//------------------------------------------------------------------------------------
//                                                                            
//                   I N T E L   P R O P R I E T A R Y                   
//                                                                       
//      COPYRIGHT (c)  2001 BY  INTEL  CORPORATION.  ALL RIGHTS          
//      RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//      BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//      RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//      LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//      MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//      THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                       
//                         INTEL  CORPORATION                            
//                                                                      
//                      2200 MISSION COLLEGE BLVD                        
//                                                                       
//                SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                       
//------------------------------------------------------------------------------------
/////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: 	dl_source.uc
//
//      Purpose: 	Contains macros for dl_sink and dl_source for Packet TX
//                  and IPv4 Blocks.
//
//      History:
//
//
//      Date            Comment                         By
//      ---------------------------------------------------------------------
//
//      11/19/2001      Created                         Senthil Nathan
//
//
/////////////////////////////////////////////////////////////////////////////

#ifndef __DL_SOURCE_UC__
#define __DL_SOURCE_UC__

#include <stdmac.uc>
#include <sig_macros.uc>
#include <dispatch_loop.h>


/////////////////////////////
//	Some local defines
/////////////////////////////

#if (DL_RING_NUMBER > 11)
#error			"For Ring Number Greater than 11, we cannot use the SCR_Ring#_Full input state. Check RING_NUMBER"
#else
#define_eval	DL_RING_FULL			SCR_Ring/**/DL_RING_NUMBER/**/_Full	
#endif


/////////////////////////////
// Global variables/Registers
/////////////////////////////


.reg 	global	dl_ring					// scratch ring number for pos rx ==>ipv4. This can be made an absolut register
.reg 	global	dl_buf_handle			// The current Buffer handle.
.reg	global	dl_eop_buf_handle		// For large packets, this is the last buffer in the chain
.reg	global	dl_next_block			// Next block that should process the buffer/packet
.reg    global  dl_exception_reg
.set            dl_exception_reg
///////////////////////////////////////////////////////////////////////////////
// _dl_pos_sink_init:
//	 	Description: 
//			Initialisation for POS dl_sink. Should be called before using dl_sink.
//			It simply sets a global variable dl_ring to the corresponding ring number
//			This will later save one instruction in dl_sink (for POS).
//
//			[Note: This is an internal macro not meant to be used by the user.
//			This macro will undergo some more changes.]
//
//	 	Outputs:
//			
//		Inputs:
//		
//		Size:   				:
//			1 instruction
//

#macro _dl_pos_sink_init[]
.begin

	alu_shf[dl_ring, --, B, DL_RING_NUMBER, <<2]	; ring number to be used in get/put 

.end
#endm

/////////////////////////////////////////////////////////////////////////////
// dl_qm_sink
//	
//	Description:
//		Send received packet to the qm block by writing scratch ring
//
//	Outputs:	
//		None
//		
//	Inputs:
//		s_hl:		SOP buffer handle register, this register value will be changed
//					after macroc is excuted.  If the caller want to keep this value  
//					the value should be stored some where 
//		e_e_hl:		Encoded EOP buffer handle: NULL for one buffer cframe
//					otherwise will be real EOP handle  
//		cls_n:		Class number
//		o_n:		Output port number
//		p_size:		packet size
//		flow_color	flow and color information
//		rx_sig:		Signal used to wait receving logid wake up  
//		scr_sig:	Scratch ring write signal
//		rt_lbl:		return lable after waked up by receiving logic
//		qm_ring_num:QM receiving ring number
//	CONSTANTS:	
//		None
//	
//
//	Size: 8 instructions
/////////////////////////////////////////////////////////////////////////////
#macro dl_qm_sink[s_hl, e_e_hl, cls_n, o_n, p_size, flow_color_id, rx_sig, scr_sig,\
					 rt_lbl, qm_ring_num]
.begin
/////////////////////////////////////////////////////////////////////////////
// Local register declaration

.reg tmp_0													; temporary gpr 0

	xbuf_alloc[$sink, 3, write]								; alloc $sink
	alu[$sink[0], --, B, s_hl]								; set SOP hndl to $sink0
	alu[$sink[1], --, B, e_e_hl]								; set EOP hndl to $sink1
#ifndef ATM_TX
	alu[tmp_0, cls_n, or, o_n, <<4 ]						; set output port/class num
#else
	alu[tmp_0, --, B, cls_n ]								; Do not set oput port num
#endif
	alu[$sink[2], tmp_0, or,1, <<31 ]							; set MSB in $sink1
	scratch[put, $sink[0], 0, qm_ring_num, 3],\
			sig_done[scr_sig]								; put to SCR ring
	ctx_arb[rx_sig, scr_sig],and,br[rt_lbl]					; ctx swap
	xbuf_free[$sink]										; free $sink
.end
#endm


///////////////////////////////////////////////////////////////////////
// Macro Name  : dl_get_egress_exception_pkt
// Description : Macro to write the two longwords of exception
//				 and trigger interrupt to activate RM ISR for 
//				 exception packets. In the current implementation
//               all packets go to queue manager and it is assumed 
//               that all packet processing has been completed
//               at the core.
// Input       : in_buffer_handle, in_exception_code
// Constant    : nil
// Size        : 4
// Branches    : 0
////////////////////////////////////////////////////////////////////////
#macro dl_get_egress_exception_pkt()


.begin
   .reg ring, temp
   .reg queue_number
   .reg $excep_ctrl_lw0 $excep_ctrl_lw1 $excep_ctrl_lw2
   .xfer_order $excep_ctrl_lw0 $excep_ctrl_lw1 $excep_ctrl_lw2
   .reg $excep_data_lw0 $excep_data_lw1 $excep_data_lw2
   .xfer_order $excep_data_lw0 $excep_data_lw1 $excep_data_lw2
   .sig scratch_rd_done meta_read_signal         

   // set the ring number

   alu[ring, --,b, EXCEPTION_RING_IN_0, <<2]

   // read one longword from exception ring 1

   scratch[get, $excep_ctrl_lw0, ring, 0, 3], sig_done[scratch_rd_done]
   ctx_arb[scratch_rd_done]

   .if( $excep_ctrl_lw0 != 0 )

	// LW0 from ring is the current buffer handle (containign SOP)

	 move[dl_buf_handle, $excep_ctrl_lw0]

	 // allocate 6 read xfer registers for reading in the meta data 

	 xbuf_alloc[$dl_meta, 7, read]

	 //read meta data

     dl_meta_load_cache[$excep_ctrl_lw0, $dl_meta, meta_read_signal, 1, 6]
	 ctx_arb[meta_read_signal]

	 // set the eop handle 

	 move[dl_eop_buf_handle, $excep_ctrl_lw1]
       

	 //Write three long words on scratch ring

	 move[$excep_data_lw0, dl_buf_handle]
	 move[$excep_data_lw1, dl_eop_buf_handle]
     dl_meta_get_output_port[queue_number]				; LW2 on the ring is Queue number for Queue Manager
	 alu_shf[queue_number, --, B, queue_number, <<4]		; fabric * 16 is the actual queue number
	 alu_shf[$excep_data_lw2, queue_number, OR, 1, <<31]		; The msb is always set to 1, to prevent producing a value of 0 on to the ring.

     alu_shf[ring, --, b, QM_RING_IN_0, <<2]

     // check for ring full condition. if full then keep trying

check_ring#:
     br_inp_state[SCR_RING/**/ETHERNET_ARP_SINK_SCR_RING/**/_STATUS, ring_full#]
 
	 scratch[put, $excep_data_lw0, 0, ring, 3], sig_done[scratch_rd_done]
	 ctx_arb[scratch_rd_done]

	 br[end#]

     // if the ring is full go back and retry to write to scratch ring
ring_full#:
     br[check_ring#]      

end#:

    xbuf_free[$dl_meta]
  .endif


.end
read_again#:


#endm /* dl_get_egress_exception_pkt() */
 

#endif	// __DL_SOURCE_UC__
