///////////////////////////////////////////////////////////////////////////////
//
//
//                  I N T E L   P R O P R I E T A R Y
//
//     COPYRIGHT [c]  2003 BY  INTEL  CORPORATION.  ALL RIGHTS
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT
//     THE PRIOR WRITTEN PERMISSION OF :
//
//                        INTEL  CORPORATION
//
//                     2200 MISSION COLLEGE BLVD
//
//               SANTA  CLARA,  CALIFORNIA  95052-8119
//
///////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: common_rx.uc
//
//      Purpose: This microblock used in oc12 application combines and
//				 multiplex two independent receive pipes:
//                > packet_rx
//                > aal5_rx
//
//	     	 All even numbered contexts are assigned to handle packet type
//               receive processing, while all odd numbered contexts are assigned
//               to handle aal5 type receive processing
//               All received frames/packets are reassembled, the collected context
//               info like input port number, packet size etc and stores it in the
//               META data.
//
//      History:
//
//
//      Date            Comment                         By
//      ---------------------------------------------------------------------
//
//      12/02/2002      Created                         Czarek Wierucki
//
/////////////////////////////////////////////////////////////////////////////

#ifndef	__COMMON_RX_UC__
#define	__COMMON_RX_UC__


#ifndef BID_NEXT_BLOCK
#define	BID_NEXT_BLOCK					BID_EDECAP
#endif

#include <dl_system.h>

/* 	Ring number to use. 0-15. */
#define_eval DL_DROP_RING			        FWD_TO_QM_SCR_RING

/* We need this define to take only implementation of the macros (not the whole ATM RX dispatch loop */
#define TAKE_MACROS_ONLY
#include "aal5_rx.uc"

#include "packet_rx.h"
#include "packet_rx_context.h"
#include "packet_rx_one_me_util.uc"

#include "pkthdr_cache.uc"

#ifdef SIMULATION
    #include "simulation.uc"
#endif

#include "system_init.uc"
#include "common_rx_init.uc"


;-----------------------------------------------------------------------------
; Global Register declaration
;-----------------------------------------------------------------------------

//	Following regsiters are used as global variables
.reg		sink_sig_mask			// mask to tell what I/Os were done



#ifdef OLD_CODE

///////////////////////////////////////////////////////////////////////////////
//
// packet_pkt_rx
//
//	 	Description:
//
// 			Packet processing entry point for packet microblock. In each iteration
//			one mpacket will be handled. Implements a state machine to receive
//			SOP, MOP and EOP mpackets.
//
//	 	Outputs:
//
//			All global variables will be affected.
//
//		Inputs:
//
//			Will use global varaibles.
//		Size:
//
//			85 instructions for min packets.
//
///////////////////////////////////////////////////////////////////////////////

#macro packet_pkt_rx()
.begin

	.reg 	ctx_st								// keeps the machine state
	.reg 	$protocol							// protocol in PPP hdr
	.reg	protocol_reg						// protocol in PPP hdr
	.reg 	drop_sop_handle, drop_eop_handle	// SOP & EOP buffers to drop
	.reg 	port_num							// inpur port number
	.reg 	ele_num								// rbuf element number
	.reg 	ele_addr							// rbuf element address
	.reg 	lm_ctx_base							// base addr of reassembly context

	.set 	$rsw2, $rsw3						// to avoid warnings

comn_rx_phase_1#:

	//	The thread autmatically wakes up on receiving an mpacket.
	//	The receive status words are available in $rsw2 and $rsw3.
	//	If no packet had been recieved within the timeout, then
	//	the status word will indicate that a NULL packet was received.

	_packet_rx_rsw_get_port_num(port_num, $rsw2)					; From which port pkt received.

	//	locate the context in local memory and setup local memory.
	//	There are totally 16 contexts, one per port. Each conext
	//	is 64 bytes (16 long words)

	immed[lm_ctx_base, CONTEXT_DESC_BASE]
	alu_shf[lm_ctx_base, lm_ctx_base, or, port_num, <<PO2_CTXSIZE]
	local_csr_wr[active_lm_addr_0, lm_ctx_base]			; setup local memory for this port

	//	check if error in receviing

	alu[--, $rsw2, AND, ALL_ERROR_CODE, <<8] 			; 0x2E00. XXX - should it be 0x2a00?
 	bne[rx_error_processing#], defer[1]				; goto err processing
	alu[sink_sig_mask, --, B, 1, <<RX_EVENT_SIG]     		; set sink_sig_mask						;

	//	a proper mpacket was received. Depending on the State of the
	//	Receive State Machine and the kind of mpacket (SOP, EOP, MOP)
	//	received branch accordingly.

	_packet_rx_get_machine_state(ctx_st, $rsw2)					; state will be returned in ctx_st

	//	There are only two states: START and PROCessing. There are four
	//	kinds of mpackets: SOP, EOP, MOP and SOP&EOP. MOP is (!SOP and !EOP).
	//	In START state SOP or SOP&EOP mpackets are expected.
	//	In Processing state MOP or EOP mpackets are expected.

	//	ctx_st is interpreted as follows.

	//	000 START 	state, 	MOP mpacket	--	Error case	--	msp
	//	001 PROC 	state, 	MOP mpacket	--	Normalcase	--	mpp
	//	010 START 	state, 	EOP mpacket	--	Error case	--	esp
	//	011 PROC 	state, 	EOP mpacket	--	Normal case	--	epp
	//	100 START 	state, 	SOP mpacket	--	Normal case	--	ssp
	//	101 PROC 	state, 	SOP mpacket	--	Error case	-- 	spp (+ normal case)
	//	110 START 	state, 	SOP&EOP 	--	Normal case	-- 	sesp
	//	111 PROC 	state, 	SOP&EOP 	--	Error case	--	sepp (+ normal case)

	//	jump to the corresponding state processing

	jump[ctx_st, jump_table#], defer[3], targets[msp_j#, mpp_j#, esp_j#, \
												epp_j#, ssp_j#, spp_j#, sesp_j#, sepp_j#]
		_packet_rx_rsw_get_element_num(ele_num, $rsw2)			; get rbuf element number
		_packet_rx_rsw_get_byte_count(byte_count, $rsw2) 			; get number of bytes received
		alu_shf[ele_addr, --, B, ele_num, <<RBUF_ELE_SIZE_LOG2]	; ele_num * size of rbuf element

//	Start of jump table.

jump_table#:
msp_j#:
	br[msp#]    ; goto mop_start_processing		--	Error case

mpp_j#:
	br[mpp#]	; goto mop_proc_processing		--	Normal case

esp_j#:
	br[esp#]	; goto eop_start_processing		--	Error case

epp_j#:
	br[epp#]	; goto eop_proc_processing		--	Normal case

ssp_j#:
	br[ssp#]	; goto sop_start_processing		-- 	Normal case

spp_j#:
	br[spp#]	; goto sop_proc_processing		--	Error + Normal case

sesp_j#:
	br[sesp#]	; goto sop_eop_start_processing	--	Normal case

sepp_j#:
	br[sepp#]	; goto sop_eop_proc_processing	--	Error + Normal case


//	Error processing code

rx_error_processing#: 

	//	We are here because receive status word indicated an error.

	//	check for NULL (i.e timeout)

	alu[--, $rsw2, AND, 1, <<9]							;[9] indicates NULL receive
	bne[null_error_processing#], defer[1]
	_packet_rx_rsw_get_element_num(ele_num, $rsw2)		; get rbuf element number

	//	Some other error. Prepare to drop the mpacket
	//	For large packets drop_eop_handle will be the last buffer
	//	in the chain

	alu[drop_sop_handle, --, B, lm_sop_buf]
	alu[drop_eop_handle, --, B, lm_cur_buf]

#ifdef _DEBUG_COUNTERS
	alu[@pkt_rx_num_err_mpkts, @pkt_rx_num_err_mpkts, +, 1]
#endif ; _DEBUG_COUNTERS

	//	Reset State Machine

	_packet_rx_set_state(STATE_START)						; reset to start state

	//	End of Phase 1. Wait next thread signal to synchronize

	ctx_arb[sig_next_thread]

	//	Drop the mpacket.

	_packet_rx_drop_current_pkt(drop_sop_handle, drop_eop_handle)

	_packet_rx_incr_counter(port_num, PACKET_PKTS_DROPPED)

	//	Even though we dropped the packet, we'll propagate a NULL
	//	packet thro' the pipeline to keep the pipeline moving. (we do this 
	//	by setting dl_next_block to BID_NULL anddl_buf_handle to DL_BUF_NULL)

	//	Continue with Phase2 processing.

	br[comn_rx_phase_2#], defer[2]
		alu[dl_next_block, --, B, BID_NULL] 			; NULL (0xFF)
		alu[dl_buf_handle, --, B, DL_BUF_NULL]     		; NULL (0xFF)

null_error_processing#: 

	//	Null mpkt processing. (i.e receive logic timed out for this thread)
	
	//	At this point, the receiving logic has not received a mpkt for certain
	//	amount of time. In order to keep the pipeline moving, the receiving logic
	//	removes the thread from freelist and wakes up the thread by sending the
	//	thread a (rx_event) signal.  

	//	Nothing to do. Just synchronize with rest of the threads
	//	and go to Phase2.

	ctx_arb[sig_next_thread]

	_packet_rx_signal_next_thread(nt_sig_reg)				; signal next thread

	br[null_phase_2_processing#],defer[2]
		alu[dl_next_block, --, B, BID_NULL] 			; next_block is NULL (0xFF)
		alu[dl_buf_handle, --, B, DL_BUF_NULL]     		; dl_buffer_handle is NULL (0xFF)


msp#:	
	//	Got MOP mpacket while expecting SOP packet - Error case 

esp#:
	//	Got EOP mpacket while expecting SOP packet - Error case

	//	Both the above error cases are handled here.
	//	Being in START state, we have absolutely nothing to do here.
	//	Just synchronize with rest of the threads and go to Phase2.

	ctx_arb[sig_next_thread]

	br[comn_rx_phase_2#],defer[2]
		alu[dl_next_block, --, B, BID_NULL]			 	; next_block is NULL (0xFF)
		alu[dl_buf_handle, --, B, DL_BUF_NULL]     		; dl_buffer_handle is NULL (0xFF)


mpp#:
	//	Got MOP mpacket while expecting a MOP/EOP mpacket: Normal case

	_packet_rx_mpp_processing()

	br[comn_rx_phase_2#]
				
sepp#:
	//	Got a SOP&EOP (both in same mpacket) while expecting MOP/EOP mpacket - Error

	
	//	We do two things. 
	//	[1] We discard the partially (so far) assembled packet.
	//	[2] And since SOP&EOP is a new but complete packet we can process it.
		
	//	[1]
	//	Reset State Machine

	_packet_rx_set_state(STATE_START)								; reset to start state
	 
	//	Prepare to drop the mpacket For large packets 
	//	drop_eop_handle will be the last buffer in the chain
	//	XXX - Why can't drop packet done befor sesp_processing?

	alu[drop_sop_handle, --, B, lm_sop_buf]
	alu[drop_eop_handle, --, B, lm_cur_buf]

	//	[2]	Process SOP&EOP mpacket

	_packet_rx_sesp_processing()

	//	drop the partially assembled pkt

	_packet_rx_drop_current_pkt(drop_sop_handle, drop_eop_handle)	

	_packet_rx_incr_counter(port_num, PACKET_PKTS_DROPPED)

	br[comn_rx_phase_2#]

epp#:
	//	Got EOP mapacket while expecting MOP/EOP - Normal case

	_packet_rx_epp_processing()

	br[comn_rx_phase_2#]

spp#:
	//	Got SOP mpacket while expecting MOP/EOP -  Error case

	//	We do two things. 
	//	[1] We discard the partially (so far) assembled packet.
	//	[2] And since SOP is a new packet we can process it.

	//	Reset State Machine. 

	_packet_rx_set_state(STATE_START)								; reset to start state

	//	Prepare to drop the packet. For large packets 
	//	drop_eop_handle will be the last buffer in the chain
	//	XXX - Why can't drop packet done befor sesp_processing? .

	alu[drop_sop_handle, --, B, lm_sop_buf]
	alu[drop_eop_handle, --, B, lm_cur_buf]

	//	Do the SOP processing for the newly received mpacket

	_packet_rx_ssp_processing()

	//	drop the partially assembled pkt

	_packet_rx_drop_current_pkt(drop_sop_handle, drop_eop_handle)	

	_packet_rx_incr_counter(port_num, PACKET_PKTS_DROPPED)

	br[comn_rx_phase_2#]

ssp#:
	//	Got SOP mpacket while expecting a SOP mpacket -  normal case

	//	Do the SOP processing for the newly received mpacket

	_packet_rx_ssp_processing()

	br[comn_rx_phase_2#]

sesp#:
	//	Got SOP mpacket while expecting a SOP mpacket -  normal case

	//	This is the critical path, where min packets are handled. 

	_packet_rx_sesp_processing()
		
comn_rx_phase_2#:

	//	Phase 2

	_packet_rx_signal_next_thread(nt_sig_reg)				; Signal next thread
	_packet_rx_free_rbuf_element(ele_num)					; Free rbuf element.

	// for NULL receive, no rbuf needs to be freed 

null_phase_2_processing#:  					

	// add current thread to freelist 0
	msf[fast_wr, --, rx_freelist, freelist_reg]			; add thread to freelist

.end
#endm


///////////////////////////////////////////////////////////////////////////////
//
// _packet_rx_sesp_processing
//
//	 	Description: 
//
//			Process a mpacket containing both SOP&EOP. i.e packets size
//			ranging from 42(min) - 128 are handled here. Min packet case 
//			is the critical path, so it will be optimised wherever possible.
//
//	 	Outputs:
//
//		Inputs:
//
//		Size: 
//
//			~43 instructions for min packet. 
//
///////////////////////////////////////////////////////////////////////////////

#macro _packet_rx_sesp_processing()

.begin

	.reg	buf_offset
	.reg	dram_handle
	.reg	header_type
	.sig	sig_dram
	.sig	sig_meta
	.sig	sig_msf_rd
	.sig	sig_buf_alloc
	.sig	sig_cntr

	#ifdef IXP2800
	; Flush meta-data belonging to channel 0 and BUF_SRAM_CHANNEL using
	; separate macro calls.
	.sig	sig_meta_ch3
	#endif
	
	//	SOP and EOP are in the same mpacket. This is the critical path
	//	i.e min packets are handled here. 

	//	check if prefetch buffer handle is zero

	br=byte[$prefetch_buf_handle, 3, 0, sesp_prf_buf_zero_proc#]

	alu[header_type, lm_header_type, and, 0xFF]		; get port header type

	//	We got a valid buffer here.

	//	Get a random offset (headroom) for this buffer. 

	_packet_rx_get_rand_buf_offset(buf_offset)

	//	prefetch_data_handle is where the buffer begins.
	//	+ buf_offset will specify where we should store data 
	//	(including PPP hdr)

	alu[dram_handle, buf_offset, +, prefetch_data_handle]

	//	move received mpkt to data buffer in DRAM

	//	_packet_rx_move_rbuf_to_pkt_buf(dram_handle, ele_addr, byte_count, rbuf_base, sig_dram)
	#ifdef	DISABLE_MAC_FILTERING

		.reg	tmp_addr, tmp_count, tmp_handle 

		alu[tmp_addr, ele_addr, +, 8]
		alu[tmp_count, byte_count, -, 8]
		alu[tmp_handle, dram_handle, +, 8]
		_packet_rx_move_rbuf_to_pkt_buf(tmp_handle, tmp_addr, tmp_count, rbuf_base, sig_dram)

	#else 

		_packet_rx_move_rbuf_to_pkt_buf(dram_handle, ele_addr, byte_count, rbuf_base, sig_dram)

	#endif // DISABLE_MAC_FILTERING

	//	Set all the required META Data. This will happen
	//	in various defer slots.
    
	alu[dl_buf_handle, --, B, $prefetch_buf_handle] 	; set dl_buf_handle 
	
	//	Find out if the received mpacket can be contained in 1 cell 
	//	(on the transmit side). Remember that for every packet there is
	//	a header ( 8 bytes) that gets prepended. (This is apart from per 
	//	cell header (8 bytes) that gets added) This Per pkt header is added
	//	on the fly. Hence it is not counted in buffer/packet size.

#define	FCELL_SIZE_WO_PPHDR		(CELL_SIZE - PER_PKT_PREPEND_BYTES)
	alu[--, byte_count, -, FCELL_SIZE_WO_PPHDR]			; > cell size?

	//	Remember if the packet can be contained in one cell, 
	//	then the cell count is 0. If it can be contained in 
	//	two cells, then cell count is 1 etc. 
	//	(less 1 than the actual) 
	ble[sesp_cell_count_compute_done#], defer[1]		; yes, 1 cell is enough.
		dl_meta_set_buffer_next(IX_NULL)				; There is no next buffer

	//	The packet cannot be contained in one cell. Calculate the number
	//	of cells in the packet. The cell count calculation is generic 
	//	with no hard coded values but is expensive. Watch out if
	//	you run out of budget.

	.reg	cell_count
	.reg	bytes_remaining
	.reg	temp_byte_count

	alu[temp_byte_count, byte_count, +, PER_PKT_PREPEND_BYTES] 	; cell includes per pkt hdr
	_packet_rx_set_first_cell_count(cell_count, bytes_remaining, temp_byte_count)
	alu[dl_buf_handle, dl_buf_handle, OR, cell_count, <<24]		;

sesp_cell_count_compute_done#:

	// 	Set the various meta data fields for consumption by rest of the
	//	pipeline

	// 	We branch here so that we can take advantage of the defer slots.
	//	This branch with defer, will finally save a branch to the end of this macro 
	//	(where 2 cycles may be wasted)

	br[continue_with_sesp#], defer[2]
		dl_meta_set_offset(buf_offset)					; set buffer offset
		dl_meta_set_buffer_size(byte_count)				; set buffer size (= byte count)

	// 	program flow continues in continue_with_sesp#:

	//	Error Handling.

sesp_prf_buf_zero_proc#:

	_packet_rx_incr_counter(port_num, PACKET_PKTS_DROPPED)

	//	We got a null buffer. So nothing to do. We still need to keep the pipeline
	//	moving. Hence set the next block to NULL (and dl_buf_handle to NULL)
	//	(Try to) Get a new prefetch buffer so that it can used next time around.

	dl_buf_alloc($prefetch_buf_handle, BUF_FREE_LIST0, sig_buf_alloc, SIG_NONE)

	ctx_arb[sig_next_thread, sig_buf_alloc], defer[2]
		alu[dl_next_block, --, B, BID_NULL]		 		; set next_block NULL
		alu[dl_buf_handle, --, B, DL_BUF_NULL]	   		; set dl_buf_handle NULL
 
	//	Get DRAM address of the new buffer 

	dl_buf_get_data(prefetch_data_handle, $prefetch_buf_handle)

	br[end#]

continue_with_sesp#:

	// 	Set the various meta data fields for consumption by rest of the
	//	pipeline (Cont...)
	dl_meta_set_free_list(FREE_LIST_ID)					; set free list. 
	dl_meta_set_rx_stat(0)								; set rx stat. Zero for now
	dl_meta_set_input_port(port_num)					; set input port number
	dl_meta_set_packet_size(byte_count)					; set packet size
	dl_meta_set_header_type(header_type)				; set header type for the port

	//	Flush the meta data to SRAM. Only the first four Long Words 
	//	(0,1,2,3) of the meta. The rest of the meta data are not touched 
	//	by packet RX.

	xbuf_alloc($sesp_meta,4, write)						; no code generated.
	xbuf_alloc($bytes_counter, 1, write)

	#ifdef IXP2800

		; Flush meta-data belonging to channel 0 and BUF_SRAM_CHANNEL using
		; separate macro calls.
		dl_meta_flush_cache($sesp_meta, dl_buf_handle, sig_meta, SIG_NONE, 0, 1)
		dl_meta_flush_cache($sesp_meta, dl_buf_handle, sig_meta_ch3, SIG_NONE, 1, 3)

	#else

		dl_meta_flush_cache($sesp_meta, dl_buf_handle, sig_meta, SIG_NONE, 0, 4)

	#endif	// IXP2800

	//	Prefetch a new buffer so that it can be used next time around.

	dl_buf_alloc($prefetch_buf_handle, BUF_FREE_LIST0, sig_buf_alloc, SIG_NONE)

	//	increment packets received and bytes received counter

	_packet_rx_incr_counter(port_num, PACKET_PKTS_RECEIVED)			; pkts rcvd++
	_packet_rx_bytes_add_counter(port_num, byte_count, $bytes_counter, PACKET_BYTES_RECEIVED, sig_cntr)	

	#ifdef	RFC_PPP_COUNTERS                                      

		//	When counters are enabled, wait for an additional signal sig_cntr

		#ifdef IXP2800
			; Flush meta-data belonging to channel 0 and BUF_SRAM_CHANNEL using
			; separate macro calls.

			ctx_arb[sig_dram, sig_next_thread, sig_meta, sig_meta_ch3, \
					sig_buf_alloc, sig_cntr], defer[1]
				alu[dl_eop_buf_handle, --, B, DL_BUF_NULL] 		; packet is in a single buffer

		#else

			ctx_arb[sig_dram, sig_next_thread, sig_meta, sig_buf_alloc, sig_cntr], defer[1]
				alu[dl_eop_buf_handle, --, B, DL_BUF_NULL] 		; packet is in a single buffer

		#endif	// IXP2800

	#else

		#ifdef IXP2800

			; Flush meta-data belonging to channel 0 and BUF_SRAM_CHANNEL using
			; separate macro calls.

			ctx_arb[sig_dram, sig_next_thread, sig_meta, sig_meta_ch3, sig_buf_alloc], defer[1]
				alu[dl_eop_buf_handle, --, B, DL_BUF_NULL] 		; packet is in a single buffer

		#else

			ctx_arb[sig_dram, sig_next_thread, sig_meta, sig_buf_alloc], defer[1]
				alu[dl_eop_buf_handle, --, B, DL_BUF_NULL] 		; packet is in a single buffer

		#endif	// IXP2800

	#endif

	xbuf_free($sesp_meta) 								; no code generated
	xbuf_free($bytes_counter)

	alu[dl_next_block, --, B, BID_NEXT_BLOCK]			; set pkt destination to L2 block
	dl_meta_set_packet_size(byte_count)	   				; set packet size (= byte count)
	dl_meta_set_input_port(port_num)       				; set input port number


sesp_set_next_block_end#:

	//	Get DRAM Address of the prefetch buffer 
	//	for use in next packet.

	dl_buf_get_data(prefetch_data_handle, $prefetch_buf_handle)

end#:

#undef	FCELL_SIZE_WO_PPHDR

.end
#endm


///////////////////////////////////////////////////////////////////////////////
//
// _packet_rx_ssp_processing
//
//	 	Description: 
//
//			Process a mpacket containing only SOP. Even though we have
//			3 times the budget (3 * 88) as min (42 byte) packets
//			we should be careful with 129 byte packet, where the budget is 
//			reduced to 1.5 * 88 (because we now have to handle two mpackets
//			in 3 * 88 budget. The second mpacket with EOP will be handled in
//			epp_processing). 
//
//			All cycle times are for IXP2400 @ 600 MHZ
//
//	 	Outputs:
//
//		Inputs:
//
//		Size: 
//
//			~50 instructions
//
///////////////////////////////////////////////////////////////////////////////

#macro _packet_rx_ssp_processing()

.begin

	.reg 	temp_byte_count
	.reg	buf_offset
	.reg	header_type
	.sig	sig_dram
	.sig	sig_msf_rd
	.sig	sig_buf_alloc
	
	//	check if prefetch buffer handle is zero

	alu[--, --, B, $prefetch_buf_handle, <<8]			; [24:0] contains buf handle
	beq[ssp_prf_buf_zero_proc#]							; goto error handling
	alu[header_type, lm_header_type,and,0xFF]			; get header type

	//	This is the first SOP mpacket within the packet

	_packet_rx_get_rand_buf_offset(buf_offset)						; get a random offset for sop

	//	prefetch_data_handle is where the buffer begins.
	//	+ buf_offset will specify where we should store data 
	//	(including PPP hdr)

	alu[lm_cur_data, buf_offset, +, prefetch_data_handle]	; add offset to buf begining

	//	move received mpkt to DRAM

	_packet_rx_move_rbuf_to_pkt_buf(lm_cur_data, ele_addr, byte_count, rbuf_base, sig_dram)

	//	Update various reassembly context data.

	alu[lm_cur_buf, --, B, $prefetch_buf_handle]		; prefetch to current 
	alu[lm_sop_buf, --, B, $prefetch_buf_handle] 		; also make it sop buffer
	alu[lm_sop_buf_offset, --, B, buf_offset]			; remember this offset as
	alu[lm_buf_offset, --, B, buf_offset]				; sop buf and cur buf offset 
	alu[lm_buf_size, --, B, byte_count]					; buf size excludes hdr lenth 
	alu[lm_pkt_size, --, B, byte_count]					; pkt size excludes hdr length

	#ifdef IXP2800

		; In IXP2800 reference designs, we write 3 LW's on the outgoing scratch ring
		; instead of 5 LW's. The additional 2 LW's contain sop buffer meta-data. We
		; need to write this meta-data to SRAM now.

		xbuf_alloc($sop_meta, 4, write)

		.sig sig_sop_meta

		; Flush 2 LW's of SOP buffer meta-data to SRAM.
		dl_meta_set_header_type(header_type)				; set header type for the port
		dl_meta_set_rx_stat(0)								; set rx stat. Zero for now
		dl_meta_set_free_list(FREE_LIST_ID)					; set free list. 
		dl_meta_set_input_port(port_num)        			; set the input port
	
		dl_meta_flush_cache($sop_meta, $prefetch_buf_handle, sig_sop_meta, SIG_NONE, 2, 2)

	#endif	// IXP2800

	//	Cell Count Calculation.
	//	On the transmit side, for every packet we add a prepend hdr
	//	(apart from a cell hdr for every cell). Include this hdr in
	//	cell count calculation. (For a given packet, we'll hit SOP
	//	condition only once so this place fits well for this 
	//	work)

	alu[temp_byte_count, byte_count, +, PER_PKT_PREPEND_BYTES]
	_packet_rx_set_first_cell_count(lm_cell_count, lm_bytes_remaining, temp_byte_count)

	//	Advance the current data pointer (where next mpacket will
	//	be stored)

	alu[lm_cur_data, lm_cur_data, +, byte_count]		; move the cur data ptr

	//	Prefetch a new buffer so that it can be used next time around.

	dl_buf_alloc($prefetch_buf_handle, BUF_FREE_LIST0, sig_buf_alloc, SIG_NONE)


	//	Move the Receive Machine to PROCessing State

	_packet_rx_set_state(STATE_PROC)

	//	End of Phase 1
	//	Wait on the all the previous I/O plus the next thread signal.
	//	Remember that we have received only a part of the packet so far.
	//	Until we receive EOP, dl_next_block and dl_buf_handle should 
	//	be null.

	//	End of Phase 1

	#ifdef IXP2800

	    ctx_arb[sig_sop_meta, sig_dram, sig_next_thread,sig_buf_alloc], defer[2]
			alu[dl_next_block, --, B, BID_NULL]		 		; set next_block NULL
			alu[dl_buf_handle, --, B, DL_BUF_NULL]	   		; set dl_buffer_handle NULL
		xbuf_free($sop_meta)

	#else

	    ctx_arb[sig_dram, sig_next_thread,sig_buf_alloc], defer[2]
			alu[dl_next_block, --, B, BID_NULL]		 		; set next_block NULL
			alu[dl_buf_handle, --, B, DL_BUF_NULL]	   		; set dl_buffer_handle NULL

	#endif	// IXP2800

	//	Phase 2

	//	Get DRAM Address of the prefetch buffer for use in next packet.

	dl_buf_get_data(prefetch_data_handle, $prefetch_buf_handle)

	br[ssp_end#]

	//	Null Buffer error condition

ssp_prf_buf_zero_proc#:

	_packet_rx_incr_counter(port_num, PACKET_PKTS_DROPPED)

	//	We got a null buffer. So nothing to do. We still need to keep the pipeline
	//	moving. Hence set the next block to NULL (and dl_buf_handle to NULL)
	//	(Try to) Get a new prefetch buffer so that it can used next time around.

	dl_buf_alloc($prefetch_buf_handle, BUF_FREE_LIST0, sig_buf_alloc, SIG_NONE)

	ctx_arb[sig_next_thread, sig_buf_alloc], defer[2]
		alu[dl_next_block, --, B, BID_NULL]		 		; set next_block NULL
		alu[dl_buf_handle, --, B, DL_BUF_NULL]	   		; set dl_buffer_handle NULL
 
	//	Get DRAM Address of the prefetch buffer for use in next packet.
		
	dl_buf_get_data(prefetch_data_handle, $prefetch_buf_handle)

ssp_end#:

.end
#endm

///////////////////////////////////////////////////////////////////////////////
//
// _packet_rx_epp_processing
//
//	 	Description: 
//
//			Process a mpacket containing only EOP. The path 129 byte packet 
//			takes here should (i.e 1 byte mpacket with EOP will be received)
//			be completed in 1.5 * 88 cycles (for IXP2400 @ 600 MHZ)
//
//	 	Outputs:
//
//		Inputs:
//
//		Size: 
//
//			?? instructions
//
///////////////////////////////////////////////////////////////////////////////

#macro _packet_rx_epp_processing()

.begin

	.reg	cur_buf
	.reg	prev_buf
	.reg 	cell_count
	.reg	buf_offset
	.reg	header_type
	.sig	sig_prev_meta
	.sig	sig_cur_meta
	.sig	sig_dram
	.sig	sig_cntr

	#ifdef IXP2800
		; Flush meta-data belonging to channel 0 and BUF_SRAM_CHANNEL using
		; separate macro calls.
		.sig	sig_prev_meta_ch3 sig_cur_meta_ch3
	#endif

	//	Got EOP mpacket while expecting MOP/EOP - Normal case
	//	Receive State Machine is in PROCessing State.

	alu[header_type, lm_header_type,and,0xFF]
		
	//	move received mpkt to DRAM
	
	_packet_rx_move_rbuf_to_pkt_buf(lm_cur_data, ele_addr, byte_count, rbuf_base, sig_dram)

 	//	increment cell count depending on the number of bytes received.
	//	embed this cell count in the buffer handle.

	_packet_rx_incr_cell_count(lm_cell_count, lm_bytes_remaining, byte_count)

	alu[cell_count, --, B, lm_cell_count]
	alu[cur_buf, --, B, lm_cur_buf] 					; get curr buffer handle
	alu[cur_buf, cur_buf, OR, cell_count, <<24] 		; embed cell count
														; bydefault EOP bit is set.
	//	update buffer size, packet size and reset
	//	state machine to START state.
	//	[ See defer slots]

	//	Check if current buffer is a sop buffer

	br_bclr[cur_buf, 30, epp_not_sop_buffer#], defer[3]	; [30] is SOP bit.
		alu[lm_buf_size, lm_buf_size, +, byte_count]	; update buffer size
		alu[lm_pkt_size, lm_pkt_size, +, byte_count]	; update packet size
		_packet_rx_set_state(STATE_START)					; Reset State Machine

epp_sop_buffer#:

	//	increment packets received and bytes received counter

	xbuf_alloc($bytes_counter, 1, write)

	_packet_rx_incr_counter(port_num, PACKET_PKTS_RECEIVED)			; pkts rcvd++
	_packet_rx_bytes_add_counter(port_num, lm_pkt_size, $bytes_counter, PACKET_BYTES_RECEIVED, sig_cntr)	


	//	Yes, this buffer has SOP. i.e this buffer now contains
	//	a complete packet (SOP ...EOP). So we need to handle only
	//	a single buffer. Set all META data fields and write the 
	//	meta data to SRAM.
		
	//	buffer offset is where data begins in the buffer.
	// 	For subsequent blocks buffer offset should point to 
	//	where IP header (or rather whatever follows PPP header)
	//	begins.
	alu[buf_offset, --, B, lm_sop_buf_offset]			; get buffer offset 
	dl_meta_set_offset(buf_offset)
	dl_meta_set_buffer_next(IX_NULL)                	; there is no next buffer
	dl_meta_set_buffer_size(lm_buf_size)  				; set buffer size
	dl_meta_set_packet_size(lm_pkt_size)				; set packet size (= buf size)
	dl_meta_set_input_port(port_num)        			; set the input port
	dl_meta_set_header_type(header_type)				; set header type for the port
	dl_meta_set_free_list(FREE_LIST_ID)					; set free list. 
	dl_meta_set_rx_stat(0)								; set rx stat. Zero for now

	//	Write the meta data to SRAM.

	xbuf_alloc($cur_meta,4 , write) 					; alloc xfer regs for flush 

	#ifdef IXP2800

		; Flush meta-data belonging to channel 0 and BUF_SRAM_CHANNEL using
		; separate macro calls.
		dl_meta_flush_cache($cur_meta, cur_buf, sig_cur_meta, SIG_NONE, 0, 1)
		dl_meta_flush_cache($cur_meta, cur_buf, sig_cur_meta_ch3, SIG_NONE, 1, 3)

	#else

		dl_meta_flush_cache($cur_meta, cur_buf, sig_cur_meta, SIG_NONE, 0, 4)

	#endif	// IXP2800

	//	We are done with this packet. Prepare to send it to
	//	Next block in the pipeline. dl_buf_handle should contain 
	//	the SOP buffer handle, dl_eop_buf_handle is not applicable
	//	because the entire packet is in a single buffer and dl_next_block
	//	is IPv4 block (based on the protocol info in PPP hdr)

	//	dl_next_block can be set only in Phase2 here because of a possible race
	//	condition. see [end#]

	//	End of Phase 1.
	//	Wait on the all the previous I/O plus the next thread signal.

	#ifdef	RFC_PPP_COUNTERS                                      

		//	When counters are enabled, wait for an additional signal sig_cntr

		#ifdef IXP2800
	
			; Flush meta-data belonging to channel 0 and BUF_SRAM_CHANNEL using
			; separate macro calls.
			ctx_arb[sig_dram, sig_next_thread, sig_cur_meta, sig_cur_meta_ch3, sig_cntr], defer[2], br[end#]
				alu[dl_eop_buf_handle, --, B, DL_BUF_NULL] 		; no buffer chain
				alu[dl_buf_handle, --, B, cur_buf] 				; 

		#else

			ctx_arb[sig_dram, sig_next_thread, sig_cur_meta, sig_cntr], defer[2], br[end#]
				alu[dl_eop_buf_handle, --, B, DL_BUF_NULL] 		; no buffer chain
				alu[dl_buf_handle, --, B, cur_buf] 				; 

		#endif	// IXP2800

	#else

		#ifdef IXP2800
			; Flush meta-data belonging to channel 0 and BUF_SRAM_CHANNEL using
			; separate macro calls.
			ctx_arb[sig_dram, sig_next_thread, sig_cur_meta, \
				sig_cur_meta_ch3], defer[2], br[end#]
				alu[dl_eop_buf_handle, --, B, DL_BUF_NULL] 		; no buffer chain
				alu[dl_buf_handle, --, B, cur_buf] 				; 

		#else

			ctx_arb[sig_dram, sig_next_thread, sig_cur_meta], defer[2], br[end#]
			alu[dl_eop_buf_handle, --, B, DL_BUF_NULL] 		; no buffer chain
			alu[dl_buf_handle, --, B, cur_buf] 				; 

		#endif	// IXP2800

	#endif	//RFC_PPP_COUNTERS

	xbuf_free($cur_meta) 								; free sram tranfer regs
	xbuf_free($bytes_counter)	

epp_not_sop_buffer#:

	//	At this point, EOP and SOP are in different buffers i.e large packet.
	//  So we need to write previous buffer's meta data & current buffer's (EOP) 
	//	meta data to SRAM.
	//	There are two cases to handle.

	//	[1] Previous Buffer is also SOP Buffer.
	//	[2] Previous Buffer is not SOP Buffer.

	//  [Recap] cur_buf contains current buffer handle. cell count
	//	had already been updated in the buffer handle.

	//	compose meta data for the previous buffer. (defer slots)

	//	Check if the previous buffer is a SOP buffer. If it is not a SOP buffer,
	//	the offset is set as NON_SOP_OFFSET(0) (in the defer slot) i.e no headroom
	//	If it is a SOP buffer we have stored the offset in lm. Use that.

	alu[prev_buf, --, B, lm_prev_buf] 

	br_bclr[prev_buf, 30, epp_prev_buffer_not_sop#], defer[3]	; [30] is SOP bit
		dl_meta_set_buffer_next(cur_buf)    			; cur buf is next for prev
		dl_meta_set_buffer_size(lm_prev_buf_size)  		; set buffer size
		dl_meta_set_offset(NON_SOP_OFFSET)				; saves an extra branch

	//	Previous buffer is a SOP buffer. Buffer offset should point to where 
	//	IP header (or rather whatever follows PPP header) begins.
	alu[buf_offset,--, B, lm_sop_buf_offset]			; strip header len
	dl_meta_set_offset(buf_offset)						; set offset
	dl_meta_set_header_type(header_type)				; set header type for the port

epp_prev_buffer_not_sop#:

	xbuf_alloc($prev_meta1, 2, write) 					; allocate xfer regs
	xbuf_alloc($cur_meta1, 2, write) 					; allocate xfer regs
	xbuf_alloc($bytes_counter2, 1, write)

	//	increment packets received and bytes received counter

	_packet_rx_incr_counter(port_num, PACKET_PKTS_RECEIVED)			; pkts rcvd++
	_packet_rx_bytes_add_counter(port_num, lm_pkt_size, $bytes_counter2, PACKET_BYTES_RECEIVED, sig_cntr)	

	//	Write META data of previous buffer to SRAM

	#ifdef IXP2800

		; Flush meta-data belonging to channel 0 and BUF_SRAM_CHANNEL using
		; separate macro calls.
		dl_meta_flush_cache($prev_meta1, prev_buf, sig_prev_meta, SIG_NONE, 0, 1)
		dl_meta_flush_cache($prev_meta1, prev_buf, sig_prev_meta_ch3, SIG_NONE, 1, 1)

	#else

		dl_meta_flush_cache($prev_meta1, prev_buf, sig_prev_meta, SIG_NONE, 0, 2)

	#endif	// IXP2800

	//	Set current buffer meta data

	dl_meta_set_buffer_next(IX_NULL)  					; this is the last buffer
	dl_meta_set_offset(NON_SOP_OFFSET) 					; this buffer !SOP buffer
	dl_meta_set_buffer_size(lm_buf_size)    			; set buffer size
	
	//	Write META data of current buffer to SRAM

	#ifdef IXP2800

		; Flush meta-data belonging to channel 0 and BUF_SRAM_CHANNEL using
		; separate macro calls.
		dl_meta_flush_cache($cur_meta1, cur_buf, sig_cur_meta, SIG_NONE, 0, 1)
		dl_meta_flush_cache($cur_meta1, cur_buf, sig_cur_meta_ch3, SIG_NONE, 1, 1)

	#else

		dl_meta_flush_cache($cur_meta1, cur_buf, sig_cur_meta, SIG_NONE, 0, 2)

	#endif	// IXP2800

	//	End of Phase 1.

	//	The way packets are sent to next block in the pipeline is through
	//	dl_sink. (see the outer loop in pos_rx.uc). 
	//	As part of optimisation, two LWs ( 1 & 2) from meta data are sent 
	//	in the scratch ring. dl_sink takes these meta data from the 
	//	previous dl_set_* calls. These meta data should belong to SOP buffer.

	//	In our case the prev dl_set_* calls belong to the EOP buffer. So we
	//	need to do a few dl_set_* calls for SOP buffer here.
	alu[buf_offset, --, B, lm_sop_buf_offset]			; get buff offset
	dl_meta_set_offset(buf_offset)						; set offset
	dl_meta_set_buffer_size(lm_sop_buf_size)   			; set sop buffer size
	dl_meta_set_input_port(port_num)         			; set input port number
	dl_meta_set_packet_size(lm_pkt_size)				; set packet size
	dl_meta_set_header_type(header_type)				; set header type
	dl_meta_set_free_list(FREE_LIST_ID)					; set free list. 
	dl_meta_set_rx_stat(0)								; set rx stat. Zero for now

	//	Wait on the all the previous I/O plus the next thread signal.

	//	We are done with this packet. Prepare to send it to
	//	next block in the pipeline. dl_buf_handle should contain 
	//	the SOP buffer handle, dl_eop_buf_handle should contain EOP buffer
	//	and dl_next_block is (usually) IPv4 block (based on the protocol 
	//	info in PPP hdr)
 
	//	dl_next_block can be set only in Phase2 here because of a possible race
	//	condition. see [end#]

	#ifdef	RFC_PPP_COUNTERS                                      

		//	When counters are enabled, wait for an additional signal sig_cntr

		#ifdef IXP2800
	
			; Flush meta-data belonging to channel 0 and BUF_SRAM_CHANNEL using
			; separate macro calls.
			ctx_arb[sig_dram, sig_next_thread, sig_prev_meta, sig_cur_meta, \
					sig_prev_meta_ch3, sig_cur_meta_ch3, sig_cntr], defer[2]
				alu[dl_buf_handle, --, B, lm_sop_buf] 			; first buffer
				alu[dl_eop_buf_handle, --, B, cur_buf]			; last buffer in buffer chain.

		#else

			ctx_arb[sig_dram, sig_next_thread, sig_prev_meta, sig_cur_meta, sig_cntr], defer[2]
				alu[dl_buf_handle, --, B, lm_sop_buf] 			; first buffer
				alu[dl_eop_buf_handle, --, B, cur_buf]			; last buffer in buffer chain.

		#endif	// IXP2800

	#else 	

		#ifdef IXP2800
	
			; Flush meta-data belonging to channel 0 and BUF_SRAM_CHANNEL using
			; separate macro calls.

			ctx_arb[sig_dram, sig_next_thread, sig_prev_meta, sig_cur_meta, \
					sig_prev_meta_ch3, sig_cur_meta_ch3], defer[2]
				alu[dl_buf_handle, --, B, lm_sop_buf] 			; first buffer
				alu[dl_eop_buf_handle, --, B, cur_buf]			; last buffer in buffer chain.
	
		#else

			ctx_arb[sig_dram, sig_next_thread, sig_prev_meta, sig_cur_meta], defer[2]
				alu[dl_buf_handle, --, B, lm_sop_buf] 			; first buffer
				alu[dl_eop_buf_handle, --, B, cur_buf]			; last buffer in buffer chain.

		#endif	// IXP2800

	#endif	// RFC_PPP_COUNTERS	

	xbuf_free($prev_meta1)  							; free $prev_meta1
	xbuf_free($cur_meta1)								; free $cur_meta1
	xbuf_free($bytes_counter2)

end#:

	alu[dl_next_block, --, B, BID_NEXT_BLOCK] 				; set pkt destination

.end
#endm


///////////////////////////////////////////////////////////////////////////////
//
// _packet_rx_mpp_processing
//
//	 	Description: 
//
//			Process a mpacket containing neither SOP nor EOP. i.e this is a 
//			MOP (middle of packet).
//
//	 	Outputs:
//
//		Inputs:
//
//		Size: 
//
//			?? instructions
//
///////////////////////////////////////////////////////////////////////////////

#macro _packet_rx_mpp_processing()

.begin

	.reg 	buf_size
	.reg	cur_buf
	.reg	cell_count
	.reg 	buf_offset
	.reg	cur_data
	.reg 	temp
	.reg 	header_type
	.reg	buf_hndl
	.sig	sig_dram
	.sig	sig_buf_alloc
	.sig	sig_meta

	#ifdef IXP2800

		; Flush meta-data belonging to channel 0 and BUF_SRAM_CHANNEL using
		; separate macro calls.
		.sig	sig_meta_ch3

	#endif
	
	//	Expect a MOP and get a MOP: Normal case

	//	Increment buffer size by number of bytes received. Buffer size
	//	is the amount of the data currently present in the buffer. We are
	//	guaranteed to have room in the buffer for this mpacket.
	//	[ See the defer slots below]

	//	Find out if buffer is full (for next mpacket). 
	//	Buffer size + headroom in the buffer will show 
	//	if the buffer is full.
	
	alu[buf_size, byte_count, +, lm_buf_size]			; buf size after THIS mpacket.
	alu[buf_size, buf_size, +, lm_buf_offset] 			; + headroom
	
	alu[--, buf_size, -, max_buf_size]					; is buffer full? for next time

	blt[mpp_buffer_not_full#],defer[2] 					; no, goto buffer not full
		alu[lm_buf_size, byte_count, +, lm_buf_size]	; increment buffer size						
		alu[lm_pkt_size, byte_count, +, lm_pkt_size]	; increment packet size

	//	At this point buffer will be full after storing the current
	//	mpacket. See if a new buffer (prefetch buffer) is available .

	alu[--, --, B, $prefetch_buf_handle, <<8]			; [23:0] contains buf handle
	beq[mpp_null_buffer#]								; null buffer error processing	

	//	New buffer is available. Normal processing.

	//	move received mpkt to DRAM

	_packet_rx_move_rbuf_to_pkt_buf(lm_cur_data, ele_addr, byte_count, rbuf_base, sig_dram)

	//	check if current buffer is a SOP buffer. (i.e a buffer containing
	//	start of packet.)

	alu[cur_buf, --, B, lm_cur_buf] 					; lm cannot be used for bit ops
	br_bclr[cur_buf, 30, mpp_buffer_full_not_sop#], defer[1]	; [30] is SOP bit
		alu_shf[cur_buf, cur_buf, AND~, 0x1, <<31]		; anyway, clear EOP [31] bit.
														; because this buffer will not
														; contain EOP.
mpp_buffer_full_sop#:

	//	[Recap] This buffer is full. This buffer contains SOP. 
	//	A new buffer is available.

	//	increment cell count depending on the number of bytes received.
	//	embed this cell count in the buffer handle.

	_packet_rx_incr_cell_count(lm_cell_count, lm_bytes_remaining, byte_count)

	alu[cell_count, --, B, lm_cell_count]				; lm can't be used with <<
	alu[cur_buf, cur_buf, OR, cell_count, <<24]			; [29:24] are cell count

	//	At this point, no meta data needs to be flushed to SRAM
	//	keep the SOP buffer offset, buffer size and move current
	//	buffer handle to the previous buffer handle in the context data
	//	cur_buf contains the current buffer handle

	//	We are done with current buffer. Make it previous buffer as well as
	//	SOP Buffer

	alu[lm_sop_buf, --, B, cur_buf]						; cur buf is SOP buf
	alu[lm_prev_buf, --, B, cur_buf]					; cur buf to prev buf
	alu[lm_prev_buf_size, --, B, lm_buf_size]			; buf size to prev

	//	Since this buffer is also SOP buffer, remember
	//	a few more fields.

	alu[lm_sop_buf_size,--, B, lm_buf_size]				; buf size to sop buf size.

	//	Ok, now make the prefetch buffer as current buffer. Any new
	//	mpacket received will be stored in this new buffer.

	alu[temp, $prefetch_buf_handle, AND~, 0x1, <<30]	; clear sop bit
	alu[lm_cur_buf, --, B, temp]						; prefetch buffer is current
	alu[lm_cur_data, --, B, prefetch_data_handle]		; prefetch data ptr is current

	//	clear cell count and byte reminder.

	alu[lm_cell_count, --, B, 0x0] 						; clear cell count
	alu[lm_bytes_remaining, --, B, 0x0] 				; clear bytes remaining

	//	Prefetch a new buffer so that it can be used next time around.

	dl_buf_alloc($prefetch_buf_handle, BUF_FREE_LIST0, sig_buf_alloc, SIG_NONE)

	//	End of Phase 1.
	//	Wait on the all the previous I/O plus the next thread signal.

	ctx_arb[sig_dram, sig_next_thread, sig_buf_alloc], defer[2]
		alu[lm_buf_size, --, B, 0] 						; set buffer size as 0
		alu[lm_buf_offset, --, B, NON_SOP_OFFSET]		; set offset as 0

	//	Get DRAM address of the new buffer 

	dl_buf_get_data(prefetch_data_handle, $prefetch_buf_handle)

	// 	go to phase 2 processing

	//	Remember that we have received only a part of the packet so far.
	//	Until we receive EOP, dl_next_block and dl_buf_handle should 
	//	be null.

	br[end#], defer[2]
		alu[dl_next_block, --, B, BID_NULL] 			; set next_block NULL
		alu[dl_buf_handle, --, B, DL_BUF_NULL]	   		; set dl_buf_handle NULL
									
mpp_buffer_full_not_sop#:
	
	//	[Recap]: Current buffer is full. But it does not contain SOP.
	//	cur_buf contains the current buffer handle.

	//	Two slightly different cases to handle here. 
	//	[1] Previous Buffer is SOP. (this will have headroom)
	//	[2] Previous Buffer is not SOP. (this will not)

	//	increment cell count depending on the number of bytes received.
	//	embed this cell count in the buffer handle.
	//	XXX - Why can't this cell count calc. be done in the common path 
	//	up above.

	_packet_rx_incr_cell_count(lm_cell_count, lm_bytes_remaining, byte_count)

	alu[cell_count, --, B, lm_cell_count]				; lm can't be used with <<
	alu[cur_buf, cur_buf, OR, cell_count, <<24]			; [29:24] are cell count

	//	Now we are ready to flush META data of previous buffer to SRAM.
	//	Remember the META data contains a hw_next pointer which, in this case, 
	//	should point to current buffer. [Various META data fields are set in
	//	the defer slots below]

	alu[dl_buf_handle,--, B, lm_prev_buf]				; prepare to set meta data

	//	Check if the previous buffer is a SOP buffer
	//	If it is not a SOP buffer, the offset is set 
	//	as NON_SOP_OFFSET(0) (in the defer slot) i.e no headroom
	//	If it is a SOP buffer we have stored the offset in
	//	lm. Use that.

	br_bclr[dl_buf_handle,30,mpp_prev_buffer_not_sop#], defer[3]
		dl_meta_set_buffer_next(cur_buf)    			; cur buf is next for prev.
		dl_meta_set_buffer_size(lm_prev_buf_size)  		; set previous buffer size
		dl_meta_set_offset(NON_SOP_OFFSET)				; saves an extra branch

	//	Here, previous buffer is a SOP buffer
	//	set header type as stored in LM.
	alu[header_type,lm_header_type, and, 0xFF]			; get header type
	alu[buf_offset, --, B, lm_sop_buf_offset]			; Get SOP buffer offset
	dl_meta_set_offset(buf_offset)						; set SOP buffer offset
	dl_meta_set_header_type(header_type)				; set header type for the port

mpp_prev_buffer_not_sop#:

	//	All the required fields of META data have been set
	//	for the previous buffer. Write it to SRAM. The META
	//	Data for !SOP buffers is only a subset of SOP META
	//	and contains only the first 2 longwords of META.

	xbuf_alloc($x_0,2, write) 							; allocate xfer registers

	#ifdef IXP2800

 		; Flush meta-data belonging to channel 0 and BUF_SRAM_CHANNEL using
		; separate macro calls.
		dl_meta_flush_cache($x_0, dl_buf_handle, sig_meta, SIG_NONE, 0, 1)	; write 1 LW
		dl_meta_flush_cache($x_0, dl_buf_handle, sig_meta_ch3, SIG_NONE, 1, 1)	; write 1 LW

	#else

		dl_meta_flush_cache($x_0, dl_buf_handle, sig_meta, SIG_NONE, 0, 2)	; write 1 LWs

	#endif	// IXP2800

	//	Ok, prepare for our next trip down this lane.

	//	We are done with current buffer. Make it previous buffer 

	alu_shf[lm_prev_buf, --, B, cur_buf]
	alu[lm_prev_buf_size,--, B, lm_buf_size]
	
	//	Ok, now make the prefetch buffer as current buffer. Any new
	//	mpacket received will be stored in this new buffer.

	alu[temp, $prefetch_buf_handle, AND~, 0x1, <<30]	; clear sop bit
	alu[lm_cur_buf, --, B, temp]						; prefetch buffer is current
	alu[lm_cur_data, --, B, prefetch_data_handle]		; prefetch data ptr is current

	//	clear cell count and byte reminder.

	alu[lm_cell_count, --, B, 0x0]  					; clear cell count
	alu[lm_bytes_remaining, --, B, 0x0] 				; clear bytes remaining

	//	Prefetch a new buffer so that it can be used next time around.

	dl_buf_alloc($prefetch_buf_handle, BUF_FREE_LIST0, sig_buf_alloc, SIG_NONE)

	//	End of Phase 1.
	//	Wait on the all the previous I/O plus the next thread signal.

	#ifdef IXP2800

		; Flush meta-data belonging to channel 0 and BUF_SRAM_CHANNEL using
		; separate macro calls.

		ctx_arb[sig_dram, sig_next_thread, sig_meta, \
				sig_meta_ch3, sig_buf_alloc],and, defer[2]
			alu[lm_buf_size, --, B, 0] 						; set buffer size as 0
			alu[lm_buf_offset, --, B, NON_SOP_OFFSET]		; set offset as 0

	#else

		ctx_arb[sig_dram, sig_next_thread, sig_meta, sig_buf_alloc],and, defer[2]
			alu[lm_buf_size, --, B, 0] 						; set buffer size as 0
			alu[lm_buf_offset, --, B, NON_SOP_OFFSET]		; set offset as 0

	#endif	// IXP2800

	xbuf_free($x_0) 									; free sram trnsfer regs 

	//	Get DRAM address of the new buffer 

	dl_buf_get_data(prefetch_data_handle, $prefetch_buf_handle)

	// 	go to phase 2 processing

	//	Remember that we have received only a part of the packet so far.
	//	Until we receive EOP, dl_next_block and dl_buf_handle should 
	//	be null.

	br[end#], defer[2]
		alu[dl_next_block, --,B, BID_NULL]			 	; NULL (0xFF)
		alu[dl_buf_handle, --, B, DL_BUF_NULL]	   		; NULL (0xFF)

mpp_buffer_not_full#:

	//	MOP received and Current buffer is not FULL. (i.e it can
	//	take current mpacket + atleast one more mpacket)

	//	move received mpkt to DRAM

	_packet_rx_move_rbuf_to_pkt_buf(lm_cur_data, ele_addr, byte_count, rbuf_base, sig_dram)

	//	increment cell count depending on the number of bytes received.

	_packet_rx_incr_cell_count(lm_cell_count, lm_bytes_remaining, byte_count)

	alu[lm_cur_data, byte_count, +, lm_cur_data]

	//	End of Phase 1.
	//	Wait on the all the previous I/O plus the next thread signal.
	//	Remember that we have received only a part of the packet so far.
	//	Until we receive EOP, dl_next_block and dl_buf_handle should 
	//	be null.

    ctx_arb[sig_dram, sig_next_thread], defer[2], br[end#]
		alu[dl_next_block, --, B, BID_NULL] 			; NULL (0xFF)
		alu[dl_buf_handle, --, B, DL_BUF_NULL]	   		; NULL (0xFF)

mpp_null_buffer#:

	//	NULL Buffer error handling. 

	_packet_rx_incr_counter(port_num, PACKET_PKTS_DROPPED)
		
	//	If buffer not available drop the current (so far assembled) 
	//	packet as well as the just received mpacket. Reset State 
	//	Machine (to	START State)

	_packet_rx_set_state(STATE_START)								; reset to start state.

	//	Prefetch a new buffer so that it can be used next time around.

	dl_buf_alloc($prefetch_buf_handle, BUF_FREE_LIST0, sig_buf_alloc, SIG_NONE)

	ctx_arb[sig_next_thread, sig_buf_alloc], and, defer[2]
		alu[drop_sop_handle, --, B, lm_sop_buf]	
		alu[drop_eop_handle, --, B, lm_cur_buf]

	//	Get DRAM address of the new buffer 

	dl_buf_get_data(prefetch_data_handle, $prefetch_buf_handle)

	//	drop current (so far assembled) packet

	_packet_rx_drop_current_pkt(drop_sop_handle, drop_eop_handle)

	alu[dl_next_block, --, B, BID_NULL]		 			; NULL (0xFF)
	alu[dl_buf_handle, --, B, DL_BUF_NULL]	   			; NULL (0xFF)

end#:

.end
#endm



#endif /* OLD_CODE */


////////////////////////////////////////////////////////////////////////////////
//	This file is the main file for AAL5 RX block.
//
//	Main entry point -- Code begins execution here.
//
///////////////////////////////////////////////////////////////////////////////


main#:

.begin

	.sig sig_sink
	.sig sink_sig  					// signal used in dl_sink for scratch_put
	.reg sink_sig_mask				// signal mask used by aal5_rx() and dl_sink()
	.sig sig_hdr					


	// System initialization should be performed only once by Thread 0 of this uE
	br!=ctx[0, SKIP_SYSTEM_INIT#]
	one_time_rx_init[]

#ifdef DO_SYSTEM_INIT

	system_init[]

#endif /* SYSTEM_INIT */

	//  Dispatch packet and aal5 rx routines based on current context number 
	//  contexts 0, 2, 4, 6 will handle packet rx
	//  contexts 1, 3, 5, 7 will handle aal5 rx

SKIP_SYSTEM_INIT#:
	br=ctx[1, AAL5_RX_LOOP#]
	br=ctx[3, AAL5_RX_LOOP#]
	br=ctx[5, AAL5_RX_LOOP#]
	br=ctx[7, AAL5_RX_LOOP#]

	//  THIS PATH IS EXECUTED FOR CONTEXTS ASSOCIATED WITH DOWNSTREAM PROCESSING !!!
	#ifndef DOWNSTREAM
		#define DOWNSTREAM
	#endif

	#ifdef UPSTREAM
		#undef UPSTREAM
	#endif

	#ifdef BUF_POOL
		#undef BUF_POOL
	#endif

    #ifdef BUF_SRAM_BASE
        #undef BUF_SRAM_BASE
    #endif

    #ifdef BUF_SDRAM_BASE
        #undef BUF_SDRAM_BASE
    #endif

    #ifdef BUF_SRAM_BASE_LW
        #undef BUF_SRAM_BASE_LW
    #endif
    
    #ifdef BUF_SDRAM_BASE_LW
        #undef BUF_SDRAM_BASE_LW
    #endif
    
    #ifdef AVLBL_PKT_BUFFERS
        #undef AVLBL_PKT_BUFFERS
    #endif
    
    #ifdef DL_REL_BASE
        #undef DL_REL_BASE
    #endif

	// redefine memory defs so they can reference correct memory resources
	#define_eval BUF_SRAM_BASE		DOWNSTREAM_BUF_LW0_SRAM_BASE
	#define_eval BUF_SDRAM_BASE		DOWNSTREAM_BUF_SDRAM_BASE

	#define BUF_SRAM_BASE_LW		(BUF_SRAM_BASE + META_DATA_SIZE)
	#define BUF_SDRAM_BASE_LW		(BUF_SDRAM_BASE + BUFFER_SIZE)

	#define AVLBL_PKT_BUFFERS		(NO_OF_PKT_BUFFERS - 1)
	#define BUF_POOL 				BUF_SDRAM_BASE_LW, BUFFER_SIZE, BUF_SRAM_BASE_LW, META_DATA_SIZE

#ifdef USE_IMPORT_VAR
	#define DL_REL_BASE				DOWNSTREAM_DL_REL_BASE
#else
	#define DL_REL_BASE		    	(BUF_SDRAM_BASE - ((BUF_SRAM_BASE & 0x3fffffff) * DL_DS_RATIO))
#endif

	//	Initialization of the packet rx portion of code
	packet_rx_init[]
	

PACKET_RX_LOOP#:

.while(1)

	; Allocate transfer registers to be used by dl_sink
	xbuf_alloc($x_dl_sink, 5, write)	// no code generated    

	; Packet processing code. For every mpacket the output of this
	; macro is dl_next_block, dl_buf_handle and dl_eop_buf_handle. 
	packet_pkt_rx()

	; Send the packet to the next block in the pipeline.
	; Even though dl_sink is called for every mpacket, actual
	; work will be done only on receiving EOP.
	dl_sink($x_dl_sink, sig_sink, sink_sig_mask)

	; To avoid warnings. The assembler can't find that we actually
	; waited on this sig_sink because of the use of signal masks.
	.io_completed sig_sink
	.io_completed drop_sig

	; Free the transfer registers
	xbuf_free($x_dl_sink)	// no code generated

.endw


	br[RX_LOOP_END#]									// skip aal5 rx loop portion


	;-----------------------------------------------------------------------------
	;	   Main AAL5 RX code starts here
	;-----------------------------------------------------------------------------
AAL5_RX_LOOP#:

	//  THIS PATH IS EXECUTED FOR CONTEXTS ASSOCIATED WITH UPSTREAM PROCESSING !!!
	#ifdef DOWNSTREAM
		#undef DOWNSTREAM
	#endif

	#ifndef UPSTREAM
		#define UPSTREAM
	#endif

	#ifdef BUF_POOL
		#undef BUF_POOL
	#endif

    #ifdef BUF_SRAM_BASE
        #undef BUF_SRAM_BASE
    #endif

    #ifdef BUF_SDRAM_BASE
        #undef BUF_SDRAM_BASE
    #endif

    #ifdef BUF_SRAM_BASE_LW
        #undef BUF_SRAM_BASE_LW
    #endif
    
    #ifdef BUF_SDRAM_BASE_LW
        #undef BUF_SDRAM_BASE_LW
    #endif
    
    #ifdef AVLBL_PKT_BUFFERS
        #undef AVLBL_PKT_BUFFERS
    #endif
    
    #ifdef DL_REL_BASE
        #undef DL_REL_BASE
    #endif
    
	// redefine memory defs so they can reference correct memory resources
	#define BUF_SRAM_BASE		UPSTREAM_BUF_LW0_SRAM_BASE
	#define BUF_SDRAM_BASE		UPSTREAM_BUF_SDRAM_BASE

	#define BUF_SRAM_BASE_LW	(BUF_SRAM_BASE + META_DATA_SIZE)
	#define BUF_SDRAM_BASE_LW	(BUF_SDRAM_BASE + BUFFER_SIZE)

	#define AVLBL_PKT_BUFFERS	(NO_OF_PKT_BUFFERS - 1)
	#define	BUF_POOL 			BUF_SDRAM_BASE_LW, BUFFER_SIZE, BUF_SRAM_BASE_LW, META_DATA_SIZE

#ifdef USE_IMPORT_VAR
	#define DL_REL_BASE			UPSTREAM_DL_REL_BASE
#else
	#define DL_REL_BASE		    (BUF_SDRAM_BASE - ((BUF_SRAM_BASE & 0x3fffffff) * DL_DS_RATIO))
#endif
	//  Initialization of the AAL5 rx portion of code.
	common_aal5_rx_init()
	move[zero, 0]

	alu[dl_meta[6], --, B, 0x0]

.while(1)

	;----------------------------------------------------------------------
	; For ONE_ME_RX_FP - Phase1#, Phase2#, Phase3# 
	; For TWO_ME_RX_FP - Phase1#, Phase2#, Phase3#, Phase4#, Phase5#  
	;----------------------------------------------------------------------
	
	; AAL5 frame reassembly.
	aal5_rx(sink_sig_mask)


	;----------------------------------------------------------------------
	; For ONE_ME_RX_FP - Phase4#
	; For TWO_ME_RX_FP - Phase6#
	;----------------------------------------------------------------------
	
	xbuf_alloc[$x_dl_sink_aal5, RX_TO_FWD_MSG_SIZE, write]	// sxfer reg to be used by dl_sink

	; Passing onto next block.
	dl_sink($x_dl_sink_aal5, sink_sig, sink_sig_mask) 
												 

	xbuf_free($x_dl_sink_aal5)				// free transfer registers

.io_completed sink_sig

.endw

RX_LOOP_END#:
    nop
.end   // MAIN# 


#endif //	__COMMON_RX_UC__
