///////////////////////////////////////////////////////////////////////////////
//
//                  I N T E L   P R O P R I E T A R Y
//
//     COPYRIGHT (c)  2003 BY  INTEL  CORPORATION.  ALL RIGHTS
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT
//     THE PRIOR WRITTEN PERMISSION OF :
//
//                        INTEL  CORPORATION
//
//                     2200 MISSION COLLEGE BLVD
//
//               SANTA  CLARA,  CALIFORNIA  95052-8119
//
///////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: common_rx_init.uc
//
//      Purpose: Performs global initialization as well as specific Packet and
//               AAL5 RX microengine initialization.
//
//      History:
//
//
//      Date            Comment                         By
//      ---------------------------------------------------------------------
//
//      12/17/2002      Created                         Czarek Wierucki
//
//
/////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
//
// _packet_rx_setup_next_thread_sig
//
//	 	Description:
//
//			Setup a global register that will be used in next thread signalling.
//
//	 	Outputs:
//			nt_sig_reg		: a register containing
//							  [7]   - next ctx  bit
//							  [6:3] - signal number
//							  [2:0] - thread number
//		Inputs:
//
//			thread_num		: Current thread number
//
//		Size:
//
//
//
///////////////////////////////////////////////////////////////////////////////

#macro	_packet_rx_setup_next_thread_sig[nt_sig_reg, thread_num]
.begin


	//  Initialize next thread signal register used for packet RX.
	//  0->2->4->6->0
	alu[nt_sig_reg, thread_num, +, 2]			; calculate next thread number
	alu[nt_sig_reg, nt_sig_reg, AND, 7]			; make sure the thread number value is [0..7]
	alu[nt_sig_reg, nt_sig_reg, OR, &sig_next_thread, <<3]	; [6:3] signal number
.end
#endm



///////////////////////////////////////////////////////////////////////////////
//	common_rx_config_msf_rx_ctrl:
//
//		Description:
//			Initialize MSF RX control MSF reg for both Packet RX and AAL5 RX.
//
//		Input:
//
//		Output:
//
//		Size: N/A
//
///////////////////////////////////////////////////////////////////////////////
#macro common_rx_config_msf_rx_ctrl()
.begin

	.reg $tmp, msf_rx_addr
	.sig msf_wr_sig

	//	[1:0]		Reserved.
	//	[3:2]		01 ==> RBUF element size is 128.
	//	[8:4]		Reserved.
	//	[9]			0  ==> CSIX Free list - Data and control cframes goto
	//				different Rx_thread_freelist. (Not used here).
	//	[12:10]		Reserved.
	//	[13]		0  ==> Tx Cbus width is 4 bit.
	//	[15:14]		Reserved.
	//	0[16]		1  ==> Tx CBus Mode is full duplex mode.
	//	1[17]		0  ==> RX Mphy related - Not used here.
	//	2[18]		0  ==> RX Mphy related - Not used here.
	//	3[19]		0  ==> RX Single Phy Mode.
	//	4[21:20]	10 ==> RX width 4 x 8.
	//	6[22]		0  ==> RX Mode is UTOPIA/POS mode.
	//	7[27:23]	Reserved.
	//	12[31:28]	0x? ==> Receive enable based on RXTX_ATM_LINE and RXTX_ETH_LINE.


	// make sure the init is done in two steps for IXP2400 - first set up all
	// configuration bits then enable all rx lines
	move($tmp, 0x00210004)

	move(msf_rx_addr, MSF_RX_CNTL)

	msf[write, $tmp, msf_rx_addr, 0, 1], sig_done[msf_wr_sig]
	ctx_arb[msf_wr_sig]

	#define_eval EN_MSF_RX_BITS_A 	1 << RXTX_ATM_LINE
	#define_eval EN_MSF_RX_BITS_E   1 << RXTX_ETH_LINE
	#define_eval EN_MSF_RX_BITS 	((EN_MSF_RX_BITS_A + EN_MSF_RX_BITS_E) << 28)
	#define_eval MSF_RX_CNTL_WORD	(0x00210004 + EN_MSF_RX_BITS)

	move($tmp, MSF_RX_CNTL_WORD)

	msf[write, $tmp, msf_rx_addr, 0, 1], sig_done[msf_wr_sig]
	ctx_arb[msf_wr_sig]

.end
#endm   //   common_rx_config_msf_rx_ctrl()



///////////////////////////////////////////////////////////////////////////////
//
// one_time_rx_init
//
//	 	Description:
//			One time initialisation code executed by Thread 0.
//
//	 	Outputs:
//
//		Inputs:
//
//		Size:
//
//
///////////////////////////////////////////////////////////////////////////////
#macro one_time_rx_init[]
.begin

	#ifdef		MAX_BUFF_SIZE
		#warning	MAX_BUFF_SIZE is already defined
	#endif

	#define		MAX_BUFF_SIZE		BUFFER_SIZE

	.reg lm_addr								; LM addr to store reg values


	move[@rand_buf_offset, MAX_BUFFER_OFFSET]	; Startng Random Buffer offset. See [Randomizing Buffer Offset]
    move[@prx_rand_buf_offset, MAX_BUFFER_OFFSET]	; Startng Random Buffer offset. See [Randomizing Buffer Offset]
	move[@rand_offset_mask, RAND_OFFSET_MASK]	; 0x1ff

	//	Thread 0 does some onetime global initialisation.
	//	For SPHY 4x8 mode there are 4 thread freelists (one per line).
	//  Look for RXTX_ATM_LINE and RXTX_ETH_LINE definition for currnet line assignment.

	//	Set timeout for all freelists.
	.if (RXTX_ETH_LINE == 0)
		_packet_rx_set_rx_thread_freelist_timeout[0, RX_THREAD_FREELIST_TIMEOUT_0] 	; timeout is in bus clk cycles
	.elif (RXTX_ATM_LINE == 0)
		_aal5_rx_set_rx_thread_freelist_timeout(0, RX_THREAD_FREELIST_TIMEOUT_0) 	; timeout is in bus clk cycles
	.else
		_packet_rx_set_rx_thread_freelist_timeout[0, 0x0]
	.endif

	.if (RXTX_ETH_LINE == 1)
		_packet_rx_set_rx_thread_freelist_timeout[1, RX_THREAD_FREELIST_TIMEOUT_0] 	; timeout is in bus clk cycles
	.elif (RXTX_ATM_LINE == 1)
		_aal5_rx_set_rx_thread_freelist_timeout(1, RX_THREAD_FREELIST_TIMEOUT_0) 	; timeout is in bus clk cycles
	.else
		_packet_rx_set_rx_thread_freelist_timeout[1, 0x0]
	.endif

	.if (RXTX_ETH_LINE == 2)
		_packet_rx_set_rx_thread_freelist_timeout[2, RX_THREAD_FREELIST_TIMEOUT_0]
	.elif (RXTX_ATM_LINE == 2)
		_aal5_rx_set_rx_thread_freelist_timeout(2, RX_THREAD_FREELIST_TIMEOUT_0) 	; timeout is in bus clk cycles
	.else
		_packet_rx_set_rx_thread_freelist_timeout[2, 0x0]
	.endif

	.if (RXTX_ETH_LINE == 3)
		_packet_rx_set_rx_thread_freelist_timeout[3, RX_THREAD_FREELIST_TIMEOUT_0]
	.elif (RXTX_ATM_LINE == 3)
		_aal5_rx_set_rx_thread_freelist_timeout(3, RX_THREAD_FREELIST_TIMEOUT_0) 	; timeout is in bus clk cycles
	.else
		_packet_rx_set_rx_thread_freelist_timeout[3, 0x0]
	.endif

	; set up some H/W stuff.
	_packet_rx_set_rx_up_ctrl[RXTX_ETH_LINE]			; set up MSF RX_UP_Control for packet rx.
	common_aal5_rx_set_rx_up_ctrl[RXTX_ATM_LINE]		; set up MSF RX_UP_Control for aal5 rx

	_packet_rx_free_all_rbuf_elements[RBUF_TOTAL_ELEMENTS]	; free all rbuf elements for both packet and aal5 rx

	_packet_rx_clear_all_context_desc[]					; clear reassembly context data in local mem
	_packet_rx_set_all_port_header_type[]				; set header type for each port

	immed32[lm_addr, L2PORT_MAP_BASE]					; L2PORT_MAP initialiation
	local_csr_wr[ACTIVE_LM_ADDR_0, lm_addr]
	nop
	nop
	nop
    alu[*l$index0[0], --, b, ETH_L2_PORT_NUM]
        
	; Constants Init in Local Memory.
	; Init the LM handle 1 with the lm addr for storing register values
	immed32[lm_addr, LM_CONSTANTS_BASE]
	alu_shf[lm_addr, --, b, lm_addr, <<2]
	local_csr_wr[ACTIVE_LM_ADDR_1, lm_addr]
	nop
	nop
	nop

	_aal5_rx_lm_reg_init()

	; absolute registers init	
	move(@rand_buf_offset, MAX_BUFFER_OFFSET)	; Starting Random Buf offset
		
	#ifdef HASH_INIT	
		; Do Hash tables and RXC init.
		_aal5_rx_hash_rxc_init()
	#endif

	; Init CAM and LM with first 16 RXC
	immed32(vc_info_sram_base, AAL5_RX_VC_INFO_SRAM_BASE)  // init vc_info_sram_base needed by _aal5_rx_init_cam()
	_aal5_rx_init_cam()

.end
#endm



///////////////////////////////////////////////////////////////////////////////
//
// packet_rx_init
//
//	 	Description: 
//
//			POS RX Thread initialization code. 
//
//	 	Outputs:
//
//			Many global registers are set to predefined values.
//
//		Inputs:
//
//
//		Size: 
//
//			??
//
///////////////////////////////////////////////////////////////////////////////
#macro packet_rx_init[]
.begin

	.reg me_num
	.reg temp_nt_sig_reg
	.sig sig_buf_alloc

#ifdef		MAX_PAYLOAD
#warning	MAX_PAYLOAD is already defined
#endif

#define		MAX_PAYLOAD		(BUFFER_SIZE - PPP_HEADER)

	move[max_buf_size, BUFFER_SIZE]						; set max buffer size
	move[prx_rbuf_base, RBUF_TBUF_BASE]            			; set up RBUF base
#ifdef RFC_PPP_COUNTERS
	immed32[counter_base, PACKET_COUNTERS_SRAM_BASE]		; counters base addr.
#endif

	move[dram_rbuf_ov,((1 << 25)|(1<<4))]	   			; [25] ref_cnt, [4] xfer regs overwrite	
												
	immed32[protocol_mask,0xFFFF]          				; protocol mask for PPP hdr.

	//	To put a thread on to free list, we need the thread number,
	//	ME number, a signal and two xfer registers. Assemble all this
	//	info in a register.

	_packet_rx_get_current_me_thread_number[me_num, thread_num]
	_packet_rx_setup_next_thread_sig[nt_sig_reg, thread_num]; setup next thread signal register

	_packet_rx_set_freelist_reg[freelist_reg, RX_EVENT_SIG, me_num, thread_num, &$prx_rsw1]
	alu[freelist_reg,--, B, freelist_reg, <<16]			; for fast_wr, ms16bits is data

	//	Initialise Dispatch Loop vaiables to known state.

	alu[dl_next_block, --, B, BID_NULL]					; set dl_next_block NULL
	alu[dl_buf_handle, --, B, DL_BUF_NULL]     			; set dl_buffer_handle NULL
	alu[dl_eop_buf_handle, --, B, DL_BUF_NULL] 			; 

	//	This call is to eliminate the warnings
	//	What variable you pass here doesn't matter. Will get better next time

	dl_meta_init_cache[dl_next_block, dl_next_block, dl_next_block, dl_next_block, 0, 0, 0, 0]	; TBD: to avoid warning. Revisit


	//	All other threads just wait for a kickstart signal
	//	This will come from system_init.

	.if(ctx() != 0)
		ctx_arb[sig_next_thread]						; wait for signal from prev thread.
	.else

		// Initialise the free buffer list for packet rx (downstream)
        // when Core Components are not used. Otherwise buffers will
        // be initialised by CC infrastructure.
        #ifdef DO_FREELIST_INIT
		    #ifdef FREELIST_INIT_VIA_SCRIPT

				// Because of design and using both DOWNSTREAM and UPSTREAM defines we 
				// have to make this ugly define here
				#define TEMPORARY_FREELIST_QDESC            DOWNSTREAM_TEMPORARY_FREELIST_QDESC

                _aal5_rx_read_freelist_qdesc()

				#undef TEMPORARY_FREELIST_QDESC

            #else
                // Create buffer list for upstream direction - only for the first thread
		        dl_buffer_init[]
            #endif
        #endif /* DO_FREELIST_INIT */
	.endif

	//	prefetch a buffer to store packets. From the handle get the location
	//	in DRAM where the packet will be stored.

	dl_buf_alloc[$prx_prefetch_buf_handle, BUF_FREE_LIST0, sig_buf_alloc, sig_buf_alloc]
	dl_buf_get_data[prefetch_data_handle, $prx_prefetch_buf_handle] ; get prefetch data handle 

	//	In SPHY 4x8 mode there is one thread freelist per port. 
	//	We use four threads per port. i.e four threads per freelist.

	/*
	 * In this case all even contexts will handle packet rx. 
	 * Add this context to the free list associated with RXTX_ETH_LINE
	 */
	alu[rx_freelist, --, B, RX_THREAD_FREELIST_/**/RXTX_ETH_LINE]

	msf[fast_wr, --, rx_freelist, freelist_reg]			; add this thread to freelist

no_msf_config#:
	//	signal next packet rx thread to wakeup.
	_packet_rx_signal_next_thread[nt_sig_reg]	

	// signal first aal5 rx thread to wakeup.
	move[temp_nt_sig_reg, 1]	 						; wake up Thread 1 signal (AAL5 RX)
	alu[temp_nt_sig_reg, temp_nt_sig_reg, OR, NEXT_THREAD_SIGNAL, <<3]	; [6:3] signal number
	_packet_rx_signal_next_thread[temp_nt_sig_reg]			

    ctx_arb[rx_event] 	    							; wait for a mpacket

#undef	MAX_PAYLOAD								

.end
#endm





////////////////////////////////////////////////////////////////////////////////
//	These are the init macros specific for AAL5 RX part of the code
////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// common_aal5_rx_set_rx_up_ctrl:
//
//		Description: Initialize MSF Rx_UP_control reg: clock decode response 
//					time, cell size, parity, CP_mode and PP_mode.
//	
//		Input: 
//			rx_up_ctrl_num:	Rx_UP_control MSF reg, range [0,3]
//	
//		Output: 
//
//		Size: N/A
//
//////////////////////////////////////////////////////////////////////////////
#macro common_aal5_rx_set_rx_up_ctrl(RX_UP_CTRL_NUM)
.begin

	.reg tmp, msf_rx_addr

	//	[0]		- 1  ==> UPCtrl_DRTime decode response time, 
	//					  0=1cycle, 1=2cycles (reset = 0, originally it was 1).
	//	[1]		- x	 ==> UPCtrl_CellSize, 0=52 bytes   1=56 bytes.
	//	[3:2]	- 10 ==> UPCtrl_Parity, single bit odd parity.
	//	[4]		- 1  ==> UPCtrl_CP_Mode, 1 = packet mode; 
	//						0 = cell mode (originally it was 1).
	//	[5]		- 1  ==> UPCtrl_PP_Mode, 0 = POS2; 1 = POS3 
	//						(reset =0 , orignally it was 1).
	//	[31:6]	- x  ==> reserved.

	move(msf_rx_addr, RX_UP_CTRL_/**/RX_UP_CTRL_NUM)

#ifdef FPGA_HEADER
	// we need to run in POS-PHY Level 3 mode to allow the reception of the cell + FPGA Header 
	#define 	 AAL5_RX_FPGA_UP_CTRL_CONFIG	((RX_CP_MODE_PACKET << RX_CP_MODE_BITS)|(RX_PARITY_ODD << RX_PARITY_BITS)|(RX_DR_TIME_2_CLOCK << RX_DR_TIME_BITS)|(1<<5))
	#define_eval AAL5_RX_UP_CTRL_CONFIG_0		AAL5_RX_FPGA_UP_CTRL_CONFIG
	#define_eval AAL5_RX_UP_CTRL_CONFIG_1		AAL5_RX_FPGA_UP_CTRL_CONFIG
	#define_eval AAL5_RX_UP_CTRL_CONFIG_2		AAL5_RX_FPGA_UP_CTRL_CONFIG
	#define_eval AAL5_RX_UP_CTRL_CONFIG_3		AAL5_RX_FPGA_UP_CTRL_CONFIG
#endif // FPGA_HEADER

	; RX_UP_CTRL_{0,3}  has only 6 bits [0:5] relevant, so we can use fast_wr.
	 immed[tmp, AAL5_RX_UP_CTRL_CONFIG_/**/RX_UP_CTRL_NUM]

	; msf[fast_wr,...] requires 
	; rx_up_ctrl_num in upper 16 bits  and address in lower 16 bits.
	alu_shf[tmp, --, b, tmp, <<16]

	msf[fast_wr, --, tmp, msf_rx_addr]				

.end
#endm   // common_aal5_rx_set_rx_up_ctrl()



///////////////////////////////////////////////////////////////////////////////
//	common_aal5_rx_dl_init:
//
//		Description:
//			Initializes some dispatch loop variables.
//
//		Input: 
//			
//		Output: 
//
//		Size: N/A
//
///////////////////////////////////////////////////////////////////////////////

#macro common_aal5_rx_dl_init()
.begin

.reg tmp_zero


	; Initialise Dispatch Loop variables to known state.
	alu[dl_next_block, --, B, BID_NULL]			// set dl_next_block NULL
	alu[dl_buf_handle, --, B, DL_BUF_NULL]     	// set dl_buffer_handle NULL
	alu[dl_eop_buf_handle, --, B, DL_BUF_NULL] 	// set dl_eop_buf_handle NULL 


	// In main code, the meta data is set only when it is required. So if we 
	// set it to other than zero value in init, then the value remains 
	// unchanged for the part that is not updated in the main code. So wrong 
	// value is written into SRAM META.
	dl_meta_init_cache(zero, zero, zero, zero, zero, zero, 0, 0)
.end
#endm  // common_aal5_rx_dl_init()
	

///////////////////////////////////////////////////////////////////////////////
//	common_aal5_rx_regs_init:
//
//		Description:
//			Initializes some registers used in aal5_rx().
//
//		Input: 
//			
//		Output: 
//
//		Size: N/A
//
///////////////////////////////////////////////////////////////////////////////

#macro common_aal5_rx_regs_init()
.begin
	.reg 	tmp


	move(mask_lower_half, 0xffff)
    move(zero, 0)

	immed32(vc_info_sram_base, AAL5_RX_VC_INFO_SRAM_BASE)  // Has both RXCs and VC Stat counters
	immed32(rbuf_base, RBUF_TBUF_BASE)						// Has RBUF base
	immed32(hash_pri_base_sram, AAL5_RX_VC_HASH_PRI_SRAM_BASE)		// Has Primary hash table base
	immed32(hash_sec_base_sram, AAL5_RX_VC_HASH_SEC_SRAM_BASE)		// Has Secondary hash table base
	immed32(port_stat_base, AAL5_RX_PORT_STAT_SRAM_BASE)   // Has Port Stat counters

	; Get the current ME number and ctx number.
	_aal5_rx_get_curr_me_ctx_num(me_num, ctx_num)
	; Setup the reg to add context onto the thread freelist.
	_aal5_rx_set_thd_freelist_reg(rx_thd_freelist_reg, me_num, ctx_num, &$rsw0, (&rsw_rcvd_sig))
	; for fast_wr, upper 16bits has data and lower 16 bits has address.
	alu[rx_thd_freelist_reg, --, B, rx_thd_freelist_reg, <<16]

	// setup next thread register
	alu[tmp, ctx_num, +, 2]								; calculate next thread number
	alu[tmp, tmp, AND, 7]								; make sure the thread number value is [0..7] 
	alu[tmp, tmp, OR, NEXT_THREAD_SIGNAL, <<3]			; [6:3] signal number
	alu[next_thread_sig_val, --, B, tmp]

	#if (RX_PHY_MODE == SPHY_4_8)

		// in this case we use only one freelist defined by RXTX_ATM_LINE
		// aal5 rx is performed by single ME
		alu[rx_freelist_no,--,B, RX_THREAD_FREELIST_/**/RXTX_ATM_LINE]

	#elif (RX_PHY_MODE == MPHY_8E_8A)

		alu[rx_freelist_no,--,B, AAL5_RX_THREAD_FREELIST]

	#else  // SPHY_1_32

		; For (TWO_ME_RX_FP, SPHY_1_32(default)) & 
		;     (ONE_ME_RX_FP, SPHY_1_32(default))
		alu[rx_freelist_no,--,B, RX_THREAD_FREELIST_0]

	#endif			

.end
#endm  // common_aal5_rx_regs_init()



/////////////////////////////////////////////////////////////////////////////////
//            Macro to initialize AAL5 RX
/////////////////////////////////////////////////////////////////////////////////
//	common_aal5_rx_init:
//
//		Description:
//			It does necessary initialization for AAL5 RX and the pipeline
//				1.  Registers init, local memory init with constant values
//				2.  RX related MSF registers init
//				3.  Does system init (scratch rings init, buffer_free_list creation, 
//					signal all MEs with the SYSTEM_INIT signal)
//				4.  Context, prefetching a buffer
//				5.  Context putting itself onto Thread freelist and waits for 
//					rsw_rcvd_sig
//
//		Inputs: 
//		
//		Outputs: 
//
//		Size: 
//			N/A
//
/////////////////////////////////////////////////////////////////////////////////						

#macro common_aal5_rx_init()

.begin

	.reg me_num										// current ME number
	.reg lm_addr									// LM addr to store reg values

	.sig bd_prefetch_sig							// signal for prefetching buffer handle

	
	; All aal rx threads would execute this part
	; Registers init

	; Init the LM handle 1 with the lm addr for storing register values
	immed32[lm_addr, LM_CONSTANTS_BASE]				
	alu_shf[lm_addr, --, b, lm_addr, <<2]
	local_csr_wr[ACTIVE_LM_ADDR_1, lm_addr]

	common_aal5_rx_regs_init()
		
	; Dispatch loop init
	common_aal5_rx_dl_init()

	.if(ctx() == 1)
		// Initialise the free buffer list for aal5 rx (upstream)
        // when Core Components are not used. Otherwise buffers will
        // be initialised by CC infrastructure.
        #ifdef DO_FREELIST_INIT
		    #ifdef FREELIST_INIT_VIA_SCRIPT

				// Because of design and using both DOWNSTREAM and UPSTREAM defines we 
				// have to make this ugly define here
				#define TEMPORARY_FREELIST_QDESC            UPSTREAM_TEMPORARY_FREELIST_QDESC

                _aal5_rx_read_freelist_qdesc()

				#undef TEMPORARY_FREELIST_QDESC

            #else
                // Create buffer list for upstream direction - only for the first thread
		        dl_buffer_init[]
            #endif
        #endif /* DO_FREELIST_INIT */
	.endif

	; All contexts just wait for a signal from their previous context 
	; Thread 1 will be signalled by Thread 0
	ctx_arb[next_thread_sig]	


	; Prefetch a buffer handle for the context
   	dl_buf_alloc($prefetch_buf_handle, BUF_FREE_LIST0, bd_prefetch_sig, bd_prefetch_sig)

	; Add thread to freelist.
	msf[fast_wr, --, rx_freelist_no, rx_thd_freelist_reg]	

	.if (ctx() == 7) 
		common_rx_config_msf_rx_ctrl()    				; trigger receive logic as last initalized context	
	.endif

	; signal the next context
	signal_ctx(next_thread_sig_val)												

	ctx_arb[rsw_rcvd_sig]
	
.end
#endm  // common_aal5_rx_init()

