/*******************************************************************************
                             Intel Proprietary

 Copyright (c) 2003 By Intel Corporation.  All rights reserved.
 No part of this program or publication may be reproduced, transmitted,
 transcribed, stored in a retrieval system, or translated into any language
 or computer language in any form or by any means, electronic, mechanical,
 magnetic, optical, chemical, manual, or otherwise, without the prior
 written permission of:
                         Intel Corporation
                         2200 Mission College Blvd.
                         Santa Clara, CA  95052-8119
*******************************************************************************/

/*
 * forwarder_pipeline.uc
 */

#ifndef FORWARDER_PIPELINE_UC
#define FORWARDER_PIPELINE_UC

#ifdef USE_IMPORT_VAR
    #include <dl_system.h>
#else
    #include <dl_system.h>
    #include <dl_mem_map.h>
#endif

#include <system_init.uc>
#include <dispatch_loop.uc>

#include <xbuf.uc>


// include ipv4_fwder_dl.uc from IPV4 forwarder
#include <ipv4_fwder_init.uc>

// include ipv4_fwder.uc from IPV4 forwarder
#include <ipv4_fwder.uc>

// include llc snap decapsulation
#include <llc_snap_decap.uc>
#include <llc_encap_util.uc>

// include ether decapsulation
#include <ether.uc>

#define  BID_NEXT_BLOCK   BID_QM	// block next to dispatch loop

// Scratch ring entry size in lw's.
#define SCR_RING_CACHE_SIZE 		3

//------------------------------------------------------------------------------
//    Signals

.sig ip_cache_sig
.set_sig ip_cache_sig

#ifdef	IPV4_START_ME

/* This signal is declared in system_init() macro */
.sig visible sys_init_signal

//	Global Variable for a Prev/Next thread signaling.
.sig write_sig_ip
.sig volatile	sig_prev
.addr 			sig_prev		INTER_THD_SIGNAL

#endif

#define  prev_thread_done_sig   sig_prev

;-----------------------------------------------------------
; Registers
;-----------------------------------------------------------

.reg dl_queue_num

;-----------------------------------------------------------
; AAL5 - IPV4 Macro
;-----------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////////////////
//
// ipv4_fwder_kick_start()
//
//	 	Description:
//
//		Kick starts IPV4 Forwarder blocks across the 4 MEs
//	(1) when the system initialization is complete.
//	(2) signal thread 0 on beginning ME after which the
//		thread order is maintained by DL macros.
//
//	All the MEs (and threads) running IPV4 Forwader call
//	this macro. Only thread0 of first ME would execute
//	this fully. Other threads of first ME and all the
//	threads on other MEs would continue and call the
//	dl_source[DL_THREAD_ORDER], and wait for signal
//	from previous thread. This would result in thread0
//	of first ME kickstarting the whole chain.
//
//
//	 	Outputs:
//			None.
//
//		Inputs:
//			None.
//
//		CONSTANTS
//			None.	
//
//		Size: 
//
//			?? instructions. (Worst case cycle count)
//
////////////////////////////////////////////////////////////////////////////////////////
#macro	ipv4_fwder_kick_start()

	// The IPV4 fwder code can run on single ME or multiple MEs. For
	// Point Reyes, we run this code on 4 MEs as a chain. These MEs
	// need to wait for the system initialization to complete which
	// is communicated by sending a signal. So, thread0 of the 
	// first ME waits for this signal. Once, this signal is received.
	// this thread kick starts the ME chain. (Actually any thread on
	// any ME can act as kick starter).

	// We need to know the ME number to get this right. If we pickup
	// different MEs, the following line has to be modified accordingly.
	// For Point Reyes, '0x1' is the first ME in the chain. 
#ifdef	IPV4_START_ME	

.begin
	br!=ctx[0,ipv4_fwder_kick_start_end#]

	// this is code for thread0

	// along with the system initialization signal, we assert a wakeup
	// signal so that very first time around, we don't need to wait. This
	// helps in kick starting the chain.
	signal_first_ctx[DL_SIG_WAKE]

	// we wait for the signal here
	ctx_arb[sys_init_signal]

	// by exiting this block, we free up the signal
.end

ipv4_fwder_kick_start_end#:

#endif	//IPV4_START_ME

	// here, initialize the thread ID.
#ifdef	IPV4_NEW_XBUF
.begin
	.reg	thread_id
	local_csr_rd[active_ctx_sts]
	immed[thread_id,0]
	alu[thread_id,0x7, AND, thread_id]
	dl_set_thread_id[thread_id]
.end
#endif

#endm

////////////////////////////////////////////////////////////////////////////////
//
// _aal5_ipv4_fwder
//
//	 	Description: 
//
//			Its a wrapper macro around  _ipv4_fwder() to satisfy the AAL5 
//			specific offset requirements.			
//
//	 	Outputs:
//			N/A
//		
//		Inputs:
//			ip_hdr_offset	-	the offset to point to the start of IP header
//								(If the frame has LLC/SNAP header, offset = 8,
//								 else if frame is not encapsulated, offset = 0)
//
//			__PACKETHDR		- 	buffer containing packet header.
//
//		Size: 
//		
///////////////////////////////////////////////////////////////////////////////

#macro _aal5_ipv4_fwder(ip_hdr_offset, __PACKETHDR)
.begin

		
	alu[--, ip_hdr_offset, -, 8]
	bne[hdr_offset_is_zero#], defer[3]
	// The macro copies only 20 bytes of IP header to the
	// destination register buffer. we need to copy the registers untouched
	// by the above macro. The registers modified is dependent on
	// where the IP Hdr starts in the buffer. 
	alu[__PACKETHDR[7],--,B,__PACKETHDR[7]]
	alu[__PACKETHDR[8],--,B,__PACKETHDR[8]]
	alu[__PACKETHDR[9],--,B,__PACKETHDR[9]]

	; Here the frame has LLC/SNAP header encapsulation.
	 _ipv4_fwder(__PKTHDR,__PKTHDR, IP_HDR_OFFSET_8, IP_HDR_OFFSET_8)

	br[end#], defer[2]	
	alu[__PACKETHDR[0],--,B,__PACKETHDR[0]]
	alu[__PACKETHDR[1],--,B,__PACKETHDR[1]]

hdr_offset_is_zero#:

	; Here the frame is not encapsulated.
	 _ipv4_fwder(__PKTHDR,__PKTHDR, IP_HDR_OFFSET_0, IP_HDR_OFFSET_0)

 	alu[__PACKETHDR[5],--,B,__PACKETHDR[5]]
	alu[__PACKETHDR[6],--,B,__PACKETHDR[6]]


end#:

.end
#endm

;-----------------------------------------------------------
; Initialize the blocks
;-----------------------------------------------------------
init#:

   dl_iphdr_cache_init[]
  .set $$iphdr0 $$iphdr1 $$iphdr2 $$iphdr3 $$iphdr4 $$iphdr5 $$iphdr6 $$iphdr7 $$iphdr8

#ifndef	__PKTHDR
#define	__PKTHDR	$$iphdr	; for convenience. DL defines this.
#else
#warning	"__PKTHDR is already defined."
#endif

#ifdef ETHER_WITHOUT_L2
    ; init l2   
#ifndef DISABLE_MAC_FILTERING
    _ether_init(SRAM_FLT_TABLE_BASE, SRAM_PORT_BIT_MAP_BASE)
#else
    ; in this case the L2 gprs are unused so set them to 0
    _ether_init(0, 0)
#endif
#endif 

	; init buffer meta data cache to Zero.
	init_meta_cache_to_zero[]

	; Temporary : Initialize EOP handle default.
	_dl_eop_buf_handle_init[]

	; ipv4_fwder initialization.
	_ipv4_fwder_init()

	; now kick start the IPV4 Forwarder.
    ipv4_fwder_kick_start()
   
#ifdef DOWNSTREAM

	; Defines SCR_RING_CACHE_SIZE gpr's.
	.reg src_sink_data[SCR_RING_CACHE_SIZE]
    .set src_sink_data[0], src_sink_data[1], src_sink_data[2]
    .reg next_thread

    ; Define signal mask gpr's.
    .reg sigmask_a
    .reg default_sigmask_a
    .reg sigmask_b
    .reg default_sigmask_b
 
    ; Initialize some gpr's that store constants.
 	; These gpr's save cycle in the critical path.
 	.reg OxFFFF
	move(OxFFFF, 0xFFFF)

	.reg OxE003FFFF
	move(OxE003FFFF, 0xE003FFFF)

	.reg drop_queue
	move(drop_queue, QM_DROP_QUEUE)

	.reg NEXT_HOP_TABLE_BASE_SRAM_GPR
	immed32(NEXT_HOP_TABLE_BASE_SRAM_GPR, NEXT_HOP_TABLE_SRAM_BASE) 

	; Allocate a register to increment eop buffer size when additional
	; ATM cell is detected.
	.reg write $eop_buf_size_incr 

  	; Allocate 4 lw's to write the 8-byte LLC/SNAP header along with
	; the first qw of the non 8-byte aligned start of payload.
	xbuf_alloc($$payload, 4, read_write)

    _llc_encap_init()

#endif

	.set dl_queue_num 

	//
	// Kick start the pipeline.
	//
.if (ctx() == 0)

#ifdef	IPV4_START_ME

		// The forwarder pipeline code runs on 2 MEs as a chain. These MEs
		// need to wait for the system initialization to complete which
		// is communicated by sending a signal. Thread0 of the
		// first ME waits for this signal. Once, this signal is received,
		// this thread kick starts the ME chain.

		.begin

			// Generate signal consumed by dl_source().
			signal_first_ctx(&sig_prev)

		.end

#endif	// IPV4_START_ME

.endif 

;-----------------------------------------------------------
; dispatch loop
;-----------------------------------------------------------
.begin
.reg sig_mask

	immed32[sig_mask, 0]

	; Get packets in order.
	dl_source[DL_THREAD_ORDER, sig_mask, __PKTHDR]

.end



loop#:

      .reg ip_hdr_offset_reg
      .set ip_hdr_offset_reg

#ifdef UPSTREAM
       .if (dl_next_block == BID_ADECAP)
       .begin
             ; Perform classification and decapsulation of headers accordingly.
             alu[ip_hdr_offset_reg, --, b, 0]

             _llc_snap_decap_classify(ip_hdr_offset_reg, __PKTHDR, AAL5_HDR_OFFSET)

             ;Process the IP header. The modified packet is in cache
             ;registers starting at ip_hdr_offset_reg.

             _aal5_ipv4_fwder(ip_hdr_offset_reg, __PKTHDR)

         	 ; write the IP header of the packet and cache it in some place
             .if (ip_hdr_offset_reg == 8)
#define_eval TRANSFER_REGISTER __PKTHDR/**/2/**/
             dl_write_hdr_cache[TRANSFER_REGISTER, dl_buf_handle, write_sig_ip, write_sig_ip] 
             .else
#define_eval TRANSFER_REGISTER __PKTHDR/**/0/**/
             dl_write_hdr_cache[TRANSFER_REGISTER, dl_buf_handle, write_sig_ip, write_sig_ip] 
             .endif

       .end
       .endif
#endif

#ifdef DOWNSTREAM
       .if (dl_next_block == BID_EDECAP)
       .begin
             ; remove warnings
             .io_completed eop_meta_wr_sig

#ifdef ETHER_WITHOUT_L2
             _ether_decap_classify(__PKTHDR,ETHER_HEADER_OFFSET)
#endif
             ;process the IP header. The modified packet is in cache
             ;registers starting at IP_HDR_OFFSET.

             _ipv4_fwder(__PKTHDR,__PKTHDR,IP_HDR_OFFSET,IP_HDR_OFFSET) 

         	 ; write the IP header of the packet and cache it in some place
             .io_completed dram_wr_sig

#define_eval TRANSFER_REGISTER __PKTHDR/**/4/**/
             dl_write_hdr_cache[TRANSFER_REGISTER, dl_buf_handle, write_sig_ip, write_sig_ip] 

             alu[src_sink_data[0], --, B, dl_buf_handle]
             alu[src_sink_data[1], --, B, dl_eop_buf_handle]
             alu[src_sink_data[2], --, B, dl_queue_num]

             _llc_encap()

       .end
       .endif
#endif


sink_the_packet#:


    .begin
		.reg sig_mask
		.reg write $buf_size_incr
		.sig sig_scr_put sig_meta
		.set_sig sig_scr_put
		.set_sig sig_meta

		// Allocate xfer registers.
		xbuf_alloc[$wxfer, FWD_TO_QM_MSG_SIZE, write]
		immed32[sig_mask, 0]


		// Send packet.
 		dl_qm_sink(sig_mask, sig_scr_put, sig_meta, $wxfer, $buf_size_incr, __PKTHDR)
 
		// Receive packet.
		dl_source(DL_THREAD_NO_ORDER, sig_mask, __PKTHDR)

		.io_completed sig_scr_put
		.io_completed $buf_size_incr

		// Free allocated xfer registers.
		xbuf_free($wxfer)

    .end

	; start again.
	br[loop#]	


term#:


   dl_iphdr_cache_fini[]

kill#:
	nop
	nop

#endif // FORWARDER_PIPELINE_UC
/*******************************  FILE HISTORY  *******************************
 * \main\1 Wed Aug 13 14:03:16 2003 adziecia
 * IGK0100025024: ATM-IPV4-ETH: simulation project check-in 
 *  
 *****************************************************************************/
