/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      xscale_me_init.c
 *
 * = DESCRIPTION
 *      This file contains functions to perform the Xscale initialization 
 *      for 10Gb Ethernet application. This application uses Resource
 *      Manager functions to patch symbols and to download the code to
 * 	microengines.
 *
 * = AUTHOR
 *	ddt
 *
 * = CHANGE HISTORY
 *      04/13/03 	Created.for 4Gb Ethernet (IXP2400)
 *	09/09/03	Updated for 10Gb Ethernet (IXP2800)
 * ============================================================================
 */

/*
 *  Include files
 *	-------------
 */

#include	"xscale_me_init.h"
#include	"xscale_me_rm_app.h"
#include	"xscale_common.h"
/*
 *  some global macro definitions
 */
void *VIRT_SRAM_CH0_BASE;
void *VIRT_SRAM_CH1_BASE;
void *VIRT_SRAM_CH2_BASE;
void *VIRT_SRAM_CH3_BASE;
void *VIRT_DRAM_BASE;
void *VIRT_SCRATCH_BASE;
void *VIRT_MSF_CSR_BASE;
void *VIRT_MSF_RBUF_BASE;
void *VIRT_CAP_CSR_GLOBAL_BASE;
void *VIRT_CAP_CSR_SCRATCH_CONFIG;
void *VIRT_XSCALE_LOCAL_CSR_BASE;

/*
 * Include ingress and egress system setup files
 */
#include "ingress_init.c"
#include "egress_init.c"

/*
 * Include the Routes file
 */
#include "rtm_routes.c"

/*
 * Obtain the memory map
 */
void	xscale_me_mmap_init(void)
{

	DBG_MSG("\nGetting memory mapping ...\n");

	VIRT_SRAM_CH0_BASE = (void *)(ix_uint32)Hal_sram_ch0_virtAddr;
	VIRT_SRAM_CH1_BASE = (void *)(ix_uint32)Hal_sram_ch1_virtAddr;
	VIRT_SRAM_CH2_BASE = (void *)(ix_uint32)Hal_sram_ch2_virtAddr;
	VIRT_SRAM_CH3_BASE = (void *)(ix_uint32)Hal_sram_ch3_virtAddr;

	VIRT_DRAM_BASE = (void *)(ix_uint32)Hal_dram_ch0_virtAddr;
	VIRT_SCRATCH_BASE = (void *)(ix_uint32)Hal_scratch_rd_wr_swap_virtAddr;
	VIRT_MSF_CSR_BASE = (void *)(ix_uint32)Hal_msf_virtAddr;
	VIRT_MSF_RBUF_BASE = (void *)((unsigned char*)VIRT_MSF_CSR_BASE + 0x2000);
	VIRT_CAP_CSR_GLOBAL_BASE = (void *)
			(((unsigned char*)((ix_uint32)Hal_cap_global_ctl_csr_virtAddr)) + 0xa00);
	VIRT_CAP_CSR_SCRATCH_CONFIG = (void *)
			(((unsigned char*)((ix_uint32)Hal_cap_global_ctl_csr_virtAddr)) + 0x800);
	VIRT_XSCALE_LOCAL_CSR_BASE = (void *)((ix_uint32)Hal_xscale_local_intctl_virtAddr);

}


/*
 * init_module for the Xscale initialization code
 * ----------------------------------------------
 */
int	system_setup(char *chip)
{

	ix_error err;

	if (!chip)
	{
		DBG_MSG("ERROR: USAGE 'system_setup(\"ingress/egress\").\n");
		return -1;
	}

	/*
	 * Init RM
	 */
	DBG_MSG("%s : Initializing RM ...\n",chip);
    	err = ix_rm_init(0);
    	if(err != IX_SUCCESS)
	{
		DBG_MSG("\n%s : RM Initialization FAILED.\n",chip);
        	return err;
	}

	/*
	 * Get Memory Map
	 */
	xscale_me_mmap_init();

	/*
	 * Xcale-ME Application Initialization
	 */
	if(!strcmp(chip,"ingress"))
	{
		/*
		 * perform Ingress initialization
		 */

		if(xscale_me_rm_app_init("ingress"))
			return -1;

		ingress_system_setup();

		rm_app_start_ingress_me();

	}
	else if(!strcmp(chip,"egress"))
	{
		/*
		 * perform Egress initialization
		 */

		if(xscale_me_rm_app_init("egress"))
			return -1;

		egress_system_setup();

		rm_app_start_egress_me();

	}
	else
	{
		DBG_MSG("ERROR: USAGE 'system_setup(\"ingress/egress\").\n");

		return -1;
	}


	DBG_MSG("\n@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
	DBG_MSG("        %s Initialization complete.         \n",chip);
	DBG_MSG("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n\n");
	return 0;

}

void system_clean(char *chip)
{
	if(xscale_me_rm_app_fini(chip))
	    DBG_MSG("Failed to clean up %s\n",chip);	

	DBG_MSG("\n@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
	DBG_MSG("        %s BYE BYE !!!         \n",chip);
	DBG_MSG("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n\n");

	return;
}

/* Shorter function names */
void iss()
{
	ix_error err;
	err = system_setup("ingress");
    		
  	return;
}

void ess()
{
	ix_error err;
	err = system_setup("egress");
    		
  	return;
}

void isc()
{
	system_clean("ingress");
	return;
}

void esc()
{
	system_clean("egress");
	return;
}


/* ========================= UTILITIES ================================================== */

/**
 * NAME: read_ipv4_counters
 *
 * DESCRIPTION: This function reads IPv4 counters 
 */
int read_ipv4_counters(void)
{

	unsigned char *base = IPV4_STATS_TABLE_BASE;

	DBG_MSG("\n=====================  IPv4 Counters for all 4 ports ==================\n");

	DBG_MSG("RX       =%12d %12d %12d %12d \n",
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_RX  ) ),
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_RX  + 64  ) ), 
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_RX  + 128  ) ),
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_RX  + 192  ) ) ) ;


	DBG_MSG("FWD      =%12d %12d %12d %12d \n",
			( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_FWD  ) ),
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_FWD  + 64  ) ), 
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_FWD  + 128  ) ),
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_FWD  + 192  ) ) ) ;


	DBG_MSG("DROP     =%12d %12d %12d %12d \n",
			( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_DROP  ) ),
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_DROP  + 64  ) ), 
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_DROP  + 128  ) ),
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_DROP  + 192  ) ) ) ;



	DBG_MSG("EXCP     =%12d %12d %12d %12d \n",
			( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_EXCP  ) ),
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_EXCP  + 64  ) ), 
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_EXCP  + 128  ) ),
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_EXCP  + 192  ) ) ) ;


	DBG_MSG("BAD_HDR  =%12d %12d %12d %12d \n",
			( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_BAD_HDR  ) ),
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_BAD_HDR  + 64  ) ), 
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_BAD_HDR  + 128  ) ),
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_BAD_HDR  + 192  ) ) ) ;


	DBG_MSG("BAD_LEN  =%12d %12d %12d %12d \n",
			( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_BAD_LEN  ) ),
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_BAD_LEN  + 64  ) ), 
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_BAD_LEN  + 128  ) ),
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_BAD_LEN  + 192  ) ) ) ;


	DBG_MSG("BAD_TTL  =%12d %12d %12d %12d \n",
			( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_BAD_TTL  ) ),
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_BAD_TTL  + 64  ) ), 
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_BAD_TTL  + 128  ) ),
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_BAD_TTL  + 192  ) ) ) ;


	DBG_MSG("NO_ROUTE =%12d %12d %12d %12d \n",
			( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_NO_ROUTE  ) ),
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_NO_ROUTE  + 64  ) ), 
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_NO_ROUTE  + 128  ) ),
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_NO_ROUTE  + 192  ) ) ) ;


	DBG_MSG("TOO_SMALL=%12d %12d %12d %12d \n",
			( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_TOO_SMALL  ) ),
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_TOO_SMALL  + 64  ) ), 
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_TOO_SMALL  + 128  ) ),
		   ( unsigned int) *(( unsigned int *) (base + OFFSET_PKTS_TOO_SMALL  + 192  ) ) ) ;

	
	return 0;
}


/**
 * NAME: read_rbuf
 *
 * DESCRIPTION: This function reads rbuf given the element size and the element number
 *              By default it will read element 0
 */
int read_rbuf( unsigned int *element_num, unsigned int rbuf_size)
{

	int i ;

	unsigned int *rbuf_base, *rbuf_add ;

	if( rbuf_size  <= 64  ) 
	{
		DBG_MSG("Assuming rbuf_size to be 128\n") ;
		rbuf_size = 128 ;
	}

	rbuf_base = (unsigned int *) MSF_RBUF_BASE ;

	rbuf_add = ( unsigned int * ) ( ( unsigned int ) rbuf_base + ( ((unsigned int ) element_num ) * rbuf_size));

	DBG_MSG("Printing the Rbuf data at address 0x%x......\n", (unsigned int) rbuf_add);

	for(i=0; i<32; i++ )
	{
		DBG_MSG("0x%x\n", *(rbuf_add) );
		rbuf_add += 1 ;
	} 

	return 0 ;
}

/**
 * NAME: read_scratch_full_reg
 *
 * DESCRIPTION: This function reads the scratch full CSR register
 */
int read_scratch_full_reg(void)
{ 
	int i;
	unsigned int *p = (unsigned int*)CAP_CSR_SCRATCH;
	
	DBG_MSG("Scratch ring full register value is = 0x%x\n", GET32(SCRATCH_RING_STATUS)) ;

	/*
	 * also dump, for each ring, the base address, size, head ptr, tail ptr
	 * - all in byte address
	 */
	DBG_MSG("\t--- All are byte addresses. Size in bytes ---\n") ;
	for(i=0;i<16;i++)
	{
		unsigned int base,head,tail;
		int size;
		/* for all 16 scratch rings */
		switch(p[0]>>30)
		{
			case 0:
				size = 128; break;
			case 1:
				size = 256; break;
			case 2:
				size = 512; break;
			case 3:
				size = 1024; break;
		}

		base = (unsigned int)SCRATCH_BASE + (p[0] & 0x3fff);
		head = base + (p[1] & 0xfff);
		tail = base + (p[2] & 0xfff);

		DBG_MSG("Ring= %2d Size= %4d Base= 0x%.8x Hptr= 0x%.8x Tptr= 0x%.8x\n",
						i,size*(sizeof(unsigned int)),base,head,tail);
		p += 4;	/* ring control reg */
	}
	return 0 ;

}

/**
 * NAME: clear_ipv4_counters
 *
 * DESCRIPTION: This function clears the ipv4 counters
 */
int clear_ipv4_counters(void)
{
		unsigned char* base = IPV4_STATS_TABLE_BASE;

        *(( unsigned int *) (base + OFFSET_PKTS_RX      ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_FWD     ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_DROP    ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_EXCP    ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_BAD_HDR ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_BAD_LEN ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_BAD_TTL ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_NO_ROUTE) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_TOO_SMALL) ) = 0  ;



        *(( unsigned int *) (base + OFFSET_PKTS_RX        + 64 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_FWD       + 64 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_DROP      + 64 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_EXCP      + 64 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_BAD_HDR   + 64 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_BAD_LEN   + 64 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_BAD_TTL   + 64 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_NO_ROUTE  + 64 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_TOO_SMALL + 64 ) )  = 0 ;


        *(( unsigned int *) (base + OFFSET_PKTS_RX        + 128 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_FWD       + 128 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_DROP      + 128 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_EXCP      + 128 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_BAD_HDR   + 128 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_BAD_LEN   + 128 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_BAD_TTL   + 128 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_NO_ROUTE  + 128 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_TOO_SMALL + 128 ) )  = 0 ;


 	*(( unsigned int *) (base + OFFSET_PKTS_RX        + 192 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_FWD       + 192 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_DROP      + 192 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_EXCP      + 192 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_BAD_HDR   + 192 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_BAD_LEN   + 192 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_BAD_TTL   + 192 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_NO_ROUTE  + 192 ) )  = 0 ;
        *(( unsigned int *) (base + OFFSET_PKTS_TOO_SMALL + 192 ) )  = 0 ;


        DBG_MSG("IPV4 Counters Reset for all 4 ports \n") ;
 
        return 0 ;

}


void	msf_print_status(void)
{
	unsigned int temp;

	temp = GET32(MSF_RX_CTRL);
	DBG_MSG("MSF : MSF RX CTRL is 0x%x\n",temp);
	
	temp = GET32(MSF_TX_CTRL);
	DBG_MSG("MSF : MSF TX CTRL is 0x%x\n",temp);

	temp = GET32(MSF_RX_UP_CTRL_0);
	DBG_MSG("MSF : Port '0' RX UP CTRL is 0x%x\n",temp);
	temp = GET32(MSF_RX_UP_CTRL_1);
	DBG_MSG("MSF : Port '1' RX UP CTRL is 0x%x\n",temp);
	temp = GET32(MSF_RX_UP_CTRL_2);
	DBG_MSG("MSF : Port '2' RX UP CTRL is 0x%x\n",temp);
	temp = GET32(MSF_RX_UP_CTRL_3);
	DBG_MSG("MSF : Port '3' RX UP CTRL is 0x%x\n",temp);
	
	temp = GET32(MSF_TX_UP_CTRL_0);
	DBG_MSG("MSF : Port '0' TX UP CTRL is 0x%x\n",temp);
	temp = GET32(MSF_TX_UP_CTRL_1);
	DBG_MSG("MSF : Port '1' TX UP CTRL is 0x%x\n",temp);
	temp = GET32(MSF_TX_UP_CTRL_2);
	DBG_MSG("MSF : Port '2' TX UP CTRL is 0x%x\n",temp);
	temp = GET32(MSF_TX_UP_CTRL_3);
	DBG_MSG("MSF : Port '3' TX UP CTRL is 0x%x\n",temp);


	temp = GET32(MSF_INTR_STATUS);
	DBG_MSG("MSF : %d mpkts lost 'cos of no RBUFs\n",(temp>>8)&0xf);
	if(temp & 0x10)
		DBG_MSG("MSF : TCW programming error.\n");
	if(temp & 0x2)
		DBG_MSG("MSF : Incorrect vertical parity on CFrame(CSIX mode only)\n");
	if(temp & 0x1)
		DBG_MSG("MSF : Incorrect horizontal parity(all modes)\n");

	temp = GET32(MSF_TX_SEQ_0);
	DBG_MSG("MSF : TBUF '0' partition seq num is %d \n", temp & 0xff);
	if (temp & 0x80000000)
		DBG_MSG("MSF : TBUF '0' partition is empty (no valid elements)\n");
	else
		DBG_MSG("MSF : TBUF '0' partition is not-empty (some valid elements)\n");

	temp = GET32(MSF_TX_SEQ_1);
	DBG_MSG("MSF : TBUF '1' partition seq num is %d \n", temp & 0xff);
	if (temp & 0x80000000)
		DBG_MSG("MSF : TBUF '1' partition is empty (no valid elements)\n");
	else
		DBG_MSG("MSF : TBUF '1' partition is not-empty (some valid elements)\n");

	temp = GET32(MSF_TX_SEQ_2);
	DBG_MSG("MSF : TBUF '2' partition seq num is %d \n", temp & 0xff);
	if (temp & 0x80000000)
		DBG_MSG("MSF : TBUF '2' partition is empty (no valid elements)\n");
	else
		DBG_MSG("MSF : TBUF '2' partition is not-empty (some valid elements)\n");

	temp = GET32(MSF_TX_SEQ_3);
	DBG_MSG("MSF : TBUF '3' partition seq num is %d \n", temp & 0xff);
	if (temp & 0x80000000)
		DBG_MSG("MSF : TBUF '3' partition is empty (no valid elements)\n");
	else
		DBG_MSG("MSF : TBUF '3' partition is not-empty (some valid elements)\n");

}


static unsigned char dump_page[4096];

void dump_memory(void *begin, int nbytes)
{
	int nl_words;
	int cnt;
	unsigned int *p = (unsigned int*)begin;
	unsigned char *buf=dump_page;

	/* somehow, I couldn't dump more than 256 bytes */
	if (nbytes > 256)
		nbytes = 256;

    nl_words = (nbytes + 0x3)/sizeof(long);
	DBG_MSG("\nDumping %d bytes from 0x%x.\n",
			(nl_words)*sizeof(long),(unsigned int)begin);

	for(cnt=0;cnt<nl_words;cnt++)
	{
		if((cnt%4) == 0)
			buf += sprintf(buf,"\n0x%8x : \t",(unsigned int)p);

		buf += sprintf(buf,"%.8x ",*p++);
	}

	DBG_MSG("%s\n\n",dump_page);

}

