/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      ingress_init.c
 *
 * = DESCRIPTION
 *      This file contains scripts to initialize the IXP2800 Ingress processor.
 *
 * = AUTHOR
 *      Ameya Varde
 *      ameya.s.varde@intel.com
 *
 * = CHANGE HISTORY
 *      03/11/2003 - Created 
 *	10/03/2003 - Updated for ipv4/mpls
 * ============================================================================
 */

#define	IXP_SCRIPTS
#define	IXP_XSCALE

#include "xscale_common.h"
#include "..\ingress\wbench_project\dispatch_loop\dl_system.h"
#include "..\..\..\..\library\microblocks_library\include\dl_ingress_system_default_2800.h"

#define log2(x)                         2

void init_sram(int init_value, int start_addr, int end_addr)
{
	int addr;
	for(addr=start_addr; addr<end_addr; addr+=4)
		set_sram(addr, init_value);
}

void init_scratch(int init_value, int start_addr, int end_addr)
{
	int addr;
	for(addr=start_addr; addr<end_addr; addr+=4)
		set_scratch(addr, init_value);
}


/*----------------------
 *	IPv4
 *----------------------
 */

int rtm_init()
{	
	int cnt, index, finaladdr;

	printf("Initializing SRAM trie table area...\n");

	cnt = 64 * 1024 + 256 + SIZE_OF_TRIE_SET * NUM_TRIE_SETS;
	finaladdr = TRIE_TABLE_SRAM_BASE + (cnt * 4);
	init_sram(0x0, TRIE_TABLE_SRAM_BASE, finaladdr);

	printf("Initialization complete.\n");

	printf("Initializing SRAM next hop table area...\n");

	cnt = NUM_NEXTHOPS * SIZE_OF_NEXTHOP_LW * 4;
	finaladdr = NEXTHOP_TABLE_SRAM_BASE + cnt;
	init_sram(0x0,NEXTHOP_TABLE_SRAM_BASE, finaladdr);

	printf("Initialization complete.\n");

	return 0;
}


int clear_ipv4_scratch_counters() {

	int finaladdr;

	printf("Initializing scratch IPv4 counters area...\n");

	finaladdr = IPV4_STATS_TABLE_BASE + IPV4_STATS_TABLE_SIZE;
	init_scratch(0x0, IPV4_STATS_TABLE_BASE, finaladdr);

	printf("Initialization complete.\n");

	return 0;
}


int dbcast_init() {

	int size, finaladdr;

	printf("Initializing SRAM directed broadcast table area...\n");

	size = DBCAST_NUM_BLOCKS * DBCAST_TABLE_BLOCK_SIZE * 4;
	finaladdr = DBCAST_TABLE_SRAM_BASE + size;
	init_sram(0x0, DBCAST_TABLE_SRAM_BASE, finaladdr);

	printf("Initialization complete.\n");

	return 0;
}


/*	This function adds an address to dbcast table. 
//	NOTE : Add utmost DBCAST_TABLE_BLOCK_SIZE-1 entries
//  per block. Each block looks as follows.
//
//	 ---------------  BASE
//	|				|
//	|---------------
//	|	x			|
//	 ---------------
//	|	 			|
//	|---------------
//	|<Next addr>	|-> contains the full address of the next block
//	 ---------------
//	|				|
//	|---------------
//	|				|
//	 ---------------
//
*/
int	dbcast_add(int ip_addr) {

	int hash_result, base, index, value;

	/* Given an IP address of form A.B.C.D produce
	 * a hash_result of form A.A^B.A^B^C.A^B^C^D	
	 */

	hash_result = (ip_addr) ^ (ip_addr >> 8) ^ (ip_addr >> 16) ^ (ip_addr>>24);

	/* Get the key which is 8 bits */
	hash_result = hash_result & 0xff;

	/* Calculate the base */
	base = DBCAST_TABLE_SRAM_BASE + (hash_result << log2(DBCAST_TABLE_BLOCK_SIZE)) * 4;

	for (index=0;index<DBCAST_TABLE_BLOCK_SIZE-1;index++)
	{
		value = get_sram(base+(index*4));
		if (value)
			continue;
		else
		{
			set_sram(base+(index*4),ip_addr);
			printf("Added 0x%x to DBCAST table at 0x%x\n",
				ip_addr,base+(index*4));
			break;
		}
	}

	if (index == (DBCAST_TABLE_BLOCK_SIZE-1))
		printf("Dbcast table is full. Can't add 0x%x\n",ip_addr);

	return 0;
}

/*	========================================================================
//	The following functions implement the DBcast processing in Local memory. 
//	The table in SRAM is [periodically] copied into the Microengine Local
//	memory, and is used for lookup. To simplify the cacheing Dbcast table, a
//	different layout is used. We will have two tables.
//
//		Entry table	:	LM_DBCAST_BLOCK_COUNT blocks, each of size
//				LM_DBCAST_BLOCK_SIZE contain the DBcast addresses. The 
//				LM_DBCAST_BLOCK_SIZE is in bytes (should be multiple of 4, 
//				and power of 2).
//		Index table	:	Before the Entry table is this table. Size of this
//				table is LM_DBCAST_BLOCK_COUNT / 4  long words. This table
//				contains the next pointers.
//		
//	Total memory required = LM_DBCAST_INDEX_TABLE_SIZE  +  
//			(LM_DBCAST_BLOCK_COUNT * LM_DBCAST_BLOCK_SIZE) in longwords.
//
//	total memory required should be <=640 LW. See ipv4.uc file for further
//	details.
//	========================================================================
*/
int lm_dbcast_init() {

	int size, finaladdr;

	printf("Initializing SRAM local memory - directed broadcast table area...\n");

	size = (LM_DBCAST_BLOCK_COUNT * LM_DBCAST_BLOCK_SIZE) + 
		LM_DBCAST_BLOCK_COUNT/4;	/* in in lw */
	size = size * 4;	/* in bytes */

	finaladdr = DBCAST_TABLE_SRAM_BASE + size;
	init_sram(0x0, DBCAST_TABLE_SRAM_BASE, finaladdr);

	/*
	// now for first 16 blocks, add links to resolve collisions. Ideally, this
	// would be done dynamically on Xscale, but for our purpose, this is OK. 
	// The collision blocks have indices starting from 16 till 31.
	*/
	set_sram(DBCAST_TABLE_SRAM_BASE, 0x13121110);
	set_sram(DBCAST_TABLE_SRAM_BASE + 4, 0x17161514);
	set_sram(DBCAST_TABLE_SRAM_BASE + 8, 0x1b1a1918);
	set_sram(DBCAST_TABLE_SRAM_BASE + 12, 0x1f1e1d1c);

	/* we will keep 0xff as invalid value for index. */
	set_sram(DBCAST_TABLE_SRAM_BASE + 16, 0xffffffff);
	set_sram(DBCAST_TABLE_SRAM_BASE + 20, 0xffffffff);
	set_sram(DBCAST_TABLE_SRAM_BASE + 24, 0xffffffff);
	set_sram(DBCAST_TABLE_SRAM_BASE + 28, 0xffffffff);
	
	return 0;
}

/*
//	This function adds an address to dbcast table. 
*/
int	lm_dbcast_add(int ip_addr) {

	int hash_result, entry_table_base, entry_base;
	int	cnt, value, entry_num;
	int mask;

	/* construct mask */
	mask = ~(~0 << log2(LM_DBCAST_BLOCK_COUNT/2));

	/* Given an IP address of form A.B.C.D produce
	 * a hash_result of form A.A^B.A^B^C.A^B^C^D	
	 */
	hash_result = (ip_addr) ^ (ip_addr >> 8) ^ (ip_addr >> 16) ^ (ip_addr>>24);

	/* Get the key which is 8 bits */
	hash_result = hash_result & mask;

	/* Calculate the base address in bytes */
	entry_table_base = DBCAST_TABLE_SRAM_BASE + LM_DBCAST_BLOCK_COUNT;

	entry_num = hash_result;

	while(entry_num != 0xff)
	{

		int tmp;

		tmp = (entry_num * LM_DBCAST_BLOCK_SIZE);
		entry_base = entry_table_base + (tmp * 4);

		for(cnt=0; cnt < LM_DBCAST_BLOCK_SIZE; cnt++)
		{
			value = get_sram(entry_base+(cnt*4));
			if (value)
				continue;		/* entry is occupied. */
			else
			{
				set_sram(entry_base+(cnt*4),ip_addr);
				printf("Added 0x%x to DBCAST table at 0x%x, entry = %d, hash = %d\n",
					ip_addr,entry_base+(cnt*4), entry_num, hash_result);
				return 1;
			}
		}

		/*
		// calculate next index. The value is in this format
		//	+------+-----+-----+------+
		//	| 3	   |  2  |  1  |  0   |
		//	+------+-----+-----+------+
		*/
		tmp = get_sram(DBCAST_TABLE_SRAM_BASE + (entry_num &~ 0x3));
		entry_num = (tmp >> (8 * (entry_num & 0x3))) & 0xff;
		
	};

	if (entry_num == 0xff)
		printf("Dbcast table is full. Can't add 0x%x\n",ip_addr);

	return 1;
}


/*----------------------
 *	QM
 *----------------------
 */
int ingress_qm_init() {
	
	int finaladdr;

	printf("Initializing SRAM queue descriptor area...\n");

	finaladdr = QD_SRAM_BASE + ((QD_TOTAL+2)*QD_SIZE);
	init_sram(0x0, QD_SRAM_BASE, finaladdr);

	printf("Initialization complete.\n");

	return 0;
}


void ingress_system_setup()
{
	int finaladdr;

	printf("\n******************************************************\n");
	printf("\n     System setup begin ...                           \n");
	printf("\n******************************************************\n");

	/*----------------------
	 *	Packet RX
	 *----------------------
	 */

	/*----------------------
	 *	IPv4
	 *----------------------
	 */

	/* Clear trie table and next hop area in SRAM. */
	rtm_init();

#ifdef LM_DBCAST_TABLE

	lm_dbcast_init();

#else

	dbcast_init();

#endif

	/* Clear IPv4 counters. */
	clear_ipv4_scratch_counters();

	/*----------------------
	 *	QM
	 *----------------------
	 */

	/* Initialize queue descriptor area in SRAM. */
	ingress_qm_init();

	printf("Adding some routes...\n");
	l3_create_routes();

#ifdef LM_DBCAST_TABLE

	/* Add dbcast entries using addresses from rtm_routes.ind file
	 * (take the destination IP address for which the 'LOCAL' flag is set,
	 * and then add corresponding network broadcast address for that
	 * network in the table).
	 */

	printf("Adding some dbcast entries...\n");
	lm_dbcast_add(0x1affffff);
	lm_dbcast_add(0x1a86ffff);
	lm_dbcast_add(0x1a877fff);
	lm_dbcast_add(0x1887ffff);
	lm_dbcast_add(0x18870aff);
	lm_dbcast_add(0x18870a00);

#else

	/* Add dbcast entries using addresses from rtm_routes.ind file
	 * (take the destination IP address for which the 'LOCAL' flag is set,
	 * and then add corresponding network broadcast address for that
	 * network in the table).
	 */

	printf("Adding some dbcast entries...\n");
	dbcast_add(0x1affffff);
	dbcast_add(0x1a86ffff);
	dbcast_add(0x1a877fff);
	dbcast_add(0x1887ffff);
	dbcast_add(0x18870aff);
	dbcast_add(0x18870a00);

#endif

	/* Set up the control block to enable only 0-31 ports for forwarding. */
	/* Control block is currently unused. */
	/* set_scratch(CONTROL_BLOCK_SCRATCH_BASE, 0xffffffff); */

	/*----------------------
	 *	MPLS
	 *----------------------
	 */
	printf("Initializing MPLS tables ...\n");
	mpls_init();



	/*----------------------
	 *	CSIX TX
	 *----------------------
	 */

	/*----------------------
	 *	COMMON
	 *----------------------
	 */

	/* If needed, set CSR to select DRAM address bits that
	 * decide DRAM bank scheduling.
	 * dc.d_0.DPL_DRAM_CTRL_RDH = 0x00021100;
	 * dc.d_1.DPL_DRAM_CTRL_RDH = 0x00021100;
	 * dc.d_2.DPL_DRAM_CTRL_RDH = 0x00021100;
	 */

	/* Initialize SRAM meta-data area. */
	finaladdr = BUF_SRAM_BASE+BUF_SRAM_SIZE;
	init_sram(0x0, BUF_SRAM_BASE, finaladdr);

	printf("\n*********************************************************\n");
	printf("\n     System setup complete.                              \n");
	printf("\n*********************************************************\n");

}


/* Shorter function names */
void iss()
{	
    ingress_system_setup();
}
