/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2400 Network Processor
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      ATM SAR Agent Core Component
 *
 * = FILENAME
 *      ix_cc_atmsar_agent.h
 *
 * = DESCRIPTION
 *      This file defines all data types and function prototypes required
 *      internaly by the ATM SAR Agent CC module.
 *
 * = AUTHOR
 *       Jaroslaw Kogut
 *       jaroslaw.kogut@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 * 
 */

#if !defined(__IX_CC_ATMSAR_AGENT_PRIVATE_H__)
#define __IX_CC_ATMSAR_AGENT_PRIVATE_H__

#include "cc/internal/ix_cc_atmsar_msg.h"
#include "cc/internal/ix_cc_atmsar_common.h"
#include "cc/ix_cc_atmsar_plugin.h"

#if defined(__cplusplus)
extern "C" {
#endif /* end defined(__cplusplus) */

/* Flag fornot to send report to msg support lib*/
#define IX_CC_ATMSAR_AGENT_SEND_NO_RESPONSE 1 


/**
 * TYPENAME: _ix_cc_atmsar_agent_context_t
 * 
 * DESCRIPTION:
 *    This structure defines format for ATM SAR AGENT core component
 *    context.
 */
typedef struct ix_s_cc_atmsar_agent_context
{
    ix_cc_handle hAtmsar_AgentHandle;      /* Handle of ATM SAR agent CC      */
    ix_uint16 subCoreInitialized;  /* the flag is set to 1 if all SoftSar core*/
    /* subcomponents are registered into ATM SAR MAIN and AGENT. Otherwise,   */
    /* it is equal 0.                                                         */
    
    _ix_cc_atmsar_engine_t engine;

} _ix_cc_atmsar_agent_context_t;


/******************************************************************************
 *                         GLOBAL VARIABLES
 *****************************************************************************/

/******************************************************************************
 *          INTERNAL FUNCTIONS - FUNCTION FOR ASYNC MSG PROCESSING
 *****************************************************************************/


/**
 * NAME: ix_cc_atmsar_agent_vc_create
 *
 * DESCRIPTION:
 *    The function adds VC into ATM SAR Agent plugins.
 * 
 * @Param:  IN  arg_pContext : pointer to ATM SAR Agent CC
 * @Param:  IN  arg_hVc : handle to VC to be created
 * @Param:  IN  arg_pVcDescr : parameters describing requested VC.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 */
ix_error ix_cc_atmsar_agent_vc_create(
void *arg_pContext,
const ix_cc_atmsar_vc_handle_t arg_hVc,
const ix_cc_atmsar_vc_descr_t* arg_pVcDescr);


/**
 * NAME: ix_cc_atmsar_agent_vc_update
 *
 * DESCRIPTION:
 *    The function updates VC in ATM SAR Agent plugins.
 * 
 * @Param:  IN  arg_pContext : pointer to ATM SAR Agent CC
 * @Param:  IN  arg_hVc : handle to VC to be created
 * @Param:  IN  arg_pUpdate : parameters describing requested VC.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 */
ix_error ix_cc_atmsar_agent_vc_update(
void *arg_pContext,
const ix_cc_atmsar_vc_handle_t arg_hVc,
const ix_cc_atmsar_update_data_t* arg_pUpdate);

/**
 * NAME: ix_cc_atmsar_agent_vc_remove
 *
 * DESCRIPTION:
 *    The function removes VC from ATM SAR Agent plugins.
 * 
 * @Param:  IN  arg_pContext : pointer to ATM SAR Agent CC
 * @Param:  IN  arg_hVc : handle to VC to be removed
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error ix_cc_atmsar_agent_vc_remove(
void *arg_pContext,
const ix_cc_atmsar_vc_handle_t arg_hVc);


/**
 * NAME: ix_cc_atmsar_agent_vc_stats
 *
 * DESCRIPTION:
 *    The function reads statistic for VC from ATM SAR Agent plugins.
 * 
 * @Param:  IN  arg_pContext : pointer to ATM SAR Agent CC
 * @Param:  IN  arg_hVc : handle to VC to be removed
 * @Param:  OUT arg_pLookupResult; pointer to buffer where VC statistic has to
 *                  be stored.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error ix_cc_atmsar_agent_vc_stats(
void *arg_pContext,
const ix_cc_atmsar_vc_handle_t arg_hVc,
ix_cc_atmsar_vc_stats_t* arg_pLookupResult);


/**
 * NAME: ix_cc_atmsar_agent_port_create
 *
 * DESCRIPTION:
 *    The function adds Port into ATM SAR Agent plugins.
 * 
 * @Param:  IN  arg_pContext : pointer to ATM SAR Agent CC
 * @Param:  IN  arg_hPort : handle to port to be created
 * @Param:  IN  arg_pPortDescr : parameters describing requested port.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error ix_cc_atmsar_agent_port_create(
void *arg_pContext,
const ix_cc_atmsar_port_handle_t arg_hPort,
const ix_cc_atmsar_port_descr_t* arg_pPortDescr);


/**
 * NAME: ix_cc_atmsar_agent_port_remove
 *
 * DESCRIPTION:
 *    The function removes port from ATM SAR Agent plugins.
 * 
 * @Param:  IN  arg_pContext : pointer to ATM SAR Agent CC
 * @Param:  IN  arg_hPort : handle to port to be removed
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error ix_cc_atmsar_agent_port_remove(
void *arg_pContext,
const ix_cc_atmsar_port_handle_t arg_hPort);


/**
 * NAME: ix_cc_atmsar_agent_port_stats
 *
 * DESCRIPTION:
 *    The function reads statistic for port from ATM SAR Agent plugins.
 * 
 * @Param:  IN  arg_pContext : pointer to ATM SAR Agent CC
 * @Param:  IN  arg_hPort : handle to port for statistic reading
 * @Param:  OUT arg_pLookupResult; pointer to buffer where port statistic has to
 *                  be stored.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error ix_cc_atmsar_agent_port_stats(
void *arg_pContext,
const ix_cc_atmsar_port_handle_t arg_hPort,
ix_cc_atmsar_port_stats_t* arg_pLookupResult);


/**
 * NAME: ix_cc_atmsar_agent_cid_create
 *
 * DESCRIPTION:
 *    The function adds cid into ATM SAR Agent plugins.
 * 
 * @Param:  IN  arg_pContext : pointer to ATM SAR Agent CC
 * @Param:  IN  arg_hCid : handle to CID to be created
 * @Param:  IN  arg_pCidDescr : parameters describing requested CID.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error ix_cc_atmsar_agent_cid_create(
void *arg_pContext,
const ix_cc_atmsar_cid_handle_t arg_hCid,
const ix_cc_atmsar_cid_descr_t* arg_pCidDescr);


/**
 * NAME: ix_cc_atmsar_agent_cid_update
 *
 * DESCRIPTION:
 *    The function updates cid in ATM SAR Agent plugins.
 * 
 * @Param:  IN  arg_pContext : pointer to ATM SAR Agent CC
 * @Param:  IN  arg_hCid : handle to CID to be created
 * @Param:  IN  arg_pCidDescr : parameters describing requested CID.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error ix_cc_atmsar_agent_cid_update(
void *arg_pContext,
const ix_cc_atmsar_cid_handle_t arg_hCid,
const ix_cc_atmsar_update_data_t* arg_pUpdate);

/**
 * NAME: ix_cc_atmsar_agent_cid_remove
 *
 * DESCRIPTION:
 *    The function removes CID from ATM SAR Agent plugins.
 * 
 * @Param:  IN  arg_pContext : pointer to ATM SAR Agent CC
 * @Param:  IN  arg_hCid : handle to CID to be removed
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error ix_cc_atmsar_agent_cid_remove(
void *arg_pContext,
const ix_cc_atmsar_cid_handle_t arg_hCid);


/**
 * NAME: ix_cc_atmsar_agent_cid_stats
 *
 * DESCRIPTION:
 *    The function reads statistic for CID from ATM SAR Agent plugins.
 * 
 * @Param:  IN  arg_pContext : pointer to ATM SAR Agent CC
 * @Param:  IN  arg_hCid : handle to CID for that statistic is read
 * @Param:  OUT arg_pLookupResult; pointer to buffer where CID statistic has to
 *                  be stored.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error ix_cc_atmsar_agent_cid_stats(
void *arg_pContext,
const ix_cc_atmsar_cid_handle_t arg_hCid,
ix_cc_atmsar_cid_stats_t* arg_pLookupResult);


/******************************************************************************
 *                       ANOTHER INTERNAL FUNCTIONS
 *****************************************************************************/


/**
 * NAME: _ix_cc_atmsar_agent_get_ctx
 *
 * DESCRIPTION:
 *    This function returns pointer to ATM SAR AGENT Core Component
 *    context.
 * 
 * @Param:  IN  arg_instance : Id of SAR Instance for that the action is
 *                  performed.
 * @Return: pointer to context of ATM SAR AGENT Core Component.
 */
_ix_cc_atmsar_agent_context_t* _ix_cc_atmsar_agent_get_ctx(
ix_cc_atmsar_instance_t arg_instance);

#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__IX_CC_ATMSAR_AGENT_PRIVATE_H__) */

