/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2400 Network Processor, Release 5
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      ATM SAR Core Component
 *
 * = FILENAME
 *      ix_cc_atmsar_common.h
 *
 * = DESCRIPTION
 *      This file defines types, definitions and functions used/shared in ATM
 *      SAR MAIN and ATM SAR AGENT core component.
 *
 * = AUTHOR
 *       Jaroslaw Kogut
 *       jaroslaw.kogut@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *     
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 * 
 */

#if !defined(__IX_CC_ATMSAR_COMMON_H__)
#define __IX_CC_ATMSAR_COMMON_H__

#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_cc.h"
#include "ix_cc_properties.h"
#include "cc/ix_cc_atmsar_plugin.h"

#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */


/******************************************************************************
 *                PREPROCESSOR DEFINITIONS: MACROS & DEFINES
 *****************************************************************************/

/* TBD: Temporaty solution - during redesing SAR in multinstaces solution.
 * Instead of the definition should be passed mechanism that find instance id
 * for the particular core component */
#define _IX_CC_ATMSAR_INSTANCE_ID_TMP 0

#define _IX_CC_ATMSAR_INSTANCE_MAX 10


/**
 * MACRO NAME: _IX_CC_ATMSAR_UNUSED_ARG
 *
 * DESCRIPTION:
 *    This macro is used to remove the warnings generated for unused
 *    arguments.
 */
#define _IX_CC_ATMSAR_UNUSED_ARG(arg_Argument) (void)arg_Argument


/******************************************************************************
 *                         TYPES DEFINITION
 *****************************************************************************/

/**
 * TYPENAME: _ix_cc_atmsar_plugin_entry_t
 * 
 * DESCRIPTION:
 *    The structure describes state of plugin, contains registered callback
 *    routines, and pointer to ATM SAR core component corelated with the
 *    instance of plugin.
 */
typedef struct ix_s_cc_atmsar_plugin_descr {
    ix_uint16  plugInId;
#define _IX_CC_ATMSAR_PLUGIN_INITFLAG_VC        0x01
#define _IX_CC_ATMSAR_PLUGIN_INITFLAG_PORT      0x02
#define _IX_CC_ATMSAR_PLUGIN_INITFLAG_CID       0x04
#define _IX_CC_ATMSAR_PLUGIN_INITFLAG_GENERAL   0x08
    ix_uint16  initFlag;
    ix_cc_atmsar_plugin_vc_create_t pCbVcCreate;
    ix_cc_atmsar_plugin_vc_update_t pCbVcUpdate;
    ix_cc_atmsar_plugin_vc_remove_t pCbVcRemove;
    ix_cc_atmsar_plugin_get_vc_stats_t pCbVcStats;
    ix_cc_atmsar_plugin_port_create_t pCbPortCreate;
    ix_cc_atmsar_plugin_port_remove_t pCbPortRemove;
    ix_cc_atmsar_plugin_get_port_stats_t pCbPortStats;
    ix_cc_atmsar_plugin_cid_create_t pCbCidCreate;
    ix_cc_atmsar_plugin_cid_update_t pCbCidUpdate;
    ix_cc_atmsar_plugin_cid_remove_t pCbCidRemove;
    ix_cc_atmsar_plugin_get_cid_stats_t pCbCidStats;
    void *pCtx;
} _ix_cc_atmsar_plugin_entry_t;


/**
 * TYPENAME: _ix_cc_atmsar_engine_t
 * 
 * DESCRIPTION:
 *    The structure contains description of all core components serviced by
 *    ATM SAR main or ATM SAR agent.
 *
 *    The membler 'threadId' is used for verifying plug-ins that try to
 *    register into the engine. Updating of the engine is allowed only
 *    if requested function is executed in the same thread.
 *
 *    Configuration parameters
 *    ------------------------
 *    The structure contains description of  configuration for SAR core
 *    component (agent or main). The 'ccNumber' describes how many core
 *    component descriptions are ready in the 'pluginTab'. The plugin tab
 *    is dinamicly created during preparing the configuration structure.
 */
typedef struct ix_s_cc_atmsar_engine
{
    const char* appName;   /**< application name: for debug purpose */
    ix_ossl_thread_t threadId; /**< thread id where the engine was created */

    /* the callbacks and core component context for handle managers */

    /* the VC handle manager */
    ix_cc_atmsar_plugin_vc_alloc_t pCbVcAlloc;
    ix_cc_atmsar_plugin_vc_free_t pCbVcFree;
    void *pVcHandleMgrCtx;

    /* the port handle manager */
    ix_cc_atmsar_plugin_port_alloc_t pCbPortAlloc;
    ix_cc_atmsar_plugin_port_free_t pCbPortFree;
    void *pPortHandleMgrCtx;

    /* the CID handle manager */
    ix_cc_atmsar_plugin_cid_alloc_t pCbCidAlloc;
    ix_cc_atmsar_plugin_cid_free_t pCbCidFree;
    void *pCidHandleMgrCtx;

    /* configuration parameters */
    ix_uint8  ccNumber;                       /**< number of core components */
    _ix_cc_atmsar_plugin_entry_t* pluginTab;   /**< pointer to table with    */
    ix_cc_atmsar_plugin_cfg_params_t cfg;  /**< common cfg for all plug-ins */

    /* core component description */

    ix_error (*pInitVerify)(void*); /**< callback that update status of initialization for the core component context */
    void *pCtx; /**< my core component context */
}_ix_cc_atmsar_engine_t;


/**
 * TYPENAME: _ix_cc_atmsar_module_id_t
 * 
 * DESCRIPTION:
 *    The type describes what kind of ATM SAR core component are allowed in the
 *    system
 */
typedef enum {
    _IX_CC_ATMSAR_MODULE_ID_MAIN,
    _IX_CC_ATMSAR_MODULE_ID_AGENT
}_ix_cc_atmsar_module_id_t;


/******************************************************************************
 *                        FUNCTIONS DECLARATION
 *****************************************************************************/


/******************************************************************************
 *           FUNCTIONS OF SAR CONTROL READ CONFIGURATION SUBMODULE
 *****************************************************************************/

/**
 * NAME: _ix_cc_atmsar_engine_read_cfg
 *
 * DESCRIPTION:
 *    This function prepares/fills configuration structure. It reads
 *    configuration file or it gets the configuration from header file.
 *    The source of configuration is switched by compilation switch.
 *    The read configuration is preapared for one instance of SAR and for AGENT
 *    or MAIN core component. It depends on input parameters.
 * 
 * @Param:  IN  arg_instance :  Identificator of instance.
 * @Param:  IN  arg_moduleId :  Identificator of Module.
 * @Param:  OUT arg_pEngine :  pointer to structure that will be felt.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCESS - ATM port successfully created
 */
ix_error _ix_cc_atmsar_engine_read_cfg(
ix_cc_atmsar_instance_t arg_instance,
_ix_cc_atmsar_module_id_t arg_moduleId,
_ix_cc_atmsar_engine_t* arg_pEngine);


/******************************************************************************
 *              FUNCTIONS OF SAR CONTROL ENGINE SUBMODULE
 *****************************************************************************/

/**
 * NAME: _ix_cc_atmsar_engine_vc_create
 *
 * DESCRIPTION:
 *    This function creates an VC with specified value in 'arg_pVcDescr'.
 * 
 * @Param:  IN  arg_pEngine :  pointer to structure contains description of
 *                  all sub-core conponets serviced by the main or agent sar
 *                  control submodule
 * @Param:  IN  arg_hVC :  handle to created VC.
 * @Param:  IN  arg_pVcDescr : parameters describing requested VC.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCESS - ATM port successfully created
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM - invalid input parameters
 *    IX_NO_RESOURCES - no resources to add the entry 
 *    IX_CC_ERROR_UNINIT - This CC has not been initialized
 */
ix_error _ix_cc_atmsar_engine_vc_create(
const _ix_cc_atmsar_engine_t *arg_pEngine,
const ix_cc_atmsar_vc_handle_t arg_phVC,
const ix_cc_atmsar_vc_descr_t* arg_pVcDescr);


/**
 * NAME: _ix_cc_atmsar_engine_vc_update
 *
 * DESCRIPTION:
 *    This function updates a VC with specified value in 'arg_pVcDescr'.
 * 
 * @Param:  IN  arg_pEngine :  pointer to structure contains description of
 *                  all sub-core conponets serviced by the main or agent sar
 *                  control submodule
 * @Param:  IN  arg_hVC :  handle to created VC.
 * @Param:  IN  arg_pVcDescr : parameters describing requested VC.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCESS - ATM port successfully created
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM - invalid input parameters
 *    IX_NO_RESOURCES - no resources to add the entry 
 *    IX_CC_ERROR_UNINIT - This CC has not been initialized
 */
ix_error _ix_cc_atmsar_engine_vc_update(
const _ix_cc_atmsar_engine_t *arg_pEngine,
const ix_cc_atmsar_vc_handle_t arg_hVc,
const ix_cc_atmsar_update_data_t* arg_pUpdate);


/**
 * NAME: _ix_cc_atmsar_engine_vc_remove
 *
 * DESCRIPTION:
 *    This function removes a VC specified by 'arg_hVc'.
 *    The function calls the core component managed by agent or main
 *    sar control. The list of sub-rutines for calling is pessed in
 *    'arg_pEngine'.
 * 
 * @Param:  IN  arg_pEngine : pointer to structure describing supported
 *                  core components.
 * @Param:  IN  arg_hVc : handle to VC to be removed
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - vc successfully removed
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM - invalid input parameters
 *    IX_CC_ERROR_UNINIT - This CC has not been initialized
 */
ix_error _ix_cc_atmsar_engine_vc_remove(
const _ix_cc_atmsar_engine_t *arg_pEngine,
ix_cc_atmsar_vc_handle_t arg_hVc);


/**
 * NAME: _ix_cc_atmsar_engine_get_port_stats
 *
 * DESCRIPTION:
 *    This function retrieves an VC statistics. 
 * 
 * @Param:  IN  arg_pEngine : pointer to structure describing supported
 *                  core components.
 * @Param:  IN  arg_hVc : handle to VC to be removed
 * @Param:  OUT arg_pLookupResult : pointer to statistic structure
 *                  where the requested data will be stored
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - statistics successfully retrieved
 *    IX_CC_INVALID_PORT - specified ATM port does not exist 
 *    IX_CC_ERROR_UNINIT - This CC has not been initialized
 */
ix_error _ix_cc_atmsar_engine_get_vc_stats(
const _ix_cc_atmsar_engine_t *arg_pEngine,
ix_cc_atmsar_vc_handle_t arg_hVc,
ix_cc_atmsar_vc_stats_t *arg_pLookupResult);


/**
 * NAME: ix_cc_atmsar_engine_port_create
 *
 * DESCRIPTION:
 *    This function creates an ATM port with default values.
 * 
 * @Param:  IN  arg_pEngine : pointer to structure describing supported
 *                  core components.
 * @Param:  IN  arg_pPortDescr : pointer to data describing parameters
 *                  for requested port
 * @Param:  OUT arg_hPort : ATM port handle
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCESS - ATM port successfully created
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM - invalid input parameters
 *    IX_NO_RESOURCES - no resources to add the entry 
 *    IX_CC_ERROR_UNINIT - This CC has not been initialized
 */
ix_error _ix_cc_atmsar_engine_port_create(
const _ix_cc_atmsar_engine_t *arg_pEngine,
ix_cc_atmsar_port_handle_t arg_hPort,
const ix_cc_atmsar_port_descr_t* arg_pPortDescr);


/**
 * NAME: ix_cc_atmsar_engine_port_remove
 *
 * DESCRIPTION:
 *    This function removes an ATM port from all VLANs.
 * 
 * @Param:  IN  arg_pEngine : pointer to structure describing supported
 *                  core components.
 * @Param:  IN  arg_port : ATM port to be removed
 * @Param:  OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - port successfully removed
 *    IX_CC_INVALID_PORT - specified port does not exist
 *    IX_CC_ERROR_UNINIT - This CC has not been initialized
 */
ix_error _ix_cc_atmsar_engine_port_remove(
const _ix_cc_atmsar_engine_t *arg_pEngine,
ix_cc_atmsar_port_handle_t arg_port);


/**
 * NAME: _ix_cc_atmsar_engine_get_port_stats
 *
 * DESCRIPTION:
 *    This function retrieves an ATM port statistics. 
 * 
 * @Param:  IN  arg_pEngine : pointer to structure describing supported
 *                  core components.
 * @Param:  IN  arg_port : ATM port ID
 * @Param:  OUT arg_pLookupResult : pointer to ix_cc_sar_port_stats structure
 *                  where the requested data will be stored
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - statistics successfully retrieved
 *    IX_CC_INVALID_PORT - specified ATM port does not exist 
 *    IX_CC_ERROR_UNINIT - This CC has not been initialized
 */
ix_error _ix_cc_atmsar_engine_get_port_stats(
const _ix_cc_atmsar_engine_t *arg_pEngine,
ix_cc_atmsar_port_handle_t arg_port,
ix_cc_atmsar_port_stats_t * arg_pLookupResult);


/**
 * NAME: _ix_cc_atmsar_engine_cid_create
 *
 * DESCRIPTION:
 *    This function creates an CID instance.
 * 
 * @Param:  IN  arg_pEngine : pointer to structure describing supported
 *                  core components.
 * @Param:  IN  arg_pCidDescr : pointer to data describing parameters
 *                  for requested CID
 * @Param:  OUT arg_hCid : CID handle
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCESS - instance successfully created
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM - invalid input parameters
 *    IX_NO_RESOURCES - no resources to add the entry 
 *    IX_CC_ERROR_UNINIT - This CC has not been initialized
 */
ix_error _ix_cc_atmsar_engine_cid_create(
const _ix_cc_atmsar_engine_t *arg_pEngine,
ix_cc_atmsar_cid_handle_t arg_hCid,
const ix_cc_atmsar_cid_descr_t* arg_pCidDescr);

/**
 * NAME: _ix_cc_atmsar_engine_cid_update
 *
 * DESCRIPTION:
 *    This function update an CID instance.
 * 
 * @Param:  IN  arg_cid : pointer to structure describing supported
 *                  core components.
 * @Param:  IN  arg_pUpdate : pointer to data describing parameters
 *                  for requested CID
 * @Param:  OUT arg_hCid : CID handle
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - instance successfully updated
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM - invalid input parameters
 *    IX_NO_RESOURCES - no resources to add the entry 
 *    IX_CC_ERROR_UNINIT - This CC has not been initialized
 */
ix_error _ix_cc_atmsar_engine_cid_update(
const _ix_cc_atmsar_engine_t *arg_pEngine,
ix_cc_atmsar_cid_handle_t arg_cid,
const  ix_cc_atmsar_update_data_t* arg_pUpdate);

/**
 * NAME: _ix_cc_atmsar_engine_cid_remove
 *
 * DESCRIPTION:
 *    This function removes the CID instance.
 * 
 * @Param:  IN  arg_pEngine : pointer to structure describing supported
 *                  core components.
 * @Param:  IN  arg_cid : ATM cid to be removed
 * @Param:  OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - port successfully removed
 *    IX_CC_INVALID_PORT - specified port does not exist
 *    IX_CC_ERROR_UNINIT - This CC has not been initialized
 */
ix_error _ix_cc_atmsar_engine_cid_remove(
const _ix_cc_atmsar_engine_t *arg_pEngine,
ix_cc_atmsar_cid_handle_t arg_cid);


/**
 * NAME: _ix_cc_atmsar_engine_get_cid_stats
 *
 * DESCRIPTION:
 *    This function retrieves an CID statistics. 
 * 
 * @Param:  IN  arg_pEngine : pointer to structure describing supported
 *                  core components.
 * @Param:  IN  arg_cid : CID handle
 * @Param:  OUT arg_pLookupResult : pointer to ix_cc_sar_cid_stats structure
 *                  where the requested data will be stored
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - statistics successfully retrieved
 *    IX_CC_ERROR_UNINIT - This CC has not been initialized
 */
ix_error _ix_cc_atmsar_engine_get_cid_stats(
const _ix_cc_atmsar_engine_t *arg_pEngine,
ix_cc_atmsar_cid_handle_t arg_cid,
ix_cc_atmsar_cid_stats_t * arg_pLookupResult);


/******************************************************************************
 *                  SUPPORT FOR DEBBUGING AND TRACES
 *****************************************************************************/

extern ix_uint32 g_atmsarDebugMask;

#define _IX_CC_ATMSAR_DBG(__mask, __debugInfo) \
    if((__mask & g_atmsarDebugMask)!= 0) ix_ossl_message_log __debugInfo;

#define _IX_CC_ATMSAR_DBG_INIT(__debugInfo) _IX_CC_ATMSAR_DBG(0x01, __debugInfo)
#define _IX_CC_ATMSAR_DBG_ENG(__debugInfo) _IX_CC_ATMSAR_DBG(0x02, __debugInfo)
#define _IX_CC_ATMSAR_DBG_MAIN(__debugInfo) _IX_CC_ATMSAR_DBG(0x04, __debugInfo)
#define _IX_CC_ATMSAR_DBG_AGENT(__debugInfo) _IX_CC_ATMSAR_DBG(0x08, __debugInfo)
#define _IX_CC_ATMSAR_DBG_TEST(__debugInfo) _IX_CC_ATMSAR_DBG(0x10, __debugInfo)

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
/******************************************************************************
 *                             TRUE/FALSE defines
 *****************************************************************************/
#define FALSE 0x0
#define TRUE  0x1
#endif

#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__IX_CC_ATMSAR_COMMON_H__) */

