/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor
 *
 * = LIBRARY
 *
 *
 * = MODULE
 *      ATM SAR Agent Core Component - CCI - init function
 *
 * = FILENAME
 *      ix_cc_atmsar_agent_cci.c
 *
 * = DESCRIPTION
 *      The file defines for core component infrestructure  interface and
 *      internal functions for initialization and shutdown.
 *
 * = AUTHOR
 *       Jaroslaw Kogut
 *       jaroslaw.kogut@intel.com
 *
 * = AKNOWLEDGEMENTS
 *
 *
 * = CREATION TIME
 *
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_atmsar_agent_cci.c,v 1.2 2003/09/10 07:37:10 rwielick Exp $";



/**
 * System defined include files
 */
#include "ix_cc_error.h"
#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_cc.h"
#include "ix_cc_macros.h"
#include "ix_cc_properties.h"
#include "ix_config.h"
#include "bindings.h"
#include "cc/ix_cc_msup.h"

/**
 * User defined include files
 */
#include "cc/ix_cc_atmsar.h"
#include "cc/ix_cc_atmsar_agent_cci.h"
#include "cc/ix_cc_atmsar_plugin.h"
#include "cc/internal/ix_cc_atmsar_msg.h"
#include "cc/internal/ix_cc_atmsar_agent.h"
#include "cc/internal/ix_cc_atmsar_common.h"

/**
 * Preprocessor symbols and Macros used in this file
 */


/**
 * Type definitions whose scope is limited to this file
 */

/**
 * Variable declarations global to this file only. Externs are followed by
 * static variables.
 */

static _ix_cc_atmsar_agent_context_t* s_pAtmSarAgentCtxTab[_IX_CC_ATMSAR_INSTANCE_MAX];

/**
 * Extern function prototypes.
 */


/**
 * Static function prototypes.
 */
static ix_error
_ix_cc_atmsar_agent_msg_handler(
ix_buffer_handle arg_hDataToken,
ix_uint32 arg_UserData,
void* arg_pContext);


/**
 * Function definitions.
 */


/**
 * NAME: _ix_cc_atmsar_agent_get_ctx
 *
 * DESCRIPTION:
 *    This function returns pointer to ATM SAR AGENT Core Component
 *    context.
 *
 * @Param:  IN  arg_instance : Id of SAR Instance for that the action is
 *                  performed.
 * @Return: pointer to context of ATM SAR AGENT Core Component.
 */
_ix_cc_atmsar_agent_context_t*
_ix_cc_atmsar_agent_get_ctx(
ix_cc_atmsar_instance_t arg_instance)
{
    return s_pAtmSarAgentCtxTab[arg_instance];
} /* end _ix_cc_atmsar_agent_get_ctx */


/******************************************************************************
 *                   DEINICJALIZATION/SHUTDOWN FUNCTIONS
 *****************************************************************************/

/**
 * NAME: _ix_cc_atmsar_agent_unregister_message_handler
 *
 * DESCRIPTION:
 *    This primitive will be used to unregister message handler. It calls CCI
 *    services in level 1 system and resore manager services in level 0 system.
 *
 * @Param:  IN  arg_pContext : pointer of core component context.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_CC_ATMSAR_ERROR_CCI - error from CCI
 */
ix_error
_ix_cc_atmsar_agent_unregister_message_handler(
_ix_cc_atmsar_agent_context_t *arg_pContext)
{
    _IX_CC_ATMSAR_UNUSED_ARG(arg_pContext);
    /* Call CCI if CCI is enabled */
#if defined(IX_EXCLUDE_CCI)
    IX_ERROR_CRT(ix_rm_message_handler_unregister(IX_CC_ATMSAR_AGENT_MSG_INPUT),
                 IX_CC_ATMSAR_ERROR_CCI,
                 IX_ERROR_LEVEL_WARNING);
#else
    IX_ERROR_CRT(ix_cci_cc_remove_message_handler(IX_CC_ATMSAR_AGENT_MSG_INPUT),
                 IX_CC_ATMSAR_ERROR_CCI,
                 IX_ERROR_LEVEL_WARNING);
#endif
    return IX_SUCCESS;
}/* end _ix_cc_atmsar_agent_unregister_message_handler */


/******************************************************************************
 *                         INITIALISATION FUNCTION
 *****************************************************************************/

/**
 * NAME: _ix_cc_atmsar_agent_retrieve_registry_data
 *
 * DESCRIPTION:
 *    This primitive will be used to retrieve data from registry for a given
 *    property type.
 *
 * @Param:  IN  arg_pPropertyName - pointer to property name string
 * @Param:  OUT arg_pPropertyValue - pointer to property value
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_CC_ATMSAR_ERROR_REGISTRY  - failed to access registry
 */
ix_error
_ix_cc_atmsar_agent_retrieve_registry_data(
const char* arg_pPropertyName,
ix_uint32* arg_pPropertyValue)
{
    ix_configuration_property_handle hAtmsar_AgentPropertyHandle;

    /* Get static data from registry */
    IX_ERROR_CRT(ix_rm_cp_property_open(IX_CP_CORE_PROPERTY_ROOT,
                                        arg_pPropertyName,
                                        &hAtmsar_AgentPropertyHandle),
                 IX_CC_ATMSAR_ERROR_REGISTRY,
                 IX_ERROR_LEVEL_WARNING);

    IX_ERROR_CRT(ix_rm_cp_property_get_value_uint32(hAtmsar_AgentPropertyHandle,
                                                    arg_pPropertyValue),
                 IX_CC_ATMSAR_ERROR_REGISTRY,
                 IX_ERROR_LEVEL_WARNING);

    IX_ERROR_CRT(ix_rm_cp_property_close(hAtmsar_AgentPropertyHandle),
                 IX_CC_ATMSAR_ERROR_REGISTRY,
                 IX_ERROR_LEVEL_WARNING);

    return IX_SUCCESS;
}/* end _ix_cc_atmsar_agent_retrieve_registry_data */


/**
 * NAME: _ix_cc_atmsar_agent_registry
 *
 * DESCRIPTION:
 *    This primitive will be used to retreive data from registry for all static
 *    data and stored in Atmsar_AgentRegistyData
 *
 * @Param:  IN  none
 * @Param:  OUT none
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_CC_ATMSAR_ERROR_REGISTRY  - not able to access registry.
 */
ix_error
_ix_cc_atmsar_agent_registry()
{
    return IX_SUCCESS;
}/* end _ix_cc_atmsar_agent_registry */


/**
 * NAME: _ix_cc_atmsar_agent_register_message_handler
 *
 * DESCRIPTION: This primitive will be used to register message handler.
 *
 * @Param:  IN  arg_pContext : pointer of core component context.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_CC_ATMSAR_ERROR_CCI - Error from CCI
 */
ix_error
_ix_cc_atmsar_agent_register_message_handler(
_ix_cc_atmsar_agent_context_t* arg_pContext)
{
    /* Call CCI message handler if CCI is enabled */
#if defined(IX_EXCLUDE_CCI)
    IX_ERROR_CRT(ix_rm_message_handler_register(IX_CC_ATMSAR_AGENT_MSG_INPUT,
                                                _ix_cc_atmsar_agent_msg_handler,
                                                arg_pContext),
                 IX_CC_ATMSAR_ERROR_CCI,
                 IX_ERROR_LEVEL_WARNING);
#else
    IX_ERROR_CRT(ix_cci_cc_add_message_handler(arg_pContext->hAtmsar_AgentHandle,
                                               IX_CC_ATMSAR_AGENT_MSG_INPUT,
                                               _ix_cc_atmsar_agent_msg_handler,
                                               IX_INPUT_TYPE_MULTI_SRC),
                 IX_CC_ATMSAR_ERROR_CCI,
                 IX_ERROR_LEVEL_WARNING);

#endif
    return IX_SUCCESS;
}/* end _ix_cc_atmsar_agent_register_message_handler */



/**
 * NAME: _ix_cc_atmsar_cb_from_main_agent_init
 *
 * DESCRIPTION:
 *    This callback function for reply from ATM SAR Main for.
 *    'AGENT INIT' message.
 *
 * @Param:  IN  arg_Result: is the ix_error from the called function.
 * @Param:  IN  arg_pContext: pointer of core component context.
 * @Param:  IN  arg_Msg: is the message.
 * @Param:  IN  arg_Msg_Len: is the length of the message
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent
 *    IX_ERROR
 */
static ix_error
_ix_cc_atmsar_cb_from_main_agent_init(
ix_error arg_result,
void *arg_pContext,
void *arg_pMsg,
ix_uint32 arg_Msg_Len)
{
    _IX_CC_ATMSAR_DBG_INIT(( "AGENT: initialized\n" ));
    return IX_SUCCESS;
} /* end _ix_cc_atmsar_cb_from_main_agent_init */

/**
 * NAME: _ix_cc_atmsar_agent_init_verify
 *
 * DESCRIPTION:
 *    This function check status of plug-in table for ATM SAR Agent
 *    subcomponent.
 *
 * @Param:  IN  arg_pCtx : pointer to ATM SAR Agent context.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
static ix_error
_ix_cc_atmsar_agent_init_verify(
void *arg_pCtx)
{
    ix_error err = IX_SUCCESS;
    char *pFncName="_ix_cc_atmsar_agent_init_verify";
    ix_uint16 i;
    _ix_cc_atmsar_plugin_entry_t *pTab;
    _ix_cc_atmsar_agent_context_t* pCtx=(_ix_cc_atmsar_agent_context_t*)arg_pCtx;
    ix_communication_id commId;
    _ix_cc_atmsar_agent_init_notify_t nfyData;

    if (pCtx==NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("Input parameter(arg_pCtx) is null"));
    }

    pTab=pCtx->engine.pluginTab;

    /* find uninitialized plug-ins */
    for (i=0; i < pCtx->engine.ccNumber ; i++)
    {
        if ((pTab[i].initFlag & _IX_CC_ATMSAR_PLUGIN_INITFLAG_GENERAL) == 0)
        {
            /* any plug-in is uninitialized */
            _IX_CC_ATMSAR_DBG_INIT(( "%s() entry %d is not initialised\n", pFncName, i ));
            return IX_SUCCESS;
        }
    }
    _IX_CC_ATMSAR_DBG_INIT(( "%s() all plug-ins initialized\n" , pFncName));
    /* OK. all is initialized, we can set that ATM SAR MAIN is ready */

    _IX_CC_ATMSAR_DBG_INIT(( "%s() s_pAtmsarAgentContext updated\n", pFncName ));
    pCtx->subCoreInitialized |= _IX_CC_ATMSAR_INIT_MASK_PLUGIN;
    if (pCtx->engine.pCbPortAlloc != NULL)
    {
        _IX_CC_ATMSAR_DBG_INIT(("   port management installed\n" ));
        pCtx->subCoreInitialized |= _IX_CC_ATMSAR_INIT_MASK_PORT_MGMNT;
    }
    if (pCtx->engine.pCbVcAlloc != NULL)
    {
        _IX_CC_ATMSAR_DBG_INIT(("   vc management installed\n" ));
        pCtx->subCoreInitialized |= _IX_CC_ATMSAR_INIT_MASK_VC_MGMNT;
    }
    if (pCtx->engine.pCbCidAlloc != NULL)
    {
        _IX_CC_ATMSAR_DBG_INIT(("   cid management installed\n" ));
        pCtx->subCoreInitialized |= _IX_CC_ATMSAR_INIT_MASK_CID_MGMNT;
    }

    nfyData.flags=pCtx->subCoreInitialized;

#ifndef ATMSAR_MAIN_REMOTE
    commId = IX_RM_COMM_MAKE_LOCAL_ID(IX_CC_ATMSAR_MAIN_MSG_ID);
#else
    commId = IX_RM_COMM_MAKE_ID(IX_CC_ATMSAR_MAIN_MSG_ID,IX_PEER_SUBSYSTEM, 0);
#endif
    err=ix_cc_msup_send_async_msg(commId,
                                  _ix_cc_atmsar_cb_from_main_agent_init,
                                  (void*)pCtx,
                                  _IX_CC_ATMSAR_MSG_AGENT_INIT,
                                  (void *)&nfyData,
                                  sizeof(nfyData));
    return IX_SUCCESS;
} /* _ix_cc_atmsar_agent_init_verify */

/******************************************************************************
 *                CORE COMPONENT INFRASTRUCTURE FUNCTIONS
 *****************************************************************************/

/**
 * NAME: ix_cc_atmsar_agent_init
 *
 * DESCRIPTION:
 *    This primitive shall be called and returned successfully before
 *    requesting any service from this core component. This primitive should be
 *    called only once to initialize this core component.This function performs
 *    allocation of memory for symbols to be patched,Creation of 64 bit
 *    counters, registration of packet and message handlers, allocation and
 *    initialization of internal data structures.
 *
 * @Param:  IN arg_hCcHandle - handle to this core component, created by core
 *            component infrastructure; This shall be used later to get other
 *            (to add event handler) services from CCI.
 * @Param:  IN-OUT arg_ppContext - Buffer free list and dynamic property data
 *            will be given during by system APPs during initialization. Upon
 *            successfull completion, it represents the location where the
 *            pointer to the control block allocated by this core component
 *            will be stored. The control block is internal to the core
 *            component and it contains information about internal data
 *            structures, allocated memory and other relevant information.
 *            Later this pointer to be passed into core component fini()
 *            function for freeing memory and other house keeping operations
 *            when this component is being destroyed.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_CC_ERROR_NULL - input parameter arg_ppContext is null
 *    IX_CC_ERROR_OOM   - memory allocation failure
 *    IX_CC_ATMSAR_ERROR_FAILED_PATCHING   - patching failures
 *    IX_CC_ERROR_OOM_64BIT_COUNTER - 64 bit counter creation failure
 *    IX_CC_ATMSAR_ERROR_REGISTRY  - registry access failures
 *    IX_CC_ATMSAR_ERROR_RTM - failure from RTM core component
 *    IX_CC_ATMSAR_ERROR_CCI - failure from CCI
 */
ix_error
ix_cc_atmsar_agent_init(
ix_cc_handle arg_hCcHandle,
void** arg_ppContext)
{
    ix_error err = IX_SUCCESS;
    _ix_cc_atmsar_agent_context_t* pContext;
    ix_cc_atmsar_instance_t instance;

    /* Variables used in retrieving static data from arg_ppContext */
    ix_cc_init_context *pInitCtx;

    _IX_CC_ATMSAR_DBG_INIT(( "ix_cc_atmsar_agent_init() START\n"));

    /* TBD Add multiinstance support */
    instance=_IX_CC_ATMSAR_INSTANCE_ID_TMP;

    /* Check for argument validity */
    if (arg_ppContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("Input parameter(arg_ppContext) is null"));

    }/* end if (arg_ppContext == NULL) */

    /* veryfication of pointer to context for the instance */
    if (s_pAtmSarAgentCtxTab[instance] != NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("Instance %d for ATM SAR Agent already initialized"));
    }

    /* Allocate memory for core component context */
    pContext = (_ix_cc_atmsar_agent_context_t*)ix_ossl_malloc(sizeof(_ix_cc_atmsar_agent_context_t));
    if (pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                                ("Failed to allocate memory for SAR CTRL context"));
    }/* end if (pContext == NULL) */

    /* initialize the structure */
    ix_ossl_memset(pContext, 0, sizeof(_ix_cc_atmsar_agent_context_t));

    /* store core component handle in _ix_cc_atmsar_agent_context_t */
    pContext->hAtmsar_AgentHandle = arg_hCcHandle;

    if ((err=ix_ossl_thread_get_id(&pContext->engine.threadId)) != IX_SUCCESS)
    {
        goto cleanupContext;
    }

    /* The pointer has to be set now for veryfing purpose */
    s_pAtmSarAgentCtxTab[instance]=pContext;

    pContext->engine.pInitVerify=_ix_cc_atmsar_agent_init_verify;
    pContext->engine.pCtx=pContext;
    pContext->engine.appName="AGENT";

    pContext->subCoreInitialized = 0;

    if ((err=ix_ossl_thread_get_id(&pContext->engine.threadId)) != IX_SUCCESS)
    {
        goto cleanupContext;
    }

    /* Get the generic init context */
    pInitCtx = (ix_cc_init_context *)*arg_ppContext;

    IX_ERROR_CG(_ix_cc_atmsar_engine_read_cfg(instance,
                                              _IX_CC_ATMSAR_MODULE_ID_AGENT,
                                              &(pContext->engine)),
                err, cleanupContext);

    /* Register Message Handler */
    IX_ERROR_CG(_ix_cc_atmsar_agent_register_message_handler(pContext), err, cleanupContext);

    if (err != IX_SUCCESS)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_OOR,
                               ("Failed to create mutext"));
        goto cleanupMessageHandler;

    }/* end if (err != IX_SUCCESS) */

    /* for testing */
    _IX_CC_ATMSAR_DBG_INIT(( "ix_cc_atmsar_agent_init() SUCCESS\n"));
    if (pContext->engine.pInitVerify == NULL)
    {
        err=IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL, ("pInitVerify is NULL"));
        goto cleanupMessageHandler;
    }

    IX_ERROR_CG(pContext->engine.pInitVerify(pContext->engine.pCtx),
                err, cleanupMessageHandler);

    *arg_ppContext = (void *)pContext;
    return IX_SUCCESS;

  cleanupMessageHandler:
    _ix_cc_atmsar_agent_unregister_message_handler(pContext);
  cleanupContext:
    ix_ossl_free(pContext);
    s_pAtmSarAgentCtxTab[instance]=NULL;
    return err;
} /* end ix_cc_atmsar_agent_init */


/**
 * NAME: ix_cc_atmsar_agent_fini
 *
 * DESCRIPTION:
 *    This primitive shall be called to terminate services from this core
 *    component. This primitive frees memory allocated during initialization,
 *    free all the created resources: delete 64 bit counter. If client claims
 *    for any services from the core component after calling this function,
 *    then the behavior is undefined.
 *
 * @Param:  IN arg_hCcHandle - handle to the core component
 * @Param:  IN arg_pContext - pointer to the control block memory allocated
 *                 earlier in ix_cc_atmsar_agent_init function.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_CC_ATMSAR_ERROR_INVALID_INPUT_PARAM - invalid input parameter(s)
 *    IX_CC_ERROR _OOM - failure of memory deletion
 *    IX_CC_ERROR_OOM_64BIT_COUNTER - failure of  64 bit counter deletion
 *    IX_CC_ATMSAR_ERROR _RTM - failure from RTM
 *    IX_CC_ATMSAR_ERROR_CCI - failure from CCI
 */
ix_error
ix_cc_atmsar_agent_fini(
ix_cc_handle arg_hCcHandle,
void* arg_pContext)
{
    ix_cc_atmsar_instance_t instance;
    _ix_cc_atmsar_agent_context_t* pContext= (_ix_cc_atmsar_agent_context_t*)arg_pContext;

    /* TBD Add multiinstance support */
    instance=_IX_CC_ATMSAR_INSTANCE_ID_TMP;

    /* Validate Parameters */
    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("arg_pContext is null"));

    }/* if (arg_pContext == NULL)  */

    if (arg_pContext != s_pAtmSarAgentCtxTab[instance])
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("arg_pContext is different than global test variable"));

    }/* if (arg_pContext != s_pAtmsarAgentContext) */


    if (arg_hCcHandle != pContext->hAtmsar_AgentHandle)
    {
        return IX_ERROR_WARNING(IX_CC_ATMSAR_ERROR_INVALID_INPUT_PARAM,
                                ("Invalid input Parameters"));
    }/* (arg_hCcHandle != pContext->hAtmsar_AgentHandle) */


    /* Unregister Message Handler */
    IX_ERROR_CR(_ix_cc_atmsar_agent_unregister_message_handler(pContext));

    /* Free Atmsar_Agent Context */
    s_pAtmSarAgentCtxTab[instance]=NULL;
    ix_ossl_free(pContext);
    return IX_SUCCESS;
} /* end ix_cc_atmsar_agent_fini */


/**
 * NAME: _ix_cc_atmsar_agent_msg_handler
 *
 * DESCRIPTION:
 *    This primitive is the message handler function for the core component.
 *    The core component receives messages from other core component through
 *    this message handler function and internally it calls appropriate library
 *    function to process the message. This message handler shall be used to
 *    update dynamic properties.
 *
 * @Param:  IN arg_hDataToken: buffer handle embedding information for the
 *                 message passed in arg_UserData.
 * @Param:  IN arg_UserData: message type
 * @Param:  IN arg_pContext: pointer to the coree component context that will
 *                 be passed to the core component when a message arrives. This
 *                 context was defined by the core component and passed to core
 *                 components infrastructure through CC init function.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_CC_ERROR_NULL - null argument
 *    IX_CC_ERROR_UNDEFINED_MSG - un supported message
 *    IX_CC_ATMSAR_ERROR_MSG_LIBRARY - error from message support library
 *    IX_CC_ATMSAR_ERROR_BUFFER_FREE - error from RM for freeing buffer
 */
ix_error
_ix_cc_atmsar_agent_msg_handler(
ix_buffer_handle arg_hDataToken,
ix_uint32 arg_UserData,
void* arg_pContext)
{
    ix_error err = IX_SUCCESS;
    void *pContext = NULL;
    void *pMsg = NULL;
    void *pRetMsg = NULL;
    ix_uint32 msgRetLength = 0;
    ix_uint32 noResponse = 0;
    _ix_cc_atmsar_agent_reply_general_t reply;


    /* Validate arg_pContext */
    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("arg_pContext is NULL"));
    }/* end if (arg_pContext == NULL) */

    if ((arg_UserData < IX_CC_COMMON_MSG_ID_FIRST) ||
        (arg_UserData >= _IX_CC_ATMSAR_MSG_LAST))
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_UNDEFINED_MSG,
                                ("Invalid Message Type"));
    }/* end if */

    IX_MSUP_EXTRACT_MSG(arg_hDataToken, &pMsg, &pContext);

    /* Check for route messages, next hop message, icmp messages and statistics message*/
    switch (arg_UserData)
    {
    case _IX_CC_ATMSAR_MSG_VC_CREATE:
    {
        _ix_cc_atmsar_agent_vc_create_request_t* pMsgVcCreate=pMsg;
        pRetMsg= &(reply.vcCreate);
        msgRetLength = sizeof(reply.vcCreate);
        err=ix_cc_atmsar_agent_vc_create(arg_pContext, pMsgVcCreate->handle,
                                         &pMsgVcCreate->descr);
        break;
    }
    case _IX_CC_ATMSAR_MSG_VC_UPDATE:
    {
        _ix_cc_atmsar_agent_vc_update_request_t* pMsgVcUpdate=pMsg;
        pRetMsg= &(reply.vcUpdate);
        msgRetLength = sizeof(reply.vcUpdate);
        err=ix_cc_atmsar_agent_vc_update(arg_pContext, pMsgVcUpdate->handle,
                                         &pMsgVcUpdate->descr);
        break;
    }
    case _IX_CC_ATMSAR_MSG_VC_REMOVE:
    {
        _ix_cc_atmsar_agent_vc_remove_request_t* pMsgVcRemove=pMsg;
        pRetMsg= &(reply.vcRemove);
        msgRetLength = sizeof(reply.vcRemove);
        err=ix_cc_atmsar_agent_vc_remove(arg_pContext, pMsgVcRemove->handle);
        break;
    }
    case _IX_CC_ATMSAR_MSG_GET_VC_STATS:
    {
        _ix_cc_atmsar_agent_vc_stats_request_t* pMsgVcStats=pMsg;
        pRetMsg= &(reply.vcStats);
        msgRetLength = sizeof(reply.vcStats);
        err=ix_cc_atmsar_agent_vc_stats(arg_pContext, pMsgVcStats->handle,
                                        &(reply.vcStats.stats));
        break;
    }
    case _IX_CC_ATMSAR_MSG_PORT_CREATE:
    {
        _ix_cc_atmsar_agent_port_create_request_t* pMsgPortCreate=pMsg;
        pRetMsg= &(reply.portCreate);
        msgRetLength = sizeof(reply.portCreate);
        err=ix_cc_atmsar_agent_port_create(arg_pContext, pMsgPortCreate->handle,
                                           &pMsgPortCreate->descr);
        break;
    }
    case _IX_CC_ATMSAR_MSG_PORT_REMOVE:
    {
        _ix_cc_atmsar_agent_port_remove_request_t* pMsgPortRemove=pMsg;
        pRetMsg= &(reply.portRemove);
        msgRetLength = sizeof(reply.portRemove);
        err=ix_cc_atmsar_agent_port_remove(arg_pContext, pMsgPortRemove->handle);
        break;
    }
    case _IX_CC_ATMSAR_MSG_GET_PORT_STATS:
    {
        _ix_cc_atmsar_agent_port_stats_request_t* pMsgPortStats=pMsg;
        pRetMsg= &(reply.portStats);
        msgRetLength = sizeof(reply.portStats);
        err=ix_cc_atmsar_agent_port_stats(arg_pContext, pMsgPortStats->handle,
                                          &(reply.portStats.stats));
        break;
    }
    case _IX_CC_ATMSAR_MSG_CID_CREATE:
    {
        _ix_cc_atmsar_agent_cid_create_request_t* pMsgCidCreate=pMsg;
        pRetMsg= &(reply.cidCreate);
        msgRetLength = sizeof(reply.cidCreate);
        err=ix_cc_atmsar_agent_cid_create(arg_pContext, pMsgCidCreate->handle,
                                          &pMsgCidCreate->descr);
        break;
    }
    case _IX_CC_ATMSAR_MSG_CID_UPDATE:
    {
        _ix_cc_atmsar_agent_cid_update_request_t* pMsgCidUpdate=pMsg;
        pRetMsg= &(reply.cidUpdate);
        msgRetLength = sizeof(reply.cidUpdate);
        err=ix_cc_atmsar_agent_cid_update(arg_pContext, pMsgCidUpdate->handle,
                                          &pMsgCidUpdate->update);
        break;
    }
    case _IX_CC_ATMSAR_MSG_CID_REMOVE:
    {
        _ix_cc_atmsar_agent_cid_remove_request_t* pMsgCidRemove=pMsg;
        pRetMsg= &(reply.cidRemove);
        msgRetLength = sizeof(reply.cidRemove);
        err=ix_cc_atmsar_agent_cid_remove(arg_pContext, pMsgCidRemove->handle);
        break;
    }
    case _IX_CC_ATMSAR_MSG_GET_CID_STATS:
    {
        _ix_cc_atmsar_agent_cid_stats_request_t* pMsgCidStats=pMsg;
        pRetMsg= &(reply.cidStats);
        msgRetLength = sizeof(reply.cidStats);
        err=ix_cc_atmsar_agent_cid_stats(arg_pContext, pMsgCidStats->handle,
                                         &(reply.cidStats.stats));
        break;
    }

    default:
        if (pMsg == NULL)
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                    ("Invalid Buffer"));
        }/* end if (pContext == NULL) */
        IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hDataToken), IX_CC_ATMSAR_ERROR_BUFFER_FREE, 0);  
        return IX_ERROR_WARNING(IX_CC_ERROR_UNDEFINED_MSG,
                                ("Invalid Message Type"));
        break;
    }/* end switch (arg_UserData) */

    if (noResponse == IX_CC_ATMSAR_AGENT_SEND_NO_RESPONSE)
    {
        IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hDataToken), IX_CC_ATMSAR_ERROR_BUFFER_FREE, 0);  
        return IX_SUCCESS;
    }/* end if no need to send message to msg support lib*/

    /* send return message, length and retrieved context to message support library */
    err = ix_cc_msup_send_reply_msg(pContext, pRetMsg, msgRetLength, err);
    if (err != IX_SUCCESS)
    {
        err = IX_ERROR_WARNING(IX_CC_ATMSAR_ERROR_MSG_LIBRARY,
                               ("Send reply message failed"));
    }/* endif (err != IX_SUCCESS) */

    IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hDataToken), IX_CC_ATMSAR_ERROR_BUFFER_FREE, 0);
    return err;
} /* end _ix_cc_atmsar_agent_msg_handler */

