/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      ATM SAR MAIN Core Component
 *
 * = FILENAME
 *      ix_cc_atmsar_main_api.c
 *
 * = DESCRIPTION
 *      TBD
 *
 * = AUTHOR
 *       Jaroslaw Kogut
 *       jaroslaw.kogut@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_atmsar_main_api.c,v 1.5 2003/12/12 18:24:48 adziecia Exp $";


/**
 * System defined include files
 */
#include "ix_cc_error.h"
#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_cc.h"
#include "ix_cc_macros.h"
#include "ix_cc_properties.h"
#include "ix_config.h"
#include "bindings.h"
#include "cc/ix_cc_msup.h"

#if (_IX_OS_TYPE_ != _IX_OS_LINUX_KERNEL_)
#include "assert.h"
#endif

/**
 * User defined include files
 */
#include "cc/ix_cc_atmsar.h"
#include "cc/internal/ix_cc_atmsar_msg.h"
#include "cc/internal/ix_cc_atmsar_main.h"
#include "cc/internal/ix_cc_atmsar_common.h"

/**
 * Preprocessor symbols and Macros used in this file
 */

/**
 * Type definitions whose scope is limited to this file
 */


/**
 * TYPENAME: _ix_cc_atmsar_async_ctx_to_agent_t
 * 
 * DESCRIPTION:
 *   The structure describes all parameters required for sending asynchronous
 *   call to ATM SAR AGENT and all parameters required for processing reply
 *   from ATM SAR AGENT and sending reply to user core component that
 *   generates the request.
 */
typedef struct{
    _ix_cc_atmsar_main_context_t *pMyContext;               /* the core component context   */
    char *pOperTxt;                               /* text description for debugging purpose */
    _ix_cc_atmsar_messages_t msgCode;
    ix_buffer_handle hBuf;                                  /* original buffer handle       */
    ix_error err;                                           /* result of local action       */
    union {
        ix_cc_atmsar_vc_handle_t hVc;                       /* for VC another operation     */
        ix_cc_atmsar_port_handle_t hPort;                   /* for port another operation   */
        ix_cc_atmsar_cid_handle_t hCid;                     /* for CID another operation    */
        _ix_cc_atmsar_async_vc_stats_reply_t *pVcStats;     /* for VC statistic operation   */
        _ix_cc_atmsar_async_port_stats_reply_t *pPortStats; /* for port statistic operation */
        _ix_cc_atmsar_async_cid_stats_reply_t *pCidStats;   /* for CID statistic operation  */
    } data;           /* the structure contains data set during processing the user request */
}_ix_cc_atmsar_async_ctx_to_agent_t;


/**
 * TYPENAME: _ix_cc_atmsar_fsm_stage_t
 * 
 * DESCRIPTION:
 *    The type defines stages of FSM during processing asynchronous user
 *    request. The FSM controls sending request to ATM SAR AGENT,
 *    processing occured errors, sending reply to user core component and etc.
 *
 *    Description of stages and operations that should be performed in
 *    the stage:
 *
 *    _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_NULL:
 *         Unused stage.
 *
 *    _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE:
 *         Fills general context of request allocated by the FSM mechanism
 *         with data appropriate to the request (set type of request, set
 *         size of structures exchange between AGENT and MAIN, If it is
 *         functions servicing FSM that create any object, it should calls
 *         allocation of handle in the stage. The new allocated handle
 *         should be stored in context data
 *
 *    _IX_CC_ATMSAR_FSM_STAGE_LOCAL_PROCESS:
 *         Request to AGENT is sent (if current instance of ATM SAR supports
 *         ATM SAR agent). In the stage ATM SAR MAIN has to call engine to
 *         process the request on local registered plugins. It reads data from
 *         context prepared in previous stage, and passes theirs into called
 *         engine function.
 *
 *    _IX_CC_ATMSAR_FSM_STAGE_ERROR_WITHOUT_SEND:
 *         The sending request to AGENT faild. FSM function has to free all
 *         resources allocated during stage _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE.
 *
 *    _IX_CC_ATMSAR_FSM_STAGE_PREPARE_REPLY:
 *         Fill buffer of reply message with data collected during processing the
 *         request. Set size of reply.
 *
 *    _IX_CC_ATMSAR_FSM_STAGE_AFTER_REPLY:
 *         The reply has sent to calling module. Free all resources extra
 *         allocated for processing the request. Check status of processing
 *         the request (arg_pFsm->pContext->err). If it is not IX_SUCCESS,
 *         free the object. 'Get statistic' for object should not cause
 *         freeing object.
 */
typedef enum _ix_cc_atmsar_e_fsm_stage {
    _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_NULL=0,
    _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE,
    _IX_CC_ATMSAR_FSM_STAGE_LOCAL_PROCESS,
    _IX_CC_ATMSAR_FSM_STAGE_ERROR_WITHOUT_SEND,
    _IX_CC_ATMSAR_FSM_STAGE_PREPARE_REPLY,
    _IX_CC_ATMSAR_FSM_STAGE_AFTER_REPLY
}_ix_cc_atmsar_fsm_stage_t;


/**
 * TYPENAME: _ix_cc_atmsar_main_fsm_t
 * 
 * DESCRIPTION:
 *
 */
typedef struct{
    _ix_cc_atmsar_async_ctx_to_agent_t *pContext;
    void *pMsg;
    void *pReplyMsg;
    union {
        _ix_cc_atmsar_agent_request_t req;
        _ix_cc_atmsar_async_reply_t reply;
    } data;
    ix_uint16 sizeBufferToSend;
    _ix_cc_atmsar_fsm_stage_t stage;        /* current stage of FSM */
}_ix_cc_atmsar_main_fsm_t;


/**
 * TYPENAME: _ix_cc_atmsar_fsm_cb_t
 * 
 * DESCRIPTION:
 *    The type defines format of callback routine called when ATM SAR Main
 *    core component processes following stages of FSM.
 */
typedef ix_error (*_ix_cc_atmsar_fsm_cb_t) (
_ix_cc_atmsar_main_fsm_t*);


/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */

/** 
 * Extern function prototypes.
 */


/** 
 * Static function prototypes.
 */

static ix_error _ix_cc_atmsar_main_rec_vc_create(_ix_cc_atmsar_main_fsm_t* arg_pFsm);
static ix_error _ix_cc_atmsar_main_rec_vc_update(_ix_cc_atmsar_main_fsm_t* arg_pFsm);
static ix_error _ix_cc_atmsar_main_rec_vc_remove(_ix_cc_atmsar_main_fsm_t* arg_pFsm);
static ix_error _ix_cc_atmsar_main_rec_vc_stats(_ix_cc_atmsar_main_fsm_t* arg_pFsm);
static ix_error _ix_cc_atmsar_main_rec_port_create(_ix_cc_atmsar_main_fsm_t* arg_pFsm);
static ix_error _ix_cc_atmsar_main_rec_port_remove(_ix_cc_atmsar_main_fsm_t* arg_pFsm);
static ix_error _ix_cc_atmsar_main_rec_port_stats(_ix_cc_atmsar_main_fsm_t* arg_pFsm);

static ix_error _ix_cc_atmsar_main_rec_cid_create(_ix_cc_atmsar_main_fsm_t* arg_pFsm);
static ix_error _ix_cc_atmsar_main_rec_cid_update(_ix_cc_atmsar_main_fsm_t* arg_pFsm);
static ix_error _ix_cc_atmsar_main_rec_cid_remove(_ix_cc_atmsar_main_fsm_t* arg_pFsm);
static ix_error _ix_cc_atmsar_main_rec_cid_stats(_ix_cc_atmsar_main_fsm_t* arg_pFsm);


static ix_error _ix_cc_atmsar_main_vc_clear(
_ix_cc_atmsar_main_context_t *arg_pMyContext,
ix_cc_atmsar_vc_handle_t arg_hVc);

static ix_error _ix_cc_atmsar_main_port_clear(
_ix_cc_atmsar_main_context_t *arg_pMyContext,
ix_uint16 arg_port);

static ix_error _ix_cc_atmsar_main_cid_clear(
_ix_cc_atmsar_main_context_t *arg_pMyContext,
ix_uint16 arg_cid);

/** 
 * Function definitions.
 */


/**
 * NAME: ix_cc_atmsar_put_async_ctx
 *
 * DESCRIPTION:
 *    The function free context after usage by asynchronous helper functions.
 *    The function services a pool of contextes.
 * 
 * @Param:  IN: pContext - context for freeing
 * 
 * @Return: None
 */
static void
_ix_cc_atmsar_main_free_ctx_to_agent(
_ix_cc_atmsar_async_ctx_to_agent_t* pContext)
{
    ix_ossl_free(pContext);
} /* end _ix_cc_atmsar_main_free_ctx_to_agent */


/**
 * NAME: _ix_cc_atmsar_main_alloc_ctx_to_agent
 *
 * DESCRIPTION:
 *    The function allocates context for asynchronous helper function.
 *    The function services a pool of contextes.
 * 
 * @Param:  none
 * 
 * @Return: pointer to context for asychronous SAR CTRL operation.
 *          NULL if no context is available.
 */
static _ix_cc_atmsar_async_ctx_to_agent_t*
_ix_cc_atmsar_main_alloc_ctx_to_agent(void)
{
    return (_ix_cc_atmsar_async_ctx_to_agent_t*)ix_ossl_malloc(sizeof(_ix_cc_atmsar_async_ctx_to_agent_t));
} /* end _ix_cc_atmsar_main_alloc_ctx_to_agent */



/**
 * NAME: _ix_cc_atmsar_main_cb_from_agent
 *
 * DESCRIPTION:
 *    This callback function for reply from ATM SAR Agent.
 * 
 * @Param:  IN  arg_Result is the ix_error from the called function.
 * @Param:  IN  arg_pContext is the user specified context
 * @Param:  IN  arg_Msg is the message.
 * @Param:  IN  arg_Msg_Len is the length of the message
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent
 *    IX_ERROR
 */
static ix_error
_ix_cc_atmsar_main_cb_from_agent(
ix_error arg_result,
void *arg_pContext,
void *arg_pMsg,
ix_uint32 arg_Msg_Len)
{
    _ix_cc_atmsar_main_fsm_t fsm;
    ix_uint16 i;
    void *pRemoteContext;
    ix_error err = IX_SUCCESS;

    fsm.pContext=arg_pContext;
    fsm.pMsg = arg_pMsg;

    if (fsm.pContext->err == IX_SUCCESS)
    {
        fsm.pContext->err=arg_result;
    }
    fsm.sizeBufferToSend = 0;
    fsm.pReplyMsg = NULL;

    
    for(i=0;i<2; i++)
    {

        if (i==0)
        {
            fsm.stage=_IX_CC_ATMSAR_FSM_STAGE_PREPARE_REPLY;
        }
        else
        {
            fsm.stage=_IX_CC_ATMSAR_FSM_STAGE_AFTER_REPLY;
        }
        switch (fsm.pContext->msgCode)
        {
        case _IX_CC_ATMSAR_MSG_VC_CREATE:

            IX_ERROR_CT(_ix_cc_atmsar_main_rec_vc_create(&fsm), err,
                        IX_CC_ATMSAR_ERROR_MSG_LIBRARY, IX_ERROR_LEVEL_GLOBAL);
            break;

        case _IX_CC_ATMSAR_MSG_VC_UPDATE:
            IX_ERROR_CT(_ix_cc_atmsar_main_rec_vc_update(&fsm), err,
                        IX_CC_ATMSAR_ERROR_MSG_LIBRARY, IX_ERROR_LEVEL_GLOBAL);
            break;
            
        case _IX_CC_ATMSAR_MSG_VC_REMOVE:
            IX_ERROR_CT(_ix_cc_atmsar_main_rec_vc_remove(&fsm), err,
                        IX_CC_ATMSAR_ERROR_MSG_LIBRARY, IX_ERROR_LEVEL_GLOBAL);
            break;
        
        case _IX_CC_ATMSAR_MSG_GET_VC_STATS:
            IX_ERROR_CT(_ix_cc_atmsar_main_rec_vc_stats(&fsm), err,
                        IX_CC_ATMSAR_ERROR_MSG_LIBRARY, IX_ERROR_LEVEL_GLOBAL);
            break;

        case _IX_CC_ATMSAR_MSG_PORT_CREATE:
            IX_ERROR_CT(_ix_cc_atmsar_main_rec_port_create(&fsm), err,
                        IX_CC_ATMSAR_ERROR_MSG_LIBRARY, IX_ERROR_LEVEL_GLOBAL);
            break;
        
        case _IX_CC_ATMSAR_MSG_PORT_REMOVE:
            IX_ERROR_CT(_ix_cc_atmsar_main_rec_port_remove(&fsm), err,
                        IX_CC_ATMSAR_ERROR_MSG_LIBRARY, IX_ERROR_LEVEL_GLOBAL);
            break;
        
        case _IX_CC_ATMSAR_MSG_GET_PORT_STATS:
            IX_ERROR_CT(_ix_cc_atmsar_main_rec_port_stats(&fsm), err,
                        IX_CC_ATMSAR_ERROR_MSG_LIBRARY, IX_ERROR_LEVEL_GLOBAL);
            break;

        case _IX_CC_ATMSAR_MSG_CID_CREATE:
            IX_ERROR_CT(_ix_cc_atmsar_main_rec_cid_create(&fsm), err,
                        IX_CC_ATMSAR_ERROR_MSG_LIBRARY, IX_ERROR_LEVEL_GLOBAL);
            break;
        
        case _IX_CC_ATMSAR_MSG_CID_UPDATE:
            IX_ERROR_CT(_ix_cc_atmsar_main_rec_cid_update(&fsm), err,
                        IX_CC_ATMSAR_ERROR_MSG_LIBRARY, IX_ERROR_LEVEL_GLOBAL);
            break;

        case _IX_CC_ATMSAR_MSG_CID_REMOVE:
            IX_ERROR_CT(_ix_cc_atmsar_main_rec_cid_remove(&fsm), err,
                        IX_CC_ATMSAR_ERROR_MSG_LIBRARY, IX_ERROR_LEVEL_GLOBAL);
            break;
        
        case _IX_CC_ATMSAR_MSG_GET_CID_STATS:
            IX_ERROR_CT(_ix_cc_atmsar_main_rec_cid_stats(&fsm), err,
                        IX_CC_ATMSAR_ERROR_MSG_LIBRARY, IX_ERROR_LEVEL_GLOBAL);
            break;

        default:
            break;
        }
        if ((err != IX_SUCCESS) && (fsm.pContext->err == IX_SUCCESS))
        {
            fsm.pContext->err=err;
        }
        if (fsm.stage==_IX_CC_ATMSAR_FSM_STAGE_PREPARE_REPLY)
        {
            /* The reply is ready in buffer. Send return message. */
            
            IX_MSUP_EXTRACT_MSG(fsm.pContext->hBuf, &fsm.pMsg, &pRemoteContext);
            if (pRemoteContext == NULL)
            {
                return IX_ERROR_GLOBAL(IX_CC_ERROR_NULL, ("pRemoteContext is NULL"));
            }
            if ( ix_cc_msup_send_reply_msg(pRemoteContext, fsm.pReplyMsg,
                                           fsm.sizeBufferToSend, fsm.pContext->err)
                 != IX_SUCCESS)
            {
                err=IX_ERROR_WARNING(IX_CC_ATMSAR_ERROR_MSG_LIBRARY, ("Send reply message failed"));
                break;
            }/* endif (err != IX_SUCCESS) */
        }
    }

    /* free uunsed hBuf of request from user module*/
    IX_ERROR_CRT(ix_rm_buffer_free_chain(fsm.pContext->hBuf), IX_CC_ATMSAR_ERROR_BUFFER_FREE, 0);
    /* free unused, processed context */
    _ix_cc_atmsar_main_free_ctx_to_agent(fsm.pContext);

    return err;
} /* end _ix_cc_atmsar_main_cb_from_agent */


/**
 * NAME: _ix_cc_atmsar_main_process_user_req
 *
 * DESCRIPTION:
 *    The function processes incoming all asynchronous messages. It calls
 *    user function for executing next stages of processing the message..
 * 
 * @Param:  IN  arg_hBuf: buffer handle embedding information for the
 *                  message.
 * @Param:  IN  arg_pMyContext: pointer to the ATM SAR Context
 * @Param:  IN  arg_pfsmCb: pointer to user callback function that
 *                  process the message
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully local processed.
 *    IX_CC_ATMSAR_ERROR_BUFFER_FREE - cannot free incoming buffer
 *    IX_CC_ATMSAR_ERROR_MSG_LIBRARY - cannot send reply to calling core
 *        component.
 *    IX_CC_ERROR_NULL - context for calling core component is NULL. It
 *        cannot send reply
 */
static ix_error
_ix_cc_atmsar_main_process_user_req(
ix_buffer_handle arg_hBuf,
_ix_cc_atmsar_main_context_t *arg_pMyContext,
_ix_cc_atmsar_fsm_cb_t arg_pfsmCb)
{
    _ix_cc_atmsar_main_fsm_t fsm;
    void *pRemoteContext;
    ix_error err = IX_SUCCESS;

#ifdef ATMSAR_MAIN_WITH_AGENT_SUPPORT
    ix_communication_id commId;	
#endif /* ATMSAR_MAIN_WITH_AGENT_SUPPORT */
    
    memset(&fsm, 0, sizeof(fsm));
   
    IX_MSUP_EXTRACT_MSG(arg_hBuf, &fsm.pMsg, &pRemoteContext);

    if (pRemoteContext == NULL)
    {
        return IX_ERROR_GLOBAL(IX_CC_ERROR_NULL, ("pRemoteContext is NULL"));
    }

    if (fsm.pMsg == NULL)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("Message pointer is NULL"));
        goto error_support_no_process_fsm;
    }

    if (arg_pfsmCb == NULL)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("fsm callback is not defined"));
        goto error_support_no_process_fsm;
    }

    /* allocate new local context for working */
    if ((fsm.pContext=_ix_cc_atmsar_main_alloc_ctx_to_agent()) == NULL)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("No available context for async. oper"));
        goto error_support_no_process_fsm;
    }

    fsm.pContext->hBuf=arg_hBuf;
    fsm.pContext->pMyContext=arg_pMyContext;
    fsm.pContext->err=IX_SUCCESS;
    fsm.stage=_IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE;

    if ((err=arg_pfsmCb(&fsm)) != IX_SUCCESS)
    {
        goto error_support_process_fsm;
    }
  
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    if ( (fsm.pContext->msgCode==0) || (fsm.sizeBufferToSend == 0) )
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("ASSERT: (fsm.pContext->msgCode==0) || (fsm.sizeBufferToSend == 0)"));
    }
#else
    assert(fsm.pContext->msgCode != 0); /* message code has to be updated in the stage */
    assert(fsm.sizeBufferToSend != 0); /* size of message to send has to be updated in the stage */
#endif
    
#ifdef ATMSAR_MAIN_WITH_AGENT_SUPPORT

#ifndef ATMSAR_AGENT_REMOTE
	commId = IX_RM_COMM_MAKE_LOCAL_ID(IX_CC_ATMSAR_AGENT_MSG_ID); 
#else
	commId = IX_RM_COMM_MAKE_ID(IX_CC_ATMSAR_AGENT_MSG_ID,IX_PEER_SUBSYSTEM, 0); 
#endif

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(commId,
                                           _ix_cc_atmsar_main_cb_from_agent,
                                           (void*)fsm.pContext,
                                           fsm.pContext->msgCode,
                                           (void *)(&fsm.data.req),
                                           fsm.sizeBufferToSend ),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING,
                 error_support_process_fsm);
#endif /* ATMSAR_MAIN_WITH_AGENT_SUPPORT */
    
    /*
     * early we send request to remote side. Now we are able to process the
     * request local
     */
    fsm.stage=_IX_CC_ATMSAR_FSM_STAGE_LOCAL_PROCESS;
    fsm.pContext->err = arg_pfsmCb(&fsm);

#ifndef ATMSAR_MAIN_WITH_AGENT_SUPPORT
    _ix_cc_atmsar_main_cb_from_agent(IX_SUCCESS, fsm.pContext, NULL, 0);
#endif /* ATMSAR_MAIN_WITH_AGENT_SUPPORT */
    return IX_SUCCESS;

  error_support_process_fsm:
    /*
     * The action is processed only if the function did not send a message to
     * agent side. And for local side, was processed only stage
     * _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE.
     */
    fsm.pContext->err=err;
    fsm.stage=_IX_CC_ATMSAR_FSM_STAGE_ERROR_WITHOUT_SEND;
    arg_pfsmCb(&fsm);
    
    /* free unused, processed context */
    _ix_cc_atmsar_main_free_ctx_to_agent(fsm.pContext);

  error_support_no_process_fsm:
    /*
     * Send return message, returned data is not processed if err code
     * is error. Therefore we send only 1 byte
     */
    IX_ERROR_CRT(ix_cc_msup_send_reply_msg(pRemoteContext, pRemoteContext, 4, err),
                 IX_CC_ATMSAR_ERROR_MSG_LIBRARY, 0);
    IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hBuf), IX_CC_ATMSAR_ERROR_BUFFER_FREE, 0);
    return IX_SUCCESS;
} /* end _ix_cc_atmsar_main_process_user_req */


/******************************************************************************
 *          INTERNAL FUNCTIONS - FUNCTION PROCESSING INCOMING MESSAGES
 *****************************************************************************/


/**
 * NAME: _ix_cc_atmsar_main_rec_vc_create
 *
 * DESCRIPTION:
 *    The callback function processes incoming asynchronous messages
 *    'VC create'. The callback is called by the fsm mechanism.
 * 
 * @Param:  IN  arg_pFsm: the pointer to structure describing fsm for
 *                  processed message.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error
_ix_cc_atmsar_main_rec_vc_create(
_ix_cc_atmsar_main_fsm_t* arg_pFsm)
{
    ix_error err = IX_SUCCESS;
    _ix_cc_atmsar_main_context_t *pMyCtx=arg_pFsm->pContext->pMyContext;
    _ix_cc_atmsar_async_vc_create_request_t* pMsg;
    /* pMyCtx contains pointer to ATM SAR MAIN context */
    
    switch (arg_pFsm->stage)
    {
    case _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE:
        /*
         * Fills general context of request allocated by the FSM mechanism
         * with data appropriate to the request (set type of request, set
         * size of structures exchange between AGENT and MAIN, If it is
         * functions servicing FSM that create any object, it should calls
         * allocation of handle in the stage. The new allocated handle
         * should be stored in context data
         */

        /* arg_pFsm->data.req contains data requested by agent to process
         * the request */
        
        pMsg=arg_pFsm->pMsg;
        arg_pFsm->pContext->pOperTxt="VC CREATE";
        arg_pFsm->pContext->msgCode=_IX_CC_ATMSAR_MSG_VC_CREATE;
        arg_pFsm->sizeBufferToSend=sizeof(arg_pFsm->data.req.vcCreate);
        memcpy(&arg_pFsm->data.req.vcCreate.descr, &(pMsg->descr),
               sizeof(arg_pFsm->data.req.vcCreate.descr));
        arg_pFsm->data.req.vcCreate.sarInstance=pMsg->sarInstance;
        arg_pFsm->pContext->data.hVc=IX_CC_ATMSAR_VCID_INVALID;
        
        if (pMyCtx->engine.pCbVcAlloc != NULL)
        {
            err=pMyCtx->engine.pCbVcAlloc(pMyCtx->engine.pVcHandleMgrCtx,
                                          &arg_pFsm->pContext->data.hVc,
                                          &(pMsg->descr));
        }
        else
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("ERROR. Functionality REMOTE VC POOL is not implemented yet\n"));
        }
        
        if ( err != IX_SUCCESS)
        {
            return err;
        }
        arg_pFsm->data.req.vcCreate.handle=arg_pFsm->pContext->data.hVc;
        break;
        
    case _IX_CC_ATMSAR_FSM_STAGE_LOCAL_PROCESS:
        /*
         * Request to AGENT is sent (if current instance of ATM SAR supports
         * ATM SAR agent). In the stage ATM SAR MAIN has to call engine to
         * process the request on local registered plugins. It reads data from
         * context prepared in previous stage, and passes theirs into called
         * engine function.
         */
        pMsg=arg_pFsm->pMsg;
        return _ix_cc_atmsar_engine_vc_create(&(pMyCtx->engine),
                                              arg_pFsm->pContext->data.hVc,
                                              &(pMsg->descr));
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_ERROR_WITHOUT_SEND:
        /*
         * The sending request to AGENT faild. FSM function has to free all
         * resources allocated during stage _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE.
         */
        if (pMyCtx->engine.pCbVcFree != NULL)
        {
            err=pMyCtx->engine.pCbVcFree(pMyCtx->engine.pVcHandleMgrCtx,
                                         arg_pFsm->pContext->data.hVc);
        }
        else
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                    ("ERROR. Functionality REMOTE VC POOL is not implemented yet\n"));
        }
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_PREPARE_REPLY:
        /*
         * Fill buffer of reply message with data collected during processing the
         * request. Set size of reply.
         */
        
        arg_pFsm->pReplyMsg = &arg_pFsm->data.reply.vcCreate;
        arg_pFsm->sizeBufferToSend = sizeof(arg_pFsm->data.reply.vcCreate);
        arg_pFsm->data.reply.vcCreate.handle = arg_pFsm->pContext->data.hVc;
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_AFTER_REPLY:
        /*
         * The reply has sent to calling module. Free all resources extra
         * allocated for processing the request. Check status of processing
         * the request (arg_pFsm->pContext->err). If it is not IX_SUCCESS,
         * free the object. 'Get statistic' for object should not cause
         * freeing object.
         */
        
        if (arg_pFsm->pContext->err != IX_SUCCESS)
        {
            _ix_cc_atmsar_main_vc_clear(arg_pFsm->pContext->pMyContext,
                                        arg_pFsm->pContext->data.hVc);            
        }
        break;

    default:
        break;
    }
    return IX_SUCCESS;
} /* end _ix_cc_atmsar_main_rec_vc_create */

/**
 * NAME: _ix_cc_atmsar_main_rec_vc_update
 *
 * DESCRIPTION:
 *    The callback function processes incoming asynchronous messages
 *    'VC update'. The callback is called by the fsm mechanism.
 * 
 * @Param:  IN  arg_pFsm: the pointer to structure describing fsm for
 *                  processed message.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error
_ix_cc_atmsar_main_rec_vc_update(
_ix_cc_atmsar_main_fsm_t* arg_pFsm)
{
    ix_error err = IX_SUCCESS;
    _ix_cc_atmsar_main_context_t *pMyCtx=arg_pFsm->pContext->pMyContext;
    _ix_cc_atmsar_async_vc_update_request_t* pMsg;
    /* pMyCtx contains pointer to ATM SAR MAIN context */
    
    switch (arg_pFsm->stage)
    {
    case _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE:
        /*
         * Fills general context of request allocated by the FSM mechanism
         * with data appropriate to the request (set type of request, set
         * size of structures exchange between AGENT and MAIN, If it is
         * functions servicing FSM that create any object, it should calls
         * allocation of handle in the stage. The new allocated handle
         * should be stored in context data
         */

        /* arg_pFsm->data.req contains data requested by agent to process
         * the request */
        
        pMsg=arg_pFsm->pMsg;

        arg_pFsm->pContext->pOperTxt="VC UPDATE";
        arg_pFsm->pContext->msgCode=_IX_CC_ATMSAR_MSG_VC_UPDATE;

        arg_pFsm->sizeBufferToSend=sizeof(arg_pFsm->data.req.vcUpdate);
        
        memcpy(&arg_pFsm->data.req.vcUpdate.updateData, &(pMsg->updateData),
               sizeof(arg_pFsm->data.req.vcUpdate.updateData));

        arg_pFsm->data.req.vcUpdate.sarInstance=pMsg->sarInstance;
        arg_pFsm->data.req.vcUpdate.handle=pMsg->handle;
        break;
        
    case _IX_CC_ATMSAR_FSM_STAGE_LOCAL_PROCESS:
        /*
         * Request to AGENT is sent (if current instance of ATM SAR supports
         * ATM SAR agent). In the stage ATM SAR MAIN has to call engine to
         * process the request on local registered plugins. It reads data from
         * context prepared in previous stage, and passes theirs into called
         * engine function.
         */
        pMsg=arg_pFsm->pMsg;
        return _ix_cc_atmsar_engine_vc_update(&(pMyCtx->engine),
                                              pMsg->handle,
                                              &(pMsg->updateData));
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_ERROR_WITHOUT_SEND:
        /*
         * The sending request to AGENT faild. FSM function has to free all
         * resources allocated during stage _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE.
         */
        if (pMyCtx->engine.pCbVcFree != NULL)
        {
            err=pMyCtx->engine.pCbVcFree(pMyCtx->engine.pVcHandleMgrCtx,
                                         arg_pFsm->pContext->data.hVc);
        }
        else
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                    ("ERROR. Functionality REMOTE VC POOL is not implemented yet\n"));
        }
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_PREPARE_REPLY:
        /*
         * Fill buffer of reply message with data collected during processing the
         * request. Set size of reply.
         */
        arg_pFsm->pReplyMsg = &arg_pFsm->data.reply.vcUpdate;
        arg_pFsm->sizeBufferToSend = sizeof(arg_pFsm->data.reply.vcUpdate);
        arg_pFsm->data.reply.vcUpdate.handle = arg_pFsm->pContext->data.hVc;

        break;

    case _IX_CC_ATMSAR_FSM_STAGE_AFTER_REPLY:
        /*
         * The reply has sent to calling module. Free all resources extra
         * allocated for processing the request. Check status of processing
         * the request (arg_pFsm->pContext->err). If it is not IX_SUCCESS,
         * free the object. 'Get statistic' for object should not cause
         * freeing object.
         */
        if (arg_pFsm->pContext->err != IX_SUCCESS)
        {
            _ix_cc_atmsar_main_vc_clear(arg_pFsm->pContext->pMyContext,
                                        arg_pFsm->pContext->data.hVc);
        }
        break;

    default:
        break;
    }
    return IX_SUCCESS;
} /* end _ix_cc_atmsar_main_rec_vc_update */


/**
 * NAME: _ix_cc_atmsar_main_rec_vc_remove
 *
 * DESCRIPTION:
 *    The callback function processes incoming asynchronous messages
 *    'VC remove'. The callback is called by the fsm mechanism.
 * 
 * @Param:  IN  arg_pFsm: the pointer to structure describing fsm for
 *                  processed message.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error
_ix_cc_atmsar_main_rec_vc_remove(
_ix_cc_atmsar_main_fsm_t* arg_pFsm)
{
    _ix_cc_atmsar_main_context_t *pMyCtx;
    switch (arg_pFsm->stage)
    {
    case _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE:
        /*
         * Fills general context of request allocated by the FSM mechanism
         * with data appropriate to the request (set type of request, set
         * size of structures exchange between AGENT and MAIN, If it is
         * functions servicing FSM that create any object, it should calls
         * allocation of handle in the stage. The new allocated handle
         * should be stored in context data
         */
        arg_pFsm->pContext->pOperTxt="VC REMOVE";
        arg_pFsm->pContext->msgCode=_IX_CC_ATMSAR_MSG_VC_REMOVE;
        arg_pFsm->sizeBufferToSend=sizeof(arg_pFsm->data.req.vcRemove);
        arg_pFsm->pContext->data.hVc=((_ix_cc_atmsar_async_vc_remove_request_t*)(arg_pFsm->pMsg))->handle;
        arg_pFsm->data.req.vcRemove.handle=arg_pFsm->pContext->data.hVc;
        break;
        
    case _IX_CC_ATMSAR_FSM_STAGE_LOCAL_PROCESS:
        /*
         * Request to AGENT is sent (if current instance of ATM SAR supports
         * ATM SAR agent). In the stage ATM SAR MAIN has to call engine to
         * process the request on local registered plugins. It reads data from
         * context prepared in previous stage, and passes theirs into called
         * engine function.
         */
        return _ix_cc_atmsar_engine_vc_remove(&(arg_pFsm->pContext->pMyContext->engine),
                                              arg_pFsm->pContext->data.hVc);
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_ERROR_WITHOUT_SEND:
        /*
         * The sending request to AGENT faild. FSM function has to free all
         * resources allocated during stage _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE.
         */

        /* TBD - it is freeing point of hVc */
        _IX_CC_ATMSAR_DBG_MAIN(("_ix_cc_atmsar_main_rec_vc_remove(): ERROR."
                                "Functionality is not implemented yet\n"));
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_PREPARE_REPLY:
        /*
         * Fill buffer of reply message with data collected during processing the
         * request. Set size of reply.
         */
        arg_pFsm->pReplyMsg = &arg_pFsm->data.reply.vcRemove;
        arg_pFsm->sizeBufferToSend = sizeof(arg_pFsm->data.reply.vcRemove);
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_AFTER_REPLY:
        /*
         * The reply has sent to calling module. Free all resources extra
         * allocated for processing the request. Check status of processing
         * the request (arg_pFsm->pContext->err). If it is not IX_SUCCESS,
         * free the object. 'Get statistic' for object should not cause
         * freeing object.
         */
        pMyCtx=arg_pFsm->pContext->pMyContext;
        if (arg_pFsm->pContext->err != IX_SUCCESS)
        {
            _ix_cc_atmsar_main_vc_clear(arg_pFsm->pContext->pMyContext,
                                        arg_pFsm->pContext->data.hVc);
        }
        if (pMyCtx->engine.pCbVcFree != NULL)
        {
            return pMyCtx->engine.pCbVcFree(pMyCtx->engine.pVcHandleMgrCtx,
                                            arg_pFsm->pContext->data.hVc);
        }
        else
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                    ("ERROR. Functionality REMOTE VC POOL is not implemented yet\n"));
        }
        break;

    default:
        break;
    }
    return IX_SUCCESS;
} /* end _ix_cc_atmsar_main_rec_vc_remove */


/**
 * NAME: _ix_cc_atmsar_main_rec_vc_stats
 *
 * DESCRIPTION:
 *    The callback function processes incoming asynchronous messages
 *    'VC stats'. The callback is called by the fsm mechanism.
 * 
 * @Param:  IN  arg_pFsm: the pointer to structure describing fsm for
 *                  processed message.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error
_ix_cc_atmsar_main_rec_vc_stats(
_ix_cc_atmsar_main_fsm_t* arg_pFsm)
{
    switch (arg_pFsm->stage)
    {
    case _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE:
        /*
         * Fills general context of request allocated by the FSM mechanism
         * with data appropriate to the request (set type of request, set
         * size of structures exchange between AGENT and MAIN, If it is
         * functions servicing FSM that create any object, it should calls
         * allocation of handle in the stage. The new allocated handle
         * should be stored in context data
         */
        arg_pFsm->pContext->pOperTxt="VC STATS";
        arg_pFsm->pContext->data.pVcStats=ix_ossl_malloc(sizeof(_ix_cc_atmsar_async_vc_stats_reply_t));
        if (arg_pFsm->pContext->data.pVcStats == NULL)
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                    ("ERROR. Can't allocate memory for statistics\n"));
        }
        ix_ossl_memset(arg_pFsm->pContext->data.pVcStats, 0,
                       sizeof(_ix_cc_atmsar_async_vc_stats_reply_t));
        arg_pFsm->pContext->msgCode=_IX_CC_ATMSAR_MSG_GET_VC_STATS;
        arg_pFsm->sizeBufferToSend=sizeof(arg_pFsm->data.req.vcStats);
        arg_pFsm->data.req.vcStats.handle=((_ix_cc_atmsar_async_vc_stats_request_t*)(arg_pFsm->pMsg))->handle;
        break;
        
    case _IX_CC_ATMSAR_FSM_STAGE_LOCAL_PROCESS:
        /*
         * Request to AGENT is sent (if current instance of ATM SAR supports
         * ATM SAR agent). In the stage ATM SAR MAIN has to call engine to
         * process the request on local registered plugins. It reads data from
         * context prepared in previous stage, and passes theirs into called
         * engine function.
         */
        return _ix_cc_atmsar_engine_get_vc_stats(&(arg_pFsm->pContext->pMyContext->engine),
                                                 arg_pFsm->data.req.vcStats.handle,
                                                 &arg_pFsm->pContext->data.pVcStats->stats);
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_ERROR_WITHOUT_SEND:
        /*
         * The sending request to AGENT faild. FSM function has to free all
         * resources allocated during stage _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE.
         */
        if ((arg_pFsm->pContext != NULL) && (arg_pFsm->pContext->data.pVcStats != NULL))
        {
            ix_ossl_free(arg_pFsm->pContext->data.pVcStats);
            arg_pFsm->pContext->data.pVcStats=NULL;
        }
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_PREPARE_REPLY:
        /*
         * Fill buffer of reply message with data collected during processing the
         * request. Set size of reply.
         */
        arg_pFsm->pReplyMsg = arg_pFsm->pContext->data.pVcStats;
        arg_pFsm->sizeBufferToSend = sizeof(_ix_cc_atmsar_async_vc_stats_reply_t);
#ifdef ATMSAR_MAIN_WITH_AGENT_SUPPORT
#ifdef ATMSAR_AGENT_ON_EGRESS
        memcpy(&arg_pFsm->pContext->data.pVcStats->stats.tx,
               &((_ix_cc_atmsar_agent_vc_stats_reply_t*)arg_pFsm->pMsg)->stats.tx,
               sizeof(arg_pFsm->pContext->data.pVcStats->stats.tx));
#else /* ATMSAR_AGENT_ON_EGRESS */
        memcpy(&arg_pFsm->pContext->data.pVcStats->stats.rx,
               &((_ix_cc_atmsar_agent_vc_stats_reply_t*)arg_pFsm->pMsg)->stats.rx,
               sizeof(arg_pFsm->pContext->data.pVcStats->stats.rx));
#endif /* ATMSAR_AGENT_ON_EGRESS */
#endif /* ATMSAR_MAIN_WITH_AGENT_SUPPORT */
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_AFTER_REPLY:
        /*
         * The reply has sent to calling module. Free all resources extra
         * allocated for processing the request. Check status of processing
         * the request (arg_pFsm->pContext->err). If it is not IX_SUCCESS,
         * free the object. 'Get statistic' for object should not cause
         * freeing object.
         */
        if ((arg_pFsm->pContext != NULL) && (arg_pFsm->pContext->data.pVcStats != NULL))
        {
            ix_ossl_free(arg_pFsm->pContext->data.pVcStats);
            arg_pFsm->pContext->data.pVcStats=NULL;
        }
        break;

    default:
        break;
    }
    return IX_SUCCESS;
} /* end _ix_cc_atmsar_main_rec_vc_stats */


/**
 * NAME: _ix_cc_atmsar_main_rec_port_create
 *
 * DESCRIPTION:
 *    The callback function processes incoming asynchronous messages
 *    'port create'. The callback is called by the fsm mechanism.
 * 
 * @Param:  IN  arg_pFsm: the pointer to structure describing fsm for
 *                  processed message.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error
_ix_cc_atmsar_main_rec_port_create(
_ix_cc_atmsar_main_fsm_t* arg_pFsm)
{
    ix_error err = IX_SUCCESS;
    _ix_cc_atmsar_async_port_create_request_t* pMsg;
    _ix_cc_atmsar_main_context_t *pMyCtx=arg_pFsm->pContext->pMyContext;
    switch (arg_pFsm->stage)
    {
    case _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE:
        /*
         * Fills general context of request allocated by the FSM mechanism
         * with data appropriate to the request (set type of request, set
         * size of structures exchange between AGENT and MAIN, If it is
         * functions servicing FSM that create any object, it should calls
         * allocation of handle in the stage. The new allocated handle
         * should be stored in context data
         */
        pMsg=arg_pFsm->pMsg;
        arg_pFsm->pContext->pOperTxt="PORT CREATE";
        arg_pFsm->pContext->msgCode=_IX_CC_ATMSAR_MSG_PORT_CREATE;
        arg_pFsm->sizeBufferToSend=sizeof(arg_pFsm->data.req.portCreate);
        memcpy(&arg_pFsm->data.req.portCreate.descr, &(pMsg->descr),
               sizeof(arg_pFsm->data.req.portCreate.descr));
        arg_pFsm->data.req.portCreate.sarInstance=pMsg->sarInstance;
        arg_pFsm->pContext->data.hPort=IX_CC_ATMSAR_PORTID_INVALID;
        if (pMyCtx->engine.pCbPortAlloc != NULL)
        {
            err=pMyCtx->engine.pCbPortAlloc(pMyCtx->engine.pPortHandleMgrCtx,
                                            &arg_pFsm->pContext->data.hPort,
                                            &(pMsg->descr));
        }
        else
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                    ("ERROR. Functionality REMOTE PORT POOL is not implemented yet\n"));
        }
            
        if ( err != IX_SUCCESS)
        {
            return err;
        }
        arg_pFsm->data.req.portCreate.handle=arg_pFsm->pContext->data.hPort;
        break;
        
    case _IX_CC_ATMSAR_FSM_STAGE_LOCAL_PROCESS:
        /*
         * Request to AGENT is sent (if current instance of ATM SAR supports
         * ATM SAR agent). In the stage ATM SAR MAIN has to call engine to
         * process the request on local registered plugins. It reads data from
         * context prepared in previous stage, and passes theirs into called
         * engine function.
         */
        pMsg=arg_pFsm->pMsg;
        return _ix_cc_atmsar_engine_port_create(&(arg_pFsm->pContext->pMyContext->engine),
                                                arg_pFsm->pContext->data.hPort,
                                                &(pMsg->descr));
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_ERROR_WITHOUT_SEND:
        /*
         * The sending request to AGENT faild. FSM function has to free all
         * resources allocated during stage _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE.
         */
        if (pMyCtx->engine.pCbPortFree != NULL)
        {
            err=pMyCtx->engine.pCbPortFree(pMyCtx->engine.pPortHandleMgrCtx,
                                           arg_pFsm->pContext->data.hPort);
        }
        else
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                    ("ERROR. Functionality REMOTE PORT POOL is not implemented yet\n"));
        }
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_PREPARE_REPLY:
        /*
         * Fill buffer of reply message with data collected during processing the
         * request. Set size of reply.
         */
        arg_pFsm->pReplyMsg = &arg_pFsm->data.reply.portCreate;
        arg_pFsm->sizeBufferToSend = sizeof(arg_pFsm->data.reply.portCreate);
        arg_pFsm->data.reply.portCreate.handle = arg_pFsm->pContext->data.hPort;
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_AFTER_REPLY:
        /*
         * The reply has sent to calling module. Free all resources extra
         * allocated for processing the request. Check status of processing
         * the request (arg_pFsm->pContext->err). If it is not IX_SUCCESS,
         * free the object. 'Get statistic' for object should not cause
         * freeing object.
         */
        if (arg_pFsm->pContext->err != IX_SUCCESS)
        {
            _ix_cc_atmsar_main_port_clear(arg_pFsm->pContext->pMyContext,
                                          arg_pFsm->pContext->data.hPort);
        }
        break;

    default:
        break;
    }
    return IX_SUCCESS;
} /* end _ix_cc_atmsar_main_rec_port_create */


/**
 * NAME: _ix_cc_atmsar_main_rec_port_remove
 *
 * DESCRIPTION:
 *    The callback function processes incoming asynchronous messages
 *    'port remove'. The callback is called by the fsm mechanism.
 * 
 * @Param:  IN  arg_pFsm: the pointer to structure describing fsm for
 *                  processed message.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error
_ix_cc_atmsar_main_rec_port_remove(
_ix_cc_atmsar_main_fsm_t* arg_pFsm)
{
    _ix_cc_atmsar_main_context_t *pMyCtx;
    switch (arg_pFsm->stage)
    {
    case _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE:
        /*
         * Fills general context of request allocated by the FSM mechanism
         * with data appropriate to the request (set type of request, set
         * size of structures exchange between AGENT and MAIN, If it is
         * functions servicing FSM that create any object, it should calls
         * allocation of handle in the stage. The new allocated handle
         * should be stored in context data
         */
        arg_pFsm->pContext->pOperTxt="PORT REMOVE";
        arg_pFsm->pContext->msgCode=_IX_CC_ATMSAR_MSG_PORT_REMOVE;
        arg_pFsm->sizeBufferToSend=sizeof(arg_pFsm->data.req.portRemove);
        arg_pFsm->pContext->data.hPort=((_ix_cc_atmsar_async_port_remove_request_t*)(arg_pFsm->pMsg))->handle;
        arg_pFsm->data.req.portRemove.handle=arg_pFsm->pContext->data.hPort;
        break;
        
    case _IX_CC_ATMSAR_FSM_STAGE_LOCAL_PROCESS:
        /*
         * Request to AGENT is sent (if current instance of ATM SAR supports
         * ATM SAR agent). In the stage ATM SAR MAIN has to call engine to
         * process the request on local registered plugins. It reads data from
         * context prepared in previous stage, and passes theirs into called
         * engine function.
         */
        return _ix_cc_atmsar_engine_port_remove(&(arg_pFsm->pContext->pMyContext->engine),
                                                arg_pFsm->pContext->data.hPort);
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_ERROR_WITHOUT_SEND:
        /*
         * The sending request to AGENT faild. FSM function has to free all
         * resources allocated during stage _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE.
         */

        /* TBD - it is freeing point of port handle */
        _IX_CC_ATMSAR_DBG_MAIN(("_ix_cc_atmsar_main_rec_port_remove(): ERROR."
                                "Functionality is not implemented yet\n"));
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_PREPARE_REPLY:
        /*
         * Fill buffer of reply message with data collected during processing the
         * request. Set size of reply.
         */
        arg_pFsm->pReplyMsg = &arg_pFsm->data.reply.portRemove;
        arg_pFsm->sizeBufferToSend = sizeof(arg_pFsm->data.reply.portRemove);
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_AFTER_REPLY:
        /*
         * The reply has sent to calling module. Free all resources extra
         * allocated for processing the request. Check status of processing
         * the request (arg_pFsm->pContext->err). If it is not IX_SUCCESS,
         * free the object. 'Get statistic' for object should not cause
         * freeing object.
         */
        pMyCtx=arg_pFsm->pContext->pMyContext;
        if (arg_pFsm->pContext->err != IX_SUCCESS)
        {
            _ix_cc_atmsar_main_port_clear(arg_pFsm->pContext->pMyContext,
                                          arg_pFsm->pContext->data.hPort);
        }
        if (pMyCtx->engine.pCbPortFree != NULL)
        {
            return pMyCtx->engine.pCbPortFree(pMyCtx->engine.pPortHandleMgrCtx,
                                              arg_pFsm->pContext->data.hPort);
        }
        else
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                    ("ERROR. Functionality REMOTE PORT POOL is not implemented yet\n"));
        }
        break;

    default:
        break;
    }
    return IX_SUCCESS;
} /* end _ix_cc_atmsar_main_rec_port_remove */


/**
 * NAME: _ix_cc_atmsar_main_rec_port_stats
 *
 * DESCRIPTION:
 *    The callback function processes incoming asynchronous messages
 *    'port stats'. The callback is called by the fsm mechanism.
 * 
 * @Param:  IN  arg_pFsm: the pointer to structure describing fsm for
 *                  processed message.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error
_ix_cc_atmsar_main_rec_port_stats(
_ix_cc_atmsar_main_fsm_t* arg_pFsm)
{
    switch (arg_pFsm->stage)
    {
    case _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE:
        /*
         * Fills general context of request allocated by the FSM mechanism
         * with data appropriate to the request (set type of request, set
         * size of structures exchange between AGENT and MAIN, If it is
         * functions servicing FSM that create any object, it should calls
         * allocation of handle in the stage. The new allocated handle
         * should be stored in context data
         */
        arg_pFsm->pContext->pOperTxt="PORT STATS";
        arg_pFsm->pContext->data.pPortStats=(_ix_cc_atmsar_async_port_stats_reply_t*)
            ix_ossl_malloc(sizeof(_ix_cc_atmsar_async_port_stats_reply_t));
        
        if (arg_pFsm->pContext->data.pPortStats == NULL)
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                    ("ERROR. Can't allocate memory for statistics\n"));
        }
        ix_ossl_memset(arg_pFsm->pContext->data.pPortStats, 0,
                       sizeof(_ix_cc_atmsar_async_port_stats_reply_t));

        arg_pFsm->pContext->msgCode=_IX_CC_ATMSAR_MSG_GET_PORT_STATS;
        arg_pFsm->sizeBufferToSend=sizeof(arg_pFsm->data.req.portStats);
        arg_pFsm->data.req.portStats.handle=((_ix_cc_atmsar_async_port_stats_request_t*)
                                             (arg_pFsm->pMsg))->handle;
        break;
        
    case _IX_CC_ATMSAR_FSM_STAGE_LOCAL_PROCESS:
        /*
         * Request to AGENT is sent (if current instance of ATM SAR supports
         * ATM SAR agent). In the stage ATM SAR MAIN has to call engine to
         * process the request on local registered plugins. It reads data from
         * context prepared in previous stage, and passes theirs into called
         * engine function.
         */
        return _ix_cc_atmsar_engine_get_port_stats(&(arg_pFsm->pContext->pMyContext->engine),
                                                   arg_pFsm->data.req.portStats.handle,
                                                   &arg_pFsm->pContext->data.pPortStats->stats);
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_ERROR_WITHOUT_SEND:
        /*
         * The sending request to AGENT faild. FSM function has to free all
         * resources allocated during stage _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE.
         */
        if ((arg_pFsm->pContext != NULL) && (arg_pFsm->pContext->data.pPortStats != NULL))
        {
            ix_ossl_free(arg_pFsm->pContext->data.pPortStats);
            arg_pFsm->pContext->data.pPortStats=NULL;
        }
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_PREPARE_REPLY:
        /*
         * Fill buffer of reply message with data collected during processing the
         * request. Set size of reply.
         */
        arg_pFsm->pReplyMsg = arg_pFsm->pContext->data.pPortStats;
        arg_pFsm->sizeBufferToSend = sizeof(_ix_cc_atmsar_async_port_stats_reply_t);

#ifdef ATMSAR_MAIN_WITH_AGENT_SUPPORT
#ifdef ATMSAR_AGENT_ON_EGRESS
        memcpy(&arg_pFsm->pContext->data.pPortStats->stats.tx,
               &((_ix_cc_atmsar_agent_port_stats_reply_t*)arg_pFsm->pMsg)->stats.tx,
               sizeof(arg_pFsm->pContext->data.pPortStats->stats.tx));
#else /* ATMSAR_AGENT_ON_EGRESS */
        memcpy(&arg_pFsm->pContext->data.pPortStats->stats.rx,
               &((_ix_cc_atmsar_agent_port_stats_reply_t*)arg_pFsm->pMsg)->stats.rx,
               sizeof(arg_pFsm->pContext->data.pPortStats->stats.rx));
#endif /* ATMSAR_AGENT_ON_EGRESS */
#endif /* ATMSAR_MAIN_WITH_AGENT_SUPPORT */
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_AFTER_REPLY:
        /*
         * The reply has sent to calling module. Free all resources extra
         * allocated for processing the request. Check status of processing
         * the request (arg_pFsm->pContext->err). If it is not IX_SUCCESS,
         * free the object. 'Get statistic' for object should not cause
         * freeing object.
         */
        if ((arg_pFsm->pContext != NULL) && (arg_pFsm->pContext->data.pPortStats != NULL))
        {
            ix_ossl_free(arg_pFsm->pContext->data.pPortStats);
            arg_pFsm->pContext->data.pPortStats=NULL;
        }
        break;

    default:
        break;

    }
    return IX_SUCCESS;
} /* end _ix_cc_atmsar_main_rec_port_stats */


/**
 * NAME: _ix_cc_atmsar_main_vc_clear
 *
 * DESCRIPTION:
 *    The function clears VC for VC that VC create operation failed.
 * 
 * @Param:  IN  arg_pMyContext: pointer to the ATM SAR Context
 * @Param:  IN  arg_pMyContext: pointer to context of calling part
 * @Param:  IN  arg_pMsg: pointer to buffer with description of created VC.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error
_ix_cc_atmsar_main_vc_clear(
_ix_cc_atmsar_main_context_t *arg_pMyContext,
ix_cc_atmsar_vc_handle_t arg_hVc)
{
    ix_error err = IX_SUCCESS;

    _IX_CC_ATMSAR_DBG_MAIN(("_ix_cc_atmsar_main_vc_clear(): "
                            "Functionality isn;t implemented yet\n"));
    /*
     *it should:
     * - call local clear operation
     * - call remote clear operation
     * - in callback - free VC handle
     */
    
    return err;
} /* end _ix_cc_atmsar_main_vc_clear */


/**
 * NAME: _ix_cc_atmsar_main_port_clear
 *
 * DESCRIPTION:
 *    The function clears VC for VC that VC create operation failed.
 * 
 * @Param:  IN  arg_pMyContext: pointer to the ATM SAR Context
 * @Param:  IN  arg_pMyContext: pointer to context of calling part
 * @Param:  IN  arg_pMsg: pointer to buffer with description of created VC.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error
_ix_cc_atmsar_main_port_clear(
_ix_cc_atmsar_main_context_t *arg_pMyContext,
ix_uint16 arg_port)
{
    ix_error err = IX_SUCCESS;

    _IX_CC_ATMSAR_DBG_MAIN(("_ix_cc_atmsar_main_port_clear(): "
                            "Functionality isn;t implemented yet\n"));
    /*
     *it should:
     * - call local clear operation
     * - call remote clear operation
     * - in callback - free VC handle
     */
    
    return err;
} /* end _ix_cc_atmsar_main_port_clear */


/******************************************************************************
 *             FUNCTION PROCESSING MESSAGES FOR CID SUPPORT
 *****************************************************************************/


/**
 * NAME: _ix_cc_atmsar_main_rec_cid_create
 *
 * DESCRIPTION:
 *    The callback function processes incoming asynchronous messages
 *    'cid create'. The callback is called by the fsm mechanism.
 * 
 * @Param:  IN  arg_pFsm: the pointer to structure describing fsm for
 *                  processed message.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error
_ix_cc_atmsar_main_rec_cid_create(
_ix_cc_atmsar_main_fsm_t* arg_pFsm)
{
    ix_error err = IX_SUCCESS;
    _ix_cc_atmsar_main_context_t *pMyCtx=arg_pFsm->pContext->pMyContext;
    _ix_cc_atmsar_async_cid_create_request_t* pMsg;
    switch (arg_pFsm->stage)
    {
    case _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE:
        /*
         * Fills general context of request allocated by the FSM mechanism
         * with data appropriate to the request (set type of request, set
         * size of structures exchange between AGENT and MAIN, If it is
         * functions servicing FSM that create any object, it should calls
         * allocation of handle in the stage. The new allocated handle
         * should be stored in context data
         */
        pMsg=arg_pFsm->pMsg;
        arg_pFsm->pContext->pOperTxt="CID CREATE";
        arg_pFsm->pContext->msgCode=_IX_CC_ATMSAR_MSG_CID_CREATE;
        arg_pFsm->sizeBufferToSend=sizeof(arg_pFsm->data.req.cidCreate);
        memcpy(&arg_pFsm->data.req.cidCreate.descr, &(pMsg->descr),
               sizeof(arg_pFsm->data.req.cidCreate.descr));
        arg_pFsm->data.req.cidCreate.sarInstance=pMsg->sarInstance;
        arg_pFsm->pContext->data.hCid=IX_CC_ATMSAR_CIDID_INVALID;
        
        if (pMyCtx->engine.cfg.aal2Included == IX_CC_ATMSAR_PLUGIN_AAL2_INLUDED_NO)
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                    ("The SAR does not support AAL2\n"));
        }
        
        if (pMyCtx->engine.pCbCidAlloc != NULL)
        {
            err=pMyCtx->engine.pCbCidAlloc(pMyCtx->engine.pCidHandleMgrCtx,
                                           &arg_pFsm->pContext->data.hCid,
                                           &(pMsg->descr));
        }
        else
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                    ("ERROR. Functionality REMOTE CID POOL is not implemented yet\n"));
        }
            
        if ( err != IX_SUCCESS)
        {
            return err;
        }
        arg_pFsm->data.req.cidCreate.handle=arg_pFsm->pContext->data.hCid;
        break;
        
    case _IX_CC_ATMSAR_FSM_STAGE_LOCAL_PROCESS:
        /*
         * Request to AGENT is sent (if current instance of ATM SAR supports
         * ATM SAR agent). In the stage ATM SAR MAIN has to call engine to
         * process the request on local registered plugins. It reads data from
         * context prepared in previous stage, and passes theirs into called
         * engine function.
         */
        pMsg=arg_pFsm->pMsg;
        return _ix_cc_atmsar_engine_cid_create(&(arg_pFsm->pContext->pMyContext->engine),
                                               arg_pFsm->pContext->data.hCid,
                                               &(pMsg->descr));
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_ERROR_WITHOUT_SEND:
        /*
         * The sending request to AGENT faild. FSM function has to free all
         * resources allocated during stage _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE.
         */
        if (arg_pFsm->pContext->data.hCid==IX_CC_ATMSAR_CIDID_INVALID)
        {
            return IX_SUCCESS;
        }

        if (pMyCtx->engine.pCbCidFree != NULL)
        {
            err=pMyCtx->engine.pCbCidFree(pMyCtx->engine.pCidHandleMgrCtx,
                                          arg_pFsm->pContext->data.hCid);
        }
        else
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                    ("ERROR. Functionality REMOTE CID POOL is not implemented yet\n"));
        }
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_PREPARE_REPLY:
        /*
         * Fill buffer of reply message with data collected during processing the
         * request. Set size of reply.
         */
        arg_pFsm->pReplyMsg = &arg_pFsm->data.reply.cidCreate;
        arg_pFsm->sizeBufferToSend = sizeof(arg_pFsm->data.reply.cidCreate);
        arg_pFsm->data.reply.cidCreate.handle = arg_pFsm->pContext->data.hCid;
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_AFTER_REPLY:
        /*
         * The reply has sent to calling module. Free all resources extra
         * allocated for processing the request. Check status of processing
         * the request (arg_pFsm->pContext->err). If it is not IX_SUCCESS,
         * free the object. 'Get statistic' for object should not cause
         * freeing object.
         */
        if (arg_pFsm->pContext->err != IX_SUCCESS)
        {
            _ix_cc_atmsar_main_cid_clear(arg_pFsm->pContext->pMyContext,
                                         arg_pFsm->pContext->data.hCid);
        }
        break;

    default:
        break;

    }
    return IX_SUCCESS;
} /* end _ix_cc_atmsar_main_rec_cid_create */

/**
 * NAME: _ix_cc_atmsar_main_rec_cid_create
 *
 * DESCRIPTION:
 *    The callback function processes incoming asynchronous messages
 *    'cid update'. The callback is called by the fsm mechanism.
 * 
 * @Param:  IN  arg_pFsm: the pointer to structure describing fsm for
 *                  processed message.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error
_ix_cc_atmsar_main_rec_cid_update(
_ix_cc_atmsar_main_fsm_t* arg_pFsm)
{
    ix_error err = IX_SUCCESS;
    _ix_cc_atmsar_main_context_t *pMyCtx=arg_pFsm->pContext->pMyContext;
    _ix_cc_atmsar_async_cid_update_request_t* pMsg;
    switch (arg_pFsm->stage)
    {
    case _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE:
        /*
         * Fills general context of request allocated by the FSM mechanism
         * with data appropriate to the request (set type of request, set
         * size of structures exchange between AGENT and MAIN, If it is
         * functions servicing FSM that create any object, it should calls
         * allocation of handle in the stage. The new allocated handle
         * should be stored in context data
         */
        pMsg=arg_pFsm->pMsg;
        arg_pFsm->pContext->pOperTxt="CID UPDATE";
        arg_pFsm->pContext->msgCode=_IX_CC_ATMSAR_MSG_CID_UPDATE;
        arg_pFsm->sizeBufferToSend=sizeof(arg_pFsm->data.req.cidUpdate);
        memcpy(&arg_pFsm->data.req.cidUpdate.update, &(pMsg->updateData),
               sizeof(arg_pFsm->data.req.cidUpdate.update));

        arg_pFsm->pContext->data.hCid=pMsg->handle;
        
        if (pMyCtx->engine.cfg.aal2Included == IX_CC_ATMSAR_PLUGIN_AAL2_INLUDED_NO)
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                    ("The SAR does not support AAL2\n"));
        }

        arg_pFsm->data.req.cidUpdate.sarInstance=pMsg->sarInstance;
        arg_pFsm->data.req.cidUpdate.handle=pMsg->handle;
        break;
        
    case _IX_CC_ATMSAR_FSM_STAGE_LOCAL_PROCESS:
        /*
         * Request to AGENT is sent (if current instance of ATM SAR supports
         * ATM SAR agent). In the stage ATM SAR MAIN has to call engine to
         * process the request on local registered plugins. It reads data from
         * context prepared in previous stage, and passes theirs into called
         * engine function.
         */
        pMsg=arg_pFsm->pMsg;
        return _ix_cc_atmsar_engine_cid_update(&(arg_pFsm->pContext->pMyContext->engine),
                                               arg_pFsm->pContext->data.hCid,
                                               &(pMsg->updateData));
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_ERROR_WITHOUT_SEND:
        /*
         * The sending request to AGENT faild. FSM function has to free all
         * resources allocated during stage _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE.
         */
        if (arg_pFsm->pContext->data.hCid==IX_CC_ATMSAR_CIDID_INVALID)
        {
            return IX_SUCCESS;
        }

        if (pMyCtx->engine.pCbCidFree != NULL)
        {
            err=pMyCtx->engine.pCbCidFree(pMyCtx->engine.pCidHandleMgrCtx,
                                          arg_pFsm->pContext->data.hCid);
        }
        else
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                    ("ERROR. Functionality REMOTE CID POOL is not implemented yet\n"));
        }
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_PREPARE_REPLY:
        /*
         * Fill buffer of reply message with data collected during processing the
         * request. Set size of reply.
         */
        arg_pFsm->pReplyMsg = &arg_pFsm->data.reply.cidUpdate;
        arg_pFsm->sizeBufferToSend = sizeof(arg_pFsm->data.reply.cidUpdate);
        arg_pFsm->data.reply.cidUpdate.handle = arg_pFsm->pContext->data.hCid;
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_AFTER_REPLY:
        /*
         * The reply has sent to calling module. Free all resources extra
         * allocated for processing the request. Check status of processing
         * the request (arg_pFsm->pContext->err). If it is not IX_SUCCESS,
         * free the object. 'Get statistic' for object should not cause
         * freeing object.
         */
        if (arg_pFsm->pContext->err != IX_SUCCESS)
        {
            _ix_cc_atmsar_main_cid_clear(arg_pFsm->pContext->pMyContext,
                                         arg_pFsm->pContext->data.hCid);
        }
        break;

    default:
        break;

    }
    return IX_SUCCESS;
} /* end _ix_cc_atmsar_main_rec_cid_update */

/**
 * NAME: _ix_cc_atmsar_main_rec_cid_remove
 *
 * DESCRIPTION:
 *    The callback function processes incoming asynchronous messages
 *    'cid remove'. The callback is called by the fsm mechanism.
 * 
 * @Param:  IN  arg_pFsm: the pointer to structure describing fsm for
 *                  processed message.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error
_ix_cc_atmsar_main_rec_cid_remove(
_ix_cc_atmsar_main_fsm_t* arg_pFsm)
{
    _ix_cc_atmsar_main_context_t *pMyCtx;
    switch (arg_pFsm->stage)
    {
    case _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE:
        /*
         * Fills general context of request allocated by the FSM mechanism
         * with data appropriate to the request (set type of request, set
         * size of structures exchange between AGENT and MAIN, If it is
         * functions servicing FSM that create any object, it should calls
         * allocation of handle in the stage. The new allocated handle
         * should be stored in context data
         */
        arg_pFsm->pContext->pOperTxt="CID REMOVE";
        arg_pFsm->pContext->msgCode=_IX_CC_ATMSAR_MSG_CID_REMOVE;
        arg_pFsm->sizeBufferToSend=sizeof(arg_pFsm->data.req.cidRemove);
        arg_pFsm->pContext->data.hCid=((_ix_cc_atmsar_async_port_remove_request_t*)(arg_pFsm->pMsg))->handle;
        arg_pFsm->data.req.cidRemove.handle=arg_pFsm->pContext->data.hCid;
        if (pMyCtx->engine.cfg.aal2Included == IX_CC_ATMSAR_PLUGIN_AAL2_INLUDED_NO)
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                    ("The SAR does not support AAL2\n"));
        }
        break;
        
    case _IX_CC_ATMSAR_FSM_STAGE_LOCAL_PROCESS:
        /*
         * Request to AGENT is sent (if current instance of ATM SAR supports
         * ATM SAR agent). In the stage ATM SAR MAIN has to call engine to
         * process the request on local registered plugins. It reads data from
         * context prepared in previous stage, and passes theirs into called
         * engine function.
         */
        return _ix_cc_atmsar_engine_cid_remove(&(arg_pFsm->pContext->pMyContext->engine),
                                               arg_pFsm->pContext->data.hCid);
        break;
        
    case _IX_CC_ATMSAR_FSM_STAGE_ERROR_WITHOUT_SEND:
        /*
         * The sending request to AGENT faild. FSM function has to free all
         * resources allocated during stage _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE.
         */

        /* TBD - it is freeing point of port handle */
        _IX_CC_ATMSAR_DBG_MAIN(("_ix_cc_atmsar_main_rec_cid_remove(): ERROR."
                                "Functionality is not implemented yet\n"));
        break;
        
    case _IX_CC_ATMSAR_FSM_STAGE_PREPARE_REPLY:
        /*
         * Fill buffer of reply message with data collected during processing the
         * request. Set size of reply.
         */
        arg_pFsm->pReplyMsg = &arg_pFsm->data.reply.cidRemove;
        arg_pFsm->sizeBufferToSend = sizeof(arg_pFsm->data.reply.cidRemove);
        break;

    case _IX_CC_ATMSAR_FSM_STAGE_AFTER_REPLY:
        /*
         * The reply has sent to calling module. Free all resources extra
         * allocated for processing the request. Check status of processing
         * the request (arg_pFsm->pContext->err). If it is not IX_SUCCESS,
         * free the object. 'Get statistic' for object should not cause
         * freeing object.
         */
        pMyCtx=arg_pFsm->pContext->pMyContext;
        if (arg_pFsm->pContext->err != IX_SUCCESS)
        {
            _ix_cc_atmsar_main_cid_clear(arg_pFsm->pContext->pMyContext,
                                         arg_pFsm->pContext->data.hCid);
        }
        if (pMyCtx->engine.pCbCidFree != NULL)
        {
            return pMyCtx->engine.pCbCidFree(pMyCtx->engine.pCidHandleMgrCtx,
                                             arg_pFsm->pContext->data.hCid);
        }
        else
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                    ("ERROR. Functionality REMOTE VC POOL is not implemented yet\n"));
        }
        break;

    default:
        break;

    }
    return IX_SUCCESS;
} /* end _ix_cc_atmsar_main_rec_cid_remove */

/**
 * NAME: _ix_cc_atmsar_main_rec_cid_stats
 *
 * DESCRIPTION:
 *    The callback function processes incoming asynchronous messages
 *    'CID STATS'. The callback is called by the fsm mechanism.
 * 
 * @Param:  IN  arg_pFsm: the pointer to structure describing fsm for
 *                  processed message.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error
_ix_cc_atmsar_main_rec_cid_stats(
_ix_cc_atmsar_main_fsm_t *arg_pFsm)
{
    switch (arg_pFsm->stage)
    {
    case _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE:
        /*
         * Fills general context of request allocated by the FSM mechanism
         * with data appropriate to the request (set type of request, set
         * size of structures exchange between AGENT and MAIN, If it is
         * functions servicing FSM that create any object, it should calls
         * allocation of handle in the stage. The new allocated handle
         * should be stored in context data
         */
        arg_pFsm->pContext->pOperTxt="CID STATS";
        arg_pFsm->pContext->data.pCidStats=(_ix_cc_atmsar_async_cid_stats_reply_t*)
            ix_ossl_malloc(sizeof(_ix_cc_atmsar_async_cid_stats_reply_t));
        
        if (arg_pFsm->pContext->data.pCidStats == NULL)
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                    ("ERROR. Can't allocate memory for statistics\n"));
        }
        ix_ossl_memset(arg_pFsm->pContext->data.pCidStats, 0,
                       sizeof(_ix_cc_atmsar_async_cid_stats_reply_t));

        arg_pFsm->pContext->msgCode=_IX_CC_ATMSAR_MSG_GET_CID_STATS;
        arg_pFsm->sizeBufferToSend=sizeof(arg_pFsm->data.req.cidStats);
        arg_pFsm->data.req.cidStats.handle=((_ix_cc_atmsar_async_cid_stats_request_t*)
                                            (arg_pFsm->pMsg))->handle;
        if (arg_pFsm->pContext->pMyContext->engine.cfg.aal2Included == IX_CC_ATMSAR_PLUGIN_AAL2_INLUDED_NO)
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                    ("The SAR does not support AAL2\n"));
        }
        break;
        
    case _IX_CC_ATMSAR_FSM_STAGE_LOCAL_PROCESS:
        /*
         * Request to AGENT is sent (if current instance of ATM SAR supports
         * ATM SAR agent). In the stage ATM SAR MAIN has to call engine to
         * process the request on local registered plugins. It reads data from
         * context prepared in previous stage, and passes theirs into called
         * engine function.
         */
        return _ix_cc_atmsar_engine_get_cid_stats(&(arg_pFsm->pContext->pMyContext->engine),
                                                  arg_pFsm->data.req.cidStats.handle,
                                                  &arg_pFsm->pContext->data.pCidStats->stats);
        break;
        
    case _IX_CC_ATMSAR_FSM_STAGE_ERROR_WITHOUT_SEND:
        /*
         * The sending request to AGENT faild. FSM function has to free all
         * resources allocated during stage _IX_CC_ATMSAR_FSM_STAGE_CONTEXT_UPDATE.
         */
        if ((arg_pFsm->pContext != NULL) && (arg_pFsm->pContext->data.pCidStats != NULL))
        {
            ix_ossl_free(arg_pFsm->pContext->data.pCidStats);
            arg_pFsm->pContext->data.pCidStats=NULL;
        }
        break;
        
    case _IX_CC_ATMSAR_FSM_STAGE_PREPARE_REPLY:
        /*
         * Fill buffer of reply message with data collected during processing the
         * request. Set size of reply.
         */
        arg_pFsm->pReplyMsg = arg_pFsm->pContext->data.pCidStats;
        arg_pFsm->sizeBufferToSend = sizeof(_ix_cc_atmsar_async_cid_stats_reply_t);

#ifdef ATMSAR_MAIN_WITH_AGENT_SUPPORT
#ifdef ATMSAR_AGENT_ON_EGRESS
        memcpy(&arg_pFsm->pContext->data.pCidStats->stats.tx,
               &((_ix_cc_atmsar_agent_cid_stats_reply_t*)arg_pFsm->pMsg)->stats.tx,
               sizeof(arg_pFsm->pContext->data.pCidStats->stats.tx));
#else /* ATMSAR_AGENT_ON_EGRESS */
        memcpy(&arg_pFsm->pContext->data.pCidStats->stats.rx,
               &((_ix_cc_atmsar_agent_cid_stats_reply_t*)arg_pFsm->pMsg)->stats.rx,
               sizeof(arg_pFsm->pContext->data.pCidStats->stats.rx));
#endif /* ATMSAR_AGENT_ON_EGRESS */
#endif /* ATMSAR_MAIN_WITH_AGENT_SUPPORT */
        break;
        
    case _IX_CC_ATMSAR_FSM_STAGE_AFTER_REPLY:
        /*
         * The reply has sent to calling module. Free all resources extra
         * allocated for processing the request. Check status of processing
         * the request (arg_pFsm->pContext->err). If it is not IX_SUCCESS,
         * free the object. 'Get statistic' for object should not cause
         * freeing object.
         */
        if ((arg_pFsm->pContext != NULL) && (arg_pFsm->pContext->data.pCidStats != NULL))
        {
            ix_ossl_free(arg_pFsm->pContext->data.pCidStats);
            arg_pFsm->pContext->data.pCidStats=NULL;
        }
        break;
        
    default:
        break;
        
    }
    return IX_SUCCESS;
} /* end _ix_cc_atmsar_main_rec_cid_stats */


/**
 * NAME: _ix_cc_atmsar_main_cid_clear
 *
 * DESCRIPTION:
 *    The function clears VC for VC that VC create operation failed.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
ix_error _ix_cc_atmsar_main_cid_clear(
_ix_cc_atmsar_main_context_t *arg_pMyContext,
ix_uint16 arg_cid)
{
    ix_error err = IX_SUCCESS;

    _IX_CC_ATMSAR_DBG_MAIN(("_ix_cc_atmsar_main_cid_clear(): "
                            "Functionality isn;t implemented yet\n"));
    /*
     *it should:
     * - call local clear operation
     * - call remote clear operation
     * - in callback - free VC handle
     */
    
    return err;
} /* end _ix_cc_atmsar_main_cid_clear */


/******************************************************************************
 *          FUNCTION PROCESSING INCOMING INITIALIZATION MESSAGE
 *****************************************************************************/

/**
 * NAME: _ix_cc_atmsar_main_rec_agent_init
 *
 * DESCRIPTION:
 *    The callback function processes incoming asynchronous messages
 *    'agent initialized'. 
 * 
 * @Param:  IN arg_hDataToken: buffer handle embedding information for the
 *                 message passed in arg_UserData.
 * @Param:  IN arg_pMyContext: pointer to the core component context that will
 *                 be passed to the core component when a message arrives. This
 *                 context was defined by the core component and passed to core
 *                 components infrastructure through CC init function.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ERROR_SEND_FAIL - failure on sending a message
 */
static ix_error
_ix_cc_atmsar_main_rec_agent_init(
ix_buffer_handle arg_hDataToken,
_ix_cc_atmsar_main_context_t *arg_pMyContext)
{
    ix_error err = IX_SUCCESS;
    void *pContext;
    _ix_cc_atmsar_agent_init_notify_t *pMsg;
    
    IX_MSUP_EXTRACT_MSG(arg_hDataToken, (void*)&pMsg, &pContext);
    arg_pMyContext->agentInitialized=pMsg->flags;

    if (arg_pMyContext->engine.pInitVerify == NULL)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL, ("pInitVerify is NULL"));
    }
    else
    {
        err=arg_pMyContext->engine.pInitVerify(arg_pMyContext->engine.pCtx);
    }
    
    /* send return message, length and retrieved context to message support library */
    if (ix_cc_msup_send_reply_msg(pContext, &err, 1, err)!=IX_SUCCESS)
    {
        err = IX_ERROR_WARNING(IX_CC_ATMSAR_ERROR_MSG_LIBRARY, ("Send reply message failed"));
    }
    IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hDataToken),
                 IX_CC_ATMSAR_ERROR_BUFFER_FREE, 0);
    return err;
} /* end _ix_cc_atmsar_main_rec_agent_init */


/******************************************************************************
 *                CORE COMPONENT INFRASTRUCTURE FUNCTIONS
 *****************************************************************************/

/**
 * NAME: ix_cc_atmsar_msg_handler
 *
 * DESCRIPTION:
 *    This primitive is the message handle function for the core component.
 *    The core component receives messages from other core component through
 *    this message handle function and internally it calls appropriate library
 *    function to process the message. This message handle shall be used to
 *    update dynamic properties.
 * 
 * @Param:  IN arg_hDataToken: buffer handle embedding information for the
 *                 message passed in arg_UserData.
 * @Param:  IN arg_UserData: message type
 * @Param:  IN arg_pContext: pointer to the coree component context that will
 *                 be passed to the core component when a message arrives. This
 *                 context was defined by the core component and passed to core
 *                 components infrastructure through CC init function.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_CC_ERROR_NULL - null argument
 *    IX_CC_ERROR_UNDEFINED_MSG - un supported message
 *    IX_CC_ATMSAR_ERROR_MSG_LIBRARY - error from message support library
 *    IX_CC_ATMSAR_ERROR_BUFFER_FREE - error from RM for freeing buffer
 */
ix_error ix_cc_atmsar_msg_handler(
ix_buffer_handle arg_hDataToken,
ix_uint32 arg_UserData,
void* arg_pContext)
{
    ix_error err = IX_SUCCESS;
    void *pContext = NULL;
    void *pMsg = NULL;

    /* Validate arg_pContext */
    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("arg_pContext is NULL"));
    }/* end if (arg_pContext == NULL) */

    if ((arg_UserData < IX_CC_COMMON_MSG_ID_FIRST) ||
        (arg_UserData >= _IX_CC_ATMSAR_MSG_LAST))
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_UNDEFINED_MSG,
                                ("Invalid Message Type"));
    }/* end if */

    if ((((_ix_cc_atmsar_main_context_t*)arg_pContext)->readyFlag == FALSE)
        && (arg_UserData!=_IX_CC_ATMSAR_MSG_AGENT_INIT))
    {
        /* send reply ERROR ATM SAR is not initialized. */
        err=IX_ERROR_WARNING(IX_CC_ERROR_UNINIT,
                             ("ATM SAR is not initialized, agentFLAGS=%X, mainFLAGS=%X, command=%d",
                              ((_ix_cc_atmsar_main_context_t*)arg_pContext)->agentInitialized,
                              ((_ix_cc_atmsar_main_context_t*)arg_pContext)->mainInitialized,
                              arg_UserData));

        IX_MSUP_EXTRACT_MSG(arg_hDataToken, &pMsg, &pContext);
        err=ix_cc_msup_send_reply_msg(pContext, &err, 1, err);
        IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hDataToken),
                     IX_CC_ATMSAR_ERROR_BUFFER_FREE, 0);
        return err;
    }
    /* Check for route messages, next hop message, icmp messages and statistics message*/
    switch (arg_UserData)
    {
    case _IX_CC_ATMSAR_MSG_VC_CREATE:
        _IX_CC_ATMSAR_DBG_MAIN(("MAIN : VC CREATE msg received\n"));
        return _ix_cc_atmsar_main_process_user_req(arg_hDataToken, arg_pContext,
                                                   _ix_cc_atmsar_main_rec_vc_create);
        /* buffer will be freed after message processing */

    case _IX_CC_ATMSAR_MSG_VC_UPDATE:
        _IX_CC_ATMSAR_DBG_MAIN(("MAIN : VC UPDATE msg received\n"));
        return _ix_cc_atmsar_main_process_user_req(arg_hDataToken, arg_pContext,
                                                   _ix_cc_atmsar_main_rec_vc_update);
        
    case _IX_CC_ATMSAR_MSG_VC_REMOVE:
        _IX_CC_ATMSAR_DBG_MAIN(("MAIN : VC REMOVE msg received\n"));
        return _ix_cc_atmsar_main_process_user_req(arg_hDataToken, arg_pContext,
                                                   _ix_cc_atmsar_main_rec_vc_remove);
        
    case _IX_CC_ATMSAR_MSG_GET_VC_STATS:
        _IX_CC_ATMSAR_DBG_MAIN(("MAIN : VC STATS msg received\n"));
        return _ix_cc_atmsar_main_process_user_req(arg_hDataToken, arg_pContext,
                                                   _ix_cc_atmsar_main_rec_vc_stats);

    case _IX_CC_ATMSAR_MSG_PORT_CREATE:
        _IX_CC_ATMSAR_DBG_MAIN(("MAIN : PORT CREATE msg received\n"));
        return _ix_cc_atmsar_main_process_user_req(arg_hDataToken, arg_pContext,
                                                   _ix_cc_atmsar_main_rec_port_create);
        
    case _IX_CC_ATMSAR_MSG_PORT_REMOVE:
        _IX_CC_ATMSAR_DBG_MAIN(("MAIN : PORT REMOVE msg received\n"));
        return _ix_cc_atmsar_main_process_user_req(arg_hDataToken, arg_pContext,
                                                   _ix_cc_atmsar_main_rec_port_remove);
        
    case _IX_CC_ATMSAR_MSG_GET_PORT_STATS:
        _IX_CC_ATMSAR_DBG_MAIN(("MAIN : PORT STATS msg received\n"));
        return _ix_cc_atmsar_main_process_user_req(arg_hDataToken, arg_pContext,
                                                   _ix_cc_atmsar_main_rec_port_stats);

    case _IX_CC_ATMSAR_MSG_CID_CREATE:
        _IX_CC_ATMSAR_DBG_MAIN(("MAIN : CID CREATE msg received\n"));
        return _ix_cc_atmsar_main_process_user_req(arg_hDataToken, arg_pContext,
                                                   _ix_cc_atmsar_main_rec_cid_create);
    case _IX_CC_ATMSAR_MSG_CID_UPDATE:
        _IX_CC_ATMSAR_DBG_MAIN(("MAIN : CID UPDATE msg received\n"));
        return _ix_cc_atmsar_main_process_user_req(arg_hDataToken, arg_pContext,
                                                   _ix_cc_atmsar_main_rec_cid_update);
        
    case _IX_CC_ATMSAR_MSG_CID_REMOVE:
        _IX_CC_ATMSAR_DBG_MAIN(("MAIN : CID REMOVE msg received\n"));
        return _ix_cc_atmsar_main_process_user_req(arg_hDataToken, arg_pContext,
                                                   _ix_cc_atmsar_main_rec_cid_remove);
        
    case _IX_CC_ATMSAR_MSG_GET_CID_STATS:
        _IX_CC_ATMSAR_DBG_MAIN(("MAIN : CID STATS msg received\n"));
        return _ix_cc_atmsar_main_process_user_req(arg_hDataToken, arg_pContext,
                                                   _ix_cc_atmsar_main_rec_cid_stats);

    case _IX_CC_ATMSAR_MSG_AGENT_INIT:
        _IX_CC_ATMSAR_DBG_MAIN(("MAIN : AGENT INIT  msg received\n"));
        return _ix_cc_atmsar_main_rec_agent_init(arg_hDataToken, arg_pContext);
               
    default:
        if (pMsg == NULL)
        {
            IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("Invalid Buffer"));
            break;
        }/* end if (pMsg == NULL) */
        err=IX_ERROR_WARNING(IX_CC_ERROR_UNDEFINED_MSG, ("Invalid Message Type"));
        break;
    }/* end switch (arg_UserData) */

    IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hDataToken),
                 IX_CC_ATMSAR_ERROR_BUFFER_FREE, 0);
    return err;
} /* end ix_cc_atmsar_msg_handler */

