/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      ATM SAR Main Core Component
 *
 * = FILENAME
 *      ix_cc_atmsar_main_lib.c
 *
 * = DESCRIPTION
 *      The file defines libarary  interface functions.
 *
 * = AUTHOR
 *       Jaroslaw Kogut
 *       jaroslaw.kogut@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_atmsar_main_lib.c,v 1.1 2003/06/25 12:06:53 rwielick Exp $";


/**
 * System defined include files
 */
#include "ix_cc_error.h"
#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_cc.h"
#include "ix_cc_macros.h"
#include "ix_cc_properties.h"
#include "ix_config.h"
#include "bindings.h"
#include "cc/ix_cc_msup.h"
#include <assert.h>
/**
 * User defined include files
 */
#include "cc/ix_cc_atmsar.h"
#include "cc/internal/ix_cc_atmsar_main.h"
#include "cc/internal/ix_cc_atmsar_msg.h"

/**
 * Preprocessor symbols and Macros used in this file
 */


/**
 * Type definitions whose scope is limited to this file
 */


/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */


/** 
 * Extern function prototypes.
 */


/** 
 * Static function prototypes.
 */


/** 
 * Function definitions.
 */


/**
 * NAME: ix_cc_atmsar_vc_create
 *
 * DESCRIPTION:
 *    Creating VC in the ATM SAR.
 * 
 * @Param:  IN  arg_pContext : pointer to core component context.
 * @Param:  IN  arg_pVcDescr : parameters describing requested VC.
 * @Param:  OUT arg_pHVc : pointer where is places handle to created VC
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM - invalid input parameters
 */
ix_error
ix_cc_atmsar_vc_create(
void* arg_pContext,
const ix_cc_atmsar_vc_descr_t* arg_pVcDescr,
ix_cc_atmsar_vc_handle_t* arg_pHVc)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_pContext);
    IX_ERROR_DCHECK_ARG_NULL(2, arg_pVcDescr);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pHVc);
    
    _ix_cc_atmsar_main_context_t *pCtx = arg_pContext;
    ix_error err = IX_SUCCESS;
        
    if (pCtx->readyFlag == FALSE)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_UNINIT,
                                ("ATM SAR is not initialized, agentFLAGS=%X, mainFLAGS=%X",
                                 pCtx->agentInitialized, pCtx->mainInitialized));

    }

    if (pCtx->engine.pCbVcAlloc != NULL)
    {
        if ((err=pCtx->engine.pCbVcAlloc(pCtx->engine.pVcHandleMgrCtx,
                                         arg_pHVc, arg_pVcDescr)) != IX_SUCCESS)
        {
            return err;
        }
    }
    else
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, (" VC manager is not registered\n"));
    }
    
    return _ix_cc_atmsar_engine_vc_create(&(pCtx->engine), *arg_pHVc, arg_pVcDescr);
} /* end ix_cc_atmsar_vc_create */



/**
 * NAME: ix_cc_atmsar_vc_update
 *
 * DESCRIPTION:
 *    Update VC with bind parameters.
 * 
 * @Param:  IN  arg_pContext : pointer to core component context.
 * @Param:  IN  arg_hVc : handler to VC to be updated
 * @Param:  IN  arg_pBindData : parameters describing remote part of data stream.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ATMSAR_ERROR_NO_RESOURCES - no resources to add the entry 
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM - invalid pointer in input parameters
 *    IX_CC_ERROR_MSG_LIBRARY - failure on sending a message 
 *    IX_CC_ERROR_UNINIT - This CC has not been initialized
 */
ix_error
ix_cc_atmsar_vc_update(
void* arg_pContext,
ix_cc_atmsar_vc_handle_t arg_hVc,
const ix_cc_atmsar_update_data_t* arg_pUpdateData)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_pContext);

    _ix_cc_atmsar_main_context_t *pCtx = arg_pContext;
        
    if (pCtx->readyFlag == FALSE)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_UNINIT,
                                ("ATM SAR is not initialized, agentFLAGS=%X, mainFLAGS=%X",
                                 pCtx->agentInitialized, pCtx->mainInitialized));

    }
    
    return _ix_cc_atmsar_engine_vc_update(&(pCtx->engine), arg_hVc, arg_pUpdateData);
} /* end ix_cc_atmsar_vc_update */


/**
 * NAME: ix_cc_atmsar_vc_remove
 *
 * DESCRIPTION:
 *    Remove VC from ATM SAR.
 * 
 * @Param:  IN  arg_sarInstance : SAR instance ID.
 * @Param:  IN  arg_hVc : handler to VC to be removed
 * @Param:  IN  arg_Callback : a client provided callback function 
 * @Param:  IN  arg_pUserContext : pointer to client defined context
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM- invalid pointer in input parameters
 *    IX_CC_ERROR_MSG_LIBRARY - failure on sending a message 
 *    IX_CC_ERROR_UNINIT - This CC has not been initialized
 */
ix_error
ix_cc_atmsar_vc_remove(
void* arg_pContext,
const ix_cc_atmsar_instance_t arg_sarInstance,
const ix_cc_atmsar_vc_handle_t arg_hVc)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_pContext);

    _ix_cc_atmsar_main_context_t *pCtx = arg_pContext;
        
    if (pCtx->readyFlag == FALSE)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_UNINIT,
                                ("ATM SAR is not initialized, agentFLAGS=%X, mainFLAGS=%X",
                                 pCtx->agentInitialized, pCtx->mainInitialized));

    }
    return IX_ERROR_GLOBAL(IX_CC_ERROR_NULL, ("function is not impleneted yet"));
} /* end ix_cc_atmsar_vc_remove */



/**
 * NAME: ix_cc_atmsar_get_vc_stats
 *
 * DESCRIPTION:
 *    Get statistic for requested VC. 
 * 
 * @Param:  IN  arg_pContext : pointer to core component context.
 * @Param:  IN  arg_hVc : handler to VC to be removed
 * @Param:  IN/OUT - pointer where statistic result is stored
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM - invalid pointer in input parameters
 *    IX_CC_ERROR_MSG_LIBRARY - failure on sending a message 
 *    IX_CC_ERROR_UNINIT - This CC has not been initialized
 */
ix_error
ix_cc_atmsar_get_vc_stats(
void* arg_pContext,
ix_cc_atmsar_vc_handle_t arg_hVc,
ix_cc_atmsar_vc_stats_t *arg_pLookupResult)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_pContext);

    _ix_cc_atmsar_main_context_t *pCtx = arg_pContext;
        
    if (pCtx->readyFlag == FALSE)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_UNINIT,
                                ("ATM SAR is not initialized, agentFLAGS=%X, mainFLAGS=%X",
                                 pCtx->agentInitialized, pCtx->mainInitialized));

    }
    return IX_ERROR_GLOBAL(IX_CC_ERROR_NULL, ("function is not impleneted yet"));
} /* end ix_cc_atmsar_get_vc_stats */



/**
 * NAME: ix_cc_atmsar_port_create
 *
 * DESCRIPTION:
 *    Create port in the ATM SAR.
 * 
 * @Param:  IN  arg_pContext : pointer to core component context.
 * @Param:  IN  arg_pPortDescr : parameters describing requested port.
 * @Param:  OUT arg_pHPort : pointer where is places handle to created port
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ATMSAR_ERROR_NO_RESOURCES - no resources to add the entry 
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM - invalid pointer in input parameters
 *    IX_CC_ERROR_MSG_LIBRARY - failure on sending a message 
 *    IX_CC_ERROR_UNINIT - This CC has not been initialized
 */
ix_error
ix_cc_atmsar_port_create(
void* arg_pContext,
const ix_cc_atmsar_port_descr_t* arg_pPortDescr,
ix_cc_atmsar_port_handle_t *arg_pHPort)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_pContext);

    _ix_cc_atmsar_main_context_t *pCtx = arg_pContext;
        
    if (pCtx->readyFlag == FALSE)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_UNINIT,
                                ("ATM SAR is not initialized, agentFLAGS=%X, mainFLAGS=%X",
                                 pCtx->agentInitialized, pCtx->mainInitialized));

    }
    return IX_ERROR_GLOBAL(IX_CC_ERROR_NULL, ("function is not impleneted yet"));
} /* end ix_cc_atmsar_port_create */



/**
 * NAME: ix_cc_atmsar_port_remove
 *
 * DESCRIPTION:
 *    Remove port from the ATM SAR. 
 * 
 * @Param:  IN  arg_pContext : pointer to core component context.
 * @Param:  IN  arg_port : ATM port to be removed
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM- invalid pointer in input parameters
 *    IX_CC_ERROR_MSG_LIBRARY - failure on sending a message 
 *    IX_CC_ERROR_UNINIT - This CC has not been initialized
 */
ix_error
ix_cc_atmsar_port_remove(
void* arg_pContext,
ix_cc_atmsar_port_handle_t arg_port)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_pContext);

    _ix_cc_atmsar_main_context_t *pCtx = arg_pContext;
        
    if (pCtx->readyFlag == FALSE)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_UNINIT,
                                ("ATM SAR is not initialized, agentFLAGS=%X, mainFLAGS=%X",
                                 pCtx->agentInitialized, pCtx->mainInitialized));

    }
    return IX_ERROR_GLOBAL(IX_CC_ERROR_NULL, ("function is not impleneted yet"));
} /* end ix_cc_atmsar_port_remove */



/**
 * NAME: ix_cc_atmsar_get_port_stats
 *
 * DESCRIPTION:
 *    Get statistic for requested port. 
 * 
 * @Param:  IN  arg_pContext : pointer to core component context.
 * @Param:  IN  arg_port : port ID
 * @Param:  IN/OUT  arg_pLookupResult : pointer to ix_cc_atmsar_port_stats
 *                  structure where the requested data will be stored
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM - invalid pointer in input parameters
 *    IX_CC_ERROR_MSG_LIBRARY - failure on sending a message 
 *    IX_CC_ERROR_UNINIT - This CC has not been initialized
 */
ix_error
ix_cc_atmsar_get_port_stats(
void* arg_pContext,
ix_cc_atmsar_port_handle_t arg_port,
ix_cc_atmsar_port_stats_t *arg_pLookupResult)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_pContext);

    _ix_cc_atmsar_main_context_t *pCtx = arg_pContext;
        
    if (pCtx->readyFlag == FALSE)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_UNINIT,
                                ("ATM SAR is not initialized, agentFLAGS=%X, mainFLAGS=%X",
                                 pCtx->agentInitialized, pCtx->mainInitialized));

    }
    return IX_ERROR_GLOBAL(IX_CC_ERROR_NULL, ("function is not impleneted yet"));
} /* end ix_cc_atmsar_get_port_stats */


/**
 * NAME: ix_cc_atmsar_cid_create
 *
 * DESCRIPTION:
 *    Create CID in the ATM SAR.
 * 
 * @Param:  IN  arg_pContext : pointer to core component context.
 * @Param:  IN  arg_pCidDescr : parameters describing requested CID.
 * @Param:  IN/OUT arg_pHCid : pointer where is places handle to created port
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ATMSAR_ERROR_NO_RESOURCES - no resources to add the entry 
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM - invalid pointer in input parameters
 *    IX_CC_ERROR_MSG_LIBRARY - failure on sending a message 
 *    IX_CC_ERROR_UNINIT - This CC has not been initialized
 */
ix_error
ix_cc_atmsar_cid_create(
void* arg_pContext,
const ix_cc_atmsar_cid_descr_t* arg_pCidDescr,
ix_cc_atmsar_cid_handle_t* arg_pHCid)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_pContext);
    IX_ERROR_DCHECK_ARG_NULL(2, arg_pCidDescr);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pHCid);
    
    _ix_cc_atmsar_main_context_t *pCtx = arg_pContext;
    ix_error err = IX_SUCCESS;
        
    if (pCtx->readyFlag == FALSE)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_UNINIT,
                                ("ATM SAR is not initialized, agentFLAGS=%X, mainFLAGS=%X",
                                 pCtx->agentInitialized, pCtx->mainInitialized));

    }

    if (pCtx->engine.pCbCidAlloc != NULL)
    {
        if ((err=pCtx->engine.pCbCidAlloc(pCtx->engine.pCidHandleMgrCtx,
                                    arg_pHCid, arg_pCidDescr)) != IX_SUCCESS)
        {
            return err;
        }
    }
    else
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, (" CID manager is not registered\n"));
    }
    
    return _ix_cc_atmsar_engine_cid_create(&(pCtx->engine), *arg_pHCid, arg_pCidDescr);
} /* end ix_cc_atmsar_cid_create */



/**
 * NAME: ix_cc_atmsar_cid_update
 *
 * DESCRIPTION:
 *    Update CID with bind parameters.
 * 
 * @Param:  IN  arg_pContext : pointer to core component context.
 * @Param:  IN  arg_hCid : handle for CID to be updated
 * @Param:  IN  arg_pUpdate : parameters describing remote part of data stream.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ATMSAR_ERROR_NO_RESOURCES - no resources to add the entry 
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM - invalid pointer in input parameters
 *    IX_CC_ERROR_MSG_LIBRARY - failure on sending a message 
 *    IX_CC_ERROR_UNINIT - This CC has not been initialized
 */
ix_error
ix_cc_atmsar_cid_update(
void* arg_pContext,
ix_cc_atmsar_cid_handle_t arg_hCid,
const ix_cc_atmsar_update_data_t* arg_pUpdateData)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_pContext);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pUpdateData);
    
    _ix_cc_atmsar_main_context_t *pCtx = arg_pContext;
        
    if (pCtx->readyFlag == FALSE)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_UNINIT,
                                ("ATM SAR is not initialized, agentFLAGS=%X, mainFLAGS=%X",
                                 pCtx->agentInitialized, pCtx->mainInitialized));

    }
    return _ix_cc_atmsar_engine_cid_update(&(pCtx->engine), arg_hCid, arg_pUpdateData);
} /* end ix_cc_atmsar_cid_update */


/**
 * NAME: ix_cc_atmsar_cid_remove
 *
 * DESCRIPTION:
 *    Remove Cid from the ATM SAR core component. 
 * 
 * @Param:  IN  arg_pContext : pointer to core component context.
 * @Param:  IN  arg_cid : CID to be removed
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM- invalid pointer in input parameters
 *    IX_CC_ERROR_MSG_LIBRARY - failure on sending a message 
 *    IX_CC_ERROR_UNINIT - This CC has not been initialized
 */
ix_error
ix_cc_atmsar_cid_remove(
void* arg_pContext,
ix_cc_atmsar_cid_handle_t arg_cid)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_pContext);

    _ix_cc_atmsar_main_context_t *pCtx = arg_pContext;
        
    if (pCtx->readyFlag == FALSE)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_UNINIT,
                                ("ATM SAR is not initialized, agentFLAGS=%X, mainFLAGS=%X",
                                 pCtx->agentInitialized, pCtx->mainInitialized));

    }
    return IX_ERROR_GLOBAL(IX_CC_ERROR_NULL, ("function is not impleneted yet"));
} /* end ix_cc_atmsar_cid_remove */


/**
 * NAME: ix_cc_atmsar_get_cid_stats
 *
 * DESCRIPTION:
 *    Get statistic for requested CID. 
 * 
 * @Param:  IN  arg_pContext : pointer to core component context.
 * @Param:  IN  arg_cid : CID handler
 * @Param:  IN/OUT  arg_pLookupResult : pointer to ix_cc_atmsar_cid_stats_t
 *                  structure where the requested data will be stored
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - message was successfully built and sent 
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM - invalid pointer in input parameters
 *    IX_CC_ERROR_MSG_LIBRARY - failure on sending a message 
 *    IX_CC_ERROR_UNINIT - This CC has not been initialized
 */
ix_error
ix_cc_atmsar_get_cid_stats(
void* arg_pContext,
ix_cc_atmsar_cid_handle_t arg_cid,
ix_cc_atmsar_cid_stats_t *arg_pLookupResult)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_pContext);

    _ix_cc_atmsar_main_context_t *pCtx = arg_pContext;
        
    if (pCtx->readyFlag == FALSE)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_UNINIT,
                                ("ATM SAR is not initialized, agentFLAGS=%X, mainFLAGS=%X",
                                 pCtx->agentInitialized, pCtx->mainInitialized));

    }
    return IX_ERROR_GLOBAL(IX_CC_ERROR_NULL, ("function is not impleneted yet"));
} /* end ix_cc_atmsar_get_cid_stats */


