////////////////////////////////////////////////////////////////////////////////
//                                                                     
//                  I N T E L   P R O P R I E T A R Y                   
//                                                                      
//     COPYRIGHT (c)  2002-2003 BY  INTEL  CORPORATION.  ALL RIGHTS          
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//     THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                      
//                        INTEL  CORPORATION                            
//                                                                     
//                     2200 MISSION COLLEGE BLVD                        
//                                                                      
//               SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                      
////////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: v6v4_natpt.uc
//
//      Purpose: This file contains macros for supporting the IPv6 translation
//               mechanism, Network Address Translation/Protocol
//               translation (NAT-PT)
//
//      History:
//
//
//      Date            Comment
//      -----------------------------------------------------------------------
//
//      12/10/2002      Created
//
//
////////////////////////////////////////////////////////////////////////////////
#ifndef __V6V4_NATPT_UC__
#define __V6V4_NATPT_UC__

//------------------------------------------------------------------------------
// Exported APIs:
// v6v4_natpt_init()
// v6v4_natpt_translate(in_ipv4_hdr, IN_IPV4_OFFSET,
//                      in_ipv6_hdr, IN_IPV6_OFFSET,
//                      out_ipv4_hdr, OUT_IPV4_OFFSET,
//                      out_ipv6_hdr, OUT_IPV6_OFFSET)
// v6v4_natpt_fini()

// Assumptions:
//

//------------------------------------------------------------------------------
// Includes.
#include "v6v4_natpt.h"
#include "v6v4_natpt_util.uc"

#include <ipv4.h>
#include <ipv6.h>
#include <ether.h>

#include <xbuf.uc>
#include <dl_meta.uc>
#include <dl_source.uc>

//------------------------------------------------------------------------------
// Definitions used by the microblock
//

#define IPV6_MAIN_HEADER_LENGTH 40
#define IPV4_MAIN_HEADER_LENGTH 20
#define IPV6_FRAGMENT_HDR_SIZE 8

#define_eval V6V4_MAIN_HDR_LEN_DIFF (IPV6_MAIN_HEADER_LENGTH-IPV4_MAIN_HEADER_LENGTH)
#define_eval V6MAIN_FRAG_HDR_LEN_SUM (IPV6_MAIN_HEADER_LENGTH+IPV6_FRAGMENT_HDR_SIZE)


//UDP header is 8 bytes, so one quad word read is sufficient
#define TRANSPORT_HDR_BYTES_TO_READ 8

#define IPV4_DONT_FRAG_BIT 14
#define IPV4_MORE_FRAG_BIT 13

//Offset (bytes) of Checksum in TCP/UDP header
#define TCP_CHECKSUM_OFFSET 16
#define UDP_CHECKSUM_OFFSET 6
#define UDP_LENGTH_OFFSET 4

//Used in the generate checksum function to read the transport payload & in the
//function that parses v6 extension headers
#define NAT_DRAM_READ_BYTES 64

//------------------------------------------------------------------------------
// Global registers.
//
// None

//------------------------------------------------------------------------------
// Global signals
//
// None

//------------------------------------------------------------------------------
// Module specific registers 
//
.reg checksum
.reg in_operand
.reg ret_reg

//Helper routine for Hardware debugging
// Print a "value" at an "offset" from the start of a predetermined sram address
#macro  _natpt_print_something(value, offset)
#ifdef AI_DEBUG
.begin

   .reg sram_address
   .reg $temp
   .sig nh_sram_signal
   .reg var

//   immed32[sram_address, NATPT_NAT_V4V6_HASH_TABLE_BASE]
   immed32[sram_address, 0x40200000]
//   alu[sram_address, sram_address, +, 16] ;move past the first 2 entries in this table
   alu_shf[var, --, B, offset, <<2]
   alu[sram_address, sram_address, +, var]
   alu[$temp, --, B, value]

    // Read IPV6_NH_SIZE_LW long words from this location 
    // into the SRAM xfer registers.

    sram[write,            \
         $temp,            \
         0,                \
         sram_address,     \
         1],               \
         ctx_swap[nh_sram_signal]

.end
#endif //AI_DEBUG

#endm

////////////////////////////////////////////////////////////////////////////////
// v6v4_natpt_init
//
//    Description:
//        Sets up global data for the v6v4_natpt microblock.
//
//    Parameters:
//        Outputs:
//        Inputs:
//
////////////////////////////////////////////////////////////////////////////////
#macro v6v4_natpt_init()

#ifdef NATPT_DEBUG_COUNTERS
    //Global register
    .reg @natpt_sram_counter_base

    immed32(@natpt_sram_counter_base, NATPT_SRAM_COUNTER_BASE)
#endif

    _natpt_hash_init()

    //Allocate the local memory buffer that will be used by the NATPT 
    //microblock for caching v6 extension headers & transport headers if needed
    xbuf_alloc(LM_NATPT_WORK_BUF, LM_BUF_LWSIZE_NATPT_WORK, read_write)
    xbuf_bind_address(LM_NATPT_WORK_BUF, LM_NATPT_WORK_BUF_BASE,
                      LM_NATPT_WORK_BUF_SIZE, LM_BUF_OFFSET_NATPT_WORK_BUF)
#endm //v6v4_natpt_init

////////////////////////////////////////////////////////////////////////////////
// v6v4_natpt_fini
//
//    Description:
//        Cleans up state created for the v6v4_natpt microblock.
////////////////////////////////////////////////////////////////////////////////
#macro v6v4_natpt_fini()
    xbuf_free(LM_NATPT_WORK_BUF)
#endm //v6v4_natpt_fini

////////////////////////////////////////////////////////////////////////////////
// v6v4_natpt_translate
//
//   Description: Translates IPv4 packets to IPv6 and vice versa. It identifies
//      the packets to translate. If a packet should be translated but there
//        is no binding for the address, this macro throws an exception to the
//        core so that the core may create a binding.
//
//   Parameters:
//        Outputs:
//             - out_ipv4_hdr   : Buffer in which the translated IPv4 header is
//                              stored.
//           - OUT_IPV4_OFFSET: The translated header should be stored in the
//                              buffer starting at this offset
//             - out_ipv6_hdr   : Buffer in which the translated IPv6 header is
//                              stored.
//           - OUT_IPV6_OFFSET: The translated header should be stored in the
//                              buffer starting at this offset
//
//        Inputs:
//             - in_ipv4_hdr   : Buffer containing the incoming IPv4 header
//             - IN_IPV4_OFFSET: Start offset of header in the buffer passed in
//             - in_ipv6_hdr   : Buffer containing the incoming IPv6 header
//             - IN_IPV6_OFFSET: Start offset of header in the buffer passed in
//
//   Note: For a particular packet, only one of the in headers and one of the
//         out headers is used. The header selected depends on the header type
//
//   Side effects:
//        Updates the meta data (packet size, start offset, header type) of
//        translated packets
//
//   Assumptions:
//         Assumes that the main IP header (not including the option headers for
//      IPv4 & the extension headers for IPv6) is present in the buffer passed
//      in. Any modifications to the IP headers are made in the output buffers
//      (also passed in).
//      If needed, this macro reads in transport headers. Any modifications to
//      the transport headers are made in transfer registers visible to this
//      microblock only and flushed back to DRAM.
//      This is based on the assumption that it is unlikely that other
//      microblocks will need to read/modify transport headers. Network headers
//      on the other hand might be read/modified by other microblocks. Hence
//      the changes are made to the buffers passed in. In this case it is the
//      responsibility of the application chaining the microblocks together to
//      figure out when to flush the header back to DRAM.
//
////////////////////////////////////////////////////////////////////////////////
#macro v6v4_natpt_translate(in_ipv4_hdr, IN_IPV4_OFFSET,\
                            in_ipv6_hdr, IN_IPV6_OFFSET,\
                            out_ipv4_hdr, OUT_IPV4_OFFSET,\
                            out_ipv6_hdr, OUT_IPV6_OFFSET)
.begin
    br!=byte[dl_next_block,0,BID_NATPT,v6v4_natpt_translate_end#]

    .reg context_id
    .reg header_type

    dl_get_thread_id(context_id)

#ifdef NATPT_DEBUG_COUNTERS
    .reg    counter_base    // base where the counters are maintained

    alu[counter_base,  --, B, @natpt_sram_counter_base]
    _natpt_increment_counter(counter_base, NATPT_RECV)
#endif

    dl_meta_get_header_type(header_type)

    .if (header_type == ETHER_IPV4_TYPE)
        .begin
            .reg dest_addr
            .reg hit
            xbuf_alloc($nat_entry, NAT_ENTRY_SIZE_LW, read)

        #if (strstr(in_ipv4_hdr, 'lmem_buf'))
            #define BUF_IN_LOCAL_MEM
            //Set the local memory index pointer to point to the IPv4 header
            xbuf_activate(in_ipv4_hdr, 0, context_id, 1)
        #endif

            xbuf_extract(dest_addr, in_ipv4_hdr, IN_IPV4_OFFSET,
                         IP_DESTINATION_ADDRESS)

            //Check if binding exists for the destination address
            //Note: We update entryUsed flag in NAT table to avoid the race
            //    condition where this function returns a hit, but the core
            //    deletes the NAT entry due to a timeout. Updating the flag
            //    ensures that the entry is valid for TIMEOUT seconds after this
            //    function returns a match.

            _natpt_v4_v6_hash(hit,$nat_entry,dest_addr,
                              NATPT_NAT_V4V6_HASH_TABLE_BASE,
                              NATPT_NAT_COLL_HASH_TABLE_BASE,
                              NATPT_NAT_TABLE_BASE)

            .if (hit == 1)
              .begin
                .reg result
                //Binding exists for this address. Translate the packet.

                #ifdef BUF_IN_LOCAL_MEM
                //Deactivate buffer. This is needed as otherwise lmem_bufx
                //will be passed in to the macro, instead of in_ipv4_hdr.
                //And as the xbuf_activate #defines lmem_bufx to *l$index0/1
                //this is what will be used if we make any xbuf_calls in that 
                //macro. These will fail as there is no buffer with that name
                    xbuf_deactivate(in_ipv4_hdr)
                #endif
                _translate_to_v6( result, out_ipv6_hdr, OUT_IPV6_OFFSET,
                                  in_ipv4_hdr, IN_IPV4_OFFSET,
                                  L2_HEADER_LENGTH, dest_addr,
                                  context_id, $nat_entry)

                .if (result != 0)
                    dl_set_exception[BID_NATPT, result]
                    alu[dl_next_block, --, B, IX_EXCEPTION]

                    #ifdef NATPT_DEBUG_COUNTERS
                        _natpt_increment_counter(counter_base,
                                                 NATPT_SENT_TO_CORE)
                    #endif
                .else
                    dl_meta_set_header_type(ETHER_IPV6_TYPE)
                    alu[dl_next_block, --, B, V6V4_NATPT_TRANSLATED_V6]

#ifdef NATPT_V6_OPTIMIZE_LOOKUP
                    //The NAT binding information includes the v6 NHID
                    //for the V6 address for which the binding is created
                    //We use this information to set the next hop ID in the
                    //meta data. By having the NAT core, cache the v6 NHID
                    //we avoid having to do a lookup for the v6 address
                    dl_meta_set_nexthop_id($nat_entry[5])
                    dl_meta_set_skip_lookup_flag()
#endif //NATPT_V6_OPTIMIZE_LOOKUP

                    #ifdef NATPT_DEBUG_COUNTERS
                        _natpt_increment_counter(counter_base,
                                                 NATPT_TRANSLATED_TO_V6)
                    #endif
                .endif //result != 0
              .end
            .else //hit != 1
/*
              .begin
                //No entry found in the NAT state. Lets look at NAPT state
                //First check if NAPT mode is enabled
                .reg $napt_mode
                .reg napt_mode_address
                .sig sig_napt_state

                //Check if we can avoid the sram read each time. Can the ME be
                //signalled whenever this mem locn is modified?
                //Since this will prolly not be modified often, cache it in LM
                //and periodically refresh?
                //OR Core can directly write to a particular register

                immed32(napt_mode_address, NAPT_ON)
                sram[read, $napt_mode, napt_mode_address, 0, 1], ctx_swap[sig_napt_state]

                .if ($napt_mode == 1)

                    .begin
                        .reg ipv4_hdr_len
                        .reg transport_protocol
                        .reg transport_id
                        .reg $napt_addr
                        .reg predefined_napt_addr

                        #define_eval NUM_TRANS_HDR_REG \
                                     (TRANSPORT_HDR_BYTES_TO_READ>>2)
                        xbuf_alloc(transport_hdr,NUM_TRANS_HDR_REG,read_write)
                        #undef NUM_TRANS_HDR_REG

                        //First check if the destination address matches the
                        //IP address set aside for NAPT
                        //Note: If this value does not change too often,
                        //      can store in LM or reg.
                        immed32(predefined_napt_addr, NAPT_ADDR)
                        sram[read, $napt_addr, predefined_napt_addr, 0, 1],
                             ctx_swap[sig_napt_state]

                        .if (dest_addr == $napt_addr)
                            //Might be worthwhile to read in the relevant parts
                            //of the packet from DRAM right now as we might
                            //need them if the packet has to be translated.
                            //Read the packet into some xbuf. define macros to
                            //access specific fields from in there.

                            // Now read in relevant parts of the transport
                            // header

                            //1. Get the offset in the packet where the
                            //   transport header begins.
                            //2. Get the type of the transport protocol
                            //3. Based on the protocol type, read in relevant
                            //   parts of the header from DRAM

                            //Get the IP header length
                            xbuf_extract(ipv4_hdr_len,
                                         in_ipv4_hdr,
                                         IN_IPV4_OFFSET,
                                         IP_VERSION_LENGTH)
                            alu[ipv4_hdr_len, ipv4_hdr_len, AND, 0xf]


                            //Now get the protocol type.
                            xbuf_extract(transport_protocol,
                                         in_ipv4_hdr,
                                         IN_IPV4_OFFSET,
                                          IP_PROTOCOL)

                            //Read in relevant part from DRAM
                            .if ((PROTO_UDP == transport_protocol) || \
                                 (PROTO_TCP == transport_protocol))
                              .begin
                                .sig sig_trans_hdr
                                .reg transport_hdr_offset

                                // Do not use the buffer offset from meta data
                                // as we don't know where in the ublock chain
                                // natpt is placed. Simply start from the base
                                // of the packet.
                                alu[transport_hdr_offset, ipv4_hdr_len, +,
                                    L2_HEADER_LENGTH]

                                //To get rid of warning. This field is filled
                                //by dl_get_bytes
                                .set transport_hdr0
                                dl_get_bytes(transport_hdr,
                                             transport_hdr_offset,
                                             TRANSPORT_HDR_BYTES_TO_READ,
                                             sig_trans_hdr,
                                             sig_trans_hdr)
//                                dram[read, 
//                                     $$transport_hdr,
//                                     sdram_offset,
//                                     transport_hdr_offset,
//                                     UDP_HDR_DRAM_RD_SIZE], sig_done[??]
                                ld_field_w_clr[    transport_id, 0011,
                                                transport_hdr0]
                              .end
                            .else
                                ;simply send the packet on. Do not try to create a binding for it
                                br[LBL_NO_BINDINGS#]
                            .endif //(PROTO_UDP == transport_protocol) ||...

                            _napt_hash(index,
                                       transport_protocol,
                                       transport_id,
                                       NAPTPT_HASH_TABLE_BASE)

                            //If we have a binding, translate, else simply
                            //pass the packet on..
                            .if (index != -1)
                                _napt_translate_to_v6(index,
                                                      transport_protocol,
                                                      transport_hdr)

                            .else
LBL_NO_BINDINGS#:
                            .endif //napt hash index == -1
                        .endif //dest_addr == $napt_addr

                        xbuf_free(transport_hdr)
                    .end
                .endif //$napt_mode == 1
                alu[dl_next_block, --, B, BID_IPV4]
                #ifdef NATPT_DEBUG_COUNTERS
                    _natpt_increment_counter(counter_base, NATPT_PASS_THROUGH)
                #endif

              .end
*/

                alu[dl_next_block, --, B, V6V4_NATPT_PASSTHROUGH_V4]
                #ifdef NATPT_DEBUG_COUNTERS
                  _natpt_increment_counter(counter_base, NATPT_PASS_THROUGH)
                #endif
            .endif //nat hash hit == 1
            xbuf_free($nat_entry)
        .end

        #ifdef BUF_IN_LOCAL_MEM
            xbuf_deactivate(in_ipv4_hdr)
            #undef BUF_IN_LOCAL_MEM
        #endif

    .elif (header_type == ETHER_IPV6_TYPE)
        .begin
            .reg ipv6_hdr_len
            .reg transport_protocol
            .reg fragment_offset_and_flag, fragment_identification
            .reg da_127_96, da_95_64, da_63_32, da_31_0
            .reg sa_127_96, sa_95_64, sa_63_32, sa_31_0
            .reg hit
            .reg map_prefix
            .reg security_hdr_found,fragment_hdr_found
        #if (strstr(in_ipv6_hdr, 'lmem_buf'))
            #define BUF_IN_LOCAL_MEM
            //Set the local memory index pointer to point to the IPv6 header
            xbuf_activate(in_ipv6_hdr, 0, context_id, 1)
        #endif
            //LM work buffer. workaround for xbuf GPR buffer problem
            xbuf_activate(LM_NATPT_WORK_BUF, 1, context_id, 0)

            //Translate only if the destination address matches the predefined
            //NATPT prefix
            xbuf_extract(da_127_96, in_ipv6_hdr, IN_IPV6_OFFSET,
                         IPV6_DST_ADDR_127_96)
            immed32(map_prefix, NATPT_PREFIX_0)
            alu[--, map_prefix, -, da_127_96]
            bne[v6v4_natpt_pass_through#]


            xbuf_extract(da_95_64, in_ipv6_hdr, IN_IPV6_OFFSET,
                         IPV6_DST_ADDR_95_64)
            immed32(map_prefix, NATPT_PREFIX_1)
            alu[--, map_prefix, -, da_95_64]
            bne[v6v4_natpt_pass_through#]

            xbuf_extract(da_63_32, in_ipv6_hdr, IN_IPV6_OFFSET,
                         IPV6_DST_ADDR_63_32)
            immed32(map_prefix, NATPT_PREFIX_2)

            alu[--, map_prefix, -, da_63_32]
            bne[v6v4_natpt_pass_through#]


            //It is a v4-mapped v6 address. Get the v4 address.
            xbuf_extract(da_31_0, in_ipv6_hdr, IN_IPV6_OFFSET,
                         IPV6_DST_ADDR_31_0)

            //We now get the v6 header length & the transport protocol. This
            //information is used further ahead.
            //1. Get the offset in the packet where the
            //   transport header begins.
            //2. Get the type of the transport protocol

            alu[fragment_offset_and_flag, --, B, 0]
            alu[fragment_identification, --, B, 0]

        #ifdef BUF_IN_LOCAL_MEM
            //Deactivate buffer. This is needed as otherwise lmem_bufx
            //will be passed in to the macro, instead of in_ipv6_hdr.
            //And as the xbuf_activate #defines lmem_bufx to *l$index0/1
            //this is what will be used if we make any xbuf_calls in that 
            //macro. These will fail as there is no buffer with that name
            xbuf_deactivate(in_ipv6_hdr)
        #endif

            //Get the IP header length
            _parse_v6_extension_headers(ipv6_hdr_len,
                                        transport_protocol,
                                        fragment_offset_and_flag,
                                        fragment_identification,
                                        fragment_hdr_found,
                                        security_hdr_found,
                                        in_ipv6_hdr,
                                        IN_IPV6_OFFSET,
                                        L2_HEADER_LENGTH,
										context_id)

        #ifdef BUF_IN_LOCAL_MEM
            xbuf_activate(in_ipv6_hdr, 0, context_id, 1)
        #endif

            //Next, check if binding for the source address exists.
            xbuf_extract(sa_127_96, in_ipv6_hdr,
                         IN_IPV6_OFFSET, IPV6_SRC_ADDR_127_96)
            xbuf_extract(sa_95_64, in_ipv6_hdr,
                         IN_IPV6_OFFSET,  IPV6_SRC_ADDR_95_64)
            xbuf_extract(sa_63_32, in_ipv6_hdr,
                         IN_IPV6_OFFSET, IPV6_SRC_ADDR_63_32)
            xbuf_extract(sa_31_0, in_ipv6_hdr,
                         IN_IPV6_OFFSET, IPV6_SRC_ADDR_31_0)

            //We don't need to read in the last LW
            #define_eval NUM_NAT_XFER_REG (NAT_ENTRY_READ_SIZE_LW -1)
            xbuf_alloc($nat_entry, NUM_NAT_XFER_REG, read)

            _natpt_v6_v4_hash(hit, $nat_entry, NUM_NAT_XFER_REG, sa_127_96,
                              sa_95_64, sa_63_32, sa_31_0,
                              NATPT_NAT_V6V4_HASH_TABLE_BASE,
                              NATPT_NAT_COLL_HASH_TABLE_BASE,
                              NATPT_NAT_TABLE_BASE)

            #undef NUM_NAT_XFER_REG

            .if (hit == 1)
                //Binding exists for this address. Translate the packet.

                //End-to-end network security not possible with NATPT
                .if (security_hdr_found == 1)
                    dl_set_exception[BID_NATPT,
                                         NATPT_EXCEPTION_IPV6_SECURITY_HEADER]
                    #ifdef NATPT_DEBUG_COUNTERS
                        _natpt_increment_counter(counter_base,
                                                 NATPT_SENT_TO_CORE)
                    #endif

                    br[v6v4_natpt_deactivate_end#], defer[1]
                    alu[dl_next_block, --, B, IX_EXCEPTION]
                .endif

                //The first thing we do is to check the transport protocol
                //This is because we do not translate ICMP packets.
                //So, if the proto is ICMP we throw it to the core right
                //away and avoid translating the packet alltogether!
                .if (PROTO_ICMP_V6 == transport_protocol)
                    dl_set_exception[BID_NATPT, NATPT_EXCEPTION_ICMP_IPV6]
                    #ifdef NATPT_DEBUG_COUNTERS
                        _natpt_increment_counter(counter_base,
                                                 NATPT_SENT_TO_CORE)
                    #endif

                    br[v6v4_natpt_deactivate_end#], defer[1]
                    alu[dl_next_block, --, B, IX_EXCEPTION]
                .endif

        #ifdef BUF_IN_LOCAL_MEM
                xbuf_deactivate(in_ipv6_hdr)
        #endif
                _translate_to_v4(out_ipv4_hdr, OUT_IPV4_OFFSET, 
                                 in_ipv6_hdr, IN_IPV6_OFFSET,
                                 da_31_0,
                                 sa_127_96, sa_95_64, sa_63_32, sa_31_0,
                                 ipv6_hdr_len, transport_protocol,
                                 fragment_hdr_found,
                                 fragment_offset_and_flag,
                                 fragment_identification,
                                 L2_HEADER_LENGTH,
                                 $nat_entry)
            .else
              .begin
                //No entry found in the NAT state. Lets look at NAPT state
                //First check if NAPT mode is enabled
/*
                .reg $napt_mode
                .reg napt_mode_address
                .sig sig_napt_state

                immed32(napt_mode_address, NAPT_ON)
                sram[read, $napt_mode, napt_mode_address, 0, 1],
                     ctx_swap[sig_napt_state]

                .if ($napt_mode == 1)

                        // Now read in relevant parts of the transport header
                        // Based on the transport protocol type, read in
                        // relevant parts of the header from DRAM

                        //Read in relevant part from DRAM
                        .if ((PROTO_UDP == transport_protocol) || \
                             (PROTO_TCP == transport_protocol))
                          .begin
                            .sig sig_trans_hdr
                            .reg transport_id
                            .reg transport_hdr_offset

                            #define_eval NUM_TRANS_HDR_REG \
                                         TRANSPORT_HDR_BYTES_TO_READ>>2
                            xbuf_alloc(    transport_hdr,
                                        NUM_TRANS_HDR_REG,
                                        read_write)
                            #undef NUM_TRANS_HDR_REG

                            // Do not use the buffer offset from meta data
                            // as we don't know where in the ublock chain
                            // natpt is placed. Simply start from the base
                            // of the packet.
                            alu[transport_hdr_offset, ipv6_hdr_len, +,
                                 L2_HEADER_LENGTH]

                            //To get rid of warning. This field is filled by
                            //dl_get_bytes
                            .set transport_hdr0
                            dl_get_bytes(transport_hdr,
                                         transport_hdr_offset,
                                         TRANSPORT_HDR_BYTES_TO_READ,
                                         sig_trans_hdr,
                                         sig_trans_hdr)
//                            dram[read, 
//                                 $$transport_hdr,
//                                 sdram_offset,
//                                 transport_hdr_offset,
//                                 TRANSPORT_HDR_BYTES_TO_READ], sig_done[??]
                            ld_field_w_clr[    transport_id, 0011,
                                            transport_hdr[0]]

                            _napt_hash(index,
                                       transport_protocol,
                                       transport_id,
                                       NAPTPT_HASH_TABLE_BASE)

                            .if (index != -1)
                                //Binding exists for this address.
                                //Translate the packet.
                                _napt_translate_to_v4(     index, in_ipv6_hdr,
                                                        da_31_0, out_ipv4_hdr)
                            .else
                                //We need to add a binding. Send an exception
                                //to the core
                                dl_set_exception[BID_NATPT,
                                                      NATPT_EXCEPTION_NOENTRY_IPV6]
                                alu[dl_next_block, --, B, IX_EXCEPTION]

                            Cannot just fall through here as next_block set to v4 below
                            Either merge with the exception path below OR
                            br to deactive_end#

                            .endif

                            xbuf_free(transport_hdr)
                          .end
                        .else //transport_protocol != TCP | UDP
                            //NAPT mode is ON but we cannot do port translation
                            //Create a NAT binding.
                            //Do we need to send a different exception to the
                            //core?
                            dl_set_exception[BID_NATPT,
                                                  NATPT_EXCEPTION_NOENTRY_IPV6]
                            alu[dl_next_block, --, B, IX_EXCEPTION]

                            Cannot just fall through here as next_block set to v4 below
                            Either merge with the exception path below OR
                            br to deactive_end#
                        .endif
                .else //($napt_mode == 1)
*/
                    //Create a NAT binding
                    //If same exception is being sent to the core, can collapse
                    //the above 3 exception paths.

                    dl_set_exception[BID_NATPT, NATPT_EXCEPTION_NOENTRY_IPV6]
                    #ifdef NATPT_DEBUG_COUNTERS
                        _natpt_increment_counter(counter_base,
                                                 NATPT_SENT_TO_CORE)
                    #endif

                    br[v6v4_natpt_deactivate_end#], defer[1]
                    alu[dl_next_block, --, B, IX_EXCEPTION]
//                .endif //($napt_mode == 1)
              .end
            .endif //NAT hash (hit == 1)

            xbuf_free($nat_entry)
        .end


        #ifdef NATPT_DEBUG_COUNTERS
            _natpt_increment_counter(counter_base, NATPT_TRANSLATED_TO_V4)
        #endif

        //set next block id indicating that the v4 fwder shud be in this path
        dl_meta_set_header_type(ETHER_IPV4_TYPE)
        br[v6v4_natpt_deactivate_end#], defer[1]
        alu[dl_next_block, --, B, V6V4_NATPT_TRANSLATED_V4]

    v6v4_natpt_pass_through#:
        alu[dl_next_block, --, B, V6V4_NATPT_PASSTHROUGH_V6]
        #ifdef NATPT_DEBUG_COUNTERS
            _natpt_increment_counter(counter_base,
                                     NATPT_PASS_THROUGH)
        #endif

    v6v4_natpt_deactivate_end#:
        #ifdef BUF_IN_LOCAL_MEM
            xbuf_deactivate(in_ipv6_hdr)
            #undef BUF_IN_LOCAL_MEM
        #endif
        xbuf_deactivate(LM_NATPT_WORK_BUF)
    .endif //header_type == ...

v6v4_natpt_translate_end#:
.end
#endm //v6v4_natpt_translate

////////////////////////////////////////////////////////////////////////////////
// _parse_v6_extension_headers
//
//    Description:
//        Walks the chain of extension headers in an IPv6 packet collecting
//        the following information, 
//          length of the entire ipv6 header
//          transport protocol and
//          fragment information
//
//    Parameters:
//        Outputs:
//           - out_ipv6_hdr_len : Length of the entire IPv6 header
//           - out_transport_proto: Protocol number of the transport PDU
//           - out_fragment_offset_and_flag: If the packet is a fragment,
//                       the first 13 bits of this parameter are set to the
//                       offset of the fragment in the packet and the LSbit
//                       indicates if More Fragments are present in the packet.
//           - out_fragment_identification:
//                       Value used to identify which packet a fragment is a
//                       part of.
//             - out_fragment_hdr_present:
//                         Flag indicating if a fragment header is present in
//                         the packet
//           - out_security_hdr_present:
//                       Set to 1 if the ESP or Auth extension headers are
//                       found. Else set to 0. If set, the other output values
//                       are undefined and should not be used.
//        Inputs:
//           - in_ipv6_hdr           : Buffer holding the IPv6 header to parse
//           - in_start_offset_in_buf: Start offset of header in the buffer
//                       passed in
//           - in_L2_hdr_length      : Length of the Layer 2 header
//           - context_id            : Thread number
////////////////////////////////////////////////////////////////////////////////
#macro _parse_v6_extension_headers( out_ipv6_hdr_len,\
                                    out_transport_proto,\
                                    out_fragment_offset_and_flag,\
                                    out_fragment_identification,\
                                    out_fragment_hdr_present,\
                                    out_security_hdr_present,\
                                    in_ipv6_hdr,\
                                    in_start_offset_in_buf,\
                                    in_L2_hdr_length,\
									context_id)
.begin
    .sig sig_ipv6_hdr

    .reg next_header
    .reg cur_hdr_len
    .reg len_hdr_in_buf
    .reg length_needed

    //Flag indicating if the entire ipv6 header is in the initial buffer
    //passed in
    .reg overflow

    //Buffer in which the remaining part of the ipv6 header (which was not
    //present in the initial buffer passed in) is read in
    //Due to register overflow, the following is ifdef'd out. We use a buffer
    //defined in local memory instead

    #ifdef NO_GPR_OVERFLOW
        //Div by 4 to get number of registers
        #define_eval NUM_OVERFLOW_REG NAT_DRAM_READ_BYTES>>2
        xbuf_alloc(overflow_buffer, NUM_OVERFLOW_REG, read_write)
        #undef NUM_OVERFLOW_REG
    #endif //NO_GPR_OVERFLOW

    //offset in the buffer at which the next header begins
    .reg buf_offset

    #if (strstr(in_ipv6_hdr, 'lmem_buf'))
        xbuf_activate(in_ipv6_hdr, 0, context_id, 1)
        #define IPV6_HDR_IN_LM
    #endif

    alu[overflow, --, B, 0]
    alu[out_ipv6_hdr_len, --, B, 0]
    alu[out_fragment_hdr_present, --, B, 0]
    alu[out_security_hdr_present, --, B, 0]

    //The macro assumes that the buffer is large enough to hold the initial
    //part of the ipv6 header
    alu[len_hdr_in_buf, --, B, IPV6_MAIN_HEADER_LENGTH]
    xbuf_extract(next_header, in_ipv6_hdr, in_start_offset_in_buf,
                 IPV6_NEXT_HEADER)

    alu[buf_offset, --, B, IPV6_MAIN_HEADER_LENGTH] ;Initialize past size
                                                    ; of main ipv6 header

ext_hdr_loop#:
    alu[length_needed, --, B, buf_offset]

    .if (next_header == IPV6_EXTN_HOP_BY_HOP)
        br[ext_hdr_readnext#]
    .elif (next_header == IPV6_EXTN_ROUTING)
        br[ext_hdr_readnext#]
    .elif (next_header == IPV6_EXTN_FRAGMENT)
        alu[length_needed, length_needed, +, 4];to account for the 4 extra
                                       ;bytes (for identification field) that
                                       ;need to be read in for fragment header
        br[ext_hdr_readnext#]

    .elif (next_header == IPV6_EXTN_DESTINATION)
        br[ext_hdr_readnext#]
    .elif (next_header == IPV6_EXTN_AUTHENTICATION)
        alu[out_security_hdr_present, --, B, 1]
        br[ext_hdr_finish#]
    .elif (next_header == IPV6_EXTN_ENCAPSULATION)
        alu[out_security_hdr_present, --, B, 1]
        br[ext_hdr_finish#]
    .else
        br[ext_hdr_finish#]
    .endif

ext_hdr_readnext#:
    alu[length_needed, length_needed, +, 4];to account for the 4 bytes that
                                           ;need to be read in for all headers
    .if (len_hdr_in_buf < length_needed)
        alu[overflow, --, B, 1]

        alu[out_ipv6_hdr_len, out_ipv6_hdr_len, +, buf_offset]

        //The entire IPv6 header has not been read in.
        //Read in the next chunk. 
        //Note: We don't really care about reading in each and every byte.
        //We begin reading from the point where the next header
        //begins, since all we are interested in finding is where the ipv6
        //header ends. (We figure this out by looking at the next header type)
        //So if the previous buffer contained part of an extension
        //header, we will actually skip that header & read in from the start
        //of the next header.

#ifdef NO_CODESTORE_OVERFLOW
        .reg start_offset_in_header

        move(start_offset_in_header, out_ipv6_hdr_len)
        alu[start_offset_in_header, start_offset_in_header, +, in_L2_hdr_length]

    #ifdef NO_GPR_OVERFLOW
        dl_get_bytes(overflow_buffer,
    #else
        dl_get_bytes(LM_NATPT_WORK_BUF,
    #endif
                     start_offset_in_header,
                     NAT_DRAM_READ_BYTES,
                     sig_ipv6_hdr,
                     sig_ipv6_hdr)
#else
        //Setup the arguments to the call
        move(in_start_offset_in_header, out_ipv6_hdr_len)
        alu[in_start_offset_in_header, in_start_offset_in_header, +,
            in_L2_hdr_length]
        move(in_bytes_to_read, NAT_DRAM_READ_BYTES)
        call(dl_get_bytes#)
#endif //NO_CODESTORE_OVERFLOW

        alu[len_hdr_in_buf, --, B, NAT_DRAM_READ_BYTES]
        alu[buf_offset, --, B, 0]
    .endif

    //Special processing for fragment header. We read in fragment offset,
    //flags & identification field right now (so we don't have to walk the
    //extension header chain again later)
    .if (next_header == IPV6_EXTN_FRAGMENT)
        .if (overflow != 0)
            #ifdef NO_GPR_OVERFLOW
                xbuf_extract(out_fragment_offset_and_flag, overflow_buffer,
                             buf_offset, 2, 2)

                xbuf_extract(out_fragment_identification, overflow_buffer,
                             buf_offset, 4, 4)
            #else
                xbuf_extract(out_fragment_offset_and_flag, LM_NATPT_WORK_BUF,
                             buf_offset, 2, 2)

                xbuf_extract(out_fragment_identification, LM_NATPT_WORK_BUF,
                             buf_offset, 4, 4)
            #endif //NO_GPR_OVERFLOW
        .else
          .begin
            .reg offset_in_frag_header

            alu[offset_in_frag_header, buf_offset, +, 2]

            xbuf_extract(out_fragment_offset_and_flag, in_ipv6_hdr,
                         in_start_offset_in_buf, offset_in_frag_header, 2)

            alu[offset_in_frag_header, buf_offset, +, 4]

            xbuf_extract(out_fragment_identification, in_ipv6_hdr,
                         in_start_offset_in_buf, offset_in_frag_header, 4)
          .end
        .endif

        //Set flag indicating that a fragment header was found
        alu[out_fragment_hdr_present, --, B, 1]
    .endif


    //extract next header & current header length from the appropriate buffer
    .begin
        .reg header_info
        .if (overflow != 0)
            #ifdef NO_GPR_OVERFLOW
                xbuf_extract(header_info, overflow_buffer, 0, buf_offset, 2)
            #else
                xbuf_extract(header_info, LM_NATPT_WORK_BUF, 0, buf_offset, 2)
            #endif
        .else
            xbuf_extract(header_info, in_ipv6_hdr, in_start_offset_in_buf,
                         buf_offset, 2)
        .endif

        ld_field_w_clr[next_header, 0001, header_info, >>8]
        ld_field_w_clr[cur_hdr_len, 0001, header_info]

        //header length does not include the first 8 bytes of the
        //extension header
        alu[cur_hdr_len, cur_hdr_len, +, 1]
        //header length is in 8 byte units (i.e 1 -> 8 bytes)
        alu_shf[cur_hdr_len, --, B, cur_hdr_len, <<3]

        //NOTE: The Fragment extension header definition does not include
        //  header length like the other extension headers. The corresponding
        //  byte in this header is 'reserved'. The reserved field is supposed
        //  to be set to 0 by the transmitter. Therefore, we do not need to do
        //  anything different for this header. We will read in the reserved
        //  field (value = 0), add 1 to it & multiply it by 8, which will
        //  result in a length of 8, which is the correct length for the
        //  fragment header!

        //Now, move past the current header
        alu[buf_offset, buf_offset, +, cur_hdr_len]
    .end

    br[ext_hdr_loop#]

ext_hdr_finish#:
    alu[out_ipv6_hdr_len, out_ipv6_hdr_len, +, buf_offset]
    alu[out_transport_proto, --, B, next_header]

    #ifdef NO_GPR_OVERFLOW
        xbuf_free(overflow_buffer)
    #endif

    #ifdef IPV6_HDR_IN_LM
        xbuf_deactivate(in_ipv6_hdr)
        #undef IPV6_HDR_IN_LM
    #endif
.end
#endm

////////////////////////////////////////////////////////////////////////////////
// _translate_to_v4
//
//    Description:
//        Internal macro called by the main NATPT microblock to translate an
//        IPv6 packet to IPv4. This macro performs the task of translating
//        individual fields in the IPv6 header to the appropriate value in the
//        IPv4 header.
//
//    Parameters:
//        Outputs:
//             - out_ipv4_hdr   : Buffer in which the translated IPv4 header is
//                              stored.
//           - OUT_IPV4_OFFSET: The translated header should be stored in the
//                              buffer starting at this offset
//        Inputs:
//             - in_ipv6_hdr    : Buffer containing the incoming IPv6 header
//             - IN_IPV6_OFFSET : Start offset of header in the buffer passed in
//           - in_v4_dest_addr: The destination address to insert in the
//                              translated IPv4 packet
//           - in_ipv6_src_addr_127_96:
//           - in_ipv6_src_addr_95_64:
//           - in_ipv6_src_addr_63_32:
//           - in_ipv6_src_addr_31_0 :
//           - in_ipv6_hdr_len:
//           - in_transport_proto:
//             - in_fragment_present:
//           - in_fragment_offset_and_flag:
//           - in_fragment_identification:
//                              The above fields in the IPV6 packet are passed
//                              in as they are extracted already in the main
//                              NATPT macro.
//           - in_L2_hdr_length:Length of the layer 2 header encapsulating
//                              the IPV6 header
//           - $nat_entry      :Transfer register containing the binding
//                              information for this packet's address
//    Side effects: Updates dl_meta_packet_size, header_type, start_offset
//                  Updates checksum field in transport header in DRAM
////////////////////////////////////////////////////////////////////////////////
#macro _translate_to_v4( out_ipv4_hdr, OUT_IPV4_OFFSET,\
                         in_ipv6_hdr, IN_IPV6_OFFSET,\
                         in_v4_dest_addr,\
                         in_ipv6_src_addr_127_96, in_ipv6_src_addr_95_64,\
                         in_ipv6_src_addr_63_32, in_ipv6_src_addr_31_0,\
                         in_ipv6_hdr_len, in_transport_proto,\
                         in_fragment_present,\
                         in_fragment_offset_and_flag,\
                         in_fragment_identification,\
                         in_L2_hdr_length,\
                         $nat_entry)
.begin
    .sig sig_nat_table_read

    .reg cur_field
    .reg staging_reg //register containing long word to write into ipv4 header
    .reg v6_payload_len, transport_pdu_len
    .reg result

    //We do not directly write to the out buffer to account for the possibility
    //that the same buffer is passed in for both the in & out headers.
    //If the same buffer is passed in, we copy the contents to a temporary
    //buffer, & only then begin writing the translated header into the
    //buffer passed.
    //Otherwise (if the buffers are different) we work directly with the
    //buffer passed in
    #if (streq(out_ipv4_hdr, in_ipv6_hdr))
        //We create a work buffer in local memory which natpt will use to read
        //in appropriate headers from DRAM. If the ipv4/ipv6 headers passed in
        //are in local memory, there is an issue. We do not know which LM
        //index is set by the application to point to in_ipv6_hdr.
        //Therefore we call xbuf_activate on that buffer in here.
        #if (strstr(in_ipv6_hdr, 'lmem_buf'))
            //The header is in local memory. Lets activate it
            xbuf_activate(in_ipv6_hdr, 0, context_id, 0)

            //Since in_ipv6_hdr == out_ipv4_hdr, we turn on this flag.
            //It will be used later
            #define V4_BUF_IN_LM
        #endif
        xbuf_activate(LM_NATPT_WORK_BUF, 1, context_id, 1)

        xbuf_copy(LM_NATPT_WORK_BUF, 0, 0, in_ipv6_hdr, 0, 0,
                  IPV6_MAIN_HEADER_LENGTH, 0)

        #define_eval work_buffer LM_NATPT_WORK_BUF

        //Note that we should activate the out buffer, so that we can write the
        //translated packet in there. BUT since the in & out buffers are the
        //same this is already done for us.
        //xbuf_activate(out_ipv4_hdr, 0, context_id, 0)
    #else
        #if (strstr(out_ipv4_hdr, 'lmem_buf'))
            xbuf_activate(out_ipv4_hdr, 0, context_id, 0)
            #define V4_BUF_IN_LM
        #endif

        #if (strstr(in_ipv6_hdr, 'lmem_buf'))
            //The header is in local memory. Lets activate it
            xbuf_activate(in_ipv6_hdr, 1, context_id, 1)
        #endif

        #define_eval work_buffer in_ipv6_hdr
    #endif

    //Now we begin the actual translation process.

    //First, the version number & header length
    //Set the header length to the minimum ipv4 header length with no
    //options (i.e 5 long words)
    immed32(staging_reg, 0x45000000)

    //ToS, Precedence (=traffic class) OR 0 based on compile time flag
    #ifdef TRANSLATE_TRAFFIC_CLASS
        ;extract traffic class
        xbuf_extract(cur_field, work_buffer, IN_IPV6_OFFSET, 0, 4)
        ld_field[staging_reg, 0100, cur_field, >>4]
    #endif

    //Now get the length. v6 payload length includes exntension hdr, so
    //subtract that out.
    //Total length = Ipv6 Payload length - length(v6 extension header) + 
    //                 v4 header length

    ;Get the v6 Extension header length
    alu[cur_field, in_ipv6_hdr_len, -, IPV6_MAIN_HEADER_LENGTH]

    //We need to update the buffer size.
    //New buffer size = Old buffer size - Extension Hdr len 
    //                    -IPV6 Main Hdr len + IPV4 Main Hdr len
    //                  = Old buffer size - Extension Hdr len 
    //                    - (IPV6 Main Hdr len - IPV4 Main Hdr len)
    .begin
        .reg buf_size

        dl_meta_get_buffer_size(buf_size)
        alu[buf_size, buf_size, -, cur_field]
        alu[buf_size, buf_size, -, V6V4_MAIN_HDR_LEN_DIFF]
        dl_meta_set_buffer_size(buf_size)
    .end

    xbuf_extract(v6_payload_len, work_buffer, IN_IPV6_OFFSET, 4, 2)

    alu[transport_pdu_len, v6_payload_len, -, cur_field]
     ;calculate total_length
    alu[cur_field, transport_pdu_len, +, IPV4_MAIN_HEADER_LENGTH]

    alu[staging_reg, staging_reg, OR, cur_field] ;OR in total length
    xbuf_insert(out_ipv4_hdr, staging_reg, OUT_IPV4_OFFSET, 0, 4)

    //Update the packet size in meta data
    dl_meta_set_packet_size(cur_field)

    //Now fill the 2nd long word
    //We want to store,
    //+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //|  Identification               |  D M|      Fragment Offset    |
    //+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    .if (in_fragment_present == 0)
        //No fragments
        //Set Identification=0, DF=1, MF=0, Frag offset=0
        immed32(staging_reg, 0x4000)
        xbuf_insert(out_ipv4_hdr, staging_reg, OUT_IPV4_OFFSET, 4, 4)
    .else
        //V6 packet contained a fragment header

        //Pick out the flags & place at right offset. Note DF is set to 0,
        //since v6 in_fragment_offset_and_flag field has 0-0-M in lowest 3 bits
        immed32(staging_reg, 0xE000)
        alu_shf[staging_reg, staging_reg, AND, in_fragment_offset_and_flag,<<13]
        alu_shf[staging_reg, staging_reg, OR, in_fragment_offset_and_flag, >>3]
        //Pick only lower 16 bits from IdentV6
        alu_shf[staging_reg, staging_reg, OR, in_fragment_identification, <<16] 

        xbuf_insert(out_ipv4_hdr, staging_reg, OUT_IPV4_OFFSET, 4, 4)
    .endif

    //Next, Set TTL=Hop Limit, Protocol=Transport proto, Checksum=0
    //We will generate the checksum once the entire header is created
    //+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //|    TTL(8)     |  Protocol(8)  |         Checksum(16)          |
    //+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    xbuf_extract(staging_reg, work_buffer, IN_IPV6_OFFSET, 7, 1) ;Pick out v6
                                                                 ;hop limit
    alu_shf[staging_reg, --, B, staging_reg, <<24]
    alu_shf[staging_reg, staging_reg, OR, in_transport_proto, <<16]
    xbuf_insert(out_ipv4_hdr, staging_reg, OUT_IPV4_OFFSET, 8, 4)

    //Add the source address
    xbuf_insert(out_ipv4_hdr, $nat_entry[NE_V4_ADDR_OFFSET], OUT_IPV4_OFFSET,
                12, 4)

    //Finally, the destination address
    xbuf_insert(out_ipv4_hdr, in_v4_dest_addr, OUT_IPV4_OFFSET, 16, 4)

    //Now generate the IP checksum
        alu[checksum, --, B, 0] ;initialize checksum

#ifdef NO_CODESTORE_OVERFLOW
        xbuf_extract(in_operand, out_ipv4_hdr, OUT_IPV4_OFFSET, 0, 4)
        _modify_checksum(checksum, '+', in_operand)
        xbuf_extract(in_operand, out_ipv4_hdr, OUT_IPV4_OFFSET, 4, 4)
        _modify_checksum(checksum, '+', in_operand)
        xbuf_extract(in_operand, out_ipv4_hdr, OUT_IPV4_OFFSET, 8, 4)
        _modify_checksum(checksum, '+', in_operand)
        xbuf_extract(in_operand, out_ipv4_hdr, OUT_IPV4_OFFSET, 12, 4)
        _modify_checksum(checksum, '+', in_operand)
        xbuf_extract(in_operand, out_ipv4_hdr, OUT_IPV4_OFFSET, 16, 4)
        _modify_checksum(checksum, '+', in_operand)
#else
        xbuf_extract(in_operand, out_ipv4_hdr, OUT_IPV4_OFFSET, 0, 4)
        call(_modify_checksum_ADD#)
        xbuf_extract(in_operand, out_ipv4_hdr, OUT_IPV4_OFFSET, 4, 4)
        call(_modify_checksum_ADD#)
        xbuf_extract(in_operand, out_ipv4_hdr, OUT_IPV4_OFFSET, 8, 4)
        call(_modify_checksum_ADD#)
        xbuf_extract(in_operand, out_ipv4_hdr, OUT_IPV4_OFFSET, 12, 4)
        call(_modify_checksum_ADD#)
        xbuf_extract(in_operand, out_ipv4_hdr, OUT_IPV4_OFFSET, 16, 4)
        call(_modify_checksum_ADD#)
#endif //NO_CODESTORE_OVERFLOW

        //Final checksum is a 16-bit complement of the sum
        _modify_checksum(checksum, 'END', --)

    xbuf_insert(out_ipv4_hdr, checksum, OUT_IPV4_OFFSET, 10, 2)

    //The translated header eventually needs to be flushed back to DRAM
    //Since DRAM writes are done in Quad-word units we will read in some bytes
    //of the transport PDU (which is right after the IP header) into the out
    //buffer (out_ipv4_hdr)
    //We need to read in upto 7 bytes at the most. So lets do this and append it
    //to the translated header.

    //Lets activate LM_NATPT_WORK_BUF, as dl_get_bytes expects to write into it.
    //Also note that past this point we do not use in_ipv6_hdr anymore, so it is
    //ok to activate LM_NATPT_WORK_BUF to index 1
    #if (streq(out_ipv4_hdr, in_ipv6_hdr) == 0)
        xbuf_activate(LM_NATPT_WORK_BUF, 1, context_id, 1)
    #endif

    //Get the offset where the transport PDU begins in the packet
    alu[result, in_ipv6_hdr_len, +, in_L2_hdr_length]

#ifdef NO_CODESTORE_OVERFLOW
    dl_get_bytes(LM_NATPT_WORK_BUF,
                 result,
                 7,
                 sig_dram_read,
                 sig_dram_read)    
#else
    move(in_start_offset_in_header, result)
    move(in_bytes_to_read, 7)
    call(dl_get_bytes#)
#endif //NO_CODESTORE_OVERFLOW

    #define_eval COPY_START_OFFSET (OUT_IPV4_OFFSET + IPV4_MAIN_HEADER_LENGTH)

    //Check if the destination is long-word aligned. If it is not, we need to
    //pass in the bytes in the first long-word, to xbuf_copy
    #define_eval UNALIGNED_BYTES (COPY_START_OFFSET % 4)
	#define_eval BYTES_TO_EXTRACT (4-UNALIGNED_BYTES)
    #if (UNALIGNED_BYTES == 0)
        xbuf_copy(out_ipv4_hdr, 0, COPY_START_OFFSET, LM_NATPT_WORK_BUF,
                  0, 0, 7, 0)
    #else
        .begin
            .reg in_prepend

            xbuf_extract(in_prepend, LM_NATPT_WORK_BUF, 0, 0, BYTES_TO_EXTRACT)
            xbuf_copy(out_ipv4_hdr, 0, COPY_START_OFFSET, LM_NATPT_WORK_BUF,
                      0, in_prepend, 7, 0)
        .end
    #endif
	#undef BYTES_TO_EXTRACT
    #undef UNALIGNED_BYTES

    #undef COPY_START_OFFSET

    //Now lets move to the transport headers. 
    //TCP & UDP checksums include some fields from the IP header
    //Since we changed the IP header we need to modify the transport checksum
    .begin
        .if (PROTO_TCP == in_transport_proto)
            //Get the offset of the TCP checksum in the packet
            //'result' is already set to the start of the TCP header
            alu[result, result, +, TCP_CHECKSUM_OFFSET]

            //Do the incremental checksum update
            //Add in the new v4 source address.
            //Remove the original v6 source address and the v4-mapped-v6 prefix
            //Note that the v4 dest address was already in the 
            //packet (lowest 32 bits of the original destination address). 
            //So nothing to be done for it.
    
            #define_eval in_ipv6_src_addr_127_96 in_ipv6_src_addr_127_96
            #define_eval in_ipv6_src_addr_95_64 in_ipv6_src_addr_95_64
            #define_eval in_ipv6_src_addr_63_32 in_ipv6_src_addr_63_32
            #define_eval in_ipv6_src_addr_31_0 in_ipv6_src_addr_31_0

            _update_TCP_checksum(in_fragment_offset_and_flag,
                                transport_pdu_len,
                                result, 
                                "$nat_entry[0]",
                                "NATPT_PREFIX_0,NATPT_PREFIX_1,NATPT_PREFIX_2, in_ipv6_src_addr_127_96, in_ipv6_src_addr_95_64,in_ipv6_src_addr_63_32, in_ipv6_src_addr_31_0")

        .elif (PROTO_UDP == in_transport_proto)
            //we need the offset to the start of the UDP packet
            //'result' is already set to this value. So we just use it here.

            _update_UDP_over_v6_checksum(result,
                                        "$nat_entry[0]",
                                        "NATPT_PREFIX_0,NATPT_PREFIX_1,NATPT_PREFIX_2, in_ipv6_src_addr_127_96, in_ipv6_src_addr_95_64,in_ipv6_src_addr_63_32, in_ipv6_src_addr_31_0",
                                        in_fragment_offset_and_flag)
        .endif
    .end


    xbuf_deactivate(LM_NATPT_WORK_BUF)

    //Update the header offset in meta data
    //The translated header is of a different size from the original.
    //To ensure that the payload is present right after the header, we change
    //the offset in the meta data, such that when the header is flushed back
    //it lies right before the payload.
    dl_meta_get_offset(cur_field)    ;get the header offset
    alu[cur_field, cur_field, +, in_ipv6_hdr_len] ;move the header offset past the v6 header
    alu[cur_field, cur_field, -, IPV4_MAIN_HEADER_LENGTH] ;move back by the size of the new header
    dl_meta_set_offset(cur_field)

LBL_END_TRANSLATE_TO_V4#:
    //if streq(out_ipv4_hdr, in_ipv6_hdr)
    //        we will deactivate LM_NATPT_WORK_BUF & the buffer passed in
    //else
    //        we will deactivate in_ipv6_hdr (since it is #defined to work_buf)
    //        & out_ipv4_hdr
    xbuf_deactivate(work_buffer)
    #ifdef V4_BUF_IN_LM
        xbuf_deactivate(out_ipv4_hdr)
    #endif

    #undef work_buffer
.end
#endm

////////////////////////////////////////////////////////////////////////////////
// _translate_to_v6
//
//    Description:
//        Internal macro called by the main NATPT microblock to translate an
//        IPv4 packet to IPv6. This macro performs the task of translating
//        individual fields in the IPv4 header to the appropriate value in the
//        IPv6 header.
//
//    Parameters:
//        Outputs:
//           - outcome        : Contains the result of the translation
//           - out_ipv6_hdr   : Buffer in which the translated IPv6 header is
//                              stored.
//           - OUT_IPV6_OFFSET: The translated header should be stored in the
//                              buffer starting at this offset
//        Inputs:
//           - in_ipv4_hdr     : Buffer containing the incoming IPv4 header
//           - IN_IPV4_OFFSET  : Start offset of header in the buffer passed in
//           - in_L2_hdr_length:Length of the layer 2 header encapsulating
//                              the IPV4 header
//           - in_dest_addr4   :The destination address in the IPv4 packet to
//                              be translated. This is used to update the
//                              transport checksum.
//           - context_id      :Id of the ME context running the code
//           - $nat_entry      :Transfer register containing the binding
//                              information for this packet's address
//    Side effects: Updates meta data (packet_size, start_offset & header_type)
//                    Updates checksum field in transport header in DRAM
////////////////////////////////////////////////////////////////////////////////
#macro _translate_to_v6( outcome, out_ipv6_hdr, OUT_IPV6_OFFSET, in_ipv4_hdr,\
                         IN_IPV4_OFFSET, in_L2_hdr_length, in_dest_addr4,\
                         context_id, $nat_entry)
.begin
    .sig sig_dram_op

    .reg ipv4_lw    //Long word from the ipv4 header
    .reg cur_field
    .reg staging_reg
    .reg total_length4, hdr_length4
    .reg transport_proto
    .reg ttl
    .reg more_frags4
    .reg fragment_offset_and_flag4
    .reg fragment_present
    .reg src_addr4
    .reg flush_offset

    //We do not directly write to the out buffer to account for the possibility
    //that the same buffer is passed in for both the in & out headers.
    //If the same buffer is passed in, we copy the contents to a temporary
    //buffer, & only then begin writing the translated header into the
    //buffer passed.
    //Otherwise (if the buffers are different) we work directly with the
    //buffer passed in
    #if (streq(out_ipv6_hdr, in_ipv4_hdr))
        //We create a work buffer in local memory which natpt will use to read
        //in appropriate headers from DRAM.
        //If the ipv4/ipv6 headers passed in are in local memory, there is
        //an issue. We do not know which LM index is set by the application to 
        //point to in_ipv4_hdr. Therefore we call xbuf_activate on that buffer  
        //in here
        #if (strstr(in_ipv4_hdr, 'lmem_buf'))
            //The header is in local memory. Lets activate it
            xbuf_activate(in_ipv4_hdr, 0, context_id, 0)

            //Since out_ipv6_hdr == in_ipv4_hdr, we turn on this flag.
            //It will be used later
            #define V6_BUF_IN_LM
        #endif
        xbuf_activate(LM_NATPT_WORK_BUF, 1, context_id, 1)

        xbuf_copy(LM_NATPT_WORK_BUF, 0, 0, in_ipv4_hdr, 0, 0, 
                  IPV4_MAIN_HEADER_LENGTH, 0)
        #define_eval work_buffer LM_NATPT_WORK_BUF

        //Note that we should activate the out buffer, so that we can write the
        //translated packet in there. BUT since the in & out buffers are the
        //same this is already done for us.
        //xbuf_activate(out_ipv6_hdr, 0, context_id, 0)
    #else
        #if (strstr(out_ipv6_hdr, 'lmem_buf'))
            //The header is in local memory. Lets activate it
            xbuf_activate(out_ipv6_hdr, 0, context_id, 0)
            #define V6_BUF_IN_LM
        #endif

        #if (strstr(in_ipv4_hdr, 'lmem_buf'))
            //The header is in local memory. Lets activate it
            xbuf_activate(in_ipv4_hdr, 1, context_id, 1)
        #endif

        #define_eval work_buffer in_ipv4_hdr
    #endif

    //The first thing we do is to check the transport protocol
    //This is because we do not translate ICMP packets.
    //So, if the proto is ICMP we throw it to the core right away and avoid
    //translating the packet alltogether!
    
    //We also store the ttl for later use
    xbuf_extract(ipv4_lw, work_buffer, IN_IPV4_OFFSET, 8, 4)
    ld_field_w_clr[transport_proto, 0001, ipv4_lw, >>16] ;right shift proto to LSB & load
    ld_field_w_clr[ttl, 0001, ipv4_lw, >>24] ;right shift TTL to LSB & load

    .if (PROTO_ICMP_V4 == transport_proto)
        br[LBL_END_TRANSLATE_TO_V6#], defer[1]
        alu[outcome, --, B, NATPT_EXCEPTION_ICMP_IPV4]
    .endif

    //Now we begin the actual translation process.
    //Add in the version (=6), 
    //the traffic class (= TOS in IPv4) and 
    //flow label (=0)
    //+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //|Version| Traffic Class |           Flow Label                  |
    //+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    //First, extract the TOS field
    xbuf_extract(ipv4_lw, work_buffer, IN_IPV4_OFFSET, 0, 4)
    ld_field_w_clr[cur_field, 0001, ipv4_lw, >>16] ;get the tos field
    alu_shf[staging_reg, --, B, 0x6, <<28]
    alu_shf[staging_reg, staging_reg, OR, cur_field, <<20] ;Or in tos

    xbuf_insert(out_ipv6_hdr, staging_reg, OUT_IPV6_OFFSET, 0, 4)

    ld_field_w_clr[total_length4, 0011, ipv4_lw]
    alu_shf[hdr_length4, 0xf, AND, ipv4_lw, >>24]
    //IPv4 Header length is in 32 bit words. Convert it to bytes
    alu_shf[hdr_length4, --, B, hdr_length4, <<2]

    //Now pick out the fields from the 2nd LW in the ipv4 header
    xbuf_extract(ipv4_lw, work_buffer, IN_IPV4_OFFSET, 4, 4)

    .begin
        .reg mask
        immed[mask, 0x1fff, <<0]
        alu_shf[fragment_offset_and_flag4, mask, AND, ipv4_lw] ;pick out frag_offset
    .end

    //Check if DF bit is clear
    br_bclr[ipv4_lw, IPV4_DONT_FRAG_BIT, LBL_ADD_FRAGMENT#], defer[2]

    alu_shf[more_frags4, 1, AND, ipv4_lw, >>IPV4_MORE_FRAG_BIT]  ;pick out the 'More Fragments' Flag

    //We merge the fragment offset & more frags field.
    //if frag_offset == 0 && more_frags == 0, it means there are no fragments
    //By merging into one field we just need to do one test to determine this.
    //Merging the fields has an added benefit. If a fragment header needs to
    //be added, we need to write in the frag_offset & more_flags fields. We 
    //merge the fields now in the same format as is needed in the fragment
    //header. This way we can simply reuse this merged field.
    alu_shf[fragment_offset_and_flag4, more_frags4, OR,
            fragment_offset_and_flag4, <<3]
    
    //DF bit is set. Now check if fragment offset = 0 & MF bit is clear
    .if (fragment_offset_and_flag4 == 0)
        //No fragments

        //Update the header offset in meta data
        //The translated header is of a different size from the original.
        //To ensure that the payload is present right after the header, we change
        //the offset in the meta data, such that when the header is flushed back
        //it lies right before the payload.
        //It is assumed here that there will be some headroom in the packet buffer,
        //so the meta data offset will not go below zero, by the following subtractions.
        dl_meta_get_offset(cur_field)    ;get the header offset
        alu[cur_field, cur_field, +, hdr_length4] ;move the header offset past the v4 header
        alu[cur_field, cur_field, -, IPV6_MAIN_HEADER_LENGTH] ;move back by the size of the new header
        dl_meta_set_offset(cur_field)

        //We need to update the buffer size.
        //New buffer size = Old buffer size - Extension Hdr len 
        //                    -IPV6 Main Hdr len + IPV4 Main Hdr len
        //                  = Old buffer size - Extension Hdr len 
        //                    - (IPV6 Main Hdr len - IPV4 Main Hdr len)
        .begin
            .reg buf_size

            dl_meta_get_buffer_size(buf_size)
            alu[buf_size, buf_size, -, hdr_length4];subtract main v4 & any options
            alu[buf_size, buf_size, +, IPV6_MAIN_HEADER_LENGTH]
            dl_meta_set_buffer_size(buf_size)
        .end

        //+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        //|         Payload Length        |  Next Header  |   Hop Limit   |
        //+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        alu[cur_field, total_length4, -, hdr_length4]
        alu_shf[staging_reg, ttl, OR, cur_field, <<16]

        br[LBL_PAST_ADD_FRAGMENT#], defer[2]
        alu_shf[staging_reg, staging_reg, OR, transport_proto, <<8]
        alu[fragment_present, --, B, 0]
    .endif

LBL_ADD_FRAGMENT#:
    //We come here if we need to add a fragment header to the IPv6 packet
    //created. Lets first create the fragment extension header
    //+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //|  Next Header  |   Reserved    |      Fragment Offset    |Res|M|
    //+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //|                         Identification                        |
    //+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //
    // M = More Fragments
    // We set Next header = transport protocol

    alu[fragment_present, --, B, 1]

    alu_shf[staging_reg, fragment_offset_and_flag4, OR, transport_proto, <<24]
    //Insert the fragment header right after the main ipv6 header (40 bytes)
    xbuf_insert(out_ipv6_hdr, staging_reg, OUT_IPV6_OFFSET, 40, 4) 

    //Update the header offset in meta data
    //The translated header is of a different size from the original.
    //To ensure that the payload is present right after the header, we change
    //the offset in the meta data, such that when the header is flushed back
    //it lies right before the payload.
    //It is assumed here that there will be some headroom in the packet buffer,
    //so the meta data offset will not go below zero, by the following subtractions.
    dl_meta_get_offset(cur_field)    ;get the header offset
    alu[cur_field, cur_field, +, hdr_length4] ;move the header offset past the v4 header
    alu[cur_field, cur_field, -, IPV6_MAIN_HEADER_LENGTH] ;move back by the size of the new header
    alu[cur_field, cur_field, -, IPV6_FRAGMENT_HDR_SIZE]
    dl_meta_set_offset(cur_field)

    //We need to update the buffer size.
    //New buffer size = Old buffer size - Extension Hdr len 
    //                    -IPV6 Main Hdr len + IPV4 Main Hdr len
    //                  = Old buffer size - Extension Hdr len 
    //                    - (IPV6 Main Hdr len - IPV4 Main Hdr len)
    .begin
        .reg buf_size
        dl_meta_get_buffer_size(buf_size)
        alu[buf_size, buf_size, -, hdr_length4];subtract main v4 & any options
        alu[buf_size, buf_size, +, V6MAIN_FRAG_HDR_LEN_SUM]
        dl_meta_set_buffer_size(buf_size)
    .end


    ld_field_w_clr[cur_field, 0011, ipv4_lw, >>16] ;load ident4
    xbuf_insert(out_ipv6_hdr, cur_field, OUT_IPV6_OFFSET, 44, 4)

    //Now setup the main header
    //+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //|         Payload Length        |  Next Header  |   Hop Limit   |
    //+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    alu[cur_field, total_length4, -, hdr_length4]
    alu[cur_field, cur_field, +, IPV6_FRAGMENT_HDR_SIZE] ;Add size of a fragment header
    alu_shf[staging_reg, ttl, OR, IPV6_EXTN_FRAGMENT, <<8]
    alu_shf[staging_reg, staging_reg, OR, cur_field, <<16]

LBL_PAST_ADD_FRAGMENT#:
    //Fill in the 2nd LW in the ipv6 header
    xbuf_insert(out_ipv6_hdr, staging_reg, OUT_IPV6_OFFSET, 4, 4) 

    //Update the packet size in meta data. 
    //cur_field holds the payload length (which includes extension headers, if
    //present). Add in the length of the main header
    alu[cur_field, cur_field, +, IPV6_MAIN_HEADER_LENGTH]
    dl_meta_set_packet_size(cur_field)

    //Next, lets fill in the source address
    move(staging_reg, NATPT_PREFIX_0)
    xbuf_insert(out_ipv6_hdr, staging_reg, OUT_IPV6_OFFSET, 8, 4)

    move(staging_reg, NATPT_PREFIX_1)
    xbuf_insert(out_ipv6_hdr, staging_reg, OUT_IPV6_OFFSET, 12, 4)

    move(staging_reg, NATPT_PREFIX_2)
    xbuf_insert(out_ipv6_hdr, staging_reg, OUT_IPV6_OFFSET, 16, 4)

    xbuf_extract(src_addr4, work_buffer, IN_IPV4_OFFSET, IP_SOURCE_ADDRESS)
    xbuf_insert(out_ipv6_hdr, src_addr4, OUT_IPV6_OFFSET, 20, 4)

    //Finally, the destination address
    xbuf_insert(out_ipv6_hdr, $nat_entry[NE_V6_ADDR_127_96_OFFSET], OUT_IPV6_OFFSET, 24, 4)
    xbuf_insert(out_ipv6_hdr, $nat_entry[NE_V6_ADDR_95_64_OFFSET], OUT_IPV6_OFFSET, 28, 4)
    xbuf_insert(out_ipv6_hdr, $nat_entry[NE_V6_ADDR_63_32_OFFSET], OUT_IPV6_OFFSET, 32, 4)
    xbuf_insert(out_ipv6_hdr, $nat_entry[NE_V6_ADDR_31_0_OFFSET], OUT_IPV6_OFFSET, 36, 4)

    //The translated header eventually needs to be flushed back to DRAM
    //Since DRAM writes are done in Quad-word units we will read in some bytes
    //of the transport PDU (which is right after the IP header) into the out
    //buffer (out_ipv6_hdr)
    //We need to read in upto 7 bytes at the most. So lets do this and append it
    //to the translated header.

    //Lets activate LM_NATPT_WORK_BUF, as dl_get_bytes expects to write into it.
    //Also note that past this point we do not use in_ipv4_hdr anymore, so it is
    //ok to activate LM_NATPT_WORK_BUF to index 1
    #if (streq(out_ipv6_hdr, in_ipv4_hdr) == 0)
        xbuf_activate(LM_NATPT_WORK_BUF, 1, context_id, 1)
    #endif

    //Get the offset where the transport PDU begins in the packet
    alu[result, hdr_length4, +, in_L2_hdr_length]

#ifdef NO_CODESTORE_OVERFLOW
    dl_get_bytes(LM_NATPT_WORK_BUF,
                 result,
                 7,
                 sig_dram_read,
                 sig_dram_read)    
#else
    move(in_start_offset_in_header, result)
    move(in_bytes_to_read, 7)
    call(dl_get_bytes#)
#endif //NO_CODESTORE_OVERFLOW

    //We need the offset in the out buffer where the transport PDU begins
    //Basically if a fragment header is present in the translated packet, 
    //we know that the transport PDU starts
    //at offset 48 (IPV6_MAIN_HEADER_LENGTH + IPV6_FRAGMENT_HDR_SIZE)
    //else at offset 40 (IPV6_MAIN_HEADER_LENGTH)
    .if (fragment_present == 1)
      .begin
        .reg offset_in_v6_hdr

        #define_eval TRANS_OFFSET (IPV6_MAIN_HEADER_LENGTH + IPV6_FRAGMENT_HDR_SIZE)
        #define_eval COPY_START_OFFSET (OUT_IPV6_OFFSET + TRANS_OFFSET)

        //Check if the destination is long-word aligned. If it is not, we need to
        //pass in the bytes in the first long-word, to xbuf_copy
        #define_eval UNALIGNED_BYTES (COPY_START_OFFSET % 4)
		#define_eval BYTES_TO_EXTRACT (4-UNALIGNED_BYTES)
        #if (UNALIGNED_BYTES == 0)
            xbuf_copy(out_ipv6_hdr, 0, COPY_START_OFFSET, LM_NATPT_WORK_BUF,
                      0, 0, 7, 0)
        #else
            .begin
                .reg in_prepend

                xbuf_extract(in_prepend, LM_NATPT_WORK_BUF, 0, 0, BYTES_TO_EXTRACT)
                xbuf_copy(out_ipv6_hdr, 0, COPY_START_OFFSET, LM_NATPT_WORK_BUF,
                          0, in_prepend, 7, 0)
            .end
        #endif
		#undef BYTES_TO_EXTRACT
        #undef UNALIGNED_BYTES

        #undef COPY_START_OFFSET


        //Only the main IPv6 header is flushed back to DRAM by the header
        //caching mechanism provided. So we will have to flush the extension
        //header to DRAM ourselves.

        //if result (i.e start of the transport PDU) is QW aligned, 
        //flush the fragment header 1 QW before the current QW
        //else, 
        //flush the 8 bytes at the current QW in the buffer (ending of frag
        //header + start of payload). The logic for flushing
        //the main IPv6 header will take care of flushing the first few bytes
        //of the fragment header.(since the last few bytes of the main header
        //are in the same QW)
        alu_shf[offset_in_v6_hdr, result, AND, 7]
        bne[LBL_NOT_QW_ALIGNED#], defer[1]
        alu[flush_offset, --, B, result]

            //The offset is QW aligned
            br[LBL_FLUSH_HEADER#], defer[2]
            alu[flush_offset, flush_offset, -, 8]
            alu[offset_in_v6_hdr, --, B, IPV6_MAIN_HEADER_LENGTH] ;start of 
                                                            ;fragment header

LBL_NOT_QW_ALIGNED#:
        alu[offset_in_v6_hdr, TRANS_OFFSET, -, offset_in_v6_hdr] ;get to start
                                ;of fragment header that lies in the current QW
        #undef TRANS_OFFSET

LBL_FLUSH_HEADER#:
        xbuf_alloc($$xfer_buf, 2, write)

        xbuf_extract($$xfer_buf0, out_ipv6_hdr, OUT_IPV6_OFFSET, offset_in_v6_hdr, 4)
        alu[offset_in_v6_hdr, offset_in_v6_hdr, +, 4]
        xbuf_extract($$xfer_buf1, out_ipv6_hdr, OUT_IPV6_OFFSET, offset_in_v6_hdr, 4)
        dl_put_quad_words($$xfer_buf0, flush_offset, 1, sig_dram_op, sig_dram_op)

        xbuf_free($$xfer_buf)
      .end
    .else
        #define_eval COPY_START_OFFSET (OUT_IPV6_OFFSET + IPV6_MAIN_HEADER_LENGTH)

        //Check if the destination is long-word aligned. If it is not, we need to
        //pass in the bytes in the first long-word, to xbuf_copy
        #define_eval UNALIGNED_BYTES (COPY_START_OFFSET % 4)
		#define_eval BYTES_TO_EXTRACT (4-UNALIGNED_BYTES)
        #if (UNALIGNED_BYTES == 0)
            xbuf_copy(out_ipv6_hdr, 0, COPY_START_OFFSET, LM_NATPT_WORK_BUF,
                      0, 0, 7, 0)
        #else
            .begin
                .reg in_prepend

                xbuf_extract(in_prepend, LM_NATPT_WORK_BUF, 0, 0, BYTES_TO_EXTRACT)
                xbuf_copy(out_ipv6_hdr, 0, COPY_START_OFFSET, LM_NATPT_WORK_BUF,
                          0, in_prepend, 7, 0)
            .end
        #endif
		#undef BYTES_TO_EXTRACT
        #undef UNALIGNED_BYTES

        #undef COPY_START_OFFSET
    .endif

    //Now lets move to the transport headers. 
    //TCP & UDP checksums include some fields from the IP header
    //Since we changed the IP header we need to modify the transport checksum
    .if (PROTO_TCP == transport_proto)
      .begin
        .reg transport_pdu_len

        //This could have been moved to the code path where frag_offset<=2,
        //if the _update_TCP_checksum function had been written out over here
        alu[transport_pdu_len, total_length4, -, hdr_length4]

        //Get the offset of the TCP checksum in the packet
        //'result' is already set to the start of the TCP header
        alu[result, result, +, TCP_CHECKSUM_OFFSET]

        //The following is needed to get the actual identifier name passed to
        //this macro.
        #define_eval actual_token in_dest_addr4

        _update_TCP_checksum(fragment_offset_and_flag4,
                            transport_pdu_len,
                            result, 
                            "NATPT_PREFIX_0,NATPT_PREFIX_1,NATPT_PREFIX_2,$nat_entry[1],$nat_entry[2],$nat_entry[3],$nat_entry[4]",
                            "actual_token")

        #undef actual_token
      .end
    .elif (PROTO_UDP == transport_proto)
      .begin

        #define_eval actual_token in_dest_addr4

        //we need the offset to the start of the UDP packet
        //'result' is already set to this value. So we just use it here.

        _update_UDP_over_v4_checksum(outcome,
                            result,
                            "NATPT_PREFIX_0,NATPT_PREFIX_1,NATPT_PREFIX_2,$nat_entry[1],$nat_entry[2],$nat_entry[3],$nat_entry[4]",
                            "actual_token",
                            fragment_offset_and_flag4,
                            src_addr4,
                            LBL_END_TRANSLATE_TO_V6#)

        #undef actual_token
      .end
    .endif

    xbuf_deactivate(LM_NATPT_WORK_BUF)

    alu[outcome, --, B, 0] ;translation was successful

LBL_END_TRANSLATE_TO_V6#:
    //if streq(out_ipv6_hdr, in_ipv4_hdr)
    //        we will deactivate LM_NATPT_WORK_BUF & the buffer passed in
    //else
    //        we will deactivate in_ipv4_hdr (since it is #defined to work_buf)
    //        & out_ipv6_hdr
    xbuf_deactivate(work_buffer)

    #ifdef V6_BUF_IN_LM
        xbuf_deactivate(out_ipv6_hdr)
    #endif

    #undef work_buffer
.end
#endm

////////////////////////////////////////////////////////////////////////////////
// _update_UDP_over_v6_checksum
//
//    Description:
//        Internal macro that updates the checksum in a UDP PDU carried in an
//        IPv6 packet. The original checksum is updated by including the fields
//        in 'in_add_list' while the fields in 'in_del_list' are excluded.
//        Note that this macro updates (& possibly even reads) the packet directly
//        from DRAM.
//
//    Parameters:
//        Outputs:
//        Inputs:
//          - in_UDP_start_offset: Offset in packet where the UDP PDU begins
//          - in_add_list        : String containing a comma separated list of
//                                 the fields that need to be included in the
//                                 checksum
//          - in_del_list        : String containing a comma separated list of
//                                 the fields that need to be removed from the
//                                 checksum
//          - in_fragment_offset_and_flag:
//                                 Field containing the fragment offset & MF
//                                 flag from the IP header
//
//     Side effects:
//         The checksum field in the UDP PDU is updated directly in DRAM
////////////////////////////////////////////////////////////////////////////////
#macro _update_UDP_over_v6_checksum(in_UDP_start_offset,\
                                    in_add_list,\
                                    in_del_list,\
                                    in_fragment_offset_and_flag)
//IPv6 mandates that the UDP checksum must be generated when a UDP packet is
//transported over IPv6. So we assume that the UDP checksum cannot be 0 and
//hence we can do an incremental update of the checksum even if the packet
//is fragmented
.begin
    .sig sig_dram_udp
    .reg offset
    .reg checksum_pos //Offset of checksum in the Quad Word read

    .if (in_fragment_offset_and_flag != 0)
        //The UDP checksum is at offset 6
        //Fragment offsets are in 8 byte octets. So checksum field can be
        //present in the fragment only if offset is 0
        alu_shf[in_fragment_offset_and_flag, --, B,
                in_fragment_offset_and_flag, >>3]

        //Test the zero flag to check if this is the first fragment.
        //Nothing to do if it is.
        beq[LBL_PAST_UDP_CHECKSUM_UPDATE#]
    .endif

    //We come here if the checksum is present in this packet (or fragment)

    //First, we need to read in the original checksum.
    alu[offset, in_UDP_start_offset, +, UDP_CHECKSUM_OFFSET]

    alu[checksum_pos, offset, AND, 7]

    //The offset needs to be quadword aligned. Actually dl_get_bytes takes care
    //of this. But since we write back the modified value to DRAM and the
    //relevant bytes might already be in the cache, we need to do this.
    //(Otherwise when the bytes are picked up from the cache they will land up
    // at offset 0 in the returned buffer, while they will be at offset 
    // checksum_pos when they are picked up from DRAM)
    alu_shf[offset, offset, AND~, 7] ;mask off the lower 3 bits

    //We read in 8 or 16 bytes i.e. one or two quad words, because we need to 
    //write back the modified checksum, possibly to DRAM (if the checksum
    //is not present in the header cache) which works in quad word
    //units.
#ifdef NO_CODESTORE_OVERFLOW
    .if (checksum_pos == 7)
        //The checksum field straddles a Quadword boundary.
        //So we read in two Quadwords.
        dl_get_bytes(LM_NATPT_WORK_BUF,
                     offset,
                     16,
                     sig_dram_udp,
                     sig_dram_udp)    
    .else
        dl_get_bytes(LM_NATPT_WORK_BUF,
                     offset,
                     8,
                     sig_dram_udp,
                     sig_dram_udp)
    .endif
#else
    move(in_start_offset_in_header, offset)

    .if (checksum_pos == 7)
        //The checksum field straddles a Quadword boundary.
        //So we read in two Quadwords.
        move(in_bytes_to_read, 16)
    .else
        move(in_bytes_to_read, 8)
    .endif
    call(dl_get_bytes#)
#endif //NO_CODESTORE_OVERFLOW

/*The checksum modification code is ifdef'd out as we lack space in the code
store. We will simply set the UDP checksum to 0 when translating from
v6 to v4. This is considered acceptable [rfc 2766, 5.3.2]
*/
#ifdef NO_CODESTORE_OVERFLOW
    //Extract the checksum from the buffer
    xbuf_extract(checksum, LM_NATPT_WORK_BUF, 0, checksum_pos, 2)

    _modify_checksum(checksum, 'BEGIN', --)

    //Subtract out the fields that were removed
    _repeat_per_elem_in_list(del_wrapper, in_del_list)

    //Add in the new fields 
    _repeat_per_elem_in_list(add_wrapper, in_add_list)

    _modify_checksum(checksum, 'END', --)
#else
    alu[checksum, --, B, 0]
#endif //NO_CODESTORE_OVERFLOW

    //Insert the updated checksum

    xbuf_insert(LM_NATPT_WORK_BUF, checksum, 0, checksum_pos, 2)


/*
    The specific case in which we invoke the xbuf_copy macro in 
    dl_put_quad_words_from_gpr is not supported.
    If enhanced versions of the macros become available use the code
    in the #if part below. The code in the #else part is correct but
    will probably not be as efficient.
*/
#ifdef ENHANCED_XBUF_API
    .if (checksum_pos == 7)
        dl_put_quad_words_from_gpr_lm(LM_NATPT_WORK_BUF,
                                      offset,
                                      2,
                                      sig_dram_udp,
                                      sig_dram_udp)
    .else
        dl_put_quad_words_from_gpr_lm(LM_NATPT_WORK_BUF,
                                      offset,
                                      1,
                                      sig_dram_udp,
                                      sig_dram_udp)
    .endif        
#else
    xbuf_alloc($$xfer_buf, 4, write)

    xbuf_extract($$xfer_buf0, LM_NATPT_WORK_BUF, 0, 0, 4)
    xbuf_extract($$xfer_buf1, LM_NATPT_WORK_BUF, 0, 4, 4)

    .if (checksum_pos == 7)
        xbuf_extract($$xfer_buf2, LM_NATPT_WORK_BUF, 0, 8, 4)
        xbuf_extract($$xfer_buf3, LM_NATPT_WORK_BUF, 0, 12, 4)

        dl_put_quad_words($$xfer_buf0, offset, 2, sig_dram_udp, sig_dram_udp)
    .else
        dl_put_quad_words($$xfer_buf0, offset, 1, sig_dram_udp, sig_dram_udp)
    .endif        
#endif //ENHANCED_XBUF_API

    xbuf_free($$xfer_buf)

LBL_PAST_UDP_CHECKSUM_UPDATE#:
.end
#endm

////////////////////////////////////////////////////////////////////////////////
// _update_UDP_over_v4_checksum
//
//    Description:
//        Internal macro that updates the checksum in a UDP PDU carried in an
//        IPv4 packet. The original checksum is updated by including the fields
//        in 'in_add_list' while the fields in 'in_del_list' are excluded.
//        Note that this macro updates (& possibly even reads) the packet directly
//        from DRAM.
//
//    Parameters:
//        Outputs:
//          - out_retval : In case the checksum is not updated, this parameter
//                         is updated with a reason code and the code flow
//                         is switched to the label passed in. Note that when
//                         the checksum is updated, this parameter is not
//                         touched.
//        Inputs:
//          - in_UDP_start_offset: Offset in packet where the UDP PDU begins
//          - in_add_list        : String containing a comma separated list of
//                                 the fields that need to be included in the
//                                 checksum
//          - in_del_list        : String containing a comma separated list of
//                                 the fields that need to be removed from the
//                                 checksum
//          - in_fragment_offset_and_flag:
//                                 Field containing the fragment offset & MF
//                                 flag from the IP header
//          - in_src_addr4       : The IP source address in the original IPv4
//                                 packet. This is needed in the case when the
//                                 checksum field has to be generated (instead
//                                 of being updated)
//          - in_err_label       : Label to jump to in case of error
//     Side effects:
//         The checksum field in the UDP PDU is updated directly in DRAM
////////////////////////////////////////////////////////////////////////////////

#macro _update_UDP_over_v4_checksum(out_retval,\
                            in_UDP_start_offset,\
                            in_add_list,\
                            in_del_list,\
                            in_fragment_offset_and_flag,\
                            in_src_addr4,\
                            in_err_label)
    //First we check if the packet is fragmented. If it is, we send it to the
    //core for processing. This is because of 2 reasons,
    //first, if the checksum is 0 we have to generate it
    //second, we cannot afford to wait for all the fragments, reassemble & then
    //generate the checksum.
    //Unlike in the v6 case, we cannot do an incremental update of fragmented
    //packets since there is no guarantee that the existing checksum will be
    //non-zero. If we pass on fragments (instead of sending to core) which do
    //not contain the checksum and then find that the first fragment(offset 0)
    //contains no checksum (i.e. checksum=0, i.e. checksum wasn't generated)
    //there is no way we can reassemble the packet(as we might have already
    //forwarded some of the fragments containing the latter half of the packet)
    //Hence, if the packet is fragmented we send it to the core.
    .if (in_fragment_offset_and_flag != 0)
        alu[out_retval, --, B, NATPT_EXCEPTION_REASSEMBLE_UDP]
        br [in_err_label]
    .endif

    //If Checksum present
    //    Do an incremental checksum update
    //Else
    //    Generate the checksum
.begin
    .sig sig_dram_udp
    .reg udp_offset
    .reg length_pos //Offset of UDP length field in quad word
    .reg udp_length

    //First, we need to read in the original checksum.
    //We might also need the UDP length(if checksum has to be generated), 
    //which is just before the checksum field in the header. 
    //So we read from an offset such that we get both the fields
    alu[udp_offset, in_UDP_start_offset, +, UDP_LENGTH_OFFSET]

    //Is it faster to just read in 16 bytes always?
    alu[length_pos, udp_offset, AND, 7]

    //The offset needs to be quadword aligned. Actually dl_get_bytes takes care
    //of this. But since we write back the modified value to DRAM and the
    //relevant bytes might already be in the cache, we need to do this.
    //(Otherwise when the bytes are picked up from the cache they will land up
    // at offset 0 in the returned buffer, while they will be at offset 
    // length_pos when they are picked up from DRAM)
    alu_shf[udp_offset, udp_offset, AND~, 7] ;mask off the lower 3 bits

    //We read in 8 or 16 bytes i.e. one or two quad words, because we need to 
    //write back the modified checksum, possibly to DRAM (if the checksum
    //is not present in the header cache) which works in quad word
    //units.
#ifdef NO_CODESTORE_OVERFLOW
    .if (length_pos > 4)
        //The length & checksum fields straddle a Quadword boundary.
        //So we read in two Quadwords.
        dl_get_bytes(LM_NATPT_WORK_BUF,
                     udp_offset,
                     16,
                     sig_dram_udp,
                     sig_dram_udp)    
    .else
        dl_get_bytes(LM_NATPT_WORK_BUF,
                     udp_offset,
                     8,
                     sig_dram_udp,
                     sig_dram_udp)
    .endif
#else
    move(in_start_offset_in_header, udp_offset)

    .if (length_pos > 4)
        //The length & checksum fields straddle a Quadword boundary.
        //So we read in two Quadwords.
        move(in_bytes_to_read, 16)
    .else
        move(in_bytes_to_read, 8)
    .endif
    call(dl_get_bytes#)
#endif //NO_CODESTORE_OVERFLOW

    //Extract the checksum from the buffer

    xbuf_extract(checksum, LM_NATPT_WORK_BUF, 2, length_pos, 2)

    .if (checksum != 0)
        _modify_checksum(checksum, 'BEGIN', --)

        //Subtract out the fields that were removed
        _repeat_per_elem_in_list(del_wrapper, in_del_list)

        //Add in the new fields 
        _repeat_per_elem_in_list(add_wrapper, in_add_list)

        _modify_checksum(checksum, 'END', --)
    .else
        //Complete UDP packet with no checksum. 
        //If this is the common case, we will handle
        //it here by generating a checksum. Else can simply throw
        //an exception if checksum is 0 & let the core handle it

//The code for UDP checksum generation is ifdef'd out due to lack of space in
//the code store. We will throw an exception to the core in this case
        #ifdef NO_CODESTORE_OVERFLOW
                //Extract the length from the buffer

                xbuf_extract(udp_length, LM_NATPT_WORK_BUF, 0, length_pos, 2)

                //Pass in the fields of the UDP pseudo header for
                //checksum calculation
                //_generate_checksum returns the checksum in 'checksum'
                _generate_checksum(in_UDP_start_offset, udp_length)
                xbuf_activate(LM_NATPT_WORK_BUF, 1, context_id, 0)

                //Now add in the fields of the pseudo-header
                //Add list contains all the newly added fields, i.e. 
                //the destination address and the V4-Mapped-V6 prefix.
                //In addition we will need to add in the transport protocol & the
                //original source address
                _repeat_per_elem_in_list(add_wrapper, in_add_list)

            //This is redundant at the moment. But left in here in case there
            //is enough space to generate the checksum, but not enough to make
            //the following macro calls
            #ifdef NO_CODESTORE_OVERFLOW
                _modify_checksum(checksum, '+', PROTO_UDP)
                _modify_checksum(checksum, '+', in_src_addr4)
            #else
                alu[in_operand, --, B, PROTO_UDP]
                call(_modify_checksum_ADD#)

                alu[in_operand, --, B, in_src_addr4]
                call(_modify_checksum_ADD#)
            #endif //NO_CODESTORE_OVERFLOW

                _modify_checksum(checksum, 'END', --)
        #else
            //Set the exception code and exit from here
            br [in_err_label], defer[1]
            alu[out_retval, --, B, NATPT_EXCEPTION_UDP_V4_CHECKSUM_ZERO]
        #endif //NO_CODESTORE_OVERFLOW
    .endif //(checksum != 0)

    //Insert the updated checksum

    xbuf_insert(LM_NATPT_WORK_BUF, checksum, 2, length_pos, 2)

    //We only modified the checksum so we only write that back

/*
    The specific case in which we invoke the xbuf_copy macro in 
    dl_put_quad_words_from_gpr is not supported.
    If enhanced versions of the macros become available use the code
    in the #if part below. The code in the #else part does the same operations
    but will probably not be as efficient.
*/
#ifdef ENHANCED_XBUF_API
    .if (length_pos == 7);checksum is in 2nd QW
      .begin
        alu[udp_offset, udp_offset, +, 8]

        xbuf_alloc($$xfer_udp_hdr, 2, write)
        xbuf_copy($$xfer_udp_hdr, 0, 0, LM_NATPT_WORK_BUF, 1, 0, 8, 0)

        dl_put_quad_words($$xfer_udp_hdr0,
                          udp_offset,
                          1,
                          sig_dram_udp,
                          sig_dram_udp)
        xbuf_free($$xfer_udp_hdr)
      .end
    .else
        .begin
            .if (length_pos == 6);checksum is in 8th byte of this QW & 1st byte of next
                //We need to write both QW back
                dl_put_quad_words_from_gpr_lm(LM_NATPT_WORK_BUF,
                                  udp_offset,
                                  2,
                                  sig_dram_udp,
                                  sig_dram_udp)
            .else
                //Checksum is in 1st QW. Only write that back
                dl_put_quad_words_from_gpr_lm(LM_NATPT_WORK_BUF,
                                  udp_offset,
                                  1,
                                  sig_dram_udp,
                                  sig_dram_udp)
            .endif
        .end
    .endif
#else
    .if (length_pos == 7);checksum is in 2nd QW
      .begin
        alu[udp_offset, udp_offset, +, 8]

        xbuf_alloc($$xfer_udp_hdr, 2, write)
        xbuf_copy($$xfer_udp_hdr, 0, 0, LM_NATPT_WORK_BUF, 1, 0, 8, 0)

        dl_put_quad_words($$xfer_udp_hdr0,
                          udp_offset,
                          1,
                          sig_dram_udp,
                          sig_dram_udp)
        xbuf_free($$xfer_udp_hdr)
      .end
    .else
        .begin
            xbuf_alloc($$xfer_buf, 4, write)

            xbuf_extract($$xfer_buf0, LM_NATPT_WORK_BUF, 0, 0, 4)
            xbuf_extract($$xfer_buf1, LM_NATPT_WORK_BUF, 0, 4, 4)

            .if (length_pos == 6);checksum is in 8th byte of this QW & 1st byte of next
            //We need to write both QW back
                xbuf_extract($$xfer_buf2, LM_NATPT_WORK_BUF, 0, 8, 4)
                xbuf_extract($$xfer_buf3, LM_NATPT_WORK_BUF, 0, 12, 4)
                dl_put_quad_words($$xfer_buf0, udp_offset, 2, sig_dram_udp, sig_dram_udp)
            .else
                //Checksum is in 1st QW. Only write that back
                dl_put_quad_words($$xfer_buf0, udp_offset, 1, sig_dram_udp, sig_dram_udp)
            .endif

            xbuf_free($$xfer_buf)
        .end
    .endif
#endif //ENHANCED_XBUF_API
.end

#endm

////////////////////////////////////////////////////////////////////////////////
// _update_TCP_checksum
//
//    Description:
//      Internal macro that updates the checksum in a TCP PDU. The original 
//      checksum is updated by including the fields in 'in_add_list' while
//      the fields in 'in_del_list' are excluded.
//      Note that this macro updates (& possibly even reads) the packet directly
//      from DRAM.
//
//    Parameters:
//        Outputs:
//        Inputs:
//          - in_fragment_offset_and_flag:
//                                 Field containing the fragment offset & MF
//                                 flag from the IP header
//          - in_transport_pdu_len: Length of the transport packet/fragment
//          - in_checksum_offset_in_packet: Offset where the TCP checksum field
//                                 would begin (if it were present) in the
//                                 packet.
//          - in_add_list        : String containing a comma separated list of
//                                 the fields that need to be included in the
//                                 checksum
//          - in_del_list        : String containing a comma separated list of
//                                 the fields that need to be removed from the
//                                 checksum
//
//     Side effects:
//         The checksum field in the TCP PDU is updated directly in DRAM
////////////////////////////////////////////////////////////////////////////////
#macro _update_TCP_checksum(in_fragment_offset_and_flag,\
                            in_transport_pdu_len,\
                            in_checksum_offset_in_packet,\
                            in_add_list,\
                            in_del_list)
    //We will first confirm if the TCP checksum field is even present in
    //this packet. The checksum might not be present if the payload was
    //fragmented. In this case the current fragment might be holding the
    //latter part of the TCP packet. The other (rare) case is when the
    //fragment holds the inital part of the TCP packet but is so small
    //that the first 18 bytes of the TCP header are not present in it.
    //(The checksum is in bytes 17-18 of the TCP packet)
    //If the checksum is not present then there is nothing more to be done 
    //for this fragment.
    .if (in_fragment_offset_and_flag != 0)
        //Fragment offsets are in 8 byte octets. So checksum field can be
        //present in the fragment only if offset is less than 2
        alu_shf[in_fragment_offset_and_flag, --, B,
                in_fragment_offset_and_flag, >>3] ;get rid of flags

        .if (in_fragment_offset_and_flag > 3)
            br[LBL_PAST_TCP_CHECKSUM_UPDATE#]
        .else
            //Next we check to see if the fragment is long enough to be
            //holding the checksum
            #define_eval LENGTH_TO_HOLD_CHECKSUM TCP_CHECKSUM_OFFSET + 2

            alu_shf[result, --, B, in_fragment_offset_and_flag, <<3]
            alu_shf[result, result, +, in_transport_pdu_len]
            .if (result < LENGTH_TO_HOLD_CHECKSUM)
                br[LBL_PAST_TCP_CHECKSUM_UPDATE#]
            .endif
            #undef LENGTH_TO_HOLD_CHECKSUM
        .endif
    .endif

    //We come here if the checksum is present in this packet (or fragment)
    .begin
        .sig sig_dram_tcp
        .reg checksum_pos //Offset of checksum in the Quad Word read

        //First we need to read in the original checksum.
        //Get the offset of the transport header in the packet.
        //Read in the checksum from the appropriate offset (16) in the
        //TCP header.

        alu[checksum_pos, in_checksum_offset_in_packet, AND, 7]

        //The offset needs to be quadword aligned. Actually dl_get_bytes takes care
        //of this. But since we write back the modified value to DRAM and the
        //relevant bytes might already be in the cache, we need to do this.
        //(Otherwise when the bytes are picked up from the cache they will land up
        // at offset 0 in the returned buffer, while they will be at offset 
        // checksum_pos when they are picked up from DRAM)
        alu_shf[in_checksum_offset_in_packet, in_checksum_offset_in_packet, AND~, 7] ;mask off the lower 3 bits

        //We read in 8 or 16 bytes i.e. one or two quad words, because we need
        //to write back the modified checksum, possibly to DRAM (if the
        //checksum is not present in the header cache) which works in quad word
        //units.

#ifdef NO_CODESTORE_OVERFLOW
        .if (checksum_pos == 7)
            //The checksum straddles a Quadword boundary.
            //So we read in two Quadwords.
            dl_get_bytes(LM_NATPT_WORK_BUF,
                         in_checksum_offset_in_packet,
                         16,
                         sig_dram_tcp,
                         sig_dram_tcp)    
        .else
            dl_get_bytes(LM_NATPT_WORK_BUF,
                         in_checksum_offset_in_packet,
                         8,
                         sig_dram_tcp,
                         sig_dram_tcp)
        .endif
#else
        move(in_start_offset_in_header, in_checksum_offset_in_packet)

        .if (checksum_pos == 7)
            //The checksum straddles a Quadword boundary.
            //So we read in two Quadwords.
            move(in_bytes_to_read, 16)
        .else
            move(in_bytes_to_read, 8)
        .endif
        call(dl_get_bytes#)
#endif //NO_CODE_STORE_OVERFLOW

        //Extract the checksum from the buffer
        xbuf_extract(checksum, LM_NATPT_WORK_BUF, 0, checksum_pos, 2)

        _modify_checksum(checksum, 'BEGIN', --)

        //Subtract out the fields that were removed
        _repeat_per_elem_in_list(del_wrapper, in_del_list)

        //Add in the new fields 
        _repeat_per_elem_in_list(add_wrapper, in_add_list)

        _modify_checksum(checksum, 'END', --)

        //Write back the modified checksum
        xbuf_insert(LM_NATPT_WORK_BUF, checksum, 0, checksum_pos, 2)

/*
    The specific case in which we invoke the xbuf_copy macro in 
    dl_put_quad_words_from_gpr is not supported.
    If enhanced versions of the macros become available use the code
    in the #if part below. The code in the #else part is correct but
    will probably not be as efficient.
*/
#ifdef ENHANCED_XBUF_API
        .if (checksum_pos == 7)
            dl_put_quad_words_from_gpr_lm(LM_NATPT_WORK_BUF,
                    in_checksum_offset_in_packet,
                    2,
                    sig_dram_tcp,
                    sig_dram_tcp)
        .else
            dl_put_quad_words_from_gpr_lm(LM_NATPT_WORK_BUF,
                    in_checksum_offset_in_packet,
                    1,
                    sig_dram_tcp,
                    sig_dram_tcp)
        .endif
#else
        .if (checksum_pos == 7)
          .begin
            xbuf_alloc($$xfer_buf, 4, write)

            xbuf_extract($$xfer_buf0, LM_NATPT_WORK_BUF, 0, 0, 4)
            xbuf_extract($$xfer_buf1, LM_NATPT_WORK_BUF, 0, 4, 4)
            xbuf_extract($$xfer_buf2, LM_NATPT_WORK_BUF, 0, 8, 4)
            xbuf_extract($$xfer_buf3, LM_NATPT_WORK_BUF, 0, 12, 4)
            dl_put_quad_words($$xfer_buf0, in_checksum_offset_in_packet, 2, sig_dram_tcp, sig_dram_tcp)

            xbuf_free($$xfer_buf)
          .end
        .else
          .begin
            xbuf_alloc($$xfer_buf, 2, write)

            xbuf_extract($$xfer_buf0, LM_NATPT_WORK_BUF, 0, 0, 4)
            xbuf_extract($$xfer_buf1, LM_NATPT_WORK_BUF, 0, 4, 4)
            dl_put_quad_words($$xfer_buf0, in_checksum_offset_in_packet, 1, sig_dram_tcp, sig_dram_tcp)

            xbuf_free($$xfer_buf)
          .end
        .endif
#endif //ENHANCED_XBUF_API

    .end

LBL_PAST_TCP_CHECKSUM_UPDATE#:
#endm

//Macros/functions to modify checksum.
//We do an incremental update which essentially amounts to
//the arithmetic difference of the old values & the new
//values from the original checksum

////////////////////////////////////////////////////////////////////////////////
// _modify_checksum
//
//    Description:
//        Internal macro to perform an incremental update on a checksum.
//
//    Parameters:
//        Outputs:
//           - io_checksum : Contains the updated checksum
//        Inputs:
//           - io_checksum : Contains the checksum to modify
//           - in_operation: Contains the operation to perform on the checksum
//                           Can be one of 'BEGIN', 'END', '+' or '-'
//           - in_operand  : Based on the operations, contains the field to
//                           add or delete
//
//     Usage:
//                  When making a series of incremental updates to the
//                  checksum, it is assumed that the caller will start
//                  by passing the original value with the 'BEGIN' operation
//                  & subsequently use the value returned by this macro. 
//                  To include a field in the checksum this macro should be
//                  called with the '+' operation. To remove a field, call 
//                  this macro with the '-' operation. The checksum update is
//                  completed by a call to this macro with the 'END' operation
////////////////////////////////////////////////////////////////////////////////
#macro _modify_checksum(io_checksum, in_operation, in_operand)
    #if (streq(in_operation, 'BEGIN') || streq(in_operation, 'END'))
    .begin
        //Here's what we need to do
        //x=~x & 0xFFFF;
        .reg mask

        immed[mask, 0xffff, <<0]
        alu[io_checksum, io_checksum, ~AND, mask]
    .end
    #endif

    #if (streq(in_operation, '+'))
    .begin
        //x+=new & 0xffff;
        //if (x & 0x10000) { x++; x&=0xffff; }
        //Note that 'new' above is a 16 bit variable whereas we are passed
        //in a 32 bit register
        .reg top16

        alu[io_checksum, io_checksum, +16, in_operand]
        br_bclr[io_checksum, 16, LBL_NO_ROLLOVER_LO#], defer[1]
        alu[top16, --, B, in_operand, >>16]

        alu[io_checksum, 1, +16, io_checksum]

LBL_NO_ROLLOVER_LO#:
        alu[io_checksum, io_checksum, +16, top16]
        br_bclr[io_checksum, 16, LBL_NO_ROLLOVER#]
        alu[io_checksum, 1, +16, io_checksum]
LBL_NO_ROLLOVER#:
    .end
    #endif

    #if (streq(in_operation, '-'))
    .begin
        //Here's what we need to do
        //x-=old & 0xffff;
        //if (x<=0) { x--; x&=0xffff; }
        //Note that 'old' above is a 16 bit variable whereas we are passed
        //in a 32 bit register
        .reg mask
        .reg low16, hi16

        immed[mask, 0xffff, <<0]
        alu[low16, in_operand, AND, mask]
        alu[io_checksum, io_checksum, -, low16] ;x=A-B
        bgt[LBL_LO16_GREATER_THAN_0#],defer[1]  ;branch if A>B i.e x>0
        alu_shf[hi16, --, B, in_operand, >>16]

        ;x <= 0
        alu[io_checksum, io_checksum, -, 1]
        alu[io_checksum, io_checksum, AND, mask]

LBL_LO16_GREATER_THAN_0#:
        alu[io_checksum, io_checksum, -, hi16] ;x=A-B
        bgt[LBL_HI16_GREATER_THAN_0#]          ;branch if A>B i.e x>0
        alu[io_checksum, io_checksum, -, 1]
        alu[io_checksum, io_checksum, AND, mask]

LBL_HI16_GREATER_THAN_0#:
    .end
    #endif
#endm

//When making a series of incremental updates to the checksum, it is assumed
//that the caller will start by invoking '_modify_checksum_BEGIN_END'
//& subsequently use the value returned by this macro. 
//The checksum update is completed by a call to this same macro.

#macro _modify_checksum_BEGIN_END(io_checksum)
.begin
    //Here's what we need to do
    //x=~x & 0xFFFF;
    .reg mask

    immed[mask, 0xffff, <<0]
    alu[io_checksum, io_checksum, ~AND, mask]
.end
#endm

//The following 'functions' perform the same operations as the macros above
//The 'functions' below assume that the following registers will be defined
//checksum
//in_operand
//NOTE: For all the functions below, the registers used are context relative,
//      so there is no issue with multiple contexts being in the same function
//        at the same time.



br[LBL_PAST_FUNCS#] ;Just to get rid of warnings while assembling stand-alone

//x+=new & 0xffff;
//if (x & 0x10000) { x++; x&=0xffff; }
//Note that 'new' above is a 16 bit variable whereas we are passed
//in a 32 bit register
_modify_checksum_ADD#:
.begin
    .reg top16

    alu[checksum, checksum, +16, in_operand]
    br_bclr[checksum, 16, LBL_NO_ROLLOVER_LO#], defer[1]
    alu[top16, --, B, in_operand, >>16]

    alu[checksum, 1, +16, checksum]

LBL_NO_ROLLOVER_LO#:
    alu[checksum, checksum, +16, top16]
    br_bclr[checksum, 16, LBL_NO_ROLLOVER#]
    alu[checksum, 1, +16, checksum]
LBL_NO_ROLLOVER#:

    rtn[ret_reg]
.end

//x-=old & 0xffff;
//if (x<=0) { x--; x&=0xffff; }
//Note that 'old' above is a 16 bit variable whereas we are passed
//in a 32 bit register
_modify_checksum_SUB#:
.begin
    .reg mask
    .reg low16, hi16

    immed[mask, 0xffff, <<0]
    alu[low16, in_operand, AND, mask]
    alu[checksum, checksum, -, low16] ;x=A-B
    bgt[LBL_LO16_GREATER_THAN_0#],defer[1]  ;branch if A>B i.e x>0
    alu_shf[hi16, --, B, in_operand, >>16]

    ;x <= 0
    alu[checksum, checksum, -, 1]
    alu[checksum, checksum, AND, mask]

LBL_LO16_GREATER_THAN_0#:
    alu[checksum, checksum, -, hi16] ;x=A-B
    bgt[LBL_HI16_GREATER_THAN_0#]          ;branch if A>B i.e x>0
    alu[checksum, checksum, -, 1]
    alu[checksum, checksum, AND, mask]

LBL_HI16_GREATER_THAN_0#:
    rtn[ret_reg]
.end
LBL_PAST_FUNCS#:

////////////////////////////////////////////////////////////////////////////////
// _generate_checksum
//
//    Description:
//        Generates checksum on the specified data
//
//    Parameters:
//        Outputs:
//              Assumes checksum has to be returned in a register named
//              'checksum', which is defined by the caller. This will save us
//              2 instructions.
//  
//        Inputs:
//           - in_start_offset: Offset in packet where the data to be
//                  included in the checksum generation begins
//           - in_data_length : Length of the data to include in the checksum.
//                  Essentially data from 'in_start_offset' to 
//                  'in_start_offset' + 'in_data_length' is included in the
//                  checksum
////////////////////////////////////////////////////////////////////////////////
#ifdef NO_CODESTORE_OVERFLOW
//Note: This macro has been ifdef'd out for lack of space in the code store.
//      We leave the code here as a reference in case this microblock is
//      chained in another application where there is space available in the
//      code store.
#macro _generate_checksum(in_start_offset, in_data_length)
.begin
//Note: The buffer below has been arbitrarily picked. This will have to be set
//      appropriately based on where this macro is actually used.
#define data_buf lmem_buf14
    //Div by 4 to get number of registers
    #define_eval NUM_OVERFLOW_REG NAT_DRAM_READ_BYTES>>2
    xbuf_alloc(data_buf, NUM_OVERFLOW_REG, read_write)
    #undef NUM_OVERFLOW_REG

    .reg bytes_remaining
    .reg cur_offset
    .reg cur_data, padded_data
    .reg start_offset
    .reg rem
    .sig sig_dram_read

    xbuf_bind_address(data_buf, 64, NAT_DRAM_READ_BYTES, 0)
    xbuf_activate(data_buf, 1, context_id, 0)

    alu[checksum, --, B, 0] ;initialize checksum
    alu[bytes_remaining, --, B, in_data_length]
    alu[start_offset, --, B, in_start_offset]

LBL_READ_DATA#:
    alu[cur_offset, --, B, 0]
    dl_get_bytes(data_buf, start_offset, NAT_DRAM_READ_BYTES,
                    sig_dram_read, sig_dram_read)

    //work with data_buf. Note: If bytes_remaining < NAT_DRAM_READ_BYTES
    //only use bytes_remaining bytes from the buffer

LBL_GEN_CS#:
    //It is assumed that NAT_DRAM_READ_BYTES is a multiple of 4
    .if ((cur_offset < NAT_DRAM_READ_BYTES)&&(cur_offset < bytes_remaining))
        xbuf_extract(cur_data, data_buf, 0, cur_offset, 4)
        alu[rem, bytes_remaining, -, cur_offset]
        .if (rem < 4)
            //If less than 4 bytes remaining, data_buf[cur_offset] holds some
            //bytes that we are not interested in. We pad the register with 0
            //Note: rem can never equal 0. We break out of the loop in this
            //case
            //Get rid of warning
            .set padded_data
            .if (rem == 1)
                ld_field_w_clr[padded_data, 1000, cur_data]
            .elif (rem == 2)
                ld_field_w_clr[padded_data, 1100, cur_data]
            .elif (rem == 3)
                ld_field_w_clr[padded_data, 1110, cur_data]
            .endif

            #ifdef NO_CODESTORE_OVERFLOW
                   _modify_checksum(out_checksum, '+', padded_data)
            #else
                   alu[in_operand, --, B, padded_data]
                   call(_modify_checksum_ADD#)
            #endif //NO_CODESTORE_OVERFLOW
        .else
            #ifdef NO_CODESTORE_OVERFLOW
                _modify_checksum(out_checksum, '+', cur_data)
            #else
                alu[in_operand, --, B, cur_data]
                call(_modify_checksum_ADD#)
            #endif //NO_CODESTORE_OVERFLOW
        .endif

        alu[cur_offset, cur_offset, +, 4]
        br[LBL_GEN_CS#]
    .endif

    alu[start_offset, start_offset, +, NAT_DRAM_READ_BYTES]
    alu[bytes_remaining, bytes_remaining, -, NAT_DRAM_READ_BYTES]
    bgt[LBL_READ_DATA#]

    //Final checksum is a 16-bit complement of the sum
    _modify_checksum_BEGIN_END(checksum)

    xbuf_deactivate(data_buf)
    xbuf_free(data_buf)
    #undef data_buf
.end
#endm
#endif //NO_CODESTORE_OVERFLOW

#endif // __V6V4_NATPT_UC__
