/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.x for the IXP2xxx Network Processors
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      NATPT Core Component - Tables for microblock
 *
 * = FILENAME
 *      ix_cc_natpt_nat_table.h
 *
 * = DESCRIPTION
 *      This file defines all data types, macros and function prototypes
 *      that support management of the NATPT tables that are shared by
 *      both CC and microblock.
 *
 * = CREATION TIME
 *      03/03/03
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 * $Id: ix_cc_natpt_nat_table.h,v 1.14 2003/04/30 01:06:33 sshegde Exp $
 */

#if !defined(__IX_CC_NATPT_NAT_TABLE_H__)
#define __IX_CC_NATPT_NAT_TABLE_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */


/* Masks for extracting bits in the ix_cc_natpt_nat_entry strcture */
#define IX_CC_NATPT_NAT_MASK_ENTRY_USED     0x8000
#define IX_CC_NATPT_NAT_MASK_ENTRY_STATIC   0x4000

/* Special value in the Nat Hash Table that is not used */
#define IX_CC_NATPT_UNUSED_ENTRY_FILL       0xffffffff


/**
 * TYPENAME: ix_cc_natpt_nat_entry
 * 
 * DESCRIPTION: This structure defines the format for an entry in the end
 * tunnel table in shared memory.
 *
 */
typedef struct ix_s_cc_natpt_nat_entry
{
    ix_uint32   v4Addr;
    ix_uint128  v6Addr;

#if IX_BYTE_ORDER == IX_LITTLE_ENDIAN 
    ix_uint16 unused:14;
    ix_uint16 isStatic:1;
    ix_uint16 isUsed:1;
#elif IX_BYTE_ORDER == IX_BIG_ENDIAN
    ix_uint16 isUsed:1;
    ix_uint16 isStatic:1;
    ix_uint16 unused:14;
#endif

    ix_uint16   v6NHID;
    ix_uint64   reserved;
} __attribute__ ((packed)) ix_cc_natpt_nat_entry;


/**
 * TYPENAME: ix_cc_natpt_nat_hash_entry
 * 
 * DESCRIPTION: This structure defines the format for an entry in the end
 * tunnel table in shared memory.
 *
 */
typedef struct ix_s_cc_natpt_nat_hash_entry
{
    ix_uint32   index; /* index into NAT Table */
    ix_uint32   next;  /* index into Collision Hash Table */
} __attribute__ ((packed)) ix_cc_natpt_nat_hash_entry;


/*
 * NAME: _READ_NAT_TABLE_ENTRY()
 *
 * DESCRIPTION: Reads a NAT table entry from shared memory and
 * places it into local memory.
 *
 * @Param:      - IN    __base
 *                      Base address of NAT entry table.
 *
 * @Param:      - IN    __index
 *                      Index of NAT entry to be read.
 *
 * @Param:      - OUT   __local
 *                      Local structure where the data will be placed.
 *                      
 */
#define _READ_NAT_TABLE_ENTRY(__base, __index, __local)                     \
    *(ix_uint64*) &(__local) = IX_RM_MEM_UINT64_READ(                       \
        (ix_uint64*) &(((ix_cc_natpt_nat_entry*)(__base))[__index]));       \
    *(((ix_uint64*) &(__local))+1) = IX_RM_MEM_UINT64_READ(                 \
        (ix_uint64*) &(((ix_cc_natpt_nat_entry*)(__base))[__index])+1);     \
    *(((ix_uint64*) &(__local))+2) = IX_RM_MEM_UINT64_READ(                 \
        (ix_uint64*) &(((ix_cc_natpt_nat_entry*)(__base))[__index])+2);


/*
 * NAME: _READ_NAT_TABLE_ENTRY_POINTER()
 *
 * DESCRIPTION: Reads a NAT table entry from shared memory and
 * places it into local memory.
 *
 * @Param:      - IN    __ptr
 *                      Pointer to NAT entry to be read.
 *
 * @Param:      - OUT   __local
 *                      Local structure where the data will be placed.
 *                      
 */
#define _READ_NAT_TABLE_ENTRY_POINTER(__ptr, __local)                       \
    *(ix_uint64*) &(__local) = IX_RM_MEM_UINT64_READ(                       \
        (ix_uint64*) (__ptr));                                              \
    *(((ix_uint64*) &(__local))+1) = IX_RM_MEM_UINT64_READ(                 \
        (ix_uint64*) (__ptr) + 1);                                          \
    *(((ix_uint64*) &(__local))+2) = IX_RM_MEM_UINT64_READ(                 \
        (ix_uint64*) (__ptr) + 2);


/*
 * NAME: _WRITE_NAT_TABLE_ENTRY()
 *
 * DESCRIPTION: Writes a NAT table entry from shared memory and
 * places it into local memory.
 *
 * @Param:      - IN    __base
 *                      Base address of NAT entry table.
 *
 * @Param:      - IN    __index
 *                      Index of NAT entry to be written.
 *
 * @Param:      - IN    __local
 *                      Local structure to be copied into shared memory.
 *
 */
#define _WRITE_NAT_TABLE_ENTRY(__base, __index, __local)                    \
    (void) IX_RM_MEM_UINT64_WRITE(                                          \
        (ix_uint64*) &(((ix_cc_natpt_nat_entry*)(__base))[__index]),        \
        *(ix_uint64*) &(__local));                                          \
    (void) IX_RM_MEM_UINT64_WRITE(                                          \
        (ix_uint64*) &(((ix_cc_natpt_nat_entry*)(__base))[__index]) + 1,    \
        *((ix_uint64*) &(__local)+1));                                      \
    (void) IX_RM_MEM_UINT64_WRITE(                                          \
        (ix_uint64*) &(((ix_cc_natpt_nat_entry*)(__base))[__index]) + 2,    \
        *((ix_uint64*) &(__local)+2));                                      

/*
 * NAME: _WRITE_NAT_TABLE_RESET_USED()
 *
 * DESCRIPTION: Writes a NAT table entry from shared memory and
 * places it into local memory.
 *
 * @Param:      - IN    __base
 *                      Base address of NAT entry table.
 *
 * @Param:      - IN    __index
 *                      Index of NAT entry to be written.
 *
 * @Param:      - IN    __local
 *                      Local structure to be copied into shared memory.
 *
 */
#define _WRITE_NAT_TABLE_RESET_USED(__base, __index, __local)               \
    (void) IX_RM_MEM_UINT8_WRITE(                                           \
        (ix_uint8*) &(((ix_cc_natpt_nat_entry*)(__base))[__index]) + 20,    \
        *((ix_uint8*) &(__local) + 20));

/*
 * NAME: _CLEAR_NAT_TABLE_ENTRY()
 *
 * DESCRIPTION: Clears a NAT table entry in shared memory.
 *
 * @Param:      - IN    __base
 *                      Base address of NAT entry table.
 *
 * @Param:      - IN    __index
 *                      Index of NAT entry to be read.
 *
 */
#define _CLEAR_NAT_TABLE_ENTRY(__base, __index)                             \
    (void) IX_RM_MEM_UINT64_WRITE(                                          \
        (ix_uint64*) &(((ix_cc_natpt_nat_entry*)(__base))[__index]),        \
        0);                                                                 \
    (void) IX_RM_MEM_UINT64_WRITE(                                          \
        (ix_uint64*) &(((ix_cc_natpt_nat_entry*)(__base))[__index]) + 1,    \
        0);                                                                 \
    (void) IX_RM_MEM_UINT64_WRITE(                                          \
        (ix_uint64*) &(((ix_cc_natpt_nat_entry*)(__base))[__index]) + 2,    \
        0);


/*
 * NAME: _NAT_TABLE_INDEX2POINTER()
 *
 * DESCRIPTION: Writes a NAT table entry from shared memory and
 * places it into local memory.
 *
 * @Param:      - IN    __base
 *                      Base address of NAT entry table.
 *
 * @Param:      - IN    __index
 *                      Index of NAT entry.
 *
 */
#define _NAT_TABLE_INDEX2POINTER(__base, __index)                    \
    (ix_uint32) &(((ix_cc_natpt_nat_entry*)(__base))[__index]);


/*
 * NAME: _READ_NAT_HASH_ENTRY()
 *
 * DESCRIPTION: Reads a NAT Hash table entry from shared memory and
 * places it into local memory.
 *
 * @Param:      - IN    __base
 *                      Base address of NAT Hash entry table.
 *
 * @Param:      - IN    __index
 *                      Index of NAT Hash entry to be read.
 *
 * @Param:      - OUT   __local
 *                      Local structure where the data will be placed.
 *                      
 */
#define _READ_NAT_HASH_ENTRY(__base, __index, __local)                      \
    *(ix_uint64*) &(__local) = IX_RM_MEM_UINT64_READ(                       \
        (ix_uint64*) &(((ix_cc_natpt_nat_hash_entry*)(__base))[__index]));


/*
 * NAME: _READ_NAT_HASH_ENTRY_POINTER()
 *
 * DESCRIPTION: Reads a NAT Hash table entry from shared memory and
 * places it into local memory.
 *
 * @Param:      - IN    __ptr
 *                      Pointer of NAT Hash entry to be read.
 *
 * @Param:      - OUT   __local
 *                      Local structure where the data will be placed.
 *                      
 */
#define _READ_NAT_HASH_ENTRY_POINTER(__ptr, __local)              \
    *(ix_uint64*) &(__local) = IX_RM_MEM_UINT64_READ((ix_uint64*)(__ptr));


/*
 * NAME: _WRITE_NAT_HASH_ENTRY()
 *
 * DESCRIPTION: Writes a NAT Hash table entry to shared memory from 
 * local memory.
 *
 * @Param:      - IN    __base
 *                      Base address of NAT Hash entry table.
 *
 * @Param:      - IN    __index
 *                      Index of NAT Hash entry to be written.
 *
 * @Param:      - IN    __local
 *                      Local structure to be copied into shared memory.
 *
 */
#define _WRITE_NAT_HASH_ENTRY(__base, __index, __local)                     \
    (void) IX_RM_MEM_UINT64_WRITE(                                          \
        (ix_uint64*) &(((ix_cc_natpt_nat_hash_entry*)(__base))[__index]),   \
        *(ix_uint64*) &(__local));


/*
 * NAME: _WRITE_NAT_HASH_ENTRY_POINTER()
 *
 * DESCRIPTION: Writes a NAT Hash table entry to shared memory from
 * local memory.
 *
 * @Param:      - IN    __ptr
 *                      Pointer where NAT Hash entry is to be written.
 *
 * @Param:      - IN    __local
 *                      Local structure to be copied into shared memory.
 *
 */
#define _WRITE_NAT_HASH_ENTRY_POINTER(__ptr, __local)                       \
    (void) IX_RM_MEM_UINT64_WRITE(                                          \
            (ix_uint64*)(__ptr), *(ix_uint64*) &(__local));


/*
 * NAME: _CLEAR_NAT_HASH_ENTRY()
 *
 * DESCRIPTION: Clears a NAT table entry in shared memory.
 *
 * @Param:      - IN    __base
 *                      Base address of NAT Hash entry table.
 *
 * @Param:      - IN    __index
 *                      Index of NAT Hash entry to be cleared.
 *
 */
#define _CLEAR_NAT_HASH_ENTRY(__base, __index)                              \
    (void) IX_RM_MEM_UINT32_WRITE(                                          \
        (ix_uint32*) &(((ix_cc_natpt_nat_hash_entry*)(__base))[__index]),   \
        IX_CC_NATPT_UNUSED_ENTRY_FILL);                                                        \
    (void) IX_RM_MEM_UINT32_WRITE(                                          \
        (ix_uint32*) &(((ix_cc_natpt_nat_hash_entry*)(__base))[__index]) + 1,\
        IX_CC_NATPT_UNUSED_ENTRY_FILL);                                                        \


/*
 * NAME: _CLEAR_NAT_HASH_ENTRY_POINTER()
 *
 * DESCRIPTION: Clears a NAT table entry in shared memory.
 *
 * @Param:      - IN    __ptr
 *                      Pointer to NAT Hash entry to be cleared.
 *
 */
#define _CLEAR_NAT_HASH_ENTRY_POINTER(__ptr)                                \
    (void) IX_RM_MEM_UINT64_WRITE(                                          \
        (ix_uint64*) __ptr, 0);


/*
 * NAME: _NAT_HASH_TABLE_INDEX2POINTER()
 *
 * DESCRIPTION: Converts NAT hash table index to pointer
 *
 * @Param:      - IN    __base
 *                      Base address of NAT hash entry table.
 *
 * @Param:      - IN    __index
 *                      Index of NAT hash entry.
 *
 */
#define _NAT_HASH_TABLE_INDEX2POINTER(__base, __index)                    \
    (ix_uint32) &(((ix_cc_natpt_nat_hash_entry*)(__base))[__index]);


/*
 * NAME: _NAT_HASH_TABLE_POINTER2INDEX()
 *
 * DESCRIPTION: Converts NAT hash table pointer to index
 *
 * @Param:      - IN    __base
 *                      Base address of NAT hash entry table.
 *
 * @Param:      - IN    __ptr
 *                      Pointer to NAT hash entry.
 *
 */
#define _NAT_HASH_TABLE_POINTER2INDEX(__base, __ptr)                    \
    (ix_uint32)((((ix_uint8 *)__ptr) - ((ix_uint8*)__base)) /           \
    (sizeof(ix_cc_natpt_nat_hash_entry)));



/**
 * NAME: _ix_cc_natpt_add_nat_hash_table_entries
 *
 * DESCRIPTION: This function is used to add a NAT Hash Entry into
 * both v6-to-v4 and v4-to-v6 Hash Tables .
 *           
 * @Param:  - IN arg_pContext - pointer to NATPT Context
 * @Param:  - IN arg_natEntry - pointer to NAT Table Entry
 * @Param:  - IN arg_index - Index where NAT Table Entry was added
 *
 * @Param:  - OUT None
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_FULL - Out of Memory
 * 
 */
extern ix_error _ix_cc_natpt_add_nat_table_entry(
                                ix_uint32* arg_tableBase,
                                ix_cc_index_pool arg_indexPool,
                                ix_cc_natpt_nat_entry* arg_natEntry,
                                ix_uint32* arg_index);

/**
 * NAME: _ix_cc_natpt_del_nat_hash_table_entries
 *
 * DESCRIPTION: This function is used to remove a NAT Hash Entry in
 * both v6-to-v4 and v4-to-v6 Hash Tables .
 *           
 * @Param:  - IN arg_pContext - pointer to NATPT Context
 * @Param:  - IN arg_natEntry - pointer to NAT Table Entry
 * @Param:  - IN arg_index - Index where NAT Table Entry was added
 *
 * @Param:  - OUT None
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_ENTRY_NOT_FOUND - Entry Not found.
 * 
 */
extern ix_error _ix_cc_natpt_del_nat_table_entry(
                                ix_uint32* arg_tableBase,
                                ix_cc_index_pool arg_indexPool,
                                ix_cc_natpt_nat_entry* arg_natEntry);

/**
 * NAME: _ix_cc_natpt_find_nat_mapping_v6tov4
 *
 * DESCRIPTION: This function is used to get IPv4 address mapping for IPv6
 * address.
 *           
 * @Param:  - IN arg_pContext - pointer to NATPT Context
 *
 * @Param:  - IN/OUT arg_natEntry - pointer to NAT Table Entry. Contains
 * only IPv6 address on input, IPv4 address is filled in on output
 *
 * @Param:  - OUT None
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_ENTRY_NOT_FOUND - Entry Not found.
 * 
 */
extern ix_error _ix_cc_natpt_find_nat_table_entry(
                                ix_uint32* arg_tableBase,
                                ix_cc_index_pool arg_indexPool,
                                ix_cc_natpt_nat_entry* arg_natEntry,
                                ix_uint32* arg_index);

/**
 * NAME: _ix_cc_natpt_add_nat_hash_table_entries
 *
 * DESCRIPTION: This function is used to add a NAT Hash Entry into
 * both v6-to-v4 and v4-to-v6 Hash Tables .
 *           
 * @Param:  - IN arg_pContext - pointer to NATPT Context
 * @Param:  - IN arg_natEntry - pointer to NAT Table Entry
 * @Param:  - IN arg_index - Index where NAT Table Entry was added
 *
 * @Param:  - OUT None
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_FULL - Out of Memory
 * 
 */
extern ix_error _ix_cc_natpt_add_nat_hash_table_entries(
                                ix_cc_natpt_context* arg_pContext,
                                ix_cc_natpt_nat_entry* arg_natEntry,
                                ix_uint32 index);

/**
 * NAME: _ix_cc_natpt_del_nat_hash_table_entries
 *
 * DESCRIPTION: This function is used to remove a NAT Hash Entry in
 * both v6-to-v4 and v4-to-v6 Hash Tables .
 *           
 * @Param:  - IN arg_pContext - pointer to NATPT Context
 * @Param:  - IN arg_natEntry - pointer to NAT Table Entry
 * @Param:  - IN arg_index - Index where NAT Table Entry was added
 *
 * @Param:  - OUT None
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_ENTRY_NOT_FOUND - Entry Not found.
 * 
 */
extern ix_error _ix_cc_natpt_del_nat_hash_table_entries(
                                ix_cc_natpt_context* arg_pContext,
                                ix_cc_natpt_nat_entry* arg_natEntry,
                                ix_uint32 index);

/**
 * NAME: _ix_cc_natpt_find_nat_mapping_v6tov4
 *
 * DESCRIPTION: This function is used to get IPv4 address mapping for IPv6
 * address.
 *           
 * @Param:  - IN arg_pContext - pointer to NATPT Context
 *
 * @Param:  - IN/OUT arg_natEntry - pointer to NAT Table Entry. Contains
 * only IPv6 address on input, IPv4 address is filled in on output
 *
 * @Param:  - OUT None
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_ENTRY_NOT_FOUND - Entry Not found.
 * 
 */
extern ix_error _ix_cc_natpt_find_nat_mapping_v6tov4(
                                ix_cc_natpt_context* arg_pContext,
                                ix_cc_natpt_nat_entry* arg_natEntry);

/**
 * NAME: _ix_cc_natpt_find_nat_mapping_v4tov6
 *
 * DESCRIPTION: This function is used to get IPv6 address mapping for IPv4
 * address.
 *           
 * @Param:  - IN arg_pContext - pointer to NATPT Context
 *
 * @Param:  - IN/OUT arg_natEntry - pointer to NAT Table Entry. Contains
 * only IPv4 address on input, IPv6 address is filled in on output
 *
 * @Param:  - OUT None
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_ENTRY_NOT_FOUND - Entry Not found.
 * 
 */
extern ix_error _ix_cc_natpt_find_nat_mapping_v4tov6(
                                ix_cc_natpt_context* arg_pContext,
                                ix_cc_natpt_nat_entry* arg_natEntry);



/**
 * NAME: _ix_cc_natpt_set_entry_timeout_event_handler
 *
 * DESCRIPTION: This function is used to create a periodic event to check
 * for expiring dynamic mappings. It is also used when time interval
 * is changed.
 *           
 * @Param:  - IN arg_pContext - pointer to NATPT Context
 *
 * @Param:  - OUT None
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_NATPT_ERROR_CCI - Adding/Updating event failed
 * 
 */
extern ix_error _ix_cc_natpt_set_entry_timeout_event_handler(
                        ix_cc_natpt_context* arg_pContext);

#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__IX_CC_NATPT_NAT_TABLE_H__) */
