/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.x for the IXP2xxx Network Processors
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      IPv6 NATPT Core Component - Definitions for Translation.
 *
 * = FILENAME
 *      ix_cc_natpt_translate.h
 *
 * = DESCRIPTION
 *      This file defines all data types, macros and function prototypes
 *      that support the handling of packets requiring translation.
 *
 * = CREATION TIME
 *      03/03/03
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 * $Id: ix_cc_natpt_translate.h,v 1.8 2003/05/10 00:37:06 sshegde Exp $
 */

#if !defined(__IX_CC_NATPT_TRANSLATE_H__)
#define __IX_CC_NATPT_TRANSLATE_H__

#include "cc/natpt/internal/ix_cc_natpt_header_defs.h"
#include "cc/natpt/internal/ix_cc_natpt_nat_table.h"

#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */


#define IX_CC_NATPT_IPV6_TRCLS           0
#define IX_CC_NATPT_IPV6_DST_ADDR        24

#define IX_CC_NATPT_IPV4_VERSION         4
#define IX_CC_NATPT_IPV4_IHL             (IX_CC_NATPT_MIN_IPV4_LEN >> 2)

#define IX_CC_NATPT_IPV4_VERS_IHL        \
        ((IX_CC_NATPT_IPV4_VERSION << 4) | IX_CC_NATPT_IPV4_IHL)

#define IX_CC_NATPT_HDRTYPE_IPV4  0x01
#define IX_CC_NATPT_HDRTYPE_IPV6  0x02


/**
 * NAME: _ix_cc_natpt_handle_ipv6_packet
 *
 * DESCRIPTION: This primitive is used to process IPv6 packets that
 * require translation.
 *           
 * @Param:  - IN arg_hPacket - handle to buffer containing the packet.
 *
 * @Param:  - IN arg_pContext - pointer to NATPT core component
 * context buffer.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * 
 */

ix_error _ix_cc_natpt_handle_ipv6_packet(
    ix_buffer_handle arg_hPacket,  
    void* arg_pContext
    );


/**
 * NAME: _ix_cc_natpt_handle_ipv4_packet
 *
 * DESCRIPTION: This primitive is used to process IPv4 packets that
 * require translation.
 *           
 * @Param:  - IN arg_hPacket - handle to buffer containing the packet.
 *
 * @Param:  - IN arg_pContext - pointer to NATPT core component
 * context buffer.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * 
 */

ix_error _ix_cc_natpt_handle_ipv4_packet(
    ix_buffer_handle arg_hPacket,  
    void* arg_pContext
    );


/**
 * NAME: _ix_cc_natpt_handle_udp_ipv4_packet
 *
 * DESCRIPTION: This primitive is used to process IPv4 UDP packets that
 * cannot be translated by the microblock because UDP checksum field
 * is zero. 
 *           
 * @Param:  - IN arg_hPacket - handle to buffer containing the packet.
 *
 * @Param:  - IN arg_pContext - pointer to NATPT core component
 * context buffer.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * 
 */

ix_error _ix_cc_natpt_handle_udp_ipv4_packet(
    ix_buffer_handle arg_hPacket,  
    void* arg_pContext
    );


/**
 * NAME: _ix_cc_natpt_get_ipv6_to_ipv4_mapping
 *
 * DESCRIPTION: This primitive is used to get IPv4 address for the IPv6
 * address. If there mapping does not exist, it is created.
 *           
 * @Param:  - IN arg_ipv6Hdr - pointer to IPv6 header.
 * @Param:  - IN arg_pContext - pointer to NATPT core component
 * context buffer.
 *
 * @Param:  - IN/OUT arg_natEntry - On input contains IPv6 address, IPv4
 * address is filled in on output.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * 
 */

ix_error _ix_cc_natpt_get_ipv6_to_ipv4_mapping(
            ix_cc_natpt_ipv6_header* arg_ipv6Hdr,
            ix_cc_natpt_nat_entry* arg_natEntry,
            ix_cc_natpt_context* arg_pContext);

/**
 * NAME: _ix_cc_natpt_get_ipv4_to_ipv6_mapping
 *
 * DESCRIPTION: This primitive is used to get IPv6 address for the IPv4
 * address.
 *           
 * @Param:  - IN arg_ipv4Hdr - pointer to IPv4 header.
 * @Param:  - IN arg_pContext - pointer to NATPT core component
 * context buffer.
 *
 * @Param:  - IN/OUT arg_natEntry - On input contains IPv4 address, IPv6
 * address is filled in on output.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * 
 */

ix_error _ix_cc_natpt_get_ipv4_to_ipv6_mapping(
            ix_cc_natpt_ipv4_header* arg_ipv4Hdr,
            ix_cc_natpt_nat_entry* arg_natEntry,
            ix_cc_natpt_context* arg_pContext);


/**
 * NAME: checksumadjust
 *
 * DESCRIPTION: 
 *      Checkusm update routing taken from RFC 1631
 *
 * @Param:  - IN unsigned char is 8 bits, long is 32 bits.
 * @Param:  - IN chksum points to the chksum in the packet
 * @Param:  - IN optr points to the old data in the packet
 * @Param:  - IN nptr points to the new data in the packet
 *
 * @Return: None.
 *
 */

void checksumadjust(
            unsigned char *chksum, 
            unsigned char *optr,
            int olen, 
            unsigned char *nptr, 
            int nlen);


/**
 * NAME: _ix_cc_natpt_validate_extension_headers
 *
 * DESCRIPTION:  This function walks through the extension header chain
 * and looks for unsupported ones. 
 *           
 * @Param:  - IN arg_pData - pointer to packet
 * @Param:  - IN arg_ipv6Hdr - pointer to IPv6 Header
 *
 * @Return: IX_SUCCESS if no unsupported extn hdrs found otherwise error
 *
 * Error Codes:
 *      IX_CC_NATPT_ERROR_NOT_IMPLEMENTED
 * 
 */

ix_error _ix_cc_natpt_validate_extension_headers(
            ix_uint8* arg_pData,
            ix_cc_natpt_ipv6_header* arg_ipv6Hdr);


#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__IX_CC_NATPT_TRANSLATE_H__) */
