/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.x for the IXP2xxx Network Processors
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      IPv6 NATPT Core Component - ICMP message translation
 *
 * = FILENAME
 *      ix_cc_natpt_icmp.c
 *
 * = DESCRIPTION
 *      The file defines functions for translation of ICMP messages that 
 *      are given by the microblock.
 *
 * = CREATION TIME
 *      03/03/03
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */

#define IX_ERROR_FILE_IDENT "$Id: ix_cc_natpt_icmp.c,v 1.24 2003/10/31 19:28:17 ktseng Exp $";


/**
 * System defined include files
 */
#include "ix_cc_error.h"
#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_cc.h"
#include "bindings.h"
#include "cc/ix_cc_natpt.h"


/**
 * User defined include files
 */
#include "cc/natpt/internal/ix_cc_natpt_init.h"
#include "cc/natpt/internal/ix_cc_natpt_lists.h"
#include "cc/natpt/internal/ix_cc_natpt_nat_table.h"
#include "cc/natpt/internal/ix_cc_natpt_utils.h"
#include "cc/natpt/internal/ix_cc_natpt_translate.h"
#include "cc/natpt/internal/ix_cc_natpt_icmp.h"


/**
 * Preprocessor symbols and Macros used in this file
 */


/**
 * Type definitions whose scope is limited to this file
 */


/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */


/** 
 * Extern function prototypes.
 */


/** 
 * Static function prototypes.
 */

PRIVATE ix_error _ix_cc_natpt_translate_icmp_header_v6tov4(
            ix_cc_natpt_icmp_header* arg_pIcmpV6,
            ix_cc_natpt_icmp_header* arg_pIcmpV4);
PRIVATE ix_error _ix_cc_natpt_translate_icmp_header_v4tov6(
            ix_cc_natpt_icmp_header* arg_pIcmpV4,
            ix_cc_natpt_icmp_header* arg_pIcmpV6);
PRIVATE void _ix_cc_natpt_icmp_v6_calculate_checksum(ix_uint8 *icmpMessage);
PRIVATE void _ix_cc_natpt_icmp_v4_calculate_checksum(ix_uint8 *icmpMessage);


/** 
 * Function definitions.
 */

/**
 * NAME: _ix_cc_natpt_icmp_v6_packet
 *
 * DESCRIPTION: This primitive is used to translate an ICMPv6 packet to ICMPv4
 * packet. Then the packet is given to IPv4 Forwarder.
 *           
 * @Param:  - IN arg_hPacket - handle to buffer containing the packet.
 *
 * @Param:  - IN arg_pContext - pointer to tunneling core component
 * context buffer.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * IX_CC_NATPT_ERROR_ICMP_TRANSLATE - could not translate
 * 
 */
ix_error _ix_cc_natpt_icmp_v6_packet(
    ix_buffer_handle arg_hPacket,  
    void* arg_pContext)
{
    ix_uint8* pData;
    ix_cc_natpt_ipv4_header ipv4Hdr;
    ix_cc_natpt_ipv6_header ipv6Hdr;
    ix_cc_natpt_icmp_header icmpV6Hdr;
    ix_cc_natpt_icmp_header icmpV4Hdr;
    ix_cc_natpt_ipv4_header errV4Hdr;
    ix_cc_natpt_ipv6_header errV6Hdr;
    ix_cc_natpt_ipv6_extn_hdr_common* extHdr;
    ix_hw_buffer_meta *pHwMeta;
    ix_cc_natpt_context* pContext;
    ix_uint16 packetSize;
    ix_uint32 csum = 0;
    ix_uint32 isEOP = 0;
    ix_uint16* pHdr;
    ix_cc_natpt_nat_entry natEntry;
    ix_uint16 extLength;
    ix_uint16 bufSize;
    ix_uint16 offset;
    ix_uint16 errSize = 0;
    ix_uint16 tmpSize;
    ix_uint16 bufInfo;
    ix_uint32 cellCount;
    int i;
    ix_buffer_handle unlinkedBuffer = IX_NULL_BUFFER_HANDLE;  

    pContext = (ix_cc_natpt_context*)arg_pContext;

    /* Get the packet meta-data */
    IX_ERROR_CRT(
        ix_rm_buffer_get_meta(arg_hPacket, (void **)&pHwMeta),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    /* read meta data sizes */
    packetSize = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_PacketSize));
    bufSize = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_BufferSize));
    offset = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_Offset));

    /* Get packet data */
    IX_ERROR_CRT(
        ix_rm_buffer_get_data(arg_hPacket, (void **)&pData),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    pData += IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_Offset));

    ix_ossl_memcpy(&ipv6Hdr, pData, IX_CC_NATPT_IPV6_HDR_LEN);

#if defined(NATPT_DEBUG)
    ix_ossl_message_log("Received Packet Src Address\n");
    _ix_cc_natpt_print_ipv6_address(&ipv6Hdr.sourceAddress);
    ix_ossl_message_log("Received Packet Dest Address\n");
    _ix_cc_natpt_print_ipv6_address(&ipv6Hdr.destAddress);
#endif /* NATPT_DEBUG */

    /* Get the mapping IPv4 address */
    IX_ERROR_CRT(
        _ix_cc_natpt_get_ipv6_to_ipv4_mapping(&ipv6Hdr, &natEntry, pContext),
        IX_CC_ERROR_OOR, IX_ERROR_LEVEL_WARNING);

#if defined(NATPT_DEBUG)
    ix_ossl_message_log("Mapping gave IPv4 address (%lx)\n", natEntry.v4Addr);
#endif /* NATPT_DEBUG */

    /* Construct the IPv4 header */
    ix_ossl_memset(&ipv4Hdr, 0, IX_CC_NATPT_MIN_IPV4_LEN);
    ipv4Hdr.vers_hdrLength = IX_CC_NATPT_IPV4_VERS_IHL;
#if defined (TRANSLATE_TRAFFIC_CLASS)
    ipv4Hdr.tos = *((ix_uint8*)&ipv6Hdr + IX_CC_NATPT_IPV6_TRCLS) << 4;
    ipv4Hdr.tos |= *((ix_uint8*)&ipv6Hdr + IX_CC_NATPT_IPV6_TRCLS + 1) >> 4;
#else
    ipv4Hdr.tos = 0;
#endif /* TRANSLATE_TRAFFIC_CLASS */
    ipv4Hdr.ttl = ipv6Hdr.hopLimit;
    ipv4Hdr.protocol = IX_CC_NATPT_IPPROTO_ICMP;

    /* update packet sizes etc */
    tmpSize = IX_CC_NATPT_IPV6_HDR_LEN - IX_CC_NATPT_MIN_IPV4_LEN;
    packetSize -= tmpSize;
    bufSize -= tmpSize;
    offset += tmpSize;

    ipv4Hdr.totalLength = IX_HTON16(packetSize);

    ipv4Hdr.checksum = 0;
    ipv4Hdr.sourceAddress = natEntry.v4Addr;
    ipv4Hdr.destAddress = ipv6Hdr.destAddress.aUint32[3];

    /* skip extension headers if any */
    if ((ipv6Hdr.nextHeader == IX_CC_NATPT_NEXT_HDR_HOPBYHOP) ||
            (ipv6Hdr.nextHeader == IX_CC_NATPT_NEXT_HDR_ROUTING) ||
            (ipv6Hdr.nextHeader == IX_CC_NATPT_NEXT_HDR_FRAGMENT) ||
            (ipv6Hdr.nextHeader == IX_CC_NATPT_NEXT_HDR_ENCAPSULATION) ||
            (ipv6Hdr.nextHeader == IX_CC_NATPT_NEXT_HDR_AUTHENTICATION) ||
            (ipv6Hdr.nextHeader == IX_CC_NATPT_NEXT_HDR_DSTOPT))
    {
        /* walk thru the chain to calculate */
        extLength = 0;
        extHdr = ( ix_cc_natpt_ipv6_extn_hdr_common*) 
                (pData + IX_CC_NATPT_IPV6_HDR_LEN);
        do {
            extLength += (8 + (extHdr->hdrLength * 8));
            if ((extHdr->nextHeader == IX_CC_NATPT_NEXT_HDR_HOPBYHOP) ||
                (extHdr->nextHeader == IX_CC_NATPT_NEXT_HDR_ROUTING) ||
                (extHdr->nextHeader == IX_CC_NATPT_NEXT_HDR_FRAGMENT) ||
                (extHdr->nextHeader == IX_CC_NATPT_NEXT_HDR_ENCAPSULATION) ||
                (extHdr->nextHeader == IX_CC_NATPT_NEXT_HDR_AUTHENTICATION) ||
                (extHdr->nextHeader == IX_CC_NATPT_NEXT_HDR_DSTOPT))
            {
                /* go to next header */
                extHdr = ( ix_cc_natpt_ipv6_extn_hdr_common* ) 
                        (pData + IX_CC_NATPT_IPV6_HDR_LEN + extLength);
                continue;
            }
            else
            {
                break;
            }
        } while (1); /* intensional; loop terminates using break */
    }
    else
    {
        extLength = 0;
    }

#if defined(NATPT_DEBUG)
    ix_ossl_message_log("Total Extension Header Length (%d)\n", extLength);
#endif /* NATPT_DEBUG */

    packetSize -= extLength;
    bufSize -= extLength;
    offset += extLength;

    /* Get the ICMP Header */
    pData += IX_CC_NATPT_IPV6_HDR_LEN;
    pData += extLength;
    ix_ossl_memcpy(&icmpV6Hdr, pData, IX_CC_NATPT_ICMP_HDR_LEN);

    /* Translate ICMP header */
    IX_ERROR_CRT(
        _ix_cc_natpt_translate_icmp_header_v6tov4(&icmpV6Hdr, &icmpV4Hdr),
        IX_CC_NATPT_ERROR_ICMP_TRANSLATE, IX_ERROR_LEVEL_WARNING);
    icmpV4Hdr.checksum = 0;

    /* If Error message, translate the inner IP header as well */
    if (icmpV6Hdr.type == IX_CC_ICMPV6_TYPE_DEST_UNREACHABLE ||
        icmpV6Hdr.type == IX_CC_ICMPV6_TYPE_PACKET_TOO_BIG ||
        icmpV6Hdr.type == IX_CC_ICMPV6_TYPE_TIME_EXCEEDED ||
        icmpV6Hdr.type == IX_CC_ICMPV6_TYPE_PARAMETER_PROBLEM)
    {
        pData += (IX_CC_NATPT_ICMP_HDR_LEN);

        ix_ossl_memcpy(&errV6Hdr, pData, IX_CC_NATPT_IPV6_HDR_LEN);

        /* Construct the IPv4 header */
        ix_ossl_memset(&errV4Hdr, 0, IX_CC_NATPT_MIN_IPV4_LEN);
        errV4Hdr.vers_hdrLength = IX_CC_NATPT_IPV4_VERS_IHL;
#if defined (TRANSLATE_TRAFFIC_CLASS)
        errV4Hdr.tos = *((ix_uint8*)&errV6Hdr + IX_CC_NATPT_IPV6_TRCLS) << 4;
        errV4Hdr.tos |= *((ix_uint8*)&errV6Hdr + IX_CC_NATPT_IPV6_TRCLS + 1) >> 4;
#else
        errV4Hdr.tos = 0;
#endif /* TRANSLATE_TRAFFIC_CLASS */
        errV4Hdr.ttl = errV6Hdr.hopLimit;
        errV4Hdr.protocol = errV6Hdr.nextHeader;
        
        errSize = IX_NTOH16(errV6Hdr.payloadLength);
        errSize += IX_CC_NATPT_MIN_IPV4_LEN;
        errV4Hdr.totalLength = IX_HTON16(errSize);

        /* update packet sizes etc */
        tmpSize = IX_CC_NATPT_IPV6_HDR_LEN - IX_CC_NATPT_MIN_IPV4_LEN;
        packetSize -= tmpSize;
        bufSize -= tmpSize;
        offset += tmpSize;

        /* Only 64 bits or 8 bytes of data; May have to free linked buffers */
        tmpSize = (IX_CC_NATPT_MIN_IPV4_LEN * 2) + IX_CC_NATPT_ICMP_HDR_LEN + 8;
        if (packetSize > tmpSize)
        {
#if defined(NATPT_DEBUG)
            ix_ossl_message_log ("More than 8 bytes of err ip packet\n");
#endif /* NATPT_DEBUG */
            packetSize = tmpSize;
            bufSize = tmpSize;
        }

        IX_ERROR_CRT(
            ix_rm_buffer_is_eop(arg_hPacket, &isEOP),
            IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
        if ( ! isEOP)
        {
#if defined(NATPT_DEBUG)
            ix_ossl_message_log ("More than one buffer\n");
#endif /* NATPT_DEBUG */
            /* unlink the buffer  and free the rest of chain */
            IX_ERROR_CRT(
                ix_rm_buffer_unlink(&arg_hPacket, &unlinkedBuffer),
                IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

#if defined(NATPT_DEBUG)
            ix_ossl_message_log ("Freeing rest of buffers\n");
#endif /* NATPT_DEBUG */
            IX_ERROR_CRT(
                ix_rm_buffer_free_chain(unlinkedBuffer),
                IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
        }

        errV4Hdr.checksum = 0;
        errV4Hdr.sourceAddress = errV6Hdr.sourceAddress.aUint32[3];
        errV4Hdr.destAddress = natEntry.v4Addr;

        /* Calculate the checksum */
        pHdr = (ix_uint16*)&errV4Hdr;
        for (i = 0; i < (IX_CC_NATPT_MIN_IPV4_LEN >> 1); i++)
        {
            csum += IX_NTOH16(*pHdr);
            pHdr++;
        }
        csum = ~((csum & 0xffff) + (csum >> 16));
        errV4Hdr.checksum = IX_HTON16((ix_uint16)csum);

        /* Set the pointer to where err IP header is written*/
        pData += (IX_CC_NATPT_IPV6_HDR_LEN - IX_CC_NATPT_MIN_IPV4_LEN);
        /* write err ip header */
        ix_ossl_memcpy(pData, &errV4Hdr, IX_CC_NATPT_MIN_IPV4_LEN);

        /* Set the pointer to where ICMPv4 header is written */
        pData -= IX_CC_NATPT_ICMP_HDR_LEN;
    }

    /* write ICMP header */
    ix_ossl_memcpy(pData, &icmpV4Hdr, IX_CC_NATPT_ICMP_HDR_LEN);

    /* update outer header length again */
    ipv4Hdr.totalLength = IX_HTON16(packetSize);

    /* Calculate the outer Ipv4 header checksum */
    pHdr = (ix_uint16*)&ipv4Hdr;
    csum = 0;
    for (i = 0; i < (IX_CC_NATPT_MIN_IPV4_LEN >> 1); i++)
    {
        csum += IX_NTOH16(*pHdr);
        pHdr++;
    }
    csum = ~((csum & 0xffff) + (csum >> 16));
    ipv4Hdr.checksum = IX_HTON16((ix_uint16)csum);

    /* Set the pointer to where IPv4 header is written */
    pData -= IX_CC_NATPT_MIN_IPV4_LEN;

    /* write IPv4 header */
    ix_ossl_memcpy(pData, &ipv4Hdr, IX_CC_NATPT_MIN_IPV4_LEN);

    /* Calculate ICMPv4 checksum */
    _ix_cc_natpt_icmp_v4_calculate_checksum(pData);

    /* Update the meta data */
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_PacketSize, IX_HTON16(packetSize));
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_BufferSize, IX_HTON16(bufSize));
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_Offset, IX_HTON16(offset));

    bufInfo = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_BufferInfo));
    IX_RM_HW_META_SET_HEADER_TYPE(bufInfo, IX_CC_NATPT_HDRTYPE_IPV4);
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_BufferInfo, IX_HTON16(bufInfo));

    cellCount = ix_cc_calculate_cell_count(bufSize, 1);
    ix_rm_buffer_cell_count_set(&arg_hPacket, cellCount);

#if defined(NATPT_DEBUG)
    ix_ossl_message_log (" After Translation Packet Details\n");
    ix_ossl_message_log ("PacketSize (0x%x) \n", 
                 IX_NTOH16(pHwMeta->m_PacketSize));
    ix_ossl_message_log ("BufferSize (0x%x) \n", 
                 IX_NTOH16(pHwMeta->m_BufferSize));
    ix_ossl_message_log ("BufferInfo (0x%x) \n", 
                 IX_NTOH16(pHwMeta->m_BufferInfo));
    ix_ossl_message_log ("Offset (0x%x) \n", 
                 IX_NTOH16(pHwMeta->m_Offset));
    ix_ossl_message_log ("Cell Count (0x%lx) \n", cellCount);
#endif /* NATPT_DEBUG */

#if defined(NATPT_DEBUG)
    while (packetSize)
    {
        for (i = 0; i < 16 && packetSize; i++)
        {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
            ix_ossl_message_log(" %02x", *pData);
#else
            printf(" %02x", *pData);
#endif
            packetSize --;
            pData ++;
        }
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_ossl_message_log("\n");
#else
        printf("\n");
#endif
    }
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_ossl_message_log("\n");
#else
    printf("\n");
#endif
#endif /* NATPT_DEBUG */

    IX_ERROR_CRT(
        ix_rm_packet_send(IX_CC_NATPT_PKT_IPV4_OUTPUT, arg_hPacket, 0),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    return IX_SUCCESS;
}
    

/**
 * NAME: _ix_cc_natpt_icmp_v4_packet
 *
 * DESCRIPTION: This primitive is used to translate an ICMPv4 packet to ICMPv6
 * packet. Then the packet is given to IPv6 Forwarder.
 *           
 * @Param:  - IN arg_hPacket - handle to buffer containing the packet.
 *
 * @Param:  - IN arg_pContext - pointer to tunneling core component
 * context buffer.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * IX_CC_NATPT_ERROR_ICMP_TRANSLATE - could not translate
 * 
 */
ix_error _ix_cc_natpt_icmp_v4_packet(
    ix_buffer_handle arg_hPacket,  
    void* arg_pContext
    )
{
    ix_uint8* pData;
    ix_cc_natpt_ipv4_header ipv4Hdr;
    ix_cc_natpt_ipv6_header ipv6Hdr;
    ix_cc_natpt_icmp_header icmpV6Hdr;
    ix_cc_natpt_icmp_header icmpV4Hdr;
    ix_cc_natpt_ipv4_header errV4Hdr;
    ix_cc_natpt_ipv6_header errV6Hdr;
    ix_uint128 ipv6Addr;
    ix_hw_buffer_meta *pHwMeta;
    ix_cc_natpt_context* pContext;
    ix_uint16 packetSize;
    ix_cc_natpt_nat_entry natEntry;
    ix_uint16 bufSize;
    ix_uint16 offset;
    ix_uint16 tmpSize;
    ix_uint16 bufInfo;

    pContext = (ix_cc_natpt_context*)arg_pContext;

    /* Get the packet meta-data */
    IX_ERROR_CRT(
        ix_rm_buffer_get_meta(arg_hPacket, (void **)&pHwMeta),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    /* read meta data sizes */
    packetSize = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_PacketSize));
    bufSize = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_BufferSize));
    offset = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_Offset));

    /* Get packet data */
    IX_ERROR_CRT(
        ix_rm_buffer_get_data(arg_hPacket, (void **)&pData),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    pData += IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_Offset));

    ix_ossl_memcpy(&ipv4Hdr, pData, IX_CC_NATPT_MIN_IPV4_LEN);

#if defined(NATPT_DEBUG)
    ix_ossl_message_log("Received Packet Src Address (%lx)\n", ipv4Hdr.sourceAddress);
    ix_ossl_message_log("Received Packet Dst Address (%lx)\n", ipv4Hdr.destAddress);
#endif /* NATPT_DEBUG */

    /* Get the mapping IPv6 address */
    IX_ERROR_CRT(
        _ix_cc_natpt_get_ipv4_to_ipv6_mapping(&ipv4Hdr, &natEntry, pContext),
        IX_CC_ERROR_OOR, IX_ERROR_LEVEL_WARNING);

#if defined(NATPT_DEBUG)
    ix_ossl_message_log("Mapping gave IPv6 address \n");
    _ix_cc_natpt_print_ipv6_address(&natEntry.v6Addr);
#endif /* NATPT_DEBUG */

    /* Construct the IPv6 header */
    ix_ossl_memset(&ipv6Hdr, 0, IX_CC_NATPT_IPV6_HDR_LEN);
    ipv6Hdr.version = 6;
#if defined (TRANSLATE_TRAFFIC_CLASS)
    ipv6Hdr.traffic_class = ipv4Hdr.tos;
#else
    ipv6Hdr.traffic_class = 0;
#endif /* TRANSLATE_TRAFFIC_CLASS */
    ipv6Hdr.traffic_class = ipv4Hdr.tos;
    ipv6Hdr.flow_label = 0;
    ipv6Hdr.nextHeader = IX_CC_NATPT_NEXT_HDR_ICMP;
    ipv6Hdr.hopLimit = ipv4Hdr.ttl;

    /* update packet sizes etc */
    tmpSize = IX_CC_NATPT_IPV6_HDR_LEN - IX_CC_NATPT_MIN_IPV4_LEN;
    packetSize += tmpSize;
    bufSize += tmpSize;
    offset -= tmpSize;
#if defined(NATPT_DEBUG)
    ix_ossl_message_log ("Outer Header changed PacketSize (0x%x) \n", packetSize);
#endif /* NATPT_DEBUG */

    ipv6Hdr.payloadLength = IX_HTON16(packetSize - IX_CC_NATPT_IPV6_HDR_LEN) ;

    MAKE_IPV4_MAPPED_IPV6_ADDRESS(ipv4Hdr.sourceAddress, &ipv6Addr);
    /*IX_NTOH128(&ipv6Addr, &ipv6Addr);*/
    ix_ossl_memcpy(&ipv6Hdr.sourceAddress, &ipv6Addr, 
                IX_CC_NATPT_IPV6ADDRLEN_BYTES);
    
    ix_ossl_memcpy(&ipv6Hdr.destAddress, &natEntry.v6Addr, 
                IX_CC_NATPT_IPV6ADDRLEN_BYTES);

    /* Get the ICMP Header */
    pData += IX_CC_NATPT_MIN_IPV4_LEN;
    ix_ossl_memcpy(&icmpV4Hdr, pData, IX_CC_NATPT_ICMP_HDR_LEN);

    /* Translate ICMP header */
    IX_ERROR_CRT(
        _ix_cc_natpt_translate_icmp_header_v4tov6(&icmpV4Hdr, &icmpV6Hdr),
        IX_CC_NATPT_ERROR_ICMP_TRANSLATE, IX_ERROR_LEVEL_WARNING);
    icmpV6Hdr.checksum = 0;

    /* If Error message, translate the inner IP header as well */
    if (icmpV4Hdr.type == IX_CC_ICMPV4_TYPE_DEST_UNREACHABLE ||
        icmpV4Hdr.type == IX_CC_ICMPV4_TYPE_TIME_EXCEEDED ||
        icmpV4Hdr.type == IX_CC_ICMPV4_TYPE_PARAMETER_PROBLEM)
    {
        pData += (IX_CC_NATPT_ICMP_HDR_LEN);

        ix_ossl_memcpy(&errV4Hdr, pData, IX_CC_NATPT_MIN_IPV4_LEN);

        /* Construct the IPv6 header */
        ix_ossl_memset(&errV6Hdr, 0, IX_CC_NATPT_IPV6_HDR_LEN);
        errV6Hdr.version = 6;
#if defined (TRANSLATE_TRAFFIC_CLASS)
        errV6Hdr.traffic_class = errV4Hdr.tos;
#else
        errV6Hdr.traffic_class = 0;
#endif /* TRANSLATE_TRAFFIC_CLASS */
        errV6Hdr.flow_label = 0;
        errV6Hdr.nextHeader = errV4Hdr.protocol;
        errV6Hdr.hopLimit = errV4Hdr.ttl;

        /* update packet sizes etc */
        tmpSize = IX_CC_NATPT_IPV6_HDR_LEN - IX_CC_NATPT_MIN_IPV4_LEN;
        packetSize += tmpSize;
        bufSize += tmpSize;
        offset -= tmpSize;

#if defined(NATPT_DEBUG)
    ix_ossl_message_log ("Inner Header changed PacketSize (0x%x) \n", packetSize);
#endif /* NATPT_DEBUG */
  
        tmpSize = IX_NTOH16(errV4Hdr.totalLength);
        tmpSize -= IX_CC_NATPT_IPV6_HDR_LEN;
        errV6Hdr.payloadLength = IX_HTON16(tmpSize);

        /*IX_NTOH128(&ipv6Addr, &natEntry.v6Addr);*/
        ix_ossl_memcpy(&errV6Hdr.sourceAddress, &natEntry.v6Addr, 
                IX_CC_NATPT_IPV6ADDRLEN_BYTES);
    
        MAKE_IPV4_MAPPED_IPV6_ADDRESS(errV4Hdr.sourceAddress, &ipv6Addr);
        /*IX_NTOH128(&ipv6Addr, &ipv6Addr);*/
        ix_ossl_memcpy(&errV6Hdr.destAddress, &ipv6Addr, 
                    IX_CC_NATPT_IPV6ADDRLEN_BYTES);

        /* Set the pointer to where err IP header is written*/
        pData -= (IX_CC_NATPT_IPV6_HDR_LEN - IX_CC_NATPT_MIN_IPV4_LEN);
        /* write err ip header */
        ix_ossl_memcpy(pData, &errV6Hdr, IX_CC_NATPT_IPV6_HDR_LEN);

        /* Set the pointer to where ICMPv6 header is written */
        pData -= IX_CC_NATPT_ICMP_HDR_LEN;

    }

    /* update length in outer ip header again*/
    ipv6Hdr.payloadLength = IX_HTON16(packetSize - IX_CC_NATPT_IPV6_HDR_LEN) ;

    /* write ICMP header */
    ix_ossl_memcpy(pData, &icmpV6Hdr, IX_CC_NATPT_ICMP_HDR_LEN);

    /* Set the pointer to where IPv4 header is written */
    pData -= IX_CC_NATPT_IPV6_HDR_LEN;
    /* write IPv6 header */
    ix_ossl_memcpy(pData, &ipv6Hdr, IX_CC_NATPT_IPV6_HDR_LEN);

    /* Calculate ICMPv6 checksum */
    _ix_cc_natpt_icmp_v6_calculate_checksum(pData);

    /* Update the meta data */
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_PacketSize, IX_HTON16(packetSize));
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_BufferSize, IX_HTON16(bufSize));
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_Offset, IX_HTON16(offset));

    bufInfo = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_BufferInfo));
    IX_RM_HW_META_SET_HEADER_TYPE(bufInfo, IX_CC_NATPT_HDRTYPE_IPV6);
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_BufferInfo, IX_HTON16(bufInfo));

#if defined(NATPT_DEBUG)
    ix_ossl_message_log (" After Translation Packet Details\n");
    ix_ossl_message_log ("PacketSize (0x%x) \n", 
                 IX_NTOH16(pHwMeta->m_PacketSize));
    ix_ossl_message_log ("BufferSize (0x%x) \n", 
                 IX_NTOH16(pHwMeta->m_BufferSize));
    ix_ossl_message_log ("BufferInfo (0x%x) \n", 
                 IX_NTOH16(pHwMeta->m_BufferInfo));
    ix_ossl_message_log ("Offset (0x%x) \n", 
                 IX_NTOH16(pHwMeta->m_Offset));
#endif /* NATPT_DEBUG */

#if defined(NATPT_DEBUG)
    while (packetSize)
    {
        int i;
        for (i = 0; i < 16 && packetSize; i++)
        {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
            ix_ossl_message_log(" %02x", *pData);
#else
            printf(" %02x", *pData);
#endif
            packetSize --;
            pData ++;
        }
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_ossl_message_log("\n");
#else
        printf("\n");
#endif
    }
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_ossl_message_log("\n");
#else
    printf("\n");
#endif
#endif /* NATPT_DEBUG */

    IX_ERROR_CRT(
        ix_rm_packet_send(IX_CC_NATPT_PKT_IPV6_OUTPUT, arg_hPacket, 0),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    return IX_SUCCESS;
}



/**
 * NAME: _ix_cc_natpt_translate_icmp_header_v6tov4
 *
 * DESCRIPTION: This function is used to translate the ICMPv6 header to 
 * ICMPv4 header.
 *           
 * @Param:  - IN arg_pIcmpV6 - pointer to ICMPv6 Header
 *
 * @Param:  - IN arg_pIcmpV4 - pointer to ICMPv4 Header
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * IX_CC_NATPT_ERROR_ICMP_TRANSLATE - invalid code or type
 * 
 */
PRIVATE ix_error _ix_cc_natpt_translate_icmp_header_v6tov4(
            ix_cc_natpt_icmp_header* arg_pIcmpV6,
            ix_cc_natpt_icmp_header* arg_pIcmpV4)
{
    switch (arg_pIcmpV6->type)
    {
        case IX_CC_ICMPV6_TYPE_DEST_UNREACHABLE:
            arg_pIcmpV4->type = IX_CC_ICMPV4_TYPE_DEST_UNREACHABLE;
            switch (arg_pIcmpV6->code)
            {
                case IX_CC_ICMPV6_DU_CODE_COMM_PROHIBITED:
                case IX_CC_ICMPV6_DU_CODE_NOROUTE:
                case IX_CC_ICMPV6_DU_CODE_NOT_ASSIGNED:
                case IX_CC_ICMPV6_DU_CODE_ADDR_UNREACHABLE:
                    arg_pIcmpV4->code = IX_CC_ICMPV4_DU_CODE_HOST_UNREACHABLE;
                    break;
                case IX_CC_ICMPV6_DU_CODE_PORT_UNREACHABLE:
                    arg_pIcmpV4->code = IX_CC_ICMPV4_DU_CODE_PORT_UNREACHABLE;
                    break;
                default:
                    /* Silently Drop the packet */
                    return IX_CC_NATPT_ERROR_ICMP_TRANSLATE;
            }
            break;
        case IX_CC_ICMPV6_TYPE_PACKET_TOO_BIG:
            arg_pIcmpV4->type = IX_CC_ICMPV4_TYPE_DEST_UNREACHABLE;
            switch (arg_pIcmpV6->code)
            {
                case IX_CC_ICMPV6_DU_CODE_COMM_PROHIBITED:
                case IX_CC_ICMPV6_DU_CODE_NOROUTE:
                case IX_CC_ICMPV6_DU_CODE_NOT_ASSIGNED:
                case IX_CC_ICMPV6_DU_CODE_ADDR_UNREACHABLE:
                    arg_pIcmpV4->code = IX_CC_ICMPV4_DU_CODE_HOST_UNREACHABLE;
                    break;
                case IX_CC_ICMPV6_DU_CODE_PORT_UNREACHABLE:
                    arg_pIcmpV4->code = IX_CC_ICMPV4_DU_CODE_PORT_UNREACHABLE;
                    break;
                default:
                    /* Silently Drop the packet */
                    return IX_CC_NATPT_ERROR_ICMP_TRANSLATE;
            }
            break;
        case IX_CC_ICMPV6_TYPE_TIME_EXCEEDED:
            arg_pIcmpV4->type = IX_CC_ICMPV4_TYPE_TIME_EXCEEDED;
            arg_pIcmpV4->code = arg_pIcmpV6->code;
            break;
        case IX_CC_ICMPV6_TYPE_PARAMETER_PROBLEM:
            switch (arg_pIcmpV6->code)
            {
                case IX_CC_ICMPV6_PP_CODE_WRONG_HDR_FIELD:
                    arg_pIcmpV4->type = IX_CC_ICMPV4_TYPE_DEST_UNREACHABLE;
                    arg_pIcmpV4->code = IX_CC_ICMPV4_DU_CODE_PROTOCOL_UNREACHABLE;
                    break;
                case IX_CC_ICMPV6_PP_CODE_NEXT_HDR_UNKNOWN:
                case IX_CC_ICMPV6_PP_CODE_OPTION_UNKNOWN:
                    arg_pIcmpV4->type = IX_CC_ICMPV4_TYPE_PARAMETER_PROBLEM;
                    arg_pIcmpV4->code = IX_CC_ICMPV4_PP_CODE_POINTER_INDICATES;
                    break;
                default:
                    /* Silently Drop the packet */
                    return IX_CC_NATPT_ERROR_ICMP_TRANSLATE;
            }
            break;
        case IX_CC_ICMPV6_TYPE_ECHO_REQUEST:
            arg_pIcmpV4->type = IX_CC_ICMPV4_TYPE_ECHO_REQUEST;
            arg_pIcmpV4->code = arg_pIcmpV6->code;
            break;
        case IX_CC_ICMPV6_TYPE_ECHO_REPLY:
            arg_pIcmpV4->type = IX_CC_ICMPV4_TYPE_ECHO_REPLY;
            arg_pIcmpV4->code = arg_pIcmpV6->code;
            break;
        default:
            /* Silently Drop the packet */
            return IX_CC_NATPT_ERROR_ICMP_TRANSLATE;
    }

    arg_pIcmpV4->extra = arg_pIcmpV6->extra;

    return IX_SUCCESS;
}


/**
 * NAME: _ix_cc_natpt_translate_icmp_header_v4tov6
 *
 * DESCRIPTION: This function is used to translate the ICMPv4 header to 
 * ICMPv6 header.
 *           
 * @Param:  - IN arg_pIcmpV6 - pointer to ICMPv6 Header
 *
 * @Param:  - IN arg_pIcmpV4 - pointer to ICMPv4 Header
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * IX_CC_NATPT_ERROR_ICMP_TRANSLATE - invalid code or type
 * 
 */
PRIVATE ix_error _ix_cc_natpt_translate_icmp_header_v4tov6(
            ix_cc_natpt_icmp_header* arg_pIcmpV4,
            ix_cc_natpt_icmp_header* arg_pIcmpV6)
{
    switch (arg_pIcmpV4->type)
    {
        case IX_CC_ICMPV4_TYPE_DEST_UNREACHABLE:
            switch (arg_pIcmpV4->code)
            {
                case IX_CC_ICMPV4_DU_CODE_NET_UNREACHABLE:
                case IX_CC_ICMPV4_DU_CODE_HOST_UNREACHABLE:
                case IX_CC_ICMPV4_DU_CODE_SOURCE_ROUTE_FAIL:
                case IX_CC_ICMPV4_DU_CODE_SIX:
                case IX_CC_ICMPV4_DU_CODE_SEVEN:
                case IX_CC_ICMPV4_DU_CODE_EIGHT:
                case IX_CC_ICMPV4_DU_CODE_ELEVEN:
                case IX_CC_ICMPV4_DU_CODE_TWELVE:
                    arg_pIcmpV6->type = IX_CC_ICMPV6_TYPE_DEST_UNREACHABLE;
                    arg_pIcmpV6->code = IX_CC_ICMPV6_DU_CODE_NOROUTE;
                    break;
                case IX_CC_ICMPV4_DU_CODE_PROTOCOL_UNREACHABLE:
                    arg_pIcmpV6->type = IX_CC_ICMPV6_TYPE_PARAMETER_PROBLEM;
                    arg_pIcmpV6->code = IX_CC_ICMPV6_PP_CODE_NEXT_HDR_UNKNOWN;
                    /* adjust pointer to next hdr field*/
                    break;
                case IX_CC_ICMPV4_DU_CODE_PORT_UNREACHABLE:
                    arg_pIcmpV6->type = IX_CC_ICMPV6_TYPE_DEST_UNREACHABLE;
                    arg_pIcmpV6->code = IX_CC_ICMPV6_DU_CODE_PORT_UNREACHABLE;
                    break;
                case IX_CC_ICMPV4_DU_CODE_FRAGMENTATION:
                    arg_pIcmpV6->type = IX_CC_ICMPV6_TYPE_PACKET_TOO_BIG;
                    arg_pIcmpV6->code = IX_CC_ICMPV6_CODE_ZERO;
                    /* adjust MTU */
                    break;
                case IX_CC_ICMPV4_DU_CODE_NINE:
                case IX_CC_ICMPV4_DU_CODE_TEN:
                    arg_pIcmpV6->type = IX_CC_ICMPV6_TYPE_DEST_UNREACHABLE;
                    arg_pIcmpV6->code = IX_CC_ICMPV6_DU_CODE_COMM_PROHIBITED;
                    break;
                default:
                    /* Silently Drop the packet */
                    return IX_CC_NATPT_ERROR_ICMP_TRANSLATE;
            }
            break;
        case IX_CC_ICMPV4_TYPE_TIME_EXCEEDED:
            arg_pIcmpV6->type = IX_CC_ICMPV6_TYPE_TIME_EXCEEDED;
            arg_pIcmpV6->code = arg_pIcmpV4->code;
            break;
        case IX_CC_ICMPV4_TYPE_PARAMETER_PROBLEM:
            arg_pIcmpV6->type = IX_CC_ICMPV6_TYPE_PARAMETER_PROBLEM;
            arg_pIcmpV6->code = IX_CC_ICMPV6_PP_CODE_NEXT_HDR_UNKNOWN;
            /* adjust pointer to corresponding field*/
            break;
        case IX_CC_ICMPV4_TYPE_ECHO_REQUEST:
            arg_pIcmpV6->type = IX_CC_ICMPV6_TYPE_ECHO_REQUEST;
            arg_pIcmpV6->code = arg_pIcmpV4->code;
            break;
        case IX_CC_ICMPV4_TYPE_ECHO_REPLY:
            arg_pIcmpV6->type = IX_CC_ICMPV6_TYPE_ECHO_REPLY;
            arg_pIcmpV6->code = arg_pIcmpV4->code;
            break;
        default:
            /* Silently Drop the packet */
            return IX_CC_NATPT_ERROR_ICMP_TRANSLATE;
    }

    arg_pIcmpV6->extra = arg_pIcmpV4->extra;

    return IX_SUCCESS;
}


/**
 * NAME: _ix_cc_natpt_icmp_v6_calculate_checksum
 *
 * DESCRIPTION:  This function calculates the checksum for the 
 * ICMP message as defined in RFC2460 and RFC2463
 *           
 * @Param:  - IN icmpMessage - pointer to ICMPv6 Message
 *
 * @Return: None
 *
 * Error Codes:
 * 
 */

PRIVATE void _ix_cc_natpt_icmp_v6_calculate_checksum(ix_uint8 *icmpMessage)
{
    ix_cc_natpt_ipv6_header *pIpHdr;
    ix_cc_natpt_icmp_header *pIcmpHdr;
    ix_cc_natpt_ipv6_pseudo_header pseudoHdr;
    ix_uint16 length;
    ix_uint16 isExtraOctet = 0;
    ix_uint32 checksum = 0;
    ix_uint16 *pCurPosn;
    ix_uint16 count;

    pIpHdr = (ix_cc_natpt_ipv6_header *)icmpMessage; 
    pIcmpHdr = (ix_cc_natpt_icmp_header *)
                (icmpMessage + IX_CC_NATPT_IPV6_HDR_LEN);    

    /* First, set the checksum field in the ICMP header to zero */
    pIcmpHdr->checksum=0;

    /* 
     * Prepare the pseudo header. This header contains the source address
     * destination address, the length of the "upper layer" payload, a 
     * zero field and a next header. See RFC 2460, Sec 8.1 for details.
     */
    ix_ossl_memcpy(&pseudoHdr.srcAddress, &pIpHdr->sourceAddress, 
            IX_CC_NATPT_IPV6ADDRLEN_BYTES);
    ix_ossl_memcpy(&pseudoHdr.dstAddress, &pIpHdr->destAddress, 
            IX_CC_NATPT_IPV6ADDRLEN_BYTES);
    
    pseudoHdr.pktLength = (ix_uint32) pIpHdr->payloadLength; 
    pseudoHdr.unused = 0;
    pseudoHdr.nextHdr = IX_CC_NATPT_NEXT_HDR_ICMP;

    /* Calculate checksum for the pseudo header */
    length = sizeof(ix_cc_natpt_ipv6_pseudo_header)/2;
    pCurPosn = (ix_uint16 *) &pseudoHdr;
    while(length)
    {
        checksum += (ix_uint16)*pCurPosn;
        pCurPosn++;
        length--;
    }
    
    /* Calculate the checksum for the rest of the ICMPv6 packet */
    pCurPosn = (ix_uint16 *)(icmpMessage + IX_CC_NATPT_IPV6_HDR_LEN);

    /* 
     * Check if the packet does not end at a 16-bit boundary. This will 
     * be used later to add any remaining octets to the sum
     */

    length = IX_NTOH16(pIpHdr->payloadLength);
    isExtraOctet = length%2;

    /*
     * The length stored in the 'length' variable is in octets. Convert
     * it into the # of 16 bit words
     */
    length=length/2;

    for(count = 0; count < length; count++, pCurPosn++)
    {
        checksum+=(ix_uint16)*pCurPosn;
    }
    
    /* Now add the remaining octet, if any, to the sum */
    if(isExtraOctet)
    {
        checksum+= *(ix_uint8 *)pCurPosn;
    }

    /*
     * Now add the upper 16 bits to the lower 16 bits until we end up with a 
     * 16 bit value. This is the 1's complement sum
     */
    while(checksum>>16)
    {
        checksum = (checksum>>16) + (checksum & 0xffff);
    }
    
    /* 
     * Finally, take the 1's complement of the calculated 1's complement sum
     * and store it in the checksum field of the ICMP header
     */
    pIcmpHdr->checksum = (ix_uint16)~checksum;

}


/**
 * NAME: _ix_cc_natpt_icmp_v4_calculate_checksum
 *
 * DESCRIPTION:  This function calculates the checksum for the 
 * ICMP message as defined in RFC1071 
 *           
 * @Param:  - IN icmpMessage - pointer to ICMPv4 Message
 *
 * @Return: None
 *
 * Error Codes:
 * 
 */

PRIVATE void _ix_cc_natpt_icmp_v4_calculate_checksum(ix_uint8 *icmpMessage)
{
    ix_cc_natpt_ipv4_header *pIpHdr;
    ix_cc_natpt_icmp_header *pIcmpHdr;
    ix_int32 sum = 0;
    ix_uint8 *pData;
    ix_uint32 size;

    pIpHdr = (ix_cc_natpt_ipv4_header *)icmpMessage; 
    pIcmpHdr = (ix_cc_natpt_icmp_header *)
                    (icmpMessage + IX_CC_NATPT_MIN_IPV4_LEN);    

    /* First, set the checksum field in the ICMP header to zero */
    pIcmpHdr->checksum=0;

    pData = (ix_uint8*) pIcmpHdr;
    size = IX_NTOH16(pIpHdr->totalLength);
    size -= IX_CC_NATPT_MIN_IPV4_LEN;

    while ( size > 1 )
    {
        sum += *(ix_uint16 *)(void *)pData;
        pData += 2;
        size -= 2;
    }
    if ( size > 0 )
    {
        sum += *(ix_uint8 *)pData;
    }
    while ( sum >> 16 ) 
    {
        sum = ( sum & 0xffff ) + ( sum >> 16 );
    }

    pIcmpHdr->checksum = (ix_uint16)~sum;
}


