/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.x for the IXP2xxx Network Processors
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      NATPT Core Component - Library Primitives.
 *
 * = FILENAME
 *      ix_cc_natpt_library.c
 *
 * = DESCRIPTION
 *      The file defines library interface functions.
 *
 * = AKNOWLEDGEMENTS
 *      
 * = CREATION TIME
 *      03/03/03
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_natpt_library.c,v 1.18 2003/05/16 21:51:03 sshegde Exp $";

/**
 * System defined include files
 */
#include "ix_cc_error.h"
#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_netmacros.h"
#include "ix_cc.h"
#include "bindings.h"
#include "cc/ix_cc_natpt.h"


/**
 * User defined include files
 */
#include "cc/natpt/internal/ix_cc_natpt_msg_handler.h"
#include "cc/natpt/internal/ix_cc_natpt_prop.h"
#include "cc/natpt/internal/ix_cc_natpt_init.h"
#include "cc/natpt/internal/ix_cc_natpt_lists.h"
#include "cc/natpt/internal/ix_cc_natpt_nat_table.h"
#include "cc/natpt/internal/ix_cc_natpt_utils.h"


/**
 * Preprocessor symbols and Macros used in this file
 */

/**
 * Type definitions whose scope is limited to this file
 */

/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */

/** 
 * Extern function prototypes.
 */


/** 
 * Static function prototypes.
 */


/** 
 * Function definitions.
 */


/**
 * NAME: ix_cc_natpt_set_config_parameters
 *
 * DESCRIPTION:This library primitive helps client to set global
 * configuration parameters
 * 
 * @Param:  - IN arg_pConfig: pointer to config parameters structure.
 * @Param:  - IN arg_pContext: pointer to NATPT context. 
 *
 * @Param:  - OUT none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM - the information pointed to by 
 *              arg_pConfig is invalid.
 */
ix_error ix_cc_natpt_set_config_parameters(
                ix_cc_natpt_config_params * arg_pConfig,
                void * arg_pContext)
{
    ix_cc_natpt_context* pContext;

    if (arg_pConfig == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pConfig"));
    }/* end if(arg_pConfig == NULL) */

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_INVALID_CONTEXT,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    pContext = (ix_cc_natpt_context*)arg_pContext;

    pContext->idleCheckInterval = arg_pConfig->idleCheckInterval;

    IX_ERROR_CR(
        _ix_cc_natpt_set_entry_timeout_event_handler(pContext));

    return IX_SUCCESS;
}


/**
 * NAME: ix_cc_natpt_get_config_parameters
 *
 * DESCRIPTION:This library primitive helps client to get global
 * configuration parameters.
 * 
 * @Param:  - IN arg_pContext: pointer to NATPT context. 
 *
 * @Param:  - OUT arg_pConfig: Pointer to configuration parameters structure 
 *                             that will be filled.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *
 * IX_CC_ERROR_NULL - invalid pointer
 */
ix_error ix_cc_natpt_get_config_parameters(
                ix_cc_natpt_config_params * arg_pConfig,
                void * arg_pContext)
{
    ix_cc_natpt_context* pContext;

    if (arg_pConfig == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pConfig"));
    }/* end if(arg_pConfig == NULL) */

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_INVALID_CONTEXT,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    pContext = (ix_cc_natpt_context*)arg_pContext;

    arg_pConfig->idleCheckInterval = pContext->idleCheckInterval;

    return IX_SUCCESS;
}



/**
 * NAME: ix_cc_natpt_add_static_mapping
 *
 * DESCRIPTION:This library primitive add static ipv6/ipv4 address
 * mapping.
 * 
 * @Param:  - IN arg_pv6v4Map: Pointer to a v6 v4 address mapping structure.
 * @Param:  - IN arg_pContext: pointer to NATPT context. 
 *
 * @Param:  - OUT none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM - the information pointed to by 
 *              arg_pv6v4Map is invalid.
 */
ix_error ix_cc_natpt_add_static_mapping(
                ix_cc_natpt_static_v6v4map * arg_pv6v4Map,
                void * arg_pContext)
{
    ix_error err = IX_SUCCESS;
    ix_cc_natpt_context* pContext;
    ix_cc_natpt_nat_entry natEntry;
    ix_uint32 natIndex;

    if (arg_pv6v4Map == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pv6v4Map"));
    }/* end if(arg_pv6v4Map == NULL) */

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_INVALID_CONTEXT,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    pContext = (ix_cc_natpt_context*)arg_pContext;


    IX_ERROR_CR(
            _ix_cc_natpt_add_v4v6map(pContext->staticMapList, arg_pv6v4Map));

    natEntry.v4Addr = IX_HTON32(arg_pv6v4Map->v4addr);
    IX_HTON128(&arg_pv6v4Map->v6addr, &natEntry.v6Addr); 
    natEntry.unused = 0;
    natEntry.isUsed = 0;
    natEntry.isStatic = 1;
    natEntry.v6NHID = 0xffff;
    
    IX_ERROR_CG(
            _ix_cc_natpt_add_nat_table_entry(
                    pContext->pMbNatTableBase, 
                    pContext->natTableIndexPool, 
                    &natEntry, &natIndex),
            err,
            cleanupList);

    IX_ERROR_CG(
            _ix_cc_natpt_add_nat_hash_table_entries(
                    pContext, &natEntry, natIndex),
            err,
            cleanupHash);

    return IX_SUCCESS;

cleanupHash:
    _ix_cc_natpt_del_nat_table_entry(pContext->pMbNatTableBase, 
            pContext->natTableIndexPool, &natEntry);
cleanupList:
    _ix_cc_natpt_remove_v4v6map(pContext->staticMapList, arg_pv6v4Map);
    return err;
}


/**
 * NAME: ix_cc_natpt_remove_static_mapping
 *
 * DESCRIPTION:This library primitive remove static ipv6/ipv4 address
 * mapping.
 * 
 * @Param:  - IN arg_pv6v4Map: Pointer to a v6 v4 address mapping structure.
 * @Param:  - IN arg_pContext: pointer to NATPT context. 
 *
 * @Param:  - OUT none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM - the information pointed to by 
 *              arg_pv6v4Map is invalid.
 */
ix_error ix_cc_natpt_remove_static_mapping(
                ix_cc_natpt_static_v6v4map * arg_pv6v4Map,
                void * arg_pContext)
{
    ix_cc_natpt_context* pContext;
    ix_cc_natpt_nat_entry natEntry;
    ix_uint32 natIndex;

    if (arg_pv6v4Map == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pv6v4Map"));
    }/* end if(arg_pv6v4Map == NULL) */

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_INVALID_CONTEXT,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    pContext = (ix_cc_natpt_context*)arg_pContext;

    IX_ERROR_CR(
            _ix_cc_natpt_remove_v4v6map(pContext->staticMapList, 
                        arg_pv6v4Map));

    natEntry.v4Addr = IX_HTON32(arg_pv6v4Map->v4addr);
    IX_HTON128(&arg_pv6v4Map->v6addr, &natEntry.v6Addr); 
    natEntry.isUsed = 0;
    natEntry.isStatic = 1;
    natEntry.v6NHID = 0xffff;
    
    IX_ERROR_CR(
            _ix_cc_natpt_find_nat_table_entry(
                    pContext->pMbNatTableBase, pContext->natTableIndexPool,
                    &natEntry, &natIndex));
    
    IX_ERROR_CR(
            _ix_cc_natpt_del_nat_hash_table_entries(
                    pContext, &natEntry, natIndex));
            
    return (_ix_cc_natpt_del_nat_table_entry(
                    pContext->pMbNatTableBase,
                    pContext->natTableIndexPool,
                    &natEntry));
}


/**
 * NAME: ix_cc_natpt_get_static_mapping
 *
 * DESCRIPTION:This library primitive get static ipv6/ipv4 address
 * mapping.
 * 
 * @Param:  - IN arg_pContext: pointer to NATPT context. 
 *
 * @Param:  - IN/OUT arg_NumItems: On input, the number of entries that the 
 *                   buffer pointed to by arg_pv6v4MapArray can accommodate.  
 *                   On output, the actual number of entries written to the 
 *                   buffer.  If IX_CC_NATPT_ERROR_BUFFER_TOO_SMALL is 
 *                   returned, then this number will be the total number 
 *                   of entries that would have been written to the buffer 
 *                   if it had been large enough.
 *
 * @Param:  - OUT arg_pv6v4MapArray : Pointer to buffer into which the array 
 *                of v6-v4 address mappings will be returned.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM - arg_NumItems is zero.
 * IX_CC_NATPT_ERROR_BUFFER_TOO_SMALL - more items exist than could fit 
 *                  into the supplied buffer (non-fatal error)
 */
ix_error ix_cc_natpt_get_static_mapping(
                ix_uint32 * arg_NumItems,
                ix_cc_natpt_static_v6v4map * arg_pv6v4MapArray,
                void * arg_pContext)
{
    ix_cc_natpt_context* pContext;

    if (arg_pv6v4MapArray == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pv6v4MapArray"));
    }/* end if(arg_pv6v4MapArray == NULL) */

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_INVALID_CONTEXT,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (arg_NumItems == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_NumItems"));
    }/* end if(arg_NumItems == NULL) */

    if (*arg_NumItems == 0)
    {
        return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM,
                               ("Invalid value: arg_NumItems"));
    }/* end if(arg_NumItems == NULL) */

    pContext = (ix_cc_natpt_context*)arg_pContext;

    return (_ix_cc_natpt_get_v4v6map(pContext->staticMapList, 
                arg_NumItems, arg_pv6v4MapArray));
}


/**
 * NAME: ix_cc_natpt_add_v4addr_pool
 *
 * DESCRIPTION:This library primitive helps client to add IPv4
 * addresses to pool.
 * 
 * @Param:  - IN arg_NumItems: Number of v4 addrs in the array pointed 
 *                             by the next argument.
 * @Param:  - IN arg_pv4addrArray: Pointer to a v4 address array.
 * @Param:  - IN arg_pContext: pointer to NATPT context. 
 *
 * @Param:  - OUT none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM - the information pointed to 
 *              by arg_pv4addrArray is invalid.
 */
ix_error ix_cc_natpt_add_v4addr_pool(
                ix_uint32 arg_NumItems,
                ix_cc_natpt_v4addr * arg_pv4addrArray,
                void * arg_pContext)
{
    ix_cc_natpt_context* pContext;

    if (arg_pv4addrArray == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pv4addrArray"));
    }/* end if(arg_pv4addrArray == NULL) */

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_INVALID_CONTEXT,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    pContext = (ix_cc_natpt_context*)arg_pContext;

    return (_ix_cc_natpt_add_v4pool(pContext->v4PoolList, arg_NumItems, 
                    arg_pv4addrArray));
}


/**
 * NAME: ix_cc_natpt_remove_v4addr_pool
 *
 * DESCRIPTION:This library primitive helps client to remove IPv4
 * addresses from pool.
 * 
 * @Param:  - IN arg_NumItems: Number of v4 addrs in the array pointed 
 *                             by the next argument.
 * @Param:  - IN arg_pv4addrArray: Pointer to a v4 address array.
 * @Param:  - IN arg_pContext: pointer to NATPT context. 
 *
 * @Param:  - OUT none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM - the information pointed to 
 *              by arg_pv4addrArray is invalid.
 */
ix_error ix_cc_natpt_remove_v4addr_pool(
                ix_uint32 arg_NumItems,
                ix_cc_natpt_v4addr * arg_pv4addrArray,
                void * arg_pContext)
{
    ix_cc_natpt_context* pContext;

    if (arg_pv4addrArray == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pv4addrArray"));
    }/* end if(arg_pv4addrArray == NULL) */

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_INVALID_CONTEXT,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    pContext = (ix_cc_natpt_context*)arg_pContext;

    return (_ix_cc_natpt_remove_v4pool(pContext->v4PoolList, arg_NumItems, 
                    arg_pv4addrArray));
}


/**
 * NAME: ix_cc_natpt_get_v4addr_pool
 *
 * DESCRIPTION:This library primitive helps client to get current
 * list of ipv4 address in pool.
 * 
 * @Param:  - IN arg_pContext: pointer to NATPT context. 
 * 
 * @Param:  - IN/OUT arg_NumItems: On input, the number of entries that the 
 *                   buffer pointed to by arg_pv4addrArray can accommodate.
 *                   On output, the actual number of entries written to the 
 *                   buffer.  If IX_CC_NATPT_ERROR_BUFFER_TOO_SMALL is 
 *                   returned, then this number will be the total number of 
 *                   entries that would have been written to the buffer if 
 *                   it had been large enough.
 *
 * @Param:  - OUT arg_pv4addrArray : Pointer to buffer into which the array 
 *                of IPv4 addresses will be returned.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM - arg_NumItems is zero.
 * IX_CC_NATPT_ERROR_BUFFER_TOO_SMALL - more items exist than could fit 
 *              into the supplied buffer (non-fatal error)
 */
ix_error ix_cc_natpt_get_v4addr_pool(
                ix_uint32 * arg_NumItems,
                ix_cc_natpt_v4addr * arg_pv4addrArray,
                void * arg_pContext)
{
    ix_cc_natpt_context* pContext;

    if (arg_pv4addrArray == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pv4addrArray"));
    }/* end if(arg_pv4addrArray == NULL) */

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_INVALID_CONTEXT,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (arg_NumItems == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_NumItems"));
    }/* end if(arg_NumItems == NULL) */

    if (*arg_NumItems == 0)
    {
        return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM,
                               ("Invalid value: arg_NumItems"));
    }/* end if(arg_NumItems == NULL) */

    pContext = (ix_cc_natpt_context*)arg_pContext;

    return (_ix_cc_natpt_get_v4pool(pContext->v4PoolList, 
                arg_NumItems, arg_pv4addrArray));
}


/**
 * NAME: ix_cc_natpt_set_napt_mode
 *
 * DESCRIPTION:This library primitive helps client to turn NAPT
 * mode ON/OFF.
 * 
 * @Param:  - IN arg_NaptMode: NAPT mode flag.
 * @Param:  - IN arg_pContext: pointer to NATPT context. 
 *
 * @Param:  - OUT none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *
 * IX_CC_ERROR_NULL - invalid pointer
 */
ix_error ix_cc_natpt_set_napt_mode(
                ix_cc_natpt_naptmode arg_NaptMode,
                void * arg_pContext)
{
    return IX_CC_NATPT_ERROR_NOT_IMPLEMENTED;
}


/**
 * NAME: ix_cc_natpt_add_v4v6port_mapping
 *
 * DESCRIPTION:This library primitive helps client to add a 
 * IPv4addr/IPv6Addr/PortNubmer mapping.
 * 
 * @Param:  - IN arg_pv4v6portMap: Pointer to a v4 to v6 address with port 
 *                                 number mapping structure.
 * @Param:  - IN arg_pContext: pointer to NATPT context. 
 *
 * @Param:  - OUT none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM - the information pointed to by 
 *              arg_pv4v6portMap is invalid.
 */
ix_error ix_cc_natpt_add_v4v6port_mapping(
                ix_cc_natpt_v4v6portmap * arg_pv4v6portMap,
                void * arg_pContext)
{
    return IX_CC_NATPT_ERROR_NOT_IMPLEMENTED;
}


/**
 * NAME: ix_cc_natpt_remove_v4v6port_mapping
 *
 * DESCRIPTION:This library primitive helps client to remove a 
 * IPv4addr/IPv6Addr/PortNubmer mapping.
 * 
 * @Param:  - IN arg_pv4v6portMap: Pointer to a v4 to v6 address with port 
 *                                 number mapping structure.
 * @Param:  - IN arg_pContext: pointer to NATPT context. 
 *
 * @Param:  - OUT none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM - the information pointed to by 
 *              arg_pv4v6portMap is invalid.
 */
ix_error ix_cc_natpt_remove_v4v6port_mapping(
                ix_cc_natpt_v4v6portmap * arg_pv4v6portMap,
                void * arg_pContext)
{
    return IX_CC_NATPT_ERROR_NOT_IMPLEMENTED;
}


/**
 * NAME: ix_cc_natpt_get_v4v6port_mapping
 *
 * DESCRIPTION:This library primitive helps client to get current
 * list of IPv4addr/IPv6Addr/PortNubmer mappings.
 * 
 * @Param:  - IN arg_pContext: pointer to NATPT context. 
 * 
 * @Param:  - IN/OUT arg_NumItems: On input, the number of entries that the 
 *                   buffer pointed to by arg_pv4v6portMapArray can 
 *                   accommodate.
 *                   On output, the actual number of entries written to the 
 *                   buffer.  If IX_CC_NATPT_ERROR_BUFFER_TOO_SMALL is 
 *                   returned, then this number will be the total number of 
 *                   entries that would have been written to the buffer if 
 *                   it had been large enough.
 *
 * @Param:  - OUT arg_pv4v6portMapArray : Pointer to buffer into which the 
 *                array of IPv4/IPv6/Port mappings will be returned.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM - arg_NumItems is zero.
 * IX_CC_NATPT_ERROR_BUFFER_TOO_SMALL - more items exist than could fit 
 *              into the supplied buffer (non-fatal error)
 */
ix_error ix_cc_natpt_get_v4v6port_mapping(
                ix_uint32 * arg_NumItems,
                ix_cc_natpt_v4v6portmap * arg_pv4v6portMapArray,
                void * arg_pContext)
{
    return IX_CC_NATPT_ERROR_NOT_IMPLEMENTED;
}


/**
 * NAME: ix_cc_natpt_get_active_sessions
 *
 * DESCRIPTION:This library primitive helps client to get current
 * list of active NATPT sessions.
 * 
 * @Param:  - IN arg_pContext: pointer to NATPT context. 
 * 
 * @Param:  - IN/OUT arg_NumItems: On input, the number of entries that the 
 *                   buffer pointed to by arg_pSessionArray can 
 *                   accommodate.
 *                   On output, the actual number of entries written to the 
 *                   buffer.  If IX_CC_NATPT_ERROR_BUFFER_TOO_SMALL is 
 *                   returned, then this number will be the total number of 
 *                   entries that would have been written to the buffer if 
 *                   it had been large enough.
 *
 * @Param:  - OUT arg_pSessionArray : Pointer to buffer into which the 
 *                array of active NATPT sessions will be returned.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM - arg_NumItems is zero.
 * IX_CC_NATPT_ERROR_BUFFER_TOO_SMALL - more items exist than could fit 
 */
ix_error ix_cc_natpt_get_active_sessions (
                ix_uint32* arg_NumItems,
                ix_cc_natpt_session* arg_pSessionArray,
                void * arg_pContext)
{
    ix_cc_natpt_context* pContext;
    ix_uint32 index;
    ix_uint16 listIndex = 0;
    ix_cc_natpt_nat_entry natEntry;
    ix_error err = IX_SUCCESS;
    ix_uint8 moreItems = 0;

    if (arg_pSessionArray == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pSessionArray"));
    }/* end if(arg_pSessionArray == NULL) */

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_INVALID_CONTEXT,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (arg_NumItems == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_NumItems"));
    }/* end if(arg_NumItems == NULL) */

    if (*arg_NumItems == 0)
    {
        return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM,
                               ("Invalid value: arg_NumItems"));
    }/* end if(arg_NumItems == NULL) */

    pContext = (ix_cc_natpt_context*)arg_pContext;

    moreItems = 0;

    err = ix_cc_index_pool_get_first(pContext->natTableIndexPool, &index);
    while (err == IX_SUCCESS)
    {
        if (listIndex < *arg_NumItems)
        {
            /* Read the entry */
            _READ_NAT_TABLE_ENTRY(pContext->pMbNatTableBase, index, natEntry);

            IX_NTOH128(&natEntry.v6Addr, &arg_pSessionArray[listIndex].v6addr); 
            arg_pSessionArray[listIndex].v4addr = IX_NTOH32(natEntry.v4Addr);
            arg_pSessionArray[listIndex].protocol = 0;
            arg_pSessionArray[listIndex].v6PortNumber = 0;
            arg_pSessionArray[listIndex].v4PortNumber = 0;
        }
        else
        {
            moreItems = 1;
        }

        listIndex++;
        err = ix_cc_index_pool_get_next(
            pContext->natTableIndexPool, index, &index);
    }

    *arg_NumItems = listIndex;

    if (moreItems)
    {
        return IX_CC_NATPT_ERROR_BUFFER_TOO_SMALL;
    }

    return IX_SUCCESS;
}
    
