/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.x for the IXP2xxx Network Processors
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      NATPT Core Component
 *
 * = FILENAME
 *      ix_cc_natpt_lists.c
 *
 * = DESCRIPTION
 *      The file contains the functions and macros needed to manage the
 *      the IPv4 address pool and Static Mappings.
 *
 * = CREATION TIME
 *      03/03/03
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_natpt_lists.c,v 1.15 2003/12/12 00:59:57 ktseng Exp $";


/**
 * System defined include files
 */
#include "ix_cc_error.h"
#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_netmacros.h"
#include "ix_cc.h"
#include "ix_cc_macros.h"
#include "ix_config.h"
#include "bindings.h"
#include "cc/ix_cc_natpt.h"

/**
 * User defined include files
 */
#include "cc/natpt/internal/ix_cc_natpt_init.h"
#include "cc/natpt/internal/ix_cc_natpt_utils.h"
#include "cc/natpt/internal/ix_cc_natpt_lists.h"


/**
 * Preprocessor symbols and Macros used in this file
 */

/**
 * Type definitions whose scope is limited to this file
 */

/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */

/** 
 * Extern function prototypes.
 */


/** 
 * Static function prototypes.
 */


/** 
 * Function definitions.
 */


/**
 * NAME: _ix_cc_natpt_create_lists
 *
 * DESCRIPTION: Creates the static mapping list and V4 address pool list.
 * 
 * @Param:  - IN natptContext: pointer to NATPT context.  
 *
 * @Param:  - OUT none
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *      IX_CC_ERROR_OOM
 * 
 */
ix_error _ix_cc_natpt_create_lists(ix_cc_natpt_context* natptContext)
{
    natptContext->staticMapList = ListCreate();

    if (natptContext->staticMapList == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                      ("Failed to allocate memory for creating a list"));
    }

    natptContext->v4PoolList = ListCreate();

    if (natptContext->v4PoolList == NULL)
    {
        ListDelete(natptContext->staticMapList);
        natptContext->staticMapList = NULL;
        return IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                      ("Failed to allocate memory for creating a list"));
    }

    return IX_SUCCESS;
}


/**
 * NAME: _ix_cc_natpt_cleanup_lists
 *
 * DESCRIPTION: Destroys the static mapping list and V4 address pool list.
 * 
 * @Param:  - IN natptContext: pointer to NATPT context.  
 *
 * @Param:  - OUT none
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 */
ix_error _ix_cc_natpt_cleanup_lists(ix_cc_natpt_context* natptContext)
{
    if (natptContext->staticMapList)
    {
        ListFreeDelete(natptContext->staticMapList);
    }

    if (natptContext->v4PoolList)
    {
        ListFreeDelete(natptContext->v4PoolList);
    }

    return IX_SUCCESS;
}


/**
 * NAME: _ix_cc_natpt_add_v4v6map
 *
 * DESCRIPTION: Adds a static mapping to static mapping list.
 * 
 * @Param:  - IN arg_stMapList: pointer to Static Mapping List.  
 * @Param:  - IN arg_v6v4Map: pointer to Static Mapping to be added.  
 *
 * @Param:  - OUT none
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *      IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM
 *      IX_CC_ERROR_OOM
 * 
 */
ix_error _ix_cc_natpt_add_v4v6map(
                LPLIST arg_stMapList,
                ix_cc_natpt_static_v6v4map* arg_v6v4Map)
{
    LPLISTITR itr;
    ix_cc_natpt_stV4V6MapEntry* pMapEntry;

    if (ListGetCount(arg_stMapList))
    {
        itr = ListItrCreate(arg_stMapList);
        if (itr == NULL)
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                            ("Failed to allocate memory for list iterator"));
        }

        while ((pMapEntry = (ix_cc_natpt_stV4V6MapEntry*) ListItrGetNext(itr)))
        {
            if ( (arg_v6v4Map->v4addr == pMapEntry->v6v4Map.v4addr) &&
                    NATPT_IS_IPV6_ADDRESS_EQUAL(&arg_v6v4Map->v6addr, 
                    &pMapEntry->v6v4Map.v6addr) )
            {
                ListItrDelete(itr);
                return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM,
                            ("Entry already exists"));
            }
        } /* while */

        ListItrDelete(itr);
    }

    pMapEntry = (ix_cc_natpt_stV4V6MapEntry*)
            ix_ossl_malloc(sizeof(ix_cc_natpt_stV4V6MapEntry));
    if ( ! pMapEntry)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                        ("Failed to allocate memory for map entry"));
    }

    pMapEntry->v6v4Map.v4addr = arg_v6v4Map->v4addr;
    NATPT_COPY_IPV6_ADDRESS(&pMapEntry->v6v4Map.v6addr, &arg_v6v4Map->v6addr);

    if ( ! ListAddItem(arg_stMapList, (void*)pMapEntry))
    {
        ix_ossl_free(pMapEntry);
        return IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                        ("Failed to add item to list"));
    }

    return IX_SUCCESS;
}


/**
 * NAME: _ix_cc_natpt_remove_v4v6map
 *
 * DESCRIPTION: Removes a static mapping from static mapping list.
 * 
 * @Param:  - IN arg_stMapList: pointer to Static Mapping List.  
 * @Param:  - IN arg_v6v4Map: pointer to Static Mapping to be removed.  
 *
 * @Param:  - OUT none
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *      IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM
 *      IX_CC_ERROR_OOM
 * 
 */
ix_error _ix_cc_natpt_remove_v4v6map(
                LPLIST arg_stMapList,
                ix_cc_natpt_static_v6v4map * arg_v6v4Map)
{
    LPLISTITR itr;
    ix_cc_natpt_stV4V6MapEntry* pMapEntry;

    if (ListGetCount(arg_stMapList))
    {
        itr = ListItrCreate(arg_stMapList);
        if (itr == NULL)
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                            ("Failed to allocate memory for list iterator"));
        }

        while ((pMapEntry = (ix_cc_natpt_stV4V6MapEntry*) ListItrGetNext(itr)))
        {
            if ( (arg_v6v4Map->v4addr == pMapEntry->v6v4Map.v4addr) &&
                    NATPT_IS_IPV6_ADDRESS_EQUAL(&arg_v6v4Map->v6addr, 
                    &pMapEntry->v6v4Map.v6addr) )
            {
                ListItrRemoveItem(itr);
                ix_ossl_free(pMapEntry);
                ListItrDelete(itr);
                return IX_SUCCESS;
            }
        } /* while */

        ListItrDelete(itr);
    }

    return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM,
                        ("Failed to find the item in list"));
}


/**
 * NAME: _ix_cc_natpt_get_v4v6map
 *
 * DESCRIPTION: Retrievs the current list of static mappings.
 * 
 * @Param:  - IN arg_stMapList: pointer to Static Mapping List.  
 *
 * @Param:  - IN/OUT arg_NumItems: On input, contains number of items
 * the array can hold. On output, contains the actual number of items.
 *
 * @Param:  - OUT arg_v6v4Map: pointer to Static Mapping array where
 * retrieved mappings are placed.  
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *      IX_CC_NATPT_ERROR_BUFFER_TOO_SMALL
 *      IX_CC_ERROR_OOM
 * 
 */
ix_error _ix_cc_natpt_get_v4v6map(
                LPLIST arg_stMapList,
                ix_uint32* arg_NumItems,
                ix_cc_natpt_static_v6v4map* arg_v6v4Map)
{
    LPLISTITR itr;
    ix_cc_natpt_stV4V6MapEntry* pMapEntry;
    ix_uint32 givenNum = *arg_NumItems;

    if (ListGetCount(arg_stMapList))
    {
        itr = ListItrCreate(arg_stMapList);
        if (itr == NULL)
        {
            *arg_NumItems = ListGetCount(arg_stMapList);
            return IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                            ("Failed to allocate memory for list iterator"));
        }

        *arg_NumItems = 0;
        while ((pMapEntry = (ix_cc_natpt_stV4V6MapEntry*) ListItrGetNext(itr)))
        {
            if (*arg_NumItems == givenNum)
            {
                *arg_NumItems = ListGetCount(arg_stMapList);
                ListItrDelete(itr);
                return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_BUFFER_TOO_SMALL,
                            ("More items exist"));
            }
            arg_v6v4Map->v4addr = pMapEntry->v6v4Map.v4addr;
            NATPT_COPY_IPV6_ADDRESS(&arg_v6v4Map->v6addr, &pMapEntry->v6v4Map.v6addr);

            (*arg_NumItems) ++;
            arg_v6v4Map ++;
        } /* while */

        ListItrDelete(itr);
    }
    else
    {
        *arg_NumItems = 0;
    }

    return IX_SUCCESS;
}


/**
 * NAME: _ix_cc_natpt_add_v4pool
 *
 * DESCRIPTION: Adds a list of V4 address to v4 address pool list.
 * 
 * @Param:  - IN arg_poolList: pointer to V4 Address Pool List.  
 * @Param:  - IN arg_NumItems: number of items in the array.  
 * @Param:  - IN arg_pv4addrArray: pointer to V4 address array.  
 *
 * @Param:  - OUT none
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *      IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM
 *      IX_CC_ERROR_OOM
 * 
 */
ix_error _ix_cc_natpt_add_v4pool(
                LPLIST arg_poolList,
                ix_uint32 arg_NumItems,
                ix_cc_natpt_v4addr* arg_pv4addrArray)
{
    LPLISTITR itr;
    ix_cc_natpt_v4PoolEntry* v4Entry;
    ix_uint8 found;
    ix_uint32 num;
    ix_cc_natpt_v4addr* addrArray;

    addrArray = arg_pv4addrArray;
    num = arg_NumItems;

    while (num)
    {
        if (ListGetCount(arg_poolList))
        {
            itr = ListItrCreate(arg_poolList);
            if (itr == NULL)
            {
                return IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                    ("Failed to allocate memory for list iterator"));
            }

            found = 0;
            while ((v4Entry = (ix_cc_natpt_v4PoolEntry*) ListItrGetNext(itr)))
            {
                if ( *addrArray == v4Entry->v4addr)
                {
                    found = 1;
                    break;
                }
            } /* while */

            ListItrDelete(itr);

            if (found) 
            {
                return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM,
                        ("Trying to add duplicate entries"));
            }
        }

        num --;
        addrArray ++;

    } /* while (num) */

    addrArray = arg_pv4addrArray;
    num = arg_NumItems;

    while (num)
    {
        v4Entry = (ix_cc_natpt_v4PoolEntry*)
                ix_ossl_malloc(sizeof(ix_cc_natpt_v4PoolEntry));
        if ( ! v4Entry)
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                        ("Failed to allocate memory for pool entry"));
        }

        v4Entry->v4addr = *addrArray;
        v4Entry->isUsed = FALSE;

        if ( ! ListAddItem(arg_poolList, (void*)v4Entry))
        {
            ix_ossl_free(v4Entry);
            return IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                        ("Failed to add item to list"));
        }

        num --;
        addrArray ++;

    } /* while (num) */

    return IX_SUCCESS;
}


/**
 * NAME: _ix_cc_natpt_remove_v4pool
 *
 * DESCRIPTION: Removes a list of v4 addresses from v4 address pool list.
 * 
 * @Param:  - IN arg_poolList: pointer to V4 Address Pool List.  
 * @Param:  - IN arg_NumItems: number of items in the array.  
 * @Param:  - IN arg_pv4addrArray: pointer to V4 address array.  
 *
 * @Param:  - OUT none
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *      IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM
 *      IX_CC_ERROR_OOM
 * 
 */
ix_error _ix_cc_natpt_remove_v4pool(
                LPLIST arg_poolList,
                ix_uint32 arg_NumItems,
                ix_cc_natpt_v4addr* arg_pv4addrArray)
{
    LPLISTITR itr;
    ix_cc_natpt_v4PoolEntry* v4Entry;
    ix_uint8 found;

    while (arg_NumItems)
    {
        if (ListGetCount(arg_poolList))
        {
            itr = ListItrCreate(arg_poolList);
            if (itr == NULL)
            {
                return IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                            ("Failed to allocate memory for list iterator"));
            }

            found = 0;
            while ((v4Entry = (ix_cc_natpt_v4PoolEntry*) ListItrGetNext(itr)))
            {
                if ( *arg_pv4addrArray == v4Entry->v4addr)
                {
                    found = 1;
                    ListItrRemoveItem(itr);
                    ix_ossl_free(v4Entry);
                    /* continue with the next item */
                    break;
                }
            } /* while */

            ListItrDelete(itr);

            if ( ! found) 
            {
                return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM,
                        ("Failed to find the item in list"));
            }

        } /* if (ListGetCount(arg_poolList)) */
        else
        {
            return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM,
                        ("No more items in the stored list"));
        }

        arg_NumItems --;
        arg_pv4addrArray ++;

    } /* while (arg_NumItems) */

    return IX_SUCCESS;
}


/**
 * NAME: _ix_cc_natpt_get_v4pool
 *
 * DESCRIPTION: Retrieves the list of v4 addresses in the pool list.
 * 
 * @Param:  - IN arg_poolList: pointer to V4 address Pool List.  
 *
 * @Param:  - IN/OUT arg_NumItems: On input, contains number of items
 * the array can hold. On output, contains the actual number of items.
 *
 * @Param:  - OUT arg_pv4addrArray: pointer to V4 address array where
 * retrieved addresses are placed.  
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *      IX_CC_NATPT_ERROR_BUFFER_TOO_SMALL
 *      IX_CC_ERROR_OOM
 * 
 */
ix_error _ix_cc_natpt_get_v4pool(
                LPLIST arg_poolList,
                ix_uint32* arg_NumItems,
                ix_cc_natpt_v4addr* arg_pv4addrArray)
{
    LPLISTITR itr;
    ix_cc_natpt_v4PoolEntry* pPoolEntry;
    ix_uint32 givenNum = *arg_NumItems;

    if (ListGetCount(arg_poolList))
    {
        itr = ListItrCreate(arg_poolList);
        if (itr == NULL)
        {
            *arg_NumItems = ListGetCount(arg_poolList);
            return IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                            ("Failed to allocate memory for list iterator"));
        }

        *arg_NumItems = 0;
        while ((pPoolEntry = (ix_cc_natpt_v4PoolEntry*) ListItrGetNext(itr)))
        {
            if (*arg_NumItems == givenNum)
            {
                *arg_NumItems = ListGetCount(arg_poolList);
                ListItrDelete(itr);
                return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_BUFFER_TOO_SMALL,
                            ("More items exist"));
            }
            *arg_pv4addrArray = pPoolEntry->v4addr;

            (*arg_NumItems) ++;
            arg_pv4addrArray ++;
        } /* while */

        ListItrDelete(itr);
    }
    else
    {
        *arg_NumItems = 0;
    }

    return IX_SUCCESS;
}


/**
 * NAME: _ix_cc_natpt_alloc_ipv4addr
 *
 * DESCRIPTION: Allocates an IPv4 address (if available) for dynamic 
 * mapping.
 * 
 * @Param:  - IN arg_poolList: pointer to V4 address Pool List.  
 *
 * @Param:  - OUT arg_ipv4Addr: pointer to allocated V4 address.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *      IX_CC_NATPT_ERROR_BUFFER_TOO_SMALL
 *      IX_CC_ERROR_OOM
 * 
 */
ix_error _ix_cc_natpt_alloc_ipv4addr(
                LPLIST arg_poolList,
                ix_uint32* arg_ipv4Addr)
{
    LPLISTITR itr;
    ix_cc_natpt_v4PoolEntry* pPoolEntry;
    ix_uint8 found = 0;

    itr = ListItrCreate(arg_poolList);
    if (itr == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_OOR,
                                ("Failed to allocate memory for list iterator"));
    }

    while ((pPoolEntry = (ix_cc_natpt_v4PoolEntry*) ListItrGetNext(itr)))
    {
        if (pPoolEntry->isUsed == 0)
        {
            pPoolEntry->isUsed = 1;
            *arg_ipv4Addr = pPoolEntry->v4addr;
            found = 1;
            break;
        }
    } /* while */
    ListItrDelete(itr);

    if (found)
        return IX_SUCCESS;

    return IX_CC_ERROR_OOR;
}        


/**
 * NAME: _ix_cc_natpt_free_ipv4addr
 *
 * DESCRIPTION: De-allocates an IPv4 address so that it is available for
 * for dynamic mapping again.
 * 
 * @Param:  - IN arg_poolList: pointer to V4 address Pool List.  
 * @Param:  - IN arg_ipv4Addr: V4 address that is to be freed.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *      IX_CC_NATPT_ERROR_BUFFER_TOO_SMALL
 *      IX_CC_ERROR_OOM
 * 
 */
ix_error _ix_cc_natpt_free_ipv4addr(
                LPLIST arg_poolList,
                ix_uint32 arg_ipv4Addr)
{
    LPLISTITR itr;
    ix_cc_natpt_v4PoolEntry* pPoolEntry;
    ix_uint8 found = 0;

    itr = ListItrCreate(arg_poolList);
    if (itr == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_OOR,
                                ("Failed to allocate memory for list iterator"));
    }

    while ((pPoolEntry = (ix_cc_natpt_v4PoolEntry*) ListItrGetNext(itr)))
    {
        if ((pPoolEntry->v4addr == arg_ipv4Addr) &&
                (pPoolEntry->isUsed == 1))
        {
            pPoolEntry->isUsed = 0;
            found = 1;
            break;
        }
    } /* while */
    ListItrDelete(itr);

    if (found)
        return IX_SUCCESS;

    return IX_CC_NATPT_ERROR_INVALID_INPUT_PARAM;
}        


