/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.x for the IXP2xxx Network Processors
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      NATPT Core Component - message handler
 *
 * = FILENAME
 *      ix_cc_natpt_msg_handler.c
 *
 * = DESCRIPTION
 *      The file defines message handler interface functions.It implements 
 *      code to handle messages sent to the NATPT core component. 
 *      Some messages expect the core component to send a response, while
 *      others simply cause an addition/update of some internal data 
 *      structure. Each message is handled individually.
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      03/03/03
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_natpt_msg_handler.c,v 1.13 2003/05/23 23:30:40 sshegde Exp $";

/**
 * System defined include files
 */

#include "ix_cc_error.h"
#include "ix_ossl.h"
#include "ix_rm.h"
#include "ix_cc.h"
#include "ix_cc_macros.h"
#include "ix_cc_properties.h"
#include "bindings.h"
#include "cc/ix_cc_msup.h"
#include "cc/ix_cc_natpt.h"

/**
 * User defined include files
 */

#include "cc/natpt/internal/ix_cc_natpt_init.h"
#include "cc/natpt/internal/ix_cc_natpt_msg_handler.h"


/**
 * Preprocessor symbols and Macros used in this file
 */

/* Flag fornot to send report to msg support lib */
#define IX_CC_NATPT_SEND_NO_RESPONSE    1 

/**
 * Type definitions whose scope is limited to this file
 */

/* Return data for getting v4/v6 static map */
typedef struct ix_s_cc_natpt_static_map_data
{
    ix_uint32 numEntries;
    ix_cc_natpt_static_v6v4map* paInfo;
} ix_cc_natpt_static_map_data;


/* Return data for getting v4 address pool */
typedef struct ix_s_cc_natpt_get_v4addr_pool_data
{
    ix_uint32 numEntries;
    ix_cc_natpt_v4addr* paInfo;
} ix_cc_natpt_get_v4addr_pool_data;


/* Return data for getting active sessions */
typedef struct ix_s_cc_natpt_get_active_sessions_data
{
    ix_uint32 numEntries;
    ix_cc_natpt_session* paInfo;
} ix_cc_natpt_get_active_sessions_data;


/* 
 * This structure defines all types of return messages that can 
 * be returned from the message handler (If the message type requires
 * a reply to be sent).
 */
typedef union ix_u_cc_natpt_msg_handler_return_data
{
    ix_cc_natpt_config_params configParamsData;
    ix_cc_natpt_static_map_data staticMapData;
    ix_cc_natpt_get_v4addr_pool_data v4addrPoolData;
    ix_cc_natpt_get_active_sessions_data activeSessionsData;
    ix_uint16 unused;
} ix_cc_natpt_msg_handler_return_data;


/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */

/** 
 * Extern function prototypes.
 */


/** 
 * Static function prototypes.
 */


/** 
 * Function definitions.
 */


/**
 * NAME: ix_cc_natpt_msg_handler
 *
 * DESCRIPTION: This primitive is the message handler function for NATPT
 * core component. The NATPT core component receives messages 
 * from other core component through this message handler function and 
 * internally it calls appropriate library function to process the message. 
 * This message handler shall be used to update dynamic properties.
 * 
 * @Param:  - IN arg_hDataToken - buffer handle embedding information for 
 *            the message passed in arg_UserData.
 *
 * @Param:  - IN  arg_UserData - message type
 *
 * @Param:  - IN - arg_pContext: pointer to NATPT core 
 *                 component context that will be passed to the core component 
 *                 when a message arrives. This context was defined by the 
 *                 core component and passed to core components infrastructure 
 *                 through ix_cc_natpt_init function.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - null argument
 * IX_CC_ERROR_UNDEFINED_MSG - un supported message
 * IX_CC_NATPT_ERROR_MSG_LIBRARY - error from message support library
 * IX_CC_NATPT_ERROR_BUFFER_FREE - error from RM for freeing buffer
 */

ix_error ix_cc_natpt_msg_handler(ix_buffer_handle arg_hDataToken,  
                                 ix_uint32 arg_UserData,
                                 void* arg_pContext
                                )
{
    ix_error err = IX_SUCCESS;

    void *pContext = NULL;
    void *pMsg = NULL;
    void *pRetMsg = NULL;

    ix_cc_natpt_msg_handler_return_data msgRetData;
    ix_uint32 msgRetLength = 0;
    ix_uint32 noResponse = 0;  

    /* Validate arg_pContext */
    if(arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                              ("arg_pContext is NULL"));
    }/* end if(arg_pContext == NULL) */

    if((arg_UserData < IX_CC_COMMON_MSG_ID_FIRST) ||
       (arg_UserData >= IX_CC_NATPT_MSG_LAST))
    {
    
        return IX_ERROR_WARNING(IX_CC_ERROR_UNDEFINED_MSG,
                                               ("Invalid Message Type"));
    }/* end if */

    /* Extract the input argument, pMsg could be NULL after this */
    IX_MSUP_EXTRACT_MSG(arg_hDataToken, &pMsg, &pContext);

    
    /* Check for route messages, next hop message, icmp messages 
     */
    switch(arg_UserData)
    {
        case IX_CC_NATPT_MSG_SET_CONFIG_PARAM:
        {
            ix_cc_natpt_config_params* pData = 
                            (ix_cc_natpt_config_params*) pMsg;
            err = ix_cc_natpt_set_config_parameters(pData, arg_pContext);
            break;
        }
        case IX_CC_NATPT_MSG_GET_CONFIG_PARAM:
        {
            err = ix_cc_natpt_get_config_parameters(
                        &msgRetData.configParamsData, arg_pContext);
            pRetMsg = &msgRetData.configParamsData;
            msgRetLength = sizeof(msgRetData.configParamsData);
            break;
        }
        case IX_CC_NATPT_MSG_ADD_STATIC_MAP:
        {
            ix_cc_natpt_static_v6v4map* pData = 
                            (ix_cc_natpt_static_v6v4map*) pMsg;
            err = ix_cc_natpt_add_static_mapping(pData, arg_pContext);
            break;
        }
        case IX_CC_NATPT_MSG_REM_STATIC_MAP:
        {
            ix_cc_natpt_static_v6v4map* pData = 
                            (ix_cc_natpt_static_v6v4map*) pMsg;
            err = ix_cc_natpt_remove_static_mapping(pData, arg_pContext);
            break;
        }
        case IX_CC_NATPT_MSG_GET_STATIC_MAP:
        {
            ix_uint32 size = IX_CC_NATPT_STATICMAP_SIZE_HINT;
            do
            {
                msgRetData.staticMapData.numEntries = size;
                size *= sizeof(ix_cc_natpt_static_v6v4map);
                msgRetData.staticMapData.paInfo =
                    (ix_cc_natpt_static_v6v4map *) ix_ossl_malloc(size);

                if(msgRetData.staticMapData.paInfo != NULL)
                {
                    err = ix_cc_natpt_get_static_mapping(
                        &msgRetData.staticMapData.numEntries,
                        msgRetData.staticMapData.paInfo,
                        arg_pContext);

                    err = IX_ERROR_GET_CODE(err);
                    if (err == IX_CC_NATPT_ERROR_BUFFER_TOO_SMALL)
                    {
                        size = msgRetData.staticMapData.numEntries;
                        ix_ossl_free(msgRetData.staticMapData.paInfo);
                    }
                }
                else
                {
                    err = IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                                           ("No memory to get static maps"));

                }
            } while (err == IX_CC_NATPT_ERROR_BUFFER_TOO_SMALL);

            pRetMsg = (void*)&msgRetData.staticMapData;
            msgRetLength = sizeof(ix_cc_natpt_static_map_data);
            break;
        }
        case IX_CC_NATPT_MSG_ADD_V4ADDR_POOL:
        {
            ix_cc_natpt_v4addr_pool_data* pData = 
                            (ix_cc_natpt_v4addr_pool_data*) pMsg;
            err = ix_cc_natpt_add_v4addr_pool(pData->numItems, 
                            pData->pv4addrArray, arg_pContext);
            ix_ossl_free(pData->pv4addrArray);
            break;
        }
        case IX_CC_NATPT_MSG_REM_V4ADDR_POOL:
        {
            ix_cc_natpt_v4addr_pool_data* pData = 
                            (ix_cc_natpt_v4addr_pool_data*) pMsg;
            err = ix_cc_natpt_remove_v4addr_pool(pData->numItems, 
                            pData->pv4addrArray, arg_pContext);
            ix_ossl_free(pData->pv4addrArray);
            break;
        }
        case IX_CC_NATPT_MSG_GET_V4ADDR_POOL:
        {
            ix_uint32 size = IX_CC_NATPT_V4ADDRPOOL_SIZE_HINT;
            do
            {
                msgRetData.v4addrPoolData.numEntries = size;
                size *= sizeof(ix_cc_natpt_v4addr);
                msgRetData.v4addrPoolData.paInfo =
                    (ix_cc_natpt_v4addr *) ix_ossl_malloc(size);

                if(msgRetData.v4addrPoolData.paInfo != NULL)
                {
                    err = ix_cc_natpt_get_v4addr_pool(
                        &msgRetData.v4addrPoolData.numEntries,
                        msgRetData.v4addrPoolData.paInfo,
                        arg_pContext);

                    err = IX_ERROR_GET_CODE(err);
                    if (err == IX_CC_NATPT_ERROR_BUFFER_TOO_SMALL)
                    {
                        size = msgRetData.v4addrPoolData.numEntries;
                        ix_ossl_free(msgRetData.v4addrPoolData.paInfo);
                    }
                }
                else
                {
                    err = IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                                           ("No memory to get static maps"));

                }
            } while (err == IX_CC_NATPT_ERROR_BUFFER_TOO_SMALL);

            pRetMsg = (void*)&msgRetData.v4addrPoolData;
            msgRetLength = sizeof(ix_cc_natpt_get_v4addr_pool_data);
            break;
        }
        case IX_CC_NATPT_MSG_SET_NAPT_MODE:
            break;

        case IX_CC_NATPT_MSG_ADD_V4V6_PORT_MAP:
            break;

        case IX_CC_NATPT_MSG_REM_V4V6_PORT_MAP:
            break;

        case IX_CC_NATPT_MSG_GET_V4V6_PORT_MAP:
            break;

        case IX_CC_NATPT_MSG_GET_ACTIVE_SESSIONS:
        {
            ix_uint32 size = IX_CC_NATPT_ACTIVE_SESSIONS_SIZE_HINT;
            do
            {
                msgRetData.activeSessionsData.numEntries = size;
                size *= sizeof(ix_cc_natpt_session);
                msgRetData.activeSessionsData.paInfo =
                    (ix_cc_natpt_session *) ix_ossl_malloc(size);

                if(msgRetData.activeSessionsData.paInfo != NULL)
                {
                    err = ix_cc_natpt_get_active_sessions(
                        &msgRetData.activeSessionsData.numEntries,
                        msgRetData.activeSessionsData.paInfo,
                        arg_pContext);

                    if (err == IX_CC_NATPT_ERROR_BUFFER_TOO_SMALL)
                    {
                        size = msgRetData.activeSessionsData.numEntries;
                        ix_ossl_free(msgRetData.activeSessionsData.paInfo);
                    }
                }
                else
                {
                    err = IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                                           ("No memory to get static maps"));

                }
            } while (err == IX_CC_NATPT_ERROR_BUFFER_TOO_SMALL);

            pRetMsg = (void*)&msgRetData.activeSessionsData;
            msgRetLength = sizeof(ix_cc_natpt_get_active_sessions_data);
            break;
        }

        default:

           if(pMsg != NULL)
           {
               ix_ossl_free(pMsg);
           }

           IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hDataToken),
                                        IX_CC_NATPT_ERROR_BUFFER_FREE, 0);  

           return IX_ERROR_WARNING(IX_CC_ERROR_UNDEFINED_MSG,
                                     ("Invalid Message Type"));
           break;
    } /* end switch(arg_UserData) */

    if(pMsg != NULL)
    {
        ix_ossl_free(pMsg);
    }

    /* 
     * Note that some messages DO NOT require a response to be sent to the
     * sender of the message. In this case, simply remove the message buffer
     * from the queue and return it to the free list and return 
     */

    if (noResponse == IX_CC_NATPT_SEND_NO_RESPONSE)
    {
        IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hDataToken),
                                          IX_CC_NATPT_ERROR_BUFFER_FREE, 0);
        return IX_SUCCESS;
    }/* end if no need to send message to msg support lib*/


    /* 
     * Send return message, length and retrieved context 
     * to message support library. The message length and the actual
     * message were built appropriately in each case statement.
     */

    err = ix_cc_msup_send_reply_msg(pContext, pRetMsg, msgRetLength, err);
    if(err != IX_SUCCESS)
    {
        err = IX_ERROR_WARNING(IX_CC_NATPT_ERROR_MSG_LIBRARY,
                                   ("Send reply message failed"));
    }/* endif (err != IX_SUCCESS) */

    IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hDataToken),
                                              IX_CC_NATPT_ERROR_BUFFER_FREE, 0);
    return err;

} /* end ix_cc_natpt_msg_handler */


/**
 * NAME: _ix_cc_natpt_cb_genaral
 *
 * DESCRIPTION: This is a general callback function. The message support
 * library  calls this function to return the status of the requested
 * asynchronous operation.
 *
 * @Param:  - IN  arg_Result: result of the asynchronous operation
 * @Param:  - IN  arg_pContext: context provided by caller
 * @Param:  - IN  arg_pMsg: return message
 * @Param:  - IN  arg_MsgLength: length of the message
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_NATPT_ERROR_CALLBACK - user callback returns error message
 * 
 */

ix_error _ix_cc_natpt_cb_general(
                ix_error arg_Result,
                void* arg_pContext,
                void* arg_pMsg,
                ix_uint32 arg_MsgLength)
{
    ix_error err = IX_SUCCESS;
 
    ix_cc_natpt_general_callback_data* pCbData = NULL;
 
    pCbData = (ix_cc_natpt_general_callback_data*)arg_pContext;
   
    if(pCbData != NULL)
    {
        err = pCbData->pUserCallback(arg_Result,pCbData->pUserContext);
        ix_ossl_free(arg_pContext);
        if(err != IX_SUCCESS)
        {
            return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_CALLBACK,
                                      ("general callback failed"));
        }/* end if (err != IX_SUCCESS) */
    }/* end if(pCbData != NULL) */
       
    return IX_SUCCESS;
}/* _ix_cc_natpt_cb_general */


/**
 * NAME: _ix_cc_natpt_cb_get_config_parameters
 *
 * DESCRIPTION: This is a callback function for getting configuration
 * parameters. The message support library  calls this function to return 
 * the status of the requested asynchronous operation.
 *
 * @Param:  - IN  arg_Result: result of the asynchronous operation
 * @Param:  - IN  arg_pContext: context provided by caller
 * @Param:  - IN  arg_pMsg: return message
 * @Param:  - IN  arg_MsgLength: length of the message
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_NATPT_ERROR_CALLBACK - user callback returns error message
 * 
 */

ix_error _ix_cc_natpt_cb_get_config_parameters(
                ix_error arg_Result,
                void* arg_pContext,
                void* arg_pMsg,
                ix_uint32 arg_MsgLength)
{
    ix_error err = IX_SUCCESS;
 
    ix_cc_natpt_get_config_parameters_callback_data* pCbData;
    ix_cc_natpt_config_params* pMsgData;

    if (arg_pContext != NULL)
    {
        pCbData = (ix_cc_natpt_get_config_parameters_callback_data*)
                                arg_pContext;
        pMsgData = (ix_cc_natpt_config_params*)arg_pMsg;
    
        err = pCbData->pUserCallback(arg_Result,pCbData->pUserContext,
                        pMsgData);
        ix_ossl_free(arg_pContext);
        if(err != IX_SUCCESS)
        {
            return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_CALLBACK,
                                      ("lookup callback failed"));
        }/* end if (err != IX_SUCCESS) */
    }/* end if(pCbData != NULL)*/
       
    return IX_SUCCESS;
}

 
/**
 * NAME: _ix_cc_natpt_cb_get_static_mapping
 *
 * DESCRIPTION: This is a callback function for getting current static
 * mapping data. The message support library  calls this function to 
 * return the status of the requested asynchronous operation.
 *
 * @Param:  - IN  arg_Result: result of the asynchronous operation
 * @Param:  - IN  arg_pContext: context provided by caller
 * @Param:  - IN  arg_pMsg: return message
 * @Param:  - IN  arg_MsgLength: length of the message
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_NATPT_ERROR_CALLBACK - user callback returns error message
 * 
 */

ix_error _ix_cc_natpt_cb_get_static_mapping(
                ix_error arg_Result,
                void* arg_pContext,
                void* arg_pMsg,
                ix_uint32 arg_MsgLength)
{
    ix_error err = IX_SUCCESS;
 
    ix_cc_natpt_get_static_mapping_callback_data* pCbData;
    ix_cc_natpt_static_map_data* pMsgData;

    if (arg_pContext != NULL)
    {
        pCbData = (ix_cc_natpt_get_static_mapping_callback_data*)arg_pContext;
        pMsgData = (ix_cc_natpt_static_map_data*)arg_pMsg;
    
        err = pCbData->pUserCallback(
            arg_Result,pCbData->pUserContext,
            pMsgData->numEntries,pMsgData->paInfo);

        ix_ossl_free(arg_pContext);
        ix_ossl_free(pMsgData->paInfo);
        if(err != IX_SUCCESS)
        {
            return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_CALLBACK,
                                      ("lookup callback failed"));
        }/* end if (err != IX_SUCCESS) */
    }/* end if(pCbData != NULL)*/
       
    return IX_SUCCESS;
}

 
/**
 * NAME: _ix_cc_natpt_cb_get_v4addr_pool
 *
 * DESCRIPTION: This is a callback function for getting current IPv4
 * address pool. The message support library  calls this function to return 
 * the status of the requested asynchronous operation.
 *
 * @Param:  - IN  arg_Result: result of the asynchronous operation
 * @Param:  - IN  arg_pContext: context provided by caller
 * @Param:  - IN  arg_pMsg: return message
 * @Param:  - IN  arg_MsgLength: length of the message
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_NATPT_ERROR_CALLBACK - user callback returns error message
 * 
 */

ix_error _ix_cc_natpt_cb_get_v4addr_pool(
                ix_error arg_Result,
                void* arg_pContext,
                void* arg_pMsg,
                ix_uint32 arg_MsgLength)
{
    ix_error err = IX_SUCCESS;
 
    ix_cc_natpt_get_v4addr_pool_callback_data* pCbData;
    ix_cc_natpt_get_v4addr_pool_data* pMsgData;

    if (arg_pContext != NULL)
    {
        pCbData = (ix_cc_natpt_get_v4addr_pool_callback_data*)arg_pContext;
        pMsgData = (ix_cc_natpt_get_v4addr_pool_data*)arg_pMsg;
    
        err = pCbData->pUserCallback(
            arg_Result,pCbData->pUserContext,
            pMsgData->numEntries,pMsgData->paInfo);

        ix_ossl_free(arg_pContext);
        ix_ossl_free(pMsgData->paInfo);
        if(err != IX_SUCCESS)
        {
            return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_CALLBACK,
                                      ("lookup callback failed"));
        }/* end if (err != IX_SUCCESS) */
    }/* end if(pCbData != NULL)*/
       
    return IX_SUCCESS;
}

 
/**
 * NAME: _ix_cc_natpt_cb_get_active_sessions
 *
 * DESCRIPTION: This is a callback function for getting active NATPT 
 * sessions. The message support library  calls this function to return 
 * the status of the requested asynchronous operation.
 *
 * @Param:  - IN  arg_Result: result of the asynchronous operation
 * @Param:  - IN  arg_pContext: context provided by caller
 * @Param:  - IN  arg_pMsg: return message
 * @Param:  - IN  arg_MsgLength: length of the message
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_NATPT_ERROR_CALLBACK - user callback returns error message
 * 
 */

ix_error _ix_cc_natpt_cb_get_active_sessions(
                ix_error arg_Result,
                void* arg_pContext,
                void* arg_pMsg,
                ix_uint32 arg_MsgLength)
{
    ix_error err = IX_SUCCESS;
 
    ix_cc_natpt_get_active_sessions_callback_data* pCbData;
    ix_cc_natpt_get_active_sessions_data* pMsgData;

    if (arg_pContext != NULL)
    {
        pCbData = (ix_cc_natpt_get_active_sessions_callback_data*)arg_pContext;
        pMsgData = (ix_cc_natpt_get_active_sessions_data*)arg_pMsg;
    
        err = pCbData->pUserCallback(
            arg_Result,pCbData->pUserContext,
            pMsgData->numEntries,pMsgData->paInfo);

        ix_ossl_free(arg_pContext);
        ix_ossl_free(pMsgData->paInfo);
        if(err != IX_SUCCESS)
        {
            return IX_ERROR_WARNING(IX_CC_NATPT_ERROR_CALLBACK,
                                      ("lookup callback failed"));
        }/* end if (err != IX_SUCCESS) */
    }/* end if(pCbData != NULL)*/
       
    return IX_SUCCESS;
}

 
