/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.x for the IXP2xxx Network Processors
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      NATPT Core Component - Tables for microblock
 *
 * = FILENAME
 *      ix_cc_natpt_nat_table.c
 *
 * = DESCRIPTION
 *      The file contains the functions and macros needed to manage the
 *      NATPT tables that are shared by both CC and microblock.
 *
 * = CREATION TIME
 *      03/03/03
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_natpt_nat_table.c,v 1.20 2003/10/31 19:27:32 ktseng Exp $";


/**
 * System defined include files
 */
#include "ix_cc_error.h"
#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_netmacros.h"
#include "ix_cc.h"
#include "ix_cc_macros.h"
#include "ix_config.h"
#include "bindings.h"
#include "cc/ix_cc_natpt.h"

/**
 * User defined include files
 */
#include "cc/natpt/internal/ix_cc_natpt_init.h"
#include "cc/natpt/internal/ix_cc_natpt_utils.h"
#include "cc/natpt/internal/ix_cc_natpt_nat_table.h"
#include "cc/natpt/internal/ix_cc_natpt_lists.h"


/**
 * Preprocessor symbols and Macros used in this file
 */

#define IX_CC_NATPT_EVENT_PRIORITY 1

/**
 * Type definitions whose scope is limited to this file
 */

/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */

/** 
 * Extern function prototypes.
 */


/** 
 * Static function prototypes.
 */


/** 
 * Function definitions.
 */


/**
 * NAME: _ix_cc_natpt_add_nat_table_entry
 *
 * DESCRIPTION: This function is used to add a new NAT Table Entry.
 *           
 * @Param:  - IN arg_tableBase - pointer to NAT Table
 * @Param:  - IN arg_indexPool - Index pool for NAT Table
 * @Param:  - IN arg_natEntry - Pointer to NAT Table Entry
 *
 * @Param:  - OUT arg_index - pointer to Index at which the Entry was added.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_FULL - Out of memory
 * 
 */
ix_error _ix_cc_natpt_add_nat_table_entry(
                                ix_uint32* arg_tableBase,
                                ix_cc_index_pool arg_indexPool,
                                ix_cc_natpt_nat_entry* arg_natEntry,
                                ix_uint32* arg_index)
{
    ix_uint32 index;

    /* Allocate an index */
    IX_ERROR_CR(
                ix_cc_index_pool_alloc(arg_indexPool, &index));

    /* Write the entry to shared memory */
    _WRITE_NAT_TABLE_ENTRY(arg_tableBase, index, *arg_natEntry);

    *arg_index = index;

    return IX_SUCCESS;
}


/**
 * NAME: _ix_cc_natpt_find_nat_table_entry
 *
 * DESCRIPTION: This function is used to find an existing NAT Table Entry.
 *           
 * @Param:  - IN arg_tableBase - pointer to NAT Table
 * @Param:  - IN arg_indexPool - Index pool for NAT Table
 * @Param:  - IN arg_natEntry - Pointer to NAT Table Entry
 *
 * @Param:  - OUT arg_index - pointer to Index at which the Entry was added.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_ENTRY_NOT_FOUND - Entry Not found.
 * 
 */
ix_error _ix_cc_natpt_find_nat_table_entry(
                                ix_uint32* arg_tableBase,
                                ix_cc_index_pool arg_indexPool,
                                ix_cc_natpt_nat_entry* arg_natEntry,
                                ix_uint32* arg_index)
{
    ix_uint32 index;
    ix_cc_natpt_nat_entry natEntry;
    ix_error err = IX_SUCCESS;

    err = ix_cc_index_pool_get_first(arg_indexPool, &index);
    while (err == IX_SUCCESS)
    {
        /* Read the entry */
        _READ_NAT_TABLE_ENTRY(arg_tableBase, index, natEntry);

        if ( (arg_natEntry->v4Addr == natEntry.v4Addr) &&
            NATPT_IS_IPV6_ADDRESS_EQUAL(&arg_natEntry->v6Addr, &natEntry.v6Addr) )
        {
            
            arg_natEntry->isStatic = natEntry.isStatic;
            arg_natEntry->isUsed = natEntry.isUsed;
            *arg_index = index;
            return IX_SUCCESS; /* found */
        }

        err = ix_cc_index_pool_get_next(arg_indexPool, index, &index);
    } /* while (err == IX_SUCCESS) */

    return IX_CC_ERROR_ENTRY_NOT_FOUND; /* not found */
}


/**
 * NAME: _ix_cc_natpt_del_nat_table_entry
 *
 * DESCRIPTION: This function is used to delete an existing NAT Table Entry.
 *           
 * @Param:  - IN arg_tableBase - pointer to NAT Table
 * @Param:  - IN arg_indexPool - Index pool for NAT Table
 * @Param:  - IN arg_natEntry - Pointer to NAT Table Entry
 *
 * @Param:  - OUT None
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_ENTRY_NOT_FOUND - Entry Not found.
 * 
 */
ix_error _ix_cc_natpt_del_nat_table_entry(
                                ix_uint32* arg_tableBase,
                                ix_cc_index_pool arg_indexPool,
                                ix_cc_natpt_nat_entry* arg_natEntry)
{
    ix_uint32 index;
    ix_cc_natpt_nat_entry natEntry;
    ix_error err = IX_SUCCESS;

    err = ix_cc_index_pool_get_first(arg_indexPool, &index);
    while (err == IX_SUCCESS)
    {
        /* Read the entry */
        _READ_NAT_TABLE_ENTRY(arg_tableBase, index, natEntry);

        if ( (arg_natEntry->v4Addr == natEntry.v4Addr) &&
            NATPT_IS_IPV6_ADDRESS_EQUAL(&arg_natEntry->v6Addr, 
                    &natEntry.v6Addr) )
        {
            /* erase the entry */
            _CLEAR_NAT_TABLE_ENTRY(arg_tableBase, index);

            /* free the index */
            IX_ERROR_CR(ix_cc_index_pool_free(arg_indexPool, index));

            return IX_SUCCESS;
        }

        err = ix_cc_index_pool_get_next(arg_indexPool, index, &index);
    } /* while (err == IX_SUCCESS) */

    return IX_CC_ERROR_ENTRY_NOT_FOUND; /* not found */
}


/**
 * NAME: _ix_cc_natpt_hash_v4addr
 *
 * DESCRIPTION: This function is used to calculate the hash value of 
 * an IPv4 address.
 *           
 * @Param:  - IN v4addr - IPv4 addresss
 *
 * @Param:  - OUT None
 *
 * @Return: 7 bit hash value in 8 bit variable
 *
 * Error Codes:
 * 
 */
PRIVATE ix_uint8 _ix_cc_natpt_hash_v4addr(ix_uint32 v4addr)
{
    ix_error err;
    ix_hash_48 hash48;

    hash48.m_LW0 = v4addr;
    hash48.m_LW1 = 0;

    err = ix_rm_hash_48_hash(&hash48);
    if (err != IX_SUCCESS)
    {
#if defined(NATPT_DEBUG)
        ix_ossl_message_log("ix_rm_hash_48_hash "
                "returned error (0x%lx)\n", err);
#endif /* NATPT_DEBUG */
    }

    return ((ix_uint8) (hash48.m_LW0 & 0x0000007f));
}


/**
 * NAME: _ix_cc_natpt_hash_v6addr
 *
 * DESCRIPTION: This function is used to calculate the hash value of 
 * an IPv6 address.
 *           
 * @Param:  - IN v6addr - IPv6 addresss
 *
 * @Param:  - OUT None
 *
 * @Return: 7 bit hash value in 8 bit variable
 *
 * Error Codes:
 * 
 */
PRIVATE ix_uint8 _ix_cc_natpt_hash_v6addr(ix_uint128 v6addr)
{
    ix_error err;
    ix_hash_128 hash128;

    hash128.m_LW0 = v6addr.aUint32[3];
    hash128.m_LW1 = v6addr.aUint32[2];
    hash128.m_LW2 = v6addr.aUint32[1];
    hash128.m_LW3 = v6addr.aUint32[0];

    err = ix_rm_hash_128_hash(&hash128);
    if (err != IX_SUCCESS)
    {
#if defined(NATPT_DEBUG)
        ix_ossl_message_log("ix_rm_hash_128_hash "
                "returned error (0x%lx)\n", err);
#endif /* NATPT_DEBUG */
    }

    return ((ix_uint8) (hash128.m_LW0 & 0x0000007f));
}


/**
 * NAME: _ix_cc_natpt_add_collision_entry
 *
 * DESCRIPTION: This function is used to add a NAT Hash Entry into 
 * collision table.
 *           
 * @Param:  - IN arg_pContext - pointer to NATPT Context
 * @Param:  - IN natIndex - Index at which NAT Table entry was added
 *
 * @Param:  - OUT location - address where NAT Hash Entry was written
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_FULL - Out of Memory
 * 
 */
PRIVATE ix_error _ix_cc_natpt_add_collision_entry(
                                ix_cc_natpt_context* arg_pContext,
                                ix_uint32 arg_natIndex,
                                ix_uint32* arg_pColIndex)
{
    ix_uint32 colIndex;
    ix_uint32 readIndex;
    ix_cc_natpt_nat_hash_entry hashEntry;

    IX_ERROR_CR(
            ix_cc_index_pool_alloc(
                    arg_pContext->natCollisionTableIndexPool, &colIndex));

    hashEntry.index = arg_natIndex;
    hashEntry.next = IX_CC_NATPT_UNUSED_ENTRY_FILL;

    _WRITE_NAT_HASH_ENTRY(arg_pContext->pMbNatCollisionTableBase,
                colIndex, hashEntry);

    if (*arg_pColIndex != IX_CC_NATPT_UNUSED_ENTRY_FILL)
    {
        do {
            readIndex = *arg_pColIndex;
            _READ_NAT_HASH_ENTRY(arg_pContext->pMbNatCollisionTableBase,
                    readIndex, hashEntry);
            *arg_pColIndex = hashEntry.next;
        } while (*arg_pColIndex != IX_CC_NATPT_UNUSED_ENTRY_FILL);

        hashEntry.next = colIndex;

        _WRITE_NAT_HASH_ENTRY(arg_pContext->pMbNatCollisionTableBase,
                readIndex, hashEntry);
    }
    else
    {
        *arg_pColIndex = colIndex;
    }

    return IX_SUCCESS;
}


/**
 * NAME: _ix_cc_natpt_del_collision_entry
 *
 * DESCRIPTION: This function is used to delete a NAT Hash Entry in
 * collision table.
 *           
 * @Param:  - IN arg_pContext - pointer to NATPT Context
 * @Param:  - IN arg_natLocation - Address of NAT Table Entry
 * @Param:  - IN arg_colLocation - Address of NAT Hash Entry in 
 * Collision Table 
 *
 * @Param:  - OUT arg_newColLocation - address of NAT Hash Entry in
 * Collision Table where next Entry exists
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_ENTRY_NOT_FOUND - Entry Not found.
 * 
 */
PRIVATE ix_error _ix_cc_natpt_del_collision_entry(
                                ix_cc_natpt_context* arg_pContext,
                                ix_uint32 arg_natIndex,
                                ix_uint32 arg_colIndex,
                                ix_uint32* arg_newColIndex)
{
    ix_cc_natpt_nat_hash_entry hashEntry;
    ix_cc_natpt_nat_hash_entry prevEntry;
    ix_uint32 prevIndex;

    /*
        read colLocation
        if there is a match for nat location, then set newColLoc to next ptr
        and clear out the entry, remember to update the previous entry's next
        pointer with the pointer of the next entry if any.
    */

    prevIndex = IX_CC_NATPT_UNUSED_ENTRY_FILL;

    do {
        _READ_NAT_HASH_ENTRY(arg_pContext->pMbNatCollisionTableBase, 
                arg_colIndex, hashEntry);
    
        if (hashEntry.index == arg_natIndex)
        {
            if (prevIndex != IX_CC_NATPT_UNUSED_ENTRY_FILL)
            {
                /* Update the previous entry for next pointer */
                _READ_NAT_HASH_ENTRY(arg_pContext->pMbNatCollisionTableBase, 
                        prevIndex, prevEntry);
                prevEntry.next = hashEntry.next;
                _WRITE_NAT_HASH_ENTRY(arg_pContext->pMbNatCollisionTableBase, 
                        prevIndex, prevEntry);
            }
            else
            {
                /* send back the location for updating main table next ptr */
                *arg_newColIndex = hashEntry.next;
            }

            _CLEAR_NAT_HASH_ENTRY(arg_pContext->pMbNatCollisionTableBase, 
                    arg_colIndex);

            IX_ERROR_CR(
                ix_cc_index_pool_free(
                     arg_pContext->natCollisionTableIndexPool, arg_colIndex));

            return IX_SUCCESS;
        }
        else
        {
            prevIndex = arg_colIndex;
            arg_colIndex = hashEntry.next;
        }
        
    } while (hashEntry.next != IX_CC_NATPT_UNUSED_ENTRY_FILL);

    return IX_CC_ERROR_ENTRY_NOT_FOUND;
}


/**
 * NAME: _ix_cc_natpt_add_nat_hash_table_entries
 *
 * DESCRIPTION: This function is used to add a NAT Hash Entry into
 * both v6-to-v4 and v4-to-v6 Hash Tables .
 *           
 * @Param:  - IN arg_pContext - pointer to NATPT Context
 * @Param:  - IN arg_natEntry - pointer to NAT Table Entry
 * @Param:  - IN arg_index - Index where NAT Table Entry was added
 *
 * @Param:  - OUT None
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_FULL - Out of Memory
 * 
 */
ix_error _ix_cc_natpt_add_nat_hash_table_entries(
                                ix_cc_natpt_context* arg_pContext,
                                ix_cc_natpt_nat_entry* arg_natEntry,
                                ix_uint32 arg_index)
{
    ix_uint8 hashValue;
    ix_cc_natpt_nat_hash_entry hashEntry;
    ix_uint32 colIndex;

    /* Add to V6-V4 hash table */
    /* Hash on v6 src address */
    /* Read the entry at that index. If there is no entry add it, otherwise, 
       If the next entry is 0 then add the entry in the collision table, and
       update the entry in the main table. Otherwise read the next entry 
       pointed to by the next pointer in the collision table. Follow the 
       chain in the collision table until it ends. Alloc another index from 
       pool andadd the entry */

    hashValue = _ix_cc_natpt_hash_v6addr(arg_natEntry->v6Addr);

    _READ_NAT_HASH_ENTRY(arg_pContext->pMbNatV6V4HashTableBase,
                    hashValue, hashEntry);

    if ( hashEntry.index == IX_CC_NATPT_UNUSED_ENTRY_FILL)
    {
        hashEntry.index = arg_index;
        hashEntry.next = IX_CC_NATPT_UNUSED_ENTRY_FILL;

        _WRITE_NAT_HASH_ENTRY(arg_pContext->pMbNatV6V4HashTableBase,
                    hashValue, hashEntry);
    }
    else if (hashEntry.next == IX_CC_NATPT_UNUSED_ENTRY_FILL)
    {
        colIndex = IX_CC_NATPT_UNUSED_ENTRY_FILL;

        IX_ERROR_CR(
            _ix_cc_natpt_add_collision_entry(arg_pContext, 
                        arg_index, &colIndex));

        hashEntry.next = colIndex;

        _WRITE_NAT_HASH_ENTRY(arg_pContext->pMbNatV6V4HashTableBase,
                    hashValue, hashEntry);
    }
    else
    {
        colIndex = hashEntry.next;

        IX_ERROR_CR(
            _ix_cc_natpt_add_collision_entry(arg_pContext, 
                        arg_index, &colIndex));
    }

    /* In a similar way Add to V4-V6 hash table */
    /* Hash on v4 dest address */
    hashValue = _ix_cc_natpt_hash_v4addr(arg_natEntry->v4Addr);

    _READ_NAT_HASH_ENTRY(arg_pContext->pMbNatV4V6HashTableBase,
                    hashValue, hashEntry);

    if ( hashEntry.index == IX_CC_NATPT_UNUSED_ENTRY_FILL)
    {
        hashEntry.index = arg_index;
        hashEntry.next = IX_CC_NATPT_UNUSED_ENTRY_FILL;

        _WRITE_NAT_HASH_ENTRY(arg_pContext->pMbNatV4V6HashTableBase,
                    hashValue, hashEntry);
    }
    else if (hashEntry.next == IX_CC_NATPT_UNUSED_ENTRY_FILL)
    {
        colIndex = IX_CC_NATPT_UNUSED_ENTRY_FILL;

        IX_ERROR_CR(
            _ix_cc_natpt_add_collision_entry(arg_pContext, 
                        arg_index, &colIndex));

        hashEntry.next = colIndex;

        _WRITE_NAT_HASH_ENTRY(arg_pContext->pMbNatV4V6HashTableBase,
                    hashValue, hashEntry);
    }
    else
    {
        colIndex = hashEntry.next;

        IX_ERROR_CR(
            _ix_cc_natpt_add_collision_entry(arg_pContext, 
                        arg_index, &colIndex));
    }

    return IX_SUCCESS;
}


/**
 * NAME: _ix_cc_natpt_del_nat_hash_table_entries
 *
 * DESCRIPTION: This function is used to remove a NAT Hash Entry in
 * both v6-to-v4 and v4-to-v6 Hash Tables .
 *           
 * @Param:  - IN arg_pContext - pointer to NATPT Context
 * @Param:  - IN arg_natEntry - pointer to NAT Table Entry
 * @Param:  - IN arg_index - Index where NAT Table Entry was added
 *
 * @Param:  - OUT None
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_ENTRY_NOT_FOUND - Entry Not found.
 * 
 */
ix_error _ix_cc_natpt_del_nat_hash_table_entries(
                                ix_cc_natpt_context* arg_pContext,
                                ix_cc_natpt_nat_entry* arg_natEntry,
                                ix_uint32 arg_index)
{
    ix_uint8 hashValue;
    ix_cc_natpt_nat_hash_entry hashEntry;
    ix_cc_natpt_nat_hash_entry colHashEntry;
    ix_uint32 colIndex;

    hashValue = _ix_cc_natpt_hash_v6addr(arg_natEntry->v6Addr);

    _READ_NAT_HASH_ENTRY(arg_pContext->pMbNatV6V4HashTableBase,
                    hashValue, hashEntry);

    /* 
       if the entry ptr matches and next ptr to collision is 0
       then just clear out the entry.
       if the entry ptr matches and next ptr to collision is non-zero
       then get the next entry from collision table after removing it
       from there and write it in the main table.
       if the entry ptr does not match (next ptr is non-zero) 
    */

    if (hashEntry.index == arg_index && 
                hashEntry.next == IX_CC_NATPT_UNUSED_ENTRY_FILL)
    {
        _CLEAR_NAT_HASH_ENTRY(arg_pContext->pMbNatV6V4HashTableBase,
                    hashValue);
    }
    else if (hashEntry.index == arg_index && 
                hashEntry.next != IX_CC_NATPT_UNUSED_ENTRY_FILL)
    {
        _READ_NAT_HASH_ENTRY(arg_pContext->pMbNatCollisionTableBase, 
                    hashEntry.next, colHashEntry);

        _CLEAR_NAT_HASH_ENTRY(arg_pContext->pMbNatCollisionTableBase, 
                    hashEntry.next);

        IX_ERROR_CR(
            ix_cc_index_pool_free(
                 arg_pContext->natCollisionTableIndexPool, hashEntry.next));

        _WRITE_NAT_HASH_ENTRY(arg_pContext->pMbNatV6V4HashTableBase,
                    hashValue, colHashEntry);
    }
    else if (hashEntry.index != arg_index && 
                hashEntry.next != IX_CC_NATPT_UNUSED_ENTRY_FILL)
    {
        colIndex = IX_CC_NATPT_UNUSED_ENTRY_FILL;

        IX_ERROR_CR(
            _ix_cc_natpt_del_collision_entry(arg_pContext, arg_index, 
                    hashEntry.next, &colIndex));

        if (colIndex != IX_CC_NATPT_UNUSED_ENTRY_FILL)
        {
            hashEntry.next = colIndex;

            _WRITE_NAT_HASH_ENTRY(arg_pContext->pMbNatV6V4HashTableBase,
                    hashValue, hashEntry);
        }
    }


    /* Repeat similarly for v4-v6 hash table */
        
    hashValue = _ix_cc_natpt_hash_v4addr(arg_natEntry->v4Addr);

    _READ_NAT_HASH_ENTRY(arg_pContext->pMbNatV4V6HashTableBase,
                    hashValue, hashEntry);

    if (hashEntry.index == arg_index && 
                hashEntry.next == IX_CC_NATPT_UNUSED_ENTRY_FILL)
    {
        _CLEAR_NAT_HASH_ENTRY(arg_pContext->pMbNatV4V6HashTableBase,
                    hashValue);
    }
    else if (hashEntry.index == arg_index && 
                hashEntry.next != IX_CC_NATPT_UNUSED_ENTRY_FILL)
    {
        _READ_NAT_HASH_ENTRY(arg_pContext->pMbNatCollisionTableBase, 
                hashEntry.next, colHashEntry);

        _CLEAR_NAT_HASH_ENTRY(arg_pContext->pMbNatCollisionTableBase, 
                hashEntry.next);

        IX_ERROR_CR(
            ix_cc_index_pool_free(
                 arg_pContext->natCollisionTableIndexPool, hashEntry.next));

        _WRITE_NAT_HASH_ENTRY(arg_pContext->pMbNatV4V6HashTableBase,
                    hashValue, colHashEntry);
    }
    else if (hashEntry.index != arg_index && 
                hashEntry.next != IX_CC_NATPT_UNUSED_ENTRY_FILL)
    {
        colIndex = IX_CC_NATPT_UNUSED_ENTRY_FILL;

        IX_ERROR_CR(
            _ix_cc_natpt_del_collision_entry(arg_pContext, arg_index, 
                    hashEntry.next, &colIndex));

        if (colIndex != IX_CC_NATPT_UNUSED_ENTRY_FILL)
        {
            hashEntry.next = colIndex;

            _WRITE_NAT_HASH_ENTRY(arg_pContext->pMbNatV4V6HashTableBase,
                    hashValue, hashEntry);
        }
    }
    
    return IX_SUCCESS;
}


/**
 * NAME: _ix_cc_natpt_find_nat_mapping_v6tov4
 *
 * DESCRIPTION: This function is used to get IPv4 address mapping for IPv6
 * address.
 *           
 * @Param:  - IN arg_pContext - pointer to NATPT Context
 *
 * @Param:  - IN/OUT arg_natEntry - pointer to NAT Table Entry. Contains
 * only IPv6 address on input, IPv4 address is filled in on output
 *
 * @Param:  - OUT None
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_ENTRY_NOT_FOUND - Entry Not found.
 * 
 */
ix_error _ix_cc_natpt_find_nat_mapping_v6tov4(
                            ix_cc_natpt_context* arg_pContext,
                            ix_cc_natpt_nat_entry* arg_natEntry)
{
    ix_uint8 hashValue;
    ix_cc_natpt_nat_hash_entry hashEntry;
    ix_cc_natpt_nat_entry natEntry;
    ix_uint32 colIndex;

    hashValue = _ix_cc_natpt_hash_v6addr(arg_natEntry->v6Addr);

    _READ_NAT_HASH_ENTRY(arg_pContext->pMbNatV6V4HashTableBase,
                    hashValue, hashEntry);

    if ( hashEntry.index == IX_CC_NATPT_UNUSED_ENTRY_FILL)
    {
        return IX_CC_ERROR_ENTRY_NOT_FOUND;
    }
    else if (hashEntry.index != IX_CC_NATPT_UNUSED_ENTRY_FILL && 
            hashEntry.next == IX_CC_NATPT_UNUSED_ENTRY_FILL)
    {
        _READ_NAT_TABLE_ENTRY(arg_pContext->pMbNatTableBase, 
                hashEntry.index, *arg_natEntry);

        return IX_SUCCESS;
    }
    else if (hashEntry.index != IX_CC_NATPT_UNUSED_ENTRY_FILL && 
            hashEntry.next != IX_CC_NATPT_UNUSED_ENTRY_FILL)
    {
        _READ_NAT_TABLE_ENTRY(arg_pContext->pMbNatTableBase,
                hashEntry.index, natEntry);

        if (NATPT_IS_IPV6_ADDRESS_EQUAL(&arg_natEntry->v6Addr, 
                &natEntry.v6Addr) )
        {
            arg_natEntry->v4Addr = natEntry.v4Addr;
            return IX_SUCCESS;
        }
        else
        {
            do {
                colIndex = hashEntry.next;
                _READ_NAT_HASH_ENTRY(arg_pContext->pMbNatCollisionTableBase,
                        colIndex, hashEntry);

                _READ_NAT_TABLE_ENTRY(arg_pContext->pMbNatTableBase,
                        hashEntry.index, natEntry);

                if (NATPT_IS_IPV6_ADDRESS_EQUAL(&arg_natEntry->v6Addr, 
                            &natEntry.v6Addr) )
                {
                    arg_natEntry->v4Addr = natEntry.v4Addr;
                    return IX_SUCCESS;
                }
            } while (hashEntry.next != IX_CC_NATPT_UNUSED_ENTRY_FILL);
        }
    }

    return IX_CC_ERROR_ENTRY_NOT_FOUND;
}


/**
 * NAME: _ix_cc_natpt_find_nat_mapping_v4tov6
 *
 * DESCRIPTION: This function is used to get IPv6 address mapping for IPv4
 * address.
 *           
 * @Param:  - IN arg_pContext - pointer to NATPT Context
 *
 * @Param:  - IN/OUT arg_natEntry - pointer to NAT Table Entry. Contains
 * only IPv4 address on input, IPv6 address is filled in on output
 *
 * @Param:  - OUT None
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_ENTRY_NOT_FOUND - Entry Not found.
 * 
 */
ix_error _ix_cc_natpt_find_nat_mapping_v4tov6(
                            ix_cc_natpt_context* arg_pContext,
                            ix_cc_natpt_nat_entry* arg_natEntry)
{
    ix_uint8 hashValue;
    ix_cc_natpt_nat_hash_entry hashEntry;
    ix_cc_natpt_nat_entry natEntry;
    ix_uint32 colIndex;

    hashValue = _ix_cc_natpt_hash_v4addr(arg_natEntry->v4Addr);

    _READ_NAT_HASH_ENTRY(arg_pContext->pMbNatV4V6HashTableBase,
                    hashValue, hashEntry);

    if ( hashEntry.index == IX_CC_NATPT_UNUSED_ENTRY_FILL)
    {
        return IX_CC_ERROR_ENTRY_NOT_FOUND;
    }
    else if (hashEntry.index != IX_CC_NATPT_UNUSED_ENTRY_FILL && 
            hashEntry.next == IX_CC_NATPT_UNUSED_ENTRY_FILL)
    {
        _READ_NAT_TABLE_ENTRY(arg_pContext->pMbNatTableBase,
                hashEntry.index, *arg_natEntry);

        return IX_SUCCESS;
    }
    else if (hashEntry.index != IX_CC_NATPT_UNUSED_ENTRY_FILL && 
            hashEntry.next != IX_CC_NATPT_UNUSED_ENTRY_FILL)
    {
        _READ_NAT_TABLE_ENTRY(arg_pContext->pMbNatTableBase, 
                hashEntry.index, natEntry);

        if (arg_natEntry->v4Addr ==  natEntry.v4Addr)
        {
            NATPT_COPY_IPV6_ADDRESS(&arg_natEntry->v6Addr, &natEntry.v6Addr);
            return IX_SUCCESS;
        }
        else
        {
            do {
                colIndex = hashEntry.next;
                _READ_NAT_HASH_ENTRY(arg_pContext->pMbNatCollisionTableBase, 
                        colIndex, hashEntry);

                _READ_NAT_TABLE_ENTRY(arg_pContext->pMbNatTableBase, 
                        hashEntry.index, natEntry);

                if (arg_natEntry->v4Addr ==  natEntry.v4Addr)
                {
                    NATPT_COPY_IPV6_ADDRESS(&arg_natEntry->v6Addr, &natEntry.v6Addr);
                    return IX_SUCCESS;
                }
            } while (hashEntry.next != IX_CC_NATPT_UNUSED_ENTRY_FILL);
        }
    }

    return IX_CC_ERROR_ENTRY_NOT_FOUND;
}


/**
 * NAME: _ix_cc_natpt_check_dynamic_mappings
 *
 * DESCRIPTION: This function is used to get IPv6 address mapping for IPv4
 * address.
 *           
 * @Param:  - IN arg_pContext - pointer to NATPT Context
 *
 * @Param:  - OUT None
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 */
ix_error _ix_cc_natpt_check_dynamic_mappings(void* arg_pContext)
{
#if !defined(IX_EXCLUDE_CCI)
    ix_cc_natpt_context* pContext = (ix_cc_natpt_context*)arg_pContext;
    ix_uint32 index;
    ix_cc_natpt_nat_entry natEntry;
    ix_error err = IX_SUCCESS;
    
#if defined(NATPT_DEBUG)
    ix_ossl_message_log("In %s\n", __FUNCTION__);
#endif /* NATPT_DEBUG */
    err = ix_cc_index_pool_get_first(pContext->natTableIndexPool, &index);
    while (err == IX_SUCCESS)
    {
        /* Read the entry */
        _READ_NAT_TABLE_ENTRY(pContext->pMbNatTableBase, index, natEntry);

#if defined(NATPT_DEBUG)
        ix_ossl_message_log("Entry at %ld and isUsed = %d isStatic = %d\n",
                      index, natEntry.isUsed, natEntry.isStatic);
#endif /* NATPT_DEBUG */
        if ( ! natEntry.isStatic)
        {
            if (natEntry.isUsed == 0)
            {
                /* Delete the entry */
                IX_ERROR_CR(
                        _ix_cc_natpt_del_nat_hash_table_entries(
                                pContext, &natEntry, index));
            
                IX_ERROR_CR(
                        _ix_cc_natpt_del_nat_table_entry(
                                pContext->pMbNatTableBase,
                                pContext->natTableIndexPool,
                                &natEntry));
                
                /* Return V4 address to pool */
                IX_ERROR_CR(
                    _ix_cc_natpt_free_ipv4addr(
                            pContext->v4PoolList, natEntry.v4Addr));
            }
            else
            {
                natEntry.isUsed = 0;
                _WRITE_NAT_TABLE_RESET_USED(pContext->pMbNatTableBase, 
                        index, natEntry);
            }
        }

        err = ix_cc_index_pool_get_next(
            pContext->natTableIndexPool, index, &index);
    }

#endif /* ! IX_EXCLUDE_CCI */

    return IX_SUCCESS;
}


/**
 * NAME: _ix_cc_natpt_set_entry_timeout_event_handler
 *
 * DESCRIPTION: This function is used to create a periodic event to check
 * for expiring dynamic mappings. It is also used when time interval
 * is changed.
 *           
 * @Param:  - IN arg_pContext - pointer to NATPT Context
 *
 * @Param:  - OUT None
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_NATPT_ERROR_CCI - Adding/Updating event failed
 * 
 */
ix_error _ix_cc_natpt_set_entry_timeout_event_handler(
                        ix_cc_natpt_context* arg_pContext)
{
#if !defined(IX_EXCLUDE_CCI)
#if defined(NATPT_DEBUG)
    ix_ossl_message_log("In %s Setting check interval to %ld seconds\n", 
                    __FUNCTION__, arg_pContext->idleCheckInterval);
#endif /* NATPT_DEBUG */
    if ( ! arg_pContext->hEventHandle)
    {
        IX_ERROR_CRT(ix_cci_cc_add_event_handler_ex(
                                      arg_pContext->hNatPtHandle,
                                      arg_pContext->idleCheckInterval * 1000,
                                      _ix_cc_natpt_check_dynamic_mappings,
                                      IX_EVENT_TYPE_PERIODIC,
                                      IX_CC_NATPT_EVENT_PRIORITY,
                                      arg_pContext,
                                      &arg_pContext->hEventHandle),
                 IX_CC_NATPT_ERROR_CCI,
                 IX_ERROR_LEVEL_WARNING);
    }
    else
    {
        IX_ERROR_CRT(ix_cci_change_event(arg_pContext->hEventHandle,
                                    arg_pContext->idleCheckInterval * 1000),
                 IX_CC_NATPT_ERROR_CCI,
                 IX_ERROR_LEVEL_WARNING);
    }
#endif /* ! IX_EXCLUDE_CCI */

    return IX_SUCCESS;
}
