/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.x for the IXP2xxx Network Processors
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      NATPT Core Component - packet handlers.
 *      This modules handles the following kinds of packets
 *          - Packets from the NATPT microblock 
 *
 * = FILENAME
 *      ix_cc_natpt_pkt_handlers.c
 *
 * = DESCRIPTION
 *      The file defines packet handler interface functions.
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      03/03/03
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */


#define IX_ERROR_FILE_IDENT "$Id: ix_cc_natpt_pkt_handlers.c,v 1.17 2003/12/12 00:59:58 ktseng Exp $";

/**
 * System defined include files
 */
#include "ix_cc_error.h"
#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_cc.h"
#include "bindings.h"
#include "cc/ix_cc_natpt.h"
#include "../../include/v6v4_natpt_core_uc.h"


/**
 * User defined include files
 */
#include "cc/natpt/internal/ix_cc_natpt_init.h"
#include "cc/natpt/internal/ix_cc_natpt_translate.h"
#include "cc/natpt/internal/ix_cc_natpt_icmp.h"

/**
 * Preprocessor symbols and Macros used in this file
 */

/**
 * Type definitions whose scope is limited to this file
 */


/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */


/** 
 * Extern function prototypes.
 */

/** 
 * Static function prototypes.
 */


/** 
 * Function definitions.
 */

/**
 * NAME: ix_cc_natpt_microblock_pkt_handler
 *
 * DESCRIPTION: This is the only registered function to receive exception 
 * packets from NATPT microblock. 
 * 
 * @Param:  - IN arg_hBuffer: handle to a buffer which contains exception 
 *                            packets from NATPT microblock
 * @Param: - IN arg_ExceptionCode: the exception code.
 * @Param: - IN arg_pContext: pointer to NATPT core component context.
 * @Param:  - OUT none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_SEND_FAIL - error from CCI
 * 
 */

ix_error 
ix_cc_natpt_microblock_pkt_handler(ix_buffer_handle arg_hBuffer,  
                                   ix_uint32 arg_ExceptionCode,
                                   void* arg_pContext)
{
    ix_error err = IX_SUCCESS;
#if defined(NATPT_DEBUG)
    ix_hw_buffer_meta *pHwMeta;

    IX_ERROR_CRT(ix_rm_buffer_get_meta(arg_hBuffer, 
                                       (void **)&pHwMeta), 
                       IX_CC_ERROR_INTERNAL,
                       0);
    ix_ossl_message_log (" *********  Packet Details  *********** \n");

    ix_ossl_message_log ("PacketSize (0x%x) \n", 
                 IX_NTOH16(pHwMeta->m_PacketSize));
    ix_ossl_message_log ("BufferSize (0x%x) \n", 
                             IX_NTOH16(pHwMeta->m_BufferSize));
    ix_ossl_message_log ("Offset (0x%x) \n", 
                 IX_NTOH16(pHwMeta->m_Offset));
    ix_ossl_message_log ("BufferInfo (0x%x) \n", 
                 IX_NTOH16(pHwMeta->m_BufferInfo));
    ix_ossl_message_log ("InputPort (0x%x) \n", 
                 IX_NTOH16(pHwMeta->m_InputPort));

    ix_ossl_message_log ("Exception Code (0x%lx) \n", 
                 arg_ExceptionCode);
#endif /* NATPT_DEBUG */

    switch(arg_ExceptionCode)
    {
        case NATPT_EXCEPTION_NOENTRY_IPV6:
        {
            IX_ERROR_CG(
                _ix_cc_natpt_handle_ipv6_packet(arg_hBuffer, arg_pContext),
                err, freePacket);

            break;
        }

#ifdef NOT_YET_SUPPORTED
        case NATPT_EXCEPTION_NOENTRY_IPV4:
        {
            /* This case won't happen now as we don't yet support 
               communication initiation from IPv4 realm */
            IX_ERROR_CG(
                _ix_cc_natpt_handle_ipv4_packet(arg_hBuffer, arg_pContext),
                err, freePacket);

            break;
        }
#endif /* NOT_YET_SUPPORTED */

        case NATPT_EXCEPTION_ICMP_IPV6:
        {
            IX_ERROR_CG(
                _ix_cc_natpt_icmp_v6_packet(arg_hBuffer, arg_pContext),
                err, freePacket);

            break;
        }

        case NATPT_EXCEPTION_ICMP_IPV4:
        {
            IX_ERROR_CG(
                _ix_cc_natpt_icmp_v4_packet(arg_hBuffer, arg_pContext),
                err, freePacket);

            break;
        }

        case NATPT_EXCEPTION_REASSEMBLE_UDP:
        {
            IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hBuffer),
                                            IX_CC_ERROR_SEND_FAIL, 0);

            return IX_ERROR_WARNING(IX_CC_ERROR_UNIMPL,
                        ("Fragmented UDP V4 Packet"));
            break;
        }

        case NATPT_EXCEPTION_UDP_V4_CHECKSUM_ZERO:
        {
            IX_ERROR_CG(
                _ix_cc_natpt_handle_udp_ipv4_packet(
                        arg_hBuffer, arg_pContext),
                err, freePacket);

            break;
        }

        case NATPT_EXCEPTION_IPV6_SECURITY_HEADER:
        {
            IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hBuffer),
                                            IX_CC_ERROR_SEND_FAIL, 0);

            return IX_ERROR_WARNING(IX_CC_ERROR_UNIMPL,
                        ("Security Extension Headers in V6 Packet"));
            break;
        }

        default:

            IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hBuffer),
                                            IX_CC_ERROR_SEND_FAIL, 0);

            return IX_ERROR_WARNING(IX_CC_ERROR_UNDEFINED_EXCEP,
                                                ("undefined exception code"));
            break;

    } /* end switch */

    if (err != IX_SUCCESS)
    {
        IX_ERROR_CRT(
            ix_rm_buffer_free_chain(arg_hBuffer),
            IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

        return err;

    }/* end if (err != IX_SUCCESS) */     

   return IX_SUCCESS;

freePacket:
    IX_ERROR_CRT(
        ix_rm_buffer_free_chain(arg_hBuffer),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    return err;

} /* end ix_cc_natpt_microblock_pkt_handler */


