/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.x for the IXP2xxx Network Processors
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      IPv6 NATPT Core Component - Translation
 *
 * = FILENAME
 *      ix_cc_natpt_translate.c
 *
 * = DESCRIPTION
 *      The file defines functions for handling of IPv6 and IPv4 packets 
 *      requiring translation.
 *
 * = CREATION TIME
 *      03/03/03
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_natpt_translate.c,v 1.26 2003/10/31 19:25:59 ktseng Exp $";


/**
 * System defined include files
 */
#include "ix_cc_error.h"
#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_cc.h"
#include "bindings.h"
#include "cc/ix_cc_natpt.h"

/**
 * User defined include files
 */
#include "cc/natpt/internal/ix_cc_natpt_init.h"
#include "cc/natpt/internal/ix_cc_natpt_lists.h"
#include "cc/natpt/internal/ix_cc_natpt_utils.h"
#include "cc/natpt/internal/ix_cc_natpt_translate.h"

/**
 * Preprocessor symbols and Macros used in this file
 */

#define IX_CC_NATPT_IPV4_DF              0x4000
#define IX_CC_NATPT_IPV4_MF              0x2000
#define IX_CC_NATPT_MIN_IPV6_MTU         1300

#define IX_CC_NATPT_TCP_CHKSUM_OFFSET       16
#define IX_CC_NATPT_UDP_CHKSUM_OFFSET       6


/**
 * Type definitions whose scope is limited to this file
 */

/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */


/** 
 * Extern function prototypes.
 */


/** 
 * Static function prototypes.
 */

PRIVATE void _ix_cc_natpt_udp_v6_calculate_checksum(ix_uint8 *udpMessage);

/** 
 * Function definitions.
 */


/**
 * NAME: _ix_cc_natpt_handle_ipv6_packet
 *
 * DESCRIPTION: This primitive is used to process IPv6 packets that
 * require translation.
 *           
 * @Param:  - IN arg_hPacket - handle to buffer containing the packet.
 *
 * @Param:  - IN arg_pContext - pointer to NATPT core component
 * context buffer.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * 
 */

ix_error _ix_cc_natpt_handle_ipv6_packet(
    ix_buffer_handle arg_hPacket,  
    void* arg_pContext
    )
{
    ix_uint8* pData;
    ix_uint8* pPtr;
    ix_cc_natpt_ipv4_header ipv4Hdr;
    ix_cc_natpt_ipv6_header ipv6Hdr;
    ix_hw_buffer_meta *pHwMeta;
    ix_cc_natpt_context* pContext;
    ix_uint16 packetSize;
    ix_uint32 csum = 0;
    ix_uint8* csumPtr;
    ix_uint8 csumBuf[2];
    ix_uint16* pHdr;
    ix_cc_natpt_nat_entry natEntry;
    ix_cc_natpt_ipv6_extn_hdr_common* extHdr;
    ix_cc_natpt_ipv6_fragment_hdr* fragHdr = NULL;
    ix_uint16 extLength;
    ix_uint16 fragOffBytes;
    ix_uint16 payLoadBytes;
    ix_uint16 bufSize;
    ix_uint16 offset;
    ix_uint16 bufInfo;
    ix_uint32 cellCount;
    int i;


    pContext = (ix_cc_natpt_context*)arg_pContext;

    /* Get the packet meta-data */
    IX_ERROR_CRT(
        ix_rm_buffer_get_meta(arg_hPacket, (void **)&pHwMeta),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    /* Get packet data */
    IX_ERROR_CRT(
        ix_rm_buffer_get_data(arg_hPacket, (void **)&pData),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    pData += IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_Offset));

    ix_ossl_memcpy(&ipv6Hdr, pData, IX_CC_NATPT_IPV6_HDR_LEN);

#if defined(NATPT_DEBUG)
    ix_ossl_message_log("Received Packet Src Address\n");
    _ix_cc_natpt_print_ipv6_address(&ipv6Hdr.sourceAddress);
    ix_ossl_message_log("Received Packet Dest Address\n");
    _ix_cc_natpt_print_ipv6_address(&ipv6Hdr.destAddress);
#endif /* NATPT_DEBUG */

    /* Check for IPSEC extension headers */
    IX_ERROR_CRT(
        _ix_cc_natpt_validate_extension_headers(pData, &ipv6Hdr),
        IX_CC_ERROR_UNIMPL, IX_ERROR_LEVEL_WARNING);

    /* Get the mapping IPv4 address */
    IX_ERROR_CRT(
        _ix_cc_natpt_get_ipv6_to_ipv4_mapping(&ipv6Hdr, &natEntry, pContext),
        IX_CC_ERROR_OOR, IX_ERROR_LEVEL_WARNING);

#if defined(NATPT_DEBUG)
    ix_ossl_message_log("Mapping gave IPv4 address (%lx)\n", natEntry.v4Addr);
#endif /* NATPT_DEBUG */

    /* Construct the IPv4 header */
    ix_ossl_memset(&ipv4Hdr, 0, IX_CC_NATPT_MIN_IPV4_LEN);
    ipv4Hdr.vers_hdrLength = IX_CC_NATPT_IPV4_VERS_IHL;

    /* Set the appropriate value for the TOS byte. */
#if defined (TRANSLATE_TRAFFIC_CLASS)
    ipv4Hdr.tos = *((ix_uint8*)&ipv6Hdr + IX_CC_NATPT_IPV6_TRCLS) << 4;
    ipv4Hdr.tos |= *((ix_uint8*)&ipv6Hdr + IX_CC_NATPT_IPV6_TRCLS + 1) >> 4;
#else
    ipv4Hdr.tos = 0;
#endif /* TRANSLATE_TRAFFIC_CLASS */

    ipv4Hdr.ttl = ipv6Hdr.hopLimit;

    /* calculate extension headers length */
    if ((ipv6Hdr.nextHeader == IX_CC_NATPT_NEXT_HDR_HOPBYHOP) ||
            (ipv6Hdr.nextHeader == IX_CC_NATPT_NEXT_HDR_ROUTING) ||
            (ipv6Hdr.nextHeader == IX_CC_NATPT_NEXT_HDR_FRAGMENT) ||
            (ipv6Hdr.nextHeader == IX_CC_NATPT_NEXT_HDR_ENCAPSULATION) ||
            (ipv6Hdr.nextHeader == IX_CC_NATPT_NEXT_HDR_AUTHENTICATION) ||
            (ipv6Hdr.nextHeader == IX_CC_NATPT_NEXT_HDR_DSTOPT))
    {
        /* walk thru the chain to calculate */
        extLength = 0;
        extHdr = ( ix_cc_natpt_ipv6_extn_hdr_common*) 
                (pData + IX_CC_NATPT_IPV6_HDR_LEN);
        /* save pointer if fragment hdr for later use */
        if (extHdr->nextHeader == IX_CC_NATPT_NEXT_HDR_FRAGMENT)
        {
            fragHdr = (ix_cc_natpt_ipv6_fragment_hdr*) extHdr;
        }

        do {
            extLength += (8 + (extHdr->hdrLength * 8));
            if ((extHdr->nextHeader == IX_CC_NATPT_NEXT_HDR_HOPBYHOP) ||
                (extHdr->nextHeader == IX_CC_NATPT_NEXT_HDR_ROUTING) ||
                (extHdr->nextHeader == IX_CC_NATPT_NEXT_HDR_FRAGMENT) ||
                (extHdr->nextHeader == IX_CC_NATPT_NEXT_HDR_ENCAPSULATION) ||
                (extHdr->nextHeader == IX_CC_NATPT_NEXT_HDR_AUTHENTICATION) ||
                (extHdr->nextHeader == IX_CC_NATPT_NEXT_HDR_DSTOPT))
            {
                ipv4Hdr.protocol = extHdr->nextHeader;     
                /* go to next header */
                extHdr = ( ix_cc_natpt_ipv6_extn_hdr_common* ) 
                        (pData + IX_CC_NATPT_IPV6_HDR_LEN + extLength);
                continue;
            }
            else
            {
                ipv4Hdr.protocol = extHdr->nextHeader;     
                break;
            }
        } while (1); /* intensional; loop terminates using break */
    }
    else
    {
        extLength = 0;
        ipv4Hdr.protocol = ipv6Hdr.nextHeader;
    }

#if defined(NATPT_DEBUG)
    ix_ossl_message_log("Total Extension Header Length (%d)\n", extLength);
#endif /* NATPT_DEBUG */

    /* update packet size */
    packetSize = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_PacketSize));
    packetSize -= IX_CC_NATPT_IPV6_HDR_LEN;
    packetSize -= extLength;
    packetSize += IX_CC_NATPT_MIN_IPV4_LEN;

    ipv4Hdr.totalLength = IX_HTON16(packetSize);

    /* Check for fragments */
    if (fragHdr)
    {
        ipv4Hdr.id = (ix_uint16) fragHdr->ident;
        ipv4Hdr.flags_fragOffset = fragHdr->fragOffset;
        if (fragHdr->mFlag)
        {
            ipv4Hdr.flags_fragOffset |= IX_CC_NATPT_IPV4_MF;
        }
    }
    else
    {
        ipv4Hdr.id = 0;
        ipv4Hdr.flags_fragOffset = IX_HTON16(IX_CC_NATPT_IPV4_DF);
    }

    ipv4Hdr.checksum = 0;
    ipv4Hdr.sourceAddress = natEntry.v4Addr;
    ipv4Hdr.destAddress = ipv6Hdr.destAddress.aUint32[3];

    /* Calculate the checksum */
    pHdr = (ix_uint16*)&ipv4Hdr;
    for (i = 0; i < (IX_CC_NATPT_MIN_IPV4_LEN >> 1); i++)
    {
        csum += IX_NTOH16(*pHdr);
        pHdr++;
    }

    csum = ~((csum & 0xffff) + (csum >> 16));
    ipv4Hdr.checksum = IX_HTON16((ix_uint16)csum);

    /* Update TCP and UDP checksum if necessary */
    if ((ipv4Hdr.protocol == IX_CC_NATPT_NEXT_HDR_TCP) && fragHdr)
    {
#if defined(NATPT_DEBUG)
        ix_ossl_message_log("TCP and Fragment Header present\n");
#endif /* NATPT_DEBUG */
        pPtr = (ix_uint8*) extHdr;

        if (fragHdr->fragOffset < 3)
        { 
            /* determine if the checksum is in this fragment */
            /* Checksum field is in 17 & 18th bytes of TCP Hdr*/
            fragOffBytes = IX_NTOH16(fragHdr->fragOffset) * 8;
            payLoadBytes = IX_NTOH16(
                    IX_RM_MEM_UINT16_READ(&pHwMeta->m_PacketSize)) -
                    (extLength + IX_CC_NATPT_IPV6_HDR_LEN);

            if ( (payLoadBytes - 
                 (IX_CC_NATPT_TCP_CHKSUM_OFFSET - fragOffBytes)) > 2)
            { 
                csumPtr = pPtr + IX_CC_NATPT_TCP_CHKSUM_OFFSET - 
                        fragOffBytes;
                ix_ossl_memcpy(csumBuf, csumPtr, 2);
                checksumadjust(csumBuf, 
                        (ix_uint8*) &ipv6Hdr.sourceAddress, 256,
                        (ix_uint8*) &ipv4Hdr.sourceAddress, 64);
                ix_ossl_memcpy(csumPtr, csumBuf, 2);
            }
        }
    } 
    else if ((ipv4Hdr.protocol == IX_CC_NATPT_NEXT_HDR_UDP) && fragHdr)
    {
#if defined(NATPT_DEBUG)
        ix_ossl_message_log("UDP and Fragment Header present\n");
#endif /* NATPT_DEBUG */
        pPtr = (ix_uint8*) extHdr;

        if (fragHdr->fragOffset == 0)
        { 
            /* determine if the checksum is in this fragment */
            /* Checksum field is in 7 & 8th bytes of UDP Hdr*/
            payLoadBytes = IX_NTOH16(
                    IX_RM_MEM_UINT16_READ(&pHwMeta->m_PacketSize)) -
                    (extLength + IX_CC_NATPT_IPV6_HDR_LEN);

            if ( (payLoadBytes - IX_CC_NATPT_UDP_CHKSUM_OFFSET) > 2)
            { 
                csumPtr = pPtr + IX_CC_NATPT_UDP_CHKSUM_OFFSET;
                ix_ossl_memcpy(csumBuf, csumPtr, 2);
                checksumadjust(csumBuf, 
                        (ix_uint8*) &ipv6Hdr.sourceAddress, 256,
                        (ix_uint8*) &ipv4Hdr.sourceAddress, 64);
                ix_ossl_memcpy(csumPtr, csumBuf, 2);
            }
        }
    }
 
    /* Copy the header to the packet */
    ix_ossl_memcpy(pData + IX_CC_NATPT_IPV6_HDR_LEN + 
            extLength - IX_CC_NATPT_MIN_IPV4_LEN,
            &ipv4Hdr, IX_CC_NATPT_MIN_IPV4_LEN);

    /* Update the meta data */
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_PacketSize, IX_HTON16(packetSize));

    bufSize = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_BufferSize));
    bufSize = bufSize - IX_CC_NATPT_IPV6_HDR_LEN - extLength + 
            IX_CC_NATPT_MIN_IPV4_LEN;
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_BufferSize, IX_HTON16(bufSize));

    offset = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_Offset));
    offset = offset + IX_CC_NATPT_IPV6_HDR_LEN + extLength - 
            IX_CC_NATPT_MIN_IPV4_LEN;
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_Offset, IX_HTON16(offset));

    bufInfo = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_BufferInfo));
    IX_RM_HW_META_SET_HEADER_TYPE(bufInfo, IX_CC_NATPT_HDRTYPE_IPV4);
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_BufferInfo, IX_HTON16(bufInfo));

    cellCount = ix_cc_calculate_cell_count(bufSize, 1);
    ix_rm_buffer_cell_count_set(&arg_hPacket, cellCount);

#if defined(NATPT_DEBUG)
    ix_ossl_message_log (" After Translation Packet Details\n");
    ix_ossl_message_log ("PacketSize (0x%x) \n", 
                 IX_NTOH16(pHwMeta->m_PacketSize));
    ix_ossl_message_log ("BufferSize (0x%x) \n", 
                 IX_NTOH16(pHwMeta->m_BufferSize));
    ix_ossl_message_log ("BufferInfo (0x%x) \n", 
                 IX_NTOH16(pHwMeta->m_BufferInfo));
    ix_ossl_message_log ("Offset (0x%x) \n", 
                 IX_NTOH16(pHwMeta->m_Offset));
    ix_ossl_message_log ("Cell Count (0x%lx) \n", cellCount);
#endif /* NATPT_DEBUG */

#if defined(NATPT_DEBUG)
    pData += (IX_CC_NATPT_IPV6_HDR_LEN + extLength - IX_CC_NATPT_MIN_IPV4_LEN);
    while (packetSize)
    {
        for (i = 0; i < 16 && packetSize; i++)
        {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
            ix_ossl_message_log(" %02x", *pData);
#else
            printf(" %02x", *pData);
#endif
            packetSize --;
            pData ++;
        }
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_ossl_message_log("\n");
#else
        printf("\n");
#endif
    }
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_ossl_message_log("\n");
#else
    printf("\n");
#endif
#endif /* NATPT_DEBUG */

    IX_ERROR_CRT(
        ix_rm_packet_send(IX_CC_NATPT_PKT_IPV4_OUTPUT, arg_hPacket, 0),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    return IX_SUCCESS;

} /* end - _ix_cc_natpt_handle_ipv6_packet */




/**
 * NAME: _ix_cc_natpt_handle_ipv4_packet
 *
 * DESCRIPTION: This primitive is used to process IPv4 packets that
 * require translation.
 *           
 * @Param:  - IN arg_hPacket - handle to buffer containing the packet.
 *
 * @Param:  - IN arg_pContext - pointer to NATPT core component
 * context buffer.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * 
 */

ix_error _ix_cc_natpt_handle_ipv4_packet(
    ix_buffer_handle arg_hPacket,  
    void* arg_pContext
    )
{
    return IX_CC_ERROR_UNIMPL;
} /* end - _ix_cc_natpt_handle_ipv6_packet */


/**
 * NAME: _ix_cc_natpt_handle_udp_ipv4_packet
 *
 * DESCRIPTION: This primitive is used to process IPv4 UDP packets that
 * cannot be translated by the microblock because UDP checksum field
 * is zero. 
 *           
 * @Param:  - IN arg_hPacket - handle to buffer containing the packet.
 *
 * @Param:  - IN arg_pContext - pointer to NATPT core component
 * context buffer.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * 
 */

ix_error _ix_cc_natpt_handle_udp_ipv4_packet(
    ix_buffer_handle arg_hPacket,  
    void* arg_pContext
    )
{
    ix_uint8* pData;
    ix_cc_natpt_ipv4_header ipv4Hdr;
    ix_cc_natpt_ipv6_header ipv6Hdr;
    ix_uint128 ipv6Addr;
    ix_hw_buffer_meta *pHwMeta;
    ix_cc_natpt_context* pContext;
    ix_uint16 packetSize;
    ix_cc_natpt_nat_entry natEntry;
    ix_uint16 bufSize;
    ix_uint16 offset;
    ix_uint16 tmpSize;
    ix_uint16 bufInfo;

    pContext = (ix_cc_natpt_context*)arg_pContext;

    /* Get the packet meta-data */
    IX_ERROR_CRT(
        ix_rm_buffer_get_meta(arg_hPacket, (void **)&pHwMeta),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    /* read meta data sizes */
    packetSize = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_PacketSize));
    bufSize = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_BufferSize));
    offset = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_Offset));

    /* Get packet data */
    IX_ERROR_CRT(
        ix_rm_buffer_get_data(arg_hPacket, (void **)&pData),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    pData += IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_Offset));

    ix_ossl_memcpy(&ipv4Hdr, pData, IX_CC_NATPT_MIN_IPV4_LEN);

#if defined(NATPT_DEBUG)
    ix_ossl_message_log("Received Packet Src Address (%lx)\n", ipv4Hdr.sourceAddress);
    ix_ossl_message_log("Received Packet Dst Address (%lx)\n", ipv4Hdr.destAddress);
#endif /* NATPT_DEBUG */

    /* Get the mapping IPv6 address */
    IX_ERROR_CRT(
        _ix_cc_natpt_get_ipv4_to_ipv6_mapping(&ipv4Hdr, &natEntry, pContext),
        IX_CC_ERROR_OOR, IX_ERROR_LEVEL_WARNING);

#if defined(NATPT_DEBUG)
    ix_ossl_message_log("Mapping gave IPv6 address \n");
    _ix_cc_natpt_print_ipv6_address(&natEntry.v6Addr);
#endif /* NATPT_DEBUG */

    /* Construct the IPv6 header */
    ix_ossl_memset(&ipv6Hdr, 0, IX_CC_NATPT_IPV6_HDR_LEN);
    ipv6Hdr.version = 6;
#if defined (TRANSLATE_TRAFFIC_CLASS)
    ipv6Hdr.traffic_class = ipv4Hdr.tos;
#else
    ipv6Hdr.traffic_class = 0;
#endif /* TRANSLATE_TRAFFIC_CLASS */

    ipv6Hdr.flow_label = 0;
    ipv6Hdr.nextHeader = ipv4Hdr.protocol;
    ipv6Hdr.hopLimit = ipv4Hdr.ttl;

    /* update packet sizes etc */
    tmpSize = IX_CC_NATPT_IPV6_HDR_LEN - IX_CC_NATPT_MIN_IPV4_LEN;
    packetSize += tmpSize;
    bufSize += tmpSize;
    offset -= tmpSize;
#if defined(NATPT_DEBUG)
    ix_ossl_message_log ("Header changed PacketSize (0x%x) \n", packetSize);
#endif /* NATPT_DEBUG */

    ipv6Hdr.payloadLength = IX_HTON16(packetSize - IX_CC_NATPT_IPV6_HDR_LEN) ;

    MAKE_IPV4_MAPPED_IPV6_ADDRESS(ipv4Hdr.sourceAddress, &ipv6Addr);
    ix_ossl_memcpy(&ipv6Hdr.sourceAddress, &ipv6Addr, 
                IX_CC_NATPT_IPV6ADDRLEN_BYTES);
    
    ix_ossl_memcpy(&ipv6Hdr.destAddress, &natEntry.v6Addr, 
                IX_CC_NATPT_IPV6ADDRLEN_BYTES);

    /* update length in outer ip header again*/
    ipv6Hdr.payloadLength = IX_HTON16(packetSize - IX_CC_NATPT_IPV6_HDR_LEN) ;

    /* Set the pointer to where IPv4 header is written */
    pData -= (IX_CC_NATPT_IPV6_HDR_LEN - IX_CC_NATPT_MIN_IPV4_LEN);
    /* write IPv6 header */
    ix_ossl_memcpy(pData, &ipv6Hdr, IX_CC_NATPT_IPV6_HDR_LEN);

    /* Calculate UDP checksum */
    _ix_cc_natpt_udp_v6_calculate_checksum(pData);

    /* Update the meta data */
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_PacketSize, IX_HTON16(packetSize));
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_BufferSize, IX_HTON16(bufSize));
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_Offset, IX_HTON16(offset));

    bufInfo = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_BufferInfo));
    IX_RM_HW_META_SET_HEADER_TYPE(bufInfo, IX_CC_NATPT_HDRTYPE_IPV6);
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_BufferInfo, IX_HTON16(bufInfo));

#if defined(NATPT_DEBUG)
    ix_ossl_message_log (" After Translation Packet Details\n");
    ix_ossl_message_log ("PacketSize (0x%x) \n", 
                 IX_NTOH16(pHwMeta->m_PacketSize));
    ix_ossl_message_log ("BufferSize (0x%x) \n", 
                 IX_NTOH16(pHwMeta->m_BufferSize));
    ix_ossl_message_log ("BufferInfo (0x%x) \n", 
                 IX_NTOH16(pHwMeta->m_BufferInfo));
    ix_ossl_message_log ("Offset (0x%x) \n", 
                 IX_NTOH16(pHwMeta->m_Offset));
#endif /* NATPT_DEBUG */

#if defined(NATPT_DEBUG)
    while (packetSize)
    {
        int i;
        for (i = 0; i < 16 && packetSize; i++)
        {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
            ix_ossl_message_log(" %02x", *pData);
#else
            printf(" %02x", *pData);
#endif
            packetSize --;
            pData ++;
        }
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_ossl_message_log("\n");
#else
        printf("\n");
#endif
    }
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_ossl_message_log("\n");
#else
    printf("\n");
#endif
#endif /* NATPT_DEBUG */

    IX_ERROR_CRT(
        ix_rm_packet_send(IX_CC_NATPT_PKT_IPV6_OUTPUT, arg_hPacket, 0),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    return IX_SUCCESS;
}



/**
 * NAME: _ix_cc_natpt_get_ipv6_to_ipv4_mapping
 *
 * DESCRIPTION: This primitive is used to get IPv4 address for the IPv6
 * address. If there mapping does not exist, it is created.
 *           
 * @Param:  - IN arg_ipv6Hdr - pointer to IPv6 header.
 * @Param:  - IN arg_pContext - pointer to NATPT core component
 * context buffer.
 *
 * @Param:  - IN/OUT arg_natEntry - On input contains IPv6 address, IPv4
 * address is filled in on output.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * 
 */

ix_error _ix_cc_natpt_get_ipv6_to_ipv4_mapping(
            ix_cc_natpt_ipv6_header* arg_ipv6Hdr,
            ix_cc_natpt_nat_entry* arg_natEntry,
            ix_cc_natpt_context* arg_pContext)
{
    ix_uint32 ipv4Addr;
    ix_uint32 natIndex;
    ix_error err;

    /* Check whether mapping exists already */
    NATPT_COPY_IPV6_ADDRESS(&arg_natEntry->v6Addr, &arg_ipv6Hdr->sourceAddress);
    if (_ix_cc_natpt_find_nat_mapping_v6tov4(arg_pContext, arg_natEntry) 
            != IX_SUCCESS)
    {
        /* Get a new IPv4 address from pool */
        IX_ERROR_CR(
            _ix_cc_natpt_alloc_ipv4addr(
                    arg_pContext->v4PoolList, &ipv4Addr));

        /* Create a new binding and update the table */
        arg_natEntry->v4Addr = ipv4Addr;
#if !(_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        NATPT_COPY_IPV6_ADDRESS(&arg_natEntry->v6Addr, &arg_ipv6Hdr->sourceAddress);
#endif
        arg_natEntry->unused = 0;
        arg_natEntry->isUsed = 0;
        arg_natEntry->isStatic = 0;
        arg_natEntry->v6NHID = 0xffff;
        natIndex = 0;
        IX_ERROR_CR(_ix_cc_natpt_add_nat_table_entry(
                            arg_pContext->pMbNatTableBase, 
                            arg_pContext->natTableIndexPool, 
                            arg_natEntry, &natIndex));

        err = _ix_cc_natpt_add_nat_hash_table_entries(
                    arg_pContext, arg_natEntry, natIndex);
        if (err != IX_SUCCESS)
        {
            _ix_cc_natpt_del_nat_table_entry(arg_pContext->pMbNatTableBase, 
                    arg_pContext->natTableIndexPool, arg_natEntry);
            return err;
        }
    }
        
    return IX_SUCCESS;
}


/**
 * NAME: _ix_cc_natpt_get_ipv4_to_ipv6_mapping
 *
 * DESCRIPTION: This primitive is used to get IPv6 address for the IPv4
 * address.
 *           
 * @Param:  - IN arg_ipv4Hdr - pointer to IPv4 header.
 * @Param:  - IN arg_pContext - pointer to NATPT core component
 * context buffer.
 *
 * @Param:  - IN/OUT arg_natEntry - On input contains IPv4 address, IPv6
 * address is filled in on output.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * 
 */

ix_error _ix_cc_natpt_get_ipv4_to_ipv6_mapping(
            ix_cc_natpt_ipv4_header* arg_ipv4Hdr,
            ix_cc_natpt_nat_entry* arg_natEntry,
            ix_cc_natpt_context* arg_pContext)
{
    /* Check whether mapping exists already */
    arg_natEntry->v4Addr = arg_ipv4Hdr->destAddress;

    /* This mapping must be present */
    return (_ix_cc_natpt_find_nat_mapping_v4tov6(arg_pContext, arg_natEntry));
}


/**
 * NAME: checksumadjust
 *
 * DESCRIPTION: 
 *      Checkusm update routing taken from RFC 1631
 *
 * @Param:  - IN unsigned char is 8 bits, long is 32 bits.
 * @Param:  - IN chksum points to the chksum in the packet
 * @Param:  - IN optr points to the old data in the packet
 * @Param:  - IN nptr points to the new data in the packet
 *
 * @Return: None.
 *
 */

void checksumadjust(
            unsigned char *chksum, 
            unsigned char *optr,
            int olen, 
            unsigned char *nptr, 
            int nlen)
{
 long x, old, new;
 x=chksum[0]*256+chksum[1];
 x=~x;
 while (olen) {
   if (olen==1) {
     old=optr[0]*256+optr[1];
     x-=old & 0xff00;
     if (x<=0) { x--; x&=0xffff; }
     break;
   }
   else {
     old=optr[0]*256+optr[1]; optr+=2;
     x-=old & 0xffff;
     if (x<=0) { x--; x&=0xffff; }
     olen-=2;
   }
 }
 while (nlen) {
   if (nlen==1) {
     new=nptr[0]*256+nptr[1];
     x+=new & 0xff00;
     if (x & 0x10000) { x++; x&=0xffff; }
     break;
   }
   else {
     new=nptr[0]*256+nptr[1]; nptr+=2;
     x+=new & 0xffff;
     if (x & 0x10000) { x++; x&=0xffff; }
     nlen-=2;
   }
 }
 x=~x;
 chksum[0]=x/256; chksum[1]=x & 0xff;
}


/**
 * NAME: _ix_cc_natpt_validate_extension_headers
 *
 * DESCRIPTION:  This function walks through the extension header chain
 * and looks for unsupported ones. 
 *           
 * @Param:  - IN arg_pData - pointer to packet
 * @Param:  - IN arg_ipv6Hdr - pointer to IPv6 Header
 *
 * @Return: IX_SUCCESS if no unsupported extn hdrs found otherwise error
 *
 * Error Codes:
 *      IX_CC_NATPT_ERROR_NOT_IMPLEMENTED
 * 
 */

ix_error _ix_cc_natpt_validate_extension_headers(
            ix_uint8* arg_pData,
            ix_cc_natpt_ipv6_header* arg_v6Hdr)
{
    ix_cc_natpt_ipv6_extn_hdr_common* extHdr;
    ix_uint16 extLength;

    if ((arg_v6Hdr->nextHeader == IX_CC_NATPT_NEXT_HDR_HOPBYHOP) ||
            (arg_v6Hdr->nextHeader == IX_CC_NATPT_NEXT_HDR_ROUTING) ||
            (arg_v6Hdr->nextHeader == IX_CC_NATPT_NEXT_HDR_FRAGMENT) ||
            (arg_v6Hdr->nextHeader == IX_CC_NATPT_NEXT_HDR_ENCAPSULATION) ||
            (arg_v6Hdr->nextHeader == IX_CC_NATPT_NEXT_HDR_AUTHENTICATION) ||
            (arg_v6Hdr->nextHeader == IX_CC_NATPT_NEXT_HDR_DSTOPT))
    {
        if ((arg_v6Hdr->nextHeader == IX_CC_NATPT_NEXT_HDR_ENCAPSULATION) ||
                (arg_v6Hdr->nextHeader == 
                IX_CC_NATPT_NEXT_HDR_AUTHENTICATION))
        {
            return IX_CC_NATPT_ERROR_NOT_IMPLEMENTED;
        }

        /* walk thru the chain */
        extLength = 0;
        extHdr = ( ix_cc_natpt_ipv6_extn_hdr_common*) 
                (arg_pData + IX_CC_NATPT_IPV6_HDR_LEN);

        do {
            extLength += (8 + (extHdr->hdrLength * 8));
            if ((extHdr->nextHeader == IX_CC_NATPT_NEXT_HDR_HOPBYHOP) ||
                (extHdr->nextHeader == IX_CC_NATPT_NEXT_HDR_ROUTING) ||
                (extHdr->nextHeader == IX_CC_NATPT_NEXT_HDR_FRAGMENT) ||
                (extHdr->nextHeader == IX_CC_NATPT_NEXT_HDR_ENCAPSULATION) ||
                (extHdr->nextHeader == IX_CC_NATPT_NEXT_HDR_AUTHENTICATION) ||
                (extHdr->nextHeader == IX_CC_NATPT_NEXT_HDR_DSTOPT))
            {
                if ((extHdr->nextHeader == 
                        IX_CC_NATPT_NEXT_HDR_ENCAPSULATION) ||
                        (extHdr->nextHeader == 
                        IX_CC_NATPT_NEXT_HDR_AUTHENTICATION))
                {
                    return IX_CC_NATPT_ERROR_NOT_IMPLEMENTED;
                }

                /* go to next header */
                extHdr = ( ix_cc_natpt_ipv6_extn_hdr_common* ) 
                        (arg_pData + IX_CC_NATPT_IPV6_HDR_LEN + extLength);
                continue;
            }
            else
            {
                break;
            }
        } while (1); /* intensional; loop terminates using break */
    }

    return IX_SUCCESS;
}


/**
 * NAME: _ix_cc_natpt_udp_v6_calculate_checksum
 *
 * DESCRIPTION:  This function calculates the checksum for the 
 * UDP message
 *           
 * @Param:  - IN udpMessage - pointer to UDP V6 Message
 *
 * @Return: None
 *
 * Error Codes:
 * 
 */

PRIVATE void _ix_cc_natpt_udp_v6_calculate_checksum(ix_uint8 *udpMessage)
{
    ix_cc_natpt_ipv6_header *pIpHdr;
    ix_cc_natpt_udp_hdr *pUdpHdr;
    ix_cc_natpt_ipv6_pseudo_header pseudoHdr;
    ix_uint16 length;
    ix_uint16 isExtraOctet = 0;
    ix_uint32 checksum = 0;
    ix_uint16 *pCurPosn;
    ix_uint16 count;

    pIpHdr = (ix_cc_natpt_ipv6_header *)udpMessage; 
    pUdpHdr = (ix_cc_natpt_udp_hdr *)
                (udpMessage + IX_CC_NATPT_IPV6_HDR_LEN);    

    /* First, set the checksum field in the UDP header to zero */
    pUdpHdr->checksum=0;

    /* 
     * Prepare the pseudo header. This header contains the source address
     * destination address, the length of the "upper layer" payload, a 
     * zero field and a next header. See RFC 2460, Sec 8.1 for details.
     */
    ix_ossl_memcpy(&pseudoHdr.srcAddress, &pIpHdr->sourceAddress, 
            IX_CC_NATPT_IPV6ADDRLEN_BYTES);
    ix_ossl_memcpy(&pseudoHdr.dstAddress, &pIpHdr->destAddress, 
            IX_CC_NATPT_IPV6ADDRLEN_BYTES);
    
    pseudoHdr.pktLength = (ix_uint32) pIpHdr->payloadLength; 
    pseudoHdr.unused = 0;
    pseudoHdr.nextHdr = IX_CC_NATPT_NEXT_HDR_UDP;

    /* Calculate checksum for the pseudo header */
    length = sizeof(ix_cc_natpt_ipv6_pseudo_header)/2;
    pCurPosn = (ix_uint16 *) &pseudoHdr;
    while(length)
    {
        checksum += (ix_uint16)*pCurPosn;
        pCurPosn++;
        length--;
    }
    
    /* Calculate the checksum for the rest of the UDP packet */
    pCurPosn = (ix_uint16 *)(udpMessage + IX_CC_NATPT_IPV6_HDR_LEN);

    /* 
     * Check if the packet does not end at a 16-bit boundary. This will 
     * be used later to add any remaining octets to the sum
     */

    length = IX_NTOH16(pIpHdr->payloadLength);
    isExtraOctet = length%2;

    /*
     * The length stored in the 'length' variable is in octets. Convert
     * it into the # of 16 bit words
     */
    length=length/2;

    for(count = 0; count < length; count++, pCurPosn++)
    {
        checksum+=(ix_uint16)*pCurPosn;
    }
    
    /* Now add the remaining octet, if any, to the sum */
    if(isExtraOctet)
    {
        checksum+= *(ix_uint8 *)pCurPosn;
    }

    /*
     * Now add the upper 16 bits to the lower 16 bits until we end up with a 
     * 16 bit value. This is the 1's complement sum
     */
    while(checksum>>16)
    {
        checksum = (checksum>>16) + (checksum & 0xffff);
    }
    
    /* 
     * Finally, take the 1's complement of the calculated 1's complement sum
     * and store it in the checksum field of the UDP header
     */
    pUdpHdr->checksum = (ix_uint16)~checksum;

}


