/*******************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 2003 By Intel Corporation.  All rights reserved.  
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *******************************************************************************/

/******************************************************************************
 *
 * File:  ipv6_fwder.c
 *
 * Abstract:
 *      Main IPv6 Forwarder Code.
 *
 * Exported APIs: 
 *   ipv6Forwarder(pIpv6HdrOut, ipv6WrOffset, pIpv6HdrIn, ipv6RdOffset,
 *                 ipv6WrMemType, ipv6RdMemType)
 *
 *****************************************************************************/


#ifndef	__IPV6_FWDER_C__
#define	__IPV6_FWDER_C__

#define DONT_HASH_LOWER_64_BITS_OF_DIP

/* Includes */
#include "ipv6_fwder_uc.h"
#include "ipv6.h"

#include "ipv6_fwder_init.c"
#include "ipv6_fwder_util.c"

/* Macro to calculate trie offset */

#define CALC_TRIE_OFFSET(__da, __shf, __trie) \
    (((((__da) >> (__shf)) & 0xff) + ((__trie) << 8)) << 2)


/******************************************************************************
 * ipv6HandleError
 *
 *		Descripton 	:
 *        Check an error result, set the exception code if needed, and set
 *        dlNextBlock.
 *		
 *		Outputs		:
 *          exception code (if result indicates exception) and dlNextBlock are
 *          updated.
 *
 *		Inputs		:
 *          error :      Error code.
 * 
 *      Returns:
 *          None.
 */
INLINE	void ipv6HandleError(
    unsigned int error)
{
    if (error == IX_DROP)
    {
        dlNextBlock = IX_DROP;
    }
    else
    {
        dl_set_exception(BID_IPV6, error);
        dlNextBlock = IX_EXCEPTION;
    }

    return;
}

/******************************************************************************
 * ipv6ValidateHeader
 *
 *		Descripton 	:
 *        Performs IPv6 header validations and returns a result code indicating
 *        whether the validations were successful, or if not, whether the
 *        packet should be dropped to sent to the XScale core as an exception.
 *		
 *		Outputs		:
 *          None.
 *
 *		Inputs		:
 *			pInIpv6     :   Buffer from which to read IPv6 header.
 *          ipv6RdOffset :  Offset of start of IPv6 header in read buffer.
 *          memType :       Memory type of IPv6 buffer.
 *          dst127_96 :     First LW of IPv6 destination address.
 *          dst95_64 :      Second LW of IPv6 destination address.
 *          dst63_32 :      Third LW of IPv6 destination address.
 *          dst31_0 :       Fourth LW of IPv6 destination address.
 *          counterBase :   Base of in counters.
 * 
 *      Returns:
 *          IPV6_SUCCESS, IX_DROP or an exception code.
 */
INLINE	unsigned int ipv6ValidateHeader(
	void *pInIpv6,
	unsigned int ipv6RdOffset,
    mem_t memType,
    unsigned int dst127_96,
    unsigned int dst95_64,
    unsigned int dst63_32,
    unsigned int dst31_0,
    unsigned int counterBase)
{
    unsigned int version;
    unsigned int src127_96;
    unsigned int nextHdr;

    version = _buf_byte_extract(
	    (UINT*)pInIpv6,
        ipv6RdOffset + IPV6_VERSION_TRAFFIC_CLASS_HI4_START,
        IPV6_VERSION_TRAFFIC_CLASS_HI4_COUNT, memType);

    if ((version >> 4) != 6)
    {
#ifdef IPV6_STATS
        /* IPV6_IN_HDR_ERRORS: Number of packets in error (version number 
         * mismatch, other format errors, hop count exceeded, errors 
         * discovered in processing their IP options, etc)
         */
        ipv6IncrementCounter(counterBase, IPV6_IN_HDR_ERRORS);
#endif //IPV6_STATS

        return IX_DROP;
    }

    if (dlMeta.packetSize < 40)
    {
#ifdef IPV6_STATS
        /* IPV6_IN_TRUNCATED: Count of IP packets discarded because
         * the packet didn't carry enough data
         */
        ipv6IncrementCounter(counterBase, IPV6_IN_TRUNCATED);
#endif //IPV6_STATS

        return IX_DROP;
    }

#ifdef RFC2373_CHECKS
    /* Validate the source address.  First check for multicast. */

    src127_96 = _buf_byte_extract(
	    (UINT*)pInIpv6,
        ipv6RdOffset + IPV6_SRC_ADDR_127_96_START,
        IPV6_SRC_ADDR_127_96_COUNT, memType);

    if ((src127_96 >> 24) == 0xff)
    {
        return IX_DROP;
    }

    /* Now check for loopback address (::1) */

    if (src127_96 == 0)
    {
        unsigned int src95_64;

        src95_64= _buf_byte_extract(
    	    (UINT*)pInIpv6,
            ipv6RdOffset + IPV6_SRC_ADDR_95_64_START,
            IPV6_SRC_ADDR_95_64_COUNT, memType);

        if (src95_64 == 0)
        {
            unsigned int src63_32;

            src63_32= _buf_byte_extract(
        	    (UINT*)pInIpv6,
                ipv6RdOffset + IPV6_SRC_ADDR_63_32_START,
                IPV6_SRC_ADDR_63_32_COUNT, memType);

            if (src63_32 == 0)
            {
                unsigned int src31_0;

                src31_0= _buf_byte_extract(
            	    (UINT*)pInIpv6,
                    ipv6RdOffset + IPV6_SRC_ADDR_31_0_START,
                    IPV6_SRC_ADDR_31_0_COUNT, memType);

                if (src31_0 == 1)
                {
                    return IX_DROP;
                }
            }
        }
    }
        
    /* Validate the destination address */

    if ((dst127_96 == 0) && (dst95_64 == 0) && (dst63_32 == 0) &&
        ((dst31_0 & ~1) == 0))
    {
#ifdef IPV6_STATS
        /* IPV6_IN_ADDR_ERRORS: Count of packets discarded because
         * the destination address was in error.
         */
     ipv6IncrementCounter(counterBase, IPV6_IN_ADDR_ERRORS);
#endif //IPV6_STATS
        return IX_DROP;
    }

    /* Check the destination address for multicast */

    if ((dst127_96 >> 24) == 0xff)
    {
        return IPV6_EXCP_MULTICAST;
    }

    /* Check the destination address for link local */

    if ((dst127_96 >> 22) == 0x3FA)
    {
        return IPV6_EXCP_LINKLOCAL;
    }

    /* Check the nextheader for Hop-by-hop options header (nextheader == 0) */

    nextHdr= _buf_byte_extract(
	    (UINT*)pInIpv6,
        ipv6RdOffset + IPV6_NEXT_HEADER_START,
        IPV6_NEXT_HEADER_COUNT, memType);

    if (nextHdr == 0)
    {
        return IPV6_EXCP_OPTIONS;
    }

#endif /* RFC2373_CHECKS */

    return IPV6_SUCCESS;

}


/******************************************************************************
 * ipv6CheckTrieCache
 *
 *		Descripton 	:
 *          Check to see if the trie cache needs to be refreshed, and if so
 *          read the updated info from SRAM.
 *		
 *		Outputs		:
 *          Trie cache is updated if necessary.
 *
 *		Inputs		:
 *          None.
 * 
 *      Returns:
 *          None.
 */
INLINE	void ipv6CheckTrieCache(void)
{
    __implicit_read(&ipv6CacheRefreshSignal);

    if (signal_test(&ipv6CacheRefreshSignal))
    {
        SIGNAL sramSig;
        unsigned int futureCount;
        __declspec(sram_read_reg) unsigned int sramTrieCache[IPV6_TRIE_CACHE_LW];

    	sram_read((__declspec(sram_read_reg) void *)sramTrieCache,
    			(volatile __declspec(sram) void *)IPV6_ROUTER_PREFIX_SRAM_BASE,
                IPV6_TRIE_CACHE_LW, sig_done, &sramSig);

        futureCount = local_csr_read(local_csr_timestamp_low);
        futureCount += FUTURE_COUNT;
        local_csr_write(
            local_csr_indirect_ctx_future_count, futureCount);

        __wait_for_all(&sramSig);

        trieCache[0] = sramTrieCache[0];
        trieCache[1] = sramTrieCache[1];
        trieCache[2] = sramTrieCache[2];
        trieCache[3] = sramTrieCache[3];
        trieCache[4] = sramTrieCache[4];
        trieCache[5] = sramTrieCache[5];
        trieCache[6] = sramTrieCache[6];
        trieCache[7] = sramTrieCache[7];
        trieCache[8] = sramTrieCache[8];
        trieCache[9] = sramTrieCache[9];
    }

    return;
}


/******************************************************************************
 * ipv6HashLookup
 *
 *		Descripton 	:
 *          Performs a hash lookup given index and two words to match.
 *          Do chaining till it is either a hit or miss. 
 *		
 *		Outputs		:
 *          Trie cache is updated if necessary.
 *
 *		Inputs		:
 *          nextHop :       Next hop ID.
 *          w0 :            First of the two 32-bit words to match.
 *          w1 :            Last of the two 32-bit words to match.
 * 
 *      Returns:
 *          Next Hop ID.
 */
INLINE	unsigned int ipv6HashLookup(
    unsigned int nextHop,
    unsigned int w0,
    unsigned int w1)
{
    SIGNAL_PAIR hashSigPair;
    SIGNAL sramSig;
    __declspec(sram_read_reg) unsigned int hashLookup[3];
    __declspec(sram_read_reg) unsigned int hashIndexRd[2];
    __declspec(sram_write_reg) unsigned int hashIndexWr[2];
    unsigned int hashIndex;
    unsigned int hashBase = IPV6_HASH_TABLE_SRAM_BASE;

    hashIndexWr[0] = w0;
    hashIndexWr[1] = w1;

    hash_64(hashIndexRd, hashIndexWr, 1, sig_done, &hashSigPair);
    wait_for_all(&hashSigPair);

    hashIndex = (hashIndexRd[0] << 2) & 0xffff;

    /* hash_index now has a 14-bit index value

     * Hash entry has 3 words. 
     * words0 and word1 for matching w0 and w1
     * word3 has: 
     *    0:15  - next index in case of chaining
     *    16:31 - routing_ptr in case of hit
     */

    /* First, check the main table */
	sram_read((__declspec(sram_read_reg) void *)hashLookup,
			(volatile __declspec(sram) void *)hashBase,
            3, ctx_swap, &sramSig);

    if (((hashLookup[0] ^ w0) | (hashLookup[1] ^ w1)) == 0)
    {
        return hashLookup[2] >> 16;
    }

    /* Point to collision table at the end of main table */
    hashBase += (1 << IPV6_NUM_HASH_BITS);
    hashIndex = ((hashLookup[2] << 2) & 0xffff);

    /* Currently, the collision list is traversed until a
     * match is found or a null next index is encountered.
     * This can be limited to a predefined number of tries,
     * after which the search can be performed in the exception
     * path.
     */

    while (hashIndex != 0)
    {
    	sram_read((__declspec(sram_read_reg) void *)hashLookup,
    			(volatile __declspec(sram) void *)hashBase,
                3, ctx_swap, &sramSig);

        if (((hashLookup[0] ^ w0) | (hashLookup[1] ^ w1)) == 0)
        {
            return hashLookup[2] >> 16;
        }

        hashIndex = ((hashLookup[2] << 2) & 0xffff);
    }

    return 0;
}


/******************************************************************************
 * ipv6RouteLookup
 *
 *		Descripton 	:
 *          Performs LPM route lookup. 
 *		
 *		Outputs		:
 *          None.
 *
 *		Inputs		:
 *			pInIpv6     :   Buffer from which to read IPv6 header.
 *          ipv6RdOffset :  Offset of start of IPv6 header in read buffer.
 *          counterBase :   Base of in counters.
 *          inPort :        Input port.
 *          nextHopId :     Next Hop Id from route lookup.
 * 
 *      Returns:
 *          Next Hop Id from lookup.
 */
INLINE	unsigned int ipv6RouteLookup(
	unsigned int dst127_96,
    unsigned int dst95_64,
    unsigned int dst63_32,
    unsigned int dst31_0)
{
    unsigned int nextTrie;
    unsigned int nextHop;
    __declspec(sram_read_reg) unsigned int curTrieEntry;
    unsigned int trieBase16;
    unsigned int trieBase8;
    unsigned int offset;
    SIGNAL sram_sig;

    trieBase8 = IPV6_TRIE_TABLE_SRAM_BASE + (4 << 16);

    /*
     * Local Memory Router's Prefix Matching 
     *

     * Match dest IP with stored Router Prefix 16,8,8,8,8,8,8 bits at a time
     * As long as the bits match, the strides take place in Local Memory
     * At the first mismatch, the cached trie entry is used to start the
     * SRAM trie-walk from that point onwards.

     * Check to see if the first long word of the router prefix is zero.
     * if it is, it was never set - simply perform a regular trie walk on 
     * the IPv6 address.
     */

    if (trieCache[0] != 0)
    {
        unsigned int result;

        result = trieCache[0] ^ dst127_96;

        if ((result >> 16) != 0)
        {
            goto sramTrieWalkHi64;
        }

        nextHop = trieCache[2] >> 16;
        nextTrie = trieCache[2] & 0xffff;
        if (nextTrie == 0)
        {
            return nextHop;
        }

        /* Check next 8 */

        if (((result >> 8) && 0xff) != 0)
        {
            goto lookup_111_104;
        }
        if ((trieCache[3] >> 16) != 0)
        {
            nextHop = trieCache[3] >> 16;
        }
        nextTrie = trieCache[3] & 0xffff;
        if (nextTrie == 0)
        {
            return nextHop;
        }

        /* Check next 8 */

        if ((result && 0xff) != 0)
        {
            goto lookup_103_96;
        }
        if ((trieCache[4] >> 16) != 0)
        {
            nextHop = trieCache[4] >> 16;
        }
        nextTrie = trieCache[4] & 0xffff;
        if (nextTrie == 0)
        {
            return nextHop;
        }

        /* Check next 8 */
        result = trieCache[1] ^ dst95_64;

        if (((result >> 24) && 0xff) != 0)
        {
            goto lookup_95_88;
        }
        if ((trieCache[5] >> 16) != 0)
        {
            nextHop = trieCache[5] >> 16;
        }
        nextTrie = trieCache[5] & 0xffff;
        if (nextTrie == 0)
        {
            return nextHop;
        }

        /* Check next 8 */

        if (((result >> 16) && 0xff) != 0)
        {
            goto lookup_87_80;
        }
        if ((trieCache[6] >> 16) != 0)
        {
            nextHop = trieCache[6] >> 16;
        }
        nextTrie = trieCache[6] & 0xffff;
        if (nextTrie == 0)
        {
            return nextHop;
        }

        /* Check next 8 */

        if (((result >> 8) && 0xff) != 0)
        {
            goto lookup_79_72;
        }
        if ((trieCache[7] >> 16) != 0)
        {
            nextHop = trieCache[7] >> 16;
        }
        nextTrie = trieCache[7] & 0xffff;
        if (nextTrie == 0)
        {
            return nextHop;
        }

        /* Check next 8 */

        if ((result && 0xff) != 0)
        {
            goto lookup_71_64;
        }
        if ((trieCache[8] >> 16) != 0)
        {
            nextHop = trieCache[8] >> 16;
        }
        nextTrie = trieCache[8] & 0xffff;
        if (nextTrie == 0)
        {
            return nextHop;
        }

        /* Matched all 64 bits of router prefix. */

        goto sramTrieWalkLo64;

    }


sramTrieWalkHi64:
    /* Read the base trie entry */
    trieBase16 = IPV6_TRIE_TABLE_SRAM_BASE;
    offset = (dst127_96 >> 14) & ~0x3;

	sram_read((__declspec(sram_read_reg) void *)&curTrieEntry,
			(volatile __declspec(sram) void *)(trieBase16 + offset),
            1, ctx_swap, &sram_sig);

    nextHop = curTrieEntry >> 16;
    nextTrie = curTrieEntry & 0xffff;
    if (nextTrie == 0)
    {
        return nextHop;
    }

lookup_111_104:
    offset = CALC_TRIE_OFFSET(dst127_96, 8, nextTrie);

	sram_read((__declspec(sram_read_reg) void *)&curTrieEntry,
			(volatile __declspec(sram) void *)(trieBase8 + offset),
            1, ctx_swap, &sram_sig);

    if ((curTrieEntry >> 16) != 0)
    {
        nextHop = curTrieEntry >> 16;
    }
    nextTrie = curTrieEntry & 0xffff;
    if (nextTrie == 0)
    {
        return nextHop;
    }

lookup_103_96:
    offset = CALC_TRIE_OFFSET(dst127_96, 0, nextTrie);

	sram_read((__declspec(sram_read_reg) void *)&curTrieEntry,
			(volatile __declspec(sram) void *)(trieBase8 + offset),
            1, ctx_swap, &sram_sig);

    if ((curTrieEntry >> 16) != 0)
    {
        nextHop = curTrieEntry >> 16;
    }
    nextTrie = curTrieEntry & 0xffff;
    if (nextTrie == 0)
    {
        return nextHop;
    }

lookup_95_88:
    offset = CALC_TRIE_OFFSET(dst95_64, 24, nextTrie);

	sram_read((__declspec(sram_read_reg) void *)&curTrieEntry,
			(volatile __declspec(sram) void *)(trieBase8 + offset),
            1, ctx_swap, &sram_sig);

    if ((curTrieEntry >> 16) != 0)
    {
        nextHop = curTrieEntry >> 16;
    }
    nextTrie = curTrieEntry & 0xffff;
    if (nextTrie == 0)
    {
        return nextHop;
    }

lookup_87_80:
    offset = CALC_TRIE_OFFSET(dst95_64, 16, nextTrie);

	sram_read((__declspec(sram_read_reg) void *)&curTrieEntry,
			(volatile __declspec(sram) void *)(trieBase8 + offset),
            1, ctx_swap, &sram_sig);

    if ((curTrieEntry >> 16) != 0)
    {
        nextHop = curTrieEntry >> 16;
    }
    nextTrie = curTrieEntry & 0xffff;
    if (nextTrie == 0)
    {
        return nextHop;
    }

lookup_79_72:
    offset = CALC_TRIE_OFFSET(dst95_64, 8, nextTrie);

	sram_read((__declspec(sram_read_reg) void *)&curTrieEntry,
			(volatile __declspec(sram) void *)(trieBase8 + offset),
            1, ctx_swap, &sram_sig);

    if ((curTrieEntry >> 16) != 0)
    {
        nextHop = curTrieEntry >> 16;
    }
    nextTrie = curTrieEntry & 0xffff;
    if (nextTrie == 0)
    {
        return nextHop;
    }

lookup_71_64:
    offset = CALC_TRIE_OFFSET(dst95_64, 0, nextTrie);

	sram_read((__declspec(sram_read_reg) void *)&curTrieEntry,
			(volatile __declspec(sram) void *)(trieBase8 + offset),
            1, ctx_swap, &sram_sig);

    if ((curTrieEntry >> 16) != 0)
    {
        nextHop = curTrieEntry >> 16;
    }
    nextTrie = curTrieEntry & 0xffff;
    if (nextTrie == 0)
    {
        return nextHop;
    }

#ifndef DONT_HASH_LOWER_64_BITS_OF_DIP
    /*
     * Hash Lookup
     *

     * Matched 64 bits, do hash on remaining 64
     * next-trie pointer is pointing to hash table
     */

    nextHop = ipv6HashLookup(nextHop, dst63_32, dst31_0);
    return nextHop;

#endif

sramTrieWalkLo64:
    /* LPM on lower 64 bits of dest IP (SRAM Trie walk) */

    /* Bits 63:56 */
    offset = CALC_TRIE_OFFSET(dst63_32, 24, nextTrie);

	sram_read((__declspec(sram_read_reg) void *)&curTrieEntry,
			(volatile __declspec(sram) void *)(trieBase8 + offset),
            1, ctx_swap, &sram_sig);

    if ((curTrieEntry >> 16) != 0)
    {
        nextHop = curTrieEntry >> 16;
    }
    nextTrie = curTrieEntry & 0xffff;
    if (nextTrie == 0)
    {
        return nextHop;
    }


    /* Bits 55:48 */
    offset = CALC_TRIE_OFFSET(dst63_32, 16, nextTrie);

	sram_read((__declspec(sram_read_reg) void *)&curTrieEntry,
			(volatile __declspec(sram) void *)(trieBase8 + offset),
            1, ctx_swap, &sram_sig);

    if ((curTrieEntry >> 16) != 0)
    {
        nextHop = curTrieEntry >> 16;
    }
    nextTrie = curTrieEntry & 0xffff;
    if (nextTrie == 0)
    {
        return nextHop;
    }

    /* Bits 47:40 */
    offset = CALC_TRIE_OFFSET(dst63_32, 8, nextTrie);

	sram_read((__declspec(sram_read_reg) void *)&curTrieEntry,
			(volatile __declspec(sram) void *)(trieBase8 + offset),
            1, ctx_swap, &sram_sig);

    if ((curTrieEntry >> 16) != 0)
    {
        nextHop = curTrieEntry >> 16;
    }
    nextTrie = curTrieEntry & 0xffff;
    if (nextTrie == 0)
    {
        return nextHop;
    }

    /* Bits 39:32 */
    offset = CALC_TRIE_OFFSET(dst63_32, 0, nextTrie);

	sram_read((__declspec(sram_read_reg) void *)&curTrieEntry,
			(volatile __declspec(sram) void *)(trieBase8 + offset),
            1, ctx_swap, &sram_sig);

    if ((curTrieEntry >> 16) != 0)
    {
        nextHop = curTrieEntry >> 16;
    }
    nextTrie = curTrieEntry & 0xffff;
    if (nextTrie == 0)
    {
        return nextHop;
    }

    /* Bits 31:24 */
    offset = CALC_TRIE_OFFSET(dst31_0, 24, nextTrie);

	sram_read((__declspec(sram_read_reg) void *)&curTrieEntry,
			(volatile __declspec(sram) void *)(trieBase8 + offset),
            1, ctx_swap, &sram_sig);

    if ((curTrieEntry >> 16) != 0)
    {
        nextHop = curTrieEntry >> 16;
    }
    nextTrie = curTrieEntry & 0xffff;
    if (nextTrie == 0)
    {
        return nextHop;
    }

    /* Bits 23:16 */
    offset = CALC_TRIE_OFFSET(dst31_0, 16, nextTrie);

	sram_read((__declspec(sram_read_reg) void *)&curTrieEntry,
			(volatile __declspec(sram) void *)(trieBase8 + offset),
            1, ctx_swap, &sram_sig);

    if ((curTrieEntry >> 16) != 0)
    {
        nextHop = curTrieEntry >> 16;
    }
    nextTrie = curTrieEntry & 0xffff;
    if (nextTrie == 0)
    {
        return nextHop;
    }

    /* Bits 15:8 */
    offset = CALC_TRIE_OFFSET(dst31_0, 8, nextTrie);

	sram_read((__declspec(sram_read_reg) void *)&curTrieEntry,
			(volatile __declspec(sram) void *)(trieBase8 + offset),
            1, ctx_swap, &sram_sig);

    if ((curTrieEntry >> 16) != 0)
    {
        nextHop = curTrieEntry >> 16;
    }
    nextTrie = curTrieEntry & 0xffff;
    if (nextTrie == 0)
    {
        return nextHop;
    }

    /* Bits 7:0 */
    offset = CALC_TRIE_OFFSET(dst31_0, 0, nextTrie);

	sram_read((__declspec(sram_read_reg) void *)&curTrieEntry,
			(volatile __declspec(sram) void *)(trieBase8 + offset),
            1, ctx_swap, &sram_sig);

    if ((curTrieEntry >> 16) != 0)
    {
        nextHop = curTrieEntry >> 16;
    }
    nextTrie = curTrieEntry & 0xffff;

    return nextHop;
}


/******************************************************************************
 * ipv6Forwarder
 *
 *		Descripton 	:
 *        Verify IPv6 header and perform LPM on the packet.
 *		
 *		Outputs		:
 *			pOutIpv6    :   Buffer used to update IPv6 header.
 *
 *		Inputs		:
 *			pInIpv6     :   Buffer from which to read IPv6 header.
 *          ipv6WrOffset :  Offset of start of IPv6 header in write buffer.
 *          ipv6RdOffset :  Offset of start of IPv6 header in read buffer.
 *          ipv6WrMemType : Memory type of IPv6 output buffer.
 *          ipv6RdMemType : Memory type of IPv6 input buffer.
 * 
 *      Returns:
 *          None.
 */
INLINE	void ipv6Forwarder(
    void *pOutIpv6,
    unsigned int ipv6WrOffset,
	void *pInIpv6,
	unsigned int ipv6RdOffset,
    mem_t ipv6WrMemType,
    mem_t ipv6RdMemType)
{
    unsigned int counterBase;
    unsigned int outCounterBase;
    unsigned int inPort;
    unsigned int dst127_96, dst95_64, dst63_32, dst31_0;
    unsigned int nextHopId;
    unsigned int result;
    unsigned int classId;

    if (dlNextBlock != BID_IPV6)
    {
        return;
    }

#ifdef IPV6_STATS
    counterBase = dlMeta.inputPort;
    counterBase <<= IPV6_IN_STATS_PER_PORT_SHIFT;
    counterBase += IPV6_STATS_TABLE_SRAM_BASE;

    /* Increment the number of packet received */
    ipv6IncrementCounter(counterBase, IPV6_IN_RECV);

    /* Increment the count, statsInOctets by the packet size. */

    ipv6AddToCounter(counterBase, IPV6_IN_OCTETS, dlMeta.packetSize);
#else
    counterBase = 0;
#endif

    /* Extract the address lwords and validate the header */

    dst127_96 = _buf_byte_extract(
	    (UINT*)pInIpv6,
        ipv6RdOffset + IPV6_DST_ADDR_127_96_START,
        IPV6_DST_ADDR_127_96_COUNT, ipv6RdMemType);

    dst95_64 = _buf_byte_extract(
	    (UINT*)pInIpv6,
        ipv6RdOffset + IPV6_DST_ADDR_95_64_START,
        IPV6_DST_ADDR_95_64_COUNT, ipv6RdMemType);

    dst63_32 = _buf_byte_extract(
	    (UINT*)pInIpv6,
        ipv6RdOffset + IPV6_DST_ADDR_63_32_START,
        IPV6_DST_ADDR_63_32_COUNT, ipv6RdMemType);

    dst31_0 = _buf_byte_extract(
	    (UINT*)pInIpv6,
        ipv6RdOffset + IPV6_DST_ADDR_31_0_START,
        IPV6_DST_ADDR_31_0_COUNT, ipv6RdMemType);

    result = ipv6ValidateHeader(
            pInIpv6, ipv6RdOffset, ipv6RdMemType, dst127_96, dst95_64,
            dst63_32, dst31_0, counterBase);
    if (result != IPV6_SUCCESS)
    {
        ipv6HandleError(result);
        return;
    }

    /* Perform lookup */
    nextHopId = ipv6RouteLookup(dst127_96, dst95_64, dst63_32, dst31_0);

    /* Refresh trie cache if necessary */
    ipv6CheckTrieCache();

    /* Post-lookup validations */

    inPort = dlMeta.inputPort | (THIS_BLADE_ID << 24);
    result = ipv6ProcessNextHopInfo(
                pInIpv6, ipv6RdOffset, ipv6RdMemType,
                #ifdef IPV6_LOADBALANCE
                dst31_0,
                #endif
                counterBase, inPort, nextHopId);
    if (result != IPV6_SUCCESS)
    {
        ipv6HandleError(result);
        return;
    }

#ifdef IPV6_STATS
    outCounterBase = IPV6_STATS_TABLE_SRAM_BASE +
                   (IPV6_IN_STATS_PER_PORT)*(PORTS_PER_BLADE);
    outCounterBase += (dlMeta.fabricPort << IPV6_OUT_STATS_PER_BLADE_SHIFT);
    outCounterBase += (dlMeta.outputPort << IPV6_OUT_STATS_PER_PORT_SHIFT);

    ipv6IncrementCounter(outCounterBase, IPV6_OUT_FORW);
#endif

    /* Update the header */

    result = ipv6UpdateHeader(
        (unsigned int *)pOutIpv6, ipv6WrOffset, ipv6WrMemType,
        (unsigned int *)pInIpv6, ipv6RdOffset, ipv6RdMemType);

    if (result != IPV6_SUCCESS)
    {
        ipv6HandleError(result);
        return;
    }

    /* Finish updating the meta-data */

    classId = _buf_byte_extract(
	    (UINT*)pInIpv6,
        ipv6RdOffset + IPV6_TRAFFIC_CLASS_LO4_FLOW_LABEL_HI4_START,
        IPV6_TRAFFIC_CLASS_LO4_FLOW_LABEL_HI4_COUNT, ipv6RdMemType);

    dlMeta.classId = classId >> 4;

#ifdef IPV6_STATS
    ipv6IncrementCounter(outCounterBase, IPV6_OUT_TRANSMITS);
    ipv6AddToCounter(outCounterBase, IPV6_OUT_OCTETS, dlMeta.packetSize);
#endif

    dlNextBlock = IPV6_NEXT_BLOCK;
    return;
}

#endif	/* __IPV6_FWDER_C__ */