/*******************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 2003 By Intel Corporation.  All rights reserved.  
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *******************************************************************************/

/******************************************************************************
 *
 * File:  ipv6_fwder_util.c
 *
 * Abstract:
 *      IPv6 Forwarder Utility Routines.
 *
 *****************************************************************************/


#ifndef	__IPV6_FWDER_UTIL_C__
#define	__IPV6_FWDER_UTIL_C__

/**
 *************************************************************************
 * Nexthop Information structure
 *
 * Purpose
 *     This structure stores the information useful in forwarding
 *     the packet.
 *
 ************************************************************************/
typedef	__declspec(packed) union 	
{
    struct
    {
        unsigned int reserved1:24;
        unsigned int flags:8;

        unsigned int fabricPortId:8;
        unsigned int nexthopIdType:4;
        unsigned int reserved2:4;
        unsigned int nexthopId:16;

        uint16_t	mtu;
        uint16_t	outputPort;

        uint32_t	reserved3;
    } regNextHop;

    struct
    {
        unsigned int status:8;
        unsigned int reserved1:16;
        unsigned int flags:8;

        unsigned int nhid0:16;
        unsigned int nhid1:16;

        unsigned int nhid2:16;
        unsigned int nhid3:16;

        uint32_t	reserved2;
    } intNextHop;

	unsigned int value[4];      /**< aggregate for the above fields */

} ipv6_nexthop_info_t;

/* Macros to check next hop flags */

#define NH_FLAG_SET(__nh, __f) \
    ((((__declspec(sram_read_reg) unsigned int *)&(__nh))[IPV6_NH_FLAGS_LW] & (__f)) != 0)

#define NH_FLAG_CLEAR(__nh, __f) \
    ((((__declspec(sram_read_reg) unsigned int *)&(__nh))[IPV6_NH_FLAGS_LW] & (__f)) == 0)


/******************************************************************************
 * ipv6IncrementCounter
 *
 *		Descripton 	:
 *        Atomically increment a 32bit counter.
 *		
 *		Outputs		:
 *          Increments the counter value in SRAM
 *
 *		Inputs		:
 *          base :       Base address of counters.
 *          offset :    Offset of counter to be incremented.
 * 
 *      Returns:
 *          None.
 */
INLINE	void ipv6IncrementCounter(
    unsigned int base,
    unsigned int offset)
{
	sram_incr((volatile void __declspec(sram) *)(base + offset));
    return;
}


/******************************************************************************
 * ipv6AddToCounter
 *
 *		Descripton 	:
 *        Increments a 32-bit counter by the amount specified.
 *		
 *		Outputs		:
 *          Updates the counter value in SRAM
 *
 *		Inputs		:
 *          base :       Base address of counters.
 *          offset :     Offset of counter to be updated.
 *          amount :     Amount to be added.
 * 
 *      Returns:
 *          None.
 */
INLINE	void ipv6AddToCounter(
    unsigned int base,
    unsigned int offset,
    unsigned int amount)
{
    __declspec(sram_write_reg) unsigned int value;
    SIGNAL addSig;

    value = amount;
	sram_add(&value, (volatile void __declspec(sram) *)(base + offset),
             ctx_swap, &addSig);
    return;
}



#ifdef IPV6_LOADBALANCE
/******************************************************************************
 * ipv6ProcessIntermediateNextHop
 *
 *		Descripton 	:
 *          Check the intermediate next hop entry and locate a real next hop
 *          entry to use. 
 *
 *
 *          Support for upto 4-equal cost next-hops
 *          This nexthop_index could point to an "intermediate entry", check
 *          to see if it does. If it does not (i.e it points to a real entry)
 *          there is not a whole lot to do, use this nexthop_index. If it does,
 *          read the intermediate entry (16 bytes) and use a generated index
 *          (last two bits of source IP address XOR'd with last two bits of
 *          destination IP address) to pick on of the four equal cost hops and
 *          save that in the nexthop_index. 
 *		
 *		Outputs		:
 *          None.
 *
 *		Inputs		:
 *			pInIpv6     :   Buffer from which to read IPv6 header.
 *          ipv6RdOffset :  Offset of start of IPv6 header in read buffer.
 *          ipv6RdMemType : Memory type of IPv6 header buffer.
 *          dst31_0 :       Last 32 bits of destination address.
 *
 *      In/Out:
 *          pNextHopId :    Next Hop Id from route lookup, which can be a 
 *                          real next hop or an intermediate nexthop.  On output,
 *                          this contains the real next hop .
 * 
 *      Returns:
 *          Index of nexthop to use, or IPV6_NH_NHID_TOTAL if none could 
 *          be found.
 */
INLINE	unsigned int ipv6ProcessIntermediateNextHop(
	void *pInIpv6,
	unsigned int ipv6RdOffset,
    mem_t memType,
	__declspec(sram_read_reg) ipv6_nexthop_info_t *pNhInfo,
    unsigned int dst31_0)
{
    unsigned int src31_0;
    unsigned int count;
    unsigned int index;

    src31_0 = _buf_byte_extract(
	    (UINT*)pInIpv6,
        ipv6RdOffset + IPV6_SRC_ADDR_31_0_START,
        IPV6_SRC_ADDR_31_0_COUNT, memType);

    count = ((src31_0 ^ dst31_0) & 0x3);
    index = count;

    do
    {
        if (pNhInfo->intNextHop.status & (1 << index))
        {
            /* Interface status is UP, so return this one. */
            return index;
        }

        /* Try the next one */
        index++;
        if (index == IPV6_NH_NHID_TOTAL)
        {
            index = 0;
        }
    } while (index != count);

    return IPV6_NH_NHID_TOTAL;
}


/******************************************************************************
 * ipv6GetRealNextHopId
 *
 *		Descripton 	:
 *          Extract a real next hop ID from an intermediate next hop entry
 *          based on the index number (0 - 3).
 *
 *		
 *		Outputs		:
 *          None.
 *
 *		Inputs		:
 *			pIntNhInfo   :  Pointer to intermediate next hop entry.
 *          index :         index of nexthop entry to retrieve.
 *
 *      Returns:
 *          Nexthop ID of real entry.
 */
INLINE	unsigned int ipv6GetRealNextHopId(
	__declspec(sram_read_reg) ipv6_nexthop_info_t *pNhInfo,
    unsigned int index)
{
    if (index == 0)
    {
        return pNhInfo->intNextHop.nhid0;
    }
    else if (index == 1)
    {
        return pNhInfo->intNextHop.nhid1;
    }
    else if (index == 2)
    {
        return pNhInfo->intNextHop.nhid2;
    }
    else
    {
        return pNhInfo->intNextHop.nhid3;
    }

}
#endif


/******************************************************************************
 * ipv6ProcessNextHopInfo
 *
 *		Descripton 	:
 *          Read L3 next-hop structure from SRAM
 *          Perform validations based on information in the next-hop structure
 *          Update next-hop-type and next-hop-id in the meta data
 *		
 *		Outputs		:
 *          Updates nextHopId, nextHopIdType, outputPort and fabricPort in
 *          the meta data.
 *
 *		Inputs		:
 *			pInIpv6     :   Buffer from which to read IPv6 header.
 *          ipv6RdOffset :  Offset of start of IPv6 header in read buffer.
 *          memType :       Memory type of IPv6 header buffer.
 *          dst31_0 :       Last 32 bits of destination address (only needed
 *                          if multiple next hops are used).
 *          counterBase :   Base of in counters.
 *          inPort :        Input port.
 *          nextHopId :     Next Hop Id from route lookup.
 * 
 *      Returns:
 *          IPV6_SUCCESS, IX_DROP or an exception code.
 */
INLINE	unsigned int ipv6ProcessNextHopInfo(
	void *pInIpv6,
	unsigned int ipv6RdOffset,
    mem_t memType,
#ifdef IPV6_LOADBALANCE
    unsigned int dst31_0,
#endif
    unsigned int counterBase,
    unsigned int inPort,
    unsigned int nextHopId)
{
	__declspec(sram_read_reg) ipv6_nexthop_info_t	nhInfo;
	ipv6_nexthop_info_t	*pNhInfo;		/* ptr to NH info in SRAM */
    SIGNAL sram_sig;

    unsigned int l2NextHop;
    unsigned int l2NextHopType;
    unsigned int result;

	pNhInfo = (ipv6_nexthop_info_t *)(IPV6_L3_NEXTHOP_TABLE_SRAM_BASE +
					(nextHopId << IPV6_NH_SIZE_SHIFT));

	sram_read((__declspec(sram_read_reg) void *)&nhInfo,
			(volatile __declspec(sram) void *)pNhInfo,
            IPV6_NH_SIZE_LW, ctx_swap, &sram_sig);

#ifdef IPV6_LOADBALANCE
    if (NH_FLAG_SET(nhInfo, IPV6_NH_FLAGS_MULTIPLE))
    {
        unsigned int realNextHop;

        result = ipv6ProcessIntermediateNextHop(
                    pInIpv6, ipv6RdOffset, memType, &nhInfo, dst31_0);

        if (result >= IPV6_NH_NHID_TOTAL)
        {
            #ifdef IPV6_STATS
                ipv6IncrementCounter(counterBase, IPV6_IN_FORW);
            #endif

            return IPV6_EXCP_LKUPFAILED;
        }

        realNextHop = ipv6GetRealNextHopId(&nhInfo, result);

    	pNhInfo = (ipv6_nexthop_info_t *)(IPV6_L3_NEXTHOP_TABLE_SRAM_BASE +
    					(realNextHop << IPV6_NH_SIZE_SHIFT));

    	sram_read((__declspec(sram_read_reg) void *)&nhInfo,
    			(volatile __declspec(sram) void *)pNhInfo,
                IPV6_NH_SIZE_LW, ctx_swap, &sram_sig);
    }
#endif /* IPV6_LOADBALANCE */

    l2NextHop = nhInfo.regNextHop.nexthopId;
    l2NextHopType = nhInfo.regNextHop.nexthopIdType;

    if (l2NextHop == 0xffff)
    {
        /* No route */
        result = IPV6_EXCP_LKUPFAILED;
    }
    else if ((nhInfo.regNextHop.outputPort | 
             (nhInfo.regNextHop.fabricPortId << 24)) == inPort)
    {
        /* Input port same as output port */
        result = IPV6_EXCP_REDIRECT;
    }
    else if (nhInfo.regNextHop.flags == 0)
    {
        /* Packet is not for local delivery */

#ifdef RFC2373_CHECKS
        /* Check if the source address is link local & drop the packet if so.
         * Note that this is the earliest we can perform this check, as we just
         * now came to know that the packet is not for local delivery
         * Check for link-local source address
         */
         unsigned int srcAddr;

        srcAddr = _buf_byte_extract(
    	    (UINT*)pInIpv6,
            ipv6RdOffset + IPV6_SRC_ADDR_127_96_START,
            IPV6_SRC_ADDR_127_96_COUNT, memType);

        if ((srcAddr >> 22) == 0x3FA)
        {
            result = IX_DROP;
        } else
#endif //RFC2373_CHECKS

        /* Check the packet size against the MTU */

        if (nhInfo.regNextHop.mtu < dlMeta.packetSize)
        {
            result = IPV6_EXCP_MTU;
        }
        else
        {
            result = IPV6_SUCCESS;
        }
    }
    else
    {
        /* Check the flags */
        if (NH_FLAG_SET(nhInfo, IPV6_NH_FLAGS_LOCAL))
        {
            if (l2NextHopType == 0)
            {
                return IPV6_EXCP_LOCAL;
            }
            else
            {
                result =  IPV6_SUCCESS;
            }
        }
        else if (NH_FLAG_SET(nhInfo, IPV6_NH_FLAGS_DOWN))
        {
            result = IPV6_EXCP_LKUPFAILED;
        }
        else if (NH_FLAG_SET(nhInfo, IPV6_NH_FLAGS_DROP))
        {
            result = IX_DROP;
        }
        else
        {
            result = IX_DROP;
        }
    }

    if (result == IPV6_SUCCESS)
    {
        dlMeta.nextHopId = l2NextHop;
        dlMeta.nhidType  = l2NextHopType;
        dlMeta.outputPort = nhInfo.regNextHop.outputPort;
        dlMeta.fabricPort = nhInfo.regNextHop.fabricPortId;
    }

#ifdef IPV6_STATS
    ipv6IncrementCounter(counterBase, IPV6_IN_FORW);
#endif

    return result;

}


/******************************************************************************
 * ipv6UpdateHeader
 *
 *		Descripton 	:
 *          Copy the IPv6 header to the output buffer if necessary.  Update
 *          the header as needed.
 *		
 *		Outputs		:
 *			pOutIpv6    :   Buffer used to update IPv6 header.
 *
 *		Inputs		:
 *			pInIpv6     :   Buffer from which to read IPv6 header.
 *          ipv6WrOffset :  Offset of start of IPv6 header in write buffer.
 *          ipv6WrMemType : Memory type of IPv6 output buffer.
 *          ipv6RdOffset :  Offset of start of IPv6 header in read buffer.
 *          ipv6RdMemType : Memory type of IPv6 input buffer.
 * 
 *      Returns:
 *          IPV6_SUCCESS or an exception code (if hop limit has expired).
 */
INLINE	unsigned int ipv6UpdateHeader(
    unsigned int *pOutIpv6,
    unsigned int ipv6WrOffset,
    mem_t ipv6WrMemType,
	unsigned int *pInIpv6,
	unsigned int ipv6RdOffset,
    mem_t ipv6RdMemType)
{
    unsigned int inStartWd;
    unsigned int outStartWd;
    unsigned int inOffset;
    unsigned int outOffset;
    unsigned int hdrValue;

    inStartWd = ipv6RdOffset >> 2;
    outStartWd = ipv6WrOffset >> 2;

    inOffset = ipv6RdOffset & 0x3;
    outOffset = ipv6WrOffset & 0x3;


    if (inOffset == outOffset)
    {
        _GET_MEM_CLASS_DATA(hdrValue, pInIpv6 + inStartWd + 0,
                            unsigned int, ipv6RdMemType);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 0, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(hdrValue, pInIpv6 + inStartWd + 1,
                            unsigned int, ipv6RdMemType);

        /* Update the hop limit if the offset is zero */
        if (inOffset == 0)
        {
            int hopLimit = hdrValue & 0xff;
            hopLimit--;
            if (hopLimit <= 0)
            {
                return IPV6_EXCP_BAD_HOPLIMIT;
            }
            hdrValue = (hdrValue & ~0xff) | hopLimit;
        }


        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 1, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(hdrValue, pInIpv6 + inStartWd + 2,
                            unsigned int, ipv6RdMemType);

        /* Update the hop limit if the offset is not zero */
        if (inOffset == 1)
        {
            int hopLimit = (hdrValue >> 24) & 0xff;
            hopLimit--;
            if (hopLimit <= 0)
            {
                return IPV6_EXCP_BAD_HOPLIMIT;
            }
            hdrValue = (hdrValue & 0xffffff) | (hopLimit << 24);
        }
        else if (inOffset == 2)
        {
            int hopLimit = (hdrValue >> 16) & 0xff;
            hopLimit--;
            if (hopLimit <= 0)
            {
                return IPV6_EXCP_BAD_HOPLIMIT;
            }
            hdrValue = (hdrValue & 0xff00ffff) | (hopLimit << 16);
        }
        else if (inOffset == 3)
        {
            int hopLimit = (hdrValue >> 8) & 0xff;
            hopLimit--;
            if (hopLimit <= 0)
            {
                return IPV6_EXCP_BAD_HOPLIMIT;
            }
            hdrValue = (hdrValue & 0xffff00ff) | (hopLimit << 8);
        }

        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 2, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(hdrValue, pInIpv6 + inStartWd + 3,
                            unsigned int, ipv6RdMemType);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 3, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(hdrValue, pInIpv6 + inStartWd + 4,
                            unsigned int, ipv6RdMemType);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 4, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(hdrValue, pInIpv6 + inStartWd + 5,
                            unsigned int, ipv6RdMemType);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 5, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(hdrValue, pInIpv6 + inStartWd + 6,
                            unsigned int, ipv6RdMemType);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 6, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(hdrValue, pInIpv6 + inStartWd + 7,
                            unsigned int, ipv6RdMemType);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 7, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(hdrValue, pInIpv6 + inStartWd + 8,
                            unsigned int, ipv6RdMemType);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 8, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(hdrValue, pInIpv6 + inStartWd + 9,
                            unsigned int, ipv6RdMemType);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 9, hdrValue,
                            unsigned int, ipv6WrMemType);

        if (inOffset != 0)
        {
            _GET_MEM_CLASS_DATA(hdrValue, pInIpv6 + inStartWd + 10,
                                unsigned int, ipv6RdMemType);
            _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 10, hdrValue,
                                unsigned int, ipv6WrMemType);
        }

    }
    else if (inOffset < outOffset)
    {
        unsigned int index;
        unsigned int value1;
        unsigned int value2;
        unsigned int shiftAmount = (outOffset - inOffset) * 8;

        if (inStartWd == 0)
        {
            value1 = 0;
        }
        else
        {
            _GET_MEM_CLASS_DATA(value1, pInIpv6 + inStartWd - 1,
                                unsigned int, ipv6RdMemType);
        }

        _GET_MEM_CLASS_DATA(value2, pInIpv6 + inStartWd + 0,
                            unsigned int, ipv6RdMemType);
        hdrValue = dbl_shr(value1, value2, shiftAmount);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 0, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(value1, pInIpv6 + inStartWd + 1,
                            unsigned int, ipv6RdMemType);
        /* Update the hop limit if the in offset is zero */
        if (inOffset == 0)
        {
            int hopLimit = value1 & 0xff;
            hopLimit--;
            if (hopLimit <= 0)
            {
                return IPV6_EXCP_BAD_HOPLIMIT;
            }
            value1 = (value1 & ~0xff) | hopLimit;
        }

        hdrValue = dbl_shr(value2, value1, shiftAmount);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 1, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(value2, pInIpv6 + inStartWd + 2,
                            unsigned int, ipv6RdMemType);
        /* Update the hop limit if the in offset is not zero */
        if (inOffset == 1)
        {
            int hopLimit = (value2 >> 24) & 0xff;
            hopLimit--;
            if (hopLimit <= 0)
            {
                return IPV6_EXCP_BAD_HOPLIMIT;
            }
            value2 = (value2 & 0xffffff) | (hopLimit << 24);
        }
        else if (inOffset == 2)
        {
            int hopLimit = (value2 >> 16) & 0xff;
            hopLimit--;
            if (hopLimit <= 0)
            {
                return IPV6_EXCP_BAD_HOPLIMIT;
            }
            value2 = (value2 & 0xff00ffff) | (hopLimit << 16);
        }

        hdrValue = dbl_shr(value1, value2, shiftAmount);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 2, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(value1, pInIpv6 + inStartWd + 3,
                            unsigned int, ipv6RdMemType);
        hdrValue = dbl_shr(value2, value1, shiftAmount);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 3, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(value2, pInIpv6 + inStartWd + 4,
                            unsigned int, ipv6RdMemType);
        hdrValue = dbl_shr(value1, value2, shiftAmount);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 4, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(value1, pInIpv6 + inStartWd + 5,
                            unsigned int, ipv6RdMemType);
        hdrValue = dbl_shr(value2, value1, shiftAmount);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 5, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(value2, pInIpv6 + inStartWd + 6,
                            unsigned int, ipv6RdMemType);
        hdrValue = dbl_shr(value1, value2, shiftAmount);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 6, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(value1, pInIpv6 + inStartWd + 7,
                            unsigned int, ipv6RdMemType);
        hdrValue = dbl_shr(value2, value1, shiftAmount);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 7, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(value2, pInIpv6 + inStartWd + 8,
                            unsigned int, ipv6RdMemType);
        hdrValue = dbl_shr(value1, value2, shiftAmount);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 8, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(value1, pInIpv6 + inStartWd + 9,
                            unsigned int, ipv6RdMemType);
        hdrValue = dbl_shr(value2, value1, shiftAmount);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 9, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(value2, pInIpv6 + inStartWd + 10,
                            unsigned int, ipv6RdMemType);
        hdrValue = dbl_shr(value1, value2, shiftAmount);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 10, hdrValue,
                            unsigned int, ipv6WrMemType);
    }
    else
    {
        unsigned int index;
        unsigned int value1;
        unsigned int value2;
        unsigned int shiftAmount = (inOffset - outOffset) * 8;


        _GET_MEM_CLASS_DATA(value1, pInIpv6 + inStartWd + 0,
                            unsigned int, ipv6RdMemType);
        _GET_MEM_CLASS_DATA(value2, pInIpv6 + inStartWd + 1,
                            unsigned int, ipv6RdMemType);
        hdrValue = dbl_shl(value1, value2, shiftAmount);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 0, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(value1, pInIpv6 + inStartWd + 2,
                            unsigned int, ipv6RdMemType);
        /* Update the hop limit */
        if (inOffset == 1)
        {
            int hopLimit = (value1 >> 24) & 0xff;
            hopLimit--;
            if (hopLimit <= 0)
            {
                return IPV6_EXCP_BAD_HOPLIMIT;
            }
            value1 = (value1 & 0xffffff) | (hopLimit << 24);
        }
        else if (inOffset == 2)
        {
            int hopLimit = (value1 >> 16) & 0xff;
            hopLimit--;
            if (hopLimit <= 0)
            {
                return IPV6_EXCP_BAD_HOPLIMIT;
            }
            value1 = (value1 & 0xff00ffff) | (hopLimit << 16);
        }
        else if (inOffset == 3)
        {
            int hopLimit = (value1 >> 8) & 0xff;
            hopLimit--;
            if (hopLimit <= 0)
            {
                return IPV6_EXCP_BAD_HOPLIMIT;
            }
            value1 = (value1 & 0xffff00ff) | (hopLimit << 8);
        }

        hdrValue = dbl_shl(value2, value1, shiftAmount);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 1, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(value2, pInIpv6 + inStartWd + 3,
                            unsigned int, ipv6RdMemType);
        hdrValue = dbl_shl(value1, value2, shiftAmount);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 2, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(value1, pInIpv6 + inStartWd + 4,
                            unsigned int, ipv6RdMemType);
        hdrValue = dbl_shl(value2, value1, shiftAmount);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 3, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(value2, pInIpv6 + inStartWd + 5,
                            unsigned int, ipv6RdMemType);
        hdrValue = dbl_shl(value1, value2, shiftAmount);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 4, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(value1, pInIpv6 + inStartWd + 6,
                            unsigned int, ipv6RdMemType);
        hdrValue = dbl_shl(value2, value1, shiftAmount);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 5, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(value2, pInIpv6 + inStartWd + 7,
                            unsigned int, ipv6RdMemType);
        hdrValue = dbl_shl(value1, value2, shiftAmount);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 6, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(value1, pInIpv6 + inStartWd + 8,
                            unsigned int, ipv6RdMemType);
        hdrValue = dbl_shl(value2, value1, shiftAmount);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 7, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(value2, pInIpv6 + inStartWd + 9,
                            unsigned int, ipv6RdMemType);
        hdrValue = dbl_shl(value1, value2, shiftAmount);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 8, hdrValue,
                            unsigned int, ipv6WrMemType);

        _GET_MEM_CLASS_DATA(value1, pInIpv6 + inStartWd + 10,
                            unsigned int, ipv6RdMemType);
        hdrValue = dbl_shl(value2, value1, shiftAmount);
        _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 9, hdrValue,
                            unsigned int, ipv6WrMemType);

        if (outOffset != 0)
        {
            _GET_MEM_CLASS_DATA(value2, pInIpv6 + inStartWd + 11,
                                unsigned int, ipv6RdMemType);
            hdrValue = dbl_shl(value1, value2, shiftAmount);
            _PUT_MEM_CLASS_DATA(pOutIpv6 + outStartWd + 10, hdrValue,
                                unsigned int, ipv6WrMemType);
        }

    }

    return IPV6_SUCCESS;
}

#endif	/* __IPV6_FWDER_UTIL_C__ */