////////////////////////////////////////////////////////////////////////////////
//                                                                     
//                  I N T E L   P R O P R I E T A R Y                   
//                                                                      
//     COPYRIGHT (c)  2001-2002 BY  INTEL  CORPORATION.  ALL RIGHTS          
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//     THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                      
//                        INTEL  CORPORATION                            
//                                                                     
//                     2200 MISSION COLLEGE BLVD                        
//                                                                      
//               SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                      
////////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: ipv6_fwder.uc
//
//      Purpose: Contains the core IPv6 forwarder code. This contains macro
//                 that perform the LPM, IPv6 header validations.
//
//      History:
//
//
//      Date            Comment                                     By
// -----------------------------------------------------------------------------
//
//      04/12/2002      Created                                     
//
//
////////////////////////////////////////////////////////////////////////////////

#ifndef    __IPV6_FWDER_UC__
#define    __IPV6_FWDER_UC__

////////////////////////////////////////////////////////////////////////////////

//    Include Files
#include "xbuf.uc"
#include "ipv6.h"
#include <ipv6_fwder_uc.h>
#include "stdmac.uc"
#include "bytefield.uc"
#include "sram.uc"

#include "ipv6_fwder_init.uc"

#include "nh_extract.uc"

#define DONT_HASH_LOWER_64_BITS_OF_DIP 1


#ifdef IPV6_STATS
#define_eval   IPV6_OUT_STATS_PER_BLADE_SHIFT  \
                    (IPV6_OUT_STATS_PER_PORT_SHIFT+PORTS_PER_BLADE_SHIFT)
#endif


#ifdef AI_DEBUG
//
// Useful Print Macros
//
// Print to an absolute address -- defined in the macro
#macro  _ipv6_print_absolute(arb)
.begin
	.reg $temp
	.reg sram_address
	.sig nh_sram_signal
	.reg var

	   immed32[sram_address, 0x007e738]
	   immed32[var, IPV6_L3_NEXTHOP_TABLE_SRAM_BASE] 
	   alu[$temp, --, B, var]

	    // Read IPV6_NH_SIZE_LW long words from this location 
	    // into the SRAM xfer registers.

	    sram[write,
	         $temp,
	         0,
	         sram_address,
	         1],
	         ctx_swap[nh_sram_signal]
.end
#endm

// Print a "value" at an "offset" from the start of the next hop table

#macro  _ipv6_print_something(value, offset)
.begin

   .reg sram_address
   .reg $temp
   .sig nh_sram_signal
   .reg var

   immed32[sram_address, IPV6_L3_NEXTHOP_TABLE_SRAM_BASE]
   alu_shf[var, --, B, offset, <<4]
   alu[sram_address, sram_address, +, var]
   alu[$temp, --, B, value]

    // Read IPV6_NH_SIZE_LW long words from this location 
    // into the SRAM xfer registers.

    sram[write,            \
         $temp,            \
         0,                \
         sram_address,     \
         1],               \
         ctx_swap[nh_sram_signal]

.end
#endm

#endif // AI_DEBUG

////////////////////////////////////////////////////////////////////////////////
//
// _ipv6_process_nexthop_info
//
// Description: 
//
//     Read L3 next-hop structure from SRAM
//     Perform validations based on information in the next-hop structure
//     Update next-hop-type and next-hop-id in the meta data
//
// Outputs:
//
//     out_result - result of this macro.  Can be:
//         IPV6_SUCCESS : if validations passed and the packet can be forwarded
//         IPV6_FAILURE : if the packet needs to be dropped
//         Other values : if not success, and not failure, it is exception code
//
//     out_output_port  : Output port number for the interface
//
//     out_blade_id     : Output blade id for the outgoing interface
//
//     out_nhid_type    : type of next-hop-id
//
//     io_nexthop_id    : table index found in L3 structure ( could be L2 
//                        or other) 
//
// Condition Code       : The last instruction of this macro sets out_result
//                        If SUCCESS, the Z flag will be set
//
// Inputs:
//
//     io_nexthop_id    : L3 next-hop table index ( found by the lookup )
//
//     in_counter_base  : Base of statistics counter for the relevant interface
//
//     IN_IPV6_OFFSET   : The relative start byte position of IP header w.r.t 
//                        start of the buffer
//     ipv6_hdr         : Location in memory where the IPv6 header lives.
//     in_packet_len    : Size of the packet being processed. (For MTU check).
//
////////////////////////////////////////////////////////////////////////////////

#macro  _ipv6_process_nexthop_info(out_result,      \
                                   out_output_port, \
                                   out_blade_id,    \
                                   out_nhid_type,   \
                                   io_nexthop_id,   \
                                   in_packet_len,   \
                                   in_port,         \
                                   in_counter_base, \
                                   ipv6_hdr,        \
                                   IN_IPV6_OFFSET)
.begin
    .reg flags
    .reg nexthop_info_base
    .reg tmp
    .sig nh_sram_signal
    .set out_blade_id

    xbuf_alloc($nexthop, 4, read)

    ;; next-hop array base in SRAM
    immed32(nexthop_info_base, IPV6_L3_NEXTHOP_TABLE_SRAM_BASE)

    // We have the index of the next hop info entry. Each entry is of
    // 16 bytes. So, to get the relative address of the entry multiply
    // the index by 16.
    alu_shf[io_nexthop_id, --, B, io_nexthop_id, <<IPV6_NH_SIZE_SHIFT]
    ;;
    ;; Read next hop information from SRAM
    sram[read, $nexthop[0], io_nexthop_id, nexthop_info_base, IPV6_NH_SIZE_LW],
               ctx_swap[nh_sram_signal]

    ;; Get nexthop-id and its type from the strucure
    xbuf_extract[out_nhid_type, $nexthop, 0, IPV6_NH_ID_TYPE]
    alu[out_nhid_type, --, B, out_nhid_type, >>4] 

    ;; extract the output port
    xbuf_extract[out_output_port, $nexthop, 0, IPV6_NH_OUTPORT]

    ;; extract the L2 next hop ID
    xbuf_extract[io_nexthop_id, $nexthop, 0, IPV6_NH_ID]

    //
    // The next hop ID after a lookup could be zero, which points to the 0th
	// entry in the L3 next hop table. If the L2 index in this entry is 0xFFFF
	// it means no route exists. Otherwise use the L2 index in this entry to 
	// send the packet to egress.
	//
    .begin
		.reg var
		immed[var, 0xffff]
		alu[--, io_nexthop_id, -, var]
    	beq[no_route#]
	.end


    // We now start processing the Flags. These Flags normally result in
    // exception packets. Normally, these Flags are clear. If any of
    // the Flags is set, we will process packet in non-critical code path.
    // The Flags are in bits 16 thru 31.
	// Note: Flags are checked before we check for a possible ICMP redirect.
	// This handles a condition where, for instance, a next hop entry might
	// have the LOCAL flag set (for locally destined packets), in which 
	// case the redirect check is unnecessary. 
	// Note: If new flags are added and checked for, make sure that this check
	// (checking for any flag set) can still be performed ahead of the ICMP
	// redirect check.
	//
    ;;
    ;; Check if any Flags are set
    xbuf_extract[flags, $nexthop, 0, IPV6_NH_FLAGS]
    alu[--, --, B, flags]
    bne[check_flags#]    

    //
    // Check if the packet has to go out thru the same interface from
    // which it is received. This requires an ICMP redirect message to
    // be sent to the sender of the packet. The interface is identified
    // by a combination of blade ID and output interface on that blade.

    // OR in the output blade id into the Most Significant byte of the 
    // output port The in port was simlarly ORed with this blade id
    xbuf_extract[out_blade_id, $nexthop, 0, IPV6_NH_BLADE_ID]
    alu_shf[tmp, out_output_port, OR, out_blade_id,<<24]

    alu[--, in_port, -, tmp]
    beq[nh_icmp_redirect#]

  
#ifdef RFC2373_CHECKS
    //At this point we know that this packet is NOT for local delivery
    //Check if the source address is link local & drop the packet if so.
    //Note that this is the earliest we can perform this check, as we just
    //now came to know that the packet is not for local delivery
    ;; Check for link-local source address
    .begin
        .reg src_addr

        xbuf_extract(src_addr, ipv6_hdr, IN_IPV6_OFFSET, IPV6_SRC_ADDR_127_96)
        sub_shf_right[--, 0x3FA, src_addr, 22]
        beq[nh_drop#]
    .end
#endif //RFC2373_CHECKS

    // If our frame size is > than the MTU, we send the packet to the core 
    ;; MTU Check
    .begin
        .reg mtu

        xbuf_extract[mtu, $nexthop, 0, IPV6_NH_MTU]
        alu[--, mtu, -, in_packet_len ]
        blt[nh_larger_than_mtu#]
    .end

    ;; 
    ;; Success
    ;;

    // We passed all the tests 
    // set the result code to SUCCESS and exit macro.
    br[proc_nexthop_info_end#], defer[2]
      immed[out_result, IPV6_SUCCESS]
      nop


    //
    // Exception packet processing code
    //
    
no_route#:
    // No route for the destination address, send to core
    br[proc_nexthop_info_end#], defer[1]
      immed[out_result, IPV6_EXCP_LKUPFAILED]

    // Check the flags in more detail. We have a particular
    // order in which we check the flags. Also, currently these 
    // flags are mutually exclusive.

check_flags#:

    ;; Check if Local Delivery bit is set
    br_bset[flags, IPV6_NH_FLAGS_LOCAL_BIT, nh_local_delivery#]

    ;; Check if port down bit is set
    br_bset[flags, IPV6_NH_FLAGS_DOWN_BIT, nh_unreacheable#]

    ;; Check if drop everything bit is set
    br_bset[flags, IPV6_NH_FLAGS_DROP_BIT, nh_drop#]

    // Currently, we have separate labels for all these exceptions
    // If it turns out that the only work we have to do is set exception 
    // code, we can get it done in the defer slots

nh_drop#:
    // Otherwise, simply drop the packet for unknown flag settings.
    // drop the packet.
    br[proc_nexthop_info_end#], defer[1]
      immed[out_result, IX_DROP]

nh_unreacheable#:
    // when the outport is down, an ICMP message will be sent to 
    // the sender of the packet. The core portion can do this.
    br[proc_nexthop_info_end#], defer[1]
      immed[out_result, IPV6_EXCP_LKUPFAILED]

nh_local_delivery#:
    // The packet is for local applications
    // If the nexthop-id-type is not forwarder, simply send the packet ahead
    alu[--, --, B, out_nhid_type]
    bne[proc_nexthop_info_end#], defer[1]
      immed[out_result, IPV6_SUCCESS]

    dl_set_exception_priority[1]
    br[proc_nexthop_info_end_no_incr#], defer[1]
      immed[out_result, IPV6_EXCP_LOCAL]

nh_larger_than_mtu#:
    br[proc_nexthop_info_end#], defer[1]
      immed[out_result, IPV6_EXCP_MTU]

nh_icmp_redirect#:
    // ICMP redirect required.
    br[proc_nexthop_info_end#], defer[1]
      immed[out_result, IPV6_EXCP_REDIRECT]

proc_nexthop_info_end#:
#ifdef IPV6_STATS
    // IPV6_IN_FORW: count of non-local packets for which a lookup is done
    //        Increment counter if it is not locally destined
    //        Ideally we would have done this *before* the lookup, but we don't
    //        know then if the packet is locally destined
    _ipv6_increment_counter(in_counter_base, IPV6_IN_FORW)
#endif //IPV6_STATS

proc_nexthop_info_end_no_incr#:
    xbuf_free($nexthop)
    
.end

#endm


///////////////////////////////////////////////////////////////////////////
//
// _ipv6_increment_counter()
//
// Description: 
//
//   Atomically increment a 32bit counter.
//
// Outputs:
//
//   Increments the counter value in SRAM
//
// Inputs:
//
//   in_addr   - base address for counters. Should be a GPR.
//   OFFSET    - a constant offset (should be 7bits maximum).
//
//////////////////////////////////////////////////////////////////////////
#macro _ipv6_increment_counter(in_addr, OFFSET)

    sram[incr, --, OFFSET, in_addr]
#endm


////////////////////////////////////////////////////////////////////////////////
//
// _ipv6_add_to_counter()
//
// Description: 
//
//   Increments a 32-bit counter by the amount specified
//
// Outputs:
//   Updates the counter value in SRAM
//
// Inputs:
//
//   in_addr   - base address for counters. Should be a GPR.
//   OFFSET    - a constant offset (should be 7bits maximum).
//   add_me    - amount to add to the counter
//    
////////////////////////////////////////////////////////////////////////////////
#macro _ipv6_add_to_counter(in_addr, OFFSET, add_me)

    .begin
        .reg $xfer
        .sig sig_count_read

        move($xfer, add_me)
        //Note that we are swapping out here. This is a performance hit
        sram[add, $xfer, OFFSET, in_addr], ctx_swap[sig_count_read]
    .end
#endm

////////////////////////////////////////////////////////////////////////////////
//
// _ipv6_header_validations[]
//
// Description: 
//
//     Performs IPv6 header validations sets up result 
//
// Outputs:
//
//    out_result     -  result of this macro. Can be one of:
//                       IX_SUCCESS
//                       IX_DROP 
//                       Exception_code -  reason for the exception
//
// Inputs
//
//    ipv6_hdr       - Buffer containing IPv6 header   
//
//    IN_IPV6_OFFSET - The relative start byte position of IP header w.r.t 
//                     start of the buffer
//
//    da_127_96      - GPR containing IP destination address bits [127..96]
//
//    da_95_64       - GPR containing IP destination address bits [95..64]
//
//    da_63_32       - GPR containing IP destination address bits [63..32]
//
//    da_31_0        - GPR containing IP destination address bits [31..0]
//
//    ( The destination IP is passed in for avoiding additional extracts )
//
////////////////////////////////////////////////////////////////////////////////

#macro _ipv6_header_validations(out_result,     \
                                ipv6_hdr,       \
                                IN_IPV6_OFFSET, \
                                da_127_96,      \
                                da_95_64,       \
                                da_63_32,       \
                                da_31_0)
.begin
    .reg sa_127_96, sa_95_64, sa_63_32, sa_31_0
    .reg version
    .reg packetsize
    .reg nextheader

    ;; If the version is  not 0x6, drop
     xbuf_extract(version, \
                  ipv6_hdr, \
                  IN_IPV6_OFFSET, \
                  IPV6_VERSION_TRAFFIC_CLASS_HI4)
 
    alu_shf_right[--, 0x6, -, version,4]
    bne[version_err#]

    ;; If the packet is too small to hold IPv6 header, drop
    dl_meta_get_packet_size(packetsize)
    alu[--, packetsize, -, 40] ; packetsize - 40
    blt[packet_size_err#]

#ifdef RFC2373_CHECKS
    ;; If the source address is multicast, drop
    // The address is multicast if the first byte is 0xFF
    // Read the first 8 bits of the IPv6 Source Address
    xbuf_extract(sa_127_96, ipv6_hdr, IN_IPV6_OFFSET, IPV6_SRC_ADDR_127_96 )

    br=byte[sa_127_96, 3, 0xFF, drop_packet#]

    ;; if the source address is loopback (::1), drop
    alu[--, --, B, sa_127_96]
    bne[validate_dest#]

    xbuf_extract(sa_95_64, ipv6_hdr, IN_IPV6_OFFSET,  IPV6_SRC_ADDR_95_64)
    alu[--, --, B, sa_95_64]
    bne[validate_dest#]

    xbuf_extract(sa_63_32, ipv6_hdr, IN_IPV6_OFFSET, IPV6_SRC_ADDR_63_32)
    alu[--, --, B, sa_63_32]
    bne[validate_dest#]

    xbuf_extract(sa_31_0, ipv6_hdr, IN_IPV6_OFFSET, IPV6_SRC_ADDR_31_0)
    alu[--, sa_31_0, -, 0x01]
    beq[drop_packet#]

validate_dest#:
    ;; If the destination address is 0 or ::1, drop 
    // This code is optimized to handle the case where there is no error
    alu[--, --, B, da_127_96] 
    bne[next_validation#]     

    alu[--, --, B, da_95_64]
    bne[next_validation#]    

    alu[--, --, B, da_63_32]
    bne[next_validation#]

    // When we reach here all bits 127-32 are 0
    ;; Check if the last 32 bits are 0x0000 or 0x0001
    // i.e., Check if the MSB 31 bits are non zero
    alu[--, da_31_0, AND~, 0x01] 
    beq[dest_error#]

next_validation#:

    //
    // Exceptions
    //
    ;; If the Destination Address is multicast
    // Multicast address will have MSB 8 bits = 0xFF
    alu_shf[--, 0xFF,  AND~, da_127_96, >>24]
    beq[exception_dest_multicast#]

    ;; If the Destination Address is link local
    sub_shf_right[--, 0x3FA, da_127_96, 22]
    beq[exception_dest_linklocal#]

    ;; If nextheader is Hop-by-hop options header (nextheader == 0)
    xbuf_extract(nextheader, ipv6_hdr, IN_IPV6_OFFSET, IPV6_NEXT_HEADER)

    alu[--, --, B, nextheader]
    beq[exception_options#]
#endif //RFC2373_CHECKS

    //
    // Success
    //

    br[ipv6_validate_header_end#], defer[1]
      immed[out_result, IPV6_SUCCESS]

packet_size_err#:
#ifdef IPV6_STATS
    //IPV6_IN_TRUNCATED: Count of IP packets discarded because
    // the packet didn't carry enough data
    _ipv6_increment_counter(counter_base, IPV6_IN_TRUNCATED)
#endif //IPV6_STATS
    br[ipv6_validate_header_end#], defer[1]
      immed[out_result, IX_DROP]

version_err#:
#ifdef IPV6_STATS
    //IPV6_IN_HDR_ERRORS: Number of packets in error (version number 
    //  mismatch, other format errors, hop count exceeded, errors 
    //  discovered in processing their IP options, etc)
    _ipv6_increment_counter(counter_base, IPV6_IN_HDR_ERRORS)
#endif //IPV6_STATS
    br[ipv6_validate_header_end#], defer[1]
      immed[out_result, IX_DROP]

#ifdef RFC2373_CHECKS
dest_error#:
#ifdef IPV6_STATS
    //IPV6_IN_ADDR_ERRORS: Count of packets discarded because
    // the destination address was in error 
    _ipv6_increment_counter(counter_base, IPV6_IN_ADDR_ERRORS)
#endif //IPV6_STATS
    //fall through

drop_packet#:
    br[ipv6_validate_header_end#], defer[1]
      immed[out_result, IX_DROP]

exception_dest_multicast#:
    br[ipv6_validate_header_end#], defer[1]
      immed[out_result, IPV6_EXCP_MULTICAST]

exception_dest_linklocal#:
    br[ipv6_validate_header_end#], defer[1]
      immed[out_result, IPV6_EXCP_LINKLOCAL]

exception_options#:
    dl_set_exception_priority[1]
    immed[out_result, IPV6_EXCP_OPTIONS]
#endif //RFC2373_CHECKS

ipv6_validate_header_end#:

.end

#endm


//////////////////////////////////////////////////////////////////////////
// ipv6_hash_lookup
//
// Description:
//
//     Performs a hash lookup given index and two words to match.
//     Do chaining till it is either a hit or miss. 
//
// Inputs
//        IPV6_HASH_BASE  : base address of the hash table
//        w0, w1          : two words to match in the hash
//
// Output 
//        routing_ptr: contains next-hop id if hit, otherwise 0
//
/////////////////////////////////////////////////////////////////////////

#macro _ipv6_hash_lookup(routing_ptr, IPV6_HASH_BASE, w0, w1 )
.begin

    .reg hash_index, tmp, hash_base 
    .reg match0, match1

    immed32[hash_base, IPV6_HASH_BASE] 

    ;;
    ;; hash 64 bits of dest IP in w0 and w1
    xbuf_alloc[$hash_index, 2, read_write]
    alu[$hash_index0, --, B, w0]
    alu[$hash_index1, --, B, w1]
    hash_64[$hash_index0, 1], sig_done[sram_swap_signal] 
    ctx_arb[sram_swap_signal]
    ld_field_w_clr[hash_index, 0011, $hash_index0, <<2]  
    xbuf_free[$hash_index]

    // hash_index now has a 14-bit index value

    // Hash entry has 3 words. 
    // words0 and word1 for matching w0 and w1
    // word3 has: 
    //    0:15  - next index in case of chaining
    //    16:31 - routing_ptr in case of hit
    
    xbuf_alloc[$hash_lookup, 3, read]

    // First, check the main table
    sram[read, $hash_lookup0, hash_base, hash_index, 3], 
            ctx_swap[sram_swap_signal]
    alu[match0, w0, XOR,$hash_lookup0]
    alu[match1, w1, XOR,$hash_lookup1]  
    alu[--, match1, OR, match0]
    beq[matched_entry#]

    // Point to collision table at the end of main table
    alu[tmp, --, B, 1, <<IPV6_NUM_HASH_BITS]
    alu[hash_base, tmp, +, hash_base]

    ld_field_w_clr[hash_index, 0011, $hash_lookup2, <<2]

    // Currently, the collision list is traversed until a
    // match is found or a null next index is encountered.
    // This can be limited to a predefined number of tries,
    // after which the search can be performed in the exception
    // path.

start_collision_loop#:

    sram[read, $hash_lookup0, hash_base, hash_index, 3], 
            ctx_swap[sram_swap_signal]
    alu[match0, w0, XOR,$hash_lookup0]
    alu[match1, w1, XOR,$hash_lookup1]  
    alu[--, match1, OR, match0]
    beq[matched_entry#]

    no_match#:
    // Miss
    ld_field_w_clr[hash_index, 0011, $hash_lookup2, <<2]   
    bne[start_collision_loop#]
 
    // End of collision list
    // Leave the routing pointer alone and exit 
    br[end_hash_lookup#]

matched_entry#:

    ld_field_w_clr[routing_ptr, 1100, $hash_lookup2, >>0]

end_hash_lookup#:

    xbuf_free[$hash_lookup]

.end
#endm

///////////////////////////////////////////////////////////////////////////////
//
// _ipv6_trie_8[]
//
// Description: 
//
// - Extracts 8 bits from the ipaddr into an index (ending SHIFT_AMT location)
// - Reads the trie entry at Current_Trie_Block[index]
// - If Next_hop field of this trie-entry!= 0, update prev_valid_next_hop 
// - Place the value of Next_trie field into next_trie_block
//
//
// Outputs:
//
// prev_valid_next_hop - Value of the trie entry of the previous trie walk step.
//                       Will be updated with the value of the trie entry
//                       corresponding to the current trie walk step
//
// next_trie_block     - Index of the next trie exracted from the trie entry
//                       corresponding to the current 8 bits of IP address.
//
// Inputs
//
// ipaddr       -  Current 32 bits of IP address that are being looked up
//
// SHIFT_AMT    -  Pointer to the LSBit of the which needs to be extracted
//                 from ipaddr for this step of the trie walk 
//                 e.g., 0 -> We extract bits 7-0 for the walk 
//                 8 -> We extract bits 15-7 for the walk
// next_trie_block   - Index of the current trie block. This entry is read to 
//                     proceed with the lookup.
//
// trie_base    -  Base address of the 0th 8 bit trie block.
//
////////////////////////////////////////////////////////////////////////////////

#macro _ipv6_trie_8(ipaddr,                \
                    SHIFT_AMT,             \
                    prev_valid_next_hop,   \
                    next_trie_block,       \
                    trie_base)
.begin
    .reg current_next_hop
    .sig sram_trie_signal
    .reg $xfer_reg
    

    // Convert trie block index to word offset into SRAM
    // next_trie_block = next_trie_block * 256
    shf_left(next_trie_block, next_trie_block, 8)                        

    // Extract the specified 8 bits from the IP address
    // Use these 8 bits to index into into the trie block
    // Note: The index is in terms of 32 bit long words
    alu_shf_right(next_trie_block, next_trie_block, +8, ipaddr, SHIFT_AMT)        

    // Convert to get a byte offset
    // i.e., Multiply by 4
    shf_left(next_trie_block, next_trie_block, 2)

    // by now, we have the trie entry to read
    sram[read, $xfer_reg, trie_base, next_trie_block, 1], \
                  ctx_swap[sram_trie_signal]

    // get the next hop pointer
    bytefield_select(current_next_hop, 1100, $xfer_reg, DO_LOAD_CC)        

    beq[read_next_trie#]

    // If the next hop is valid at this trie level, we need to update the 
    // prev_valid_next_hop with the new value so that it gets carried forward
    move[prev_valid_next_hop,current_next_hop]

read_next_trie#:
    // get the next trie pointer, set cond code
    bytefield_select(next_trie_block, 0011, $xfer_reg, DO_LOAD_CC) 

.end
#endm


////////////////////////////////////////////////////////////////////////////////
//
// ipv6_route_lookup
//
// Description: Perform ip longest prefix match lookup. 
//              Uses tables written by core Route Table Manager.
// Outputs:
//    Next_hop_ptr: GPR which stores the result of the lookup.
//                  Bits 15-0 contain the index into the next hop table 
//
// Inputs:
//    da_127_96            : GPR with IP destination address bits [127..96]
//    da_95_64             : GPR with IP destination address bits [95..64]
//    da_63_32             : GPR with IP destination address bits [63..32]
//    da_31_0              : GPR with IP destination address bits [31..0]
//    IPV6_TRIE_TABLE_BASE : Constant base address of IPv6 Trie Table
//                           The trie table consists of an initial 16 bit trie 
//                           block (64k) followed by 8 bit (1k) trie blocks.
// Side Effect:
//                         : Future Count may be reset
// Example Usage:
//
// ipv6_route_lookup(next_hop_ptr, ip_da0, ip_da1, ip_da2, ip_da3, 0x20000)
// 
// Comments                : Worst case 7 SRAM accesses to matach a 64 bit  
//                           route-entry
//
///////////////////////////////////////////////////////////////////////////////
#ifdef SPLIT_IPV6_TRIE_TABLE
#macro _ipv6_route_lookup(next_hop_ptr,         \
                          da_127_96,            \
                          da_95_64,             \
                          da_63_32,             \
                          da_31_0,              \
                          IPV6_TRIE_TABLE_BASE_16, \
						  IPV6_TRIE_TABLE_BASE_8) 
.begin
    .reg base_16bit_trie, base_8bit_tries
    .reg next_trie_ptr
    .reg cache_update_state 
    .reg result, offset, tmp, tmp1
    .sig volatile sram_cache_update
    .sig sram_swap_signal
    .reg $curr_trie_entry 
    .reg $t0 $t1 $t2 $t3 $t4 $t5 $t6 $t7 $t8 $t9
    .xfer_order $t0 $t1 $t2 $t3 $t4 $t5 $t6 $t7 $t8 $t9

    immed32[next_hop_ptr, 0];

    // base_16bit_trie can be either a static constant or from 
    // a .import_var. It points to the base address of the 16 bit 
    // 64k entry trie table
    immed32(base_16bit_trie, IPV6_TRIE_TABLE_BASE_16)

    // base_8bit_tries points to the base of the 8 bit trie block memory
	immed32(base_8bit_tries, IPV6_TRIE_TABLE_BASE_8)

    // We store 64-bit router prefix and 8 trie-entries corresponding to
    // that route in local memory trie-cache
     
    // The Trie cache in Local Memory contains 10 long-words
    // +-----+-----+-----+-----+-----+-----+-----+
    // | lw0 | lw1 | lw2 | lw3 | lw4 |.....| lw9 |
    // +-----+-----+-----+-----+-----+-----+-----+
    // |<- Router->|<----- Trie-Entries -------->|
    //     Prefix                                 

    #define_eval LM_TRIE_ENTRY_BASE (IPV6_TRIE_CACHE_LM_BASE +8)
    immed[tmp, IPV6_TRIE_CACHE_LM_BASE ]
    immed[tmp1, LM_TRIE_ENTRY_BASE]
    local_csr_wr[ACTIVE_LM_ADDR_0, tmp ]
    local_csr_wr[ACTIVE_LM_ADDR_1, tmp1 ]
    nop
    nop
    nop 

start_LPM#:    

    //
    // Local Memory Router's Prefix Matching 
    //

    // Match dest IP with stored Router Prefix 16,8,8,8,8,8,8 bits at a time
    // As long as the bits match, the strides take place in Local Memory
    // At the first mismatch, the cached trie entry is used to start the
    // SRAM trie-walk from that point onwards.

    // Unroll the loop completely

    ;; check to see if the first long word of the router prefix is zero.
    ;; if it is, it was never set - simply perform a regular trie walk on 
    ;; the IPv6 address.

    alu[--, --, B, *l$index0]
    beq[SRAM_trie_walk_high_64#]

    ;;
    ;; compare first lw of destIP with first lw of prefix
    alu[result, *l$index0, XOR, da_127_96]

    ;; did first 16 bit match
    ld_field_w_clr[tmp, 1100 , result], load_cc 
    bne[lookup_127_112#]

    // Yes they matched, countinue matching
    // Store next-hop and next-trie portions
    ld_field_w_clr[next_hop_ptr,  1100, *l$index1[0]] 
    alu[next_trie_ptr, 0,  +16, *l$index1[0]]
    beq[set_route_ptr#]

    ;; do the next 8 bits match
    ld_field_w_clr[tmp, 0010 , result], load_cc 
    bne[lookup_111_104#] ;; no
    ;; yes
    ld_field_w_clr[tmp,  1100, *l$index1[1]] , load_cc
    beq[continue1#]
    alu[next_hop_ptr, --, B, tmp]
    continue1#:
    alu[next_trie_ptr, 0,  +16, *l$index1[1]]         
    beq[set_route_ptr#]

    ;; next 8
    ld_field_w_clr[tmp, 0001 , result], load_cc
    bne[lookup_103_96#]
    ld_field_w_clr[tmp,  1100, *l$index1[2]] , load_cc
    beq[continue2#]
    alu[next_hop_ptr, --, B, tmp]
    continue2#:
    alu[next_trie_ptr, 0,  +16, *l$index1[2]]         
    beq[set_route_ptr#]

    ;;
    ;; Compare second lw of destIP with second lw of prefix
    alu[result, *l$index0[1], XOR, da_95_64]

    ;; next 8
    ld_field_w_clr[tmp, 1000 , result], load_cc
    bne[lookup_95_88#] 
    ld_field_w_clr[tmp,  1100, *l$index1[3]] , load_cc
    beq[continue3#]
    alu[next_hop_ptr, --, B, tmp]
    continue3#:
    alu[next_trie_ptr, 0,  +16, *l$index1[3]]         
    beq[set_route_ptr#]

    ;; next 8
    ld_field_w_clr[tmp, 0100 , result], load_cc
    bne[lookup_87_80#]
    ld_field_w_clr[tmp,  1100, *l$index1[4]] , load_cc
    beq[continue4#]
    alu[next_hop_ptr, --, B, tmp]
    continue4#:
    alu[next_trie_ptr, 0,  +16, *l$index1[4]]         
    beq[set_route_ptr#]
 
    ;; next 8
    ld_field_w_clr[tmp, 0010 , result], load_cc
    bne[lookup_79_72#]
    ld_field_w_clr[tmp,  1100, *l$index1[5]] , load_cc
    beq[continue5#]
    alu[next_hop_ptr, --, B, tmp]
    continue5#:
    alu[next_trie_ptr, 0,  +16, *l$index1[5]]         
    beq[set_route_ptr#]

    ;; next 8
    ld_field_w_clr[tmp, 0001 , result], load_cc
    bne[lookup_71_64#]
    ld_field_w_clr[tmp,  1100, *l$index1[6]] , load_cc
    beq[continue6#]
    alu[next_hop_ptr, --, B, tmp]
    continue6#:
    alu[next_trie_ptr, 0,  +16, *l$index1[6]]         
    beq[set_route_ptr#]

    // Matched all 64 bits of route's prefix.
    // This must be an edge router at SLA boundary
    // Continue LPM on remaining 64 bits as they may have
    // sub divided their own subnet using lower 64 bits.
    br[SRAM_trie_walk_lower_64#]


    ;;
    ;; LPM on first 64 bits of dest IP (SRAM Trie walk)
    ;;

SRAM_trie_walk_high_64#:

    lookup_127_112#:

    // Extract first 16 bits from IP addr and use them to index into 
    // the 16 bit table. Each trie entry is 4 bytes long. So the index
    // needs to be multiplied by 4 to get to correct SRAM location

    ;; 16-bit trie-block offset
    alu_shf_right(offset, 0x3, ~AND, da_127_96, 14)   
    // Read the trie entry at that 16 bit offset
    sram[read, $curr_trie_entry, base_16bit_trie, offset, 1], 
             ctx_swap[sram_swap_signal] 

    // Get next hop pointer
    bytefield_select(next_hop_ptr,1100,$curr_trie_entry, NO_LOAD_CC)   
    // get next trie pointer 
    bytefield_select(next_trie_ptr, 0011, $curr_trie_entry, DO_LOAD_CC) 

    // next_trie_ptr = 0, no route found
    beq[set_route_ptr#]                                                                                        
    // Start 8 bit trie walk
    // Completely unroll the lookup loop for efficiency

    lookup_111_104#:
    _ipv6_trie_8(da_127_96, 8, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_103_96#:
    _ipv6_trie_8(da_127_96, 0, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_95_88#:
    _ipv6_trie_8(da_95_64, 24, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_87_80#:
    _ipv6_trie_8(da_95_64, 16, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_79_72#:
    _ipv6_trie_8(da_95_64, 8, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_71_64#:
    _ipv6_trie_8(da_95_64, 0, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

#ifndef DONT_HASH_LOWER_64_BITS_OF_DIP

HASH_lower_64#:

    ;;
    ;; Hash Lookup
    ;;

    // Matched 64 bits, do hash on remaining 64
    // next-trie pointer is pointing to hash table

    _ipv6_hash_lookup(next_hop_ptr,IPV6_HASH_TABLE_SRAM_BASE,da_63_32,da_31_0)

    br[set_route_ptr#]

#endif

    ;;
    ;; LPM on lower 64 bits of dest IP (SRAM Trie walk)
    ;;

SRAM_trie_walk_lower_64#:

    lookup_63_56#:
    _ipv6_trie_8(da_63_32, 24, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_55_48#:
    _ipv6_trie_8(da_63_32, 16, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_47_40#:
    _ipv6_trie_8(da_63_32, 8, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_39_32#:
    _ipv6_trie_8(da_63_32, 0, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_31_24#:
    _ipv6_trie_8(da_31_0, 24, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_23_16#:
    _ipv6_trie_8(da_31_0, 16, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_15_8#:
    _ipv6_trie_8(da_31_0, 8, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_7_0#:
    _ipv6_trie_8(da_31_0, 0, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

set_route_ptr#:



    // extract the last valid next hop
    shf_right(next_hop_ptr, next_hop_ptr, 16)  
    //
    // Following code periodically refreshes the 
    // local-memory trie cache. Currently, we do not
    // expect the routing information to change dynamically

    // If the future_count has expired, start read to 
    // refresh the trie-cache   
    //

    .set_sig ipv6_cache_refresh_sig
    br_!signal[ipv6_cache_refresh_sig, lookup_end#]

    // Signal is set, we need to copy from SRAM
    ;;
    ;; Set LM base pointer 
    immed[tmp, IPV6_TRIE_CACHE_LM_BASE]
    local_csr_wr[active_lm_addr_1, tmp]

    immed32[offset, IPV6_ROUTER_PREFIX_SRAM_BASE]
    ;; Start the read to get router's prefix
    alu[--, --, B, 0x1A, <<21]
    sram[read, $t0,  0, offset, max_10],  
             indirect_ref, sig_done[sram_cache_update]

    // Set up signal for the next update
    .begin
        local_csr_rd[timestamp_low]
        immed[tmp, 0]
        immed32[tmp1, FUTURE_COUNT] 
        alu[tmp, tmp, +, tmp1]
        local_csr_wr[INDIRECT_CTX_FUTURE_COUNT, tmp]
    .end

    ctx_arb[sram_cache_update]

    alu[*l$index1++, --, B, $t0]
    alu[*l$index1++, --, B, $t1]
    alu[*l$index1++, --, B, $t2]
    alu[*l$index1++, --, B, $t3]
    alu[*l$index1++, --, B, $t4]
    alu[*l$index1++, --, B, $t5]
    alu[*l$index1++, --, B, $t6]
    alu[*l$index1++, --, B, $t7]
    alu[*l$index1++, --, B, $t8]
    alu[*l$index1++, --, B, $t9]


lookup_end#:

.end
#endm

#else
#macro _ipv6_route_lookup(next_hop_ptr,         \
                          da_127_96,            \
                          da_95_64,             \
                          da_63_32,             \
                          da_31_0,              \
                          IPV6_TRIE_TABLE_BASE) 
.begin
    .reg base_16bit_trie, base_8bit_tries
    .reg next_trie_ptr
    .reg cache_update_state 
    .reg result, offset, tmp, tmp1
    .sig volatile sram_cache_update
    .sig sram_swap_signal
    .reg $curr_trie_entry 
    .reg $t0 $t1 $t2 $t3 $t4 $t5 $t6 $t7 $t8 $t9
    .xfer_order $t0 $t1 $t2 $t3 $t4 $t5 $t6 $t7 $t8 $t9

    immed32[next_hop_ptr, 0];

    // base_16bit_trie can be either a static constant or from 
    // a .import_var. It points to the base address of the 16 bit 
    // 64k entry trie table
    immed32(base_16bit_trie, IPV6_TRIE_TABLE_BASE)

    // The 16 bit 64k entry trie block is
    // followed by 8 bit (256 entry) trie blocks
    // base_8bit_tries points to the base of the 8 bit trie block memory
    // Add 0x40000 (64k * 4) to the table base to get location of 
    // 8 bit trie blocks
	immed32(tmp, 0x40000)
	alu[base_8bit_tries, base_16bit_trie, +, tmp]

    // We store 64-bit router prefix and 8 trie-entries corresponding to
    // that route in local memory trie-cache
     
    // The Trie cache in Local Memory contains 10 long-words
    // +-----+-----+-----+-----+-----+-----+-----+
    // | lw0 | lw1 | lw2 | lw3 | lw4 |.....| lw9 |
    // +-----+-----+-----+-----+-----+-----+-----+
    // |<- Router->|<----- Trie-Entries -------->|
    //     Prefix                                 

    #define_eval LM_TRIE_ENTRY_BASE (IPV6_TRIE_CACHE_LM_BASE +8)
    immed[tmp, IPV6_TRIE_CACHE_LM_BASE ]
    immed[tmp1, LM_TRIE_ENTRY_BASE]
    local_csr_wr[ACTIVE_LM_ADDR_0, tmp ]
    local_csr_wr[ACTIVE_LM_ADDR_1, tmp1 ]
    nop
    nop
    nop 

start_LPM#:    

    //
    // Local Memory Router's Prefix Matching 
    //

    // Match dest IP with stored Router Prefix 16,8,8,8,8,8,8 bits at a time
    // As long as the bits match, the strides take place in Local Memory
    // At the first mismatch, the cached trie entry is used to start the
    // SRAM trie-walk from that point onwards.

    // Unroll the loop completely

    ;; check to see if the first long word of the router prefix is zero.
    ;; if it is, it was never set - simply perform a regular trie walk on 
    ;; the IPv6 address.

    alu[--, --, B, *l$index0]
    beq[SRAM_trie_walk_high_64#]

    ;;
    ;; compare first lw of destIP with first lw of prefix
    alu[result, *l$index0, XOR, da_127_96]

    ;; did first 16 bit match
    ld_field_w_clr[tmp, 1100 , result], load_cc 
    bne[lookup_127_112#]

    // Yes they matched, countinue matching
    // Store next-hop and next-trie portions
    ld_field_w_clr[next_hop_ptr,  1100, *l$index1[0]] 
    alu[next_trie_ptr, 0,  +16, *l$index1[0]]
    beq[set_route_ptr#]

    ;; do the next 8 bits match
    ld_field_w_clr[tmp, 0010 , result], load_cc 
    bne[lookup_111_104#] ;; no
    ;; yes
    ld_field_w_clr[tmp,  1100, *l$index1[1]] , load_cc
    beq[continue1#]
    alu[next_hop_ptr, --, B, tmp]
    continue1#:
    alu[next_trie_ptr, 0,  +16, *l$index1[1]]         
    beq[set_route_ptr#]

    ;; next 8
    ld_field_w_clr[tmp, 0001 , result], load_cc
    bne[lookup_103_96#]
    ld_field_w_clr[tmp,  1100, *l$index1[2]] , load_cc
    beq[continue2#]
    alu[next_hop_ptr, --, B, tmp]
    continue2#:
    alu[next_trie_ptr, 0,  +16, *l$index1[2]]         
    beq[set_route_ptr#]

    ;;
    ;; Compare second lw of destIP with second lw of prefix
    alu[result, *l$index0[1], XOR, da_95_64]

    ;; next 8
    ld_field_w_clr[tmp, 1000 , result], load_cc
    bne[lookup_95_88#] 
    ld_field_w_clr[tmp,  1100, *l$index1[3]] , load_cc
    beq[continue3#]
    alu[next_hop_ptr, --, B, tmp]
    continue3#:
    alu[next_trie_ptr, 0,  +16, *l$index1[3]]         
    beq[set_route_ptr#]

    ;; next 8
    ld_field_w_clr[tmp, 0100 , result], load_cc
    bne[lookup_87_80#]
    ld_field_w_clr[tmp,  1100, *l$index1[4]] , load_cc
    beq[continue4#]
    alu[next_hop_ptr, --, B, tmp]
    continue4#:
    alu[next_trie_ptr, 0,  +16, *l$index1[4]]         
    beq[set_route_ptr#]
 
    ;; next 8
    ld_field_w_clr[tmp, 0010 , result], load_cc
    bne[lookup_79_72#]
    ld_field_w_clr[tmp,  1100, *l$index1[5]] , load_cc
    beq[continue5#]
    alu[next_hop_ptr, --, B, tmp]
    continue5#:
    alu[next_trie_ptr, 0,  +16, *l$index1[5]]         
    beq[set_route_ptr#]

    ;; next 8
    ld_field_w_clr[tmp, 0001 , result], load_cc
    bne[lookup_71_64#]
    ld_field_w_clr[tmp,  1100, *l$index1[6]] , load_cc
    beq[continue6#]
    alu[next_hop_ptr, --, B, tmp]
    continue6#:
    alu[next_trie_ptr, 0,  +16, *l$index1[6]]         
    beq[set_route_ptr#]

    // Matched all 64 bits of route's prefix.
    // This must be an edge router at SLA boundary
    // Continue LPM on remaining 64 bits as they may have
    // sub divided their own subnet using lower 64 bits.
    br[SRAM_trie_walk_lower_64#]


    ;;
    ;; LPM on first 64 bits of dest IP (SRAM Trie walk)
    ;;

SRAM_trie_walk_high_64#:

    lookup_127_112#:

    // Extract first 16 bits from IP addr and use them to index into 
    // the 16 bit table. Each trie entry is 4 bytes long. So the index
    // needs to be multiplied by 4 to get to correct SRAM location

    ;; 16-bit trie-block offset
    alu_shf_right(offset, 0x3, ~AND, da_127_96, 14)   
    // Read the trie entry at that 16 bit offset
    sram[read, $curr_trie_entry, base_16bit_trie, offset, 1], 
             ctx_swap[sram_swap_signal] 

    // Get next hop pointer
    bytefield_select(next_hop_ptr,1100,$curr_trie_entry, NO_LOAD_CC)   
    // get next trie pointer 
    bytefield_select(next_trie_ptr, 0011, $curr_trie_entry, DO_LOAD_CC) 

    // next_trie_ptr = 0, no route found
    beq[set_route_ptr#]                                                                                        
    // Start 8 bit trie walk
    // Completely unroll the lookup loop for efficiency

    lookup_111_104#:
    _ipv6_trie_8(da_127_96, 8, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_103_96#:
    _ipv6_trie_8(da_127_96, 0, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_95_88#:
    _ipv6_trie_8(da_95_64, 24, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_87_80#:
    _ipv6_trie_8(da_95_64, 16, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_79_72#:
    _ipv6_trie_8(da_95_64, 8, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_71_64#:
    _ipv6_trie_8(da_95_64, 0, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

#ifndef DONT_HASH_LOWER_64_BITS_OF_DIP

HASH_lower_64#:

    ;;
    ;; Hash Lookup
    ;;

    // Matched 64 bits, do hash on remaining 64
    // next-trie pointer is pointing to hash table

    _ipv6_hash_lookup(next_hop_ptr,IPV6_HASH_TABLE_SRAM_BASE,da_63_32,da_31_0)

    br[set_route_ptr#]

#endif

    ;;
    ;; LPM on lower 64 bits of dest IP (SRAM Trie walk)
    ;;

SRAM_trie_walk_lower_64#:

    lookup_63_56#:
    _ipv6_trie_8(da_63_32, 24, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_55_48#:
    _ipv6_trie_8(da_63_32, 16, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_47_40#:
    _ipv6_trie_8(da_63_32, 8, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_39_32#:
    _ipv6_trie_8(da_63_32, 0, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_31_24#:
    _ipv6_trie_8(da_31_0, 24, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_23_16#:
    _ipv6_trie_8(da_31_0, 16, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_15_8#:
    _ipv6_trie_8(da_31_0, 8, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

    lookup_7_0#:
    _ipv6_trie_8(da_31_0, 0, next_hop_ptr, next_trie_ptr, base_8bit_tries)
    beq[set_route_ptr#]

set_route_ptr#:



    // extract the last valid next hop
    shf_right(next_hop_ptr, next_hop_ptr, 16)  
    //
    // Following code periodically refreshes the 
    // local-memory trie cache. Currently, we do not
    // expect the routing information to change dynamically

    // If the future_count has expired, start read to 
    // refresh the trie-cache   
    //

    .set_sig ipv6_cache_refresh_sig
    br_!signal[ipv6_cache_refresh_sig, lookup_end#]

    // Signal is set, we need to copy from SRAM
    ;;
    ;; Set LM base pointer 
    immed[tmp, IPV6_TRIE_CACHE_LM_BASE]
    local_csr_wr[active_lm_addr_1, tmp]

    immed32[offset, IPV6_ROUTER_PREFIX_SRAM_BASE]
    ;; Start the read to get router's prefix
    alu[--, --, B, 0x1A, <<21]
    sram[read, $t0,  0, offset, max_10],  
             indirect_ref, sig_done[sram_cache_update]

    // Set up signal for the next update
    .begin
        local_csr_rd[timestamp_low]
        immed[tmp, 0]
        immed32[tmp1, FUTURE_COUNT] 
        alu[tmp, tmp, +, tmp1]
        local_csr_wr[INDIRECT_CTX_FUTURE_COUNT, tmp]
    .end

    ctx_arb[sram_cache_update]

    alu[*l$index1++, --, B, $t0]
    alu[*l$index1++, --, B, $t1]
    alu[*l$index1++, --, B, $t2]
    alu[*l$index1++, --, B, $t3]
    alu[*l$index1++, --, B, $t4]
    alu[*l$index1++, --, B, $t5]
    alu[*l$index1++, --, B, $t6]
    alu[*l$index1++, --, B, $t7]
    alu[*l$index1++, --, B, $t8]
    alu[*l$index1++, --, B, $t9]


lookup_end#:

.end
#endm

#endif

////////////////////////////////////////////////////////////////////////////////
//
// _ipv6_process_intermediate_nexthop_info
//
// Description: 
//
//   Read L3 next-hop structure from SRAM
//      Check to see if it is an intermediate entry, if it is, extract a 
//   "real" nexthop entry (L3 table index). 
//
//
//  Support for upto 4-equal cost next-hops
//     This nexthop_index could point to an "intermediate entry", check to 
//  see if it does. If it does not (i.e it points to a real entry) there 
//  is not a whole lot to do, use this nexthop_index. If it does, read 
//  the intermediate entry (16 bytes) and use a generated index (last two 
//  bits of source IP address XOR'd with last two bits of destination IP 
//  address) to pick on of the four equal cost hops and save that in the 
//  nexthop_index. 
//
// Outputs:
//     out_result - result of this macro.  Can be:
//     IPV6_SUCCESS : if the macro succeeded.
//     IPV6_FAILURE : if the macro failed.
//     Exception code - depending on the exception being raised.
//     out_nhid_type    : type of next-hop-id
//
// Inputs:
//     ipv6_hdr      : Memory location where the IPv6 header lives.
//     nexthop_id    : L3 next-hop table index ( found by the lookup )
//     IN_IPV6_OFFSET: Offset of the IPv6 header in the packet (memory).
//     da_31_0       : The least significant 32 bits of the destination 
//                     address.
//
////////////////////////////////////////////////////////////////////////////////
#ifdef IPV6_LOADBALANCE

#macro  _ipv6_process_intermediate_nexthop_info(out_result,     \
                                                ipv6_hdr,       \
                                                IN_IPV6_OFFSET, \
                                                nexthop_index,  \
                                                da_31_0)
.begin
    .reg sa_31_0, status
    .reg count, real_index, tmp
    .reg retval

    nh_read_entry($nexthop_entry, nexthop_index)
    
    // Is the entry an "intermediate" entry?
    nh_is_intermediate_entry($nexthop_entry, retval)

    .if(retval == IPV6_SUCCESS)
        
        // yes it is an intermediate entry.
        // pick a "real" next hop entry. See comments above for 
        // how this is done. Read the source address out of the 
        // IPv6 header and continue

        xbuf_extract(sa_31_0, ipv6_hdr, IN_IPV6_OFFSET, IPV6_SRC_ADDR_31_0)        

        // perform an XOR operation and since we are interested only 
        // in the last two bits (there are only 4 possible next hops in 
        // an intermediate entry).

        alu[count, sa_31_0, XOR, da_31_0]
        alu[count, 0x03, AND, count]

        // Check the "status (UP/DOWN of this entry) if it is DOWN, 
        // try to find the next best entry by walking down the list. 
        // Remember all the entries are equal cost here, so it doesn't 
        // matter which one is picked.
        
        nh_extract_status($nexthop_entry, status)

        alu[--, count, AND, 0x0000000F]
        
        alu_shf[real_index, status, AND, 0x01, <<indirect]

        // the link is UP, good, read the "real" next hop and move to 
        // the encap block
        bne[set_meta_data#]

link_down#:
        
        alu[tmp, --, B, count]

loop#:
        alu[count, count, +, 1]

        // have we wrapped around?
        .if (count == IPV6_NH_NHID_TOTAL)
            immed[count, 0x0]
        .endif

        alu[--, count, AND, 0x0000000F]

        alu_shf[real_index, status, AND, 0x01, <<indirect]

        // is the bit set, do we have a link that is UP?
        bne[set_meta_data#]

        // have we tried all the bits
        alu[--, tmp, - , count]
        
        // All the interfaces are DOWN, drop the packet. 
        // Raise an exception - No Route.
        beq[done#], defer[1]
            immed[out_result, IPV6_EXCP_LKUPFAILED]
        
        // try again..
        br[loop#]
        
set_meta_data#:

        alu[real_index, --, B, count]

        nh_extract_nhid_at_index($nexthop_entry, real_index, nexthop_index)

        immed[out_result, IPV6_SUCCESS]
        nh_free_entry($nexthop_entry)

    .else

        //This is a real entry
        immed[out_result, IPV6_SUCCESS]

    .endif

done#:
    // return

.end
#endm

#endif

//////////////////////////////////////////////////////////////////////////////
//
// ipv6_forwarder[]
//
// Description: 
//
//     Verify IP header and perform LPM on the packet
//
// Outputs:
//            
//     out_ipv6_hdr    : Buffer for the modified IPv6 header
//                           (Hop-limit field gets modified) 
//
//     OUT_IPV6_OFFSET : Byte offset of the 1st byte of IPv6 header in buffer
//
//
// Side Effects:
//
//     dl_next_block is set to IX_DROP, IX_EXCEPTION, or IPV6_NEXT_BLOCK
//     In case if exception, the exception code in copied in metadata
//
//     nexthop_id, and nexthop-id-type found in the L3 table are copied 
//     to the metadata
//
// Inputs:
//
//     ipv6_hdr     : Buffer containing the cached IPv6 header
//
//     IN_IPV6_OFFSET  : Byte offset of the 1st byte of IPv6 header in buffer
//
///////////////////////////////////////////////////////////////////////////////

#macro ipv6_forwarder(out_ipv6_hdr, OUT_IPV6_OFFSET, ipv6_hdr, IN_IPV6_OFFSET)
.begin
    .reg  in_port         // port from which this packet is rx'ed
    .reg  out_port        // port to which this packet is tx'ed
    .reg  out_blade_id    // blade to which this packet is tx'ed
    .reg  nexthop_index   // index of next hop information
    .reg  nhid_type       // type of the nexthop information
    .reg  context_id
    .reg  result          
    .reg  class_id
    .reg  hop_count
    .reg  packet_size
    .reg  da_127_96, da_95_64, da_63_32, da_31_0
    .reg  offset

    ;; Check if the next block is set to IPv6
    br!=byte[dl_next_block, 0, BID_IPV6, ipv6_forwarder_finish#]


    dl_get_thread_id(context_id)
    // First, set the local memory index pointer to point to IPv6 header
    xbuf_activate(ipv6_hdr, 1, context_id, 0)

    // Stats are maintained per port
    // Build Stat Counter base for this input port.

    dl_meta_get_input_port[in_port]    

    //Get the packet size
    dl_meta_get_packet_size(packet_size)

#ifdef IPV6_STATS
    .reg    counter_base    // base where the stats are maintained

    // Build Stat Counter base for this port.
    alu[counter_base,  --, B, @ipv6_sram_counter_base]
    alu_shf[offset, --, B, in_port, <<IPV6_IN_STATS_PER_PORT_SHIFT]
    alu[counter_base, counter_base, +, offset]

    // Update the number of packets received.
    _ipv6_increment_counter(counter_base, IPV6_IN_RECV)

    //Increment the count, statsInOctets by the packet size.
    _ipv6_add_to_counter(counter_base, IPV6_IN_OCTETS, packet_size)
#endif //IPV6_STATS

    // combine the inport with blade Id by loading bladeId
    // in most significant byte.
    alu_shf[in_port,in_port,OR,THIS_BLADE_ID,<<24]

    // Extract the 128 bit destination address from the header cache
    // These will be used by header validation as well as lookup macro
    xbuf_extract(da_127_96, ipv6_hdr, IN_IPV6_OFFSET, IPV6_DST_ADDR_127_96)
    xbuf_extract(da_95_64, ipv6_hdr, IN_IPV6_OFFSET,  IPV6_DST_ADDR_95_64)
    xbuf_extract(da_63_32, ipv6_hdr, IN_IPV6_OFFSET, IPV6_DST_ADDR_63_32)
    xbuf_extract(da_31_0, ipv6_hdr, IN_IPV6_OFFSET, IPV6_DST_ADDR_31_0)

    ;;
    ;; IPv6 header Validations
    ;;
    _ipv6_header_validations(result,         \
                             ipv6_hdr,       \
                             IN_IPV6_OFFSET, \
                             da_127_96,      \
                             da_95_64,       \
                             da_63_32,       \
                             da_31_0)

    br!=byte[result, 0, IPV6_SUCCESS, handle_error_exception#]

    ;;
    ;; IPv6 LPM
    ;;

	#ifdef SPLIT_IPV6_TRIE_TABLE
    _ipv6_route_lookup(nexthop_index,             \
                       da_127_96,                 \
                       da_95_64,                  \
                       da_63_32,                  \
                       da_31_0,                   \
                       IPV6_TRIE_TABLE_SRAM_BASE_16, \
                       IPV6_TRIE_TABLE_SRAM_BASE_8)
	#else
    _ipv6_route_lookup(nexthop_index,             \
                       da_127_96,                 \
                       da_95_64,                  \
                       da_63_32,                  \
                       da_31_0,                   \
                       IPV6_TRIE_TABLE_SRAM_BASE)
	#endif

    // Lookup might have moved the local memory pointer. Reset it.
    xbuf_activate(ipv6_hdr, 1, context_id, 0)

#ifdef IPV6_LOADBALANCE

    _ipv6_process_intermediate_nexthop_info(result,         \
                                            ipv6_hdr,       \
                                            IN_IPV6_OFFSET, \
                                            nexthop_index,  \
                                            da_31_0)

    .if( result != IPV6_SUCCESS)
#ifdef IPV6_STATS
         // IPV6_IN_FORW: count of non-local packets for which a lookup is done
         // Increment counter if it is not locally destined
         // Ideally we would have done this *before* the lookup, but we don't
         // know then if the packet is locally destined
         _ipv6_increment_counter(counter_base, IPV6_IN_FORW)
#endif //IPV6_STATS
        br[handle_error_exception#]
    .endif

#endif  // IPV6_LOADBALANCE

    ;;
    ;; Post lookup Validations
    ;;
    _ipv6_process_nexthop_info(result,        \
                               out_port,      \
                               out_blade_id,  \
                               nhid_type,     \
                               nexthop_index, \
                               packet_size,   \
                               in_port,       \
                               counter_base,  \
                               ipv6_hdr,      \
                               IN_IPV6_OFFSET)

    br!=byte[result, 0, IPV6_SUCCESS, handle_error_exception#]

#ifdef IPV6_STATS
    //IPV6_OUT_FORW: Count of IP packets for which a path to the destination
    //               was found
    //IMP: Increment this counter for outgoing interface!
    .reg outgoing_intf_counter_base
    .set outgoing_intf_counter_base
    .begin
        .reg blade_offset, port_offset

        alu[outgoing_intf_counter_base, --, B, @ipv6_out_sram_counter_base]
        alu_shf[blade_offset, \
                --,           \
                B,            \
                out_blade_id, \
                <<IPV6_OUT_STATS_PER_BLADE_SHIFT]
        alu_shf[port_offset, --, B, out_port, <<IPV6_OUT_STATS_PER_PORT_SHIFT]
        alu[outgoing_intf_counter_base, \
            outgoing_intf_counter_base, \
            +,                          \
            blade_offset]
        alu[outgoing_intf_counter_base, \
            outgoing_intf_counter_base, \
            +,                          \
            port_offset]
    .end

     _ipv6_increment_counter(outgoing_intf_counter_base, IPV6_OUT_FORW)
#endif //IPV6_STATS



    ;; Check Hop-Count
    // If hop count == 1 send the packet to the XScale
    // Else decrement the hop count and place it back in the packet haeder
    xbuf_extract(hop_count, ipv6_hdr, IN_IPV6_OFFSET, IPV6_HOP_LIMIT)

    // Decrement hop count, and check if it has expired
    alu[hop_count, hop_count, -, 1]
    ble[hop_count_expired#]

    ;;
    ;; Success
    ;;

    ;; Copy In header to Out header
    xbuf_copy(out_ipv6_hdr,0,OUT_IPV6_OFFSET,ipv6_hdr,IN_IPV6_OFFSET,0,40,0)

    ;; Updated hop count back into the header
    xbuf_insert(out_ipv6_hdr, hop_count, IN_IPV6_OFFSET, IPV6_HOP_LIMIT)
    
    //
    // Update Metadata
    //

    ;; Store the class-id in meta data
    // Currently, we have only 8 classes, so we get only lower 4 bits 
    xbuf_extract(class_id,       \
                 ipv6_hdr,       \
                 IN_IPV6_OFFSET, \
                 IPV6_TRAFFIC_CLASS_LO4_FLOW_LABEL_HI4)
    alu[class_id, --, B, class_id, >>4]
    dl_meta_set_class_id(class_id)
   
    ;; Save next-hop-id and nexthop-id-type in metadata
    dl_meta_set_nexthop_id(nexthop_index)
    dl_meta_set_nexthop_id_type(nhid_type)

    ;; Save output port & blade id
    dl_meta_set_output_port(out_port)
    dl_meta_set_fabric_port(out_blade_id)

#ifdef IPV6_STATS
    //IPV6_OUT_TRANSMITS: Count of packets supplied to lower layer
    //IPV6_OUT_OCTETS: Number of bytes in IP packets delivered to
    //                 lower layers for transmission
    //We only update outgoing counters if it is on this blade
    _ipv6_increment_counter(outgoing_intf_counter_base, IPV6_OUT_TRANSMITS)
    _ipv6_add_to_counter(outgoing_intf_counter_base,IPV6_OUT_OCTETS,packet_size)
#endif //IPV6_STATS

    ;; Set the dl_next_block and finish processing
    br[ipv6_forwarder_finish#], defer[1]
      alu[dl_next_block, --, B, IPV6_NEXT_BLOCK]; 

hop_count_expired#:


    // If the hop count expired, set the right exception and setup
    // the dl_next_block to direct the packet to the XScale
    br[handle_exception#], defer[1]
      immed32[result, IPV6_EXCP_BAD_HOPLIMIT]

handle_error_exception#:
    br!=byte[result, 0, IX_DROP, handle_exception#]
    br[ipv6_forwarder_finish#], defer[1]
        alu[dl_next_block, --, B, IX_DROP]

handle_exception#:
    dl_set_exception[BID_IPV6, result]    
    alu[dl_next_block, --, B, IX_EXCEPTION]

ipv6_forwarder_finish#:
  
    xbuf_deactivate(ipv6_hdr)

.end

#endm

#endif    //    end __IPV6_FWDER_UC__

