////////////////////////////////////////////////////////////////////////////////
//                                                                     
//                  I N T E L   P R O P R I E T A R Y                   
//                                                                      
//     COPYRIGHT (c)  2001-2002 BY  INTEL  CORPORATION.  ALL RIGHTS          
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//     THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                      
//                        INTEL  CORPORATION                            
//                                                                     
//                     2200 MISSION COLLEGE BLVD                        
//                                                                      
//               SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                      
////////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: nh_extract.uc
//
//      Purpose: 
//            This file defines macros for accessing fields in the nexthop 
//         structure. The nexthop table consists of structures of two 
//         different types - "real" and "intermediate". The structure of 
//         both types of entries are described below.
//
//      History:
//
//
//      Date            Comment                                     By
// ---------------------------------------------------------------------------- 
//
//      09/11/2002      Created                                     
//
//
////////////////////////////////////////////////////////////////////////////////


#ifndef __NH_EXTRACT_UC__
#define __NH_EXTRACT_UC__

#include "ipv6.h"


//
//    Structure of a "real" next hop entry. 
//  Each entry is 4 long words (16 bytes) long.
//
//  Byte 3                  2                  1               0        
// |----------------|----------------|----------------|----------------|
// +-------------------------------------------------------------------+
// |               Reserved          |       Generic Flags             |
// |----------------------------------------------------  -------------|
// | Blade ID       |    Type |  Res |       Next Hop ID               | 
// |-------------------------------------------------------------------|
// |          MTU                    |          Output Port            |
// |-------------------------------------------------------------------|
// |                             Reserved                              |
// |-------------------------------------------------------------------|
//
//
//  Structure of an "intermediate" nexthop entry. 
//  Each entry is 4 long words (16 bytes) long.
// 
//  Byte 3                  2                  1               0        
// |----------------|----------------|----------------|----------------|
// +-------------------------------------------------------------------+
// | Status         |     Reserved   |       Generic Flags             |
// |----------------------------------------------------  -------------|
// |         Next Hop ID0            |        Next Hop ID1             | 
// |-------------------------------------------------------------------|
// |         Next Hop ID2            |        Next Hop ID3             | 
// |-------------------------------------------------------------------|
// |                                Reserved                           |
// |-------------------------------------------------------------------|
//

//
// Macro to read a nexthop entry and return the result of the read operation
// in a transfer register. 
//
// Inputs: $nexthop - symbolic name for SRAM transfer registers
//         next_hop_l2_id - the L2 index for the next hop that needs to be read.
//
// Outputs: result - IPV6_SUCCESS if the entry was successfully read
//                      IPV6_FAILURE if there was some error reading the entry.
//

#macro _nh_set_regs($nexthop)
    .set $nexthop/**/0, $nexthop/**/1, $nexthop/**/2, $nexthop/**/3
#endm

#macro nh_read_entry($nexthop, next_hop_l2_id)

    // Allocate a set of SRAM xfer registers.
    xbuf_alloc($nexthop, IPV6_NH_SIZE_LW, read)

.begin

    .reg sram_entry_offset, sram_address, sram_size
    .sig nh_sram_signal

    // First calculate the offset of the entry in the table (in bytes)
    // sram_entry_offset = next_hop_l2_id * 16
    alu_shf[sram_entry_offset, --, B, next_hop_l2_id, <<IPV6_NH_SIZE_SHIFT]    

    immed32[sram_address, IPV6_L3_NEXTHOP_TABLE_SRAM_BASE]
    //immed32[sram_size, IPV6_L3_NEXTHOP_TABLE_SIZE]

    alu[sram_address, sram_address, +, sram_entry_offset]



    // Read IPV6_NH_SIZE_LW long words from this location 
    // into the SRAM xfer registers.

    sram[read,             \
         $nexthop/**/0,    \
         0,                \
         sram_address,     \
         IPV6_NH_SIZE_LW], \
         ctx_swap[nh_sram_signal]

    // The 16 bytes of this entry should be in the $nexthop SRAM xfer 
    // register now!
.end

#endm

// This macro must be called to free up the SRAM transfer register 
// allocated using nh_read_entry(). See above.
#macro nh_free_entry($nexthop)
    xbuf_free($nexthop)
#endm


// nh_extract_macros
// 
// All of the nh_extract_* macros listed below work on the SRAM xfer 
// registers allocated earlier (see the above macro) and produce a desired 
// output field.

// Extract the Generic Flags field.
#macro nh_extract_flags($nexthop_entry, flags)
    _nh_set_regs($nexthop_entry)    
    xbuf_extract(flags, $nexthop_entry, 0, IPV6_NH_FLAGS)
#endm

#macro nh_extract_nhid($nexthop_entry, nhid)
    _nh_set_regs($nexthop_entry)
    xbuf_extract(nhid, $nexthop_entry, 0, IPV6_NH_ID)
#endm

// Extract the NextHop ID field.
#macro nh_extract_nhid_type($nexthop_entry, nhid_type)
.begin
    _nh_set_regs($nexthop_entry)
    .reg tmp_nhid_type

    xbuf_extract(tmp_nhid_type, $nexthop_entry, 0, IPV6_NH_ID_TYPE)

    // Also shift the 4-bits to the bottom 4 bits in the register
    alu_shf[nhid_type, --, B, tmp_nhid_type, >>4]
.end
#endm

// Extract the blade ID field.
#macro nh_extract_blade_id($nexthop_entry, blade_id)
    _nh_set_regs($nexthop_entry)
    xbuf_extract(blade_id, $nexthop_entry, 0, IPV6_NH_BLADE_ID)
#endm

// Extract the ouput_port field.
#macro nh_extract_output_port($nexthop_entry, port)
    _nh_set_regs($nexthop_entry)
    xbuf_extract(port, $nexthop_entry, 0, IPV6_NH_OUTPORT)
#endm

// Extract the MTU field.
#macro nh_extract_mtu($nexthop_entry, mtu)
    _nh_set_regs($nexthop_entry)
    xbuf_extract(mtu, $nexthop_entry, 0, IPV6_NH_MTU)
#endm

// Extract the status field.
#macro nh_extract_status($nexthop_entry, status)
    _nh_set_regs($nexthop_entry)
    xbuf_extract(status, $nexthop_entry, 0, IPV6_NH_STATUS)
#endm

// This macro is used if the entry is an "intermediate" entry. index is the 
// index of the "real" entry that is embedded in this entry and needs 
// to be returned in variable nhid. 
// 
// Output: result - IPV6_SUCCESS, if the read was successful.
//                     IPV6_FAILURE, if the opreation failed.

#macro nh_extract_nhid_at_index($nexthop_entry, index, nhid)
.begin
    _nh_set_regs($nexthop_entry)
    
    .if (index == 0)
        xbuf_extract(nhid, $nexthop_entry, 0, IPV6_NH_NHID0)
    .elif (index == 1)
        xbuf_extract(nhid, $nexthop_entry, 0, IPV6_NH_NHID1)
    .elif (index == 2)
        xbuf_extract(nhid, $nexthop_entry, 0, IPV6_NH_NHID2)
    .elif (index == 3)
        xbuf_extract(nhid, $nexthop_entry, 0, IPV6_NH_NHID3)
    .endif

.end
#endm

//
// These macros make it easy to inspect certain bits
// result is 1 if the bit is set, 0 otherwise
//
#macro _nh_is_bit_set(field, bit_mask, result)
.begin

    alu[--, field, AND, bit_mask]
    beq[bit_not_set#]

    immed[result, IPV6_SUCCESS]
    br[return_result#]

bit_not_set#:
    immed[result, IPV6_FAILURE]
    // fall through

return_result#:
    // done, return result
.end
#endm


// All the macros that look at the "flag" field in the nexthop entry

#macro _nh_is_flag_set($nexthop_entry, mask, result)
    _nh_set_regs($nexthop_entry)
    .reg flag

    nh_extract_flags($nexthop_entry, flag)
    _nh_is_bit_set(flag, mask, result)
#endm

// Is the "LOCAL" flag set?
#macro nh_is_local($nexthop_entry, result)
    _nh_set_regs($nexthop_entry)
    _nh_is_flag_set($nexthop_entry, IPV6_NH_FLAGS_LOCAL, result)
#endm

// Is the "DOWN" flag set?
#macro nh_is_down($nexthop_entry, result)
    _nh_set_regs($nexthop_entry)
    _nh_is_flag_set($nexthop_entry, IPV6_NH_FLAGS_DOWN, result)
#endm

// Is the "DROP" flag set?
#macro nh_is_drop($nexthop_entry, result)
    _nh_set_regs($nexthop_entry)
    _nh_is_flag_set($nexthop_entry, IPV6_NH_FLAGS_DROP, result)
#endm

// "MULTIPLE" flag set ?
#macro nh_is_intermediate_entry($nexthop_entry, result)
    _nh_set_regs($nexthop_entry)
    _nh_is_flag_set($nexthop_entry, IPV6_NH_FLAGS_MULTIPLE, result)
#endm


// Macros to test NextHop ID type

#macro _nh_is_id_type_set($nexthop_entry, mask, result)
    .reg nhid
    _nh_set_regs($nexthop_entry)

    nh_extract_nhid_type($nexthop_entry, nhid)
    _nh_is_bit_set(nhid, mask, result)
#endm

// "FORWARDER" Entry type?
#macro nh_is_id_fwdr($nexthop_entry, result)
    _nh_set_regs($nexthop_entry)
    _nh_is_id_type_set($nexthop_entry, IPV6_NH_ID_TYPE_FWRD, result)
#endm

// "TUNNEL" Entry Type?
#macro nh_is_id_tunnel($nexthop_entry, result)
    _nh_set_regs($nexthop_entry)
    _nh_is_id_type_set($nexthop_entry, IPV6_NH_ID_TYPE_TUNNEL, result)
#endm

// "COMPRESS" Entry Type?
#macro nh_is_id_compress($nexthop_entry, result)
    _nh_set_regs($nexthop_entry)
    _nh_is_id_type_set($nexthop_enrty, IPV6_NH_ID_TYPE_COMPRESS, result)
#endm



#endif //__NH_EXTRACT_UC__
