/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 5
 *
 *       
 *
 * = MODULE
 *      ICMPv6 module
 *
 * = FILENAME
 *      icmp.h
 *
 * = DESCRIPTION
 *    
 *    This file contains the definition of ICMPv6 message building
 *    functions. ICMPv6 messages supported are ICMPv6 parameter problem
 *    message, ICMPv6 destination unreachable message, ICMPv6 time
 *    exceeded error message and ICMPv6 redirect message. This file
 *    also contains internal functions for building ICMPv6 messages
 *    and queueing these into ICMPv6 queue.     
 *
 *
 * 
 * = CHANGE HISTORY
 *
 * ============================================================================
 * $Id: icmp6.h,v 1.13 2003/11/01 02:30:23 rranjeet Exp $
 */

#if !defined(__ICMPv6_H__)
#define __ICMPv6_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */

#include "utils.h"
#include "ix_cc_ipv6_init.h"

/**
 * Preprocessor symbol and macro definitions.
 */

/**
 * Type definitions.
 */
#define IX_CC_ICMPV6_MIN_RATE_LIMIT_TIME 1
#define IX_CC_ICMPV6_MAX_RATE_LIMIT_TIME 32
#define IX_CC_ICMPV6_DEFAULT_RATE_LIMIT_TIME 1
#define IX_CC_ICMPV6_ICMPQ_DEFAULT_SIZE 100
#define IX_CC_ICMPV6_DUMP_MSG_SIZE 32

#define IX_CC_ICMPV6_PKT_MTU 1280
#define IX_CC_ICMPV6_PAYLOAD_LIMIT 1218 /*1280 (MTU) - 18(ethernet hdr)
                                                -40(IPv6 hdr) - 4(ICMPv6 hdr)
                                                = 1218*/
/*length of the field always following the ICMP header for error messages*/
#define IX_CC_ICMPV6_REASON_FIELD_LENGTH 4

/*The following defines how many bytes of the payload can actually be used 
 * to fill in the error packet. This figure accounts for
 * the reason/parameter fieid that always follow the ICMP header. The contents
 * of this field depends upon the type of error ICMP message being sent*/
#define IX_CC_ICMPV6_ERR_PKT_PAYLOAD_LIMIT \
                IX_CC_ICMPV6_PAYLOAD_LIMIT - IX_CC_ICMPV6_REASON_FIELD_LENGTH 

/*ICMPv6 Message types*/
#define IX_CC_ICMPV6_TYPE_DEST_UNREACHABLE      1
#define IX_CC_ICMPV6_TYPE_PACKET_TOO_BIG        2
#define IX_CC_ICMPV6_TYPE_TIME_EXCEEDED         3
#define IX_CC_ICMPV6_TYPE_PARAMETER_PROBLEM     4

#define IX_CC_ICMPV6_CODE_ZERO              0

/*Codes for the Destination Unreachable Message type*/
#define IX_CC_ICMPV6_DU_CODE_NOROUTE            0
#define IX_CC_ICMPV6_DU_CODE_COMM_PROHIBITED    1
#define IX_CC_ICMPV6_DU_CODE_ADDR_UNREACHABLE   3
#define IX_CC_ICMPV6_DU_CODE_PORT_UNREACHABLE   4
    
/*Codes for the Time Exceeded Message type*/
#define IX_CC_ICMPV6_TE_CODE_HOP_LIMIT_EXCEEDED  0
#define IX_CC_ICMPV6_TE_CODE_REASSEMBLY_EXCEEDED 1

/*Codes for the Parameter  Problem Message type*/
#define IX_CC_ICMPV6_PP_CODE_WRONG_HDR_FIELD     0
#define IX_CC_ICMPV6_PP_CODE_NEXT_HDR_UNKNOWN    1
#define IX_CC_ICMPV6_PP_CODE_OPTION_UNKNOWN      2
    
#if 0
#define _IX_CC_IPV6_ICMPv6QUEUE_LINE_SIZE 50
#define _IX_CC_IPV6_ICMPv6QUEUE_TITLE_STRING  \
"Type Code Checksum \n"
#define _IX_CC_IPV6_ICMPv6QUEUE_TITLE_SIZE sizeof(_IX_CC_IPV6_ICMPv6QUEUE_TITLE_STRING)
#define _IX_CC_IPV6_ICMPv6QUEUE_TERM_STRING "\n"
#define _IX_CC_IPV6_ICMPv6QUEUE_TERM_SIZE sizeof(_IX_CC_IPV6_ICMPv6QUEUE_TERM_STRING)
#endif

/**
 * Prototypes for interface functions.
 */

/**
 * NAME: _ix_cc_ipv6_icmp_init
 *
 * DESCRIPTION: This routine initializes member variables to d
 *              default state. This function initializes 
 *              g_IcmpQueueDepth to IX_CC_IPV6_ICMPv6Q_MAXSIZE(8),
 *              g_IcmpSleepTime to 1 second and g_IcmpQueueDrain
 *              to IX_CC_IPV6_ICMPv6Q_MAXSIZE. This function also
 *              calls initialization function for queue
 * 
 * @Param:  - IN None
 *            
 * @Param:  - INOUT  - None
 * @Param:  - OUT - None  
 *
 *
 */

ix_error _ix_cc_ipv6_icmp_init(ix_cc_ipv6_context *pIpv6Context);

void _ix_cc_ipv6_icmp_fini(void);

/**
 * NAME: ix_cc_ipv6_icmp_param_problem
 *
 * DESCRIPTION: This function will build a complete ICMPv6 parameter problem
 *              error message for the packet passed in the buffer handle.  
 * 
 * @Param:  - IN arg_hBuffer - the buffer handle containing the packet that is 
 *        causing the ICMPv6 message to be generated 
 * @Param: -  IN arg_Param - Byte offset in the IP header that is
 *             causing the error. This parameter is in 
 *                      host order
 * @Param:  - INOUT  - None
 * @Param:  - OUT - None  
 *
 *
 */

ix_error _ix_cc_ipv6_icmp_param_problem(
            ix_buffer_handle arg_hBuffer, 
            ix_uint8 arg_Reason, 
            ix_uint32 arg_Param 
            );

/**
 * NAME: ix_cc_ipv6_icmp_pkt_too_big
 * DESCRIPTION: This function will build a complete ICMPv6 packet-too-big 
 *              messagefor the packet passed in the buffer handle
 * 
 * @Param:  - IN arg_hBuffer - the buffer handle containing the packet that is 
 *        causing the ICMPv6 message to be generated 
 * @Param: - IN arg_Reason - The actual MTU of the outgoing link.
 *
 */
ix_error _ix_cc_ipv6_icmp_pkt_too_big(
            ix_buffer_handle arg_hBuffer, 
            ix_uint16 arg_Reason );


/**
 * NAME: _ix_cc_ipv6_icmp_dest_unreachable
 *
 * DESCRIPTION: This function will build a complete ICMPv6 Destination
 *              Unreachable error message for the packet passed in
 *              the buffer handle.  
 * 
 * @Param:  - IN arg_hBuffer - the buffer handle containing the packet
 *             that is causing the ICMPv6 message to be generated
 *  
 *            IN arg_Reason - the reason that the destination is
 *               unreachable.For different values for this field,
 *               refer to header_defs.h file
 *
 * @Param:  - INOUT  - None
 * @Param:  - OUT - None  
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 *          
 */
ix_error _ix_cc_ipv6_icmp_dest_unreachable( 
                ix_buffer_handle arg_hBuffer,
                ix_uint8 arg_Reason);


/**
 * NAME: _ix_cc_ipv6_icmp_redirect
 *
 * DESCRIPTION: This function will build a complete ICMPv6 redirect error
 *        message for the packet passed in the buffer handle.  
 * 
 * @Param:  - IN arg_buffer_handle - the buffer handle containing the packet
 *           that is causing the ICMPv6 message to be generated
 * 
 *            IN arg_Gateway - the address of the gateway to which the traffic
 *            for the network specified in the IP destination address of
 *            the original datagram's data should be sent. This 
 *          parameter is expected in host byte order. When building the
 *            the message and sending it out, it is converted to
 *            network byte order
 *
 * @Param:  - INOUT  - None
 * @Param:  - OUT - None  
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 *
ix_error _ix_cc_ipv6_icmp_redirect( 
            ix_buffer_handle arg_hBuffer, 
            ix_uint32 arg_Gateway 
            );
*/

/**
 * NAME: ix_cc_ipv6_icmp_time_exceeded
 *
 * DESCRIPTION: This function will build a complete ICMPv6 time exceeded error
 *        message for the packet passed in the buffer handle.  
 * 
 * @Param:  - IN arg_hBuffer - the buffer handle containing the packet that
 *         is causing the ICMPv6 message to be generated 
 *           
 *            IN arg_Reason - the reason that the error message is being 
 *            generated. The two possible codes are 'hop limit exceeded' and
 *            'fragment reassembly time exceeded'. Defined in header_defs.h
 *
 * @Param:  - INOUT  - None
 * @Param:  - OUT - None  
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_cc_ipv6_icmp_time_exceeded( 
                ix_buffer_handle arg_hBuffer,
                ix_uint8 arg_Reason);

/*
* The rest of header file contains definition of APIs used
*     by above defined functions 
*/

/**
 * NAME: _ix_cc_ipv6_icmp_print_message
 *
 * DESCRIPTION: This function prints the header and the contents of 
 *              the ICMPv6 payload
 *         
 * 
 * @Param:  - IN arg_pMessage - the byte array containing an ICMPv6 message,
 *                              that is,an ICMPv6 header and its payload
 *         arg_Size - the number of bytes in the ICMPv6 message 
 *           
 *
 * @Param:  - INOUT  - None
 * @Param:  - OUT - None  
 *
 * @Return: None
 */
void _ix_cc_ipv6_icmp_print_message( 
            const ix_uint8 *arg_pMessage, 
            ix_uint32 arg_Size 
            );


/**
 * NAME: _ix_cc_ipv6_icmp_dump_message
 *
 * DESCRIPTION: This function prints the ICMPv6 header 
 *  into a string buffer
 * 
 * @Param:  - IN arg_pMessage - the byte array containing an ICMPv6 message,
 *         that is, an ICMPv6 header 
 *          IN arg_MessageSize - the number of bytes in the ICMPv6 message
 *            IN   arg_BufferSize - the size of buffer which will get a dump
 *               of ICMPv6 header
 *       
 *
 * @Param:  - INOUT  - None
 * @Param:  - OUT - arg_pBuffer - the string buffer where the header
 *         is printed
 *
 * @Return: IX_SUCCESS if successful or a valid error token on failure
 */
ix_error _ix_cc_ipv6_icmp_dump_message(const ix_uint8 *arg_pMessage,
                                    ix_uint32 arg_MessageSize,
                       ix_uint32 arg_BufferSize,
                                    ix_uint8 *arg_pBuffer
                    );




/**
 * NAME: _ix_cc_ipv6_icmp_build_err_message
 *
 * DESCRIPTION: This function builds a complete ICMPv6 error message and 
 *         queues it so it can be sent. Note that the enqueue operation 
 *         may fail due to queue being full. In that case, the ICMPv6 
 *         message will not be sent and the error packet will simply be 
 *         dropped. When forming the icmp message, all data is filled up 
 *         in network byte order
 * 
 * @Param:  - IN - arg_Type - the type field for the ICMPv6 header
 *            IN - arg_Code - the code field for the ICMPv6 header 
 *          IN - arg_Param - the data for the parameter field of the header
 *                       This parameter is in host byte order
 *           arg_hBuffer - the buffer containing the IP packet
 *             causing error
 *                 arg_SendMsg - This value indicates if ICMPv6 message has to 
 *                                be sent.
 * @Param:  - INOUT  - None
 * @Param:  - OUT 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure
 */
ix_error _ix_cc_ipv6_icmp_build_err_message( 
                ix_uint8 arg_Type, 
                ix_uint8 arg_Code, 
                ix_uint32 arg_Param,
                ix_uint32 arg_SendMsg, 
                ix_buffer_handle arg_hBuffer 
                );


ix_error _ix_cc_ipv6_icmp_build_info_message(
                ix_ipv6_addr    arg_SrcAddr,
                ix_ipv6_addr    arg_DstAddr,
                ix_uint8        arg_Type,
                ix_uint8        arg_Code,
                ix_uint32       arg_Length,
                ix_uint8        *arg_Buffer,
                ix_uint16       arg_InputPort);

/**
 * NAME: _ix_cc_ipv6_icmp_copy_payload
 *
 * DESCRIPTION: This function attempts to copy the number of bytes
 *              specified in length plus icmp header length.
 * 
 * @Param:  - IN - arg_pSource -  byte array of data to copy
 *                 arg_Length -maximum number of bytes to copy 
 *           
 *           
 * @Param:  - INOUT  - None
 * @Param:  - OUT arg_pDest - byte array to copy data into 
 *
 * @Return: the number of bytes actually copied
 */
ix_uint32 _ix_cc_ipv6_icmp_copy_payload( 
                ix_uint8 *arg_pSource, 
                ix_uint8 *arg_pDest, 
                ix_uint32 arg_Length 
                );

/**
 * NAME: _ix_cc_ipv6_icmp_ok_to_send
 *
 * DESCRIPTION: This function verifies that the given packet can cause 
 *              an ICMPv6 error message to be generated. RFC2463 states
 *              several rules (section 2.4)
 *        for when a packet must generate an ICMPv6 error message. 
 *              This routine compares the error packet to those rules
 *              and determines if an ICMPv6 error
 *        message may or may not be sent.
 *        
 * 
 * @Param:  - IN - arg_msgType -  The type of ICMP message being generated
 * @Param:  - IN - arg_hBuffer -  the buffer containing the packet
 *             to verify
 *           
 *           
 * @Param:  - INOUT  - None
 * @Param:  - OUT  - arg_pOkToSend =1 if packet may generate an ICMPv6 error,
 *                 0 otherwise 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token on failure
 */
ix_error _ix_cc_ipv6_icmp_ok_to_send(
                           ix_uint16 arg_msgType,
                           ix_buffer_handle arg_hBuffer,
                           ix_uint32 *arg_pOkToSend
                           );

/**
 * NAME: _ix_cc_ipv6_icmp_send_message
 *
 * DESCRIPTION: This function attempts to dequeue ICMPv6 messages and
 *              send them to the destination. This is the ICMPv6
 *              queue processing loop. This is the event handler
 *              function that will be registered in the init 
 *              function of IPv4 Forwarder Core Component
 *        
 * @Param:  - IN - arg_pContext - Context returned by
 *                       ix_cci_cc_add_event_handler( ) function
 *           
 *           
 * @Param:  - INOUT  - None
 * @Param:  - OUT  - None 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error on failure
 */

ix_error _ix_cc_ipv6_icmp_send_err_message(void *arg_pContext);
                

/*
*
*/

ix_uint32 _ix_cc_ipv6_icmp_calculate_checksum(ix_uint8 *icmpMessage);

ix_error _ix_cc_ipv6_icmp_validate_checksum(ix_uint8    *arg_pIpv6Pkt,
                                            ix_uint16   arg_icmpLength,
                                            ix_uint8    *arg_pIcmpHdr);

ix_error _ix_cc_ipv6_icmp_build_buffer_chain(ix_uint8 *arg_pIpv6Pkt,
                                             ix_uint16 arg_inputPort,
                                             ix_buffer_handle *arg_hBuffer);

ix_error _ix_cc_ipv6_icmp_set_rate_limit(ix_uint16 rate);
ix_error _ix_cc_ipv6_icmp_get_rate_limit(ix_uint16 *rate);

/* This function simply generates the ICMP message buffer. It doesn't transmit
 * it. The caller has the responsibility of transmitting the buffer*/ 
ix_error _ix_cc_ipv6_icmp_build_info_message_ex(
                ix_ipv6_addr    arg_SrcAddr,
                ix_ipv6_addr    arg_DstAddr,
                ix_uint8        arg_Type,
                ix_uint8        arg_Code,
                ix_uint32       arg_Length,
                ix_uint8        *arg_packet,
                ix_uint16       arg_InputPort,
                ix_buffer_handle *arg_HwBuffer);

/**
 * Exported variables.
 */
extern ix_uint16 g_Icmpv6QueueDepth;
extern ix_uint16 g_Icmpv6QueueDrain;
extern ix_uint16 g_Icmpv6SleepTime;



#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */


#endif /* end !defined(__ICMPv6_H__) */


