/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 *
 * = MODULE
 *      ICMP module of IPv6 Fwdr CC
 *
 * = FILENAME
 *      utils.h
 *
 * = DESCRIPTION
 *   This file contains utility functions like calculation of IP
 *   packet checksum and calculation of Identification field
 *   to be filled up in IP header for formation of
 *   ICMP messages   
 *
 * 
 *      
 *
 * = CREATION TIME
 *      12/03/2002 10:37 AM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */

#if !defined(__UTILS_H__)
#define __UTILS_H__

#include <assert.h>

#include "ix_cc_error.h"
#include "v6_header_defs.h"
#include "cc/ix_cc_rtmv6.h"
#include "cc/ix_cc_ipv6.h"
#include "cc/ipv6/internal/ix_cc_ipv6_init.h"
#include "cc/ipv6/internal/ix_cc_ipv6_prop.h"
#include "ix_netmacros.h"


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */

/* execute expression "__expr"; on error goto label "__label" */
#define IX_IPV6_ERROR_CGT(__expr, __err, __label) \
        {                                         \
            __err = __expr;                       \
            if(__err != IX_SUCCESS)               \
            goto __label;                         \
        }

extern ix_cc_ipv6_registry_data g_Ipv6RegistryData;

/**
 * Prototypes for interface functions.
 */

/**
 * NAME: _ix_cc_ipv6_init_ip_header
 *
 * DESCRIPTION: This function copies the data (size is equal to the basic
 *              IPv6 header length) from the data pointed to by arg_pData 
 *              into the IPv6 header structure pointed to by arg_pIpHdr
 * 
 * @Param:  - IN arg_pData - pointer to data to copy onto IPv6 header
 *		
 * @Param:  - OUT  arg_pIphdr - Pointer to IPv6 header
 *
 */

void _ix_cc_ipv6_init_ip_header(ix_uint8 *arg_pData, 
                                ix_cc_ipv6_header *arg_pIphdr);

/**
 * NAME: _ix_cc_print_ipv6_address
 *
 * DESCRIPTION: This function prints an IPv6 address.
 * 
 * @Param:  - IN pAddr - pointer to the IPv6 address.
 *
 * return:
 * none.
 *
 */
void _ix_cc_print_ipv6_address(ix_uint128 *pAddr);


/**
 * NAME: _ix_cc_print_ipv6_header
 *
 * DESCRIPTION: This utility function prints the contents of an IPv6 header
 * 
 * @Param:    IN arg_pHdr - Pointer to IPv6 header
 *
 * @Param:  - INOUT  - None
 * @Param:  - OUT - None
 *
 */

void _ix_cc_ipv6_print_iphdr(ix_cc_ipv6_header *arg_pHdr);


/* 
 * The following macro's expect the IPv6 address to be in 
 * Host Byte Order Only.
 */

#define IX_127_96(ipaddr) (ipaddr.aUint32[0])
#define IX_95_64(ipaddr)  (ipaddr.aUint32[1])
#define IX_63_32(ipaddr)  (ipaddr.aUint32[2])
#define IX_31_0(ipaddr)   (ipaddr.aUint32[3])

/* NAME: IS_IPV6_ADDRESS_ZERO
 *
 * DESCRIPTION: Check to see if an IPv6 address is zero.
 *
 * @Param: IN arg_src - Pointer to the IPv6 address.
 *
 * return: 1 if it is.
 *         0 Otherwise
 * 
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
ix_uint8
IS_IPV6_ADDRESS_ZERO(ix_uint128 *arg_ipv6Addr);


/* NAME: ZERO_IPV6_ADDRESS
 *
 * DESCRIPTION: Zero's out an IPv6 address
 *
 * @Param: IN arg_src - Pointer to the IPv6 address.
 *
 * return: none.
 *
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
void
ZERO_IPV6_ADDRESS(ix_uint128 *arg_ipv6Addr);

/* NAME: COPY_IPV6_ADDRESS
 *
 * DESCRIPTION: Copies the source IPv6 address into the destination.
 *
 * @Param: IN arg_dst - Pointer to the destination IPv6 address.
 * @Param: IN arg_src - Pointer to the source IPv6 address.
 *
 * return: none.
 *
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
void
COPY_IPV6_ADDRESS(ix_uint128 *arg_dst, ix_uint128 *arg_src);

/* 
 * NAME: IS_IPV6_ADDRESS_EQUAL
 *
 * DESCRIPTION: Compare two IPv6 addresses. 
 *
 * @Param: IN arg_src - Pointer to the IPv6 address.
 * @Param: OUT arg_dst - Pointer to where the result of the operation 
 *                        should be stored.
 * return: 1 if they are equal
 *         0 Otherwise
 *
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
ix_uint8
IS_IPV6_ADDRESS_EQUAL(ix_uint128 *arg_src, ix_uint128 *arg_dst);

/* 
 * NAME: IS_IPV6_ADDRESS_MUTLICAST
 *
 * DESCRIPTION: Checks to see if an IPv6 address is Multicast. 
 * If the Most Significant Byte of the address is 0xFF, it is.
 *
 * @Param: IN arg_src - The IPv6 address.
 *
 * return: 1 if it is a Multicast address.
 *         0 Otherwise
 *
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
ix_uint8
IS_IPV6_ADDRESS_MULTICAST(ix_uint128 ipv6Addr);

/* 
 * NAME: IS_IPV6_ADDRESS_LINK_LOCAL
 *
 * DESCRIPTION: Checks to see if an IPv6 address is Link Local
 *
 * @Param: IN arg_src - The IPv6 address.
 *
 * return: 1 if it is a Link Local
 *         0 Otherwise
 *
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
ix_uint8
IS_IPV6_ADDRESS_LINK_LOCAL(ix_uint128 ipv6Addr);

/* 
 * NAME: IS_IPV6_ADDRESS_SITE_LOCAL
 *
 * DESCRIPTION: Checks to see if an IPv6 address is Site Local
 *
 * @Param: IN arg_src - The IPv6 address.
 *
 * return: 1 if it is a Site Local
 *         0 Otherwise
 *
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
ix_uint8
IS_IPV6_ADDRESS_SITE_LOCAL(ix_uint128 ipv6Addr);

/* 
 * NAME: IS_IPV6_ADDRESS_LOOPBACK
 *
 * DESCRIPTION: Checks to see if an IPv6 address is Loopback (::1). 
 *
 * @Param: IN arg_src - The IPv6 address.
 *
 * return: 1 if it is a Multicast address.
 *         0 Otherwise
 *
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
ix_uint8
IS_IPV6_ADDRESS_LOOPBACK(ix_uint128 ipv6Addr);

/* 
 * NAME: SET_LOOPBACK_ADDRESS
 *
 * DESCRIPTION: Sets the address to loopback (::1)
 *
 * @Param: IN arg_src - The IPv6 address.
 *
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
void
SET_LOOPBACK_ADDRESS(ix_uint128 *ipv6Addr);


/* Reference: Per RFC 2373 - This is what a multicast address looks like
 *     
 *   8 bits      4        4                   112 bits
 *  ---------------------------------------------------------------------
 *  |11111111| 000T | scope |            group ID                      |
 *  ---------------------------------------------------------------------
 * A multicast address can be "well known" (T = 0) or transient (T = 1).
 *
 * Based on the above definition:
 * 
 * "all nodes" multicast addresses can be 
 *        FF01:0:0:0:0:0:0:1  (node-local scope)
 *        FF02:0:0:0:0:0:0:1  (link-local scope)
 *
 * "all routers" multicast addresses can be
 *        FF01:0:0:0:0:0:0:2  (node-local scope)
 *        FF02:0:0:0:0:0:0:2  (link-local scope)
 *        FF05:0:0:0:0:0:0:2  (site-local scope)
 *
 * "solicited node" multicast address
 *        FF02:0:0:0:0:1:FFXX:XXXX 
 *  where XX:XXXX consists of the last 24 bits of the node's unicast 
 *  or anycast address
 */


/* 
 * NAME: IS_IPV6_ADDRESS_ALL_NODE_MULTICAST
 *
 * DESCRIPTION: Checks to see if an IPv6 address is an 
 * "all node" multicast address.
 *
 * @Param: IN ipv6Addr - The IPv6 address.
 *
 * return: 1 if true
 *         0 if false
 *
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
ix_uint8
IS_IPV6_ADDRESS_ALL_NODE_MULTICAST(ix_uint128 ipv6Addr);

/* Is it "all node" node-local multicast ? */
#define IS_IPV6_ADDRESS_ALL_NODE_NL_MULTICAST(ipv6Addr) \
    (IS_IPV6_ADDRESS_ALL_NODE_MULTICAST(ipv6Addr) && \
                                ((IX_127_96(ipv6Addr) >> 16) == 0xFF01))
           
/* Is it "all node" link-local multicast ? */
#define IS_IPV6_ADDRESS_ALL_NODE_LL_MULTICAST(ipv6Addr) \
    (IS_IPV6_ADDRESS_ALL_NODE_MULTICAST(ipv6Addr) && \
                                ((IX_127_96(ipv6Addr) >> 16) == 0xFF02))

/* 
 * NAME: IS_IPV6_ADDRESS_ALL_ROUTER_MULTICAST
 *
 * DESCRIPTION: Checks to see if an IPv6 address is an 
 * "all routers" multicast address.
 *
 * @Param: IN ipv6Addr - The IPv6 address.
 *
 * return: 1 if true
 *         0 if false
 *
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
ix_uint8
IS_IPV6_ADDRESS_ALL_ROUTER_MULTICAST(ix_uint128 ipv6Addr);

/* Is it "all router" node-local multicast ? */
#define IS_IPV6_ADDRESS_ALL_ROUTER_NL_MULTICAST(ipv6Addr) \
    (IS_IPV6_ADDRESS_ALL_ROUTER_MULTICAST(ipv6Addr) && \
                                ((IX_127_96(ipv6Addr) >> 16) == 0xFF01))

/* Is it "all router" link-local multicast ? */
#define IS_IPV6_ADDRESS_ALL_ROUTER_LL_MULTICAST(ipv6Addr) \
    (IS_IPV6_ADDRESS_ALL_ROUTER_MULTICAST(ipv6Addr) && \
                                ((IX_127_96(ipv6Addr) >> 16) == 0xFF02))

/* Is it "all router" site-local multicast ? */
#define IS_IPV6_ADDRESS_ALL_ROUTER_SL_MULTICAST(ipv6Addr) \
    (IS_IPV6_ADDRESS_ALL_ROUTER_MULTICAST(ipv6Addr) && \
                                ((IX_127_96(ipv6Addr) >> 16) == 0xFF05))
/* 
 * NAME: IS_IPV6_ADDRESS_SOLICITED_NODE_MULTICAST
 *
 * DESCRIPTION: Checks to see if an IPv6 address is a 
 * "solicited node" multicast address.
 *
 * @Param: IN ipv6Addr - The IPv6 address.
 * @Param: IN portIpv6Addr - The Ports IPv6 Address
 *
 * return: 1 if true
 *         0 if false
 *
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
ix_uint8
IS_IPV6_ADDRESS_SOLICITED_NODE_MULTICAST(ix_uint128 ipv6Addr, 
                                         ix_uint128 portIpv6Addr);

/**
 * Exported variables.
 */

#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__UTILS_H__) */
