/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor
 *
 * = MODULE
 *      IPv6 Forwarder Header Validations
 *
 * = FILENAME
 *      v6_header_validation.h
 *
 * = DESCRIPTION
 *     This file contains definition of functions used for IPv6 header 
 *     validation according to RFC2460
 *      
 * = CREATION TIME
 *      12/13/2002
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */

#if !defined(__V6_HEADER_VALIDATION_H__)
#define __V6_HEADER_VALIDATION_H__

#if defined(__cplusplus)

extern "C"
{
#endif /* end defined(__cplusplus) */

/**
 * Preprocessor symbol and macro definitions.
 */
#include "ix_cc_error.h"
#include "v6_header_defs.h"
#include "utils.h"
#include "ix_netmacros.h"


/**
 * Type definitions.
 */

/**
 * Prototypes for interface functions.
 */

/**
 * NAME: _ix_cc_ipv6_IPheader_validate
 *
 * DESCRIPTION: This function validates the IPv6 header according to RFC 2460.
 * 
 * @Param:  - IN  arg_pIpHdr  - a pointer to IPv6 header	  
 *                arg_hBuffer - a buffer containing the packet
 *                              with IPv6 header
 *                arg_inCounters - argument to indicate "maintain in counters"
 *                arg_outCounters - argument to indicate "maintain out counters"
 * @Param:  - INOUT  None
 * @Param:  - OUT arg_pDropped - value to tell whether the packet
 *				should be dropped or forwarded  
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token
 *             for failure.
 *
 * Error codes: IX_CC_IPV6_ERROR_INVALID_HEADER
 *              IX_CC_ERROR_INTERNAL
 */

ix_error _ix_cc_ipv6_IPheader_validate(ix_cc_ipv6_header *arg_pIpHdr,
			               ix_buffer_handle arg_hBuffer,
                                       ix_uint32 *arg_pDropped,
				       ix_uint8 arg_inCounters,
				       ix_uint8 arg_outCounters);


/**
 * NAME:_ix_cc_ipv6_validate_IP_addresses 
 *
 * DESCRIPTION: This function checks the Destination
 *              IPv6 address and Source IPv6 address in the
 *              given buffer to make sure that they do not contain any 
 *              unrouteable addresses according to RFC 2460 
 * 
 * @Param:  - IN  arg_pIpHdr - a pointer to IPv6 header
 *		  arg_hBuffer - a buffer containing the packet
 *                              with IPv6 header
 * @Param:  - INOUT  None
 * @Param:  - OUT arg_pDropped - value to tell whether the packet
 *				should be dropped or forwarded  
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token
 *             for failure.
 * Error codes: IX_CCI_ERR_SEND_FAIL
 *              IX_CC_ERROR_INTERNAL
 */
ix_error _ix_cc_ipv6_validate_IP_addresses(ix_cc_ipv6_header *arg_pIpHdr, 
                                           ix_buffer_handle arg_hBuffer,
                                           ix_uint32 *arg_pDropped, 
					   ix_uint8 arg_inCounters,
					   ix_uint8 arg_outCounters);

/**
 * NAME:_ix_cc_ipv6_check_for_solnode_address
 *
 * DESCRIPTION: This function checks the Destination
 *              IPv6 address is a Solicited Node Address as defined
 *              in RFC 2373.
 * 
 * @Param:  - IN  arg_destAddr - The Destination IPv6 address that needs
 *                               to be checked.
 *
 * @Return: IX_SUCCESS - If it is a solicited node multicast address.
 *          IX_CC_ERROR_ENTRY_NOT_FOUND - If it is not a multicast address.
 *          IX_CC_ERROR_INTERNAL - Internal error.
 */

ix_error _ix_cc_ipv6_check_for_solnode_address(ix_buffer_handle arg_hBuffer,
                                               ix_uint128 arg_destAddr);

/**
 * NAME: _ix_cc_ipv6_inspect_pkt_hdr
 *
 * DESCRIPTION: This function inspects an IPv6 packet header for possible
 *              extension headers and calls the appropriate functions to
 *              deal with the headers it finds.
 *
 * @Param:  - IN  arg_hBuffer - The buffer containing the IPv6 packet.
 * @Param:  - OUT arg_pDrop -   If 1, the packet has been dropped. The caller
 *                             should simply return.
 *
 * @Return: IX_SUCCESS - If successful.
 */
ix_error _ix_cc_ipv6_inspect_pkt_hdr(ix_buffer_handle arg_hBuffer,
                                     ix_uint8 *arg_pDrop);

/**
 * Exported variables.
 */

#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__V6_HEADER_VALIDATION_H__) */
