/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002-2003 Intel Corporation. All rights reserved.

 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor
 *
 * 
 *
 * = MODULE
 *     IPv6 Core Component - ICMP module 
 *
 * = FILENAME
 *      icmp6.c
 *
 *
 * = DESCRIPTION
 *   This file contains the implementation of ICMP message building 
 *   functions.
 *   ICMP messages supported are ICMP parameter problem message, ICMP 
 *   destination unreachable message and the ICMP time exceeded error message.
 *
 * = CREATION TIME
 *      6/15/2002 5:56:24 PM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */
#define IX_ERROR_FILE_IDENT "$Id: icmp6.c,v 1.34 2003/12/02 01:08:37 ktseng Exp $ $Author: ktseng $ $Date: 2003/12/02 01:08:37 $ $Revision: 1.34 $"


/**
 * User defined include files required.
 */
#include "ix_error.h"
#include "ix_cc_error.h"
#include "ix_cc_properties.h"
#include "cc/ipv6/internal/icmp6.h"
#include "cc/ipv6/internal/utils.h"
#include "cc/ipv6/internal/ipv6_fwdr.h"
#include "cc/ipv6/internal/ix_cc_ipv6_init.h"

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
#include "cc/ipv6/internal/ix_cc_ipv6_slist.h"
#else
#include "cc/ix_cc_slist.h"
#endif

#if !defined(IX_EXCLUDE_CCI)
#include "cci/cci_types.h" 
#endif /* IX_EXCLUDE_CCI */

#ifdef MEAT_TEST
 LPLIST g_Context_Icmp;
#endif

 /**
 * Preprocessor symbols and macros used in this file.
 */
#define IX_CC_ICMP_HW_BUFFER_SIZE 2048

#define ICMP_LOCK() ix_ossl_mutex_lock(g_icmp_mutex, IX_OSSL_WAIT_FOREVER)
#define ICMP_UNLOCK() ix_ossl_mutex_unlock(g_icmp_mutex)

/**
 * Types definitions whose scope is limited to this file.
 */

/**
 * Variable declarations global to this file only.  Externs are followed by
 * static variables.
 */
ix_cc_properties_node *g_Ipv6Properties;
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
extern ix_cc_ipv6_context *g_pIpv6Context;
#else
ix_cc_ipv6_context *g_pIpv6Context;
#endif

/**
 * ICMP error message queue. Used to implement rate limiting for error
 * messages. Info messages are not rate limited
 */
LPLIST g_IcmpMsgQ;

/**
* g_Icmpv6QueueDepth represents the depth of the ICMP error message queue.
* That is, the maximum number of ICMP messages that can fit in the queue,
* not the number of messages currently in the queue
**/
ix_uint16 g_Icmpv6QueueDepth;

/**
* g_Icmpv6QueueDrain represents the maximum number of ICMP error messages 
* to send (from the ICMP message queue) each time the queue processing
* event fires. Not used currently. It may be used in the future if a different
* queue management algorithm is implemented.
**/
ix_uint16 g_Icmpv6QueueDrain;

/**
*  g_Icmpv6SleepTime represents the number of seconds between when the event
*  will fire causing the ICMP messages to be dequeued and sent. The lower 
*  the value of g_Icmpv6SleepTime, the more frequent the ICMP messages will
*   be dequeued    and sent.
**/
ix_uint16 g_Icmpv6SleepTime;

/**
 * Lock for accessing the queue
 */
static ix_ossl_mutex_t g_icmp_mutex;

/**
 * Event handle used for setting queue events
 */
#if !defined(IX_EXCLUDE_CCI)
ix_event_handle g_IcmpQEventHdl;
#endif /*IX_EXCLUDE_CCI*/

/**
 * Local function prototypes
 */
static
ix_uint8 _ix_cc_ipv6_icmp_is_icmp_err_pkt(ix_uint8  *arg_pIpv6Pkt);

static
ix_error _ix_cc_ipv6_icmp_queue_message(ix_buffer_handle arg_hBuffer);

ix_error _ix_cc_ipv6_icmp_send_err_message(void *arg_pContext);

/**
 * Function definitions.
 */

/* Initialization function */
ix_error _ix_cc_ipv6_icmp_init(ix_cc_ipv6_context *arg_pIpv6Context)
{
    ix_error errVal;

    if(!arg_pIpv6Context)
        return IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM;

    g_pIpv6Context = arg_pIpv6Context;

    /*Initialize the Rate Limiting mechanism*/
    g_Icmpv6QueueDepth = IX_CC_ICMPV6_ICMPQ_DEFAULT_SIZE;
    g_Icmpv6SleepTime = IX_CC_ICMPV6_DEFAULT_RATE_LIMIT_TIME;

    if(ix_ossl_mutex_init(IX_OSSL_MUTEX_UNLOCK, &g_icmp_mutex) != IX_SUCCESS)
        return IX_CC_ERROR_OOM;
    
    ICMP_LOCK();
    g_IcmpMsgQ = ListCreate();
    if(!g_IcmpMsgQ)
    {
        ICMP_UNLOCK();
        ix_ossl_mutex_fini(g_icmp_mutex);
        return IX_CC_ERROR_OOM;
    }

    /*Set up the event timer for ICMP*/
#if !defined(IX_EXCLUDE_CCI)
    errVal = ix_cci_cc_add_event_handler(g_pIpv6Context->hIpv6Handle,
                                         g_Icmpv6SleepTime,
                                         _ix_cc_ipv6_icmp_send_err_message,
                                         IX_EVENT_TYPE_ONESHOT,
                                         1, &(g_IcmpQEventHdl));
    if(errVal != IX_SUCCESS)
    {
        ListDelete(g_IcmpMsgQ);
        ICMP_UNLOCK();
        ix_ossl_mutex_fini(g_icmp_mutex);
        return errVal;
    }
#elif defined(MEAT_TEST)
    g_Context_Icmp = ListCreate();
    if(!g_Context_Icmp)
    {
        ListDelete(g_IcmpMsgQ);
        ICMP_UNLOCK();
        return IX_CC_ERROR_OOM;
    }
    
    ListAddItem(g_Context_Icmp, _ix_cc_ipv6_icmp_send_err_message);
#endif

    ICMP_UNLOCK();
    
#if defined(IPV6_DEBUG)
    ix_ossl_message_log("ICMPv6 Initialized successfully.\n");
#endif /* IPV6_DEBUG */
    return IX_SUCCESS;
}

void _ix_cc_ipv6_icmp_fini()
{
    /*_ix_cc_ipv6_icmp_cleanup_q();*/
    ICMP_LOCK();
#if !defined(IX_EXCLUDE_CCI)    
    ix_cci_cc_remove_event_handler(g_pIpv6Context->hIpv6Handle,
                                   g_IcmpQEventHdl);
#endif/*IX_EXCLUDE_CCI*/

    ListDelete(g_IcmpMsgQ);
    ICMP_UNLOCK();
    ix_ossl_mutex_fini(g_icmp_mutex);
}

/* 
 * This function builds ICMP Parameter Problem message for the packet 
 * passed in the buffer handle 
 */
ix_error _ix_cc_ipv6_icmp_param_problem(
                                    ix_buffer_handle arg_hBuffer, 
                                    ix_uint8         arg_Reason,
                                    ix_uint32        arg_Param )
{
    ix_uint32 OkToSend; 

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("Entering ICMP Param Problem.\n");
#endif /* IPV6_DEBUG */

    IX_ERROR_CR(_ix_cc_ipv6_icmp_ok_to_send(
                                        IX_CC_ICMPV6_TYPE_PARAMETER_PROBLEM,
                                        arg_hBuffer, &OkToSend));
    if (!OkToSend)
    {
        /* Free the buffer */
        IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hBuffer), 
                     IX_CC_ERROR_INTERNAL, 0);

#if defined(IPV6_DEBUG)
        ix_ossl_message_log("ICMP Param Problem: Pkt Validation failed.\n");
#endif /* IPV6_DEBUG */

        return IX_SUCCESS;
    }

    IX_ERROR_CR(_ix_cc_ipv6_icmp_build_err_message
                           (IX_CC_ICMPV6_TYPE_PARAMETER_PROBLEM, 
                           arg_Reason, arg_Param, 1, arg_hBuffer)); 

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("Exiting ICMP Param Problem.\n");
#endif /* IPV6_DEBUG */
    return IX_SUCCESS;
}

/* 
 * This function attempts to build ICMP destination unreachable error 
 * message for the packet passed in buffer handle 
 */
ix_error _ix_cc_ipv6_icmp_dest_unreachable(
                ix_buffer_handle arg_hBuffer,
                ix_uint8 arg_Reason)
{
    
    ix_uint32 OkToSend; 

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("Entering ICMP Destination Unreacheble.\n");
#endif /* IPV6_DEBUG */

    IX_ERROR_CR(_ix_cc_ipv6_icmp_ok_to_send(IX_CC_ICMPV6_TYPE_DEST_UNREACHABLE,
                                            arg_hBuffer, &OkToSend));

    if (!OkToSend)
    {

        /* Free the buffer */
        IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hBuffer), 
                               IX_CC_ERROR_INTERNAL, 0);
#if defined(IPV6_DEBUG)
        ix_ossl_message_log("ICMP Dest unreach: Pkt Validation failed.\n");
#endif /* IPV6_DEBUG */

        return IX_SUCCESS;
    }

    IX_ERROR_CR(_ix_cc_ipv6_icmp_build_err_message
                          (IX_CC_ICMPV6_TYPE_DEST_UNREACHABLE, 
                          arg_Reason, 0, 1, arg_hBuffer));

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("Exiting ICMP Destination Unreacheble.\n");
#endif /* IPV6_DEBUG */

    return IX_SUCCESS;
}

/*
 * This function attempts to build ICMP time Exceeded error message 
 * for the packet passed in as buffer handle
 */
ix_error _ix_cc_ipv6_icmp_time_exceeded(
                ix_buffer_handle arg_hBuffer,
                ix_uint8 arg_Reason)
{
    ix_uint32 OkToSend; 

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("Entering ICMP Time Exceeded.\n");
#endif /* IPV6_DEBUG */

    IX_ERROR_CR(_ix_cc_ipv6_icmp_ok_to_send(IX_CC_ICMPV6_TYPE_TIME_EXCEEDED,
                                            arg_hBuffer, &OkToSend));

    if (!OkToSend)
    {
        /* Free the buffer */
        IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hBuffer), 
                                                  IX_CC_ERROR_INTERNAL, 0);

#if defined(IPV6_DEBUG)
        ix_ossl_message_log("ICMP time exceeded: Pkt Validation failed.\n");
#endif /* IPV6_DEBUG */

        return IX_SUCCESS;
    }

    IX_ERROR_CR(_ix_cc_ipv6_icmp_build_err_message
                                (IX_CC_ICMPV6_TYPE_TIME_EXCEEDED,
                                arg_Reason, 0, 1, arg_hBuffer));

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("Exiting ICMP Time Exceeded.\n");
#endif /* IPV6_DEBUG */

    return IX_SUCCESS;
}

/*
 * This function attempts to build a complete ICMP packet-too-big message
 * for the packet passed in as the buffer handle
 */
ix_error _ix_cc_ipv6_icmp_pkt_too_big(
                        ix_buffer_handle arg_hBuffer,
            ix_uint16 arg_Reason)
{
    ix_uint32 OkToSend; 

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("Entering ICMP Packet Too Big.\n");
#endif /* IPV6_DEBUG */

    IX_ERROR_CR(_ix_cc_ipv6_icmp_ok_to_send(IX_CC_ICMPV6_TYPE_PACKET_TOO_BIG,
                                            arg_hBuffer, &OkToSend));
    if (!OkToSend)
    {
        /* Free the buffer */
        IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hBuffer), 
                                                IX_CC_ERROR_INTERNAL, 0);
#if defined(IPV6_DEBUG)
        ix_ossl_message_log("ICMP Pkt too big: Pkt Validation failed.\n");
#endif /* IPV6_DEBUG */

        return IX_SUCCESS;
    }

    IX_ERROR_CR(_ix_cc_ipv6_icmp_build_err_message
                             (IX_CC_ICMPV6_TYPE_PACKET_TOO_BIG,
                             IX_CC_ICMPV6_CODE_ZERO, arg_Reason, 
                             1, arg_hBuffer));
   
#if defined(IPV6_DEBUG)
    ix_ossl_message_log("Exiting ICMP Packet Too Big.\n");
#endif /* IPV6_DEBUG */
    return IX_SUCCESS;
}


/*
 * This function prints the header and the contents of the ICMP payload 
 * on the screen
 */
void _ix_cc_ipv6_icmp_print_message(
            const ix_uint8 *arg_pMessage, 
            ix_uint32 arg_size )
{
    ix_cc_ipv6_icmp_header icmpHdr;
    register ix_uint32 i=0;

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    if (arg_pMessage == NULL)
    {
        IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                        ("Input parameter(arg_pMessage) is null"));
        return;
    }
    if (arg_size >= IX_CC_ICMPV6_HDR_LEN)
    {
        IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                        ("Input parameter(arg_size) is invalid"));
        return;
    }
#else
    assert( arg_pMessage );
    assert( arg_size >= IX_CC_ICMPV6_HDR_LEN );
#endif
    
    ix_ossl_memcpy(&icmpHdr, arg_pMessage, IX_CC_ICMPV6_HDR_LEN);
    
#if defined(IPV6_DEBUG)
    ix_ossl_message_log("type: %d , code: %d, csum: %d ", 
                            icmpHdr.type, icmpHdr.code, icmpHdr.checksum);
    
    ix_ossl_message_log("contents:\n" );
#endif /* IPV6_DEBUG */

    for (i = 0;i < arg_size; i++)
    {
#if defined(IPV6_DEBUG)
        ix_ossl_message_log("%02x ", arg_pMessage[i]);
#endif /* IPV6_DEBUG */

        if ((i+1) % 4 == 0)
        {
#if defined(IPV6_DEBUG)
            ix_ossl_message_log("\n");
#endif /* IPV6_DEBUG */
        }
    }

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("\n");
#endif /* IPV6_DEBUG */
}

/**
*This function builds a complete ICMP error message and send it to the IPv6
*forwarder to perform lookup and forward the packet.
*Currently, we do not support rate limiting, hence 'arg_SendMsg' 
*is ignored. Else it would instruct the function whether to add the packet
*to the queue or send it right away.
**/
ix_error _ix_cc_ipv6_icmp_build_err_message(
                ix_uint8 arg_Type,
                ix_uint8 arg_Code,
                ix_uint32 arg_Param,
                ix_uint32 arg_SendMsg,
                ix_buffer_handle arg_hBuffer
                )
{
    ix_error errVal;
    ix_uint8 icmpMessage[IX_CC_ICMPV6_PKT_MTU];
    ix_cc_ipv6_header errPktIpHdr;/*IP hdr of the packet causing the error*/
    ix_cc_ipv6_header *pIpHdr; /*IP hdr of the new ICMP pkt*/
    ix_cc_ipv6_icmp_header *pIcmpHdr;
    ix_buffer_handle  prevBufHdl = 0, newBufHdl = 0;
    ix_hw_buffer_meta *pHwMeta;
    void *pData;
    ix_uint16 inPort;               /*The port through which the pkt arrived*/
    ix_uint128 ipv6Addr[PROP_MAX_IPV6_ADDRESSES];
    ix_uint128 tempIpv6Addr;
    ix_uint8  numIPAddrs;
    ix_uint16 pktLen, dataLen, prevDataLen;
    ix_uint8  *pCurPosn;
    ix_uint32  isSOP;
    ix_uint32  numBytesCopied = 0;

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("Entered Build Err Message\n");
#endif /* IPV6_DEBUG */

    /* First, set the pointers to the correct location */
    pIpHdr = (ix_cc_ipv6_header *)icmpMessage;
    pIcmpHdr = (ix_cc_ipv6_icmp_header *)(icmpMessage + IX_CC_IPV6_HDR_LEN);

    /*
     * Get data from the error packet, this involves getting the meta data,
     * and the IPv6 header of the error packet 
     */
    IX_IPV6_ERROR_CGT(ix_rm_buffer_get_meta( arg_hBuffer, (void **)&pHwMeta), 
                 errVal, LBL_ERR);
    /* Get the incoming port from the meta data. This will determine the 
     * source address of the outgoing ICMP packet
     */
    inPort = IX_NTOH16(pHwMeta->m_InputPort);

    /* Get the packet data. This will give us the IPv6 header of the error
     * packet
     */
    IX_IPV6_ERROR_CGT(ix_cc_hw_get_packet_data (arg_hBuffer, &pData),
                 errVal, LBL_ERR);

    /* Copy the IPv6 header.*/
    _ix_cc_ipv6_init_ip_header( (ix_uint8 *) pData, &errPktIpHdr);

    /* 
     * Now build the complete ICMP packet. This is mainly used for calculating
     * ICMP checksum. Also, in case there isn't enough space in the 
     * current hardware buffers, this ICMP message is used to create a 
     * completely new hardware buffer chain
     */

    /* 
     * IPV6 Header Initialization:
     * This is the outgoing IPv6 packet header, this will contain the 
     * ICMPv6 packet and the original "error" packet in its payload
     */
    pIpHdr->version=6;
    pIpHdr->traffic_class = 0;
    pIpHdr->flow_label = 0;

    /* Calculate the size of the IPv6 payload. If the error packet is 
     * bigger than the max allowed ICMP payload, set the payload length
     * to (ICMP header length + Reason field length + max ICMP payload length).
     * Else, if error pkt is small enough, set IPv6 payload length to
     * (ICMP header length + reason field length + err pkt length)
     */

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("PacketSize (0x%x), BufferSize (0x%x), OffSet (0x%x\n",
                                          IX_NTOH16(pHwMeta->m_PacketSize),
                                          IX_NTOH16(pHwMeta->m_BufferSize),
                                          IX_NTOH16(pHwMeta->m_Offset));
#endif /* IPV6_DEBUG */

    if(IX_NTOH16(pHwMeta->m_PacketSize) >= IX_CC_ICMPV6_ERR_PKT_PAYLOAD_LIMIT)
    {
        pIpHdr->payloadLength = IX_HTON16(IX_CC_ICMPV6_HDR_LEN + 
                                          IX_CC_ICMPV6_PAYLOAD_LIMIT); 
        pktLen = IX_CC_ICMPV6_ERR_PKT_PAYLOAD_LIMIT;
    }
    else
    {
        pIpHdr->payloadLength = IX_HTON16(IX_CC_ICMPV6_HDR_LEN + 
                                        IX_NTOH16(pHwMeta->m_PacketSize) + 4);
        pktLen = IX_NTOH16(pHwMeta->m_PacketSize);
    }

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("PayLoadLength (%d), pktLen (%d) \n",
                                           pIpHdr->payloadLength, pktLen);
#endif /* IPV6_DEBUG */

    pIpHdr->nextHeader = IX_CC_IPV6_NEXT_HDR_ICMP;
    pIpHdr->hopLimit = IX_CC_IPV6_HOP_LIMIT_MAX;
    
    /*
     * Now we need the IP address of the incoming port to use as the source
     * address
     */
    IX_IPV6_ERROR_CGT(_ix_cc_ipv6_get_port_ip_address(
                                            g_Ipv6RegistryData.localBladeId,
                                            inPort, &numIPAddrs, ipv6Addr), 
                 errVal, LBL_ERR);
    
    /*
     * Convert the address to network byte order and copy it to the source
     * address field of the IPv6 header
     */
#if 0
    /* This is a hack for now. Until we add support for IPv6 
     * addresses in the system app, we avoid checking for 0 addresses
     * because that is a possiblity! 
     */
    if((numIPAddrs == 0) || (numIPAddrs > PROP_MAX_IPV6_ADDRESSES))
#else
    if(numIPAddrs > PROP_MAX_IPV6_ADDRESSES)
#endif
    {
#if defined(IPV6_DEBUG)
        ix_ossl_message_log("ICMPv6: Invalid number of IP addresses returned:"
                            " %d\n", numIPAddrs);
#endif /* IPV6_DEBUG */

        goto LBL_ERR;
    }

    /* The above call returns an array of IPv6 addresses (each port can 
     * have multiple IPv6 addresses. In this case, simply use the first 
     * IP address (as the source) from the list returned
     */
    IX_HTON128(ipv6Addr, &tempIpv6Addr);

    ix_ossl_memcpy(&pIpHdr->srcAddress, &tempIpv6Addr, 
		           IX_CC_IPV6_ADDRESS_LENGTH_BYTES);

    /* The destination IP address comes from the original packets source
     * IP address
     */
    ix_ossl_memcpy(&pIpHdr->dstAddress, &errPktIpHdr.srcAddress, 
			              IX_CC_IPV6_ADDRESS_LENGTH_BYTES);

    /* ICMPv6 Header Initialization:
     * Initialize the ICMPv6 header as much as possible.
     */
    pIcmpHdr->type = arg_Type;
    pIcmpHdr->code = arg_Code;
    pIcmpHdr->checksum = 0;

    /* ICMP Payload Creation */
     
    /* 
     * Copy the param into the appropriate field just after the ICMP header 
     */
    pCurPosn = icmpMessage+IX_CC_ICMPV6_HDR_LEN+IX_CC_IPV6_HDR_LEN;

    ix_ossl_memcpy(pCurPosn, &arg_Param, sizeof(ix_uint32));

    *(ix_uint16 *)pCurPosn = IX_HTON32(*pCurPosn);
    pCurPosn+=sizeof(ix_uint32);

    /* Copy the payload into the packet. This might involve copying packet 
     * data from multiple hardware buffers in case the packet overflows one
     * or more buffers
     */

     /* Get the available data within the first buffer */
     dataLen = IX_NTOH16(pHwMeta->m_BufferSize);

    /* 
     * If the packet fits within the buffer, set the data length to the
     * packet size
     */

    if(dataLen > pktLen)
    {
        dataLen = pktLen;
    }

    prevDataLen = dataLen;

    /* 
     * New buffer handle, starts out with the handle to the original
     * buffer 
     */
    newBufHdl = arg_hBuffer;
    
    isSOP = TRUE;
    
    while(1)
    {
	ix_ossl_memcpy(pCurPosn, (ix_uint8 *)pData, dataLen);

        pCurPosn+=dataLen;

        pktLen-=dataLen;

        numBytesCopied += dataLen;

        if(pktLen == 0)
        {
            break;  /*The entire packet has been copied.*/
        }
        
        /* Now get the next buffer */
        prevBufHdl = newBufHdl;

        IX_IPV6_ERROR_CGT(ix_rm_buffer_get_next( prevBufHdl, &newBufHdl), 
                 errVal, LBL_ERR);

        if (newBufHdl == 0)
        {
            /* End of buffer chain, nothing more to copy */
#if defined(IPV6_DEBUG)
            ix_ossl_message_log("detected end of buffer chain, breaking.. \n");
#endif /* IPV6_DEBUG */
            break;
        }
        
        IX_IPV6_ERROR_CGT(ix_rm_buffer_get_meta( newBufHdl, (void **)&pHwMeta), 
                 errVal, LBL_ERR);

        IX_IPV6_ERROR_CGT(ix_cc_hw_get_packet_data(newBufHdl, &pData),
                 errVal, LBL_ERR);
        
        prevDataLen = dataLen;
        dataLen = IX_NTOH16(pHwMeta->m_BufferSize);

        /* Are we crossing out PAYLOAD_LIMIT ? */
        if(dataLen > pktLen)
        {
            dataLen = pktLen;
        }

        isSOP = FALSE;

    } /*while(1)*/
    
    pIpHdr->payloadLength = IX_HTON16(IX_CC_ICMPV6_HDR_LEN + numBytesCopied + 4);
    /*Now calculate the checksum on the entire packet*/
    _ix_cc_ipv6_icmp_calculate_checksum(icmpMessage);

    /* 
     * Build a new buffer chain (newBufHdl) and free the old buffer
     * chain (arg_hBuffer).
     */

    IX_IPV6_ERROR_CGT(_ix_cc_ipv6_icmp_build_buffer_chain(icmpMessage, 
                                                          inPort,
                                                          &newBufHdl), 
                              errVal, LBL_ERR);

    /*
     * Successfully constructed a new buffer chain and all the data
     * has been copied. Send it to the IPv6 forwarder to ship it out the
     * door.
     */


    IX_IPV6_ERROR_CGT(_ix_cc_ipv6_icmp_queue_message(newBufHdl), 
                                                errVal, LBL_FREE_NEW_BUF);

    /* 
     * Free the original buffer.
     */
    ix_rm_buffer_free_chain(arg_hBuffer);

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("Exiting Build Err message function.\n");
#endif /* IPV6_DEBUG */

    return IX_SUCCESS;   

LBL_FREE_NEW_BUF:
    ix_rm_buffer_free_chain(newBufHdl);

LBL_ERR:
    ix_rm_buffer_free_chain(arg_hBuffer);

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("Exiting Build Err message function. Err Code: %ld.\n",
                        (ix_uint32)errVal);
#endif /* IPV6_DEBUG */

    return errVal;
}

/*
 * This function builds and sends informational ICMP messages
 */
ix_error _ix_cc_ipv6_icmp_build_info_message(
                ix_ipv6_addr    arg_SrcAddr,
                ix_ipv6_addr    arg_DstAddr,
                ix_uint8        arg_Type,
                ix_uint8        arg_Code,
                ix_uint32       arg_Length,
                ix_uint8        *arg_Buffer,
                ix_uint16       arg_InputPort)
{
    ix_buffer_handle hBuffer = IX_NULL_BUFFER_HANDLE;
    ix_error errVal;
    
    IX_ERROR_CR(_ix_cc_ipv6_icmp_build_info_message_ex(arg_SrcAddr,
                                                            arg_DstAddr,
                                                            arg_Type,
                                                            arg_Code,
                                                            arg_Length,
                                                            arg_Buffer,
                                                            arg_InputPort,
                                                            &hBuffer));
    
    /* Finally, send the packet */
    if(hBuffer != IX_NULL_BUFFER_HANDLE)
    {
        errVal = _ix_cc_ipv6_pass_packet(hBuffer);
    }
    else
    {
#if defined(IPV6_DEBUG)
        ix_ossl_message_log ("_ix_cc_ipv6_icmp_build_info_message: "
                "build_buffer_chain returned IX_NULL_BUFFER_HANDLE \n");
#endif /* IPV6_DEBUG */
    }

    return errVal;
}

/* This function simply builds an informational ICMP message. It doesn't 
 * transmit the message. Instead, it returns the buffer pointer for the
 * message, and it is the caller's responsibility to transmit the message*/
ix_error _ix_cc_ipv6_icmp_build_info_message_ex(
                ix_ipv6_addr    arg_SrcAddr,
                ix_ipv6_addr    arg_DstAddr,
                ix_uint8        arg_Type,
                ix_uint8        arg_Code,
                ix_uint32       arg_Length,
                ix_uint8        *arg_packet,
                ix_uint16       arg_InputPort,
                ix_buffer_handle *arg_HwBuffer)
{
    ix_error errVal;
    ix_cc_ipv6_header *pIpHdr; 
    ix_cc_ipv6_icmp_header *pIcmpHdr;
    ix_uint8 icmpMessage[IX_CC_ICMPV6_PKT_MTU], *pCurLocation;

    if(!arg_packet || (arg_Length > IX_CC_ICMPV6_PAYLOAD_LIMIT))
    {
        return IX_CC_ERROR_INTERNAL;
    }
 
    /* 
     * First, set the pointers to the correct location
     */
    pIpHdr = (ix_cc_ipv6_header *)icmpMessage;
    pIcmpHdr = (ix_cc_ipv6_icmp_header *)(icmpMessage + IX_CC_IPV6_HDR_LEN);
    
    /*
     * Initialize the IPv6 header at this point. We can't fully initialize it
     * just yet.
     */
    pIpHdr->version= 6;
    pIpHdr->traffic_class = 0;
    pIpHdr->flow_label = 0;

    pIpHdr->payloadLength = IX_HTON16(IX_CC_ICMPV6_HDR_LEN + arg_Length);
    pIpHdr->nextHeader = IX_CC_IPV6_NEXT_HDR_ICMP;
    pIpHdr->hopLimit = IX_CC_IPV6_HOP_LIMIT_MAX;

    /* Start building the IPv6 header */
    IX_HTON128(&arg_SrcAddr, &pIpHdr->srcAddress);
    IX_HTON128(&arg_DstAddr, &pIpHdr->dstAddress);

    /* Initialize the ICMP header */
    pIcmpHdr->type = arg_Type;
    pIcmpHdr->code = arg_Code;
    pIcmpHdr->checksum = 0;

    pCurLocation = icmpMessage + IX_CC_IPV6_HDR_LEN + IX_CC_ICMPV6_HDR_LEN;

    ix_ossl_memcpy(pCurLocation, arg_packet, arg_Length);

    /* Now calculate the checksum on the entire packet */
    _ix_cc_ipv6_icmp_calculate_checksum(icmpMessage);

    errVal = _ix_cc_ipv6_icmp_validate_checksum((ix_uint8 *)pIpHdr, 
                                              IX_NTOH16(pIpHdr->payloadLength),
                                              (ix_uint8 *)pIcmpHdr);
    if(errVal != IX_SUCCESS)
    {
        goto LBL_ERR;
    }

    /* 
     * Allocate a buffer chain to carry this payload into the Network.
     */
    IX_IPV6_ERROR_CGT(_ix_cc_ipv6_icmp_build_buffer_chain(icmpMessage, 
                                                          arg_InputPort,
                                                          arg_HwBuffer),
                      errVal, LBL_ERR);


    return IX_SUCCESS;

LBL_ERR:

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("ICMP6: Build info message failed\n");
#endif /* IPV6_DEBUG */

    if(*arg_HwBuffer != IX_NULL_BUFFER_HANDLE)
    {
        ix_rm_buffer_free_chain(*arg_HwBuffer);
        *arg_HwBuffer = IX_NULL_BUFFER_HANDLE;
    }

    return errVal;
}

/**
 * This function calculates the checksum for the ICMP message as
 * defined in RFC2460 and RFC2463
 **/
ix_uint32 _ix_cc_ipv6_icmp_calculate_checksum(ix_uint8 *icmpMessage)
{
    ix_cc_ipv6_header *pIpHdr;
    ix_cc_ipv6_icmp_header *pIcmpHdr;
    ix_cc_ipv6_pseudo_header pseudoHdr;
    ix_uint16 length;
    ix_uint16 isExtraOctet = 0;
    ix_uint32 checksum = 0;
    ix_uint16 *pCurPosn;
    ix_uint16 count;

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("Entering %s\n", __FUNCTION__);
#endif /* IPV6_DEBUG */

    pIpHdr = (ix_cc_ipv6_header *)icmpMessage; 
    pIcmpHdr = (ix_cc_ipv6_icmp_header *)(icmpMessage + IX_CC_IPV6_HDR_LEN);    

    /* First, set the checksum field in the ICMP header to zero */
    pIcmpHdr->checksum=0;

    /* 
     * Prepare the pseudo header. This header contains the source address
     * destination address, the length of the "upper layer" payload, a 
     * zero field and a next header. See RFC 2460, Sec 8.1 for details.
     */
    ix_ossl_memcpy(&pseudoHdr.srcAddress, &pIpHdr->srcAddress, 
            IX_CC_IPV6_ADDRESS_LENGTH_BYTES);

    ix_ossl_memcpy(&pseudoHdr.dstAddress, &pIpHdr->dstAddress, 
            IX_CC_IPV6_ADDRESS_LENGTH_BYTES);
    
    pseudoHdr.pktLength = (ix_uint32) pIpHdr->payloadLength; 
    pseudoHdr.unused = 0;
    pseudoHdr.nextHdr = IX_CC_IPV6_NEXT_HDR_ICMP;

    /* Calculate checksum for the pseudo header */
    length = sizeof(ix_cc_ipv6_pseudo_header)/2;  /* number of 16 bit words */
    pCurPosn = (ix_uint16 *) &pseudoHdr;
    while(length)
    {
        checksum += (ix_uint16)*pCurPosn;
        pCurPosn++;
        length--;
    }
    
    /* Calculate the checksum for the rest of the ICMPv6 packet */
    pCurPosn = (ix_uint16 *)(icmpMessage + IX_CC_IPV6_HDR_LEN);

    /* 
     * Check if the packet does not end at a 16-bit boundary. This will 
     * be used later to add any remaining octets to the sum
     */

    length = IX_NTOH16(pIpHdr->payloadLength);
    isExtraOctet = length%2;

    /*
     * The length stored in the 'length' variable is in octets. Convert
     * it into the # of 16 bit words
     */
    length=length/2;

    for(count = 0; count < length; count++, pCurPosn++)
    {
        checksum+=(ix_uint16)*pCurPosn;
    }
    
    /* Now add the remaining octet, if any, to the sum */
    if(isExtraOctet)
    {
        checksum+= *(ix_uint8 *)pCurPosn;
    }

    /*
     * Now add the upper 16 bits to the lower 16 bits until we end up with a 
     * 16 bit value. This is the 1's complement sum
     */
    while(checksum>>16)
    {
        checksum = (checksum>>16) + (checksum & 0xffff);
    }
    
    /* 
     * Finally, take the 1's complement of the calculated 1's complement sum
     * and store it in the checksum field of the ICMP header
     */
    pIcmpHdr->checksum = (ix_uint16)~checksum;

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("Checksum 0x%x (%d) calculated successfully.\n", 
                                  pIcmpHdr->checksum, pIcmpHdr->checksum);
#endif /* IPV6_DEBUG */

    return IX_SUCCESS;
}

/**
 * TYPENAME: _ix_cc_ipv6_icmp_validate_checksum
 * 
 * DESCRIPTION:
 *
 *      arg_pIpv6Pkt   - pointer to the IPv6 packet.
 *      arg_icmpLength - the length of the entire ICMP packet (including
 *                       the ICMP header).
 *      arg_pIcmpHdr   - Pointer to the ICMP packet with in the IPv6 pkt.
 */
ix_error _ix_cc_ipv6_icmp_validate_checksum(ix_uint8    *arg_pIpv6Pkt,
                                            ix_uint16   arg_icmpLength,
                                            ix_uint8    *arg_pIcmpHdr)
{
    ix_uint16   *pCurPosn;
    ix_cc_ipv6_pseudo_header pseudoHdr;
    ix_cc_ipv6_header        *pIpv6Hdr;
    ix_uint16   length, isExtraOctet;
    ix_uint32   sum = 0;

    if(!arg_pIpv6Pkt || !arg_pIcmpHdr)
        return IX_CC_ERROR_NULL;

    pIpv6Hdr = (ix_cc_ipv6_header *)arg_pIpv6Pkt;

    if( (arg_pIcmpHdr < arg_pIpv6Pkt) ||
        (arg_pIcmpHdr > 
            (arg_pIpv6Pkt + IX_CC_IPV6_HDR_LEN + pIpv6Hdr->payloadLength )))
    {
        /*the Icmp header pointer points to a wrong location*/
        return IX_CC_ERROR_INTERNAL;
    }
    ix_ossl_memcpy(&pseudoHdr.srcAddress, &pIpv6Hdr->srcAddress, 
                    IX_CC_IPV6_ADDRESS_LENGTH_BYTES);
    ix_ossl_memcpy(&pseudoHdr.dstAddress, &pIpv6Hdr->dstAddress,
                    IX_CC_IPV6_ADDRESS_LENGTH_BYTES);

    pseudoHdr.pktLength = pIpv6Hdr->payloadLength;
    pseudoHdr.unused = 0;
    pseudoHdr.nextHdr = IX_CC_IPV6_NEXT_HDR_ICMP; 
    
    /*calculate sum of the psuedoheader*/
    length = sizeof(ix_cc_ipv6_pseudo_header) / 2;/*# of 16 bit chunks*/
    pCurPosn = (ix_uint16 *) &pseudoHdr;
    while(length)
    {
        sum += (ix_uint16)*pCurPosn;
        pCurPosn++;
        length--;
    }
    
    /*now calculate sum of the icmp header*/
    isExtraOctet = arg_icmpLength%2;
    length = arg_icmpLength/2;
    
    pCurPosn = (ix_uint16 *) arg_pIcmpHdr;
    while(length)
    {
        sum += (ix_uint16) *pCurPosn;
        pCurPosn++;
        length--;
    }

    if(isExtraOctet)
        sum += *(ix_uint8 *)pCurPosn;
    
    /*Now add the upper 16 bits to the lower 16 bits until we end up with a 
     *16 bit value. */
    while(sum>>16)
    {
        sum = (sum>>16) + (sum & 0xffff);
    }
 
    /* finally, if the lower 16 bits of sum are equal to ffff, the checksum
     * is valid*/
    if( (sum ^ 0xffff) != 0)
    {
#if defined(IPV6_DEBUG)
        ix_ossl_message_log("!! Checksum Calculation Failed !!\n");
#endif /* IPV6_DEBUG */

        return IX_CC_ERROR_INTERNAL; 
    }

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("ICMP Checksum Verified !! \n");
#endif /* IPV6_DEBUG */

    return IX_SUCCESS;
}

/*
 * Validates the packet causing the error.
 */
ix_error _ix_cc_ipv6_icmp_ok_to_send(ix_uint16 msgType,
                                      ix_buffer_handle arg_hBuffer,
                                      ix_uint32 *arg_pOkToSend)
{
    ix_uint8 *inPacket;
   
    ix_cc_ipv6_header *pInIpHdr;
    void *pData = 0;
        
    *arg_pOkToSend = 0;
    
    IX_ERROR_CRT(ix_cc_hw_get_packet_data (arg_hBuffer, &pData),
                IX_CC_ERROR_INTERNAL, 0);
   
    inPacket = (ix_uint8 *)pData;
    pInIpHdr = (ix_cc_ipv6_header *)inPacket;
    
    if(_ix_cc_ipv6_icmp_is_icmp_err_pkt(inPacket) == TRUE)
    {
#if defined(IPV6_DEBUG)
        ix_ossl_message_log("Trying to send ICMP message for an ICMP error\n");
#endif /* IPV6_DEBUG */

        *arg_pOkToSend = 0;
        return IX_SUCCESS;
    }

    if (IS_IPV6_ADDRESS_MULTICAST(pInIpHdr->dstAddress) && 
        ((IX_NTOH16(msgType) != IX_CC_ICMPV6_TYPE_PARAMETER_PROBLEM) || 
        (IX_NTOH16(msgType) != IX_CC_ICMPV6_TYPE_PACKET_TOO_BIG)))
    {
#if defined(IPV6_DEBUG)
        ix_ossl_message_log("Trying to send ICMP message for IPV6 multicast\n");
#endif /* IPV6_DEBUG */

        *arg_pOkToSend = 0;   
        return IX_SUCCESS;
    }
    
    *arg_pOkToSend = 1;   
    return IX_SUCCESS;    
}

/*
 * This function checks the IPv6 packet and verifies whether the packet is
 * an ICMP error message. If it is indeed an ICMP error message, this 
 * function returns TRUE, else it returns FALSE
 */
ix_uint8 _ix_cc_ipv6_icmp_is_icmp_err_pkt(ix_uint8  *arg_pIpv6Pkt)
{
    ix_cc_ipv6_header   *pIpv6Header;
    ix_cc_ipv6_icmp_header *pIcmpHeader;
    ix_cc_ipv6_options_hdr_common *pIpOptHdr;
    ix_uint8            *pCurPosn;
    ix_uint8            nextHeader;
    ix_uint32           curLength;
    ix_uint8            found = FALSE;

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("Entering %s\n", __FUNCTION__);
#endif /* IPV6_DEBUG */
    
    pCurPosn = arg_pIpv6Pkt;
    pIpv6Header = (ix_cc_ipv6_header *) pCurPosn;
    nextHeader = pIpv6Header->nextHeader;
    
    pCurPosn += IX_CC_IPV6_HDR_LEN;

    while( (nextHeader != IX_CC_IPV6_NEXT_HDR_NONE) && !found)
    {
        switch(nextHeader)
        {
            case IX_CC_IPV6_NEXT_HDR_HOPBYHOP: /*fall thru*/
            case IX_CC_IPV6_NEXT_HDR_ROUTING: /*fall thru*/
            case IX_CC_IPV6_NEXT_HDR_FRAGMENT: /*fall thru*/
            case IX_CC_IPV6_NEXT_HDR_DSTOPT:
                pIpOptHdr = (ix_cc_ipv6_options_hdr_common *) pCurPosn;
                nextHeader = pIpOptHdr->nextHeader;
                
                /*Length of the header is in 'num of 8 octets', excluding
                 *the first 8 octets. So add 1 to length, before multiplying
                 *by 8*/
                curLength = (1+IX_NTOH16(pIpOptHdr->hdrLength))*8;
                pCurPosn += curLength;
                break;
            case IX_CC_IPV6_NEXT_HDR_ICMP:
                found = TRUE;
                break;
            default:
                /*don't care what other upper layer protocol headers there are.
                 * This isn't an ICMP packet, so we shouldn't process any 
                 * further*/
                nextHeader = IX_CC_IPV6_NEXT_HDR_NONE;
#if defined(IPV6_DEBUG)
                ix_ossl_message_log("ICMPv6: Unknown/UpperLayer Next hdr detected\n");
#endif /* IPV6_DEBUG */
                break;
        }/*switch*/        
    }/*while*/
 
    if(found)
    {
       pIcmpHeader = (ix_cc_ipv6_icmp_header *) pCurPosn;
       /* If the type value falls in the range of informational messages,
        * this message is valid*/
       if(IX_NTOH16(pIcmpHeader->type) >= 128)
           found = FALSE;
    }
    return found;
}/*_ix_cc_ipv6_icmp_is_icmp_err_pkt*/

/*
 * Allocates buffers and populates them with the IPv6 packet
 */
ix_error _ix_cc_ipv6_icmp_build_buffer_chain(ix_uint8 *arg_pIpv6Pkt, 
                                             ix_uint16 arg_inputPort,
                                             ix_buffer_handle *arg_hBuffer)
{
    ix_error    errVal;
    ix_cc_ipv6_header *pIpHdr; 
    ix_buffer_handle hFirstBuffer = IX_NULL_BUFFER_HANDLE;
    ix_buffer_handle hCurrBuffer = IX_NULL_BUFFER_HANDLE;
    ix_buffer_handle hPrevBuffer = IX_NULL_BUFFER_HANDLE;
    ix_hw_buffer_meta *pHwMeta;
    void        *pData; 
    ix_uint8    *pCurPosn;
    ix_uint16    totalLen, pktLen, curLen;
    ix_uint32    firstBuffer = 1;
    ix_uint32   cell_count = 0;
    ix_uint16   dataOffset = 0;

    /* used to suppress compile warnings */
    ix_uint16 temp;

    pIpHdr = (ix_cc_ipv6_header *) arg_pIpv6Pkt;

    totalLen = pktLen = IX_NTOH16(pIpHdr->payloadLength) + IX_CC_IPV6_HDR_LEN;
    pCurPosn = (ix_uint8 *)pIpHdr;

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("Building chain: totalLen (0x%x), pktLen (0x%x)\n", 
                                                            totalLen, pktLen);
#endif /* IPV6_DEBUG */

    while(pktLen)
    {
        IX_IPV6_ERROR_CGT(ix_rm_buffer_alloc(g_pIpv6Context->hBufferFreeList, 
                                             &hCurrBuffer),
                                             errVal, LBL_ERR);

        if(firstBuffer)
        {
            firstBuffer = 0;
            hFirstBuffer = hPrevBuffer = hCurrBuffer;
            dataOffset = IX_CC_SOP_HW_BUFFER_PREPEND + 14;

            if(pktLen > (IX_CC_ICMP_HW_BUFFER_SIZE-dataOffset))
            {
                curLen = IX_CC_ICMP_HW_BUFFER_SIZE-dataOffset;
            }
            else
            {
                curLen = pktLen;
            }

            cell_count = ix_cc_calculate_cell_count(curLen, 1);

#if defined(IPV6_DEBUG)
            ix_ossl_message_log("First Buffer: dataOffset (0x%lx), hCurrBuffer (0x%lx)\n", (ix_uint32)dataOffset, (ix_uint32)hCurrBuffer);
#endif /* IPV6_DEBUG */

            /* Write the cell count into the First buffer handle */
            errVal = ix_rm_buffer_cell_count_set(&hFirstBuffer, cell_count);
            if (errVal != IX_SUCCESS)
            {
#if defined(IPV6_DEBUG)
                ix_ossl_message_log("First Buffer: failed cell_count_set \n");
#endif /* IPV6_DEBUG */

                goto LBL_ERR;
            }
        }
        else
        {
            /* This is not the first buffer in the chain */
            dataOffset = 0;
            if(pktLen > IX_CC_ICMP_HW_BUFFER_SIZE)
            {
                curLen = IX_CC_ICMP_HW_BUFFER_SIZE;
            }
            else
            {
                curLen = pktLen;
            }

            cell_count = ix_cc_calculate_cell_count(curLen, 0);

            /* Write the cell count into the buffer handle */
            errVal = ix_rm_buffer_cell_count_set(&hCurrBuffer, cell_count);
            if (errVal != IX_SUCCESS)
            {
#if defined(IPV6_DEBUG)
                ix_ossl_message_log("Current Buffer: failed cell_count_set \n");
#endif /* IPV6_DEBUG */

                goto LBL_ERR;
            }

#if defined(IPV6_DEBUG)
            ix_ossl_message_log("Other Buffer: dataOffset (0x%lx), hCurrBuffer (0x%lx\n", (ix_uint32)dataOffset, (ix_uint32)hCurrBuffer);
#endif /* IPV6_DEBUG */
            
            /*link this buffer into the buffer chain*/
            IX_IPV6_ERROR_CGT(ix_rm_buffer_link(&hPrevBuffer, &hCurrBuffer),
                              errVal, LBL_ERR);
            
            hPrevBuffer = hCurrBuffer;    
        }

       /*
        * Update the meta data associated with the packet
        */
       IX_IPV6_ERROR_CGT(ix_rm_buffer_get_meta(hCurrBuffer, (void **)&pHwMeta),
                         errVal, LBL_ERR);

       pHwMeta->m_HwNext = 0xFF;      /* Set next buffer to NULL */
       pHwMeta->m_BufferInfo |= 0x2;  /* Set Header Type to IPv6 */
 
       /* Set the input port to what was provided */
       temp = IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_InputPort, IX_HTON16(arg_inputPort));
       temp = IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_PacketSize, 
                                              IX_HTON16((ix_uint16)totalLen));
       temp = IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_BufferSize, 
                                              IX_HTON16(curLen));
       temp = IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_Offset, 
                                              IX_HTON16(dataOffset));

       /* Copy the data into the buffer */
       IX_IPV6_ERROR_CGT(ix_cc_hw_get_packet_data(hCurrBuffer, &pData),
                         errVal, LBL_ERR);

       ix_ossl_memcpy(pData, pCurPosn, curLen);

       /* Next Loop.. */
       pktLen -= curLen;
       pCurPosn += curLen;
       
    } /*while*/

    *arg_hBuffer = hFirstBuffer;
    return IX_SUCCESS;

LBL_ERR:
#if defined(IPV6_DEBUG)
    ix_ossl_message_log("ICMP6: Build Buffer Chain failed\n");
#endif /* IPV6_DEBUG */

    /* 
     * This check has been put for a special case where it is possible that 
     * hCurrBuffer and hFirstBuffer may be same. In that case, the buffer free should 
     * happen only once and not twice. 
     */
    if(hCurrBuffer == hFirstBuffer)
    {
       if(hCurrBuffer != IX_NULL_BUFFER_HANDLE)
       {
           ix_rm_buffer_free_chain(hCurrBuffer);
           hCurrBuffer = IX_NULL_BUFFER_HANDLE;
           hFirstBuffer = IX_NULL_BUFFER_HANDLE;
       }
    }
    else
    {
       if(hCurrBuffer != IX_NULL_BUFFER_HANDLE)
       {
           ix_rm_buffer_free_chain(hCurrBuffer);
           hCurrBuffer = IX_NULL_BUFFER_HANDLE;
       }
       if(hFirstBuffer != IX_NULL_BUFFER_HANDLE)
       {
           ix_rm_buffer_free_chain(hFirstBuffer);
           hFirstBuffer = IX_NULL_BUFFER_HANDLE;
       }
    }
    *arg_hBuffer = IX_NULL_BUFFER_HANDLE;

    return errVal;
}

static
ix_error _ix_cc_ipv6_icmp_queue_message(ix_buffer_handle arg_hBuffer)
{
    ix_uint16   count;
    ix_buffer_handle hBuf;

    ICMP_LOCK();

    /* count is the size of the queue */
    count = ListGetCount(g_IcmpMsgQ);

    /* If the queue is full, remove the first element */
    while(count > IX_CC_ICMPV6_ICMPQ_DEFAULT_SIZE)
    {
        hBuf = (ix_buffer_handle)ListGetFirst(g_IcmpMsgQ);
        if(hBuf)
        {
            ListRemoveItem(g_IcmpMsgQ, (void *)hBuf);
            ix_rm_buffer_free_chain(hBuf);
            hBuf = IX_NULL_BUFFER_HANDLE;
        }
        count = ListGetCount(g_IcmpMsgQ);
    }

    /* Add the newest buffer to the queue */
    if(ListAddItem(g_IcmpMsgQ, (void *)arg_hBuffer) == 0)
    {
        ICMP_UNLOCK();
        return IX_CC_ERROR_OOM;
    }

    /* 
     * Addig the unlock here for the lock which was taken at the start of 
     * this function.
     */
    
    ICMP_UNLOCK();


#if defined(IPV6_DEBUG)
            ix_ossl_message_log("ICMPv6 message queue'd \n");
#endif /* IPV6_DEBUG */

    return IX_SUCCESS;
}

/*
 * This function has been rewritten and tested for LINUX to send all 
 * packets in single shot. Old function which send 1 pkt in a single expiry 
 * is in the else part. 
 * */

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
ix_error _ix_cc_ipv6_icmp_send_err_message(void *arg_pContext)
{
    ix_error    errVal = IX_SUCCESS;
    ix_buffer_handle hBuf;
    ix_uint16   count, i;

    /*First schedule the next event, then do the rest of the processing*/
#if !defined(IX_EXCLUDE_CCI)
    errVal = ix_cci_cc_add_event_handler(g_pIpv6Context->hIpv6Handle,
                                            g_Icmpv6SleepTime,
                                            _ix_cc_ipv6_icmp_send_err_message,
                                            IX_EVENT_TYPE_ONESHOT,
                                            1, &(g_IcmpQEventHdl));
    if(errVal != IX_SUCCESS)
        return IX_CC_IPV6_ERROR_CCI;
#elif defined(MEAT_TEST)
    ListAddItem(g_Context_Icmp, _ix_cc_ipv6_icmp_send_err_message);
#endif


    /*Now send the first packet in the queue*/
    ICMP_LOCK(); 
    if(!g_IcmpMsgQ)
    {
        ICMP_UNLOCK();
        return IX_CC_ERROR_NULL;
    }

    count = ListGetCount(g_IcmpMsgQ);
    if(count == 0)
        goto LBL_EXIT;

    for(i=0; i<g_Icmpv6QueueDepth; i++)
    {
       hBuf = (ix_buffer_handle)ListGetFirst(g_IcmpMsgQ);
       if(!hBuf)
       {
           errVal = IX_NULL_BUFFER_HANDLE;
           goto LBL_EXIT;
       }    

       errVal = _ix_cc_ipv6_pass_packet(hBuf);

       /* Free the buffer chain if this function returns error */
       if(errVal != IX_SUCCESS)
       {
           ix_rm_buffer_free_chain(hBuf);
       }
#if defined(IPV6_DEBUG)
       else
       {
          ix_ossl_message_log("Message sent to core for transmission \n");
       }
#endif /* IPV6_DEBUG */

       ListRemoveItem(g_IcmpMsgQ, (void *)hBuf);
    }

LBL_EXIT:
    ICMP_UNLOCK();
    return errVal;
}

#else

ix_error _ix_cc_ipv6_icmp_send_err_message(void *arg_pContext)
{
    ix_error    errVal = IX_SUCCESS;
    ix_buffer_handle hBuf;
    ix_uint16   count;

    /*First schedule the next event, then do the rest of the processing*/
#if !defined(IX_EXCLUDE_CCI)
    errVal = ix_cci_cc_add_event_handler(g_pIpv6Context->hIpv6Handle,
                                            g_Icmpv6SleepTime*1000,
                                            _ix_cc_ipv6_icmp_send_err_message,
                                            IX_EVENT_TYPE_ONESHOT,
                                            1, &(g_IcmpQEventHdl));
    if(errVal != IX_SUCCESS)
        return IX_CC_IPV6_ERROR_CCI;
#elif defined(MEAT_TEST)
    ListAddItem(g_Context_Icmp, _ix_cc_ipv6_icmp_send_err_message);
#endif

    /*Now send the first packet in the queue*/
    ICMP_LOCK(); 
    if(!g_IcmpMsgQ)
    {
        ICMP_UNLOCK();
        return IX_CC_ERROR_NULL;
    }

    count = ListGetCount(g_IcmpMsgQ);
    if(count == 0)
        goto LBL_EXIT;

    hBuf = (ix_buffer_handle)ListGetFirst(g_IcmpMsgQ);
    if(!hBuf)
    {
        errVal = IX_NULL_BUFFER_HANDLE;
        goto LBL_EXIT;
    } 
    ListRemoveItem(g_IcmpMsgQ, (void *)hBuf);

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("ICMPv6 Invoking PASS PACKET with handle 0x%x \n", hBuf);
#endif /* IPV6_DEBUG */
    IX_IPV6_ERROR_CGT(_ix_cc_ipv6_pass_packet(hBuf), 
                                                errVal, LBL_FREE_BUF);

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("Message sent to core for transmission \n");
#endif /* IPV6_DEBUG */

LBL_FREE_BUF:
    ix_rm_buffer_free(hBuf);
LBL_EXIT:
    ICMP_UNLOCK();
    return errVal;
}
#endif /* _IX_OS_LINUX_KERNEL_ */

ix_error _ix_cc_ipv6_icmp_set_rate_limit(ix_uint16 rate)
{
    if((rate < IX_CC_ICMPV6_MIN_RATE_LIMIT_TIME) ||
       (rate > IX_CC_ICMPV6_MAX_RATE_LIMIT_TIME))
        return IX_CC_IPV6_ERROR_INVALID_RATE_LIMIT_TIME;
    
    ICMP_LOCK();
    g_Icmpv6SleepTime = rate;
    ICMP_UNLOCK();

    return IX_SUCCESS;
}

ix_error _ix_cc_ipv6_icmp_get_rate_limit(ix_uint16 *rate)
{
    ICMP_LOCK();
    *rate = g_Icmpv6SleepTime ;
    ICMP_UNLOCK();
    return IX_SUCCESS;
}


