/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *   Copyright (C) 2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 6
 *
 * = MODULE
 *      IPv6 Forwarder Core Component
 *
 * = FILENAME
 *      ipv6_fwdr.c
 *
 * = DESCRIPTION
 *   This file implements the forwarding functionality for packets it 
 *   receives from:
 *
 *   - The IPv6 microblock (running on the microengines)
 *   - Other core components (running on the x-scale along with this CC)
 *   - The stack driver.
 *   
 * = CREATION TIME
 *      12/12/2002
 *
 * = CHANGE HISTORY
 * ============================================================================
 */

#define IX_ERROR_FILE_IDENT "$Id: ipv6_fwder.c,v 1.18 2003/12/12 00:49:53 ktseng Exp $";

/**
 * System defined include files required.
 */

/**
 * User defined include files required.
 */
#include "ix_cc_error.h"
#include "cc/ipv6/internal/ipv6_fwdr.h"

/**
 * Preprocessor symbols and macros used in this file.
 */

/**
 * Types definitions whose scope is limited to this file.
 */

/**
 * Variable declarations global to this file only.  Externs are followed by
 * static variables.
 */

/**
 * Extern function prototypes.
 */

/* Function to process link local packets */
extern ix_error _ix_cc_ipv6_process_link_local_pkt(ix_buffer_handle arg_hBuffer,
                                          ix_uint128 arg_ipv6Addr);

/**
 * Static function prototypes.
 */

/**
 * Function definitions.
 */

/**
 * NAME: _ix_cc_ipv6_pass_packet
 *
 * DESCRIPTION: This function hands the given packet off to an outgoing target.  * The packet must be completely formed. 
 *
 * @Param:  - IN arg_hBuffer - The buffer to forward
 * @Param:  - INOUT  - None
 * @Param:  - OUT - None
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token
 *          for failure.
 *
 * Note: This routine will not free the hwBuffer if it returns an error. It
 * is the responsibilty of the caller to do so. If the routine returns 
 * success, the packet is either freed or sent to a different CC for 
 * furthur processing and the caller should not free the packet.
 */
ix_error _ix_cc_ipv6_pass_packet(ix_buffer_handle arg_hBuffer)
                                 
{
    ix_error err = IX_SUCCESS;
    void *pData;
    ix_uint32 oMtu;
    ix_uint16 bufSize;
    ix_cc_ipv6_header ipv6Hdr;
    ix_cc_rtmv6_next_hop_info nexthopInfo;
    ix_hw_buffer_meta *pHwMeta;
    ix_uint128 dest_address;
    ix_uint16 port, bladeID, l2Index;
#ifdef IPV6_STATS
    ix_uint32 index;
#endif /* IPV6_STATS */

    /* Extract the payload, in this case the IPv6 header, from the HW buffer */
    err = ix_cc_hw_get_packet_data(arg_hBuffer, &pData);
    if (err)
    {
         return IX_ERROR_NEW(IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
    }

    /* Get the meta data for the packet */
    err = ix_rm_buffer_get_meta(arg_hBuffer, (void **)&pHwMeta);
    if (err)
    {
        return IX_ERROR_NEW(IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
    }

#ifdef IPV6_STATS

    /* Increment the counter that tracks the number of datagrams
     * that were turned in for transmission by upper layer protocols
     * like ICMPv6 
     */
     g_pIpv6Context->ipv6FwdStats.ipv6CoreOutRequests++;

#endif /* IPV6_STATS */

    /* Initialize the IPv6 header */
    _ix_cc_ipv6_init_ip_header((ix_uint8 *)(pData), &ipv6Hdr);

    IX_NTOH128(&ipv6Hdr.dstAddress, &dest_address);

    /* Perform a lookup opertion */
    IX_ERROR_CR(ix_cc_rtmv6_lookup(g_pIpv6Context->hRtm, 
                                   dest_address, 
	                           &nexthopInfo));

    /* If there is no route ahead OR the "drop" flag is set, drop this pkt */
    if ((nexthopInfo.l2Index == IX_CC_RTMV6_L2INDEX_NO_ROUTE) ||
        (nexthopInfo.flags & IPV6_NH_FLAGS_DROP))
    {
#if defined(IPV6_DEBUG)
	ix_ossl_message_log("RTMv6 Lookup: Unreachable Host");
        _ix_cc_print_ipv6_address(&dest_address);
#endif /* IPV6_DEBUG */

#ifdef IPV6_STATS

    /* Increment the total number of locally
     * generated datagrams for which no route was found. 
     * NOTE: Since no outgoing interface was found, interface
     * specific counters do not make sense here.
     */
    g_pIpv6Context->ipv6FwdStats.ipv6CoreOutNoRoutes++;

#endif /* IPV6_STATS */

        ix_rm_buffer_free_chain(arg_hBuffer);

        return IX_SUCCESS;
    }

    /* outgoing MTU is in host byte order */
    oMtu = nexthopInfo.mtu;

    /* Size of the packet */
    bufSize = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_PacketSize));

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("Packet size: %d, mtu: %ld \n", bufSize, oMtu); 
#endif /* IPV6_DEBUG */

    /* 
     * By definition, the MTU specifies the maximum size IP datagram
     * that can be handled by this outgoing network interface
     */

    /* check if mtu < Packet size */
    if (oMtu < bufSize)
    {
        /* If outgoing MTU is less than the size of the packet, the packet
	 * cannot be sent. In this case, send an ICMP message back to the 
	 * sender of this packet. 
	 */

       ix_rm_buffer_free_chain(arg_hBuffer);

#if defined(IPV6_DEBUG)
       ix_ossl_message_log(" packet too big, dropping packet \n");
#endif /* IPV6_DEBUG */
    }
    else 
    {
        /* send the packet out to the queue manager */

        port = (ix_uint16)nexthopInfo.portID;
	bladeID = (ix_uint8)nexthopInfo.bladeID;
	l2Index = (ix_uint16)nexthopInfo.l2Index;

#ifdef IPV6_STATS

        /* Increment the counter for number of packets for which 
	 * a route was successfully found.
	 */
	g_pIpv6Context->ipv6FwdStats.ipv6OutCore[index].ipv6CoreOutForwDgrms++;
 
#if defined(IPV6_DEBUG)
        ix_ossl_message_log("NextHop.PortID (0x%x), NextHop.BladeID (0x%x) \n",
                                                 port, bladeID);
#endif /* IPV6_DEBUG */

	/* Compute the index of the blade/port for which this counter 
	 * needs to be incremented.
	 */
	index = (MAX_NUMBER_OF_PORTS_PER_BLADE * bladeID) + port;

        /* Count of the number of packets supplied to the lower layers
	 * for transmision
	 */

	g_pIpv6Context->ipv6FwdStats.ipv6OutCore[index].ipv6CoreOutTransmits++;

#endif /* IPV6_STATS */

#if defined(IPV6_DEBUG)
        ix_ossl_message_log("Setting up meta data and passing packet to QM \n");
#endif /* IPV6_DEBUG */

	/* 
	 * Setup the packet meta data before sending it off to the Queue
	 * Manager. All the other meta data fields should be setup by ICMP6.
	 */
        pHwMeta->m_NextHopID = IX_HTON16(l2Index);
	pHwMeta->m_OutputPort = pHwMeta->m_InputPort;
	pHwMeta->m_FabricPort = (ix_uint8)bladeID;

#if defined(IPV6_DEBUG)
        ix_ossl_message_log("Sending Packet to QM with buffer Handle (0x%lx)\n",                                                      (ix_uint32)arg_hBuffer);
#endif /* IPV6_DEBUG */

        /* Send the packet to queue Manager */
        err = ix_rm_packet_send(IX_CC_IPV6_PKT_QM_OUTPUT, arg_hBuffer, 0);

        if (err)
        {
            /* Free the buffer and return error to caller */
            ix_rm_buffer_free_chain(arg_hBuffer);
            return IX_ERROR_NEW(IX_CC_ERROR_SEND_FAIL, IX_ERROR_LEVEL_WARNING);
        }
    }

    return IX_SUCCESS;
}

/**
 * NAME: _ix_cc_ipv6_forward
 *
 * DESCRIPTION: This function performs forwarding of packets coming 
 * from the IPv6 microblock and other core Components other than Stack 
 * Driver core component
 * 
 * @Param:  - IN - arg_hBuffer - a buffer handle containing the packet that 
 *                 needs to be forwarded  
 * @Param:  - IN - arg_inCounters - do we need to maintain in counters.
 * @Param:  - IN - arg_outCounters - do we need to maintain out counters.
 *		 
 * @Param:  - OUT - None  
 *
 * @Return: IX_SUCCESS if successful 
 *          valid ix_error token for failure.
 *
 * NOTE: arg_hBuffer treatment:
 * arg_hBuffer will be freed if the function returns an error. In cases
 * where it returns success, the buffer is probably going to a different
 * core component (unless the packet is being dropped).
 * Overall the caller of this function does not have to worry about freeing
 * the buffer.
 */

ix_error _ix_cc_ipv6_forward (ix_buffer_handle arg_hBuffer, 
                              ix_uint8 arg_inCounters,
			      ix_uint8 arg_outCounters)
{
    void *pData;
    ix_hw_buffer_meta *pHwMeta; 
    ix_uint16 packetSize, ingressPort;
    ix_uint32 ingressPortStatus = 0, egressPortStatus = 0;        
    ix_cc_ipv6_header ipv6Hdr;
    ix_uint32 dropped;
    ix_cc_rtmv6_next_hop_info nexthopInfo;
    ix_error err = IX_SUCCESS;
    ix_uint16 l2Index;
    ix_uint16 port;
    ix_uint128 dest_address;
    ix_uint8 bladeID;
    ix_uint32 index = 0;


    /* Here is how this function works:
     * The IPv6 packet is contained in one or more (chained) HW buffers
     * and arg_hBuffer is a handle to this chain of HW buffers.
     * 
     * The forwarder really only looks at the IPv6 header and other extension
     * headers and does not care about the payload of the packet.
     * To extract the header, it uses the buffer manipulation functions 
     * and all the meta-data functions. The IPv6 packet is considered
     * data payload of the HW buffer.
     */
    
    /* extract IPv6 header */
    err = ix_cc_hw_get_packet_data(arg_hBuffer, &pData);

    /* Attempt to free the buffer if get data fails */
    if (err)
    {
        ix_rm_buffer_free_chain(arg_hBuffer);
        return IX_ERROR_NEW(IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);        
    }
   
    /* Get the meta data corresponding to this packet */
    err = ix_rm_buffer_get_meta(arg_hBuffer, (void **)&pHwMeta);
    if (err)
    {
        ix_rm_buffer_free_chain(arg_hBuffer);
        return IX_ERROR_NEW(IX_CC_ERROR_INTERNAL,IX_ERROR_LEVEL_WARNING);
    }

    /* 
     * Get the input interface 
     * Note: m_InputPort is ix_uint16 
     */

    ingressPort = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_InputPort));

#if defined(IPV6_DEBUG)
    ix_ossl_message_log( "IPv6 Forwarder got a packet from port %d\n",
                                                                ingressPort );
#endif /* IPV6_DEBUG */

    /* Storing packet size in host byte order */
    packetSize = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_PacketSize));

#ifdef IPV6_STATS
    if (arg_inCounters)
    {
        /* Increment the total number of IP datagrams received */
	g_pIpv6Context->ipv6FwdStats.ipv6InCore[ingressPort].ipv6CoreInRecv++;

	/* Increment the total octet count of incoming datagrams received */
	g_pIpv6Context->ipv6FwdStats.ipv6InCore[ingressPort].ipv6CoreInOctets +=
	                                              packetSize;
    }
#endif /* IPV6_STATS */

    /*
     * Is the packet smaller than the minimum IPv6 packet size?
     */

    if (packetSize <  IX_CC_IPV6_HDR_LEN) 
    {
#ifdef IPV6_STATS
        /* Increment the total number of truncated pkts */
	g_pIpv6Context->ipv6FwdStats.ipv6InCore[ingressPort].ipv6CoreInTruncatedPkts++;
#endif /* IPV6_STATS */
	/* drop the packet */
        ix_rm_buffer_free_chain(arg_hBuffer);
        return IX_SUCCESS;
    }

    /* Initialize the IPv6 header */
    _ix_cc_ipv6_init_ip_header((ix_uint8 *)(pData), &ipv6Hdr);
    
#if defined(IPV6_DEBUG)
    ix_ossl_message_log( "Validating header ... \n");
#endif /* IPV6_DEBUG */

    /* Validate the IPv6 header - RFC 2460 */
    IX_ERROR_CR(_ix_cc_ipv6_IPheader_validate(&ipv6Hdr, 
                                              arg_hBuffer, 
					      &dropped, 
					      arg_inCounters, 
					      arg_outCounters));
    
    if (dropped)
    {
	return IX_SUCCESS;
    }

#if defined(IPV6_DEBUG)
    ix_ossl_message_log( "Validating addresses ... \n");
#endif /* IPV6_DEBUG */

    /* Validate the IPv6 addresses in the packet - RFC 2373 checks */
    IX_ERROR_CR(_ix_cc_ipv6_validate_IP_addresses(&ipv6Hdr, 
                                                  arg_hBuffer, 
						  &dropped, 
						  arg_inCounters,
						  arg_outCounters));
    if (dropped)
    {
	return IX_SUCCESS;
    }

    /*  
     * If the next header field in the IPv6 header indicates the presence
     * of any extension header, process the extension header.
     * TBD
     */

#if defined(IPV6_DEBUG)
    ix_ossl_message_log( "Checking for extension headers ... \n");
#endif /* IPV6_DEBUG */

    if (ipv6Hdr.nextHeader != IX_CC_IPV6_NEXT_HDR_NONE)
    {
        IX_ERROR_CR(_ix_cc_ipv6_parse_ext_headers(&ipv6Hdr, 
	                                         arg_hBuffer, 
						 &dropped));
        if (dropped)
        {
            return IX_SUCCESS;
        }

	/* Simply continue on with forwarding */
    }

#ifdef IPV6_STATS
    /* Increment a count of the number of packets forwarding was attempted
     * on.
     */
    g_pIpv6Context->ipv6FwdStats.ipv6InCore[ingressPort].ipv6CoreInForwDgrms++;
#endif /* IPV6_STATS */
    
    /* Perform RTMv6 lookup: The result of the lookup is the nexthop
     * information structure 
     */

#if defined(IPV6_DEBUG)
    ix_ossl_message_log( "Performing address lookup ... \n");
#endif /* IPV6_DEBUG */

    IX_NTOH128(&ipv6Hdr.dstAddress, &dest_address);

    IX_ERROR_CR(ix_cc_rtmv6_lookup(g_pIpv6Context->hRtm, 
                                   dest_address, 
	                           &nexthopInfo));
    if (err != IX_SUCCESS)
    {
        /* Lookup failed, drop the packet */
#if defined(IPV6_DEBUG)
        ix_ossl_message_log("lookup failed in ipv6_forward \n");
#endif /* IPV6_DEBUG */

        ix_rm_buffer_free_chain(arg_hBuffer);
        return IX_ERROR_NEW(IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    }


    /* Perform Post Lookup Next Hop validation here */

    /* If route entry is invalid, send a "destination unreachable"
     * ICMP error message back to the sender and drop the packet 
     */
    if (nexthopInfo.l2Index == IX_CC_RTMV6_L2INDEX_NO_ROUTE)
    {

#ifdef IPV6_STATS

	/* Increment the count of number of packets for which no route
    	 * was found
	 */
	g_pIpv6Context->ipv6FwdStats.ipv6InCore[ingressPort].ipv6CoreInNoRoutes++;
#endif /* IPV6_STATS */

#if defined(IPV6_DEBUG)
	ix_ossl_message_log("RTMv6 Lookup: Unreachable Host");
#endif /* IPV6_DEBUG */

#if defined(IPV6_DEBUG)
        _ix_cc_print_ipv6_address(&dest_address);
#endif /* IPV6_DEBUG */
		 
	err = _ix_cc_ipv6_icmp_dest_unreachable(arg_hBuffer,
                                        IX_CC_ICMPV6_DU_CODE_ADDR_UNREACHABLE);
        if (err)
        {
            /* The input buffer has been freed. Just return an error */
            return IX_ERROR_NEW(IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
        }

        return IX_SUCCESS;
    }

    /* Make sure that the MTU of the outgoing interface is large
     * enough to support this packet.
     */

    else if (nexthopInfo.mtu < packetSize)
    {
        /* Send an ICMP error message to the sender saying 
	 * "Packet Too Big" 
	 */
        err = _ix_cc_ipv6_icmp_pkt_too_big(arg_hBuffer, nexthopInfo.mtu);

        if (err)
        {
            /* The input buffer has been freed. Just return an error */
            return IX_ERROR_NEW(IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
        }

        return IX_SUCCESS;
    }

    /* If route entry indicates local delivery, send the packet to the 
     * stack driver 
     */
    else if (nexthopInfo.flags & IPV6_NH_FLAGS_LOCAL)
    {
	/* send the packet to the stack driver */

#ifdef IPV6_STATS
	/* Keep count of the number of packets for local delivery */
	g_pIpv6Context->ipv6FwdStats.ipv6InCore[ingressPort].ipv6CoreInDelivers++;
#endif /* IPV6_STATS */
		
#if defined(IPV6_DEBUG)
	ix_ossl_message_log("Packet for Local delivery\n");
#endif /* IPV6_DEBUG */
     
	/* Send the packet to Stack driver CC through high 
	 * priority queue 
	 */
        err = ix_rm_message_send(IX_CC_IPV6_PKT_HIGH_PRIORITY_STKDRV_OUTPUT,
                                 arg_hBuffer, 
				 0);
        if (err)
        {
            /* Free the buffer and return error to caller */
            ix_rm_buffer_free_chain(arg_hBuffer);
            return IX_ERROR_NEW(IX_CC_ERROR_SEND_FAIL, IX_ERROR_LEVEL_WARNING);
        }            
    }
    else if (nexthopInfo.flags & IPV6_NH_FLAGS_DROP)
    {
#if defined(IPV6_DEBUG)
        ix_ossl_message_log("Route entry indicates drop target\n");
#endif /* IPV6_DEBUG */
        ix_rm_buffer_free_chain(arg_hBuffer);
    }
    else
    {
        l2Index = (ix_uint16)nexthopInfo.l2Index;
        port = (ix_uint16)nexthopInfo.portID;
	bladeID = (ix_uint8)nexthopInfo.bladeID;

	/* Compute the index of the blade/port for which this counter 
	 * needs to be incremented.
	 */
	index = (MAX_NUMBER_OF_PORTS_PER_BLADE * bladeID) + port;

#ifdef IPV6_STATS

        /* Increment the counter for number of packets for which 
	 * a route was successfully found.
	 */
	g_pIpv6Context->ipv6FwdStats.ipv6OutCore[index].ipv6CoreOutForwDgrms++;

#endif /* IPV6_STATS */

        /* If the hop limit has expired, send an "time exceeded" ICMP
	 * error message to the sender and drop the packet
	 */
        if (ipv6Hdr.hopLimit <= 1)
        {
#ifdef IPV6_STATS
            /* Increment the total count of packets with bad headers */
	    g_pIpv6Context->ipv6FwdStats.ipv6InCore[ingressPort].ipv6CoreInHdrErrors++;
#endif /* IPV6_STATS */
            err = _ix_cc_ipv6_icmp_time_exceeded(arg_hBuffer, 
	                              IX_CC_ICMPV6_TE_CODE_HOP_LIMIT_EXCEEDED);

            if (err)
            {
                /* The input buffer has been freed. Just return an error */
                return IX_ERROR_NEW(IX_CC_ERROR_INTERNAL, 
		                                 IX_ERROR_LEVEL_WARNING);
            }            

            return IX_SUCCESS;
        }

        ipv6Hdr.hopLimit--;

	/* At this point everything looks good, we extract information
	 * out of the Next Hop structure and populate the meta 
	 * data associated with this buffer. The buffer is probably
	 * transmitted over the CSIX fabric over to the egress side for
	 * transmission
	 */

	/* 
	 * Extract relevant information out of the next hop structure
	 */

        /* Fill the L2 next hop ID in the meta data */
        pHwMeta->m_NextHopID = IX_HTON16(l2Index);

        /* Fill out buffer meta with output port **/
        pHwMeta->m_OutputPort = IX_HTON16(port);

        err = _ix_cc_ipv6_get_port_status(ingressPort, &ingressPortStatus);
        if (err != IX_SUCCESS)
	{
            ix_rm_buffer_free_chain(arg_hBuffer);
	    return IX_ERROR_NEW(IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
	}
            
        err =  _ix_cc_ipv6_get_port_status((ix_uint16)nexthopInfo.portID,
                                                           &egressPortStatus);
	if (err != IX_SUCCESS)
	{
            ix_rm_buffer_free_chain(arg_hBuffer);

	}			

        /* Copy the IP header into the buffer */
        ix_ossl_memcpy(pData, &ipv6Hdr, IX_CC_IPV6_HDR_LEN);
           
	/* Here the assumption is as follows:
	 * If the port status is DISABLED, it means that we don't want to 
	 * forward packet (packet forward is DISABLED) either coming in or
	 * going out of disabled port.
	 */
        if ((ingressPortStatus == IX_CC_IPV6_DISABLE_PORT_STATUS) || 
	    (egressPortStatus == IX_CC_IPV6_DISABLE_PORT_STATUS))
        {
#if defined(IPV6_DEBUG)
            ix_ossl_message_log(" Port disabled destination unreachable\n" );
#endif /* IPV6_DEBUG */

	    err = _ix_cc_ipv6_icmp_dest_unreachable(arg_hBuffer,
                                         IX_CC_ICMPV6_DU_CODE_COMM_PROHIBITED);
	    if (err != IX_SUCCESS)
	    {
                /* The input buffer has been freed. Just return an error */
	        return IX_ERROR_NEW(IX_CC_ERROR_INTERNAL, 
		                     IX_ERROR_LEVEL_WARNING);
            }

            return IX_SUCCESS;
        }

        /* Now to check for redirect: Check if ingress interface is same as 
         * egress interface. 
	 */
        /* ingressPort is in host order (16 bits)
         * nexthopInfo.portID is in host order (32 bits)
	 *
	 * The check to see if the incoming interface and the outgoing
	 * interface are the same is to see if the portID's match and 
	 * the BladeID's match.
	 */
        if ((ingressPort == (ix_uint16)nexthopInfo.portID) &&
            (g_Ipv6RegistryData.localBladeId == nexthopInfo.bladeID))
        {
#if defined(IPV6_DEBUG)
            ix_ossl_message_log("Redirect\n");
#endif /* IPV6_DEBUG */

#if defined(ICMPV6_REDIRECT_IMPLEMENTED)

            /* nexthopIpAddress will be in host byte order */
            nexthopIpAddress = nexthopInfo.ipv6Addr;

            err = _ix_cc_ipv6_icmp_redirect(arg_hBuffer, nexthopIpAddress);

            if (err)
            {
                /* The input buffer has been freed. Just return an error */
		return IX_ERROR_NEW(IX_CC_ERROR_INTERNAL,
 						IX_ERROR_LEVEL_WARNING);
            }
#endif /* ICMPV6_REDIRECT_IMPLEMTENTED */

            return IX_SUCCESS;		
        }

#ifdef IPV6_STATS

        /* Count of the number of packets supplied to the lower layers
	 * for transmision
	 */
	g_pIpv6Context->ipv6FwdStats.ipv6OutCore[index].ipv6CoreOutTransmits++;
#endif /* IPV6_STATS */

#if defined(IPV6_DEBUG)
        ix_ossl_message_log("Sending Packet to QM with buffer Handle (0x%lx)\n", (ix_uint32)arg_hBuffer);
#endif /* IPV6_DEBUG */

        /* Send the packet to queue Manager */
        err = ix_rm_packet_send(IX_CC_IPV6_PKT_QM_OUTPUT, arg_hBuffer, 0);

        if (err)
        {
            /** Free the buffer and return error to caller **/
            ix_rm_buffer_free_chain(arg_hBuffer);
            return IX_ERROR_NEW(IX_CC_ERROR_SEND_FAIL, IX_ERROR_LEVEL_WARNING);
        }


    }

    return IX_SUCCESS;
}

/**
 * NAME: _ix_cc_ipv6_forward_pkts_from_stkdrvr
 *
 * DESCRIPTION: This function forwards the packets received from stack driver.
 *             
 * @Param:  - IN - arg_hBuffer - a buffer handle containing the packet that 
 *            needs to be forwarded  
 *		 
 * @Param:  - INOUT  - None
 * @Param:  - OUT - None  
 *
 * @Return: IX_SUCCESS if successful 
 *          valid ix_error token for failure.
 */                    
ix_error _ix_cc_ipv6_forward_pkts_from_stkdrvr(ix_buffer_handle arg_hBuffer)
{
    
    void *pData; 
    ix_uint32 ingressPortStatus = 0, egressPortStatus = 0;     
    ix_hw_buffer_meta *pHwMeta; 
    ix_uint32 dropped;
    ix_uint16 ingressPort;
    ix_cc_ipv6_header ipv6Hdr;
    ix_cc_rtmv6_next_hop_info nexthopInfo;
    ix_error err = IX_SUCCESS;
    ix_uint16 l2Index;
    ix_uint16 port;
    ix_uint128 dest_address;
    ix_uint32 packetSize;
#ifdef IPV6_STATS
    ix_uint32 index = 0;
#endif /* IPV6_STATS */
    ix_uint8 bladeID;

    /* For packets coming from stack driver CC, the IPv6 forwarder CC 
     * does not perform the following -
     * - No header validation is performed on the packets.
     * - The Hop Limit will not be decremented.
     * - No ICMP error messages will be generated to avoid conflict with the
     *   ICMP messages generated by the TCP/IP stack.
     */ 

    /* extract data from buffer handle */
    err = ix_cc_hw_get_packet_data(arg_hBuffer, &pData);
    if (err)
    {
        ix_rm_buffer_free_chain(arg_hBuffer);
        return IX_ERROR_NEW(IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
    }

      
    /* Use ix_rm_buffer_get_meta to get input interface number */
    err = ix_rm_buffer_get_meta( arg_hBuffer, (void **)&pHwMeta);
    if (err)
    {
        ix_rm_buffer_free_chain(arg_hBuffer);
	return IX_ERROR_NEW(IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
    }

    /* IngressPort value is in host byte order */
    ingressPort = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_InputPort));

#ifdef IPV6_STATS

    /* Increment the counter that tracks the number of datagrams
     * that were turned in for transmission by upper layer protocols
     * like ICMPv6 
     */
     g_pIpv6Context->ipv6FwdStats.ipv6CoreOutRequests++;

#endif /* IPV6_STATS */

    /* Storing packet size in host byte order */
    packetSize = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_PacketSize));

#if defined(IPV6_DEBUG)
    ix_ossl_message_log( "IPv6 Forwarder got a packet from"
                         " stack driver Core Component from port  %x\n",
                                                            ingressPort );
#endif /* IPV6_DEBUG */

    /* Initialize the IPv6 header from the packet data */
    _ix_cc_ipv6_init_ip_header((ix_uint8 *)(pData), &ipv6Hdr);

    IX_ERROR_CR(_ix_cc_ipv6_validate_IP_addresses(&ipv6Hdr, 
                                                  arg_hBuffer, 
                                                  &dropped, 
						  0,
						  0));
    
    if (dropped)
    {
#if defined(IPV6_DEBUG)
        ix_ossl_message_log( "Invalid IPv6 address, dropping\n" );
#endif /* IPV6_DEBUG */

        return IX_SUCCESS;
    }
    
    /* If the next header field in the IPv6 header indicates the presence
     * of an extension header, process the extension header
     */
    if (ipv6Hdr.nextHeader != IX_CC_IPV6_NEXT_HDR_NONE)
    {
        IX_ERROR_CR(_ix_cc_ipv6_parse_ext_headers(&ipv6Hdr, 
	                                         arg_hBuffer, 
						 &dropped));
        if (dropped)
        {
            return IX_SUCCESS;
        }

	/* Simply continue on with forwarding */
    }

    IX_NTOH128(&ipv6Hdr.dstAddress, &dest_address);

    /* Perform a route lookup on destination IP address */
    err = ix_cc_rtmv6_lookup(g_pIpv6Context->hRtm, 
                             dest_address,
			     &nexthopInfo);

    if (err != IX_SUCCESS)
    {
#if defined(IPV6_DEBUG)
        /* Lookup failed, drop the packet */
        ix_ossl_message_log("lookup failed in ipv6_forward \n");
#endif /* IPV6_DEBUG */

        ix_rm_buffer_free_chain(arg_hBuffer);
        return IX_ERROR_NEW(IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    }
    
    /* Check if route entry is invalid */
    if (nexthopInfo.l2Index == IX_CC_RTMV6_L2INDEX_NO_ROUTE)
    {
#if defined(IPV6_DEBUG)
	ix_ossl_message_log("RTMv6 Lookup: Unreachable Host.");
#endif /* IPV6_DEBUG */

#if defined(IPV6_DEBUG)
        _ix_cc_print_ipv6_address(&dest_address);
#endif /* IPV6_DEBUG */

        ix_rm_buffer_free_chain(arg_hBuffer);
        return IX_SUCCESS;	
    }

    /* Make sure that the MTU of the outgoing interface is large
     * enough to support this packet.
     */
    else if (nexthopInfo.mtu < packetSize)
    {
        ix_rm_buffer_free_chain(arg_hBuffer);
        return IX_SUCCESS;
    }

    /* This is unexpected and should not really happen, but we check it 
     * anyway, just in case it does happen.
     * If this is true, what this means is a packet that came from the stack
     * driver above, is destined for local delivery and needs to be sent
     * back to the stack driver.
     */
    else if(nexthopInfo.flags & IPV6_NH_FLAGS_LOCAL)
    {
        /* send it back to the stack driver */
#if defined(IPV6_DEBUG)
        ix_ossl_message_log("Packet for Local delivery\n" );
#endif /* IPV6_DEBUG */

	/* Send the packet to Stack driver CC through high priority queue */
        err = ix_rm_message_send(IX_CC_IPV6_PKT_HIGH_PRIORITY_STKDRV_OUTPUT,
                                 arg_hBuffer, 
				 0);

    	/* If error free the buffer */
        if (err)
        {
            ix_rm_buffer_free_chain(arg_hBuffer);
	    return IX_ERROR_NEW(IX_CC_ERROR_SEND_FAIL, IX_ERROR_LEVEL_WARNING);
        }
    }
    else if (nexthopInfo.flags & IPV6_NH_FLAGS_DROP)
    {
#if defined(IPV6_DEBUG)
        ix_ossl_message_log("Route entry indicates drop packet\n" );
#endif /* IPV6_DEBUG */

	ix_rm_buffer_free_chain(arg_hBuffer );
    }
    else
    {
        if (ipv6Hdr.hopLimit == 0 )
        {
            ix_rm_buffer_free_chain(arg_hBuffer);
            return IX_SUCCESS;            
	    
        }
        else /* If Hop Limit has not expired */
        {
            /* 
	     * Fill out the buffer handle with nexthop ID (L2 Index) and
             * output port ID, which were obtained as a result 
	     * of the lookup 
	     */

            l2Index = (ix_uint16)nexthopInfo.l2Index;
            port = (ix_uint16)nexthopInfo.portID;
	    bladeID = (ix_uint8)nexthopInfo.bladeID;

            pHwMeta->m_NextHopID = IX_HTON16(l2Index);
	    pHwMeta->m_OutputPort = IX_HTON16(port);
 	    pHwMeta->m_FabricPort = bladeID;
            
	    /* Copy the header into the hardware packet */
            ix_ossl_memcpy(pData, &ipv6Hdr, IX_CC_IPV6_HDR_LEN);

	    /* Get the status of the ingress port */
            err = _ix_cc_ipv6_get_port_status(ingressPort, &ingressPortStatus);
            if (err != IX_SUCCESS)
            {
                ix_rm_buffer_free_chain(arg_hBuffer);
	        return IX_ERROR_NEW(IX_CC_ERROR_INTERNAL, 
		                                     IX_ERROR_LEVEL_WARNING);
            }	

	    /* Get the status of the egress port */
            err = _ix_cc_ipv6_get_port_status((ix_uint16)nexthopInfo.portID,
                                              &egressPortStatus);
            if (err != IX_SUCCESS)
            {
                ix_rm_buffer_free_chain(arg_hBuffer);
		return IX_ERROR_NEW(IX_CC_ERROR_INTERNAL, 
				    IX_ERROR_LEVEL_WARNING);
            }	

            /* If either ingress or egress port is disabled, drop the buffer */
            if ((ingressPortStatus == IX_CC_IPV6_DISABLE_PORT_STATUS) || 
	        (egressPortStatus == IX_CC_IPV6_DISABLE_PORT_STATUS))
            {
#if defined(IPV6_DEBUG)
                ix_ossl_message_log("Port disabled: destination unreachable\n");
#endif /* IPV6_DEBUG */

		ix_rm_buffer_free_chain(arg_hBuffer);
                return IX_SUCCESS;                               
            }

	    /* Redirect Check:
	     * Check to see if the ingress interface (combination of blade
	     * ID and port ID) is the same as the egress interface 
	     */

            /* ingressPort is in Host byte order */            
            if ((ingressPort == (ix_uint16)nexthopInfo.portID )&&
                (g_Ipv6RegistryData.localBladeId == nexthopInfo.bladeID))
            {
#if defined(IPV6_DEBUG)
                ix_ossl_message_log("Redirect check passed \n");
#endif /* IPV6_DEBUG */

                ix_rm_buffer_free_chain(arg_hBuffer);
                return IX_SUCCESS;
            }

#ifdef IPV6_STATS
	    /* Compute the index of the blade/port for which this counter 
	     * needs to be incremented.
	     */
	     index = (MAX_NUMBER_OF_PORTS_PER_BLADE * bladeID) + port;

            /* Count of the number of packets supplied to the lower layers
    	     * for transmision
	     */
	    g_pIpv6Context->ipv6FwdStats.ipv6OutCore[index].ipv6CoreOutTransmits++;
#endif /* IPV6_STATS */
    
#if defined(IPV6_DEBUG)
            ix_ossl_message_log("Sending Packet to QM with buffer Handle (0x%lx)\n", (ix_uint32)arg_hBuffer);
#endif /* IPV6_DEBUG */

            /* Send the packet to queue Manager */
            err = ix_rm_packet_send(IX_CC_IPV6_PKT_QM_OUTPUT, arg_hBuffer, 0);

            if (err)
            {
                /* Free the buffer and return error to caller */
                ix_rm_buffer_free_chain(arg_hBuffer);
                return IX_ERROR_NEW(IX_CC_ERROR_SEND_FAIL, 
		                                  IX_ERROR_LEVEL_WARNING);
            }
        }
    }

    return IX_SUCCESS;
}
