/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 6
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      IPv6 Forwarder Core Component - library primitives
 *
 * = FILENAME
 *      ix_cc_ipv6_library.c
 *
 * = DESCRIPTION
 *      The file defines library interface functions.
 *
 * = AKNOWLEDGEMENTS
 *      
 * = CREATION TIME
 *      12/03/2002
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_ipv6_library.c,v 1.26 2003/12/12 00:49:53 ktseng Exp $";

/**
 * System defined include files
 */


/**
 * User defined include files
 */

#include "ix_cc_error.h"
#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_netmacros.h"
#include "cc/ix_cc_rtmv6.h"

#include "ix_cc.h"
#include "bindings.h"
#include "cc/ix_cc_msup.h"
#include "ipv6_fwder_uc.h"
#include "cc/ix_cc_ipv6.h"
#include "cc/ipv6/internal/utils.h"
#include "cc/ipv6/internal/ix_cc_ipv6_msg_handler.h"
#include "cc/ipv6/internal/ix_cc_ipv6_prop.h"
#include "cc/ipv6/internal/ix_cc_ipv6_init.h"
#include "cc/ipv6/internal/ix_cc_ipv6_nd.h"
#include "cc/ipv6/internal/icmp6.h"

/**
 * Preprocessor symbols and Macros used in this file
 */

/**
 * Type definitions whose scope is limited to this file
 */

extern ix_uint16 g_Icmpv6SleepTime;
extern ix_uint16 g_Icmpv6QueueDrain;
extern ix_uint16 g_Icmpv6QueueDepth;
extern ix_cc_ipv6_context* g_pIpv6Context;

/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */


/** 
 * Extern function prototypes.
 */


/** 
 * Static function prototypes.
 */
static 
ix_error _ix_cc_check_prefix_notifications(void);

/**
 * TYPENAME: ix_cc_route_update_element
 *
 * DESCRIPTION: This structure defines the format of each element stored to 
 * provide route updates.
 *
 */
typedef struct ix_s_cc_route_update_element
{
    ix_uint128 prefix;
    ix_uint32 current_nhid;

} ix_cc_route_update_element;

/**
 * TYPENAME: ix_cc_route_update_member
 *
 * DESCRIPTION: This structure defines the format of each member (each client
 * that has registered for route updates).
 */
typedef struct ix_s_cc_route_update_member
{
    LPLIST prefixes;
    ix_uint32 member_msg_id;

} ix_cc_route_update_member;

static LPLIST g_RouteUpdateList = NULL;
static ix_ossl_mutex_t g_RouteUpdateLock;

/** 
 * Function definitions.
 */


/**
 * NAME: ix_cc_ipv6_add_prefix
 *
 * DESCRIPTION:This library primitive helps to add a prefix in RTMv6. 
 * This API internally calls ix_cc_rtmv6_add_prefix API of RTMv6 core 
 * component to add prefix in RTM database.
 * 
 * @Param:  - IN arg_IpAddr: network IPv6 address identifying the prefix.
 * @Param:  - IN arg_PrefixLength: prefix length for the prefix.
 * @Param:  - IN arg_NextHopId: next hop identifier for prefix.
 * @Param:  - IN arg_pContext: pointer to IPv6 context
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - invalid next hop identifier
 * IX_CC_ERROR_DUPLICATE_ROUTE - prefix was already added to this NHID
 * IX_CC_ERROR_CONFLICTING_ROUTE - prefix was already added to another NHID
 * IX_CC_IPV6_ERROR_RTM - RTM internal failure
 * 
 */

ix_error ix_cc_ipv6_add_prefix(ix_uint128 arg_IpAddr,
                              ix_uint8 arg_PrefixLength,
                              ix_cc_rtmv6_nhid arg_NextHopId,
                              void *arg_pContext)
{
    ix_error err = IX_SUCCESS;

    /* IPv6 Address is in Host Byte Order */
    IX_ERROR_C(ix_cc_rtmv6_add_prefix(((ix_cc_ipv6_context*)arg_pContext)->hRtm,
                                      arg_IpAddr,
                                      arg_PrefixLength,
                                      arg_NextHopId), err);
    if(err != IX_SUCCESS)
    {
        switch(IX_ERROR_GET_CODE(err))
        {
            case IX_CC_ERROR_ENTRY_NOT_FOUND:
            case IX_CC_ERROR_DUPLICATE_ENTRY:
            case IX_CC_ERROR_CONFLICTING_ENTRY:
		return err;

            default:
            	return IX_ERROR_SET_CODE(err, IX_CC_IPV6_ERROR_RTM);
        }/* switch (err) */
    }/* end if (err != IX_SUCCESS) */     

   

    return err;

} /* end ix_cc_ipv6_add_prefix */

/**
 * NAME: ix_cc_ipv6_delete_prefix
 *
 * DESCRIPTION: This library primitive helps to delete a prefix from RTMv6.
 * This API internally calls ix_cc_rtmv6_delete_prefix API of RTM core 
 * component to delete prefix from RTMv6 database.
 * 
 * @Param:  - IN arg_IpAddr: network IPv6 address identifying the prefix
 * @Param:  - IN arg_PrefixLength: prefix length for this prefix
 * @Param:  - IN arg_pContext: pointer to IPv6 context
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - entry not found in RTM
 * IX_CC_IPV6_ERROR_RTM - RTM internal failure
 * 
 */

ix_error ix_cc_ipv6_delete_prefix(ix_uint128 arg_IpAddr,
                                 ix_uint8 arg_PrefixLength,
                                 void *arg_pContext)
{

    ix_uint32 errCode;
    ix_error err = IX_SUCCESS;


    /* IPv6 address is in Host Byte Order */
    IX_ERROR_C(ix_cc_rtmv6_delete_prefix(( 
                                    (ix_cc_ipv6_context*)arg_pContext)->hRtm,
                                    arg_IpAddr,
                                    arg_PrefixLength), err);

    if(err != IX_SUCCESS)
    {
        errCode = IX_ERROR_GET_CODE(err);
        if(errCode == IX_CC_ERROR_ENTRY_NOT_FOUND)
        {
            return err;
        }/* end if (err == IX_CC_ERROR_ENTRY_NOT_FOUND) */
        else
        {
 	    return IX_ERROR_SET_CODE(err, IX_CC_IPV6_ERROR_RTM); 
        }/* end else(err == IX_CC_ERROR_ENTRY_NOT_FOUND) */

	return err;
    }/* end if (err != IX_SUCCESS) */     

    err = _ix_cc_check_prefix_notifications();
    if (err)
    {
#if defined(IPV6_DEBUG)
        ix_ossl_message_log ("Failed to check for prefix notifications \n");
#endif /* IPV6_DEBUG */
    }

    return err;

} /* end ix_cc_ipv6_delete_prefix */



/**
 * NAME: ix_cc_ipv6_update_prefix
 *
 * DESCRIPTION:This library primitive helps to update a prefix in RTMv6.
 * This API internally calls ix_cc_rtmv6_update_prefix API of RTM core 
 * component to update a prefix in RTM database. 
 * 
 * @Param:  - IN arg_IpAddr: network IPv6 address for the prefix to be updated
 * @Param:  - IN arg_PrefixLength: network mask for the prefix to be updated
 * @Param:  - IN arg_NextHopId: hop identifier for prefix to be updated
 * @Param:  - IN arg_pContext: pointer to IPv6 context
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - invalid next hop identifier/prefix
 * IX_CC_IPV6_ERROR_INVALID_MASK - invalid netmask
 * IX_CC_ERROR_CONFLICTING_ENTRY - prefix was already added to another NHID
 * IX_CC_IPV6_ERROR_RTM - RTM internal failure
 * 
 */

ix_error ix_cc_ipv6_update_prefix(ix_uint128 arg_IpAddr,
                                 ix_uint8 arg_PrefixLength,
                                 ix_cc_rtmv6_nhid arg_NextHopId,
                                 void *arg_pContext)
{
    ix_error err = IX_SUCCESS;

    /* IPv6 address is in Host Byte Order */
    IX_ERROR_C(ix_cc_rtmv6_update_prefix((
                                (ix_cc_ipv6_context*)arg_pContext)->hRtm,
                                arg_IpAddr,
                                arg_PrefixLength,
                                arg_NextHopId), err);
    if(err != IX_SUCCESS)
    {
        switch(IX_ERROR_GET_CODE(err))
        {
            case IX_CC_ERROR_ENTRY_NOT_FOUND:
            case IX_CC_ERROR_CONFLICTING_ENTRY:
		return err;
            case IX_CC_RTMV6_ERROR_INVALID_PREFIX:
            	return IX_ERROR_SET_CODE(err, IX_CC_IPV6_ERROR_INVALID_PREFIX);
            default:
            	return IX_ERROR_SET_CODE(err, IX_CC_IPV6_ERROR_RTM);
        }/* switch (err) */

        return err;
    }/* end if (err != IX_SUCCESS) */ 

    err = _ix_cc_check_prefix_notifications();
    if (err != IX_SUCCESS)
    {
#if defined(IPV6_DEBUG)
        ix_ossl_message_log ("Failed to check for prefix notifications \n");
#endif /* IPV6_DEBUG */
    }

    return err;

} /* end ix_cc_ipv6_update_prefix */


/**
 * NAME: ix_cc_ipv6_lookup_prefix
 *
 * DESCRIPTION: This primitive helps a client to look up routing information 
 * for a given IPv6 address. This internally call ix_cc_rtmv6_lookup to 
 * get lookup result from RTM.
 * 
 * @Param:  - IN arg_IpAddr: network IPv6 address to be looked up.
 * @Param:  - IN arg_pContext: pointer to IPv6 context
 * @Param:  - OUT arg_pNextHopInfo - next hop information result for the 
 *            given IPv6 address
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - null pinter
 * IX_CC_IPV6_ERROR_RTM - RTM internal error
 * 
 */

ix_error ix_cc_ipv6_lookup_prefix(ix_uint128 arg_IpAddr,
                                 ix_cc_rtmv6_next_hop_info *arg_pNextHopInfo,
                                 void *arg_pContext)
{
    ix_uint32 errCode;
    ix_error err = IX_SUCCESS;

    /* IPv6 address is in Host Byte Order */
    IX_ERROR_C(ix_cc_rtmv6_lookup(((ix_cc_ipv6_context*)arg_pContext)->hRtm,
                                   arg_IpAddr,
                                   arg_pNextHopInfo), err);

    if(err != IX_SUCCESS)
    {
        errCode = IX_ERROR_GET_CODE(err);
        if(errCode == IX_CC_ERROR_NULL)
        {
            return err;
        }/* end if (err == IX_CC_ERROR_NULL) */
        else
        {
            return IX_ERROR_SET_CODE(err, IX_CC_IPV6_ERROR_RTM); 
        }/* end else(err == IX_CC_ERROR_NULL) */
    }/* end if (err != IX_SUCCESS) */     

 
    return err;

} /* end ix_cc_ipv6_lookup_prefix */

/**
 * NAME: ix_cc_ipv6_purge_prefixes
 *
 * DESCRIPTION: This primitive helps a client to remove all routes from RTM.
 * 
 * @Param:  - IN arg_pContext: pointer to IPv6 context
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - null pointer
 * IX_CC_ERROR_ALIGN - pointer not aligned properly
 * IX_CC_IPV6_ERROR_RTM - RTM failure
 */

ix_error ix_cc_ipv6_purge_prefixes(void *arg_pContext)
{
    IX_ERROR_CR(ix_cc_rtmv6_purge_routes(
                               ((ix_cc_ipv6_context*)arg_pContext)->hRtm));
    
    /* Add invalid routes */
   IX_ERROR_CR(_ix_cc_ipv6_add_invalid_routes());

   return IX_SUCCESS;

} /* end ix_cc_ipv6_purge_prefixes */

/**
 * NAME: ix_cc_ipv6_dump_prefixes
 *
 * DESCRIPTION: This primitive helps to dump all routes of RTMv6 in memory
 * Client shall use IX_CC_RTMV6_DUMP_ROUTE_SIZE to determine how many 
 * bytes to allocate.
 * 
 * @Param:  - IN arg_pBuffer - pointer to a block of memory where routes 
 *            are to be stored.
 * @Param:  - IN arg_Size - size of buffer in bytes.
 * @Param:  - IN arg_pContext: pointer to IPv6 context
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - null pointer
 * IX_CC_ERROR_ALIGN - pointer not aligned properly
 * IX_CC_ERROR_FULL -  buffer is too small
 */

ix_error ix_cc_ipv6_dump_prefixes(char* arg_pBuffer,
                                ix_uint32 arg_Size,
                                void *arg_pContext)
{

    if(arg_pBuffer == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                  ("Null Pointer: arg_pNextHopInfo"));
 
    }/* end if (err != IX_SUCCESS) */  
 
    IX_ERROR_CR(ix_cc_rtmv6_dump_routes(
                                 ((ix_cc_ipv6_context*)arg_pContext)->hRtm,
                                 arg_pBuffer,
                                 arg_Size));
    return IX_SUCCESS;
} /* end ix_cc_ipv6_dump_prefixes */


/**
 * NAME: ix_cc_ipv6_add_next_hop
 *
 * DESCRIPTION: This primitive helps a client to add next hop information 
 * to RTM database. Client needs to add next hop to RTM in order to 
 * add routes referring to that next hop. 
 * 
 * 
 * @Param:  - IN arg_NextHopId: identifier of next hop to be added.
 * @Param:  - IN arg_pNextHopData: defined by RTM.
 * @Param:  - IN arg_pContext: pointer to IPv6 context
 *
 * @Param:  - OUT  none 
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_ERROR_DUPLICATE_ENTRY - next hop was already added to this NHID
 * IX_CC_ERROR_CONFLICTING_ENTRY - next hop was already added to another NHID
 * IX_CC_IPV6_ERROR_RTM - RTM internal failure
 * IX_CC_IPV6_ERROR_INVALID_NH_INFO - invalid next hop information
 */

ix_error ix_cc_ipv6_add_next_hop(ix_cc_rtmv6_nhid arg_NextHopId,
                                 ix_uint32 arg_NumNextHops,
                                 ix_cc_rtmv6_next_hop_info* arg_pNextHopInfo,
                                 void *arg_pContext)
{
    ix_uint32 errCode;
    ix_error err = IX_SUCCESS;

    if(arg_pNextHopInfo == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                  ("Null Pointer: arg_pNextHopInfo"));
 
    }/* end if(arg_pNextHopInfo == NULL) */     

    /* Make sure that the number of next hops is in range (1 -- 4) */
    if ((arg_NumNextHops > 4) || (arg_NumNextHops < 1))
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
                    ("Valid range for number of Next Hops is 1 - 4 \n"));
    }
 
    IX_ERROR_C(ix_cc_rtmv6_add_next_hop(
                                ((ix_cc_ipv6_context*)arg_pContext)->hRtm,
                                arg_NextHopId,
				arg_NumNextHops,
                                arg_pNextHopInfo),
                            err);

    if(err != IX_SUCCESS)
    {
        errCode = IX_ERROR_GET_CODE(err);
        switch(errCode)
        {
            case IX_CC_RTMV6_ERROR_INVALID_NH_INFO:
                return IX_ERROR_SET_CODE(err, IX_CC_IPV6_ERROR_INVALID_NH_INFO);
                break;
            case IX_CC_ERROR_DUPLICATE_ENTRY:
            case IX_CC_ERROR_CONFLICTING_ENTRY:
                return err;
                break;
            default:
                return IX_ERROR_SET_CODE(err, IX_CC_IPV6_ERROR_RTM);
                break;

        }/* switch (err) */
    }/* end if (err != IX_SUCCESS) */     

    return err;
} /* end ix_cc_ipv6_add_next_hop */

/**
 * NAME: ix_cc_ipv6_delete_next_hop
 *
 * DESCRIPTION: This primitive helps a client to delete next hop information 
 * from RTM database. Client needs to remove all routes associated with 
 * next hop before calling this primitive. If not, client gets an error.
 * This API internally calls ix_cc_rtmv6_delete_next_hop API of RTMv6
 * to delete next hop information from RTMv6.
 * 
 * @Param:  - IN arg_NextHopId: identifier of next hop to be removed.
 * @Param:  - IN arg_pContext: pointer to IPv6 context
 * @Param:  - OUT -none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_IPV6_ERROR_RTM - RTM internal failure
 * IX_CC_ERROR_ENTRY_NOT_FOUND - invalid next hop id
 * IX_CC_IPV6_ERROR_NEXT_HOP_ID_IN_USE -next hop id is in use
 */

ix_error ix_cc_ipv6_delete_next_hop(ix_cc_rtmv6_nhid arg_NextHopId,
                                    void *arg_pContext)
{
    ix_uint32 errCode;
    ix_error err = IX_SUCCESS;


    IX_ERROR_C(ix_cc_rtmv6_delete_next_hop(
                                ((ix_cc_ipv6_context*)arg_pContext)->hRtm,
                                arg_NextHopId),
                                err);

    if(err != IX_SUCCESS)
    {
        errCode = IX_ERROR_GET_CODE(err);
        switch(errCode)
        {
            case IX_CC_ERROR_ENTRY_NOT_FOUND:
                return err;
                break;
            case IX_CC_RTMV6_ERROR_NHID_IN_USE:
                return IX_ERROR_SET_CODE(err, 
                                    IX_CC_IPV6_ERROR_NEXT_HOP_ID_IN_USE);
                break;
            default:
                return IX_ERROR_SET_CODE(err, IX_CC_IPV6_ERROR_RTM);
                break;

        }/* switch (err) */
    }/* end if (err != IX_SUCCESS) */     

    return err;

} /* end ix_cc_ipv6_delete_next_hop */

/**
 * NAME: ix_cc_ipv6_update_next_hop
 *
 * DESCRIPTION: This primitive helps a client to update next hop information 
 * to RTM database.
 * 
 * @Param:  - IN arg_NextHopId: identifier of next hop to be updated.
 * @Param:  - IN arg_pNextHopInfo: next hop information to be updated
 * @Param:  - IN arg_pContext: pointer to IPv6 context
 * @Param:  - OUT  none 
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_IPV6_ERROR_RTM - RTM internal failure
 * IX_CC_IPV6_ERROR_INVALID_NH_INFO - invalid next hop information
 * IX_CC_ERROR_ENTRY_NOT_FOUND - invalid next hop id
 */

ix_error ix_cc_ipv6_update_next_hop(ix_cc_rtmv6_nhid arg_NextHopId,
                               ix_uint32 arg_numNextHops,
                               ix_cc_rtmv6_next_hop_info* arg_pNextHopInfo,
                               void *arg_pContext)
{
    ix_uint32 errCode;
    ix_error err = IX_SUCCESS;
    
    if(arg_pNextHopInfo == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                  ("Null Pointer: arg_pNextHopInfo"));
 
    }/* end if(arg_pNextHopInfo == NULL) */     
 
    IX_ERROR_C(ix_cc_rtmv6_update_next_hop(
                                   ((ix_cc_ipv6_context*)arg_pContext)->hRtm,
                                   arg_NextHopId,
                                   arg_pNextHopInfo),
                    err);

    if(err != IX_SUCCESS)
    {
        errCode = IX_ERROR_GET_CODE(err);
        switch(errCode)
        {
            case IX_CC_RTMV6_ERROR_INVALID_NH_INFO:
                return IX_ERROR_SET_CODE(err, IX_CC_IPV6_ERROR_INVALID_NH_INFO);
                break;
            case IX_CC_ERROR_ENTRY_NOT_FOUND:
                return err;
                break;
            default:
                return IX_ERROR_SET_CODE(err, IX_CC_IPV6_ERROR_RTM);
                break;

        }/* switch (err) */
    }/* end if (err != IX_SUCCESS) */     

    return err;
} /* end ix_cc_ipv6_update_next_hop */

/**
 * NAME: ix_cc_ipv6_get_next_hop
 *
 * DESCRIPTION: This library primitive helps to retrieve next hop information 
 * from RTMv6. This API internally calls ix_cc_rtmv6_get_next_hop API of 
 * RTM core component to retrieve next hop information. 
 * 
 * @Param:  - IN arg_NextHopId: Identifier of the next hop information to 
 *            be retrieved from RTMv6.
 * @Param:  - IN arg_pContext: pointer to IPv6 context
 * @Param:  - OUT arg_pNextHopInfo: Result will be placed into this structure
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_IPV6_ERROR_RTM - RTM internal failure
 * IX_CC_ERROR_ENTRY_NOT_FOUND - invalid next hop id
 * IX_CC_ERROR_NULL - null pointer
 */

ix_error ix_cc_ipv6_get_next_hop(ix_cc_rtmv6_nhid arg_NextHopId,
                                 ix_cc_rtmv6_next_hop_info* arg_pNextHopInfo,
                                 void *arg_pContext
                                )
{
    ix_uint32 errCode;
    ix_error err = IX_SUCCESS;

    if(arg_pNextHopInfo == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                  ("Null Pointer: arg_pNextHopInfo"));
 
    }/* end if(arg_pNextHopInfo == NULL) */     

    IX_ERROR_C(ix_cc_rtmv6_get_next_hop(
                                   ((ix_cc_ipv6_context*)arg_pContext)->hRtm,
                                   arg_NextHopId,
                                   arg_pNextHopInfo),
                    err);

  
    if(err != IX_SUCCESS)
    {
        errCode = IX_ERROR_GET_CODE(err);
        if(errCode == IX_CC_ERROR_ENTRY_NOT_FOUND)
        {
            return err;
        }/* end if (err == IX_CC_RTMV6_ERROR_INVALID_NHID) */
        else
        {
            return IX_ERROR_SET_CODE(err, IX_CC_IPV6_ERROR_RTM);
 
        }/* end else(err == IX_CC_RTMV6_ERROR_INVALID_NHID) */
    }/* end if (err != IX_SUCCESS) */     
    return err;

} /* end ix_cc_ipv6_get_next_hop */


/**
 * NAME: ix_cc_ipv6_purge_rtm
 *
 * DESCRIPTION: This primitive helps a client to remove all routes 
 * and next hops from RTMv6 database.
 * 
 * @Param:  - IN arg_pContext: pointer to IPv6 context
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_IPV6_ERROR_RTM - RTM failure
 * 
 */

ix_error ix_cc_ipv6_purge_rtm(void *arg_pContext)
{

    IX_ERROR_CRT(ix_cc_rtmv6_purge(((ix_cc_ipv6_context*)arg_pContext)->hRtm),
                 IX_CC_IPV6_ERROR_RTM, 0);

    /* Add reserved next hops */
    IX_ERROR_CR(_ix_cc_ipv6_add_reserved_next_hops());

    /* Add reserved prefixes */
    IX_ERROR_CR(_ix_cc_ipv6_add_reserved_prefixes());
    /* Add invalid routes */
    IX_ERROR_CR(_ix_cc_ipv6_add_invalid_routes());

    return IX_SUCCESS;
} /* end ix_cc_ipv6_purge_rtm */


/**
 * NAME: ix_cc_ipv6_dump_next_hops
 *
 * DESCRIPTION: This API internally calls ix_cc_rtmv6_dump_next_hop API of 
 * RTMv6 core component. Client shall use IX_CC_RTMV6_DUMP_NEXT_HOP_SIZE 
 * to determine how many bytes to allocate.
 * 
 * @Param:  - IN arg_pBuffer - pointer to a block of memory where next hops 
 *            are to be stored.
 * @Param:  - IN arg_Size - size of buffer in bytes
 * @Param:  - IN arg_pContext: pointer to IPv6 context
 *
 * @Param:  - OUT none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - null pointer
 * IX_CC_ERROR_ALIGN - pointer not aligned properly
 * IX_CC_ERROR_FULL -  buffer is too small
 * 
 */

ix_error ix_cc_ipv6_dump_next_hops(char* arg_pBuffer,
                                   ix_uint32 arg_Size,
                                   void *arg_pContext)
{
    if(arg_pBuffer == NULL)
    {
       return IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("Null pointer: arg_pBuffer"));
    }/* end if (err != IX_SUCCESS) */     

 
    IX_ERROR_CR(ix_cc_rtmv6_dump_next_hops(
                                     ((ix_cc_ipv6_context*)arg_pContext)->hRtm,
                                     arg_pBuffer,
                                     arg_Size));
      return IX_SUCCESS;
} /* end ix_cc_ipv6_dump_next_hops */

/**
 * NAME: ix_cc_ipv6_set_mtu
 *
 * DESCRIPTION: This primitive helps client to update MTU for a given next 
 * hop. Before calling this, client needs to call ix_cc_ipv6_add_next_hop 
 * to add next hop identifier in RTMv6 database. The validity of MTU 
 * shall be done by the client.
 * 
 * @Param:  - IN arg_NextHopId: identifier of next hop.
 * @Param:  - IN arg_Mtu: new maximum transmission unit for the given next hop.
 * @Param:  - IN arg_pContext: pointer to IPv6 context
 * @Param:  - OUT none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *
 * IX_CC_IPV6_ERROR_RTM - RTM failure
 * IX_CC_ERROR_ENTRY_NOT_FOUND - invalid next hop id 
 * IX_CC_IPV6_ERROR_INVALID_NH_INFO - invalid mtu
 */

ix_error ix_cc_ipv6_set_mtu(ix_cc_rtmv6_nhid arg_NextHopId,
                            ix_uint32 arg_Mtu,
                            void *arg_pContext)
{
    ix_uint32 errCode;
    ix_error err = IX_SUCCESS;


    err = ix_cc_rtmv6_set_mtu(((ix_cc_ipv6_context*)arg_pContext)->hRtm,
                                 arg_NextHopId,
                                 arg_Mtu);
                             

  
    if(err != IX_SUCCESS)
    {
        errCode = IX_ERROR_GET_CODE(err);
        switch(errCode)
        {
            case IX_CC_ERROR_ENTRY_NOT_FOUND:
            return err;
            break;
            case IX_CC_RTMV6_ERROR_INVALID_NH_INFO:
            return IX_ERROR_SET_CODE(err,IX_CC_IPV6_ERROR_INVALID_NH_INFO);
            break;
            default:
            return IX_ERROR_SET_CODE(err,IX_CC_IPV6_ERROR_RTM);
            break; 
        }/* end switch(err) */
    }/* end if (err != IX_SUCCESS) */     
    return err;

} /* end ix_cc_ipv6_set_mtu  */

/**
 * NAME: ix_cc_ipv6_set_flags
 *
 * DESCRIPTION:This primitive helps client to update flags for a given 
 * next hop. Before calling this primitive, client needs to call 
 * ix_cc_ipv6_add_next_hop to add next hop identifier in RTM database. 
 * 
 * @Param:  - IN arg_NextHopId: identifier of next hop.
 * @Param:  - IN arg_Flags: new flags for the given next hop 
 * @Param:  - IN arg_pContext: pointer to IPv6 context
 *
 * @Param:  - OUT none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_IPV6_ERROR_RTM - RTM failure
 * IX_CC_ERROR_ENTRY_NOT_FOUND - invalid next hop id 
 * IX_CC_IPV6_ERROR_INVALID_FLAGS - invalid flags 
 * 
 */

ix_error ix_cc_ipv6_set_flags(ix_cc_rtmv6_nhid arg_NextHopId,
                              ix_uint32 arg_Flags,
                              void *arg_pContext
                             )
{
    ix_uint32 errCode;
    ix_error err = IX_SUCCESS;

    if((arg_Flags !=IPV6_NH_FLAGS_LOCAL) &&
      (arg_Flags !=IPV6_NH_FLAGS_DOWN) &&
      (arg_Flags !=IPV6_NH_FLAGS_DROP))
    {
        err = IX_ERROR_WARNING(IX_CC_IPV6_ERROR_INVALID_FLAGS,
                                       ("Invalid flags"));
        return err;
    }/* end if flags bit */


    err = ix_cc_rtmv6_set_flags(((ix_cc_ipv6_context*)arg_pContext)->hRtm,
                                  arg_NextHopId,
                                  arg_Flags);

    if(err != IX_SUCCESS)
    {
        errCode = IX_ERROR_GET_CODE(err);
        if(errCode == IX_CC_ERROR_ENTRY_NOT_FOUND)
        {
            return err;
        }/* end if (err == IX_CC_RTMV6_ERROR_INVALID_NHID) */
        else
        {
            return IX_ERROR_SET_CODE(err,IX_CC_IPV6_ERROR_RTM); 
        }/* end else(err == IX_CC_RTMV6_ERROR_INVALID_NHID) */
    }/* end if (err != IX_SUCCESS) */     
    return err;

} /* end ix_cc_ipv6_set_flags */



/**
 * NAME: ix_cc_ipv6_get_rate_limit_time
 *
 * DESCRIPTION: This library primitive helps to retrieve the number of 
 * seconds between when the event will fire causing the ICMP messages to 
 * be dequeued and sent. The lower the value of arg_pSleepTime, the more 
 * frequently the ICMP messages will be dequeued and sent.  
 * 
 * @Param:  - IN arg_pContext: pointer to IPv6 context
 *
 * @Param:  - OUT arg_pSleepTime: ICMP sleep period 
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - null pointer
 * 
 */

ix_error ix_cc_ipv6_get_rate_limit_time(ix_uint16* arg_pSleepTime,
                                        void *arg_pContext)
{
    
    if(arg_pSleepTime == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid Pointer: arg_pSleepTime")); 
    }/* end if (arg_pSleepTime == NULL) */

    return _ix_cc_ipv6_icmp_get_rate_limit(arg_pSleepTime);
} /* end ix_cc_ipv6_get_rate_limit_time  */



/**
 * NAME: ix_cc_ipv6_set_rate_limit_time
 *
 * DESCRIPTION: This library primitive helps to set the period between 
 * events firing to process the ICMP error message queue. The lower the 
 * value of arg_SleepTime, the more frequently the output queue will 
 * be processed. The minimum value of arg_SleepTime is 1 sec and maximum 
 * value can be 32sec.
 * 
 * @Param:  - IN arg_SleepTime: the number of seconds to wait between processing
 * the ICMP error message queue.
 * @Param:  - IN arg_pContext: pointer to IPv6 context
 *
 * @Param:  - OUT none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_IPV6_INVALID_RATE_LIMIT_TIME - invalid sleep time
 * 
 */

ix_error ix_cc_ipv6_set_rate_limit_time(ix_uint16 arg_SleepTime,
                                        void *arg_pContext)
{
    return _ix_cc_ipv6_icmp_set_rate_limit(arg_SleepTime);
} /* end ix_cc_ipv6_set_rate_limit_time */

/**
 * NAME: ix_cc_ipv6_get_queue_depth
 *
 * DESCRIPTION: This library primitive helps to retrieves the depth of 
 * the ICMP error message queue. That is, the maximum number of ICMP 
 * messages that can fit in the queue, not the number of messages currently 
 * in the queue.
 * 
 * @Param:  - IN arg_pContext: pointer to IPv6 context
 *
 * @Param:  - OUT arg_pQueueDepth: the current number of entries 
 * that the ICMP error message queue can hold.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - null pointer
 */

ix_error ix_cc_ipv6_get_queue_depth(ix_uint16* arg_pQueueDepth,
                                    void *arg_pContext)
{
    if(arg_pQueueDepth == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pQueueDepth"));
    }/* end if(arg_pQueueDepth == NULL)  */

    *arg_pQueueDepth = g_Icmpv6QueueDepth;

    return IX_SUCCESS;

} /* end ix_cc_ipv6_get_queue_depth  */

/**
 * NAME: ix_cc_ipv6_get_packets_to_drain
 *
 * DESCRIPTION:This library primitive helps to retrieve the maximum number 
 * of ICMP error messages to send (from the ICMP message queue) each 
 * time the queue processing event fires. 
 * 
 * @Param:  - IN arg_pContext: pointer to IPv6 context
 * @Param:  - OUT arg_pPacketsToDrain: the maximum number of packets to 
 * send each time the event fires for processing the queue.  
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * 
 */

ix_error ix_cc_ipv6_get_packets_to_drain(ix_uint16* arg_pPacketsToDrain,
                                         void *arg_pContext)
{
    if(arg_pPacketsToDrain == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pPacketsToDrain "));
    }/* end if(arg_pPacketsToDrain == NULL)  */

    *arg_pPacketsToDrain = g_Icmpv6QueueDrain;

    return IX_SUCCESS;

} /* end ix_cc_ipv6_get_packets_to_drain  */

/**
 * NAME: ix_cc_ipv6_set_packets_to_drain
 *
 * DESCRIPTION: This library primitive helps to set the maximum number 
 * of ICMP error messages to send each time the queue processing event fires. 
 * The value of arg_PacketsToDrain must not exceed the depth of the queue.
 * 
 * @Param:  - IN arg_PacketsToDrain: The maximum number of packets to 
 *            send each time the queue processing event fires. 
 * @Param:  - IN arg_pContext: pointer to IPv6 context
 *
 * @Param:  - OUT none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_IPV6_ERROR_INVALID_PACKETS_TO_DRAIN - invalid value
 * 
 */

ix_error ix_cc_ipv6_set_packets_to_drain(ix_uint16 arg_PacketsToDrain,
                                         void *arg_pContext)
{
    if ((arg_PacketsToDrain < 1) || (arg_PacketsToDrain > g_Icmpv6QueueDepth))
    {
        
        return IX_ERROR_WARNING(IX_CC_IPV6_ERROR_INVALID_PACKETS_TO_DRAIN,
            ("ICMP: ix_cc_ipv6_set_packets_to_drain invalid parameter: %d\n",
	                arg_PacketsToDrain));
        
    }

    ix_ossl_mutex_lock(((ix_cc_ipv6_context*)arg_pContext)->ipv6Mutex,
                       IX_OSSL_WAIT_FOREVER);

    g_Icmpv6QueueDrain = arg_PacketsToDrain;

    ix_ossl_mutex_unlock(((ix_cc_ipv6_context*)arg_pContext)->ipv6Mutex);
    
    return IX_SUCCESS;

} /* end ix_cc_ipv6_set_packets_to_drain */

/**
 * NAME: ix_cc_ipv6_get_statistics
 *
 * DESCRIPTION: This library primitive helps to retrieve statistics report 
 * from IPV6 Forwarder core component. This primitive will give the statistics 
 * of IPV6 microblock, RTM and IPV6 Forwarder core component. Client needs 
 * to allocate memory to hold data.
 * 
 * @Param:  - IN arg_pContext: pointer to IPv6 context
 *
 * @Param:  - OUT arg_pBuffer: pointer to memory to hold statistics data.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - null pointer
 * IX_CC_IPV6_ERROR_RTM  - RTM failure
 * IX_CC_ERROR_INTERNAL - internal failure
 */

ix_error ix_cc_ipv6_get_statistics(ix_cc_ipv6_stats_data *arg_pBuffer,
                                   void *arg_pContext)
{
#ifdef IPV6_STATS
    ix_cc_rtmv6_statistics rtmStats;
    ix_uint32 index, portIndex, bladeIndex, counter;
    ix_uint32 baseAddress, outBaseAddress;
    ix_uint64 inMbStats[IX_CC_IPV6_NUMBER_OF_IN_MB_COUNTERS];
    ix_uint64 outMbStats[IX_CC_IPV6_NUMBER_OF_OUT_MB_COUNTERS];
    ix_error err = IX_SUCCESS;
    ix_cc_ipv6_context *pContext;

    ix_cc_in_ipv6_stats_data *pInStat;
    ix_cc_out_ipv6_stats_data *pOutStats;
    ix_uint32 outCounterChunkSize;

    if(arg_pBuffer == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pBuffer "));
    }/* end if(arg_pBuffer == NULL)  */

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pContext "));
    }/* end if(arg_pContext == NULL)  */

    pContext = (ix_cc_ipv6_context *)arg_pContext;

    /* 
     * Incoming Statistics first 
     */

    for (portIndex = 0; portIndex < MAX_NUMBER_OF_PORTS_PER_BLADE; portIndex++)
    {
       baseAddress = (ix_uint32)((ix_uint32 *)pContext->pMbStatsBase + 
                       (portIndex * IX_CC_IPV6_TOTAL_NUMBER_OF_IN_MB_COUNTERS));

        pInStat = &arg_pBuffer->ipv6InStats[portIndex];

	for (counter = 0; counter < IX_CC_IPV6_NUMBER_OF_IN_MB_COUNTERS; 
	                                                           counter++)
	{
            err = ix_rm_counter_64bit_get_value(
                  pContext->inCounter64BitHandle[portIndex].handles[counter],
                  &inMbStats[counter]);

            if(err != IX_SUCCESS)
            {
                return IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
                                    ("Failed :ix_rm_counter_64bit_get_value"));
            }/* end if (err != IX_SUCCESS) */ 
	}

	/* Microblock maintained 64-bit incoming counters */
	pInStat->ipv6MbInReceives = inMbStats[0] +
	      IX_RM_MEM_UINT32_READ((ix_uint32 *)(baseAddress + IPV6_IN_RECV)) +
	      pContext->ipv6FwdStats.ipv6InCore[portIndex].ipv6CoreInRecv;

	pInStat->ipv6MbInOctets = inMbStats[1] + 
            IX_RM_MEM_UINT32_READ((ix_uint32 *)(baseAddress + IPV6_IN_OCTETS)) +
	    pContext->ipv6FwdStats.ipv6InCore[portIndex].ipv6CoreInOctets;

	pInStat->ipv6MbInForwDatagrams = 
              IX_RM_MEM_UINT32_READ((ix_uint32 *)(baseAddress + IPV6_IN_FORW)) +
	      pContext->ipv6FwdStats.ipv6InCore[portIndex].ipv6CoreInForwDgrms;

	pInStat->ipv6MbInAddrErrors = 
       IX_RM_MEM_UINT32_READ((ix_uint32 *)(baseAddress + IPV6_IN_ADDR_ERRORS)) +
	    pContext->ipv6FwdStats.ipv6InCore[portIndex].ipv6CoreInAddrErrors;

	pInStat->ipv6MbInTruncated = 
         IX_RM_MEM_UINT32_READ((ix_uint32 *)(baseAddress + IPV6_IN_TRUNCATED)) +
	   pContext->ipv6FwdStats.ipv6InCore[portIndex].ipv6CoreInTruncatedPkts;

	/* 
	 * This counter is updated both by the microblock and the core,
	 * so add the two values together.
	 */
	pInStat->ipv6CoreMbInHdrErrors = 
         IX_RM_MEM_UINT32_READ((ix_uint32 *)(baseAddress + IPV6_IN_HDR_ERRORS));

	pInStat->ipv6CoreMbInHdrErrors += 
	  pContext->ipv6FwdStats.ipv6InCore[portIndex].ipv6CoreInHdrErrors;
	  

        /* Core maintained incoming stats */
	pInStat->ipv6CoreInNoRoutes = 
	  pContext->ipv6FwdStats.ipv6InCore[portIndex].ipv6CoreInNoRoutes;
	 
	pInStat->ipv6CoreInUnknownProtos = 
	  pContext->ipv6FwdStats.ipv6InCore[portIndex].ipv6CoreInUnknownProtos;

	pInStat->ipv6CoreInDiscards = 
	  pContext->ipv6FwdStats.ipv6InCore[portIndex].ipv6CoreInDiscards;

	pInStat->ipv6CoreInMulticast = 
	  pContext->ipv6FwdStats.ipv6InCore[portIndex].ipv6CoreInMulticast;

	pInStat->ipv6CoreInMulticastOctets = 
	pContext->ipv6FwdStats.ipv6InCore[portIndex].ipv6CoreInMulticastOctets;
    }
     
    /* 
     * Outgoing Statistics
     */

    outBaseAddress = (ix_uint32)((ix_uint32 *)pContext->pMbStatsBase + 
              (IX_CC_IPV6_TOTAL_NUMBER_OF_IN_MB_COUNTERS * 
	                                   MAX_NUMBER_OF_PORTS_PER_BLADE));
    
    outCounterChunkSize = (IX_CC_IPV6_TOTAL_NUMBER_OF_OUT_MB_COUNTERS * 
                                                       sizeof (ix_uint32));

    for (bladeIndex = 0; bladeIndex < MAX_NUMBER_OF_BLADES; bladeIndex++)
    {
        /* Get out stats for all ports on this bladeIndex */
        for (portIndex = 0; portIndex < MAX_NUMBER_OF_PORTS_PER_BLADE; 
	                                                     portIndex++)
	{
	    index = (MAX_NUMBER_OF_PORTS_PER_BLADE * bladeIndex) + portIndex;

            pOutStats = &arg_pBuffer->ipv6OutStats[index];

	    /* base address of out stats for this {blade/port} index */
	    baseAddress = outBaseAddress + (outCounterChunkSize * index);

	    for (counter = 0; counter < IX_CC_IPV6_NUMBER_OF_OUT_MB_COUNTERS;
	                                                    counter++)
            {
                err = ix_rm_counter_64bit_get_value(
                       pContext->outCounter64BitHandle[index].handles[counter],
                       &outMbStats[counter]);

                if(err != IX_SUCCESS)
                {
                    return IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
                                    ("Failed :ix_rm_counter_64bit_get_value"));
                }/* end if (err != IX_SUCCESS) */ 
	    }

	    /* Fill out the out counters */
	    pOutStats->ipv6MbOutForwDatagrams = 
             IX_RM_MEM_UINT32_READ((ix_uint32 *)(baseAddress + IPV6_OUT_FORW)) +
	      pContext->ipv6FwdStats.ipv6OutCore[index].ipv6CoreOutForwDgrms;

	    pOutStats->ipv6MbOutTransmits = outMbStats[0] +
        IX_RM_MEM_UINT32_READ((ix_uint32 *)(baseAddress + IPV6_OUT_TRANSMITS)) +
	      pContext->ipv6FwdStats.ipv6OutCore[index].ipv6CoreOutTransmits;
							      

	    pOutStats->ipv6MbOutOctets = outMbStats[1] +
          IX_RM_MEM_UINT32_READ((ix_uint32 *)(baseAddress + IPV6_OUT_OCTETS)) +
	       pContext->ipv6FwdStats.ipv6OutCore[portIndex].ipv6CoreOutOctets;


	    pOutStats->ipv6CoreOutRequests = 
	                          pContext->ipv6FwdStats.ipv6CoreOutRequests;

	    pOutStats->ipv6CoreOutNoRoutes = 
	                          pContext->ipv6FwdStats.ipv6CoreOutNoRoutes;

	    pOutStats->ipv6CoreOutDiscards = 
	       pContext->ipv6FwdStats.ipv6OutCore[index].ipv6CoreOutDiscards;

	    pOutStats->ipv6CoreOutMulticast = 
	       pContext->ipv6FwdStats.ipv6OutCore[index].ipv6CoreOutMulticast;

	    pOutStats->ipv6CoreOutMulticastOctets = 
          pContext->ipv6FwdStats.ipv6OutCore[index].ipv6CoreOutMulticastOctets;
        }
    }

    /* Get RTM statistics */
    err = ix_cc_rtmv6_get_statistics(pContext->hRtm, &rtmStats);

    if(err != IX_SUCCESS)
    {
        return IX_ERROR_WARNING(IX_CC_IPV6_ERROR_RTM,
                                ("Failed :ix_cc_rtmv6_get_statistics"));
    }/* end if (err != IX_SUCCESS) */ 

    /* Fillup buffer with RTM statistics */
    arg_pBuffer->ipv6NumberOfRoutes = rtmStats.numberOfRoutes;
    arg_pBuffer->ipv6NumberOfNextHops = rtmStats.numberOfNextHops;

    return IX_SUCCESS;
#else 

    return IX_CC_ERROR_UNIMPL;

#endif /* IPV6_STATS */

} /* end ix_cc_ipv6_get_statistics  */

/**
 * NAME: ix_cc_ipv6_set_property
 *
 * DESCRIPTION: This library primitive helps to set dynamic properties of 
 * IPv6 Forwarder core component. 
 * 
 * @Param:  - IN arg_PropId: Identifier of a property or properties(ORed) 
 *            to be updated for a port
 * @Param:  - IN arg_pPropData: pointer to dynamic property data structure
 * @Param:  - IN arg_pContext: pointer to IPv6 context
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_NUL - invalid pointer
 * IX_ERROR_FULL - table is full
 * IX_CC_ERROR_ENTRY_NOT_FOUND - entry not found in table
 * IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM - inavlid property id
 */

ix_error ix_cc_ipv6_set_property(ix_uint32 arg_PropId,
                                 ix_cc_properties* arg_pPropData,
                                 void *arg_pContext)
{
    ix_uint128 ipv6Addr;
    ix_uint32 loop_var;

    /* Validate arg_pPropData */
    if(arg_pPropData == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                ("Invalid pointer: arg_pPropData"));
    }/* end if (arg_PropData == NULL) */  

    if(((arg_PropId & IX_CC_SET_PROPERTY_ADD_INTERFACE_IPV6) != 
       IX_CC_SET_PROPERTY_ADD_INTERFACE_IPV6) &&
       ((arg_PropId & IX_CC_SET_PROPERTY_DEL_INTERFACE_IPV6) != 
       IX_CC_SET_PROPERTY_DEL_INTERFACE_IPV6) &&
       ((arg_PropId & IX_CC_SET_PROPERTY_PHYSICAL_IF_STATUS) != 
       IX_CC_SET_PROPERTY_PHYSICAL_IF_STATUS))
    {
	/* Unexpected, or unknown property ID */
        return IX_ERROR_WARNING(IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM,
                                          ("Invalid property Id"));

    }/* end if invalid property */ 


    for (loop_var = 0;
         loop_var < arg_pPropData->no_virtual_if;
         loop_var++)
    {
        /* Check if the present address is IPv6 address */
        if (arg_pPropData->ip_prop[loop_var].ip_version
                != IX_CC_IP_VERSION_IPV6)
        {
            continue;
        }

        /* Convert IP Address into network byte order */
        IX_HTON128(&arg_pPropData->ip_prop[loop_var].protocol.ipv6_prop.ipv6_address.addr,
            &ipv6Addr);

        /* arg_PropId can have multiple properties */  
        if((arg_PropId & IX_CC_SET_PROPERTY_ADD_INTERFACE_IPV6) == 
            IX_CC_SET_PROPERTY_ADD_INTERFACE_IPV6)
        {
            IX_ERROR_CR(_ix_cc_ipv6_property_update_ip_address(
                            IX_CC_SET_PROPERTY_ADD_INTERFACE_IPV6,
                            arg_pPropData->blade_id,
                            arg_pPropData->port_id,
                            ipv6Addr));
        }/* end if add ipv6 address */

        if((arg_PropId & IX_CC_SET_PROPERTY_DEL_INTERFACE_IPV6) == 
            IX_CC_SET_PROPERTY_DEL_INTERFACE_IPV6)
        {
            IX_ERROR_CR(_ix_cc_ipv6_property_update_ip_address(
                            IX_CC_SET_PROPERTY_DEL_INTERFACE_IPV6,
                            arg_pPropData->blade_id,
                            arg_pPropData->port_id,
                            ipv6Addr));
        }/* end if delete ipv6 address */
    }

    if((arg_PropId & IX_CC_SET_PROPERTY_PHYSICAL_IF_STATUS) == 
                                      IX_CC_SET_PROPERTY_PHYSICAL_IF_STATUS)
    {
        IX_ERROR_CR(_ix_cc_ipv6_property_update_port_status(
	                                   arg_pPropData->port_id,
                                           arg_pPropData->physical_if_state));
    }/* end if delete ipv6 address */
    
    return IX_SUCCESS;
} /* end ix_cc_ipv6_set_property */

/**
 * NAME: ix_cc_ipv6_send_icmp_error
 *
 * DESCRIPTION: This library primitive helps the client to send an ICMP
 * error message.
 *
 * @Param:  - IN arg_IcmpErrType: The ICMP error type to include in the message
 * @Param:  - IN arg_IcmpErrCode: The ICMP error code to include in the message
 * @Param:  - IN arg_hBuffer: The IPv6 packet causing the error.
 * @Param:  - IN arg_Param: Contains extra information if required by the ICMP
 *                          error message type.
 * @Param:  - IN arg_pUserContext: pointer to client defined context.
 *
 * @Param:  - OUT - none
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_IPV6_ERROR_MSG_LIBRARY- failure from Message Support Library
 *
 */
ix_error ix_cc_ipv6_send_icmp_error(
                    ix_cc_ipv6_icmp_err_type    arg_IcmpErrType,
                    ix_cc_ipv6_icmp_err_code    arg_IcmpErrCode,
                    ix_buffer_handle            arg_hBuffer,
                    ix_uint32                   arg_Param,
                    void*                       arg_pUserContext)
{
    ix_uint8    code;
    if( (arg_IcmpErrType < IX_CC_IPV6_ICMP_ERR_FIRST) ||
        (arg_IcmpErrType >= IX_CC_IPV6_ICMP_ERR_LAST))
    {
        return IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM;
    }

    if((arg_IcmpErrCode < IX_CC_IPV6_ICMP_CODE_FIRST) ||
        (arg_IcmpErrCode >= IX_CC_IPV6_ICMP_CODE_LAST))
    {
        return IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM;
    }

    /*Translate the error code according to the specified error type*/
    switch(arg_IcmpErrType)
    {
        case IX_CC_IPV6_ICMP_ERR_DEST_UNREACHABLE:
            switch(arg_IcmpErrCode)
            {
                case IX_CC_IPV6_ICMP_CODE_DU_NO_ROUTE:
                    code = IX_CC_ICMPV6_DU_CODE_NOROUTE;
                case IX_CC_IPV6_ICMP_CODE_DU_COMM_PROHIBITED:
                    code = IX_CC_ICMPV6_DU_CODE_COMM_PROHIBITED;
                case IX_CC_IPV6_ICMP_CODE_DU_ADDR_UNREACHABLE:
                    code = IX_CC_ICMPV6_DU_CODE_ADDR_UNREACHABLE;
                case IX_CC_IPV6_ICMP_CODE_DU_PORT_UNREACHABLE:
                    code = IX_CC_ICMPV6_DU_CODE_PORT_UNREACHABLE;
                default:
                    return IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM;
            }

            IX_ERROR_CR(_ix_cc_ipv6_icmp_dest_unreachable(
                        arg_hBuffer, code));
            break;
        case IX_CC_IPV6_ICMP_ERR_PKT_TOO_BIG:
            if(arg_IcmpErrCode != IX_CC_IPV6_ICMP_CODE_PKT_TOO_BIG)
                return IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM;

            IX_ERROR_CR(_ix_cc_ipv6_icmp_pkt_too_big(
                             arg_hBuffer, (ix_uint8)arg_Param));
            break;
        case IX_CC_IPV6_ICMP_ERR_TIME_EXCEEDED:
            switch(arg_IcmpErrCode)
            {
                case IX_CC_IPV6_ICMP_CODE_TE_HOP_LIMIT_EXCEEDED:
                    code = IX_CC_ICMPV6_TE_CODE_HOP_LIMIT_EXCEEDED;
                case IX_CC_IPV6_ICMP_CODE_TE_REASSEMBLY_EXCEEDED:
                    code = IX_CC_ICMPV6_TE_CODE_REASSEMBLY_EXCEEDED;
                default:
                    return IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM;
            }

            IX_ERROR_CR(_ix_cc_ipv6_icmp_time_exceeded(
                                        arg_hBuffer, code));  
            break;
        case IX_CC_IPV6_ICMP_ERR_PARAM_PROBLEM:
            switch(arg_IcmpErrCode)
            {
                case IX_CC_IPV6_ICMP_CODE_PP_WRONG_HDR_FIELD:
                    code = IX_CC_ICMPV6_PP_CODE_WRONG_HDR_FIELD;
                case IX_CC_IPV6_ICMP_CODE_PP_NEXT_HDR_UNKNOWN:
                    code = IX_CC_ICMPV6_PP_CODE_NEXT_HDR_UNKNOWN;
                case IX_CC_IPV6_ICMP_CODE_PP_OPTION_UNKNOWN:
                    code = IX_CC_ICMPV6_PP_CODE_OPTION_UNKNOWN;
                default:
                    return IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM;
            }

            IX_ERROR_CR(_ix_cc_ipv6_icmp_param_problem(
                            arg_hBuffer, code, arg_Param));
            break;
        default:
            return IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM;
            break;
    }

    return IX_SUCCESS;
}


/**
 * NAME: ix_cc_ipv6_send_icmp_info_message
 *
 * DESCRIPTION: This library primitive helps the client to send an ICMP
 *              informational message.
 *
 * @Param:  - IN arg_SrcIpAddr: The IPv6 address to be included in the source
 *                              address field of the final IPv6 packet.
 * @Param:  - IN arg_DstAddr: The destination IPv6 address.
 * @Param:  - IN arg_Type: The type of the ICMP message. This value will be
 *                          included in the ICMP header.
 * @Param:  - IN arg_Code: The ICMP message code corresponding to the message
 *                          type.
 * @Param:  - IN arg_PayloadLength: The length of the ICMP payload.
 * @Param:  - IN arg_pPayload: The payload to include in the ICMP packet.
 * @Param:  - IN arg_pUserContext: pointer to client defined context.
 *
 * @Param:  - OUT - none
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_IPV6_ERROR_MSG_LIBRARY- failure from Message Support Library
 *
 */
ix_error ix_cc_ipv6_send_icmp_info_message(
                                    ix_uint128  arg_SrcIpAddr,
                                    ix_uint128  arg_DstIpAddr,
                                    ix_uint8    arg_Type,
                                    ix_uint8    arg_Code,
                                    ix_uint32   arg_PayloadLength,
                                    ix_uint8    *arg_pPayload,
                                    void        *arg_pUserContext)
{
    ix_uint16 inPort = 0;

    if(( arg_Type < 128) || (arg_Type >255))
    {
        /*The type is outside the valid ICMP informational message range*/
        return IX_CC_ERROR_RANGE;
    }
            
    if(!arg_pPayload || !arg_pUserContext)
    {
        return IX_CC_ERROR_NULL;
    }

    IX_ERROR_CR(_ix_cc_ipv6_icmp_build_info_message(
                        arg_SrcIpAddr, arg_DstIpAddr, arg_Type, arg_Code,
                        arg_PayloadLength, arg_pPayload, inPort));

    return IX_SUCCESS;
}

/**
 * NAME: ix_cc_ipv6_perform_addr_resolution
 *
 * DESCRIPTION: This library primitive helps the cllient to request
 *              address resolution to be performed for a packet destined
 *              to a neighbor.
 *
 * @Param:  - IN arg_hBuffer: The hardware buffer containing the ethernet
 *                            packet to be sent to the neighbor
 * @Param:  - IN arg_pUserContext: pointer to client defined context.
 *
 * @Param:  - OUT - none
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_IPV6_ERROR_MSG_LIBRARY- failure from Message Support Library
 *
 */
ix_error ix_cc_ipv6_perform_addr_resolution(
                                         ix_buffer_handle arg_hBuffer,
                                         void* arg_pUserContext)
{
    IX_ERROR_CR(ix_cc_nd_perform_addr_resolution(arg_hBuffer,
                                                 arg_pUserContext));

    return IX_SUCCESS;
}

/**
 * NAME: ix_cc_ipv6_add_neighbor
 *
 * DESCRIPTION: This library primitive helps the cllient to add a
 *              neighbor to the L2 table
 *
 * @Param:  - IN arg_L2Index: The index into the L2 table at which this
 *                            information should be added.
 * @Param   - IN arg_NbrIp:   The IPv6 address of the neighbor.
 * @Param   - IN arg_OutputPort: The egress port from which all traffic
 *                            destined for the neighbor should exit.
 * @Param   - IN arg_L2Addr: The MAC address of the neighbor.
 * @Param:  - IN arg_pUserContext: pointer to client defined context.
 *
 * @Param:  - OUT - none
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_IPV6_ERROR_MSG_LIBRARY- failure from Message Support Library
 *
 */
ix_error ix_cc_ipv6_add_neighbor(
                                    ix_uint16 arg_L2Index,
                                    ix_uint128 arg_NbrIp,
                                    ix_uint16 arg_OutputPort,
                                    ix_uint8 *arg_L2Addr,
                                    void* arg_pUserContext)
{
    IX_ERROR_CR(ix_cc_nd_add_neighbor(arg_L2Index,
                                     arg_NbrIp,
                                     arg_OutputPort,
                                     arg_L2Addr,
                                     arg_pUserContext));
    return IX_SUCCESS;
}

/**
 * NAME: ix_cc_ipv6_del_neighbor
 *
 * DESCRIPTION: This library primitive helps the cllient to add a
 *              neighbor to the L2 table
 *
 * @Param:  - IN arg_L2Index: The index into the L2 table at which this
 *                            information should be added.
 * @Param:  - IN arg_pUserContext: pointer to client defined context.
 *
 * @Param:  - OUT - none
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_IPV6_ERROR_MSG_LIBRARY- failure from Message Support Library
 *
 */
ix_error ix_cc_ipv6_del_neighbor(
                                 ix_uint16 arg_L2Index,
                                 void* arg_pUserContext)
{
    IX_ERROR_CR(ix_cc_nd_del_neighbor(arg_L2Index, arg_pUserContext));
    return IX_SUCCESS;
}


/* 
 * The overall architecture of this is as follows:
 *
 * Each "member" structure contains a list of "elements" for each prefix/length
 * a member is interested in. A member is some other CC like (NATPT) that is interested
 * in route updates. A new member structure is created for each new registered member.
 * All the member structures are added to a global linked list. 
 * 
 * Each element (on a member linked list) contains a prefix/prefix_length and the 
 * current_next_hop_id (initialized to whatever the result of a match_prefix was
 * when the element was created). 
 *
 * At the end of ix_cc_ipv6_update_prefix() and ix_cc_ipv6_delete_prefix(), a function
 * is called to "walk" down the list of members and for each member check and see
 * if each elements current_next_hop_id is still current. If it is not (i.e. it got
 * changed either because of a prefix update or a prefix delete), a message is 
 * sent to the member (at the registered message ID). The message contains the 
 * new next hop ID along with the prefix and prefix_length.
 */

/**
 * NAME: ix_cc_ipv6_prefix_update_register
 *
 * DESCRIPTION: This primitive will allow a client to register for prefix update
 * notifications in the future. The client will be returned a handle which needs
 * to be used to inform the IPv6 CC about particular prefixes that this client
 * is interested in hearing about (see ix_cc_prefix_update_notify and
 * ix_cc_prefix_update_cancel).
 *
 * @Param:  - IN arg_msgID: User supplied message ID to which route updates
 * notifications will be sent.
 *
 * @Param:  - OUT arg_pHandle: Handle to be used for notifications/un-registration.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_IPV6_INVALID_INPUT_PARAM :
 * IX_CC_ERROR_OOR : Failed to allocate a resource during registration.
 *
 */
ix_error ix_cc_ipv6_prefix_update_register(ix_uint32 arg_msgID,
                                           route_update_handle *arg_pHandle)
{
    ix_cc_route_update_member *pMember = NULL;

    if (arg_pHandle == NULL)
    {
        return IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM;
    }

    /* Check to see if the global list exists, if it does not, create one now */
    if (!g_RouteUpdateList)
    {
        g_RouteUpdateList = ListCreate();
        if (g_RouteUpdateList == NULL)
        {
	    return IX_CC_ERROR_OOR;
        }
	
	if (ix_ossl_mutex_init(IX_OSSL_MUTEX_UNLOCK, &g_RouteUpdateLock) != IX_SUCCESS)
	{
	    return IX_CC_ERROR_OOR;
	}
    }

    /* Create a structure for this new member and initialize its element list */
    pMember = ix_ossl_malloc(sizeof(ix_cc_route_update_member));
    if (pMember == NULL)
    {
         goto cleanup;
    }

    ix_ossl_memset(pMember, 0, sizeof(ix_cc_route_update_member));

    pMember->prefixes = ListCreate();
    if (pMember->prefixes == NULL)
    {
        goto cleanup;
    }

    pMember->member_msg_id = arg_msgID;

    ix_ossl_mutex_lock(g_RouteUpdateLock, IX_OSSL_WAIT_FOREVER);

    /* Add it to the global list */
    if (!ListAddItem(g_RouteUpdateList, (void *)pMember))
    {
	ix_ossl_mutex_unlock(g_RouteUpdateLock);
        goto cleanup;
    }

    ix_ossl_mutex_unlock(g_RouteUpdateLock);

    /* Return a pointer to this member structure as a handle to the caller */
    *arg_pHandle = (route_update_handle)pMember;

    return IX_SUCCESS;

cleanup:

    ix_ossl_mutex_fini(g_RouteUpdateLock);

    if (pMember)
    {
        ix_ossl_free(pMember);
    }

    if (ListGetCount(g_RouteUpdateList) == 0)
    {
        ListDelete(g_RouteUpdateList);
    }

    return IX_CC_ERROR_OOR;
}

/**
 * NAME: ix_cc_ipv6_prefix_update_unregister
 *
 * DESCRIPTION: This primitive will allow a client to un-register with the IPv6 CC
 * After this call returns, the client will no longer receive any prefix update
 * notification messages.
 *
 * @Param:  - IN arg_handle: Handle returned when the client registered for
 * route update notifications.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 */
ix_error ix_cc_ipv6_prefix_update_unregister(route_update_handle arg_handle)
{
    ix_cc_route_update_member *pMember = NULL;

    if (arg_handle == NULL)
    {
        return IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM;
    }
    
    if (g_RouteUpdateList == NULL)
    {
        return IX_CC_ERROR_NULL;
    }

    pMember = (ix_cc_route_update_member *)arg_handle;

    ix_ossl_mutex_lock(g_RouteUpdateLock, IX_OSSL_WAIT_FOREVER);

    /* Delete the member and all the elements hanging off of it */
    ListFreeDelete(pMember->prefixes);

    /* Finally remove and delete this member from the global list */
    if (!ListRemoveItem(g_RouteUpdateList, (void *)pMember))
    {
        ix_ossl_mutex_unlock(g_RouteUpdateLock);
        return IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM;
    }

    /* 
     * If this was the last member on the global list, delete the list and
     * cleanup
     */
    if (ListGetCount(g_RouteUpdateList) == 0)
    {
        ListDelete(g_RouteUpdateList);
	ix_ossl_mutex_fini(g_RouteUpdateLock);

	g_RouteUpdateList = NULL;
    }
    else
    {
        ix_ossl_mutex_unlock(g_RouteUpdateLock);
    }
   
    return IX_SUCCESS;
}

/**
 * NAME: ix_cc_ipv6_prefix_update_notify
 *
 * DESCRIPTION: This primitive will allow a client to inform the IPv6 CC about
 * specific prefixes it cares about. When the prefix is updated, the client
 * will recieve a message giving the client the new next hop ID associated with that
 * prefix/prefix_length combination. (Also see ix_cc_route_udpate_cancel).
 *
 * @Param:  - IN arg_handle: Handle returned when the client registered for
 * route update notifications.
 * @Param:  - IN arg_prefix: The prefix that the client is interested in.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 */
ix_error ix_cc_ipv6_prefix_update_notify(route_update_handle arg_handle,
                                         ix_uint128 arg_prefix)
{
    ix_cc_route_update_member *pMember = NULL;
    ix_cc_route_update_element *pElement = NULL;
    ix_cc_ipv6_prefix_change_notify notify_msg;
    ix_uint32 member_msg_id;

    
    /* check to make sure that the input handle is non null */
    if (arg_handle == NULL)
    {
        return IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM;
    }

    /* Member structure */
    pMember = (ix_cc_route_update_member *)arg_handle;
    member_msg_id = pMember->member_msg_id;
    
    /* Create a new element and add it to this member's list */
    pElement = (ix_cc_route_update_element *)
                                ix_ossl_malloc(sizeof(ix_cc_route_update_element));
    if (pElement == NULL)
    {
        return IX_CC_ERROR_OOM;
    }

    ix_ossl_memset(pElement, 0, sizeof(ix_cc_route_update_element));

    /* Copy the prefix that we intend to monitor */
    ix_ossl_memcpy(&pElement->prefix, &arg_prefix, sizeof(ix_uint128));

    /* Get the current next hop ID corresponding to this route and store it */
    IX_ERROR_CR(ix_cc_rtmv6_match_prefix(g_pIpv6Context->hRtm, 
                                         arg_prefix, 
					 &pElement->current_nhid));

    ix_ossl_mutex_lock(g_RouteUpdateLock, IX_OSSL_WAIT_FOREVER);

    /* Add this to the member's prefix list */
    if(!ListAddItem(pMember->prefixes, (void *)pElement))
    {
        ix_ossl_mutex_unlock(g_RouteUpdateLock);
        ix_ossl_free(pElement);
	return IX_CC_ERROR_INTERNAL;
    }

    ix_ossl_mutex_unlock(g_RouteUpdateLock);

    /* There may have been a delay and a possible prefix update between the time
     * the user performed a first lookup and the time this registration request
     * was received. If there was only one prefix update, it is possible that the 
     * user will never receive information about the latest Next Hop ID and might
     * continue to use the old (which, obviously, is incorrect)
     */

    ix_ossl_memcpy(&notify_msg.prefix, &pElement->prefix, sizeof(ix_uint128));
    notify_msg.prefixLength = 0; 
    notify_msg.nextHopId = pElement->current_nhid;

    /* Send message! */
    if (ix_cc_msup_send_msg(member_msg_id,
                          IX_CC_IPV6_MSG_PREFIX_CHANGE_NOTIFY,
                          &notify_msg,
			  sizeof(ix_cc_ipv6_prefix_change_notify)) != IX_SUCCESS)
    {
        return IX_CC_ERROR_SEND_FAIL;
    }

    return IX_SUCCESS;
}

/**
 * NAME: ix_cc_ipv6_prefix_update_cancel
 *
 * DESCRIPTION: This primitive will allow a client cancel any previous notification
 * request for prefix updateds (see ix_cc_route_update_notify).
 *
 * @Param:  - IN arg_handle: Handle returned when the client registered for
 * route update notifications.
 * @Param:  - IN arg_prefix: The prefix that the client is interested in.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 */
ix_error ix_cc_ipv6_prefix_update_cancel(route_update_handle arg_handle,
                                         ix_uint128 arg_prefix)
{
    ix_cc_route_update_member *pMember = NULL;
    ix_cc_route_update_element *pElement = NULL;
    LPLISTITR iterator;
    ix_uint8 found = 0;
    
    /* check to make sure that the input handle is non null */
    if (arg_handle == NULL)
    {
        return IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM;
    }
    
    pMember = (ix_cc_route_update_member *)arg_handle;

    ix_ossl_mutex_lock(g_RouteUpdateLock, IX_OSSL_WAIT_FOREVER);

    /* Create an iterator to traverse the list */
    if ((iterator = ListItrCreate(pMember->prefixes)) == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_OOM,
			        ("List creation for ipv6 failed \n"));
    }
    
    /* Create a new element and add it to this member's list */
    while ((pElement = (ix_cc_route_update_element *)
                                             ListItrGetNext(iterator)) != NULL)
    {
        if (!memcmp(&pElement->prefix, &arg_prefix, sizeof(ix_uint128)))
        {
	    /* Bingo we found the element */
	    found = 1;
            break; 
	}
    }

    /* Cleanup the iterator */
    ListItrDelete(iterator);

    if (!found)
    {
        ix_ossl_mutex_unlock(g_RouteUpdateLock);
        return IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM;
    }

    /* Remove this to the member's prefix list */
    if(!ListRemoveItem(pMember->prefixes, (void *)pElement))
    {
        ix_ossl_mutex_unlock(g_RouteUpdateLock);
	return IX_CC_ERROR_INTERNAL;
    }

    ix_ossl_mutex_unlock(g_RouteUpdateLock);
    ix_ossl_free(pElement);

    return IX_SUCCESS;
}

/* 
 * This is an internal function that will walk to the global "route update" list
 * and call ix_cc_rtmv6_match_prefix on each element. If the new Next Hop ID
 * doesn't match the old, a message is sent to the "requester" informing it about
 * this change in the Next Hop ID. The message consists of the prefix, prefix_length
 * and the new next hop ID
 */
static 
ix_error _ix_cc_check_prefix_notifications(void)
{
    ix_cc_route_update_member *pMember = NULL;
    ix_cc_route_update_element *pElement = NULL;
    LPLISTITR member_iterator;
    LPLISTITR element_iterator;
    ix_uint32 new_nhid, current_nhid;
    ix_cc_ipv6_prefix_change_notify notify_msg;
    ix_uint32 member_msg_id;

    if (g_RouteUpdateList == NULL)
    {
	/* List is NULL, nothing to do */
        return IX_SUCCESS;
    }

    ix_ossl_mutex_lock(g_RouteUpdateLock, IX_OSSL_WAIT_FOREVER);

    if (ListGetCount(g_RouteUpdateList) == 0)
    {
        ix_ossl_mutex_unlock(g_RouteUpdateLock);

	/* List is NULL, nothing to do */
        return IX_SUCCESS;
    }
    
    /* Create an iterator to traverse the list of members */
    if ((member_iterator = ListItrCreate(g_RouteUpdateList)) == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_OOM,
			        ("List creation for ipv6 failed \n"));
    }

    while ((pMember = ListItrGetNext(member_iterator)) != NULL)
    {
        if (ListGetCount(pMember->prefixes) == 0)
	{
	    continue;   /* no elements, continue through to the next element */
	}

	member_msg_id = pMember->member_msg_id;

	/* Create a new element iterator for this member's elements */
        if ((element_iterator = ListItrCreate(pMember->prefixes)) == NULL)
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_OOM,
			            ("List creation for ipv6 failed \n"));
        }

	while ((pElement = ListItrGetNext(element_iterator)) != NULL)
	{
	    current_nhid = pElement->current_nhid;

            /* Get the latest next hop id for this prefix and see if it has changed */
            if (ix_cc_rtmv6_match_prefix(g_pIpv6Context->hRtm, 
	                                         pElement->prefix,
	                                         &new_nhid) != IX_SUCCESS)
            {
                ix_ossl_mutex_unlock(g_RouteUpdateLock);
	        return IX_CC_ERROR_INTERNAL;
	    }

	    if (new_nhid != current_nhid)
	    {
	        /* 
		 * The next hop id changes, update local structures and 
		 * trigger a message send to the interested party.
		 */
		pElement->current_nhid = new_nhid;

		ix_ossl_memcpy(&notify_msg.prefix, &pElement->prefix, sizeof(ix_uint128));
                notify_msg.prefixLength = 0; 
		notify_msg.nextHopId = new_nhid;

		/* Send message! */
		if (ix_cc_msup_send_msg(member_msg_id,
                                IX_CC_IPV6_MSG_PREFIX_CHANGE_NOTIFY,
				&notify_msg,
				sizeof(ix_cc_ipv6_prefix_change_notify)) != IX_SUCCESS) 
                {
                    ix_ossl_mutex_unlock(g_RouteUpdateLock);
		    return IX_CC_ERROR_SEND_FAIL;
		}
	    }
	}

	ListItrDelete(element_iterator);
    }

    ListItrDelete(member_iterator);

    ix_ossl_mutex_unlock(g_RouteUpdateLock);
    
    return IX_SUCCESS;
}
