/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 6
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      IPv6 Forwarder Core Component - message handler
 *      This modules implements code to handle messages sent to the IPv6
 *      core component. Messages includes things like "ADD ROUTE" or "ADD
 *      NEXT HOP" etc.
 *      Some messages expect the core component to send a response, while
 *      others simply cause an addition/update of some internal data 
 *      structure. Each message is handled individually.
 *
 * = FILENAME
 *      ix_cc_ipv6_msg_handler.c
 *
 * = DESCRIPTION
 *      The file defines message handler interface functions.
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      12/03/2002
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_ipv6_msg_handler.c,v 1.20 2003/11/04 02:37:18 ktseng Exp $";

/**
 * System defined include files
 */

/**
 * User defined include files
 */

#include "ix_cc_error.h"
#include "ix_ossl.h"
#include "ix_rm.h"

#include "ix_cc.h"
#include "ix_cc_macros.h"
#include "ix_cc_properties.h"

#include "cc/ix_cc_rtmv6.h"
#include "cc/ix_cc_msup.h"

#include "bindings.h"

#include "cc/ix_cc_ipv6.h"
#include "cc/ipv6/internal/ix_cc_ipv6_init.h"
#include "cc/ipv6/internal/ix_cc_ipv6_msg_handler.h"


/**
 * Preprocessor symbols and Macros used in this file
 */

/* Flag fornot to send report to msg support lib */
#define IX_CC_IPV6_SEND_NO_RESPONSE    1 

/**
 * Type definitions whose scope is limited to this file
 */


/* 
 * This structure defines all types of return messages that can 
 * be returned from the message handler (If the message type requires
 * a reply to be sent).
 */
typedef union ix_u_cc_ipv6_msg_handler_return_data
{
    ix_cc_rtmv6_next_hop_info nextHopInfo[4];
    ix_cc_ipv6_dump_data dumpData;
    ix_uint16 icmpData;
} ix_cc_ipv6_msg_handler_return_data;

/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */

/** 
 * Extern function prototypes.
 */


/** 
 * Static function prototypes.
 */


/** 
 * Function definitions.
 */


/**
 * NAME: ix_cc_ipv6_msg_handler
 *
 * DESCRIPTION: This primitive is the message handler function for IPv6 
 * Forwarder core component. IPV6 Forwarder core component receives messages 
 * from other core component through this message handler function and 
 * internally it calls appropriate library function to process the message. 
 * This message handler shall be used to update dynamic properties.
 * 
 * @Param:  - IN arg_hDataToken - buffer handle embedding information for 
 *            the message passed in arg_UserData.
 *
 * @Param:  - IN  arg_UserData - message type
 *
 * @Param:  - IN - arg_pContext: pointer to IPv6 Forwarder core 
 *                 component context that will be passed to the core component 
 *                 when a message arrives. This context was defined by the 
 *                 core component and passed to core components infrastructure 
 *                 through ix_cc_ipv6_init function.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - null argument
 * IX_CC_ERROR_UNDEFINED_MSG - un supported message
 * IX_CC_IPV6_ERROR_MSG_LIBRARY - error from message support library
 * IX_CC_IPV6_ERROR_BUFFER_FREE - error from RM for freeing buffer
 */

ix_error ix_cc_ipv6_msg_handler(ix_buffer_handle arg_hDataToken,  
                                ix_uint32 arg_UserData,
                                void* arg_pContext
                                )
{
    ix_error err = IX_SUCCESS;

    void *pContext = NULL;
    void *pMsg = NULL;
    void *pRetMsg = NULL;

    ix_cc_ipv6_msg_handler_return_data msgRetData;
    ix_uint32 msgRetLength = 0;
    ix_uint32 noResponse = 0;  

    /* Validate arg_pContext */
    if(arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                              ("arg_pContext is NULL"));
    }/* end if(arg_pContext == NULL) */

    if((arg_UserData < IX_CC_COMMON_MSG_ID_FIRST) ||
       (arg_UserData >= IX_CC_IPV6_MSG_LAST))
    {
    
        return IX_ERROR_WARNING(IX_CC_ERROR_UNDEFINED_MSG,
                                               ("Invalid Message Type"));
    }/* end if */

    /* Extract the input argument, pMsg could be NULL after this */
    if(arg_UserData != IX_CC_IPV6_MSG_PERFORM_ADDR_RESOLUTION)
        IX_MSUP_EXTRACT_MSG(arg_hDataToken, &pMsg, &pContext);

    /* Messages other than the following types expect an "input argument" */
    if((arg_UserData != IX_CC_IPV6_MSG_PURGE_ROUTES) &&
       (arg_UserData != IX_CC_IPV6_MSG_DUMP_ROUTES) && 
       (arg_UserData != IX_CC_IPV6_MSG_PURGE_RTM) &&
       (arg_UserData != IX_CC_IPV6_MSG_DUMP_NEXT_HOPS) &&
       (arg_UserData != IX_CC_IPV6_MSG_GET_RATE_LIMIT_TIME) &&
       (arg_UserData != IX_CC_IPV6_MSG_GET_QUEUE_DEPTH) &&
       (arg_UserData != IX_CC_IPV6_MSG_GET_PACKETS_TO_DRAIN) &&
       (arg_UserData != IX_CC_IPV6_MSG_PERFORM_ADDR_RESOLUTION) &&
       (arg_UserData != IX_CC_IPV6_MSG_GET_STATISTICS)) 
    {
        if(pMsg == NULL)
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("Invalid Buffer"));

        }/* end if(pMsg == NULL) */
    }/* end if */
    
    /* Check for route messages, next hop message, icmp messages 
     * and statistics message
     */
    switch(arg_UserData)
    {
        case IX_CC_COMMON_MSG_ID_PROP_UPDATE: 
        {
            ix_cc_properties_msg *pData = (ix_cc_properties_msg *)pMsg;

	    /* no response required for property updates */
            noResponse = IX_CC_IPV6_SEND_NO_RESPONSE;

            if(((pData->propID & IX_CC_SET_PROPERTY_ADD_INTERFACE_IPV6) 
	                        == IX_CC_SET_PROPERTY_ADD_INTERFACE_IPV6) ||
               ((pData->propID & IX_CC_SET_PROPERTY_DEL_INTERFACE_IPV6) 
	                        == IX_CC_SET_PROPERTY_DEL_INTERFACE_IPV6) ||   
               ((pData->propID & IX_CC_SET_PROPERTY_PHYSICAL_IF_STATUS) 
	                        == IX_CC_SET_PROPERTY_PHYSICAL_IF_STATUS))
            {
                err = ix_cc_ipv6_set_property(pData->propID,
		                              &pData->propData,
					      arg_pContext); 
            }/* end if propId */
            else
            {
                err = IX_ERROR_WARNING(IX_CC_ERROR_UNDEFINED_MSG,
                                       ("Invalid Message Type"));
 
            }/*end if invalid propId*/

            break;
        }

        /* RTM route related message */

	/* Add a new prefix */
        case IX_CC_IPV6_MSG_ADD_PREFIX:
        {
            ix_cc_ipv6_add_prefix_data *pData = 
	                                (ix_cc_ipv6_add_prefix_data *)pMsg;

	    err = ix_cc_ipv6_add_prefix(pData->ipv6Addr,
	                               pData->prefixLength,
				       pData->nextHopId,
				       arg_pContext);
            break;
        }

	/* Delete an existing prefix */
        case IX_CC_IPV6_MSG_DELETE_PREFIX:
        {
            ix_cc_ipv6_delete_prefix_data *pData = 
	                                (ix_cc_ipv6_delete_prefix_data *)pMsg;
            err = ix_cc_ipv6_delete_prefix(pData->ipv6Addr,
	                                   pData->prefixLength,
					   arg_pContext);
            break;
        }

	/* Update an existing prefix */
        case IX_CC_IPV6_MSG_UPDATE_PREFIX:
        {
            ix_cc_ipv6_add_prefix_data *pData = 
	                                  (ix_cc_ipv6_add_prefix_data*)pMsg;

	    err = ix_cc_ipv6_update_prefix(pData->ipv6Addr,
	                                   pData->prefixLength,
					   pData->nextHopId,
					   arg_pContext);
            break;
        }

	/* Lookup a prefix, needs to return Next-hop information */
        case IX_CC_IPV6_MSG_LOOKUP_PREFIX:
        {
            ix_uint128 *pData = (ix_uint128 *)pMsg;

            msgRetLength = sizeof(ix_cc_rtmv6_next_hop_info);

            err = ix_cc_ipv6_lookup_prefix(*pData,
	                                   &msgRetData.nextHopInfo[0],
					   arg_pContext);

            pRetMsg = (void*)&msgRetData.nextHopInfo;

            break;
        }

	/* Purge all routes from RTMv6. No response is required */
        case IX_CC_IPV6_MSG_PURGE_ROUTES:
	{
       	    err = ix_cc_ipv6_purge_prefixes(arg_pContext);
            noResponse = IX_CC_IPV6_SEND_NO_RESPONSE;

            break;
        }

	/* Dump all routes that exists in RTMv6 */
        case IX_CC_IPV6_MSG_DUMP_ROUTES:
        {
            ix_cc_rtmv6_statistics rtmData;
            ix_uint32 value;

	    /* Figure out how much memory to allocate first */
            err = ix_cc_rtmv6_get_statistics(
	                  ((ix_cc_ipv6_context *)arg_pContext)->hRtm, &rtmData);

            if(err == IX_SUCCESS)
            {
                /* Allocate memory for data */
                value = rtmData.numberOfRoutes;
                value = IX_CC_RTMV6_DUMP_ROUTE_SIZE(value);

		/* Return a buffer of size "value" */
                msgRetData.dumpData.size = value;
                msgRetData.dumpData.pBuffer = 
		             (char *)ix_ossl_malloc(msgRetData.dumpData.size);

                if(msgRetData.dumpData.pBuffer != NULL)
                {
                    msgRetLength = sizeof(ix_cc_ipv6_dump_data);
                    err = ix_cc_ipv6_dump_prefixes(msgRetData.dumpData.pBuffer,
		                        msgRetData.dumpData.size,arg_pContext);
                    pRetMsg = (void*)&msgRetData.dumpData;
                }
                else
                {
                    err = IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                                           ("No memory to dump routes"));

                }
            }/* end if statistics*/
            break;
        }

        /* RTMv6 Next Hop related messages */

	/* Add a new next hop to the table */
        case IX_CC_IPV6_MSG_ADD_NEXT_HOP:
        {
            ix_cc_ipv6_add_next_hop_data *pData = 
	                                  (ix_cc_ipv6_add_next_hop_data *)pMsg;

	    err = ix_cc_ipv6_add_next_hop(pData->nextHopId,
	    	                          pData->numNextHops,
	                                  pData->nextHopInfo,
					  arg_pContext);
            break;
        }

	/* Delete an existing next hop */
        case IX_CC_IPV6_MSG_DELETE_NEXT_HOP:
        {
            ix_cc_rtmv6_nhid *pData = (ix_cc_rtmv6_nhid *)pMsg;

            err = ix_cc_ipv6_delete_next_hop(*pData, arg_pContext);

            break;
        }

	/* Update an existing next hop with "new" next hop information */
        case IX_CC_IPV6_MSG_UPDATE_NEXT_HOP:
        {
            ix_cc_ipv6_add_next_hop_data *pData = 
	                                 (ix_cc_ipv6_add_next_hop_data *)pMsg;

	    err = ix_cc_ipv6_update_next_hop(pData->nextHopId,
	                                     pData->numNextHops,
	                                     pData->nextHopInfo,
					     arg_pContext);
            break;
        }

	/* Return nexthop information given a nexthop ID */
        case IX_CC_IPV6_MSG_GET_NEXT_HOP:
        {
            ix_cc_rtmv6_nhid *pData = (ix_cc_rtmv6_nhid*)pMsg;

            msgRetLength = sizeof(ix_cc_rtmv6_next_hop_info);

            err = ix_cc_ipv6_get_next_hop(*pData, 
	                                  &msgRetData.nextHopInfo[0],
					  arg_pContext);

            pRetMsg = (void*)&msgRetData.nextHopInfo;

            break;
        }

	/* Cleanup the RTMv6 */
        case IX_CC_IPV6_MSG_PURGE_RTM:
	{
       	    err = ix_cc_ipv6_purge_rtm(arg_pContext);
            noResponse = IX_CC_IPV6_SEND_NO_RESPONSE;
            break;
        }

	/* Dump all next hops */
        case IX_CC_IPV6_MSG_DUMP_NEXT_HOPS:
        {
            ix_cc_rtmv6_statistics rtmData;

	    /* Figure out how much memory to allocate */
            err = ix_cc_rtmv6_get_statistics(
                        ((ix_cc_ipv6_context*)arg_pContext)->hRtm, &rtmData);

            if(err == IX_SUCCESS)
            {
		/* This will compute the number of next hops as follows:
		 * "numberOfNextHops" is the total number of next hops added
		 * some of which might be pointing to upto 4-equal cost
		 * next hops. "numberOfMultipleNextHops" contains that number.
		 * Therefore, the total number of next hops (we avoid printing
		 * the "intermediate" next hops) is 
		 * total = (numberOfNextHops - numOfMultipleNextHops) + 
		 *                (numberOfMultipleNextHops * 5)
		 */
                msgRetData.dumpData.size = 
		       (rtmData.numberOfNextHops - rtmData.numberOfMultipleNextHops) +
		           rtmData.numberOfMultipleNextHops * 5;

                msgRetData.dumpData.size = 
		      IX_CC_RTMV6_DUMP_NEXT_HOP_SIZE(msgRetData.dumpData.size);

                msgRetData.dumpData.pBuffer = 
		              (char *)ix_ossl_malloc(msgRetData.dumpData.size);

                if(msgRetData.dumpData.pBuffer != NULL)
                {
                    msgRetLength = sizeof(ix_cc_ipv6_dump_data);

                    err = ix_cc_ipv6_dump_next_hops(
		                     msgRetData.dumpData.pBuffer,
				     msgRetData.dumpData.size,
				     arg_pContext);

                    pRetMsg = (void*)&msgRetData.dumpData;
                }
                else
                {
                    err = IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                                           ("No memory to dump next hops"));

                }
            }/* end if statistics*/
            break;
        }

	/* Set the MTU for an interface */
        case IX_CC_IPV6_MSG_SET_MTU:
        {
            ix_cc_ipv6_set_mtu_data *pData = (ix_cc_ipv6_set_mtu_data *)pMsg;
            err = ix_cc_ipv6_set_mtu(pData->nextHopId,
	                             pData->mtu,
				     arg_pContext);
            break;
        }

        case IX_CC_IPV6_MSG_SET_FLAGS:
        {
            ix_cc_ipv6_set_flags_data *pData = 
	                                  (ix_cc_ipv6_set_flags_data*)pMsg;

            err = ix_cc_ipv6_set_flags(pData->nextHopId,
	                               pData->flags,
				       arg_pContext);
            break;
        }

        /* 
	 * ICMP related messages 
	 */

	/* Get sleep time */
        case IX_CC_IPV6_MSG_GET_RATE_LIMIT_TIME: 
	{
            msgRetLength = sizeof(ix_uint16);

            err = ix_cc_ipv6_get_rate_limit_time(&msgRetData.icmpData,
	                                         arg_pContext);

            pRetMsg = (void*)&msgRetData.icmpData;
            break; 
        }

	/* Set sleep time. No response required */
        case IX_CC_IPV6_MSG_SET_RATE_LIMIT_TIME:
        { 
            ix_uint16 *pData = (ix_uint16*)pMsg;

            err = ix_cc_ipv6_set_rate_limit_time(*pData, arg_pContext);

            noResponse = IX_CC_IPV6_SEND_NO_RESPONSE;

            break;
        }

	/* Get depth of ICMP queue. */
        case IX_CC_IPV6_MSG_GET_QUEUE_DEPTH: 
	{
            msgRetLength = sizeof(ix_uint16);

            err = ix_cc_ipv6_get_queue_depth(&msgRetData.icmpData,
	                                     arg_pContext);

            pRetMsg = (void*)&msgRetData.icmpData;
            break; 
        }

	/* Get the number of packets to drain when the timer goes off */
        case IX_CC_IPV6_MSG_GET_PACKETS_TO_DRAIN: 
	{
            msgRetLength = sizeof(ix_uint16);
            err = ix_cc_ipv6_get_packets_to_drain(&msgRetData.icmpData,
	                                          arg_pContext);
            pRetMsg = (void*)&msgRetData.icmpData;
            break; 
        }

	/* Set the number of packets to drain, when the timer goes off */
        case IX_CC_IPV6_MSG_SET_PACKETS_TO_DRAIN: 
        {
            ix_uint16 *pData = (ix_uint16*)pMsg;
            err = ix_cc_ipv6_set_packets_to_drain(*pData,
	                                          arg_pContext);

            noResponse = IX_CC_IPV6_SEND_NO_RESPONSE;
            break; 
        }

        /* Send an ICMP error message */
        case IX_CC_IPV6_MSG_SEND_ICMP_ERROR:
        {
            ix_cc_ipv6_icmp_err_data *pData =
                            (ix_cc_ipv6_icmp_err_data *)pMsg;
            err = ix_cc_ipv6_send_icmp_error(
                                            pData->icmpErrType,
                                            pData->icmpErrCode,
                                            pData->hBuffer,
                                            pData->param,
                                            arg_pContext);
            break;
        }
        /* Send an ICMP informational message */
        case IX_CC_IPV6_MSG_SEND_ICMP_INFO:
        {
            ix_cc_ipv6_icmp_info_data *pData =
                            (ix_cc_ipv6_icmp_info_data *)pMsg;
            err = ix_cc_ipv6_send_icmp_info_message(
                                        pData->srcIpAddr,
                                        pData->dstIpAddr,
                                        pData->type,
                                        pData->code,
                                        pData->payloadLength,
                                        pData->pPayload,
                                        arg_pContext);
            break;
        }

	/* Get Core + Microblock maintained Statistics Counters */
        case IX_CC_IPV6_MSG_GET_STATISTICS: 
        {
            msgRetLength = sizeof(ix_cc_ipv6_stats_data);
            
            msgRetData.dumpData.size = msgRetLength;
            msgRetData.dumpData.pBuffer = 
		              (char *)ix_ossl_malloc(msgRetData.dumpData.size);
            ix_ossl_memset(msgRetData.dumpData.pBuffer, 0, 
                                                     msgRetData.dumpData.size);
            pRetMsg = (void*)&msgRetData.dumpData;
            msgRetLength = sizeof(ix_cc_ipv6_dump_data);

            if (msgRetData.dumpData.pBuffer != NULL) 
            {
                err = ix_cc_ipv6_get_statistics(
                          (ix_cc_ipv6_stats_data *)msgRetData.dumpData.pBuffer,
	                   arg_pContext);

            }
            else
            {
                err = IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                                           ("No memory to dump next hops"));
            }

            break; 
        }

        /* Perform address resolution */
        case IX_CC_IPV6_MSG_PERFORM_ADDR_RESOLUTION:
        {
            /* We assume the called function will free the buffer */
            err = ix_cc_ipv6_perform_addr_resolution(arg_hDataToken, 
                                                     pContext);

            return(IX_SUCCESS);
        }

        /* Add neighbor entry to the L2 table */
        case IX_CC_IPV6_MSG_ADD_NBR:
        {
            ix_cc_ipv6_add_nbr_data *pData = (ix_cc_ipv6_add_nbr_data *)pMsg;
            err = ix_cc_ipv6_add_neighbor(pData->l2Index,
                                                pData->nbrIp,
                                                pData->outputPort,
                                                pData->l2Addr,
                                                pContext);
            noResponse = IX_CC_IPV6_SEND_NO_RESPONSE;
            break;
        }

        /* Delete nbr entry from the L2 table */
        case IX_CC_IPV6_MSG_DEL_NBR:
        {
            ix_uint16 *pData = (ix_uint16 *)pMsg;
            err = ix_cc_ipv6_del_neighbor(*pData, pContext);
            noResponse = IX_CC_IPV6_SEND_NO_RESPONSE;
            break;
        }

        default:
	   
	   /* 
	    * The EXTRACT_MSG() macro at the beginning of this function
	    * expects the caller to free the "pMsg" pointer. Do it now.
	    */
 
           if (pMsg != NULL)
	   {
	       ix_ossl_free(pMsg);
	   }

           IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hDataToken),
	                                    IX_CC_IPV6_ERROR_BUFFER_FREE,0);  

           return IX_ERROR_WARNING(IX_CC_ERROR_UNDEFINED_MSG,
                                     ("Invalid Message Type"));
           break;
    } /* end switch(arg_UserData) */

    /* 
     * The EXTRACT_MSG() macro at the beginning of this function
     * expects the caller to free the "pMsg" pointer. Do it now.
     */
 
    if (pMsg != NULL)
    {
        ix_ossl_free(pMsg);
    }

    /* 
     * Note that some messages DO NOT require a response to be sent to the
     * sender of the message. In this case, simply remove the message buffer
     * from the queue and return it to the free list and return 
     */

    if (noResponse == IX_CC_IPV6_SEND_NO_RESPONSE)
    {
        IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hDataToken),
	                                      IX_CC_IPV6_ERROR_BUFFER_FREE,0);  
        return IX_SUCCESS;
    }/* end if no need to send message to msg support lib*/


    /* 
     * Send return message, length and retrieved context 
     * to message support library. The message length and the actual
     * message were built appropriately in each case statement.
     */

    err = ix_cc_msup_send_reply_msg(pContext, pRetMsg, msgRetLength, err);
    if(err != IX_SUCCESS)
    {
        err = IX_ERROR_WARNING(IX_CC_IPV6_ERROR_MSG_LIBRARY,
                                   ("Send reply message failed"));
    }/* endif (err != IX_SUCCESS) */

    IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hDataToken),
                                              IX_CC_IPV6_ERROR_BUFFER_FREE,0);
    return err;

} /* end ix_cc_ipv6_msg_handler */

/**
 * NAME: _ix_cc_ipv6_cb_genaral
 *
 * DESCRIPTION: This is a general callback function. Message support library 
 * calls this function to return the status of the requested asynchrouns 
 * operation 
 *
 * @Param:  - IN  arg_Result: result of the asynchronous operation
 * @Param:  - IN  arg_pContext: context provided by caller
 * @Param:  - IN  arg_pMsg: return message
 * @Param:  - IN  arg_MsgLength: length of the message
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_IPV6_ERROR_CALLBACK - user callback returns error message
 */

ix_error _ix_cc_ipv6_cb_general(ix_error arg_Result,
                                void* arg_pContext,
                                void* arg_pMsg,
                                ix_uint32 arg_MsgLength)
{
    ix_error err = IX_SUCCESS;
 
    ix_cc_ipv6_general_callback_data *pCbData = NULL;
 
    pCbData = (ix_cc_ipv6_general_callback_data *)arg_pContext;
   
    if(pCbData != NULL)
    {
        err = pCbData->pUserCallback(arg_Result, pCbData->pUserContext);

        ix_ossl_free(arg_pContext);

        if(err != IX_SUCCESS)
        {
            return IX_ERROR_WARNING(IX_CC_IPV6_ERROR_CALLBACK,
                                      ("general callback failed"));

        }/* end if (err != IX_SUCCESS) */
    }/* end if(pCbData != NULL) */
       
    return IX_SUCCESS;
} /* _ix_cc_ipv6_cb_general */


/**
 * NAME: _ix_cc_ipv6_cb_lookup_rtm
 *
 * DESCRIPTION: This is a callback function for lookup rtm and get next 
 * hop message helpers. Message support library calls this function to return 
 * the status of the requested asynchrouns operation 
 *
 * @Param:  - IN  arg_Result: result of the asynchronous operation
 * @Param:  - IN  arg_pContext: context provided by caller
 * @Param:  - IN  arg_pMsg: return message
 * @Param:  - IN  arg_MsgLength: length of the message
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_IPV6_ERROR_CALLBACK - user callback returns error message
 * 
 */

ix_error _ix_cc_ipv6_cb_lookup_rtm(ix_error arg_Result,
                                   void* arg_pContext,
                                   void* arg_pMsg,
                                   ix_uint32 arg_MsgLength)
{
    ix_error err = IX_SUCCESS;
 
    ix_cc_ipv6_lookup_callback_data *pCbData = NULL;
    ix_cc_rtmv6_next_hop_info *pNextHopInfo = NULL;

    pCbData = (ix_cc_ipv6_lookup_callback_data*)arg_pContext;
    pNextHopInfo = (ix_cc_rtmv6_next_hop_info*)arg_pMsg;
    
   
    if(pCbData != NULL)
    {
        err = pCbData->pUserCallback(arg_Result,
	                             pCbData->pUserContext,
				     pNextHopInfo);
        ix_ossl_free(arg_pContext);

        if(err != IX_SUCCESS)
        {
            return IX_ERROR_WARNING(IX_CC_IPV6_ERROR_CALLBACK,
                                      ("lookup callback failed"));

        }/* end if (err != IX_SUCCESS) */
    }/* end if(pCbData != NULL)*/
       
    return IX_SUCCESS;
}/* _ix_cc_ipv6_cb_lookup_rtm */

/**
 * NAME: _ix_cc_ipv6_cb_dump_rtm
 *
 * DESCRIPTION: This is a callback function for dump(routes/next hops) 
 * message helpers. Message support library calls this function to return 
 * the status of the requested asynchrouns operation 
 *
 * @Param:  - IN  arg_Result: result of the asynchronous operation
 * @Param:  - IN  arg_pContext: context provided by caller
 * @Param:  - IN  arg_pMsg: return message
 * @Param:  - IN  arg_MsgLength: length of the message
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_IPV6_ERROR_CALLBACK - user callback returns error message
 * 
 */

ix_error _ix_cc_ipv6_cb_dump_rtm(ix_error arg_Result,
                                 void* arg_pContext,
                                 void* arg_pMsg,
                                 ix_uint32 arg_MsgLength)
{
    ix_error err = IX_SUCCESS;
 
    ix_cc_ipv6_dump_callback_data* pCbData = NULL;
    ix_cc_ipv6_dump_data* pDumpData = NULL;


    pCbData = (ix_cc_ipv6_dump_callback_data*)arg_pContext;
    pDumpData = (ix_cc_ipv6_dump_data*)arg_pMsg;
   
    if(pCbData != NULL)
    {
        err = pCbData->pUserCallback(arg_Result,
	                             pCbData->pUserContext,
				     pDumpData);
        ix_ossl_free(arg_pContext);
        ix_ossl_free(pDumpData->pBuffer);

        if(err != IX_SUCCESS)
        {
            return IX_ERROR_WARNING(IX_CC_IPV6_ERROR_CALLBACK,
                                      ("dump callback failed"));
        }/* end if (err != IX_SUCCESS) */

    }/* end if (pCbData != NULL)*/
       
    return IX_SUCCESS;
}/* _ix_cc_ipv6_cb_dump_rtm */

/**
 * NAME: _ix_cc_ipv6_cb_icmp
 *
 * DESCRIPTION: This is a callback function for ICMP message helpers. Message 
 * support library calls this function to return the status of the requested 
 * asynchrouns operation 
 *
 * @Param:  - IN  arg_Result: result of the asynchronous operation
 * @Param:  - IN  arg_pContext: context provided by caller
 * @Param:  - IN  arg_pMsg: return message
 * @Param:  - IN  arg_MsgLength: length of the message
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_IPV6_ERROR_CALLBACK - user callback returns error message
 */

ix_error _ix_cc_ipv6_cb_icmp(ix_error arg_Result,
                             void* arg_pContext,
                             void* arg_pMsg,
                             ix_uint32 arg_MsgLength)
{
    ix_error err = IX_SUCCESS;
    ix_cc_ipv6_icmp_callback_data *pCbData = NULL;
    ix_uint16 *pRetData;

    pCbData = (ix_cc_ipv6_icmp_callback_data*)arg_pContext;
    pRetData = (ix_uint16 *)arg_pMsg;
   
    if(pCbData != NULL)
    {
        err = pCbData->pUserCallback(arg_Result,
	                             pCbData->pUserContext,
				     *pRetData);
        ix_ossl_free(arg_pContext);

        if(err != IX_SUCCESS)
        {        
            return IX_ERROR_WARNING(IX_CC_IPV6_ERROR_CALLBACK,
                                      ("icmp callback failed"));
        }/* end if (err != IX_SUCCESS) */

    }/* end if (pCbData != NULL)*/
       
    return IX_SUCCESS;
}/* _ix_cc_ipv6_cb_icmp */

/**
 * NAME: _ix_cc_ipv6_cb_statistics
 *
 * DESCRIPTION: This is a callback function for statistics message helper. 
 * Message support library calls this function to return the status of the 
 * requested asynchrouns operation 
 *
 * @Param:  - IN  arg_Result: result of the asynchronous operation
 * @Param:  - IN  arg_pContext: context provided by caller
 * @Param:  - IN  arg_pMsg: return message
 * @Param:  - IN  arg_MsgLength: length of the message
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_IPV6_ERROR_CALLBACK - user callback returns error message
 */

ix_error _ix_cc_ipv6_cb_statistics(ix_error arg_Result,
                                   void* arg_pContext,
                                   void* arg_pMsg,
                                   ix_uint32 arg_MsgLength)
{
    ix_error err = IX_SUCCESS;
 
    ix_cc_ipv6_statistics_callback_data* pCbData = NULL;
    ix_cc_ipv6_stats_data* pStatsData = NULL;
    ix_cc_ipv6_dump_data* pDumpData = NULL;

    pCbData = (ix_cc_ipv6_statistics_callback_data*)arg_pContext;
    pDumpData = (ix_cc_ipv6_dump_data*)arg_pMsg;
    pStatsData = (ix_cc_ipv6_stats_data*)pDumpData->pBuffer;
   
    if(pCbData != NULL)
    {
        err = pCbData->pUserCallback(arg_Result,
	                             pCbData->pUserContext,
				     pStatsData);
        ix_ossl_free(arg_pContext);
        ix_ossl_free(pDumpData->pBuffer);

        if(err != IX_SUCCESS)
        {        
            return IX_ERROR_WARNING(IX_CC_IPV6_ERROR_CALLBACK,
                                      ("statistics callback failed"));
        }/* end if (err != IX_SUCCESS) */    

    }/* end if (pCbData != NULL)*/
       
    return IX_SUCCESS;
}/* _ix_cc_ipv6_cb_statistics */
