/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 6
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      IPv6 Forwarder Core Component - packet handlers.
 *      This modules handles four kinds of packets
 *      - Packets from the IPv6 microblock classified as "high priority" 
 *        This includes packets that need to be forwarder by the core.
 *      - Packets from the IPv6 microblock classified as "Low priority"
 *        This includes packets with size larger than MTU etc. Typical 
 *        case handling involves sending an icmp message with an error
 *        message back to the sender of the packet.
 *      - Packets from the local stack driver that need to be forwarded.
 *      - Packets from other core components (for instance, from the tunneling
 *        core component).
 *
 * = FILENAME
 *      ix_cc_ipv6_pkt_handlers.c
 *
 * = DESCRIPTION
 *      The file defines packet handler interface functions.
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      12/03/2002
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */

#define IX_ERROR_FILE_IDENT "$Id: ix_cc_ipv6_pkt_handlers.c,v 1.21 2003/12/12 00:49:54 ktseng Exp $";


/**
 * System defined include files
 */


/**
 * User defined include files
 */
#include "ix_cc_error.h"

#include "ix_rm.h"
#include "ix_ossl.h"
#include "cc/ix_cc_rtmv6.h"

#include "ix_cc.h"
#include "bindings.h"
#include "ipv6_fwder_uc.h"
#include "cc/ix_cc_ipv6.h"
#include "cc/ipv6/internal/ix_cc_ipv6_init.h"
#include "cc/ipv6/internal/icmp6.h"
#include "cc/ipv6/internal/fragmentation_support.h"
#include "cc/ipv6/internal/v6_header_defs.h"
#include "cc/ipv6/internal/v6_header_validation.h"
#include "ipv6_fwder_uc.h"

/**
 * Preprocessor symbols and Macros used in this file
 */

/**
 * Type definitions whose scope is limited to this file
 */


/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */


/** 
 * Extern function prototypes.
 */

extern ix_cc_ipv6_context* g_pIpv6Context;

extern ix_error _ix_cc_ipv6_forward(ix_buffer_handle arg_hBuffer, 
                                    ix_uint8 arg_inCounters,
                                    ix_uint8 arg_outCounters);

extern 
ix_error _ix_cc_ipv6_forward_pkts_from_stkdrvr(ix_buffer_handle arg_hBuffer);  

/** 
 * Static function prototypes.
 */

/**
 * NAME: _ix_cc_ipv6_process_link_local_pkt
 *
 * DESCRIPTION: This function processes a link local packet. This function
 * is called when a LINK_LOCAL exception is raised by the microblock. It checks
 * to make sure that the packet is addressed to one of it's interfaces and if so
 * either sends the packet to Neighbor discovery (link local packet containing
 * ICMP headers) or up to the stack driver.
 * 
 * @Param:  - IN arg_hBuffer: handle to a buffer which contains exception 
 *                            packets from IPv6 microblock
 * @Param: - IN arg_ipv6Addr: The destination IPv6 address of the packet.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_SEND_FAIL - error from CCI
 * 
 */
ix_error 
_ix_cc_ipv6_process_link_local_pkt(ix_buffer_handle arg_hBuffer,
                                   ix_uint128 arg_ipv6Addr)
{
    ix_uint128 portIpv6Addr[PROP_MAX_IPV6_ADDRESSES];
    ix_uint8 numElements;
    ix_uint16 portNum;
    ix_hw_buffer_meta *pHwMeta;
    ix_uint8 i;
    ix_uint8 dropped = 0;

    /* get the meta data for the port */
    IX_ERROR_CRT(ix_rm_buffer_get_meta(arg_hBuffer, 
                                       (void **)&pHwMeta), 
				       IX_CC_ERROR_INTERNAL,
				       0);

    portNum = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_InputPort));

    /* Get the IPv6 address for this port */
    IX_ERROR_CRT(_ix_cc_ipv6_get_port_ip_address(THIS_BLADE_ID,
                                                 portNum,
						 &numElements,
                                                 portIpv6Addr),
                                                 IX_CC_ERROR_INTERNAL,
                                                 0);
    for (i = 0; i < numElements; i++)
    {
        if (IS_IPV6_ADDRESS_EQUAL(&arg_ipv6Addr, &portIpv6Addr[i]))
        {
	    /* Packet is addressed to this node */

	    /* If NBR Sol or NBR adv send to Neighbor Discovery */
	    _ix_cc_ipv6_inspect_pkt_hdr(arg_hBuffer, &dropped);
	    
	    /* The packet was used, return from this function */ 
	    if (dropped)
	    {
	       return IX_SUCCESS;
	    }
	    else
	    {
                /* Else Send the packet to the stack driver */
                IX_ERROR_CRT(ix_rm_message_send(
	                           IX_CC_IPV6_PKT_HIGH_PRIORITY_STKDRV_OUTPUT,
		                   arg_hBuffer, 
			           0), 
			           IX_CC_ERROR_SEND_FAIL, 
			           0);
            }
        }
    }

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("Dropping link local packet\n");
#endif /* IPV6_DEBUG */

    /* Drop the packet */
    IX_ERROR_CR(ix_rm_buffer_free_chain(arg_hBuffer));

    return IX_SUCCESS;

}

/**
 * NAME: _ix_cc_ipv6_process_local_packet
 *
 * DESCRIPTION: This function processes a locally destined packet. This function
 * is called when a LOCAL exception is raised by the microblock. It either sends 
 * the packet to Neighbor discovery (link local packet containing ICMP headers) 
 * or up to the stack driver.
 * 
 * @Param:  - IN arg_hBuffer: handle to a buffer which contains exception 
 *                            packets from IPv6 microblock
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_SEND_FAIL - error from CCI
 * 
 */
ix_error
_ix_cc_ipv6_process_local_packet(ix_buffer_handle arg_hBuffer)
{
    ix_uint8 dropped = 0;

    /* If NBR Sol or NBR adv send to Neighbor Discovery */
    _ix_cc_ipv6_inspect_pkt_hdr(arg_hBuffer, &dropped);
	    
    /* The packet was used, return from this function */ 
    if (dropped)
    {
       return IX_SUCCESS;
    }
    else
    {
        /* Else Send the packet to the stack driver */
        IX_ERROR_CRT(ix_rm_message_send(
	                           IX_CC_IPV6_PKT_HIGH_PRIORITY_STKDRV_OUTPUT,
		                   arg_hBuffer, 
			           0), 
			           IX_CC_ERROR_SEND_FAIL, 
   		                   0);
    }

    return IX_SUCCESS;
}

/** 
 * Function definitions.
 */

/**
 * NAME: ix_cc_ipv6_microblock_high_priority_pkt_handler
 *
 * DESCRIPTION: This is the only registered function to receive exception 
 * packets from high priority queue of IPv6 microblock. This function 
 * sends the packet to stack driver (ie, recieves only local delivery packets). 
 * 
 * @Param:  - IN arg_hBuffer: handle to a buffer which contains exception 
 *                            packets from IPv6 microblock
 * @Param: - IN arg_ExceptionCode: the exception code.
 * @Param: - IN arg_pContext: pointer to IPV6 Forwarder core component context.
 * @Param:  - OUT none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_SEND_FAIL - error from CCI
 * 
 */

ix_error 
ix_cc_ipv6_microblock_high_priority_pkt_handler(ix_buffer_handle arg_hBuffer,  
                                                ix_uint32 arg_ExceptionCode,
                                                void* arg_pContext)
{
    ix_error err = IX_SUCCESS;
    ix_hw_buffer_meta *pHwMeta;

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("**** IPv6 High Priority Pkt Handler Invoked **** \n");
#endif /* IPV6_DEBUG */

    /* get the meta data for the port */
    IX_ERROR_CRT(ix_rm_buffer_get_meta(arg_hBuffer, 
                                       (void **)&pHwMeta), 
				       IX_CC_ERROR_INTERNAL,
				       0);
#if defined(IPV6_DEBUG)

    ix_ossl_message_log (" *********  Packet Details  *********** \n");

    ix_ossl_message_log ("PacketSize (0x%x) \n", 
			     IX_NTOH16(pHwMeta->m_PacketSize));
    ix_ossl_message_log ("BufferSize (0x%x) \n", 
                             IX_NTOH16(pHwMeta->m_BufferSize));
    ix_ossl_message_log ("Offset (0x%x) \n", 
			     IX_NTOH16(pHwMeta->m_Offset));
    ix_ossl_message_log ("BufferInfo (0x%x) \n", 
			     IX_NTOH16(pHwMeta->m_BufferInfo));
    ix_ossl_message_log ("InputPort (0x%x) \n", 
			     IX_NTOH16(pHwMeta->m_InputPort));
			       
#endif /* IPV6_DEBUG */

    switch(arg_ExceptionCode)
    {
        case IPV6_EXCP_OPTIONS:
	{
	    /* 
	     * Detected the presence of options header in the IPv6 packets.
	     * The packet is sent to the core for handling
	     */

#if defined(IPV6_DEBUG)
            ix_ossl_message_log ("IPV6_EXCP_OPTIONS\n");
#endif /* IPV6_DEBUG */

            return _ix_cc_ipv6_forward(arg_hBuffer, 0, 0);
            break;
        }

        case IPV6_EXCP_LOCAL:
	{
	    /* 
	     * This packet is for local delivery. Send the packet to the 
	     * stack driver.
	     * NOTE: The microblock already checks to see if the NHID Type
	     * is forwarder. If not, the packet is sent to the next microblock
	     * in the group. This handles instances where an IPv6 packet might
	     * be encpasulated in a V4 packet. The packet might be destinted
	     * to the local node, but still need to go to the tunneling micro
	     * block before it can be sent to the local stack driver. As far
	     * as the core goes though, nothing changes.
	     */
 
	    /* This packet might be an ICMPv6 or a packet for neighbor 
	     * discovery. We intercept the packet and make sure it is not
	     * one of those before sending it to the stack driver.
	     */

#if defined(IPV6_DEBUG)
            ix_ossl_message_log ("IPV6_EXCP_LOCAL\n");
#endif /* IPV6_DEBUG */

            err = _ix_cc_ipv6_process_local_packet(arg_hBuffer);
            break;
        }

        default:


            IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hBuffer),
	                                        IX_CC_ERROR_SEND_FAIL, 0);

            return IX_ERROR_WARNING(IX_CC_ERROR_UNDEFINED_EXCEP,
                                                ("undefined exception code"));
            break;

    } /* end switch */

    if (err != IX_SUCCESS)
    {
        IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hBuffer),
                                             IX_CC_ERROR_SEND_FAIL, 0);

        return IX_ERROR_WARNING(IX_CC_ERROR_SEND_FAIL,
                                      ("packet send failed - high priority"));

    }/* end if (err != IX_SUCCESS) */     
    
   return IX_SUCCESS;

} /* end ix_cc_ipv6_microblock_high_priority_pkt_handler */


/**
 * NAME: ix_cc_ipv6_microblock_low_priority_pkt_handler
 *
 * DESCRIPTION: This is the registered function to receive exception packets 
 * from IPv6 microblock. This function internally calls/performs different 
 * functions/operations based on exception code for the given packet.
 * 
 * @Param:  - IN arg_hBuffer: handle to a buffer which contains exception 
 *                            packets from IPv6 microblock
 * @Param: - IN arg_ExceptionCode: Exception codes generated by IPv6 forwarder 
 *                                 microblock.
 * @Param: - IN arg_pContext: pointer to IPV6 Forwarder core component context
 * @Param:  - OUT none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_SEND_FAIL - error from CCI
 * IX_CC_ERROR_UNDEFINED_EXCEP - unsupported exception code
 * IX_CC_ERROR_INTERNAL - an internal error
 * IX_CC_IPV6_ERROR_RTM - error from RTMv6
 */


ix_error 
ix_cc_ipv6_microblock_low_priority_pkt_handler(ix_buffer_handle arg_hBuffer,  
                                               ix_uint32 arg_ExceptionCode,
                                               void* arg_pContext)
{
    ix_error err = IX_SUCCESS;
    ix_cc_ipv6_header *pipv6Hdr, ipv6Hdr;
    ix_cc_rtmv6_next_hop_info nextHopInfo;
    ix_uint128 ipv6Addr;
    ix_uint16 ingressPort;

    ix_hw_buffer_meta *pHwMeta;

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("**** IPv6 Low Priority Pkt Handler Invoked **** \n");
#endif /* IPV6_DEBUG */

    /* Get the IPv6 header */
    IX_ERROR_CRT(ix_cc_hw_get_packet_data(arg_hBuffer,
                                          (void**)&pipv6Hdr),
				          IX_CC_ERROR_INTERNAL,
				          0);

    /* get the meta data for the port */
    IX_ERROR_CRT(ix_rm_buffer_get_meta(arg_hBuffer, 
                                       (void **)&pHwMeta), 
				       IX_CC_ERROR_INTERNAL,
				       0);

    ingressPort = IX_NTOH16(pHwMeta->m_InputPort);

#if defined(IPV6_DEBUG)

    ix_ossl_message_log (" *********  Packet Details  *********** \n");

    ix_ossl_message_log ("PacketSize (0x%x) \n", 
			     IX_NTOH16(pHwMeta->m_PacketSize));
    ix_ossl_message_log ("BufferSize (0x%x) \n", 
                             IX_NTOH16(pHwMeta->m_BufferSize));
    ix_ossl_message_log ("Offset (0x%x) \n", 
			     IX_NTOH16(pHwMeta->m_Offset));
    ix_ossl_message_log ("BufferInfo (0x%x) \n", 
			     IX_NTOH16(pHwMeta->m_BufferInfo));
    ix_ossl_message_log ("InputPort (0x%x) \n", 
			     IX_NTOH16(pHwMeta->m_InputPort));
			       
#endif /* IPV6_DEBUG */

    /* Memcopy the Header to get over the quad word alignment issue */
    ix_ossl_memcpy(&ipv6Hdr, pipv6Hdr, IX_CC_IPV6_HDR_LEN);

    IX_NTOH128(&ipv6Hdr.dstAddress, &ipv6Addr);

    switch(arg_ExceptionCode)
    {
        case IPV6_EXCP_LINKLOCAL:
	{
#if defined(IPV6_DEBUG)
            ix_ossl_message_log ("IPV6_EXCP_LINKLOCAL\n");
#endif /* IPV6_DEBUG */

	    /* Destination address is link-local. If the packets destination
	     * address matches the link-local address of the input port,
	     * send the packet to the stack driver. Else drop it.
	     */

	    err = _ix_cc_ipv6_process_link_local_pkt(arg_hBuffer, ipv6Addr);

            break;
	}

        case IPV6_EXCP_MTU:
        {
#if defined(IPV6_DEBUG)
            ix_ossl_message_log ("IPV6_EXCP_MTU\n");
#endif /* IPV6_DEBUG */

	    /* 
	     * Packet size exceeds MTU of the link, send an ICMP 
	     * "packet too big" message to the sender 
	     */

            err = ix_cc_ipv6_lookup_prefix(ipv6Addr,
	                                   &nextHopInfo,
		                           arg_pContext);
            if (err != IX_SUCCESS)
            {
                break;
            }

            _ix_cc_ipv6_icmp_pkt_too_big(arg_hBuffer, nextHopInfo.mtu);

            /* The input buffer has been freed by the above function call */
            return(IX_SUCCESS);
        }

	case IPV6_EXCP_BAD_HOPLIMIT:
	{
#if defined(IPV6_DEBUG)
            ix_ossl_message_log ("IPV6_EXCP_HOPLIMIT\n");
#endif /* IPV6_DEBUG */

	    /* Packet hop limit is invalid (less than 0). Send an 
	     * ICMP message to the sender of the packet.
	     */

#ifdef IPV6_STATS

           /* Increment the count of packets with bad headers. */
	   g_pIpv6Context->ipv6FwdStats.ipv6InCore[ingressPort].ipv6CoreInHdrErrors++;

#endif /* IPV6_STATS */
            err = _ix_cc_ipv6_icmp_time_exceeded(arg_hBuffer,
	                         IX_CC_ICMPV6_TE_CODE_HOP_LIMIT_EXCEEDED);
            /* The input buffer has been freed by the above function call */
            return(IX_SUCCESS);
	}

        case IPV6_EXCP_REDIRECT:
        {
#if defined(IPV6_DEBUG)
            ix_ossl_message_log ("IPV6_EXCP_REDIRECT\n");
#endif /* IPV6_DEBUG */

#if defined(ICMPV6_REDIRECT_IMPLEMENTED)

            /* 
    	     * Get IPv6 address from IP Header and call route lookup to get 
	     * next hop info from RTM 
	     */

            IX_ERROR_CR(ix_cc_ipv6_lookup_prefix(ipv6Addr,
	                                         &nextHopInfo,
					         arg_pContext));


            err = _ix_cc_ipv6_icmp_redirect(arg_hBuffer,
	                                    nextHopInfo.ipv6Addr);
#else
            
            /* For now we drop the packet */
    	    IX_ERROR_CR(ix_rm_buffer_free_chain(arg_hBuffer));

#endif /* ICMPV6_REDIRECT_IMPLEMENTED */
            /* The input buffer has been freed by the above function call */
            return(IX_SUCCESS);
        }

        case IPV6_EXCP_LKUPFAILED:
	{
#if defined(IPV6_DEBUG)
            ix_ossl_message_log ("IPV6_EXCP_LKUPFAILED\n");
#endif /* IPV6_DEBUG */

            /* 
	     * Lookup on the packet resulted in a NHID 0 and port -1, 
	     * OR 
	     * The IPV6_NH_FLAGS_DOWN bit is set.
	     *
	     * Lookup Failed. Send an ICMP message saying "Host Unreachable".
	     */

#ifdef IPV6_STATS
            /* 
	     * Increment the count of number of packets for which no route
	     * was found
	     */
	    g_pIpv6Context->ipv6FwdStats.ipv6InCore[ingressPort].ipv6CoreInNoRoutes++;
	     
#endif /* IPV6_STATS */
            err = _ix_cc_ipv6_icmp_dest_unreachable(arg_hBuffer, 
	                                 IX_CC_ICMPV6_DU_CODE_ADDR_UNREACHABLE);
            /* The input buffer has been freed by the above function call */
            return(IX_SUCCESS);
        }

        case IPV6_EXCP_MULTICAST:
	{
#if defined(IPV6_DEBUG)
            ix_ossl_message_log ("IPV6_EXCP_MULTICAST\n");
#endif /* IPV6_DEBUG */

	    /* 
	     * Increment the total number of multicast packets recd on
	     * this input port.
	     */
#ifdef IPV6_STATS
            
	    g_pIpv6Context->ipv6FwdStats.ipv6InCore[ingressPort].ipv6CoreInMulticast++;
	    g_pIpv6Context->ipv6FwdStats.ipv6InCore[ingressPort].ipv6CoreInMulticastOctets = g_pIpv6Context->ipv6FwdStats.ipv6InCore[ingressPort].ipv6CoreInMulticastOctets + IX_NTOH16(pHwMeta->m_PacketSize);

#endif /* IPV6_STATS */

	    /* Multicast packet, handle it appropriately */
            return _ix_cc_ipv6_forward(arg_hBuffer, 0, 0); 
        }

        default:
#if defined(IPV6_DEBUG)
            ix_ossl_message_log(" *** Unexpected EXCEPTION raised  *** \n");
#endif /* IPV6_DEBUG */

            IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hBuffer), 
	                                           IX_CC_ERROR_SEND_FAIL,0);

            return IX_ERROR_WARNING(IX_CC_ERROR_UNDEFINED_EXCEP,
                                               ("undefined exception code"));
        break;
    }/* end switch */
  
    if (err)
    {
        /* Free the buffer */
	IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hBuffer), 
	                                          IX_CC_ERROR_INTERNAL, 0);

        return err;
    }
    return IX_SUCCESS;
} /* end ix_cc_ipv6_microblock_low_priority_pkt_handler */


/**
 * NAME: ix_cc_ipv6_stkdrv_pkt_handler
 *
 * DESCRIPTION: This is the registered function to receive packets from 
 * stack driver core component. 
 *
 * @Param:  - IN arg_hBuffer: handle to a buffer which contains packet
 * from stack driver core component. - 
 *
 * @Param:  - IN arg_ExceptionCode:  will be Ignored 
 *
 * @Param:  - IN arg_pContext: pointer to IPV6 Forwarder core component context.
 * @Param:  - OUT none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_SEND_FAIL - error from CCI
 * IX_CC_ERROR_INTERNAL - an internal error
 * IX_CC_ERROR_ALIGN - pointer not aligned properly
 * IX_CC_ERROR_NULL - null pointer
 * 
 */

ix_error ix_cc_ipv6_stkdrv_pkt_handler(ix_buffer_handle arg_hBuffer,  
                                       ix_uint32 arg_ExceptionCode,
                                       void* arg_pContext)
{
    /* Call stkdrv function */
    return _ix_cc_ipv6_forward_pkts_from_stkdrvr(arg_hBuffer);
   
} /* end ix_cc_ipv6_stkdrv_pkt_handler */


/**
 * NAME: ix_cc_ipv6_common_pkt_handler
 *
 * DESCRIPTION: This is the registered function to receive packets 
 * from any core component other than stack driver.
 * 
 * @Param:  - IN arg_hBuffer: handle to a buffer which contains packet 
 *                            from any core component other than stack driver.
 * @Param:  - INarg_ExceptionCode:  will be Ignored
 * @Param:  - INarg_pContext: pointer to IPV6 Forwarder core component context 
 * @Param:  - OUT none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - an internal error
 * IX_CC_ERROR_ALIGN - pointer not aligned properly
 * IX_CC_ERROR_NULL - null pointer
 * IX_CC_ERROR_SEND_FAIL - error from CCI
 */

ix_error ix_cc_ipv6_common_pkt_handler(ix_buffer_handle arg_hBuffer,  
                                       ix_uint32 arg_ExceptionCode,
                                       void* arg_pContext)
{
    /* Call ipv6 forward function */
    return _ix_cc_ipv6_forward(arg_hBuffer, 1, 0); 
   
} /* end ix_cc_ipv6_common_pkt_handler */
