/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 6
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      IPv6 Forwarder Core Component - dynamic properties table management
 *
 * = FILENAME
 *      ix_cc_ipv6_prop.c
 *
 * = DESCRIPTION
 *      The file defines dynamic properties table management interface 
 *      functions.
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      12/03/2002
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_ipv6_prop.c,v 1.11 2003/11/01 02:16:54 rranjeet Exp $";

/**
 * System defined include files
 */

/**
 * User defined include files
 */

#include "ix_cc_error.h"

#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_netmacros.h"

#include "cc/ix_cc_rtmv6.h"

#include "ix_cc.h"
#include "ix_cc_properties.h"
#include "bindings.h"
#include "cc/ix_cc_ipv6.h"
#include "cc/ipv6/internal/ix_cc_ipv6_init.h"
#include "cc/ipv6/internal/ix_cc_ipv6_prop.h"
#include "cc/ipv6/internal/utils.h"


/**
 * Preprocessor symbols and Macros used in this file
 */

/**
 * Type definitions whose scope is limited to this file
 */

/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */

extern ix_cc_ipv6_context* g_pIpv6Context;
extern ix_cc_ipv6_registry_data g_Ipv6RegistryData;


/** 
 * Extern function prototypes.
 */


/** 
 * Static function prototypes.
 */

/** 
 * Function definitions.
 */


/**
 * NAME: _ix_cc_ipv6_get_port_ip_address
 *
 * DESCRIPTION: This primitive will be used to get an array of 
 *              IPv6 address of a port. The maximum number of addresses
 *              returned is defined by PROP_MAX_IPV6_ADDRESSES
 * 
 * @Param:  - IN arg_BladeId - blade number of ip address to be retrieved
 * @Param:  - IN arg_PortId - port number of ip address to be retrieved
 *
 * @Param:  - OUT arg_numElements - number of elements in the array.
 * @Param:  - IN OUT arg_pIpAddress - Array of IPv6 addresses of 
 *                                    given blade and port
 *
 * @Return: IX_SUCCESS if successfull or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM - invalid blade/port number
 * IX_CC_ERROR_ENTRY_NOT_FOUND - no data in location
 */
ix_error _ix_cc_ipv6_get_port_ip_address(ix_uint16 arg_BladeId,
                                         ix_uint16 arg_PortId, 
                                         ix_uint8   *arg_pNumElements,
                                         ix_uint128 *arg_pIpAddr)
{
    ix_uint32 keyIndex;
    ix_uint32 i;

    /* Validate Port and blade Ids */
    if((arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
       (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
    {
        return IX_ERROR_WARNING(IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM,
                 ("Invalid Blade Id % or Port Id %d",arg_BladeId,arg_PortId));

    }/* endif ((arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
      (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1)) */

    /* Zero out the IPv6 address array before doing anything else.
     * It is, of course, assumed that the caller has allocated an 
     * array of PROP_MAX_IPV6_ADDRESSES elements.
     */
    ix_ossl_memset(arg_pIpAddr, 0, 
                            sizeof(ix_uint128) * PROP_MAX_IPV6_ADDRESSES);

    /* Calculate the index of in the Property Table.
     * Each blade has 16 ports and there are a total of 64 blades.
     */

    keyIndex = arg_BladeId << 4 | arg_PortId;

    /* find out the number of elements in the list for this blade/port */
    *arg_pNumElements = 
                  g_pIpv6Context->ipv6PropTable[keyIndex].ipv6_num_elements;
     
    for (i = 0; i < *arg_pNumElements; i++)
    {
        IX_NTOH128(&(g_pIpv6Context->ipv6PropTable[keyIndex].ipv6_address[i]),
                    (arg_pIpAddr + i));
    }

    return IX_SUCCESS;

}/* end _ix_cc_ipv6_get_port_ip_address */

/**
 * NAME: _ix_cc_ipv6_prop_get_port_status
 *
 * DESCRIPTION: This primitive will be used to get status of local port
 *           .
 * 
 * @Param:  - IN arg_PortId - port number of status to be retrieved
 *
 * @Param:  - OUT arg_pPortStatus - status value of given port
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM - invalid port number
 * 
 */

ix_error _ix_cc_ipv6_get_port_status(ix_uint16 arg_PortId, 
                                    ix_uint32* arg_pPortStatus)
{
    ix_uint32 currentValue;
    ix_cc_ipv6_mb_ctrl* pData;

    /* Initialize arg_pPrortStatus with disable */
    *arg_pPortStatus = IX_CC_IPV6_DISABLE_PORT_STATUS;

    /* Validate Port Id */
    if(arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1)
    {
        return IX_ERROR_WARNING(IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM,
                                                     ("Invalid Port Id"));
    }/* endif (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1) */

    pData = g_pIpv6Context->pMbCbBase;

    currentValue = IX_IPV6_MEM_UINT32_READ(&pData->portForward);
    currentValue = IX_NTOH32(currentValue);

    *arg_pPortStatus = (currentValue >> arg_PortId) & 1;
   
    return IX_SUCCESS;
}/* end _ix_cc_ipv6_get_port_status */

/**
 * NAME: _ix_cc_ipv6_property_update_port_status
 *
 * DESCRIPTION: This primitive will be used to set status of local port.
 *           .
 * 
 * @Param:  - IN arg_PortId - port number 
 * @Param:  - IN arg_pPortStatus - status value of given port
 * @Param:  - OUT none
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM - invalid port number
 * 
 */
ix_error _ix_cc_ipv6_property_update_port_status(
                                    ix_uint16 arg_PortId, 
                                    ix_uint32 arg_PortStatus)
{

    ix_uint32 portMask;
    ix_cc_ipv6_mb_ctrl* pData;
    ix_uint32 data;
    
    /* Validate Port Id */
    if(arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1)
    {
        return IX_ERROR_WARNING(IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM,
                                ("Invalid Port Id"));
    }/* endif (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1) */

    portMask = 1 << (ix_uint32)arg_PortId;

    /* Calucate Port mask to set port status from pMbCbBase */
    pData = g_pIpv6Context->pMbCbBase;
    data = IX_IPV6_MEM_UINT32_READ(&pData->portForward);
    data = IX_NTOH32(data);
     
    if(arg_PortStatus == IX_CC_IPV6_ENABLE_PORT_STATUS)
    {
        data |= portMask;
    }/* end if (arg_PortStatus == IX_CC_IPV6_ENABLE_PORT_STATUS) */
    else
    {
        data &= ~portMask;
    }/* end else (arg_PortStatus == IX_CC_IPV6_ENABLE_PORT_STATUS) */

    /* write it back */
    data = IX_HTON32(data);

    ix_ossl_mutex_lock(g_pIpv6Context->ipv6Mutex,IX_OSSL_MUTEX_LOCK);

    IX_IPV6_MEM_UINT32_WRITE(&pData->portForward,data);

    ix_ossl_mutex_unlock(g_pIpv6Context->ipv6Mutex);
   
    return IX_SUCCESS;

}/* end ix_cc_ipv6_property_update_port_status */


/**
 * NAME: _ix_cc_ipv6_property_update_ip_address
 *
 * DESCRIPTION: This primitive will be used to add/delete ip address and 
 * net mask in dynamic property table and directed broadcast table
 *           
 * @Param:  - IN arg_PropId - property id : add/delete ip address
 * @Param:  - IN arg_BladeId - blade number
 * @Param:  - IN arg_PordId - port number
 * @Param:  - IN arg_Ipv6Addr - ip address of port
 *
 * @Param:  - OUT none
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM - invalid blade/port number.
 * 
 */

ix_error _ix_cc_ipv6_property_update_ip_address(ix_uint16 arg_PropId,
                                                ix_uint16 arg_BladeId,
                                                ix_uint16 arg_PortId,
                                                ix_uint128 arg_Ipv6Address)
{
    ix_uint32 keyIndex;
    ix_uint8 currentElements;
    ix_uint32 i;
    ix_uint128 tempIpv6Addr;
    ix_uint32 loop_var;

    /* Validate input arguments */
    if ((arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
        (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
    {
        return IX_ERROR_WARNING(IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM,
                                ("Invalid Blade Id or Port Id"));
    }/* endif (arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
      (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1)) */

    /* Calculate the index in the Property Table
     * There are 16 Ports per Blade and a total of 64 blades.
     */

#if defined(IPV6_DEBUG)
    ix_ossl_message_log ("For Blade (0x%x) Port (0x%x) \n", 
                                                     arg_BladeId, arg_PortId);
    ix_ossl_message_log ("The IpAddr is ... \n");
    _ix_cc_print_ipv6_address(&arg_Ipv6Address);
#endif /* IPV6_DEBUG */

    keyIndex = (arg_BladeId << 4) | arg_PortId;    
                                       
    currentElements = g_pIpv6Context->ipv6PropTable[keyIndex].ipv6_num_elements;

    if (arg_PropId == IX_CC_SET_PROPERTY_ADD_INTERFACE_IPV6)
    {
        if (currentElements == PROP_MAX_IPV6_ADDRESSES)
        {
            return IX_ERROR_WARNING(IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM,
                                ("There are %d elements already", 
				                  PROP_MAX_IPV6_ADDRESSES));
        }
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
	/* Check if this address is already present on this port */
        IX_HTON128(&arg_Ipv6Address, &tempIpv6Addr);
        for (loop_var = 0;
             loop_var < currentElements;
             loop_var++)
        {
            if(IS_IPV6_ADDRESS_EQUAL(
                   &g_pIpv6Context->ipv6PropTable[keyIndex].ipv6_address[loop_var],
                   &tempIpv6Addr))
            {
                return IX_ERROR_WARNING(IX_CC_IPV6_ERROR_DUPLICATE_ADDRESS_DETECTED,
                           ("Address already present"));
	    }
        }
#endif
        /* store ipv6 address in the dynamic property table */
        ix_ossl_mutex_lock(g_pIpv6Context->ipv6Mutex, IX_OSSL_WAIT_FOREVER);

	/* Copy the address */
        IX_HTON128(&arg_Ipv6Address, 
	&g_pIpv6Context->ipv6PropTable[keyIndex].ipv6_address[currentElements]);

        /* One more element has been added */
        g_pIpv6Context->ipv6PropTable[keyIndex].ipv6_num_elements++;

        ix_ossl_mutex_unlock(g_pIpv6Context->ipv6Mutex); 
    }/* end if (arg_PropId == IX_CC_PROPERTY_ADD_IPV6_ADDRESS)*/

    if (arg_PropId == IX_CC_SET_PROPERTY_DEL_INTERFACE_IPV6)
    {
        IX_HTON128(&arg_Ipv6Address, &tempIpv6Addr);
        for (i = 0; i < currentElements; i++)
	{
	    /* First look for the element */
            if(IS_IPV6_ADDRESS_EQUAL(
	            &g_pIpv6Context->ipv6PropTable[keyIndex].ipv6_address[i], 
		    &tempIpv6Addr))
            {
	        break;
	    }
	}

	if (i == currentElements)
	{
	    /* end of loop reached, element doesn't exists in the prop table */
            return IX_ERROR_WARNING(IX_CC_IPV6_ERROR_INVALID_INPUT_PARAM,
                                              ("Ipv6 address doesn't exist"));
	}

        ix_ossl_mutex_lock(g_pIpv6Context->ipv6Mutex, IX_OSSL_WAIT_FOREVER);

	if (i == currentElements - 1)
	{
	   /* This was the last element in the list. Zero it out and 
	    * decrement the number of elements
	    */
            ZERO_IPV6_ADDRESS(
	         &g_pIpv6Context->ipv6PropTable[keyIndex].ipv6_address[i]);
	}
	else
	{
	   /* The element is in the middle of this array and the code
	    * always expects this to be contigous. So move the last element
	    * in this ones place and zero out the last element.
	    */
           COPY_IPV6_ADDRESS(&g_pIpv6Context->ipv6PropTable[keyIndex].ipv6_address[currentElements -1], &g_pIpv6Context->ipv6PropTable[keyIndex].ipv6_address[i]);
            ZERO_IPV6_ADDRESS(
	         &g_pIpv6Context->ipv6PropTable[keyIndex].ipv6_address[currentElements - 1]);
	}

	/* One more element has been added */
	g_pIpv6Context->ipv6PropTable[keyIndex].ipv6_num_elements--;

        ix_ossl_mutex_unlock(g_pIpv6Context->ipv6Mutex);

    }/* endif  */
 
    return IX_SUCCESS;    

}/* end _ix_cc_ipv6_property_update_ip_address */


/**
 * NAME: _ix_cc_ipv6_prop_table_lookup_ip_address
 *
 * DESCRIPTION: This primitive will be used to lookup the ipv6 address
 * in dynamic property table
 *           
 * @Param:  - IN arg_IpAddr - ip address(non zero) to be searched in table.
 *
 * @Param:  - OUT none
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - no entry in table 
 * 
 */

ix_error _ix_cc_ipv6_prop_table_lookup_ip_address(ix_uint128 arg_IpAddr)
{
    ix_uint32 keyIndex;
    ix_uint128 ipv6_address;
    ix_uint8 currentElements;
    ix_uint32 i;
    
    /* Convert the IP address into network byte order */   
    IX_HTON128(&arg_IpAddr, &ipv6_address);

    for(keyIndex = 0; keyIndex < IPV6_TOTAL_NUMBER_OF_PORTS; keyIndex++)
    {
        currentElements = 
	         g_pIpv6Context->ipv6PropTable[keyIndex].ipv6_num_elements; 

        for (i = 0; i < currentElements; i++)
	{
            if(IS_IPV6_ADDRESS_EQUAL(
	            &g_pIpv6Context->ipv6PropTable[keyIndex].ipv6_address[i],
		    &ipv6_address))
            {
                return IX_SUCCESS;

            }/* end if */

        }
    } /* end of table search */
    return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,("no entry in table"));

} /* end _ix_cc_ipv6_prop_table_lookup_ip_address */
