/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 6
 *
 * = MODULE
 *      IPv6 Forwarder Core Component
 *
 * = FILENAME
 *      v6_header_validation.c
 *
 * = DESCRIPTION
 *      This file contains the implementation of IP header validation 
 *      and address validation functions according to RFC2460 and RFC2373
 *
 *
 * = CREATION TIME
 *      12/12/2002
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */

#define IX_ERROR_FILE_IDENT "$Id: v6_header_validation.c,v 1.16 2003/12/12 00:49:54 ktseng Exp $";

/**
 * System defined include files required.
 */

/**
 * User defined include files required.

 */
#include "ix_cc_error.h"
#include "cc/ipv6/internal/v6_header_validation.h"
#include "cc/ipv6/internal/ix_cc_ipv6_nd.h"
#include "cc/ipv6/internal/ix_cc_ipv6_icmp_nd.h"
#include "cc/ipv6/internal/utils.h"
#include "cc/ipv6/internal/icmp6.h"
#include "cc/ix_cc_ipv6.h"


/**
 * Function definitions.
 */
/* Function to process a link local packet */
extern ix_error _ix_cc_ipv6_process_link_local_pkt(ix_buffer_handle arg_hBuffer,
                                          ix_uint128 arg_ipv6Addr);

/**
 * NAME: _ix_cc_ipv6_IPheader_validate
 *
 * DESCRIPTION: This function validates the IP header according to RFC 2460
 * 
 * @Param:  - IN  arg_pIpHdr  - a pointer to IPv6 header
 *                arg_hBuffer - a buffer containing the packet with IP header
 *                arg_inCounters - indicates "maintain in counters"
 *                arg_outCounters - indicated "maintain out counters"
 *
 * @Param:  - INOUT  None
 * @Param:  - OUT arg_pDropped - value to tell whether the packet should be 
 *                               dropped or forwarded  
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token
 *             for failure.
 *
 * Error codes: IX_CC_ERROR_INTERNAL
 */
ix_error _ix_cc_ipv6_IPheader_validate(ix_cc_ipv6_header *arg_pIpHdr,
                                       ix_buffer_handle arg_hBuffer,
                                       ix_uint32 *arg_pDropped, 
                                       ix_uint8 arg_inCounters,
                                       ix_uint8 arg_outCounters)
{
#ifdef IPV6_STATS
    ix_error err = IX_SUCCESS;
    ix_uint16 ingressPort;
    ix_hw_buffer_meta *pHwMeta;
#endif /* IPV6_STATS */

    *arg_pDropped = 0;

    if (arg_pIpHdr->version != 6)
    {
        if (arg_inCounters)
        {
#ifdef IPV6_STATS

            /* Get the meta data corresponding to this packet */
            err = ix_rm_buffer_get_meta(arg_hBuffer, (void **)&pHwMeta);
            if (err)
            {
                goto drop;
            }

            /* 
             * Get the input interface 
             * Note: m_InputPort is ix_uint16 
             */

            ingressPort = IX_NTOH16(pHwMeta->m_InputPort);

            /* Increment count of packets with Header Errors */
            g_pIpv6Context->ipv6FwdStats.ipv6InCore[ingressPort].ipv6CoreInHdrErrors++;

#endif /* IPV6_STATS */
        }
        /* Invalid version in the header */
        goto drop;
    }

    /* packet header forward it */
    return IX_SUCCESS;

drop:

    *arg_pDropped = 1;
    
    /* need to delete this buffer */
    IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hBuffer), IX_CC_ERROR_INTERNAL, 0);

    return IX_SUCCESS;

}


/**
 * NAME:_ix_cc_ipv6_validate_IP_addresses 
 *
 * DESCRIPTION: This function checks the Destination IPv6 address and Source 
 *              IPv6 address in the given buffer to make sure that they do 
 *              not contain any unrouteable addresses according to RFC 2373
 * 
 * @Param:  - IN  arg_pIpHdr  - a pointer to IPv6 header
 *                arg_hBuffer - a buffer containing the packet with IPv6 header
 *
 * @Param:  - OUT arg_pDropped - value to tell whether the packet
 *                               should be dropped or forwarded  
 *
 * @Return: IX_SUCCESS if successful 
 *          valid token if failure
 *
 * Error codes: IX_CCI_ERR_SEND_FAIL
 *              IX_CC_ERROR_INTERNAL
 */
ix_error _ix_cc_ipv6_validate_IP_addresses(ix_cc_ipv6_header *arg_pIpHdr, 
                                           ix_buffer_handle arg_hBuffer,
                                           ix_uint32 *arg_pDropped, 
                                           ix_uint8 arg_inCounters,
                                           ix_uint8 arg_outCounters)
{
    ix_uint128 source_addr;
    ix_uint128 dest_addr;
#if defined(RFC2373_CHECKS)
    ix_hw_buffer_meta *pHwMeta;
    ix_uint32 stackDeliv = 0;
    ix_uint32 neighDeliv = 0;
    ix_error err;    
    ix_uint16 ingressPort;
#endif

    /* default */
    *arg_pDropped = 0;

    /* Convert from Network Byte to Host Byte Order */
    IX_NTOH128(&arg_pIpHdr->srcAddress, &source_addr);
    IX_NTOH128(&arg_pIpHdr->dstAddress, &dest_addr);


#if defined(IPV6_DEBUG)
    /* 
     * For debug purposes, print the source and destination addresses
     */
    ix_ossl_message_log("source and destination address in the header \n");
    _ix_cc_print_ipv6_address (&source_addr);
    _ix_cc_print_ipv6_address (&dest_addr);
#endif /* IPV6_DEBUG */
    

    /* RFC 2373 checks:
     *
     * 1. If the source address is a Multicast address, DROP
     * 2. If the source address is loopback (::1), DROP
     * 3. If the destination address is loopback (::1), DROP
     * 4. If the destination address is 0, DROP
     * 5. If the destination address is Multicast -
     *    - If it is "all node NL" or "all router" multicast - STACK DRIVER
     *    - If it is "all node LL" or "Solicited Node" multicast - 
     *                                                   NEIGHBOR DISCOVERY CC
     *    - Else, DROP
     * 6. If the destination address is link local, cannot forward this 
     *    packet to other links. In this case the following is done:
     *    - If it is addressed to this node:
     *        - If NBR SOl or NBR Adv, send it to NEIGHBOR DISCOVERY
     *        - Else send to stack driver.
     *    - If not addressed to this node - DROP
     * 7. If the destination address is site local, cannot forward this
     *    packet to other sites - DROP.
     */
    
#if defined(RFC2373_CHECKS)

    /* Get the meta data corresponding to this packet */
    err = ix_rm_buffer_get_meta(arg_hBuffer, (void **)&pHwMeta);
    if (err)
    {
        goto cleanup;
    }

    /* Get the input interface */
    ingressPort = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_InputPort));

    /* Case 1 and Case 2 */
    if (IS_IPV6_ADDRESS_MULTICAST(source_addr) ||
        IS_IPV6_ADDRESS_LOOPBACK(source_addr))
    {
#if defined(IPV6_DEBUG)
    ix_ossl_message_log("source address is either loopback or multicast, dropping pkt\n");
#endif /* IPV6_DEBUG */
        goto cleanup;
    }

    /* Case 3 and Case 4 */
    if (IS_IPV6_ADDRESS_LOOPBACK(dest_addr)    ||
        IS_IPV6_ADDRESS_ZERO(&dest_addr))
    {
#ifdef IPV6_STATS
    /* Increment the count of packets with Address Errors */
    g_pIpv6Context->ipv6FwdStats.ipv6InCore[ingressPort].ipv6CoreInAddrErrors++;
#endif /* IPV6_STATS */

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("dest address is either loopback or zero, dropping pkt \n");
#endif /* IPV6_DEBUG */

        goto cleanup;
    }

    /* Case 6 */
    if (IS_IPV6_ADDRESS_LINK_LOCAL(dest_addr))
    {
#if defined(IPV6_DEBUG)
    ix_ossl_message_log("dest address link local, processing.. pkt \n");
#endif /* IPV6_DEBUG */

        /* Check if the packet is for local delivery */
        err = _ix_cc_ipv6_process_link_local_pkt(arg_hBuffer, dest_addr);
        if (err != IX_SUCCESS)
        {
#ifdef IPV6_STATS
            /* Increment the count of packets with Address Errors */
            g_pIpv6Context->ipv6FwdStats.ipv6InCore[ingressPort].ipv6CoreInAddrErrors++;
#endif /* IPV6_STATS */
            goto cleanup;
        }
        else
        {
            /* Simply return from the Ipv6 forwarder */
            *arg_pDropped = 1;
            return IX_SUCCESS;
        }
    }

    /* 
     * Checks for cases 5 
     *
     * If dest is "all router" (node-local, link-local or site-local)
     * multicast, send it to the stack driver.
     */
    if ((IS_IPV6_ADDRESS_ALL_ROUTER_NL_MULTICAST(dest_addr)) || 
        (IS_IPV6_ADDRESS_ALL_ROUTER_LL_MULTICAST(dest_addr)) ||
        (IS_IPV6_ADDRESS_ALL_ROUTER_SL_MULTICAST(dest_addr)))
    {
#if defined(SEND_ROUTER_ADV)
       neighDeliv = 1;
#else
       stackDeliv = 1;
#endif /* SEND_ROUTE_ADV */
        goto deliver;
    } 

    /* 
     * More checks for cases 5 
     *
     * If dest is "all node" (node-local) multicast
     * send to stack driver.
     */
    if (IS_IPV6_ADDRESS_ALL_NODE_NL_MULTICAST(dest_addr))
    {
        stackDeliv = 1;
        goto deliver;
    }

    if (IS_IPV6_ADDRESS_ALL_NODE_LL_MULTICAST(dest_addr))
    {
        /* deliver to stack driver */
        neighDeliv = 1;
        goto deliver;
    }

    /* 
     * If dest is a "solicited node" multicast address, send it to
     * neighbor discovery.
     */
    err = _ix_cc_ipv6_check_for_solnode_address(arg_hBuffer, dest_addr);

    if (err == IX_SUCCESS)
    {
        /* deliver to neighbor discovery */
        neighDeliv = 1;
        goto deliver;
    }
    else if (err != IX_CC_ERROR_ENTRY_NOT_FOUND)
    {
       goto cleanup;
    }

    if (IS_IPV6_ADDRESS_MULTICAST(dest_addr))
    {
#if defined(IPV6_DEBUG)
    ix_ossl_message_log("destination is a multicast we don't care about, dropping pkt\n");
#endif /* IPV6_DEBUG */

       /* 
        * Destination is a multicast address we don't care about, simply
        * drop the packet 
        */
       goto cleanup;
    }

#endif /* RFC2373_CHECKS */

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("returning from validate_addresses successfully \n");
#endif /* IPV6_DEBUG */

    /* Simply return to the IPv6 forwarder and let it deal with the packet */
    return IX_SUCCESS;

#if defined(RFC2373_CHECKS)
deliver:

#ifdef IPV6_STATS
    /* Increment the count of packets  delivered to upper layer protocols */
    g_pIpv6Context->ipv6FwdStats.ipv6InCore[ingressPort].ipv6CoreInDelivers++;
#endif /* IPV6_STATS */

    if (stackDeliv)
    {
#if defined(IPV6_DEBUG)
        ix_ossl_message_log("Delivering packet to stack driver \n");
#endif /* IPV6_DEBUG */

        /* After this, simply return from the IPv6 forwarder */
        *arg_pDropped = 1;

        /* Local delivery - send through high priority queue */
        err = ix_rm_message_send(IX_CC_IPV6_PKT_HIGH_PRIORITY_STKDRV_OUTPUT,
                                 arg_hBuffer, 0);
        if (err)
        {
            /* Free the buffer chain and return error to caller */
            IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hBuffer),
                                IX_CC_ERROR_INTERNAL, 0);

            return IX_ERROR_CHAIN(err, IX_CC_ERROR_SEND_FAIL,0, 
                              ("Failed to send packet to next target\n"));
        }     
    }

    if (neighDeliv)
    {
#if defined(IPV6_DEBUG)
        ix_ossl_message_log("Delivering packet to neighbor discovery \n");
#endif /* IPV6_DEBUG */

        /* After this, simply return from the IPv6 forwarder */
        *arg_pDropped = 1;

        /* Send packet to Neighbor Discovery for furthur handling */
        err = ix_cc_nd_process_incoming_packet(arg_hBuffer);
        if (err)
        {
            /* Free the buffer and return error to caller */
            IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hBuffer),
                                IX_CC_ERROR_INTERNAL, 0);

            return IX_ERROR_CHAIN(err, IX_CC_ERROR_SEND_FAIL,0, 
                                  ("Failed to send packet to next target\n"));
        }
    }

    /* Packet processing complete, return to the IPv6 forwarder */
    return IX_SUCCESS;

cleanup:

    /* 
     * Common cleanup area, free the buffer chain and return from 
     * the function
     */

    IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hBuffer), 
                                                IX_CC_ERROR_INTERNAL, 0);

    /* After this, simply return from the IPv6 forwarder */
    *arg_pDropped = 1;

    return IX_SUCCESS;
#endif    
}

/**
 * NAME:_ix_cc_ipv6_check_for_solnode_address
 *
 * DESCRIPTION: This function checks the Destination
 *              IPv6 address is a Solicited Node Address as defined
 *              in RFC 2373.
 *
 * @Param:  - IN  arg_destAddr - The buffer containing the IPv6 packet.
 * @Param:  - IN  arg_destAddr - The Destination IPv6 address that needs
 *                               to be checked.
 *
 * @Return: IX_SUCCESS - If it is a solicited node multicast address.
 *          IX_CC_ERROR_ENTRY_NOT_FOUND - If it is not a multicast address.
 *          IX_CC_ERROR_INTERNAL - Internal error.
 */

ix_error _ix_cc_ipv6_check_for_solnode_address(ix_buffer_handle arg_hBuffer,
                                                ix_uint128 arg_destAddr)
{
    ix_uint128 portIpv6Addr[PROP_MAX_IPV6_ADDRESSES];
    ix_uint8 numElements = 0;
    ix_hw_buffer_meta *pHwMeta;
    ix_uint16 portNum;
    ix_uint8 i;

    /* 
     * Get all the IPv6 addresses for the port on which this packet was
     * received.
     */
    
    IX_ERROR_CRT(ix_rm_buffer_get_meta(arg_hBuffer,
                                       (void **)&pHwMeta),
                                       IX_CC_ERROR_INTERNAL,
                 0);

    portNum = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_InputPort));

    /* Get the list of Ipv6 addresses for this port */
    IX_ERROR_CRT(_ix_cc_ipv6_get_port_ip_address(THIS_BLADE_ID,
                                                 portNum,
                         &numElements,
                         portIpv6Addr),
                         IX_CC_ERROR_INTERNAL,
                         0);

    for (i = 0; i < numElements; i++)
    {
        /* check to see if this qualifies as a solicited node address */
        if(IS_IPV6_ADDRESS_SOLICITED_NODE_MULTICAST(arg_destAddr, 
                                                    portIpv6Addr[i]))
        {
            return IX_SUCCESS;
        }
    }

#if defined(IPV6_DEBUG)
    ix_ossl_message_log("Entry was not a solicited node address \n");
#endif /*IPV6_DEBUG */

    return IX_CC_ERROR_ENTRY_NOT_FOUND;
}

/**
 * NAME: _ix_cc_ipv6_inspect_pkt_hdr
 *
 * DESCRIPTION: This function inspects an IPv6 packet header for possible
 *              extension headers and calls the appropriate functions to 
 *              deal with the headers it finds.
 *
 * @Param:  - IN  arg_hBuffer - The buffer containing the IPv6 packet.
 * @Param:  - OUT arg_pDrop -   If 1, the packet has been dropped. The caller
 *                             should simply return.
 *
 * @Return: IX_SUCCESS - If successful.
 */
ix_error _ix_cc_ipv6_inspect_pkt_hdr(ix_buffer_handle arg_hBuffer,
                                     ix_uint8 *arg_pDrop)
{
    ix_buffer_handle  curBufHdl, newBufHdl;
    ix_hw_buffer_meta *pHwMeta;
    void *pData, *pCurPosn;
    ix_uint32 pktLen, curDataLen;
    ix_cc_ipv6_header *pIpv6Header;
    ix_cc_ipv6_options_hdr_common *pIpOptHdr;
    ix_uint8 nextHeader;
    ix_uint8 found = FALSE;
    ix_uint8 ipHdr = TRUE;
    ix_cc_ipv6_icmp_header *pIcmpHdr;
    ix_uint32 bytesRemaining = 0; /* used only if a header spans buffers */
    char buf[256];           /* used only of a header spans buffers */
    void *bufPtr = buf;      /* points to the current location of the buffer */
    ix_uint32 offset_of_param = 0;

    IX_ERROR_CRT(ix_rm_buffer_get_meta(arg_hBuffer,
                                       (void **)&pHwMeta),
                                       IX_CC_ERROR_INTERNAL,
                                       0);

    pktLen = IX_NTOH16(pHwMeta->m_PacketSize);
    curDataLen = IX_NTOH16(pHwMeta->m_BufferSize);
    curBufHdl = arg_hBuffer;
    
    while(pktLen > 0)
    {
        pktLen -= curDataLen;
        IX_ERROR_CRT(ix_cc_hw_get_packet_data(curBufHdl, 
                                              &pData),
                                              IX_CC_ERROR_INTERNAL, 
                                              0);
        pCurPosn = pData;
        
        /* In case any header was not completely included in a buffer, it 
         * got copied into buf. We now need to copy the remaining portion 
         * of it here.
         */

        if (bytesRemaining)
        {
           ix_ossl_memcpy(bufPtr, pCurPosn, bytesRemaining);
           bufPtr = buf;
        }
       
        /* The first header in the packet will be the IPv6 header */
        if(ipHdr)
        {
            /* If the header was split across multiple buffers, it will
             * be reassembled in 'buf'. In this case, use bufPtr to 
             * access the header
             */
            if(bytesRemaining)
            {
                pIpv6Header = (ix_cc_ipv6_header *)bufPtr;
                pCurPosn += bytesRemaining;
                curDataLen -= bytesRemaining;
            }
            else
            {
                pIpv6Header = (ix_cc_ipv6_header *)pCurPosn;
                pCurPosn += IX_CC_IPV6_HDR_LEN;
                curDataLen -= IX_CC_IPV6_HDR_LEN;
            }

            nextHeader = pIpv6Header->nextHeader;
            ipHdr = FALSE;
        }
        else
        {
            /* If the header was split across multiple buffers, it will
             * be reassembled in 'buf'. In this case, use bufPtr to 
             * access the header
             */
            if(bytesRemaining)
            {
                pIpOptHdr = (ix_cc_ipv6_options_hdr_common*)bufPtr;
                pCurPosn += bytesRemaining;
                curDataLen -= bytesRemaining;
            }
            else
            {
                pIpOptHdr = (ix_cc_ipv6_options_hdr_common*)pCurPosn;
                pCurPosn += (1+IX_NTOH16(pIpOptHdr->hdrLength))*8;
                curDataLen -= (1+IX_NTOH16(pIpOptHdr->hdrLength))*8;
            }

            nextHeader = pIpOptHdr->nextHeader;
        }

        /* 
         * Walk the headers untill either desired header is reached,
         * or there are no more headers
         */
        while(curDataLen && (nextHeader != IX_CC_IPV6_NEXT_HDR_NONE) && !found)
        {
            switch(nextHeader)
            {
                case IX_CC_IPV6_NEXT_HDR_HOPBYHOP: /*fall thru*/
                case IX_CC_IPV6_NEXT_HDR_ROUTING:  /*fall thru*/
                case IX_CC_IPV6_NEXT_HDR_FRAGMENT: /*fall thru*/
                case IX_CC_IPV6_NEXT_HDR_DSTOPT:

                    /* Note: In the current implementation we don't handle
                     * the above extension headers - even if they preceed 
                     * an ICMPv6 header in the packet. 
                     *
                     * As defined in RFC 2460 we simply send an ICMPv6 error 
                     * code - "Unrecognized header type encountered" to the 
                     * source.
                     *
                     * Note: We don't "walk" the list of extension headers. 
                     * We simply look at the next header
                     * field in the IPv6 header (which is at offset 6) and 
                     * send an ICMP error message with that offset as a 
                     * parameter.
                     */

                    offset_of_param = 6;   /* offset of the next header field */
                    if (_ix_cc_ipv6_icmp_param_problem(arg_hBuffer,
                                         IX_CC_ICMPV6_PP_CODE_NEXT_HDR_UNKNOWN,
                                         offset_of_param) != IX_SUCCESS)
                    {
#if defined(IPV6_DEBUG)
                        ix_ossl_message_log("Failed to send ICMP Param Problem error \n");
#endif /* IPV6_DEBUG */
                    }

                    /* The buffer will be dropped by the ICMP function called above,
                     * nothing to do, simply return
                     */

                    *arg_pDrop = 1;
                    return IX_SUCCESS;

                case IX_CC_IPV6_NEXT_HDR_ICMP:
                    found = TRUE;
                    break;

                default:
                    curDataLen = 0; /*This will break out of the loop*/
                    pktLen = 0;
                    break;
            } /* switch */
        } /* while */

        if((nextHeader == IX_CC_IPV6_NEXT_HDR_NONE) || found)
            break;

        IX_ERROR_CRT(ix_rm_buffer_get_next( curBufHdl, &newBufHdl),
	                         IX_CC_IPV6_ERROR_CCI, 0);

        if(newBufHdl == IX_NULL_BUFFER_HANDLE)
	    break; /*no more buffers in the chain*/

        curBufHdl = newBufHdl;

	    IX_ERROR_CRT(ix_rm_buffer_get_meta( curBufHdl, (void **)&pHwMeta),
	                         IX_CC_IPV6_ERROR_CCI, 0);

        curDataLen = IX_NTOH16(pHwMeta->m_BufferSize);

    } /* while */

    if(!found)
    {
        /* No extension or ICMP headers were found, continue processing pkt */
        return IX_SUCCESS;
    }

    /*
     * If an ICMP header was found, we need to check if it contains a 
     * neighbor discovery related packet
     */

    pIcmpHdr = (ix_cc_ipv6_icmp_header *)pCurPosn;

    if((pIcmpHdr->type == IX_CC_IPV6_ICMP_TYPE_NBR_SOL) ||
       (pIcmpHdr->type == IX_CC_IPV6_ICMP_TYPE_NBR_ADV))
    {
        *arg_pDrop = 1;
        return ix_cc_nd_process_incoming_packet(arg_hBuffer);
    }
    else 
    if((pIcmpHdr->type == IX_CC_IPV6_ICMP_TYPE_RTR_SOL) ||
       (pIcmpHdr->type == IX_CC_IPV6_ICMP_TYPE_RTR_ADV))
    {

#if defined(SEND_ROUTER_ADV)
        *arg_pDrop = 1;
        return ix_cc_nd_process_incoming_packet(arg_hBuffer);
#endif

    }

    return IX_SUCCESS;
}
