/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2003 Intel Corporation. All rights reserved.
 * 
 * = MODULE
 *     utility functions.
 *
 * = FILENAME
 *      utils.c
 *
 * = DESCRIPTION
 *   This file contains a set of utility functions to calculate to assist
 *   the IPv6 forwarder core component with the forwarding process.
 *
 * = CREATION TIME
 *      12/12/2002 
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */


#define IX_ERROR_FILE_IDENT "$Id: v6_utils.c,v 1.2 2003/11/01 01:05:49 rranjeet Exp $";

/**
 * System defined include files required.
 */
/**
 * User defined include files required.
 */

#include "ix_cc_error.h"
#include "cc/ipv6/internal/utils.h"

/**
 * Preprocessor symbols and macros used in this file.
 */

/**
 * Types definitions whose scope is limited to this file.
 */

/**
 * Variable declarations global to this file only.  Externs are followed by
 * static variables.
 */

/* Initialize the IPv6 header from the data sent in. */
void _ix_cc_ipv6_init_ip_header(ix_uint8 *arg_pData, 
                                ix_cc_ipv6_header *arg_pIpHdr)
{

    ix_ossl_memcpy(arg_pIpHdr, arg_pData, IX_CC_IPV6_HDR_LEN);
}

#if defined(IPV6_DEBUG)
/* This function prints an IPv6 address */
void _ix_cc_print_ipv6_address(ix_uint128 *pAddr)
{
    ix_uint8 i;

    for (i = 0; i < 4; i++)
    {
        ix_ossl_message_log("IPv6[%d] %lx \n", i, pAddr->aUint32[i]);
    }

    return;
}

/* This function prints the contents of an IPv6 header */
void _ix_cc_print_ipv6_header(ix_cc_ipv6_header *arg_pHdr)
{
    
    ix_ossl_message_log("IPv6 Header: \n");
    ix_ossl_message_log("Version (0x%x) \n", arg_pHdr->version);
    ix_ossl_message_log("Traffic Class (0x%x) \n", arg_pHdr->traffic_class);
    ix_ossl_message_log("flow label (0x%x) \n", arg_pHdr->flow_label);
    ix_ossl_message_log("Pay load length (0x%x) \n", arg_pHdr->payloadLength);
    ix_ossl_message_log("Next header (0x%x) \n", arg_pHdr->nextHeader);
    ix_ossl_message_log("Hop limit (0x%x) \n", arg_pHdr->hopLimit);

    ix_ossl_message_log("Source Address : \n");
    _ix_cc_print_ipv6_address(&arg_pHdr->srcAddress);

    ix_ossl_message_log("Destination Address : \n");
    _ix_cc_print_ipv6_address(&arg_pHdr->dstAddress);

    return;
}
#endif /* IPV6_DEBUG */


/* NAME: IS_IPV6_ADDRESS_ZERO
 *
 * DESCRIPTION: Check to see if an IPv6 address is zero.
 *
 * @Param: IN arg_src - Pointer to the IPv6 address.
 *
 * return: 1 if it is.
 *         0 Otherwise
 * 
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
ix_uint8
IS_IPV6_ADDRESS_ZERO(ix_uint128 *arg_ipv6Addr)
{
    ix_uint8 i;
    ix_uint32 value;
    
    for (i = 0; i < 4; i++)
    {
	value = IX_IPV6_MEM_UINT32_READ(&arg_ipv6Addr->aUint32[i])
        if (value != 0x0)
	{
	    return 0;   /* FALSE */
	}
    }

    return 1;   /* TRUE */
}


/* NAME: ZERO_IPV6_ADDRESS
 *
 * DESCRIPTION: Zero's out an IPv6 address
 *
 * @Param: IN arg_src - Pointer to the IPv6 address.
 *
 * return: none.
 *
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
void
ZERO_IPV6_ADDRESS(ix_uint128 *arg_ipv6Addr)
{
    ix_uint8 i;

    for (i = 0; i < 4; i++)
    {
         IX_IPV6_MEM_UINT32_WRITE(&arg_ipv6Addr->aUint32[i], 0x0)
    }

    return;
}

/* NAME: COPY_IPV6_ADDRESS
 *
 * DESCRIPTION: Copies the source IPv6 address into the destination.
 *
 * @Param: IN arg_dst - Pointer to the destination IPv6 address.
 * @Param: IN arg_src - Pointer to the source IPv6 address.
 *
 * return: none.
 *
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
void
COPY_IPV6_ADDRESS(ix_uint128 *arg_dst, ix_uint128 *arg_src)
{
    ix_uint8 i;

    for (i = 0; i < 4; i++)
    {
        arg_dst->aUint32[i] = IX_IPV6_MEM_UINT32_READ(&arg_src->aUint32[i])
    }

    return;
}

/* 
 * NAME: IS_IPV6_ADDRESS_EQUAL
 *
 * DESCRIPTION: Compare two IPv6 addresses. 
 *
 * @Param: IN arg_src - Pointer to the IPv6 address.
 * @Param: OUT arg_dst - Pointer to where the result of the operation 
 *                        should be stored.
 * return: 1 if they are equal
 *         0 Otherwise
 *
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
ix_uint8
IS_IPV6_ADDRESS_EQUAL(ix_uint128 *arg_src, ix_uint128 *arg_dst)
{
    ix_uint8 i;
    ix_uint32 value1;
    ix_uint32 value2;

    for (i = 0; i < 4; i++)
    {
        value1 = IX_IPV6_MEM_UINT32_READ(&arg_src->aUint32[i])
        value2 = IX_IPV6_MEM_UINT32_READ(&arg_dst->aUint32[i])

        if (value1 != value2)
        {
	    return 0;    /* FALSE */
	}
    }

    return 1;
}

/* 
 * NAME: IS_IPV6_ADDRESS_MUTLICAST
 *
 * DESCRIPTION: Checks to see if an IPv6 address is Multicast. 
 * If the Most Significant Byte of the address is 0xFF, it is.
 *
 * @Param: IN arg_src - The IPv6 address.
 *
 * return: 1 if it is a Multicast address.
 *         0 Otherwise
 *
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
ix_uint8
IS_IPV6_ADDRESS_MULTICAST(ix_uint128 ipv6Addr)
{
    /* If bits 127 - 96 are 0xFFXXXXXX, return TRUE */
    return ((IX_127_96(ipv6Addr) >> 24) == 0xFF);
}

/* 
 * NAME: IS_IPV6_ADDRESS_LINK_LOCAL
 *
 * DESCRIPTION: Checks to see if an IPv6 address is Link Local
 *
 * @Param: IN arg_src - The IPv6 address.
 *
 * return: 1 if it is a Link Local
 *         0 Otherwise
 *
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
ix_uint8
IS_IPV6_ADDRESS_LINK_LOCAL(ix_uint128 ipv6Addr)
{
    /* The "prefix bits" have to be 1111 1110 10 */
    return (((IX_127_96(ipv6Addr) & 0xFFC00000) >> 22) == 0x3FA);
}

/* 
 * NAME: IS_IPV6_ADDRESS_SITE_LOCAL
 *
 * DESCRIPTION: Checks to see if an IPv6 address is Site Local
 *
 * @Param: IN arg_src - The IPv6 address.
 *
 * return: 1 if it is a Site Local
 *         0 Otherwise
 *
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
ix_uint8
IS_IPV6_ADDRESS_SITE_LOCAL(ix_uint128 ipv6Addr)
{
    /* The "prefix bits" have to be 1111 1110 11 */
    return (((IX_127_96(ipv6Addr) & 0xFFC00000) >> 22) == 0x3FB);
}


/* 
 * NAME: IS_IPV6_ADDRESS_LOOPBACK
 *
 * DESCRIPTION: Checks to see if an IPv6 address is Loopback (::1). 
 *
 * @Param: IN arg_src - The IPv6 address.
 *
 * return: 1 if it is a Multicast address.
 *         0 Otherwise
 *
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
ix_uint8
IS_IPV6_ADDRESS_LOOPBACK(ix_uint128 ipv6Addr)
{
    /* 
     * This is what a loopback address looks like 
     * 0000:0000:0000:0000:0000:0000:0000:0001
     */

    if ((IX_127_96(ipv6Addr) != 0x0) ||
        (IX_95_64(ipv6Addr) != 0x0)  ||
        (IX_63_32(ipv6Addr) != 0x0))
    {
        return 0;    /* FALSE */
    }

    /* Make sure the bottom 32 bits are 0x00000001 */
    if (IX_31_0(ipv6Addr) != 0x1)
    {
        return 0;    /* FALSE */
    }

    return 1;
}


/* Reference: Per RFC 2373 - This is what a multicast address looks like
 *     
 *   8 bits      4        4                   112 bits
 *  ---------------------------------------------------------------------
 *  |11111111| 000T | scope |            group ID                      |
 *  ---------------------------------------------------------------------
 * A multicast address can be "well known" (T = 0) or transient (T = 1).
 *
 * Based on the above definition:
 * 
 * "all nodes" multicast addresses can be 
 *        FF01:0:0:0:0:0:0:1  (node-local scope)
 *        FF02:0:0:0:0:0:0:1  (link-local scope)
 *
 * "all routers" multicast addresses can be
 *        FF01:0:0:0:0:0:0:2  (node-local scope)
 *        FF02:0:0:0:0:0:0:2  (link-local scope)
 *        FF05:0:0:0:0:0:0:2  (site-local scope)
 *
 * "solicited node" multicast address
 *        FF02:0:0:0:0:1:FFXX:XXXX 
 *  where XX:XXXX consists of the last 24 bits of the node's unicast 
 *  or anycast address
 */


/* 
 * NAME: IS_IPV6_ADDRESS_ALL_NODE_MULTICAST
 *
 * DESCRIPTION: Checks to see if an IPv6 address is an 
 * "all node" multicast address.
 *
 * @Param: IN ipv6Addr - The IPv6 address.
 *
 * return: 1 if true
 *         0 if false
 *
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
ix_uint8
IS_IPV6_ADDRESS_ALL_NODE_MULTICAST(ix_uint128 ipv6Addr)
{
    /* checks if the top 32 bits are FF0X0000 */
    if ((IX_127_96(ipv6Addr) & 0xFFFCFFFF) != 0xFF000000)
    {
        return 0;   /* FALSE */
    }

    if ((IX_95_64(ipv6Addr) != 0x0) || 
        (IX_63_32(ipv6Addr) != 0x0))
    {
        return 0;
    }

    if (IX_31_0(ipv6Addr) != 0x01)
    {
        return 0;
    }

    return 1; /* TRUE */
}

/* 
 * NAME: IS_IPV6_ADDRESS_ALL_ROUTER_MULTICAST
 *
 * DESCRIPTION: Checks to see if an IPv6 address is an 
 * "all routers" multicast address.
 *
 * @Param: IN ipv6Addr - The IPv6 address.
 *
 * return: 1 if true
 *         0 if false
 *
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
ix_uint8
IS_IPV6_ADDRESS_ALL_ROUTER_MULTICAST(ix_uint128 ipv6Addr)
{
    /* checks if the top 32 bits are FF0X0000 */
    if ((IX_127_96(ipv6Addr) & 0xFFF8FFFF) != 0xFF000000)
    {
        return 0;   /* FALSE */
    }

    if ((IX_95_64(ipv6Addr) != 0x0) || 
        (IX_63_32(ipv6Addr) != 0x0))
    {
        return 0;
    }

    if (IX_31_0(ipv6Addr) != 0x02)
    {
        return 0;
    }

    return 1; /* TRUE */
}

/* 
 * NAME: IS_IPV6_ADDRESS_SOLICITED_NODE_MULTICAST
 *
 * DESCRIPTION: Checks to see if an IPv6 address is a 
 * "solicited node" multicast address.
 *
 * @Param: IN ipv6Addr - The IPv6 address.
 * @Param: IN portIpv6Addr - The Ports IPv6 Address
 *
 * return: 1 if true
 *         0 if false
 *
 * NOTE: THIS MACRO EXPECTS THE IPV6 ADDRESS IN HOST BYTE ORDER
 */
ix_uint8
IS_IPV6_ADDRESS_SOLICITED_NODE_MULTICAST(ix_uint128 ipv6Addr, 
                                         ix_uint128 portIpv6Addr)
{
    ix_uint32 last24;

    /* checks if the top 32 bits are FF0X0000 */
    if (IX_127_96(ipv6Addr) != 0xFF020000)
    {
        return 0;   /* FALSE */
    }

    if (IX_95_64(ipv6Addr) != 0x0) 
    {
        return 0;
    }

    if (IX_63_32(ipv6Addr) != 0x1) 
    {
        return 0;
    }

    /* The last 32 bits should look like this - 
     *
     * FFXX:XXXX where XX:XXXX are the last 24 bits of the ports IP address
     */
  
    last24 = IX_31_0(portIpv6Addr) & 0x00FFFFFFFF;

    last24 = 0xFF000000 | last24;

    if (IX_31_0(ipv6Addr) != last24)
    {
        return 0;
    }

    return 1; /* TRUE */
}
