/*******************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 2003 By Intel Corporation.  All rights reserved.  
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *******************************************************************************/

/******************************************************************************
 *
 * File:  v6v4_tunnel_decap.c
 *
 * Abstract:
 *      Decapsulates IPv6 packets that have been encapsulated in IPv4.
 *
 * Exported APIs: 
 *   v6v4_tunnel_decap(pIpv6HdrOut, pIpv6HdrIn, pIpv4HdrIn, 
 *                   ipv6_wr_offset, ipv6_rd_offset, ipv4_rd_offset,
 *                   ipv6_wr_mem_type, ipv6_rd_mem_type, ipv4_rd_mem_type)
 *
 *****************************************************************************/


#ifndef	__V6V4_TUNNEL_DECAP_C__
#define	__V6V4_TUNNEL_DECAP_C__

/* Includes */
#include "v6v4_tunnel_decap_uc.h"
#include "v6v4_tunnel_util.c"


#ifdef V6V4_INGRESS_LIST_TRIE

/******************************************************************************
 * v6v4_tunnel_ingress_validate
 *
 *		Descripton 	:
 *         Trie table lookup for ingress source list.  This is a simple trie
 *         scheme using a 256 entry table and trie block lookups.  Maximum
 *         number of reads is 7 for the trie table blocks, plus zero or more
 *         reads to check the index table.  Note that the lookup process differs
 *         from the forwarder LPM lookup in that the shortest prefix that allows
 *         the tunnel is considered a successful validation.
 *		
 *		Outputs		:
 *          None
 *
 *		Inputs		:
 *			srcAddr :       IP source address.
 *          tunnelIndex :   Index of ending tunnel endpoint that received the
 *                          packet with source address srcAddr.
 * 
 *      Returns:
 *          V6V4_VALIDATE_SUCCESS or V6V4_VALIDATE_FAIL.
 */
INLINE	int v6v4_tunnel_ingress_validate(
    unsigned int srcAddr,
    unsigned int tunnelIndex)
{
    __declspec(sram_read_reg) unsigned int isl_data[V6V4_INGRESS_LIST_INDEX_LW];
    __declspec(sram) unsigned char * pTrieBase;
    __declspec(sram) unsigned char * pIndexBase;

    SIGNAL sram_sig;

    unsigned int offset;
    unsigned int trieNum = 0;
    unsigned int lsbNum = 24;
    unsigned int currentIndex;
    unsigned int tunnelId;
    unsigned int trieIndex;

    pTrieBase = (__declspec(sram) unsigned char *)TUNNEL_INGRESS_LIST_SRAM_BASE;
    pIndexBase = (__declspec(sram) unsigned char *)V4V6_ISL_INDEX_BLOCKS_BASE;

    /* Form the offset into the base from 31:24 for the first read. */
    offset = (srcAddr >> V6V4_ISL_TRIE_BASE_SHF) & ~3;

    while (trieNum < V6V4_MAX_TRIES)
    {
        /* Read the entry */
        sram_read((__declspec(sram_read_reg) void *)isl_data,
            (volatile __declspec(sram) void *)(pTrieBase + offset), 1,
            ctx_swap, &sram_sig);

        if (trieNum == 0)
        {
            pTrieBase = (__declspec(sram) unsigned char *)
                V4V6_ISL_TRIE_BLOCKS_BASE;
        }

        /* Get the index value from the entry.  If it is not zero, then we
         * need to check the associated index block for the tunnel.
         */
        trieIndex = isl_data[0] & 0xffff;
        currentIndex = isl_data[0] >> 16;

        while (currentIndex != 0)
        {
            unsigned int islIndex;

            /* Read the index block. */
            offset = currentIndex << V6V4_ISL_INDEX_SHF;

            sram_read((__declspec(sram_read_reg) void *)isl_data,
                (volatile __declspec(sram) void *)(pIndexBase + offset),
                V6V4_INGRESS_LIST_INDEX_LW, ctx_swap, &sram_sig);

            /* Check the first index location in the block. */
            tunnelId = isl_data[0] & 0xffff;
            if (tunnelId == tunnelIndex)
            {
                return V6V4_VALIDATE_SUCCESS;
            }

            /* Check the remaining index locations. */
            #if (V6V4_INGRESS_LIST_INDEX_LW > 1)
                tunnelId = isl_data[1] >> 16;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
                tunnelId = isl_data[1] & 0xffff;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
            #endif

            /* Check the remaining index locations. */
            #if (V6V4_INGRESS_LIST_INDEX_LW > 2)
                tunnelId = isl_data[2] >> 16;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
                tunnelId = isl_data[2] & 0xffff;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
            #endif

            /* Check the remaining index locations. */
            #if (V6V4_INGRESS_LIST_INDEX_LW > 3)
                tunnelId = isl_data[3] >> 16;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
                tunnelId = isl_data[3] & 0xffff;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
            #endif

            /* Check the remaining index locations. */
            #if (V6V4_INGRESS_LIST_INDEX_LW > 4)
                tunnelId = isl_data[4] >> 16;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
                tunnelId = isl_data[4] & 0xffff;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
            #endif

            /* Check the remaining index locations. */
            #if (V6V4_INGRESS_LIST_INDEX_LW > 5)
                tunnelId = isl_data[5] >> 16;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
                tunnelId = isl_data[5] & 0xffff;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
            #endif

            /* Check the remaining index locations. */
            #if (V6V4_INGRESS_LIST_INDEX_LW > 6)
                tunnelId = isl_data[6] >> 16;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
                tunnelId = isl_data[6] & 0xffff;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
            #endif

            /* Check the remaining index locations. */
            #if (V6V4_INGRESS_LIST_INDEX_LW > 7)
                tunnelId = isl_data[7] >> 16;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
                tunnelId = isl_data[7] & 0xffff;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
            #endif

            /* Check the remaining index locations. */
            #if (V6V4_INGRESS_LIST_INDEX_LW > 8)
                tunnelId = isl_data[8] >> 16;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
                tunnelId = isl_data[8] & 0xffff;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
            #endif

            /* Check the remaining index locations. */
            #if (V6V4_INGRESS_LIST_INDEX_LW > 9)
                tunnelId = isl_data[9] >> 16;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
                tunnelId = isl_data[9] & 0xffff;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
            #endif

            /* Check the remaining index locations. */
            #if (V6V4_INGRESS_LIST_INDEX_LW > 10)
                tunnelId = isl_data[10] >> 16;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
                tunnelId = isl_data[10] & 0xffff;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
            #endif

            /* Check the remaining index locations. */
            #if (V6V4_INGRESS_LIST_INDEX_LW > 11)
                tunnelId = isl_data[11] >> 16;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
                tunnelId = isl_data[11] & 0xffff;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
            #endif

            /* Check the remaining index locations. */
            #if (V6V4_INGRESS_LIST_INDEX_LW > 12)
                tunnelId = isl_data[12] >> 16;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
                tunnelId = isl_data[12] & 0xffff;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
            #endif

            /* Check the remaining index locations. */
            #if (V6V4_INGRESS_LIST_INDEX_LW > 13)
                tunnelId = isl_data[13] >> 16;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
                tunnelId = isl_data[13] & 0xffff;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
            #endif

            /* Check the remaining index locations. */
            #if (V6V4_INGRESS_LIST_INDEX_LW > 14)
                tunnelId = isl_data[14] >> 16;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
                tunnelId = isl_data[14] & 0xffff;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
            #endif

            /* Check the remaining index locations. */
            #if (V6V4_INGRESS_LIST_INDEX_LW > 15)
                tunnelId = isl_data[15] >> 16;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
                tunnelId = isl_data[15] & 0xffff;
                if (tunnelId == tunnelIndex) return V6V4_VALIDATE_SUCCESS;
            #endif


            /* Get next index block */
            currentIndex = isl_data[0] >> 16;
        }

        /* Read the next trie block */

        if (trieNum < (V6V4_MAX_TRIES - 1))
        {
            if (trieIndex == 0)
            {
                /* There are no more tries for this prefix, so the lookup
                 * failed.
                 */
                return V6V4_VALIDATE_FAILURE;
            }

            /* Calculate the offset of the next trie entry to read */

            lsbNum -= 4;

            if (lsbNum != 0)
            {
                offset = (trieIndex << V6V4_ISL_TRIE_SHF) + 
                         (srcAddr >> (lsbNum - 2)) & 0x3c;
            }
            else
            {
                offset = (trieIndex << V6V4_ISL_TRIE_SHF) + 
                         (srcAddr << 2) & 0x3c;
            }
        }
    }

    return V6V4_VALIDATE_FAILURE;

}

#else
/******************************************************************************
 * v6v4_tunnel_ingress_check_addr
 *
 *		Descripton 	:
 *         Check a particular address in the ingress list block.
 *		
 *		Outputs		:
 *          None
 *
 *		Inputs		:
 *          pIslBlock :     Ingress list block.
 *          addrNum :       Address number.
 *			srcAddr :       IP source address.
 * 
 *      Returns:
 *          V6V4_VALIDATE_SUCCESS or V6V4_VALIDATE_FAIL.
 */
INLINE	int v6v4_tunnel_ingress_check_addr(
    __declspec(sram_read_reg) unsigned int *pIslBlock,
    unsigned int addrNum,
    unsigned int srcAddr)
{
    unsigned int addrIndex;
    unsigned int maskIndex;
    unsigned int shiftAmt;
    unsigned int maskLen;
    unsigned int addressValue;

    /* Set addrIndex to the index in the ISL block of the longword that
     * contains the address being compared.  Set maskIndex to the index
     * in the ISL block of the longword that contains the corresponding
     * mask length.
     */
    if (addrNum < 4)
    {
        addrIndex = addrNum + 2;
        maskIndex = V6V4_ISL_PREFIX_1_4;
    }
    else if (addrNum < 8)
    {
        addrIndex = addrNum + 3;
        maskIndex = V6V4_ISL_PREFIX_5_8;
    }
    else
    {
        addrIndex = addrNum + 4;
        maskIndex = V6V4_ISL_PREFIX_9_12;
    }

    shiftAmt = ((3 - (addrNum % 4)) * 8);

    /* If this address entry is valid, compare it to srcAddr */
    if (pIslBlock[V6V4_ISL_FLAGS] & (1 << addrNum))
    {
        maskLen = (pIslBlock[maskIndex] >> shiftAmt) & 0xff;
        addressValue = srcAddr & (0xffffffff << maskLen);
        if (addressValue == pIslBlock[addrIndex])
        {
            return V6V4_VALIDATE_SUCCESS;
        }
    }

    return V6V4_VALIDATE_FAILURE;
}


/******************************************************************************
 * v6v4_tunnel_ingress_validate
 *
 *		Descripton 	:
 *         Array lookup for ingress source list.  This uses a simple list of
 *         blocks, each containing up to 12 prefixes that are valid for the
 *         tunnel.
 *		
 *		Outputs		:
 *          None
 *
 *		Inputs		:
 *			srcAddr :       IP source address.
 *          tunnelEntry :   Tunnel next hop entry containing ingress list
 *                          block index.
 * 
 *      Returns:
 *          V6V4_VALIDATE_SUCCESS or V6V4_VALIDATE_FAIL.
 */
INLINE	int v6v4_tunnel_ingress_validate(
    unsigned int srcAddr,
    unsigned int tunnelEntry)
{
    __declspec(sram_read_reg) unsigned int islBlock[V6V4_INGRESS_LIST_MAX];
    __declspec(sram) unsigned char * pIslBase;

    SIGNAL sram_sig;

    unsigned int index;
    unsigned int addrNum;

    pIslBase = (__declspec(sram) unsigned char *)TUNNEL_INGRESS_LIST_SRAM_BASE;

    index = tunnelEntry & 0xffff;

    while (index != 0)
    {
        sram_read((__declspec(sram_read_reg) void *)islBlock,
            (volatile __declspec(sram) void *)(pIslBase + (index << V6V4_ISL_SHF)),
            V6V4_INGRESS_LIST_MAX, ctx_swap, &sram_sig);

        if (v6v4_tunnel_ingress_check_addr(islBlock, 0, srcAddr) ==
                V6V4_VALIDATE_SUCCESS)
            return V6V4_VALIDATE_SUCCESS;

        #if (V6V4_MAX_ADDRESS > 1)
        if (v6v4_tunnel_ingress_check_addr(islBlock, 1, srcAddr) ==
                V6V4_VALIDATE_SUCCESS)
            return V6V4_VALIDATE_SUCCESS;
        #endif

        #if (V6V4_MAX_ADDRESS > 2)
        if (v6v4_tunnel_ingress_check_addr(islBlock, 2, srcAddr) ==
                V6V4_VALIDATE_SUCCESS)
            return V6V4_VALIDATE_SUCCESS;
        #endif

        #if (V6V4_MAX_ADDRESS > 3)
        if (v6v4_tunnel_ingress_check_addr(islBlock, 3, srcAddr) ==
                V6V4_VALIDATE_SUCCESS)
            return V6V4_VALIDATE_SUCCESS;
        #endif

        #if (V6V4_MAX_ADDRESS > 4)
        if (v6v4_tunnel_ingress_check_addr(islBlock, 4, srcAddr) ==
                V6V4_VALIDATE_SUCCESS)
            return V6V4_VALIDATE_SUCCESS;
        #endif

        #if (V6V4_MAX_ADDRESS > 5)
        if (v6v4_tunnel_ingress_check_addr(islBlock, 5, srcAddr) ==
                V6V4_VALIDATE_SUCCESS)
            return V6V4_VALIDATE_SUCCESS;
        #endif

        #if (V6V4_MAX_ADDRESS > 6)
        if (v6v4_tunnel_ingress_check_addr(islBlock, 6, srcAddr) ==
                V6V4_VALIDATE_SUCCESS)
            return V6V4_VALIDATE_SUCCESS;
        #endif

        #if (V6V4_MAX_ADDRESS > 7)
        if (v6v4_tunnel_ingress_check_addr(islBlock, 7, srcAddr) ==
                V6V4_VALIDATE_SUCCESS)
            return V6V4_VALIDATE_SUCCESS;
        #endif

        #if (V6V4_MAX_ADDRESS > 8)
        if (v6v4_tunnel_ingress_check_addr(islBlock, 8, srcAddr) ==
                V6V4_VALIDATE_SUCCESS)
            return V6V4_VALIDATE_SUCCESS;
        #endif

        #if (V6V4_MAX_ADDRESS > 9)
        if (v6v4_tunnel_ingress_check_addr(islBlock, 9, srcAddr) ==
                V6V4_VALIDATE_SUCCESS)
            return V6V4_VALIDATE_SUCCESS;
        #endif

        #if (V6V4_MAX_ADDRESS > 10)
        if (v6v4_tunnel_ingress_check_addr(islBlock, 10, srcAddr) ==
                V6V4_VALIDATE_SUCCESS)
            return V6V4_VALIDATE_SUCCESS;
        #endif

        #if (V6V4_MAX_ADDRESS > 11)
        if (v6v4_tunnel_ingress_check_addr(islBlock, 11, srcAddr) ==
                V6V4_VALIDATE_SUCCESS)
            return V6V4_VALIDATE_SUCCESS;
        #endif


        /* Go on to the next block if necessary */
#ifdef V6V4_INGRESS_LIST_MULTI_BLOCKS
        index = islBlock[V6V4_ISL_NEXT_BLOCK] >> 16;
#else
        index = 0;
#endif

    }

    return V6V4_VALIDATE_FAILURE;
}
#endif /* V6V4_INGRESS_LIST_TRIE */


/******************************************************************************
 * v6v4_tunnel_decap_set_exception
 *
 *		Descripton 	:
 *        Sets the exception code and increments the exception counter.
 *		
 *		Outputs		:
 *          None
 *
 *		Inputs		:
 *			excpCode   :    Exception code.
 * 
 *      Returns:
 *          None.
 */
INLINE	void v6v4_tunnel_decap_set_exception(
	unsigned int excpCode)
{

#ifdef V6V4_DECAP_COUNTERS
	sram_incr((volatile void __declspec(sram) *)
                (V6V4_DECAP_COUNTER_BASE + V6V4_DECAP_EXCP_COUNT));
#endif

    dl_set_exception(BID_V6V4_TUNNEL_DECAP, excpCode);
    dlNextBlock = IX_EXCEPTION;
    return;
}


/******************************************************************************
 * v6v4_tunnel_decap_drop_packet
 *
 *		Descripton 	:
 *        Sets next block to indicate drop and increments the drop counter.
 *		
 *		Outputs		:
 *          None
 *
 *		Inputs		:
 *          None.
 * 
 *      Returns:
 *          None.
 */
INLINE	void v6v4_tunnel_decap_drop_packet(void)
{

#ifdef V6V4_DECAP_COUNTERS
	sram_incr((volatile void __declspec(sram) *)
                (V6V4_DECAP_COUNTER_BASE + V6V4_DECAP_DROP_COUNT));
#endif

    dlNextBlock = IX_DROP;
    return;
}



/******************************************************************************
 * v6v4_tunnel_decap_auto
 *
 *		Descripton 	:
 *        Validates automatic tunneled packets requiring decapsulation at
 *        this node.
 *		
 *		Outputs		:
 *          None
 *
 *		Inputs		:
 *          pInIpv4     :   Buffer containing IPv4 header.
 *			pInIpv6     :   Buffer containing IPv6 header.
 *          ipv4RdOffset :  Offset of start of IPv4 header in read buffer.
 *          ipv6RdOffset :  Offset of start of IPv6 header in read buffer.
 *          ipv4RdMemType : Memory type of IPv4 buffer.
 *          ipv6RdMemType : Memory type of IPv6 buffer.
 *          tunnelEntry :   Tunnel next hop entry.
 * 
 *      Returns:
 *          None.
 */
INLINE	void v6v4_tunnel_decap_auto(
	void *pInIpv4,
	void *pInIpv6,
	unsigned int ipv4RdOffset,
	unsigned int ipv6RdOffset,
    mem_t ipv4RdMemType,
    mem_t ipv6RdMemType,
    unsigned int tunnelEntry)
{
    unsigned int ipv4DestAddr;
    unsigned int embeddedIpv4Addr;
    unsigned int ipv4SrcAddr;
    int result;

    if ((tunnelEntry & (1 << V6V4_DECAP_FLAGS_AUTO)) == 0)
    {
        v6v4_tunnel_decap_drop_packet();
        return;
    }

    embeddedIpv4Addr = _buf_byte_extract(
	    (UINT*)pInIpv6,
        ipv6RdOffset + IPV6_DST_ADDR_START + IPV6_V4ADDR_OFFSET_AUTO,
        IPV6_EMBEDDED_IPV4_LENGTH, ipv6RdMemType);

    ipv4DestAddr = _buf_byte_extract(
	    (UINT*)pInIpv4, ipv4RdOffset + IPV4_DESTINATION_ADDRESS_START_BYTE,
        IPV4_DESTINATION_ADDRESS_LEN, ipv4RdMemType);

    if (embeddedIpv4Addr != ipv4DestAddr)
    {
        v6v4_tunnel_decap_drop_packet();
        return;
    }

#ifdef V6V4_DECAP_VERIFY_IPV6_SOURCE
    result = v6v4_tunnel_ipv6_source_verify(
                pInIpv6, ipv6RdOffset, ipv6RdMemType);
    if (result == V6V4_VALIDATE_FAILURE)
    {
        v6v4_tunnel_decap_drop_packet();
        return;
    }
#endif

#ifdef V6V4_DECAP_VERIFY_IPV4_SOURCE
    ipv4SrcAddr = _buf_byte_extract(
	    (UINT*)pInIpv4, ipv4RdOffset + IPV4_SOURCE_ADDRESS_START_BYTE,
        IPV4_SOURCE_ADDRESS_LEN, ipv4RdMemType);

    result = v6v4_tunnel_ipv4_addr_verify(ipv4SrcAddr, 0, V6V4_DO_VERIFY);
    if (result == V6V4_VALIDATE_FAILURE)
    {
        v6v4_tunnel_decap_drop_packet();
        return;
    }
#endif

    v6v4_tunnel_decap_set_exception(V6V4_DECAP_EXCEPTION_AUTO);
    return;
}


/******************************************************************************
 * v6v4_tunnel_decap
 *
 *		Descripton 	:
 *        Handles IPv6 packets that have been tunneled over IPv4 and
 *        require decapsulation.  The IPv4 forwarder should have previously
 *        handled the packet and set the next hop ID in the meta data.
 *		
 *		Outputs		:
 *          None
 *
 *		Inputs		:
 *			pOutIpv6    :   Buffer used to update IPv6 header (only used
 *                          if TOS option is selected).
 *			pInIpv6     :   Buffer into which to load/read IPv6 header.
 *          pInIpv4     :   Buffer from which to read IPv4 header.
 *          ipv6WrOffset :  Offset of start of IPv6 header in write buffer.
 *          ipv6RdOffset :  Offset of start of IPv6 header in read buffer.
 *          ipv4RdOffset :  Offset of start of IPv4 header in read buffer.
 *          ipv6WrMemType : Memory type of IPv6 output buffer.
 *          ipv6RdMemType : Memory type of IPv6 input buffer.
 *          ipv4RdMemType : Memory type of IPv4 input buffer.
 * 
 *      Returns:
 *          None.
 */
INLINE	void v6v4_tunnel_decap(
    void *pOutIpv6,
	void *pInIpv6,
	void *pInIpv4,
    unsigned int ipv6WrOffset,
	unsigned int ipv6RdOffset,
	unsigned int ipv4RdOffset,
    mem_t ipv6WrMemType,
    mem_t ipv6RdMemType,
    mem_t ipv4RdMemType)
{
    unsigned int ipv4Proto;
    unsigned int fragInfo;
    unsigned int tunnelEntry;
    unsigned int result;
    SIGNAL sram_sig;

    __declspec(sram_read_reg) unsigned int tunnelNhInfo;
    __declspec (sram) unsigned int *pTunnelNextHopBase = 
        (__declspec (sram) unsigned int *)TUNNEL_DECAP_NEXTHOP_SRAM_BASE;

    if (dlNextBlock != BID_V6V4_TUNNEL_DECAP)
    {
        return;
    }

    if (dlMeta.nhidType != NEXTHOP_TYPE_TUNNEL)
    {
        dlNextBlock = V6V4_NEXT_NO_DECAP;
        return;
    }

    ipv4Proto = _buf_byte_extract(
	    (UINT*)pInIpv4, ipv4RdOffset + IPV4_PROTOCOL_START_BYTE,
        IPV4_PROTOCOL_LEN, ipv4RdMemType);

    /* Check IPv4 protocol for tunneling */

    if (ipv4Proto != V6V4_PROTOCOL)
    {
        v6v4_tunnel_decap_set_exception(V6V4_DECAP_EXCEPTION_LOCAL);
        return;
    }

    /* Check for a fragmented packet */

    fragInfo = _buf_byte_extract(
	    (UINT*)pInIpv4, ipv4RdOffset + IPV4_FLAGS_FRAGMENT_START_BYTE,
        IPV4_FLAGS_FRAGMENT_LEN, ipv4RdMemType);

    if ((fragInfo << V6V4_FRAGMENT_SHF) != 0)
    {
        v6v4_tunnel_decap_set_exception(V6V4_DECAP_EXCEPTION_FRAGMENT);
        return;
    }

    /* Read the tunnel next hop entry and check the validity flag */

    sram_read(&tunnelNhInfo, (volatile void __declspec(sram) *)
        (&pTunnelNextHopBase[dlMeta.nextHopId]),
        V6V4_DECAP_TNH_LW_COUNT, ctx_swap, &sram_sig);

    tunnelEntry = tunnelNhInfo;

    if ((tunnelEntry & (1 << V6V4_DECAP_FLAGS_VALID)) == 0)
    {
        v6v4_tunnel_decap_drop_packet();
        return;
    }

    /* Set the header type to indicate a decapsulated ipv6 packet */

    dlMeta.headerType = V6V4_HDR_TYPE_DECAP;

    /* Load the IPv6 header */

    dl_load_ipv6_hdr(pInIpv6, ipv6RdOffset, ipv6RdMemType, V6V4_IPV6_HDR_SIZE);

    /* Check for IPv4-compatible address, and if found, perform automatic
     * tunnel validations
     */

    if (v6v4_ipv4_compatible_destination(pInIpv6, ipv6RdOffset, ipv6RdMemType))
    {
        v6v4_tunnel_decap_auto(
            pInIpv4, pInIpv6, ipv4RdOffset, ipv6RdOffset,
            ipv4RdMemType, ipv6RdMemType, tunnelEntry);

        return;
    }

    /* Not an automatic tunnel.  Validate the IPv6 source address */

#ifdef V6V4_DECAP_VERIFY_IPV6_SOURCE
    result = v6v4_tunnel_ipv6_source_verify(
                pInIpv6, ipv6RdOffset, ipv6RdMemType);
    if (result == V6V4_VALIDATE_FAILURE)
    {
        v6v4_tunnel_decap_drop_packet();
        return;
    }
#endif

    /* If necessary, check for and validate 6to4 source and destination
     * addresses.
     */

    if ((tunnelEntry & (1 << V6V4_DECAP_FLAGS_6TO4)) != 0)
    {
        result = v6v4_tunnel_6to4_verify(
            pInIpv6, ipv6RdOffset + IPV6_SRC_ADDR_START, ipv6RdMemType, 0);
        if (result == V6V4_VALIDATE_FAILURE)
        {
            v6v4_tunnel_decap_drop_packet();
            return;
        }

        result = v6v4_tunnel_6to4_verify(
            pInIpv6, ipv6RdOffset + IPV6_DST_ADDR_START, ipv6RdMemType, 0);
        if (result == V6V4_VALIDATE_FAILURE)
        {
            v6v4_tunnel_decap_drop_packet();
            return;
        }
    }

	/* Validate the source address using the ingress source list if required */

    if ((tunnelEntry & (1 << V6V4_DECAP_FLAGS_SRC_VALIDATE)) != 0)
    {
        unsigned int ipv4SrcAddr;

        ipv4SrcAddr = _buf_byte_extract(
    	    (UINT*)pInIpv4, ipv4RdOffset + IPV4_SOURCE_ADDRESS_START_BYTE,
            IPV4_SOURCE_ADDRESS_LEN, ipv4RdMemType);

        #ifdef V6V4_INGRESS_LIST_TRIE
		    result = v6v4_tunnel_ingress_validate(
                        ipv4SrcAddr, dlMeta.nextHopId);
        #else
		    result = v6v4_tunnel_ingress_validate(ipv4SrcAddr, tunnelEntry);
        #endif

        if (result == V6V4_VALIDATE_FAILURE)
        {
            v6v4_tunnel_decap_drop_packet();
            return;
        }
    }
#ifdef V6V4_DECAP_VERIFY_IPV4_SOURCE
    else
    {
        /* Ingress list not used for this tunnel, so just do the basic
         * RFC 2893 validations.
         */
        unsigned int ipv4SrcAddr;
        int result;

        ipv4SrcAddr = _buf_byte_extract(
    	    (UINT*)pInIpv4, ipv4RdOffset + IPV4_SOURCE_ADDRESS_START_BYTE,
            IPV4_SOURCE_ADDRESS_LEN, ipv4RdMemType);

        result = v6v4_tunnel_ipv4_addr_verify(ipv4SrcAddr, 0, V6V4_DO_VERIFY);
        if (result == V6V4_VALIDATE_FAILURE)
        {
            v6v4_tunnel_decap_drop_packet();
            return;
        }
    }
#endif

    /* Copy TOS byte to IPv6 traffic class if requested.  Note this is not the
	 * behavior defined by RFC 2893, but can be used if future RFCs specify it.
     */

    if ((tunnelEntry & (1 << V6V4_DECAP_FLAGS_TOS_V6)) != 0)
    {
        unsigned int trClass;
        unsigned int tos;

        trClass = _buf_byte_extract(
    	    (UINT*)pInIpv6,
            ipv6RdOffset + IPV6_VERSION_TRAFFIC_CLASS_FLOW_LABEL_START,
            IPV6_VERSION_TRAFFIC_CLASS_FLOW_LABEL_COUNT, ipv6RdMemType);

        tos = _buf_byte_extract(
    	    (UINT*)pInIpv4, ipv6RdOffset + IPV4_TYPE_OF_SERVICE_START_BYTE,
            IPV4_TYPE_OF_SERVICE_LEN, ipv4RdMemType);

        trClass  = (trClass & 0xf00fffff) | (tos << 16);

        v6v4_insert_lword(pOutIpv6,
                  ipv6WrOffset + IPV6_VERSION_TRAFFIC_CLASS_FLOW_LABEL_START,
                  trClass, 0, ipv6WrMemType);
    }

    /* Decapsulate the packet */

    dlMeta.bufferSize -= V6V4_IPV4_HDR_SIZE;
    dlMeta.packetSize -= V6V4_IPV4_HDR_SIZE;

#ifdef V6V4_UPDATE_OFFSET
    dlMeta.offset += V6V4_IPV4_HDR_SIZE;
#endif

    dlNextBlock = V6V4_NEXT_IPV6;
    return;
}

#endif	/* __V6V4_TUNNEL_DECAP_C__ */