/*******************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 2003 By Intel Corporation.  All rights reserved.  
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *******************************************************************************/

/******************************************************************************
 *
 * File:  v6v4_tunnel_encap.c
 *
 * Abstract:
 *      Encapsulates IPv6 packets that need to be routed over IPv4 to reach
 *      the IPv6 next hop.
 *
 * Exported APIs: 
 *   v6v4_tunnel_encap(pIpv4HdrOut, pIpv6HdrIn, ipv4_wr_offset, ipv6_rd_offset,
 *                     ipv4_rd_mem_type, ipv6_rd_mem_type)
 *
 *****************************************************************************/


#ifndef	__V6V4_TUNNEL_ENCAP_C__
#define	__V6V4_TUNNEL_ENCAP_C__

/* Includes */
#include "v6v4_tunnel_encap_uc.h"
#include "ipv6.h"
#include "v6v4_tunnel_util.c"

unsigned int v6v4_id_base;
unsigned int v6v4_id_seq;

/* Macros for checking tunnel encap flags */


#define ENCAP_FLAG_SET(__tnh, __f) \
    ((((__declspec(sram_read_reg) unsigned int *)&(__tnh))[V6V4_ENCAP_TNH_FLAGS_LW] & (1 << (__f))) != 0)

#define ENCAP_FLAG_CLEAR(__tnh, __f) \
    ((((__declspec(sram_read_reg) unsigned int *)&(__tnh))[V6V4_ENCAP_TNH_FLAGS_LW] & (1 << (__f))) == 0)



/******************************************************************************
 * v6v4_tunnel_encap_init
 *
 *		Descripton 	:
 *        Initializes the global data for the encapsulation block.
 *		
 *		Outputs		:
 *          None
 *
 *		Inputs		:
 *          None.
 * 
 *      Returns:
 *          None.
 */
INLINE	void v6v4_tunnel_encap_init(void)
{
    unsigned int cxtStatus;

    cxtStatus = local_csr_read(local_csr_active_ctx_sts);
    v6v4_id_base = cxtStatus & 0xff;
    v6v4_id_seq = 0;
    
    return;
}


/******************************************************************************
 * v6v4_tunnel_encap_set_exception
 *
 *		Descripton 	:
 *        Sets the exception code and increments the exception counter.
 *		
 *		Outputs		:
 *          None
 *
 *		Inputs		:
 *			excpCode   :    Exception code.
 * 
 *      Returns:
 *          None.
 */
INLINE	void v6v4_tunnel_encap_set_exception(
	unsigned int excpCode)
{

#ifdef V6V4_ENCAP_COUNTERS
	sram_incr((volatile void __declspec(sram) *)
                (V6V4_ENCAP_COUNTER_BASE + V6V4_ENCAP_EXCP_COUNT));
#endif

    dl_set_exception(BID_V6V4_TUNNEL_ENCAP, excpCode);
    dlNextBlock = IX_EXCEPTION;
    return;
}


/******************************************************************************
 * v6v4_tunnel_encap_drop_packet
 *
 *		Descripton 	:
 *        Sets next block to indicate drop and increments the drop counter.
 *		
 *		Outputs		:
 *          None
 *
 *		Inputs		:
 *          None.
 * 
 *      Returns:
 *          None.
 */
INLINE	void v6v4_tunnel_encap_drop_packet(void)
{

#ifdef V6V4_ENCAP_COUNTERS
	sram_incr((volatile void __declspec(sram) *)
                (V6V4_ENCAP_COUNTER_BASE + V6V4_ENCAP_DROP_COUNT));
#endif

    dlNextBlock = IX_DROP;
    return;
}


/******************************************************************************
 * v6v4_tunnel_encap
 *
 *		Descripton 	:
 *        Handles IPv6 packets that require encapsulation in IPv4 in order to
 *        be routed to the IPv6 next hop.  The IPv6 forwarder should have
 *        previously handled the packet and set the next hop ID in the meta
 *        data.
 *		
 *		Outputs		:
 *          None
 *
 *		Inputs		:
 *			pOutIpv4    :   Buffer used to create the IPv4 header.
 *			pInIpv6     :   Buffer into which to load/read IPv6 header.
 *          ipv4WrOffset :  Offset of start of IPv4 header in write buffer.
 *          ipv6RdOffset :  Offset of start of IPv6 header in read buffer.
 *          ipv4WrMemType : Memory type of IPv4 header buffer.
 *          ipv6RdMemType : Memory type of IPv6 header buffer.
 * 
 *      Returns:
 *          None.
 */
INLINE	void v6v4_tunnel_encap(
    void *pOutIpv4,
	void *pInIpv6,
    unsigned int ipv4WrOffset,
	unsigned int ipv6RdOffset,
    mem_t ipv4WrMemType,
    mem_t ipv6RdMemType)
{
    SIGNAL sram_sig;
    unsigned int pktLength;
    unsigned int checksum;
    unsigned int hdrValue;
    unsigned int hdrValue1;
    unsigned int remoteAddr;

    __declspec(sram_read_reg) tunnel_encap_nh_t tunnelNhInfo;

    __declspec (sram) tunnel_encap_nh_t *pTunnelNextHopBase = 
        (__declspec (sram) tunnel_encap_nh_t *)TUNNEL_ENCAP_NEXTHOP_SRAM_BASE;

    if (dlNextBlock != BID_V6V4_TUNNEL_ENCAP)
    {
        return;
    }

    if (dlMeta.nhidType != NEXTHOP_TYPE_TUNNEL)
    {
        dlNextBlock = V6V4_NEXT_NO_ENCAP;
        return;
    }

    /* Read the tunnel next hop entry and check the validity flag */

    sram_read((__declspec(sram_read_reg) void *)&tunnelNhInfo,
              (volatile void __declspec(sram) *)
                (&pTunnelNextHopBase[dlMeta.nextHopId]),
              V6V4_ENCAP_TNH_LW_COUNT, ctx_swap, &sram_sig);

    if (ENCAP_FLAG_CLEAR(tunnelNhInfo, V6V4_ENCAP_FLAGS_VALID))
    {
        v6v4_tunnel_encap_drop_packet();
        return;
    }

    /* Check MTU, and send to XScale if the encapsulated packet is too large. */

    pktLength = _buf_byte_extract(
	    (UINT*)pInIpv6, ipv6RdOffset + IPV6_PAYLOAD_LENGTH_START,
        IPV6_PAYLOAD_LENGTH_COUNT, ipv6RdMemType);

    pktLength += V6V4_ENCAP_HDR_SIZE;

    if (tunnelNhInfo.pathMtu < pktLength)
    {
        if (pktLength > V6V4_MIN_IPV6_MTU)
        {
            v6v4_tunnel_encap_set_exception(V6V4_ENCAP_EXCEPTION_MTU);
            return;
        }
    }

    /* Verify 6to4 source and destination if necessary. */

    if (ENCAP_FLAG_SET(tunnelNhInfo, V6V4_ENCAP_FLAGS_6TO4))
    {
        unsigned int result;

        result = v6v4_tunnel_6to4_verify(
            pInIpv6, ipv6RdOffset + IPV6_SRC_ADDR_START, ipv6RdMemType, 0);
        if (result == V6V4_VALIDATE_FAILURE)
        {
            v6v4_tunnel_decap_drop_packet();
            return;
        }

        result = v6v4_tunnel_6to4_verify(
            pInIpv6, ipv6RdOffset + IPV6_DST_ADDR_START, ipv6RdMemType, 0);
        if (result == V6V4_VALIDATE_FAILURE)
        {
            v6v4_tunnel_decap_drop_packet();
            return;
        }
    }

    /* Construct the IPv4 header */

	/* LW 0 contains the version, header length, TOS and total packet
	 * length.

     * If the TOS_V6 config flag is set, then special handling
	 * is done for the TOS byte.  Note this is not the behavior defined
	 * by RFC 2893, but can be used if future RFCs specify it.
     */

    hdrValue = V6V4_VERS_IHL_TOS;

    if (ENCAP_FLAG_SET(tunnelNhInfo, V6V4_ENCAP_FLAGS_TOS_V6))
    {
        unsigned int trClass;

        trClass = _buf_byte_extract(
    	    (UINT*)pInIpv6,
            ipv6RdOffset + IPV6_VERSION_TRAFFIC_CLASS_FLOW_LABEL_START,
            2, ipv6RdMemType);

        hdrValue |= ((trClass >> 4) & 0xff);
    }
    else
    {
        hdrValue |= tunnelNhInfo.tos;
    }

    checksum = pktLength + hdrValue;
    hdrValue = (hdrValue << 16) | pktLength;

    v6v4_insert_lword(pOutIpv4, ipv4WrOffset + 0, hdrValue, 0, ipv4WrMemType);

	/* LW 1 contains the ID, flags and fragment offset.  We must determine
	 * whether or not to set the DONT_FRAGMENT flag.  If the tunnel has
	 * the NO_DF flag set or if the tunnel MTU is less than or equal to the
	 * minimum MTU we don't set the flag.
     */

    hdrValue1 = (v6v4_id_base << 8) | (v6v4_id_seq & 0xff);
    v6v4_id_seq += 2;

    checksum += hdrValue1;
    hdrValue1 <<= 16;

    if (ENCAP_FLAG_CLEAR(tunnelNhInfo, V6V4_ENCAP_FLAGS_NO_DF))
    {
        if (tunnelNhInfo.pathMtu > V6V4_MIN_IPV6_MTU)
        {
            hdrValue1 |= IPV4_DONT_FRAGMENT;
            checksum += (unsigned short)hdrValue1;
        }
    }

    v6v4_insert_lword(pOutIpv4, ipv4WrOffset + 4,
                      hdrValue1, hdrValue, ipv4WrMemType);

    /* Finish up the checksum before writing the rest of the header
     * (we could write LW 3 & 4 but if the buffer is not LW aligned
     * it is easier to write them in order).
     */

	/* LW 3 cotains the source address. */

    checksum += (tunnelNhInfo.localAddress >> 16);
    checksum += (unsigned short)tunnelNhInfo.localAddress;

    /* LW 4 contains the destination address, which can either come from
     * the tunnel next hop or from the packet IPv6 destination address.
     */

    if (ENCAP_FLAG_CLEAR(tunnelNhInfo, V6V4_ENCAP_FLAGS_AUTO))
    {
        /* Configured tunnel - get the remote address from the next hop data */
        remoteAddr = tunnelNhInfo.remoteAddress;
    }
    else if (ENCAP_FLAG_SET(tunnelNhInfo, V6V4_ENCAP_FLAGS_6TO4))
    {
        /* 6to4 automatic tunnel - extract the remote address from the IPv6
         * destination.
         */
        remoteAddr = _buf_byte_extract((UINT*)pInIpv6, 
            ipv6RdOffset + IPV6_DST_ADDR_START +IPV6_V4ADDR_OFFSET_6TO4,
            IPV6_EMBEDDED_IPV4_LENGTH, ipv6RdMemType);
    }
    else
    {
        /* Automatic tunnel - extract the remote address from the IPv6
         * destination */

        int verifyAll;
        unsigned int result;

        remoteAddr = _buf_byte_extract((UINT*)pInIpv6,
            ipv6RdOffset + IPV6_DST_ADDR_START +IPV6_V4ADDR_OFFSET_AUTO,
            IPV6_EMBEDDED_IPV4_LENGTH, ipv6RdMemType);

        /* Validate the extracted address */

        #ifdef V6V4_ENCAP_VERIFY_DEST_AUTO
            verifyAll = V6V4_DO_VERIFY;
        #else
            verifyAll = V6V4_DONT_VERIFY;
        #endif

        result = v6v4_tunnel_ipv4_addr_verify(
                    remoteAddr, tunnelNhInfo.snBroadcast, verifyAll);
        if (result == V6V4_VALIDATE_FAILURE)
        {
            v6v4_tunnel_decap_drop_packet();
            return;
        }
    }

    checksum += (remoteAddr >> 16);
    checksum += (unsigned short)remoteAddr;

    /* Lastly, set LW 2, which contains the TTL, protocol and checksum. */

    hdrValue = (tunnelNhInfo.ttl << 8) | V6V4_PROTOCOL;
    checksum += (unsigned short)hdrValue;

    /* Finish the checksum calculation */

    while ((checksum >> 16) != 0)
    {
        checksum = (checksum & 0xffff) + (checksum >> 16);
    }

    checksum = ~checksum;
    hdrValue = (hdrValue << 16) | (unsigned short)checksum;

    /* Write the rest of the header */

    v6v4_insert_lword(pOutIpv4, ipv4WrOffset + 8,
                      hdrValue, hdrValue1, ipv4WrMemType);

    v6v4_insert_lword(pOutIpv4, ipv4WrOffset + 12,
                      tunnelNhInfo.localAddress, hdrValue, ipv4WrMemType);

    v6v4_insert_lword(pOutIpv4, ipv4WrOffset + 16,
                      remoteAddr, tunnelNhInfo.localAddress,
                      ipv4WrMemType);

    if ((ipv4WrOffset & 3) != 0)
    {
        /* This code is needed only if the IPv4 header is not 32-bit
         * aligned.  In this case we still need to get the last part
         * of the last LW of the IPv4 header into the buffer, so we\
         * get the first LW of the IPv6 header to merge with.
         */
        hdrValue = _buf_byte_extract(
            (UINT*)pInIpv6, ipv6RdOffset, 4, ipv6RdMemType);

        v6v4_insert_lword(pOutIpv4, ipv4WrOffset + 20,
                          hdrValue, remoteAddr, ipv4WrMemType);
    }

    dlMeta.bufferSize += V6V4_IPV4_HDR_SIZE;
    dlMeta.packetSize += V6V4_IPV4_HDR_SIZE;
    dlMeta.headerType |= V6V4_HDR_TYPE_ENCAP;

#ifdef V6V4_UPDATE_OFFSET
    dlMeta.offset -= V6V4_IPV4_HDR_SIZE;
#endif

    dlMeta.nextHopId = 0xffffffff;

    dlNextBlock = V6V4_NEXT_IPV4;
    return;
}


#endif	/* __V6V4_TUNNEL_ENCAP_C__ */