/*******************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 2003 By Intel Corporation.  All rights reserved.  
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *******************************************************************************/

/******************************************************************************
 *
 * File:  v6v4_tunnel_util.c
 *
 * Abstract:
 *      Contains utility routines used by the IPv6/v4 tunneling code.
 *
 *****************************************************************************/


#ifndef	__V6V4_TUNNEL_UTIL_C__
#define	__V6V4_TUNNEL_UTIL_C__

#include "v6v4_tunnel.h"
#include "v6v4_tunnel_microc.h"
#include "ipv6.h"

/* Defines needed for validation */

#define V6V4_IPV4_MCAST_MASK 0xE
#define V6V4_IPV4_MCAST_SHF  28

#define V6V4_LOOPBACK_IPV4   0x7f000001
#define V6V4_BROADCAST_IPV4  0xffffffff

#define V6V4_IPV6_MCAST_MASK 0xFF
#define V6V4_IPV6_MCAST_SHF  24

/* Prefix values for private IPv4 addresses */

#define V6V4_IPV4_PRIVATE_24 10           // 10/8 prefix
#define V6V4_IPV4_PRIVATE_20 0xac1        // 172.16/12 prefix
#define V6V4_IPV4_PRIVATE_16 0xc0a8       // 192.168/16 prefix

/* Return values for boolean routines */

#define V6V4_TRUE	1
#define V6V4_FALSE  0


/******************************************************************************
 * v6v4_tunnel_ipv4_addr_verify
 *
 *		Descripton 	:
 *        Verify an IPv4 source or destination address according to RFC 2893.
 *		
 *		Outputs		:
 *          None
 *
 *		Inputs		:
 *          ipAddr :            Address to be verified.
 *          subnetBroadcast:    Subnet broadcast address to used for additional
 *                              verification.
 *          verifyAll :         Flag indicating whether all checks are to be
 *                              performed, or only those checks not typically
 *                              performed by the forwarder for destination
 *                              addresses.
 * 
 *      Returns:
 *          V6V4_VALIDATE_SUCCESS or V6V4_VALIDATE_FAILURE.
 */
INLINE	int v6v4_tunnel_ipv4_addr_verify(
    unsigned int ipAddr,
    unsigned int subnetBroadcast,
    int verifyAll)
{
    if (verifyAll)
    {
        if (ipAddr == 0)
        {
            return V6V4_VALIDATE_FAILURE;
        }

        if (ipAddr == V6V4_LOOPBACK_IPV4)
        {
            return V6V4_VALIDATE_FAILURE;
        }
    }

    if (ipAddr == V6V4_BROADCAST_IPV4)
    {
        return V6V4_VALIDATE_FAILURE;
    }

    if ((ipAddr >> V6V4_IPV4_MCAST_SHF) == V6V4_IPV4_MCAST_MASK)
    {
        return V6V4_VALIDATE_FAILURE;
    }

    if (subnetBroadcast != 0)
    {
        if (subnetBroadcast == ipAddr)
        {
            return V6V4_VALIDATE_FAILURE;
        }
    }

    return V6V4_VALIDATE_SUCCESS;
}


/******************************************************************************
 * v6v4_tunnel_ipv6_source_verify
 *
 *		Descripton 	:
 *        Verify the tunnel source IPv6 address according to RFC 2893.
 *		
 *		Outputs		:
 *          None
 *
 *		Inputs		:
 *          pIpv6Hdr :          Buffer containing IPv6 header.
 *          hdrOffset :         Offset of header in buffer
 *          memType :           Memory type of buffer
 * 
 *      Returns:
 *          V6V4_VALIDATE_SUCCESS or V6V4_VALIDATE_FAILURE.
 */
INLINE	int v6v4_tunnel_ipv6_source_verify(
    void* pIpv6Hdr,
    unsigned int hdrOffset,
    mem_t memType)
{
    unsigned int value;

	/* Extract the first longword of the source address and check the first
	 * byte for multicast.
	 */

    value = _buf_byte_extract(
	    (UINT*)pIpv6Hdr, hdrOffset + IPV6_SRC_ADDR_START + 0, 4,
        memType);

    if ((value >> V6V4_IPV4_MCAST_SHF) == V6V4_IPV6_MCAST_MASK)
    {
        return V6V4_VALIDATE_FAILURE;
    }

	/* The remaining checks are for a loopback (::1), unspecified (::0)
	 * address, or IPv4-compatible address.  For all of these, the first
	 * 3 longwords are zero.
	 */

    if (value != 0)
    {
        return V6V4_VALIDATE_SUCCESS;
    }

    value = _buf_byte_extract(
	    (UINT*)pIpv6Hdr, hdrOffset + IPV6_SRC_ADDR_START + 41, 4,
        memType);

    if (value != 0)
    {
        return V6V4_VALIDATE_SUCCESS;
    }

    value = _buf_byte_extract(
	    (UINT*)pIpv6Hdr, hdrOffset + IPV6_SRC_ADDR_START + 8, 4,
        memType);

    if (value != 0)
    {
        return V6V4_VALIDATE_SUCCESS;
    }

    /* Get 4th longword of address */

    value = _buf_byte_extract(
	    (UINT*)pIpv6Hdr, hdrOffset + IPV6_SRC_ADDR_START + 12, 4,
        memType);

	/* Check for loopback or unspecified address (:: 0 or ::1) */

    if ((value & ~0x00000001) == 0)
    {
        return V6V4_VALIDATE_FAILURE;
    }

    /* If we get to here, we have an IPv4-compatible address, and the
	 * embedded address is in sa_value.  Check it for acceptability.
     */

    return v6v4_tunnel_ipv4_addr_verify(value, 0, V6V4_DO_VERIFY);
}

/******************************************************************************
 * v6v4_ipv4_compatible_destination
 *
 *		Descripton 	:
 *        Determine if the destination address in the specified IPv6 header
 *        is IPv4-compatible.
 *		
 *		Outputs		:
 *          None
 *
 *		Inputs		:
 *          pIpv6Hdr :          Buffer containing IPv6 header.
 *          hdrOffset :         Offset of header in buffer
 *          memType :           Memory type of buffer
 * 
 *      Returns:
 *          V6V4_TRUE if the address is IPv4-compatible, V6V4_FALSE otherwise.
 */
INLINE	int v6v4_ipv4_compatible_destination(
    void* pIpv6Hdr,
    unsigned int hdrOffset,
    mem_t memType)
{
    unsigned int value;

    value = _buf_byte_extract(
	    (UINT*)pIpv6Hdr, hdrOffset + IPV6_DST_ADDR_START + 0, 4,
        memType);

    if (value != V6V4_COMPATIBLE_PREFIX_0_1)
    {
        return V6V4_FALSE;
    }

    value = _buf_byte_extract(
	    (UINT*)pIpv6Hdr, hdrOffset + IPV6_DST_ADDR_START + 4, 4,
        memType);

    if (value != V6V4_COMPATIBLE_PREFIX_2_3)
    {
        return V6V4_FALSE;
    }

    value = _buf_byte_extract(
	    (UINT*)pIpv6Hdr, hdrOffset + IPV6_DST_ADDR_START + 8, 4,
        memType);

    if (value != V6V4_COMPATIBLE_PREFIX_4_5)
    {
        return V6V4_FALSE;
    }

    return V6V4_TRUE;
}

/******************************************************************************
 * v6v4_tunnel_6to4_verify
 *
 *		Descripton 	:
 *        Determine if the specified IPv6 address is a 6to4 address, and if so
 *        validate it according to RFC 3056.
 *		
 *		Outputs		:
 *          None
 *
 *		Inputs		:
 *          pIpv6Hdr :          Buffer containing IPv6 header.
 *          addrOffset :        Offset of address in buffer.
 *          memType :           Memory type of buffer.
 *          subnetBroadcast :   Subnet broadcast to use for validating
 *                              embedded IPv4 address.
 * 
 *      Returns:
 *          V6V4_VALIDATE_SUCCESS or V6V4_VALIDATE_FAILURE.
 */
INLINE	int v6v4_tunnel_6to4_verify(
    void* pIpv6Hdr,
    unsigned int addrOffset,
    mem_t memType,
    unsigned int subnetBroadcast)
{
    unsigned int prefix;
    unsigned int embeddedAddr;

    /* Get first 2 bytes of address and check for 6to4 prefix */

    prefix = _buf_byte_extract(
	    (UINT*)pIpv6Hdr, addrOffset, V6V4_6TO4_PREFIX_LENGTH_BYTES,
        memType);

    if (prefix != V6V4_6TO4_PREFIX)
    {
        return V6V4_VALIDATE_SUCCESS;
    }

    /* Get the embedded IPv4 address and validate it */

    embeddedAddr = _buf_byte_extract(
	    (UINT*)pIpv6Hdr, addrOffset + V6V4_6TO4_PREFIX_LENGTH_BYTES, 
        IPV6_EMBEDDED_IPV4_LENGTH, memType);

    if (v6v4_tunnel_ipv4_addr_verify(embeddedAddr, subnetBroadcast,
            V6V4_DO_VERIFY) == V6V4_VALIDATE_FAILURE)
    {
        return V6V4_VALIDATE_FAILURE;
    }

    /* Check for "24-bit block" private addresses (10/8 prefix) */

    if ((embeddedAddr >> 24) == V6V4_IPV4_PRIVATE_24)
    {
        return V6V4_VALIDATE_FAILURE;
    }

    /* Check for "20-bit block" private addresses (172.16/12 prefix) */

    if ((embeddedAddr >> 20) == V6V4_IPV4_PRIVATE_20)
    {
        return V6V4_VALIDATE_FAILURE;
    }

    /* Check for "16-bit block" private addresses (192.168/16 prefix) */

    if ((embeddedAddr >> 16) == V6V4_IPV4_PRIVATE_16)
    {
        return V6V4_VALIDATE_FAILURE;
    }

    return V6V4_VALIDATE_SUCCESS;
}

/******************************************************************************
 * v6v4_insert_lword
 *
 *		Descripton 	:
 *        Insert a portion of a longword into a buffer.  This routine updates a
 *        32-bit location in a buffer using the 2 values "value" and
 *        "prevValue".  The parameter "startByte" indicates the byte offset in
 *        the buffer at which "value" should start.  The word that contains the
 *        first byte of "value" is written using the appropriate bytes from
 *        "prevValue" and "value".
 *		
 *		Outputs		:
 *          None
 *
 *		Inputs		:
 *          pOutBuf :           Buffer into which to insert bytes.
 *          startByte :         Byte offset at which to insert the data.
 *          value :             Value to be inserted.
 *          prevValue :         Buffer value preceding this one.
 *          memType :           Memory type for pOutBuf.
 * 
 *      Returns:
 *          None
 */
INLINE	void v6v4_insert_lword(
    void* pOutBuf,
    unsigned int startByte,
    unsigned int value,
    unsigned int prevValue,
	mem_t memType)
{
    unsigned int startWord = startByte >> 2;
    unsigned int byteOffset = startByte & 0x3;
    unsigned int * pData = (unsigned int *)pOutBuf;

    if (byteOffset == 0)
    {
        _PUT_MEM_CLASS_DATA(pData + startWord, value, unsigned int, memType);
    }
    else if (byteOffset == 1)
    {
        _PUT_MEM_CLASS_DATA(pData + startWord,
                    (value >> 8) | (prevValue << 24), unsigned int, memType);
    }
    else if (byteOffset == 2)
    {
        _PUT_MEM_CLASS_DATA(pData + startWord,
                    (value >> 16) | (prevValue << 16), unsigned int, memType);
    }
    else
    {
        _PUT_MEM_CLASS_DATA(pData + startWord,
                    (value >> 24) | (prevValue << 8), unsigned int, memType);
    }

    return;
}

#endif /* __V6V4_TUNNEL_UTIL_C__ */
