////////////////////////////////////////////////////////////////////////////////
//                                                                     
//                  I N T E L   P R O P R I E T A R Y                   
//                                                                      
//     COPYRIGHT (c)  2001-2002 BY  INTEL  CORPORATION.  ALL RIGHTS          
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//     THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                      
//                        INTEL  CORPORATION                            
//                                                                     
//                     2200 MISSION COLLEGE BLVD                        
//                                                                      
//               SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                      
////////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: v6v4_tunnel_decap.uc
//
//      Purpose: This file contains the macro for handling tunneled IPv6
//               packets that require decapsulation. 
//
//      History:
//
//
//      Date            Comment                         		
//      -----------------------------------------------------------------------
//
//      09/05/2002      Created
//
//
////////////////////////////////////////////////////////////////////////////////

#ifndef __V6V4_TUNNEL_DECAP_UC__
#define __V6V4_TUNNEL_DECAP_UC__


//------------------------------------------------------------------------------
// Exported APIs:
// v6v4_tunnel_decap(out_ipv6_hdr, in_ipv6_hdr, in_ipv4_hdr, 
//                   IPV6_START_WR, IPV6_START_RD, IPV4_START_RD)

//------------------------------------------------------------------------------
// Assumptions:
//    Assumes in_ipv4_hdr contains the IPv4 header and that a previous IPv4
//    forwarding operation has set the meta-data variables next_hop_id, and
//    next_hop_id_type.
//
//    Requires a global register context_id to exist for use in xbuf_activate.
//    Requires a macro dl_load_ipv6_hdr that can be invoked to load the
//    IPv6 header if the packet requires decapsulation.

//------------------------------------------------------------------------------
// Includes

#include "v6v4_tunnel_decap_uc.h"
#include "ipv6.h"
#include "ipv4.h"
#include "ixp_hw.uc"
#include "v6v4_tunnel_util.uc"

//------------------------------------------------------------------------------
// Global definitions

#ifdef V6V4_INGRESS_LIST_TRIE

#include "sram.uc"

// Calculate some convienient numbers based on the trie and index
// configuration.

// Number of indexes that can be stored in one index block
#define_eval V6V4_ISL_INDEX_PER_BLOCK ((V6V4_INGRESS_LIST_INDEX_LW * 2) - 1)

// Base address of trie blocks (follows the 1024-byte root)
#define_eval V4V6_ISL_TRIE_BLOCKS_BASE \
             (TUNNEL_INGRESS_LIST_SRAM_BASE + 1024)

// Size of trie blocks
#define_eval V6V4_ISL_TRIE_BLOCKS_SIZE \
             (TUNNEL_INGRESS_LIST_NUM_TRIE_BLOCKS * 64)

// Base address of index blocks (follows the blocks)
#define_eval V4V6_ISL_INDEX_BLOCKS_BASE \
             (V4V6_ISL_TRIE_BLOCKS_BASE + V6V4_ISL_TRIE_BLOCKS_SIZE)

#endif // ifdef V6V4_INGRESS_LIST_TRIE


//------------------------------------------------------------------------------
// Global registers
//
// None

//------------------------------------------------------------------------------
// Global signals

//------------------------------------------------------------------------------
// 
// None

////////////////////////////////////////////////////////////////////////////////
// v6v4_tunnel_decap_init
//
//    Description:
//        Set up global data for the v6v4_tunnel_decap microblock.
//
//    Parameters:
//        Outputs:
//        Inputs:
//
////////////////////////////////////////////////////////////////////////////////
#macro v6v4_tunnel_decap_init()
#endm


#ifdef V6V4_INGRESS_LIST_TRIE
////////////////////////////////////////////////////////////////////////////////
// v6v4_tunnel_ingress_validate
//		Description:
//         Trie table lookup for ingress source list.  This is a simple trie
//         scheme using a 256 entry table and trie block lookups.  Maximum
//         number of reads is 7 for the trie table blocks, plus zero or more
//         reads to check the index table.  Note that the lookup process differs
//         from the forwarder LPM lookup in that the shortest prefix that allows
//         the tunnel is considered a successful validation.
//
//		Outputs:
//
//		Inputs:
//			in_ip_sa     : GPR with IP source address.
//          tunnel_index : Index of ending tunnel endpoint that received the
//                         packet with source address in_ip_as.
//          SUCCESS_LBL  : Label to branch to if source address is found
//                         in trie blocks, and allowed for the indicated
//                         tunnel. 
//
////////////////////////////////////////////////////////////////////////////////
#macro v6v4_tunnel_ingress_validate(in_ip_sa, tunnel_index, SUCCESS_LBL)
.begin
    .reg trie_base, index_base, offset, lookup 
	.reg prev_index, current_index
    .sig tmp_signal

	xbuf_alloc($rd_xfer, V6V4_INGRESS_LIST_INDEX_LW, read)

    // Get base addresses of tables
	immed32(trie_base, TUNNEL_INGRESS_LIST_SRAM_BASE);
	immed32(index_base, V4V6_ISL_INDEX_BLOCKS_BASE)


    // Form the offset into the base from 31:24 for the first read
	alu_shf[offset, 0x3, ~AND, in_ip_sa, >>22]


    #define_eval TRIE_NUM 0
    #define_eval LSB_NUM 24

	#while (TRIE_NUM < V6V4_MAX_TRIES)

        sram_read($rd_xfer0, trie_base, offset, 1, tmp_signal, tmp_signal,
	              optimize_mem);
	
        #if (TRIE_NUM == 0)
	        immed32(trie_base, V4V6_ISL_TRIE_BLOCKS_BASE)
        #endif

        // Check the index value in the current trie.  If zero, read the
		// next trie.

	    ld_field_w_clr[lookup, 0011, $rd_xfer0]
		alu_shf[current_index, --, B, $rd_xfer0, >>16]
        beq[read_next_trie/**/TRIE_NUM/**/#]

	    // else, we need to check the index blocks for the tunnel ID

  	read_isl_index_block/**/TRIE_NUM/**/#:

        // Read the index block.
        alu_shf[offset, --, B, current_index, <<V6V4_ISL_INDEX_SHF]

	    sram_read($rd_xfer0, index_base, offset, V6V4_INGRESS_LIST_INDEX_LW,
	              tmp_signal, tmp_signal, optimize_mem);

        // Check the first index location in the block.

        ld_field_w_clr[current_index, 0011, $rd_xfer0]
	    alu[--, tunnel_index, -, current_index]
	    beq[SUCCESS_LBL]

        // Check the remaining index locations.

	    #define_eval ADDRESS_REG 1

        #while (ADDRESS_REG < V6V4_INGRESS_LIST_INDEX_LW)

            ld_field_w_clr[current_index, 0011, $rd_xfer/**/ADDRESS_REG]
	        alu[--, tunnel_index, -, current_index]
	        beq[SUCCESS_LBL]

	        alu_shf[current_index, --, B, $rd_xfer/**/ADDRESS_REG, >>16]
	        alu[--, tunnel_index, -, current_index]
	        beq[SUCCESS_LBL]

	        #define_eval ADDRESS_REG (ADDRESS_REG + 1)
        #endloop

	    #undef ADDRESS_REG

        // Continue checking if there are additional blocks.  Next block
		// index is in the upper 16 bits of the first longword.  If there
		// are no more blocks, read the next trie.

	    alu[current_index, --, B, $rd_xfer0, >>16]
	    bne[read_isl_index_block/**/TRIE_NUM/**/#]


    read_next_trie/**/TRIE_NUM/**/#:

        #if (TRIE_NUM < (V6V4_MAX_TRIES - 1))
	        #define_eval LSB_NUM (LSB_NUM - 4)

            // Check the next trie, if there is one.
		    alu[--, lookup, OR, 0]
		    beq[tunnel_ingress_trie_lookup_end#]

            // Get offset from ip address and trie block address
		    // First get byte offset of entry within the trie block

		    #if (LSB_NUM > 0)
		        #define_eval SHIFT_AMT (LSB_NUM - 2)
		        alu_shf[offset, 0x3c, AND, in_ip_sa, >>SHIFT_AMT]
            #else
		        #define_eval SHIFT_AMT 2
		        alu_shf[offset, 0x3c, AND, in_ip_sa, <<SHIFT_AMT]
            #endif

		    // Now add the offset to the byte offset of the trie block relative
		    // to the base of the trie blocks.
		    alu_shf[lookup, --, B, lookup, <<V6V4_ISL_TRIE_SHF]
		    alu[offset, offset, +, lookup]
        #endif

		#define_eval TRIE_NUM (TRIE_NUM + 1)

    #endloop

    #undef TRIE_NUM
    #undef LSB_NUM
	#undef SHIFT_AMT

tunnel_ingress_trie_lookup_end#:
    xbuf_free[$rd_xfer]

.end

#endm

#else
////////////////////////////////////////////////////////////////////////////////
// v6v4_tunnel_ingress_validate
//		Description:
//         Array lookup for ingress source list.  This uses a simple list of
//         blocks, each containing up to 12 prefixes that are valid for the
//         tunnel.
//
//		Outputs:
//
//		Inputs:
//			in_ip_sa     : GPR with IP source address.
//          tnh_entry    : Tunnel next hop entry containing ingress list
//                         block index.
//          SUCCESS_LBL  : Label to branch to if source address is found
//                         in the list.
//
////////////////////////////////////////////////////////////////////////////////
#macro v6v4_tunnel_ingress_validate(in_ip_sa, tnh_entry, SUCCESS_LBL)
.begin
	 
    .reg index			// Block index
    .reg mask_len   	// Holds mask length from list block
    .reg mask_value		// Holds mask value for address being checked
    .reg ingress_list_base // Holds base address of ingress list blocks
    .reg address_value  // Holds address value from list block
    .sig sram_sig		// Signal for reading ingress list block

    ld_field_w_clr[index, 0011, tnh_entry], load_cc
    beq[v6v4_isl_validate_end#]

    // Allocate registers for reading the ingress list.

	#if (V6V4_INGRESS_LIST_MAX < 9)
        xbuf_alloc[$ingress_block, V6V4_INGRESS_LIST_MAX, read]
    #else
        xbuf_alloc[$ingress_block, V6V4_ISL_LW_COUNT, read]
    #endif

    immed32(ingress_list_base, TUNNEL_INGRESS_LIST_SRAM_BASE)

v6v4_read_ingress_source_block#:
	// Shift the index to get the byte offset.

    alu_shf[index, --, B,index, <<V6V4_ISL_SHF]

	#if (V6V4_INGRESS_LIST_MAX < 9)
        sram[read, $ingress_block0, index, ingress_list_base,
            V6V4_INGRESS_LIST_MAX], ctx_swap[sram_sig]
	#else
	    #define_eval OV_FLAG (1 << (REF_CNT_OV_OFFSET - REF_CNT_OFFSET))
		#define_eval REFCNT_VALUE ((V6V4_INGRESS_LIST_MAX - 1) | OV_FLAG)

		alu_shf[--, --, B, REFCNT_VALUE, <<REF_CNT_OFFSET] 
        sram[read, $ingress_block[0], index, ingress_list_base,
            max_16], ctx_swap[sram_sig], indirect_ref

		#undef OV_FLAG
		#undef REFCNT_VALUE
	#endif

    #define_eval ADDRESS_INDEX 0

    #while(ADDRESS_INDEX < V6V4_MAX_ADDRESS)
        // Get prefix masks.

        #if (ADDRESS_INDEX < 4)
	        #define_eval ADDRESS_REG (ADDRESS_INDEX + 2)
			#define_eval MASK_REG V6V4_ISL_PREFIX_1_4
        #elif (ADDRESS_INDEX < 8)
	        #define_eval ADDRESS_REG (ADDRESS_INDEX + 3)
			#define_eval MASK_REG V6V4_ISL_PREFIX_5_8
        #else
	        #define_eval ADDRESS_REG (ADDRESS_INDEX + 4)
			#define_eval MASK_REG V6V4_ISL_PREFIX_9_12
        #endif

		#define_eval SHIFT_AMT ((3 - (ADDRESS_INDEX % 4)) * 8)

        br_bclr[$ingress_block/**/[V6V4_ISL_FLAGS], ADDRESS_INDEX,
		    v6v4_next_addr_/**/ADDRESS_INDEX/**/#]

		ld_field_w_clr[mask_len, 0001, $ingress_block/**/[MASK_REG], >>SHIFT_AMT]
        immed[mask_value, 0xffffffff]

        // Set shift amount for indirect shift (extract from prefix mask).

        alu[--, mask_len, OR, 0]
        alu_shf[address_value, in_ip_sa, AND, mask_value, <<indirect]
        alu[--, address_value, -, $ingress_block/**/[ADDRESS_REG]]
        beq[SUCCESS_LBL]

	v6v4_next_addr_/**/ADDRESS_INDEX/**/#:

        #define_eval ADDRESS_INDEX (ADDRESS_INDEX + 1)
        #undef ADDRESS_REG
		#undef MASK_REG
		#undef SHIFT_AMT

    #endloop

    #ifdef V6V4_INGRESS_LIST_MULTI_BLOCKS
        // Get next block index if there is one.

        alu_shf[index, --, B, $ingress_block/**/V6V4_ISL_NEXT_BLOCK, >>16]
        bne[v6v4_read_ingress_source_block#]
    #endif

    #undef ADDRESS_INDEX
    xbuf_free[$ingress_block]

v6v4_isl_validate_end#:
.end
#endm
#endif


////////////////////////////////////////////////////////////////////////////////
// v6v4_tunnel_decap
//
//    Description:
//        Handles IPv6 packets that have been tunneled over IPv4 and require
//        decapsulation.  The IPv4 forwarder should have previously handled
//        the packet and set the next hop ID in the meta data.
//
//    Parameters:
//        Outputs:
//            - Meta-data offset, packet_size and header_type are set after
//              decapsulation.
//        Inputs:
//            - out_ipv6_hdr   : Buffer to use to load IPv6 header.
//            - in_ipv6_hdr    : Buffer name to use to read IPv6 header
//                               after load.
//            - in_ipv4_hdr    : Buffer name to use to read the IPv4 header.
//            - IPV6_START_WR  : Starting location of the IPv6 header in
//                               out_ipv6_hdr.
//            - IPV6_START_RD  : Starting location of the IPv6 header in
//                               in_ipv6_hdr.
//            - IPV4_START_RD  : Starting location of the IPv4 header in
//                               in_ipv4_hdr.
//            - next_hop_id from meta-data
//            - Tunnel next hop info is in SRAM
//
////////////////////////////////////////////////////////////////////////////////
#macro v6v4_tunnel_decap(out_ipv6_hdr, in_ipv6_hdr, in_ipv4_hdr, \
                         IPV6_START_WR, IPV6_START_RD, IPV4_START_RD)
.begin
    .reg tunnel_info			// GPR to hold tunnel info from next hop
    .reg context_id 			// Holds thread ID for xbuf_activate

    // Check if this block is supposed to process the packet.

    alu[--, dl_next_block, -, BID_V6V4_TUNNEL_DECAP]
	bne[v6v4_tunnel_decap_end#]

	.begin
        .reg nexthop_type

		dl_meta_get_nexthop_id_type(nexthop_type)

		alu[--, nexthop_type, -, NEXTHOP_TYPE_TUNNEL]
		bne[v6v4_no_decap#]
    .end

    // Activate the IPv4 header buffer.

	dl_get_thread_id(context_id)

    xbuf_activate(in_ipv4_hdr, 0, context_id, 0)

	//--------------------------------------------------------------------------
    // Check protocol type to determine if the packet is a tunneled packet.
	//--------------------------------------------------------------------------

	.begin
		.reg proto			 // Holds protocol from IPv4 header

		xbuf_extract(proto, in_ipv4_hdr, IPV4_START_RD, IP_PROTOCOL)
		alu[--, proto, -, V6V4_PROTOCOL]
		bne[v6v4_local#]
    .end

	//--------------------------------------------------------------------------
    // Check for a fragmented packet.
	//--------------------------------------------------------------------------

    .begin
		.reg fragment_info   // Holds fragment info from IPv4 header

		xbuf_extract(fragment_info, in_ipv4_hdr, IPV4_START_RD,
		             IP_FLAGS_FRAGMENT)

		alu_shf[--, --, B, fragment_info, <<V6V4_FRAGMENT_SHF]
		bne[v6v4_fragment#]
    .end

	//--------------------------------------------------------------------------
	// Read the tunneling info from SRAM
	//--------------------------------------------------------------------------

	.begin
        .reg nexthop_info_base      // base of the tunnel next hop info table
        .reg nexthop_index          // Next hop index
        .sig nexthop_sig            // Signal for completion of next hop read

		// Xfer reg for tunnel next hop read.

		xbuf_alloc[$next_hop, V6V4_DECAP_TNH_LW_COUNT, read]

		immed32(nexthop_info_base, TUNNEL_DECAP_NEXTHOP_SRAM_BASE)

		// Get the nexthop_index from the dispatch loop meta data.

		dl_meta_get_nexthop_id(nexthop_index)

		// Shift by the appropriate amount to get the byte address.

		alu_shf[nexthop_index, --, B, nexthop_index, <<V6V4_DECAP_TNH_SHF]

		// read next hop information from SRAM.

		sram[read, $next_hop[0], nexthop_index, nexthop_info_base,
		    V6V4_DECAP_TNH_LW_COUNT], ctx_swap[nexthop_sig]

		move(tunnel_info, $next_hop[0])

		xbuf_free[$next_hop]
    .end


	//--------------------------------------------------------------------------
	// Check validity flag.  If not set, then the tunnel has been deleted
	// or the tunnel info is out of sync with the route tables.
	//--------------------------------------------------------------------------

	br_bclr[tunnel_info, V6V4_DECAP_FLAGS_VALID, v6v4_decap_drop#]

	//--------------------------------------------------------------------------
	// Update header type to indicate a tunneled packet that requires
	// decapsulation.
	//--------------------------------------------------------------------------

	dl_meta_set_header_type(V6V4_HDR_TYPE_DECAP)

	//--------------------------------------------------------------------------
    // Load the IPv6 header.
	//--------------------------------------------------------------------------

    xbuf_deactivate(in_ipv4_hdr)

    dl_load_ipv6_hdr(in_ipv6_hdr, IPV6_START_RD, V6V4_IPV6_HDR_SIZE)

    xbuf_activate(in_ipv4_hdr, 0, context_id, 0)
    xbuf_activate(in_ipv6_hdr, 1, context_id, 1)

	//--------------------------------------------------------------------------
    // Perform automatic tunneling validations.
	//--------------------------------------------------------------------------


    .begin
		.reg ipv6_dest_addr  // Holds portion of IPv6 destination address
		.reg ipv4_dest_addr  // Holds IPv4 destination address

		// Check for an IPv4-compatible destination address.

        v6v4_is_address_ipv4_compatible(ipv6_dest_addr, in_ipv6_hdr,
		        IPV6_START_RD, V6V4_VERIFY_DST, v6v4_decap_verify_ipv6_src#)

		// Packet has an automatic tunnel destination.  Drop if the node
		// doesn't want them.
		 
		br_bclr[tunnel_info, V6V4_DECAP_FLAGS_AUTO, v6v4_decap_drop#]

		// The embedded IPv4 address (which was placed in ipv6_dest_addr
		// by the previous macro) must be the same as the packet destination
		// address.

		xbuf_extract(ipv4_dest_addr, in_ipv4_hdr, IPV4_START_RD,
		             IP_DESTINATION_ADDRESS)

		alu[--, ipv6_dest_addr, -, ipv4_dest_addr]
		bne[v6v4_decap_drop#]

	    // Before we are done, verify the source addresses if necessary.

	    #ifdef V6V4_DECAP_VERIFY_IPV6_SOURCE
	        v6v4_tunnel_ipv6_source_verify(in_ipv6_hdr, IPV6_START_RD,
			                               v6v4_decap_drop#)
	    #endif

        #ifdef V6V4_DECAP_VERIFY_IPV4_SOURCE
        .begin
	        .reg src_addr

            xbuf_extract(src_addr, in_ipv4_hdr, IPV4_START_RD,
			            IP_SOURCE_ADDRESS)

            v6v4_tunnel_ipv4_addr_verify(src_addr, 0, V6V4_DO_VERIFY,
		                                 v6v4_decap_drop#)
        .end
        #endif

		dl_set_exception[BID_V6V4_TUNNEL_DECAP, V6V4_DECAP_EXCEPTION_AUTO]
	    br[v6v4_decap_exception#]
    .end

	//--------------------------------------------------------------------------
	// Packet is not addressed to a tunnel endpoint, so no further processing is
	// required.
	//--------------------------------------------------------------------------

v6v4_no_decap#:
	br[v6v4_tunnel_decap_end#], defer[1]
	alu[dl_next_block, --, B, V6V4_NEXT_NO_DECAP]

	//--------------------------------------------------------------------------
	// Packet is not tunneled, so it must be for local delivery.  Send
	// to the core component.
	//--------------------------------------------------------------------------

v6v4_local#:
	dl_set_exception[BID_V6V4_TUNNEL_DECAP, V6V4_DECAP_EXCEPTION_LOCAL]
	br[v6v4_decap_exception#]


	//--------------------------------------------------------------------------
	// Packet is fragmented.  Send to the core component.
	//--------------------------------------------------------------------------

v6v4_fragment#:
	dl_set_exception[BID_V6V4_TUNNEL_DECAP, V6V4_DECAP_EXCEPTION_FRAGMENT]

	//--------------------------------------------------------------------------
	// Count exceptions.
	//--------------------------------------------------------------------------

v6v4_decap_exception#:
    #ifdef V6V4_DECAP_COUNTERS
    .begin
	    .reg counter_base

		immed32(counter_base, V6V4_DECAP_COUNTER_BASE)
		v6v4_tunnel_increment_counter(counter_base, V6V4_DECAP_EXCP_COUNT)
    .end
    #endif

	br[v6v4_tunnel_decap_end#], defer[1]
	alu[dl_next_block, --, B, IX_EXCEPTION]

	//--------------------------------------------------------------------------
	// Verify IPv6 source address if necessary.
	//--------------------------------------------------------------------------

v6v4_decap_verify_ipv6_src#:

	#ifdef V6V4_DECAP_VERIFY_IPV6_SOURCE
	    v6v4_tunnel_ipv6_source_verify(in_ipv6_hdr, IPV6_START_RD,
		                               v6v4_decap_drop#)
	#endif


	//--------------------------------------------------------------------------
	// Verify 6to4 source and destination if necessary.
	//--------------------------------------------------------------------------

v6v4_decap_verify_6to4#:

    br_bclr[tunnel_info, V6V4_DECAP_FLAGS_6TO4, v6v4_decap_validate_source#]

    v6v4_tunnel_6to4_verify(in_ipv6_hdr, 0, IPV6_START_RD,
	                        V6V4_VERIFY_SRC, v6v4_decap_drop#)

    v6v4_tunnel_6to4_verify(in_ipv6_hdr, 0, IPV6_START_RD,
	                        V6V4_VERIFY_DST, v6v4_decap_drop#)

	//--------------------------------------------------------------------------
	// Validate the source address using the ingress source list if required.
	//--------------------------------------------------------------------------

v6v4_decap_validate_source#:

    br_bclr[tunnel_info, V6V4_DECAP_FLAGS_SRC_VALIDATE,
	        v6v4_decap_verify_src_type#]

    .begin
	    .reg src_addr

         #ifdef V6V4_INGRESS_LIST_TRIE
		 .reg tunnel_id
         #endif

		xbuf_extract(src_addr, in_ipv4_hdr, IPV4_START_RD, IP_SOURCE_ADDRESS)

        #ifdef V6V4_INGRESS_LIST_TRIE
		    dl_meta_get_nexthop_id(tunnel_id)

		    v6v4_tunnel_ingress_validate(
			    src_addr, tunnel_id, v6v4_decapsulate#)
        #else
		    v6v4_tunnel_ingress_validate(
			    src_addr, tunnel_info, v6v4_decapsulate#)
        #endif
    .end


	//--------------------------------------------------------------------------
	// Drop Packet Handler.
	//--------------------------------------------------------------------------

v6v4_decap_drop#:
    #ifdef V6V4_DECAP_COUNTERS
    .begin
	    .reg counter_base

		immed32(counter_base, V6V4_DECAP_COUNTER_BASE)
		v6v4_tunnel_increment_counter(counter_base, V6V4_DECAP_DROP_COUNT)
    .end
    #endif

	br[v6v4_tunnel_decap_end#], defer[1]
	alu[dl_next_block, --, B, IX_DROP]

	//--------------------------------------------------------------------------
    // If validation was not performed using the ingress source list, check
	// the address for acceptability according to RFC 2893.
	//--------------------------------------------------------------------------

v6v4_decap_verify_src_type#:

    #ifdef V6V4_DECAP_VERIFY_IPV4_SOURCE
    .begin
	    .reg src_addr

        xbuf_extract(src_addr, in_ipv4_hdr, IPV4_START_RD, IP_SOURCE_ADDRESS)

        v6v4_tunnel_ipv4_addr_verify(src_addr, 0, V6V4_DO_VERIFY,
		                             v6v4_decap_drop#)
    .end
    #endif

	//--------------------------------------------------------------------------
	// Decapsulate the packet.
	//--------------------------------------------------------------------------

v6v4_decapsulate#:

    // Copy TOS byte to IPv6 traffic class if requested.  Note this is not the
	// behavior defined by RFC 2893, but can be used if future RFCs specify it.
	
    br_bclr[tunnel_info, V6V4_DECAP_FLAGS_TOS_V6, v6v4_decap_update_meta#], defer[1]
	alu[dl_next_block, -- ,B, V6V4_NEXT_IPV6]

    .begin
	    .reg tr_class tos_byte

		#define IPV6_TRF_CLASS  0, 2

		xbuf_extract(tr_class, in_ipv6_hdr, IPV6_START_RD, IPV6_TRF_CLASS)

	    xbuf_activate(out_ipv6_hdr, 1, context_id, 0)
	    xbuf_extract(tos_byte, in_ipv4_hdr, IPV4_START_RD, IP_TYPE_OF_SERVICE)

        alu_shf[tr_class, tr_class, AND~, 0xff, <<4]
		alu_shf[tr_class, tr_class, OR, tos_byte, <<4]

		xbuf_insert(out_ipv6_hdr, tr_class, IPV6_START_WR, IPV6_TRF_CLASS)

		xbuf_deactivate(out_ipv6_hdr)
		#undef IPV6_TRF_CLASS
    .end

v6v4_decap_update_meta#:

	// Update buffer size.

	.begin
		.reg buf_size
		dl_meta_get_buffer_size(buf_size)
		sub(buf_size, buf_size, V6V4_IPV4_HDR_SIZE)
		dl_meta_set_buffer_size(buf_size)
	.end

	// Update packet_size.

    .begin
        .reg cur_length
        dl_meta_get_packet_size(cur_length)
        sub(cur_length, cur_length, V6V4_IPV4_HDR_SIZE)
	    dl_meta_set_packet_size(cur_length)
    .end

#ifdef V6V4_UPDATE_OFFSET
	// Update buffer offset.

    .begin
        .reg buf_offset
        dl_meta_get_offset(buf_offset)
        add(buf_offset, buf_offset, V6V4_IPV4_HDR_SIZE)
	    dl_meta_set_offset(buf_offset)
    .end
#endif

v6v4_tunnel_decap_end#:

    xbuf_deactivate(in_ipv4_hdr)

.end
#endm // v6v4_tunnel_decap


#endif // __V6V4_TUNNEL_DECAP_UC__
