////////////////////////////////////////////////////////////////////////////////
//                                                                     
//                  I N T E L   P R O P R I E T A R Y                   
//                                                                      
//     COPYRIGHT (c)  2001-2003 BY  INTEL  CORPORATION.  ALL RIGHTS          
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//     THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                      
//                        INTEL  CORPORATION                            
//                                                                     
//                     2200 MISSION COLLEGE BLVD                        
//                                                                      
//               SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                      
////////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: v6v4_tunnel_encap.uc
//
//      Purpose: This file contains the macro for handling IPv6
//               packets that require encapsulation in IPv4.
//
//      History:
//
//
//      Date            Comment                         		
//      -----------------------------------------------------------------------
//
//      09/05/2002      Created
//
//
////////////////////////////////////////////////////////////////////////////////

#ifndef __V6V4_TUNNEL_ENCAP_UC__
#define __V6V4_TUNNEL_ENCAP_UC__


//------------------------------------------------------------------------------
// Exported APIs:
// v6v4_tunnel_encap(out_ipv4_hdr, in_ipv6_hdr, IPV4_START_WR, IPV6_START_RD)

// Assumptions:
//    Assumes in_ipv6_hdr contains the IPv6 header and that a previous IPv6
//    forwarding operation has set the meta-data variables next_hop_id, and
//    next_hop_id_type.
//
//    Requires a global register context_id to exist for use in xbuf_activate.

//------------------------------------------------------------------------------
// Includes.

#include "v6v4_tunnel_encap_uc.h"
#include "v6v4_tunnel_util.uc"

//------------------------------------------------------------------------------
// Global definitions
//
// None


//------------------------------------------------------------------------------
// Global registers.

.reg global v6v4_id_base	// Holds base value for ID word in IPv4 header
.reg global v6v4_id_seq     // Holds sequence number for ID word in IPv4 header

//------------------------------------------------------------------------------
// Global signals
//
// None

//------------------------------------------------------------------------------
// 

////////////////////////////////////////////////////////////////////////////////
// v6v4_tunnel_encap_init
//
//    Description:
//        Set up global data for the v6v4_tunnel_encap microblock.
//
//    Parameters:
//        Outputs:
//            - Global register v6v4_id_base is set up to the base value
//              that is used to form the ID field in the encapsulated packets.
//        Inputs:
//
////////////////////////////////////////////////////////////////////////////////
#macro v6v4_tunnel_encap_init()
.begin
	.reg ctx_sts  // Holds result of reading active_ctx_sts

	local_csr_rd[ACTIVE_CTX_STS]
	immed[ctx_sts, 0]

	// The active context status register has the microengine number and
	// active context number in the low-order 8 bits.  Get those into the
	// ID base register.

	ld_field_w_clr[v6v4_id_base, 0001, ctx_sts]

	immed[v6v4_id_seq, 0]

.end
#endm


////////////////////////////////////////////////////////////////////////////////
// v6v4_tunnel_encap
//
//    Description:
//        Handles IPv6 packets that require tunneling over IPv4 to reach the
//        next hop.  The IPv6 forwarder should have previously handled
//        the packet and set the next hop ID in the meta data.
//
//    Parameters:
//        Outputs:
//            - Meta-data packet_size and header_type are set after
//              decapsulation.
//        Inputs:
//            - next_hop_id from meta-data
//            - Next hop info is in SRAM
//            - out_ipv4_hdr   : Buffer to use to write the IPv4 header.
//            - in_ipv6_hdr    : Buffer name to use to read IPv6 header.
//            - IPV4_START_WR  : Starting location of the IPv4 header in
//                               out_ipv4_hdr.
//            - IPV6_START_RD  : Starting location of the IPv6 header in
//                               in_ipv6_hdr.
//
////////////////////////////////////////////////////////////////////////////////
#macro v6v4_tunnel_encap(out_ipv4_hdr, in_ipv6_hdr, \
                         IPV4_START_WR, IPV6_START_RD)
.begin
    .reg packet_size
	.reg mtu
	.reg context_id

    // Check if this block is supposed to process the packet.

    alu[--, dl_next_block, -, BID_V6V4_TUNNEL_ENCAP]
	bne[v6v4_tunnel_encap_end#]

	.begin
        .reg nexthop_type

		dl_meta_get_nexthop_id_type(nexthop_type)

		alu[--, nexthop_type, -, NEXTHOP_TYPE_TUNNEL]
		bne[v6v4_no_encap#]
    .end


	// Xfer reg for tunnel next hop read

	xbuf_alloc[$next_hop, V6V4_ENCAP_TNH_LW_COUNT, read]

	#define_eval tunnel_flags_reg $next_hop/**/[V6V4_ENCAP_TNH_FLAGS_LW]

    // Activate the header buffers.

	dl_get_thread_id(context_id)

    xbuf_activate(in_ipv6_hdr, 1, context_id, 0)

	//--------------------------------------------------------------------------
	// Read the tunneling info from SRAM
	//--------------------------------------------------------------------------

	.begin
        .reg nexthop_info_base      // base of the tunnel next hop info table
        .reg nexthop_index          // Next hop index
        .sig nexthop_sig            // Signal for completion of next hop read

		immed32(nexthop_info_base, TUNNEL_ENCAP_NEXTHOP_SRAM_BASE)

		// Get the nexthop_index from the dispatch loop meta data.

		dl_meta_get_nexthop_id(nexthop_index)

		// Shift by the appropriate amount to get the byte address.

		alu_shf[nexthop_index, --, B, nexthop_index, <<V6V4_ENCAP_TNH_SHF]

		// read next hop information from SRAM.

		sram[read, $next_hop[0], nexthop_index, nexthop_info_base,
		    V6V4_ENCAP_TNH_LW_COUNT], ctx_swap[nexthop_sig]

    .end


	//--------------------------------------------------------------------------
	// Check validity flag.  If not set, then the tunnel has been deleted
	// or the tunnel info is out of sync with the route tables.
	//--------------------------------------------------------------------------

	br_bclr[tunnel_flags_reg, V6V4_ENCAP_FLAGS_VALID, v6v4_encap_drop#]

	//--------------------------------------------------------------------------
    // Check MTU, and send to XScale if the encapsulated packet is too large.
	//--------------------------------------------------------------------------

    .begin
        .reg min_mtu

        xbuf_extract(packet_size, in_ipv6_hdr, IPV6_START_RD,
                     IPV6_PAYLOAD_LENGTH)

        add(packet_size, packet_size, V6V4_ENCAP_HDR_SIZE)

        xbuf_extract(mtu, $next_hop, 0, V6V4_ENCAP_TNH_MTU)
        alu[--, mtu, -, packet_size]
        bhs[v6v4_encap_verify_6to4#]

        immed[min_mtu, V6V4_MIN_IPV6_MTU]
        alu[--, min_mtu, -, packet_size]
        blo[v6v4_packet_too_big#]
    .end

	//--------------------------------------------------------------------------
	// Verify 6to4 source and destination if necessary.
	//--------------------------------------------------------------------------

v6v4_encap_verify_6to4#:

    br_bclr[tunnel_flags_reg, V6V4_ENCAP_FLAGS_6TO4, v6v4_encapsulate#]

	.begin
	    .reg sn_bcast

		xbuf_extract(sn_bcast, $next_hop, 0, V6V4_ENCAP_TNH_SUBNET_BC)

        v6v4_tunnel_6to4_verify(in_ipv6_hdr, sn_bcast, IPV6_START_RD,
	                            V6V4_VERIFY_SRC, v6v4_encap_drop#)

        v6v4_tunnel_6to4_verify(in_ipv6_hdr, sn_bcast, IPV6_START_RD,
	                            V6V4_VERIFY_DST, v6v4_encap_drop#)
    .end

	//--------------------------------------------------------------------------
    // Construct the IPv4 header.
	//--------------------------------------------------------------------------

v6v4_encapsulate#:

    .begin
		.reg checksum     // Holds working checksum.
		.reg header_word  // Work register used in creating the IPvr header.
		.reg csum_upper   // Used to fold in upper 16 bits of checksum.

        xbuf_activate(out_ipv4_hdr, 0, context_id, 0)

		// LW 0 contains the version, header length, TOS and total packet
		// length.

        // If the TOS_V6 config flag is set, then special handling
		// is done for the TOS byte.  Note this is not the behavior defined
		// by RFC 2893, but can be used if future RFCs specify it.

		br_bclr[tunnel_flags_reg, V6V4_ENCAP_FLAGS_TOS_V6, v6v4_encap_tos_val#],
		       defer[1]

		immed[checksum, V6V4_VERS_IHL_TOS]

        // If this option is selected then the TOS byte is set from the
		// IPv6 traffic class.

		#define IPV6_TRF_CLASS  0, 2

        xbuf_extract(header_word, in_ipv6_hdr, IPV6_START_RD, IPV6_TRF_CLASS)

        br[v6v4_encap_word0#], defer[1]
        ld_field[checksum, 0001, header_word, >>4]

		#undef IPV6_TRF_CLASS

v6v4_encap_tos_val#:
		xbuf_extract(header_word, $next_hop, 0, V6V4_ENCAP_TNH_TOS)
        ld_field[checksum, 0001, header_word]

v6v4_encap_word0#:
		alu_shf[header_word, packet_size, OR, checksum, <<16]
		alu[checksum, checksum, +16, header_word]

		xbuf_insert(out_ipv4_hdr, header_word, IPV4_START_WR, IPV4_HDR_LW_0)

		// LW 1 contains the ID, flags and fragment offset.  We must determine
		// whether or not to set the DONT_FRAGMENT flag.  If the tunnel has
		// the NO_DF flag set or if the tunnel MTU is less than or equal to the
		// minimum MTU we don't set the flag.

		br_bset[tunnel_flags_reg, V6V4_ENCAP_FLAGS_NO_DF, v6v4_encap_no_df#],
		       defer[3]

		ld_field_w_clr[header_word, 0100, v6v4_id_seq, <<16]
		alu_shf[header_word, header_word, OR, v6v4_id_base, <<24]
		alu[v6v4_id_seq, v6v4_id_seq, +, 2]

        .begin
		    .reg min_mtu

            immed[min_mtu, V6V4_MIN_IPV6_MTU]
	        alu[--, min_mtu, -, mtu]
	        bhs[v6v4_encap_no_df#]

	        alu_shf[header_word, header_word, OR, 1, <<IPV4_DONT_FRAGMENT_BIT]
        .end

v6v4_encap_no_df#:
		xbuf_insert(out_ipv4_hdr, header_word, IPV4_START_WR, IPV4_HDR_LW_1)

		alu[checksum, checksum, +16, header_word]
		alu_shf[header_word, --, B, header_word, >>16]
		alu[checksum, checksum, +16, header_word]

		// LW 3 cotains the source address (we'll do LW 2 last).

		xbuf_extract(header_word, $next_hop, 0,
		             V6V4_ENCAP_TNH_LOCAL_ADDR_31_16)
		alu[checksum, checksum, +16, header_word]
		alu[checksum, checksum, +16,
		    $next_hop/**/[V6V4_ENCAP_TNH_LOCAL_ADDRESS_LW]]

        xbuf_insert(out_ipv4_hdr, 
		            $next_hop/**/[V6V4_ENCAP_TNH_LOCAL_ADDRESS_LW],
		            IPV4_START_WR, IPV4_HDR_LW_3)

		// LW 4 contains the destination address, which can either come from
		// the tunnel next hop or from the packet IPv6 destination address.

		br_bset[tunnel_flags_reg, V6V4_ENCAP_FLAGS_AUTO, v6v4_encap_6to4#]

		// Get destination address from tunnel next hop info.

		xbuf_insert(out_ipv4_hdr,
		            $next_hop/**/[V6V4_ENCAP_TNH_REMOTE_ADDRESS_LW],
		            IPV4_START_WR, IPV4_HDR_LW_4)

		xbuf_extract(header_word, $next_hop, 0,
		             V6V4_ENCAP_TNH_REMOTE_ADDR_31_16)
		br[v6v4_encap_checksum#], defer[2]
		alu[checksum, checksum, +16, header_word]
		alu[checksum, checksum, +16,
		    $next_hop/**/[V6V4_ENCAP_TNH_REMOTE_ADDRESS_LW]]

	v6v4_encap_6to4#:
		br_bclr[tunnel_flags_reg, V6V4_ENCAP_FLAGS_6TO4, v6v4_encap_auto#]

		// Get destination address from bits 111:80 of packet IPv6 destination.
		// The extracted address has already been validated.

		xbuf_extract(header_word, in_ipv6_hdr, IPV6_START_RD,
		             IPV6_6TO4_V4_DEST_ADDR_HI)
        xbuf_insert(out_ipv4_hdr, header_word, IPV4_START_WR, IPV4_HDR_WD_8)

		alu[checksum, checksum, +16, header_word]

		xbuf_extract(header_word, in_ipv6_hdr, IPV6_START_RD,
		             IPV6_6TO4_V4_DEST_ADDR_LO)

        xbuf_insert(out_ipv4_hdr, header_word, IPV4_START_WR, IPV4_HDR_WD_9)

		br[v6v4_encap_checksum#], defer[1]
		alu[checksum, checksum, +16, header_word]

	v6v4_encap_auto#:
		// Get destination address from last 32 bits of IPv6 destination.

		xbuf_extract(header_word, in_ipv6_hdr, IPV6_START_RD,
		             IPV6_AUTO_V4_DEST_ADDR)

        // Validate the extracted address.

        .begin
		    .reg sn_bcast

			#ifdef V6V4_ENCAP_VERIFY_DEST_AUTO
			    #define_eval VERIFY_ALL V6V4_DO_VERIFY
			#else
			    #define_eval VERIFY_ALL V6V4_DONT_VERIFY
            #endif

		    xbuf_extract(sn_bcast, $next_hop, 0, V6V4_ENCAP_TNH_SUBNET_BC)
		    v6v4_tunnel_ipv4_addr_verify(header_word, sn_bcast, VERIFY_ALL,
		                                 v6v4_encap_drop#)
        .end

        xbuf_insert(out_ipv4_hdr, header_word, IPV4_START_WR, IPV4_HDR_LW_4)

		alu[checksum, checksum, +16, header_word]
		alu_shf[header_word, --, B, header_word, >>16]
		alu[checksum, checksum, +16, header_word]

	v6v4_encap_checksum#:
		// Lastly, set LW 2, which contains the TTL, protocol and checksum.

        xbuf_extract(header_word, $next_hop, 0, V6V4_ENCAP_TNH_TTL)
		alu_shf[header_word, V6V4_PROTOCOL, OR, header_word, <<8]
		alu[checksum, checksum, +16, header_word]

    v6v4_encap_csum_get_upper#:
		alu_shf[csum_upper, --, B, checksum, >>16]
		beq[v6v4_encap_csum_invert#]

        br[v6v4_encap_csum_get_upper#], defer[2]
		// Clear upper 16 bits of checksum before adding in the overflow --
		// we know csum_upper has zeros in the 16 high order bits so it's
		// convenient to use that.
		ld_field[checksum, 1100, csum_upper]
		alu[checksum, checksum, +16, csum_upper]

		// Get 1's complement of the checksum to finish.

    v6v4_encap_csum_invert#:
		alu[checksum, --, ~B, checksum]
		ld_field[checksum, 1100, header_word, <<16]

        xbuf_insert(out_ipv4_hdr, checksum, IPV4_START_WR, IPV4_HDR_LW_2)

		#undef tunnel_flags_reg

    .end

	//--------------------------------------------------------------------------
	// Update the meta-data.
	//--------------------------------------------------------------------------

v6v4_encap_update_meta#:

	// Update buffer size.

	.begin
		.reg buf_size
		dl_meta_get_buffer_size(buf_size)
		add(buf_size, buf_size, V6V4_IPV4_HDR_SIZE)
		dl_meta_set_buffer_size(buf_size)
	.end

	// Update header type.

	.begin
		.reg hdr_type
		dl_meta_get_header_type(hdr_type)
		alu[hdr_type, hdr_type, OR, V6V4_HDR_TYPE_ENCAP]
		dl_meta_set_header_type(hdr_type)
	.end

	// Update packet_size.

    dl_meta_get_packet_size(packet_size)
    add(packet_size, packet_size, V6V4_IPV4_HDR_SIZE)
    dl_meta_set_packet_size(packet_size)


#ifdef V6V4_UPDATE_OFFSET
	// Update buffer offset.

    .begin
        .reg buf_offset
        dl_meta_get_offset(buf_offset)
        sub(buf_offset, buf_offset, V6V4_IPV4_HDR_SIZE)
	    dl_meta_set_offset(buf_offset)
    .end
#endif

    // Set next hop ID to -1 for the IPv4 forwarder.
	.begin
	    .reg nh_id
        immed[nh_id, 0xffff]
		dl_meta_set_nexthop_id(nh_id)
    .end

	// Set next block ID.

	br[v6v4_tunnel_encap_end#], defer[1]
	alu[dl_next_block, -- ,B, V6V4_NEXT_IPV4]


	//--------------------------------------------------------------------------
	// Packet is not addressed to a tunnel endpoint.  Bypass.
	//--------------------------------------------------------------------------

v6v4_no_encap#:
	br[v6v4_tunnel_encap_end#], defer[1]
	alu[dl_next_block, --, B, V6V4_NEXT_NO_ENCAP]

	//--------------------------------------------------------------------------
	// Drop Packet Handler.
	//--------------------------------------------------------------------------

v6v4_encap_drop#:
    #ifdef V6V4_ENCAP_COUNTERS
    .begin
	    .reg counter_base

		immed32(counter_base, V6V4_ENCAP_COUNTER_BASE)
		v6v4_tunnel_increment_counter(counter_base, V6V4_ENCAP_DROP_COUNT)
    .end
    #endif

	br[v6v4_tunnel_encap_end#], defer[1]
	alu[dl_next_block, --, B, IX_DROP]

	//--------------------------------------------------------------------------
	// Packet is not tunneled, so it must be for local delivery.  Send
	// to the core component.
	//--------------------------------------------------------------------------

v6v4_packet_too_big#:
	dl_set_exception[BID_V6V4_TUNNEL_ENCAP, V6V4_ENCAP_EXCEPTION_MTU]
	alu[dl_next_block, --, B, IX_EXCEPTION]

    #ifdef V6V4_ENCAP_COUNTERS
    .begin
	    .reg counter_base

		immed32(counter_base, V6V4_ENCAP_COUNTER_BASE)
		v6v4_tunnel_increment_counter(counter_base, V6V4_ENCAP_EXCP_COUNT)
    .end
    #endif


v6v4_tunnel_encap_end#:

    xbuf_deactivate(out_ipv4_hdr)
    xbuf_deactivate(in_ipv6_hdr)

    xbuf_free[$next_hop]

.end
#endm // v6v4_tunnel_encap


#endif // __V6V4_TUNNEL_ENCAP_UC__