////////////////////////////////////////////////////////////////////////////////
//                                                                     
//                  I N T E L   P R O P R I E T A R Y                   
//                                                                      
//     COPYRIGHT (c)  2001-2002 BY  INTEL  CORPORATION.  ALL RIGHTS          
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//     THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                      
//                        INTEL  CORPORATION                            
//                                                                     
//                     2200 MISSION COLLEGE BLVD                        
//                                                                      
//               SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                      
////////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: v6v4_tunnel_util.uc
//
//      Purpose: This file contains utility macros used by the IPv6/v4
//               tunneling code. 
//
//      History:
//
//
//      Date            Comment                         		
//      -----------------------------------------------------------------------
//
//      10/09/2002      Created
//
//
////////////////////////////////////////////////////////////////////////////////

#ifndef __V6V4_TUNNEL_UTIL_UC__
#define __V6V4_TUNNEL_UTIL_UC__


//------------------------------------------------------------------------------
// Exported APIs:
// v6v4_tunnel_decap(out_ipv6_hdr, in_ipv6_hdr, in_ipv4_hdr, 
//                   IPV6_START_WR, IPV6_START_RD, IPV4_START_RD)

//------------------------------------------------------------------------------
// Assumptions:
//    Assumes in_ipv4_hdr contains the IPv4 header and that a previous IPv4
//    forwarding operation has set the meta-data variables next_hop_id, and
//    next_hop_id_type.
//
//    Requires a global register context_id to exist for use in xbuf_activate.
//    Requires a macro dl_load_ipv6_hdr that can be invoked to load the
//    IPv6 header if the packet requires decapsulation.

//------------------------------------------------------------------------------
// Includes

#include <v6v4_tunnel.h>
#include "ipv6.h"
#include "ipv4.h"

//------------------------------------------------------------------------------
// Global definitions

#define V6V4_IPV4_MCAST_MASK 0xE
#define V6V4_IPV4_MCAST_SHF  28

#define V6V4_LOOPBACK_IPV4   0x7f000001

// Prefix values for private IPv4 addresses

#define V6V4_IPV4_PRIVATE_24 10           // 10/8 prefix
#define V6V4_IPV4_PRIVATE_20 0xac1        // 172.16/12 prefix
#define V6V4_IPV4_PRIVATE_16 0xc0a8       // 192.168/16 prefix

// IPv6 header locations for surce and destination addresses.

#define V6V4_IPV6_SRC_ADDR_START     8
#define V6V4_IPV6_DST_ADDR_START     24

// Defines for address selection

#define V6V4_VERIFY_SRC              0
#define V6V4_VERIFY_DST              1

// Defines for selective verification

#define V6V4_DONT_VERIFY             0
#define V6V4_DO_VERIFY               1

//------------------------------------------------------------------------------
// Global registers
//
// None

//------------------------------------------------------------------------------
// Global signals

//------------------------------------------------------------------------------
// 
// None

////////////////////////////////////////////////////////////////////////////////
// v6v4_tunnel_ipv4_source_verify
//
//    Description:
//        Verify the tunnel source IPv4 address according to RFC 2893.
//
//    Parameters:
//        Outputs:
//        Inputs:
//            in_addr      : IPv4 source address.
//            subnet_bcast : Subnet broadcast address to which in_addr is
//                           to be compared.
//            VERIFY_ALL   : Set to 1 if all checks are to be performed, or
//                           0 if only those checks not typically performed
//                           by the forwarder for destination addresses are to
//                           be done.
//            FAIL_LBL     : Label to branch to if address is not acceptable.
//            
//
////////////////////////////////////////////////////////////////////////////////
#macro v6v4_tunnel_ipv4_addr_verify(in_addr, subnet_bcast, VERIFY_ALL, FAIL_LBL)
.begin
    .reg temp

#if (VERIFY_ALL == V6V4_DO_VERIFY)
    // Check for unspecified address (0.0.0.0)

    alu[--, in_addr, -, 0]
	beq[FAIL_LBL]

	// Check for loopback address (127.0.0.1)

    immed32(temp, V6V4_LOOPBACK_IPV4)
    alu[--, in_addr, -, temp]
    beq[FAIL_LBL]
#endif

	// Check for broadcast address (255.255.255.255)

    alu[temp, --, ~B, 0]
    alu[--, in_addr, -, temp]
    beq[FAIL_LBL]

	// Check for multicast address (224.x.x.x to 239.x.x.x)

    alu_shf_right(temp, V6V4_IPV4_MCAST_MASK, -, in_addr, V6V4_IPV4_MCAST_SHF)
    beq[FAIL_LBL]

	// If a subnet broadcast address was specified, compare to the IPv4
	// address.

#if (!isnum(subnet_bcast))
    alu[--, subnet_bcast, -, in_addr]
	beq[FAIL_LBL]
#endif

.end
#endm


////////////////////////////////////////////////////////////////////////////////
// v6v4_tunnel_ipv6_source_verify
//
//    Description:
//        Verify the tunnel source IPv6 address according to RFC 2893.
//
//    Parameters:
//        Outputs:
//        Inputs:
//            ipv6_hdr    : Buffer containing IPv6 header.
//            IPV6_OFFSET : Offset to start of IPv6 header.
//            FAIL_LBL    : Label to branch to if address is not acceptable.
//            
//
////////////////////////////////////////////////////////////////////////////////
#macro v6v4_tunnel_ipv6_source_verify(ipv6_hdr, IPV6_OFFSET, FAIL_LBL)
.begin
    .reg sa_value

    // Check for multicast address (first byte == 0xff)

    xbuf_extract(sa_value, ipv6_hdr, IPV6_OFFSET, IPV6_SRC_ADDR_127_96)

    alu_shf_right(--, 0xFF, -, sa_value, 24)
    beq[FAIL_LBL]
	
	alu[--, --, B, sa_value]
	bne[v4v6_tunnel_ipv6_source_verify_end#]

	xbuf_extract(sa_value, ipv6_hdr, IPV6_OFFSET, IPV6_SRC_ADDR_95_64)
	alu[--, --, B, sa_value]
	bne[v4v6_tunnel_ipv6_source_verify_end#]

	xbuf_extract(sa_value, ipv6_hdr, IPV6_OFFSET, IPV6_SRC_ADDR_63_32)
	alu[--, --, B, sa_value]
	bne[v4v6_tunnel_ipv6_source_verify_end#]

	xbuf_extract(sa_value, ipv6_hdr, IPV6_OFFSET, IPV6_SRC_ADDR_31_0)

	// Check for loopback or unspecified address (:: 0 or ::1)

    alu[--, sa_value, AND~, 0x1]
	beq[FAIL_LBL]

    // If we get to here, we have an IPv4-compatible address, and the
	// embedded address is in sa_value.  Check it for acceptability.

    v6v4_tunnel_ipv4_addr_verify(sa_value, 0, V6V4_DO_VERIFY, FAIL_LBL)

v4v6_tunnel_ipv6_source_verify_end#:

.end
#endm


////////////////////////////////////////////////////////////////////////////////
// v6v4_tunnel_6to4_v4addr_verify
//
//    Description:
//        Verify that an IPv4 address embedded in a 6to4 address is acceptable
//        according to RFC 3056.
//
//    Parameters:
//        Outputs:
//        Inputs:
//            in_addr      : Embedded IPv4 address.
//            subnet_bcast : Subnet broadcast address to which in_addr is
//                           to be compared.
//            FAIL_LBL     : Label to branch to if address is not acceptable.
//            
//
////////////////////////////////////////////////////////////////////////////////
#macro v6v4_tunnel_6to4_v4addr_verify(in_addr, subnet_bcast, FAIL_LBL)
.begin
    .reg prefix

	// The 6to4 validations include those specified in RFC 2893 as well as
	// subnet broadcast and private addresses.

    v6v4_tunnel_ipv4_addr_verify(in_addr, subnet_bcast, V6V4_DO_VERIFY,
	                             FAIL_LBL)

    // Check for "24-bit block" private addresses (10/8 prefix)

    alu_shf[prefix, --, B, in_addr, >>24]
	alu[--, prefix, -, V6V4_IPV4_PRIVATE_24]
	beq[FAIL_LBL]

    // Check for "20-bit block" private addresses (172.16/12 prefix)

	immed[prefix, V6V4_IPV4_PRIVATE_20]
	alu_shf_right(--, prefix, -, in_addr, 20)
	beq[FAIL_LBL]

    // Check for "16-bit block" private addresses (192.168/16 prefix)

	immed[prefix, V6V4_IPV4_PRIVATE_16]
	alu_shf_right(--, prefix, -, in_addr, 16)
	beq[FAIL_LBL]

.end
#endm


/////////////////////////////////////////////////////////////////////////////////////
// v6v4_is_address_6to4
//
//    Description:
//        Check the indicated IPv6 address for a 6to4 address,  If not, branch
//        to NOT_6TO4_LBL.  If so, return the embedded address in out_addr.
//
//    Parameters:
//        Outputs:
//            out_addr     : Register in which to return the embedded address.
//
//        Inputs:
//            ipv6_hdr     : Buffer containing IPv6 header.
//            IPV6_OFFSET  : Offset to start of IPv6 header.
//            WHICH_ADDR   : 1 = Dest, 0 = Src
//            NOT_6TO4_LBL : Label to branch to if address is not 6to4.
//            
//
////////////////////////////////////////////////////////////////////////////////
#macro v6v4_is_address_6to4(out_addr, ipv6_hdr, IPV6_OFFSET, WHICH_ADDR,\
       NOT_6TO4_LBL)
.begin
	.reg pfx_6to4

#if (WHICH_ADDR == V6V4_VERIFY_SRC)
    #define_eval PFX_START V6V4_IPV6_SRC_ADDR_START
#else
    #define_eval PFX_START V6V4_IPV6_DST_ADDR_START
#endif

    xbuf_extract(out_addr, ipv6_hdr, IPV6_OFFSET, PFX_START, 
	             V6V4_6TO4_PREFIX_LENGTH_BYTES)

    immed[pfx_6to4, V6V4_6TO4_PREFIX]
    alu[--, out_addr, -, pfx_6to4]
	bne[NOT_6TO4_LBL]

	#define V4ADDR_START (PFX_START + V6V4_6TO4_PREFIX_LENGTH_BYTES)

	xbuf_extract(out_addr, ipv6_hdr, IPV6_OFFSET, V4ADDR_START, 4)

	#undef PFX_START
	#undef V4ADDR_START

.end
#endm


////////////////////////////////////////////////////////////////////////////////
// v6v4_tunnel_6to4_verify
//
//    Description:
//        Check the indicated IPv6 address for a 6to4 address,
//        and if detected, validate the embedded IPv4 address for acceptability
//        according to RFC 3056.
//
//    Parameters:
//        Outputs:
//        Inputs:
//            ipv6_hdr     : Buffer containing IPv6 header.
//            subnet_bcast : Subnet broadcast address to which in_addr is
//                           to be compared.
//            IPV6_OFFSET  : Offset to start of IPv6 header.
//            WHICH_ADDR   : 0 = Source, 1 = Dest.
//            FAIL_LBL     : Label to branch to if address is not acceptable.
//            
//
////////////////////////////////////////////////////////////////////////////////
#macro v6v4_tunnel_6to4_verify(ipv6_hdr, subnet_bcast, IPV6_OFFSET, \
       WHICH_ADDR, FAIL_LBL)
.begin
    .reg sa_value

    // Check if the address is a 6to4 address.

    v6v4_is_address_6to4(sa_value, ipv6_hdr, IPV6_OFFSET, WHICH_ADDR,
	                     v6v4_tunnel_6to4_verify_end#)

    // If the address is 6to4, then the embedded IPv4 address is now in
	// sa_value.  Validate it for acceptability according to RFC 3056.

	v6v4_tunnel_6to4_v4addr_verify(sa_value, subnet_bcast, FAIL_LBL)

v6v4_tunnel_6to4_verify_end#:

.end
#endm


/////////////////////////////////////////////////////////////////////////////////////
// v6v4_is_address_ipv4_compatible
//
//    Description:
//        Check the indicated IPv6 address for an IPv4-compatible address,
//        If not, branch  to NOT_IPV4_COMP_LBL.  If so, return the embedded
//        address in out_addr.
//
//    Parameters:
//        Outputs:
//            out_addr     : Register in which to return the embedded address.
//
//        Inputs:
//            ipv6_hdr          : Buffer containing IPv6 header.
//            IPV6_OFFSET       : Offset to start of IPv6 header.
//            WHICH_ADDR        : 1 = Dest, 0 = Src
//            NOT_IPV4_COMP_LBL : Label to branch to if address is not
//                                IPv4-compatible.
//            
//
////////////////////////////////////////////////////////////////////////////////
#macro v6v4_is_address_ipv4_compatible(out_addr, ipv6_hdr, IPV6_OFFSET,\
       WHICH_ADDR, NOT_IPV4_COMP_LBL)
.begin

#if (WHICH_ADDR == V6V4_VERIFY_SRC)
    #define_eval PFX_START V6V4_IPV6_SRC_ADDR_START
#else
    #define_eval PFX_START V6V4_IPV6_DST_ADDR_START
#endif

	xbuf_extract(out_addr, ipv6_hdr, IPV6_OFFSET, PFX_START, 4)

	alu[--, out_addr, -, V6V4_COMPATIBLE_PREFIX_0_1]
	bne[NOT_IPV4_COMP_LBL]

	#define_eval PFX_START (PFX_START + 4)

	xbuf_extract(out_addr, ipv6_hdr, IPV6_OFFSET, PFX_START, 4)

	alu[--, out_addr, -, V6V4_COMPATIBLE_PREFIX_2_3]
	bne[NOT_IPV4_COMP_LBL]

	#define_eval PFX_START (PFX_START + 4)

	xbuf_extract(out_addr, ipv6_hdr, IPV6_OFFSET, PFX_START, 4)

	alu[--, out_addr, -, V6V4_COMPATIBLE_PREFIX_4_5]
	bne[NOT_IPV4_COMP_LBL]

	// Address is IPv4-compatible.  Extract the embedded address

	#define_eval V4ADDR_START (PFX_START + 4)

	xbuf_extract(out_addr, ipv6_hdr, IPV6_OFFSET, V4ADDR_START, 4)

	#undef PFX_START
	#undef V4ADDR_START

.end
#endm


///////////////////////////////////////////////////////////////////////////
// v6v4_tunnel_increment_counter
//
//    Description:
//        Atomically increment a tunneling counter.
//
//    Parameters:
//        Outputs:
//            Increments counter value in SRAM.
//
//        Inputs:
//            cntr_base    : Base address for counters.
//            OFFSET       : Constant offset to counter to be incremented.
//            
////////////////////////////////////////////////////////////////////////////////
#macro v6v4_tunnel_increment_counter(cntr_base, OFFSET)

    sram[incr, --, OFFSET, cntr_base]
#endm

#endif // __V6V4_TUNNEL_DECAP_UC__