/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2400 Network Processor, Release 6
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      Core Component Infrastructure - Ingress source validation list
 *      internal API.
 *
 * = FILENAME
 *      ix_cc_v6v4_ingress_list.h
 *
 * = DESCRIPTION
 *      This file defines all data types, macros and function prototypes
 *      that support management of the ingress source validation list for
 *      tunneling decapsulation.  Specific implementations of the functions
 *      defined in this file exist for each supported format of the ingress
 *      list.
 *
 * = CREATION TIME
 *      11/25/2002 1:42 PM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 * $Id: ix_cc_v6v4_ingress_list.h,v 1.3 2003/01/30 18:31:53 cabell Exp $
 */

#if !defined(__IX_CC_V6V4_INGRESS_LIST_H__)
#define __IX_CC_V6V4_INGRESS_LIST_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */

/**
 * TYPENAME: ix_cc_v6v4_ingress_list_data
 * 
 * DESCRIPTION: This structure holds pointers to the resources allocated
 * to manage the ingress source list.
 *
 */

typedef struct ix_s_cc_v6v4_ingress_list_data
{
    /* base address of shared memory allocated for the ingress lists */
    ix_uint32* pMbIslBase;

    /* internal context information */
    void* pIslContext;
} ix_cc_v6v4_ingress_list_data;

/**
 * NAME: _ix_cc_v6v4_ingress_list_setup
 *
 * DESCRIPTION: This primitive is used to setup the ingress source list.
 * This includes allocating all shared memory and internal memory used
 * to manage the list.
 *           
 * @Param:  - IN None.
 *
 * @Param:  - OUT arg_pData - pointer to location in which to return the
 * pointers to the allocated resources.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_OOR - out of resources 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */

typedef ix_error (*_ix_cc_v6v4_ingress_list_setup)(
    ix_cc_v6v4_ingress_list_data* arg_pData
    );


/**
 * NAME: _ix_cc_v6v4_ingress_list_cleanup
 *
 * DESCRIPTION: This primitive is used to cleanup the ingress source list.
 * This includes freeing all shared memory and internal memory used
 * to manage the list.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */

typedef ix_error (*_ix_cc_v6v4_ingress_list_cleanup)(
    ix_cc_v6v4_ingress_list_data* arg_pData
    );


/**
 * NAME: _ix_cc_v6v4_ingress_list_add_prefix
 *
 * DESCRIPTION: This primitive is used to add a prefix to the ingress list
 * for a tunnel.  The caller must have already validated the tunnel index
 * and the address for acceptability.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_TunnelIndex - end tunnel index.
 *
 * @Param:  - IN arg_pEntry - pointer to ingress list entry.
 *
 * @Param:  - IN/OUT arg_pTunnelInfo - on input, this points to the 16-bit
 * word in the tunnel configuration entry pertaining to the ingress
 * list.  On output, it contains the value that should be written to the
 * 16-bit word in the tunnel configuration entry.  The 16-bit word is a
 * value assigned by the ingress list handler and may change when a prefix
 * is added or deleted, and must be zero only if there are no prefixes
 * allowed for the tunnel.  The interpretation of non-zero values is format-
 * specific.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_DUPLICATE_ENTRY - entry already exists 
 * IX_CC_ERROR_FULL - cannot add any more entries 
 * 
 */

typedef ix_error (*_ix_cc_v6v4_ingress_list_add_prefix)(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_uint16 arg_TunnelIndex,
    ix_cc_v6v4_ingress_source_entry* arg_pEntry,
    ix_uint16* arg_pTunnelInfo
    );



/**
 * NAME: _ix_cc_v6v4_ingress_list_del_prefix
 *
 * DESCRIPTION: This primitive is used to delete a prefix from the ingress
 * list for a tunnel.  The caller must have already validated the tunnel
 * index.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_TunnelIndex - end tunnel index.
 *
 * @Param:  - IN arg_pEntry - pointer to ingress list entry.
 *
 * @Param:  - IN/OUT arg_pTunnelInfo - on input, this points to the 16-bit
 * word in the tunnel configuration entry pertaining to the ingress
 * list.  On output, it contains the value that should be written to the
 * 16-bit word in the tunnel configuration entry.  The 16-bit word is a
 * value assigned by the ingress list handler and may change when a prefix
 * is added or deleted, and must be zero only if there are no prefixes
 * allowed for the tunnel.  The interpretation of non-zero values is format-
 * specific.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - cannot find specified entry 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */

typedef ix_error (*_ix_cc_v6v4_ingress_list_del_prefix)(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_uint16 arg_TunnelIndex,
    ix_cc_v6v4_ingress_source_entry* arg_pEntry,
    ix_uint16* arg_pTunnelInfo
    );


/**
 * NAME: _ix_cc_v6v4_ingress_list_get_prefixes
 *
 * DESCRIPTION: This primitive is used to retrieve the prefixes allowed for
 * a particular tunnel.  The caller must have already validated the tunnel
 * index.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_TunnelIndex - end tunnel index.
 *
 * @Param:  - IN arg_TunnelInfo - the 16-bit word in the tunnel
 * configuration entry pertaining to the ingress list.
 *
 * @Param:  - IN/OUT arg_pNumEntries - On input, points to the number of
 * entries the buffer pointed to by arg_paEntries can accommodate.  On output,
 * contains the actual number of prefixes in the ingress list for the
 * tunnel.
 *
 * @Param:  - OUT arg_paEntries - pointer to array into which to place the
 * entries.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_BUFFER_TOO_SMALL - buffer too small to contain the complete
 *                                     list.
 * 
 */

typedef ix_error (*_ix_cc_v6v4_ingress_list_get_prefixes)(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_uint16 arg_TunnelIndex,
    ix_uint16 arg_TunnelInfo,
    ix_uint32* arg_pNumEntries,
    ix_cc_v6v4_ingress_source_entry* arg_paEntries
    );


/**
 * NAME: _ix_cc_v6v4_ingress_list_del_tunnel
 *
 * DESCRIPTION: This primitive is used to delete all prefixes assigned to
 * a tunnel.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_TunnelIndex - end tunnel index.
 *
 * @Param:  - IN arg_pTunnelInfo - the 16-bit word in the tunnel configuration
 * entry pertaining to the ingress list.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */

typedef ix_error (*_ix_cc_v6v4_ingress_list_del_tunnel)(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_uint16 arg_TunnelIndex,
    ix_uint16 arg_pTunnelInfo
    );


/**
 * NAME: _ix_cc_v6v4_ingress_list_lookup
 *
 * DESCRIPTION: This primitive is used to lookup a prefix in the ingress list
 * to check whether it is allowed for a particular tunnel.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_IpAddr - IP source address.
 *
 * @Param:  - IN arg_TunnelIndex - end tunnel index.
 *
 * @Param:  - IN arg_TunnelInfo - the 16-bit word in the tunnel configuration
 * entry pertaining to the ingress list.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - address is not allowed for the tunnel 
 * 
 */

typedef ix_error (*_ix_cc_v6v4_ingress_list_lookup)(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_uint32 arg_IpAddr,
    ix_uint16 arg_TunnelIndex,
    ix_uint16 arg_TunnelInfo
    );


/**
 * TYPENAME: ix_cc_v6v4_ingress_list_handlers
 * 
 * DESCRIPTION: This structure defines the internal function table
 * for processing the ingress source validation list.
 *
 */

typedef struct ix_s_cc_v6v4_ingress_list_handlers
{
    _ix_cc_v6v4_ingress_list_setup        setup;
    _ix_cc_v6v4_ingress_list_cleanup      cleanup;
    _ix_cc_v6v4_ingress_list_add_prefix   addPrefix;
    _ix_cc_v6v4_ingress_list_del_prefix   delPrefix;
    _ix_cc_v6v4_ingress_list_get_prefixes getPrefixes;   
    _ix_cc_v6v4_ingress_list_del_tunnel   delTunnel;   
    _ix_cc_v6v4_ingress_list_lookup       lookup;   
} ix_cc_v6v4_ingress_list_handlers;


#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__IX_CC_V6V4_INGRESS_LIST_H__) */
