/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2400 Network Processor, Release 6
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      Core Component Infrastructure - Reassembly internal API.
 *
 * = FILENAME
 *      ix_cc_v6v4_reassembly.h
 *
 * = DESCRIPTION
 *      This file defines all data types, macros and function prototypes
 *      that support reassembly of tunneled packets.
 *
 * = CREATION TIME
 *      01/06/2003 10:43 AM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 * $Id: ix_cc_v6v4_reassembly.h,v 1.3 2003/01/30 18:33:31 cabell Exp $
 */

#if !defined(__IX_CC_V6V4_REASSEMBLY_H__)
#define __IX_CC_V6V4_REASSEMBLY_H__

#include "cc/ix_cc_hash_table128.h"
#include "cc/ix_cc_index_pool.h"

#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */

/*
 * NAME: ix_cc_v6v4_reassembly
 *
 * DESCRIPTION: Public handle for using the reassembly utility.
 */
typedef struct ix_s_cc_v6v4_reassembly * ix_cc_v6v4_reassembly;


/**
 * Type definitions.
 */

/**
 * TYPENAME: ix_cc_v6v4_reasm_packet_info
 * 
 * DESCRIPTION: This struct contains information about a packet being
 * reassembled. 
 *
 */

typedef struct ix_s_cc_v6v4_reasm_packet_info
{
	ix_uint32 srcAddr;
    ix_uint32 dstAddr;
    ix_uint16 idNum;
    ix_uint8 protocol;
    ix_uint16 bytesReceived;
} ix_cc_v6v4_reasm_packet_info;


/**
 * Prototypes for interface functions.
 */


/**
 * NAME: _ix_cc_v6v4_reasm_init
 *
 * DESCRIPTION: This primitive is used to initialize the IPv4 packet
 * reassembly resources.
 *           
 * @Param:  - IN arg_ReasmTimeout - number of ticks after a fragment
 * arrives to wait for the next fragment before releasing the reassembly
 * data for a packet.
 *
 * @Param:  - IN arg_hFreeList - handle to buffer free list.
 *
 * @Param:  - OUT arg_phReasm - pointer to location in which to return the
 * reassembly context.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_OOR - out of resources 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */

ix_error _ix_cc_v6v4_reasm_init(
    ix_uint32 arg_ReasmTimeout,
    ix_buffer_free_list_handle arg_hFreeList,
    ix_cc_v6v4_reassembly* arg_phReasm
    );


/**
 * NAME: _ix_cc_v6v4_reasm_cleanup
 *
 * DESCRIPTION: This primitive is used to release resources previously
 * allocated for IPv4 packet reassembly.
 *           
 * @Param:  - IN arg_hReasm - reassembly context handle returned from
 * _ix_cc_v6v4_reasm_init.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */

ix_error _ix_cc_v6v4_reasm_cleanup(
    ix_cc_v6v4_reassembly arg_hReasm
    );


/**
 * NAME: _ix_cc_v6v4_reasm_add_fragment
 *
 * DESCRIPTION: This primitive is used to add a new fragment to its reassembly
 * buffer.  If this is the first fragment received for the packet, new
 * reassembly resources are allocated.
 *           
 * @Param:  - IN arg_hReasm - reassembly context handle returned from
 * _ix_cc_v6v4_reasm_init.
 *
 * @Param:  - IN arg_hFragment - handle to buffer containing the fragment.
 *
 * @Param:  - OUT arg_phPacket - If the return value from this function is
 * IX_SUCCESS then this output parameter will be set to the  buffer handle
 * of the reassembled packet if the reassembly is complete, or to
 * IX_NULL_BUFFER_HANDLE if more fragments are needed before the packet
 * can be processed further.  If the return value from this function is not
 * IX_SUCCESS then this parameter has no meaning.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_OOR - out of resources
 * IX_CC_ERROR_OOM - out of memory
 * IX_CC_ERROR_INTERNAL - internal error occurred
 * 
 */

ix_error _ix_cc_v6v4_reasm_add_fragment(
    ix_cc_v6v4_reassembly arg_hReasm,
    ix_buffer_handle arg_hFragment,
    ix_buffer_handle* arg_phPacket
    );


/**
 * NAME: _ix_cc_v6v4_reassembly_tick
 *
 * DESCRIPTION: This primitive is called periodically  to check for packets
 * that never received all fragments in the designated period of time.
 *           
 * @Param:  - IN arg_hReasm - reassembly context handle returned from
 * _ix_cc_v6v4_reasm_init.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */

ix_error _ix_cc_v6v4_reassembly_tick(
    ix_cc_v6v4_reassembly arg_hReasm
    );


/**
 * NAME: _ix_cc_v6v4_reassembly_num_packets
 *
 * DESCRIPTION: This routine exists primarily for testing.  It returns
 * a the number of packets currently being reassembled.
 *           
 * @Param:  - IN arg_hReasm - reassembly context handle returned from
 * _ix_cc_v6v4_reasm_init.
 *
 * @Param:  - OUT arg_pNumPackets - pointer to variable to receive number
 * of packets being reassembled.
 *
 * @Return: IX_SUCCESS.
 *
 */

ix_error _ix_cc_v6v4_reassembly_num_packets(
    ix_cc_v6v4_reassembly arg_hReasm,
    ix_uint32* arg_pNumPackets
    );

/**
 * NAME: _ix_cc_v6v4_reassembly_packet_info
 *
 * DESCRIPTION: This routine exists primarily for testing.  It returns
 * information about a packet currently being reassembled.
 *           
 * @Param:  - IN arg_hReasm - reassembly context handle returned from
 * _ix_cc_v6v4_reasm_init.
 *
 * @Param:  - IN arg_Index - zero-relative index number.  Note that if
 * packets are removed while iterating through the list, info about the
 * same packet might be returned twice.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - packet not found 
 * 
 */

ix_error _ix_cc_v6v4_reassembly_packet_info(
    ix_cc_v6v4_reassembly arg_hReasm,
    ix_uint32 arg_Index,
    ix_cc_v6v4_reasm_packet_info* arg_pInfo
    );




#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__IX_CC_V6V4_REASSEMBLY_H__) */
