/* ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2400 Network Processor, Release 6
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      Core Component Infrastructure - Header Validation
 *
 * = FILENAME
 *      ix_cc_v6v4_validate.h
 *
 * = DESCRIPTION
 *      This file defines all data types, macros and function prototypes
 *      that support header validation for tunneling.
 *
 * = CREATION TIME
 *      12/09/2002 2:44 PM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 * $Id: ix_cc_v6v4_validate.h,v 1.3 2003/01/30 18:33:31 cabell Exp $
 */

#if !defined(__IX_CC_V6V4_VALIDATE_H__)
#define __IX_CC_V6V4_VALIDATE_H__

#define IX_CC_V6V4_VERIFY_IPV6_SOURCE
#define IX_CC_V6V4_VERIFY_IPV4_SOURCE

#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_header_defs.h"

#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */

/**
 * NAME: IX_CC_V6V4_IPV4_COMPATIBLE
 *
 * DESCRIPTION: This macro checks an address to determine if it is IPv4-
 * compatible.
 *           .
 * @Param:  - IN __pAdr - pointer to address in packet buffer.
 *
 */

#define IX_CC_V6V4_IPV4_COMPATIBLE(__pAdr)                                  \
    ((IX_RM_MEM_UINT32_READ(&(__pAdr).aUint32[0]) == 0) &&                   \
     (IX_RM_MEM_UINT32_READ(&(__pAdr).aUint32[1]) == 0) &&                   \
     (IX_RM_MEM_UINT32_READ(&(__pAdr).aUint32[2]) == 0))


/**
 * NAME: _ix_cc_v6v4_validate_ipv4_addr
 *
 * DESCRIPTION: This primitive to validate that an IPv4 address is suitable
 * for use as a source address.
 *           .
 * 
 * @Param:  - IN arg_ipAddr - address to be validated 
 *
 * @Param:  - IN arg_subnetBroadcast - subnet broadcast address to be checked
 * against in addition to the known invalid addresses. 
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_INVALID_ADDRESS  - address is invalid
 * 
 */


ix_error _ix_cc_v6v4_validate_ipv4_addr(
    ix_uint32 arg_ipAddr,
    ix_uint32 arg_subnetBroadcast
    );
    

/**
 * NAME: _ix_cc_v6v4_validate_ipv6_source
 *
 * DESCRIPTION: This primitive is used to validate the IPv6 source address
 * in a packet being decapsulated.
 *           
 * @Param:  - IN arg_pIpv6Hdr - pointer to IPv6 header in packet buffer.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * IX_CC_V6V4_ERROR_INVALID_ADDRESS  - source address is invalid
 * 
 */

ix_error _ix_cc_v6v4_validate_ipv6_source(
    ix_cc_v6v4_ipv6_header* arg_pIpv6Hdr
    );


/**
 * NAME: _ix_cc_v6v4_validate_auto
 *
 * DESCRIPTION: This primitive is used to validate IPv4 packets tunneled
 * to a local interface using an IPv4-compatible address.
 *           
 * @Param:  - IN arg_pIpv4Hdr - pointer to IPv4 header in packet buffer.
 *
 * @Param:  - IN arg_pIpv6Hdr - pointer to IPv6 header in packet buffer.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * IX_CC_V6V4_ERROR_INVALID_ADDRESS  - source or destination address
 * is invalid
 * 
 */

ix_error _ix_cc_v6v4_validate_auto(
    ix_cc_v6v4_ipv4_header* arg_pIpv4Hdr,
    ix_cc_v6v4_ipv6_header* arg_pIpv6Hdr
    );


/**
 * NAME: _ix_cc_v6v4_validate_6to4
 *
 * DESCRIPTION: This primitive is used to check the IPv6 source and
 * destination addresses of a packet being encapsulated or decapsulated
 * for 6to4 addresses.  If found, the embedded address is validated
 * according to RFC 3056.
 *           
 * @Param:  - IN arg_pIpv6Hdr - pointer to IPv6 header in packet buffer.
 *
 * @Param:  - IN arg_subnetBroadcast - subnet broadcast address to be checked
 * against in addition to the known invalid addresses. 
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * IX_CC_V6V4_ERROR_INVALID_ADDRESS  - source or destination 6to4 address
 * is invalid
 * 
 */

ix_error _ix_cc_v6v4_validate_6to4(
    ix_cc_v6v4_ipv6_header* arg_pIpv6Hdr,
    ix_uint32 arg_subnetBroadcast
    );


#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__IX_CC_V6V4_VALIDATE_H__) */
