/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 6
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      IPv6/v4 Tunneling Core Component - Decapsulation
 *
 * = FILENAME
 *      ix_cc_v6v4_decap.c
 *
 * = DESCRIPTION
 *      The file defines functions for handling of IPv4 packets requiring
 *      decapsulation.
 *
 * = CREATION TIME
 *      12/31/2002 12:14 PM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */



#define IX_ERROR_FILE_IDENT "$Id: ix_cc_v6v4_decap.c,v 1.6 2003/05/27 17:59:32 cabell Exp $";


/**
 * System defined include files
 */


/**
 * User defined include files
 */
#include "ix_cc_error.h"

#include "ix_rm.h"
#include "ix_ossl.h"

#include "ix_cc.h"
#include "bindings.h"
#include "cc/ix_cc_v6v4.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_init.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_decap.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_config.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_validate.h"




/**
 * Preprocessor symbols and Macros used in this file
 */

#define IX_CC_V6V4_IPV6_TRCLS_MASK      0x0f


/**
 * NAME: IX_CC_V6V4_UPDATE_IPV6_TRFCLASS
 *
 * DESCRIPTION: This macro updates the traffic class in the IPv6 header.
 *           .
 * @Param:  - IN __pHdr - pointer to ipv6 header.
 *
 * @Param:  - IN __val - 8-bit value to set.
 *
 */

#define IX_CC_V6V4_UPDATE_IPV6_TRFCLASS(__pHdr, __val)                      \
{                                                                           \
    *(__pHdr) &= ~IX_CC_V6V4_IPV6_TRCLS_MASK;                               \
    *(__pHdr) |= ((__val) >> 4);                                            \
    *((__pHdr) + 1) &= IX_CC_V6V4_IPV6_TRCLS_MASK;                          \
    *((__pHdr) + 1) |= ((__val) << 4);                                      \
}

/**
 * Type definitions whose scope is limited to this file
 */


/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */


/** 
 * Extern function prototypes.
 */


extern ix_cc_v6v4_context* g_pV6v4Context;
extern ix_error _ix_cc_v6v4_forward(ix_buffer_handle arg_hBuffer);

/** 
 * Static function prototypes.
 */


/** 
 * Function definitions.
 */


/**
 * NAME: _ix_cc_v6v4_handle_ipv4_packet
 *
 * DESCRIPTION: This primitive is used to process IPv4 packets that
 * require decapsulation.
 *           
 * @Param:  - IN arg_hPacket - handle to buffer containing the packet.
 *
 * @Param:  - IN arg_pContext - pointer to tunneling core component
 * context buffer.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * 
 */

ix_error _ix_cc_v6v4_handle_ipv4_packet(
    ix_buffer_handle arg_hPacket,  
    void* arg_pContext
    )
{
	ix_uint8* pData;
	ix_cc_v6v4_ipv4_header ipv4Hdr;
    ix_cc_v6v4_ipv6_header ipv6Hdr;
    ix_hw_buffer_meta *pHwMeta;
    ix_uint16 nextHop;
    ix_cc_v6v4_end_tunnel tunnelInfo;
    ix_cc_v6v4_context* pContext;
    ix_error err;

    pContext = (ix_cc_v6v4_context*)arg_pContext;

    /* Get the packet meta-data and get next hop ID */

    IX_ERROR_CRT(
        ix_rm_buffer_get_meta(arg_hPacket, (void **)&pHwMeta),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    nextHop = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_NextHopID));

    /* Get the tunnel info based on next hop ID */

    ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

    _READ_END_TUNNEL(pContext->pMbEndTunnelBase, nextHop, tunnelInfo);

    ix_ossl_mutex_unlock(pContext->v6v4Mutex);

    if ((tunnelInfo.flags & IX_HTON16(IX_CC_V6V4_FLAG_TUNNEL_VALID)) == 0)
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_TUNNEL_ENTRY,
                            ("Valid flag not set on tunnel entry"));
    }

    /* Get packet data */

    IX_ERROR_CRT(
        ix_rm_buffer_get_data(arg_hPacket, (void **)&pData),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    pData += IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_Offset));
	ix_ossl_memcpy(&ipv4Hdr, pData, IX_CC_V6V4_MIN_IPV4_LEN);

    pData += IX_CC_V6V4_MIN_IPV4_LEN;
    ix_ossl_memcpy(&ipv6Hdr, pData, IX_CC_V6V4_MIN_IPV6_LEN);

    /* Check for automatic tunneled packet */

    if (IX_CC_V6V4_IPV4_COMPATIBLE(ipv6Hdr.destAddress))
    {
        if ((tunnelInfo.flags & IX_HTON16(IX_CC_V6V4_DECAP_AUTO)) == 0)
        {
            return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_ADDRESS,
                                ("Automatic tunneling not enabled on interface"));
        }

        IX_ERROR_CR(_ix_cc_v6v4_validate_auto(&ipv4Hdr, &ipv6Hdr));

        IX_ERROR_CR(_ix_cc_v6v4_decapsulate(arg_hPacket, arg_pContext));

        IX_ERROR_CRT(
            ix_rm_packet_send(
                IX_CC_V6V4_PKT_LOW_PRIORITY_STKDRV_OUTPUT,
                arg_hPacket, 0),
            IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

        return IX_SUCCESS;
    }

    /* Validate IPv6 source address */

#if defined (IX_CC_V6V4_VERIFY_IPV6_SOURCE)
    IX_ERROR_CR(_ix_cc_v6v4_validate_ipv6_source(&ipv6Hdr));
#endif

    if ((tunnelInfo.flags & IX_HTON16(IX_CC_V6V4_DECAP_6TO4)) != 0)
    {
        IX_ERROR_CR(_ix_cc_v6v4_validate_6to4(&ipv6Hdr, 0));
    }

    /* Validate IPv4 source addresses */

    if ((tunnelInfo.flags & IX_HTON16(IX_CC_V6V4_SRC_VALIDATE)) != 0)
    {
        /* Validate using the ingress source validation list */

        ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

        err = (*pContext->ingressListHandlers.lookup)(
                &pContext->islInfo, IX_NTOH32(ipv4Hdr.sourceAddress), nextHop,
                tunnelInfo.ingressList);

        ix_ossl_mutex_unlock(pContext->v6v4Mutex);

        if (err != IX_SUCCESS)
        {
            if (IX_ERROR_GET_CODE(err) == IX_CC_ERROR_ENTRY_NOT_FOUND)
            {
                return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_ADDRESS,
                                    ("Source address not allowed"));
            }
            return err;
        }
    }
    else
    {
        /*
         * This tunnel does not use the ingress list, so just do the
         * basic validation
         */
#if defined (IX_CC_V6V4_VERIFY_IPV4_SOURCE)
        IX_ERROR_CR(_ix_cc_v6v4_validate_ipv4_addr(
			IX_NTOH32(ipv4Hdr.sourceAddress), 0));
#endif
    }

    pContext->ccStats.pktsDecapsulated++;

    if ((tunnelInfo.flags & IX_HTON16(IX_CC_V6V4_DECAP_TOS_V6)) != 0)
    {
        IX_CC_V6V4_UPDATE_IPV6_TRFCLASS(pData, ipv4Hdr.tos);
    }

    IX_ERROR_CR(_ix_cc_v6v4_decapsulate(arg_hPacket, arg_pContext));

    IX_ERROR_CRT(
        ix_rm_packet_send(IX_CC_V6V4_PKT_IPV6_OUTPUT, arg_hPacket, 0),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_handle_ipv4_packet */


/**
 * NAME: _ix_cc_v6v4_decapsulate
 *
 * DESCRIPTION: This primitive is used to adjust the meta data remove the
 * IPv4 header from a tunneled packet.
 *           
 * @Param:  - IN arg_hPacket - handle to buffer containing the packet.
 *
 * @Param:  - IN arg_pContext - pointer to tunneling core component
 * context buffer.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * 
 */

ix_error _ix_cc_v6v4_decapsulate(
    ix_buffer_handle arg_hPacket,  
    void* arg_pContext
    )
{
    ix_hw_buffer_meta *pHwMeta;
    ix_uint16 size;
    ix_uint16 offset;
    ix_uint16 bufInfo;
    ix_uint32 hdrType;

    /* Adjust the packet's offset, packet size and buffer size */

    IX_ERROR_CRT(
        ix_rm_buffer_get_meta(arg_hPacket, (void **)&pHwMeta),
        IX_CC_ERROR_INTERNAL,
        IX_ERROR_LEVEL_WARNING);

    size = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_PacketSize));
    size -= IX_CC_V6V4_MIN_IPV4_LEN;
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_PacketSize, IX_HTON16(size));

    size = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_BufferSize));
    size -= IX_CC_V6V4_MIN_IPV4_LEN;
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_BufferSize, IX_HTON16(size));

    offset = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_Offset));
    offset += IX_CC_V6V4_MIN_IPV4_LEN;
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_Offset, IX_HTON16(offset));

#if defined(IX_CC_V6V4_HDRTYPE_DECAP)
    bufInfo = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_BufferInfo));
    hdrType = IX_CC_V6V4_HDRTYPE_DECAP;
    IX_RM_HW_META_SET_HEADER_TYPE(bufInfo, hdrType);
#endif

    return IX_SUCCESS;
} /* end - _ix_cc_v6v4_decapsulate */


