/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 6
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      IPv6/v4 Tunneling Core Component - Encapsulation
 *
 * = FILENAME
 *      ix_cc_v6v4_encap.c
 *
 * = DESCRIPTION
 *      The file defines functions for handling of IPv6 packets requiring
 *      encapsulation.
 *
 * = CREATION TIME
 *      01/10/2003 08:28 AM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */



#define IX_ERROR_FILE_IDENT "$Id: ix_cc_v6v4_encap.c,v 1.6 2003/05/27 17:59:32 cabell Exp $";


/**
 * System defined include files
 */


/**
 * User defined include files
 */
#include "ix_cc_error.h"

#include "ix_rm.h"
#include "ix_ossl.h"

#include "ix_cc.h"
#include "bindings.h"
#include "cc/ix_cc_v6v4.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_init.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_encap.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_config.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_validate.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_icmp.h"




/**
 * Preprocessor symbols and Macros used in this file
 */

#define IX_CC_V6V4_IPV6_TRCLS           0
#define IX_CC_V6V4_IPV6_DST_ADDR        24
#define IX_CC_V6V4_6TO4_V4ADDR_OFFFSET  2
#define IX_CC_V6V4_AUTO_V4ADDR_OFFFSET  12

#define IX_CC_V6V4_IPV4_VERSION         4
#define IX_CC_V6V4_IPV4_IHL             (IX_CC_V6V4_MIN_IPV4_LEN >> 2)

#define IX_CC_V6V4_IPV4_VERS_IHL        \
        ((IX_CC_V6V4_IPV4_VERSION << 4) | IX_CC_V6V4_IPV4_IHL)

#define IX_CC_V6V4_IPV4_DF              0x4000
#define IX_CC_V6V4_MIN_IPV6_MTU         1300


/**
 * Type definitions whose scope is limited to this file
 */


/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */

PRIVATE ix_uint16 g_V6v4Id = 1;


/** 
 * Extern function prototypes.
 */


extern ix_cc_v6v4_context* g_pV6v4Context;
extern ix_error _ix_cc_v6v4_forward(ix_buffer_handle arg_hBuffer);

/** 
 * Static function prototypes.
 */


/** 
 * Function definitions.
 */


/**
 * NAME: _ix_cc_v6v4_handle_ipv6_packet
 *
 * DESCRIPTION: This primitive is used to process IPv6 packets that
 * require encapsulation.
 *           
 * @Param:  - IN arg_hPacket - handle to buffer containing the packet.
 *
 * @Param:  - IN arg_pContext - pointer to tunneling core component
 * context buffer.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * 
 */

ix_error _ix_cc_v6v4_handle_ipv6_packet(
    ix_buffer_handle arg_hPacket,  
    void* arg_pContext
    )
{
	ix_uint8* pData;
    ix_cc_v6v4_ipv4_header ipv4Hdr;
    ix_cc_v6v4_ipv6_header ipv6Hdr;
    ix_hw_buffer_meta *pHwMeta;
    ix_uint16 nextHop;
    ix_cc_v6v4_start_tunnel tunnelInfo;
    ix_cc_v6v4_context* pContext;
    ix_uint16 packetSize;
    ix_uint32 csum = 0;
    ix_uint16* pHdr;
    int i;
    ix_uint16 bufSize;
    ix_uint16 offset;
    ix_uint16 bufInfo;
    ix_uint32 hdrType;

    pContext = (ix_cc_v6v4_context*)arg_pContext;

    /* Get the packet meta-data and get next hop ID */

    IX_ERROR_CRT(
        ix_rm_buffer_get_meta(arg_hPacket, (void **)&pHwMeta),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    nextHop = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_NextHopID));

    /* Get the tunnel info based on next hop ID */

    ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

    _READ_START_TUNNEL(pContext->pMbStartTunnelBase, nextHop, tunnelInfo);

    ix_ossl_mutex_unlock(pContext->v6v4Mutex);

    if ((tunnelInfo.flags & IX_HTON16(IX_CC_V6V4_FLAG_TUNNEL_VALID)) == 0)
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_TUNNEL_ENTRY,
                            ("Valid flag not set on tunnel entry"));
    }

    /* Check packet size against path MTU */

    packetSize = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_PacketSize));
    packetSize += IX_CC_V6V4_MIN_IPV4_LEN;

    if ((packetSize > IX_NTOH16(tunnelInfo.pathMtu)) &&
        (packetSize > IX_CC_V6V4_MIN_IPV6_MTU))
    {
        /* Send "Packet too Big" ICMP error */
        IX_ERROR_CR(
            _ix_cc_v6v4_pkt_too_big(arg_hPacket, arg_pContext));

        return IX_SUCCESS;
    }

    /* Get packet data */

    IX_ERROR_CRT(
        ix_rm_buffer_get_data(arg_hPacket, (void **)&pData),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    pData += IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_Offset));

    ix_ossl_memcpy(&ipv6Hdr, pData, IX_CC_V6V4_MIN_IPV6_LEN);

    /* Validate 6to4 addresses if necessary */
    if ((tunnelInfo.flags & IX_HTON16(IX_CC_V6V4_ENCAP_6TO4)) != 0)
    {
        IX_ERROR_CR(_ix_cc_v6v4_validate_6to4(
            &ipv6Hdr, IX_NTOH32(tunnelInfo.subnetBroadcast)));
    }

    pContext->ccStats.pktsEncapsulated++;

    /* Construct the IPv4 header */

    ipv4Hdr.vers_hdrLength = IX_CC_V6V4_IPV4_VERS_IHL;

    /**
     * Get the appropriate value for the TOS byte.  If the flag
     * IX_CC_V6V4_ENCAP_TOS_V6 is set in the tunnel next hop then
     * we set tos to the value of the IPv6 traffic class.  If not,
     * then tos is set to the value specified in the tunnel next hop.
     */
    if ((tunnelInfo.flags & IX_HTON16(IX_CC_V6V4_ENCAP_TOS_V6)) != 0)
    {
        ipv4Hdr.tos = *((ix_uint8*)&ipv6Hdr + IX_CC_V6V4_IPV6_TRCLS) << 4;
        ipv4Hdr.tos |= *((ix_uint8*)&ipv6Hdr + IX_CC_V6V4_IPV6_TRCLS + 1) >> 4;
    }
    else
    {
        ipv4Hdr.tos = tunnelInfo.tos;
    }

    ipv4Hdr.totalLength = IX_HTON16(packetSize);
    ipv4Hdr.id = IX_HTON16(g_V6v4Id);

    /**
     * The ID number is incremented by 2 to keep the numbers assigned
     * by the cc separate from those assigned by the microblock.
     */
    g_V6v4Id += 2;

    /**
     * Determine whether to set the DONT_FRAGMENT flag -
     * if the IX_CC_V6V4_ENCAP_NO_DF is not set and the tunnel's path
     * mtu is larger than the minimum tunnel MTU then we can set the
     * flag.
     */

    if (((tunnelInfo.flags & IX_HTON16(IX_CC_V6V4_ENCAP_NO_DF)) == 0) &&
        (IX_NTOH16(tunnelInfo.pathMtu) > IX_CC_V6V4_MIN_IPV6_MTU))
    {
        ipv4Hdr.flags_fragOffset = IX_HTON16(IX_CC_V6V4_IPV4_DF);
    }
    else
    {
        ipv4Hdr.flags_fragOffset = 0;
    }


    ipv4Hdr.ttl = tunnelInfo.ttl;
    ipv4Hdr.protocol = IX_CC_V6V4_IP_PROTO_IPV6;
    ipv4Hdr.checksum = 0;
    ipv4Hdr.sourceAddress = tunnelInfo.localAddress;

    /**
     * The destination address is obtained either from the tunnel next
     * hop info (if the IX_CC_V6V4_ENCAP_AUTO flag is not set) or from
     * the IPv6 destination address if IX_CC_V6V4_ENCAP_AUTO is set.
     */

    if ((tunnelInfo.flags & IX_HTON16(IX_CC_V6V4_ENCAP_AUTO)) == 0)
    {
        ipv4Hdr.destAddress = tunnelInfo.remoteAddress;
    }
    else if ((tunnelInfo.flags & IX_HTON16(IX_CC_V6V4_ENCAP_6TO4)) != 0)
    {
        /**
         * 6to4 tunnel - get address from the 32 bits following the
         * 16-bit 2002:: prefix in the IPv6 destination address.
         */

        *((ix_uint16*)&ipv4Hdr.destAddress) = ipv6Hdr.destAddress.aUint16[1];
        *((ix_uint16*)&ipv4Hdr.destAddress + 1) =
            ipv6Hdr.destAddress.aUint16[2];
    }
    else
    {
        /**
         * Automatic tunnel - get address from the last bits of the
         * IPv6 destination address.
         */

        ipv4Hdr.destAddress = ipv6Hdr.destAddress.aUint32[3];

        /**
         * Validate the embedded address - we skip this step for
         * the 6to4 tunnels because it is already done in the previous
         * 6to4 address validation.
         */
        IX_ERROR_CR(_ix_cc_v6v4_validate_ipv4_addr(
            IX_NTOH32(ipv4Hdr.destAddress),
            IX_NTOH32(tunnelInfo.subnetBroadcast)));
    }

    /* Calculate the checksum */

    pHdr = (ix_uint16*)&ipv4Hdr;
    for (i = 0; i < (IX_CC_V6V4_MIN_IPV4_LEN >> 1); i++)
    {
        csum += IX_NTOH16(*pHdr);
        pHdr++;
    }

    csum = ~((csum & 0xffff) + (csum >> 16));
    ipv4Hdr.checksum = IX_HTON16((ix_uint16)csum);

    /* Append the header to the packet */

    ix_ossl_memcpy(
        pData - IX_CC_V6V4_MIN_IPV4_LEN,
        &ipv4Hdr, IX_CC_V6V4_MIN_IPV4_LEN);

    /* Update the meta data */

    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_PacketSize, IX_HTON16(packetSize));

    bufSize = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_BufferSize));
    bufSize += IX_CC_V6V4_MIN_IPV4_LEN;
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_BufferSize, IX_HTON16(bufSize));

    offset = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_Offset));
    offset -= IX_CC_V6V4_MIN_IPV4_LEN;
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_Offset, IX_HTON16(offset));

#if defined(IX_CC_V6V4_HDRTYPE_ENCAP)
    bufInfo = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_BufferInfo));
    hdrType = IX_RM_HW_META_GET_HEADER_TYPE(bufInfo);
    hdrType |= IX_CC_V6V4_HDRTYPE_ENCAP;
    IX_RM_HW_META_SET_HEADER_TYPE(bufInfo, hdrType);
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_BufferInfo, IX_HTON16(bufInfo));
#endif

    IX_ERROR_CRT(
        ix_rm_packet_send(IX_CC_V6V4_PKT_IPV4_OUTPUT, arg_hPacket, 0),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_handle_ipv6_packet */



