/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 6
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      IPv6/v4 Tunneling Core Component - ICMP error message detection
 *
 * = FILENAME
 *      ix_cc_v6v4_icmp.c
 *
 * = DESCRIPTION
 *      The file defines functions for detection and handling of ICMP
 *      error messages received over tunnels.
 *
 * = CREATION TIME
 *      12/30/2002 2:20 PM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */



#define IX_ERROR_FILE_IDENT "$Id: ix_cc_v6v4_icmp.c,v 1.4 2003/01/28 17:05:50 cabell Exp $";


/**
 * System defined include files
 */


/**
 * User defined include files
 */
#include "ix_cc_error.h"

#include "ix_rm.h"
#include "ix_ossl.h"

#include "ix_cc.h"
#include "bindings.h"
#include "cc/ix_cc_v6v4.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_init.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_icmp.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_header_defs.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_config.h"


/**
 * Preprocessor symbols and Macros used in this file
 */


/**
 * Type definitions whose scope is limited to this file
 */


/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */


/** 
 * Extern function prototypes.
 */


extern ix_cc_v6v4_context* g_pV6v4Context;
extern ix_error _ix_cc_v6v4_forward(ix_buffer_handle arg_hBuffer);

/** 
 * Static function prototypes.
 */

PRIVATE ix_error _ix_cc_v6v4_icmpv6_cb(
    ix_error arg_Result,
    void* arg_pContext);



/** 
 * Function definitions.
 */


/**
 * NAME: _ix_cc_v6v4_icmp_check
 *
 * DESCRIPTION: This primitive is used to check a packet to determine if it is
 * an ICMP error message received for tunneled packet.
 *           
 * @Param:  - IN arg_hBuffer - handle to buffer containing the packet.
 *
 * @Param:  - IN arg_pContext - pointer to tunneling core component
 * context buffer.
 *
 * @Param:  - OUT arg_pSend - pointer to location in which to return an
 * indicator of whether the packet is to be sent up to the stack driver or
 * whether it was handled internally.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_OOR - out of resources 
 * 
 */

ix_error _ix_cc_v6v4_icmp_check(
    ix_buffer_handle arg_hPacket,  
    void* arg_pContext,
    ix_uint8* arg_pSend
    )
{
    void *pData;
    ix_uint8 proto;
    ix_uint8 type;
    ix_uint8 code;
    ix_cc_v6v4_ipv4_header* pIpv4Hdr;
    ix_hw_buffer_meta *pHwMeta;
    ix_uint16 length;
    ix_uint16 offset;
    ix_cc_v6v4_icmp_header* pIcmpHdr;
    ix_cc_ipv6_icmp_err_type icmpv6Type;
    ix_cc_ipv6_icmp_err_code icmpv6Code;

    *arg_pSend = 1;

    /* Get the packet data, which should point to the IPv4 header */

    IX_ERROR_CRT(
        ix_rm_buffer_get_meta(arg_hPacket, (void **)&pHwMeta),
        IX_CC_ERROR_INTERNAL,
        IX_ERROR_LEVEL_WARNING);

    IX_ERROR_CRT(
        ix_rm_buffer_get_data(arg_hPacket, &pData),
        IX_CC_ERROR_INTERNAL,
        IX_ERROR_LEVEL_WARNING);

    pData += IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_Offset));

    pIpv4Hdr = (ix_cc_v6v4_ipv4_header*)pData;

    proto = IX_RM_MEM_UINT8_READ(&pIpv4Hdr->protocol);

    if (proto != IX_CC_V6V4_IP_PROTO_ICMP)
    {
        return IX_SUCCESS;
    }

    /**
     * This is an ICMP packet, and we need to determine if it is one
     * that should be mapped to an ICMPv6 packet.  If so, the original
     * packet must have been a tunneled packet, and the total length of
     * the ICMP packet must be large enough to include the IPv6 header.
     * Furthermore, the ICMP type must be one for which the mapping might
     * have at least some usefulness to the original IPv6 sender.
     */

    /* Get length and compare to minimum required */
    length = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_PacketSize));

    if ((length - IX_CC_V6V4_MIN_IPV4_LEN - IX_CC_V6V4_ICMP_HDR_LEN) <
        (IX_CC_V6V4_MIN_IPV4_LEN + IX_CC_V6V4_MIN_IPV6_LEN))
    {
        return IX_SUCCESS;
    }

    /* Get protocol field of original packet */
    pIpv4Hdr = (ix_cc_v6v4_ipv4_header*)((ix_uint32)pData +
                     IX_CC_V6V4_MIN_IPV4_LEN + IX_CC_V6V4_ICMP_HDR_LEN);
    proto = IX_RM_MEM_UINT8_READ(&pIpv4Hdr->protocol);

    if (proto != IX_CC_V6V4_IP_PROTO_IPV6)
    {
        return IX_SUCCESS;
    }

    /* Get ICMP header and check the type */
    pIcmpHdr = (ix_cc_v6v4_icmp_header*)(
                (ix_uint32)pData + IX_CC_V6V4_MIN_IPV4_LEN);
    type = IX_RM_MEM_UINT8_READ(&pIcmpHdr->type);
    
    if ((type != IX_CC_V6V4_ICMP_DEST_UNREACH) &&
        (type != IX_CC_V6V4_ICMP_TIME_EXCEEDED))
    {
        if (type == IX_CC_V6V4_ICMP_PARAM_PROBLEM)
        {
            /* *
             * Parameter problem with encapsulating header - don't
             * really need to send to stack, but can't effectively
             * translate to ICMPv6 either.
             */
            ix_rm_buffer_free_chain(arg_hPacket);
            *arg_pSend = 0;
        }
        return IX_SUCCESS;
    }

    /* Get the ICMP code and map to ICMPv6 */

    code = IX_RM_MEM_UINT8_READ(&pIcmpHdr->code);

    if (type == IX_CC_V6V4_ICMP_DEST_UNREACH)
    {
        icmpv6Type = IX_CC_IPV6_ICMP_ERR_DEST_UNREACHABLE;
        switch (code)
        {
            case IX_CC_V6V4_ICMP_DU_NETWORK:
                icmpv6Code = IX_CC_IPV6_ICMP_CODE_DU_NO_ROUTE;
                break;

            case IX_CC_V6V4_ICMP_DU_HOST:
            case IX_CC_V6V4_ICMP_DU_PROTOCOL:
                icmpv6Code = IX_CC_IPV6_ICMP_CODE_DU_ADDR_UNREACHABLE;
                break;

            case IX_CC_V6V4_ICMP_DU_PORT:
                icmpv6Code = IX_CC_IPV6_ICMP_CODE_DU_PORT_UNREACHABLE;
                break;

            case IX_CC_V6V4_ICMP_DU_COMM_PROHIB:
                icmpv6Code = IX_CC_IPV6_ICMP_CODE_DU_COMM_PROHIBITED;
                break;

            default:
                return IX_SUCCESS;
        }
    }
    else
    {
        icmpv6Type = IX_CC_IPV6_ICMP_ERR_TIME_EXCEEDED;
        switch (code)
        {
            case IX_CC_V6V4_ICMP_TE_TTL:
                icmpv6Code = IX_CC_IPV6_ICMP_CODE_TE_HOP_LIMIT_EXCEEDED;
                break;

            case IX_CC_V6V4_ICMP_TE_FRAG_TIMER:
                icmpv6Code = IX_CC_IPV6_ICMP_CODE_TE_REASSEMBLY_EXCEEDED;
                break;

            default:
                return IX_SUCCESS;
        }
    }

    *arg_pSend = 0;

    /* Update the meta-data to point to the IPv6 packet */

    length -= (IX_CC_V6V4_MIN_IPV4_LEN + IX_CC_V6V4_ICMP_HDR_LEN +
               IX_CC_V6V4_MIN_IPV4_LEN);
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_PacketSize, IX_HTON16(length));

    length = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_BufferSize));
    length -= (IX_CC_V6V4_MIN_IPV4_LEN + IX_CC_V6V4_ICMP_HDR_LEN +
             IX_CC_V6V4_MIN_IPV4_LEN);
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_BufferSize, IX_HTON16(length));

    offset = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_Offset));
    offset += (IX_CC_V6V4_MIN_IPV4_LEN + IX_CC_V6V4_ICMP_HDR_LEN +
               IX_CC_V6V4_MIN_IPV4_LEN);
    IX_RM_MEM_UINT16_WRITE(&pHwMeta->m_Offset, IX_HTON16(offset));

    /* Send the packet to the IPv6 forwarder */

    IX_ERROR_CRT(
        ix_cc_ipv6_async_send_icmp_error(
            icmpv6Type, icmpv6Code, arg_hPacket, 0,
            _ix_cc_v6v4_icmpv6_cb, arg_pContext),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

	return IX_SUCCESS;
}

/**
 * NAME: _ix_cc_v6v4_pkt_too_big
 *
 * DESCRIPTION: This primitive is used to send an ICMPv6 Packet too Big
 * message.
 *           
 * @Param:  - IN arg_hPacket - handle to buffer containing the IPv6 packet.
 *
 * @Param:  - IN arg_pContext - pointer to tunneling core component
 * context buffer.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * 
 */

ix_error _ix_cc_v6v4_pkt_too_big(
    ix_buffer_handle arg_hPacket,  
    void* arg_pContext
    )
{
    ix_hw_buffer_meta *pHwMeta;
    ix_uint16 mtu = 0;
    ix_uint16 nextHop;
    ix_cc_v6v4_context* pContext;

    pContext = (ix_cc_v6v4_context*)arg_pContext;

    IX_ERROR_CRT(
        ix_rm_buffer_get_meta(arg_hPacket, (void **)&pHwMeta),
        IX_CC_ERROR_INTERNAL,
        IX_ERROR_LEVEL_WARNING);

    /**
     * The meta-data next hop ID indicates the tunnel over which the packet
     * was to be sent.
     */

    nextHop = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_NextHopID));

    mtu = _READ_ENCAP_MTU(pContext->pMbStartTunnelBase, nextHop);        

    IX_ERROR_CRT(
        ix_cc_ipv6_async_send_icmp_error(
            IX_CC_IPV6_ICMP_ERR_PKT_TOO_BIG,
            IX_CC_IPV6_ICMP_CODE_PKT_TOO_BIG,
            arg_hPacket,
            mtu,
            _ix_cc_v6v4_icmpv6_cb,
            arg_pContext),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    return IX_SUCCESS;
}


/**
 * NAME: _ix_cc_v6v4_icmpv6_cb
 *
 * DESCRIPTION: Callback for messages to IPv6 forwarder for sending ICMPv6
 * error messages.
 *           
 * @Param:  - IN arg_Result - result code.
 *
 * @Param:  - IN arg_pContext - pointer to tunneling core component
 * context buffer.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_icmpv6_cb(
        ix_error arg_Result,
        void* arg_pContext)
{
    return IX_SUCCESS;
}
