/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      IPv6/v4 Tunneling Core Component - Ingress Source List Management
 *      (Default format)
 *
 * = FILENAME
 *      ix_cc_v6v4_ingress_list_def.c
 *
 * = DESCRIPTION
 *      The file contains the functions and macros needed to manage the
 *      default format ingress source validation list for tunnel
 *      decapsulation.
 *
 * = CREATION TIME
 *      11/25/2002 2:15 PM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */

#define IX_ERROR_FILE_IDENT "$Id: ix_cc_v6v4_ingress_list_def.c,v 1.7 2003/10/31 17:27:14 ktseng Exp $";



/**
 * System defined include files
 */

/**
 * User defined include files
 */
#include "ix_cc_error.h"

#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_netmacros.h"

#include "ix_cc.h"
#include "ix_cc_macros.h"
#include "ix_config.h"
#include "bindings.h"
#include "cc/ix_cc_v6v4.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_init.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_fini.h" 
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_config.h" 
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_ingress_list_def.h" 


/**
 * Preprocessor symbols and Macros used in this file
 */

#define IX_CC_V6V4_ISL_MAX_ADDRS        12

/*
 * NAME: IX_CC_V6V4_ISL_GROUP_INDEX
 *
 * DESCRIPTION: Get the index of the group number where the indicated
 * address resides in the ISL block.
 *
 * @Param:      - IN    __adr
 *                      address number (0 - 11).
 */
#define IX_CC_V6V4_ISL_GROUP_INDEX(__adr)   ((__adr) / 4)

/*
 * NAME: IX_CC_V6V4_ISL_ENTRY_INDEX
 *
 * DESCRIPTION: Get the index of the address and corresponding mask info
 * based on the address number withing the ISL block.
 *
 * @Param:      - IN    __adr
 *                      address number (0 - 11).
 */
#define IX_CC_V6V4_ISL_ENTRY_INDEX(__adr)   ((__adr) % 4)

/*
 * NAME: IX_CC_V6V4_ISL_GET_MAX_ADDRS()
 *
 * DESCRIPTION: Calculates the maximum number of addresses allowed in
 * an ingress list block based on the format specification.
 *
 * @Param:      - IN    __fmt
 *                      Format specification from registry.
 *
 */
#define IX_CC_V6V4_ISL_GET_MAX_ADDRS(__fmt)                                 \
    (((__fmt) == IX_CC_V6V4_ISL_DEFAULT_NO_MAX) ?                           \
        IX_CC_V6V4_ISL_MAX_ADDRS : (__fmt))



/*
 * NAME: IX_CC_V6V4_ISL_GET_MAX_DWORDS()
 *
 * DESCRIPTION: Calculates the number of 32-bit words that are significant
 *              in each ingress list block, based on the format specification.
 *
 * @Param:      - IN    __fmt
 *                      Format specification from registry.
 * @Param:      - OUT   __dw
 *                      Variable to receive result.
 *
 */
#define IX_CC_V6V4_ISL_GET_MAX_DWORDS(__fmt, __dw)                          \
    if ((__fmt) == IX_CC_V6V4_ISL_DEFAULT_NO_MAX)                           \
        (__dw) = sizeof(ix_cc_v6v4_ingress_list_block)/sizeof(ix_uint32);   \
    else if ((__fmt) < IX_CC_V6V4_ISL_DEFAULT_MAX_5)                        \
        (__dw) = (__fmt) + 2;                                               \
    else if ((__fmt) < IX_CC_V6V4_ISL_DEFAULT_MAX_9)                        \
        (__dw) = (__fmt) + 3;                                               \
    else (__dw) = (__fmt) + 4;



/*
 * NAME: IX_CC_V6V4_ISL_GET_BLOCK_ADDR()
 *
 * DESCRIPTION: Calculates the address of the indicated ingress list block
 * in shared memory.
 *
 * @Param:      - IN    __base
 *                      Base address of ingress list blocks.
 * @Param:      - IN    __ix
 *                      Index of the desired block.
 *
 */
#define IX_CC_V6V4_ISL_GET_BLOCK_ADDR(__base, __ix)                         \
    &(((ix_cc_v6v4_ingress_list_block*)(__base))[__ix])

/*
 * NAME: IX_CC_V6V4_ISL_SET_VALID
 *
 * DESCRIPTION: Set the validity flag for an address entry in a block.
 *
 * @Param:      - IN    __blk
 *                      Block in which to set the flag.
 * @Param:      - IN    __adr
 *                      Address number (0 - 11).
 *
 */
#define IX_CC_V6V4_ISL_SET_VALID(__blk, __adr) \
    (__blk).validityFlags |= IX_HTON16((ix_uint16)(1 << __adr))

/*
 * NAME: IX_CC_V6V4_ISL_CLEAR_VALID
 *
 * DESCRIPTION: Clear the validity flag for an address entry in a block.
 *
 * @Param:      - IN    __blk
 *                      Block in which to clear the flag.
 * @Param:      - IN    __adr
 *                      Address number (0 - 11).
 *
 */
#define IX_CC_V6V4_ISL_CLEAR_VALID(__blk, __adr) \
    (__blk).validityFlags &= ~IX_HTON16((ix_uint16)(1 << __adr))

/*
 * NAME: IX_CC_V6V4_ISL_CHECK_VALID
 *
 * DESCRIPTION: Check the validity flag for an address entry in a block.
 *
 * @Param:      - IN    __blk
 *                      Block in which to check the flag.
 * @Param:      - IN    __adr
 *                      Address number (0 - 11).
 *
 */
#define IX_CC_V6V4_ISL_CHECK_VALID(__blk, __adr) \
    (((__blk).validityFlags & IX_HTON16((ix_uint16)(1 << __adr))) != 0)

/*
 * NAME: IX_CC_V6V4_ISL_SET_ENTRY
 *
 * DESCRIPTION: Set the address and mask length for an address entry in a
 * local copy of an ISL block.
 *
 * @Param:      - IN    __blk
 *                      Block in which to set the address & mask length.
 * @Param:      - IN    __adr
 *                      Address number (0 - 11).
 * @Param:      - IN    __e
 *                      Entry pointer.
 *
 */
#define IX_CC_V6V4_ISL_SET_ENTRY(__blk, __adr, __e)                     \
{                                                                       \
    ix_uint8 __g = IX_CC_V6V4_ISL_GROUP_INDEX(__adr);                   \
    ix_uint8 __ei = IX_CC_V6V4_ISL_ENTRY_INDEX(__adr);                  \
    ix_uint8 __m = 32 - (__e)->maskLength;                              \
    (__blk).addrGroup[__g].maskInfo[__ei] = __m;                        \
    (__blk).addrGroup[__g].ipAddr[__ei] =                               \
        IX_HTON32((__e)->ipAddr & (0xffffffff << __m));                 \
}


/*
 * NAME: IX_CC_V6V4_ISL_COMP_ADDR
 *
 * DESCRIPTION: Compare the address and mask length for an address entry in a
 * local copy of a block to specified values.
 *
 * @Param:      - IN    __blk
 *                      Block in which to get the address & mask info.
 * @Param:      - IN    __adr
 *                      Address number (0 - 11).
 * @Param:      - IN    __e
 *                      Entry pointer.
 * @Param:      - IN    __res
 *                      Result of comparison.
 *
 */
#define IX_CC_V6V4_ISL_COMP_ADDR(__blk, __adr, __e, __res)              \
{                                                                       \
    ix_uint8 __g = IX_CC_V6V4_ISL_GROUP_INDEX(__adr);                   \
    ix_uint8 __ei = IX_CC_V6V4_ISL_ENTRY_INDEX(__adr);                  \
    ix_uint8 __m = 32 - (__e)->maskLength;                              \
    __res = (((__blk).addrGroup[__g].maskInfo[__ei] == __m) &&          \
             ((__blk).addrGroup[__g].ipAddr[__ei] ==                    \
              IX_HTON32((__e)->ipAddr & (0xffffffff << __m))));         \
}


/*
 * NAME: IX_CC_V6V4_ISL_GET_ENTRY
 *
 * DESCRIPTION: Get the address and mask length for an address entry from
 * a local copy of an ISL block.
 *
 * @Param:      - IN    __blk
 *                      Block in which to set the address & mask length.
 * @Param:      - IN    __adr
 *                      Address number (0 - 11).
 * @Param:      - IN    __e
 *                      Entry to receive the data.
 *
 */
#define IX_CC_V6V4_ISL_GET_ENTRY(__blk, __adr, __e)                     \
{                                                                       \
    ix_uint8 __g = IX_CC_V6V4_ISL_GROUP_INDEX(__adr);                   \
    ix_uint8 __ei = IX_CC_V6V4_ISL_ENTRY_INDEX(__adr);                  \
    (__e)->maskLength = 32 - (__blk).addrGroup[__g].maskInfo[__ei];     \
    (__e)->ipAddr = IX_NTOH32((__blk).addrGroup[__g].ipAddr[__ei]);     \
}


/*
 * NAME: IX_CC_V6V4_ISL_WRITE_ENTRY
 *
 * DESCRIPTION: Write the specified address and mask length into an
 * ingress list block.
 *
 * @Param:      - IN    __base
 *                      Base address of ingress list blocks.
 * @Param:      - IN    __ix
 *                      Block index.
 * @Param:      - IN    __adr
 *                      Address number (0 - 11).
 * @Param:      - IN    __e
 *                      Entry pointer.
 *
 */
#define IX_CC_V6V4_ISL_WRITE_ENTRY(__base, __ix, __adr, __e)                \
{                                                                           \
    ix_cc_v6v4_ingress_list_block* __pEntry =                               \
        ((ix_cc_v6v4_ingress_list_block*)(__base)) + (__ix);                \
    ix_uint8 __g = IX_CC_V6V4_ISL_GROUP_INDEX(__adr);                       \
    ix_uint8 __ei = IX_CC_V6V4_ISL_ENTRY_INDEX(__adr);                      \
    ix_uint8 __m = 32 - (__e)->maskLength;                                  \
    (void) IX_RM_MEM_UINT8_WRITE(                                           \
        &__pEntry->addrGroup[__g].maskInfo[__ei], __m);                     \
    (void) IX_RM_MEM_UINT32_WRITE(                                          \
        &__pEntry->addrGroup[__g].ipAddr[__ei],                             \
        IX_HTON32((__e)->ipAddr & (0xffffffff << __m)));                    \
}

/*
 * NAME: IX_CC_V6V4_ISL_WRITE_NEXT_BLOCK
 *
 * DESCRIPTION: Write a block index into the next block field of the
 * of the indicated block.
 *
 * @Param:      - IN    __base
 *                      Base address of ingress list blocks.
 * @Param:      - IN    __ix
 *                      Index of block to receive next block index.
 * @Param:      - IN    __nb
 *                      Index of next block.
 *
 */
#define IX_CC_V6V4_ISL_WRITE_NEXT_BLOCK(__base, __ix, __nb)                 \
{                                                                           \
    ix_cc_v6v4_ingress_list_block* __pEntry =                               \
        ((ix_cc_v6v4_ingress_list_block*)(__base)) + (__ix);                \
    (void) IX_RM_MEM_UINT16_WRITE(&__pEntry->nextBlock, __nb);              \
}

/*
 * NAME: IX_CC_V6V4_ISL_WRITE_FLAGS
 *
 * DESCRIPTION: Write the validity flags of the indicated block.
 *
 * @Param:      - IN    __base
 *                      Base address of ingress list blocks.
 * @Param:      - IN    __ix
 *                      Index of block to receive next block index.
 * @Param:      - IN    __f
 *                      Validity flags to be written.
 *
 */
#define IX_CC_V6V4_ISL_WRITE_FLAGS(__base, __ix, __f)                       \
{                                                                           \
    ix_cc_v6v4_ingress_list_block* __pEntry =                               \
        ((ix_cc_v6v4_ingress_list_block*)(__base)) + (__ix);                \
    (void) IX_RM_MEM_UINT16_WRITE(&__pEntry->validityFlags, __f);           \
}


/*
 * NAME: IX_CC_V6V4_ISL_READ_NEXT_BLOCK
 *
 * DESCRIPTION: Read the next block field of the indicated block.
 *
 * @Param:      - IN    __base
 *                      Base address of ingress list blocks.
 * @Param:      - IN    __ix
 *                      Index of block from which to read next block index.
 *
 */
#define IX_CC_V6V4_ISL_READ_NEXT_BLOCK(__base, __ix)                        \
    IX_RM_MEM_UINT16_READ(                                                  \
        (ix_uint16*)((ix_cc_v6v4_ingress_list_block*)(__base) + (__ix)))


/**
 * Type definitions whose scope is limited to this file
 */

/**
 * TYPENAME: ix_cc_v6v4_ingress_list_block
 * 
 * DESCRIPTION: This structure defines the format for an ingress source
 * validation list block .
 *
 */
typedef struct ix_s_cc_v6v4_ingress_list_block
{
    ix_uint16                   nextBlock;
    ix_uint16                   validityFlags;
    struct {
        ix_uint8 maskInfo[4];
        ix_uint32 ipAddr[4]; }  addrGroup[3];
} ix_cc_v6v4_ingress_list_block;


/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
extern ix_cc_v6v4_registry_data g_V6v4RegistryData;
#else
ix_cc_v6v4_registry_data g_V6v4RegistryData;
#endif

/** 
 * Extern function prototypes.
 */


/** 
 * Static function prototypes.
 */


/** 
 * Function definitions.
 */


/**
 * NAME: _ix_cc_v6v4_isl_def_write_block
 *
 * DESCRIPTION: This primitive is used to write a local copy of an
 * ingress list block to shared memory.
 *           
 * @Param:  - In arg_pBase - Base of ingress list blocks in shared memory.
 *
 * @Param:  - IN arg_Index - Ingress list block index.
 *
 * @Param:  - IN arg_pLocal - pointer to local copy of the block.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS.
 *
 * Error Codes:
 * 
 */
PRIVATE ix_error _ix_cc_v6v4_isl_def_write_block(
    ix_uint32* arg_pBase,
    ix_uint16 arg_Index,
    ix_cc_v6v4_ingress_list_block* arg_pLocal
    )
{
    ix_int32 dWords;
    ix_uint64* pShared;
    ix_uint64* pLocal;

    IX_CC_V6V4_ISL_GET_MAX_DWORDS(
            g_V6v4RegistryData.ingressListFormat, dWords);

    pShared = (ix_uint64*)
        IX_CC_V6V4_ISL_GET_BLOCK_ADDR(arg_pBase, arg_Index);
    pLocal = (ix_uint64*)arg_pLocal;

    while (dWords > 0)
    {
        (void) IX_RM_MEM_UINT64_WRITE(pShared, *pLocal);
        pShared++;
        pLocal++;
        dWords -= 2;
    }

    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_isl_def_write_block */


/**
 * NAME: _ix_cc_v6v4_isl_def_read_block
 *
 * DESCRIPTION: This primitive is used to read an ingress list block from
 * shared memory into a local buffer.
 *           
 * @Param:  - In arg_pBase - Base of ingress list blocks in shared memory.
 *
 * @Param:  - IN arg_Index - Ingress list block index.
 *
 * @Param:  - IN arg_pLocal - pointer to local buffer into which to read the
 * block.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS.
 *
 * Error Codes:
 * 
 */
PRIVATE ix_error _ix_cc_v6v4_isl_def_read_block(
    ix_uint32* arg_pBase,
    ix_uint16 arg_Index,
    ix_cc_v6v4_ingress_list_block* arg_pLocal
    )
{
    ix_int32 dWords;
    ix_uint64* pShared;
    ix_uint64* pLocal;

    IX_CC_V6V4_ISL_GET_MAX_DWORDS(
        g_V6v4RegistryData.ingressListFormat, dWords);

    pShared = (ix_uint64*)
        IX_CC_V6V4_ISL_GET_BLOCK_ADDR(arg_pBase, arg_Index);
    pLocal = (ix_uint64*)arg_pLocal;

    while (dWords > 0)
    {
        *pLocal = IX_RM_MEM_UINT64_READ(pShared);
        pShared++;
        pLocal++;
        dWords -= 2;
    }

    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_isl_def_read_block */


/**
 * NAME: _ix_cc_v6v4_isl_def_new_block
 *
 * DESCRIPTION: This primitive is used to allocate a new block and initialize
 * it with the first entry.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_pEntry - pointer to ingress list entry to be added to
 * the new block.
 *
 * @Param:  - OUT arg_pIndex - pointer to location to receive index of new
 * block.
 *
 * @Param:  - OUT None.
 *
 * @Return: 
 * IX_CC_ERROR_FULL - cannot add any more entries 
 *
 * Error Codes:
 * 
 */
PRIVATE ix_error _ix_cc_v6v4_isl_def_new_block(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_cc_v6v4_ingress_source_entry* arg_pEntry,
    ix_uint16* arg_pIndex
    )
{
    ix_uint32 index;
    ix_uint8 addrNum;
    ix_cc_v6v4_ingress_list_block islBlock;

   /* Allocate a new block */

    IX_ERROR_CR(
        ix_cc_index_pool_alloc(
            (ix_cc_index_pool)arg_pData->pIslContext, &index)
    );

    ix_ossl_memset(&islBlock, 0, sizeof(islBlock));

    addrNum = 0;
    IX_CC_V6V4_ISL_SET_ENTRY(islBlock, addrNum, arg_pEntry);
    IX_CC_V6V4_ISL_SET_VALID(islBlock, addrNum);

    _ix_cc_v6v4_isl_def_write_block(
            arg_pData->pMbIslBase, (ix_uint16)index, &islBlock);

    *arg_pIndex = (ix_uint16)index;
    return IX_SUCCESS;
} /* end - _ix_cc_v6v4_isl_def_new_block */


/**
 * NAME: _ix_cc_v6v4_isl_def_add_prefix
 *
 * DESCRIPTION: This primitive is used to add a prefix to the ingress list
 * for a tunnel.  The caller must have already validated the tunnel index
 * and the address for acceptability.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_TunnelIndex - end tunnel index (not used for this
 * ISL format).
 *
 * @Param:  - IN arg_pEntry - pointer to ingress list entry.
 *
 * @Param:  - IN/OUT arg_pTunnelInfo - on input, this points to the 16-bit
 * word in the tunnel configuration entry pertaining to the ingress
 * list.  On output, it contains the value that should be written to the
 * 16-bit word in the tunnel configuration entry.  The 16-bit word is a
 * value assigned by the ingress list handler and may change when a prefix
 * is added or deleted, and must be zero only if there are no prefixes
 * allowed for the tunnel.  The interpretation of non-zero values is format-
 * specific.  For the default format, it represents the block ID of the
 * first block of the ingress list.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_DUPLICATE_ENTRY - entry already exists 
 * IX_CC_ERROR_FULL - cannot add any more entries 
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_isl_def_add_prefix(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_uint16 arg_TunnelIndex,
    ix_cc_v6v4_ingress_source_entry* arg_pEntry,
    ix_uint16* arg_pTunnelInfo
    )
{
    ix_uint16 index = IX_NTOH16(*arg_pTunnelInfo);
    ix_cc_v6v4_ingress_list_block islBlock;
    ix_uint8 addrNum;
    ix_uint8 maxAddrs;
    ix_uint16 unusedIndex;
    ix_uint8 unusedAddrNum;
    ix_uint16 validityFlags;
    ix_uint8 result;
    ix_uint16 lastIndex;

    /* 
     * The tunnel info contains the index of the ingress list block
     * currently in use by the tunnel.
     */
    if (index == 0)
    {
        /* Allocate a new block and initialize it with the entry. */
        IX_ERROR_CR (
            _ix_cc_v6v4_isl_def_new_block(arg_pData, arg_pEntry, &index));
        *arg_pTunnelInfo = IX_HTON16(index);

        return IX_SUCCESS;
    }

    /*
     * The tunnel has an ingress list block.
     */

    maxAddrs = IX_CC_V6V4_ISL_GET_MAX_ADDRS(
                g_V6v4RegistryData.ingressListFormat);
    unusedIndex = 0;
    lastIndex = 0;

    while (index != 0)
    {
        _ix_cc_v6v4_isl_def_read_block(
            arg_pData->pMbIslBase, index, &islBlock);

        for (addrNum = 0; addrNum < maxAddrs; addrNum++)
        {
            if (IX_CC_V6V4_ISL_CHECK_VALID(islBlock, addrNum))
            {
                IX_CC_V6V4_ISL_COMP_ADDR(
                    islBlock, addrNum, arg_pEntry, result);

                if (result)
                {
                    /* Entry already exists */
                    return IX_CC_ERROR_DUPLICATE_ENTRY;
                }
            }
            else if (unusedIndex == 0)
            {
                /*
                 * Save this value as the first location in which to add
                 * a new entry.
                 */
                 unusedIndex = index;
                 unusedAddrNum = addrNum;
                 validityFlags = islBlock.validityFlags;
            }
        }

        lastIndex = index;
        index = IX_NTOH16(islBlock.nextBlock);
    }

    if (unusedIndex == 0)
    {
        /*
         * If the format does not limit the size of the list, allocate a
         * new block to hold the entry.
         */
        if (g_V6v4RegistryData.ingressListFormat ==
            IX_CC_V6V4_ISL_DEFAULT_NO_MAX)
        {
            IX_ERROR_CR(
               _ix_cc_v6v4_isl_def_new_block(arg_pData, arg_pEntry, &index)
            );

            /* Write index value into previous block */

            IX_CC_V6V4_ISL_WRITE_NEXT_BLOCK(
                arg_pData->pMbIslBase, lastIndex, IX_HTON16(index));

            return IX_SUCCESS;
        }
        else
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_FULL,
                                    ("Ingress list is full"));
        }
    }

    /*
     * Add the entry in the first unused location we found in the list
     */

    /* Write address and mask info */

    IX_CC_V6V4_ISL_WRITE_ENTRY(
        arg_pData->pMbIslBase, unusedIndex, unusedAddrNum, arg_pEntry);

    /* Now write the validity flags */

    validityFlags |= IX_HTON16((ix_uint16)(1 << unusedAddrNum));
    IX_CC_V6V4_ISL_WRITE_FLAGS(
        arg_pData->pMbIslBase, unusedIndex, validityFlags);

    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_isl_def_add_prefix */


/**
 * NAME: _ix_cc_v6v4_isl_def_del_prefix
 *
 * DESCRIPTION: This primitive is used to delete a prefix from the ingress
 * list for a tunnel.  The caller must have already validated the tunnel
 * index.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_TunnelIndex - end tunnel index (not used for this
 * ISL format).
 *
 * @Param:  - IN arg_pEntry - pointer to ingress list entry.
 *
 * @Param:  - IN/OUT arg_pTunnelInfo - on input, this points to the 16-bit
 * word in the tunnel configuration entry pertaining to the ingress
 * list.  On output, it contains the value that should be written to the
 * 16-bit word in the tunnel configuration entry.  The 16-bit word is a
 * value assigned by the ingress list handler and may change when a prefix
 * is added or deleted, and must be zero only if there are no prefixes
 * allowed for the tunnel.  The interpretation of non-zero values is format-
 * specific.  For the default format, it represents the block ID of the
 * first block of the ingress list.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - cannot find specified entry 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_isl_def_del_prefix(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_uint16 arg_TunnelIndex,
    ix_cc_v6v4_ingress_source_entry* arg_pEntry,
    ix_uint16* arg_pTunnelInfo
    )
{
    ix_uint16 index = IX_NTOH16(*arg_pTunnelInfo);
    ix_cc_v6v4_ingress_list_block islBlock;
    ix_uint8 addrNum;
    ix_uint8 maxAddrs;
    ix_uint8 result;
    ix_uint16 lastIndex;
    
    maxAddrs = IX_CC_V6V4_ISL_GET_MAX_ADDRS(
                g_V6v4RegistryData.ingressListFormat);
    lastIndex = 0;
    result = 0;

    while (index != 0)
    {
        _ix_cc_v6v4_isl_def_read_block(arg_pData->pMbIslBase, index, &islBlock);

        for (addrNum = 0; addrNum < maxAddrs; addrNum++)
        {
            if (IX_CC_V6V4_ISL_CHECK_VALID(islBlock, addrNum))
            {
                IX_CC_V6V4_ISL_COMP_ADDR(
                    islBlock, addrNum, arg_pEntry, result);
            }

            if (result)
            {
                /*
                 * Found the entry.
                 */
                IX_CC_V6V4_ISL_CLEAR_VALID(islBlock, addrNum);
                IX_CC_V6V4_ISL_WRITE_FLAGS(
                    arg_pData->pMbIslBase, index, islBlock.validityFlags);

                /* Check for empty block */

                if (islBlock.validityFlags == 0)
                {
                    if (lastIndex != 0)
                    {
                        /*
                         * Update the next block field in the previous
                         * block
                         */

                        IX_CC_V6V4_ISL_WRITE_NEXT_BLOCK(
                            arg_pData->pMbIslBase, lastIndex,
                            islBlock.nextBlock);
                    }
                    else
                    {
                        /* 
                         * Freeing the initial block, so return the next block
                         * to be written to the tunnel configuration.
                         */
                        *arg_pTunnelInfo = islBlock.nextBlock;
                    }

                    /* Free the block index */

                    IX_ERROR_CRT(
                        ix_cc_index_pool_free(
                            (ix_cc_index_pool)arg_pData->pIslContext, index),
                        IX_CC_ERROR_INTERNAL,
                        IX_ERROR_LEVEL_WARNING
                    );
                }

                return IX_SUCCESS;
            }
        }

        lastIndex = index;
        index = IX_NTOH16(islBlock.nextBlock);
    }

    return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                            ("Ingress list entry not found"));

} /* end - _ix_cc_v6v4_isl_def_del_prefix */



/**
 * NAME: _ix_cc_v6v4_isl_def_get_prefixes
 *
 * DESCRIPTION: This primitive is used to retrieve the prefix from the ingress
 * list for a tunnel.  The caller must have already validated the tunnel
 * index.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_TunnelIndex - end tunnel index (not used for this
 * ISL format).
 *
 * @Param:  - IN arg_TunnelInfo - the 16-bit word in the tunnel
 * configuration entry pertaining to the ingress list.
 *
 * @Param:  - IN/OUT arg_pNumEntries - On input, points to the number of
 * entries the buffer pointed to by arg_pEntry can accommodate.  On output,
 * contains the actual number of prefixes in the ingress list for the
 * tunnel.
 *
 * @Param:  - OUT arg_paEntries - pointer to array into which to place the
 * entries.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_BUFFER_TOO_SMALL - buffer too small to contain the complete
 *                                     list.
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_isl_def_get_prefixes(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_uint16 arg_TunnelIndex,
    ix_uint16 arg_TunnelInfo,
    ix_uint32* arg_pNumEntries,
    ix_cc_v6v4_ingress_source_entry* arg_paEntries
    )
{
    ix_uint16 index = IX_NTOH16(arg_TunnelInfo);
    ix_cc_v6v4_ingress_list_block islBlock;
    ix_uint8 addrNum;
    ix_uint8 maxAddrs;
    ix_uint16 listIndex;
    ix_cc_v6v4_ingress_source_entry* pEntry = arg_paEntries;
    ix_error err;
    
    maxAddrs = IX_CC_V6V4_ISL_GET_MAX_ADDRS(
                g_V6v4RegistryData.ingressListFormat);
    listIndex = 0;

    while (index != 0)
    {
        _ix_cc_v6v4_isl_def_read_block(arg_pData->pMbIslBase, index, &islBlock);

        for (addrNum = 0; addrNum < maxAddrs; addrNum++)
        {
            if (IX_CC_V6V4_ISL_CHECK_VALID(islBlock, addrNum))
            {
                if (listIndex < *arg_pNumEntries)
                {
                    IX_CC_V6V4_ISL_GET_ENTRY(islBlock, addrNum, pEntry);
                    pEntry++;
                }
                listIndex++;
            }
        }

        index = IX_NTOH16(islBlock.nextBlock);
    }

    err = (listIndex > *arg_pNumEntries) ?
        IX_ERROR_WARNING(IX_CC_V6V4_ERROR_BUFFER_TOO_SMALL, ("Buffer too small")) :
        IX_SUCCESS;

    *arg_pNumEntries = listIndex;
    return err;

} /* end - _ix_cc_v6v4_isl_def_get_prefixes */


/**
 * NAME: _ix_cc_v6v4_isl_def_del_tunnel
 *
 * DESCRIPTION: This primitive is used to delete all prefixes assigned to
 * a tunnel.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_TunnelIndex - end tunnel index (not used for this
 * ISL format).
 *
 * @Param:  - IN arg_pTunnelInfo - the 16-bit word in the tunnel configuration
 * entry pertaining to the ingress list.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_isl_def_del_tunnel(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_uint16 arg_TunnelIndex,
    ix_uint16 arg_pTunnelInfo
    )
{
    ix_uint16 index = IX_NTOH16(arg_pTunnelInfo);
    ix_uint16 nextBlock;
    
    while (index != 0)
    {
        nextBlock =
            IX_CC_V6V4_ISL_READ_NEXT_BLOCK(arg_pData->pMbIslBase, index);

        IX_CC_V6V4_ISL_WRITE_FLAGS(arg_pData->pMbIslBase, index, 0);

        IX_ERROR_CRT(
            ix_cc_index_pool_free(
                (ix_cc_index_pool)arg_pData->pIslContext, index),
            IX_CC_ERROR_INTERNAL,
            IX_ERROR_LEVEL_WARNING
        );

        index = IX_NTOH16(nextBlock);
    }

    return IX_SUCCESS;
} /* end - _ix_cc_v6v4_isl_def_del_tunnel */



/**
 * NAME: _ix_cc_v6v4_isl_def_lookup
 *
 * DESCRIPTION: This primitive is used to lookup a prefix in the ingress list
 * to check whether it is allowed for a particular tunnel.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_IpAddr - IP source address.
 *
 * @Param:  - IN arg_TunnelIndex - end tunnel index (not used for this
 * ISL format).
 *
 * @Param:  - IN arg_TunnelInfo - the 16-bit word in the tunnel configuration
 * entry pertaining to the ingress list.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - address is not allowed for the tunnel 
 * 
 */

ix_error _ix_cc_v6v4_isl_def_lookup(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_uint32 arg_IpAddr,
    ix_uint16 arg_TunnelIndex,
    ix_uint16 arg_TunnelInfo)
{
    ix_uint16 index = IX_NTOH16(arg_TunnelInfo);
    ix_cc_v6v4_ingress_list_block islBlock;
    ix_uint8 addrNum;
    ix_uint8 maxAddrs;
    ix_cc_v6v4_ingress_source_entry entry;
    
    maxAddrs = IX_CC_V6V4_ISL_GET_MAX_ADDRS(
                g_V6v4RegistryData.ingressListFormat);

    while (index != 0)
    {
        _ix_cc_v6v4_isl_def_read_block(arg_pData->pMbIslBase, index, &islBlock);

        for (addrNum = 0; addrNum < maxAddrs; addrNum++)
        {
            if (IX_CC_V6V4_ISL_CHECK_VALID(islBlock, addrNum))
            {
                IX_CC_V6V4_ISL_GET_ENTRY(islBlock, addrNum, &entry);
                if ((arg_IpAddr & (0xffffffff << (32 - entry.maskLength))) ==
                    entry.ipAddr)
                {
                    return IX_SUCCESS;
                }
            }
        }

        index = IX_NTOH16(islBlock.nextBlock);
    }

    return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND, ("Entry not found in ingress list"));

} /* end - _ix_cc_v6v4_isl_def_lookup */


/**
 * NAME: _ix_cc_v6v4_isl_def_setup
 *
 * DESCRIPTION: This primitive is used to setup the ingress source list
 * (default format). This includes allocating all shared memory and internal
 * memory used to manage the list.
 *           
 * @Param:  - IN None.
 *
 * @Param:  - OUT arg_pData - pointer to location in which to return the
 * return pointers to the allocated resources.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_OOR - out of resources 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_isl_def_setup(
    ix_cc_v6v4_ingress_list_data* arg_pData)
{
    ix_error err;
    ix_cc_index_pool indexPool;

    IX_ERROR_CR(
        _ix_cc_v6v4_alloc_shared_table(
                g_V6v4RegistryData.ingressListSizeHint,
                sizeof(ix_cc_v6v4_ingress_list_block),
                g_V6v4RegistryData.ingressListMemType,
                g_V6v4RegistryData.ingressListChannel,
                (void**)&arg_pData->pMbIslBase)
    );

    IX_ERROR_CG(
        ix_cc_index_pool_init(
            g_V6v4RegistryData.ingressListSizeHint, 1,
            &indexPool),
        err,
        cleanupMem);

    /*
     * All we need for managing the default format is the
     * index pool.
     */
    arg_pData->pIslContext = (void*)indexPool;

    return IX_SUCCESS;

cleanupMem:
    ix_rm_mem_free(arg_pData->pMbIslBase);
    return err;

}/* end _ix_cc_v6v4_isl_def_setup */


/**
 * NAME: _ix_cc_v6v4_isl_def_cleanup
 *
 * DESCRIPTION: This primitive is used to cleanup the ingress source list.
 * (default format).  The shared memory and internal memory allocated to
 * to manage the list is freed.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */

ix_error _ix_cc_v6v4_isl_def_cleanup(
    ix_cc_v6v4_ingress_list_data* arg_pData)
{
    /* Free shared memory */

    IX_ERROR_CRT(
        ix_rm_mem_free(arg_pData->pMbIslBase),
        IX_CC_ERROR_INTERNAL,
        IX_ERROR_LEVEL_LOCAL);

    IX_ERROR_CRT(
        ix_cc_index_pool_fini((ix_cc_index_pool)arg_pData->pIslContext),
        IX_CC_ERROR_INTERNAL,
        IX_ERROR_LEVEL_WARNING);

    return IX_SUCCESS;

}/* end _ix_cc_v6v4_isl_def_cleanup */




/**
 * NAME: _ix_cc_v6v4_ingress_list_init_def_handlers
 *
 * DESCRIPTION: This primitive is used to initialize the ingress list
 * handlers for the default format.
 *           
 * @Param:  - IN None.
 *
 * @Param:  - OUT arg_pHandlers - pointer to handler list to be
 * initialized.
 *
 * @Return: IX_SUCCESS.
 *
 * Error Codes:
 * 
 */

ix_error _ix_cc_v6v4_ingress_list_init_def_handlers(
    ix_cc_v6v4_ingress_list_handlers* arg_pHandlers)
{
    arg_pHandlers->setup = _ix_cc_v6v4_isl_def_setup;
    arg_pHandlers->cleanup = _ix_cc_v6v4_isl_def_cleanup;
    arg_pHandlers->addPrefix = _ix_cc_v6v4_isl_def_add_prefix;
    arg_pHandlers->delPrefix = _ix_cc_v6v4_isl_def_del_prefix;
    arg_pHandlers->getPrefixes = _ix_cc_v6v4_isl_def_get_prefixes;
    arg_pHandlers->delTunnel = _ix_cc_v6v4_isl_def_del_tunnel;
    arg_pHandlers->lookup = _ix_cc_v6v4_isl_def_lookup;

    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_ingress_list_init_def_handlers */

#if defined(IX_DEBUG)
extern ix_cc_v6v4_context* g_pV6v4Context;
ix_error ix_cc_v6v4_isl_dump_used_blocks(void)
{
    ix_cc_index_pool indexPool;
    ix_uint16 index;
    int inUse;
    ix_uint16 total = 0;

    indexPool = (ix_cc_index_pool)g_pV6v4Context->islInfo.pIslContext;

    for (index = 1; index <= g_V6v4RegistryData.ingressListSizeHint; index++)
    {
        IX_ERROR_CRT(
            ix_cc_index_pool_in_use(indexPool, index, &inUse),
            IX_CC_ERROR_INTERNAL,
            IX_ERROR_LEVEL_WARNING);

        if (inUse)
        {
            ix_ossl_message_log("Ingress list block %d in use\n", index);
            total++;
        }
    }

    ix_ossl_message_log("Total ingress list blocks in use = %d\n", total);

    return IX_SUCCESS;

}

#endif
