/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      IPv6/v4 Tunneling Core Component - Ingress Source List Management
 *      (Trie format)
 *
 * = FILENAME
 *      ix_cc_v6v4_ingress_list_trie.c
 *
 * = DESCRIPTION
 *      The file contains the functions and macros needed to manage the
 *      trie format ingress source validation list for tunnel decapsulation.
 *
 * = CREATION TIME
 *      12/02/2002 9:24 AM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_v6v4_ingress_list_trie.c,v 1.8 2003/10/31 17:27:05 ktseng Exp $";



/**
 * System defined include files
 */

/**
 * User defined include files
 */
#include "ix_cc_error.h"

#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_netmacros.h"

#include "ix_cc.h"
#include "ix_cc_macros.h"
#include "ix_config.h"
#include "bindings.h"
#include "cc/ix_cc_v6v4.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_init.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_fini.h" 
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_config.h" 
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_ingress_list_trie.h" 


/**
 * Preprocessor symbols and Macros used in this file
 */

#define IX_CC_V6V4_ISL_TRIE_BASE_ENTRIES    256
#define IX_CC_V6V4_ISL_TRIE_ENTRIES         16

#define IX_CC_V6V4_ISL_TRIE_BASE_SIZE                                       \
        (IX_CC_V6V4_ISL_TRIE_BASE_ENTRIES * sizeof(ix_uint32))
#define IX_CC_V6V4_ISL_TRIE_SIZE                                            \
        (IX_CC_V6V4_ISL_TRIE_ENTRIES * sizeof(ix_uint32))

#define IX_CC_V6V4_TRIE_INFO_HASH_SIZE      128
#define IX_CC_V6V4_ISL_MIN_TRIE_BLOCKS      8

#define IX_CC_V6V4_ISL_MAX_CHAIN            6
#define IX_CC_V6V4_MAX_BASE_EXPANSIONS      128
#define IX_CC_V6V4_MAX_EXPANSIONS           8
#define IX_CC_V6V4_MAX_INDEX_BLOCK_SIZE     64

#define IX_CC_V6V4_ISL_BASE_BITS            8
#define IX_CC_V6V4_ISL_TRIE_BITS            4

/*
 * NAME: IX_CC_V6V4_ISL_INDEX_BLOCK_SIZE
 *
 * DESCRIPTION: Get the size of an index block based on the registry
 * format value.
 *
 * @Param:      - IN    __r
 *                      format value selected in registry.
 */
#define IX_CC_V6V4_ISL_INDEX_BLOCK_SIZE(__r)                                \
    (8 << ((__r) - IX_CC_V6V4_ISL_TRIE_FIRST));                                                   \


/*
 * NAME: IX_CC_V6V4_ISL_TRIE_INFO_SIZE
 *
 * DESCRIPTION: Get the size of the variable portion of trie info block
 * based on the registry
 * format value.
 *
 * @Param:      - IN    __r
 *                      format value selected in registry.
 */
#define IX_CC_V6V4_ISL_TRIE_INFO_SIZE(__r)                                  \
    (4 << ((__r) - IX_CC_V6V4_ISL_TRIE_FIRST))                                                  \


/*
 * NAME: IX_CC_V6V4_ISL_TRIE_INFO_TOTAL_SIZE
 *
 * DESCRIPTION: Get the total size of an trie info block based on the
 * registry format value.
 *
 * @Param:      - IN    __r
 *                      format value selected in registry.
 */
#define IX_CC_V6V4_ISL_TRIE_INFO_TOTAL_SIZE(__r)                            \
    (IX_CC_V6V4_ISL_TRIE_INFO_SIZE(__r) +                                   \
     sizeof(ix_cc_v6v4_isl_trie_info))


/*
 * NAME: IX_CC_V6V4_ISL_NUM_TRIE_BLOCKS
 *
 * DESCRIPTION: Get the total number of trie blocks based on the size
 * information in the registry.    This is just a guess and might need to be
 * adjusted.
 *
 * @Param:      - IN    __r
 *                      size hint in registry.
 */
#define IX_CC_V6V4_ISL_NUM_TRIE_BLOCKS(__r)                                 \
    (ix_uint16)((((__r) / 2) < IX_CC_V6V4_ISL_MIN_TRIE_BLOCKS) ?            \
                IX_CC_V6V4_ISL_MIN_TRIE_BLOCKS : (__r) / 2)


/*
 * NAME: IX_CC_V6V4_ISL_NUM_INDEX_BLOCKS
 *
 * DESCRIPTION: Get the total number of index blocks based on the size
 * information in the registry.
 *
 * @Param:      - IN    __r
 *                      size hint in registry.
 */
#define IX_CC_V6V4_ISL_NUM_INDEX_BLOCKS(__r)   (ix_uint16)(__r)


/*
 * NAME: IX_CC_V6V4_ISL_NUM_TRIE_INFO_BLOCKS
 *
 * DESCRIPTION: Get the total number of trie info blocks based on the size
 * information in the registry.  This is just a guess and might need to be
 * adjusted.
 *
 * @Param:      - IN    __r
 *                      size hint in registry.
 */
#define IX_CC_V6V4_ISL_NUM_TRIE_INFO_BLOCKS(__r)  ((__r) / 3)


/*
 * NAME: IX_CC_V6V4_ISL_TRIE_INFO_HASH_INDEX
 *
 * DESCRIPTION: Get the hash index of a trie info block based on the
 * trie index and entry index.
 *
 * @Param:      - IN    __ip
 *                      32-bit IP address.
 * @Param:      - IN    __ml
 *                      8-bit mask length (1 - 32)
 */
#define IX_CC_V6V4_ISL_TRIE_INFO_HASH_INDEX(__ip, __ml)                     \
    ((ix_uint8)((((__ip) >> 24) ^ ((__ip) >> 16) ^ ((__ip) >> 8) ^          \
                 (__ip) ^ (__ml)) & 0x7f))

/*
 * NAME: IX_CC_V6V4_ISL_CHECK_TRIE_INFO
 *
 * DESCRIPTION: Check a trie info to see if a prefix length is valid for
 * a given location.
 *
 * @Param:      - IN    __ti
 *                      Trie Info block.
 * @Param:      - IN    __ix
 *                      Index of tunnel in corresponding index block.
 * @Param:      - IN    __pf
 *                      Prefix bits used in last trie (-1)
 */
#define IX_CC_V6V4_ISL_CHECK_TRIE_INFO(__ti, __ix, __pf)                    \
    (*((ix_uint8*)((__ti) + 1) + (__ix)) & (1 << (__pf)))


/*
 * NAME: IX_CC_V6V4_ISL_GET_TRIE_INFO_DATA
 *
 * DESCRIPTION: Get the data byte corresponding to a given location in
 * the corresponding index block.
 *
 * @Param:      - IN    __ti
 *                      Trie Info block.
 * @Param:      - IN    __ix
 *                      Index of tunnel in corresponding index block.
 * @Param:      - IN    __pf
 *                      Prefix bits used in last trie (-1)
 */
#define IX_CC_V6V4_ISL_GET_TRIE_INFO_DATA(__ti, __ix)                       \
    (*((ix_uint8*)((__ti) + 1) + (__ix)))


/*
 * NAME: IX_CC_V6V4_ISL_TRIE_OFFSET
 *
 * DESCRIPTION: Get the offset (in 32-bit words) of a trie block relative to
 * the start of the trie shared memory.
 *
 * @Param:      - IN    __ti
 *                      Trie index.
 */
#define IX_CC_V6V4_ISL_TRIE_OFFSET(__ti)                                    \
    (IX_CC_V6V4_ISL_TRIE_BASE_ENTRIES +                                     \
     (IX_CC_V6V4_ISL_TRIE_ENTRIES * (__ti)))

/*
 * NAME: IX_CC_V6V4_ISL_TRIE_ADDR
 *
 * DESCRIPTION: Get the address of a trie block.
 *
 * @Param:      - IN    __base
 *                      Base address of ingress list blocks.
 * @Param:      - IN    __ti
 *                      Trie index.
 */
#define IX_CC_V6V4_ISL_TRIE_ADDR(__base, __ti)                              \
    &(__base)[IX_CC_V6V4_ISL_TRIE_OFFSET(__ti)]

/*
 * NAME: IX_CC_V6V4_ISL_NEXT_TRIE_FROM_ENTRY
 *
 * DESCRIPTION: Get the next trie from the trie entry value.
 *
 * @Param:      - IN    __ev
 *                      32-bit entry value from trie block.
 */
#define IX_CC_V6V4_ISL_NEXT_TRIE_FROM_ENTRY(__ev)                          \
    ((ix_uint16)(IX_NTOH32(__ev) & 0xffff))

/*
 * NAME: IX_CC_V6V4_ISL_INDEX_FROM_ENTRY
 *
 * DESCRIPTION: Get the index value from the trie entry value.
 *
 * @Param:      - IN    __ev
 *                      32-bit entry value from trie block.
 */
#define IX_CC_V6V4_ISL_INDEX_FROM_ENTRY(__ev)                               \
    ((ix_uint16)(IX_NTOH32(__ev) >> 16))

/*
 * NAME: IX_CC_V6V4_ISL_WRITE_BASE_TRIE_INDEX
 *
 * DESCRIPTION: Write the specified index value into the base trie block.
 *
 * @Param:      - IN    __base
 *                      Base address of ingress list blocks.
 * @Param:      - IN    __ei
 *                      Entry index (0 - 255).
 * @Param:      - IN    __ix
 *                      Index value.
 *
 */
#define IX_CC_V6V4_ISL_WRITE_BASE_TRIE_INDEX(__base, __ei, __ix)            \
    (void) IX_RM_MEM_UINT16_WRITE(                                          \
        (ix_uint16*)(&(__base)[__ei]), (__ix))

/*
 * NAME: IX_CC_V6V4_ISL_WRITE_BASE_TRIE_NEXT
 *
 * DESCRIPTION: Write the specified next value into the base trie block.
 *
 * @Param:      - IN    __base
 *                      Base address of ingress list blocks.
 * @Param:      - IN    __ei
 *                      Entry index (0 - 255).
 * @Param:      - IN    __nt
 *                      Next trie value.
 *
 */
#define IX_CC_V6V4_ISL_WRITE_BASE_TRIE_NEXT(__base, __ei, __nt)             \
    (void) IX_RM_MEM_UINT16_WRITE(                                          \
        (ix_uint16*)(&(__base)[__ei]) + 1, (__nt))

/*
 * NAME: IX_CC_V6V4_ISL_READ_BASE_TRIE_INDEX
 *
 * DESCRIPTION: Read an index value from the base trie block.
 *
 * @Param:      - IN    __base
 *                      Base address of ingress list blocks.
 * @Param:      - IN    __ei
 *                      Entry index (0 - 255).
 *
 */
#define IX_CC_V6V4_ISL_READ_BASE_TRIE_INDEX(__base, __ei)                   \
    IX_RM_MEM_UINT16_READ((ix_uint16*)(&(__base)[__ei]))

/*
 * NAME: IX_CC_V6V4_ISL_READ_BASE_TRIE_NEXT
 *
 * DESCRIPTION: Read a next trie value from the base trie block.
 *
 * @Param:      - IN    __base
 *                      Base address of ingress list blocks.
 * @Param:      - IN    __ei
 *                      Entry index (0 - 255).
 *
 */
#define IX_CC_V6V4_ISL_READ_BASE_TRIE_NEXT(__base, __ei)                    \
    IX_RM_MEM_UINT16_READ((ix_uint16*)(&(__base)[__ei]) + 1)

/*
 * NAME: IX_CC_V6V4_ISL_WRITE_TRIE_INDEX
 *
 * DESCRIPTION: Write the specified index value into a trie block.
 *
 * @Param:      - IN    __base
 *                      Base address of ingress list blocks.
 * @Param:      - IN    __ti
 *                      Trie index.
 * @Param:      - IN    __ei
 *                      Entry index (0 - 255).
 * @Param:      - IN    __ix
 *                      Index value.
 *
 */
#define IX_CC_V6V4_ISL_WRITE_TRIE_INDEX(__base, __ti, __ei, __ix)           \
    (void) IX_RM_MEM_UINT16_WRITE(                                          \
        (ix_uint16*)(IX_CC_V6V4_ISL_TRIE_ADDR(__base, __ti) + (__ei)),      \
        (__ix))

/*
 * NAME: IX_CC_V6V4_ISL_WRITE_TRIE_NEXT
 *
 * DESCRIPTION: Write the specified next value into a trie block.
 *
 * @Param:      - IN    __base
 *                      Base address of ingress list blocks.
 * @Param:      - IN    __ti
 *                      Trie index.
 * @Param:      - IN    __ei
 *                      Entry index (0 - 255).
 * @Param:      - IN    __nt
 *                      Next trie value.
 *
 */
#define IX_CC_V6V4_ISL_WRITE_TRIE_NEXT(__base, __ti, __ei, __nt)            \
    (void) IX_RM_MEM_UINT16_WRITE(                                          \
        (ix_uint16*)(IX_CC_V6V4_ISL_TRIE_ADDR(__base, __ti) + (__ei)) + 1,  \
        (__nt))

/*
 * NAME: IX_CC_V6V4_ISL_READ_TRIE_INDEX
 *
 * DESCRIPTION: Read an index value from a trie block.
 *
 * @Param:      - IN    __base
 *                      Base address of ingress list blocks.
 * @Param:      - IN    __ti
 *                      Trie index.
 * @Param:      - IN    __ei
 *                      Entry index (0 - 15).
 *
 */
#define IX_CC_V6V4_ISL_READ_TRIE_INDEX(__base, __ti, __ei)                  \
    IX_RM_MEM_UINT16_READ(                                                  \
        (ix_uint16*)(IX_CC_V6V4_ISL_TRIE_ADDR(__base, __ti) + (__ei)))


/*
 * NAME: IX_CC_V6V4_ISL_READ_TRIE_NEXT
 *
 * DESCRIPTION: Read a next trie value from a trie block.
 *
 * @Param:      - IN    __base
 *                      Base address of ingress list blocks.
 * @Param:      - IN    __ti
 *                      Trie index.
 * @Param:      - IN    __ei
 *                      Entry index (0 - 15).
 *
 */
#define IX_CC_V6V4_ISL_READ_TRIE_NEXT(__base, __ti, __ei)                   \
    IX_RM_MEM_UINT16_READ(                                                  \
        (ix_uint16*)(IX_CC_V6V4_ISL_TRIE_ADDR(__base, __ti) + (__ei)) + 1)


/*
 * NAME: IX_CC_V6V4_ISL_READ_INDEX_BLOCK
 *
 * DESCRIPTION: Read an index block.
 *
 * @Param:      - IN    __cxt
 *                      Context information.
 * @Param:      - IN    __bi
 *                      Block index.
 * @Param:      - IN    __buf
 *                      buffer to receive the data.
 *
 */
#define IX_CC_V6V4_ISL_READ_INDEX_BLOCK(__cxt, __bi, __buf)                 \
{                                                                           \
    ix_uint8 __cnt = (__cxt)->indexBlockSize >> 3;                          \
    ix_uint64* __pBlk = (ix_uint64*)(__cxt)->pIndexBase + ((__bi) * __cnt); \
    ix_uint8 __i;                                                           \
    for (__i = 0; __i < __cnt; __i++)                                       \
    {                                                                       \
        *((ix_uint64*)(__buf) + __i) = IX_RM_MEM_UINT64_READ(__pBlk);       \
        __pBlk++;                                                           \
    }                                                                       \
}


/*
 * NAME: IX_CC_V6V4_ISL_CLEAR_INDEX_BLOCK
 *
 * DESCRIPTION: Clear an index block.
 *
 * @Param:      - IN    __cxt
 *                      Context information.
 * @Param:      - IN    __bi
 *                      Block index.
 *
 */
#define IX_CC_V6V4_ISL_CLEAR_INDEX_BLOCK(__cxt, __bi)                       \
{                                                                           \
    ix_uint8 __cnt = (__cxt)->indexBlockSize >> 3;                          \
    ix_uint64* __pBlk = (ix_uint64*)(__cxt)->pIndexBase + ((__bi) * __cnt); \
    ix_uint8 __i;                                                           \
    for (__i = 0; __i < __cnt; __i++)                                       \
    {                                                                       \
        (void)IX_RM_MEM_UINT64_WRITE(__pBlk, 0);                            \
        __pBlk++;                                                           \
    }                                                                       \
}


/*
 * NAME: IX_CC_V6V4_ISL_WRITE_INDEX_ENTRY
 *
 * DESCRIPTION: Write a 16-bit entry in an index block.
 *
 * @Param:      - IN    __cxt
 *                      Context information.
 * @Param:      - IN    __bi
 *                      Block index.
 * @Param:      - IN    __ei
 *                      Entry index.
 * @Param:      - IN    __ev
 *                      Entry value.
 *
 */
#define IX_CC_V6V4_ISL_WRITE_INDEX_ENTRY(__cxt, __bi, __ei, __ev)           \
    (void)IX_RM_MEM_UINT16_WRITE(                                           \
        (ix_uint16*)(__cxt)->pIndexBase +                                   \
                    ((__bi) * ((__cxt)->indexBlockSize >> 1)) + (__ei),     \
        (__ev))

/*
 * NAME: IX_CC_V6V4_ISL_WRITE_INDEX_NEXT
 *
 * DESCRIPTION: Write the next block value in an index block.
 *
 * @Param:      - IN    __cxt
 *                      Context information.
 * @Param:      - IN    __bi
 *                      Block index.
 * @Param:      - IN    __ev
 *                      Entry value.
 *
 */
#define IX_CC_V6V4_ISL_WRITE_INDEX_NEXT(__cxt, __bi, __ev)                  \
    (void)IX_RM_MEM_UINT16_WRITE(                                           \
        (ix_uint16*)(__cxt)->pIndexBase +                                   \
                    ((__bi) * ((__cxt)->indexBlockSize >> 1)), (__ev))

/*
 * NAME: IX_CC_V6V4_ISL_READ_INDEX_NEXT
 *
 * DESCRIPTION: Read the next block value from an index block.
 *
 * @Param:      - IN    __cxt
 *                      Context information.
 * @Param:      - IN    __bi
 *                      Block index.
 * @Param:      - IN    __ev
 *                      Entry value.
 *
 */
#define IX_CC_V6V4_ISL_READ_INDEX_NEXT(__cxt, __bi)                         \
    IX_RM_MEM_UINT16_READ(                                                  \
        (ix_uint16*)(__cxt)->pIndexBase +                                   \
                    ((__bi) * ((__cxt)->indexBlockSize >> 1)))

/**
 * Type definitions whose scope is limited to this file
 */

/**
 * TYPENAME: ix_cc_v6v4_ingress_list_trie_entry
 * 
 * DESCRIPTION: This structure defines the format for an entry in an
 * ingress source validation list trie table.
 *
 */
typedef struct ix_s_cc_v6v4_ingress_list_trie_entry
{
    ix_uint16 index;
    ix_uint16 nextTrie;
} ix_cc_v6v4_ingress_list_trie_entry;


/**
 * TYPENAME: ix_cc_v6v4_isl_trie_info
 * 
 * DESCRIPTION: This structure defines the format for the information
 * saved for prefixes that have overlapping entries.  The fixed length
 * information defined in the structure is followed by a variable number
 * of bytes that contain information about the prefix lengths that are
 * valid for the tunnel entry in the corresponding index block.  The size
 * of this variable length data depends on the length of the index block,
 * which is determined by a registry setting.
 *
 */
typedef struct ix_s_cc_v6v4_isl_trie_info
{
    struct ix_s_cc_v6v4_isl_trie_info* pHashNext;
    struct ix_s_cc_v6v4_isl_trie_info* pNext;
    ix_uint32 ipAddr;
    ix_uint8  maskLength;
    ix_uint16 index;
} ix_cc_v6v4_isl_trie_info;


/**
 * TYPENAME: ix_cc_v6v4_isl_index_update
 * 
 * DESCRIPTION: This structure contains information to assist in updating
 * the ingress list index blocks and associated trie info blocks.  Some
 * fields apply to adds only.
 *
 */
typedef struct ix_s_cc_v6v4_isl_index_update
{
    /* The index block actually used to store the tunnel index */
    ix_uint16 indexBlock;

    /* Index block previous to the one used for this tunnel */
    ix_uint16 prevIndexBlock;

    /* Indicates whether a new index block was allocated */
    ix_uint8  newIndexBlock;

    /* Indicates whether a new trie info block was allocated */
    ix_uint8  newTrieInfo;

    /* Location in index block where tunnel resides or will reside */
    ix_uint8  indexLocation;

    /* Indicates whether tunnel already exists for the prefix */
    ix_uint8  existing;

    /* The ip address associated with this update */
    ix_uint32 ipAddr;

    /* The mask length associated with this update */
    ix_uint8 maskLength;

    /* The hash index for this ip address and mask length */
    ix_uint8 hashIndex;

    /* Indicates if this is the only entry in the block */
    ix_uint8 only;

    /* The trie info block corresponding to indexBlock */
    ix_cc_v6v4_isl_trie_info* pTrieInfo;

    /* Existing trie info block to which to chain trie info block */
    ix_cc_v6v4_isl_trie_info* pPrevTrieInfo;

} ix_cc_v6v4_isl_index_update;


/**
 * TYPENAME: ix_cc_v6v4_isl_trie_update
 * 
 * DESCRIPTION: This structure contains information to assist in updating
 * the ingress list trie blocks.  Some fields apply to adds only.
 *
 */
typedef struct ix_s_cc_v6v4_isl_trie_update
{
    /* The ip address associated with this update */
    ix_uint32 ipAddr;

    /* The trie blocks used to make this prefix */
    ix_uint16 aTries[IX_CC_V6V4_ISL_MAX_CHAIN];

    /* The total number of tries beyond the base required */
    ix_uint8  numTries;

    /* The number of tries in the chain that are newly allocated */
    ix_uint8  numNewTries;

    /* The number of significant bits (-1) of the trie block used by
     * the prefix
     */
    ix_uint8 prefixOffset;

    /* The number of entries in the trie block that are affected */
    ix_uint8 numExpansions;

    /* Update info for the index blocks affected by this update */
    ix_cc_v6v4_isl_index_update* paIndexInfo;
} ix_cc_v6v4_isl_trie_update;


/**
 * TYPENAME: ix_cc_v6v4_ingress_list_trie_context
 * 
 * DESCRIPTION: This structure defines the format for the context
 * information needed to manage the trie format ingress source list.
 *
 */
typedef struct ix_s_cc_v6v4_ingress_list_trie_context
{
    /* pIndexBase: base address of index blocks */
    ix_uint32*          pIndexBase;

    /* numTries: the total number of 4-bit trie blocks */
    ix_uint16           numTries;

    /* numIndexBlocks: the total number of index blocks */
    ix_uint16           numIndexBlocks;

    /* indexBlockSize: the size of each index block */
    ix_uint16           indexBlockSize;

    /* trieInfoSize: the size of each trie info block */
    ix_uint16           trieInfoSize;

    /* trieBlockPool: the index pool for allocating 4-bit trie blocks */
    ix_cc_index_pool    trieBlockPool;

    /* indexBlockPool: the index pool for allocating index blocks */
    ix_cc_index_pool    indexBlockPool;

    /* trieInfoPool: the index pool for allocating trie info blocks */
    ix_cc_index_pool    trieInfoPool;

    /* pTrieInfoBase: base address of pool of all trie info blocks */
    ix_uint8*           pTrieInfoBase;

    /* aTrieInfoHash: hash table containing in-use trie info blocks */
    ix_cc_v6v4_isl_trie_info* 
                        aTrieInfoHash[IX_CC_V6V4_TRIE_INFO_HASH_SIZE];

} ix_cc_v6v4_ingress_list_trie_context;


/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
extern ix_cc_v6v4_registry_data g_V6v4RegistryData;
#else
ix_cc_v6v4_registry_data g_V6v4RegistryData;
#endif

/** 
 * Extern function prototypes.
 */


/** 
 * Static function prototypes.
 */


/** 
 * Function definitions.
 */


/**
 * NAME: _ix_cc_v6v4_isl_init_trie_info
 *
 * DESCRIPTION: This helper routine is used to initialize a newly allocated
 * trie info block.
 *           
 * @Param:  - In arg_pContext - pointer to ingress list context information.
 *
 * @Param:  - IN arg_pIndexInfo - pointer to helper structure containing
 * information about the specific index block to which the tunnel is being
 * added.  The pointer to the new trie info block is also in this structure.
 *
 * @Return: None.
 *
 */

PRIVATE void _ix_cc_v6v4_isl_init_trie_info(
    ix_cc_v6v4_ingress_list_trie_context* arg_pContext,
    ix_cc_v6v4_isl_index_update* arg_pIndexInfo
    )
{
    ix_cc_v6v4_isl_trie_info* pTrieInfo = arg_pIndexInfo->pTrieInfo;

    ix_ossl_memset(pTrieInfo, 0, arg_pContext->trieInfoSize);

    /*
     * If the trie info is newly allocated but the index block is not,
     * then we need to initialize the data portion of the trie info 
     * to correspond to the tunnels already in the block (by default
     * they will all correspond to the maximum prefix offset).
     */
    if (!arg_pIndexInfo->newIndexBlock)
    {
        ix_uint16 indexBlock[IX_CC_V6V4_MAX_INDEX_BLOCK_SIZE/sizeof(ix_uint16)];
        ix_uint8 maxLoc = arg_pContext->indexBlockSize / 2;
        ix_uint8 loc;
        ix_uint8 mask;
        ix_uint8* pTrieInfoData;

        pTrieInfoData = (ix_uint8*)(pTrieInfo + 1);

        mask = (arg_pIndexInfo->maskLength <= IX_CC_V6V4_ISL_BASE_BITS) ?
            (ix_uint8)(1 << (IX_CC_V6V4_ISL_BASE_BITS - 1)) :
            (ix_uint8)(1 << (IX_CC_V6V4_ISL_TRIE_BITS - 1));


        IX_CC_V6V4_ISL_READ_INDEX_BLOCK(
            arg_pContext, arg_pIndexInfo->indexBlock, indexBlock);

        for (loc = 1; loc < maxLoc; loc++)
        {
            if (indexBlock[loc] != 0)
            {
                pTrieInfoData[loc] |= mask;
            }
        }
    }
} /* end - _ix_cc_v6v4_isl_init_trie_info */

/**
 * NAME: _ix_cc_v6v4_isl_find_trie_info
 *
 * DESCRIPTION: This helper routine is used to allocate the trie info block
 * associated with an index block if necessary.
 *           
 * @Param:  - In arg_pContext - pointer to ingress list context information.
 *
 * @Param:  - IN arg_Reserve - reserve a new trie info if not found.
 *
 * @Param:  - IN/OUT arg_pIndexInfo - pointer to helper structure containing
 * information about the specific index block to which the tunnel is being
 * added. If arg_Reserve is set to 1, this routine will allocate a new trie
 * info block if one was not found, and supply the pointer.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_FULL - cannot add any more entries (only if reserve option is
 * requested)
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_isl_find_trie_info(
    ix_cc_v6v4_ingress_list_trie_context* arg_pContext,
    ix_uint8 arg_Reserve,
    ix_cc_v6v4_isl_index_update* arg_pIndexInfo
    )
{
    ix_cc_v6v4_isl_trie_info* pTrieInfo;
    ix_uint32 trieInfoIndex;

    /* If a trie info already exists, get the pointer */

    arg_pIndexInfo->hashIndex = IX_CC_V6V4_ISL_TRIE_INFO_HASH_INDEX(
        arg_pIndexInfo->ipAddr, arg_pIndexInfo->maskLength);

    pTrieInfo = arg_pContext->aTrieInfoHash[arg_pIndexInfo->hashIndex];

    while (pTrieInfo != NULL)
    {
        if ((pTrieInfo->ipAddr == arg_pIndexInfo->ipAddr) &&
            (pTrieInfo->maskLength == arg_pIndexInfo->maskLength))
        {
            do
            {
                if (pTrieInfo->index == arg_pIndexInfo->indexBlock)
                {
                    arg_pIndexInfo->pTrieInfo = pTrieInfo;
                    return IX_SUCCESS;
                }
                arg_pIndexInfo->pPrevTrieInfo = pTrieInfo;
                pTrieInfo = pTrieInfo->pNext;
            } while (pTrieInfo != NULL);
        }
        else
        {
            arg_pIndexInfo->pPrevTrieInfo = pTrieInfo;
            pTrieInfo = pTrieInfo->pHashNext;
        }
    }

    /*
     * There is no trie info for this entry yet.  Allocate one if this
     * prefix requires it.
     */

    if (arg_Reserve)
    {
        ix_uint32 offset;

        IX_ERROR_CR(
            ix_cc_index_pool_alloc(arg_pContext->trieInfoPool, &trieInfoIndex)
        );

        offset = trieInfoIndex * arg_pContext->trieInfoSize;

        arg_pIndexInfo->pTrieInfo = (ix_cc_v6v4_isl_trie_info*)
            (&arg_pContext->pTrieInfoBase[offset]);
        arg_pIndexInfo->newTrieInfo = 1;

        _ix_cc_v6v4_isl_init_trie_info(arg_pContext, arg_pIndexInfo);
    }

    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_isl_find_trie_info */


/**
 * NAME: _ix_cc_v6v4_isl_find_tunnel
 *
 * DESCRIPTION: This helper routine is used to find a tunnel associated with
 * a prefix and optionally reserve a location to add the tunnel if not found.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_TunnelIndex - tunnel index to be located.
 *
 * @Param:  - IN arg_Reserve - reserve a location for the tunnel if not
 * found.
 *
 * @Param:  - IN/OUT arg_pIndexInfo - pointer to helper structure containing
 * information about the operation.  On input this contains the existing index
 * block allocated for the prefix, if any. On ouput, the remaining information
 * about where to record the tunnel index is added (if the reserve option is
 * chosen).
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_FULL - cannot add any more entries (only if reserve option is
 * requested)
 * 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - tunnel not found (only if reserve option is
 * not requested)
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_isl_find_tunnel(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_uint16 arg_TunnelIndex,
    ix_uint8 arg_Reserve,
    ix_cc_v6v4_isl_index_update* arg_pIndexInfo
    )
{
    ix_uint16 indexBlock[IX_CC_V6V4_MAX_INDEX_BLOCK_SIZE/sizeof(ix_uint16)];
    ix_uint32 blockIndex = arg_pIndexInfo->indexBlock;
    ix_cc_v6v4_ingress_list_trie_context* pContext;
    ix_uint8 loc;
    ix_uint8 maxLoc;
    ix_uint16 prevBlock = 0;
    ix_uint16 availBlock = 0;
    ix_uint8 availLoc;
    ix_uint8 onlyEntry;
    ix_uint16 tunnelId = IX_HTON16(arg_TunnelIndex);

    pContext = (ix_cc_v6v4_ingress_list_trie_context*)arg_pData->pIslContext;
    maxLoc = pContext->indexBlockSize / 2;

    while (blockIndex != 0)
    {
        /* Read index block */

        IX_CC_V6V4_ISL_READ_INDEX_BLOCK(pContext, blockIndex, indexBlock);

        /* Look for the tunnel index */

        onlyEntry = 1;

        for (loc = 1; loc < maxLoc; loc++)
        {
            if (indexBlock[loc] == tunnelId)
            {
                /* Found it */
                arg_pIndexInfo->prevIndexBlock = prevBlock;
                arg_pIndexInfo->indexBlock = (ix_uint16)blockIndex;
                arg_pIndexInfo->indexLocation = loc;
                arg_pIndexInfo->existing = 1;

                if (!arg_Reserve)
                {
                    for (loc++; onlyEntry && (loc < maxLoc); loc++)
                    {
                        if (indexBlock[loc] != 0)
                        {
                            onlyEntry = 0;
                        }
                    }
                    arg_pIndexInfo->only = onlyEntry;
                }

                return IX_SUCCESS;
            }
            else if (indexBlock[loc] != 0)
            {
                onlyEntry = 0;
            }
            else if (arg_Reserve && (availBlock == 0))
            {
                /* Found an available slot - save the info */
                availBlock = (ix_uint16)blockIndex;
                availLoc = loc;
            }
        }

        prevBlock = (ix_uint16)blockIndex;
        blockIndex = indexBlock[0];
    }

    if (arg_Reserve)
    {
        if (availBlock != 0)
        {
            /*
             * Didn't find the tunnel in any of the blocks, but found a place
             * to put it.  Don't care about prevBlock in this case, since we
             * won't be adding or removing an index block.
             */
            arg_pIndexInfo->indexBlock = availBlock;
            arg_pIndexInfo->indexLocation = availLoc;

            return IX_SUCCESS;
        }

        /*
         * We need to allocate a new block.
         */
        IX_ERROR_CR(
            ix_cc_index_pool_alloc(pContext->indexBlockPool, &blockIndex)
        );

        arg_pIndexInfo->prevIndexBlock = prevBlock;
        arg_pIndexInfo->indexBlock = (ix_uint16)blockIndex;
        arg_pIndexInfo->indexLocation = 1;
        arg_pIndexInfo->newIndexBlock = 1;

        return IX_SUCCESS;
    }
    else
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
            ("Tunnel not associated with prefix"));
    }

} /* end - _ix_cc_v6v4_isl_find_tunnel */


/**
 * NAME: _ix_cc_v6v4_isl_search_trie
 *
 * DESCRIPTION: This primitive is used to search a trie block for prefixes
 * allowed for a particular tunnel, and if found, write them to an output
 * buffer.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_TunnelIndex - end tunnel index.
 *
 * @Param:  - IN arg_pIndexInfo - pointer to structure used to maintain
 * information about the prefixes as they are searched. On input, the 
 * maskLength parameter is set to indicate the prefix length we are working
 * with, and ipAddr contains the address bits corresponding to the previous
 * tries.  The remaining fields can be used as working variables by the
 * routine, but are not significant upon return.
 *
 * @Param:  - IN arg_TrieIndex - index of trie to be searched.
 *
 * @Param:  - IN/OUT arg_pNumEntries - On input, points to the number of
 * entries the buffer pointed to by arg_paEntries can accommodate.  On output,
 * contains the actual number of prefixes in the ingress list for the
 * tunnel.
 *
 * @Param:  - OUT arg_paEntries - pointer to array into which to place the
 * entries.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_BUFFER_TOO_SMALL - buffer not large enough to contain
 * all entries.
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_isl_search_trie(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_uint16 arg_TunnelIndex,
    ix_cc_v6v4_isl_index_update* arg_pIndexInfo,
    ix_uint16 arg_TrieIndex,
    ix_uint32* arg_pNumEntries,
    ix_cc_v6v4_ingress_source_entry* arg_paEntries
    )
{
    ix_error err;
    ix_cc_v6v4_ingress_source_entry* pEntry;    /* current entry in output 
                                                   buffer */
    ix_uint16 entry;         /* current index in trie block */
    ix_uint32 trieEntry;    /* 32-bit value of trie entry */
    ix_uint16 nextTrie;     /* 16-bit next trie index (trieEntry 31:16) */
    ix_uint32* pTrieEntry;  /* address of current entry in trie block */
    ix_uint32 numEntries;   /* number of entries found so far */
    ix_uint32 entryCount;   /* max entries left to find */
    ix_uint8 trieBits;      /* 8 for base, 4 otherwise */
    ix_uint16 trieSize;     /* 256 for base, 16 otherwise */
    ix_uint8 shiftAmount;   /* shift amount for constructing ip address */
    ix_uint32 addrMask;     /* mask used in constructing ip address */


    /*
     * Calculate parameters dependent on whether we are searching the
     * base or a 4-bit trie.
     */
    shiftAmount = 32 - arg_pIndexInfo->maskLength;
    if (arg_pIndexInfo->maskLength == IX_CC_V6V4_ISL_BASE_BITS)
    {
        trieSize = IX_CC_V6V4_ISL_TRIE_BASE_ENTRIES;
        trieBits = IX_CC_V6V4_ISL_BASE_BITS;
        pTrieEntry = arg_pData->pMbIslBase;
        addrMask = 0;
    }
    else
    {
        trieSize = IX_CC_V6V4_ISL_TRIE_ENTRIES;
        trieBits = IX_CC_V6V4_ISL_TRIE_BITS;
        pTrieEntry = IX_CC_V6V4_ISL_TRIE_ADDR(
            arg_pData->pMbIslBase, arg_TrieIndex);
        addrMask = (0xffffffff << (shiftAmount + trieBits));
    }

    /*
     * Set up remaining parameters
     */
    numEntries = 0;
    pEntry = arg_paEntries;
    
    /* Search the trie for entries with a nonn-zero index */

    for (entry = 0; entry < trieSize; entry++)
    {
        /* Set the IP address prefix corresponding to this entry */

        arg_pIndexInfo->ipAddr &= addrMask;
        arg_pIndexInfo->ipAddr |= (entry << shiftAmount);

        trieEntry = IX_RM_MEM_UINT32_READ(pTrieEntry);
        arg_pIndexInfo->indexBlock =
            IX_CC_V6V4_ISL_INDEX_FROM_ENTRY(trieEntry);

        if (arg_pIndexInfo->indexBlock != 0)
        {
            /*
             * Look for the tunnel
             */
            arg_pIndexInfo->indexLocation = 0;
            arg_pIndexInfo->existing = 0;
            arg_pIndexInfo->pTrieInfo = NULL;

            err =  _ix_cc_v6v4_isl_find_tunnel(
                    arg_pData, arg_TunnelIndex, 0, arg_pIndexInfo);

            if (err == IX_SUCCESS)
            {
                _ix_cc_v6v4_isl_find_trie_info(
                    (ix_cc_v6v4_ingress_list_trie_context*)
                        arg_pData->pIslContext,
                    0, arg_pIndexInfo);

                if (arg_pIndexInfo->pTrieInfo != NULL)
                {
                    /* 
                     * Add an entry in the user buffer for each prefix offset
                     * flagged in the trie info byte for this location. Note
                     * we only record the entry for the first expanded address
                     * to avoid writing equivalent entries to the buffer.
                     */
                    ix_uint8 prefixBits;
                    ix_uint8 i;

                    prefixBits = IX_CC_V6V4_ISL_GET_TRIE_INFO_DATA(
                        arg_pIndexInfo->pTrieInfo,
                        arg_pIndexInfo->indexLocation);

                    for (i = 0; i < trieBits; i++)
                    {
                        if ((prefixBits & (1 << i)) &&
                            ((arg_pIndexInfo->ipAddr & 
                             (~addrMask >> (i + 1))) == 0))
                        {
                            /* Add this prefix to the buffer */
                            if (numEntries < *arg_pNumEntries)
                            {
                                pEntry->ipAddr = arg_pIndexInfo->ipAddr;
                                pEntry->maskLength =
                                    arg_pIndexInfo->maskLength -
                                    trieBits + 1 + i;
                                pEntry++;
                            }
                            numEntries++;
                        }
                    }
                }
                else
                {
                    if (numEntries < *arg_pNumEntries)
                    {
                        /* 
                         * Add an entry in the user buffer for the prefix length
                         * corresponding to this trie
                         */
                        pEntry->ipAddr = arg_pIndexInfo->ipAddr;
                        pEntry->maskLength = arg_pIndexInfo->maskLength;
                        pEntry++;
                    }
                    numEntries++;
                }
            } /* end - if (err == IX_SUCCESS) */
        } /* end - if (arg_pIndexInfo->indexBlock != 0) */

        nextTrie = IX_CC_V6V4_ISL_NEXT_TRIE_FROM_ENTRY(trieEntry);

        if (nextTrie != 0)
        {
            /*
             * Examine next trie
             */
            entryCount = *arg_pNumEntries - numEntries;

            arg_pIndexInfo->maskLength += IX_CC_V6V4_ISL_TRIE_BITS;

            IX_ERROR_CR(
                _ix_cc_v6v4_isl_search_trie(
                    arg_pData, arg_TunnelIndex, arg_pIndexInfo,
                    nextTrie, &entryCount, pEntry));

            arg_pIndexInfo->maskLength -= IX_CC_V6V4_ISL_TRIE_BITS;

            pEntry += entryCount;
            numEntries += entryCount;
        }

        pTrieEntry++;

#if !defined(IX_DEBUG)
        if (numEntries >= *arg_pNumEntries)
        {
            break;
        }
#endif

    } /* end - for */

    if (numEntries <= *arg_pNumEntries)
    {
        err = IX_SUCCESS;
    }
    else
    {
        err = IX_ERROR_WARNING(
            IX_CC_V6V4_ERROR_BUFFER_TOO_SMALL, ("Buffer too small"));
    }

    *arg_pNumEntries = numEntries;
    return err;

} /* end - _ix_cc_v6v4_isl_search_trie */


/**
 * NAME: _ix_cc_v6v4_isl_delete_tries
 *
 * DESCRIPTION: This helper routine is used to check the tries for a prefix
 * for which a tunnel was just removed, and determine if one or more trie
 * blocks can be freed.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_pUpdateInfo - pointer to helper structure containing
 * information about the operation.  The information in this structure has
 * already been set up -- all trie blocks have been allocated.  All this
 * routine needs to do is write the information to memory to commit the
 * update.
 *
 *
 * @Return: None.
 *
 */

PRIVATE void _ix_cc_v6v4_isl_delete_tries(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_cc_v6v4_isl_trie_update* arg_pUpdateInfo
    )
{
    ix_cc_v6v4_ingress_list_trie_context* pContext;
    ix_uint8 trieNum;
    ix_uint8 entryIndex;
    ix_uint32 trieEntry;
    ix_uint32 *pTrieEntry;
    ix_uint32 *pTrieEnd;
    ix_uint8 done = 0;

    pContext = (ix_cc_v6v4_ingress_list_trie_context*)arg_pData->pIslContext;

    trieNum = arg_pUpdateInfo->numTries;

    while (!done && (trieNum > 0))
    {
        trieNum--;

        /* Check for non-zero entries in the trie block */
        pTrieEntry = IX_CC_V6V4_ISL_TRIE_ADDR(
                        arg_pData->pMbIslBase,
                        arg_pUpdateInfo->aTries[trieNum]);
        pTrieEnd = pTrieEntry + IX_CC_V6V4_ISL_TRIE_ENTRIES;
        trieEntry = 0;

        while ((trieEntry == 0) && (pTrieEntry < pTrieEnd))
        {
            trieEntry = IX_RM_MEM_UINT32_READ(pTrieEntry);
            pTrieEntry++;
        }

        if (trieEntry == 0)
        {
            /* Trie block can be freed */

            entryIndex =
                (ix_uint8)(arg_pUpdateInfo->ipAddr >> (24 - (trieNum * 4)));

            if (trieNum == 0)
            {
                /* Updating the base */

                IX_CC_V6V4_ISL_WRITE_BASE_TRIE_NEXT(
                    arg_pData->pMbIslBase, entryIndex, 0);
            }
            else
            {
                /* Updating previous 4-bit trie */

                entryIndex &= 0xf;

                IX_CC_V6V4_ISL_WRITE_TRIE_NEXT(
                    arg_pData->pMbIslBase,
                    arg_pUpdateInfo->aTries[trieNum - 1],
                    entryIndex, 0);
            }

            ix_cc_index_pool_free(pContext->trieBlockPool,
                                  arg_pUpdateInfo->aTries[trieNum]);

        } /* end - if (trieEntry = 0) */
        else
        {
            done = 1;
        }
    } /* end - while (trieNum > 0) */

} /* end - _ix_cc_v6v4_isl_delete_tries */


/**
 * NAME: _ix_cc_v6v4_isl_clear_trie_info
 *
 * DESCRIPTION: This helper routine is used to clear the appropriate flag in
 * a trie info block to invalidate a prefix length for an entry in an index
 * block.
 *           
 * @Param:  - In arg_pContext - pointer to ingress list context information.
 *
 * @Param:  - IN arg_PrefixBits - Number of bits (-1) in the final trie that
 * are used by the prefix: 0-7 for prefix length <=8, 0-3 otherwise.
 *
 * @Param:  - IN arg_pIndexInfo - pointer to helper structure containing
 * information about the specific index block to which the tunnel is being
 * added.
 *
 *
 * @Return: Value of trie info data byte after clearing the bit for this
 * prefix.
 *
 */

PRIVATE ix_uint8 _ix_cc_v6v4_isl_clear_trie_info(
    ix_cc_v6v4_ingress_list_trie_context* arg_pContext,
    ix_uint8 arg_PrefixBits,
    ix_cc_v6v4_isl_index_update* arg_pIndexInfo
    )
{
    ix_uint8 mask;
    ix_uint8 trieInfoData;
    ix_cc_v6v4_isl_trie_info* pTrieInfo;
    ix_cc_v6v4_isl_trie_info* pPrevInfo;
    ix_uint8* pTrieInfoData;
    ix_uint8 i;
    ix_uint8 empty = 1;
    ix_uint8 count;
    ix_uint32* pData;
    ix_uint32 offset;

    pTrieInfo = arg_pIndexInfo->pTrieInfo;
    pTrieInfoData = (ix_uint8*)(pTrieInfo + 1) +
                    (arg_pIndexInfo->indexLocation);

    mask = (1 << arg_PrefixBits);
    *pTrieInfoData &= ~mask;
    trieInfoData = *pTrieInfoData;

    count = (arg_pContext->trieInfoSize - sizeof(ix_cc_v6v4_isl_trie_info)) /
            sizeof(ix_uint32);
    pData = (ix_uint32*)(pTrieInfo + 1);
    for (i = 0; empty && (i < count); i++)
    {
        empty = (*pData++ == 0);
    }

    if (empty)
    {
        /* Delete the trie info */

        if (arg_pIndexInfo->pPrevTrieInfo)
        {
            pPrevInfo = arg_pIndexInfo->pPrevTrieInfo;

            if ((pPrevInfo->ipAddr == pTrieInfo->ipAddr) &&
                (pPrevInfo->maskLength == pTrieInfo->maskLength))
            {
                pPrevInfo->pNext = pTrieInfo->pNext;
            }
            else if (pTrieInfo->pNext != NULL)
            {
                pTrieInfo->pNext->pHashNext = pTrieInfo->pHashNext;
                pPrevInfo->pHashNext = pTrieInfo->pNext;
            }
            else
            {
                pPrevInfo->pHashNext = pTrieInfo->pHashNext;
            }
        }
        else if (pTrieInfo->pNext != NULL)
        {
            pTrieInfo->pNext->pHashNext = pTrieInfo->pHashNext;
            arg_pContext->aTrieInfoHash[arg_pIndexInfo->hashIndex] =
                pTrieInfo->pNext;
        }
        else
        {
            arg_pContext->aTrieInfoHash[arg_pIndexInfo->hashIndex] =
                pTrieInfo->pHashNext;
        }

        offset = ((ix_uint32)pTrieInfo -
                  (ix_uint32)arg_pContext->pTrieInfoBase) /
                 arg_pContext->trieInfoSize;

        ix_cc_index_pool_free(
            arg_pContext->trieInfoPool, (ix_uint16)offset);

        arg_pIndexInfo->pTrieInfo = NULL;
    }

    return trieInfoData;

} /* end - _ix_cc_v6v4_isl_clear_trie_info */


/**
 * NAME: _ix_cc_v6v4_isl_delete_tunnel_from_block
 *
 * DESCRIPTION: This helper routine is used to remove a tunnel from an
 * an index block and update the associated trie info block accordingly.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_TrieIndex - index of trie block that points to the
 * index block.
 *
 * @Param:  - IN arg_EntryIndex - index withing trie block contains the
 * entry.
 *
 * @Param:  - IN arg_PrefixOffset - number of prefix bits (-1) in the trie
 * block that are significant for the tunnel being removed.
 *
 * @Param:  - IN arg_pIndexInfo - pointer to helper structure containing
 * information about the operation.  The information in this structure has
 * already been set up. All this routine needs to do is write the information
 * to memory to commit the update.
 *
 * @Param:  - OUT arg_pDelPrefix - pointer to location to to update with
 * a flag that indicates whethter the prefix can be removed (no more tunnels
 * allow it).
 *
 *
 * @Return: None.
 *
 */

PRIVATE void _ix_cc_v6v4_isl_delete_tunnel_from_block(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_uint16 arg_TrieIndex,
    ix_uint8 arg_EntryIndex,
    ix_uint8 arg_PrefixOffset,
    ix_cc_v6v4_isl_index_update* arg_pIndexInfo,
    ix_uint8* arg_pDelPrefix
    )
{
    ix_cc_v6v4_ingress_list_trie_context* pContext;
    ix_uint16 nextBlock;
    ix_uint8 trieInfoData;


    pContext = (ix_cc_v6v4_ingress_list_trie_context*)arg_pData->pIslContext;
    *arg_pDelPrefix = 0;

    /* Update the trie info if necessary */

    if (arg_pIndexInfo->pTrieInfo != NULL)
    {
        trieInfoData = _ix_cc_v6v4_isl_clear_trie_info(
            pContext, arg_PrefixOffset, arg_pIndexInfo);
    }
    else
    {
        trieInfoData = 0;
    }

    if (trieInfoData == 0)
    {
        /*
         * Clear the tunnel from the designated location in the index
         * block
         */
        IX_CC_V6V4_ISL_WRITE_INDEX_ENTRY(
            pContext,
            arg_pIndexInfo->indexBlock, arg_pIndexInfo->indexLocation, 0);

        if (arg_pIndexInfo->only)
        {
            /*
             * If the index block is now empty, write the next block into
             * the previous block for the prefix or into the trie entry
             * if there is no existing index block for the prefix.
             */
            nextBlock = IX_CC_V6V4_ISL_READ_INDEX_NEXT(
                    pContext, arg_pIndexInfo->indexBlock);

            if (arg_pIndexInfo->prevIndexBlock != 0)
            {
                IX_CC_V6V4_ISL_WRITE_INDEX_NEXT(
                    pContext, arg_pIndexInfo->prevIndexBlock,
                    nextBlock);
            }
            else
            {
                if (arg_TrieIndex == 0)
                {
                    IX_CC_V6V4_ISL_WRITE_BASE_TRIE_INDEX(
                        arg_pData->pMbIslBase, arg_EntryIndex, nextBlock);
                }
                else
                {
                    IX_CC_V6V4_ISL_WRITE_TRIE_INDEX(
                        arg_pData->pMbIslBase, arg_TrieIndex,
                        arg_EntryIndex, nextBlock);
                }
                if (nextBlock == 0)
                {
                    *arg_pDelPrefix = 1;
                }
            }

            ix_cc_index_pool_free(
                pContext->indexBlockPool, arg_pIndexInfo->indexBlock);

        } /* end - if (pIndexInfo->only) */
    } /* end - if (trieInfoData == 0) */

} /* end - _ix_cc_v6v4_isl_delete_tunnel_from_block */


/**
 * NAME: _ix_cc_v6v4_isl_search_and_del_tunnel
 *
 * DESCRIPTION: This primitive is used to search a trie block for prefixes
 * that are allowed for a particular tunnel, and if found, delete the tunnel.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_TunnelIndex - index of end tunnel to be removed.
 *
 * @Param:  - IN arg_pIndexInfo - pointer to structure used to maintain
 * information about the prefixes as they are searched. On input, the 
 * maskLength parameter is set to indicate the prefix length we are working
 * with, and ipAddr contains the address bits corresponding to the previous
 * tries.  The remaining fields can be used as working variables by the
 * routine, but are not significant upon return.
 *
 * @Param:  - IN arg_TrieIndex - index of trie to be searched.
 *
 * @Param:  - IN/OUT arg_pNumEntries - On input, points to the maximun number
 * of entries to remove.  On output, contains the actual number of entries
 * removed.
 *
 * @Param:  - OUT arg_pEmpty - Indicates whether trie block is empty
 * after tunnel was removed.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error.
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_isl_search_and_del_tunnel(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_uint16 arg_TunnelIndex,
    ix_cc_v6v4_isl_index_update* arg_pIndexInfo,
    ix_uint16 arg_TrieIndex,
    ix_uint32* arg_pNumEntries,
    ix_uint8* arg_pEmpty
    )
{
    ix_cc_v6v4_ingress_list_trie_context* pContext;
    ix_error err;
    ix_uint16 entry;         /* current index in trie block */
    ix_uint32 trieEntry;    /* 32-bit value of trie entry */
    ix_uint16 nextTrie;     /* 16-bit next trie index (trieEntry 31:16) */
    ix_uint32* pTrieEntry;  /* address of current entry in trie block */
    ix_uint32 numEntries;   /* number of entries found so far */
    ix_uint32 entryCount;   /* max entries left to find */
    ix_uint8 trieBits;      /* 8 for base, 4 otherwise */
    ix_uint16 trieSize;     /* 256 for base, 16 otherwise */
    ix_uint8 shiftAmount;   /* shift amount for constructing ip address */
    ix_uint32 addrMask;     /* mask used in constructing ip address */
    ix_uint8 trieEmpty;     /* indicates whether trie block is empty */


    pContext = (ix_cc_v6v4_ingress_list_trie_context*)arg_pData->pIslContext;

    /*
     * Calculate parameters dependent on whether we are searching the
     * base or a 4-bit trie.
     */
    shiftAmount = 32 - arg_pIndexInfo->maskLength;
    if (arg_pIndexInfo->maskLength == IX_CC_V6V4_ISL_BASE_BITS)
    {
        trieSize = IX_CC_V6V4_ISL_TRIE_BASE_ENTRIES;
        trieBits = IX_CC_V6V4_ISL_BASE_BITS;
        pTrieEntry = arg_pData->pMbIslBase;
        addrMask = 0;
    }
    else
    {
        trieSize = IX_CC_V6V4_ISL_TRIE_ENTRIES;
        trieBits = IX_CC_V6V4_ISL_TRIE_BITS;
        pTrieEntry = IX_CC_V6V4_ISL_TRIE_ADDR(
            arg_pData->pMbIslBase, arg_TrieIndex);
        addrMask = (0xffffffff << (shiftAmount + trieBits));
    }

    /*
     * Set up remaining parameters
     */
    numEntries = 0;
    trieEmpty = 1;
    
    /* Search the trie for entries with a nonn-zero index */

    for (entry = 0; entry < trieSize; entry++)
    {
        /* Set the IP address prefix corresponding to this entry */

        arg_pIndexInfo->ipAddr &= addrMask;
        arg_pIndexInfo->ipAddr |= (entry << shiftAmount);

        trieEntry = IX_RM_MEM_UINT32_READ(pTrieEntry);
        arg_pIndexInfo->indexBlock =
            IX_CC_V6V4_ISL_INDEX_FROM_ENTRY(trieEntry);

        if (arg_pIndexInfo->indexBlock != 0)
        {
            /*
             * Look for the tunnel
             */
            arg_pIndexInfo->indexLocation = 0;
            arg_pIndexInfo->prevIndexBlock = 0;
            arg_pIndexInfo->existing = 0;
            arg_pIndexInfo->only = 0;
            arg_pIndexInfo->pTrieInfo = NULL;
            arg_pIndexInfo->pPrevTrieInfo = NULL;

            err =  _ix_cc_v6v4_isl_find_tunnel(
                    arg_pData, arg_TunnelIndex, 0, arg_pIndexInfo);

            if (err == IX_SUCCESS)
            {
                ix_uint8 delPrefix;

                _ix_cc_v6v4_isl_find_trie_info(
                    (ix_cc_v6v4_ingress_list_trie_context*)
                        arg_pData->pIslContext,
                    0, arg_pIndexInfo);

                if (arg_pIndexInfo->pTrieInfo != NULL)
                {
                    /* 
                     * Remove the tunnel for all flagged prefix lengths.
                     * When we call _ix_cc_v6v4_isl_delete_tunnel_from_block
                     * for the last non-zero bit, we will clear the tunnel
                     * from the index block, and remove the index block
                     * and/or trie info if necesary.
                     */
                    ix_uint8 prefixBits;
                    ix_uint8 i;

                    prefixBits = IX_CC_V6V4_ISL_GET_TRIE_INFO_DATA(
                        arg_pIndexInfo->pTrieInfo,
                        arg_pIndexInfo->indexLocation);

                    for (i = 0; i < trieBits; i++)
                    {
                        if (prefixBits & (1 << i))
                        {
                            if ((arg_pIndexInfo->ipAddr & 
                                (~addrMask >> (i + 1))) == 0)
                            {
                                numEntries++;
                            }
                            _ix_cc_v6v4_isl_delete_tunnel_from_block(
                                arg_pData, arg_TrieIndex, (ix_uint8)entry, i,
                                arg_pIndexInfo, &delPrefix);
                        }
                    }
                }
                else
                {
                    _ix_cc_v6v4_isl_delete_tunnel_from_block(
                        arg_pData, arg_TrieIndex, (ix_uint8)entry,
                        (ix_uint8)(trieBits - 1),
                        arg_pIndexInfo, &delPrefix);
                    numEntries++;
                }

                if (!delPrefix)
                {
                    trieEmpty = 0;
                }
            } /* end - if (err == IX_SUCCESS) */
        } /* end - if (arg_pIndexInfo->indexBlock != 0) */

        /*
         * If there is a next trie, check it next.
         */

        nextTrie = IX_CC_V6V4_ISL_NEXT_TRIE_FROM_ENTRY(trieEntry);

        if (nextTrie != 0)
        {
            ix_uint8 empty;

            /*
             * Examine next trie
             */
            entryCount = *arg_pNumEntries - numEntries;
            arg_pIndexInfo->maskLength += IX_CC_V6V4_ISL_TRIE_BITS;

            IX_ERROR_CR(
                _ix_cc_v6v4_isl_search_and_del_tunnel(
                    arg_pData, arg_TunnelIndex, arg_pIndexInfo,
                    nextTrie, &entryCount, &empty));

            arg_pIndexInfo->maskLength -= IX_CC_V6V4_ISL_TRIE_BITS;
            numEntries += entryCount;

            if (empty)
            {
                if (arg_TrieIndex == 0)
                {
                    /* Updating the base */

                    IX_CC_V6V4_ISL_WRITE_BASE_TRIE_NEXT(
                        arg_pData->pMbIslBase, entry, 0);
                }
                else
                {
                    /* Updating previous 4-bit trie */

                    IX_CC_V6V4_ISL_WRITE_TRIE_NEXT(
                        arg_pData->pMbIslBase, arg_TrieIndex, entry, 0);
                }

                ix_cc_index_pool_free(pContext->trieBlockPool, nextTrie);
            }
            else
            {
                trieEmpty = 0;
            }
        } /* end if (nextTrie != 0) */

        pTrieEntry++;

#if !defined(IX_DEBUG)
        if (numEntries >= *arg_pNumEntries)
        {
            break;
        }
#endif

    } /* end - for */

    *arg_pEmpty = trieEmpty;

    if (numEntries <= *arg_pNumEntries)
    {
        *arg_pNumEntries = numEntries;
        return IX_SUCCESS;
    }

    return IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
            ("Ingress list size inconsistent with tunnel entry"));

} /* end - _ix_cc_v6v4_isl_search_and_del_tunnel */


/**
 * NAME: _ix_cc_v6v4_isl_delete_tunnel
 *
 * DESCRIPTION: This helper routine is used to update the index blocks and
 * trie info blocks required to remove the association between a tunnel
 * and a prefix.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_pUpdateInfo - pointer to helper structure containing
 * information about the operation.  The information in this structure has
 * already been set up -- all index blocks and trie blocks have been located.
 * All this routine needs to do is write the information to memory to commit
 * the update.
 *
 *
 * @Return: None.
 *
 */

PRIVATE void _ix_cc_v6v4_isl_delete_tunnel(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_cc_v6v4_isl_trie_update* arg_pUpdateInfo
    )
{
    ix_cc_v6v4_ingress_list_trie_context* pContext;
    ix_uint8 shiftAmount = 24;
    ix_uint8 index;
    ix_uint8 startingIndex;
    ix_uint8 endingIndex;
    ix_cc_v6v4_isl_index_update* pIndexInfo;
    ix_uint16 trieIndex;
    ix_uint8 delPrefix;
    ix_uint8 numDeletes = 0;


    pContext = (ix_cc_v6v4_ingress_list_trie_context*)arg_pData->pIslContext;

    shiftAmount -= (arg_pUpdateInfo->numTries * 4);
    startingIndex = (ix_uint8)(arg_pUpdateInfo->ipAddr >> shiftAmount);

    if (arg_pUpdateInfo->numTries > 0)
    {
        startingIndex &= 0xf;
    }

    endingIndex = startingIndex + arg_pUpdateInfo->numExpansions;
    pIndexInfo = arg_pUpdateInfo->paIndexInfo;

    trieIndex = (arg_pUpdateInfo->numTries == 0) ?
        0 : arg_pUpdateInfo->aTries[arg_pUpdateInfo->numTries - 1];

    for (index = startingIndex; index < endingIndex; index++)
    {
        _ix_cc_v6v4_isl_delete_tunnel_from_block(
            arg_pData,
            trieIndex,
            index,
            arg_pUpdateInfo->prefixOffset,
            pIndexInfo, &delPrefix);

        if (delPrefix)
        {
            numDeletes++;
        }

        pIndexInfo++;

    } /* end - for */

    /*
     * If the index block was removed for all prefixes, then the one or more
     * trie blocks at the end of the chain could be empty. Check for this
     * and delete if detected.
     */
    if ((numDeletes == arg_pUpdateInfo->numExpansions) &&
        (arg_pUpdateInfo->numTries != 0))
    {
        _ix_cc_v6v4_isl_delete_tries(arg_pData, arg_pUpdateInfo);
    }

} /* end - _ix_cc_v6v4_isl_delete_tunnel */


/**
 * NAME: _ix_cc_v6v4_isl_get_index_blocks
 *
 * DESCRIPTION: This helper routine is used to retrieve the index blocks and
 * trie info blocks associated with a tunnel/prefix binding.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_TunnelIndex - tunnel index of interest.
 *
 * @Param:  - IN/OUT arg_pUpdateInfo - pointer to helper structure containing
 * information about the operation.  On input this contains the information
 * recorded by the trie retrieval procedure. On ouput, the information about
 * the index blocks and trie info blocks associated with the tunnel.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - tunnel not found 
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_isl_get_index_blocks(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_uint16 arg_TunnelIndex,
    ix_cc_v6v4_isl_trie_update* arg_pUpdateInfo
    )
{
    ix_cc_v6v4_ingress_list_trie_context* pContext;
    ix_uint8 shiftAmount = 24;
    ix_uint8 index;
    ix_uint8 startingIndex;
    ix_uint8 endingIndex;
    ix_uint32 offset = 0;
    ix_uint16 indexBlock;
    ix_cc_v6v4_isl_index_update* pIndexInfo;

    pContext = (ix_cc_v6v4_ingress_list_trie_context*)arg_pData->pIslContext;

    shiftAmount -= (arg_pUpdateInfo->numTries * 4);
    startingIndex = (ix_uint8)(arg_pUpdateInfo->ipAddr >> shiftAmount);

    if (arg_pUpdateInfo->numTries > 0)
    {
        startingIndex &= 0xf;
    }

    endingIndex = startingIndex + arg_pUpdateInfo->numExpansions;
    pIndexInfo = arg_pUpdateInfo->paIndexInfo;

    for (index = startingIndex; index < endingIndex; index++)
    {
        pIndexInfo->ipAddr =
            (arg_pUpdateInfo->ipAddr | (offset << shiftAmount));

        pIndexInfo->maskLength = (arg_pUpdateInfo->numTries * 4) + 8;

        /* Read the entry and extract the index */
        if (arg_pUpdateInfo->numTries == 0)
        {
            /* Prefix is represented in base (no 4-bit tries) */
            indexBlock = IX_CC_V6V4_ISL_READ_BASE_TRIE_INDEX(
                arg_pData->pMbIslBase, index);
        }
        else
        {
            /* Read entry from last trie */
            indexBlock = IX_CC_V6V4_ISL_READ_TRIE_INDEX(
                arg_pData->pMbIslBase,
                arg_pUpdateInfo->aTries[arg_pUpdateInfo->numTries - 1],
                index);
        }

        pIndexInfo->indexBlock = IX_NTOH16(indexBlock);

        /* Find the tunnel in the index blocks */ 
        IX_ERROR_CR(
            _ix_cc_v6v4_isl_find_tunnel(
                arg_pData, arg_TunnelIndex, 0, pIndexInfo)
        );

        /*
         * Locate the trie info block if available, and check if
         * the correct prefix length is represented for the tunnel.
         */ 
        _ix_cc_v6v4_isl_find_trie_info(pContext, 0, pIndexInfo);

        if (pIndexInfo->pTrieInfo != NULL)
        {
            if (IX_CC_V6V4_ISL_CHECK_TRIE_INFO(
                    pIndexInfo->pTrieInfo, pIndexInfo->indexLocation,
                    arg_pUpdateInfo->prefixOffset) == 0)
            {
                return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                            ("Tunnel not associated with prefix"));
            }
        }
        else if (arg_pUpdateInfo->numExpansions > 1)
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                        ("Tunnel not associated with prefix"));
        }

        pIndexInfo++;
        offset++;

    } /* end - for */

    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_isl_get_index_blocks */


/**
 * NAME: _ix_cc_v6v4_isl_get_tries
 *
 * DESCRIPTION: This helper routine is used to retrieve the trie blocks
 * associated with a prefix into the trie table.  The helper structure is
 * updated with the results.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - OUT arg_pUpdateInfo - pointer to helper structure to be updated
 * with info about the insert.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - entries for prefix not found 
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_isl_get_tries(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_cc_v6v4_isl_trie_update* arg_pUpdateInfo
    )
{
    ix_uint8 index;
    ix_uint16 nextTrie;
    ix_uint8 curTrie = 0;
    ix_uint8 shiftAmount = 24;

    if (arg_pUpdateInfo->numTries == 0)
    {
        return IX_SUCCESS;
    }

    /*
     * Read the base entry
     */
    index = (ix_uint8)(arg_pUpdateInfo->ipAddr >> shiftAmount);
    nextTrie = IX_CC_V6V4_ISL_READ_BASE_TRIE_NEXT(arg_pData->pMbIslBase, index);

    /*
     * Get all 4-bit tries.
     */
    while ((nextTrie != 0) && (curTrie < arg_pUpdateInfo->numTries))
    {
        nextTrie = IX_NTOH16(nextTrie);

        arg_pUpdateInfo->aTries[curTrie] = nextTrie;
        curTrie++;

        shiftAmount -= 4;
        index = (ix_uint8)(arg_pUpdateInfo->ipAddr >> shiftAmount) & 0xf;

        nextTrie = IX_CC_V6V4_ISL_READ_TRIE_NEXT(
            arg_pData->pMbIslBase, nextTrie, index);
    } /* end - while */

    if (curTrie < arg_pUpdateInfo->numTries)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                ("Prefix not found in table"));
    }


    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_isl_get_tries */


/**
 * NAME: _ix_cc_v6v4_isl_link_tries
 *
 * DESCRIPTION: This helper routine is used to update the next trie pointers
 * in the trie blocks involved in the add-prefix update.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_pUpdateInfo - pointer to helper structure containing
 * information about the operation.  The information in this structure has
 * already been set up -- all trie blocks have been allocated.  All this
 * routine needs to do is write the information to memory to commit the
 * update.
 *
 *
 * @Return: None.
 *
 */

PRIVATE void _ix_cc_v6v4_isl_link_tries(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_cc_v6v4_isl_trie_update* arg_pUpdateInfo
    )
{
    ix_cc_v6v4_ingress_list_trie_context* pContext;
    ix_uint8 numUpdates;
    ix_uint8 trieNum;
    ix_uint8 entryIndex;

    pContext = (ix_cc_v6v4_ingress_list_trie_context*)arg_pData->pIslContext;

    numUpdates = arg_pUpdateInfo->numNewTries;
    trieNum = arg_pUpdateInfo->numTries;

    while (numUpdates > 0)
    {
        trieNum--;
        entryIndex =
            (ix_uint8)(arg_pUpdateInfo->ipAddr >> (24 - (trieNum * 4)));
        if (trieNum == 0)
        {
            /* Updating the base */

            IX_CC_V6V4_ISL_WRITE_BASE_TRIE_NEXT(
                arg_pData->pMbIslBase, entryIndex,
                IX_HTON16(arg_pUpdateInfo->aTries[trieNum]));
        }
        else
        {
            /* Updating previous 4-bit trie */

            entryIndex &= 0xf;

            IX_CC_V6V4_ISL_WRITE_TRIE_NEXT(
                arg_pData->pMbIslBase,
                arg_pUpdateInfo->aTries[trieNum - 1],
                entryIndex,
                IX_HTON16(arg_pUpdateInfo->aTries[trieNum]));
        }

        numUpdates--;
    }

} /* end - _ix_cc_v6v4_isl_link_tries */


/**
 * NAME: _ix_cc_v6v4_isl_set_trie_info
 *
 * DESCRIPTION: This helper routine is used to set the appropriate flag in
 * a trie info block to associate an entry in an index block with the
 * correct prefix length.
 *           
 * @Param:  - In arg_pContext - pointer to ingress list context information.
 *
 * @Param:  - IN arg_PrefixBits - Number of bits (-1) in the final trie that
 * are used by the prefix: 0-7 for prefix length <=8, 0-3 otherwise.
 *
 * @Param:  - IN arg_pIndexInfo - pointer to helper structure containing
 * information about the specific index block to which the tunnel is being
 * added.
 *
 *
 * @Return: None.
 *
 */

PRIVATE void _ix_cc_v6v4_isl_set_trie_info(
    ix_cc_v6v4_ingress_list_trie_context* arg_pContext,
    ix_uint8 arg_PrefixBits,
    ix_cc_v6v4_isl_index_update* arg_pIndexInfo
    )
{
    ix_uint8 mask;
    ix_cc_v6v4_isl_trie_info* pTrieInfo;
    ix_cc_v6v4_isl_trie_info* pPrevInfo;

    mask = (1 << arg_PrefixBits);

    pTrieInfo = arg_pIndexInfo->pTrieInfo;
    *((ix_uint8*)(pTrieInfo + 1) + (arg_pIndexInfo->indexLocation)) |= mask;

    if (arg_pIndexInfo->newTrieInfo)
    {
        pTrieInfo->ipAddr = arg_pIndexInfo->ipAddr;
        pTrieInfo->maskLength = arg_pIndexInfo->maskLength;
        pTrieInfo->index = arg_pIndexInfo->indexBlock;
        
        /* Save the new trie info in the hash table */
        if (arg_pIndexInfo->pPrevTrieInfo)
        {
            pPrevInfo = arg_pIndexInfo->pPrevTrieInfo;

            if ((pPrevInfo->ipAddr == pTrieInfo->ipAddr) &&
                (pPrevInfo->maskLength == pTrieInfo->maskLength))
            {
                pPrevInfo->pNext = pTrieInfo;
            }
            else
            {
                pPrevInfo->pHashNext = pTrieInfo;
            }
        }
        else
        {
            pTrieInfo->pHashNext =
                arg_pContext->aTrieInfoHash[arg_pIndexInfo->hashIndex];
            arg_pContext->aTrieInfoHash[arg_pIndexInfo->hashIndex] = pTrieInfo;
        }
    }

} /* end - _ix_cc_v6v4_isl_set_trie_info */


/**
 * NAME: _ix_cc_v6v4_isl_add_tunnel
 *
 * DESCRIPTION: This helper routine is used to update the index blocks and
 * trie info blocks required to associate a tunnel with a prefix.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_TunnelIndex - tunnel index to be added.
 *
 * @Param:  - IN arg_pUpdateInfo - pointer to helper structure containing
 * information about the operation.  The information in this structure has
 * already been set up -- all index blocks and trie blocks have been allocated.
 * All this routine needs to do is write the information to memory to commit
 * the update.
 *
 *
 * @Return: None.
 *
 */

PRIVATE void _ix_cc_v6v4_isl_add_tunnel(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_uint16 arg_TunnelIndex,
    ix_cc_v6v4_isl_trie_update* arg_pUpdateInfo
    )
{
    ix_cc_v6v4_ingress_list_trie_context* pContext;
    ix_uint8 shiftAmount = 24;
    ix_uint8 index;
    ix_uint8 startingIndex;
    ix_uint8 endingIndex;
    ix_cc_v6v4_isl_index_update* pIndexInfo;


    pContext = (ix_cc_v6v4_ingress_list_trie_context*)arg_pData->pIslContext;

    shiftAmount -= (arg_pUpdateInfo->numTries * 4);
    startingIndex = (ix_uint8)(arg_pUpdateInfo->ipAddr >> shiftAmount);

    if (arg_pUpdateInfo->numTries > 0)
    {
        startingIndex &= 0xf;
    }

    endingIndex = startingIndex + arg_pUpdateInfo->numExpansions;
    pIndexInfo = arg_pUpdateInfo->paIndexInfo;

    for (index = startingIndex; index < endingIndex; index++)
    {

        if (pIndexInfo->newIndexBlock)
        {
            /* Clear the index block if it is new */
            IX_CC_V6V4_ISL_CLEAR_INDEX_BLOCK(
                pContext, pIndexInfo->indexBlock);
        }

        if (!pIndexInfo->existing)
        {
            /*
             * Write the tunnel into the designated location in the index
             * block
             */
            IX_CC_V6V4_ISL_WRITE_INDEX_ENTRY(
                pContext, pIndexInfo->indexBlock,
                pIndexInfo->indexLocation, IX_HTON16(arg_TunnelIndex));
        }

        if (pIndexInfo->newIndexBlock)
        {
            /*
             * If the index block is new, write the index into the previous
             * block for the prefix or into the trie entry if there is no
             * existing index block for the prefix.
             */
            if (pIndexInfo->prevIndexBlock != 0)
            {
                IX_CC_V6V4_ISL_WRITE_INDEX_NEXT(
                    pContext, pIndexInfo->prevIndexBlock,
                    IX_HTON16(pIndexInfo->indexBlock));
            }
            else if (arg_pUpdateInfo->numTries == 0)
            {
                IX_CC_V6V4_ISL_WRITE_BASE_TRIE_INDEX(
                    arg_pData->pMbIslBase,
                    index,
                    IX_HTON16(pIndexInfo->indexBlock));
            }
            else
            {
                IX_CC_V6V4_ISL_WRITE_TRIE_INDEX(
                    arg_pData->pMbIslBase,
                    arg_pUpdateInfo->aTries[arg_pUpdateInfo->numTries - 1],
                    index,
                    IX_HTON16(pIndexInfo->indexBlock));
            }
        } /* end - if (pIndexInfo->newIndexBlock) */

        /* Update the trie info if necessary */

        if (pIndexInfo->pTrieInfo != NULL)
        {
            _ix_cc_v6v4_isl_set_trie_info(
                pContext, arg_pUpdateInfo->prefixOffset, pIndexInfo);
        }

        pIndexInfo++;

    } /* end - for */

} /* end - _ix_cc_v6v4_isl_add_tunnel */


/**
 * NAME: _ix_cc_v6v4_isl_alloc_index_blocks
 *
 * DESCRIPTION: This helper routine is used to allocate the index blocks and
 * trie info blocks required to associate a tunnel with a prefix.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_TunnelIndex - tunnel index to be added.
 *
 * @Param:  - IN/OUT arg_pUpdateInfo - pointer to helper structure containing
 * information about the operation.  On input this contains the information
 * recorded by the trie allocate procedure. On ouput, the information about the
 * index block allocations is added.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_DUPLICATE_ENTRY - entry already exists 
 * IX_CC_ERROR_FULL - cannot add any more entries 
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_isl_alloc_index_blocks(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_uint16 arg_TunnelIndex,
    ix_cc_v6v4_isl_trie_update* arg_pUpdateInfo
    )
{
    ix_cc_v6v4_ingress_list_trie_context* pContext;
    ix_uint8 shiftAmount = 24;
    ix_uint8 index;
    ix_uint8 startingIndex;
    ix_uint8 endingIndex;
    ix_uint32 offset = 0;
    ix_cc_v6v4_isl_index_update* pIndexInfo;
    ix_error err;
    ix_uint8 reserve;
    ix_uint16 indexBlock;

    pContext = (ix_cc_v6v4_ingress_list_trie_context*)arg_pData->pIslContext;

    shiftAmount -= (arg_pUpdateInfo->numTries * 4);
    startingIndex = (ix_uint8)(arg_pUpdateInfo->ipAddr >> shiftAmount);

    if (arg_pUpdateInfo->numTries > 0)
    {
        startingIndex &= 0xf;
    }

    endingIndex = startingIndex + arg_pUpdateInfo->numExpansions;
    pIndexInfo = arg_pUpdateInfo->paIndexInfo;

    for (index = startingIndex; index < endingIndex; index++)
    {
        pIndexInfo->ipAddr =
            (arg_pUpdateInfo->ipAddr | (offset << shiftAmount));

        pIndexInfo->maskLength = (arg_pUpdateInfo->numTries * 4) + 8;

        /* Read the entry and extract the index */
        if (arg_pUpdateInfo->numTries == 0)
        {
            /* Prefix is represented in base (no 4-bit tries) */
            indexBlock = IX_CC_V6V4_ISL_READ_BASE_TRIE_INDEX(
                arg_pData->pMbIslBase, index);
        }
        else
        {
            /* Read entry from last trie */
            indexBlock = IX_CC_V6V4_ISL_READ_TRIE_INDEX(
                arg_pData->pMbIslBase,
                arg_pUpdateInfo->aTries[arg_pUpdateInfo->numTries - 1],
                index);
        }

        pIndexInfo->indexBlock = IX_NTOH16(indexBlock);

        /* Reserve a location in the index block(s) for the tunnel */ 
        IX_ERROR_CG(
            _ix_cc_v6v4_isl_find_tunnel(
                arg_pData, arg_TunnelIndex, 1, pIndexInfo),
            err,
            cleanup
        );

        /* Locate the trie info, reserving one if necessary */
        reserve = (arg_pUpdateInfo->numExpansions > 1) ? 1 : 0;
        IX_ERROR_CG(
            _ix_cc_v6v4_isl_find_trie_info(pContext, reserve, pIndexInfo),
            err,
            cleanup
        );

        /* Check for duplicate entry */
        if (pIndexInfo->existing)
        {
            if (pIndexInfo->pTrieInfo != NULL)
            {
                if (IX_CC_V6V4_ISL_CHECK_TRIE_INFO(
                        pIndexInfo->pTrieInfo, pIndexInfo->indexLocation,
                        arg_pUpdateInfo->prefixOffset))
                {
                    err = IX_CC_ERROR_DUPLICATE_ENTRY;
                    goto cleanup;
                }
            }
            else if (arg_pUpdateInfo->numExpansions > 1)
            {
                err = IX_CC_ERROR_DUPLICATE_ENTRY;
                goto cleanup;
            }
        }

        pIndexInfo++;
        offset++;

    } /* end - for */

    return IX_SUCCESS;

cleanup:
    pIndexInfo = arg_pUpdateInfo->paIndexInfo;
    for (index = startingIndex; index < endingIndex; index++)
    {
        if (pIndexInfo->newTrieInfo)
        {
            ix_uint32 offset = 
                ((ix_uint32)pIndexInfo->pTrieInfo -
                 (ix_uint32)pContext->pTrieInfoBase) /
                pContext->trieInfoSize;

            ix_cc_index_pool_free(pContext->trieInfoPool, offset);
        }

        if (pIndexInfo->newIndexBlock)
        {
            ix_cc_index_pool_free(
                pContext->indexBlockPool, pIndexInfo->indexBlock);
        }
        pIndexInfo++;
    }

    return err;
} /* end - _ix_cc_v6v4_isl_alloc_index_blocks */


/**
 * NAME: _ix_cc_v6v4_isl_alloc_tries
 *
 * DESCRIPTION: This helper routine is used to allocate the trie blocks needed
 * to insert a prefix into the trie table.  The helper structure is updated
 * with the results.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - OUT arg_pUpdateInfo - pointer to helper structure to be updated
 * with info about the insert.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_FULL - cannot add any more entries 
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_isl_alloc_tries(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_cc_v6v4_isl_trie_update* arg_pUpdateInfo
    )
{
    ix_cc_v6v4_ingress_list_trie_context* pContext;
    ix_uint8 index;
    ix_uint16 nextTrie;
    ix_uint8 curTrie = 0;
    ix_uint8 shiftAmount = 24;
    ix_error err;
    ix_uint32 newTrieIndex;

    if (arg_pUpdateInfo->numTries == 0)
    {
        return IX_SUCCESS;
    }

    pContext = (ix_cc_v6v4_ingress_list_trie_context*)arg_pData->pIslContext;

    /*
     * Read the base entry
     */
    index = (ix_uint8)(arg_pUpdateInfo->ipAddr >> shiftAmount);
    nextTrie = IX_CC_V6V4_ISL_READ_BASE_TRIE_NEXT(arg_pData->pMbIslBase, index);

    /*
     * Get all tries for this prefix that are already allocated.
     */
    while ((nextTrie != 0) && (curTrie < arg_pUpdateInfo->numTries))
    {
        nextTrie = IX_NTOH16(nextTrie);

        arg_pUpdateInfo->aTries[curTrie] = nextTrie;
        curTrie++;

        shiftAmount -= 4;
        index = (ix_uint8)(arg_pUpdateInfo->ipAddr >> shiftAmount) & 0xf;

        nextTrie = IX_CC_V6V4_ISL_READ_TRIE_NEXT(
            arg_pData->pMbIslBase, nextTrie, index);
    } /* end - for */

    /*
     * Allocate the remaining tries needed.
     */
    while (curTrie < arg_pUpdateInfo->numTries)
    {
        IX_ERROR_CG(
            ix_cc_index_pool_alloc(pContext->trieBlockPool, &newTrieIndex),
            err,
            cleanupTries
        );

        arg_pUpdateInfo->aTries[curTrie] = (ix_uint16)newTrieIndex;
        curTrie++;
        arg_pUpdateInfo->numNewTries++;

    } /* end - for */

    return IX_SUCCESS;

cleanupTries:
    while (arg_pUpdateInfo->numNewTries > 0)
    {
        curTrie--;
        ix_cc_index_pool_free(
            pContext->trieBlockPool, arg_pUpdateInfo->aTries[curTrie]);
        arg_pUpdateInfo->numNewTries--;
    }

    return err;

} /* end - _ix_cc_v6v4_isl_alloc_tries */


/**
 * NAME: _ix_cc_v6v4_isl_trie_add_prefix
 *
 * DESCRIPTION: This primitive is used to add a prefix to the ingress list
 * for a tunnel.  The caller must have already validated the tunnel index
 * and the address for acceptability.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_TunnelIndex - end tunnel index (not used for this
 * ISL format).
 *
 * @Param:  - IN arg_pEntry - pointer to ingress list entry.
 *
 * @Param:  - IN/OUT arg_pTunnelInfo - on input, this points to the 16-bit
 * word in the tunnel configuration entry pertaining to the ingress
 * list.  On output, it contains the value that should be written to the
 * 16-bit word in the tunnel configuration entry.  The 16-bit word is a
 * value assigned by the ingress list handler and may change when a prefix
 * is added or deleted, and must be zero only if there are no prefixes
 * allowed for the tunnel.  The interpretation of non-zero values is format-
 * specific.  For the trie format, it represents the number of prefixes
 * allowed for the tunnel.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_DUPLICATE_ENTRY - entry already exists 
 * IX_CC_ERROR_FULL - cannot add any more entries 
 * IX_CC_ERROR_OOM - memory allocation failed 
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_isl_trie_add_prefix(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_uint16 arg_TunnelIndex,
    ix_cc_v6v4_ingress_source_entry* arg_pEntry,
    ix_uint16* arg_pTunnelInfo
    )
{
    ix_cc_v6v4_ingress_list_trie_context* pContext;
    ix_cc_v6v4_isl_trie_update updateInfo;
    ix_cc_v6v4_isl_index_update indexInfo[4];
    ix_error err;
    ix_uint32 indexInfoSize;
    ix_uint16 numTunnels = IX_NTOH16(*arg_pTunnelInfo);

    /*
     * Set up the update information structure
     */
    ix_ossl_memset(&updateInfo, 0, sizeof(updateInfo));

    updateInfo.ipAddr = arg_pEntry->ipAddr &
                        (0xffffffff << (32 - arg_pEntry->maskLength));

    if (arg_pEntry->maskLength <= 8)
    {
        updateInfo.prefixOffset = arg_pEntry->maskLength - 1;
        updateInfo.numExpansions =
            IX_CC_V6V4_MAX_BASE_EXPANSIONS >> updateInfo.prefixOffset;
    }
    else
    {
        updateInfo.prefixOffset = (arg_pEntry->maskLength - 1) % 4;
        updateInfo.numExpansions =
            IX_CC_V6V4_MAX_EXPANSIONS >> updateInfo.prefixOffset;
        updateInfo.numTries = ((arg_pEntry->maskLength - 9) / 4) + 1;
    }

    indexInfoSize = updateInfo.numExpansions *
                    sizeof(ix_cc_v6v4_isl_index_update);

    if (updateInfo.numExpansions > 4)
    {
        updateInfo.paIndexInfo =
            (ix_cc_v6v4_isl_index_update*)ix_ossl_malloc(indexInfoSize);

        if (updateInfo.paIndexInfo == NULL)
        {
            return IX_ERROR_WARNING(
                IX_CC_ERROR_OOM,
                ("Cannot allocate memory required to complete operation"));
        }
    }
    else
    {
        updateInfo.paIndexInfo = indexInfo;
    }

    ix_ossl_memset(updateInfo.paIndexInfo, 0, indexInfoSize);

    /*
     * Allocate any new trie blocks required to add this prefix.
     */
    IX_ERROR_CG(
        _ix_cc_v6v4_isl_alloc_tries(arg_pData, &updateInfo),
        err,
        cleanupInfo
    );

    /*
     * Allocate any new index blocks and trie info blocks that are needed.
     */
    IX_ERROR_CG(
        _ix_cc_v6v4_isl_alloc_index_blocks(
            arg_pData, arg_TunnelIndex, &updateInfo),
        err,
        freeTries
    );

    /*
     * Update the index blocks and trie info blocks.
     */
    _ix_cc_v6v4_isl_add_tunnel(
        arg_pData, arg_TunnelIndex, &updateInfo);

    /*
     * Update the trie blocks.
     */
    if (updateInfo.numNewTries != 0)
    {
        _ix_cc_v6v4_isl_link_tries(arg_pData, &updateInfo);
    }

    /* Store the number of prefixes allowed for this tunnel in the
     * tunnel entry.
     */
    *arg_pTunnelInfo = IX_HTON16((ix_uint16)(numTunnels + 1));

    err = IX_SUCCESS;
    goto cleanupInfo;

freeTries:
    pContext = (ix_cc_v6v4_ingress_list_trie_context*)arg_pData->pIslContext;
    while (updateInfo.numNewTries > 0)
    {
        updateInfo.numTries--;
        ix_cc_index_pool_free(
            pContext->trieBlockPool, updateInfo.aTries[updateInfo.numTries]);
        updateInfo.numNewTries--;
    }

cleanupInfo:
    if (updateInfo.numExpansions > 4)
    {
        ix_ossl_free(updateInfo.paIndexInfo);
    }

    return err;
} /* end - _ix_cc_v6v4_isl_trie_add_prefix */


/**
 * NAME: _ix_cc_v6v4_isl_trie_del_prefix
 *
 * DESCRIPTION: This primitive is used to delete a prefix from the ingress
 * list for a tunnel.  The caller must have already validated the tunnel
 * index.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_TunnelIndex - end tunnel index (not used for this
 * ISL format).
 *
 * @Param:  - IN arg_pEntry - pointer to ingress list entry.
 *
 * @Param:  - IN/OUT arg_pTunnelInfo - on input, this points to the 16-bit
 * word in the tunnel configuration entry pertaining to the ingress
 * list.  On output, it contains the value that should be written to the
 * 16-bit word in the tunnel configuration entry.  The 16-bit word is a
 * value assigned by the ingress list handler and may change when a prefix
 * is added or deleted, and must be zero only if there are no prefixes
 * allowed for the tunnel.  The interpretation of non-zero values is format-
 * specific.  For the trie format, it represents the number of prefixes
 * allowed for the tunnel.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - cannot find specified entry 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * IX_CC_ERROR_OOM - memory allocation failed 
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_isl_trie_del_prefix(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_uint16 arg_TunnelIndex,
    ix_cc_v6v4_ingress_source_entry* arg_pEntry,
    ix_uint16* arg_pTunnelInfo
    )
{
    ix_cc_v6v4_isl_trie_update updateInfo;
    ix_cc_v6v4_isl_index_update indexInfo[4];
    ix_error err;
    ix_uint32 indexInfoSize;
    ix_uint16 numTunnels = IX_NTOH16(*arg_pTunnelInfo);

    /*
     * Set up the update information structure
     */
    ix_ossl_memset(&updateInfo, 0, sizeof(updateInfo));

    updateInfo.ipAddr = arg_pEntry->ipAddr &
                        (0xffffffff << (32 - arg_pEntry->maskLength));

    if (arg_pEntry->maskLength <= 8)
    {
        updateInfo.prefixOffset = arg_pEntry->maskLength - 1;
        updateInfo.numExpansions =
            IX_CC_V6V4_MAX_BASE_EXPANSIONS >> updateInfo.prefixOffset;
    }
    else
    {
        updateInfo.prefixOffset = (arg_pEntry->maskLength - 1) % 4;
        updateInfo.numExpansions =
            IX_CC_V6V4_MAX_EXPANSIONS >> updateInfo.prefixOffset;
        updateInfo.numTries = ((arg_pEntry->maskLength - 9) / 4) + 1;
    }

    indexInfoSize = updateInfo.numExpansions *
                    sizeof(ix_cc_v6v4_isl_index_update);

    if (updateInfo.numExpansions > 4)
    {
        updateInfo.paIndexInfo =
            (ix_cc_v6v4_isl_index_update*)ix_ossl_malloc(indexInfoSize);

        if (updateInfo.paIndexInfo == NULL)
        {
            return IX_ERROR_WARNING(
                IX_CC_ERROR_OOM,
                ("Cannot allocate memory required to complete operation"));
        }
    }
    else
    {
        updateInfo.paIndexInfo = indexInfo;
    }

    ix_ossl_memset(updateInfo.paIndexInfo, 0, indexInfoSize);

    /*
     * Retrieve the trie blocks associated with this prefix.
     */
    IX_ERROR_CG(
        _ix_cc_v6v4_isl_get_tries(arg_pData, &updateInfo),
        err,
        cleanupInfo
    );

    /*
     * Retrieve the associated index blocks and trie info blocks.
     */
    IX_ERROR_CG(
        _ix_cc_v6v4_isl_get_index_blocks(
            arg_pData, arg_TunnelIndex, &updateInfo),
        err,
        cleanupInfo
    );

    /*
     * Update the index blocks and trie info blocks.
     */
    _ix_cc_v6v4_isl_delete_tunnel(arg_pData, &updateInfo);

    /*
     * Store the number of prefixes allowed for this tunnel in the
     * tunnel entry.
     */
    *arg_pTunnelInfo = IX_HTON16((ix_uint16)(numTunnels - 1));

    err = IX_SUCCESS;

cleanupInfo:
    if (updateInfo.numExpansions > 4)
    {
        ix_ossl_free(updateInfo.paIndexInfo);
    }

    return err;
} /* end - _ix_cc_v6v4_isl_trie_del_prefix */



/**
 * NAME: _ix_cc_v6v4_isl_trie_get_prefixes
 *
 * DESCRIPTION: This primitive is used to retrieve the prefixes allowed for
 * a particular tunnel.  The caller must have already validated the tunnel
 * index.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_TunnelIndex - end tunnel index.
 *
 * @Param:  - IN arg_TunnelInfo - the 16-bit word in the tunnel
 * configuration entry pertaining to the ingress list.
 *
 * @Param:  - IN/OUT arg_pNumEntries - On input, points to the number of
 * entries the buffer pointed to by arg_paEntries can accommodate.  On output,
 * contains the actual number of prefixes in the ingress list for the
 * tunnel.
 *
 * @Param:  - OUT arg_paEntries - pointer to array into which to place the
 * entries.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_BUFFER_TOO_SMALL - buffer too small to contain the complete
 *                                     list.
 * IX_CC_ERROR_INTERNAL - number of entries doesn't match tunnel entry
 * (debug only).
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_isl_trie_get_prefixes(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_uint16 arg_TunnelIndex,
    ix_uint16 arg_TunnelInfo,
    ix_uint32* arg_pNumEntries,
    ix_cc_v6v4_ingress_source_entry* arg_paEntries
    )
{
    ix_error err;
    ix_cc_v6v4_isl_index_update indexInfo;

    ix_ossl_memset(&indexInfo, 0, sizeof(indexInfo));
    indexInfo.maskLength = IX_CC_V6V4_ISL_BASE_BITS;

#if defined(IX_DEBUG)

    err = _ix_cc_v6v4_isl_search_trie(
            arg_pData, arg_TunnelIndex, &indexInfo,
            0, arg_pNumEntries, arg_paEntries);

    if (*arg_pNumEntries != IX_NTOH16(arg_TunnelInfo))
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
               ("Actual number of ISL entries does not match tunnel entry"));
    }

    return err;

#else
    if (*arg_pNumEntries < IX_NTOH16(arg_TunnelInfo))
    {
        *arg_pNumEntries = IX_NTOH16(arg_TunnelInfo);
        return IX_CC_V6V4_ERROR_BUFFER_TOO_SMALL;
    }

    *arg_pNumEntries = IX_NTOH16(arg_TunnelInfo);

    IX_ERROR_CR(
        _ix_cc_v6v4_isl_search_trie(
            arg_pData, arg_TunnelIndex, &indexInfo,
            0, arg_pNumEntries, arg_paEntries));

    return IX_SUCCESS;

#endif

} /* end - _ix_cc_v6v4_isl_trie_get_prefixes */


/**
 * NAME: _ix_cc_v6v4_isl_trie_del_tunnel
 *
 * DESCRIPTION: This primitive is used to delete all prefixes assigned to
 * a tunnel.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_TunnelIndex - end tunnel index.
 *
 * @Param:  - IN arg_pTunnelInfo - the 16-bit word in the tunnel configuration
 * entry pertaining to the ingress list.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_isl_trie_del_tunnel(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_uint16 arg_TunnelIndex,
    ix_uint16 arg_pTunnelInfo
    )
{
    ix_cc_v6v4_isl_index_update indexInfo;
    ix_uint8 empty;
    ix_uint32 numEntries;

    ix_ossl_memset(&indexInfo, 0, sizeof(indexInfo));
    indexInfo.maskLength = IX_CC_V6V4_ISL_BASE_BITS;

#if defined(IX_DEBUG)
    numEntries = 0xffffffff;
#else
    numEntries = IX_NTOH16(arg_pTunnelInfo);
#endif

    IX_ERROR_CR(
        _ix_cc_v6v4_isl_search_and_del_tunnel(
            arg_pData, arg_TunnelIndex, &indexInfo,
            0, &numEntries, &empty));

    return IX_SUCCESS;
} /* end - _ix_cc_v6v4_isl_trie_del_tunnel */



/**
 * NAME: _ix_cc_v6v4_isl_trie_lookup
 *
 * DESCRIPTION: This primitive is used to lookup a prefix in the ingress list
 * to check whether it is allowed for a particular tunnel.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - IN arg_IpAddr - IP source address.
 *
 * @Param:  - IN arg_TunnelIndex - end tunnel index.
 *
 * @Param:  - IN arg_TunnelInfo - the 16-bit word in the tunnel configuration
 * entry pertaining to the ingress list.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - address is not allowed for the tunnel 
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_isl_trie_lookup(
    ix_cc_v6v4_ingress_list_data* arg_pData,
    ix_uint32 arg_IpAddr,
    ix_uint16 arg_TunnelIndex,
    ix_uint16 arg_TunnelInfo)
{
    ix_cc_v6v4_isl_index_update indexInfo;
    ix_error err;
    ix_uint8 index;
    ix_uint16 nextTrie;
    ix_uint8 curTrie = 0;
    ix_uint8 shiftAmount = 24;
    ix_uint16 indexBlock;

    ix_ossl_memset(&indexInfo, 0, sizeof(indexInfo));

    /*
     * Read the base entry and check the associated index block, if any.
     */
    index = (ix_uint8)(arg_IpAddr >> shiftAmount);

    indexBlock =
        IX_CC_V6V4_ISL_READ_BASE_TRIE_INDEX(arg_pData->pMbIslBase, index);

    if (indexBlock != 0)
    {
        indexInfo.indexBlock = IX_NTOH16(indexBlock);
        err = _ix_cc_v6v4_isl_find_tunnel(
                arg_pData, arg_TunnelIndex, 0, &indexInfo);
        if (err == IX_SUCCESS)
        {
            return IX_SUCCESS;
        }

        if (IX_ERROR_GET_CODE(err) != IX_CC_ERROR_ENTRY_NOT_FOUND)
        {
            return err;
        }
    }

    /* Couldn't find the tunnel in the base, so continue looking at
     * the next trie
     */
    nextTrie = IX_CC_V6V4_ISL_READ_BASE_TRIE_NEXT(arg_pData->pMbIslBase, index);

    while ((nextTrie != 0) && (curTrie < IX_CC_V6V4_ISL_MAX_CHAIN))
    {
        nextTrie = IX_NTOH16(nextTrie);

        shiftAmount -= 4;
        index = (ix_uint8)(arg_IpAddr >> shiftAmount) & 0xf;

        indexBlock = IX_CC_V6V4_ISL_READ_TRIE_INDEX(
            arg_pData->pMbIslBase, nextTrie, index);

        if (indexBlock != 0)
        {
            indexInfo.indexBlock = IX_NTOH16(indexBlock);
            err = _ix_cc_v6v4_isl_find_tunnel(
                    arg_pData, arg_TunnelIndex, 0, &indexInfo);
            if (err == IX_SUCCESS)
            {
                return IX_SUCCESS;
            }

            if (IX_ERROR_GET_CODE(err) != IX_CC_ERROR_ENTRY_NOT_FOUND)
            {
                return err;
            }
        }

        nextTrie = IX_CC_V6V4_ISL_READ_TRIE_NEXT(
            arg_pData->pMbIslBase, nextTrie, index);
        curTrie++;

    } /* end - for */

    return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND, ("Entry not found in ingress list"));

} /* end - _ix_cc_v6v4_isl_trie_lookup */


/**
 * NAME: _ix_cc_v6v4_isl_trie_setup
 *
 * DESCRIPTION: This primitive is used to setup the ingress source list
 * (default format). This includes allocating all shared memory and internal
 * memory used to manage the list.
 *           
 * @Param:  - IN None.
 *
 * @Param:  - OUT arg_pData - pointer to location in which to return the
 * return pointers to the allocated resources.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_OOR - out of resources 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_isl_trie_setup(
    ix_cc_v6v4_ingress_list_data* arg_pData)
{
    ix_error err;
    ix_uint32 memSize;
    ix_uint32 numTrieInfo;
    ix_cc_v6v4_ingress_list_trie_context* pContext;

    /*
     * Allocate the context data.
     */

    pContext = (ix_cc_v6v4_ingress_list_trie_context*)
        ix_ossl_malloc(sizeof(ix_cc_v6v4_ingress_list_trie_context));

    if (pContext == NULL)
    {
        return IX_ERROR_WARNING(
            IX_CC_ERROR_OOM,
            ("Failed to allocate memory for ingress source list context"));
    }

    /*
     * Allocate the trie info block array.
     */

    pContext->trieInfoSize = IX_CC_V6V4_ISL_TRIE_INFO_TOTAL_SIZE(
        g_V6v4RegistryData.ingressListFormat);
    numTrieInfo = IX_CC_V6V4_ISL_NUM_TRIE_INFO_BLOCKS(
        g_V6v4RegistryData.ingressListSizeHint);

    memSize = numTrieInfo * pContext->trieInfoSize;
    pContext->pTrieInfoBase = ix_ossl_malloc(memSize);

    if (pContext->pTrieInfoBase == NULL)
    {
        err = IX_ERROR_WARNING(
            IX_CC_ERROR_OOM,
            ("Failed to allocate memory for ingress source list trie info"));
        goto cleanupContext;
    }

    ix_ossl_memset(pContext->pTrieInfoBase, 0, memSize);
    ix_ossl_memset(&pContext->aTrieInfoHash, 0, sizeof(pContext->aTrieInfoHash));

    /* 
     * Calculate the amount of shared memory we need.  The memory consists
     * of the 1K base block, several 64-byte trie blocks and several index
     * blocks to hold lists of tunnel indexes 
     */

    pContext->numTries = IX_CC_V6V4_ISL_NUM_TRIE_BLOCKS(
        g_V6v4RegistryData.ingressListSizeHint);
    pContext->numIndexBlocks = IX_CC_V6V4_ISL_NUM_INDEX_BLOCKS(
        g_V6v4RegistryData.ingressListSizeHint);
    pContext->indexBlockSize = IX_CC_V6V4_ISL_INDEX_BLOCK_SIZE(
        g_V6v4RegistryData.ingressListFormat);

    memSize = IX_CC_V6V4_ISL_TRIE_BASE_SIZE +
              (IX_CC_V6V4_ISL_TRIE_SIZE * pContext->numTries) +
              (pContext->indexBlockSize * pContext->numIndexBlocks);

    IX_ERROR_CG(
        _ix_cc_v6v4_alloc_shared_memory(
                memSize,
                g_V6v4RegistryData.ingressListMemType,
                g_V6v4RegistryData.ingressListChannel,
                (void**)&arg_pData->pMbIslBase),
        err,
        cleanupTrieInfo;
    );

    /*
     * Calculate the base address of the index blocks and save in
     * the context data
     */

    pContext->pIndexBase = arg_pData->pMbIslBase +
        IX_CC_V6V4_ISL_TRIE_BASE_ENTRIES +
        (pContext->numTries * IX_CC_V6V4_ISL_TRIE_ENTRIES);

    /*
     * Allocate the index pools for the trie blocks, index blocks and
     * trie info blocks.
     */

    IX_ERROR_CG(
        ix_cc_index_pool_init(
            pContext->numTries, 1, &pContext->trieBlockPool),
        err,
        cleanupSharedMem);

    IX_ERROR_CG(
        ix_cc_index_pool_init(
            pContext->numIndexBlocks, 1, &pContext->indexBlockPool),
        err,
        cleanupTrieBlockPool);

    IX_ERROR_CG(
        ix_cc_index_pool_init(
            numTrieInfo, 0, &pContext->trieInfoPool),
        err,
        cleanupIndexBlockPool);

    arg_pData->pIslContext = (void*)pContext;

    return IX_SUCCESS;

cleanupIndexBlockPool:
    ix_cc_index_pool_fini(pContext->indexBlockPool);
cleanupTrieBlockPool:
    ix_cc_index_pool_fini(pContext->trieBlockPool);
cleanupSharedMem:

    ix_rm_mem_free(arg_pData->pMbIslBase);
cleanupTrieInfo:
    ix_ossl_free(pContext->pTrieInfoBase);
cleanupContext:
    ix_ossl_free(pContext);

    return err;

}/* end _ix_cc_v6v4_isl_trie_setup */


/**
 * NAME: _ix_cc_v6v4_isl_trie_cleanup
 *
 * DESCRIPTION: This primitive is used to cleanup the ingress source list.
 * (default format).  The shared memory and internal memory allocated to
 * to manage the list is freed.
 *           
 * @Param:  - In arg_pData - pointer to buffer containing the pointers to the
 * resources that were previously allocated to manage the ingress list.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_isl_trie_cleanup(
    ix_cc_v6v4_ingress_list_data* arg_pData)
{
    ix_cc_v6v4_ingress_list_trie_context* pContext;

    pContext = (ix_cc_v6v4_ingress_list_trie_context*)arg_pData->pIslContext;

    /* Free index pools */

    IX_ERROR_CRT(
        ix_cc_index_pool_fini(pContext->trieInfoPool),
        IX_CC_ERROR_INTERNAL,
        IX_ERROR_LEVEL_WARNING);

    IX_ERROR_CRT(
        ix_cc_index_pool_fini(pContext->indexBlockPool),
        IX_CC_ERROR_INTERNAL,
        IX_ERROR_LEVEL_WARNING);

    IX_ERROR_CRT(
        ix_cc_index_pool_fini(pContext->trieBlockPool),
        IX_CC_ERROR_INTERNAL,
        IX_ERROR_LEVEL_WARNING);

    /* Free shared memory */

    IX_ERROR_CRT(
        ix_rm_mem_free(arg_pData->pMbIslBase),
        IX_CC_ERROR_INTERNAL,
        IX_ERROR_LEVEL_LOCAL);

    /* Free trie info array */

    ix_ossl_free(pContext->pTrieInfoBase);

    /* Free context */

    ix_ossl_free(pContext);

    return IX_SUCCESS;

}/* end _ix_cc_v6v4_isl_trie_cleanup */




/**
 * NAME: _ix_cc_v6v4_ingress_list_init_trie_handlers
 *
 * DESCRIPTION: This primitive is used to initialize the ingress list
 * handlers for the default format.
 *           
 * @Param:  - IN None.
 *
 * @Param:  - OUT arg_pHandlers - pointer to handler list to be
 * initialized.
 *
 * @Return: IX_SUCCESS.
 *
 * Error Codes:
 * 
 */

ix_error _ix_cc_v6v4_ingress_list_init_trie_handlers(
    ix_cc_v6v4_ingress_list_handlers* arg_pHandlers)
{
    arg_pHandlers->setup = _ix_cc_v6v4_isl_trie_setup;
    arg_pHandlers->cleanup = _ix_cc_v6v4_isl_trie_cleanup;
    arg_pHandlers->addPrefix = _ix_cc_v6v4_isl_trie_add_prefix;
    arg_pHandlers->delPrefix = _ix_cc_v6v4_isl_trie_del_prefix;
    arg_pHandlers->getPrefixes = _ix_cc_v6v4_isl_trie_get_prefixes;
    arg_pHandlers->delTunnel = _ix_cc_v6v4_isl_trie_del_tunnel;
    arg_pHandlers->lookup = _ix_cc_v6v4_isl_trie_lookup;

    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_ingress_list_init_trie_handlers */


#if defined(IX_DEBUG)
extern ix_cc_v6v4_context* g_pV6v4Context;
ix_error ix_cc_v6v4_isl_dump_used_tries(void)
{
    ix_cc_v6v4_ingress_list_trie_context* pContext;
    ix_uint16 index;
    int inUse;
    ix_uint16 total = 0;

    pContext = (ix_cc_v6v4_ingress_list_trie_context*)
        g_pV6v4Context->islInfo.pIslContext;

    for (index = 1; index <= pContext->numTries; index++)
    {
        IX_ERROR_CRT(
            ix_cc_index_pool_in_use(pContext->trieBlockPool, index, &inUse),
            IX_CC_ERROR_INTERNAL,
            IX_ERROR_LEVEL_WARNING);

        if (inUse)
        {
            ix_ossl_message_log("Trie block %d in use\n", index);
            total++;
        }
    }

    ix_ossl_message_log("Total trie blocks in use = %d\n", total);

    return IX_SUCCESS;

}

ix_error ix_cc_v6v4_isl_dump_used_index_blocks(void)
{
    ix_cc_v6v4_ingress_list_trie_context* pContext;
    ix_uint16 index;
    int inUse;
    ix_uint16 total = 0;

    pContext = (ix_cc_v6v4_ingress_list_trie_context*)
        g_pV6v4Context->islInfo.pIslContext;

    for (index = 1; index <= pContext->numIndexBlocks; index++)
    {
        IX_ERROR_CRT(
            ix_cc_index_pool_in_use(pContext->indexBlockPool, index, &inUse),
            IX_CC_ERROR_INTERNAL,
            IX_ERROR_LEVEL_WARNING);

        if (inUse)
        {
            ix_ossl_message_log("Index block %d in use\n", index);
            total++;
        }
    }

    ix_ossl_message_log("Total index blocks in use = %d\n", total);

    return IX_SUCCESS;

}

ix_error ix_cc_v6v4_isl_dump_used_trie_info_blocks(void)
{
    ix_cc_v6v4_ingress_list_trie_context* pContext;
    ix_uint16 index;
    int inUse;
    ix_uint16 total = 0;
    ix_uint32 numTrieInfo = IX_CC_V6V4_ISL_NUM_TRIE_INFO_BLOCKS(
        g_V6v4RegistryData.ingressListSizeHint);

    pContext = (ix_cc_v6v4_ingress_list_trie_context*)
        g_pV6v4Context->islInfo.pIslContext;

    for (index = 0; index < numTrieInfo; index++)
    {
        IX_ERROR_CRT(
            ix_cc_index_pool_in_use(pContext->trieInfoPool, index, &inUse),
            IX_CC_ERROR_INTERNAL,
            IX_ERROR_LEVEL_WARNING);

        if (inUse)
        {
            ix_ossl_message_log("Trie info block %d in use\n", index);
            total++;
        }
    }

    ix_ossl_message_log("Total trie info blocks in use = %d\n", total);

    return IX_SUCCESS;

}
#endif
