/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 6
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      IPv6/v4 Tunneling Core Component - library primitives
 *
 * = FILENAME
 *      ix_cc_v6v4_library.c
 *
 * = DESCRIPTION
 *      The file defines library interface functions.
 *
 * = CREATION TIME
 *      11/19/2002 8:48 AM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */


#define IX_ERROR_FILE_IDENT "$Id: ix_cc_v6v4_library.c,v 1.7 2003/11/01 02:34:29 rranjeet Exp $";



/**
 * System defined include files
 */


/**
 * User defined include files
 */

#include "ix_cc_error.h"
#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_netmacros.h"

#include "ix_cc.h"
#include "bindings.h"
#include "cc/ix_cc_v6v4.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_msg_handler.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_prop.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_init.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_config.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_validate.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_tunnel_prop.h"


/**
 * Preprocessor symbols and Macros used in this file
 */

/**
 * Type definitions whose scope is limited to this file
 */

extern ix_cc_v6v4_context* g_pV6v4Context;
extern ix_cc_v6v4_registry_data g_V6v4RegistryData;

/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */


/** 
 * Extern function prototypes.
 */


/** 
 * Static function prototypes.
 */


/** 
 * Function definitions.
 */


/**
 * NAME: ix_cc_v6v4_add_end_tunnel
 *
 * DESCRIPTION:This library primitive adds a new end tunnel endpoint. 
 * 
 * @Param:  - IN arg_pConfig: pointer to configuration information.
 * @Param:  - IN arg_pContext: pointer to tunneling context
 *
 * @Param:  - OUT arg_phTunnel: pointer to location in which to return
 * tunneling component's handle to tunnel.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_ERROR_FULL - out of resources
 * IX_CC_ERROR_V6V4_INVALID_INPUT_PARAM - input parameter is invalid
 * 
 */

ix_error ix_cc_v6v4_add_end_tunnel(
    ix_cc_v6v4_end_tunnel_config* arg_pConfig,
    void* arg_pContext,
    ix_cc_v6v4_tunnel_handle* arg_phTunnel
    )
{
    ix_cc_v6v4_context* pContext;
    ix_uint32 index;
    ix_cc_v6v4_end_tunnel tunnelEntry;
    ix_error err;

    if (arg_pConfig == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pConfig"));
    }/* end if(arg_pConfig == NULL) */

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (arg_phTunnel == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_phTunnel"));
    }/* end if(arg_phTunnel == NULL) */

    /* Validate the flags */
    if (!IX_CC_V6V4_END_TUNNEL_FLAGS_VALID(arg_pConfig->flags))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                ("Invalid end tunnel flags"));
    }

    tunnelEntry.ingressList = 0;
    tunnelEntry.flags = IX_HTON16(
        (ix_uint16)(arg_pConfig->flags | IX_CC_V6V4_FLAG_TUNNEL_VALID));

    pContext = (ix_cc_v6v4_context*)arg_pContext;

    /* Allocate an index */
    ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

    IX_ERROR_CG(
        ix_cc_index_pool_alloc(pContext->endTunnelIndexPool, &index),
        err, indexError);

    ix_ossl_mutex_unlock(pContext->v6v4Mutex);

    /* Write the entry to shared memory */
    _WRITE_END_TUNNEL(pContext->pMbEndTunnelBase, index, tunnelEntry);

    *arg_phTunnel = IX_CC_V6V4_MAKE_END_TUNNEL_HANDLE((ix_uint16)index);

    return IX_SUCCESS;

indexError:
    ix_ossl_mutex_unlock(pContext->v6v4Mutex);
    return err;

} /* end: ix_cc_v6v4_add_end_tunnel */


/**
 * NAME: ix_cc_v6v4_delete_end_tunnel
 *
 * DESCRIPTION:This library primitive deletes an end tunnel endpoint. 
 * 
 * @Param:  - IN arg_hTunnel: Handle to end tunnel.
 * @Param:  - IN arg_pContext: pointer to tunneling context
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_ERROR_V6V4_INVALID_INPUT_PARAM - input parameter is invalid
 * 
 */

ix_error ix_cc_v6v4_delete_end_tunnel(
    ix_cc_v6v4_tunnel_handle arg_hTunnel,
    void* arg_pContext
    )
{
    ix_uint16 index;
    ix_cc_v6v4_end_tunnel tunnelEntry;
    ix_cc_v6v4_context* pContext;
    ix_error err = IX_SUCCESS;

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (!IX_CC_V6V4_END_TUNNEL_HANDLE_VALID(arg_hTunnel))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Invalid parameter: arg_hTunnel"));
    }

    pContext = (ix_cc_v6v4_context*)arg_pContext;

    index = IX_CC_V6V4_INDEX_FROM_HANDLE(arg_hTunnel);

    /* 
     * Free the index.  If it is not valid, this routine will return
     * an error.
     */
    ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

    IX_ERROR_CG(
        ix_cc_index_pool_free(pContext->endTunnelIndexPool, index),
        err, deleteEndTunnel_finish);

    /* Read the entry */

    _READ_END_TUNNEL(pContext->pMbEndTunnelBase, index, tunnelEntry);

    /* Clear the entry in shared memory */

    _CLEAR_END_TUNNEL(pContext->pMbEndTunnelBase, index);

    /* Delete the ingress source list if there is one */
    if (tunnelEntry.ingressList != 0)
    {
        err = (*pContext->ingressListHandlers.delTunnel)(
                &pContext->islInfo, index, tunnelEntry.ingressList);
    }

deleteEndTunnel_finish:
    ix_ossl_mutex_unlock(pContext->v6v4Mutex);
    return err;

} /* end: ix_cc_v6v4_delete_end_tunnel */


/**
 * NAME: ix_cc_v6v4_set_decap_tos_option
 *
 * DESCRIPTION:This library primitive sets the TOS option for decapsulation.
 * 
 * @Param:  - IN arg_hTunnel: Handle to end tunnel.
 * @Param:  - IN arg_Selection: IX_CC_V6V4_SELECT to select the option or
 * IX_CC_V6V4_CLEAR to deselect the option.
 * @Param:  - IN arg_pContext: pointer to tunneling context
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_ERROR_V6V4_INVALID_INPUT_PARAM - input parameter is invalid
 * 
 */

ix_error ix_cc_v6v4_set_decap_tos_option(
    ix_cc_v6v4_tunnel_handle arg_hTunnel,
    ix_uint8 arg_Selection,
    void* arg_pContext
    )
{
    ix_uint16 index;
    ix_cc_v6v4_end_tunnel tunnelEntry;
    ix_cc_v6v4_context* pContext;
    int inUse;
    ix_error err = IX_SUCCESS;

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (!IX_CC_V6V4_END_TUNNEL_HANDLE_VALID(arg_hTunnel))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Invalid parameter: arg_hTunnel"));
    }

    pContext = (ix_cc_v6v4_context*)arg_pContext;

    index = IX_CC_V6V4_INDEX_FROM_HANDLE(arg_hTunnel);

    ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

    err = ix_cc_index_pool_in_use(
			pContext->endTunnelIndexPool, index, &inUse);

	if (err != IX_SUCCESS)
	{
		if (IX_ERROR_GET_CODE(err) == IX_CC_ERROR_RANGE)
		{
			err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
					("Tunnel handle is invalid"));
		}
		else
		{
			err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
					("Internal error validating tunnel handle"));
		}
		goto decapTos_finish;
	}
    else if (!inUse)
    {
        err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Specified tunnel is not in use"));
        goto decapTos_finish;
    }

    /* Read the entry */

    _READ_END_TUNNEL(pContext->pMbEndTunnelBase, index, tunnelEntry);

    if (arg_Selection == IX_CC_V6V4_SELECT)
    {
        tunnelEntry.flags |= IX_HTON16(IX_CC_V6V4_DECAP_TOS_V6);
    }
    else if (arg_Selection == IX_CC_V6V4_CLEAR)
    {
        tunnelEntry.flags &= IX_HTON16(~IX_CC_V6V4_DECAP_TOS_V6);
    }
    else
    {
        err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Invalid parameter: selection"));
        goto decapTos_finish;
    }

    /* Write the flags to shared memory */
    _WRITE_DECAP_FLAGS(pContext->pMbEndTunnelBase, index, tunnelEntry.flags);

decapTos_finish:
    ix_ossl_mutex_unlock(pContext->v6v4Mutex);
    return err;

} /* end: ix_cc_v6v4_set_decap_tos_option */


/**
 * NAME: ix_cc_v6v4_set_src_validation
 *
 * DESCRIPTION:This library primitive sets the ingress source validation
 * option for decapsulation.
 * 
 * @Param:  - IN arg_hTunnel: Handle to end tunnel.
 * @Param:  - IN arg_Selection: IX_CC_V6V4_SELECT to select the option or
 * IX_CC_V6V4_CLEAR to deselect the option.
 * @Param:  - IN arg_pContext: pointer to tunneling context
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_ERROR_V6V4_INVALID_INPUT_PARAM - input parameter is invalid
 * 
 */

ix_error ix_cc_v6v4_set_src_validation(
    ix_cc_v6v4_tunnel_handle arg_hTunnel,
    ix_uint8 arg_Selection,
    void* arg_pContext
    )
{
    ix_uint16 index;
    ix_cc_v6v4_end_tunnel tunnelEntry;
    ix_cc_v6v4_context* pContext;
    int inUse;
    ix_error err = IX_SUCCESS;

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (!IX_CC_V6V4_END_TUNNEL_HANDLE_VALID(arg_hTunnel))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Invalid parameter: arg_hTunnel"));
    }

    pContext = (ix_cc_v6v4_context*)arg_pContext;

    index = IX_CC_V6V4_INDEX_FROM_HANDLE(arg_hTunnel);

    ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

    err = ix_cc_index_pool_in_use(
			pContext->endTunnelIndexPool, index, &inUse);

	if (err != IX_SUCCESS)
	{
		if (IX_ERROR_GET_CODE(err) == IX_CC_ERROR_RANGE)
		{
			err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
					("Tunnel handle is invalid"));
		}
		else
		{
			err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
					("Internal error validating tunnel handle"));
		}
		goto decapSrcVal_finish;
	}
    else if (!inUse)
    {
        err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Specified tunnel is not in use"));
        goto decapSrcVal_finish;
    }

    /* Read the entry */

    _READ_END_TUNNEL(pContext->pMbEndTunnelBase, index, tunnelEntry);

    if (arg_Selection == IX_CC_V6V4_SELECT)
    {
        tunnelEntry.flags |= IX_HTON16(IX_CC_V6V4_SRC_VALIDATE);
    }
    else if (arg_Selection == IX_CC_V6V4_CLEAR)
    {
        tunnelEntry.flags &= IX_HTON16(~IX_CC_V6V4_SRC_VALIDATE);
    }
    else
    {
        err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Invalid parameter: selection"));
        goto decapSrcVal_finish;
    }

    /* Write the flags to shared memory */
    _WRITE_DECAP_FLAGS(pContext->pMbEndTunnelBase, index, tunnelEntry.flags);

decapSrcVal_finish:
    ix_ossl_mutex_unlock(pContext->v6v4Mutex);
    return err;

} /* end: ix_cc_v6v4_set_src_validation */


/**
 * NAME: ix_cc_v6v4_get_end_tunnel
 *
 * DESCRIPTION:This library primitive returns the configuration information
 * for a specific end tunnel endpoint.
 * 
 * @Param:  - IN arg_pContext: pointer to tunneling context
 *
 * @Param:  - IN/OUT arg_pTunnelInfo: Pointer to buffer to contain tunnel
 * information.  As input, the client initializes arg_pTunnelInfo->hTunnel
 * with the handle to the tunnel it is interested in.  As output, the core
 * component supplies the remaining information in the structure.
 * 
 * @Param:  - OUT - none.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_ERROR_V6V4_INVALID_INPUT_PARAM - input parameter is invalid
 * 
 */

ix_error ix_cc_v6v4_get_end_tunnel(
    ix_cc_v6v4_end_tunnel_info* arg_pTunnelInfo,
    void* arg_pContext
    )
{
    ix_uint16 index;
    ix_cc_v6v4_end_tunnel tunnelEntry;
    ix_cc_v6v4_context* pContext;
    int inUse;
    ix_error err = IX_SUCCESS;

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (arg_pTunnelInfo == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pTunnelInfo"));
    }/* end if(arg_pTunnelInfo == NULL) */

    if (!IX_CC_V6V4_END_TUNNEL_HANDLE_VALID(arg_pTunnelInfo->hTunnel))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
           ("Invalid parameter: arg_pTunnelInfo->arg_hTunnel"));
    }

    pContext = (ix_cc_v6v4_context*)arg_pContext;

    index = IX_CC_V6V4_INDEX_FROM_HANDLE(arg_pTunnelInfo->hTunnel);

    ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

    err = ix_cc_index_pool_in_use(
            pContext->endTunnelIndexPool, index, &inUse);

	if (err != IX_SUCCESS)
	{
		if (IX_ERROR_GET_CODE(err) == IX_CC_ERROR_RANGE)
		{
			err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
					("Tunnel handle is invalid"));
		}
		else
		{
			err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
					("Internal error validating tunnel handle"));
		}
	}
    else if (!inUse)
    {
        err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Specified tunnel is not in use"));
    }
    else
    {
        /* Read the entry */

        _READ_END_TUNNEL(pContext->pMbEndTunnelBase, index, tunnelEntry);

        /* Extract the flags */
        arg_pTunnelInfo->config.flags =
            IX_NTOH16(tunnelEntry.flags) & IX_CC_V6V4_END_TUNNEL_FLAGS_ALL;
    }

    ix_ossl_mutex_unlock(pContext->v6v4Mutex);
    return err;

} /* end: ix_cc_v6v4_get_end_tunnel */


/**
 * NAME: ix_cc_v6v4_dump_end_tunnels
 *
 * DESCRIPTION:This library primitive returns a list of all end tunnel
 * endpoints.
 * 
 * @Param:  - IN arg_pContext: pointer to tunneling context
 *
 * @Param:  - IN/OUT arg_NumEntries: On input, the maximum number of entries
 * that can be written to the output buffer.  On output, the number of actual
 * end tunnel endpoints.
 *
 * @Param:  - OUT arg_paInfo: pointer to buffer into which to return the array
 * of end tunnel endpoints.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_ERROR_V6V4_BUFFER_TOO_SMALL - buffer too small
 * 
 */

ix_error ix_cc_v6v4_dump_end_tunnels(
    ix_uint32* arg_pNumEntries,
    ix_cc_v6v4_end_tunnel_info * arg_paInfo,
    void* arg_pContext
    )
{
    ix_uint32 index;
    ix_uint16 listIndex = 0;
    ix_cc_v6v4_end_tunnel tunnelEntry;
    ix_cc_v6v4_context* pContext;
    ix_error err = IX_SUCCESS;

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (arg_pNumEntries == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pNumEntries"));
    }/* end if(arg_pNumEntries == NULL) */

    if (arg_paInfo == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_paInfo"));
    }/* end if(arg_paInfo == NULL) */


    pContext = (ix_cc_v6v4_context*)arg_pContext;

    ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

    err = ix_cc_index_pool_get_first(pContext->endTunnelIndexPool, &index);
    while (err == IX_SUCCESS)
    {
        if (listIndex < *arg_pNumEntries)
        {
            /* Read the entry */

            _READ_END_TUNNEL(pContext->pMbEndTunnelBase, index, tunnelEntry);

            arg_paInfo[listIndex].hTunnel =
                IX_CC_V6V4_MAKE_END_TUNNEL_HANDLE(index);

            arg_paInfo[listIndex].config.flags =
                IX_NTOH16(tunnelEntry.flags) & IX_CC_V6V4_END_TUNNEL_FLAGS_ALL;
        }

        listIndex++;
        err = ix_cc_index_pool_get_next(
            pContext->endTunnelIndexPool, index, &index);
    }

    /*
     * If we successfully retrieved all the tunnels, the error code
     * should be IX_CC_ERROR_ENTRY_NOT_FOUND.  If not, something
     * unexpected has occurred.
     */
    if (IX_ERROR_GET_CODE(err) != IX_CC_ERROR_ENTRY_NOT_FOUND)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
              ("Unexpected error while reading end tunnel indexes"));
    }
    else
    {
        err = (listIndex > *arg_pNumEntries) ?
            IX_ERROR_WARNING(IX_CC_V6V4_ERROR_BUFFER_TOO_SMALL,
                         ("Buffer too small")) :
            IX_SUCCESS;
        *arg_pNumEntries = listIndex;
    }

    ix_ossl_mutex_unlock(pContext->v6v4Mutex);
    return err;

} /* end: ix_cc_v6v4_dump_end_tunnels */


/**
 * NAME: ix_cc_v6v4_add_allowed_source
 *
 * DESCRIPTION:This library primitive adds an entry to the ingress source
 * validation list for an end tunnel endpoint.
 * 
 * @Param:  - IN arg_hTunnel: Handle to end tunnel.
 * @Param:  - IN arg_pEntry: pointer to buffer containing the address and
 * prefix length to add to the ingress list.
 * @Param:  - IN arg_pContext: pointer to tunneling context
 *
 * @Param:  - OUT - none.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_ERROR_OOR - out of resources
 * IX_CC_ERROR_V6V4_INVALID_INPUT_PARAM - input parameter is invalid
 * 
 */

ix_error ix_cc_v6v4_add_allowed_source(
    ix_cc_v6v4_tunnel_handle arg_hTunnel,
    ix_cc_v6v4_ingress_source_entry* arg_pEntry,
    void* arg_pContext
    )
{
    ix_uint16 index;
    ix_cc_v6v4_end_tunnel tunnelEntry;
    ix_cc_v6v4_context* pContext;
    ix_uint16 tunnelIsl;
    int inUse;
    ix_error err = IX_SUCCESS;

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (arg_pEntry == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pEntry"));
    }/* end if(arg_pEntry == NULL) */

    if (!IX_CC_V6V4_END_TUNNEL_HANDLE_VALID(arg_hTunnel))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Invalid parameter: arg_hTunnel"));
    }

    /* Validate the source address and mask length */

    IX_ERROR_CR(_ix_cc_v6v4_validate_ipv4_addr(arg_pEntry->ipAddr, 0));
    IX_ERROR_CHECK_ARG_RANGE(2, arg_pEntry->maskLength, 1, 32, "%d");

    pContext = (ix_cc_v6v4_context*)arg_pContext;

    index = IX_CC_V6V4_INDEX_FROM_HANDLE(arg_hTunnel);

    ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

    err = ix_cc_index_pool_in_use(
			pContext->endTunnelIndexPool, index, &inUse);

	if (err != IX_SUCCESS)
	{
		if (IX_ERROR_GET_CODE(err) == IX_CC_ERROR_RANGE)
		{
			err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
					("Tunnel handle is invalid"));
		}
		else
		{
			err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
					("Internal error validating tunnel handle"));
		}
		goto addAllowedSrc_finish;
	}
    else if (!inUse)
    {
        err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Specified tunnel is not in use"));
        goto addAllowedSrc_finish;
    }

    /* Read the tunnel entry */

    _READ_END_TUNNEL(pContext->pMbEndTunnelBase, index, tunnelEntry);

    tunnelIsl = tunnelEntry.ingressList;

    /* Call the format-specific handler to add the prefix to the list */

    IX_ERROR_CG(
        (*pContext->ingressListHandlers.addPrefix)(
            &pContext->islInfo, index, arg_pEntry, &tunnelIsl),
        err, addAllowedSrc_finish);

    /* Update the tunnel entry if necessary */

    if (tunnelIsl != tunnelEntry.ingressList)
    {
        _WRITE_DECAP_INGRESS_LIST(
            pContext->pMbEndTunnelBase, index, tunnelIsl);
    }

addAllowedSrc_finish:
    ix_ossl_mutex_unlock(pContext->v6v4Mutex);
    return err;

} /* end: ix_cc_v6v4_add_allowed_source */


/**
 * NAME: ix_cc_v6v4_delete_allowed_source
 *
 * DESCRIPTION:This library primitive deletes an entry from the ingress source
 * validation list for an end tunnel endpoint.
 * 
 * @Param:  - IN arg_hTunnel: Handle to end tunnel.
 * @Param:  - IN arg_pEntry: pointer to buffer containing the address and
 * prefix length to delete from the ingress list.
 * @Param:  - IN arg_pContext: pointer to tunneling context
 *
 * @Param:  - OUT - none.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_ERROR_V6V4_INVALID_INPUT_PARAM - input parameter is invalid
 * 
 */

ix_error ix_cc_v6v4_delete_allowed_source(
    ix_cc_v6v4_tunnel_handle arg_hTunnel,
    ix_cc_v6v4_ingress_source_entry* arg_pEntry,
    void* arg_pContext
    )
{
    ix_uint16 index;
    ix_cc_v6v4_end_tunnel tunnelEntry;
    ix_cc_v6v4_context* pContext;
    ix_uint16 tunnelIsl;
    int inUse;
    ix_error err = IX_SUCCESS;

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (arg_pEntry == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pEntry"));
    }/* end if(arg_pEntry == NULL) */

    if (!IX_CC_V6V4_END_TUNNEL_HANDLE_VALID(arg_hTunnel))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Invalid parameter: arg_hTunnel"));
    }

    /* Validate the source address and mask length */

    IX_ERROR_CR(_ix_cc_v6v4_validate_ipv4_addr(arg_pEntry->ipAddr, 0));
    IX_ERROR_CHECK_ARG_RANGE(2, arg_pEntry->maskLength, 1, 32, "%d");

    pContext = (ix_cc_v6v4_context*)arg_pContext;

    index = IX_CC_V6V4_INDEX_FROM_HANDLE(arg_hTunnel);

    ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

    err = ix_cc_index_pool_in_use(
			pContext->endTunnelIndexPool, index, &inUse);

	if (err != IX_SUCCESS)
	{
		if (IX_ERROR_GET_CODE(err) == IX_CC_ERROR_RANGE)
		{
			err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
					("Tunnel handle is invalid"));
		}
		else
		{
			err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
					("Internal error validating tunnel handle"));
		}
		goto delAllowedSrc_finish;
	}
    else if (!inUse)
    {
        err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Specified tunnel is not in use"));
        goto delAllowedSrc_finish;
    }

    /* Read the tunnel entry */

    _READ_END_TUNNEL(pContext->pMbEndTunnelBase, index, tunnelEntry);

    tunnelIsl = tunnelEntry.ingressList;

    /*
     * Call the format-specific handler to remove the prefix from the
     * list for this tunnel
     */

    IX_ERROR_CG(
        (*pContext->ingressListHandlers.delPrefix)(
            &pContext->islInfo, index, arg_pEntry, &tunnelIsl),
        err, delAllowedSrc_finish);

    /* Update the tunnel entry if necessary */

    if (tunnelIsl != tunnelEntry.ingressList)
    {
        _WRITE_DECAP_INGRESS_LIST(
            pContext->pMbEndTunnelBase, index, tunnelIsl);
    }

delAllowedSrc_finish:
    ix_ossl_mutex_unlock(pContext->v6v4Mutex);
    return err;

} /* end: ix_cc_v6v4_delete_allowed_source */


/**
 * NAME: ix_cc_v6v4_clear_allowed_sources
 *
 * DESCRIPTION:This library primitive deletes all entries from the ingress
 * source validation list for an end tunnel endpoint. 
 * 
 * @Param:  - IN arg_hTunnel: Handle to end tunnel.
 * @Param:  - IN arg_pContext: pointer to tunneling context
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_ERROR_V6V4_INVALID_INPUT_PARAM - input parameter is invalid
 * 
 */

ix_error ix_cc_v6v4_clear_allowed_sources(
    ix_cc_v6v4_tunnel_handle arg_hTunnel,
    void* arg_pContext
    )
{
    ix_uint16 index;
    ix_cc_v6v4_end_tunnel tunnelEntry;
    ix_cc_v6v4_context* pContext;
    int inUse;
    ix_error err = IX_SUCCESS;

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (!IX_CC_V6V4_END_TUNNEL_HANDLE_VALID(arg_hTunnel))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Invalid parameter: arg_hTunnel"));
    }

    pContext = (ix_cc_v6v4_context*)arg_pContext;

    index = IX_CC_V6V4_INDEX_FROM_HANDLE(arg_hTunnel);

    ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

    err = ix_cc_index_pool_in_use(
			pContext->endTunnelIndexPool, index, &inUse);

	if (err != IX_SUCCESS)
	{
		if (IX_ERROR_GET_CODE(err) == IX_CC_ERROR_RANGE)
		{
			err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
					("Tunnel handle is invalid"));
		}
		else
		{
			err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
					("Internal error validating tunnel handle"));
		}
		goto clearAllowedSrc_finish;
	}
    else if (!inUse)
    {
        err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Specified tunnel is not in use"));
        goto clearAllowedSrc_finish;
    }

    /* Read the entry */

    _READ_END_TUNNEL(pContext->pMbEndTunnelBase, index, tunnelEntry);

    /* Delete the ingress source list if there is one */
    if (tunnelEntry.ingressList != 0)
    {
        _WRITE_DECAP_INGRESS_LIST(pContext->pMbEndTunnelBase, index, 0);

        err = (*pContext->ingressListHandlers.delTunnel)(
                &pContext->islInfo, index, tunnelEntry.ingressList);
    }

clearAllowedSrc_finish:
    ix_ossl_mutex_unlock(pContext->v6v4Mutex);
    return err;

} /* end: ix_cc_v6v4_clear_allowed_sources */


/**
 * NAME: ix_cc_v6v4_get_allowed_sources
 *
 * DESCRIPTION:This library primitive returns a list of all entries in the
 * ingress source list for an end tunnel endpoint.
 * 
 * @Param:  - IN arg_hTunnel: Handle to end tunnel.
 * @Param:  - IN arg_pContext: pointer to tunneling context
 *
 * @Param:  - IN/OUT arg_pNumEntries: On input, points to the maximum number
 * of entries that can be written to the output buffer.  On output, points to
 * the number of actual entries in the ingress source list for the tunnel.
 *
 * @Param:  - OUT arg_paEntries: pointer to buffer into which to return the
 * array of ingress source list entries.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_ERROR_V6V4_BUFFER_TOO_SMALL - buffer too small
 * IX_CC_ERROR_V6V4_INVALID_INPUT_PARAM - input parameter is invalid
 * 
 */

ix_error ix_cc_v6v4_get_allowed_sources(
    ix_cc_v6v4_tunnel_handle arg_hTunnel,
    ix_uint32* arg_pNumEntries,
    ix_cc_v6v4_ingress_source_entry* arg_paEntries,
    void* arg_pContext
    )
{
    ix_uint16 index;
    ix_cc_v6v4_end_tunnel tunnelEntry;
    ix_cc_v6v4_context* pContext;
    int inUse;
    ix_error err = IX_SUCCESS;

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (arg_pNumEntries == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pNumEntries"));
    }/* end if(arg_pNumEntries == NULL) */

    if (arg_paEntries == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_paEntries"));
    }/* end if(arg_paEntries == NULL) */

    if (!IX_CC_V6V4_END_TUNNEL_HANDLE_VALID(arg_hTunnel))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Invalid parameter: arg_hTunnel"));
    }

    pContext = (ix_cc_v6v4_context*)arg_pContext;

    index = IX_CC_V6V4_INDEX_FROM_HANDLE(arg_hTunnel);

    ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

    err = ix_cc_index_pool_in_use(
			pContext->endTunnelIndexPool, index, &inUse);

	if (err != IX_SUCCESS)
	{
		if (IX_ERROR_GET_CODE(err) == IX_CC_ERROR_RANGE)
		{
			err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
					("Tunnel handle is invalid"));
		}
		else
		{
			err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
					("Internal error validating tunnel handle"));
		}
		goto getAllowedSrc_finish;
	}
    else if (!inUse)
    {
        err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Specified tunnel is not in use"));
        goto getAllowedSrc_finish;
    }

    /* Read the tunnel entry */

    _READ_END_TUNNEL(pContext->pMbEndTunnelBase, index, tunnelEntry);

    /* Call the format-specific handler to get the prefixes */

    err = (*pContext->ingressListHandlers.getPrefixes)(
            &pContext->islInfo, index, tunnelEntry.ingressList,
            arg_pNumEntries, arg_paEntries);

getAllowedSrc_finish:
    ix_ossl_mutex_unlock(pContext->v6v4Mutex);
    return err;

} /* end: ix_cc_v6v4_get_allowed_sources */


/**
 * NAME: ix_cc_v6v4_add_start_tunnel
 *
 * DESCRIPTION:This library primitive adds a new start tunnel endpoint. 
 * 
 * @Param:  - IN arg_pConfig: pointer to configuration information.
 * @Param:  - IN arg_pContext: pointer to tunneling context
 *
 * @Param:  - OUT arg_phTunnel: pointer to location in which to return
 * tunneling component's handle to tunnel.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_ERROR_OOR - out of resources
 * IX_CC_ERROR_V6V4_INVALID_INPUT_PARAM - input parameter is invalid
 * 
 */

ix_error ix_cc_v6v4_add_start_tunnel(
    ix_cc_v6v4_start_tunnel_config* arg_pConfig,
    void* arg_pContext,
    ix_cc_v6v4_tunnel_handle* arg_phTunnel
    )
{
    ix_cc_v6v4_context* pContext;
    ix_uint32 index;
    ix_cc_v6v4_start_tunnel tunnelEntry;
    ix_uint16 flags;
    ix_error err = IX_SUCCESS;
    ix_uint8 trackMtu;
    ix_uint8 trackBroadcast;
    ix_uint32 subnetBroadcast;
    ix_uint16 mtu;

    if (arg_pConfig == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pConfig"));
    }/* end if(arg_pConfig == NULL) */

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (arg_phTunnel == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_phTunnel"));
    }/* end if(arg_phTunnel == NULL) */

    flags = arg_pConfig->flags;

    /* Validate the flags */
    if (!IX_CC_V6V4_START_TUNNEL_FLAGS_VALID(flags))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                ("Invalid start tunnel flags"));
    }

    pContext = (ix_cc_v6v4_context*)arg_pContext;

    /**
     * Set what we can from arg_pConfig -- the rest we have to do after
     * obtaining the lock since we must fetch information from the 
     * property tables.
     */

    tunnelEntry.remoteAddress = IX_HTON32(arg_pConfig->remoteIpv4Address);
    tunnelEntry.ttl = arg_pConfig->ttl;
    tunnelEntry.tos = arg_pConfig->tos;
    tunnelEntry.reserved1 = 0;
    tunnelEntry.reserved2[0] = 0;
    tunnelEntry.reserved2[1] = 0;
    tunnelEntry.reserved2[2] = 0;

    ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

    /* 
     * Get the subnet broadcast address to assist in validation of
     * the remote address.
     */
    IX_ERROR_CG(
        _ix_cc_v6v4_prop_get_port_broadcast_address(
            arg_pConfig->localInterface.bladeId,
            arg_pConfig->localInterface.portId,
            &subnetBroadcast),
        err, addStartTunnel_finish);

    if ((flags & IX_CC_V6V4_ENCAP_AUTO) == 0)
    {
        /* Validate the remote address if this is a configured tunnel */
        IX_ERROR_CG(_ix_cc_v6v4_validate_ipv4_addr(
            arg_pConfig->remoteIpv4Address, IX_NTOH32(subnetBroadcast)),
            err, addStartTunnel_finish);
    }

    if (arg_pConfig->pathMtu == IX_CC_V6V4_PATH_MTU_INTERFACE)
    {
        /* 
         * CC will manage the path MTU, so set it based on the local
         * interface MTU.
         */
        IX_ERROR_CG(
            _ix_cc_v6v4_prop_get_port_mtu(
                arg_pConfig->localInterface.bladeId,
                arg_pConfig->localInterface.portId,
                &mtu),
            err, addStartTunnel_finish);

        tunnelEntry.pathMtu = IX_HTON16(mtu);
        trackMtu = IX_CC_V6V4_SELECT;

        /*
         * Set the NO_DF flag in this case.  Since we will be maintaining
         * the MTU as the link MTU rather than the true path MTU, we
         * must not set the DONT_FRAGMENT flag.
         */
        flags |= IX_CC_V6V4_ENCAP_NO_DF;
    }
    else
    {
        /* 
         * Client will manage the path MTU, so just set it to the indicated
         * value.
         */
        tunnelEntry.pathMtu = IX_HTON16(arg_pConfig->pathMtu);
        trackMtu = IX_CC_V6V4_CLEAR;

        /*
         * Set the NO_DF flag if the tunnel is 6to4, since RFC 3056 specifies
         * DONT_FRAGMENT should not be set.
         */
        if ((flags & IX_CC_V6V4_ENCAP_6TO4) != 0)
        {
            flags |= IX_CC_V6V4_ENCAP_NO_DF;
        }
    }

    /* Now we can set the flags in the entry */

    tunnelEntry.flags =
        IX_HTON16((ix_uint16)(flags | IX_CC_V6V4_FLAG_TUNNEL_VALID));

    if (arg_pConfig->subnetBroadcast == IX_CC_V6V4_BROADCAST_INTERFACE)
    {
        /* 
         * CC will manage the subnet broadcast, so set it based on the local
         * interface broadcast address.
         */
        tunnelEntry.subnetBroadcast = subnetBroadcast;
        trackBroadcast = IX_CC_V6V4_SELECT;
    }
    else
    {
        /* 
         * Client will manage the subnet broadcast, so just set it to the
         * indicated value.
         */
        tunnelEntry.subnetBroadcast = IX_HTON32(arg_pConfig->subnetBroadcast);
        trackBroadcast = IX_CC_V6V4_CLEAR;
    }

    IX_ERROR_CG(
        _ix_cc_v6v4_get_port_ip_address(
            arg_pConfig->localInterface.bladeId,
            arg_pConfig->localInterface.portId,
            &tunnelEntry.localAddress),
        err, addStartTunnel_finish);

    /* Allocate an index */
    IX_ERROR_CG(
        ix_cc_index_pool_alloc(pContext->startTunnelIndexPool, &index),
        err, addStartTunnel_finish);

    /* Write the entry to shared memory */
    _WRITE_START_TUNNEL(pContext->pMbStartTunnelBase, index, tunnelEntry);

    /* Add the tunnel to the list of tunnels configured on the interface */
    IX_ERROR_CG(_ix_cc_v6v4_add_tunnel_to_interface(
                    arg_pConfig->localInterface.bladeId,
                    arg_pConfig->localInterface.portId,
                    (ix_uint16)index,
                    trackMtu,
                    trackBroadcast),
                err,
                clearTunnelEntry);

    *arg_phTunnel = IX_CC_V6V4_MAKE_START_TUNNEL_HANDLE(index);

addStartTunnel_finish:
    ix_ossl_mutex_unlock(pContext->v6v4Mutex);
    return err;


clearTunnelEntry:
    _CLEAR_START_TUNNEL(pContext->pMbStartTunnelBase, index);
    ix_cc_index_pool_free(pContext->startTunnelIndexPool, index);
    ix_ossl_mutex_unlock(pContext->v6v4Mutex);
    return err;

} /* end: ix_cc_v6v4_add_start_tunnel */


/**
 * NAME: ix_cc_v6v4_delete_start_tunnel
 *
 * DESCRIPTION:This library primitive deletes a start tunnel endpoint. 
 * 
 * @Param:  - IN arg_hTunnel: Handle to start tunnel.
 * @Param:  - IN arg_pContext: pointer to tunneling context
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_ERROR_V6V4_INVALID_INPUT_PARAM - input parameter is invalid
 * 
 */

ix_error ix_cc_v6v4_delete_start_tunnel(
    ix_cc_v6v4_tunnel_handle arg_hTunnel,
    void* arg_pContext
    )
{
    ix_uint16 index;
    ix_uint16 bladeId;
    ix_uint16 portId;
    ix_cc_v6v4_context* pContext;
    ix_cc_v6v4_start_tunnel tunnelEntry;
    int inUse;
    ix_error err = IX_SUCCESS;

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (!IX_CC_V6V4_START_TUNNEL_HANDLE_VALID(arg_hTunnel))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Invalid parameter: arg_hTunnel"));
    }

    pContext = (ix_cc_v6v4_context*)arg_pContext;

    index = IX_CC_V6V4_INDEX_FROM_HANDLE(arg_hTunnel);

    ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

    err = ix_cc_index_pool_in_use(
			pContext->startTunnelIndexPool, index, &inUse);

	if (err != IX_SUCCESS)
	{
		if (IX_ERROR_GET_CODE(err) == IX_CC_ERROR_RANGE)
		{
			err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
					("Tunnel handle is invalid"));
		}
		else
		{
			err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
					("Internal error validating tunnel handle"));
		}
		goto delStartTunnel_finish;
	}
    else if (!inUse)
    {
        err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Specified tunnel is not in use"));
        goto delStartTunnel_finish;
    }

    /* Read the entry */

    _READ_START_TUNNEL(pContext->pMbStartTunnelBase, index, tunnelEntry);

    /* Find the interface based on the local address */

    IX_ERROR_CGT(
        _ix_cc_v6v4_prop_table_lookup_ip_address(
            tunnelEntry.localAddress,
            &bladeId, &portId),
        err, IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING,
        delStartTunnel_finish);

    /* Remove the tunnel from the interface */

    IX_ERROR_CG(
        _ix_cc_v6v4_remove_tunnel_from_interface(bladeId, portId, index),
        err, delStartTunnel_finish);

    /* Free the index */

    IX_ERROR_CG(
        ix_cc_index_pool_free(pContext->startTunnelIndexPool, index),
        err, delStartTunnel_finish);

    /* Clear the valid flag */

    tunnelEntry.flags &= IX_HTON16(~IX_CC_V6V4_FLAG_TUNNEL_VALID);
    _WRITE_ENCAP_FLAGS(pContext->pMbStartTunnelBase, index, tunnelEntry.flags);

delStartTunnel_finish:
    ix_ossl_mutex_unlock(pContext->v6v4Mutex);
    return err;

} /* end: ix_cc_v6v4_delete_start_tunnel */


/**
 * NAME: ix_cc_v6v4_set_ttl
 *
 * DESCRIPTION:This library primitive sets the TTL value to be used for
 * encapsulation on a given tunnel.
 * 
 * @Param:  - IN arg_hTunnel: Handle to start tunnel.
 * @Param:  - IN arg_Ttl: Time-to-live value.
 * @Param:  - IN arg_pContext: pointer to tunneling context
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_ERROR_V6V4_INVALID_INPUT_PARAM - input parameter is invalid
 * 
 */

ix_error ix_cc_v6v4_set_ttl(
    ix_cc_v6v4_tunnel_handle arg_hTunnel,
    ix_uint8 arg_Ttl,
    void* arg_pContext
    )
{
    ix_uint16 index;
    ix_cc_v6v4_context* pContext;
    int inUse;
    ix_error err = IX_SUCCESS;

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (!IX_CC_V6V4_START_TUNNEL_HANDLE_VALID(arg_hTunnel))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Invalid parameter: arg_hTunnel"));
    }

    pContext = (ix_cc_v6v4_context*)arg_pContext;

    index = IX_CC_V6V4_INDEX_FROM_HANDLE(arg_hTunnel);

    ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

    err = ix_cc_index_pool_in_use(
			pContext->startTunnelIndexPool, index, &inUse);

	if (err != IX_SUCCESS)
	{
		if (IX_ERROR_GET_CODE(err) == IX_CC_ERROR_RANGE)
		{
			err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
					("Tunnel handle is invalid"));
		}
		else
		{
			err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
					("Internal error validating tunnel handle"));
		}
	}
    else if (!inUse)
    {
        err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Specified tunnel is not in use"));
    }
    else
    {
        /* Write the flags to shared memory */
        _WRITE_ENCAP_TTL(pContext->pMbStartTunnelBase, index, arg_Ttl);
    }

    ix_ossl_mutex_unlock(pContext->v6v4Mutex);
    return err;

} /* end: ix_cc_v6v4_set_ttl */


/**
 * NAME: ix_cc_v6v4_set_encap_tos_option
 *
 * DESCRIPTION:This library primitive specifies the manner in which the
 * TOS byte is encoded during encapsulation for a given tunnel.
 * 
 * @Param:  - IN arg_hTunnel: Handle to start tunnel.
 * @Param:  - IN arg_Selection: value indicating which behavior is to be used.
 * @Param:  - IN arg_pContext: pointer to tunneling context
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_ERROR_V6V4_INVALID_INPUT_PARAM - input parameter is invalid
 * 
 */

ix_error ix_cc_v6v4_set_encap_tos_option(
    ix_cc_v6v4_tunnel_handle arg_hTunnel,
    ix_uint8 arg_Selection,
    void* arg_pContext
    )
{
    ix_uint16 index;
    ix_cc_v6v4_start_tunnel tunnelEntry;
    ix_cc_v6v4_context* pContext;
    int inUse;
    ix_error err = IX_SUCCESS;

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (!IX_CC_V6V4_START_TUNNEL_HANDLE_VALID(arg_hTunnel))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Invalid parameter: arg_hTunnel"));
    }

    pContext = (ix_cc_v6v4_context*)arg_pContext;

    index = IX_CC_V6V4_INDEX_FROM_HANDLE(arg_hTunnel);

    ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

    err = ix_cc_index_pool_in_use(
			pContext->startTunnelIndexPool, index, &inUse);

	if (err != IX_SUCCESS)
	{
		if (IX_ERROR_GET_CODE(err) == IX_CC_ERROR_RANGE)
		{
			err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
					("Tunnel handle is invalid"));
		}
		else
		{
			err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
					("Internal error validating tunnel handle"));
		}
		goto encapTosOpt_finish;
	}
    else if (!inUse)
    {
        err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Specified tunnel is not in use"));
        goto encapTosOpt_finish;
    }

    /* Read the entry */

    _READ_START_TUNNEL(pContext->pMbStartTunnelBase, index, tunnelEntry);

    if (arg_Selection == IX_CC_V6V4_SELECT)
    {
        tunnelEntry.flags |= IX_HTON16(IX_CC_V6V4_ENCAP_TOS_V6);
    }
    else if (arg_Selection == IX_CC_V6V4_CLEAR)
    {
        tunnelEntry.flags &= IX_HTON16(~IX_CC_V6V4_ENCAP_TOS_V6);
    }
    else
    {
        err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Invalid parameter: selection"));
        goto encapTosOpt_finish;
    }

    /* Write the flags to shared memory */
    _WRITE_ENCAP_FLAGS(pContext->pMbStartTunnelBase, index, tunnelEntry.flags);

encapTosOpt_finish:
    ix_ossl_mutex_unlock(pContext->v6v4Mutex);
    return err;

} /* end: ix_cc_v6v4_set_encap_tos_option */


/**
 * NAME: ix_cc_v6v4_set_tos
 *
 * DESCRIPTION:This library primitive specifies the value to use to encode
 * the TOS byte during encapsulation, if the tunnel is not configured to 
 * encode the TOS byte from the IPv6 header.
 * 
 * @Param:  - IN arg_hTunnel: Handle to start tunnel.
 * @Param:  - IN arg_Tos: TOS value.
 * @Param:  - IN arg_pContext: pointer to tunneling context
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_ERROR_V6V4_INVALID_INPUT_PARAM - input parameter is invalid
 * 
 */

ix_error ix_cc_v6v4_set_tos(
    ix_cc_v6v4_tunnel_handle arg_hTunnel,
    ix_uint8 arg_Tos,
    void* arg_pContext
    )
{
    ix_uint16 index;
    ix_cc_v6v4_context* pContext;
    int inUse;
    ix_error err = IX_SUCCESS;

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (!IX_CC_V6V4_START_TUNNEL_HANDLE_VALID(arg_hTunnel))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Invalid parameter: arg_hTunnel"));
    }

    pContext = (ix_cc_v6v4_context*)arg_pContext;

    index = IX_CC_V6V4_INDEX_FROM_HANDLE(arg_hTunnel);

    ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

    err = ix_cc_index_pool_in_use(
			pContext->startTunnelIndexPool, index, &inUse);

	if (err != IX_SUCCESS)
	{
		if (IX_ERROR_GET_CODE(err) == IX_CC_ERROR_RANGE)
		{
			err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
					("Tunnel handle is invalid"));
		}
		else
		{
			err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
					("Internal error validating tunnel handle"));
		}
	}
    else if (!inUse)
    {
        err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Specified tunnel is not in use"));
    }
    else
    {
        /* Write the Tos to shared memory */
        _WRITE_ENCAP_TOS(pContext->pMbStartTunnelBase, index, arg_Tos);
    }

    ix_ossl_mutex_unlock(pContext->v6v4Mutex);
    return err;

} /* end: ix_cc_v6v4_set_tos */



/**
 * NAME: ix_cc_v6v4_set_mtu
 *
 * DESCRIPTION:This library primitive sets the path MTU value for a given
 * tunnel.
 * 
 * @Param:  - IN arg_hTunnel: Handle to start tunnel.
 * @Param:  - IN arg_Mtu: Path MTU value.
 * @Param:  - IN arg_pContext: pointer to tunneling context
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_ERROR_V6V4_INVALID_INPUT_PARAM - input parameter is invalid
 * 
 */

ix_error ix_cc_v6v4_set_mtu(
    ix_cc_v6v4_tunnel_handle arg_hTunnel,
    ix_uint16 arg_Mtu,
    void* arg_pContext
    )
{
    ix_uint16 index;
    ix_cc_v6v4_context* pContext;
    ix_cc_v6v4_start_tunnel tunnelEntry;
    ix_uint16 bladeId;
    ix_uint16 portId;
    ix_uint16 flags;
    ix_uint8 trackMtu;
    int inUse;
    ix_uint16 mtu;
    ix_error err = IX_SUCCESS;

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (!IX_CC_V6V4_START_TUNNEL_HANDLE_VALID(arg_hTunnel))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Invalid parameter: arg_hTunnel"));
    }

    pContext = (ix_cc_v6v4_context*)arg_pContext;

    index = IX_CC_V6V4_INDEX_FROM_HANDLE(arg_hTunnel);

    ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

    err = ix_cc_index_pool_in_use(
			pContext->startTunnelIndexPool, index, &inUse);

	if (err != IX_SUCCESS)
	{
		if (IX_ERROR_GET_CODE(err) == IX_CC_ERROR_RANGE)
		{
			err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
					("Tunnel handle is invalid"));
		}
		else
		{
			err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
					("Internal error validating tunnel handle"));
		}
		goto setMtu_finish;
	}
    else if (!inUse)
    {
        err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Specified tunnel is not in use"));
        goto setMtu_finish;
    }

    /* Read the entry flags */

    _READ_START_TUNNEL(pContext->pMbStartTunnelBase, index, tunnelEntry);

    flags = tunnelEntry.flags;

    IX_ERROR_CGT(
        _ix_cc_v6v4_prop_table_lookup_ip_address(
            tunnelEntry.localAddress, &bladeId, &portId),
        err, IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING, setMtu_finish);

    if (arg_Mtu == IX_CC_V6V4_PATH_MTU_INTERFACE)
    {
        
        /* 
         * CC will manage the path MTU, so set it based on the local
         * interface MTU.
         */
        IX_ERROR_CG(
            _ix_cc_v6v4_prop_get_port_mtu(bladeId, portId, &mtu),
            err, setMtu_finish);

        tunnelEntry.pathMtu = IX_HTON16(mtu);
        flags |= IX_HTON16(IX_CC_V6V4_ENCAP_NO_DF);
        trackMtu = IX_CC_V6V4_SELECT;
    }
    else
    {
        /* 
         * Client will manage the path MTU, so just set it to the indicated
         * value.
         */
        tunnelEntry.pathMtu = IX_HTON16(arg_Mtu);
        trackMtu = IX_CC_V6V4_CLEAR;

        /*
         * Clear the NO_DF flag if the tunnel is not 6to4
         */
        if ((flags & IX_CC_V6V4_ENCAP_6TO4) == 0)
        {
            flags &= IX_HTON16(~IX_CC_V6V4_ENCAP_NO_DF);
        }
    }

    /* Update the interface MTU tracking info for this tunnel */

    IX_ERROR_CG(
        _ix_cc_v6v4_set_track_mtu(bladeId, portId, index, trackMtu),
        err, setMtu_finish);

    /* Write the flags to shared memory if necessary */
    if (flags != tunnelEntry.flags)
    {
        _WRITE_ENCAP_FLAGS(pContext->pMbStartTunnelBase, index, flags);
    }

    /* Write the MTU to shared memory */
    _WRITE_ENCAP_MTU(pContext->pMbStartTunnelBase, index, tunnelEntry.pathMtu);

setMtu_finish:
    ix_ossl_mutex_unlock(pContext->v6v4Mutex);
    return err;

} /* end: ix_cc_v6v4_set_mtu */


/**
 * NAME: ix_cc_v6v4_set_subnet_broadcast
 *
 * DESCRIPTION:This library primitive sets the subnet broadcast address
 * for a given tunnel.
 * 
 * @Param:  - IN arg_hTunnel: Handle to start tunnel.
 * @Param:  - IN arg_SubnetBroadcast: Subnet broadcast address.
 * @Param:  - IN arg_pContext: pointer to tunneling context
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_ERROR_V6V4_INVALID_INPUT_PARAM - input parameter is invalid
 * 
 */

ix_error ix_cc_v6v4_set_subnet_broadcast(
    ix_cc_v6v4_tunnel_handle arg_hTunnel,
    ix_uint32 arg_SubnetBroadcast,
    void* arg_pContext
    )
{
    ix_uint16 index;
    ix_cc_v6v4_context* pContext;
    ix_cc_v6v4_start_tunnel tunnelEntry;
    ix_uint16 bladeId;
    ix_uint16 portId;
    ix_uint8 trackBroadcast;
    int inUse;
    ix_error err = IX_SUCCESS;

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (!IX_CC_V6V4_START_TUNNEL_HANDLE_VALID(arg_hTunnel))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Invalid parameter: arg_hTunnel"));
    }

    pContext = (ix_cc_v6v4_context*)arg_pContext;

    index = IX_CC_V6V4_INDEX_FROM_HANDLE(arg_hTunnel);

    ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

    err = ix_cc_index_pool_in_use(
            pContext->startTunnelIndexPool, index, &inUse);

	if (err != IX_SUCCESS)
	{
		if (IX_ERROR_GET_CODE(err) == IX_CC_ERROR_RANGE)
		{
			err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
					("Tunnel handle is invalid"));
		}
		else
		{
			err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
					("Internal error validating tunnel handle"));
		}
		goto setSubnetBroadcast_finish;
	}
    else if (!inUse)
    {
        err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Specified tunnel is not in use"));
        goto setSubnetBroadcast_finish;
    }

    _READ_START_TUNNEL(pContext->pMbStartTunnelBase, index, tunnelEntry);

    IX_ERROR_CGT(
        _ix_cc_v6v4_prop_table_lookup_ip_address(
            tunnelEntry.localAddress,
            &bladeId, &portId),
        err, IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING,
        setSubnetBroadcast_finish);

    if (arg_SubnetBroadcast == IX_CC_V6V4_BROADCAST_INTERFACE)
    {
        /* 
         * CC will manage the subnet broadcast, so set it based on the local
         * interface MTU.
         */

        IX_ERROR_CG(
            _ix_cc_v6v4_prop_get_port_broadcast_address(
                bladeId, portId,
                &tunnelEntry.subnetBroadcast),
            err, setSubnetBroadcast_finish);

        trackBroadcast = IX_CC_V6V4_SELECT;
    }
    else
    {
        /* 
         * Client will manage the subnet broadcast, so just set it to the
         * indicated value.
         */
        tunnelEntry.subnetBroadcast = IX_HTON32(arg_SubnetBroadcast);
        trackBroadcast = IX_CC_V6V4_CLEAR;
    }

    /* Update the interface broadcast tracking info for this tunnel */

    IX_ERROR_CG(
        _ix_cc_v6v4_set_track_broadcast(bladeId, portId, index, trackBroadcast),
        err, setSubnetBroadcast_finish);

    /* Write the address to shared memory */
    _WRITE_ENCAP_SUBNET_BROADCAST(
        pContext->pMbStartTunnelBase, index, tunnelEntry.subnetBroadcast);

setSubnetBroadcast_finish:
    ix_ossl_mutex_unlock(pContext->v6v4Mutex);
    return err;

} /* end: ix_cc_v6v4_set_subnet_broadcast */


/**
 * NAME: ix_cc_v6v4_get_start_tunnel
 *
 * DESCRIPTION:This library primitive returns the configuration information
 * for a specific start tunnel endpoint.
 * 
 * @Param:  - IN arg_pContext: pointer to tunneling context
 *
 * @Param:  - IN/OUT arg_pTunnelInfo: Pointer to buffer to contain tunnel
 * information.  As input, the client initializes arg_pTunnelInfo->hTunnel
 * with the handle to the tunnel it is interested in.  As output, the core
 * component supplies the remaining information in the structure.
 * 
 * @Param:  - OUT - none.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_ERROR_V6V4_INVALID_INPUT_PARAM - input parameter is invalid
 * 
 */

ix_error ix_cc_v6v4_get_start_tunnel(
    ix_cc_v6v4_start_tunnel_info* arg_pTunnelInfo,
    void* arg_pContext
    )
{
    ix_uint16 index;
    ix_cc_v6v4_start_tunnel tunnelEntry;
    ix_cc_v6v4_context* pContext;
    int inUse;
    ix_error err = IX_SUCCESS;

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (arg_pTunnelInfo == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pTunnelInfo"));
    }/* end if(arg_pTunnelInfo == NULL) */

    if (!IX_CC_V6V4_START_TUNNEL_HANDLE_VALID(arg_pTunnelInfo->hTunnel))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
           ("Invalid parameter: arg_pTunnelInfo->arg_hTunnel"));
    }

    pContext = (ix_cc_v6v4_context*)arg_pContext;

    index = IX_CC_V6V4_INDEX_FROM_HANDLE(arg_pTunnelInfo->hTunnel);

    ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

    err = ix_cc_index_pool_in_use(
			pContext->startTunnelIndexPool, index, &inUse);

	if (err != IX_SUCCESS)
	{
		if (IX_ERROR_GET_CODE(err) == IX_CC_ERROR_RANGE)
		{
			err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
					("Tunnel handle is invalid"));
		}
		else
		{
			err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
					("Internal error validating tunnel handle"));
		}
		goto getStartTunnel_finish;
	}
    else if (!inUse)
    {
        err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                               ("Specified tunnel is not in use"));
        goto getStartTunnel_finish;
    }

    /* Read the entry */

    _READ_START_TUNNEL(pContext->pMbStartTunnelBase, index, tunnelEntry);

    /* Extract the information */
    arg_pTunnelInfo->config.subnetBroadcast =
        IX_NTOH32(tunnelEntry.subnetBroadcast);
    arg_pTunnelInfo->config.pathMtu = IX_NTOH16(tunnelEntry.pathMtu);
    arg_pTunnelInfo->config.flags =
        IX_NTOH16(tunnelEntry.flags) & IX_CC_V6V4_START_TUNNEL_FLAGS_ALL;
    arg_pTunnelInfo->config.ttl = tunnelEntry.ttl;
    arg_pTunnelInfo->config.tos = tunnelEntry.tos;
    arg_pTunnelInfo->config.remoteIpv4Address =
        IX_NTOH32(tunnelEntry.remoteAddress);

    IX_ERROR_CGT(
        _ix_cc_v6v4_prop_table_lookup_ip_address(
            tunnelEntry.localAddress,
            &arg_pTunnelInfo->config.localInterface.bladeId,
            &arg_pTunnelInfo->config.localInterface.portId),
        err, IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING,
        getStartTunnel_finish);

    /* Check whether CC is tracking mtu/bcast for this tunnel */

    IX_ERROR_CGT(
        _ix_cc_v6v4_get_tunnel_track_info(
            arg_pTunnelInfo->config.localInterface.bladeId,
            arg_pTunnelInfo->config.localInterface.portId,
            index,
            &arg_pTunnelInfo->trackMtu,
            &arg_pTunnelInfo->trackBroadcast),
        err, IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING,
        getStartTunnel_finish);

getStartTunnel_finish:
    ix_ossl_mutex_unlock(pContext->v6v4Mutex);
    return err;

} /* end: ix_cc_v6v4_get_start_tunnel */


/**
 * NAME: ix_cc_v6v4_dump_start_tunnels
 *
 * DESCRIPTION:This library primitive returns a list of all start tunnel
 * endpoints.
 * 
 * @Param:  - IN arg_pContext: pointer to tunneling context
 *
 * @Param:  - IN/OUT arg_pNumEntries: On input, points to the maximum number
 * of entries that can be written to the output buffer.  On output, points to
 * the number of actual start tunnel endpoints.
 *
 * @Param:  - OUT arg_paInfo: pointer to buffer into which to return the array
 * of tart tunnel endpoints.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_ERROR_V6V4_BUFFER_TOO_SMALL - buffer too small
 * 
 */

ix_error ix_cc_v6v4_dump_start_tunnels(
    ix_uint32* arg_pNumEntries,
    ix_cc_v6v4_start_tunnel_info * arg_paInfo,
    void* arg_pContext
    )
{
    ix_uint32 index;
    ix_uint16 listIndex = 0;
    ix_cc_v6v4_start_tunnel tunnelEntry;
    ix_cc_v6v4_context* pContext;
    ix_error err = IX_SUCCESS;

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (arg_pNumEntries == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pNumEntries"));
    }/* end if(arg_pNumEntries == NULL) */

    if (arg_paInfo == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_paInfo"));
    }/* end if(arg_paInfo == NULL) */


    pContext = (ix_cc_v6v4_context*)arg_pContext;

    ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

    err = ix_cc_index_pool_get_first(pContext->startTunnelIndexPool, &index);
    while (err == IX_SUCCESS)
    {
        if (listIndex < *arg_pNumEntries)
        {
            /* Read the entry */

            _READ_START_TUNNEL(
                pContext->pMbStartTunnelBase, index, tunnelEntry);

            arg_paInfo[listIndex].hTunnel =
                IX_CC_V6V4_MAKE_START_TUNNEL_HANDLE(index);

            arg_paInfo[listIndex].config.subnetBroadcast =
                IX_NTOH32(tunnelEntry.subnetBroadcast);
            arg_paInfo[listIndex].config.pathMtu =
                IX_NTOH16(tunnelEntry.pathMtu);
            arg_paInfo[listIndex].config.flags =
                IX_NTOH16(tunnelEntry.flags) &
                IX_CC_V6V4_START_TUNNEL_FLAGS_ALL;
            arg_paInfo[listIndex].config.ttl = tunnelEntry.ttl;
            arg_paInfo[listIndex].config.tos = tunnelEntry.tos;
            arg_paInfo[listIndex].config.remoteIpv4Address =
                IX_NTOH32(tunnelEntry.remoteAddress);

            IX_ERROR_CGT(
                _ix_cc_v6v4_prop_table_lookup_ip_address(
                    tunnelEntry.localAddress,
                    &arg_paInfo[listIndex].config.localInterface.bladeId,
                    &arg_paInfo[listIndex].config.localInterface.portId),
                err, IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING,
                dumpStartTunnels_finish);

            /* Check whether CC is tracking mtu/bcast for this tunnel */

            IX_ERROR_CGT(
                _ix_cc_v6v4_get_tunnel_track_info(
                    arg_paInfo[listIndex].config.localInterface.bladeId,
                    arg_paInfo[listIndex].config.localInterface.portId,
                    (ix_uint16)index,
                    &arg_paInfo[listIndex].trackMtu,
                    &arg_paInfo[listIndex].trackBroadcast),
                err, IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING,
                dumpStartTunnels_finish);

        }

        listIndex++;
        err = ix_cc_index_pool_get_next(
            pContext->startTunnelIndexPool, index, &index);
    }

    if (IX_ERROR_GET_CODE(err) != IX_CC_ERROR_ENTRY_NOT_FOUND)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
                               ("Can't locate end tunnel indexes"));
    }
    else
    {
        err = (listIndex > *arg_pNumEntries) ?
            IX_ERROR_WARNING(IX_CC_V6V4_ERROR_BUFFER_TOO_SMALL,
                         ("Buffer too small")) :
            IX_SUCCESS;
        *arg_pNumEntries = listIndex;
    }

dumpStartTunnels_finish:
    ix_ossl_mutex_unlock(pContext->v6v4Mutex);
    return err;

} /* end: ix_cc_v6v4_dump_start_tunnels */


/**
 * NAME: ix_cc_v6v4_get_statistics
 *
 * DESCRIPTION:This library primitive returns the tunneling statistics.
 * 
 * @Param:  - IN arg_pContext: pointer to tunneling context
 *
 * @Param:  - OUT arg_pStats: pointer to buffer into which to return
 * the statistics.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - invalid pointer
 * IX_CC_ERROR_V6V4_INVALID_INPUT_PARAM - input parameter is invalid
 * 
 */

ix_error ix_cc_v6v4_get_statistics(
    ix_cc_v6v4_statistics * arg_pStats,
    void* arg_pContext
    )
{
    ix_uint32* pStats;
    ix_cc_v6v4_context* pContext;
    ix_uint32 value;

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */

    if (arg_pStats == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pStats "));
    }/* end if(arg_pStats == NULL)  */

    pContext = (ix_cc_v6v4_context*)arg_pContext;

    ix_ossl_mutex_lock(pContext->v6v4Mutex, IX_OSSL_WAIT_FOREVER);

    /* Get Microblock statistics */

    pStats = pContext->pMbDecapStatsBase;

    value = IX_RM_MEM_UINT32_READ(pStats);
    arg_pStats->decapMbDropPkts = IX_NTOH32(value);
    value = IX_RM_MEM_UINT32_READ(pStats + 1);
    arg_pStats->decapMbExcpPkts = IX_NTOH32(value);

    pStats = pContext->pMbEncapStatsBase;

    value = IX_RM_MEM_UINT32_READ(pStats);
    arg_pStats->encapMbDropPkts = IX_NTOH32(value);
    value = IX_RM_MEM_UINT32_READ(pStats + 1);
    arg_pStats->encapMbExcpPkts = IX_NTOH32(value);

    /* Get core component statistics */

    arg_pStats->localPkts = pContext->ccStats.localPkts;
    arg_pStats->autoTunnelPkts = pContext->ccStats.autoTunnelPkts;
    arg_pStats->fragmentsRcvd = pContext->ccStats.fragmentsRcvd;
    arg_pStats->reassembledPkts = pContext->ccStats.reassembledPkts;
    arg_pStats->pktTooBigMsgsSent = pContext->ccStats.pktTooBigMsgsSent;
    arg_pStats->pktsEncapsulated = pContext->ccStats.pktsEncapsulated;
    arg_pStats->pktsDecapsulated = pContext->ccStats.pktsDecapsulated;

    ix_ossl_mutex_unlock(pContext->v6v4Mutex);
    return IX_SUCCESS;

} /* end: ix_cc_v6v4_get_statistics */


/**
 * NAME: ix_cc_v6v4_set_property
 *
 * DESCRIPTION: This library primitive helps to set dynamic properties of 
 * the tunneling core component. 
 * 
 * @Param:  - IN arg_PropId: Identifier of a property or properties(ORed)
 * to be updated for a port
 * @Param:  - IN arg_pPropData: pointer to dynamic property data structure
 * @Param:  - IN arg_pContext: pointer to tunneling context
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_NUL - invalid pointer
 * IX_CC_IPV4_ERROR_INVALID_INPUT_PARAM - invalid input param
 * IX_ERROR_FULL - table is full
 * IX_CC_ERROR_ENTRY_NOT_FOUND - entry not found in table
 *
 */
ix_error ix_cc_v6v4_set_property(
    ix_uint32 arg_PropId,
    ix_cc_properties* arg_pPropData, 
    void* arg_pContext
    )
{
    ix_uint32 ipAddr;
    ix_uint32 netMask;

    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("Invalid pointer: arg_pContext"));
    }/* end if(arg_pContext == NULL) */


    /* Validate arg_pPropData */
    if (arg_pPropData == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("Invalid pointer: arg_pPropData"));
    }/* end if (arg_PropData == NULL) */  

    if (((arg_PropId & IX_CC_SET_PROPERTY_ADD_INTERFACE_IPV4) != 
         IX_CC_SET_PROPERTY_ADD_INTERFACE_IPV4) &&
        ((arg_PropId & IX_CC_SET_PROPERTY_DEL_INTERFACE_IPV4) != 
         IX_CC_SET_PROPERTY_DEL_INTERFACE_IPV4) &&
        ((arg_PropId & IX_CC_SET_PROPERTY_ADD_BROADCAST_IPV4) != 
         IX_CC_SET_PROPERTY_ADD_BROADCAST_IPV4) &&
        ((arg_PropId & IX_CC_SET_PROPERTY_DEL_BROADCAST_IPV4) != 
         IX_CC_SET_PROPERTY_DEL_BROADCAST_IPV4) &&
        ((arg_PropId & IX_CC_SET_PROPERTY_MTU) != 
         IX_CC_SET_PROPERTY_MTU))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                ("Invalid property Id"));

    }/* end if invalid property */
  
    /* arg_PropId can have multiple properties */
    if ((arg_PropId & IX_CC_SET_PROPERTY_ADD_INTERFACE_IPV4) == 
        IX_CC_SET_PROPERTY_ADD_INTERFACE_IPV4)
    {
        ix_uint32 loop_var;

        for (loop_var = 0;
             loop_var < arg_pPropData->no_virtual_if;
             loop_var++)
        {
            if (arg_pPropData->ip_prop[loop_var].ip_version
                == IX_CC_IP_VERSION_IPV4)
            {
                /* Convert IP Address into network byte order */
                ipAddr = IX_HTON32(arg_pPropData->ip_prop[loop_var].protocol.ipv4_prop.ipv4_address);
                netMask = IX_HTON32(arg_pPropData->ip_prop[loop_var].protocol.ipv4_prop.ipv4_subnet_mask);
                IX_ERROR_CR(_ix_cc_v6v4_property_update_ip_address(
                                IX_CC_SET_PROPERTY_ADD_INTERFACE_IPV4,
                                (ix_uint16)arg_pPropData->blade_id,
                                (ix_uint16)arg_pPropData->port_id,
                                ipAddr,
                                netMask));
            }
        }

        /* Update all start tunnels configured on this interface */
        IX_ERROR_CR(_ix_cc_v6v4_update_tunnels_ip_addr(
                        (ix_uint16)arg_pPropData->blade_id,
                        (ix_uint16)arg_pPropData->port_id,
                        ipAddr));
    }/* end if add ipv4 address */

    if ((arg_PropId & IX_CC_SET_PROPERTY_DEL_INTERFACE_IPV4) == 
        IX_CC_SET_PROPERTY_DEL_INTERFACE_IPV4)
    {
        ix_uint16 numTunnels;
        ix_uint32 loop_var;

        for (loop_var = 0;
             loop_var < arg_pPropData->no_virtual_if;
             loop_var++)
        {
            if (arg_pPropData->ip_prop[loop_var].ip_version
                == IX_CC_IP_VERSION_IPV4)
            {
                /* Convert IP Address into network byte order */
                ipAddr = IX_HTON32(arg_pPropData->ip_prop[loop_var].protocol.ipv4_prop.ipv4_address);
                netMask = IX_HTON32(arg_pPropData->ip_prop[loop_var].protocol.ipv4_prop.ipv4_subnet_mask);
                IX_ERROR_CR(_ix_cc_v6v4_property_update_ip_address(
                                IX_CC_SET_PROPERTY_DEL_INTERFACE_IPV4,
                                (ix_uint16)arg_pPropData->blade_id,
                                (ix_uint16)arg_pPropData->port_id,
                                ipAddr,
                                netMask));
            }
        }

        /* 
         * Sanity check -- there shouldn't be any tunnels left on the
         * interface if it is being removed
         */
        IX_ERROR_CRT(_ix_cc_v6v4_get_num_tunnels_for_interface(
                        (ix_uint16)arg_pPropData->blade_id,
                        (ix_uint16)arg_pPropData->port_id,
                        &numTunnels),
            IX_CC_ERROR_INTERNAL,
            IX_ERROR_LEVEL_WARNING);

        if (numTunnels != 0)
        {
            return IX_ERROR_WARNING(
                        IX_CC_ERROR_INTERNAL,
                        ("Deleted interface still has tunnels"));
        }
    }/* end if delete v6v4 address */

    if ((arg_PropId & IX_CC_SET_PROPERTY_ADD_BROADCAST_IPV4) == 
        IX_CC_SET_PROPERTY_ADD_BROADCAST_IPV4)
    {
        ix_uint32 loop_var;

        for (loop_var = 0;
             loop_var < arg_pPropData->no_virtual_if;
             loop_var++)
        {
            if (arg_pPropData->ip_prop[loop_var].ip_version
                == IX_CC_IP_VERSION_IPV4)
            {
                /* Convert IP Address into network byte order */
                ipAddr =
                    IX_HTON32(arg_pPropData->ip_prop[loop_var].protocol.ipv4_prop.ipv4_broadcast);
                IX_ERROR_CR(_ix_cc_v6v4_property_update_broadcast(
                                IX_CC_SET_PROPERTY_ADD_BROADCAST_IPV4,
                                (ix_uint16)arg_pPropData->blade_id,
                                (ix_uint16)arg_pPropData->port_id,
                                ipAddr));
            }
        }
        /* Update all start tunnels configured on this interface */
        IX_ERROR_CR(_ix_cc_v6v4_update_tunnels_broadcast(
                        (ix_uint16)arg_pPropData->blade_id,
                        (ix_uint16)arg_pPropData->port_id,
                        ipAddr));
    }/* end if add ipv4 broadcast address */

    if ((arg_PropId & IX_CC_SET_PROPERTY_DEL_BROADCAST_IPV4) == 
        IX_CC_SET_PROPERTY_DEL_BROADCAST_IPV4)
    {
        ix_uint32 loop_var;

        for (loop_var = 0;
             loop_var < arg_pPropData->no_virtual_if;
             loop_var++)
        {
            if (arg_pPropData->ip_prop[loop_var].ip_version
                == IX_CC_IP_VERSION_IPV4)
            {
               /* Convert IP Address into network byte order */
               ipAddr =
                 IX_HTON32(
                     arg_pPropData->ip_prop[loop_var].protocol.ipv4_prop.ipv4_broadcast);
               IX_ERROR_CR(_ix_cc_v6v4_property_update_broadcast(
                               IX_CC_SET_PROPERTY_DEL_BROADCAST_IPV4,
                               (ix_uint16)arg_pPropData->blade_id,
                               (ix_uint16)arg_pPropData->port_id,
                               ipAddr));
            }
        }

        /* Update all start tunnels configured on this interface */
        IX_ERROR_CR(_ix_cc_v6v4_update_tunnels_broadcast(
                        (ix_uint16)arg_pPropData->blade_id,
                        (ix_uint16)arg_pPropData->port_id,
                        ipAddr));
    }/* end if delete ipv4 broadcast address */

    if ((arg_PropId & IX_CC_SET_PROPERTY_MTU) == IX_CC_SET_PROPERTY_MTU)
    {
        IX_ERROR_CR(_ix_cc_v6v4_property_update_mtu(
                                   (ix_uint16)arg_pPropData->blade_id,
                                   (ix_uint16)arg_pPropData->port_id,
                                   arg_pPropData->mtu));
        /* Update all start tunnels configured on this interface */
        IX_ERROR_CR(_ix_cc_v6v4_update_tunnels_mtu(
                        (ix_uint16)arg_pPropData->blade_id,
                        (ix_uint16)arg_pPropData->port_id,
                        arg_pPropData->mtu));
    }/* end if set MTU */


    return IX_SUCCESS;
} /* end ix_cc_v6v4_set_property */
