/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 6
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      IPv6/v4 Tunneling Core Component - message handler
 *
 * = FILENAME
 *      ix_cc_v6v4_msg_handler.c
 *
 * = DESCRIPTION
 *      The file defines message handler interface functions.
 *
 * = CREATION TIME
 *      11/15/2002 2:55 PM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */



#define IX_ERROR_FILE_IDENT "$Id: ix_cc_v6v4_msg_handler.c,v 1.3 2003/01/30 19:08:47 cabell Exp $";

/**
 * System defined include files
 */

/**
 * User defined include files
 */

#include "ix_cc_error.h"
#include "ix_ossl.h"
#include "ix_rm.h"

#include "ix_cc.h"
#include "ix_cc_macros.h"
#include "ix_cc_properties.h"

#include "cc/ix_cc_msup.h"

#include "bindings.h"

#include "cc/ix_cc_v6v4.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_init.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_msg_handler.h"

/**
 * Preprocessor symbols and Macros used in this file
 */
/* Flag for not to send report to msg support lib*/
#define IX_CC_V6V4_SEND_NO_RESPONSE 1 

/**
 * Type definitions whose scope is limited to this file
 */

/* Return data for IX_CC_V6V4_MSG_DUMP_END_TUNNELS */
typedef struct ix_s_cc_v6v4_dump_end_tunnel_data
{
    ix_uint32 numEntries;
    ix_cc_v6v4_end_tunnel_info* paInfo;
} ix_cc_v6v4_dump_end_tunnel_data;


/* Return data for IX_CC_V6V4_MSG_DUMP_START_TUNNELS */
typedef struct ix_s_cc_v6v4_dump_start_tunnel_data
{
    ix_uint32 numEntries;
    ix_cc_v6v4_start_tunnel_info* paInfo;
} ix_cc_v6v4_dump_start_tunnel_data;


/* Return data for IX_CC_V6V4_MSG_GET_ALLOWED_SOURCES */
typedef struct ix_s_cc_v6v4_get_allowed_sources_data
{
    ix_uint32 numEntries;
    ix_cc_v6v4_ingress_source_entry* paEntries;
} ix_cc_v6v4_get_allowed_sources_data;


typedef union ix_u_cc_v6v4_msg_handler_return_data
{
    ix_cc_v6v4_tunnel_handle hTunnel;
    ix_cc_v6v4_end_tunnel_info endTunnelData;
    ix_cc_v6v4_start_tunnel_info startTunnelData;
    ix_cc_v6v4_dump_end_tunnel_data dumpEndTunnelData;
    ix_cc_v6v4_dump_start_tunnel_data dumpStartTunnelData;
    ix_cc_v6v4_get_allowed_sources_data getAllowedSrcsData;
    ix_cc_v6v4_statistics statsData;
} ix_cc_v6v4_msg_handler_return_data;

/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */

extern ix_cc_v6v4_registry_data g_V6v4RegistryData;

/** 
 * Extern function prototypes.
 */


/** 
 * Static function prototypes.
 */


/** 
 * Function definitions.
 */


/**
 * NAME: ix_cc_v6v4_msg_handler
 *
 * DESCRIPTION: This primitive is the message handler function for IPv6/v4
 * Tunneling core component. The tunneling core component receives messages
 * from other core components through this message handler function and
 * internally it calls the appropriate library function to process the
 * message. This message handler is also used to update dynamic properties.
 * 
 * @Param: - IN arg_hMsgInfo - buffer handle embedding information pertinent
 * to the message.
 *
 * @Param: - IN arg_MsgType - message type
 *
 * @Param: - IN arg_pContext: pointer to Tunneling core component context that 
 * was defined by the core component and passed to the core component
 * infrastructure through ix_cc_v6v4_init().
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - null argument
 * IX_CC_ERROR_UNDEFINED_MSG - un supported message
 * IX_CC_V6V4_ERROR_MSG_LIBRARY - error from message support library
 * IX_CC_V6V4_ERROR_BUFFER_FREE - error from RM for freeing buffer
 */


ix_error ix_cc_v6v4_msg_handler(
    ix_buffer_handle arg_hMsgInfo,
    ix_uint32 arg_MsgType,
    void* arg_pContext
    )
{

    ix_error err = IX_SUCCESS;
    void *pContext = NULL;
    void *pMsg = NULL;
    void *pRetMsg = NULL;
    ix_cc_v6v4_msg_handler_return_data msgRetData;
    ix_uint32 msgRetLength = 0;
    ix_uint32 noResponse = 0;  

    
    /* Validate arg_pContext */
    if(arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                   ("arg_pContext is NULL"));
    }/* end if(arg_pContext == NULL) */

    if((arg_MsgType < IX_CC_COMMON_MSG_ID_FIRST) ||
       (arg_MsgType >= IX_CC_V6V4_MSG_LAST))
    {
    
        return IX_ERROR_WARNING(IX_CC_ERROR_UNDEFINED_MSG,
                                  ("Invalid Message Type"));
    }/* end if */

    IX_MSUP_EXTRACT_MSG(arg_hMsgInfo,&pMsg,&pContext);

    if((arg_MsgType != IX_CC_V6V4_MSG_DUMP_END_TUNNELS) &&
       (arg_MsgType != IX_CC_V6V4_MSG_DUMP_START_TUNNELS) &&
       (arg_MsgType != IX_CC_V6V4_MSG_GET_STATISTICS)) 
    {
        if(pMsg == NULL)
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                      ("Invalid Buffer"));
        }/* end if(pMsg == NULL) */
    }/* end if */
    
    /* Check the message type */
    switch(arg_MsgType)
    {
        case IX_CC_COMMON_MSG_ID_PROP_UPDATE: 
        {
            ix_cc_properties_msg* pData = (ix_cc_properties_msg*)pMsg;
            noResponse = IX_CC_V6V4_SEND_NO_RESPONSE;
            if(((pData->propID & IX_CC_SET_PROPERTY_ADD_INTERFACE_IPV4) == IX_CC_SET_PROPERTY_ADD_INTERFACE_IPV4) ||
               ((pData->propID & IX_CC_SET_PROPERTY_DEL_INTERFACE_IPV4) == IX_CC_SET_PROPERTY_DEL_INTERFACE_IPV4) ||   
               ((pData->propID & IX_CC_SET_PROPERTY_ADD_BROADCAST_IPV4) == IX_CC_SET_PROPERTY_ADD_BROADCAST_IPV4) ||   
               ((pData->propID & IX_CC_SET_PROPERTY_DEL_BROADCAST_IPV4) == IX_CC_SET_PROPERTY_DEL_BROADCAST_IPV4) ||   
               ((pData->propID & IX_CC_SET_PROPERTY_MTU) == IX_CC_SET_PROPERTY_MTU))
            {
                err = ix_cc_v6v4_set_property(pData->propID,&pData->propData,arg_pContext); 
            }/* end if propId */
            else
            {
                err = IX_ERROR_WARNING(IX_CC_ERROR_UNDEFINED_MSG,
                                       ("Invalid Message Type"));
 
            }/*end if invalid propId*/

            break;
        }


        case IX_CC_V6V4_MSG_ADD_END_TUNNEL:
        {
            ix_cc_v6v4_end_tunnel_config* pData = (ix_cc_v6v4_end_tunnel_config*)pMsg;
	        err = ix_cc_v6v4_add_end_tunnel(pData, arg_pContext, &msgRetData.hTunnel);
            pRetMsg = &msgRetData.hTunnel;
            msgRetLength = sizeof(msgRetData.hTunnel);
            break;
        }
        case IX_CC_V6V4_MSG_DELETE_END_TUNNEL:
        {
            ix_cc_v6v4_tunnel_handle* pData = (ix_cc_v6v4_tunnel_handle*) pMsg;
            err = ix_cc_v6v4_delete_end_tunnel(*pData, arg_pContext);
            break;
        }
        case IX_CC_V6V4_MSG_SET_DECAP_TOS_OPTION:
        {
            ix_cc_v6v4_set_decap_tos_data* pData = (ix_cc_v6v4_set_decap_tos_data*)pMsg;
	        err = ix_cc_v6v4_set_decap_tos_option(pData->hTunnel,pData->selection,arg_pContext);
            break;
        }
        case IX_CC_V6V4_MSG_SET_SRC_VALIDATION:
        {
            ix_cc_v6v4_set_src_validation_data* pData = (ix_cc_v6v4_set_src_validation_data*)pMsg;
	        err = ix_cc_v6v4_set_src_validation(pData->hTunnel,pData->selection,arg_pContext);
            break;
        }
        case IX_CC_V6V4_MSG_GET_END_TUNNEL:
        {
            ix_cc_v6v4_tunnel_handle* pData = (ix_cc_v6v4_tunnel_handle*) pMsg;
            msgRetData.endTunnelData.hTunnel = *pData;
	        err = ix_cc_v6v4_get_end_tunnel(&msgRetData.endTunnelData,arg_pContext);
            pRetMsg = &msgRetData.endTunnelData;
            msgRetLength = sizeof(msgRetData.endTunnelData);
            break;
        }
        case IX_CC_V6V4_MSG_ADD_ALLOWED_SOURCE:
        {
            ix_cc_v6v4_add_allowed_source_data* pData = (ix_cc_v6v4_add_allowed_source_data*)pMsg;
	        err = ix_cc_v6v4_add_allowed_source(pData->hTunnel,&pData->islEntry,arg_pContext);
            break;
        }
        case IX_CC_V6V4_MSG_DELETE_ALLOWED_SOURCE:
        {
            ix_cc_v6v4_delete_allowed_source_data* pData = (ix_cc_v6v4_delete_allowed_source_data*)pMsg;
	        err = ix_cc_v6v4_delete_allowed_source(pData->hTunnel,&pData->islEntry,arg_pContext);
            break;
        }
        case IX_CC_V6V4_MSG_CLEAR_ALLOWED_SOURCES:
        {
            ix_cc_v6v4_tunnel_handle* pData = (ix_cc_v6v4_tunnel_handle*) pMsg;
            err = ix_cc_v6v4_clear_allowed_sources(*pData, arg_pContext);
            break;
        }
        case IX_CC_V6V4_MSG_GET_ALLOWED_SOURCES:
        {
            ix_uint32 size = IX_CC_V6V4_ALLOWED_SRCS_HINT;
            ix_cc_v6v4_tunnel_handle* pData = (ix_cc_v6v4_tunnel_handle*) pMsg;

            do
            {
                msgRetData.getAllowedSrcsData.numEntries = size;
                size *= sizeof(ix_cc_v6v4_ingress_source_entry);
                msgRetData.getAllowedSrcsData.paEntries =
                    (ix_cc_v6v4_ingress_source_entry *)ix_ossl_malloc(size);

                if(msgRetData.getAllowedSrcsData.paEntries != NULL)
                {
	                err = ix_cc_v6v4_get_allowed_sources(
                        *pData,
                        &msgRetData.getAllowedSrcsData.numEntries,
                        msgRetData.getAllowedSrcsData.paEntries,
                        arg_pContext);

                    if (err == IX_CC_V6V4_ERROR_BUFFER_TOO_SMALL)
                    {
                        size = msgRetData.getAllowedSrcsData.numEntries;
                        ix_ossl_free(msgRetData.getAllowedSrcsData.paEntries);
                    }
                }
                else
                {
                    err = IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                                           ("No memory to dump routes"));

                }
            } while (err == IX_CC_V6V4_ERROR_BUFFER_TOO_SMALL);

            pRetMsg = (void*)&msgRetData.getAllowedSrcsData;
            msgRetLength = sizeof(ix_cc_v6v4_get_allowed_sources_data);
            break;
        }
        case IX_CC_V6V4_MSG_DUMP_END_TUNNELS:
        {
            ix_uint32 size = (g_V6v4RegistryData.endTunnelSize / 2) + 1;

            do
            {
                msgRetData.dumpEndTunnelData.numEntries = size;
                size *= sizeof(ix_cc_v6v4_end_tunnel_info);
                msgRetData.dumpEndTunnelData.paInfo =
                    (ix_cc_v6v4_end_tunnel_info *)ix_ossl_malloc(size);

                if(msgRetData.dumpEndTunnelData.paInfo != NULL)
                {
	                err = ix_cc_v6v4_dump_end_tunnels(
                        &msgRetData.dumpEndTunnelData.numEntries,
                        msgRetData.dumpEndTunnelData.paInfo,
                        arg_pContext);

                    if (err == IX_CC_V6V4_ERROR_BUFFER_TOO_SMALL)
                    {
                        size = msgRetData.dumpEndTunnelData.numEntries;
                        ix_ossl_free(msgRetData.dumpEndTunnelData.paInfo);
                    }
                }
                else
                {
                    err = IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                                           ("No memory to dump routes"));

                }
            } while (err == IX_CC_V6V4_ERROR_BUFFER_TOO_SMALL);

            pRetMsg = (void*)&msgRetData.dumpEndTunnelData;
            msgRetLength = sizeof(ix_cc_v6v4_dump_end_tunnel_data);
            break;
        }
        case IX_CC_V6V4_MSG_ADD_START_TUNNEL:
        {
            ix_cc_v6v4_start_tunnel_config* pData = (ix_cc_v6v4_start_tunnel_config*)pMsg;
	        err = ix_cc_v6v4_add_start_tunnel(pData, arg_pContext, &msgRetData.hTunnel);
            pRetMsg = &msgRetData.hTunnel;
            msgRetLength = sizeof(msgRetData.hTunnel);
            break;
        }
        case IX_CC_V6V4_MSG_DELETE_START_TUNNEL:
        {
            ix_cc_v6v4_tunnel_handle* pData = (ix_cc_v6v4_tunnel_handle*) pMsg;
            err = ix_cc_v6v4_delete_start_tunnel(*pData, arg_pContext);
            break;
        }
        case IX_CC_V6V4_MSG_SET_TTL:
        {
            ix_cc_v6v4_set_ttl_data* pData = (ix_cc_v6v4_set_ttl_data*)pMsg;
	        err = ix_cc_v6v4_set_ttl(pData->hTunnel,pData->ttl,arg_pContext);
            break;
        }
        case IX_CC_V6V4_MSG_SET_ENCAP_TOS_OPTION:
        {
            ix_cc_v6v4_set_encap_tos_data* pData = (ix_cc_v6v4_set_encap_tos_data*)pMsg;
	        err = ix_cc_v6v4_set_encap_tos_option(
                pData->hTunnel,pData->selection,arg_pContext);
            break;
        }
        case IX_CC_V6V4_MSG_SET_TOS:
        {
            ix_cc_v6v4_set_tos_data* pData = (ix_cc_v6v4_set_tos_data*)pMsg;
	        err = ix_cc_v6v4_set_tos(
                pData->hTunnel,pData->tos,arg_pContext);
            break;
        }
        case IX_CC_V6V4_MSG_SET_MTU:
        {
            ix_cc_v6v4_set_mtu_data* pData = (ix_cc_v6v4_set_mtu_data*)pMsg;
	        err = ix_cc_v6v4_set_mtu(pData->hTunnel,pData->mtu,arg_pContext);
            break;
        }
        case IX_CC_V6V4_MSG_SET_SUBNET_BROADCAST:
        {
            ix_cc_v6v4_set_subnet_broadcast_data* pData =
                (ix_cc_v6v4_set_subnet_broadcast_data*)pMsg;
	        err = ix_cc_v6v4_set_subnet_broadcast(
                    pData->hTunnel,pData->subnetBroadcast,arg_pContext);
            break;
        }
        case IX_CC_V6V4_MSG_GET_START_TUNNEL:
        {
            ix_cc_v6v4_tunnel_handle* pData = (ix_cc_v6v4_tunnel_handle*) pMsg;
            msgRetData.startTunnelData.hTunnel = *pData;
	        err = ix_cc_v6v4_get_start_tunnel(&msgRetData.startTunnelData,arg_pContext);
            pRetMsg = &msgRetData.startTunnelData;
            msgRetLength = sizeof(msgRetData.startTunnelData);
            break;
        }
        case IX_CC_V6V4_MSG_DUMP_START_TUNNELS:
        {
            ix_uint32 size = (g_V6v4RegistryData.startTunnelSize / 2) + 1;

            do
            {
                msgRetData.dumpStartTunnelData.numEntries = size;
                size *= sizeof(ix_cc_v6v4_start_tunnel_info);
                msgRetData.dumpStartTunnelData.paInfo =
                    (ix_cc_v6v4_start_tunnel_info *)ix_ossl_malloc(size);

                if(msgRetData.dumpStartTunnelData.paInfo != NULL)
                {
	                err = ix_cc_v6v4_dump_start_tunnels(
                        &msgRetData.dumpStartTunnelData.numEntries,
                        msgRetData.dumpStartTunnelData.paInfo,
                        arg_pContext);

                    if (err == IX_CC_V6V4_ERROR_BUFFER_TOO_SMALL)
                    {
                        size = msgRetData.dumpStartTunnelData.numEntries;
                        ix_ossl_free(msgRetData.dumpStartTunnelData.paInfo);
                    }
                }
                else
                {
                    err = IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                                           ("No memory to dump routes"));

                }
            } while (err == IX_CC_V6V4_ERROR_BUFFER_TOO_SMALL);

            pRetMsg = (void*)&msgRetData.dumpStartTunnelData;
            msgRetLength = sizeof(ix_cc_v6v4_dump_start_tunnel_data);
            break;
        }
        case IX_CC_V6V4_MSG_GET_STATISTICS:
        {
	        err = ix_cc_v6v4_get_statistics(&msgRetData.statsData,arg_pContext);
            pRetMsg = &msgRetData.statsData;
            msgRetLength = sizeof(msgRetData.statsData);
            break;
        }
        default:
           IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hMsgInfo),IX_CC_V6V4_ERROR_BUFFER_FREE,0);  
           return IX_ERROR_WARNING(IX_CC_ERROR_UNDEFINED_MSG,
                                     ("Invalid Message Type"));
           break;
    }/* end switch(arg_MsgType) */

    if (pMsg != NULL)
    {
        /* ix_cc_msup_extract_msg copies the message parameters into a
         * allocated local buffer, which must be freed by the caller.
         */
        ix_ossl_free(pMsg);
    }


    if(noResponse == IX_CC_V6V4_SEND_NO_RESPONSE)
    {
        IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hMsgInfo),IX_CC_V6V4_ERROR_BUFFER_FREE,0);  
        return IX_SUCCESS;
    }/* end if no need to send message to msg support lib*/


    /* send return message, length and retrieved context to message support library */
    err = ix_cc_msup_send_reply_msg(pContext,pRetMsg,msgRetLength,err);
    if(err != IX_SUCCESS)
    {
        err = IX_ERROR_WARNING(IX_CC_V6V4_ERROR_MSG_LIBRARY,
                                   ("Send reply message failed"));
    }/* endif (err != IX_SUCCESS) */

    IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hMsgInfo),IX_CC_V6V4_ERROR_BUFFER_FREE,0);
    return err;

} /* end ix_cc_v6v4_msg_handler */





/**
 * NAME: _ix_cc_v6v4_cb_genaral
 *
 * DESCRIPTION: This is a general callback function. The message support
 * library  calls this function to return the status of the requested
 * asynchronous operation.
 * @Param:  - IN  arg_Result: result of the asynchronous operation
 * @Param:  - IN  arg_pContext: context provided by caller
 * @Param:  - IN  arg_pMsg: return message
 * @Param:  - IN  arg_MsgLength: length of the message
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_CALLBACK - user callback returns error message
 * 
 */

ix_error _ix_cc_v6v4_cb_general(
    ix_error arg_Result,
    void* arg_pContext,
    void* arg_pMsg,
    ix_uint32 arg_MsgLength
    )
{
    ix_error err = IX_SUCCESS;
 
    ix_cc_v6v4_general_callback_data* pCbData = NULL;
 
    pCbData = (ix_cc_v6v4_general_callback_data*)arg_pContext;
   
    if(pCbData != NULL)
    {
        err = pCbData->pUserCallback(arg_Result,pCbData->pUserContext);
        ix_ossl_free(arg_pContext);
        if(err != IX_SUCCESS)
        {
            return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_CALLBACK,
                                      ("general callback failed"));
        }/* end if (err != IX_SUCCESS) */
    }/* end if(pCbData != NULL) */
       
    return IX_SUCCESS;
}/* _ix_cc_v6v4_cb_general */




/**
 * NAME: _ix_cc_v6v4_cb_add_tunnel
 *
 * DESCRIPTION: This is a callback function for add tunnel message helpers.
 * The message support library calls this function to return the status of the
 * requested asynchrouns operation.
 * @Param:  - IN  arg_Result: result of the asynchronous operation
 * @Param:  - IN  arg_pContext: context provided by caller
 * @Param:  - IN  arg_pMsg: return message
 * @Param:  - IN  arg_MsgLength: length of the message
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_CALLBACK - user callback returns error message
 * 
 */
ix_error _ix_cc_v6v4_cb_add_tunnel(
    ix_error arg_Result,
    void* arg_pContext,
    void* arg_pMsg,
    ix_uint32 arg_MsgLength
    )
{
    ix_error err = IX_SUCCESS;
 
    ix_cc_v6v4_add_tunnel_callback_data* pCbData;
    ix_cc_v6v4_tunnel_handle* pMsgData;

    if (arg_pContext != NULL)
    {
        pCbData = (ix_cc_v6v4_add_tunnel_callback_data*)arg_pContext;
        pMsgData = (ix_cc_v6v4_tunnel_handle*)arg_pMsg;
    
        err = pCbData->pUserCallback(arg_Result,pCbData->pUserContext,*pMsgData);
        ix_ossl_free(arg_pContext);
        if(err != IX_SUCCESS)
        {
            return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_CALLBACK,
                                      ("lookup callback failed"));
        }/* end if (err != IX_SUCCESS) */
    }/* end if(pCbData != NULL)*/
       
    return IX_SUCCESS;
}/* _ix_cc_v6v4_cb_add_tunnel */


/**
 * NAME: _ix_cc_v6v4_cb_end_tunnel
 *
 * DESCRIPTION: This is a callback function for add tunnel message helpers.
 * The message support library calls this function to return the status of the
 * requested asynchrouns operation.
 * @Param:  - IN  arg_Result: result of the asynchronous operation
 * @Param:  - IN  arg_pContext: context provided by caller
 * @Param:  - IN  arg_pMsg: return message
 * @Param:  - IN  arg_MsgLength: length of the message
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_CALLBACK - user callback returns error message
 * 
 */
ix_error _ix_cc_v6v4_cb_end_tunnel(
    ix_error arg_Result,
    void* arg_pContext,
    void* arg_pMsg,
    ix_uint32 arg_MsgLength
    )
{
    ix_error err = IX_SUCCESS;
 
    ix_cc_v6v4_end_tunnel_callback_data* pCbData;
    ix_cc_v6v4_end_tunnel_info* pMsgData;

    if (arg_pContext != NULL)
    {
        pCbData = (ix_cc_v6v4_end_tunnel_callback_data*)arg_pContext;
        pMsgData = (ix_cc_v6v4_end_tunnel_info*)arg_pMsg;
    
        err = pCbData->pUserCallback(arg_Result,pCbData->pUserContext,pMsgData);
        ix_ossl_free(arg_pContext);
        if(err != IX_SUCCESS)
        {
            return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_CALLBACK,
                                      ("lookup callback failed"));
        }/* end if (err != IX_SUCCESS) */
    }/* end if(pCbData != NULL)*/
       
    return IX_SUCCESS;
}/* _ix_cc_v6v4_cb_end_tunnel */


/**
 * NAME: _ix_cc_v6v4_cb_get_sources
 *
 * DESCRIPTION: This is a callback function for get allowed sources message
 * helpers. The message support library calls this function to return the
 * status of the requested asynchrouns operation.
 * @Param:  - IN  arg_Result: result of the asynchronous operation
 * @Param:  - IN  arg_pContext: context provided by caller
 * @Param:  - IN  arg_pMsg: return message
 * @Param:  - IN  arg_MsgLength: length of the message
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_CALLBACK - user callback returns error message
 * 
 */
ix_error _ix_cc_v6v4_cb_get_sources(
    ix_error arg_Result,
    void* arg_pContext,
    void* arg_pMsg,
    ix_uint32 arg_MsgLength
    )
{
    ix_error err = IX_SUCCESS;
 
    ix_cc_v6v4_get_sources_callback_data* pCbData;
    ix_cc_v6v4_get_allowed_sources_data* pMsgData;

    if (arg_pContext != NULL)
    {
        pCbData = (ix_cc_v6v4_get_sources_callback_data*)arg_pContext;
        pMsgData = (ix_cc_v6v4_get_allowed_sources_data*)arg_pMsg;
    
        err = pCbData->pUserCallback(
            arg_Result,pCbData->pUserContext,
            pMsgData->numEntries,pMsgData->paEntries);

        ix_ossl_free(arg_pContext);
        ix_ossl_free(pMsgData->paEntries);
        if(err != IX_SUCCESS)
        {
            return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_CALLBACK,
                                      ("lookup callback failed"));
        }/* end if (err != IX_SUCCESS) */
    }/* end if(pCbData != NULL)*/
       
    return IX_SUCCESS;
}/* _ix_cc_v6v4_cb_get_sources */


/**
 * NAME: _ix_cc_v6v4_cb_dump_end_tunnels
 *
 * DESCRIPTION: This is a callback function for dump end tunnel message
 * helpers. The message support library calls this function to return the
 * status of the requested asynchrouns operation.
 * @Param:  - IN  arg_Result: result of the asynchronous operation
 * @Param:  - IN  arg_pContext: context provided by caller
 * @Param:  - IN  arg_pMsg: return message
 * @Param:  - IN  arg_MsgLength: length of the message
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_CALLBACK - user callback returns error message
 * 
 */
ix_error _ix_cc_v6v4_cb_dump_end_tunnels(
    ix_error arg_Result,
    void* arg_pContext,
    void* arg_pMsg,
    ix_uint32 arg_MsgLength
    )
{
    ix_error err = IX_SUCCESS;
 
    ix_cc_v6v4_dump_end_tunnels_callback_data* pCbData;
    ix_cc_v6v4_dump_end_tunnel_data* pMsgData;

    if (arg_pContext != NULL)
    {
        pCbData = (ix_cc_v6v4_dump_end_tunnels_callback_data*)arg_pContext;
        pMsgData = (ix_cc_v6v4_dump_end_tunnel_data*)arg_pMsg;
    
        err = pCbData->pUserCallback(
            arg_Result,pCbData->pUserContext,
            pMsgData->numEntries,pMsgData->paInfo);

        ix_ossl_free(arg_pContext);
        ix_ossl_free(pMsgData->paInfo);
        if(err != IX_SUCCESS)
        {
            return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_CALLBACK,
                                      ("lookup callback failed"));
        }/* end if (err != IX_SUCCESS) */
    }/* end if(pCbData != NULL)*/
       
    return IX_SUCCESS;
}/* _ix_cc_v6v4_cb_dump_end_tunnels */


/**
 * NAME: _ix_cc_v6v4_cb_start_tunnel
 *
 * DESCRIPTION: This is a callback function for get start tunnel message
 * helpers. The message support library calls this function to return the
 * status of the requested asynchrouns operation.
 * @Param:  - IN  arg_Result: result of the asynchronous operation
 * @Param:  - IN  arg_pContext: context provided by caller
 * @Param:  - IN  arg_pMsg: return message
 * @Param:  - IN  arg_MsgLength: length of the message
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_CALLBACK - user callback returns error message
 * 
 */
ix_error _ix_cc_v6v4_cb_start_tunnel(
    ix_error arg_Result,
    void* arg_pContext,
    void* arg_pMsg,
    ix_uint32 arg_MsgLength
    )
{
    ix_error err = IX_SUCCESS;
 
    ix_cc_v6v4_start_tunnel_callback_data* pCbData;
    ix_cc_v6v4_start_tunnel_info* pMsgData;

    if (arg_pContext != NULL)
    {
        pCbData = (ix_cc_v6v4_start_tunnel_callback_data*)arg_pContext;
        pMsgData = (ix_cc_v6v4_start_tunnel_info*)arg_pMsg;
    
        err = pCbData->pUserCallback(arg_Result,pCbData->pUserContext,pMsgData);
        ix_ossl_free(arg_pContext);
        if(err != IX_SUCCESS)
        {
            return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_CALLBACK,
                                      ("lookup callback failed"));
        }/* end if (err != IX_SUCCESS) */
    }/* end if(pCbData != NULL)*/
       
    return IX_SUCCESS;
}/* _ix_cc_v6v4_cb_start_tunnel */


/**
 * NAME: _ix_cc_v6v4_cb_dump_start_tunnels
 *
 * DESCRIPTION: This is a callback function for dump start tunnel message
 * helpers. The message support library calls this function to return the
 * status of the requested asynchrouns operation.
 * @Param:  - IN  arg_Result: result of the asynchronous operation
 * @Param:  - IN  arg_pContext: context provided by caller
 * @Param:  - IN  arg_pMsg: return message
 * @Param:  - IN  arg_MsgLength: length of the message
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_CALLBACK - user callback returns error message
 * 
 */
ix_error _ix_cc_v6v4_cb_dump_start_tunnels(
    ix_error arg_Result,
    void* arg_pContext,
    void* arg_pMsg,
    ix_uint32 arg_MsgLength
    )
{
    ix_error err = IX_SUCCESS;
 
    ix_cc_v6v4_dump_start_tunnels_callback_data* pCbData;
    ix_cc_v6v4_dump_start_tunnel_data* pMsgData;

    if (arg_pContext != NULL)
    {
        pCbData = (ix_cc_v6v4_dump_start_tunnels_callback_data*)arg_pContext;
        pMsgData = (ix_cc_v6v4_dump_start_tunnel_data*)arg_pMsg;
    
        err = pCbData->pUserCallback(
            arg_Result,pCbData->pUserContext,
            pMsgData->numEntries,pMsgData->paInfo);

        ix_ossl_free(arg_pContext);
        ix_ossl_free(pMsgData->paInfo);
        if(err != IX_SUCCESS)
        {
            return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_CALLBACK,
                                      ("lookup callback failed"));
        }/* end if (err != IX_SUCCESS) */
    }/* end if(pCbData != NULL)*/
       
    return IX_SUCCESS;
}/* _ix_cc_v6v4_cb_dump_end_tunnels */


/**
 * NAME: _ix_cc_v6v4_cb_statistics
 *
 * DESCRIPTION: This is a callback function for get statistics message
 * helpers. The message support library calls this function to return the
 * status of the requested asynchrouns operation.
 * @Param:  - IN  arg_Result: result of the asynchronous operation
 * @Param:  - IN  arg_pContext: context provided by caller
 * @Param:  - IN  arg_pMsg: return message
 * @Param:  - IN  arg_MsgLength: length of the message
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_CALLBACK - user callback returns error message
 * 
 */
ix_error _ix_cc_v6v4_cb_statistics(
    ix_error arg_Result,
    void* arg_pContext,
    void* arg_pMsg,
    ix_uint32 arg_MsgLength
    )
{
    ix_error err = IX_SUCCESS;
 
    ix_cc_v6v4_statistics_callback_data* pCbData;
    ix_cc_v6v4_statistics* pMsgData;

    if (arg_pContext != NULL)
    {
        pCbData = (ix_cc_v6v4_statistics_callback_data*)arg_pContext;
        pMsgData = (ix_cc_v6v4_statistics*)arg_pMsg;
    
        err = pCbData->pUserCallback(arg_Result,pCbData->pUserContext,pMsgData);
        ix_ossl_free(arg_pContext);
        if(err != IX_SUCCESS)
        {
            return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_CALLBACK,
                                      ("lookup callback failed"));
        }/* end if (err != IX_SUCCESS) */
    }/* end if(pCbData != NULL)*/
       
    return IX_SUCCESS;
}/* _ix_cc_v6v4_cb_options */
