/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 6
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      IPv6/v4 Tunneling Core Component - packet handlers
 *
 * = FILENAME
 *      ix_cc_v6v4_pkt_handlers.c
 *
 * = DESCRIPTION
 *      The file defines packet handler interface functions.
 *
 * = CREATION TIME
 *      11/15/2002 2:13 PM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */



#define IX_ERROR_FILE_IDENT "$Id: ix_cc_v6v4_pkt_handlers.c,v 1.5 2003/02/07 18:40:04 cabell Exp $";


/**
 * System defined include files
 */


/**
 * User defined include files
 */
#include "ix_cc_error.h"

#include "ix_rm.h"
#include "ix_ossl.h"

#include "ix_cc.h"
#include "bindings.h"
#include "cc/ix_cc_v6v4.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_init.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_icmp.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_decap.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_encap.h"


/**
 * Preprocessor symbols and Macros used in this file
 */

/* Exception codes - must match the defines in the microcode headers */
#define V6V4_DECAP_EXCEPTION_LOCAL      0x80
#define V6V4_DECAP_EXCEPTION_AUTO       0x81
#define V6V4_DECAP_EXCEPTION_FRAGMENT   0x82
#define V6V4_ENCAP_EXCEPTION_MTU        0x90

/* Next hop type for tunnels */
#define IX_CC_V6V4_NEXT_HOP_TUNNEL      0x2

/**
 * Type definitions whose scope is limited to this file
 */


/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */


/** 
 * Extern function prototypes.
 */


extern ix_cc_v6v4_context* g_pV6v4Context;
extern ix_error _ix_cc_v6v4_forward(ix_buffer_handle arg_hBuffer);

/** 
 * Static function prototypes.
 */


/** 
 * Function definitions.
 */




/**
 * NAME: ix_cc_v6v4_microblock_pkt_handler
 *
 * DESCRIPTION: This is the registered function to receive exception packets from
 * the tunneling microblocks. 
 * 
 * @Param:  - IN arg_hPacket: handle to a buffer which contains exception packets
 * from tunneling microblock
 * @Param: - IN arg_ExceptionCode: exception code set by microblock.
 * @Param: - IN arg_pContext: pointer to Tunneling core component context that 
 * was defined by the core component and passed to the core component
 * infrastructure through ix_cc_v6v4_init().
 * @Param:  - OUT
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_SEND_FAIL - send packet or message failure
 * IX_CC_ERROR_UNDEFINED_EXCEP - unsupported exception code
 * IX_CC_ERROR_INTERNAL - an internal error
 * IX_CC_ERROR_ALIGN - pointer not aligned properly
 * IX_CC_ERROR_NULL - null pointer
 * 
 */

ix_error ix_cc_v6v4_microblock_pkt_handler(
    ix_buffer_handle arg_hPacket,  
    ix_uint32 arg_ExceptionCode,
    void* arg_pContext
    )
{
    ix_error err = IX_SUCCESS;
    ix_uint8 sendUp;
    ix_cc_v6v4_context* pContext;
    ix_buffer_handle hReassembledPacket;

    pContext = (ix_cc_v6v4_context*)arg_pContext;

    switch(arg_ExceptionCode)
    {
        case V6V4_DECAP_EXCEPTION_LOCAL:
            IX_ERROR_CG(
                _ix_cc_v6v4_icmp_check(arg_hPacket, arg_pContext, &sendUp),
                err, freePacket);

            if (sendUp)
            {
                pContext->ccStats.localPkts++;
                IX_ERROR_CGT(
                    ix_rm_message_send(
                        IX_CC_V6V4_PKT_HIGH_PRIORITY_STKDRV_OUTPUT,
                        arg_hPacket, 0),
                    err, IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING,
                    freePacket);
            }
            break;

        case V6V4_DECAP_EXCEPTION_AUTO:
            pContext->ccStats.autoTunnelPkts++;

            IX_ERROR_CG(
                _ix_cc_v6v4_decapsulate(arg_hPacket, arg_pContext),
                err, freePacket);

            IX_ERROR_CGT(
                ix_rm_message_send(
                    IX_CC_V6V4_PKT_HIGH_PRIORITY_STKDRV_OUTPUT,
                    arg_hPacket, 0),
                err, IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING,
                freePacket);

            break;

        case V6V4_DECAP_EXCEPTION_FRAGMENT:
            pContext->ccStats.fragmentsRcvd++;

            err = _ix_cc_v6v4_reasm_add_fragment(
                    pContext->hReassembly, arg_hPacket,
                    &hReassembledPacket);

            IX_ERROR_CRT(
                ix_rm_buffer_free_chain(arg_hPacket),
                IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

            if (err == IX_SUCCESS)
            {
                if (hReassembledPacket != IX_NULL_BUFFER_HANDLE)
                {
                    pContext->ccStats.reassembledPkts++;

                    IX_ERROR_CG(
                        _ix_cc_v6v4_handle_ipv4_packet(
                            hReassembledPacket, arg_pContext),
                        err, freeReassembledPacket);
                }
            }
                    
            break;

        case V6V4_ENCAP_EXCEPTION_MTU:
            pContext->ccStats.pktTooBigMsgsSent++;

            IX_ERROR_CG(
                _ix_cc_v6v4_pkt_too_big(arg_hPacket, arg_pContext),
                err, freePacket);
            break;

        default:
            IX_ERROR_CRT(
                ix_rm_buffer_free_chain(arg_hPacket),
                IX_CC_ERROR_INTERNAL,0);

            return IX_ERROR_WARNING(IX_CC_ERROR_UNDEFINED_EXCEP,
                                   ("undefined exception code"));
            break;
    }/* end switch */

    return IX_SUCCESS;

freeReassembledPacket:
    IX_ERROR_CRT(
        ix_rm_buffer_free_chain(hReassembledPacket),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    return err;

freePacket:
    IX_ERROR_CRT(
        ix_rm_buffer_free_chain(arg_hPacket),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    return err;

} /* end ix_cc_v6v4_microblock_pkt_handler */



/**
 * NAME: ix_cc_v6v4_ipv6_pkt_handler
 *
 * DESCRIPTION: This is the registered function to receive IPv6 packets from
 * other core components. 
 * 
 * @Param:  - IN arg_hPacket: handle to a buffer which contains IPv6 packet.
 * @Param: - IN arg_ExceptionCode: not used.
 * @Param: - IN arg_pContext: pointer to Tunneling core component context that 
 * was defined by the core component and passed to the core component
 * infrastructure through ix_cc_v6v4_init().
 * @Param:  - OUT
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_SEND_FAIL - error from CCI
 * IX_CC_ERROR_INTERNAL - an internal error
 * IX_CC_ERROR_ALIGN - pointer not aligned properly
 * IX_CC_ERROR_NULL - null pointer
 * 
 */

ix_error ix_cc_v6v4_ipv6_pkt_handler(
    ix_buffer_handle arg_hPacket,  
    ix_uint32 arg_ExceptionCode,
    void* arg_pContext
    )
{
    ix_hw_buffer_meta *pHwMeta;
    ix_uint8 nextHopType;
    ix_error err;

    /**
     * The meta-data next hop ID type indicates whether the packet was
     * received on a local interface that supports tunneling.  If not,
     * we don't need to handle the packet.
     */

    IX_ERROR_CGT(
        ix_rm_buffer_get_meta(arg_hPacket, (void **)&pHwMeta),
        err,
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING,
        freePacket);

    nextHopType = pHwMeta->m_NextHopIDType;
    if (nextHopType != IX_CC_V6V4_NEXT_HOP_TUNNEL)
    {
        IX_ERROR_CGT(
            ix_rm_packet_send(
                IX_CC_V6V4_PKT_DEFAULT_OUTPUT, arg_hPacket, 0),
            err, IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING,
            freePacket);
    }
    else
    {
        IX_ERROR_CG(
            _ix_cc_v6v4_handle_ipv6_packet(arg_hPacket, arg_pContext),
            err, freePacket);
    }

    return IX_SUCCESS;

freePacket:
    IX_ERROR_CRT(
        ix_rm_buffer_free_chain(arg_hPacket),
        IX_CC_ERROR_INTERNAL,0);

    return err;

} /* end ix_cc_v6v4_ipv6_pkt_handler */


/**
 * NAME: ix_cc_v6v4_ipv4_pkt_handler
 *
 * DESCRIPTION: This is the registered function to receive IPv4 packets from
 * other core components. 
 * 
 * @Param:  - IN arg_hPacket: handle to a buffer which contains IPv4 packet.
 * @Param: - IN arg_ExceptionCode: not used.
 * @Param: - IN arg_pContext: pointer to Tunneling core component context that 
 * was defined by the core component and passed to the core component
 * infrastructure through ix_cc_v6v4_init().
 * @Param:  - OUT
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_CCI - error from CCI
 * 
 */

ix_error ix_cc_v6v4_ipv4_pkt_handler(
    ix_buffer_handle arg_hPacket,  
    ix_uint32 arg_ExceptionCode,
    void* arg_pContext
    )
{
    ix_hw_buffer_meta *pHwMeta;
    ix_uint8* pData;
    ix_uint8 nextHopType;
    ix_uint8 proto;
    ix_cc_v6v4_ipv4_header* pIpv4Hdr;
    ix_error err;

    /**
     * The meta-data next hop ID type indicates whether the packet was
     * received on a local interface that supports tunneling.  If not,
     * we don't need to handle the packet.
     */

    IX_ERROR_CGT(
        ix_rm_buffer_get_meta(arg_hPacket, (void **)&pHwMeta),
        err,
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING,
        freePacket);

    nextHopType = pHwMeta->m_NextHopIDType;
    if (nextHopType != IX_CC_V6V4_NEXT_HOP_TUNNEL)
    {
        IX_ERROR_CGT(
            ix_rm_packet_send(
                IX_CC_V6V4_PKT_DEFAULT_OUTPUT, arg_hPacket, 0),
            err, IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING,
            freePacket);
    }
    else
    {
        /* Get the packet data and check for a tunneled packet */

        IX_ERROR_CRT(
            ix_rm_buffer_get_data(arg_hPacket, (void **)&pData),
            IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

        pIpv4Hdr = (ix_cc_v6v4_ipv4_header*)(pData +
			IX_NTOH16(IX_RM_MEM_UINT16_READ(&pHwMeta->m_Offset)));

        proto = IX_RM_MEM_UINT8_READ(&pIpv4Hdr->protocol);

        if (proto != IX_CC_V6V4_IP_PROTO_IPV6)
        {
            IX_ERROR_CGT(
                ix_rm_packet_send(
                    IX_CC_V6V4_PKT_LOW_PRIORITY_STKDRV_OUTPUT,
                    arg_hPacket, 0),
                err, IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING,
                freePacket);
        }
        else
        {
            IX_ERROR_CG(
                _ix_cc_v6v4_handle_ipv4_packet(arg_hPacket, arg_pContext),
                err, freePacket);
        }
    }

    return IX_SUCCESS;

freePacket:
    IX_ERROR_CRT(
        ix_rm_buffer_free_chain(arg_hPacket),
        IX_CC_ERROR_INTERNAL,0);

    return err;

} /* end ix_cc_v6v4_ipv4_pkt_handler */
