/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 6
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      IPv6/v4 Tunneling Core Component - dynamic properties table management
 *
 * = FILENAME
 *      ix_cc_v6v4_prop.c
 *
 * = DESCRIPTION
 *      The file contains dynamic properties table management interface
 *      functions.
 *
 * = CREATION TIME
 *      11/14/2002 11:25 AM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */


#define IX_ERROR_FILE_IDENT "$Id: ix_cc_v6v4_prop.c,v 1.4 2003/03/12 16:31:35 cabell Exp $" 

/**
 * System defined include files
 */

/**
 * User defined include files
 */

#include "ix_cc_error.h"

#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_netmacros.h"

#include "ix_cc.h"
#include "ix_cc_properties.h"
#include "bindings.h"
#include "cc/ix_cc_v6v4.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_init.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_prop.h"





/**
 * Preprocessor symbols and Macros used in this file
 */

/**
 * Type definitions whose scope is limited to this file
 */

/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */
extern ix_cc_v6v4_context* g_pV6v4Context;
extern ix_cc_v6v4_registry_data g_V6v4RegistryData;
/** 
 * Extern function prototypes.
 */


/** 
 * Static function prototypes.
 */


/** 
 * Function definitions.
 */


/**
 * NAME: _ix_cc_v6v4_prop_get_port_ip_address
 *
 * DESCRIPTION: This primitive retrieves the IP address of a port
 *           .
 * 
 * @Param:  - IN arg_BladeId - blade number of ip address to be retrieved
 * @Param:  - IN arg_PortId - port number of ip address to be retrieved
 *
 * @Param:  - OUT arg_pIpAddress - IP address of given blade and port
 *
 * @Return: IX_SUCCESS if successfull or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM - invalid blade/port number
 * IX_CC_ERROR_ENTRY_NOT_FOUND - no data in location
 */

ix_error _ix_cc_v6v4_get_port_ip_address(
    ix_uint16 arg_BladeId,
    ix_uint16 arg_PortId, 
    ix_uint32 *arg_pIpAddr
    )
{
    ix_uint32 keyIndex;

    /* Validate Port and blade Ids */
    if ((arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
        (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
    {
        return IX_ERROR_WARNING(
            IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
            ("Invalid Blade Id % or Port Id %d",arg_BladeId,arg_PortId));
    }/* endif ((arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
               (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
      */

    /*
     * Calculate keyIndex from Blade and Port Id
     * 6 bits for Blade Id(totally 64 blades) and each blade can have up to
     * 16 ports
     * 4 bits to store portIds
     */

    keyIndex = (arg_BladeId << 4) | arg_PortId;
     
    /* store ipAddr from dynamic property table to arg_pIpAddr */
    if (g_pV6v4Context->aIfPropTable[keyIndex].ipAddr == 0)
    {
        *arg_pIpAddr = 0;
        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                                ("No Ip Address in location"));
    }
    else
    {
        *arg_pIpAddr = g_pV6v4Context->aIfPropTable[keyIndex].ipAddr;
    }
    
    return IX_SUCCESS;
}/* end _ix_cc_v6v4_get_port_ip_address */



/**
 * NAME: _ix_cc_v6v4_prop_get_port_broadcast_address
 *
 * DESCRIPTION: This primitive is used to get the broadcast address of a
 * port
 *           .
 * @Param:  - IN arg_BladeId - blade number for interface
 * @Param:  - IN arg_PortId - port number for interface
 *
 * @Param:  - OUT arg_pBdcastAddr - Broadcast address of given blade and port
 *
 * @Return: IX_SUCCESS if successfull or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM - invalid blade/port number
 * IX_CC_ERROR_ENTRY_NOT_FOUND - no data in location
 */

ix_error _ix_cc_v6v4_prop_get_port_broadcast_address(
    ix_uint16 arg_BladeId,
    ix_uint16 arg_PortId, 
    ix_uint32 *arg_pBdcastAddr
    )
{
    ix_uint32 keyIndex;

    /* Validate Port and blade Ids */
    if ((arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
        (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
    {
        return IX_ERROR_WARNING(
            IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
            ("Invalid Blade Id % or Port Id %d",arg_BladeId,arg_PortId));
    }/* endif ((arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
               (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
      */

    /*
     * Calculate keyIndex from Blade and Port Id
     * 6 bits for Blade Id(totally 64 blades) and each blade can have up to
     * 16 ports
     * 4 bits to store portIds
     */

    keyIndex = (arg_BladeId << 4) | arg_PortId;
  
    /* ipAddr from dynamic property table must be non-zero for this entry
     * to be valid
     */
    if (g_pV6v4Context->aIfPropTable[keyIndex].ipAddr == 0)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                                ("Interface is not valid"));
    }
    else
    {
        *arg_pBdcastAddr =
            g_pV6v4Context->aIfPropTable[keyIndex].broadcast;
    }
    
    return IX_SUCCESS;
}/* end _ix_cc_v6v4_prop_get_port_broadcast_address */



/**
 * NAME: _ix_cc_v6v4_prop_get_port_mtu
 *
 * DESCRIPTION: This primitive is used to get the mtu for a port.
 *           .
 * 
 * @Param:  - IN arg_BladeId - blade number of interface
 * @Param:  - IN arg_PortId - port number of interface
 *
 * @Param:  - OUT arg_pMtu - Mtu for of given blade and port
 *
 * @Return: IX_SUCCESS if successfull or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM - invalid blade/port number
 * IX_CC_ERROR_ENTRY_NOT_FOUND - no data in location
 */

ix_error _ix_cc_v6v4_prop_get_port_mtu(
    ix_uint16 arg_BladeId,
    ix_uint16 arg_PortId, 
    ix_uint16 *arg_pMtu
    )
{
    ix_uint32 keyIndex;

    /* Validate Port and blade Ids */
    if ((arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
        (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
    {
        return IX_ERROR_WARNING(
            IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
            ("Invalid Blade Id % or Port Id %d",arg_BladeId,arg_PortId));
    }/* endif ((arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
               (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
      */

    /*
     * Calculate keyIndex from Blade and Port Id
     * 6 bits for Blade Id(totally 64 blades) and each blade can have up to
     * 16 ports
     * 4 bits to store portIds
     */

    keyIndex = (arg_BladeId << 4) | arg_PortId;
     
    /* ipAddr from dynamic property table must be non-zero for this entry
     * to be valid
     */
    if (g_pV6v4Context->aIfPropTable[keyIndex].ipAddr == 0)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                                ("No MTU in location"));
    }
    else
    {
        *arg_pMtu =
            g_pV6v4Context->aIfPropTable[keyIndex].mtu;
    }
    
    return IX_SUCCESS;
}/* end _ix_cc_v6v4_prop_get_port_mtu */



/**
 * NAME: _ix_cc_v6v4_property_update_ip_address
 *
 * DESCRIPTION: This primitive is used to add/delete ip address and
 * net mask in the dynamic property table.
 *           .
 * 
 * @Param:  - IN arg_PropId - property id : add/delete IP address
 * @Param:  - IN arg_BladeId - blade number
 * @Param:  - IN arg_PordId - port number
 * @Param:  - IN arg_IpAddr - IP address
 * @Param:  - IN arg_NetMask - Subnet mask
 *
 * @Param:  - OUT none
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM - invalid blade/port number 
 * 
 */

ix_error _ix_cc_v6v4_property_update_ip_address(
    ix_uint16 arg_PropId,
    ix_uint32 arg_BladeId,
    ix_uint32 arg_PortId,
    ix_uint32 arg_IpAddr,
    ix_uint32 arg_NetMask
    )
{
    ix_uint32 keyIndex;

    /* Validate input arguments */
    if((arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
      (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                                ("Invalid Blade Id or Port Id"));
    }/* endif (arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
      (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
      */

    /*
     * Calculate keyIndex from Blade and Port Id
     * 6 bits for Blade Id(totally 64 blades) and each blade can have up to
     * 16 ports
     * 4 bits to store portIds
     */
    keyIndex = (arg_BladeId << 4) | arg_PortId;    

    if (arg_PropId == IX_CC_SET_PROPERTY_ADD_INTERFACE_IPV4)
    {
        /* store ipAddr and netMask to  dynamic property table */
        ix_ossl_mutex_lock(g_pV6v4Context->v6v4Mutex,IX_OSSL_WAIT_FOREVER);
        g_pV6v4Context->aIfPropTable[keyIndex].ipAddr = arg_IpAddr;
        g_pV6v4Context->aIfPropTable[keyIndex].netMask = arg_NetMask;
        ix_ossl_mutex_unlock(g_pV6v4Context->v6v4Mutex); 


    }/* end if (arg_PropId == IX_CC_PROPERTY_ADD_V6V4_ADDRESS)*/

    if (arg_PropId == IX_CC_SET_PROPERTY_DEL_INTERFACE_IPV4)
    {
        ix_ossl_mutex_lock(g_pV6v4Context->v6v4Mutex,IX_OSSL_WAIT_FOREVER);
        g_pV6v4Context->aIfPropTable[keyIndex].ipAddr =  0;
        g_pV6v4Context->aIfPropTable[keyIndex].netMask = 0;
        ix_ossl_mutex_unlock(g_pV6v4Context->v6v4Mutex);
    }/* endif  */
 
    return IX_SUCCESS;    
}/* end _ix_cc_v6v4_property_update_ip_address */



/**
 * NAME: _ix_cc_v6v4_property_update_broadcast
 *
 * DESCRIPTION: This primitive is used to add/delete broadcast address
 * in the dynamic property table.
 *           .
 * 
 * @Param:  - IN arg_PropId - property id : add/delete broadcast
 * @Param:  - IN arg_BladeId - blade number
 * @Param:  - IN arg_PordId - port number
 * @Param:  - IN arg_BroadcastAddr - broadcast address
 *
 * @Param:  - OUT none
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM - invalid blade/port number 
 * 
 */
ix_error _ix_cc_v6v4_property_update_broadcast(
    ix_uint16 arg_PropId,
    ix_uint32 arg_BladeId,
    ix_uint32 arg_PortId,
    ix_uint32 arg_BroadcastAddr
    )
{
    ix_uint32 keyIndex;

    /* Validate input arguments */
    if((arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
      (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                                ("Invalid Blade Id or Port Id"));
    }/* endif (arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
      (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
      */

    /*
     * Calculate keyIndex from Blade and Port Id
     * 6 bits for Blade Id(totally 64 blades) and each blade can have up to
     * 16 ports
     * 4 bits to store portIds
     */
    keyIndex = (arg_BladeId << 4) | arg_PortId;    

    if(arg_PropId == IX_CC_SET_PROPERTY_ADD_BROADCAST_IPV4)
    {
        /* store broadcast address in  dynamic property table */
        ix_ossl_mutex_lock(g_pV6v4Context->v6v4Mutex,IX_OSSL_WAIT_FOREVER);
        g_pV6v4Context->aIfPropTable[keyIndex].broadcast = arg_BroadcastAddr;
        ix_ossl_mutex_unlock(g_pV6v4Context->v6v4Mutex); 


    }/* end if (arg_PropId == IX_CC_SET_PROPERTY_ADD_BROADCAST_IPV4)*/

    if(arg_PropId == IX_CC_SET_PROPERTY_DEL_BROADCAST_IPV4)
    {
        ix_ossl_mutex_lock(g_pV6v4Context->v6v4Mutex,IX_OSSL_WAIT_FOREVER);
        g_pV6v4Context->aIfPropTable[keyIndex].broadcast =  0;
        ix_ossl_mutex_unlock(g_pV6v4Context->v6v4Mutex);
    }/* endif  */
 
    return IX_SUCCESS;    
}


/**
 * NAME: _ix_cc_v6v4_property_update_mtu
 *
 * DESCRIPTION: This primitive is used to update the MTU of an interface
 * in the dynamic property table.
 *           .
 * 
 * @Param:  - IN arg_PropId - property id : set MTU
 * @Param:  - IN arg_BladeId - blade number
 * @Param:  - IN arg_PordId - port number
 * @Param:  - IN arg_Mtu - MTU
 *
 * @Param:  - OUT none
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM - invalid blade/port number 
 * 
 */

ix_error _ix_cc_v6v4_property_update_mtu(
    ix_uint32 arg_BladeId,
    ix_uint32 arg_PortId,
    ix_uint16 arg_Mtu
    )
{
    ix_uint32 keyIndex;

    /* Validate input arguments */
    if((arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
      (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                                ("Invalid Blade Id or Port Id"));
    }/* endif (arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
      (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
      */

    /*
     * Calculate keyIndex from Blade and Port Id
     * 6 bits for Blade Id(totally 64 blades) and each blade can have up to
     * 16 ports
     * 4 bits to store portIds
     */
    keyIndex = (arg_BladeId << 4) | arg_PortId;    

    /* store new mtu in dynamic property table */
    ix_ossl_mutex_lock(g_pV6v4Context->v6v4Mutex,IX_OSSL_WAIT_FOREVER);
    g_pV6v4Context->aIfPropTable[keyIndex].mtu = arg_Mtu;
    ix_ossl_mutex_unlock(g_pV6v4Context->v6v4Mutex); 

    return IX_SUCCESS;    
}/* end _ix_cc_v6v4_property_update_mtu */







/**
 * NAME: _ix_cc_v6v4_prop_table_lookup_ip_address
 *
 * DESCRIPTION: This primitive is used to lookup an ip address in the
 * dynamic property table and return the associated blade ID and port ID.
 *           
 * @Param:  - IN arg_IpAddr - ip address(non zero) in network byte order
 * to be searched in table.
 *
 * @Param:  - OUT arg_pBladeId - pointer to location in which to return the
 * blade ID associated with the IP address.
 *
 * @Param:  - OUT arg_pPortId - pointer to location in which to return the
 * port ID associated with the IP address.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - no entry in table 
 * 
 */

ix_error _ix_cc_v6v4_prop_table_lookup_ip_address(
    ix_uint32 arg_IpAddr,
    ix_uint16* arg_pBladeId,
    ix_uint16* arg_pPortId)
{
    ix_uint32 keyIndex;

    for(keyIndex = 0; keyIndex < TOTAL_NUMBER_OF_PORTS; keyIndex++)
    {
        if(g_pV6v4Context->aIfPropTable[keyIndex].ipAddr == arg_IpAddr)
        {
            *arg_pBladeId = (ix_uint16)(keyIndex >> 4);
            *arg_pPortId = (ix_uint16)(keyIndex & 0xF);

            return IX_SUCCESS;
        }/* end if */
    }/* end of table search */

    return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,("no entry in table"));
}/* end _ix_cc_v6v4_prop_table_lookup_ip_address */
