/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      IPv6/v4 Tunneling Core Component - IPv4 reassembly support.
 *
 * = FILENAME
 *      ix_cc_v6v4_reassembly.c
 *
 * = DESCRIPTION
 *      The file contains the functions and macros needed to reassembly
 *      of IPv4 packets.
 *
 * = CREATION TIME
 *      01/06/2003 12:39 PM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_v6v4_reassembly.c,v 1.6 2003/10/31 17:26:34 ktseng Exp $";

/**
 * System defined include files
 */

/**
 * User defined include files
 */
#include "ix_cc_error.h"

#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_netmacros.h"

#include "ix_cc.h"
#include "ix_cc_macros.h"
#include "ix_config.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_reassembly.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_header_defs.h"


/**
 * Preprocessor symbols and Macros used in this file
 */

#define IX_CC_V6V4_REASM_HASH_BUCKETS       64
#define IX_CC_V6V4_REASM_MAX_PACKETS        256
#define IX_CC_V6V4_REASM_FRAG_MAP_BLOCKS    128

#define IPV4_HDR_MF                         0x2000

#define IX_CC_V6V4_REASM_START_OFFSET       192


/*
 * NAME: _IX_CC_V6V4_REASM_LIST_INIT()
 *
 * DESCRIPTION: Initializes the reassemby list header.
 *
 * @Param:      - IN    __lh
 *                      List header.
 */
#define _IX_CC_V6V4_REASM_LIST_INIT(__lh)                                   \
    (__lh).pFirst = &(__lh).pTail;                                          \
    (__lh).pTail = NULL;                                                    \
    (__lh).pLast = &(__lh).pFirst;


/*
 * NAME: _IX_CC_V6V4_REASM_LIST_INSERT()
 *
 * DESCRIPTION: Appends a reassembly data structure at the end of
 * the list.
 *
 * @Param:      - IN    __lh
 *                      List header.
 *
 * @Param:      - IN    __prd
 *                      Pointer to reassembly data structure to insert.
 */
#define _IX_CC_V6V4_REASM_LIST_INSERT(__lh, __prd)                          \
    (__prd)->pNext = (ix_cc_v6v4_reasm_data*)&((__lh).pTail);               \
    (__prd)->pPrev = (ix_cc_v6v4_reasm_data*)((__lh).pLast);                \
    ((ix_cc_v6v4_reasm_data*)(__lh).pLast)->pNext = (__prd);                \
    (__lh).pLast = (__prd);


/*
 * NAME: _IX_CC_V6V4_REASM_LIST_REMOVE()
 *
 * DESCRIPTION: Removes a reassembly data structure from the list.
 *
 * @Param:      - IN    __prd
 *                      Pointer to reassembly data structure to remove.
 */
#define _IX_CC_V6V4_REASM_LIST_REMOVE(__prd)                                \
    (__prd)->pPrev->pNext = (__prd)->pNext;                                 \
    (__prd)->pNext->pPrev = (__prd)->pPrev;



/**
 * Type definitions whose scope is limited to this file
 */

/**
 * TYPENAME: ix_cc_v6v4_reasm_frag_map
 * 
 * DESCRIPTION: This structure contains information to track which
 * fragments have been received for a packet begin reassembled.
 *
 */

typedef struct ix_s_cc_v6v4_reasm_frag_map
{
    struct ix_s_cc_v6v4_reasm_frag_map* pNext;
    ix_uint16                           start;
    ix_uint16                           end;
} ix_cc_v6v4_reasm_frag_map;

/**
 * TYPENAME: ix_cc_v6v4_reasm_data
 * 
 * DESCRIPTION: This structure holds the information needed to
 * reassemble one packet.
 *
 */

typedef struct ix_s_cc_v6v4_reasm_data
{
    struct ix_s_cc_v6v4_reasm_data* pNext;
    struct ix_s_cc_v6v4_reasm_data* pPrev;
    ix_uint128                      key;
    ix_cc_v6v4_reasm_frag_map       fragMap;
    ix_uint16                       hdrSize;
    ix_uint16                       dataSize;
    ix_uint32                       ticks;
    ix_buffer_handle                hStartOfChain;
    ix_buffer_handle                hEndOfChain;
    ix_buffer_handle                hLastBuffer;
} ix_cc_v6v4_reasm_data;

/**
 * TYPENAME: ix_cc_v6v4_reasm_list
 * 
 * DESCRIPTION: This structure defines a list header for reassembly
 * data structures.
 *
 */

typedef struct ix_s_cc_v6v4_reasm_list
{
    void* pFirst;
    void* pTail;
    void* pLast;
} ix_cc_v6v4_reasm_list;

/**
 * TYPENAME: ix_cc_v6v4_reassembly
 * 
 * DESCRIPTION: This structure holds the global information needed to
 * support packet reassembly.
 *
 */

struct ix_s_cc_v6v4_reassembly
{
    ix_buffer_free_list_handle  hBufferFreeList;
    ix_uint32                   freeListId;
    ix_uint32                   bufferSize;
    ix_uint32                   reassemblyTimeout;
    ix_cc_hash_table128         hReassemblyDataHash;
    ix_cc_v6v4_reasm_list       reassemblyDataList;
    ix_cc_index_pool            fragMapIndexPool;
    ix_cc_v6v4_reasm_frag_map   aFragMap[IX_CC_V6V4_REASM_FRAG_MAP_BLOCKS];
};

/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */


/** 
 * Extern function prototypes.
 */


/** 
 * Static function prototypes.
 */

PRIVATE ix_error _ix_cc_v6v4_reasm_free_data(
    ix_cc_v6v4_reassembly arg_hReasm,
    ix_cc_v6v4_reasm_data* arg_pData
    );

PRIVATE ix_error _ix_cc_v6v4_reasm_copy_fragment(
    ix_cc_v6v4_reassembly arg_hReasm,
    ix_cc_v6v4_reasm_data* arg_pData,
    ix_cc_v6v4_ipv4_header* arg_pFragHdr,
    ix_buffer_handle arg_hFragment
    );

PRIVATE ix_error _ix_cc_v6v4_reasm_link_buffer(
    ix_cc_v6v4_reassembly arg_hReasm,
    ix_cc_v6v4_reasm_data* arg_pData,
    ix_uint16 arg_BufSize);

PRIVATE ix_error _ix_cc_v6v4_reasm_copy_frag_data(
    ix_buffer_handle arg_hPacket,
    ix_buffer_handle arg_hFragment,
    ix_uint16 arg_PktOffset,
    ix_uint16 arg_FragOffset,
    ix_uint16* arg_ByteCount
    );

PRIVATE ix_error _ix_cc_v6v4_reasm_update_frag_map(
    ix_cc_v6v4_reassembly arg_hReasm,
    ix_cc_v6v4_reasm_data* arg_pData,
    ix_uint16 arg_FragOffset,
    ix_uint16 arg_DataBytes
    );

PRIVATE ix_error _ix_cc_v6v4_reasm_packet_complete(
    ix_cc_v6v4_reassembly arg_hReasm,
    ix_cc_v6v4_reasm_data* arg_pData,
    ix_hw_buffer_meta* arg_pFragMeta
    );

/** 
 * Function definitions.
 */

/**
 * NAME: _ix_cc_v6v4_reasm_init
 *
 * DESCRIPTION: This primitive is used to initialize the IPv4 packet
 * reassembly resources.
 *           
 * @Param:  - IN arg_ReasmTimeout - number of ticks after a fragment
 * arrives to wait for the next fragment before releasing the reassembly
 * data for a packet.
 *
 * @Param:  - IN arg_hFreeList - handle to buffer free list.
 *
 * @Param:  - OUT arg_phReasm - pointer to location in which to return the
 * reassembly context.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_OOR - out of resources 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */

ix_error _ix_cc_v6v4_reasm_init(
    ix_uint32 arg_ReasmTimeout,
    ix_buffer_free_list_handle arg_hFreeList,
    ix_cc_v6v4_reassembly* arg_phReasm
    )
{
    ix_error err;
    ix_cc_v6v4_reassembly pReasmContext;
    ix_buffer_free_list_info bufferInfo;

    IX_ERROR_CRT(
        ix_rm_buffer_free_list_get_info(arg_hFreeList, &bufferInfo),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    if (arg_ReasmTimeout == 0)
    {
        return IX_ERROR_WARNING(
                    IX_CC_ERROR_RANGE,
                    ("Reassembly timeout of zero is not allowed"));
    }/* end if (arg_ReasmTimeout == 0) */

    pReasmContext = (ix_cc_v6v4_reassembly)
        ix_ossl_malloc(sizeof(struct ix_s_cc_v6v4_reassembly));

    if (pReasmContext == NULL)
    {
        return IX_ERROR_WARNING(
                    IX_CC_ERROR_OOM,
                    ("Failed to allocate memory for reassembly context"));
    }/* end if (pReasmContext == NULL) */

    /* initialize the structure */
    ix_ossl_memset(pReasmContext, 0, sizeof(ix_cc_v6v4_reassembly));

    pReasmContext->reassemblyTimeout = arg_ReasmTimeout;
    pReasmContext->bufferSize = bufferInfo.m_DataElementSize;
    pReasmContext->freeListId = bufferInfo.m_FreeListInfo1;

    _IX_CC_V6V4_REASM_LIST_INIT(pReasmContext->reassemblyDataList);

    /* Allocate the hash resources */

    IX_ERROR_CG(ix_cc_hash_table128_init(
                    IX_CC_V6V4_REASM_HASH_BUCKETS,
                    IX_CC_V6V4_REASM_MAX_PACKETS,
                    IX_CC_HASH_TABLE_ALGO_XOR8,
                    &pReasmContext->hReassemblyDataHash),
                err,
                cleanupContext);

    /* Allocate the index pool for the map blocks */

    IX_ERROR_CG(ix_cc_index_pool_init(
                    IX_CC_V6V4_REASM_FRAG_MAP_BLOCKS,
                    0,
                    &pReasmContext->fragMapIndexPool),
                err,
                cleanupHash);

    *arg_phReasm = pReasmContext;
    return IX_SUCCESS;

cleanupHash:
    ix_cc_hash_table128_fini(pReasmContext->hReassemblyDataHash);

cleanupContext:
    ix_ossl_free(pReasmContext);
    return err;
} /* end - _ix_cc_v6v4_reasm_init */



/**
 * NAME: _ix_cc_v6v4_reasm_cleanup
 *
 * DESCRIPTION: This primitive is used to release resources previously
 * allocated for IPv4 packet reassembly.
 *           
 * @Param:  - IN arg_hReasm - reassembly context handle returned from
 * _ix_cc_v6v4_reasm_init.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */

ix_error _ix_cc_v6v4_reasm_cleanup(
    ix_cc_v6v4_reassembly arg_hReasm
    )
{
    ix_cc_v6v4_reasm_data* pData;
    ix_cc_v6v4_reasm_data* pNext;

    if (arg_hReasm == NULL)
    {
        return IX_ERROR_WARNING(
                    IX_CC_ERROR_NULL,
                    ("Reassembly context is NULL"));
    } /* if (arg_hReasm == NULL) */

    pData = (ix_cc_v6v4_reasm_data*)arg_hReasm->reassemblyDataList.pFirst;
    while (pData->pNext != NULL)
    {
        pNext = pData->pNext;
        _ix_cc_v6v4_reasm_free_data(arg_hReasm, pData);
        pData = pNext;
    }

    IX_ERROR_CRT(ix_cc_index_pool_fini(arg_hReasm->fragMapIndexPool),
                 IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    IX_ERROR_CRT(ix_cc_hash_table128_fini(arg_hReasm->hReassemblyDataHash),
                 IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    ix_ossl_free(arg_hReasm);
    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_reasm_cleanup */



/**
 * NAME: _ix_cc_v6v4_reasm_add_fragment
 *
 * DESCRIPTION: This primitive is used to add a new fragment to its reassembly
 * buffer.  If this is the first fragment received for the packet, new
 * reassembly resources are allocated.
 *           
 * @Param:  - IN arg_hReasm - reassembly context handle returned from
 * _ix_cc_v6v4_reasm_init.
 *
 * @Param:  - IN arg_hFragment - handle to buffer containing the fragment.
 *
 * @Param:  - OUT arg_phPacket - If the return value from this function is
 * IX_SUCCESS then this output parameter will be set to the  buffer handle
 * of the reassembled packet if the reassembly is complete, or to
 * IX_NULL_BUFFER_HANDLE if more fragments are needed before the packet
 * can be processed further.  If the return value from this function is not
 * IX_SUCCESS then this parameter has no meaning.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_OOR - out of resources
 * IX_CC_ERROR_OOM - out of memory
 * IX_CC_ERROR_INTERNAL - internal error occurred
 * 
 */

ix_error _ix_cc_v6v4_reasm_add_fragment(
    ix_cc_v6v4_reassembly arg_hReasm,
    ix_buffer_handle arg_hFragment,
    ix_buffer_handle* arg_phPacket
    )
{
    ix_error err;
	ix_uint8* pFragData;
    ix_hw_buffer_meta *pFragMeta;
    ix_hw_buffer_meta *pMeta;
    ix_cc_v6v4_ipv4_header ipv4Hdr;
    ix_cc_v6v4_ipv4_header* pIpv4Hdr;
    ix_uint128 key;
    ix_cc_v6v4_reasm_data* pData;
    int exists;
    ix_uint8 hdrSize;

    /**
     * Get packet data and extract the required header information.
     */
    IX_ERROR_CRT(
        ix_rm_buffer_get_meta(arg_hFragment, (void **)&pFragMeta),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    IX_ERROR_CRT(
        ix_rm_buffer_get_data(arg_hFragment, (void **)&pFragData),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    pIpv4Hdr = (ix_cc_v6v4_ipv4_header*)
		(pFragData + IX_NTOH16(IX_RM_MEM_UINT16_READ(&pFragMeta->m_Offset)));

    hdrSize = (pIpv4Hdr->vers_hdrLength & 0xf) << 2;

    ix_ossl_memcpy(&ipv4Hdr, pIpv4Hdr, hdrSize);

    key.aUint32[0] = ipv4Hdr.id;
    key.aUint32[1] = ipv4Hdr.protocol;
    key.aUint32[2] = ipv4Hdr.sourceAddress;
    key.aUint32[3] = ipv4Hdr.destAddress;

    IX_ERROR_CR(ix_cc_hash_table128_lookup(
                    arg_hReasm->hReassemblyDataHash, key,
                    &exists, (void **)&pData));

    if (!exists)
    {
        pData = (ix_cc_v6v4_reasm_data*)
                    ix_ossl_malloc(sizeof(ix_cc_v6v4_reasm_data));
        if (pData == NULL)
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_OOM,
                    ("Unable to allocate memory for reassembly resources"));
        }

        ix_ossl_memset(pData, 0, sizeof(ix_cc_v6v4_reasm_data));
        pData->ticks = arg_hReasm->reassemblyTimeout;
        pData->key = key;

        IX_ERROR_CG(ix_cc_hash_table128_add_entry(
                        arg_hReasm->hReassemblyDataHash, key, pData),
                    err, freeData);

        IX_ERROR_CGT(ix_rm_buffer_alloc(
                        arg_hReasm->hBufferFreeList, &pData->hLastBuffer),
                    err, IX_CC_ERROR_OOR, IX_ERROR_LEVEL_WARNING,
                    freeHash);

        IX_ERROR_CGT(
            ix_rm_buffer_get_meta(pData->hLastBuffer, (void **)&pMeta),
            err, IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING,
            freeBuffer);

        IX_RM_MEM_UINT32_WRITE(&pMeta->m_Offset,
                IX_NTOH16(IX_CC_V6V4_REASM_START_OFFSET));
        IX_RM_MEM_UINT32_WRITE(&pMeta->m_HwNext,
                IX_HTON32(IX_NULL_BUFFER_HANDLE));
        IX_RM_MEM_UINT32_WRITE(&pMeta->m_PacketNext, 0);

        _IX_CC_V6V4_REASM_LIST_INSERT(
            arg_hReasm->reassemblyDataList, pData);
    }

    IX_ERROR_CG(_ix_cc_v6v4_reasm_copy_fragment(
            arg_hReasm, pData, &ipv4Hdr, arg_hFragment),
            err, freePacket);

    if ((pData->fragMap.pNext == NULL) &
        (pData->fragMap.end == pData->dataSize))
    {
        /**
         * Packet is completely reassembled.  Update the meta data
         * and header fields appropriately.
         */

        IX_ERROR_CR(_ix_cc_v6v4_reasm_packet_complete(
                        arg_hReasm, pData, pFragMeta));

        *arg_phPacket = pData->hStartOfChain;
        pData->hStartOfChain = IX_NULL_BUFFER_HANDLE;
        _ix_cc_v6v4_reasm_free_data(arg_hReasm, pData);
    }
    else
    {
        *arg_phPacket = IX_NULL_BUFFER_HANDLE;
    }

    return IX_SUCCESS;

freePacket:
    _ix_cc_v6v4_reasm_free_data(arg_hReasm, pData);
    return err;

freeBuffer:
    ix_rm_buffer_free_chain(pData->hLastBuffer);

freeHash:
    ix_cc_hash_table128_delete_entry(arg_hReasm->hReassemblyDataHash, key);

freeData:
    ix_ossl_free(pData);
    return err;

} /* end - _ix_cc_v6v4_reasm_add_fragment */


/**
 * NAME: _ix_cc_v6v4_reasm_copy_fragment
 *
 * DESCRIPTION: This primitive is called to copy a fragment into the
 * correct location in the reassembly buffer.  Note that we could just
 * save the fragments and link them into a larger chain, but we don't do
 * this because if the fragments are small, a lot of buffer space would
 * be consumed during the time the packet is reassembled.
 *           
 * @Param:  - IN arg_hReasm - reassembly context handle returned from
 * _ix_cc_v6v4_reasm_init.
 *
 * @Param:  - IN arg_pData - pointer to reassembly data.
 *
 * @Param:  - IN arg_pFragHdr - pointer to aligned copy of IPv4 header
 * of fragment.
 *
 * @Param:  - IN arg_hFragment - handle to buffer containing fragment.
 *
 * @Param:  - OUT arg_pComplete - indicates whether the packet is completely
 * reassembled with the addition of this fragment.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_reasm_copy_fragment(
    ix_cc_v6v4_reassembly arg_hReasm,
    ix_cc_v6v4_reasm_data* arg_pData,
    ix_cc_v6v4_ipv4_header* arg_pFragHdr,
    ix_buffer_handle arg_hFragment
    )
{
    ix_uint16 fragOffset;    /* offset into packet where fragment starts */
    ix_uint16 dataBytes;     /* data bytes in fragment */
    ix_uint16 hdrBytes;      /* header bytes in fragment */
    ix_uint16 pktOffset;     /* offset into packet where current packet buffer
                              * starts */
    ix_uint16 pktStart;      /* offset into current packet buffer where
                              * next data byte is to be written */
    ix_uint16 curPktOffset;  /* offset into the packet where the next 
                                fragment byte is to be copied */
    ix_uint16 curFragOffset; /* offset into the fragment where the next
                                byte is to be copied */
    ix_uint16 pktBufSize;    /* number of data bytes available in current
                                packet buffer */
    ix_uint16 bytesCopied;   /* current number of bytes copied from the fragment
                                into the packet buffer */
    ix_uint16 byteCount;     /* number of bytes copied into the current
                                packet buffer */
    ix_buffer_handle hBuffer;
    ix_buffer_handle hPrevBuffer;
    ix_hw_buffer_meta *pMeta;
    ix_uint8* pPktData;
    int lastFrag;


    hdrBytes = ((arg_pFragHdr->vers_hdrLength) & 0xf) << 2;
    dataBytes = IX_NTOH16(arg_pFragHdr->totalLength) - hdrBytes;
    fragOffset = IX_NTOH16(arg_pFragHdr->flags_fragOffset);
    lastFrag = ((fragOffset & IPV4_HDR_MF) == 0);
    fragOffset = (fragOffset & 0x1fff) << 3;

    hBuffer = arg_pData->hStartOfChain;
    hPrevBuffer = IX_NULL_BUFFER_HANDLE;
    bytesCopied = 0;
    pktOffset = 0;
    curPktOffset = fragOffset;
    curFragOffset = hdrBytes;

    while (bytesCopied < dataBytes)
    {
        if (hBuffer != IX_NULL_BUFFER_HANDLE)
        {
            /* Still working on previously chained buffers */

            IX_ERROR_CRT(
                ix_rm_buffer_get_meta(hBuffer, (void **)&pMeta),
                IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
        }
        else if ((arg_pData->hLastBuffer != IX_NULL_BUFFER_HANDLE) &&
                (arg_pData->hLastBuffer != hPrevBuffer))
        {
            /*
             * We have reached the end of the chained buffers, but we
             * still have the last buffer allocated, which hasn't been
             * added to the chain yet, since we don't have the final
             * length (hence we can't calculate the cell count).
             */
             hBuffer = arg_pData->hLastBuffer;

            IX_ERROR_CRT(
                ix_rm_buffer_get_meta(hBuffer, (void **)&pMeta),
                IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
        }
        else
        {
            /*
             * We need another buffer (the fragment contains bytes
             * beyond what the current chain holds.  First link the
             * last fragment to the existing chain.
             */
            IX_RM_MEM_UINT16_WRITE(&pMeta->m_BufferSize, IX_HTON16(pktBufSize));

            IX_ERROR_CR(_ix_cc_v6v4_reasm_link_buffer(
                arg_hReasm, arg_pData, pktBufSize));

            IX_ERROR_CRT(ix_rm_buffer_alloc(
                            arg_hReasm->hBufferFreeList, &hBuffer),
                        IX_CC_ERROR_OOR, IX_ERROR_LEVEL_WARNING);

            arg_pData->hLastBuffer = hBuffer;

            IX_ERROR_CRT(
                ix_rm_buffer_get_meta(hBuffer, (void **)&pMeta),
                IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

            IX_RM_MEM_UINT32_WRITE(&pMeta->m_Offset, 0);
            IX_RM_MEM_UINT32_WRITE(&pMeta->m_HwNext,
                    IX_HTON32(IX_NULL_BUFFER_HANDLE));
            IX_RM_MEM_UINT32_WRITE(&pMeta->m_PacketNext, 0);
        }

        pktBufSize = (ix_uint16)(arg_hReasm->bufferSize -
            IX_NTOH16(IX_RM_MEM_UINT16_READ(&pMeta->m_Offset)));

        if ((pktOffset + pktBufSize) > curPktOffset)
        {
            /* Copy as much as possible into this buffer */

            pktStart = curPktOffset - pktOffset;
            byteCount = pktBufSize - pktStart;
            if (byteCount > (dataBytes - bytesCopied))
            {
                byteCount = dataBytes - bytesCopied;
            }

            IX_ERROR_CR(
                _ix_cc_v6v4_reasm_copy_frag_data(
                    hBuffer, arg_hFragment,
                    pktStart, curFragOffset, &byteCount));

            bytesCopied += byteCount;
            curPktOffset += byteCount;
            curFragOffset += byteCount;
        }

        pktOffset += pktBufSize;

        hPrevBuffer = hBuffer;
        IX_ERROR_CR(ix_rm_buffer_get_next(hPrevBuffer, &hBuffer));
    } /* end - while (bytesCopied < dataBytes) */

    /**
     * If this is the first fragment, copy the header to the packet
     * buffer
     */
    if (fragOffset == 0)
    {
        arg_pData->hdrSize = hdrBytes;

        hBuffer = (arg_pData->hStartOfChain == IX_NULL_BUFFER_HANDLE) ?
            arg_pData->hLastBuffer : arg_pData->hStartOfChain;

        IX_ERROR_CRT(
            ix_rm_buffer_get_meta(hBuffer, (void **)&pMeta),
            IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

        IX_ERROR_CRT(
            ix_rm_buffer_get_data(hBuffer, (void **)&pPktData),
            IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

		pPktData += IX_NTOH16(IX_RM_MEM_UINT16_READ(&pMeta->m_Offset));

        ix_ossl_memcpy(pPktData - hdrBytes, arg_pFragHdr, hdrBytes);
    }

    /**
     * If this is the last fragment, update the packet size and
     * link the fragment into the chain.
     */
    if (lastFrag)
    {
        arg_pData->dataSize = fragOffset + dataBytes;

        if (arg_pData->hLastBuffer != IX_NULL_BUFFER_HANDLE)
        {
            ix_uint16 bufSize = pktStart + byteCount;

            if (arg_pData->hLastBuffer != hPrevBuffer)
            {
                return IX_ERROR_WARNING(
                    IX_CC_ERROR_INTERNAL, ("Internal error during reassembly"));
            }

            IX_RM_MEM_UINT16_WRITE(&pMeta->m_BufferSize, IX_HTON16(bufSize));

            IX_ERROR_CR(_ix_cc_v6v4_reasm_link_buffer(
                arg_hReasm, arg_pData, bufSize));
        }
    }

    IX_ERROR_CR(_ix_cc_v6v4_reasm_update_frag_map(
                    arg_hReasm, arg_pData, fragOffset, dataBytes));

    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_reasm_copy_fragment */


/**
 * NAME: _ix_cc_v6v4_reasm_link_buffer
 *
 * DESCRIPTION: This primitive is called to link the last allocated packet
 * buffer into the existing buffer chain.  This is called once we have enough
 * fragments to determine the buffer size (so the cell count can be
 * set accurately).
 *           
 * @Param:  - IN arg_hReasm - reassembly context handle returned from
 * _ix_cc_v6v4_reasm_init.
 *
 * @Param:  - IN arg_pData - pointer to reassembly data.
 *
 * @Param:  - IN arg_BufSize - size of last buffer.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */
PRIVATE ix_error _ix_cc_v6v4_reasm_link_buffer(
    ix_cc_v6v4_reassembly arg_hReasm,
    ix_cc_v6v4_reasm_data* arg_pData,
    ix_uint16 arg_BufSize)
{
    ix_uint32 cellCount;

    if (arg_pData->hEndOfChain != IX_NULL_BUFFER_HANDLE)
    {
        /* Link to end of existing chain */
        cellCount = ix_cc_calculate_cell_count((ix_uint32)arg_BufSize, 0);

        ix_rm_buffer_cell_count_set(&arg_pData->hLastBuffer, cellCount);

        IX_ERROR_CRT(
            ix_rm_buffer_link(&arg_pData->hEndOfChain, &arg_pData->hLastBuffer),
            IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    }
    else
    {
        /* 
         * No existing chain, so we have just a single buffer.
         */
        cellCount = ix_cc_calculate_cell_count((ix_uint32)arg_BufSize, 1);
        ix_rm_buffer_cell_count_set(&arg_pData->hLastBuffer, cellCount);

        arg_pData->hStartOfChain = arg_pData->hLastBuffer;
    }

    arg_pData->hEndOfChain = arg_pData->hLastBuffer;
    arg_pData->hLastBuffer = 0;
    return IX_SUCCESS;
} /* end - _ix_cc_v6v4_reasm_link_buffer */


/**
 * NAME: _ix_cc_v6v4_reasm_copy_frag_data
 *
 * DESCRIPTION: This primitive is called to copy as many bytes as
 * possible from the fragment buffer(s) into one of the packet buffers.
 * The copy operation will either fill the packet buffer or will copy
 * all bytes of the fragment between the starting offset and the end
 * of the fragment. 
 *           
 * @Param:  - IN arg_hPacket - handle to packet buffer.
 *
 * @Param:  - IN arg_hFragment - handle to fragment buffer.
 *
 * @Param:  - IN arg_PktOffset - offset into packet buffer where first
 * byte should be written.
 *
 * @Param:  - IN arg_FragOffset - offset into fragment where first byte
 * should be read.
 *
 * @Param:  - OUT arg_ByteCount - on input, points to the maximum number
 * of bytes to be copied.  On output, points to the actual number of
 * bytes copied.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_reasm_copy_frag_data(
    ix_buffer_handle arg_hPacket,
    ix_buffer_handle arg_hFragment,
    ix_uint16 arg_PktOffset,
    ix_uint16 arg_FragOffset,
    ix_uint16* arg_ByteCount
    )
{
    ix_hw_buffer_meta *pPktMeta;
    ix_hw_buffer_meta *pFragMeta;
    ix_uint8* pPktData;
    ix_uint8* pFragData;
    ix_uint16 bytesRemaining;
    ix_uint16 bytesCopied;
    ix_uint16 fragOffset;
    ix_uint16 curFragOffset;
    ix_uint16 pktOffset;
    ix_uint16 fragBufSize;
    ix_uint16 fragStart;
    ix_uint16 bytesToCopy;
    ix_buffer_handle hFragBuffer;


    IX_ERROR_CRT(
        ix_rm_buffer_get_meta(arg_hPacket, (void **)&pPktMeta),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    IX_ERROR_CRT(
        ix_rm_buffer_get_data(arg_hPacket, (void **)&pPktData),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

	pPktData += IX_NTOH16(IX_RM_MEM_UINT16_READ(&pPktMeta->m_Offset));


    /* Position to starting byte in fragment buffer */

    bytesRemaining = *arg_ByteCount;
    bytesCopied = 0;
    fragOffset = 0;
    curFragOffset = arg_FragOffset;
    pktOffset = arg_PktOffset;
    hFragBuffer = arg_hFragment;

    while ((bytesRemaining != 0) && (hFragBuffer != IX_NULL_BUFFER_HANDLE))
    {
        IX_ERROR_CRT(
            ix_rm_buffer_get_data(hFragBuffer, (void **)&pFragData),
            IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

        IX_ERROR_CRT(
            ix_rm_buffer_get_meta(hFragBuffer, (void **)&pFragMeta),
            IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

		pFragData += IX_NTOH16(IX_RM_MEM_UINT16_READ(&pFragMeta->m_Offset));

        fragBufSize = IX_NTOH16(
            IX_RM_MEM_UINT16_READ(&pFragMeta->m_BufferSize));

        if ((fragOffset + fragBufSize) > curFragOffset)
        {
            /* Copy as much as possible from this buffer */

            fragStart = curFragOffset - fragOffset;
            if ((fragBufSize - fragStart) < bytesRemaining)
            {
                bytesToCopy = fragBufSize - fragStart;
            }
            else
            {
                bytesToCopy = bytesRemaining;
            }

            ix_ossl_memcpy(
                pPktData + pktOffset, pFragData + fragStart, bytesToCopy);

            bytesRemaining -= bytesToCopy;
            bytesCopied += bytesToCopy;
            pktOffset += bytesToCopy;
            curFragOffset += bytesToCopy;
        }

        fragOffset += fragBufSize;
        IX_ERROR_CR(ix_rm_buffer_get_next(hFragBuffer, &hFragBuffer));

    } /* end - while (bytesCopied < dataBytes) */

    *arg_ByteCount = bytesCopied;
    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_reasm_copy_frag_data */


/**
 * NAME: _ix_cc_v6v4_reasm_update_frag_map
 *
 * DESCRIPTION: This primitive is called to update the map of which
 * fragments have arrived.
 *           
 * @Param:  - IN arg_hReasm - reassembly context handle returned from
 * _ix_cc_v6v4_reasm_init.
 *
 * @Param:  - IN arg_pData - pointer to reassembly data.
 *
 * @Param:  - IN arg_FragOffset - fragment offset received (value from
 * fragment header * 8).
 *
 * @Param:  - IN arg_DataBytes - number of data bytes in fragment.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_reasm_update_frag_map(
    ix_cc_v6v4_reassembly arg_hReasm,
    ix_cc_v6v4_reasm_data* arg_pData,
    ix_uint16 arg_FragOffset,
    ix_uint16 arg_DataBytes
    )
{
    ix_cc_v6v4_reasm_frag_map* pFragMap = &arg_pData->fragMap;
    ix_cc_v6v4_reasm_frag_map* pFrag;
    ix_uint16 endOffset = arg_FragOffset + arg_DataBytes;
    ix_uint32 fragIndex;
    ix_uint8 done = 0;

    while (!done)
    {
        done = 1;

        if (endOffset < pFragMap->start)
        {
            /**
             * Fragment is before this one, and not contiguous.
             * Allocate a new element for the map.
             */
            IX_ERROR_CR(ix_cc_index_pool_alloc(
                            arg_hReasm->fragMapIndexPool, &fragIndex));

            pFrag = &arg_hReasm->aFragMap[fragIndex];

            *pFrag = *pFragMap;
            pFragMap->pNext = pFrag;
            pFragMap->start = arg_FragOffset;
            pFragMap->end = endOffset;
        }
        else if (arg_FragOffset <= pFragMap->end)
        {
            /**
             * Fragment is is contiguous, so merge them.
             */
            if (arg_FragOffset < pFragMap->start)
            {
                pFragMap->start = arg_FragOffset;
            }
            if (endOffset > pFragMap->end)
            {
                pFragMap->end = endOffset;
            }

            pFrag = pFragMap->pNext;
            while ((pFrag != NULL) && (pFrag->start <= pFragMap->end))
            {
                if (pFrag->start <= pFragMap->end)
                {
                    if (pFrag->end > pFragMap->end)
                    {
                        pFragMap->end = pFrag->end;
                    }

                }

                pFragMap->pNext = pFrag->pNext;

                IX_ERROR_CR(ix_cc_index_pool_free(
                        arg_hReasm->fragMapIndexPool,
                            (ix_uint32)(pFrag - &arg_hReasm->aFragMap[0])));

                pFrag = pFragMap->pNext;
            }
        }
        else if (pFragMap->pNext == NULL)
        {
            /**
             * Fragment is at the end, and not contiguous.
             * Allocate a new element for the map.
             */
            IX_ERROR_CR(ix_cc_index_pool_alloc(
                            arg_hReasm->fragMapIndexPool, &fragIndex));

            pFrag = &arg_hReasm->aFragMap[fragIndex];

            pFragMap->pNext = pFrag;
            pFrag->pNext = NULL;
            pFrag->start = arg_FragOffset;
            pFrag->end = endOffset;
        }
        else
        {
            done = 0;
            pFragMap = pFragMap->pNext;
        }
    }

    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_reasm_update_frag_map */


/**
 * NAME: _ix_cc_v6v4_reasm_packet_complete
 *
 * DESCRIPTION: This primitive is called to update the map of which
 * fragments have arrived.
 *           
 * @Param:  - IN arg_hReasm - reassembly context handle returned from
 * _ix_cc_v6v4_reasm_init.
 *
 * @Param:  - IN arg_pData - pointer to reassembly data.
 *
 * @Param:  - IN arg_pFragMeta - pointer to fragment meta data.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_reasm_packet_complete(
    ix_cc_v6v4_reassembly arg_hReasm,
    ix_cc_v6v4_reasm_data* arg_pData,
    ix_hw_buffer_meta* arg_pFragMeta
    )
{
    ix_buffer_handle hBuffer;
    ix_hw_buffer_meta *pMeta;
	ix_uint8* pData;
    ix_cc_v6v4_ipv4_header* pIpv4Hdr;
    ix_uint16 offset;
    ix_uint16 packetSize;
    ix_uint16 bufferSize;
    ix_uint16 bufferInfo;
    int firstBuffer = 1;
    ix_uint32 csum = 0;
    ix_uint16* pHdr;
    ix_uint8 i;
#if defined(IX_DEBUG)
    ix_uint16 remainingBytes;
#endif


    hBuffer = arg_pData->hStartOfChain;

#if defined(IX_DEBUG)
    if (hBuffer == IX_NULL_BUFFER_HANDLE)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
           ("Start of buffer chain found to be NULL when reassembly complete"));
    }
#endif

    packetSize = arg_pData->hdrSize + arg_pData->dataSize;

    /**
     * Update the appropriate header fields for the reassembled
     * packet.
     */
    
    IX_ERROR_CRT(
        ix_rm_buffer_get_meta(hBuffer, (void **)&pMeta),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    IX_ERROR_CRT(
        ix_rm_buffer_get_data(hBuffer, (void **)&pData),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    /* Offset still points to payload, so adjust the pointer backwards
     * to point to the header
     */
    pIpv4Hdr = (ix_cc_v6v4_ipv4_header*)(pData +
		IX_NTOH16(IX_RM_MEM_UINT16_READ(&pMeta->m_Offset)) - arg_pData->hdrSize);

    /* Set the fragmentation info and total length */
    IX_RM_MEM_UINT16_WRITE(&pIpv4Hdr->flags_fragOffset, 0);
    IX_RM_MEM_UINT16_WRITE(&pIpv4Hdr->totalLength, IX_HTON16(packetSize));

    /* Calculate the checksum */
    IX_RM_MEM_UINT16_WRITE(&pIpv4Hdr->checksum, 0);
    pHdr = (ix_uint16*)pIpv4Hdr;

    for (i = 0; i < (arg_pData->hdrSize >> 1); i++)
    {
        csum += IX_NTOH16(IX_RM_MEM_UINT16_READ(pHdr));
        pHdr++;
    }

    while ((csum >> 16) != 0)
    {
        csum = (csum & 0xffff) + (csum >> 16);
    }
    csum = ~csum;
    IX_RM_MEM_UINT16_WRITE(&pIpv4Hdr->checksum, IX_HTON16((ix_uint16)csum));

    /**
     * Update the meta data for each buffer
     */

#if defined(IX_DEBUG)
    remainingBytes = packetSize;
#endif

    bufferInfo = IX_NTOH16(IX_RM_MEM_UINT16_READ(&arg_pFragMeta->m_BufferInfo));
    IX_RM_HW_META_SET_FREE_LIST_ID(bufferInfo, arg_hReasm->freeListId);
    bufferInfo = IX_HTON16(bufferInfo);

    while (hBuffer != IX_NULL_BUFFER_HANDLE)
    {
		if (!firstBuffer)
		{
        	IX_ERROR_CRT(
           	    ix_rm_buffer_get_meta(hBuffer, (void **)&pMeta),
                IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
        }

        /**
         * Copy the required fields from the fragment meta data.
         * Note this must change if the fields in the meta data get
         * moved around.
         */
        IX_RM_MEM_UINT16_WRITE(&pMeta->m_BufferInfo, bufferInfo);

        IX_RM_MEM_UINT32_WRITE(&pMeta->m_InputPort,
             IX_RM_MEM_UINT32_READ(&arg_pFragMeta->m_InputPort));

        IX_RM_MEM_UINT32_WRITE(&pMeta->m_NextHopID,
             IX_RM_MEM_UINT32_READ(&arg_pFragMeta->m_NextHopID));

        /*
         * Cannot use shared memory access macros for these
         * since they contain bit fields.
         */
        pMeta->m_NextHopIDType = arg_pFragMeta->m_NextHopIDType;
        pMeta->m_ColorID = arg_pFragMeta->m_ColorID;
        pMeta->m_FlowID = arg_pFragMeta->m_FlowID;

        IX_RM_MEM_UINT32_WRITE(&pMeta->m_ClassID,
             IX_RM_MEM_UINT32_READ(&arg_pFragMeta->m_ClassID));

        /* Update the necessary data */

        if (firstBuffer)
        {
            offset = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pMeta->m_Offset));
            offset -= arg_pData->hdrSize;
            IX_RM_MEM_UINT16_WRITE(&pMeta->m_Offset, IX_HTON16(offset));

            bufferSize = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pMeta->m_BufferSize));
            bufferSize += arg_pData->hdrSize;
            IX_RM_MEM_UINT16_WRITE(&pMeta->m_BufferSize, IX_HTON16(bufferSize));
        }
#if defined(IX_DEBUG)
        else
        {
            bufferSize = IX_NTOH16(IX_RM_MEM_UINT16_READ(&pMeta->m_BufferSize));
        }
        remainingBytes -= bufferSize;
#endif

        IX_RM_MEM_UINT16_WRITE(&pMeta->m_PacketSize, IX_HTON16(packetSize));

        IX_ERROR_CR(ix_rm_buffer_get_next(hBuffer, &hBuffer));
        firstBuffer = 0;
    }

#if defined(IX_DEBUG)
    if (remainingBytes != 0)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
               ("Unexpected error while reassembling packet"));
    }
#endif

    return IX_SUCCESS;
} /* end - _ix_cc_v6v4_reasm_packet_complete */


/**
 * NAME: _ix_cc_v6v4_reassembly_tick
 *
 * DESCRIPTION: This primitive is called periodically  to check for packets
 * that never received all fragments in the designated period of time.
 *           
 * @Param:  - In arg_hReasm - reassembly context handle returned from
 * _ix_cc_v6v4_reasm_init.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */

ix_error _ix_cc_v6v4_reassembly_tick(
    ix_cc_v6v4_reassembly arg_hReasm
    )
{
    ix_cc_v6v4_reasm_data* pData;
    ix_cc_v6v4_reasm_data* pNext;

    pData = (ix_cc_v6v4_reasm_data*)arg_hReasm->reassemblyDataList.pFirst;

    while (pData->pNext != NULL)
    {
        pNext = pData->pNext;

        if (pData->ticks == 0)
        {
            _ix_cc_v6v4_reasm_free_data(arg_hReasm, pData);
        }
        else
        {
            pData->ticks--;
        }

        pData = pNext;
    }

    return IX_SUCCESS;
} /* end - _ix_cc_v6v4_reassembly_tick */


/**
 * NAME: _ix_cc_v6v4_reasm_free_data
 *
 * DESCRIPTION: This primitive is called to free the resources for
 * the indicated reassembly data.
 *           
 * @Param:  - IN arg_hReasm - reassembly context handle.
 *
 * @Param:  - IN arg_pData - reassembly data to be released.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error occurred 
 * 
 */

PRIVATE ix_error _ix_cc_v6v4_reasm_free_data(
    ix_cc_v6v4_reassembly arg_hReasm,
    ix_cc_v6v4_reasm_data* arg_pData)
{
    ix_cc_v6v4_reasm_frag_map* pFragMap = arg_pData->fragMap.pNext;

    while (pFragMap != NULL)
    {
        arg_pData->fragMap.pNext = pFragMap->pNext;
        
        IX_ERROR_CRT(ix_cc_index_pool_free(
                arg_hReasm->fragMapIndexPool,
                    (ix_uint32)(pFragMap - &arg_hReasm->aFragMap[0])),
                IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

        pFragMap = arg_pData->fragMap.pNext;
    }
  
    if (arg_pData->hStartOfChain != IX_NULL_BUFFER_HANDLE)
    {
        ix_rm_buffer_free_chain(arg_pData->hStartOfChain);
    }

    if (arg_pData->hLastBuffer != IX_NULL_BUFFER_HANDLE)
    {
        ix_rm_buffer_free_chain(arg_pData->hLastBuffer);
    }

    _IX_CC_V6V4_REASM_LIST_REMOVE(arg_pData);

    IX_ERROR_CR(ix_cc_hash_table128_delete_entry(
                    arg_hReasm->hReassemblyDataHash, arg_pData->key));

    ix_ossl_free(arg_pData);

    return IX_SUCCESS;
} /* end - _ix_cc_v6v4_reasm_free_data */


/**
 * NAME: _ix_cc_v6v4_reassembly_num_packets
 *
 * DESCRIPTION: This routine exists primarily for testing.  It returns
 * a the number of packets currently being reassembled.
 *           
 * @Param:  - IN arg_hReasm - reassembly context handle returned from
 * _ix_cc_v6v4_reasm_init.
 *
 * @Param:  - OUT arg_pNumPackets - pointer to variable to receive number
 * of packets being reassembled.
 *
 * @Return: IX_SUCCESS.
 *
 * 
 */

ix_error _ix_cc_v6v4_reassembly_num_packets(
    ix_cc_v6v4_reassembly arg_hReasm,
    ix_uint32* arg_pNumPackets
    )
{
    ix_cc_v6v4_reasm_data* pData;
    ix_uint32 count = 0;

    pData = (ix_cc_v6v4_reasm_data*)arg_hReasm->reassemblyDataList.pFirst;

    while (pData->pNext != NULL)
    {
		count++;
        pData = pData->pNext;
    }

    *arg_pNumPackets = count;
    return IX_SUCCESS;

} /* end _ix_cc_v6v4_reassembly_num_packets */


/**
 * NAME: _ix_cc_v6v4_reassembly_packet_info
 *
 * DESCRIPTION: This routine exists primarily for testing.  It returns
 * information about a packet currently being reassembled.
 *           
 * @Param:  - IN arg_hReasm - reassembly context handle returned from
 * _ix_cc_v6v4_reasm_init.
 *
 * @Param:  - IN arg_Index - zero-relative index number.  Note that if
 * packets are removed while iterating through the list, info about the
 * same packet might be returned twice.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - packet not found 
 * 
 */

ix_error _ix_cc_v6v4_reassembly_packet_info(
    ix_cc_v6v4_reassembly arg_hReasm,
    ix_uint32 arg_Index,
    ix_cc_v6v4_reasm_packet_info* arg_pInfo
    )
{
    ix_cc_v6v4_reasm_data* pData;
    ix_cc_v6v4_reasm_frag_map* pFrag;
    ix_uint32 count = 0;

    pData = (ix_cc_v6v4_reasm_data*)arg_hReasm->reassemblyDataList.pFirst;

    while (pData->pNext != NULL)
    {
        if (count == arg_Index)
        {
            arg_pInfo->srcAddr = IX_NTOH32(pData->key.aUint32[2]);
            arg_pInfo->dstAddr = IX_NTOH32(pData->key.aUint32[3]);
            arg_pInfo->idNum = IX_NTOH16((ix_uint16)(pData->key.aUint32[0]));
            arg_pInfo->protocol = (ix_uint8)pData->key.aUint32[1];
            arg_pInfo->bytesReceived = 0;
            pFrag = &pData->fragMap;
            while (pFrag != NULL)
            {
                arg_pInfo->bytesReceived += (pFrag->end - pFrag->start);
                pFrag = pFrag->pNext;
            }
            return IX_SUCCESS;
        }
		count++;
        pData = pData->pNext;
    }

    return IX_CC_ERROR_ENTRY_NOT_FOUND;

} /* end _ix_cc_v6v4_reassembly_packet_info */


