/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 6
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      IPv6/v4 Tunneling Core Component - tunnel property maintenance.
 *
 * = FILENAME
 *      ix_cc_v6v4_tunnel_prop.c
 *
 * = DESCRIPTION
 *      The file contains support routines for managing property information
 *      for start tunnels.
 *
 * = CREATION TIME
 *      11/22/2002 2:29 PM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */


#define IX_ERROR_FILE_IDENT "$Id: ix_cc_v6v4_tunnel_prop.c,v 1.3 2003/01/30 19:10:21 cabell Exp $" 

/**
 * System defined include files
 */

/**
 * User defined include files
 */

#include "ix_cc_error.h"

#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_netmacros.h"

#include "ix_cc.h"
#include "ix_cc_properties.h"
#include "bindings.h"
#include "cc/ix_cc_v6v4.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_init.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_tunnel_prop.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_config.h"




/**
 * Preprocessor symbols and Macros used in this file
 */

#define IX_CC_V6V4_TRACK_MTU                0x01
#define IX_CC_V6V4_TRACK_BROADCAST          0x02


/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */
extern ix_cc_v6v4_context* g_pV6v4Context;
extern ix_cc_v6v4_registry_data g_V6v4RegistryData;
/** 
 * Extern function prototypes.
 */


/** 
 * Static function prototypes.
 */


/** 
 * Function definitions.
 */


/**
 * NAME: _ix_cc_v6v4_add_tunnel_to_interface
 *
 * DESCRIPTION: This primitive is used to associate a tunnel with an
 * interface.
 *           
 * @Param:  - IN arg_BladeId - blade ID for the interface.
 *
 * @Param:  - IN arg_PortId - port ID for the interface.
 *
 * @Param:  - IN arg_TunnelIndex - index of the start tunnel.
 *
 * @Param:  - IN arg_TrackMtu - indicates whether to maintain the tunnel
 * MTU as the interface MTU.
 *
 * @Param:  - IN arg_TrackBroadcast - indicates whether to maintain the tunnel
 * subnet broadcast address as the interface broadcast address.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - no entry in table 
 * 
 */

ix_error _ix_cc_v6v4_add_tunnel_to_interface(
    ix_uint16 arg_BladeId,
    ix_uint16 arg_PortId,
    ix_uint16 arg_TunnelIndex,
    ix_uint8 arg_TrackMtu,
    ix_uint8 arg_TrackBroadcast)
{
    ix_cc_v6v4_tunnel_group* pGroup;
    ix_cc_v6v4_tunnel_group** ppGroup;
    ix_uint32 keyIndex;

    /* Validate Port and blade Ids */
    if((arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
      (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
    {
        return IX_ERROR_WARNING(
            IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
            ("Invalid Blade Id % or Port Id %d",arg_BladeId,arg_PortId));
    }/* endif ((arg_BladeId > MAX_NUMBER_OF_BLADES - 1)... */

    /*
     * Calculate keyIndex from Blade and Port Id
     * 6 bits for Blade Id(totally 64 blades) and each blade can have up to
     * 16 ports
     * 4 bits to store portIds
     */

    keyIndex = (arg_BladeId << 4) | arg_PortId;

    ppGroup = &g_pV6v4Context->aIfTunnelTable[keyIndex];
    pGroup = *ppGroup;

    while ((pGroup != NULL) && (pGroup->numEntries == IX_CC_V6V4_TUNNEL_LIST_ENTRIES))
    {
        ppGroup = &pGroup->pNext;
        pGroup = pGroup->pNext;
    }

    if (pGroup == NULL)
    {
        /* Need to allocate a new block */
        pGroup = (ix_cc_v6v4_tunnel_group*)
            ix_ossl_malloc(sizeof(ix_cc_v6v4_tunnel_group));

        if (pGroup == NULL)
        {
            return IX_ERROR_WARNING(
                    IX_CC_ERROR_OOM,
                    ("Failed to allocate memory for tunneling config info"));
        }
        ix_ossl_memset(pGroup,0,sizeof(ix_cc_v6v4_tunnel_group));
        *ppGroup = pGroup;
    }

    pGroup->aTunnel[pGroup->numEntries] = arg_TunnelIndex;
    if (arg_TrackMtu)
    {
        pGroup->aFlags[pGroup->numEntries] |= IX_CC_V6V4_TRACK_MTU;
    }
    if (arg_TrackBroadcast)
    {
        pGroup->aFlags[pGroup->numEntries] |= IX_CC_V6V4_TRACK_BROADCAST;
    }
    pGroup->numEntries++;

    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_add_tunnel_to_interface */


/**
 * NAME: _ix_cc_v6v4_remove_tunnel_from_interface
 *
 * DESCRIPTION: This primitive is used to remove a tunnel from an interface.
 *           
 * @Param:  - IN arg_BladeId - blade ID for the interface.
 *
 * @Param:  - IN arg_PortId - port ID for the interface.
 *
 * @Param:  - IN arg_TunnelIndex - index of the start tunnel.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - no entry in table 
 * 
 */

ix_error _ix_cc_v6v4_remove_tunnel_from_interface(
    ix_uint16 arg_BladeId,
    ix_uint16 arg_PortId,
    ix_uint16 arg_TunnelIndex)
{
    ix_cc_v6v4_tunnel_group* pGroup;
    ix_cc_v6v4_tunnel_group** ppGroup;
    ix_uint32 keyIndex;
    ix_uint16 index;
    int found;

    /* Validate Port and blade Ids */
    if((arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
      (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
    {
        return IX_ERROR_WARNING(
            IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
            ("Invalid Blade Id % or Port Id %d",arg_BladeId,arg_PortId));
    }/* endif ((arg_BladeId > MAX_NUMBER_OF_BLADES - 1)... */

    /*
     * Calculate keyIndex from Blade and Port Id
     * 6 bits for Blade Id(totally 64 blades) and each blade can have up to
     * 16 ports
     * 4 bits to store portIds
     */

    keyIndex = (arg_BladeId << 4) | arg_PortId;

    ppGroup = &g_pV6v4Context->aIfTunnelTable[keyIndex];
    pGroup = *ppGroup;
    found = 0;

    while (!found && (pGroup != NULL))
    {
        index = 0;
        while (!found && (index < pGroup->numEntries))
        {
            if (pGroup->aTunnel[index] == arg_TunnelIndex)
            {
                found = 1;
            }
            else
            {
                index++;
            }
        }

        if (!found)
        {
            ppGroup = &pGroup->pNext;
            pGroup = pGroup->pNext;
        }
    }

    if (!found)
    {
        return IX_ERROR_WARNING(
                IX_CC_ERROR_ENTRY_NOT_FOUND,
                ("Specified tunnel not configured on indicated interface"));
    }

    if ((pGroup->numEntries - 1) == index)
    {
        pGroup->aTunnel[index] = 0;
        pGroup->aFlags[index] = 0;
    }
    else
    {
        pGroup->aTunnel[index] = pGroup->aTunnel[pGroup->numEntries - 1];
        pGroup->aFlags[index] = pGroup->aFlags[pGroup->numEntries - 1];
        pGroup->aTunnel[pGroup->numEntries - 1] = 0;
        pGroup->aFlags[pGroup->numEntries - 1] = 0;
    }

    if (--pGroup->numEntries == 0)
    {
        *ppGroup = pGroup->pNext;
        ix_ossl_free(pGroup);
    }

    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_remove_tunnel_from_interface */


/**
 * NAME: _ix_cc_v6v4_set_track_mtu
 *
 * DESCRIPTION: This primitive is used to enable or disable MTU tracking
 * for a tunnel.
 *           
 * @Param:  - IN arg_BladeId - blade ID for the interface.
 *
 * @Param:  - IN arg_PortId - port ID for the interface.
 *
 * @Param:  - IN arg_TunnelIndex - index of the start tunnel.
 *
 * @Param:  - IN arg_Selection - index of the start tunnel.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - no entry in table 
 * 
 */

ix_error _ix_cc_v6v4_set_track_mtu(
    ix_uint16 arg_BladeId,
    ix_uint16 arg_PortId,
    ix_uint16 arg_TunnelIndex,
    ix_uint8 arg_Selection)
{
    ix_cc_v6v4_tunnel_group* pGroup;
    ix_uint32 keyIndex;
    ix_uint16 index;
    int found;

    /* Validate Port and blade Ids */
    if((arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
      (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
    {
        return IX_ERROR_WARNING(
            IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
            ("Invalid Blade Id % or Port Id %d",arg_BladeId,arg_PortId));
    }/* endif ((arg_BladeId > MAX_NUMBER_OF_BLADES - 1)... */

    /*
     * Calculate keyIndex from Blade and Port Id
     * 6 bits for Blade Id(totally 64 blades) and each blade can have up to
     * 16 ports
     * 4 bits to store portIds
     */

    keyIndex = (arg_BladeId << 4) | arg_PortId;

    pGroup = g_pV6v4Context->aIfTunnelTable[keyIndex];
    found = 0;

    while (!found && (pGroup != NULL))
    {
        index = 0;
        while (!found && (index < pGroup->numEntries))
        {
            if (pGroup->aTunnel[index] == arg_TunnelIndex)
            {
                found = 1;
            }
            else
            {
                index++;
            }
        }

        if (!found)
        {
            pGroup = pGroup->pNext;
        }
    }

    if (!found)
    {
        return IX_ERROR_WARNING(
                IX_CC_ERROR_ENTRY_NOT_FOUND,
                ("Specified tunnel not configured on indicated interface"));
    }

    if (arg_Selection == IX_CC_V6V4_SELECT)
    {
        pGroup->aFlags[index] |= IX_CC_V6V4_TRACK_MTU;
    }
    else if (arg_Selection == IX_CC_V6V4_CLEAR)
    {
        pGroup->aFlags[index] &= ~IX_CC_V6V4_TRACK_MTU;
    }
    else
    {
        return IX_ERROR_WARNING(
                IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                ("Selection parameter is invalid"));
    }

    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_set_track_mtu */


/**
 * NAME: _ix_cc_v6v4_set_track_broadcast
 *
 * DESCRIPTION: This primitive is used to enable or disable broadcast tracking
 * for a tunnel.
 *           
 * @Param:  - IN arg_BladeId - blade ID for the interface.
 *
 * @Param:  - IN arg_PortId - port ID for the interface.
 *
 * @Param:  - IN arg_TunnelIndex - index of the start tunnel.
 *
 * @Param:  - IN arg_Selection - index of the start tunnel.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - no entry in table 
 * 
 */

ix_error _ix_cc_v6v4_set_track_broadcast(
    ix_uint16 arg_BladeId,
    ix_uint16 arg_PortId,
    ix_uint16 arg_TunnelIndex,
    ix_uint8 arg_Selection)
{
    ix_cc_v6v4_tunnel_group* pGroup;
    ix_uint32 keyIndex;
    ix_uint16 index;
    int found;

    /* Validate Port and blade Ids */
    if((arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
      (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
    {
        return IX_ERROR_WARNING(
            IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
            ("Invalid Blade Id % or Port Id %d",arg_BladeId,arg_PortId));
    }/* endif ((arg_BladeId > MAX_NUMBER_OF_BLADES - 1)... */

    /*
     * Calculate keyIndex from Blade and Port Id
     * 6 bits for Blade Id(totally 64 blades) and each blade can have up to
     * 16 ports
     * 4 bits to store portIds
     */

    keyIndex = (arg_BladeId << 4) | arg_PortId;

    pGroup = g_pV6v4Context->aIfTunnelTable[keyIndex];
    found = 0;

    while (!found && (pGroup != NULL))
    {
        index = 0;
        while (!found && (index < pGroup->numEntries))
        {
            if (pGroup->aTunnel[index] == arg_TunnelIndex)
            {
                found = 1;
            }
            else
            {
                index++;
            }
        }

        if (!found)
        {
            pGroup = pGroup->pNext;
        }
    }

    if (!found)
    {
        return IX_ERROR_WARNING(
                IX_CC_ERROR_ENTRY_NOT_FOUND,
                ("Specified tunnel not configured on indicated interface"));
    }

    if (arg_Selection == IX_CC_V6V4_SELECT)
    {
        pGroup->aFlags[index] |= IX_CC_V6V4_TRACK_BROADCAST;
    }
    else if (arg_Selection == IX_CC_V6V4_CLEAR)
    {
        pGroup->aFlags[index] &= ~IX_CC_V6V4_TRACK_BROADCAST;
    }
    else
    {
        return IX_ERROR_WARNING(
                IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                ("Selection parameter is invalid"));
    }

    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_set_track_broadcast */


/**
 * NAME: _ix_cc_v6v4_get_tunnel_track_info
 *
 * DESCRIPTION: This primitive is used to retrieve the current MTU and
 * broadcast tracking settings for a tunnel.
 *           
 * @Param:  - IN arg_BladeId - blade ID for the interface.
 *
 * @Param:  - IN arg_PortId - port ID for the interface.
 *
 * @Param:  - IN arg_TunnelIndex - index of the start tunnel.
 *
 * @Param:  - OUT arg_pTrackMtu - pointer to location in which to return
 * track MTU setting for the tunnel.
 *
 * @Param:  - OUT arg_pTrackMtu - pointer to location in which to return
 * track broadcast setting for the tunnel.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - no entry in table 
 * 
 */

ix_error _ix_cc_v6v4_get_tunnel_track_info(
    ix_uint16 arg_BladeId,
    ix_uint16 arg_PortId,
    ix_uint16 arg_TunnelIndex,
    ix_uint8* arg_pTrackMtu,
    ix_uint8* arg_pTrackBroadcast)
{
    ix_cc_v6v4_tunnel_group* pGroup;
    ix_uint32 keyIndex;
    ix_uint16 index;
    int found;

    /* Validate Port and blade Ids */
    if((arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
      (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
    {
        return IX_ERROR_WARNING(
            IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
            ("Invalid Blade Id % or Port Id %d",arg_BladeId,arg_PortId));
    }/* endif ((arg_BladeId > MAX_NUMBER_OF_BLADES - 1)... */

    /*
     * Calculate keyIndex from Blade and Port Id
     * 6 bits for Blade Id(totally 64 blades) and each blade can have up to
     * 16 ports
     * 4 bits to store portIds
     */

    keyIndex = (arg_BladeId << 4) | arg_PortId;

    pGroup = g_pV6v4Context->aIfTunnelTable[keyIndex];
    found = 0;

    while (!found && (pGroup != NULL))
    {
        index = 0;
        while (!found && (index < pGroup->numEntries))
        {
            if (pGroup->aTunnel[index] == arg_TunnelIndex)
            {
                found = 1;
            }
            else
            {
                index++;
            }
        }

        if (!found)
        {
            pGroup = pGroup->pNext;
        }
    }

    if (!found)
    {
        return IX_ERROR_WARNING(
                IX_CC_ERROR_ENTRY_NOT_FOUND,
                ("Specified tunnel not configured on indicated interface"));
    }

    *arg_pTrackMtu =
        ((pGroup->aFlags[index] & IX_CC_V6V4_TRACK_MTU) != 0);

    *arg_pTrackBroadcast =
        ((pGroup->aFlags[index] & IX_CC_V6V4_TRACK_BROADCAST) != 0);

    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_get_tunnel_track_info */



/**
 * NAME: _ix_cc_v6v4_update_tunnels_ip_addr
 *
 * DESCRIPTION: This primitive is used to update the local IP address
 * for all tunnels associated with an interface.
 *           
 * @Param:  - IN arg_BladeId - blade ID for the interface.
 *
 * @Param:  - IN arg_PortId - port ID for the interface.
 *
 * @Param:  - IN arg_IpAddr - IP address to be set.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - no entry in table 
 * 
 */

ix_error _ix_cc_v6v4_update_tunnels_ip_addr(
    ix_uint16 arg_BladeId,
    ix_uint16 arg_PortId,
    ix_uint32 arg_IpAddr)
{
    ix_cc_v6v4_tunnel_group* pGroup;
    ix_uint32 keyIndex;
    ix_uint16 index;

    /* Validate Port and blade Ids */
    if((arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
      (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
    {
        return IX_ERROR_WARNING(
            IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
            ("Invalid Blade Id % or Port Id %d",arg_BladeId,arg_PortId));
    }/* endif ((arg_BladeId > MAX_NUMBER_OF_BLADES - 1)... */

    /*
     * Calculate keyIndex from Blade and Port Id
     * 6 bits for Blade Id(totally 64 blades) and each blade can have up to
     * 16 ports
     * 4 bits to store portIds
     */

    keyIndex = (arg_BladeId << 4) | arg_PortId;

    pGroup = g_pV6v4Context->aIfTunnelTable[keyIndex];

    while (pGroup != NULL)
    {
        for (index = 0; index < pGroup->numEntries; index++)
        {
            _WRITE_ENCAP_LOCAL_ADDR(g_pV6v4Context->pMbStartTunnelBase,
                pGroup->aTunnel[index], arg_IpAddr);
        }

        pGroup = pGroup->pNext;
    }

    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_update_tunnels_ip_addr */


/**
 * NAME: _ix_cc_v6v4_update_tunnels_mtu
 *
 * DESCRIPTION: This primitive is used to update the path MTU for all
 * tunnels associated with an interface that are configured for MTU tracking.
 *           
 * @Param:  - IN arg_BladeId - blade ID for the interface.
 *
 * @Param:  - IN arg_PortId - port ID for the interface.
 *
 * @Param:  - IN arg_Mtu - MTU to be set.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - no entry in table 
 * 
 */

ix_error _ix_cc_v6v4_update_tunnels_mtu(
    ix_uint16 arg_BladeId,
    ix_uint16 arg_PortId,
    ix_uint16 arg_Mtu)
{
    ix_cc_v6v4_tunnel_group* pGroup;
    ix_uint32 keyIndex;
    ix_uint16 index;
    ix_uint16 mtu = IX_HTON16(arg_Mtu);

    /* Validate Port and blade Ids */
    if((arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
      (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
    {
        return IX_ERROR_WARNING(
            IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
            ("Invalid Blade Id % or Port Id %d",arg_BladeId,arg_PortId));
    }/* endif ((arg_BladeId > MAX_NUMBER_OF_BLADES - 1)... */

    /*
     * Calculate keyIndex from Blade and Port Id
     * 6 bits for Blade Id(totally 64 blades) and each blade can have up to
     * 16 ports
     * 4 bits to store portIds
     */

    keyIndex = (arg_BladeId << 4) | arg_PortId;

    pGroup = g_pV6v4Context->aIfTunnelTable[keyIndex];

    while (pGroup != NULL)
    {
        for (index = 0; index < pGroup->numEntries; index++)
        {
            if (pGroup->aFlags[index] & IX_CC_V6V4_TRACK_MTU)
            {
                _WRITE_ENCAP_MTU(g_pV6v4Context->pMbStartTunnelBase,
                    pGroup->aTunnel[index], mtu);
            }
        }

        pGroup = pGroup->pNext;
    }

    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_update_tunnels_mtu */


/**
 * NAME: _ix_cc_v6v4_update_tunnels_broadcast
 *
 * DESCRIPTION: This primitive is used to update the subnet broadcast
 * address for all tunnels associated with an interface that are configured
 * for broadcast tracking.
 *           
 * @Param:  - IN arg_BladeId - blade ID for the interface.
 *
 * @Param:  - IN arg_PortId - port ID for the interface.
 *
 * @Param:  - IN arg_Broadcast - broadcast address to be set.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_ENTRY_NOT_FOUND - no entry in table 
 * 
 */

ix_error _ix_cc_v6v4_update_tunnels_broadcast(
    ix_uint16 arg_BladeId,
    ix_uint16 arg_PortId,
    ix_uint32 arg_Broadcast)
{
    ix_cc_v6v4_tunnel_group* pGroup;
    ix_uint32 keyIndex;
    ix_uint16 index;

    /* Validate Port and blade Ids */
    if((arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
      (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
    {
        return IX_ERROR_WARNING(
            IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
            ("Invalid Blade Id % or Port Id %d",arg_BladeId,arg_PortId));
    }/* endif ((arg_BladeId > MAX_NUMBER_OF_BLADES - 1)... */

    /*
     * Calculate keyIndex from Blade and Port Id
     * 6 bits for Blade Id(totally 64 blades) and each blade can have up to
     * 16 ports
     * 4 bits to store portIds
     */

    keyIndex = (arg_BladeId << 4) | arg_PortId;

    pGroup = g_pV6v4Context->aIfTunnelTable[keyIndex];

    while (pGroup != NULL)
    {
        for (index = 0; index < pGroup->numEntries; index++)
        {
            if (pGroup->aFlags[index] & IX_CC_V6V4_TRACK_BROADCAST)
            {
                _WRITE_ENCAP_SUBNET_BROADCAST(
                    g_pV6v4Context->pMbStartTunnelBase,
                    pGroup->aTunnel[index], arg_Broadcast);
            }
        }

        pGroup = pGroup->pNext;
    }

    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_update_tunnels_mtu */


/**
 * NAME: _ix_cc_v6v4_get_num_tunnels_for_interface
 *
 * DESCRIPTION: This primitive is used to check how many tunnels are
 * configured on an interface.
 *           
 * @Param:  - IN arg_BladeId - blade ID for the interface.
 *
 * @Param:  - IN arg_PortId - port ID for the interface.
 *
 * @Param:  - OUT arg_pNumTunnels - pointer to location in which to return
 * the number of tunnels.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM - blade or port invalid 
 * IX_CC_ERROR_NULL - arn_NumTunnels is NULL 
 * 
 */

ix_error _ix_cc_v6v4_get_num_tunnels_for_interface(
    ix_uint16 arg_BladeId,
    ix_uint16 arg_PortId,
    ix_uint16* arg_pNumTunnels)
{
    ix_uint32 keyIndex;
    ix_cc_v6v4_tunnel_group* pGroup;
    ix_uint16 count;

    /* Validate input arguments */

    if (arg_pNumTunnels == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("arg_pNumTunnels is NULL"));
    }/* endif (arg_pNumTunnels == NULL) */

    if((arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
      (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_INPUT_PARAM,
                                ("Invalid Blade Id or Port Id"));
    }/* endif (arg_BladeId > MAX_NUMBER_OF_BLADES - 1) ||
      (arg_PortId > MAX_NUMBER_OF_PORTS_PER_BLADE - 1))
      */

    /*
     * Calculate keyIndex from Blade and Port Id
     * 6 bits for Blade Id(totally 64 blades) and each blade can have up to
     * 16 ports
     * 4 bits to store portIds
     */
    keyIndex = (arg_BladeId << 4) | arg_PortId;    

    pGroup = g_pV6v4Context->aIfTunnelTable[keyIndex];
    count = 0;

    while (pGroup != NULL)
    {
        count += (ix_uint16)pGroup->numEntries;
        pGroup = pGroup->pNext;
    }

    *arg_pNumTunnels = count;

    return IX_SUCCESS;
}/* end _ix_cc_v6v4_get_num_tunnels_for_interface */


/**
 * NAME: _ix_cc_v6v4_free_tunnel_interface_info
 *
 * DESCRIPTION: This primitive is used to free all resources allocated
 * to maintain the tunnel-interface associations.
 *           
 * @Param:  - IN None.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS.
 *
 * Error Codes:
 * 
 */

ix_error _ix_cc_v6v4_free_tunnel_interface_info()
{
    ix_uint32 keyIndex;
    ix_cc_v6v4_tunnel_group* pGroup;

    for(keyIndex = 0; keyIndex < TOTAL_NUMBER_OF_PORTS; keyIndex++)
    {
        while (g_pV6v4Context->aIfTunnelTable[keyIndex] != NULL)
        {
            pGroup = g_pV6v4Context->aIfTunnelTable[keyIndex];
            g_pV6v4Context->aIfTunnelTable[keyIndex] = pGroup->pNext;
            ix_ossl_free(pGroup);
        }/* end while */
    }/* end of table search */

    return IX_SUCCESS;
}/* end _ix_cc_v6v4_free_tunnel_interface_info */
