/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 6
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      IPv6/v4 Tunneling Core Component - validation primitives
 *
 * = FILENAME
 *      ix_cc_v6v4_validate.c
 *
 * = DESCRIPTION
 *      The file IPv6v4 tunneling header validation functions.
 *
 * = CREATION TIME
 *      12/11/2002 9:14 AM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */


#define IX_ERROR_FILE_IDENT "$Id: ix_cc_v6v4_validate.c,v 1.4 2003/05/27 17:59:33 cabell Exp $";


/**
 * System defined include files
 */


/**
 * User defined include files
 */

#include "ix_cc_error.h"
#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_netmacros.h"

#include "ix_cc.h"
#include "bindings.h"
#include "cc/ix_cc_v6v4.h"
#include "cc/v6v4_tunnel/internal/ix_cc_v6v4_validate.h"


/**
 * Preprocessor symbols and Macros used in this file
 */

#define IX_CC_V6V4_ZERO_ADDR            0
#define IX_CC_V6V4_BROADCAST_ADDR       0xffffffff
#define IX_CC_V6V4_LOOPBACK_ADDR        0x7f000001
#define IX_CC_V6V4_MULTICAST_PREFIX     0xe
#define IX_CC_V6V4_MULTICAST_SHIFT      28

#define IX_CC_V6V4_IPV4_PRIVATE_24      10
#define IX_CC_V6V4_IPV4_PRIVATE_20      0xac1
#define IX_CC_V6V4_IPV4_PRIVATE_16      0xc0a8

/**
 * Type definitions whose scope is limited to this file
 */

/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */


/** 
 * Extern function prototypes.
 */


/** 
 * Static function prototypes.
 */

PRIVATE ix_error _ix_cc_v6v4_verify_6to4_embedded_addr(
    ix_uint32 arg_IpAddr,
    ix_uint32 arg_SubnetBroadcast);

/** 
 * Function definitions.
 */


/**
 * NAME: _ix_cc_v6v4_validate_ipv4_addr
 *
 * DESCRIPTION: This primitive to validate that an IPv4 address is suitable
 * for use as a source address.
 *           .
 * 
 * @Param:  - IN arg_IpAddr - address to be validated 
 *
 * @Param:  - IN arg_SubnetBroadcast - subnet broadcast address to be checked
 * against in addition to the known invalid addresses. 
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_INVALID_ADDRESS  - address is invalid
 * 
 */


ix_error _ix_cc_v6v4_validate_ipv4_addr(
    ix_uint32 arg_IpAddr,
    ix_uint32 arg_SubnetBroadcast
    )
{
    if ((arg_IpAddr == IX_CC_V6V4_ZERO_ADDR) ||
        (arg_IpAddr == IX_CC_V6V4_BROADCAST_ADDR) ||
        (arg_IpAddr == IX_CC_V6V4_LOOPBACK_ADDR))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_ADDRESS,
                            ("Invalid IPv4 address"));
    }

    if ((arg_IpAddr >> IX_CC_V6V4_MULTICAST_SHIFT) ==
         IX_CC_V6V4_MULTICAST_PREFIX)
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_ADDRESS,
                            ("Invalid IPv4 address (multicast)"));
    }

    if ((arg_SubnetBroadcast != 0) && (arg_SubnetBroadcast == arg_IpAddr))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_ADDRESS,
                            ("Invalid IPv4 address (subnet broadcast)"));
    }

    return IX_SUCCESS;
} /* end - _ix_cc_v6v4_validate_ipv4_addr */
    

/**
 * NAME: _ix_cc_v6v4_validate_ipv6_source
 *
 * DESCRIPTION: This primitive is used to validate the IPv6 source address
 * in a packet being decapsulated.
 *           
 * @Param:  - IN arg_pIpv6Hdr - pointer to IPv6 header in packet buffer.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * IX_CC_V6V4_ERROR_INVALID_ADDRESS - packet failed validation 
 * 
 */

ix_error _ix_cc_v6v4_validate_ipv6_source(
    ix_cc_v6v4_ipv6_header* arg_pIpv6Hdr
    )
{
    ix_uint32 addrVal;

    addrVal = IX_NTOH32(IX_RM_MEM_UINT32_READ(
                            &arg_pIpv6Hdr->sourceAddress.aUint32[0]));

    /* Check for multicast - first byte = 0xff */
    if ((addrVal >> 24) == 0xff)
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_ADDRESS,
                            ("Invalid IPv6 address (multicast)"));
    }

    if (addrVal != 0)
    {
        return IX_SUCCESS;
    }

    addrVal = IX_NTOH32(IX_RM_MEM_UINT32_READ(
                            &arg_pIpv6Hdr->sourceAddress.aUint32[1]));
    if (addrVal != 0)
    {
        return IX_SUCCESS;
    }

    addrVal = IX_NTOH32(IX_RM_MEM_UINT32_READ(
                            &arg_pIpv6Hdr->sourceAddress.aUint32[2]));
    if (addrVal != 0)
    {
        return IX_SUCCESS;
    }

    addrVal = IX_NTOH32(IX_RM_MEM_UINT32_READ(
                            &arg_pIpv6Hdr->sourceAddress.aUint32[3]));

    /* Check for loopback (::1) or unspecified (::0) address */
    if ((addrVal & ~0x00000001) == 0)
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_ADDRESS,
                            ("Invalid IPv6 address (loopback/unspecified)"));
    }

    /**
     * If we get to here, the address must be an IPv4-compatible address.
     * Validate the embedded address.
     */

    IX_ERROR_CR(_ix_cc_v6v4_validate_ipv4_addr(addrVal, 0));

    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_validate_ipv6_source */


/**
 * NAME: _ix_cc_v6v4_validate_auto
 *
 * DESCRIPTION: This primitive is used to validate IPv4 packets tunneled
 * to a local interface using an IPv4-compatible address.
 *           
 * @Param:  - IN arg_pIpv4Hdr - pointer to IPv4 header in packet buffer.
 *
 * @Param:  - IN arg_pIpv6Hdr - pointer to IPv6 header in packet buffer.
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * IX_CC_V6V4_ERROR_INVALLID_PACKET - packet failed validation 
 * 
 */

ix_error _ix_cc_v6v4_validate_auto(
    ix_cc_v6v4_ipv4_header* arg_pIpv4Hdr,
    ix_cc_v6v4_ipv6_header* arg_pIpv6Hdr
    )
{
    ix_uint32 ipv4Addr;
    ix_uint32 embeddedIpv4;

    /* The IPv4 address embedded in the IPv6 destination address must
     * be the same as the IPv4 destination address
     */
    ipv4Addr = IX_RM_MEM_UINT32_READ(&arg_pIpv4Hdr->destAddress);
    embeddedIpv4 = IX_RM_MEM_UINT32_READ(&arg_pIpv6Hdr->destAddress.aUint32[3]);
    if (ipv4Addr != embeddedIpv4)
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_ADDRESS,
                            ("Invalid embedded address for auto tunnel"));
    }

#if defined(IX_CC_V6V4_VERIFY_IPV6_SOURCE)
    IX_ERROR_CR(_ix_cc_v6v4_validate_ipv6_source(arg_pIpv6Hdr));
#endif

#if defined(IX_CC_V6V4_VERIFY_IPV4_SOURCE)
    ipv4Addr = IX_NTOH32(IX_RM_MEM_UINT32_READ(&arg_pIpv4Hdr->sourceAddress));
    IX_ERROR_CR(_ix_cc_v6v4_validate_ipv4_addr(ipv4Addr, 0));
#endif

    return IX_SUCCESS;
} /* end - _ix_cc_v6v4_validate_auto */


/**
 * NAME: _ix_cc_v6v4_validate_6to4
 *
 * DESCRIPTION: This primitive is used to check the IPv6 source and
 * destination addresses of a packet being encapsulated or decapsulated
 * for 6to4 addresses.  If found, the embedded address is validated
 * according to RFC 3056.
 *           
 * @Param:  - IN arg_pIpv6Hdr - pointer to IPv6 header in packet buffer.
 *
 * @Param:  - IN arg_SubnetBroadcast - subnet broadcast address to be checked
 * against in addition to the known invalid addresses. 
 *
 * @Param:  - OUT None.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_INTERNAL - internal error 
 * IX_CC_V6V4_ERROR_INVALID_ADDRESS  - source or destination 6to4 address
 * is invalid
 * 
 */

ix_error _ix_cc_v6v4_validate_6to4(
    ix_cc_v6v4_ipv6_header* arg_pIpv6Hdr,
    ix_uint32 arg_SubnetBroadcast
    )
{
    ix_uint16 prefix;
    ix_uint32 ipAddr;
    ix_uint16 ipAddr_lo;

    prefix = IX_NTOH16(IX_RM_MEM_UINT16_READ(&arg_pIpv6Hdr->sourceAddress));
    if (prefix == IX_CC_V6V4_6TO4_PREFIX)
    {
        ipAddr = IX_NTOH16(IX_RM_MEM_UINT16_READ(
                        &arg_pIpv6Hdr->sourceAddress.aUint16[1]));
        ipAddr_lo = IX_NTOH16(IX_RM_MEM_UINT16_READ(
                        &arg_pIpv6Hdr->sourceAddress.aUint16[2]));
        ipAddr = (ipAddr <<16) | ipAddr_lo;
        
        IX_ERROR_CR(_ix_cc_v6v4_verify_6to4_embedded_addr(
                        ipAddr, arg_SubnetBroadcast));
    }

    prefix = IX_NTOH16(IX_RM_MEM_UINT16_READ(&arg_pIpv6Hdr->destAddress));
    if (prefix == IX_CC_V6V4_6TO4_PREFIX)
    {
        ipAddr = IX_NTOH16(IX_RM_MEM_UINT16_READ(
                        &arg_pIpv6Hdr->destAddress.aUint16[1]));
        ipAddr_lo = IX_NTOH16(IX_RM_MEM_UINT16_READ(
                        &arg_pIpv6Hdr->destAddress.aUint16[2]));
        ipAddr = (ipAddr <<16) | ipAddr_lo;
        
        IX_ERROR_CR(_ix_cc_v6v4_verify_6to4_embedded_addr(
                        ipAddr, arg_SubnetBroadcast));
    }

    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_validate_6to4 */


/**
 * NAME: _ix_cc_v6v4_verify_6to4_embedded_addr
 *
 * DESCRIPTION: This primitive to validate that the IPv4 address embedded in
 * a 6to4 address is acceptable according to RFC 3056.
 *           .
 * 
 * @Param:  - IN arg_IpAddr - embedded IPv4 address to be validated 
 *
 * @Param:  - IN arg_SubnetBroadcast - subnet broadcast address to be checked
 * against in addition to the known invalid addresses. 
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_V6V4_ERROR_INVALID_ADDRESS  - address is invalid
 * 
 */


ix_error _ix_cc_v6v4_verify_6to4_embedded_addr(
    ix_uint32 arg_IpAddr,
    ix_uint32 arg_SubnetBroadcast
    )
{
    IX_ERROR_CR(
        _ix_cc_v6v4_validate_ipv4_addr(arg_IpAddr, arg_SubnetBroadcast));

    /**
     *  Check for private addresses:
     *      10/8
     *      172.16/12
     *      192.168/16
     */

     if (((arg_IpAddr >> 24) == IX_CC_V6V4_IPV4_PRIVATE_24) ||
         ((arg_IpAddr >> 20) == IX_CC_V6V4_IPV4_PRIVATE_20) ||
         ((arg_IpAddr >> 16) == IX_CC_V6V4_IPV4_PRIVATE_16))
    {
        return IX_ERROR_WARNING(IX_CC_V6V4_ERROR_INVALID_ADDRESS,
                            ("Invalid embedded 6to4 address (private)"));
    }

    return IX_SUCCESS;

} /* end - _ix_cc_v6v4_verify_6to4_embedded_addr */
