//----------------------------------------------------------------------------
//
//                   I N T E L   P R O P R I E T A R Y
//
//      COPYRIGHT (c)  2001-2003 BY  INTEL  CORPORATION.  ALL RIGHTS
//      RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY
//      BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A
//      RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER
//      LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,
//      MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT
//      THE PRIOR WRITTEN PERMISSION OF :
//
//                         INTEL  CORPORATION
//
//                      2200 MISSION COLLEGE BLVD
//
//                SANTA  CLARA,  CALIFORNIA  95052-8119
//
//----------------------------------------------------------------------------
/////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: 	mpls_ftn.uc
//
//      Purpose: 	MPLS FTN code.
//
// 		Assumptions:
// 				This file may be included by others
//				(and thus uses header guards)
//
//
//
/////////////////////////////////////////////////////////////////////////////


#ifndef MPLS_FTN_UC
#define MPLS_FTN_UC 1


//--------------------------------------------------------------------------
// Included files:

#include "dl_system.h"			// System-wide definitions.
#include "mpls_defs.h"          // definitions common to the MPLS microblock
								// and core component
#include "mpls.h"				// MPLS-specific definitions.
#include <stdmac.uc>
#include <dl_meta.uc>
#include <dl_misc.uc>
#include <dram.uc>
#include <sram.uc>
#include <xbuf.uc>
#include "mpls_util.uc"			// MPLS-specific macros.





//--------------------------------------------------------------------------
// mpls_ftn
//
//    Description:
//		The MPLS FTN microblock acts (in co-operation with the IPv4 forwarder)
//				as the ingress LER MPLS forwarder
//
//    Parameters:
//      Outputs: none
//
//      In/Outs: none
//
//      Inputs:  HDR_CACHE - pointer to the packet header cache buffer
//					in local memory.
//
//      Constants:
//
//				 HDR_OFFSET - offset within header cache buffer
//					where the packet header data begins.
//
//		Assumptions:
//
//		1. It is assumed that the packet header (or only a portion of the
//			packet header) is cached in local memory.
//		2. The MPLS FTN as an ingress LER depends on IPv4/IPv6 microblock
//			which performs the longest prefix match (LPM) lookup on
//			the packet's destination address and sets dl_nexthop_id_type
//			to NHID_MPLS and dl_nexthop_id which has a meaning of an index
//			to the NHLFE table.
//		3. The MPLS FTN collects various statistics relevant to MPLS
//			packet forwarding, however they can be compiled out of code for
//			efficiency reasons (statistics are collected if
//			the __COUNTERS__ compilation flag is defined).
//		4. There are some global registers used:
//				@ftn_nhlfe_addr - FTN NHLFE table base
//				@out_port_stat - output port statistics table base (used
//					only if the __COUNTERS__ compilation flag is defined)
//
//--------------------------------------------------------------------------
#macro mpls_ftn(HDR_CACHE, HDR_OFFSET)
.begin

/*
 * HDR_WORD_OFFSET - offset within header cache buffer
 *   where the packet header data begins (in 32-bit words).
 */
#define_eval HDR_WORD_OFFSET HDR_OFFSET/4

.reg nhlfe_index

/*
 ---------------------------------------------------------------------------
     Main entry point -- Code begins execution here.
 ---------------------------------------------------------------------------
 */

main#:

/*
 * Check for our next hop ID type.
 * If doesn't match, execute an empty bypass path.
 */

	br!=byte[dl_next_block, 0, BID_MPLSFTN, bypass_path#]

    .reg nexthop_id_type

    dl_meta_get_nexthop_id_type(nexthop_id_type)

	alu[dl_next_block, --, B , BID_MPLSFTN_NEXT1]
    br!=byte[nexthop_id_type, 0, NEXT_HOP_ID_MPLS_TYPE, bypass_path#]

; PHASE 1

/*
 * Next hop ID is an input to the MPLS ftn,
 * it comes from the LPM function.
 * We map next hop ID into FEC ID.
 */
.reg nexthop_id

	dl_meta_get_nexthop_id(nexthop_id)

/*
 * Allocate FTN_CACHE buffer for FTN NHLFE
 */
	xbuf_alloc(FTN_CACHE, 8, read)

.begin
.reg fec_id ftn_base_address

/*
 * nexthop_id is an index to the FTN NHLFE table.
 * Since FTN NHLFE entry size is 32-byte,
 * we have got to multiply nexthop_id by 32 in order
 * to get pointer to the entry - fec_id.
 */
    alu_shf[fec_id, --, B , nexthop_id, <<MPLS_NHLFE_INDEX_SHIFT]

/*
 * NHLFE processing starts here.
 * Fetch NHLFE, base address of the NHLFE table
 * is stored in global register @ftn_nhlfe_addr,
 * fec_id holds the offset.
 */
	move(ftn_base_address, @ftn_nhlfe_addr)
	_mpls_fetch_ftn_nhlfe(ftn_base_address, fec_id)

/*
 * FTN NHLFE structure:
 *
 ------------------------------------------------------------------------------
  LW  Bits    Size   Field           Description
 ------------------------------------------------------------------------------
  0   31:24   8      flags           bit 31: MPLS_FTN_NHLFE_VALID_BIT - if set, the
                                     entry is valid, otherwise the entry
									 is unallocated and bits 31:24 of the flags
									 field should be equal to 0
									 bit 30: MPLS_FTN_NHLFE_SET_BIT - entry type:
									     0 - regular NHLFE
										 1 - NHLFE set
									 bits 29:28 - MPLS_FTN_NHLFE_TUNNEL_MODEL:
									     00 - pipe model
										 01 - uniform model
										 10 - short pipe model without PHP
										 11 - short pipe model with PHP
									 bits 27:26 - MPLS_FTN_NHLFE_LSP_TYPE
									     00 - non-diffserv LSP
										 10 - E-LSP
										 11 - L-LSP
									 bit 25: MPLS_FTN_NHLFE_TC_ENABLED_BIT -
									 traffic conditioning enabled
									 bit 24: MPLS_FTN_NHLFE_PRECONFIGURED_EXP_MARKING
 ------------------------------------------------------------------------------
  0   23:22   2      reserved        0
 ------------------------------------------------------------------------------
  0   21:19   3      push_count      Number of labels to be added to the label
									 stack, it can be 1 to 4.
 ------------------------------------------------------------------------------
  0   18:16   3      l3Protocol      The L3 protocol type of the MPLS payload:
  									     0 - unknown
										 6 - IP_4
										 7 - IP_6
 ------------------------------------------------------------------------------
  0   15:0    16     nextHopID       ID used on the Egress Sausalito
									 to lookup the outgoing link layer
									 information
 ------------------------------------------------------------------------------
  1   31:16   16     maxLabPktSize   Effective maximum frame payload size
									 for labeled packets (LSP MTU),
									 for the outPortID port
 ------------------------------------------------------------------------------
  1   15:0    16     outPortID       Outgoing port ID - local port index on
  									 outgoing blade
 ------------------------------------------------------------------------------
  2   31:12   20     topLabel        Top of stack label - euther mpls shim or
  									 ATM label
 ------------------------------------------------------------------------------
  2	  11:8	  4		 reserved		 0
 ------------------------------------------------------------------------------
  2	  7:4	  4		 outPortType     Outgoing port type
 ------------------------------------------------------------------------------
  2	  3:0	  4		 TTL			 TTL field decrement value
 ------------------------------------------------------------------------------
  3   31:12   20     label1          Label to be pushed below the topLabel, if
  									 the pushCount field value is greater than 1
 ------------------------------------------------------------------------------
  3	  11:8	  4		 reserved		 0
 ------------------------------------------------------------------------------
  3   7:0     8      bladeID         Outgoing blade ID w.r.t. CSIX fabric
 ------------------------------------------------------------------------------
  4   31:12   20     label2          Label to be pushed below Label1, if the
  									 pushCount field value is greater than 2
 ------------------------------------------------------------------------------
  4	  11:8	  4      reserved		 0
 ------------------------------------------------------------------------------
  4   7:0     8      flowIdHigh      MSB 8 bits of 20-bit flow id
                                     (not used in this release)
 ------------------------------------------------------------------------------
  5   31:12   20     label3          Label to be pushed below Label2, if the
  									 pushCount field value equals 4
 ------------------------------------------------------------------------------
  5   11:0    12	 flowIdLow		 LSB 12 bits of flow id
                                     (not used in this release)
 ------------------------------------------------------------------------------
  6   31:16   16     Exp2phb         Index into the EXP2PHB table
                                     (not used in this release)
 ------------------------------------------------------------------------------
  6   15:0    16     Phb2exp         Index into the PHB2EXP table
                                     (not used in this release)
 ------------------------------------------------------------------------------
  7   31:0    32     reserved        0
 ------------------------------------------------------------------------------
 *
 * In case of NHLFE set the layout of the table looks different.
 * It is presented below:
 *
 ------------------------------------------------------------------------------
  LW  Bits    Size   Field           Description
 ------------------------------------------------------------------------------
  0   31:24   8      flags           bit 31: MPLS_ILM_VALID_BIT - if set, the
                                     entry is valid, otherwise the entry
									 is unallocated and bits 31:24 of the flags
									 field should be equal to 0
									 bit 30: MPLS_ILM_NHLFE_SET_BIT - entry type:
									     0 - regular NHLFE
										 1 - NHLFE set
                                     bits 29:24 - reserved
 ------------------------------------------------------------------------------
  0   23:8    16     exp2phb         Index into the EXP2PHB table
                                     (not used in this release)
 ------------------------------------------------------------------------------
  0   7:0     8      reserved        set to 0
 ------------------------------------------------------------------------------
  1   31:0    32     phbMask1        Bit mask denoting the PHBs supported by
  									 the NHLFE pointed to by this set element.
									 Used for choosing one of the NHLFE entries
									 present in the set by the MPLS forwarder.
 ------------------------------------------------------------------------------
  2   31:0    32     phbMask2        Bit mask denoting the PHBs supported by
  									 the NHLFE pointed to by this set element.
									 Used for choosing one of the NHLFE entries
									 present in the set by the MPLS forwarder.
 ------------------------------------------------------------------------------
  3   31:0    32     phbMask2        Bit mask denoting the PHBs supported by
  									 the NHLFE pointed to by this set element.
									 Used for choosing one of the NHLFE entries
									 present in the set by the MPLS forwarder.
 ------------------------------------------------------------------------------
  4   31:0    32     phbMask2        Bit mask denoting the PHBs supported by
  									 the NHLFE pointed to by this set element.
									 Used for choosing one of the NHLFE entries
									 present in the set by the MPLS forwarder.
 ------------------------------------------------------------------------------
  5   31:12   20     outSegID1       Index to the 1st NHLFE belonging
                                     to the set.
 ------------------------------------------------------------------------------
  5   11:0    12     outSegID2       Index to the 2nd NHLFE belonging
                                     to the set (12 most significant bits).
 ------------------------------------------------------------------------------
  6   31:24   8      outSegID2       Index to the 2nd NHLFE belonging
                                     to the set (8 least significant bits).
 ------------------------------------------------------------------------------
  6   23:4   20      outSegID3       Index to the 3rd NHLFE belonging
                                     to the set.
 ------------------------------------------------------------------------------
  6   3:0    4       outSegID4       Index to the 4th NHLFE belonging
                                     to the set (4 most significant bits).
 ------------------------------------------------------------------------------
  7   31:16  16      outSegID4       Index to the 4th NHLFE belonging
                                     to the set (16 least significant bits).
 ------------------------------------------------------------------------------
  7   15:0   16      reserved        set to 0
 ------------------------------------------------------------------------------
*/


/*
 * Check if the FTN NHLFE entry is valid. If not drop the packet.
 */
	br_bclr[FTN_CACHE[0], MPLS_FTN_NHLFE_VALID_BIT, drop#]

    alu[nhlfe_index, --, b, nexthop_id]
/*
 * Check FTN NHLFE entry type. If this is NHLFE set choose one entry,
 * otherwise skip this step.
 */
	br_bclr[FTN_CACHE[0], MPLS_FTN_NHLFE_SET_BIT, regular_nhlf_entry#]

/*
 * NHLFE set is used for many purposes: diffserv, load_balancing, fault
 * tolerance, etc.. Different algorithms are used to select NHLFE entry
 * depending on application purpose.
 * For now we just select the first entry.
 */

 	_mpls_select_nhlfe_from_set(FTN_CACHE, nhlfe_index)

    alu_shf[fec_id, --, B , nhlfe_index, <<MPLS_NHLFE_INDEX_SHIFT]

	_mpls_fetch_ftn_nhlfe(ftn_base_address, fec_id)


.end 	// end of fec_id ftn_base_address

/*
 * Check if the FTN NHLFE entry is valid. If not drop the packet.
 */
	br_bclr[FTN_CACHE[0], MPLS_FTN_NHLFE_VALID_BIT, drop#]

regular_nhlf_entry#:


; PHASE 2
/*
 * From this point we assume that packet is received correctly.
 * Further packet's processing depends on operation
 * to be performed on the packet's header. In this case, PUSH is the
 * only operation.
 */
.reg push_count

/*
 * Here we handle IP headers which enter MPLS domain (MPLS labels
 * will be pushed onto the IP header)
 */

.begin

.reg ttl_value
	alu_shf_right(ttl_value, 0xFF, AND, HDR_CACHE[6], 24)

push_handling#:
.begin //PUSH handling

/*
 * Extract TTL value from IPv4 header, decrement it by the value
 * obtained from NHLF entry and check if it is greater than zero.
 * If not TTL_EXPIRED exception is invoked.
 */
.reg nhlfe_ttl

	_mpls_ftn_ttl_extract(nhlfe_ttl)
	alu_op(ttl_value, ttl_value, - , nhlfe_ttl)
	ble[ttl_expired#]

/*
 * Check tunnel model. If it is uniform we only have to set
 * decremented ttl_value in all the pushed labels,
 * otherwise we set PIPE_TTL value in all the pushed labels
 * and decremented ttl_value in IPv4 header,
 * IPv4 header checksum is also updated.
 * If the outgoing LSP is subject to DiffServ processing, the EXP bits
 * for the labels in the packet's label stack are determined.
 */
.reg tunnel_model

	_mpls_ftn_tunnel_model_extract(tunnel_model)
	br=byte[tunnel_model, 0, MPLS_UNIFORM_TUNNEL_MODEL, uniform_tunnel_model#]

	_mpls_update_ipv4_header_lm(HDR_CACHE, HDR_WORD_OFFSET, ttl_value)

	_mpls_push_labels_lm(push_count, labels_to_write, PIPE, NON_DIFFSERV,
						HDR_CACHE, HDR_WORD_OFFSET, FTN_CACHE, ttl_value)

	br[end_push#]

uniform_tunnel_model#:

	_mpls_push_labels_lm(push_count, labels_to_write, UNIFORM, NON_DIFFSERV,
						HDR_CACHE, HDR_WORD_OFFSET, FTN_CACHE, ttl_value)

end_push#:

.end	// end of tunnel_model nhlfe_ttl

.end	// end of ttl_value


/*
 * Update dl_size
 * Update meta data.
 * Set packet header type.
 * Check if packet size does not exceed Effective Maximum Frame
 * Payload Size for Labeled Packets parameter, if it does
 * FTN_LABELED_PACKET_TOO_BIG exception is invoked.
 * Optionally update mplsOutIfcPkts and mplsOutSegPkts counters.
 */

common_code#:

.begin
.reg dl_packet_offset dl_packet_size dl_buf_size

	dl_meta_get_offset(dl_packet_offset)
	alu_shf_left(dl_packet_offset, dl_packet_offset, - , push_count, 2)
	dl_meta_set_offset(dl_packet_offset)

	dl_meta_get_packet_size(dl_packet_size)
	alu_shf_left(dl_packet_size, dl_packet_size, + , push_count, 2)
	dl_meta_set_packet_size(dl_packet_size)

	dl_meta_get_buffer_size(dl_buf_size)
	alu_shf_left(dl_buf_size, dl_buf_size, + , push_count, 2)
	dl_meta_set_buffer_size(dl_buf_size)

.reg max_lab_pkt_size

	dl_meta_set_header_type(MPLS_MPLS_HEADER_TYPE)

	_mpls_ftn_max_lab_pkt_size_extract(max_lab_pkt_size)
	alu_op(--, dl_packet_size, -, max_lab_pkt_size)
	bgt[labeled_packet_too_big#]

	_mpls_update_meta_data(FTN_CACHE)


/*
 * Optionally update FTN counters.
 */
#ifdef __COUNTERS__
	_mpls_update_incoming_counters_ftn(nexthop_id, dl_packet_size)
#endif

send_to_next_block#:

	br[bypass_path#]

.end	// end of dl_packet_offset dl_packet_size dl_buf_size max_lab_pkt_size

; ----------------------------------------------------------------------------
; Exception handling area starts here.
; ----------------------------------------------------------------------------
ttl_expired#:

	dl_set_exception(BID_MPLSFTN, TTL_EXPIRED)
	immed32(dl_next_block, IX_EXCEPTION)

#ifdef __COUNTERS__
.begin
.reg out_counter_addr

    immed32(out_counter_addr, MPLS_FTN_STAT_BASE)

    add_shf_left(out_counter_addr, out_counter_addr, nexthop_id, 4)
    add_shf_left(out_counter_addr, out_counter_addr, nexthop_id, 3)

	sram_incr(out_counter_addr, MPLS_IN_SEGMENT_ERROR_PKTS)

	#ifdef MPLS_STATS_FULL

    .reg dl_packet_size
	.sig sig_stat

	dl_meta_get_packet_size(dl_packet_size)

	xbuf_alloc($packet_size, 1, write)
	move($packet_size[0], dl_packet_size)

		// Update number of octets sent on this FTN
		sram_add($packet_size[0], out_counter_addr,
				MPLS_IN_SEGMENT_ERROR_OCTETS, sig_stat, sig_stat, ___)
	xbuf_free($packet_size)

	#endif
.end
#endif

	br[bypass_path#]

labeled_packet_too_big#:
/*
 * Labeled packet is sent to the core component if it exceeds
 * Effective Maximum Frame Payload Size for Labeled Packets parameter.
 * mplsOutIfcFragments counter is optionally updated.
 */
	dl_set_exception(BID_MPLSFTN, FTN_LABELED_PACKET_TOO_BIG)

	dl_meta_set_nexthop_id(nhlfe_index);
    alu[nhlfe_index, --, B, nhlfe_index, >>16]
	dl_meta_set_output_port(nhlfe_index);

	immed32(dl_next_block, IX_EXCEPTION)

#ifdef __COUNTERS__
.begin
.reg out_counter_addr

    immed32(out_counter_addr, MPLS_FTN_STAT_BASE)

    add_shf_left(out_counter_addr, out_counter_addr, nexthop_id, 4)
    add_shf_left(out_counter_addr, out_counter_addr, nexthop_id, 3)

	sram_incr(out_counter_addr, MPLS_IN_SEGMENT_PKTS)

	#ifdef MPLS_STATS_FULL

    .reg dl_packet_size
	.sig sig_stat

	dl_meta_get_packet_size(dl_packet_size)

	xbuf_alloc($packet_size, 1, write)
	move($packet_size[0], dl_packet_size)

		// Update number of octets sent on this FTN
		sram_add($packet_size[0], out_counter_addr,
				MPLS_IN_SEGMENT_OCTETS, sig_stat, sig_stat, ___)
	xbuf_free($packet_size)

	#endif
.end
#endif

	br[bypass_path#]

; ----------------------------------------------------------------------------
; Drop handling area starts here.
; ----------------------------------------------------------------------------


drop#:

/*
 * Drop the packet and exit.
 */
	immed32(dl_next_block, IX_DROP)

bypass_path#:
xbuf_free(FTN_CACHE)

.end
#endm


#endif /* MPLS_FTN_UC */
