//----------------------------------------------------------------------------
//
//                   I N T E L   P R O P R I E T A R Y
//
//      COPYRIGHT (c)  2001-2003 BY  INTEL  CORPORATION.  ALL RIGHTS
//      RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY
//      BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A
//      RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER
//      LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,
//      MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT
//      THE PRIOR WRITTEN PERMISSION OF :
//
//                         INTEL  CORPORATION
//
//                      2200 MISSION COLLEGE BLVD
//
//                SANTA  CLARA,  CALIFORNIA  95052-8119
//
//----------------------------------------------------------------------------
/////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: 	mpls_ilm.uc
//
//      Purpose: 	MPLS ILM code.
//
// 		Assumptions:
// 				This file may be included by others
//				(and thus uses header guards)
//
//
//
/////////////////////////////////////////////////////////////////////////////



#ifndef MPLS_ILM_UC
#define MPLS_ILM_UC 1


//--------------------------------------------------------------------------
// Included files:
#include "dl_system.h"			// System-wide definitions.
#include "mpls_defs.h"          // definitions common to the MPLS microblock
								// and core component
#include "mpls.h"				// MPLS-specific definitions.
#include <stdmac.uc>
#include <dl_meta.uc>
#include <dl_misc.uc>
#include <dram.uc>
#include <sram.uc>
#include <xbuf.uc>
#include "mpls_util.uc"			// MPLS-specific macros.


//--------------------------------------------------------------------------
// mpls_ilm
//
//    Description:
//		MPLS ILM comprises MPLS forwarding on transit LSR and and egress
//		LER nodes.
//
//    Parameters:
//      Outputs: none
//
//      In/Outs: none
//
//      Inputs:  HDR_CACHE - pointer to the packet header cache buffer
//					in local memory.
//
//      Constants:
//
//				 HDR_OFFSET - offset within header cache buffer
//					where the packet header data begins.
//
//		Assumptions:
//
//		1. It is assumed that the packet header (or only a portion of the
//			packet header) is cached in local memory.
//		2. The MPLS ILM collects various statistics relevant to MPLS
//			packet forwarding, however they can be compiled out of code for
//			efficiency reasons (statistics are collected if
//			the __COUNTERS__ compilation flag is defined).
//		3. There are some global registers used:
//				@ilm_nhlfe_addr - ILM table base address
//				@in_port_stat - input port statistics table base (used only
//					if the __COUNTERS__ compilation flag is defined)
//
//--------------------------------------------------------------------------
#macro mpls_ilm(HDR_CACHE, HDR_OFFSET)
.begin

#define_eval HDR_WORD_OFFSET HDR_OFFSET/4

.reg extern  dl_next_block


/*
 ---------------------------------------------------------------------------
     Main entry point -- Code begins execution here.
 ---------------------------------------------------------------------------
*/

main#:

/*
 * Check if dl_next_block is BID_MPLSILM, otherwise execute an empty
 * bypass path.
 */

	br!=byte[dl_next_block, 0, BID_MPLSILM, bypass_path#]

; PHASE 1

.reg dl_packet_size
.reg input_port
.reg ilm_base_addr


/*
 * Read configuration variables from local memory.
 * They are:
 *   a) in case of per platform label space:
 *        - min_label - minimum label value allowed on this device;
 *                      legal values are <0x10, 0xFFFFF>
 *        - max_label - maximum label value allowed on this device;
 *                      legal values are <0x10, 0xFFFFF>
 *   b) in case of per interface label space:
 *        - min_label - minimum label value allowed on this interface;
 *                      legal values are <0x10, 0xFFFFF>
 *        - max_label - maximum label value allowed on this interface;
 *                      legal values are <0x10, 0xFFFFF>
 *        - per interface ILM Table offset
 *        - per interface ILM Counters Table offset (only if statistics
 *          are gathered)
 */
.reg min_label
.reg max_label
.reg lm_address
.reg label_space_base

	immed32(label_space_base, LM_LABEL_SPACE_ADDR)
	dl_meta_get_input_port(input_port)



#ifdef PER_INTERFACE_LABEL_SPACE

.reg ilm_per_port_offset

    add_shf_left(label_space_base, label_space_base, input_port, 4)
    local_csr_wr[ACTIVE_LM_ADDR_1, label_space_base]

    alu[ilm_base_addr, --, B, @ilm_nhlfe_addr]

    alu_shf_left(lm_address, ctx_rel_lm_addr, OR, 4, 2)

	dl_meta_get_packet_size(dl_packet_size)

/*
 * Calculate HDR_POINTER, a pointer to the beginning of the packet
 * header inside the header cache buffer in local memory
 */


#ifdef __COUNTERS__

    alu[min_label,               --, B, *l$index1++]
    local_csr_wr[ACTIVE_LM_ADDR_1, lm_address]

#else

    local_csr_wr[ACTIVE_LM_ADDR_1, lm_address]
    alu[min_label,               --, B, *l$index1++]

#endif


    ; the following instructions still use previous value of *l$index1
    alu[max_label,               --, B, *l$index1++]
    alu[ilm_per_port_offset,     --, B, *l$index1++]

#ifdef __COUNTERS__
    .reg ilm_counter_per_port_offset
    alu[ilm_counter_per_port_offset, --, B, *l$index1]
#endif


#else // PER_PLATFORM_LABEL_SPACE

	local_csr_wr[ACTIVE_LM_ADDR_1, label_space_base]
	alu[ilm_base_addr, --, B, @ilm_nhlfe_addr]

    alu_shf_left(lm_address, ctx_rel_lm_addr, OR, 4, 2)

/*
 * Calculate HDR_POINTER, a pointer to the beginning of the packet
 * header inside the header cache buffer in local memory
 */
    local_csr_wr[ACTIVE_LM_ADDR_1, lm_address]

	// the following instructions still use previous value of *l$index1
    alu[min_label, --, B, *l$index1++]
    alu[max_label, --, B, *l$index1]

	dl_meta_get_packet_size(dl_packet_size)
#endif


/*
 * Increase the input port packets counter
 * if the __COUNTERS__ compilation flag is defined.
 */
#ifdef __COUNTERS__
    .reg ilm_counter_offset 

	_mpls_update_rx_port_counters(input_port, dl_packet_size)
#endif

/*
 * recursive lookup for pop handling
 */
recursive_lookup#:

.reg mpls_label
.reg ilm_index
.reg ilm_offset


/*
 * Check whether the top label on the packet's label stack belongs
 * to the reserved label range. If so jump to the reserved label
 * processing section.
 */

	alu_shf[mpls_label, --, B, HDR_POINTER, >>12]
	alu_op(--, mpls_label, -, MPLS_RES_LABEL_MAX)
	ble[reserved_label_processing#]

	alu_op(--, mpls_label, -, max_label)
	bgt[failed_lookup#]
	alu_op(ilm_index, mpls_label, -, min_label)
	blt[failed_lookup#]

    alu_shf[ilm_offset, --, B , ilm_index, <<MPLS_NHLFE_INDEX_SHIFT]

#ifdef __COUNTERS__
    .reg @ilm_counter_addr
    immed32(@ilm_counter_addr, MPLS_ILM_STAT_BASE)
    add_shf_left(ilm_counter_offset, @ilm_counter_addr, ilm_index, 4)
    add_shf_left(ilm_counter_offset, ilm_counter_offset, ilm_index, 3)
#endif

#ifdef PER_INTERFACE_LABEL_SPACE
   /*
    * Having index into the ILM table, calculate the pointer.
    */
    alu[ilm_offset, ilm_offset, +, ilm_per_port_offset]

    #ifdef __COUNTERS__
        alu[ilm_counter_offset, ilm_counter_offset, +, ilm_counter_per_port_offset]
	#endif
#endif


 /* ILM table structure:
  ------------------------------------------------------------------------------
  LW  Bits    Size   Field           Description
  ------------------------------------------------------------------------------
  0	  31:24	  8		 flags		 	 bit 31: MPLS_ILM_VALID_BIT - if set, the
                                     entry is valid, otherwise the entry
									 is unallocated
									 bit 30: MPLS_ILM_NHLFE_SET_BIT - entry type:
									     0 - regular NHLFE
										 1 - NHLFE set
  									 bits 29:28 - MPLS_ILM_TUNNEL_MODEL
									     00 - pipe model
										 01 - uniform model
										 10 - short pipe model without PHP
										 11 - short pipe model with PHP
									 bits 27:26 - MPLS_ILM_LSP_TYPE
									     00 - non-diffserv LSP
										 10 - E-LSP
										 11 - L-LSP
									 bit 25: MPLS_ILM_TC_ENABLED_BIT - traffic
									 conditioning enabled
									 bit 24: MPLS_ILM_CNTSKIP_BIT - do not
									 update inSeg counters
  ------------------------------------------------------------------------------
  0	  23:22	  2		 operation		 The code of the operation to be performed
                                     on the label stack, one of:
									 MPLS_LABELOP_SWAP = 0
									 MPLS_LABELOP_SWAP_PUSH = 1
									 MPLS_LABELOP_POP = 2
									 MPLS_LABELOP_POP_FORWARD = 3
  ------------------------------------------------------------------------------
  0   21:19   3      push_count      Number of labels to push,
									 it can be 1 to 3. 
  ------------------------------------------------------------------------------
  0   18:16   3      l3Protocol      The L3 protocol type of the MPLS payload:
  									     0 - unknown
										 6 - IP_4
										 7 - IP_6
  ------------------------------------------------------------------------------
  0   15:0    16     nextHopID       ID used on the Egress to lookup the
 									 outgoing link layer information
  ------------------------------------------------------------------------------
  1   31:16   16     maxLabPktSize   Effective maximum frame payload size
									 for labeled packets (LSP MTU),
									 corresponding to the output
									 port parameter
  ------------------------------------------------------------------------------
  1   15:0    16     outPortID       Outgoing port ID - local port index on 
   									 outgoing blade
  ------------------------------------------------------------------------------
  2   31:12   20     topLabel        Top of stack label
  ------------------------------------------------------------------------------
  2	  11:8	  4		 reserved		 0
  ------------------------------------------------------------------------------
  2	  7:4	  4		 outPortType     Outgoing port type
  ------------------------------------------------------------------------------
  2	  3:0	  4		 TTL			 TTL field decrement value
  ------------------------------------------------------------------------------
  3   31:12   20     label1          Label to be pushed below the topLabel, if
  									 the pushCount field value is greater than 1
  ------------------------------------------------------------------------------
  3   11:8    4      reserved		 0
  ------------------------------------------------------------------------------
  3   7:0     8      bladeID         Outgoing blade ID w.r.t. CSIX fabric
  ------------------------------------------------------------------------------
  4   31:12   20     label2          Label to be pushed below Label1, if the
  									 pushCount field value is greater than 2
  ------------------------------------------------------------------------------
  4   11:0	  12	 reserved 		 0
  ------------------------------------------------------------------------------
  5   31:12   20     label3          Label to be pushed below Label2, if the
  									 pushCount field value equals 4
  ------------------------------------------------------------------------------
  5   11:0    12	 reserved		 0
  ------------------------------------------------------------------------------
  6   31:16   16     Exp2phb         Index into the EXP2PHB table
                                     (not used in this release)
 ------------------------------------------------------------------------------
  6   15:0    16     Phb2exp         Index into the PHB2EXP table
                                     (not used in this release)
 ------------------------------------------------------------------------------
  7   31:0    32     reserved        0
 ------------------------------------------------------------------------------
 *
 * In case of NHLFE set the layout of the table looks different.
 * It is presented below:
 *
 ------------------------------------------------------------------------------
  LW  Bits    Size   Field           Description
 ------------------------------------------------------------------------------
  0   31:24   8      flags           bit 31: MPLS_ILM_VALID_BIT - if set, the
                                     entry is valid, otherwise the entry
									 is unallocated and bits 31:24 of the flags
									 field should be equal to 0
									 bit 30: MPLS_ILM_NHLFE_SET_BIT - entry type:
									     0 - regular NHLFE
										 1 - NHLFE set
                                     bits 29:24 - reserved
 ------------------------------------------------------------------------------
  0   23:8    16     exp2phb         Index into the EXP2PHB table
                                     (not used in this release)
 ------------------------------------------------------------------------------
  0   7:0     8      reserved        set to 0
 ------------------------------------------------------------------------------
  1   31:0    32     phbMask1        Bit mask denoting the PHBs supported by
  									 the NHLFE pointed to by this set element.
									 Used for choosing one of the NHLFE entries
									 present in the set by the MPLS forwarder.
 ------------------------------------------------------------------------------
  2   31:0    32     phbMask2        Bit mask denoting the PHBs supported by
  									 the NHLFE pointed to by this set element.
									 Used for choosing one of the NHLFE entries
									 present in the set by the MPLS forwarder.
 ------------------------------------------------------------------------------
  3   31:0    32     phbMask2        Bit mask denoting the PHBs supported by
  									 the NHLFE pointed to by this set element.
									 Used for choosing one of the NHLFE entries
									 present in the set by the MPLS forwarder.
 ------------------------------------------------------------------------------
  4   31:0    32     phbMask2        Bit mask denoting the PHBs supported by
  									 the NHLFE pointed to by this set element.
									 Used for choosing one of the NHLFE entries
									 present in the set by the MPLS forwarder.
 ------------------------------------------------------------------------------
  5   31:12   20     outSegID1       Index to the 1st NHLFE belonging
                                     to the set.
 ------------------------------------------------------------------------------
  5   11:0    12     outSegID2       Index to the 2nd NHLFE belonging
                                     to the set (12 most significant bits).
 ------------------------------------------------------------------------------
  6   31:24   8      outSegID2       Index to the 2nd NHLFE belonging
                                     to the set (8 least significant bits).
 ------------------------------------------------------------------------------
  6   23:4   20      outSegID3       Index to the 3rd NHLFE belonging
                                     to the set.
 ------------------------------------------------------------------------------
  6   3:0    4       outSegID4       Index to the 4th NHLFE belonging
                                     to the set (4 most significant bits).
 ------------------------------------------------------------------------------
  7   31:16  16      outSegID4       Index to the 4th NHLFE belonging
                                     to the set (16 least significant bits).
 ------------------------------------------------------------------------------
  7   15:0   16      reserved        set to 0
 ------------------------------------------------------------------------------
*/

/*
 * Allocate ILM_CACHE buffer for ILM NHLFE
 * Fetch 1 LW from ILM NHLFE table
 */
	xbuf_alloc(ILM_CACHE, 8,read)

	_mpls_fetch_ilm_nhlfe(ILM_CACHE[0], ilm_base_addr, ilm_offset, 1)

/*
 * Check whether ILM table entry is valid.
 */
	br_bclr[ILM_CACHE[0], MPLS_ILM_VALID_BIT, error_packet#]

/*
 * Check ILM NHLFE entry type. If this is NHLFE set choose one entry,
 * otherwise skip this step.
 */
	br_bclr[ILM_CACHE[0], MPLS_ILM_NHLFE_SET_BIT, regular_nhlf_entry#]

/*
 * Read the rest of the ILM NHLFE SET table.
 */
	alu[ilm_offset, ilm_offset, +, 4]

#ifdef ILM_SRAM
	_mpls_fetch_ilm_nhlfe(ILM_CACHE[1], ilm_base_addr, ilm_offset, 7)
#else  // ILM_DRAM
	_mpls_fetch_ilm_nhlfe(ILM_CACHE[2], ilm_base_addr, ilm_offset, 6)
#endif

/*
 * NHLFE set is used for many purposes: diffserv, load_balancing, fault
 * tolerance, etc.. Different algorithms are used to select NHLFE entry
 * depending on application purpose.
 * For now we just select the first entry.
 */
 	_mpls_select_nhlfe_from_set(ILM_CACHE, ilm_index)

    alu_shf[ilm_offset, --, B , ilm_index, <<MPLS_NHLFE_INDEX_SHIFT]

	alu[ilm_base_addr, --, B, @ilm_nhlfe_set_addr]

/* 31th bit denotes that the ilm_index points at ilm_set nhlfe table */
    alu[ilm_index, ilm_index, OR, 1, <<31]

	_mpls_fetch_ilm_nhlfe(ILM_CACHE[0], ilm_base_addr, ilm_offset, 1)

/*
 * Check whether ILM table entry is valid.
 */
	br_bclr[ILM_CACHE[0], MPLS_ILM_VALID_BIT, error_packet#]

/*
 * Optionally update incoming_label_map counters.
 */
#ifdef __COUNTERS__
.begin
   .reg @ilm_set_counter_addr
    immed32(@ilm_set_counter_addr, MPLS_ILM_NHLFE_SET_STAT_BASE)
    add_shf_left(ilm_counter_offset, @ilm_set_counter_addr, ilm_index, 4)
    add_shf_left(ilm_counter_offset, ilm_counter_offset, ilm_index, 3)
	_mpls_update_incoming_counters_ilm(ilm_counter_offset, dl_packet_size)
	br[phase_2#]
.end
#endif

regular_nhlf_entry#:

/*
 * Optionally update incoming_label_map counters.
 */
#ifdef __COUNTERS__
    _mpls_update_incoming_counters_ilm(ilm_counter_offset, dl_packet_size)
#endif


; PHASE 2
phase_2#:
/*
 * The second phase of packet's processing depends on the label stack
 * operation specified in the ILM entry. If operation is not POP, the
 * rest of ILM NHLFE table is read.
 */
.reg operation
	_mpls_ilm_operation_extract(operation)

.reg tunnel_model
.reg nhlfe_ttl
.reg i_ttl

	alu_op(i_ttl, HDR_POINTER, AND, 0xFF)
	_mpls_ilm_tunnel_model_extract(tunnel_model)

.set ILM_CACHE[1] ILM_CACHE[2] ILM_CACHE[3] ILM_CACHE[4] ILM_CACHE[5]
.set nhlfe_ttl

	br=byte[operation, 0, MPLS_LABELOP_POP, pop_handling#]

/*
 * Read the rest of ILM NHLFE table if operation is not POP
 */
	alu[ilm_offset, ilm_offset, +, 4]
#ifdef MPLS_DIFFSERV_FEATURE
    #ifdef ILM_SRAM
        _mpls_fetch_ilm_nhlfe(ILM_CACHE[1], ilm_base_addr, ilm_offset, 6)
    #else  // ILM_DRAM
        _mpls_fetch_ilm_nhlfe(ILM_CACHE[2], ilm_base_addr, ilm_offset, 6)
    #endif
#else
    #ifdef ILM_SRAM
        _mpls_fetch_ilm_nhlfe(ILM_CACHE[1], ilm_base_addr, ilm_offset, 5)
    #else  // ILM_DRAM
        _mpls_fetch_ilm_nhlfe(ILM_CACHE[2], ilm_base_addr, ilm_offset, 4)
    #endif
#endif


	_mpls_ilm_ttl_extract(nhlfe_ttl)

/*
 * Operation can be POP_FORWARD, SWAP or SWAP_PUSH
 */
	br=byte[operation, 0, MPLS_LABELOP_POP_FORWARD, pop_forward_handling#]

	alu_op(i_ttl, i_ttl, - , nhlfe_ttl)
	ble[ilm_ttl_expired#]

	br=byte[operation, 0, MPLS_LABELOP_SWAP_PUSH, swap_push_handling#]

/*
 *------------------------------------------------------------------------------
 *	SWAP HANDLING
 *------------------------------------------------------------------------------
 */
swap_handling#:
.begin

	_mpls_swap_labels_lm(NON_DIFFSERV, HDR_POINTER, ILM_CACHE, i_ttl)

	br[ilm_common_code#]

.end

/*
 *------------------------------------------------------------------------------
 *	SWAP PUSH HANDLING
 *------------------------------------------------------------------------------
 */
swap_push_handling#:
.begin
.reg push_count

	br=byte[tunnel_model, 0, MPLS_UNIFORM_TUNNEL_MODEL, sp_ut#]

;swap_push handling, pipe/short pipe tunnel model
.begin //SWAP_PUSH handling, PIPE/SHORT PIPE tunnel model

	_mpls_swap_push_labels_lm(push_count, PIPE, NON_DIFFSERV,
						HDR_POINTER, ILM_CACHE, i_ttl)

	br[sp_update_dl_size#]

.end //SWAP_PUSH handling, PIPE/SHORT PIPE tunnel model

;swap_push handling, uniform tunnel model
sp_ut#:

.begin //SWAP_PUSH handling, UNIFORM tunnel model

	_mpls_swap_push_labels_lm(push_count, UNIFORM, NON_DIFFSERV,
						HDR_POINTER, ILM_CACHE, i_ttl)

.end //SWAP_PUSH handling, UNIFORM tunnel model

sp_update_dl_size#:

	_mpls_update_dl_size(push_count)
	
	br[ilm_common_code#]
.end	// end of SWAP_PUSH handling

;------------------------------------------------------------------------------
; 	POP_FORWARD HANDLING
;------------------------------------------------------------------------------
pop_forward_handling#:
.begin
.reg o_ttl
	alu[o_ttl, i_ttl, -, nhlfe_ttl]
	ble[ilm_ttl_expired#]

/*
 * When POP_FORWARD operation is decoded, the top label's TTL
 * is saved and then the top label is popped from the label stack.
 * The packet's meta data is updated accordingly.
 */
.set labels_in_cache
	alu[labels_in_cache, labels_in_cache, -, 1]

.begin
.reg dl_packet_offset dl_packet_size dl_buf_size

		dl_meta_get_offset(dl_packet_offset)
		alu_op(dl_packet_offset, dl_packet_offset, + , 4)
		dl_meta_set_offset(dl_packet_offset)

		dl_meta_get_packet_size(dl_packet_size)
		alu_op(dl_packet_size, dl_packet_size, - , 4)
		dl_meta_set_packet_size(dl_packet_size)

		dl_meta_get_buffer_size(dl_buf_size)
		alu_op(dl_buf_size, dl_buf_size, - , 4)
		dl_meta_set_buffer_size(dl_buf_size)
.end

/*
 * Check whether the label stack is empty (IP header is exposed).
 * If so jump to IP exposed section.
 */
.begin
.reg bottom_of_stack
	alu_shf[bottom_of_stack, --, B, 1, <<8]
	alu[--, bottom_of_stack, AND, HDR_POINTER++]
.end

	bne[pop_fwd_ipv4_exposed#]

; ----------------------------------------------------------------------------
; OPERATION POP_FORWARD
; MPLS packet processing
; ----------------------------------------------------------------------------
/*
 * Paste the previously copied ttl value on the top label provided
 * the tunnel model is uniform.
 */
	immed[labels_to_write, 0]
	br!=byte[tunnel_model, 0, MPLS_UNIFORM_TUNNEL_MODEL, ilm_common_code#]

	alu_op(HDR_POINTER, HDR_POINTER, AND~, 0xFF)
	alu_op(HDR_POINTER, HDR_POINTER, OR, o_ttl)
	immed[labels_to_write, 1]

	br[ilm_common_code#]


; ----------------------------------------------------------------------------
; OPERATION POP_FORWARD
; IP packet processing (IP exposed)
; ----------------------------------------------------------------------------
pop_fwd_ipv4_exposed#:

/*
 * Set packet header type depending payload type: IPv4, IPv6, etc.
 */
.begin
.reg l3prot

	_mpls_ilm_l3protocol_extract(l3prot)
	_mpls_set_header_type(l3prot)

.end	// end of l3prot

	br=byte[tunnel_model, 0, MPLS_UNIFORM_TUNNEL_MODEL, pop_fwd_ipv4_uniform#]

/*
 * Set meta data variables according to the values in NHLFE.
 */
        immed[labels_to_write, 0]


pop_fwd_ipv4_uniform#:
	_mpls_cache_ip_header(labels_in_cache)

/*
 * Set outgoing TTL value in the packet's IP header TTL field,
 * recalculate IP header checksum.
 */
	_mpls_update_ipv4_header_lm(HDR_CACHE, HDR_WORD_OFFSET, o_ttl)


pop_fwd_update_meta#:

	_mpls_update_meta_data(ILM_CACHE)

	alu[dl_next_block, --, B , BID_MPLSILM_NEXT1]
	br[bypass_path#]

.end


;------------------------------------------------------------------------------
; 	POP HANDLING
;------------------------------------------------------------------------------
pop_handling#:
	alu[--, i_ttl, -, 1]
	ble[ilm_ttl_expired#]

/*
 * When POP operation is decoded, the top label's TTL
 * is saved and then the top label is popped from the label stack.
 * The packet's meta data is updated accordingly.
 */
.set labels_in_cache
	alu[labels_in_cache, labels_in_cache, -, 1]

.begin
.reg dl_packet_offset dl_buf_size

		dl_meta_get_offset(dl_packet_offset)
		alu_op(dl_packet_offset, dl_packet_offset, + , 4)
		dl_meta_set_offset(dl_packet_offset)

		dl_meta_get_packet_size(dl_packet_size)
		alu_op(dl_packet_size, dl_packet_size, - , 4)
		dl_meta_set_packet_size(dl_packet_size)

		dl_meta_get_buffer_size(dl_buf_size)
		alu_op(dl_buf_size, dl_buf_size, - , 4)
		dl_meta_set_buffer_size(dl_buf_size)
.end

/*
 * Check whether the label stack is empty (IP header is exposed).
 * If so jump to IP exposed section.
 */
.begin
.reg bottom_of_stack
	alu_shf[bottom_of_stack, --, B, 1, <<8]
	alu[--, bottom_of_stack, AND, HDR_POINTER++]
.end

	bne[pop_ipv4_exposed#]

; ----------------------------------------------------------------------------
; OPERATION POP
; MPLS packet processing
; ----------------------------------------------------------------------------
/*
 * Paste the previously copied ttl value on the top label provided
 * the tunnel model is uniform.
 */
	br!=byte[tunnel_model, 0, MPLS_UNIFORM_TUNNEL_MODEL, check_cache_content#]

	alu_op(HDR_POINTER, HDR_POINTER, AND~, 0xFF)
	alu_op(HDR_POINTER, HDR_POINTER, OR, i_ttl)

/*
 * If decoded operation is POP the next label is looked up (provided
 * it is in cache, if not the packet is directed to the core component)
 * and the packet processing starts from the beginning.
 */
check_cache_content#:
	alu[--, labels_in_cache, -, 1]
	ble[out_of_cache#]
	br[recursive_lookup#]


; ----------------------------------------------------------------------------
; OPERATION POP
; IP packet processing (IP exposed)
; ----------------------------------------------------------------------------
pop_ipv4_exposed#:

	_mpls_cache_ip_header(labels_in_cache)

/*
 * Set packet header type depending payload type: IPv4, IPv6, etc.
 */
.begin
.reg l3prot

	_mpls_ilm_l3protocol_extract(l3prot)
	_mpls_set_header_type(l3prot)

.end	// end of l3prot

	br!=byte[tunnel_model, 0, MPLS_UNIFORM_TUNNEL_MODEL, pop_ipv4_pipe#]

/*
 * Set outgoing TTL value in the packet's IP header TTL field,
 * recalculate IP header checksum.
 */
	_mpls_update_ipv4_header_lm(HDR_CACHE, HDR_WORD_OFFSET, i_ttl)

/*
 * If the tunnel model is pipe and the operation is POP
 * the dl_nexthop_id variable is set to (-1), and the dl_nexthop_id_type
 * variable is set to NHID_IP(0) - based on these value the packet will be
 * ultimately processed by the IPv4 forwarder. Meanwhile, the packet exits
 * the ILM Classifier, and is passed either to the IPv4 forwarder.
 */
pop_ipv4_pipe#:
	dl_meta_set_nexthop_id(0xFFFF)
	dl_meta_set_nexthop_id_type(0)

    ; go to exit 2
	alu[dl_next_block, --, B, BID_MPLSILM_NEXT2]
	br[bypass_path#]


/*
 * Update meta data.
 * Check if packet size does not exceed Effective Maximum Frame
 * Payload Size for Labeled Packets parameter, if it does
 * ILM_LABELED_PACKET_TOO_BIG exception is invoked.
 * Optionally update mplsOutIfcPkts and mplsOutSegPkts counters.
 */
ilm_common_code#:

	_mpls_update_meta_data(ILM_CACHE)

.reg max_lab_pkt_size

	dl_meta_get_packet_size(dl_packet_size)
	_mpls_ilm_max_lab_pkt_size_extract(max_lab_pkt_size)
	alu_op(--, dl_packet_size, -, max_lab_pkt_size)
	bgt[ilm_labeled_packet_too_big#]

	alu[dl_next_block, --, B , BID_MPLSILM_NEXT1]
	br[bypass_path#]

; ----------------------------------------------------------------------------
; Reserved MPLS labels handling
; ----------------------------------------------------------------------------
reserved_label_processing#:

	alu[--, --, B, mpls_label]
	bne[label_1#]

/*
 * A value of 0 is only legal at the bottom of the label stack.
 * If this is not the bottom of the label stack drop the packet.
 * Otherwise, read the special ILM entry for this label and
 * continue normal processing.
 */
	br_bclr[HDR_POINTER, MPLS_STACK_BIT, drop#]

    dl_meta_set_header_type(MPLS_IPV4_HEADER_TYPE)

	/*
	 * When POP operation is decoded, the top label's TTL
	 * is saved and then the top label is popped from the label stack.
	 * The packet's meta data is updated accordingly.
	 */
	.set labels_in_cache
     alu[labels_in_cache, labels_in_cache, -, 1]

	.begin
	.reg dl_packet_offset dl_packet_size dl_buf_size

		dl_meta_get_offset(dl_packet_offset)
		alu_op(dl_packet_offset, dl_packet_offset, + , 4)
		dl_meta_set_offset(dl_packet_offset)

		dl_meta_get_packet_size(dl_packet_size)
		alu_op(dl_packet_size, dl_packet_size, - , 4)
		dl_meta_set_packet_size(dl_packet_size)

		dl_meta_get_buffer_size(dl_buf_size)
		alu_op(dl_buf_size, dl_buf_size, - , 4)
		dl_meta_set_buffer_size(dl_buf_size)
	.end

	_mpls_cache_ip_header(labels_in_cache)

	dl_meta_set_nexthop_id(0xFFFF)
	dl_meta_set_nexthop_id_type(0)

	br[exit_2#]


label_1#:
	alu[--, mpls_label, -, 1]
	bne[label_2#]

/*
 * A value of 1 is not legal at the bottom of the label stack.
 * If this is the bottom of the label stack drop the packet.
 * Otherwise, set exception and send it to core.
 */
	br_bset[HDR_POINTER, MPLS_STACK_BIT, drop#]


	dl_set_exception(BID_MPLSILM, ROUTER_ALERT)
	immed32(dl_next_block, IX_EXCEPTION)
	br[bypass_path#]

label_2#:
	alu[--, mpls_label, -, 2]
	bne[label_3#]

	br_bclr[HDR_POINTER, MPLS_STACK_BIT, drop#]

    dl_meta_set_header_type(MPLS_IPV6_HEADER_TYPE)

	/*
	 * When POP operation is decoded, the top label's TTL
	 * is saved and then the top label is popped from the label stack.
	 * The packet's meta data is updated accordingly.
	 */
	.set labels_in_cache
     alu[labels_in_cache, labels_in_cache, -, 1]

	.begin
	.reg dl_packet_offset dl_packet_size dl_buf_size

		dl_meta_get_offset(dl_packet_offset)
		alu_op(dl_packet_offset, dl_packet_offset, + , 4)
		dl_meta_set_offset(dl_packet_offset)

		dl_meta_get_packet_size(dl_packet_size)
		alu_op(dl_packet_size, dl_packet_size, - , 4)
		dl_meta_set_packet_size(dl_packet_size)

		dl_meta_get_buffer_size(dl_buf_size)
		alu_op(dl_buf_size, dl_buf_size, - , 4)
		dl_meta_set_buffer_size(dl_buf_size)
	.end

	_mpls_cache_ip_header(labels_in_cache)

	dl_meta_set_nexthop_id(0xFFFF)
	dl_meta_set_nexthop_id_type(0)

/*
 * A value of 3 to 15 is reserved. Packets with such labels are dropped.
 */
	br[exit_3#]

label_3#:

	br[drop#]

; ----------------------------------------------------------------------------
; Microblock's outputs
; ----------------------------------------------------------------------------

exit_2#:
	alu[dl_next_block, --, B, BID_MPLSILM_NEXT2]
	br[bypass_path#]

exit_3#:
	alu[dl_next_block, --, B, BID_MPLSILM_NEXT3]
	br[bypass_path#]

; ----------------------------------------------------------------------------
; Exception handling
; ----------------------------------------------------------------------------

ilm_ttl_expired#:
	dl_set_exception(BID_MPLSILM, TTL_EXPIRED)
	immed32(dl_next_block, IX_EXCEPTION)

#ifdef __COUNTERS__
.begin
    .reg  @ilm_set_counter_addr

    immed32(@ilm_set_counter_addr, MPLS_ILM_STAT_BASE)

    #ifdef PER_INTERFACE_LABEL_SPACE
    alu[@ilm_set_counter_addr, @ilm_set_counter_addr, +, ilm_counter_per_port_offset]
    #endif

	; 31 bit set if ILM belongs to ILM set
	br_bclr[ilm_index, 31, cnt_regular_entry#]
    immed32(@ilm_set_counter_addr, MPLS_ILM_NHLFE_SET_STAT_BASE)

cnt_regular_entry#:
    add_shf_left(ilm_counter_offset, @ilm_set_counter_addr, ilm_index, 4)
    add_shf_left(ilm_counter_offset, ilm_counter_offset, ilm_index, 3)

	sram_incr(ilm_counter_offset, MPLS_IN_SEGMENT_ERROR_PKTS)

	#ifdef MPLS_STATS_FULL

	.sig sig_stat
	xbuf_alloc($packet_size, 1, write)
	move($packet_size[0], dl_packet_size)

		// Update number of octets sent on this input ILM
		sram_add($packet_size[0], ilm_counter_offset,
				MPLS_IN_SEGMENT_ERROR_OCTETS, sig_stat, sig_stat, ___)
	xbuf_free($packet_size)

	#endif

.end
#endif
	br[bypass_path#]

out_of_cache#:
	alu[labels_to_write, --, B, 1]
	alu[labels_in_cache, --, B, 1]
	dl_set_exception(BID_MPLSILM, TOO_MANY_POPS)
	immed32(dl_next_block, IX_EXCEPTION)
	br[bypass_path#]

ilm_labeled_packet_too_big#:
/*
 * Labeled packet is sent to the core component if it exceeds
 * Effective Maximum Frame Payload Size for Labeled Packets parameter.
 * mplsOutIfcFragments counter is optionally updated.
 */
	dl_set_exception(BID_MPLSILM, ILM_LABELED_PACKET_TOO_BIG)

	dl_meta_set_nexthop_id(ilm_index)
    alu[ilm_index, --, B, ilm_index, >>16]
	dl_meta_set_output_port(ilm_index)

	immed32(dl_next_block, IX_EXCEPTION)
	br[bypass_path#]

/*
 * Drop packet if the corresponding entry in routing table does not exist.
 * Optionally update per interface counter.
 */
failed_lookup#:
#ifdef __COUNTERS__
	_mpls_update_failed_lookup_counter(input_port)
	br[drop#]
#endif

/*
 * Drop packet that cannot be forwarded due to errors in routing tables.
 * Optionally update per interface counter.
 */
error_packet#:
#ifdef __COUNTERS__
	_mpls_update_in_ifc_error_counter(input_port)
#endif

drop#:

/*
 * Drop packet and exit.
 */
	immed32(dl_next_block, IX_DROP)

bypass_path#:
xbuf_free(ILM_CACHE)

.end
#endm


#endif /* MPLS_ILM_UC */
