/*
 * ============================================================================
 * COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *                                                                          */
/** \file      ix_cc_mpls_fwder.h
 *
 *  \brief     The file defines mpls forwarding slow path.
 *
 */
#if !defined(IX_CC_MPLS_FWDER_H)
#define      IX_CC_MPLS_FWDER_H


/*
 * misc
 */
#define IX_MPLS_RES_LABEL_MAX                    15
#define IX_MPLS_RES_LABEL_IPv4_EXPLICIT_NULL     0
#define IX_MPLS_RES_LABEL_ROUTER_ALARM           1
#define IX_MPLS_RES_LABEL_IPv6_EXPLICIT_NULL     2
#define IX_MPLS_RES_LABEL_IMPLICIT_NULL          3

#define IX_MPLS_FEC_ID_SHIFT                     5
#define IX_MPLS_PIPE_TTL                         0xFF

/* this value define which bit of ilm_index is used to
 * define which table it refers to. It can point at ILM table
 * or ILM_SET table. Microcode must set this bit to let Core
 * Component known from which table it must get data
 */
#define MPLS_BIT_ILM_SET_NOTIFICATION            31
#define MPLS_FLAG_TUNNEL_MASK                    0x30

/**
 * MPLS label structure
 */
typedef struct ix_s_cc_mpls_label_internal
{

    ix_uint32       label          :20,
                    exp            :3,
                    s              :1,
                    ttl            :8;

} ix_cc_mpls_label_internal;




/**
 * ICMP sending options
 */
typedef enum ix_e_icmp_send
{
  IX_CC_MPLS_DO_NOT_SEND_ICMP,
  IX_CC_MPLS_SEND_ICMP

} ix_icmp_send;




/*
 * FUNCTION: _ix_cc_mpls_decap ()  */
/**
 * ABSTRACT: Pops all MPLS labels from packet
 *
 * \param  arg_pData            - pointer to the body of a processed packet
 * \param  arg_pHwMeta          - a pointer to the meta data of a processed packet
 * \param  arg_ptMplsLabelStack - label stack popped from packet
 * \param  arg_pLabelNum        - stack size
 * \param  arg_maxLabelNum      - max number of labels that can be popped *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_decap (ix_cc_mpls_label_internal **arg_pData,
                   ix_hw_buffer_meta          *arg_pHwMeta,
                   ix_cc_mpls_label_internal  *arg_ptMplsLabelStack,
                   ix_uint16                  *arg_pLabelNum,
                   ix_uint16                   arg_maxLabelNum);




/*
 * FUNCTION: _ix_cc_mpls_encap()  */
/**
 * ABSTRACT: Pushes MPLS labels into packet
 *
 * \param  arg_pData            - pointer to the body of a processed packet
 * \param  arg_pHwMeta          - a pointer to the meta data of a processed packet
 * \param  arg_ptMplsLabelStack - label stack popped from packet
 * \param  arg_pLabelNum        - stack size
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_encap(ix_cc_mpls_label_internal **arg_pData,
                  ix_hw_buffer_meta          *arg_pHwMeta,
                  ix_uint16                   arg_maxLabelNum,
                  ix_cc_mpls_label_internal  *arg_ptMplsLabelStack);




/*
 * FUNCTION: _ix_cc_mpls_ilm_fragmentation ()  */
/**
 * ABSTRACT: Fragments MPLS packet coming from ilm forwarder
 *
 * DESCRIPTION:  Function retrieves ilm_id from meta data
 *               Label stack is popped
 *               IPv4 packet is fragmented
 *               Every fragment is encapsulated with label stack
 *               Every fragment is sent to mpls forwarder
 *
 * \param  arg_hBuffer   - buffer handle of a processed packet
 * \param  arg_pContext  - a pointer to the MPLS context
 * \param  arg_OutputId  - communication ID of destination Core Componenet
 * \param  arg_OutputBid - if the packet goes to the microcode this is the block id of the block which 
 *                         will take it over.
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_ilm_fragmentation (ix_buffer_handle   arg_hBuffer,
                               ix_uint16          arg_OutputId,
                               ix_uint16          arg_OutputBid,
                               void              *arg_pContext);



/*
 * FUNCTION: _ix_cc_mpls_ftn_fragmentation ()  */
/**
 * ABSTRACT: Fragments MPLS packet
 *
 * DESCRIPTION:  Label stack is popped
 *               IPv4 packet is fragmented
 *               Every fragment is encapsulated with label stack
 *               Every fragment is sent to mpls forwarder
 *
 * \param  arg_hBuffer   - buffer handle of a processed packet
 * \param  arg_pContext  - a pointer to the MPLS context
 * \param  arg_OutputId  - communication ID of destination Core Componenet
 * \param  arg_OutputBid - if the packet goes to the microcode this is the block id of the block which 
 *                         will take it over.
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_ftn_fragmentation (ix_buffer_handle   arg_hBuffer,
                               ix_uint16          arg_OutputId,
                               ix_uint16          arg_OutputBid,
                               void              *arg_pContext);



/*
 * FUNCTION: _ix_cc_mpls_fragmentation ()  */
/**
 * ABSTRACT: Fragments MPLS packet
 *
 * DESCRIPTION:  Label stack is popped
 *               IPv4 packet is fragmented
 *               Every fragment is encapsulated with label stack
 *               Every fragment is sent to mpls forwarder
 *
 * \param  arg_hBuffer   - buffer handle of a processed packet
 * \param  arg_pContext  - a pointer to the MPLS context
 * \param  arg_OutputId  - communication ID of destination Core Componenet
 * \param  arg_OutputBid - if the packet goes to the microcode this is the block id of the block which 
 *                         will take it over.
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_fragmentation (ix_buffer_handle     arg_hBuffer,
                           void                *arg_pData,
                           ix_hw_buffer_meta   *arg_pHwMeta,
                           ix_uint32            arg_maxIpPktSize,
                           ix_uint16            arg_OutputId,
                           ix_uint16            arg_OutputBid,
#if defined (MPLS_COUNTERS_FEATURE)
                           ix_uint16           *arg_pFragsNum,
                           ix_uint32           *arg_pFragsAddOctets,
#endif
                           void                *arg_pContext);


/*
 * FUNCTION: _ix_cc_mpls_ilm_forwarder()  */
/**
 * ABSTRACT: implements mpls slow path ILM forwarding
 *
 * \param  arg_hBuffer  - buffer handle of a processed packet
 * \param  arg_icmpSend - specifies weather or not generate ICMP packets
 * \param  arg_pContext - a pointer to the MPLS context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_ilm_forwarder(ix_buffer_handle    arg_hBuffer,
                          ix_icmp_send        arg_icmpSend,
                          void               *arg_pContext);






/*
 * FUNCTION: _ix_cc_mpls_ftn_forwarder()  */
/**
 * ABSTRACT: implements mpls slow path FTN forwarding
 *
 * \param  arg_hBuffer  - buffer handle of a processed packet
 * \param  arg_icmpSend - specifies weather or not generate ICMP packets
 * \param  arg_pContext - a pointer to the MPLS context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_ftn_forwarder(ix_buffer_handle    arg_hBuffer,
                          ix_icmp_send        arg_icmpSend,
                          void               *arg_pContext);



#endif /* end !defined(IX_CC_MPLS_FWDER_H) */
