/*
 * ============================================================================
 * COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *                                                                          */
/** \file      ix_cc_mpls_icmp.h
 *
 *  \brief     The file contains functions to handle icmp generation
 *
 */
#if !defined(IX_CC_MPLS_ICMP_H)
#define      IX_CC_MPLS_ICMP_H


#if defined(__cplusplus)
extern "C"
{
#endif



/* maximum number of MPLS labels on the poacket - used to reserve space in tables */
#define   MAX_MPLS_LABS       16

/* ttl set in mpls icmp message */
#define   ICMP_MPLS_TTL       0xFF



/*
 * FUNCTION: _ix_cc_mpls_icmp_init()  */
/**
 * ABSTRACT: Initialization functions to generate ICMP 
 *
 * DESCRIPTION: creates queue for icmp messages.
 *
 * \param  arg_pContext - a pointer to the MPLS context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_icmp_init(void *arg_pContext);



/*
 * FUNCTION: _ix_cc_mpls_icmp_fini()  */
/**
 * ABSTRACT: removes resources taken to generate ICMP 
 *
 * DESCRIPTION: releases the queue for icmp messages.
 *
 * \param  arg_pContext - a pointer to the MPLS context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_icmp_fini(void *arg_pContext);





/*
 * FUNCTION: _ix_cc_mpls_icmp_print_queue()  */
/**
 * ABSTRACT: prints the content of ICMP queue
 *
 * \param  arg_pContext - a pointer to the MPLS context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
#if defined (IX_MPLS_DEBUG)
void
_ix_cc_mpls_icmp_print_queue (void *arg_pContext);
#endif





/*
 * FUNCTION: _ix_cc_mpls_icmp_dest_unreachable ()  */
/**
 * ABSTRACT: builds icmp packet denoting destination unreachability
 *
 * DESCRIPTION: This function attempts to build ICMP dest unreaacheable
 *              error message for the packet passed in as buffer handle
 *
 * \param  arg_pContext - a pointer to the MPLS context
 * \param  arg_hBuffer  - packet buffer handle.
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_icmp_dest_unreachable(ix_buffer_handle    arg_hBuffer,
                                  ix_uint8            arg_Reason,
                                  ix_uint32           arg_Param,
                                  void               *arg_pContext);




/*
 * FUNCTION: _ix_cc_mpls_icmp_time_exceeded ()  */
/**
 * ABSTRACT: builds icmp packet denoting destination unreachability
 *
 * DESCRIPTION: This function attempts to build ICMP
 *              time Exceeded error message for the packet passed in
 *              as buffer handle
 *
 * \param  arg_pContext - a pointer to the MPLS context
 * \param  arg_hBuffer  - packet buffer handle.
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_icmp_time_exceeded(ix_buffer_handle   arg_hBuffer,
                               void              *arg_pContext);





/*
 * FUNCTION: _ix_cc_mpls_icmp_send_message()  */
/**
 * ABSTRACT: function called periodically by the framework to handle enqueued ICMP messages
 *
 * DESCRIPTION:  This function attempts to dequeue ICMP messages and
 *               send them to the destination.
 *
 * \param  arg_pContext - a pointer to the MPLS context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_icmp_send_message (void *arg_pContext);





/*
 * FUNCTION: ix_cc_mpls_get_sleep_time()  */
/**
 * ABSTRACT: gets the sleeping time between calling _ix_cc_mpls_icmp_send_message();
 *
 * DESCRIPTION: This library primitive helps to retrieve the number of seconds
 *              between when the event will fire causing the ICMP messages 
 *              to be dequeued and sent. The lower the value of arg_pSleepTime, 
 *              the more frequently the ICMP messages will be dequeued and sent.  
 *
 * \param  arg_pContext   - pointer to MPLS context
 * \param  arg_pSleepTime - returned ICMP sleep period 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 *
 *   Error Codes:
 * 
 * \retval    IX_CC_ERROR_NULL - null pointer
 */
#ifndef IX_EXCLUDE_CCI
ix_error
ix_cc_mpls_get_sleep_time (ix_uint16 *arg_pSleepTime,
                           void      *arg_pContext);
#endif



/*
 * FUNCTION: ix_cc_mpls_set_sleep_time()  */
/**
 * ABSTRACT: sets the sleeping time between calling _ix_cc_mpls_icmp_send_message();
 *
 * DESCRIPTION: This library primitive helps to set the period between 
 *              events firing to process the ICMP error message queue.
 *              The lower the value of arg_SleepTime, the more frequently
 *              the output queue will be processed. 
 *              The minimum value of arg_SleepTime is 1 sec and 
 *              maximum value can be 32sec.
 *
 * \param  arg_SleepTime- the number of seconds to wait between processing 
 *                        the ICMP error message queue.
 * \param  arg_pContext - pointer to MPLS context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 *
 *   Error Codes:
 * 
 * \retval    IX_CC_MPLS_INVALID_SLEEP_TIME  - null pointer
 */
#ifndef IX_EXCLUDE_CCI
ix_error
ix_cc_mpls_set_sleep_time (ix_uint16  arg_SleepTime,
                           void      *arg_pContext);
#endif


/*
 * FUNCTION: ix_cc_mpls_get_queue_depth()  */
/**
 * ABSTRACT: gets the current length of ICMP queue
 *
 * DESCRIPTION: This library primitive helps to retrieves 
 *              the depth of the ICMP error message queue. 
 *              That is, the maximum number of ICMP messages that can 
 *              fit in the queue, not the number of messages currently in the queue.
 *
 * \param  arg_pQueueDepth - the current number of entries 
 *                           that the ICMP error message queue can hold.
 * \param  arg_pContext    - pointer to MPLS context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 *
 *   Error Codes:
 * 
 * \retval    IX_CC_ERROR_NULL - null pointer
 */
ix_error
ix_cc_mpls_get_queue_depth (ix_uint16 *arg_pQueueDepth,
                            void      *arg_pContext);





/*
 * FUNCTION: ix_cc_mpls_get_packets_to_drain()  */
/**
 * ABSTRACT: gets the max number of ICMP messages sent by _ix_cc_mpls_icmp_send_message() in one turn
 *
 * DESCRIPTION:This library primitive helps to retrieve 
 *             the maximum number of ICMP error messages
 *             to send (from the ICMP message queue ) each time the 
 *             queue processing event fires. 
 *
 * \param  arg_pPacketsToDrain - the maximum number of packets 
 *                               to send each time the event fires for processing the queue.
 * \param  arg_pContext        - pointer to MPLS context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 *
 *   Error Codes:
 * 
 * \retval    IX_CC_ERROR_NULL - null pointer
 */
ix_error
ix_cc_mpls_get_packets_to_drain (ix_uint16 *arg_pPacketsToDrain,
                                 void      *arg_pContext);






/*
 * FUNCTION: ix_cc_mpls_set_packets_to_drain()  */
/**
 * ABSTRACT: sets the max number of ICMP messages sent by _ix_cc_mpls_icmp_send_message() in one turn
 *
 * DESCRIPTION: This library primitive helps to set the maximum number 
 *              of ICMP error messages to send each time the queue processing event fires.
 *              The value of arg_PacketsToDrain must not exceed the depth of the queue.
 *
 * \param   arg_PacketsToDrain - The maximum number of packets to send 
 *                               each time the queue processing event fires. 
 * \param  arg_pContext        - pointer to MPLS context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 *
 *   Error Codes:
 * 
 * \retval    IX_CC_MPLS_ERROR_INVALID_PACKETS_TO_DRAIN  - null pointer
 */
ix_error
ix_cc_mpls_set_packets_to_drain (ix_uint16  arg_PacketsToDrain,
                                 void      *arg_pContext);





#if defined(__cplusplus)
}
#endif 

#endif /* end !defined(IX_CC_MPLS_ICMP) */
