/*
 * ============================================================================
 * COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *                                                                          */
/** \file      ix_cc_mpls_nhlfe.c
 *
 *  \brief     The file contains CCI API function exposed by
 *             MPLS Core Component and used to manage NHLFE table.
 *
 */
/* $Id: ix_cc_mpls.h,v 1.16.2.1 2004/03/12 10:40:45 wneubaue Exp $ */

#if !defined(IX_CC_MPLS_H)
#define      IX_CC_MPLS_H

#if defined(__cplusplus)
extern "C"
{
#endif 




#define MPLS_MAX_PHB_NUM         4
#define MPLS_MAX_EXP_NUM         8
#define MPLS_MAX_NHLFE_IN_SET    4

/*! Maximum number of ports per blade. Currently the default is
 *  defined as 4.
 */
#define MPLS_MAX_LOCAL_PORTS    4


/* Maximum number of NHLFE table entries.*/
#define MPLS_MAX_NHLFE_ENTRIES          65536

/* Minimum incoming label value */
#define MPLS_MIN_LABEL                  16

/* Maximum incoming label value */
#define MPLS_MAX_LABEL                  65536

/* Maximum number of entries in NHLFE set tables */
#define MPLS_SET_TABLE_SIZE             65536

#ifndef MPLS_DEFAULT_SRAM_CH_NUM
#define MPLS_DEFAULT_SRAM_CH_NUM        0 
#endif

#ifndef MPLS_ICMP_EVENT_PRIORITY
#define MPLS_ICMP_EVENT_PRIORITY        1
#endif

#ifndef IX_CC_MPLS_ICMP_MIN_SLEEP_TIME
#define IX_CC_MPLS_ICMP_MIN_SLEEP_TIME  100
#endif

#ifndef IX_CC_MPLS_ICMP_MAX_SLEEP_TIME
#define IX_CC_MPLS_ICMP_MAX_SLEEP_TIME  10000
#endif

/*
 * TYPENAME: ix_cc_mpls_label  */
/**
 * ABSTRACT: MPLS label structure
 */
typedef struct ix_s_cc_mpls_label
{
    ix_uint32       label          :20,
                    reserved       :12;
} ix_cc_mpls_label;


/*
 * TYPENAME: ix_cc_mpls_error  */
/**
 * DESCRIPTION: MPLS error codes
 */
enum ix_e_cc_mpls_error {
    /* all error groups are defined in ix_error.h file */
    IX_CC_MPLS_ERROR_FIRST = IX_ERROR_MAKE_GROUP(CC_MPLS),
    IX_CC_MPLS_ERROR_INVALID_HANDLE,
    IX_CC_MPLS_ERROR_RM_PATCH_SYMBOL,
    IX_CC_MPLS_ERROR_RM_GET_PHYS_OFFSET,
    IX_CC_MPLS_ERROR_RM_REG_HDLR,
    IX_CC_MPLS_ERROR_RM_REM_HDLR,
    IX_CC_MPLS_ERROR_CCI_REG_HDLR,
    IX_CC_MPLS_ERROR_CCI_REM_HDLR,
    IX_CC_MPLS_ERROR_RM_FREE_BUF,
    IX_CC_MPLS_ERROR_MSUP_EXTRACT_MSG,
    IX_CC_MPLS_ERROR_MSUP_SEND_REPLY,
    IX_CC_MPLS_ERROR_INVALID_ENTRY,
    IX_CC_MPLS_ERROR_OSSL_MUTEX_INIT,
    IX_CC_MPLS_ERROR_OSSL_MUTEX_FINI,
    IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK,
    IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK,
    IX_CC_MPLS_ERROR_INVALID_EXCEP_CODE,
    IX_CC_MPLS_ERROR_ENTRY_NHLFE,
    IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET,
    IX_CC_MPLS_ERROR_ENTRY_USED,
    IX_CC_MPLS_ERROR_STATS_NOT_SUPPORTED,
    IX_CC_MPLS_ERROR_RTM_ADD_ENTRY,
    IX_CC_MPLS_ERROR_RTM_ADD_ROUTE,
    IX_CC_MPLS_ERROR_INVALID_PACKETS_TO_DRAIN,
    IX_CC_MPLS_ERROR_LAST
}ix_cc_mpls_error;



/*
 * TYPENAME: ix_cc_mpls_nhlfe_handle  */
/**
 * ABSTRACT: handle to CC NHLFE table entry
 *
 */
typedef struct ix_s_cc_mpls_nhlfe_handle {
    
    ix_uint32 bladeId  :  8,
              outSegId : 24; /* Index to NHLFE table */
                                
} ix_cc_mpls_nhlfe_handle;




/*
 * TYPENAME: ix_cc_mpls_lsp_handle  */
/**
 * ABSTRACT: handle to LSP table entry
 *
 */
typedef struct ix_s_cc_mpls_lsp_handle {

    ix_uint32 bladeId  :  8,
              LspIndex : 24; /* Index to CC LSP table */

} ix_cc_mpls_lsp_handle;



/*
 * TYPENAME:  ix_cc_mpls_exp2phb Data Type */
/**
 * ABSTRACT:  Data structure of EXP to PHB mapping.
 *
 * DESCRIPTION: This type specifies an Exp2Phb table entry from the control
 *              plane perspective. It is used by the MPLS CC API.
 */
typedef struct ix_cc_s_mpls_exp2phb
{
    ix_uint8 class_id;  /*< DiffServ class id */
    ix_uint8 color;     /*< drop precedence */
    ix_uint8 phb;       /*< PHB value */
} ix_cc_mpls_exp2phb;


/*
 * TYPENAME:  ix_cc_mpls_phb2exp Data Type */
/**
 * ABSTRACT:  Data structure of EXP to PHB mapping.
 *
 * DESCRIPTION: PHB to EXP map table entry type definition
 */
typedef struct ix_cc_s_mpls_phb2exp
{
    /*! outgoing EXP for each label stack element */
    ix_uint8 out[4];
} ix_cc_mpls_phb2exp;


/*
 * TYPENAME: ix_cc_mpls_fec         */
/**
 * ABSTRACT: This type specifies an MPLS Forwarding Equivalence Class.
 *
 * DESCRIPTION: FEC parameter structure type. This type doesn't have operation
 *              field as it is only possible to push a label on a non MPLS packet.
 */
typedef struct ix_s_cc_mpls_fec
{
    #define IX_CC_MPLS_FEC_TYPE_IPV4       1
    #define IX_CC_MPLS_FEC_TYPE_IPV6       2

   /*! FEC type: IPv4, IPv6, 6-tuple, etc. */
   ix_uint32 type;
   union
   {
        struct ix_s_ip4fec
        {
            ix_uint32        ipAddress;
            ix_uint32        netMask;
            ix_cc_rtmv4_nhid nexthopId;
        } ip4fec;
        struct ix_s_ip6fec
        {
            ix_uint8  ip6Addr[16];
            ix_uint32 prefixLen;
        } ip6fec;
        /* other FECs to be defined */
   } fec;
} ix_cc_mpls_fec;




/*
 * TYPENAME: ix_cc_mpls_ilm         */
/**
 * ABSTRACT: This type specifies an ILM key for MPLS CC.
 */
typedef struct ix_s_cc_mpls_ilm
{
    ix_cc_mpls_label label;
    ix_uint16        in_port;     /* 0 for per-platform label space */

    /* LSP operation field definitions */
    #define IX_MPLS_LABELOP_SWAP         0
    #define IX_MPLS_LABELOP_SWAP_PUSH    1
    #define IX_MPLS_LABELOP_POP          2
    #define IX_MPLS_LABELOP_POP_FORWARD  3
    #define IX_MPLS_LABELOP_PUSH         4

    ix_uint8         operation;   /* Label operation */
} ix_cc_mpls_ilm;



/*
 * TYPENAME: ix_cc_mpls_nhlfe  */
/**
 * ABSTRACT: NHLFE data type used by the MPLS CC API
 *
 */
typedef struct ix_s_cc_mpls_nhlfe 
{
    ix_uint8  pushCount;       /* number of labels to push, from 1 to 4 */
    ix_uint8  l3Protocol;      /* type of the MPLS payload 
                                *    MPLS_L3_UNKNOWN
                                *    MPLS_L3_IP_V4
                                *    MPLS_L3_IP_V6
                                */
    ix_uint32 nextHopId;       /* index to the L2 table on the egress blade */

    ix_uint16 bladeId;	       /* outgoing blade ID */
    ix_uint16 outPortId;	   /* outgoing port ID */
    ix_uint8  outPortType;     /* outgoing port type - Eth, POS, ATM */

    ix_cc_mpls_label topLabel; /* label pushed on the top of the stack */  
    ix_cc_mpls_label label1;   /* label pushed below top of the stack */
    ix_cc_mpls_label label2;   /* label pushed below label1 */
    ix_cc_mpls_label label3;   /* label pushed below label2 */
    
    ix_uint32 mplsPolicy;      /* criteria for chosing set element */

} ix_cc_mpls_nhlfe;




/* LSP flags field definitions */
#define MPLS_FLAG_TUNNEL_PIPE                 0x00
#define MPLS_FLAG_TUNNEL_UNIFORM              0x10
#define MPLS_FLAG_TUNNEL_SHORT_PIPE           0x20

#define MPLS_FLAG_LSP                         0x00  /* non-Diffserv */
#define MPLS_FLAG_E_LSP                       0x08
#define MPLS_FLAG_L_LSP                       0x0C

#define MPLS_FLAG_TC_ENABLED                  0x02  /* traffic conditioning */
#define MPLS_FLAG_PRECONF_EXP_MARK            0x01  /* perform EXP marking */




/* TYPENAME: ix_cc_mpls_lsp_param     */
/**
 * ABSTRACT: LSP parameter structure type
 *
 * DESCRIPTION: This type specifies the control-plane level view of an LSP.
 *              atributes
 */
typedef struct ix_s_cc_mpls_lsp_param 
{
    ix_uint32 flags;            /* LSP flags                        */
    ix_uint16 maxLabPktSize;    /* LSP MTU                          */
    ix_uint8  ttl;              /* ttl decrement value              */
    ix_uint32 flowID;           /* flow ID for DiffServ             */
    ix_uint16 exp2phbIdx;       /* index into the EXP2PHB table     */
    ix_uint16 phb2expIdx;       /* index into the PHB2EXP table     */
} ix_cc_mpls_lsp_param;





/*
 * TYPENAME: ix_cc_mpls_lsp    */
/**
 * ABSTRACT: LSP data type
 *
 * DESCRIPTION: This type specifies the control-plane level view of an LSP.
 *              It is used by the MPLS CC API.
 */
typedef struct ix_s_cc_mpls_lsp
{
    #define IX_CC_MPLS_LSP_FEC_TYPE          1
    #define IX_CC_MPLS_LSP_ILM_TYPE          2

    ix_uint8 lspType;    /* LSP_FEC(1), LSP_ILM(2) */

    union
    {
        ix_cc_mpls_fec  fec;
        ix_cc_mpls_ilm  ilm;
    } inSeg;

    #define IX_CC_MPLS_LSP_NHLFE_SET_HANDLE_TYPE 1
    #define IX_CC_MPLS_LSP_NHLFE_HANDLE_TYPE     2
    #define IX_CC_MPLS_LSP_NHLFE_DATA_TYPE       3

    ix_uint8                 nhlfeInfoType;  /* NHLFE_SET_HANDLE_TYPE(1),
                                              * NHLFE_HANDLE_TYPE(2),
                                              * NHLFE_DATA_TYPE(3)
                                              */
    ix_cc_mpls_nhlfe_handle  nhlfeSetHandle;

    ix_uint16                nhlfe_num;      /* number of nhlfes in the table */
    union
    {
        ix_cc_mpls_nhlfe_handle  nhlfeHandle[MPLS_MAX_NHLFE_IN_SET];
        ix_cc_mpls_nhlfe        *nhlfeData[MPLS_MAX_NHLFE_IN_SET];
    } outSeg;

    ix_cc_mpls_lsp_param         lspParams[MPLS_MAX_NHLFE_IN_SET];

} ix_cc_mpls_lsp;



/*
 * TYPENAME: ix_cc_mpls_label_range
/**
 * ABSTRACT: defines the range of labels for port. Used in ix_cc_mpls_params
 */
typedef struct ix_s_cc_mpls_label_range
{
    ix_cc_mpls_label min_label;
    ix_cc_mpls_label max_label;
} ix_cc_mpls_label_range;



/*
 * TYPENAME: ix_cc_mpls_lm_exp2phb Data Type */
/**
 * ABSTRACT: Data structure of EXP to PHB mapping.
 *
 */
typedef struct ix_cc_s_mpls_lm_exp2phb
{
    ix_uint32   color_id   :2,
                class_id   :6,
                reserved   :18,
                phbid      :6;                        
} ix_cc_mpls_lm_exp2phb;


/*
 * TYPENAME: ix_cc_mpls_lm_phb2exp
/**
 * ABSTRACT: Data structure of EXP to PHB mapping.
 *
 */
typedef struct ix_s_cc_mpls_lm_phb2exp
{
	ix_uint32   reserved3: 5,
                exp_3:     3,
                reserved2: 5,
                exp_2:     3,
                reserved1: 5,
                exp_1:     3,
                reserved0: 5,
                exp_0:     3;
} ix_cc_mpls_lm_phb2exp;



/*
 * TYPENAME: ix_cc_mpls_params  */
/**
 * ABSTRACT: This type specifies the control-plane level view of MPLS parameters.
 */
typedef struct ix_s_cc_mpls_params
{
    #define IX_CC_MPLS_PER_PLATFORM_LABEL_SPACE      0
    #define IX_CC_MPLS_PER_INTERFACE_LABEL_SPACE     1

    /*! per platform (0), per interface(1) */    
    ix_uint32             labelSpaceMode;
    ix_uint32             maxLocalPorts;
    struct {
        ix_cc_mpls_label  min_label;
        ix_cc_mpls_label  max_label;
    } labelRange[MPLS_MAX_LOCAL_PORTS];
    
    ix_cc_mpls_lm_exp2phb   preconfExp2Phb;
    ix_cc_mpls_lm_phb2exp   preconfPhb2Exp2[MPLS_MAX_PHB_NUM];
} ix_cc_mpls_params;



    
/*
 * TYPENAME: ix_cc_mpls_messages  */
/**
 * DESCRIPTION: defines all the messages recognised by MPLS CC
 */
typedef enum ix_e_cc_mpls_messages
{
    IX_CC_MPLS_MSG_FIRST = 0,
    /*!  This message creates an LSP. */
    IX_CC_MPLS_MSG_LSP_CREATE,
    /*!  This message deletes an LSP. */
    IX_CC_MPLS_MSG_LSP_DELETE,
    /*!  This message modifies an LSP. */
    IX_CC_MPLS_MSG_LSP_MODIFY,
    /*!  This message reads an LSP description. */
    IX_CC_MPLS_MSG_LSP_QUERY,
    /*!  This message reads an LSP statistics. */
    IX_CC_MPLS_MSG_LSP_STATS_QUERY,
    /*!  This message deletes all LSPs. */
    IX_CC_MPLS_MSG_LSP_PURGE,
    /*!  This message creates an NHLFE table entry. */
    IX_CC_MPLS_MSG_NHLFE_CREATE,
    /*!  This message deletes an NHLFE table entry. */
    IX_CC_MPLS_MSG_NHLFE_DELETE,
    /*!  This message reads an NHLFE table entry. */
    IX_CC_MPLS_MSG_NHLFE_QUERY,
    /*!  This message reads an NHLFE table entry statistics. */
    IX_CC_MPLS_MSG_NHLFE_STATS_QUERY,
    /*!  This message deletes all NHLFE table entries. */
    IX_CC_MPLS_MSG_NHLFE_PURGE,
    /*!  This message creates an NHLFE set entry. */
    IX_CC_MPLS_MSG_NHLFE_SET_CREATE,
    /*!  This message deletes an NHLFE set entry. */
    IX_CC_MPLS_MSG_NHLFE_SET_DELETE,
    /*!  This message modifies an NHLFE set entry. */
    IX_CC_MPLS_MSG_NHLFE_SET_MODIFY,
    /*!  This message reads an NHLFE set entry. */
    IX_CC_MPLS_MSG_NHLFE_SET_QUERY,
    /*!  This message modifies an EXP2PHB table entry . */
    IX_CC_MPLS_MSG_EXP2PHB_MODIFY,
    /*!  This message reads an EXP2PHB table entry. */
    IX_CC_MPLS_MSG_EXP2PHB_QUERY,
    /*!  This message modifies a PHB2EXP table entry. */
    IX_CC_MPLS_MSG_PHB2EXP_MODIFY,
    /*!  This message reads a PHB2EXP table entry. */
    IX_CC_MPLS_MSG_PHB2EXP_QUERY,
    /*!  This message reads MPLS Parameters table. */
    IX_CC_MPLS_MSG_PARAM_QUERY,
    /*!  This message reads MPLS port statistics. */
    IX_CC_MPLS_MSG_PORT_STATS_QUERY,
    /*!  This message reads MPLS internal context. */
    IX_CC_MPLS_MSG_GET_MPLS_CONTEXT,
    /* Add new messages before this */
    IX_CC_MPLS_MSG_LAST
} ix_cc_mpls_messages;




/* 
 * TYPENAME:  ix_cc_mpls_lsp_stats */
/**
 * ABSTRACT:  Data structure of statistics for ILM entry.
 *
 * DESCRIPTION: This type specifies the control plane view of
 *              an LSP input statistics, connected with an ILM table entry.
 *              It is used by the MPLS CC API.
 */
typedef struct ix_s_cc_mpls_lsp_stats
{
   ix_uint64 mplsRxOctets; 
   ix_uint64 mplsRxPkts;
   ix_uint64 mplsErrorOctets;	
   ix_uint64 mplsErrorPkts;
   ix_uint64 mplsInTxOctets;
   ix_uint64 mplsInTxPkts;	
} ix_cc_mpls_lsp_stats;




/* 
 * TYPENAME:  ix_s_cc_mpls_nhlfe_stats */
/**
 * ABSTRACT:  Data structure of an NHLFE Counters table entry
 *
 * DESCRIPTION: statistics connected with packets forwarded according 
 *              to a given NHLFE entry. Its entries correspond 1:1
 *              to the NHLFE table entries
 */
typedef struct ix_s_cc_mpls_nhlfe_stats
{
    /*! 64-bit counter accumulating the number of octets received 
     *  through this entry.
     */
    ix_uint64 mplsNhlfeOctets;

    /*
     *! 64-bits counter accumulating the number of packets 
     *  received through this entry
     */
    ix_uint64 mplsNhlfePkts;

    /*! 64-bits counter accumulating the number of octets that were not 
     *  sent through this entry because their TTL expired or they required 
     *  fragmentation but had the DF bit set in the IP header
     */
    ix_uint64 mplsNhlfeErrorOctets;

    /*! 64-bits counter accumulating the number of packets that were 
     *  not sent through this entry because their TTL expired or they 
     *  required fragmentation but had the DF bit set in the IP header
     */
    ix_uint64 mplsNhlfeErrorPkts;

    /*! a 32-bits coutner accumulating the number of octets being the difference
     * between the sum of fragments length and the fragmented packet length
     * sent through this entry. This counter is maintained by the MPLS CC.
     */
    ix_uint64 mplsNhlfeFragmOctetsDelta;

    /*! a 32-bits counter accumunlating the number of fragments sent
     * through entry, excluding the first fragment. This counter is
     * maintained by the MPLS CC.
     */
    ix_uint64 mplsNhlfeFragmPktsDelta;
    
} ix_cc_mpls_nhlfe_stats;
    



/*
 * TYPENAME:  ix_cc_mpls_input_port_stats */
/**
 * ABSTRACT:  MPLS InPort Counters table
 *
 * DESCRIPTION:
 */
typedef struct ix_s_cc_mpls_in_port_stats
{
  /*! 32-bit counter accumulating the number of 
   * MPLS labeled packets received on this interface
   */
  ix_uint32 mplsInIfcRxPkts;
  /*! 32-bit counter accumulating the number packets that were
   *  discarded due to failed MPLS label lookup on this interface
   */
  ix_uint32 mplsInIfcFailedLookupPkts;
  /*! 32-bit counter accumulating the number packets that were
   *  discarded due to errors other than failed MPLS label 
   *  lookup on this interface
   */
  ix_uint32 mplsInIfcErrorPkts;
  /*! 32-bit counter accumulating the number of octets received
   *  in MPLS labeled packets on this interface.
   */
  ix_uint32 mplsInIfcRxOctets;
} ix_cc_mpls_input_port_stats;



/*
 *   CCI API
 *************************************************************************************************/



/*
 * FUNCTION: ix_cc_mpls_init()  */
/**
 * ABSTRACT: function initializes the MPLS CC
 *
 * DESCRIPTION: This function initializes the MPLS core component.
 *              It should be called and returned successfully before any other 
 *              function in the core component can be called. It initializes 
 *              the ILM and NHLFE tables, and MPLS parameters. Then it performs 
 *              static configuration of the ILM and FTN Forwarder microblocks,
 *              allocating and patching required variables and symbols into 
 *              the microblocks by calling Resource Manager's API function,
 *              ix_rm_ueng_patch_symbols().  Finally, it registers the message 
 *              and packet handlers for its defined message inputs and packet input.
 *
 * \param  arg_CcHandle  - handle to the core component, created by the core component 
 8                         infrastructure; this shall be used later to get other services
 *                         from the core component infrastructure.
 * \param  arg_ppContext - handle of the buffer free list and dynamic property data,
 *                         given by the system application during initialization.
 *                         Upon successful completion, it represents the location where 
 *                         the pointer to the control block allocated by MPLS forwarder 
 *                         core component will be stored. The control block is internal
 *                         to the MPLS core component and  contains information about 
 *                         MPLS internal data structures, allocated memory and other 
 *                         relevant information. Later this pointer will be passed to 
 *                         the ix_cc_mpls_fini() function for freeing memory and other 
 *                         house-keeping operations when the MPLS core component is being 
 *                         destroyed
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL - invalid pointer
 * \n         IX_CC_ERROR _OOM   - memory allocation failure      
 * \n         IX_CC_ERROR _FAILED_PATCHING   - patching failures                      
 * \n         IX_CC_ERROR _64BIT_COUNTER_- 64 bit counter creation failure
 * \n         IX_CC_ERROR_CCI - failure from CCI
 */
ix_error
ix_cc_mpls_init(ix_cc_handle   arg_hCcHandle,
                void         **arg_ppContext);





/*
 * FUNCTION: ix_cc_mpls_fini()  */
/**
 * ABSTRACT: The function terminates the MPLS core component.
 *
 * DESCRIPTION: This function terminates the MPLS core component.
 *              It will be executed when the execution engine running 
 *              the core component is being shut down.  This function will free 
 *              all allocated memory and resources allocated by the ix_cc_mpls_init()
 *              function, as well as other functions of the core component.  
 *
 * \param  arg_CcHandle - handle to the core component.
 * \param  arg_pContext - pointer to the control block memory allocated earlier in
 *                        ix_cc_mpls_init(). The termination routine will use it to 
 *                        de-allocate the control block memory.
 * 
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL - invalid pointer
 * \n         IX_CC_ERROR_OOM - memory free failure
 * \n         IX_CC_ERROR_CCI - failure from CCI
 * \n         IX_CC_ERROR_OOM_64BIT_COUNTER - 64 bit counter deletion failure
 */
ix_error
ix_cc_mpls_fini(ix_cc_handle  arg_hCcHandle,
				void         *arg_pContext);




/*
 * TYPENAME:    ix_cc_mpls_cb_general */
/**
 * ABSTRACT:    prototype for user Callback of async messaging function.
 *
 * DESCRIPTION: Client provided callback function.
 *              The core component invokes this function to report
 *              the status of some API calls 
 *
 * \param  arg_Result    - indicates error conditions for the call
 * \param  arg_pContext  - pointer to client-defined context, provided as
 *                         a parameter in the API call. Used by the caller 
 *                         to match asynchronous request with a response.
 */
typedef ix_error (*ix_cc_mpls_cb_general) (ix_error  arg_Result,
                                           void     *arg_pContext);


/*
 *   LSP CREATION
 *************************************************************************************************/



/*
 * FUNCTION: ix_cc_mpls_lsp_create()  */
/**
 * ABSTRACT: This function creates an LSP description defining a FEC to NHLFE, 
 *           or ILM to NHLFE mapping.
 *
 * DESCRIPTION:This function creates an LSP description defining 
 *             FEC to NHLFE, or ILM to NHLFE mapping. 
 *             If arg_Lsp.lspType is equal to LSP_FEC, the function checks 
 *             whether an IPv4 routing table entry for the given FEC exists;
 *             if so, an error is returned. Otherwise, the specified NHLFE
 *             table entry (or entries) is created if arg_Lsp.nhlfeInfoType
 *             equals to NHLFE_DATA. Next the IPv4 next_hop and routing 
 *             table entries for the given FEC are added. Finally, a CC_LSP 
 *             table entry linking the FEC ID with the NHLFE outSegID is created,
 *             and the LSP handle is returned to the caller.
 *
 *             If arg_Lsp.lspType is equal to LSP_ILM, the function checks 
 *             whether an ILM table entry for the given label and input port 
 *             exists; if so, an error is returned. Otherwise, the specified 
 *             NHLFE table entry (or entries) is created if arg_Lsp.nhlfeInfoType 
 *             equals to NHLFE_DATA. Next the ILM table entry for the given label
 *             and input port is added. Finally, a CC_LSP table entry linking 
 *             the ILM entry with the NHLFE outSegID is created, and the LSP 
 *             handle is returned to the caller.
 *
 * \param  arg_pLsp       - pointer to an ix_s_cc_mpls_lsp structure, 
 *                          defining an LSP
 * \param  arg_pContext   - pointer to the MPLS core component context 
 * \param  arg_pLspHandle - pointer to the handle identifying the created LSP
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_DUPLICATE_ENTRY      - the specified FEC or label already exists
 * \n         IX_CC_ERROR_FULL                 - no place in the ILM or NHLFE table
 * \n         IX_CC_ERROR_RANGE                - LSP parameters are out of range
 * \n         IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK - mutex lock failed
 * \n
 * \n NHLFE error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_RANGE                - arg_NhlfeSize is beyond 
 *                                               the range from (1 to 4)
 * \n         IX_CC_ERROR_FULL                 - NHLFE table is full
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - one of the set NHLFEs does not exist
 * \n         IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET - one of the set NHLFEs is an NHLFE set entry
 * \n         IX_CC_MPLS_ERROR_ENTRY_NHLFE     - this is not an NHLFE set entry
 * \n         IX_CC_MPLS_ERROR_INVALID_HANDLE  - invalid NHLFE parameters
 */
ix_error
ix_cc_mpls_lsp_create(ix_cc_mpls_lsp        *arg_Lsp,
                      ix_cc_mpls_lsp_handle *arg_pLspHandle,
                      void                  *arg_pContext);



/*
 * TYPENAME:    ix_cc_mpls_cb_lsp_create      */
/**
 * ABSTRACT:    Client provided callback function.
 *
 * DESCRIPTION: Client provided callback function.
 *              The core component invokes this function to report
 *              the status of an LSP entry creation
 *
 * \param  arg_Result    - indicates error conditions for the call
 * \param  arg_pContext  - pointer to client-defined context, provided as
 *                         a parameter in the API call. Used by the caller 
 *                         to match asynchronous request with a response.
 * \param  arg_nhlfeHandles_num - if greater than 1 it means that entry 
 *                         created entry is associated with NHLFE SET
 *                         In such case it denotes the number of member NHLFE
 *                         in a set.
 * \param  arg_pnhlfeHandles - pointer to nhlfe associated with the current LSP entry.
 *                         If LSP entry is associated with NHLFE SET this is the
 *                         pointer to the table with handles to all members of 
 *                         NHLFE SET
 * \param  arg_nhlfeSetHandle - if created entry is associated with NHLFE SET
 *                         this is the handle to that NHLFE_SET
 * \param  arg_LspHandle - the handle identifying the created LSP.
 */
typedef ix_error (*ix_cc_mpls_cb_lsp_create) (ix_error                 arg_Result,
                                              void                    *arg_pContext,
                                              ix_cc_mpls_nhlfe_handle  arg_nhlfeSetHandle,
                                              ix_uint16                arg_nhlfeHandles_num,
                                              ix_cc_mpls_nhlfe_handle  arg_nhlfeHandles[MPLS_MAX_NHLFE_IN_SET],
                                              ix_cc_mpls_lsp_handle    arg_LspHandle);



/*
 * FUNCTION: ix_cc_mpls_async_lsp_create()   */
/**
 * ABSTRACT: Sends message to create an LSP entry
 *
 * DESCRIPTION: This message helper function creates an LSP description 
 *              defining a FEC to NHLFE, or ILM to NHLFE mapping, and creates
 *              corresponding IP Routing table, ILM and NHLFE entries (if necessary).
 *              See ix_cc_mpls_lsp_create() for a detailed description
 *
 * \param arg_pLsp         - pointer to an ix_s_cc_mpls_lsp structure, defining an LSP
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will be passed to 
 *                           the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_lsp_create(ix_cc_mpls_lsp            *arg_pLsp,
                            ix_cc_mpls_cb_lsp_create   arg_Callback,
                            void                      *arg_pUserContext);




/*
 *   LSP DELETION
 *************************************************************************************************/

				


/*
 * FUNCTION: ix_cc_mpls_lsp_delete()  */
/**
 * ABSTRACT: This function deletes an LSP description from the CC_ILM table,
 *           together with its corresponding IP Routing table,
 *           ILM and NHLFE entries (if necessary)
 *
 * \param  arg_LspHandle - handle identifying an LSP. 
 * \param  arg_pContext  - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_MPLS_ERROR_INVALID_HANDLE  - invalid arg_LspHandle
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - no such entry 
 * \n         IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK - mutex lock failed
 * \n
 * \n LSP add error codes:
 * \n         IX_CC_ERROR_OOM_SYSTEM           - memory allocation failure
 */
ix_error
ix_cc_mpls_lsp_delete(ix_cc_mpls_lsp_handle  arg_LspHandle,
                      void                  *arg_pContext);



/*
 * FUNCTION: ix_cc_mpls_async_lsp_delete()   */
/**
 * ABSTRACT: Sends message to delete LSP entry
 *
 * DESCRIPTION: This message helper function deletes an LSP description,
 *              together with its corresponding IP Routing table,
 *              ILM and NHLFE entries (if necessary).
 *              See ix_cc_mpls_lsp_delete() for a detailed description
 *
 * \param arg_LspHandle    - handle identifying an LSP. 
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will be passed
 *                           to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_lsp_delete(ix_cc_mpls_lsp_handle  arg_LspHandle,
                            ix_cc_mpls_cb_general  arg_Callback,
                            void                  *arg_pUserContext);



/*
 *   LSP MODIFICATION
 *************************************************************************************************/


/*
 * FUNCTION: ix_cc_mpls_lsp_modify()  */
/**
 * ABSTRACT: This function modifies mapping of NHLFEs for an LSP entry
 *
 * DESCRIPTION:This function modifies the NHLFE set mapping for a given LSP.
 *             The new NHLFE set can be specified either as one NHLFE 
 *             handle pointing to a NHLFE set entry, or as up to four 
 *             NHLFE handles pointing to regular (non-set) NHLFEs. 
 *
 * \param  arg_LspHandle        - handle identifying an LSP
 * \param  arg_nhlfeInfoType    - specifies weather function should use handle to predefined
 *                                NHLFE SET or the table of single NHLFEs during LSP modification.
 * \param  arg_nhlfeHandles_num - the number of NHLFE in a table to be cludtered in a set and
 *                                associated with the requested LSP entry.
 * \param  arg_nhlfeHandles     - table of nhlfe handles to be associated with the LSP entry.
 *                                If LSP entry is to be associated with NHLFE SET
 *                                this is the table with handles 
 *                                to all members of this NHLFE SET. 
 * \param  arg_pnhlfeSetHandle  - pointer to a placeholder of handle to predefined NHLFE SET
 *                                if LSP entry is to be associated with allready
 *                                created NHLFE SET this is the pointer to the handle to that 
 *                                NHLFE_SET 
 * \param  arg_lspParams        - pointer to the array of LSP parameters for each element of
 *                                the new NHLFE set
 * \param  arg_pContext         - pointer to the MPLS core component context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - LSP or NHLFE does not exist
 * \n         IX_CC_ERROR_RANGE                - LSP parameters are out of range
 * \n         IX_CC_MPLS_ERROR_INVALID_HANDLE  - invalid arg_LspHandle
 * \n         IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK - mutex lock failed
 * \n         IX_CC_MPLS_ERROR_INVALID_ENTRY   - invalid LSP parameters
 * \n
 * \n RTMv4 error codes:
 * \n         IX_CC_ERROR_NULL
 * \n         IX_CC_ERROR_ALIGN
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND
 * \n         IX_CC_ERROR_INTERNAL
 * \n         IX_CC_ERROR_DUPLICATE_ENTRY
 * \n         IX_CC_ERROR_CONFLICTING_ENTRY
 * \n         IX_CC_ERROR_FULL
 * \n         IX_CC_RTMV4_ERROR_INVALID_NH_INFO
 */
ix_error
ix_cc_mpls_lsp_modify(ix_cc_mpls_lsp_handle     arg_LspHandle,
                      ix_uint8                  arg_nhlfeInfoType,
                      ix_cc_mpls_nhlfe_handle  *arg_pNhlfeSetHandle,
                      ix_uint16                 arg_nhlfeHandles_num,
                      ix_cc_mpls_nhlfe_handle   arg_nhlfeHandles[MPLS_MAX_NHLFE_IN_SET],
                      ix_cc_mpls_lsp_param      arg_lspParams[MPLS_MAX_NHLFE_IN_SET],
                      void                     *arg_pContext);



/*
 * TYPENAME:    ix_cc_mpls_cb_lsp_modify      */
/**
 * ABSTRACT:    Client provided callback function.
 *
 * DESCRIPTION: Client provided callback function.
 *              The core component invokes this function to report
 *              the status of an LSP entry modification
 *
 * \param  arg_Result    - indicates error conditions for the call
 * \param  arg_pContext  - pointer to client-defined context, provided as
 *                         a parameter in the API call. Used by the caller 
 *                         to match asynchronous request with a response.
 * \param  arg_nhlfeSetHandle - if during modification NHLFE SET has been created
 *                         this is the handle to that NHLFE_SET
 */
typedef ix_error (*ix_cc_mpls_cb_lsp_modify) (ix_error                 arg_Result,
                                              void                    *arg_pContext,
                                              ix_cc_mpls_nhlfe_handle  arg_nhlfeSetHandle);


/*
 * FUNCTION: ix_cc_mpls_async_lsp_modify()   */
/**
 * ABSTRACT: Sends message to modify LSP entry
 *
 * DESCRIPTION: This message helper function modifies the NHLFE set mapping 
 *              for a given LSP. See ix_cc_mpls_lsp_modify()
 *              for a detailed description.
 *
 * \param  arg_LspHandle        - handle identifying an LSP
 * \param  arg_nhlfeInfoType    - specifies weather function should use handle to predefined
 *                                NHLFE SET or the table of single NHLFEs during LSP modification.
 * \param  arg_nhlfeHandles_num - the number of NHLFE in a table to be cludtered in a set and
 *                                associated with the requested LSP entry.
 * \param  arg_nhlfeHandles     - table of nhlfe handles to be associated with the LSP entry.
 *                                If LSP entry is to be associated with NHLFE SET
 *                                this is the table with handles 
 *                                to all members of this NHLFE SET. 
 * \param  arg_pnhlfeSetHandle  - pointer to a placeholder of handle to predefined NHLFE SET
 *                                if LSP entry is to be associated with allready
 *                                created NHLFE SET this is the pointer to the handle to that 
 *                                NHLFE_SET 
 * \param  arg_lspParams        - pointer to the array of LSP parameters for each element of
 *                                the new NHLFE set
 * \param  arg_Callback         - pointer to the client provided callback function
 * \param  arg_pUserContext     - pointer to the client defined context that will 
 *                                be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_lsp_modify(ix_cc_mpls_lsp_handle    arg_LspHandle,
                            ix_uint8                 arg_nhlfeInfoType,
                            ix_cc_mpls_nhlfe_handle  arg_nhlfeSetHandle,
                            ix_uint16                arg_nhlfeHandles_num,
                            ix_cc_mpls_nhlfe_handle  arg_nhlfeHandles[MPLS_MAX_NHLFE_IN_SET],
                            ix_cc_mpls_lsp_param     arg_lspParams[MPLS_MAX_NHLFE_IN_SET],
                            ix_cc_mpls_cb_lsp_modify arg_Callback,
                            void                    *arg_pUserContext);



/*
 *   LSP QUERY
 *************************************************************************************************/


/*
 * FUNCTION: ix_cc_mpls_lsp_query()  */
/**
 * ABSTRACT: This function reads and LSP description.
 *
 * DESCRIPTION: This function reads and LSP description. 
 *
 * \param  arg_LspHandle - handle identifying an LSP
 * \param  arg_pLsp      - pointer to the buffer for storing the LSP 
 *                         information returned by the function.
 * \param  arg_pContext  - pointer to the MPLS core component context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_MPLS_ERROR_INVALID_HANDLE  - invalid arg_LspHandle
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - LSP or NHLFE does not exist
 * \n         IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK - mutex lock failed
 */
ix_error
ix_cc_mpls_lsp_query(ix_cc_mpls_lsp_handle  arg_LspHandle,
                     ix_cc_mpls_lsp        *arg_pLsp,
                     void                  *arg_pContext);


/*
 * TYPENAME: ix_cc_mpls_cb_lsp_query      */
/**
 * ABSTRACT: Client provided callback function.
 *
 * DESCRIPTION: Client provided callback function.
 *              The core component invokes this function to report
 *              the status of LSP reading operation
 *
 * \param  arg_Result    - indicates error conditions for the call
 * \param  arg_pContext  - pointer to client-defined context, provided as
 *                         a parameter in the API call. Used by the caller 
 *                         to match asynchronous request with a response.
 * \param  arg_Lsp       - pointer to the buffer storing the LSP information 
 *                         returned by the ix_cc_mpls_lsp_query() function.
 */
typedef ix_error (*ix_cc_mpls_cb_lsp_query) (ix_error         arg_Result,
                                              void           *arg_pContext,
                                              ix_cc_mpls_lsp *arg_Lsp);



/*
 * FUNCTION: ix_cc_mpls_async_lsp_query()   */
/**
 * ABSTRACT: Sends message to get LSP
 *
 * DESCRIPTION: This message helper function reads and LSP description. 
 *
 * \param arg_LspHandle         - handle identifying an LSP
 * \param arg_Callback          - pointer to the client provided callback function
 * \param arg_pUserContext      - pointer to the client defined context that will 
 *                                be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_lsp_query(ix_cc_mpls_lsp_handle    arg_LspHandle,
                           ix_cc_mpls_cb_lsp_query  arg_Callback,
                           void                    *arg_pUserContext);



/*
 *   LSP STATISTICS
 *************************************************************************************************/


/*
 * FUNCTION: ix_cc_mpls_lsp_stats_query()  */
/**
 * ABSTRACT: This function reads LSP statistics.
 *
 * DESCRIPTION: This function gets LSP statistics.
 *
 * \param  arg_LspHandle     - handle identifying an LSP
 * \param  arg_LspStatistics - pointer to the buffer for storing 
 *                             the LSP statistics returned by the function.  
 * \param  arg_pContext      - pointer to the MPLS core component context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                     - invalid pointer
 * \n         IX_CC_MPLS_ERROR_INVALID_HANDLE      - invalid arg_LspHandle
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND          - LSP does not exist
 * \n         IX_CC_MPLS_ERROR_STATS_NOT_SUPPORTED - statistics are not supported
 * \n         IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK     - mutex lock failed
 */
ix_error
ix_cc_mpls_lsp_stats_query(ix_cc_mpls_lsp_handle  arg_LspHandle,
                           ix_cc_mpls_lsp_stats  *arg_pLspStatistics,
                           void                  *arg_pContext);




/*
 * TYPENAME: ix_cc_mpls_cb_lsp_stats_query        */
/**
 * ABSTRACT: Client provided callback function.
 *
 * DESCRIPTION: Client provided callback function.
 *              The core component invokes this function to report
 *              the status of LSP statistics reading operation.
 *
 * \param  arg_Result        - indicates error conditions for the call
 * \param  arg_pContext      - pointer to client-defined context, provided as
 *                             a parameter in the API call. Used by the caller 
 *                             to match asynchronous request with a response.
 * \param  arg_LspStatistics - pointer to the buffer storing the LSP information 
 *                             returned by the ix_cc_mpls_lsp_stats_query() function.
 */
typedef ix_error (*ix_cc_mpls_cb_lsp_stats_query) (ix_error              arg_Result,
                                                   void                 *arg_pContext,
                                                   ix_cc_mpls_lsp_stats *arg_LspStatistics);




/*
 * FUNCTION: ix_cc_mpls_async_lsp_stats_query()   */
/**
 * ABSTRACT: Sends message to get LSP statistics
 *
 * DESCRIPTION: This message helper function reads and LSP description. 
 *
 * \param arg_LspHandle         - handle identifying an LSP
 * \param arg_Callback          - pointer to the client provided callback function
 * \param arg_pUserContext      - pointer to the client defined context that will 
 *                                be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_lsp_stats_query(ix_cc_mpls_lsp_handle          arg_LspHandle,
                                 ix_cc_mpls_cb_lsp_stats_query  arg_Callback,
                                 void                          *arg_pUserContext);


/*
 *   LSP PURGE
 *************************************************************************************************/



/*
 * FUNCTION: ix_cc_mpls_lsp_purge()  */
/**
 * ABSTRACT: This function removes all entries from LSP table.
 *
 * DESCRIPTION: This function deletes all LSP definitions,
 *              together with their corresponding IP Routing table,
 *              ILM and NHLFE entries. 
 *
 * \param  arg_pContext      - pointer to the MPLS core component context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK - mutex lock failed
 */
ix_error
ix_cc_mpls_lsp_purge(void *arg_pContext);




/*
 * FUNCTION: ix_cc_mpls_async_lsp_purge()   */
/**
 * ABSTRACT: Sends message to purge LSP table
 *
 * DESCRIPTION: This message helper function deletes all LSP definitions,
 *              together with its corresponding IP Routing table,
 *              ILM and NHLFE entries. 
 *
 * \param arg_Callback          - pointer to the client provided callback function
 * \param arg_pUserContext      - pointer to the client defined context that will 
 *                                be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_lsp_purge(ix_cc_mpls_cb_general   arg_Callback,
                           void                   *arg_pUserContext);



/*
 *   NHLFE CREATION
 *************************************************************************************************/


/*
 * FUNCTION: ix_cc_mpls_nhlfe_create()  */
/**
 * ABSTRACT: This function creates NHLFE & CC_NHLFE entries.
 *
 * DESCRIPTION: This function creates a NHLFE and corresponding CC_NHLFE 
 *              entry. On creation, the CC_NHLFE entry refCount 
 *              field is set to 0. It is incremented by each LSP creation 
 *              that specifies this NHLFE as its outgoing segment, 
 *              and decremented by such an LSP deletion.
 *
 * \param  arg_Nhlfe        - new NHLFE data to be added
 * \param  arg_pNhlfeHandle - pointer to the handle identifying the created NHLFE.  
 * \param  arg_pContext     - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n           IX_CC_ERROR_NULL - invalid pointer
 * \n           IX_CC_ERROR_FULL - no place in the NHLFE table
 */
ix_error
ix_cc_mpls_nhlfe_create(ix_cc_mpls_nhlfe         arg_Nhlfe,
                        ix_cc_mpls_nhlfe_handle *arg_pNhlfeHandle,
                        void                    *arg_pContext);



/*
 * TYPENAME: ix_cc_mpls_cb_nhlfe_create      */
/**
 * ABSTRACT: Client provided callback function.
 *
 * DESCRIPTION: Client provided callback function.
 *              The core component invokes this function to report
 *              the status of new NHLFE entry creation
 *
 * \param  arg_Result      - indicates error conditions for the call
 * \param  arg_pContext    - pointer to client-defined context, provided as
 *                           a parameter in the API call. Used by the caller 
 *                           to match asynchronous request with a response.
 * \param  arg_NhlfeHandle - the handle identifying the created NHLFE returned 
 *                           by the function.
 */
typedef ix_error (*ix_cc_mpls_cb_nhlfe_create) (ix_error                 arg_Result,
                                                void                    *arg_pContext,
                                                ix_cc_mpls_nhlfe_handle  arg_NhlfeHandle);



/*
 * FUNCTION: ix_cc_mpls_async_nhlfe_create   */
/**
 * ABSTRACT: Sends message to MPLS CC to create NHLFE
 *
 * DESCRIPTION: This message helper function creates an NHLFE table entry.
 *              See ix_cc_mpls_nhlfe_create() for a detailed description.
 *
 * \param arg_pNhlfe            - pointer to the ix_s_cc_mpls_nhlfe structure, defining an NHLFE
 * \param arg_Callback          - pointer to the client provided callback function
 * \param arg_pUserContext      - pointer to the client defined context that will 
 *                                be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_nhlfe_create(ix_cc_mpls_nhlfe 			 *arg_pNhlfe,
                              ix_cc_mpls_cb_nhlfe_create  arg_Callback,
                              void                       *arg_pUserContext);



/*
 *   NHLFE DELETION
 *************************************************************************************************/



/*
 * FUNCTION: ix_cc_mpls_async_nhlfe_delete        */
/**
 * ABSTRACT: Sends message to MPLS CC to delete NHLFE
 *
 * DESCRIPTION: This message helper function deletes an NHLFE entry. 
 *              See ix_cc_mpls_nhlfe_delete() for a detailed description. 
 *
 * \param arg_NhlfeHandle  - handle identifying an NHLFE. 
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_nhlfe_delete(ix_cc_mpls_nhlfe_handle arg_NhlfeHandle,
                              ix_cc_mpls_cb_general   arg_Callback,
                              void                   *arg_pUserContext);




/*
 * FUNCTION: ix_cc_mpls_nhlfe_delete()  */
/**
 * ABSTRACT: This function deletes an NHLFE.
 *
 * DESCRIPTION: This function deletes an NHLFE and corresponding 
 *              CC_NHLFE entries if they are not referenced by 
 *              any LSP definition (that is, the CC_NHLFE entry 
 *              refCount field is equal to 0). This function cannot 
 *              be used for deleting NHLFEs containing set descriptions;
 *              for such entries use ix_cc_mpls_nhlfe_set_delete().
 *
 * \param  arg_NhlfeHandle - handle identifying an NHLFE. 
 * \param  arg_pContext    - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE does not exist
 * \n         IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET - this is an NHLFE set entry
 * \n         IX_CC_MPLS_ERROR_ENTRY_USED      - entry referenced by an LSP
 */
ix_error
ix_cc_mpls_nhlfe_delete(ix_cc_mpls_nhlfe_handle  arg_NhlfeHandle,
                        void                    *arg_pContext);



/*
 *   NHLFE QUERY
 *************************************************************************************************/


/*
 * FUNCTION: ix_cc_mpls_nhlfe_query()  */
/**
 * ABSTRACT: This function gets an NHLFE.
 *
 * DESCRIPTION: This message helper function reads a regular (non-set) 
 *              NHLFE entry. This function cannot be used for reading 
 *              NHLFEs containing set descriptions; for such entries 
 *              use ix_cc_mpls_nhlfe_set_query().
 *
 * \param  arg_NhlfeHandle - handle identifying an NHLFE. 
 * \param  arg_pNhlfe      - pointer to the buffer for storing 
 *                           the NHLFE information returned by the function.  
 * \param  arg_pContext    - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE does not exist
 * \n         IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET - this is an NHLFE set entry
 */
ix_error
ix_cc_mpls_nhlfe_query(ix_cc_mpls_nhlfe_handle  arg_NhlfeHandle,
                       ix_cc_mpls_nhlfe        *arg_pNhlfe,
                       void                    *arg_pContext);


/*
 * TYPENAME: ix_cc_mpls_cb_nhlfe_query        */
/**
 * ABSTRACT: Client provided callback function.
 *
 * DESCRIPTION: Client provided callback function.
 *              The core component invokes this function to report
 *              the status of NHLFE search operation
 *
 * \param  arg_Result    - indicates error conditions for the call
 * \param  arg_pContext  - pointer to client-defined context, provided as
 *                         a parameter in the API call. Used by the caller 
 *                         to match asynchronous request with a response.
 * \param  arg_Nhlfe     - pointer to the buffer storing the NHLFE 
 *                         information returned by the function.
 */
typedef ix_error (*ix_cc_mpls_cb_nhlfe_query) (ix_error          arg_Result,
                                               void             *arg_pContext,
                                               ix_cc_mpls_nhlfe *arg_Nhlfe);


/*
 * FUNCTION: ix_cc_mpls_async_nhlfe_query        */
/**
 * ABSTRACT: Sends message to MPLS CC to get NHLFE
 *
 * DESCRIPTION: This message helper function reads a regular NHLFE entry.
 *
 * \param arg_NhlfeHandle  - handle identifying an NHLFE. 
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_nhlfe_query(ix_cc_mpls_nhlfe_handle    arg_NhlfeHandle,
                             ix_cc_mpls_cb_nhlfe_query  arg_Callback,
                             void                      *arg_pUserContext);


/*
 *   NHLFE STATS QUERY
 *************************************************************************************************/


/*
 * FUNCTION: ix_cc_mpls_nhlfe_stats_query()  */
/**
 * ABSTRACT: This function gets an NHLFE statistics.
 *
 * DESCRIPTION: This function reads NHLFE statistics. 
 *              This function cannot be used for reading NHLFEs 
 *              containing set descriptions; statistics are valid 
 *              only for regular NHLFEs.
 *
 * \param  arg_NhlfeHandle      - handle identifying an NHLFE. 
 * \param  arg_pNhlfeStatistics - pointer to the buffer for storing the 
 *                                NHLFE statistics returned by the function.  
 * \param  arg_pContext         - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE does not exist
 * \n         IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET - this is an NHLFE set entry
 */
ix_error
ix_cc_mpls_nhlfe_stats_query(ix_cc_mpls_nhlfe_handle  arg_NhlfeHandle,
                             ix_cc_mpls_nhlfe_stats  *arg_pNhlfeStatistics,
                             void                    *arg_pContext);



/*
 * TYPENAME: ix_cc_mpls_cb_nhlfe_stats_query        */
/**
 * ABSTRACT: Client provided callback function.
 *
 * DESCRIPTION: Client provided callback function.
 *              The core component invokes this function to report
 *              the status of NHLFE statistics search operation
 *
 * \param  arg_Result     - indicates error conditions for the call
 * \param  arg_pContext   - pointer to client-defined context, provided as
 *                          a parameter in the API call. Used by the caller 
 *                          to match asynchronous request with a response.
 * \param  arg_NhlfeStats - pointer to the buffer storing the LSP statistics 
 *                          information returned by the function.
 */
typedef ix_error (*ix_cc_mpls_cb_nhlfe_stats_query) (ix_error                arg_Result,
                                                     void                   *arg_pContext,
                                                     ix_cc_mpls_nhlfe_stats *arg_NhlfeStats);




/*
 * FUNCTION: ix_cc_mpls_async_nhlfe_stats_query        */
/**
 * ABSTRACT: Sends message to MPLS CC to get NHLFE statistics
 *
 * DESCRIPTION: This message helper function reads NHLFE statistics. 
 *
 * \param arg_NhlfeHandle  - handle identifying an NHLFE. 
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_nhlfe_stats_query(ix_cc_mpls_nhlfe_handle          arg_NhlfeHandle,
                                   ix_cc_mpls_cb_nhlfe_stats_query  arg_Callback,
                                   void                            *arg_pUserContext);


/*
 *   NHLFE PURGE
 *************************************************************************************************/


/*
 * FUNCTION: ix_cc_mpls_nhlfe_purge()  */
/**
 * ABSTRACT: This function deletes all NHLFE table entries.
 *
 * DESCRIPTION: This function deletes all NHLFE table entries.
 *              If any of the entries has a non-zero refCount field,
 *              no entry is deleted. This function can be used for
 *              clearing the NHLFE table before any NHLFE 
 *              was assigned to an LSP.
 *
 * \param  arg_pContext         - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_MPLS_ERROR_ENTRY_USED      - an NHLFE is referenced by an LSP
 */
ix_error
ix_cc_mpls_nhlfe_purge(void *arg_pContext);



/*
 * FUNCTION: ix_cc_mpls_async_nhlfe_stats_query        */
/**
 * ABSTRACT: Sends message to MPLS CC to delete all NHLFEs
 *
 * DESCRIPTION: This message helper function deletes all NHLFE table entries. 
 *
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_nhlfe_purge(ix_cc_mpls_cb_general   arg_Callback,
                             void                   *arg_pUserContext);


/*
 *   NHLFE SET CREATION
 *************************************************************************************************/


/*
 * FUNCTION: ix_cc_mpls_nhlfe_set_create()  */
/**
 * ABSTRACT: Creates an NHLFE table entry containing NHLFE set 
 *
 * DESCRIPTION: Creates an NHLFE table entry containing NHLFE set comprising up to
 *              four regular NHLFEs - the NHLFE handles specyfying the set cannot
 *              point to other sets.
 *
 * \param  arg_nhlfeSetSize      - number of elements in the NHLFE set
 * \param  arg_nhlfeSetHandles[] - pointer to the array of handles defining 
 *                                 the new NHLFE set 
 * \param  arg_NhlfeHandle       - pointer to the handle identifying 
 *                                 the created NHLFE.  
 * \param  arg_pContext          - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_RANGE                - arg_NhlfeSize is beyond 
 *                                               the range from (1 to 4)
 * \n         IX_CC_ERROR_FULL                 - NHLFE table is full
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - one of the set NHLFEs does not exist
 * \n         IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET - one of the set NHLFEs is an NHLFE set entry
 */
ix_error
ix_cc_mpls_nhlfe_set_create(ix_uint16                arg_nhlfeSetSize,
                            ix_cc_mpls_nhlfe_handle  arg_nhlfeSetHandles[],
                            ix_cc_mpls_nhlfe_handle *arg_NhlfeHandle,
                            void                    *arg_pContext);


/*
 * TYPENAME: ix_cc_mpls_cb_nhlfe_set_create          */
/**
 * ABSTRACT: Client provided callback function.
 *
 * DESCRIPTION: Client provided callback function.
 *              The core component invokes this function to report
 *              the status of NHLFE set creation
 *
 * \param  arg_Result      - indicates error conditions for the call
 * \param  arg_pContext    - pointer to client-defined context, provided as
 *                           a parameter in the API call. Used by the caller 
 *                           to match asynchronous request with a response.
 * \param  arg_NhlfeHandle - the handle identifying the created NHLFE set 
 *                           entry returned by the function.
 */
typedef ix_error (*ix_cc_mpls_cb_nhlfe_set_create) (ix_error                 arg_Result,
                                                    void                    *arg_pContext,
                                                    ix_cc_mpls_nhlfe_handle  arg_NhlfeHandle);



/*
 * FUNCTION: ix_cc_mpls_async_nhlfe_set_create        */
/**
 * ABSTRACT: Sends message to MPLS CC to create NHLFE SET
 *
 * DESCRIPTION: This message helper function creates an NHLFE table entry 
 *              containing an NHLFE set. See ix_cc_mpls_nhlfe_set_create()
 *              for a detailed description.
 *
 * \param arg_nhlfeSetSize  - number of elements in the NHLFE set
 * \param arg_nhlfeHandle[] - pointer to the array of handles defining the new NHLFE set
 * \param arg_Callback      - pointer to the client provided callback function
 * \param arg_pUserContext  - pointer to the client defined context that will 
 *                            be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_nhlfe_set_create(ix_uint16                   arg_NhlfeSetSize,
                                  ix_cc_mpls_nhlfe_handle     arg_NhlfeSetHandles[],
                                  ix_cc_mpls_cb_nhlfe_create  arg_Callback,
                                  void                       *arg_pUserContext);


/*
 *   NHLFE SET DELETION
 *************************************************************************************************/


/*
 * FUNCTION: ix_cc_mpls_async_nhlfe_set_delete        */
/**
 * ABSTRACT: Sends message to MPLS CC to delete NHLFE SET
 *
 * DESCRIPTION: This message helper function deletes an NHLFE entry containing an NHLFE set.
 *              See ix_cc_mpls_nhlfe_set_delete() for a detailed description. 
 *
 * \param arg_nhlfeHandle  - handle defining the NHLFE to be deleted
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_nhlfe_set_delete(ix_cc_mpls_nhlfe_handle arg_NhlfeHandle,
                                  ix_cc_mpls_cb_general   arg_Callback,
                                  void                   *arg_pUserContext);



/*
 * FUNCTION: ix_cc_mpls_nhlfe_set_delete()  */
/**
 * ABSTRACT: Deletes an NHLFE table entry containing NHLFE set.
 *
 * DESCRIPTION: This function deletes an NHLFE entry containing 
 *              an NHLFE set. It cannot be used for deleting 
 *              a regular NHLFE; for such an entry use ix_cc_mpls_nhlfe_delete().
 *
 * \param  arg_NhlfeHandle - handle identifying an NHLFE set entry 
 * \param  arg_pContext    - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE does not exist
 * \n         IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET - this is not an NHLFE set entry
 */
ix_error
ix_cc_mpls_nhlfe_set_delete(ix_cc_mpls_nhlfe_handle  arg_NhlfeHandle,
                            void                    *arg_pContext);



/*
 *   NHLFE SET MODIFICATION
 *************************************************************************************************/


/*
 * FUNCTION: ix_cc_mpls_nhlfe_set_modify()  */
/**
 * ABSTRACT: This function modifies the NHLFE table entry containing an NHLFE set.
 *
 * DESCRIPTION: This function modifies the NHLFE table entry containing an NHLFE set.
 *              The NHLFE handles comprising the old set definition are replaced 
 *              by the handles specified by the arg_nhlfeSetSize and 
 *              arg_nhlfeSetHandles[] function arguments. The new set handles 
 *              must point to regular (non-set) NHLFEs.
 *
 * \param  arg_NhlfeHandle        -	handle identifying an NHLFE entry 
 *                                  containing an NHLFE set
 * \param  arg_nhlfeSetSize       -	number of elements in the new NHLFE set
 * \param  arg_nhlfeSetHandles[]  - pointer to the array of handles defining 
 *                                  the new NHLFEs for the set 
 * \param  arg_pContext           - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE does not exist
 * \n         IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET - this is not an NHLFE set entry
 */
ix_error
ix_cc_mpls_nhlfe_set_modify(ix_cc_mpls_nhlfe_handle  arg_NhlfeHandle,
                            ix_uint16                arg_nhlfeSetSize,
                            ix_cc_mpls_nhlfe_handle  arg_nhlfeSetHandles[MPLS_MAX_NHLFE_IN_SET],
                            void                    *arg_pContext);




/*
 * FUNCTION: ix_cc_mpls_async_nhlfe_set_modify        */
/**
 * ABSTRACT: Sends message to MPLS CC to modify NHLFE SET
 *
 * DESCRIPTION: This message helper function modifies the NHLFE table entry
 *              containing an NHLFE set. See ix_cc_mpls_nhlfe_set_modify()
 *              for a detailed description.
 *
 * \param arg_nhlfeHandle       - handle identifying an NHLFE entry containing an NHLFE set
 * \param arg_nhlfeSetSize      - number of elements in the new NHLFE set
 * \param arg_nhlfeSetHandles[] - pointer to the array of handles defining the new NHLFEs for the set 
 * \param arg_Callback          - pointer to the client provided callback function
 * \param arg_pUserContext      - pointer to the client defined context that will 
 *                                be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_nhlfe_set_modify(ix_cc_mpls_nhlfe_handle arg_NhlfeHandle,
                                  ix_uint16               arg_NhlfeSetSize,
                                  ix_cc_mpls_nhlfe_handle arg_NhlfeSetHandles[MPLS_MAX_NHLFE_IN_SET],
                                  ix_cc_mpls_cb_general   arg_Callback,
                                  void                   *arg_pUserContext);


/*
 *   NHLFE SET QUERY
 *************************************************************************************************/


/*
 * FUNCTION: ix_cc_mpls_nhlfe_set_query()  */
/**
 * ABSTRACT: This function gets an NHLFE entry containing an NHLFE set.
 *
 * DESCRIPTION: This function reads an NHLFE entry containing an NHLFE set.
 *              It returns only the NHLFE handles of the set elements.
 *              The definition of each set element must be read separately 
 *              by calling the  ix_cc_mpls_nhlfe_query() function for 
 *              each returned handle.
 *
 * \param  arg_NhlfeHandle     - handle identifying an NHLFE entry containing 
 *                               an NHLFE set
 * \param  arg_SetSize         - pointer to the buffer storing the number of
 *                               the NHLFE set handles returned by the function.
 * \param  arg_NhlfeSetHandles - pointer to the buffer for storing the NHLFE
 *                               handles identifying the set elements returned 
 *                               by the function.  
 * \param  arg_pContext        - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE does not exist
 * \n         IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET - this is not an NHLFE set entry
 */
ix_error
ix_cc_mpls_nhlfe_set_query(ix_cc_mpls_nhlfe_handle   arg_NhlfeHandle,
                           ix_uint16                *arg_nhlfeSetSize,
                           ix_cc_mpls_nhlfe_handle   arg_nhlfeSetHandles[MPLS_MAX_NHLFE_IN_SET],
                           void                     *arg_pContext);


/*
 * TYPENAME: ix_cc_mpls_cb_nhlfe_set_query          */
/**
 * ABSTRACT: Client provided callback function.
 *
 * DESCRIPTION: Client provided callback function.
 *              The core component invokes this function to report
 *              the status of NHLFE set query
 *
 * \param  arg_Result          - indicates error conditions for the call
 * \param  arg_pContext        - pointer to client-defined context, provided as
 *                               a parameter in the API call. Used by the caller 
 *                               to match asynchronous request with a response.
 * \param  arg_SetSize         - the number of the NHLFE set handles.
 * \param  arg_NhlfeSetHandles - pointer to the buffer storing the NHLFE handles 
 *                               identifying the set of elements returned by the function.
 */
typedef ix_error (*ix_cc_mpls_cb_nhlfe_set_query) (ix_error                 arg_Result,
                                                   void                    *arg_pContext,
                                                   ix_uint16                arg_SetSize,
                                                   ix_cc_mpls_nhlfe_handle  arg_NhlfeSetHandles[MPLS_MAX_NHLFE_IN_SET]);



/*
 * FUNCTION: ix_cc_mpls_async_nhlfe_set_query        */
/**
 * ABSTRACT: Sends message to MPLS CC to get NHLFE SET
 *
 * DESCRIPTION: This message helper function reads an NHLFE entry containing an NHLFE set.
 *              See ix_cc_mpls_nhlfe_set_query() for a detailed description.
 *
 * \param arg_nhlfeHandle       - handle identifying an NHLFE entry containing an NHLFE set
 * \param arg_Callback          - pointer to the client provided callback function
 * \param arg_pUserContext      - pointer to the client defined context that will 
 *                                be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_nhlfe_set_query(ix_cc_mpls_nhlfe_handle        arg_NhlfeHandle,
                                 ix_cc_mpls_cb_nhlfe_set_query 	arg_Callback,
                                 void					       *arg_pUserContext);


/*
 *   EXP2PHB MODIFICATION
 *************************************************************************************************/


/*
 * FUNCTION: ix_cc_mpls_exp2phb_modify ()  */
/**
 * ABSTRACT: This function modifies an element of  EXP2PHB mapping table.
 *
 * DESCRIPTION: This function modifies an EXP2PHB table entry. 
 *              Note: all EXP2PHB table entries are created 
 *              during initialization and set to defaults.
 *
 * \param  arg_Index    - index identifying an EXP2PHB table entry
 * \param  arg_exp2phb  - pointer to the table containing the mappings 
 *                        for one EXP2PHB table entry
 * \param  arg_pContext - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE does not exist
 */
ix_error
ix_cc_mpls_exp2phb_modify (ix_uint16           arg_Index,
                           ix_cc_mpls_exp2phb  arg_exp2phb[MPLS_MAX_EXP_NUM],
                           void               *arg_pContext);


/*
 * FUNCTION: ix_cc_mpls_async_exp2phb_modify        */
/**
 * ABSTRACT: Sends message to MPLS CC to modify a mapping in exp2phb table
 *
 * DESCRIPTION: This message helper function modifies an EXP2PHB table entry. 
 *              Note: all EXP2PHB table entries are created during initialization 
 *              and set to defaults.
 *
 * \param arg_Index        - index identifying an EXP2PHB table entry
 * \param arg_exp2phb      - pointer to the table containing the mappings for one EXP2PHB 
 *                           table entry
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_exp2phb_modify(ix_uint16              arg_Index,
                                ix_cc_mpls_exp2phb     arg_exp2phb[MPLS_MAX_EXP_NUM],
                                ix_cc_mpls_cb_general  arg_Callback,
                                void                  *arg_pUserContext);


/*
 *   EXP2PHB QUERY
 *************************************************************************************************/


/*
 * FUNCTION: ix_cc_mpls_exp2phb_query  ()  */
/**
 * ABSTRACT: This function gets an element of  EXP2PHB mapping table.
 *
 * DESCRIPTION: This function reads an EXP2PHB table entry. 
 *
 * \param  arg_Index    - index identifying an EXP2PHB table entry
 * \param  arg_exp2phb  - pointer to the buffer storing the table
 *                        containing the mappings for one EXP2PHB table entry.  
 * \param  arg_pContext - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE does not exist
 */
ix_error
ix_cc_mpls_exp2phb_query (ix_uint16            arg_Index,
                          ix_cc_mpls_exp2phb   arg_exp2phb[MPLS_MAX_EXP_NUM],
                          void                *arg_pContext);
 
/*
 * TYPENAME: ix_cc_mpls_cb_exp2phb_query          */
/**
 * ABSTRACT: Client provided callback function.
 *
 * DESCRIPTION: Client provided callback function.
 *              The core component invokes this function to pass
 *              an EXP2PHB entry connected with the NHLFE.
 *
 * \param  arg_Result          - indicates error conditions for the call
 * \param  arg_pContext        - pointer to client-defined context, provided as
 *                               a parameter in the API call. Used by the caller 
 *                               to match asynchronous request with a response.
 * \param  arg_exp2phb         - pointer to the buffer storing the table containing 
 *                               the mappings for one EXP2PHB table entry.
 */
typedef ix_error (*ix_cc_mpls_cb_exp2phb_query) (ix_error            arg_Result,
                                                 void               *arg_pContext,
                                                 ix_cc_mpls_exp2phb *arg_exp2phb);



/*
 * FUNCTION: ix_cc_mpls_async_exp2phb_query        */
/**
 * ABSTRACT: Sends message to MPLS CC to get a mapping from exp2phb table
 *
 * DESCRIPTION: This message helper function reads an EXP2PHB table entry. 
 *
 * \param arg_Index        - index identifying an EXP2PHB table entry
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_exp2phb_query(ix_uint16                    arg_Index,
                               ix_cc_mpls_cb_exp2phb_query  arg_Callback,
                               void                        *arg_pUserContext);

/*
 *   PHB2EXP MODIFICATION
 *************************************************************************************************/


/*
 * FUNCTION: ix_cc_mpls_phb2exp_modify ()  */
/**
 * ABSTRACT: This function modifies a PHB2EXP table entry.
 *
 * DESCRIPTION: This function modifies a PHB2EXP table entry.
 *              Note: all PHB2EXP table entries are created during 
 *              initialization and set to defaults. 
 *
 * \param  arg_Index    - index identifying a PHB2EXP table entry
 * \param  arg_phb2exp  - pointer to the mappings for one PHB2EXP table entry
 * \param  arg_pContext - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE does not exist
 */
ix_error
ix_cc_mpls_phb2exp_modify (ix_uint16           arg_Index,
                           ix_cc_mpls_phb2exp  arg_phb2exp[MPLS_MAX_PHB_NUM],
                           void               *arg_pContext);



/*
 * FUNCTION: ix_cc_mpls_async_phb2exp_modify       */
/**
 * ABSTRACT: Sends message to MPLS CC to modify a mapping in phb2exp table
 *
 * DESCRIPTION: This message helper function modifies a PHB2EXP table entry.
 *              Note: all PHB2EXP table entries are created during 
 *              initialization and set to defaults. 
 *
 * \param arg_Index        - index identifying an EXP2PHB table entry
 * \param arg_phb2exp      - pointer to the table containing the mappings for 
 *                           a PHB2EXP table entry
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_phb2exp_modify(ix_uint16              arg_Index,
                                ix_cc_mpls_phb2exp     arg_phb2exp[MPLS_MAX_PHB_NUM],
                                ix_cc_mpls_cb_general  arg_Callback,
                                void                  *arg_pUserContext);

/*
 *   PHB2EXP QUERY
 *************************************************************************************************/



/*
 * FUNCTION: ix_cc_mpls_phb2exp_query()  */
/**
 * ABSTRACT: This function gets a PHB2EXP table entry.
 *
 * DESCRIPTION: This function reads a PHB2EXP table entry
 *
 * \param  arg_Index    - index identifying a PHB2EXP table entry
 * \param  arg_pContext - pointer to the MPLS core component context
 * \param  arg_phb2exp  - pointer to the buffer storing the mappings
 *                        for one PHB2EXP table entry returned 
 *                        by the function.  
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE does not exist
 */
ix_error
ix_cc_mpls_phb2exp_query (ix_uint16            arg_Index,
                          ix_cc_mpls_phb2exp   arg_phb2exp[MPLS_MAX_PHB_NUM],
                          void                *arg_pContext);



/*
 * TYPENAME:    ix_cc_mpls_cb_phb2exp_query          */
/**
 * ABSTRACT:    Client provided callback function.
 *
 * DESCRIPTION: Client provided callback function.
 *              The core component invokes this function to pass
 *              an PHB2EXP entry connected with the NHLFE.
 *
 * \param  arg_Result          - indicates error conditions for the call
 * \param  arg_pContext        - pointer to client-defined context, provided as
 *                               a parameter in the API call. Used by the caller 
 *                               to match asynchronous request with a response.
 * \param  arg_exp2phb         - pointer to the buffer storing the table containing 
 *                               the mappings for one PHB2EXP table entry.
 */
typedef ix_error (*ix_cc_mpls_cb_phb2exp_query) (ix_error            arg_Result,
                                                 void               *arg_pContext,
                                                 ix_cc_mpls_phb2exp *arg_phb2exp);




/*
 * FUNCTION: ix_cc_mpls_async_phb2exp_query       */
/**
 * ABSTRACT: Sends message to MPLS CC to get a mapping from phb2exp table
 *
 * DESCRIPTION: This message helper function reads a PHB2EXP table entry. 
 *
 * \param arg_Index        - index identifying an EXP2PHB table entry
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_phb2exp_query(ix_uint16                    arg_Index,
                               ix_cc_mpls_cb_phb2exp_query  arg_Callback,
                               void                        *arg_pUserContext);


/*
 *   MPLS PARAM QUERY
 *************************************************************************************************/



/*
 * FUNCTION: ix_cc_mpls_param_query()  */
/**
 * ABSTRACT: This function reads MPLS parameters
 *
 * DESCRIPTION: This function reads MPLS parameters
 *
 * \param  arg_pMplstParameters - pointer to the buffer storing the MPLS 
 *                                parameters returned by the function.  
 * \param  arg_pContext         - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL            - invalid pointer
 */
ix_error
ix_cc_mpls_param_query (ix_cc_mpls_params  *arg_pMplsParameters,
                        void               *arg_pContext);



/*
 * TYPENAME: ix_cc_mpls_cb_param_query          */
/**
 * ABSTRACT: Client provided callback function.
 *
 * DESCRIPTION: Client provided callback function.
 *              The core component packet handler invokes this function to return
 *              an MPLS parameters.
 *
 * \param  arg_Result          - indicates error conditions for the call
 * \param  arg_pContext        - pointer to client-defined context, provided as
 *                               a parameter in the API call. Used by the caller 
 *                               to match asynchronous request with a response.
 * \param  arg_pMplsParameters - pointer to the buffer storing the MPLS parameters 
 *                               returned by the function
 */
typedef ix_error (*ix_cc_mpls_cb_param_query) (ix_error           arg_Result,
                                               void              *arg_pContext,
                                               ix_cc_mpls_params *arg_pMplsParameters);


/*
 * FUNCTION: ix_cc_mpls_async_param_query       */
/**
 * ABSTRACT: Sends message to MPLS CC to get MPLS CC parameters
 *
 * DESCRIPTION: This message helper function reads MPLS parameters.
 *
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_param_query(ix_cc_mpls_cb_param_query  arg_Callback,
                             void                      *arg_pUserContext);




/*
 *   MPLS PORT STATS QUERY
 *************************************************************************************************/



/*
 * FUNCTION: ix_cc_mpls_port_stats_query ()  */
/**
 * ABSTRACT: This function reads MPLS statistics for a given input port.
 *
 * DESCRIPTION: This function reads MPLS statistics for a given input port.
 *
 * \param  arg_PortIndex           - local port index.
 * \param  arg_pMplsPortStatistics - pointer to the buffer storing the MPLS 
 *                                   port statistics returned by the function.
 * \param  arg_pContext            - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL            - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND - arg_PortIndex is out of range
 */
ix_error
ix_cc_mpls_port_stats_query (ix_uint16                     arg_PortIndex,
                             ix_cc_mpls_input_port_stats  *arg_pMplsPortStatistics,
                             void                         *arg_pContext);



/*
 * TYPENAME: ix_cc_mpls_cb_port_stats_query          */
/**
 * ABSTRACT: Client provided callback function.
 *
 * DESCRIPTION: Client provided callback function.
 *              The core component packet handler invokes this function to return
 *              an MPLS port statistics.
 *
 * \param  arg_Result     - indicates error conditions for the call
 * \param  arg_pContext   - pointer to client-defined context, provided as
 *                          a parameter in the API call. Used by the caller 
 *                          to match asynchronous request with a response.
 * \param  arg_pMplsStats - pointer to the buffer storing the MPLS port 
 *                          statistics returned by the function
 */
typedef ix_error (*ix_cc_mpls_cb_port_stats_query) (ix_error                     arg_Result,
                                                    void                        *arg_pContext,
                                                    ix_cc_mpls_input_port_stats *arg_pMplsStats);



/*
 * FUNCTION: ix_cc_mpls_async_port_stats_query       */
/**
 * ABSTRACT: Sends message to MPLS CC to get MPLS statistics from the given port
 *
 * DESCRIPTION: This message helper function reads MPLS statistics for a given input port. 
 *
 * \param arg_PortIndex    - local port index
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_port_stats_query(ix_uint16                        arg_PortIndex,
                                  ix_cc_mpls_cb_port_stats_query   arg_Callback,
                                  void                            *arg_pUserContext);


/*
 *   MPLS GET MPLS CONTEXT
 *************************************************************************************************/


/*
 * TYPENAME: ix_cc_mpls_cb_get_mpls_context          */
/**
 * ABSTRACT: Client provided callback function to get mpls context
 *
 * DESCRIPTION: Client provided callback function.
 *              The core component packet handler invokes this function to return
 *              an MPLS internal context.
 *
 * \param  arg_Result       - indicates error conditions for the call
 * \param  arg_pContext     - pointer to client-defined context, provided as
 *                            a parameter in the API call. Used by the caller 
 *                            to match asynchronous request with a response.
 * \param  arg_pMplsContext - pointer to the buffer storing the MPLS context
 */
typedef ix_error (*ix_cc_mpls_cb_get_mpls_context) (ix_error               arg_Result,
                                                    void                  *arg_pContext,
                                                    void                  *arg_pMplsContext);



/*
 * FUNCTION: ix_cc_mpls_async_get_mpls_context       */
/**
 * ABSTRACT: This function reads MPLS internal context structure 
 *
 * DESCRIPTION: This function reads MPLS internal context structure 
 *              It is used during Unit Tests.
 *
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_get_mpls_context(ix_cc_mpls_cb_get_mpls_context   arg_Callback,
                                  void                            *arg_pUserContext);







#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(IX_CC_MPLS_H) */








