/*
 * ============================================================================
 * COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *                                                                          */
/** \file      ix_cc_mpls_lsp.c
 *
 *  \brief     The file contains CCI API functions (init and fini) exposed by
 *             MPLS Core Component
 *
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_mpls.c,v 1.12.2.1 2004/03/12 10:40:14 wneubaue Exp $";


#include <ix_error.h>
#include <ix_types.h>

/* include IXA SDK framework */
#include <ix_ossl.h>
#include "ix_rm.h"

#if !defined(IX_EXCLUDE_CCI)
#include "ix_cci.h"
#endif 

/* include CC support */
#include "ix_cc_error.h"
#include "ix_cc.h"

#include "bindings.h"

#include "cc/ix_cc_rtmv4.h" 
#include "cc/ix_cc_ipv4.h"
#include <cc/ix_cc_reg_util.h>


#include "../../include/mpls_defs.h"
#include "cc/ix_cc_mpls.h"
#include "cc/internal/ix_cc_mpls_msg.h"
#include "cc/internal/ix_cc_mpls_icmp.h"
#include "cc/internal/ix_cc_mpls_lsp.h"
#include "cc/internal/ix_cc_mpls_nhlfe.h"
#include "cc/internal/ix_cc_mpls_pkt_handler.h"
#include "cc/internal/ix_cc_mpls_internal.h"





/**
 * Private functions declarations
 */
PRIVATE ix_error
_ix_cc_mpls_register_msg_handler(ix_cc_handle        arg_hCcHandle,
                                 ix_cc_mpls_context *arg_pContext);
PRIVATE ix_error
_ix_cc_mpls_unregister_msg_handler();

PRIVATE ix_error 
_ix_cc_mpls_register_packet_handlers(ix_cc_mpls_context *arg_pContext);

PRIVATE ix_error 
_ix_cc_mpls_unregister_packet_handlers();

PRIVATE ix_error 
_ix_cc_mpls_add_icmp_event_handlers(ix_cc_mpls_context *arg_pContext);

PRIVATE ix_error
_ix_cc_mpls_delete_icmp_event_handlers (ix_cc_mpls_context* arg_pContext);



/*
 * FUNCTION: ix_cc_mpls_init()  */
/**
 * ABSTRACT: function initializes the MPLS CC
 *
 * DESCRIPTION: This function initializes the MPLS core component.
 *              It should be called and returned successfully before any other 
 *              function in the core component can be called. It initializes 
 *              the ILM and NHLFE tables, and MPLS parameters. Then it performs 
 *              static configuration of the ILM and FTN Forwarder microblocks,
 *              allocating and patching required variables and symbols into 
 *              the microblocks by calling Resource Manager's API function,
 *              ix_rm_ueng_patch_symbols().  Finally, it registers the message 
 *              and packet handlers for its defined message inputs and packet input.
 *
 * \param  arg_CcHandle  - handle to the core component, created by the core component 
 *                         infrastructure; this shall be used later to get other services
 *                         from the core component infrastructure.
 * \param  arg_ppContext - handle of the buffer free list and dynamic property data,
 *                         given by the system application during initialization.
 *                         Upon successful completion, it represents the location where 
 *                         the pointer to the control block allocated by MPLS forwarder 
 *                         core component will be stored. The control block is internal
 *                         to the MPLS core component and  contains information about 
 *                         MPLS internal data structures, allocated memory and other 
 *                         relevant information. Later this pointer will be passed to 
 *                         the ix_cc_mpls_fini() function for freeing memory and other 
 *                         house-keeping operations when the MPLS core component is being 
 *                         destroyed
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL - invalid pointer
 * \n         IX_CC_ERROR _OOM   - memory allocation failure      
 * \n         IX_CC_ERROR _FAILED_PATCHING   - patching failures                      
 * \n         IX_CC_ERROR _64BIT_COUNTER_- 64 bit counter creation failure
 * \n         IX_CC_ERROR_CCI - failure from CCI
 */
ix_error
ix_cc_mpls_init(ix_cc_handle   arg_hCcHandle,
                void         **arg_ppContext)

{
    ix_error err      = IX_SUCCESS;
    ix_error err_exit = IX_SUCCESS;
    ix_cc_mpls_context* pContext = NULL;
    ix_cc_init_context* pInitContext = NULL;

    /* Check the input arguments */
    if (arg_hCcHandle == IX_NULL_HANDLE) 
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE,
                                ("ix_cc_mpls_init(): got invalid CC handle"));

    if (*arg_ppContext == (void *)NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                    ("ix_cc_mpls_init(): got null pointer to context pointer store"));

    if (arg_ppContext == (void **)NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                    ("ix_cc_mpls_init(): got null pointer to init context "));

    /* Get memory for context structure */ 
    pContext = (ix_cc_mpls_context*)ix_ossl_malloc(sizeof(ix_cc_mpls_context));
    if (NULL == pContext)
        return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
               ("ix_cc_mpls_init(): Out of system memory in ix_cc_mpls_init"));

    /* Initialize the structure */
    ix_ossl_memset((void *)pContext, 0,
                   (ix_ossl_size_t)sizeof(ix_cc_mpls_context));

    /* store mpls component handle in ix_cc_mpls_context */
    pContext->hMplsHandle = arg_hCcHandle;


    /* Create mutex for critical section. */
    IX_ERROR_C(ix_ossl_mutex_init(OS_MUTEX_UNLOCK, &(pContext->mutexId)),err);
    if (err != IX_SUCCESS)
    {
        err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_OSSL_MUTEX_INIT,
                               ("ix_cc_mpls_init(): can't initialize LSP mutex"));
        goto err_init;
    }

    IX_ERROR_C(ix_ossl_mutex_lock(pContext->mutexId,
               IX_OSSL_WAIT_FOREVER), err);
    if (err != IX_SUCCESS)
    {
        err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK,
                               ("ix_cc_mpls_init(): can't lock mutex"));
        goto err_mutex_init;
    }

    /* Retrieve static configuration data */
    IX_ERROR_C(_ix_cc_mpls_get_static_data(pContext), err);
    if (err != IX_SUCCESS) goto err_nhlfe_init;


    /* Initialize the NHLFE library */
    IX_ERROR_C(ix_cc_mpls_nhlfe_init(pContext), err);
    if (err != IX_SUCCESS) goto err_nhlfe_init;


    /* Initialize the LSP library */
    IX_ERROR_C(ix_cc_mpls_lsp_init(pContext), err);
    if (err != IX_SUCCESS) goto err_lsp_init;


    /* Patch various symbols into MPLS microblock */
    IX_ERROR_C(_ix_cc_mpls_patch_symbols(pContext),err);
    if (err != IX_SUCCESS)
    {
        err = IX_ERROR_WARNING(err, ("ix_cc_mpls_init(): Error ! Could not patch symbols."));
        goto err_patch;
    }

    /* Register message handler */
    IX_ERROR_C(_ix_cc_mpls_register_msg_handler(arg_hCcHandle, pContext), err);
    if (err != IX_SUCCESS) 
        goto err_patch;

    /* Register packet handler */
    IX_ERROR_C(_ix_cc_mpls_register_packet_handlers(pContext), err);
    if (err != IX_SUCCESS) 
        goto err_pkt_handler;
    
    _ix_cc_mpls_icmp_init (pContext);

    /* Register event handlers */
    err = _ix_cc_mpls_add_icmp_event_handlers(pContext);
    if (err != IX_SUCCESS) 
        goto err_icmp_handler;


    /* Obtain blade ID */
    pInitContext = (ix_cc_init_context*)*arg_ppContext;
    if ((pInitContext != NULL) &&
        (pInitContext->pPropertyList != NULL) &&
        (pInitContext->pPropertyList->pPropertyInfo != NULL))
    {
        pContext->localBladeId = pInitContext->pPropertyList->pPropertyInfo->blade_id;
    }

    #if IX_MPLS_DEBUG    
    ix_ossl_message_log("ix_cc_mpls_init(): Blade Id = %p \n", pContext->localBladeId);
    ix_ossl_message_log("ix_cc_mpls_init(): pContext = 0x%x \n", (int) pContext);
    #endif
    
    /* Assign output argument */
    *arg_ppContext = (void *)pContext;

    /* unlock mutex */
    IX_ERROR_C(ix_ossl_mutex_unlock(pContext->mutexId), err);
    if (err != IX_SUCCESS) goto err_unlock_mutex;

    return IX_SUCCESS;

err_unlock_mutex:
    err_exit = _ix_cc_mpls_delete_icmp_event_handlers (pContext);
    if (err_exit != IX_SUCCESS)
          ix_ossl_message_log("ix_cc_mpls_init(): Warning: can't unregister ICMP event handler.\n");


err_icmp_handler:
    _ix_cc_mpls_icmp_fini (pContext);

    err_exit = _ix_cc_mpls_unregister_packet_handlers(arg_hCcHandle, pContext);
    if (err_exit != IX_SUCCESS)
          ix_ossl_message_log("ix_cc_mpls_init(): Warning: can't unregister packet handler.\n");


err_pkt_handler:
    err_exit = _ix_cc_mpls_unregister_msg_handler(arg_hCcHandle, pContext);
    if (err_exit != IX_SUCCESS)
          ix_ossl_message_log("ix_cc_mpls_init(): Warning: can't unregister message handler.\n");

err_patch:
    _ix_cc_mpls_lsp_fini(pContext);


err_lsp_init:
    _ix_cc_mpls_nhlfe_fini(pContext);


err_nhlfe_init:
    err_exit = ix_ossl_mutex_unlock(pContext->mutexId);
    if (err_exit != IX_SUCCESS)
          ix_ossl_message_log("ix_cc_mpls_init(): Warning: can't unlock mutex.\n");


err_mutex_init:
    /* Remove mutex for critical section. */
    err_exit = ix_ossl_mutex_fini(pContext->mutexId);
    if (err_exit != IX_SUCCESS)
          ix_ossl_message_log("ix_cc_mpls_init(): Warning: can't remove mutex.\n");

err_init:
    ix_ossl_free(pContext);

    *arg_ppContext = NULL;

    return err;

} /* ix_cc_mpls_init */





/*
 * FUNCTION: ix_cc_mpls_fini()  */
/**
 * ABSTRACT: The function terminates the MPLS core component.
 *
 * DESCRIPTION: This function terminates the MPLS core component.
 *              It will be executed when the execution engine running 
 *              the core component is being shut down.  This function will free 
 *              all allocated memory and resources allocated by the ix_cc_mpls_init()
 *              function, as well as other functions of the core component.  
 *
 * \param  arg_CcHandle - handle to the core component.
 * \param  arg_pContext - pointer to the control block memory allocated earlier in
 *                        ix_cc_mpls_init(). The termination routine will use it to 
 *                        de-allocate the control block memory.
 * 
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL - invalid pointer
 * \n         IX_CC_ERROR_OOM - memory free failure
 * \n         IX_CC_ERROR_CCI - failure from CCI
 * \n         IX_CC_ERROR_OOM_64BIT_COUNTER - 64 bit counter deletion failure
 */
ix_error
ix_cc_mpls_fini(ix_cc_handle  arg_hCcHandle,
                void         *arg_pContext)
{
    ix_error            err;
    ix_cc_mpls_context *pContext = arg_pContext;

    /* Check the input arguments */
    if (arg_hCcHandle == IX_NULL_HANDLE) 
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE,
                         ("ix_cc_mpls_fini(): got invalid CC handle"));

    /* Check whether arg_pContext is valid */
    if (arg_pContext == (void *)NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                          ("ix_cc_mpls_fini(): got null CC context"));

    IX_ERROR_C(ix_ossl_mutex_lock(pContext->mutexId,
               IX_OSSL_WAIT_FOREVER), err);
    if (err != IX_SUCCESS)
        ix_ossl_message_log("ix_cc_mpls_fini(): Warning: can't lock mutex.\n");

    err = _ix_cc_mpls_delete_icmp_event_handlers (arg_pContext);
    if (err != IX_SUCCESS)
          ix_ossl_message_log("ix_cc_mpls_fini(): Warning: can't unregister ICMP event handler.\n");
    
    _ix_cc_mpls_icmp_fini(arg_pContext);

    err = _ix_cc_mpls_unregister_packet_handlers();
    if (err != IX_SUCCESS)
          ix_ossl_message_log("ix_cc_mpls_fini(): Warning: can't unregister packet handler.\n");

    err = _ix_cc_mpls_unregister_msg_handler();
    if (err != IX_SUCCESS)
          ix_ossl_message_log("ix_cc_mpls_fini(): Warning: can't unregister message handler.\n");
    
    _ix_cc_mpls_lsp_fini(arg_pContext);

    /* Remove the NHLFE table. */
    _ix_cc_mpls_nhlfe_fini(arg_pContext);
    
    /* Exit the critical section */
    ix_ossl_mutex_unlock(pContext->mutexId);

    /* Remove mutex for critical section. */
    err = ix_ossl_mutex_fini(pContext->mutexId);
    if (err != IX_SUCCESS)
          ix_ossl_message_log("ix_cc_mpls_fini(): Warning: can't remove mutex.\n");

    /* Free the control block structure itself */
    ix_ossl_free((void *)pContext);

    return (IX_SUCCESS);
} /* ix_cc_mpls_fini */





/*
 * FUNCTION: ix_cc_mpls_param_query()  */
/**
 * ABSTRACT: This function reads MPLS parameters
 *
 * DESCRIPTION: This function reads MPLS parameters
 *
 * \param  arg_pMplstParameters - pointer to the buffer storing the MPLS 
 *                                parameters returned by the function.  
 * \param  arg_pContext         - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL            - invalid pointer
 */
ix_error
ix_cc_mpls_param_query (ix_cc_mpls_params *arg_pMplsParameters,
                        void              *arg_pContext)
{
    ix_uint32 idx;
    ix_cc_mpls_context *pContext = arg_pContext;

    /* Check whether arg_pContext is valid */
    if (arg_pContext == (void *)NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
           ("ix_cc_mpls_param_query (): got null MPLS CC context"));

    /* Check whether arg_pContext is valid */
    if (arg_pMplsParameters == (void *)NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
             ("ix_cc_mpls_param_query (): got null pointer to MPLS parameters structure"));

    arg_pMplsParameters->labelSpaceMode = pContext->labelSpaceMode;
    arg_pMplsParameters->maxLocalPorts  = pContext->maxLocalPorts;

    for (idx = 0; idx < pContext->maxLocalPorts; idx++) {
        if (pContext->labelSpaceMode == IX_CC_MPLS_PER_INTERFACE_LABEL_SPACE)
        {
            arg_pMplsParameters->labelRange[idx].min_label.label = (pContext->pLabelSpaceTable)[idx].min_label;
            arg_pMplsParameters->labelRange[idx].max_label.label = (pContext->pLabelSpaceTable)[idx].max_label;
        } else {
            arg_pMplsParameters->labelRange[idx].min_label.label = (pContext->pLabelSpaceTable)[0].min_label;
            arg_pMplsParameters->labelRange[idx].max_label.label = (pContext->pLabelSpaceTable)[0].max_label;
        }
    }

    ix_ossl_memset(&(arg_pMplsParameters->preconfExp2Phb), 0, sizeof(ix_cc_mpls_lm_exp2phb));
    ix_ossl_memset(&(arg_pMplsParameters->preconfPhb2Exp2),0, MPLS_MAX_PHB_NUM * 
                                                              sizeof(ix_cc_mpls_lm_exp2phb));

    return IX_SUCCESS;
} /* ix_cc_mpls_param_query */






/*
 * FUNCTION: ix_cc_mpls_port_stats_query ()  */
/**
 * ABSTRACT: This function reads MPLS statistics for a given input port.
 *
 * DESCRIPTION: This function reads MPLS statistics for a given input port.
 *
 * \param  arg_PortIndex           - local port index.
 * \param  arg_pMplsPortStatistics - pointer to the buffer storing the MPLS 
 *                                   port statistics returned by the function.
 * \param  arg_pContext            - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL            - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND - arg_PortIndex is out of range
 */
ix_error
ix_cc_mpls_port_stats_query (ix_uint16                     arg_PortIndex,
                             ix_cc_mpls_input_port_stats  *arg_pMplsPortStatistics,
                             void                         *arg_pContext)
{
    /* Check whether arg_pContext is valid */
    if (arg_pContext == (void *)NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
             ("ix_cc_mpls_port_stats_query(): got null MPLS CC context"));

    /* Check whether arg_pContext is valid */
    if (arg_pMplsPortStatistics == (void *)NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
             ("ix_cc_mpls_port_stats_query(): got null pointer to a port statistics structure"));

    return IX_SUCCESS;
} /* ix_cc_mpls_port_stats_query */




/*
 * FUNCTION: _ix_cc_mpls_prop_get_uint_value   */
/**
 * ABSTRACT: Get the value of a specific subproperty
 *
 * DESCRIPTION: Get's the value of a specific subproperty.
 *              The type is validated and an error is passed if it is not
 *              uint32.
 *
 * \param  arg_parent Handle to the parent property.
 * \param  arg_name   Name of the sub-property
 * \param  arg_value  Output parameter for the value.
 *
 * \return IX_SUCCESS upon success or an ix_error with one of the following values.
 * \retval IX_CC_ERROR_ENTRY_NOT_FOUND The property is not the correct type (uint32)
 */
ix_error
_ix_cc_mpls_prop_get_uint_value(ix_configuration_property_handle  arg_parent,
                                const char                       *arg_name,
                                ix_uint32                        *arg_value)
{
    ix_configuration_property_handle prop;
    ix_cp_property_info prop_info;

    /* Validate arguments */
    IX_ERROR_CHECK_ARG_NULL(1, arg_parent);
    IX_ERROR_CHECK_ARG_NULL(2, arg_name);
    IX_ERROR_CHECK_ARG_NULL(3, arg_value);

    IX_ERROR_CR(ix_rm_cp_property_open(arg_parent, 
                                       arg_name,
                                       &prop));
    IX_ERROR_CR(ix_rm_cp_property_get_info(prop, &prop_info));
        
    /* if the value is the correct type, uint32, then get the value
       otherwise return an error */
    if(prop_info.m_Options & IX_CP_OPTIONS_FLAGS_DATA_UINT32){
        IX_ERROR_CR(ix_rm_cp_property_get_value_uint32(prop,
                                                       arg_value));
        
    } else {
        return IX_ERROR_LOCAL(IX_CC_ERROR_ENTRY_NOT_FOUND,
                              ("_ix_cc_mpls_prop_get_uint_value(): Invalid registry property"
                               " type.  MPLS PORT subproperties must be uint32."));
    }   
    IX_ERROR_CR(ix_rm_cp_property_close(prop));
    return IX_SUCCESS;
} /* _ix_cc_mpls_prop_get_uint_value */






/*
 * FUNCTION: _ix_cc_mpls_get_static_data ()  */
/**
 * ABSTRACT: This function retrieves all MPLS static data from system registry
 *
 * DESCRIPTION: This function retrieves all MPLS static data from 
 *              system registry (uses defaults if registry is absent) 
 *              and store them into the corresponding field in the context 
 *              structure passed in.
 *
 * \param  arg_pContext  - MPLS CC context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_get_static_data(void *arg_pContext)
{
    ix_error  err;
    ix_cc_mpls_context *pContext = (ix_cc_mpls_context*) arg_pContext;

#ifdef IX_INCLUDE_REGISTRY
    ix_uint32 RegData;
    ix_configuration_property_handle hProperty;

	#if IX_MPLS_DEBUG    
    ix_ossl_message_log("_ix_cc_mpls_get_static_data: IX_INCLUDE_REGISTRY...\n");
	#endif


    IX_ERROR_CR(ix_cc_reg_get_prop_from_path(IX_MPLS_PROP_ROOT, 
                                             &hProperty));

    /* get the property IX_MPLS_PROP_MEV2_NUM_MASK */
    IX_ERROR_CRT(_ix_cc_mpls_prop_get_uint_value(hProperty,
                                                 IX_MPLS_PROP_MEV2_NUM_MASK,
                                                 &(pContext->mev2NumMask)),
                 IX_CC_ERROR_ENTRY_NOT_FOUND,
                 IX_ERROR_LEVEL_REMOTE);

    
    /* get the property IX_MPLS_PROP_LABEL_SPACE_CONFIG */
    IX_ERROR_CRT(_ix_cc_mpls_prop_get_uint_value(hProperty,
                                                 IX_MPLS_PROP_LABEL_SPACE_CONFIG,
                                                 &(pContext->labelSpaceMode)),
                 IX_CC_ERROR_ENTRY_NOT_FOUND,
                 IX_ERROR_LEVEL_REMOTE);

    
    /* get the property IX_MPLS_PROP_MAX_LOCAL_PORTS */
    IX_ERROR_CRT(_ix_cc_mpls_prop_get_uint_value(hProperty,
                                                 IX_MPLS_PROP_MAX_LOCAL_PORTS,
                                                 &(pContext->maxLocalPorts)),
                 IX_CC_ERROR_ENTRY_NOT_FOUND,
                 IX_ERROR_LEVEL_REMOTE);

    /* get the property IX_MPLS_PROP_MAX_GLOBAL_PORTS */
    IX_ERROR_CRT(_ix_cc_mpls_prop_get_uint_value(hProperty,
                                                 IX_MPLS_PROP_MAX_GLOBAL_PORTS,
                                                 &(pContext->maxGlobalPorts)),
                 IX_CC_ERROR_ENTRY_NOT_FOUND,
                 IX_ERROR_LEVEL_REMOTE);

    /* get the property IX_MPLS_PROP_ICMPQ_MAXSIZE */
    IX_ERROR_CRT(_ix_cc_mpls_prop_get_uint_value(hProperty,
                                                 IX_MPLS_PROP_ICMPQ_MAXSIZE,
                                                 &RegData),
                 IX_CC_ERROR_ENTRY_NOT_FOUND,
                 IX_ERROR_LEVEL_REMOTE);
    
    pContext->icmpQueueMaxSize = (ix_uint16)RegData;

    /* get the property IX_MPLS_PROP_ICMP_DEFAULT_SLEEP_TIME */
    IX_ERROR_CRT(_ix_cc_mpls_prop_get_uint_value(hProperty,
                                                 IX_MPLS_PROP_ICMP_DEFAULT_SLEEP_TIME,
                                                 &RegData),
                 IX_CC_ERROR_ENTRY_NOT_FOUND,
                 IX_ERROR_LEVEL_REMOTE);
    
    pContext->icmpSleepTime = (ix_uint16)RegData;


#else /* IX_INCLUDE_REGISTRY */

    pContext->mev2NumMask        = MPLS_MEV2_NUM_MASK;
    pContext->maxLocalPorts      = MPLS_MAX_LOCAL_PORTS;
    pContext->maxGlobalPorts     = MPLS_MAX_GLOBAL_PORTS;
    pContext->icmpQueueMaxSize   = MPLS_ICMP_QUEUE_MAXSIZE;
    pContext->icmpSleepTime      = MPLS_ICMP_QUEUE_SLEEPTIME;
    
#endif /* IX_INCLUDE_REGISTRY */
    
    err = _ix_cc_mpls_nhlfe_get_static_data(arg_pContext);
    if (err != IX_SUCCESS) 
        return err;

    err = _ix_cc_mpls_lsp_get_static_data(arg_pContext);
    if (err != IX_SUCCESS) 
        return err;

    return (IX_SUCCESS);
} /* _ix_cc_mpls_get_static_data */





/*
 * FUNCTION: _ix_cc_mpls_patch_symbols()  */
/**
 * ABSTRACT: This function frees all memory and resources taken by MPLS CC
 *
 * DESCRIPTION: This function frees all memory and resources taken by MPLS CC
 *              Finally, it frees the CC context structure.
 *
 * \param  arg_pContext  - MPLS CC context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_patch_symbols(void *arg_pContext)
{
    ix_error err = IX_SUCCESS;
    
    err = _ix_cc_mpls_nhlfe_patch_symbols(arg_pContext);
    if (err != IX_SUCCESS) 
	return err;

    err = _ix_cc_mpls_lsp_patch_symbols(arg_pContext);
    if (err != IX_SUCCESS) 
	return err;

    return err;
} /* _ix_cc_mpls_patch_symbols */





/*
 * FUNCTION: _ix_cc_mpls_register_msg_handler()  */
/**
 * ABSTRACT: This function registers message handler fo MPLS core component
 *
 * \param  arg_hCcHandle - handle to the MPLS core component
 * \param  arg_pContext  - MPLS CC context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
PRIVATE ix_error
_ix_cc_mpls_register_msg_handler(ix_cc_handle        arg_hCcHandle,
                                 ix_cc_mpls_context *arg_pContext)
{

#ifndef IX_EXCLUDE_CCI

    /* make compiler happy */
    arg_pContext = arg_pContext;

    IX_ERROR_CRT(ix_cci_cc_add_message_handler(arg_hCcHandle,
                                               IX_CC_MPLS_MSG_INPUT,
                                               ix_cc_mpls_msg_handler,
                                               IX_INPUT_TYPE_SINGLE_SRC),
                 IX_CC_MPLS_ERROR_CCI_REG_HDLR, 
                 IX_ERROR_LEVEL_REMOTE);
    
#else  /* IX_EXCLUDE_CCI */
    
    IX_ERROR_CRT(ix_rm_message_handler_register(IX_CC_MPLS_MSG_INPUT,
                                                ix_cc_mpls_msg_handler,
                                                arg_pContext),
                 IX_CC_MPLS_ERROR_RM_REG_HDLR, 
                 IX_ERROR_LEVEL_REMOTE);

#endif /* IX_EXCLUDE_CCI */

    return (IX_SUCCESS);
} /* _ix_cc_mpls_register_msg_handler */




/*
 * FUNCTION: _ix_cc_mpls_unregister_msg_handler()  */
/**
 * ABSTRACT: This function unregisters message handler fo MPLS core component
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
PRIVATE ix_error
_ix_cc_mpls_unregister_msg_handler()
{

    ix_error err;

#ifndef IX_EXCLUDE_CCI

    /* Remove message handler from CCI */
    IX_ERROR_CT(ix_cci_cc_remove_message_handler(IX_CC_MPLS_MSG_INPUT),
                err,
                IX_CC_MPLS_ERROR_CCI_REM_HDLR, 
                IX_ERROR_LEVEL_REMOTE);
    if (err != IX_SUCCESS)
    {
        ix_ossl_message_log("Warning: CCI: message handler cannot be removed\n");
    }


#else /* IX_EXCLUDE_CCI */

    /* Remove message handler from RM */
    IX_ERROR_CT(ix_rm_message_handler_unregister(IX_CC_MPLS_MSG_INPUT),
                err,
                IX_CC_MPLS_ERROR_RM_REM_HDLR, 
                IX_ERROR_LEVEL_REMOTE);
    if (err != IX_SUCCESS)
    {
        ix_ossl_message_log("Warning: RM: message handler cannot be removed\n");
    }

#endif /* IX_EXCLUDE_CCI */

    return err;
} /* _ix_cc_mpls_unregister_msg_handler */



/*
 * FUNCTION: _ix_cc_mpls_register_packet_handlers()  */
/**
 * ABSTRACT: This function registers packet handler fo MPLS core component
 *
 * DESCRIPTION: This primitive will be used to register packet handlers. 
 *              It registers with CCI in level 1 system and register
 *              with RM in level 0 system.
 *
 * \param  arg_pContext  - MPLS CC context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
PRIVATE ix_error 
_ix_cc_mpls_register_packet_handlers(ix_cc_mpls_context* arg_pContext)
{

/* Call CCI packet handler if CCI is enabled */
#if defined(IX_EXCLUDE_CCI)

    IX_ERROR_CRT(ix_rm_packet_handler_register(IX_CC_MPLS_ILM_PKT_MICROBLOCK_INPUT,
                                               ix_cc_mpls_microblock_pkt_handler,
                                               arg_pContext),
                 IX_CC_MPLS_ERROR_RM_REG_HDLR,
                 IX_ERROR_LEVEL_REMOTE);

    IX_ERROR_CRT(ix_rm_packet_handler_register(IX_CC_MPLS_FTN_PKT_MICROBLOCK_INPUT,
                                               ix_cc_mpls_microblock_pkt_handler,
                                               arg_pContext),
                 IX_CC_MPLS_ERROR_RM_REG_HDLR,
                 IX_ERROR_LEVEL_REMOTE);

    IX_ERROR_CRT(ix_rm_packet_handler_register(IX_CC_MPLS_ILM_INPUT,
                                               ix_cc_mpls_ilm_pkt_handler,
                                               arg_pContext),
                 IX_CC_MPLS_ERROR_RM_REG_HDLR,
                 IX_ERROR_LEVEL_REMOTE);

    IX_ERROR_CRT(ix_rm_packet_handler_register(IX_CC_MPLS_FTN_INPUT,
                                               ix_cc_mpls_ftn_pkt_handler,
                                               arg_pContext),
                 IX_CC_MPLS_ERROR_RM_REG_HDLR,
                 IX_ERROR_LEVEL_REMOTE);

#else
    
    IX_ERROR_CRT(ix_cci_cc_add_packet_handler(arg_pContext->hMplsHandle,
                                              IX_CC_MPLS_ILM_PKT_MICROBLOCK_INPUT,
                                              ix_cc_mpls_microblock_pkt_handler,
                                              IX_INPUT_TYPE_SINGLE_SRC),
                 IX_CC_MPLS_ERROR_CCI_REG_HDLR,
                 IX_ERROR_LEVEL_REMOTE);

    IX_ERROR_CRT(ix_cci_cc_add_packet_handler(arg_pContext->hMplsHandle,
                                              IX_CC_MPLS_FTN_PKT_MICROBLOCK_INPUT,
                                              ix_cc_mpls_microblock_pkt_handler,
                                              IX_INPUT_TYPE_SINGLE_SRC),
                 IX_CC_MPLS_ERROR_CCI_REG_HDLR,
                 IX_ERROR_LEVEL_REMOTE);

    IX_ERROR_CRT(ix_cci_cc_add_packet_handler(arg_pContext->hMplsHandle,
                                              IX_CC_MPLS_ILM_INPUT,
                                              ix_cc_mpls_ilm_pkt_handler,
                                              IX_INPUT_TYPE_SINGLE_SRC),
                 IX_CC_MPLS_ERROR_CCI_REG_HDLR,
                 IX_ERROR_LEVEL_REMOTE);

    IX_ERROR_CRT(ix_cci_cc_add_packet_handler(arg_pContext->hMplsHandle,
                                              IX_CC_MPLS_FTN_INPUT,
                                              ix_cc_mpls_ftn_pkt_handler,
                                              IX_INPUT_TYPE_SINGLE_SRC),
                 IX_CC_MPLS_ERROR_CCI_REG_HDLR,
                 IX_ERROR_LEVEL_REMOTE);


#endif

    return IX_SUCCESS;

}/* _ix_cc_mpls_register_packet_handlers */





/*
 * FUNCTION: _ix_cc_mpls_unregister_packet_handlers()  */
/**
 * ABSTRACT: This function unregisters packet handler fo MPLS core component
 *
 * DESCRIPTION: This primitive will be used to unregister packet handlers. 
 *              It unregisters with CCI in level 1 system and register
 *              with RM in level 0 system.
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
PRIVATE ix_error 
_ix_cc_mpls_unregister_packet_handlers()
{
    ix_error err;

#ifndef IX_EXCLUDE_CCI

    /* Remove packet handler from CCI */
    IX_ERROR_CT(ix_cci_cc_remove_packet_handler(IX_CC_MPLS_FTN_PKT_MICROBLOCK_INPUT),
                err,
                IX_CC_MPLS_ERROR_CCI_REM_HDLR,
                IX_ERROR_LEVEL_REMOTE);
    if (err != IX_SUCCESS)
    {
        ix_ossl_message_log("Warning: CCI: packet handler cannot be removed\n");
    }

        /* Remove packet handler from CCI */
    IX_ERROR_CT(ix_cci_cc_remove_packet_handler(IX_CC_MPLS_ILM_PKT_MICROBLOCK_INPUT),
                err,
                IX_CC_MPLS_ERROR_CCI_REM_HDLR,
                IX_ERROR_LEVEL_REMOTE);
    if (err != IX_SUCCESS)
    {
        ix_ossl_message_log("Warning: CCI: packet handler cannot be removed\n");
    }

    /* Remove packet handler from CCI */
    IX_ERROR_CT(ix_cci_cc_remove_packet_handler(IX_CC_MPLS_FTN_INPUT),
                err,
                IX_CC_MPLS_ERROR_CCI_REM_HDLR,
                IX_ERROR_LEVEL_REMOTE);
    if (err != IX_SUCCESS)
    {
        ix_ossl_message_log("Warning: CCI: packet handler cannot be removed\n");
    }


    /* Remove packet handler from CCI */
    IX_ERROR_CT(ix_cci_cc_remove_packet_handler(IX_CC_MPLS_ILM_INPUT),
                err,
                IX_CC_MPLS_ERROR_CCI_REM_HDLR,
                IX_ERROR_LEVEL_REMOTE);
    if (err != IX_SUCCESS)
    {
        ix_ossl_message_log("Warning: CCI: packet handler cannot be removed\n");
    }


#else /* IX_EXCLUDE_CCI */

    /* Remove message handler from RM */
    IX_ERROR_CT(ix_rm_packet_handler_unregister(IX_CC_MPLS_FTN_PKT_MICROBLOCK_INPUT),
                err,
                IX_CC_MPLS_ERROR_RM_REM_HDLR,
                IX_ERROR_LEVEL_REMOTE);
    if (err != IX_SUCCESS)
    {
        ix_ossl_message_log("Warning: RM: packet handler cannot be removed\n");
    }

    /* Remove message handler from RM */
    IX_ERROR_CT(ix_rm_packet_handler_unregister(IX_CC_MPLS_ILM_PKT_MICROBLOCK_INPUT),
                err,
                IX_CC_MPLS_ERROR_RM_REM_HDLR,
                IX_ERROR_LEVEL_REMOTE);
    if (err != IX_SUCCESS)
    {
        ix_ossl_message_log("Warning: RM: packet handler cannot be removed\n");
    }

    /* Remove message handler from RM */
    IX_ERROR_CT(ix_rm_packet_handler_unregister(IX_CC_MPLS_FTN_INPUT),
                err,
                IX_CC_MPLS_ERROR_RM_REM_HDLR,
                IX_ERROR_LEVEL_REMOTE);
    if (err != IX_SUCCESS)
    {
        ix_ossl_message_log("Warning: RM: packet handler cannot be removed\n");
    }

    /* Remove message handler from RM */
    IX_ERROR_CT(ix_rm_packet_handler_unregister(IX_CC_MPLS_ILM_INPUT),
                err,
                IX_CC_MPLS_ERROR_RM_REM_HDLR,
                IX_ERROR_LEVEL_REMOTE);
    if (err != IX_SUCCESS)
    {
        ix_ossl_message_log("Warning: RM: packet handler cannot be removed\n");
    }

#endif /* IX_EXCLUDE_CCI */


    return err;
}; /* _ix_cc_mpls_unregister_packet_handlers */


/*
 * FUNCTION: _ix_cc_mpls_add_icmp_event_handlers()  */
/**
 * ABSTRACT: This function registers ICMP event handlers. 
 *
 * DESCRIPTION: This primitive will be used to register ICMP event handlers. 
 *              It registers with CCI in level 1 system and register
 *              with RM in level 0 system.
 *
 * \param  arg_pContext  - MPLS CC context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 *
 * \n  Error Codes:
 * \n      IX_CC_IPV4_ERROR_CCI - error from CCI/resource manager
 */
PRIVATE ix_error
_ix_cc_mpls_add_icmp_event_handlers(ix_cc_mpls_context* arg_pContext)
{

    ix_uint32 eventInterval;
    
    /* Calculate event interval in milliseconds, 
       data in icmpSleepTime is in seconds */
    eventInterval = arg_pContext->icmpSleepTime;

#if !defined(IX_EXCLUDE_CCI) /* compile time flag */

#if defined (IX_MPLS_DEBUG)
    ix_ossl_message_log ("_ix_cc_mpls_add_icmp_event_handlers(): "
                             "eventInterval = %d\n", eventInterval);
#endif /* IX_MPLS_DEBUG */
    
    IX_ERROR_CRT(ix_cci_cc_add_event_handler(arg_pContext->hMplsHandle,
                                             eventInterval,
                                             _ix_cc_mpls_icmp_send_message,
                                             IX_EVENT_TYPE_PERIODIC,
                                             MPLS_ICMP_EVENT_PRIORITY,
                                             &arg_pContext->hEventHandle),
                 IX_CC_MPLS_ERROR_CCI_REG_HDLR,
                 IX_ERROR_LEVEL_WARNING);

#endif

    return IX_SUCCESS;
} /* _ix_cc_mpls_add_icmp_event_handler */





/*
 * FUNCTION: _ix_cc_mpls_delete_icmp_event_handler()  */
/**
 * ABSTRACT: This primitive will be used to deregister ICMP event handlers. 
 *
 * \param  arg_pContext  - MPLS CC context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 *
 * \n  Error Codes:
 * \n      IX_CC_IPV4_ERROR_CCI - error from CCI/resource manager
 */
PRIVATE ix_error
_ix_cc_mpls_delete_icmp_event_handlers (ix_cc_mpls_context* arg_pContext)
{
        
#if !defined(IX_EXCLUDE_CCI) /* compile time flag */

    IX_ERROR_CRT(ix_cci_cc_remove_event_handler(arg_pContext->hMplsHandle,
                                                arg_pContext->hEventHandle),
                 IX_CC_MPLS_ERROR_CCI_REM_HDLR,
                 IX_ERROR_LEVEL_WARNING);
#endif

    return IX_SUCCESS;

} /* _ix_cc_mpls_delete_icmp_event_handler */



