/*
 * ============================================================================
 * COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *                                                                          */
/** \file      ix_cc_mpls_fragmentation.c
 *
 *  \brief     MPLS fragmentation module
 *
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_mpls_fragmentation.c,v 1.14.2.1 2004/03/12 10:40:14 wneubaue Exp $"


#include <ix_error.h>
#include <ix_types.h>

#include <definitions.h>

/* include IXA SDK framework */
#include <ix_ossl.h>
#include "ix_rm.h"

#if !defined(IX_EXCLUDE_CCI)
#include "ix_cci.h"
#endif 


/* include CC support */
#include "ix_cc_error.h"
#include "ix_cc.h"

#include "bindings.h"

#include "cc/ix_cc_rtmv4.h" 
#include "cc/ipv4/internal/fragmentation_support.h"

#include "../../include/mpls_defs.h"
#include "cc/ix_cc_mpls.h"
#include "cc/internal/ix_cc_mpls_icmp.h"
#include "cc/internal/ix_cc_mpls_lsp.h"
#include "cc/internal/ix_cc_mpls_nhlfe.h"
#include "cc/internal/ix_cc_mpls_fwder.h"
#include "cc/internal/ix_cc_mpls_internal.h"


/*
 * FUNCTION: _ix_cc_mpls_decap ()  */
/**
 * ABSTRACT: Pops all MPLS labels from packet
 *
 * \param  arg_pData            - pointer to the body of a processed packet
 * \param  arg_pHwMeta          - a pointer to the meta data of a processed packet
 * \param  arg_ptMplsLabelStack - label stack popped from packet
 * \param  arg_pLabelNum        - stack size
 * \param  arg_maxLabelNum      - max number of labels that can be popped *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_decap (ix_cc_mpls_label_internal **arg_pData,
                   ix_hw_buffer_meta          *arg_pHwMeta,
                   ix_cc_mpls_label_internal  *arg_ptMplsLabelStack,
                   ix_uint16                  *arg_pLabelNum,
                   ix_uint16                   arg_maxLabelNum)
{
    ix_error        err = IX_SUCCESS;
    ix_uint16       labelNum = 0;
    ix_uint16       buffOffset;
    ix_uint16       pktSize;
    ix_uint16       buffSize;

    buffOffset = IX_NTOH16 (IX_RM_MEM_UINT16_READ (&arg_pHwMeta->m_Offset));
    pktSize    = IX_NTOH16 (IX_RM_MEM_UINT16_READ (&arg_pHwMeta->m_PacketSize));
    buffSize   = IX_NTOH16 (IX_RM_MEM_UINT16_READ (&arg_pHwMeta->m_BufferSize));

    while (labelNum < arg_maxLabelNum)
    {
        ix_uint8  *pByte = (ix_uint8*)*arg_pData;

        ix_ossl_memcpy(arg_ptMplsLabelStack + labelNum, 
                      *arg_pData,
                       sizeof(ix_cc_mpls_label_internal));
        labelNum++;
        (*arg_pData)++;
  
        if (arg_ptMplsLabelStack[labelNum - 1].s == 1) break;
    } 

    if (labelNum > arg_maxLabelNum) {
        #if defined (IX_MPLS_DEBUG)
        ix_ossl_message_log ("Too many labels. Max can be %d labels. Dropping packet\n",arg_maxLabelNum);
        #endif

        return IX_ERROR_WARNING (IX_CC_ERROR_FULL, ("_ix_cc_mpls_decap(): too many mpls labels"));

    }
    buffOffset += (labelNum * sizeof (ix_cc_mpls_label));
    IX_RM_MEM_UINT16_WRITE (&arg_pHwMeta->m_Offset, IX_HTON16 (buffOffset));

    pktSize -= (labelNum * sizeof (ix_cc_mpls_label));
    IX_RM_MEM_UINT16_WRITE (&arg_pHwMeta->m_PacketSize, IX_HTON16 (pktSize));

    buffSize -= (labelNum * sizeof (ix_cc_mpls_label));
    IX_RM_MEM_UINT16_WRITE (&arg_pHwMeta->m_BufferSize, IX_HTON16 (buffSize));

    *arg_pLabelNum = labelNum;

    return IX_SUCCESS;
}	/* _ix_cc_mpls_decap */








/*
 * FUNCTION: _ix_cc_mpls_encap()  */
/**
 * ABSTRACT: Pushes MPLS labels into packet
 *
 * \param  arg_pData            - pointer to the body of a processed packet
 * \param  arg_pHwMeta          - a pointer to the meta data of a processed packet
 * \param  arg_ptMplsLabelStack - label stack popped from packet
 * \param  arg_pLabelNum        - stack size
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_encap(ix_cc_mpls_label_internal **arg_pData,
                  ix_hw_buffer_meta          *arg_pHwMeta,
                  ix_uint16                   arg_maxLabelNum,
                  ix_cc_mpls_label_internal  *arg_ptMplsLabelStack)
{

  ix_error        err      = IX_SUCCESS;
  ix_uint16       labelNum = arg_maxLabelNum;
  ix_uint16       buffOffset;
  ix_uint16       pktSize;
  ix_uint16       buffSize;


  /* Update buffer offset */
  buffOffset = IX_NTOH16 (IX_RM_MEM_UINT16_READ (&arg_pHwMeta->m_Offset));
  pktSize    = IX_NTOH16 (IX_RM_MEM_UINT16_READ (&arg_pHwMeta->m_PacketSize));
  buffSize   = IX_NTOH16 (IX_RM_MEM_UINT16_READ (&arg_pHwMeta->m_BufferSize));

  while (labelNum > 0)
  {
      (*arg_pData)--;
      labelNum--;
      ix_ossl_memcpy(*arg_pData,
                     arg_ptMplsLabelStack + labelNum,
                     sizeof(ix_cc_mpls_label_internal));
  }

  buffOffset -= (arg_maxLabelNum * sizeof (ix_cc_mpls_label));
  IX_RM_MEM_UINT16_WRITE (&arg_pHwMeta->m_Offset, IX_HTON16 (buffOffset));

  pktSize    += (arg_maxLabelNum * sizeof (ix_cc_mpls_label));
  IX_RM_MEM_UINT16_WRITE (&arg_pHwMeta->m_PacketSize, IX_HTON16 (pktSize));

  buffSize   += (arg_maxLabelNum * sizeof (ix_cc_mpls_label));
  IX_RM_MEM_UINT16_WRITE (&arg_pHwMeta->m_BufferSize, IX_HTON16 (buffSize));

  return IX_SUCCESS;

}	/* _ix_cc_mpls_encap */




/*
 * FUNCTION: _ix_cc_mpls_ilm_fragmentation ()  */
/**
 * ABSTRACT: Fragments MPLS packet coming from ilm forwarder
 *
 * DESCRIPTION:  Function retrieves ilm_id from meta data
 *               Label stack is popped
 *               IPv4 packet is fragmented
 *               Every fragment is encapsulated with label stack
 *               Every fragment is sent to mpls forwarder
 *
 * \param  arg_hBuffer   - buffer handle of a processed packet
 * \param  arg_pContext  - a pointer to the MPLS context
 * \param  arg_OutputId  - communication ID of destination Core Componenet
 * \param  arg_OutputBid - if the packet goes to the microcode this is the block id of the block which 
 *                         will take it over.
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_ilm_fragmentation (ix_buffer_handle   arg_hBuffer,
                               ix_uint16          arg_OutputId,
                               ix_uint16          arg_OutputBid,
                               void              *arg_pContext)
{
    ix_error                   err;

    ix_uint32                  IlmId;
    void                      *pData   = NULL;
    ix_hw_buffer_meta         *pHwMeta = NULL;
    ix_cc_mpls_ilm_internal   *pIlm;
    ix_cc_mpls_context        *pContext = arg_pContext;
#if defined (MPLS_COUNTERS_FEATURE)
    ix_cc_mpls_ilm_counters_internal  *pIlmCountersEntry;
    ix_uint16                          fragsNum = 0;
    ix_uint32                          fragsAddOctets = 0;
#endif
    
#if defined (IX_MPLS_DEBUG)
    ix_ossl_message_log ("_ix_cc_mpls_ilm_fragmentation: start\n");
#endif

    /* Get meta data for buffer */
    IX_ERROR_CRT(ix_rm_buffer_get_meta (arg_hBuffer, (void **) &pHwMeta),
                 IX_CC_ERROR_INTERNAL, 
                 IX_ERROR_LEVEL_WARNING);

    IX_ERROR_CRT(ix_cc_hw_get_packet_data (arg_hBuffer, &pData),
                 IX_CC_ERROR_INTERNAL, 
                 IX_ERROR_LEVEL_WARNING);

#if defined (IX_MPLS_DEBUG)
    show_metadata(pHwMeta);
    show_mem ((ix_uint32 *) pData, 32);
#endif
    
    IlmId = ((ix_uint32)IX_NTOH16 (IX_RM_MEM_UINT16_READ (&pHwMeta->m_OutputPort)))*65536 +
            ((ix_uint32)IX_NTOH16 (IX_RM_MEM_UINT16_READ (&pHwMeta->m_NextHopID)));

    if (IlmId & (1 << MPLS_BIT_ILM_SET_NOTIFICATION)) {
        pIlm = (ix_cc_mpls_ilm_internal*) &pContext->pIlmSetBaseAddr[IlmId & ~(1<<31)];
    } else {
        pIlm = (ix_cc_mpls_ilm_internal*) &pContext->pIlmBaseAddr[IlmId];
    }

#if defined (IX_MPLS_DEBUG)
    ix_ossl_message_log("_ix_cc_mpls_ilm_fragmentation: IlmId   : 0x%x\n", IlmId);
    ix_ossl_message_log("_ix_cc_mpls_ilm_fragmentation: ILMEntry: 0x%x\n", pIlm);
#endif

    if  ((!(pIlm ->flags & MPLS_FLAG_VALID_BIT))  ||
            pIlm ->flags & MPLS_FLAG_SET_BIT) 
    {
#if defined (IX_MPLS_DEBUG)
        ix_ossl_message_log ("_ix_cc_mpls_ilm_fragmentation(): ILM entry "
                             "received from microcode is invalid. dropping...\n");
#endif

        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND, 
                                ("_ix_cc_mpls_ilm_fragmentation(): Can't go on with this ILM handle"));
    };

    pHwMeta->m_NextHopID  = (ix_uint16)pIlm->nextHopId;
    pHwMeta->m_OutputPort = (ix_uint16)pIlm->outPortId;

    err =  _ix_cc_mpls_fragmentation (arg_hBuffer,
                                      pData,
                                      pHwMeta,
                                      pIlm->maxLabPktSize,
                                      arg_OutputId,
                                      arg_OutputBid,
#if defined (MPLS_COUNTERS_FEATURE)
                                     &fragsNum,
                                     &fragsAddOctets,
#endif
                                      arg_pContext);

#if defined (MPLS_COUNTERS_FEATURE)
    if (IlmId & (1 << MPLS_BIT_ILM_SET_NOTIFICATION))
    {
        pIlmCountersEntry =  (ix_cc_mpls_ilm_counters_internal*)((ix_uint8*)(pContext->pIlmNhlfeSetCounters) + 
                              (IlmId & ~(1<<31)) * sizeof(ix_cc_mpls_ilm_counters_internal));
    }
    else
    {
        pIlmCountersEntry =  (ix_cc_mpls_ilm_counters_internal*)((ix_uint8*)(pContext->pIlmCounters) + 
                             pContext->pLabelSpaceTable[pHwMeta->m_InputPort].stats_offset +
                             (IlmId - pContext->pLabelSpaceTable[pHwMeta->m_InputPort].min_label)
                              * sizeof(ix_cc_mpls_ilm_counters_internal));
    }
    ix_rm_atomic_sram_add(fragsNum - 1, &pIlmCountersEntry->mplsInSegFragmPktDelta);
    ix_rm_atomic_sram_add(fragsAddOctets - pHwMeta->m_PacketSize, &pIlmCountersEntry->mplsInSegFragmOctDelta);
#endif

   
    return err;
} /* _ix_cc_mpls_ilm_fragmentation */




/*
 * FUNCTION: _ix_cc_mpls_ftn_fragmentation ()  */
/**
 * ABSTRACT: Fragments MPLS packet coming from ftn forwarder
 *
 * DESCRIPTION:  Function retrieves nhlfe_id from meta data
 *               Label stack is popped
 *               IPv4 packet is fragmented
 *               Every fragment is encapsulated with label stack
 *               Every fragment is sent to mpls forwarder
 *
 * \param  arg_hBuffer   - buffer handle of a processed packet
 * \param  arg_pContext  - a pointer to the MPLS context
 * \param  arg_OutputId  - communication ID of destination Core Componenet
 * \param  arg_OutputBid - if the packet goes to the microcode this is the block id of the block which 
 *                         will take it over.
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_ftn_fragmentation (ix_buffer_handle   arg_hBuffer,
                               ix_uint16          arg_OutputId,
                               ix_uint16          arg_OutputBid,
                               void              *arg_pContext)
{
    ix_error                   err;

    ix_uint16                  NhlfeId;
    void                      *pData   = NULL;
    ix_hw_buffer_meta         *pHwMeta = NULL;
    ix_cc_mpls_nhlfe_internal *pNhlfe;
    ix_cc_mpls_context        *pContext = arg_pContext;
#if defined (MPLS_COUNTERS_FEATURE)
    ix_cc_mpls_nhlfe_counters_internal  *pFtnCountersEntry;
    ix_uint16                            fragsNum;
    ix_uint32                            fragsAddOctets;
#endif

#if defined (IX_MPLS_DEBUG)
    ix_ossl_message_log ("_ix_cc_mpls_ftn_fragmentation:start\n");
#endif

    /* Get meta data for buffer */
    IX_ERROR_CRT(ix_rm_buffer_get_meta (arg_hBuffer, (void **) &pHwMeta),
                 IX_CC_ERROR_INTERNAL, 
                 IX_ERROR_LEVEL_WARNING);

    IX_ERROR_CRT(ix_cc_hw_get_packet_data (arg_hBuffer, &pData),
                 IX_CC_ERROR_INTERNAL, 
                 IX_ERROR_LEVEL_WARNING);

#if defined (IX_MPLS_DEBUG)
    show_metadata(pHwMeta);
    show_mem ((ix_uint32 *) pData, 32);
#endif
    
    NhlfeId      = IX_NTOH16 (IX_RM_MEM_UINT16_READ (&pHwMeta->m_OutputPort))*65536 +
                   IX_NTOH16 (IX_RM_MEM_UINT16_READ (&pHwMeta->m_NextHopID));

    pNhlfe = (ix_cc_mpls_nhlfe_internal*) &pContext->pNhlfeBaseAddr[NhlfeId];

    if  ((!(pNhlfe ->flags & MPLS_FLAG_VALID_BIT))  ||
            pNhlfe ->flags & MPLS_FLAG_SET_BIT) 
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND, 
                                ("_ix_cc_mpls_ftn_fragmentation(): Can't go on with this NHLFE handle"));
    }

    pHwMeta->m_NextHopID  = (ix_uint16)pNhlfe->nextHopId;
    pHwMeta->m_OutputPort = (ix_uint16)pNhlfe->outPortId;

    err =  _ix_cc_mpls_fragmentation (arg_hBuffer,
                                      pData,
                                      pHwMeta,
                                      pNhlfe->maxLabPktSize,
                                      arg_OutputId,
                                      arg_OutputBid,
#if defined (MPLS_COUNTERS_FEATURE)
                                     &fragsNum,
                                     &fragsAddOctets,
#endif
                                      arg_pContext);

#if defined (MPLS_COUNTERS_FEATURE)
    pFtnCountersEntry =  (ix_cc_mpls_nhlfe_counters_internal*)((ix_uint8*)(pContext->pFtnCounters) + 
                          NhlfeId * sizeof(ix_cc_mpls_ilm_counters_internal));
    ix_rm_atomic_sram_add(fragsNum - 1, &pFtnCountersEntry->mplsNhlfeFragmPktsDelta);
    ix_rm_atomic_sram_add(fragsAddOctets - pHwMeta->m_PacketSize, &pFtnCountersEntry->mplsNhlfeFragmOctetsDelta);
#endif

    return err;
} /* _ix_cc_mpls_ftn_fragmentation */





/*
 * FUNCTION: _ix_cc_mpls_fragmentation ()  */
/**
 * ABSTRACT: Fragments MPLS packet
 *
 * DESCRIPTION:  Label stack is popped
 *               IPv4 packet is fragmented
 *               Every fragment is encapsulated with label stack
 *               Every fragment is sent to mpls forwarder
 *
 * \param  arg_hBuffer   - buffer handle of a processed packet
 * \param  arg_pContext  - a pointer to the MPLS context
 * \param  arg_OutputId  - communication ID of destination Core Componenet
 * \param  arg_OutputBid - if the packet goes to the microcode this is the block id of the block which 
 *                         will take it over.
 * \param  arg_pFragsNum - returns the number of fragments the packet has been split into
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_fragmentation (ix_buffer_handle     arg_hBuffer,
                           void                *arg_pData,
                           ix_hw_buffer_meta   *arg_pHwMeta,
                           ix_uint32            arg_maxIpPktSize,
                           ix_uint16            arg_OutputId,
                           ix_uint16            arg_OutputBid,
#if defined (MPLS_COUNTERS_FEATURE)
                           ix_uint16           *arg_pFragsNum,
                           ix_uint32           *arg_pFragsAddOctets,
#endif
                           void                *arg_pContext)
{
    ix_error            err = IX_SUCCESS;

    ix_cc_mpls_label_internal  MplsLabelStack[MAX_MPLS_LABS];
    ix_cc_mpls_context        *pContext = arg_pContext;
    
    ix_uint8        fabricPort;
    ix_uint16       outputPort;
    ix_uint16       labelNum;
    ix_uint8        DFSet = 0;
    ix_uint32       maxIpPktSize;
    
    ix_cc_fragv4_fragment     *pFirst   = NULL;
    ix_cc_fragv4_fragment     *pNewFrag = NULL;
    ix_cc_fragv4_fragment     *pTemp;

    ix_uint8                  *pIpHdr   = NULL;


#ifdef IX_MPLS_DEBUG
    {
         ix_ossl_message_log ("---------  PACKET TO FRAGMENT ------------- \n");
         show_mem ((ix_uint32 *) arg_pData, 32);
    }
#endif


    fabricPort   = IX_RM_MEM_UINT8_READ  (&arg_pHwMeta->m_FabricPort);
    outputPort   = IX_RM_MEM_UINT16_READ (&arg_pHwMeta->m_OutputPort);


    if (IX_RM_HW_META_GET_HEADER_TYPE(arg_pHwMeta->m_BufferInfo) == MPLS_MPLS_HEADER_TYPE)
    {
        /* Remove MPLS labels */
        err = _ix_cc_mpls_decap ((ix_cc_mpls_label_internal**)(&arg_pData),
                                  arg_pHwMeta,
                                  MplsLabelStack,
                                 &labelNum, 
                                  MAX_MPLS_LABS);
        if (err != IX_SUCCESS)
        {
#if defined (IX_MPLS_DEBUG)
            ix_ossl_message_log ("_ix_cc_mpls_fragmentation(): Packet decapsulation failed."
                                 " Dropping packet\n");
#endif
            return err;
        }
    }
    else
    {
        labelNum = 0;
    }


   /* 
    * Fragment IP packet. 
    * The output buffer contains ipv4 packet and mpls labels
    */
    if (arg_maxIpPktSize < labelNum * sizeof (ix_cc_mpls_label_internal))
    {
        ix_ossl_message_log ("MTU is too low to carry all labels on a packet. dropping \n");
        return IX_ERROR_NEW (IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
    }
    maxIpPktSize = arg_maxIpPktSize - labelNum * sizeof (ix_cc_mpls_label_internal);
    

    /* Check if fragment fits in buffer */
    if (maxIpPktSize < IX_CC_FRAGV4_IP_HEADER_MIN_SIZE)
    {
        return IX_ERROR_NEW (IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
    }

    /* Check if IPv4 header can be fragmented (if DF bit is set */
    /* Get data from buffer */

#ifdef IX_MPLS_DEBUG
    {
         ix_ossl_message_log ("---------  packet after decap ------------- \n");
         show_mem ((ix_uint32 *) arg_pData, 32);
         ix_ossl_message_log ("m_Offset       = 0x%x\n", 
                               (int)IX_NTOH16 (IX_RM_MEM_UINT16_READ(&arg_pHwMeta->m_Offset)));
         ix_ossl_message_log ("m_PacketSize   = 0x%x\n", 
                               (int)IX_NTOH16 (IX_RM_MEM_UINT16_READ(&arg_pHwMeta->m_PacketSize)));
    }
#endif

    /* Get pointer to IP header */
    pIpHdr = (ix_uint8 *) arg_pData;

    /* Check ip header version. If not 4, packet must be droped */
    if (((ix_cc_ipv4_header *) pIpHdr)->version != 4)
    {

#if defined (IX_MPLS_DEBUG)
        ix_ossl_message_log ("_ix_cc_mpls_fragmentation(): IP version (%d) incorrect.  Dropping packet\n", 
                             ((ix_cc_ipv4_header *) pIpHdr)->version);
#endif

        return IX_ERROR_NEW (IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
    }

    /* Check ipv4 header protocol field. IPCM packet cannot be fragmented 
       (to avoid loop) and must be droped */
    if (((ix_cc_ipv4_header *) pIpHdr)->protocol == IX_IP_PROTOICMP)
    {

#if defined (IX_MPLS_DEBUG)
        ix_ossl_message_log ("_ix_cc_mpls_fragmentation(): IP packet carries ICMP. "
                             "ICMP cannot be fragmented. Dropping packet\n");
#endif
        
        return IX_ERROR_NEW (IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
    }


    /* Check DF flag - if packet cannot be fragmented, ICMP message is sent */
    if (_ix_cc_ipv4_header_get_DF (arg_hBuffer, pIpHdr, &DFSet) != IX_SUCCESS)
    {

#if defined (IX_MPLS_DEBUG)
        ix_ossl_message_log ("_ix_cc_mpls_fragmentation(): Cannot determine DF bit. Dropping packet\n");
#endif

        return IX_ERROR_NEW (IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
    }

    if (DFSet)
    {
#if defined (IX_MPLS_DEBUG)
        ix_ossl_message_log ("_ix_cc_mpls_fragmentation(): DF bit is SET\n");
#endif

        /* Since DF bit is set, an ICMP message must be generated 
         * for the original encapsulated packet
       . */

        IX_RM_HW_META_SET_HEADER_TYPE(arg_pHwMeta->m_BufferInfo, MPLS_IPV4_HEADER_TYPE);

	return _ix_cc_mpls_icmp_dest_unreachable(arg_hBuffer,
                                                 IX_ICMP_UNREACH_NEED_FRAG,
                                                 0xFFFF & maxIpPktSize,
                                                 pContext);
    }	
    else
    {
#if defined (IX_MPLS_DEBUG)
        ix_ossl_message_log ("_ix_cc_mpls_fragmentation(): DF bit is NOT set\n");
#endif
    } /* if (DFSet) */



    /* Fragment IPv4 packet */

    IX_ERROR_CR (ix_cc_fragv4_fragment_make_from_buffer (arg_hBuffer,
                                                         0, &pNewFrag));


    #if defined (IX_MPLS_DEBUG)
    ix_ossl_message_log ("arg_hBuffer = 0x%x\n",     arg_hBuffer);
    ix_ossl_message_log ("pFirst->hBuffer = 0x%x\n", pFirst->hBuffer);
    ix_ossl_message_log ("\n\n");
    #endif


    /* fragment newly created fragment into smaller fragments */
    if ((err = ix_cc_fragv4_fragment_divide (pNewFrag,
                                            maxIpPktSize,
                                            g_pIpv4Context->hBufferFreeList,
                                            &pFirst)) != IX_SUCCESS)
    {
        /*
         * Free the fragment element that holds the input packet.
         * (Caller to free the input ix_buffer for error return)
         */
        ix_cc_fragv4_fragment_delete_frag_only(pNewFrag);
#if defined (IX_MPLS_DEBUG)
        ix_ossl_message_log ("_ix_cc_mpls_fragmentation(): Fragmentation failed\n");
#endif
        return IX_ERROR_NEW (IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
    }


    /* traverse fragment list */
    pTemp = pFirst;

#if defined (MPLS_COUNTERS_FEATURE)
    *arg_pFragsNum = 0;
    *arg_pFragsAddOctets = 0;
#endif

    while (pTemp != NULL)
    {
        ix_cc_fragv4_fragment  *pFragDelete = pTemp ;

        if (pTemp->hBuffer != 0)
        {
            ix_cc_hw_get_packet_data (pTemp->hBuffer, &arg_pData);
            ix_rm_buffer_get_meta (pTemp->hBuffer, (void **) &arg_pHwMeta);

#ifdef IX_MPLS_DEBUG
            {
                ix_ossl_message_log ("traverse fragment list\n");
                ix_ossl_message_log ("Network packet\n");
                ix_ossl_message_log ("---------  before encap ------------- \n");
                ix_ossl_message_log ("pTemp->hBuffer = 0x%x\n", pTemp->hBuffer);
                ix_ossl_message_log ("m_Offset       = 0x%x\n", 
                                      (int)IX_NTOH16 (IX_RM_MEM_UINT16_READ(&arg_pHwMeta->m_Offset)));
                ix_ossl_message_log ("m_PacketSize   = 0x%x\n", 
                                      (int)IX_NTOH16 (IX_RM_MEM_UINT16_READ(&arg_pHwMeta->m_PacketSize)));
            }
#endif

            if (IX_RM_HW_META_GET_HEADER_TYPE(arg_pHwMeta->m_BufferInfo) == MPLS_MPLS_HEADER_TYPE)
            {

                /* append labels for every fragment */
                _ix_cc_mpls_encap ((ix_cc_mpls_label_internal**)&arg_pData,
                                   arg_pHwMeta,
                                   labelNum,
                                   MplsLabelStack);
            }

#ifdef IX_MPLS_DEBUG
            {
                ix_ossl_message_log ("---------  after encap ------------- \n");
                ix_ossl_message_log ("pTemp->hBuffer = 0x%x\n", pTemp->hBuffer);
                ix_ossl_message_log ("m_Offset       = 0x%x\n", 
                                      (int)IX_NTOH16 (IX_RM_MEM_UINT16_READ(&arg_pHwMeta->m_Offset)));
                ix_ossl_message_log ("m_PacketSize   = 0x%x\n", 
                                      (int)IX_NTOH16 (IX_RM_MEM_UINT16_READ(&arg_pHwMeta->m_PacketSize)));
            }
#endif

            err = ix_rm_packet_send(arg_OutputId, pTemp->hBuffer, arg_OutputBid);

#ifdef IX_MPLS_DEBUG
            if (err != IX_SUCCESS) 
                ix_ossl_message_log ("_ix_cc_mpls_fragmentation(): Error "
                                      "in sending one of the fragments. Dropping the rest.\n");
#endif
        }

        pTemp = (ix_cc_fragv4_fragment *) pTemp->pNext;

        /* Free the buffer and return error to caller */
        ix_cc_fragv4_fragment_delete_frag_only (pFragDelete);

#if defined (MPLS_COUNTERS_FEATURE)
        (*arg_pFragsNum) ++;
        (*arg_pFragsAddOctets) += IX_NTOH16 (IX_RM_MEM_UINT16_READ(&arg_pHwMeta->m_PacketSize));
#endif

    }

    /* the associated buffer is not used anywhere else */
    /* so it is deleted here */
    IX_ERROR_CR(ix_cc_fragv4_fragment_delete_frag_and_data_buff(pNewFrag));

    return IX_SUCCESS;

} /* _ix_cc_mpls_fragmentation */















