/*
 * ============================================================================
 * COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *                                                                          */
/** \file      ix_cc_mpls_fwder.c
 *
 *  \brief     The file defines mpls forwarding slow path.
 *
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_mpls_fwder.c,v 1.26.2.1 2004/03/12 10:40:14 wneubaue Exp $";


#include <ix_error.h>
#include <ix_types.h>

/* include IXA SDK framework */
#include <ix_ossl.h>
#include "ix_rm.h"

#if !defined(IX_EXCLUDE_CCI)
#include "ix_cci.h"
#endif 

#include "ix_netmacros.h"

/* include CC support */
#include "ix_cc_error.h"
#include "ix_cc.h"

#include "definitions.h"
#include "bindings.h"

#include "cc/ix_cc_rtmv4.h" 
#include "cc/ipv4/internal/utils.h"
#include "cc/ipv4/internal/header_defs.h"
#include "cc/ipv6/internal/v6_header_defs.h"

#include "../../include/mpls_defs.h"
#include "cc/ix_cc_mpls.h"
#include "cc/internal/ix_cc_mpls_icmp.h"
#include "cc/internal/ix_cc_mpls_lsp.h"
#include "cc/internal/ix_cc_mpls_nhlfe.h"
#include "cc/internal/ix_cc_mpls_fwder.h"
#include "cc/internal/ix_cc_mpls_internal.h"



/**
 * Private functions
 */
PRIVATE void
_ix_cc_mpls_lsp_copy_ilm_to_nhlfe(ix_cc_mpls_ilm_internal   *arg_pIlmEntry,
                                  ix_cc_mpls_nhlfe_internal *arg_pNhlfe);


PRIVATE ix_uint8
_ix_cc_mpls_reserved_label_handling(ix_cc_mpls_label_internal **arg_ppMplsLabel,
                                    ix_hw_buffer_meta          *arg_pHwMeta);

PRIVATE ix_uint32
_ix_cc_mpls_fwder_pop_handling(ix_cc_mpls_label_internal **arg_ppMplsLabel,
                               ix_hw_buffer_meta          *arg_pHwMeta,
                               ix_uint32                   arg_NhlfeFlags,
                               ix_uint8                    arp_l3protocol,
                               ix_uint8                   *arg_newTTL);

PRIVATE void
_ix_cc_mpls_fwder_swap_handling(ix_cc_mpls_label_internal *arg_pMplsLabel,
                                ix_cc_mpls_nhlfe_internal *arg_pNhlfe,
                                ix_uint32                  arg_oTTL);


PRIVATE void
_ix_cc_mpls_fwder_push_handling(ix_cc_mpls_label_internal **arg_ppMplsLabel,
                                ix_hw_buffer_meta          *arg_pHwMeta,
                                ix_cc_mpls_nhlfe_internal  *arg_pNhlfe,
                                ix_uint8                    arg_EndofStack,
                                ix_uint32                   arg_oTTL);


/*
 * FUNCTION: _ix_cc_mpls_ilm_forwarder()  */
/**
 * ABSTRACT: implements mpls slow path ILM forwarding
 *
 *
 * \param  arg_pContext - a pointer to the MPLS context
 * \param  arg_hBuffer  - packet buffer handle.
 * \param  arg_icmpSend - specifies weather or not generate ICMP packets
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_ilm_forwarder(ix_buffer_handle    arg_hBuffer,
                          ix_icmp_send        arg_icmpSend,
                          void               *arg_pContext)
{

  ix_error  err = IX_SUCCESS;

  void                             *pData;        /* pointer to the packet contents */
  ix_hw_buffer_meta                *pHwMeta;      /* pointer to the packet meta data */
  ix_cc_mpls_label_internal        MplsLabel;     /* pointer to MPLS label stack */
  ix_cc_mpls_context               *pContext = arg_pContext;
  ix_cc_mpls_nhlfe_internal         Nhlfe;
  ix_cc_mpls_ilm_internal          *pIlmEntry;
#if defined (MPLS_COUNTERS_FEATURE)
  ix_cc_mpls_ilm_counters_internal *pIlmCountersEntry;
#endif
    
  /* Dispatch loop data */
  ix_uint32       in_port;
  ix_uint16       PacketOutput     = IX_CC_MPLS_DROP_OUTPUT - 1; /* initial value */
  ix_uint16       PacketOutput_bid = IX_CC_MPLS_DROP_OUTPUT - 1; /* - prevents from random packet dropping */
  ix_uint32       NextLookup;  /* MPLS lookup loop maintanance */

  /* Get data from buffer */
  IX_ERROR_CRT(ix_cc_hw_get_packet_data(arg_hBuffer, &pData), IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_LOCAL);

  /* Get meta data for buffer */
  IX_ERROR_CRT(ix_rm_buffer_get_meta (arg_hBuffer, (void **) &pHwMeta), IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_LOCAL);

  /* Set meta data */
  pHwMeta->m_NextHopIDType = NEXT_HOP_ID_MPLS_TYPE;
  IX_RM_MEM_UINT16_WRITE (&pHwMeta->m_NextHopID, IX_HTON16 (0xFFFF));


  in_port = IX_NTOH16 (IX_RM_MEM_UINT16_READ (&pHwMeta->m_InputPort));

  NextLookup = 0;
  /* Recursive ILM lookup */
  do 
  {
      ix_uint8                          iTTL;

      NextLookup = 0;

      /* Read the top mpls label */
      ix_ossl_memcpy(&MplsLabel, pData, sizeof(ix_cc_mpls_label_internal));

#if defined (IX_MPLS_DEBUG)
      ix_ossl_message_log ("\n\n******** Recursive ILM lookup *********\n");
      ix_ossl_message_log ("MplsLabel.label = 0x%x\n", (int) MplsLabel.label);
      ix_ossl_message_log ("MplsLabel.exp = 0x%x\n", (int) MplsLabel.exp);
      ix_ossl_message_log ("MplsLabel.s = 0x%x\n", (int) MplsLabel.s);
      ix_ossl_message_log ("MplsLabel.ttl = 0x%x\n", (int) MplsLabel.ttl);
#endif

      if (MplsLabel.label < IX_MPLS_RES_LABEL_MAX)
      {
#if defined (IX_MPLS_DEBUG)
          ix_ossl_message_log ("_ix_cc_mpls_ilm_forwarder(): got reserved label.\n");
#endif

      } else if ((MplsLabel.label < pContext->pLabelSpaceTable[in_port].min_label) ||
                 (MplsLabel.label > pContext->pLabelSpaceTable[in_port].max_label))
    
      {
#if defined (IX_MPLS_DEBUG)
          ix_ossl_message_log ("_ix_cc_mpls_ilm_forwarder(): got the label from out of range.\n");
#endif
          PacketOutput     = IX_CC_MPLS_DROP_OUTPUT;
          PacketOutput_bid = IX_CC_MPLS_DROP_OUTPUT;
      }
      else
      {
         pIlmEntry = (ix_cc_mpls_ilm_internal*)((ix_uint8*)pContext->pIlmBaseAddr + 
                      pContext->pLabelSpaceTable[in_port].table_offset +
                     (MplsLabel.label - pContext->pLabelSpaceTable[in_port].min_label)
                          * sizeof(ix_cc_mpls_ilm_internal));

         if (pIlmEntry->flags & MPLS_FLAG_VALID_BIT)
         {
             if (pIlmEntry->flags & MPLS_FLAG_SET_BIT)
             {
                 ix_cc_mpls_ilm_set_internal *pIlmSetEntry = (ix_cc_mpls_ilm_set_internal*) pIlmEntry;
                 ix_cc_mpls_nhlfe_internal   *pNhlfeEntry;

                 pNhlfeEntry = (ix_cc_mpls_nhlfe_internal*)(&pContext->pIlmSetBaseAddr[((ix_uint32)(pIlmSetEntry->OutSegId1))]);

                 Nhlfe = *pNhlfeEntry;

             } else
                 _ix_cc_mpls_lsp_copy_ilm_to_nhlfe(pIlmEntry, &Nhlfe);

             /* Check if the nhlfe is invalid */
             if (!(Nhlfe.flags & MPLS_FLAG_VALID_BIT))
             {
#if defined (IX_MPLS_DEBUG)
                 ix_ossl_message_log ("_ix_cc_mpls_ilm_forwarder(): NHLFE is invalid\n");
#endif
                 PacketOutput     = IX_CC_MPLS_DROP_OUTPUT;
                 PacketOutput_bid = IX_CC_MPLS_DROP_OUTPUT;
             }

         }
         else
         {
#if defined (IX_MPLS_DEBUG)
             ix_ossl_message_log ("_ix_cc_mpls_ilm_forwarder(): ILM is invalid\n");
#endif
             PacketOutput     = IX_CC_MPLS_DROP_OUTPUT;
             PacketOutput_bid = IX_CC_MPLS_DROP_OUTPUT;
         }
      }
    
      /* LOOKUP FAILED */
      if (PacketOutput == IX_CC_MPLS_DROP_OUTPUT)
      {
          /* Failed reserved label lookup */
#if defined (IX_MPLS_DEBUG)
          ix_ossl_message_log ("_ix_cc_mpls_ilm_forwarder(): Lookup failed\n");
#endif
          
          break;
      } 

      /* LOOKUP SUCCEEDED */
#if defined (MPLS_COUNTERS_FEATURE)
      if (pIlmEntry->flags & MPLS_FLAG_SET_BIT)
      {
          ix_cc_mpls_ilm_set_internal *pIlmSetEntry = (ix_cc_mpls_ilm_set_internal*) pIlmEntry;

          pIlmCountersEntry =  (ix_cc_mpls_ilm_counters_internal*)((ix_uint8*)(pContext->pIlmNhlfeSetCounters) + 
                                ((ix_uint32)(pIlmSetEntry->OutSegId1)) * sizeof(ix_cc_mpls_ilm_counters_internal));
      }
      else
      {
          pIlmCountersEntry =  (ix_cc_mpls_ilm_counters_internal*)((ix_uint8*)(pContext->pIlmCounters) + 
                                pContext->pLabelSpaceTable[in_port].stats_offset +
                               (MplsLabel.label - pContext->pLabelSpaceTable[in_port].min_label)
                                   * sizeof(ix_cc_mpls_ilm_counters_internal));
      }
      ix_rm_atomic_sram_add(1, &pIlmCountersEntry->mplsInSegmentPkts);
      ix_rm_atomic_sram_add(IX_NTOH16 (IX_RM_MEM_UINT16_READ (&pHwMeta->m_PacketSize)),
                            &pIlmCountersEntry->mplsInSegmentOctets);
#endif


      /*** PHASE 2 ***/

#if defined (IX_MPLS_DEBUG)
      ix_ossl_message_log ("*** PHASE 2 ***\n");
#endif

      iTTL = (ix_uint8)MplsLabel.ttl;

      if (MplsLabel.label < IX_MPLS_RES_LABEL_MAX)
      {
          ix_uint32 ExitCode;
#if defined (IX_MPLS_DEBUG)
          ix_ossl_message_log ("_ix_cc_mpls_ilm_forwarder(): got reserved label.\n");
#endif

          ExitCode = _ix_cc_mpls_reserved_label_handling((ix_cc_mpls_label_internal**)&pData,
                                                         pHwMeta);
          switch(ExitCode) {
              case 1:
                      pHwMeta->m_NextHopID     = 0xFFFF;
                      pHwMeta->m_NextHopIDType = NEXT_HOP_ID_IPV4_TYPE;
                      PacketOutput     = IX_CC_MPLS_ILM_FORWARDER_IPV4_OUTPUT;
                      PacketOutput_bid = IX_CC_MPLS_ILM_FORWARDER_IPV4_OUTPUT_BID;
                      break;

              case 2: 
                      pHwMeta->m_NextHopID     = 0xFFFF;
                      pHwMeta->m_NextHopIDType = NEXT_HOP_ID_IPV6_TYPE;
                      PacketOutput     = IX_CC_MPLS_ILM_FORWARDER_IPV6_OUTPUT;
                      PacketOutput_bid = IX_CC_MPLS_ILM_FORWARDER_IPV6_OUTPUT_BID;
                      break;

              default:PacketOutput     = IX_CC_MPLS_DROP_OUTPUT;
                      PacketOutput_bid = IX_CC_MPLS_DROP_OUTPUT;
          } 
      } else if (Nhlfe.operation == IX_MPLS_LABELOP_POP) {

          NextLookup = _ix_cc_mpls_fwder_pop_handling(
                              (ix_cc_mpls_label_internal**)&pData,
                                                            pHwMeta,
                                                 (ix_uint8) Nhlfe.flags,
                                                 (ix_uint8) Nhlfe.l3Protocol,
                                                           &iTTL);

          if (((Nhlfe.flags & MPLS_FLAG_TUNNEL_UNIFORM) == 0) && (iTTL <= Nhlfe.ttl))
          {
              if (arg_icmpSend)
                  err = _ix_cc_mpls_icmp_time_exceeded(arg_hBuffer,
                                                       arg_pContext);

              PacketOutput     = IX_CC_MPLS_DROP_OUTPUT;
              PacketOutput_bid = IX_CC_MPLS_DROP_OUTPUT;
#if defined (MPLS_COUNTERS_FEATURE)
              if (pIlmEntry->flags & MPLS_FLAG_SET_BIT)
              {
                  ix_cc_mpls_ilm_set_internal *pIlmSetEntry = (ix_cc_mpls_ilm_set_internal*) pIlmEntry;
          
                  pIlmCountersEntry =  (ix_cc_mpls_ilm_counters_internal*)((ix_uint8*)(pContext->pIlmNhlfeSetCounters) + 
                                        ((ix_uint32)(pIlmSetEntry->OutSegId1)) * sizeof(ix_cc_mpls_ilm_counters_internal));
              }
              else
              {
                  pIlmCountersEntry =  (ix_cc_mpls_ilm_counters_internal*)((ix_uint8*)(pContext->pIlmCounters) + 
                                        pContext->pLabelSpaceTable[in_port].stats_offset +
                                       (MplsLabel.label - pContext->pLabelSpaceTable[in_port].min_label)
                                          * sizeof(ix_cc_mpls_ilm_counters_internal));
              }
              ix_rm_atomic_sram_add(1, &pIlmCountersEntry->mplsInSegErrorPkts);
              ix_rm_atomic_sram_add(IX_NTOH16 (IX_RM_MEM_UINT16_READ (&pHwMeta->m_PacketSize)),
                                    &pIlmCountersEntry->mplsInSegErrorOctets);
#endif
              break;
          }

          if (!NextLookup)  {
              
              pHwMeta->m_NextHopID     = 0xFFFF;

              switch(Nhlfe.l3Protocol) {

                  case MPLS_L3_IP_V4:
                    PacketOutput     = IX_CC_MPLS_ILM_FORWARDER_IPV4_OUTPUT;
                    PacketOutput_bid = IX_CC_MPLS_ILM_FORWARDER_IPV4_OUTPUT_BID;

                    break;

                  case MPLS_L3_IP_V6:

                  default:
#if defined (MPLS_COUNTERS_FEATURE)
                      if (pIlmEntry->flags & MPLS_FLAG_SET_BIT)
                      {
                          ix_cc_mpls_ilm_set_internal *pIlmSetEntry = (ix_cc_mpls_ilm_set_internal*) pIlmEntry;
                          
                          pIlmCountersEntry =  (ix_cc_mpls_ilm_counters_internal*)((ix_uint8*)(pContext->pIlmNhlfeSetCounters) + 
                                                ((ix_uint32)(pIlmSetEntry->OutSegId1) * sizeof(ix_cc_mpls_ilm_counters_internal)));
                      }
                      else
                      {
                          pIlmCountersEntry =  (ix_cc_mpls_ilm_counters_internal*)((ix_uint8*)(pContext->pIlmCounters) + 
                                                pContext->pLabelSpaceTable[in_port].stats_offset +
                                               (MplsLabel.label - pContext->pLabelSpaceTable[in_port].min_label)
                                                    * sizeof(ix_cc_mpls_ilm_counters_internal));
                      }
                      ix_rm_atomic_sram_add(1, &pIlmCountersEntry->mplsInSegErrorPkts);
                      ix_rm_atomic_sram_add(IX_NTOH16 (IX_RM_MEM_UINT16_READ (&pHwMeta->m_PacketSize)),
                                            &pIlmCountersEntry->mplsInSegErrorOctets);
#endif
                    PacketOutput     = IX_CC_MPLS_DROP_OUTPUT;
                    PacketOutput_bid = IX_CC_MPLS_DROP_OUTPUT;
              }
              
              break;
          }

      } else if (iTTL <= Nhlfe.ttl) {

          if (arg_icmpSend)
                err = _ix_cc_mpls_icmp_time_exceeded(arg_hBuffer,
                                                     arg_pContext);

#if defined (MPLS_COUNTERS_FEATURE)
          if (pIlmEntry->flags & MPLS_FLAG_SET_BIT)
          {
              ix_cc_mpls_ilm_set_internal *pIlmSetEntry = (ix_cc_mpls_ilm_set_internal*) pIlmEntry;
              
              pIlmCountersEntry =  (ix_cc_mpls_ilm_counters_internal*)((ix_uint8*)(pContext->pIlmNhlfeSetCounters) + 
                                    ((ix_uint32)(pIlmSetEntry->OutSegId1)) * sizeof(ix_cc_mpls_ilm_counters_internal));
          }
          else
          {
              pIlmCountersEntry =  (ix_cc_mpls_ilm_counters_internal*)((ix_uint8*)(pContext->pIlmCounters) + 
                                    pContext->pLabelSpaceTable[in_port].stats_offset +
                                   (MplsLabel.label - pContext->pLabelSpaceTable[in_port].min_label)
                                      * sizeof(ix_cc_mpls_ilm_counters_internal));
          }
          ix_rm_atomic_sram_add(1, &pIlmCountersEntry->mplsInSegErrorPkts);
          ix_rm_atomic_sram_add(IX_NTOH16 (IX_RM_MEM_UINT16_READ (&pHwMeta->m_PacketSize)),
                                &pIlmCountersEntry->mplsInSegErrorOctets);
#endif
          PacketOutput     = IX_CC_MPLS_DROP_OUTPUT;
          PacketOutput_bid = IX_CC_MPLS_DROP_OUTPUT;

      } else {

          ix_uint8 oTTL = iTTL - (ix_uint8)Nhlfe.ttl;

          pHwMeta->m_FabricPort = (ix_uint8)Nhlfe.bladeId;
          pHwMeta->m_OutputPort = (ix_uint16)Nhlfe.outPortId;
          pHwMeta->m_NextHopID  = (ix_uint16)Nhlfe.nextHopId;

          
          if (Nhlfe.operation == IX_MPLS_LABELOP_POP_FORWARD) {

              _ix_cc_mpls_fwder_pop_handling(
                             (ix_cc_mpls_label_internal**) &pData,
                                                            pHwMeta,
                                                (ix_uint8)  Nhlfe.flags,
                                                (ix_uint8)  Nhlfe.l3Protocol,
                                                           &oTTL);

              PacketOutput     = IX_CC_MPLS_ILM_FORWARDER_OUTPUT;
              PacketOutput_bid = IX_CC_MPLS_ILM_FORWARDER_OUTPUT_BID;


          } else  if (Nhlfe.operation == IX_MPLS_LABELOP_SWAP_PUSH) {

              ix_uint8 oTTL_push;

              if ((Nhlfe.flags & MPLS_FLAG_TUNNEL_UNIFORM) == 0)
                  oTTL_push = IX_MPLS_PIPE_TTL;
              else
                  oTTL_push = oTTL;

              _ix_cc_mpls_fwder_swap_handling(pData,
                                             &Nhlfe,
                                              oTTL);

              _ix_cc_mpls_fwder_push_handling((ix_cc_mpls_label_internal**)&pData,
                                               pHwMeta,
                                              &Nhlfe,
                                               0,
                                               oTTL_push);

              PacketOutput     = IX_CC_MPLS_ILM_FORWARDER_OUTPUT;
              PacketOutput_bid = IX_CC_MPLS_ILM_FORWARDER_OUTPUT_BID;

          } else {
              
              _ix_cc_mpls_fwder_swap_handling(pData,
                                             &Nhlfe,
                                              oTTL);

              PacketOutput     = IX_CC_MPLS_ILM_FORWARDER_OUTPUT;
              PacketOutput_bid = IX_CC_MPLS_ILM_FORWARDER_OUTPUT_BID;
          } 

          break;
      }       
      if (PacketOutput == IX_CC_MPLS_DROP_OUTPUT)
      {
          /* Failed reserved label lookup */
#if defined (IX_MPLS_DEBUG)
          ix_ossl_message_log ("_ix_cc_mpls_ilm_forwarder(): Forwarding failed - dropping\n");
#endif

          break;
      }


  } while (NextLookup);


  /* FORWARDING FAILED */
  if (PacketOutput != IX_CC_MPLS_DROP_OUTPUT)
  {
      ix_uint16  packetSize;

      packetSize = IX_NTOH16 (IX_RM_MEM_UINT16_READ (&pHwMeta->m_PacketSize));

      if (packetSize > Nhlfe.maxLabPktSize  &&  PacketOutput != IX_CC_MPLS_ILM_FORWARDER_IPV4_OUTPUT)
      {
#if defined (MPLS_COUNTERS_FEATURE)
          ix_uint16 fragsNum;
          ix_uint32 fragsAddOctets;
#endif
          
#if defined (IX_MPLS_DEBUG)
          ix_ossl_message_log ("_ix_cc_mpls_ilm_forwarder(): Fragmentation\n");
#endif

          err = _ix_cc_mpls_fragmentation (arg_hBuffer,
                                           pData,
                                           pHwMeta,
                                           Nhlfe.maxLabPktSize,
                                           PacketOutput,
                                           PacketOutput_bid,
#if defined (MPLS_COUNTERS_FEATURE)
                                           &fragsNum,
                                           &fragsAddOctets,
#endif
                                           pContext);
          
#if defined (MPLS_COUNTERS_FEATURE)
          if (pIlmEntry->flags & MPLS_FLAG_SET_BIT)
          {
              ix_cc_mpls_ilm_set_internal *pIlmSetEntry = (ix_cc_mpls_ilm_set_internal*) pIlmEntry;
              
              pIlmCountersEntry =  (ix_cc_mpls_ilm_counters_internal*)((ix_uint8*)(pContext->pIlmNhlfeSetCounters) + 
                                    ((ix_uint32)(pIlmSetEntry->OutSegId1)) * sizeof(ix_cc_mpls_ilm_counters_internal));
          }
          else
          {
              pIlmCountersEntry =  (ix_cc_mpls_ilm_counters_internal*)((ix_uint8*)(pContext->pIlmCounters) + 
                                    pContext->pLabelSpaceTable[in_port].stats_offset +
                                   (MplsLabel.label - pContext->pLabelSpaceTable[in_port].min_label)
                                     * sizeof(ix_cc_mpls_ilm_counters_internal));
          }
          ix_rm_atomic_sram_add(fragsNum - 1, &pIlmCountersEntry->mplsInSegFragmPktDelta);
          ix_rm_atomic_sram_add(fragsAddOctets - 1, &pIlmCountersEntry->mplsInSegFragmOctDelta);
#endif
      }
      else
      {
          err = ix_rm_packet_send(PacketOutput,
                                  arg_hBuffer,
                                  PacketOutput_bid);
      }

      if (err != IX_SUCCESS) 
          return IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
                                  ("_ix_cc_mpls_ilm_forwarder(): failed to send packet to the WRED microblock"));
  }
  else
  {
      IX_ERROR_CRT(ix_rm_buffer_free_chain( arg_hBuffer ),
                   IX_CC_ERROR_INTERNAL, 0 );
  }


  return IX_SUCCESS;

} /* _ix_cc_mpls_ilm_forwarder */






/*
 * FUNCTION: _ix_cc_mpls_ftn_forwarder()  */
/**
 * ABSTRACT: implements mpls slow path FTN forwarding
 *
 *
 * \param  arg_pContext - a pointer to the MPLS context
 * \param  arg_hBuffer  - packet buffer handle.
 * \param  arg_icmpSend - specifies weather or not generate ICMP packets
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_ftn_forwarder(ix_buffer_handle    arg_hBuffer,
                          ix_icmp_send        arg_icmpSend,
                          void               *arg_pContext)
{

    ix_error        err = IX_SUCCESS;
    ix_uint32       HeaderType;
    
    ix_uint16       packetSize;

    void                      *pData      = NULL;
    ix_cc_mpls_nhlfe          *pNhlfe     = NULL;
    ix_hw_buffer_meta         *pHwMeta    = NULL;
    ix_cc_ipv4_header         *pIpv4Hdr;
    ix_cc_ipv6_header         *pIpv6Hdr;
    
    ix_cc_mpls_nhlfe_internal *pNhlfeEntry;
    ix_cc_mpls_context        *pContext = arg_pContext;

    /* Dispatch loop data */
    ix_uint32  iTTL; /* input  TTL */
    ix_uint32  oTTL; /* output TTL */

#if defined (IX_MPLS_DEBUG)
    ix_ossl_message_log ("_ix_cc_mpls_ftn_forwarder(): starting forwarding ...\n");
#endif

    /* Get data from buffer */
    err = ix_cc_hw_get_packet_data (arg_hBuffer, &pData);
    if (err != IX_SUCCESS) 
        return IX_ERROR_NEW (IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    /* Get meta data for buffer */
    err = ix_rm_buffer_get_meta (arg_hBuffer, (void **) &pHwMeta);
    if (err != IX_SUCCESS)
        return IX_ERROR_NEW (IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    /*** PHASE 1 ***/

    /* Check next hop id type */
    if (pHwMeta->m_NextHopIDType != NEXT_HOP_ID_MPLS_TYPE)
    {
#if defined (IX_MPLS_DEBUG)
        ix_ossl_message_log ("pHwMeta->m_NextHopIDType != NEXT_HOP_ID_MPLS_TYPE\n");
#endif
        
        return ix_rm_packet_send(IX_CC_MPLS_FTN_FORWARDER_OUTPUT,
                                arg_hBuffer,
                                IX_CC_MPLS_FTN_FORWARDER_OUTPUT_BID);
    }


#if defined (IX_MPLS_DEBUG)
    ix_ossl_message_log ("_ix_cc_mpls_ftn_forwarder(): got packet with NextHopID = %d \n ",
                         IX_NTOH16 (IX_RM_MEM_UINT16_READ(&pHwMeta->m_NextHopID)));
#endif

    
    pNhlfeEntry = &pContext->pNhlfeBaseAddr[pHwMeta->m_NextHopID];

#if defined (IX_MPLS_DEBUG)
    ix_ossl_message_log ("_ix_cc_mpls_ftn_forwarder(): taking NHLFE from 0x%x \n ",pNhlfeEntry);
#endif
    
    /* Check if the entry is valid */
    if ((pNhlfeEntry->flags & MPLS_FLAG_VALID_BIT) == 0)
    {

#if defined (IX_MPLS_DEBUG)
        ix_ossl_message_log("_ix_cc_mpls_ftn_forwarder(): Nhlfe is invalid\n");
#endif

        ix_rm_buffer_free_chain(arg_hBuffer);
        return IX_SUCCESS;
    }

    /* Check set flag */
    if (pNhlfeEntry->flags & MPLS_FLAG_SET_BIT != 0)
    {
        ix_cc_mpls_nhlfe_set_internal *pNhlfeSetEntry = (ix_cc_mpls_nhlfe_set_internal*) pNhlfeEntry;

#if defined (IX_MPLS_DEBUG)
        ix_ossl_message_log ("_ix_cc_mpls_ftn_forwarder(): this NHLFE is a set\n");
#endif

        pNhlfeEntry = &pContext->pNhlfeBaseAddr[((ix_uint32)(pNhlfeSetEntry->outSegId1))];

        if ((pNhlfeEntry ->flags & MPLS_FLAG_VALID_BIT) == 0)
        {

#if defined (IX_MPLS_DEBUG)
            ix_ossl_message_log("_ix_cc_mpls_ftn_forwarder(): NhlfeSet  subelement is invalid\n");
#endif

            ix_rm_buffer_free_chain(arg_hBuffer);
            return IX_SUCCESS;
        }
    }

#if defined (MPLS_COUNTERS_FEATURE)
    ix_rm_atomic_sram_add(1, &pContext->pFtnCounters[IX_NTOH16 (IX_RM_MEM_UINT16_READ(&pHwMeta->m_NextHopID))].mplsNhlfePkts);
    ix_rm_atomic_sram_add(IX_NTOH16 (IX_RM_MEM_UINT16_READ(&pHwMeta->m_PacketSize)),
                          &pContext->pFtnCounters[IX_NTOH16 (IX_RM_MEM_UINT16_READ(&pHwMeta->m_NextHopID))].mplsNhlfeOctets);
#endif
 

    /*** PHASE 2 ***/

    HeaderType = IX_RM_HW_META_GET_HEADER_TYPE(pHwMeta->m_BufferInfo);

    if (HeaderType == MPLS_IPV4_HEADER_TYPE) {
        pIpv4Hdr = pData;
        iTTL     = pIpv4Hdr->ttl;
    } else
    if (HeaderType == MPLS_IPV6_HEADER_TYPE) {
        pIpv6Hdr = pData;
        iTTL     = pIpv6Hdr->hopLimit;
    } else {
#if defined (IX_MPLS_DEBUG)
        ix_ossl_message_log("_ix_cc_mpls_ftn_forwarder(): unknown header type.\n");
#endif

        ix_rm_buffer_free_chain(arg_hBuffer);
        return IX_SUCCESS;
    }
 
    if (iTTL <= pNhlfeEntry->ttl)
    {
        err = _ix_cc_mpls_icmp_time_exceeded(arg_hBuffer,
                                             arg_pContext);
        ix_rm_buffer_free_chain(arg_hBuffer);

#if defined (MPLS_COUNTERS_FEATURE)
        ix_rm_atomic_sram_add(1, &pContext->pFtnCounters[IX_NTOH16 (IX_RM_MEM_UINT16_READ(&pHwMeta->m_NextHopID))].mplsNhlfeErrorPkts);
        ix_rm_atomic_sram_add(IX_NTOH16 (IX_RM_MEM_UINT16_READ(&pHwMeta->m_PacketSize)),
                              &pContext->pFtnCounters[IX_NTOH16 (IX_RM_MEM_UINT16_READ(&pHwMeta->m_NextHopID))].mplsNhlfeErrorOctets);
#endif
        return IX_SUCCESS;
    }

    oTTL = iTTL - pNhlfeEntry->ttl;

    if ((pNhlfeEntry->flags & MPLS_FLAG_TUNNEL_MASK) != MPLS_FLAG_TUNNEL_UNIFORM)
    {
        if (HeaderType == MPLS_IPV4_HEADER_TYPE) 
        {   
            pIpv4Hdr->ttl = (ix_uint8)oTTL;
            pIpv4Hdr->checksum = 0;
            pIpv4Hdr->checksum = _ix_cc_ipv4_calculate_checksum (
                                   (ix_uint8*)pIpv4Hdr,
                                   pIpv4Hdr->headerLength * 4);
        } else if (HeaderType == MPLS_IPV6_HEADER_TYPE) {
            pIpv6Hdr->hopLimit = (ix_uint8)oTTL;
        }

        oTTL = IX_MPLS_PIPE_TTL;
    }

    _ix_cc_mpls_fwder_push_handling((ix_cc_mpls_label_internal**)&pData,
                                     pHwMeta,
                                     pNhlfeEntry,
                                     1,
                                     oTTL);


    IX_RM_HW_META_SET_HEADER_TYPE(pHwMeta->m_BufferInfo,
                                  MPLS_MPLS_HEADER_TYPE);

    packetSize = IX_NTOH16 (IX_RM_MEM_UINT16_READ (&pHwMeta->m_PacketSize));


    if (packetSize > pNhlfeEntry->maxLabPktSize)
    {
#if defined (MPLS_COUNTERS_FEATURE)
          ix_uint16 fragsNum;
          ix_uint32 fragsAddOctets;
#endif
          err = _ix_cc_mpls_fragmentation(arg_hBuffer,
                                           pData,
                                           pHwMeta,
                                           pNhlfeEntry->maxLabPktSize,
                                           IX_CC_MPLS_ILM_FORWARDER_OUTPUT,
                                           IX_CC_MPLS_FTN_FORWARDER_OUTPUT_BID,
#if defined (MPLS_COUNTERS_FEATURE)
                                          &fragsNum,
                                          &fragsAddOctets, 
#endif
                                           arg_pContext);

#if defined (MPLS_COUNTERS_FEATURE)
          ix_rm_atomic_sram_add(fragsNum,
                                &pContext->pFtnCounters[IX_NTOH16 (IX_RM_MEM_UINT16_READ(&pHwMeta->m_NextHopID))].mplsNhlfeFragmPktsDelta);
          ix_rm_atomic_sram_add(fragsAddOctets,
                                &pContext->pFtnCounters[IX_NTOH16 (IX_RM_MEM_UINT16_READ(&pHwMeta->m_NextHopID))].mplsNhlfeFragmOctetsDelta);
#endif

    } 
    else 
    {
        err = ix_rm_packet_send(IX_CC_MPLS_FTN_FORWARDER_OUTPUT,
                                arg_hBuffer,
                                IX_CC_MPLS_FTN_FORWARDER_OUTPUT_BID);

    }

    if (err != IX_SUCCESS)
    {

#if defined (MPLS_COUNTERS_FEATURE)
        ix_rm_atomic_sram_add(1, &pContext->pFtnCounters[IX_NTOH16 (IX_RM_MEM_UINT16_READ(&pHwMeta->m_NextHopID))].mplsNhlfeErrorPkts);
        ix_rm_atomic_sram_add(IX_NTOH16 (IX_RM_MEM_UINT16_READ(&pHwMeta->m_PacketSize)),
                              &pContext->pFtnCounters[IX_NTOH16 (IX_RM_MEM_UINT16_READ(&pHwMeta->m_NextHopID))].mplsNhlfeErrorOctets);
#endif
        
        return IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
                                ("_ix_cc_mpls_ilm_forwarder(): failed to send packet to the WRED microblock"));
    }

    return IX_SUCCESS;

} /* _ix_cc_mpls_ftn_forwarder */








/*
 * FUNCTION: _ix_cc_mpls_reserved_label_handling()  */
/**
 * ABSTRACT: function deals with MPLS packets with reserved labels
 *
 * \param  arg_ppMplsLabel - pointer to dram area with mpls label stack
 * \param  arg_pHwMeta     - pointer to the meta data of the packet
 * \param  arg_pContext    - a pointer to the MPLS context
 *
 * \return info if the next go in the loop is needed
 * \retval 0 - if this is packet to be dropped
 * \retval 1 - if label was popped and next run in the pop loop is required.
 */
PRIVATE ix_uint8
_ix_cc_mpls_reserved_label_handling(ix_cc_mpls_label_internal **arg_ppMplsLabel,
                                    ix_hw_buffer_meta          *arg_pHwMeta)

{
  ix_cc_mpls_label_internal        MplsLabel;

  /* Read the top mpls label */
  ix_ossl_memcpy(&MplsLabel, (*arg_ppMplsLabel), sizeof(ix_cc_mpls_label_internal));

  if ((MplsLabel.label == 0) & (MplsLabel.s == 1))
  {
      _ix_cc_mpls_fwder_pop_handling(arg_ppMplsLabel,
                                     arg_pHwMeta,
                                     MPLS_FLAG_TUNNEL_PIPE,
                                     MPLS_L3_IP_V4,
                                     0);
      return 1;

  } else if ((MplsLabel.label == 2) & (MplsLabel.s == 1)) {

      _ix_cc_mpls_fwder_pop_handling(arg_ppMplsLabel,
                                     arg_pHwMeta,
                                     MPLS_FLAG_TUNNEL_PIPE,
                                     MPLS_L3_IP_V6,
                                     0);
      return 2;
  
  } else   if ((MplsLabel.label == 1) & (MplsLabel.s == 1)) {

  } else {
      
  }

  return 0;
}





PRIVATE ix_uint32
_ix_cc_mpls_fwder_pop_handling(ix_cc_mpls_label_internal **arg_ppMplsLabel,
                               ix_hw_buffer_meta          *arg_pHwMeta,
                               ix_uint32                   arg_NhlfeFlags,
                               ix_uint8                    arp_l3protocol,
                               ix_uint8                   *arg_pNewTTL)
{
    ix_cc_mpls_label_internal MplsLabel;
    ix_uint16                 packetOffset;
    ix_uint16                 packetSize;
    ix_uint16                 bufferSize;
    ix_uint32                 EndofStack;

   /* Pop one label, update packet size and offset */
    packetOffset = IX_NTOH16 (IX_RM_MEM_UINT16_READ (&arg_pHwMeta->m_Offset));
    packetSize   = IX_NTOH16 (IX_RM_MEM_UINT16_READ (&arg_pHwMeta->m_PacketSize));
    bufferSize   = IX_NTOH16 (IX_RM_MEM_UINT16_READ (&arg_pHwMeta->m_BufferSize));

    /* Move header pointer to the next label */
    EndofStack = (**arg_ppMplsLabel).s;
    (*arg_ppMplsLabel)++;

    packetOffset += sizeof (ix_cc_mpls_label_internal);
    packetSize   -= sizeof (ix_cc_mpls_label_internal);
    bufferSize   -= sizeof (ix_cc_mpls_label_internal);

    /* Save meta data */
    IX_RM_MEM_UINT16_WRITE (&arg_pHwMeta->m_PacketSize, IX_HTON16 (packetSize));
    IX_RM_MEM_UINT16_WRITE (&arg_pHwMeta->m_Offset,     IX_HTON16 (packetOffset));
    IX_RM_MEM_UINT16_WRITE (&arg_pHwMeta->m_BufferSize, IX_HTON16 (bufferSize));

    /* Check bottom of stack and pop label */
        
    if (EndofStack == 1) {

        ix_cc_ipv4_header *pIpv4Hdr = (ix_cc_ipv4_header*)(*arg_ppMplsLabel);

        switch(arp_l3protocol) {

             case MPLS_L3_IP_V4:
                    IX_RM_HW_META_SET_HEADER_TYPE(arg_pHwMeta->m_BufferInfo,
                                                  MPLS_IPV4_HEADER_TYPE);

                    if (arg_NhlfeFlags & MPLS_FLAG_TUNNEL_UNIFORM)
                    {
                         pIpv4Hdr->ttl      = (ix_uint8)*arg_pNewTTL;
                         pIpv4Hdr->checksum = 0;
                         pIpv4Hdr->checksum = _ix_cc_ipv4_calculate_checksum (
                                              (ix_uint8*)pIpv4Hdr,
                                              pIpv4Hdr->headerLength * 4);
                    } else {
                         *arg_pNewTTL        = pIpv4Hdr->ttl;
                    }
                    break;


             case MPLS_L3_IP_V6:
                    IX_RM_HW_META_SET_HEADER_TYPE(arg_pHwMeta->m_BufferInfo,
                                                  MPLS_IPV6_HEADER_TYPE);
                    break;
        }
        
        return 0;

    } else {
        ix_ossl_memcpy(&MplsLabel, *arg_ppMplsLabel, sizeof(ix_cc_mpls_label_internal));
        if (arg_NhlfeFlags & MPLS_FLAG_TUNNEL_UNIFORM) {
            MplsLabel.ttl = *arg_pNewTTL;
            ix_ossl_memcpy(*arg_ppMplsLabel, &MplsLabel, sizeof(ix_cc_mpls_label_internal));
        } else {
            *arg_pNewTTL = MplsLabel.ttl;
        }

    }
    return 1;
}



PRIVATE void
_ix_cc_mpls_fwder_swap_handling(ix_cc_mpls_label_internal *arg_pMplsLabel,
                                ix_cc_mpls_nhlfe_internal *arg_pNhlfe,
                                ix_uint32                  arg_oTTL)
{
    ix_cc_mpls_label_internal MplsLabel;

    #if defined (IX_MPLS_DEBUG)
    ix_ossl_message_log ("_ix_cc_mpls_fwder_swap_handling(): processing\n");
    #endif

    ix_ossl_memcpy(&MplsLabel, arg_pMplsLabel, sizeof(ix_cc_mpls_label_internal));


    if (arg_pNhlfe->operation == IX_MPLS_LABELOP_SWAP)
        /* Swap bottom label */
        MplsLabel.label  = arg_pNhlfe->topLabel;
    else if (arg_pNhlfe->operation == IX_MPLS_LABELOP_SWAP_PUSH) 
    {
	switch (arg_pNhlfe->pushCount)
        {
        case 0:
           /* Swap bottom label */
           MplsLabel.label  = arg_pNhlfe->topLabel;
           break;
        case 1:
           /* Swap bottom label */
           MplsLabel.label  = arg_pNhlfe->label1;
           break;
        case 2:
           /* Swap bottom label */
           MplsLabel.label  = arg_pNhlfe->label2;
           break;
        case 3:
           /* Swap bottom label */
           MplsLabel.label  = arg_pNhlfe->label3;
           break;
        }
    }
    MplsLabel.ttl = arg_oTTL;

    ix_ossl_memcpy(arg_pMplsLabel, &MplsLabel, sizeof(ix_cc_mpls_label_internal));

    return;
}





PRIVATE void
_ix_cc_mpls_fwder_push_handling(ix_cc_mpls_label_internal **arg_ppMplsLabel,
                                ix_hw_buffer_meta          *arg_pHwMeta,
                                ix_cc_mpls_nhlfe_internal  *arg_pNhlfe,
                                ix_uint8                    arg_EndofStack,
                                ix_uint32                   arg_oTTL)
{
    ix_uint8                  pushCount = (ix_uint8)arg_pNhlfe->pushCount;
    ix_cc_mpls_label_internal MplsLabelStack[4];


    #if defined (IX_MPLS_DEBUG)
    ix_ossl_message_log ("_ix_cc_mpls_fwder_push_handling(): processing\n");
    #endif

    if (pushCount < 1) return;

    #if defined (IX_MPLS_DEBUG)
    ix_ossl_message_log ("pushCount = 0x%x\n", (int) pushCount);
    #endif

    switch (pushCount) {
        case 4:
          MplsLabelStack[3].label = arg_pNhlfe->label3;
          MplsLabelStack[3].ttl   = arg_oTTL;
          MplsLabelStack[3].s     = arg_EndofStack;
        case 3:
          MplsLabelStack[2].label = arg_pNhlfe->label2;
          MplsLabelStack[2].ttl   = arg_oTTL;
          MplsLabelStack[2].s     = (pushCount > 3) ? 0:arg_EndofStack;
        case 2:          
          MplsLabelStack[1].label = arg_pNhlfe->label1;
          MplsLabelStack[1].ttl   = arg_oTTL;
          MplsLabelStack[1].s     = (pushCount > 2) ? 0:arg_EndofStack;
        case 1:
          MplsLabelStack[0].label = arg_pNhlfe->topLabel;
          MplsLabelStack[0].ttl   = arg_oTTL;    
          MplsLabelStack[0].s     = (pushCount > 1) ? 0:arg_EndofStack;
        default:
          ;
    }

    _ix_cc_mpls_encap (arg_ppMplsLabel, 
                       arg_pHwMeta,
                       pushCount,
                       MplsLabelStack);

    return;
}





/*
 * FUNCTION: _ix_cc_mpls_lsp_copy_ilm_to_nhlfe()  */
/**
 * ABSTRACT: copy the content of ilm entry to nhlfe
 *
 * \param  arg_pNhlfe    - a pointer to the NHLFE Table context
 * \param  arg_LspHandle - a handle identifying an CC LSP entry.
 *
 */
PRIVATE void
_ix_cc_mpls_lsp_copy_ilm_to_nhlfe(ix_cc_mpls_ilm_internal   *arg_pIlmEntry,
                                  ix_cc_mpls_nhlfe_internal *arg_pNhlfe)
{
   arg_pNhlfe->bladeId       = arg_pIlmEntry->bladeId;
   arg_pNhlfe->exp2phb       = arg_pIlmEntry->exp2phb;
   arg_pNhlfe->phb2exp       = arg_pIlmEntry->phb2exp;
   arg_pNhlfe->l3Protocol    = arg_pIlmEntry->l3Protocol;
   arg_pNhlfe->label1        = arg_pIlmEntry->label1;
   arg_pNhlfe->label2        = arg_pIlmEntry->label2;
   arg_pNhlfe->label3        = arg_pIlmEntry->label3;
   arg_pNhlfe->maxLabPktSize = arg_pIlmEntry->maxLabPktSize;
   arg_pNhlfe->nextHopId     = arg_pIlmEntry->nextHopId;
   arg_pNhlfe->operation     = arg_pIlmEntry->operation;
   arg_pNhlfe->outPortId     = arg_pIlmEntry->outPortId;
   arg_pNhlfe->outPortType   = arg_pIlmEntry->outPortType;
   arg_pNhlfe->pushCount     = arg_pIlmEntry->pushCount;
   arg_pNhlfe->topLabel      = arg_pIlmEntry->topLabel;
   arg_pNhlfe->ttl           = arg_pIlmEntry->ttl;
   arg_pNhlfe->flags         = arg_pIlmEntry->flags;
}



