/*
 * ============================================================================
 * COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *                                                                          */
/** \file      ix_cc_mpls_icmp.c
 *
 *  \brief     The file contains functions to handle icmp generation
 *
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_mpls_icmp.c,v 1.11.2.1 2004/03/12 10:40:15 wneubaue Exp $";


#include <ix_error.h>
#include <ix_types.h>

/* include IXA SDK framework */
#include <ix_ossl.h>
#include "ix_rm.h"

#if !defined(IX_EXCLUDE_CCI)
#include "ix_cci.h"
#endif 

#include "definitions.h"
#include "ix_netmacros.h"

/* include CC support */
#include "ix_cc_error.h"
#include "ix_cc.h"

#include "bindings.h"

#include "cc/ix_cc_rtmv4.h" 
#include "cc/ipv4/internal/icmp.h"

#include "../../include/mpls_defs.h"
#include "cc/ix_cc_mpls.h"
#include "cc/internal/ix_cc_mpls_icmp.h"
#include "cc/internal/ix_cc_mpls_lsp.h"
#include "cc/internal/ix_cc_mpls_nhlfe.h"
#include "cc/internal/ix_cc_mpls_fwder.h"
#include "cc/internal/ix_cc_mpls_internal.h"


/**
 * Private functions declarations
 */

/* This macro returns the no. of items currently in queue */
#define IX_CC_MPLS_ICMP_GET_QUEUE_SIZE(arg_HeadIdx,                              \
                                       arg_TailIdx,                              \
                                       arg_MaxSize)                              \
                                                                                 \
((arg_HeadIdx >= arg_TailIdx) ? (arg_HeadIdx - arg_TailIdx)                \
                                 : (arg_MaxSize + arg_HeadIdx - arg_TailIdx))



/* This macro checks if queue is empty */
#define IX_CC_MPLS_ICMP_IS_QUEUE_EMPTY(arg_HeadIdx, \
                                       arg_TailIdx) \
        (arg_HeadIdx == arg_TailIdx)


/* This macro checks if queue is full */
#define IX_CC_MPLS_ICMP_IS_QUEUE_FULL(arg_HeadIdx, \
                                      arg_TailIdx, \
                                      arg_MaxSize) \
        ((IX_CC_MPLS_ICMP_GET_QUEUE_SIZE(arg_HeadIdx, arg_TailIdx, arg_MaxSize)) + 1 == arg_MaxSize)




PRIVATE ix_error
_ix_cc_mpls_icmp_queue_init (ix_cc_mpls_context *arg_pMplsContext);



PRIVATE ix_error
_ix_cc_mpls_icmp_build_message(ix_uint8            arg_Type,
                               ix_uint8            arg_Code,
                               ix_uint32           arg_Param,
                               ix_buffer_handle    arg_hBuffer,
                               ix_cc_mpls_context *arg_pContext);

PRIVATE ix_uint32
_ix_cc_mpls_icmp_queue_enqueue (ix_buffer_handle     arg_hBuffer,
                                ix_cc_mpls_context  *arg_pMplsContext);

PRIVATE ix_uint32
_ix_cc_mpls_icmp_queue_dequeue (ix_buffer_handle   *arg_phBuffer,
                                ix_cc_mpls_context *arg_pMplsContext);





/*
 * FUNCTION: _ix_cc_mpls_icmp_init()  */
/**
 * ABSTRACT: Initialization functions to generate ICMP 
 *
 * DESCRIPTION: creates queue for icmp messages.
 *
 * \param  arg_pContext - a pointer to the MPLS context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_icmp_init(void *arg_pContext)
{
  ix_error            err;
  ix_cc_mpls_context *pContext  = (ix_cc_mpls_context*)arg_pContext;
  ix_uint32           icmpQSize = pContext->icmpQueueMaxSize * sizeof (ix_buffer_handle);

  pContext->icmpQheadIdx   = 0;
  pContext->icmpQtailIdx   = 0;
  pContext->icmpQueueDrain = pContext->icmpQueueMaxSize;

  pContext->icmp_ipID = 0x6000;

  /* Create mutex for critical section. */
  IX_ERROR_C(ix_ossl_mutex_init(OS_MUTEX_UNLOCK, &(pContext->ICMPmutexId)),err);
  if (err != IX_SUCCESS)
  {
      ix_ossl_free(pContext);
      return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_OSSL_MUTEX_INIT,
             ("_ix_cc_mpls_icmp_init(): can't initialize mutex"));
  }

  /* Allocate memory for pIcmpQueue */
  pContext->pIcmpQueue = (ix_buffer_handle *)ix_ossl_malloc (icmpQSize);
  
  if (pContext->pIcmpQueue  == NULL)
    return (IX_ERROR_REMOTE (IX_CC_ERROR_OOM_SYSTEM,
			     ("_ix_cc_mpls_icmp_init(): Not enough memory to create ICMP queue")));

  /* Initialize the structure */
  ix_ossl_memset ((void *) pContext->pIcmpQueue, 0, icmpQSize);

  return IX_SUCCESS;
}





/*
 * FUNCTION: _ix_cc_mpls_icmp_fini()  */
/**
 * ABSTRACT: removes resources taken to generate ICMP 
 *
 * DESCRIPTION: releases the queue for icmp messages.
 *
 * \param  arg_pContext - a pointer to the MPLS context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_icmp_fini(void *arg_pContext)
{
  ix_error            err;
  ix_buffer_handle    hBuffer;
  ix_cc_mpls_context *pContext  = (ix_cc_mpls_context*)arg_pContext;
  ix_uint32           icmpQSize = pContext->icmpQueueMaxSize * sizeof (ix_buffer_handle);

  IX_ERROR_CRT(ix_ossl_mutex_lock(pContext->ICMPmutexId, IX_OSSL_WAIT_FOREVER),
               IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK, IX_ERROR_LEVEL_LOCAL);

  #if defined (IX_MPLS_DEBUG)
  ix_ossl_message_log ("_ix_cc_mpls_icmp_fini(): purging ICMP queue\n");
  #endif
  while (!IX_CC_MPLS_ICMP_IS_QUEUE_EMPTY(pContext->icmpQheadIdx,
                                         pContext->icmpQtailIdx))
  {
    if (_ix_cc_mpls_icmp_queue_dequeue (&hBuffer, pContext) == 0)
        _ix_cc_mpls_ftn_forwarder (hBuffer, IX_CC_MPLS_DO_NOT_SEND_ICMP, pContext);

  }

  pContext->icmpQueueDrain   = 0;

  /* Allocate memory for pIcmpQueue */
  ix_ossl_free(pContext->pIcmpQueue);
  
  pContext->pIcmpQueue = NULL;

  IX_ERROR_CT(ix_ossl_mutex_unlock(pContext->ICMPmutexId), err,
               IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK, IX_ERROR_LEVEL_LOCAL);


  IX_ERROR_CRT(ix_ossl_mutex_fini(pContext->ICMPmutexId),
               IX_CC_MPLS_ERROR_OSSL_MUTEX_FINI, 
               IX_ERROR_LEVEL_LOCAL);

  return IX_SUCCESS;
}






/*
 * FUNCTION: _ix_cc_mpls_icmp_queue_enqueue()  */
/**
 * ABSTRACT: enqueues ICMP packet. increases  arg_pMplsContext->icmpQhead.
 *
 * \param  arg_pContext - a pointer to the MPLS context
 * \param  arg_hBuffer  - the buffer handle to be put in the ICMP queue
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \retval 1 if operation has been unsuccessful because ICMP queue was full.
 * \retval 0 if operation has been successful.
 */
PRIVATE ix_uint32
_ix_cc_mpls_icmp_queue_enqueue (ix_buffer_handle    arg_hBuffer,
                                ix_cc_mpls_context *arg_pContext)
{

  if (IX_CC_MPLS_ICMP_IS_QUEUE_FULL(arg_pContext->icmpQheadIdx,
                                    arg_pContext->icmpQtailIdx,
                                    arg_pContext->icmpQueueMaxSize))
  {
      #if defined (IX_MPLS_DEBUG)
      ix_ossl_message_log ("_ix_cc_mpls_icmp_queue_enqueue(): operation failed\n");
      #endif
      return 1;
  }


  arg_pContext->pIcmpQueue[arg_pContext->icmpQheadIdx] = arg_hBuffer;
  arg_pContext->icmpQheadIdx = (++arg_pContext->icmpQheadIdx) %
                                       (arg_pContext->icmpQueueMaxSize);

  #if defined (IX_MPLS_DEBUG)
  {
      void           *pData;
      ix_ossl_message_log ("_ix_cc_mpls_icmp_queue_enqueue(): operation succeeded\n");
      IX_ERROR_CRT (ix_cc_hw_get_packet_data (arg_hBuffer, &pData),
                    IX_CC_ERROR_INTERNAL, 0);

      show_mem (pData, 100);

      _ix_cc_mpls_icmp_print_queue (arg_pContext);
  }
  #endif /* IX_MPLS_DEBUG */

  return 0;

}	/* _ix_cc_mpls_icmp_queue_enqueue */






/*
 * FUNCTION: _ix_cc_mpls_icmp_queue_dequeue ()  */
/**
 * ABSTRACT: dequeues ICMP packet. increases  arg_pMplsContext->icmpQtail.
 *
 * \param  arg_pContext - a pointer to the MPLS context
 * \param  arg_phBuffer - a pointer to the buffer handle taken from the queue
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \retval 1 if operation has been unsuccessful because ICMP queue was empty.
 * \retval 0 if operation has been successful.
 */
PRIVATE ix_uint32
_ix_cc_mpls_icmp_queue_dequeue (ix_buffer_handle   *arg_phBuffer,
                                ix_cc_mpls_context *arg_pContext)
{

  if (IX_CC_MPLS_ICMP_IS_QUEUE_EMPTY(arg_pContext->icmpQheadIdx,
                                     arg_pContext->icmpQtailIdx))
  {
      #if defined (IX_MPLS_DEBUG)
      ix_ossl_message_log ("_ix_cc_mpls_icmp_queue_dequeue (): operation failed\n");
      #endif
      return 1;
  }

  *arg_phBuffer = arg_pContext->pIcmpQueue[arg_pContext->icmpQtailIdx];
  arg_pContext->icmpQtailIdx = (++arg_pContext->icmpQtailIdx) % 
                                      (arg_pContext->icmpQueueMaxSize);


  #if defined (IX_MPLS_DEBUG)
  {
      void              *pData   = NULL;
      ix_hw_buffer_meta *pHwMeta = NULL;

      ix_ossl_message_log ("_ix_cc_mpls_icmp_queue_dequeue(): operation succeeded;"
                           "*arg_phBuffer = 0x%08x\n", *arg_phBuffer);

      /* Get meta data for buffer */
      ix_rm_buffer_get_meta (*arg_phBuffer, (void **) &pHwMeta);

      ix_cc_hw_get_packet_data ((*arg_phBuffer), &pData);
      show_metadata(pHwMeta);
      show_mem (pData, pHwMeta->m_BufferSize);
  }

  #endif /* IX_MPLS_DEBUG */
  
  return 0;

}	/* _ix_cc_mpls_icmp_queue_dequeue */








/*
 * FUNCTION: _ix_cc_mpls_icmp_print_queue()  */
/**
 * ABSTRACT: prints the content of ICMP queue
 *
 * \param  arg_pContext - a pointer to the MPLS context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
#if defined (IX_MPLS_DEBUG)
void
_ix_cc_mpls_icmp_print_queue (void *arg_pContext)
{
  ix_cc_mpls_context  *pContext = arg_pContext;

  register ix_uint32 i = 0;
  register ix_uint32 j = 0;

  ix_error    err;
  ix_uint32   size;

  IX_ERROR_C(ix_ossl_mutex_lock(pContext->ICMPmutexId, IX_OSSL_WAIT_FOREVER), err);
  if (err != IX_SUCCESS)
    _ix_cc_mpls_icmp_print_queue ("_ix_cc_mpls_icmp_print_queue() can't lock mutext I won't show anything\n");


  size = IX_CC_MPLS_ICMP_GET_QUEUE_SIZE(pContext->icmpQheadIdx,
				                        pContext->icmpQtailIdx,
                                        pContext->icmpQueueMaxSize);

  ix_ossl_message_log ("Head: %d\n", pContext->icmpQheadIdx);
  ix_ossl_message_log ("Tail: %d\n", pContext->icmpQtailIdx);

  while (i == pContext->icmpQheadIdx)
  {
    void   *pData;

    ix_ossl_message_log("\nelement %d \n", j++);
    
    err = ix_cc_hw_get_packet_data(pContext->pIcmpQueue[i], &pData);

    if (err != IX_SUCCESS) 
       ix_ossl_message_log("_ix_cc_mpls_icmp_print_queue(): ix_cc_hw_get_packet_data failed - nothing to show \n");
    else
       show_mem (pData, 100);

    i = (i+1) % (pContext->icmpQueueMaxSize);
  }

  IX_ERROR_C(ix_ossl_mutex_unlock(pContext->ICMPmutexId), err);

  if (err != IX_SUCCESS)
      _ix_cc_mpls_icmp_print_queue ("_ix_cc_mpls_icmp_print_queue() can't unlock mutext !!!\n");

  return;
} /* _ix_cc_mpls_icmp_print_queue */

#endif








/*
 * FUNCTION: _ix_cc_mpls_icmp_dest_unreachable ()  */
/**
 * ABSTRACT: builds icmp packet denoting destination unreachability
 *
 * DESCRIPTION: This function attempts to build ICMP dest unreaacheable
 *              error message for the packet passed in as buffer handle
 *
 * \param  arg_pContext - a pointer to the MPLS context
 * \param  arg_hBuffer  - packet buffer handle.
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_icmp_dest_unreachable(ix_buffer_handle    arg_hBuffer,
                                  ix_uint8            arg_Reason,
                                  ix_uint32           arg_Param,
                                  void               *arg_pContext)
{

  ix_error err = IX_SUCCESS;
  ix_cc_mpls_context *pContext  = (ix_cc_mpls_context*)arg_pContext;

  return _ix_cc_mpls_icmp_build_message(IX_ICMP_UNREACH,
                                        arg_Reason,
                                        arg_Param,
                                        arg_hBuffer,
                                        pContext);
} /* _ix_cc_mpls_icmp_dest_unreachable */







/*
 * FUNCTION: _ix_cc_mpls_icmp_time_exceeded ()  */
/**
 * ABSTRACT: builds icmp packet denoting destination unreachability
 *
 * DESCRIPTION: This function attempts to build ICMP
 *              time Exceeded error message for the packet passed in
 *              as buffer handle
 *
 * \param  arg_pContext - a pointer to the MPLS context
 * \param  arg_hBuffer  - packet buffer handle.
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_icmp_time_exceeded(ix_buffer_handle   arg_hBuffer,
                               void              *arg_pContext)
{
  ix_error err = IX_SUCCESS;
  ix_cc_mpls_context *pContext  = (ix_cc_mpls_context*)arg_pContext;

  return _ix_cc_mpls_icmp_build_message(IX_ICMP_TIMEXCEED,
                                        IX_ICMP_TIMEXCEED_INTRANS, 0,
                                        arg_hBuffer, pContext);
} /* _ix_cc_mpls_icmp_time_exceeded */







/*
 * FUNCTION: _ix_cc_mpls_icmp_build_message()  */
/**
 * ABSTRACT: builds and queues new ICMP message
 *
 * DESCRIPTION: This function builds a complete ICMP message and
 *              queues it so it can be sent.
 *
 * \param  arg_Type     -
 * \param  arg_Code     -
 * \param  arg_Param    -
 * \param  arg_pContext - a pointer to the MPLS context
 * \param  arg_hBuffer  - packet buffer handle.
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
PRIVATE ix_error
_ix_cc_mpls_icmp_build_message (ix_uint8            arg_Type,
                                ix_uint8            arg_Code,
                                ix_uint32           arg_Param,
                                ix_buffer_handle    arg_hBuffer,
                                ix_cc_mpls_context *arg_pContext)
{
    ix_error        err = IX_SUCCESS;
    ix_uint32       isEOP;
    
    ix_cc_ipv4_header         inIpHdr;
    ix_hw_buffer_meta        *pHwMeta     = NULL;
    ix_cc_rtmv4_next_hop_info nexthopInfo;
    ix_uint8                  icmpMessage[IX_CC_IPV4_IP_HDR_PAYLOAD_LIMIT];
    ix_cc_mpls_label_internal mplsLabelStack[MAX_MPLS_LABS];
    ix_cc_ipv4_header        *pOutIpHdr   = (ix_cc_ipv4_header *) icmpMessage;
    ix_cc_ipv4_icmp_header   *pOutIcmpHdr = (ix_cc_ipv4_icmp_header *) (icmpMessage + MIN_IP_LEN);


    ix_uint16       inPort;
    void           *pData = NULL;
    ix_uint32       payloadLength;
    ix_uint16       length;
    ix_uint8        HeaderType;
    ix_uint32       ipAddr;
    ix_uint32       cell_count = 0;
    ix_uint16       labelNum = 0;

    #if defined(IX_MPLS_DEBUG)
    ix_ossl_message_log ("_ix_cc_mpls_icmp_build_message(): building ICMP packet\n");
    #endif


   /* Get data from buffer */
    err = ix_cc_hw_get_packet_data (arg_hBuffer, &pData);
    if (err != IX_SUCCESS)
         return IX_ERROR_NEW (IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);


   /* Get meta data for buffer */

    err = ix_rm_buffer_get_meta (arg_hBuffer, (void **) &pHwMeta);

    if (err != IX_SUCCESS)
         return IX_ERROR_NEW (IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    HeaderType = IX_RM_HW_META_GET_HEADER_TYPE(pHwMeta->m_BufferInfo);

   /* Get the incoming IP header from the incoming packet */
   /* Allocate the outgoing ICMP message */
   /* Initialize outgoing ICMP message headers */
   /* Copy data from incoming packet to outgoing ICMP message */

    if (HeaderType == MPLS_MPLS_HEADER_TYPE) {
 
        #if defined(IX_MPLS_DEBUG)
        ix_ossl_message_log ("stripping all llabels\n");
        #endif

        err = _ix_cc_mpls_decap ((ix_cc_mpls_label_internal**)(&pData),
                                  pHwMeta,
                                  mplsLabelStack,
                                 &labelNum,
                                  (ix_uint32) MAX_MPLS_LABS);

        if (err != IX_SUCCESS) 
            return err;

        IX_RM_HW_META_SET_HEADER_TYPE(pHwMeta->m_BufferInfo, MPLS_IPV4_HEADER_TYPE);


    } else if (HeaderType != MPLS_IPV4_HEADER_TYPE) {

        #if defined(IX_MPLS_DEBUG)
        ix_ossl_message_log ("Packet of unknown type.\n");
        #endif
        return IX_ERROR_NEW (IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
    }

   /* Note m_InputPort is ix_uint16 and is in Network byte order */
    inPort = IX_NTOH16 (pHwMeta->m_InputPort);

    #if defined (IX_MPLS_DEBUG)
    ix_ossl_message_log ("_ix_cc_mpls_icmp_build_message(): mpls label removed\n");
    show_mem ((ix_uint32 *) pData, 0x40);
    #endif

    _ix_cc_ipv4_init_ip_header ((ix_uint8 *) pData, &inIpHdr);

    #if defined (IX_MPLS_DEBUG)
    ix_ossl_message_log ("_ix_cc_mpls_icmp_build_message() :ip header initialized\n");
    show_mem ((ix_uint32 *) &inIpHdr, 0x40);
    #endif


    #if (IX_BYTE_ORDER == IX_BIG_ENDIAN)

   /* Check ipv4 header flags field. Do not send ICMP for fragments. */
    if (inIpHdr.flags & 1) {
        #if defined(IX_MPLS_DEBUG)
        ix_ossl_message_log ("No route to the source. Dropping ICMP packet.\n");
        #endif
        return IX_ERROR_NEW (IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
    }
    #endif


   /* Initialize an ICMP header to have type, code and param field
    * as passed in the argument */

    pOutIcmpHdr->type = arg_Type;
    pOutIcmpHdr->code = arg_Code;
    pOutIcmpHdr->extra = IX_HTON32 (arg_Param);
    pOutIcmpHdr->checksum = 0;

    ix_ossl_memcpy (icmpMessage + MIN_IP_LEN + IX_ICMP_HDR_LEN,
		    (ix_uint8 *) pData, (inIpHdr.headerLength * 4) + 8);

    payloadLength = (inIpHdr.headerLength * 4) + 8;

    pOutIpHdr->version      = 4;
    pOutIpHdr->headerLength = 5;
    pOutIpHdr->tos          = 0;
    pOutIpHdr->ttl          = ICMP_MPLS_TTL;
    pOutIpHdr->protocol     = IX_IP_PROTOICMP;

    #if (IX_BYTE_ORDER == IX_BIG_ENDIAN)
    pOutIpHdr->flags = 0;
    pOutIpHdr->offset = 0;
    #else
    pOutIpHdr->fragOffset = 0;
    #endif

    IX_ERROR_CRT (_ix_cc_ipv4_get_port_ip_address
		  ((ix_uint16) arg_pContext->localBladeId, inPort, &ipAddr),
		  IX_CC_ERROR_INTERNAL, 0);

   /* propData returns IP address in host byte order 
    * Before storing the source IP address in packet, it has
    * to be converted to network byte order \
    */

    pOutIpHdr->sourceAddress = IX_HTON32 (ipAddr);
    pOutIpHdr->destAddress = inIpHdr.sourceAddress;

    /** ID is 16 bits identifying value assigned by the sender**/
    /** to aid in assembling the fragments of the datagram **/

    if (++arg_pContext->icmp_ipID >= 0x8000) arg_pContext->icmp_ipID  = 0x6000;
    pOutIpHdr->id = arg_pContext->icmp_ipID;

    length = (ix_uint16) (MIN_IP_LEN + IX_ICMP_HDR_LEN + payloadLength);

    pOutIpHdr->totalLength = IX_HTON16 (length);
    pOutIpHdr->checksum = 0;
    pOutIpHdr->checksum = IX_HTON16 (_ix_cc_ipv4_calculate_checksum
                                   ((ix_uint8 *) pOutIpHdr, pOutIpHdr->headerLength * 4));

    pOutIcmpHdr->checksum = IX_HTON16 (_ix_cc_ipv4_calculate_checksum
                                     ((ix_uint8 *) (icmpMessage + MIN_IP_LEN),
                                      IX_ICMP_HDR_LEN + payloadLength));

   /* Set output interface as input interface number in meta data of packet */

    pHwMeta->m_BufferSize = IX_HTON16 (length);
    pHwMeta->m_PacketSize = IX_HTON16 (length);

    ix_ossl_memcpy ((ix_uint8 *) pData,
		    (void *) icmpMessage,
		    MIN_IP_LEN + IX_ICMP_HDR_LEN +
                   ((inIpHdr.headerLength * 4) + 8));

   /** I need to fill up meta data before sending buffer out so that
    ** L2 can use it **/
   /** Get the nexthop information from RTM **/
   /** RTM expect destination IP address in host order **/

    IX_ERROR_CRT(ix_cc_rtmv4_lookup(g_pIpv4Context->hRtm, 
                 IX_NTOH32(pOutIpHdr->destAddress),
                &nexthopInfo),
                 IX_CC_ERROR_INTERNAL, 0);

    if ((nexthopInfo.l2Index == IX_CC_RTMV4_L2INDEX_NO_ROUTE) || 
        (nexthopInfo.flags & IPV4_NH_FLAGS_DROP_BIT))
    {
        #if defined(IX_MPLS_DEBUG)
        ix_ossl_message_log ("No route to the source. Dropping ICMP packet.\n");
        #endif

        IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hBuffer),
                     IX_CC_ERROR_INTERNAL, 0);
        return IX_SUCCESS;
    }

    pHwMeta->m_NextHopIDType = nexthopInfo.l2IndexType;


   /*
    * Since our host TCP/IP stack forwards IP broadcasts for us (no broad-
    * cast forwarding in IPv4 CC) we can get here with a packet coming
    * from the host's stack what means the input port is not set.
    * So don't use the input port from meta-data, but the port from the
    * nexthopInfo.
    */
    pHwMeta->m_OutputPort = IX_HTON16((ix_uint16)nexthopInfo.portID);
    pHwMeta->m_NextHopID  = IX_HTON16((ix_uint16)nexthopInfo.l2Index);
    pHwMeta->m_FabricPort = (ix_uint8)nexthopInfo.bladeID;


    IX_ERROR_CRT (ix_rm_buffer_is_eop (arg_hBuffer, &isEOP),
		  IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);


   /*
    * If the buffer is not the end of the packet (the incoming packet
    * was bigger than could fit in a single buffer), free up the rest of
    * the chain.  We don't need the rest of the chain.
    */
    if (!isEOP)
    {
      ix_buffer_handle hUnlinked = IX_NULL_BUFFER_HANDLE;

      IX_ERROR_CRT (ix_rm_buffer_unlink (&arg_hBuffer, &hUnlinked),
		    IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
      IX_ERROR_CRT (ix_rm_buffer_free_chain (hUnlinked),
		    IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
    }


    /* Set cellcount field in buffer */
    cell_count = ix_cc_calculate_cell_count ((ix_uint32) length, 1);


    /* Write the cell count into the buffer handle */
    IX_ERROR_CRT (ix_rm_buffer_cell_count_set (&arg_hBuffer, cell_count),
		  IX_CC_ERROR_INTERNAL, 0);


    if (_ix_cc_mpls_icmp_queue_enqueue (arg_hBuffer, arg_pContext))
    {
         ix_ossl_message_log ("_ix_cc_mpls_icmp_build_message (): ICMP queue full!!\n");
         return IX_ERROR_NEW (IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
    }

    return IX_SUCCESS;

}  /* _ix_cc_mpls_icmp_build_message */








/*
 * FUNCTION: _ix_cc_mpls_icmp_send_message()  */
/**
 * ABSTRACT: function called periodically by the framework to handle enqueued ICMP messages
 *
 * DESCRIPTION:  This function attempts to dequeue ICMP messages and
 *               send them to the destination.
 *
 * \param  arg_pContext - a pointer to the MPLS context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_icmp_send_message (void *arg_pContext)
{

  ix_uint32           idx = 0;
  ix_buffer_handle    hBuffer;
  ix_hw_buffer_meta  *pHwMeta     = NULL;
  ix_uint8            HeaderType;
  ix_cc_mpls_context *pContext = (ix_cc_mpls_context*) arg_pContext;


  /* Check if context returned by ix_cci_cc_add_event_handler function is not null */
  if (pContext == NULL)
      return IX_ERROR_WARNING (IX_CC_ERROR_NULL, ("It is a null context"));

  IX_ERROR_CRT(ix_ossl_mutex_lock(pContext->ICMPmutexId, IX_OSSL_WAIT_FOREVER),
               IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK, IX_ERROR_LEVEL_LOCAL);

  while ((idx < pContext->icmpQueueDrain) && 
         !IX_CC_MPLS_ICMP_IS_QUEUE_EMPTY(pContext->icmpQheadIdx,
                                         pContext->icmpQtailIdx))
  {

    if (_ix_cc_mpls_icmp_queue_dequeue (&hBuffer, pContext) == 0)
    {

      ix_error err;
      #if defined (IX_MPLS_DEBUG)
      {
	      void *pData = NULL;
          ix_cc_hw_get_packet_data (hBuffer, &pData);
          ix_ossl_message_log ("_ix_cc_mpls_icmp_send_message(): sending a packet\n");
          show_mem ((ix_uint32 *) pData, 100);
      }
      #endif

     /* Get meta data for buffer */
      err = ix_rm_buffer_get_meta (hBuffer, (void **) &pHwMeta);
      if (err == IX_SUCCESS) 
      {
           HeaderType = IX_RM_HW_META_GET_HEADER_TYPE(pHwMeta->m_BufferInfo);

           if (HeaderType == MPLS_IPV4_HEADER_TYPE)
           {
               ix_rm_packet_send(IX_CC_MPLS_FTN_INPUT,
                                 hBuffer,
                                 IX_CC_MPLS_FTN_INPUT);
           }
           else if (HeaderType == MPLS_MPLS_HEADER_TYPE)
           { 
               ix_rm_packet_send(IX_CC_MPLS_ILM_INPUT,
                                 hBuffer,
                                 IX_CC_MPLS_ILM_INPUT);
           }
           else
           {
#if defined (IX_MPLS_DEBUG)
                ix_ossl_message_log ("_ix_cc_mpls_icmp_send_message(): dropping packet\n");
#endif
                ix_rm_buffer_free_chain (hBuffer);
           }
      }
      idx++;
    }

  }

  IX_ERROR_CRT(ix_ossl_mutex_unlock(pContext->ICMPmutexId),
               IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK, IX_ERROR_LEVEL_LOCAL);

  return IX_SUCCESS;

}	/* _ix_cc_mpls_icmp_send_message */








/*
 * FUNCTION: ix_cc_mpls_get_sleep_time()  */
/**
 * ABSTRACT: gets the sleeping time between calling _ix_cc_mpls_icmp_send_message();
 *
 * DESCRIPTION: This library primitive helps to retrieve the number of seconds
 *              between when the event will fire causing the ICMP messages 
 *              to be dequeued and sent. The lower the value of arg_pSleepTime, 
 *              the more frequently the ICMP messages will be dequeued and sent.  
 *
 * \param  arg_pContext   - pointer to MPLS context
 * \param  arg_pSleepTime - returned ICMP sleep period 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 *
 *   Error Codes:
 * 
 * \retval    IX_CC_ERROR_NULL - null pointer
 */
ix_error
ix_cc_mpls_get_sleep_time (ix_uint16 *arg_pSleepTime,
                           void      *arg_pContext)
{

  if (arg_pSleepTime == NULL)
     return IX_ERROR_WARNING (IX_CC_ERROR_NULL,
			     ("Invalid Pointer: arg_pSleepTime"));

  *arg_pSleepTime = ((ix_cc_mpls_context *) arg_pContext)->icmpSleepTime;


  return IX_SUCCESS;
}	/* end ix_cc_mpls_get_sleep_time  */







/*
 * FUNCTION: ix_cc_mpls_set_sleep_time()  */
/**
 * ABSTRACT: sets the sleeping time between calling _ix_cc_mpls_icmp_send_message();
 *
 * DESCRIPTION: This library primitive helps to set the period between 
 *              events firing to process the ICMP error message queue.
 *              The lower the value of arg_SleepTime, the more frequently
 *              the output queue will be processed. 
 *              The minimum value of arg_SleepTime is 1 sec and 
 *              maximum value can be 32sec.
 *
 * \param  arg_SleepTime- the number of seconds to wait between processing 
 *                        the ICMP error message queue.
 * \param  arg_pContext - pointer to MPLS context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 *
 *   Error Codes:
 * 
 * \retval    IX_CC_MPLS_ERROR_INVALID_ENTRY  - null pointer
 */
#ifndef IX_EXCLUDE_CCI
ix_error
ix_cc_mpls_set_sleep_time (ix_uint16  arg_SleepTime,
                           void      *arg_pContext)
{
  ix_error            err;
  ix_uint32           eventInterval;
  ix_cc_handle        hEventHandle;
  ix_cc_mpls_context *pContext = arg_pContext;

  if ((arg_SleepTime < IX_CC_MPLS_ICMP_MIN_SLEEP_TIME) ||
      (arg_SleepTime > IX_CC_MPLS_ICMP_MAX_SLEEP_TIME))
         return IX_ERROR_WARNING (IX_CC_MPLS_ERROR_INVALID_ENTRY,
			     ("ICMP: ix_cc_mpls_set_sleep_time: invalid parameters"));

  IX_ERROR_CRT(ix_ossl_mutex_lock(pContext->ICMPmutexId, IX_OSSL_WAIT_FOREVER),
               IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK, IX_ERROR_LEVEL_LOCAL);
  
  pContext->icmpSleepTime = arg_SleepTime;
  
  hEventHandle = pContext->hEventHandle;
  
  /* Calculate event interval in milliseconds, 
     data in icmpSleepTime is in seconds  */
  eventInterval = arg_SleepTime;

  IX_ERROR_C(ix_cci_change_event (hEventHandle, eventInterval), err);

  IX_ERROR_CRT(ix_ossl_mutex_unlock(pContext->ICMPmutexId),
               IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK, IX_ERROR_LEVEL_LOCAL);

  return err;

}	/* end ix_cc_mpls_set_sleep_time */
#endif





/*
 * FUNCTION: ix_cc_mpls_get_queue_depth()  */
/**
 * ABSTRACT: gets the current length of ICMP queue
 *
 * DESCRIPTION: This library primitive helps to retrieves 
 *              the depth of the ICMP error message queue. 
 *              That is, the maximum number of ICMP messages that can 
 *              fit in the queue, not the number of messages currently in the queue.
 *
 * \param  arg_pQueueDepth - the current number of entries 
 *                           that the ICMP error message queue can hold.
 * \param  arg_pContext    - pointer to MPLS context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 *
 *   Error Codes:
 * 
 * \retval    IX_CC_ERROR_NULL - null pointer
 */
ix_error
ix_cc_mpls_get_queue_depth (ix_uint16 *arg_pQueueDepth,
                            void      *arg_pContext)
{
  ix_cc_mpls_context *pContext = (ix_cc_mpls_context*) arg_pContext;

  if (pContext  == NULL)
    return IX_ERROR_WARNING (IX_CC_ERROR_NULL,
			     ("Invalid pointer: arg_pQueueDepth"));

  IX_ERROR_CRT(ix_ossl_mutex_lock(pContext->ICMPmutexId, IX_OSSL_WAIT_FOREVER),
               IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK, IX_ERROR_LEVEL_LOCAL);

  *arg_pQueueDepth = IX_CC_MPLS_ICMP_GET_QUEUE_SIZE(pContext->icmpQheadIdx,
                                                    pContext->icmpQtailIdx,
                                                    pContext->icmpQueueMaxSize);

  IX_ERROR_CRT(ix_ossl_mutex_unlock(pContext->ICMPmutexId),
               IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK, IX_ERROR_LEVEL_LOCAL);

  return IX_SUCCESS;

}	/* end ix_cc_mpls_get_queue_depth  */







/*
 * FUNCTION: ix_cc_mpls_get_packets_to_drain()  */
/**
 * ABSTRACT: gets the max number of ICMP messages sent by _ix_cc_mpls_icmp_send_message() in one turn
 *
 * DESCRIPTION:This library primitive helps to retrieve 
 *             the maximum number of ICMP error messages
 *             to send (from the ICMP message queue ) each time the 
 *             queue processing event fires. 
 *
 * \param  arg_pPacketsToDrain - the maximum number of packets 
 *                               to send each time the event fires for processing the queue.
 * \param  arg_pContext        - pointer to MPLS context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 *
 *   Error Codes:
 * 
 * \retval    IX_CC_ERROR_NULL - null pointer
 */
ix_error
ix_cc_mpls_get_packets_to_drain (ix_uint16 *arg_pPacketsToDrain,
                                 void      *arg_pContext)
{
  if (arg_pPacketsToDrain == NULL)
  {
    return IX_ERROR_WARNING (IX_CC_ERROR_NULL,
			     ("Invalid pointer: arg_pPacketsToDrain "));
  }	/* end if(arg_pPacketsToDrain == NULL)  */

  *arg_pPacketsToDrain = ((ix_cc_mpls_context *) arg_pContext)->icmpQueueDrain;

  return IX_SUCCESS;

}	/* end ix_cc_mpls_get_packets_to_drain  */






/*
 * FUNCTION: ix_cc_mpls_set_packets_to_drain()  */
/**
 * ABSTRACT: sets the max number of ICMP messages sent by _ix_cc_mpls_icmp_send_message() in one turn
 *
 * DESCRIPTION: This library primitive helps to set the maximum number 
 *              of ICMP error messages to send each time the queue processing event fires.
 *              The value of arg_PacketsToDrain must not exceed the depth of the queue.
 *
 * \param   arg_PacketsToDrain - The maximum number of packets to send 
 *                               each time the queue processing event fires. 
 * \param  arg_pContext        - pointer to MPLS context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 *
 *   Error Codes:
 * 
 * \retval    IX_CC_MPLS_ERROR_INVALID_PACKETS_TO_DRAIN  - null pointer
 */
ix_error
ix_cc_mpls_set_packets_to_drain (ix_uint16  arg_PacketsToDrain,
                                 void      *arg_pContext)
{
  ix_cc_mpls_context *pContext = (ix_cc_mpls_context*) arg_pContext;

  if ((arg_PacketsToDrain < 1) || (arg_PacketsToDrain > g_IcmpQueueDepth))
     return IX_ERROR_WARNING (IX_CC_MPLS_ERROR_INVALID_PACKETS_TO_DRAIN,
              ("ix_cc_mpls_set_packets_to_drain(): invalid parameter: %d\n", arg_PacketsToDrain));
  
  IX_ERROR_CRT(ix_ossl_mutex_lock(pContext->ICMPmutexId, IX_OSSL_WAIT_FOREVER),
               IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK, IX_ERROR_LEVEL_LOCAL);

  ((ix_cc_mpls_context *) arg_pContext)->icmpQueueDrain = arg_PacketsToDrain;

  IX_ERROR_CRT(ix_ossl_mutex_unlock(pContext->ICMPmutexId),
               IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK, IX_ERROR_LEVEL_LOCAL);

  return IX_SUCCESS;

}	/* end ix_cc_mpls_set_packets_to_drain */

