/*
 * ============================================================================
 * COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *                                                                          */
/** \file      ix_cc_mpls_lsp.c
 *
 *  \brief     The file contains CCI API function exposed by
 *             MPLS Core Component used to manage LSP table
 *
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_mpls_lsp.c,v 1.19.2.1 2004/03/12 10:40:15 wneubaue Exp $";

#include <ix_types.h>
#include <ix_error.h>

#include <ix_ossl.h>
#include <ix_rm.h>

#if !defined(IX_EXCLUDE_CCI)
#include <ix_cci.h>
#endif 

#include <ix_cc_error.h>
#include <ix_cc.h>

#include <bindings.h>
#include "definitions.h"

#include <cc/ix_cc_rtmv4.h>
#include <cc/ix_cc_reg_util.h>

#include <../../include/mpls_defs.h>
#include <cc/ix_cc_mpls.h>
#include "cc/internal/ix_cc_mpls_lsp.h"
#include "cc/internal/ix_cc_mpls_nhlfe.h"
#include "cc/internal/ix_cc_mpls_internal.h"
#include "cc/ipv4/internal/ix_cc_ipv4_init.h"

#if (OS == WIN_32)
#define __FUNCTION__ ""
#endif


PRIVATE ix_error
_ix_cc_mpls_lsp_verify_input(ix_cc_mpls_lsp     *arg_pLsp,
                             ix_cc_mpls_context *arg_pContext);

PRIVATE ix_error
_ix_cc_mpls_lsp_add_to_free_list(ix_cc_mpls_lsp_handle  arg_LspHandle,
                                 ix_cc_mpls_context    *arg_pContext);

PRIVATE ix_error
_ix_cc_mpls_ilm_set_add_to_free_list(ix_uint32           arg_IlmSetIdx,
                                     ix_cc_mpls_context *arg_pContext);

PRIVATE void
_ix_cc_mpls_lsp_copy_nhlfe_to_ilm(ix_cc_mpls_ilm_internal  *pNewIlmEntry,
                                  ix_cc_mpls_lsp_param      arg_lspParams,
                                  ix_cc_mpls_nhlfe         *Nhlfe);

PRIVATE ix_error
_ix_cc_mpls_lsp_get_label_space_config(void *arg_pContext);


PRIVATE ix_error
_ix_cc_mpls_lsp_copy_nhlfe_set_to_ilm(ix_cc_mpls_ilm_set_internal  *arg_pIlmSetEntry,
                                      ix_uint16                     arg_IlmSetSize,
                                      ix_cc_mpls_lsp_param          arg_lspParams[MPLS_MAX_NHLFE_IN_SET],
                                      ix_cc_mpls_ilm_internal      *arg_NewIlmSetEntry[MPLS_MAX_NHLFE_IN_SET],
                                      ix_uint32                     arg_NewIlmSetIndices[MPLS_MAX_NHLFE_IN_SET],
                                      ix_cc_mpls_nhlfe_handle       arg_NhlfeHandles[MPLS_MAX_NHLFE_IN_SET],
                                      ix_cc_mpls_context           *arg_pContext);

/* for the purpose of this file only */
#define ON_ERROR_GOTO(arg_destination, msg)                                       \
        if (err != IX_SUCCESS) {                                                  \
                    err = IX_ERROR_WARNING(IX_ERROR_GET_CODE(err),                \
                    (__FUNCTION__"(): "msg));       \
                    goto arg_destination;                                         \
        }



/*
 * FUNCTION: ix_cc_mpls_lsp_create()  */
/**
 * ABSTRACT: This function creates an LSP description defining a FEC to NHLFE, 
 *           or ILM to NHLFE mapping.
 *
 * DESCRIPTION:This function creates an LSP description defining 
 *             FEC to NHLFE, or ILM to NHLFE mapping. 
 *             If arg_Lsp.lspType is equal to LSP_FEC, the function checks 
 *             whether an IPv4 routing table entry for the given FEC exists;
 *             if so, an error is returned. Otherwise, the specified NHLFE
 *             table entry (or entries) is created if arg_Lsp.nhlfeInfoType
 *             equals to NHLFE_DATA. Next the IPv4 next_hop and routing 
 *             table entries for the given FEC are added. Finally, a CC_LSP 
 *             table entry linking the FEC ID with the NHLFE outSegID is created,
 *             and the LSP handle is returned to the caller.
 *
 *             If arg_Lsp.lspType is equal to LSP_ILM, the function checks 
 *             whether an ILM table entry for the given label and input port 
 *             exists; if so, an error is returned. Otherwise, the specified 
 *             NHLFE table entry (or entries) is created if arg_Lsp.nhlfeInfoType 
 *             equals to NHLFE_DATA. Next the ILM table entry for the given label
 *             and input port is added. Finally, a CC_LSP table entry linking 
 *             the ILM entry with the NHLFE outSegID is created, and the LSP 
 *             handle is returned to the caller.
 *
 * \param  arg_pLsp       - pointer to an ix_s_cc_mpls_lsp structure, 
 *                          defining an LSP
 * \param  arg_pContext   - pointer to the MPLS core component context 
 * \param  arg_pLspHandle - pointer to the handle identifying the created LSP
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_DUPLICATE_ENTRY      - the specified FEC or label already exists
 * \n         IX_CC_ERROR_FULL                 - no place in the ILM or NHLFE table
 * \n         IX_CC_ERROR_RANGE                - LSP parameters are out of range
 * \n         IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK - mutex lock failed
 * \n
 * \n NHLFE error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_RANGE                - arg_NhlfeSize is beyond 
 *                                               the range from (1 to 4)
 * \n         IX_CC_ERROR_FULL                 - NHLFE table is full
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - one of the set NHLFEs does not exist
 * \n         IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET - one of the set NHLFEs is an NHLFE set entry
 * \n         IX_CC_MPLS_ERROR_ENTRY_NHLFE     - this is not an NHLFE set entry
 * \n         IX_CC_MPLS_ERROR_INVALID_HANDLE  - invalid NHLFE parameters
 */
ix_error
ix_cc_mpls_lsp_create(ix_cc_mpls_lsp          *arg_pLsp,
                      ix_cc_mpls_lsp_handle   *arg_pLspHandle,
                      void                    *arg_pContext) 

{
    ix_error   err;
    ix_uint32  idx;
    ix_cc_mpls_context           *pContext = arg_pContext;
    ix_cc_rtmv4_next_hop_info     nextHopInfo;
    ix_cc_mpls_nhlfe              Nhlfe;

    ix_uint16                     NhlfeHandles_num;
    ix_cc_mpls_nhlfe_handle       NhlfeHandles[MPLS_MAX_NHLFE_IN_SET];

    ix_cc_mpls_cc_lsp            *pCcLspEntry;
    ix_cc_mpls_cc_lsp_free_slot  *pFreeLspListOldHead;
    ix_cc_mpls_ilm_internal      *pNewIlmEntry;

#ifdef MPLS_COUNTERS_FEATURE
    ix_cc_mpls_ilm_counters_internal   *pIlmStats;
    ix_cc_mpls_nhlfe_counters_internal *pFtnStats;
    ix_counter_64bit_handle            *p64BitCountersHdls;
#endif

    /* verify input arguments */
    if (arg_pLspHandle == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("ix_cc_mpls_lsp_create(): got null pointer to LSP handle storage"));

    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_lsp_create(): got null Null MPLS context"));

    if (pContext->pCcLspBaseAddr == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("ix_cc_mpls_lsp_create(): "
                                "CC LSP table has not been initialized."));

    /* check if there is a free slot in CC LSP table */
    if (pContext->pFreeLspListHead == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_FULL,
                                ("ix_cc_mpls_lsp_create(): no place in the CC LSP table."));

    /* verify parameters in LSP structure */
    err = _ix_cc_mpls_lsp_verify_input(arg_pLsp, arg_pContext);
    if (err != IX_SUCCESS)
        return err;

    IX_ERROR_C(ix_ossl_mutex_lock(pContext->lspMutexId, IX_OSSL_WAIT_FOREVER), err);
    
    if (err != IX_SUCCESS)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK,
                                ("ix_cc_mpls_lsp_create(): can't lock mutex"));


    pCcLspEntry = &pContext->pCcLspBaseAddr[pContext->pFreeLspListHead->LspIndex];

    if ((pCcLspEntry->flags & MPLS_FLAG_VALID_BIT) != 0)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                               ("ix_cc_mpls_lsp_create(): Entry already used"));
        goto err_exit;
    }

    if  (arg_pLsp->lspType == IX_CC_MPLS_LSP_ILM_TYPE)
    {
        /* get new ILM table's entry */
        pNewIlmEntry = (ix_cc_mpls_ilm_internal*)((ix_uint8*)pContext->pIlmBaseAddr + 
                        pContext->pLabelSpaceTable[arg_pLsp->inSeg.ilm.in_port].table_offset +
                        (arg_pLsp->inSeg.ilm.label.label - pContext->pLabelSpaceTable[arg_pLsp->inSeg.ilm.in_port].min_label)
                              * sizeof(ix_cc_mpls_ilm_internal));

#if defined (IX_MPLS_DEBUG)
        ix_ossl_message_log("ix_cc_mpls_lsp_create(): Creating ILM at 0x%x\n", pNewIlmEntry);
        ix_ossl_message_log("Adding label : %d\n",arg_pLsp->inSeg.ilm.label.label);
#endif

        if ((pNewIlmEntry->flags & MPLS_FLAG_VALID_BIT) != 0)
        {
            err = IX_ERROR_WARNING(IX_CC_ERROR_DUPLICATE_ENTRY,
                                   ("ix_cc_mpls_lsp_create(): ILM entry allready exists"));
            goto err_exit;
        }

    }


    /* Add Nhlfe */
    if (arg_pLsp->nhlfeInfoType == IX_CC_MPLS_LSP_NHLFE_DATA_TYPE)
    {
        NhlfeHandles_num = arg_pLsp->nhlfe_num;
        for (idx = 0; idx < NhlfeHandles_num; idx ++) 
        {
            err = ix_cc_mpls_nhlfe_create(
                            *arg_pLsp->outSeg.nhlfeData[idx],
                            &NhlfeHandles[idx],
                             arg_pContext);
 
            ON_ERROR_GOTO(err_free_nhlfe, "could not create NHLFE entry");
        }

        Nhlfe = *arg_pLsp->outSeg.nhlfeData[0];
       
        /* Add Nhlfe Set if necesery */
        if (NhlfeHandles_num >= 2) 
        {
            err = ix_cc_mpls_nhlfe_set_create(NhlfeHandles_num,
                                              NhlfeHandles,
                                              &pCcLspEntry->NhlfeHandle,
                                              arg_pContext);

            ON_ERROR_GOTO (err_free_nhlfe, "could not create NHLFE set");
        
            arg_pLsp->nhlfeSetHandle = pCcLspEntry->NhlfeHandle;

        } else {

            pCcLspEntry->NhlfeHandle = NhlfeHandles[0];
        }

    } else {
        if (arg_pLsp->nhlfeInfoType == IX_CC_MPLS_LSP_NHLFE_SET_HANDLE_TYPE)
        {
            err = ix_cc_mpls_nhlfe_set_query(arg_pLsp->nhlfeSetHandle,
                                             &NhlfeHandles_num,
                                             NhlfeHandles,
                                             arg_pContext);

            ON_ERROR_GOTO (err_exit, "specified NHLFE set does not exist");

            pCcLspEntry->NhlfeHandle = arg_pLsp->nhlfeSetHandle;

        }
        else if (arg_pLsp->nhlfeInfoType == IX_CC_MPLS_LSP_NHLFE_HANDLE_TYPE)
        {
            NhlfeHandles_num = arg_pLsp->nhlfe_num;
            ix_ossl_memcpy(NhlfeHandles, arg_pLsp->outSeg.nhlfeHandle,
                           sizeof(ix_cc_mpls_nhlfe_handle) * NhlfeHandles_num);

            /* Add Nhlfe Set if necesery */
            if (NhlfeHandles_num >= 2)
            {
                err = ix_cc_mpls_nhlfe_set_create(NhlfeHandles_num,
                                                  NhlfeHandles,
                                                  &pCcLspEntry->NhlfeHandle,
                                                  arg_pContext);

                ON_ERROR_GOTO (err_exit, "could not create NHLFE SET");
  
                arg_pLsp->nhlfeSetHandle = pCcLspEntry->NhlfeHandle;
            }
            else
            {
                pCcLspEntry->NhlfeHandle = arg_pLsp->outSeg.nhlfeHandle[0];
            }
        }
        err = ix_cc_mpls_nhlfe_query(NhlfeHandles[0],
                                    &Nhlfe,
                                     arg_pContext);

        ON_ERROR_GOTO (err_free_nhlfe_set, "could not find NHLFE set entry");
    }

    
    if  (arg_pLsp->lspType == IX_CC_MPLS_LSP_FEC_TYPE)
    {
        /* Add next hop info and route entries */
    
        /* set ipv4 address */
        nextHopInfo.ipAddr      = arg_pLsp->inSeg.fec.fec.ip4fec.ipAddress;
        /* set mpls next hop id type */
        nextHopInfo.l2IndexType = NEXT_HOP_ID_MPLS_TYPE;
        nextHopInfo.l2Index     = pCcLspEntry->NhlfeHandle.outSegId;
        nextHopInfo.bladeID     = Nhlfe.bladeId;
        nextHopInfo.mtu         = arg_pLsp->lspParams[0].maxLabPktSize;
        nextHopInfo.portID      = Nhlfe.outPortId;
        nextHopInfo.flags       = 0;

        /* add RTMv4 next hop */
        IX_ERROR_CGT(ix_cc_rtmv4_add_next_hop(
                     pContext->hRtm,
                     arg_pLsp->inSeg.fec.fec.ip4fec.nexthopId,
                     &nextHopInfo),
                     err,
                     IX_CC_MPLS_ERROR_RTM_ADD_ENTRY,
                     IX_ERROR_LEVEL_REMOTE,
                     err_add_next_hop);

        /* add RTMv4 route */
        IX_ERROR_CGT(ix_cc_rtmv4_add_route(
                     pContext->hRtm,
                     arg_pLsp->inSeg.fec.fec.ip4fec.ipAddress,
                     arg_pLsp->inSeg.fec.fec.ip4fec.netMask,
                     arg_pLsp->inSeg.fec.fec.ip4fec.nexthopId),
                     err,
                     IX_CC_MPLS_ERROR_RTM_ADD_ROUTE,
                     IX_ERROR_LEVEL_REMOTE,
                     err_add_route);

        pCcLspEntry->inHandle.fecData.NextHopId = arg_pLsp->inSeg.fec.fec.ip4fec.nexthopId; 
        pCcLspEntry->inHandle.fecData.NetMask   = arg_pLsp->inSeg.fec.fec.ip4fec.netMask;
        pCcLspEntry->inHandle.fecData.NetAddr   = arg_pLsp->inSeg.fec.fec.ip4fec.ipAddress;

        for (idx = 0; idx < NhlfeHandles_num; idx ++)
        {
            _ix_cc_mpls_nhlfe_internal_modify(NhlfeHandles[idx],
                                              0xFFFF,
                                              arg_pLsp->lspParams[0],
                                              arg_pContext);
        }
#if defined (MPLS_COUNTERS_FEATURE)
        
        pFtnStats = (ix_cc_mpls_nhlfe_counters_internal*)((ix_uint8*)(pContext->pFtnCounters) + 
                     pCcLspEntry->NhlfeHandle.outSegId * sizeof(ix_cc_mpls_nhlfe_counters_internal));
		       
        p64BitCountersHdls = (ix_counter_64bit_handle*)((ix_uint8*)(pContext->pFtn64BitCountersHdls) + 
                       pCcLspEntry->NhlfeHandle.outSegId * sizeof(ix_counter_64bit_handle) * MPLS_NHLFE_64BIT_COUNTERS_NUM);

        /* Clear 32-bit counters */
        ix_ossl_memset(pFtnStats, 0, sizeof(ix_cc_mpls_nhlfe_counters_internal));


        /* Warning: 64-bit counters 32-bit base have to be at the beginning
        of counters structure (now only one 64-bit counter is used */
        IX_ERROR_CGT(ix_rm_counter_64bit_new(MPLS_NHLFE_64BIT_COUNTERS_NUM,
                                            (ix_uint32*)pFtnStats,
                                            MPLS_NHLFE_32BIT_COUNTER_OVERFLOW_TIME,
                                            p64BitCountersHdls),
                     err,
                     IX_CC_ERROR_OOM_64BIT_COUNTER,
                     IX_ERROR_LEVEL_REMOTE,
                     err_add_route);
#endif

    }
    else
    {
       
        if (NhlfeHandles_num >= 2)
        {
            ix_uint32                      NewIlmSetIdices[MPLS_MAX_NHLFE_IN_SET];
            ix_cc_mpls_ilm_internal       *NewIlmSetEntry[MPLS_MAX_NHLFE_IN_SET];
            ix_cc_mpls_ilm_set_free_slot  *pCurrentFreeIlmSet = pContext->pFreeIlmSetListHead;

            for (idx = 0; idx < NhlfeHandles_num; idx ++)
            {
                if (pCurrentFreeIlmSet == NULL)
                {
                    err = IX_ERROR_WARNING(IX_CC_ERROR_FULL,
                                           ("ix_cc_mpls_lsp_create(): no space to create ilm set entry."));
                    goto err_create_ilm;
                }
                NewIlmSetIdices[idx] = pCurrentFreeIlmSet->IlmSetIndex; 
                pCurrentFreeIlmSet   = pCurrentFreeIlmSet->pNextFreeSlot;

                /* get new ILM SET table's entry */
                NewIlmSetEntry[idx] = pContext->pIlmSetBaseAddr + 
                                      NewIlmSetIdices[idx];
            }

            err = _ix_cc_mpls_lsp_copy_nhlfe_set_to_ilm(
                          (ix_cc_mpls_ilm_set_internal*)pNewIlmEntry,
                                                        NhlfeHandles_num,
                                                        arg_pLsp->lspParams,
                                                        NewIlmSetEntry,
                                                        NewIlmSetIdices,
                                                        NhlfeHandles,
                                                        pContext);
            
            ON_ERROR_GOTO (err_create_ilm, "Could not copy NHLFE contents to ilm entry");

            for (idx = 0; idx < NhlfeHandles_num; idx ++)
            {
                NewIlmSetEntry[idx]->operation = arg_pLsp->inSeg.ilm.operation;
                    
                pCurrentFreeIlmSet            = pContext->pFreeIlmSetListHead;
                pContext->pFreeIlmSetListHead = pCurrentFreeIlmSet->pNextFreeSlot;
                ix_ossl_free(pCurrentFreeIlmSet);
            }
                

#if defined (MPLS_COUNTERS_FEATURE)
            {
                ix_cc_mpls_ilm_set_internal *pIlmSetEntry = (ix_cc_mpls_ilm_set_internal*) pNewIlmEntry;

                pIlmStats = (ix_cc_mpls_ilm_counters_internal*)((ix_uint8*)(pContext->pIlmNhlfeSetCounters) + 
                               ((ix_uint32)(pIlmSetEntry->OutSegId1)) * sizeof(ix_cc_mpls_ilm_counters_internal));
		       
                p64BitCountersHdls = (ix_counter_64bit_handle*)((ix_uint8*)(pContext->pIlmNhlfeSet64BitCountersHdls) + 
                                      ((ix_uint32)(pIlmSetEntry->OutSegId1)) * sizeof(ix_counter_64bit_handle) * MPLS_ILM_64BIT_COUNTERS_NUM);

                /* Clear 32-bit counters */
                ix_ossl_memset(pIlmStats, 0, sizeof(ix_cc_mpls_ilm_counters_internal));


                /* Warning: 64-bit counters 32-bit base have to be at the beginning
                   of counters structure (now only one 64-bit counter is used */
                IX_ERROR_CGT(ix_rm_counter_64bit_new(MPLS_ILM_64BIT_COUNTERS_NUM,
                                                     (ix_uint32*)pIlmStats,
                                                     MPLS_ILM_32BIT_COUNTER_OVERFLOW_TIME,
                                                     p64BitCountersHdls),
                             err,
                             IX_CC_ERROR_OOM_64BIT_COUNTER,
                             IX_ERROR_LEVEL_REMOTE,
                             err_add_route);
            }
#endif

        }
        else
        {
            _ix_cc_mpls_lsp_copy_nhlfe_to_ilm( pNewIlmEntry, 
                                               arg_pLsp->lspParams[0],
                                              &Nhlfe);

            pNewIlmEntry->operation = arg_pLsp->inSeg.ilm.operation;

#if defined (MPLS_COUNTERS_FEATURE)

            pIlmStats = (ix_cc_mpls_ilm_counters_internal*)((ix_uint8*)(pContext->pIlmCounters) + 
                        pContext->pLabelSpaceTable[arg_pLsp->inSeg.ilm.in_port].stats_offset +
                       (arg_pLsp->inSeg.ilm.label.label - pContext->pLabelSpaceTable[arg_pLsp->inSeg.ilm.in_port].min_label)
                              * sizeof(ix_cc_mpls_ilm_counters_internal));
		       
            p64BitCountersHdls = (ix_counter_64bit_handle*)((ix_uint8*)(pContext->pIlm64BitCountersHdls) + 
                       pContext->pLabelSpaceTable[arg_pLsp->inSeg.ilm.in_port].stats_offset +
                       (arg_pLsp->inSeg.ilm.label.label - pContext->pLabelSpaceTable[arg_pLsp->inSeg.ilm.in_port].min_label)
                       * sizeof(ix_counter_64bit_handle) * MPLS_ILM_64BIT_COUNTERS_NUM);

            /* Clear 32-bit counters */
            ix_ossl_memset(pIlmStats, 0, sizeof(ix_cc_mpls_ilm_counters_internal));


            /* Warning: 64-bit counters 32-bit base have to be at the beginning
               of counters structure (now only one 64-bit counter is used */
            IX_ERROR_CGT(ix_rm_counter_64bit_new(MPLS_ILM_64BIT_COUNTERS_NUM,
                                                 (ix_uint32*)pIlmStats,
                                                 MPLS_ILM_32BIT_COUNTER_OVERFLOW_TIME,
                                                 p64BitCountersHdls),
                         err,
                         IX_CC_ERROR_OOM_64BIT_COUNTER,
                         IX_ERROR_LEVEL_REMOTE,
                         err_add_route);
#endif

        }
        
        pCcLspEntry->inHandle.ilmData.label   = arg_pLsp->inSeg.ilm.label;
        pCcLspEntry->inHandle.ilmData.in_port = arg_pLsp->inSeg.ilm.in_port;
    }

    /* Fill in CC LSP structure */
    pCcLspEntry->type        = arg_pLsp->lspType;
    pCcLspEntry->flags       = MPLS_FLAG_VALID_BIT;
    
    err = _ix_cc_mpls_nhlfe_inc_ref_count(pCcLspEntry->NhlfeHandle, arg_pContext);
    ON_ERROR_GOTO (err_add_route, "could not increse NHLFE's refCount");

    err = _ix_cc_mpls_nhlfe_add_lsp(pCcLspEntry, arg_pContext);
    ON_ERROR_GOTO (err_add_lsp, "could not add LSP to the list");

    
    /* get free CC LSP table's entry */
    arg_pLspHandle->bladeId  = pContext->localBladeId;
    arg_pLspHandle->LspIndex = pContext->pFreeLspListHead->LspIndex;

    /* remove CC Lsp entry from the free list */
    pFreeLspListOldHead = pContext->pFreeLspListHead;
    pContext->pFreeLspListHead = pContext->pFreeLspListHead->pNextFreeSlot;
    ix_ossl_free(pFreeLspListOldHead);
    
    /* Return Nhlfe handles */
    if (arg_pLsp->nhlfeInfoType == IX_CC_MPLS_LSP_NHLFE_DATA_TYPE)
        for (idx = 0; idx < NhlfeHandles_num; idx ++) 
            arg_pLsp->outSeg.nhlfeHandle[idx] = NhlfeHandles[idx];

    goto err_exit;


err_add_lsp:
    if (_ix_cc_mpls_nhlfe_dec_ref_count(pCcLspEntry->NhlfeHandle,
                                        arg_pContext) != IX_SUCCESS)
          ix_ossl_message_log("ix_cc_mpls_lsp_create(): Warning: can't decrease "
                              "LSP reference count.\n");

    
err_add_route:
    if (ix_cc_rtmv4_delete_next_hop(pContext->hRtm,
                                    arg_pLsp->inSeg.fec.fec.ip4fec.nexthopId) != IX_SUCCESS)
          ix_ossl_message_log("ix_cc_mpls_lsp_create(): Warning: can't delete "
                              "next hop.\n");

err_create_ilm:
err_add_next_hop:
err_free_nhlfe_set:
    if ((arg_pLsp->nhlfeInfoType == IX_CC_MPLS_LSP_NHLFE_DATA_TYPE) ||
        (arg_pLsp->nhlfeInfoType == IX_CC_MPLS_LSP_NHLFE_HANDLE_TYPE))

       if (arg_pLsp->nhlfe_num >= 2) {
            if (ix_cc_mpls_nhlfe_set_delete(pCcLspEntry->NhlfeHandle,
                                            arg_pContext) != IX_SUCCESS)
                ix_ossl_message_log("ix_cc_mpls_lsp_create(): Warning: can't delete "
                              "created NHLFE set.\n");
       }


err_free_nhlfe:
    if (arg_pLsp->nhlfeInfoType == IX_CC_MPLS_LSP_NHLFE_DATA_TYPE)

        for (idx; idx > 0; idx --) 
        {
            if(ix_cc_mpls_nhlfe_delete(NhlfeHandles[idx-1],
                                       arg_pContext) != IX_SUCCESS)
	    		ix_ossl_message_log("ix_cc_mpls_lsp_create(): Warning: can't delete "
                                    "created NHLFEs .\n");
        }

err_exit:
    IX_ERROR_CRT(ix_ossl_mutex_unlock(pContext->lspMutexId),
                 IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK, 
                 IX_ERROR_LEVEL_REMOTE);

    return err;
} /* ix_cc_mpls_lsp_create */





/*
 * FUNCTION: ix_cc_mpls_lsp_delete()  */
/**
 * ABSTRACT: This function deletes an LSP description from the CC_ILM table,
 *           together with its corresponding IP Routing table,
 *           ILM and NHLFE entries (if necessary)
 *
 * \param  arg_LspHandle - handle identifying an LSP. 
 * \param  arg_pContext  - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_MPLS_ERROR_INVALID_HANDLE  - invalid arg_LspHandle
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - no such entry 
 * \n         IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK - mutex lock failed
 * \n
 * \n LSP add error codes:
 * \n         IX_CC_ERROR_OOM_SYSTEM           - memory allocation failure
 */
ix_error
ix_cc_mpls_lsp_delete(ix_cc_mpls_lsp_handle  arg_LspHandle,
                      void                  *arg_pContext)
{

    ix_error                  err;
    ix_cc_mpls_ilm_internal  *pIlmEntry;
    ix_cc_mpls_cc_lsp        *pCcLspEntry;
    ix_cc_mpls_context       *pContext = arg_pContext;

#ifdef MPLS_COUNTERS_FEATURE
    ix_counter_64bit_handle  *p64BitCountersHdls;
    ix_uint32                 idx;
#endif


    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_lsp_delete(): got null Null MPLS context"));

    if (pContext->pCcLspBaseAddr == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("ix_cc_mpls_lsp_delete(): "
                                "CC LSP table has not been initialized."));

    /* add free CC LSP table's entry */
    if (arg_LspHandle.bladeId != pContext->localBladeId)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE, 
                                ("ix_cc_mpls_lsp_delete(): wrong bladeId"));

    IX_ERROR_C(ix_ossl_mutex_lock(pContext->lspMutexId, IX_OSSL_WAIT_FOREVER), err);

    if (err != IX_SUCCESS)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK,
                                ("ix_cc_mpls_lsp_delete(): can't lock mutex"));


    pCcLspEntry = &pContext->pCcLspBaseAddr[arg_LspHandle.LspIndex];

    if ((pCcLspEntry->flags & MPLS_FLAG_VALID_BIT) == 0)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                               ("ix_cc_mpls_lsp_delete(): LSP entry not found"));
        goto err_exit;
    }
    
    if  (pCcLspEntry->type == IX_CC_MPLS_LSP_ILM_TYPE)
    {
        ix_uint32 idx;

        /* get new ILM table's entry */
        pIlmEntry = (ix_cc_mpls_ilm_internal*)((ix_uint8*)pContext->pIlmBaseAddr + 
                     pContext->pLabelSpaceTable[pCcLspEntry->inHandle.ilmData.in_port].table_offset +
                    (pCcLspEntry->inHandle.ilmData.label.label - pContext->pLabelSpaceTable[pCcLspEntry->inHandle.ilmData.in_port].min_label)
                           * sizeof(ix_cc_mpls_ilm_internal));

        if (pIlmEntry->flags & MPLS_FLAG_SET_BIT)
        {
            ix_cc_mpls_ilm_set_internal   *pIlmSetEntry = (ix_cc_mpls_ilm_set_internal*) pIlmEntry;
            ix_cc_mpls_ilm_internal       *pNewIlmEntry;
            ix_uint32                      IlmSetIndices[MPLS_MAX_NHLFE_IN_SET];
            
            if (pIlmSetEntry->SetSize <= 0)
            {
                ix_ossl_message_log("ILM SET contains %d ILM entries\n", pIlmSetEntry->SetSize);
            }

            IlmSetIndices[0] =  pIlmSetEntry->OutSegId1;
            IlmSetIndices[1] = (pIlmSetEntry->OutSegId2Hi << 8)  | pIlmSetEntry->OutSegId2Lo;
            IlmSetIndices[2] =  pIlmSetEntry->OutSegId3;
            IlmSetIndices[3] = (pIlmSetEntry->OutSegId4Hi << 16) | pIlmSetEntry->OutSegId4Lo;

            /* free ILM entries */
            for (idx = 0; idx < pIlmSetEntry->SetSize; idx ++)
            {
                /* get new ILM SET table's entry */
                pNewIlmEntry = pContext->pIlmSetBaseAddr + 
                    IlmSetIndices[idx] *
                    sizeof(ix_cc_mpls_ilm_internal);

                ix_ossl_memset(pNewIlmEntry, 0, sizeof(ix_cc_mpls_ilm_internal));

                if (_ix_cc_mpls_ilm_set_add_to_free_list(IlmSetIndices[idx], pContext) != IX_SUCCESS)
                {
                    ix_ossl_message_log("ix_cc_mpls_lsp_create(): Warning: can't free ilm entries "
                                        "Out of memoty - some ILMs are lost.\n");
                    
                    break;
                }
            }
        }
        
#if defined (MPLS_COUNTERS_FEATURE)

        if (pIlmEntry->flags & MPLS_FLAG_SET_BIT)
        {
            ix_cc_mpls_ilm_set_internal   *pIlmSetEntry = (ix_cc_mpls_ilm_set_internal*) pIlmEntry;
            
            p64BitCountersHdls = (ix_counter_64bit_handle*)((ix_uint8*)(pContext->pIlmNhlfeSet64BitCountersHdls) + 
                                 ((ix_uint32)(pIlmSetEntry->OutSegId1)) * sizeof(ix_counter_64bit_handle) * MPLS_ILM_64BIT_COUNTERS_NUM);
        }
        else
        {
            p64BitCountersHdls = (ix_counter_64bit_handle*)((ix_uint8*)pContext->pIlm64BitCountersHdls + 
                       pContext->pLabelSpaceTable[pCcLspEntry->inHandle.ilmData.in_port].stats_offset +
                      (pCcLspEntry->inHandle.ilmData.label.label - pContext->pLabelSpaceTable[pCcLspEntry->inHandle.ilmData.in_port].min_label)
	                       * sizeof(ix_counter_64bit_handle) * MPLS_ILM_64BIT_COUNTERS_NUM);
        }
        
        for (idx = 0; idx < MPLS_ILM_64BIT_COUNTERS_NUM; idx ++) {
            /* Warning: 64-bit counters 32-bit base have to be at the beginning
            of counters structure (now only one 64-bit counter is used */
            err = ix_rm_counter_64bit_delete(*(p64BitCountersHdls + idx));
            if (err != IX_SUCCESS)
                ix_ossl_message_log("ix_cc_mpls_lsp_delete(): Warning: Could not remove 64bit counter handle.\n");
        }
                                                                         
#endif

        if ((pIlmEntry->flags && MPLS_FLAG_VALID_BIT) == 0) 
            ix_ossl_message_log("ix_cc_mpls_lsp_delete(): Warning: Ilm entry connected with "
                                "requested Lsp was marked as unused.\n");

        /* clear ILM entry */
        pIlmEntry->flags = 0;
        ix_ossl_memset(pIlmEntry, 0, sizeof(ix_cc_mpls_ilm_internal));
  
 
    } else  {

        /* delete RTMv4 route */
        err = ix_cc_rtmv4_delete_route(
                    pContext->hRtm,
                    pCcLspEntry->inHandle.fecData.NetAddr,
                    pCcLspEntry->inHandle.fecData.NetMask);
        
        if (err != IX_SUCCESS)
            ix_ossl_message_log("ix_cc_mpls_lsp_delete(): Warning: can't delete "
                                 "created IPv4 route.\n");

        /* delete RTMv4 next hop */
        err = ix_cc_rtmv4_delete_next_hop(
                    pContext->hRtm,
                    pCcLspEntry->inHandle.fecData.NextHopId);

        if (err != IX_SUCCESS)
            ix_ossl_message_log("ix_cc_mpls_lsp_delete(): Warning: can't delete "
                                 "created IPv4 next hop.\n");

#if defined (MPLS_COUNTERS_FEATURE)

        p64BitCountersHdls = (ix_counter_64bit_handle*)((ix_uint8*)(pContext->pFtn64BitCountersHdls) + 
                       pCcLspEntry->NhlfeHandle.outSegId * sizeof(ix_counter_64bit_handle) * MPLS_NHLFE_64BIT_COUNTERS_NUM);

        for (idx = 0; idx < MPLS_NHLFE_64BIT_COUNTERS_NUM; idx ++) {
            /* Warning: 64-bit counters 32-bit base have to be at the beginning
            of counters structure (now only one 64-bit counter is used) */
            err = ix_rm_counter_64bit_delete(*(p64BitCountersHdls + idx));
            if (err != IX_SUCCESS)
                ix_ossl_message_log("ix_cc_mpls_lsp_delete(): Warning: Could not remove 64bit counter handle.\n");
        }
                                                                         
#endif
    }

    err = _ix_cc_mpls_nhlfe_dec_ref_count(pCcLspEntry->NhlfeHandle, arg_pContext);
    if (err != IX_SUCCESS)
        ix_ossl_message_log("ix_cc_mpls_lsp_delete(): Warning: can't decrement "
                            "LSP refCount.\n");

    err = _ix_cc_mpls_nhlfe_del_lsp(pCcLspEntry, arg_pContext);
    if (err != IX_SUCCESS)
        ix_ossl_message_log("ix_cc_mpls_lsp_delete(): Warning: can't remove "
                            "LSP from the list.\n");

    pCcLspEntry->flags &= ~MPLS_FLAG_VALID_BIT;
    
    if (err != IX_SUCCESS)
          ix_ossl_message_log("ix_cc_mpls_lsp_delete(): Warning: can't decrease "
                              "reference count.\n");

    err = _ix_cc_mpls_lsp_add_to_free_list(arg_LspHandle,
                                           arg_pContext);

err_exit:

    IX_ERROR_CRT(ix_ossl_mutex_unlock(pContext->lspMutexId),
                 IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK, 
                 IX_ERROR_LEVEL_REMOTE);
    return err;

} /* ix_cc_mpls_lsp_delete */






/*
 * FUNCTION: ix_cc_mpls_lsp_modify()  */
/**
 * ABSTRACT: This function modifies mapping of NHLFEs for an LSP entry
 *
 * DESCRIPTION:This function modifies the NHLFE set mapping for a given LSP.
 *             The new NHLFE set can be specified either as one NHLFE 
 *             handle pointing to a NHLFE set entry, or as up to four 
 *             NHLFE handles pointing to regular (non-set) NHLFEs. 
 *
 * \param  arg_LspHandle        - handle identifying an LSP
 * \param  arg_nhlfeInfoType    - specifies weather function should use handle to predefined
 *                                NHLFE SET or the table of single NHLFEs during LSP modification.
 * \param  arg_nhlfeHandles_num - the number of NHLFE in a table to be cludtered in a set and
 *                                associated with the requested LSP entry.
 * \param  arg_nhlfeHandles     - table of nhlfe handles to be associated with the LSP entry.
 *                                If LSP entry is to be associated with NHLFE SET
 *                                this is the table with handles 
 *                                to all members of this NHLFE SET. 
 * \param  arg_nhlfeSetHandle   - (handle to predefined NHLFE SET)
 *                                handle to NHLFE set, the LSP entry is to be associated with.
 *                                If the LSP should be associated with the simple NHLFE,
 *                                the argument returns 0 handle (bladeId and outSegId == 0).
 * \param  arg_lspParamsSel     - flags denoting which fields of lsp has to be modified
 * \param  arg_lspParams        - pointer to the array of LSP parameters for each element of
 *                                the new NHLFE set
 * \param  arg_pContext         - pointer to the MPLS core component context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - LSP or NHLFE does not exist
 * \n         IX_CC_ERROR_RANGE                - LSP parameters are out of range
 * \n         IX_CC_MPLS_ERROR_INVALID_HANDLE  - invalid arg_LspHandle
 * \n         IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK - mutex lock failed
 * \n         IX_CC_MPLS_ERROR_INVALID_ENTRY   - invalid LSP parameters
 * \n
 * \n RTMv4 error codes:
 * \n         IX_CC_ERROR_NULL
 * \n         IX_CC_ERROR_ALIGN
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND
 * \n         IX_CC_ERROR_INTERNAL
 * \n         IX_CC_ERROR_DUPLICATE_ENTRY
 * \n         IX_CC_ERROR_CONFLICTING_ENTRY
 * \n         IX_CC_ERROR_FULL
 * \n         IX_CC_RTMV4_ERROR_INVALID_NH_INFO
 */
ix_error
ix_cc_mpls_lsp_modify(ix_cc_mpls_lsp_handle     arg_LspHandle,
                      ix_uint8                  arg_nhlfeInfoType,
                      ix_cc_mpls_nhlfe_handle  *arg_pNhlfeSetHandle,
                      ix_uint16                 arg_nhlfeHandles_num,
                      ix_cc_mpls_nhlfe_handle   arg_nhlfeHandles[MPLS_MAX_NHLFE_IN_SET],
                      ix_cc_mpls_lsp_param	arg_lspParams[MPLS_MAX_NHLFE_IN_SET],
                      void                     *arg_pContext)
{

    ix_error   err;
    ix_cc_mpls_cc_lsp        *pCcLspEntry;
    ix_cc_mpls_context       *pContext = arg_pContext;
    ix_cc_mpls_nhlfe          Nhlfe;
    ix_uint16                 NhlfeHandles_num;
    ix_cc_mpls_nhlfe_handle   NhlfeHandles[MPLS_MAX_NHLFE_IN_SET];

    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_lsp_modify(): got null Null MPLS context"));

    if (pContext->pCcLspBaseAddr == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("ix_cc_mpls_lsp_modify(): "
                                "CC LSP table has not been initialized."));
    if ((arg_nhlfeHandles_num < 1) || 
        (arg_nhlfeHandles_num > MPLS_MAX_NHLFE_IN_SET))
        return IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
                                ("ix_cc_mpls_lsp_modify(): nhlfeSetSize out of range"));

    /* add free CC LSP table's entry */
    if (arg_LspHandle.bladeId != pContext->localBladeId)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE,
                                ("ix_cc_mpls_lsp_modify(): wrong bladeId"));

    if (arg_pNhlfeSetHandle == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_lsp_modify(): Invalid pointer to the NHLFE SET handle"));

    IX_ERROR_C(ix_ossl_mutex_lock(pContext->lspMutexId, IX_OSSL_WAIT_FOREVER), err);

    if (err != IX_SUCCESS)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK,
                                ("ix_cc_mpls_lsp_modify(): can't lock mutex"));


    pCcLspEntry = &pContext->pCcLspBaseAddr[arg_LspHandle.LspIndex];

    if ((pCcLspEntry->flags & MPLS_FLAG_VALID_BIT) == 0)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                               ("ix_cc_mpls_lsp_modify(): LSP entry not found"));
        goto err_exit;
    }
    
    if (arg_nhlfeInfoType == IX_CC_MPLS_LSP_NHLFE_HANDLE_TYPE)
    {
        if (arg_nhlfeHandles == NULL)
        {
            err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                   ("ix_cc_mpls_lsp_modify(): NULL NHLFE handle"));
            goto err_exit;
        }

        NhlfeHandles_num = arg_nhlfeHandles_num;
        ix_ossl_memcpy(NhlfeHandles, arg_nhlfeHandles,
                       sizeof(ix_cc_mpls_nhlfe_handle)* NhlfeHandles_num);

        if (NhlfeHandles_num >= 2)
        {
            err = ix_cc_mpls_nhlfe_set_create(NhlfeHandles_num,
                                              NhlfeHandles,
                                              arg_pNhlfeSetHandle,
                                              arg_pContext);

            ON_ERROR_GOTO (err_exit, "could not create NHLFE set");
        }
        else
        {
            if  (pCcLspEntry->type == IX_CC_MPLS_LSP_ILM_TYPE)
            {
                err = ix_cc_mpls_nhlfe_query(NhlfeHandles[0],
                                             &Nhlfe,
                                             arg_pContext);
                ON_ERROR_GOTO (err_exit, "invalid first NHLFE handle");
            }

            if (NhlfeHandles[0].bladeId != pContext->localBladeId)
            {
                err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE,
                                       ("ix_cc_mpls_lsp_modify(): invalid NHLFE handle"));
                ON_ERROR_GOTO (err_exit, "invalid first NHLFE handle");
            }
            *arg_pNhlfeSetHandle = NhlfeHandles[0];
        }
    }
    else if (arg_nhlfeInfoType == IX_CC_MPLS_LSP_NHLFE_SET_HANDLE_TYPE)
    {
        err = ix_cc_mpls_nhlfe_set_query(*arg_pNhlfeSetHandle,
                                         &NhlfeHandles_num,
                                         NhlfeHandles,
                                         arg_pContext);
        ON_ERROR_GOTO (err_exit, "could not query NHLFE set");
    }
    else
    {
        err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_ENTRY,
                               ("ix_cc_mpls_lsp_modify(): invalid NHLFE data type."));
        goto err_exit;
    }

    if  (pCcLspEntry->type == IX_CC_MPLS_LSP_ILM_TYPE)
    {
        ix_cc_mpls_ilm_internal      *pIlmEntry;
        ix_uint32                      idx;
        ix_uint32                      IlmSetIndices[MPLS_MAX_NHLFE_IN_SET];
        ix_cc_mpls_ilm_set_internal   *pIlmSetEntry;
        ix_cc_mpls_ilm_internal       *NewIlmSetEntry[MPLS_MAX_NHLFE_IN_SET];
        ix_cc_mpls_ilm_set_free_slot  *pCurrentFreeIlmSet;

        /* get ILM table's entry */
        pIlmEntry = (ix_cc_mpls_ilm_internal*)((ix_uint8*)pContext->pIlmBaseAddr + 
                    pContext->pLabelSpaceTable[pCcLspEntry->inHandle.ilmData.in_port].table_offset +
                    pCcLspEntry->inHandle.ilmData.label.label * sizeof(ix_cc_mpls_ilm_internal));
	
        pIlmSetEntry = (ix_cc_mpls_ilm_set_internal*)pIlmEntry;

        if ((pIlmEntry->flags & MPLS_FLAG_VALID_BIT) == 0)
        {
            ix_ossl_message_log("ix_cc_mpls_lsp_modify(): Warning: Ilm entry connected with "
                                "requested Lsp was marked as unused.\n");
        }
        

        if (pIlmSetEntry->flags & MPLS_FLAG_SET_BIT)
        {
            IlmSetIndices[0] =  pIlmSetEntry->OutSegId1;
            IlmSetIndices[1] = (pIlmSetEntry->OutSegId2Hi << 8)  | pIlmSetEntry->OutSegId2Lo;
            IlmSetIndices[2] =  pIlmSetEntry->OutSegId3;
            IlmSetIndices[3] = (pIlmSetEntry->OutSegId4Hi << 16) | pIlmSetEntry->OutSegId4Lo;
            
            for (idx=0; idx<pIlmSetEntry->SetSize; idx ++)
            {
                NewIlmSetEntry[idx] = pContext->pIlmSetBaseAddr + 
                    IlmSetIndices[idx] *
                    sizeof(ix_cc_mpls_ilm_internal);
                
                ix_ossl_memset(NewIlmSetEntry[idx], 0, sizeof(ix_cc_mpls_ilm_internal));
                
                if (_ix_cc_mpls_ilm_set_add_to_free_list(IlmSetIndices[idx], pContext) != IX_SUCCESS)
                {
                    NhlfeHandles_num = (ix_uint16)idx;
                    ix_ossl_message_log("ix_cc_mpls_lsp_modify(): Warning: can't free ilm entries "
                                        "Out of memoty - some ILMs are lost.\n");
                    
                    break;
                }
            }
        }
        
        
        if ((arg_nhlfeInfoType == IX_CC_MPLS_LSP_NHLFE_HANDLE_TYPE  &&  NhlfeHandles_num == 1) ||
            !(pIlmSetEntry->flags & MPLS_FLAG_SET_BIT))
        {
            _ix_cc_mpls_lsp_copy_nhlfe_to_ilm(pIlmEntry, arg_lspParams[0], &Nhlfe);
	    pIlmEntry->flags &= ~MPLS_FLAG_SET_BIT;
        }
        else
        {
	    pCurrentFreeIlmSet = pContext->pFreeIlmSetListHead;
            /* check if there's enough space for all ILM entries */
            for (idx = 0; idx < NhlfeHandles_num; idx ++)
            {
                if (pCurrentFreeIlmSet == NULL)
                {
                    err = IX_ERROR_WARNING(IX_CC_ERROR_FULL,
                                           ("ix_cc_mpls_lsp_modify(): no space to create ilm set entry."));
                    goto err_free_nhlfe_set;
                }
                IlmSetIndices[idx] = pCurrentFreeIlmSet->IlmSetIndex;
                pCurrentFreeIlmSet = pCurrentFreeIlmSet->pNextFreeSlot;
                
                /* get new ILM SET table's entry */
                NewIlmSetEntry[idx] = pContext->pIlmSetBaseAddr + 
                    IlmSetIndices[idx] *
                    sizeof(ix_cc_mpls_ilm_internal);
            }
            
            /* get ILM entries from freelist */
            for (idx = 0; idx < NhlfeHandles_num; idx ++)
            {
                pCurrentFreeIlmSet            = pContext->pFreeIlmSetListHead;
                pContext->pFreeIlmSetListHead = pCurrentFreeIlmSet->pNextFreeSlot;
                ix_ossl_free(pCurrentFreeIlmSet);
            }
        
            err = _ix_cc_mpls_lsp_copy_nhlfe_set_to_ilm((ix_cc_mpls_ilm_set_internal*)pIlmEntry,
                                                        NhlfeHandles_num,
                                                        arg_lspParams,
                                                        NewIlmSetEntry,
                                                        IlmSetIndices,
                                                        NhlfeHandles,
                                                        pContext);
        }
    }
    else if (pCcLspEntry->type == IX_CC_MPLS_LSP_FEC_TYPE)
    {

        ix_cc_rtmv4_next_hop_info nextHopInfo;

        err = ix_cc_rtmv4_get_next_hop (pContext->hRtm,
                                        pCcLspEntry->inHandle.fecData.NextHopId,
                                       &nextHopInfo);

        ON_ERROR_GOTO (err_free_nhlfe_set, "Invalid NextHop associated with the given FEC");

        /* pointer to NHLFE or NHLFE set */
        nextHopInfo.l2Index = arg_pNhlfeSetHandle->outSegId;

        err = ix_cc_rtmv4_update_next_hop (pContext->hRtm,
                                           pCcLspEntry->inHandle.fecData.NextHopId,
                                           &nextHopInfo);

        ON_ERROR_GOTO (err_free_nhlfe_set, "Could not modify NextHop Info associated with the given FEC");

        if (NhlfeHandles_num >= 2)
        {
            ix_uint32 idx;
            for (idx = 0; idx < NhlfeHandles_num; idx ++)
            {
                _ix_cc_mpls_nhlfe_internal_modify(NhlfeHandles[idx],
                                                  0xFFFF,
                                                  arg_lspParams[idx],
                                                  arg_pContext);
            }
        }

    };

    err = _ix_cc_mpls_nhlfe_dec_ref_count(pCcLspEntry->NhlfeHandle, arg_pContext);
    ON_ERROR_GOTO (err_free_nhlfe_set, "Can't decrease LSP reference count");

    err = _ix_cc_mpls_nhlfe_del_lsp(pCcLspEntry, arg_pContext);
    ON_ERROR_GOTO (err_incr_ref_cnt, "Can't remove LSP from the list");

    
    pCcLspEntry->NhlfeHandle = *arg_pNhlfeSetHandle;

    err = _ix_cc_mpls_nhlfe_inc_ref_count(pCcLspEntry->NhlfeHandle, arg_pContext);
    if (err != IX_SUCCESS)
          ix_ossl_message_log("ix_cc_mpls_lsp_modify(): Warning: can't increase "
                              "LSP reference count.\n");

    err = _ix_cc_mpls_nhlfe_add_lsp(pCcLspEntry, arg_pContext);
    if (err != IX_SUCCESS)
          ix_ossl_message_log("ix_cc_mpls_lsp_modify(): Warning: can't add "
                              "LSP to the list.\n");
    
    
    /* modification of set into simple NHLFE has to return NULL in the arg_pNhlfeSetHandle */
    if (arg_nhlfeHandles_num == 1)
    {
        arg_pNhlfeSetHandle->bladeId = 0;
        arg_pNhlfeSetHandle->outSegId = 0;
    }
    
    goto err_exit;

    
    
err_incr_ref_cnt:
    if (_ix_cc_mpls_nhlfe_inc_ref_count(pCcLspEntry->NhlfeHandle, arg_pContext) != IX_SUCCESS)
          ix_ossl_message_log("ix_cc_mpls_lsp_modify(): Warning: can't increase "
                              "LSP reference count.\n");
   

err_free_nhlfe_set:
    if (arg_nhlfeInfoType == IX_CC_MPLS_LSP_NHLFE_HANDLE_TYPE) 
    {
        if (arg_nhlfeHandles_num >= 2)
        {
            err = ix_cc_mpls_nhlfe_set_delete(
                         *arg_pNhlfeSetHandle,
                          arg_pContext);

            if (err != IX_SUCCESS)
                ix_ossl_message_log("ix_cc_mpls_lsp_modify(): Warning: can't delete "
                                    "created NHLFE set.\n");
       }
    }
    
err_exit:
    IX_ERROR_CRT(ix_ossl_mutex_unlock(pContext->lspMutexId),
                 IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK, 
                 IX_ERROR_LEVEL_REMOTE);

	return err;
} /* ix_cc_mpls_lsp_modify */






/*
 * FUNCTION: ix_cc_mpls_lsp_query()  */
/**
 * ABSTRACT: This function reads and LSP description.
 *
 * DESCRIPTION: This function reads and LSP description. 
 *
 * \param  arg_LspHandle - handle identifying an LSP
 * \param  arg_pLsp      - pointer to the buffer for storing the LSP 
 *                         information returned by the function.
 * \param  arg_pContext  - pointer to the MPLS core component context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_MPLS_ERROR_INVALID_HANDLE  - invalid arg_LspHandle
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - LSP or NHLFE does not exist
 * \n         IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK - mutex lock failed
 */
ix_error
ix_cc_mpls_lsp_query(ix_cc_mpls_lsp_handle  arg_LspHandle,
                     ix_cc_mpls_lsp        *arg_pLsp,
                     void                  *arg_pContext)
{

    ix_error                      err      = IX_SUCCESS;
    ix_cc_mpls_context           *pContext = arg_pContext;
    ix_cc_mpls_cc_lsp            *pCcLspEntry;
    ix_cc_mpls_ilm_internal      *pIlmEntry;
    ix_uint16                     NhlfeHandles_num;
    ix_cc_mpls_nhlfe_handle       NhlfeHandles[MPLS_MAX_NHLFE_IN_SET];

    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_lsp_query(): got null MPLS context"));

    /* verify input arguments */
    if (arg_pLsp == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_lsp_query(): got null LSP pointer"));

    if (pContext->pCcLspBaseAddr == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("ix_cc_mpls_lsp_query(): "
                                "CC LSP table has not been initialized."));

    /* add free CC LSP table's entry */
    if (arg_LspHandle.bladeId != pContext->localBladeId)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_lsp_query(): wrong bladeId"));

    IX_ERROR_C(ix_ossl_mutex_lock(pContext->lspMutexId, IX_OSSL_WAIT_FOREVER), err);

    if (err != IX_SUCCESS)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK,
                                ("ix_cc_mpls_lsp_query(): can't lock mutex"));

    pCcLspEntry = &pContext->pCcLspBaseAddr[arg_LspHandle.LspIndex];

    if ((pCcLspEntry->flags & MPLS_FLAG_VALID_BIT) == 0)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                               ("ix_cc_mpls_lsp_query(): LSP entry is not valid"));
        goto err_exit;
    }

    arg_pLsp->lspType        = pCcLspEntry->type;
    arg_pLsp->nhlfeInfoType  = IX_CC_MPLS_LSP_NHLFE_HANDLE_TYPE;

    if  (pCcLspEntry->type == IX_CC_MPLS_LSP_ILM_TYPE) 
    {
        /* get new ILM table's entry */
        pIlmEntry = (ix_cc_mpls_ilm_internal*)((ix_uint8*)pContext->pIlmBaseAddr + 
                     pContext->pLabelSpaceTable[pCcLspEntry->inHandle.ilmData.in_port].table_offset +
                    (pCcLspEntry->inHandle.ilmData.label.label - pContext->pLabelSpaceTable[pCcLspEntry->inHandle.ilmData.in_port].min_label)
                           * sizeof(ix_cc_mpls_ilm_internal));

        if ((pIlmEntry->flags && MPLS_FLAG_VALID_BIT) == 0) {

            ix_ossl_message_log("ix_cc_mpls_lsp_query(): Warning: Ilm entry connected with "
                                "requested Lsp was marked as unused.\n");
            err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                                   ("ix_cc_mpls_lsp_query(): given entry doesn not exist"));
            
            goto err_exit;
        }
        
        arg_pLsp->inSeg.ilm.in_port   = pCcLspEntry->inHandle.ilmData.in_port;
        arg_pLsp->inSeg.ilm.label     = pCcLspEntry->inHandle.ilmData.label;
        arg_pLsp->inSeg.ilm.operation = (ix_uint8)(pIlmEntry->operation);


        if ((pIlmEntry->flags && MPLS_FLAG_SET_BIT) != 0) 
        {
            ix_cc_mpls_ilm_set_internal *pIlmSetEntry = (
                     ix_cc_mpls_ilm_set_internal *)pIlmEntry;
            /* get new ILM table's entry */
            pIlmEntry = pContext->pIlmSetBaseAddr + 
                            ((ix_uint32)(pIlmSetEntry->OutSegId1)) * 
                            sizeof(ix_cc_mpls_ilm_internal);          
        }
     
        /* LSP flags                        */
        arg_pLsp->lspParams[0].flags         = pIlmEntry->flags;
        /* LSP MTU                          */
        arg_pLsp->lspParams[0].maxLabPktSize = (ix_uint16)(pIlmEntry->maxLabPktSize);
        /* ttl decrement value              */
        arg_pLsp->lspParams[0].ttl           = (ix_uint8)(pIlmEntry->ttl);
        /* index into the EXP2PHB table     */
        arg_pLsp->lspParams[0].exp2phbIdx    = (ix_uint16)(pIlmEntry->exp2phb);
        /* index into the PHB2EXP table     */
        arg_pLsp->lspParams[0].phb2expIdx    = (ix_uint16)(pIlmEntry->phb2exp);

    } else  {
        arg_pLsp->inSeg.fec.type = IX_CC_MPLS_FEC_TYPE_IPV4;
        arg_pLsp->inSeg.fec.fec.ip4fec.ipAddress = pCcLspEntry->inHandle.fecData.NetAddr;
        arg_pLsp->inSeg.fec.fec.ip4fec.netMask   = pCcLspEntry->inHandle.fecData.NetMask;
        arg_pLsp->inSeg.fec.fec.ip4fec.nexthopId = pCcLspEntry->inHandle.fecData.NextHopId;
    }    

    err = ix_cc_mpls_nhlfe_set_query(pCcLspEntry->NhlfeHandle,
                                    &NhlfeHandles_num,
                                     NhlfeHandles,
                                     arg_pContext);

    switch (IX_ERROR_GET_CODE(err)) 
    {
        case IX_CC_ERROR_ENTRY_NOT_FOUND:
        
            err = IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                   ("ix_cc_mpls_lsp_query(): Invalid NHLFE associated with the current LSP"));
            goto err_exit;

            break;

        case IX_CC_MPLS_ERROR_ENTRY_NHLFE:

            arg_pLsp->nhlfeSetHandle.bladeId  = 0;
            arg_pLsp->nhlfeSetHandle.outSegId = 0;
            arg_pLsp->nhlfe_num = 1;
            arg_pLsp->outSeg.nhlfeHandle[0] = pCcLspEntry->NhlfeHandle;

            err = IX_SUCCESS;
            if  (pCcLspEntry->type == IX_CC_MPLS_LSP_FEC_TYPE) 
            {
                err = _ix_cc_mpls_nhlfe_query_params( pCcLspEntry->NhlfeHandle,
                                                     &(arg_pLsp->lspParams[0]),
                                                      arg_pContext);

                ON_ERROR_GOTO (err_exit, "cant read an NHLFE entry attached to the given LSP FEC");
            }
            break;

        case IX_SUCCESS:

            arg_pLsp->nhlfe_num = NhlfeHandles_num;

            ix_ossl_memcpy(arg_pLsp->outSeg.nhlfeHandle,
                           NhlfeHandles,
                           NhlfeHandles_num * sizeof(ix_cc_mpls_nhlfe_handle));

            arg_pLsp->nhlfeSetHandle = pCcLspEntry->NhlfeHandle;

            break;
    }

err_exit:

    IX_ERROR_CRT(ix_ossl_mutex_unlock(pContext->lspMutexId),
                 IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK, 
                 IX_ERROR_LEVEL_REMOTE);

    return err;
} /* ix_cc_mpls_lsp_query */






/*
 * FUNCTION: ix_cc_mpls_lsp_stats_query()  */
/**
 * ABSTRACT: This function reads LSP statistics.
 *
 * DESCRIPTION: This function gets LSP statistics.
 *
 * \param  arg_LspHandle     - handle identifying an LSP
 * \param  arg_LspStatistics - pointer to the buffer for storing 
 *                             the LSP statistics returned by the function.  
 * \param  arg_pContext      - pointer to the MPLS core component context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                     - invalid pointer
 * \n         IX_CC_MPLS_ERROR_INVALID_HANDLE      - invalid arg_LspHandle
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND          - LSP does not exist
 * \n         IX_CC_MPLS_ERROR_STATS_NOT_SUPPORTED - statistics are not supported
 * \n         IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK     - mutex lock failed
 */
ix_error
ix_cc_mpls_lsp_stats_query(ix_cc_mpls_lsp_handle  arg_LspHandle,
                           ix_cc_mpls_lsp_stats  *arg_pLspStatistics,
                           void                  *arg_pContext)
{

#if !defined(MPLS_COUNTERS_FEATURE)

    return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_STATS_NOT_SUPPORTED,
                            ("ix_cc_mpls_lsp_stats_query(): Statistics are not supported."));

#else

    ix_error                      err      = IX_SUCCESS;
    ix_cc_mpls_context           *pContext = arg_pContext;
    ix_uint32                     idx;
    ix_cc_mpls_cc_lsp            *pCcLspEntry;
    ix_cc_mpls_ilm_internal      *pIlmEntry;
    ix_counter_64bit_handle      *p64BitCountersHdls;

    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_lsp_stats_query(): got null MPLS context"));

    /* verify input arguments */
    if (arg_pLspStatistics == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_lsp_stats_query(): got null LSP statistics pointer"));

    if (pContext->pCcLspBaseAddr == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_lsp_stats_query(): CC LSP table has not been initialized."));

    /* add free CC LSP table's entry */
    if (arg_LspHandle.bladeId != pContext->localBladeId)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE, 
                                ("ix_cc_mpls_lsp_stats_query(): wrong bladeId"));


    IX_ERROR_C(ix_ossl_mutex_lock(pContext->lspMutexId, IX_OSSL_WAIT_FOREVER), err);

    if (err != IX_SUCCESS)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK,
                                ("ix_cc_mpls_lsp_stats_query(): can't lock mutex"));

    pCcLspEntry = &pContext->pCcLspBaseAddr[arg_LspHandle.LspIndex];

    if ((pCcLspEntry->flags & MPLS_FLAG_VALID_BIT) == 0)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                               ("ix_cc_mpls_lsp_stats_query(): improper LSP entry"));
        goto err_exit;
    }

    if (pCcLspEntry->type == IX_CC_MPLS_LSP_FEC_TYPE)
    {
        ix_cc_mpls_nhlfe_stats NhlfeStats;

        ix_cc_mpls_nhlfe_stats_query(pCcLspEntry->NhlfeHandle, &NhlfeStats, arg_pContext);

        arg_pLspStatistics->mplsRxOctets    = NhlfeStats.mplsNhlfeOctets;
        arg_pLspStatistics->mplsRxPkts      = NhlfeStats.mplsNhlfePkts;
        arg_pLspStatistics->mplsErrorOctets = NhlfeStats.mplsNhlfeErrorOctets;
        arg_pLspStatistics->mplsErrorPkts   = NhlfeStats.mplsNhlfeErrorPkts;
        arg_pLspStatistics->mplsInTxOctets  = NhlfeStats.mplsNhlfeOctets + NhlfeStats.mplsNhlfeFragmOctetsDelta;
        arg_pLspStatistics->mplsInTxPkts    = NhlfeStats.mplsNhlfePkts + NhlfeStats.mplsNhlfeFragmPktsDelta;
        
        err = IX_SUCCESS;
        goto err_exit;
    }
    
    pIlmEntry = (ix_cc_mpls_ilm_internal*)((ix_uint8*)pContext->pIlmBaseAddr + 
                 pContext->pLabelSpaceTable[pCcLspEntry->inHandle.ilmData.in_port].table_offset +
                (pCcLspEntry->inHandle.ilmData.label.label - pContext->pLabelSpaceTable[pCcLspEntry->inHandle.ilmData.in_port].min_label)
                 * sizeof(ix_cc_mpls_ilm_internal));

    if ((pIlmEntry->flags & MPLS_FLAG_VALID_BIT) == 0)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                               ("ix_cc_mpls_lsp_stats_query(): improper ILM entry"));
        goto err_exit;
    }

    if (pIlmEntry->flags & MPLS_FLAG_SET_BIT) 
    {
        ix_cc_mpls_ilm_set_internal *pIlmSetEntry = (ix_cc_mpls_ilm_set_internal*) pIlmEntry;

        p64BitCountersHdls = (ix_counter_64bit_handle*)((ix_uint8*)pContext->pIlmNhlfeSet64BitCountersHdls + 
                              ((ix_uint32)(pIlmSetEntry->OutSegId1)) * sizeof(ix_counter_64bit_handle) * MPLS_ILM_64BIT_COUNTERS_NUM);
    }
    else
    {
        p64BitCountersHdls = (ix_counter_64bit_handle*)((ix_uint8*)pContext->pIlm64BitCountersHdls + 
                              pContext->pLabelSpaceTable[pCcLspEntry->inHandle.ilmData.in_port].stats_offset +
                             (pCcLspEntry->inHandle.ilmData.label.label - pContext->pLabelSpaceTable[pCcLspEntry->inHandle.ilmData.in_port].min_label)
 	                          * sizeof(ix_counter_64bit_handle) * MPLS_ILM_64BIT_COUNTERS_NUM);
    }

    for(idx = 0; idx < MPLS_ILM_64BIT_COUNTERS_NUM; idx ++)
    {
        err = ix_rm_counter_64bit_get_value(*(p64BitCountersHdls + idx), (ix_uint64*)arg_pLspStatistics + idx);
        if (err != IX_SUCCESS)
        {
            err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE, 
                                  ("ix_cc_mpls_lsp_stats_query(): Invalid counter's handle"));
            break;
        }
    }

    arg_pLspStatistics->mplsInTxOctets  += arg_pLspStatistics->mplsRxOctets;
    arg_pLspStatistics->mplsInTxPkts    += arg_pLspStatistics->mplsRxPkts;
    
    
err_exit:
    IX_ERROR_CRT(ix_ossl_mutex_unlock(pContext->lspMutexId),
                 IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK, 
                 IX_ERROR_LEVEL_REMOTE);
    return err;

#endif
} /* ix_cc_mpls_lsp_stats_query */








/*
 * FUNCTION: ix_cc_mpls_lsp_purge()  */
/**
 * ABSTRACT: This function removes all entries from LSP table.
 *
 * DESCRIPTION: This function deletes all LSP definitions,
 *              together with their corresponding IP Routing table,
 *              ILM and NHLFE entries. 
 *
 * \param  arg_pContext      - pointer to the MPLS core component context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK - mutex lock failed
 */
ix_error
ix_cc_mpls_lsp_purge(void *arg_pContext)
{
    ix_error                 err;
    ix_uint32                idx;
    ix_cc_mpls_context      *pContext = arg_pContext;
    ix_cc_mpls_cc_lsp       *pCcLspEntry;
    ix_cc_mpls_lsp_handle    LspHandle;

    
    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_lsp_purge(): got null MPLS context"));

    if (pContext->pCcLspBaseAddr == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_lsp_purge(): CC LSP table has not been initialized."));

    LspHandle.bladeId = pContext->localBladeId;

    IX_ERROR_C(ix_ossl_mutex_lock(pContext->lspMutexId, IX_OSSL_WAIT_FOREVER), err);
    
    if (err != IX_SUCCESS)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK,
                                ("ix_cc_mpls_lsp_purge(): can't lock mutex"));

    /* Check if all entries have refCount field set to zero */
    for (idx = 0; idx < pContext->maxLspEntries; idx++)
    {
        pCcLspEntry = &pContext->pCcLspBaseAddr[idx];
        if ((pCcLspEntry->flags & MPLS_FLAG_VALID_BIT) == 0) 
            continue;

        LspHandle.LspIndex = idx;
        err = ix_cc_mpls_lsp_delete(LspHandle, arg_pContext);
    }

    IX_ERROR_CRT(ix_ossl_mutex_unlock(pContext->lspMutexId),
                 IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK, 
                 IX_ERROR_LEVEL_REMOTE);

    return err;
} /* ix_cc_mpls_lsp_purge */





/*
 * FUNCTION: ix_cc_mpls_lsp_init()  */
/**
 * ABSTRACT: function initializes the LSP part of MPLS CC
 *
 * DESCRIPTION: This function creates and initializes the LSP table. It should
 *              be called and returned successfully before any other function in
 *              this library can be called. It returns the base address of the
 *              LSP table.
 *
 * \param  arg_pContext   - pointer to the MPLS CC context
 *                          that will be passed to the core component API function
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n           IX_CC_ERROR_NULL            - invalid pointer
 * \n           IX_CC_ERROR_DUPLICATE_ENTRY - LSP table already exist
 * \n           IX_CC_ERROR_OOM_SYSTEM      - memory allocation failure
 * \n           IX_CC_ERROR_OOM_SRAM        - memory allocation failure
 */
ix_error
ix_cc_mpls_lsp_init(void *arg_pContext)
{

    ix_uint32                idx;
    ix_error                 err       = IX_SUCCESS;
    ix_cc_mpls_lsp_handle    LspHandle;
    ix_cc_mpls_context      *pContext  = arg_pContext;
    

    /* Check the input arguments */
    if (pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("ix_cc_mpls_lsp_init(): NULL MPLS context"));

    /* Create mutex for critical section. */
    IX_ERROR_C(ix_ossl_mutex_init(OS_MUTEX_UNLOCK, &(pContext->lspMutexId)),err);
    if (err != IX_SUCCESS)
    {
        ix_ossl_free(pContext);
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_OSSL_MUTEX_INIT,
                                ("ix_cc_mpls_lsp_init(): can't initialize LSP mutex"));
    }

    /* Allocate memory for label config table */
    err = ix_rm_mem_alloc(IX_MEMORY_TYPE_SRAM,
                          pContext->LabelSpaceTableMemoryChannel,
                          pContext->maxLocalPorts *
                          sizeof(ix_cc_mpls_label_space),
                          (void **)&pContext->pLabelSpaceTable);

    if (err != IX_SUCCESS) {
         _ix_cc_mpls_lsp_free_resources(pContext);
         return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
                                 ("ix_cc_mpls_lsp_init(): Out of Memory "
                                  "to allocate table with Label Space config"));
    }

    ix_ossl_memset((void*)pContext->pLabelSpaceTable, 0,
                   (ix_ossl_size_t)pContext->maxLocalPorts *
                            sizeof(ix_cc_mpls_label_space));

    _ix_cc_mpls_lsp_get_label_space_config(pContext);

    
    /* Allocate memory and initialize CC LSP table */
    pContext->pCcLspBaseAddr = ix_ossl_malloc(pContext->maxLspEntries *
                                                    sizeof(ix_cc_mpls_cc_lsp));

    if (pContext->pCcLspBaseAddr == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
                                ("ix_cc_mpls_lsp_init(): Out of Memory for Allocating CC LSP table"));

    ix_ossl_memset((void*)pContext->pCcLspBaseAddr,0,
                   (ix_ossl_size_t) (pContext->maxLspEntries *
                                     sizeof(ix_cc_mpls_cc_lsp)));

    LspHandle.bladeId = pContext->localBladeId;
    pContext->pFreeLspListHead = NULL;

    for (idx = 1; idx < pContext->maxLspEntries; idx++)
    {
        LspHandle.LspIndex = idx;
        _ix_cc_mpls_lsp_add_to_free_list(LspHandle, pContext);
    }

    
    /* Allocate memory and initialize ILM table */
    err = ix_rm_mem_alloc(pContext->IlmTableMemoryType,
                          pContext->IlmTableMemoryChannel,
                          pContext->maxIlmEntries *
                          sizeof(ix_cc_mpls_ilm_internal),
                          (void **)&pContext->pIlmBaseAddr);
    
    if (err != IX_SUCCESS) {
         _ix_cc_mpls_lsp_free_resources(pContext);
         return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
                                 ("ix_cc_mpls_lsp_init(): Out of Memory for Allocating CC LSP table"));
    }
    
    ix_ossl_memset((void*)pContext->pIlmBaseAddr, 0,
                   (ix_ossl_size_t)(pContext->maxIlmEntries *
                                    sizeof(ix_cc_mpls_ilm_internal)));


    /* Allocate memory and initialize ILM SET table */
    err = ix_rm_mem_alloc(pContext->IlmTableMemoryType,
                          pContext->IlmTableMemoryChannel,
                          pContext->maxIlmSetEntries *
                          sizeof(ix_cc_mpls_ilm_internal),
                          (void **)&pContext->pIlmSetBaseAddr);
    
    if (err != IX_SUCCESS) {
         _ix_cc_mpls_lsp_free_resources(pContext);
         return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
                                 ("ix_cc_mpls_lsp_init(): Out of Memory for Allocating CC LSP table"));
    }

   
    ix_ossl_memset((void*)pContext->pIlmSetBaseAddr, 0,
                   (ix_ossl_size_t)(pContext->maxIlmSetEntries *
                                    sizeof(ix_cc_mpls_ilm_internal)));

    pContext->pFreeIlmSetListHead = NULL;

    for (idx = 1; idx < pContext->maxIlmSetEntries; idx++)
        _ix_cc_mpls_ilm_set_add_to_free_list(idx, pContext);



#ifdef MPLS_COUNTERS_FEATURE

    {
        /* Allocate memory and initialize 32-bit ILM counters */
        err = ix_rm_mem_alloc(IX_MEMORY_TYPE_SRAM,
                                     pContext->IlmCountersMemoryChannel,
                                     pContext->maxIlmEntries *
                                     sizeof(ix_cc_mpls_ilm_counters_internal),
                                     (void **)&pContext->pIlmCounters);
        if (err != IX_SUCCESS)
        {
             _ix_cc_mpls_lsp_free_resources(pContext);
             return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
                                     ("ix_cc_mpls_lsp_init(): Out of Memory "
                                      "for Allocating 32 bit ILM counters table"));
        }
        ix_ossl_memset((void *)pContext->pIlmCounters, 0,
                       (ix_ossl_size_t)(pContext->maxIlmEntries *
                                        sizeof(ix_cc_mpls_ilm_counters_internal)));

        
        /* Allocate memory and initialize 64-bit ILM counters handler array */
        pContext->pIlm64BitCountersHdls = (ix_counter_64bit_handle *)
                                          ix_ossl_malloc( pContext->maxIlmEntries * 
                                                          sizeof(ix_counter_64bit_handle) * MPLS_ILM_64BIT_COUNTERS_NUM);
        if (pContext->pIlm64BitCountersHdls == NULL)
        {
            _ix_cc_mpls_lsp_free_resources(pContext);
            return IX_ERROR_REMOTE(IX_CC_ERROR_OOM_SYSTEM, 
                     ("ix_cc_mpls_lsp_init(): Out of Memory for allocating 64 bit ILM counters handles"));
        }
        ix_ossl_memset((void *)pContext->pIlm64BitCountersHdls, 0,
                       pContext->maxIlmEntries * sizeof(ix_counter_64bit_handle) * MPLS_ILM_64BIT_COUNTERS_NUM);
                                                                              

        /* Allocate memory and initialize 32-bit ILM NHLFE set counters */
        err = ix_rm_mem_alloc(IX_MEMORY_TYPE_SRAM,
                                     pContext->IlmCountersMemoryChannel,
                                     pContext->maxIlmSetEntries *
                                     sizeof(ix_cc_mpls_ilm_counters_internal),
                                     (void **)&pContext->pIlmNhlfeSetCounters);
        if (err != IX_SUCCESS)
        {
             _ix_cc_mpls_lsp_free_resources(pContext);
             return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
                                     ("ix_cc_mpls_lsp_init(): Out of Memory for "
                                      "Allocating 32 bit ILM NHLFE set counters table"));
        }
        ix_ossl_memset((void *)pContext->pIlmNhlfeSetCounters, 0,
                       (ix_ossl_size_t)(pContext->maxIlmSetEntries *
                                        sizeof(ix_cc_mpls_ilm_counters_internal)));

        
        /* Allocate memory and initialize 64-bit ILM NHLFE set counters handler array */
        pContext->pIlmNhlfeSet64BitCountersHdls = (ix_counter_64bit_handle *)
                                          ix_ossl_malloc( pContext->maxIlmSetEntries * 
                                                          sizeof(ix_counter_64bit_handle) * MPLS_ILM_64BIT_COUNTERS_NUM);
        if (pContext->pIlmNhlfeSet64BitCountersHdls == NULL)
        {
            _ix_cc_mpls_lsp_free_resources(pContext);
            return IX_ERROR_REMOTE(IX_CC_ERROR_OOM_SYSTEM, 
                     ("ix_cc_mpls_lsp_init(): Out of Memory for allocating 64 bit ILM NHLFE set counters handles"));
        }
        ix_ossl_memset((void *)pContext->pIlmNhlfeSet64BitCountersHdls, 0,
                       pContext->maxIlmSetEntries * sizeof(ix_counter_64bit_handle) * MPLS_ILM_64BIT_COUNTERS_NUM);


        /* Allocate memory and initialize 32-bit input port counters */
        err = ix_rm_mem_alloc(IX_MEMORY_TYPE_SRAM,
                                     pContext->IlmCountersMemoryChannel,
                                     pContext->maxLocalPorts *
                                     sizeof(ix_cc_mpls_input_port_stats),
                                     (void **)&pContext->pInputPortCounters);
        if (err != IX_SUCCESS)
        {
             _ix_cc_mpls_lsp_free_resources(pContext);
             return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
                                     ("ix_cc_mpls_lsp_init(): Out of Memory "
                                      "for allocating 32 bit Input Port Counters table"));
        }
        ix_ossl_memset((void *)pContext->pInputPortCounters, 0,
                       (ix_ossl_size_t)(pContext->maxLocalPorts *
                                        sizeof(ix_cc_mpls_input_port_stats)));

        
        /* Allocate memory and initialize 64-bit input port counters handler array */
        pContext->pInputPort64BitCountersHdls = (ix_counter_64bit_handle *)
                                                 ix_ossl_malloc( pContext->maxLocalPorts * 
                                                                 sizeof(ix_counter_64bit_handle) * MPLS_ILM_64BIT_COUNTERS_NUM);
        if (pContext->pInputPort64BitCountersHdls == NULL)
        {
            _ix_cc_mpls_lsp_free_resources(pContext);
            return IX_ERROR_REMOTE(IX_CC_ERROR_OOM_SYSTEM, 
                     ("ix_cc_mpls_lsp_init(): Out of Memory for allocating 64 bit input port counters handles"));
        }
        ix_ossl_memset((void *)pContext->pInputPort64BitCountersHdls, 0,
                       pContext->maxLocalPorts * sizeof(ix_counter_64bit_handle) * MPLS_ILM_64BIT_COUNTERS_NUM);
                                                                                     
    }
#endif /* MPLS_COUNTERS_FEATURE */

    /* initialize handle to RTM */
    pContext->hRtm =  g_pIpv4Context->hRtm;
    
    return err;
} /* ix_cc_mpls_lsp_init */






/*
 * FUNCTION: _ix_cc_mpls_lsp_free_resources()  */
/**
 * ABSTRACT: These function releases all resources taken by LSP library
 *
 * DESCRIPTION: This function frees all memory and resources taken by
 *              LSP library
 *
 * \param  arg_pContext - a pointer to the MPLS LSP context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
void
_ix_cc_mpls_lsp_free_resources(void *arg_pContext)
{

    ix_cc_mpls_context* pContext = (ix_cc_mpls_context*) arg_pContext;

    if (pContext == NULL) {
        ix_ossl_message_log("_ix_cc_mpls_lsp_free_resources(): Warning -"
                            " got NULL MPLS context.\n");
        return;
    }

    if (pContext->pCcLspBaseAddr == NULL) {

        ix_ossl_message_log("_ix_cc_mpls_lsp_free_resources(): Warning -"
                            " CC LSP table has not been initialized.\n");
    } else  {

        if (ix_cc_mpls_lsp_purge(pContext) != IX_SUCCESS) 
            ix_ossl_message_log("_ix_cc_mpls_lsp_free_resources(): Warning :"
                                "can't purge LSP table.\n");

        ix_ossl_free(pContext->pCcLspBaseAddr);
    }

    if (ix_cc_mpls_lsp_purge(arg_pContext) != IX_SUCCESS) {
        ix_ossl_message_log("_ix_cc_mpls_lsp_free_resources(): Warning -"
                            "error during purging LSP table.\n");
    };


#ifdef MPLS_COUNTERS_FEATURE

    /* Free 64-bit ILM counters handle array */
    if (pContext->pIlm64BitCountersHdls != NULL)
    {
        ix_ossl_free(pContext->pIlm64BitCountersHdls);
    }
    
    /* Free 32-bit ILM counters table */
    if (pContext->pIlmCounters != NULL)
    {
        if (ix_rm_mem_free(pContext->pIlmCounters) != IX_SUCCESS)
			ix_ossl_message_log("_ix_cc_mpls_lsp_free_resources(): Warning -"
                                " Resource Manager failed to free ILM counters table.\n");
    }

    
    /* Free 64-bit ILM NHLFE set counters handle array */
    if (pContext->pIlmNhlfeSet64BitCountersHdls != NULL)
    {
        ix_ossl_free(pContext->pIlmNhlfeSet64BitCountersHdls);
    }
    
    /* Free 32-bit ILM counters table */
    if (pContext->pIlmNhlfeSetCounters != NULL)
    {
        if (ix_rm_mem_free(pContext->pIlmNhlfeSetCounters) != IX_SUCCESS)
			ix_ossl_message_log("_ix_cc_mpls_lsp_free_resources(): Warning -"
                                " Resource Manager failed to free ILM NHLFE set counters table.\n");
    }


    /* Free 64-bit input port counters handle array */
    if (pContext->pInputPort64BitCountersHdls != NULL)
    {
        ix_ossl_free(pContext->pInputPort64BitCountersHdls);
    }
    
    /* Free 32-bit ILM counters table */
    if (pContext->pInputPortCounters != NULL)
    {
        if (ix_rm_mem_free(pContext->pInputPortCounters) != IX_SUCCESS)
			ix_ossl_message_log("_ix_cc_mpls_lsp_free_resources(): Warning -"
                                " Resource Manager failed to free input port counters table.\n");
    }

#endif /* MPLS_COUNTERS_FEATURE */

    /* Free memory used by LSP ILM SET array */
    if (pContext->pIlmSetBaseAddr != NULL)
    {
        if (ix_rm_mem_free(pContext->pIlmSetBaseAddr) != IX_SUCCESS)
			ix_ossl_message_log("_ix_cc_mpls_lsp_free_resources(): Warning -"
                                " Resource Manager ILM SET table free failed.\n");
    }
    
    /* Free memory used by LSP ILM array */
    if (pContext->pIlmBaseAddr != NULL)
    {
        if (ix_rm_mem_free(pContext->pIlmBaseAddr) != IX_SUCCESS)
			ix_ossl_message_log("_ix_cc_mpls_lsp_free_resources(): Warning -"
                                " Resource Manager ILM table free failed.\n");
    }

    pContext->pIlmBaseAddr = NULL;

    if (pContext->pLabelSpaceTable != NULL)
    {
        if (ix_rm_mem_free(pContext->pLabelSpaceTable) != IX_SUCCESS)
			ix_ossl_message_log("_ix_cc_mpls_lsp_free_resources(): Warning -"
               " Resource Manager failed to free Label Space config table.\n");
    }


    /* Free free LSP CC list */
    if (pContext->pFreeLspListHead != NULL)
    {
        ix_cc_mpls_cc_lsp_free_slot* pFreeLspEntry = pContext->pFreeLspListHead;
        ix_cc_mpls_cc_lsp_free_slot* pNextFreeLspEntry = NULL;

        while (pFreeLspEntry != NULL)
        {
            /* Remove slot from free NHLFE entries list */
            pNextFreeLspEntry = pFreeLspEntry->pNextFreeSlot;
            ix_ossl_free((void *)pFreeLspEntry);
            pFreeLspEntry = pNextFreeLspEntry;
        }
        pContext->pFreeLspListHead = NULL;

    }

    return;
} /* _ix_cc_mpls_lsp_free_resources */






/*
 * FUNCTION: _ix_cc_mpls_lsp_fini()  */
/**
 * ABSTRACT: removes resources taken to generate LSP
 *
 * DESCRIPTION: releases the LSP table.
 *
 * \param  arg_pContext - a pointer to the MPLS context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_lsp_fini(void *arg_pContext)
{
    ix_cc_mpls_context* pContext = (ix_cc_mpls_context*) arg_pContext;
    ix_error            err;
    
    IX_ERROR_CRT(ix_ossl_mutex_lock(pContext->lspMutexId, IX_OSSL_WAIT_FOREVER),
                 IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK, IX_ERROR_LEVEL_LOCAL);
    
    _ix_cc_mpls_lsp_free_resources(pContext);
        
    IX_ERROR_CT(ix_ossl_mutex_unlock(pContext->lspMutexId), err,
                IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK, IX_ERROR_LEVEL_LOCAL);

    /* Remove mutex for critical section. */
    IX_ERROR_CRT(ix_ossl_mutex_fini(pContext->lspMutexId),
                 IX_CC_MPLS_ERROR_OSSL_MUTEX_FINI, 
                 IX_ERROR_LEVEL_REMOTE);
    
    return IX_SUCCESS;
} /* _ix_cc_mpls_lsp_fini */






/*
 * FUNCTION: _ix_cc_mpls_lsp_get_label_space_config()  */
/**
 * ABSTRACT: gets range of label per port which will be supported
 *
 * DESCRIPTION: This function retrieves label range info per each port and
 *              fills label space table.
 *
 * \param  arg_pContext  - pointer to component context memory
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 *
 * \n Error codes:
 * \n       IX_CC_ERROR_UNINIT - MPLS Context parameters uninitialized
 */
PRIVATE ix_error
_ix_cc_mpls_lsp_get_label_space_config(void *arg_pContext)
{
    ix_int32            idx = 0;
    ix_uint32           NextOffset;
    ix_cc_mpls_context *pContext  = (ix_cc_mpls_context*)arg_pContext;

#ifdef IX_INCLUDE_REGISTRY

    ix_configuration_property_handle hCoreRoot = 0;
    ix_configuration_property_handle hProperty = 0;
    ix_cp_property_info              prop_info;

    IX_ERROR_CR(ix_cc_reg_get_prop_from_path(IX_MPLS_PROP_ILM_LABEL_SPACE_INFO, 
                                             &hCoreRoot));

    while(!ix_rm_cp_property_get_subproperty(hCoreRoot,
                                             idx,
                                            &hProperty)){
        
        char       port_str[5];
        ix_uint32  port_idx;
        ix_uint32  temp_value;
        
        idx++;
        
        if (hProperty == 0) break;

        IX_ERROR_CRT(ix_rm_cp_property_get_info(hProperty,
                                               &prop_info),
                 IX_CC_ERROR_ENTRY_NOT_FOUND,
                 IX_ERROR_LEVEL_REMOTE);

        /*  make sure it starts with PORT  */
        if(strncmp(prop_info.m_aName, "PORT", 4))  continue;

        port_idx = atoi(strncpy(port_str, prop_info.m_aName + 4, 5));

        /* get MPLS_MIN_LABEL */
        IX_ERROR_CRT(_ix_cc_mpls_prop_get_uint_value(hProperty,
                                                     "MPLS_MIN_LABEL",
                                                     &temp_value),
                 IX_CC_ERROR_ENTRY_NOT_FOUND,
                 IX_ERROR_LEVEL_REMOTE);
        
        pContext->pLabelSpaceTable[port_idx].min_label = temp_value;
        
        /* get MPLS_MAX_LABEL */
        IX_ERROR_CRT(_ix_cc_mpls_prop_get_uint_value(hProperty,
                                                     "MPLS_MAX_LABEL",
                                                     &temp_value),
                 IX_CC_ERROR_ENTRY_NOT_FOUND,
                 IX_ERROR_LEVEL_REMOTE);

        pContext->pLabelSpaceTable[port_idx].max_label = temp_value;

    }

    if (pContext->labelSpaceMode == IX_CC_MPLS_PER_PLATFORM_LABEL_SPACE)
    {
       idx = 0;
    } else {
       idx = pContext->maxLocalPorts - 1;
    }

    /* chack completness of the table */
    for (; idx >= 0; idx--)
    {

        if ((pContext->pLabelSpaceTable[idx].max_label == 0) ||
            (pContext->pLabelSpaceTable[idx].min_label == 0))
        {

                ix_ossl_message_log("_ix_cc_mpls_lsp_get_label_space_config(): Warning: "
                       "Registry doesn't have information on label space for port %d", idx);

                return IX_ERROR_WARNING(IX_CC_ERROR_UNINIT,
                                        ("_ix_cc_mpls_lsp_get_label_space_config(): Registry doesn't "
                                         "have information on label space for all ports"));
        }
    }

#endif

    /* calculate table offset */
    for (idx = 0, NextOffset = 0; idx < (pContext->maxLocalPorts); idx++) {

#ifndef IX_INCLUDE_REGISTRY
        pContext->pLabelSpaceTable[idx].min_label = MPLS_MIN_LABEL;
        pContext->pLabelSpaceTable[idx].max_label = MPLS_MAX_LABEL;
#endif /* IX_INCLUDE_REGISTRY */

        if (pContext->labelSpaceMode == IX_CC_MPLS_PER_INTERFACE_LABEL_SPACE) 
        {
            pContext->pLabelSpaceTable[idx].table_offset = NextOffset * sizeof(ix_cc_mpls_ilm_internal);
            pContext->pLabelSpaceTable[idx].stats_offset = NextOffset * sizeof(ix_cc_mpls_ilm_counters_internal);
            NextOffset += ((pContext->pLabelSpaceTable[idx].max_label -
                            pContext->pLabelSpaceTable[idx].min_label) + 1);
        } else {
            pContext->pLabelSpaceTable[idx].table_offset = 0;
            pContext->pLabelSpaceTable[idx].stats_offset = 0;        
            NextOffset = ((pContext->pLabelSpaceTable[0].max_label -
                           pContext->pLabelSpaceTable[0].min_label) + 1);
        }
    } 

    pContext->maxIlmEntries = NextOffset;

    return (IX_SUCCESS);
} /* _ix_cc_mpls_lsp_get_label_space_config */



/*
 * FUNCTION: _ix_cc_mpls_lsp_get_static_data()  */
/**
 * ABSTRACT: reads LSP-relevant data from MPLS configuration data external resource
 *
 * DESCRIPTION: This function retrieves all LSP-relevant static data from
 *              either system repository or ix_cc_mpls.h and store them
 *              into the corresponding field in the control block
 *              structure passed in.
 *
 * \param  arg_pContext  - pointer to component context memory
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 *
 * \n Error codes:
 * \n       IX_CC_ERROR_ENTRY_NOT_FOUND - could not find given property
 */
ix_error
_ix_cc_mpls_lsp_get_static_data(void *arg_pContext)
{

    ix_cc_mpls_context *pContext = (ix_cc_mpls_context*)arg_pContext;


#ifdef IX_INCLUDE_REGISTRY

    ix_configuration_property_handle hProperty;

    IX_ERROR_CR(ix_cc_reg_get_prop_from_path(IX_MPLS_PROP_ILM_ROOT, 
                                             &hProperty));

    /* get MAX_LSP_ENTRIES */
    IX_ERROR_CRT(_ix_cc_mpls_prop_get_uint_value(hProperty,
                                                 IX_MPLS_PROP_MAX_LSP_ENTRIES,
                                                 &pContext->maxLspEntries),
                 IX_CC_ERROR_ENTRY_NOT_FOUND,
                 IX_ERROR_LEVEL_REMOTE);


    /* get MAX_ILM_SET_ENTRIES */
    IX_ERROR_CRT(_ix_cc_mpls_prop_get_uint_value(hProperty,
                                                 IX_MPLS_PROP_MAX_ILM_SET_ENTRIES,
                                                 &pContext->maxIlmSetEntries),
                 IX_CC_ERROR_ENTRY_NOT_FOUND,
                 IX_ERROR_LEVEL_REMOTE);


    /* get ILM table memory type */
    IX_ERROR_CRT(_ix_cc_mpls_prop_get_uint_value(hProperty,
                                                 IX_MPLS_PROP_ILM_TABLE_MEM_TYPE,
                                                 (ix_uint32*)(&pContext->IlmTableMemoryType)),
                 IX_CC_ERROR_ENTRY_NOT_FOUND,
                 IX_ERROR_LEVEL_REMOTE);

    /* get ILM table memory channel */
    IX_ERROR_CRT(_ix_cc_mpls_prop_get_uint_value(hProperty,
                                                 IX_MPLS_PROP_ILM_TABLE_MEM_CHANNEL,
                                                 &pContext->IlmTableMemoryChannel),
                 IX_CC_ERROR_ENTRY_NOT_FOUND,
                 IX_ERROR_LEVEL_REMOTE);

    /* get ILM counters table memory channel */
    IX_ERROR_CRT(_ix_cc_mpls_prop_get_uint_value(hProperty,
                                                 IX_MPLS_PROP_ILM_LABEL_RANGE_TABLE_MEM_CHANNEL,
                                                 &pContext->LabelSpaceTableMemoryChannel),
                 IX_CC_ERROR_ENTRY_NOT_FOUND,
                 IX_ERROR_LEVEL_REMOTE);


#ifdef MPLS_COUNTERS_FEATURE

    /* get ILM counters table memory channel */
    IX_ERROR_CRT(_ix_cc_mpls_prop_get_uint_value(hProperty,
                                                 IX_MPLS_PROP_ILM_COUNTERS_TABLE_MEM_CHANNEL,
                                                 &pContext->IlmCountersMemoryChannel),
                 IX_CC_ERROR_ENTRY_NOT_FOUND,
                 IX_ERROR_LEVEL_REMOTE);

#endif /* MPLS_COUNTERS_FEATURE */


    
#else /* IX_INCLUDE_REGISTRY */

    pContext->IlmCountersMemoryChannel     = MPLS_ILM_COUNTERS_TABLE_MEM_CHANNEL;
    pContext->LabelSpaceTableMemoryChannel = MPLS_ILM_LABEL_SPACE_TABLE_MEM_TYPE;
    pContext->IlmTableMemoryType           = MPLS_ILM_TABLE_MEM_TYPE;
    pContext->IlmTableMemoryChannel        = MPLS_ILM_TABLE_MEM_CHANNEL;
    pContext->maxIlmSetEntries             = MPLS_ILM_MAX_SET_ENTRIES;

#endif /* IX_INCLUDE_REGISTRY */

    return (IX_SUCCESS);
} /* _ix_cc_mpls_lsp_get_static_data */




/*
 * FUNCTION: _ix_cc_mpls_lsp_patch_symbols()  */
/**
 * ABSTRACT: These function patches all LSP related symbols in the microcode
 *
 * DESCRIPTION: Patching of various symbols and memory bases into MPLS microblock
 *
 * \param  arg_pContext - a pointer to the MPLS context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n       IX_CC_MPLS_ERROR_RM_GET_PHYS_OFFSET - error by physical offset
 *                                                reading
 * \n       IX_CC_MPLS_ERROR_RM_PATCH_SYMBOL    - error by patching imported
 *                                                symbols into the microcode
 */
ix_error 
_ix_cc_mpls_lsp_patch_symbols(void *arg_pContext)
{

    ix_cc_mpls_context *pContext = (ix_cc_mpls_context*)arg_pContext;
    ix_imported_symbol  _Symbols[IX_MPLS_LSP_SYMBOLS_PATCHED_NUM];
    
    ix_uint32 symbolIndex = 0;
    ix_uint32 mev2Id;
    ix_uint32 memOffset;

    /* ==================================================================  */
    /* Obtain the physical base offset of the ilm table                    */
    IX_ERROR_CRT( ix_rm_get_phys_offset(pContext->pIlmBaseAddr, NULL, NULL,
                                        &memOffset, NULL ),
                  IX_CC_MPLS_ERROR_RM_GET_PHYS_OFFSET,
                  IX_ERROR_LEVEL_REMOTE);
    
    /* Pack the SRAM channel number into the offset */
    if(pContext->IlmTableMemoryType == IX_MEMORY_TYPE_SRAM )
    {
        IX_CC_SET_SRAM_CHANNEL_INTO_OFFSET(pContext->IlmTableMemoryChannel, memOffset);
    }
    
    /* Prepare the symbol array */
    MPLS_SET_UBLOCK_SYMBOLS(_Symbols[symbolIndex], memOffset, IX_MPLS_SYMBOL_ILM_BASE);
    symbolIndex++;

    /* ==================================================================  */
    /* Obtain the physical base offset of the ilm set table */
    IX_ERROR_CRT( ix_rm_get_phys_offset(pContext->pIlmSetBaseAddr, NULL, NULL,
                                        &memOffset, NULL ),
                  IX_CC_MPLS_ERROR_RM_GET_PHYS_OFFSET,
                  IX_ERROR_LEVEL_REMOTE);
    
    /* Pack the SRAM channel number into the offset */
    if(pContext->IlmTableMemoryType == IX_MEMORY_TYPE_SRAM )
    {
        IX_CC_SET_SRAM_CHANNEL_INTO_OFFSET(pContext->IlmTableMemoryChannel, memOffset);
    }
    
    /* Prepare the symbol array */
    MPLS_SET_UBLOCK_SYMBOLS(_Symbols[symbolIndex], memOffset, IX_MPLS_SYMBOL_ILM_SET_BASE);
    symbolIndex++;


    /* ==================================================================  */
    /* Obtain the physical base offset of the nhlfe table */
    IX_ERROR_CRT( ix_rm_get_phys_offset(pContext->pLabelSpaceTable, NULL, NULL,
                                        &memOffset, NULL ),
                  IX_CC_MPLS_ERROR_RM_GET_PHYS_OFFSET,
                  IX_ERROR_LEVEL_REMOTE);
    
    IX_CC_SET_SRAM_CHANNEL_INTO_OFFSET(pContext->LabelSpaceTableMemoryChannel, memOffset);
    
    /* Prepare the symbol array */
    MPLS_SET_UBLOCK_SYMBOLS(_Symbols[symbolIndex], memOffset, IX_MPLS_SYMBOL_ILM_PARAM_BASE);
    symbolIndex++;


#ifdef MPLS_COUNTERS_FEATURE

    /* ==================================================================  */
    /* Obtain the physical base offset of the ILM counters table           */
    IX_ERROR_CRT( ix_rm_get_phys_offset(pContext->pIlmCounters, NULL, NULL,
                                        &memOffset, NULL ),
                  IX_CC_MPLS_ERROR_RM_GET_PHYS_OFFSET,
                  IX_ERROR_LEVEL_REMOTE);
    
    /* Pack the SRAM channel number into the offset */
    IX_CC_SET_SRAM_CHANNEL_INTO_OFFSET(pContext->IlmCountersMemoryChannel, memOffset);
    
    /* Prepare the symbol array */
    MPLS_SET_UBLOCK_SYMBOLS(_Symbols[symbolIndex], memOffset, IX_MPLS_SYMBOL_ILM_STAT_BASE);
    symbolIndex++;

    /* ==================================================================  */
    /* Obtain the physical base offset of the ILM NHLFE set counters table */
    IX_ERROR_CRT( ix_rm_get_phys_offset(pContext->pIlmNhlfeSetCounters, NULL, NULL,
                                        &memOffset, NULL ),
                  IX_CC_MPLS_ERROR_RM_GET_PHYS_OFFSET,
                  IX_ERROR_LEVEL_REMOTE);
    
    /* Pack the SRAM channel number into the offset */
    IX_CC_SET_SRAM_CHANNEL_INTO_OFFSET(pContext->IlmCountersMemoryChannel, memOffset);
    
    /* Prepare the symbol array */
    MPLS_SET_UBLOCK_SYMBOLS(_Symbols[symbolIndex], memOffset, IX_MPLS_SYMBOL_ILM_NHLFE_SET_STAT_BASE);
    symbolIndex++;

    /* ==================================================================  */
    /* Obtain the physical base offset of the input port counters table    */
    IX_ERROR_CRT( ix_rm_get_phys_offset(pContext->pInputPortCounters, NULL, NULL,
                                        &memOffset, NULL ),
                  IX_CC_MPLS_ERROR_RM_GET_PHYS_OFFSET,
                  IX_ERROR_LEVEL_REMOTE);
    
    /* Pack the SRAM channel number into the offset */
    IX_CC_SET_SRAM_CHANNEL_INTO_OFFSET(pContext->IlmCountersMemoryChannel, memOffset);
    
    /* Prepare the symbol array */
    MPLS_SET_UBLOCK_SYMBOLS(_Symbols[symbolIndex], memOffset, IX_MPLS_INPUT_PORT_STAT_BASE);
    symbolIndex++;
    
#endif /* MPLS_COUNTERS_FEATURE */

        
    /* Patch them into each microengine used by the microblock */
    for(mev2Id = 0; mev2Id < IX_MAX_MICROENGINE_NUMBER; mev2Id++)
    {
        if (pContext->mev2NumMask & (0x1 << mev2Id))
        {
            IX_ERROR_CRT(ix_rm_ueng_patch_symbols(g_aMicroengineNumberMapping[mev2Id],
                                                  IX_MPLS_LSP_SYMBOLS_PATCHED_NUM,
                                                  _Symbols ),
                         IX_CC_MPLS_ERROR_RM_PATCH_SYMBOL,
                         IX_ERROR_LEVEL_REMOTE);
        }
    }

    return (IX_SUCCESS);
} /* _ix_cc_mpls_lsp_patch_symbols */





/*
 * FUNCTION: _ix_cc_mpls_lsp_verify_input()  */
/**
 * ABSTRACT: Verifies Lsp data.
 *
 * DESCRIPTION: Verifies Lsp data.
 *
 * \param  arg_Lsp       - an ix_cc_mpls_lsp structure
 *                         describing an Lsp entry.
 * \param  arg_pContext  - pointer to component context memory
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n       IX_CC_ERROR_NULL               - invalid pointer
 * \n       IX_CC_ERROR_RANGE              - paramters are out of permitted range
 * \n       IX_CC_MPLS_ERROR_INVALID_ENTRY - invalid LSP parameters
 */
PRIVATE ix_error
_ix_cc_mpls_lsp_verify_input(ix_cc_mpls_lsp     *arg_pLsp,
                             ix_cc_mpls_context *arg_pContext)
{
    ix_uint8 i;

    /* check number of NHLFE entries */
    if ((arg_pLsp->nhlfe_num < 1) || 
        (arg_pLsp->nhlfe_num > MPLS_MAX_NHLFE_IN_SET))
        return IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
                                ("_ix_cc_mpls_lsp_verify_input(): nhlfeSetSize out of range"));

    /* check NHLFE data type */
    if ((arg_pLsp->nhlfeInfoType != IX_CC_MPLS_LSP_NHLFE_SET_HANDLE_TYPE) &&
        (arg_pLsp->nhlfeInfoType != IX_CC_MPLS_LSP_NHLFE_HANDLE_TYPE) &&
        (arg_pLsp->nhlfeInfoType != IX_CC_MPLS_LSP_NHLFE_DATA_TYPE))
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_ENTRY,
                                ("_ix_cc_mpls_lsp_verify_input(): unknown nhlfeInfoType"));
    /* check NHLFE data presence */
    if ((arg_pLsp->nhlfeInfoType == IX_CC_MPLS_LSP_NHLFE_DATA_TYPE) &&
        (arg_pLsp->outSeg.nhlfeData == NULL))
        return IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
                                ("_ix_cc_mpls_lsp_verify_input(): NULL nhlfe data table"));

    switch (arg_pLsp->lspType)
    {
    case IX_CC_MPLS_LSP_ILM_TYPE:

        /* check operation type */
        switch (arg_pLsp->inSeg.ilm.operation)
        {
        case IX_MPLS_LABELOP_SWAP:
        case IX_MPLS_LABELOP_SWAP_PUSH:
        case IX_MPLS_LABELOP_POP:
        case IX_MPLS_LABELOP_POP_FORWARD:
            break;

        case IX_MPLS_LABELOP_PUSH:
            return IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
                                    ("ix_cc_mpls_lsp_create(): PUSH operation set for ILM LSP."));
        default:
            return IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
                                    ("ix_cc_mpls_lsp_create(): unknown operation."));
        }
    
        /* check push count for PUSH operation */
        if ((arg_pLsp->inSeg.ilm.operation == IX_MPLS_LABELOP_SWAP_PUSH)  &&
            (arg_pLsp->nhlfeInfoType == IX_CC_MPLS_LSP_NHLFE_DATA_TYPE))
        {
            for (i=0; i<arg_pLsp->nhlfe_num; i++)
            {
                if ((arg_pLsp->outSeg.nhlfeData[i]->pushCount < 1)  &&
                    (arg_pLsp->outSeg.nhlfeData[i]->pushCount > 3))
                    return IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
                                            ("_ix_cc_mpls_lsp_verify_input(): push_count out of range"));
            }
        }
    
        /* check port number */
        if  (arg_pLsp->inSeg.ilm.in_port > arg_pContext->maxLocalPorts)
            return IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
                                    ("_ix_cc_mpls_lsp_verify_input(): ilm port value out of range"));

        /* check label */
        if  ((arg_pLsp->inSeg.ilm.label.label < arg_pContext->pLabelSpaceTable[arg_pLsp->inSeg.ilm.in_port].min_label) ||
              (arg_pLsp->inSeg.ilm.label.label > arg_pContext->pLabelSpaceTable[arg_pLsp->inSeg.ilm.in_port].max_label))
            return IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
                                    ("_ix_cc_mpls_lsp_verify_input(): label value out of range"));

        /* check TTL in ILM */
        if  ((arg_pLsp->inSeg.ilm.operation == IX_MPLS_LABELOP_POP) &&
             (arg_pLsp->lspParams[0].ttl != 1))
            return IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
                                    ("_ix_cc_mpls_lsp_verify_input(): improper ilm ttl value for pop operation"));


#if defined(MPLS_COUNTERS_FEATURE)
        if (arg_pContext->pIlmCounters == NULL)
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                    ("ix_cc_mpls_lsp_create(): ILM counters table has not been created."));
#endif
        break;

        
    case IX_CC_MPLS_LSP_FEC_TYPE:

        /* check push count for PUSH operation */
        if (arg_pLsp->nhlfeInfoType == IX_CC_MPLS_LSP_NHLFE_DATA_TYPE)
        {
            for (i=0; i<arg_pLsp->nhlfe_num; i++)
            {
                if ((arg_pLsp->outSeg.nhlfeData[i]->pushCount < 1)  &&
                    (arg_pLsp->outSeg.nhlfeData[i]->pushCount > 4))
                    return IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
                                            ("_ix_cc_mpls_lsp_verify_input(): push_count out of range"));
            }
        }

#if defined(MPLS_COUNTERS_FEATURE)
        if (arg_pContext->pFtnCounters == NULL)
            return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                    ("ix_cc_mpls_lsp_create(): FTN counters table has not been created."));
#endif
        break;

        
    default:
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_ENTRY,
                                ("_ix_cc_mpls_lsp_verify_input(): unknown lspType"));
    }

    return IX_SUCCESS;
} /* _ix_cc_mpls_lsp_verify_input */





/*
 * FUNCTION: _ix_cc_mpls_lsp_copy_nhlfe_to_ilm()  */
/**
 * ABSTRACT: copy the content of nhlfe to ilm entry
 *
 * \param  arg_LspHandle - a handle identifying an CC LSP entry.
 * \param  arg_pNhlfe    - a pointer to the NHLFE Table context
 */
PRIVATE void
_ix_cc_mpls_lsp_copy_nhlfe_to_ilm(ix_cc_mpls_ilm_internal  *arg_pIlmEntry,
                                  ix_cc_mpls_lsp_param      arg_lspParams,
                                  ix_cc_mpls_nhlfe         *arg_pNhlfe)
{
    arg_pIlmEntry->bladeId       = arg_pNhlfe->bladeId;
    arg_pIlmEntry->exp2phb       = arg_lspParams.exp2phbIdx;
    arg_pIlmEntry->l3Protocol    = arg_pNhlfe->l3Protocol;
    arg_pIlmEntry->label1        = arg_pNhlfe->label1.label;
    arg_pIlmEntry->label2        = arg_pNhlfe->label2.label;
    arg_pIlmEntry->label3        = arg_pNhlfe->label3.label;
    arg_pIlmEntry->maxLabPktSize = arg_lspParams.maxLabPktSize;
    arg_pIlmEntry->nextHopId     = arg_pNhlfe->nextHopId;
    arg_pIlmEntry->outPortId     = arg_pNhlfe->outPortId;
    arg_pIlmEntry->outPortType   = arg_pNhlfe->outPortType;
    arg_pIlmEntry->phb2exp       = arg_lspParams.phb2expIdx;
    arg_pIlmEntry->pushCount     = arg_pNhlfe->pushCount;
    arg_pIlmEntry->topLabel      = arg_pNhlfe->topLabel.label;
    arg_pIlmEntry->ttl           = arg_lspParams.ttl;
    arg_pIlmEntry->flags         = MPLS_FLAG_VALID_BIT | arg_lspParams.flags;
} /* _ix_cc_mpls_lsp_copy_nhlfe_to_ilm */





/*
 * FUNCTION: _ix_cc_mpls_lsp_copy_nhlfe_set_to_ilm()  */
/**
 * ABSTRACT: copy the content of nhlfes from a set to an ilm entry
 *
 * \param  arg_pIlmEntry    - pointer to the ilm entry
 * \param  arg_NhlfeSetSize - set size
 * \param  arg_NhlfeHandles - the nhlfe set
 *
 * \return IX_SUCCESS
 *
 * NHLFE error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE does not exist
 * \n         IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET - this is an NHLFE set entry
 */
PRIVATE ix_error
_ix_cc_mpls_lsp_copy_nhlfe_set_to_ilm(ix_cc_mpls_ilm_set_internal  *arg_pIlmSetEntry,
                                      ix_uint16                     arg_IlmSetSize,
                                      ix_cc_mpls_lsp_param          arg_lspParams[MPLS_MAX_NHLFE_IN_SET],
                                      ix_cc_mpls_ilm_internal      *arg_NewIlmSetEntry[MPLS_MAX_NHLFE_IN_SET],
                                      ix_uint32                     arg_NewIlmSetIndices[MPLS_MAX_NHLFE_IN_SET],
                                      ix_cc_mpls_nhlfe_handle       arg_NhlfeHandles[MPLS_MAX_NHLFE_IN_SET],
                                      ix_cc_mpls_context           *arg_pContext)
{
    ix_uint32        idx;
    ix_cc_mpls_nhlfe NhlfeSet[MPLS_MAX_NHLFE_IN_SET];

    for (idx = 0; idx < arg_IlmSetSize; idx++) {
        ix_error err;
        err = ix_cc_mpls_nhlfe_query(arg_NhlfeHandles[idx],
                                     &NhlfeSet[idx],
                                     arg_pContext);
        if (err != IX_SUCCESS)
            return err;
    }

    for (idx = 0; idx < arg_IlmSetSize; idx++)
        _ix_cc_mpls_lsp_copy_nhlfe_to_ilm(arg_NewIlmSetEntry[idx],
                                          arg_lspParams[idx],
                                         &NhlfeSet[idx]);

    arg_pIlmSetEntry -> flags       =  MPLS_FLAG_VALID_BIT | MPLS_FLAG_SET_BIT;
    arg_pIlmSetEntry -> SetSize     =  arg_IlmSetSize;
    arg_pIlmSetEntry -> OutSegId1   =  arg_NewIlmSetIndices[0];
    arg_pIlmSetEntry -> OutSegId2Hi = (arg_NewIlmSetIndices[1] & 0xFFF00) >> 8;
    arg_pIlmSetEntry -> OutSegId2Lo =  arg_NewIlmSetIndices[1] & 0x000FF;
    arg_pIlmSetEntry -> OutSegId3   =  arg_NewIlmSetIndices[2];
    arg_pIlmSetEntry -> OutSegId4Hi = (arg_NewIlmSetIndices[3] & 0xF0000) >> 16;
    arg_pIlmSetEntry -> OutSegId4Lo =  arg_NewIlmSetIndices[3] & 0x0FFFF;

    return IX_SUCCESS;
} /* _ix_cc_mpls_lsp_copy_nhlfe_set_to_ilm */



/*
 * FUNCTION: _ix_cc_mpls_lsp_add_to_free_list()  */
/**
 * ABSTRACT: Adds new element to the list of free CC LSP entries
 *
 * DESCRIPTION: Adds new element to the list of free CC LSP entries
 *
 * \param  arg_LspHandle - a handle identifying an CC LSP entry.
 * \param  arg_pContext  - a pointer to the MPLS CC context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n       IX_CC_ERROR_OOM_SYSTEM - memory allocation failure
 */
PRIVATE ix_error
_ix_cc_mpls_lsp_add_to_free_list(ix_cc_mpls_lsp_handle  arg_LspHandle,
                                 ix_cc_mpls_context    *arg_pContext)
{
    ix_cc_mpls_cc_lsp_free_slot  *pFreeLspListNewHead;

    pFreeLspListNewHead = (ix_cc_mpls_cc_lsp_free_slot*)
                          ix_ossl_malloc(sizeof(ix_cc_mpls_cc_lsp_free_slot));

    if (pFreeLspListNewHead == NULL)
        return (IX_ERROR_REMOTE(IX_CC_ERROR_OOM_SYSTEM, 
               ("_ix_cc_mpls_lsp_add_to_free_list(): "
                "No memory to create the list of free LSP elements")));

    pFreeLspListNewHead->LspIndex      = arg_LspHandle.LspIndex;
    pFreeLspListNewHead->pNextFreeSlot = arg_pContext->pFreeLspListHead;
    arg_pContext->pFreeLspListHead     = pFreeLspListNewHead;

    return IX_SUCCESS;
} /* _ix_cc_mpls_lsp_add_to_free_list */








/*
 * FUNCTION: _ix_cc_mpls_ilm_set_add_to_free_list()  */
/**
 * ABSTRACT: Adds new element to the list of free ILM SET entries
 *
 * DESCRIPTION: Adds new element to the list of free ILM SET entries
 *
 * \param  arg_IlmSetIdx - a handle identifying an element in ILM SET table
 * \param  arg_pContext  - a pointer to the MPLS CC context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n       IX_CC_ERROR_OOM_SYSTEM - memory allocation failure
 */
PRIVATE ix_error
_ix_cc_mpls_ilm_set_add_to_free_list(ix_uint32           arg_IlmSetIdx,
                                     ix_cc_mpls_context *arg_pContext)
{
    ix_cc_mpls_ilm_set_free_slot  *pFreeIlmSetListNewHead;

    pFreeIlmSetListNewHead = (ix_cc_mpls_ilm_set_free_slot*)
                                ix_ossl_malloc(sizeof(ix_cc_mpls_ilm_set_free_slot));

    if (pFreeIlmSetListNewHead == NULL)
        return (IX_ERROR_REMOTE(IX_CC_ERROR_OOM_SYSTEM, 
               ("_ix_cc_mpls_ilm_set_add_to_free_list(): "
                "No memory to create the list of free ILM SET elements")));

    pFreeIlmSetListNewHead->IlmSetIndex   = arg_IlmSetIdx;
    pFreeIlmSetListNewHead->pNextFreeSlot = arg_pContext->pFreeIlmSetListHead;
    arg_pContext->pFreeIlmSetListHead     = pFreeIlmSetListNewHead;

    return IX_SUCCESS;
} /* _ix_cc_mpls_ilm_set_add_to_free_list */



#undef ON_ERROR_GOTO

