/*
 * ============================================================================
 * COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *                                                                          */
/** \file      ix_cc_mpls_msg_handler.c
 *
 *  \brief     The file contains CCI API function (msg handler) exposed by
 *             MPLS Core Component
 *
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_mpls_msg_handler.c,v 1.11.2.1 2004/03/12 10:40:15 wneubaue Exp $";

#include <ix_types.h>
#include <ix_error.h>

#include <ix_ossl.h>
#include <ix_rm.h>

#if !defined(IX_EXCLUDE_CCI)
#include <ix_cci.h>
#endif 

#include <ix_cc_error.h>
#include <ix_cc.h>

#include <bindings.h>

#include <cc/ix_cc_msup.h>

#include <cc/ix_cc_rtmv4.h>

/**
 * User defined include files required
 */
#include "../../include/mpls_defs.h"
#include "cc/ix_cc_mpls.h"
#include "cc/internal/ix_cc_mpls_msg.h"
#include "cc/internal/ix_cc_mpls_lsp.h"
#include "cc/internal/ix_cc_mpls_nhlfe.h"
#include "cc/internal/ix_cc_mpls_internal.h"


/*
 * FUNCTION: ix_cc_mpls_msg_handler()  */
/**
 * ABSTRACT: This is the message handler function for the MPLS core component.
 *
 * DESCRIPTION: This is the message handler function for the MPLS core component.
 *              The MPLS CC receives messages from other core component through
 *              this message handler function, and internally it calls appropriate
 *              library functions to process the message.
 *
 * \param  arg_hDataToken - buffer handle embedding information for 
 *                          the message passed in arg_UserData.
 * \param  arg_UserData   - message type.
 * \param  arg_pContext   - pointer to the MPLS Forwarder core component context
 *                          that will be passed to the core component when a message 
 *                          arrives. This context was defined by the core component 
 *                          and passed to core components infrastructure through 
 *                          the ix_cc_mpls_init function.
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_INVALID_MESSAGE     - unsupported message
 * \n      IX_CC_MPLS_ERROR_INVALID_HANDLE - buffer handle is unusable
 * \n      IX_CC_ERROR_INVALID_CONTEXT     - context is unusable
 * \n      IX_CC_ERROR_NULL                - invalid pointer
 */
ix_error
ix_cc_mpls_msg_handler(ix_buffer_handle arg_hDataToken,
                       ix_uint32        arg_UserData,
                       void             *arg_pContext)
{
    ix_error   err = IX_SUCCESS;
    void      *pUserContext = NULL;
    void      *pMsg         = NULL;
    void      *pRetMsg      = NULL;
    ix_uint32  RetMsgLength = 0;

    ix_cc_mpls_lsp                         Lsp;
    ix_cc_mpls_lsp_stats                   LspStatistics;
    ix_cc_mpls_nhlfe                       Nhlfe;
    ix_cc_mpls_nhlfe_handle                NhlfeHandle;
    ix_cc_mpls_nhlfe_stats                 NhlfeStatistics;
    ix_cc_mpls_params                      MplsParameters;
    ix_cc_mpls_input_port_stats            MplsPortStatistics;
    ix_cc_mpls_phb2exp                     Phb2ExpTable[MPLS_MAX_PHB_NUM];
    ix_cc_mpls_exp2phb                     Exp2PhbTable[MPLS_MAX_EXP_NUM];
    ix_cc_mpls_message_lsp_create_rep      LspCreateRep;
    ix_cc_mpls_message_nhlfe_set_query_rep NhlfeSetQueryRep;


    #if defined (IX_MPLS_MSG_DEBUG)
    ix_ossl_message_log("ix_cc_mpls_msg_handler:  I have received a piece of message...\n");
    #endif

    /* Check the input arguments */
    if (NULL == arg_pContext) {
        ix_rm_buffer_free_chain(arg_hDataToken);
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("ix_cc_mpls_msg_handler(): Error - got NULL context"));

    }

    /* Check the input arguments */
    if (arg_hDataToken == IX_NULL_BUFFER_HANDLE)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE,
                                ("ix_cc_mpls_msg_handler(): Invalid buffer handle."));


    /* Get the message sent by the helper function */
    IX_ERROR_CT(ix_cc_msup_extract_msg(arg_hDataToken, &pMsg, &pUserContext),
                err,
                IX_CC_MPLS_ERROR_MSUP_EXTRACT_MSG,
                IX_ERROR_LEVEL_REMOTE);

    if (err != IX_SUCCESS) 
    {
#if defined (IX_MPLS_DEBUG)
        ix_ossl_message_log("ix_cc_mpls_msg_handler(): Failed to extract message\n");
#endif

        /* Free the ix_buffer */
        ix_rm_buffer_free_chain(arg_hDataToken);

        return err;
    }

    
    /* Validate message buffer */
    if(pMsg == NULL)
    {
#if defined (IX_MPLS_DEBUG)
        ix_ossl_message_log("ix_cc_mpls_msg_handler(): Invalid buffer\n");
#endif

        /* Free the ix_buffer */
        ix_rm_buffer_free_chain(arg_hDataToken);

        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_msg_handler(): Invalid Buffer"));
    }


    /* Process the message */
    switch(arg_UserData)
    {

    case IX_CC_MPLS_MSG_LSP_CREATE:
    {
        ix_uint8 idx;
        ix_cc_mpls_message_lsp_create *pMsgBody = 
                           (ix_cc_mpls_message_lsp_create*)pMsg;

        if (pMsgBody->Lsp.nhlfeInfoType == IX_CC_MPLS_LSP_NHLFE_DATA_TYPE)
           for(idx = 0; idx < pMsgBody->Lsp.nhlfe_num; idx ++)
               pMsgBody->Lsp.outSeg.nhlfeData[idx] = &(pMsgBody->NhlfeSet[idx]);

        err = ix_cc_mpls_lsp_create(&pMsgBody->Lsp,
                                    &LspCreateRep.LspHandle,
                                     arg_pContext);

        LspCreateRep.nhlfe_num       = pMsgBody->Lsp.nhlfe_num;
        LspCreateRep.nhlfeSetHandle  = pMsgBody->Lsp.nhlfeSetHandle;
        
        ix_ossl_memcpy(LspCreateRep.nhlfeHandle,
                       pMsgBody->Lsp.outSeg.nhlfeHandle,
                       sizeof(ix_cc_mpls_nhlfe_handle) * MPLS_MAX_NHLFE_IN_SET);

        pRetMsg = (void*)(&LspCreateRep);
        RetMsgLength = sizeof(LspCreateRep);

        break;
    }
    case IX_CC_MPLS_MSG_LSP_DELETE:
    {
        ix_cc_mpls_lsp_handle *pLspHandle = (ix_cc_mpls_lsp_handle*)pMsg;

        err = ix_cc_mpls_lsp_delete(*pLspHandle,
                                     arg_pContext);

        break;
    }
    case IX_CC_MPLS_MSG_LSP_MODIFY:
    {
        ix_cc_mpls_message_lsp_modify *pMsgBody =
                     (ix_cc_mpls_message_lsp_modify*)pMsg;

        err = ix_cc_mpls_lsp_modify(pMsgBody->LspHandle,
                                    pMsgBody->nhlfeInfoType,
                                  &(pMsgBody->NhlfeSetHandle),
                                    pMsgBody->nhlfeHandles_num,
                                    pMsgBody->nhlfeHandles,
                                    pMsgBody->lspParams,
                                    arg_pContext);
        
        pRetMsg      = (void*)&(pMsgBody->NhlfeSetHandle);
        RetMsgLength = sizeof(ix_cc_mpls_nhlfe_handle);

        break;        
    }
    case IX_CC_MPLS_MSG_LSP_QUERY:
    {
        ix_cc_mpls_lsp_handle *pLspHandle = (ix_cc_mpls_lsp_handle*)pMsg;

        err = ix_cc_mpls_lsp_query(*pLspHandle,
                                   &Lsp,
                                   arg_pContext);

        pRetMsg = (void*)(&Lsp);
        RetMsgLength = sizeof(Lsp);

        break;
    }
    case IX_CC_MPLS_MSG_LSP_STATS_QUERY:
    {
        ix_cc_mpls_lsp_handle *pLspHandle = (ix_cc_mpls_lsp_handle*)pMsg;

        err = ix_cc_mpls_lsp_stats_query(*pLspHandle,
                                         &LspStatistics,
                                         arg_pContext);

        pRetMsg = (void*)(&LspStatistics);
        RetMsgLength = sizeof(LspStatistics);

        break;
    }
    case IX_CC_MPLS_MSG_LSP_PURGE:
    {
        err = ix_cc_mpls_lsp_purge(arg_pContext);
        break;
    }
    case IX_CC_MPLS_MSG_NHLFE_CREATE:
    {
        ix_cc_mpls_nhlfe MsgBody = *((ix_cc_mpls_nhlfe*)pMsg);

        err = ix_cc_mpls_nhlfe_create( MsgBody,
                                      &NhlfeHandle,
                                      arg_pContext);

        pRetMsg = (void*)(&NhlfeHandle);
        RetMsgLength = sizeof(NhlfeHandle);

        break;
    }
    case IX_CC_MPLS_MSG_NHLFE_DELETE:
    {
        ix_cc_mpls_nhlfe_handle *pNhlfeHandle =
			            (ix_cc_mpls_nhlfe_handle*)pMsg;

        err = ix_cc_mpls_nhlfe_delete(*pNhlfeHandle,
                                       arg_pContext);

        break;
    }
    case IX_CC_MPLS_MSG_NHLFE_QUERY:
    {
        ix_cc_mpls_nhlfe_handle *pNhlfeHandle =
			            (ix_cc_mpls_nhlfe_handle*)pMsg;

        err = ix_cc_mpls_nhlfe_query(*pNhlfeHandle,
                                     &Nhlfe,
                                     arg_pContext);

        pRetMsg = (void*)(&Nhlfe);
        RetMsgLength = sizeof(Nhlfe);

        break;
    }
    case IX_CC_MPLS_MSG_NHLFE_STATS_QUERY:
    {
        ix_cc_mpls_nhlfe_handle *pNhlfeHandle = 
			            (ix_cc_mpls_nhlfe_handle*)pMsg;

        err = ix_cc_mpls_nhlfe_stats_query(*pNhlfeHandle,
                                           &NhlfeStatistics,
                                           arg_pContext);

        pRetMsg = (void*)(&NhlfeStatistics);
        RetMsgLength = sizeof(NhlfeStatistics);

        break;
    }
    case IX_CC_MPLS_MSG_NHLFE_PURGE:
    {

        err = ix_cc_mpls_nhlfe_purge(arg_pContext);

        break;
    }
    case IX_CC_MPLS_MSG_NHLFE_SET_CREATE:
    {
        ix_cc_mpls_message_nhlfe_set_create *pMsgBody =
               (ix_cc_mpls_message_nhlfe_set_create*) pMsg;

        err = ix_cc_mpls_nhlfe_set_create(pMsgBody->NhlfeSetSize,
                                          pMsgBody->NhlfeSetHandleTable,
                                          &NhlfeHandle,
                                          arg_pContext);

        pRetMsg = (void*)(&NhlfeHandle);
        RetMsgLength = sizeof(NhlfeHandle);

        break;
    }
    case IX_CC_MPLS_MSG_NHLFE_SET_DELETE:
    {
        ix_cc_mpls_nhlfe_handle *pNhlfeHandle = 
			            (ix_cc_mpls_nhlfe_handle*)pMsg;

        err = ix_cc_mpls_nhlfe_set_delete(*pNhlfeHandle,
                                          arg_pContext);
        break;
    }
    case IX_CC_MPLS_MSG_NHLFE_SET_MODIFY:
    {
        ix_cc_mpls_message_nhlfe_set_modify *pMsgBody =
               (ix_cc_mpls_message_nhlfe_set_modify*) pMsg;

        err = ix_cc_mpls_nhlfe_set_modify(pMsgBody->NhlfeHandle,
                                          pMsgBody->NhlfeSetSize,
                                          pMsgBody->NhlfeSetHandleTable,
                                          arg_pContext);
        break;
    }
    case IX_CC_MPLS_MSG_NHLFE_SET_QUERY:
    {
        ix_cc_mpls_nhlfe_handle *pNhlfeHandle = 
			            (ix_cc_mpls_nhlfe_handle*)pMsg;

        err = ix_cc_mpls_nhlfe_set_query(*pNhlfeHandle,
                                         &(NhlfeSetQueryRep.NhlfeSetSize),
                                         NhlfeSetQueryRep.NhlfeSetHandleTable,
                                         arg_pContext);


        pRetMsg = (void*)(&NhlfeSetQueryRep);
        RetMsgLength = sizeof(NhlfeSetQueryRep);
        break;
    }
    case IX_CC_MPLS_MSG_EXP2PHB_MODIFY:
    {
        ix_cc_mpls_message_exp2phb_modify *pMsgBody =
            (ix_cc_mpls_message_exp2phb_modify*) pMsg;

        err = ix_cc_mpls_exp2phb_modify(pMsgBody->Index,
                                        pMsgBody->Exp2PhbTable,
                                        arg_pContext);
        break;
    }
    case IX_CC_MPLS_MSG_EXP2PHB_QUERY:
    {
        ix_uint16 Index = *((ix_uint16*)pMsg);

        err = ix_cc_mpls_exp2phb_query(Index,
                                       Exp2PhbTable,
                                       arg_pContext);

        pRetMsg = (void*)(&Exp2PhbTable);
        RetMsgLength = sizeof(Exp2PhbTable);
        break;
    }
    case IX_CC_MPLS_MSG_PHB2EXP_MODIFY:
    {

        ix_cc_mpls_message_phb2exp_modify *pMsgBody =
            (ix_cc_mpls_message_phb2exp_modify*) pMsg;
        
        err = ix_cc_mpls_phb2exp_modify(pMsgBody->Index,
                                        pMsgBody->Phb2ExpTable,
                                        arg_pContext);
        break;
    }
    case IX_CC_MPLS_MSG_PHB2EXP_QUERY:
    {
        ix_uint16* Index = (ix_uint16*)pMsg;

        err = ix_cc_mpls_phb2exp_query(*Index,
                                       Phb2ExpTable,
                                       arg_pContext);
        pRetMsg = (void*)(&Phb2ExpTable);
        RetMsgLength = sizeof(Phb2ExpTable);

        break;
    }
    case IX_CC_MPLS_MSG_PARAM_QUERY:
    {

        err = ix_cc_mpls_param_query(&MplsParameters,
                                      arg_pContext);
        pRetMsg = &MplsParameters;
        RetMsgLength = sizeof(MplsParameters);

        break;
    }
    case IX_CC_MPLS_MSG_PORT_STATS_QUERY:
    {
        ix_uint16* PortIndex = (ix_uint16*)pMsg;

        err = ix_cc_mpls_port_stats_query(*PortIndex,
                                          &MplsPortStatistics,
                                          arg_pContext);
        pRetMsg = &MplsPortStatistics;
        RetMsgLength = sizeof(ix_cc_mpls_input_port_stats);

        break;
    }
    case IX_CC_MPLS_MSG_GET_MPLS_CONTEXT:
    {
        pRetMsg = &arg_pContext;
        RetMsgLength = sizeof(ix_cc_mpls_context*);
        break;
    }
    default:
        err = IX_ERROR_LOCAL(IX_CC_ERROR_UNDEFINED_MSG,
                             ("ix_cc_mpls_msg_handler(): Undefined message type."));
        break;
    }


    /* send return message, length and retrieved context to message support library */
    err = ix_cc_msup_send_reply_msg(pUserContext, pRetMsg, RetMsgLength, err);
    if(err != IX_SUCCESS)
    {
        err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_MSUP_SEND_REPLY,
                               ("ix_cc_mpls_msg_handler(): Send reply message failed"));
    }

    ix_ossl_free(pMsg);

    /* free buffer */
    IX_ERROR_CRT(ix_rm_buffer_free_chain(arg_hDataToken),
                 IX_CC_MPLS_ERROR_RM_FREE_BUF, IX_ERROR_LEVEL_REMOTE);

    return err;
} /* ix_cc_mpls_msg_handler */
