/*
 * ============================================================================
 * COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *                                                                          */
/** \file      ix_cc_mpls_msg_helper.c
 *
 *  \brief     The file contains headers of MPLS CC message helper functions 
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_mpls_msg_helper.c,v 1.10.2.1 2004/03/12 10:40:15 wneubaue Exp $";

#include <ix_types.h>
#include <ix_error.h>

#include <ix_ossl.h>
#include <ix_rm.h>

#if !defined(IX_EXCLUDE_CCI)
#include <ix_cci.h>
#endif 

#include <ix_cc_error.h>
#include <ix_cc.h>

#include <bindings.h>

#include <cc/ix_cc_msup.h>

#include <cc/ix_cc_rtmv4.h>

/* Include MPLS-related files */
#include "../../include/mpls_defs.h"
#include "cc/ix_cc_mpls.h"
#include "cc/internal/ix_cc_mpls_msg.h"
#include "cc/internal/ix_cc_mpls_lsp.h"
#include "cc/internal/ix_cc_mpls_nhlfe.h"
#include "cc/internal/ix_cc_mpls_internal.h"




/*
 * FUNCTION: _ix_cc_mpls_cb_general()   */
/**
 * ABSTRACT: internal callback for the async messaging API
 * 
 * DESCRIPTION: This function is a generic internal callback used by 
 *              a number of async API calls 
 *
 * \param  arg_Result   - result of the asynchronous operation
 * \param  arg_pContext - context provided by caller
 * \param  arg_Msg      - return message
 * \param  arg_MsgLen   - length of the message
 * 
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 */
PRIVATE ix_error 
_ix_cc_mpls_cb_general(ix_error  arg_Result,
                       void*     arg_pContext,
                       void*     arg_pMsg,
                       ix_uint32 arg_MsgLength)
{
    ix_error                   err         = IX_SUCCESS;
    ix_cc_mpls_message_ctx    *pContext  = arg_pContext;
    ix_cc_mpls_cb_general      Callback;

    /* Retrieve the user callback */
    Callback = pContext->pCallBack;
    
    if (Callback == NULL)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cb_general(): Callback is NULL"));
    }
    else 
    {
        err = Callback(arg_Result, pContext->pUserContext);
    }
    ix_ossl_free(arg_pContext);

    return err;
}



/*
 * FUNCTION: _ix_cc_mpls_cbi_lsp_create()   */
/**
 * ABSTRACT: internal callback for the async messaging API
 * 
 * DESCRIPTION: This function is an internal callback for the async 
 *              messaging API, ix_cc_mpls_async_lsp_create().
 *
 * \param  arg_Result   - result of the asynchronous operation
 * \param  arg_pContext - context provided by caller
 * \param  arg_Msg      - return message
 * \param  arg_MsgLen   - length of the message
 * 
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 */
PRIVATE ix_error 
_ix_cc_mpls_cbi_lsp_create(ix_error  arg_Result,
                           void*     arg_pContext,
                           void*     arg_pMsg,
                           ix_uint32 arg_MsgLength)
{
    ix_error                   err         = IX_SUCCESS;
    ix_cc_mpls_message_ctx    *pContext  = arg_pContext;
    ix_cc_mpls_cb_lsp_create   Callback;

    /* Retrieve the user callback */
    Callback = pContext->pCallBack;

    if (NULL == Callback)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_lsp_create(): Callback is NULL"));
    }
    else if ( arg_pMsg == NULL )
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_lsp_create(): Returned data is NULL."));
    }
    else if ( arg_MsgLength != sizeof(ix_cc_mpls_message_lsp_create_rep) )
    {
        err = IX_ERROR_WARNING( IX_CC_ERROR_INTERNAL,
                               ("_ix_cc_mpls_cbi_lsp_create(): message length " \
                               "doesn't match expected size"));
    }
    else 
    {
        ix_cc_mpls_message_lsp_create_rep LspCreateRep = 
                *(ix_cc_mpls_message_lsp_create_rep *) arg_pMsg;

        err = Callback(arg_Result,
                       pContext->pUserContext,
                       LspCreateRep.nhlfeSetHandle,
                       LspCreateRep.nhlfe_num,
                       LspCreateRep.nhlfeHandle,
                       LspCreateRep.LspHandle);
 
    }

    ix_ossl_free(arg_pContext);

    return err;
}





/*
 * FUNCTION: _ix_cc_mpls_cbi_lsp_modify()   */
/**
 * ABSTRACT: internal callback for the async messaging API
 * 
 * DESCRIPTION: This function is an internal callback for the async 
 *              messaging API, ix_cc_mpls_async_lsp_modify().
 *
 * \param  arg_Result   - result of the asynchronous operation
 * \param  arg_pContext - context provided by caller
 * \param  arg_Msg      - return message
 * \param  arg_MsgLen   - length of the message
 * 
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 */
PRIVATE ix_error 
_ix_cc_mpls_cbi_lsp_modify(ix_error  arg_Result,
                           void*     arg_pContext,
                           void*     arg_pMsg,
                           ix_uint32 arg_MsgLength)
{
    ix_error                   err         = IX_SUCCESS;
    ix_cc_mpls_message_ctx    *pContext  = arg_pContext;
    ix_cc_mpls_cb_lsp_modify   Callback;

    /* Retrieve the user callback */
    Callback = pContext->pCallBack;

    if (NULL == Callback)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
           (                   "_ix_cc_mpls_cbi_lsp_modify(): Callback is NULL"));
    }
    else if ( arg_pMsg == NULL )
    {
        err = IX_ERROR_WARNING( IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_lsp_modify(): Returned data is NULL."));
    }
    else if ( arg_MsgLength != sizeof(ix_cc_mpls_nhlfe_handle) )
    {
        err = IX_ERROR_WARNING( IX_CC_ERROR_INTERNAL,
                ("_ix_cc_mpls_cbi_lsp_create(): message length " \
                 "doesn't match expected size"));
    }
    else 
    {
        ix_cc_mpls_nhlfe_handle LspModifyRep = 
                *(ix_cc_mpls_nhlfe_handle*) arg_pMsg;

        err = Callback(arg_Result,
                       pContext->pUserContext,
                       LspModifyRep);
    }

    ix_ossl_free(arg_pContext);

    return err;
}






/*
 * FUNCTION: _ix_cc_mpls_cbi_lsp_query()   */
/**
 * ABSTRACT: internal callback for the async messaging API
 * 
 * DESCRIPTION: This function is an internal callback for the async 
 *              messaging API, ix_cc_mpls_async_lsp_query().
 *
 * \param  arg_Result   - result of the asynchronous operation
 * \param  arg_pContext - context provided by caller
 * \param  arg_Msg      - return message
 * \param  arg_MsgLen   - length of the message
 * 
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 */
PRIVATE ix_error 
_ix_cc_mpls_cbi_lsp_query(ix_error  arg_Result,
                          void*     arg_pContext,
                          void*     arg_pMsg,
                          ix_uint32 arg_MsgLength)
{
    ix_error                   err      = IX_SUCCESS;
    ix_cc_mpls_message_ctx    *pContext = arg_pContext;
    ix_cc_mpls_cb_lsp_query    Callback;

    /* Retrieve the user callback */
    Callback = pContext->pCallBack;

    if (NULL == Callback)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_lsp_query(): Callback is NULL"));
    }
    else if ( arg_pMsg == NULL )
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_lsp_query(): Returned data is NULL."));
    }
    else if ( arg_MsgLength != sizeof(ix_cc_mpls_lsp) )
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
                               ("_ix_cc_mpls_cbi_lsp_query(): message length " \
                               "doesn't match expected size"));
    }
    else
    {
        err = Callback(arg_Result,
                       pContext->pUserContext,
                       (ix_cc_mpls_lsp*)arg_pMsg);
    }

    ix_ossl_free(arg_pContext);

    return err;
}




/*
 * FUNCTION: _ix_cc_mpls_cbi_lsp_stats_query()   */
/**
 * ABSTRACT: internal callback for the async messaging API
 * 
 * DESCRIPTION: This function is an internal callback for the async 
 *              messaging API, _ix_cc_mpls_async_lsp_stats_query().
 *
 * \param  arg_Result   - result of the asynchronous operation
 * \param  arg_pContext - context provided by caller
 * \param  arg_Msg      - return message
 * \param  arg_MsgLen   - length of the message
 * 
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 */
PRIVATE ix_error 
_ix_cc_mpls_cbi_lsp_stats_query(ix_error  arg_Result,
                                void*     arg_pContext,
                                void*     arg_pMsg,
                                ix_uint32 arg_MsgLength)
{
    ix_error                       err      = IX_SUCCESS;
    ix_cc_mpls_message_ctx        *pContext = arg_pContext;
    ix_cc_mpls_cb_lsp_stats_query  Callback;

    /* Retrieve the user callback */
    Callback = pContext->pCallBack;

    if (NULL == Callback)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_lsp_stats_query(): Callback is NULL"));
    }
    else if ( arg_pMsg == NULL )
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_lsp_stats_query(): Returned data is NULL."));
    }
    else if ( arg_MsgLength != sizeof(ix_cc_mpls_lsp_stats) )
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
                               ("_ix_cc_mpls_cbi_lsp_stats_query(): message length " \
                               "doesn't match expected size"));
    }
    else 
    {
        err = Callback(arg_Result,
                       pContext->pUserContext,
                       (ix_cc_mpls_lsp_stats*)arg_pMsg);
    }

    ix_ossl_free(arg_pContext);

    return err;
}




/*
 * FUNCTION: _ix_cc_mpls_cbi_nhlfe_create()   */
/**
 * ABSTRACT: internal callback for the async messaging API
 * 
 * DESCRIPTION: This function is an internal callback for the async 
 *              messaging API, _ix_cc_mpls_async_nhlfe_create().
 *
 * \param  arg_Result   - result of the asynchronous operation
 * \param  arg_pContext - context provided by caller
 * \param  arg_Msg      - return message
 * \param  arg_MsgLen   - length of the message
 * 
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 */
PRIVATE ix_error 
_ix_cc_mpls_cbi_nhlfe_create(ix_error  arg_Result,
                                void*     arg_pContext,
                                void*     arg_pMsg,
                                ix_uint32 arg_MsgLength)
{
    ix_error                    err      = IX_SUCCESS;
    ix_cc_mpls_message_ctx     *pContext = arg_pContext;
    ix_cc_mpls_cb_nhlfe_create  Callback;

    /* Retrieve the user callback */
    Callback = pContext->pCallBack;

    if (NULL == Callback)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_nhlfe_create(): Callback is NULL"));
    }
    else if ( arg_pMsg == NULL )
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_nhlfe_create(): Returned data is NULL."));
    }
    else if ( arg_MsgLength != sizeof(ix_cc_mpls_nhlfe_handle) )
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
                               ("_ix_cc_mpls_cbi_nhlfe_create(): message length " \
                               "doesn't match expected size"));
    }
    else
    {
        err = Callback(arg_Result,
                       pContext->pUserContext,
                       *((ix_cc_mpls_nhlfe_handle*)arg_pMsg));
    }

    ix_ossl_free(arg_pContext);

    return err;
}





/*
 * FUNCTION: _ix_cc_mpls_cbi_nhlfe_query()   */
/**
 * ABSTRACT: internal callback for the async messaging API
 * 
 * DESCRIPTION: This function is an internal callback for the async 
 *              messaging API, _ix_cc_mpls_async_nhlfe_query().
 *
 * \param  arg_Result   - result of the asynchronous operation
 * \param  arg_pContext - context provided by caller
 * \param  arg_Msg      - return message
 * \param  arg_MsgLen   - length of the message
 * 
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 */
PRIVATE ix_error 
_ix_cc_mpls_cbi_nhlfe_query(ix_error  arg_Result,
                            void*     arg_pContext,
                            void*     arg_pMsg,
                            ix_uint32 arg_MsgLength)
{
    ix_error                   err      = IX_SUCCESS;
    ix_cc_mpls_message_ctx    *pContext = arg_pContext;
    ix_cc_mpls_cb_nhlfe_query  Callback;

    /* Retrieve the user callback */
    Callback = pContext->pCallBack;

    if (NULL == Callback)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_nhlfe_query(): Callback is NULL"));
    } 
    else if ( arg_pMsg == NULL )
    {
       err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                              ("_ix_cc_mpls_cbi_nhlfe_query(): Returned data is NULL."));
    }
    else if ( arg_MsgLength != sizeof(ix_cc_mpls_nhlfe) )
    {
       err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
                              ("_ix_cc_mpls_cbi_nhlfe_query(): message length " \
                              "doesn't match expected size"));
    }
    else
    {
        err = Callback(arg_Result,
                       pContext->pUserContext,
                       (ix_cc_mpls_nhlfe*)arg_pMsg);
    }

    ix_ossl_free(arg_pContext);

    return err;
}



/*
 * FUNCTION: _ix_cc_mpls_cbi_nhlfe_stats_query()   */
/**
 * ABSTRACT: internal callback for the async messaging API
 * 
 * DESCRIPTION: This function is an internal callback for the async 
 *              messaging API, _ix_cc_mpls_async_nhlfe_stats_query().
 *
 * \param  arg_Result   - result of the asynchronous operation
 * \param  arg_pContext - context provided by caller
 * \param  arg_Msg      - return message
 * \param  arg_MsgLen   - length of the message
 * 
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 */
PRIVATE ix_error 
_ix_cc_mpls_cbi_nhlfe_stats_query(ix_error  arg_Result,
                                  void*     arg_pContext,
                                  void*     arg_pMsg,
                                  ix_uint32 arg_MsgLength)
{
    ix_error                         err      = IX_SUCCESS;
    ix_cc_mpls_message_ctx          *pContext = arg_pContext;
    ix_cc_mpls_cb_nhlfe_stats_query  Callback;

    /* Retrieve the user callback */
    Callback = pContext->pCallBack;

    if (NULL == Callback)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_nhlfe_stats_query(): Callback is NULL"));
    }
    else if ( arg_pMsg == NULL )
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_nhlfe_stats_query(): Returned data is NULL."));
    }
    else if ( arg_MsgLength != sizeof(ix_cc_mpls_nhlfe_stats) )
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
                               ("_ix_cc_mpls_cbi_nhlfe_stats_query(): message length " \
                               "doesn't match expected size"));
    }
    else
    {
        err = Callback(arg_Result,
                       pContext->pUserContext,
                       (ix_cc_mpls_nhlfe_stats*)arg_pMsg);
    }

    ix_ossl_free(arg_pContext);

    return err;
}






/*
 * FUNCTION: _ix_cc_mpls_cbi_nhlfe_set_create()   */
/**
 * ABSTRACT: internal callback for the async messaging API
 * 
 * DESCRIPTION: This function is an internal callback for the async 
 *              messaging API, _ix_cc_mpls_async_nhlfe_set_create().
 *
 * \param  arg_Result   - result of the asynchronous operation
 * \param  arg_pContext - context provided by caller
 * \param  arg_Msg      - return message
 * \param  arg_MsgLen   - length of the message
 * 
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 */
PRIVATE ix_error 
_ix_cc_mpls_cbi_nhlfe_set_create(ix_error  arg_Result,
                                 void*     arg_pContext,
                                 void*     arg_pMsg,
                                 ix_uint32 arg_MsgLength)
{
    ix_error                         err      = IX_SUCCESS;
    ix_cc_mpls_message_ctx          *pContext = arg_pContext;
    ix_cc_mpls_cb_nhlfe_set_create   Callback;

    /* Retrieve the user callback */
    Callback = pContext->pCallBack;

    if (NULL == Callback)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_nhlfe_set_create(): Callback is NULL"));
    }    
    else if ( arg_pMsg == NULL )
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_nhlfe_set_create(): Returned data is NULL."));
    }
    else if ( arg_MsgLength != sizeof(ix_cc_mpls_nhlfe_handle) )
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
                               ("_ix_cc_mpls_cbi_nhlfe_set_create(): message length " \
                               "doesn't match expected size"));
    }
    else
    {
        err = Callback(arg_Result,
                       pContext->pUserContext,
                       *(ix_cc_mpls_nhlfe_handle*)arg_pMsg);
    }

    ix_ossl_free(arg_pContext);

    return err;
}




/*
 * FUNCTION: _ix_cc_mpls_cbi_nhlfe_set_query()   */
/**
 * ABSTRACT: internal callback for the async messaging API
 * 
 * DESCRIPTION: This function is an internal callback for the async 
 *              messaging API, _ix_cc_mpls_async_nhlfe_set_query().
 *
 * \param  arg_Result   - result of the asynchronous operation
 * \param  arg_pContext - context provided by caller
 * \param  arg_Msg      - return message
 * \param  arg_MsgLen   - length of the message
 * 
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 */
PRIVATE ix_error 
_ix_cc_mpls_cbi_nhlfe_set_query(ix_error  arg_Result,
                                void*     arg_pContext,
                                void*     arg_pMsg,
                                ix_uint32 arg_MsgLength)
{
    ix_error                       err      = IX_SUCCESS;
    ix_cc_mpls_message_ctx        *pContext = arg_pContext;
    ix_cc_mpls_cb_nhlfe_set_query  Callback;

    /* Retrieve the user callback */
    Callback = pContext->pCallBack;

    if (NULL == Callback)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_nhlfe_set_query(): Callback is NULL"));
    }
    else if ( arg_pMsg == NULL )
    {        
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_nhlfe_set_query(): Returned data is NULL."));
    }
    else if ( arg_MsgLength != sizeof(ix_cc_mpls_message_nhlfe_set_query_rep) )
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
                               ("_ix_cc_mpls_cbi_nhlfe_set_query(): message length " \
                               "doesn't match expected size"));
    }
    else 
    {
        ix_cc_mpls_message_nhlfe_set_query_rep *NhlfeSetQueryRep = 
            (ix_cc_mpls_message_nhlfe_set_query_rep *) arg_pMsg;

        err = Callback(arg_Result, 
                       pContext->pUserContext,
                       NhlfeSetQueryRep->NhlfeSetSize,
                       NhlfeSetQueryRep->NhlfeSetHandleTable);
    }

    ix_ossl_free(arg_pContext);

    return err;
}




/*
 * FUNCTION: _ix_cc_mpls_cbi_exp2phb_query()   */
/**
 * ABSTRACT: internal callback for the async messaging API
 * 
 * DESCRIPTION: This function is an internal callback for the async 
 *              messaging API, _ix_cc_mpls_async_exp2phb_query().
 *
 * \param  arg_Result   - result of the asynchronous operation
 * \param  arg_pContext - context provided by caller
 * \param  arg_Msg      - return message
 * \param  arg_MsgLen   - length of the message
 * 
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 */
PRIVATE ix_error 
_ix_cc_mpls_cbi_exp2phb_query(ix_error  arg_Result,
                              void*     arg_pContext,
                              void*     arg_pMsg,
                              ix_uint32 arg_MsgLength)
{
    ix_error                     err      = IX_SUCCESS;
    ix_cc_mpls_message_ctx      *pContext = arg_pContext;
    ix_cc_mpls_cb_exp2phb_query  Callback;

    /* Retrieve the user callback */
    Callback = pContext->pCallBack;

    if (NULL == Callback)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_exp2phb_query(): Callback is NULL"));
    }
    else if ( arg_pMsg == NULL )
    {       
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_exp2phb_query(): Returned data is NULL."));
    }
    else if ( arg_MsgLength != MPLS_MAX_PHB_NUM * sizeof(ix_cc_mpls_phb2exp) )
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
                               ("_ix_cc_mpls_cbi_exp2phb_query(): message length " \
                               "doesn't match expected size"));
    }
    else
    {
        err = Callback(arg_Result,
                       pContext->pUserContext,
                       (ix_cc_mpls_exp2phb*) arg_pMsg);
    }
    ix_ossl_free(arg_pContext);

    return err;
}





/*
 * FUNCTION: _ix_cc_mpls_cbi_phb2exp_query()   */
/**
 * ABSTRACT: internal callback for the async messaging API
 * 
 * DESCRIPTION: This function is an internal callback for the async 
 *              messaging API, _ix_cc_mpls_async_phb2exp_query().
 *
 * \param  arg_Result   - result of the asynchronous operation
 * \param  arg_pContext - context provided by caller
 * \param  arg_Msg      - return message
 * \param  arg_MsgLen   - length of the message
 * 
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 */
PRIVATE ix_error 
_ix_cc_mpls_cbi_phb2exp_query(ix_error  arg_Result,
                              void*     arg_pContext,
                              void*     arg_pMsg,
                              ix_uint32 arg_MsgLength)
{
    ix_error                     err      = IX_SUCCESS;
    ix_cc_mpls_message_ctx      *pContext = arg_pContext;
    ix_cc_mpls_cb_phb2exp_query  Callback;

    /* Retrieve the user callback */
    Callback = pContext->pCallBack;

    if (NULL == Callback)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_phb2exp_query(): Callback is NULL"));
    }
    else if ( arg_pMsg == NULL )
    {       
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_phb2exp_query(): Returned data is NULL."));
    }
    else if ( arg_MsgLength != MPLS_MAX_EXP_NUM * sizeof(ix_cc_mpls_exp2phb) )
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
                               ("_ix_cc_mpls_cbi_phb2exp_query(): message length " \
                               "doesn't match expected size"));
    }
    else 
    {
        err = Callback(arg_Result, 
                       pContext->pUserContext,
                       (ix_cc_mpls_phb2exp*) arg_pMsg);
    }

    ix_ossl_free(arg_pContext);

    return err;
}





/*
 * FUNCTION: _ix_cc_mpls_cbi_param_query()   */
/**
 * ABSTRACT: internal callback for the async messaging API
 * 
 * DESCRIPTION: This function is an internal callback for the async 
 *              messaging API, _ix_cc_mpls_async_param_query().
 *
 * \param  arg_Result   - result of the asynchronous operation
 * \param  arg_pContext - context provided by caller
 * \param  arg_Msg      - return message
 * \param  arg_MsgLen   - length of the message
 * 
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 */
PRIVATE ix_error 
_ix_cc_mpls_cbi_param_query(ix_error  arg_Result,
                            void*     arg_pContext,
                            void*     arg_pMsg,
                            ix_uint32 arg_MsgLength)
{
    ix_error                   err      = IX_SUCCESS;
    ix_cc_mpls_message_ctx    *pContext = arg_pContext;
    ix_cc_mpls_cb_param_query  Callback;

    /* Retrieve the user callback */
    Callback = pContext->pCallBack;

    if (NULL == Callback)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_param_query(): Callback is NULL"));
    }
    else if ( arg_pMsg == NULL )
    {        
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_param_query(): Returned data is NULL."));
    }
    else if ( arg_MsgLength != sizeof(ix_cc_mpls_params) )
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
                               ("_ix_cc_mpls_cbi_param_query(): message length " \
                               "doesn't match expected size"));
    }
    else 
    {
        err = Callback(arg_Result, 
                       pContext->pUserContext,
                       (ix_cc_mpls_params*) arg_pMsg);
    }

    ix_ossl_free(arg_pContext);

    return err;
}





/*
 * FUNCTION: _ix_cc_mpls_cbi_port_stats_query()   */
/**
 * ABSTRACT: internal callback for the async messaging API
 * 
 * DESCRIPTION: This function is an internal callback for the async 
 *              messaging API, _ix_cc_mpls_async_port_stats_query().
 *
 * \param  arg_Result   - result of the asynchronous operation
 * \param  arg_pContext - context provided by caller
 * \param  arg_Msg      - return message
 * \param  arg_MsgLen   - length of the message
 * 
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 */
PRIVATE ix_error 
_ix_cc_mpls_cbi_port_stats_query(ix_error  arg_Result,
                                 void*     arg_pContext,
                                 void*     arg_pMsg,
                                 ix_uint32 arg_MsgLength)
{
    ix_error                        err      = IX_SUCCESS;
    ix_cc_mpls_message_ctx         *pContext = arg_pContext;
    ix_cc_mpls_cb_port_stats_query  Callback;

    /* Retrieve the user callback */
    Callback = pContext->pCallBack;

    if (NULL == Callback)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_port_stats_query(): Callback is NULL"));
    }
    else if ( arg_pMsg == NULL )
    {        
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_port_stats_query(): Returned data is NULL."));
    }
    else if ( arg_MsgLength != sizeof(ix_cc_mpls_input_port_stats) )
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
                               ("_ix_cc_mpls_cbi_port_stats_query(): message length " \
                               "doesn't match expected size"));
    }
    else 
    {
        err = Callback(arg_Result, 
                       pContext->pUserContext,
                       (ix_cc_mpls_input_port_stats*) arg_pMsg);
    }

    ix_ossl_free(arg_pContext);

    return err;
}



/*
 * FUNCTION: _ix_cc_mpls_cbi_get_mpls_context()   */
/**
 * ABSTRACT: internal callback for the async messaging API
 * 
 * DESCRIPTION: This function is an internal callback for the async 
 *              messaging API, ix_cc_mpls_async_get_mpls_context().
 *
 * \param  arg_Result   - result of the asynchronous operation
 * \param  arg_pContext - context provided by caller
 * \param  arg_Msg      - return message
 * \param  arg_MsgLen   - length of the message
 * 
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 */
PRIVATE ix_error 
_ix_cc_mpls_cbi_get_mpls_context(ix_error  arg_Result,
                                 void*     arg_pContext,
                                 void*     arg_pMsg,
                                 ix_uint32 arg_MsgLength)
{
    ix_error                        err      = IX_SUCCESS;
    ix_cc_mpls_message_ctx         *pContext = arg_pContext;
    ix_cc_mpls_cb_get_mpls_context  Callback;

    /* Retrieve the user callback */
    Callback = pContext->pCallBack;

    if (NULL == Callback)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_get_mpls_context(): Callback is NULL"));
    }
    else if ( arg_pMsg == NULL )
    {        
        err = IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                               ("_ix_cc_mpls_cbi_get_mpls_context(): Returned data is NULL."));
    }
    else if ( arg_MsgLength != sizeof(ix_cc_mpls_context*) )
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
                               ("_ix_cc_mpls_cbi_get_mpls_context(): message length " \
                               "doesn't match expected size"));
    }
    else 
    {
        err = Callback(arg_Result, 
                       pContext->pUserContext,
                       *((ix_cc_mpls_context**) arg_pMsg));
    }

    ix_ossl_free(arg_pContext);

    return err;
}





/*
 * FUNCTION: ix_cc_mpls_async_lsp_create()   */
/**
 * ABSTRACT: Sends message to create LSP entry
 *
 * DESCRIPTION: This message helper function creates an LSP description 
 *              defining a FEC to NHLFE, or ILM to NHLFE mapping, and creates
 *              corresponding IP Routing table, ILM and NHLFE entries (if necessary).
 *              See ix_cc_mpls_lsp_create() for a detailed description
 *
 * \param arg_pLsp         - pointer to an ix_s_cc_mpls_lsp structure, defining an LSP
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will be passed to 
 *                           the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_lsp_create(ix_cc_mpls_lsp            *arg_pLsp,
                            ix_cc_mpls_cb_lsp_create   arg_Callback,
                            void                      *arg_pUserContext)
{
    ix_error                       err     = IX_SUCCESS;
    ix_uint32                      idx;
    ix_cc_mpls_message_ctx        *pIntCtx = NULL;
    ix_cc_mpls_message_lsp_create  LspMsg;

    if (NULL == arg_Callback) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
	              ("ix_cc_mpls_async_lsp_create(): function was called with NULL callback"));

    /* create ix_cc_mpls_message_ctx */
    pIntCtx = ix_ossl_malloc(sizeof(ix_cc_mpls_message_ctx));
    
    if (NULL == pIntCtx) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
	              ("ix_cc_mpls_async_lsp_create(): Fail to allocate system memory"));

    LspMsg.Lsp = *arg_pLsp;
    if (LspMsg.Lsp.nhlfeInfoType == IX_CC_MPLS_LSP_NHLFE_DATA_TYPE)
        for(idx = 0; idx < LspMsg.Lsp.nhlfe_num; idx ++)
            LspMsg.NhlfeSet[idx] = *(arg_pLsp->outSeg.nhlfeData[idx]);

    pIntCtx->pCallBack    = arg_Callback;
    pIntCtx->pUserContext = arg_pUserContext;

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_MPLS_MSG_INPUT,
                                           (ix_cc_msghlp_callback)_ix_cc_mpls_cbi_lsp_create,
                                           pIntCtx,
                                           IX_CC_MPLS_MSG_LSP_CREATE,
                                           (void*)(&LspMsg),
                                           sizeof(LspMsg)),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING, 
                 cleanup_context);

    return IX_SUCCESS;

cleanup_context:

    if(pIntCtx != NULL)
    {
        ix_ossl_free(pIntCtx);
    }/* end if cbContext != NULL */
    return err;
}





/*
 * FUNCTION: ix_cc_mpls_async_lsp_delete()   */
/**
 * ABSTRACT: Sends message to delete LSP entry
 *
 * DESCRIPTION: This message helper function deletes an LSP description,
 *              together with its corresponding IP Routing table,
 *              ILM and NHLFE entries (if necessary).
 *              See ix_cc_mpls_lsp_delete() for a detailed description
 *
 * \param arg_LspHandle    - handle identifying an LSP. 
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will be passed
 *                           to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_lsp_delete(ix_cc_mpls_lsp_handle  arg_LspHandle,
                            ix_cc_mpls_cb_general  arg_Callback,
                            void                  *arg_pUserContext)
{
    ix_error                  err     = IX_SUCCESS;
    ix_cc_mpls_message_ctx   *pIntCtx = NULL;

    if (NULL == arg_Callback) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
	              ("ix_cc_mpls_async_lsp_delete(): function was called with NULL callback"));

    /* create ix_cc_mpls_message_ctx */
    pIntCtx = ix_ossl_malloc(sizeof(ix_cc_mpls_message_ctx));
    
    if (NULL == pIntCtx) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
	              ("ix_cc_mpls_async_lsp_delete(): Fail to allocate system memory"));
    
    pIntCtx->pCallBack    = arg_Callback;
    pIntCtx->pUserContext = arg_pUserContext;

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_MPLS_MSG_INPUT,
                                           (ix_cc_msghlp_callback)_ix_cc_mpls_cb_general,
                                           pIntCtx,
                                           IX_CC_MPLS_MSG_LSP_DELETE,
                                           (void*)(&arg_LspHandle),
                                           sizeof(arg_LspHandle)),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING, 
                 cleanup_context);

    return IX_SUCCESS;

cleanup_context:

    if(pIntCtx != NULL)
    {
        ix_ossl_free(pIntCtx);
    }/* end if cbContext != NULL */
    return err;
}




/*
 * FUNCTION: ix_cc_mpls_async_lsp_modify()   */
/**
 * ABSTRACT: Sends message to modify LSP entry
 *
 * DESCRIPTION: This message helper function modifies the NHLFE set mapping 
 *              for a given LSP. See ix_cc_mpls_lsp_modify()
 *              for a detailed description.
 *
 * \param  arg_LspHandle        - handle identifying an LSP
 * \param  arg_nhlfeInfoType    - specifies weather function should use handle to predefined
 *                                NHLFE SET or the table of single NHLFEs during LSP modification.
 * \param  arg_nhlfeHandles_num - the number of NHLFE in a table to be cludtered in a set and
 *                                associated with the requested LSP entry.
 * \param  arg_nhlfeHandles     - table of nhlfe handles to be associated with the LSP entry.
 *                                If LSP entry is to be associated with NHLFE SET
 *                                this is the table with handles 
 *                                to all members of this NHLFE SET. 
 * \param  arg_pnhlfeSetHandle  - pointer to a placeholder of handle to predefined NHLFE SET
 *                                if LSP entry is to be associated with allready
 *                                created NHLFE SET this is the pointer to the handle to that 
 *                                NHLFE_SET 
 * \param  arg_lspParams        - pointer to the array of LSP parameters for each element of
 *                                the new NHLFE set
 * \param arg_Callback          - pointer to the client provided callback function
 * \param arg_pUserContext      - pointer to the client defined context that will 
 *                                be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_lsp_modify(ix_cc_mpls_lsp_handle    arg_LspHandle,
                            ix_uint8                 arg_nhlfeInfoType,
                            ix_cc_mpls_nhlfe_handle  arg_nhlfeSetHandle,
                            ix_uint16                arg_nhlfeHandles_num,
                            ix_cc_mpls_nhlfe_handle  arg_nhlfeHandles[MPLS_MAX_NHLFE_IN_SET],
                            ix_cc_mpls_lsp_param     arg_lspParams[MPLS_MAX_NHLFE_IN_SET],
                            ix_cc_mpls_cb_lsp_modify arg_Callback,
                            void                    *arg_pUserContext)
{
    ix_error                       err     = IX_SUCCESS;
    ix_cc_mpls_message_ctx        *pIntCtx = NULL;
    ix_cc_mpls_message_lsp_modify  MsgData;

    if (NULL == arg_Callback) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
	              ("ix_cc_mpls_async_lsp_modify(): function was called with NULL callback"));

    /* create ix_cc_mpls_message_ctx */
    pIntCtx = ix_ossl_malloc(sizeof(ix_cc_mpls_message_ctx));
    
    if (NULL == pIntCtx) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
	              ("ix_cc_mpls_async_lsp_modify(): Fail to allocate system memory"));
    
    MsgData.LspHandle        = arg_LspHandle;
    MsgData.nhlfeInfoType    = arg_nhlfeInfoType;
    MsgData.NhlfeSetHandle   = arg_nhlfeSetHandle;
    MsgData.nhlfeHandles_num = arg_nhlfeHandles_num;

    ix_ossl_memcpy(&(MsgData.nhlfeHandles), arg_nhlfeHandles,
                    sizeof(ix_cc_mpls_nhlfe_handle)*arg_nhlfeHandles_num);
    ix_ossl_memcpy(&(MsgData.lspParams), arg_lspParams,
                    sizeof(ix_cc_mpls_lsp_param)*arg_nhlfeHandles_num);

    pIntCtx->pCallBack    = arg_Callback;
    pIntCtx->pUserContext = arg_pUserContext;

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_MPLS_MSG_INPUT,
                                           (ix_cc_msghlp_callback)_ix_cc_mpls_cbi_lsp_modify,
                                           pIntCtx,
                                           IX_CC_MPLS_MSG_LSP_MODIFY,
                                           (void*)(&MsgData),
                                           sizeof(MsgData)),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING, 
                 cleanup_context);

    return IX_SUCCESS;

cleanup_context:

    if(pIntCtx != NULL)
    {
        ix_ossl_free(pIntCtx);
    }/* end if cbContext != NULL */
    return err;
}




/*
 * FUNCTION: ix_cc_mpls_async_lsp_query()   */
/**
 * ABSTRACT: Sends message to get LSP
 *
 * DESCRIPTION: This message helper function reads and LSP description. 
 *
 * \param arg_LspHandle         - handle identifying an LSP
 * \param arg_Callback          - pointer to the client provided callback function
 * \param arg_pUserContext      - pointer to the client defined context that will 
 *                                be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_lsp_query(ix_cc_mpls_lsp_handle    arg_LspHandle,
                           ix_cc_mpls_cb_lsp_query  arg_Callback,
                           void                    *arg_pUserContext)
{
    ix_error                       err     = IX_SUCCESS;
    ix_cc_mpls_message_ctx        *pIntCtx = NULL;

    if (NULL == arg_Callback) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
	              ("ix_cc_mpls_async_lsp_query(): function was called with NULL callback"));

    /* create ix_cc_mpls_message_ctx */
    pIntCtx = ix_ossl_malloc(sizeof(ix_cc_mpls_message_ctx));
    
    if (NULL == pIntCtx) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
	              ("ix_cc_mpls_async_lsp_query(): Fail to allocate system memory"));
    
    pIntCtx->pCallBack    = arg_Callback;
    pIntCtx->pUserContext = arg_pUserContext;

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_MPLS_MSG_INPUT,
                                           (ix_cc_msghlp_callback)_ix_cc_mpls_cbi_lsp_query,
                                           pIntCtx,
                                           IX_CC_MPLS_MSG_LSP_QUERY,
                                           (void*)(&arg_LspHandle),
                                           sizeof(arg_LspHandle)),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING, 
                 cleanup_context);

    return IX_SUCCESS;

cleanup_context:

    if(pIntCtx != NULL)
    {
        ix_ossl_free(pIntCtx);
    }/* end if cbContext != NULL */
    return err;
}







/*
 * FUNCTION: ix_cc_mpls_async_lsp_stats_query()   */
/**
 * ABSTRACT: Sends message to get LSP statistics
 *
 * DESCRIPTION: This message helper function reads and LSP description. 
 *
 * \param arg_LspHandle         - handle identifying an LSP
 * \param arg_Callback          - pointer to the client provided callback function
 * \param arg_pUserContext      - pointer to the client defined context that will 
 *                                be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_lsp_stats_query(ix_cc_mpls_lsp_handle          arg_LspHandle,
                                 ix_cc_mpls_cb_lsp_stats_query  arg_Callback,
                                 void                          *arg_pUserContext)
{
    ix_error                       err     = IX_SUCCESS;
    ix_cc_mpls_message_ctx        *pIntCtx = NULL;

    if (NULL == arg_Callback) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
	              ("ix_cc_mpls_async_lsp_stats_query(): function was called with NULL callback"));

    /* create ix_cc_mpls_message_ctx */
    pIntCtx = ix_ossl_malloc(sizeof(ix_cc_mpls_message_ctx));
    
    if (NULL == pIntCtx) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
	              ("ix_cc_mpls_async_lsp_query(): Fail to allocate system memory"));
    
    pIntCtx->pCallBack    = arg_Callback;
    pIntCtx->pUserContext = arg_pUserContext;

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_MPLS_MSG_INPUT,
                                           (ix_cc_msghlp_callback)_ix_cc_mpls_cbi_lsp_stats_query,
                                           pIntCtx,
                                           IX_CC_MPLS_MSG_LSP_STATS_QUERY,
                                           (void*)(&arg_LspHandle),
                                           sizeof(arg_LspHandle)),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING, 
                 cleanup_context);

    return IX_SUCCESS;

cleanup_context:

    if(pIntCtx != NULL)
    {
        ix_ossl_free(pIntCtx);
    }/* end if cbContext != NULL */
    return err;
}





/*
 * FUNCTION: ix_cc_mpls_async_lsp_purge()   */
/**
 * ABSTRACT: Sends message to purge LSP table
 *
 * DESCRIPTION: This message helper function deletes all LSP definitions,
 *              together with its corresponding IP Routing table,
 *              ILM and NHLFE entries. 
 *
 * \param arg_Callback          - pointer to the client provided callback function
 * \param arg_pUserContext      - pointer to the client defined context that will 
 *                                be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_lsp_purge(ix_cc_mpls_cb_general   arg_Callback,
                           void                   *arg_pUserContext)

{
    ix_error                       err     = IX_SUCCESS;
    ix_cc_mpls_message_ctx        *pIntCtx = NULL;

    if (NULL == arg_Callback) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
	              ("ix_cc_mpls_async_lsp_purge(): function was called with NULL callback"));

    /* create ix_cc_mpls_message_ctx */
    pIntCtx = ix_ossl_malloc(sizeof(ix_cc_mpls_message_ctx));
    
    if (NULL == pIntCtx) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
	              ("ix_cc_mpls_async_lsp_purge(): Fail to allocate system memory"));
    
    pIntCtx->pCallBack    = arg_Callback;
    pIntCtx->pUserContext = arg_pUserContext;

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_MPLS_MSG_INPUT,
                                           (ix_cc_msghlp_callback)_ix_cc_mpls_cb_general,
                                           pIntCtx,
                                           IX_CC_MPLS_MSG_LSP_STATS_QUERY,
                                           (void*)(NULL),
                                           0),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING, 
                 cleanup_context);

    return IX_SUCCESS;

cleanup_context:

    if(pIntCtx != NULL)
    {
        ix_ossl_free(pIntCtx);
    }/* end if cbContext != NULL */
    return err;
}






/*
 * FUNCTION: ix_cc_mpls_async_nhlfe_create   */
/**
 * ABSTRACT: Sends message to MPLS CC to create NHLFE
 *
 * DESCRIPTION: This message helper function creates an NHLFE table entry.
 *              See ix_cc_mpls_nhlfe_create() for a detailed description.
 *
 * \param arg_pNhlfe            - pointer to the ix_s_cc_mpls_nhlfe structure, defining an NHLFE
 * \param arg_Callback          - pointer to the client provided callback function
 * \param arg_pUserContext      - pointer to the client defined context that will 
 *                                be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_nhlfe_create(ix_cc_mpls_nhlfe 			 *arg_pNhlfe,
                              ix_cc_mpls_cb_nhlfe_create  arg_Callback,
                              void                       *arg_pUserContext)
{
    ix_error                       err     = IX_SUCCESS;
    ix_cc_mpls_message_ctx        *pIntCtx = NULL;

    if (NULL == arg_Callback) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
	              ("ix_cc_mpls_async_nhlfe_create(): function was called with NULL callback"));

    /* create ix_cc_mpls_message_ctx */
    pIntCtx = ix_ossl_malloc(sizeof(ix_cc_mpls_message_ctx));
    
    if (NULL == pIntCtx) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
	              ("ix_cc_mpls_async_nhlfe_create(): Fail to allocate system memory"));
    
    pIntCtx->pCallBack    = arg_Callback;
    pIntCtx->pUserContext = arg_pUserContext;

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_MPLS_MSG_INPUT,
                                           (ix_cc_msghlp_callback)_ix_cc_mpls_cbi_nhlfe_create,
                                           pIntCtx,
                                           IX_CC_MPLS_MSG_NHLFE_CREATE,
                                           (void*)(arg_pNhlfe),
                                           sizeof(*arg_pNhlfe)),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING, 
                 cleanup_context);

    return IX_SUCCESS;

cleanup_context:

    if(pIntCtx != NULL)
    {
        ix_ossl_free(pIntCtx);
    }/* end if cbContext != NULL */
    return err;
}






/*
 * FUNCTION: ix_cc_mpls_async_nhlfe_delete        */
/**
 * ABSTRACT: Sends message to MPLS CC to delete NHLFE
 *
 * DESCRIPTION: This message helper function deletes an NHLFE entry. 
 *              See ix_cc_mpls_nhlfe_delete() for a detailed description. 
 *
 * \param arg_NhlfeHandle  - handle identifying an NHLFE. 
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_nhlfe_delete(ix_cc_mpls_nhlfe_handle arg_NhlfeHandle,
                              ix_cc_mpls_cb_general   arg_Callback,
                              void                   *arg_pUserContext)
{
    ix_error                       err     = IX_SUCCESS;
    ix_cc_mpls_message_ctx        *pIntCtx = NULL;

    if (NULL == arg_Callback) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
	              ("ix_cc_mpls_async_nhlfe_delete(): function was called with NULL callback"));

    /* create ix_cc_mpls_message_ctx */
    pIntCtx = ix_ossl_malloc(sizeof(ix_cc_mpls_message_ctx));
    
    if (NULL == pIntCtx) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
	              ("ix_cc_mpls_async_nhlfe_delete(): Fail to allocate system memory"));
    
    pIntCtx->pCallBack    = arg_Callback;
    pIntCtx->pUserContext = arg_pUserContext;

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_MPLS_MSG_INPUT,
                                           (ix_cc_msghlp_callback)_ix_cc_mpls_cb_general,
                                           pIntCtx,
                                           IX_CC_MPLS_MSG_NHLFE_DELETE,
                                           (void*)(&arg_NhlfeHandle),
                                           sizeof(arg_NhlfeHandle)),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING, 
                 cleanup_context);

    return IX_SUCCESS;

cleanup_context:

    if(pIntCtx != NULL)
    {
        ix_ossl_free(pIntCtx);
    }/* end if cbContext != NULL */
    return err;
}





/*
 * FUNCTION: ix_cc_mpls_async_nhlfe_query        */
/**
 * ABSTRACT: Sends message to MPLS CC to get NHLFE
 *
 * DESCRIPTION: This message helper function reads a regular NHLFE entry.
 *
 * \param arg_NhlfeHandle  - handle identifying an NHLFE. 
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_nhlfe_query(ix_cc_mpls_nhlfe_handle    arg_NhlfeHandle,
                             ix_cc_mpls_cb_nhlfe_query  arg_Callback,
                              void                     *arg_pUserContext)
{
    ix_error                       err     = IX_SUCCESS;
    ix_cc_mpls_message_ctx        *pIntCtx = NULL;

    if (NULL == arg_Callback) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
	              ("ix_cc_mpls_async_nhlfe_query(): function was called with NULL callback"));

    /* create ix_cc_mpls_message_ctx */
    pIntCtx = ix_ossl_malloc(sizeof(ix_cc_mpls_message_ctx));

    if (NULL == pIntCtx)
	       return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
	              ("ix_cc_mpls_async_nhlfe_query(): Fail to allocate system memory"));

    pIntCtx->pCallBack    = arg_Callback;
    pIntCtx->pUserContext = arg_pUserContext;

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_MPLS_MSG_INPUT,
                                           (ix_cc_msghlp_callback)_ix_cc_mpls_cbi_nhlfe_query,
                                           pIntCtx,
                                           IX_CC_MPLS_MSG_NHLFE_QUERY,
                                           (void*)(&arg_NhlfeHandle),
                                           sizeof(arg_NhlfeHandle)),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING, 
                 cleanup_context);

    return IX_SUCCESS;

cleanup_context:

    if(pIntCtx != NULL)
    {
        ix_ossl_free(pIntCtx);
    }/* end if cbContext != NULL */
    return err;
}




/*
 * FUNCTION: ix_cc_mpls_async_nhlfe_stats_query        */
/**
 * ABSTRACT: Sends message to MPLS CC to get NHLFE statistics
 *
 * DESCRIPTION: This message helper function reads NHLFE statistics. 
 *
 * \param arg_NhlfeHandle  - handle identifying an NHLFE. 
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_nhlfe_stats_query(ix_cc_mpls_nhlfe_handle          arg_NhlfeHandle,
                                   ix_cc_mpls_cb_nhlfe_stats_query  arg_Callback,
                                   void                            *arg_pUserContext)
{
    ix_error                       err     = IX_SUCCESS;
    ix_cc_mpls_message_ctx        *pIntCtx = NULL;

    if (NULL == arg_Callback) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
	              ("ix_cc_mpls_async_nhlfe_stats_query(): function was called with NULL callback"));

    /* create ix_cc_mpls_message_ctx */
    pIntCtx = ix_ossl_malloc(sizeof(ix_cc_mpls_message_ctx));

    if (NULL == pIntCtx)
	       return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
	              ("ix_cc_mpls_async_nhlfe_stats_query(): Fail to allocate system memory"));

    pIntCtx->pCallBack    = arg_Callback;
    pIntCtx->pUserContext = arg_pUserContext;

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_MPLS_MSG_INPUT,
                                           (ix_cc_msghlp_callback)_ix_cc_mpls_cbi_nhlfe_stats_query,
                                           pIntCtx,
                                           IX_CC_MPLS_MSG_NHLFE_STATS_QUERY,
                                           (void*)(&arg_NhlfeHandle),
                                           sizeof(arg_NhlfeHandle)),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING, 
                 cleanup_context);

    return IX_SUCCESS;

cleanup_context:

    if(pIntCtx != NULL)
    {
        ix_ossl_free(pIntCtx);
    }/* end if cbContext != NULL */
    return err;
}





/*
 * FUNCTION: ix_cc_mpls_async_nhlfe_stats_query        */
/**
 * ABSTRACT: Sends message to MPLS CC to delete all NHLFEs
 *
 * DESCRIPTION: This message helper function deletes all NHLFE table entries. 
 *
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_nhlfe_purge(ix_cc_mpls_cb_general   arg_Callback,
                             void                   *arg_pUserContext)
{
    ix_error                       err     = IX_SUCCESS;
    ix_cc_mpls_message_ctx        *pIntCtx = NULL;

    if (NULL == arg_Callback) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
	              ("ix_cc_mpls_async_nhlfe_purge(): function was called with NULL callback"));

    /* create ix_cc_mpls_message_ctx */
    pIntCtx = ix_ossl_malloc(sizeof(ix_cc_mpls_message_ctx));

    if (NULL == pIntCtx)
	       return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
	              ("ix_cc_mpls_async_nhlfe_purge(): Fail to allocate system memory"));

    pIntCtx->pCallBack    = arg_Callback;
    pIntCtx->pUserContext = arg_pUserContext;

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_MPLS_MSG_INPUT,
                                           (ix_cc_msghlp_callback)_ix_cc_mpls_cb_general,
                                           pIntCtx,
                                           IX_CC_MPLS_MSG_NHLFE_PURGE,
                                           (void*)(NULL),
                                           0),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING, 
                 cleanup_context);

    return IX_SUCCESS;

cleanup_context:

    if(pIntCtx != NULL)
    {
        ix_ossl_free(pIntCtx);
    }/* end if cbContext != NULL */
    return err;
}




/*
 * FUNCTION: ix_cc_mpls_async_nhlfe_set_create        */
/**
 * ABSTRACT: Sends message to MPLS CC to create NHLFE SET
 *
 * DESCRIPTION: This message helper function creates an NHLFE table entry 
 *              containing an NHLFE set. See ix_cc_mpls_nhlfe_set_create()
 *              for a detailed description.
 *
 * \param arg_nhlfeSetSize  - number of elements in the NHLFE set
 * \param arg_nhlfeHandle[] - pointer to the array of handles defining the new NHLFE set
 * \param arg_Callback      - pointer to the client provided callback function
 * \param arg_pUserContext  - pointer to the client defined context that will 
 *                            be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_nhlfe_set_create(ix_uint16                   arg_NhlfeSetSize,
                                  ix_cc_mpls_nhlfe_handle     arg_NhlfeSetHandles[],
                                  ix_cc_mpls_cb_nhlfe_create  arg_Callback,
                                  void                       *arg_pUserContext)
{
    ix_error                             err     = IX_SUCCESS;
    ix_cc_mpls_message_ctx              *pIntCtx = NULL;
    ix_cc_mpls_message_nhlfe_set_create  MsgData;

    if (NULL == arg_Callback) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
	              ("ix_cc_mpls_async_nhlfe_set_create(): function was called with NULL callback"));

    /* create ix_cc_mpls_message_ctx */
    pIntCtx = ix_ossl_malloc(sizeof(ix_cc_mpls_message_ctx));

    if (NULL == pIntCtx)
	       return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
	              ("ix_cc_mpls_async_nhlfe_set_create(): Fail to allocate system memory"));

    MsgData.NhlfeSetSize  = arg_NhlfeSetSize;
    ix_ossl_memcpy(&(MsgData.NhlfeSetHandleTable),
                   arg_NhlfeSetHandles,
                   sizeof(ix_uint32)*arg_NhlfeSetSize);

    pIntCtx->pCallBack    = arg_Callback;
    pIntCtx->pUserContext = arg_pUserContext;

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_MPLS_MSG_INPUT,
                                           (ix_cc_msghlp_callback)_ix_cc_mpls_cbi_nhlfe_set_create,
                                           pIntCtx,
                                           IX_CC_MPLS_MSG_NHLFE_SET_CREATE,
                                           (void*)(&MsgData),
                                           sizeof(MsgData)),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING, 
                 cleanup_context);

    return IX_SUCCESS;

cleanup_context:

    if(pIntCtx != NULL)
    {
        ix_ossl_free(pIntCtx);
    }/* end if cbContext != NULL */
    return err;
}





/*
 * FUNCTION: ix_cc_mpls_async_nhlfe_set_delete        */
/**
 * ABSTRACT: Sends message to MPLS CC to delete NHLFE SET
 *
 * DESCRIPTION: This message helper function deletes an NHLFE entry containing an NHLFE set.
 *              See ix_cc_mpls_nhlfe_set_delete() for a detailed description. 
 *
 * \param arg_nhlfeHandle  - handle defining the NHLFE to be deleted
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_nhlfe_set_delete(ix_cc_mpls_nhlfe_handle arg_NhlfeHandle,
                                  ix_cc_mpls_cb_general   arg_Callback,
                                  void                   *arg_pUserContext)
{
    ix_error                             err     = IX_SUCCESS;
    ix_cc_mpls_message_ctx              *pIntCtx = NULL;

    if (NULL == arg_Callback) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
	              ("ix_cc_mpls_async_nhlfe_set_delete(): function was called with NULL callback"));

    /* create ix_cc_mpls_message_ctx */
    pIntCtx = ix_ossl_malloc(sizeof(ix_cc_mpls_message_ctx));

    if (NULL == pIntCtx)
	       return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
	              ("ix_cc_mpls_async_nhlfe_set_delete(): Fail to allocate system memory"));

    pIntCtx->pCallBack    = arg_Callback;
    pIntCtx->pUserContext = arg_pUserContext;

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_MPLS_MSG_INPUT,
                                           (ix_cc_msghlp_callback)_ix_cc_mpls_cb_general,
                                           pIntCtx,
                                           IX_CC_MPLS_MSG_NHLFE_SET_DELETE,
                                           (void*)(&arg_NhlfeHandle),
                                           sizeof(arg_NhlfeHandle)),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING, 
                 cleanup_context);

    return IX_SUCCESS;

cleanup_context:

    if(pIntCtx != NULL)
    {
        ix_ossl_free(pIntCtx);
    }/* end if cbContext != NULL */
    return err;
}






/*
 * FUNCTION: ix_cc_mpls_async_nhlfe_set_modify        */
/**
 * ABSTRACT: Sends message to MPLS CC to modify NHLFE SET
 *
 * DESCRIPTION: This message helper function modifies the NHLFE table entry
 *              containing an NHLFE set. See ix_cc_mpls_nhlfe_set_modify()
 *              for a detailed description.
 *
 * \param arg_nhlfeHandle       - handle identifying an NHLFE entry containing an NHLFE set
 * \param arg_nhlfeSetSize      - number of elements in the new NHLFE set
 * \param arg_nhlfeSetHandles[] - pointer to the array of handles defining the new NHLFEs for the set 
 * \param arg_Callback          - pointer to the client provided callback function
 * \param arg_pUserContext      - pointer to the client defined context that will 
 *                                be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_nhlfe_set_modify(ix_cc_mpls_nhlfe_handle arg_NhlfeHandle,
                                  ix_uint16               arg_NhlfeSetSize,
                                  ix_cc_mpls_nhlfe_handle arg_NhlfeSetHandles[],
                                  ix_cc_mpls_cb_general   arg_Callback,
                                  void                   *arg_pUserContext)
{
    ix_error                             err     = IX_SUCCESS;
    ix_cc_mpls_message_ctx              *pIntCtx = NULL;
    ix_cc_mpls_message_nhlfe_set_modify  MsgData;

    if (NULL == arg_Callback) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
	              ("ix_cc_mpls_async_nhlfe_set_modify(): function was called with NULL callback"));

    /* create ix_cc_mpls_message_ctx */
    pIntCtx = ix_ossl_malloc(sizeof(ix_cc_mpls_message_ctx));

    if (NULL == pIntCtx)
	       return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
	              ("ix_cc_mpls_async_nhlfe_set_modify(): Fail to allocate system memory"));

    MsgData.NhlfeHandle         = arg_NhlfeHandle;
    MsgData.NhlfeSetSize        = arg_NhlfeSetSize;
    ix_ossl_memcpy(&(MsgData.NhlfeSetHandleTable),
                   arg_NhlfeSetHandles,
                   sizeof(ix_uint32)*arg_NhlfeSetSize);

    pIntCtx->pCallBack    = arg_Callback;
    pIntCtx->pUserContext = arg_pUserContext;

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_MPLS_MSG_INPUT,
                                           (ix_cc_msghlp_callback)_ix_cc_mpls_cb_general,
                                           pIntCtx,
                                           IX_CC_MPLS_MSG_NHLFE_SET_MODIFY,
                                           (void*)(&MsgData),
                                           sizeof(MsgData)),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING, 
                 cleanup_context);

    return IX_SUCCESS;

cleanup_context:

    if(pIntCtx != NULL)
    {
        ix_ossl_free(pIntCtx);
    }/* end if cbContext != NULL */
    return err;
}




/*
 * FUNCTION: ix_cc_mpls_async_nhlfe_set_query        */
/**
 * ABSTRACT: Sends message to MPLS CC to get NHLFE SET
 *
 * DESCRIPTION: This message helper function reads an NHLFE entry containing an NHLFE set.
 *              See ix_cc_mpls_nhlfe_set_query() for a detailed description.
 *
 * \param arg_nhlfeHandle       - handle identifying an NHLFE entry containing an NHLFE set
 * \param arg_Callback          - pointer to the client provided callback function
 * \param arg_pUserContext      - pointer to the client defined context that will 
 *                                be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_nhlfe_set_query(ix_cc_mpls_nhlfe_handle        arg_NhlfeHandle,
                                 ix_cc_mpls_cb_nhlfe_set_query 	arg_Callback,
                                 void					       *arg_pUserContext)
{
    ix_error                             err     = IX_SUCCESS;
    ix_cc_mpls_message_ctx              *pIntCtx = NULL;

    if (NULL == arg_Callback) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
	              ("ix_cc_mpls_async_nhlfe_set_query(): function was called with NULL callback"));

    /* create ix_cc_mpls_message_ctx */
    pIntCtx = ix_ossl_malloc(sizeof(ix_cc_mpls_message_ctx));

    if (NULL == pIntCtx)
	       return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
	              ("ix_cc_mpls_async_nhlfe_set_query(): Fail to allocate system memory"));

    pIntCtx->pCallBack    = arg_Callback;
    pIntCtx->pUserContext = arg_pUserContext;

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_MPLS_MSG_INPUT,
                                           (ix_cc_msghlp_callback)_ix_cc_mpls_cbi_nhlfe_set_query,
                                           pIntCtx,
                                           IX_CC_MPLS_MSG_NHLFE_SET_QUERY,
                                           (void*)(&arg_NhlfeHandle),
                                           sizeof(arg_NhlfeHandle)),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING, 
                 cleanup_context);

    return IX_SUCCESS;

cleanup_context:

    if(pIntCtx != NULL)
    {
        ix_ossl_free(pIntCtx);
    }/* end if cbContext != NULL */
    return err;
}




/*
 * FUNCTION: ix_cc_mpls_async_exp2phb_modify        */
/**
 * ABSTRACT: Sends message to MPLS CC to modify a mapping in exp2phb table
 *
 * DESCRIPTION: This message helper function modifies an EXP2PHB table entry. 
 *              Note: all EXP2PHB table entries are created during initialization 
 *              and set to defaults.
 *
 * \param arg_Index        - index identifying an EXP2PHB table entry
 * \param arg_exp2phb      - pointer to the table containing the mappings for one EXP2PHB 
 *                           table entry
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_exp2phb_modify(ix_uint16              arg_Index,
                                ix_cc_mpls_exp2phb     arg_exp2phb[8],
                                ix_cc_mpls_cb_general  arg_Callback,
                                void                  *arg_pUserContext)
{
    ix_error                             err     = IX_SUCCESS;
    ix_cc_mpls_message_ctx              *pIntCtx = NULL;
    ix_cc_mpls_message_exp2phb_modify    MsgData;

    if (NULL == arg_Callback) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
	              ("ix_cc_mpls_async_exp2phb_modify(): function was called with NULL callback"));

    /* create ix_cc_mpls_message_ctx */
    pIntCtx = ix_ossl_malloc(sizeof(ix_cc_mpls_message_ctx));

    if (NULL == pIntCtx)
	       return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
	              ("ix_cc_mpls_async_exp2phb_modify(): Fail to allocate system memory"));

    MsgData.Index        = arg_Index;
    ix_ossl_memcpy(&(MsgData.Exp2PhbTable),
                   arg_exp2phb,
                   sizeof(ix_cc_mpls_exp2phb)*8);

    pIntCtx->pCallBack    = arg_Callback;
    pIntCtx->pUserContext = arg_pUserContext;

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_MPLS_MSG_INPUT,
                                           (ix_cc_msghlp_callback)_ix_cc_mpls_cb_general,
                                           pIntCtx,
                                           IX_CC_MPLS_MSG_EXP2PHB_MODIFY,
                                           (void*)(&MsgData),
                                           sizeof(MsgData)),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING, 
                 cleanup_context);

    return IX_SUCCESS;

cleanup_context:

    if(pIntCtx != NULL)
    {
        ix_ossl_free(pIntCtx);
    }/* end if cbContext != NULL */
    return err;
}





/*
 * FUNCTION: ix_cc_mpls_async_exp2phb_query        */
/**
 * ABSTRACT: Sends message to MPLS CC to get a mapping from exp2phb table
 *
 * DESCRIPTION: This message helper function reads an EXP2PHB table entry. 
 *
 * \param arg_Index        - index identifying an EXP2PHB table entry
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_exp2phb_query(ix_uint16                    arg_Index,
                               ix_cc_mpls_cb_exp2phb_query  arg_Callback,
                               void                        *arg_pUserContext)
{
    ix_error                             err     = IX_SUCCESS;
    ix_cc_mpls_message_ctx              *pIntCtx = NULL;

    if (NULL == arg_Callback) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
	              ("ix_cc_mpls_async_exp2phb_query(): function was called with NULL callback"));

    /* create ix_cc_mpls_message_ctx */
    pIntCtx = ix_ossl_malloc(sizeof(ix_cc_mpls_message_ctx));

    if (NULL == pIntCtx)
	       return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
	              ("ix_cc_mpls_async_exp2phb_query(): Fail to allocate system memory"));

    pIntCtx->pCallBack    = arg_Callback;
    pIntCtx->pUserContext = arg_pUserContext;

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_MPLS_MSG_INPUT,
                                           (ix_cc_msghlp_callback)_ix_cc_mpls_cbi_exp2phb_query,
                                           pIntCtx,
                                           IX_CC_MPLS_MSG_EXP2PHB_QUERY,
                                           (void*)(&arg_Index),
                                           sizeof(arg_Index)),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING, 
                 cleanup_context);

    return IX_SUCCESS;

cleanup_context:

    if(pIntCtx != NULL)
    {
        ix_ossl_free(pIntCtx);
    }/* end if cbContext != NULL */
    return err;
}





/*
 * FUNCTION: ix_cc_mpls_async_phb2exp_modify       */
/**
 * ABSTRACT: Sends message to MPLS CC to modify a mapping in phb2exp table
 *
 * DESCRIPTION: This message helper function modifies a PHB2EXP table entry.
 *              Note: all PHB2EXP table entries are created during 
 *              initialization and set to defaults. 
 *
 * \param arg_Index        - index identifying an EXP2PHB table entry
 * \param arg_phb2exp      - pointer to the table containing the mappings for 
 *                           a PHB2EXP table entry
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_phb2exp_modify(ix_uint16              arg_Index,
                                ix_cc_mpls_phb2exp     arg_phb2exp[MPLS_MAX_PHB_NUM],
                                ix_cc_mpls_cb_general  arg_Callback,
                                void                  *arg_pUserContext)
{
    ix_error                             err     = IX_SUCCESS;
    ix_cc_mpls_message_ctx              *pIntCtx = NULL;
    ix_cc_mpls_message_phb2exp_modify    MsgData;

    if (NULL == arg_Callback) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
	              ("ix_cc_mpls_async_exp2phb_modify(): function was called with NULL callback"));

    /* create ix_cc_mpls_message_ctx */
    pIntCtx = ix_ossl_malloc(sizeof(ix_cc_mpls_message_ctx));

    if (NULL == pIntCtx)
	       return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
	              ("ix_cc_mpls_async_exp2phb_modify(): Fail to allocate system memory"));

    MsgData.Index        = arg_Index;
    ix_ossl_memcpy(&(MsgData.Phb2ExpTable),
                   arg_phb2exp,
                   sizeof(ix_cc_mpls_phb2exp)*8);
    
    pIntCtx->pCallBack    = arg_Callback;
    pIntCtx->pUserContext = arg_pUserContext;

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_MPLS_MSG_INPUT,
                                           (ix_cc_msghlp_callback)_ix_cc_mpls_cb_general,
                                           pIntCtx,
                                           IX_CC_MPLS_MSG_PHB2EXP_MODIFY,
                                           (void*)(&MsgData),
                                           sizeof(MsgData)),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING, 
                 cleanup_context);

    return IX_SUCCESS;

cleanup_context:

    if(pIntCtx != NULL)
    {
        ix_ossl_free(pIntCtx);
    }/* end if cbContext != NULL */
    return err;
}






/*
 * FUNCTION: ix_cc_mpls_async_phb2exp_query       */
/**
 * ABSTRACT: Sends message to MPLS CC to get a mapping from phb2exp table
 *
 * DESCRIPTION: This message helper function reads a PHB2EXP table entry. 
 *
 * \param arg_Index        - index identifying an EXP2PHB table entry
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_phb2exp_query(ix_uint16                    arg_Index,
                               ix_cc_mpls_cb_phb2exp_query  arg_Callback,
                               void                        *arg_pUserContext)
{
    ix_error                             err     = IX_SUCCESS;
    ix_cc_mpls_message_ctx              *pIntCtx = NULL;

    if (NULL == arg_Callback) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
	              ("ix_cc_mpls_async_phb2exp_query(): function was called with NULL callback"));

    /* create ix_cc_mpls_message_ctx */
    pIntCtx = ix_ossl_malloc(sizeof(ix_cc_mpls_message_ctx));

    if (NULL == pIntCtx)
	       return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
	              ("ix_cc_mpls_async_phb2exp_query(): Fail to allocate system memory"));

    pIntCtx->pCallBack    = arg_Callback;
    pIntCtx->pUserContext = arg_pUserContext;

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_MPLS_MSG_INPUT,
                                           (ix_cc_msghlp_callback)_ix_cc_mpls_cbi_phb2exp_query,
                                           pIntCtx,
                                           IX_CC_MPLS_MSG_PHB2EXP_QUERY,
                                           (void*)(&arg_Index),
                                           sizeof(arg_Index)),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING, 
                 cleanup_context);

    return IX_SUCCESS;

cleanup_context:

    if(pIntCtx != NULL)
    {
        ix_ossl_free(pIntCtx);
    }/* end if cbContext != NULL */
    return err;
}





/*
 * FUNCTION: ix_cc_mpls_async_param_query       */
/**
 * ABSTRACT: Sends message to MPLS CC to get MPLS CC parameters
 *
 * DESCRIPTION: This message helper function reads MPLS parameters.
 *
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_param_query(ix_cc_mpls_cb_param_query  arg_Callback,
                             void                      *arg_pUserContext)
{
    ix_uint32                            MsgID   = 1;
    ix_error                             err     = IX_SUCCESS;
    ix_cc_mpls_message_ctx              *pIntCtx = NULL;

    if (NULL == arg_Callback) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
	              ("ix_cc_mpls_async_param_query(): function was called with NULL callback"));

    /* create ix_cc_mpls_message_ctx */
    pIntCtx = ix_ossl_malloc(sizeof(ix_cc_mpls_message_ctx));

    if (NULL == pIntCtx)
	       return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
	              ("ix_cc_mpls_async_param_query(): Fail to allocate system memory"));

    pIntCtx->pCallBack    = arg_Callback;
    pIntCtx->pUserContext = arg_pUserContext;

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_MPLS_MSG_INPUT,
                                           (ix_cc_msghlp_callback)_ix_cc_mpls_cbi_param_query,
                                           pIntCtx,
                                           IX_CC_MPLS_MSG_PARAM_QUERY,
                                           (void*)&MsgID,
                                           sizeof(MsgID)),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING, 
                 cleanup_context);

    return IX_SUCCESS;

cleanup_context:

    if(pIntCtx != NULL)
    {
        ix_ossl_free(pIntCtx);
    }/* end if cbContext != NULL */
    return err;
}






/*
 * FUNCTION: ix_cc_mpls_async_port_stats_query       */
/**
 * ABSTRACT: Sends message to MPLS CC to get MPLS statistics from the given port
 *
 * DESCRIPTION: This message helper function reads MPLS statistics for a given input port. 
 *
 * \param arg_PortIndex    - local port index
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_port_stats_query(ix_uint16                        arg_PortIndex,
                                  ix_cc_mpls_cb_port_stats_query   arg_Callback,
                                  void                            *arg_pUserContext)
{
    ix_error                             err     = IX_SUCCESS;
    ix_cc_mpls_message_ctx              *pIntCtx = NULL;

    if (NULL == arg_Callback) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
	              ("ix_cc_mpls_async_port_stats_query(): function was called with NULL callback"));

    /* create ix_cc_mpls_message_ctx */
    pIntCtx = ix_ossl_malloc(sizeof(ix_cc_mpls_message_ctx));

    if (NULL == pIntCtx)
	       return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
	              ("ix_cc_mpls_async_port_stats_query(): Fail to allocate system memory"));

    pIntCtx->pCallBack    = arg_Callback;
    pIntCtx->pUserContext = arg_pUserContext;

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_MPLS_MSG_INPUT,
                                           (ix_cc_msghlp_callback)_ix_cc_mpls_cbi_port_stats_query,
                                           pIntCtx,
                                           IX_CC_MPLS_MSG_PORT_STATS_QUERY,
                                           (void*)(&arg_PortIndex),
                                           sizeof(arg_PortIndex)),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING, 
                 cleanup_context);

    return IX_SUCCESS;

cleanup_context:

    if(pIntCtx != NULL)
    {
        ix_ossl_free(pIntCtx);
    }/* end if cbContext != NULL */
    return err;
}





/*
 * FUNCTION: ix_cc_mpls_async_get_mpls_context       */
/**
 * ABSTRACT: This function reads MPLS internal context structure 
 *
 * DESCRIPTION: This function reads MPLS internal context structure 
 *              It is used during Unit Tests.
 *
 * \param arg_Callback     - pointer to the client provided callback function
 * \param arg_pUserContext - pointer to the client defined context that will 
 *                           be passed to the callback function
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_SEND_FAIL - failure from Message Support Library
 */
ix_error
ix_cc_mpls_async_get_mpls_context(ix_cc_mpls_cb_get_mpls_context   arg_Callback,
                                  void                            *arg_pUserContext)
{
    ix_error                             err     = IX_SUCCESS;
    ix_cc_mpls_message_ctx              *pIntCtx = NULL;

    if (NULL == arg_Callback) 
	       return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
	              ("ix_cc_mpls_async_get_mpls_context(): function was called with NULL callback"));

    /* create ix_cc_mpls_message_ctx */
    pIntCtx = ix_ossl_malloc(sizeof(ix_cc_mpls_message_ctx));

    if (NULL == pIntCtx)
	       return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
	              ("ix_cc_mpls_async_get_mpls_context(): Fail to allocate system memory"));

    pIntCtx->pCallBack    = arg_Callback;
    pIntCtx->pUserContext = arg_pUserContext;

    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_MPLS_MSG_INPUT,
                                           (ix_cc_msghlp_callback)_ix_cc_mpls_cbi_get_mpls_context,
                                           pIntCtx,
                                           IX_CC_MPLS_MSG_GET_MPLS_CONTEXT,
                                           pIntCtx,
                                           sizeof(ix_cc_mpls_message_ctx)),
                 err,
                 IX_CC_ERROR_SEND_FAIL,
                 IX_ERROR_LEVEL_WARNING, 
                 cleanup_context);

    return IX_SUCCESS;

cleanup_context:

    if(pIntCtx != NULL)
    {
        ix_ossl_free(pIntCtx);
    }/* end if cbContext != NULL */
    return err;
}




