/*
 * ============================================================================
 * COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *                                                                          */
/** \file      ix_cc_mpls_nhlfe.c
 *
 *  \brief     The file contains CCI API function exposed by
 *             MPLS Core Component and used to manage NHLFE table.
 *
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_mpls_nhlfe.c,v 1.13.2.1 2004/03/12 10:40:16 wneubaue Exp $";


#include <ix_types.h>
#include <ix_error.h>
#include <definitions.h>

#include <ix_ossl.h>
#include <ix_rm.h>

#if !defined(IX_EXCLUDE_CCI)
#include <ix_cci.h>
#endif 

#include <ix_cc_error.h>
#include <ix_cc.h>

#include <bindings.h>

#include <cc/ix_cc_rtmv4.h>
#include <cc/ix_cc_reg_util.h>

/* Include MPLS-related files */
#include "../../include/mpls_defs.h"
#include "cc/ix_cc_mpls.h"
#include "cc/internal/ix_cc_mpls_lsp.h"
#include "cc/internal/ix_cc_mpls_nhlfe.h"
#include "cc/internal/ix_cc_mpls_internal.h"




PRIVATE ix_error
_ix_cc_mpls_nhlfe_allocate_memory(ix_cc_mpls_context *arg_pContext);


PRIVATE ix_error
_ix_cc_mpls_nhlfe_add_slot(ix_cc_mpls_nhlfe_handle   arg_outSegId,

                           ix_cc_mpls_context       *arg_pNhlfeContext);

PRIVATE ix_error
_ix_cc_mpls_nhlfe_verify_input(ix_cc_mpls_nhlfe arg_Nhlfe);


/*
 * FUNCTION: ix_cc_mpls_nhlfe_init()  */
/**
 * ABSTRACT: function initializes the NHLFE part of MPLS CC
 *
 * DESCRIPTION: This function creates and initializes the NHLFE table. It should
 *              be called and returned successfully before any other function in
 *              this library can be called. It returns the base address of the
 *              NHLFE table.
 *
 * \param  arg_pContext   - pointer to the MPLS CC context
 *                          that will be passed to the core component API function
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n           IX_CC_ERROR_NULL            - invalid pointer
 * \n           IX_CC_ERROR_DUPLICATE_ENTRY - NHLFE table already exist
 * \n           IX_CC_ERROR_OOM_SYSTEM      - memory allocation failure
 * \n           IX_CC_ERROR_OOM_SRAM        - memory allocation failure
 */
ix_error
ix_cc_mpls_nhlfe_init(void* arg_pContext)
{

    ix_error err = IX_SUCCESS;
    ix_cc_mpls_context      *pContext = arg_pContext;
    ix_cc_mpls_nhlfe_handle  NhlfeHandle;
    ix_uint32 i;

    /* Check the input arguments */
    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("ix_cc_mpls_nhlfe_init(): got NULL MPLS context"));

    /* Create mutex for critical section. */
    IX_ERROR_C(ix_ossl_mutex_init(OS_MUTEX_UNLOCK, &(pContext->nhlfeMutexId)),err);
    if (err != IX_SUCCESS)
    {
        ix_ossl_free(pContext);
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_OSSL_MUTEX_INIT,
               ("ix_cc_mpls_nhlfe_init(): can't initialize LSP mutex"));
    }

    /* Allocate needed memory */
    IX_ERROR_C(_ix_cc_mpls_nhlfe_allocate_memory(arg_pContext), err);
    if (err != IX_SUCCESS)
    {
        /* Free all memories allocated earlier */
        _ix_cc_mpls_nhlfe_free_resources(arg_pContext);
        return(err);
    }


    /* Create CC NHLFE Entries freelist */
    for (i=1; i<=pContext->maxNhlfeEntries; i++)
    {
        NhlfeHandle.bladeId = pContext->localBladeId;
        NhlfeHandle.outSegId = i;
        _ix_cc_mpls_nhlfe_add_slot(NhlfeHandle, pContext);
    }
    
    return err;
} /* ix_cc_mpls_nhlfe_init */




/*
 * FUNCTION: ix_cc_mpls_nhlfe_create()  */
/**
 * ABSTRACT: This function creates NHLFE & CC_NHLFE entries.
 *
 * DESCRIPTION: This function creates a NHLFE and corresponding CC_NHLFE 
 *              entry. On creation, the CC_NHLFE entry refCount 
 *              field is set to 0. It is incremented by each LSP creation 
 *              that specifies this NHLFE as its outgoing segment, 
 *              and decremented by such an LSP deletion.
 *
 * \param  arg_Nhlfe        - pointer to the ix_s_cc_mpls_nhlfe structure,
 *                            defining an NHLFE
 * \param  arg_pNhlfeHandle - pointer to the handle identifying the created NHLFE.  
 * \param  arg_pContext     - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n           IX_CC_ERROR_NULL - invalid pointer
 * \n           IX_CC_ERROR_FULL - no place in the NHLFE table
 * \n           IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK - mutex lock failed
 */
ix_error
ix_cc_mpls_nhlfe_create(ix_cc_mpls_nhlfe         arg_Nhlfe,
                        ix_cc_mpls_nhlfe_handle *arg_pNhlfeHandle,
                        void                    *arg_pContext)
{
    ix_cc_mpls_context           * pContext           = arg_pContext;
    ix_cc_mpls_cc_nhlfe_free_slot* pFreeSlotListOldHead;
    ix_cc_mpls_cc_nhlfe          * pCcNhlfeNewEntry;
    ix_cc_mpls_nhlfe_internal    * pNhlfeNewEntry;
    ix_error                       err;
    
    /* verify input arguments */
    if (arg_pNhlfeHandle == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_nhlfe_create(): Null pointer to NHLFE structure"));

    
    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_nhlfe_create(): Null MPLS context"));


    /* verify parameters in NHLFE structure */
    err = _ix_cc_mpls_nhlfe_verify_input(arg_Nhlfe);
    if (err != IX_SUCCESS)
        return err;

    /* check if there is a free slot in CC NHLFE table */
    if (pContext->pFreeNhlfeListTail == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_FULL, 
                                ("ix_cc_mpls_nhlfe_create(): no place in the CC NHLFE/NHLFE tables"));


    IX_ERROR_C(ix_ossl_mutex_lock(pContext->nhlfeMutexId, IX_OSSL_WAIT_FOREVER), err);
    
    if (err != IX_SUCCESS)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK,
                                ("ix_cc_mpls_nhlfe_create(): can't lock mutex"));
    
    /* get free CC NHLFE table's entry */
    arg_pNhlfeHandle->outSegId = pContext->pFreeNhlfeListHead->NhlfeHandle.outSegId;
    arg_pNhlfeHandle->bladeId  = pContext->localBladeId;
    pCcNhlfeNewEntry = &pContext->pCcNhlfeBaseAddr[arg_pNhlfeHandle->outSegId];
    ix_ossl_memset(pCcNhlfeNewEntry, 0, sizeof(ix_cc_mpls_cc_nhlfe));
    
    /* get new NHLFE table's entry */
    pNhlfeNewEntry = &pContext->pNhlfeBaseAddr[arg_pNhlfeHandle->outSegId];
    ix_ossl_memset(pNhlfeNewEntry, 0, sizeof(ix_cc_mpls_nhlfe_internal));
    
    /* remove entry from list */
    pFreeSlotListOldHead = pContext->pFreeNhlfeListHead;
    pContext->pFreeNhlfeListHead = pContext->pFreeNhlfeListHead->pNextFreeSlot;
    if (pContext->pFreeNhlfeListHead == NULL)
    {
        pContext->pFreeNhlfeListTail = NULL;
    }
    ix_ossl_free(pFreeSlotListOldHead);

    /* Copy data to CC NHLFE Table entry */
    pCcNhlfeNewEntry->flags    = MPLS_CC_NHLFE_FLAG_NHLFE;
    pCcNhlfeNewEntry->refCount = 0;
    pCcNhlfeNewEntry->lspList  = NULL;

    /* Copy data to NHLFE Table entry */
    pNhlfeNewEntry->flags         = MPLS_FLAG_VALID_BIT;
    pNhlfeNewEntry->pushCount     = (arg_Nhlfe.pushCount      & 0x07);
    pNhlfeNewEntry->l3Protocol    = (arg_Nhlfe.l3Protocol     & 0x07);
    pNhlfeNewEntry->nextHopId     = (arg_Nhlfe.nextHopId      & 0x0000FFFF);
    pNhlfeNewEntry->maxLabPktSize = 0;
    pNhlfeNewEntry->outPortId     =  arg_Nhlfe.outPortId;
    pNhlfeNewEntry->topLabel      =  arg_Nhlfe.topLabel.label;
    pNhlfeNewEntry->outPortType   = (arg_Nhlfe.outPortType    & 0x0F);
    pNhlfeNewEntry->ttl           = 0;
    pNhlfeNewEntry->label1        =  arg_Nhlfe.label1.label;
    pNhlfeNewEntry->label2        =  arg_Nhlfe.label2.label;
    pNhlfeNewEntry->label3        =  arg_Nhlfe.label3.label;
    pNhlfeNewEntry->bladeId       = (arg_Nhlfe.bladeId        & 0x00FF);
    pNhlfeNewEntry->flowIdHigh    = 0;
    pNhlfeNewEntry->flowIdLow     = 0;
    pNhlfeNewEntry->exp2phb       = 0;
    pNhlfeNewEntry->phb2exp       = 0;

#if defined (IX_MPLS_DEBUG)
    ix_ossl_message_log("ix_cc_mpls_nhlfe_create(): Creating Nhlfe at 0x%x    handle = 0x%x\n",
                        pNhlfeNewEntry, *arg_pNhlfeHandle);
#endif

    IX_ERROR_CRT(ix_ossl_mutex_unlock(pContext->nhlfeMutexId),
                 IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK, 
                 IX_ERROR_LEVEL_REMOTE);
        
    return err;
} /* ix_cc_mpls_nhlfe_create */






/*
 * FUNCTION: ix_cc_mpls_nhlfe_delete()  */
/**
 * ABSTRACT: This function deletes an NHLFE.
 *
 * DESCRIPTION: This function deletes an NHLFE and corresponding 
 *              CC_NHLFE entries if they are not referenced by 
 *              any LSP definition (that is, the CC_NHLFE entry 
 *              refCount field is equal to 0). This function cannot 
 *              be used for deleting NHLFEs containing set descriptions;
 *              for such entries use ix_cc_mpls_nhlfe_set_delete().
 *
 * \param  arg_NhlfeHandle - handle identifying an NHLFE. 
 * \param  arg_pContext    - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE does not exist
 * \n         IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET - this is an NHLFE set entry
 * \n         IX_CC_MPLS_ERROR_ENTRY_USED      - entry referenced by an LSP or NHLFE SET
 * \n         IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK - mutex lock failed
 */
ix_error
ix_cc_mpls_nhlfe_delete(ix_cc_mpls_nhlfe_handle  arg_NhlfeHandle,
                        void                    *arg_pContext)
{
    ix_cc_mpls_cc_nhlfe *pCcNhlfe;
    ix_cc_mpls_context  *pContext = arg_pContext;
    ix_error             err;
    
    /* Verify the input arguments */
    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_nhlfe_delete(): Null MPLS context argument"));

    if (pContext->pCcNhlfeBaseAddr == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_nhlfe_delete(): Null CC NHLFE table address"));

    if (arg_NhlfeHandle.bladeId != pContext->localBladeId)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE, 
                                ("ix_cc_mpls_nhlfe_delete(): wrong bladeId"));

    IX_ERROR_C(ix_ossl_mutex_lock(pContext->nhlfeMutexId, IX_OSSL_WAIT_FOREVER), err);

    if (err != IX_SUCCESS)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK,
                                ("ix_cc_mpls_nhlfe_delete(): can't lock mutex"));

    
    /* add free CC LSP table's entry */
    /* Check whether this is a valid entry */
    pCcNhlfe = &(pContext->pCcNhlfeBaseAddr[arg_NhlfeHandle.outSegId]);
    switch (pCcNhlfe->flags)
    {
        /* Valid entry */
    case MPLS_CC_NHLFE_FLAG_NHLFE:
        break;

        /* Check whether this is a valid entry */
    case MPLS_CC_NHLFE_FLAG_FREE:
        err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                               ("ix_cc_mpls_nhlfe_delete(): invalid NHLFE entry (empty entry)"));
        goto err_exit;

        /* Check whether this is not a set entry */
    case MPLS_CC_NHLFE_FLAG_NHLFE_SET:
        err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET,
                               ("ix_cc_mpls_nhlfe_delete(): invalid NHLFE entry (set entry)"));
        goto err_exit;

        /* Check whether this is not a NHLFE entry used in a set */
    case MPLS_CC_NHLFE_FLAG_NHLFE_USED_IN_SET:
        err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_ENTRY_USED,
                               ("ix_cc_mpls_nhlfe_delete(): invalid NHLFE entry (entry used in set)"));
        goto err_exit;
        
    default:
        err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                               ("ix_cc_mpls_nhlfe_delete(): invalid NHLFE entry (unspecified entry type)"));
        goto err_exit;
    }
    
    /* Check if CC NHLFE entry is referenced by LSPs */
    if (pCcNhlfe->refCount > 0)
    {
        err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_ENTRY_USED,
                               ("ix_cc_mpls_nhlfe_delete(): invalid NHLFE entry (already referenced by LSP)"));
        goto err_exit;
    }

    /* Add CC NHLFE entry back to freeSlotList */
    pCcNhlfe->flags = MPLS_CC_NHLFE_FLAG_FREE;
    _ix_cc_mpls_nhlfe_add_slot(arg_NhlfeHandle, arg_pContext);

err_exit:
    IX_ERROR_CRT(ix_ossl_mutex_unlock(pContext->nhlfeMutexId),
                 IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK, 
                 IX_ERROR_LEVEL_REMOTE);
    
    return err;
} /* ix_cc_mpls_nhlfe_delete */





/*
 * FUNCTION: ix_cc_mpls_nhlfe_query()  */
/**
 * ABSTRACT: This function gets an NHLFE.
 *
 * DESCRIPTION: This message helper function reads a regular (non-set) 
 *              NHLFE entry. This function cannot be used for reading 
 *              NHLFEs containing set descriptions; for such entries 
 *              use ix_cc_mpls_nhlfe_set_query().
 *
 * \param  arg_NhlfeHandle - handle identifying an NHLFE. 
 * \param  arg_pNhlfe      - pointer to the buffer for storing 
 *                           the NHLFE information returned by the function.  
 * \param  arg_pContext    - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE does not exist
 * \n         IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET - this is an NHLFE set entry
 * \n         IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK - mutex lock failed
 */
ix_error
ix_cc_mpls_nhlfe_query(ix_cc_mpls_nhlfe_handle  arg_NhlfeHandle,
                       ix_cc_mpls_nhlfe        *arg_pNhlfe,
                       void                    *arg_pContext)
{
    ix_cc_mpls_cc_nhlfe       *pCcNhlfeEntry;
    ix_cc_mpls_nhlfe_internal *pNhlfeEntry;
    ix_cc_mpls_context        *pContext = arg_pContext;
    ix_error                   err;
    
    /* Verify the input arguments */
    if (arg_pNhlfe == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_nhlfe_query(): Null MPLS pointer to NHLFE handle"));

    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_nhlfe_query(): Null MPLS context argument"));

    if (pContext->pCcNhlfeBaseAddr == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_nhlfe_query(): Null CC NHLFE table address"));

    if (arg_NhlfeHandle.bladeId != pContext->localBladeId)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE, 
                                ("ix_cc_mpls_nhlfe_query(): wrong bladeId"));

    IX_ERROR_C(ix_ossl_mutex_lock(pContext->nhlfeMutexId, IX_OSSL_WAIT_FOREVER), err);

    if (err != IX_SUCCESS)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK,
                                ("ix_cc_mpls_nhlfe_query(): can't lock mutex"));

    /* Check whether this is a valid entry */
    pCcNhlfeEntry = &pContext->pCcNhlfeBaseAddr[arg_NhlfeHandle.outSegId];
    switch (pCcNhlfeEntry->flags)
    {
        /* Valid entry */
    case MPLS_CC_NHLFE_FLAG_NHLFE:
    case MPLS_CC_NHLFE_FLAG_NHLFE_USED_IN_SET:
        break;

        /* Check whether this is a valid entry */
    case MPLS_CC_NHLFE_FLAG_FREE:
        err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                               ("ix_cc_mpls_nhlfe_query(): invalid NHLFE entry (empty entry)"));
        goto err_exit;

        /* Check whether this is not a set entry */
    case MPLS_CC_NHLFE_FLAG_NHLFE_SET:
        err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET,
                               ("ix_cc_mpls_nhlfe_query(): invalid NHLFE entry (set entry)"));
        goto err_exit;
        
    default:
        err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                               ("ix_cc_mpls_nhlfe_query(): invalid NHLFE entry (unspecified entry type)"));
        goto err_exit;
    }
    
    pNhlfeEntry = &pContext->pNhlfeBaseAddr[arg_NhlfeHandle.outSegId];
    arg_pNhlfe->pushCount      = (ix_uint8)pNhlfeEntry->pushCount;
    arg_pNhlfe->l3Protocol     = (ix_uint8)pNhlfeEntry->l3Protocol;
    arg_pNhlfe->nextHopId      = pNhlfeEntry->nextHopId;
    arg_pNhlfe->outPortId      = (ix_uint16)pNhlfeEntry->outPortId;
    arg_pNhlfe->topLabel.label = pNhlfeEntry->topLabel;
    arg_pNhlfe->outPortType    = (ix_uint8)pNhlfeEntry->outPortType;
    arg_pNhlfe->label1.label   = pNhlfeEntry->label1;
    arg_pNhlfe->label2.label   = pNhlfeEntry->label2;
    arg_pNhlfe->label3.label   = pNhlfeEntry->label3;
    arg_pNhlfe->bladeId        = (ix_uint16)pNhlfeEntry->bladeId;
    
err_exit:
    IX_ERROR_CRT(ix_ossl_mutex_unlock(pContext->nhlfeMutexId),
                 IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK, 
                 IX_ERROR_LEVEL_REMOTE);

    return err;
} /* ix_cc_mpls_nhlfe_query */






/*
 * FUNCTION: ix_cc_mpls_nhlfe_stats_query()  */
/**
 * ABSTRACT: This function gets an NHLFE statistics.
 *
 * DESCRIPTION: This function reads NHLFE statistics. 
 *              This function cannot be used for reading NHLFEs 
 *              containing set descriptions; statistics are valid 
 *              only for regular NHLFEs.
 *
 * \param  arg_NhlfeHandle      - handle identifying an NHLFE. 
 * \param  arg_pNhlfeStatistics - pointer to the buffer for storing the 
 *                                NHLFE statistics returned by the function.  
 * \param  arg_pContext         - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE does not exist
 * \n         IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET - this is an NHLFE set entry
 * \n         IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK - mutex lock failed
 */
ix_error
ix_cc_mpls_nhlfe_stats_query(ix_cc_mpls_nhlfe_handle  arg_NhlfeHandle,
                             ix_cc_mpls_nhlfe_stats  *arg_pNhlfeStatistics,
                             void                    *arg_pContext)
{
    ix_cc_mpls_cc_nhlfe               *pCcNhlfeEntry;
    ix_cc_mpls_context                *pContext  = arg_pContext;
    ix_error                           err;

#if defined(MPLS_COUNTERS_FEATURE)
    ix_uint32                          idx;
    ix_counter_64bit_handle           *pFtn64BitCountersHdls;
#endif

    /* Verify the input arguments */
    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_nhlfe_stats_query(): Null MPLS context argument"));

    if (pContext->pCcNhlfeBaseAddr == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_nhlfe_stats_query(): Null CC NHLFE table address"));

    if (arg_NhlfeHandle.bladeId != pContext->localBladeId)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE, 
                                ("ix_cc_mpls_nhlfe_stats_query(): wrong bladeId"));

    if (arg_pNhlfeStatistics == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_nhlfe_stats_query(): Null MPLS statistics pointer"));

    IX_ERROR_C(ix_ossl_mutex_lock(pContext->nhlfeMutexId, IX_OSSL_WAIT_FOREVER), err);

    if (err != IX_SUCCESS)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK,
                                ("ix_cc_mpls_nhlfe_stats_query(): can't lock mutex"));

    /* Check whether this is a valid entry */
    pCcNhlfeEntry = &pContext->pCcNhlfeBaseAddr[arg_NhlfeHandle.outSegId];
    switch (pCcNhlfeEntry->flags)
    {
        /* Valid entry */
    case MPLS_CC_NHLFE_FLAG_NHLFE:
        break;

        /* Check whether this is a valid entry */
    case MPLS_CC_NHLFE_FLAG_FREE:
        err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                               ("ix_cc_mpls_nhlfe_stats_query(): invalid NHLFE entry (empty entry)"));
        goto err_exit;

        /* Check whether this is not a set entry */
    case MPLS_CC_NHLFE_FLAG_NHLFE_SET:
        err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET,
                               ("ix_cc_mpls_nhlfe_stats_query(): invalid NHLFE entry (set entry)"));
        goto err_exit;

        /* Check whether this is not a NHLFE entry used in a set */
    case MPLS_CC_NHLFE_FLAG_NHLFE_USED_IN_SET:
        err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET,
                               ("ix_cc_mpls_nhlfe_stats_query(): invalid NHLFE entry (entry used in set)"));
        goto err_exit;
        
    default:
        err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                               ("ix_cc_mpls_nhlfe_stats_query(): invalid NHLFE entry (unspecified entry type)"));
        goto err_exit;
    }

#if defined(MPLS_COUNTERS_FEATURE)

    pFtn64BitCountersHdls = (ix_counter_64bit_handle*)((ix_uint8*)pContext->pFtn64BitCountersHdls + 
                            arg_NhlfeHandle.outSegId * sizeof(ix_counter_64bit_handle) * MPLS_NHLFE_64BIT_COUNTERS_NUM);
    
    for(idx = 0; idx < MPLS_NHLFE_64BIT_COUNTERS_NUM; idx ++)
    {
        err = ix_rm_counter_64bit_get_value(*(pFtn64BitCountersHdls + idx), (ix_uint64*)arg_pNhlfeStatistics + idx);
        if (err != IX_SUCCESS)
        {
            err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE, 
                                   ("ix_cc_mpls_lsp_stats_query(): Invalid counter's handle"));
            break;
        }
    }
#endif

err_exit:
    IX_ERROR_CRT(ix_ossl_mutex_unlock(pContext->nhlfeMutexId),
                 IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK, 
                 IX_ERROR_LEVEL_REMOTE);

    return err;
} /* ix_cc_mpls_nhlfe_stats_query */




/*
 * FUNCTION: ix_cc_mpls_nhlfe_purge()  */
/**
 * ABSTRACT: This function deletes all NHLFE table entries.
 *
 * DESCRIPTION: This function deletes all NHLFE table entries.
 *              If any of the entries has a non-zero refCount field,
 *              no entry is deleted. This function can be used for
 *              clearing the NHLFE table before any NHLFE 
 *              was assigned to an LSP.
 *
 * \param  arg_pContext         - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_MPLS_ERROR_ENTRY_USED      - an NHLFE is referenced by an LSP
 * \n         IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK - mutex lock failed
 */
ix_error
ix_cc_mpls_nhlfe_purge(void* arg_pContext)
{

    ix_cc_mpls_cc_nhlfe     *pCcNhlfe;
    ix_cc_mpls_nhlfe_handle  NhlfeHandle;
    ix_cc_mpls_context      *pContext = arg_pContext;
    ix_uint32                i;
    ix_error                 err;
    
    /* Verify the input arguments */
    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_nhlfe_purge(): Null MPLS context argument"));
    }
    if (pContext->pCcNhlfeBaseAddr == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_nhlfe_purge(): Null CC NHLFE table address"));
    }

    IX_ERROR_C(ix_ossl_mutex_lock(pContext->nhlfeMutexId, IX_OSSL_WAIT_FOREVER), err);

    if (err != IX_SUCCESS)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK,
                                ("ix_cc_mpls_nhlfe_purge(): can't lock mutex"));

    /* Check if all entries have refCount field set to zero */
    for (i=1; i<=pContext->maxNhlfeEntries; i++)
    {
        pCcNhlfe = &pContext->pCcNhlfeBaseAddr[i];
        if (pCcNhlfe->refCount > 0)
        {
            err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_ENTRY_USED,
                                   ("ix_cc_mpls_nhlfe_purge(): NHLFE already referenced by LSP"));
            goto err_exit;
        }
    }

    /* No NHLFE has been assigned to LSP - proceed with purging */
    for (i=1; i<=pContext->maxNhlfeEntries; i++)
    {
        pCcNhlfe = &pContext->pCcNhlfeBaseAddr[i];
        /* Add CC NHLFE entry back to freeSlotList */
        NhlfeHandle.bladeId = pContext->localBladeId;
        NhlfeHandle.outSegId = i;
        if (pCcNhlfe->flags != MPLS_CC_NHLFE_FLAG_FREE)
        {
            if (err = _ix_cc_mpls_nhlfe_add_slot(NhlfeHandle, arg_pContext) != IX_SUCCESS)
                goto err_exit;
        }
 
        pCcNhlfe->flags = MPLS_CC_NHLFE_FLAG_FREE;
    }

err_exit:
    IX_ERROR_CRT(ix_ossl_mutex_unlock(pContext->nhlfeMutexId),
                 IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK, 
                 IX_ERROR_LEVEL_REMOTE);

    return err;
} /* ix_cc_mpls_nhlfe_purge */






/*
 * FUNCTION: ix_cc_mpls_nhlfe_set_create()  */
/**
 * ABSTRACT: Creates an NHLFE table entry containing NHLFE set 
 *
 * DESCRIPTION: Creates an NHLFE table entry containing NHLFE set comprising up to
 *              four regular NHLFEs - the NHLFE handles specyfying the set cannot
 *              point to other sets.
 *
 * \param  arg_nhlfeSetSize      - number of elements in the NHLFE set
 * \param  arg_nhlfeSetHandles[] - pointer to the array of handles defining 
 *                                 the new NHLFE set 
 * \param  arg_NhlfeHandle       - pointer to the handle identifying 
 *                                 the created NHLFE.  
 * \param  arg_pContext          - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_RANGE                - arg_NhlfeSize is beyond 
 *                                               the range from (1 to 4)
 * \n         IX_CC_ERROR_FULL                 - NHLFE table is full
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - one of the set NHLFEs does not exist
 * \n         IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET - one of the set NHLFEs is an NHLFE set entry
 * \n         IX_CC_MPLS_ERROR_INVALID_HANDLE  - invalid NHLFE parameters
 * \n         IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK - mutex lock failed
 */
ix_error
ix_cc_mpls_nhlfe_set_create(ix_uint16                arg_nhlfeSetSize,
                            ix_cc_mpls_nhlfe_handle  arg_nhlfeSetHandles[],
                            ix_cc_mpls_nhlfe_handle *arg_NhlfeHandle,
                            void                    *arg_pContext)
{
    ix_uint32 i;
    ix_cc_mpls_cc_nhlfe           * pCcNhlfeNewEntry;
    ix_cc_mpls_cc_nhlfe           * pCcNhlfeEntry;
    ix_cc_mpls_nhlfe_set_internal * pNhlfeSetNewEntry;
    ix_cc_mpls_cc_nhlfe_free_slot * pFreeSlotListOldHead;
    ix_cc_mpls_context            *pContext = arg_pContext;
    ix_error                       err;
    
    if (arg_NhlfeHandle == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_nhlfe_set_create(): Null pointer to NHLFE handle"));
    }
    if (arg_pContext == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_nhlfe_set_create(): Null MPLS context"));
    }
    if (arg_nhlfeSetSize < MPLS_NHLFE_MIN_SET_SIZE  ||  arg_nhlfeSetSize > MPLS_NHLFE_MAX_SET_SIZE)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_RANGE, 
                                ("ix_cc_mpls_nhlfe_set_create(): improper NHLFE set size"));
    }

    IX_ERROR_C(ix_ossl_mutex_lock(pContext->nhlfeMutexId, IX_OSSL_WAIT_FOREVER), err);
    
    if (err != IX_SUCCESS)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK,
                                ("ix_cc_mpls_nhlfe_set_create(): can't lock mutex"));

    for (i=0; i<arg_nhlfeSetSize; i++)
    {
        if (arg_nhlfeSetHandles[i].bladeId != pContext->localBladeId)
        {
            err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE,
                                   ("ix_cc_mpls_nhlfe_set_create(): invalid NHLFE handle (invalid bladeId)"));
            goto err_exit;
        }

        pCcNhlfeEntry = &(pContext->pCcNhlfeBaseAddr[arg_nhlfeSetHandles[i].outSegId]);
        switch (pCcNhlfeEntry->flags)
        {
        case MPLS_CC_NHLFE_FLAG_NHLFE:
            break;

            /* NHLFE entry is empty */
        case MPLS_CC_NHLFE_FLAG_FREE:
            err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                                   ("ix_cc_mpls_nhlfe_set_create(): invalid NHLFE entry (empty entry)"));
            goto err_exit;

            /* entry already contains NHLFE SET */
        case MPLS_CC_NHLFE_FLAG_NHLFE_SET:
        case MPLS_CC_NHLFE_FLAG_NHLFE_USED_IN_SET:
            err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET,
                                   ("ix_cc_mpls_nhlfe_set_create(): invalid NHLFE entry (set entry)"));
            goto err_exit;
            
        default:
            err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                                   ("ix_cc_mpls_nhlfe_set_create(): invalid NHLFE entry (unspecified entry type)"));
            goto err_exit;
        }
    }

    /* check if there is a free slot in CC NHLFE table */
    if (pContext->pFreeNhlfeListTail == NULL)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_FULL,
                               ("ix_cc_mpls_nhlfe_set_create(): full CC NHLFE table"));
        goto err_exit;
    }

    /* get free CC NHLFE table's entry */
    arg_NhlfeHandle->outSegId = pContext->pFreeNhlfeListHead->NhlfeHandle.outSegId;
    arg_NhlfeHandle->bladeId  = pContext->localBladeId;
    pCcNhlfeNewEntry = &pContext->pCcNhlfeBaseAddr[pContext->pFreeNhlfeListHead->NhlfeHandle.outSegId];
    ix_ossl_memset(pCcNhlfeNewEntry, 0, sizeof(ix_cc_mpls_phb2exp_internal));
    
    /* get new NHLFE table's entry */
    pNhlfeSetNewEntry = (ix_cc_mpls_nhlfe_set_internal*) &pContext->pNhlfeBaseAddr[arg_NhlfeHandle->outSegId];
    ix_ossl_memset(pNhlfeSetNewEntry, 0, sizeof(ix_cc_mpls_nhlfe_set_internal));
    
    /* remove entry from list */
    pFreeSlotListOldHead = pContext->pFreeNhlfeListHead;
    pContext->pFreeNhlfeListHead = pContext->pFreeNhlfeListHead->pNextFreeSlot;
    if (pContext->pFreeNhlfeListHead == NULL)
    {
        pContext->pFreeNhlfeListTail = NULL;
    }
    ix_ossl_free(pFreeSlotListOldHead);

    /* Set flags in CC NHLFE entries */
    pCcNhlfeNewEntry->flags = MPLS_CC_NHLFE_FLAG_NHLFE_SET;
    pCcNhlfeNewEntry->refCount = 0;
    for (i=0; i<arg_nhlfeSetSize; i++)
    {
        pCcNhlfeEntry = &(pContext->pCcNhlfeBaseAddr[arg_nhlfeSetHandles[i].outSegId]);
        pCcNhlfeEntry->flags = MPLS_CC_NHLFE_FLAG_NHLFE_USED_IN_SET;
    }

    /* Fill the NHLFE Table's set entry */
    pNhlfeSetNewEntry->flags = MPLS_FLAG_VALID_BIT | MPLS_FLAG_SET_BIT;
    pNhlfeSetNewEntry->exp2phb = 0;
    pNhlfeSetNewEntry->setSize = arg_nhlfeSetSize;
    pNhlfeSetNewEntry->phbMask1 = 0;
    pNhlfeSetNewEntry->phbMask2 = 0;
    pNhlfeSetNewEntry->phbMask3 = 0;
    pNhlfeSetNewEntry->phbMask4 = 0;
    switch(arg_nhlfeSetSize)
    {
    case 4:
        pNhlfeSetNewEntry->outSegId4Hi = (arg_nhlfeSetHandles[3].outSegId & 0xF0000) >> 16;
        pNhlfeSetNewEntry->outSegId4Lo =  arg_nhlfeSetHandles[3].outSegId & 0x0FFFF;
    case 3:
        pNhlfeSetNewEntry->outSegId3   =  arg_nhlfeSetHandles[2].outSegId;
    case 2:
        pNhlfeSetNewEntry->outSegId2Hi = (arg_nhlfeSetHandles[1].outSegId & 0xFFF00) >> 8;
        pNhlfeSetNewEntry->outSegId2Lo =  arg_nhlfeSetHandles[1].outSegId & 0x000FF;
    case 1:
        pNhlfeSetNewEntry->outSegId1   =  arg_nhlfeSetHandles[0].outSegId;
    }
    pNhlfeSetNewEntry->reserved2 = 0;
    
err_exit:
    IX_ERROR_CRT(ix_ossl_mutex_unlock(pContext->nhlfeMutexId),
                 IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK, 
                 IX_ERROR_LEVEL_REMOTE);

    return err;
} /* ix_cc_mpls_nhlfe_set_create */






/*
 * FUNCTION: ix_cc_mpls_nhlfe_set_delete()  */
/**
 * ABSTRACT: Deletes an NHLFE table entry containing NHLFE set.
 *
 * DESCRIPTION: This function deletes an NHLFE entry containing 
 *              an NHLFE set. It cannot be used for deleting 
 *              a regular NHLFE; for such an entry use ix_cc_mpls_nhlfe_delete().
 *
 * \param  arg_NhlfeHandle - handle identifying an NHLFE set entry 
 * \param  arg_pContext    - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE does not exist
 * \n         IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET - this is not an NHLFE set entry
 * \n         IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK - mutex lock failed
 */
ix_error
ix_cc_mpls_nhlfe_set_delete(ix_cc_mpls_nhlfe_handle  arg_NhlfeHandle,
                            void                    *arg_pContext)
{

    ix_cc_mpls_cc_nhlfe           *pCcNhlfe;
    ix_cc_mpls_nhlfe_set_internal *pNhlfeSet;
    ix_cc_mpls_context            *pContext = arg_pContext;
    ix_uint32                      outSegId;
    ix_error                       err;
    
    /* Verify the input arguments */
    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_nhlfe_set_delete(): Null MPLS context argument"));

    if (pContext->pCcNhlfeBaseAddr == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_nhlfe_set_delete(): Null CC NHLFE table address"));
    
    if (arg_NhlfeHandle.bladeId != pContext->localBladeId)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE, 
                                ("ix_cc_mpls_nhlfe_set_delete(): wrong bladeId"));
    
    IX_ERROR_C(ix_ossl_mutex_lock(pContext->nhlfeMutexId, IX_OSSL_WAIT_FOREVER), err);

    if (err != IX_SUCCESS)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK,
                                ("ix_cc_mpls_nhlfe_set_delete(): can't lock mutex"));

    /* Check whether this is a valid set entry */
    pCcNhlfe = &pContext->pCcNhlfeBaseAddr[arg_NhlfeHandle.outSegId];
    switch (pCcNhlfe->flags)
    {
        /* Valid set entry */
    case MPLS_CC_NHLFE_FLAG_NHLFE_SET:
        break;
        
        /* Check whether this is a valid entry */
    case MPLS_CC_NHLFE_FLAG_FREE:
        err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                               ("ix_cc_mpls_nhlfe_set_delete(): invalid NHLFE entry (empty entry)"));
        goto err_exit;

        /* Check whether this is not a regular entry */
    case MPLS_CC_NHLFE_FLAG_NHLFE:
        err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET,
                               ("ix_cc_mpls_nhlfe_set_delete(): invalid NHLFE entry (set entry)"));
        goto err_exit;

        /* Check whether this is not a NHLFE entry used in a set */
    case MPLS_CC_NHLFE_FLAG_NHLFE_USED_IN_SET:
        err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET,
                               ("ix_cc_mpls_nhlfe_set_delete(): invalid NHLFE entry (entry used in set)"));
        goto err_exit;
        
    default:
        err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                               ("ix_cc_mpls_nhlfe_set_delete(): invalid NHLFE entry (unspecified entry type)"));
        goto err_exit;
    }
    
    /* Check if CC NHLFE entry is referenced by LSPs */
    if (pCcNhlfe->refCount > 0)
    {
        err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_ENTRY_USED,
                               ("ix_cc_mpls_nhlfe_set_delete(): entry referenced by LSP"));
        goto err_exit;
    }

    /* delete references to all set entries */
    pNhlfeSet = (ix_cc_mpls_nhlfe_set_internal*) &pContext->pNhlfeBaseAddr[arg_NhlfeHandle.outSegId];
    switch (pNhlfeSet->setSize)
    {
    case 4:
        outSegId = (pNhlfeSet->outSegId4Hi << 16) | pNhlfeSet->outSegId4Lo;
        pContext->pCcNhlfeBaseAddr[outSegId].flags = MPLS_CC_NHLFE_FLAG_NHLFE;
    case 3:
        outSegId = pNhlfeSet->outSegId3;
        pContext->pCcNhlfeBaseAddr[outSegId].flags = MPLS_CC_NHLFE_FLAG_NHLFE;
    case 2:
        outSegId = (pNhlfeSet->outSegId2Hi << 8) | pNhlfeSet->outSegId2Lo;
        pContext->pCcNhlfeBaseAddr[outSegId].flags = MPLS_CC_NHLFE_FLAG_NHLFE;
    case 1:
        outSegId = pNhlfeSet->outSegId1;
        pContext->pCcNhlfeBaseAddr[outSegId].flags = MPLS_CC_NHLFE_FLAG_NHLFE;
    }
    /* Add CC NHLFE entry back to freeSlotList */
    pCcNhlfe->flags = MPLS_CC_NHLFE_FLAG_FREE;

    _ix_cc_mpls_nhlfe_add_slot(arg_NhlfeHandle, pContext);

err_exit:
    IX_ERROR_CRT(ix_ossl_mutex_unlock(pContext->nhlfeMutexId),
                 IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK, 
                 IX_ERROR_LEVEL_REMOTE);
    
    return err;
} /* ix_cc_mpls_set_delete */




/*
 * FUNCTION: ix_cc_mpls_nhlfe_set_modify()  */
/**
 * ABSTRACT: This function modifies the NHLFE table entry containing an NHLFE set.
 *
 * DESCRIPTION: This function modifies the NHLFE table entry containing an NHLFE set.
 *              The NHLFE handles comprising the old set definition are replaced 
 *              by the handles specified by the arg_nhlfeSetSize and 
 *              arg_nhlfeSetHandles[] function arguments. The new set handles 
 *              must point to regular (non-set) NHLFEs.
 *
 * \param  arg_NhlfeHandle        -	handle identifying an NHLFE entry 
 *                                  containing an NHLFE set
 * \param  arg_nhlfeSetSize       -	number of elements in the new NHLFE set
 * \param  arg_nhlfeSetHandles[]  - pointer to the array of handles defining 
 *                                  the new NHLFEs for the set 
 * \param  arg_pContext           - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE does not exist
 * \n         IX_CC_MPLS_ERROR_ENTRY_NHLFE     - this is not a NHLFE set entry
 * \n         IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET - one of the entries belonging to the set
 *                                               is a set entry
 * \n         IX_CC_ERROR_RANGE                - improper set size
 * \n         IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK - mutex lock failed
 */
ix_error
ix_cc_mpls_nhlfe_set_modify(ix_cc_mpls_nhlfe_handle  arg_NhlfeHandle,
                            ix_uint16                arg_nhlfeSetSize,
                            ix_cc_mpls_nhlfe_handle  arg_nhlfeSetHandles[],
                            void                    *arg_pContext)
{
    ix_cc_mpls_cc_nhlfe           *pCcNhlfeEntry;
    ix_cc_mpls_nhlfe_set_internal *pNhlfeSet;
    ix_cc_mpls_context            *pContext = arg_pContext;
    ix_uint32                      i, outSegId[MPLS_MAX_NHLFE_IN_SET];
    ix_error                       err;
    
    /* Verify the input arguments */
    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_nhlfe_set_modify(): Null MPLS context argument"));

    if (pContext->pCcNhlfeBaseAddr == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_nhlfe_set_modify(): Null CC NHLFE table address"));

    if (arg_NhlfeHandle.bladeId != pContext->localBladeId)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE, 
                                ("ix_cc_mpls_nhlfe_set_modify(): wrong bladeId"));

    if (arg_nhlfeSetHandles == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("ix_cc_mpls_nhlfe_set_modify(): Null NHLFE set handles table"));

    if ((arg_nhlfeSetSize < 1)  ||  (arg_nhlfeSetSize > MPLS_MAX_NHLFE_IN_SET))
        return IX_ERROR_WARNING(IX_CC_ERROR_RANGE, 
                                ("ix_cc_mpls_nhlfe_set_modify(): wrong bladeId"));
    
    IX_ERROR_C(ix_ossl_mutex_lock(pContext->nhlfeMutexId, IX_OSSL_WAIT_FOREVER), err);

    if (err != IX_SUCCESS)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK,
                                ("ix_cc_mpls_nhlfe_set_modify(): can't lock mutex"));

    /* Check whether this is a valid NHLFE set entry */
    pCcNhlfeEntry = &pContext->pCcNhlfeBaseAddr[arg_NhlfeHandle.outSegId];
    switch (pCcNhlfeEntry->flags)
    {
        /* Valid entry */
    case MPLS_CC_NHLFE_FLAG_NHLFE_SET:
        break;

        /* Check whether this is a valid entry */
    case MPLS_CC_NHLFE_FLAG_FREE:
        err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                               ("ix_cc_mpls_nhlfe_set_modify(): invalid NHLFE entry (empty entry)"));
        goto err_exit;

        /* Check whether this is not a NHLFE  entry */
    case MPLS_CC_NHLFE_FLAG_NHLFE:
    case MPLS_CC_NHLFE_FLAG_NHLFE_USED_IN_SET:
        err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_ENTRY_NHLFE,
                               ("ix_cc_mpls_nhlfe_set_modify(): invalid NHLFE entry (plain NHLFE entry)"));
        goto err_exit;
        
    default:
        err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                               ("ix_cc_mpls_nhlfe_set_modify(): invalid NHLFE entry (unspecified entry type)"));
        goto err_exit;
    }

    /* Check if CC NHLFE entry is referenced by LSPs */
    if (pCcNhlfeEntry->refCount > 0)
    {
        err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_ENTRY_USED,
                               ("ix_cc_mpls_nhlfe_set_modify(): invalid NHLFE entry (entry referenced by LSP)"));
        goto err_exit;
    }

    /* get reference to the set */
    pNhlfeSet = (ix_cc_mpls_nhlfe_set_internal*) &pContext->pNhlfeBaseAddr[arg_NhlfeHandle.outSegId];
    
    /* check if all NHLFE entries in the new set are proper */
    for (i=0; i<arg_nhlfeSetSize; i++)
    {
        if (arg_nhlfeSetHandles[i].bladeId != pContext->localBladeId)
        {
	    err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE,
                                   ("ix_cc_mpls_nhlfe_set_modify(): invalid NHLFE entry (invalid NHLFE handle (invalid bladeId))"));
            goto err_exit;
        }
        pCcNhlfeEntry = &(pContext->pCcNhlfeBaseAddr[arg_nhlfeSetHandles[i].outSegId]);
        switch (pCcNhlfeEntry->flags)
        {
        case MPLS_CC_NHLFE_FLAG_NHLFE:
            break;

            /* NHLFE entry is emplty */
        case MPLS_CC_NHLFE_FLAG_FREE:
	    err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                                   ("ix_cc_mpls_nhlfe_set_modify(): invalid NHLFE entry (empty entry)"));
            goto err_exit;

            /* entry already contains NHLFE set */
        case MPLS_CC_NHLFE_FLAG_NHLFE_SET:
	    err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET,
                                   ("ix_cc_mpls_nhlfe_set_modify(): invalid NHLFE entry (set entry)"));
            goto err_exit;

	    /* entry already used in other NHLFE set */
        case MPLS_CC_NHLFE_FLAG_NHLFE_USED_IN_SET:
            /* check if the entry is one of the old entries */
            switch(pNhlfeSet->setSize)
            {
            case 4:
                if (arg_nhlfeSetHandles[i].outSegId == ((pNhlfeSet->outSegId4Hi << 16) | pNhlfeSet->outSegId4Lo))
                    break;
            case 3:
                if (arg_nhlfeSetHandles[i].outSegId == pNhlfeSet->outSegId3)
                    break;
            case 2:
                if (arg_nhlfeSetHandles[i].outSegId == ((pNhlfeSet->outSegId2Hi << 8) | pNhlfeSet->outSegId2Lo))
                    break;
            case 1:
                if (arg_nhlfeSetHandles[i].outSegId == pNhlfeSet->outSegId1)
                    break;
            default:
                err = IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
                                       ("ix_cc_mpls_nhlfe_set_modify(): invalid NHLFE set size"));
                goto err_exit;
            }
            break;
            
        default:
            err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                                   ("ix_cc_mpls_nhlfe_set_modify(): invalid NHLFE entry (unspecified entry type)"));
            goto err_exit;
        }

        /* Check if CC NHLFE entry is referenced by LSPs */
        if (pCcNhlfeEntry->refCount > 0)
        {
            err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_ENTRY_USED,
                                   ("ix_cc_mpls_nhlfe_set_modify(): entry referenced by LSP"));
            goto err_exit;
        }
    }
    
    /* delete references to old set entries */
    switch(pNhlfeSet->setSize)
    {
    case 4: outSegId[3] = (pNhlfeSet->outSegId4Hi << 16) | pNhlfeSet->outSegId4Lo;
    case 3: outSegId[2] =  pNhlfeSet->outSegId3;
    case 2: outSegId[1] = (pNhlfeSet->outSegId2Hi << 8) | pNhlfeSet->outSegId2Lo;
    case 1: outSegId[0] =  pNhlfeSet->outSegId1;
    }
    for (i=0; i<pNhlfeSet->setSize; i++)
    {
        pContext->pCcNhlfeBaseAddr[outSegId[i]].flags = MPLS_CC_NHLFE_FLAG_NHLFE;
    }
    
    /* Set flags for new NHLFE entries used in the set */
    for (i=0; i<arg_nhlfeSetSize; i++)
    {
        pCcNhlfeEntry = &(pContext->pCcNhlfeBaseAddr[arg_nhlfeSetHandles[i].outSegId]);
        pCcNhlfeEntry->flags = MPLS_CC_NHLFE_FLAG_NHLFE_USED_IN_SET;
    }

    /* Fill the NHLFE Table's set entry */
    pNhlfeSet->flags = MPLS_FLAG_VALID_BIT | MPLS_FLAG_SET_BIT;
    pNhlfeSet->exp2phb = 0;
    pNhlfeSet->setSize = arg_nhlfeSetSize;
    pNhlfeSet->phbMask1 = 0;
    pNhlfeSet->phbMask2 = 0;
    pNhlfeSet->phbMask3 = 0;
    pNhlfeSet->phbMask4 = 0;
    pNhlfeSet->outSegId1   =  arg_nhlfeSetHandles[0].outSegId;
    pNhlfeSet->outSegId2Hi = (arg_nhlfeSetHandles[1].outSegId & 0xFFF00) >> 8;
    pNhlfeSet->outSegId2Lo =  arg_nhlfeSetHandles[1].outSegId & 0x000FF;
    pNhlfeSet->outSegId3   =  arg_nhlfeSetHandles[2].outSegId;
    pNhlfeSet->outSegId4Hi = (arg_nhlfeSetHandles[3].outSegId & 0xF0000) >> 16;
    pNhlfeSet->outSegId4Lo =  arg_nhlfeSetHandles[3].outSegId & 0x0FFFF;
    pNhlfeSet->reserved2 = 0;

err_exit:    
    IX_ERROR_CRT(ix_ossl_mutex_unlock(pContext->nhlfeMutexId),
                 IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK, 
                 IX_ERROR_LEVEL_REMOTE);

    return err;
} /* ix_cc_mpls_nhlfe_set_modify */





/*
 * FUNCTION: _ix_cc_mpls_nhlfe_internal_modify()  */
/**
 * ABSTRACT: This function modifies the NHLFE parameters connected with LSP
 *
 * DESCRIPTION: This function modifies the NHLFE parameters that user passes
 *              as LSP information. Thus this function is invoked internally
 *              by ix_cc_mpls_lsp_create / modify.
 *
 * \param  arg_NhlfeHandle        -	handle identifying an NHLFE entry 
 *                                  containing an NHLFE set
 * \param  arg_lspParamsSel       - flags denoting which fields of lsp has to be modified
 * \param  arg_lspParams          - LSP parameters that should be kept in NHLFE 
 *                                  structure
 * \param  arg_pContext           - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE does not exist
 * \n         IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET - this is not an NHLFE set entry
 */
ix_error
_ix_cc_mpls_nhlfe_internal_modify(ix_cc_mpls_nhlfe_handle  arg_NhlfeHandle,
                                  ix_uint16                arg_lspParamsSel,
                                  ix_cc_mpls_lsp_param     arg_lspParams,
                                  void                    *arg_pContext)
{
    ix_cc_mpls_cc_nhlfe        *pCcNhlfeEntry;
    ix_cc_mpls_nhlfe_internal  *pNhlfe;
    ix_cc_mpls_context         *pContext = arg_pContext;
    
    /* Verify the input arguments */
    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("_ix_cc_mpls_nhlfe_internal_modify(): Null MPLS context argument"));

    if (pContext->pCcNhlfeBaseAddr == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
                                ("_ix_cc_mpls_nhlfe_internal_modify(): Null CC NHLFE table address"));

    if (arg_NhlfeHandle.bladeId != pContext->localBladeId)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE, 
                                ("_ix_cc_mpls_nhlfe_internal_modify(): wrong bladeId"));

    /* Check whether this is a valid entry */
    pCcNhlfeEntry = &pContext->pCcNhlfeBaseAddr[arg_NhlfeHandle.outSegId];

    if  ((pCcNhlfeEntry->flags != MPLS_CC_NHLFE_FLAG_NHLFE) && 
         (pCcNhlfeEntry->flags != MPLS_CC_NHLFE_FLAG_NHLFE_SET))

        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                                ("_ix_cc_mpls_nhlfe_internal_modify(): You cant change this type of NHLFE"));
       
    pNhlfe = (ix_cc_mpls_nhlfe_internal*) &pContext->pNhlfeBaseAddr[arg_NhlfeHandle.outSegId];

#if defined (IX_MPLS_DEBUG)
    ix_ossl_message_log("_ix_cc_mpls_nhlfe_internal_modify(): Modifying Nhlfe at 0x%x    handle = 0x%x\n",
                        pNhlfe, arg_NhlfeHandle);
#endif


    /* Fill the NHLFE Table's set entry */
    if (arg_lspParamsSel | MPLS_SET_FLAGS)  pNhlfe->flags          = pNhlfe->flags | arg_lspParams.flags;
    if (arg_lspParamsSel | MPLS_SET_PHB)    pNhlfe->phb2exp        = arg_lspParams.phb2expIdx;
    if (arg_lspParamsSel | MPLS_SET_EXP)    pNhlfe->exp2phb        = arg_lspParams.exp2phbIdx;
    if (arg_lspParamsSel | MPLS_SET_MTU)    pNhlfe->maxLabPktSize  = arg_lspParams.maxLabPktSize;
    if (arg_lspParamsSel | MPLS_SET_FLOWID)
    {
                                            pNhlfe->flowIdHigh     = (arg_lspParams.flowID & 0x000FF000) >> 12;
                                            pNhlfe->flowIdLow      = (arg_lspParams.flowID & 0x00000FFF);
    }
    if (arg_lspParamsSel | MPLS_SET_TTL)    pNhlfe->ttl            = arg_lspParams.ttl;

    return IX_SUCCESS;
} /* ix_cc_mpls_nhlfe_internal_modify */







/*
 * FUNCTION: _ix_cc_mpls_nhlfe_query_params()  */
/**
 * ABSTRACT: This function reads NHLFE parameters which are seen as atributes of LSP
 *
 * DESCRIPTION: This function reads NHLFE parameters which are seen as atributes of LSP.
 *              This function is invoked internally by ix_cc_mpls_lsp_query.
 *
 * \param  arg_NhlfeHandle        -	handle identifying an NHLFE entry 
 *                                  containing an NHLFE set
 * \param  arg_lspParams          - structere where LSP parameters are put
 * \param  arg_pContext           - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE does not exist
 * \n         IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET - this is not an NHLFE set entry
 */
ix_error
_ix_cc_mpls_nhlfe_query_params(ix_cc_mpls_nhlfe_handle  arg_NhlfeHandle,
                               ix_cc_mpls_lsp_param    *arg_plspParams,
                               void                    *arg_pContext)
{
    ix_cc_mpls_cc_nhlfe        *pCcNhlfeEntry;
    ix_cc_mpls_nhlfe_internal  *pNhlfe;
    ix_cc_mpls_context         *pContext = arg_pContext;
    
    /* Verify the input arguments */
    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
           ("_ix_cc_mpls_nhlfe_query_params(): Null MPLS context argument"));

    if (pContext->pCcNhlfeBaseAddr == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
           ("_ix_cc_mpls_nhlfe_query_params(): Null CC NHLFE table address"));

    if (arg_NhlfeHandle.bladeId != pContext->localBladeId)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE, 
           ("_ix_cc_mpls_nhlfe_query_params(): wrong bladeId"));

    /* Check whether this is a valid entry */
    pCcNhlfeEntry = &pContext->pCcNhlfeBaseAddr[arg_NhlfeHandle.outSegId];
    if  ((pCcNhlfeEntry->flags != MPLS_CC_NHLFE_FLAG_NHLFE) && 
         (pCcNhlfeEntry->flags != MPLS_CC_NHLFE_FLAG_NHLFE_SET))

        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND, 
           ("_ix_cc_mpls_nhlfe_query_params(): You cant read lsp params from NHLFE SET"));
   

    pNhlfe = (ix_cc_mpls_nhlfe_internal*) &pContext->pNhlfeBaseAddr[arg_NhlfeHandle.outSegId];

    arg_plspParams->flags          = pNhlfe->flags;
    arg_plspParams->exp2phbIdx     = (ix_uint16)pNhlfe->exp2phb;
    arg_plspParams->maxLabPktSize  = (ix_uint16)pNhlfe->maxLabPktSize;
    arg_plspParams->flowID         = pNhlfe->flowIdLow | (pNhlfe->flowIdHigh <<12);
    arg_plspParams->phb2expIdx     = (ix_uint16)pNhlfe->phb2exp;
    arg_plspParams->ttl            = (ix_uint8)pNhlfe->ttl;
    
    return IX_SUCCESS;
} /* _ix_cc_mpls_nhlfe_query_params */





/*
 * FUNCTION: ix_cc_mpls_nhlfe_set_query()  */
/**
 * ABSTRACT: This function gets an NHLFE entry containing an NHLFE set.
 *
 * DESCRIPTION: This function reads an NHLFE entry containing an NHLFE set.
 *              It returns only the NHLFE handles of the set elements.
 *              The definition of each set element must be read separately 
 *              by calling the  ix_cc_mpls_nhlfe_query() function for 
 *              each returned handle.
 *
 * \param  arg_NhlfeHandle     - handle identifying an NHLFE entry containing 
 *                               an NHLFE set
 * \param  arg_SetSize         - pointer to the buffer storing the number of
 *                               the NHLFE set handles returned by the function.
 * \param  arg_NhlfeSetHandles - pointer to the buffer for storing the NHLFE
 *                               handles identifying the set elements returned 
 *                               by the function.  
 * \param  arg_pContext        - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL              - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND   - NHLFE does not exist
 * \n         IX_CC_MPLS_ERROR_ENTRY_NHLFE  - this is not an NHLFE set entry
 * \n         IX_CC_ERROR_RANGE             - improper set size
 * \n         IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK - mutex lock failed
 */
ix_error
ix_cc_mpls_nhlfe_set_query(ix_cc_mpls_nhlfe_handle   arg_NhlfeHandle,
                           ix_uint16                *arg_nhlfeSetSize,
                           ix_cc_mpls_nhlfe_handle   arg_nhlfeSetHandles[],
                           void                     *arg_pContext)
{
    ix_cc_mpls_cc_nhlfe           *pCcNhlfeEntry;
    ix_cc_mpls_nhlfe_set_internal *pNhlfeSetEntry;
    ix_cc_mpls_context            *pContext = arg_pContext;
    ix_uint32                      i;
    ix_error                       err;
    
    /* Verify the input arguments */
    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("Null MPLS context argument"));

    if (pContext->pCcNhlfeBaseAddr == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("Null CC NHLFE table address"));

    if (arg_NhlfeHandle.bladeId != pContext->localBladeId)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE, 
                   ("ix_cc_mpls_nhlfe_set_query(): wrong bladeId"));

    if (arg_nhlfeSetSize == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("Null pointer to set size variable"));

    if (arg_nhlfeSetHandles == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("Null NHLFE set handles table"));

    IX_ERROR_C(ix_ossl_mutex_lock(pContext->nhlfeMutexId, IX_OSSL_WAIT_FOREVER), err);

    if (err != IX_SUCCESS)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK,
                      ("ix_cc_mpls_nhlfe_set_query(): can't lock mutex"));

    /* Check whether this is a valid entry */
    pCcNhlfeEntry = &pContext->pCcNhlfeBaseAddr[arg_NhlfeHandle.outSegId];
    switch (pCcNhlfeEntry->flags)
    {
        /* Valid entry */
    case MPLS_CC_NHLFE_FLAG_NHLFE_SET:
        break;

        /* Check whether this is a valid entry */
    case MPLS_CC_NHLFE_FLAG_FREE:
        err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                  ("ix_cc_mpls_nhlfe_set_query(): invalid NHLFE entry (empty entry)"));
        goto err_exit;

        /* Check whether this is not a regular NHLFE entry */
    case MPLS_CC_NHLFE_FLAG_NHLFE:
        err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_ENTRY_NHLFE,
                  ("ix_cc_mpls_nhlfe_set_query(): invalid NHLFE entry (plain NHLFE entry)"));
        goto err_exit;

        /* Check whether this is not a NHLFE entry used in a set */
    case MPLS_CC_NHLFE_FLAG_NHLFE_USED_IN_SET:
        err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET,
                  ("ix_cc_mpls_nhlfe_set_query(): invalid NHLFE entry (entry used in set)"));
        goto err_exit;
        
    default:
        err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                  ("ix_cc_mpls_nhlfe_set_query(): invalid NHLFE entry (unspecified entry type)"));
        goto err_exit;
    }

    /* Get NHLFE set entry */
    pNhlfeSetEntry = (ix_cc_mpls_nhlfe_set_internal*) &pContext->pNhlfeBaseAddr[arg_NhlfeHandle.outSegId];
    *arg_nhlfeSetSize = (ix_uint16)pNhlfeSetEntry->setSize;

    /* return parameters */
    switch(*arg_nhlfeSetSize)
    {
    case 4:
        arg_nhlfeSetHandles[3].bladeId  = pContext->localBladeId;
        arg_nhlfeSetHandles[3].outSegId = (pNhlfeSetEntry->outSegId4Hi << 16) | pNhlfeSetEntry->outSegId4Lo;
    case 3:
        arg_nhlfeSetHandles[2].bladeId  = pContext->localBladeId;
        arg_nhlfeSetHandles[2].outSegId = pNhlfeSetEntry->outSegId3;
    case 2: 
        arg_nhlfeSetHandles[1].bladeId  = pContext->localBladeId;
        arg_nhlfeSetHandles[1].outSegId = (pNhlfeSetEntry->outSegId2Hi << 8) | pNhlfeSetEntry->outSegId2Lo;
    case 1:
        arg_nhlfeSetHandles[0].bladeId  = pContext->localBladeId;
        arg_nhlfeSetHandles[0].outSegId = pNhlfeSetEntry->outSegId1;
	break;
    default:
        err = IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
                  ("ix_cc_mpls_nhlfe_set_query(): invalid NHLFE set size"));
        goto err_exit;
    }

    /* Check if all NHLFE entries in the set are valid */
    for (i=0; i<*arg_nhlfeSetSize; i++)
    {
        if (arg_nhlfeSetHandles[i].bladeId != pContext->localBladeId)
        {
            err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE,
                  ("ix_cc_mpls_nhlfe_set_query(): invalid NHLFE handle (invalid bladeId)"));
            goto err_exit;
        }

        pCcNhlfeEntry = &(pContext->pCcNhlfeBaseAddr[arg_nhlfeSetHandles[i].outSegId]);
        switch (pCcNhlfeEntry->flags)
        {
        case MPLS_CC_NHLFE_FLAG_NHLFE:
        case MPLS_CC_NHLFE_FLAG_NHLFE_USED_IN_SET:
            break;

            /* NHLFE entry is emplty */
        case MPLS_CC_NHLFE_FLAG_FREE:
            err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                  ("ix_cc_mpls_nhlfe_set_query(): invalid NHLFE entry (empty entry)"));
            goto err_exit;

            /* entry already contains NHLFE set */
        case MPLS_CC_NHLFE_FLAG_NHLFE_SET:
            err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_ENTRY_NHLFE_SET,
                  ("ix_cc_mpls_nhlfe_set_query(): invalid NHLFE entry (set entry)"));
            goto err_exit;

        default:
            err = IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                  ("ix_cc_mpls_nhlfe_set_query(): invalid NHLFE entry (unspecified entry type)"));
            goto err_exit;
        }
    }

err_exit:
    IX_ERROR_CRT(ix_ossl_mutex_unlock(pContext->nhlfeMutexId),
                 IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK, 
                 IX_ERROR_LEVEL_REMOTE);

    return err;
} /* ix_cc_mpls_nhlfe_set_query */




/*
 * FUNCTION: ix_cc_mpls_exp2phb_modify ()  */
/**
 * ABSTRACT: This function modifies an element of  EXP2PHB mapping table.
 *
 * DESCRIPTION: This function modifies an EXP2PHB table entry. 
 *              Note: all EXP2PHB table entries are created 
 *              during initialization and set to defaults.
 *
 * \param  arg_Index    - index identifying an EXP2PHB table entry
 * \param  arg_exp2phb  - pointer to the table containing the mappings 
 *                        for one EXP2PHB table entry
 * \param  arg_pContext - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE does not exist
 */
ix_error
ix_cc_mpls_exp2phb_modify (ix_uint16           arg_Index,
                           ix_cc_mpls_exp2phb  arg_exp2phb[MPLS_MAX_EXP_NUM],
                           void               *arg_pContext)
{

    /* Verify the input arguments */
    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("ix_cc_mpls_exp2phb_modify(): "
                                                   "Null MPLS context argument"));

    if (arg_exp2phb == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("ix_cc_mpls_exp2phb_modify(): "
                                                   "Null arg_exp2phb argument"));

    return IX_ERROR_WARNING(IX_CC_ERROR_UNIMPL, ("ix_cc_mpls_exp2phb_modify(): "
                                                 "function not supported in the release"));

} /* ix_cc_mpls_exp2phb_modify */




/*
 * FUNCTION: ix_cc_mpls_exp2phb_query  ()  */
/**
 * ABSTRACT: This function gets an element of  EXP2PHB mapping table.
 *
 * DESCRIPTION: This function reads an EXP2PHB table entry. 
 *
 * \param  arg_Index    - index identifying an EXP2PHB table entry
 * \param  arg_exp2phb  - pointer to the buffer storing the table
 *                        containing the mappings for one EXP2PHB table entry.  
 * \param  arg_pContext - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE does not exist
 */
ix_error
ix_cc_mpls_exp2phb_query (ix_uint16             arg_Index,
                          ix_cc_mpls_exp2phb    arg_exp2phb[MPLS_MAX_EXP_NUM],
                          void                 *arg_pContext)
{

    /* Verify the input arguments */
    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("ix_cc_mpls_exp2phb_query(): "
                                                   "Null MPLS context argument"));

    if (arg_exp2phb == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("ix_cc_mpls_exp2phb_query(): "
                                                   "Null arg_exp2phb argument"));

    return IX_ERROR_WARNING(IX_CC_ERROR_UNIMPL, ("ix_cc_mpls_exp2phb_query(): "
                                                 "function not supported in the release"));
}



/*
 * FUNCTION: ix_cc_mpls_phb2exp_modify ()  */
/**
 * ABSTRACT: This function modifies a PHB2EXP table entry.
 *
 * DESCRIPTION: This function modifies a PHB2EXP table entry.
 *              Note: all PHB2EXP table entries are created during 
 *              initialization and set to defaults. 
 *
 * \param  arg_Index    - index identifying a PHB2EXP table entry
 * \param  arg_phb2exp  - pointer to the mappings for one PHB2EXP table entry
 * \param  arg_pContext - pointer to the MPLS core component context 
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE does not exist
 */
ix_error
ix_cc_mpls_phb2exp_modify (ix_uint16           arg_Index,
                           ix_cc_mpls_phb2exp  arg_phb2exp[MPLS_MAX_PHB_NUM],
                           void               *arg_pContext)
{

    /* Verify the input arguments */
    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("ix_cc_mpls_phb2exp_modify (): "
                                                   "Null MPLS context argument"));

    if (arg_phb2exp == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("ix_cc_mpls_phb2exp_modify (): "
                                                   "Null arg_phb2exp argument"));

    return IX_ERROR_WARNING(IX_CC_ERROR_UNIMPL, ("ix_cc_mpls_phb2exp_modify (): "
                                                 "function not supported in the release"));

} /* ix_cc_mpls_set_phb2exp */




/*
 * FUNCTION: ix_cc_mpls_phb2exp_query()  */
/**
 * ABSTRACT: This function gets a PHB2EXP table entry.
 *
 * DESCRIPTION: This function reads a PHB2EXP table entry
 *
 * \param  arg_Index    - index identifying a PHB2EXP table entry
 * \param  arg_pContext - pointer to the MPLS core component context
 * \param  arg_phb2exp  - pointer to the buffer storing the mappings
 *                        for one PHB2EXP table entry returned 
 *                        by the function.  
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_NULL                 - invalid pointer
 * \n         IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE does not exist
 */
ix_error
ix_cc_mpls_phb2exp_query(ix_uint16            arg_Index,
                         ix_cc_mpls_phb2exp   arg_phb2exp[MPLS_MAX_PHB_NUM],
                         void                *arg_pContext)
{

    /* Verify the input arguments */
    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("ix_cc_mpls_phb2exp_query(): "
                                                   "Null MPLS context argument"));

    if (arg_phb2exp == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("ix_cc_mpls_phb2exp_query(): "
                                                   "Null arg_phb2exp argument"));

    return IX_ERROR_WARNING(IX_CC_ERROR_UNIMPL, ("ix_cc_mpls_phb2exp_query(): "
                                                 "function not supported in the release"));
} /* ix_cc_mpls_phb2exp_query */






/*
 * FUNCTION: _ix_cc_mpls_nhlfe_allocate_memory()  */
/**
 * ABSTRACT: These function allocates resources for NHLFE library
 *
 * DESCRIPTION: These function allocates resources needed initially
 *              by NHLFE library
 *
 * \param  arg_pContext - a pointer to the MPLS NHLFE context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_OOM_SRAM   - memory allocation failure
 * \n         IX_CC_ERROR_OOM_SYSTEM - memory allocation failure
 */
PRIVATE ix_error
_ix_cc_mpls_nhlfe_allocate_memory(ix_cc_mpls_context *arg_pContext)
{

    ix_cc_mpls_context        *pContext = arg_pContext;

    /* Allocate memory and initialize CC NHLFE table */
    /* +1 -> entry no. 0 is left free */
    arg_pContext->pCcNhlfeBaseAddr = ix_ossl_malloc((arg_pContext->maxNhlfeEntries+1) *
                                                    sizeof(ix_cc_mpls_cc_nhlfe));
    if (arg_pContext->pCcNhlfeBaseAddr == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
                                ("Allocating CC NHLFE table -> out of memory"));
    }
    
    ix_ossl_memset((void*)arg_pContext->pCcNhlfeBaseAddr,
                   0,
                   (ix_ossl_size_t)
                   ((arg_pContext->maxNhlfeEntries+1) *
                    sizeof(ix_cc_mpls_cc_nhlfe)));

    /* Allocate memory and initialize NHLFE table */
    IX_ERROR_CRT(ix_rm_mem_alloc(arg_pContext->nhlfeTableMemoryType,
                                 arg_pContext->nhlfeTableMemoryChannel,
                                 (arg_pContext->maxNhlfeEntries+1) *
                                 sizeof(ix_cc_mpls_nhlfe_internal),
                                 (void **)&arg_pContext->pNhlfeBaseAddr),
                 IX_CC_ERROR_OOM_SRAM,
                 IX_ERROR_LEVEL_REMOTE);
    ix_ossl_memset((void*)arg_pContext->pNhlfeBaseAddr,
                   0,
                   (ix_ossl_size_t)
                   ((arg_pContext->maxNhlfeEntries+1) *
                    sizeof(ix_cc_mpls_nhlfe_internal)));

#ifdef MPLS_COUNTERS_FEATURE

    {
        ix_error err;
        
        /* Allocate memory and initialize 32-bit FTN counters */
        err = ix_rm_mem_alloc(IX_MEMORY_TYPE_SRAM,
                                     pContext->nhlfeCountersMemoryChannel,
                                     (pContext->maxNhlfeEntries+1) *
                                     sizeof(ix_cc_mpls_nhlfe_counters_internal),
                                     (void **)&pContext->pFtnCounters);
        if (err != IX_SUCCESS)
        {
             _ix_cc_mpls_nhlfe_free_resources(pContext);
             return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
                  ("_ix_cc_mpls_nhlfe_allocate_memory(): Out of Memory for Allocating 32 bit FTN counters table"));
        }
        ix_ossl_memset((void *)pContext->pFtnCounters, 0,
                       (ix_ossl_size_t)((pContext->maxNhlfeEntries+1) *
                                        sizeof(ix_cc_mpls_nhlfe_counters_internal)));

        
        /* Allocate memory and initialize 64-bit FTN counters handler array */
        pContext->pFtn64BitCountersHdls = (ix_counter_64bit_handle *)
                                          ix_ossl_malloc((pContext->maxNhlfeEntries+1) * 
                                          sizeof(ix_counter_64bit_handle) * MPLS_NHLFE_64BIT_COUNTERS_NUM);
        if (pContext->pFtn64BitCountersHdls == NULL)
        {
            _ix_cc_mpls_nhlfe_free_resources(pContext);
            return IX_ERROR_REMOTE(IX_CC_ERROR_OOM_SYSTEM, 
                     ("_ix_cc_mpls_nhlfe_allocate_memory(): Out of Memory for allocating 64 bit FTN counters handles"));
        }
        ix_ossl_memset((void *)pContext->pFtn64BitCountersHdls, 0,
                       (pContext->maxNhlfeEntries+1) * sizeof(ix_counter_64bit_handle) * MPLS_NHLFE_64BIT_COUNTERS_NUM);

    }

#endif /* MPLS_COUNTERS_FEATURE */

    return IX_SUCCESS;
} /* _ix_cc_mpls_nhlfe_allocate_memory */






/*
 * FUNCTION: _ix_cc_mpls_nhlfe_free_resources()  */
/**
 * ABSTRACT: These function releases all resources taken by NHLFE library
 *
 * DESCRIPTION: This function frees all memories and resources taken by
 *              NHLFE library
 *
 * \param  arg_pContext - a pointer to the MPLS LSP context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
void
_ix_cc_mpls_nhlfe_free_resources(void *arg_pContext)
{

    ix_cc_mpls_context        *pContext = arg_pContext;

#ifdef MPLS_COUNTERS_FEATURE

    /* Free 64-bit outseg counters handle array */
    if (pContext->pFtn64BitCountersHdls != NULL)
    {
        ix_error err;

        /* First delete each NHLFE entry (related 64-bit outseg counters
           resources will be fried from Resource Manager) */
        err = ix_cc_mpls_nhlfe_purge (arg_pContext);

        if (err != IX_SUCCESS)
             ix_ossl_message_log("_ix_cc_mpls_nhlfe_free_resources(): "
                   "Warning: Nhlfe table purge failed. \n");

        /* Now free memory used by outseg counters handlers arrays */
        ix_ossl_free(pContext->pFtn64BitCountersHdls);
    }
    /* Free memory used by 32-bit outseg counters (and also for 64-bit
       outseg counters 32-bit base) array  */
    if (pContext->pFtnCounters != NULL)
    {
        if (ix_rm_mem_free(pContext->pFtnCounters) != IX_SUCCESS)
        {
			ix_ossl_message_log("Warning: Resource Manager "
                                "memory free failed.\n");
        }
    }

#endif /* MPLS_COUNTERS_FEATURE */

    /* Free memory used by NHLFE entries array*/
    if (pContext->pNhlfeBaseAddr != NULL)
    {
        if (ix_rm_mem_free(pContext->pNhlfeBaseAddr) != IX_SUCCESS)
        {
			ix_ossl_message_log("Warning: Resource Manager "
                                "memory free failed.\n");
        }
    }
    /* Free free slot list if used */
    if (pContext->pFreeNhlfeListHead != NULL)
    {
        ix_cc_mpls_cc_nhlfe_free_slot* pFreeSlot =
            pContext->pFreeNhlfeListHead;
        ix_cc_mpls_cc_nhlfe_free_slot* pNextFreeSlot = NULL;

        while (pFreeSlot != NULL)
        {
            pNextFreeSlot = pFreeSlot->pNextFreeSlot;
            /* Remove slot from free NHLFE entries list */
            ix_ossl_free((void *)pFreeSlot);
            pFreeSlot = pNextFreeSlot;
        }
        pContext->pFreeNhlfeListHead = NULL;
        pContext->pFreeNhlfeListTail = NULL;
    }

    return;
} /* _ix_cc_mpls_nhlfe_free_resources */




/*
 * FUNCTION: _ix_cc_mpls_nhlfe_fini()  */
/**
 * ABSTRACT: removes resources taken to generate NHLFE
 *
 * DESCRIPTION: releases the NHLFE table.
 *
 * \param  arg_pContext - a pointer to the MPLS context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_nhlfe_fini(void *arg_pContext)
{
    ix_cc_mpls_context* pContext = (ix_cc_mpls_context*) arg_pContext;
    ix_error            err;
    
    IX_ERROR_CRT(ix_ossl_mutex_lock(pContext->nhlfeMutexId, IX_OSSL_WAIT_FOREVER),
                 IX_CC_MPLS_ERROR_OSSL_MUTEX_LOCK, IX_ERROR_LEVEL_LOCAL);
    
    _ix_cc_mpls_nhlfe_free_resources(pContext);
        
    IX_ERROR_CT(ix_ossl_mutex_unlock(pContext->nhlfeMutexId), err,
                IX_CC_MPLS_ERROR_OSSL_MUTEX_UNLOCK, IX_ERROR_LEVEL_LOCAL);

    /* Remove mutex for critical section. */
    IX_ERROR_CRT(ix_ossl_mutex_fini(pContext->nhlfeMutexId),
                 IX_CC_MPLS_ERROR_OSSL_MUTEX_FINI, 
                 IX_ERROR_LEVEL_REMOTE);
    
    return IX_SUCCESS;
} /* _ix_cc_mpls_nhlfe_fini */




/*
 * FUNCTION: _ix_cc_mpls_nhlfe_patch_symbols()  */
/**
 * ABSTRACT: These function patches all NHLFE related symbols in microcode
 *
 * DESCRIPTION: Patching of various symbols and memory bases into MPLS
 *              microblock
 *
 * \param  arg_pContext - a pointer to the MPLS NHLFE context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n       IX_CC_MPLS_ERROR_RM_GET_PHYS_OFFSET - error by physical offset
 *                                                reading
 * \n       IX_CC_MPLS_ERROR_RM_PATCH_SYMBOL    - error by patching imported
 *                                                symbols into the microcode
 * \n       IX_CC_ERROR_RANGE                   - counters not in SRAM
 */
ix_error 
_ix_cc_mpls_nhlfe_patch_symbols(void *arg_pContext )
{

    ix_cc_mpls_context        *pContext = (ix_cc_mpls_context*)arg_pContext;
    ix_imported_symbol        _Symbols[MPLS_NHLFE_SYMBOLS_PATCHED_NUM];
    
    ix_uint32 symbolIndex = 0;
    ix_uint32 mev2Id;
    ix_uint32 memOffset;

    /* Obtain the physical base offset of parameters for current blade */
    IX_ERROR_CRT( ix_rm_get_phys_offset(pContext->pNhlfeBaseAddr, NULL, NULL,
                                        &memOffset, NULL ),
                  IX_CC_MPLS_ERROR_RM_GET_PHYS_OFFSET,
                  IX_ERROR_LEVEL_REMOTE);
    
    /* Pack the SRAM channel number into the offset */
    if(pContext->nhlfeTableMemoryType == IX_MEMORY_TYPE_SRAM )
    {
        IX_CC_SET_SRAM_CHANNEL_INTO_OFFSET(pContext->nhlfeTableMemoryChannel,
                                           memOffset );
    }
    
    /* Prepare the symbol array */
    MPLS_SET_UBLOCK_SYMBOLS(_Symbols[symbolIndex], memOffset,
                            IX_MPLS_SYMBOL_FTN_NHLFE_BASE );
    symbolIndex++;

#ifdef MPLS_COUNTERS_FEATURE

    /* ==================================================================  */
    /* Obtain the physical base offset of the FTN counters table           */
    IX_ERROR_CRT( ix_rm_get_phys_offset(pContext->pFtnCounters, NULL, NULL,
                                        &memOffset, NULL ),
                  IX_CC_MPLS_ERROR_RM_GET_PHYS_OFFSET,
                  IX_ERROR_LEVEL_REMOTE);
    
    /* Pack the SRAM channel number into the offset */
    if(pContext->nhlfeTableMemoryType != IX_MEMORY_TYPE_SRAM )
    {
        IX_ERROR_WARNING(IX_CC_ERROR_RANGE, ("Counters have to be in SRAM"));
    }
    
    IX_CC_SET_SRAM_CHANNEL_INTO_OFFSET(pContext->nhlfeCountersMemoryChannel,
                                       memOffset);
    
    /* Prepare the symbol array */
    MPLS_SET_UBLOCK_SYMBOLS(_Symbols[symbolIndex], memOffset,
                            IX_MPLS_SYMBOL_FTN_STAT_BASE);
    symbolIndex++;
    
#endif /* MPLS_COUNTERS_FEATURE */

        
    /* Patch them into each microengine used by the microblock */
    for(mev2Id = 0; mev2Id < IX_MAX_MICROENGINE_NUMBER; mev2Id++)
    {
        if (pContext->mev2NumMask & (0x1 << mev2Id))
        {
            IX_ERROR_CRT(ix_rm_ueng_patch_symbols(g_aMicroengineNumberMapping[mev2Id],
                                                  MPLS_NHLFE_SYMBOLS_PATCHED_NUM, _Symbols ),
                         IX_CC_MPLS_ERROR_RM_PATCH_SYMBOL,
                         IX_ERROR_LEVEL_REMOTE
                        );
        }
    }

    return IX_SUCCESS;
} /* _ix_cc_mpls_nhlfe_patch_symbols */




/*
 * FUNCTION: _ix_cc_mpls_nhlfe_add_slot()  */
/**
 * ABSTRACT: Adds new slot into free CC NHLFE entries list
 *
 * DESCRIPTION: Adds new slot into free CC NHLFE entries list
 *
 * \param  arg_NhlfeHandle   - an index identifying an CC NHLFE entry.
 * \param  arg_pNhlfeContext - a pointer to the NHLFE Table context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n         IX_CC_ERROR_OOM_SYSTEM - memory allocation failure
 */
PRIVATE ix_error
_ix_cc_mpls_nhlfe_add_slot(ix_cc_mpls_nhlfe_handle  arg_NhlfeHandle,
                           ix_cc_mpls_context      *arg_pContext)
{
    ix_cc_mpls_cc_nhlfe_free_slot* pNewFreeSlotListEntry = NULL;

    if ((pNewFreeSlotListEntry = (ix_cc_mpls_cc_nhlfe_free_slot*)
         ix_ossl_malloc(sizeof(ix_cc_mpls_cc_nhlfe_free_slot))) == NULL)
    {
        return(IX_ERROR_REMOTE(IX_CC_ERROR_OOM_SYSTEM, ("Out of memory")));
    }
    pNewFreeSlotListEntry->NhlfeHandle = arg_NhlfeHandle;
    pNewFreeSlotListEntry->pNextFreeSlot = NULL;

    if (arg_pContext->pFreeNhlfeListTail == NULL)
    {
        arg_pContext->pFreeNhlfeListHead = pNewFreeSlotListEntry;
        arg_pContext->pFreeNhlfeListTail = pNewFreeSlotListEntry;
    }
    else
    {
        arg_pContext->pFreeNhlfeListTail->pNextFreeSlot =
            pNewFreeSlotListEntry;
        arg_pContext->pFreeNhlfeListTail = pNewFreeSlotListEntry;
    }

    return IX_SUCCESS;
} /* _ix_cc_mpls_nhlfe_add_slot */




/*
 * FUNCTION: _ix_cc_mpls_nhlfe_verify_input()  */
/**
 * ABSTRACT: Verifies outseg data.
 *
 * DESCRIPTION: Verifies outseg data.
 *
 * \param  arg_Nhlfe - an ix_cc_mpls_nhlfe structure
 *                     describing an NHLFE entry or NHLFE set.
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n       IX_CC_ERROR_RANGE - paramters are out of permitted range
 */
PRIVATE ix_error
_ix_cc_mpls_nhlfe_verify_input(ix_cc_mpls_nhlfe arg_Nhlfe)
{

    if (arg_Nhlfe.pushCount > 4)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
                                ("pushCOunt is out of range -> possible values 1..4"));
    }

    /* check L3 protocol type */
    if ((arg_Nhlfe.l3Protocol != MPLS_L3_IP_V4) && 
        (arg_Nhlfe.l3Protocol != MPLS_L3_IP_V6))
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
                                ("unknown L3 protocol type"));
    }

    /* check nextHopID value */
    if (arg_Nhlfe.nextHopId > 0x0000FFFF)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
                                ("nextHopID out of range"));
    }

    return IX_SUCCESS;
} /* _ix_cc_mpls_nhlfe_verify_input */




/*
 * FUNCTION: _ix_cc_mpls_nhlfe_inc_ref_count()  */
/**
 * ABSTRACT: increases ref count of a given CC NHLFE entry
 *
 * \param  arg_NhlfeHandle - handle to NHLFE entry which ref count will be increased
 * \param  arg_pContext    - pointer to the MPLS CC context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * 
 * \n Error codes:
 * \n       IX_CC_ERROR_NULL            - invalid pointer
 * \n       IX_CC_ERROR_ENTRY_NOT_FOUND - NHLFE does not exist
 */
ix_error
_ix_cc_mpls_nhlfe_inc_ref_count(ix_cc_mpls_nhlfe_handle  arg_NhlfeHandle,
                                void                    *arg_pContext)
{
    ix_cc_mpls_cc_nhlfe        *pCcNhlfeEntry;
    ix_cc_mpls_context         *pContext = arg_pContext;
    
    /* Verify the input arguments */
    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
           ("_ix_cc_mpls_nhlfe_inc_ref_count(): Null MPLS context argument"));

    if (pContext->pCcNhlfeBaseAddr == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
           ("_ix_cc_mpls_nhlfe_inc_ref_count(): Null CC NHLFE table address"));

    if (arg_NhlfeHandle.bladeId != pContext->localBladeId)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE, 
           ("_ix_cc_mpls_nhlfe_inc_ref_count(): wrong bladeId"));

    /* Check whether this is a valid entry */
    pCcNhlfeEntry = &pContext->pCcNhlfeBaseAddr[arg_NhlfeHandle.outSegId];
    
    if  ((pCcNhlfeEntry->flags != MPLS_CC_NHLFE_FLAG_NHLFE) && 
         (pCcNhlfeEntry->flags != MPLS_CC_NHLFE_FLAG_NHLFE_SET) &&
 	 (pCcNhlfeEntry->flags != MPLS_CC_NHLFE_FLAG_NHLFE_USED_IN_SET))

         return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND, 
           ("_ix_cc_mpls_nhlfe_inc_ref_count(): specified NHLFE does not exist"));
   
    pCcNhlfeEntry->refCount ++;
    
    return IX_SUCCESS;    
} /* _ix_cc_mpls_nhlfe_inc_ref_count */




/*
 * FUNCTION: _ix_cc_mpls_nhlfe_dec_ref_count()  */
/**
 * ABSTRACT: decreases ref count of a given CC NHLFE entry
 *
 * \param  arg_NhlfeHandle - handle to NHLFE entry which ref count will be decreased
 * \param  arg_pContext    - pointer to the MPLS CC context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * 
 * \n Error codes:
 * \n       IX_CC_ERROR_NULL                 - invalid pointer
 * \n       IX_CC_MPLS_ERROR_INVALID_HANDLE  - NHLFE handle is invalid
 * \n       IX_CC_ERROR_ENTRY_NOT_FOUND      - NHLFE doed not exist
 */
ix_error
_ix_cc_mpls_nhlfe_dec_ref_count(ix_cc_mpls_nhlfe_handle  arg_NhlfeHandle,
                                void                    *arg_pContext)
{
    ix_cc_mpls_cc_nhlfe        *pCcNhlfeEntry;
    ix_cc_mpls_context         *pContext = arg_pContext;
    
    /* Verify the input arguments */
    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
           ("_ix_cc_mpls_nhlfe_dec_ref_count(): Null MPLS context argument"));

    if (pContext->pCcNhlfeBaseAddr == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
           ("_ix_cc_mpls_nhlfe_dec_ref_count(): Null CC NHLFE table address"));

    if (arg_NhlfeHandle.bladeId != pContext->localBladeId)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE, 
           ("_ix_cc_mpls_nhlfe_dec_ref_count(): wrong bladeId"));

    /* Check whether this is a valid entry */
    pCcNhlfeEntry = &pContext->pCcNhlfeBaseAddr[arg_NhlfeHandle.outSegId];
    
    if  ((pCcNhlfeEntry->flags != MPLS_CC_NHLFE_FLAG_NHLFE) && 
         (pCcNhlfeEntry->flags != MPLS_CC_NHLFE_FLAG_NHLFE_SET) &&
	 (pCcNhlfeEntry->flags != MPLS_CC_NHLFE_FLAG_NHLFE_USED_IN_SET))
        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND, 
           ("_ix_cc_mpls_nhlfe_dec_ref_count(): specified NHLFE does not exist"));
   
    if  (pCcNhlfeEntry->refCount  == 0) 
        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND, 
           ("_ix_cc_mpls_nhlfe_dec_ref_count(): refcount of this NHLFE entry is allready zero"));

    pCcNhlfeEntry->refCount --;
    
    return IX_SUCCESS;    
} /* _ix_cc_mpls_nhlfe_dec_ref_count */





/*
 * FUNCTION: _ix_cc_mpls_nhlfe_add_lsp()  */
/**
 * ABSTRACT: Add LSP pointer to the list of CC NHLFE structure
 *
 * \param  arg_pCcLsp      - pointer to the CC LSP structure
 * \param  arg_pContext    - pointer to the MPLS CC context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * 
 * \n Error codes:
 * \n       IX_CC_ERROR_NULL            - invalid pointer
 * \n       IX_CC_MPLS_ERROR_INVALID_HANDLE - invalid MPLS handle
 */
ix_error
_ix_cc_mpls_nhlfe_add_lsp(ix_cc_mpls_cc_lsp *arg_pCcLsp,
                          void              *arg_pContext)
{
    ix_cc_mpls_cc_nhlfe        *pCcNhlfeEntry;
    ix_cc_mpls_context         *pContext = arg_pContext;
    ix_cc_mpls_lsp_list_ptr    *pLspElem;
    ix_uint8                    fec_on_list = 0;
        
    /* Verify the input arguments */
    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
           ("_ix_cc_mpls_nhlfe_add_lsp(): Null MPLS context argument"));

    if (pContext->pCcNhlfeBaseAddr == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
           ("_ix_cc_mpls_nhlfe_add_lsp(): Null CC NHLFE table address"));
    
    if (arg_pCcLsp == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
           ("_ix_cc_mpls_nhlfe_add_lsp(): Null CC LSP entry"));

    if (arg_pCcLsp->NhlfeHandle.bladeId != pContext->localBladeId)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE, 
           ("_ix_cc_mpls_nhlfe_add_lsp(): Wrong bladeId"));

    /* Find free entry on the list */
    pCcNhlfeEntry = &pContext->pCcNhlfeBaseAddr[arg_pCcLsp->NhlfeHandle.outSegId];

    for (pLspElem = &pCcNhlfeEntry->lspList;
         *pLspElem != NULL;
         pLspElem = &(*pLspElem)->pNextElem)
    {
        if ((*pLspElem)->pCcLsp->type == IX_CC_MPLS_LSP_FEC_TYPE)
            fec_on_list++;
    }

    if (fec_on_list > 1)
        return IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL, 
           ("_ix_cc_mpls_nhlfe_add_lsp(): %d FEC entries on the list", fec_on_list));

    if (fec_on_list == 1  &&  arg_pCcLsp->type == IX_CC_MPLS_LSP_FEC_TYPE)
        return IX_ERROR_WARNING(IX_CC_ERROR_DUPLICATE_ENTRY, 
           ("_ix_cc_mpls_nhlfe_add_lsp(): NHLFE already used by FEC"));
    
    /* Add the LSP to the list */
    *pLspElem = ix_ossl_malloc(sizeof(struct ix_s_cc_mpls_lsp_list));
    (*pLspElem)->pCcLsp = arg_pCcLsp;
    (*pLspElem)->pNextElem = NULL;
    
    return IX_SUCCESS;
} /* _ix_cc_mpls_nhlfe_add_lsp */





/*
 * FUNCTION: _ix_cc_mpls_nhlfe_del_lsp()  */
/**
 * ABSTRACT: Remove LSP pointer from the list of CC NHLFE structure
 *
 * \param  arg_pCcLsp      - pointer to the CC LSP structure
 * \param  arg_pContext    - pointer to the MPLS CC context
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * 
 * \n Error codes:
 * \n       IX_CC_ERROR_NULL            - invalid pointer
 * \n       IX_CC_MPLS_ERROR_INVALID_HANDLE - invalid MPLS handle
 */
ix_error
_ix_cc_mpls_nhlfe_del_lsp(ix_cc_mpls_cc_lsp *arg_pCcLsp,
                          void              *arg_pContext)
{
    ix_cc_mpls_cc_nhlfe        *pCcNhlfeEntry;
    ix_cc_mpls_context         *pContext = arg_pContext;
    ix_cc_mpls_lsp_list_ptr     lspElem, prevLspElem;
        
    /* Verify the input arguments */
    if (arg_pContext == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, 
           ("_ix_cc_mpls_nhlfe_add_lsp(): Null MPLS context argument"));

    if (pContext->pCcNhlfeBaseAddr == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
           ("_ix_cc_mpls_nhlfe_add_lsp(): Null CC NHLFE table address"));
    
    if (arg_pCcLsp == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
           ("_ix_cc_mpls_nhlfe_add_lsp(): Null CC LSP entry"));

    if (arg_pCcLsp->NhlfeHandle.bladeId != pContext->localBladeId)
        return IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_HANDLE, 
           ("_ix_cc_mpls_nhlfe_add_lsp(): Wrong bladeId"));

    /* Check whether this is a valid entry */
    pCcNhlfeEntry = &pContext->pCcNhlfeBaseAddr[arg_pCcLsp->NhlfeHandle.outSegId];

    if (pCcNhlfeEntry == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
           ("_ix_cc_mpls_nhlfe_add_lsp(): Null CC NHLFE entry"));
    
    if (pCcNhlfeEntry->lspList == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
           ("_ix_cc_mpls_nhlfe_add_lsp(): LSP list empty"));

    /* Find LSP on the list */
    for (lspElem = pCcNhlfeEntry->lspList, prevLspElem = NULL;
         lspElem != NULL;
         prevLspElem = lspElem, lspElem = lspElem->pNextElem)
    {
        /* LSP found on the list */
        if (lspElem->pCcLsp == arg_pCcLsp)
        {
            /* remove LSP from the list */
            if (prevLspElem != NULL)
                prevLspElem->pNextElem = lspElem->pNextElem;
            if (pCcNhlfeEntry->lspList == lspElem)
                pCcNhlfeEntry->lspList = lspElem->pNextElem;

            ix_ossl_free(lspElem);
            
            return IX_SUCCESS;
        }
    }

    /* LSP is not on the list */
    return IX_CC_ERROR_ENTRY_NOT_FOUND;
} /* _ix_cc_mpls_nhlfe_del_lsp */





/*
 * FUNCTION: _ix_cc_mpls_nhlfe_get_static_data()  */
/**
 * ABSTRACT: reads NHLFE-relevant data from MPLS configuration data external resource
 *
 * DESCRIPTION: This function retrieves all NHLFE-relevant static data from
 *              either system repository or ix_cc_mpls.h and store them
 *              into the corresponding field in the control block
 *              structure passed in.
 *
 * \param  arg_pContext  - pointer to component context memory
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 */
ix_error
_ix_cc_mpls_nhlfe_get_static_data(void *arg_pContext)
{

    ix_cc_mpls_context        *pContext = (ix_cc_mpls_context*)arg_pContext;

#ifdef IX_INCLUDE_REGISTRY

    ix_configuration_property_handle hProperty;

    IX_ERROR_CR(ix_cc_reg_get_prop_from_path(IX_MPLS_PROP_NHLFE_ROOT, 
                                             &hProperty));


    /* get the property IX_MPLS_PROP_NHLFE_TABLE_MEM_TYPE */
    IX_ERROR_CRT(_ix_cc_mpls_prop_get_uint_value(hProperty,
                                                 IX_MPLS_PROP_NHLFE_TABLE_MEM_TYPE,
                                                 (ix_uint32*)&(pContext->nhlfeTableMemoryType)),
                 IX_CC_ERROR_ENTRY_NOT_FOUND,
                 IX_ERROR_LEVEL_REMOTE);


    /* get the property IX_MPLS_PROP_NHLFE_TABLE_MEM_CHANNEL */
    IX_ERROR_CRT(_ix_cc_mpls_prop_get_uint_value(hProperty,
                                                 IX_MPLS_PROP_NHLFE_TABLE_MEM_CHANNEL,
                                                 &(pContext->nhlfeTableMemoryChannel)),
                 IX_CC_ERROR_ENTRY_NOT_FOUND,
                 IX_ERROR_LEVEL_REMOTE);


    /* get the property IX_MPLS_PROP_MAX_NHLFE_ENTRIES */
    IX_ERROR_CRT(_ix_cc_mpls_prop_get_uint_value(hProperty,
                                                 IX_MPLS_PROP_MAX_NHLFE_ENTRIES,
                                                 &(pContext->maxNhlfeEntries)),
                 IX_CC_ERROR_ENTRY_NOT_FOUND,
                 IX_ERROR_LEVEL_REMOTE);


#ifdef MPLS_COUNTERS_FEATURE

    /* get the property IX_MPLS_PROP_NHLFE_COUNTERS_MEM_CHANNEL */
    IX_ERROR_CRT(_ix_cc_mpls_prop_get_uint_value(hProperty,
                                                 IX_MPLS_PROP_NHLFE_COUNTERS_MEM_CHANNEL,
                                                 &(pContext->nhlfeCountersMemoryChannel)),
                 IX_CC_ERROR_ENTRY_NOT_FOUND,
                 IX_ERROR_LEVEL_REMOTE);

#endif /* MPLS_COUNTERS_FEATURE */

    
#else /* IX_INCLUDE_REGISTRY */

    pContext->nhlfeTableMemoryType       = MPLS_NHLFE_TABLES_MEM_TYPE;
    pContext->nhlfeTableMemoryChannel    = MPLS_NHLFE_TABLES_MEM_CHANNEL;
    pContext->maxNhlfeEntries            = MPLS_NHLFE_MAX_NHLFE_ENTRIES;
    pContext->nhlfeCountersMemoryChannel = MPLS_NHLFE_COUNTERS_MEM_CHANNEL;

#endif /* IX_INCLUDE_REGISTRY */

    return IX_SUCCESS;
} /* _ix_cc_mpls_nhlfe_get_static_data */








