/*
 * ============================================================================
 * COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *                                                                          */
/** \file      ix_cc_mpls_pkt_handler.c
 *
 *  \brief     The file contains implementation of CC packet handler
 *
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_mpls_pkt_handler.c,v 1.12.2.1 2004/03/12 10:40:16 wneubaue Exp $";


#include <ix_types.h>
#include <ix_error.h>

#include <ix_ossl.h>
#include <ix_rm.h>

#if !defined(IX_EXCLUDE_CCI)
#include <ix_cci.h>
#endif 

#include <ix_cc_error.h>
#include <ix_cc.h>

#include <bindings.h>

#include <cc/ix_cc_msup.h>

#include <cc/ix_cc_rtmv4.h>

#include "../../include/mpls_defs.h"
#include "cc/ix_cc_mpls.h"
#include "cc/internal/ix_cc_mpls_lsp.h"
#include "cc/internal/ix_cc_mpls_nhlfe.h"
#include "cc/internal/ix_cc_mpls_pkt_handler.h"
#include "cc/internal/ix_cc_mpls_internal.h"

#include "cc/internal/ix_cc_mpls_icmp.h"
#include "cc/internal/ix_cc_mpls_fwder.h"

/*
 * IPv4-related include files required
 */
#include "cc/ipv4/internal/header_defs.h"
#include "cc/ipv4/internal/icmp.h"
#include "cc/ipv4/internal/utils.h"
#include "cc/ipv4/internal/fragmentation_support.h"



/*
 * FUNCTION: ix_cc_mpls_microblock_pkt_handler()  */
/**
 * ABSTRACT: microblock packet handler function for MPLS Core Component. 
 *
 * DESCRIPTION: This is the registered function to receive exception 
 *              packets from the ILM and FTN Forwarder microblocks.
 *              This function internally calls different functions and
 *              performs various operations based on the exception code
 *              for a given packet.
 *
 * \param  arg_hBuffer       - handle to a buffer which contains exception 
 *                             packets from the ILM and FTN Forwarder microblocks.
 * \param  arg_ExceptionCode - Exception codes  generated by the ILM and FTN Forwarder
 *                             microblocks
 * \param  arg_pContext      - pointer to the MPLS core component context that will 
 *                             be passed to the core component when a packet arrives.
 *                             This context was defined by the core component and 
 *                             passed to core components infrastructure through 
 *                             the ix_cc_mpls_init() function.
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 *         IX_CC_ERROR_INVALID_CODE    - unsupported exception code
 *         IX_CC_ERROR_INVALID_BUFFER  - buffer handle is unusable
 *         IX_CC_ERROR_INVALID_CONTEXT - context is unusable
 *         IX_CC_ERROR_NULL            - invalid pointer
 */
ix_error
ix_cc_mpls_microblock_pkt_handler(ix_buffer_handle arg_hBuffer,
                                  ix_uint32        arg_ExceptionCode,
                                  void            *arg_pContext)
{

  ix_error        err   = IX_SUCCESS;
 
#if defined (IX_MPLS_DEBUG)
  ix_ossl_message_log("ix_cc_mpls_microblock_pkt_handler(): got sth\n");
#endif

  switch (arg_ExceptionCode)
  {
    case ILM_LABELED_PACKET_TOO_BIG:

        #if defined (IX_MPLS_DEBUG)
        ix_ossl_message_log ("ix_cc_mpls_microblock_pkt_handler(): "
                             "ILM_LABELED_PACKET_TOO_BIG exception received\n");
        #endif

        err = _ix_cc_mpls_ilm_fragmentation (arg_hBuffer,
                                             IX_CC_MPLS_FRAGMENT_OUTPUT,
                                             IX_CC_MPLS_FRAGMENT_OUTPUT_BID,
                                             (ix_cc_mpls_context *) arg_pContext);

        break;

    case FTN_LABELED_PACKET_TOO_BIG:

        #if defined (IX_MPLS_DEBUG)
        ix_ossl_message_log ("ix_cc_mpls_microblock_pkt_handler(): "
                             "FTN_LABELED_PACKET_TOO_BIG exception received\n");
        #endif


        err = _ix_cc_mpls_ftn_fragmentation (arg_hBuffer,
                                             IX_CC_MPLS_FRAGMENT_OUTPUT,
                                             IX_CC_MPLS_FRAGMENT_OUTPUT_BID,
                                             (ix_cc_mpls_context *) arg_pContext);

        break;

    case TTL_EXPIRED:

        #if defined (IX_MPLS_DEBUG)
        ix_ossl_message_log ("ix_cc_mpls_microblock_pkt_handler(): "
                             "TTL_EXPIRED exception received\n");
        #endif

        err = _ix_cc_mpls_icmp_time_exceeded(arg_hBuffer,
                                            (ix_cc_mpls_context *) arg_pContext);

        break;

    case TOO_MANY_POPS:
        {
            void                      *pData     = NULL;
            ix_cc_mpls_label_internal  MplsLabel;
 
            #if defined (IX_MPLS_DEBUG)
            ix_ossl_message_log ("ix_cc_mpls_microblock_pkt_handler(): "
                                 "TOO_MANY_POPS exception received\n");
            #endif

           /* Get data from buffer */
            err = ix_cc_hw_get_packet_data (arg_hBuffer, &pData);
            if (err != IX_SUCCESS) 
            {
                err = IX_ERROR_NEW (IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
            }

            ix_ossl_memcpy(&MplsLabel, pData, sizeof(ix_cc_mpls_label_internal));

            /* after 4 pops in microcode we have to check ttl
             * of the fifth label (incidentally for pop operation
             * nhlfe.ttl must be always 1
             */
            if (MplsLabel.ttl <= 1)
            {
                err = _ix_cc_mpls_icmp_time_exceeded(arg_hBuffer,
                                                     arg_pContext);
                if (err != IX_SUCCESS)
                {
                    ix_rm_buffer_free_chain(arg_hBuffer);
                }
                return IX_SUCCESS;
            }

            err = _ix_cc_mpls_ilm_forwarder(arg_hBuffer, IX_CC_MPLS_SEND_ICMP,
                                            (ix_cc_mpls_context *) arg_pContext);

            break;
        }
    case ROUTER_ALERT:

    default:

        err = IX_ERROR_WARNING(IX_CC_MPLS_ERROR_INVALID_EXCEP_CODE,
  	                       ("ix_cc_mpls_microblock_pkt_handler(): "
                               "Unknown exception code received from microblock."));

        break;

    }	/* end switch */


    if (err != IX_SUCCESS) 
    {
      ix_rm_buffer_free_chain(arg_hBuffer);
    }

    return err;

}	







/*
 * FUNCTION: ix_cc_mpls_ilm_pkt_handler()  */
/**
 * ABSTRACT: packet handler function for MPLS Core Component. 
 *
 * DESCRIPTION: This is the registered function to receive  
 *              packets from other CC. This function forwards MPLS packets
 *              in the same way as microblock 
 *
 * \param  arg_hBuffer       - handle to a buffer which contains exception 
 *                             packets from the ILM and FTN Forwarder microblocks.
 * \param  arg_ExceptionCode - Exception codes  generated by the ILM and FTN Forwarder
 *                             microblocks
 * \param  arg_pContext      - pointer to the MPLS core component context that will 
 *                             be passed to the core component when a packet arrives.
 *                             This context was defined by the core component and 
 *                             passed to core components infrastructure through 
 *                             the ix_cc_mpls_init() function.
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 *         IX_CC_ERROR_INVALID_BUFFER  - buffer handle is unusable
 *         IX_CC_ERROR_INVALID_CONTEXT - context is unusable
 *         IX_CC_ERROR_NULL            - invalid pointer
 */
ix_error
ix_cc_mpls_ilm_pkt_handler(ix_buffer_handle arg_hBuffer,
                           ix_uint32        arg_ExceptionCode,
                           void            *arg_pContext)
{

    ix_error        err = IX_SUCCESS;

#if defined (IX_MPLS_DEBUG)
    ix_ossl_message_log("ix_cc_mpls_ilm_pkt_handler(): got sth\n");
#endif

    err = _ix_cc_mpls_ilm_forwarder(arg_hBuffer,
                                    IX_CC_MPLS_SEND_ICMP,
                                    (ix_cc_mpls_context *) arg_pContext);

    if (err != IX_SUCCESS)
    {
        ix_rm_buffer_free_chain (arg_hBuffer);
    }

    return err;
}	





/*
 * FUNCTION: ix_cc_mpls_ftn_pkt_handler()  */
/**
 * ABSTRACT: packet handler function for MPLS Core Component. 
 *
 * DESCRIPTION: This is the registered function to receive  
 *              packets from other CC. This function forwards MPLS packets
 *              in the same way as microblock 
 *
 * \param  arg_hBuffer       - handle to a buffer which contains exception 
 *                             packets from the ILM and FTN Forwarder microblocks.
 * \param  arg_ExceptionCode - Exception codes  generated by the ILM and FTN Forwarder
 *                             microblocks
 * \param  arg_pContext      - pointer to the MPLS core component context that will 
 *                             be passed to the core component when a packet arrives.
 *                             This context was defined by the core component and 
 *                             passed to core components infrastructure through 
 *                             the ix_cc_mpls_init() function.
 *
 * \return IX_SUCCESS if the operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 *         IX_CC_ERROR_INVALID_BUFFER  - buffer handle is unusable
 *         IX_CC_ERROR_INVALID_CONTEXT - context is unusable
 *         IX_CC_ERROR_NULL            - invalid pointer
 */
ix_error
ix_cc_mpls_ftn_pkt_handler(ix_buffer_handle arg_hBuffer,
                           ix_uint32        arg_ExceptionCode,
                           void            *arg_pContext)
{

    ix_error        err = IX_SUCCESS;

#if defined (IX_MPLS_DEBUG)
    ix_ossl_message_log("ix_cc_mpls_ftn_pkt_handler(): got packet\n");
#endif

    err = _ix_cc_mpls_ftn_forwarder(arg_hBuffer,
                                    IX_CC_MPLS_SEND_ICMP,
                                   (ix_cc_mpls_context *) arg_pContext);
    if (err != IX_SUCCESS) 
    {
        ix_rm_buffer_free_chain (arg_hBuffer);
    }

    return err;
}	

