/*
 * ============================================================================
 * COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *                                                                           */
/** \file    ix_cc_classifier_6t.h
 *
 *  \brief   The file contains definitions of common functions that are exposed 
 *           by all core components(i.e. init, fini, etc.)
 */
#if !defined(IX_CC_CLASSIFIER_6T_H)
#define      IX_CC_CLASSIFIER_6T_H

#if defined(__cplusplus)
extern "C"{
#endif

#include "cc/ix_cc_classifier_6t_stats.h"
#include "cc/ix_cc_classifier_6t_api.h"
#include "cc/ix_cc_classifier_6t_hash.h"

/* defines */

#define LEVEL1 			 /* level 1 support */
#define LEVEL0  		 /* level 0 support */

/* temp defines for global CC constants */
#define IX_IMPORTED_SYMBOL_NAME_SIZE    128
#define IX_CCI_ERR_SEND_FAIL            0x1000


#if defined(CLASSIFIER_6T_DRAM_HASH_TABLE)
#define IX_CC_CLASSIFIER_6T_HASH_TABLE_MEM_TYPE     IX_MEMORY_TYPE_DRAM
#else
#define IX_CC_CLASSIFIER_6T_HASH_TABLE_MEM_TYPE     IX_MEMORY_TYPE_SRAM
#endif


/*! number of 6-tuple classifier symbols to patch. */
#if !defined(CLASSIFIER_6T_PACKET_COUNTER_FEATURE)
    #define IX_CC_CLASSIFIER_6T_PATCH_SYMBOLS               6
#else
    #define IX_CC_CLASSIFIER_6T_PATCH_SYMBOLS               8
#endif

/*! size of Hash Table index in bits (used to mask out result of hash operation)
 *  config data set if the system registry is not included */
#define IX_CC_CLASSIFIER_6T_HASH_INDEX_SIZE				0x08


 /*! channel in which hash table is to be hols
 *  config data set if the system registry is not included */
#define IX_CC_CLASSIFIER_6T_HASH_TABLE_CHANNEL_NUM      0x00


#if defined(CLASSIFIER_6T_PACKET_COUNTER_FEATURE)
/*!
 *  channel in which hash table is to be hols
 *  config data set if the system registry is not included
 */
#define IX_CC_CLASSIFIER_6T_STATS_TABLE_CHANNEL_NUM      0x00
#endif


/*! Size (in number of entries) of auxiliary hash table. 
 *  config data set if the system registry is not included */
#define IX_CC_CLASSIFIER_6T_AUX_HASH_TABLE_SIZE         0x20
/*! defines color which the packet is marked when 
 *  the default rule is applied                      */
#define IX_CC_CLASSIFIER_6T_DEFAULT_COLOR_ID            GREEN_PACKET
/*! defines flow ID for the default rule             */
#define IX_CC_CLASSIFIER_6T_DEFAULT_FLOW_ID             0x01
/*! defines class ID for the default rule            */
#define IX_CC_CLASSIFIER_6T_DEFAULT_CLASS_ID            0x08
/*! defines next hop ID for the default rule         */
#define IX_CC_CLASSIFIER_6T_DEFAULT_NEXTHOP_ID          0xFFFF
/*! defines next hop ID type for the default rule    */
#define IX_CC_CLASSIFIER_6T_DEFAULT_NEXTHOP_ID_TYPE		0x00
/*! defines next block for packets the default rule    */


enum ix_e_cc_classifier_6t_messages
{
    IX_CC_CLASSIFIER_6T_MSG_ADD_ENTRY      = IX_CC_COMMON_MSG_ID_LAST,
    IX_CC_CLASSIFIER_6T_MSG_REMOVE_ENTRY,
    IX_CC_CLASSIFIER_6T_MSG_UPDATE_ENTRY,
    IX_CC_CLASSIFIER_6T_MSG_SEARCH_ENTRY

#if defined(CLASSIFIER_6T_PACKET_COUNTER_FEATURE)
    ,                                        
    IX_CC_CLASSIFIER_6T_MSG_GET_STATISTICS,
    IX_CC_CLASSIFIER_6T_MSG_GET_STATISTICS_DEF
#endif
};


/* global setting defines - probably belong to the global header or registry */
#define INGRESS 0
#define EGRESS  1
#define IX_SWITCH_FABRIC_NO_LOOPBACK    0
#define IX_SWITCH_FABRIC_LOOPBACK       1


/* TYPENAME:  ix_cc_classifier_6t_ublock_config */
/**
 * ABSTRACT:  Data structure of 6-tuple classifier core.
 *
 * DESCRIPTION: Classifier core componenet uses this structure
 *              to hold configuration of the microblock working
 *              on a data plane.
 */
typedef struct {
    ix_uint32 	HashIndexSize;
    ix_uint32	HashTableBase;
    ix_uint32   HashTableMemoryType;
#if defined(CLASSIFIER_6T_PACKET_COUNTER_FEATURE)
    ix_uint32   StatsTableBase;
    ix_uint32   DefaultStatAddr;
#endif /* #if defined(CLASSIFIER_6T_PACKET_COUNTER_FEATURE) */
    ix_uint32	DefaultFlowID;
    ix_uint16	DefaultClassID;
    ix_uint16	DefaultNextHopID;
    ix_uint8	DefaultNextHopIDType:4, DefaultColorID:4;
    ix_uint32	_uEngMask;
} ix_cc_classifier_6t_ublock_config;





/* TYPENAME: ix_cc_classifier_6t_context  */
/** 
 * ABSTRACT: Data structure of 6-tuple classifier context.
 *
 * DESCRIPTION: Structure used by an execution engine to hold context 
 *              information for 6-tuple classifier core componenet. 
 *
 */
typedef struct {
    /*! id of the blade that the classifier is running on */
    ix_uint8                              localBladeId;
    /*! indicates ingress (0) or egress (1) side we are running on */
    ix_uint8                              IngressEgress;
    /*! pointer to the microblock configuration structure */
    ix_cc_classifier_6t_ublock_config     *pUblockConfig;
    /*! Pointer at Hash Table. It is shared between core componenet and micro blocks */
    _ix_cc_classifier_6t_hash_entry       *HashTableAddress;
#if defined(CLASSIFIER_6T_PACKET_COUNTER_FEATURE)
    ix_uint32                             StatsTableChannelNum;
    /*! Pointer to the Statsistics Table. */
    ix_cc_classifier_6t_statistics        *StatsTableAddress;
    ix_cc_classifier_6t_statistics        *DefaultStatAddress;
    ix_uint8                              DefaultStatEnable;
#endif /* #if defined(CLASSIFIER_6T_PACKET_COUNTER_FEATURE) */
    /*! mutex used during critical sections in API calls */
    ix_ossl_mutex_t                       API_mutex;
    /*! handle of 6-tuple classifier produced by the execution engine */	
    ix_cc_handle                          hClassifier6tHandle;
    ix_uint32                             HashTableChannelNum;
    /*! The hash table is organised as one address area logically
     *  divided to two parts. First part (primary table) is addressed by Hash Index
     *  calculated based on 6-tuple from IP header. If two or more
     *  different packets produce to the same Hash Index their respective entries are 
     *  connected into list. The first element is in the primary table 
     *  whereas the remaining elements are in the auxiliary table (occupying area 
     *  beyond space addressed by hash index). This register contains information on
     *  how many vacant entries remains on the auxiliary table */
    ix_uint16                             AuxEntryEmpty;
    /*! This register contains information on
     *  how much room (how many entries) there are reserved on the auxiliary table. */
    ix_uint16                             AuxEntryNo;
    /*! name of the image of machine code to be load on a microengine.  */
} _ix_cc_classifier_6t_context;



/* TYPENAME: ix_cc_classifier_6t_error */
/**
 * ABSTRACT: This enummeration defines errors generated by
 *           6 tuple classifier core component.
 */
enum ix_cc_classifier_6t_error {
    IX_CC_CLASSIFIER_6T_ERROR_FIRST = IX_ERROR_MAKE_GROUP(CC_CLASSIFIER_6T),
	
/*! returned when one of the function's arguments was improper.
    e.g. when type of lookup result is neither IX_CC_CLASSIFIER_6T_FWD nor
    IX_CC_CLASSIFIER_6T_QOS                                                     */
    IX_CC_CLASSIFIER_6T_ERROR_PARAM,

/*! returned after an error in opening a branch of system registry tree.        */
    IX_CC_CLASSIFIER_6T_ERROR_REPOSITORY_OPEN,

/*! returned after an error in reading data from system registry.               */
    IX_CC_CLASSIFIER_6T_ERROR_REPOSITORY_READ,

/*! returned after an error in changing virtual address to physical offset.     */
    IX_CC_CLASSIFIER_6T_ERROR_VIRTUAL_TO_PHYSICAL,

/*! returned after an error in changing virtual address to physical offset.     */
    IX_CC_CLASSIFIER_6T_ERROR_SET_UCODE_FILE,

/*! returned after an error in patching microcode.                              */
    IX_CC_CLASSIFIER_6T_ERROR_PATCH_SYMBOLS,

/*! returned after a failed attempt to free memory resources.                   */
    IX_CC_CLASSIFIER_6T_ERROR_FREE_MEM,

/*! returned after a failed attempt to free buffer.                             */
    IX_CC_CLASSIFIER_6T_ERROR_BUFF_FREE,

/*! returned by function which got invalid buffer handle.                       */
    IX_CC_CLASSIFIER_6T_ERROR_INVALID_BUF_HANDLE,

/*! returned by function which got invalid core componenet handle.              */
    IX_CC_CLASSIFIER_6T_ERROR_INVALID_CC_HANDLE,

/*! returned after the failure of user callback function.                       */
    IX_CC_CLASSIFIER_6T_ERROR_USER_CB,

/*! when performing api calls xscale uses hardware hash unit.
 *  Having called that hash it has to wait for result periodically 
 *  polling its internal register to find out if operation has ben completed
 *  This error occurs when api function having waited time defined by
 *  IX_CC_CLASSIFIER_6T_HASH_TIME_OUT did not get the response from hash unit.  */
    IX_CC_CLASSIFIER_6T_ERROR_HASH_TIME_OUT,

/*! This code denotes is returned when a function would init memory structure
 *  which has allready been initialized                                         */
    IX_CC_ERROR_CLASSIFIER_6T_NOT_NULL,

/*! signifies that it was impossible to create Hash Table of zero size 
 *  because HashIndexSize equals zero    */
    IX_CC_CLASSIFIER_6T_ERROR_HASH_IDX_SIZE,

/*! returned after attempt to destroy Hash Table which has not been created       */
    IX_CC_CLASSIFIER_6T_ERROR_INV_HASH_TABLE, 

/*! returned after attempt to create Hash Table which has already been created    */
    IX_CC_CLASSIFIER_6T_ERROR_HASH_TABLE_EXISTS, 

#if defined(CLASSIFIER_6T_PACKET_COUNTER_FEATURE)
/*! returned after attempt to destroy Statistics Table which has not been created      */
    IX_CC_CLASSIFIER_6T_ERROR_INV_STATS_TABLE,

/*! returned after attempt to create Statistics Table which has already been created   */
    IX_CC_CLASSIFIER_6T_ERROR_STATS_TABLE_EXISTS,

/*! returned after attempt to destroy Default Statistics Entry which has not been created      */
    IX_CC_CLASSIFIER_6T_ERROR_INV_DEFAULT_STAT,

/*! returned after attempt to create Default Statistics Entry which has already been created   */
    IX_CC_CLASSIFIER_6T_ERROR_DEFAULT_STAT_EXISTS,

/*! returned after attempt to create Default Statistics Entry which has already been created   */
    IX_CC_CLASSIFIER_6T_ERROR_STATS_NOT_SUPPORTED,

#endif /* #if defined(CLASSIFIER_6T_PACKET_COUNTER_FEATURE) */

    IX_CC_CLASSIFIER_6T_ERROR_LAST
};

/*
 * TYPENAME: ix_cc_classifier_6t_header     */
/**
 * ABSTRACT: This struct defines the format of IP header 
 *
 */
typedef struct ix_s_classifier_6t_header{

	#if IX_BYTE_ORDER == IX_LITTLE_ENDIAN
	    /*! IP version & Header length in words */	
        ix_uint8    headerLength:4,   version:4;
	#elif IX_BYTE_ORDER == IX_BIG_ENDIAN
		/*! IP version & Header length in words */
		ix_uint8    version:4,        headerLength:4;
	#endif

    /*! Type of service */
	ix_uint8    tos;
    
	/*! Length of packet */
	ix_uint16   totalLength;
    
	/*! Unique ID */
	ix_uint16   id;

	#if IX_BYTE_ORDER == IX_LITTLE_ENDIAN
	/*! Offset in buffer to first IP option */
	ix_uint16   fragOffset:13, flags:3;
	#elif IX_BYTE_ORDER == IX_BIG_ENDIAN
	ix_uint16   flags:3, fragOffset:13;
	#endif

    /*! Time to live (in hops) */
	ix_uint8    ttl;
    /*! Protocol carried in IP packet */
	ix_uint8    protocol;
    /*! One's complement checksum */
	ix_uint16   checksum;  	         
    /*! IPv4 source address           */
	ix_uint32   sourceAddress;       
    /*! IPv4 destination address      */
	ix_uint32   destAddress;
    /*! options of IP header plus TCP/UDP ports (can be 0-44 bytes long) */
	ix_uint8    options[44];

} ix_cc_classifier_6t_header;


/*
 * FUNCTION: ix_cc_classifier_6t_init() */
/**
 * ABSTRACT: the function performs initialization of the 6-tuple core component
 *
 * DESCRIPTION: The function initializes the core component. It should be 
 *              called and returned successfully before any other function 
 *              in the core component can be called. The function:
 *               Allocates RAM for the hash table and initializes 
 *                this table with all zeros. 
 *               Patches 6-tuple microcode with imported variables
 *               Initializes a hardware hash unit with the 128-bit 
 *                multiplicand
 *               Registers a packet handler ix_cc_classifier_6t_pkt_handler()
 *                for inputs: 
 *                o IX_CC_CLASSIFIER_6T_EXEPTION_PKT_INPUT (single source mode)
 *                o IX_CC_CLASSIFIER_6T_LOCAL_PKT_INPUT (single source mode)
 *               Registers a message handler ix_cc_classifier_6t_msg_handler()
 *                for inputs:
 *                o IX_CC_CLASSIFIER_6T_MSG_INPUT (multiple sources mode)
 *               Allocates a control block, and returns a pointer to this 
 *                block in arg_ppContext. 
 * 
 * \param  arg_cc_handle - handle of the 6-tuple classifier 
 *                         given by CCI 
 * \param  arg_ppContext - address to the pointer to the 6-tuple
 *                         classifier Core Context
 *
 * \return IX_SUCCESS if operation has been successful or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_cc_handle is not a valid classifier core componenet handle
 * \n      IX_CC_ERROR_OOM   - there was not enough system memory to create classifier context
 *                             or other data structure
 * \n      any valid ix_error if code fails
 */
ix_error
ix_cc_classifier_6t_init(ix_cc_handle arg_cc_handle, void **arg_ppContext);




/*
 * FUNCTION: ix_cc_classifier_6t_fini()   */
/**
 * ABSTRACT: the function shuts down 6-tuple core component
 *
 * DESCRIPTION: Shut down function invoked by the Execution Engine of the CCI
 * 
 * \param  arg_cc_handle - handle of the 6-tuple classifier given by CCI 
 * \param  arg_ppContext - pointer to the 6-tuple classifier Core Context
 *
 * \return IX_SUCCESS if operation has been successful or a valid error token for failure.
 */
ix_error
ix_cc_classifier_6t_fini(ix_cc_handle arg_cc_handle, void *arg_pContext);



/*
 * FUNCTION: ix_cc_classifier_6t_msg_handler()  */
/**
 * ABSTRACT: message handler function for 6-tuple Core Component. 
 *
 * DESCRIPTION: Message handler function invoked by the Execution Engine of the CCI
 *              6-tuple Core Component receives messages from message helper functions
 *              and internally it calls appropriate library function to process the 
 *              messages. 
 * 
 * \param  arg_hDataToken - handle of the IX buffer - actual message buffer 
 * \param  arg_UserData   - idetifies message id.
 * \param  arg_pContext   - pointer to the context of 6 tuple classifier 
 *                          Core Component created in init function
 *
 * \return IX_SUCCESS if operation has been successful or a valid error token for failure.
 */
ix_error
ix_cc_classifier_6t_msg_handler(ix_buffer_handle arg_Msg, 
                                ix_uint32        arg_MsgId,
                                void            *arg_pContext);



/*
 * FUNCTION: ix_cc_classifier_6t_pkt_handler()            */
/**
 * ABSTRACT: Packet handler function invoked by the Execution Engine of the CCI
 *
 * DESCRIPTION: This is registered function to receive packets from any core component
 *         
 * \param  arg_hDataToken    - handle of the IX buffer - represents packet in transmission 
 * \param  arg_ExceptionCode - exception code  - not used in this function
 * \param  arg_pContext      - pointer to the context of 6-tuple Classifier Core Component 
 *                             created in init function
 *
 * \return IX_SUCCESS if operation has been successful or a valid error token for failure.
 */
ix_error ix_cc_classifier_6t_pkt_handler(ix_buffer_handle arg_hDataToken, 
                                         ix_uint32        arg_ExceptionCode, 
                                         void            *arg_pContext);

#if defined(__cplusplus)
}
#endif

#endif /* end !defined(IX_CC_CLASSIFIER_6T_H) */
