/*
 * ============================================================================
 * COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *                                                                            */
/** \file     ix_cc_classifier_6t_api.h
 *
 *  \brief    The file contains definitions used by API functions of
 *            6 tuple classifier Core Component
 */
#if !defined(IX_CC_CLASSIFIER_6T_API_H)
#define      IX_CC_CLASSIFIER_6T_API_H


/*
 * TYPENAME: ix_cc_classifier_6t_table_type      */
/**
 * ABSTRACT:
 *
 * DESCRIPTION:  This data type defines an exact-match rule classification pattern.
 *               This is a packed structure, passed on as a byte string in
 *               a call to add/update/search/remove entry messages
 *
 */
typedef struct ix_s_cc_classifier_6t_pattern
{
  ix_uint32 ipv4SourceAddress : 32;
  ix_uint32 ipv4DestinationAddress : 32;
  ix_uint32 inputPort : 16,
            dscp      : 6,
            padding   : 2,
            protocol  : 8;
  ix_uint32 l4SourcePort : 16;
  ix_uint32 l4DestinationPort : 16;
} ix_cc_classifier_6t_pattern;

/*
 * TYPENAME: ix_cc_classifier_6t_table_type   */
/**
 * ABSTRACT:
 *
 * DESCRIPTION: A classification pattern can be associated with two lookup results:
 *              QoS and forwarding info. Both QoS and forwarding rules can be
 *              configured independently, typically from different applications.
 *              To facilitate configuration, the core component provides an abstraction
 *              of two logical tables
 */
typedef enum {
	/*! QoS rule                             */
    IX_CC_CLASSIFIER_6T_QOS,
	/*! forwarding rule                      */
    IX_CC_CLASSIFIER_6T_FWD
} ix_cc_classifier_6t_table_type;


/*
 * TYPENAME: ix_cc_classifier_6t_result      */
/**
 * ABSTRACT:
 *
 * DESCRIPTION: QoS lookup result
 *
 */
typedef struct ix_s_cc_classifier_6t_qos_result
{
  /*! QoS flow id used to police traffic 	 */
  ix_uint32                   flowId;
  /*! relative identifier of an output queue */
  ix_uint16                   classId;
  /*! packet drop precedence level 	         */
  ix_uint8                    colorId;

  /*! identifier of a QoS microblock. One of
   *   IX_CC_CLASSIFIER_6T_METER_BID  
   *   IX_CC_CLASSIFIER_6T_MARKER_BID
   *   IX_CC_CLASSIFIER_6T_FORWARD_BID
   *   IX_CC_CLASSIFIER_6T_DEFAULT_BID
   */
  ix_uint8                    OutputId;

} ix_cc_classifier_6t_qos_result;

/*
 * TYPENAME: ix_cc_classifier_6t_result      */
/**
 * ABSTRACT:
 *
 */
typedef struct ix_s_cc_classifier_6t_fwd_result
{
	ix_uint16 NextHopId:16;						/* forwarding lookup identifier */
	ix_uint8  NextHopIdType:4, Reserved:4;
} ix_cc_classifier_6t_fwd_result;


/*
 * TYPENAME: ix_cc_classifier_6t_result       */
/**
 * ABSTRACT:
 *
 * DESCRIPTION: the structure that is provided to the API procedures
 *
 */
typedef struct ix_s_cc_classifier_6t_result
{
  ix_cc_classifier_6t_table_type type;
  
  /* indicates whether the microblock should
   * gather statistics for the rule
   */
  #if defined(CLASSIFIER_6T_PACKET_COUNTER_FEATURE)
  ix_uint8  stats_enable; 
  #endif /* #if defined(CLASSIFIER_6T_PACKET_COUNTER_FEATURE) */

  union {
    ix_cc_classifier_6t_qos_result qos;
    ix_cc_classifier_6t_fwd_result fwd;
  } result;

} ix_cc_classifier_6t_result;



/*
 * FUNCTION: ix_cc_classifier_6t_add_entry()   */
/**
 * ABSTRACT:
 *
 * DESCRIPTION: This function adds a new entry in a hash table.
 *              The core component uses the classification pattern,
 *              arg_pKey, to prepare a hash input. Next, it employs
 *              a hardware hash unit to generate a hash table index.
 *              If the hash table already contains a specified
 *              classification pattern, the core component reports
 *              an error. It does not update lookup results associated
 *              with the existing rule. Otherwise, if a hash lookup fails,
 *              the core component adds a new entry to the hash table.
 *              The lookup result, arg_pLookupResult, is stored in
 *              the newly added entry.
 *
 * \param  arg_pKey          - pointer to the classification pattern structure,
 *                             which defines an exact-match rule.
 * \param  arg_pLookupResult - identifies a logical lookup table;
 *                             (IX_CC_CLASSIFIER_6T_QOS or IX_CC_CLASSIFIER_6T_FWD)
 *                             and points at the structure of lookup results
 *                             to be added to the specified classification pattern
 * \param  arg_pContext      - pointer to the core componenet context
 *
 * \return IX_SUCCESS if the entry was successfully added or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_DUPLICATE_ENTRY - a classification pattern is already configured
 * \n      IX_CC_CLASSIFIER_6T_ERROR_TIME_OUT - time out during hash index computation
 *         or other valid ix_error code.
 */
ix_error
ix_cc_classifier_6t_add_entry(ix_cc_classifier_6t_pattern    *arg_pKey,
                              ix_cc_classifier_6t_result     *arg_pLookupResult,
							  void *arg_pContext);


/*
 * FUNCTION: ix_error ix_cc_classifier_6t_remove_entry()       */
/**
 * ABSTRACT:
 *
 * DESCRIPTION: This function removes an entry from the hash table. The core
 *              component uses the classification pattern, arg_pKey, to prepare
 *              a hash input. Next, it employs a hardware hash unit to generate
 *              a hash table index. If the hash table already contains a specified
 *              classification pattern, the core component removes an entry.
 *              Otherwise, if a hash lookup fails, the core component reports an
 *              error and no entry is removed from the hash table.
 *
 * \param  arg_pKey     - pointer to the classification pattern structure,
 *                        which defines an exact-match rule.
 * \param  arg_Type     - identifier of a logical lookup table;
 *                        it should be IX_CC_CLASSIFIER_6T_QOS or IX_CC_CLASSIFIER_6T_FWD
 * \param  arg_pContext - pointer to the core componenet context
 *
 * \return IX_SUCCESS if entry successfully removed or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n       IX_CC_ERROR_NULL   - arg_pContext is not valid classifier core component context.
 * \n       IX_CC_ERROR_ENTRY_NOT_FOUND  - a classification pattern does not exist
 * \n       IX_CC_CLASSIFIER_6T_ERROR_TIME_OUT - time out during hash index computation
 *          or other valid ix_error code.
 */
ix_error
ix_cc_classifier_6t_remove_entry(ix_cc_classifier_6t_pattern    *arg_pKey,
                                 ix_cc_classifier_6t_table_type  arg_Type,
								 void *arg_pContext);

/*
 * FUNCTION: ix_error ix_cc_classifier_6t_update_entry()   */
/**
 * ABSTRACT:
 *
 * DESCRIPTION: This function updated lookup results in a hash table entry.
 *              The core component uses the classification pattern, arg_pKey,
 *              to prepare a hash input. Next, it employs a hardware hash unit
 *              to generate a hash table index. If the hash lookup fails
 *              (classification pattern is not configured), the core component
 *              reports an error. On successful lookup, the core component updates
 *              result fields in the hash table entry.
 *
 * \param  arg_pKey          - pointer to the classification pattern structure,
 *                             which defines an exact-match rule.
 * \param  arg_pLookupResult - identifier of a logical lookup table;
 *                             (IX_CC_CLASSIFIER_6T_QOS or IX_CC_CLASSIFIER_6T_FWD )
 *                             and points at the structure of updated lookup
 *                             results to be updated for specified classification pattern.
 * \param  arg_pContext      - pointer to the core componenet context
 *
 * \return IX_SUCCESS if entry was successfully updated or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_ERROR_ENTRY_NOT_FOUND - a classification pattern is not configured
 * \n      IX_CC_CLASSIFIER_6T_ERROR_TIME_OUT - time out during hash index computation
 *         or other valid ix_error code.
 */
ix_error
ix_cc_classifier_6t_update_entry(ix_cc_classifier_6t_pattern  *arg_pKey,
                                 ix_cc_classifier_6t_result   *arg_pLookupResult,
								 void *arg_pContext);



/*
 * FUNCTION: ix_error ix_cc_classifier_6t_search_entry()   */
/**
 * ABSTRACT: This function searches Hash Table for given entry.
 *
 * DESCRIPTION: This function searches Hash Table for given entry.
 *              The core component uses the classification
 *              pattern, arg_pKey, to prepare a hash input. Next, it employs
 *              a hardware hash unit to generate a hash table index.
 *              Then, the algorithm searches through the chain of elements
 *              starting from the entry pointed by Hash index and going by
 *              all entries linked to it.
 *              If the requested classification pattern is found, a lookup result associated
 *              with this pattern is returned. Otherwise, if a hash table
 *              does not contain the classification pattern, the core component
 *              returns error IX_CC_ERROR_ENTRY_NOT_FOUND.
 *
 * \param  arg_pKey          - pointer to the classification pattern structure,
 *                             which defines an exact-match rule.
 * \param  arg_pLookupResult - identifier of a logical lookup table
 *                             (IX_CC_CLASSIFIER_6T_QOS or IX_CC_CLASSIFIER_6T_FWD)
 *                             and a pointer to a placeholder for the lookup
 *                             results associated with the specified classification pattern.
 * \param  arg_pContext      - pointer to the core componenet context
 *
 * \return IX_SUCCESS if lookup was successful  or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pContext is not valid classifier core component context.
 * \n      IX_CC_CLASSIFIER_6T_ERROR_TIME_OUT - time out during hash index computation
 * \n      IX_CC_ERROR_ENTRY_NOT_FOUND - entry does not exist.
 *         or other valid ix_error code.
 */
ix_error
ix_cc_classifier_6t_search_entry(ix_cc_classifier_6t_pattern    *arg_pKey,
                                 ix_cc_classifier_6t_result     *arg_pLookupResult,
								 void *arg_pContext);



/* --------------------------------------------------------------------------------------- */
#if defined(CLASSIFIER_6T_PACKET_COUNTER_FEATURE)
/* --------------------------------------------------------------------------------------- */

/*
 * FUNCTION: ix_error ix_cc_classifier_6t_get_statistics()    */
/**
 * ABSTRACT: This function gets statistics associated with a specified
 *           classification pattern.
 *
 * DESCRIPTION: This function does a lookup in the Hash Table. If the microblock
 *              supports stastistics and the requested classification is found,
 *              statistics associated with this pattern are returned. Otherwise,
 *              the core component returns an error code and the arg_pStatistics
 *              argument is left unaltered.
 *
 * \param  arg_pKey          - pointer to the classification pattern structure,
 *                             which defines an exact-match rule.
 * \param  arg_pStatistics   - pointer to a placeholder for the statistics
 *                             associated with the classification pattern.
 * \param  arg_pContext      - pointer to the core componenet context
 *
 * \return IX_SUCCESS if lookup was successful  or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pKey, arg_pStatistics or arg_pContext argument is NULL.
 * \n      IX_CC_ERROR_ENTRY_NOT_FOUND - entry does not exist.
 * \n      IX_CC_CLASSIFIER_6T_ERROR_STATS_NOT_SUPPORTED - statistics associated with
 * \n                                                      given pattern are not gathered.
 *         or other valid ix_error code.
 */
ix_error
ix_cc_classifier_6t_get_statistics(ix_cc_classifier_6t_pattern    *arg_pKey,
                                   ix_cc_classifier_6t_statistics *arg_pStatistics,
                                   void *arg_pContext);

/*
 * FUNCTION: ix_error ix_cc_classifier_6t_get_statistics_def()    */
/**
 * ABSTRACT: This function gets statistics associated with the default rule.
 *
 * DESCRIPTION: If the microblock supports stastistics the function returns
 *              statistics associated with the default rule-. Otherwise,
 *              the core component returns an error code and the arg_pStatistics
 *              argument is left unaltered.
 *
 * \param  arg_pStatistics   - pointer to a placeholder for the statistics
 *                             associated with the classification pattern.
 * \param  arg_pContext      - pointer to the core componenet context
 *
 * \return IX_SUCCESS if lookup was successful  or a valid error token for failure.
 * \n
 * \n Error codes:
 * \n      IX_CC_ERROR_NULL  - arg_pStatistics or arg_pContext argument is NULL.
 * \n      IX_CC_ERROR_ENTRY_NOT_FOUND - entry does not exist.
 * \n      IX_CC_CLASSIFIER_6T_ERROR_STATS_NOT_SUPPORTED - statistics associated with
 * \n                                                      the default rule are not gathered.
 *         or other valid ix_error code.
 */
ix_error ix_cc_classifier_6t_get_statistics_def(ix_cc_classifier_6t_statistics *arg_pStatistics,
                                                void *arg_pContext);


/* --------------------------------------------------------------------------------------- */
#endif
/* --------------------------------------------------------------------------------------- */

#endif
