/*
 * ============================================================================
 * COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *                                                                            */
/** \file   ix_cc_classifier_6t_hash.c
 * 
 *  \brief  The file contains internal functions used to control hash table
 *          by 6 tuple classifier Core Component
 *
 */

#include "ix_error.h"

#include "ix_ossl.h"
#include "ix_cc.h"
#include "ix_cci.h"

#include "ix_cc_error.h"

#include "bindings.h"
#include "cc/ix_cc_classifier_6t.h"



/*
 * FUNCTION: _ix_cc_classifier_6t_create_hash_table
 */
ix_error
_ix_cc_classifier_6t_create_hash_table(
			ix_uint32   arg_IndexSize,
			ix_uint32   arg_AuxListSize,
			ix_uint32	arg_MemoryType,
			ix_uint32	arg_MemChannel,
			_ix_cc_classifier_6t_hash_entry   **HashTable)
{

	ix_uint32   i;
	ix_error    err;
	_ix_cc_classifier_6t_hash_entry *index;
	
	if (NULL != *HashTable)
		return IX_ERROR_WARNING(IX_CC_CLASSIFIER_6T_ERROR_HASH_TABLE_EXISTS,
		                 ("ix_cc_classifier_6t_create_hash_table(): Error ! Hash Table has already been created."));

	if (arg_IndexSize == 0)
		return IX_ERROR_WARNING(IX_CC_CLASSIFIER_6T_ERROR_HASH_IDX_SIZE,
		                 ("ix_cc_classifier_6t_create_hash_table(): Error ! size of HashIndex = 0. Empty Hash Table cannot be created."));

	/* allocate memory for Hash Table */
    
    err = ix_rm_mem_alloc (arg_MemoryType, arg_MemChannel,
                         (arg_AuxListSize + (1 << arg_IndexSize)) * sizeof(_ix_cc_classifier_6t_hash_entry),
                         (void**)HashTable);
	

	if (err != IX_SUCCESS)
		return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SRAM,
		                 ("ix_cc_classifier_6t_create_hash_table(): Error! Not enough memory to create Hash Table."));

	index = (*HashTable);

	/* Clear all Hash Table */
	for(i=0; i< (arg_AuxListSize + (1 << arg_IndexSize)); i++, index++)
        _ix_cc_classifier_6t_zero_out_hash_entry(index);

	return IX_SUCCESS;
};






/*
 * FUNCTION: _ix_cc_classifier_6t_destroy_hash_table
 */
ix_error
_ix_cc_classifier_6t_destroy_hash_table(_ix_cc_classifier_6t_hash_entry  **HashTable) {

	ix_error err = IX_SUCCESS;

	if (NULL == *HashTable)
		return IX_ERROR_WARNING(IX_CC_CLASSIFIER_6T_ERROR_INV_HASH_TABLE,
		                 ("ix_cc_classifier_6t_destroy_hash_table(): Error ! Hash Table has not been created."));

    err = ix_rm_mem_free((void*)(*HashTable));
	if (err != IX_SUCCESS)
		return IX_ERROR_WARNING(IX_CC_CLASSIFIER_6T_ERROR_FREE_MEM,
		                 ("ix_cc_classifier_6t_destroy_hash_table(): Error ! Cannot free memory resources taken by Hash Table."));

	*HashTable = NULL;

	return err;
};



/*
 * FUNCTION: _ix_cc_classifier_6t_find_free_element 
 */
_ix_cc_classifier_6t_hash_entry*
_ix_cc_classifier_6t_find_free_element(
    _ix_cc_classifier_6t_hash_entry  *FirstElement,
    _ix_cc_classifier_6t_hash_entry  *LastElement)
{
	for ( ; FirstElement < LastElement; FirstElement++) {
		if (_ix_cc_classifier_6t_isEmptyElement(FirstElement))
			return FirstElement;
	}
	return NULL;
}


/*
 * FUNCTION: _ix_cc_classifier_6t_get_hash_entry 
 */
ix_error
_ix_cc_classifier_6t_get_hash_entry(
		       ix_cc_classifier_6t_result      *arg_pLookupResult,
		const _ix_cc_classifier_6t_hash_entry  *HashElement)
{

	switch(arg_pLookupResult->type) {
		case IX_CC_CLASSIFIER_6T_QOS:
            if (_ix_cc_classifier_6t_isEmptyEntry(IX_CC_CLASSIFIER_6T_QOS, HashElement) == _IX_CC_CLASSIFIER_6T_ENTRY_EMPTY)
				return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
			       ("_ix_cc_classifier_6t_get_hash_entry(): Error ! Given entry does not exist."));

			arg_pLookupResult->result.qos.flowId   = HashElement->hash_flow;
			arg_pLookupResult->result.qos.classId  = (ix_uint16)((HashElement->hash_class_color_nextblock >> _IX_CC_CLASSIFIER_6T_CLASS_ID_SHIFT) & _IX_CC_CLASSIFIER_6T_CLASS_ID_MASK);
			arg_pLookupResult->result.qos.colorId  = (ix_uint8) ((HashElement->hash_class_color_nextblock >> _IX_CC_CLASSIFIER_6T_COLOR_ID_SHIFT) & _IX_CC_CLASSIFIER_6T_COLOR_ID_MASK);
                        #if defined(CLASSIFIER_6T_PACKET_COUNTER_FEATURE)
                        arg_pLookupResult->stats_enable        = ((HashElement->hash_class_color_nextblock >> _IX_CC_CLASSIFIER_6T_STATS_ENABLE_SHIFT) & _IX_CC_CLASSIFIER_6T_STATS_ENABLE_MASK);			arg_pLookupResult->result.qos.OutputId = (ix_uint8)  (HashElement->hash_class_color_nextblock & _IX_CC_CLASSIFIER_6T_NEXTBLOCK_ID_MASK);
                        #endif
			break;


		case IX_CC_CLASSIFIER_6T_FWD:
            if (_ix_cc_classifier_6t_isEmptyEntry(IX_CC_CLASSIFIER_6T_FWD, HashElement) == _IX_CC_CLASSIFIER_6T_ENTRY_EMPTY)
				return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
			       ("_ix_cc_classifier_6t_get_hash_entry(): Error ! Given entry does not exist."));

			arg_pLookupResult->result.fwd.NextHopId     = HashElement->hash_nexthop;
			arg_pLookupResult->result.fwd.NextHopIdType = (ix_uint8)((HashElement->hash_class_color_nextblock >>_IX_CC_CLASSIFIER_6T_NEXTHOP_ID_TYPE_SHIFT) & _IX_CC_CLASSIFIER_6T_NEXTHOP_ID_TYPE_MASK);
			break;

		default:
            return IX_ERROR_WARNING(IX_CC_CLASSIFIER_6T_ERROR_PARAM,
		       ("_ix_cc_classifier_6t_get_hash_entry(): Error ! Wrong parameter."));
	};

	return IX_SUCCESS;
}




/*
 * FUNCTION: _ix_cc_classifier_6t_set_hash_entry_lookup_result 
 */
ix_error
_ix_cc_classifier_6t_set_hash_entry_lookup_result(
	 const  ix_cc_classifier_6t_result      *arg_pLookupResult,
	       _ix_cc_classifier_6t_hash_entry  *HashElement)
{
	switch(arg_pLookupResult->type) {

        case IX_CC_CLASSIFIER_6T_QOS:
            if (_ix_cc_classifier_6t_isEmptyEntry(IX_CC_CLASSIFIER_6T_QOS, HashElement) == _IX_CC_CLASSIFIER_6T_ENTRY_EXISTS)
				return IX_ERROR_WARNING(IX_CC_ERROR_DUPLICATE_ENTRY,
			       ("_ix_cc_classifier_6t_set_hash_entry(): Error ! Given entry does allredy aexist."));
            #ifdef IX_CLASSIFIER_6T_DEBUG 
            printf("arg_p_result: \n");
            show_mem(arg_pLookupResult, sizeof(ix_cc_classifier_6t_result));
            printf("arg_p_result.flowid: %d\n",arg_pLookupResult->result.qos.flowId);
            printf("arg_p_result.OutputID: %d\n",arg_pLookupResult->result.qos.OutputId );
	    #endif

			HashElement->hash_flow                   =  arg_pLookupResult->result.qos.flowId;
            #ifdef IX_CLASSIFIER_6T_DEBUG 
            printf(" new flow id : %d \n",HashElement->hash_flow);
            #endif

			HashElement->hash_class_color_nextblock &=  _IX_CC_CLASSIFIER_6T_EMPTY_CLASS_COLOR_NEXTBLOCK_FIELD;
			HashElement->hash_class_color_nextblock |=  ((arg_pLookupResult->result.qos.classId & _IX_CC_CLASSIFIER_6T_CLASS_ID_MASK) << _IX_CC_CLASSIFIER_6T_CLASS_ID_SHIFT)   |
                                                        #if defined(CLASSIFIER_6T_PACKET_COUNTER_FEATURE)
                                                        ((arg_pLookupResult->stats_enable & _IX_CC_CLASSIFIER_6T_STATS_ENABLE_MASK) << _IX_CC_CLASSIFIER_6T_STATS_ENABLE_SHIFT) |
                                                        #endif 
                                                        ((arg_pLookupResult->result.qos.colorId & _IX_CC_CLASSIFIER_6T_COLOR_ID_MASK) << _IX_CC_CLASSIFIER_6T_COLOR_ID_SHIFT)   |
                                                         (arg_pLookupResult->result.qos.OutputId & _IX_CC_CLASSIFIER_6T_NEXTBLOCK_ID_MASK);
            #ifdef IX_CLASSIFIER_6T_DEBUG 
            printf(" new hash_class_color_nextblock : %d \n",HashElement->hash_class_color_nextblock );
            #endif
			break;

		case IX_CC_CLASSIFIER_6T_FWD:
            if (_ix_cc_classifier_6t_isEmptyEntry(IX_CC_CLASSIFIER_6T_FWD, HashElement) == _IX_CC_CLASSIFIER_6T_ENTRY_EXISTS)
				return IX_ERROR_WARNING(IX_CC_ERROR_DUPLICATE_ENTRY,
			       ("_ix_cc_classifier_6t_set_hash_entry(): Error ! Given entry does allready exist."));
            
            HashElement->hash_class_color_nextblock &= _IX_CC_CLASSIFIER_6T_EMPTY_NEXT_HOP_ID_FIELD;

            HashElement->hash_class_color_nextblock |= ((arg_pLookupResult->result.fwd.NextHopIdType & 
                                                         _IX_CC_CLASSIFIER_6T_NEXTHOP_ID_TYPE_MASK)
                                                                                 << _IX_CC_CLASSIFIER_6T_NEXTHOP_ID_TYPE_SHIFT);

            HashElement->hash_nexthop = arg_pLookupResult->result.fwd.NextHopId;

            if (_ix_cc_classifier_6t_isEmptyEntry(IX_CC_CLASSIFIER_6T_QOS, HashElement) == _IX_CC_CLASSIFIER_6T_ENTRY_EMPTY)
            {
              	HashElement->hash_class_color_nextblock &= _IX_CC_CLASSIFIER_6T_EMPTY_NEXTBLOCK_FIELD;
                HashElement->hash_class_color_nextblock |= (IX_CC_CLASSIFIER_6T_FORWARDER_BID & _IX_CC_CLASSIFIER_6T_NEXTBLOCK_ID_MASK);

            }

            break;


        default:
            return IX_ERROR_WARNING(IX_CC_CLASSIFIER_6T_ERROR_PARAM,
                ("_ix_cc_classifier_6t_set_hash_entry(): Error ! Wrong parameter."));

	};

	return IX_SUCCESS;
}

/*
 * FUNCTION: _ix_cc_classifier_6t_clear_hash_entry_lookup_result
 */
ix_error
_ix_cc_classifier_6t_clear_hash_entry_lookup_result(
	      ix_cc_classifier_6t_table_type  arg_Type,
	     _ix_cc_classifier_6t_hash_entry  *HashElement)
{
	switch(arg_Type) {
		case IX_CC_CLASSIFIER_6T_QOS:

            if (_ix_cc_classifier_6t_isEmptyEntry(IX_CC_CLASSIFIER_6T_QOS, HashElement) == _IX_CC_CLASSIFIER_6T_ENTRY_EMPTY)
				return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
			       ("_ix_cc_classifier_6t_clear_hash_entry(): Error ! Given entry does not exist."));

			HashElement->hash_flow                    =  _IX_CC_CLASSIFIER_6T_EMPTY_QOS_ENTRY;
			HashElement->hash_class_color_nextblock  &=  _IX_CC_CLASSIFIER_6T_EMPTY_CLASS_COLOR_NEXTBLOCK_FIELD;
			break;

		case IX_CC_CLASSIFIER_6T_FWD:
            if (_ix_cc_classifier_6t_isEmptyEntry(IX_CC_CLASSIFIER_6T_FWD, HashElement) == _IX_CC_CLASSIFIER_6T_ENTRY_EMPTY)
				return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
			       ("_ix_cc_classifier_6t_clear_hash_entry(): Error ! Given entry does not exist."));

			HashElement->hash_nexthop                = _IX_CC_CLASSIFIER_6T_EMPTY_FWD_ENTRY;
			HashElement->hash_class_color_nextblock &= _IX_CC_CLASSIFIER_6T_EMPTY_NEXT_HOP_ID_FIELD;

            if (_ix_cc_classifier_6t_isEmptyEntry(IX_CC_CLASSIFIER_6T_QOS, HashElement) == _IX_CC_CLASSIFIER_6T_ENTRY_EMPTY)
            {
              	HashElement->hash_class_color_nextblock &= _IX_CC_CLASSIFIER_6T_EMPTY_NEXTBLOCK_FIELD;
            }

            break;

		default:
			return IX_ERROR_WARNING(IX_CC_CLASSIFIER_6T_ERROR_PARAM,
			       ("_ix_cc_classifier_6t_clear_hash_entry(): Error ! Improper lookup type."));

	};

	return IX_SUCCESS;
}

/*
 * FUNCTION: _ix_cc_classifier_6t_zero_out_hash_entry
 */
void
_ix_cc_classifier_6t_zero_out_hash_entry(
	     _ix_cc_classifier_6t_hash_entry  *HashElement)
{
	HashElement->NextHashEntry               = (ix_uint32)NULL_NEXT_ENTRY;
	HashElement->hash_src                    = _IX_CC_CLASSIFIER_6T_EMPTY_FIELD_ENTRY;
	HashElement->hash_dst                    = _IX_CC_CLASSIFIER_6T_EMPTY_FIELD_ENTRY;
	HashElement->hash_ports                  = _IX_CC_CLASSIFIER_6T_EMPTY_FIELD_ENTRY;
	HashElement->hash_port_dscp_protocol     = _IX_CC_CLASSIFIER_6T_EMPTY_FIELD_ENTRY;
	HashElement->hash_flow                   = _IX_CC_CLASSIFIER_6T_EMPTY_QOS_ENTRY;
	HashElement->hash_class_color_nextblock &= _IX_CC_CLASSIFIER_6T_EMPTY_CLASS_COLOR_NEXTBLOCK_FIELD;
	HashElement->hash_class_color_nextblock &= _IX_CC_CLASSIFIER_6T_EMPTY_NEXT_HOP_ID_FIELD;
	HashElement->hash_nexthop                = _IX_CC_CLASSIFIER_6T_EMPTY_FWD_ENTRY;

	return;
}

/*
 * FUNCTION: _ix_cc_classifier_6t_set_hash_entry_pattern
 */
void
_ix_cc_classifier_6t_set_hash_entry_pattern(
				ix_cc_classifier_6t_pattern      *arg_pKey,
				_ix_cc_classifier_6t_hash_entry  *HashElement)
{
	HashElement->hash_src                = arg_pKey->ipv4SourceAddress;
	HashElement->hash_dst                = arg_pKey->ipv4DestinationAddress;
	HashElement->hash_ports              = ((((arg_pKey->l4SourcePort) & _IX_CC_CLASSIFIER_6T_SRC_PORT_MASK) << _IX_CC_CLASSIFIER_6T_SRC_PORT_SHIFT) |
										    ((arg_pKey->l4DestinationPort) & _IX_CC_CLASSIFIER_6T_DST_PORT_MASK));
	HashElement->hash_port_dscp_protocol = (((arg_pKey->inputPort & _IX_CC_CLASSIFIER_6T_INPUT_PORT_MASK) << _IX_CC_CLASSIFIER_6T_INPUT_PORT_SHIFT) |
										    ((arg_pKey->dscp & _IX_CC_CLASSIFIER_6T_DSCP_MASK) << _IX_CC_CLASSIFIER_6T_DSCP_SHIFT) |
											(arg_pKey->protocol & _IX_CC_CLASSIFIER_6T_PROTOCOL_MASK));
	return;
}




/*
 * FUNCTION: _ix_cc_classifier_6t_zero_out_hash_entry_pattern
 */
void
_ix_cc_classifier_6t_zero_out_hash_entry_pattern(
				_ix_cc_classifier_6t_hash_entry  *HashElement)
{
	HashElement->NextHashEntry              = (ix_uint32)NULL_NEXT_ENTRY;
	HashElement->hash_src                   = _IX_CC_CLASSIFIER_6T_EMPTY_FIELD_ENTRY;
	HashElement->hash_dst                   = _IX_CC_CLASSIFIER_6T_EMPTY_FIELD_ENTRY;
	HashElement->hash_ports                 = _IX_CC_CLASSIFIER_6T_EMPTY_FIELD_ENTRY;
	HashElement->hash_port_dscp_protocol    = _IX_CC_CLASSIFIER_6T_EMPTY_FIELD_ENTRY;
	return;
}

/*
 * FUNCTION: _ix_cc_classifier_6t_match_entry_pattern
 */
ix_uint32
_ix_cc_classifier_6t_match_entry_pattern(
				ix_cc_classifier_6t_pattern      *arg_pKey,
				_ix_cc_classifier_6t_hash_entry  *HashElement)
{
	if ((HashElement->hash_src   == arg_pKey->ipv4SourceAddress) &&
	    (HashElement->hash_dst   == arg_pKey->ipv4DestinationAddress) &&
	    (HashElement->hash_ports == (((arg_pKey->l4SourcePort & _IX_CC_CLASSIFIER_6T_SRC_PORT_MASK) << _IX_CC_CLASSIFIER_6T_SRC_PORT_SHIFT)
		                            | (arg_pKey->l4DestinationPort & _IX_CC_CLASSIFIER_6T_DST_PORT_MASK))) &&
	    ((HashElement->hash_port_dscp_protocol & 0xFFFF) ==  (((arg_pKey->dscp & _IX_CC_CLASSIFIER_6T_DSCP_MASK) << _IX_CC_CLASSIFIER_6T_DSCP_SHIFT)
							                                 | (arg_pKey->protocol & _IX_CC_CLASSIFIER_6T_PROTOCOL_MASK))))
	{
		if (((HashElement->hash_port_dscp_protocol >> _IX_CC_CLASSIFIER_6T_INPUT_PORT_SHIFT) 
               & _IX_CC_CLASSIFIER_6T_INPUT_PORT_MASK) == 0xFFFF) 
			return _IX_CC_MATCH;
		if (((HashElement->hash_port_dscp_protocol >> _IX_CC_CLASSIFIER_6T_INPUT_PORT_SHIFT)
               & _IX_CC_CLASSIFIER_6T_INPUT_PORT_MASK) == (arg_pKey->inputPort & _IX_CC_CLASSIFIER_6T_INPUT_PORT_MASK))
			return _IX_CC_MATCH;
	}
	return _IX_CC_DIFFERENT;
}




/*
 * FUNCTION: _ix_cc_classifier_6t_isEmptyElement
 */
ix_uint32
_ix_cc_classifier_6t_isEmptyElement(
	const  _ix_cc_classifier_6t_hash_entry  *HashElement)
{
	if ((_ix_cc_classifier_6t_isEmptyEntry(IX_CC_CLASSIFIER_6T_FWD, HashElement) == _IX_CC_CLASSIFIER_6T_ENTRY_EMPTY) &&
        (_ix_cc_classifier_6t_isEmptyEntry(IX_CC_CLASSIFIER_6T_QOS, HashElement) == _IX_CC_CLASSIFIER_6T_ENTRY_EMPTY))
        return _IX_CC_CLASSIFIER_6T_ENTRY_EMPTY;
    else
	    return _IX_CC_CLASSIFIER_6T_ENTRY_EXISTS;
}


/*
 * FUNCTION: _ix_cc_classifier_6t_isEmptyEntry
 */
ix_uint32
_ix_cc_classifier_6t_isEmptyEntry(ix_cc_classifier_6t_table_type arg_Type,
                                  const _ix_cc_classifier_6t_hash_entry *HashElement)
{ 
    switch(arg_Type) 
    {
        case IX_CC_CLASSIFIER_6T_FWD:
            if (HashElement->hash_nexthop == _IX_CC_CLASSIFIER_6T_EMPTY_FWD_ENTRY)
            {
                return _IX_CC_CLASSIFIER_6T_ENTRY_EMPTY;
            }
            break;                

        case IX_CC_CLASSIFIER_6T_QOS:
            if (HashElement->hash_flow   == _IX_CC_CLASSIFIER_6T_EMPTY_QOS_ENTRY)
            {
                return _IX_CC_CLASSIFIER_6T_ENTRY_EMPTY;
            }
            break;

        default:
            return IX_ERROR_WARNING(IX_CC_CLASSIFIER_6T_ERROR_PARAM,
	           ("_ix_cc_classifier_6t_isEmptyEntry(): Error ! Improper lookup type."));
            break;
    }
    return _IX_CC_CLASSIFIER_6T_ENTRY_EXISTS;
}


