/*
 * ============================================================================
 * COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *                                                                            */
/** \file   ix_cc_classifier_6t_stats.c
 * 
 *  \brief  The file contains internal functions used to manage statistics
 *          by 6 tuple classifier Core Component
 *
 */
#if defined(CLASSIFIER_6T_PACKET_COUNTER_FEATURE)

#include "ix_types.h"
#include "ix_error.h"
#include "ix_ossl.h"
#include "ix_rm.h"

#include "ix_cc.h"
#include "ix_cc_error.h"

#include "cc/ix_cc_classifier_6t.h"
#include "cc/ix_cc_classifier_6t_stats.h"

/*
 * FUNCTION: ix_cc_classifier_6t_create_stats_table
 */
ix_error
ix_cc_classifier_6t_create_stats_table(
			ix_uint32   arg_IndexSize,
			ix_uint32   arg_AuxListSize,
			ix_uint32   arg_MemChannel,
			ix_cc_classifier_6t_statistics   **StatsTable,
			ix_cc_classifier_6t_statistics   **DefaultStat)
{
    ix_error                          err;
    const ix_memory_type              memType = IX_MEMORY_TYPE_SRAM;
    ix_cc_classifier_6t_statistics    *index;
    ix_uint32                         i;

	if (NULL != *StatsTable)
		return IX_ERROR_WARNING(IX_CC_CLASSIFIER_6T_ERROR_STATS_TABLE_EXISTS,
		                 ("ix_cc_classifier_6t_create_stats_table(): Error ! Statistics Table has already been created."));

	if (NULL != *DefaultStat)
		return IX_ERROR_WARNING(IX_CC_CLASSIFIER_6T_ERROR_DEFAULT_STAT_EXISTS,
		                 ("ix_cc_classifier_6t_create_stats_table(): Error ! Default Statistics Table entry has already been created."));

	/* allocate memory for Statistics */
	err = ix_rm_mem_alloc (memType, arg_MemChannel,
                         (arg_AuxListSize + (1 << arg_IndexSize)) * sizeof(ix_cc_classifier_6t_statistics),
                         (void**)StatsTable);
	

	if (err != IX_SUCCESS)
		return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SRAM,
		                 ("ix_cc_classifier_6t_create_stats_table(): Error! Not enough memory to create Statistics Table."));

	index = (*StatsTable);

	/* Clear all Statistics Table entries */
	for(i=0; i< (arg_AuxListSize + (1 << arg_IndexSize) + 1); i++, index++)
	{
		index->packets = 0;
		index->bytes = 0;
	};

	/* allocate memory for Default Statistics Entry */
	err = ix_rm_mem_alloc (memType, arg_MemChannel,
                         sizeof(ix_cc_classifier_6t_statistics),
                         (void**)DefaultStat);

	if (err != IX_SUCCESS)
		return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SRAM,
		                 ("ix_cc_classifier_6t_create_stats_table(): Error! Not enough memory to create Default Statistics Entry."));

	(*DefaultStat)->packets = 0;
	(*DefaultStat)->bytes = 0;

	return IX_SUCCESS;
}


/*
 * FUNCTION: ix_cc_classifier_6t_destroy_stats_table
 */
ix_error
ix_cc_classifier_6t_destroy_stats_table(ix_cc_classifier_6t_statistics  **StatsTable) {

    ix_error err = IX_SUCCESS;

    if (NULL == *StatsTable)
        return IX_ERROR_WARNING(IX_CC_CLASSIFIER_6T_ERROR_INV_STATS_TABLE,
                   ("ix_cc_classifier_6t_destroy_stats_table(): Error ! Statistics Table has not been created."));

    err = ix_rm_mem_free((void*)(*StatsTable));
    if (err != IX_SUCCESS)
        return IX_ERROR_WARNING(IX_CC_CLASSIFIER_6T_ERROR_FREE_MEM,
                   ("ix_cc_classifier_6t_destroy_stats_table(): Error ! Cannot free memory resources taken by Statistics Table."));

    *StatsTable = NULL;

    return err;
};


/*
 * FUNCTION: ix_cc_classifier_6t_destroy_default_stat
 */
ix_error
ix_cc_classifier_6t_destroy_default_stat(ix_cc_classifier_6t_statistics  **DefaultStat) {

    ix_error err = IX_SUCCESS;

    if (NULL == *DefaultStat)
        return IX_ERROR_WARNING(IX_CC_CLASSIFIER_6T_ERROR_INV_DEFAULT_STAT,
                   ("ix_cc_classifier_6t_destroy_default_stat(): Error ! Default Statistics Entry has not been created."));

    err = ix_rm_mem_free((void*)(*DefaultStat));
    if (err != IX_SUCCESS)
        return IX_ERROR_WARNING(IX_CC_CLASSIFIER_6T_ERROR_FREE_MEM,
                   ("ix_cc_classifier_6t_destroy_default_stat(): Error ! Cannot free memory resources taken by Default Statistics Entry."));

    *DefaultStat = NULL;

    return err;
};

#endif
