/*
 * ==========================================================================
 * COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *                                                                          */
/** \file     ix_cc_classifier_6t_utils.c
 *
 *  \brief    The file contains functions that represent internal functionality of
 *            6-tuple Classifier core. The functions in this file are not exposed
 *            to the clients of the classifier, and are not dependant on the CCI
 *
 */

#define IX_ERROR_FILE_IDENT "$Id: ix_cc_classifier_6t_utils.c,v 1.33.6.1 2004/03/12 10:31:44 wneubaue Exp $";

#include "ix_types.h"
#include "ix_error.h"

#include "ix_ossl.h"
#include "ix_rm.h"
#include "ix_cci.h"
#include "ix_cc.h"
#include "ix_cc_error.h"

#include "diffserv.h"
#include "cc/ix_cc_classifier_6t.h"
#include "cc/ix_cc_classifier_6t_utils.h"
#include "../../include/classifier_6t_defs.h"

#include "bindings.h"

/* array of import variable names for the 6-tuple classifier microblock */
static const char sz_aImpVar[][IX_IMPORTED_SYMBOL_NAME_SIZE] =
               {"CLASSIFIER_6T_HASH_TABLE_BASE",
                "CLASSIFIER_6T_DEFAULT_COLOR_ID",
                "CLASSIFIER_6T_DEFAULT_FLOW_ID",
                "CLASSIFIER_6T_DEFAULT_CLASS_ID",
                "CLASSIFIER_6T_DEFAULT_NEXTHOP_ID",
                "CLASSIFIER_6T_DEFAULT_NEXTHOP_ID_TYPE",
                "CLASSIFIER_6T_STATS_SRAM_BASE",
                "CLASSIFIER_6T_DEFAULT_STATS_ADDRESS"};

#if defined(IX_EXCLUDE_CCI)

ix_error
_ix_cc_classifier_6t_reg_pkt_msg_handlers(_ix_cc_classifier_6t_context * pContext)
{

    #ifdef IX_CLASSIFIER_6T_DEBUG
    ix_ossl_message_log("_ix_cc_classifier_6t_reg_pkt_msg_handlers(): CCI excluded ;((((\n");
    #endif

    /* register message handler */
    IX_ERROR_CR( ix_rm_message_handler_register(
        IX_CC_CLASSIFIER_6T_MSG_INPUT,
        ix_cc_classifier_6t_msg_handler,
        pContext )
      );

    /* first registration for the handle of exception packets input */
    IX_ERROR_CR( ix_rm_packet_handler_register(
        IX_CC_CLASSIFIER_6T_LOCAL_PKT_INPUT,
        ix_cc_classifier_6t_pkt_handler,
        pContext )
      );

    /* registration for the handle of stack driver input */
    IX_ERROR_CR( ix_rm_packet_handler_register(
        IX_CC_CLASSIFIER_6T_MICROBLOCK_PKT_INPUT,
        ix_cc_classifier_6t_pkt_handler,
        pContext )
      );


    return IX_SUCCESS;
}


#else /* !IX_EXCLUDE_CCI */

ix_error
_ix_cc_classifier_6t_reg_pkt_msg_handlers(ix_cc_handle arg_hComponent)
{

    #ifdef IX_CLASSIFIER_6T_DEBUG
    ix_ossl_message_log("_ix_cc_classifier_6t_reg_pkt_msg_handlers(): CCI included :>\n");
    #endif
    /* register message handler */
	IX_ERROR_CR( ix_cci_cc_add_message_handler(
        arg_hComponent,
        IX_CC_CLASSIFIER_6T_MSG_INPUT,
        ix_cc_classifier_6t_msg_handler,
        IX_INPUT_TYPE_SINGLE_SRC)
    );

	/* first registration for the handle of exception packets input */
	IX_ERROR_CR(ix_cci_cc_add_packet_handler(
        arg_hComponent,
        IX_CC_CLASSIFIER_6T_MICROBLOCK_PKT_INPUT,
        ix_cc_classifier_6t_pkt_handler,
        IX_INPUT_TYPE_SINGLE_SRC)
    );

    /* registration for the handle of stack driver input */
	IX_ERROR_CR(ix_cci_cc_add_packet_handler(
        arg_hComponent,
        IX_CC_CLASSIFIER_6T_LOCAL_PKT_INPUT,
        ix_cc_classifier_6t_pkt_handler,
        IX_INPUT_TYPE_SINGLE_SRC)
    );

    return IX_SUCCESS;
}

#endif

/*
 * FUNCTION: _ix_cc_classifier_6t_dereg_pkt_msg_handlers()
 */
ix_error _ix_cc_classifier_6t_dereg_pkt_msg_handlers()
{
    #ifdef IX_CLASSIFIER_6T_DEBUG
    ix_ossl_message_log("_ix_cc_classifier_6t_dereg_pkt_msg_handlers(): executing :>\n");
    #endif


	#if defined(IX_EXCLUDE_CCI)
		/* deregister message handler */
 		IX_ERROR_CR(ix_rm_message_handler_unregister(IX_CC_CLASSIFIER_6T_MSG_INPUT));

        /* deregister packet handlers */

 		/* first deregistration for the exception packet handle */
 		IX_ERROR_CR(ix_rm_packet_handler_unregister(IX_CC_CLASSIFIER_6T_MICROBLOCK_PKT_INPUT));

 		/* second deregistration for the handle of stack driver input */
 		IX_ERROR_CR(ix_rm_packet_handler_unregister(IX_CC_CLASSIFIER_6T_LOCAL_PKT_INPUT));

 	#else

 		/* deregister message handler */
 		IX_ERROR_CR(ix_cci_cc_remove_message_handler(IX_CC_CLASSIFIER_6T_MSG_INPUT));

        /* deregister packet handlers */

		/* first deregistration for the exception packet handle */
		IX_ERROR_CR(ix_cci_cc_remove_packet_handler(IX_CC_CLASSIFIER_6T_MICROBLOCK_PKT_INPUT));

		/* second deregistration for the handle of stack driver input */
		IX_ERROR_CR(ix_cci_cc_remove_packet_handler(IX_CC_CLASSIFIER_6T_LOCAL_PKT_INPUT));

	#endif

    return IX_SUCCESS;
}

/* only include this source code if registry support is present */
#if defined( IX_INCLUDE_REGISTRY )


/* -------------------------- _ix_cc_classifier_6t_get_property_by_handle -- */
PRIVATE ix_error _ix_cc_classifier_6t_get_property_by_handle(
    ix_configuration_property_handle arg_hProp,
    ix_uint32 *                      arg_pPropValue )
  {
    ix_cp_property_info _info;

    /* validity check for arg_hPropValue */
    IX_ERROR_CHECK_ARG_NULL( 2, arg_pPropValue );

    /* validity check for arg_hProp */
    IX_ERROR_CHECK_ARG_NULL( 1, arg_hProp );

    IX_ERROR_CR(
         ix_rm_cp_property_get_info( arg_hProp, &_info )
       );

    /* validity check for property type */
    if ( _info.m_Options & IX_CP_OPTIONS_FLAGS_DATA_UINT32 )
      {
        IX_ERROR_CR(
            ix_rm_cp_property_get_value_uint32(
              arg_hProp, arg_pPropValue )
          );
      }
    else
      {
        return IX_ERROR_LOCAL(
            IX_CC_CLASSIFIER_6T_ERROR_REPOSITORY_READ,
            ( "Invalid CLASSIFIER_6T registry property type. All subproperties " \
              "must be uint32" )
          );
      }

    return IX_SUCCESS;
  }

/* ------------------------------------ _ix_cc_classifier_6t_get_property -- */
PRIVATE ix_error
_ix_cc_classifier_6t_get_property(const char       *arg_pPropName,
                                        ix_uint32  *arg_pPropValue)
{
    ix_configuration_property_handle _hPropHandle;

    /* validity check for arg_pPropName */
    IX_ERROR_CHECK_ARG_NULL( 1, arg_pPropName );

    IX_ERROR_CRT( ix_rm_cp_property_open(
        IX_CP_CORE_PROPERTY_ROOT, arg_pPropName, &_hPropHandle ),
        IX_CC_CLASSIFIER_6T_ERROR_REPOSITORY_READ, IX_ERROR_LEVEL_LOCAL );

    IX_ERROR_CR( _ix_cc_classifier_6t_get_property_by_handle(
      _hPropHandle, arg_pPropValue ) );

    IX_ERROR_CRT( ix_rm_cp_property_close( _hPropHandle ),
        IX_CC_CLASSIFIER_6T_ERROR_REPOSITORY_READ, IX_ERROR_LEVEL_LOCAL );

#if IX_DEBUG

    ix_ossl_message_log("classifier_6t_property: "
                        "%s = %ld\n", arg_pPropName, *arg_pPropValue);

#endif

    return IX_SUCCESS;
}


#endif /* IX_INCLUDE_REGISTRY */


/* ----------------------------------------- _ix_cc_classifier_6t_context -- */

ix_error
_ix_cc_classifier_6t_get_config( _ix_cc_classifier_6t_context * arg_pContext )
  {
    ix_error _result  = IX_SUCCESS;

#if defined( IX_INCLUDE_REGISTRY )

    ix_uint32 _t;
    IX_ERROR_CR( _ix_cc_classifier_6t_get_property(
                       IX_CC_CLASSIFIER_6T_REGISTRY_MEV2_NUM_MASK,
                       &arg_pContext->pUblockConfig->_uEngMask )
               );

    IX_ERROR_CR( _ix_cc_classifier_6t_get_property(
                       IX_CC_CLASSIFIER_6T_REGISTRY_AUX_HASH_TABLE_SIZE,
                       (ix_uint32 *)&arg_pContext->AuxEntryNo )
               );
    arg_pContext->AuxEntryEmpty = arg_pContext->AuxEntryNo;


    IX_ERROR_CR( _ix_cc_classifier_6t_get_property(
                       IX_CC_CLASSIFIER_6T_REGISTRY_HASH_TABLE_CHANNEL_NUM,
                       &arg_pContext->HashTableChannelNum )
               );

    #if defined (CLASSIFIER_6T_PACKET_COUNTER_FEATURE)
    arg_pContext->StatsTableChannelNum = arg_pContext->HashTableChannelNum ;
    #endif

    IX_ERROR_CR( _ix_cc_classifier_6t_get_property(
                       IX_CC_CLASSIFIER_6T_REGISTRY_DEFAULT_COLOR_ID, &_t )
               );
    arg_pContext->pUblockConfig->DefaultColorID = (ix_uint8)_t;


    IX_ERROR_CR( _ix_cc_classifier_6t_get_property(
                       IX_CC_CLASSIFIER_6T_REGISTRY_DEFAULT_FLOW_ID, &_t )
               );
    arg_pContext->pUblockConfig->DefaultFlowID = (ix_uint8)_t;


    IX_ERROR_CR( _ix_cc_classifier_6t_get_property(
                       IX_CC_CLASSIFIER_6T_REGISTRY_DEFAULT_CLASS_ID,
                       &_t )
               );
    arg_pContext->pUblockConfig->DefaultClassID = (ix_uint16)_t;

    IX_ERROR_CR( _ix_cc_classifier_6t_get_property(
                       IX_CC_CLASSIFIER_6T_REGISTRY_DEFAULT_NEXTHOP_ID,
                       &_t )
               );
    arg_pContext->pUblockConfig->DefaultNextHopID = (ix_uint16)_t;

    IX_ERROR_CR( _ix_cc_classifier_6t_get_property(
                       IX_CC_CLASSIFIER_6T_REGISTRY_DEFAULT_NEXTHOP_ID_TYPE, &_t )
               );
    arg_pContext->pUblockConfig->DefaultNextHopIDType = (ix_uint8)_t;


#else
    /* read config data from config header file */
    arg_pContext->AuxEntryEmpty                       = IX_CC_CLASSIFIER_6T_AUX_HASH_TABLE_SIZE;
    arg_pContext->AuxEntryNo                          = IX_CC_CLASSIFIER_6T_AUX_HASH_TABLE_SIZE;
    arg_pContext->HashTableChannelNum                 = IX_CC_CLASSIFIER_6T_HASH_TABLE_CHANNEL_NUM;
    arg_pContext->pUblockConfig->DefaultColorID       = IX_CC_CLASSIFIER_6T_DEFAULT_COLOR_ID;
    arg_pContext->pUblockConfig->DefaultFlowID        = IX_CC_CLASSIFIER_6T_DEFAULT_FLOW_ID;
    arg_pContext->pUblockConfig->DefaultClassID       = IX_CC_CLASSIFIER_6T_DEFAULT_CLASS_ID;
    arg_pContext->pUblockConfig->DefaultNextHopID     = IX_CC_CLASSIFIER_6T_DEFAULT_NEXTHOP_ID;
    arg_pContext->pUblockConfig->DefaultNextHopIDType = IX_CC_CLASSIFIER_6T_DEFAULT_NEXTHOP_ID_TYPE;
    arg_pContext->pUblockConfig->_uEngMask            = IX_CC_CLASSIFIER_6T_DEFAULT_MES_NUMBERS;

    #if defined (CLASSIFIER_6T_PACKET_COUNTER_FEATURE)
    arg_pContext->StatsTableChannelNum                = IX_CC_CLASSIFIER_6T_STATS_TABLE_CHANNEL_NUM;
    #endif

#endif

    arg_pContext->pUblockConfig->HashIndexSize        = CLASSIFIER_6T_HASH_INDEX_SIZE;
    return _result;
  }

/*
 * FUNCTION: _ix_cc_classifier_6t_patch_microblock()
 */

ix_error
_ix_cc_classifier_6t_patch_microblock(
                       const _ix_cc_classifier_6t_context* pContext)
{

     ix_cc_classifier_6t_ublock_config* pConfig = pContext->pUblockConfig;
     ix_uint16                          index   = 0;

     /* patch config data to the classifier_6t microblocks  */

     /* 1. allocate data structures for import variables */
     ix_imported_symbol aImportSymbols[IX_CC_CLASSIFIER_6T_PATCH_SYMBOLS];

     /* populate array of import variables */
     SetUblockSymbols(
                      aImportSymbols[index],
                      pConfig->HashTableBase,
                      &sz_aImpVar[index]);

     index++;

     SetUblockSymbols(
                      aImportSymbols[index],
                      pConfig->DefaultColorID,
                      &sz_aImpVar[index]);
     index++;

     SetUblockSymbols(
                      aImportSymbols[index],
                      pConfig->DefaultFlowID,
                      &sz_aImpVar[index]);
     index++;

     SetUblockSymbols(
                      aImportSymbols[index],
                      pConfig->DefaultClassID,
                      &sz_aImpVar[index]);
     index++;

     SetUblockSymbols(aImportSymbols[index],
                      pConfig->DefaultNextHopID,
                      &sz_aImpVar[index]);
     index++;

     SetUblockSymbols(aImportSymbols[index],
                      pConfig->DefaultNextHopIDType,
                      &sz_aImpVar[index]);

     index++;

#if defined(CLASSIFIER_6T_PACKET_COUNTER_FEATURE)

     SetUblockSymbols(aImportSymbols[index],
                      pConfig->StatsTableBase,
                      &sz_aImpVar[index]);

     index++;

     SetUblockSymbols(aImportSymbols[index],
                      pConfig->DefaultStatAddr,
                      &sz_aImpVar[index]);
#endif

     for(index = 0; index < IX_MAX_MICROENGINE_NUMBER; index++)
     {
         if (pContext->pUblockConfig->_uEngMask & (0x1 << index))
         {
             ix_rm_ueng_patch_symbols(g_aMicroengineNumberMapping[index],
                                      IX_CC_CLASSIFIER_6T_PATCH_SYMBOLS,
                                      aImportSymbols);
         }
     }

     return IX_SUCCESS;
}

