#ifndef CLASSIFIER_6T_UC
#define CLASSIFIER_6T_UC

/*******************************************************************************
                             Intel Proprietary

 Copyright (c) 1998-2003 By Intel Corporation.  All rights reserved.
 No part of this program or publication may be reproduced, transmited,
 transcribed, stored in a retrieval system, or translated into any language
 or computer language in any form or by any means, electronic, mechanical,
 magnetic, optical, chemical, manual, or otherwise, without the prior
 written permission of:
                         Intel Corporation
                         2200 Mission College Blvd.
                         Santa Clara, CA  95052-8119
*******************************************************************************/
/* #define IX_ERROR_FILE_IDENT "$Id: classifier_6t.uc,v 1.31.6.1 2004/03/12 10:28:15 wneubaue Exp $" */

// FILE  classifier_6t.uc
//
//   Implements 6-tuple IPv4 classifier micro-block


// Exported APIs:
//
//   classifier_6t(in_ip, IPHDR_RD_START_BYTE)
//   classifier_6t_init()
//   classifier_6t_term()

// ASSUMPTIONS
//
//   IP packets have headers without IP otions. If packet does not conform to
//   this restrictions classifier sends them to block with
//   CLASSIFIER_6T_INV_IP_NEXT_BLOCK id. This way they can be directed to be
//   handled by a core component. Morover, IP must carry TCP/UDP/SCTP
//   protocols. This must be checked before the classifier.
//


//------------------------------------------------------------------------------
// Included files:

#include <stdmac.uc>
#include <ipv4.h>
#include <classifier_6t_defs.h>
#include <classifier_6t.h>      // definitions relevant to the classifier

//------------------------------------------------------------------------------
// Global definitions:

#ifdef _HASH_ENTRY
	#error "_HASH_ENTRY constant has already been defined"
#endif

#ifdef CLASSIFIER_6T_DRAM_HASH_TABLE
	#define _HASH_ENTRY  $$hash_entry
#endif

#ifdef CLASSIFIER_6T_SRAM_HASH_TABLE
	#define _HASH_ENTRY  $hash_entry
#endif


//------------------------------------------------------------------------------
// #macro	classifier_6t_init[]

// classifier_6t_init
//
//    Description:
//       Initializes hardware for classifier (sets up hash multiplcands)
//
#macro	classifier_6t_init[]

	br!=ctx[0,init_relative#]

init_absolute#:


	.begin
	.reg mult_init
		move(mult_init, CLASSIFIER_6T_HASH_MUL_LW0)
		alu[--,--,B,mult_init]
		cap[fast_wr,ALU,HASH_MULTIPLIER_128_0]
		move(mult_init, CLASSIFIER_6T_HASH_MUL_LW1)
		alu[--,--,B,mult_init]
		cap[fast_wr,ALU,HASH_MULTIPLIER_128_1]
		move(mult_init, CLASSIFIER_6T_HASH_MUL_LW2)
		alu[--,--,B,mult_init]
		cap[fast_wr,ALU,HASH_MULTIPLIER_128_2]
		move(mult_init, CLASSIFIER_6T_HASH_MUL_LW3)
		alu[--,--,B,mult_init]
		cap[fast_wr,ALU,HASH_MULTIPLIER_128_3]
	.end

init_relative#:
#endm


//------------------------------------------------------------------------------
// #macro	classifier_6t_term[]

// classifier_6t_term
//
// Description:
//   This macro dismisses any resources used by 6-tuple classifier
//
#macro	classifier_6t_term[]
#endm



//------------------------------------------------------------------------------
// #macro	classifier_6t(in_ip, IPHDR_RD_START_BYTE)

// classifier_6t
//
// Description:
//   This macro implements an exact match 6-tuple classifier
//   The purpose of a 6-tuple classifier is to classify incoming packets
//   into flows based on 6 fields from their IPv4 headers.
//
//  Inputs
//    in_ip          - Transfer registers containing the IP
//                     packet. The register buffer is assumed to be of
//                     the format in_ip0, in_ip1, in_ip2 ...in_ip7
//    dl_input_port  - physical input port number.
//    dl_next_block  - identification of the next block which the packet is
//                     to be handled by. Packet is processed by a classifier
//                     micro block only if this variable has BID_CLASSIFIER_6T
//                     value.
//
//  Outputs
//    dl_flow_id     - identification of a packet flow which the current
//                     packet belongs has been assigned to.
//    dl_class_id    - identification of a queue to be used for the current
//                     packet. It determines global priority of the packet with
//                     which packet is handled in queues.
//    dl_color_id    - packet color for the current packet. It determines level
//                     of conformance to SLA e.i. probability of dropping in
//                     high congestion conditions.
//    dl_next_hop_id - id used on the egress side to look up the outgoing link
//                     layer information.
//    dl_next_block  - can assume:
//                       * CLASSIFIER_6T_DEFAULT_NEXT_BLOCK set if there is no
//                         matching entry in the hash table for the current
//                         packet
//                       * CLASSIFIER_6T_INVALID_NEXT set if there are option
//                         fields in the current packet header or if the current
//                         packet does not carry IP.
//                       * hash[matching entry].nextblock - if the current
//                         packet maches an entry in the hash table the next
//                         block is indicated by one of the fields from that
//                         entry.
//                       * Left unchanged if packet was directed to another
//                         block
//
//  Constants:
//    IPHDR_RD_START_BYTE - byte address relative to start of in_ip
//
//  The classifier performs reading from HASH TABLE from RAM which has
//	the following structure:
//
/*******************************************************************************

       +-----> memory address growth
       |
       |   bit32,               16,   10,8,       0
       |      +-----------------------------------+
       |    0 |       Next hash table entry       |
       v      +-----------------------------------+
            1 |         IP source address         |
              +-----------------------------------+
            2 |       IP destination address      |
              +-----------------+-----------------+
            3 |    source port  |destination port |
              +-----------------+-----+--+--------+
            4 |     input port  |DSCP |RS|protocol|
              +-----------------+-----+--+--------+
            5 |              flow ID              |
              +-----------------+--------+--------+
            6 |     class ID    |nht/col |next_bl |
              +-----------------+--------+--------+
            7 |                 |   next hop id   |
              +-----------------+-----------------+

  RS field: 2 bits reserved, must be zeros

  nht/col field:

           bit 7  6  5  4  3  2  1   0
              +-----------+-----+-----+
              |     nht   |  rs |color|
              +-----------+-----+-----+
                    |        |     |
                    |        |     +-> bits 0:1 - color_id
                    |        +-------> bits 2:3 - reserved
                    +----------------> bits 4:7 - next_hop_id_type

*******************************************************************************/

#macro	classifier_6t(in_ip, IPHDR_RD_START_BYTE)
.begin

	// check block ID. If doesn't match ours, execute bypass path.
	br!=byte[dl_next_block, 0, BID_CLASSIFIER_6T, classifier_end#]


#ifdef CLASSIFIER_6T_PACKET_COUNTER_FEATURE
	// Registers and signals used to handle statistics.
	.reg read $test_pkt_count
	.reg $test_byte_count
	.sig pkt_count_sig
	.sig byte_count_sig
	.reg packet_size
	.reg stats_base 
	.reg counter_offset
	.reg stats_offset
	.reg stats_enable
#endif

classifier_path#:

	// The relative start byte position
	// of IP header w.r.t start of the register buffer.
	#define_eval	RD_START		IPHDR_RD_START_BYTE


	//-------------------------------
	#ifndef CLASSIFIER_6T_HASH_BY_CRC
	//-------------------------------
	.begin
	.reg version_length
		; read two fields from IPv4 header: protocol version and header length
		xbuf_extract(version_length, in_ip, RD_START, IP_VERSION_LENGTH)
		; check if packet is IPv4 with the header 20 bytes in size.
		alu[--, version_length, -, IPV4_NO_OPTIONS]
		bne[invalid_ip#]
	.end
	//-------------------------------
	#endif
	//-------------------------------


	.sig hash_done
	.sig hash_entry_read_dn
	// fields of the IP header's 6-tuple
	.reg src_address dst_address tcp_udp_prt protocol_TOS input_port
	.reg hash_base hash_offset hash_key

	dl_meta_get_input_port(input_port)

	// packet_size is only needed when we gather statistics.
#ifdef CLASSIFIER_6T_PACKET_COUNTER_FEATURE
	dl_meta_get_packet_size(packet_size)
#endif

	;----------------------------------------------------------
	;	extract 6 tuples
	;----------------------------------------------------------
	.begin
	.reg proto_field TOS_field flags_field

		; extract ipv4 flags field
		xbuf_extract(flags_field, in_ip, RD_START, IP_FLAGS_FRAGMENT)

		; flags_field consists of 16 bits:
        ;
        ;    Bits: 0   1   2                      15
        ;        +---+---+---+------- - - ---------+
        ;        |res|DF |MF |     offset          |
        ;        +---+---+---+------- - - ---------+
		;
		;      Bit 0:    reserved, must be zero
		;      Bit 1:    (DF) 0 = May Fragment,  1 = Don't Fragment.
		;      Bit 2:    (MF) 0 = Last Fragment, 1 = More Fragments.
		;      Bit 3-15: This field indicates where in the datagram 
		;                this fragment belongs. The fragment offset 
		;                is measured in units of 8 octets (64 bits).
		;                The first fragment has offset zero.
        ;
		; if having masked out DF bit the remaining value is not zero
		; it means that packet is fragmented and default rule should be
		; applied.
		#define FLAG_DF_BIT 1
			br_bset[flags_field, FLAG_DF_BIT, send_to_default_block#]
		#undef FLAG_DF_BIT

		// extract protocol field
		xbuf_extract(proto_field, in_ip, RD_START, IP_PROTOCOL)
		// extract TOS field
		xbuf_extract(TOS_field,   in_ip, RD_START, IP_TYPE_OF_SERVICE)
		// extract the source address
		xbuf_extract(src_address, in_ip, RD_START, IP_SOURCE_ADDRESS)
		// extract the destination address
		xbuf_extract(dst_address, in_ip, RD_START, IP_DESTINATION_ADDRESS)
		// extract tcp / udp ports
		xbuf_extract(tcp_udp_prt, in_ip, RD_START, TCP_UDP_PORTS)

		alu[TOS_field, TOS_field, and, 0xFC]  ; take only DSCP field (6 bits)
		alu_shf[protocol_TOS, proto_field, or, TOS_field, <<8]

		;-----------------------------------------------------------------------
		;	issue a hash operation
		;-----------------------------------------------------------------------


		#ifdef CLASSIFIER_6T_HASH_BY_CRC

		//----------------------------------------------------------------------
		//		hash index calculation with crc instruction
		//----------------------------------------------------------------------
		.begin
		.reg crc_result

			local_csr_wr[crc_remainder, 0x0] ;clear CRC remainder
			nop ; let compiler remove these nops
			nop
			nop
			crc_le[crc_32, --, src_address]
			nop ; let compiler remove these nops
			crc_le[crc_32, --, dst_address]
			nop ; let compiler remove these nops
			crc_le[crc_32, --, protocol_TOS]
			nop ; let compiler remove these nops
			crc_le[crc_32, --, tcp_udp_prt]
			nop ; let compiler remove these nops
			nop
			nop
			nop
			nop ; five intervening instructions before reading result

			.reg version_length
			; read two fields from IPv4 header: protocol version and header length
			xbuf_extract(version_length, in_ip, RD_START, IP_VERSION_LENGTH)
			; check if packet is IPv4 with the header 20 bytes in size.
			alu[--, version_length, -, IPV4_NO_OPTIONS]
			bne[invalid_ip#]

			immed32(hash_base, CLASSIFIER_6T_HASH_TABLE_BASE)

			local_csr_rd[crc_remainder] ;get the new remainder
			immed[crc_result, 0x0000]

			#define_eval _BIT_SHIFT (32 - CLASSIFIER_6T_HASH_INDEX_SIZE)
				alu_shf[hash_key, --, B, crc_result, >>_BIT_SHIFT]
			#undef _BIT_SHIFT

		.end


		#else

		//----------------------------------------------------------------------
		//		Hash-unit supported calculation
		//----------------------------------------------------------------------

			xbuf_alloc($hash, 4, read_write)

			alu[$hash[0], --, B, src_address]
			alu[$hash[1], --, B, dst_address]
			alu[$hash[2], --, B, protocol_TOS]
			alu[$hash[3], --, B, tcp_udp_prt]

			hash_128[$hash[0], 1], sig_done[hash_done]

			immed32(hash_base, CLASSIFIER_6T_HASH_TABLE_BASE]

			ctx_arb[hash_done]


			; set index to hash table by taking certain number of
			; lower bits from computed hash. (default 16)
			#define_eval CLASSIFIER_6T_HASH_INDEX_MASK ((1 << CLASSIFIER_6T_HASH_INDEX_SIZE) - 1)
			alu_op(hash_key, $hash[0], AND, CLASSIFIER_6T_HASH_INDEX_MASK)
			#undef CLASSIFIER_6T_HASH_INDEX_MASK

			xbuf_free($hash)

		#endif

		#undef RD_START

	.end // proto_field   TOS_field

	; hash_offset equals hash_key multiplied by the size of one hash entry.
	alu[hash_offset, --, B, hash_key, <<CLASSIFIER_6T_HASH_TABLE_ENTRY_LOG_SIZE]

    //------------------------------------------
	#ifdef CLASSIFIER_6T_PACKET_COUNTER_FEATURE

		alu_shf[stats_offset, //
		        --, B, hash_key, <<CLASSIFIER_6T_STATISTICS_ENTRY_LOG_SIZE]
	#endif
    //------------------------------------------


	; hash entry occupies 8 LW in ram.
	xbuf_alloc(_HASH_ENTRY, 8, read)

	#ifdef CLASSIFIER_6T_DRAM_HASH_TABLE
		;get hash table entry using hash_key as index
		;each entry is 8 LW i.e. 32 bytes
		dram[read, _HASH_ENTRY[0], hash_base, hash_offset, 4], //
                                                  sig_done[hash_entry_read_dn]
		ctx_arb[hash_entry_read_dn]
	#endif

	#ifdef CLASSIFIER_6T_SRAM_HASH_TABLE
		;get hash table entry using hash_key as index
		;each entry is 8 LW i.e. 32 bytes
		sram[read, _HASH_ENTRY[0], hash_base, hash_offset, 8], //
                                                  ctx_swap[hash_entry_read_dn]
	#endif



6_tuple_match_loop#:


	.reg compare hash_interface

	alu[--, src_address, -, _HASH_ENTRY[1]]
	bne[dont_match#]
	alu[--, dst_address, -, _HASH_ENTRY[2]]
	bne[dont_match#]
	alu[--, tcp_udp_prt, -, _HASH_ENTRY[3]]
	bne[dont_match#]

	alu[compare, _HASH_ENTRY[4], XOR, protocol_TOS]
	alu[--, --, B, compare, <<16]
	bne[dont_match#]

	; if (hash_table_entry.input_port == -1) goto match#
	immed[compare, 0xFFFF]
	alu_shf[hash_interface, --, B, _HASH_ENTRY[4], >>16]
	alu[--, hash_interface, -, compare]
	beq[match#]

	; if (hash_table_entry.if_num == packet.interface) goto dont_match#
	alu[--, input_port, -, hash_interface]
	bne[dont_match#]


match#:

	.reg class_id color_id next_hop_id_type

	dl_meta_set_flow_id(_HASH_ENTRY[5])
	alu_shf[class_id, --, B, _HASH_ENTRY[6], >>16]
	dl_meta_set_class_id(class_id)

	alu_shf[color_id, CLASSIFIER_6T_COLOR_ID_MASK,
						                  and, _HASH_ENTRY[6], >>8]
	dl_meta_set_color(color_id)
	alu_shf[next_hop_id_type, CLASSIFIER_6T_NEXTHOP_ID_TYPE_MASK,
						                  and, _HASH_ENTRY[6], >>12]
	dl_meta_set_nexthop_id_type(next_hop_id_type)

	dl_meta_set_nexthop_id(_HASH_ENTRY[7])

	;dl_next_block = hash_table_entry.next_block
	alu[dl_next_block, 0xff, AND, _HASH_ENTRY[6]]


	#ifdef CLASSIFIER_6T_PACKET_COUNTER_FEATURE

		alu_shf[stats_enable, _HASH_ENTRY[6], AND, 1, <<15]
		br_bclr[stats_enable, CLASSIFIER_6T_STATISTICS_ENABLE_BIT, \
															classifier_end#]
		immed32(stats_base, CLASSIFIER_6T_STATS_SRAM_BASE)

		alu[counter_offset, stats_offset, +, CLASSIFIER_6T_PACKET_COUNTER_LOW]
		sram[test_and_incr, $test_pkt_count, stats_base, \
							counter_offset], sig_done[pkt_count_sig]

		alu[counter_offset, stats_offset, +, CLASSIFIER_6T_BYTE_COUNTER_LOW]
		move($test_byte_count, packet_size)
		sram[test_and_add, $test_byte_count, stats_base, \
							counter_offset], sig_done[byte_count_sig]

		br[finish_stats_setting#]

	#else

        br[classifier_end#]

	#endif




dont_match#:

; It is possible that some packets may have the same hash_key indicating
; the same field in hash table structure. To differentiate between them,
; entries in hash table can be connected into lists. So if packet's header
; doesn't match the entry in hash table we must check next entry
; pointer to which is stored in the first lword of the current hast entry.

	alu[hash_offset, --, B, _HASH_ENTRY[0]]
	; if pointer is null skip reading next entry from the hash table
	beq[send_to_default_block#]

	#ifdef CLASSIFIER_6T_DRAM_HASH_TABLE
		;get hash table entry using hash_key as index
		;each entry is 8 LW i.e. 32 bytes
		dram[read, _HASH_ENTRY[0], 0x00, hash_offset, 4], //
                                                  sig_done[hash_entry_read_dn]
	#endif

	#ifdef CLASSIFIER_6T_SRAM_HASH_TABLE
		;get hash table entry using hash_key as index
		;each entry is 8 LW i.e. 32 bytes
		sram[read, _HASH_ENTRY[0], 0x00, hash_offset, 8], //
                                                  sig_done[hash_entry_read_dn]
	#endif

	#ifdef CLASSIFIER_6T_PACKET_COUNTER_FEATURE
		#ifdef HASH_OFFS_TO_STATS_OFFS_SHIFT
		#error I can't use HASH_OFFS_TO_STATS_OFFS_SHIFT
		#endif
		#define_eval HASH_OFFS_TO_STATS_OFFS_SHIFT                         \
		                   (CLASSIFIER_6T_HASH_TABLE_ENTRY_LOG_SIZE -      \
						    CLASSIFIER_6T_STATISTICS_ENTRY_LOG_SIZE)

		alu[stats_offset, hash_offset, -, hash_base]
		alu[stats_offset, --, B, stats_offset, >>HASH_OFFS_TO_STATS_OFFS_SHIFT]

		#undef HASH_OFFS_TO_STATS_OFFS_SHIFT
	#endif

    ctx_arb[hash_entry_read_dn], br[6_tuple_match_loop#]


	xbuf_free(_HASH_ENTRY)



invalid_ip#:
	dl_set_exception(BID_CLASSIFIER_6T, 0);
	immed[dl_next_block, CLASSIFIER_6T_INV_IP_NEXT_BLOCK]
	br[classifier_end#]



send_to_default_block#:

    #ifdef CLASSIFIER_6T_PACKET_COUNTER_FEATURE
	#ifdef CLASSIFIER_6T_DEFAULT_STATS_ENABLE

		immed32(stats_base, CLASSIFIER_6T_DEFAULT_STATS_ADDRESS)
		immed[stats_offset, 0]
		sram[test_and_incr, $test_pkt_count, stats_base, \
				CLASSIFIER_6T_PACKET_COUNTER_LOW], sig_done[pkt_count_sig]
		move($test_byte_count, packet_size)
		sram[test_and_add, $test_byte_count, stats_base, \
				CLASSIFIER_6T_BYTE_COUNTER_LOW], sig_done[byte_count_sig]
    #endif
    #endif


	.reg default_flow_id
	.reg default_next_hop_id


    immed[default_flow_id, CLASSIFIER_6T_DEFAULT_FLOW_ID]
    immed[default_next_hop_id, CLASSIFIER_6T_DEFAULT_NEXTHOP_ID]

	dl_meta_set_color(CLASSIFIER_6T_DEFAULT_COLOR_ID)
	dl_meta_set_flow_id(default_flow_id)
	dl_meta_set_class_id(CLASSIFIER_6T_DEFAULT_CLASS_ID)
	dl_meta_set_nexthop_id(default_next_hop_id)
	dl_meta_set_nexthop_id_type(CLASSIFIER_6T_DEFAULT_NEXTHOP_ID_TYPE)

    immed[dl_next_block, CLASSIFIER_6T_DEFAULT_NEXT_BLOCK]

    #ifndef CLASSIFIER_6T_DEFAULT_STATS_ENABLE
	    br[classifier_end#]
    #endif


#ifdef CLASSIFIER_6T_PACKET_COUNTER_FEATURE

finish_stats_setting#:
	ctx_arb[pkt_count_sig]
	alu[--, $test_pkt_count, + , 1]
	bvc[pkt_count_no_overflow#]
	alu[counter_offset, stats_offset, +, CLASSIFIER_6T_PACKET_COUNTER_HIGH]
	sram[incr, --, stats_base, counter_offset]

pkt_count_no_overflow#:
	ctx_arb[byte_count_sig]
	alu[--, $test_byte_count, + , packet_size]
	bvc[byte_count_no_overflow#]
	alu[counter_offset, stats_offset, +, CLASSIFIER_6T_BYTE_COUNTER_HIGH]
	sram[incr, --, stats_base, counter_offset]

byte_count_no_overflow#:

#endif

classifier_end#:


.end
#endm


#endif	//  CLASSIFIER_6T_UC
