/*******************************************************************************
  INTEL CONFIDENTIAL

  Copyright 2003 Intel Corporation All Rights Reserved.

  The source code contained or described herein and all documents related to
  the source code ("Material") are owned by Intel Corporation or its 
  suppliers or licensors.

  Title to the Material remains with Intel Corporation or its suppliers and
  licensors. The Material contains trade secrets and proprietary and
  confidential information of Intel or its suppliers and licensors.
  The Material is protected by worldwide copyright and trade secret laws and
  treaty provisions. No part of the Material may be used, copied, reproduced,
  modified, published, uploaded, posted, transmitted, distributed,
  or disclosed in any way without Intel's prior express written permission.

  No license under any patent, copyright, trade secret or other intellectual
  property right is granted to or conferred upon you by disclosure
  or delivery of the Materials, either expressly, by implication, inducement,
  estoppel or otherwise. Any license under such intellectual property rights
  must be express and approved by Intel in writing.

*******************************************************************************/

/**
 * ============================================================================
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2000 Network Processor
 *
 * = FILE
 *     dscp_classifier_api.h
 *
 * = DESCRIPTION
 *      The DSCP classifier core component Library API definition. It provides
 *      direct C-style calls to the classifier core component. The library 
 *      functions correspond one-to-one with message helper API primitives. The
 *      differece is that library funcitons are executed in a caller context.
 *      In addition, the Libaray API is synchronous. There are no callback
 *      functions to report operation results.
 *
 * ============================================================================
 */

#if !defined( _DSCP_CLASSIFIER_API_H__INCLUDED_ )
#define _DSCP_CLASSIFIER_API_H__INCLUDED_

#if defined( __cplusplus )
extern "C" {
#endif /* __cplusplus */

/**
 * @brief  This function assigns interface-specific classification rules to an
 *         input port.
 *
 * If the input port already has its own classification rules, the function 
 * returns an error code. Otherwise, the function copies classification results
 * given in arg_pCassifResults table to the appropriate entries in the SRAM 
 * configuration table. Next the function marks the input port as configured.
 *
 * If the microblock supports statistics gathering and the statistics flag in
 * any of the altered SRAM configuration entries changes from zero to one, the
 * function zeroes the corresponding entries in the statistics table in SRAM.
 *
 * @param  arg_inputPort       interface for which the classification reults
 *                             are added 
 *
 * @param  arg_pClassifResults 64-element table indexed with DSCP value. Each
 *                             entry of this table contains QoS parameters
 *                             assigned to packets received from the input port
 *                             and carrying the DSCP value.
 *
 * @param  arg_pContext        pointer to the control block internal to the
 *                             core component.
 *
 * @raturn IX_SUCCESS for successful or a valid ix_error token for failure. The
 *         following error codes might be returned via the token:
 *\n
 *\n       IX_CC_DSCP_ERROR_INVALID_PARAM   - invalid parameter value specified
 *                                            (i.e. the input port number is 
 *                                            out of range)
 *\n       IX_CC_ERROR_NULL                 - arg_pContext argument is NULL,
 *\n       IX_CC_ERROR_DUPLICATE_ENTRY      - an entry for the key already 
 *                                            configured.
 */
ix_error ix_cc_classifier_dscp_add_if_config( 
             ix_uint16                      arg_inputPort,
             ix_cc_classifier_dscp_result * arg_pClassifResults,
             void *                         arg_pContext
           );

/**
 * @brief  This funciton removes a set of classification rules assigned to an 
 *         interface.
 *
 * If the input port does not have its own classifiraction rules, the function
 * returns an error code. Otherwise, the function changes configuration of the 
 * input port to the default configuration. It copies the default 
 * classification rules to the appropriate entries of the SRAM configuration
 * table. Next the function marks the input port as not configured (i.e. using
 * the default configuration).
 *
 * @param  arg_inputPort       identifies input port whose configuration is 
 *                             deleted
 *
 * @param  arg_pContext        pointer to the control block internal to the 
 *                             core component
 *
 * @return IX_SUCCESS for successful or a valid ix_error token for failure. The
 *         following error codes might be returned via the token:
 *\n
 *\n       IX_CC_ERROR_ENTRY_NOT_FOUND      - interface-specific configuration 
 *                                            for this port has not been found,
 *\n       IX_CC_ERROR_NULL                 - arg_pContext argument is NULL,
 *\n       IX_CC_DSCP_ERROR_INVALID_PARAM   - invalid parameter value specified
 *                                            (i.e. the input port number is
 *                                            out of range)           
 */
ix_error ix_cc_classifier_dscp_remove_if_config( 
             ix_uint16                      arg_inputPort,
             void *                         arg_pContext
           );

/**
 * @brief  This function updates a set of classification rules associated with
 *         an interface.
 *
 * If the input port does not have its own classification rules, the function
 * returns an error code. Otherwise, the function copies the new classification
 * rules to the appropriate entries of the SRAM configuration table.
 *
 * If the microblock supports statistics gathering and the statistics flag in
 * any of the altered SRAM configuration entries changes from zero to one, the 
 * function zeroes the corresponding entries in the statistics table in SRAM. 
 *
 * @param  arg_inputPort       inteface for which the classification results
 *                             are added.
 *
 * @param  arg_pClassifResults a 64-element table indexed with DSCP value. Each
 *                             entry of this table contains QoS parameters 
 *                             assigned to packet received from the input port
 *                             and carrying the DSCP value.
 * 
 * @param  arg_pContext        pointer to the control block internal to the 
 *                             core component.
 *
 * @return IX_SUCCESS if successful or a valid ix_error token for failure. The
 *         following error codes might be returned via the token:
 *\n
 *\n       IX_CC_ERROR_ENTRY_NOT_FOUND      - interface-specific rules for the
 *                                            port not exist,
 *\n       IX_CC_ERROR_NULL                 - one of the input arguments of the
 *                                            function is NULL,
 *\n       IX_CC_DSCP_ERROR_INVALID_PARAM   - invalid parameter value specified
 *                                            (i.e. the input port number is 
 *                                            out of range).
 */
ix_error ix_cc_classifier_dscp_update_if_config( 
             ix_uint16                      arg_inputPort,
             ix_cc_classifier_dscp_result * arg_pClassifResults,
             void *                         arg_pContext
           );

/**
 * @brief  The function returns interface-specific configuration of an input 
 *         port.
 *
 * If the input port does not have its own classification rules, the function
 * returns an error code. 
 *
 * @param  arg_inputPort       interface for which the classification results
 *                             are retrieved. 
 *
 * @param  arg_pClassifResults pointer to a placeholder for the interafce-
 *                             -specific classification rules. It is a 
 *                             64-element table. Each element contains QoS
 *                             parameters assigned to packets received from the
 *                             input port and carrying the DSCP value. The 
 *                             caller must only allocate the memory for the 
 *                             table. When a response from the core component
 *                             is received, the table is filled in and passed
 *                             to the client callback function.
 *
 * @param  arg_pContext        pointer to the control block internal to the 
 *                             core component.
 *
 * @return IX_SUCCESS if successful or a valid ix_error token for failure. The
 *         following error codes might be returned via the token:
 *\n
 *\n       IX_CC_ERROR_ENTRY_NOT_FOUND      - interface-specific rules for the
 *                                            port not exist,
 *\n       IX_CC_DSCP_ERROR_INVALID_PARAM   - invalid parameter value specified
 *                                            (i.e. the input port number is
 *                                            out of range),
 *\n       IX_CC_ERROR_NULL                 - one of the input arguments of the 
 *                                            function is specified as NULL.
 */
ix_error ix_cc_classifier_dscp_get_if_config( 
             ix_uint16                      arg_inputPort,
             ix_cc_classifier_dscp_result * arg_pClassifResults,
             void *                         arg_pContext
           );

/**
 * @brief  The function changes default classification rules.
 *
 * This function changes default classification rules  for the interfaces that
 * do not have their own interface-specific rules. The function copies the new 
 * rules to the entries in SRAM configuration table that contain classification 
 * results for input ports using default rules. The function stores the new 
 * default rules in the control block.
 *
 * If the microblock supports statistics gathering and the statistics flag in
 * any of the altered SRAM configuration entries changes from zero to one, the
 * function zeroes the corresponding entries in the statistics table in SRAM.
 *
 * @param  arg_pClassifResults a 64-element table indexed with DSCP value and 
 *                             containing the new set of classification rules.
 *                             Each entry of this table contains QoS parameters
 *                             assigned to oackets carrying the DSCP value.
 *
 * @param  arg_pContext        pointer to the control block internal to the 
 *                             core component.
 *
 * @return IX_SUCCESS if successful or a valid ix_error token for failure. The
 *         following error codes might be returned via the token:
 *\n
 *\n       IX_CC_ERROR_NULL                 - one of the input arguments of the
 *                                            function is NULL.
 */
ix_error ix_cc_classifier_dscp_set_def_rules(
             ix_cc_classifier_dscp_result * arg_pClassifResults,
             void *                         arg_pContext
           );

/**
 * @brief  The function returns default classification rules.
 *
 * This function returns all the 64 classification rules applicable to the
 * input port without interface specific configuration.
 *
 * @param  arg_pClassifResults pointer to a placeholder for the interface-
 *                             -specific classification rules. It is a 
 *                             64-element table. Each element contains QoS
 *                             parameters assigned to packets carrying the DSCP
 *                             value. The caller must only allocate memory for
 *                             the table. When a response from the core 
 *                             component is received, the table is filled in 
 *                             and passed to the client callback function.
 *
 * @param  arg_pContext        pointer to the control block internal to the 
 *                             core component.
 *
 * @return IX_SUCCESS if successful or a valid ix_error token for failure. The
 *         following error codes might be returned via the token:
 *\n
 *\n       IX_CC_ERROR_NULL                 - one of the input arguments of the
 *                                            function is NULL.
 */
ix_error ix_cc_classifier_dscp_get_def_rules(
             ix_cc_classifier_dscp_result * arg_pClassifResults,
             void *                         arg_pContext
           );

/**
 * @brief  The function returns statistics gathered by the microblock.
 *
 * If the microblock supports statistics gathering, the core component returns
 * statistics gatherd for the <input port, DSCP value> pair. Otherwise, the 
 * core component returns an error code and the arg_pStatistics argument is 
 * left unaltered.
 *
 * The funciton uses SRAM read operations to get 64-bit counter values, 
 * becouse the operation does not change the statistics and does not have to be
 * synchronized with statistics updates. This means that core component can
 * retrieve the counter in the middle of a read-modify-write atomic operation 
 * performed by the microblock. In such case, the SRAN controller returns the
 * counter value before the atomic operation.
 *
 * @param  arg_inputPort       interface for which the statistics are 
 *                             requested.
 *
 * @param  arg_dscp            a DSCP value for which the statistics are 
 *                             requested.
 *
 * @param  arg_pStatistics     pointer to a placeholder for the statistics 
 *                             values associated with the 
 *                             <input port, DSCP value> pair. The caller must 
 *                             alloate memory for the data structure.
 *
 * @param  arg_pContext        pointer to the control block internal to the
 *                             core component.
 *
 * @return IX_SUCCESS if successful or a valid ix_error token for failure. The
 *         following error codes might be returned via the token:
 *\n
 *\n       IX_CC_ERROR_NULL                 - one of the arguments of the 
 *                                            function is NULL,
 *\n       IX_CC_DSCP_ERROR_INVALID_PARAM   - invalid parameter value specified
 *                                            (i.e. the input port number or
 *                                            DSCP value are out of range),
 *\n       IX_CC_DSCP_ERROR_STATS_NOT_SUPPORTED - the microblock does not 
 *                                            support statistics gathering.
 */
ix_error ix_cc_classifier_dscp_get_statistics(
             ix_uint16                          arg_inputPort,
             ix_uint8                           arg_dscp,
             ix_cc_classifier_dscp_statistics * arg_pStatistics,
             void *                             arg_pContext
           );

#if defined( __cplusplus )
}
#endif /* __cplusplus */

#endif /* _DSCP_CLASSIFIER_API_H__INCLUDED_ */
