/*******************************************************************************
  INTEL CONFIDENTIAL

  Copyright 2003 Intel Corporation All Rights Reserved.

  The source code contained or described herein and all documents related to
  the source code ("Material") are owned by Intel Corporation or its 
  suppliers or licensors.

  Title to the Material remains with Intel Corporation or its suppliers and
  licensors. The Material contains trade secrets and proprietary and
  confidential information of Intel or its suppliers and licensors.
  The Material is protected by worldwide copyright and trade secret laws and
  treaty provisions. No part of the Material may be used, copied, reproduced,
  modified, published, uploaded, posted, transmitted, distributed,
  or disclosed in any way without Intel's prior express written permission.

  No license under any patent, copyright, trade secret or other intellectual
  property right is granted to or conferred upon you by disclosure
  or delivery of the Materials, either expressly, by implication, inducement,
  estoppel or otherwise. Any license under such intellectual property rights
  must be express and approved by Intel in writing.

*******************************************************************************/

/**
 * ============================================================================
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2000 Network Processor
 *
 * = FILE
 *     dscp_classifier_defs.h
 *
 * = DESCRIPTION
 *      DSCP Classifier external API definition.
 *
 * ============================================================================
 */

#if !defined( _DSCP_CLASSIFIER_DEFS_H__INCLUDED_ )
#define _DSCP_CLASSIFIER_DEFS_H__INCLUDED_

/**
 * Number of symbols to be patched by the DSCP classifier core component.
 */
#if defined(CLASSIFIER_DSCP_PACKET_COUNTER_FEATURE)
#define CLASSIFIER_DSCP_PATCH_SYMBOLS       2
#else
#define CLASSIFIER_DSCP_PATCH_SYMBOLS       1
#endif

/**
 * Names of the symbols to be patched by the DSCP classifier core component.
 */
#define CLASSIFIER_DSCP_CONFIG_SRAM_BASE    "CLASSIFIER_DSCP_CONFIG_SRAM_BASE"

#if defined(CLASSIFIER_DSCP_PACKET_COUNTER_FEATURE)
#define CLASSIFIER_DSCP_STATS_SRAM_BASE     "CLASSIFIER_DSCP_STATS_SRAM_BASE"
#endif

/**
 * This enumeration defines messages specific to the DSCP core component.
 */
typedef enum ix_e_cc_classifier_dscp_messages
  {
    /**
     * This message adds an interface-specific set of classification rules.
     */
    IX_CC_CLASSIFIER_DSCP_MSG_ADD_IF_CONFIG = IX_CC_COMMON_MSG_ID_LAST,

    /**
     * This message deletes an interface-specific set of classification rules.
     */
    IX_CC_CLASSIFIER_DSCP_MSG_REMOVE_IF_CONFIG,
    
    /**
     * This message updates an interface-specific set of classification rules.
     */
    IX_CC_CLASSIFIER_DSCP_MSG_UPDATE_IF_CONFIG,
    
    /**                                                  
     * This message retrieves an interface-specific set of classification 
     * rules.
     */
    IX_CC_CLASSIFIER_DSCP_MSG_GET_IF_CONFIG,
    
    /**
     * This message changes default classification rules for the interfaces 
     * that do not have interface-specific rules configured.
     */
    IX_CC_CLASSIFIER_DSCP_MSG_SET_DEFAULT_RULES,
    
    /**
     * This message retrieves default classification rules for the interfaces
     * that do not have interface-specific rules configured.
     */
    IX_CC_CLASSIFIER_DSCP_MSG_GET_DEFAULT_RULES,
    
    /**
     * This message retrieves statistics associated with a <input port, 
     * DSCP value> pair.
     */
    IX_CC_CLASSIFIER_DSCP_MSG_GET_STATISTICS,

    /**
     * !!! Last message of the DSCP classifier core component. Add new 
     *     messages before this one !!!     
     */
    IX_CC_CLASSIFIER_DSCP_MSG_LAST

  } ix_cc_classifier_dscp_messages;

/**
 * Maximum number of the DSCP value.
 */
#define DSCP_MAX_VALUE                    (ix_uint32)(64)

#define DSCP_IPv4_HEADER_TYPE             0x06
#define DSCP_IPv6_HEADER_TYPE             0x07

/**
 * This macro frees a memory block allocated in the system area and sets the 
 * pointer to NULL.
 *
 * @param  __ptr               address of a memory block to be freed.
 */
#define _SYSTEM_DELETE( __ptr ) \
  { \
    ix_ossl_free( (void *)(__ptr) ); \
    (void *)(__ptr) = NULL; \
  } \

/**
 * The macro calculates an offset of the specified entry in a table.
 *
 * @param  __index             index of the entry.
 * @param  __size              size of the entry in bytes.
 *
 * @return The return value is an 32-bit offset value.
 */
#define _GET_OFFSET( __index, __size ) \
  ((ix_uint32)(__index) * (ix_uint32)(__size))

/**
 * The macro sets a bit field.
 *
 * @param  __val               value to be resolved as a boolean expression.
 */
#define _BIT_FLD( __val ) \
  (((ix_uint8)(__val) != 0) ? 1 : 0)

/**
 * This macro returns a pointer of the specified entry in a table. This is 
 * common macro used to obtain data from vairous tables.
 *
 * @param  __base              base address of an array.
 * @param  __index             index of the entry.
 * @param  __size              size of the entry in bytes.
 *
 * @return Pointer to the entry in the specified table/array.
 */
#define _GET_PTR( __base, __index, __size ) \
  ((ix_uint8 *)(__base) + _GET_OFFSET( __index, __size ))

/**
 * The macro returns the pointer to the structure describing core component's
 * control block. 
 */
#define _CONTEXT( __cntxt ) \
  ((_ix_cc_classifier_dscp_context *)(__cntxt))

/**
 * This macro returns the configuration structure of the DSCP classifier core
 * component. Pointer to this structure is held in core component's control 
 * block, which is internal to core component. 
 *
 * @param  arg_cntxt           pointer to the control block, internal to the
 *                             core component.
 *
 * @return If core component has been configured successfuly, this macro 
 *         returns pointer to the configuration structure maintained in the
 *         system memory. Otherwise, it returns NULL.
 */
#define _CONFIG( __cntxt ) \
    _CONTEXT( __cntxt )->_config


#if defined(CLASSIFIER_DSCP_PACKET_COUNTER_FEATURE)

/**
 * The macro returns statistics table entry size.
 *
 * @param  __cntxt             pointer to the control block.
 *
 * @return Size of the statistics table entry in bytes.
 */
#define STATS_ENTRY_SIZE( __cntxt ) \
    (DSCP_MAX_VALUE * _CONFIG( __cntxt )->_stats_entry_size )

#endif

/**
 * The macro returns configuration table entry size.
 *
 * @param  __cntxt             pointer to the control block.
 *
 * @return Size of the configuration table entry in bytes.
 */
#define TABLE_ENTRY_SIZE( __cntxt ) \
    (DSCP_MAX_VALUE * _CONFIG( __cntxt )->_table_entry_size )

#if defined( CLASSIFIER_DSCP_PACKET_COUNTER_FEATURE )

/**
 * The macro returns pointer to the specified entry in the statistics table.
 *
 * @param  __cntxt             pointer to the control block.
 * @param  __input_port        number of the port available in the system.
 *
 * @return Pointer to the specified entry in the statistics table.
 */
#define _STATS_ENTRY( __cntxt, __input_port ) \
  (_ix_cc_classifier_dscp_stats_entry *) \
      _GET_PTR( \
          _CONTEXT( __cntxt )->_dscp_stats_sram_base, \
          __input_port, \
          STATS_ENTRY_SIZE( __cntxt ) \
        ) \

#endif 

/**
 * The macro returns pointer to the specified entry in the configuration table.
 *
 * @param  __cntxt             pointer to the control block.
 * @param  __input_port        number of the port available in the system.
 *
 * @return Pointer to the specified entry in the configuration table.
 */
#define _TABLE_ENTRY( __cntxt, __input_port ) \
  (_ix_cc_classifier_dscp_table_entry *) \
      _GET_PTR( \
          _CONTEXT( __cntxt )->_dscp_table_sram_base, \
          __input_port, \
          TABLE_ENTRY_SIZE( __cntxt ) \
        ) \

/**
 * The macro locks a mutex. The handle to the object is stored in a control 
 * block of the core component.
 *
 * @param  __cntxt             pointer to the control block.
 *
 * @return IX_SUCCESS if successful or a valid ix_error token for failure.
 */
#define _MUTEX_LOCK( __cntxt ) \
    ix_ossl_mutex_lock( _CONTEXT( __cntxt )->_mutex, IX_OSSL_WAIT_FOREVER )

/**
 * The macro unlocks the mutex. The handle to the object is stored in a control
 * block of the core component.
 *
 * @param  __cntxt             pointer to the control block.
 *
 * @return IX_SUCCESS if successful or a valid ix_error token for failure.
 */
#define _MUTEX_UNLOCK( __cntxt ) \
    ix_ossl_mutex_unlock( _CONTEXT( __cntxt )->_mutex )

/**
 * This macro invokes user-defined callback function after message data 
 * processing.
 *
 * @param  arg_cntxt           pointer to the control block, internal to the
 *                             core component.
 * 
 * @param  arg_rslt            message processing operation state.
 *
 * @return IX_SUCCESS if successful or a valid error token for failure.
 */
#define _CALLBACK( arg_cntxt, arg_rslt ) \
  ((ix_cc_classifier_dscp_callback) \
    (_CB_CONTEXT( arg_cntxt )->_fn_callback))( arg_rslt, \
     _CB_CONTEXT( arg_cntxt )->_context ) \

/**
 * This macro invokes user-defined callback function after message data 
 * processing. Use this macro with extendent version of the callback function.
 *
 * @param  arg_cntxt           pointer to the control block, internal to the
 *                             core component.
 * @param  arg_data            reply message data.
 * @param  arg_rslt            message processing operation state.
 *
 * @return IX_SUCCESS if successful or a valid ix_error token for failure.
 */
#define _CALLBACK_EX( arg_cntxt, arg_data, arg_rslt ) \
  ((ix_cc_classifier_dscp_callback_ex) \
    (_CB_CONTEXT( arg_cntxt )->_fn_callback))( arg_rslt, arg_data, \
     _CB_CONTEXT( arg_cntxt )->_context ) \

/**
 * This macro creates a callback context for a message.
 *
 * @param  __fncb              address of a user-defined callback function.
 * @param  __cntxt             pointer to a user-defined application context.
 */
#define _CREATE_CB_CONTEXT( __fncb, __cntxt ) \
  _T( cb_context ) * __callback_cntxt__; \
  \
  __callback_cntxt__ = ix_ossl_malloc( _SIZE_OF( cb_context ) ); \
  if ( __callback_cntxt__ == NULL ) \
    { \
      return IX_ERROR_WARNING( \
          IX_CC_ERROR_OOM_SYSTEM, \
          ( "not enought free system memory available" ) \
         ); \
    } \
  \
  __callback_cntxt__->_fn_callback = (void *)__fncb; \
  __callback_cntxt__->_context     = __cntxt; \
    
/**
 * This macro frees memory allocated for a callback context. It is designed to 
 * be used inside of the message helper library functions.
 */
#define _DELETE_CB_CONTEXT() \
  _SYSTEM_DELETE( __callback_cntxt__ )

/**
 * This macro frees memory allocated for the specified callback context. 
 *
 * @param  __cb_cntxt         pointer to the callback context to be freed.
 */
#define _DELETE_CB_CONTEXT_EX( __callback_cntxt ) \
   _SYSTEM_DELETE( __callback_cntxt )

/**
 *  The macro returns pointer to the callback context.
 *
 *  @param  __cb_cntxt
 */
#define _CB_CONTEXT( __cb_cntxt ) \
   ((_T( cb_context ) *)(__cb_cntxt))

/**
 * This macro sends a message to the DSCP classifier core component.
 *
 * @param  __msg_type          message type identifier.
 * @param  __data              pointer to the message data.
 * @param  __size              size of message data.
 */
#define _SEND_MESSAGE( __msg, __data, __size ) \
  { \
    ix_error __send_result = \
      ix_cc_msup_send_async_msg( \
          IX_CC_CLASSIFIER_DSCP_MSG_INPUT, \
          _ix_cc_classifier_dscp_callback, \
          (void *)__callback_cntxt__, \
          __msg, (void *)__data, __size ); \
    \
    if ( __send_result != IX_SUCCESS ) \
      { \
        _DELETE_CB_CONTEXT(); \
        \
        return IX_ERROR_WARNING( \
            IX_CC_ERROR_SEND_FAIL, \
            ( "failed to send asynchronous message (" #__msg ")" ) \
          ); \
      } \
  } \

/**
 * This macro sends a message to the DSCP classifier core component. Reply 
 * message wil be send via extended callback function.
 *
 * @param  __msg_type          message type identifier.
 * @param  __data              pointer to the message data.
 * @param  __size              size of the message data.
 */
#define _SEND_MESSAGE_EX( __msg, __data, __size ) \
  { \
    ix_error __result__ = \
      ix_cc_msup_send_async_msg( \
          IX_CC_CLASSIFIER_DSCP_MSG_INPUT, \
          _ix_cc_classifier_dscp_callback_ex, \
          (void *)__callback_cntxt__, \
          __msg, (void *)__data, __size ); \
    \
    if ( __result__ != IX_SUCCESS ) \
      { \
        _DELETE_CB_CONTEXT(); \
        \
        return IX_ERROR_WARNING( \
            IX_CC_ERROR_SEND_FAIL, \
            ( "failed to send asynchronous message (" #__msg ")" ) \
          ); \
      } \
   } \

#define _MSG_DATA( __msg ) \
    ((_ix_cc_classifier_dscp_msg_data *)(__msg))

/**
 * This macro extracts input port number from the message.
 *
 * @param  __msg               pointer to the message data.
 *
 * @return Number of an input port extracted from the message.
 */
#define MSG_DATA_INPUT_PORT( __msg ) \
    _MSG_DATA( __msg )->_input_port

/**
 * This macro extracts classification rules from the message.
 *
 * @param  __msg               pointer to the message buffer.
 *
 * @return Pointer to the classification rules table extracted
 *         from the message.
 */
#define MSG_DATA_RESULTS( __msg ) \
    _MSG_DATA( __msg )->_data_ex._results

/**
 * This macro extracts a DSCP value from the message.
 *
 * @param  __msg               pointer to the message buffer.
 *
 * @return The DSCP value extracted from the message.
 */
#define MSG_DATA_DSCP_VALUE( __msg ) \
    _MSG_DATA( __msg )->_data_ex._dscp

/** 
 * This macro invokes 'add interface configuration' library API function.
 *
 * @param  __msg               pointer to the message data buffer.
 * @param  __cntxt             pointer to the core component's control block.
 * @param  __rslt              output argument - operation state.
 *
 * @see ix_cc_classifier_dscp_add_if_config()
 */
#define _CLASSIFIER_DSCP_ADD_IF_CONFIG( __msg, __cntxt, __rslt ) \
  __rslt = ix_cc_classifier_dscp_add_if_config( \
      MSG_DATA_INPUT_PORT( __msg ), MSG_DATA_RESULTS( __msg ), \
      (void *)(__cntxt) ) \

/**
 * This macro invokes 'remove interface configuration' library API funciton.
 *
 * @param  __msg
 * @param  __cntxt
 * @param  __rslt
 *
 * @see ix_cc_classifier_dscp_remove_if_config
 */
#define _CLASSIFIER_DSCP_REMOVE_IF_CONFIG( __msg, __cntxt, __rslt ) \
  __rslt = ix_cc_classifier_dscp_remove_if_config( *(ix_uint16 *)__msg, \
    (void *)(__cntxt) ) \

/**
 * This macro invokes 'update interface configuraiton' library API function.
 *
 * @param  __msg
 * @param  __cntxt
 * @param  __rslt
 *
 * @see ix_cc_classifier_dscp_update_if_config
 */
#define _CLASSIFIER_DSCP_UPDATE_IF_CONFIG( __msg, __cntxt, __rslt ) \
  __rslt = ix_cc_classifier_dscp_update_if_config( \
      MSG_DATA_INPUT_PORT( __msg ), \
      MSG_DATA_RESULTS( __msg ), (void *)(__cntxt) ) \

/**
 * This macro invokes 'get internface configuration' libray API function.
 *
 * @param  __msg
 * @param  __cntxt
 * @param  __rslt
 *
 * @see ix_cc_classifier_dscp_get_if_config
 */
#define _CLASSIFIER_DSCP_GET_IF_CONFIG( __msg, __rslt, __cntxt, __err ) \
  __err = ix_cc_classifier_dscp_get_if_config( \
      MSG_DATA_INPUT_PORT( __msg ), \
      __rslt, (void *)(__cntxt) ) \

/**
 * This macro invokes 'set default rules' library API function.
 *
 * @param  __msg
 * @param  __cntxt
 * @param  __rslt
 *
 * @see ix_cc_classifier_dscp_set_def_rules
 */
#define _CLASSIFIER_DSCP_SET_DEF_RULES( __msg, __cntxt, __rslt ) \
  __rslt = ix_cc_classifier_dscp_set_def_rules( \
    *(ix_cc_classifier_dscp_result **)(__msg), (void *)(__cntxt) )\

/**
 * This macro invokes 'get default rules' library API function.
 *
 * @param  __msg
 * @param  __cntxt
 * @param  __rslt
 *
 * @see ix_cc_classifier_dscp_get_def_rules
 */
#define _CLASSIFIER_DSCP_GET_DEF_RULES( __rslt, __cntxt, __err ) \
  __err = ix_cc_classifier_dscp_get_def_rules( __rslt, (void *)(__cntxt) )\

/**
 * This macro invokes 'get statistics' library API function.
 *
 * @param  __msg
 * @param  __cntxt
 * @param  __rslt
 *
 * @see ix_cc_classifier_dscp_get_statistics()
 */
#define _CLASSIFIER_DSCP_GET_STATISTICS( __msg, __cntxt, __out, __rslt ) \
  __rslt = ix_cc_classifier_dscp_get_statistics( \
      MSG_DATA_INPUT_PORT( __msg ), MSG_DATA_DSCP_VALUE( __msg ), \
      (ix_cc_classifier_dscp_statistics *)(__out), (void *)(__cntxt) ) \

/**
 * The macro returns size of a internal datatype in bytes. 
 */
#define _SIZE_OF( __type_name ) \
  sizeof( _ix_cc_classifier_dscp_##__type_name )

#define _T( __type_name ) \
  _ix_cc_classifier_dscp_##__type_name

#define _ENTRY( __ptr ) \
    ((_ix_cc_classifier_dscp_table_entry *)(__ptr))

/* _RESULT */
#define _RESULT( __ptr ) \
    ((ix_cc_classifier_dscp_result *)(__ptr))

/* _B_INDEX */
#define _B_INDEX( __in_port ) \
    (((ix_uint16)(__in_port)) / 8)

/* _B_PTR */
#define _B_PTR( __in_port, __cntxt ) \
    ((ix_uint8 *)(_CONTEXT( __cntxt )->_cfg_bitmap + _B_INDEX( __in_port )))

/* _B_MASK */
#define _B_MASK( __in_port ) \
    (ix_uint8)(1 << (((ix_uint16)(__in_port)) & 7))

/* __IS_SET */
#define __IS_SET( __in_port, __cntxt ) \
    (((*_B_PTR( __in_port, __cntxt )) & _B_MASK( __in_port )) != 0)

/* __SET_BIT */
#define __SET_BIT( __in_port, __cntxt ) \
    ((*_B_PTR( __in_port, __cntxt )) |= _B_MASK( __in_port ))

/* __RESET_BIT */
#define __RESET_BIT( __in_port, __cntxt ) \
    ((*_B_PTR( __in_port, __cntxt )) &= ~_B_MASK( __in_port ))

/* RESULT_TO_ENTRY */
#define RESULT_TO_ENTRY( __to, __from, __ipc, __cntxt ) \
    _ix_cc_classifier_dscp_convert_a( \
        (_ix_cc_classifier_dscp_table_entry *)(__to), \
        ( ix_cc_classifier_dscp_result *)(__from), \
        (ix_uint32)(__ipc), (void *)(__cntxt) \
      ) \

/* ENTRY_TO_RESULT */
#define ENTRY_TO_RESULT( __to, __from, __cntxt ) \
    _ix_cc_classifier_dscp_convert_b( \
        ( ix_cc_classifier_dscp_result * )(__to), \
        (_ix_cc_classifier_dscp_table_entry *)(__from), \
        (void *)(__cntxt) \
      ) \

#endif /* _DSCP_CLASSIFIER_DEFS_H__INCLUDED_ */
