/*******************************************************************************
  INTEL CONFIDENTIAL

  Copyright 2003 Intel Corporation All Rights Reserved.

  The source code contained or described herein and all documents related to
  the source code ("Material") are owned by Intel Corporation or its 
  suppliers or licensors.

  Title to the Material remains with Intel Corporation or its suppliers and
  licensors. The Material contains trade secrets and proprietary and
  confidential information of Intel or its suppliers and licensors.
  The Material is protected by worldwide copyright and trade secret laws and
  treaty provisions. No part of the Material may be used, copied, reproduced,
  modified, published, uploaded, posted, transmitted, distributed,
  or disclosed in any way without Intel's prior express written permission.

  No license under any patent, copyright, trade secret or other intellectual
  property right is granted to or conferred upon you by disclosure
  or delivery of the Materials, either expressly, by implication, inducement,
  estoppel or otherwise. Any license under such intellectual property rights
  must be express and approved by Intel in writing.

*******************************************************************************/

/**
 * ============================================================================
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2000 Network Processor
 *
 * = FILE
 *     dscp_classifier_types.h
 *
 * = DESCRIPTION
 *      The DSCP classifier core component internals. This file provides 
 *      definitions of the internal data types of the DSCP classifier core 
 *      component.
 *
 * ============================================================================
 */

#if !defined( _DSCP_CLASSIFIER_TYPES_H__INCLUDED_ )
#define _DSCP_CLASSIFIER_TYPES_H__INCLUDED_

/**
 * This data type is internal to the DSCP classifier core component. It defines 
 * core component's configuration data structure.
 */
typedef struct _ix_s_cc_classifier_dscp_config
  {    
    /** 
     * Initial value of flow identifier.
     */
    ix_uint32   _default_flow_id;

    /**
     * Initial value of class identifier.
     */
    ix_uint32   _default_class_id    : 16,
                _reserved            : 10,
                _default_flag_ipc    : 1,
                _default_flag_stats  : 1,
                _default_flag_tc     : 1,
                _default_flag_remark : 1,
                _default_color_id    : 2; 
    
    /**
     * Number of SRAM channel taken from the registry.
     */
    ix_uint32   _sram_channel;

    /**
     * The number of input ports in the system.
     */
    ix_uint32   _in_ports;

    /**
     * Number of bytes occupied by a single entry in a configuration table.
     */
    ix_uint32   _table_entry_size;

#if defined(CLASSIFIER_DSCP_PACKET_COUNTER_FEATURE)

    /**
     * Number of bytes occupied by a single entry in a statistics table.
     */
    ix_uint32   _stats_entry_size;

#endif

    /**
     * MEv2 mask
     */
    ix_uint32   _mev2_mask;

  } _ix_cc_classifier_dscp_config;

/**
 * This data type is internal to the DSCP classifier core component. It defines
 * a single entry in the classifier's configuration table.
 */
typedef struct _ix_s_cc_classifier_dscp_table_entry
  {
    /**
     * Identifier of a traffic flow to which the packet belongs.
     */
    ix_uint32   _id_flow;
    
    /**
     * A relative queue number within an output port.
     */
    ix_uint32   _id_class        : 16,
                
    /**
     * Reserved bits, not used now.
     */               
                _reserved        : 10,

    /**
     * Input Port Configured flag. This flag indicates whether the input port 
     * corresponding to the entry has its own configuration.
     */
                _flag_ipc        : 1,
                
    /**
     * This flag indicates whether the microblock should gather statistics for 
     * the rule.
     */                
                _flag_stats      : 1,

    /**
     * Trafic Conditioning flag. This flag indicates whether the packets must 
     * be metered (meaningful only if _flag_ipc flag is set).
     */
                _flag_tc         : 1,
                
    /** 
     * This flag indicates whether DSCP value in the packets must be remarked
     * (meaningful only if _flag_ipc flag is set).
     */
                _flag_remark     : 1,
     
    /**
     * Packet drop precendence level: green, yellow or red color.
     */
                _id_color        : 2;

  } _ix_cc_classifier_dscp_table_entry;

#if defined(CLASSIFIER_DSCP_PACKET_COUNTER_FEATURE)

/**
 * This data type is internal to the DSCP classifier core component. It defines
 * a single entry in the classifier's statistics table.
 */
typedef struct _ix_s_cc_classifier_dscp_stats_entry
  {
    /**
     * This counter stores a number of packets matching the rule.
     */
    ix_uint64  _packet_counter;

    /**
     * This counter keeps a number of bytes matching the rule.
     */
    ix_uint64  _byte_counter;

  } _ix_cc_classifier_dscp_stats_entry;

#endif

/**
 * This datatype is internal to the DSCP classifier core component.
 */
typedef struct _ix_s_cc_classifier_dscp_msg_data
  {
    /**
     * Number of the input port in the system.
     */
    ix_uint16 _input_port;

    union _ix_u_data    
      {
        ix_cc_classifier_dscp_result * _results;
        ix_uint8                       _dscp;
      } _data_ex;

  } _ix_cc_classifier_dscp_msg_data;

/**
 * This data type is internal to the DSCP classifier core component.
 */
typedef struct _ix_s_classifier_dscp_cb_context
  {
    /**
     * Address of a user defined callback function.
     */
    void * _fn_callback;
    
    /**
     * Pointer to the user defined application context.
     */
    void * _context;

  } _ix_cc_classifier_dscp_cb_context;

/**
 * This data type is internal to the DSCP classifier core component. It defines
 * core component's execution context (control block).
 */
typedef struct _ix_s_cc_classifier_dscp_context
  {
    /**
     * Default classification rules for the interfaces that do not have thier
     * own interface-specific classification rules.
     */
    _ix_cc_classifier_dscp_table_entry * _dscp_default_rules;
     
    /**
     * Handle to the core component.
     */ 
    ix_cc_handle                         _handle;
         
    /**
     * Base address of the classifier's configuration table.
     */
    void *                               _dscp_table_sram_base;

    /**
     * Default configuration of the core component.
     */
    _ix_cc_classifier_dscp_config      * _config;

#if defined(CLASSIFIER_DSCP_PACKET_COUNTER_FEATURE)       
    
    /**
     * Base address of the classifier's statistics table.
     */
    void *                               _dscp_stats_sram_base;

#endif 

    /**
     * Handle of a symchornization object
     */
    ix_ossl_mutex_t                      _mutex;

    /**
     * Bitamp indicating which input port has interface-specific configuration.
     */
    ix_uint8 *                           _cfg_bitmap;

  } _ix_cc_classifier_dscp_context;

/* IX_LITTLE_ENDIAN */
#if   IX_BYTE_ORDER == IX_LITTLE_ENDIAN 
  
  typedef struct _ix_s_64bit_counter
    {
      ix_uint32  _l;
      ix_uint32  _h;
    } _ix_64bit_counter;

/* IX_BIG_ENDIAN */
#elif IX_ORDER_TYPE == IX_BIG_ENDIAN

  typedef struct _ix_s_64bit_counter
    {
      ix_uint32  _h;
      ix_uint32  _l;
    } _ix_64bit_counter;

#else
  #error "unsupported byte ordering"
#endif

#endif /* _DSCP_CLASSIFIER_TYPES_H__INCLUDED_ */
