/*******************************************************************************
  INTEL CONFIDENTIAL

  Copyright 2003 Intel Corporation All Rights Reserved.

  The source code contained or described herein and all documents related to
  the source code ("Material") are owned by Intel Corporation or its 
  suppliers or licensors.

  Title to the Material remains with Intel Corporation or its suppliers and
  licensors. The Material contains trade secrets and proprietary and
  confidential information of Intel or its suppliers and licensors.
  The Material is protected by worldwide copyright and trade secret laws and
  treaty provisions. No part of the Material may be used, copied, reproduced,
  modified, published, uploaded, posted, transmitted, distributed,
  or disclosed in any way without Intel's prior express written permission.

  No license under any patent, copyright, trade secret or other intellectual
  property right is granted to or conferred upon you by disclosure
  or delivery of the Materials, either expressly, by implication, inducement,
  estoppel or otherwise. Any license under such intellectual property rights
  must be express and approved by Intel in writing.

*******************************************************************************/

/**
 * ============================================================================
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2000 Network Processor
 *
 * = FILE
 *     ix_cc_dscp_classifier.h
 *
 * = DESCRIPTION
 *     The DSCP classifier core component initialization/finalization.
 *
 * ============================================================================
 */

#if !defined( _IX_CC_DSCP_CLASSIFIER_H__INCLUDED__frmwrk_ )
#define _IX_CC_DSCP_CLASSIFIER_H__INCLUDED__frmwrk_

/**
 *  This enumeration defines error codes specific to the DSCP classifier core 
 *  component.
 */
typedef enum ix_e_cc_classifier_dscp_error_code
  {
    IX_CC_CLASSIFIER_DSCP_ERROR_FIRST = IX_ERROR_MAKE_GROUP( CC_CLASSIFIER_DSCP ),
    IX_CC_CLASSIFIER_DSCP_SUCCESS     = IX_CC_CLASSIFIER_DSCP_ERROR_FIRST,

    /**
     * Invalid arg_ppContext specified as an argument in the core component 
     * initialization function. *arg_ppContext should be NULL.
     */
    IX_CC_CLASSIFIER_DSCP_ERROR_NOT_NULL,
    
    /**
     * This error occures only if we are trying to configure already set up a 
     * core component's configuration table.
     */
    IX_CC_CLASSIFIER_DSCP_ERROR_CONFIG_TABLE_EXIST,
    
    /**
     * This error indicates a failure while reading static configuration data 
     * from system application registry.
     */
    IX_CC_CLASSIFIER_DSCP_ERROR_REGISTRY,
    
    /**
     * Configuration of the DSCP Marker micoblock failed.
     */
    IX_CC_CLASSIFIER_DSCP_ERROR_PATCH_SYMBOLS,

    /**
     * Some error occured while freeing a buffer
     */
    IX_CC_CLASSIFIER_DSCP_ERROR_BUFF_FREE,

    /**
     * Classifier does not support statistics gathering.
     */
    IX_CC_CLASSIFIER_DSCP_ERROR_STATS_NOT_SUPPORTED,

    /**
     * Invalid buffer handle specified as an argument of a function
     */
    IX_CC_CLASSIFIER_DSCP_ERROR_INVALID_HANDLE,

    IX_CC_CLASSIFIER_DSCP_ERROR_INVALID_ENTRY_SIZE,

    /* unknown and last error codes */
    IX_CC_CLASSIFIER_DSCP_ERROR_UNKNOWN,
    IX_CC_CLASSIFIER_DSCP_ERROR_LAST
  
  } ix_cc_classifier_dscp_error_code;

#if defined( __cplusplus )
extern "C" {
#endif /* __cplusplus */

/**
 * @brief  Initialization of the DSCP classifer core component
 *
 * This function initializes the core component. It should be called and 
 * returned successfuly before any other function in the core component can be 
 * called. The function:
 *
 *      - allocates SRAM memory for the configuration table and initializes it 
 *        with default configuration rule read from the system registry,
 *
 *      - if the microblock supports statistics gathering, the function 
 *        allocates SRAM memory for 64-bit statistics table and initializes it 
 *        with zero values,
 *
 *      - patches DSCP calssifier microcode with imported variables taken from 
 *        the system application registry,
 *
 *      - registers a packet handler for inputs (single source mode),
 *
 *      - registers a message handler for inputs (multiple source mode),
 *
 *      - allocates and initlizes a bitmap to store information which input 
 *        ports have its own interface-specific set of classification rules,
 *
 *      - allocates a control block, and returns a pointer to this block via 
 *        arg_ppContext function's argument.
 *
 * @param  arg_hCC           handle to the core component (cannot be NULL).
 *
 * @param  arg_ppContext     a placeholder where the pointer to the control 
 *                           block allocated by the core component will be 
 *                           stored. The control block is internal to the core
 *                           component and contains internal data structures.
 *
 * @return IX_SUCCESS if successful or a valid ix_error token for failure. The 
 *         following error codes might be returned via the token:
 *\n
 *\n       IX_CC_ERROR_NULL                      - one of the input arguments 
 *                                                 is NULL (either arg_hCC or 
 *                                                 arg_ppContext).
 *\n       IX_CC_ERROR_OOM_SYSTEM                - system memory allocation 
 *                                                 failure.
 *\n       IX_CC_ERROR_OOM_SRAM                  - SRAM memory allocation 
 *                                                 failure.
 *\n       IX_CC_DSCP_ERROR_CONFIG_TABLE_EXISTS  - configuration table already
 *                                                 exists.
 *\n       IX_CC_DSCP_ERROR_NOT_NULL             - *arg_ppContext isn't NULL.  
 *\n       IX_CC_DSCP_ERROR_REPOSITORY           - access failure to the system
 *                                                 application registry.
 *\n       IX_CC_DSCP_ERROR_PATCH_SYMBOLS        - symbols patching to the 
 *                                                 microblock failed.
 *\n       IX_OSSL_ERROR_MUTEX_INIT_FAILED       - the function failed to 
 *                                                 create a mutex object.
 */
ix_error ix_cc_classifier_dscp_init( ix_cc_handle arg_hCC,
                                     void **      arg_ppContext );

/**
 * @brief  Finalization of the DSCP classifer core component
 *
 * The function terminates the DSCP classifier core component. It will be 
 * executed when the execution engine running the core component is being shut 
 * down. This function:
 *
 *      - frees SRAM memory allocated for configuration table and statistics
 *        table (if statistics gathering enabled),
 *
 *      - removes packet and message handlers,
 *
 *      - frees memory allocated for the control block.
 *
 * @param  arg_hCC           handle to the core component (cannot be NULL).
 *
 * @param  arg_pContext      pointer to the internal control block.
 *
 * @return IX_SUCCESS if successful or a vaild ix_error token for failure. The 
 *         following error codes might be returned via the token:
 *\n
 *\n       IX_CC_ERROR_NULL                      - arg_pContext is NULL. 
 *\n       IX_CC_DSCP_ERROR_INVALID_CC_HANDLE    - invalid core component
 *                                                 handle specified in the
 *                                                 arg_hCC argument.
 */
ix_error ix_cc_classifier_dscp_fini( ix_cc_handle arg_hCC,
                                     void *       arg_pContext );

/**
 * @brief  Message handling function of the DSCP classifier core component.
 *
 * The core component uses one message handler for processing all types of 
 * messages. The DSCP classifer CC receives messages from the DiffServ 
 * application. The suit of supported messages directly reflects a subset of 
 * Lookup Library API primitives pertinent to DSCP classification. The handler
 * function parses message payloads and calls appropriate DSCP classificator
 * CC Library API functions to process the message. The supported messages are
 * listed below:
 *\n  
 *\n       IX_CC_DSCP_MSG_ADD_IF_CONFIG     - add an interface-specific set of
 *                                            classification rules.
 *\n       IX_CC_DSCP_MSG_REMOVE_IF_CONFIG  - delete an interface-specific set 
 *                                            of classification rules.
 *\n       IX_CC_DSCP_MSG_UPDATE_IF_CONFIG  - update an interface-specific set 
 *                                            of classification rules.
 *\n       IX_CC_DSCP_MSG_GET_IF_CONFIG     - return an interface-specific set 
 *                                            of classification rules.      
 *\n       IX_CC_DSCP_MSG_SET_DEFAULT_RULES - change default classification 
 *                                            rules for the interfaces that do
 *                                            not have interface-specific rules
 *                                            configured,
 *\n       IX_CC_DSCP_MSG_GET_DEFAULT_RULES - return default classification 
 *                                            rules for the interfaces that do
 *                                            not have interface-specific rules
 *                                            configured,
 *\n       IX_CC_DSCP_MSG_GET_STATS         - return statistics associated with 
 *                                            a <input port, DSCP value> pair.
 *
 * @param  arg_hBuffer      buffer handle embedding information for the message
 *                          passed in the arg_nMsgType argument.
 *
 * @param  arg_nMsgType     message type (see above),
 *
 * @param  arg_pContext     pointer to the control block that is passing to the
 *                          core component when a message arrives.
 *
 * @return IX_SUCCES if successful or a valid ix_error token for failure. The 
 *         following error codes might be returned via the token:
 *\n
 *\n       IX_CC_ERROR_NULL                 - the arg_pContext argument is 
 *                                            specified as NULL or a hash table 
 *                                            has not been created,
 *\n       IX_CC_ERROR_UNDEFINED_MSG        - unknown message type recieved,
 *\n       IX_CC_ERROR_SEND_FAIl            - reply message sending failure,
 *\n       IX_CC_DSCP_ERROR_INVALID_HANDLE  - buffer handle is NULL,
 *\n       IX_CC_DSCP_ERROR_FREE_BUFFER     - buffer memory deallocation 
 *                                            failure.
 */
ix_error ix_cc_classifier_dscp_msg_handler( ix_buffer_handle arg_hBuffer,
                                            ix_uint32        arg_nMsgType,
                                            void *           arg_pContext );

/** 
 * @brief  Packet handling function of the DSCP classifier core component.
 *
 * The core component uses packet handler for processing local IPv6 and IPv4
 * packets generated by Stack Driver. Packets other than IPv6 or IPv4 are sent
 * to IX_CC_CLASSIFIER)DSCP_INV_IP_OUTPUT, and the handler reports an error.
 *
 * For IPv6 or IPv4 packets, the function retrieves DSCP value from the IP 
 * header and lookups configuration entry in the SRAM configuration table. Next
 * the packet handler updates QoS related metadata variables according to the
 * lookup results.
 *
 * If the Traffic Condition flag is set in the configuration entry, the 
 * function sends the packet to the IX_CC_CLASSIFIER_DSCP_METER output. 
 *
 * If the remark flag is set int the configuration entry, the function changes
 * DSCP value carried in the packet. Next it sends the packet to the forwarder
 * core component (either IX_CC_CLASSIFIER_DSCP_IP4_FWD_OUTPUT or 
 * IX_CC_CLASSIFIER_DSCP_IP6_FWD_OUTPUT).
 *
 * If CLASSIFIER_DSCP_PACKET_COUNTER_FEATURE global flag is asserted and 
 * statistics for the rule are enabled, the packet handler updates per-rule 
 * statistics. The core component implements the same statistics update 
 * algorithm as the microblock. This ensures synchronization in access to the
 * statistics variables.
 *
 * @param  arg_hBuffer      buffer handle to the IP packet,
 *
 * @param  arg_nExCode      equals to zero when Stack Driver originates the 
 *                          packet,
 *
 * @param  arg_pContext     pointer to a control blok that is passing to the 
 *                          core component when a message arrives.
 *
 * @return IX_SUCCESS if successful or a valid ix_error token for failure. The
 *         folowing error codes might be returned via the token:
 *\n
 *\n       IX_CC_DSCP_ERROR_INVALID_HANDLE  - buffer handle is NULL,
 *\n       IX_CC_ERROR_NULL                 - the arg_pContext arguemtn is NULL
 *                                            or the hash table has not been 
 *                                            created,
 *\n       IX_CC_ERROR_SEND_FAIL            - reply message sending failure,
 *\n       IX_CC_ERROR_INTERNAL             - the function failed to get the 
 *                                            packet meta data or the packet 
 *                                            payload.
 */
ix_error ix_cc_classifier_dscp_pkt_handler( ix_buffer_handle arg_hBuffer,
                                            ix_uint32        arg_nExCode,
                                            void *           arg_pContext );

#if defined( __cplusplus )
}
#endif /* __cplusplus */

#endif /* _IX_CC_DSCP_CLASSIFIER_H__INCLUDED__frmwrk_ */
