/*******************************************************************************
  INTEL CONFIDENTIAL

  Copyright 2003 Intel Corporation All Rights Reserved.

  The source code contained or described herein and all documents related to
  the source code ("Material") are owned by Intel Corporation or its 
  suppliers or licensors.

  Title to the Material remains with Intel Corporation or its suppliers and
  licensors. The Material contains trade secrets and proprietary and
  confidential information of Intel or its suppliers and licensors.
  The Material is protected by worldwide copyright and trade secret laws and
  treaty provisions. No part of the Material may be used, copied, reproduced,
  modified, published, uploaded, posted, transmitted, distributed,
  or disclosed in any way without Intel's prior express written permission.

  No license under any patent, copyright, trade secret or other intellectual
  property right is granted to or conferred upon you by disclosure
  or delivery of the Materials, either expressly, by implication, inducement,
  estoppel or otherwise. Any license under such intellectual property rights
  must be express and approved by Intel in writing.

*******************************************************************************/

/**
 * ============================================================================
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2000 Network Processor
 *
 * = FILE
 *     ix_cc_dscp_classifier_msghlp.h
 *
 * = DESCRIPTION
 *        DSCP Marker message helper API definition. This is wrapper library 
 *     that  facilitates  sending  messages  to  the  DSCP  classifier  core 
 *     component.  All  API  functions  are  asynchronous: a core  component 
 *     reports operation status in a callback routine.
 *
 * ============================================================================
 */

#if !defined( _IX_CC_DSCP_CLASSIFIER_MSGHLP_H__INCLUDED_ )
#define _IX_CC_DSCP_CLASSIFIER_MSGHLP_H__INCLUDED_

/**
 * @brief Classification Result Data Type
 * 
 * Representation of the classification result for <input port, DSCP value> 
 * pair.
 */
typedef struct ix_s_cc_classifier_dscp_result
  {
    /**
     * Flow identifier
     */
    ix_uint32 flowId;   

    /**
     * Class identyfier
     */
    ix_uint16 classId;
    
    /**
     * Color identifier 
     */
    ix_uint8  colorId;

    /**
     * Statistics Flag. This field indicates whether statistics shall be 
     * gathered for the rule.
     */
    ix_uint8  statistics;
    
    /**
     * Traffic Conditioning Flag. This field indicates whether metering is the 
     * next processing step.
     */
    ix_uint8  TC;
    
    /**
     * Remark Flag. This field indicates whether the DSCP value should be 
     * remarked.
     */
    ix_uint8  remark;

  } ix_cc_classifier_dscp_result;

/**
 * @brief Statistics Data Type
 *
 * Statistics associated with <input port, DSCP value> pair consist of two 
 * 64-bit counters: packet counter and byte counter.
 */
typedef struct ix_s_cc_classifier_dscp_statistics
  {
    /**
     * number of packets for the pair
     */
    ix_uint64 packets;

    /**
     * number of bytes for the pair
     */
    ix_uint64 bytes;

  } ix_cc_classifier_dscp_statistics;

#if defined( __cplusplus )
extern "C" {
#endif /* __cplusplus */

/**
 * @brief  Definition of a default callback function
 *
 * This function is invoking by core component to report an operation status.
 *
 * @param  arg_Result       indiacates an error conditions for the call: 
 *                          IX_SUCCESS if successful or a vaild ix_error 
 *                          token for failure. 
 *
 * @param  arg_pUserContext pointer to the client-defined context, provided as
 *                          a parameter in the API call. Used by the caller to 
 *                          match asynchronous request with response.
 *
 * @return IX_SUCCESS if successful or a valid ix_error token for failure.
 */
typedef ix_error (* ix_cc_classifier_dscp_callback)( ix_error arg_Result,
                                                     void *   arg_pUserContext );

/**
 * @brief  Definition of a data retrieving callback function
 *
 * This function is invoking by core component to report a status of data
 * retrieving operation.
 *
 * @param  arg_Result       indicates an error conditions for the call:
 *                          IX_SUCCESS if successful or a valid ix_error token
 *                          for failure.
 *
 * @param  arg_pStatistics  if the error code is IX_SUCCESS, the argument 
 *                          contains statistics associated with the requested
 *                          <input port, DSCP value> pair. Otherwise, the 
 *                          argument's value is NULL.
 *
 * @param  arg_pUserContext pointer to the client-defined context, provided as
 *                          a parameter in the API call. Used by the caller to
 *                          match asynchronous request with a response.
 *
 * @result IX_SUCCESS if successful or a valid ix_error token for failure.
 */
typedef ix_error (* ix_cc_classifier_dscp_callback_ex)( ix_error arg_Result,
                                                        void * arg_pData,
                                                        void * arg_pUserContext );
/**
 * @brief  Sends "add interface configuration" message
 *
 * This function builds and sends "add inteface configuration" message to the
 * classifier core component. If an interface-specific configuration already 
 * exists for the input port, the core component returns an error code. 
 * Otherwise it uses the arg_pClassifResults argument (a set of classification
 * results), to fill in all the configuration entries for the interface.
 *
 * @param  arg_inputPort       interface for which the classification results 
 *                             are added
 *
 * @param  arg_pResults        64-element table indexed with DSCP value. Each 
 *                             entry of this table contains QoS parameters 
 *                             assigned to packets received from the input port 
 *                             and carrying the DSCP value (cannot be NULL).
 *                          
 * @param  arg_Callback        a client provided callback function. The core 
 *                             component invokes this function to report an 
 *                             operation status (cannot be NULL).
 *
 * @param  arg_pUserContext    pointer to client defined context. This value is 
 *                             opaque to the core component. The context is 
 *                             returned in a callback function.
 *
 * @result IX_SUCCESS if successful or a valid ix_error token for failure. The
 *         following error codes might be returned via the token:
 *\n
 *\n       IX_CC_ERROR_OOM_SYSTEM                - not enought system memory 
 *                                                 available to build and asend
 *                                                 a message.
 *\n       IX_CC_ERROR_NULL                      - one of the input arguments 
 *                                                 is NULL.
 *\n       IX_CC_ERROR_SEND_FAIL                 - failure at sending a message
 *
 * @see    ix_cc_classifeir_dscp_add_if_config
 */
ix_error ix_cc_classifier_dscp_async_add_if_config( 
            ix_uint16                           arg_inputPort,
            ix_cc_classifier_dscp_result      * arg_pResults,
            ix_cc_classifier_dscp_callback      arg_Callback,
            void                              * arg_pUserContext
          );

/**
 * @brief  Sends "remove interface configuration" message
 *
 * This function builds and sends "remove interface configuration" message to 
 * the classifier core component. If the input port has interface-specific 
 * configuration, the core component deletes interface-specific set of 
 * classification rules for the input port and applies the default settings.
 * 
 * @param arg_inputPort        identifies the input port whose configuration is
 *                             deleted.
 *
 * @param arg_Callback         a client provided callback function. The core 
 *                             component invokes this funciton to report an 
 *                             operation status (cannot be NULL).
 *
 * @param arg_pUserContext     pointer to client defined context. This value is
 *                             opaque to the core component. The context is 
 *                             returned in a callback function.
 *
 * @result IX_SUCCESS if successful or a valid ix_error token for failure. The
 *         following error codes might be returned via the token:
 *\n
 *\n       IX_CC_ERROR_OOM_SYSTEM                - not enought system memory 
 *                                                 available to build and asend
 *                                                 a message.
 *\n       IX_CC_ERROR_NULL                      - one of the input arguments 
 *                                                 is NULL.
 *\n       IX_CC_ERROR_SEND_FAIL                 - failure at sending a message
 *
 * @see    ix_cc_classifier_dscp_remove_if_config
 */
ix_error ix_cc_classifier_dscp_async_remove_if_config(
            ix_uint16                           arg_inputPort,
            ix_cc_classifier_dscp_callback      arg_Callback,
            void                              * arg_pUserContext
          );

/**
 * @brief  Sends "update interface configuration" message
 *
 * This function builds and sends "update interface configuration" message to
 * the classifier core component. If the input port has interface specific 
 * configuration, the core component changes the set of classification rules 
 * for the port. Otherwise it reports an error.
 *
 * @param  arg_inputPort       interface for which the classification results
 *                             are changed.
 *
 * @param  arg_pClassifResults 64-element table indexed with DSCP value and
 *                             containing the new set of classification rules. 
 *                             Each entry of this table conatains QoS 
 *                             parameters assigned to packets reveived form the
 *                             input port and carrying the DSCP value.
 *
 * @param  arg_Callback        a client provided callback function. The core 
 *                             component invokes this function to report an 
 *                             operation status.
 *
 * @param  arg_pUserContext    pointer to client defined context. This value is
 *                             opaque to the core component. The context is 
 *                             returned in a callback function.
 *
 * @result IX_SUCCESS if successful or a valid ix_error token for failure. The
 *         following error codes might be returned via the token:
 *\n
 *\n       IX_CC_ERROR_OOM_SYSTEM                - not enought system memory 
 *                                                 available to build and asend
 *                                                 a message.
 *\n       IX_CC_ERROR_NULL                      - one of the input arguments 
 *                                                 is NULL.
 *\n       IX_CC_ERROR_SEND_FAIL                 - failure at sending a message
 *
 * @see    ix_cc_classifier_dscp_update_if_config
 */
ix_error ix_cc_classifier_dscp_async_update_if_config(
            ix_uint16                           arg_inputPort,
            ix_cc_classifier_dscp_result      * arg_pClassifResults,
            ix_cc_classifier_dscp_callback      arg_Callback,
            void                              * arg_pUserContext
          );

/**
 * @brief  Sends "get interface configuration" message
 *
 * This function builds and sends "get interface configuration" message to the
 * classifier core component. If the input port has interface-specific 
 * configuration, the core component returns all 64 classificationrules. 
 * Otherwise it reports an error.
 *
 * @param  arg_inputPort       interface for which the calssification rules are
 *                             retrieved.
 *
 * @param  arg_Callback        a client provided callback function. The core 
 *                             component invokes this function to report an 
 *                             operation status and return input port 
 *                             configuration.
 *
 * @param  arg_pUserContext    pointer to client defined context. This value is 
 *                             opaque to the core component. The context is 
 *                             returned in a callback function.
 *
 * @result IX_SUCCESS if successful or a valid ix_error token for failure. The
 *         following error codes might be returned via the token:
 *\n
 *\n       IX_CC_ERROR_OOM_SYSTEM                - not enought system memory 
 *                                                 available to build and asend
 *                                                 a message.
 *\n       IX_CC_ERROR_NULL                      - one of the input arguments 
 *                                                 is NULL.
 *\n       IX_CC_ERROR_SEND_FAIL                 - failure at sending a message
 *
 * @see    ix_cc_classifier_dscp_get_if_config
 */
ix_error ix_cc_classifier_dscp_async_get_if_config(
            ix_uint16                           arg_inputPort,
            ix_cc_classifier_dscp_callback_ex   arg_Callback,
            void                              * arg_pUserContext
          );

/**
 * @brief  Sends "set default rules" message
 *
 * This helper function builds and sends "set default rules" message to the 
 * classifier core component. The core component changes default configuration
 * for the interface that do not have their own interface-specific rules.
 *
 * @param  arg_pClassifResults 64-element table indexed with DSCP value and 
 *                             conataining the new set of classification rules.
 *                             Each entry of this table contains QoS parameters
 *                             assigned to packets carrying the DSCP value.
 *
 * @param  arg_Callback        a client provided callback function. The core 
 *                             component invokes this  function to report an
 *                             operation status.
 *
 * @param  arg_pUserContext    pointer to client defined context. This vlaue is
 *                             opaque to the core component. The context is 
 *                             returned in a callback function.
 *
 * @result IX_SUCCESS if successful or a valid ix_error token for failure. The
 *         following error codes might be returned via the token:
 *\n
 *\n       IX_CC_ERROR_OOM_SYSTEM                - not enought system memory 
 *                                                 available to build and asend
 *                                                 a message.
 *\n       IX_CC_ERROR_NULL                      - one of the input arguments 
 *                                                 is NULL.
 *\n       IX_CC_ERROR_SEND_FAIL                 - failure at sending a message
 *
 * @see    ix_cc_classifier_dscp_set_def_rules
 */
ix_error ix_cc_classifier_dscp_async_set_def_rules(
            ix_cc_classifier_dscp_result      * arg_pClassifResults,
            ix_cc_classifier_dscp_callback      arg_Callback,
            void                              * arg_pUserContext
          );

/**
 * @brief  Sends "get default rules" message
 *
 * This helper function builds and sends a "get default rules" message to the
 * classifier core component. The core component returns all the 64 
 * classification rules applicable to the input port without interface specific
 * configuration. The configuration is returned in a callback function.
 *
 * @param  arg_pCallback       a client provided callback function. The core 
 *                             component invikes this function to report an 
 *                             operation status and return input port 
 *                             configuration.
 *
 * @param  arg_pUserContext    pointer to the client deifned context. This is
 *                             opaque to the core component. The context is 
 *                             returned in a callback function.
 *
 * @result IX_SUCCESS if successful or a valid ix_error token for failure. The
 *         following error codes might be returned via the token:
 *\n
 *\n       IX_CC_ERROR_OOM_SYSTEM                - not enought system memory 
 *                                                 available to build and asend
 *                                                 a message.
 *\n       IX_CC_ERROR_NULL                      - one of the input arguments 
 *                                                 is NULL.
 *\n       IX_CC_ERROR_SEND_FAIL                 - failure at sending a message
 *
 * @see    ix_cc_classifier_dscp_get_def_rules
 */
ix_error ix_cc_classifier_dscp_async_get_def_rules(
            ix_cc_classifier_dscp_callback_ex   arg_Callback,
            void                              * arg_pUserContext
          );

/**
 * @brief ix_cc_classifier_dscp_async_get_statisitics
 *
 * This helper function builds and sends a "get statistics" message to the 
 * classifier core component. If the microblock supports statistics gathering, 
 * the core component retrieves statistics for this  <input port, DSCP value>
 * pair and sends them back to the caller.
 *
 * Note that the function returns statistics regardless of the type of 
 * configuration applied to the port (default or interface-specific). 
 * It returns statistics even if statistics gathering is disabled in the port
 * configuration.
 *
 * If the microblock does not support statistics gahtering, the core component
 * returns an error token.
 *
 * @param  arg_inputPort       interface for which statistics are requested
 *
 * @param  arg_dscp            DSCP value for which statistics are requested
 *
 * @param  arg_Callback        a client provided callback function. The core 
 *                             component invokes this function to report an 
 *                             operation status.
 *
 * @param  arg_pUserContext    pointer to client defined context. This value is
 *                             opaque to the core component. The context is 
 *                             returned ina callback function.
 *
 * @result IX_SUCCESS if successful or a valid ix_error token for failure. The
 *         following error codes might be returned via the token:
 *\n
 *\n       IX_CC_ERROR_OOM_SYSTEM                - not enought system memory 
 *                                                 available to build and asend
 *                                                 a message.
 *\n       IX_CC_ERROR_NULL                      - one of the input arguments 
 *                                                 is NULL.
 *\n       IX_CC_ERROR_SEND_FAIL                 - failure at sending a message
 *
 * @see    ix_cc_classifier_dscp_get_statistics
 */
ix_error ix_cc_classifier_dscp_async_get_statistics(
            ix_uint16                           arg_inputPort,
            ix_uint8                            arg_dscp,
            ix_cc_classifier_dscp_callback_ex   arg_Callback,
            void                              * arg_pUserContext 
          );

#if defined( __cplusplus )
  }
#endif /* __cplusplus */

#endif /* _IX_CC_DSCP_CLASSIFIER_MSGHLP_H__INCLUDED_ */
