/*******************************************************************************
  INTEL CONFIDENTIAL

  Copyright 2003 Intel Corporation All Rights Reserved.

  The source code contained or described herein and all documents related to
  the source code ("Material") are owned by Intel Corporation or its 
  suppliers or licensors.

  Title to the Material remains with Intel Corporation or its suppliers and
  licensors. The Material contains trade secrets and proprietary and
  confidential information of Intel or its suppliers and licensors.
  The Material is protected by worldwide copyright and trade secret laws and
  treaty provisions. No part of the Material may be used, copied, reproduced,
  modified, published, uploaded, posted, transmitted, distributed,
  or disclosed in any way without Intel's prior express written permission.

  No license under any patent, copyright, trade secret or other intellectual
  property right is granted to or conferred upon you by disclosure
  or delivery of the Materials, either expressly, by implication, inducement,
  estoppel or otherwise. Any license under such intellectual property rights
  must be express and approved by Intel in writing.

*******************************************************************************/

/**
 * ============================================================================
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2000 Network Processor
 *
 * = FILE
 *     ix_cc_dscp_classifier_init.c
 *
 * = DESCRIPTION
 *     Implementation of the initialization function of the DSCP classifier 
 *     core component.
 *
 * ============================================================================
 */

#define IX_ERROR_FILE_IDENT \
  "$Id: ix_cc_dscp_classifier_init.c,v 1.8.6.1 2004/03/12 09:25:35 wneubaue Exp $"



#include "cc/internal/dscp_classifier_internal.h"


/* ------------------------------------------- ix_cc_classifier_dscp_init -- */

ix_error ix_cc_classifier_dscp_init( ix_cc_handle arg_hCC,
                                     void **      arg_ppContext )
  {
    _ix_cc_classifier_dscp_config  * _cfg;
    _ix_cc_classifier_dscp_context * _cntxt;
    _T( table_entry )              * _p;

    ix_uint32 _t;
    ix_error  _result = IX_SUCCESS;

    #ifdef IX_DSCP_CLASSIFIER_DEBUG
    printf("ix_cc_classifier_dscp_init\n");
    #endif
    /* validity check for arg_hCC */
    if ( arg_hCC == IX_NULL_HANDLE )
      {
        return IX_ERROR_WARNING(
            IX_CC_ERROR_NULL,
            ( "invalid core component handle specified" )
          );        
      }
    
    /* validity check for arg_ppContext */
    IX_ERROR_CHECK_ARG_NULL( 2, arg_ppContext );
    
    /* -- CONTROL BLOCK -- */

    /* memory allocation for the core component's control block */
    _cntxt = ix_ossl_malloc( _SIZE_OF( context ) );
    if ( _cntxt == NULL )
      {
        return IX_ERROR_WARNING(
            IX_CC_ERROR_OOM_SYSTEM,
            ( "not enough free system memory available, cannot allocate " \
              "the control block of the core component" )
          );
      }    

    #ifdef IX_DSCP_CLASSIFIER_DEBUG
    printf("ix_cc_classifier_dscp_init: memset\n");
    #endif
    
    /* resetting to zero all control data */
    ix_ossl_memset( _cntxt, 0, _SIZE_OF( context )) ;
        
    /* -- SYNCHRONIZATION OBJECT -- */
    
    #ifdef IX_DSCP_CLASSIFIER_DEBUG
    printf("ix_cc_classifier_dscp_init: mutex init\n");
    #endif

    _result = ix_ossl_mutex_init( IX_OSSL_MUTEX_UNLOCK, &_cntxt->_mutex );
    if ( _result != IX_SUCCESS )
      {
        _ix_cc_classifier_dscp_free_memory( (void **)&_cntxt );
        return _result;
      }
   
    /* -- UBLOCK CONFIGURATION -- */

    #ifdef IX_DSCP_CLASSIFIER_DEBUG
    printf("ix_cc_classifier_dscp_init: ublock config\n");
    #endif

    /* memory allocation for the ublock configuration structure */
    _cfg = ix_ossl_malloc( sizeof(_ix_cc_classifier_dscp_config) );
    if ( _cfg == NULL )
      {
        _ix_cc_classifier_dscp_free_memory( (void **)&_cntxt );
        
        return IX_ERROR_WARNING(
            IX_CC_ERROR_OOM_SYSTEM,
            ( "not enough free system memory available, cannot allocate " \
              "buffer for the configuration block" )
          );
      }          
    
    /* storing handle to the core component and pointer to the ublock 
       configuration structure in the core component's control block */
    _cntxt->_handle = arg_hCC;
    _cntxt->_config = _cfg;
           
    #ifdef IX_DSCP_CLASSIFIER_DEBUG
    printf("ix_cc_classifier_dscp_init: update config 0x%x\n",_cfg);
    #endif

    /* updating ublock configuration form the system application registry */
    _result = _ix_cc_classifier_dscp_update_config( _cfg );
    if ( _result != IX_SUCCESS )
    {
        #ifdef IX_DSCP_CLASSIFIER_DEBUG
        printf("ix_cc_classifier_dscp_init: failed to read static configuration from the registry\n");
        #endif
        _ix_cc_classifier_dscp_free_memory( (void **)&_cntxt );
                
        return IX_ERROR_WARNING(
            IX_CC_CLASSIFIER_DSCP_ERROR_REGISTRY,
            ( "failed to read static configuration from the registry" )
          );
    }
       
    /* -- PORTS CONFIGURATION BITMAP -- */

    #ifdef IX_DSCP_CLASSIFIER_DEBUG
    printf("ix_cc_classifier_dscp_init: port config\n");
    #endif
    
    /* calculate the size of the configuration bitmap (bits in the bitmap 
       corespond to the ports in the system) */       
    _t = (_cfg->_in_ports / 8) + ((_cfg->_in_ports & 7) != 0);
   
    /* allocating memory for the ports configuration bitmap */    
    _cntxt->_cfg_bitmap = ix_ossl_malloc( _t );
    if ( _cntxt->_cfg_bitmap == NULL )
      {
        _ix_cc_classifier_dscp_free_memory( (void **)&_cntxt );

        return IX_ERROR_WARNING(
            IX_CC_ERROR_OOM_SYSTEM,
            ( "not enough free system memory available" )
          );
      }

    #ifdef IX_DSCP_CLASSIFIER_DEBUG
    printf("ix_cc_classifier_dscp_init: resetting to zero all bits in the bitmap \n");
    #endif

    /* resetting to zero all bits in the bitmap (ports are not configured) */
    ix_ossl_memset( _cntxt->_cfg_bitmap, 0, _t );

    /* -- DEFAULT CONFIGURATION RULES -- */

    /* calculate the size of the 64-element table containing default 
       classification rules */
    _t = TABLE_ENTRY_SIZE( _cntxt );
    

    #ifdef IX_DSCP_CLASSIFIER_DEBUG
    printf("ix_cc_classifier_dscp_init: memory allocation for the default classification rules  \n");
    #endif

    /* memory allocation for the default classification rules */
    _cntxt->_dscp_default_rules = ix_ossl_malloc( _t );
    if ( _cntxt->_dscp_default_rules == NULL )
      {
        _ix_cc_classifier_dscp_free_memory( (void **)&_cntxt );

        return IX_ERROR_WARNING(
            IX_CC_ERROR_OOM_SYSTEM,
            ( "not enough free system memory available" )
          );
      }

    /* copying the default data to the table */
    _p = _cntxt->_dscp_default_rules;

    for ( _t = 0; _t < DSCP_MAX_VALUE; _t++ )
    {
       ix_ossl_memcpy( _p, _cfg, _SIZE_OF( table_entry ) );
       ((ix_uint8 *)_p) += _cfg->_table_entry_size;
    }
                                
    /* -- INTERFACE-SPECIFIC CONFIGURATIO TABLE -- */

    /* calculating the size of the table */
    _t = _cfg->_in_ports * TABLE_ENTRY_SIZE( _cntxt );

    _result = ix_rm_mem_alloc( 
        IX_MEMORY_TYPE_SRAM, 
        _cfg->_sram_channel, 
        _t, &_cntxt->_dscp_table_sram_base );

    if ( _result != IX_SUCCESS )
      {
        _ix_cc_classifier_dscp_free_memory( (void **)&_cntxt );
                
        return IX_ERROR_WARNING(
            IX_CC_ERROR_OOM_SRAM,
            ( "not enough SRAM memory available, cannot create " \
              "interface-specific configuration table" )
          );
      }

#if defined(CLASSIFIER_DSCP_PACKET_COUNTER_FEATURE)
    
    #ifdef IX_DSCP_CLASSIFIER_DEBUG
    printf("ix_cc_classifier_dscp_init: STATISTICS TABLE \n");
    #endif

    /* -- STATISTICS TABLE -- */
    
    _t = _cfg->_in_ports * STATS_ENTRY_SIZE( _cntxt );
    _result = ix_rm_mem_alloc( 
        IX_MEMORY_TYPE_SRAM, 
        _cfg->_sram_channel, 
        _t, &_cntxt->_dscp_stats_sram_base );
                   
    if ( _result != IX_SUCCESS )
      {                  
        _ix_cc_classifier_dscp_free_memory( (void **)&_cntxt );            
          
        return IX_ERROR_WARNING(
            IX_CC_ERROR_OOM_SRAM,
            ( "not enought SRAM memory available, cannot create " \
              " statistics table" )
          );
      }

    ix_ossl_memset( _cntxt->_dscp_stats_sram_base, 0, _t );

#endif 

    _ix_cc_classifier_dscp_update_entries( (void *)_cntxt );

	/* patch symbols, register message and packets handlers */
    _result = _ix_cc_classifier_dscp_patch_symbols( _cntxt );
    if ( _result == IX_SUCCESS )
      {
        _result = _ix_cc_classifier_dscp_register_pkt_handlers( _cntxt );
        if ( _result == IX_SUCCESS )
          {
            _result = _ix_cc_classifier_dscp_register_msg_handler( _cntxt );
            if ( _result == IX_SUCCESS )
              {
                *arg_ppContext = (void *)_cntxt;
                return IX_SUCCESS;
              }
            _ix_cc_classifier_dscp_unregister_pkt_handlers();
          } 
      }

    _ix_cc_classifier_dscp_free_memory( (void **)&_cntxt );
    return _result;
  }
