/*******************************************************************************
  INTEL CONFIDENTIAL

  Copyright 2003 Intel Corporation All Rights Reserved.

  The source code contained or described herein and all documents related to
  the source code ("Material") are owned by Intel Corporation or its 
  suppliers or licensors.

  Title to the Material remains with Intel Corporation or its suppliers and
  licensors. The Material contains trade secrets and proprietary and
  confidential information of Intel or its suppliers and licensors.
  The Material is protected by worldwide copyright and trade secret laws and
  treaty provisions. No part of the Material may be used, copied, reproduced,
  modified, published, uploaded, posted, transmitted, distributed,
  or disclosed in any way without Intel's prior express written permission.

  No license under any patent, copyright, trade secret or other intellectual
  property right is granted to or conferred upon you by disclosure
  or delivery of the Materials, either expressly, by implication, inducement,
  estoppel or otherwise. Any license under such intellectual property rights
  must be express and approved by Intel in writing.

*******************************************************************************/

/**
 * ============================================================================
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2000 Network Processor
 *
 * = FILE
 *     ix_cc_dscp_classifier_pkt_handlers.c
 *
 * = DESCRIPTION
 *      Packet handler of the DSCP Classifier Core Component implementation.
 *
 * ============================================================================
 */

#define IX_ERROR_FILE_IDENT \
    "$Id: ix_cc_dscp_classifier_pkt_handlers.c,v 1.6.6.1 2004/03/12 09:25:35 wneubaue Exp $"


#include "cc/internal/dscp_classifier_internal.h"


/**
 * IP version 4 packet header
 */
typedef struct _ix_s_ipv4_header
  {
    ix_uint8  _version  :  4,       
              _ihl      :  4;
    ix_uint8  _tos;
    ix_uint16 _total_length;
    ix_uint16 _id;
    ix_uint16 _flags    :  3,
              _offset   : 13;
    ix_uint8  _ttl;
    ix_uint8  _protocol;
    ix_uint16 _chsum;
    ix_uint32 _addr_source;
    ix_uint32 _addr_dest;                      
  
  } _ix_ipv4_header;

/**
 * IP version 6 packet header
 */
typedef struct _ix_s_ipv6_header
  {
    ix_uint32 _version       :  4,
              _tos           :  8,  /* traffic class */
              _flow_label    : 20;
    ix_uint16 _payload_length;
    ix_uint8  _next_header;
    ix_uint8  _hop_limit;
    ix_uint32 _addr_source;
    ix_uint32 _addr_dest;
  
  } _ix_ipv6_header;

/* ------------------------------------ ix_cc_classifier_dscp_pkt_handler -- */

ix_error 
ix_cc_classifier_dscp_pkt_handler( ix_buffer_handle   arg_hBuffer,
                                   ix_uint32          arg_nExCode, 
                                   void             * arg_pContext )
  {
    ix_hw_buffer_meta * _meta;
    
    ix_communication_id _comm_id = IX_CC_PKT_DROP;
    void *              _data;
    
    ix_uint32           _dscp;
    ix_uint16           _size;        

    ix_error            _error;
    _T( table_entry ) * _entry;

    /* validity check for arg_pContext */
    IX_ERROR_CHECK_ARG_NULL( 3, arg_pContext );

    /* validity check for arg_hBuffer */
    IX_ERROR_CHECK_ARG_NULL( 1, arg_hBuffer );

    _error = ix_cc_hw_get_packet_data( arg_hBuffer, &_data );
    if ( _error != IX_SUCCESS )
      {
        IX_ERROR_CRT( ix_rm_buffer_free_chain( arg_hBuffer ),
            IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING );

        return IX_ERROR_CHAIN(
            _error, IX_CC_ERROR_INTERNAL,
            IX_ERROR_LEVEL_WARNING,
            ( "cannot get packet data from the buffer" )
          );
      }
    
    _error = ix_rm_buffer_get_meta( arg_hBuffer, (void **)&_meta );
    if ( _error != IX_SUCCESS )
      {
        IX_ERROR_CRT( ix_rm_buffer_free_chain( arg_hBuffer ),
            IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING );

        return IX_ERROR_CHAIN(
            _error, IX_CC_ERROR_INTERNAL,
            IX_ERROR_LEVEL_WARNING,
            ( "cannot get packet data from the buffer" )
          );
      }

    switch ( _meta->m_BufferInfo & 0xff )
      {
      case DSCP_IPv4_HEADER_TYPE:
        {
          _dscp = ((_ix_ipv4_header *)_data)->_tos;
          break;
        }
      case DSCP_IPv6_HEADER_TYPE:
        {
          _dscp = ((_ix_ipv6_header *)_data)->_tos;
          break;
        }
      default:
        {          
          /* unsupported packet type - drop it */
          _error = ix_rm_packet_send( 
              IX_CC_CLASSIFIER_DSCP_INV_IP_OUTPUT, arg_hBuffer, 0 );
          
          IX_ERROR_CRT( ix_rm_buffer_free_chain( arg_hBuffer ),
              IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING );

          return IX_ERROR_CHAIN( _error, IX_CC_ERROR_INTERNAL, 
              IX_ERROR_LEVEL_WARNING, ( "unsupported packet type recieved" ) );
        }
    }                   

    _dscp >>= 2;
    _size   = ((ix_uint16)*((ix_uint8 *)_data) & 0x0f) << 2;
       
    if ( _MUTEX_LOCK( arg_pContext ) == IX_SUCCESS )
      {
        _entry = _TABLE_ENTRY( arg_pContext, _meta->m_InputPort );

        _meta->m_ColorID = _entry->_id_color;
        _meta->m_FlowID  = _entry->_id_flow;
        _meta->m_ClassID = (ix_uint16)_entry->_id_class;

        _MUTEX_UNLOCK( arg_pContext );
      }

#if defined( CLASSIFIER_DSCP_PACKET_COUNTER_FEATURE )    

    if ( _entry->_flag_stats )
      {        
        ix_uint32           _last;
        _T( stats_entry ) * _stats =
            _STATS_ENTRY( arg_pContext, _meta->m_InputPort );

        /* packet counter */

        _last = ix_rm_atomic_sram_test_and_add( 
            1, &((_ix_64bit_counter *)_stats)->_l );

        if ( (++_last) == 0 )
          {
            ix_rm_atomic_sram_add( 1, &((_ix_64bit_counter *)_stats)->_h );
          }

        ((ix_uint8 *)_stats) += sizeof( ix_uint64 );
        
        /* byte counter */

        _last = ix_rm_atomic_sram_test_and_add( 
            _size, &((_ix_64bit_counter *)_stats)->_l );

        if ( (++_last) == 0 )
          {
            ix_rm_atomic_sram_add( 1, &((_ix_64bit_counter *)_stats)->_l );
          }
      }

#endif /* CLASSIFIER_DSCP_PACKET_COUNTER_FEATURE */
   
    if ( _entry->_flag_ipc )
      {
        switch ( _meta->m_BufferInfo & 0xff )
          {
          case DSCP_IPv4_HEADER_TYPE:
            _comm_id = IX_CC_CLASSIFIER_DSCP_IP4_DEFAULT_OUTPUT;
            break;
          case DSCP_IPv6_HEADER_TYPE:
            _comm_id = IX_CC_CLASSIFIER_DSCP_IP6_DEFAULT_OUTPUT;
            break;
          }
      }
    else
    if ( _entry->_flag_tc )
      {
        _comm_id = IX_CC_CLASSIFIER_DSCP_METER_OUTPUT;
      }
    else
    if ( _entry->_flag_remark )
      {
        _dscp = (_entry->_id_flow & 0x3f) << 2;
        switch ( _meta->m_BufferInfo & 0xff )
          {
          case DSCP_IPv4_HEADER_TYPE:
            _comm_id = IX_CC_CLASSIFIER_DSCP_IP4_DEFAULT_OUTPUT;                        
            
            ((_ix_ipv4_header *)_data)->_tos = (ix_uint8)
              (_dscp | (((_ix_ipv4_header *)_data)->_tos & 0x03));

            break;
          case DSCP_IPv6_HEADER_TYPE:
            _comm_id = IX_CC_CLASSIFIER_DSCP_IP6_DEFAULT_OUTPUT;

            ((_ix_ipv6_header *)_data)->_tos = (ix_uint8)
              (_dscp | (((_ix_ipv6_header *)_data)->_tos & 0x03));

            break;
          }
      }

    _error = ix_rm_packet_send( _comm_id, arg_hBuffer, 0 );
    if ( _error != IX_SUCCESS )
      {
        IX_ERROR_CRT( ix_rm_buffer_free( arg_hBuffer ),
            IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING );

        return IX_ERROR_CHAIN( _error, 
            IX_CC_ERROR_SEND_FAIL, IX_ERROR_LEVEL_WARNING, 
            ( "failed to send packet to the next target" )
          );
      }

    return IX_SUCCESS;
  }
