#ifndef DSCP_CLASSIFIER_UC
#define DSCP_CLASSIFIER_UC	1

/*******************************************************************************
                             Intel Proprietary

 Copyright (c) 1998-2003 By Intel Corporation.  All rights reserved.
 No part of this program or publication may be reproduced, transmited,
 transcribed, stored in a retrieval system, or translated into any language
 or computer language in any form or by any means, electronic, mechanical,
 magnetic, optical, chemical, manual, or otherwise, without the prior
 written permission of:
                         Intel Corporation
                         2200 Mission College Blvd.
                         Santa Clara, CA  95052-8119
*******************************************************************************/


// FILE  dscp_classifier.uc
//
//
// Exported APIs:
//
//   dscp_classifier(in_ip_header, IPHDR_RD_START_BYTE)
//
//
// ASSUMPTIONS
//
//   Handles both IPv4 and IPv6 packets.
//


//------------------------------------------------------------------------------
// Included files:

#include "dl_system.h"
#include "dscp_classifier.h"
#include <stdmac.uc>
#include <xbuf.uc>
#include <dl_meta.uc>


//------------------------------------------------------------------------------
//#macro dscp_classifier(in_ip_header, IPHDR_RD_START_BYTE)
//
// Description:
//
//	dscp_classifier sets QoS related variables based on the DSCP value
//  retrieved from the IPv4/IPv6 packet's header and directs the packet
//  to the proper next block for further processing.
//
//  Inputs:
//    in_ip_header   - IP header cache. It can span transfer registers, GPRs
//                     LocalMem area. It can be refered to by xbuf_xxx macros.
//    dl_next_block  - indicates where the current packet is addressed.
//                     BID_CLASSIFIER_DSCP value indicates that the current
//                     packet should be processed by a dscp_classifier
//                     microblock.
//    dl_header_type - used to determine the version of ip (because of different
//                     positions of dscp fields in the headers).
//                     It is read using dl_meta_get_header_type macro.
//    dl_input_port  - logical number of an incoming interface, on which
//                     a packet was received.
//                     It is read using dl_meta_get_input_port macro.
//
//  Outputs:
//    dl_next_block  - identifier of a microblock that should continue with
//                     packet processing, dscp_classifier sets
//                     one of the following:
//                         - CLASSIFIER_DSCP_METER_OUTPUT
//                         - CLASSIFIER_DSCP_MARKER_OUTPUT
//                         - CLASSIFIER_DSCP_IP4_FWD_OUTPUT
//                         - CLASSIFIER_DSCP_IP6_FWD_OUTPUT
//                         - CLASSIFIER_DSCP_INV_IP_OUTPUT
//
//  Constans
//    IPHDR_RD_START_BYTE - byte address relative to the start of in_ip_header
//
#macro dscp_classifier(in_ip_header, IPHDR_RD_START_BYTE)
.begin

	// Check block ID. If doesn't match ours, execute bypass path.
	br!=byte[dl_next_block, 0, BID_CLASSIFIER_DSCP, bypass_path#]

.reg header_type
.reg input_port
	dl_meta_get_input_port(input_port)
	dl_meta_get_header_type(header_type)

	// Extract IP_ver and ToS values from the IP header (2 bytes @ offset 0).
.reg dscp_field
	xbuf_extract(dscp_field, in_ip_header, IPHDR_RD_START_BYTE, 0, 2)

	br=byte[header_type, 0, CLASSIFIER_DSCP_IPV4_HEADER_TYPE, ip_ver4#]
	br=byte[header_type, 0, CLASSIFIER_DSCP_IPV6_HEADER_TYPE, ip_ver6#]

	immed[dl_next_block, CLASSIFIER_DSCP_INV_IP_OUTPUT]
	br[bypass_path#]


ip_ver6#:
        // DSCP field position in an IPv6 header:
        //
        //    IP       DSCP
        //    ver    bits 6..11
        //  +------+------------+----------+
        //  | 0110 |   010100   |  xxxxxx  |
        //  +------+------------+----------+
        //
        //
	alu_shf[dscp_field, 0x3F, AND, dscp_field, >>6]
	br[dscp_extracted#]

ip_ver4#:
        // DSCP field position in an IPv4 header:
        //
        //    IP             DSCP
        //    ver    len   bits 2..7
        //  +------+------+--------+----+
        //  | 0100 | 0101 | 000000 | xx |
        //  +------+------+--------+----+
        //
	alu_shf[dscp_field, 0x3F, AND, dscp_field, >>2]

dscp_extracted#:

	// Calculate the Classification Result Entry address.
.reg table_index
	alu_shf[table_index, dscp_field, OR, input_port, <<6]
.reg table_offset
	alu_shf[table_offset, --, B, table_index, <<3]

	// Get the Classification Result Entry from SRAM.
	// Classification Result Entry layout:
	//     31                                                  0
	//     +---------------------------------------------------+
	// LW0 |                     flow_id                       |
	//     +------------+------+-----+---+----+---+------------+
	// LW1 |  class_id  | res. | IPC | S | TC | R |  color_id  |
	//     +------------+------+-----+---+----+---+------------+

// Classification Result Entry definition:
// +------------------+------------------+------------------------------------+
// | field name       |    location      |             description            |
// +------------------+------------------+------------------------------------+
// | flow_id          | LW0, bits 31..0  | Identifier of a traffic flow to    |
// |                  |                  | which the packet belongs.          |
// |                  |                  | Subsequent blocks can use this     |
// |                  |                  | value to select a policy instance. |
// +------------------+------------------+------------------------------------+
// | class_id         | LW1, bits 31..16 | A relative queue number within     |
// |                  |                  | an output port.                    |
// +------------------+------------------+------------------------------------+
// | reserved         | LW1, bits 15..6  | Not used now.                      |
// +------------------+------------------+------------------------------------+
// | Input Port       | LW1, bit 5       | Input Port Configured flag         |
// | Configured (IPC) |                  | indicating whether the input port  |
// | flag             |                  | corresponding to the entry has     |
// |                  |                  | its own configuration.             |
// +------------------+------------------+------------------------------------+
// | Statistics flag  | LW1, bit 4       | Indicates whether the microblock   |
// |                  |                  | should gather statistics for       |
// |                  |                  | the rule. This parameter is used   |
// |                  |                  | only when the microblock is        |
// |                  |                  | compiled with statistics handling. |
// +------------------+------------------+------------------------------------+
// | TC flag          | LW1, bit 3       | Traffic Conditioning flag          |
// |                  |                  | indicating whether the packets     |
// |                  |                  | must be metered.                   |
// +------------------+------------------+------------------------------------+
// | Remark flag      | LW1, bit 2       | Indicates whether DSCP value       |
// |                  |                  | in the packets must be remarked.   |
// +------------------+------------------+------------------------------------+
// | color_id         | LW1, bits 1..0   | Packet's drop precedence level:    |
// |                  |                  | green, yellow, or red color.       |
// +------------------+------------------+------------------------------------+

xbuf_alloc($classification_result, 2, read)

.reg table_base
	immed32(table_base, CLASSIFIER_DSCP_CONFIG_SRAM_BASE)

.sig config_read_sig
	sram[read, $classification_result[0], table_base, table_offset, 2],\
          ctx_swap[config_read_sig]


	// First stage of 64-bit counters update.
	// Update lower parts of packet and byte counters
	// and get premodified values.
#ifdef CLASSIFIER_DSCP_PACKET_COUNTER_FEATURE

.reg read $test_pkt_count
.reg $test_byte_count
.sig pkt_count_sig
.sig byte_count_sig
.reg packet_size

.set $test_pkt_count
.set $test_byte_count
.set_sig pkt_count_sig
.set_sig byte_count_sig
.set packet_size

	br_bclr[$classification_result[1], CLASSIFIER_DSCP_STATISTICS_FLAG_BIT,\
	          skip_count1#]

	dl_meta_get_packet_size(packet_size)

	immed32(table_base, CLASSIFIER_DSCP_STATS_SRAM_BASE)
	alu_shf[table_offset, --, B, table_index, <<4]
.reg counter_offset
	alu[counter_offset, table_offset, +, CLASSIFIER_DSCP_PACKET_COUNTER_LOW]
	sram[test_and_incr, $test_pkt_count, table_base, counter_offset],\
          sig_done[pkt_count_sig]

	move($test_byte_count, packet_size)
	alu[counter_offset, table_offset, +, CLASSIFIER_DSCP_BYTE_COUNTER_LOW]
	sram[test_and_add, $test_byte_count, table_base, counter_offset],\
	      sig_done[byte_count_sig]

skip_count1#:

#endif // CLASSIFIER_DSCP_PACKET_COUNTER_FEATURE


	// Set QoS related variables.
	dl_meta_set_flow_id($classification_result[0])
.reg class
	alu_shf[class, --, B, $classification_result[1], >>16]
	dl_meta_set_class_id(class)
.reg color
	alu[color, $classification_result[1], AND, 0x3]
	dl_meta_set_color(color)

	// Check if the Input Port Configured (IPC) flag is set.
	br_bclr[$classification_result[1], CLASSIFIER_DSCP_IPC_FLAG_BIT, \
		ipc_not_set#]

	// Set next block ID.
	br_bset[$classification_result[1], CLASSIFIER_DSCP_TC_FLAG_BIT, exit1#]
	br_bset[$classification_result[1], CLASSIFIER_DSCP_REMARK_FLAG_BIT, exit2#]
	br=byte[header_type, 0, CLASSIFIER_DSCP_IPV4_HEADER_TYPE, exit3#]

	immed[dl_next_block, CLASSIFIER_DSCP_IP6_FWD_OUTPUT]
	br[bid_set#]
exit1#:
	immed[dl_next_block, CLASSIFIER_DSCP_METER_OUTPUT]
	br[bid_set#]
exit2#:
	immed[dl_next_block, CLASSIFIER_DSCP_MARKER_OUTPUT]
	br[bid_set#]
exit3#:
	immed[dl_next_block, CLASSIFIER_DSCP_IP4_FWD_OUTPUT]
	br[bid_set#]
//exit4#:
ipc_not_set#:
	br=byte[header_type, 0, CLASSIFIER_DSCP_IPV4_HEADER_TYPE, exit5#]
	immed[dl_next_block, CLASSIFIER_DSCP_IP6_DEFAULT_OUTPUT]
	br[bid_set#]
exit5#:
	immed[dl_next_block, CLASSIFIER_DSCP_IP4_DEFAULT_OUTPUT]

bid_set#:


	// Second stage of 64-bit counters update.
	// If overflow of the lower counter part is detected,
	// issue increment operation for the higher counter part.
#ifdef CLASSIFIER_DSCP_PACKET_COUNTER_FEATURE

	br_bclr[$classification_result[1], CLASSIFIER_DSCP_STATISTICS_FLAG_BIT,\
	          skip_count2#]

	ctx_arb[pkt_count_sig]
	alu[--, $test_pkt_count, + , 1]
	bvc[pkt_count_no_overflow#]
	alu[counter_offset, table_offset, +, CLASSIFIER_DSCP_PACKET_COUNTER_HIGH]
	sram[incr, --, table_base, counter_offset]

pkt_count_no_overflow#:
	ctx_arb[byte_count_sig]
	alu[--, $test_byte_count, + , packet_size]
	bvc[byte_count_no_overflow#]
	alu[counter_offset, table_offset, +, CLASSIFIER_DSCP_BYTE_COUNTER_HIGH]
	sram[incr, --, table_base, counter_offset]

byte_count_no_overflow#:
skip_count2#:

.io_completed $test_pkt_count
.io_completed $test_byte_count
.io_completed pkt_count_sig
.io_completed byte_count_sig

#endif // CLASSIFIER_DSCP_PACKET_COUNTER_FEATURE

xbuf_free($classification_result)

bypass_path#:

.end
#endm


#endif	// DSCP_CLASSIFIER_UC
