#ifndef SCHEDULER_DIFFSERV_UC
#define SCHEDULER_DIFFSERV_UC

/*******************************************************************************
                             Intel Proprietary

 Copyright (c) 1998-2003 By Intel Corporation.  All rights reserved.  
 No part of this program or publication may be reproduced, transmited,
 transcribed, stored in a retrieval system, or translated into any language
 or computer language in any form or by any means, electronic, mechanical,
 magnetic, optical, chemical, manual, or otherwise, without the prior
 written permission of:
                         Intel Corporation
                         2200 Mission College Blvd.
                         Santa Clara, CA  95052-8119
*******************************************************************************/

// FILE  scheduler_diffserv.uc
//

// 	DiffServ scheduler for Egress Sausalito
// 
// This scheduler runs on the egress Sausalito. It is a frame or packet based 
// scheduler. It implements WRR scheduling on the ports and DRR on the 
// different priority two groups of queues within a port. 
// Currently 16 ports and 16 queues per port are supported
//
// 
// The following are the issues with implementing DRR under the IXP2400 environment
// 
//      1)  The packet size is not available for a queue when it is being scheduled.
//          Once the dequeue is issued, the packet size is received N beats later 
//          where each beat is 88 cycles.
//
//      2)  The scheduler schedules a packet every beat (88 cycles). This means that
//          for large packets, the scheduler is running faster than the transmit
//
//      3)  The exact packet size is not available. The packet size is in multiples 
//			of CHUNK_SIZE which is MTU/128. The queue credit increments are also 
//			given in CHUNK_SIZE units.
//
//      4)  During a DRR round, a queue may go empty or come alive. While a queue is
//          empty it should not receive credit. But queues that frequently go empty
//          should not affect the bandwidth allocation.
//
// The above issues are worked around with modifications to DRR as follows
//    
//      1) We use a scheme of negative credits. The criteria for a queue to be 
//		   eligible to send is that it has data, flow control is off on the port and 
//		   the credits for the queue are positive. A packet is transmitted from a 
//		   queue if it meets the above criteria. Once the packet length is received, 
//		   (N beats later), the packet length is decremented from the current credit 
//		   of the queue. When the current credit of the queue goes negative, it can 
//		   no longer transmit. When all the queues on a port go negative, one DRR 
//		   round is over. Each queue gets another round of credit at this point. 
//		   To ensure that all the queues are schedulable with one round of credit, 
//		   we need to keep the minimum quantum for a queue as (N *  MTU)/CHUNK_SIZE.
//
//      2) If the queue between TX and QM gets full due to large packets or because 
//		   the scheduler is running slightly faster, then the QM will not dequeue the
//		   packet and instead return a 0 for the packet size.
//
//      3) The algorithm will round robin among ports first and queues next. i.e. if 
//         queue i of port j is scheduled, the next queue scheduled will be queue k 
//		   of port j + 1. When the scheduler comes back to port j, the next queue 
//		   scheduled in port j will be port i+1. This increases the probability that 
//		   the packet length is back by the time the queue 
//  
//      4) While a queue is empty or flow control is on, it's credit remains 
//		   untouched. If a queue comes alive in the middle of a round, it is allowed 
//		   to participate right away with the available credit. 
//         
//  


//------------------------------------------------------------------------------
// Included files:

// include stdmac.uc in IXPblocks portable library
#include <stdmac.uc>

// include localmem.uc in IXPblocks portable library 
// for read/write local memory
#include <localmem.uc>          

// include file system.h for system constants          
#include "dl_system.h"          

#include "definitions.h"

// header file with constants for algorithm
#include "scheduler_diffserv.h" 

// include scheduler thread algorithm
#include "scheduler_diffserv_ddr.uc" 

// include the queue manager message handling code
#include "scheduler_diffserv_qm.uc" 



//------------------------------------------------------------------------------
// Global registers


; 1 word (NUMBER_OF_PORTS bits are used). If bit is set then port has data.
.reg @port_empty_vector

.reg @pending_requests

.reg @low_priority_mask

.reg @high_priority_mask


#ifdef SCHEDULER_DIFFSERV_DEBUG
	.reg @debug_in_flight_defer
	.reg @debug_pending_defer
	.reg @debug_packets_scheduled
	.reg @debug_packets_confirmed
	.reg @debug_no_packets
#endif

#ifdef SCHEDULER_DELAY
    .reg @scheduler_delay_counter
#endif

/////////////////////////////////////////////////////////////////////////////////////
// 
// _scheduler_get_credit_increment()
//
// Description:
// 	
//		Get the DRR credit quantum for a given queue. This should be read from SRAM  
//		control block. For now we will simply use a fixed number equal to the minimum 
//		credit increment
//
// Outputs: 
//
//		out_credit_increment:	DRR credit quantum for a specific queue
//
// Inputs:  
//
//		in_queue_id:			queue_id
//
// Size: 
//
//		1 instruction
// 
//
/////////////////////////////////////////////////////////////////////////////////////
	
#macro _scheduler_get_credit_increment(out_credit_increment, in_queue_id)

#ifdef USE_IMPORT_VAR
.begin
  .reg sram_block_base, offset, $credit0, queue_offset
  .sig sram_sig


  // get the sram address to read from 

  immed32(sram_block_base, SCHED_WEIGHT_CREDIT_BASE)

  // get the queue offset in SRAM. The first NUM_OF_PORTS LW addressess
  // are for weights on each port

  alu[queue_offset,--,b, NUMBER_OF_PORTS, <<2]

  // get the offset based on queue number

  alu[offset, --,b, in_queue_id, <<2]

  // the final offset of a give queue in SRAM
  alu[offset, offset, +, queue_offset]

  sram[read, $credit0, sram_block_base, offset, 1], ctx_swap[sram_sig]

  // $credit0 contains weight for the port

  alu[out_credit_increment, --,b, $credit0]


.end

#else

// In simulation mode define the credit increment like this

#define_eval    MIN_CREDIT_INCREMENT  ((MTU * 8) / (1 << BITS_FOR_PACKET_LENGTH))

  immed32(credit_increment, MIN_CREDIT_INCREMENT)

#endif

#endm

/////////////////////////////////////////////////////////////////////////////////////
// 
// _scheduler_get_port_weight()
//
// Description:
// 	
//		 Get the WRR weight for a given port. This should be read from a SRAM control 
//		 block. For now use a fixed value
//
// Outputs: 
//
//		weight:					WRR weight in number of packets
//
// Inputs:  
//
//		port:					port_number			
//
// Size: 
//
//		1 instruction
// 
//
/////////////////////////////////////////////////////////////////////////////////////
	
#macro _scheduler_get_port_weight(out_weight, in_port]

#ifdef USE_IMPORT_VAR
.begin
  .reg sram_block_base, offset, $credit0
  .sig sram_sig


  // get the sram address to read from 

  immed32(sram_block_base, SCHED_WEIGHT_CREDIT_BASE)
  alu[offset, --,b, in_port, <<2]
  sram[read, $credit0, sram_block_base, offset, 1], ctx_swap[sram_sig]

  // $credit0 contains weight for the port

  alu[out_weight, --,b, $credit0]

.end

#else

   // In simulation we will set this up as 1 for every port

  immed[out_weight, 1]

#endif

#endm


/////////////////////////////////////////////////////////////////////////////////////
// 
// _scheduler_init_rings()
//
// Description:
// 	
//		 Initialize the scratch ring and NN rings
//
#macro _scheduler_init_rings()
.begin
.reg 	ctx_enable_data

	//
	// Set up the CTX_ENABLES local csr for NN ring 
	// 
	// bit 20 NN_MODE = 0 : next neighbor register are written
	//                                      by previous ME 
	//
	// bits [19:18] NN_RING_EMPTY = 0 : NN_EMPTY asserts when
	//                                                  NN_PUT == NN_GET
	//													(default)
	// bits [15:8] CTX enables for contexts 0 to 7
	//

    immed32(ctx_enable_data, 0xFF00)			
    local_csr_wr[CTX_ENABLES, ctx_enable_data] 	
    local_csr_wr[NN_PUT, 0]						; initialize NN_PUT
    local_csr_wr[NN_GET, 0]						; initialize NN_GET

.end
#endm


/////////////////////////////////////////////////////////////////////////////////////
// 
// scheduler_init()
//
// Description:
// 	
//		 Initialize the DRR scheduler
//
//
// Outputs:				
//
//		port_credit_vector		Current WRR credit vector for port
//		port_init_credit_vector Initial WRR credit vector for port
//		ring					Scratch ring for deq requests
//		port_mask				Mask used to round robin among ports			
//		minus_two				A constant 0xfffffffe for computinf next port mask	
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro scheduler_init(port_credit_vector, port_init_credit_vector, ring, port_mask,\
					 minus_two)

.begin
	
	.reg	i
	.reg    credit
	.reg 	credit_increment
	.reg 	weight
	.reg	new_schedule_vector

	// ((1 << number_of_queues_per_port) - 1)

	immed32(new_schedule_vector, ((1 << NUMBER_OF_QUEUES_PER_PORT) - 1))

	// This is a constant 0xfffffffe which is used to compute the next port mask. 
	// This is stored in a register since it takes two instructions to compute

	immed32(minus_two, 0xfffffffe)

	// Store the initial bit vector for ports with credit in a register

	immed32(port_init_credit_vector, PORT_INITIAL_CREDIT_VECTOR)
	immed32(port_credit_vector, PORT_INITIAL_CREDIT_VECTOR)

	// Id of the scratch ring to talk to the Queue Manager

	alu_shf[ring, --, b, RING_ID, <<2]

	// This portMask is used to control the round robin scheduling among ports
	// Initialize it to all 1's

	alu[port_mask, --, ~B, 0]

	// Initialize the port empty vector to 0. All ports are initially empty

	immed[@port_empty_vector,0x0] 

	immed[@pending_requests, 0x0]

	immed[port_lm_base, PORT_LM_BASE_OFFSET]
	immed[queue_lm_base, QUEUE_LM_BASE_OFFSET]

	move(@low_priority_mask,  (~HIGH_PRIORITY_MASK))

	move(@high_priority_mask, HIGH_PRIORITY_MASK)

	// Set the base address for ports in local memory. This has
	// a 3 cycle latency which should be covered by the branch in
	// the while

	localmem_set_address(PORT_LM_BASE_OFFSET, 0, LM_HANDLE_0)

	nop
	nop

	// For each port set up the configuration

	immed[i, 0]

	.while (i < NUMBER_OF_PORTS)		

		// Set schedule vector and queue mask to all 1's. Every queue has enough 
		// credits

		alu[*l$index0++, --,  B, new_schedule_vector]  ; write schedule vector

		alu[*l$index0++, --,  B, @high_priority_mask]  ; write high queue mask 
		                                               ; - in negative form
		alu[*l$index0++, --,  B, @low_priority_mask]   ; write lo queue mask
		                                               ; - in negative form

		// Set all ports to be empty

		alu[*l$index0++, --, B, 0] 		; QueueEmpty Vector 

		// Set the packets scheduled to be  0

		alu[*l$index0++, --, B, 0] 		; packets scheduled 

		_scheduler_get_port_weight(weight, i)		


		alu[*l$index0++, --, B, weight] ; current port weight 
		                                ; for WRR among ports


		alu[*l$index0++, --, B, weight]	; port weight quantum for WRR
		
		alu[*l$index0++, --, B, 0x00]	; pending requests on port
										; we start with no scheduled 
										; requests

		// increment i

		alu[i, i, +, 1]

	.endw
	
	// initialize the loop variable
	
	immed[i, 0] 

	// For each queue set up the current credit and quantum. Get the quantum
	// from the SRAM control block

	.while (i < (NUMBER_OF_PORTS * NUMBER_OF_QUEUES_PER_PORT))
		
		_scheduler_get_credit_increment(credit_increment, i)

		alu[credit, --, B, credit_increment, <<16]
		alu[credit, credit, OR, credit_increment]


		// Set the credit increment and the current credit
		alu[*l$index0++, --, B, credit] ; credit & credit increment
		alu[*l$index0++, --, B, 0] 	    ; packet count

			
		// increment i

		alu[i, i, +, 1]

	.endw

.end

#endm



/////////////////////////////////////////////////////////////////////////////////////
//
// main() : Code execution starts here
//
/////////////////////////////////////////////////////////////////////////////////////

#define INTER_THREAD_SIGNAL 	8

.sig volatile	prev_thread_signal
.addr 			prev_thread_signal		INTER_THREAD_SIGNAL

br=ctx[0, CONTROL_THREAD#]

// If thread number is 1, go to the scheduler routine. 

br=ctx[ 1, SCHEDULER_THREAD#]

// If thread number is 2, go to the Queue Manager Message Handler routine

br=ctx[ 2, QM_HANDLER_THREAD#]

// All other threads should simply abort

ctx_arb[kill]

/////////////////////////////////////////////////////////////////////////////////////	

// The xfer registers 0..15 will be used as a target for a reflector write by 
// the transmit microengine. They will contain the count of the packets transmitted 
// and will be accessed in absolute indexed mode. Since xfer registers 0..15 belong 
// to thread 0 in relative mode, this thread will initialize them to 0 by reading in 
// 0's into these registers initially. It will signal the scheduler thread once done

/////////////////////////////////////////////////////////////////////////////////////

CONTROL_THREAD#:

.begin

	.reg visible $$txd_p0, $$txd_p1, $$txd_p2, $$txd_p3, $$txd_p4, $$txd_p5, $$txd_p6, \
				 $$txd_p7, $$txd_p8, $$txd_p9, $$txd_p10, $$txd_p11, $$txd_p12, \
				 $$txd_p13, $$txd_p14,$$txd_p15
	
	.xfer_order	 $$txd_p0, $$txd_p1, $$txd_p2, $$txd_p3, $$txd_p4, $$txd_p5, $$txd_p6, \
				 $$txd_p7, $$txd_p8, $$txd_p9, $$txd_p10, $$txd_p11, $$txd_p12, \
				 $$txd_p13, $$txd_p14, $$txd_p15

	.sig 		 read_signal1, read_signal2
	.reg 		 offset

	// Set up the registers for next neighbor and scratch rings

	_scheduler_init_rings()

	alu[offset, --, B, &$$txd_p0]

	// set up offset to point to a location in SRAM where 8 0's are stored

	immed32[offset, SRAM_ZERO_BLOCK]

	// read in the 8 zero's

	sram[read, $$txd_p0, offset, 0,  8], sig_done[read_signal1]
	sram[read, $$txd_p8, offset, 32, 8], sig_done[read_signal2]

	// wait for the IO to complete. It is safe to let other threads run now

	ctx_arb[read_signal1, read_signal2]

	// signal the scheduler thread 

	#define_eval INTER_THREAD_SIGNAL_REG_VALUE (0x80 | (8 << 3))

	local_csr_wr[SAME_ME_SIGNAL, INTER_THREAD_SIGNAL_REG_VALUE]


	#ifdef SCHEDULER_DIFFSERV_DEBUG
		alu[@debug_in_flight_defer, --, B, 0]
		alu[@debug_pending_defer, --, B, 0]
		alu[@debug_packets_confirmed, --, B, 0]
		alu[@debug_packets_scheduled, --, B, 0]
		alu[@debug_no_packets, --, B, 0]
	#endif

        #ifdef SCHEDULER_DELAY
                alu[@scheduler_delay_counter, --, B, 0]
        #endif

	ctx_arb[kill]

.end

/////////////////////////////////////////////////////////////////////////////////////

SCHEDULER_THREAD#:

.begin
	
	// Scheduler thread
	.reg	port_credit_vector
	.reg	port_init_credit_vector
	.reg    queue_lm_base
	.reg	port_lm_base
	.reg 	current_port_mask
	.reg 	minus_two
	.reg    ring
	.reg	$deq1, $deq2, $deq3, $deq4
	.sig	deq_signal1, deq_signal2, deq_signal3, deq_signal4

	// Wait for a signal from thread 0

	ctx_arb[prev_thread_signal]


	// Execute in thread 1 in an infinite loop

	scheduler_init[port_credit_vector,
	               port_init_credit_vector,
				   ring, current_port_mask,
				   minus_two]


SCHEDULE#:
	
	schedule[port_credit_vector, port_init_credit_vector, ring, current_port_mask, \
			 queue_lm_base, port_lm_base, $deq1, minus_two, deq_signal1, voluntary]

	schedule[port_credit_vector, port_init_credit_vector, ring, current_port_mask, \
			 queue_lm_base, port_lm_base, $deq2, minus_two, deq_signal2, voluntary]

	schedule[port_credit_vector, port_init_credit_vector, ring, current_port_mask, \
			 queue_lm_base, port_lm_base, $deq3, minus_two, deq_signal3, voluntary]

	schedule[port_credit_vector, port_init_credit_vector, ring, current_port_mask, \
			 queue_lm_base, port_lm_base, $deq4, minus_two, deq_signal4, deq_signal1]


SCHEDULE_LOOP#:

	schedule[port_credit_vector, port_init_credit_vector, ring, current_port_mask, \
			 queue_lm_base, port_lm_base, $deq1, minus_two, deq_signal1, deq_signal2]

	schedule[port_credit_vector, port_init_credit_vector, ring, current_port_mask, \
	 		 queue_lm_base, port_lm_base, $deq2, minus_two, deq_signal2, deq_signal3]

	schedule[port_credit_vector, port_init_credit_vector, ring, current_port_mask, \
			 queue_lm_base, port_lm_base, $deq3, minus_two, deq_signal3, deq_signal4]

	schedule[port_credit_vector, port_init_credit_vector, ring, current_port_mask, \
			 queue_lm_base, port_lm_base, $deq4, minus_two, deq_signal4, deq_signal1]

	br [SCHEDULE_LOOP#]

.end

/////////////////////////////////////////////////////////////////////////////////////

QM_HANDLER_THREAD#:

	// Execute in an infinite loop handling QM messages

	_scheduler_qm_message_handler()

/////////////////////////////////////////////////////////////////////////////////////



#endif   // SCHEDULER_DIFFSERV_UC