#ifndef SCHEDULER_DIFFSERV_QM_UC
#define SCHEDULER_DIFFSERV_QM_UC

/*******************************************************************************
                             Intel Proprietary

 Copyright (c) 1998-2003 By Intel Corporation.  All rights reserved.  
 No part of this program or publication may be reproduced, transmited,
 transcribed, stored in a retrieval system, or translated into any language
 or computer language in any form or by any means, electronic, mechanical,
 magnetic, optical, chemical, manual, or otherwise, without the prior
 written permission of:
                         Intel Corporation
                         2200 Mission College Blvd.
                         Santa Clara, CA  95052-8119
*******************************************************************************/

// FILE  scheduler_diffserv_qm.uc
//
//
// Macros for Egress diffserv QM message handling 
//


/////////////////////////////////////////////////////////////////////////////////////
//
//
// The QM sends messages either on a scratch ring or a NN ring. Each message is 2 
// words. First word is for enqueue, the next for dequeue 
//
// Each word is as follows
//
//						--	Bit 31 : Valid bit. If 0, then dont use this word
//						--	Bit 30 : Enqueue/Dequeue Transition bit
//						--	Bit 29 : Invalid dequeue 
//						--	Bits 27..20 : packet length in CHUNK_SIZE units
//						--  Bits 19..0  : Queue Id
//
// For each beat, there may be an enqueue transition AND a dequeue transition in the
// worst case. 
//
//
/////////////////////////////////////////////////////////////////////////////////////

// Bit number for detecting enqueue transition messages (empty queue getting data)

#define QM_TRANSITION_BIT           30

// Bit number for detecting valid messages - if set message is valid

#define QM_VALID_MESSAGE_BIT	    31

// Bit number for invalid dequeues

#define QM_INVALID_DEQUEUE_BIT		29


/////////////////////////////////////////////////////////////////////////////////////
//
// _scheduler_qm_get_packet_length()
//
// Description:
//		Get the packet length from the Queue Manager Message, bits 16-23
//
// Outputs: 
//		out_packet_length:		return length of the packet
//
// Inputs:  
//		in_message:				QM Message
//
//
// Size: 
//		1 instruction
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _scheduler_qm_get_packet_length(out_packet_length, in_message)

	ld_field_w_clr[out_packet_length, 0100, in_message]

#endm

/////////////////////////////////////////////////////////////////////////////////////
//
// _scheduler_qm_get_port_number()
//
// Description:
//		Get the port number from the Queue Manager Message, bits 4..7 
//
// Outputs: 
//		out_port_number:		port number from 0..15
//
// Inputs:  
//		in_message:				QM Message
//
// Size: 
//		1 instruction
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _scheduler_qm_get_port_number(out_port_number, in_message)

	alu_shf[out_port_number, ((1 << NUMBER_OF_BITS_FOR_PORT) - 1), AND, in_message, >>NUMBER_OF_BITS_FOR_QUEUE]

#endm

/////////////////////////////////////////////////////////////////////////////////////
//
// _scheduler_qm_get_queue_number()
//
// Description:
//		Get the queue number from the Queue Manager Message, bits 0..3 
//
// Outputs: 
//		out_queue_number:		queue number from 0..15 
//
// Inputs:  
//		in_message:				QM Message
//
// Size: 
//		1 instruction
//
/////////////////////////////////////////////////////////////////////////////////////


#macro _scheduler_qm_get_queue_number(out_queue_number, in_message)

	alu[out_queue_number, ((1 << NUMBER_OF_BITS_FOR_QUEUE) - 1), AND, in_message]

#endm

/////////////////////////////////////////////////////////////////////////////////////
//
// _scheduler_qm_get_port_lm_offset()
//
// Description:
//		Get the local memory offset for the port in the QM message
//
// Outputs: 
//		out_port_lm_offset:		local memory offset for port in QM message
//
// Inputs:  
//		in_message:				QM Message
//		in_port_lm_mask			Mask used to extract offset from QM message 
//		in_port_lm_base			Base in local memory for port data structure
//
// Size: 
//		1 instruction
//
/////////////////////////////////////////////////////////////////////////////////////


#macro _scheduler_qm_get_port_lm_offset(out_port_lm_offset, in_message, \
										in_port_lm_base, in_port_lm_mask) 

 	// The port number is bits 4..7. Each entry is 32 bytes (left shift of 5). 
	// Combining the two gives us a left shift of 1 and a mask of 0x1e0 for 
	// zeroing out other bits
	alu_shf[out_port_lm_offset, in_port_lm_mask, AND, in_message, <<1]

	// Add the local memory base
	alu[out_port_lm_offset, out_port_lm_offset, +, in_port_lm_base]

#endm

/////////////////////////////////////////////////////////////////////////////////////
//
// _scheduler_qm_get_queue_lm_offset[]
//
// Description:
//		Get the local memory offset for the queue in the QM message
//
// Outputs: 
//		out_queue_lm_offset:		local memory offset for queue in QM message
//
// Inputs:  
//		in_message:					QM Message
//		in_queue_lm_mask			Mask used to extract offset from QM message 
//		in_queue_lm_base			Base in local memory for queue data structure
//
// Size: 
//		2 instruction
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _scheduler_qm_get_queue_lm_offset(out_queue_lm_offset, in_message, \
										in_queue_lm_base, in_queue_lm_offset_mask)

	// compute the queue local memory offset. We use the queue id (0..255) and
	// multiply it by size of each entry (8 bytes) to compute the local memory 
	// offset. queue id is in first 8 bits. We left shift by QUEUE_LM_ENTRY_LOG_SIZE
	// and that puts the lm offset in the first 11 bits
	alu_shf[out_queue_lm_offset, in_queue_lm_offset_mask, AND, in_message, <<QUEUE_LM_ENTRY_LOG_SIZE]

	// Add the local memory base

	alu[out_queue_lm_offset, out_queue_lm_offset, +, in_queue_lm_base] 	

#endm


/////////////////////////////////////////////////////////////////////////////////////
//
// _scheduler_qm_handle_enq_transition()
//
// Description:
//		Handle an enq transition message from the QM
//
// Inputs:  
//		in_message:				QM Message
//		in_port_lm_mask			Mask used to extract port lm offset from QM message 
//
// Size: 
//		10 instruction
//
/////////////////////////////////////////////////////////////////////////////////////
#macro _scheduler_qm_handle_enq_transition(in_message,      \
                                           in_port_lm_mask, \
										   queue_lm_base,   \
										   queue_lm_mask)
.begin

	.reg 	port_lm_offset		// offset in local memory for port data structure
	.reg	queue_lm_offset 	// offset into local mem for port data structure
	.reg	queue_number		// queue number in enq transition message
	.reg	queue_bitmask		// temporary variable storing 1 << queue_number
	.reg	port_number			// port number in enq transition message
	.reg	port_bitmask		// temporary variable storing 1 << port_number


	// compute queue group/port structure local memory offset

	_scheduler_qm_get_port_lm_offset(port_lm_offset, in_message, port_lm_base, \
									in_port_lm_mask) 

	// Set up port offset index into local memory. 3 cycle latency for this to take
	// effect. 

	localmem_set_address(0, port_lm_offset, LM_HANDLE_1)


	// compute queue structure local memory offset. 2 instructions
	_scheduler_qm_get_queue_lm_offset(queue_lm_offset, in_message, queue_lm_base, \
									  queue_lm_mask)

	// Set up queue offset index into local memory. 3 cycle latency for this to
	// take effect. 
	localmem_set_address(0, queue_lm_offset, LM_HANDLE_0)


	// compute the queue number within the port/queue group

	_scheduler_qm_get_queue_number(queue_number, in_message)

	// compute 1 << queueNumber

	alu[--, queue_number, OR, 0]
	alu_shf[queue_bitmask, -- , B, 1, <<indirect]

	// compute the queue group number from 0..31

	_scheduler_qm_get_port_number(port_number, in_message)

	// compute 1 << port number. Cant use global in al_shf instruction

	alu[--, port_number, OR, 0]
	alu_shf[port_bitmask, -- , B, 1, <<indirect]

	// Set the bit for this queue in the queue empty vector.for the group 

	alu[*l$index1[PORT_QUEUE_EMPTY_VECTOR_INDEX], \
		*l$index1[PORT_QUEUE_EMPTY_VECTOR_INDEX] , OR , queue_bitmask]

	// Also if the bit for this group was 0 in the parent vector, we need to set it.
	// Rather than check for the bit being 0, we will simply set the bit	
	
	alu[ @port_empty_vector, @port_empty_vector, OR, port_bitmask]

	// LM[queue_no].packetCount++
	alu[*l$index0[QUEUE_PACKET_COUNT], *l$index0[QUEUE_PACKET_COUNT], +, 1]


.end		
#endm


/////////////////////////////////////////////////////////////////////////////////////
//
// _scheduler_qm_adjust_credit[]
//
// Description:
//		Adjusts the DRR credit for a queue. If queue is negative it marks it as non 
//		schedulable and gives it another round of credit. If all queues in the group
//      are not schedulable, the DRR round begins
//
// Outputs: 
//
//								None
// Inputs:  
//		in_queue_number
//
// Constants
//		EXIT_LABEL 					Label to branch to 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _scheduler_qm_adjust_credit(in_queue_number, EXIT_LABEL)
.begin 

	.reg	credit_increment
	.reg	schedule_vector			// schedule vector for port
	.reg	packet_length
	.reg	queue_bit_mask
	.reg    active_queues


	// Compute (1 << queueNumber)
	alu[--, in_queue_number, OR, 0]
	alu_shf[queue_bit_mask, --, B, 1, <<indirect]

	// mark queue as non schedulable
	alu[schedule_vector, \
	    *l$index1[PORT_SCHEDULE_VECTOR_INDEX], AND~, queue_bit_mask]


	// check if any queue is schedulable
	alu[active_queues, \
	    schedule_vector, AND, *l$index1[PORT_QUEUE_EMPTY_VECTOR_INDEX]]


	alu[--, queue_bit_mask, AND, @high_priority_mask]

	// Now check if the entire schedule vector is 0, which means the round is over
	beq [low_priority_queue#] , defer[3]
	
	// defer 1
	alu[*l$index1[PORT_SCHEDULE_VECTOR_INDEX], --, B, schedule_vector]


	// defer 2
	ld_field_w_clr[credit_increment, 1100, *l$index0[QUEUE_CREDIT_INCREMENT_INDEX], \
                                                  <<QUEUE_CURRENT_CREDIT_OFFSET]

	// defer 3
	alu[*l$index0[QUEUE_CURRENT_CREDIT_INDEX], credit_increment, +, \
							*l$index0[QUEUE_CURRENT_CREDIT_INDEX]]


high_priority_queue#:

	alu[--, active_queues, AND, @high_priority_mask]
	bne[EXIT_LABEL]	
	br[EXIT_LABEL], defer[1]
	// Round is over. Write HIGH_PRIORITY_MASK to the port schedule vector
	alu_shf[*l$index1[PORT_SCHEDULE_VECTOR_INDEX], \
            *l$index1[PORT_SCHEDULE_VECTOR_INDEX], OR,  @high_priority_mask]

low_priority_queue#:

	alu[--, active_queues, AND, @low_priority_mask]
	bne[EXIT_LABEL]
	br[EXIT_LABEL], defer[1]
	// Round is over. Write ~HIGH_PRIORITY_MASK to the port schedule vector
	alu_shf[*l$index1[PORT_SCHEDULE_VECTOR_INDEX], \
			*l$index1[PORT_SCHEDULE_VECTOR_INDEX], OR, @low_priority_mask]


.end
#endm  
		
/////////////////////////////////////////////////////////////////////////////////////
//
// _scheduler_qm_message_handler()
//
// Description:
// 	
//		Thread that handles QM messages
//
// Size: 
//
//		42 instruction 
//
/////////////////////////////////////////////////////////////////////////////////////


#macro _scheduler_qm_message_handler()

.begin

	.reg  	message					// QM  message read from ring
	.reg	port_lm_mask			// Mask to get lm offset for port struct
	.reg	queue_lm_mask			// Mask to get local mem offset for queue struct
	.reg	queue_lm_base			// Base in local mem for queue data structures
	.reg	port_lm_base			// Base in local mem for port data structures
	.reg	packet_length			// Packet Length
	.reg	queue_bitmask			// (1 << queueNumber)
	.reg	port_lm_offset			// offset into local mem for port data structure
	.reg	queue_lm_offset 		// offset into local mem for port data structure
	.reg	queue_number			// queue number 

	// Compute port local memory mask. Set this up in a register so we dont waste 
	// cycles computing it 

	immed32[port_lm_mask, PORT_LM_BASE_MASK]
	immed32[port_lm_base, PORT_LM_BASE_OFFSET]

	// Queue Local Memory Mask and base. Constants stored in registers to save 
	// instructions

	immed32[queue_lm_mask, QUEUE_LM_BASE_MASK]
	immed32[queue_lm_base, QUEUE_LM_BASE_OFFSET] 

QM_SWAP_OUT#:

	// Swap out. The first time the thread runs it will swap out immediately

	ctx_arb[voluntary]

	// First Read the QM message from a scratch or NN ring. In this case we 
	// use a NN ring

CHECK_NN_EMPTY#:

	// Check if the ring is empty. If it is not, then branch to swapping out

	br_inp_state[NN_EMPTY, QM_SWAP_OUT#]

	// Read the enqueue message

	alu[message, --, B, *n$index++] 

	// Check if it is an enqueue transition.  

	br_bclr[message, QM_TRANSITION_BIT, check_dequeue#]

	#ifdef SCHEDULER_DIFFSERV_DEBUG
		alu[@debug_packets_scheduled, @debug_packets_scheduled, +, 1]
	#endif

	;---------------------------------------------------------------------------
	; Now handle the enqueue transition 
	;---------------------------------------------------------------------------

	_scheduler_qm_handle_enq_transition(message, 
	                                    port_lm_mask,
										queue_lm_base,
									    queue_lm_mask)


check_dequeue#:

	// Read the dequeue message 
	alu[message, --, B, *n$index++] 

	// Check if it is an enqueue transition.  

	br_bclr[message, QM_TRANSITION_BIT, check_nn_empty#]

	 #ifdef SCHEDULER_DIFFSERV_DEBUG
		alu[@debug_packets_confirmed, @debug_packets_confirmed, +, 1]
	#endif

	;---------------------------------------------------------------------------
	; Now handle the dequeue transition 
	;---------------------------------------------------------------------------

	// compute queue structure local memory offset. 2 instructions
	_scheduler_qm_get_queue_lm_offset(queue_lm_offset, message, queue_lm_base, \
									  queue_lm_mask)

	// Set up queue offset index into local memory. 3 cycle latency for this to
	// take effect. 
	localmem_set_address(0, queue_lm_offset, LM_HANDLE_0)


	// compute port local memory offset
	_scheduler_qm_get_port_lm_offset(port_lm_offset, message, port_lm_base, \
										port_lm_mask)


	// one outstanding deque message consumed
	alu[@pending_requests, @pending_requests, -, 1]

	// get the packet length from the message 
	_scheduler_qm_get_packet_length(packet_length, message)

	alu[*l$index0[QUEUE_CURRENT_CREDIT_INDEX], \
					*l$index0[QUEUE_CURRENT_CREDIT_INDEX], -, packet_length]

	ble[adjust#]

	ctx_arb[voluntary], br[check_nn_empty#]


	;---------------------------------------------------------------------------
	; Adjust the DRR credit
	;---------------------------------------------------------------------------

	// if credit is not positive, we need to mark the queue as non schedulable
	// and give it another round of credit to the respective queue group

adjust#:
	// Set up port offset index into local memory. 3 cycle latency for this to 
	// take effect. 
	localmem_set_address(0, port_lm_offset, LM_HANDLE_1)

	// Compute queue number
	_scheduler_qm_get_queue_number(queue_number, message)

	_scheduler_qm_adjust_credit(queue_number, QM_SWAP_OUT#)


.end

#endm
	

#endif 		//  SCHEDULER_DIFFSERV_QM_UC
