#ifndef TC_METER_UC
#define TC_METER_UC

/*******************************************************************************
                             Intel Proprietary

 Copyright (c) 1998-2003 By Intel Corporation.  All rights reserved.
 No part of this program or publication may be reproduced, transmited,
 transcribed, stored in a retrieval system, or translated into any language
 or computer language in any form or by any means, electronic, mechanical,
 magnetic, optical, chemical, manual, or otherwise, without the prior
 written permission of:
                         Intel Corporation
                         2200 Mission College Blvd.
                         Santa Clara, CA  95052-8119
*******************************************************************************/

//
// "$Id: tc_meter.uc,v 1.19.4.1 2004/03/12 10:23:36 wneubaue Exp $"

// FILE  tc_meter.uc
//   This block implements a SRTC / TRTC meter in conformity with RFC 2697/2698.
//

// Exported APIs:
//   tc_meter()

// Descripion:
//   	This is the microblock for measuring temporary characteristics of
//   packet flows and verifing them against an agreed profile. It does so
//   by splitting flows into conformance levels (colors) depending on packets
//   being in- or out-of-profile. The algorithm of this block must be
//   configured to work as single or two rate three color meter.
//   At the compilation time you can also choose one of the two modes
//   of operation: color-blind or color-aware. Additional compile time
//   option allows to add the code for collecting statistics of measured flows.
//
//   The complete Three Color Meter consists of following files:
//	    tc_meter.h        - defines options of a compilation and other
//                          parameters needed for a tc_meter to be compiled.
//	    tc_meter.uc       - this file.
//	    tc_meter_init.uc  - macros which initialize or finish a tc_meter block
//      tc_meter_utils.uc - part of the meter algorithm where stc_meter/TRTCM
//                          algorithms are implemented. They perform actual
//                          measuring

//   WARNING - In order to obtain an efficient program this code must be
//   compiled with the workbench's code optimization option enabled.
//   This will ensure that all possible instructions of external macros
//   such as dl_get_flow_id(flow_id) or dl_meta_get_packet_size(packet_size)
//   will be put in free defer slots.


//------------------------------------------------------------------------------
// Included files:
#include <tc_meter.h>
#include <tc_meter_init.uc>
#include <tc_meter_utils.uc>
#include <tc_meter_defs.h>



//------------------------------------------------------------------------------
// Global signals

// Unique signal reserved only for tc_meter block on all ME using it.
// It is used to ensure that only one meter of all working will
// have exclusive access to SRAM.
.sig visible volatile tc_meter_next_me_signal
#ifdef TC_METER_NEXT_ME_SIG
.addr tc_meter_next_me_signal     TC_METER_NEXT_ME_SIG
#endif


.sig volatile global tc_meter_next_thread_signal
#ifdef  TC_METER_NEXT_THREAD_SIG
.addr tc_meter_next_thread_signal TC_METER_NEXT_THREAD_SIG
#else
#define	TC_METER_NEXT_THREAD_SIG &tc_meter_next_thread_signal
#endif


//------------------------------------------------------------------------------
// Global registers

// _tc_meter_pir_extract_mask register contains bit mask to extract pir field
// from the meter entry. It is initialized in init_tc_meter macro.
//.reg _tc_meter_pir_extract_mask


//------------------------------------------------------------------------------
// #macro tc_meter()

// tc_meter
//
// Description:
//   Main macro performing the thread synchronization and
//   data preparation for the actual SRTCM/TRTCM algorithm. For the sake of
//   readability the part of the code pertinent to the actual flow metering has
//   been moved to 'tc_meter_utils.uc' file.
//
// Inputs:
//    dl_flow_id     - the identification of a packet flow which allows to find
//                     required entry in the tc_meter table.
//    dl_packet_size - the size of the current packet - used to consume tokens
//                     from buckets and assign a new packet color
//    dl_color_id    - the initial color of a received packet. This is used only
//                     by color aware version of the algorithm.
//                     NOTE: this variable should be one of the values:
//                     GREEN_PACKET, YELLOW_PACKET, RED_PACKET
//    dl_next_block  - the identification of the next block which the packet is
//                     to be handled by.
//                     BID_TC_METERETER value indicates that the current packet
//                     should be processed by a tc_meter micro block.
//
// Outputs:
//   dl_color_id     - the color which has been designated for the packet.
//                     It is set by _tc_meter_meetering() macro.
//   dl_flow_id      - new dscp field that corresponds to the designated color
//                     for a given flow.
//   dl_next_block   - the identification of the next block which the packet is
//                     to be directed to. If the current packet has been handled
//                     by a tc_meter it is set to TC_METER_NEXT1 or 
//                     TC_METER_NEXT2 block depending on the flow specific 
//                     option set in the tc_meter table.
//
//
//------------------------------------------------------------------------------
//
// Three Color Meter block uses a table stored in SRAM which is implemented as
// a simple array indexed with a flow_id. A table entry
// contains meter parameters which are illustrated in the figure below.
//
// METER TABLE in SRAM (common for srt and trc modes of operation)
//
//
// +-----> memory address growth
// |
// |             lword1         lword2
// |      +----+-----+------+-----------------+
// |    0 |CIR | PIR |crate |       CBS       |
// v      +----+-----+------+---+---+---+-----+
//      1 |    EBS / PBS    | G | Y | R |prate|
//        +-----------------+---+---+---+-----+
//      2 |LAST TSTAMP low  |LAST TSTAMP high |   <--- 64 bit timestamp
//        +-----------------+-----------------+        of the last packet
//      3 |        TC       |     TE / TP     |        in this flow
//        +-----------------+-----------------+
//      4 |packet green cnt | byte green cnt  |
//        +-----------------+-----------------+
//      5 |packet yellow cnt| byte yellow cnt |
//        +-----------------+-----------------+
//      6 |packet red cnt   | byte red cnt    |
//        +-----------------+-----------------+
//      8 |              reserved             |
//        +-----------------------------------+
//
//
//  crate (cir rate scale) 
//
//       bit 7  6  5                 0
//           +------+----------------+
//           |unused|   rate scale   |
//           +------+----------------+
//
//  prate (pir rate scale) fields:
//
//       bit 7  6  5                 0
//           +------+----------------+
//           |flags |   rate scale   |
//           +------+----------------+
//            | | |        |
//            | | |        +----> number of bits timestamp will be shifted
//            | | |
//            | | +-> bit 5 - unused
//            | +---> bit 6 - enables statistics gathering
//            |                  0 meter gathers statistics
//            |                  1 meter doesn't gather statistics
//            +-----> bit 7 - indicates the type of the meter
//                               0 meter is single rate (SRTCM)
//                               1 meter is two rate (TRTCM)
//
//
// TRTCM METER TABLE in LOCAL MEMORY
//
//  in order to optimize tc_meter algorithm the meter data
//	arrangement in LM is different:
//
//
//             lword1             lword2
//       +-----------------+-----------------+
//     0 |LAST TSTAMP low  |LAST TSTAMP high |   <--- 64 bit timestamp
//       +-----------------+-----------------+        of the last packet
//     1 |        TC       |     TE / TP     |        in this flow
//       +---+---+---+-----+-----------------+
//     2 | G | Y | R |flags|       CIR       |
//       +---+---+---+-----+-----------------+
//     3 |       PIR       |       CBS       |
//       +-----------------+-----------------+
//     4 |     EBS / PBS   |packet green cnt |
//       +-----------------+-----------------+
//     5 | byte green cnt  |packet yellow cnt|
//       +-----------------+-----------------+
//     6 | byte yellow cnt |packet red cnt   |
//       +-----------------+-----------------+
//     7 | byte red cnt    |     ref count   |   <--- ref count field contains
//       +-----------------+-----------------+        the number of threads
//                                                    using the same tc_meter
//                                                    entry in local memory.
//
//  flags field:
//
//       bit 7  6  5                 0
//           +------+----------------+
//           |flags |   reserved     |
//           +------+----------------+
//            | | |
//            | | +-> bit 5 - unused
//            | +---> bit 6 - enables statistics gathering
//            |                  0 meter gathers statistics
//            |                  1 meter doesn't gather statistics
//            +-----> bit 7 - indicates the type of the meter
//                               0 meter is single rate (SRTCM)
//                               1 meter is two rate (TRTCM)
//
//
//
//  G, Y, R fields:
//
//       bit  7                 2    0
//           +------------------+----+
//           |        DSCP      |OUT |
//           +------------------+----+
//                      |         |
//                      |         +----> bits 0-1 - points where to divert
//                      |                           packets in this color:
//                      |                             00 to TRTCM_NEXT1 block
//                      |                             01 to TRTCM_NEXT1 block
//                      |                             10 to TRTCM_NEXT2 block
//                      |                             11 to TRTCM_NEXT3 block
//                      +--------------> bits 2-7 - DSCP field
//
//
//
//

#macro tc_meter()
.begin


	 ;----------------------------------------------------------
	 ;	signal declarations
	 ;----------------------------------------------------------
	.sig  tc_meter_sram_read_dn
	.sig  tc_meter_sram_write_dn

	// registers for holding the address of the tc_meter data in SRAM
	.reg tc_meter_offset
	.reg tc_meter_base

	// registers indicate the time of packet arrival.
	// they are initialized globally by the thread 0 and used by the others
	// (thereby the same value is taken to process all 8 packets)
	.reg @current_timestamp_l
	.reg @current_timestamp_h
	.reg current_timestamp_l
	.reg current_timestamp_h


	// in the event of compliling this code with TC_METER_COLOR_AWARE_FEATURE
	// this register contains values GREEN_PACKET, YELLOW_PACKET, RED_PACKET
	// which tell tc_meter_color_selection macro what was the input color of
	// the processed packet.
	#ifdef TC_METER_COLOR_AWARE_FEATURE
	.reg color_id
	#endif

     ;----------------------------------------------------------
	 ; dispatch loop variables
     ;----------------------------------------------------------
	.reg flow_id
	.reg packet_size


tc_meter_path#:

;-------------------------------------------------------------------------------
; CAM CLEAR STAGE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	br=ctx[0,cam_clear_thread0#], defer[2]
	// initialize variables responible for holding the address
	// to the tc_meter's data in SRAM
	immed[tc_meter_base, (TC_METER_TABLE_SRAM_BASE & 0xFFFF)]
	immed_w1[tc_meter_base, ((TC_METER_TABLE_SRAM_BASE >> 16) & 0xFFFF)]


	   	dl_meta_get_flow_id(flow_id)
		dl_meta_get_packet_size(packet_size)

		ctx_arb[tc_meter_next_thread_signal], br[sram_read_stage#]

	cam_clear_thread0#:


	   	dl_meta_get_flow_id(flow_id)
		dl_meta_get_packet_size(packet_size)

		// Wait for inter-ME signal and do not allow the other threads
		// to begin next stage
       	ctx_arb[tc_meter_next_me_signal, tc_meter_next_thread_signal]

		//get the current time
		local_csr_rd[TIMESTAMP_LOW]
		immed[@current_timestamp_l,0]
		local_csr_rd[TIMESTAMP_HIGH]
		immed[@current_timestamp_h,0]

		// Clear the CAM & reset LRU just once at the beginning but after
		// synchronization in order to prevent an interference with
		// the previous blocks
		cam_clear




sram_read_stage#:
;-------------------------------------------------------------------------------
; SRAM READ STAGE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	; send TC_METER_NEXT_THREAD_SIGNAL to the next thread
	local_csr_wr[SAME_ME_SIGNAL, (0x80 | ((&tc_meter_next_thread_signal) <<3)) ]

	; compare dl_next_block with BID_TC_METER
	; If dl_next_block doesn't match ours, execute the bypass path.
	br!=byte[dl_next_block, 0, BID_TC_METER, bypass_path#]

	.begin
	xbuf_alloc($meter, 14,  read)

	.begin
	// position number of the cam entry for the current flow
	.reg cam_entry_number flow_id_plus_one

		.begin
		.reg cam_result

			alu[flow_id_plus_one, flow_id, +, 1]

		    cam_lookup[cam_result, flow_id_plus_one], lm_addr0[0]

			// jump if lookup's hit
			br_bset[cam_result, 7, cam_hit#], defer[2]
	    	// cam entry number is used to save new flow_id to CAM
			alu_shf[cam_entry_number, 0xF, AND, cam_result, >>3]
			// initialize variables responible for holding the address
			// to the tc_meter's data in SRAM
			alu_shf[tc_meter_offset, --, B, flow_id, <<TC_METER_ENTRY_LOG_SRAM_SIZE]

		.end

	cam_miss#:

		// ---------------------------------
		#ifdef TC_METER_PACKET_COUNTER_FEATURE

			alu_shf[--, --, B, 0x1d, <<21]; read 14 words from sram = 00011011B
			sram[read, $meter[0], tc_meter_base, tc_meter_offset, max_14],
			     indirect_ref, sig_done[tc_meter_sram_read_dn]

		#else

			sram[read, $meter[0], tc_meter_base, tc_meter_offset, 8],
			     sig_done[tc_meter_sram_read_dn]

		#endif
		// ---------------------------------

        .reg _tc_meter_pir_extract_mask
        immed32(_tc_meter_pir_extract_mask, TC_METER_PIR_EXTRACT_MASK);

		ctx_arb[tc_meter_next_thread_signal, tc_meter_sram_read_dn], and, defer[2]
	    // if lookup misses cam entry number contains LRU,
		// and we put our read data to CAM in this position
		cam_write[cam_entry_number, flow_id_plus_one, TC_METER_CAM_VALID]
		// (LM[lru_entry].ref_count) = 1
		alu[*l$index0[TC_METER_LM_REF_COUNT], --, B, 0x01]

		// for optimization we do not have to copy all read data into local
		// memory. Because we assume that there is one timestamp for
		// every 8 threads and some of the fields of TC_METER table entry are
		// used only once to calculate new values of TC and TE there is no need
		// to copy them to local memory. In other words, it is sufficient
		// to take them directly from transfer registers and forget about them.

		alu[*l$index0[TC_METER_LM_CIR],
            --, B,   $meter[0], >>TC_METER_CIR_EXTRACT_SHIFT]

		alu[*l$index0[TC_METER_LM_PIR],
		    _tc_meter_pir_extract_mask, AND, $meter[0], >>TC_METER_PIR_EXTRACT_SHIFT]

		alu[*l$index0[TC_METER_LM_DSCP],         --, B, $meter[3]]

		// ---------------------------------
		#ifdef TC_METER_PACKET_COUNTER_FEATURE

			alu[*l$index0[TC_METER_LM_GREEN_PACKETS],  --, B, $meter[8]]
			alu[*l$index0[TC_METER_LM_GREEN_BYTES],    --, B, $meter[9]]
			alu[*l$index0[TC_METER_LM_YELLOW_PACKETS], --, B, $meter[10]]
			alu[*l$index0[TC_METER_LM_YELLOW_BYTES],   --, B, $meter[11]]
			alu[*l$index0[TC_METER_LM_RED_PACKETS],    --, B, $meter[12]]
			alu[*l$index0[TC_METER_LM_RED_BYTES],      --, B, $meter[13]]

		#endif
		// -----------


	.end


;-------------------------------------------------------------------------------
; METERING STAGE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;




#ifdef TC_METER_TRTCM_FEATURE
#ifdef TC_METER_SRTCM_FEATURE

	alu[--, *l$index0[TC_METER_LM_METER_TYPE], AND, TC_METER_TRTCM_ENABLE_FLAG]
	bne[use_TRTCM_algorithm#], defer[2]

#endif
#endif

	alu[current_timestamp_l,  --, B, @current_timestamp_l]
	alu[current_timestamp_h,  --, B, @current_timestamp_h]


;-------------------------------------------------------------------------------
#ifdef TC_METER_SRTCM_FEATURE
;-------------------------------------------------------------------------------

use_SRTCM_algorithm#:

	_SRTCM_token_update(
                   *l$index0[TC_METER_LM_TC],         /* out_new_tc           */
                   *l$index0[TC_METER_LM_TE],         /* out_new_te           */
                   *l$index0[TC_METER_LM_TIMESTAMP_L],/* out_last_timestamp_l */
                   *l$index0[TC_METER_LM_TIMESTAMP_H],/* out_last_timestamp_h */
                   current_timestamp_l,               /* in_curr_timestamp_l  */
                   current_timestamp_h,               /* in_curr_timestamp_h  */
                   $meter[4],                         /* in_last_timestamp_l  */
                   $meter[5],                         /* in_last_timestamp_h  */
                   $meter[6],                         /* in_curr_tc           */
                   $meter[7],                         /* in_curr_te           */
                   $meter[1],                         /* in_cbs               */
                   $meter[2],                         /* in_ebs               */
                   *l$index0[TC_METER_LM_CIR],        /* in_cir               */
                   $meter[0])                         /* in_cir_rate_scale    */

#ifdef TC_METER_TRTCM_FEATURE
	br[start_tc_meter#]
#endif
#endif


;-------------------------------------------------------------------------------
#ifdef TC_METER_TRTCM_FEATURE
;-------------------------------------------------------------------------------


use_TRTCM_algorithm#:

	_TRTCM_token_update(
	               *l$index0[TC_METER_LM_TC],         /* out_new_tc           */
                   *l$index0[TC_METER_LM_TP],         /* out_new_tp           */
                   *l$index0[TC_METER_LM_TIMESTAMP_L],/* out_last_timestamp_l */
                   *l$index0[TC_METER_LM_TIMESTAMP_H],/* out_last_timestamp_h */
                    current_timestamp_l,              /* in_curr_timestamp_l  */
                    current_timestamp_h,              /* in_curr_timestamp_h  */
                    $meter[4],                        /* in_last_timestamp_l  */
                    $meter[5],                        /* in_last_timestamp_h  */
                    $meter[6],                        /* in_curr_tc           */
                    $meter[7],                        /* in_curr_tp           */
                    $meter[1],                        /* in_cbs               */
                    $meter[2],                        /* in_pbs               */
                    *l$index0[TC_METER_LM_CIR],       /* in_cir               */
                    *l$index0[TC_METER_LM_PIR],       /* in_pir               */
                    $meter[0],                        /* in_cir_rate_scale    */
                    $meter[3])                        /* in_pir_rate_scale    */

;-------------------------------------------------------------------------------
#endif
;-------------------------------------------------------------------------------


start_tc_meter#:


	//----------------------------
	#ifdef TC_METER_COLOR_AWARE_FEATURE
	dl_meta_get_color(color_id)
	#endif
	//----------------------------

	_tc_metering(*l$index0[TC_METER_LM_TC],
	             *l$index0[TC_METER_LM_TE],
                 *l$index0[TC_METER_LM_DSCP],
                  packet_size,
                  color_id)

	xbuf_free($meter)
	.end

;-------------------------------------------------------------------------------
; SRAM WRITE STAGE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	br=ctx[7, thread_7#], defer[1] // Special case for thread 7.
	local_csr_wr[SAME_ME_SIGNAL, (0x80 | (&tc_meter_next_thread_signal <<3))]

		//	if (--(LM[lru_entry].ref_count)) != 0
		//		goto tc_meter_meter_final_synchronization#
		alu[*l$index0[TC_METER_LM_REF_COUNT], *l$index0[TC_METER_LM_REF_COUNT], -, 1]
		; skip writeback and signal to the next thread
		bne[tc_meter_final_synchronization#]


		xbuf_alloc($meter, 10,  write)

		//writeback meter state to sram
		alu[$meter[0], --, B, *l$index0[TC_METER_LM_TIMESTAMP_L]]
		alu[$meter[1], --, B, *l$index0[TC_METER_LM_TIMESTAMP_H]]

	thread7_writeback#:

		alu[$meter[2], --, B, *l$index0[TC_METER_LM_TC]]
		alu[$meter[3], --, B, *l$index0[TC_METER_LM_TE]]

		alu[tc_meter_offset, tc_meter_offset, +, TC_METER_WRITE_ACCESS_OFFSET]

		// ---------------------------------
		#ifdef TC_METER_PACKET_COUNTER_FEATURE
		// ---------------------------------

		alu[$meter[4], --, B, *l$index0[TC_METER_LM_GREEN_PACKETS]] ;pkt_green_cnt
		alu[$meter[5], --, B, *l$index0[TC_METER_LM_GREEN_BYTES]]   ;byte_green_cnt
		alu[$meter[6], --, B, *l$index0[TC_METER_LM_YELLOW_PACKETS]];pkt_yellow_cnt
		alu[$meter[7], --, B, *l$index0[TC_METER_LM_YELLOW_BYTES]]  ;byte_yellow_cnt
		alu[$meter[8], --, B, *l$index0[TC_METER_LM_RED_PACKETS]]   ;pkt_red_cnt
		alu[$meter[9], --, B, *l$index0[TC_METER_LM_RED_BYTES]]     ;byte_red_cnt

		alu_shf[--, --, B, 0x19, <<21] //write ten words to sram = 00011001B

		sram[write, $meter[0], tc_meter_offset, tc_meter_base, max_10],
		     indirect_ref, sig_done[tc_meter_sram_write_dn]


		// ----------------------------------------------
		#else  // TC_METER_PACKET_COUNTER_FEATURE NOT DEFINED
		// ----------------------------------------------

		//write four words to sram
		sram[write, $meter[0], tc_meter_offset, tc_meter_base, 4],
                                              sig_done[tc_meter_sram_write_dn]

		//---------------
		#endif
		//---------------

		xbuf_free($meter)

		#ifndef TC_METER_DISABLE_FIN_SYNC
	  		ctx_arb[tc_meter_sram_write_dn, tc_meter_next_thread_signal],
			                                             AND, br[tc_meter_end#]
		#else
	  		ctx_arb[tc_meter_sram_write_dn], br[tc_meter_final_synchronization#]
		#endif


	thread_7#:

		// Signal the thread 0 on DL_NEXT_ME with tc_meter_next_me_signal
		// WARNING: Beacause this signal should get to the next me after sram
		//          write in this thread it must be at least 13 cycles from here
		//          to the sram read instruction on next me.
		//			Otherwise we will have problem with inter-ME signaling -
		//			signal will be received by next ME before our sram write
		//          operation.
		// signal_me[DL_NEXT_ME, &tc_meter_next_me_signal]
		.sig remote tc_meter_next_me_signal

		cap[fast_wr,
		    (&remote(tc_meter_next_me_signal, DL_NEXT_ME) | (DL_NEXT_ME <<7)),
			INTERTHREAD_SIG]

		// thread 7 always perform writeback

		br[thread7_writeback#], defer[2]
		alu[$meter[0], --, B, *l$index0[TC_METER_LM_TIMESTAMP_L]]
		alu[$meter[1], --, B, *l$index0[TC_METER_LM_TIMESTAMP_H]]


	cam_hit#:


		ctx_arb[tc_meter_next_thread_signal], br[start_tc_meter#] ,defer[1]
		; increase the counter of threads using the same tc_meter entry
		; in local memory.  -  LM[lru_entry].ref_count++
		alu[*l$index0[TC_METER_LM_REF_COUNT], 1, +, *l$index0[TC_METER_LM_REF_COUNT]]



bypass_path#:
;-------------------------------------------------------------------------------
; SRAM READ STAGE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	ctx_arb[tc_meter_next_thread_signal]
	// signal_next_ctx[&tc_meter_next_thread_signal]
	local_csr_wr[SAME_ME_SIGNAL, (0x80 | (&tc_meter_next_thread_signal <<3))]

;-------------------------------------------------------------------------------
; SRAM WRITE STAGE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	br!=ctx[7, tc_meter_final_synchronization#]

	// Signal the thread 0 on DL_NEXT_ME with tc_meter_next_me_signal
	// the same as - signal_me[DL_NEXT_ME, &tc_meter_next_me_signal]
	cap[fast_wr,((&remote(tc_meter_next_me_signal, DL_NEXT_ME))
	             | (DL_NEXT_ME <<7)), INTERTHREAD_SIG]






tc_meter_final_synchronization#:

#ifndef TC_METER_DISABLE_FIN_SYNC
	ctx_arb[tc_meter_next_thread_signal]

tc_meter_end#:
	// signal_next_ctx[&tc_meter_next_thread_signal]
	local_csr_wr[SAME_ME_SIGNAL, (0x80 | (&tc_meter_next_thread_signal <<3))]

#endif



.end
#endm  // macro meter()




#endif // TC_METER_UC
