#ifndef TC_METER_UTILS_UC
#define TC_METER_UTILS_UC

/*******************************************************************************
                             Intel Proprietary

 Copyright (c) 1998-2003 By Intel Corporation.  All rights reserved.
 No part of this program or publication may be reproduced, transmited,
 transcribed, stored in a retrieval system, or translated into any language
 or computer language in any form or by any means, electronic, mechanical,
 magnetic, optical, chemical, manual, or otherwise, without the prior
 written permission of:
                         Intel Corporation
                         2200 Mission College Blvd.
                         Santa Clara, CA  95052-8119
*******************************************************************************/
/* "$Id: tc_meter_utils.uc,v 1.17.6.1 2004/03/12 10:23:37 wneubaue Exp $" */


// FILE  tc_meter_utils.uc
//   Implements single and two rate meter algorithms
//
// Exported APIs:
//   _srtcm_token_update(out_new_tc, out_new_te,
//                       out_last_timestamp_l, out_last_timestamp_h,
//                       in_curr_timestamp_l,  in_curr_timestamp_h,
//                       in_last_timestamp_l,  in_last_timestamp_h,
//                       in_curr_tc, in_curr_te,
//                       in_cbs,     in_ebs,
//                       in_cir,
//                       in_rate_scale)
//
//   _trtcm_token_update(out_new_tc,           out_new_tp,
//                       out_last_timestamp_l, out_last_timestamp_h,
//                       in_curr_timestamp_l,  in_curr_timestamp_h,
//                       in_last_timestamp_l,  in_last_timestamp_h,
//                       in_curr_tc,           in_curr_tp,
//                       in_cbs,               in_pbs,
//                       in_cir,               in_pir,
//                       in_pir_rate,          in_cir_rate)
//
//   _tc_metering(out_in_curr_tc,   out_in_curr_te_tp,
//                in_dscp_values,   in_packet_size,
//                in_color_id)
//
// Description:
// 		This file implements srtcm algorithm which is used by the
//   'tcm_meter()' macro to calculate 'level of conformance ' (colors) of 
//   transmitted packets. For the sake of readability and performance the 
//   algorithm has been divided into two macros. First one is 
//   _??tcm_token_update() macro which is used to update values of (TE / TC)
//   buckets. Second part of the algorithm is covered by tc_metering 
//   macro which performs actual color-of-the-packet calculation.
//
//   The behavior of this algorithm depends on
//   applied compilation options which allows to build
//
//      - color-aware meter, with statistics
//      - color-blind meter, with statistics
//      - color-aware meter, without statistics
//      - color-blind meter, without statistics
//
//   See "tc_meter.h" for detail.


#include "dl_system.h"


//------------------------------------------------------------------------------
//#macro _srtcm_token_update(out_new_tc, out_new_te,                     \
//                           out_last_timestamp_l, out_last_timestamp_h, \
//                           in_curr_timestamp_l, in_curr_timestamp_h,   \
//                           in_last_timestamp_l, in_last_timestamp_h,   \
//                           in_curr_tc, in_curr_te,                     \
//                           in_cbs, in_ebs,                             \
//                           in_cir,                                     \
//                           in_rate_scale)

// _srtcm_token_update
//   Calculates how many tokens are to be added to TC, TE buckets depending
//   on the formula (in_curr_timestamp - in_last_timestamp) * in_cir

// Inputs:
//
//   in_curr_timestamp_l
//   in_curr_timestamp_h  - current timestamp read by the thread 0 and
//                          used by all the others.
//   in_last_timestamp_l
//   in_last_timestamp_h  - stamp with the time at which the last packet has
//                          been handled in this flow
//   in_curr_tc           - current content of the bucket with green tockens
//   in_curr_te           - current content of the bucket with yellow tockens
//   in_cbs, in_ebs       - the max value of green and yellow buckets
//   in_cir,              - commited information rate for this flow
//   in_rate_scale        - time scaling factor
//
// Outputs:
//
//   out_new_tc           - calculated new value of green tockens in the bucket
//   out_new_te           - new amount of yellow tockens in the bucket
//   out_last_timestamp_l - new last timestamp to be written back to LM
//   out_last_timestamp_h - new last timestamp to be written back to LM
//
//
#macro _srtcm_token_update(out_new_tc,           out_new_te,            \
                           out_last_timestamp_l, out_last_timestamp_h,  \
                           in_curr_timestamp_l,  in_curr_timestamp_h,   \
                           in_last_timestamp_l,  in_last_timestamp_h,   \
                           in_curr_tc, in_curr_te,                      \
                           in_cbs, in_ebs,                              \
                           in_cir,                                      \
                           in_rate_scale)
.begin
    // 5 lower bits contain rate adjust
    alu[--, in_rate_scale, OR, 0]
    dbl_shf[out_last_timestamp_l, in_curr_timestamp_h, in_curr_timestamp_l,
                                                                    >>indirect]
    alu[--, in_rate_scale, OR, 0]
    alu_shf[out_last_timestamp_h,--,B,in_curr_timestamp_h,>>indirect]

    // if the inter_arrival time exceeds 32 bits set TC and TE to maximum values
    .reg inter_arrival_l
    .reg inter_arrival_h

    alu[inter_arrival_l, out_last_timestamp_l, -,      in_last_timestamp_l]
    alu[inter_arrival_h, out_last_timestamp_h, -carry, in_last_timestamp_h]
    bne[set_max_tc_te#]

    // Multiply of cir and inter_arrival:  burst_update = cir * inter_arrival
    .reg burst_update burst_update_h
    mul_step[in_cir, inter_arrival_l], 32x32_start
    mul_step[in_cir, inter_arrival_l], 32x32_step1
    mul_step[in_cir, inter_arrival_l], 32x32_step2
    mul_step[in_cir, inter_arrival_l], 32x32_step3
    mul_step[in_cir, inter_arrival_l], 32x32_step4
    mul_step[burst_update, --], 32x32_last
    mul_step[burst_update_h, --], 32x32_last2

    // if burst_update exceeds 32 bits set TC and TE to maximum values
    bne[set_max_tc_te#]


new_tc_te#:
    .reg overflow_tokens

    ; Increase TC (green) bucket with burst_update tokens
    alu[out_new_tc, in_curr_tc, +, burst_update]
    ; if carry occurs it means that TC exceeds CBS, and we must fill TE bucket
    bcs[tc_overflow#], defer[1]
    ; overflow_tokens = tc - cbs
    alu[overflow_tokens, out_new_tc, -, in_cbs]

    ; if overflow is negative, we do not need to update TE bucket
    blo[macro_end#], defer[1]
    alu[out_new_te, --, B, in_curr_te]

tc_overflow#:

    ; Update TE bucket with overflow tokens from TC: te = te + overflow_tokens
    alu[out_new_te, in_curr_te, +, overflow_tokens]

    ; if carry occurs updated TE exceeds EBS and we must set TE to maximum
    bcs[te_overflow#], defer[1]
    ;overflow_tokens = TE - EBS
    alu[overflow_tokens, out_new_te, -, in_ebs]

    ; if overflow is negative, TE bucket is ok and we can finish
    blo[macro_end#], defer[1]

te_overflow#:
set_max_tc_te#:

    ; set TC bucket to max bucket size: TC = CBS
    alu[out_new_tc, --, B , in_cbs]
    ; set TE tokens to max bucket size TE = EBS
    alu[out_new_te, --, B , in_ebs]


macro_end#:



.end
#endm


//------------------------------------------------------------------------------
//#macro _trtcm_token_update(out_new_tc,           out_new_tp,                \
//                           out_last_timestamp_l, out_last_timestamp_h,      \
//                           in_curr_timestamp_l,  in_curr_timestamp_h,       \
//                           in_last_timestamp_l,  in_last_timestamp_h,       \
//                           in_curr_tc,           in_curr_tp,                \
//                           in_cbs,               in_pbs,                    \
//                           in_cir,               in_pir,                    \
//                           in_pir_rate,          in_cir_rate)

// _trtcm_token_update
//   Calculates how many tokens are to be added to TC, TP buckets depending
//   on the formulas:
//      TC =+ (in_curr_timestamp - in_last_timestamp) * in_cir
//      TP =+ (in_curr_timestamp - in_last_timestamp) * in_pir

// Inputs:
//
//   in_curr_timestamp_l
//   in_curr_timestamp_h  - current timestamp read by the thread 0 and
//                          used by all the others.
//   in_last_timestamp_l
//   in_last_timestamp_h  - stamp with the time at which the last packet has
//                          been handled in this flow
//   in_curr_tc           - current content of the bucket with green tockens
//   in_curr_tp           - current content of the bucket with yellow tockens
//   in_cbs, in_pbs       - the max value of green and yellow buckets
//   in_cir,              - commited information rate for this flow
//   in_pir,              - commited information rate for this flow
//   in_pir_rate          - time scaling factor for calculation tc update
//   in_cir_rate          - time scaling factor for calculation tp update
//
// Outputs:
//
//   out_new_tc           - calculated new value of green tockens in the bucket
//   out_new_tp           - calculated new value of yellow tockens in the bucket
//   out_last_timestamp_l - new last timestamp to be written back to LM
//   out_last_timestamp_h - new last timestamp to be written back to LM
//
//
#macro _trtcm_token_update(out_new_tc,           out_new_tp,                \
                           out_last_timestamp_l, out_last_timestamp_h,      \
                           in_curr_timestamp_l,  in_curr_timestamp_h,       \
                           in_last_timestamp_l,  in_last_timestamp_h,       \
                           in_curr_tc,           in_curr_tp,                \
                           in_cbs,               in_pbs,                    \
                           in_cir,               in_pir,                    \
                           in_cir_rate,          in_pir_rate)
.begin
.reg pir_to_cir_rate

        ; calculate the shift of the timestamp for TP computation
        alu[--, in_pir_rate, OR, 0]     ;5 lower bits contain rate adjust
        dbl_shf[out_last_timestamp_l,
                in_curr_timestamp_h, in_curr_timestamp_l, >>indirect]

        alu[pir_to_cir_rate, in_pir_rate, OR, 0]     ;5 lower bits contain rate adjust
        dbl_shf[out_last_timestamp_h,
                0,                   in_curr_timestamp_h, >>indirect]

        .reg inter_arrival_l
        .reg inter_arrival_h

        alu[inter_arrival_l, out_last_timestamp_l, -,      in_last_timestamp_l]
        alu[inter_arrival_h, out_last_timestamp_h, -carry, in_last_timestamp_h]
        ; if burst_update exceeds 32 bits set TP to maximum value
        bne[set_max_tp#]

        .reg burst_update burst_update_h

calculate_tp#:
        //Multiply of pir and inter_arrival:  burst_update = pir * inter_arrival
        mul_step[in_pir, inter_arrival_l], 32x32_start
        mul_step[in_pir, inter_arrival_l], 32x32_step1
        mul_step[in_pir, inter_arrival_l], 32x32_step2
        mul_step[in_pir, inter_arrival_l], 32x32_step3
        mul_step[in_pir, inter_arrival_l], 32x32_step4
        mul_step[burst_update, --], 32x32_last
        mul_step[burst_update_h, --], 32x32_last2

        //      if burst_update exceeds 32 bits set TP to maximum value
        bne[set_max_tp#]

        alu[out_new_tp, in_curr_tp, +, burst_update]

        //      if burst_update + TP exceeds 32 bits set TP to maximum value
        bcs[set_max_tp#]

        alu[--, out_new_tp, -, in_pbs]
        blo[calculate_tc#]

set_max_tp#:

        alu[out_new_tp, --, B, in_pbs]

calculate_tc#:

        ; calculate the shift of the timestamp for TC computation
        alu[--, in_cir_rate, OR, 0]    ;5 lower bits contain rate adjust
        dbl_shf[inter_arrival_l,
                in_curr_timestamp_h, in_curr_timestamp_l, >>indirect]


        ;5 lower bits contain rate adjust
        alu[--, in_cir_rate, OR, 0]
        dbl_shf[inter_arrival_h, 0, in_curr_timestamp_h, >>indirect]

        .begin
           .reg cir_last_timestamp_l
           .reg cir_last_timestamp_h
           .reg temp_last_timestamp_h

            alu[temp_last_timestamp_h, --, B, in_last_timestamp_h]
            alu[pir_to_cir_rate, in_cir_rate, -, pir_to_cir_rate]

            ;5 lower bits contain rate adjust
            alu[--, pir_to_cir_rate, OR, 0]
            dbl_shf[cir_last_timestamp_l, \
                    temp_last_timestamp_h, in_last_timestamp_l, >>indirect]

            ;5 lower bits contain rate adjust
            alu[--, pir_to_cir_rate, OR, 0]
            dbl_shf[cir_last_timestamp_h, \
                    0, in_last_timestamp_h, >>indirect]

            alu[inter_arrival_l, \
                inter_arrival_l, -, cir_last_timestamp_l]

            alu[inter_arrival_h, \
                inter_arrival_h, -carry, cir_last_timestamp_h]
        
            ; if burst_update exceeds 32 bits set TP to maximum value
            bne[set_max_tc#]

        .end

        //Multiply of cir and inter_arrival:  burst_update = cir * inter_arrival
        mul_step[in_cir, inter_arrival_l], 32x32_start
        mul_step[in_cir, inter_arrival_l], 32x32_step1
        mul_step[in_cir, inter_arrival_l], 32x32_step2
        mul_step[in_cir, inter_arrival_l], 32x32_step3
        mul_step[in_cir, inter_arrival_l], 32x32_step4
        mul_step[burst_update, --], 32x32_last
        mul_step[burst_update_h, --], 32x32_last2

        //      if burst_update exceeds 32 bits set TC to maximum value
        bne[set_max_tc#]

        alu[out_new_tc, in_curr_tc, +, burst_update]

        //      if burst_update + TC exceeds 32 bits set TC to maximum value
        bcs[set_max_tc#]

        alu[--, out_new_tc, -, in_cbs]
        blo[macro_end#]

set_max_tc#:

        alu[out_new_tc, --, B, in_cbs]

macro_end#:


.end
#endm


//------------------------------------------------------------------------------
//#macro _tc_metering(out_in_curr_tc,       \
//                    out_in_curr_te,       \
//                    in_dscp_values,       \
//                    in_packet_size,       \
//                    in_color_id)
//
// _tc_metering
//   Performs actual flow measuring, and packet-color determination.
//   This macro performs color-of-a-packet calculation based on the values tc,
//   te, and packet_size. It modifies statistics if enabled.
//
// Parameters:
//   #define  TC_METER_COLOR_AWARE_FEATURE       - enables color-aware mode
//   #define  TC_METER_PACKET_COUNTER_FEATURE    - enables statistics gathering
//
// Inputs:
//   out_in_curr_tc       - number of tockens currently available in green
//                          bucket
//   out_in_curr_te       - number of tockens currently available in yellow
//                          bucket
//   in_color_id          - in the case of color aware mode this register must
//                          contain one of the values GREEN_PACKET,
//                          YELLOW_PACKET, RED_PACKET which tell what was
//                          the input color of the processed packet.
//   in_dscp_values       - values of dscp one of which will be used to denote
//                          Diffserv Behavior Aggregate and put in dl_flow_id
//   in_packet_size       - size of the current packet
//
// Local memory inputs:
//   meter entry cache in LM - data of the meter statistics for
//                          the flow which the currently processed packet
//                          belongs to. Active_lm_addr_0 index must point at
//                          the beginning of them.
//
// Outputs:
//   in_out_curr_tc       - number of tockens left in green bucket.
//   in_out_curr_te       - number of tockens left in yellow bucket.
//
// Dispatch loop outputs:
//   color_id             - the newly calculated color for the current packet
//                          It is one of the values GREEN_PACKET, YELLOW_PACKET,
//                          RED_PACKET
//   flow_id              - output flow_id denotes Diffserv Behavior Aggregate
//                          (encoded as DSCP)
//
// Comments:
//   Modifies data in LM (only if statistics collecting is enabled):
//
//   the number of green packets clasified in this flow:
//   l$index0[TC_METER_LM_GREEN_PACKETS]
//   the number of bytes transmitted in this flow in green packets
//   l$index0[TC_METER_LM_GREEN_BYTES]
//   the number of yellow packets clasified in this flow.
//   l$index0[TC_METER_LM_YELLOW_PACKETS]
//   the number of bytes transmitted in this flow in yellow packets
//   l$index0[TC_METER_LM_YELLOW_BYTES]
//   the number of red packets clasified in this flow.
//   l$index0[TC_METER_LM_RED_PACKETS]
//   the number of bytes transmitted in this flow in red packets
//   l$index0[TC_METER_LM_RED_BYTES]
//
#macro _tc_metering(out_in_curr_tc,       \
                    out_in_curr_te,       \
                    in_dscp_values,       \
                    in_packet_size,       \
                    in_color_id)
.begin


#ifdef TC_METER_TRTCM_FEATURE
#ifdef TC_METER_SRTCM_FEATURE

     alu[--, *l$index0[TC_METER_LM_METER_TYPE], AND, TC_METER_TRTCM_ENABLE_FLAG]
     bne[use_trtcm_algorithm#]

#endif
#endif


.reg DSCP ;value to be written to output flow_id (in packet meta data)


//------------------------------------------------------------------------------
#ifdef TC_METER_SRTCM_FEATURE
//------------------------------------------------------------------------------

use_srtcm_algorithm#:

        //-----------------------------
        #ifdef TC_METER_COLOR_AWARE_FEATURE

        alu[--, in_color_id, -, RED_PACKET]
        beq[Red#]
        alu[--, in_color_id, -, YELLOW_PACKET]
        beq[CheckYellow#]

        #endif
        //-----------------------------


        CheckGreen#:
             // if (tc < packet_size) goto Green#
             alu[--, out_in_curr_tc, -, in_packet_size]
             bhs[Green#]

        CheckYellow#:
             // packet is definitely not green
             // if (te < packet_size) goto Yellow#
             alu[--, out_in_curr_te, -, in_packet_size]
             bhs[Yellow#], defer[1]
             ; this part of the code pertains the processing of yellow packets
             ; it is moved here to reduce the number of cycles in Yellow# branch
             ; by one at the expense of an increse in Red# branch

             // DSCP = DSCPyellow ( DSCP_adjust >> 18)
             alu_shf[DSCP, 0x3F, AND, in_dscp_values, >>18]
             // packet has to be red if the thread got here

             //----------------------
             #ifdef TC_METER_TRTCM_FEATURE
             br[Red#]
             #endif
             //----------------------

#endif

//------------------------------------------------------------------------------
#ifdef TC_METER_TRTCM_FEATURE
//------------------------------------------------------------------------------


use_trtcm_algorithm#:
            //-----------------------------
            #ifdef TC_METER_COLOR_AWARE_FEATURE

            alu[--, in_color_id, -, RED_PACKET]
            beq[Red#]

            #endif
            //-----------------------------


            ; if (tc < packet_size) goto Red#
            alu[--, out_in_curr_te, -, in_packet_size]
            blo[Red#]


            //-----------------------------
            #ifdef TC_METER_COLOR_AWARE_FEATURE

            alu[--, in_color_id, -, YELLOW_PACKET]
            beq[Yellow#], defer[1]
            // DSCP = DSCPyellow ( DSCP_adjust >> 18)
            alu_shf[DSCP, 0x3F, AND, in_dscp_values, >>18]

            #endif
            //-----------------------------



            alu[--, out_in_curr_tc, -, in_packet_size]
            blo[Yellow#], defer[1]
            // DSCP = DSCPyellow ( DSCP_adjust >> 18)
            alu_shf[DSCP, 0x3F, AND, in_dscp_values, >>18]

            br[Green#],defer[1]
            alu[out_in_curr_te,  out_in_curr_te, -, packet_size]

//------------------------------------------------------------------------------
#endif
//------------------------------------------------------------------------------


        Red#:
            ;processing red packets

            // DSCP = DSCPred  (DSCP_adjust >> 10)
            alu_shf[DSCP, 0x3F, AND, in_dscp_values, >>10]

            dl_meta_set_flow_id(DSCP)
            dl_meta_set_color(RED_PACKET)

            //------------------------------------------------------------------
            #ifdef TC_METER_PACKET_COUNTER_FEATURE

                ; read the status to determine whether or not to gather
                ; statistics if bit cleared skip statistics
                br_bclr[in_dscp_values, TC_METER_RED_COUNTER_FLAG, red_end#]

            red_pkt_cnt_update#:
                //update counters in LM
                alu[*l$index0[TC_METER_LM_RED_PACKETS], \
                    *l$index0[TC_METER_LM_RED_PACKETS], +, 1]
                bcs[red_64bit_pkt_cnt_update#]


            red_byte_cnt_update#:
                alu[*l$index0[TC_METER_LM_RED_BYTES],  \
                    *l$index0[TC_METER_LM_RED_BYTES], +, in_packet_size]
                bcs[red_64bit_byte_cnt_update#]


            red_end#:

            #endif // TC_METER_PACKET_COUNTER_FEATURE
            //-------------------------------------------------------------------

            // check next block for this packet
            br_bclr[in_dscp_values, 9, macro_end#], defer[1]
            immed[dl_next_block, TC_METER_NEXT1]

            br_bclr[in_dscp_values, 8, macro_end#], defer[1]
            immed[dl_next_block, TC_METER_NEXT2]

            br[macro_end#], defer[1]
            immed[dl_next_block, TC_METER_NEXT3]



            //------------------------------------------------------------------
            #ifdef TC_METER_PACKET_COUNTER_FEATURE
            .begin
            .reg count_base count_offset

                ; here we update 64-bit statistics - this the exception code

            red_64bit_pkt_cnt_update#:
                move(count_base, (TC_METER_64BIT_STAT_SRAM_BASE + 16))
                alu_shf[count_offset,
                        --, B, flow_id, <<TC_METER_64BIT_STAT_LOG_SRAM_SIZE]

                sram[incr, --, count_base, count_offset]
                br[red_byte_cnt_update#]


            red_64bit_byte_cnt_update#:
                move(count_base, (TC_METER_64BIT_STAT_SRAM_BASE + 20))
                alu_shf[count_offset,
                    --, B, flow_id, <<TC_METER_64BIT_STAT_LOG_SRAM_SIZE]

                sram[incr, --, count_base, count_offset]
                br[red_end#]

            .end
            #endif // TC_METER_PACKET_COUNTER_FEATURE
            //------------------------------------------------------------------




        Green#:

            ;processing green packets

            //--------------------------------
            #ifndef TC_METER_COLOR_AWARE_FEATURE


            ; make use of the fact that if it is color aware mode
            ; packet must be already marked as geen and
            ; there is no need to mark it again

            dl_meta_set_color(GREEN_PACKET)

            #endif
            //--------------------------------

            alu[out_in_curr_tc, out_in_curr_tc,  -, in_packet_size]

            // DSCP = DSCPgreen ( DSCP_adjust >> 26)
            alu_shf[DSCP, 0x3F, AND, in_dscp_values, >>26]
            dl_meta_set_flow_id[DSCP]    // flow_id = new DSCP



            //------------------------------------------------------------------
            #ifdef TC_METER_PACKET_COUNTER_FEATURE

                ; read the status to determin whether or not
                ; to gather statistics if bit cleared skip statistics
                br_bclr[in_dscp_values, TC_METER_GREEN_COUNTER_FLAG, green_end#]


            green_pkt_cnt_update#:
                ; update counters in LM
                alu[*l$index0[TC_METER_LM_GREEN_PACKETS],\
                    *l$index0[TC_METER_LM_GREEN_PACKETS], +, 1]
                bcs[green_64bit_pkt_cnt_update#]

            green_byte_cnt_update#:
                alu[*l$index0[TC_METER_LM_GREEN_BYTES],  \
                    *l$index0[TC_METER_LM_GREEN_BYTES], +, packet_size]
                bcs[green_64bit_byte_cnt_update#]

            green_end#:

            #endif
            //------------------------------------------------------------------



            // check next block for this packet (in meter entry)
            br_bclr[in_dscp_values, 25, macro_end#], defer[1]
            immed[dl_next_block, TC_METER_NEXT1]
            br_bclr[in_dscp_values, 24, macro_end#], defer[1]
            immed[dl_next_block, TC_METER_NEXT2]
            br[macro_end#], defer[1]
            immed[dl_next_block, TC_METER_NEXT3]




            //------------------------------------------------------------------
            #ifdef TC_METER_PACKET_COUNTER_FEATURE
            .begin
            .reg count_base count_offset

                ; here we update 64-bit statistics - this is an exception code

            green_64bit_pkt_cnt_update#:
                move(count_base, (TC_METER_64BIT_STAT_SRAM_BASE + 0))
                alu_shf[count_offset, --, B, flow_id,
                                           <<TC_METER_64BIT_STAT_LOG_SRAM_SIZE]
                sram[incr, --, count_base, count_offset]
                br[green_byte_cnt_update#]

            green_64bit_byte_cnt_update#:
                move(count_base, (TC_METER_64BIT_STAT_SRAM_BASE + 4))
                alu_shf[count_offset, --, B, flow_id,
                                           <<TC_METER_64BIT_STAT_LOG_SRAM_SIZE]
                sram[incr, --, count_base, count_offset]
                br[green_end#]

            .end
            #endif // TC_METER_PACKET_COUNTER_FEATURE
            //------------------------------------------------------------------



        Yellow#:
        ;processing yellow packets

            alu[out_in_curr_te,  out_in_curr_te, -, packet_size]

            // set flow_id = DSCP yellow
            dl_meta_set_flow_id[DSCP]
            dl_meta_set_color(YELLOW_PACKET)



            //------------------------------------------------------------------
            #ifdef TC_METER_PACKET_COUNTER_FEATURE

              ; read the status to determine whether or not
              ; to gather statistics - if bit cleared skip statistics
              br_bclr[in_dscp_values, TC_METER_YELLOW_COUNTER_FLAG, yellow_end#]

                //update counters in LM
            yellow_pkt_cnt_update#:
                alu[*l$index0[TC_METER_LM_YELLOW_PACKETS],\
                    *l$index0[TC_METER_LM_YELLOW_PACKETS], +, 1]
                bcs[yellow_64bit_pkt_cnt_update#]

            yellow_byte_cnt_update#:
                alu[*l$index0[TC_METER_LM_YELLOW_BYTES],  \
                    *l$index0[TC_METER_LM_YELLOW_BYTES], +, packet_size]
                bcs[yellow_64bit_byte_cnt_update#]


            yellow_end#:

            #endif // TC_METER_PACKET_COUNTER_FEATURE
            //------------------------------------------------------------------


            // check next block for this packet (in meter entry)
            br_bclr[in_dscp_values, 17, macro_end#], defer[1]
            immed[dl_next_block, TC_METER_NEXT1]

            br_bclr[in_dscp_values, 16, macro_end#], defer[1]
            immed[dl_next_block, TC_METER_NEXT2]

            br[macro_end#], defer[1]
            immed[dl_next_block, TC_METER_NEXT3]


            //------------------------------------------------------------------
            #ifdef TC_METER_PACKET_COUNTER_FEATURE
            .begin
            .reg count_base count_offset

                ; here we update 64-bit statistics - this is an exception code

            yellow_64bit_pkt_cnt_update#:
                move(count_base, (TC_METER_64BIT_STAT_SRAM_BASE + 8))
                alu_shf[count_offset, --, B, flow_id,
                                            <<TC_METER_64BIT_STAT_LOG_SRAM_SIZE]
                sram[incr, --, count_base, count_offset]
                br[yellow_byte_cnt_update#]

            yellow_64bit_byte_cnt_update#:
                move(count_base, (TC_METER_64BIT_STAT_SRAM_BASE + 12))
                alu_shf[count_offset, --, B, flow_id,
                                            <<TC_METER_64BIT_STAT_LOG_SRAM_SIZE]
                sram[incr, --, count_base, count_offset]

                // check next block for this packet (in meter entry)
                br_bclr[in_dscp_values, 17, macro_end#], defer[1]
                immed[dl_next_block, TC_METER_NEXT1]

                br_bclr[in_dscp_values, 16, macro_end#], defer[1]
                immed[dl_next_block, TC_METER_NEXT2]

                immed[dl_next_block, TC_METER_NEXT3]

            .end
            #endif // TC_METER_PACKET_COUNTER_FEATURE
            //------------------------------------------------------------------


        macro_end#:


.end
#endm



#endif   //  TC_METER_UTILS_UC

