/*******************************************************************************
  INTEL CONFIDENTIAL

  Copyright 2000 - 2003 Intel Corporation All Rights Reserved.

  The source code contained or described herein and all documents related to
  the source code ("Material") are owned by Intel Corporation or its 
  suppliers or licensors.

  Title to the Material remains with Intel Corporation or its suppliers and
  licensors. The Material contains trade secrets and proprietary and
  confidential information of Intel or its suppliers and licensors.
  The Material is protected by worldwide copyright and trade secret laws and
  treaty provisions. No part of the Material may be used, copied, reproduced,
  modified, published, uploaded, posted, transmitted, distributed,
  or disclosed in any way without Intel's prior express written permission.

  No license under any patent, copyright, trade secret or other intellectual
  property right is granted to or conferred upon you by disclosure
  or delivery of the Materials, either expressly, by implication, inducement,
  estoppel or otherwise. Any license under such intellectual property rights
  must be express and approved by Intel in writing.

*******************************************************************************/

/**
 * ============================================================================
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2000 Network Processor
 *
 * = FILENAME
 *      ix_cc_tc_meter_api.h
 *
 * = DESCRIPTION
 *      TC METER Core Component internals - Library API
 *      This file defines function prototypes of the TC METER CC Libary API.
 *
 * ============================================================================
 */

#if !defined(IX_CC_TC_METER_API_H)
#define      IX_CC_TC_METER_API_H

#if defined( __cplusplus )
extern "C" {
#endif /* __cplusplus */




/**
 * @brief  This function adds a new entry in the TC METER meter table.
 *
 * The core component uses <code>arg_nInstance</code> to select a row in the 
 * meter table. If the row is already configured, the function reports an 
 * error. In that case, it does not override instance parameters stored in the
 * table. Otherwise, if an entry is empty, the core component copies meter 
 * instance parameters from <code>arg_pParameters</code> to the selected entry.
 *
 * @param  arg_nInstance    index of a meter table entry that should be 
 *                          configured with new parameters. The argument must
 *                          not be higher than the value defined in 
 *                          \\TC METER\TABLE_SRAM_SIZE system property
 * 
 * @param  arg_pParams      pointer to the structure of parameters defining
 *                          a meter instance.
 *
 * @param  arg_pUserContext pointer to the control block allocated during core
 *                          component's initialization process.
 *
 * @return <code>IX_SUCCESS</code> if successful or a valid <code>ix_error</code>
 *         token for failure. The following error codes might be returned via
 *         the token:
 *\n
 *\n       IX_CC_ERROR_RANGE               - entry index is out of range.  
 *\n       IX_CC_ERROR_NULL                - arg_pParams is NULL.  
 *\n       IX_CC_ERROR_CONFLICTING_ENTRY   - entry already exist.  
 *\n       IX_CC_TC METER_ERROR_INVALID_PARAM - arg_pParams structure contains 
 *                                           invalid values.
 *
 * @see    <code>ix_cc_tc_meter_parameters</code>
 */
ix_error
ix_cc_tc_meter_add_entry(ix_uint32                  arg_nInstance, 
                         ix_cc_tc_meter_parameters *arg_pParams, 
                         void                      *arg_pUserContext);

/**
 * @brief  This function removes an entry from the TC METER meter table.
 *
 * The  function  checks if an entry, pointed by <code>arg_nInstance</code>, is 
 * valid. If true, it zeros the CIR field (to mark the entry as empty). It also
 * zeros the statistics. The rest of fields are left unchanged. 
 *
 * If the  referred entry  is  already marked as empty, the function returns an 
 * error token. 
 *
 * The  core  component  uses  zero  CIR value to mark the empty entries in the 
 * meter instance  table. However,  when the microblock processes a packet,  it 
 * does not check whether the CIR field is zero, it is whether a packet matches
 * an empty entry in the meter table. It behaves as if the entry was not empty.
 *
 * This function assumes that the  application that removes a meter instance is 
 * responsible for  changing the configuration of the  classifier block in such 
 * way that no packets match the empty meter entries.
 *
 * @param  arg_nInstance    index of a meter table entry that should be removed.
 *                          The argument must not be higher than the value 
 *                          defined in \\TC METER\TABLE_SRAM_SIZE system property.
 *
 * @param  arg_pUserContext pointer to the control block allocated during core 
 *                          component's initialization process.
 *
 * @return <code>IX_SUCCESS</code> if successful or a valid <code>ix_error</code>
 *         token for failure. The following error codes might be returned via
 *         the token:
 *\n
 *\n       IX_CC_ERROR_RANGE           - entry index out of range.
 *\n       IX_CC_ERROR_NULL            - arg_pUserContext is NULL.
 *\n       IX_CC_ERROR_ENTRY_NOT_FOUND - entry doesn't exists (has not been 
 *                                       initialized).
 */
ix_error
ix_cc_tc_meter_remove_entry(ix_uint32  arg_nInstance, 
                            void      *arg_pUserContext);

/**
 * @brief  This function updates an entry in the TC METER meter table.
 *
 * The function checks if a meter entry, pointed by <code>arg_nInstance</code>,
 * is valid.  If  true,  the  core component updates a meter entry with values 
 * provided in <code>arg_pParams</code>. If  the  referred entry is empty, the 
 * core  component  reports  an  error. In such case, meter parameters are not 
 * modified and the table entry remains empty.
 *
 * @param  arg_nInstance    index of a meter table entry that should be updated. 
 *                          The argument must not be higher than the value 
 *                          defined in \\TC METER\TABLE_SRAM_SIZE system property.
 *
 * @param  arg_pParams      pointer to the structure of updated parameters for
 *                          a meter instance.
 *
 * @param  arg_pUserContext pointer to the control block allocated during core 
 *                          component's initialization process 
 *
 * @return <code>IX_SUCCESS</code> if successful or a valid <code>ix_error</code>
 *         token for failure. The following error codes might be returned via 
 *         the token:
 *\n
 *\n       IX_CC_ERROR_RANGE               - entry index is out of range.
 *\n       IX_CC_ERROR_NULL                - arg_pParams or arg_pUserContext is 
 *                                           NULL.
 *\n       IX_CC_ERROR_ENTRY_NOT_FOUND     - entry doesn't exists (has not been 
 *\n                                        initialized).
 *\n       IX_CC_TC METER_ERROR_INVALID_PARAM - arg_pParams strucutre contains 
 *                                           invalid values.
 */
ix_error ix_cc_tc_meter_update_entry( ix_uint32                arg_nInstance,
                                   ix_cc_tc_meter_parameters * arg_pParams,
                                   void *                   arg_pUserContext );
/**
 * @brief  The function reads statistics for the specified entry
 *
 * The function first checks if the meter entry, pointed by 
 * <code>arg_nInstance</code>, is valid. If the referred table entry is empty,
 * the functioin returns an error via a token. Otherwise, it retrievs 64-bit
 * statistic counters associated with specified entry and fills in 
 * <code>arg_pParams</code> structure with valid data.
 *
 * @param  arg_nInstance    index of a meter table entry for reading 
 *                          statistics.
 *
 * @param  arg_pStats       pointer to the statistics structure that should be
 *                          filled with data by the function.
 *
 * @param  arg_pUserContext pointer to the control block allocated during core
 *                          component's initialization process.
 *
 * @return <code>IX_SUCCESS</code> if successful or a valid <code>ix_error</code>
 *         token for failure. The following error codes might be returned via 
 *         the token:
 *\n
 *\n       IX_CC_ERROR_RANGE           - entry index is out of range.
 *\n       IX_CC_ERROR_NULL            - arg_pStats is NULL.
 *\n       IX_CC_ERROR_ENTRY_NOT_FOUND - entry doesn't exists (has not been 
 *                                       initialized).
 *
 * @see ix_cc_tc_meter_statistics
 */
ix_error
ix_cc_tc_meter_get_statistics(ix_uint32                  arg_nInstance, 
                              ix_cc_tc_meter_statistics *arg_pStats,
                              void                      *arg_pUserContext);

/** 
 * @brief  This function gets parameters of an entry from the TC METER table.
 *
 * The core component first checks if an entry, pointed by arg_Instance, is 
 * valid. If true, it reads the entry parameters to the arg_pParams argument. 
 * Otherwise, if the referred entry is marked empty, the core component
 * reports an error in a callback function. 
 *
 *
 * @param  arg_nInstance index of an TC METER table entry. 
 *                       The value must not be higher than TC METER\TABLE_SRAM_SIZE 
 *                       system property.
 *
 * @param  arg_pParams  pointer to the parameters structure which should be 
 *                      filled with entry parameters.
 *
 * @param  arg_pContext  pointer to the control block allocated in
 *                       <code>ix_cc_tc_meter_init()</code>
 *
 * @return IX_SUCCESS if successful or a valid ix_error token for failure. The 
 *         following error codes might be returned via the token.
 *\n        
 *         IX_CC_ERROR_ENTRY_NOT_FOUND - entry does not exist
 *         IX_CC_ERROR_RANGE           - TC METER instance out of range
 *
 * @see    ix_cc_tc_meter_add_entry()
 */
ix_error
ix_cc_tc_meter_get_parameters(ix_uint32                  arg_nInstance, 
                              ix_cc_tc_meter_parameters *arg_pParams,
                              void                      *arg_pContext);

#if defined( __cplusplus )
}
#endif /* __cplusplus */

#endif /* IX_CC_TC_METER_API_H */
