/*******************************************************************************
  INTEL CONFIDENTIAL

  Copyright 2000 - 2003 Intel Corporation All Rights Reserved.

  The source code contained or described herein and all documents related to
  the source code ("Material") are owned by Intel Corporation or its 
  suppliers or licensors.

  Title to the Material remains with Intel Corporation or its suppliers and
  licensors. The Material contains trade secrets and proprietary and
  confidential information of Intel or its suppliers and licensors.
  The Material is protected by worldwide copyright and trade secret laws and
  treaty provisions. No part of the Material may be used, copied, reproduced,
  modified, published, uploaded, posted, transmitted, distributed,
  or disclosed in any way without Intel's prior express written permission.

  No license under any patent, copyright, trade secret or other intellectual
  property right is granted to or conferred upon you by disclosure
  or delivery of the Materials, either expressly, by implication, inducement,
  estoppel or otherwise. Any license under such intellectual property rights
  must be express and approved by Intel in writing.

*******************************************************************************/

/**
 * ============================================================================
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2000 Network Processor
 *
 * = FILENAME
 *      ix_cc_tc_meter.h
 *
 * = DESCRIPTION
 *     Three Color Meter Core Component External API
 *     This file defines function prototypes of the external 
 *     API and error codes for the TCM Core Component.
 *
 * ============================================================================
 */

#if !defined(IX_CC_TC_METER_H)
#define      IX_CC_TC_METER_H


/**
 * This enumeration defines error codes specific to the TCM core component.
 */
typedef enum ix_e_cc_tc_meter_error_code
  {
    IX_CC_TC_METER_ERROR_FIRST   = IX_ERROR_MAKE_GROUP(CC_TC_METER),
    IX_CC_TC_METER_ERROR_SUCCESS = IX_CC_TC_METER_ERROR_FIRST,

    /**
     * Symbols patching procedure failure. This error code occurs only if
     * configuration of the TC Meter microblock failed.
     */
    IX_CC_TC_METER_ERROR_FAILED_PATCHING,

    /**
     * Static configuration reading failed (only if registry available).
     */
    IX_CC_TC_METER_ERROR_REGISTRY,

    /**
     * Internal CCI library failure (only for level 1).
     */
    IX_CC_TC_METER_ERROR_CCI,

    /**
     * Invalid packet recieved by ix_cc_tc_meter_pkt_handler.
     */
    IX_CC_TC_METER_ERROR_INVALID_PACKET,

    /**
     * The ix_cc_tc_meter_parameters structure contains invalid meter
     * configuration (i.e. CIR == 0).
     */
    IX_CC_TC_METER_ERROR_INVALID_PARAM,

    /**
     * Invalid TC Meter core component handle specified. The core component's
     * handle doesn't match one hold in cc's context.
     */
    IX_CC_TC_METER_ERROR_INVALID_HANDLE,

    /**
     * Invalid hardware buffer specified in packet or message handler.
     */
    IX_CC_TC_METER_ERROR_INVALID_BUFFER,

    /* unknown error code */
    IX_CC_TC_METER_ERROR_UNKNOWN,
    IX_CC_TC_METER_ERROR_LAST
  } ix_cc_tc_meter_error_code;


#if defined( __cplusplus )
extern "C" {
#endif /* __cplusplus */

/**
 * @brief  Initialization of the TC Meter Core Component
 *
 * The function initializes the core component. It should be called and return
 * successfully before any other function in the core component can be called.
 * It allocates the memory for the symbols and internal structs. The function
 * registers message and packet handlers, too.
 *
 * @param  arg_hCC           handle to the core component,
 *
 * @param  arg_ppContext     a placeholder where the pointer to the control
 *                           block will be stored.
 *
 * @return <code>IX_SUCCESS</code> if successful or a valid <code>ix_error</code>
 *         token for failure. The following error codes might be returned via
 *         the token.
 *\n
 *\n       IX_CC_ERROR_NULL                  - NULL parameter used (either
 *                                             arg_hCC or arg_ppContext).
 *\n       IX_CC_ERROR_OOM_SYSTEM            - local memory allocation failure
 *\n       IX_CC_ERROR_INTERNAL              - error in reading static
 *                                             configuration.
 *\n       IX_CC_ERROR_OOM_SRAM              - SRAM memory allocation failure.
 *\n       IX_CC_ERROR_INTERNAL              - packet or message handler
 *                                             registration failure.
 *\n       IX_CC_TC_METER_ERROR_FAILED_PATCHING - symbol patching procedure failure.
 *\n       IX_CC_TC_METER_ERROR_CCI             - internal CCI error (level 1 only).
 *
 * @see    <code>ix_cc_tc_meter_fini()</code>
 */
ix_error
ix_cc_tc_meter_init(ix_cc_handle   arg_hCC,
                    void         **arg_ppContext);


/**
 * @brief  De-initialization of the TC Meter Core Component
 *
 * The function terminates the TC Meter core component. It is executed when the
 * execution engine running the core component is being shut down. This
 * function is performing the SRAM memory de-allocation, removing message and
 * packet handlers, freeing an internal control block.
 *
 * @param  arg_hCC          handle to the core component
 *
 * @param  arg_pContext     pointer to the control block that was allocated
 *                          in <code>ix_cc_tc_meter_init()</code> function.
 *
 * @return <code>IX_SUCCESS</code> if successful or a valid <code>ix_error</code>
 *         token for failure. The following error codes might be returned via
 *         the token.
 *\n
 *\n       IX_CC_ERROR_NULL                 - NULL parameter used (either
 *                                            arg_hCC or arg_ppContext).
 *\n       IX_CC_ERROR_INTERNAL             - message or packet handler
 *                                            unregistration failure (level 0).
 *\n       IX_CC_ERROR_OOM_SRAM             - de-allocation of SRAM memory
 *                                            failed.
 *\n       IX_CC_TC_METER_ERROR_CCI            - message or packet handler
 *                                            unregistration failure (level 1).
 *\n       IX_CC_TC_METER_ERROR_INVALID_HANDLE - the core component's handle
 *                                            (arg_hCC) does match the control
 *                                            block (arg_pContext)
 *
 * @see    <code>ix_cc_tc_meter_init()</code>
 */
ix_error
ix_cc_tc_meter_fini(ix_cc_handle  arg_hCC,
                    void         *arg_pContext);


/**
 * @brief  The TC Meter Core Component message handler
 *
 * The core component uses one message handler for processing all types of
 * messages. The handler receives messages from a single source: DiffServ
 * application. A set of supported messages covers managing TC Meter instances
 * and configuring their parameters as specified in [RFC2697]. Internally,
 * the handler function parses message payloads and calls appropriate Library
 * API functions to perform the requested operation.
 *
 * @param  arg_hData        handle to the buffer embedding information for the
 *                          message passed in <code>arg_nMsgCode</code>
 *
 * @param  arg_nMsgType     message type
 *
 * @param  arg_pContext     pointer to a control block that was allocated
 *                          during initialization process
 *
 * @return <code>IX_SUCCESS</code> if successful or a valid <code>ix_error</code>
 *         token for failure. The folowing error codes might be returned via
 *         the token.
 *\n
 *\n       IX_CCI_ERROR_SEND_FAIL - failure to send the packet to the
 *                                  microblock.
 *
 * @see    <code>ix_cc_tc_meter_init()</code>,
 *         <code>ix_cc_tc_meter_pkt_handler()</code>
 */
ix_error
ix_cc_tc_meter_msg_handler(ix_buffer_handle  arg_hData,
                           ix_uint32         arg_nMsgType,
                           void             *arg_pContext);

/**
 * @brief The TC Meter Core Component packet handler
 *
 * The core component uses one packet handler for processing all packets. When
 * invoked, the handler simply redirects a packet to the TC Meter microblock. The
 * core component does not execute the metering algorithm in order to avoid
 * synchronization with the microblock.
 *
 * @param arg_hData         buffer handle to the received packet
 *
 * @param arg_nExCode       an exception code thrown by a microblock;
 *                          not used by the TC Meter core component
 *
 * @param arg_pUserContext  pointer to the control block that was allocated
 *                          during initialization process
 *
 * @return <code>IX_SUCCESS</code> if successful or a valid <code>ix_error</code>
 *         token for failure. The following error codes might be returned via
 *         the token.
 *\n
 *\n       IX_CC_ERROR_NULL          - arg_pContext or buffer handle is NULL.
 *\n       IX_CC_ERROR_UNDEFINED_MSG - invalid message type received.
 *\n       IX_CC_ERROR_INTERNAL      - function was unable to free the buffer.
 *\n       IX_CC_ERROR_SEND_FAIL     - reply message sending error  .
 *
 * @see    <code>ix_cc_tc_meter_init()</code>,
 *         <code>ix_cc_tc_meter_msg_handler()</code>
 */
ix_error
ix_cc_tc_meter_pkt_handler(ix_buffer_handle  arg_hData,
                           ix_uint32         arg_nExCode,
                           void             *arg_pUserContext);

#if defined( __cplusplus )
}
#endif /* __cplusplus */

#endif /* IX_CC_TC_METER_H */
