/*******************************************************************************
  INTEL CONFIDENTIAL

  Copyright 2000 - 2003 Intel Corporation All Rights Reserved.

  The source code contained or described herein and all documents related to
  the source code ("Material") are owned by Intel Corporation or its 
  suppliers or licensors.

  Title to the Material remains with Intel Corporation or its suppliers and
  licensors. The Material contains trade secrets and proprietary and
  confidential information of Intel or its suppliers and licensors.
  The Material is protected by worldwide copyright and trade secret laws and
  treaty provisions. No part of the Material may be used, copied, reproduced,
  modified, published, uploaded, posted, transmitted, distributed,
  or disclosed in any way without Intel's prior express written permission.

  No license under any patent, copyright, trade secret or other intellectual
  property right is granted to or conferred upon you by disclosure
  or delivery of the Materials, either expressly, by implication, inducement,
  estoppel or otherwise. Any license under such intellectual property rights
  must be express and approved by Intel in writing.

*******************************************************************************/

/**
 * ============================================================================
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2000 Network Processor
 *
 * = FILENAME
 *      ix_cc_tc_meter_msg_helpers.h
 *
 * = DESCRIPTION
 *      TCM Core Component Message Helper API
 *      This file provides function prototypes and datatype 
 *      definitions of the TCM Core Component Message Helper API. 
 *
 * ============================================================================
 */

#if !defined(IX_CC_TC_METER_MSG_HELPERS_H)
#define      IX_CC_TC_METER_MSG_HELPERS_H



/*
 * TYPENAME: ix_cc_tc_meter_output  */
/**
 * ABSTRACT: This enumeration defines outputs from the microblock.
 *
 * DESCRIPTION: 
 */
typedef enum ix_e_cc_tc_meter_output
{
    IX_CC_TC_METER_NEXT1 = 1,
    IX_CC_TC_METER_NEXT2,
    IX_CC_TC_METER_NEXT3
} ix_cc_tc_meter_output;


/*
 * TYPENAME: ix_cc_tc_meter_type   */
/**
 * ABSTRACT:
 *
 * DESCRIPTION: 
 */
typedef enum {
	/*! single rate three color meter        */
    IX_CC_TC_METER_SRTCM_TYPE = 0,
	/*! two rate three color meter           */
    IX_CC_TC_METER_TRTCM_TYPE
} ix_cc_tc_meter_type;


/**
 * This data type defines a structure of externally visible parameters, 
 * configurable for a meter instance.
 */
typedef struct ix_s_cc_srtcm_parameters
{
    ix_uint32          cir;          /**< Committed Information Rate in kbits/sec (cannot be zero). */
    ix_uint32          cbs;          /**< Committed Burst Size in bytes.        */
    ix_uint32          ebs;          /**< Excess Burst Size in bytes.           */
} ix_cc_srtcm_parameters;




/**
 * This data type defines a structure of externally visible parameters, 
 * configurable for a meter instance.
 */
typedef struct ix_s_cc_trtcm_parameters
{
    ix_uint32          cir;          /**< Committed Information Rate in kbits/sec (cannot be zero). */
    ix_uint32          pir;          /**< Peak Inforamtion Rate in kbit/sec.    */
    ix_uint32          cbs;          /**< Committed Burst Size in bytes.        */
    ix_uint32          pbs;          /**< Peak Burst Size in bytes.             */
} ix_cc_trtcm_parameters;




typedef struct ix_s_cc_tc_meter_parameters
{
    ix_uint8              meter_type_id;

    union {
        ix_cc_srtcm_parameters srtcm;
        ix_cc_trtcm_parameters trtcm;
    } meter_type;

    ix_uint8              greenDscp;    /**< DSCP value to mark for green packets. */
    ix_cc_tc_meter_output greenOutput;  /**< Next block for green packets.         */
    ix_uint8              yellowDscp;   /**< DSCP value to mark for yellow packets.*/ 
    ix_cc_tc_meter_output yellowOutput; /**< Next block for yellow packets.        */
    ix_uint8              redDscp;      /**< DSCP value to mark for red packets.   */
    ix_cc_tc_meter_output redOutput;    /**< Next block for red packets.           */
    ix_uint8              statsEnabled; /**< if true, statistics are gathered (boolean value).         */

} ix_cc_tc_meter_parameters;




/**
 * This data type defines a structure of statistic counters associated with 
 * a meter instance. 
 */
typedef struct ix_s_cc_tc_meter_statistics
  {
    /** 
     * Number of green packets metered.
     */
    ix_uint64 greenPacketCount;	 
    
    /** 
     * Number of green bytes metered.
     */
    ix_uint64 greenByteCount;	 
    
    /** 
     * Number of yellow packets metered.
     */
    ix_uint64 yellowPacketCount; 

    /** 
     * Number of yellow bytes metered.
     */
    ix_uint64 yellowByteCount;	 

    /** 
     * Number of red packets metered.
     */
    ix_uint64 redPacketCount;	 

    /** 
     * Number of red bytes metered.
     */
    ix_uint64 redByteCount;	 

  } ix_cc_tc_meter_statistics;

/**
 * @brief Statistics read callback function definition
 *
 * This callback function definition is provided for statistics reading 
 * operation only.
 *
 * @param arg_Result       indicates error conditions for the call: IX_SUCCESS 
 *                         if successful or a valid ix_error token for failure. 
 *                         The following error codes might be returned via the
 *                         token:
 * \n
 * \n                      IX_CC_ERROR_ENTRY_NOT_FOUND - 
 *                                entry is empty (not yet configured).
 *
 * \n                      IX_CC_ERROR_RANGE           - 
 *                                entry index is out of range; the instance 
 *                                identifier is higher than the maximum number 
 *                                of instances that can be configured in the 
 *                                block.
 * \n                        
 * @param arg_pStatsResult pointer to the statistics structure filled with 
 *                         valid data if a meter entry is not empty.
 *
 * @param arg_pContext     pointer to the client-defined context, provided as
 *                         a parameter in the API call. Used by the caller to
 *                         match asynchronous request with a response.
 *
 * @see   <code>ix_cc_tc_meter_async_get_statistics()</code>,
 *        <code>ix_cc_tc_meter_statistics</code>
 */
typedef ix_error (* ix_cc_tc_meter_callback_gs)(
    ix_error                 arg_Result, 
    ix_cc_tc_meter_statistics * arg_pStatsResult, 
    void *                   arg_pContext );

/**
 * @brief Parameters read callback function definition
 *
 * This callback function definition is provided for parameters reading 
 * operation only.
 *
 * @param arg_Result       indicates error conditions for the call: IX_SUCCESS 
 *                         if successful or a valid ix_error token for failure. 
 *                         The following error codes might be returned via the
 *                         token:
 * \n
 * \n                      IX_CC_ERROR_ENTRY_NOT_FOUND - 
 *                                entry is empty (not yet configured).
 *
 * \n                      IX_CC_ERROR_RANGE           - 
 *                                entry index is out of range; the instance 
 *                                identifier is higher than the maximum number 
 *                                of instances that can be configured in the 
 *                                block.
 * \n                        
 * @param arg_pParamsResult pointer to the parameters structure filled with 
 *                         valid data if a meter entry is not empty.
 *
 * @param arg_pContext     pointer to the client-defined context, provided as
 *                         a parameter in the API call. Used by the caller to
 *                         match asynchronous request with a response.
 *
 * @see   <code>ix_cc_tc_meter_async_get_parameters()</code>,
 *        <code>ix_cc_tc_meter_parameters</code>
 */
typedef ix_error (* ix_cc_tc_meter_callback_gp)(
    ix_error                 arg_Result, 
    ix_cc_tc_meter_parameters * arg_pParamsResult, 
    void *                   arg_pContext );

/** 
 * @brief Default callback function definition 
 * 
 * This callback function definition is provided for standard operations 
 * on SRTCM table (adding, removing and updating an entry).
 *
 * @param arg_Result       indicates error conditions for the call: 
 *                         IX_SUCCESS if successful or a valid 
 *                         ix_error token for failure. The 
 *                         following error codes might be returned via the
 *                         token (depending on operation type):
 * \n
 * \n                      IX_CC_ERROR_RANGE - entry index is out of range.
 * \n                      IX_CC_ERROR_NULL               - 
 *                                one of the library API's function argument
 *                                is  NULL.
 * \n                      IX_CC_ERROR_CONFLICTING_ENTRY  - 
 *                                entry already exist.
 * \n                      IX_CC_ERROR_ENTRY_NOT_FOUND    - entry doesn't exist
 *                                             (not yet configured).
 * \n                      IX_CC_SRTCM_ERROR_INVALID_PARAM - 
 *                                             ix_cc_tc_meter_parameters struct 
 *                                             contains invalid values.
 * \n
 * @param arg_pUserContext pointer to the client-defined context, provided as
 *                         a parameter in the API call. Used by the caller to
 *                         match asynchronous request with a response.
 *
 * @see   ix_cc_tc_meter_async_add_entry(),
 *        ix_cc_tc_meter_async_remove_entry(),
 *        ix_cc_tc_meter_async_update_entry(),
 *        ix_cc_tc_meter_parameters
 */
typedef ix_error (* ix_cc_tc_meter_callback_op)( ix_error arg_Result, 
                                              void *   arg_pContext );

#if defined( __cplusplus )
extern "C" {
#endif /* __cplusplus */

/**  
 * @brief Sends an entry add message to the SRTCM core component
 *
 * This helper function builds and sends an entry add message to the SRTCM Core
 * Component. The core component uses <code>arg_nInstance</code> to select a 
 * row in the meter table. If the row is already configured, the core component
 * reports an error (via callback). In this case, it does not override instance 
 * parameters stored in the table. Otherwise, if an entry is empty, the core 
 * component copies data from <code>arg_pParams</code> to the selected entry. 
 *
 * @param arg_nInstance     index of a meter table entry that should be
 *                          configured with new parameters.
 *
 * @param arg_pParams       pointer to the structure of parameters defining
 *                          a meter instance.
 *
 * @param arg_Callback      a client provided callback function. The core 
 *                          component invokes this function to report an
 *                          operation status.
 *
 * @param arg_pContext      pointer to a client defined context. The context
 *                          is returned in a callback function
 *
 * @return IX_SUCCESS if successful or a valid ix_error
 *         token for failure. The folowing error codes might be returned via 
 *         the token.
 * \n
 * \n
 * @see    <code>ix_cc_tc_meter_callback_op</code>
 */
ix_error ix_cc_tc_meter_async_add_entry( ix_uint32                arg_nInstance,
                                      ix_cc_tc_meter_parameters * arg_pParams,
                                      ix_cc_tc_meter_callback_op  arg_Callback,
                                      void *                   arg_pContext );
/**
 * @brief Sends a entry remove message to the SRTCM core component
 *
 * This helper function builds and sends an entry remove message to the SRTCM 
 * core component. The core component first checks if an entry, pointed by 
 * arg_nInstance, is valid. If true, it marks the entry as empty and zeroes the 
 * statistics. If the referred entry is already marked emtpy, the core component
 * reports an error in a callback function.
 *
 * @param arg_nInstance     index of a meter table entry that sould be removed.
 
 * @param arg_Callback      a client provided callback function. The core 
 *                          component invokes this function to report an 
 *                          operation status.
 *
 * @param arg_pUserContext  pointer to client defined context. This value is 
 *                          opaque to the core component. The context is 
 *                          returned in a callback funtion.
 *
 * @return IX_SUCCESS if successful or a valid ix_error token for failure.
 *
 * @see   <code>ix_cc_tc_meter_callback_op</code>
 */
ix_error ix_cc_tc_meter_async_remove_entry(
    ix_uint32 arg_nInstance,
    ix_cc_tc_meter_callback_op arg_Callback,
    void * arg_pUserContext
  );


/**
 * @brief Sends an entry update message to the SRTCM core component
 *
 * This helper function builds and sends an entry update message to the SRTCM 
 * core component. The core component checks if an SRTCM entry, pointed by 
 * arg_nInstance, is valid. If true, the core component updates a meter entry 
 * with values provided in arg_pParams. If the referred entry is empty, the core 
 * component reports an error. Is such case, meter parameters are not modified 
 * and the table entry remains empty.
 *
 * @param  arg_nInstance    index of a meter table entry that should be updated.
 *
 * @param  arg_pParams      pointer to the structure of updated parameters for
 *                          a meter instance. 
 *
 * @param  arg_Callback     a client provided callback function. The core 
 *                          component invokes this function to report an 
 *                          operation status. 
 *         
 * @param  arg_pUserContext pointer to a client defined context. This value is 
 *                          opaque to the core component. The context is 
 *                          returned in a callback function.
 *
 * @return IX_SUCCESS if successful or a valid ix_error token for failure.
 *
 * @see    <code>ix_cc_tc_meter_callback_op</code>\n
 *         <code>ix_cc_tc_meter_parameters</code>
 */
ix_error ix_cc_tc_meter_async_update_entry(
    ix_uint32 arg_nInstance,
    ix_cc_tc_meter_parameters * arg_pParams,  
    ix_cc_tc_meter_callback_op arg_Callback,
    void * arg_pUserContext
  );


/**
 * @brief Sends a statistics read message to the SRTCM core component
 *
 * This helper function builds and sends a statistics read message to the SRTCM 
 * core component. The core component checks fi the meter entry, pointed by 
 * arg_nInstance, is valid. If the referred table entry is empty, the core 
 * component reports an error (via a callback function). Otherwise, it retrieves
 * 64-bit statistics counters associated with this entry and returns them in a 
 * callback function. 
 *
 * @param  arg_nInstance    index of a meter table entry that is queried for
 *                          statistics.
 *         
 * @param  arg_pStats       pointer to the statistics placeholder that should
 *                          be filled with data by the core component.
 *         
 * @param  arg_Callback     a client provided callback function. The core 
 *                          component invokes this function to report an 
 *                          operation status and statistics counters.
 *
 * @param  arg_pUserContext pointer to the client defined context. This value
 *                          is opaque to the core componet. The context is
 *                          returned in a callback function.
 *
 * @return IX_SUCCESS if successful or a valid ix_error token for failure. 
 *         The following error codes might be returned via the token.
 *
 * @see    <code>ix_cc_tc_meter_callback_op</code>\n
 *         <code>ix_cc_tc_meter_statistics</code>
 */
ix_error ix_cc_tc_meter_async_get_statistics(
    ix_uint32 arg_nInstance,
    ix_cc_tc_meter_statistics * arg_pStats,
    ix_cc_tc_meter_callback_gs arg_Callback,
    void * arg_pUserContext
  );

/**
 * @brief Sends a parameters read message to the SRTCM core component
 *
 * This helper function builds and sends a parameters read message to the SRTCM 
 * core component. The core component checks if the meter entry, pointed by 
 * arg_nInstance, is valid. If the referred table entry is empty, the core 
 * component reports an error (via a callback function). Otherwise, it retrieves
 * parameters values associated with this entry and returns them in a callback
 * function. 
 *
 * @param  arg_nInstance    index of a meter table entry that is queried for
 *                          parameters.
 *         
 * @param  arg_pParams      pointer to the parameters placeholder that should
 *                          be filled with data by the core component.
 *         
 * @param  arg_Callback     a client provided callback function. The core 
 *                          component invokes this function to report an 
 *                          operation status and parameters values.
 *
 * @param  arg_pUserContext pointer to the client defined context. This value
 *                          is opaque to the core componet. The context is
 *                          returned in a callback function.
 *
 * @return IX_SUCCESS if successful or a valid ix_error token for failure. 
 *         The following error codes might be returned via the token.
 *
 * @see    <code>ix_cc_tc_meter_callback_gp</code>\n
 *         <code>ix_cc_tc_meter_parameters</code>
 */
ix_error ix_cc_tc_meter_async_get_parameters(
    ix_uint32                arg_nInstance,
    ix_cc_tc_meter_parameters * arg_pParams,
    ix_cc_tc_meter_callback_gp  arg_Callback,
    void *                   arg_pUserContext );

#if defined( __cplusplus )
}
#endif /* __cplusplus */

#endif /* IX_CC_TC_METER_MSG_HELPERS_H */
