/*******************************************************************************
  INTEL CONFIDENTIAL

  Copyright 2000 - 2003 Intel Corporation All Rights Reserved.

  The source code contained or described herein and all documents related to
  the source code ("Material") are owned by Intel Corporation or its 
  suppliers or licensors.

  Title to the Material remains with Intel Corporation or its suppliers and
  licensors. The Material contains trade secrets and proprietary and
  confidential information of Intel or its suppliers and licensors.
  The Material is protected by worldwide copyright and trade secret laws and
  treaty provisions. No part of the Material may be used, copied, reproduced,
  modified, published, uploaded, posted, transmitted, distributed,
  or disclosed in any way without Intel's prior express written permission.

  No license under any patent, copyright, trade secret or other intellectual
  property right is granted to or conferred upon you by disclosure
  or delivery of the Materials, either expressly, by implication, inducement,
  estoppel or otherwise. Any license under such intellectual property rights
  must be express and approved by Intel in writing.

*******************************************************************************/

/**
 * ============================================================================
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2000 Network Processor
 *
 * = FILENAME
 *      ix_cc_tc_meter_api.c
 *
 * ============================================================================
 */

#define IX_ERROR_FILE_IDENT \
  "$Id: ix_cc_tc_meter_api.c,v 1.10.2.1 2004/03/12 09:56:48 wneubaue Exp $"

#include <math.h>

#include <ix_types.h>
#include <ix_error.h>

#include <ix_ossl.h>
#include <ix_rm.h>
#include <ix_cc.h>
#include <ix_cc_error.h>

#include <bindings.h>

#include "cc/ix_cc_tc_meter.h"
#include "cc/ix_cc_tc_meter_msg_helpers.h"
#include "cc/internal/ix_cc_tc_meter_defs.h"
#include "cc/internal/ix_cc_tc_meter_types.h"




/**
 * This macro changes the rate in kbits/s into rate expressed in
 * bytes/timestamp.
 *
 * @param arg_nRateScale the rate value to be converted
 */
#define _CHANGE_UNIT( arg_nRateScale ) \
  (_SHIFT_LEFT( arg_nRateScale ) / TCM_RATE_UNIT_CHANGE_FACTOR)

/**
 * This macro calculates 2 raised to the power of <code>arg_nPow</code>. The
 * result is floating point double.
 *
 * @param arg_nPow the exponent value of the expression.
 */
#define _SHIFT_LEFT( arg_nPow ) \
  (double)_SHIFT_LEFT_64BIT( 1, arg_nPow )

/**
 * This macro calculates a new Committed Information Rate expressed in kbits/s.
 */
#define _NEW_CIR( arg_nOldCir, arg_nRateScale ) \
    (ix_uint32)((double)(arg_nOldCir) * _CHANGE_UNIT( arg_nRateScale ))


/* _NEW_RATE_SCALE */
#define _NEW_RATE_SCALE( arg_nCir ) \
  (ix_uint8)_ix_cc_tc_meter_calc_rate( arg_nCir )

/* ----------------------------------------------- _ix_cc_tc_meter_calc_rate -- */

PRIVATE ix_int32 
_ix_cc_tc_meter_calc_rate( ix_uint32 arg_nCIR )
{
    double    new_CIR = (double)arg_nCIR / TCM_RATE_UNIT_CHANGE_FACTOR;
    ix_int32  result  = 0;

    while ( (new_CIR - floor( new_CIR )) > (0.01 * new_CIR) )
      {
        new_CIR *= 2; result++;
      }

    return result;
}

/* ------------------------------------------------ ix_cc_tc_meter_add_entry -- */

ix_error
ix_cc_tc_meter_add_entry(ix_uint32                  arg_nInstance,
                         ix_cc_tc_meter_parameters *arg_pParams,
                         void                      *arg_pContext)
  {
    _ix_cc_tc_meter_table_entry * _pT;
    _ix_cc_tc_meter_stats_entry * _pS;

    #if defined (IX_TC_METER_DEBUG)
        printf("ix_cc_tc_meter_add_entry(): checking parameter's sanity.\n");
        printf("ix_cc_tc_meter_add_entry(): checking if context is null.\n");
    #endif

    /* validity check for the arg_pContext */
    IX_ERROR_CHECK_ARG_NULL( 3, arg_pContext );

    #if defined (IX_TC_METER_DEBUG)
        printf("ix_cc_tc_meter_add_entry(): checking arg_nInstance.\n");
    #endif

    /* validity check for arg_nInsance */
    IX_ERROR_CHECK_ARG_MAX( 1, arg_nInstance, _MAX_FLOW_IDS( arg_pContext ), "%u" );

    #if defined (IX_TC_METER_DEBUG)
        printf("ix_cc_tc_meter_add_entry(): checking arg_pParams.\n");
    #endif

    /* validity check for arg_pParams */
    IX_ERROR_CHECK_ARG_NULL( 2, arg_pParams );

    #if defined (IX_TC_METER_DEBUG)
        printf("ix_cc_tc_meter_add_entry(): checking meter's type.\n");
    #endif

    if ((arg_pParams->meter_type_id != IX_CC_TC_METER_SRTCM_TYPE) &&
        (arg_pParams->meter_type_id != IX_CC_TC_METER_TRTCM_TYPE))
    {
        return IX_ERROR_WARNING(
                    IX_CC_TC_METER_ERROR_INVALID_PARAM,
                    ( "ix_cc_tc_meter_add_entry(): unknown type of the meter instance." ));
    }

    #if defined (IX_TC_METER_DEBUG)
        printf("ix_cc_tc_meter_add_entry(): checking that pir < cir.\n");
    #endif

    if ((arg_pParams->meter_type_id == IX_CC_TC_METER_TRTCM_TYPE) &&
        (arg_pParams->meter_type.trtcm.pir < arg_pParams->meter_type.trtcm.cir))
    {
        return IX_ERROR_WARNING(
                    IX_CC_TC_METER_ERROR_INVALID_PARAM,
                    ( "ix_cc_tc_meter_add_entry(): for the given entry PIR is less than CIR." ));
    }


    _pT = _TABLE_ENTRY( arg_nInstance, arg_pContext );


    #if defined (IX_TC_METER_DEBUG)
        printf(" ix_cc_tc_meter_add_entry(): Add new meter entry at 0x%x\n",_pT);
        printf(" ----- before ------  \n");
        show_mem((void*)_pT, sizeof(_ix_cc_tc_meter_table_entry));
    #endif


    /* check if entry is available */
    if (_pT->_valid) 
    {
        return IX_ERROR_WARNING(
                   IX_CC_ERROR_CONFLICTING_ENTRY,
                   ("ix_cc_tc_meter_add_entry(): entry already exists"));
    }

    _pT->_dscp_green   = arg_pParams->greenDscp;
    _pT->_output_green = arg_pParams->greenOutput;

    _pT->_dscp_red     = arg_pParams->redDscp;
    _pT->_output_red   = arg_pParams->redOutput;

    _pT->_dscp_yellow   = arg_pParams->yellowDscp;
    _pT->_output_yellow = arg_pParams->yellowOutput;

    /* initialize SRTCM table entry */

    if (arg_pParams->meter_type_id == IX_CC_TC_METER_SRTCM_TYPE)
    {

        _pT->_tc         = arg_pParams->meter_type.srtcm.cbs;
        _pT->_te_tp      = arg_pParams->meter_type.srtcm.ebs;

        _pT->_cbs        = arg_pParams->meter_type.srtcm.cbs;
        _pT->_ebs_pbs    = arg_pParams->meter_type.srtcm.ebs;

        _pT->_cir_rate   = _NEW_RATE_SCALE(arg_pParams->meter_type.srtcm.cir);
        
        _pT->_cir        = _NEW_CIR(arg_pParams->meter_type.srtcm.cir, _pT->_cir_rate);

        #if defined (IX_TC_METER_DEBUG)
            printf(" CBS      0x%x:  \n",arg_pParams->meter_type.srtcm.cbs);
            printf(" EBS      0x%x:  \n",arg_pParams->meter_type.srtcm.ebs);
            printf(" ---------------------- \n");
            printf(" cir kb/s 0x%x\n",arg_pParams->meter_type.srtcm.cir);
            printf(" CIR      0x%x:  \n",_NEW_CIR(arg_pParams->meter_type.srtcm.cir, _pT->_cir_rate));
            printf(" CIR rate 0x%x:  \n",_NEW_RATE_SCALE(arg_pParams->meter_type.srtcm.cir));
        #endif

    } else {

        _pT->_tc         = arg_pParams->meter_type.trtcm.cbs;
        _pT->_te_tp      = arg_pParams->meter_type.trtcm.pbs;

        _pT->_cbs        = arg_pParams->meter_type.trtcm.cbs;
        _pT->_ebs_pbs    = arg_pParams->meter_type.trtcm.pbs;

        _pT->_cir_rate   = _NEW_RATE_SCALE(arg_pParams->meter_type.trtcm.cir);
        _pT->_pir_rate   = _NEW_RATE_SCALE(arg_pParams->meter_type.trtcm.pir);

        _pT->_cir        = _NEW_CIR(arg_pParams->meter_type.trtcm.cir, _pT->_cir_rate);
        _pT->_pir        = _NEW_CIR(arg_pParams->meter_type.trtcm.pir, _pT->_pir_rate);

        #if defined (IX_TC_METER_DEBUG)
            printf(" CBS      0x%x:  \n",arg_pParams->meter_type.trtcm.cbs);
            printf(" PBS      0x%x:  \n",arg_pParams->meter_type.trtcm.pbs);
            printf(" ---------------------- \n");
            printf(" cir kb/s 0x%x\n",arg_pParams->meter_type.trtcm.cir);
            printf(" CIR      0x%x:  \n",_NEW_CIR(arg_pParams->meter_type.trtcm.cir, _pT->_cir_rate));
            printf(" CIR rate 0x%x:  \n",_NEW_RATE_SCALE(arg_pParams->meter_type.trtcm.cir));
            printf(" ---------------------- \n");
            printf(" PIR kb/s 0x%x\n",arg_pParams->meter_type.trtcm.pir);
            printf(" PIR      0x%x:  \n",_NEW_CIR(arg_pParams->meter_type.trtcm.pir, _pT->_pir_rate));
            printf(" PIR rate 0x%x:  \n",_NEW_RATE_SCALE(arg_pParams->meter_type.trtcm.pir));
        #endif
    }

    _pT->_meter_type    = arg_pParams->meter_type_id;
    _pT->_valid         = TC_METER_ENTRY_VALID;
    _pT->_stats_enabled = _BIT_FLD(arg_pParams->statsEnabled);


    #if defined (IX_TC_METER_DEBUG)
        printf(" ----- after ------  \n");
        show_mem((void*)_pT, sizeof(_ix_cc_tc_meter_table_entry));

        printf(" parameter structure:  \n");
        show_mem((void*)arg_pParams, sizeof(ix_cc_tc_meter_parameters));

        printf(" ---------------------- \n");
        printf(" gDSCP   0x%x:  \n",arg_pParams->greenDscp);
        printf(" yDSCP   0x%x:  \n",arg_pParams->yellowDscp);
        printf(" rDSCP   0x%x:  \n",arg_pParams->redDscp);

    #endif

 
    /* initialize statistics entry */

    _pS = _STATS_ENTRY( arg_nInstance, arg_pContext );

    _pS->_green_pkts   = 0;
    _pS->_green_bytes  = 0;
    _pS->_red_bytes    = 0;
    _pS->_red_pkts     = 0;
    _pS->_yellow_bytes = 0;
    _pS->_yellow_pkts  = 0;

    return IX_SUCCESS;
  }

/* --------------------------------------------- ix_cc_tc_meter_remove_entry -- */

ix_error
ix_cc_tc_meter_remove_entry(ix_uint32  arg_nInstance,
                            void      *arg_pContext)
{
    _ix_cc_tc_meter_table_entry * _pT;
    _ix_cc_tc_meter_stats_entry * _pS;

    /* validity check for the arg_pContext */
    IX_ERROR_CHECK_ARG_NULL( 2, arg_pContext );

    /* validity check for arg_nInsance */
    IX_ERROR_CHECK_ARG_MAX( 1, arg_nInstance,
      _MAX_FLOW_IDS( arg_pContext ), "%u" );

    _pT = _TABLE_ENTRY( arg_nInstance, arg_pContext );
    
    /* check if entry is available */
    if (!_pT->_valid) 
    {
        return IX_ERROR_WARNING(
            IX_CC_ERROR_ENTRY_NOT_FOUND,
            ( "ix_cc_tc_meter_remove_entry(): entry doesn't exists" )
          );
    }

    _pT->_valid        = 0;


    if (_pT->_meter_type == IX_CC_TC_METER_SRTCM_TYPE)
    {

        _pT->_tc         = 0;
        _pT->_te_tp      = 0;

        _pT->_cbs        = 0;
        _pT->_ebs_pbs    = 0;

        _pT->_cir_rate   = 0;
        
        _pT->_cir        = 0xFFFF;

    } else {

        _pT->_tc         = 0;
        _pT->_te_tp      = 0;

        _pT->_cbs        = 0;
        _pT->_ebs_pbs    = 0;

        _pT->_cir_rate   = 0;
        _pT->_pir_rate   = 0;

        _pT->_cir        = 0xFFFF;
        _pT->_pir        = 0xFFFF;
    }


    /* reset lower part of the statistics */
    _pT->_green_bytes  = 0;
    _pT->_green_pkts   = 0;
    _pT->_red_bytes    = 0;
    _pT->_red_pkts     = 0;
    _pT->_yellow_bytes = 0;
    _pT->_yellow_pkts  = 0;

    _pS = _STATS_ENTRY( arg_nInstance, arg_pContext );

    /* reset higher part of the statistics */
    _pS->_red_bytes    = 0;
    _pS->_red_pkts     = 0;
    _pS->_green_bytes  = 0;
    _pS->_green_pkts   = 0;
    _pS->_yellow_bytes = 0;
    _pS->_yellow_pkts  = 0;

    return IX_SUCCESS;
  }

/* --------------------------------------------- ix_cc_tc_meter_update_entry -- */

ix_error
ix_cc_tc_meter_update_entry(ix_uint32                  arg_nInstance,
                            ix_cc_tc_meter_parameters *arg_pParams,
                            void                      *arg_pContext)
{
    _ix_cc_tc_meter_stats_entry * _pS;
    _ix_cc_tc_meter_table_entry * _pT;

    /* validity check for the arg_pContext */
    IX_ERROR_CHECK_ARG_NULL( 3, arg_pContext );

    /* validity check for arg_nInsance */
    IX_ERROR_CHECK_ARG_MAX( 1, arg_nInstance,
      _MAX_FLOW_IDS( arg_pContext ), "%u" );

    /* validity check for arg_pParams */
    IX_ERROR_CHECK_ARG_NULL( 2, arg_pParams );

    _pT = _TABLE_ENTRY( arg_nInstance, arg_pContext );

    if (!_pT->_valid)
    {
        return IX_ERROR_WARNING(
                    IX_CC_ERROR_ENTRY_NOT_FOUND,
                    ( "ix_cc_tc_meter_update_entry(): entry doesn't exists" ));
    }


    _pT->_dscp_green    = arg_pParams->greenDscp;
    _pT->_output_green  = arg_pParams->greenOutput;

    _pT->_dscp_red      = arg_pParams->redDscp;
    _pT->_output_red    = arg_pParams->redOutput;

    _pT->_dscp_yellow   = arg_pParams->yellowDscp;
    _pT->_output_yellow = arg_pParams->yellowOutput;


    if (arg_pParams->meter_type_id == IX_CC_TC_METER_SRTCM_TYPE)
    {

        _pT->_tc         = arg_pParams->meter_type.srtcm.cbs;
        _pT->_te_tp      = arg_pParams->meter_type.srtcm.ebs;

        _pT->_cbs        = arg_pParams->meter_type.srtcm.cbs;
        _pT->_ebs_pbs    = arg_pParams->meter_type.srtcm.ebs;

        _pT->_cir_rate   = _NEW_RATE_SCALE(arg_pParams->meter_type.srtcm.cir);

        _pT->_cir        = _NEW_CIR(arg_pParams->meter_type.srtcm.cir, _pT->_cir_rate);

        

    } else {

        _pT->_tc         = arg_pParams->meter_type.trtcm.cbs;
        _pT->_te_tp      = arg_pParams->meter_type.trtcm.pbs;

        _pT->_cbs        = arg_pParams->meter_type.trtcm.cbs;
        _pT->_ebs_pbs    = arg_pParams->meter_type.trtcm.pbs;

        _pT->_cir_rate   = _NEW_RATE_SCALE(arg_pParams->meter_type.trtcm.cir);
        _pT->_pir_rate   = _NEW_RATE_SCALE(arg_pParams->meter_type.trtcm.pir);

        _pT->_cir        = _NEW_CIR(arg_pParams->meter_type.trtcm.cir, _pT->_cir_rate);
        _pT->_pir        = _NEW_CIR(arg_pParams->meter_type.trtcm.pir, _pT->_pir_rate);
        
    }

    _pT->_meter_type     = arg_pParams->meter_type_id;
    _pT->_stats_enabled  = _BIT_FLD(arg_pParams->statsEnabled);


    /* zero statistics if necessery */
    if ( !_pT->_stats_enabled && arg_pParams->statsEnabled )
    {
        _pT->_green_bytes  = 0;
        _pT->_green_pkts   = 0;
        _pT->_red_bytes    = 0;
        _pT->_red_pkts     = 0;
        _pT->_yellow_pkts  = 0;
        _pT->_yellow_bytes = 0;

        _pS = _STATS_ENTRY( arg_nInstance, arg_pContext );

        _pS->_green_bytes  = 0;
        _pS->_green_pkts   = 0;
        _pS->_red_bytes    = 0;
        _pS->_red_pkts     = 0;
        _pS->_yellow_bytes = 0;
        _pS->_yellow_pkts  = 0;
    }

    return IX_SUCCESS;
}

/* ------------------------------------------------ ix_cc_strcm_get_stats -- */

ix_error
ix_cc_tc_meter_get_statistics(ix_uint32                  arg_nInstance,
                              ix_cc_tc_meter_statistics *arg_pStats,
                              void                      *arg_pContext)
{
    _ix_cc_tc_meter_table_entry * _pT;
    _ix_cc_tc_meter_stats_entry * _pS;

    /* validity check for the arg_pContext */
    IX_ERROR_CHECK_ARG_NULL( 3, arg_pContext );

    /* validity check for arg_nInsance */
    IX_ERROR_CHECK_ARG_MAX( 1, arg_nInstance,
      _MAX_FLOW_IDS( arg_pContext ), "%u" );

    /* validity check for arg_pParams */
    IX_ERROR_CHECK_ARG_NULL( 2, arg_pStats );

    _pT = _TABLE_ENTRY(
      arg_nInstance, arg_pContext );

    #if defined (IX_TC_METER_DEBUG)
    printf("ix_cc_tc_meter_get_statistics: taking from %x\n",_pT );
    #endif

    if (!_pT->_valid) 
    {
        return IX_ERROR_WARNING(
            IX_CC_ERROR_ENTRY_NOT_FOUND,
            ( "ix_cc_tc_meter_get_stats(): entry doesn't exists")
          );
    }

    _pS = _STATS_ENTRY( arg_nInstance, arg_pContext );

    #if defined (IX_TC_METER_DEBUG)
    printf("ix_cc_tc_meter_get_statistics: taking 64 bits from %x\n",_pS );
    #endif

    arg_pStats->greenPacketCount =
      IX_CC_TC_METER_MAKE_UINT64(_pT->_green_pkts , _pS->_green_pkts);
    arg_pStats->greenByteCount   =
      IX_CC_TC_METER_MAKE_UINT64(_pT->_green_bytes, _pS->_green_bytes);
    arg_pStats->redPacketCount   =
      IX_CC_TC_METER_MAKE_UINT64(_pT->_red_pkts , _pS->_red_pkts);
    arg_pStats->redByteCount     =
      IX_CC_TC_METER_MAKE_UINT64(_pT->_red_bytes, _pS->_red_bytes);
    arg_pStats->yellowByteCount  =
      IX_CC_TC_METER_MAKE_UINT64(_pT->_yellow_bytes, _pS->_yellow_bytes);
    arg_pStats->yellowPacketCount =
      IX_CC_TC_METER_MAKE_UINT64(_pT->_yellow_pkts, _pS->_yellow_pkts);

    #if defined (IX_TC_METER_DEBUG)
    printf("green : %d,   %d\n",(ix_uint32)(arg_pStats->greenPacketCount), (ix_uint32)(arg_pStats->greenByteCount));
    printf("yellow: %d,   %d\n",(ix_uint32)(arg_pStats->yellowPacketCount),(ix_uint32)(arg_pStats->yellowByteCount));
    printf("red   : %d,   %d\n",(ix_uint32)(arg_pStats->redPacketCount),   (ix_uint32)(arg_pStats->redByteCount ));
    #endif

    return IX_SUCCESS;
}



/**
 * @brief  This function gets parameters of an entry from the TC Meter table.
 *
 * The core component first checks if an entry, pointed by arg_Instance, is
 * valid. If true, it reads the entry parameters to the arg_pParams argument.
 * Otherwise, if the referred entry is marked empty, the core component
 * reports an error in a callback function.
 *
 *
 * @param  arg_nInstance index of an TC Meter table entry.
 *                       The value must not be higher than TC_METER\TABLE_SRAM_SIZE
 *                       system property.
 *
 * @param  arg_pParams  pointer to the parameters structure which should be
 *                      filled with entry parameters.
 *
 * @param  arg_pContext  pointer to the control block allocated in
 *                       <code>ix_cc_tc_meter_init()</code>
 *
 * @return IX_SUCCESS if successful or a valid ix_error token for failure. The
 *         following error codes might be returned via the token.
 *\n
 *         IX_CC_ERROR_ENTRY_NOT_FOUND - entry does not exist
 *         IX_CC_ERROR_RANGE           - TC Meter instance out of range
 *
 * @see    ix_cc_tc_meter_add_entry()
 */
ix_error
ix_cc_tc_meter_get_parameters(ix_uint32                  arg_nInstance,
                              ix_cc_tc_meter_parameters *arg_pParams,
                              void                      *arg_pContext)
{
    _ix_cc_tc_meter_table_entry * _pT;

    /* validity check for arg_pContext */
    IX_ERROR_CHECK_ARG_NULL( 3, arg_pContext );

    /* validity check for arg_nInstance */
    IX_ERROR_CHECK_ARG_MAX( 1, arg_nInstance,
      _MAX_FLOW_IDS( arg_pContext ), "%u" );

    /* validity check for arg_pParams */
    IX_ERROR_CHECK_ARG_NULL( 2, arg_pParams );

    _pT = _TABLE_ENTRY( arg_nInstance, arg_pContext );

    /* check if entry had been configured */
    if (!_pT->_valid) 
    {
        return IX_ERROR_WARNING(
            IX_CC_ERROR_ENTRY_NOT_FOUND,
            ( "ix_cc_tc_meter_get_parameters(): entry doesn't exists" )
          );
    }

    _pT->_dscp_green    = arg_pParams->greenDscp;
    _pT->_output_green  = arg_pParams->greenOutput;

    _pT->_dscp_red      = arg_pParams->redDscp;
    _pT->_output_red    = arg_pParams->redOutput;

    _pT->_dscp_yellow   = arg_pParams->yellowDscp;
    _pT->_output_yellow = arg_pParams->yellowOutput;


    if (_pT->_meter_type == IX_CC_TC_METER_SRTCM_TYPE)
    {
        arg_pParams->meter_type.srtcm.cbs = _pT->_cbs;
        arg_pParams->meter_type.srtcm.ebs = _pT->_ebs_pbs;
        
        arg_pParams->meter_type.srtcm.cir = (ix_uint32) (_pT->_cir * 
                                            TCM_RATE_UNIT_CHANGE_FACTOR / (2 << _pT->_cir_rate));

    } else {
        arg_pParams->meter_type.trtcm.cbs = _pT->_cbs;
        arg_pParams->meter_type.trtcm.pbs = _pT->_ebs_pbs;
        
        arg_pParams->meter_type.trtcm.cir = (ix_uint32) (_pT->_cir * 
                                            TCM_RATE_UNIT_CHANGE_FACTOR / (2 << _pT->_cir_rate));

        arg_pParams->meter_type.trtcm.pir = (ix_uint32) (_pT->_pir * 
                                            TCM_RATE_UNIT_CHANGE_FACTOR / (2 << _pT->_pir_rate));

    }

    _pT->_meter_type = arg_pParams->meter_type_id;
    arg_pParams->statsEnabled = (ix_uint8)((_pT->_stats_enabled != 0 ) ? 1 : 0);

    return IX_SUCCESS;
}
