#ifndef WRED_H
#define WRED_H

/*******************************************************************************
                             Intel Proprietary

 Copyright (c) 1998-2003 By Intel Corporation.  All rights reserved.  
 No part of this program or publication may be reproduced, transmited,
 transcribed, stored in a retrieval system, or translated into any language
 or computer language in any form or by any means, electronic, mechanical,
 magnetic, optical, chemical, manual, or otherwise, without the prior
 written permission of:
                         Intel Corporation
                         2200 Mission College Blvd.
                         Santa Clara, CA  95052-8119
*******************************************************************************/

/* FILE  wred.h
 *
 * Contains definitions used by a WRED micro-block
 * The following tokens change the WRED behavior:
 *
 * #define WRED_INGRESS_MODE     - WRED defaults to egress mode, this 
 *                                 macro is used to change it to 
 *                                 ingress mode. In egres mode WRED computes
 *                                 queue_id (absolute queue number for a packet)
 *                                 based on output ports whereas in igress mode
 *                                 on fabric ports
 *
 * #define WRED_DISABLE_FIN_SYNC - disables final thread synchronization
 *                                 if the next block is microblock 
 *                                 which carries out thread synchronization
 *                                 at the begining this flag can be used 
 *                                 to improve performance. (avoid double 
 *                                 synchronization one after the other)
 *
 * #define WRED_DRAM_TABLE       - sets WRED table in DRAM (it is important when
 *                                 the load of SRAM controller is too heavy).
 *                                 Note that statistics table is always in SRAM.
 *
 * #define WRED_PACKET_COUNTER_FEATURE  - includes code to gather statistics
 */


/* seed to init random number generator with */
#define        WRED_INIT_RANDOM        (0x78922634)



#ifndef USE_IMPORT_VAR


    /*
     * FIRST_WRED_ME is a system dependent definition.
     * Therefore it should be made either in dl_system.h or as a compilation
     * option. Used during initialization defines which ME send 
     * wrd_next_me_signal to.
     */
    #ifndef    FIRST_WRED_ME
        #error "FIRST_WRED_ME not defined."
    #endif


    /*
     * WRED_TABLE_RAM_BASE is a system dependent definition.
     * Therefore it should be made either in dl_system.h or as a compilation
     * option. It is the base address of WRED table mantained in RAM
     */
    #ifndef    WRED_TABLE_RAM_BASE
        #error    "WRED_TABLE_RAM_BASE not defined."
    #endif


    /*
     * WRED_64BIT_STAT_SRAM_BASE is a system dependent definition.
     * Therefore it should be made either in dl_system.h or as a compilation
     * option. It defines the base address of the table where the most
     * significant long words of wred statistics are held.
     */
    #ifdef WRED_PACKET_COUNTER_FEATURE 
    #ifndef    WRED_64BIT_STAT_SRAM_BASE
        #error "WRED_64BIT_STAT_SRAM_BASE not defined."
    #endif
    #endif

    /*
     * QD_SRAM_BASE is a system dependent definition.
     * Therefore it should be made either in dl_system.h or as a compilation
     * option. It is Base Address for Queue Descriptors
     */
    #ifndef    QD_SRAM_BASE
        #error    "QD_SRAM_BASE not defined."
    #endif



#else

    .import_var WRED_TABLE_RAM_BASE 
    #ifdef WRED_PACKET_COUNTER_FEATURE 
    .import_var WRED_64BIT_STAT_SRAM_BASE
    #endif

#endif



/*
 * QD_LOG_SRAM_SIZE is the size of a queue descriptor expressed in a power of 2.
 * It should be defiend in the includes of a queue manager.
 */
#ifndef    QD_LOG_SRAM_SIZE
    #warning "QD_LOG_SRAM_SIZE (size of Queue Descriptor) not defined."
    #warning "Probably Queue Manager files are not included."
    #warning "Assumes default value: QD_LOG_SRAM_SIZE = 4 (for INGRESS)"
    #define  QD_LOG_SRAM_SIZE  5
#endif



/*
 * WRED_NEXT1 is a system dependent definition.
 * Therefore it should be made either in dl_system.h or as a compilation option.
 */
#ifndef    WRED_NEXT1
    #error "WRED_NEXT1 not defined."
#endif


/*
 * WRED_NEXT2 is a system dependent definition.
 * Therefore it should be made either in dl_system.h or as a compilation option.
 */
#ifndef    WRED_NEXT2
    #error "WRED_NEXT2 not defined."
#endif


/*
 * WRED_NEXT_ME_SIG is a system dependent definition.
 * Therefore it should be made either in dl_system.h or as a compilation option.
 * signal used to synchronize wred crtitical sections on different ME
 */
#ifndef    WRED_NEXT_ME_SIG
    #warning "WRED_NEXT_ME_SIG not defined."
#endif

// -----------------------------------------------------------------------------
#ifndef WRED_CORE_PROBABILITY_CALCULATION
// -----------------------------------------------------------------------------

    /*
     *  WRED table in local memeory
     */
    #define        WRED_LM_COUNT             0
    #define        WRED_LM_G_COUNT           0   /* (WRED_LM_COUNT + 0) */
    #define        WRED_LM_Y_COUNT           1   /* (WRED_LM_COUNT + 1) */
    #define        WRED_LM_R_COUNT           2   /* (WRED_LM_COUNT + 2) */

    #define        WRED_LM_CONST             0
    #define        WRED_LM_G_CONST           0   /* (WRED_LM_CONST + 0) */
    #define        WRED_LM_Y_CONST           1   /* (WRED_LM_CONST + 1) */
    #define        WRED_LM_R_CONST           2   /* (WRED_LM_CONST + 2) */

    #define        WRED_LM_Q_AVG_LEN         3
    #define        WRED_LM_SERV_TIME         3
    #define        WRED_LM_WEIGHT            3
    #define        WRED_LM_STATUS            3

    #define        WRED_LM_MAX_TH            4
    #define        WRED_LM_G_MAX_TH          4   /* (WRED_LM_MAX_TH + 0) */
    #define        WRED_LM_Y_MAX_TH          5   /* (WRED_LM_MAX_TH + 1) */
    #define        WRED_LM_R_MAX_TH          6   /* (WRED_LM_MAX_TH + 2) */

    #define        WRED_LM_MIN_TH            4
    #define        WRED_LM_G_MIN_TH          4   /* (LM_MIN_TH + 0) */
    #define        WRED_LM_Y_MIN_TH          5   /* (LM_MIN_TH + 1) */
    #define        WRED_LM_R_MIN_TH          6   /* (LM_MIN_TH + 2) */


    #define        WRED_LM_Q_CURR_LEN        7

    #define        WRED_LM_PACKETS_CNT       8
    #define        WRED_LM_G_PACKETS_CNT     8   /* (WRED_LM_PACKETS_CNT + 0) */
    #define        WRED_LM_Y_PACKETS_CNT     9   /* (WRED_LM_PACKETS_CNT + 1) */
    #define        WRED_LM_R_PACKETS_CNT     10  /* (WRED_LM_PACKETS_CNT + 2) */


    #define        WRED_LM_BYTES_CNT         12
    #define        WRED_LM_G_BYTES_CNT       12  /* (WRED_LM_BYTES_CNT + 0) */
    #define        WRED_LM_Y_BYTES_CNT       13  /* (WRED_LM_BYTES_CNT + 1) */
    #define        WRED_LM_R_BYTES_CNT       14  /* (WRED_LM_BYTES_CNT + 2) */

    // -------------------------------------------------------------------------
    #ifdef WRED_DRAM_TABLE

        #define    WRED_READ_XFER                 $$wred
        #define    WRED_READ_XFER_G_PACKETS_CNT   $$wred[7]
        #define    WRED_READ_XFER_Y_PACKETS_CNT   $$wred[8]
        #define    WRED_READ_XFER_R_PACKETS_CNT   $$wred[9]
        #define    WRED_READ_XFER_G_BYTES_CNT     $$wred[10]
        #define    WRED_READ_XFER_Y_BYTES_CNT     $$wred[11]
        #define    WRED_READ_XFER_R_BYTES_CNT     $$wred[12]
        #define    WRED_READ_XFER_G_COUNT         $$wred[5]
        #define    WRED_READ_XFER_Y_COUNT         $$wred[6]
        #define    WRED_READ_XFER_R_COUNT         $$wred[6]
        #define    WRED_READ_XFER_G_MAX_TH        $$wred[0]
        #define    WRED_READ_XFER_Y_MAX_TH        $$wred[1]
        #define    WRED_READ_XFER_R_MAX_TH        $$wred[2]
        #define    WRED_READ_XFER_WEIGHT          $$wred[3]
        #define    WRED_READ_XFER_G_CONST         $$wred[3]
        #define    WRED_READ_XFER_Y_CONST         $$wred[4]
        #define    WRED_READ_XFER_R_CONST         $$wred[4]
        #define    WRED_READ_XFER_Q_AVG_LEN       $$wred[5]
        
        #define    WRED_WRITE_XFER                $$wred
        #define    WRED_WRITE_XFER_YR_CONST       $$wred[0]
        #define    WRED_WRITE_XFER_G_COUNT_AVG    $$wred[1]
        #define    WRED_WRITE_XFER_YR_COUNT       $$wred[2]
        #define    WRED_WRITE_XFER_G_PACKETS_CNT  $$wred[3]
        #define    WRED_WRITE_XFER_Y_PACKETS_CNT  $$wred[4]
        #define    WRED_WRITE_XFER_R_PACKETS_CNT  $$wred[5]
        #define    WRED_WRITE_XFER_G_BYTES_CNT    $$wred[6]
        #define    WRED_WRITE_XFER_Y_BYTES_CNT    $$wred[7]
        #define    WRED_WRITE_XFER_R_BYTES_CNT    $$wred[8]

        #ifdef WRED_PACKET_COUNTER_FEATURE
            #define     WRED_WRITE_XFER_UNUSED    $$wred[9]
		    #define     WRED_READ_SIZE          7
            #define     WRED_READ_MAX_SIZE      14
            #define     WRED_WRITE_SIZE         5
            #define     WRED_WRITE_MAX_SIZE     10
        #else    
            #define     WRED_WRITE_XFER_UNUSED    $$wred[3]
            #define     WRED_READ_SIZE          4
            #define     WRED_READ_MAX_SIZE      8
            #define     WRED_WRITE_SIZE         2
            #define     WRED_WRITE_MAX_SIZE     4
        #endif

    // -------------------------------------------------------------------------
    #else

        #define    WRED_READ_XFER                 $wred
        #define    WRED_READ_XFER_G_PACKETS_CNT   $wred[7]
        #define    WRED_READ_XFER_Y_PACKETS_CNT   $wred[8]
        #define    WRED_READ_XFER_R_PACKETS_CNT   $wred[9]
        #define    WRED_READ_XFER_G_BYTES_CNT     $wred[10]
        #define    WRED_READ_XFER_Y_BYTES_CNT     $wred[11]
        #define    WRED_READ_XFER_R_BYTES_CNT     $wred[12]
        #define    WRED_READ_XFER_G_COUNT         $wred[5]
        #define    WRED_READ_XFER_Y_COUNT         $wred[6]
        #define    WRED_READ_XFER_R_COUNT         $wred[6]
        #define    WRED_READ_XFER_G_MAX_TH        $wred[0]
        #define    WRED_READ_XFER_Y_MAX_TH        $wred[1]
        #define    WRED_READ_XFER_R_MAX_TH        $wred[2]
        #define    WRED_READ_XFER_WEIGHT          $wred[3]
        #define    WRED_READ_XFER_G_CONST         $wred[3]
        #define    WRED_READ_XFER_Y_CONST         $wred[4]
        #define    WRED_READ_XFER_R_CONST         $wred[4]
        #define    WRED_READ_XFER_Q_AVG_LEN       $wred[5]

        #define    WRED_WRITE_XFER                $wred
        #define    WRED_WRITE_XFER_G_COUNT_AVG    $wred[0]
        #define    WRED_WRITE_XFER_YR_COUNT       $wred[1]
        #define    WRED_WRITE_XFER_G_PACKETS_CNT  $wred[2]
        #define    WRED_WRITE_XFER_Y_PACKETS_CNT  $wred[3]
        #define    WRED_WRITE_XFER_R_PACKETS_CNT  $wred[4]
        #define    WRED_WRITE_XFER_G_BYTES_CNT    $wred[5]
        #define    WRED_WRITE_XFER_Y_BYTES_CNT    $wred[6]
        #define    WRED_WRITE_XFER_R_BYTES_CNT    $wred[7]


        #ifdef WRED_PACKET_COUNTER_FEATURE
            /* 12 words by indirect instruction */
            #define     WRED_READ_SIZE          0x1C
            #define     WRED_READ_MAX_SIZE      13
            #define     WRED_WRITE_SIZE         8
            #define     WRED_WRITE_MAX_SIZE     8
        #else    
            #define     WRED_READ_SIZE          7
            #define     WRED_READ_MAX_SIZE      7
            #define     WRED_WRITE_SIZE         2
            #define     WRED_WRITE_MAX_SIZE     2
        #endif

    #endif
    // -------------------------------------------------------------------------



// -----------------------------------------------------------------------------
#else  // WRED_CORE_PROBABILITY_CALCULATION
// -----------------------------------------------------------------------------

    #define        WRED_LM_PROB              0
    #define        WRED_LM_STATUS            0
    #define        WRED_LM_COUNT             1
    #define        WRED_LM_PACKETS_CNT       2
    #define        WRED_LM_BYTES_CNT         3

     // -------------------------------------------------------------------------
    #ifdef WRED_DRAM_TABLE

        #define        WRED_READ_XFER                 $$wred
        #define        WRED_READ_XFER_PROB            $$wred[0]
        #define        WRED_READ_XFER_COUNT           $$wred[1]
        #define        WRED_READ_XFER_PACKETS_CNT     $$wred[2]
        #define        WRED_READ_XFER_BYTES_CNT       $$wred[3]

        #define        WRED_WRITE_XFER                $$wred
        #define        WRED_WRITE_XFER_PROB           $$wred[0]
        #define        WRED_WRITE_XFER_COUNT          $$wred[1]
        #define        WRED_WRITE_XFER_PACKETS_CNT    $$wred[2]
        #define        WRED_WRITE_XFER_BYTES_CNT      $$wred[3]

        #ifdef WRED_PACKET_COUNTER_FEATURE
            #define    WRED_READ_SIZE         2
            #define    WRED_READ_MAX_SIZE     4
            #define    WRED_WRITE_SIZE        2
            #define    WRED_WRITE_MAX_SIZE    4
        #else    
            #define    WRED_READ_SIZE         1
            #define    WRED_READ_MAX_SIZE     2
            #define    WRED_WRITE_SIZE        1
            #define    WRED_WRITE_MAX_SIZE    2
        #endif

    // -------------------------------------------------------------------------
    #else

        #define        WRED_READ_XFER                 $wred
        #define        WRED_READ_XFER_PROB            $wred[0]
        #define        WRED_READ_XFER_COUNT           $wred[1]
        #define        WRED_READ_XFER_PACKETS_CNT     $wred[2]
        #define        WRED_READ_XFER_BYTES_CNT       $wred[3]

        #define        WRED_WRITE_XFER                $wred
        #define        WRED_WRITE_XFER_COUNT          $wred[0]
        #define        WRED_WRITE_XFER_PACKETS_CNT    $wred[1]
        #define        WRED_WRITE_XFER_BYTES_CNT      $wred[2]

        #ifdef WRED_PACKET_COUNTER_FEATURE
            #define    WRED_READ_SIZE         4
            #define    WRED_READ_MAX_SIZE     4
            #define    WRED_WRITE_SIZE        3
            #define    WRED_WRITE_MAX_SIZE    3
        #else    
            #define    WRED_READ_SIZE         2
            #define    WRED_READ_MAX_SIZE     2
            #define    WRED_WRITE_SIZE        1
            #define    WRED_WRITE_MAX_SIZE    1
        #endif

    #endif
    // -------------------------------------------------------------------------



// -----------------------------------------------------------------------------
#endif // WRED_CORE_PROBABILITY_CALCULATION
// -----------------------------------------------------------------------------

#define        WRED_LM_REF_COUNT           15

/*
 * a value written into the cam_state bits of each cam_entry
 */
#define        WRED_CAM_STATE_VALUE        0x0F



#endif /*  WRED_H  */
