#ifndef WRED_UC
#define WRED_UC

/*******************************************************************************
                             Intel Proprietary

 Copyright (c) 1998-2003 By Intel Corporation.  All rights reserved.
 No part of this program or publication may be reproduced, transmited,
 transcribed, stored in a retrieval system, or translated into any language
 or computer language in any form or by any means, electronic, mechanical,
 magnetic, optical, chemical, manual, or otherwise, without the prior
 written permission of:
                         Intel Corporation
                         2200 Mission College Blvd.
                         Santa Clara, CA  95052-8119
*******************************************************************************/

// FILE  wred.uc
//   this block implements WRED algorithm in conformity with S.Floyd and
//   V.Jacobson 'Random Early Detection Gateways for Congestion Avoidance'.

// Exported APIs:
//
//   wred()



//------------------------------------------------------------------------------
// Included files:
#include <wred.h>
#include <wred_init.uc>
#include <wred_utils.uc>
#include <wred_defs.h>

//------------------------------------------------------------------------------
// Global signals:

.sig visible volatile wred_next_me_signal
#ifdef WRED_NEXT_ME_SIG
.addr wred_next_me_signal WRED_NEXT_ME_SIG
#endif


.sig volatile global wred_next_thread_signal
#ifdef WRED_NEXT_THREAD_SIG
.addr wred_next_thread_signal WRED_NEXT_THREAD_SIG
#else
#define    WRED_NEXT_THREAD_SIG &wred_next_thread_signal
#endif

//------------------------------------------------------------------------------
// Global variables:
.reg queue_base



//------------------------------------------------------------------------------
// #macro wred()

// wred
//
// Description:
//   main macro performing the thread synchronization and
//   data preparation for the wred algorithm. For the sake of
//   readability part of the code pertinent to the average queue
//   calculation / estimation as well as the drop decision making has
//   been moved to 'wred_util.uc' file.
//
// Dispatch loop inputs:
//    dl_class_id    - the identification of a queue to be used for the current
//                     packet
//    dl_output_port - the number of the port on which packet is to be
//                     transmitted in a given blade
//    dl_fabric_port -
//    dl_color_id    - the color of the packet.
//                     NOTE: this variable should be one of the values:
//                     GREEN_PACKET, YELLOW_PACKET, RED_PACKET.
//    dl_packet_size - number of bytes in the currently processed packet,
//                     used for statistics
//    dl_next_block  - the identification of the next block which the packet is
//                     directed to.
//                     BID_WRED value indicates that the current packet
//                     should be processed by a wred micro block.
//
// Dispatch loop outputs:
//   dl_next_block   - the identification of the next block which the packet is
//                     directed to. If the packet has been handled by a wred
//                     the variable is set to WRED_NEXT1 or WRED_NEXT2 depending
//                     whether or not the packet has been destined to be dropped
//                     If the packet has not been handled by a wred entity
//                     the variable is left unchanged.
//
// The wred block uses a table stored in SRAM/DRAM which is implemented as
// a simple array indexed with a class_id + (port_id << WRED_MAX_QUEUE).
// A table entry contains WRED parameters which are illustrated
// in the figure below.
//
// -----------------------------------------------------------------------------
//  WRED TABLE when average queue length is calculated by microcode
// -----------------------------------------------------------------------------
//
//  SRAM/DRAM  picture:
//
//    32               16                 0
//    +-----------------+-----------------+
//  0 |   green max th  |  green min th   |
//    +-----------------+-----------------+
//  1 |   yellow max th |  yellow min th  |
//    +-----------------+-----------------+
//  2 |   red max th    |  red min th     |
//    +-----------------+--------+--------+
//  3 |   green const   |srv time| sweight|
//    +-----------------+--------+--------+
//  4 |   yellow const  |    red const    |
//    +-----------------+-----------------+
//  5 |   green count   |  average length |
//    +-----------------+-----------------+
//  6 |   yellow count  |    red count    |
//    +-----------------+-----------------+
//  7 |           green packets           |
//    +-----------------------------------+
//  8 |           yellow packets          |
//    +-----------------------------------+
//  9 |           red packets             |
//    +-----------------------------------+
// 10 |           green bytes             |
//    +-----------------------------------+
// 11 |           yellow bytes            |
//    +-----------------------------------+
// 12 |           red bytes               |
//    +-----------------------------------+
// 13 |              unused               |
//    +-----------------------------------+
// 14 |              unused               |
//    +-----------------------------------+
// 15 |              unused               |
//    +-----------------------------------+
//
//  sweight field:
//
//       bit 7  6  5              bit0
//           +------+----------------+
//           |status|     weight     |
//           +------+----------------+
//            | | |
//            | | +-> bit 5 - unused
//            | +---> bit 6 - unused
//            +-----> bit 7 - enable statistics gathering
//
//
//
//  in order to optimize algorithm and because we have to keep additional
//  data about the state of a queue the data arrangement in LM look
//  a bit different:
//
//    32               16       8         0
//    +-----------------+-----------------+
//  0 |   green const   |   green count   |
//    +-----------------+-----------------+
//  1 |   yellow const  |   yellow count  |
//    +-----------------+-----------------+
//  2 |   red const     |   red count     |
//    +-----------------+--------+--------+
//  3 |   avg_len       |srv time| sweight|
//    +-----------------+--------+--------+
//  4 |   green max th  |  green_min_th   |
//    +-----------------+-----------------+
//  5 |   yellow max th |  yellow_min_th  |
//    +-----------------+-----------------+
//  6 |   red max th    |  red_min_th     |
//    +-----------------+-----------------+
//  7 |   curr_length   |        -        |
//    +-----------------+-----------------+
//  8 |           green packets           |
//    +-----------------------------------+
//  9 |           yellow packets          |
//    +-----------------------------------+
// 10 |           red packets             |
//    +-----------------------------------+
// 11 |                 -                 |
//    +-----------------------------------+
// 12 |           green bytes             |
//    +-----------------------------------+
// 13 |           yellow bytes            |
//    +-----------------------------------+
// 14 |           red bytes               |
//    +-----------------------------------+
// 15 |              ref_count            |
//    +-----------------------------------+
//
//    ref count - field which contains the number of threads using
//              the same WRED entry in local memory.
//
//
//   This wred implementation assumes that queue descriptor has the following
//  format:
//
//   |                 ...                 |
//   +-------------------------------------+
// 2 |                 ...                 |
//   +---------+---------------------------+
// 3 | 8 bits  | # of packets in the queue |
//   +---------+---------------------------+
// 4 |last queue idle timestamp  low 32bits|
//   +-------------------------------------+
// 5 |last queue idle timestamp high 32bits|
//   +-------------------------------------+
//   |                 ...                 |
//
//
//
// -----------------------------------------------------------------------------
//  WRED TABLE when average queue length is calculated by CC
// -----------------------------------------------------------------------------
//
//  SRAM/DRAM picture:
//  
//
//    32               16                 0
//    +---------------+-+-----------------+
//  0 |   reserved    |S| green drop prob |
//    +---------------+-+-----------------+
//  1 |            green count            |
//    +-----------------------------------+
//  2 |            green packets          |
//    +-----------------------------------+
//  3 |            green bytes            |
//    +---------------+-+-----------------+
//  4 |   reserved    |S| yellow drop prob|
//    +---------------+-+-----------------+
//  5 |           yellow count            |
//    +-----------------------------------+
//  6 |           yellow packets          |
//    +-----------------------------------+
//  7 |           yellow bytes            |
//    +---------------+-+-----------------+
//  8 |   reserved    |S|   red drop prob |
//    +---------------+-+-----------------+
//  9 |           red count               |
//    +-----------------------------------+
// 10 |           red packets             |
//    +-----------------------------------+
// 11 |           red bytes               |
//    +-----------------------------------+
// 12 |              unused               |
//    +-----------------------------------+
// 13 |              unused               |
//    +-----------------------------------+
// 14 |              unused               |
//    +-----------------------------------+
// 15 |              unused               |
//    +-----------------------------------+
//
//  fields description:
//
//       S  +-> bit denotes that statistics gathering is enabled
//
//   reserved - means that microblocks overwrite those bits in upredictable
//              manner
//   unused   - means not used and not changed by microblock.
//
//
//
//  in order to optimize algorithm the data arrangement in LM looks
//  a bit different:
//
//
//    32               16       8         0
//    +---------------+-+-----------------+
//  0 |   reserved    |S| drop probability|
//    +---------------+-+-----------------+
//  1 |              count                |
//    +-----------------------------------+
//  2 |             packets               |
//    +-----------------------------------+
//  3 |              bytes                |
//    +-----------------------------------+
//  4 |             unused                |
//    +-----------------------------------+
//  5 |             unused                |
//    +-----------------------------------+
//  6 |             unused                |
//    +-----------------------------------+
//  7 |             unused                |
//    +-----------------------------------+
//  8 |             unused                |
//    +-----------------------------------+
//  9 |             unused                |
//    +-----------------------------------+
// 10 |             unused                |
//    +-----------------------------------+
// 11 |             unused                |
//    +-----------------------------------+
// 12 |             unused                |
//    +-----------------------------------+
// 13 |             unused                |
//    +-----------------------------------+
// 14 |             unused                |
//    +-----------------------------------+
// 15 |             ref_count             |
//    +-----------------------------------+
//
//    ref count - field which contains the number of threads using
//              the same WRED entry in local memory.
//
//
#macro wred()
.begin

     ;----------------------------------------------------------
     ;    signal declarations
     ;----------------------------------------------------------
    .sig   wred_ram_read_dn1  // denotes the complishing of wred entry reading
    .sig   wred_ram_read_dn2  // denotes the complishing of queue entry reading
    .sig   wred_ram_write_dn  // denotes the complishing of wred entry writing

    // registers responible for holding the address of the wred's data in RAM
    .reg wrd_offset wrd_base

    // registers indicate the time of packet arrivals.
    // they are initialized globally by the thread 0 and used by the others
    // (thereby the same value is taken to process all 8 packets)
    .reg @current_timestamp_l @current_timestamp_h

     ;----------------------------------------------------------
     ; wred specific and dispatch loop variables
     ;----------------------------------------------------------
     .reg queue_id  // abolute number of a queue
                    // queue_id = (output_port_id << WRED_MAX_QUEUES) + class_id
     .reg color_id  // color of the current packet
     .reg color_offset // color id multiplied by 4 
     .reg class_id   // relative queue number for a port
     .reg port_id    // output port for a packet

wred_path#:

;-------------------------------------------------------------------------------
; CAM CLEAR STAGE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    // get dispatch loop variables
    // compiler should move following macros into free time slots
    // of the first branch

    #ifdef WRED_INGRESS_MODE
        dl_meta_get_fabric_port(port_id)
    #else
        dl_meta_get_output_port(port_id)
    #endif
    dl_meta_get_class_id(class_id)


    br=ctx[0,cam_clear_th0#]

        ctx_arb[wred_next_thread_signal], defer[2], br[ram_read_stage#]
        // queue_id = port_id*max_queue + class_id
        alu[queue_id, --,B, port_id, <<WRED_MAX_QUEUE]
        alu[queue_id, class_id, +, queue_id]



    cam_clear_th0#:

        // Wait for inter-ME signal and do not allow the other threads to work
           ctx_arb[wred_next_me_signal, wred_next_thread_signal], and, defer[2]
        // queue_id = port_id*max_queue + class_id
        alu[queue_id, --,B, port_id, <<WRED_MAX_QUEUE]
        alu[queue_id, class_id, +, queue_id]



        // Clear the CAM & reset LRU just once at the beginning but after
        // synchronization to prevent the interference with other micro blocks
        cam_clear



ram_read_stage#:
;-------------------------------------------------------------------------------
; RAM READ STAGE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    // send WRED_NEXT_THREAD_SIGNAL to the next thread
    local_csr_wr[SAME_ME_SIGNAL, (0x80 | ((&wred_next_thread_signal) <<3))]
    // compare dl_next_block with BID_WRED
    // to check block ID. If doesn't match ours, execute the bypass path.
    alu[--,dl_next_block,-,BID_WRED]
    // If dl_next_block doesn't match ours, execute the bypass path.
    bne[bypass_path#]

    // initialize variables responible for holding the address to
    // the wred's data in RAM
    immed32(wrd_base, WRED_TABLE_RAM_BASE)
    alu[wrd_offset,-- , B, queue_id, <<WRED_ENTRY_LOG_RAM_SIZE]

    #ifdef WRED_DIFFSERV_DEBUG
        alu[@debug_came_in, @debug_came_in, +, 1]
    #endif


    xbuf_alloc(WRED_READ_XFER, WRED_READ_MAX_SIZE, read)

    // --------------------------------------------
    #ifndef WRED_CORE_PROBABILITY_CALCULATION
    xbuf_alloc($queue, 3, read)
    #endif
    // --------------------------------------------

    .begin
    .reg cam_result
    .reg cam_entry_number

        alu[wrd_base, wrd_base, +,  wrd_offset]

        // --------------------------------------------
        #ifdef WRED_CORE_PROBABILITY_CALCULATION
            dl_meta_get_color(color_id)
            alu[color_offset, --, B, color_id, <<4]
            alu[wrd_base, wrd_base, +, color_offset]
        #endif
        // --------------------------------------------
                    
        // perform CAM lookup and load the base address
        // to wred and queue data in LM
        cam_lookup[cam_result, wrd_base], lm_addr0[0]

        .begin
        .reg lm_addr
            // cam entry number is used to save new queue_id to CAM
            alu_shf[cam_entry_number, 0xF, AND, cam_result, >>3]

            // --------------------------------------------
            #ifndef WRED_CORE_PROBABILITY_CALCULATION

            // Calculation of the address pointing to wred entry in local memory
            // This address is shifted 0, 4 or 8 bytes depending on the color_id
            // of the current packet - it will be used later on to invoke
            // algorithmic drop macro.
            dl_meta_get_color(color_id)
            alu[color_offset, --, B, color_id, <<2]

            local_csr_rd[active_lm_addr_0]
            immed[lm_addr, 0]

            alu[lm_addr, lm_addr, +, color_offset]
            local_csr_wr[active_lm_addr_1, lm_addr]

            // check if lookup's missed
            br_bclr[cam_result, 7, cam_miss#]

            // --------------------------------------------
            #else

            // check if lookup's missed
            br_bclr[cam_result, 7, cam_miss#]

            #endif
            // --------------------------------------------



        .end



    cam_hit#:

        ctx_arb[wred_next_thread_signal], defer[1],

        // --------------------------------------
        #ifndef WRED_CORE_PROBABILITY_CALCULATION
            br[average_calculation#]
        #else
            br[dropping_stage#]
        #endif
        // --------------------------------------

        ; defer 1: LRU[lru_entry].ref_count++
        alu[*l$index0[WRED_LM_REF_COUNT], 1, +, *l$index0[WRED_LM_REF_COUNT]]




    cam_miss#:


        // --------------------------------------------------------------------
        #ifdef WRED_DRAM_TABLE
        // --------------------------------------------------------------------

                dram[read,
                     WRED_READ_XFER[0],
                     wrd_base, 0, WRED_READ_SIZE],  sig_done[wred_ram_read_dn1]

        #else


            // ----------------------------------------------------------------
            #if (WRED_READ_MAX_SIZE > 8)

                // read thirteen words from sram (c+1)
                alu_shf[--, --, B, WRED_READ_SIZE, <<21]   
                sram[read,
                     WRED_READ_XFER[0],
                     wrd_base, 0,
                     max_/**/WRED_READ_MAX_SIZE/**/],
                                     indirect_ref, sig_done[wred_ram_read_dn1]

            #else

                sram[read,
                     WRED_READ_XFER[0],
                     wrd_base, 0, WRED_READ_SIZE],  sig_done[wred_ram_read_dn1]

            #endif
            // ----------------------------------------------------------------



        // --------------------------------------------------------------------
        #endif
        // --------------------------------------------------------------------



        // --------------------------------------------------------------------
        #ifndef WRED_CORE_PROBABILITY_CALCULATION
        // --------------------------------------------------------------------

            // registers responible for holding
            // the address of the queue's data in SRAM
            .begin
            .reg queue_offset
                alu[queue_offset,-- , B, queue_id, <<QD_LOG_SRAM_SIZE]
                sram[read, $queue0, queue_base, queue_offset, 3], \
                                                   sig_done[wred_ram_read_dn2]
            .end //queue_base ;queue_offset

            #define WRED_WAIT_SIGNALS wred_ram_read_dn1, wred_ram_read_dn2

        #else

            #define WRED_WAIT_SIGNALS wred_ram_read_dn1

        // ---------------------------------------------------------------------
        #endif
        // ---------------------------------------------------------------------


        ctx_arb[wred_next_thread_signal, WRED_WAIT_SIGNALS], and, defer[2]

        #undef WRED_WAIT_SIGNALS

        // if lookup misses cam entry number contains LRU, and we put our
        // read data to CAM in this position
        cam_write[cam_entry_number, wrd_base, WRED_CAM_STATE_VALUE]
        // (LM[lru_entry].ref_count) = 1
        alu[*l$index0[WRED_LM_REF_COUNT], --, B, 0x01]


    .end


    // -------------------------------------------------------------------------
    #ifndef WRED_CORE_PROBABILITY_CALCULATION
    // -------------------------------------------------------------------------

    ;copy green_count from transfer register to local memory
    ld_field_w_clr[*l$index0[WRED_LM_G_COUNT], 0011, WRED_READ_XFER_G_COUNT, >>16]
    ;copy yellow_count from transfer register to local memory
    ld_field_w_clr[*l$index0[WRED_LM_Y_COUNT], 0011, WRED_READ_XFER_Y_COUNT, >>16]
    ;copy red_count from transfer register to local memory
    ld_field_w_clr[*l$index0[WRED_LM_R_COUNT], 0011, WRED_READ_XFER_R_COUNT]

    ;copy green_const from transfer register to local memory
    ld_field[*l$index0[WRED_LM_G_CONST],  1100, WRED_READ_XFER_G_CONST]
    ;copy yellow_const from transfer register to local memory
    ld_field[*l$index0[WRED_LM_Y_CONST],  1100, WRED_READ_XFER_Y_CONST]
    ;copy red_const from transfer register to local memory
    ld_field[*l$index0[WRED_LM_R_CONST],  1100, WRED_READ_XFER_R_CONST, <<16]


    ; green_max_th  / green_min_th
    alu[*l$index0[WRED_LM_G_MAX_TH], --, B, WRED_READ_XFER_G_MAX_TH]
    ; yellow_max_th / yellow_min_th
    alu[*l$index0[WRED_LM_Y_MAX_TH], --, B, WRED_READ_XFER_Y_MAX_TH]
    ; red_max_th    / red_min_th
    alu[*l$index0[WRED_LM_R_MAX_TH], --, B, WRED_READ_XFER_R_MAX_TH]

    ;load weight, status, and service_time
    ld_field_w_clr[*l$index0[WRED_LM_WEIGHT],     0011, WRED_READ_XFER_WEIGHT]

    ;load average length
    ld_field[*l$index0[WRED_LM_Q_AVG_LEN], 1100, WRED_READ_XFER_Q_AVG_LEN, <<16]

    #ifdef WRED_PACKET_COUNTER_FEATURE
        alu[*l$index0[WRED_LM_G_PACKETS_CNT],  --, B, WRED_READ_XFER_G_PACKETS_CNT]
        alu[*l$index0[WRED_LM_Y_PACKETS_CNT],  --, B, WRED_READ_XFER_Y_PACKETS_CNT]
        alu[*l$index0[WRED_LM_R_PACKETS_CNT],  --, B, WRED_READ_XFER_R_PACKETS_CNT]
        alu[*l$index0[WRED_LM_G_BYTES_CNT],    --, B, WRED_READ_XFER_G_BYTES_CNT]
        alu[*l$index0[WRED_LM_Y_BYTES_CNT],    --, B, WRED_READ_XFER_Y_BYTES_CNT]
        alu[*l$index0[WRED_LM_R_BYTES_CNT],    --, B, WRED_READ_XFER_R_BYTES_CNT]
    #endif

    ;load queue current length
    ld_field_w_clr[*l$index0[WRED_LM_Q_CURR_LEN], 1100, $queue0, <<16], load_cc

    // -------------------------------------------------------------------------
    #else
    // -------------------------------------------------------------------------

    ;copy green drop prob from transfer register to local memory
    ld_field_w_clr[*l$index0[WRED_LM_PROB],  1111, WRED_READ_XFER_PROB]

    ;copy red_count from transfer register to local memory
    ld_field_w_clr[*l$index0[WRED_LM_COUNT], 0011, WRED_READ_XFER_COUNT]


    #ifdef WRED_PACKET_COUNTER_FEATURE
        alu[*l$index0[WRED_LM_PACKETS_CNT],  --, B, WRED_READ_XFER_PACKETS_CNT]
        alu[*l$index0[WRED_LM_BYTES_CNT],    --, B, WRED_READ_XFER_BYTES_CNT]
    #endif

    // -------------------------------------------------------------------------
    #endif
    // -------------------------------------------------------------------------
    






;------------------------------------------------------------------------------
; WRED STAGE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


// --------------------------------------
#ifndef WRED_CORE_PROBABILITY_CALCULATION
// --------------------------------------

    .begin
    .reg  avg_len curr_len random

        ; if (curr_len !=0) goto average_calculation#:
        bne[average_calculation#]; let's compiler fill these defers.

        ; if queue is empty perform average queue estimation
        ; and put the packet into a queue
        _wred_avg_len_estimate_and_put(avg_len,  \
                                       $queue1,  \
                                       $queue2,  \
                                       wred_stage_end#)

        xbuf_free(WRED_READ_XFER)
        xbuf_free($queue)


    average_calculation#:


        ld_field_w_clr[avg_len, 0011, *l$index0[WRED_LM_Q_AVG_LEN], >>16]
        ld_field_w_clr[curr_len, 0110, *l$index0[WRED_LM_Q_CURR_LEN], >>8]

        _wred_avg_len_calculation(random,
                                  avg_len,
                                  curr_len,
                                  *l$index0[WRED_LM_WEIGHT])


    dropping_stage#:

        _wred_algorithmic_drop(avg_len,
                               queue_id,
                               color_offset,
                               random)


    wred_stage_end#:

        ld_field[*l$index0[WRED_LM_Q_AVG_LEN], 1100, avg_len, <<16]

    .end


// --------------------------------------
#else 

    .begin
    .reg  random

    dropping_stage#:

        local_csr_rd[PSEUDO_RANDOM_NUMBER]
        immed   [random,0x0]  //a CSR read is always followed by "immed" 
        immed_w1[random,0x0]

        _wred_algorithmic_drop(0,
                               queue_id,
                               color_offset,
                               random)

    .end


// --------------------------------------
#endif
// --------------------------------------


;------------------------------------------------------------------------------
; RAM WRITE STAGE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

ram_write_stage#:

    .begin
    .reg temp

    br=ctx[7, thread_7#], defer[1]
    // signal_next_ctx[&wred_next_thread_signal]
    local_csr_wr[SAME_ME_SIGNAL, (0x80 | (&wred_next_thread_signal <<3))]

        //    if (--(LM[lru_entry].ref_count)) != 0
        //        goto wred_final_synchronization#
        alu[*l$index0[WRED_LM_REF_COUNT], *l$index0[WRED_LM_REF_COUNT], -, 1]
        bne[wred_final_synchronization#]

        xbuf_alloc(WRED_WRITE_XFER, WRED_WRITE_MAX_SIZE, write)

        // --------------------------------------
        #ifndef WRED_CORE_PROBABILITY_CALCULATION
            ld_field_w_clr[temp, 1100, *l$index0[WRED_LM_G_COUNT], <<16]
            ld_field[temp, 0011, *l$index0[WRED_LM_Q_AVG_LEN], >>16]
        #else
			#ifdef WRED_DRAM_TABLE
            alu[WRED_WRITE_XFER_PROB,  --, B, *l$index0[WRED_LM_PROB]]
            #endif
            alu[WRED_WRITE_XFER_COUNT, --, B, *l$index0[WRED_LM_COUNT]]
        #endif
        // --------------------------------------


    thread7_write_back#:

        // ----------------------------------------------------------
        #ifndef WRED_CORE_PROBABILITY_CALCULATION

            alu[WRED_WRITE_XFER_G_COUNT_AVG, --, B, temp]
			
			#ifdef WRED_DRAM_TABLE

               ;copy yellow and redconst from local memory to transfer register
               ld_field[temp, 1100, *l$index0[WRED_LM_Y_CONST]]
               ld_field[temp, 0011, *l$index0[WRED_LM_R_CONST], >>16]
               alu[WRED_WRITE_XFER_YR_CONST, --, B, temp]
               .set WRED_WRITE_XFER_UNUSED

            #endif

            ld_field[temp, 1100, *l$index0[WRED_LM_Y_COUNT], <<16]
            ld_field[temp, 0011, *l$index0[WRED_LM_R_COUNT]]
            alu[WRED_WRITE_XFER_YR_COUNT, --, B, temp]

            #ifdef WRED_PACKET_COUNTER_FEATURE
                alu[WRED_WRITE_XFER_G_PACKETS_CNT, --, B, \
                    *l$index0[WRED_LM_G_PACKETS_CNT]]
                alu[WRED_WRITE_XFER_Y_PACKETS_CNT, --, B, \
                    *l$index0[WRED_LM_Y_PACKETS_CNT]]
                alu[WRED_WRITE_XFER_R_PACKETS_CNT, --, B, \
                    *l$index0[WRED_LM_R_PACKETS_CNT]]
                alu[WRED_WRITE_XFER_G_BYTES_CNT, --, B,   \
                    *l$index0[WRED_LM_G_BYTES_CNT]]
                alu[WRED_WRITE_XFER_Y_BYTES_CNT, --, B,   \
                    *l$index0[WRED_LM_Y_BYTES_CNT]]
                alu[WRED_WRITE_XFER_R_BYTES_CNT, --, B,   \
                    *l$index0[WRED_LM_R_BYTES_CNT]]
            #endif

        // ----------------------------------------------------------
        #else

            #ifdef WRED_PACKET_COUNTER_FEATURE
                alu[WRED_WRITE_XFER_PACKETS_CNT, --, B,  \
                    *l$index0[WRED_LM_PACKETS_CNT]]
                alu[WRED_WRITE_XFER_BYTES_CNT, --, B,    \
                    *l$index0[WRED_LM_BYTES_CNT]]
            #endif

        #endif  
        // ----------------------------------------------------------

        #ifndef WRED_CORE_PROBABILITY_CALCULATION
			#ifdef WRED_DRAM_TABLE 
            #define WRED_WRITE_OFFSET 16
			#else
            #define WRED_WRITE_OFFSET 20
			#endif
		#else
		    #ifdef WRED_DRAM_TABLE
            #define WRED_WRITE_OFFSET 0
			#else
            #define WRED_WRITE_OFFSET 4
			#endif
		#endif

        // ------------------------------------------- write entry back to sram
        #ifdef WRED_DRAM_TABLE

            dram[write, 
        #else

            sram[write,
		#endif
                 WRED_WRITE_XFER[0],
                 wrd_base, WRED_WRITE_OFFSET, WRED_WRITE_SIZE],
                 sig_done[wred_ram_write_dn]
        // --------------------------------------------------------------------

        #undef WRED_WRITE_OFFSET


        xbuf_free(WRED_WRITE_XFER)

        #ifndef WRED_DISABLE_FIN_SYNC
             ctx_arb[wred_ram_write_dn, wred_next_thread_signal], br[wred_end#]
        #else
             ctx_arb[wred_ram_write_dn], br[wred_final_synchronization#]
        #endif


    thread_7#:

        // Signal the thread 0 on DL_NEXT_ME with wred_next_me_signal
        // WARNING: Beacause this signal should get to the next me after ram
        //          write in this thread it must be at least 13 cycles from here
        //          to the ram read instruction on next me.
        //            Otherwise we will have problem with inter-ME signaling -
        //            signal will be received by next ME before our ram write
        //          operation.
        // signal_me[DL_NEXT_ME, &wred_next_me_signal]
        cap[fast_wr,((&wred_next_me_signal)| (DL_NEXT_ME <<7)), INTERTHREAD_SIG]

          // --------------------------------------
        #ifndef WRED_CORE_PROBABILITY_CALCULATION

            // thread 7 always have to copy data to LM
            br[thread7_write_back#], defer[2]
            ld_field_w_clr[temp, 1100, *l$index0[WRED_LM_G_COUNT], <<16]
            ld_field[temp, 0011, *l$index0[WRED_LM_Q_AVG_LEN], >>16]

        #else
            // thread 7 always have to copy data to LM

            alu[WRED_WRITE_XFER_COUNT, --, B, *l$index0[WRED_LM_COUNT]]
			#ifdef WRED_DRAM_TABLE
            alu[WRED_WRITE_XFER_PROB,  --, B, *l$index0[WRED_LM_PROB]]
            #endif

            br[thread7_write_back#]

        #endif
          // --------------------------------------

    .end








bypass_path#:

;------------------------------------------------------------------------------
; RAM READ STAGE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    ctx_arb[wred_next_thread_signal]
    // signal_next_ctx[&wred_next_thread_signal]
    local_csr_wr[SAME_ME_SIGNAL, (0x80 | (&wred_next_thread_signal <<3))]

;------------------------------------------------------------------------------
; RAM WRITE STAGE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    br!=ctx[7, wred_final_synchronization#]


    // Signal the thread 0 on DL_NEXT_ME with wred_next_me_signal
    // signal_me[DL_NEXT_ME, &wred_next_me_signal]
    cap[fast_wr,((&wred_next_me_signal) | (DL_NEXT_ME <<7)), INTERTHREAD_SIG]








wred_final_synchronization#:

#ifndef WRED_DISABLE_FIN_SYNC
    ctx_arb[wred_next_thread_signal]

wred_end#:
    // signal_next_ctx[&wred_next_thread_signal]
    local_csr_wr[SAME_ME_SIGNAL, (0x80 | (&wred_next_thread_signal <<3))]

#endif



.end
#endm  // macro wred()


#endif  //  WRED_UC
