#ifndef WRED_UTILS_UC
#define WRED_UTILS_UC

/*******************************************************************************
                             Intel Proprietary

 Copyright (c) 1998-2003 By Intel Corporation.  All rights reserved.  
 No part of this program or publication may be reproduced, transmited,
 transcribed, stored in a retrieval system, or translated into any language
 or computer language in any form or by any means, electronic, mechanical,
 magnetic, optical, chemical, manual, or otherwise, without the prior
 written permission of:
                         Intel Corporation
                         2200 Mission College Blvd.
                         Santa Clara, CA  95052-8119
*******************************************************************************/

// FILE  wred_utils.uc
//   This file implements wred algorithm and average queue update macros
//

// Exported APIs:
//   _wred_avg_len_calculation(out_random,
//                             out_in_avg_len,
//                             in_curr_len,
//                             in_weight)
//   _wred_algorithmic_drop(in_avg_len,   \
//                          in_queue_id,  \
//                          in_color_id,  \
//                          in_random)
//   _wred_avg_len_estimate_and_put(out_in_avg_len,      \
//                                  in_last_idle_time_l, \
//                                  in_last_idle_time_h, \
//                                  EXIT_LABEL)

// Description:
//   WRED is intended to detect congestion early and convey
//   congestion notification to the end-hosts. To do this WRED maintains 
//   an expotentially weighted moving average of the queue length which is
//   used to assess incipient congestion.





//------------------------------------------------------------------------------
//#macro _wred_avg_len_calculation(out_random,       \
//                                 out_in_avg_len,   \
//                                 in_curr_len,      \
//                                 in_weight)

// _wred_avg_len_calculation
//   Calculates new queue average lenght in case the queue is not empy
//
// Inputs:
//   in_curr_len    - current lenght of the queue
//   out_in_avg_len - average lenght of the queue
//
//   wred entry cache in LM - data of the wred read from RAM wred table for 
//                    the queue which the currently processed packet 
//                    has been directed to. The data should be stored in local 
//                    memory. Active_lm_addr_0 index must point at 
//                    the beginning of them.
//
// Outputs:
//   out_random     - 16bit random value - it is computed in this macro
//                    to fill up free differs
//   out_in_avg_len - average lenght of the queue
//
#macro _wred_avg_len_calculation(out_random,       \
                                 out_in_avg_len,   \
                                 in_curr_len,      \
                                 in_weight)
.begin
.reg change_factor
.reg sign_bit

    ; change_factor = curr_len - avg_len
    alu[change_factor, in_curr_len, -, out_in_avg_len]

    ; we copy sign bit to every position in sign_bit register
    ; it helps make the next division sign sensitive (right shift)
    alu[sign_bit, --, B, change_factor]
    asr[sign_bit, sign_bit, >>31]

    ; avg_len = avg_len + (change_factor >> weight)
    alu[--, in_weight, OR, 0]
    dbl_shf[change_factor, sign_bit, change_factor, >>indirect]

    alu[out_in_avg_len, out_in_avg_len, +, change_factor]
    alu[out_in_avg_len, out_in_avg_len, +carry, 0]

    alu[--, --, B, out_in_avg_len, >>16]

    beq[macro_end#], defer[3]
    
	local_csr_rd[PSEUDO_RANDOM_NUMBER]
    // a CSR read is always followed by "immed" instruction
    immed[out_random,0x0]  
    immed_w1[out_random,0x0]

    immed[out_in_avg_len, 0xffff]
        
    macro_end#:

.end
#endm





//------------------------------------------------------------------------------
//#macro _wred_avg_len_estimate_and_put(out_in_avg_len,      \
//                                      in_last_idle_time_l, \
//                                      in_last_idle_time_h, \
//                                      EXIT_LABEL)

// _wred_avg_len_estimate_and_put
//   Calculates new queue average lenght. It should be issued in the case 
//  the queue goes empy. On exiting it marks the packet to be passed.
//
//
// Inputs:
//
//   out_in_avg_len      - average lenght of the queue
//   in_last_idle_time_l
//   in_last_idle_time_h - timestamp when the queue become empty
//   EXIT_LABEL             - label which points where to go after getting the
//                         job finished.
//
//
// Local memory inputs:
//
//   Wred entry  - data of the wred read from RAM wred table for 
//                 the queue which the currently processed packet 
//                 has been directed to. The data should be stored in local 
//                 memory. Active_lm_addr_0 index must point at 
//                 the beginning of them.
//
//
// Outputs:
//   out_in_avg_len     - average lenght of the queue
//
#macro _wred_avg_len_estimate_and_put(out_in_avg_len,      \
                                      in_last_idle_time_l, \
                                      in_last_idle_time_h, \
                                      EXIT_LABEL)
.begin
.reg    m
.reg    timestamp_l
.reg    timestamp_h
.reg    service_time 
.reg    change_factor


    //get the current time
    local_csr_rd[TIMESTAMP_LOW]
    immed[timestamp_l,0]
    local_csr_rd[TIMESTAMP_HIGH]
    immed[timestamp_h,0]

    ; timestamp = current_timestamp - last_queue_idle_timestamp
    alu[timestamp_l, timestamp_l, -,      in_last_idle_time_l]
    alu[timestamp_h, timestamp_h, -carry, in_last_idle_time_h]
        
    alu[service_time, --, B, *l$index0[WRED_LM_SERV_TIME], >>8]

    ; m = timestamp / (2 ^ service_time)
    alu[--, service_time, OR, 0]
    dbl_shf[m, timestamp_h, timestamp_l,>>indirect]
    alu[--, service_time, OR, 0]
    alu_shf[timestamp_h,--,B,timestamp_h,>>indirect]
    ;if (m > 0xFFFFFFFF)  goto  set_zero#
    bne[set_zero#]

    ; if (m < (2^weight)-1) goto newton#
    alu[--, *l$index0[WRED_LM_WEIGHT], OR, 0]
    alu[--, --, B, m, >>indirect]
    bne[skip_newton#], defer[1]
    ld_field_w_clr[out_in_avg_len,  0011, *l$index0[WRED_LM_Q_AVG_LEN],  >>16]

        newton#:
            .begin
            .reg temp
                ; m = m * avg_len
                mul_step[m, out_in_avg_len], 16x16_start
                mul_step[m, out_in_avg_len], 16x16_step1
                mul_step[m, out_in_avg_len], 16x16_step2
                mul_step[change_factor, --], 16x16_last

                ; avg_len = avg_len - ((m * avg_len) >> (weight + 1))
                alu[temp, *l$index0[WRED_LM_WEIGHT], +, 1]
                alu[--, temp, OR, 0]
                alu[change_factor, --, B, change_factor, >>indirect]
                br[EXIT_LABEL], defer[2]
                alu[out_in_avg_len, out_in_avg_len, -, change_factor]
                ; set next block to pass exit
                immed[dl_next_block, WRED_NEXT1]
            .end



        set_zero#:

            br[EXIT_LABEL], defer[2]
                alu[out_in_avg_len, --, B, 0]
                ; set next block to pass exit
                immed[dl_next_block, WRED_NEXT1]



        skip_newton#:
            .begin
            .reg p
                ; p = ( m + m >> 1)  >> weight
                alu[p, --, B, m, >>1]
                alu[p, p, +, m]
                alu[--, *l$index0[WRED_LM_WEIGHT], OR, 0]
                alu[p , --, B, p, >>indirect]

                ; set next block to pass exit
                immed[dl_next_block, WRED_NEXT1]

                br[EXIT_LABEL], defer[2]
                ; avg_len = avg_len >> p
                alu[--, p, OR, 0]
                alu[out_in_avg_len, --, B, out_in_avg_len, >>indirect]
            .end            



.end
#endm




//------------------------------------------------------------------------------
//#macro _wred_algorithmic_drop(in_avg_len,   \
//                              in_queue_id,  \
//                              in_color_id,  \
//                              in_random)

// _wred_algorithmic_drop
//   Calculates the dropping probability for the current packet and 
//   decides whether or not to drop it. 

// Inputs:
//   in_avg_len     - average lenght of the queue
//   in_queue_id    - id of the queue the current packet has to be sent
//                    it is used to update 32 highier bits of statistics
//   in_color_id    - color of the current packet multiplied by 4 
//                    it is used to update statistics
//   in_random      - 16bit random value - it is computed in 
//                    _wred_avg_len_calculation macro to save 2 free time slots
//                    in its internal branch instruction and thereby lessen
//                    the length of the longest path of the whole WRED algorithm
//
//
// Dispatch loop inputs:
//
//   dl_packet_size - number of bytes in the currently processed packet,
//                    used for statistics
//
//
// Local memory inputs:
//
//   wred entry cache in LM - data of the wred read from RAM wred table for 
//                    the queue which the currently processed packet 
//                    has been directed to. The data should be stored in local 
//                    memory. Active_lm_addr_0 index must point at 
//                    the beginning of them. Active_lm_addr_1 is 
//                    the Active_lm_addr_0 increased by 0, 4, 8 depending on
//                    the current packet color.
//
//
// Dispatch loop outputs:
//
//   dl_next_block  - the identification of the next block which the packet is
//                    directed to. If the packet is passed it is set to
//                    WRED_NEXT1. If a packet is dropped the variable is set to
//                    WRED_NEXT2
//
// Local memory outputs:
//
//   wred entry cache in LM - macro modifies statistics in cache
//
//                    

#macro _wred_algorithmic_drop(in_avg_len,   \
                              in_queue_id,  \
                              in_color_id,  \
                              in_random)


.begin
.reg count drop_prob_count drop_prob

// -----------------------------------------
#ifndef WRED_CORE_PROBABILITY_CALCULATION
.reg max_th min_th const 

    
        ld_field_w_clr[max_th, 0011, *l$index1[WRED_LM_MAX_TH], >>16]

        alu[--, in_avg_len, -, max_th]
        bgt[drop#]

        ld_field_w_clr[min_th, 0011, *l$index1[WRED_LM_MIN_TH]]
    
        ; avg_len and min_th are fixed point values with 8-bit fraction part
        alu[drop_prob, in_avg_len, -, min_th]
        ble[pass#]

        ; const is fixed point value with 16-bit fraction part 
        ; it holds probability in range 0.0000 - 0.FFFF (decimal part zero)
        ld_field_w_clr[const,  0011, *l$index1[WRED_LM_CONST], >>16]

        ; multiply two 16-bit long fixed point values 
        mul_step[drop_prob, const], 16x16_start
        mul_step[drop_prob, const], 16x16_step1
        mul_step[drop_prob, const], 16x16_step2
        mul_step[drop_prob, --], 16x16_last

        ; product (drop probability threshold) has always zero decimal part 
        ; now we truncate 8 LSB bits from the fraction part
        alu[drop_prob, --, B, drop_prob, >>8]

        ; count is 16-bit decimal value
        ld_field_w_clr[count,  0011, *l$index1[WRED_LM_COUNT]]

// -----------------------------------------
#else

        // drop probability computed by CC
        #define in_drop_prob     *l$index0[WRED_LM_PROB]
        #define count            *l$index0[WRED_LM_COUNT]

        .begin
        .reg _max_prob
            ; is drop probability grater than max value ?
            immed[_max_prob, 0xFFFF]
            alu[drop_prob, _max_prob, and, in_drop_prob]
            beq[pass#]

            alu[--, drop_prob, -, _max_prob]
            bhs[drop#]
        .end

#endif
// -----------------------------------------

        mul_step[drop_prob, count], 16x16_start
        mul_step[drop_prob, count], 16x16_step1
        mul_step[drop_prob, count], 16x16_step2
        mul_step[drop_prob_count, --], 16x16_last
        ; drop_prob_count is a fixed point value with 16-bit fraction part

        ; if drop_prob_count > 0.FFFF (decimal part non-zero), we drop packet
        alu_shf[--, --, B, drop_prob_count, >>16]
        bne[drop#]

        ; multiply two 16-bit long fractions with empty decimal part
        mul_step[drop_prob_count, in_random], 16x16_start
        mul_step[drop_prob_count, in_random], 16x16_step1
        mul_step[drop_prob_count, in_random], 16x16_step2
        mul_step[drop_prob_count, --], 16x16_last

        ; truncate product back to 16-bit long fraction part
        alu_shf[drop_prob_count, --, B, drop_prob_count, >>16]

        ; add the result to intial dropping threshold (drop_prob)
        alu[drop_prob, drop_prob, +, drop_prob_count]

        ; drop packet, if a randomly drawn value is below dropping threshold
        alu[--, in_random, -, drop_prob]
        bcc[drop#]

        ; count++

        // -----------------------------------------
        #ifndef WRED_CORE_PROBABILITY_CALCULATION

            alu[count, count, +, 1]
            br[wred_finish#], defer[2]
            ld_field[*l$index1[WRED_LM_COUNT], 0011, count]

        // -----------------------------------------
        #else

            br[wred_finish#], defer[2]
            alu[*l$index0[WRED_LM_COUNT], *l$index0[WRED_LM_COUNT], +, 1]
        #endif        
        // -----------------------------------------

        ; defer 2: set next block by overwriting dl_next_block
        immed[dl_next_block, WRED_NEXT1]

    pass#:
        // --------------------------------------------
        #ifdef WRED_DIFFSERV_DEBUG
            alu[@debug_came_out, @debug_came_out, +, 1]
        #endif
        // --------------------------------------------

        br[wred_finish#], defer[2]

        ; defer [1] count = 1

        // -----------------------------------------
        #ifndef WRED_CORE_PROBABILITY_CALCULATION
            ld_field[*l$index1[WRED_LM_COUNT], 0011, 1]
            ; set next block by overwriting dl_next_block
        #else
            ld_field[*l$index0[WRED_LM_COUNT], 0011, 1]
            ; set next block by overwriting dl_next_block
        #endif
        // -----------------------------------------

        ; defer [2]
        immed[dl_next_block, WRED_NEXT1]



    // ---------------------------------
    #ifdef WRED_PACKET_COUNTER_FEATURE
    // ---------------------------------
    ; here we update 64-bit statistics - this is an exception code

    .begin
    .reg count_base count_offset

    pkt_32high_cnt_update#:
        // counter_offset = WRED_64BIT_STAT_SRAM_BASE + 
        //                  WRED_64BIT_STAT_SRAM_SIZE * queue_id +
        //                  (4 bytes) * color_id.
        alu_shf[count_base,   --, B, in_color_id, <<2]
        alu_shf[count_offset, --, B, in_queue_id, \
                                <<WRED_64BIT_STAT_LOG_SRAM_SIZE]
        alu[count_offset, count_offset, +, count_base]
        immed32(count_base, WRED_64BIT_STAT_SRAM_BASE)
        sram[incr, --, count_base, count_offset]
        br[byte_32low_cnt_update#]

    byte_32high_cnt_update#:
        // counter_offset = WRED_64BIT_STAT_SRAM_BASE + 
        //                  WRED_64BIT_STAT_SRAM_SIZE * queue_id +
        //                  (4 bytes) * color_id + 12.
        alu_shf[count_base,   --, B, in_color_id, <<2]
        alu_shf[count_offset, --, B, in_queue_id, \
                                <<WRED_64BIT_STAT_LOG_SRAM_SIZE]
        alu[count_offset, count_offset, +, count_base]
        immed32(count_base, WRED_64BIT_STAT_SRAM_BASE)
        alu[count_base, count_offset, +, count_base]
        sram[incr, --, 12, count_base]
        br[skip_statistics#]

    .end

    // -----------
    #endif
    // -----------


    drop#:

        // ---------------------------------
        #ifdef WRED_PACKET_COUNTER_FEATURE
        // ---------------------------------
        .begin
        .reg packet_size


        pkt_32low_cnt_update#:

            // -----------------------------------------
            #ifndef WRED_CORE_PROBABILITY_CALCULATION
            br_bclr[*l$index0[WRED_LM_STATUS], 7, skip_statistics#]
            alu[*l$index1[WRED_LM_PACKETS_CNT], \
                *l$index1[WRED_LM_PACKETS_CNT], +, 1]
            #else
            br_bclr[*l$index0[WRED_LM_STATUS], 16, skip_statistics#]
            alu[*l$index0[WRED_LM_PACKETS_CNT], \
                *l$index0[WRED_LM_PACKETS_CNT], +, 1]

            #endif
            // -----------------------------------------
            bcs[pkt_32high_cnt_update#]

        byte_32low_cnt_update#:

            dl_meta_get_packet_size(packet_size)
            // -----------------------------------------
            #ifndef WRED_CORE_PROBABILITY_CALCULATION
            alu[*l$index1[WRED_LM_BYTES_CNT], \
                *l$index1[WRED_LM_BYTES_CNT], +, packet_size]
            #else
            alu[*l$index0[WRED_LM_BYTES_CNT], \
                *l$index0[WRED_LM_BYTES_CNT], +, packet_size]
            #endif
            // -----------------------------------------
            bcs[byte_32high_cnt_update#]

        skip_statistics#:

        .end
        // -----------
        #endif
        // -----------

        ; count=1

        // -----------------------------------------
        #ifndef WRED_CORE_PROBABILITY_CALCULATION
            ld_field[*l$index1[WRED_LM_COUNT], 0011, 1]
            ; set next block by overwriting dl_next_block
        #else
            ld_field[*l$index0[WRED_LM_COUNT], 0011, 1]
            ; set next block by overwriting dl_next_block
        #endif
        // -----------------------------------------

        ; set next block by overwriting dl_next_block
        immed[dl_next_block, WRED_NEXT2]

        // --------------------------------------------
        #ifdef WRED_DIFFSERV_DEBUG
            alu[@debug_dropped, @debug_dropped, +, 1]
        #endif
        // --------------------------------------------

wred_finish#:

.end
#endm




#endif   //  WRED_UTILS_UC
