/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      ATM RX Core Component.
 *
 * = FILENAME
 *      ix_cc_atmrx_cid.c
 *
 * = DESCRIPTION
 *      functions provide CID support functionality
 *
 * = AUTHOR
 *       Jaroslaw Kogut
 *       jaroslaw.kogut@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      11/07/2002
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_atmrx_cid.c,v 1.2 2003/09/16 17:01:16 rwielick Exp $";

/**
 * System defined include files
 */
#include "ix_cc_error.h"
#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_cc.h"
#include "ix_cc_macros.h"
#include "ix_config.h"
#include "bindings.h"
#include "dl_system_config.h"
#include "cc/ix_cc_atmsar.h"
#include "cc/ix_cc_atmsar_plugin.h"

/**
 * User defined include files
 */
#include "cc/ix_cc_atmrx_cci.h"
#include "cc/internal/ix_cc_atmrx_private.h"

#include "cc/internal/ix_cc_atmrx_hash.h"

#ifdef AAL2_INCLUDED

/**
 * Preprocessor symbols and Macros used in this file
 */

/**
 * Type definitions whose scope is limited to this file
 */

/**
 * TYPENAME: _ix_cc_atmrx_cid_key_t
 *
 * DESCRIPTION:
 *    Contains CID key:
 *     key:    VPI(8) VCI(16) CID(8)
 *     keyExt: portNo(11) - least significant
 */
typedef struct
{
    ix_uint32 key;
    ix_uint32 keyExt;
} _ix_cc_atmrx_cid_key_t;


/**
 * TYPENAME: _ix_cc_atmrx_cid_hash_pos_t
 *
 * DESCRIPTION:
 *    Contains CID key:
 *     key:    VPI(8) VCI(16) CID(8)
 *     keyExt: portNo(11) - least significant
 */
typedef struct
{
#define _IX_CC_ATMRX_CID_HASH_POS_PRI 1
#define _IX_CC_ATMRX_CID_HASH_POS_SEC 2
    ix_uint8  tableId; /* Primary or Secondary CID Hash table */
    ix_uint8  key;     /* Key number in Hash Bucket */
    ix_uint16 idx;     /* index in hash table */
} _ix_cc_atmrx_cid_hash_pos_t;
    

/** 
 * Extern function prototypes.
 */


/** 
 * Static function prototypes.
 */


/** 
 * Function definitions.
 */



/**
 * NAME: _ix_cc_atmrx_plugin_cid_alloc
 *
 * DESCRIPTION:
 *    This is function for plug-in registering. The ATM SAR
 *    calls the function if it wants to allocate CID for CID creation.
 *    The called function should allocate and return new CID handle.
 *    The 'arg_pCidDescr' agrument is passed to the function because
 *    parameters of CID could influent on assigned handle
 * 
 * @Param:  IN  arg_pInstance :  pointer to structure describing serviced
 *                  instance.
 * @Param:  OUT arg_pCid : poiter to data where assigned handle is stored
 *                  CID handle
 * @Param:  IN  arg_pCidDescr : pointer to data describing parameters
 *                  for requested CID
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - ATM port successfully created
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM - invalid input parameters
 */
ix_error
_ix_cc_atmrx_plugin_cid_alloc(
void *arg_pInstance,
ix_cc_atmsar_cid_handle_t *arg_pCid,
const ix_cc_atmsar_cid_descr_t *arg_pCidDescr)
{
    ix_uint16 *pHead, cidHandle;
    _ix_cc_atmrx_context_t *pCtx=arg_pInstance;
    
    IX_ERROR_DCHECK_ARG_NULL(1, arg_pInstance);
    IX_ERROR_DCHECK_ARG_NULL(2, arg_pCid);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pCidDescr);

    *arg_pCid = IX_CC_ATMSAR_CIDID_INVALID;

    /* Allocate an entry from free list */
    pHead = &pCtx->freeCidCtxs;
    if (*pHead == IX_CC_ATM_LIST_INVALID_HANDLE)
    {
        return IX_ERROR_WARNING(IX_CC_ATMSAR_ERROR_INVALID_PARAM,
                                ("No free CID handles\n"));
    }

    cidHandle = *pHead;

    /* remove an entry from free CID list */
    ix_cc_atm_list_remove(cidHandle, 0, pHead, &pCtx->listCid);

    *arg_pCid = cidHandle;
    _IX_CC_ATMRX_DBG_CID(("_ix_cc_atmrx_plugin_cid_alloc: CID %d ALLOCATED\n", *arg_pCid));
    return IX_SUCCESS;
} /* end _ix_cc_atmrx_plugin_cid_alloc */


/**
 * NAME: _ix_cc_atmrx_plugin_cid_free
 *
 * DESCRIPTION:
 *    This is function for plug-in registering. The ATM SAR
 *    calls the function if it wants to free CID for CID destroying.
 * 
 * @Param:  IN  arg_pInstance :  pointer to structure describing serviced
 *                  instance.
 * @Param:  IN  arg_hCid : handle of freeing CID.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - ATM port successfully created
 */
ix_error
_ix_cc_atmrx_plugin_cid_free(
void *arg_pInstance,
ix_cc_atmsar_cid_handle_t arg_hCid)
{
    ix_uint16 *pHead;
    _ix_cc_atmrx_context_t *pCtx=arg_pInstance;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_pInstance);

    if (arg_hCid == IX_CC_ATMSAR_CIDID_INVALID)
    {
        return IX_SUCCESS;
    }

    pHead = &pCtx->freeCidCtxs;

    /* insert an entry to the free CID list */
    if (ix_cc_atm_list_insert(arg_hCid, 0, pHead, &pCtx->listCid) != IX_CC_ATM_LIST_OK)
    {
        _IX_CC_ATMRX_DBG_CID(("_ix_cc_atmrx_plugin_cid_free() insert CID %d failed\n", arg_hCid));
    }

    _IX_CC_ATMRX_DBG_CID(("_ix_cc_atmrx_plugin_cid_free\n"));
    return IX_SUCCESS;
} /* end _ix_cc_atmrx_plugin_cid_free */



/******************************************************************************
 *                     CID CONTEXT TABLE SUPPORT FUNCTIONS
 ******************************************************************************/


ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_aal2_cid_rx_ctx,    buffer_handle,    3,  0,     0xFFFFFFFF);
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_aal2_cid_rx_ctx,    suId,             4,  0,     0xFFFFFFFF);


/**
 * NAME: _ix_cc_atmrx_aal2_cid_rx_ctx_get_entry
 *
 * DESCRIPTION:
 *    Find pointer to entry in aal2 context table. 
 * 
 * @Param:  - IN arg_entryId - Id of requested entry
 * @Param:  - IN arg_pContext - current context
 *
 * @Return: NULL if error occurs, otherwise pointer to aal2 cid context table entry.
 */
ix_uint32*
_ix_cc_atmrx_aal2_cid_rx_ctx_get_entry(
_ix_cc_atmrx_context_t *arg_pContext,
ix_uint16 arg_entryId)
{
    return arg_pContext->pSymbolTable[_IX_CC_ATMRX_AAL2_CID_RX_CTX_BASE_SYMBOL_ID]+((AAL2_CID_RX_CTX_ENTRY_SIZE>>2)*arg_entryId);
} /* end _ix_cc_atmrx_aal2_cid_rx_ctx_get_entry */


/**
 * NAME: _ix_cc_atmrx_aal2_cid_rx_ctx_init
 *
 * DESCRIPTION:
 *    Initialise AAL2 CID Context table. 
 * 
 * @Param:  IN arg_pContext : current context
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 */
ix_error
_ix_cc_atmrx_aal2_cid_rx_ctx_init(
_ix_cc_atmrx_context_t *arg_pContext)
{
    ix_uint32 *pEntry;
    ix_uint32 i;
    ix_error err = IX_SUCCESS;
    ix_hash_48 hash48;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_pContext);

    hash48.m_LW0 = HASH_MULTIPLIER_LOW;
    hash48.m_LW1 = HASH_MULTIPLIER_HIGH;

    err =  ix_rm_hash_48_multiplier_set(&hash48);
    
    if (err == IX_SUCCESS)
    {
        for (i=0; i<MAX_AAL2_CHANNELS; i++)
        {
            pEntry=_ix_cc_atmrx_aal2_cid_rx_ctx_get_entry(arg_pContext, i);
            _ix_cc_atmrx_aal2_cid_rx_ctx_set_buffer_handle(pEntry, 0xFF);
        }
    }

    return err;
} /* end _ix_cc_atmrx_aal2_cid_rx_ctx_init */


/**
 * NAME: _ix_cc_atmrx_aal2_cid_rx_ctx_insert
 *
 * DESCRIPTION:
 *    Function sets initial parameters for the CID in AAL2 CID Context table.
 * 
 * @Param:  IN  arg_pContext : current context
 * @Param:  IN  arg_hCid :  handle to created CID.
 * @Param:  IN  arg_pCidDescr : parameters describing requested CID.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCESS - CID entry successfully initialized
 */
static ix_error
_ix_cc_atmrx_aal2_cid_rx_ctx_insert(
_ix_cc_atmrx_context_t *arg_pContext,
const ix_cc_atmsar_cid_handle_t arg_hCid,
const ix_cc_atmsar_cid_descr_t *arg_pCidDescr)
{
    ix_uint32 *pEntry;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_pContext);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pCidDescr);

    pEntry = _ix_cc_atmrx_aal2_cid_rx_ctx_get_entry(arg_pContext, arg_hCid);

    ix_ossl_memset((void*)pEntry, 0, AAL2_CID_RX_CTX_ENTRY_SIZE);

    _ix_cc_atmrx_aal2_cid_rx_ctx_set_buffer_handle(pEntry, 0xFF);
    return IX_SUCCESS;
} /* end _ix_cc_atmrx_aal2_cid_rx_ctx_insert */


/**
 * NAME: _ix_cc_atmrx_aal2_cid_rx_ctx_remove
 *
 * DESCRIPTION:
 *    The function sets CID as invalid.
 * 
 * @Param:  IN  arg_pContext : current context
 * @Param:  IN  arg_hCid :  handle to removed VC.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - ATM port successfully inserted
 */
static ix_error
_ix_cc_atmrx_aal2_cid_rx_ctx_remove(
_ix_cc_atmrx_context_t *arg_pContext,
const ix_cc_atmsar_cid_handle_t arg_hCid)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_pContext);

    _IX_CC_ATMRX_DBG_CID(("_ix_cc_atmrx_aal2_cid_rx_ctx_remove\n"));
    
    return IX_SUCCESS;
} /* end _ix_cc_atmrx_aal2_cid_rx_ctx_remove */


/******************************************************************************
 *                   PRIMARY CID HASH TABLE SUPPORT FUNCTIONS
 ******************************************************************************/


ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_cid_hash_pri,    ptr1,    1,  0,     0xFFFF);
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_cid_hash_pri,    ptr2,    3,  0,     0xFFFF);
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_cid_hash_pri,    ptr3,    5,  0,     0xFFFF);
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_cid_hash_pri,    next,    7,  0,     0xFFFF);
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_cid_hash_pri,    key1,    0,  0,     0xFFFFFFFF);
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_cid_hash_pri,    keyExt1, 1,  16,    0xFFFF);
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_cid_hash_pri,    key2,    2,  0,     0xFFFFFFFF);
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_cid_hash_pri,    keyExt2, 3,  16,    0xFFFF);
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_cid_hash_pri,    key3,    4,  0,     0xFFFFFFFF);
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_cid_hash_pri,    keyExt3, 5,  16,    0xFFFF);


/**
 * NAME: _ix_cc_atmrx_cid_hash_pri_get_entry
 *
 * DESCRIPTION:
 *    Find pointer to entry in CID primary hash table. 
 * 
 * @Param:  IN  arg_pContext : current context
 * @Param:  IN  arg_entryId - Id of requested entry
 *
 * @Return: NULL if error occurs, otherwise pointer to secondary bucket.
 */
ix_uint32*
_ix_cc_atmrx_cid_hash_pri_get_entry(
_ix_cc_atmrx_context_t *arg_pContext,
ix_uint16 arg_entryId)
{
    return arg_pContext->pSymbolTable[_IX_CC_ATMRX_AAL2_CID_HASH_PRIM_SYMBOL_ID]+((AAL2_CID_HASH_PRIM_BKT_SZ_BYTES>>2)*arg_entryId);
} /* end _ix_cc_atmrx_cid_hash_pri_get_entry */


/**
 * NAME: _ix_cc_atmrx_cid_hash_pri_init
 *
 * DESCRIPTION:
 *    Initialize primary CID hash table. 
 * 
 * @Param:  IN  arg_pContext : current context
 *
 * @Return: none
 */
void
_ix_cc_atmrx_cid_hash_pri_init(
_ix_cc_atmrx_context_t *arg_pContext)
{
    ix_uint32 *pEntry;
    ix_uint32 i;
    
    for (i=0; i<AAL2_CID_HASH_PRIM_BKT_TOTAL; i++)
    {
        pEntry=_ix_cc_atmrx_cid_hash_pri_get_entry(arg_pContext, i);
        _ix_cc_atmrx_cid_hash_pri_set_ptr1(pEntry, IX_CC_ATMSAR_CIDID_INVALID);
        _ix_cc_atmrx_cid_hash_pri_set_ptr2(pEntry, IX_CC_ATMSAR_CIDID_INVALID);
        _ix_cc_atmrx_cid_hash_pri_set_ptr3(pEntry, IX_CC_ATMSAR_CIDID_INVALID);
    }
} /* end _ix_cc_atmrx_cid_hash_pri_init */


/******************************************************************************
 *                  SECONDARY CID HASH TABLE SUPPORT FUNCTIONS
 ******************************************************************************/


ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_cid_hash_sec,    ptr1,    1,  0,     0xFFFF);
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_cid_hash_sec,    ptr2,    3,  0,     0xFFFF);
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_cid_hash_sec,    ptr3,    5,  0,     0xFFFF);
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_cid_hash_sec,    next,    7,  0,     0xFFFF);
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_cid_hash_sec,    list,    1,  0,     0x0FFFF); /* used for linking into free entry list*/
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_cid_hash_sec,    key1,    0,  0,     0xFFFFFFFF);
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_cid_hash_sec,    keyExt1, 1,  16,    0xFFFF);
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_cid_hash_sec,    key2,    2,  0,     0xFFFFFFFF);
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_cid_hash_sec,    keyExt2, 3,  16,    0xFFFF);
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_cid_hash_sec,    key3,    4,  0,     0xFFFFFFFF);
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmrx_cid_hash_sec,    keyExt3, 5,  16,    0xFFFF);


/**
 * NAME: _ix_cc_atmrx_cid_hash_sec_get_entry
 *
 * DESCRIPTION:
 *    Find pointer to entry in CID secondary hash table. 
 * 
 * @Param:  IN arg_pContext - current context
 * @Param:  IN arg_entryId - Id of requested entry
 *
 * @Return: NULL if error occurs, otherwise pointer to secondary bucket.
 */
ix_uint32*
_ix_cc_atmrx_cid_hash_sec_get_entry(
_ix_cc_atmrx_context_t *arg_pContext,
ix_uint16 arg_entryId)
{
    if (arg_entryId >= AAL2_CID_HASH_SEC_BKT_TOTAL)
    {
        return NULL;
    }
    return arg_pContext->pSymbolTable[_IX_CC_ATMRX_AAL2_CID_HASH_SEC_SYMBOL_ID] + ((AAL2_CID_HASH_SEC_BKT_SZ_BYTES>>2)*arg_entryId);
} /* end _ix_cc_atmrx_cid_hash_sec_get_entry */


/**
 * NAME: _ix_cc_atmrx_cid_hash_sec_free_entry
 *
 * DESCRIPTION:
 *    Put bucket to free list.
 * 
 * @Param:  IN arg_pContext - current context
 * @Param:  IN arg_pEntry - pointer to hash bucket
 *
 * @Return: none
 */
void
_ix_cc_atmrx_cid_hash_sec_free_entry(
_ix_cc_atmrx_context_t *arg_pContext,
ix_uint16 arg_entry)
{
    ix_uint32 *pEntry;

    pEntry = _ix_cc_atmrx_cid_hash_sec_get_entry(arg_pContext, arg_entry);
    if (pEntry == NULL)
    {
        return;
    }
    
    _ix_cc_atmrx_cid_hash_sec_set_list(pEntry, arg_pContext->freeSecondaryBucket);
    arg_pContext->freeSecondaryBucket = arg_entry;
} /* end _ix_cc_atmrx_cid_hash_sec_free_entry */


/**
 * NAME: _ix_cc_atmrx_cid_hash_sec_alloc_entry
 *
 * DESCRIPTION:
 *    Allocate CID secondary bucket for use.
 * 
 * @Param:  IN arg_pContext - current context
 *
 * @Return: ID of allocated entry. If no free entry function returns
 *          _IX_CC_ATMRX_CID_HASH_SEC_IDX_INVALID.
 */
ix_uint16
_ix_cc_atmrx_cid_hash_sec_alloc_bucket(
_ix_cc_atmrx_context_t *arg_pContext)
{
    ix_uint16 retValue;
    ix_uint32* pEntry;
    
    if (arg_pContext->freeSecondaryBucket == _IX_CC_ATMRX_CID_HASH_SEC_IDX_INVALID)
    {
        return _IX_CC_ATMRX_CID_HASH_SEC_IDX_INVALID;
    }
    
    retValue = arg_pContext->freeSecondaryBucket;
    pEntry = _ix_cc_atmrx_cid_hash_sec_get_entry(arg_pContext, retValue);
    arg_pContext->freeSecondaryBucket = _ix_cc_atmrx_cid_hash_sec_get_list(pEntry);

    /* ix_ossl_memset((void*)pEntry, 0, AAL2_CID_HASH_SEC_BKT_SZ_BYTES); */
    return retValue;
} /* end _ix_cc_atmrx_cid_hash_sec_alloc_bucket */


/**
 * NAME: _ix_cc_atmrx_cid_hash_sec_init
 *
 * DESCRIPTION:
 *    Initialize secondary CID hash table. 
 * 
 * @Param:  IN arg_pContext - current context
 *
 * @Return: none
 */
void
_ix_cc_atmrx_cid_hash_sec_init(
_ix_cc_atmrx_context_t *arg_pContext)
{
    ix_uint32 *pEntry;
    ix_uint32  i;
    
    /*
     * Initialize secondary hash free list
     * NOTE: index 0 from secondary hash table is unused. Index 0 means invalid
     * index in hash tables
     */

    for (i=0; i<AAL2_CID_HASH_SEC_BKT_TOTAL; i++)
    {
        pEntry=_ix_cc_atmrx_cid_hash_sec_get_entry(arg_pContext, i);
        _ix_cc_atmrx_cid_hash_sec_set_ptr1(pEntry, IX_CC_ATMSAR_CIDID_INVALID);
        _ix_cc_atmrx_cid_hash_sec_set_ptr2(pEntry, IX_CC_ATMSAR_CIDID_INVALID);
        _ix_cc_atmrx_cid_hash_sec_set_ptr3(pEntry, IX_CC_ATMSAR_CIDID_INVALID);
        _ix_cc_atmrx_cid_hash_sec_free_entry(arg_pContext, i);
    }
} /* end _ix_cc_atmrx_cid_hash_sec_init */


/******************************************************************************
 *                          CID HASH TABLE SUPPORT
 ******************************************************************************/


/**
 * NAME: _ix_cc_atmrx_hash_48
 *
 * DESCRIPTION:
 *    Calculates hardware based hash_48
 * 
 * @Param:  IN  arg_pLw :  pointer to table of 2 long words
 * 
 * @Return: value of calculated HASH_48
 */
static ix_uint32
_ix_cc_atmrx_hash_48(
const ix_uint32 *arg_pLw)
{
    ix_hash_48 hash48;
    ix_error err;

	if (arg_pLw[0] == 0  &&  arg_pLw[1] == 0)
		return IX_ERROR_WARNING(IX_CC_ERROR_NULL,
		                 ("_ix_cc_atmrx_hash_48(): Error ! Null argument passed."));
   
    hash48.m_LW0 = arg_pLw[0];
    hash48.m_LW1 = arg_pLw[1];
    err = ix_rm_hash_48_hash(&hash48);
    
    return hash48.m_LW0;

} /* end _ix_cc_atmrx_hash_48 */


/**
 * NAME: _ix_cc_atmrx_cid_hash_search_cid_by_ptr
 *
 * DESCRIPTION:
 * 
 * @Param:  IN arg_pContext - current context
 * @Param:  IN  arg_pCidKey : pointer to structure describing CID key.
 * @Param:  IN  arg_priIdx : index to the CID primary hash table entry.
 * @Param:  OUT arg_pPos:    position of found CID.
 * 
 * @Return: 1 - key exists, 0 otherwise
 */
static ix_uint8
_ix_cc_atmrx_cid_hash_search_cid_by_ptr(
_ix_cc_atmrx_context_t *arg_pContext,
ix_uint16 arg_ptr,
ix_uint16 arg_priIdx,
_ix_cc_atmrx_cid_hash_pos_t *arg_pPos)
{
    ix_uint32 *pEntry;
    ix_uint16 next;

    /* verify hash entry from CID primary hash table */
    pEntry = _ix_cc_atmrx_cid_hash_pri_get_entry(arg_pContext, arg_priIdx);

    if (_ix_cc_atmrx_cid_hash_pri_get_ptr1(pEntry) == arg_ptr)
    {
        arg_pPos->tableId = _IX_CC_ATMRX_CID_HASH_POS_PRI;
        arg_pPos->idx = arg_priIdx;
        arg_pPos->key = 1;
        return 1;
    }
    if (_ix_cc_atmrx_cid_hash_pri_get_ptr2(pEntry) == arg_ptr)
    {
        arg_pPos->tableId = _IX_CC_ATMRX_CID_HASH_POS_PRI;
        arg_pPos->idx = arg_priIdx;
        arg_pPos->key = 2;
        return 1;
    }
    if (_ix_cc_atmrx_cid_hash_pri_get_ptr3(pEntry) == arg_ptr)
    {
        arg_pPos->tableId = _IX_CC_ATMRX_CID_HASH_POS_PRI;
        arg_pPos->idx = arg_priIdx;
        arg_pPos->key = 3;
        return 1;
    }

    next = _ix_cc_atmrx_cid_hash_pri_get_next(pEntry);

    while (next != _IX_CC_ATMRX_CID_HASH_SEC_IDX_INVALID)
    {
        /* verify hash entry from CID secondary hash table */
        pEntry = _ix_cc_atmrx_cid_hash_sec_get_entry(arg_pContext, next);

        if (_ix_cc_atmrx_cid_hash_sec_get_ptr1(pEntry) == arg_ptr)
        {
            arg_pPos->tableId = _IX_CC_ATMRX_CID_HASH_POS_SEC;
            arg_pPos->idx = next;
            arg_pPos->key = 1;
            return 1;
        }
        if (_ix_cc_atmrx_cid_hash_sec_get_ptr2(pEntry) == arg_ptr)
        {
            arg_pPos->tableId = _IX_CC_ATMRX_CID_HASH_POS_SEC;
            arg_pPos->idx = next;
            arg_pPos->key = 2;
            return 1;
        }
        if (_ix_cc_atmrx_cid_hash_sec_get_ptr3(pEntry) == arg_ptr)
        {
            arg_pPos->tableId = _IX_CC_ATMRX_CID_HASH_POS_SEC;
            arg_pPos->idx = next;
            arg_pPos->key = 3;
            return 1;
        }
        /* get next entry from CID secondary hash table */
        next = _ix_cc_atmrx_cid_hash_sec_get_next(pEntry);
    }
    return 0;
} /* end _ix_cc_atmrx_cid_hash_search_cid_by_ptr */

    
/**
 * NAME: _ix_cc_atmrx_cid_hash_search_cid_by_key
 *
 * DESCRIPTION:
 *    Searches primary and secondary CID Hash tables for key equal to the given one.
 * 
 * @Param:  IN  arg_pContext - current context
 * @Param:  IN  arg_pCidKey : pointer to structure describing CID key.
 * @Param:  IN  arg_hashIdx : index to the CID primary hash table entry.
 * @Param:  OUT arg_pPos:    position of found CID.
 * 
 * @Return: 1 - key exists, 0 otherwise
 */

static ix_uint8
_ix_cc_atmrx_cid_hash_search_cid_by_key(
_ix_cc_atmrx_context_t *arg_pContext,
const _ix_cc_atmrx_cid_key_t *arg_pCidKey,
const ix_uint16 arg_priIdx,
_ix_cc_atmrx_cid_hash_pos_t *arg_pPos)
{
    ix_uint32 *pEntry;
    ix_uint16 next;

    /* verify hash entry from CID primary hash table */
    pEntry = _ix_cc_atmrx_cid_hash_pri_get_entry(arg_pContext, arg_priIdx);
    
    if ((_ix_cc_atmrx_cid_hash_pri_get_key1(pEntry) == arg_pCidKey->key) &&
        (_ix_cc_atmrx_cid_hash_pri_get_keyExt1(pEntry) == arg_pCidKey->keyExt))
    {
        arg_pPos->tableId = _IX_CC_ATMRX_CID_HASH_POS_PRI;
        arg_pPos->idx = arg_priIdx;
        arg_pPos->key = 1;
        return 1;
    }
    if ((_ix_cc_atmrx_cid_hash_pri_get_key2(pEntry) == arg_pCidKey->key) &&
        (_ix_cc_atmrx_cid_hash_pri_get_keyExt2(pEntry) == arg_pCidKey->keyExt))
    {
        arg_pPos->tableId = _IX_CC_ATMRX_CID_HASH_POS_PRI;
        arg_pPos->idx = arg_priIdx;
        arg_pPos->key = 2;
        return 1;
    }
    if ((_ix_cc_atmrx_cid_hash_pri_get_key3(pEntry) == arg_pCidKey->key) &&
        (_ix_cc_atmrx_cid_hash_pri_get_keyExt3(pEntry) == arg_pCidKey->keyExt))
    {
        arg_pPos->tableId = _IX_CC_ATMRX_CID_HASH_POS_PRI;
        arg_pPos->idx = arg_priIdx;
        arg_pPos->key = 3;
        return 1;
    }

    next = _ix_cc_atmrx_cid_hash_pri_get_next(pEntry);

    while (next != _IX_CC_ATMRX_CID_HASH_SEC_IDX_INVALID)
    {
        /* verify hash entry from CID secondary hash table */
        pEntry = _ix_cc_atmrx_cid_hash_sec_get_entry(arg_pContext, next);

        if ((_ix_cc_atmrx_cid_hash_sec_get_key1(pEntry) == arg_pCidKey->key) &&
            (_ix_cc_atmrx_cid_hash_sec_get_keyExt1(pEntry) == arg_pCidKey->keyExt))
        {
            arg_pPos->tableId = _IX_CC_ATMRX_CID_HASH_POS_SEC;
            arg_pPos->idx = next;
            arg_pPos->key = 1;
            return 1;
        }
        if ((_ix_cc_atmrx_cid_hash_sec_get_key2(pEntry) == arg_pCidKey->key) &&
            (_ix_cc_atmrx_cid_hash_sec_get_keyExt2(pEntry) == arg_pCidKey->keyExt))
        {
            arg_pPos->tableId = _IX_CC_ATMRX_CID_HASH_POS_SEC;
            arg_pPos->idx = next;
            arg_pPos->key = 2;
            return 1;
        }
        if ((_ix_cc_atmrx_cid_hash_sec_get_key3(pEntry) == arg_pCidKey->key) &&
            (_ix_cc_atmrx_cid_hash_sec_get_keyExt3(pEntry) == arg_pCidKey->keyExt))
        {
            arg_pPos->tableId = _IX_CC_ATMRX_CID_HASH_POS_SEC;
            arg_pPos->idx = next;
            arg_pPos->key = 3;
            return 1;
        }

        /* get next entry from CID secondary hash table */
        next = _ix_cc_atmrx_cid_hash_sec_get_next(pEntry);
    }
    return 0;
} /* end _ix_cc_atmrx_cid_hash_search_cid_by_key */


/**
 * NAME: _ix_cc_atmrx_cid_hash_get_key
 *
 * DESCRIPTION:
 *    Gets CID Hash key (port/vpi/vci/cid)
 * 
 * @Param:  IN  arg_pContext :  current context.
 * @Param:  IN  arg_pSpec :  handle to created Cid.
 * @Param:  OUT arg_pCidKey :  CID Key (portNo/VPI/VCI/CID)
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCESS - CID successfully inserted into hash table
 *    IX_ERROR_WARNING - no vc handle
 */
static ix_error
_ix_cc_atmrx_cid_hash_get_key(
_ix_cc_atmrx_context_t *arg_pContext,
const ix_cc_atmsar_cid_descr_t *arg_pCidDescr,
_ix_cc_atmrx_cid_key_t *arg_pCidKey)
{
    ix_uint16 priIdx;
    ix_cc_atmsar_vc_handle_t hVc;
    _ix_cc_atmrx_hash_pos_t hashVcPos;
    _ix_cc_atmrx_vc_key_t vcKey;
    ix_uint16 vci;
    ix_uint8 vpi;
    ix_uint16 portNo;
    
    /* Get VC key from Primary Hash Table */
    hVc = arg_pCidDescr->vc;
    priIdx = arg_pContext->vcPriIdx[hVc];
    if (_ix_cc_atmrx_hash_search_vc_by_ptr(arg_pContext, hVc, priIdx, &hashVcPos) == 0)
    {
        return IX_ERROR_WARNING (IX_CC_ATMSAR_ERROR_NO_RESOURCES,
                                 ("No vc handle in RX hash"));
    }
    _ix_cc_atmrx_hash_get_vc(arg_pContext, &hashVcPos, &vcKey);
    vci = vcKey.key & 0xFFFF;
    vpi = (vcKey.key >> 16) & 0xFF;
    portNo = ((vcKey.key >> 24) & 0x7F)  |  ((vcKey.keyExt << 7) & 0x780);

    /* Create CID key from VC key and CID */
    arg_pCidKey->key = ((vpi << 24)  |  (vci << 8)  |  (arg_pCidDescr->cid));
    arg_pCidKey->keyExt = portNo & 0x07FF;
    
    return IX_SUCCESS;
} /* end _ix_cc_atmrx_cid_hash_get_key */


/**
 * NAME: _ix_cc_atmrx_cid_hash_get_idx
 *
 * DESCRIPTION:
 *    Calculates hash value based on CID key (port/vpi/vci/cid)
 * 
 * @Param:  IN arg_pCidKey :  CID Key (portNo/VPI/VCI/CID)
 * 
 * @Return: Index to primary CID hash table
 */
static ix_uint16
_ix_cc_atmrx_cid_hash_get_idx(
const _ix_cc_atmrx_cid_key_t *arg_pCidKey)
{
    ix_uint32 lw[2];
    
    /* Calculate Hash48 based on 2 long words:
       L0 (least significant: VPI VCI CID
       L1 (most significant:  portNo */
    lw[0] = arg_pCidKey->key;
    lw[1] = arg_pCidKey->keyExt;
    
    return (_ix_cc_atmrx_hash_48 (lw) >> AAL2_CID_HASH_PRIM_ENTRY_SHF)
        & (AAL2_CID_HASH_PRIM_ENTRY_MASK);
} /* end _ix_cc_atmrx_cid_hash_get_idx */


/**
 * NAME: _ix_cc_atmrx_cid_hash_set_cid
 *
 * DESCRIPTION:
 *    Function fills CID hash entry
 *    the function fills hash entry structure. Argument 'arg_pPos' describes
 *    place of setting, arg_pVcKey and arg_hVc describes what is setting.
 *
 * @Param:  IN  arg_pContext: current context
 * @Param:  IN  arg_pPos : pointer to structure describing location.
 * @Param:  IN  arg_pCidKey : CID key to be inserted
 * @Param:  IN  arg_hCid : 
 * 
 * @Return: None
 */

static void
_ix_cc_atmrx_cid_hash_set_cid(
_ix_cc_atmrx_context_t *arg_pContext,
const _ix_cc_atmrx_cid_hash_pos_t *arg_pPos,
const _ix_cc_atmrx_cid_key_t *arg_pCidKey,
const ix_cc_atmsar_cid_handle_t arg_hCid)
{
    ix_uint32 *pEntry;

    switch (arg_pPos->tableId)
    {
    case _IX_CC_ATMRX_CID_HASH_POS_PRI:
        pEntry = _ix_cc_atmrx_cid_hash_pri_get_entry(arg_pContext, arg_pPos->idx);
        switch (arg_pPos->key)
        {
        case 1:
            _ix_cc_atmrx_cid_hash_pri_set_ptr1(pEntry, arg_hCid);
            _ix_cc_atmrx_cid_hash_pri_set_key1(pEntry, arg_pCidKey->key);
            _ix_cc_atmrx_cid_hash_pri_set_keyExt1(pEntry, arg_pCidKey->keyExt);
            break;
        case 2:
            _ix_cc_atmrx_cid_hash_pri_set_ptr2 (pEntry, arg_hCid);
            _ix_cc_atmrx_cid_hash_pri_set_key2 (pEntry, arg_pCidKey->key);
            _ix_cc_atmrx_cid_hash_pri_set_keyExt2 (pEntry, arg_pCidKey->keyExt);
            break;
        case 3:
            _ix_cc_atmrx_cid_hash_pri_set_ptr3(pEntry, arg_hCid);
            _ix_cc_atmrx_cid_hash_pri_set_key3(pEntry, arg_pCidKey->key);
            _ix_cc_atmrx_cid_hash_pri_set_keyExt3(pEntry, arg_pCidKey->keyExt);
            break;
        default:
            break;
        }
        break;
    case _IX_CC_ATMRX_CID_HASH_POS_SEC:
        pEntry = _ix_cc_atmrx_cid_hash_sec_get_entry(arg_pContext, arg_pPos->idx);
        switch (arg_pPos->key)
        {
        case 1:
            _ix_cc_atmrx_cid_hash_sec_set_ptr1(pEntry, arg_hCid);
            _ix_cc_atmrx_cid_hash_sec_set_key1(pEntry, arg_pCidKey->key);
            _ix_cc_atmrx_cid_hash_sec_set_keyExt1(pEntry, arg_pCidKey->keyExt);
            break;
        case 2:
            _ix_cc_atmrx_cid_hash_sec_set_ptr2(pEntry, arg_hCid);
            _ix_cc_atmrx_cid_hash_sec_set_key2(pEntry, arg_pCidKey->key);
            _ix_cc_atmrx_cid_hash_sec_set_keyExt2(pEntry, arg_pCidKey->keyExt);
            break;
        case 3:
            _ix_cc_atmrx_cid_hash_sec_set_ptr3(pEntry, arg_hCid);
            _ix_cc_atmrx_cid_hash_sec_set_key3(pEntry, arg_pCidKey->key);
            _ix_cc_atmrx_cid_hash_sec_set_keyExt3(pEntry, arg_pCidKey->keyExt);
            break;
        default:
            break;
        }
        break;
    default:
        break;
    }
} /* end _ix_cc_atmrx_cid_hash_set_cid */


/**
 * NAME: _ix_cc_atmrx_cid_hash_insert
 *
 * DESCRIPTION:
 *    Function inserts CID into CID Hash table.
 * 
 * @Param:  IN  arg_pContext : current context.
 * @Param:  IN  arg_hCid :  handle to created CID.
 * @Param:  IN  arg_pCidDescr : parameters describing requested CID.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCESS - CID successfully inserted into hash table
 */
static ix_error
_ix_cc_atmrx_cid_hash_insert(
_ix_cc_atmrx_context_t *arg_pContext,
const ix_cc_atmsar_cid_handle_t arg_hCid,
const ix_cc_atmsar_cid_descr_t *arg_pCidDescr)
{
    ix_uint32 *pEntry;
    ix_uint16 next;
    ix_uint16 priIdx;
    _ix_cc_atmrx_cid_key_t cidKey;
    _ix_cc_atmrx_cid_key_t cidEmptyKey;
    _ix_cc_atmrx_cid_hash_pos_t hashCidPos;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_pContext);
    IX_ERROR_DCHECK_ARG_NULL(2, arg_pCidDescr);

    _ix_cc_atmrx_cid_hash_get_key(arg_pContext, arg_pCidDescr, &cidKey);
    priIdx = _ix_cc_atmrx_cid_hash_get_idx(&cidKey);
    _IX_CC_ATMRX_DBG_HASH(("atm_cid_hash_insert: priIdx=%d\n", priIdx));

    if ((cidKey.key == 0) && (cidKey.keyExt == 0))
    {
        return IX_ERROR_WARNING(IX_CC_ATMSAR_ERROR_INVALID_PARAM,
                                ("Invalid CID key"));
    }
    if (_ix_cc_atmrx_cid_hash_search_cid_by_key(arg_pContext, &cidKey, priIdx, &hashCidPos) == 1)
    {
        return IX_ERROR_WARNING(IX_CC_ATMSAR_ERROR_INVALID_PARAM,
                                ("CID with the given CID key already exists"));
    }

    /* find free entry */
    cidEmptyKey.key = 0;
    cidEmptyKey.keyExt = 0;
    if (_ix_cc_atmrx_cid_hash_search_cid_by_key(arg_pContext, &cidEmptyKey, priIdx, &hashCidPos) == 1)
    {
        /* mark that this entry is not empty */
        arg_pContext->cidPriIdx[arg_hCid] = priIdx;

        /* insert CID description into CID hash table entry */
        _ix_cc_atmrx_cid_hash_set_cid(arg_pContext, &hashCidPos, &cidKey, arg_hCid);
        
        _IX_CC_ATMRX_DBG_HASH(("_ix_cc_atmrx_cid_hash_insert: CID added do %s hash table as %d entry, idx=%d\n",
                     (hashCidPos.tableId==_IX_CC_ATMRX_CID_HASH_POS_PRI) ? "primary" : "secondary",
                     hashCidPos.key, hashCidPos.idx));
        return IX_SUCCESS;
    }

    /* No free entry in bucket found
     * Allocate new bucket, put CID into its first entry and link the bucket to other buckets
     */
    hashCidPos.tableId = _IX_CC_ATMRX_CID_HASH_POS_SEC;
    hashCidPos.key = 1;
    hashCidPos.idx = _ix_cc_atmrx_cid_hash_sec_alloc_bucket(arg_pContext);
    if (hashCidPos.idx == _IX_CC_ATMRX_HASH_SEC_IDX_INVALID)
    {
        return IX_ERROR_WARNING(IX_CC_ATMSAR_ERROR_NO_RESOURCES,
                                ("No secondary CID hash bucket"));
    }

    /* mark that this entry is not empty */
    arg_pContext->cidPriIdx[arg_hCid] = priIdx;
    
    /* insert CID description into CID hash table entry */
    _ix_cc_atmrx_cid_hash_set_cid(arg_pContext, &hashCidPos, &cidKey, arg_hCid);

    /* link new hash entry into the list */
    pEntry = _ix_cc_atmrx_cid_hash_pri_get_entry(arg_pContext, priIdx);
    next = _ix_cc_atmrx_cid_hash_pri_get_next(pEntry);
    if (next == _IX_CC_ATMRX_CID_HASH_SEC_IDX_INVALID)
    {
        _IX_CC_ATMRX_DBG_HASH(("_ix_cc_atmrx_cid_hash_insert: secondary bucket (idx=%d) linked to primary (idx=%d)\n",
                     hashCidPos.idx, priIdx));
        _ix_cc_atmrx_cid_hash_pri_set_next(pEntry, hashCidPos.idx);
    }
    else
    {
        /* find secondary hash entry */
        while (next != _IX_CC_ATMRX_CID_HASH_SEC_IDX_INVALID)
        {
            _IX_CC_ATMRX_DBG_HASH(("_ix_cc_atmrx_cid_hash_insert: secondary linked %d\n", next));
            priIdx = next;
            pEntry = _ix_cc_atmrx_cid_hash_sec_get_entry(arg_pContext, priIdx);
            next = _ix_cc_atmrx_cid_hash_sec_get_next(pEntry);
        }
        _IX_CC_ATMRX_DBG_HASH(("_ix_cc_atmrx_cid_hash_insert: secondary bucket (idx=%d) linked to secondary (idx=%d)\n",
                     hashCidPos.idx, priIdx));
        _ix_cc_atmrx_cid_hash_sec_set_next(pEntry, hashCidPos.idx);
    }

    return IX_SUCCESS;
} /* end _ix_cc_atmrx_cid_hash_insert */



/**
 * NAME: _ix_cc_atmrx_cid_hash_remove
 *
 * DESCRIPTION:
 *    Removes CID from hash CID tables 
 *    The empty hash entry will be freed.
 * 
 * @Param:  IN  arg_pContext : current context.
 * @Param:  IN  arg_hCid :  handle to removed CID.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - ATM port successfully inserted
 */
static ix_error
_ix_cc_atmrx_cid_hash_remove(
_ix_cc_atmrx_context_t *arg_pContext,
const ix_cc_atmsar_cid_handle_t arg_hCid)
{
    ix_uint32 *pEntry;
    ix_uint16 priIdx;
    ix_uint8  removeSecondaryFlag;
    ix_uint16 prevSecondary;
    ix_uint16 currentSecondary;
    ix_uint16 cidPtr;
    ix_uint16 next;
    _ix_cc_atmrx_cid_key_t cidKey;
    _ix_cc_atmrx_cid_hash_pos_t hashCidPos;
    _ix_cc_atmrx_cid_hash_pos_t hashCidPosLast;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_pContext);

    priIdx = arg_pContext->cidPriIdx[arg_hCid];
    _IX_CC_ATMRX_DBG_HASH(("_ix_cc_atmrx_cid_hash_remove: hCid=%d, priIdx=%d\n", arg_hCid, priIdx));
    if (_ix_cc_atmrx_cid_hash_search_cid_by_ptr(arg_pContext, arg_hCid, priIdx, &hashCidPos) == 0)
    {
        return IX_ERROR_WARNING(IX_CC_ATMSAR_ERROR_NO_RESOURCES,
                                ("No CID handle in RX CID hash"));
    }

    /* remove CID description from hash entry */
    cidKey.key = 0;
    cidKey.keyExt = 0;
    _ix_cc_atmrx_cid_hash_set_cid(arg_pContext, &hashCidPos, &cidKey, IX_CC_ATMSAR_CIDID_INVALID);

    /* find last entry in hash list */
    pEntry = _ix_cc_atmrx_cid_hash_pri_get_entry(arg_pContext, priIdx);
    next = _ix_cc_atmrx_cid_hash_pri_get_next(pEntry);
    if (next == _IX_CC_ATMRX_CID_HASH_SEC_IDX_INVALID)
    {
        return IX_SUCCESS;
    }

    /* Secondary CID Hash Table Entries */
    removeSecondaryFlag = 0;
    currentSecondary = _IX_CC_ATMRX_CID_HASH_SEC_IDX_INVALID;
    /* Find last secondary hash entry */
    while (next != _IX_CC_ATMRX_CID_HASH_SEC_IDX_INVALID)
    {
        _IX_CC_ATMRX_DBG_HASH (("_ix_cc_atmrx_cid_hash_remove: found CID secondary entry %d\n", next));
        currentSecondary = next; /* remember next idx */
        pEntry = _ix_cc_atmrx_cid_hash_sec_get_entry(arg_pContext, next);
        next = _ix_cc_atmrx_cid_hash_sec_get_next(pEntry);
        if (next!= _IX_CC_ATMRX_CID_HASH_SEC_IDX_INVALID)
        {
            /* unlink operation on secondary entry */
            removeSecondaryFlag = 1;
            prevSecondary = currentSecondary;
        }
    }

    /*
     * 'pEntry' it points on last secondary hash entry
     * 'priIdx' contains index to last secondary hash entry
     * 'currentSecondary' - point to current secondary hash entry
     * 'removeSecondaryFlag' =
     *      1  - unlink from secondary entry
     *      0  - unlink from primary entry
     * 'prevSecondary' - points on secondary entry from which unlink
     *    operation will be performed. The value in the variable is valid
     *    if 'removeSecondaryFlag' is equal '1'.
     */

    /* check that removed CID was in the last secondary bucket */
    if ((hashCidPos.tableId != _IX_CC_ATMRX_CID_HASH_POS_SEC) ||
        (hashCidPos.idx != priIdx))
    {
        /*
         * Removed Cid was not in last secondary cid hash bucket. We need to move
         * any entry from last secondary cid hash bucket to freed slot in cid hash
         * bucket.
         * We try to find no empty slot in last secondary hash bucket.
         */
        hashCidPosLast.tableId = _IX_CC_ATMRX_CID_HASH_POS_SEC;
        hashCidPosLast.idx = currentSecondary;
        if ((cidPtr = _ix_cc_atmrx_cid_hash_sec_get_ptr1(pEntry)) != IX_CC_ATMSAR_CIDID_INVALID)
        {
            hashCidPosLast.key = 1;
            cidKey.key = _ix_cc_atmrx_cid_hash_sec_get_key1(pEntry);
            cidKey.keyExt = _ix_cc_atmrx_cid_hash_sec_get_keyExt1(pEntry);
        }
        else if ((cidPtr = _ix_cc_atmrx_cid_hash_sec_get_ptr2(pEntry)) != IX_CC_ATMSAR_VCID_INVALID)
        {
            hashCidPosLast.key = 2;
            cidKey.key = _ix_cc_atmrx_cid_hash_sec_get_key2(pEntry);
            cidKey.keyExt = _ix_cc_atmrx_cid_hash_sec_get_keyExt2(pEntry);
        }
        else if ((cidPtr = _ix_cc_atmrx_cid_hash_sec_get_ptr3(pEntry)) != IX_CC_ATMSAR_VCID_INVALID)
        {
            hashCidPosLast.key = 3;
            cidKey.key = _ix_cc_atmrx_cid_hash_sec_get_key3(pEntry);
            cidKey.keyExt = _ix_cc_atmrx_cid_hash_sec_get_keyExt3(pEntry);
        }
        else
        {
            _IX_CC_ATMRX_DBG_HASH(("_ix_cc_atmrx_cid_hash_remove: ERROR!!! Something's wrong with the key\n"));
            hashCidPosLast.key = 0;
            cidPtr = IX_CC_ATMSAR_VCID_INVALID;
            cidKey.key = 0;
            cidKey.keyExt = 0;
        }

        _IX_CC_ATMRX_DBG_HASH(("_ix_cc_atmrx_cid_hash_remove: found key to move %d\n", hashCidPosLast.key));
        /*
         * move cid description from last secondary entry to free slot in any
         * previous hash entry (secondary or primary)
         */
        _ix_cc_atmrx_cid_hash_set_cid(arg_pContext, &hashCidPos, &cidKey, cidPtr);
            
        /* remove CID description from last secondary hash entry */
        cidKey.key = 0;
        cidKey.keyExt = 0;
        _ix_cc_atmrx_cid_hash_set_cid(arg_pContext, &hashCidPosLast, &cidKey, IX_CC_ATMSAR_CIDID_INVALID);
    }

    if ((_ix_cc_atmrx_cid_hash_sec_get_ptr1(pEntry) != IX_CC_ATMSAR_VCID_INVALID) ||
        (_ix_cc_atmrx_cid_hash_sec_get_ptr2(pEntry) != IX_CC_ATMSAR_VCID_INVALID) ||
        (_ix_cc_atmrx_cid_hash_sec_get_ptr3(pEntry) != IX_CC_ATMSAR_VCID_INVALID))
    {
        /*
         * Any slot in the last secondary hash entry used.
         * Don't free the entry
         */
        _IX_CC_ATMRX_DBG_HASH(("atmrx_hash_remove: Exit. Last secondary is not empty\n"));
        return IX_SUCCESS;
    }

    if (removeSecondaryFlag == 1)
    {
        /* No key in last cid hash secondary bucket - unlink it from previous cid hash bucket */
        pEntry = _ix_cc_atmrx_cid_hash_sec_get_entry(arg_pContext, prevSecondary);
        _ix_cc_atmrx_cid_hash_sec_set_next(pEntry, _IX_CC_ATMRX_CID_HASH_SEC_IDX_INVALID);
    }
    else
    {
        /* No key in last cid hash primary bucket - unlink it from previous cid hash bucket */
        pEntry = _ix_cc_atmrx_cid_hash_pri_get_entry(arg_pContext, priIdx);
        _ix_cc_atmrx_cid_hash_pri_set_next(pEntry, _IX_CC_ATMRX_CID_HASH_SEC_IDX_INVALID);
    }
    /* Unlink the freed bucket */
    _ix_cc_atmrx_cid_hash_sec_free_entry(arg_pContext, currentSecondary);
    _IX_CC_ATMRX_DBG_HASH(("_ix_cc_atmrx_cid_hash_remove: Exit. Last secondary is freed, prev=%d, curr=%d\n",
                 prevSecondary, currentSecondary));
    
    return IX_SUCCESS;
} /* end _ix_cc_atmrx_cid_hash_remove */


    
/**
 * TYPENAME: _ix_cc_atmrx_plugin_cid_create
 *
 * DESCRIPTION:
 *    This is function for plug-in registering. The ATM SAR
 *    calls the function if it services "CID CREATE" operation. The called
 *    function should perform the operation on internal structures.
 * 
 * @Param:  IN  arg_pInstance :  pointer to structure describing serviced
 *                  instance.
 * @Param:  IN  arg_pCidDescr : pointer to data describing parameters
 *                  for requested CID
 * @Param:  OUT arg_hCid : CID handle
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCESS - CID successfully created
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM - invalid input parameters
 */
ix_error
_ix_cc_atmrx_plugin_cid_create(
void *arg_pInstance,
ix_cc_atmsar_cid_handle_t arg_hCid,
const ix_cc_atmsar_cid_descr_t *arg_pCidDescr)
{
    /*ix_uint16 i;*/
    ix_counter_64bit_handle *p64BitCountersHdls;
    _ix_cc_atmrx_context_t* pCtx=arg_pInstance;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_pInstance);
    IX_ERROR_DCHECK_ARG_NULL(2, arg_pCidDescr);

    _IX_CC_ATMRX_DBG_CID(("_ix_cc_atmrx_plugin_cid_create\n"));

    /* Prepare CID info data and insert the entry into the hash table */
    IX_ERROR_CRT(_ix_cc_atmrx_aal2_cid_rx_ctx_insert(pCtx, arg_hCid, arg_pCidDescr),
                  IX_CC_ATMSAR_ERROR_INVALID_PARAM,
                  IX_ERROR_LEVEL_GLOBAL);
    IX_ERROR_CRT(_ix_cc_atmrx_cid_hash_insert(pCtx, arg_hCid, arg_pCidDescr),
                  IX_CC_ATMSAR_ERROR_INVALID_PARAM,
                  IX_ERROR_LEVEL_GLOBAL);
    
    /* clear statistic counters for CID */
    if (pCtx->counters)
    {
        p64BitCountersHdls = &(pCtx->hCidCounters[ATMRX_CID_STAT_COUNTERS_NO*arg_hCid]);
        /* TBD */
        /* Reset counters */
        /*
          for (i=0; i<ATMRX_CID_STAT_COUNTERS_NO; i++)
          {
          IX_ERROR_CRT (ix_rm_counter_64bit_set_value (*(p64BitCountersHdls+i), 0),
          IX_CC_ATMRX_ERROR_RM_SET_64BIT_COUNTER,
          IX_ERROR_LEVEL_REMOTE);
          }
        */
    }
                
    return IX_SUCCESS;
} /* end _ix_cc_atmrx_plugin_cid_create */

/**
 * TYPENAME: _ix_cc_atmrx_plugin_cid_update
 *
 * DESCRIPTION:
 *    This is function for plug-in registering. The ATM SAR
 *    calls the function if it services "CID UPDATE" operation. The called
 *    function should perform the operation on internal structures.
 * 
 * @Param:  IN  arg_pInstance :  pointer to structure describing serviced
 *                  instance.
 * @Param:  IN  arg_pCidDescr : pointer to data describing parameters
 *                  for requested CID
 * @Param:  OUT arg_hCid : CID handle
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCESS - CID successfully created
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM - invalid input parameters
 */
ix_error
_ix_cc_atmrx_plugin_cid_update(
void *arg_pInstance,
ix_cc_atmsar_cid_handle_t arg_hCid,
const ix_cc_atmsar_update_data_t *arg_pUpdate)
{
    ix_uint32 *pEntry;
    ix_cc_atmrx_update_data_t *pUpdateData = (ix_cc_atmrx_update_data_t*)arg_pUpdate;
    ix_counter_64bit_handle *p64BitCountersHdls;
    _ix_cc_atmrx_context_t* pCtx=arg_pInstance;
    
    IX_ERROR_DCHECK_ARG_NULL(1, arg_pInstance);
    IX_ERROR_DCHECK_ARG_NULL(2, arg_pUpdate);

    _IX_CC_ATMRX_DBG_CID(("_ix_cc_atmrx_plugin_cid_update\n"));
    
    _IX_CC_ATMRX_DBG_CID(("_ix_cc_atmrx_plugin_cid_update: WORD0: %d WORD1: %d\n",
                          pUpdateData->lw0.suId,
                          pUpdateData->lw1.optionalInfo));

    pEntry = _ix_cc_atmrx_aal2_cid_rx_ctx_get_entry(pCtx, arg_hCid);
    _ix_cc_atmrx_aal2_cid_rx_ctx_set_suId(pEntry, pUpdateData->lw0.suId);
    
    return IX_SUCCESS;
} /* end _ix_cc_atmrx_plugin_cid_update */

/**
 * TYPENAME: _ix_cc_atmrx_plugin_cid_remove
 *
 * DESCRIPTION:
 *    This is function for plug-in registering. The ATM SAR
 *    calls the function if it services "CID REMOVE" operation. The called
 *    function should perform the operation on internal structures.
 * 
 * @Param:  IN  arg_pInstance :  pointer to structure describing serviced
 *                  instance.
 * @Param:  IN  arg_hCid : CID handle for removed CID
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - CID successfully removed
 */
ix_error
_ix_cc_atmrx_plugin_cid_remove(
void *arg_pInstance,
ix_cc_atmsar_cid_handle_t arg_hCid)
{
    _ix_cc_atmrx_context_t* pCtx=arg_pInstance;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_pInstance);

    _IX_CC_ATMRX_DBG_CID(("_ix_cc_atmrx_plugin_cid_remove\n"));

    /* Remove CID from CID hash table and CID Context table */
    IX_ERROR_CRT (_ix_cc_atmrx_cid_hash_remove(pCtx, arg_hCid),
                  IX_CC_ATMSAR_ERROR_INVALID_PARAM,
                  IX_ERROR_LEVEL_GLOBAL);
    IX_ERROR_CRT (_ix_cc_atmrx_aal2_cid_rx_ctx_remove(pCtx, arg_hCid),
                  IX_CC_ATMSAR_ERROR_INVALID_PARAM,
                  IX_ERROR_LEVEL_GLOBAL);

    return IX_SUCCESS;
} /* end _ix_cc_atmrx_plugin_cid_remove */


/**
 * TYPENAME: _ix_cc_atmrx_plugin_get_cid_stats
 *
 * DESCRIPTION:
 *    This is function for plug-in registering. The ATM SAR
 *    calls the function if it services "GET CID STATS" operation. The called
 *    function should perform the operation on internal structures.
 * 
 * @Param:  IN  arg_pInstance :  pointer to structure describing serviced
 *                  instance.
 * @Param:  IN  arg_cid : CID handle
 * @Param:  OUT arg_pLookupResult : pointer to ix_cc_atmsar_cid_stats_t
 *                  structure where the requested data will be stored
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - statistics successfully retrieved
 *    IX_CC_INVALID_PORT - specified ATM port does not exist 
 */
ix_error
_ix_cc_atmrx_plugin_get_cid_stats(
void *arg_pInstance,
ix_cc_atmsar_cid_handle_t arg_cid,
ix_cc_atmsar_cid_stats_t *arg_pLookupResult)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_pInstance);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pLookupResult);
    /* TBD */
    _IX_CC_ATMRX_DBG_CID(("_ix_cc_atmrx_plugin_get_cid_stats\n"));
    return IX_SUCCESS;
} /* end _ix_cc_atmrx_plugin_get_cid_stats */

#endif /* AAL2_INCLUDED */



