/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      ix_cc_csix_rx_int.h
 *
 * = DESCRIPTION
 *      This file contains parameters used internally by Csix Rx core 
 *      component.
 *
 * = CHANGE HISTORY
 *      8/22/2002 - Created.
 *
 * ============================================================================
 * $Id: ix_cc_csix_rx_int.h,v 1.6 2002/10/09 00:39:33 sharma Exp $
 */

#if !defined(__IX_CC_CSIX_RX_INT_H__)
#define __IX_CC_CSIX_RX_INT_H__



/**
 * System defined include files required.
 */
#include "ix_types.h"
#include "cc/ix_cc_csix_rx.h"

#include "ix_ossl.h" /* for malloc and free */


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */



/**
 * Prototypes for internal functions.
 */



/* Size defines for callback context tables */
#define CSIX_RX_NUM_ENTRIES_CB_CTX_TBL_COMMON           128 /* must be power of 2 */
#define CSIX_RX_NUM_ENTRIES_CB_CTX_TBL_GET_STAT         CSIX_RX_NUM_ENTRIES_CB_CTX_TBL_COMMON
#define CSIX_RX_SIZE_CB_CTX_TBL_GET_STAT                (CSIX_RX_NUM_ENTRIES_CB_CTX_TBL_GET_STAT * sizeof(ix_cc_csix_rx_cb_ctx_entry))



/* various size definitions for 32bit and 64-bit counter support */
#define CSIX_RX_NUM_COUNTERS_PER_VOQ    4   /*enum defined in main header file*/
#define CSIX_RX_TOTAL_NUM_COUNTERS                      (CC_CSIX_RX_NUM_VOQS * CSIX_RX_NUM_COUNTERS_PER_VOQ)
#define CSIX_RX_SIZE_32BIT_COUNTER_BUF                  (CSIX_RX_TOTAL_NUM_COUNTERS * sizeof(ix_uint32))
#define CSIX_RX_SIZE_64BIT_COUNTER_HDL                  (CSIX_RX_TOTAL_NUM_COUNTERS * sizeof(ix_counter_64bit_handle))



/**
 * Callback context entry structure
 */
typedef struct s_ix_cc_csix_rx_cb_ctx_entry
{
    ix_uint8 flags;     /* entry flag */
    void *pUserCtx;     /* user context */
    void *pUserCb;      /* user callback */
} ix_cc_csix_rx_cb_ctx_entry;



/* values for flags field of the callback context entry structure */
#define CC_CSIX_RX_ENTRY_FLAGS_NOT_USED     0x00 /* free entry */
#define CC_CSIX_RX_ENTRY_FLAGS_USED_STATIC  0x01 /* used static entry */
#define CC_CSIX_RX_ENTRY_FLAGS_USED_DYNAMIC 0x02 /* used dynamic entry */



/**
 * Callback context entry table structure
 */
typedef struct ix_s_cc_csix_rx_cb_ctx_tbl
{
    ix_uint32 index;                 /* Array index pointer */
    ix_cc_csix_rx_cb_ctx_entry *pTable; /* free cb context entry array */ 
} ix_cc_csix_rx_cb_ctx_tbl;



/**
 * Control block structure
 */
typedef struct ix_s_cc_csix_rx_ctl_blk
{

    /**
     * MEv2 number for Csix Rx microblock.
     */
    ix_uint32 mev2Num;

    /**
     * Csix Rx microblock Context memory.
     */
    ix_uint32 *pRxCtx;

    /**
     * Base address of 64-bit counter handle memory. This is the
	 * based address to an array of counter handles
     */
    ix_counter_64bit_handle *pCounterHandlers64Bit;

    /**
     * Base address of 32-bit counter memory.
     */
    ix_uint32 *pCounterBase32Bit;

    /**
     * Free callback context entry table for each async helper API.
     */
    ix_cc_csix_rx_cb_ctx_tbl *pCbCtxTblGetStatInfo;

    /**
     * Mutex id used by the component in critical sections.
     */
    ix_ossl_mutex_t mid;

} ix_cc_csix_rx_ctl_blk;



/**
 * MACRO NAME: IX_CC_CSIX_RX_ALLOC_ENTRY
 *
 * DESCRIPTION: This macro allocates a user specified structure.
 *              The structure is either allocated from the free entry array
 *              passed in by the user or from OSSL if one is not 
 *              available from the array.
 *
 * @Param: arg_pTbl - IN: pointer to the free entry array.
 * @Param: arg_CurIndex - IN: current index to the array.
 * @Param: arg_TblSize - IN: size of the table (MUST be a power of 2)
 * @Param: arg_Type - IN: data type of the array element.
 *
 * @Param: arg_pEntry - OUT: pointer to the free entry.
 *
 * @Return:  - None
 */
#define IX_CC_CSIX_RX_ALLOC_ENTRY(                                   \
    arg_pTbl,                                                        \
    arg_CurIndex,                                                    \
    arg_TblSize,                                                     \
    arg_Type,                                                        \
    arg_pEntry)                                                      \
{                                                                    \
                                                                     \
    arg_pEntry = arg_pTbl + arg_CurIndex;                            \
                                                                     \
    if (arg_pEntry->flags == CC_CSIX_RX_ENTRY_FLAGS_NOT_USED)        \
    {                                                                \
        arg_pEntry->flags = CC_CSIX_RX_ENTRY_FLAGS_USED_STATIC;      \
        arg_CurIndex = (arg_CurIndex + 1) & (arg_TblSize - 1);       \
    }                                                                \
    else                                                             \
    {                                                                \
        if ((arg_pEntry = (arg_Type *)ix_ossl_malloc(sizeof(arg_Type))) != NULL) \
        {                                                            \
            arg_pEntry->flags = CC_CSIX_RX_ENTRY_FLAGS_USED_DYNAMIC; \
        }                                                            \
    }                                                                \
}



/**
 * MACRO NAME: IX_CC_CSIX_RX_FREE_ENTRY
 *
 * DESCRIPTION: This macro frees a structure.
 *              The structure may be freed back to system memory if 
 *              was allocated from it.
 *
 * @Param: arg_pEntry - IN: pointer to the entry to be freed.
 *
 * @Return:  - None
 */
#define IX_CC_CSIX_RX_FREE_ENTRY(                                      \
    arg_pEntry)                                                        \
{                                                                      \
                                                                       \
    if (arg_pEntry->flags == CC_CSIX_RX_ENTRY_FLAGS_USED_STATIC)       \
    {                                                                  \
        arg_pEntry->flags = CC_CSIX_RX_ENTRY_FLAGS_NOT_USED;           \
    }                                                                  \
    else if (arg_pEntry->flags == CC_CSIX_RX_ENTRY_FLAGS_USED_DYNAMIC) \
    {                                                                  \
        ix_ossl_free(arg_pEntry);                                      \
    }                                                                  \
}



#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__IX_CC_CSIX_RX_INT_H__) */
