/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2X00 Network Processor, Release 5
 *
 * = FILENAME
 *      ix_cc_csix_rx_core.c
 *
 * = DESCRIPTION
 *      This file contains Library APIs and related 
 *      internal functions of CSIX Rx core component.
 *
 * = CHANGE HISTORY
 *      8/21/2002 - Created.
 *
 * ============================================================================
 * $Id: ix_cc_csix_rx_core.c,v 1.5 2002/10/17 19:38:55 sharma Exp $
 */

/**
 * System defined include files required.
 */


/**
 * User defined include files required.
 */
#include "ix_cc_error.h"

#include "ix_types.h"
#include "ix_error.h"
#include "ix_cc.h"
#include "bindings.h"
#include "ix_cc_properties.h"

#include "ix_rm.h"
#include "ix_ossl.h"

#ifndef IX_EXCLUDE_CCI
#include "ix_cci.h"
#endif

#include "cc/ix_cc_csix_rx.h"
#include "cc/internal/ix_cc_csix_rx_int.h"



/**
 * Variable declarations global to this file only.  Extern are followed
 * by static variables.
 */



/**
 * Pre-processor symbols and macros used in this file.
 */



/**
 * Extern function prototypes
 */


/**
 * Static function prototypes
 */



/**
 * function definitions.
 */



/**
 * NAME: ix_cc_csix_rx_get_statistics_info
 * 
 * DESCRIPTION: This function retrieves a statistics counter value of 
 *              Csix Rx CC.
 * 
 * @Param: arg_Entity - IN This is the type of counter that is being
 *						queried
 * @Param: arg_Index - IN This is the particular queue for which the 
 *					   counter/s is/are being requested.
 * @Param: arg_pBuffer - IN
 * @Param: arg_pContext - IN: pointer to the component context allocated
 *                            in the component's init function. 
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_cc_csix_rx_get_statistics_info(
    ix_cc_csix_rx_statistics_info arg_Entity,
    ix_uint32 arg_Index,
    ix_cc_csix_rx_statistics_info_data *arg_pBuffer,
    void *arg_pContext)
{

    ix_counter_64bit_handle *pH64Bit;
    ix_error err = IX_SUCCESS;
    ix_uint32 numCounters;
    ix_cc_csix_rx_ctl_blk *pCcContext;
    ix_uint32 i;


    /* Check any NULL pointer argument */
    if ((arg_pBuffer == (ix_cc_csix_rx_statistics_info_data *)NULL) ||
        (arg_pContext == NULL))
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                              ("NULL argument")));
    }

    /*
     * Check the ranges of the arg_Entity and arg_Index.
     * where arg_Index stores the VoQ number and arg_Entity stores 
     * the counter number.
     */
    if ((arg_Entity >= IX_CC_CSIX_RX_COUNTER_LAST) || 
        (arg_Index >= CC_CSIX_RX_NUM_VOQS))
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_RANGE, 
                              ("Invalid argument range")));
    }

    /* Cast the input context number to pointer to cc context */
    pCcContext = (ix_cc_csix_rx_ctl_blk *)arg_pContext;

    /*
     * Calculate the base address of the counter handle array that holds
     * counter handles pertinent to this VoQ based on VoQ number 
     * indicated by arg_Index.
     */
    pH64Bit = pCcContext->pCounterHandlers64Bit + (arg_Index * CSIX_RX_NUM_COUNTERS_PER_VOQ);


    /* Determine how many counter are to be read based on the entity type */
    if (arg_Entity == IX_CC_CSIX_RX_ALL_COUNTERS)
    { 
        /* All counters of the specified VOQ are to be returned */
        numCounters = CSIX_RX_NUM_COUNTERS_PER_VOQ;
    }
    else
    {
        /* Only the specified VOQ counter is to be returned */
        numCounters = 1;

        /*
         * set the handler pointer to the associated location from the 
         * base using the counter number indicated by the arg_entity.
         */
        pH64Bit = pH64Bit + arg_Entity;
    }

    /*
     * Enter critical section.  It is possible that the mutex can not
     * be obtained if component has been destroyed and the mutex was 
     * freed by the fini function.)
     */
    IX_ERROR_CRT(ix_ossl_mutex_lock(pCcContext->mid,
                                    IX_OSSL_WAIT_FOREVER),
                 IX_CC_CSIX_RX_ERROR_OSSL_MUTEX_LOCK, 
                 IX_ERROR_LEVEL_WARNING);

    /* Call RM to read the counter value */
    for (i = 0; i < numCounters; i++)
    {
        /* Call RM to retrieve the 64-bit counter value */
        IX_ERROR_CT(ix_rm_counter_64bit_get_value(*(pH64Bit + i),
                                                  (ix_uint64 *)(arg_pBuffer->pDataBuffer) + i),
                    err,
                    IX_CC_CSIX_RX_ERROR_RM_GET_64BIT_COUNTER, 
                    IX_ERROR_LEVEL_WARNING);
        if (err != IX_SUCCESS)
        {
            break;
        }
    }

    /* Exit the critical section */
    ix_ossl_mutex_unlock(pCcContext->mid);

    return(err);

} /* ix_cc_csix_rx_get_statistics_info */


