/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2X00 Network Processor, Release 5
 *
 * = FILENAME
 *      ix_cc_pos_core.c
 *
 * = DESCRIPTION
 *      This file contains level-0 only interface APIs and related 
 *      internal functions of POS Rx core component.
 *
 * 
 *
 * = CHANGE HISTORY
 *      7/20/2002 - Created.
 *
 * ============================================================================
 * $Id: ix_cc_pos_rx_core.c,v 1.16 2002/10/17 19:28:37 sharma Exp $
 */

/**
 * System defined include files required.
 */

/**
 * User defined include files required.
 */
#include "ix_cc_error.h"
#include "ix_error.h"

#include "ix_cc.h"
#include "bindings.h"
#include "ix_cc_properties.h"
#include "ix_rm.h"
#include "ix_ossl.h"

#include "cc/ix_cc_pos_rx.h"
#include "cc/internal/ix_cc_pos_rx_int.h"

/**
 * Variable declarations global to this file only.  Extern are followed
 * by static variables.
 */


/**
 * Pre-processor symbols and macros used in this file.
 */



/**
 * Extern function prototypes
 */


/**
 * Static function prototypes 
 */
PRIVATE ix_error _ix_cc_pos_rx_get_channel_number(ix_cc_pos_rx_ctl_blk *, ix_uint32 , ix_uint32 *);


/**
 * function definitions.
 */



/**
 * NAME: ix_cc_pos_rx_get_statistics_info
 * 
 * DESCRIPTION: This function retrieves statistics counter/s value of 
 *              POS Rx CC.
 * 
 * @Param: arg_Entity - IN
 * @Param: arg_Index - IN
 * @Param: arg_pBuffer - IN
 * @Param: arg_pContext - IN pointer to the component context allocated
 *                            in the component's init function. 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_cc_pos_rx_get_statistics_info(
    ix_cc_pos_rx_statistics_info arg_Entity,
    ix_uint32 arg_Index,
    ix_cc_pos_rx_statistics_info_data *arg_pBuffer,
	void *arg_pContext)
{
    ix_uint32 numCounters, i;
    ix_counter_64bit_handle *pH64Bit;
	ix_uint32 chNum;
    ix_error err = IX_SUCCESS;
    ix_cc_pos_rx_ctl_blk *pCcContext;
	

    /* Check any NULL pointer argument for arg_pBuffer */
    if (arg_pBuffer == (ix_cc_pos_rx_statistics_info_data *)NULL)
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                              ("NULL argument for buffer")));
    }

	/* Check any NULL pointer argument for arg_pContext*/
    if (arg_pContext == NULL)
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                              ("NULL argument for context")));
    }

    /*
     * Check the ranges of the arg_Entity
     * where arg_Entity stores the counter number. Port check??
     */
    if (arg_Entity >= IX_CC_POS_RX_COUNTER_LAST)
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_RANGE, 
                              ("Invalid argument range for Counter Type")));
    }

	/*Cast the input context to pointer to cc context */
    pCcContext = (ix_cc_pos_rx_ctl_blk *)arg_pContext;


	/*check to see if the portId is valid by searching through the array of valid channels*/
	/* Get the channel number based on the port id in arg_Index */
    IX_ERROR_CR(_ix_cc_pos_rx_get_channel_number(pCcContext,
                                                  arg_Index, 
                                                  &chNum));
	
   
    /*
     * Calculate the base address of the counter handle array that holds
     * counter handles pertinent to this channel based on port number 
     * indicated by arg_Index.
     */
    pH64Bit = pCcContext->pCounterHandlers64Bit + (chNum * POS_RX_NUM_COUNTERS_PER_PORT);


    /* Determine how many counter are to be read based on the entity type */
    if (arg_Entity == IX_CC_POS_RX_ALL_COUNTERS)
    { 
        /* All counters of the specified port are to be returned */
        numCounters = POS_RX_NUM_COUNTERS_PER_PORT;
    }
    else
    {
        /* Only the specified VOQ counter is to be returned */
        numCounters = 1;

        /*
         * set the handler pointer to the associated location from the 
         * base using the counter number indicated by the arg_entity.
         */
        pH64Bit = pH64Bit + arg_Entity;
    }
	
	/*
     * Enter critical section.  It is possible that the mutex can not
     * be obtained if component has been destroyed and the mutex was 
     * freed by the fini function.)
     */
    IX_ERROR_CRT(ix_ossl_mutex_lock(pCcContext->mid,
                                    IX_OSSL_WAIT_FOREVER),
                 IX_CC_POS_RX_ERROR_OSSL_MUTEX_LOCK, 
                 IX_ERROR_LEVEL_WARNING);
    
	/* Call RM to read the counter value */
    for (i = 0; i < numCounters; i++)
    {
        /* Call RM to retrieve the 64-bit counter value */
        IX_ERROR_CGT(ix_rm_counter_64bit_get_value(*(pH64Bit + i),
                                                   (ix_uint64 *)(arg_pBuffer->pDataBuffer) + i),
                     err,
                     IX_CC_POS_RX_ERROR_RM_GET_64BIT_COUNTER, 
                     IX_ERROR_LEVEL_WARNING,
                     cleanup);
    }


cleanup:

    /* Exit the critical section */
    ix_ossl_mutex_unlock(pCcContext->mid);

    return(err);

} /* ix_cc_pos_rx_get_statistics_info */


/**
 * NAME: ix_cc_pos_rx_get_interface_state
 * 
 * DESCRIPTION: This function retrieves the link status of a POS
 *              interface port.
 * 
 * @Param: arg_PortId - IN: port id
 * @Param: arg_pIfState - OUT: location where the interface state will be 
 *                            stored.
 * @Param: arg_pContext - IN: pointer to the component context allocated
 *                            in the component's init function. 
 * 
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_cc_pos_rx_get_interface_state(
    ix_uint32 arg_PortId, 
    ix_cc_pos_rx_if_state *arg_pIfState,
    void *arg_pContext)
{

    ix_uint32 chNum;
    ix_cc_pos_rx_ctl_blk *pCcContext;



    /* Check any NULL pointer argument */
    if ((arg_pIfState == (ix_cc_pos_rx_if_state *)NULL) ||
        (arg_pContext == (void *)NULL))
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                              ("NULL argument")));
    }

    /* Cast the input context to pointer to cc context */
    pCcContext = (ix_cc_pos_rx_ctl_blk *)arg_pContext;

    /* Get the channel number based on the port id */
    IX_ERROR_CR(_ix_cc_pos_rx_get_channel_number(pCcContext,
                                                  arg_PortId, 
                                                  &chNum));

    *arg_pIfState = pCcContext->portState[chNum];

    return(IX_SUCCESS);

} /* ix_cc_pos_rx_get_interface_state */





/**
 * NAME: ix_cc_pos_rx_set_property
 * 
 * DESCRIPTION: This function sets the properties (ony sets the IF_STATE property 
 * (but can be extended).
 * 
 * @Param: arg_PropId - IN: property id
 * @Param: arg_pProperty - IN: This is pointer to the property structure passed in
 * @Param: arg_pContext - IN: pointer to the component context allocated
 *                            in the component's init function. 
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_cc_pos_rx_set_property(
    ix_uint32 arg_PropId,
    ix_cc_properties *arg_pProperty,
    void *arg_pContext)
{

    ix_uint32 chNum;
	ix_cc_pos_rx_ctl_blk *pCcContext;


    /* Check any NULL pointer argument */
    if (arg_pProperty == (ix_cc_properties *)NULL)
	{
        return(IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                              ("NULL argument - pointer to property")));
    }	
		
		
		
    if (arg_pContext == NULL)
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                              ("NULL argument - pointer to context")));
    }

    /* Check the property type. Currently we support only the IF_STATUS */
	/*IX_CC_SET_PROPERTY_PHYSICAL_IF_STATUS is a bit mask*/
    if ((arg_PropId & IX_CC_SET_PROPERTY_PHYSICAL_IF_STATUS) != IX_CC_SET_PROPERTY_PHYSICAL_IF_STATUS)
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_UNIMPL,
                              ("Indicated property is not supported.")));
    }

    /* Cast the input context  to pointer to cc context */
    pCcContext = (ix_cc_pos_rx_ctl_blk *)arg_pContext;

    /* Get the channel number based on the port id */
    IX_ERROR_CR(_ix_cc_pos_rx_get_channel_number(pCcContext,
                                                  arg_pProperty->port_id, 
                                                  &chNum));

    /*
     * Enter critical section.  It is possible that the mutex can not 
     * be obtained if component has been destroyed and the mutex was 
     * freed by the fini function.)
     */
    IX_ERROR_CRT(ix_ossl_mutex_lock(pCcContext->mid, IX_OSSL_WAIT_FOREVER),
                 IX_CC_POS_RX_ERROR_OSSL_MUTEX_LOCK, 
                 IX_ERROR_LEVEL_WARNING);

    /* Check for desired state value and set the status of the port */
    if (arg_pProperty->physical_if_state == IX_CC_PHYSICAL_IF_STATUS_UP) 
	{
		pCcContext->portState[chNum]= IX_CC_POS_RX_IF_STATE_UP;

	}
	else if(arg_pProperty->physical_if_state == IX_CC_PHYSICAL_IF_STATUS_DOWN)
    {
        /*Update the port/channel status flag */
        pCcContext->portState[chNum] = IX_CC_POS_RX_IF_STATE_DOWN;
    }
    else 
    {
		/* Exit the critical section */	
		ix_ossl_mutex_unlock(pCcContext->mid);
        return(IX_ERROR_WARNING(IX_CC_POS_RX_ERROR_INVALID_IF_STATE,
                                  ("Invalid value for interface state.")));
    }

    
    /* Exit the critical section */
    ix_ossl_mutex_unlock(pCcContext->mid);

    return(IX_SUCCESS);

} /* ix_cc_pos_rx_set_property */



/**
 * NAME: _ix_cc_pos_rx_get_channel_number
 * 
 * DESCRIPTION: This function retrieves channel number based on the port id.
 *              
 * 
 * @Param: arg_pCtlBlk - IN: the component control block structure
 * @Param: arg_PortId - IN: the Port Id for which we need the channel number
 * @Param: arg_pBuffer - OUT: location where channel num is to be stored.
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
PRIVATE ix_error _ix_cc_pos_rx_get_channel_number(
    ix_cc_pos_rx_ctl_blk *arg_pCtlBlk, 
    ix_uint32 arg_PortId, 
    ix_uint32 *arg_ChNum)
{

    ix_uint32 i;


    for (i = 0; i < arg_pCtlBlk->numChannels; i++)
    {
        if (arg_pCtlBlk->portId[i] == arg_PortId)
        {
            *arg_ChNum = i;
            return(IX_SUCCESS);
        }
    }

    return(IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                          ("No channel found for the specifed port id.")));
}
