/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2X00 Network Processor, Release 5
 *
 * = FILENAME
 *      ix_cc_pos_mhlp.c
 *
 * = DESCRIPTION
 *      This file contains level-1 only messaging interface APIs and 
 *      related internal async callback functions of Pos Rx CC.
 *
 *
 * = CHANGE HISTORY
 *      7/18/2002 - Created.
 *
 * ============================================================================
 * $Id: ix_cc_pos_rx_mhlp.c,v 1.13 2002/10/17 19:28:37 sharma Exp $
 */

/**
 * System defined include files required.
 */

/**
 * User defined include files required.
 */
#include "ix_cc_error.h"
#include "ix_types.h"
#include "ix_error.h"
#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_cc.h"

#include "cc/ix_cc_msup.h"
#include "cc/ix_cc_pos_rx.h"
#include "cc/internal/ix_cc_pos_rx_int.h"
#include "bindings.h"




/**
 * Variable declarations global to this file only.  Extern are followed
 * by static variables.
 */
extern ix_cc_pos_rx_ctl_blk *g_pCcPosRxCtlBlk; /* control block */



/**
 * Pre-processor symbols and macros used in this file.
 */



/**
 * Extern function prototypes and variables
 */



/**
 * Static function prototypes 
 */
PRIVATE ix_error _ix_cc_pos_rx_icb_get_statistics_info(ix_error, void *, void *, ix_uint32);
PRIVATE ix_error _ix_cc_pos_rx_icb_get_interface_state(ix_error, void *, void *, ix_uint32);


/**
 * function definitions.
 */



/**
 * NAME: ix_cc_pos_rx_async_get_statistics_info
 * 
 * DESCRIPTION: This function retrieves statistics (counters) information of 
 *              the Pos Rx Core Component.
 * 
 * @Param: arg_pMgInfoContext - IN
 * @Param: arg_Callback - IN user callback, we define the signature, user implements it, 
 * 			we call it when we are ready with the reply
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_cc_pos_rx_async_get_statistics_info(
    ix_cc_pos_rx_statistics_info_context *arg_pMgInfoCtx, 
    ix_cc_pos_rx_cb_get_statistics_info arg_Callback)
{

    ix_cc_pos_rx_cb_ctx_entry *entry;
    ix_error err = IX_SUCCESS;


    /* Check any NULL pointer argument */
    if ((arg_pMgInfoCtx == NULL) || (arg_Callback == NULL))
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                              ("NULL argument")));
    }

    /*
     * Check the ranges of the arg_Entity, where arg_Entity stores 
     * the counter number. ??Port range check
     */
    if (arg_pMgInfoCtx->entity >= IX_CC_POS_RX_COUNTER_LAST)
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_RANGE, 
                              ("Invalid counter type")));
    }


    /*
     * Enter critical section.  It is possible that the mutex can not 
     * be obtained if component has been destroyed and the mutex was 
     * freed by the fini function.)
     */

    
    IX_ERROR_CRT(ix_ossl_mutex_lock(g_pCcPosRxCtlBlk->mid, IX_OSSL_WAIT_FOREVER),
                 IX_CC_POS_RX_ERROR_OSSL_MUTEX_LOCK, 
                 IX_ERROR_LEVEL_WARNING);
     
    /*
     * Get a free callback context entry.
     */
    IX_CC_POS_RX_ALLOC_ENTRY(g_pCcPosRxCtlBlk->pCbCtxTblGetStatInfo->pTable, 
                              g_pCcPosRxCtlBlk->pCbCtxTblGetStatInfo->index, 
                              POS_RX_NUM_ENTRIES_CB_CTX_TBL_GET_STAT,
                              ix_cc_pos_rx_cb_ctx_entry,
                              entry);

    if (entry == NULL)
    {
		        /* Exit critical section */
        ix_ossl_mutex_unlock(g_pCcPosRxCtlBlk->mid);
        return(IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
                               ("Out of memory for callback context entry.")));
    }

    /* Store the user callback and context into the entry */
    entry->pUserCtx = arg_pMgInfoCtx->pUserContext;
    entry->pUserCb = (void *)arg_Callback;;

    /* Call msup to send the message */ 
    IX_ERROR_CGT(ix_cc_msup_send_async_msg(IX_CC_POS_RX_MSG_INPUT,
                                           _ix_cc_pos_rx_icb_get_statistics_info,
                                           (void *)entry, 
                                           IX_CC_POS_RX_MSG_GET_STATISTICS_INFO,
                                           (void *)arg_pMgInfoCtx,
                                           sizeof(ix_cc_pos_rx_statistics_info_context)),
                 err,
                 IX_CC_POS_RX_ERROR_MSUP_SEND_MSG, 
                 IX_ERROR_LEVEL_WARNING,
                 cleanup);


cleanup:

    /* if the result is not success then we need to free the entry */ 
    if (err != IX_SUCCESS)
    {
        if (entry != NULL)
        {
            IX_CC_POS_RX_FREE_ENTRY(entry);
        }
    }

    /* Exit critical section */
    ix_ossl_mutex_unlock(g_pCcPosRxCtlBlk->mid);


    return(err);

} /* ix_cc_pos_rx_async_get_statistics_info */



/**
 * NAME: ix_cc_pos_rx_async_get_interface_state
 * 
 * DESCRIPTION: This function retrieves the interface state of a port.
 * 
 * @Param: arg_pMgInfoContext - IN
 * @Param: arg_Callback - IN
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_cc_pos_rx_async_get_interface_state(
    ix_cc_pos_rx_if_state_context *arg_pStateContext, 
    ix_cc_pos_rx_cb_get_interface_state arg_Callback)
{

    ix_cc_pos_rx_cb_ctx_entry *entry;
    ix_error err = IX_SUCCESS;



    /* Check any NULL pointer argument */
    if ((arg_pStateContext == NULL) || (arg_Callback == NULL))
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                              ("NULL argument")));
    }

    /*
     * Enter critical section.  It is possible that the mutex can not 
     * be obtained if component has been destroyed and the mutex was 
     * freed by the fini function.)
     */
    IX_ERROR_CRT(ix_ossl_mutex_lock(g_pCcPosRxCtlBlk->mid, 
                                    IX_OSSL_WAIT_FOREVER),
                 IX_CC_POS_RX_ERROR_OSSL_MUTEX_LOCK, 
                 IX_ERROR_LEVEL_WARNING);

    /*
     * Get a free callback context entry.
     */
    IX_CC_POS_RX_ALLOC_ENTRY(g_pCcPosRxCtlBlk->pCbCtxTblGetIfState->pTable, 
                              g_pCcPosRxCtlBlk->pCbCtxTblGetIfState->index, 
                              POS_RX_NUM_ENTRIES_CB_CTX_TBL_GET_IF,
                              ix_cc_pos_rx_cb_ctx_entry,
                              entry);

    if (entry == NULL)
    {
        /* Exit critical section */
        ix_ossl_mutex_unlock(g_pCcPosRxCtlBlk->mid);

        return(IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
                               ("Out of memory for callback context entry")));
    }

    /* Store the user callback and context into the entry */
    entry->pUserCtx = arg_pStateContext->pUserContext;
    entry->pUserCb = (void *)arg_Callback;;

    /* Call msup to send the message */
    IX_ERROR_CT(ix_cc_msup_send_async_msg(IX_CC_POS_RX_MSG_INPUT,
                                          _ix_cc_pos_rx_icb_get_interface_state,
                                          (void *)entry,
                                          IX_CC_POS_RX_MSG_GET_INTERFACE_STATE,
                                          (void *)arg_pStateContext,
                                          sizeof(ix_cc_pos_rx_if_state_context)),
                err,
                IX_CC_POS_RX_ERROR_MSUP_SEND_MSG, 
                IX_ERROR_LEVEL_WARNING);

    if (err != IX_SUCCESS)
    {
        IX_CC_POS_RX_FREE_ENTRY(entry);
    }

    /* Exit critical section */
    ix_ossl_mutex_unlock(g_pCcPosRxCtlBlk->mid);

    return(err);

} /* ix_cc_pos_rx_async_get_interface_state */










/**********************
 * Internal functions *
 **********************/


/**
 * NAME: _ix_cc_pos_rx_icb_get_statistics_info
 * 
 * DESCRIPTION: This function is an internal callback for the async 
 *              messaging API, ix_cc_pos_rx_async_get_statistics_info().
 * 
 * @Param: arg_Result - IN
 * @Param: arg_pContext - IN
 * @Param: arg_Msg - IN
 * @Param: arg_MsgLen - IN
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
PRIVATE ix_error _ix_cc_pos_rx_icb_get_statistics_info(
    ix_error arg_Result,
    void *arg_pContext,
    void *arg_Msg,
    ix_uint32 arg_MsgLen)
{
    ix_cc_pos_rx_cb_get_statistics_info userCb;
    void *userCtx;

    /* Retrieve the user callback and context from the context pointer */
    userCb = ((ix_cc_pos_rx_cb_ctx_entry*) arg_pContext)->pUserCb;
    userCtx = ((ix_cc_pos_rx_cb_ctx_entry *)arg_pContext)->pUserCtx;

    /*
     * Enter critical section.  It is possible that the mutex can not 
     * be obtained if component has been destroyed and the mutex was 
     * freed by the fini function.
     */
    IX_ERROR_CRT(ix_ossl_mutex_lock(g_pCcPosRxCtlBlk->mid, IX_OSSL_WAIT_FOREVER),
                 IX_CC_POS_RX_ERROR_OSSL_MUTEX_LOCK, 
                 IX_ERROR_LEVEL_WARNING);

    /* Free the callback context entry */
    IX_CC_POS_RX_FREE_ENTRY((ix_cc_pos_rx_cb_ctx_entry *)arg_pContext);

    /* Exit the critical section */
    ix_ossl_mutex_unlock(g_pCcPosRxCtlBlk->mid);

    /* Invoke the user callback to pass back the result and data */
    IX_ERROR_CRT((userCb)(arg_Result, userCtx, (ix_uint64 *)arg_Msg, arg_MsgLen), 
		IX_CC_POS_RX_ERROR_USER_CB, IX_ERROR_LEVEL_WARNING);

    return(IX_SUCCESS);

} /* _ix_cc_pos_rx_icb_get_statistics_info */


/**
 * NAME: _ix_cc_pos_rx_icb_get_interface_state
 * 
 * DESCRIPTION: This function is an internal callback for the async 
 *              messaging API, ix_cc_pos_rx_async_get_interface_state().
 * 
 * @Param: arg_Result - IN
 * @Param: arg_pContext - IN
 * @Param: arg_Msg - IN
 * @Param: arg_MsgLen - IN
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error _ix_cc_pos_rx_icb_get_interface_state(
    ix_error arg_Result,
    void *arg_pContext,
    void *arg_Msg,
    ix_uint32 arg_MsgLen)
{

    ix_cc_pos_rx_cb_get_interface_state userCb;
    void *userCtx;



    /* Retrieve the user callback and context from the context pointer */
    userCb = ((ix_cc_pos_rx_cb_ctx_entry *)arg_pContext)->pUserCb;
    userCtx = ((ix_cc_pos_rx_cb_ctx_entry *)arg_pContext)->pUserCtx;

    /*
     * Enter critical section.  It is possible that the mutex can not 
     * be obtained if component has been destroyed and the mutex was 
     * freed by the fini function.
     */
    IX_ERROR_CRT(ix_ossl_mutex_lock(g_pCcPosRxCtlBlk->mid, 
                                    IX_OSSL_WAIT_FOREVER),
                 IX_CC_POS_RX_ERROR_OSSL_MUTEX_LOCK, 
                 IX_ERROR_LEVEL_WARNING);

    /* Free the callback context entry */
    IX_CC_POS_RX_FREE_ENTRY(((ix_cc_pos_rx_cb_ctx_entry *)arg_pContext));

    /* Exit the critical section */
    ix_ossl_mutex_unlock(g_pCcPosRxCtlBlk->mid);

    /* Invoke the user callback to pass back the result and data */
    IX_ERROR_CRT((userCb)(arg_Result, 
                          userCtx, 
                          (ix_cc_pos_rx_if_state *)arg_Msg),
                 IX_CC_POS_RX_ERROR_USER_CB, 
                 IX_ERROR_LEVEL_WARNING);

    return(IX_SUCCESS);

} /* _ix_cc_pos_rx_icb_get_statistics_info */
