#ifndef	_AAL5_RX_UC_
#define	_AAL5_RX_UC_

/*******************************************************************************
                             Intel Proprietary

 Copyright (c) 1998-2002 By Intel Corporation.  All rights reserved.  
 No part of this program or publication may be reproduced, transmited,
 transcribed, stored in a retrieval system, or translated into any language
 or computer language in any form or by any means, electronic, mechanical,
 magnetic, optical, chemical, manual, or otherwise, without the prior
 written permission of:
                         Intel Corporation
                         2200 Mission College Blvd.
                         Santa Clara, CA  95052-8119
*******************************************************************************/

/*
 *
 *      File Name: 	aal5_rx.uc
 *
 *      Purpose: 	This is the main file for AAL5 RX microblock.
 *					It has the main aal5_rx() macro which does the AAL5 reassembly 
 *					function.
 *
 *					This basically gets the cell from MSF, processes the cell,
 *					computes CRC, updates the meta data in SRAM, writes cell to 
 *					DRAM. In the process, it does necessary error handling.
 *					Finally, if its a valid packet, on EOP cell, AAL5 RX block, 
 *					using dl_sink()	passes the required data to the next block 
 *					in the pipeline (here IPv4 microblock) via scratch ring.
 *					
 *
 *      History:
 *
 *
 *      Date            Comment                         By
 *      ------------------------------------------------------------------------
 *
 *		04/16/2002		Created							
 *      06/20/2003		Merged code with ITP		
 */


////////////////////////////////////////////////////////////////////////////////
//	This file is the main file for AAL5 RX block.
////////////////////////////////////////////////////////////////////////////////



;-----------------------------------------------------------------------------
;                  				INCLUDE FILES 
;-----------------------------------------------------------------------------


// Header file with MSF configuration constants and aal5 rx constants.
#include <aal5_rx.h>

// File with aal5 rx init.
#include <aal5_rx_init.uc>

// File with macros internal to aal5 rx microblock.
#include <aal5_rx_util.uc>


#ifdef ONE_ME_RX_FP

	#include "aal5_rx_one_me_fp_util.uc"

#else  // TWO_ME_RX_FP

	#include "aal5_rx_two_me_fp_util.uc"

#endif 

#include "aal5_rx_oam.uc"


#ifdef DO_SYSTEM_INIT

    // File with system init macros.
    #include "system_init.uc"

#endif

// File with standard IXP macros.
#include <stdmac.uc>

#include <dispatch_loop.uc>

#ifndef USE_HASH_LOOKUP
    #ifndef SIMPLE_LOOKUP
        #error "You must define either USE_HASH_LOOKUP or SIMPLE_LOOKUP !!!"
    #endif
#endif

#ifndef OPTIONAL_L2PORT
    #ifdef OPTIONAL_SUID
        #define USE_OPTIONAL_INFO    
    #endif
#else
    #define USE_OPTIONAL_INFO    
#endif

;-----------------------------------------------------------------------------
; Global register declaration, initialized during aal5_rx_init[]
;-----------------------------------------------------------------------------

.reg @rand_buf_offset 		// starting random buffer offset

.reg vc_info_sram_base		// base of VC INFO Table (has both RXC & VC STAT counters)
.reg rbuf_base				// base address of rbuf
.reg hash_pri_base_sram		// base of primary hash table
.reg hash_sec_base_sram		// base of secondary hash table
.reg port_stat_base			// base for port statistics counters

.reg mask_lower_half		// mask=0xffff (cannot place this value in LM, as
							// you cannot have two operands in LM in an instr)

.reg rx_thd_freelist_reg	// value used to put thread onto thread_freelist

.reg rx_freelist_no			// RX_THREAD_FREELIST number (we have [0,3] freelist)
.reg next_thread_sig_val	// The value to signal the next thread
.reg ctx_num				// number of current context
.reg clp                    // Cell Loss Priority bit
.reg FPGA_port              // FPGA port number
.reg zero                   // register with 0 value used because alu() can't work with two constants
.reg l2port                 // L2Port - initialized to 0
.reg suid                   // SUID - initialized to 0

;-----------------------------------------------------------------------------
; Global Register declaration
;-----------------------------------------------------------------------------

.reg volatile read $rsw0, $rsw1	// to read RBUF receive status word used 
.xfer_order $rsw0, $rsw1		// in aal5_rx_init[] to set rx_thd_freelist_reg

.reg read $prefetch_buf_handle	// has prefetched buf_handle


;-----------------------------------------------------------------------------
; Signal declaration
;-----------------------------------------------------------------------------

// Signal received when RSW is autopushed.
.sig volatile rsw_rcvd_sig			


// Inter context signal, volatile required as it is used for inter context 
// signalling.
.sig volatile next_thread_sig
.addr next_thread_sig NEXT_THREAD_SIGNAL

.sig visible sys_init_signal
#ifdef USE_ME_INIT_SIGNAL
.addr sys_init_signal ME_INIT_SIGNAL
#endif


#ifdef TWO_ME_RX_FP
	// AAL5 RX Inter ME signals, volatile required as they are used for 
	// inter ME signalling.

	// For entering critical section 1 - access of RXC1.
	.sig	volatile  enter_cs1_inter_me_sig
	.addr	enter_cs1_inter_me_sig 	ENTER_CS1_INTER_ME_SIGNAL

	// For entering critical section 2 - access of RXC2.
	.sig	volatile  enter_cs2_inter_me_sig
	.addr	enter_cs2_inter_me_sig 	ENTER_CS2_INTER_ME_SIGNAL

	// For entering critical section 3 - dl_sink[] and putting onto
	// thread freelist.
	.sig 	volatile enter_cs3_inter_me_sig
	.addr enter_cs3_inter_me_sig ENTER_CS3_INTER_ME_SIGNAL

#endif   // TWO_ME_RX_FP



///////////////////////////////////////////////////////////////////////////////
// Main AAL5 RX macro
///////////////////////////////////////////////////////////////////////////////
//
//	AAL5_rx:
//
//		Description:
//			It does the main AAL5 reassembly function. 
//
//		Inputs: 
//			sig_mask:	This signal mask is used to set rsw_rcvd_sig after
//						thread puts itself onto the thread freelist.
//
//						The thread waits on this sig_mask in dl_sink() along
//						with scratch put(if any)
//
//		Outputs:
//			None:
//
//		Size: 
//			
///////////////////////////////////////////////////////////////////////////////
#macro aal5_rx(sig_mask)
.begin


;-----------------------------------------------------------------------------
; register declarations
;-----------------------------------------------------------------------------

.reg element_num		// rbuf_ele_num from RSW
.reg rbuf_offset		// rbuf_ele_address derived from rbuf_ele_num of RSW
.reg eop_bit			// PTI 3rd bit(EOP bit) from RSW 
.reg pti_3bits			// PTI 3 bits from RSW
.reg sop_bit			// has RXC:exp_SOP bit
.reg input_port			// channel_num from RSW

.reg sop_eop_bits		// combined sop bit:eop bit used for jump table

.reg vc_flowid			// the index into the VC_info/RXC table, range [1, 64K-1]

.reg vc_st_aal_type		// Combined VC_status (valid/not_valid) & AAL type 
						// supported by this particular VC

.reg vc_key				// 32 bit key consisting of vpi(8)/vci(16)/input_port(4)

.reg rx_sig_mask		// signal mask
.reg rxc_abs_addr		// absolute address of VC_info/RXC for a particular VC

.reg tmp


#ifdef FPGA_HEADER

.reg read 	$$FPGA_header $$cell_in[13]

.xfer_order $$FPGA_header $$cell_in

#else	//without FPGA_HEADER

	xbuf_alloc[$$cell_in, 13, read]	// dxfer reg to read in data from the RBUF element.

#endif



;-----------------------------------------------------------------------------
; signals declaration
;-----------------------------------------------------------------------------

.sig rbuf_rd_sig				// signal for reading RBUF into sxfer regs
.sig bd_prefetch_sig			// signal for prefetching buffer handle
.sig rxc_write_sig				// signal for writing back RXC to SRAM
.sig meta1_sig					// signal for writing META1 to SRAM
.sig meta2_sig					// signal for writing META2 to SRAM
.sig data_write_sig				// signal for writing data to DRAM

#ifdef DL_SPLIT_META_DATA

    .sig meta1_lw0_sig				// For splited meta data we have to wait for
    .sig meta2_lw0_sig				// additional signals.

#else

    #ifndef meta1_lw0_sig
        #define meta1_lw0_sig  --
        #define meta2_lw0_sig  --
    #endif
  
#endif


;-----------------------------------------------------------------------------
; Phase1#
;-----------------------------------------------------------------------------
	

// To get rid of warnings: $rsw0/$rsw1 used before being set.
.set $rsw0, $rsw1

	; Get element_num, rbuf_offset (from element_num)

	; Extract Element number.
	alu_shf[element_num, 0x7f, AND, $rsw0, >>24] 

	; Get RBUF address for the first byte in the element
#ifdef IXP2800
    /* For IXP2800 offset is calculated by shifting element num by 6 positions */
	alu_shf[rbuf_offset, --, B, element_num, <<6]	 
#else
	alu_shf[rbuf_offset, --, B, element_num, <<RBUF_ELE_SIZE_LOG2]	 
#endif





#ifdef NUM_OF_INPUTPORTS_16	

	// This is the phase 1 for the application where 16 input ports are supported
	// (ie 4 bits of port(channel) )
	_aal5_rx_phase1_inputports16();

#else // NUM_OF_INPUTPORTS_16 

	// This is the phase 1 for the application where we have more than 
	// 16 ports are supported.
	_aal5_rx_phase1();

#endif // NUM_OF_INPUTPORTS_16 



;-----------------------------------------------------------------------------
; Phase2#
;-----------------------------------------------------------------------------

	
#ifdef ONE_ME_RX_FP

	; Signal the next thread.
	signal_ctx(next_thread_sig_val)

	#ifndef AAL2_INCLUDED
		; We cannot free rbuf if the cell contains AAL2 but we don't know it yet
		; Free up the rbuf element (2).
		_aal5_rx_free_rbuf_element(element_num)
	#endif //AAL2_INCLUDED

	; Get RXC into LM (31 - cam_miss).
	_aal5_rx_get_rxc(rxc_abs_addr, vc_flowid)  


#else	// TWO_ME_RX_FP 

PHASE2#:

	#ifndef AAL2_INCLUDED
	; Free up the rbuf element (2).
	_aal5_rx_free_rbuf_element(element_num)

	#endif

	; Get RXC into LM (12 cycles - CAM_MISS).
	_aal5_rx_get_rxc_cs1(rxc_abs_addr, vc_flowid)

#endif  // TWO_ME_RX_FP


;-----------------------------------------------------------------------------
; Phase 2_1#:	OAM receive processing
;-----------------------------------------------------------------------------
	_aal5_rx_oam_pm_processing(rx_sig_mask, rxc_abs_addr, $$cell_in, input_port,\
							   vc_flowid, pti_3bits, lm_sop_alloc_valid_aal_flist,\
							   $prefetch_buf_handle, MGT_CELL#)


;-----------------------------------------------------------------------------
; for ONE_ME_RX_FP - Phase3# 
; for TWO_ME_RX_FP - Phase3#, Phase4#, Phase5#  
;-----------------------------------------------------------------------------

AAL_TYPE_CHK#:

#ifdef TWO_ME_RX_FP

	// Decrement the reference count by 1.
	alu[*l$index0[RXC_LM_REF_CNT_OFFSET], *l$index0[RXC_LM_REF_CNT_OFFSET], -, 1]

#endif


	; Combining three checks into one -

	; 1. Check whether this VC is valid or not (this is required as out of 64K, 
	;    some of them might be disabled) Valid VC state = 0.

	; 2. Check whether its an OAM cell or a user data cell.
	;    The 1st bit (management cell/user cell) of PTI.
	;    If it is set, then its a management cell

	; 3. Check whether AAL_type = AAL5 (0x000),

	; if yes branch out of the normal path.

	alu_shf[vc_st_aal_type, 0x7, AND, lm_sop_alloc_valid_aal_flist, >>20] 

#ifdef AAL2_INCLUDED

	alu[--, vc_st_aal_type, -, AAL_TYPE_AAL2] ; a valid AAL2 VC is coded as '0100'
	bne[NOT_AAL2_VC#]

	; Pass the cell to the ME executing the AAL0/AAL2 sublayers
	; Here the AAL5 ME acts as the ATM layer

#ifdef AAL2_RNC

    _aal5_rx_to_atm_rx($rsw0, $rsw1, lm_crc_residue, suid, NN_RING_TO_AAL2_FULL#)   

#else

	dl_aal5_to_aal2_sink(vc_flowid, vc_st_aal_type, $rsw0, $rsw1, \
							NN_RING_TO_AAL2_FULL#)
#endif

	; When the request is sent we should jump to the end of this loop
	br[NON_AAL5#]

NN_RING_TO_AAL2_FULL#:

	_aal5_rx_free_rbuf_element(element_num)
	br[NON_AAL5#]

    
NOT_AAL2_VC#:
	; AAL5 VC

	; Now we can free the element
	; Free up the rbuf element (2).
	_aal5_rx_free_rbuf_element(element_num)
	
	#endif //AAL2_INCLUDED

	alu_shf[--, vc_st_aal_type, OR, pti_3bits, >>2] 
	bne[INV_VC_MGT_CELL_NON_AAL5_ERR#] 


	; This counter counts number of valid cells(AAL5, including signalling cells) 
	; rcvd on a per vc_flow basis.
	_aal5_rx_incr_vc_counter[vc_flowid, AAL5_RX_CELL_CNT]

	; Do a check for SOP & EOP bits to determine one of the four separate 
	; code paths.
	; (1) SOP_EOP  (2) EOP_ONLY (3) SOP_ONLY (4) MOP_ONLY

	; Get RXC:exp_SOP bit.
	alu_shf[sop_bit, --, b, lm_sop_alloc_valid_aal_flist, >>31]
	; Extract the 3rd bit (EOP bit) of PTI.
	alu_shf[eop_bit, 0x1, AND, pti_3bits]	
	; Combine sop_bit:eop_bit.
	alu_shf[sop_eop_bits, eop_bit, OR, sop_bit, <<1]

	;(sop:eop) 00 - MOP 		(!EOP and !SOP)
	;(sop:eop) 01 - EOP 		(EOP_only) 
	;(sop:eop) 10 - SOP 		(SOP_only)
	;(sop:eop) 11 - EOP_SOP 	(EOP_SOP)
	jump[sop_eop_bits, J_TABLE_BASE#], defer[3], targets[MOP_J#, EOP_J#, SOP_J#, EOP_SOP_J#] 
	; Init the CSR CRC_REMAINDER with the RXC:crc_residue.
	// This requires 3 nops between local csr wr and crc instr (instr which is inside
	// following macros)
	local_csr_wr[CRC_REMAINDER, lm_crc_residue]     		
	; Signal the next thread.
	signal_ctx(next_thread_sig_val)
	; Adding next_thread_sig to sig_mask.
	alu[rx_sig_mask, --, b, 1,  <<&next_thread_sig]

							
J_TABLE_BASE#:

MOP_J#:
	br[MOP#]
EOP_J#:
	br[EOP#]
SOP_J#:
	br[SOP#]
EOP_SOP_J#:
	br[EOP_SOP#]


;-------------------------------------------------------------------------------
MOP#:

	_aal5_rx_mop_only_proc(rx_sig_mask, rxc_abs_addr, $$cell_in, input_port,  
			$prefetch_buf_handle)

	br[AAL5_RX_END#], defer[2]
	; Signal next thread.
	signal_ctx(next_thread_sig_val)
	; Set dl_sink sig_mask register with RSW signal.
	alu[sig_mask, --, B, 1, <<(&rsw_rcvd_sig)]  

;-------------------------------------------------------------------------------
EOP#:

	_aal5_rx_eop_only_proc(rx_sig_mask, rxc_abs_addr, $$cell_in, input_port, vc_flowid,
	    pti_3bits)

	br[AAL5_RX_END#], defer[2]
	; Signal next thread.
	signal_ctx(next_thread_sig_val)
	; Set dl_sink sig_mask register with RSW signal.
	alu[sig_mask, --, B, 1, <<(&rsw_rcvd_sig)]  

;-------------------------------------------------------------------------------
SOP#:

	_aal5_rx_sop_only_proc(rx_sig_mask, rxc_abs_addr, $$cell_in, $prefetch_buf_handle)

	br[AAL5_RX_END#], defer[2]
	; Signal next thread.
	signal_ctx(next_thread_sig_val)
	; Set dl_sink sig_mask register with RSW signal.
	alu[sig_mask, --, B, 1, <<(&rsw_rcvd_sig)]  

;-------------------------------------------------------------------------------
EOP_SOP#:

	_aal5_rx_eop_sop_proc(rx_sig_mask, rxc_abs_addr, $$cell_in, input_port, vc_flowid,
			pti_3bits, $prefetch_buf_handle)

	br[AAL5_RX_END#], defer[2]
	; Signal next thread.
	signal_ctx(next_thread_sig_val)
	; Set dl_sink sig_mask register with RSW signal.
	alu[sig_mask, --, B, 1, <<(&rsw_rcvd_sig)]  



;------------------------------------------------------------------------------
; ERROR Conditions
;------------------------------------------------------------------------------


;-------------------------------------------------------------------------------		
RSW_ERR#:
;-------------------------------------------------------------------------------

	; Check for Parity Error (13th bit).
	// Err bit - This is the logical OR of Par Err.(Parity Err (11th bit) - If this 
	// bit is set, it means that a parity error was detected).


	alu_shf[--, 0x1, AND, $rsw0, >>13]			
	bne[RSW_PAR_ERR#]

	; Check for null/idle cell Rx (9th bit).
	// null/idle cell Rx - If this bit is set, it means that the Rx_Thread_Freelist 
	// timeout expired before any more data was received, and that a null RSW is 
	// being pushed in order to keep the receive pipeline flowing. The rest of the 
	// fields in RSW must be ignored; there is no data or RBUF entry associated with 
	// a null RSW.

	alu_shf[--, 0x1, AND, $rsw0, >>9]				
	bne[RSW_IDLE_RX#]

;-----------------------------------------------------------------------------
RSW_CELL_SZ_ERR#:


	; Here its cell_size error, ie cell_size is less than 52B.

	_aal5_rx_incr_port_counter[input_port, AAL5_RX_HW_ERROR]
	
	alu[dl_next_block, --, B, BID_NULL]		 		; set next_block NULL

	// Do necessary signalling to keep moving the pipeline.
	br[ERR_PROC#]

;-----------------------------------------------------------------------------
RSW_PAR_ERR#:


	_aal5_rx_incr_port_counter[input_port, AAL5_RX_HW_ERROR]

	alu[dl_next_block, --, B, BID_NULL]		 		; set next_block NULL
		
	// Do necessary signalling to keep moving the pipeline.
	br[ERR_PROC#]

;-----------------------------------------------------------------------------
RSW_IDLE_RX#:


	// Do necessary signalling to keep moving the pipeline.
	
	_aal5_rx_err_phase1()

	; no RBUF element is assigned, so we should not free the RBUF element.
	_aal5_rx_err_phase2_no_rbuf_free()

	; Do necessay signalling to bypass OAM processing
	_aal5_rx_err_phase2_1()


	br[ERR_PHASE3#], defer[1]
	alu[dl_next_block, --, B, BID_NULL]		 		; set next_block NULL



#ifdef USE_HASH_LOOKUP

;-----------------------------------------------------------------------------
HASH_MISS_ERR#:
;-----------------------------------------------------------------------------

	_aal5_rx_incr_port_counter[input_port, AAL5_RX_VC_ERROR]

	// Do necessary signalling to keep moving the pipeline.

	; Hash Miss Err Phase1.

#ifdef NUM_OF_INPUTPORTS_16

	_aal5_rx_hash_miss_inputports16()

#else 

	_aal5_rx_hash_miss()

#endif


	alu[dl_next_block, --, B, BID_NULL]	 	; set next_block NULL

	br[ERR_PHASE2#]

#endif  // USE_HASH_LOOKUP


;-----------------------------------------------------------------------------
ERR_PROC#:
;-----------------------------------------------------------------------------

ERR_PHASE1#:
	_aal5_rx_err_phase1()

ERR_PHASE2#:
	_aal5_rx_err_phase2()

ERR_PHASE2_1#:
	_aal5_rx_err_phase2_1()


ERR_PHASE3#:
	
	; For TWO_ME_RX_FP, it covers Phase3, Phase4, Phase5.
	; For ONE_ME_RX_FP, it covers Phase3.

	; This is applicable to RSW err & Hash miss err (We dont have RXC in LM here) 
	; non_AAL5 error & non_valid_VC. (We have RXC in LM here)
	;
	; (We dont care here whether we have RXC here or not).
	; In either case there is no need to write back RXC to SRAM.
	;
	; In case of RSW Error, CRC check or pkt_length check would help to catch
	; HW RX error.

	; Signal next thread.
	signal_ctx(next_thread_sig_val)

	; Do necessary signalling to keep moving the pipeline.
	_aal5_rx_err_phase3()   

		
	br[AAL5_RX_END#], defer[2]
	; Signal next thread.
	signal_ctx(next_thread_sig_val)
	; Set dl_sink sig_mask register with RSW signal.
	alu[sig_mask, --, B, 1, <<(&rsw_rcvd_sig)]  
	
;-----------------------------------------------------------------------------		
INV_VC_MGT_CELL_NON_AAL5_ERR#:
;-----------------------------------------------------------------------------

	; Here we have got either management cell OR NON_VALID VC OR NON_AAL5 cells.

	; Check whether its management cell (if bit is set, its a mgt cell).
	br_bset[pti_3bits, 2, MGT_CELL#]

	; Check whether this VC is not valid (this is required as out of 64K, 
	; some of them might be disabled) (if bit is set, its not a valid VC)).
	br_bset[lm_sop_alloc_valid_aal_flist, 23, VC_NOT_VALID#]

;-------------------------------------------------------------------------------
NON_AAL5#:


	; For now, nothing to do, basically dropping the packet.

	; For now, do necessary signalling to keep moving the pipeline.

	br[ERR_PHASE3#], defer[1]
	alu[dl_next_block, --, B, BID_NULL]		 		; set next_block NULL


;-------------------------------------------------------------------------------
VC_NOT_VALID#:
	
	; Nothing to do, basically dropping the packet.

	; Do necessary signalling to keep moving the pipeline.

	br[ERR_PHASE3#], defer[1]
	alu[dl_next_block, --, B, BID_NULL]		 		; set next_block NULL


;-------------------------------------------------------------------------------
MGT_CELL#:	


#ifdef ATM_OAM_ENABLED

	; Do necessary signalling to keep moving the pipeline.
	; Signal next thread.
	signal_ctx(next_thread_sig_val)
	; Set dl_sink sig_mask register with RSW signal.
	alu[sig_mask, --, B, 1, <<(&rsw_rcvd_sig)]  

#else // ATM_OAM_ENABLED

	; Nothing to do, basically dropping the packet. If management
	; cell are to be handled then enable ATM_OAM_ENABLED flag wherein
	; the management cells will be sent as exception to Xscale

	; Just incrementing the counter for statistics purposes.
	_aal5_rx_incr_vc_counter[vc_flowid, AAL5_RX_MGT_CELL_COUNT]

	br[ERR_PHASE3#], defer[1]
	alu[dl_next_block, --, B, BID_NULL]		 		; set next_block NULL

#endif // ATM_OAM_ENABLED	

	
;-------------------------------------------------------------------------------


AAL5_RX_END#:

#ifndef FPGA_HEADER
	xbuf_free($$cell_in)
#endif

	; Add thread to freelist.
	msf[fast_wr, --, rx_freelist_no, rx_thd_freelist_reg]	

.end
#endm 	//   aal5_rx()


#ifndef TAKE_MACROS_ONLY
;-----------------------------------------------------------------------------
;	   Main code starts here
;-----------------------------------------------------------------------------

MAIN#:

.begin

.sig sink_sig  					// signal used in dl_sink for scratch_put
.reg sink_sig_mask				// signal mask used by aal5_rx() and dl_sink()
					

	; AAL5 RX Init.
	aal5_rx_init()

.while(1)

	;----------------------------------------------------------------------
	; For ONE_ME_RX_FP - Phase1#, Phase2#, Phase3# 
	; For TWO_ME_RX_FP - Phase1#, Phase2#, Phase3#, Phase4#, Phase5#  
	;----------------------------------------------------------------------
	
	; AAL5 frame reassembly.
	aal5_rx(sink_sig_mask)

	;----------------------------------------------------------------------
	; For ONE_ME_RX_FP - Phase4#
	; For TWO_ME_RX_FP - Phase6#
	;----------------------------------------------------------------------
	
	xbuf_alloc[$x_dl_sink, RX_TO_FWD_MSG_SIZE, write]	// sxfer reg to be used by dl_sink

	; Passing onto next block.
	dl_sink($x_dl_sink, sink_sig, sink_sig_mask) 
												 


	xbuf_free($x_dl_sink)				// free transfer registers

.io_completed sink_sig



.endw

.end   // MAIN# 

;-----------------------------------------------------------------------------
;		Main code ends here
;-----------------------------------------------------------------------------
#endif /* TAKE_MACROS_ONLY */

#endif  // 	_AAL5_RX_UC_