//------------------------------------------------------------------------------
//                                                                      
//                   I N T E L   P R O P R I E T A R Y                   
//                                                                       
//      COPYRIGHT (c)  2002 BY  INTEL  CORPORATION.  ALL RIGHTS          
//      RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//      BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//      RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//      LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//      MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//      THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                       
//                         INTEL  CORPORATION                            
//                                                                      
//                      2200 MISSION COLLEGE BLVD                        
//                                                                       
//                SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                       
//------------------------------------------------------------------------------
//
//
//      File Name: aal5_rx_oam.uc
//
//      Purpose: 
// 		Code to perform ATM OAM F5 and F4 flow handling
//
//		The following assumptions are made by this microblock.
//		1.	All logic needed for ATM OAM is implemented by an OAM handler on
//			the Xscale core processor. The fast path code below limits itself
//			to identifying ATM OAM cells and sending them as exceptions to 
//			Xscale core. Additionally, the code performs updates to the OAM 
//			PM Rx counters for F4 and F5 flows.
//		2.	Either VP or VC PM are enabled simultaneously for a VCC and not both. 
//	 		Hence we will increment either the VC or the VP counters. 
//		3.	Either segment or ETE PM are enabled for a VP or VC but not both.
//			We only maintain one set of PM counters to be used as either segment
//			or ETE OAM procedure
//
//
//
/////////////////////////////////////////////////////////////////////////////////
#ifndef	_AAL5_RX_OAM_UC_
#define	_AAL5_RX_OAM_UC_

///////////////////////////////////////////////////////////////////////////////
//	MACRO
//		_aal5_rx_oam_init()
//
//	DESCRIPTION 
//		Initialize data structures used by ATM OAM Rx and kickstart it
//
//	ARGUMENTS None
//
//		
///////////////////////////////////////////////////////////////////////////////
#macro _aal5_rx_oam_init()
.begin
	#ifdef ATM_OAM_ENABLED

	; Initiaze the PM counters cache 
	.reg lm_cache_addr							; Needed since to store constants
	immed[lm_cache_addr, OAM_TAG_CACHE_LM_BASE]	; Local LM_addr_0 with tage base
	local_csr_wr[active_lm_addr_0, lm_cache_addr]	
	nop
	nop
	nop
	alu[*l$index0++, --, B, 0]
	alu[*l$index0++, --, B, 0]
	alu[*l$index0++, --, B, 0]
	alu[*l$index0++, --, B, 0]
	alu[*l$index0++, --, B, 0]
	alu[*l$index0++, --, B, 0]
	alu[*l$index0++, --, B, 0]
	alu[*l$index0++, --, B, 0]	
	#endif // ATM_OAM_ENABLED
.end
#endm
///////////////////////////////////////////////////////////////////////////////
//	MACRO
//		_aal5_rx_oam_pm_processing
//
//	DESCRIPTION 
//		Perform fast path processing for ATM OAM
//
//	ARGUMENTS
//			in_rx_sig_mask :   mask to be set with various I/O signals.
//
//			in_rxc_abs_addr :    absolute rxc addr used to write back RXC to SRAM.
//
//			$$inAtmCell :  dram xfer regs used to write to DRAM.
//
//			in_port_num : The port through which the packet came in.
//
//			in_vc_flowid :      index of RXC into RXC table (VC info table)
//
//			in_pti :        3 bits of PTI of ATM header.
//
//			in_atm_oam_options: OAM options enabled for this VC
//		
//			$in_prefetch_buf_handle:	Prefetched buffer handle for generatin exception
//		
//			MGT_CELL_PROC_LABEL	: Branch to take when generating exception to Xscale
//		
//
//		
///////////////////////////////////////////////////////////////////////////////

#macro _aal5_rx_oam_pm_processing(in_rx_sig_mask, in_rxc_abs_addr, $$inAtmCell,\
								  in_port_num, in_vc_flowid, in_pti, in_atm_oam_options,\
								  $in_prefetch_buf_handle, MGT_CELL_PROC_LABEL)
.begin
	// Include code in this file only if OAM functions enabled
	#ifdef ATM_OAM_ENABLED
	.reg cell_type 			; Bit map to determine the type of cell (see below)
							; bit 0 - OAM cell type 0 - F4 OAM 1 - F5 OAM
							; bit 1 - FPM cell ?	0 - NO     1 - YES
							; bit 2 - OAM cell ?    0 - NO     1 - YES
	.reg oam_rx_tbl_base 	; SRAM table base for F4/F5 OAM Rx counter table
	.reg oam_rx_tbl_offset	; SRAM offset for F4/F5 OAM Rx counter table						
	.reg bedc				; BEDC computed using BIP-16 on current cell
	.reg pm_ctr_cache_tag	; To distinguish F4 and F5 PM counters
	.reg sig_wait			; Signal mask to wait on for SRAM accesses
	.reg old_lm_addr0		; To store the old LM pointer 0 address
	.reg vpi				; VPI of VC on which cell received
	.sig oam_rd_wr_sig		; To wait for PM counter read from SRAM

	xbuf_alloc($oam_pm_ctrs,2,read_write)
	xbuf_alloc($$cell_out, 12, write)

	; Signal the next thread.
	signal_ctx(next_thread_sig_val)
	
	; Load the original LM address pointers
	local_csr_rd[active_lm_addr_0]
	immed[old_lm_addr0, 0]
	
	_aal5_rx_oam_analyze_atm_cell($$inAtmCell, in_pti, in_atm_oam_options,\
					cell_type, CHECK_FPM_OAM_CELL#, HANDLE_USER_CELL#, NON_FPM_OAM_CELL#)

CHECK_FPM_OAM_CELL#:

	; OAM cells are sent to Xscale core as exception.
	; Check if the cell is a OAM FPM cell. FPM cells
	; have OAM type 0010 and Function type 0000. 
	br=byte[$$inAtmCell/**/1, 3, OAM_FPM_CELL_TYPE, HANDLE_FPM_CELL#]

NON_FPM_OAM_CELL#:
	; This is a F4 or F5 OAM cell but not an FPM cell
	; F5 OAM cells are handled as user cells for F4 flow. Hence if it is a 
	; F5 OAM cell (non FPM - since PM is not supported simultaneously on F4
	; and F5) use it to collect F4 PM stats before sending to Xscale core. 
	br_bset[cell_type, F5_OAM_CELL_BIT_POS, CHECK_F4_PM_ENABLED#]

	; F4 OAM cells are sent directly to Xscale core since no 
	; PM processing is needed for F4 non FPM cells. F4 cells 
	; are not user cells for either F4 or F5 flow. Wait for 
	; previous thread to complete
	ctx_arb[next_thread_sig], br[SEND_EXCEPTION#]

HANDLE_FPM_CELL#:	
	
	; When FPM cells are sent to Xscale core, the current accumulated PM
	; counters are also piggybacked in the exception sent to Xscale. These
	; piggy backed counters are used to generate BR cells
	; Check if F5 cell bit is set.
	br_bclr[cell_type, F5_OAM_CELL_BIT_POS, CHECK_F4_PM_ENABLED#], defer[1]
		alu[cell_type, cell_type, OR, FPM_CELL_BIT]	; Mark Cell as OAM FPM cell.

	; Handle F5 FPM cells below
	; Handling as user cell is required for updation of PM counters only.
	; Assumption is that PM for F5 and F4 cells will not be enabled 
	; simultaneously and hence F5 FPM cells need for be counted in the
	; F4 PM counters since F5 FPM cells indicate that F5 PM is enabled
	; for the VC and hence no PM is enabled for the VP hosting this VC
	
HANDLE_USER_CELL#:
	; Assumptions:
	; 1.Either VP or VC PM are enabled simultaneously for a VCC and not both. 
	;   Hence we will increment either the VC or the VP counters. 
	; 2.Either segment or ETE PM are enabled for a VP or VC but not both.
	;   We only maintain one set of PM counters to be used as either segment or
	;   ETE OAM procedure

	; Is ETE or SEG PM enabled for this VC? If not check if VPI PM enabled
	br_bclr[in_atm_oam_options, F5_PM_INDICATOR_BIT_POS, CHECK_F4_PM_ENABLED#]

	; VCI PM enabled (Either segment or ETE allowed, but not both)
	; VCI PM counters placed accessed using same index as that used for VC
	; table lookup. Load VC Rx OAM table base and offset
	; Rx Table size = 16 bytes

	move(oam_rx_tbl_base, VC_RX_OAM_TBL_BASE)
	br[COLLECT_PM_OAM_STATS#], defer[2]			; defer slots used 
		alu[oam_rx_tbl_offset, --, B, in_vc_flowid, <<VC_OAM_RX_TBL_SZ_SHFT]
		alu[pm_ctr_cache_tag, in_vc_flowid, OR, VC_OAM_LM_TAG, <<24]

CHECK_F4_PM_ENABLED#:
	; Is ETE or SEG PM enabled for this VP else end OAM PM processing
	br_bset[in_atm_oam_options, F4_PM_INDICATOR_BIT_POS, VP_PM_ENABLED#], defer[3]
		; Use the three defer slots. If the vp_pm_enabled does not get taken
		; there is no side effect since, in that case OAM processing ends.
		; VP OAM PM enabled for VP. Load VP OAM Rx Table base and offset
		; Rx Table Size = 16 bytes
		; oam_rx_tbl_offset = (in_port * MAX_VPI_PER_PORT + in_vpi) * Rx Table size
		alu[oam_rx_tbl_offset, --, B, in_port_num, <<MAX_VPI_PER_PORT_SHIFT]
		alu[vpi, --, B, $$inAtmCell/**/0, >>20]
		alu[oam_rx_tbl_offset, oam_rx_tbl_offset, +, vpi]

	; Neither VP nor VC PM enabled. Wait for previous thread to complete
	ctx_arb[next_thread_sig], br[END_ATM_OAM_RX#]		

VP_PM_ENABLED#:
	alu[oam_rx_tbl_offset, --, B, oam_rx_tbl_offset, <<VP_OAM_RX_TBL_SZ_SHFT] 
	; Create LM cache tag for VP OAM PM counters in LM cache
	alu[pm_ctr_cache_tag, oam_rx_tbl_offset, OR, VP_OAM_LM_TAG, <<24]

	move(oam_rx_tbl_base, VP_RX_OAM_TBL_BASE)
		
COLLECT_PM_OAM_STATS#:
	alu[sig_wait, --, B, 1, <<&(next_thread_sig)]
	;This lookup function will set the ACTIVE_LM_ADDR_0 to point to the
	;matching LM entry belonging to another thread in case of a hit.
	;In case of a miss, it will set ACTIVE_LM_ADDR_0 to point to the blank
	;LM entry of self thread. In case of a hit, branch done to pm_cache_hit
	_aal5_rx_oam_lookup_lm_pm_cache(pm_ctr_cache_tag, PM_CACHE_HIT#)

PM_CACHE_MISS#:
	sram[read, $oam_pm_ctrs0, oam_rx_tbl_base, oam_rx_tbl_offset, LWCOUNT2],\
		sig_done[oam_rd_wr_sig]
	
	; Wait for SRAM read to complete
	alu[sig_wait, sig_wait, OR, 1, <<&(oam_rd_wr_sig)]

PM_CACHE_HIT#:

	br_bset[cell_type, FPM_CELL_BIT_POSITION, NO_BEDC_NEEDED#], defer[2]
		; Increment reference count for the cached LM entry
		alu[lm_ref_count, lm_ref_count, +, 1]
		; Initialize BEDC to 0
		alu[bedc, --, B, 0]

	; Compute the BEDC while the sram read completes
	; BEDC is computed using even parity BIP-16 over
	; the monitored cell block

	_aal5_rx_oam_compute_bedc(bedc, $$inAtmCell)

NO_BEDC_NEEDED#:

	; Wait for all SRAM write operations to complete and wait for
	; previous thread signal to enter critical section for updating
	; counters
	ctx_arb[--], defer[1]	
		local_csr_wr[active_ctx_wakeup_events, sig_wait]

	.io_completed oam_rd_wr_sig

	; Check if SRAM read was lauched for counters
	alu[--, sig_wait, AND, 1, <<&(oam_rd_wr_sig)]
	beq[NO_SRAM_READ_LAUNCHED#]
	; This is the local memory cache lookup missed case 
	; wherein the OAM PM counters are to be cached by this thread
	; Store counters read from SRAM into LM
	; Due to conditional nature in which these registers are read
	; from SRAM, the assembler may not know that the read xfer 
	; registers are used only when they are read. Disable this warning
	#pragma warning(disable:4701)
	ld_field_w_clr[lm_trcc0,   0011, $oam_pm_ctrs0]
	ld_field_w_clr[lm_trcc0_1, 0011, $oam_pm_ctrs0, >>16]
	ld_field_w_clr[lm_bedc,    0011, $oam_pm_ctrs1]
	#pragma warning(default:4701)
NO_SRAM_READ_LAUNCHED#:
	.io_completed oam_rd_wr_sig
	br_bclr[cell_type, FPM_CELL_BIT_POSITION, UPDATE_COUNTERS#]

	; FPM cell received
	; Reception of a FPM cell indicates completion of a block
	; of cells. The current accumulated counters are copied out
	; and stored in old_xxxx to piggyback onto the FPM cell sent
	; to Xscale core. The counters in cache are zeroed out to 
	; start afresh for next block of cells

	alu[$$cell_out9,  --, B, lm_trcc0]		; Copy out accumulated counters
	alu[$$cell_out10, --, B, lm_trcc0_1]	; Copy out accumulated counters
	alu[$$cell_out11, --, B, lm_bedc]		; Copy out accumulated BEDC
	alu[lm_trcc0, 	--, B, 0]				; Zero out counters
	br[DECR_REF_CNT#], defer[2]				; Fill the two defer slots and
		alu[lm_trcc0_1, --, B, 0]			; zero out the counters
		alu[lm_bedc, 	--, B, 0]
	
UPDATE_COUNTERS#:
	br_bset[$$inAtmCell/**/0, 0, DECR_REF_CNT#], defer[2]
		alu[lm_bedc, lm_bedc, XOR, bedc]		; Update BEDC in LM cache
		alu[lm_trcc0_1, lm_trcc0_1, +, 1]		; Incr. trcc0_1 
    alu[lm_trcc0, lm_trcc0, +, 1]				; Incr. trcc0, (if CLP==0)

DECR_REF_CNT#:
	; The last thread working on this record flushes it back to SRAM
	; Decrement reference count and if count is 0 flush to SRAM
	br!=byte[lm_ref_count, 0, 1, END_ATM_OAM_RX#], defer[1]
	   	alu[lm_ref_count, lm_ref_count, -, 1]	
	.reg	temp
	ld_field_w_clr[temp, 1100, lm_trcc0_1, <<16]
	alu[$oam_pm_ctrs0, lm_trcc0, OR, temp]
	ld_field_w_clr[$oam_pm_ctrs1, 0011, lm_bedc] 

	sram[write, $oam_pm_ctrs0, oam_rx_tbl_base, oam_rx_tbl_offset, LWCOUNT2],\
		sig_done[oam_rd_wr_sig]

	alu[lm_ref_count, --, B, lm_ref_count, >>16]
	_aal5_rx_oam_clear_cache_tag(lm_ref_count)
	ctx_arb[oam_rd_wr_sig]
	
END_ATM_OAM_RX#:
	; Check if the cell is a OAM cell. If so, send
	; exception to Xscale OAM handling blocks.

	br_bclr[cell_type, OAM_CELL_TYPE_BIT_POSITION, END#], defer[3]
		; Restore the LM address pointers to original values
		local_csr_wr[active_lm_addr_0, old_lm_addr0]	
		nop
		nop

SEND_EXCEPTION#:
	; Copy the first 9 longwords of payload from OAM cell to the outgoing
	; packet. When FPM cells are received the last three longwords are used
	; to piggyback the current accumulated OAM counters for sending to Xcscale
	; core. If this is not an FPM cell, then copy the remaining bytes also
	; from incoming cell. Piggybacking does not lead to any loss of data since
	; the last three longwords of the ATM FPM cell are padding bytes.	
	#define_eval	i	0
	#while (i < 9)
		alu[$$cell_out/**/i/**/, --, B, $$inAtmcell/**/i/**/]
		#define_eval	i	(i + 1)
	#endloop
	#undef	i

	.set $$cell_out9 $$cell_out10 $$cell_out11
	; Check if the OAM cell is a FPM cell; i.e. whether counters have been 
	; piggybacked already on the outgoing packet. If not do copy them in now.
	br_bset[cell_type, FPM_CELL_BIT_POSITION, FPM_COUNTER_PIGGYBACKED#], defer[3]
		; Signal the next thread.
		signal_ctx(next_thread_sig_val)
		; Adding next_thread_sig to sig_mask.
		alu[in_rx_sig_mask, --, b, 1,  <<NEXT_THREAD_SIGNAL]
		; Init. exceptoin code registers for sending exception
		immed[dl_exception_reg, 0]
	; This is not a FPM cell and no counters were piggybacked,
	; copy longwords from incoming OAM cell to outgoing packet.
	alu[$$cell_out9,  --, B, $$inAtmcell/**/10]
	alu[$$cell_out10, --, B, $$inAtmcell/**/11]
	alu[$$cell_out11, --, B, $$inAtmcell/**/12]

FPM_COUNTER_PIGGYBACKED#:
	_aal5_rx_incr_vc_counter[vc_flowid, AAL5_RX_MGT_CELL_COUNT]

	; Process the management cell.
	_aal5_rx_mgt_cell_proc(in_rx_sig_mask, in_rxc_abs_addr,\
						 $$cell_out, $in_prefetch_buf_handle)

	; Set exception block ID as OAM handler and exception code 
	; as ATM_RX_OAM_TRAFFIC
	dl_exception_set_high_priority[IX_CC_ATM_RX_MICROBLOCK_ID,\
								   IX_ATM_OAM_CELL_RECEIVED]

	; Branch to management cell handling label in main processing 
	; function, set the LM address pointer to original value before
	; branching. However make sure that there are at least one cycles
	; executed there before the LM address pointer is accessed.
	br[MGT_CELL_PROC_LABEL]

END#:
	nop
	xbuf_free($$cell_out)
	xbuf_free($oam_pm_ctrs)
	#endif // ATM_OAM_ENABLED
.end
#endm //_aal5_rx_oam_pm_processing
///////////////////////////////////////////////////////////////////////////////
//	MACRO
//		_aal5_rx_oam_compute_bedc
//
//	DESCRIPTION 
//	Compute BEDC using BIP-16 over ATM cell
//
//	ARGUMENTS
//		out_bedc	: BEDC computed over the cell
//		inAtmCell	: Cell on which to compute BEDC
//
///////////////////////////////////////////////////////////////////////////////
#macro	_aal5_rx_oam_compute_bedc(out_bedc, $$inAtmCell)
.begin
	.begin
		.reg temp					; temporary register

		alu[temp, --, B, $$inAtmCell/**/1/**/]

		; ATM cell in $$inAtmCell0 - $$inAtmCell11. Assuming PHY
		; device configured to skip the HEC and hence payload 
		; in MSF aligned on 4 byte boundary.

		#define_eval i 2
		#while ( i < 13 )
			alu[temp, temp, XOR, $$inAtmCell/**/i/**/]	; XOR 32 bit fields
		#define_eval i (i + 1)
		#endloop
		#undef i 										; cleanup name space
		ld_field_w_clr[out_bedc, 0011, temp]			; XOR the 16 bits
		alu[temp, --, B, temp, >>16]
		alu[out_bedc, out_bedc, XOR, temp]				; fields for BIP-16
	.end
.end
#endm
///////////////////////////////////////////////////////////////////////////////
//	MACRO
//		_aal5_rx_oam_analyze_atm_cell
//
//	DESCRIPTION 
//	Analyze the ATM cell to determine if it is a F4 or F5 oam cell.
//	If F4 or F5 OAM cell determine if the cell needs to be handled or
//  dropped depending on the ATM options (i.e. whether the flow is enabled)
//	ARGUMENTS
//			inAtmHdr :   ATM header LW
//
//			in_pti :        3 bits of PTI of ATM header.
//
//			in_atm_oam_options: OAM options enabled for this VC
//
//			out_cell_type :    Analyzed cell type
//
//			CHECK_FPM_CELL_BRANCH :  Branch to take for OAM cells
//
//			USER_CELL_BRANCH : Branch to take for user cells
//
//			NON_FPM_CELL_BRANCH:	Branch to take when OAM cell received
//									on VC not configured to handle that cell
//
///////////////////////////////////////////////////////////////////////////////
#macro _aal5_rx_oam_analyze_atm_cell($$inAtmHdr, in_pti, in_atm_opts, out_cell_type,\
		 CHECK_FPM_CELL_BRANCH, USER_CELL_BRANCH, NON_FPM_CELL_BRANCH)
.begin
	.reg vci 									; VCI

	ld_field_w_clr[vci, 0011, $$inAtmHdr/**/0/**/, >>4]	; Extract the VCI.

	; Classify cell into F4, F5 OAM cell or user cell
CHECK_VPI_SEG#:
	; Is this a VP SEG OAM cell (vci==3) ? If not check if VP ETE OAM cell
	alu[--, vci, -, ATM_VCI_F4_SEG_OAM]
	bne[CHECK_VPI_ETE#], defer[1]
		; Initialize cell type to non OAM cell
		immed[out_cell_type, 0]

	; Is VP configured as SEG end point ? If not then dont check if this 
	; is a FPM cell. Just send it up to Xscale core.
	br_bclr[in_atm_opts, OAM_SEG_ENDPOINT_BIT, NON_FPM_CELL_BRANCH], defer[2]
		alu[out_cell_type, --, B, F4_OAM_CELL_BIT]		; F4 SEG OAM cell
		alu[out_cell_type, out_cell_type, OR, OAM_CELL_TYPE_BIT]
	br[CHECK_FPM_CELL_BRANCH]

CHECK_VPI_ETE#:
	; Is this a VP ETE OAM cell (vci==4) ? If not check if VC OAM cell 
	alu[--, vci, -, ATM_VCI_F4_ETE_OAM]
	bne[CHECK_VCI_ETE#]
	; Is VP configured as ETE end point ? If not then dont check if this is
	; a FPM cell. Just send it up to Xscale core
	br_bclr[in_atm_opts, OAM_ETE_ENDPOINT_BIT, NON_FPM_CELL_BRANCH], defer[2]
		alu[out_cell_type, --, B, F4_OAM_CELL_BIT]		; F4 ETE OAM cell
		alu[out_cell_type, out_cell_type, OR, OAM_CELL_TYPE_BIT]
	br[CHECK_FPM_CELL_BRANCH]

CHECK_VCI_ETE#:	
    ; Is this a VC ETE OAM cell (pti==5) ? If not check if VC SEG OAM ?
	br!=byte[in_pti, 0, ATM_PTI_F5_ETE_OAM, CHECK_VCI_SEG#]
	; Is VC configured as ETE end point ?  If not then dont check if this is
	; a FPM cell. Just send it up to Xscale core
    br_bclr[in_atm_opts, OAM_ETE_ENDPOINT_BIT, NON_FPM_CELL_BRANCH], defer[2]
		alu[out_cell_type, --, B, F5_OAM_CELL_BIT]		; F5 ETE OAM cell
		alu[out_cell_type, out_cell_type, OR, OAM_CELL_TYPE_BIT]
    br[CHECK_FPM_CELL_BRANCH]

CHECK_VCI_SEG#:
    ; Is this a VC SEG OAM cell (pti==4) ? If not it must be a user cell
	br!=byte[in_pti, 0, ATM_PTI_F5_SEG_OAM, USER_CELL_BRANCH]
	; Is VC configured as SEG end point ?  If not then dont check if this is
	; a FPM cell. Just send it up to Xscale core
	br_bclr[in_atm_opts, OAM_SEG_ENDPOINT_BIT, NON_FPM_CELL_BRANCH], defer[2]
		alu[out_cell_type, --, B, F5_OAM_CELL_BIT]		; F5 SEG OAM cell		
		alu[out_cell_type, out_cell_type, OR, OAM_CELL_TYPE_BIT]
.end
#endm
///////////////////////////////////////////////////////////////////////////////
//	MACRO
//		_aal5_rx_oam_lookup_lm_pm_cache
//
//	DESCRIPTION 
//		Search through LM tag cache to find if the desired counters are
//		already cached by some thread. If so then set the active LM addr 0
//		pointer to the cached counter block.
//		
//
//	ARGUMENTS
//		in_cache_tag				- Cache tag to search for
//		return_label				- Label to branch in case of cache hit
//
///////////////////////////////////////////////////////////////////////////////
#macro _aal5_rx_oam_lookup_lm_pm_cache(in_cache_tag, ret_lbl)
.begin
	.reg lm_cache_addr					; Cache local memory addr
	
	; Local LM_addr_0<-tagbase
	immed[lm_cache_addr, OAM_TAG_CACHE_LM_BASE]
	local_csr_wr[active_lm_addr_0, lm_cache_addr]	

	; Loading CSR needs 3 cycles. Make sure first
	; access is after 3 cycles. Loop through counter
	; cache entries for all previous threads to see
	; if any of them has cached the desired counters.
	; Only need to look for previous threads since threads 
	; ahead of this thread would not have had a 
	; chance to execute yet.
	.begin
		.reg idx							; Case index
		alu[idx, 7, -, ctx_num]				; (7 - thread num) * 8	
		alu_shf[idx, --, B, idx, <<3]
		jump[idx, LOOKUP_TABLE_BASE#],
			targets[CTX0#, CTX1#, CTX2#, CTX3#,\
					CTX4#, CTX5#, CTX6#, CTX7#]
	.end

LOOKUP_TABLE_BASE#:
CTX7#:	_aal5_oam_rx_chk_tag(in_cache_tag, 6, ret_lbl)
CTX6#:	_aal5_oam_rx_chk_tag(in_cache_tag, 5, ret_lbl)
CTX5#:	_aal5_oam_rx_chk_tag(in_cache_tag, 4, ret_lbl)
CTX4#:	_aal5_oam_rx_chk_tag(in_cache_tag, 3, ret_lbl)
CTX3#:	_aal5_oam_rx_chk_tag(in_cache_tag, 2, ret_lbl)
CTX2#:	_aal5_oam_rx_chk_tag(in_cache_tag, 1, ret_lbl)
CTX1#:	_aal5_oam_rx_chk_tag(in_cache_tag, 0, ret_lbl)
CTX0#:	; Cache missed

	.reg idx							; Case index
	alu[idx, --, B, ctx_num, <<2]		; (7 - thread number) * 4	
	jump[idx, MISS_TABLE_BASE#],
		targets[M0#, M1#, M2#, M3#, M4#, M5#, M6#, M7#]

MISS_TABLE_BASE#:
M0#:	_aal5_oam_rx_upd_tag_on_miss(in_cache_tag, 0, END#)
M1#:	_aal5_oam_rx_upd_tag_on_miss(in_cache_tag, 1, END#)
M2#:	_aal5_oam_rx_upd_tag_on_miss(in_cache_tag, 2, END#)
M3#:	_aal5_oam_rx_upd_tag_on_miss(in_cache_tag, 3, END#)
M4#:	_aal5_oam_rx_upd_tag_on_miss(in_cache_tag, 4, END#)
M5#:	_aal5_oam_rx_upd_tag_on_miss(in_cache_tag, 5, END#)
M6#:	_aal5_oam_rx_upd_tag_on_miss(in_cache_tag, 6, END#)
M7#:	_aal5_oam_rx_upd_tag_on_miss(in_cache_tag, 7, END#)
END#:
	nop							; Allow LM pointer to stabilize
	immed[lm_cache_addr, OAM_TAG_CACHE_LM_BASE]
	alu[lm_ref_count, lm_cache_addr, +, idx]
	alu[lm_ref_count, --, B, lm_ref_count, <<16]
.end
#endm //_aal5_oam_tx_lookup_lm_pm_cache
//------------------------------------------------------------------
//	MACRO
//		_aal5_oam_rx_chk_tag
//
//	DESCRIPTION 
//		Match the cache tag at indicated index. Branch to match lbl
//		if matched. Addittionally load the LM pointer with cache
//		address for that counter type and index
//		
//	ARGUMENTS
//		in_tag				- Cache tag to match
//		CHK_IDX				- Index to match
//		match_lbl			- Label to branch to on match
//------------------------------------------------------------------
#macro _aal5_oam_rx_chk_tag(in_tag, CHK_IDX, MATCH_LBL)
.begin
	alu[--, in_tag, -, *l$index0[/**/CHK_IDX/**/]]	
	bne[ctx/**/CHK_IDX/**/#]
	immed[lm_cache_addr, OAM_CTR_CACHE_/**/CHK_IDX/**/_ADDR]
	local_csr_wr[active_lm_addr_0, lm_cache_addr]
	br[MATCH_LBL], defer[2]
		nop
		nop
	nop
.end
#endm //_aal5_oam_rx_chk_tag
//------------------------------------------------------------------
//	MACRO
//		_aal5_oam_rx_upd_tag_on_miss
//
//	DESCRIPTION 
//		Update the cache tag at indicated index. Branch to match lbl
//		if matched. Addittionally load the LM pointer with cache
//		address for that counter type and index
//		
//	ARGUMENTS
//		in_tag				- Cache tag to match
//		IDX					- Index to match
//		ret_lbl				- Label to branch to on return
//------------------------------------------------------------------
#macro	_aal5_oam_rx_upd_tag_on_miss(in_tag, IDX, RET_LBL)
.begin
	; Set the cache tag for this entry
	immed[lm_cache_addr, OAM_CTR_CACHE_/**/IDX/**/_ADDR]
	br[RET_LBL], defer[2]
		local_csr_wr[active_lm_addr_0, lm_cache_addr]
		; Since 3 cycles have not elapsed since last inst.
		; the LM pointer is still pointing to TAG cache
	    alu[*l$index0[/**/IDX/**/], --, B, in_tag]
.end
#endm //_aal5_oam_rx_upd_tag_on_miss
///////////////////////////////////////////////////////////////////////////////
//	MACRO
//		_aal5_rx_oam_clear_cache_tag
//
//	DESCRIPTION 
//	Clear entry matching given tag address
//
//	ARGUMENTS
//
//		
///////////////////////////////////////////////////////////////////////////////
#macro _aal5_rx_oam_clear_cache_tag(in_tag_addr)
.begin
	local_csr_wr[active_lm_addr_0, in_tag_addr]
	; The lm address pointer is still the old one since it has not yet been
	; three cycles since last instruction was invoked.
	immed[lm_ref_count, 0x00]
	nop
	nop
	; Clear the cache tag
	alu[*l$index0, --, B, 0]
.end
#endm
///////////////////////////////////////////////////////////////////////////////
//_aal5_rx_err_phase2_1
//		
//Description:
//
//		It does necessary signalling and wait (error processing)
//		for phase 2_1 for ONE_ME_RX_FP and TWO_ME_RX_FP
//
//Input:  N/A
//
//Output: N/A
//
//
///////////////////////////////////////////////////////////////////////////////
#macro _aal5_rx_err_phase2_1()
.begin
	
	#ifdef ATM_OAM_ENABLED

	// Same for both ONE_ME_RX_FP and TWO_ME_RX_FP

	; signal next thread
	signal_ctx(next_thread_sig_val)

	; To be in sync with ctx_arb[rxc_rd, nxt_thread]
	ctx_arb[next_thread_sig]

	#endif // ATM_OAM_ENABLED
.end
#endm   //   _aal5_rx_err_phase2_1()
#endif  // _AAL5_RX_OAM_UC