#ifndef _AAL5_RX_TWO_ME_FP_UTIL_UC_
#define _AAL5_RX_TWO_ME_FP_UTIL_UC_

/*******************************************************************************
                             Intel Proprietary

 Copyright (c) 1998-2002 By Intel Corporation.  All rights reserved.  
 No part of this program or publication may be reproduced, transmited,
 transcribed, stored in a retrieval system, or translated into any language
 or computer language in any form or by any means, electronic, mechanical,
 magnetic, optical, chemical, manual, or otherwise, without the prior
 written permission of:
                         Intel Corporation
                         2200 Mission College Blvd.
                         Santa Clara, CA  95052-8119
*******************************************************************************/

/*
 *
 *      File Name: 	aal5_rx_two_me_fp_util.uc
 *
 *      Purpose: 	This file contains utility macros exclusively for 
 *					TWO ME RX FP (Functional Pipeline).
 *					
 *					Option (TWO_ME_RX_FP , SPHY_1_32) on (2 MEs) for OC-48 SPHY_1_32
 *
 *      History:
 *
 *
 *      Date            Comment                         By
 *      ------------------------------------------------------------------------
 *
 *		09/19/2002		Created							
 *   	06/20/2003		Merged code with ITP		
 */


////////////////////////////////////////////////////////////////////////////////
//    _aal5_rx_get_rxc_cs1:
//
//	  Description:
//
//			This is for TWO_ME_RX_FP only.
//	
//			It checks whether the RXC1 is in LM or not using CAM_lookup.
//			If its CAM_HIT, we are done. If its CAM_MISS, we get the RXC1 from 
//			SRAM to LM, and update the CAM.
//			
//			(Here RXC1 consists of first 2 LWs of the reassembly context).
//
//		Input:
//			rxc_index :   index of RXC into the RXC table	
//		
//		Outout:
//			new_rxc_tag :	absolute address of RXC	(= CAM tag)
//
//		Size:
//			12 cycles (worst case, CAM_MISS), 9 cycles (best case, CAM_HIT)
///////////////////////////////////////////////////////////////////////////////

#macro _aal5_rx_get_rxc_cs1(new_rxc_tag, rxc_index)
.begin

	.sig rxc_read_sig			; signal to read in new RXC from SRAM

	.reg cam_result, cam_entry

	.reg tmp

	; As VC Info is 16 LWs, so <<4, but it wants in bytes so <<6.

	; cam tag = rxc_base + (index <<6) { where index = [0, 64K-1] }
	; cam tag = rxc_abs_addr - this is because CAM tag cannot be 0.

	alu_shf[tmp, --, b, rxc_index, <<AAL5_RX_VC_INFO_ENTRY_SIZE_LOG2]
	alu[new_rxc_tag, vc_info_sram_base, +, tmp]


	; CAM lookup for RXC using rxc_ptr.
	cam_lookup[cam_result, new_rxc_tag], lm_addr0[0]

	; Check for cam_hit or cam_miss.
	br_bset[cam_result, 7, CAM_HIT#]

CAM_MISS#:	

	; Get the cam entry
	alu_shf[cam_entry, 0xf, and, cam_result, >>CAM_ENTRY_SHF_VAL]	

	; Update the cam entry with the new RXC first before ctx_arb, so that
	; next thread gets a hit for the same RXC. Or the same entry is not
	; returned as LRU with the cam_miss the next time.

	; CAM_STATE_VALID state is not used now.
	cam_write[cam_entry, new_rxc_tag, CAM_STATE_VALID] 
	
	; Set reference count to 1.
	alu[*l$index0[RXC_LM_REF_CNT_OFFSET], --, b, 0x1]

	; Allocate xfer registers for reading context.
	xbuf_alloc($$rxc_in, 2, read)
	
	; Read in new RXC.
	sram[read,$$rxc_in[0], new_rxc_tag, 0, 2], sig_done[rxc_read_sig]

	; wait for signals.
	ctx_arb[rxc_read_sig, next_thread_sig], AND

	; Copy RXC from SRAM to LM.
	br[GET_RXC_DONE#], defer[2]
	alu[lm_sop_alloc_valid_aal_flist,   --,b, $$rxc_in[0]]		
	alu[lm_crc_residue,                 --,b, $$rxc_in[1]]		

	xbuf_free($$rxc_in)

CAM_HIT#:

	; Increment the reference count by 1.
	alu[*l$index0[RXC_LM_REF_CNT_OFFSET], 1, +, *l$index0[RXC_LM_REF_CNT_OFFSET]]

	; Wait for next_thread_signal.
	ctx_arb[next_thread_sig]

GET_RXC_DONE#:

.end
#endm   //   _aal5_rx_get_rxc_cs1()

////////////////////////////////////////////////////////////////////////////////
//    _aal5_rx_write_rxc_to_sram_cs1:
//
//	  Description:
//
//			This is for TWO_ME_RX_FP.
//	
//			It writes back RXC1 to SRAM
//			
//			(Here RXC1 consists of first 2 LWs of the RXC).
//
//		Input:
//			$rxc_out :  sxfer regs
//
//			rxc_add : the absolute address of RXC1
//
//			signal_number : signal number to use for SRAM write
//
//		Outout:
//			None
//
//		Size:
//			3 cycles 
///////////////////////////////////////////////////////////////////////////////

#macro _aal5_rx_write_rxc_to_sram_cs1($rxc_out, rxc_add, signal_number)
.begin

	; Write back RXC1:LW0 and RXC1:LW1 to SRAM
	alu[$rxc_out[0], --, B, lm_sop_alloc_valid_aal_flist]   
	alu[$rxc_out[1], --, B, lm_crc_residue]	 

	sram[write,$rxc_out[0], rxc_add, 0, 2], sig_done[signal_number]

.end
#endm   //   _aal5_rx_write_rxc_to_sram_cs1()

////////////////////////////////////////////////////////////////////////////////
//    _aal5_rx_get_rxc_cs2:
//		Description:
//
//			This is for TWO_ME_RX_FP.
//	
//			It checks whether the RXC2 is in LM or not using CAM_lookup.
//			If its CAM_HIT, we are done. If its CAM_MISS, we get the RXC2 from 
//			SRAM to LM, and update the CAM.
//
//			(Here RXC2 consists of last 7 LWs of RXC).
//
//		Input:
//			new_rxc_tag :	absolute address of RXC2	(= CAM tag)
//
//		Output:
//			None
//
//		Size:
//			15 cycles (worst case, CAM_MISS), 7 cycles (best case, CAM_HIT)
///////////////////////////////////////////////////////////////////////////////

#macro _aal5_rx_get_rxc_cs2(new_rxc_tag)
.begin

	.sig rxc_read_sig			; signal to read in new RXC from SRAM

	.reg cam_result, cam_entry, tmp

	; cam_tag = rxc_abs_addr - this is because CAM tag cannot be 0

	; CAM lookup for RXC using rxc_ptr.
	cam_lookup[cam_result, new_rxc_tag], lm_addr0[0]

	; Check for cam_hit or cam_miss.
	br_bset[cam_result, 7, CAM_HIT#]


CAM_MISS#:	

	; Get the cam entry.
	alu_shf[cam_entry, 0xf, and, cam_result, >>CAM_ENTRY_SHF_VAL]	

	; Update the cam entry with the new RXC first before ctx_arb, so that
	; next thread gets a hit for the same RXC. Or the same entry is not
	; returned as LRU with the cam_miss the next time.

	; CAM_STATE_VALID state is not used now.
	cam_write[cam_entry, new_rxc_tag, CAM_STATE_VALID] 
	
	; Set reference count to 1.
	alu[*l$index0[RXC_LM_REF_CNT_OFFSET], --, b, 0x1]

	; allocate xfer registers for reading context.
#ifdef USE_OPTIONAL_INFO
	xbuf_alloc($rxc_in, 8, read)
#else
	xbuf_alloc($rxc_in, 7, read)
#endif
	
	; Read in new RXC.	
#ifdef USE_OPTIONAL_INFO
	sram[read, $rxc_in[0], new_rxc_tag, 8, 8], sig_done[rxc_read_sig]
#else
	sram[read, $rxc_in[0], new_rxc_tag, 8, 7], sig_done[rxc_read_sig]
#endif

	; Wait for signals.
	ctx_arb[rxc_read_sig, next_thread_sig], AND

	; Copy RXC from SRAM to LM.
	alu[lm_curr_buf_handle,             --, B, $rxc_in[0]]		; LW 2
	alu[lm_sop_buf_handle,              --, B, $rxc_in[1]]		; LW 3
	alu[lm_sop_buf_sz_offset,           --, B, $rxc_in[2]]		; LW 4
	alu[lm_buf_offset_curr_pkt_sz,      --, B, $rxc_in[3]]		; LW 5
	alu[lm_drp_ex_hdr_prv_prv_buf_addr, --, B, $rxc_in[4]]		; LW 6

#ifdef USE_OPTIONAL_INFO
    alu[lm_optional_info,               --, B, $rxc_in[7]]		; LW 9
#endif

	br[GET_RXC_DONE#], defer[2]
	alu[lm_prv_bf_sz_remain_cell_cnt,   --, B, $rxc_in[5]]		; LW 7
	alu[lm_prv_buf_handle,              --, B, $rxc_in[6]]		; LW 8

	xbuf_free($rxc_in)

CAM_HIT#:

	; Increment the reference count by 1.
	alu[*l$index0[RXC_LM_REF_CNT_OFFSET], 1, +, *l$index0[RXC_LM_REF_CNT_OFFSET]]

	; Wait for next_thread_signal.
	ctx_arb[next_thread_sig]

GET_RXC_DONE#:

.end
#endm   //  _aal5_rx_get_rxc_cs2()

////////////////////////////////////////////////////////////////////////////////
//    _aal5_rx_write_rxc_to_sram_cs2:
//
//	  Description:
//
//			This is for TWO_ME_RX_FP.
//	
//			It writes back RXC2 to SRAM
//			
//			(Here RXC2 consists of last 7 LWs of the RXC).
//
//		Input:
//			$rxc_out :  sxfer regs
//
//			rxc_add : the absolute address of RXC1
//
//			signal_number : signal number to use for SRAM write
//
//		Outout:
//			None
//
//		Size:
//			8 cycles 
///////////////////////////////////////////////////////////////////////////////

#macro _aal5_rx_write_rxc_to_sram_cs2($rxc_out, rxc_add, signal_number)
.begin

	; Write back RXC2:LW2 - LW8 to SRAM

	alu[$rxc_out[0], --, B, lm_curr_buf_handle]
	alu[$rxc_out[1], --, B, lm_sop_buf_handle]
	alu[$rxc_out[2], --, B, lm_sop_buf_sz_offset]
	alu[$rxc_out[3], --, B, lm_buf_offset_curr_pkt_sz]
	alu[$rxc_out[4], --, B, lm_drp_ex_hdr_prv_prv_buf_addr]
	alu[$rxc_out[5], --, B, lm_prv_bf_sz_remain_cell_cnt]
	alu[$rxc_out[6], --, B, lm_prv_buf_handle]

	sram[write,$rxc_out[0], rxc_add, 8, 7], sig_done[signal_number]

.end
#endm   //   _aal5_rx_write_rxc_to_sram_cs2()




#endif  // _AAL5_RX_TWO_ME_FP_UTIL_UC_
