 /*****************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 1999-2003 By Intel Corporation.  All rights reserved.
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *****************************************************************************/

/*****************************************************************************
 * File:  csix_rx.c
 *
 * Abstract: 
 *        CSIX RX microblock recevies cframes from the switch fabrics in the 
 *        egress pipeline, reassembly cframes into packet then forwards the 
 *		  next microblock in the pipeline. This file contains the main (outer)
 *        loop of CSIX RX micrblock.
 *
 * Contents:
 *        Declaration of following global variables:
 *            rxCtxLm
 *            dlBufHandle
 *            dlEopBufHandle
 *            prefetchHandle
 *            dlRing
 *            rxFreelistReg
 *            dramHandle
 *            randBufOffset
 *            rsw
 *            prpnd
 *
 *        Declaration of following global SIGNALs: 
 *            sig_nextThrd_1
 *            sig_nextThrd_2
 *            sig_nextThrd_3
 *            sig_dramWrite
 *            sig_rxEvent
 *
 *        Declaration of following global SIGNAL_MASK: 
 *            sinkSigMask;
 *
 *        Definitions of following main function:
 *            main()
 *        Definitions of following local function:
 *            csix_rx      
 *
 *****************************************************************************/

#ifndef __CSIX_RX_C__
#define __CSIX_RX_C__

#define	INLINE	 static __forceinline

#define TMP_FIX_DL_METAFLUSH
#define USE_SXFER_FOR_DLMETA  /* this get better performance */

// queue number in the third word of microblocks interface message uses
// classId from cframe extension header (although 8 bits in extension,
// header but actually only 4 bits as it was prepared in ingress pipeline
#define QUEUE_NUMBER_USES_CLASS_ID_IN_EXTENSION_HEADER

/*
 * MSF related constants
 * will go to harware.h file later
 */

#include "csix_rx.h"
#include "csix_rx_init.c"
#include "csix_rx_util.c"
#include "csix_rx_err_proc.c"
#include "csix_rx_sep_proc.c"
#include "csix_rx_sop_proc.c"
#include "csix_rx_eop_proc.c"
#include "csix_rx_mop_proc.c"
#include "csix_rx_buf_zero_proc.c"


SIGNAL_MASK	sinkSigMask;
SIGNAL sig_nextThrd_1;
SIGNAL sig_nextThrd_2;
SIGNAL sig_nextThrd_3;
SIGNAL_PAIR sig_dramWrite;

/*
 * This signal is sent by the MSF when pushing the RBUF element status,
 * indicating that a new RBUF element is available.
 */
SIGNAL	sig_rxEvent;

__declspec(local_mem shared aligned(16)) rxc_t rxCtxLm[TOTAL_CAM_ENTRY];

dl_buf_handle_t dlBufHandle;			
dl_buf_handle_t dlEopBufHandle;		

__declspec(gp_reg) int dlNextBlock;			


/* buffer handle and datahandle for next iteration */
__declspec(sram_read_reg) dl_buf_handle_t prefetchHandle;
uint32_t prefetchDataHndl;

uint32_t dlRing;			
uint32_t rxFreelistReg;
uint32_t dramHandle;

__declspec(gp_reg shared) uint32_t randBufOffset;
__declspec(gp_reg shared) same_me_signal_t ntSigGpr_1;
__declspec(gp_reg shared) same_me_signal_t ntSigGpr_2;
__declspec(gp_reg shared) same_me_signal_t ntSigGpr_3;

/* define registers to receive the RBUF element status pushed by the MSF */
__declspec(sram_read_reg) rsw_t rsw;
__declspec(sram_read_reg) prepend_t prpnd;


/*
 * used to watch any variable's value while debugging,
 * define _DEBUG_ to use this.
 */
#define SCRATCH_DUMP_BASE 14336
void Dump(int argAddr, int argData)
{
#ifdef _DEBUG_
	__declspec(sram_write_reg) int data = argData;
	SIGNAL sig_DumpScrWrite;

	scratch_write(&data, (__declspec(scratch) void *)(SCRATCH_DUMP_BASE+argAddr*4), \
			1, ctx_swap, &sig_DumpScrWrite);
#endif
}

/*****************************************************************************
 * Abstract:
 *    Infinite CSIX frame receiving processing.
 *
 * Side Effects:
 *    None.
 *
 * Assumptions:
 *    This function will work properly only after proper initialization of
 *    hardware and software.
 *
 *****************************************************************************/

INLINE void 
CsixRx()
{
	uint32_t eleAddr=rsw.element << PO2_RBUF_SIZE;
	uint32_t tmOffset;
	uint32_t classId;
	uint32_t payloadLen;
	uint32_t extClass;

	/* This thread is waked up by receiving logic.
	 * A cframe is coming.  For the null cframe case, 
	 * keep context pipeline.
	 * The RSW is in the rsw data structure
	 */
	
	extClass=rsw.extClass;
	payloadLen=rsw.payloadLen;

	/* If the CSIX frame contains no RSW errors */
	if ((rsw.null == 0) && (rsw.err == 0))
	{
		if(prefetchHandle.lw_offset)
		{
			int ingressBlade;		//ingress blade number
			SIGNAL sig_msfRead;
		
			/* start of normal_proc_phase_1 */
			/* get prepend header (at max 4 LW) */ 
			msf_read((__declspec(sram_read_reg) void *)&prpnd, 
					(volatile void *)(RBUF_TBUF_BASE + eleAddr), 4, sig_done,
					 &sig_msfRead);

			wait_for_all(&sig_nextThrd_1, &sig_msfRead);
			__implicit_read(&sig_msfRead);

			/* start of normal_prc_phase_2 */
			SignalNextCtx(ntSigGpr_1);
			tmOffset=prpnd.alignOffset;
			ingressBlade=prpnd.ingressSrcBlade;
			classId=prpnd.classId; 
		
			if((prpnd.sop == 1) && (prpnd.eop == 1))
			{
				CsixRxSepProc(eleAddr, payloadLen, tmOffset, extClass, ingressBlade);
			}
			else if(prpnd.sop == 1)
			{
				CsixRxSopProc(eleAddr, payloadLen, tmOffset, extClass, ingressBlade);
			}
			else if(prpnd.eop == 1)
			{
				CsixRxEopProc(eleAddr, payloadLen, tmOffset, extClass, ingressBlade);
			}
			else
			{
				CsixRxMopProc(eleAddr, payloadLen, tmOffset, extClass, ingressBlade);
			}
		}
		else 
		{
			CsixRxBufZeroProc(eleAddr, payloadLen, extClass);
		}
	}
	else
	{
		CsixRxErrProc(eleAddr, extClass);
	}
} /* CsixRx() */


/*****************************************************************************
 * Abstract:
 *    This function in the main function of CSIX RX microblock. It first 
 *    initialized CSIX RX microblock, then enter infinite CSIX frame 
 *    receiving process.
 *
 * Side Effects:
 * 		All global variables will be affected
 *
 * Assumptions:
 *      None 
 *
 * Return: 
 * 		Nothing
 *
 *****************************************************************************/

void 
main()
{	
	CsixRxInit();		/* in csix_rx_init.c */

	__implicit_write((void *)&rsw);

	while (1)
	{		
		CsixRx();	
	}
	
	__implicit_read(&sig_nextThrd_1);
	__implicit_read(&sig_nextThrd_2);
	__implicit_read(&sig_nextThrd_3);

	__free_write_buffer((void *)&rsw);
} /* main */

#endif // __CSIX_RX_C__ 
