 /*****************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 1999-2003 By Intel Corporation.  All rights reserved.
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *****************************************************************************/

/*****************************************************************************
 * File:  csix_rx.h
 *
 * Abstract: 
 *        Header file for CSIX RX microblock in egress pipeline.
 *
 * Contents:
 *        Definitions of constants used in CSIX RX microblock
 *        Declaration of external variables and functions used in 
 *            CSIX RX microblock 
 *
 *****************************************************************************/



#ifndef __CSIX_RX_H__
#define __CSIX_RX_H__

/*
 * include files
 */
#include <ixp.h>
#include <dl_system.h>
#include <sig_functions.h>
#include <dl_buf.c>
#include <dl_meta.c>
#include <dl_source.c>
#include <sig_functions.c>
#include "hardware.h"

/*************************************************************************************/

/*
 *	ISO C99: 7.18 Integer types <stdint.h> : For now define them here
 */

#ifndef _STDINT_H
#define _STDINT_H   1

/* Exact integral types.  */
/* Signed.  */

typedef signed char     int8_t;
typedef short int       int16_t;
typedef int             int32_t;
typedef long long int   int64_t;

/* Unsigned.  */

typedef unsigned char           uint8_t;
typedef unsigned short int      uint16_t;
typedef unsigned int            uint32_t;
typedef unsigned long long int  uint64_t;

#endif /* _STDINT_H */

/*
 * RX_THREAD_FREELIST register field bits definition  
 */
#define SIGNAL_NUMBER_BITS				12		/* 12-15 bits Signal number */
#define ME_NUMBER_BITS					7		/* 7-9 bits Microengine number */
#define THREAD_BITS						4		/* 4-6 bits Thread number */

/*
 * MSF related user definitions.
 */
/* Thread time-out interval (in bus clocks) */
#define RX_THREAD_FREELIST_TIMEOUT_0_DATA	1024
#define RX_THREAD_FREELIST_TIMEOUT_1_DATA	0
#define RX_THREAD_FREELIST_TIMEOUT_2_DATA	0
#define RX_THREAD_FREELIST_TIMEOUT_3_DATA	0

#define RX_UP_CTRL_0_DATA					0x0d

/* Mask to wrap around at MAX_BUFFER_OFFSET (in dl_system.h) */
#define	RAND_OFFSET_MASK		(MAX_BUFFER_OFFSET - 1)
/* Setup maximum payload size */
#define MAX_PAYLOAD				(BUFFER_SIZE - PPP_HEADER)

/* User defined RBUF definitions */
#define	RBUF_TOTAL_BYTES		8192
#define RBUF_SIZE				128
#define RBUF_SIZE_IN_8B			16
#define	RBUF_TOTAL_ELEMENTS		(RBUF_TOTAL_BYTES / RBUF_SIZE)

/* the size of each RXC in longwords (32 bits) */
#define RXC_SIZE				8

/* bit shift to convert element number to address and vice vesa */
#define PO2_RBUF_SIZE			7

/* per packet header size */
#define	PER_PKT_PREPEND_BYTES	8				
/* cell size */
#define CELL_SIZE				120  			

#ifdef IXP2800
#define MSF_RX_CONTROL_DATA_DISABLE		0x00000055
#define MSF_RX_CONTROL_DATA_ENABLE		0x80000055
#else	/* IXP2800 */
#define MSF_RX_CONTROL_DATA_DISABLE		0x00410204
#define MSF_RX_CONTROL_DATA_ENABLE		0x10410204
#endif /* IXP2800 */

#define MSF_CSIX_TYPE_MAP_DATA	0x08

#define HWM_CONTROL_DATA		0x0F

#define TOTAL_CAM_ENTRY		16

#define TOTAL_RXC_ENTRY_IN_SRAM		1024

#define	NORMAL_STATE		1
#define SOP_PRPEND_HDR 		16
#define NO_SOP_PRPEND_HDR 	8

#define	IX_DL_BUF_NULL		IX_NULL


#define SCRATCH_RING_FULL(N) inp_state_scr_ring##N##_full
#define CSIX_RX_SINK_RING_FULL(N) SCRATCH_RING_FULL(N)

#define DDR_BANK_SIZE		128
#define NO_OF_DDR_BANKS		4
#define MAX_BUFFER_OFFSET	(DDR_BANK_SIZE * NO_OF_DDR_BANKS)

#define PKT_DIV		8

#ifdef COUNTERS
#define		PKT_RECEIVED_CNT_IDX	0x00
#define		PKT_DROPPED_CNT_IDX		0x04
#define		BYTE_RECEIVED_CNT_IDX	0x08
#define 	RESERVED_CNT_IDX		0x0C
#endif /* COUNTERS */

#ifdef CELL_MODE
// definition of ATM cell size
#define CELL_SIZE_CNST 	48
#endif /* CELL_MODE */

typedef enum {
	STATE_START =0,
	STATE_PROC
} RX_STATE;

#define IX_BUF_MAX 	BUFFER_SIZE



/*
 * CSIX RX microc code related user definitions 
 */

/*************************************************************************
 * Abstract
 *    Receive status word
 *
 * Purpose/Intended Use:
 *    This structure holds the information in receive status word (RSW)
 *    in CSIX RX mode
 * 
 ************************************************************************/
typedef __declspec(packed) struct rsw_s
{
 union
 {
  struct
  {
    uint32_t reserved02   	: 1;
    uint32_t element      	: 7;
    uint32_t payloadLen  	: 8;
    uint32_t csixReserved	: 1;
    uint32_t private        : 1;
    uint32_t err    	    : 1;
    uint32_t lenErr			: 1;
    uint32_t hpErr		    : 1;
    uint32_t vpErr	 		: 1;
    uint32_t null         	: 1;
    uint32_t reserved01		: 5;
    uint32_t type			: 4;
	uint32_t extClass		: 8;
	uint32_t extPrivate		: 2;
	uint32_t extCsixRsved   : 10;
	uint32_t extDestAddr	: 12; 
  };
  uint32_t value[2];
 };
} rsw_t;

/*************************************************************************
 * Abstract
 *    Prepend header
 *
 * Purpose/Intended Use:
 *    This structure holds the information in prepend header of CSIX 
 *    frame
 * 
 ************************************************************************/

typedef __declspec(packed) struct prepend_s
{
 union
 {
  struct
  {
    uint32_t prpnd0;
    uint32_t prpnd1;
    uint32_t prpnd2;
    uint32_t prpnd3;
  };
  struct
  {
  	uint32_t alignOffset		: 8;
	uint32_t seqNumb			: 16;
	uint32_t eop				: 1;
	uint32_t sop				: 1;
	uint32_t ingressSrcBlade	: 6;
	uint32_t nextHopId			: 16;
	uint32_t classId			: 16;
	uint32_t inputPort			: 16;
	uint32_t outputPort			: 16;
	uint32_t flowId;
  };
  uint32_t value[4];
 };
} prepend_t;

/*************************************************************************
 * Abstract
 *    Receive context
 *
 * Purpose/Intended Use:
 *    This structure holds the information in receive context which are  
 *    stired in local memory or sram
 * 
 ************************************************************************/
typedef __declspec(packed) struct rxc_s
{
 union
 {
  struct
  {
  	uint32_t pktSize			: 16;	//packet number
  	uint32_t cellSeqNum			: 16;   //cell sequence number
	dl_buf_handle_t sopBufHandle;
	dl_buf_handle_t curBufHandle;
  	uint32_t curDataHandle;				 	//current data handle
  	uint32_t curBufSize			: 16;	//current buffer size
  	uint32_t sopBufOffset		: 16; 	//current buffer offset
  	uint32_t classId			: 16; 	//Class ID
  	uint32_t outputPort			: 16; 	//output port number
  	uint32_t flowId;
  	dl_buf_handle_t prvBufHandle;			//previuos buffer handle
  };
  struct
  {
  	uint32_t reserved08;
  	uint32_t reserved07;
  	uint32_t reserved06;
  	uint32_t reserved05;
  	uint32_t reserved04;
  	uint32_t reserved03;
  	uint32_t reserved02;
  	uint32_t remainder			:  8;	//8 bits reminder 	
  	uint32_t reserved01			: 24;	
  };
  uint32_t value[8];
 };
} rxc_t;


/* for details on these definitions, refer to files csix_rx.c, csix_rx_util.c. */
extern SIGNAL sig_nextThrd_1;
extern SIGNAL sig_nextThrd_2;
extern SIGNAL sig_nextThrd_3;
extern SIGNAL_PAIR sig_dramWrite;


extern SIGNAL sig_rxEvent;
extern SIGNAL_MASK	sinkSigMask;
extern __declspec(sram_read_reg) rsw_t rsw;
extern uint32_t rxFreelistReg;
extern uint32_t dramHandle;

extern __declspec(sram_read_reg) prepend_t prpnd;

extern __declspec(local_mem shared aligned(16)) rxc_t rxCtxLm[TOTAL_CAM_ENTRY];

extern __declspec(sram_read_reg) dl_buf_handle_t prefetchHandle;
extern uint32_t prefetchDataHndl;

extern uint32_t	dlRing;				
extern __declspec(gp_reg) int dlNextBlock;			
extern dl_buf_handle_t dlBufHandle;			
extern dl_buf_handle_t dlEopBufHandle;		

extern __declspec(gp_reg shared) uint32_t randBufOffset;
extern __declspec(gp_reg shared) same_me_signal_t ntSigGpr_1;
extern __declspec(gp_reg shared) same_me_signal_t ntSigGpr_2;
extern __declspec(gp_reg shared) same_me_signal_t ntSigGpr_3;

extern dl_meta_t dlMeta;

#define NON_SOP_OFFSET		0 

/* will be moved to dl_system.h later */
#define			UC_NULL							0xFF		/* NULL for microcode */
#define			BID_NULL						UC_NULL		/* NULL Block ID */
#define			BID_EXCEPTION					0x1			/* Exception Block ID */
#define			BID_DROP						0x2			/* Drop Block ID */

#define	DL_BUF_NULL			UC_NULL 	/* NULL Buffer representation */

/*
 * Block IDs for various blocks. Block IDs Start from 0x20. (0 to 0x1f and 0xFF
 * are reserved for IX_*)
 */
#define			BID_POS							0x21
#define			BID_IPV4						0x22
#define			BID_QM							0x23

/* Following define the port type */
#define			HDR_ETHER						0x01
#define 		HDR_PPP							0x02

void FreeRbufElement(int);
void FreeAllRbufElements();
void AddThreadToFreelist();
void CsixRxInitAllRxc();

uint32_t GetRandBufOffset();
void MoveRbufToPktBuf(uint32_t, int, uint32_t);
void DropCurrentPkt(dl_buf_handle_t sopBuf, dl_buf_handle_t curBuf, int voqNum);

void CsixRxInit();
void CsixRx();

void Dump(int, int);

void ConfigMsfRxCtrl();
void SetCsixTypeMap(uint32_t msfCsixTypeMapData);
void SetAllCamEntryState(uint32_t stateValue);
void CsixRxSink(dl_buf_handle_t sopBufHndl, dl_buf_handle_t	eopBufHndl,
		uint32_t packetSize, uint32_t cellCount, uint32_t class,
		uint32_t port, uint32_t ring, uint32_t	voqNum,	SIGNAL_MASK	sigMask);
void GetCellCount(uint32_t *pCellCount, uint32_t *pRemainder,
		uint32_t byteCount);

void CsixRxSepProc(uint32_t eleAddr, uint32_t payloadLen, 
						uint32_t tmOffset, uint32_t extClass,
						uint32_t ingressBlade);
void CsixRxSopProc(uint32_t eleAddr, uint32_t payloadLen,
						uint32_t tmOffset, uint32_t extClass,
						uint32_t ingressBlade);
void CsixRxEopProc(uint32_t eleAddr, uint32_t payloadLen, 
						uint32_t tmOffset, uint32_t extClass, 
						uint32_t ingressBlade);
void CsixRxMopProc(uint32_t eleAddr, uint32_t payloadLen, 
						uint32_t tmOffset, uint32_t extClass, 
						uint32_t ingressBlade);
void CsixRxBufZeroProc(uint32_t eleAddr, uint32_t payloadLen, 
						uint32_t extClass);
void CsixRxErrProc(uint32_t eleAddr, uint32_t extClass);


#endif // __CSIX_RX_H__ 
