 /*****************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 1999-2003 By Intel Corporation.  All rights reserved.
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *****************************************************************************/

/*****************************************************************************
 * File:  csix_rx_util.c
 *
 * Abstract: 
 *        MicroC file for CSIX RX microblock utility functions.
 *
 * Contents:
 *        Definitions of following local functions:
 *            ConfigMsfRxCtrl         
 *            SetCsixTypeMap
 *            SetAllCamEntryState       
 *            CsixRxInitAllRxc         
 *            MoveRbufToPktBuf         
 *            FreeRbufElement     
 *            FreeAllRbufElements         
 *            AddThreadToFreelist        
 *            GetRandBufOffset   
 *            DropCurrentPkt       
 *
 *****************************************************************************/
#ifndef __CSIX_RX_UTIL_C__
#define __CSIX_RX_UTIL_C__

#include "csix_rx.h"

/*****************************************************************************
 * Abstract:
 *    Initialized hardware MSF registers.
 *
 * Side Effects:
 *    None
 *
 * Assumptions:
 *    None
 *
 *****************************************************************************/
INLINE void
ConfigMsfRxCtrl(
)
{
	 SIGNAL	sig_msfWrite;
	 __declspec(sram_write_reg) int msfRxControlReg = MSF_RX_CONTROL_DATA_DISABLE;
	// disable first, required by hardware
	msf_write(&msfRxControlReg, (volatile void *)MSF_RX_CNTL, 1, sig_done, 
				&sig_msfWrite);
	wait_for_all(&sig_msfWrite);
	__implicit_read(&msfRxControlReg);		
	__implicit_read(&sig_msfWrite);

	// Free all RBUF elements.
	FreeAllRbufElements();

	// set HWM_CONTROL msf reg
	msfRxControlReg = HWM_CONTROL_DATA;
	msf_write(&msfRxControlReg, (volatile void *)HWM_CONTROL, 1, sig_done, 
				&sig_msfWrite);
	wait_for_all(&sig_msfWrite);
	__implicit_read(&msfRxControlReg);				
	__implicit_read(&sig_msfWrite);

	SetCsixTypeMap(MSF_CSIX_TYPE_MAP_DATA);

	// Do a fast write to the RX_THREAD_FREELIST_TIMEOUT_0, 1, 2, 3 CSR. 
	// Note ms 16 bits is data, and ls 16 bits is address
	msf_fast_write((void *)(RX_THREAD_FREELIST_TO_0 | (RX_THREAD_FREELIST_TIMEOUT_0_DATA << 16)));
	msf_fast_write((void *)(RX_THREAD_FREELIST_TO_1 | (RX_THREAD_FREELIST_TIMEOUT_1_DATA << 16)));
	msf_fast_write((void *)(RX_THREAD_FREELIST_TO_2 | (RX_THREAD_FREELIST_TIMEOUT_2_DATA << 16)));
	msf_fast_write((void *)(RX_THREAD_FREELIST_TO_3 | (RX_THREAD_FREELIST_TIMEOUT_3_DATA << 16)));
	
	// enable
	msfRxControlReg = MSF_RX_CONTROL_DATA_ENABLE;
	msf_write(&msfRxControlReg, (volatile void *)MSF_RX_CNTL, 1, sig_done, 
				&sig_msfWrite);
	wait_for_all(&sig_msfWrite);			
	__implicit_read(&msfRxControlReg);		
	__implicit_read(&sig_msfWrite);
} /* ConfigMsfRxCtrl */


INLINE void SetCsixTypeMap(uint32_t msfCsixTypeMapData)
{
	 SIGNAL	sig_msfWrite;
	 __declspec(sram_write_reg) int msfCsixTypeMap = msfCsixTypeMapData;
	msf_write(&msfCsixTypeMap, (volatile void *)CSIX_TYPE_MAP, 1, sig_done, 
				&sig_msfWrite);
	wait_for_all(&sig_msfWrite);	
	__implicit_read(&msfCsixTypeMap);		
	__implicit_read(&sig_msfWrite);
} /* SetCsixTypeMap */


INLINE void SetAllCamEntryState(uint32_t stateValue)
{
	// Used to index CAM locations.
	int index;

	// Write 16 data values to the 16 CAM entries.
	for(index=0; index<TOTAL_CAM_ENTRY; index++)
		cam_write(index, index, stateValue);
} /* SetAllCamEntryState */

// Initialize all RXC's in LM & sram 
INLINE void CsixRxInitAllRxc()
{
	int i;
	
	for(i=0; i<TOTAL_CAM_ENTRY; i++)
	{			
		rxCtxLm[i].value[0]=0;
		rxCtxLm[i].value[1]=0;
		rxCtxLm[i].value[2]=0;
		rxCtxLm[i].value[3]=0;
		rxCtxLm[i].value[4]=0;
		rxCtxLm[i].value[5]=0;
		rxCtxLm[i].value[6]=0;
		rxCtxLm[i].value[7]=0;
	}

} /* CsixRxInitAllRxc */


INLINE void MoveRbufToPktBuf(uint32_t dramAddr, int elementAddr, uint32_t byteCount)
{
	dram_rbuf_tbuf_ind_t ind;

	ind.value=0;		// reset
	ind.ov_buf_addr = 1;
	ind.buf_addr = RBUF_TBUF_BASE + elementAddr;
	ind.ov_ref_count = 1;
	ind.ref_count = (byteCount - 1) >> 3;
	dram_rbuf_read_ind((volatile __declspec(dram) void *)dramAddr, 16, ind, sig_done, 
						&sig_dramWrite);
} /* MoveRbufToPktBuf */

INLINE void FreeRbufElement(int elementNum)
{
	// Do a fast write to the RBUF_ELEMENT_DONE CSR.
	// Note ms 16 bits is data, i.e. elementNum.
	// and ls 16 bits is address, i.e. RBUF_ELEMENT_DONE.
	msf_fast_write((void *)(RBUF_ELEMENT_DONE | (elementNum << 16)));
} /* FreeRbufElement */

INLINE void FreeAllRbufElements()
{
	int elementNum;

	for(elementNum = 0; elementNum < RBUF_TOTAL_ELEMENTS; elementNum++)
		FreeRbufElement(elementNum);
} /* FreeAllRbufElements */

INLINE void AddThreadToFreelist()
{
	// Do a fast write to the RX_THREAD_FREELIST_0 CSR. 
	// Note ms 16 bits is data, and ls 16 bits is address,
	// rxFreelistReg has been initialized to that value in CsixRxInit.
	msf_fast_write((void *)rxFreelistReg);
} /* AddThreadToFreelist */


INLINE uint32_t GetRandBufOffset()
{
	// Possible values for randBufOffset : 128, 256, 384, 512
	randBufOffset = (randBufOffset & RAND_OFFSET_MASK) + DDR_BANK_SIZE;

	return randBufOffset;
} /* GetRandBufOffset */


INLINE void DropCurrentPkt(dl_buf_handle_t sopBuf, dl_buf_handle_t curBuf, int voqNum)
{

	//	There are two conditions to handle
	//	1. Only a single buffer needs to be dropped.
	//	2. A buffer chain needs to be dropped. (large packet)
	//
	//	In case 1, sop_handle = curr_handle and both will have the 
	//	SOP bit set. (This is a little peculiar but it makes life
	//	easier elsewhere)
	//
	//	In case 2, sop_handle != curr_handle and only sop_handle
	//	will have the SOP bit set.

#ifdef COUNTERS
	sram_incr((volatile void *)(CSIX_RX_COUNTER_BASE + (voqNum << 4) + 
				PKT_DROPPED_CNT_IDX));
#endif	/* COUNTERS */

	if (curBuf.lw_offset == sopBuf.lw_offset) 
		Dl_BufDrop(curBuf);
	else
		Dl_BufDropChain(sopBuf, curBuf);

} /* DropCurrentPkt */


/////////////////////////////////////////////////////////////////////////////
// CsixRxSink
//	
//	Description:
//		Send received packet to the next processing block by writing scratch ring
//
//	Outputs:	
//		None
//		
//	Inputs:
//		sopBufHndl:		SOP buffer handle register, this register value will be modified
//						after macroc is excuted.  If the caller want to keep this value  
//						the value should be stored some where 
//		eopBufHndl:		Encoded EOP buffer handle: NULL for one buffer cframe
//						otherwise will be real EOP handle  
//		packetSize:		packet size 
//		class:			Class number
//		port:			Output port number
// 		cellCount:		cell count for SOP buffer : used in cell mode 
//		ring:		 	scratch ring to sink 
//		voqNum:			virtual queue number
//		sigMask:		signal maske to pass to dl_sink
//
//	CONSTANTS:	
//		None
//	
//
//	Size: 10 instructions
/////////////////////////////////////////////////////////////////////////////
INLINE void 
CsixRxSink(
	dl_buf_handle_t sopBufHndl,
	dl_buf_handle_t	eopBufHndl,
	uint32_t	packetSize,
	uint32_t	cellCount,
	uint32_t 	class,
	uint32_t	port,
	uint32_t	ring,
	uint32_t	voqNum,			
	SIGNAL_MASK	sigMask)
{
	dl_buf_handle_t pureSopHndl;
	// Check scratch ring fullness

	if(!(inp_state_test(CSIX_RX_SINK_RING_FULL(CSIX_RX_RING_OUT))))
	{	

		pureSopHndl.value = sopBufHndl.value & 0xC0FFFFFF;		/* clean out cell count */
#ifdef CELL_MODE
		// work in cell mode, put cell count in the SOP buf handle	
		/*  set cellCount field */
		pureSopHndl.value |= (cellCount << 24);
#else
		pureSopHndl.value |= ((packetSize >> PKT_DIV) + 1) << 24;	//set pkt size to SOP hndl
#endif														


		dl_sink(pureSopHndl, eopBufHndl, class, port, ring, sigMask);
		__implicit_read(&sig_rxEvent);
	}
	else //	the scratch ring is full, the packet has to be droped
	{
		DropCurrentPkt(sopBufHndl, eopBufHndl, voqNum);	//drop current pkt
		wait_for_all(&sig_rxEvent);
		__implicit_read(&sig_rxEvent);
	}

}

/////////////////////////////////////////////////////////////////////////////
// GetCellCount
//	
//	Description:
//		Compute the cell count 
//
//	Outputs:	
//		None
//		
//	Inputs:
//		pCellCount:	pointer to the the cell count
//		pRemainder:	pointer to the remainder	 
//		byteCount:	The byte count used to compute new cell count
//
//	CONSTANTS:	
//		None
//	
//	Globale variables:
//		CELL_SIZE_CNST:  This is cached cell size
/////////////////////////////////////////////////////////////////////////////

INLINE void 
GetCellCount(
	uint32_t *pCellCount,
	uint32_t *pRemainder,
	uint32_t byteCount)
{
#ifdef CELL_MODE
	*pRemainder = *pRemainder + (uint32_t)byteCount;
	while((*pRemainder - CELL_SIZE_CNST) > 0)
	{
		*pRemainder = *pRemainder - CELL_SIZE_CNST;
		*pCellCount = *pCellCount + 1;
	}
#endif /* CELL_MODE */
} /* GetCellCount */


#endif // __CSIX_RX_UTIL_C__


