 /*****************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 1998-2002 By Intel Corporation.  All rights reserved.
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *****************************************************************************/

/*****************************************************************************
 * File:	packet_rx.c
 *
 * Abstract: Packet RX is protocol (POS & ethernet) agnostic Receive/Reassembly
 *			 Code. This runs on one ME and can support up to 16 channels(ports).
 *			 Typically this block will be the first block in the pipeline and
 *			 after reassembly the packets will go to the next block (typically POS 
 *			 or ethernet block)
 *
 * Contents:
 *        contains the main (outer) loop.
 *
 *****************************************************************************/

#define	INLINE	 static __forceinline


#include "packet_rx.h"
#include "packet_rx_init.c"
#include "packet_rx_util.c"


SIGNAL_MASK	sinkSigMask;
SIGNAL 		nextThreadSig;

// This signal is sent by the MSF when pushing the RBUF element status,
// indicating that a new RBUF element is available.

SIGNAL 		sigRxEvent;


// Define 'TOTAL_PORTS' locations in local memory.
// Each location is a structure of type rxc_t.
// These locations represent 'TOTAL_PORTS' Receive Contexts.

__declspec(local_mem shared aligned(8)) rxc_t rxc[TOTAL_PORTS];

dl_buf_handle_t dlBufHandle;			
dl_buf_handle_t dlEopBufHandle;		

//	XXX - Why do we need it in gp_reg? 

__declspec(gp_reg) int dlNextBlock;			


// Buffer handle and datahandle for next iteration.

__declspec(sram_read_reg) dl_buf_handle_t prefetchBufHandle;
int prefetchDataHandle;

int	dlRing;			//	ring between packet rx and Ipv4: XXX - to be rename d
int rxFreeListReg;
int rxFreeList;     // Rx Free List

// Absolute variables that store DDR DRAM bank scheduling information.

__declspec(gp_reg shared) unsigned int randBufOffset;
//The value of ntSigGpr is different for each thread in SPHY_4_8 and MPHY_4
#if ((RX_PHY_MODE == SPHY_4_8) || (RX_PHY_MODE == MPHY_4))
__declspec(gp_reg) same_me_signal_t ntSigGpr;
#else
__declspec(gp_reg shared) same_me_signal_t ntSigGpr;
#endif

// Define registers to receive the RBUF element status pushed by the MSF.

volatile __declspec(sram_read_reg) rsw_pos_phy_t rsw[2];

/*****************************************************************************
 * Abstract:
 *		PacketRx is the reasembly function. This will be called in an infinite 
 *		loop
 *
 * Side Effects:
 * 		All global variables will be affected
 *
 * Assumptions:
 *    This function will work properly only after proper initialization of
 *    hardware and software.
 *
 *****************************************************************************/

INLINE void 
PacketRx()
{

	int portNum;

	//	At this point we have received an RBUF element from MSF. (See packet_rx_init
	//	if it is first time execution or dl_sink otherwise.)

	// 	Signal next thread. This has some latency (~10 cycles) so we signal as early
	//	as possible so that by the time we ctx swap/arb the next thread is ready to 
	//	run.

	SignalNextCtx(ntSigGpr);

	// Set input port number.
	portNum = rsw[0].channel;
	
	// If the POS packet contains no RSW errors.
	if (rsw[0].null==0 && rsw[0].rxErr==0 && rsw[0].parErr==0)
	{
		// compState is a 3 bit value: SOP bit EOP bit RXC state bit
		int compState = (rsw[0].sop << 2) | (rsw[0].eop << 1) | (rxc[portNum].state);

		//  compState
		//	000			START state, MOP mpacket : Error case	- msp
		//	001			PROC  state, MOP mpacket : Normal case	- mpp
		//	010			START state, EOP mpacket : Error case	- esp
		//	011			PROC  state, EOP mpacket : Normal case	- epp
		//	100			START state, SOP mpacket : Normal case	- ssp
		//	101			PROC  state, SOP mpacket : Error case	- spp (+ normal case)
		//	110			START state, SOP&EOP pkt : Normal case	- sesp
		//	111			PROC  state, SOP&EOP pkt : Error case	- sepp (+ normal case)

		int byteCount 	= rsw[0].byteCount;
		int eleAddr 	= rsw[0].element << PO2_RBUF_SIZE;

		switch (compState)
		{
			case 0: Msp(portNum, eleAddr, byteCount); break;
			case 1: Mpp(portNum, eleAddr, byteCount); break;
			case 2: Esp(portNum, eleAddr, byteCount); break;
			case 3: Epp(portNum, eleAddr, byteCount); break;
			case 4: Ssp(portNum, eleAddr, byteCount); break;
			case 5: Spp(portNum, eleAddr, byteCount); break;
			case 6: Sesp(portNum, eleAddr, byteCount); break;
			case 7: Sepp(portNum, eleAddr, byteCount); break;
		}

		// Free the RBUF element.
		FreeRbufElement(rsw[0].element);
	}
	else 
	{
		// For all Rx errors, except NULL packet error.
		if (rsw[0].null == 0)
		{
			// Reset state machine to STATE_START.

			rxc[portNum].state = STATE_START;

			// Wait for next thread signal.

			wait_for_all(&nextThreadSig);

			//	XXX - should we drop the packet???
			// DropCurrentPkt(drop_sop_handle, drop_eop_handle)

			dlNextBlock = IX_NULL;
			dlBufHandle.lw_offset = IX_NULL;

			// Free the RBUF element.
			FreeRbufElement(rsw[0].element);
		}
		// NULL packet error.
		else
		{
			// Wait for next thread signal.
			wait_for_all(&nextThreadSig);

			dlNextBlock = IX_NULL;
			dlBufHandle.lw_offset = IX_NULL;
		}
	}

	// Processing common to RSW error and non-error packets.

	//	Signal Next Thread.

	SignalNextCtx(ntSigGpr);

	// Add the thread to the receive freelist.

	AddThreadToFreelist(rxFreeList, rxFreeListReg);

	//	Reset the signal mask.

	sinkSigMask = 1 << __signal_number(&sigRxEvent);

}

/*****************************************************************************
 * Abstract:
 *		Main loop. Each iteration will process a mpacket (<= 128 bytes). PacketRx
 *		will do the reassembly and dl_sink will feed data to the next block in 
 *		the pipeline
 *
 * Side Effects:
 * 		All global variables will be affected
 *
 * Assumptions:
 *
 *****************************************************************************/
#ifdef WESTPORT
void
Ethernet_Rx()
#else
void main()
#endif
{	
	//	XXX - The compiler doesn't know the live range of rsw (sram read regsiters)
	//	problem: rsw[2] and prefetchBufHandle gets the same read xfer register.
	//	So we manually assign the registers.



	PacketRxInit();					// In pos_rx_init.c

	while (1)
	{		
		__implicit_write((void*)&rsw);

		PacketRx();					

		DlSink(sinkSigMask);		// In dl_source.c

		__implicit_read(&sigRxEvent);
	}


	//	XXX - Live range of the rsw registers ends here.

	__free_write_buffer((void *)&rsw[0]);
	__free_write_buffer((void *)&rsw[1]);
}