///////////////////////////////////////////////////////////////////////////////
//
//
//                  I N T E L   P R O P R I E T A R Y
//
//     COPYRIGHT [c]  2002 BY  INTEL  CORPORATION.  ALL RIGHTS
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT
//     THE PRIOR WRITTEN PERMISSION OF :
//
//                        INTEL  CORPORATION
//
//                     2200 MISSION COLLEGE BLVD
//
//               SANTA  CLARA,  CALIFORNIA  95052-8119
//
///////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: packet_rx_two_me_util.uc
//
//      Purpose: Contains macros used when packet rx runs on two ME's.
//
//      History:
//
//      Date            Comment                         By
//      ---------------------------------------------------------------------
//
//      11/11/2002      Created                         asv
//
/////////////////////////////////////////////////////////////////////////////

#ifndef	__PACKET_RX_TWO_ME_UTIL_UC__
#define	__PACKET_RX_TWO_ME_UTIL_UC__

#include <dl_system.h>
#include <stdmac.uc>
#include <dispatch_loop.uc>
#include <hardware.h>
#include <packet_rx.h>
#include <packet_rx_context.h>
#include <packet_rx_common_util.uc>


//--------------------------------
//	First Packet RX ME - Begin
//--------------------------------
#ifdef FIRST_PACKET_RX_ME

//	Some Local Definitions

//	From context size, compute the power of 2. Needed for later computation.
#define_eval 	PO2_CTXSIZE		log2(CONTEXT_SIZE)

//	Global Signals

.sig volatile sig_next_thread					// next thread signal
.sig volatile rx_event							// receive wakeup signal

.sig drop_sig									// used when freeing buffer chains

//	needed because <<&rx_event is not recognised by assembler.
.addr rx_event RX_EVENT_SIG

//	Global Regsiters

//	Following registers are used as constants to save instructions
//	If we run out of GPRs, these are prime candidates for removal.
//	If you run out of GPRs use the following criteria to decide
//	which register to remove
//
//					Instructions	where is it
//					   saved			saved
// ----------------------------------------------------------------------------
//	free_list_reg		4 			all paths
//	nt_sig_reg			2 + 2 		all paths (twice per iteration)
//	dram_rbuf_ov		2			all paths
//	prx_rbuf_base			1			critical path (i.e min pkt case)
//	@max_buf_size		1 			(not in critical path)
//	protocol_mask		1			all paths (only when using 4 Byte PPP hdr)

.reg	prx_rbuf_base						// base address of rbuf
.reg 	thread_num						// Current thread number
.reg	max_buf_size					// The max buf data size
.reg	freelist_reg					// to put a thread in free list.
.reg	nt_sig_reg						// used in signalling next thread
.reg 	dram_rbuf_ov					// for moving data from rbuf->dram
.reg 	protocol_mask					// PPP hdr has 16bit protocol field
.reg	rx_freelist						// RX free list gpr

#ifdef	PACKET_RX_COUNTERS
.reg	counter_base					// base addr for counters.
#endif

//	Following regsiters are used as global variables

.reg read 	$prx_prefetch_buf_handle		// buffer in which next pkt stored.
										// "read" for protection
.reg	prefetch_buf_handle_gpr			// gpr that holds $prx_prefetch_buf_handle
.reg	prefetch_data_handle			// dram addr of above buffer
.reg	byte_count						// bytes in current mpacket.
.reg	sink_sig_mask					// mask to tell what I/Os were done
.reg	$prx_rsw1, $prx_rsw2					// rbuf read status words
.xfer_order $prx_rsw1, $prx_rsw2

xbuf_alloc($drop, 3, write)				// used when freeing buffer chains

#if 	!IS_IXPTYPE(__IXP28XX)
.reg	@rand_offset_mask				// mask for wrapping around @ 512
.reg	@prx_rand_buf_offset				// offset to use in next(SOP) buffer.
#endif

#ifdef STRIP_4BYTES_CRC

.reg	second_lm_base					// Second Local memory context data base

#endif

#ifdef 	_DEBUG_COUNTERS_
.reg 	@pkt_rx_num_proper_mpkts_rxed	// Good mpackets received
.reg 	@pkt_rx_num_err_mpkts_rxed		// error mpackets received
.reg 	@pkt_rx_buf_alloc_fail			// number of buffer alloc failures
.reg	@pkt_rx_buf_alloc_def
.reg	@pkt_rx_buf_drop
	alu[@pkt_rx_num_proper_mpkts_rxed, --, B, 0]
	alu[@pkt_rx_num_err_mpkts_rxed, --, B, 0]
	alu[@pkt_rx_buf_alloc_fail, --, B, 0]
	alu[@pkt_rx_buf_alloc_def, --, B, 0]
	alu[@pkt_rx_buf_drop, --, b, 0]

#endif // _DEBUG_COUNTERS_

///////////////////////////////////////////////////////////////////////////////
//
// packet_pkt_rx
//
//	 	Description: 
//
// 			Packet processing entry point for packet microblock. In each 
//			iteration one mpacket will be handled. Implements a state machine 
//			to receive SOP, MOP and EOP mpackets.
//
//	 	Outputs:
//
//			All global variables will be affected.
//
//		Inputs:
//
//			Will use global varaibles.
//
///////////////////////////////////////////////////////////////////////////////
#macro packet_pkt_rx()
.begin

	.reg 	ctx_st								// keeps the machine state
	.reg 	$protocol							// protocol in PPP hdr
	.reg	protocol_reg						// protocol in PPP hdr
	.reg 	drop_sop_handle, drop_eop_handle	// SOP & EOP buffers to drop
	.reg 	port_num							// inpur port number
	.reg 	ele_num								// rbuf element number
	.reg 	lm_ctx_base							// base addr of reassembly data

	.set 	$prx_rsw1, $prx_rsw2						// to avoid warnings

	// Signal used while pre-fetching a buffer.
	.sig sig_buf_alloc

comn_rx_phase_1#:

	//	Wakeup next thread. Next thread will be ready by the time
	//	this thread goes to sleep.

	_packet_rx_signal_next_thread(nt_sig_reg)		; Signal next thread

	//	The thread autmatically wakes up on receiving an mpacket.
	//	The receive status words are available in $prx_rsw1 and $prx_rsw2.
	//	If no packet had been recieved within the timeout, then
	//	the status word will indicate that a NULL packet was received.

	_packet_rx_rsw_get_port_num(port_num, $prx_rsw1)	; From which port pkt came

	//	locate the context in local memory and setup local memory. 
	//	There are totally 16 contexts, one per port. Each conext 
	//	is 64 bytes (16 long words)

	alu_shf_left(lm_ctx_base, CONTEXT_DESC_BASE, or, port_num, PO2_CTXSIZE)
	local_csr_wr[active_lm_addr_0, lm_ctx_base]		; setup local memory

#ifdef STRIP_4BYTES_CRC
.reg	second_ctx_base

	alu_shf[second_ctx_base, second_lm_base, or, port_num, <<PO2_SECOND_CTXSIZE]
	local_csr_wr[active_lm_addr_1, second_ctx_base]		; setup local memory index 1

#endif


	//	In phase 2 we wait on these I/O signals: rx_event, scratch_put 
	//	and optionally a sig_buf_alloc signal.

	//	Even though the threads receive rbufs in order (because the threads
	//	were put in free list in order) they may not wake up inorder
	//	because of the additional wait for scratch put and sig_buf_alloc. 
	//	So inorder to ensure thread-ordering we also need to wait on next
	//	thread signal.

	alu[sink_sig_mask, --, B, ((1<<RX_EVENT_SIG) | (1 << (&sig_next_thread)))]


	//	check if error in receviing

	alu[--, $prx_rsw1, AND, ALL_ERROR_CODE, <<8]

	bne[rx_error_processing#]						; goto err processing

	//	a proper mpacket was received. Depending on the State of the 
	//	Receive State Machine and the kind of mpacket (SOP, EOP, MOP)
	//	received branch accordingly.

	_packet_rx_get_machine_state(ctx_st, $prx_rsw1)		; ctx_st will have state

	_packet_rx_debug_incr_counter(@pkt_rx_num_proper_mpkts_rxed)

	//	There are only two states: START and PROCessing. There are four
	//	kinds of mpackets: SOP, EOP, MOP and SOP&EOP. MOP is (!SOP and !EOP).
	//	In START state SOP or SOP&EOP mpackets are expected.
	//	In Processing state MOP or EOP mpackets are expected.

	//	ctx_st is interpreted as follows.

	//	000 START 	state, 	MOP mpacket	--	Error case	--	msp
	//	001 PROC 	state, 	MOP mpacket	--	Normalcase	--	mpp
	//	010 START 	state, 	EOP mpacket	--	Error case	--	esp
	//	011 PROC 	state, 	EOP mpacket	--	Normal case	--	epp
	//	100 START 	state, 	SOP mpacket	--	Normal case	--	ssp
	//	101 PROC 	state, 	SOP mpacket	--	Error case	-- 	spp (+ normal case)
	//	110 START 	state, 	SOP&EOP 	--	Normal case	-- 	sesp
	//	111 PROC 	state, 	SOP&EOP 	--	Error case	--	sepp(+ normal case)

	//	jump to the corresponding state processing

	jump[ctx_st, jump_table#], defer[2], \
	targets[msp_j#, mpp_j#, esp_j#, epp_j#, ssp_j#, spp_j#, sesp_j#, sepp_j#]
		_packet_rx_rsw_get_element_num(ele_num, $prx_rsw1)		; rbuf element
		_packet_rx_rsw_get_byte_count(byte_count, $prx_rsw1) 	; bytes received

//	Start of jump table. 

jump_table#:
msp_j#:
	br[msp#]    ; goto mop_start_processing		--	Error case

mpp_j#:
	br[mpp#]	; goto mop_proc_processing		--	Normal case

esp_j#:
	br[esp#]	; goto eop_start_processing		--	Error case

epp_j#:
	br[epp#]	; goto eop_proc_processing		--	Normal case

ssp_j#:
	br[ssp#]	; goto sop_start_processing		-- 	Normal case

spp_j#:
	br[spp#]	; goto sop_proc_processing		--	Error + Normal case

sesp_j#:
	br[sesp#]	; goto sop_eop_start_processing	--	Normal case

sepp_j#:
	br[sepp#]	; goto sop_eop_proc_processing	--	Error + Normal case


//	Error processing code

rx_error_processing#: 

	//	We are here because receive status word indicated an error.

	//	check for NULL (i.e timeout)

	alu[--, $prx_rsw1, AND, 1, <<9]						;[9] indicates NULL receive
	bne[null_error_processing#], defer[1]
		_packet_rx_rsw_get_element_num(ele_num, $prx_rsw1)	; get rbuf element num

	//	Some other error. We need to drop any partially assembled packets.
	//	If we are in START state then we do not have any partially assembled 
	//	packets. lm_state will have the current state. Only two states are 
	//	used. So checking Bit[0] is enough.

	_packet_rx_debug_incr_counter(@pkt_rx_num_err_mpkts_rxed)

	br_bclr[lm_state, 0, nothing_to_drop_in_error#]	; check if START state (0)

	//	Prepare to drop partially assembled packet. For large packets 
	//	drop_eop_handle will be the last buffer in the chain

	alu[drop_sop_handle, --, B, lm_sop_buf]
	alu[drop_eop_handle, --, B, lm_cur_buf]

	//	Reset State Machine

	_packet_rx_set_state(STATE_START)				; reset to start state

	//	End of Phase 1. Wait next thread signal to synchronize

	ctx_arb[sig_next_thread]

	; Signal next thread
	_packet_rx_signal_next_thread(nt_sig_reg)

	//	Drop the mpacket.

	_packet_rx_drop_current_pkt(drop_sop_handle, drop_eop_handle, $drop, sink_sig_mask, drop_sig)

	_packet_rx_incr_counter(port_num, PACKET_PKTS_DROPPED)

	//	Even though we dropped the packet, we'll propagate a NULL
	//	packet thro' the pipeline to keep the pipeline moving.

	//	Continue with Phase2 processing.

	// Free RBUF element.
	_packet_rx_free_rbuf_element(ele_num)

	br[comn_rx_phase_2#]

nothing_to_drop_in_error#:

	//	End of Phase 1. Wait next thread signal to synchronize

	ctx_arb[sig_next_thread]

	; Signal next thread
	_packet_rx_signal_next_thread(nt_sig_reg)

	// Free RBUF element.
	_packet_rx_free_rbuf_element(ele_num)

	br[comn_rx_phase_2#], defer[2]
		alu[dl_next_block, --, B, BID_NULL] 			; NULL (0xFF)
		alu[dl_buf_handle, --, B, DL_BUF_NULL]     		; NULL (0xFF)

null_error_processing#: 

	//	Null mpkt processing. (i.e receive logic timed out for this thread)
	
	//	At this point, the receiving logic has not received a mpkt for certain
	//	amount of time. In order to keep the pipeline moving, the receiving 
	//	logic removes the thread from freelist and wakes up the thread by 
	//	sending the thread a (rx_event) signal.

	//	Nothing to do. Just synchronize with rest of the threads
	//	and go to Phase2.

	ctx_arb[sig_next_thread]

	; Signal next thread
	_packet_rx_signal_next_thread(nt_sig_reg)

	br[comn_rx_phase_2#]

msp#:	
	//	Got MOP mpacket while expecting SOP packet - Error case 

esp#:
	//	Got EOP mpacket while expecting SOP packet - Error case

	//	Both the above error cases are handled here.
	//	Being in START state, we have absolutely nothing to do here.
	//	Just synchronize with rest of the threads and go to Phase2.

	ctx_arb[sig_next_thread]

	; Signal next thread
	_packet_rx_signal_next_thread(nt_sig_reg)

	// Free RBUF element.
	_packet_rx_free_rbuf_element(ele_num)

	br[comn_rx_phase_2#]

mpp#:
	//	Got MOP mpacket while expecting a MOP/EOP mpacket: Normal case

	_packet_rx_mpp_processing()

	br[comn_rx_phase_2#]
				
sepp#:

	//	Got a SOP&EOP (both in same mpacket) while expecting MOP/EOP 
	//	mpacket - Error
	
	//	We do two things. 
	//	[1] We discard the partially (so far) assembled packet.
	//	[2] And since SOP&EOP is a new but complete packet we can process it.
		
	//	[1]
	//	Reset State Machine

	_packet_rx_set_state(STATE_START)				; reset to start state
	 
	//	Prepare to drop the mpacket For large packets 
	//	drop_eop_handle will be the last buffer in the chain
	//	XXX - Why can't drop packet done befor sesp_processing?

	alu[drop_sop_handle, --, B, lm_sop_buf]
	alu[drop_eop_handle, --, B, lm_cur_buf]

	//	[2]	Process SOP&EOP mpacket

	_packet_rx_sesp_processing()

	//	drop the partially assembled pkt

	_packet_rx_drop_current_pkt(drop_sop_handle, drop_eop_handle, $drop, sink_sig_mask, drop_sig)

	_packet_rx_incr_counter(port_num, PACKET_PKTS_DROPPED)

	br[comn_rx_phase_2#]

epp#:
	//	Got EOP mapacket while expecting MOP/EOP - Normal case

	_packet_rx_epp_processing()

	br[comn_rx_phase_2#]

spp#:
	//	Got SOP mpacket while expecting MOP/EOP -  Error case

	//	We do two things. 
	//	[1] We discard the partially (so far) assembled packet.
	//	[2] And since SOP is a new packet we can process it.

	//	Reset State Machine. 

	_packet_rx_set_state(STATE_START)				; reset to start state

	//	Prepare to drop the packet. For large packets 
	//	drop_eop_handle will be the last buffer in the chain
	//	XXX - Why can't drop packet done befor sesp_processing? .

	alu[drop_sop_handle, --, B, lm_sop_buf]
	alu[drop_eop_handle, --, B, lm_cur_buf]

	//	Do the SOP processing for the newly received mpacket

	_packet_rx_ssp_processing()

	//	drop the partially assembled pkt

	_packet_rx_drop_current_pkt(drop_sop_handle, drop_eop_handle, $drop, sink_sig_mask, drop_sig)

	_packet_rx_incr_counter(port_num, PACKET_PKTS_DROPPED)

	br[comn_rx_phase_2#]

ssp#:
	//	Got SOP mpacket while expecting a SOP mpacket -  normal case

	//	Do the SOP processing for the newly received mpacket

	_packet_rx_ssp_processing()

	br[comn_rx_phase_2#]

sesp#:
	//	Got SOP mpacket while expecting a SOP mpacket -  normal case

	//	This is the critical path, where min packets are handled. 

	// Some Ethernet MAC device could not strip 4 bytes CRC from incoming
	// Ethernet frame.  Add CRC strip option here.

#ifdef STRIP_4BYTES_CRC

	alu[byte_count, byte_count, -, 4]

#endif


	_packet_rx_sesp_processing()
	
comn_rx_phase_2#:

	// add current thread to freelist 0

	msf[fast_wr, --, rx_freelist, freelist_reg]		; add thread to freelist

	// 	Wait to receive next m-packet signal and possibly 
	//	prefetch buffer signal.

	ctx_arb[--], defer[1]
	local_csr_wr[active_ctx_wakeup_events, sink_sig_mask]

	.io_completed sig_buf_alloc
	.io_completed drop_sig

.end
#endm


///////////////////////////////////////////////////////////////////////////////
//
// _packet_rx_write_to_nn_ring
//
//	 	Description: Write information worth 5 LW's to the nn ring so that
//      the next me in the rx context pipeline can read it and continue
//      processing.
//
//      The format is
//          LW0 (31:0) : in_dram_handle
//          LW1 (31:0) : in_curr_buf_handle
//          LW2 (31:0) : in_sop_buf_handle
//          LW3 (31:16): in_sop_buf_size
//          LW3 (15:15): in_eop_flag
//          LW3 (14:8) : in_rbuf_elem
//          LW3 (7:0)  : in_byte_count
//          LW4 (31:16): in_inport
//          LW4 (15:0) : in_packet_len
//
//	 	Outputs: None.
//
//		Inputs:
//			in_dram_handle: DRAM address where the m-packet will be written.
//          in_curr_buf_handle: current buffer handle of packet (will be eop
//            buf handle when called by eop macro. (only valid if 
//            in_eop_flag is 1).
//          in_sop_buf_handle: sop buffer handle of packet. (only valid if 
//            in_eop_flag is 1).
//          in_byte_count: number of bytes to transfer from RBUF to DRAM.
//          in_rbuf_elem: RBUF element number containing the m-packet.
//          in_eop_flag: bit indicating if this is the last m-packet for the
//            current packet.
//          in_sop_buf_size: Size in bytes of sop buffer.(only valid if 
//            in_eop_flag is 1).
//          in_packet_len: Size in bytes of entire packet.(only valid if 
//            in_eop_flag is 1).
//          in_inport: Input port number of the packet.(only valid if 
//            in_eop_flag is 1).
//          in_buf_offset: Applicable only for IXP2400 projects.
//            SOP buffer offset in bytes. Constant in IXP28xx projects.
//            (only valid if in_eop_flag is 1).
//			in_cksum: Applicable when Checksum processing is defined. Pass
//					checksum in LSB 16 bits
///////////////////////////////////////////////////////////////////////////////
#macro _packet_rx_write_to_nn_ring(in_dram_handle, in_curr_buf_handle, \
			in_sop_buf_handle, in_byte_count, in_rbuf_elem, in_eop_flag, \
			in_sop_buf_size, in_packet_len, in_inport, in_buf_offset, in_cksum)
.begin

	.reg tmp

writing_to_nn_ring#:
	br_inp_state[NN_FULL, writing_to_nn_ring#]

	alu[*n$index++, --, b, in_dram_handle]

#ifdef	INCLUDE_PKT_CKSUM
	.begin
	.reg	temp1
		#if !IS_IXPTYPE(__IXP28XX)
		alu_shf[temp1, in_buf_offset, OR, in_cksum, <<16]
		alu[*n$index++, --, b, temp1]
		#else
		alu_shf[*n$index++, --, b, in_cksum, <<16]
		#endif	//__IXP28XX
	.end
#else

	#if !IS_IXPTYPE(__IXP28XX)
	alu[*n$index++, --, b, in_buf_offset]
	#endif

#endif	// INCLUDE_PKT_CKSUM

	alu[*n$index++, --, b, in_curr_buf_handle]
	alu[*n$index++, --, b, in_sop_buf_handle]

	alu[tmp, --, b, in_byte_count]
	alu[tmp, tmp, or, in_rbuf_elem, <<8]
	alu[tmp, tmp, or, in_eop_flag, <<15]
	alu[*n$index++, tmp, or, in_sop_buf_size, <<16]

	alu[tmp, --, b, in_packet_len]
	alu[*n$index++, tmp, or, in_inport, <<16]

.end
#endm


///////////////////////////////////////////////////////////////////////////////
//
// _packet_rx_sesp_processing
//
//	 	Description: 
//
//			Process a mpacket containing both SOP&EOP. i.e packets size
//			ranging from 42(min) - 128 are handled here. Min packet case 
//			is the critical path, so it will be optimised wherever possible.
//
//	 	Outputs:
//
//		Inputs:
//
//		Size: 
//
///////////////////////////////////////////////////////////////////////////////

#macro _packet_rx_sesp_processing()

.begin

	.reg	buf_offset
	.reg	header_type
	.reg	dram_handle
	
	//	SOP and EOP are in the same mpacket. This is the critical path
	//	i.e min packets are handled here.
	

	; Set 'prefetch_buf_handle_gpr' to the pre-fetched buffer handle.
	; See macro _packet_rx_buf_alloc() for why we use this instruction.

	#ifndef FREELIST_MANAGER

	alu[prefetch_buf_handle_gpr, prefetch_buf_handle_gpr, xor, \
		$prx_prefetch_buf_handle]

	#endif

	//	check if prefetch buffer handle is zero

	br=byte[prefetch_buf_handle_gpr, 3, 0, sesp_prf_buf_zero_proc#]

	//	We got a valid buffer here.

	//	Get a random offset (headroom) for this buffer. 

	_packet_rx_get_rand_buf_offset(buf_offset)

	//	prefetch_data_handle is where the buffer begins.
	//	+ buf_offset will specify where we should store data 
	//	(including PPP hdr)

	alu[dl_buf_handle, --, B, prefetch_buf_handle_gpr] 	; set dl_buf_handle 
	dl_buf_get_data(prefetch_data_handle, prefetch_buf_handle_gpr)	

	//	Prefetch a new buffer so that it can be used next time around.

	_packet_rx_buf_alloc(BUF_FREE_LIST0, sig_buf_alloc, SIG_NONE, \
							sink_sig_mask)

	//	Find out if the received mpacket can be contained in 1 cell 
	//	(on the transmit side). Remember that for every packet there is
	//	a header ( 8 bytes) that gets prepended. (This is apart from per 
	//	cell header (8 bytes) that gets added) This Per pkt header is added
	//	on the fly. Hence it is not counted in buffer/packet size.

	#define	FCELL_SIZE_WO_PPHDR		(CELL_SIZE - PER_PKT_PREPEND_BYTES)

	alu[--, byte_count, -, FCELL_SIZE_WO_PPHDR]			; > cell size?

	//	Remember if the packet can be contained in one cell, 
	//	then the cell count is 0. If it can be contained in 
	//	two cells, then cell count is 1 etc. 
	//	(less 1 than the actual) 
	ble[continue_with_sesp#], defer[1]		; yes, 1 cell is enough.
		alu[dram_handle, buf_offset, +, prefetch_data_handle]

	//	The packet cannot be contained in one cell. Calculate the number
	//	of cells in the packet. The cell count calculation is generic 
	//	with no hard coded values but is expensive. Watch out if
	//	you run out of budget.

	.reg	cell_count
	.reg	bytes_remaining
	.reg	temp_byte_count

	alu[temp_byte_count, byte_count, +, PER_PKT_PREPEND_BYTES]
	_packet_rx_set_first_cell_count(cell_count, bytes_remaining, temp_byte_count)
	alu[dl_buf_handle, dl_buf_handle, OR, cell_count, <<24]

	br[continue_with_sesp#]

	//	Error Handling.

sesp_prf_buf_zero_proc#:

	_packet_rx_incr_counter(port_num, PACKET_PKTS_DROPPED)

	_packet_rx_debug_incr_counter(@pkt_rx_buf_alloc_fail)

	//	We got a null buffer. So nothing to do. We still need to keep the 
	//	pipeline moving. Get a new prefetch buffer so that it can used next 
	//	time around.

	_packet_rx_buf_alloc(BUF_FREE_LIST0, sig_buf_alloc, SIG_NONE, \
							sink_sig_mask)

	ctx_arb[sig_next_thread]

	; Signal next thread
	_packet_rx_signal_next_thread(nt_sig_reg)

	// Free RBUF element.
	_packet_rx_free_rbuf_element(ele_num)
 
	br[end#]

continue_with_sesp#:

#ifdef	INCLUDE_PKT_CKSUM
	// Write information to nn ring so that the 2nd me in the rx context
	// pipeline can continue processing.
	_packet_rx_write_to_nn_ring(dram_handle, DL_BUF_NULL, dl_buf_handle,
		byte_count, ele_num, 1, byte_count,
		byte_count, port_num, buf_offset, $prx_rsw2)
#else
	// Write information to nn ring so that the 2nd me in the rx context
	// pipeline can continue processing.
	_packet_rx_write_to_nn_ring(dram_handle, DL_BUF_NULL, dl_buf_handle,
		byte_count, ele_num, 1, byte_count,
		byte_count, port_num, buf_offset,0)
#endif
	//	increment packets received and bytes received counter
	_packet_rx_incr_counter(port_num, PACKET_PKTS_RECEIVED)		; pkts rcvd++

	ctx_arb[sig_next_thread]

	; Signal next thread
	_packet_rx_signal_next_thread(nt_sig_reg)

end#:

#undef	FCELL_SIZE_WO_PPHDR

.end
#endm


///////////////////////////////////////////////////////////////////////////////
//
// _packet_rx_ssp_processing
//
//	 	Description: 
//
//			Process a mpacket containing only SOP. Even though we have
//			3 times the budget (3 * 88) as min (42 byte) packets
//			we should be careful with 129 byte packet, where the budget is 
//			reduced to 1.5 * 88 (because we now have to handle two mpackets
//			in 3 * 88 budget. The second mpacket with EOP will be handled in
//			epp_processing). 
//
//	 	Outputs:
//
//		Inputs:
//
//		Size: 
//
///////////////////////////////////////////////////////////////////////////////

#macro _packet_rx_ssp_processing()

.begin

	.reg 	temp_byte_count
	.reg	buf_offset
	.reg	header_type

	; Set 'prefetch_buf_handle_gpr' to the pre-fetched buffer handle.
	; See macro _packet_rx_buf_alloc() for why we use this instruction.

	#ifndef FREELIST_MANAGER

	alu[prefetch_buf_handle_gpr, prefetch_buf_handle_gpr, xor, \
		$prx_prefetch_buf_handle]

	#endif

	dl_buf_get_data(prefetch_data_handle, prefetch_buf_handle_gpr)
	
	//	check if prefetch buffer handle is zero

	alu[--, --, B, prefetch_buf_handle_gpr, <<8]	; [24:0] is buf handle
	beq[ssp_prf_buf_zero_proc#]						; goto error handling
	alu[header_type, lm_header_type, and, 0xFF]		; get header type

	//	This is the first SOP mpacket within the packet

	_packet_rx_get_rand_buf_offset(buf_offset)		; get a random offset

	//	prefetch_data_handle is where the buffer begins.
	//	+ buf_offset will specify where we should store data 
	//	(including PPP hdr)

	alu[lm_cur_data, buf_offset, +, prefetch_data_handle]	

	.reg lm_cur_data_gpr
	alu[lm_cur_data_gpr, --, b, lm_cur_data]

	//	Update various reassembly context data.

	alu[lm_cur_buf, --, B, prefetch_buf_handle_gpr]	; prefetch to current 
	alu[lm_sop_buf, --, B, prefetch_buf_handle_gpr] ; also make it sop buf.
	alu[lm_sop_buf_offset, --, B, buf_offset]		; remember this offset as
	alu[lm_buf_offset, --, B, buf_offset]			; sop buf & cur buf offset 
	alu[lm_buf_size, --, B, byte_count]				; buf size excludes hdr len 
	alu[lm_pkt_size, --, B, byte_count]				; pkt size excludes hdr len


#ifdef STRIP_4BYTES_CRC

	alu_shf[lm_pr_pr_handle, --, B, lm_cur_buf]		; setup pr-previous buffer handle

#endif

#ifdef	INCLUDE_PKT_CKSUM
	alu[lm_pkt_cksum, 0x0, +16, $prx_rsw2]
#endif

	//	Cell Count Calculation.
	//	On the transmit side, for every packet we add a prepend hdr
	//	(apart from a cell hdr for every cell). Include this hdr in
	//	cell count calculation. (For a given packet, we'll hit SOP
	//	condition only once so this place fits well for this 
	//	work)

	alu[temp_byte_count, byte_count, +, PER_PKT_PREPEND_BYTES]
	_packet_rx_set_first_cell_count(lm_cell_count, lm_bytes_remaining, \
									temp_byte_count)

	//	Advance the current data pointer (where next mpacket will
	//	be stored)

	alu[lm_cur_data, lm_cur_data, +, byte_count]	; move the cur data ptr

	//	Prefetch a new buffer so that it can be used next time around.

	_packet_rx_buf_alloc(BUF_FREE_LIST0, sig_buf_alloc, SIG_NONE, \
						sink_sig_mask)

	//	Move the Receive Machine to PROCessing State

	_packet_rx_set_state(STATE_PROC)

	// Write information to nn ring so that the 2nd me in the rx context 
	// pipeline can continue processing.
	_packet_rx_write_to_nn_ring(lm_cur_data_gpr, 0, 0, \
			byte_count, ele_num, 0, 0, \
			0, 0, 0, 0)

	//	End of Phase 1
	ctx_arb[sig_next_thread]


	; Signal next thread
	_packet_rx_signal_next_thread(nt_sig_reg)

	//	Phase 2

	br[ssp_end#]

	//	Null Buffer error condition

ssp_prf_buf_zero_proc#:

	_packet_rx_incr_counter(port_num, PACKET_PKTS_DROPPED)

	_packet_rx_debug_incr_counter(@pkt_rx_buf_alloc_fail)

	//	We got a null buffer. So nothing to do. We still need to keep the 
	//	pipeline moving. Get a new prefetch buffer so that it can used 
	//	next time around.

	_packet_rx_buf_alloc(BUF_FREE_LIST0, sig_buf_alloc, SIG_NONE, \
							sink_sig_mask)

	ctx_arb[sig_next_thread]
	
	; Signal next thread
	_packet_rx_signal_next_thread(nt_sig_reg)

	// Free RBUF element.
	_packet_rx_free_rbuf_element(ele_num)

ssp_end#:

.end
#endm

///////////////////////////////////////////////////////////////////////////////
//
// _packet_rx_epp_processing
//
//	 	Description: 
//
//			Process a mpacket containing only EOP. 
//
//	 	Outputs:
//
//		Inputs:
//
//		Size: 
//
///////////////////////////////////////////////////////////////////////////////

#macro _packet_rx_epp_processing()

.begin

	.reg	cur_buf
	.reg	prev_buf
	.reg 	cell_count
	.reg	buf_offset
	.reg	header_type
	.sig	sig_prev_meta
	.sig	sig_cur_meta

#ifdef STRIP_4BYTES_CRC
	.reg	temp_count
#endif

	#if (defined(DL_SPLIT_META_DATA) && (BUF_SRAM_SPLIT_LW == 1))
	; Flush meta-data belonging to channel 0 and BUF_SRAM_CHANNEL using
	; separate macro calls.
	.sig	sig_prev_meta1 sig_cur_meta1
	#endif

	.reg lm_cur_data_gpr \
		 lm_sop_buf_gpr \
		 lm_pkt_size_gpr \
		 lm_buf_size_gpr \
		 lm_buf_offset_gpr

	alu[header_type, lm_header_type, and, 0xFF]

#ifdef	INCLUDE_PKT_CKSUM
	// first of all, update the checksum.
	_packet_rx_incr_cksum(lm_pkt_cksum, lm_pkt_cksum, $prx_rsw2)

	// set dispatch loop checksum field. We will pass the packet
	// to next stage. A possible optimization is to place this
	// instruction in defer slot.
	dl_meta_set_cksum(lm_pkt_cksum)
#endif

	// Check if strip 4 bytes CRC from incoming packet is defined
	 
#ifdef STRIP_4BYTES_CRC
	.sig sig_pr_pr_handle

	// First, check if current mpakct size is greater than 4
	// Compute the difference between the 4 and byte count of current packet

	alu[temp_count, 4, -, byte_count] 

	bge[mpacket_size_less_than_or_equal_to_4#]	; things become complicated
	
	// Here, the mpacket size is greater than 4, Adjust byte_count and
	// other processing is the same as without CRC stripping
	
	alu[byte_count, byte_count, -, 4]

	br[normal_epp_processing#]
	 	
mpacket_size_less_than_or_equal_to_4#:

	// No data should be written to the DRAM from this m-packet
	// So set byte count as 0 

	alu[byte_count, --, b, 0]

	// Check if this is the first mpacket in current data buffer
	// If this is not the first mpacket in curent buffer, Things are not too complicated.
	// Just need to adjust cell count and set byte_count as zero.

	alu[--, lm_buf_size, -, 0]					

	beq[this_is_first_mpacket_in_this_buffer#]	; things become more complicated

	// Here, this in not first m-packet in current buffer.
	// Adjust current buffer cell count

	alu[--, lm_bytes_remaining, -, temp_count]
	bge[no_cell_count_change#]
	
	// Cell count change is needed because remainder is less than the difference
	alu[lm_cell_count, lm_cell_count, -, 1]

no_cell_count_change#:

	// Set byte count is zero so the second ME will not move the data from current
	// RBUF element to DRAM 

	br[normal_epp_processing_1#], defer[2]
		alu[lm_buf_size, lm_buf_size, -, temp_count]	; update buffer size
		alu[lm_pkt_size, lm_pkt_size, -, temp_count]	; update packet size


this_is_first_mpacket_in_this_buffer#:

	// This is the first mpacket in current buffer and mpacket size is less 
	// than or equal to 4. In this case, some CRC bytes (1-3) are in the previous buffer.
	// These CRC bytes need to be removed from previous buffer.  Current buffer needs 
	// to be freed instead of in the buffer train.

	// Get SOP handel

	alu[lm_sop_buf_gpr, --, b, lm_sop_buf]

	// Adjust previous buffer size 

	alu[lm_prev_buf_size, lm_prev_buf_size, -, temp_count]

	// Adjust packet size
	alu[lm_pkt_size, lm_pkt_size, -, temp_count]

	alu[prev_buf, --, B, lm_prev_buf] 

	br_bclr[prev_buf, 30, crc_strip_epp_prev_buffer_not_sop#], defer[3]	; [30] is SOP
		dl_meta_set_buffer_size(lm_prev_buf_size)  	; set buffer size
		dl_meta_set_offset(NON_SOP_OFFSET)			; saves an extra branch
		_packet_rx_set_state(STATE_START)			; Reset State Machine

crc_strip_epp_prev_buffer_sop#:
	
	// For IXP2800 sillicon implementation, the BUF offset is fixed 384 so no
	// SOP cell count update is needed.  If BUF offset is changed, cell count
	// update may be needed

	// If prev buffer is sop buffer, CELL_SIZE is 120, offset is 128 and temp_count
	// is greater than 0 then the SOP cell count need to be updated. Other CELL size
	// will be worked out when it is avalable. Only if previous buffer is SOP buffer 
	// this cell count adjustment is necessary

	
#ifndef IXP2800	 
 
#if (CELL_SIZE == 120)

#if (BUFFER_SIZE == 2048)	
	// Check if sop offste is 128

	alu[--, lm_sop_buf_offset, -, 128]
	bne[no_sop_cell_count_adjust#]
	
	// check if tem_count is greater than 1
	
	alu[--, temp_count, -, 0]
	beq[no_sop_cell_count_adjust#]
	
	// Adjust the cell count and write to the SOP buffer handle

.set cell_count

	alu[cell_count, cell_count, -, 1]	

	alu[lm_sop_buf_gpr, lm_sop_buf_gpr, and~, 0x3F, <<24]
	alu[lm_sop_buf_gpr, lm_sop_buf_gpr, OR, cell_count, <<24] 	; embed cell count

#else

#Redo the cell count update for other buffer size

#endif

#else

#Redo the cell count update code for other cell size

#endif

#endif


no_sop_cell_count_adjust#:

	//	Previous buffer is a SOP buffer. Buffer offset should point to where 
	//	incoming packet begins.

	
	// Previous buffer is SOP and also EOP, set EOP bit

	alu[lm_sop_buf_gpr, lm_sop_buf_gpr, or, 1, <<31]

	alu[lm_sop_buf_size, lm_sop_buf_size, -, temp_count]

	alu[buf_offset, --, B, lm_sop_buf_offset]			; get sop buffer offset

	dl_meta_set_offset(buf_offset)						; set offset


crc_strip_epp_prev_buffer_not_sop#:

	dl_meta_set_header_type(header_type)				; set header type
	dl_meta_set_buffer_next(IX_NULL)  				; this is the last buffer

	// Now previous buffer becomes EOP buf, the EOP bit needs to be set

	alu[prev_buf, prev_buf, or, 1, <<31]

	alu[lm_pkt_size_gpr, --, b, lm_pkt_size]

	xbuf_alloc($prev_meta1_x, 2, write) 				; allocate xfer regs

	//	Write META data of previous buffer to SRAM

	dl_meta_flush_cache($prev_meta1_x, prev_buf, sig_prev_meta, SIG_NONE, 0, 2)

	// set previous buffer handle

	dl_meta_set_buffer_next(prev_buf)

	xbuf_alloc($pr_pr_x, 1, write) 						; allocate  1 xfer reg

	// Update the pr-previuos buffer next buffer handle using previous buffer handle (EOP bit set)

	dl_meta_flush_cache($pr_pr_x, lm_pr_pr_handle, sig_pr_pr_handle, SIG_NONE, 0, 1)

	// Current buffer needs to be freed

	alu[cur_buf, --, b, lm_cur_buf]						; get current buffer handle

	dl_drop_buffer(cur_buf)								; drop current buffer

	// Write information to nn ring so that the 2nd me in the rx context 
	// pipeline can continue processing.

#ifdef	INCLUDE_PKT_CKSUM
	_packet_rx_write_to_nn_ring(0, prev_buf, lm_sop_buf_gpr, \
			byte_count, ele_num, 1, lm_sop_buf_size, \
			lm_pkt_size_gpr, port_num, lm_buf_offset_gpr,lm_pkt_cksum)

#else
	_packet_rx_write_to_nn_ring(0, prev_buf, lm_sop_buf_gpr, \
			byte_count, ele_num, 1, lm_sop_buf_size, \
			lm_pkt_size_gpr, port_num, lm_buf_offset_gpr,0)
#endif

	#if (defined(DL_SPLIT_META_DATA) && (BUF_SRAM_SPLIT_LW == 1))

	; Flush meta-data belonging to channel 0 and BUF_SRAM_CHANNEL using
	; separate macro calls.

	ctx_arb[sig_next_thread, sig_prev_meta, sig_prev_meta1, sig_pr_pr_handle]

	#else

	ctx_arb[sig_next_thread, sig_prev_meta, sig_pr_pr_handle]

	#endif

	xbuf_free($prev_meta1_x)  						; free $prev_meta1
	
	xbuf_free($pr_pr_x)  							; free $pr_pr_x
	
	br[end#]

#endif //STRIP_4BYTES_CRC	
	

normal_epp_processing#:

	//	Got EOP mpacket while expecting MOP/EOP - Normal case
	//	Receive State Machine is in PROCessing State.

	//	increment cell count depending on the number of bytes received.
	//	embed this cell count in the buffer handle.

	_packet_rx_incr_cell_count(lm_cell_count, lm_bytes_remaining, byte_count)
		
normal_epp_processing_1#:

	alu[cell_count, --, B, lm_cell_count]
	alu[cur_buf, --, B, lm_cur_buf] 				; get curr buffer handle
	alu[cur_buf, cur_buf, OR, cell_count, <<24] 	; embed cell count
													; by default EOP is set.
	//	update buffer size, packet size and reset
	//	state machine to START state.
	//	[See defer slots]

	//	Check if current buffer is a sop buffer

	br_bclr[cur_buf, 30, epp_not_sop_buffer#], defer[3]	; [30] is SOP bit.
		alu[lm_buf_size, lm_buf_size, +, byte_count]	; update buffer size
		alu[lm_pkt_size, lm_pkt_size, +, byte_count]	; update packet size
		_packet_rx_set_state(STATE_START)				; Reset State Machine

epp_sop_buffer#:

	_packet_rx_incr_counter(port_num, PACKET_PKTS_RECEIVED)	; pkts rcvd++

	//	Yes, this buffer has SOP. i.e this buffer now contains
	//	a complete packet (SOP ...EOP). So we need to handle only
	//	a single buffer. Set all META data fields and write the 
	//	meta data to SRAM.
		
	//	buffer offset is where data begins in the buffer.
	// 	For subsequent blocks buffer offset should point to 
	//	where IP header (or rather whatever follows PPP header)
	//	begins.
	alu[buf_offset, --, B, lm_sop_buf_offset]		; get buffer offset 
	dl_meta_set_offset(buf_offset)
	dl_meta_set_buffer_next(IX_NULL)               	; there is no next buffer
	dl_meta_set_buffer_size(lm_buf_size)  			; set buffer size
	dl_meta_set_packet_size(lm_pkt_size)			; set packet size
	dl_meta_set_input_port(port_num)        		; set the input port
	dl_meta_set_header_type(header_type)			; set header type
	dl_meta_set_free_list(FREE_LIST_ID)				; set free list. 
	dl_meta_set_rx_stat(0)							; set rx stat. Zero for now

	//	Write the meta data to SRAM.

	xbuf_alloc($cur_meta, 4, write) 				; alloc xfer regs for flush

	dl_meta_flush_cache($cur_meta, cur_buf, sig_cur_meta, SIG_NONE, 0, 4)

	// Copy registers that will be used during nn transfer into gpr's.
	alu[lm_cur_data_gpr, --, b, lm_cur_data]
	alu[lm_buf_size_gpr, --, b, lm_buf_size]
	alu[lm_pkt_size_gpr, --, b, lm_pkt_size]
	alu[lm_buf_offset_gpr, --, b, lm_sop_buf_offset]

	//	End of Phase 1.
	//	Wait on the all the previous I/O plus the next thread signal.

	// Write information to nn ring so that the 2nd me in the rx context 
	// pipeline can continue processing.
#ifdef	INCLUDE_PKT_CKSUM
	_packet_rx_write_to_nn_ring(lm_cur_data_gpr, DL_BUF_NULL, cur_buf, \
			byte_count, ele_num, 1, lm_buf_size_gpr, \
			lm_pkt_size_gpr, port_num, lm_buf_offset_gpr,lm_pkt_cksum)
#else
	_packet_rx_write_to_nn_ring(lm_cur_data_gpr, DL_BUF_NULL, cur_buf, \
			byte_count, ele_num, 1, lm_buf_size_gpr, \
			lm_pkt_size_gpr, port_num, lm_buf_offset_gpr,0)
#endif

	#if (defined(DL_SPLIT_META_DATA) && (BUF_SRAM_SPLIT_LW == 1))
	; Flush meta-data belonging to channel 0 and BUF_SRAM_CHANNEL using
	; separate macro calls.
	ctx_arb[sig_next_thread, sig_cur_meta, sig_cur_meta1]
	#else
	ctx_arb[sig_next_thread, sig_cur_meta]
	#endif

	xbuf_free($cur_meta) 							; free sram tranfer regs

	br[end#]
	

epp_not_sop_buffer#:

	//	At this point, EOP and SOP are in different buffers i.e large packet.
	//  So we need to write previous buffer's meta data & current buffer's 
	//	(EOP) meta data to SRAM.
	//	There are two cases to handle.

	//	[1] Previous Buffer is also SOP Buffer.
	//	[2] Previous Buffer is not SOP Buffer.

	//  [Recap] cur_buf contains current buffer handle. cell count
	//	had already been updated in the buffer handle.

	//	compose meta data for the previous buffer. (defer slots)

	//	Check if the previous buffer is a SOP buffer. If it is not a 
	//	SOP buffer, the offset is set as NON_SOP_OFFSET(0) 
	//	(in the defer slot) i.e no headroom. If it is a SOP buffer 
	//	we have stored the offset in lm. Use that.

	xbuf_alloc($prev_meta1, 4, write) 				; allocate xfer regs

	alu[prev_buf, --, B, lm_prev_buf] 

	br_bclr[prev_buf, 30, epp_prev_buffer_not_sop#], defer[3]	; [30] is SOP
		dl_meta_set_buffer_next(cur_buf)    		; cur buf is next for prev
		dl_meta_set_buffer_size(lm_prev_buf_size)  	; set buffer size
		dl_meta_set_offset(NON_SOP_OFFSET)			; saves an extra branch

epp_prev_buffer_sop#:

	//	Previous buffer is a SOP buffer. Buffer offset should point to where 
	//	IP header (or rather whatever follows PPP header) begins.
	alu[buf_offset, --, B, lm_sop_buf_offset]		; strip header len
	dl_meta_set_offset(buf_offset)					; set offset
	dl_meta_set_header_type(header_type)			; set header type
	dl_meta_set_rx_stat(0)							; set rx stat. Zero for now
	dl_meta_set_free_list(FREE_LIST_ID)				; set free list. 
	dl_meta_set_packet_size(lm_pkt_size)			; set packet size
	dl_meta_set_input_port(port_num)        		; set the input port

	//	Write META data of previous buffer to SRAM
	dl_meta_flush_cache($prev_meta1, prev_buf, sig_prev_meta, SIG_NONE, 0, 4)

	br[epp_prev_buffer_continue#]

epp_prev_buffer_not_sop#:

	//	Write META data of previous buffer to SRAM
	dl_meta_flush_cache($prev_meta1, prev_buf, sig_prev_meta, SIG_NONE, 0, 2)

epp_prev_buffer_continue#:

	xbuf_alloc($cur_meta_1, 2, write) 				; allocate xfer regs

	//	Set current buffer meta data

	dl_meta_set_buffer_next(IX_NULL)  				; this is the last buffer
	dl_meta_set_offset(NON_SOP_OFFSET) 				; this buffer !SOP buffer
	dl_meta_set_buffer_size(lm_buf_size)    		; set buffer size
	
	//	Write META data of current buffer to SRAM
	dl_meta_flush_cache($cur_meta_1, cur_buf, sig_cur_meta, SIG_NONE, 0, 2)

	//	End of Phase 1.
	//	Wait on the all the previous I/O plus the next thread signal.

	// Copy registers that will be used during nn transfer into gpr's.
	alu[lm_cur_data_gpr, --, b, lm_cur_data]
	alu[lm_sop_buf_gpr, --, b, lm_sop_buf]
	alu[lm_pkt_size_gpr, --, b, lm_pkt_size]
	alu[lm_buf_offset_gpr, --, b, lm_sop_buf_offset]

	// Write information to nn ring so that the 2nd me in the rx context 
	// pipeline can continue processing.
#ifdef	INCLUDE_PKT_CKSUM
	_packet_rx_write_to_nn_ring(lm_cur_data_gpr, cur_buf, lm_sop_buf_gpr, \
			byte_count, ele_num, 1, lm_sop_buf_size, \
			lm_pkt_size_gpr, port_num, lm_buf_offset_gpr,lm_pkt_cksum)
#else
	_packet_rx_write_to_nn_ring(lm_cur_data_gpr, cur_buf, lm_sop_buf_gpr, \
			byte_count, ele_num, 1, lm_sop_buf_size, \
			lm_pkt_size_gpr, port_num, lm_buf_offset_gpr,0)
#endif
	#if (defined(DL_SPLIT_META_DATA) && (BUF_SRAM_SPLIT_LW == 1))
	; Flush meta-data belonging to channel 0 and BUF_SRAM_CHANNEL using
	; separate macro calls.
	ctx_arb[sig_next_thread, sig_prev_meta, sig_cur_meta, \
			sig_prev_meta1, sig_cur_meta1]
	#else
	ctx_arb[sig_next_thread, sig_prev_meta, sig_cur_meta]
	#endif

	xbuf_free($prev_meta1)  						; free $prev_meta1
	xbuf_free($cur_meta_1)							; free $cur_meta_1

end#:
	
	; Signal next thread
	_packet_rx_signal_next_thread(nt_sig_reg)

.end
#endm


///////////////////////////////////////////////////////////////////////////////
//
// _packet_rx_mpp_processing
//
//	 	Description: 
//
//			Process a mpacket containing neither SOP nor EOP. i.e this is a 
//			MOP (middle of packet).
//
//	 	Outputs:
//
//		Inputs:
//
//		Size: 
//
///////////////////////////////////////////////////////////////////////////////

#macro _packet_rx_mpp_processing()

.begin

	.reg 	buf_size
	.reg	cur_buf
	.reg	cell_count
	.reg 	buf_offset
	.reg	cur_data
	.reg 	temp
	.reg 	header_type
	.reg	buf_hndl
	.sig	sig_meta

	#if (defined(DL_SPLIT_META_DATA) && (BUF_SRAM_SPLIT_LW == 1))
	; Flush meta-data belonging to channel 0 and BUF_SRAM_CHANNEL using
	; separate macro calls.
	.sig	sig_meta1
	#endif

	//	Expect a MOP and get a MOP: Normal case

#ifdef	INCLUDE_PKT_CKSUM
	// first of all, update the checksum.
	_packet_rx_incr_cksum(lm_pkt_cksum, lm_pkt_cksum, $prx_rsw2)
#endif

	//	Increment buffer size by number of bytes received. Buffer size
	//	is the amount of the data currently present in the buffer. We are
	//	guaranteed to have room in the buffer for this mpacket.
	//	[ See the defer slots below]

	//	Find out if buffer is full (for next mpacket). 
	//	Buffer size + headroom in the buffer will show 
	//	if the buffer is full.
	
	alu[buf_size, byte_count, +, lm_buf_size]		; buf size after THIS mpkt
	alu[buf_size, buf_size, +, lm_buf_offset] 		; + headroom
	
	alu[--, buf_size, -, max_buf_size]				; is buffer full?
													; (for next time)
	blt[mpp_buffer_not_full#], defer[2] 				; no, goto buffer not full
		alu[lm_buf_size, byte_count, +, lm_buf_size]	; incr buffer size
		alu[lm_pkt_size, byte_count, +, lm_pkt_size]	; incr packet size

	//	At this point buffer will be full after storing the current
	//	mpacket. See if a new buffer (prefetch buffer) is available .

	; Set 'prefetch_buf_handle_gpr' to the pre-fetched buffer handle.
	; See macro _packet_rx_buf_alloc() for why we use this instruction.

	#ifndef FREELIST_MANAGER

	alu[prefetch_buf_handle_gpr, prefetch_buf_handle_gpr, xor, \
		$prx_prefetch_buf_handle]

	#endif

	dl_buf_get_data(prefetch_data_handle, prefetch_buf_handle_gpr)

	alu[--, --, B, prefetch_buf_handle_gpr, <<8]	; [23:0] is buf handle
	beq[mpp_null_buffer#]							; null buffer error.

	//	New buffer is available. Normal processing.

	// Write information to nn ring so that the 2nd me in the rx context 
	// pipeline can continue processing.
	_packet_rx_write_to_nn_ring(lm_cur_data, 0, 0, \
			byte_count, ele_num, 0, 0, \
			0, 0, 0, 0)

	//	check if current buffer is a SOP buffer. (i.e a buffer containing
	//	start of packet.)

	alu[cur_buf, --, B, lm_cur_buf] 				; lm cannot be used
	br_bclr[cur_buf, 30, mpp_buffer_full_not_sop#], defer[1]	; [30] is SOP
		alu_shf[cur_buf, cur_buf, AND~, 0x1, <<31]	; anyway, clear EOP [31]
													; because this buffer 
													; will not contain EOP
mpp_buffer_full_sop#:

	//	[Recap] This buffer is full. This buffer contains SOP. 
	//	A new buffer is available.

	//	increment cell count depending on the number of bytes received.
	//	embed this cell count in the buffer handle.

	_packet_rx_incr_cell_count(lm_cell_count, lm_bytes_remaining, byte_count)

	alu[cell_count, --, B, lm_cell_count]			; lm can't be used with <<
	alu[cur_buf, cur_buf, OR, cell_count, <<24]		; [29:24] cell count

	//	Keep the SOP buffer offset, buffer size and move current
	//	buffer handle to the previous buffer handle in the context data
	//	cur_buf contains the current buffer handle

	//	We are done with current buffer. Make it previous buffer as well as
	//	SOP Buffer

	alu[lm_sop_buf, --, B, cur_buf]					; cur buf is SOP buf
	alu[lm_prev_buf, --, B, cur_buf]				; cur buf to prev buf
	alu[lm_prev_buf_size, --, B, lm_buf_size]		; buf size to prev

	//	Since this buffer is also SOP buffer, remember
	//	a few more fields.

	alu[lm_sop_buf_size, --, B, lm_buf_size]		; buf size to sop buf size

	//	Ok, now make the prefetch buffer as current buffer. Any new
	//	mpacket received will be stored in this new buffer.

	alu[temp, prefetch_buf_handle_gpr, AND~, 0x1, <<30]	; clear sop bit
	alu[lm_cur_buf, --, B, temp]					
	alu[lm_cur_data, --, B, prefetch_data_handle]	

	// Write prefetched (i.e. now the current) buffer handle
	// in the next buffer field of the meta-data for the sop buffer.
	// This write is necessary to maintain the buffer chain if the packet
	// is dropped.

	.reg sop_buf
	xbuf_alloc($x_1, 1, write)	 					; allocate xfer registers

	alu[sop_buf, --, b, lm_sop_buf]
	dl_meta_set_buffer_next(lm_cur_buf) 				

	dl_meta_flush_cache($x_1, sop_buf, sig_meta, SIG_NONE, 0, 1)

	//	clear cell count and byte reminder.

	alu[lm_cell_count, --, B, 0x0] 					; clear cell count
	alu[lm_bytes_remaining, --, B, 0x0] 			; clear bytes remaining

	//	Prefetch a new buffer so that it can be used next time around.

	_packet_rx_buf_alloc(BUF_FREE_LIST0, sig_buf_alloc, SIG_NONE, sink_sig_mask)

	//	End of Phase 1.
	//	Wait on the all the previous I/O plus the next thread signal.

	ctx_arb[sig_next_thread, sig_meta], defer[2]
		alu[lm_buf_size, --, B, 0] 					; set buffer size as 0
		alu[lm_buf_offset, --, B, NON_SOP_OFFSET]	; set offset as 0

	xbuf_free($x_1)
	
	; Signal next thread
	_packet_rx_signal_next_thread(nt_sig_reg)

	// 	go to phase 2 processing

	//	Remember that we have received only a part of the packet so far.

	br[end#]
									
mpp_buffer_full_not_sop#:
	
	//	[Recap]: Current buffer is full. But it does not contain SOP.
	//	cur_buf contains the current buffer handle.

	//	Two slightly different cases to handle here. 
	//	[1] Previous Buffer is SOP. (this will have headroom)
	//	[2] Previous Buffer is not SOP. (this will not)

	//	increment cell count depending on the number of bytes received.
	//	embed this cell count in the buffer handle.
	//	XXX - Why can't this cell count calc. be done in the common path 
	//	up above.

	_packet_rx_incr_cell_count(lm_cell_count, lm_bytes_remaining, byte_count)

	alu[cell_count, --, B, lm_cell_count]			; lm can't be used with <<
	alu[cur_buf, cur_buf, OR, cell_count, <<24]		; [29:24] are cell count

	//	Now we are ready to flush META data of previous buffer to SRAM.
	//	Remember the META data contains a hw_next pointer which, in this case, 
	//	should point to current buffer. [Various META data fields are set in
	//	the defer slots below]

	alu[dl_buf_handle, --, B, lm_prev_buf]			; prepare to set meta data

#ifdef STRIP_4BYTES_CRC

	alu[lm_pr_pr_handle, --, B, lm_prev_buf]		; previous handl to pr-previous handle

#endif

	//	Check if the previous buffer is a SOP buffer
	//	If it is not a SOP buffer, the offset is set 
	//	as NON_SOP_OFFSET(0) (in the defer slot) i.e no headroom
	//	If it is a SOP buffer we have stored the offset in
	//	lm. Use that.

	br_bclr[dl_buf_handle, 30, mpp_prev_buffer_not_sop#], defer[3]
		dl_meta_set_buffer_next(cur_buf)    		; cur buf is next for prev
		dl_meta_set_buffer_size(lm_prev_buf_size)  	; set previous buffer size
		dl_meta_set_offset(NON_SOP_OFFSET)			; saves an extra branch

	//	Here, previous buffer is a SOP buffer
	//	set header type as stored in LM.
	alu[header_type, lm_header_type, and, 0xFF]		; get header type
	alu[buf_offset, --, B, lm_sop_buf_offset]		; Get SOP buffer offset
	dl_meta_set_offset(buf_offset)					; set SOP buffer offset
	dl_meta_set_header_type(header_type)			; set header type

mpp_prev_buffer_not_sop#:

	//	All the required fields of META data have been set
	//	for the previous buffer. Write it to SRAM. The META
	//	Data for !SOP buffers is only a subset of SOP META
	//	and contains only the first 2 longwords of META.

	xbuf_alloc($x_0, 2, write) 						; allocate xfer registers

	dl_meta_flush_cache($x_0, dl_buf_handle, sig_meta, SIG_NONE, 0, 2)

	//	Ok, prepare for our next trip down this lane.

	//	We are done with current buffer. Make it previous buffer 

	alu_shf[lm_prev_buf, --, B, cur_buf]
	alu[lm_prev_buf_size, --, B, lm_buf_size]
	
	//	Ok, now make the prefetch buffer as current buffer. Any new
	//	mpacket received will be stored in this new buffer.

	alu[temp, prefetch_buf_handle_gpr, AND~, 0x1, <<30]	; clear sop bit
	alu[lm_cur_buf, --, B, temp]						
	alu[lm_cur_data, --, B, prefetch_data_handle]		

	// Write prefetched (i.e. now the current) buffer handle
	// in the next buffer field of the meta-data for the previous buffer.
	// This write is necessary to maintain the buffer chain if the packet
	// is dropped.

	.reg prev_buf
	xbuf_alloc($x_2, 1, write)	 					; allocate xfer registers
	.sig sig_meta_prev

	alu[prev_buf, --, b, lm_prev_buf]
	dl_meta_set_buffer_next(lm_cur_buf) 				

	dl_meta_flush_cache($x_2, prev_buf, sig_meta_prev, SIG_NONE, 0, 1)

	//	clear cell count and byte reminder.

	alu[lm_cell_count, --, B, 0x0]  				; clear cell count
	alu[lm_bytes_remaining, --, B, 0x0] 			; clear bytes remaining

	//	Prefetch a new buffer so that it can be used next time around.

	_packet_rx_buf_alloc(BUF_FREE_LIST0, sig_buf_alloc, SIG_NONE, \
							sink_sig_mask)

	//	End of Phase 1.
	//	Wait on the all the previous I/O plus the next thread signal.

	#if (defined(DL_SPLIT_META_DATA) && (BUF_SRAM_SPLIT_LW == 1))

	; Flush meta-data belonging to channel 0 and BUF_SRAM_CHANNEL using
	; separate macro calls.

	ctx_arb[sig_next_thread, sig_meta, sig_meta1, sig_meta_prev], defer[2]
		alu[lm_buf_size, --, B, 0] 					; set buffer size as 0
		alu[lm_buf_offset, --, B, NON_SOP_OFFSET]	; set offset as 0

	#else

	ctx_arb[sig_next_thread, sig_meta, sig_meta_prev], defer[2]
		alu[lm_buf_size, --, B, 0] 					; set buffer size as 0
		alu[lm_buf_offset, --, B, NON_SOP_OFFSET]	; set offset as 0

	#endif
	
	; Signal next thread
	_packet_rx_signal_next_thread(nt_sig_reg)

	xbuf_free($x_0) 								; free sram trnsfer regs 
	xbuf_free($x_2)

	// 	go to phase 2 processing

	//	Remember that we have received only a part of the packet so far.

	br[end#]

mpp_buffer_not_full#:

	//	MOP received and Current buffer is not FULL. (i.e it can
	//	take current mpacket + atleast one more mpacket)

	// Write information to nn ring so that the 2nd me in the rx context 
	// pipeline can continue processing.
	_packet_rx_write_to_nn_ring(lm_cur_data, 0, 0, \
			byte_count, ele_num, 0, 0, \
			0, 0, 0, 0)

	//	increment cell count depending on the number of bytes received.

	_packet_rx_incr_cell_count(lm_cell_count, lm_bytes_remaining, byte_count)

	alu[lm_cur_data, byte_count, +, lm_cur_data]

	//	End of Phase 1.
	//	Wait on the all the previous I/O plus the next thread signal.
	//	Remember that we have received only a part of the packet so far.

	ctx_arb[sig_next_thread]
	
	; Signal next thread
	_packet_rx_signal_next_thread(nt_sig_reg)

	br[end#]

mpp_null_buffer#:

	//	NULL Buffer error handling. 

	_packet_rx_incr_counter(port_num, PACKET_PKTS_DROPPED)
	
	_packet_rx_debug_incr_counter(@pkt_rx_buf_alloc_fail)

	//	If buffer not available drop the current (so far assembled) 
	//	packet as well as the just received mpacket. Reset State 
	//	Machine (to	START State)

	_packet_rx_set_state(STATE_START)				; reset to start state.
	
	//	Prefetch a new buffer so that it can be used next time around.

	_packet_rx_buf_alloc(BUF_FREE_LIST0, sig_buf_alloc, SIG_NONE, \
							sink_sig_mask)

	ctx_arb[sig_next_thread], defer[2]
		alu[drop_sop_handle, --, B, lm_sop_buf]	
		alu[drop_eop_handle, --, B, lm_cur_buf]
	
	; Signal next thread
	_packet_rx_signal_next_thread(nt_sig_reg)

	// Free RBUF element.
	_packet_rx_free_rbuf_element(ele_num)

	//	drop current (so far assembled) packet

	_packet_rx_drop_current_pkt(drop_sop_handle, drop_eop_handle, $drop, sink_sig_mask, drop_sig)

end#:

.end
#endm

#endif // FIRST_PACKET_RX_ME
//--------------------------------
//	First Packet RX ME - End
//--------------------------------


//--------------------------------
//	Second Packet RX ME - Begin
//--------------------------------
#ifdef SECOND_PACKET_RX_ME

//	Global Signals

.sig volatile sig_next_thread			// next thread signal

// Global Registers

.reg	prx_rbuf_base						// base address of rbuf
.reg	nt_sig_reg						// used in signalling next thread
.reg 	dram_rbuf_ov					// for moving data from rbuf to dram

#ifdef	PACKET_RX_COUNTERS
.reg	counter_base					// base addr for counters.
#endif

#ifdef _DEBUG_COUNTERS_
.reg @pkt_rx_num_pkts					; number of pkts send to next stage
.reg @pkt_rx_drop						; number of pkts dropped.
.reg @pkt_rx_exception					; number of exception pkts.

immed[@pkt_rx_num_pkts, 0]
immed[@pkt_rx_drop, 0]
immed[@pkt_rx_exception, 0]

#endif // _DEBUG_COUNTERS_

///////////////////////////////////////////////////////////////////////////////
//
// packet_rx_2_init
//
//	 	Description: 
//
//			One time initialisation code for the second me of Packet RX.
//
//	 	Outputs:
//
//			Global variables will be affected.
//
//		Inputs:
//
//			Will use global variables.
//
///////////////////////////////////////////////////////////////////////////////
#macro packet_rx_2_init()
.begin

	move(dram_rbuf_ov, ((1 << 25)|(1<<4)))	   ; [25] ref_cnt, [4] xfer regs overwrite

	move(prx_rbuf_base, RBUF_TBUF_BASE)            ; set up RBUF base

	; Get current me and thread number.
	.reg me_num thread_num
	_packet_rx_get_current_me_thread_number[me_num, thread_num]

	; Based on RX_PHY_MODE, setup thread order.
	_packet_rx_setup_next_thread_sig[nt_sig_reg, thread_num]

	#ifdef	PACKET_RX_COUNTERS
	; Initialize counters base address.
	immed32(counter_base, PACKET_COUNTERS_SRAM_BASE)
	#endif

	; Init meta-data gpr's. Some of these will be used to write to the 
	; scratch ring.

	.reg zero
	immed[zero, 0]
	dl_meta_init_cache(zero, zero, zero, zero, zero, zero, zero, zero)

#ifdef	INCLUDE_PKT_CKSUM
	dl_meta_set_cksum(0x0)
#endif
	; For next iteration, set meta-data fields that remain constant
	; and will be sent over the scratch ring.

	#if IS_IXPTYPE(__IXP28XX)
	; Set buf_offset. Constant in this design (set to 384).
	.reg buf_offset
	_packet_rx_get_rand_buf_offset(buf_offset)
	dl_meta_set_offset(buf_offset)
	#endif

	; Set freelist id.
	dl_meta_set_free_list(FREE_LIST_ID)
	; Set rx stat to 0 for now.
	dl_meta_set_rx_stat(0)
	; Set header type to packet type coming in on port 0 for now.
	; Will need to change if different ports handle different packet types.
	; For ex: If port 0 handles PPP and port 1 handles Ethernet.
	dl_meta_set_header_type(PORT_0_TYPE)	

	_dl_pos_sink_init()

	.if(ctx()==0)

		; Init nn ring.
		.reg ctx_enable
		local_csr_rd[ctx_enables]
		immed[ctx_enable, 0]
		alu_shf[ctx_enable, ctx_enable, AND~, 3, <<18]
		alu_shf[ctx_enable, ctx_enable, OR, 3, <<18]
		local_csr_wr[ctx_enables, ctx_enable]
		local_csr_wr[nn_get, 0]
		local_csr_wr[nn_put, 0]
		alu[--, --, b, 0]
		alu[--, --, b, 0]
		alu[--, --, b, 0]

		;Ensure 8 cycle delay between a ctx_enable fast wr and a ctx arb
		alu[--, --, b, 0]
		alu[--, --, b, 0]
		alu[--, --, b, 0]
		alu[--, --, b, 0]
		alu[--, --, b, 0]

		; Wait for system initialization complete signal.
		.sig volatile system_init_sig
		.addr system_init_sig ME_INIT_SIGNAL
		ctx_arb[system_init_sig]

	.else
		ctx_arb[sig_next_thread]

	.endif

.end
#endm

///////////////////////////////////////////////////////////////////////////////
//
// packet_pkt_rx_2
//
//	 	Description: 
//
// 			Packet processing entry point for second me of packet rx.
//			In each iteration one mpacket will be handled.
//
//	 	Outputs:
//
//			Global variables will be affected.
//
//		Inputs:
//
//			Will use global variables.
//
///////////////////////////////////////////////////////////////////////////////
#macro packet_pkt_rx_2()
.begin

	.reg dram_handle \
		 byte_count \
		 ele_num \
		 ele_addr \
		 sop_buf_size \
		 packet_size \
		 input_port

	.sig sig_dram

	; Read from the nn ring. The first me of the context pipeline
	; puts data on the ring. For the format refer to macro description
	; _packet_rx_write_to_nn_ring()
	br_inp_state[nn_empty, null_request#]

	; Copy data from nn ring into gpr's.
	alu[dram_handle, --, b, *n$index++]

#ifdef	INCLUDE_PKT_CKSUM
	; extract cksum field
	.begin
	.reg temp1 cksum
		alu[temp1, -- , B, *n$index++]
		; cksum is in upper 16 bits
		ld_field_w_clr[cksum, 0011, temp1, >>16]
		
		; set cksum in DL variables
		dl_meta_set_cksum(cksum)

		#if !IS_IXPTYPE(__IXP28XX)
		; set SOP buffer offset. For 2800, this is always
		; constant.
		dl_meta_set_offset(temp1)
		#endif
	.end
#else
	#if !IS_IXPTYPE(__IXP28XX)
	; set SOP buffer offset. For 2800, this is always
	; constant.
	dl_meta_set_offset(*n$index++)
	#endif
#endif

reading_from_nn_ring#:
	br_inp_state[nn_empty, reading_from_nn_ring#]

	.reg tmp0 tmp1

	; Copy data from nn ring into gpr's.
	alu[dl_eop_buf_handle, --, b, *n$index++]
	alu[dl_buf_handle, --, b, *n$index++]
	alu[tmp0, --, b, *n$index++]
	alu[tmp1, --, b, *n$index++]

	ld_field_w_clr[byte_count, 0001, tmp0]
	alu[ele_num, 0x7F, and, tmp0, >>8]
	ld_field_w_clr[sop_buf_size, 0011, tmp0, >>16]

	ld_field_w_clr[packet_size, 0011, tmp1]
	ld_field_w_clr[input_port, 0011, tmp1, >>16]

	; Check if 'eop_flag' is set.
	br_bclr[tmp0, 15, dram_write#]

	; If yes, set dl_next_block.
	alu[dl_next_block, --, b, BID_NEXT_BLOCK]	

dram_write#:

	; Convert RBUF element number to RBUF element address.
	alu_shf[ele_addr, --, B, ele_num, <<PO2_RSIZE]	; ele_num * size 

#ifndef STRIP_4BYTES_CRC

	; Write the RBUF element to DRAM.
	_packet_rx_move_rbuf_to_pkt_buf(dram_handle, ele_addr, byte_count, \
									prx_rbuf_base, sig_dram)

	ctx_arb[sig_next_thread, sig_dram], defer[2]

	; Set meta-data fields that will be sent over the scratch ring.
	dl_meta_set_buffer_size(sop_buf_size)
	dl_meta_set_packet_size(packet_size)

#else //STRIP_4BYTES_CRC

	// Here, the strip 4 bytes CRC is defined.  In this case, if the byte_count
	// is zero, the mpacket ahould not be write to the DRAM, otherwise, the operation
	// is the same as normal case

	alu[--, byte_count, -, 0]
	beq[no_dram_write#]

	; Write the RBUF element to DRAM.
	_packet_rx_move_rbuf_to_pkt_buf(dram_handle, ele_addr, byte_count, \
									prx_rbuf_base, sig_dram)

	ctx_arb[sig_next_thread, sig_dram], defer[2], br[continue_operation#]

	; Set meta-data fields that will be sent over the scratch ring.
	dl_meta_set_buffer_size(sop_buf_size)
	dl_meta_set_packet_size(packet_size)

no_dram_write#:

	// No Dram write, only waite for previous thread signal

	ctx_arb[sig_next_thread ], defer[2]

	; Set meta-data fields that will be sent over the scratch ring.
	dl_meta_set_buffer_size(sop_buf_size)
	dl_meta_set_packet_size(packet_size)

continue_operation#:
#endif //STRIP_4BYTES_CRC

	; Free RBUF element.
	_packet_rx_free_rbuf_element(ele_num)

	br[end#], defer[2]

	; Set meta-data fields that will be sent over the scratch ring.
	dl_meta_set_input_port(input_port)

	; Signal next thread.
	_packet_rx_signal_next_thread(nt_sig_reg)

null_request#:

	; For next iteration, initialize meta-data cache.
	.reg zero
	immed[zero, 0]
	dl_meta_init_cache(zero, zero, zero, zero, zero, zero, zero, zero)

	; For next iteration, set meta-data fields that remain constant
	; and will be sent over the scratch ring.

	#if IS_IXPTYPE(__IXP28XX)
	; Set buf_offset. Constant in this design (set to 384).
	.reg buf_offset
	_packet_rx_get_rand_buf_offset(buf_offset)
	dl_meta_set_offset(buf_offset)
	#endif

	; Set freelist id.
	dl_meta_set_free_list(FREE_LIST_ID)

	; Set rx stat to 0 for now.
	dl_meta_set_rx_stat(0)

	; Set header type to packet type coming in on port 0 for now.
	; Will need to change if different ports handle different packet types.
	; For ex: If port 0 handles PPP and port 1 handles Ethernet.
	dl_meta_set_header_type(PORT_0_TYPE)	

	ctx_arb[sig_next_thread], defer[2]

	alu[dl_buf_handle, --, B, DL_BUF_NULL]
	alu[dl_eop_buf_handle, --, B, DL_BUF_NULL]

	; Signal next thread.
	_packet_rx_signal_next_thread(nt_sig_reg)

end#:

.end
#endm

#endif // SECOND_PACKET_RX_ME
//--------------------------------
//	Second Packet RX ME - End
//--------------------------------

#endif	//	__PACKET_RX_TWO_ME_UTIL_UC__
