/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      ATM TX Core Component
 *
 * = FILENAME
 *      ix_cc_atmtx_plugin.c
 *
 * = DESCRIPTION
 *      The file defines functions that are registered in ATM SAR as
 *      plugin functions.
 *
 * = AUTHOR
 *       Jaroslaw Kogut
 *       jaroslaw.kogut@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      11/07/2002
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */
#define IX_ERROR_FILE_IDENT "$Id: ix_cc_atmtx_plugin.c,v 1.3 2003/12/12 19:11:57 adziecia Exp $";

/**
 * System defined include files
 */
#include "ix_cc_error.h"
#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_cc.h"
#include "ix_cc_macros.h"
#include "ix_config.h"
#include "bindings.h"
#include "cc/ix_cc_atmsar.h"
#include "cc/ix_cc_atmsar_plugin.h"

/**
 * User defined include files
 */
#include "cc/ix_cc_atmtx_cci.h"
#include "cc/internal/ix_cc_atmtx_private.h"

/**
 * Preprocessor symbols and Macros used in this file
 */


/**
 * Type definitions whose scope is limited to this file
 */

/*                               funcPfx,          varName, lw, offset, mask */
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmtx_ctx,        vpi,   7,  20,       0x0FF);
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmtx_ctx,        vci,   7,  4,        0x0FFFF);
ATMSAR_BITS_SUPPORT_FUNC_DEFINE(_ix_cc_atmtx_ctx,   aal2_bit,   2,  31,       0x01);


/**
 * Variable declarations global to this file only. Externs are followed by 
 * static variables.
 */

/** 
 * Extern function prototypes.
 */


/** 
 * Static function prototypes.
 */


/** 
 * Function definitions.
 */




/******************************************************************************
 *                    ATM SAR PLUG-IN FUNCTIONS
 *****************************************************************************/

/**
 * NAME: _ix_cc_atmtx_ctx_get_entry
 *
 * DESCRIPTION:
 *    Find pointer to ATM TX Context entry. 
 * 
 * @Param:  IN  arg_pContext : pointer of core component context.
 * @Param:  IN  arg_entryId : Id of requested context
 *
 * @Return: NULL if error occurs, otherwise pointer to secondary bucket.
 */
static ix_uint32*
_ix_cc_atmtx_ctx_get_entry(
_ix_cc_atmtx_context_t *arg_pContext,
ix_uint16 arg_entryId)
{
    if (arg_entryId>=arg_pContext->pluginCfg.numberOfVcs)
    {
        return NULL;
    }
    return (arg_pContext->pSymbolTable[_IX_CC_ATMTX_AAL5_TX_CTX_SRAM_BASE_SYMBOL_ID]+((AAL5_TX_CTX_SIZE>>2)*arg_entryId));
} /* end _ix_cc_atmtx_ctx_get_entry */


/**
 * NAME: _ix_cc_atmtx_plugin_vc_create
 *
 * DESCRIPTION:
 *    This is function for plug-in registering. The ATM SAR
 *    calls the function if it services "VC create" operation. The called
 *    function should perform the operation on internal structures.
 * 
 * @Param:  IN  arg_pInstance :  pointer to structure describing serviced
 *                  instance.
 * @Param:  IN  arg_hVC :  handle to created VC.
 * @Param:  IN  arg_pVcDescr : parameters describing requested VC.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCESS - ATM port successfully created
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM - invalid input parameters
 *    IX_CC_ATMSAR_ERROR_NO_RESOURCES - no resources to add the entry 
 */
ix_error _ix_cc_atmtx_plugin_vc_create(
void* arg_pInstance,
const ix_cc_atmsar_vc_handle_t arg_hVc,
const ix_cc_atmsar_vc_descr_t* arg_pVcDescr)
{
    ix_uint32* pEntry;
    ix_uint32 newEntry[AAL5_TX_CTX_SIZE>>2];
    _ix_cc_atmtx_context_t *pCtx=arg_pInstance;
    ix_counter_64bit_handle *p64BitCountersHdls;
    
    _IX_CC_ATMTX_DBG_VC(("_ix_cc_atmtx_plugin_vc_create\n"));

    /* get pointer to context data in SRAM memory */
    pEntry=_ix_cc_atmtx_ctx_get_entry(pCtx, arg_hVc);
    if (pEntry == NULL)
    {
        return IX_ERROR_GLOBAL(IX_CC_ATMSAR_ERROR_INVALID_PARAM,
                               ("VC %d already used, returned context is null", arg_hVc));
    }

    /* set data context in SRAM memory*/
    ix_ossl_memset((void*)newEntry, 0, AAL5_TX_CTX_SIZE);
#ifdef AAL2_INCLUDED
    if (arg_pVcDescr->aalType == IX_CC_ATMSAR_AAL2)
    {
        _ix_cc_atmtx_ctx_set_aal2_bit(newEntry, 1);
    }
#endif /* AAL2_INCLUDED */    
    _ix_cc_atmtx_ctx_set_vpi(newEntry, arg_pVcDescr->spec.vpi);
    _ix_cc_atmtx_ctx_set_vci(newEntry, arg_pVcDescr->spec.vci);
    ix_ossl_memcpy((void*)pEntry, (void*)newEntry, AAL5_TX_CTX_SIZE);
    
    if (pCtx->counters)
    {
        /* clear statistic counters for VC */
        p64BitCountersHdls=&(pCtx->hPktsPerVcq[arg_hVc]);
        IX_ERROR_CRT(ix_rm_counter_64bit_set_value(*(p64BitCountersHdls), 0),
                     IX_CC_ATMTX_ERROR_RM_SET_64BIT_COUNTER, 
                     IX_ERROR_LEVEL_REMOTE);
        
        p64BitCountersHdls=&(pCtx->hCellsPerVcq[arg_hVc]);
        IX_ERROR_CRT(ix_rm_counter_64bit_set_value(*(p64BitCountersHdls), 0),
                     IX_CC_ATMTX_ERROR_RM_SET_64BIT_COUNTER,
                     IX_ERROR_LEVEL_REMOTE);
    } /* if counters enabled */
    
    return IX_SUCCESS;
}

/**
 * NAME: _ix_cc_atmtx_plugin_vc_update
 *
 * DESCRIPTION:
 *    This is function for plug-in registering. The ATM SAR
 *    calls the function if it services "VC update" operation. The called
 *    function should perform the operation on internal structures.
 * 
 * @Param:  IN  arg_pInstance :  pointer to structure describing serviced
 *                  instance.
 * @Param:  IN  arg_hVC :  handle to created VC.
 * @Param:  IN  arg_pVcDescr : parameters describing requested VC.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCESS - ATM port successfully created
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM - invalid input parameters
 *    IX_CC_ATMSAR_ERROR_NO_RESOURCES - no resources to add the entry 
 */
ix_error _ix_cc_atmtx_plugin_vc_update(
void* arg_pInstance,
const ix_cc_atmsar_vc_handle_t arg_hVc,
const ix_cc_atmsar_update_data_t* arg_pUpdate)
{
    ix_uint32* pEntry;
    _ix_cc_atmtx_context_t *pCtx=arg_pInstance;
    
    _IX_CC_ATMTX_DBG_VC(("_ix_cc_atmtx_plugin_vc_update\n"));

    /* get pointer to context data in SRAM memory */
    pEntry=_ix_cc_atmtx_ctx_get_entry(pCtx, arg_hVc);
    if (pEntry == NULL)
    {
        return IX_ERROR_GLOBAL(IX_CC_ATMSAR_ERROR_INVALID_PARAM,
                               ("VC %d not defined.", arg_hVc));
    }

    return IX_SUCCESS;
}

/**
 * NAME: _ix_cc_atmtx_plugin_vc_remove
 *
 * DESCRIPTION:
 *    This is function for plug-in registering. The ATM SAR
 *    calls the function if it services "VC remove" operation. The called
 *    function should perform the operation on internal structures.
 * 
 * @Param:  IN  arg_pInstance :  pointer to structure describing serviced
 *                  instance.
 * @Param:  IN  arg_hVc : handle to VC to be removed
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - vc successfully removed
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM - invalid input parameters
 */
ix_error _ix_cc_atmtx_plugin_vc_remove(
void* arg_pInstance,
ix_cc_atmsar_vc_handle_t arg_hVc)
{
    _IX_CC_ATMTX_DBG_VC(("_ix_cc_atmtx_plugin_vc_remove\n"));
    return IX_SUCCESS;
}


/**
 * NAME: ix_cc_atmtx_vc_show_stats
 *
 * DESCRIPTION:
 *    The function shows statistic counters for VC.
 * 
 * @Param:  IN  arg_pContext : pointer of core component context.
 * @Param:  IN  arg_hVc : handle to VC to be removed
 * 
 * @Return: none
 */
void ix_cc_atmtx_vc_show_stats(
_ix_cc_atmtx_context_t *arg_pContext,
ix_cc_atmsar_vc_handle_t arg_hVc)
{
    ix_uint64 packets;
    ix_uint64 cells;
    ix_counter_64bit_handle *p64BitCountersHdls;

    _IX_CC_ATMTX_DBG_VC(("ATM TX: VC[%-5d] STATS\n", arg_hVc));

    if (arg_pContext == NULL)
    {
        _IX_CC_ATMTX_DBG_VC(("ATM TX context pointer is NULL\n"));
        return;
    }
    
    if (arg_hVc>=arg_pContext->pluginCfg.numberOfVcs)
    {
        _IX_CC_ATMTX_DBG_VC(("Wrong VC ID. Statistic is not presented\n"));
        return;
    }

    if (arg_pContext->counters)
    {
        p64BitCountersHdls=&(arg_pContext->hPktsPerVcq[arg_hVc]);
        ix_rm_counter_64bit_get_value(*(p64BitCountersHdls),
                                      &packets);
        
        p64BitCountersHdls=&(arg_pContext->hCellsPerVcq[arg_hVc]);
        ix_rm_counter_64bit_get_value(*(p64BitCountersHdls),
                                      &cells);
        
        _IX_CC_ATMTX_DBG_VC(("  cell     %lx%lx\n", ((ix_int32*)&cells)[0],((ix_int32*)&cells)[1]));
        _IX_CC_ATMTX_DBG_VC(("  packets  %lx%lx\n", ((ix_int32*)&packets)[0],((ix_int32*)&packets)[1]));
    }
    else
    {
        _IX_CC_ATMTX_DBG_VC(("  Statistics disabled...\n"));
    }    
}


/**
 * NAME: _ix_cc_atmtx_plugin_get_vc_stats
 *
 * DESCRIPTION:
 *    This is function for plug-in registering. The ATM SAR
 *    calls the function if it services "GET VC STATS" operation. The called
 *    function should perform the operation on internal structures.
 * 
 * @Param:  IN  arg_pInstance :  pointer to structure describing serviced
 *                  instance.
 * @Param:  IN  arg_hVc : handle to VC to be removed
 * @Param:  OUT arg_pLookupResult : pointer to statistic structure
 *                  where the requested data will be stored
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - statistics successfully retrieved
 *    IX_CC_ATMSAR_ERROR_INVALID_PARAM - invalid input parameters
 */
ix_error _ix_cc_atmtx_plugin_get_vc_stats(
void* arg_pInstance,
ix_cc_atmsar_vc_handle_t arg_hVc,
ix_cc_atmsar_vc_stats_t *arg_pLookupResult)
{
    _ix_cc_atmtx_context_t *pCtx=arg_pInstance;
    ix_counter_64bit_handle *p64BitCountersHdls;

    _IX_CC_ATMTX_DBG_VC(("_ix_cc_atmtx_plugin_get_vc_stats\n"));

    if (pCtx->counters)
    {

        p64BitCountersHdls=&(pCtx->hPktsPerVcq[arg_hVc]);
        
        IX_ERROR_CRT(ix_rm_counter_64bit_get_value(*(p64BitCountersHdls),
                                                   &(arg_pLookupResult->tx.packets)),
                     IX_CC_ATMTX_ERROR_RM_GET_64BIT_COUNTER, 
                     IX_ERROR_LEVEL_REMOTE);
        
        p64BitCountersHdls=&(pCtx->hCellsPerVcq[arg_hVc]);
        
        IX_ERROR_CRT(ix_rm_counter_64bit_get_value(*(p64BitCountersHdls),
                                                   &(arg_pLookupResult->tx.cells)),
                     IX_CC_ATMTX_ERROR_RM_GET_64BIT_COUNTER, 
                     IX_ERROR_LEVEL_REMOTE);
    }
    else
    {
        _IX_CC_ATMTX_DBG_VC(("   Statistics disabled...\n"));
    }
    
    return IX_SUCCESS;
}


/**
 * NAME: ix_cc_atmtx_port_show_stats
 *
 * DESCRIPTION:
 *    The function shows statistic counters for port.
 * 
 * @Param:  IN  arg_pContext : pointer of core component context.
 * @Param:  IN  arg_port : ATM port ID
 * 
 * @Return: none
 */
void ix_cc_atmtx_port_show_stats(
_ix_cc_atmtx_context_t *arg_pContext,
ix_cc_atmsar_port_handle_t arg_port)
{
    ix_uint64 packets;
    ix_uint64 cells;
    ix_counter_64bit_handle *p64BitCountersHdls;

    _IX_CC_ATMTX_DBG_PORT(("ATM TX: PORT[%-5d] STATS\n", arg_port));

    if (arg_pContext == NULL)
    {
        _IX_CC_ATMTX_DBG_PORT(("ATM TX context pointer is NULL\n"));
        return;
    }
    
    if (arg_port>=arg_pContext->pluginCfg.numberOfPorts)
    {
        _IX_CC_ATMTX_DBG_PORT(("Wrong VC ID. Statistic is not presented\n"));
        return;
    }

    if (arg_pContext->counters)
    {
        p64BitCountersHdls=&(arg_pContext->hPktsPerOutPort[arg_port]);
        ix_rm_counter_64bit_get_value(*(p64BitCountersHdls),
                                      &packets);
        
        p64BitCountersHdls=&(arg_pContext->hCellsPerOutPort[arg_port]);
        ix_rm_counter_64bit_get_value(*(p64BitCountersHdls),
                                      &cells);
        
        _IX_CC_ATMTX_DBG_PORT(("  cell     %lx%lx\n", ((ix_int32*)&cells)[0],((ix_int32*)&cells)[1]));
        _IX_CC_ATMTX_DBG_PORT(("  packets  %lx%lx\n", ((ix_int32*)&packets)[0],((ix_int32*)&packets)[1]));
    }
    else
    {
        _IX_CC_ATMTX_DBG_PORT(("  Statistics disabled...\n"));
    }
}


/**
 * NAME: _ix_cc_atmtx_plugin_get_port_stats
 *
 * DESCRIPTION:
 *    This is function for plug-in registering. The ATM SAR
 *    calls the function if it services "GET PORT STATS" operation. The called
 *    function should perform the operation on internal structures.
 * 
 * @Param:  IN  arg_pInstance :  pointer to structure describing serviced
 *                  instance.
 * @Param:  IN  arg_port : ATM port ID
 * @Param:  OUT arg_pLookupResult : pointer to ix_cc_sar_port_stats structure
 *                  where the requested data will be stored
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 *    IX_SUCCESS - statistics successfully retrieved
 *    IX_CC_INVALID_PORT - specified ATM port does not exist 
 */
ix_error _ix_cc_atmtx_plugin_get_port_stats(
void* arg_pInstance,
ix_cc_atmsar_port_handle_t arg_port,
ix_cc_atmsar_port_stats_t *arg_pLookupResult)
{
    _ix_cc_atmtx_context_t *pCtx=arg_pInstance;
    ix_counter_64bit_handle *p64BitCountersHdls;
    
    _IX_CC_ATMTX_DBG_PORT(("_ix_cc_atmtx_plugin_get_port_stats\n"));
    
    p64BitCountersHdls=&(pCtx->hPktsPerOutPort[arg_port]);

    if (pCtx->counters)
    {
        
        IX_ERROR_CRT(ix_rm_counter_64bit_get_value(*(p64BitCountersHdls),
                                                   &(arg_pLookupResult->tx.packets)),
                     IX_CC_ATMTX_ERROR_RM_GET_64BIT_COUNTER, 
                     IX_ERROR_LEVEL_REMOTE);
        
        p64BitCountersHdls=&(pCtx->hCellsPerOutPort[arg_port]);
        
        IX_ERROR_CRT(ix_rm_counter_64bit_get_value(*(p64BitCountersHdls),
                                                   &(arg_pLookupResult->tx.cells)),
                     IX_CC_ATMTX_ERROR_RM_GET_64BIT_COUNTER, 
                     IX_ERROR_LEVEL_REMOTE);
        
    }
    else
    {
        _IX_CC_ATMTX_DBG_PORT(("  Statistics disabled...\n"));    
    }
    
    return IX_SUCCESS;    
}


