/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2X00 Network Processor, Release 5
 *
 * = FILENAME
 *      ix_cc_csix_tx.c
 *
 * = DESCRIPTION
 *      This file contains CSIX Tx CC interface APIs, init and fini, and 
 *      internal functions.
 *
 * = CHANGE HISTORY
 *      6/10/2002 - Created.
 *
 * ============================================================================
 * $Id: ix_cc_csix_tx.c,v 1.42 2003/09/22 17:20:30 ktseng Exp $
 */

/**
 * System defined include files required.
 */


/**
 * User defined include files required.
 */
#include "ix_cc_error.h"

#include "ix_types.h"
#include "ix_error.h"
#include "ix_cc.h"
#include "bindings.h"
#include "ix_cc_properties.h"

#include "ix_rm.h"
#include "ix_ossl.h"

#ifndef IX_EXCLUDE_CCI
#include "ix_cci.h"
#endif

#include "cc/ix_cc_csix_tx.h"
#include "cc/internal/ix_cc_csix_tx_int.h"


/**
 * Exported variable declarations global to this component only.
 */
ix_cc_csix_tx_ctl_blk *g_pCcCsixTxCtlBlk = NULL; /* control block */



/**
 * Static function prototypes
 */
PRIVATE ix_error _ix_cc_csix_tx_get_static_data(ix_cc_csix_tx_ctl_blk *);
PRIVATE void _ix_cc_csix_tx_free_ctl_blk(ix_cc_csix_tx_ctl_blk *);
PRIVATE ix_error _ix_cc_csix_tx_create_cb_ctx_tables(ix_cc_csix_tx_ctl_blk *);



/**
 * function definitions.
 */


/**
 * NAME: ix_cc_csix_tx_init
 * 
 * DESCRIPTION: This function initializes the csix tx core component.
 * 
 * @Param: arg_CcHandle - IN: cc handle.
 * @Param: arg_ppContext - IN: pointer to cc init context memory.
 *                         OUT: pointer to component context memory.
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_cc_csix_tx_init(
    ix_cc_handle arg_CcHandle, 
    void **arg_ppContext)
{

    ix_cc_csix_tx_ctl_blk *pCtlBlk = NULL;
    ix_error err = IX_SUCCESS;
    ix_uint32 i;

    /* Variables used in patching symbols */
    ix_imported_symbol symbol[CSIX_TX_NUM_SYMBOLS_PATCHED];
    ix_memory_type memType;
    ix_uint32 memCh;
    ix_uint32 memOffsetTxCtx;
    ix_uint32 memOffset32BitCounter;
    void *pPhyAddr;

    /* Variables used in retrieving static data from ix_cc_init_context */
    ix_cc_init_context *pInitCtx;
    ix_cc_properties_node *pPropNode;



    /* Check whether arg_ppContext is invalid */
    if (arg_ppContext == (void **)NULL)
    {
        return(IX_ERROR_LOCAL(IX_CC_ERROR_NULL,
                              ("NULL argument")));
    }

    /* Get the generic init context */
    pInitCtx = (ix_cc_init_context *)*arg_ppContext;

    /* Allocate the control block structure */
    if ((pCtlBlk = (ix_cc_csix_tx_ctl_blk *)ix_ossl_malloc(sizeof(ix_cc_csix_tx_ctl_blk))) == NULL)
    {
        return(IX_ERROR_REMOTE(IX_CC_ERROR_OOM_SYSTEM,
                               ("Out of memory")));
    }

    /* Initialize the structure */
    ix_ossl_memset((void *)pCtlBlk, 
                   0, 
                   (ix_ossl_size_t)sizeof(ix_cc_csix_tx_ctl_blk));

    /* Retrieve static configuration data */
    IX_ERROR_C(_ix_cc_csix_tx_get_static_data(pCtlBlk), err);
    if (err != IX_SUCCESS)
    {
        /* Free all memories allocated earlier */
        _ix_cc_csix_tx_free_ctl_blk(pCtlBlk);

	return(err);
    }

    /**
     * Retrieve the blade id information.
     */
    if ((pPropNode = pInitCtx->pPropertyList) == NULL)
    {
        /* free all memories allocated earlier */
        _ix_cc_csix_tx_free_ctl_blk(pCtlBlk);

        return(IX_ERROR_REMOTE(IX_CC_ERROR_ENTRY_NOT_FOUND, 
                               ("Property node can not be found from the ix_cc_init_context structure.")));
    }

    /**
     * Store the local blade id.  
     */
    pCtlBlk->ingressBladeId = pPropNode->pPropertyInfo->blade_id;

    /* Allocate CSIX Tx microblock context memory */
    IX_ERROR_CT(ix_rm_mem_alloc(IX_MEMORY_TYPE_SRAM,
                                CC_CSIX_TX_DEFAULT_SRAM_CH_NUM, 
                                CC_CSIX_TX_SIZE_CTX_MEM, 
                                (void **)&pCtlBlk->pTxCtx),
                err,
                IX_CC_ERROR_OOM_SRAM,
                IX_ERROR_LEVEL_REMOTE);
    if (err != IX_SUCCESS)
    {
        /* Free all memories allocated earlier */
        _ix_cc_csix_tx_free_ctl_blk(pCtlBlk);

	return(err);
    }

    /* Initialize the memory */
    ix_ossl_memset((void *)pCtlBlk->pTxCtx,
                   0, 
                   CC_CSIX_TX_SIZE_CTX_MEM);

    /* Allocate memory for 32-bit counters */
#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

    IX_ERROR_CT(ix_rm_mem_alloc(IX_MEMORY_TYPE_SRAM,
                                CC_CSIX_TX_DEFAULT_SRAM_CH_NUM_COUNTERS,
                                CSIX_TX_SIZE_32BIT_COUNTER_BUF,
                                (void **)&pCtlBlk->pCounterBase32Bit),
                err,
                IX_CC_ERROR_OOM_SRAM,
                IX_ERROR_LEVEL_REMOTE);

#else

    IX_ERROR_CT(ix_rm_mem_alloc(IX_MEMORY_TYPE_SRAM,
                                CC_CSIX_TX_DEFAULT_SRAM_CH_NUM,
                                CSIX_TX_SIZE_32BIT_COUNTER_BUF,
                                (void **)&pCtlBlk->pCounterBase32Bit),
                err,
                IX_CC_ERROR_OOM_SRAM,
                IX_ERROR_LEVEL_REMOTE);

#endif /* (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */
    if (err != IX_SUCCESS)
    {
        /* Free all memories allocated earlier */
        _ix_cc_csix_tx_free_ctl_blk(pCtlBlk);

	return(err);
    }

    /* Initialize the memory */
    ix_ossl_memset((void *)pCtlBlk->pCounterBase32Bit,
                   0, 
                   CSIX_TX_SIZE_32BIT_COUNTER_BUF);

    /* Allocate memory for 64-bit counter handler array */
    if (((pCtlBlk->pCounterHandlers64Bit = (ix_counter_64bit_handle *)ix_ossl_malloc(CSIX_TX_SIZE_64BIT_COUNTER_HDL))) == NULL)
    {
        /* free all memories allocated earlier */
        _ix_cc_csix_tx_free_ctl_blk(pCtlBlk);

        return(IX_ERROR_REMOTE(IX_CC_ERROR_OOM_SYSTEM,
                               ("Out of memory")));
    }

    /* Allocate 64-bit counters from RM */
    IX_ERROR_CT(ix_rm_counter_64bit_new(CSIX_TX_TOTAL_NUM_COUNTERS,
                                        pCtlBlk->pCounterBase32Bit,
                                        CC_CSIX_TX_32BIT_COUNTER_OVERFLOW_TIME,
                                        pCtlBlk->pCounterHandlers64Bit),
                err,
                IX_CC_ERROR_OOM_64BIT_COUNTER, 
                IX_ERROR_LEVEL_REMOTE);
    if (err != IX_SUCCESS)
    {
        /* Free all memories allocated earlier */
        _ix_cc_csix_tx_free_ctl_blk(pCtlBlk);

	return(err);
    }

    /**
     * patch various symbols and memory bases into CSIX Tx microblock:
     * 1. ingress blad id.
     * 2. microblock Tx context memory.
     * 3. 32-bit counter memory base. 
     */

    /* First obtain the base offsets */
    IX_ERROR_CT(ix_rm_get_phys_offset(pCtlBlk->pTxCtx,
                                      &memType,
                                      &memCh,
                                      &memOffsetTxCtx,
                                      &pPhyAddr),
                err,
                IX_CC_CSIX_TX_ERROR_RM_GET_PHYS_OFFSET, 
                IX_ERROR_LEVEL_REMOTE);
    if (err != IX_SUCCESS)
    {
        /* Free all memories allocated earlier */
        _ix_cc_csix_tx_free_ctl_blk(pCtlBlk);

	return(err);
    }

    /* Pack the SRAM channel number into the offset */
    IX_CC_SET_SRAM_CHANNEL_INTO_OFFSET(memCh, memOffsetTxCtx);

    IX_ERROR_CT(ix_rm_get_phys_offset(pCtlBlk->pCounterBase32Bit,
                                      &memType,
                                      &memCh,
                                      &memOffset32BitCounter,
                                      &pPhyAddr),
                err,
                IX_CC_CSIX_TX_ERROR_RM_GET_PHYS_OFFSET, 
                IX_ERROR_LEVEL_REMOTE);
    if (err != IX_SUCCESS)
    {
        /* Free all memories allocated earlier */
        _ix_cc_csix_tx_free_ctl_blk(pCtlBlk);

	return(err);
    }

    /* Pack the SRAM channel number into the offset */
    IX_CC_SET_SRAM_CHANNEL_INTO_OFFSET(memCh, memOffset32BitCounter);

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

    /* Prepare the symbol array */
    symbol[0].m_Value = pCtlBlk->ingressBladeId;
    symbol[0].m_Name = SYMBOL_CSIX_TX_INGRESS_BLADE_ID;

    symbol[1].m_Value = memOffsetTxCtx;
    symbol[1].m_Name = SYMBOL_CSIX_TX_CTX_BASE;

#else /* (_IX_HARDWARE_TYPE_ != _IX_HW_2800_) */

    /* Prepare the symbol array */
    symbol[0].m_Value = pCtlBlk->ingressBladeId;
    symbol[0].m_Name = SYMBOL_CSIX_TX_INGRESS_BLADE_ID;

    symbol[1].m_Value = memOffsetTxCtx;
    symbol[1].m_Name = SYMBOL_CSIX_TX_CTX_BASE;

    symbol[2].m_Value = memOffset32BitCounter; 
    symbol[2].m_Name = SYMBOL_CSIX_TX_COUNTERS_BASE;

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_2800_) */

    /* Patch them into each microengine used by the microblock */
    for (i = 0; i < IX_CC_CSIX_TX_MAX_ME_NUM; i++)
    {
        if (pCtlBlk->mev2Num & (0x1 << i))
        {
            IX_ERROR_CT(ix_rm_ueng_patch_symbols(g_aMicroengineNumberMapping[i],
                                                 CSIX_TX_NUM_SYMBOLS_PATCHED,
                                                 symbol),
                        err,
                        IX_CC_CSIX_TX_ERROR_RM_PATCH_SYMBOL, 
                        IX_ERROR_LEVEL_REMOTE);
           
            if (err != IX_SUCCESS)
            {
                /* Free all memories allocated earlier */
                _ix_cc_csix_tx_free_ctl_blk(pCtlBlk);

	        return(err);
            }
        }
    }

    /**
     * Create mutex for critical section. 
     */
    IX_ERROR_CT(ix_ossl_mutex_init(IX_OSSL_MUTEX_UNLOCK, 
                                   &(pCtlBlk->mid)),
                err,
                IX_CC_CSIX_TX_ERROR_OSSL_MUTEX_INIT, 
                IX_ERROR_LEVEL_REMOTE);
    if (err != IX_SUCCESS)
    {
        /* Free all memories allocated earlier */
        _ix_cc_csix_tx_free_ctl_blk(pCtlBlk);

	return(err);
    }

    /**
     * Allocate free callback context entries for each async messaging
     * helper API function.
     */
    IX_ERROR_C(_ix_cc_csix_tx_create_cb_ctx_tables(pCtlBlk), err);
    if (err != IX_SUCCESS)
    {
        /* Free all memories allocated earlier */
        _ix_cc_csix_tx_free_ctl_blk(pCtlBlk);

	return(err);
    }

#ifndef IX_EXCLUDE_CCI

    /* Register message handler */
    IX_ERROR_CT(ix_cci_cc_add_message_handler(arg_CcHandle,
                                              IX_CC_CSIX_TX_MSG_INPUT,
                                              ix_cc_csix_tx_msg_handler,
                                              IX_INPUT_TYPE_SINGLE_SRC),
                err,
                IX_CC_CSIX_TX_ERROR_CCI_REG_HDLR, 
                IX_ERROR_LEVEL_REMOTE);
    if (err != IX_SUCCESS)
    {
        /* Free all memories allocated earlier */
        _ix_cc_csix_tx_free_ctl_blk(pCtlBlk);

	return(err);
    }

#else  /* IX_EXCLUDE_CCI */

    /* Register message handler */
    IX_ERROR_CT(ix_rm_message_handler_register(IX_CC_CSIX_TX_MSG_INPUT,
                                               ix_cc_csix_tx_msg_handler,
                                               pCtlBlk),
                err,
                IX_CC_CSIX_TX_ERROR_RM_REG_HDLR, 
                IX_ERROR_LEVEL_REMOTE);
    if (err != IX_SUCCESS)
    {
        /* Free all memories allocated earlier */
        _ix_cc_csix_tx_free_ctl_blk(pCtlBlk);

	return(err);
    }

#endif /* IX_EXCLUDE_CCI */

    /* assign output argument */
    *arg_ppContext = (void *)pCtlBlk;

    /**
     * Store the address of the control block structure to a global 
     * varible that can be used later in helper API functions.
     */
    g_pCcCsixTxCtlBlk = pCtlBlk;

    return(IX_SUCCESS);

} /* ix_cc_csix_tx_init */



/**
 * NAME: ix_cc_csix_tx_fini
 * 
 * DESCRIPTION: This function terminates the csix tx core component.
 * 
 * @Param: arg_CcHandle - IN: cc handle created by CCI.
 * @Param: arg_pContext - IN: pointer to component context memory.
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_cc_csix_tx_fini(
    ix_cc_handle arg_CcHandle, 
    void *arg_pContext)
{

    ix_error err = IX_SUCCESS;


    /* Check whether arg_pContext is valid */
    if (arg_pContext == (void *)NULL)
    {
        return(IX_ERROR_LOCAL(IX_CC_ERROR_NULL,
                              ("NULL argument")));
    }

    /* Start the critical section */
    ix_ossl_mutex_lock(((ix_cc_csix_tx_ctl_blk *)arg_pContext)->mid, 
                       IX_OSSL_WAIT_FOREVER);

#ifndef IX_EXCLUDE_CCI

    /* Remove message handler from CCI */
    IX_ERROR_CT(ix_cci_cc_remove_message_handler(IX_CC_CSIX_TX_MSG_INPUT),
                err,
                IX_CC_CSIX_TX_ERROR_CCI_REM_HDLR, 
                IX_ERROR_LEVEL_REMOTE);
    if (err != IX_SUCCESS)
    {
        /* Exit the critical section */
        ix_ossl_mutex_unlock(((ix_cc_csix_tx_ctl_blk *)arg_pContext)->mid);

	return(err);
    }

#else /* IX_EXCLUDE_CCI */

    /* Remove message handler from CCI */
    IX_ERROR_CT(ix_rm_message_handler_unregister(IX_CC_CSIX_TX_MSG_INPUT),
                err,
                IX_CC_CSIX_TX_ERROR_RM_REM_HDLR, 
                IX_ERROR_LEVEL_REMOTE);
    if (err != IX_SUCCESS)
    {
        /* Exit the critical section */
        ix_ossl_mutex_unlock(((ix_cc_csix_tx_ctl_blk *)arg_pContext)->mid);

	return(err);
    }

#endif /* IX_EXCLUDE_CCI */

    /* Exit the critical section */
    ix_ossl_mutex_unlock(((ix_cc_csix_tx_ctl_blk *)arg_pContext)->mid);

    /* Free all memories and resources indicated in the arg_pContext */
    _ix_cc_csix_tx_free_ctl_blk((ix_cc_csix_tx_ctl_blk *)arg_pContext);

    /* Reset the global pointer of the control block structure */
    g_pCcCsixTxCtlBlk = NULL;

    return(IX_SUCCESS);

} /* ix_cc_csix_tx_fini */





/**********************
 * Internal functions *
 **********************/


/**
 * NAME: _ix_cc_csix_tx_get_static_data
 * 
 * DESCRIPTION: This function retrieves all CSIX Tx static data from 
 *              either system repository or ix_cc_csix_tx.h and store them 
 *              into the corresponding field in the control block 
 *              structure passed in.
 * 
 * @Param: arg_pCtlBlk - IN: pointer to component context memory.
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
PRIVATE ix_error _ix_cc_csix_tx_get_static_data(
    ix_cc_csix_tx_ctl_blk *arg_pCtlBlk)
{

#ifdef IX_INCLUDE_REGISTRY

    ix_configuration_property_handle hCoreRoot = IX_CP_CORE_PROPERTY_ROOT;
    ix_configuration_property_handle hProperty;

    /* open the property */
    IX_ERROR_CRT(ix_rm_cp_property_open(hCoreRoot,
                                        PROP_CSIX_TX_MEV2_NUM_MASK,
                                        &hProperty),
                 IX_CC_ERROR_ENTRY_NOT_FOUND, 
                 IX_ERROR_LEVEL_REMOTE);

    /* read the property */
    IX_ERROR_CRT(ix_rm_cp_property_get_value_uint32(hProperty, 
                                                    &(arg_pCtlBlk->mev2Num)),
                 IX_CC_ERROR_ENTRY_NOT_FOUND, 
                 IX_ERROR_LEVEL_REMOTE);

#else /* IX_INCLUDE_REGISTRY */

    arg_pCtlBlk->mev2Num = CSIX_TX_MEV2_NUMBER;

#endif /* IX_INCLUDE_REGISTRY */

    return(IX_SUCCESS);

} /* _ix_cc_csix_tx_get_static_data */




/**
 * NAME: _ix_cc_csix_tx_create_cb_ctx_tables
 * 
 * DESCRIPTION: This function creates callback context table that 
 *              contain free callback context entry structure for 
 *              being used by each async messaging API that has 
 *              callback defined.
 * 
 * @Param: arg_pCtlBlk - IN: pointer to component context memory.
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
PRIVATE ix_error _ix_cc_csix_tx_create_cb_ctx_tables(
    ix_cc_csix_tx_ctl_blk *arg_pCtlBlk)
{

    /**
     * Create table for ix_cc_csix_tx_async_get_statistisc_inof().
     */

    /* Allocate cb_ctx_table structure first */
    if (((arg_pCtlBlk->pCbCtxTblGetStatInfo = (ix_cc_csix_tx_cb_ctx_tbl *)ix_ossl_malloc(sizeof(ix_cc_csix_tx_cb_ctx_tbl)))) == NULL)
    {
        return(IX_ERROR_REMOTE(IX_CC_ERROR_OOM_SYSTEM,
                               ("Out of memory")));
    } 
    
    /* Initialize the structure */
    ix_ossl_memset((void *)arg_pCtlBlk->pCbCtxTblGetStatInfo, 
                   0, 
                   (ix_ossl_size_t)sizeof(ix_cc_csix_tx_cb_ctx_tbl));

    /* Allocate the callback context entries and put them in the table structure */
    if (((arg_pCtlBlk->pCbCtxTblGetStatInfo->pTable = (ix_cc_csix_tx_cb_ctx_entry *)ix_ossl_malloc(CSIX_TX_SIZE_CB_CTX_TBL_GET_STAT))) == NULL)
    {
        ix_ossl_free(arg_pCtlBlk->pCbCtxTblGetStatInfo);
        return(IX_ERROR_REMOTE(IX_CC_ERROR_OOM_SYSTEM,
                               ("Out of memory")));
    }

    return(IX_SUCCESS);

} /* _ix_cc_csix_tx_create_cb_ctx_tables */






/**
 * NAME: _ix_cc_csix_tx_free_ctl_blk
 * 
 * DESCRIPTION: This function frees all memories and resources specified 
 *              by each pointer member of the control block structure.
 *              Finally, it frees the control block structure.
 * 
 * @Param: arg_pCtlBlk - IN: pointer to component context memory.
 * 
 * @Return: None
 */
PRIVATE void _ix_cc_csix_tx_free_ctl_blk(
    ix_cc_csix_tx_ctl_blk *arg_pCtlBlk)
{
    ix_uint32 i;


    /* Free the microblock context memory */
    if (arg_pCtlBlk->pTxCtx != NULL)
    {
        if (ix_rm_mem_free(arg_pCtlBlk->pTxCtx) != IX_SUCCESS)
        {
            ix_ossl_message_log("Resource manager memory free failed.\n");
        }
    }

    /* Free 64-bit counter handle array */
    if (arg_pCtlBlk->pCounterHandlers64Bit != NULL)
    {
        /* First delete each 64-bit counter from Resource Manager */
        for (i = 0; i < CSIX_TX_TOTAL_NUM_COUNTERS; i++)
        {
            ix_rm_counter_64bit_delete(*(arg_pCtlBlk->pCounterHandlers64Bit + i));
        }

        /* Then free the counter handler array */
        ix_ossl_free(arg_pCtlBlk->pCounterHandlers64Bit);
    }

    /* Free 32-bit counter memory */
    if (arg_pCtlBlk->pCounterBase32Bit != NULL)
    {
        if (ix_rm_mem_free(arg_pCtlBlk->pCounterBase32Bit) != IX_SUCCESS)
        {
            ix_ossl_message_log("Resource manager memory free failed.\n");
        }
    }

    /* Free async callback context table and entries */
    if (arg_pCtlBlk->pCbCtxTblGetStatInfo != NULL)
    {
        /* Free the entries first */
        if (arg_pCtlBlk->pCbCtxTblGetStatInfo->pTable != NULL)
        {
            ix_ossl_free(arg_pCtlBlk->pCbCtxTblGetStatInfo->pTable);
        }

        /* Free the table structure */
        ix_ossl_free(arg_pCtlBlk->pCbCtxTblGetStatInfo);
    }

    /* Destroy the critical section mutex */
    if (arg_pCtlBlk->mid != NULL)
    {
        if (ix_ossl_mutex_fini(arg_pCtlBlk->mid) != IX_SUCCESS)
        {
            ix_ossl_message_log("OSSL mutex fini failed.\n");
        }
    }

    /* free the control block structure itself */
    ix_ossl_free((void *)arg_pCtlBlk);

} /* _ix_cc_csix_tx_free_ctl_blk */

