/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2X00 Network Processor, Release 5
 *
 * = FILENAME
 *      ix_cc_atm_pos_tx_core.c
 *
 * = DESCRIPTION
 *      This file contains Library APIs and related 
 *      internal functions of ATM/POS Tx core component.
 *
 * = CHANGE HISTORY
 *      6/10/2002 - Created.
 *
 * ============================================================================
 * $Id: ix_cc_atm_pos_tx_core.c,v 1.30 2003/10/29 02:17:14 ktseng Exp $
 */

/**
 * System defined include files required.
 */
#include <ix_ossl.h>
#include <ix_rm.h>
#ifndef IX_EXCLUDE_CCI
#include <ix_cci.h>
#endif

/**
 * User defined include files required.
 */
#include "ix_cc_error.h"

#include "ix_types.h"
#include "ix_error.h"
#include "ix_cc.h"
#include "bindings.h"
#include "ix_cc_properties.h"

#include "cc/ix_cc_atm_pos_tx.h"
#include "cc/internal/ix_cc_atm_pos_tx_int.h"

#if (defined IX_PLATFORM_2401) || (defined IX_PLATFORM_2801)

/* Platform specific
 * IXDP2401 platform has a different framer driver
 */
#if (_IX_BOARD_TYPE_ == _IX_IXDP2401_) || (_IX_BOARD_TYPE_ == _IX_IXDP2801_)
#include "pos_atm_media_drv.h"
#else
/* IXF device driver common includes */
#include "common_types.h"
#include "common_def.h"
#include "ixf_api_d.h" /* BroadBand Utility Definitions and Function headers */
#include "ixf_api_sonet_d.h"
#include "ixf_api.h"
#include "ixf_api_sonet.h"
#include "ixf6048d.h" /* to fix header dependency problem of IXFAPI */
#include "ixf6048f.h"
#endif /* _IX_BOARD_TYPE_ */

#else  /* IX_PLATFORM_2401 || IX_PLATFORM_2801 */

/* IXF device driver common includes */
#include "common_types.h"
#include "common_def.h"
#include "ixf_api_d.h" /* BroadBand Utility Definitions and Function headers */
#include "ixf_api_sonet_d.h"
#include "ixf_api.h"
#include "ixf_api_sonet.h"
#include "ixf6048d.h" /* to fix header dependency problem of IXFAPI */
#include "ixf6048f.h"

#endif /* IX_PLATFORM_2401 || IX_PLATFORM_2801 */


/**
 * Static function prototypes
 */
PRIVATE ix_error _ix_cc_atm_pos_get_channel_number(ix_cc_atm_pos_tx_ctl_blk *, ix_uint32 , ix_uint32 *);


/**
 * function definitions.
 */


/**
 * NAME: ix_cc_atm_pos_tx_get_statistics_info
 * 
 * DESCRIPTION: This function retrieves a statistics counter value of 
 *              ATM/POS Tx CC.
 * 
 * @Param: arg_Entity - IN: type of statistics info.
 * @Param: arg_Index - IN: index number to the statistics info.  Can be
 *                         either port number or VCQ number depending 
 *                         on the entity type.
 * @Param: arg_pBuffer - IN: pointer to the location for storing the data.
 * @Param: arg_pContext - IN: pointer to the component context allocated
 *                            in the component's init function. 
 * 
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_cc_atm_pos_tx_get_statistics_info(
    ix_cc_atm_pos_tx_statistics_info arg_Entity,
    ix_uint32 arg_Index,
    ix_cc_atm_pos_tx_statistics_info_data *arg_pBuffer,
    void *arg_pContext)
{

    ix_counter_64bit_handle *pH64Bit;
    ix_uint32 chNum;
    ix_uint32 numCounters;
    ix_uint32 counter;
    ix_cc_atm_pos_tx_ctl_blk *pCcContext;
    ix_uint32 i;



    /* Check any NULL pointer argument */
    if (NULL == arg_pBuffer)
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("NULL buffer argument")));
    }

    /* Check NULL arg_pContext argument */
    if (NULL == arg_pContext)
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("NULL context argument")));
    }

    /*
     * Check the ranges of the arg_Entity and arg_Index.
     * arg_Index contain the either port id or VCQ number depending on
     * the counter type indicated in the arg_Entity. 
     */
    switch(arg_Entity)
    {
        case IX_CC_POS_TX_PACKETS:
        case IX_CC_POS_TX_DROP_PACKETS:
        case IX_CC_POS_TX_BYTES:
        case IX_CC_POS_TX_DROP_BYTES:
        case IX_CC_POS_TX_ALL_COUNTERS:
            /*
             * IX_CC_POS_TX_PACKETS is the first POS counter in the enum
             * definition.  Use it to derive the counter number.
             */
            counter = arg_Entity - IX_CC_POS_TX_PACKETS;
            break;

        case IX_CC_ATM_TX_PACKETS_VCQ:
        case IX_CC_ATM_TX_CELLS_VCQ:
        case IX_CC_ATM_TX_PACKETS_PORT:
        case IX_CC_ATM_TX_CELLS_PORT:
        case IX_CC_ATM_TX_ALL_COUNTERS_VCQ:
        case IX_CC_ATM_TX_ALL_COUNTERS_PORT:
            return(IX_ERROR_WARNING(IX_CC_ERROR_UNIMPL, 
                                    ("ATM counters have not been implemented in current release.")));
            break;
        default:
            return(IX_ERROR_WARNING(IX_CC_ERROR_RANGE, 
                                    ("Invalid statistics counter type.")));
            break;
    }

    /* Get the component control block pointer */
    pCcContext = (ix_cc_atm_pos_tx_ctl_blk *)arg_pContext;

    /* Get the channel number based on the port id in arg_Index */
    IX_ERROR_CR(_ix_cc_atm_pos_get_channel_number(pCcContext,
                                                  arg_Index, 
                                                  &chNum));

    /*
     * Calculate the base address of the counter handle array that holds
     * counter handles pertinent to this port based on channel number 
     * derived from arg_Index.
     */
    pH64Bit = pCcContext->pPosCounterHandlers64Bit + (chNum * POS_TX_NUM_COUNTERS_PER_CH);

    /*
     * Determine how many counter are to be read based on the entity type. 
     * (Note: This portion of code needs to modifiled when ATM feature is
     * added later in PR6. For now it will be OK for POS.)
     */
    if (arg_Entity == IX_CC_POS_TX_ALL_COUNTERS)
    { 
        /* All counters of the specified port are to be returned */
        numCounters = POS_TX_NUM_COUNTERS_PER_PORT;
    }
    else
    {
        /* Only the specified counter is to be returned */
        numCounters = 1;

        /*
         * set the handler pointer to the associated location from the 
         * base using the counter number indicated by the arg_entity.
         */
        pH64Bit = pH64Bit + counter;
    }

    /* Call RM to read the counter value */
    for (i = 0; i < numCounters; i++)
    {
        /* Call RM to retrieve the 64-bit counter value */
        IX_ERROR_CRT(ix_rm_counter_64bit_get_value(*(pH64Bit + i),
                                                   (ix_uint64 *)(arg_pBuffer->pDataBuffer) + i),
                    IX_CC_ATM_POS_TX_ERROR_RM_GET_64BIT_COUNTER, 
                    IX_ERROR_LEVEL_WARNING);
    }

    return(IX_SUCCESS);

} /* ix_cc_csix_tx_get_statistics_info */




/**
 * NAME: ix_cc_atm_pos_tx_get_interface_state
 * 
 * DESCRIPTION: This function retrieves the link status of a ATM or POS
 *              interface port.
 * 
 * @Param: arg_PortId - IN: port id
 * @Param: arg_pIfState - OUT: location where the interface state will be 
 *                            stored.
 * @Param: arg_pContext - IN: pointer to the component context allocated
 *                            in the component's init function. 
 * 
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_cc_atm_pos_tx_get_interface_state(
    ix_uint32 arg_PortId, 
    ix_cc_atm_pos_tx_if_state *arg_pIfState,
    void *arg_pContext)
{

    ix_uint32 chNum;
    ix_cc_atm_pos_tx_ctl_blk *pCcContext;



    /* Check any NULL pointer argument */
    if (NULL == arg_pIfState)
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("NULL ix_cc_atm_pos_tx_if_state argument")));
    }

    if (NULL == arg_pContext)
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("NULL context argument")));
    }

    /* Get the component control block pointer */
    pCcContext = (ix_cc_atm_pos_tx_ctl_blk *)arg_pContext;

    /* Get the channel number based on the port id */
    IX_ERROR_CR(_ix_cc_atm_pos_get_channel_number(pCcContext,
                                                  arg_PortId, 
                                                  &chNum));

    /*
     * Enter critical section.
     */
    IX_ERROR_CRT(ix_ossl_mutex_lock(pCcContext->mid, 
                                    IX_OSSL_WAIT_FOREVER),
                 IX_CC_ATM_POS_TX_ERROR_OSSL_MUTEX_LOCK, 
                 IX_ERROR_LEVEL_WARNING);

#if (defined IX_PLATFORM_2401) || (defined IX_PLATFORM_2801)

    /* Platform specific
     * IXDP2401 and IXDP2801 platforms have a different framer driver
     */
#if (_IX_BOARD_TYPE_ == _IX_IXDP2401_) 
    pCcContext->portState[chNum] = PosAtm_GetLinkState(pCcContext->posFd,
                                                       chNum);
#elif (_IX_BOARD_TYPE_ == _IX_IXDP2801_)
    if (IX_POS_DB1_MASK & (1 << chNum))
    {
        pCcContext->portState[chNum] =
            PosAtm_GetLinkState(pCcContext->posFd[0], chNum);
    }
    else /* IX_POS_DB2_MASK & (1 << chNum) */
    {
        pCcContext->portState[chNum] =
            PosAtm_GetLinkState(pCcContext->posFd[1], chNum-4);
    }
#else
    /**
     * Poll the framer chip so it will call our registered callback to 
     * update the internal state variable.
     */
    IxfApiChipIsr(pCcContext->pChipData);
#endif
#else /* IX_PLATFORM_2401 || IX_PLATFORM_2801 */
    /**
     * Poll the framer chip so it will call our registered callback to 
     * update the internal state variable.
     */
    IxfApiChipIsr(pCcContext->pChipData);
#endif /* IX_PLATFORM_2401 || IX_PLATFORM_2801 */
    
    /* Exit critical section */
    ix_ossl_mutex_unlock(pCcContext->mid);

    *arg_pIfState = pCcContext->portState[chNum];

    return(IX_SUCCESS);

} /* ix_cc_atm_pos_tx_get_interface_state */




/**
 * NAME: ix_cc_atm_pos_tx_set_property
 * 
 * DESCRIPTION: This function sets the link state of an ATM or POS
 *              interface port.
 * 
 * @Param: arg_PortId - IN: port id
 * @Param: arg_pProperty - IN: pointer to the ix_cc_properties structure
 *                             that contains the link state to be set.
 * @Param: arg_pContext - IN: pointer to the component context allocated
 *                            in the component's init function. 
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_cc_atm_pos_tx_set_property(
    ix_uint32 arg_PropId,
    ix_cc_properties *arg_pProperty,
    void *arg_pContext)
{

    ix_uint32 chNum;
    ix_cc_atm_pos_tx_ctl_blk *pCcContext;



    /* Check any NULL pointer argument */
    if (NULL == arg_pProperty)
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("NULL ix_cc_properties argument")));
    }

    /* Check any NULL pointer argument */
    if (NULL == arg_pContext)
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("NULL context argument")));
    }

    /* Check the property type. Currently we support only the IF_STATUS */
    if ((arg_PropId & IX_CC_SET_PROPERTY_PHYSICAL_IF_STATUS) != IX_CC_SET_PROPERTY_PHYSICAL_IF_STATUS)
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_UNIMPL,
                                ("Indicated property is not supported.")));
    }

    /* Get the component control block pointer */
    pCcContext = (ix_cc_atm_pos_tx_ctl_blk *)arg_pContext;

    /* Get the channel number based on the port id */
    IX_ERROR_CR(_ix_cc_atm_pos_get_channel_number(pCcContext,
                                                  arg_pProperty->port_id, 
                                                  &chNum));

    /*
     * Enter critical section.
     */
    IX_ERROR_CRT(ix_ossl_mutex_lock(pCcContext->mid, IX_OSSL_WAIT_FOREVER),
                 IX_CC_ATM_POS_TX_ERROR_OSSL_MUTEX_LOCK, 
                 IX_ERROR_LEVEL_WARNING);

    /* Call IXF6048 API to set the state of the port */
    if (arg_pProperty->physical_if_state == IX_CC_PHYSICAL_IF_STATUS_UP)
    {
#if (defined IX_PLATFORM_2401) || (defined IX_PLATFORM_2801)

    /* Platform specific
     * IXDP2401 platform has a different framer driver
     */
#if (_IX_BOARD_TYPE_ == _IX_IXDP2401_)
        if (PosAtm_SetLinkState(pCcContext->posFd,
                                chNum, POS_ATM_STATE_UP) < 0)
#elif (_IX_BOARD_TYPE_ == _IX_IXDP2801_)
        if (
            ((IX_POS_DB1_MASK & (1 << chNum)) &&
             (PosAtm_SetLinkState(pCcContext->posFd[0],
                                  chNum, POS_ATM_STATE_UP) < 0))
            ||
            ((IX_POS_DB2_MASK & (1 << chNum)) &&
             (PosAtm_SetLinkState(pCcContext->posFd[1],
                                  chNum-4, POS_ATM_STATE_UP) < 0))
            )
#else
        if (ixf6048_SetChanOnline(pCcContext->pChipData, chNum) != bb_NO_ERROR)
#endif
#else /* IX_PLATFORM_2401 || IX_PLATFORM_2801 */
        if (ixf6048_SetChanOnline(pCcContext->pChipData, chNum) != bb_NO_ERROR)
#endif /* IX_PLATFORM_2401 || IX_PLATFORM_2801 */
        {
            /* Exit the critical section */
            ix_ossl_mutex_unlock(pCcContext->mid);

            return(IX_ERROR_WARNING(IX_CC_ATM_POS_TX_ERROR_FRAMER_SET_CHIP_ONLINE,
                                    ("Can't set channel to UP state.")));
        }

        /* Update the channel status flag */
        pCcContext->portState[chNum] = IX_CC_ATM_POS_TX_IF_STATE_UP;
    }
    else if (arg_pProperty->physical_if_state == IX_CC_PHYSICAL_IF_STATUS_DOWN)
    {
#if (defined IX_PLATFORM_2401) || (defined IX_PLATFORM_2801)

    /* Platform specific
     * IXDP2401 platform has a different framer driver
     */
#if (_IX_BOARD_TYPE_ == _IX_IXDP2401_)
        if (PosAtm_SetLinkState(pCcContext->posFd,
                                chNum, POS_ATM_STATE_DOWN) < 0)
#elif (_IX_BOARD_TYPE_ == _IX_IXDP2801_)
        if (
            ((IX_POS_DB1_MASK & (1 << chNum)) &&
             (PosAtm_SetLinkState(pCcContext->posFd[0],
                                  chNum, POS_ATM_STATE_DOWN) < 0))
            ||
            ((IX_POS_DB2_MASK & (1 << chNum)) &&
             (PosAtm_SetLinkState(pCcContext->posFd[1],
                                  chNum-4, POS_ATM_STATE_DOWN) < 0))
            )
#else
        if (ixf6048_SetChanOffline(pCcContext->pChipData, chNum) != bb_NO_ERROR)
#endif
#else /* IX_PLATFORM_2401 || IX_PLATFORM_2801 */
        if (ixf6048_SetChanOffline(pCcContext->pChipData, chNum) != bb_NO_ERROR)
#endif /* IX_PLATFORM_2401 || IX_PLATFORM_2801 */
        {
            /* Exit the critical section */
            ix_ossl_mutex_unlock(pCcContext->mid);

            return(IX_ERROR_WARNING(IX_CC_ATM_POS_TX_ERROR_FRAMER_SET_CHIP_OFFLINE,
                                    ("Can't set channel to DOWN state.")));
        }

        /* Update the channel status flag */
        pCcContext->portState[chNum] = IX_CC_ATM_POS_TX_IF_STATE_DOWN;
    }
    else
    {
        /* Exit the critical section */
        ix_ossl_mutex_unlock(pCcContext->mid);

        return(IX_ERROR_WARNING(IX_CC_ATM_POS_TX_ERROR_INVALID_IF_STATE,
                                ("Invalid interface state.")));
    }
  
    /* Exit the critical section */
    ix_ossl_mutex_unlock(pCcContext->mid);

    return(IX_SUCCESS);

} /* ix_cc_atm_pos_tx_set_property */






/**
 * NAME: _ix_cc_atm_pos_tx_get_channel_number
 * 
 * DESCRIPTION: This function retrieves channel number based on the port id.
 * 
 * @Param: arg_pCtlBlk - IN: the component control block structure
 * @Param: arg_PortId - IN: the port id to be search
 * @Param: arg_ChNum - OUT: location where channel num is to be stored.
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
PRIVATE ix_error _ix_cc_atm_pos_get_channel_number(
    ix_cc_atm_pos_tx_ctl_blk *arg_pCtlBlk, 
    ix_uint32 arg_PortId, 
    ix_uint32 *arg_ChNum)
{

    ix_uint32 i;


    for (i = 0; i < arg_pCtlBlk->numChannels; i++)
    {
        if (arg_pCtlBlk->portId[i] == arg_PortId)
        {
            *arg_ChNum = i;
            return(IX_SUCCESS);
        }
    }

    return(IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                            ("Could not find the specifed port id.")));
}
