/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2X00 Network Processor, Release 5
 *
 * = FILENAME
 *      ix_cc_atm_pos_tx_hdlr.c
 *
 * = DESCRIPTION
 *      This file defines various message handler routines.
 *
 * = CHANGE HISTORY
 *      6/10/2002 - Created.
 *
 * ============================================================================
 * $Id: ix_cc_atm_pos_tx_hdlr.c,v 1.21 2002/10/08 20:45:22 ktseng Exp $
 */

/**
 * System defined include files required.
 */


/**
 * User defined include files required.
 */
#include "ix_cc_error.h"

#include "ix_types.h"
#include "ix_error.h"
#include "ix_cc.h"
#include "bindings.h"
#include "ix_cc_properties.h"

#include "ix_rm.h"
#include "ix_ossl.h"

#ifndef IX_EXCLUDE_CCI
#include "ix_cci.h"
#endif

#include "cc/ix_cc_atm_pos_tx.h"
#include "cc/internal/ix_cc_atm_pos_tx_int.h"

#include "cc/ix_cc_msup.h"


/**
 * NAME: ix_cc_atm_pos_tx_msg_handler
 * 
 * DESCRIPTION: This function is the common message handler routine for
 *              ATM/POS Tx CC.
 * 
 * @Param: arg_Msg - IN: buffer handle that contains message-specific 
 *                       information embedded.
 * @Param: arg_UserData - IN: message type. 
 * @Param: arg_pContext - IN: pointer to the component context allocated
 *                            in the component's init function. 
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_cc_atm_pos_tx_msg_handler(
    ix_buffer_handle arg_Msg, 
    ix_uint32 arg_UserData,
    void *arg_pContext)
{
 
    void *msgData;
    ix_error err1 = IX_SUCCESS;
    ix_error err2 = IX_SUCCESS;
    void *pContext;



    /* Check whether arg_pContext is invalid */
    if (NULL == arg_pContext)
    {
        /* Free the ix_buffer (ignore the return code from ix_rm_buffer_free_chain) */
        ix_rm_buffer_free_chain(arg_Msg);

        return(IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("NULL argument")));
    }

    /* Get the message sent by the helper function */
    IX_ERROR_CT(ix_cc_msup_extract_msg(arg_Msg, &msgData, &pContext),
                err1,
                IX_CC_ATM_POS_TX_ERROR_MSUP_EXTRACT_MSG, 
                IX_ERROR_LEVEL_WARNING);

    if (err1 != IX_SUCCESS)
    {
        /* Free the ix_buffer (ignore the return code from ix_rm_buffer_free_chain) */
        ix_rm_buffer_free_chain(arg_Msg);

        return(err1);
    }

    /* Process the message */
    switch(arg_UserData)
    {
        case IX_CC_ATM_POS_TX_MSG_GET_STATISTICS_INFO:
        {
            ix_cc_atm_pos_tx_statistics_info entity;
            ix_uint32 index;
            ix_uint32 numCountersReturned;
            ix_cc_atm_pos_tx_statistics_info_data buffer;

            /* Retrieve embedded information */
            entity = ((ix_cc_atm_pos_tx_statistics_info_context *)msgData)->entity;
            index = ((ix_cc_atm_pos_tx_statistics_info_context *)msgData)->index;

            /* Obtain the number of counters to be returned */
            switch(entity)
            {
                case IX_CC_ATM_TX_ALL_COUNTERS_VCQ:
                    numCountersReturned = ATM_TX_NUM_COUNTERS_PER_VCQ;
                    break;
                case IX_CC_ATM_TX_ALL_COUNTERS_PORT:
                    numCountersReturned = ATM_TX_NUM_COUNTERS_PER_PORT;
                    break;
                case IX_CC_POS_TX_ALL_COUNTERS:
                    numCountersReturned = POS_TX_NUM_COUNTERS_PER_PORT;
                    break;
                default:
                    /* for all single counter */
                    numCountersReturned = 1;
                    break;
            }

            {
                ix_uint64 dataBuffer[numCountersReturned];

                /* Invoke the corresponding library API */
                buffer.dataLength = sizeof(dataBuffer);
                buffer.pDataBuffer = (void *)dataBuffer;
                err2 = ix_cc_atm_pos_tx_get_statistics_info(entity,
                                                            index, 
                                                            &buffer,
                                                            arg_pContext);

                /* Send back the result */
                IX_ERROR_CT(ix_cc_msup_send_reply_msg(pContext,
                                                      (void *)dataBuffer, 
                                                      sizeof(dataBuffer),
                                                      err2),
	                    err1,
                            IX_CC_ATM_POS_TX_ERROR_MSUP_SEND_REPLY, 
                            IX_ERROR_LEVEL_WARNING);
            }
            break;
        }
        case IX_CC_ATM_POS_TX_MSG_GET_INTERFACE_STATE:
        {
            ix_uint32 portId;
            ix_cc_atm_pos_tx_if_state ifState;

            /* Retrieve embedded information */
            portId = ((ix_cc_atm_pos_tx_if_state_context *)msgData)->portId;

            /* Invoke the corresponding library API */
            err2 = ix_cc_atm_pos_tx_get_interface_state(portId,
                                                        &ifState,
                                                        arg_pContext);

            /* Send back the result */
            IX_ERROR_CT(ix_cc_msup_send_reply_msg(pContext,
                                                  (void *)&ifState, 
                                                  sizeof(ix_cc_atm_pos_tx_if_state),
                                                  err2),
	                err1,
                        IX_CC_ATM_POS_TX_ERROR_MSUP_SEND_REPLY, 
                        IX_ERROR_LEVEL_WARNING);
            break;
        }
        default :
            err1 = IX_ERROR_WARNING(IX_CC_ERROR_UNDEFINED_MSG, 
                                    ("Undefined message type"));
            break;
    }

    /* Free the ix_buffer here */
    IX_ERROR_CT(ix_rm_buffer_free_chain(arg_Msg),
                err2,
                IX_CC_ATM_POS_TX_ERROR_RM_FREE_BUF, 
                IX_ERROR_LEVEL_WARNING);

    /* Always return the first error encountered */
    if (err1 != IX_SUCCESS)
    {
        /* Ignore the return code from ix_rm_buffer_free. Just return the first error */
        return(err1);
    }

    return(err2);

} /* ix_cc_atm_pos_tx_msg_handler */




/**
 * NAME: ix_cc_atm_pos_tx_property_msg_handler
 * 
 * DESCRIPTION: This function is the port message handler routine for
 *              ATM/POS Tx CC.
 * 
 * @Param: arg_Msg - IN: buffer handle that contains message-specific 
 *                       information embedded.
 * @Param: arg_UserData - IN: message type. 
 * @Param: arg_pContext - IN: pointer to the component context allocated
 *                            in the component's init function. 
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_cc_atm_pos_tx_property_msg_handler(
    ix_buffer_handle arg_Msg, 
    ix_uint32 arg_UserData,
    void *arg_pContext)
{
 
    ix_cc_properties_msg *ccPropMsg;
    ix_error err1 = IX_SUCCESS;
    ix_error err2 = IX_SUCCESS;
    void *pContext;



    /* Check whether arg_pContext is invalid */
    if (NULL == arg_pContext)
    {
        /* Free the ix_buffer. Ignore the return code from rm */
        ix_rm_buffer_free_chain(arg_Msg);

        return(IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("NULL argument")));
    }

    /* Get the message sent by the helper function */
    IX_ERROR_CT(ix_cc_msup_extract_msg(arg_Msg, (void **)&ccPropMsg, &pContext),
                err1,
                IX_CC_ATM_POS_TX_ERROR_MSUP_EXTRACT_MSG, 
                IX_ERROR_LEVEL_WARNING);

    if (err1 != IX_SUCCESS)
    {
        /* Free the ix_buffer (ignore the return code from ix_rm_buffer_free_chain) */
        ix_rm_buffer_free_chain(arg_Msg);

        return(err1);
    }

    /* Make sure the blade id is our local blade id */
    if (ccPropMsg->propData.blade_id != ((ix_cc_atm_pos_tx_ctl_blk *)arg_pContext)->bladeId)
    {
        /* Free the ix_buffer. Ignore the retunr code from rm */
        ix_rm_buffer_free_chain(arg_Msg);

        return(IX_ERROR_WARNING(IX_CC_ATM_POS_TX_ERROR_INVALID_BLADE_ID,
                                ("Blade id received does not match with local blade id.")));
    }

    /* Process the message */
    switch(arg_UserData)
    {
        case IX_CC_COMMON_MSG_ID_PROP_UPDATE:
        {
            /* Invoke the corresponding library API */
            IX_ERROR_C(ix_cc_atm_pos_tx_set_property(ccPropMsg->propID, 
                                                     &(ccPropMsg->propData),
                                                     arg_pContext),
	               err1);
            break;
        }
        default :
            err1 = IX_ERROR_WARNING(IX_CC_ERROR_UNDEFINED_MSG, 
                                    ("Undefined message type"));
            break;
    }

    /* Free the ix_buffer */
    IX_ERROR_CT(ix_rm_buffer_free_chain(arg_Msg),
                err2,
                IX_CC_ATM_POS_TX_ERROR_RM_FREE_BUF, 
                IX_ERROR_LEVEL_WARNING);

    /* Always return the first error encountered */
    if (err1 != IX_SUCCESS)
    {
        /* Ignore the return code from ix_rm_buffer_free_chain. Just return the first error */
        return(err1);
    }

    return(err2);

} /* ix_cc_atm_pos_tx_property_msg_handler */

