/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2X00 Network Processor, Release 5
 *
 * = FILENAME
 *      ix_cc_atm_pos_tx.c
 *
 * = DESCRIPTION
 *      This file contains messaging interface APIs and 
 *      related internal async callback functions of ATM/POS Tx CC.
 *
 * = CHANGE HISTORY
 *      6/10/2002 - Created.
 *
 * ============================================================================
 * $Id: ix_cc_atm_pos_tx_mhlp.c,v 1.19 2002/10/08 20:45:22 ktseng Exp $
 */

/**
 * System defined include files required.
 */


/**
 * User defined include files required.
 */
#include "ix_cc_error.h"

#include "ix_types.h"
#include "ix_error.h"
#include "ix_cc.h"
#include "bindings.h"
#include "ix_cc_properties.h"

#include "ix_rm.h"
#include "ix_ossl.h"

#ifndef IX_EXCLUDE_CCI
#include "ix_cci.h"
#endif

#include "cc/ix_cc_atm_pos_tx.h"
#include "cc/internal/ix_cc_atm_pos_tx_int.h"

#include "cc/ix_cc_msup.h"


/**
 * Variable declarations global to this file only.  Extern are followed
 * by static variables.
 */
extern ix_cc_atm_pos_tx_ctl_blk *g_pCcAtmPosTxCtlBlk; /* control block */



/**
 * Pre-processor symbols and macros used in this file.
 */



/**
 * Extern function prototypes and variables
 */



/**
 * Static function prototypes
 */
ix_error _ix_cc_atm_pos_tx_icb_get_statistics_info(ix_error, void *, void *, ix_uint32);
ix_error _ix_cc_atm_pos_tx_icb_get_interface_state(ix_error, void *, void *, ix_uint32);



/**
 * function definitions.
 */



/**
 * NAME: ix_cc_atm_pos_tx_async_get_statistics_info
 * 
 * DESCRIPTION: This function retrieves a statistics information of 
 *              the ATM/POS Tx CC.
 * 
 * @Param: arg_pMgInfoContext - IN: pointer to the management info 
 *                                  context strcuture.
 * @Param: arg_Callback - IN: user callback.
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_cc_atm_pos_tx_async_get_statistics_info(
    ix_cc_atm_pos_tx_statistics_info_context *arg_pMgInfoCtx, 
    ix_cc_atm_pos_tx_cb_get_statistics_info arg_Callback)
{

    ix_cc_atm_pos_tx_cb_ctx_entry *entry;
    ix_error err = IX_SUCCESS;



    /* Check NULL pointer argument */
    if (arg_pMgInfoCtx == NULL)
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("NULL management context argument")));
    }

    /* Check NULL pointer argument */
    if (arg_Callback == NULL)
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("NULL callback argument")));
    }

    /* Check the range of the counter type stored in arg_Entity */
    if (arg_pMgInfoCtx->entity >= IX_CC_ATM_POS_TX_COUNTER_LAST)
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_RANGE, 
                                ("Invalid statistics counter type.")));
    }

    /*
     * Enter critical section.  It is possible that the mutex can not 
     * be obtained if component has been destroyed and the mutex was 
     * freed by the fini function.)
     */
    IX_ERROR_CRT(ix_ossl_mutex_lock(g_pCcAtmPosTxCtlBlk->mid, 
                                    IX_OSSL_WAIT_FOREVER),
                 IX_CC_ATM_POS_TX_ERROR_OSSL_MUTEX_LOCK, 
                 IX_ERROR_LEVEL_WARNING);

    /*
     * Get a free callback context entry.
     */
    IX_CC_ATM_POS_TX_ALLOC_ENTRY(g_pCcAtmPosTxCtlBlk->pCbCtxTblGetStatInfo->pTable, 
                              g_pCcAtmPosTxCtlBlk->pCbCtxTblGetStatInfo->index, 
                              ATM_POS_TX_NUM_ENTRIES_CB_CTX_TBL_GET_STAT,
                              ix_cc_atm_pos_tx_cb_ctx_entry,
                              entry);

    if (entry == NULL)
    {
        /* Exit critical section */
        ix_ossl_mutex_unlock(g_pCcAtmPosTxCtlBlk->mid);

        return(IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
                                ("Out of memory for callback context entry")));
    }

    /* Store the user callback and context into the entry */
    entry->pUserCtx = arg_pMgInfoCtx->pUserContext;
    entry->pUserCb = (void *)arg_Callback;;

    /* Call msup to send the message */
    IX_ERROR_CT(ix_cc_msup_send_async_msg(IX_CC_ATM_POS_TX_MSG_INPUT,
                                          _ix_cc_atm_pos_tx_icb_get_statistics_info,
                                          (void *)entry,
                                          IX_CC_ATM_POS_TX_MSG_GET_STATISTICS_INFO,
                                          (void *)arg_pMgInfoCtx,
                                          sizeof(ix_cc_atm_pos_tx_statistics_info_context)),
                err,
                IX_CC_ATM_POS_TX_ERROR_MSUP_SEND_MSG, 
                IX_ERROR_LEVEL_WARNING);

    if (err != IX_SUCCESS)
    {
        IX_CC_ATM_POS_TX_FREE_ENTRY(entry);
    }

    /* Exit critical section */
    ix_ossl_mutex_unlock(g_pCcAtmPosTxCtlBlk->mid);

    return(err);

} /* ix_cc_atm_pos_tx_async_get_statistics_info */





/**
 * NAME: ix_cc_atm_pos_tx_async_get_interface_state
 * 
 * DESCRIPTION: This function retrieves the interface state of a port.
 * 
 * @Param: arg_pStateContext - IN: pointer to the state context structure.
 * @Param: arg_Callback - IN: user callback.
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_cc_atm_pos_tx_async_get_interface_state(
    ix_cc_atm_pos_tx_if_state_context *arg_pStateContext, 
    ix_cc_atm_pos_tx_cb_get_interface_state arg_Callback)
{

    ix_cc_atm_pos_tx_cb_ctx_entry *entry;
    ix_error err = IX_SUCCESS;



    /* Check NULL pointer argument */
    if (arg_pStateContext == NULL)
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("NULL interface state context argument")));
    }

    /* Check NULL pointer argument */
    if (arg_Callback == NULL)
    {
        return(IX_ERROR_WARNING(IX_CC_ERROR_NULL,
                                ("NULL callback argument")));
    }

    /*
     * Enter critical section.  It is possible that the mutex can not 
     * be obtained if component has been destroyed and the mutex was 
     * freed by the fini function.)
     */
    IX_ERROR_CRT(ix_ossl_mutex_lock(g_pCcAtmPosTxCtlBlk->mid, 
                                    IX_OSSL_WAIT_FOREVER),
                 IX_CC_ATM_POS_TX_ERROR_OSSL_MUTEX_LOCK, 
                 IX_ERROR_LEVEL_WARNING);

    /*
     * Get a free callback context entry.
     */
    IX_CC_ATM_POS_TX_ALLOC_ENTRY(g_pCcAtmPosTxCtlBlk->pCbCtxTblGetIfState->pTable, 
                              g_pCcAtmPosTxCtlBlk->pCbCtxTblGetIfState->index, 
                              ATM_POS_TX_NUM_ENTRIES_CB_CTX_TBL_GET_IF,
                              ix_cc_atm_pos_tx_cb_ctx_entry,
                              entry);

    if (entry == NULL)
    {
        /* Exit critical section */
        ix_ossl_mutex_unlock(g_pCcAtmPosTxCtlBlk->mid);

        return(IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
                                ("Out of memory for callback context entry")));
    }

    /* Store the user callback and context into the entry */
    entry->pUserCtx = arg_pStateContext->pUserContext;
    entry->pUserCb = (void *)arg_Callback;

    /* Call msup to send the message */
    IX_ERROR_CT(ix_cc_msup_send_async_msg(IX_CC_ATM_POS_TX_MSG_INPUT,
                                          _ix_cc_atm_pos_tx_icb_get_interface_state,
                                          (void *)entry,
                                          IX_CC_ATM_POS_TX_MSG_GET_INTERFACE_STATE,
                                          (void *)arg_pStateContext,
                                          sizeof(ix_cc_atm_pos_tx_if_state_context)),
                err,
                IX_CC_ATM_POS_TX_ERROR_MSUP_SEND_MSG, 
                IX_ERROR_LEVEL_WARNING);

    if (err != IX_SUCCESS)
    {
        IX_CC_ATM_POS_TX_FREE_ENTRY(entry);
    }

    /* Exit critical section */
    ix_ossl_mutex_unlock(g_pCcAtmPosTxCtlBlk->mid);

    return(err);

} /* ix_cc_atm_pos_tx_async_get_interface_state */







/**********************
 * Internal functions *
 **********************/


/**
 * NAME: _ix_cc_atm_pos_tx_icb_get_statistics_info
 * 
 * DESCRIPTION: This function is an internal callback for the async 
 *              messaging API, ix_cc_atm_pos_tx_async_get_statistics_info().
 * 
 * @Param: arg_Result - IN
 * @Param: arg_pContext - IN
 * @Param: arg_Msg - IN
 * @Param: arg_MsgLen - IN
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error _ix_cc_atm_pos_tx_icb_get_statistics_info(
    ix_error arg_Result,
    void *arg_pContext,
    void *arg_Msg,
    ix_uint32 arg_MsgLen)
{

    ix_cc_atm_pos_tx_cb_get_statistics_info userCb;
    void *userCtx;



    /* Retrieve the user callback and context from the context pointer */
    userCb = ((ix_cc_atm_pos_tx_cb_ctx_entry *)arg_pContext)->pUserCb;
    userCtx = ((ix_cc_atm_pos_tx_cb_ctx_entry *)arg_pContext)->pUserCtx;

    /*
     * Enter critical section.  It is possible that the mutex can not 
     * be obtained if component has been destroyed and the mutex was 
     * freed by the fini function.
     */
    IX_ERROR_CRT(ix_ossl_mutex_lock(g_pCcAtmPosTxCtlBlk->mid, 
                                    IX_OSSL_WAIT_FOREVER),
                 IX_CC_ATM_POS_TX_ERROR_OSSL_MUTEX_LOCK, 
                 IX_ERROR_LEVEL_WARNING);

    /* Free the callback context entry */
    IX_CC_ATM_POS_TX_FREE_ENTRY(((ix_cc_atm_pos_tx_cb_ctx_entry *)arg_pContext));

    /* Exit the critical section */
    ix_ossl_mutex_unlock(g_pCcAtmPosTxCtlBlk->mid);

    /* Invoke the user callback to pass back the result and data */
    IX_ERROR_CRT((userCb)(arg_Result, 
                          userCtx, 
                          (ix_uint64 *)arg_Msg,
                          arg_MsgLen),
                 IX_CC_ATM_POS_TX_ERROR_USER_CB, 
                 IX_ERROR_LEVEL_WARNING);

    return(IX_SUCCESS);

} /* _ix_cc_atm_pos_tx_icb_get_statistics_info */



/**
 * NAME: _ix_cc_atm_pos_tx_icb_get_interface_state
 * 
 * DESCRIPTION: This function is an internal callback for the async 
 *              messaging API, ix_cc_atm_pos_tx_async_get_interface_state().
 * 
 * @Param: arg_Result - IN
 * @Param: arg_pContext - IN
 * @Param: arg_Msg - IN
 * @Param: arg_MsgLen - IN
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error _ix_cc_atm_pos_tx_icb_get_interface_state(
    ix_error arg_Result,
    void *arg_pContext,
    void *arg_Msg,
    ix_uint32 arg_MsgLen)
{

    ix_cc_atm_pos_tx_cb_get_interface_state userCb;
    void *userCtx;



    /* Retrieve the user callback and context from the context pointer */
    userCb = ((ix_cc_atm_pos_tx_cb_ctx_entry *)arg_pContext)->pUserCb;
    userCtx = ((ix_cc_atm_pos_tx_cb_ctx_entry *)arg_pContext)->pUserCtx;

    /*
     * Enter critical section.  It is possible that the mutex can not 
     * be obtained if component has been destroyed and the mutex was 
     * freed by the fini function.
     */
    IX_ERROR_CRT(ix_ossl_mutex_lock(g_pCcAtmPosTxCtlBlk->mid, 
                                    IX_OSSL_WAIT_FOREVER),
                 IX_CC_ATM_POS_TX_ERROR_OSSL_MUTEX_LOCK, 
                 IX_ERROR_LEVEL_WARNING);

    /* Free the callback context entry */
    IX_CC_ATM_POS_TX_FREE_ENTRY(((ix_cc_atm_pos_tx_cb_ctx_entry *)arg_pContext));

    /* Exit the critical section */
    ix_ossl_mutex_unlock(g_pCcAtmPosTxCtlBlk->mid);

    /* Invoke the user callback to pass back the result and data */
    IX_ERROR_CRT((userCb)(arg_Result, 
                          userCtx, 
                          (ix_cc_atm_pos_tx_if_state *)arg_Msg),
                 IX_CC_ATM_POS_TX_ERROR_USER_CB, 
                 IX_ERROR_LEVEL_WARNING);

    return(IX_SUCCESS);

} /* _ix_cc_atm_pos_tx_icb_get_statistics_info */

