#ifndef	_AAL5_TX_CRC_
#define	_AAL5_TX_CRC_

/*******************************************************************************
                             Intel Proprietary

 Copyright (c) 1998-2002 By Intel Corporation.  All rights reserved.
 No part of this program or publication may be reproduced, transmited,
 transcribed, stored in a retrieval system, or translated into any language
 or computer language in any form or by any means, electronic, mechanical,
 magnetic, optical, chemical, manual, or otherwise, without the prior
 written permission of:
                         Intel Corporation
                         2200 Mission College Blvd.
                         Santa Clara, CA  95052-8119
*******************************************************************************/

/*                                                                   
 *                                                                   
 *      File Name: aal5_tx_crc.uc                                         
 *                                                                   
 *      Description: This file contains crc related macros.
 *
 */                                                                   

#include <aal5_tx.h>
#include <aal5_tx_regs.uc>

//------------------------------------------------------------------
// _aal5_tx_prepare_cell_in_lm()
//
//    Description: Prepare a 48 byte cell in a thread specific
//    location in local memory.
//
//    Parameters:
//        Inputs:
//        in_curr_payload_addr_offset:
//            Offset from where the DRAM payload starts (0-7).
//        in_cell_payload_in_curr_buf:
//            Cell payload in current buffer.
//        in_cell_payload_in_next_buf:
//            Cell payload in next buffer.
//        in_payload_reg_base_addr_lw:
//             Absolute address of $$payload0.
//        in_next_payload_reg_base_lw:
//             Absolute address of $$payloadx which contains
//             payload from next buffer. 
//
//------------------------------------------------------------------
#macro _aal5_tx_prepare_cell_in_lm( \
			in_curr_payload_addr_offset, \
			in_cell_payload_in_curr_buf, \
			in_cell_payload_in_next_buf, \
			in_payload_reg_base_addr_lw, \
			in_next_payload_reg_base_lw)
.begin

	.reg lm_addr_base

	; Depending on the current thread, set the cell payload base in LM.
	alu_shf_left(lm_addr_base, CELL_BASE_LM, +, thread_num, 6)

	; Needs 3 nops or eqv. instructions before *l$index0 is used.
	local_csr_wr[active_lm_addr_0, lm_addr_base]

	.begin
	.reg lm_addr_base_plus_32

	; Set *l$index1 at an offset of 32 bytes from *l$index0.
	alu[lm_addr_base_plus_32, lm_addr_base, +, 32]

	; Needs 3 nops or eqv. instructions before *l$index1 is used.
	local_csr_wr[active_lm_addr_1, lm_addr_base_plus_32]

	.end // lm_addr_base_plus_32


	.reg byte_offset

	.reg index_val

	; Set t_index to point to first $$payload with data.
	alu[index_val, --, B, in_curr_payload_addr_offset, >>2]
	alu[index_val, index_val, +, in_payload_reg_base_addr_lw]

	; Set byte_index to start of data within the $$payload
	; pointed by t_index.
	alu[byte_offset, 0x3, AND, in_curr_payload_addr_offset]
	alu[index_val, byte_offset, OR, index_val, <<2]

	; Needs 3 nops or eqv. instructions before byte_align_be
	; or *$$index is used.
	local_csr_wr[t_index_byte_index, index_val]


	.reg jmp_tbl_index
	.reg cell_payload_rem
	.reg last_lw_of_payload

	alu[jmp_tbl_index, --, B, in_cell_payload_in_curr_buf, >>2]

	_aal5_tx_byte_align_cell_payload(
		last_lw_of_payload,
		jmp_tbl_index,
		CURR, 
		--, 
		in_cell_payload_in_curr_buf,
		byte_offset)

	; if (in_cell_payload_in_next_buf > 0),
	; current cell is split across buffers.
	alu[--, in_cell_payload_in_next_buf, -, 0]
	ble[prepare_rest_of_cell#]

	 // Unfilled defer slot.

	.reg first_lw_of_cell_payload_in_next_buf

	.reg new_cell_payload_in_next_buf

	alu[new_cell_payload_in_next_buf, \
		in_cell_payload_in_next_buf, +, cell_payload_rem]

	_aal5_tx_align_first_lw_of_cell_payload_in_next_buf(
		first_lw_of_cell_payload_in_next_buf, 
		byte_offset,
		jmp_tbl_index,
		last_lw_of_payload,
		new_cell_payload_in_next_buf,
		in_next_payload_reg_base_lw)

	_aal5_tx_byte_align_cell_payload(
		last_lw_of_payload,
		jmp_tbl_index,
		NEXT, 
		first_lw_of_cell_payload_in_next_buf,
		new_cell_payload_in_next_buf,
		byte_offset)

prepare_rest_of_cell#:

	_aal5_tx_write_last_lw_of_payload(jmp_tbl_index, last_lw_of_payload)


	; Move CONST_BASE_LM to gpr lm_addr_base to reset
	; *l$index1 to its original value.
	move(lm_addr_base, CONST_BASE_LM)

	; Reset *l$index1 to its original value.
	; Needs 3 nops or eqv. instructions before *l$index1 is used.
	local_csr_wr[active_lm_addr_1, lm_addr_base]

	; Reset *l$index0 to its original value.
	; Needs 3 nops or eqv. instructions before *l$index0 is used.
	cam_lookup[--, vcq], lm_addr0[0]

.end
#endm


//------------------------------------------------------------------
// _aal5_tx_byte_align_cell_payload()
//
//    Description: Byte align cell payload in local memory.
//
//    Parameters:
//        Outputs:
//        out_last_lw_of_payload:
//            Last lw of payload that may need padding or aligning
//            with first lw of payload in next buffer.
//
//         I/O:
//        io_jmp_tbl_index:
//            Index of jump table. Set to index of next jump table
//            after use.
//
//        Inputs:
//        BUF_TYPE:
//            CURR indicates 'cell_payload' belongs to current buffer.
//            NEXT indicates 'cell_payload' belongs to next buffer.
//        in_first_lw_of_cell_payload_in_next_buf:
//            Valid only when BUF_TYPE is NEXT. This is the first lw
//             of cell payload in next buffer.
//        in_cell_payload:
//            Cell payload of current or next buffer.
//        in_byte_offset:
//            Offset of start of payload (0-3).
//
//------------------------------------------------------------------
#macro _aal5_tx_byte_align_cell_payload( \
			out_last_lw_of_payload, \
			io_jmp_tbl_index, \
			BUF_TYPE, \
			in_first_lw_of_cell_payload_in_next_buf, \
			in_cell_payload, \
			in_byte_offset)

.begin

	.reg cell_payload_floor

	jump[io_jmp_tbl_index, jmp_base#],
		targets[br_jmp_0#, br_jmp_1#, br_jmp_2#,  br_jmp_3#, \
			    br_jmp_4#, br_jmp_5#, br_jmp_6#,  br_jmp_7#, \
				br_jmp_8#, br_jmp_9#, br_jmp_10#, br_jmp_11#,\
				br_jmp_12#],defer[3]

	; Prepare jmp_tbl_index for the next jump table.
	alu[cell_payload_floor, -- , B, io_jmp_tbl_index, <<2]
	alu[cell_payload_rem, in_cell_payload, - , cell_payload_floor]
	alu[io_jmp_tbl_index, in_byte_offset, OR, cell_payload_rem, <<2]

jmp_base#:

#define_eval i 0		
#while (i <= 12)		
br_jmp_/**/i#:		
	br[jmp_/**/i#]	
	#define_eval i (i+1)	
#endloop
#undef i

#define_eval i	0
#while (i <= 12)		
		
jmp_/**/i#:

	#define			lm_ptr				*l$index0		
	#define_eval	lm_offset			0
	#define_eval	lm_offset_global	0

	#if (i != 0)

		#if (streq('BUF_TYPE', 'CURR'))
		byte_align_be[--, *$$index++]
		#else // streq('BUF_TYPE' == 'NEXT')
		byte_align_be[--, in_first_lw_of_cell_payload_in_next_buf]
		#endif // streq('BUF_TYPE')

		#if (streq('BUF_TYPE', 'CURR'))

		#repeat (i - 1)
		byte_align_be[lm_ptr[lm_offset], *$$index++]
		_aal5_tx_inc_lm_offset()
		#endloop	

		byte_align_be[lm_ptr[lm_offset], *$$index]		
		_aal5_tx_inc_lm_offset()


		; Set the cell payload base in LM to start of new LM lw.
		alu_shf_left(lm_addr_base, lm_addr_base, +, lm_offset_global, 2)

		br[end_jmp#], defer[2]

		; Needs 3 nops or eqv. instructions before *l$index0 is used.
		local_csr_wr[active_lm_addr_0, lm_addr_base]

		; Store the last lw of payload in a gpr for macro
		; prepare_rest_of_cell()
		; Also increment *$$index to point to the next lw.
		alu[out_last_lw_of_payload, --, B, *$$index++]


		#else // streq('BUF_TYPE' == 'NEXT')


		#repeat (i - 1)
		byte_align_be[*l$index0, *$$index]
		alu[--, --, B, *l$index0++]
		byte_align_be[--, *$$index++]
		#endloop	
				
		byte_align_be[*l$index0, *$$index]

		br[end_jmp#], defer[2]

		alu[--, --, B, *l$index0++]

		; Store the last lw of payload in a gpr for macro
		; prepare_rest_of_cell()
		; Also increment *$$index to point to the next lw.
		alu[out_last_lw_of_payload, --, B, *$$index++]

		#endif // streq('BUF_TYPE')

	#else // i==0

		br[end_jmp#], defer[1]

		// Unfilled defer slot.

		#if (streq('BUF_TYPE', 'CURR'))

		; Store the last lw of payload in a gpr for macro
		; prepare_rest_of_cell()
		; Also increment *$$index to point to the next lw.
		alu[out_last_lw_of_payload, --, B, *$$index++]

		#else // streq('BUF_TYPE' == 'NEXT')

		; Store the last lw of payload in a gpr for macro
		; prepare_rest_of_cell()
		alu[out_last_lw_of_payload, --, B, \
			in_first_lw_of_cell_payload_in_next_buf]

		#endif // streq('BUF_TYPE')

	#endif // i

	#define_eval i (i+1)

	#undef lm_ptr
	#undef lm_offset
	#undef lm_offset_global

#endloop
#undef i

end_jmp#:

.end
#endm


//------------------------------------------------------------------
// _aal5_tx_inc_lm_offset()
//
//    Description: Increment local memory pointer and switch to next
//    local memory pointer if needed. Does not produce any code.
//
//    Parameters: None.
//
//------------------------------------------------------------------
#macro _aal5_tx_inc_lm_offset()

	#define_eval lm_offset_global (lm_offset_global+1)

	#define_eval lm_offset (lm_offset+1)
	#if(lm_offset == 8)
		#undef			lm_ptr
		#define			lm_ptr		*l$index1
		#define_eval	lm_offset	0		
	#endif

#endm


//------------------------------------------------------------------
// _aal5_tx_align_first_lw_of_cell_payload_in_next_buf()
//
//    Description: Align first lw of cell payload in next buffer.
//
//    Parameters:
//        Outputs:
//        out_first_lw_of_cell_payload_in_next_buf:
//            This is the first lw of cell payload in next buffer.
//        out_byte_offset:
//            Offset of start of payload (0-3).
//
//         I/O:
//        io_jmp_tbl_index:
//            Index of jump table. Set to index of next jump table
//            after use.
//
//        Inputs:
//        in_last_lw_of_payload:
//            Last lw of payload that may need aligning
//            with first lw of payload in next buffer.
//        in_cell_payload:
//            Cell payload of current or next buffer.
//        in_next_payload_reg_base_lw:
//             Absolute address of $$payloadx which contains
//             payload from next buffer. 
//
//------------------------------------------------------------------
#macro _aal5_tx_align_first_lw_of_cell_payload_in_next_buf( \
			out_first_lw_of_cell_payload_in_next_buf, \
			out_byte_offset, \
			io_jmp_tbl_index, \
			in_last_lw_of_payload, \
			in_cell_payload, \
			in_next_payload_reg_base_lw)
.begin

	.reg index_val

	alu[out_byte_offset, --, B, io_jmp_tbl_index, >>2]

	jump[io_jmp_tbl_index, jmp_base#],
		targets[br_jmp_0#, br_jmp_1#, br_jmp_2#,  br_jmp_3#, \
			    br_jmp_4#, br_jmp_5#, br_jmp_6#,  br_jmp_7#, \
				br_jmp_8#, br_jmp_9#, br_jmp_10#, br_jmp_11#,\
				br_jmp_12#, br_jmp_13#, br_jmp_14#, br_jmp_15#], \
		defer[3]

	; Set byte_index to start of data within the $$payload
	; pointed by t_index.
	; out_byte_offset = 0x3 & (4 - io_jmp_tbl_index>>2)
	alu[out_byte_offset, 4, -, out_byte_offset]
	alu[out_byte_offset, 0x3, AND, out_byte_offset]

	; Set t_index to point to first $$payload with data.
	alu[index_val, out_byte_offset, OR, in_next_payload_reg_base_lw, <<2]

jmp_base#:		

#define_eval i 0		
#while (i <= 15)		
br_jmp_/**/i#:		
	br[jmp_/**/i#]	
	#define_eval i (i+1)	
#endloop
#undef i


#define_eval i 0		
#while (i <= 3)		
jmp_/**/i#:

	; Needs 3 nops or eqv. instructions before byte_align_be
	; or *$$index is used.
	local_csr_wr[t_index_byte_index, index_val]
	nop

	br[end_jmp#], defer[2]

	; Prepare io_jmp_tbl_index for the next jump table.
	alu[io_jmp_tbl_index, --, B, in_cell_payload, >>2]

	; Set first_lw_of_cell_payload_in_next_buf to
	; first lw from next buf.
	ld_field_w_clr[out_first_lw_of_cell_payload_in_next_buf, \
		1111, *$$index++]

	#define_eval i (i+1)	
#endloop
#undef i


#define_eval i 4		
#while (i <= 7)		
jmp_/**/i#:

	#define_eval shift_val	(7-i)*8
	ld_field_w_clr[out_first_lw_of_cell_payload_in_next_buf, \
		0001, in_last_lw_of_payload, >>shift_val]

	br[end_jmp#], defer[2]

	; Prepare io_jmp_tbl_index for the next jump table.
	alu[io_jmp_tbl_index, --, B, in_cell_payload, >>2]

	; Needs 3 nops or eqv. instructions before byte_align_be
	; or *$$index is used.
	local_csr_wr[t_index_byte_index, index_val]

	#define_eval i (i+1)	
	#undef shift_val
#endloop
#undef i


#define_eval i 8		
#while (i <= 10)		
jmp_/**/i#:

	#define_eval shift_val	(10-i)*8
	ld_field_w_clr[out_first_lw_of_cell_payload_in_next_buf, \
		0011, in_last_lw_of_payload, >>shift_val]

	br[end_jmp#], defer[2]

	; Prepare io_jmp_tbl_index for the next jump table.
	alu[io_jmp_tbl_index, --, B, in_cell_payload, >>2]

	; Needs 3 nops or eqv. instructions before byte_align_be
	; or *$$index is used.
	local_csr_wr[t_index_byte_index, index_val]

	#define_eval i (i+1)	
	#undef shift_val
#endloop
#undef i


#define_eval i 11		
jmp_/**/i#:

	ld_field_w_clr[out_first_lw_of_cell_payload_in_next_buf,\
		0010, in_last_lw_of_payload, <<8]
	ld_field[out_first_lw_of_cell_payload_in_next_buf, \
		0001, *$$index, >>24]

	br[end_jmp#], defer[2]

	; Prepare io_jmp_tbl_index for the next jump table.
	alu[io_jmp_tbl_index, --, B, in_cell_payload, >>2]

	; Needs 3 nops or eqv. instructions before byte_align_be
	; or *$$index is used.
	local_csr_wr[t_index_byte_index, index_val]

#undef i


#define_eval i 12		
jmp_/**/i#:

	ld_field_w_clr[out_first_lw_of_cell_payload_in_next_buf, \
		0111, in_last_lw_of_payload, >>8]

	br[end_jmp#], defer[2]

	; Prepare io_jmp_tbl_index for the next jump table.
	alu[io_jmp_tbl_index, --, B, in_cell_payload, >>2]

	; Needs 3 nops or eqv. instructions before byte_align_be
	; or *$$index is used.
	local_csr_wr[t_index_byte_index, index_val]

#undef i


#define_eval i 13
jmp_/**/i#:

	ld_field_w_clr[out_first_lw_of_cell_payload_in_next_buf, \
		0111, in_last_lw_of_payload]

	br[end_jmp#], defer[2]

	; Prepare io_jmp_tbl_index for the next jump table.
	alu[io_jmp_tbl_index, --, B, in_cell_payload, >>2]

	; Needs 3 nops or eqv. instructions before byte_align_be
	; or *$$index is used.
	local_csr_wr[t_index_byte_index, index_val]

#undef i


#define_eval i 14		
jmp_/**/i#:

	ld_field_w_clr[out_first_lw_of_cell_payload_in_next_buf, \
		0110, in_last_lw_of_payload, <<8]
	ld_field[out_first_lw_of_cell_payload_in_next_buf, \
		0001, *$$index, >>24]

	br[end_jmp#], defer[2]

	; Prepare io_jmp_tbl_index for the next jump table.
	alu[io_jmp_tbl_index, --, B, in_cell_payload, >>2]

	; Needs 3 nops or eqv. instructions before byte_align_be
	; or *$$index is used.
	local_csr_wr[t_index_byte_index, index_val]

#undef i


#define_eval i 15		
jmp_/**/i#:

	ld_field_w_clr[out_first_lw_of_cell_payload_in_next_buf, \
		0100, in_last_lw_of_payload, <<16]
	ld_field[out_first_lw_of_cell_payload_in_next_buf, \
		0011, *$$index, >>16]

	br[end_jmp#], defer[2]

	; Prepare io_jmp_tbl_index for the next jump table.
	alu[io_jmp_tbl_index, --, B, in_cell_payload, >>2]

	; Needs 3 nops or eqv. instructions before byte_align_be
	; or *$$index is used.
	local_csr_wr[t_index_byte_index, index_val]

#undef i

end_jmp#:

.end
#endm


//------------------------------------------------------------------
// _aal5_tx_write_last_lw_of_payload()
//
//    Description: Write the last lw of cell payload in the correct
//    byte position(s) in local memory.
//
//    Parameters:
//        Inputs:
//        in_jmp_tbl_index:
//            Index of jump table.
//        in_last_lw_of_payload:
//            Last lw of payload.
//
//------------------------------------------------------------------
#macro _aal5_tx_write_last_lw_of_payload( \
			in_jmp_tbl_index, \
			in_last_lw_of_payload)
.begin

	jump[in_jmp_tbl_index, jmp_base#],
		targets[br_jmp_0#, br_jmp_1#, br_jmp_2#,  br_jmp_3#, \
			    br_jmp_4#, br_jmp_5#, br_jmp_6#,  br_jmp_7#, \
				br_jmp_8#, br_jmp_9#, br_jmp_10#, br_jmp_11#,\
				br_jmp_12#, br_jmp_13#, br_jmp_14#, br_jmp_15#]

	 // Unfilled defer slot.

jmp_base#:		

#define_eval i 0		
#while (i <= 15)		
br_jmp_/**/i#:		
	br[jmp_/**/i#]	
	#define_eval i (i+1)	
#endloop
#undef i


#define_eval i 0		
#while (i <= 3)		
jmp_/**/i#:

	br[end_jmp#], defer[2]

	nop
	nop

	#define_eval i (i+1)	
#endloop
#undef i


#define_eval i 4		
#while (i <= 7)		
jmp_/**/i#:

	#define_eval shift_val	(i-4)*8

	br[end_jmp#], defer[2]

	ld_field[*l$index0, 1000, \
		in_last_lw_of_payload, <<shift_val]
	nop

	#define_eval i (i+1)
	#undef shift_val	
#endloop
#undef i


#define_eval i 8		
#while (i <= 10)		
jmp_/**/i#:

	#define_eval shift_val	(i-8)*8

	br[end_jmp#], defer[2]

	ld_field[*l$index0, 1100, \
		in_last_lw_of_payload, <<shift_val]
	nop

	#define_eval i (i+1)
	#undef shift_val

#endloop
#undef i


#define_eval i 11		
jmp_/**/i#:

	br[end_jmp#], defer[2]

	ld_field[*l$index0, 1000, \
		in_last_lw_of_payload, <<24]

	ld_field[*l$index0, 0100, *$$index, >>8]
#undef i


#define_eval i 12		
jmp_/**/i#:

	br[end_jmp#], defer[2]

	ld_field[*l$index0, 1110, \
		in_last_lw_of_payload, <<0]
	nop
#undef i


#define_eval i 13
jmp_/**/i#:

	br[end_jmp#], defer[2]

	ld_field[*l$index0, 1110, \
		in_last_lw_of_payload, <<8]
	nop
#undef i


#define_eval i 14		
jmp_/**/i#:

	br[end_jmp#], defer[2]

	ld_field[*l$index0, 1100, \
		in_last_lw_of_payload, <<16]

	ld_field[*l$index0, 0010, *$$index, >>16]
#undef i


#define_eval i 15		
jmp_/**/i#:

	br[end_jmp#], defer[2]

	ld_field[*l$index0, 1000, \
		in_last_lw_of_payload, <<24]

	ld_field[*l$index0, 0110, *$$index, >>8]
#undef i

end_jmp#:

.end
#endm


//------------------------------------------------------------------
// _aal5_tx_compute_crc()
//
//    Description: Compute CRC.
//                 Also copy the 52 byte cell into s-xfer-wr regs. 
//
//    Parameters:
//        Inputs:
//        in_cell_payload:
//            Cell payload in bytes of current plus next buffer.
//        in_cell_lw:
//            Actual cell payload.
//
//------------------------------------------------------------------
#macro _aal5_tx_compute_crc(in_cell_payload, in_cell_lw)
.begin

	.reg lm_addr_base
	.reg padding

	; Depending on the current thread, set the cell payload base in LM.
	alu_shf_left(lm_addr_base, CELL_BASE_LM, +, thread_num, 6]
// SCR6027: Added OAM Changes
#ifdef ATM_OAM_ENABLED
	alu[atm_cell_copy_header, --, B, *l$index0[7]]
#endif

	alu[--, in_cell_payload, -, 40]
	bgt[cell_payload_gt_40#], defer[3]

	; Needs 3 nops or eqv. instructions before *l$index0 is used.
	local_csr_wr[active_lm_addr_0, lm_addr_base]

	immed32(padding, AAL5_TX_CELL_PADDING)

	; Copy the cell header into x-fer-write reg.
	; Set the AUU field in the cell header to 1 to indicate end of packet.
	alu[in_cell_lw/**/[WRD_OFFSET_0], 0x2, OR, *l$index0[7]]
	; Note: Since the above alu[] is the 3rd instruction after issuing
	; the local_csr_wr, the original value of *l$index0 is still active.
// SCR6027: Added OAM Changes
#ifdef ATM_OAM_ENABLED
	alu[atm_cell_copy_header, atm_cell_copy_header, OR, 0x02]
	alu[$$atm_cell_copy[0], --, B, atm_cell_copy_header]
#endif

	#define_eval crc_cell_size_lw 10
	#define_eval m 0

	#while (m < crc_cell_size_lw)
		#define_eval m_plus_1	(m+WRD_OFFSET_1)
		; Compute CRC on cell lw.
		crc_be[crc_32, in_cell_lw/**/[m_plus_1]/**/, *l$index0[m]]

		// SCR6027: Added OAM Changes		
		#ifdef ATM_OAM_ENABLED
			#define_eval k		(m + 1)

			#if (m == 0)
				alu[cur_bedc, --, B, *l$index0[m]]
			#else
				alu[cur_bedc, cur_bedc, XOR, *l$index0[m]]
			#endif

			alu[$$atm_cell_copy[k], --, B, *l$index0[m]]

			#undef k
		#endif
		
		; Reset lm location to cell padding value.
		alu[*l$index0[m], --, B, padding]

		#undef m_plus_1
		#define_eval m (m+1)
	#endloop

	#undef m
	#undef crc_cell_size_lw

	crc_be[crc_32, in_cell_lw/**/[WRD_OFFSET_11], txc_packet_size]

	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
		alu[cur_bedc, cur_bedc, XOR, txc_packet_size]
		alu[$$atm_cell_copy[11], --, B, txc_packet_size]

		.set $$atm_cell_copy[12]
	#endif

	.set in_cell_lw/**/[WRD_OFFSET_12]

	br[end_compute_crc#], defer[2]

	; Reset lm location to cell padding value.
	alu[*l$index0[10], --, B, padding]

	; Reset lm location to cell padding value.
	alu[*l$index0[11], --, B, padding]


cell_payload_gt_40#:

	; Copy the cell header into x-fer-write reg.
	alu[in_cell_lw/**/[WRD_OFFSET_0], --, B, *l$index0[7]]
	; Note: Since the above alu[] is the 3rd instruction after issuing
	; the local_csr_wr, the original value of *l$index0 is still active.

// SCR6027: Added OAM Changes
#ifdef ATM_OAM_ENABLED
	alu[$$atm_cell_copy[0], --, B, atm_cell_copy_header]
#endif

	#define_eval crc_cell_size_lw 12
	#define_eval m 0

	#while (m < crc_cell_size_lw)
		#define_eval m_plus_1	(m+WRD_OFFSET_1)

		; Compute CRC on cell lw.
		crc_be[crc_32, in_cell_lw/**/[m_plus_1]/**/, *l$index0[m]]
		// SCR6027: Added OAM Changes	
		#ifdef ATM_OAM_ENABLED

			#define_eval	k	(m + 1)

			#if (m == 0)
				alu[cur_bedc, --, B, *l$index0[m]]
			#else
				alu[cur_bedc, cur_bedc, XOR, *l$index0[m]]
			#endif

			alu[$$atm_cell_copy[k], --, B, *l$index0[m]]

			#undef k

		#endif

		; Reset lm location to cell padding value.
		alu[*l$index0[m], --, B, padding]

		#undef m_plus_1
		#define_eval m (m+1)
	#endloop

	#undef m
	#undef crc_cell_size_lw

end_compute_crc#:

	; Needs 3 nops or eqv. instructions before *l$index0 is used.
	cam_lookup[--, vcq], lm_addr0[0]	
	alu[--, --, b, 0]				     ; This is to avoid optimization problems
.end
#endm


//------------------------------------------------------------------
// _aal5_tx_store_crc()
//
//    Description: Store CRC residue in TXC.
//
//    Parameters:
//        Inputs:
//        in_cell_payload:
//            Cell payload of current plus next buffer.
//        in_cell_lw:
//            Actual cell payload.
//
//------------------------------------------------------------------
#macro _aal5_tx_store_crc(in_cell_payload, in_cell_lw12)
.begin

	alu[--, --, b, 0]				     ; This is to avoid optimization problems
	alu[--, in_cell_payload, -, 40]
	bgt[end_store_crc_residue#], defer[2]

	 // Unfilled defer slot.

	; Reading CRC_Remainder afer 5 nop's since crc_be(]
	; was last used and storing it in TXC:crc.
	local_csr_rd[CRC_Remainder]
	immed[*l$index0[8], 0x0000]	

	; Invert CRC remainder and store it in s-xfer-wr register.
	alu[in_cell_lw12, --, ~B, *l$index0[8]]	
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
		.reg	temp_bedc
		alu[temp_bedc, --, ~B, *l$index0[8]]
		alu[cur_bedc, cur_bedc, XOR, temp_bedc]
		; Store a copy of the CRC longword for OAM processing
		alu[$$atm_cell_copy[12], --, ~B, *l$index0[8]]
	#endif

end_store_crc_residue#:

.end
#endm


//------------------------------------------------------------------
// _aal5_tx_compute_and_store_crc()
//
//    Description: Compute CRC.
//                 Copy the 52 byte cell into s-xfer-wr regs. 
//                 Store CRC residue in TXC.
//
//    Parameters:
//        Inputs:
//        CELL_SIZE: 
//            Cell size in bytes. (can be FORTY_EIGHT or FORTY).
//		  in_curr_payload_addr_offset:
//            Offset from where the DRAM payload starts (0-7).
//        in_cell_lw:
//            Actual cell payload.
//        CRC_DONE:
//            Label to branch to upon completion.
//
//------------------------------------------------------------------
#macro _aal5_tx_compute_and_store_crc(CELL_SIZE, in_curr_payload_addr_offset, \
			in_cell_lw, CRC_DONE)
.begin
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED

	#if (streq('CELL_SIZE', 'FORTY_EIGHT'))
	; Copy the cell header into s-xfer-wr reg.
	alu[$$atm_cell_copy[0], --, B, *l$index0[7]]
	alu[atm_cell_copy_header, --, B, *l$index0[7]]

	#else // streq('CELL_SIZE' == 'FORTY')

	; Copy the cell header into x-fer-write reg.
	; Set the AUU field in the cell header to 1 to indicate end of packet.
	alu[$$atm_cell_copy[0], 0x2, OR, *l$index0[7]]
	alu[atm_cell_copy_header, 0x2, OR, *l$index0[7]]

	#endif // streq('CELL_SIZE')
	#endif // ATM_OAM_ENABLED

	; Jump based on the dram payload offset.
	jump[in_curr_payload_addr_offset, jmp_base#],
		targets[br_jmp_0#, br_jmp_1#, br_jmp_2#,  br_jmp_3#, \
			    br_jmp_4#, br_jmp_5#, br_jmp_6#,  br_jmp_7#],\
		defer[3]

	; Set tcw_payload_offset to the starting offset of payload.
	; This gpr is used in the TCW when validating the 52-byte cell.
	alu[tcw_payload_offset, 0x3, AND, curr_payload_addr_offset]
	alu[tcw_payload_offset, --, B, tcw_payload_offset, <<13]

	#if (streq('CELL_SIZE', 'FORTY_EIGHT'))

	; Copy the cell header into s-xfer-wr reg.
	alu[in_cell_lw/**/[WRD_OFFSET_0], --, B, *l$index0[7]]

	#else // streq('CELL_SIZE' == 'FORTY')

	; Copy the cell header into x-fer-write reg.
	; Set the AUU field in the cell header to 1 to indicate end of packet.
	alu[in_cell_lw/**/[WRD_OFFSET_0], 0x2, OR, *l$index0[7]]

	#endif // streq('CELL_SIZE')

jmp_base#:		

#define_eval i 0		
#while (i <= 7)
br_jmp_/**/i#:		
	br[jmp_/**/i#]	
	#define_eval i (i+1)	
#endloop
#undef i

jmp_0#:

	.begin

	#if (streq('CELL_SIZE', 'FORTY_EIGHT'))
	// 12 lw's of the cell contain aligned payload.
	#define_eval ALIGNED_CELL_BLOCK	11
	#else // streq('CELL_SIZE' == 'FORTY')
	// 10 lw's of the cell contain aligned payload.
	#define_eval ALIGNED_CELL_BLOCK	9
	#endif // streq('CELL_SIZE')

	#define_eval m 0

	; Compute CRC on the aligned payload of the cell and store the
	; payload in $cell_lw.
	#while (m <= ALIGNED_CELL_BLOCK)
		#define_eval m_plus_1	(m+WRD_OFFSET_1)

		crc_be[crc_32, in_cell_lw/**/[m_plus_1]/**/, $$payload/**/[m]/**/]
		// SCR6027: Added OAM Changes		
		#ifdef ATM_OAM_ENABLED
		; Update BEDC with current payload long word and create a copy
		; of payload for OAM processing
		#if (m == 0)
			ld_field_w_clr[cur_bedc, 1111, $$payload/**/[m]/**/]
		#else // (m == 0)
			alu[cur_bedc, cur_bedc, XOR, $$payload/**/[m]/**/]
		#endif // (m == 0)
		alu[$$atm_cell_copy[m_plus_1], --, B, $$payload/**/[m]/**/]
		#else // ATM_OAM_ENABLED
		nop
		#endif // ATM_OAM_ENABLED

		#undef m_plus_1
		#define_eval m (m+1)
	#endloop

	#undef m

	#undef ALIGNED_CELL_BLOCK

	#if (streq('CELL_SIZE', 'FORTY_EIGHT'))

	; Need 5 nops or eqv. instructions between crc_be[]
	; and reading CRC remainder.
	nop
	nop
	nop

	br[CRC_DONE], defer[2]

	; Read CRC_Remainder afer 5 nop's since crc_be[]
	; was last used and store it in TXC:crc.
	local_csr_rd[CRC_Remainder]
	immed[*l$index0[8], 0x0000]	

	#else // streq('CELL_SIZE' == 'FORTY')

	; Compute CRC on packet size and store packet size in 
	; $cell_lw/**/11/**/
	crc_be[crc_32, in_cell_lw/**/[WRD_OFFSET_11]/**/, txc_packet_size]
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	; Compute BEDC over LW11 and create a copy for OAM processing
	alu[cur_bedc, cur_bedc, XOR, txc_packet_size]
	alu[$$atm_cell_copy[11], --, B, txc_packet_size]
	#else
	nop
	#endif

	; Need 5 nops or eqv. instructions between crc_be[]
	; and reading CRC remainder.
	nop
	nop
	nop
	nop

	; Reading CRC_Remainder afer 5 nop's since crc_be[]
	; was last used.
	.reg crc_rem
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	.reg	temp_bedc
	#endif

	local_csr_rd[CRC_Remainder]
	immed[crc_rem, 0x0000]	
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	; Compute BEDC on cell CRC and create copy for OAM PM processing
	alu[temp_bedc, --, ~B, crc_rem]
	alu[$$atm_cell_copy[12], --, ~B, crc_rem]
	#endif // ATM_OAM_ENABLED
	
	br[CRC_DONE], defer[2]

	; Invert CRC remainder and store it in $cell_lw/**/12/**/.
	alu[in_cell_lw/**/[WRD_OFFSET_12]/**/, --, ~B, crc_rem]
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	; Update BEDC with the Cell CRC
	alu[cur_bedc, cur_bedc, XOR, temp_bedc]
	#else // ATM_OAM_ENABLED
	nop
	#endif // ATM_OAM_ENABLED
	
	#endif // streq('CELL_SIZE')

	.end

jmp_1#:

	.begin

	; Compute CRC on the first lw (3-bytes) of payload and store the payload
	; in $cell_lw.
	crc_be[crc_32, in_cell_lw/**/[WRD_OFFSET_1]/**/, $$payload[0]], bytes_1_3
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	; Compute BEDC on 3 bytes of payload longword 
	; and create copy for OAM processing
	ld_field_w_clr[cur_bedc, 0111, $$payload[0]]
	alu[$$atm_cell_copy[1], --, B, $$payload[0]]
	#else  // ATM_OAM_ENABLED
	nop
	#endif // ATM_OAM_ENABLED

	#if (streq('CELL_SIZE', 'FORTY_EIGHT'))
	// 11 lw's of the cell contain aligned payload.
	#define_eval ALIGNED_CELL_BLOCK	11
	#else // streq('CELL_SIZE' == 'FORTY')
	// 9 lw's of the cell contain aligned payload.
	#define_eval ALIGNED_CELL_BLOCK	9
	#endif // streq('CELL_SIZE')

	#define_eval m 1

	; Compute CRC on the aligned payload of the cell and store the
	; payload in $cell_lw.
	#while (m <= ALIGNED_CELL_BLOCK)
		#define_eval m_plus_1	(m+WRD_OFFSET_1)

		crc_be[crc_32, in_cell_lw/**/[m_plus_1]/**/, $$payload/**/[m]/**/]
		// SCR6027: Added OAM Changes
		#ifdef ATM_OAM_ENABLED
		; Update BEDC for current longword and create copy for OAM processing
		alu[cur_bedc, cur_bedc, XOR, $$payload/**/[m]/**/]
		alu[$$atm_cell_copy/**/[m_plus_1]/**/, --, B, $$payload/**/[m]/**/]
		#else  // ATM_OAM_ENABLED
		nop
		#endif // ATM_OAM_ENABLED
		
		#undef m_plus_1
		#define_eval m (m+1)
	#endloop

	#undef m

	#define_eval m	(ALIGNED_CELL_BLOCK+1)
	#define_eval m_plus_1	(m+WRD_OFFSET_1)

	; Compute CRC on last lw (1-byte) of payload and store payload in $cell_lw.
	crc_be[crc_32, in_cell_lw/**/[m_plus_1]/**/, $$payload/**/[m]/**/], byte_0

	#if (streq('CELL_SIZE', 'FORTY'))

	; We need to align the CPCS-UU byte, CPI byte and msb byte of packet size
	; with the last 1 byte of payload.
	.reg payload_pkt_size_tmp

	; Store the last byte of payload in a gpr.
	; Also set CPCS-UU and CPI to 0.
	ld_field_w_clr[payload_pkt_size_tmp, 1000, $$payload/**/[m]/**/]

	; Store msb byte of packet_size in the gpr.
	; When USE_NON_DEFAULT_UUI is not enabled, this sets CPCS-UU
	; and CPI to 0
	ld_field[payload_pkt_size_tmp, 0111, txc_packet_size, >>8]

	; Copy the gpr to $cell_lw.
	alu[in_cell_lw/**/[m_plus_1]/**/, --, B, payload_pkt_size_tmp]
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	; Compute BEDC on current longword and create copy for OAM processing
	alu[cur_bedc, cur_bedc, XOR, payload_pkt_size_tmp]
	alu[$$atm_cell_copy/**/[m_plus_1]/**/, --, B, payload_pkt_size_tmp]
	#endif

	#endif // streq('CELL_SIZE')

	#if (streq('CELL_SIZE', 'FORTY_EIGHT'))

	; Need 5 nops or eqv. instructions between crc_be[]
	; and reading CRC remainder.
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	.begin
		.reg	temp_bedc
		; Create copy of last byte of payload
		ld_field_w_clr[temp_bedc, 1000, $$payload[m]]
		alu[$$atm_cell_copy[m_plus_1], --, B, temp_bedc]
		; Update BEDC with last byte of payload
		alu[cur_bedc, cur_bedc, XOR, temp_bedc]
		; Rotate BEDC to compensate for LW alignment
		alu[cur_bedc, --, B, cur_bedc, <<rot8]
	.end
	#else
	nop
	nop
	nop
	nop
	#endif
	
	br[CRC_DONE], defer[2]

	; Read CRC_Remainder afer 5 nop's since crc_be[]
	; was last used and store it in TXC:crc.
	local_csr_rd[CRC_Remainder]
	immed[*l$index0[8], 0x0000]	

	#else // streq('CELL_SIZE' == 'FORTY')

	.reg pkt_size_crc_rem_tmp

	; Compute CRC on packet size and store packet size in 
	; $cell_lw. Note that we have 2-bytes to store the 
	; packet size. The msb byte is stored in the above $cell_lw/**/12/**/.
	; The lsb byte is stored in $cell_lw/**/12/**/.
	crc_be[crc_32, --, txc_packet_size]
	ld_field_w_clr[pkt_size_crc_rem_tmp, 1000, txc_packet_size, <<24]

	; Need 5 nops or eqv. instructions between crc_be[]
	; and reading CRC remainder.
	nop
	nop
	nop
	nop

	; Reading CRC_Remainder afer 5 nop's since crc_be[]
	; was last used.
	.reg crc_rem
	local_csr_rd[CRC_Remainder]
	immed[crc_rem, 0x0000]	

	; Invert CRC remainder.
	alu[crc_rem, --, ~B, crc_rem]

	ld_field[pkt_size_crc_rem_tmp, 0111, crc_rem, >>8]
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	.begin
		.reg temp_bedc
		; Compute BEDC over LW12 and LW13
		alu[cur_bedc, cur_bedc, XOR, pkt_size_crc_rem_tmp]
		ld_field_w_clr[temp_bedc, 1000, crc_rem]
		alu[cur_bedc, cur_bedc, XOR, temp_bedc]
		; Rotate BEDC to compensate for LW alignment
		alu[cur_bedc, --, B, cur_bedc, <<rot8]
		; Create a copy of LW12 and 13 for OAM processing
		ld_field_w_clr[$$atm_cell_copy[12], 1111, pkt_size_crc_rem_tmp]
		ld_field_w_clr[$$atm_cell_copy[13], 1000, crc_rem, <<24]
	.end
	#endif // ATM_OAM_ENABLED

	br[CRC_DONE], defer[2]

	; Store packet size and CRC remainder in $cell_lw's.
	ld_field_w_clr[in_cell_lw/**/[WRD_OFFSET_12]/**/, 1111, pkt_size_crc_rem_tmp]
	ld_field_w_clr[in_cell_lw/**/[WRD_OFFSET_13]/**/, 1000, crc_rem, <<24]

	#endif // streq('CELL_SIZE')
	#undef m
	#undef m_plus_1
	#undef ALIGNED_CELL_BLOCK
	.end

jmp_2#:

	.begin

	; Compute CRC on the first lw (2-bytes) of payload and store the payload
	; in $cell_lw.
	crc_be[crc_32, in_cell_lw/**/[WRD_OFFSET_1]/**/, $$payload[0]], bytes_2_3
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	; Update BEDC with two bytes for current payload LW
	ld_field_w_clr[cur_bedc, 0011, $$payload[0]]
	; Create a copy of the two bytes for OAM processing
	alu[$$atm_cell_copy[1], --, B, $$payload[0]]
	#else // ATM_OAM_ENABLED
	nop
	#endif // ATM_OAM_ENABLED
	
	#if (streq('CELL_SIZE', 'FORTY_EIGHT'))
	// 11 lw's of the cell contain aligned payload.
	#define_eval ALIGNED_CELL_BLOCK	11
	#else // streq('CELL_SIZE' == 'FORTY')
	// 9 lw's of the cell contain aligned payload.
	#define_eval ALIGNED_CELL_BLOCK	9
	#endif // streq('CELL_SIZE')

	#define_eval m 1

	; Compute CRC on the aligned payload of the cell and store the
	; payload in $cell_lw.
	#while (m <= ALIGNED_CELL_BLOCK)
		#define_eval m_plus_1	(m+WRD_OFFSET_1)

		crc_be[crc_32, in_cell_lw/**/[m_plus_1]/**/, $$payload/**/[m]/**/]
		// SCR6027: Added OAM Changes
		#ifdef ATM_OAM_ENABLED
		; Compute BEDC and create copy of current long word
		alu[cur_bedc, cur_bedc, XOR, $$payload/**/[m]/**/]
		alu[$$atm_cell_copy/**/[m_plus_1]/**/, --, B, $$payload/**/[m]/**/]
		#else // ATM_OAM_ENABLED
		nop
		#endif // ATM_OAM_ENABLED

		#undef m_plus_1
		#define_eval m (m+1)
	#endloop

	#undef m

	#define_eval m	(ALIGNED_CELL_BLOCK+1)
	#define_eval m_plus_1	(m+WRD_OFFSET_1)

	.reg payload_tmp

	; Compute CRC on last lw (2-bytes) of payload and store the payload in $cell_lw.
	crc_be[crc_32, payload_tmp, $$payload/**/[m]/**/], bytes_0_1
	
	#if (streq('CELL_SIZE', 'FORTY_EIGHT'))
	; Only two bytes valid in the last longwords
	immed_w0[payload_tmp, 0x0000]
	#else
	; Update payload_tmp with the CPCS-UU and CPI bytes. When USE_NON_DEFAULT_UUI
	; is not enabled, the CPCS-UU will be set to 0. CPI is always set to 0
	; This is used only for 'CELL_SIZE' == 'FORTY' case
	ld_field[payload_tmp, 0011, txc_packet_size, >>16]
	#endif
	
	alu[in_cell_lw/**/[m_plus_1]/**/, --, B, payload_tmp]
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	alu[cur_bedc, cur_bedc, XOR, payload_tmp]
	alu[$$atm_cell_copy/**/[m_plus_1]/**/, --, B, payload_tmp]
	#endif // ATM_OAM_ENABLED

	#undef m
	#undef m_plus_1

	#undef ALIGNED_CELL_BLOCK

	#if (streq('CELL_SIZE', 'FORTY_EIGHT'))

	; Need 5 nops or eqv. instructions between crc_be[]
	; and reading CRC remainder.
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	alu[cur_bedc, --, B, cur_bedc, <<rot16]
	#else // ATM_OAM_ENABLED
	nop
	nop
	#endif // ATM_OAM_ENABLED
	nop

	br[CRC_DONE], defer[2]

	; Read CRC_Remainder afer 5 nop's since crc_be[]
	; was last used and store it in TXC:crc.
	local_csr_rd[CRC_Remainder]
	immed[*l$index0[8], 0x0000]	

	#else // streq('CELL_SIZE' == 'FORTY')

	.reg pkt_size_crc_rem_tmp

	; Compute CRC on packet size and store packet size in 
	; $cell_lw/**/12/**/.
	crc_be[crc_32, --, txc_packet_size]
	ld_field_w_clr[pkt_size_crc_rem_tmp, 1100, txc_packet_size, <<16]

	; Need 5 nops or eqv. instructions between crc_be[]
	; and reading CRC remainder.
	nop
	nop
	nop
	nop

	; Reading CRC_Remainder afer 5 nop's since crc_be[]
	; was last used.
	.reg crc_rem
	local_csr_rd[CRC_Remainder]
	immed[crc_rem, 0x0000]	

	; Invert CRC remainder.
	alu[crc_rem, --, ~B, crc_rem]

	ld_field[pkt_size_crc_rem_tmp, 0011, crc_rem, >>16]
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	.begin
		.reg	temp_bedc
		; Compute BEDC on LW12 and 13 of payload
		alu[cur_bedc, cur_bedc, XOR, pkt_size_crc_rem_tmp]
		ld_field_w_clr[temp_bedc, 1100, crc_rem, <<16]
		alu[cur_bedc, cur_bedc, XOR, temp_bedc]
		; Rotate BEDC to compensate for LW alignment
		alu[cur_bedc, --, B, cur_bedc, <<rot16]
		; Create a copy of LW12 and 13 for OAM processing
		ld_field_w_clr[$$atm_cell_copy[12], 1111, pkt_size_crc_rem_tmp]
		ld_field_w_clr[$$atm_cell_copy[13], 1100, crc_rem, <<16]
	.end
	#endif // ATM_OAM_ENABLED
	
	br[CRC_DONE], defer[2]

	; Store packet size and CRC remainder in $cell_lw's.
	ld_field_w_clr[in_cell_lw/**/[WRD_OFFSET_12]/**/, 1111, pkt_size_crc_rem_tmp]
	ld_field_w_clr[in_cell_lw/**/[WRD_OFFSET_13]/**/, 1100, crc_rem, <<16]

	#endif // streq('CELL_SIZE')

	.end

jmp_3#:

	.begin

	; Compute CRC on the first lw (1-byte) of payload and store the payload
	; in $cell_lw.
	crc_be[crc_32, in_cell_lw/**/[WRD_OFFSET_1]/**/, $$payload[0]], byte_3
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	; Compute BEDC on first byte of payload and create a copy for OAM processing
	ld_field_w_clr[cur_bedc, 0001, $$payload[0]]
	alu[$$atm_cell_copy[1], --, B, $$payload[0]]
	#else  // ATM_OAM_ENABLED
	nop
	#endif // ATM_OAM_ENABLED
	
	#if (streq('CELL_SIZE', 'FORTY_EIGHT'))
	// 11 lw's of the cell contain aligned payload.
	#define_eval ALIGNED_CELL_BLOCK	11
	#else // streq('CELL_SIZE' == 'FORTY')
	// 9 lw's of the cell contain aligned payload.
	#define_eval ALIGNED_CELL_BLOCK	9
	#endif // streq('CELL_SIZE')

	#define_eval m 1

	; Compute CRC on the aligned payload of the cell and store the
	; payload in $cell_lw.
	#while (m <= ALIGNED_CELL_BLOCK)
		#define_eval m_plus_1	(m+WRD_OFFSET_1)

		crc_be[crc_32, in_cell_lw/**/[m_plus_1]/**/, $$payload/**/[m]/**/]
		// SCR6027: Added OAM Changes
		#ifdef ATM_OAM_ENABLED
		; Update BEDC for current longword
		alu[cur_bedc, cur_bedc, XOR, $$payload[m]]
		; Copy out longword for OAM processing
		alu[$$atm_cell_copy[m_plus_1], --, B, $$payload[m]]
		#else // ATM_OAM_ENABLED
		nop
		#endif // ATM_OAM_ENABLED

		#undef m_plus_1
		#define_eval m (m+1)
	#endloop

	#undef m

	#define_eval m	(ALIGNED_CELL_BLOCK+1)
	#define_eval m_plus_1	(m+WRD_OFFSET_1)

	.reg payload_tmp

	; Compute CRC on last lw (3-bytes) of payload and store the payload in $cell_lw.
	crc_be[crc_32, payload_tmp, $$payload/**/[m]/**/], bytes_0_2
	#if (streq('CELL_SIZE', 'FORTY_EIGHT'))
	; Only three bytes valid in this longword
	immed_b0[payload_tmp, 0x00]
	#else // streq('CELL_SIZE', 'FORTY_EIGHT')
	; Set the lsb byte to CPCS-UU for the 'CELL_SIZE' == 'FORTY'
	; case. When USE_NON_DEFAULT_UUI is not defined, this sets
	; UU to 0
	ld_field[payload_tmp, 0001, txc_packet_size, >>24]
	#endif //streq('CELL_SIZE', 'FORTY_EIGHT')
	
	alu[in_cell_lw/**/[m_plus_1]/**/,--, B, payload_tmp]
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	; Copy out payload LW for OAM processing
	alu[$$atm_cell_copy[m_plus_1], --, B, payload_tmp]
	; Compute BEDC over the payload LW
	alu[cur_bedc, cur_bedc, XOR, payload_tmp]
	#endif // ATM_OAM_ENABLED

	#undef m
	#undef m_plus_1

	#undef ALIGNED_CELL_BLOCK

	#if (streq('CELL_SIZE', 'FORTY_EIGHT'))

	; Need 5 nops or eqv. instructions between crc_be[]
	; and reading CRC remainder.
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	alu[cur_bedc, --, B, cur_bedc, <<rot24]
	#else // ATM_OAM_ENABLED
	nop
	nop
	#endif // ATM_OAM_ENABLED

	br[CRC_DONE], defer[2]

	; Read CRC_Remainder afer 5 nop's since crc_be[]
	; was last used and store it in TXC:crc.
	local_csr_rd[CRC_Remainder]
	immed[*l$index0[8], 0x0000]	

	#else // streq('CELL_SIZE' == 'FORTY')

	.reg pkt_size_crc_rem_tmp

	; Compute CRC on packet size and store packet size in 
	; $cell_lw/**/12/**/.
	; The msb byte of $cell_lw/**/12/**/ will always be set
	; to 0. This byte holds the CPI field. It will be always
	; be 0 since the packet size is only 2 bytes so the msb
	; byte of pkt_size_crc_rem_tmp will always be set to 0.
	crc_be[crc_32, --, txc_packet_size]
	ld_field_w_clr[pkt_size_crc_rem_tmp, 1110, txc_packet_size, <<8]

	; Need 5 nops or eqv. instructions between crc_be[]
	; and reading CRC remainder.
	nop
	nop
	nop
	nop

	; Reading CRC_Remainder afer 5 nop's since crc_be[]
	; was last used.
	.reg crc_rem
	local_csr_rd[CRC_Remainder]
	immed[crc_rem, 0x0000]	

	; Invert CRC remainder.
	alu[crc_rem, --, ~B, crc_rem]

	ld_field[pkt_size_crc_rem_tmp, 0001, crc_rem, >>24]
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	.begin
		.reg	temp_bedc
		; Compute BEDC on payload LW12 and LW13
		alu[cur_bedc, cur_bedc, XOR, pkt_size_crc_rem_tmp]
		ld_field_w_clr[temp_bedc, 1110, crc_rem, <<8]
		alu[cur_bedc, cur_bedc, XOR, temp_bedc]
		; Rotate BEDC to compensate for LW alignment
		alu[cur_bedc, --, B, cur_bedc, <<rot24]
		; Copy out LW12 and LW13 for OAM processing
		ld_field_w_clr[$$atm_cell_copy[12], 1111, pkt_size_crc_rem_tmp]
		ld_field_w_clr[$$atm_cell_copy[13], 1110, crc_rem, <<8]
	.end
	#endif // ATM_OAM_ENABLED

	br[CRC_DONE], defer[2]

	; Store packet size and CRC remainder in $cell_lw's.
	ld_field_w_clr[in_cell_lw/**/[WRD_OFFSET_12]/**/, 1111, pkt_size_crc_rem_tmp]
	ld_field_w_clr[in_cell_lw/**/[WRD_OFFSET_13]/**/, 1110, crc_rem, <<8]

	#endif // streq('CELL_SIZE')

	.end

jmp_4#:

	.begin

	#if (streq('CELL_SIZE', 'FORTY_EIGHT'))
	// 12 lw's of the cell contain aligned payload.
	#define_eval ALIGNED_CELL_BLOCK	12
	#else // streq('CELL_SIZE' == 'FORTY')
	// 10 lw's of the cell contain aligned payload.
	#define_eval ALIGNED_CELL_BLOCK	10
	#endif // streq('CELL_SIZE')

	#define_eval m 1

	; Compute CRC on the aligned payload of the cell and store the
	; payload in $cell_lw.
	#while (m <= ALIGNED_CELL_BLOCK)
		#define_eval m_plus_0	(m+WRD_OFFSET_0)

		crc_be[crc_32, in_cell_lw/**/[m_plus_0]/**/, $$payload/**/[m]/**/]
		// SCR6027: Added OAM Changes
		#ifdef ATM_OAM_ENABLED
		; Update BEDC for the current longword
		#if (m == 1)
			alu[cur_bedc, --, B, $$payload[m]]
		#else // m == 1
			alu[cur_bedc, cur_bedc, XOR, $$payload[m]]
		#endif // m == 1
		; Copy out the current longword for OAM processing
		alu[$$atm_cell_copy[m], --, B, $$payload[m]]
		#else // ATM_OAM_ENABLED
		nop
		#endif // ATM_OAM_ENABLED

		#undef m_plus_0		
		#define_eval m (m+1)
	#endloop

	#undef m

	#undef ALIGNED_CELL_BLOCK

	#if (streq('CELL_SIZE', 'FORTY_EIGHT'))

	; Need 5 nops or eqv. instructions between crc_be[]
	; and reading CRC remainder.
	nop
	nop
	nop

	br[CRC_DONE], defer[2]

	; Read CRC_Remainder afer 5 nop's since crc_be[]
	; was last used and store it in TXC:crc.
	local_csr_rd[CRC_Remainder]
	immed[*l$index0[8], 0x0000]	

	#else // streq('CELL_SIZE' == 'FORTY')

	; Compute CRC on packet size and store packet size in 
	; $cell_lw/**/11/**/
	crc_be[crc_32, in_cell_lw/**/[WRD_OFFSET_11]/**/, txc_packet_size]
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	; Compute BEDC and copy out LW11 for OAM processing
	alu[cur_bedc, cur_bedc, XOR, txc_packet_size]
	alu[$$atm_cell_copy[11], --, B, txc_packet_size]
	#else // ATM_OAM_ENABLED
	nop
	nop
	#endif // ATM_OAM_ENABLED

	; Need 5 nops or eqv. instructions between crc_be[]
	; and reading CRC remainder.
	nop
	nop
	nop

	; Reading CRC_Remainder afer 5 nop's since crc_be[]
	; was last used.
	.reg crc_rem
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	.reg	temp_bedc
	#endif

	local_csr_rd[CRC_Remainder]
	immed[crc_rem, 0x0000]	
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	; Update BEDC with CRC of atm cell and copy out 
	; Cell CRC for OAM processing
	alu[temp_bedc, --, ~B, crc_rem]
	alu[$$atm_cell_copy[12], --, ~B, crc_rem]
	#endif // ATM_OAM_ENABLED

	br[CRC_DONE], defer[2]

	; Invert CRC remainder and store it in $cell_lw/**/12/**/.
	alu[in_cell_lw/**/[WRD_OFFSET_12]/**/, --, ~B, crc_rem]
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	; Update BEDC with cell CRC
	alu[cur_bedc, cur_bedc, XOR, temp_bedc]
	#else // ATM_OAM_ENABLED
	nop
	#endif // ATM_OAM_ENABLED

	#endif // streq('CELL_SIZE')

	.end

jmp_5#:

	.begin

	; Compute CRC on the first lw (3-bytes) of payload and store the payload
	; in $cell_lw.
	crc_be[crc_32, in_cell_lw/**/[WRD_OFFSET_1]/**/, $$payload[1]], bytes_1_3
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	; Compute BEDC on first 3 bytes of payload
	ld_field_w_clr[cur_bedc, 0111, $$payload[1]]
	; Copy out first three bytes for OAM Processing
	alu[$$atm_cell_copy[1], --, B, $$payload[1]]
	#else // ATM_OAM_ENABLED
	nop
	#endif // ATM_OAM_ENABLED

	#if (streq('CELL_SIZE', 'FORTY_EIGHT'))
	// 11 lw's of the cell contain aligned payload.
	#define_eval ALIGNED_CELL_BLOCK	12
	#else // streq('CELL_SIZE' == 'FORTY')
	// 9 lw's of the cell contain aligned payload.
	#define_eval ALIGNED_CELL_BLOCK	10
	#endif // streq('CELL_SIZE')

	#define_eval m 2

	; Compute CRC on the aligned payload of the cell and store the
	; payload in $cell_lw.
	#while (m <= ALIGNED_CELL_BLOCK)
		#define_eval m_plus_0	(m+WRD_OFFSET_0)

		crc_be[crc_32, in_cell_lw/**/[m_plus_0]/**/, $$payload/**/[m]/**/]
		// SCR6027: Added OAM Changes
		#ifdef ATM_OAM_ENABLED
		; Update BEDC for payload LW and copy out LW for OAM processing
		alu[cur_bedc, cur_bedc, XOR, $$payload[m]]
		alu[$$atm_cell_copy/**/[m_plus_0]/**/, --, B, $$payload/**/[m]/**/]
		#else // ATM_OAM_ENABLED
		nop
		#endif // ATM_OAM_ENABLED
		
		#undef m_plus_0
		#define_eval m (m+1)
	#endloop

	#undef m

	#define_eval m	(ALIGNED_CELL_BLOCK+1)
	#define_eval m_plus_0	(m+WRD_OFFSET_0)

	; Compute CRC on last lw (1-byte) of payload and store the payload in $cell_lw.
	crc_be[crc_32, in_cell_lw/**/[m_plus_0]/**/, $$payload/**/[m]/**/], byte_0

	#if (streq('CELL_SIZE', 'FORTY'))

	; We need to align the CPCS-UU byte, CPI byte and msb byte of packet size
	; with the last 1 byte of payload.
	.reg payload_pkt_size_tmp

	; Store the last byte of payload in a gpr.
	; Also set CPCS-UU and CPI to 0.
	ld_field_w_clr[payload_pkt_size_tmp, 1000, $$payload/**/[m]/**/]

	; Store msb byte of packet_size in the gpr and CPCS-UU, CPI
	; from txc_packet_size into payload. When USE_NON_DEFAULT_UUI
	; flag is not defined, this will be set to 0.
	ld_field[payload_pkt_size_tmp, 0111, txc_packet_size, >>8]

	; Copy the gpr to $cell_lw.
	alu[in_cell_lw/**/[m_plus_0]/**/, --, B, payload_pkt_size_tmp]
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	; Compute BEDC on cell longword and copy out for OAM processing
	alu[$$atm_cell_copy[m], --, B, payload_pkt_size_tmp]
	alu[cur_bedc, cur_bedc, XOR, payload_pkt_size_tmp]
	#endif // ATM_OAM_ENABLED

	#endif // streq('CELL_SIZE')


	#if (streq('CELL_SIZE', 'FORTY_EIGHT'))

	; Need 5 nops or eqv. instructions between crc_be[]
	; and reading CRC remainder.
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	.begin
		.reg	temp_bedc

		; Copy out last byte for OAM processing
		alu[$$atm_cell_copy[m], --, B, $$payload[m]]

		; For payload size 48, the cell ends here update BEDC with
		; last byte of payload
		ld_field_w_clr[temp_bedc, 1000, $$payload[m]]
		alu[cur_bedc, cur_bedc, XOR, temp_bedc]

		; Rotate BEDC to compensate for alignment offset
		alu[cur_bedc, --, B, cur_bedc, <<rot8]
	.end
	#else // ATM_OAM_ENABLED
	nop
	nop
	nop
	nop
	#endif // ATM_OAM_ENABLED

	br[CRC_DONE], defer[2]

	; Read CRC_Remainder afer 5 nop's since crc_be[]
	; was last used and store it in TXC:crc.
	local_csr_rd[CRC_Remainder]
	immed[*l$index0[8], 0x0000]	

	#else // streq('CELL_SIZE' == 'FORTY')

	.reg pkt_size_crc_rem_tmp

	; Compute CRC on packet size and store packet size in 
	; $cell_lw. Note that we have 2-bytes to store the 
	; packet size. The msb byte is stored in the above $cell_lw/**/12/**/.
	; The lsb byte is stored in $cell_lw/**/12/**/.
	crc_be[crc_32, --, txc_packet_size]
	ld_field_w_clr[pkt_size_crc_rem_tmp, 1000, txc_packet_size, <<24]

	; Need 5 nops or eqv. instructions between crc_be[]
	; and reading CRC remainder.
	nop
	nop
	nop
	nop

	; Reading CRC_Remainder afer 5 nop's since crc_be[]
	; was last used.
	.reg crc_rem
	local_csr_rd[CRC_Remainder]
	immed[crc_rem, 0x0000]	

	; Invert CRC remainder.
	alu[crc_rem, --, ~B, crc_rem]

	ld_field[pkt_size_crc_rem_tmp, 0111, crc_rem, >>8]
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	.begin
		.reg	temp_bedc
		; Update BEDC for LW12 and LW13
		alu[cur_bedc, cur_bedc, XOR, pkt_size_crc_rem_tmp]
		ld_field_w_clr[temp_bedc, 1000, crc_rem, <<24]
		alu[cur_bedc, cur_bedc, XOR, temp_bedc]
		; Rotate BEDC to compensate for LW alignment
		alu[cur_bedc, --, B, cur_bedc, <<rot8]
		; Copy out LW12 and LW13 for OAM processing
		ld_field_w_clr[$$atm_cell_copy[12], 1111, pkt_size_crc_rem_tmp]
		ld_field_w_clr[$$atm_cell_copy[13], 1000, crc_rem, <<24]	
	.end
	#endif // ATM_OAM_ENABLED

	br[CRC_DONE], defer[2]

	; Store packet size and CRC remainder in $cell_lw's.
	ld_field_w_clr[in_cell_lw/**/[WRD_OFFSET_12]/**/, 1111, pkt_size_crc_rem_tmp]
	ld_field_w_clr[in_cell_lw/**/[WRD_OFFSET_13]/**/, 1000, crc_rem, <<24]

	#endif // streq('CELL_SIZE')

	#undef m
	#undef m_plus_0
	#undef ALIGNED_CELL_BLOCK
	.end

jmp_6#:

	.begin

	; Compute CRC on the first lw (2-bytes) of payload and store the payload
	; in $cell_lw.
	crc_be[crc_32, in_cell_lw/**/[WRD_OFFSET_1]/**/, $$payload[1]], bytes_2_3
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	; Update BEDC for this payload LW and 
	; make a copy for OAM processing
	ld_field_w_clr[cur_bedc, 0011, $$payload[1]]
	alu[$$atm_cell_copy[1], --, B, $$payload[1]]
	#else // ATM_OAM_ENABLED
	nop
	#endif // ATM_OAM_ENABLED

	#if (streq('CELL_SIZE', 'FORTY_EIGHT'))
	// 11 lw's of the cell contain aligned payload.
	#define_eval ALIGNED_CELL_BLOCK	12
	#else // streq('CELL_SIZE' == 'FORTY')
	// 9 lw's of the cell contain aligned payload.
	#define_eval ALIGNED_CELL_BLOCK	10
	#endif // streq('CELL_SIZE')

	#define_eval m 2

	; Compute CRC on the aligned payload of the cell and store the
	; payload in $cell_lw.
	#while (m <= ALIGNED_CELL_BLOCK)

		#define_eval m_plus_0	(m+WRD_OFFSET_0)
		crc_be[crc_32, in_cell_lw/**/[m_plus_0]/**/, $$payload/**/[m]/**/]
		// SCR6027: Added OAM Changes
		#ifdef ATM_OAM_ENABLED
		; Update BEDC for this payload LW and 
		; make a copy for OAM processing
		alu[cur_bedc, cur_bedc, XOR, $$payload[m]]
		alu[$$atm_cell_copy/**/[m_plus_0]/**/, --, B, $$payload/**/[m]/**/]
		#else // ATM_OAM_ENABLED
		nop
		#endif // ATM_OAM_ENABLED

		#undef m_plus_0

		#define_eval m (m+1)
	#endloop

	#undef m

	#define_eval m	(ALIGNED_CELL_BLOCK+1)
	#define_eval m_plus_0	(m+WRD_OFFSET_0)

	.reg payload_tmp

	; Compute CRC on last lw (2-bytes) of payload and store the payload in $cell_lw.
	crc_be[crc_32, payload_tmp, $$payload/**/[m]/**/], bytes_0_1
	
	#if (streq('CELL_SIZE', 'FORTY_EIGHT'))

	; Only two bytes of payload_tmp are valid
	immed_w0[payload_tmp, 0x0000]

	#else // streq('CELL_SIZE', 'FORTY_EIGHT')

	; Update payload_tmp with the CPCS-UU and CPI bytes from txc_packet_size. 
	; for 'CELL_SIZE' == 'FORTY' case. When USE_NON_DEFAULT_UUI is not enabled
	; this will set the CPCS-UU and CPI to 0. CRC for the non default UUI and
	; CPI will get computed in the 'FORTY' case since, the CRC is updated using
	; entire txc_packet_size
	ld_field[payload_tmp, 0011, txc_packet_size, >>16]

	#endif // streq('CELL_SIZE', 'FORTY_EIGHT'
	; Setting the 2 lsb bytes to 0, sets the CPCS-UU and CPI bytes to 0 for the
	; 'CELL_SIZE' == 'FORTY' case.
	alu[in_cell_lw/**/[m_plus_0]/**/, --, B, payload_tmp]
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	; Store copy of last atm cell longword for OAM processing
	alu[$$atm_cell_copy[m], --, B, payload_tmp]
	; Update BEDC for this payload longword
	alu[cur_bedc, cur_bedc, XOR, payload_tmp]
	#endif // ATM_OAM_ENABLED
	
	#undef m
	#undef m_plus_0

	#undef ALIGNED_CELL_BLOCK

	#if (streq('CELL_SIZE', 'FORTY_EIGHT'))

	; Need 5 nops or eqv. instructions between crc_be[]
	; and reading CRC remainder.
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	; Rotate BEDC to compensate for LW alignment
	alu[cur_bedc, --, B, cur_bedc, <<rot16]
	#else // ATM_OAM_ENABLED
	nop
	nop
	#endif // ATM_OAM_ENABLED

	br[CRC_DONE], defer[2]

	; Read CRC_Remainder afer 5 nop's since crc_be[]
	; was last used and store it in TXC:crc.
	local_csr_rd[CRC_Remainder]
	immed[*l$index0[8], 0x0000]	

	#else // streq('CELL_SIZE' == 'FORTY')

	.reg pkt_size_crc_rem_tmp

	; Compute CRC on packet size and store packet size in 
	; $cell_lw/**/12/**/.
	crc_be[crc_32, --, txc_packet_size]
	ld_field_w_clr[pkt_size_crc_rem_tmp, 1100, txc_packet_size, <<16]

	; Need 5 nops or eqv. instructions between crc_be[]
	; and reading CRC remainder.
	nop
	nop
	nop
	nop

	; Reading CRC_Remainder afer 5 nop's since crc_be[]
	; was last used.
	.reg crc_rem
	local_csr_rd[CRC_Remainder]
	immed[crc_rem, 0x0000]	

	; Invert CRC remainder.
	alu[crc_rem, --, ~B, crc_rem]

	ld_field[pkt_size_crc_rem_tmp, 0011, crc_rem, >>16]
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	.begin
		.reg	temp_bedc
		; Update BEDC with LW 12 data
		alu[cur_bedc, cur_bedc, XOR, pkt_size_crc_rem_tmp]
		; Update BEDC with LW 13 data
		ld_field_w_clr[temp_bedc, 1100, crc_rem, <<16]
		alu[cur_bedc, cur_bedc, XOR, temp_bedc]
		; Rotate BEDC to correct byte offset
		alu[cur_bedc, --, B, cur_bedc, <<rot16]
		; Store copies of LW12 and 13 for OAM processing
		ld_field_w_clr[$$atm_cell_copy[12], 1111, pkt_size_crc_rem_tmp]
		ld_field_w_clr[$$atm_cell_copy[13], 1100, crc_rem, <<16]
	.end
	#endif // ATM_OAM_ENABLED

	br[CRC_DONE], defer[2]

	; Store packet size and CRC remainder in $cell_lw's.
	ld_field_w_clr[in_cell_lw/**/[WRD_OFFSET_12]/**/, 1111, pkt_size_crc_rem_tmp]
	ld_field_w_clr[in_cell_lw/**/[WRD_OFFSET_13]/**/, 1100, crc_rem, <<16]

	#endif // streq('CELL_SIZE')

	.end

jmp_7#:

	.begin

	; Compute CRC on the first lw (1-byte) of payload and store the payload
	; in $cell_lw.
	crc_be[crc_32, in_cell_lw/**/[WRD_OFFSET_1]/**/, $$payload[1]], byte_3
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	; Update BEDC for this payload LW and 
	; make a copy for OAM processing
	ld_field_w_clr[cur_bedc, 0001, $$payload[1]]
	alu[$$atm_cell_copy[1], --, B, $$payload[1]]
	#else // ATM_OAM_ENABLED
	nop
	#endif // ATM_OAM_ENABLED


	#if (streq('CELL_SIZE', 'FORTY_EIGHT'))
	// 11 lw's of the cell contain aligned payload.
	#define_eval ALIGNED_CELL_BLOCK	12
	#else // streq('CELL_SIZE' == 'FORTY')
	// 9 lw's of the cell contain aligned payload.
	#define_eval ALIGNED_CELL_BLOCK	10
	#endif // streq('CELL_SIZE')

	#define_eval m 2

	; Compute CRC on the aligned payload of the cell and store the
	; payload in $cell_lw.
	#while (m <= ALIGNED_CELL_BLOCK)

		#define_eval m_plus_0	(m+WRD_OFFSET_0)
		crc_be[crc_32, in_cell_lw/**/[m_plus_0]/**/, $$payload/**/[m]/**/]
		// SCR6027: Added OAM Changes
		#ifdef ATM_OAM_ENABLED
		; Update BEDC for this payload LW and 
		; make a copy for OAM processing
		alu[cur_bedc, cur_bedc, XOR, $$payload[m]]
		alu[$$atm_cell_copy/**/[m_plus_0]/**/, --, B, $$payload/**/[m]/**/]
		#else // ATM_OAM_ENABLED
		nop
		#endif // ATM_OAM_ENABLED
		
		#undef m_plus_0

		#define_eval m (m+1)
	#endloop

	#undef m

	#define_eval m	(ALIGNED_CELL_BLOCK+1)
	#define_eval m_plus_0	(m+WRD_OFFSET_0)

	.reg payload_tmp

	; Compute CRC on last lw (3-bytes) of payload and store the payload in $cell_lw.
	crc_be[crc_32, payload_tmp, $$payload/**/[m]/**/], bytes_0_2
	#if (streq('CELL_SIZE', 'FORTY_EIGHT'))
	; Only three bytes valid in this payload LW
	immed_b0[payload_tmp, 0x00]

	#else  // streq('CELL_SIZE', 'FORTY_EIGHT')

	; Update payload_tmp with the CPCS-UU from txc_packet_size for 
	; 'CELL_SIZE' == 'FORTY' case. When USE_NON_DEFAULT_UUI is not
	; enabled this will set the CPCS-UU to 0. CRC for the non 
	; default UUI will get computed in the 'FORTY' case since, 
	; the CRC is updated using entire txc_packet_size
	ld_field[payload_tmp, 0001, txc_packet_size, >>24]

	#endif // streq('CELL_SIZE', 'FORTY_EIGHT')

	alu[in_cell_lw/**/[m_plus_0]/**/, --, B, payload_tmp]
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	; Store copy of longword for OAM processing
	alu[$$atm_cell_copy[m], --, B, payload_tmp]
	; Update BEDC for this payload longword
	alu[cur_bedc, cur_bedc, XOR, payload_tmp]
	#endif // ATM_OAM_ENABLED
	
	#undef m
	#undef m_plus_0

	#undef ALIGNED_CELL_BLOCK

	#if (streq('CELL_SIZE', 'FORTY_EIGHT'))

	; Need 5 nops or eqv. instructions between crc_be[]
	; and reading CRC remainder.
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	alu[cur_bedc, --, B, cur_bedc, <<rot24]
	#else // ATM_OAM_ENABLED
	nop
	nop
	#endif // ATM_OAM_ENABLED

	br[CRC_DONE], defer[2]

	; Read CRC_Remainder afer 5 nop's since crc_be[]
	; was last used and store it in TXC:crc.
	local_csr_rd[CRC_Remainder]
	immed[*l$index0[8], 0x0000]	

	#else // streq('CELL_SIZE' == 'FORTY')

	.reg pkt_size_crc_rem_tmp

	; Compute CRC on packet size and store packet size in 
	; $cell_lw/**/12/**/.
	; The msb byte of $cell_lw/**/12/**/ will always be set
	; to 0. This byte holds the CPI field. It will be always
	; be 0 since the packet size is only 2 bytes so the msb
	; byte of pkt_size_crc_rem_tmp will always be set to 0.
	crc_be[crc_32, --, txc_packet_size]
	ld_field_w_clr[pkt_size_crc_rem_tmp, 1110, txc_packet_size, <<8]

	; Need 5 nops or eqv. instructions between crc_be[]
	; and reading CRC remainder.
	nop
	nop
	nop
	nop

	; Reading CRC_Remainder afer 5 nop's since crc_be[]
	; was last used.
	.reg crc_rem
	local_csr_rd[CRC_Remainder]
	immed[crc_rem, 0x0000]	

	; Invert CRC remainder.
	alu[crc_rem, --, ~B, crc_rem]

	ld_field[pkt_size_crc_rem_tmp, 0001, crc_rem, >>24]
	// SCR6027: Added OAM Changes
	#ifdef ATM_OAM_ENABLED
	.begin
		.reg	temp_bedc
		; Update BEDC for LW12
		alu[cur_bedc, cur_bedc, XOR, pkt_size_crc_rem_tmp]
		; Update BEDC for LW13
		ld_field_w_clr[temp_bedc, 1110, crc_rem, <<8]
		alu[cur_bedc, cur_bedc, XOR, temp_bedc]
		; Rotate BEDC to compensate for alignment offset
		alu[cur_bedc, --, B, cur_bedc, <<rot24]
		; Copy out LW12 and 13 for OAM processing
		ld_field_w_clr[$$atm_cell_copy[12], 1111, pkt_size_crc_rem_tmp]
		ld_field_w_clr[$$atm_cell_copy[13], 1110, crc_rem, <<8]
	.end
	#endif // ATM_OAM_ENABLED
	br[CRC_DONE], defer[2]

	; Store packet size and CRC remainder in the last lw of the cell.
	ld_field_w_clr[in_cell_lw/**/[WRD_OFFSET_12]/**/, 1111, pkt_size_crc_rem_tmp]
	ld_field_w_clr[in_cell_lw/**/[WRD_OFFSET_13]/**/, 1110, crc_rem, <<8]

	#endif // streq('CELL_SIZE')

	.end

.end
#endm


#endif // _AAL5_TX_CRC_UC_