///////////////////////////////////////////////////////////////////////
//                                                                   
//                  I N T E L   P R O P R I E T A R Y                
//                                                                   
//     COPYRIGHT (c)  2001-2002 BY  INTEL  CORPORATION.  ALL RIGHTS  
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY   
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A 
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER 
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL, 
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT 
//     THE PRIOR WRITTEN PERMISSION OF :                             
//                                                                   
//                        INTEL  CORPORATION                         
//                                                                   
//                     2200 MISSION COLLEGE BLVD                     
//                                                                   
//               SANTA  CLARA,  CALIFORNIA  95052-8119               
//                                                                   
///////////////////////////////////////////////////////////////////////
//                                                                    
//                                                                   
//      File Name: aal5_oam_tx.uc                                         
//                                                                   
//      Description: This is the main file. It contains initialization
//                   phase macro calls and recurring phase macro calls.
//                                                                   
///////////////////////////////////////////////////////////////////////

#ifndef	_AAL5_OAM_TX_UC_
#define _AAL5_OAM_TX_UC_

#include <crc.uc>
//------------------------------------------------------------------
// _aal5_oam_tx_init()
//
//    Description: Perform one time initialization required for PM
//				   Clear cache tag, set up global registers, load
//				   CRC-10 lookup tables,
//
//    Parameters:
//			None
//------------------------------------------------------------------
#macro _aal5_oam_tx_init() 
.begin
	#ifdef ATM_OAM_ENABLED
	.reg lm_cache_addr
	.reg old_lm_addr0
	.reg old_lm_addr1
	.reg crc10_tbl_sram_base
	.sig sram_rd_sig
	.reg i
	.reg j
	.reg crc10_tbl_ptr

	xbuf_alloc($crc_val, 8, read)

	; Load the original LM address pointers
	local_csr_rd[active_lm_addr_0]
	immed[old_lm_addr0, 0]
	local_csr_rd[active_lm_addr_1]
	immed[old_lm_addr1, 0]

	; Setup LM address pointer 
	move(lm_cache_addr, CRC_10_TABLE_ADDR)
	local_csr_wr[active_lm_addr_0, lm_cache_addr]			
	
	; Setup GPR to store the LM base address of CRC
	; table for use by CRC computation routines
	move(@crc10_table_base_addr, lm_cache_addr)
	
	; Setup address in SRAM containing the CRC-10 table
	move(crc10_tbl_ptr, CRC_10_TABLE_SRAM_ADDR)
	
	; Initialize loop variables. CRC-10 table is 128 LW
	; long. We use two loops . Outer loop runs 16 times
	; and fetches 8 LW from SRAM on each iteration. The
	; 8 LW are stored in local memory by the inner loop
	; which runs 8 times.
	immed[i, 0x00]
	
	.repeat
		; Setup pointer to register file to read 
		; CRC table from SRAM into local memory
		; Need three cycles before the indexed registers
		; can be accessed.
		alu[j, --, B, &$crc_val[0], <<2]	; Get the byte address
		alu[j, j, OR, thread_num, <<6]	; 16 xfer register/context
		local_csr_wr[T_INDEX, j]
	
		sram[read, $crc_val[0], crc10_tbl_ptr, 0, LWCOUNT8],\
			 sig_done[sram_rd_sig]
		ctx_arb[sram_rd_sig], defer[1]
		immed[j, 0x00]
		
		.repeat
			alu[j, j, +, 1]
			alu[*l$index0++, --, B, *$index++]
		.until	(j >= 8)
		
		; Update variable for next iteration
		alu[i, i, +, 1]
		
		; advance the SRAM pointer by 8 Long words
		alu[crc10_tbl_ptr, crc10_tbl_ptr, +, (LWCOUNT8 * BYTES_PER_LW)]
	
	.until(i >= 16)

	; Clear counter cache 
	immed[lm_cache_addr, OAM_TX_VC_TAG_BASE]		; Local LM_addr_1 <- base
	local_csr_wr[active_lm_addr_1, lm_cache_addr]	

	immed[lm_cache_addr, OAM_TX_VP_TAG_BASE]		; Local LM_addr_0 <- base
	local_csr_wr[active_lm_addr_0, lm_cache_addr]	
	nop
	immed[i, 0]

	; Clear the VP and VC PM counters cache in local memory
	.repeat
		alu[*l$index1++, --, B, 0]
		alu[*l$index0++, --, B, 0]
		alu[i, i, +, 1]
	.until(i >= 8)

	; Restore the LM address pointers to original values
	local_csr_wr[active_lm_addr_0, old_lm_addr0]	
	local_csr_wr[active_lm_addr_1, old_lm_addr1]	
	nop
	nop
	nop
	#endif // ATM_OAM_ENABLED
.end
#endm
//------------------------------------------------------------------
// _aal5_tx_oam_pm_processing()
//
//    Description: Main entry macro for F4 and F5 OAM PM processing.
//                 Invokes macros to read current PM counters from 
//                 SRAM, update and write them back to SRAM. If block
//                 size for FPM cell generation is reached, a FPM cell
//                 is generated. 
//
//    Parameters:
//			in_vcq	
//					VCQ# on which the current cell is being transmitted
//------------------------------------------------------------------
#macro _aal5_tx_oam_pm_processing(in_vcq)
.begin
	#ifdef ATM_OAM_ENABLED
	.reg	vp_ctr_base				; VP OAM Tx Counter table base address
	.reg	vp_ctr_offset			; VP OAM Tx Counter table index
	.reg	vc_ctr_base				; VC OAM Tx Counter table base address
	.reg	vc_ctr_offset			; VC OAM Tx Counter table index
	.reg	old_lm_addr0			; Original LM address pointer 0
	.reg	old_lm_addr1			; Original LM address pointer 1
	.reg	return_address
	.reg	crcResult
	.reg	data

	; Load the original LM address pointers
	local_csr_rd[active_lm_addr_0]
	immed[old_lm_addr0, 0]
	local_csr_rd[active_lm_addr_1]
	immed[old_lm_addr1, 0]

	xbuf_alloc($$oam_tx_ctrs, LWCOUNT5, read)
	xbuf_alloc($oam_tx_ctrs, LWCOUNT5, write)


	; Phase 2_1

	; Load F5 flow counters and options from SRAM
	_aal5_oam_tx_load_F5_pm_ctrs(in_vcq, vc_ctr_base, vc_ctr_offset)

	; Load any cells buffered due to FPM cell transmission on 
	; this VC. When PM cells are generated the ATM cell scheduled for
	; transmission in that slot is buffered to transmit FPM cell. The
	; current cell is buffered and the buffered cell is sent instead.
	_aal5_oam_tx_load_delayed_cell()

	; Phase 2_2
	; Load F4 OAM counters from SRAM.	
	_aal5_oam_tx_load_F4_pm_ctrs(vp_ctr_base, vp_ctr_offset)

	; Counter updates and send FPM cell if block size reached.
	_aal5_oam_tx_update_pm_ctrs(in_vcq)

	; Flush cached counters back to SRAM.
	_aal5_oam_tx_write_pm_ctrs(vp_ctr_base, vp_ctr_offset,\
								vc_ctr_base, vc_ctr_offset)

	xbuf_free($$oam_tx_ctrs)
	xbuf_free($oam_tx_ctrs)

	; Restore the LM address pointers to original values
	local_csr_wr[active_lm_addr_1, old_lm_addr1]	
	local_csr_wr[active_lm_addr_0, old_lm_addr0]	

	br[END#]
	
STUB_CRC_10_ADDRESS#:
	_stub_crc_10_fcs()

END#:
	#endif // ATM_OAM_ENABLED
.end
#endm
//------------------------------------------------------------------
// _aal5_oam_tx_load_F5_pm_ctrs()
//
//    Description: Macro to load the F5 OAM PM counters and the PM
//                 options indicating whether PM is enabled for VC 
//
//    Parameters:
//			in_vcq	
//					VCQ# on which the current cell is being transmitted
//			out_oam_tx_tbl_base
//					Updated with SRAM base address of F5 counter table
//			out_oam_tx_tbl_offset
//					Updated with start offset in SRAM F5 counter table
//					for this VCQ#
//------------------------------------------------------------------
#macro	_aal5_oam_tx_load_F5_pm_ctrs(in_vcq, out_oam_tx_tbl_base,\
									 out_oam_tx_tbl_offset)
.begin

	.reg temp_var						; temporary variable
	.sig sram_read_sig					; To wait on SRAM read

	; Phase 3 start

	; Inter thread signalling takes 10 cycles. Set signal for next thread here
	; to avoid this latency. Make sure that there are no context arbs before
	; the end of the critical section 

	signal_ctx(next_thread)

	; This is either a F5 OAM cell or a user cell. In either case
	; load the VC OAM Tx Table to load the OAM options for the cell
	; The VC related PM counters are also loaded simultaneously and
	; updated into local memory cache.

	move(out_oam_tx_tbl_base, VC_TX_OAM_TBL_BASE)

	; F5 flow counters are stored in 5 long words in SRAM (20 bytes)

	alu[out_oam_tx_tbl_offset, --, B, in_vcq, <<4]
	alu[temp_var, --, B, in_vcq, <<2]
	alu[out_oam_tx_tbl_offset, out_oam_tx_tbl_offset, +, temp_var]

	; In this critical section, the LM counter cache is updated with
	; the VC table counters. Look if any other thread is already 
	; cacheing this VCQ entry.	
	; This lookup function will set the ACTIVE_LM_ADDR_1 to point to the
	; matching LM entry belonging to another thread in case of a hit.
	; In case of a miss, it will set ACTIVE_LM_ADDR_1 to point to the blank
	; LM entry of self thread. In case of a hit, branch done to pm_cache_hit

	_aal5_oam_tx_lkup_lm_pm_cache(in_vcq, PM_CACHE_HIT#, 1, VC)

PM_CACHE_MISS#:

	; Cache missed. This thread will fetch the counters from SRAM

	sram[read, $$oam_tx_ctrs[0], out_oam_tx_tbl_base,\
	     out_oam_tx_tbl_offset, LWCOUNT5], sig_done[sram_read_sig]


	immed[temp_var, OAM_TX_VC_TAG_BASE]
	alu[temp_var, temp_var, OR, thread_num, <<2]

	ctx_arb[prev_thread_done_sig, sram_read_sig], defer[2]
		; Initialize flags. Currently two flags defined. One to indicate
		; if F4 counters are alread loaded (bit 0), and second to indicate
		; if the buffered cell descriptor in LM cache has changed (bit 1)
		; Clear both flags to indicate false for both conditions.
		; The uppermost 16 bits are used to store the tag address. This 
		; is used by the last thread to clear the cache entry.
		ld_field_w_clr[lm_vc_ref_cnt_flgs, 1100, temp_var, <<16]
		; Increment reference count and swap out to wait for SRAM
		; read completion. Also wait for signal from previous thread
		; to enter next phase of the pipeline. Update in lm_vc_ref_cnt_flgs
		alu[lm_vc_ref_cnt_flgs, lm_vc_ref_cnt_flgs, +, 1]  	

	; Store counters read from SRAM into LM
	; counter 0 -> TUCC0+1 |   TUCC0
	; counter 1 -> BEDC    | txN | MCSN
	; counter 2 -> lm_vc_del_bfr
	; counter 3 -> BEDC bfr| FLIP,OPTIONS | PAYLO 
	; counter 4 -> lm_vc_del_bfr_hdr
	
	ld_field_w_clr[lm_vc_tucc0,    0011, $$oam_tx_ctrs[0]]       ; lm_vc_tucc0
	ld_field_w_clr[lm_vc_tucc0_1,  0011, $$oam_tx_ctrs[0], >>16] ; lm_vc_tucc0_1
	ld_field_w_clr[lm_vc_txN_opts, 0001, $$oam_tx_ctrs[1], >>8]  ; lm_vc_txN_opts
	ld_field_w_clr[lm_vc_mcsn,     0001, $$oam_tx_ctrs[1]]       ; lm_vc_mcsn
	ld_field_w_clr[lm_vc_bedc,     0011, $$oam_tx_ctrs[1], >>16] ; lm_vc_bedc
	ld_field_w_clr[lm_vc_del_bfr,  1111, $$oam_tx_ctrs[2]]       ; lm_vc_del_bfr
	ld_field[lm_vc_txN_opts,       0110, $$oam_tx_ctrs[3], <<8]  ; lm_vc_txN_opts
	br[END#], defer[2]
	; Below two instructions placed in the defer shadow of branch.
	ld_field[lm_vc_bedc,              1100, $$oam_tx_ctrs[3]]	  ; lm_vc_bedc
	ld_field_w_clr[lm_vc_del_bfr_hdr, 1111, $$oam_tx_ctrs[4]]    ; lm_vc_del_bfr_hdr

PM_CACHE_HIT#:

	; Counters already cached. Increment the reference count for the 
	; counters and wait for previous thread signal to enter next phase

	ctx_arb[prev_thread_done_sig], defer[1]
		alu[lm_vc_ref_cnt_flgs, lm_vc_ref_cnt_flgs, +, 1]  ; lm_vc_ref_cnt_flgs

END#:

	; The BEDC field computed over the current cell is placed in cur_bedc
	; The two 16 bit words are XORed to get the BIP-16 BEDC for this cell

	ld_field_w_clr[temp_var, 0011, cur_bedc, >>16]
	alu[cur_bedc, cur_bedc, XOR, temp_var]

.end
#endm //_aal5_oam_tx_load_F5_pm_ctrs
//------------------------------------------------------------------
// _aal5_oam_tx_load_delayed_cell()
//
//    Description: Macro to load the delayed cell if any for this VC
//                 The buffer in which the delayed cell is placed is
//				   stored along with the F5 PM counters. If there is
//				   a buffered cell, the current cell is buffered and
//				   the buffered cell is scheduled for transmission.
//
//    Parameters:
//				   None.
//------------------------------------------------------------------
#macro _aal5_oam_tx_load_delayed_cell()
.begin

	.reg 	dram_addr				; DRAM address to load ATM cell
	.reg	temp_var				; Temporary scratch variable
	.sig	dram_write_sig			; Signal for completion of DRAM write
	.sig	tbuf_write_sig			; Signal for completion of TBUF write

	; Check if a delayed buffer exists. If delayed buffer exists
	; use that for transmission instead of the current buffer.
	; Store current ATM cell into the delayed buffer

	alu[--, lm_vc_del_bfr, -, UNALLOCATED]   ; lm_vc_del_bfr
	beq[NO_DELAYED_CELL#]

	; Send delayed buffer to MSF. Write the current cell to delayed
	; buffer. The delayed buffer has space to hold two cells. A flip
	; flop variable is used to determine the offset from where to 
	; write the delayed cell i.e. offset 0 or offset 64. This way
	; the delayed cell can be transferred to MSF and the current cell
	; to the delayed buffer without having to read back the delayed cell.

	dl_buf_get_data[dram_addr, lm_vc_del_bfr] ; lm_vc_del_bfr

	; Check if the delayed buffer is at offset 0 or 64 in lm_vc_txN_opts
	br_bclr[lm_vc_txN_opts, BFRD_CELL_FLIP_FLOP_BT, WR_MSF#]

	; Delayed buffer is at offset 64
	alu[dram_addr, dram_addr, +, CELL_1_OFFSET]

WR_MSF#:

	; Set up the indirect references needed for transferring from
	; DRAM to MSF. User the tbuf_add_plus_4 to write to MSF. Here
	; we overwrite the existing cell in MSF with the delayed cell

	.begin
		.reg indir_ref					; indirect reference reg
		.reg tbuf_addr_mask				; mask for valid TBUF locs

		; The global masks pointed to local memory pointer 1
		; cannot be used since local memory pointer 1 is 
		; modified. Recompute the mask to be used based on
		; the PHY mode and compute the tbuf address.
		#if (TX_PHY_MODE == SPHY_1_32)

		; Mask that limits the TBUF address space
		; used to transmit a cell to 8KB.
		immed[tbuf_addr_mask, 0x1FFF]

		#endif // (TX_PHY_MODE == SPHY_1_32)

		#if ((TX_PHY_MODE == SPHY_4_8) || (TX_PHY_MODE == MPHY_4) || (TX_PHY_MODE == SPHY_4_8_SHARED))

		; Mask that limits the TBUF address space 
		; used to transmit a cell to 2KB.
		immed[tbuf_addr_mask, 0x7FF]

		#endif // ((TX_PHY_MODE == SPHY_4_8) || (TX_PHY_MODE == MPHY_4))

	 	; Compute the TBUF address of the available TBUF element.
		alu[indir_ref, tbuf_addr_mask, AND, \
			$avail_tbuf_element, <<TBUF_ELE_SIZE_LOG2]

		; Write to address tbuf_addr + 4. The prepend 
		; length is then set to 4 (ATM header) and the
		; prepend offset is set to 4
		alu[indir_ref, indir_ref, +, tbuf_element_base]

		; Set the indirect reference bit for the tbuf address
		alu[indir_ref, 0x10, OR, indir_ref, <<5]

		; Set up indirect reference to send 8 quad words
		; (64 bytes). Reference count is one less than 
		; actual number of quad words Set bit 25 for the
		; reference count indirection.

		alu[--, indir_ref, OR, 0x17, <<21]
		dram[tbuf_wr, --, dram_addr, 0, max_8], indirect_ref,\
										sig_done[tbuf_write_sig]
	.end

	; Swap the payload offset of current cell with 
	; that stored for delayed cell in lm_vc_txN_opts.
	alu[temp_var, --, B, tcw_payload_offset]
	ld_field_w_clr[tcw_payload_offset, 0010, lm_vc_txN_opts]
	ld_field[lm_vc_txN_opts, 0010, temp_var]

	; Swap the ATM header of current cell with that 
	; stored for the delayed cell in lm_vc_del_bfr_hdr
	alu[temp_var, --, B, lm_vc_del_bfr_hdr]
	alu[lm_vc_del_bfr_hdr, --, B, atm_cell_copy_header]
	alu[atm_cell_copy_header, --, B, temp_var]

	; Swap the BEDC computed for current cell with the 
	; BEDC stored for the delayed cell in lm_vc_bedc
	alu[temp_var, --, B, cur_bedc]
	ld_field_w_clr[cur_bedc, 0011, lm_vc_bedc, >>16]
	ld_field[lm_vc_bedc, 1100, temp_var, <<16]

	; Verify if we need to store the current cell. If 
	; the current cell is the flag buffer descriptor, 
	; then just invalidate the delayed buffer and go. 
	alu[--, lm_vc_del_bfr_hdr, -, 0]
	bne[CURRENT_CELL_VALID#], defer[1]
		; Also mark the delayed buffer as changed since 
		; at least the flip flop bit would have toggled
		; and the header could have changed (different
		; CLP etc.) even if we did not release the 
		; delayed buffer yet. This ensures a writeback
		; to SRAM when all references to this cached 
		; record have been exhausted.
		alu[lm_vc_ref_cnt_flgs, lm_vc_ref_cnt_flgs, OR,\
		1, <<OAM_TX_DEL_BFR_CH_BT_PS]
		
	; Wait for the TBUF write to complete to free delayed buffer.
	ctx_arb[tbuf_write_sig], defer[2]

		; Save the delayed buffer which is to be deallocated
		; in a temporary variable. Clear delayed buffer pointer.
		alu[temp_var, --, B, lm_vc_del_bfr]

		; Mark lm_vc_del_bfr as UNALLOCATED. The delayed
		; buffer changed has already been set
		immed[lm_vc_del_bfr, UNALLOCATED]

	; Drop the delayed buffer (lm_vc_del_bfr)
	dl_buf_free(temp_var, BUF_FREE_LIST0)

	; Reset the flip flop bit and the TCW payload offset
	; stored in the lm_vc_txN_opts register. 
	ld_field_w_clr[temp_var, 0001, lm_vc_txN_opts, >>16]

	br[END_LOAD_DELAYED_CELL#], defer[2]
		alu[temp_var, temp_var, AND, 0x0F]
		ld_field[lm_vc_txN_opts, 0100, temp_var, <<16]

CURRENT_CELL_VALID#:

	; Copy current cell to the delayed buffer. 
	; Check if the delayed cell was at offset 0 or CELL_1_OFFSET
	; If the delayed cell was at offset 0, we write this cell at
	; CELL_1_OFFSET else write to offset 0
	; Get DRAM address of lm_vc_del_bfr
	dl_buf_get_data[dram_addr, lm_vc_del_bfr]
	; Check offset to write from lm_vc_txN_opts flip flop bit
	br_bset[lm_vc_txN_opts, BFRD_CELL_FLIP_FLOP_BT, wr_dram#], defer[1]
		; Toggle the flip flop bit to indicate
		; the location of the delayed buffer
		alu[temp_var, --, B, 1, <<BFRD_CELL_FLIP_FLOP_BT]

	alu[dram_addr, dram_addr, +, CELL_1_OFFSET]

WR_DRAM#:

	; Toggle flip flop bit in lm_vc_TxN_opts
	alu[lm_vc_txN_opts, lm_vc_txN_opts, XOR, temp_var]

	; Write 64 bytes of data to DRAM. 48 bytes payload + 
	; 4 byte ATM header + 4 bytes of pre-pend length + 4
	; bytes written at end for quad word alignment reason

	dram[write, $$prepend_header, dram_addr, 0, LWCOUNT8],\
		 sig_done[dram_write_sig]

	; Wait for TBUF write and DRAM write to complete.
	ctx_arb[tbuf_write_sig, dram_write_sig],
		br[END_LOAD_DELAYED_CELL#]

NO_DELAYED_CELL#:
	; No delayed cell.
	.io_completed dram_write_sig tbuf_write_sig

END_LOAD_DELAYED_CELL#:
.end
#endm //_aal5_oam_tx_load_delayed_cell
//------------------------------------------------------------------
// _aal5_oam_tx_load_F4_pm_ctrs()
//
//    Description: Macro to load the F4 OAM PM counters and the PM
//                 options indicating whether PM is enabled for VP
//
//    Parameters:
//			out_oam_tx_tbl_base
//					Updated with SRAM base address of F4 counter table
//			out_oam_tx_tbl_offset
//					Updated with start offset in SRAM F4 counter table
//------------------------------------------------------------------
#macro	_aal5_oam_tx_load_F4_pm_ctrs(out_oam_tx_tbl_base,\
									 out_oam_tx_tbl_ofst )
.begin	
	.reg temp_var					; Temporary variable
	.reg vpi						; VPI of this VCQ#
	.sig sram_read_sig				; Wait completion of SRAM read

	; Phase 4 start
	; Signal next thread to avoid signal setting latency.
	; Ensure that there are no context arbs before critical section
	; is exited.
	signal_ctx(next_thread)

	; The assumption is that either F4 or F5 PM is enabled for a VCC
	; and not both. In line with bellcore recommendations [GR-1248-Core]
	; (Section 7.1.3.1 p.7-26) 

	; Check if VP OAM PM enabled. If F5 PM is enabled for the VC then
	; no F4 PM can be enabled since as stated above, simulatenous PM 
	; on F4 and F5 flows is not requires.

	alu[--, F5_PM_IND_BITS, AND, lm_vc_txN_opts,>>16]
	bne[NO_VP_PM_ENABLED#]

	alu[vpi, --, B, atm_cell_copy_header, >>20]		; Get VPI from cell
	; VP OAM Tx Table index = in_port * MAX_VPI_PER_PORT + in_vpi
	alu[out_oam_tx_tbl_ofst, --, B, outport, <<MAX_VPI_PER_PORT_SHIFT]
	alu[out_oam_tx_tbl_ofst, out_oam_tx_tbl_ofst, +, vpi]

	; VP OAM Tx Table offset = VP OAM Tx Table index * VP OAM Tx Table Size
	alu[out_oam_tx_tbl_ofst, --, B, out_oam_tx_tbl_ofst, <<VP_OAM_TX_TBL_SZ_SHF]

	move(out_oam_tx_tbl_base, VP_TX_OAM_TBL_BASE)

	;This lookup function will set the ACTIVE_LM_ADDR_0 to point to the
	;matching LM entry belonging to another thread in case of a hit.
	;In case of a miss, it will set ACTIVE_LM_ADDR_0 to point to the blank
	;LM entry of self thread. In case of a hit, branch done to pm_cache_hit

	_aal5_oam_tx_lkup_lm_pm_cache(out_oam_tx_tbl_ofst, PM_CACHE_HIT#, 0, VP)

PM_CACHE_MISS#:

	immed[temp_var, OAM_TX_VP_TAG_BASE]
	alu[temp_var, temp_var, OR, thread_num, <<2]


	sram[read, $$oam_tx_ctrs[0], out_oam_tx_tbl_base, out_oam_tx_tbl_ofst, LWCOUNT2],\
		sig_done[sram_read_sig]

	ctx_arb[sram_read_sig, prev_thread_done_sig], defer[2]
		; Store the tag of this cache location. This will be used 
		; to clear the cache when the reference count hits 0.
		; Store in lm_vp_ref_count field
		ld_field_w_clr[lm_vp_ref_count, 1100, temp_var, <<16]
		; Increment reference count and swap out to wait for SRAM
		; read completion. Also wait for signal from previous thread
		; to enter next phase of the pipeline. Update in lm_vp_ref_count
		alu[lm_vp_ref_count, lm_vp_ref_count, +, 1]		

	; This is the local memory cache lookup missed case 
	; wherein the OAM PM counters are to be cached by this thread
	; Store counters read from SRAM into LM

	; Extract the TUCC counters accumulated thus far for this VP
	ld_field_w_clr[lm_vp_tucc0, 0011, $$oam_tx_ctrs[0]]			; lm_vp_tucc0

	br[END#], defer[2]
		; Block size is computed as 2^txN. where txN ranges from 7 to 15
		; for block sizes from 128 to 32768. Update TxN and VP counter
		; options in lm_vp_bedc_txN_mcsn
		; Load the current MCSN value for this F4 flow in lm_vp_bedc_txN_mcsn
		; Upload the BIP-16 BEDC value accumulated 
		; thus far for this VP in lm_vp_bedc_txN_mcsn
		ld_field_w_clr[lm_vp_tucc0_1, 0011, $$oam_tx_ctrs[0], >>16]	; lm_vp_tucc0_1
		ld_field_w_clr[lm_vp_bedc_txN_mcsn, 1111, $$oam_tx_ctrs[1]]

PM_CACHE_HIT#:
	; Incr. ref count in lm_vp_ref_count
	alu[lm_vp_ref_count, lm_vp_ref_count, +, 1]	

NO_VP_PM_ENABLED#:

	.set vp_ctr_base vp_ctr_offset

    ; Phase 3 completed. 
	; Wait for previous thread to complete phase 1; Also wait for
	; any SRAM reads launched for reading in PM counters on cache miss
	ctx_arb[prev_thread_done_sig]

END#:
	.io_completed sram_read_sig
.end
#endm //_aal5_oam_tx_load_F4_pm_ctrs
//------------------------------------------------------------------
// _aal5_oam_tx_update_pm_ctrs()
//
//    Description: Macro to update F4 or F5 PM counters depending
//                 on options enabled. Generate FPM cell if required.
//
//    Parameters:
//				in_vcq	-	VCQ# on which cell being transmitted
//
//------------------------------------------------------------------
#macro _aal5_oam_tx_update_pm_ctrs(in_vcq)
.begin
	.reg	vci						; VCI of cell to transmit
	.reg	pti						; PTI of cell to transmit
	.reg	dram_addr				; DRAM address of delayed bfr
	.sig 	dram_write_sig 			; Sig. for DRAM write done
	.sig	tbuf_write_sig			; Sig. for FPM cell write to MSF
	
	ld_field_w_clr[vci, 0011, atm_cell_copy_header, >>4]

	; Check if this is a F4 OAM cell. If so bypass the
	; OAM Tx processing since F4 cells are not user cells for 
	; any flow and no OAM PM processing is needed for these.
	; F4 OAM SEG OAM cell ?
	alu[--, vci, -, F4_SEG_OAM_VCI_VAL]
	beq[END_PM_CTR_UPDATE#]

	; F4 OAM ETE OAM cell ?
	alu[--, vci, -, F4_ETE_OAM_VCI_VAL]
	beq[END_PM_CTR_UPDATE#]

	; VP or VC counters have been loaded into local memory
	; Verify if counters need update. Note. The assumption
	; is that either F4 or F5 PM is enabled for a VCC and
	; not both. In line with bellcore recommendations 
	; [GR-1248-Core] (Section 7.1.3.1 p.7-26) 

	; Check if VP OAM PM enabled. If F5 PM is enabled for the
	; VC then no F4 PM can be enabled since as stated above,
	; simulatenous PM on F4 and F5 flows is not requires.
	alu[--, F5_PM_IND_BITS, AND, lm_vc_txN_opts,>>16]
	beq[NO_VC_PM_ENABLED#]

	alu[pti , 0x7, and, atm_cell_copy_header, >>1]	; Extrct PTI

	; F5 PM is enabled for this VCC. However counter should only be 
	; updated for user cells of F5 flow. Check if the cell is a VC 
	; OAM cell. VC OAM cells are not user cells for F5 flow. Skip 
	; counter updations for these cells.
	br=byte[pti, 0, F5_ETE_OAM_PTI_VAL, END_PM_CTR_UPDATE#]
	br=byte[pti, 0, F5_SEG_OAM_PTI_VAL, END_PM_CTR_UPDATE#]

	; Check if there is already a buffered cell for this VC. If so skip
	; PM cell generation since only one cell can be buffered at any time
	; Check lm_vc_del_bfr
	alu[--, --, B, lm_vc_del_bfr]	
	bne[UPD_VC_PM_CTRS#]

	; Check if this is the last thread in this critical section.
	; Only the last thread will actually generate the PM cell.
	; This allows maintainence of the cell order without requiring
	; the threads to load buffered cell twice. The last thread just
	; needs to buffer the current cell and generate a PM cell on
	; the port. Check reference count in lm_vc_ref_cnt_flgs									
	br!=byte[lm_vc_ref_cnt_flgs, 0, 1, UPD_VC_PM_CTRS#]					
	
	; Verify if the block size has been reached. If so transmit
	; a PM cell instead of the user cell. The user cell is delayed 
	; and placed in the delayed buffer ptr in this VCQ
	.begin
		.reg block_size
		alu[--, lm_vc_txN_opts, AND, 0xFF]			; Setup indirect ref val
		alu[block_size, --, B, 1,<<indirect]		; block size = 2^lm_txN
		alu[--, block_size, -, lm_vc_tucc0_1]
	.end

	; Check if block size number of cells have 
	; been sent out on the monitored VC
	; if ( block_size > lm_tucc0_1 )
	; 	dont generate PM cell. 
	ble[GEN_PM_CELL#]
	; Block size has not reached. Update the 
	; VC PM counters depending on the CLP bit.

UPD_VC_PM_CTRS#:
	
	; Update current BEDC using lm_vc_bedc
	alu[cur_bedc, cur_bedc, XOR, lm_vc_bedc]

	br_bset[atm_cell_copy_header, 0, end_pm_ctr_update#], defer[2]
		ld_field[lm_vc_bedc, 0011, cur_bedc]		; Update lm_vc_bedc	
		alu[lm_vc_tucc0_1, lm_vc_tucc0_1, +, 1]		; lm_vc_tucc0_1 
    br[end_pm_ctr_update#], defer[1]
		alu[lm_vc_tucc0, lm_vc_tucc0, +, 1]			; lm_vc_tucc0(CLP==0)

NO_VC_PM_ENABLED#:
	; Check if F4 PM is enabled for this VC.
	alu[--, F4_PM_IND_BITS, AND, lm_vc_txN_opts,>>16]
	beq[END_PM_CTR_UPDATE#]

	; Check if there is already a buffered cell for this VC. If so skip
	; PM cell generation since only one cell can be buffered at any time
	; Check lm_vc_del_bfr
	alu[--, --, B, lm_vc_del_bfr]	
	bne[UPD_VP_PM_CTRS#]

	; Check if this is the last thread in this critical section.
	; Only the last thread will actually generate the PM cell.
	; This allows maintainence of the cell order without requiring
	; the threads to load buffered cell twice. The last thread just
	; needs to buffer the current cell and generate a PM cell on
	; the port. Check reference count in lm_vc_ref_cnt_flgs									
	br!=byte[lm_vp_ref_count, 0, 1, UPD_VP_PM_CTRS#]				

	.begin
		.reg block_size
		alu[block_size, --, B, lm_vp_bedc_txN_mcsn, >>8]
		alu[--, block_size, AND, 0xFF]					; Setup indirect ref
		alu[block_size, --, B, 1,<<indirect]			; block = 2^lm_txN
		alu[--, block_size, -, lm_vp_tucc0_1]
	.end
	ble[GEN_PM_CELL#]

UPD_VP_PM_CTRS#:	
	alu[cur_bedc, --, B, cur_bedc, <<16]		
	br_bset[atm_cell_copy_header, 0, END_PM_CTR_UPDATE#], defer[3]
		alu[cur_bedc, cur_bedc, XOR, lm_vp_bedc_txN_mcsn]
		ld_field[lm_vp_bedc_txN_mcsn, 1100, cur_bedc]	; Update BEDC
		alu[lm_vp_tucc0_1, lm_vp_tucc0_1, +, 1]			; Incr. lm_vp_tucc0_1
	br[END_PM_CTR_UPDATE#], defer[1]
	    alu[lm_vp_tucc0, lm_vp_tucc0, +, 1]				; lm_vp_tucc0(CLP=0)
		
GEN_PM_CELL#:
	
	; Allocate a buffer to form the FPM cell.
	dl_buf_alloc($$oam_tx_ctrs[0], BUF_FREE_LIST0,\
				dram_write_sig, dram_write_sig)

	; If no buffer avail, try again. Store fetched buffer
	; in lm_vc_del_bfr field
	alu[lm_vc_del_bfr, --, B, $$oam_tx_ctrs[0]]
	beq[GEN_PM_CELL#]						
	
	; Get DRAM address of buffer handle in lm_vc_del_bfr
	dl_buf_get_data[dram_addr, lm_vc_del_bfr]

	; Determine DRAM offset to write from flip flop bit in lm_vc_txN_opts
	br_bclr[lm_vc_txN_opts, BFRD_CELL_FLIP_FLOP_BT, WR_DRAM#], defer[3]
	
		; Save the TCW payload offset of the buffered cell in lm_vc_txN_opts
		ld_field[lm_vc_txN_opts, 0010, tcw_payload_offset]

		; Save the BEDC value computed for the delayed buffer(from lm_vc_bedc)
		ld_field[lm_vc_bedc, 1100, cur_bedc, <<16]

		; Save the ATM header of the buffered cell in lm_vc_del_bfr_hdr
		ld_field_w_clr[lm_vc_del_bfr_hdr, 1111, atm_cell_copy_header]
	
	; Dram address offset 1 i.e 64 bytes to be used
	alu[dram_addr, dram_addr, +, CELL_1_OFFSET]

WR_DRAM#:

	; Save the current cell in the delayed buffer.
	dram[write, $$prepend_header, dram_addr, 0, LWCOUNT8],\
								 sig_done[dram_write_sig]

	; Buffer current cell and mark the buffer handle as NULL
	; Mark delayed buffer as changed in lm_vc_ref_cnt_flgs
	alu[lm_vc_ref_cnt_flgs, lm_vc_ref_cnt_flgs, OR,\
					 1, <<OAM_TX_DEL_BFR_CH_BT_PS]

	; Generate the PM cell to be transmitted in $$cell_out0
	; to $$cell_out12. Store the current cell descriptor in
	; delayed descriptor for the VC
	xbuf_alloc($pm_cell, (LWCOUNT8 + LWCOUNT6), write)

	; $pm_cell[13] is not used but needs to be declared for 
	; quad word alignment since a write64 is used to write
	; the FPM cell to MSF

	.set $pm_cell[13]	

	_aal5_oam_tx_gen_pm_cell($pm_cell, tbuf_write_sig)

	; Clear the local memory counters since one block of cell
	; processing is complete and the counters need to start 
	; afresh for the next cell block. Check if F4 PM enabled 
	; in lm_vc_txN_opts
	alu[--, F5_PM_IND_BITS, AND, lm_vc_txN_opts, >>16]
	beq[CLR_VP_PM_CTRS#]
	
	; Zero out the lm_vc_tucc0_1, lm_vc_tucc0 and lm_vc_bedc
	immed[lm_vc_tucc0_1, 0]	
	br[INSERT_FLAG_DESCR#], defer[2]
		immed[lm_vc_tucc0, 0]
		immed_w0[lm_vc_bedc, 0]

CLR_VP_PM_CTRS#:
	; Zero out the lm_vp_tucc0_1, lm_vp_tucc0 and lm_vp_bedc
	immed[lm_vp_tucc0_1, 0]
	immed[lm_vp_tucc0,   0]
	immed_w1[lm_vp_bedc_txN_mcsn, 0x00]

INSERT_FLAG_DESCR#:
	; Wait for completion of FPM cell write to MSF and 
	; the delayed cell write to delayed DRAM buffer.
	ctx_arb[dram_write_sig, tbuf_write_sig]

	; Insert the flag descriptor into QM input to allow the buffered cell
	; i.e. cell which was destined to be transmitted in this slot to be pushed
	; out. This will allow the VC flow to be conformant to the established
	; traffic contract even with performance management enabled.
	_aal5_oam_tx_insert_flag_descr(in_vcq)

END_PM_CTR_UPDATE#:
	xbuf_free($pm_cell)

	.io_completed dram_write_sig tbuf_write_sig
.end
#endm //_aal5_oam_tx_update_pm_ctrs
//------------------------------------------------------------------
// _aal5_oam_tx_write_pm_ctrs()
//
//    Description: Macro to flush F4 or F5 PM counters depending
//                 on options enabled. When F4 counters are flushed,
//				   if a delayed buffer was set/cleared on the VC on
//				   which the cell was transmitted, then the VC counters
//				   for that VC are also flushed.
//
//    Parameters:
//				in_vp_ctr_base		- F4 counter table base address
//				in_vp_ctr_offset	- F4 counter entry offset
//				in_vc_ctr_base		- F5 counter table base address
//				in_vc_ctr_offset	- F5 counter entry offset
//------------------------------------------------------------------
#macro _aal5_oam_tx_write_pm_ctrs(in_vp_ctr_base, in_vp_ctr_offset,\
								  in_vc_ctr_base, in_vc_ctr_offset)
.begin
	.reg	temp				; Temporary variable
	.reg	sig_wait			; Signals to wait on
	.sig	vp_sram_write		; Signal for VP OAM Tx Table write comp
	.sig	vc_sram_write		; Signal for VC OAM Tx Table write comp
	.reg	sram_base			; SRAM base address for counter table
	.reg	sram_offset			; SRAM entry offset for counter

	; Time to flush the counters. Depending on whether F4 or F5 flows
	; were enabled, flush the cached counters to either VP or VC OAM Tx Table
	; The base address and offset are already passed into this macro

	; Check if F5 flow counters need to be synched back to SRAM.
	; by checking F5 PM indicator bits in lm_vc_txN_opts
	alu[--, F5_PM_IND_BITS, AND, lm_vc_txN_opts, >>16]
	beq[CHECK_F4_CTRS#]

	; Decrement the LM reference count. If count 
	; reaches zero then flush the counters back
	; to SRAM as required. Use lm_vc_ref_cnt_flgs
	alu[lm_vc_ref_cnt_flgs, lm_vc_ref_cnt_flgs, -, 1]
	br!=byte[lm_vc_ref_cnt_flgs, 0, 0, no_flush_to_sram#], defer[2]

	; Load the base address and offset where 
	; the counters are to be flushed.
	alu[sram_base, --, B, in_vc_ctr_base]
	alu[sram_offset, --, B, in_vc_ctr_offset]

	; F5 flow enabled. Store counters in transfer
	; registeres and write back to SRAM.
	ld_field_w_clr[temp, 0011, lm_vc_tucc0]				; lm_vc_tucc0
	alu[$oam_tx_ctrs[0], temp, OR, lm_vc_tucc0_1, <<16]	; lm_vc_tucc0_1

	ld_field_w_clr[temp, 0001, lm_vc_mcsn]				; lm_vc_mcsn
	ld_field[temp, 0010, lm_vc_txN_opts, <<8]			; lm_vc_txN_opts
	alu[$oam_tx_ctrs[1], temp, OR, lm_vc_bedc, <<16]		; lm_vc_bedc

	; Write back delayed cells, buffer descriptor
	alu[$oam_tx_ctrs[2], --, B, lm_vc_del_bfr]			; lm_vc_del_bfr

	; Write back the payload offset and options and flip flop bit
	; of the delayed buffer stored in lm_vc_txN_opts
	ld_field_w_clr[temp, 0011, lm_vc_txN_opts, >>8]		; lm_vc_txN_opts
	; Write back the BEDC computed on the delayed buffer 
	ld_field[temp, 1100, lm_vc_bedc]					; lm_vc_bedc
	alu[$oam_tx_ctrs[3], --, B, temp]

	; Write back the ATM header of the delayed buffer
	alu[$oam_tx_ctrs[4], --, B, lm_vc_del_bfr_hdr] 		; lm_vc_del_bfr_hdr

	sram[write, $oam_tx_ctrs[0], sram_base,\
	   	 sram_offset, LWCOUNT5], sig_done[vc_sram_write]
	
	ld_field_w_clr[temp, 0011, lm_vc_ref_cnt_flgs,>>16]	; lm_vc_ref_cnt_flgs

	; Clear the cache tag since no thread needs it anymore
	_aal5_oam_tx_clear_tag(temp, 1)
	
	ctx_arb[vc_sram_write], br[END#]	 	
	
CHECK_F4_CTRS#:
	
	; Check the counter types to flush
	alu[--, F4_PM_IND_BITS, AND, lm_vc_txN_opts,>>16]
	beq[NO_FLUSH_TO_SRAM#], defer[1]
		immed[sig_wait, 0]

	; Decrement reference count for this counter cache
	alu[lm_vp_ref_count, lm_vp_ref_count, -, 1]

	; If VP reference count is zero, the VP counters
	; are flushed to SRAM
	br!=byte[lm_vp_ref_count, 0, 0, CHECK_DELAYED_BFR#], defer[2]

	; Load base address where the counters are to be flushed
	alu[sram_base, --, B, in_vp_ctr_base]
	alu[sram_offset, --, B, in_vp_ctr_offset]

	; Flush counters to VP OAM Tx table
	; LW0 -> TUCC0 + TUCC0+1
	ld_field_w_clr[temp, 0011, lm_vp_tucc0]				; lm_vp_tucc0
	alu[$oam_tx_ctrs[0], temp, OR, lm_vp_tucc0_1, <<16]	; lm_vp_tucc0_1

	; LW1 -> MCSN + txN + BEDC
	alu[$oam_tx_ctrs[1], --, B, lm_vp_bedc_txN_mcsn]
	sram[write, $oam_tx_ctrs[0], sram_base,\
	   	 sram_offset, LWCOUNT2], sig_done[vp_sram_write]
	; Mark this SRAM write as an event to wait for
	alu[sig_wait, sig_wait, OR, 1, <<(&vp_sram_write)]

	; Clear the LM cache tag for this VP	
	ld_field_w_clr[temp, 0011, lm_vp_ref_count,>>16]
	_aal5_oam_tx_clear_tag(temp, 0)

CHECK_DELAYED_BFR#:

	; If the state of the buffered cell has changed - write back
	; the buffered cell also. The bfrd cell needs to be written
	; back to the VC OAM Tx Table for the VCQ from which cell 
	; slot to send PM cell was stolen. 	
	; The VC delayed buffer has to be flushed back to SRAM 
	; if the reference count has gone to zero AND the delayed
	; buffer has changed since it was read in from SRAM.

	; Decrement reference counte for the VC counters record.
	; in lm_vc_ref_cnt_flgs
	alu[lm_vc_ref_cnt_flgs, lm_vc_ref_cnt_flgs, -, 1]

	; Check if buffered cell has changed since read from SRAM
	br_bclr[lm_vc_ref_cnt_flgs, OAM_TX_DEL_BFR_CH_BT_PS, CHECK_TAG_UPDATE#]

	; The buffered cell has changed, however the buffered cell is 
	; flushed back to SRAM only if the reference count has gone to
	; zero. If reference count is non zero, then just wait for VP
	; counters flush to SRAM if SRAM write performed.

	br!=byte[lm_vc_ref_cnt_flgs, 0, 0, NO_TAG_UPDATE#], defer[2]
		; load SRAM base address and offset where
		; to flush the delayed buffer.
		alu[sram_base, --, B, in_vc_ctr_base]
		alu[sram_offset, --, B, in_vc_ctr_offset]		

	; Buffered cel status did change and reference count is  0. 
	; Flush the delayed buffer into VC counters block. Only the 
	; delayed buffer is of interest since VP and VC OAM cant be
	; simultaneously enabled. Hence set other counters are dont
	; care even though written though written back to SRAM
	; Load lm_vc_del_bfr

	; Write back delayed cells, buffer descriptor
	alu[$oam_tx_ctrs[2], --, B, lm_vc_del_bfr]			; lm_vc_del_bfr

	; Write back the payload offset, options and flip flop
	; bit of the delayed buffer stored in lm_vc_txN_opts
	ld_field_w_clr[temp, 0011, lm_vc_txN_opts, >>8]		; lm_vc_txN_opts

	; Write back the BEDC computed on the delayed buffer 
	ld_field[temp, 1100, lm_vc_bedc]					; lm_vc_bedc
	alu[$oam_tx_ctrs[3], --, B, temp]

	; Write back the ATM header of the delayed buffer
	alu[$oam_tx_ctrs[4], --, B, lm_vc_del_bfr_hdr] 		; lm_vc_del_bfr_hdr

	.set $oam_tx_ctrs[0] $oam_tx_ctrs[1]

	sram[write, $oam_tx_ctrs[0], sram_base,\
	   	 sram_offset, LWCOUNT5], sig_done[vc_sram_write]
	; Mark this SRAM write as an event to wait for completion
	alu[sig_wait, sig_wait, OR, 1, <<(&vc_sram_write)]

CHECK_TAG_UPDATE#:

	; Check if tag update is needed. Only if reference count is 0
	br!=byte[lm_vc_ref_cnt_flgs, 0, 0, NO_TAG_UPDATE#]
						
	; The VC counters have been flushed out. Clear the cache tag
	ld_field_w_clr[temp, 0011, lm_vc_ref_cnt_flgs, >>16]
	
	_aal5_oam_tx_clear_tag(temp, 1)

NO_TAG_UPDATE#:
	; Any events to wait for completion ?
	alu[--, --, B, sig_wait]
	beq[NO_FLUSH_TO_SRAM#]

	; Wait for all SRAM write operations to complete.
	ctx_arb[--], defer[1]	
		local_csr_wr[active_ctx_wakeup_events, sig_wait]

	; no counters to flush
NO_FLUSH_TO_SRAM#:
end#:
	.io_completed vp_sram_write vc_sram_write
.end
#endm //_aal5_oam_tx_write_pm_ctrs
//------------------------------------------------------------------
// _aal5_oam_tx_gen_pm_cell()
//
//    Description: Generate FPM cell for either F4 or F5 flow
//				   - Update with accumulated counters
//				   - Update CRC and transmit to MSF.
//    Parameters:
//				$in_cell_lw			- SRAM xfer register for FPM cell
//				in_tbuf_write_sig	- Signal to wait for MSF write
//------------------------------------------------------------------
#macro 	_aal5_oam_tx_gen_pm_cell($in_cell_lw, in_tbuf_write_sig)
.begin
	.reg atmPload1				; To store payload LW1 temporarily
	.reg atmPload2				; To store payload LW2 temporarily
	.reg temp_var				; Temporary variable

	; Generate the OAM cell in the $atm_cell transfer registers
	; The transfer registers already contains the cell which was
	; originally supposed to have gone out in this slot.

	.begin
		.reg vci 				; VCI of cell to transmit
		.reg vpi 				; VPI of cell to transmit
		.reg pti				; Payload type indicator

		; Extract VPI, VCI and PTI from ATM header cell which
		; was originally scheduled for transmission in this slot


		; Determine type (F4/F5) of OAM FPM cell to generate

		br_bclr[lm_vc_txN_opts, F5_ETE_PM_IND_BIT_POS, CHK_VC_SEG#], defer[3]
			ld_field_w_clr[vci, 0011, atm_cell_copy_header, >>4]
			alu[vpi, --, b, atm_cell_copy_header, >>20]
			alu[pti, 0x07, AND, atm_cell_copy_header, >>1]

		; F5 ETE FPM Cell to be generated.
		; Same VCI/VPI; PTI change to 0x05
		br[BUILD_F5_PM_CELL_BODY#], defer[2]
			alu[vci, 0x0A, OR, vci, <<4]			
			alu[$in_cell_lw[WRD_OFFSET_0], vci, OR, vpi, <<20]
		
CHK_VC_SEG#:

		br_bclr[lm_vc_txN_opts, F5_SEG_PM_IND_BIT_POS, CHECK_VP_ETE#]

		; F5 SEG FPM Cell to be generated.
		; Same VPI/VCI; PTI change to 0x04
		alu[vci, 0x08, OR, vci, <<4]			
		alu[$in_cell_lw[WRD_OFFSET_0], vci, OR, vpi, <<20]

BUILD_F5_PM_CELL_BODY#:

		; Build LW1 of OAM cell. 
		ld_field_w_clr[atmPload1, 1000, OAM_FPM_CELL_TYPE, <<24]
		ld_field[atmPload1, 0100, lm_vc_mcsn, <<16]
		ld_field[atmPload1, 0011, lm_vc_tucc0_1]
		; Increment MCSN 
		alu[lm_vc_mcsn, lm_vc_mcsn, +, 1]
		
		br[compute_crc#], defer[2]
			; Build LW2 of OAM cell. 	
			ld_field_w_clr[atmPload2, 1100, lm_vc_bedc, <<16]
			ld_field[atmPload2, 0011, lm_vc_tucc0]
	
CHECK_VP_ETE#:
		br_bclr[lm_vc_txN_opts, F4_ETE_PM_IND_BIT_POS, CHECK_VP_SEG#]

		; F4 ETE FPM Cell to be generated.Same VPI/PTI. VCI change to 0x04
		br[build_f4_pm_cell_body#], defer[2]
			alu[vpi, 0x40, OR, vpi, <<20]
			alu[$in_cell_lw[WRD_OFFSET_0], vpi, OR, pti, <<1]

CHECK_VP_SEG#:
		; F4 SEG FPM Cell to be generated. Same VPI/PTI; VCI change to 0x03
		alu[vpi, 0x30, OR, vpi, <<20]
		alu[$in_cell_lw[WRD_OFFSET_0], vpi, OR, pti, <<1]

BUILD_F4_PM_CELL_BODY#:
		; Build LW1 of OAM cell. 
		ld_field_w_clr[atmPload1, 1000, OAM_FPM_CELL_TYPE, <<24]
		ld_field[atmPload1, 0100, lm_vp_bedc_txN_mcsn, <<16]
		ld_field[atmPload1, 0011, lm_vp_tucc0_1]
	
		; Build LW2 of OAM cell. 	
		ld_field_w_clr[atmPload2, 1100, lm_vp_bedc_txN_mcsn]
		ld_field[atmPload2, 0011, lm_vp_tucc0]

		; Increment MCSN 
		; lm_mcsn = (lm_mcsn + 1) % MAX_MCSN_VALUE		
		ld_field_w_clr[temp_var, 0001, lm_vp_bedc_txN_mcsn]
		alu[temp_var, temp_var, +, 1]
		ld_field[lm_vp_bedc_txN_mcsn, 0001, temp_var]
	.end

	; $outAtmCell0                  - ATM header             - 4 bytes
	; $outAtmCell1 to $outAtmCell3  - FPM cell data          - 12 bytes
	; $outAtmCell4 to $outAtmCell12 - filled with 0x6A6A6A6A - 36 bytes
	; HEC byte will be inserted by the PHY.

COMPUTE_CRC#:

	;Initialize CRC remainder to 0 - Need 3 cycles to take effect
	crc_write(0)

	alu[$in_cell_lw[WRD_OFFSET_1], --, B, atmPload1]
	alu[$in_cell_lw[WRD_OFFSET_2], --, B, atmPload2]	

	; CRC computation library messes up the active_lm_addr_0 register
	; The current register value is saved here and is reloaded at the
	; end of the CRC counter update completion to allow correct flush
	; of counters after PM cell generation to SRAM.

	local_csr_rd[active_lm_addr_0]
	immed[temp_var, 0]

	; Compute CRC-10 on payload LW1
	crc_10(crcResult, atmPload1, big_endian, 0, 3)		

	; Compute CRC-10 on payload LW2
	crc_10(crcResult, atmPload2, big_endian, 0, 3)

	; TSTP (32 bits;LW3) - Set to 0xFFFFFFFF
	alu[$in_cell_lw[WRD_OFFSET_3], --, ~B, 0]	
	
	; _crc_10_fcs is an internal function of the CRC library
	; however since the byte values and the actual bytes are
	; know a priori, using this function leads to a saving of
	; 18 cycles for each set of 4 bytes. In this case, this
	; saves (9)	* 18 = 162 cycles
	.reg	i
	alu[i, --, B, &$in_cell_lw[WRD_OFFSET_4], <<2]	; Get the type address
	alu[i, i, OR, thread_num, <<6]				; 16 registers per context
	local_csr_wr[T_INDEX, i]

	; Let the assembler know that $in_cell_lw[4] - $in_cell_lw[11] will
	; be initialized using the T_INDEX pointer
	.set	$in_cell_lw[WRD_OFFSET_4],  $in_cell_lw[WRD_OFFSET_5]
	.set	$in_cell_lw[WRD_OFFSET_6],  $in_cell_lw[WRD_OFFSET_7] 
	.set	$in_cell_lw[WRD_OFFSET_8],  $in_cell_lw[WRD_OFFSET_9]
	.set	$in_cell_lw[WRD_OFFSET_10], $in_cell_lw[WRD_OFFSET_11]

	; Initialize loop variable
	immed[i, 0x00]

	.repeat
		_call_crc_10_fcs(0xFF)
		alu[i, i, +, 1]
	.until(i >= 4)

	; Next 8 Long words have filler bytes 0x6A
	immed32[atmPload1, 0x6A6A6A6A]
	
	; Initialize LW 4 to 11 with filler bytes 0x6A
	; and compute CRC over these bytes.
	immed[i, 0x04]
	.repeat
		; Copy filler bytes to outgoing OAM cell
		alu[*$index++, --, B, atmPload1]
		#define_eval j	0
		#while (j < 4)
			; Compute CRC-10 on each fill byte
			_call_crc_10_fcs(0x6A)
		#define_eval j (j + 1)
		#endloop
		#undef j
		; Increment loop variable
		alu[i, i, +, 1]
	.until (i >= 12)

	; Last LW has two upper bytes as filler bytes
	; and CRC-10 remainder in two lower bytes
	_call_crc_10_fcs(0x6A)
	_call_crc_10_fcs(0x6A)

	; Consider the last 6 reserved bits of FPM cell in CRC

	_call_crc_10_fcs(0x00)
	_call_crc_10_fcs(0x00)

	; restore the active_lm_addr_0 pointer to value
	; before initiating CRC-10 operation.
	local_csr_wr[active_lm_addr_0, temp_var]

	; Update cell with CRC-10 value
	immed[atmPload1, 0x6A6A]
	alu[$in_cell_lw[WRD_OFFSET_12], crcResult, OR, atmPload1, <<16]
	.begin
		.reg tbuf_addr_plus_4
		.reg tbuf_addr_mask

		; The global masks pointed to local memory pointer 1
		; cannot be used since local memory pointer 1 is 
		; modified. Recompute the mask to be used based on
		; the PHY mode and compute the tbuf address.
		#if (TX_PHY_MODE == SPHY_1_32)

		; Mask that limits the TBUF address space used to
		; transmit a cell to 8KB.
		immed[tbuf_addr_mask, 0x1FFF]

		#endif // (TX_PHY_MODE == SPHY_1_32)

		#if ((TX_PHY_MODE == SPHY_4_8) || (TX_PHY_MODE == MPHY_4))

		; Mask that limits the TBUF address space used to
		; transmit a cell to 2KB.
		immed[tbuf_addr_mask, 0x7FF]

		#endif // ((TX_PHY_MODE == SPHY_4_8) || (TX_PHY_MODE == MPHY_4))

	 	; Compute the TBUF address of the available TBUF element.
		alu[tbuf_addr_plus_4, tbuf_addr_mask, AND, \
			$avail_tbuf_element, <<TBUF_ELE_SIZE_LOG2]
#ifndef FPGA_HEADER
		alu[tbuf_addr_plus_4, tbuf_addr_plus_4, +, BYTES_PER_LW]	
#else
		alu[$in_cell_lw[0], 0x03, ~AND, txc_port_and_clp, <<1]
#endif
		; Write the OAM PM cell to MSF.
		msf[write64, $in_cell_lw[0], tbuf_element_base,\
			tbuf_addr_plus_4, QWCOUNT7], sig_done[in_tbuf_write_sig]
		immed[tcw_payload_offset, 0]
	.end
.end
#endm //_aal5_oam_tx_gen_pm_cell
//------------------------------------------------------------------
// _aal5_oam_tx_insert_flag_descr()
//
//    Description: Insert a request to QM input ring to request a
//				   cell slot for pushing out the buffered cell
//
//    Parameters:
//				in_vcq				- VCQ on which to request tx
//------------------------------------------------------------------
#macro _aal5_oam_tx_insert_flag_descr(in_vcq)
.begin
	.reg aal5_to_qm_rng 			; QM Input scratch ring
	.sig aal5_to_qm_send_done		; Signal for scratch put

	xbuf_alloc($message, LWCOUNT3, write)	; Message to QM 

	; Compute the ring address by multiplying the base by 4
	alu[aal5_to_qm_rng, --, B, AAL5_TO_QM_RING_OUT, <<2]

	; Check to make sure there is room on the ring (idle loop 
	; till ring has space)
RNG_FULL#:
	br_inp_state[SCR_Ring/**/AAL5_TO_QM_RING_OUT/**/_Full, RNG_FULL#]

	; Send a enqueue message to QM with dummy flag descriptor
	; buffer handle. This causes QM to schedule a slot for the
	; buffered cell on this VC.
	immed32[$message[0], ATM_OAM_FLAG_DESCR]		; Flag message
	alu[$message[1], --, B, IX_NULL]				; EOP->invalid
	alu[$message[2], in_vcq, OR, 1, <<18]			; cell count=1

	; Write the flag descriptor to the scratch ring
	scratch[put, $message[0], 0, aal5_to_qm_rng, LWCOUNT3], \
	    sig_done[aal5_to_qm_send_done]
	; Wait for scratch put to complete
	ctx_arb[aal5_to_qm_send_done]
	xbuf_free($message)
.end
#endm //_aal5_oam_tx_insert_flag_descr
//------------------------------------------------------------------
//	MACRO
//		oam_pm_lookup_lm_pm_cache
//
//	DESCRIPTION 
//		Search through LM tag cache to find if the desired counters 
//		are	already cached by some thread. If so then set the active
//		LM addr 0 pointer to the cached counter block.
//		
//	ARGUMENTS
//		in_cache_tag				- Cache tag to search for
//		ret_lbl						- Label to branch in case of hit
//		PTR							- LM pointer to use (0/1)
//		TYPE						- lookup VC or VP cache ?
//------------------------------------------------------------------
#macro _aal5_oam_tx_lkup_lm_pm_cache(in_cache_tag, ret_lbl, PTR, TYPE)
.begin
	.reg lm_cache_addr					; Cache local memory addr
	
	; Local LM_addr_0<-tagbase
	immed[lm_cache_addr, OAM_TX_/**/TYPE/**/_TAG_BASE]
	local_csr_wr[active_lm_addr_/**/PTR/**/, lm_cache_addr]	

	; Loading CSR needs 3 cycles. Make sure first
	; access is after 3 cycles. Loop through counter
	; cache entries for all previous threads to see
	; if any of them has cached the desired counters.
	; Only need to look for previous threads since threads 
	; ahead of this thread would not have had a 
	; chance to execute yet.
	.begin
		.reg idx							; Case index
		alu[idx, 7, -, thread_num]			; (7 - thread num) * 8	
		alu_shf[idx, --, B, idx, <<3]
		jump[idx, LOOKUP_TABLE_BASE#],
			targets[CTX0#, CTX1#, CTX2#, CTX3#,\
					CTX4#, CTX5#, CTX6#, CTX7#]
	.end

LOOKUP_TABLE_BASE#:
CTX7#:	_aal5_oam_tx_chk_tag(in_cache_tag, 6, ret_lbl, PTR, TYPE)
		nop
CTX6#:	_aal5_oam_tx_chk_tag(in_cache_tag, 5, ret_lbl, PTR, TYPE)
		nop
CTX5#:	_aal5_oam_tx_chk_tag(in_cache_tag, 4, ret_lbl, PTR, TYPE)
		nop
CTX4#:	_aal5_oam_tx_chk_tag(in_cache_tag, 3, ret_lbl, PTR, TYPE)
		nop
CTX3#:	_aal5_oam_tx_chk_tag(in_cache_tag, 2, ret_lbl, PTR, TYPE)
		nop
CTX2#:	_aal5_oam_tx_chk_tag(in_cache_tag, 1, ret_lbl, PTR, TYPE)
		nop
CTX1#:	_aal5_oam_tx_chk_tag(in_cache_tag, 0, ret_lbl, PTR, TYPE)
		nop
CTX0#:	; Cache missed
	.begin
		.reg idx							; Case index
		alu[idx, --, B, thread_num, <<2]	; (7 - thread number) * 4	
		jump[idx, MISS_TABLE_BASE#],
			targets[M0#, M1#, M2#, M3#, M4#, M5#, M6#, M7#]
	.end
MISS_TABLE_BASE#:
M0#:	_aal5_oam_tx_upd_tag_on_miss(in_cache_tag, 0, END#, PTR, TYPE)
M1#:	_aal5_oam_tx_upd_tag_on_miss(in_cache_tag, 1, END#, PTR, TYPE)
M2#:	_aal5_oam_tx_upd_tag_on_miss(in_cache_tag, 2, END#, PTR, TYPE)
M3#:	_aal5_oam_tx_upd_tag_on_miss(in_cache_tag, 3, END#, PTR, TYPE)
M4#:	_aal5_oam_tx_upd_tag_on_miss(in_cache_tag, 4, END#, PTR, TYPE)
M5#:	_aal5_oam_tx_upd_tag_on_miss(in_cache_tag, 5, END#, PTR, TYPE)
M6#:	_aal5_oam_tx_upd_tag_on_miss(in_cache_tag, 6, END#, PTR, TYPE)
M7#:	_aal5_oam_tx_upd_tag_on_miss(in_cache_tag, 7, END#, PTR, TYPE)
END#:
.end
#endm //_aal5_oam_tx_lkup_lm_pm_cache
//------------------------------------------------------------------
//	MACRO
//		_aal5_oam_tx_chk_tag
//
//	DESCRIPTION 
//		Match the cache tag at indicated index. Branch to match lbl
//		if matched. Addittionally load the LM pointer with cache
//		address for that counter type and index
//		
//	ARGUMENTS
//		in_tag				- Cache tag to match
//		CHK_IDX				- Index to match
//		match_lbl			- Label to branch to on match
//		PTR					- LM pointer to use (0/1)
//		TYPE				- lookup VC or VP cache ?
//------------------------------------------------------------------
#macro _aal5_oam_tx_chk_tag(in_tag, CHK_IDX, MATCH_LBL, PTR, TYPE)
.begin
	alu[--, in_tag, -, *l$index/**/PTR/**/[/**/CHK_IDX/**/]]	
	bne[ctx/**/CHK_IDX/**/#]
	immed[lm_cache_addr, OAM_TX_/**/TYPE/**/_CTR_CTX/**/CHK_IDX/**/]
	local_csr_wr[active_lm_addr_/**/PTR/**/, lm_cache_addr]
	br[MATCH_LBL], defer[2]
		nop
		nop
.end
#endm //_aal5_oam_tx_chk_tag
//------------------------------------------------------------------
//	MACRO
//		_aal5_oam_tx_upd_tag_on_miss
//
//	DESCRIPTION 
//		Update the cache tag at indicated index. Branch to match lbl
//		if matched. Addittionally load the LM pointer with cache
//		address for that counter type and index
//		
//	ARGUMENTS
//		in_tag				- Cache tag to match
//		IDX					- Index to match
//		ret_lbl				- Label to branch to on return
//		PTR					- LM pointer to use (0/1)
//		TYPE				- lookup VC or VP cache ?
//------------------------------------------------------------------
#macro	_aal5_oam_tx_upd_tag_on_miss(in_tag, IDX, RET_LBL, PTR, TYPE)
.begin
	; Set the cache tag for this entry
    alu[*l$index/**/PTR/**/[/**/IDX/**/], --, B, in_tag]
	br[RET_LBL], defer[2]
		immed[lm_cache_addr, OAM_TX_/**/TYPE/**/_CTR_CTX/**/IDX/**/]
		local_csr_wr[active_lm_addr_/**/PTR/**/, lm_cache_addr]
.end
#endm //_aal5_oam_tx_upd_tag_on_miss
//------------------------------------------------------------------
//	MACRO
//		_aal5_oam_tx_clear_tag
//
//	DESCRIPTION 
//		Clear the cache tag entry at address passed
//		
//	ARGUMENTS
//		cache_addr			- Cache address to clear
//		PTR					- LM pointer to use (0/1)
//------------------------------------------------------------------
#macro _aal5_oam_tx_clear_tag (cache_addr, PTR)
.begin
	local_csr_wr[active_lm_addr_/**/PTR/**/, cache_addr]	
	nop
	nop
	nop
	alu[*l$index/**/PTR/**/, --, B, 0]
.end
#endm //_aal5_oam_tx_clear_tag
//------------------------------------------------------------------
//	MACRO
//		_aal5_oam_tx_bypass_processing
//
//	DESCRIPTION 
//		Bypass OAM Tx processing
//		
//	ARGUMENTS
//		None.
//------------------------------------------------------------------
#macro _aal5_oam_tx_bypass_processing()
.begin

	; Phase 3
	signal_ctx(next_thread)

	#repeat 10
	nop
	#endloop

	ctx_arb[prev_thread_done_sig]

	; Phase 4
	signal_ctx(next_thread)

	#repeat 10
	nop
	#endloop

	ctx_arb[prev_thread_done_sig]

.end
#endm   //_aal5_oam_tx_bypass_processing
//------------------------------------------------------------------
//	MACRO
//		_call_crc_10_fcs
//
//	DESCRIPTION 
//		Invoke the CRC-10 computation subroutine. To save code store
//		
//	ARGUMENTS
//			d	-	data byte on which to compute CRC
//------------------------------------------------------------------
#macro _call_crc_10_fcs(d)
.begin

	; Store data byte in global register used as parameter for
	; subroutine. This macro serves as a stack push routine,
	; where in parameter and return address are placed on the
	; global GPRs before branching to code segment to compute 
	; CRC. The CRC-10 code segmeng will return to RTN_LABEL#

	alu[data, --, B, d]
	load_addr[return_address, RTN_LABEL#]

	; Jump into the subroutine
	br[STUB_CRC_10_ADDRESS#]

RTN_LABEL#:

.end
#endm
//------------------------------------------------------------------
//	MACRO
//		_stub_crc_10_fcs
//
//	DESCRIPTION 
//		CRC-10 computation subroutine. Includes the CRC-10 compute
//		macro, and returns to address stored in return_address GPR
//		
//	ARGUMENTS
//			crcResult 	- 	Global GPR used to return CRC result
//			data		- 	Global GPR containing data to work on
//------------------------------------------------------------------
#macro	_stub_crc_10_fcs()
.subroutine
	; Expand CRC-10 computation macro
	_crc_10_fcs(crcResult, data)
	; Return to caller
	rtn[return_address]
.endsub
#endm
#endif  // _AAL5_OAM_TX_UC