#ifndef	_AAL5_TX_UTIL_COMMON_UC_
#define _AAL5_TX_UTIL_COMMON_UC_

/*******************************************************************************
                             Intel Proprietary

 Copyright (c) 1998-2002 By Intel Corporation.  All rights reserved.
 No part of this program or publication may be reproduced, transmited,
 transcribed, stored in a retrieval system, or translated into any language
 or computer language in any form or by any means, electronic, mechanical,
 magnetic, optical, chemical, manual, or otherwise, without the prior
 written permission of:
                         Intel Corporation
                         2200 Mission College Blvd.
                         Santa Clara, CA  95052-8119
*******************************************************************************/

/*                                                                   
 *                                                                   
 *      File Name: aal5_tx_util_common.uc
 *                                                                   
 *      Description: This file contains utility macros that are used
 *                   during the processing phase.
 *
 */                                                                   

#include <aal5_tx.h>
#include <aal5_tx_regs.uc>

//------------------------------------------------------------------
// _aal5_tx_compute_curr_dram_indirect_read_format()
//
//    Description: Compute cell payload in current buffer in
//    quad-words taking 8-byte alignment into consideration.
//
//    Parameters:
//        Inputs:
//        in_cell_payload_in_curr_buf:
//            Cell payload in current buffer in bytes.
//
//------------------------------------------------------------------
#macro _aal5_tx_compute_curr_dram_indirect_read_format( \
			in_cell_payload_in_curr_buf)
.begin
	
	#define REF_COUNT_OVERRIDE_BIT 0x10
	
	.reg ref_cnt

	; Add 2 extra lw's (8 bytes) for possibly non 8 byte aligned start
	; of payload. If the start of payload is not 8-byte aligned we
	; might need to read at most 8 additional bytes from DRAM.

	; Convert (in_cell_payload_in_curr_buf + 8) bytes to qw's.
	; Ceil the bytes to the closest qw.
	; i.e. compute (((in_cell_payload_in_curr_buf + 8) - 1) >> 3) + 1
	; Note: We do not need to add the right most '+1' since ref_cnt
	; needs to be one less than the actual value.
	alu[ref_cnt, in_cell_payload_in_curr_buf, +, 7]
	alu[ref_cnt, REF_COUNT_OVERRIDE_BIT, OR, ref_cnt, >>3]

	; This should always be the last line of this macro.
	alu[--, --, B, ref_cnt, <<21]

	#undef REF_COUNT_OVERRIDE_BIT

.end
#endm


//------------------------------------------------------------------
// _aal5_tx_compute_next_dram_indirect_read_format()
//
//    Description: Compute cell payload in next buffer in quad-words.
//    Also compute absolute DRAM read-xfer-reg address that will hold
//    the cell payload in next buffer.
//
//    Parameters:
//        Outputs:
//        out_next_payload_reg_base_lw:
//            Absolute DRAM read-xfer-reg address that will hold
//            the cell payload in next buffer.
//
//        Inputs:
//        in_cell_payload_in_curr_buf:
//            Cell payload in current buffer in bytes.
//        in_cell_payload_in_next_buf:
//            Cell payload in next buffer in bytes.
//
//------------------------------------------------------------------
#macro _aal5_tx_compute_next_dram_indirect_read_format( \
			out_next_payload_reg_base_lw, \
			in_cell_payload_in_curr_buf, \
			in_cell_payload_in_next_buf)
.begin

	#define REF_COUNT_OVERRIDE_BIT		0x10
	#define XFER_REG_ADDR_OVERRIDE_BIT	0x10

	.reg indirect_reg

	.reg ref_cnt

	; Convert in_cell_payload_in_next_buf bytes to qw's.
	; i.e compute ((in_cell_payload_in_next_buf - 1) >> 3) + 1
	; Note: We do not need to add the right most '+1' since ref_cnt
	; needs to be one less than the actual value.
	alu[ref_cnt, in_cell_payload_in_next_buf, -, 1]
	alu[ref_cnt, REF_COUNT_OVERRIDE_BIT, OR, ref_cnt, >>3]

	alu[indirect_reg, XFER_REG_ADDR_OVERRIDE_BIT, OR, ref_cnt, <<21]
	
	; Set absolute d-xfer-read register address.
	; Add 2 extra lw's (8 bytes) for possibly non 8 byte aligned
	; start of payload. Ceil this number to the closest lw.
	; i.e. compute ((((in_cell_payload_in_curr_buf + 8) - 1) >> 3) + 1) * 2
	alu[out_next_payload_reg_base_lw, in_cell_payload_in_curr_buf, +, 7]
	alu[out_next_payload_reg_base_lw, --, B, \
			out_next_payload_reg_base_lw, >>3]
	alu[out_next_payload_reg_base_lw, out_next_payload_reg_base_lw, +, 1]
	alu[out_next_payload_reg_base_lw, --, B, \
			out_next_payload_reg_base_lw, <<1]
	alu[out_next_payload_reg_base_lw, payload_reg_base_addr_lw, +, \
			out_next_payload_reg_base_lw]

	; This should always be the last line of this macro.
	alu[indirect_reg, indirect_reg, OR, \
		out_next_payload_reg_base_lw, <<5]

	; Note: The current value of out_next_payload_reg_base_lw
	; is used in macro align_first_lw_of_cell_payload_in_next_buf().
	; This is done to avoid recomputing next_payload_reg_base_lw.

	#undef REF_COUNT_OVERRIDE_BIT
	#undef XFER_REG_ADDR_OVERRIDE_BIT

.end
#endm


//------------------------------------------------------------------
// Pipestage 1 - phase 4
//------------------------------------------------------------------
// _aal5_tx_read_cell_payload_and_update_txc1()
//
//    Description: Read payload from DRAM.
//                 Update TXC.
//                 Prefetch next buffers' meta data if needed.
//
//    Parameters: None.
//
//------------------------------------------------------------------
#macro _aal5_tx_read_cell_payload_and_update_txc()
.begin

	.set	$$payload[0]  $$payload[1]  $$payload[2]  $$payload[3]  \
			$$payload[4]  $$payload[5]  $$payload[6]  $$payload[7]  \
			$$payload[8]  $$payload[9]  $$payload[10] $$payload[11] \
			$$payload[12] $$payload[13] $$payload[14] $$payload[15]

	signal_ctx(next_thread)

	; Set during sram[read] of the next buffers' meta data.
	.set $prefetch_meta[0] $prefetch_meta[1]

	; Is lm:txc_end_of_payload bit set? If yes, this is the 
	; additional cell sent by the QM; branch to addn_cell#.
	br_bset[*l$index0[5], EOP_BIT, addn_cell#]

cell_with_payload#:

	.reg txc_bytes_done_in_packet
	.reg txc_bytes_rem_in_curr_buf
	.reg txc_next_buf_size
	.reg bytes_rem_in_packet

	; if (txc_bytes_rem_in_curr_buf == 0), switch to prefetched buffer.
	alu[txc_bytes_rem_in_curr_buf, 0, +16, *l$index0[0]]
	bne[prefetch_update_done#], defer[3]

	; Store lm:txc_packet_size in a gpr.
	ld_field_w_clr[txc_packet_size, 0011, *l$index0[6], >>16]

	; Store lm:txc_bytes_done_in_packet in a gpr.
	ld_field_w_clr[txc_bytes_done_in_packet, 0011, *l$index0[4]]

	; Compute bytes remaining in packet.
	alu[bytes_rem_in_packet, txc_packet_size, -, txc_bytes_done_in_packet]

	.reg txc_bytes_done_in_next_buf

	; txc_bytes_done_in_next_buf = lm:txc_bytes_done_in_next_buf
	alu[txc_bytes_done_in_next_buf, 0x3F, AND, *l$index0[2], >>25]

	; lm:txc_curr_buf_hdl = lm:txc_next_buf_hdl
	; lm:txc_bytes_done_in_next_buf = 0
	alu[*l$index0[2], *l$index0[3], AND~, 0xFF, <<24]

	; txc_next_buf_size = lm:txc_next_buf_size
	alu[txc_next_buf_size, --, B, *l$index0[4], >>16]

	; txc_bytes_rem_in_curr_buf = txc_next_buf_size - txc_bytes_done_in_next_buf
	alu[txc_bytes_rem_in_curr_buf, txc_next_buf_size, -, txc_bytes_done_in_next_buf]

	; lm:txc_bytes_rem_in_curr_buf = txc_bytes_rem_in_curr_buf
	ld_field[*l$index0[0], 0011, txc_bytes_rem_in_curr_buf]

	; lm:txc_curr_payload_addr =
	; dl_buf_get_data_from_meta(lm:txc_curr_buf_hdl)
	dl_buf_get_data_from_meta(*l$index0[1], *l$index0[2])

	; lm:txc_new_curr_buf_hdl = 1
	; Note: This instruction comes after the dl_buf_get_data_from_meta()
	; macro call, since the macro call assumes all 32 bits of *l$index0[2]
	; contain the buffer handle.
	alu[*l$index0[2], *l$index0[2], OR, 1, <<24]

	; lm:txc_curr_payload_addr = lm:txc_curr_payload_addr + \
	; txc_bytes_done_in_next_buf
	alu[*l$index0[1], *l$index0[1], +, txc_bytes_done_in_next_buf]

	; lm:txc_next_buf_hdl = lm:txc_next_next_buf_hdl
	alu[*l$index0[3], *l$index0[5], AND~, 0xFF, <<24]

prefetch_update_done#:

	; At this point the current cells' payload starts at DRAM
	; address lm:txc_curr_payload_addr.

	; if ((lm:txc_bytes_rem_in_curr_buf < bytes_rem_in_packet) &&
	;	  lm:txc_new_curr_buf_hdl == 1),
	; prefetch next buffers' meta data.
	alu[--, txc_bytes_rem_in_curr_buf, -, bytes_rem_in_packet]
	bge[check_if_last_cell#], defer[3]

	; lm:txc_bytes_done_in_packet += 48
	alu[txc_bytes_done_in_packet, txc_bytes_done_in_packet, +, 48]
	ld_field[*l$index0[4], 0011, txc_bytes_done_in_packet]

	; cell_payload_in_curr_buf = 48
	immed[cell_payload_in_curr_buf, 48]

	; if (lm:txc_new_curr_buf_hdl == 1),
	; prefetch next buffers' meta data.
	br_bclr[*l$index0[2], 24, check_if_last_cell#]

	; Set the prefetched_buffer flag to 1.
	immed[prefetched_buffer, 1]

	; lm:txc_new_curr_buf_hdl = 0
	alu[*l$index0[2], *l$index0[2], AND~, 0x01, <<24]

	; Add meta_rd_sig to sigmask_a.
	alu[sigmask_a, sigmask_a, OR, 1, <<(&meta_rd_sig)]

	; Prefetch next buffers' meta data from SRAM.
	sram[read, $prefetch_meta[0], *l$index0[3], 0, 2],\
		sig_done[meta_rd_sig]

check_if_last_cell#:

	; if (lm:txc_bytes_done_in_packet >= lm:txc_packet_size), this
	; is the last cell with payload for this packet.
	alu[--, txc_bytes_done_in_packet, -, txc_packet_size]
	blt[check_free_buffer#], defer[2]

	 // Unfilled defer slot.

	; Set curr_payload_addr_offset to the DRAM address quadword offset.
	alu[curr_payload_addr_offset, 0x7, AND, *l$index0[1]]

	; Store output port in a gpr for use in pipestage 2.
	ld_field_w_clr[outport, 0011, *l$index0[0], >>16]


	; lm:txc_end_of_payload = 1
	alu[*l$index0[5], *l$index0[5], OR, 1, <<EOP_BIT]

	; cell_payload_in_curr_buf =
	;	48 + lm:txc_packet_size - lm:txc_bytes_done_in_packet
	alu[cell_payload_in_curr_buf, 48, +, txc_packet_size]
	alu[cell_payload_in_curr_buf, \
			cell_payload_in_curr_buf, -, txc_bytes_done_in_packet]

check_free_buffer#:

	; Store total cell_payload of the cell in a gpr
	; for use in pipestage 2.
	alu[cell_payload, --, B, cell_payload_in_curr_buf]

	; if (cell_payload_in_curr_buf >= txc_bytes_rem_in_curr_buf), 
	; free current buffer.

	; Note that if an additional cell is going to be scheduled for
	; this packet, the eop buffer will not be freed here. Instead
	; it will be freed by the thread that processes the additional
	; cell. We insure this by incrementing the eop buffer size by 1
	; in the LLC encap microblock if an additional cell is needed
	; for the packet. Then, the condition below will not hold true
	; since 1 more byte is remaining in the eop buffer and the buffer
	; will not be freed below. Incrementing the eop buffer size by 1
	; does not affect any functionality.
	alu[cell_payload_in_next_buf, \
			cell_payload_in_curr_buf, - , txc_bytes_rem_in_curr_buf]
	blt[read_curr_payload#], defer[2]

	 // Unfilled defer slot.

	; Store freelist id in register to use in pipestage 2.
	alu[freelist, 0xF, AND, *l$index0[5], >>24]

	; Add curr_buf_rd_sig and curr_buf_rd_sig+1 to sigmask_b.
	alu[sigmask_b, sigmask_b, OR, 0x3, <<(&curr_buf_rd_sig)]


	; cell_payload_in_curr_buf = txc_bytes_rem_in_curr_buf
	alu[cell_payload_in_curr_buf, --, B, txc_bytes_rem_in_curr_buf]

	; This buffer will be freed by this thread after reading the
	; cell payload from it.
	; Store current buffer handle in free_curr_buf in LW's for
	; the dl_buf_free() macro.
	; Also set bit 31 of free_curr_buf. This bit will be tested
	; in pipestage 2 before freeing this buffer.
	ld_field_w_clr[free_curr_buf, 0111, *l$index0[2]]
	alu[free_curr_buf, Ox80000000, OR, free_curr_buf, >>2]

check_split_cell#:

	; if (cell_payload_in_next_buf > 0),
	; current cell is split across buffers.
	alu[--, cell_payload_in_next_buf, -, 0]
	ble[read_curr_payload#]

	; lm:txc_bytes_done_in_next_buf = cell_payload_in_next_buf
	; Note that lm:txc_bytes_done_in_next_buf is 0 until here.
	alu[*l$index0[2], *l$index0[2], OR, cell_payload_in_next_buf, <<25]

	ld_field_w_clr[txc_next_buf_size, 0011, *l$index0[4], >>16]

	; if (txc_next_buf_size == cell_payload_in_next_buf),
	; free next buffer too.

	; Note that if an additional cell is going to be scheduled for
	; this packet, the eop buffer will not be freed here. Instead
	; it will be freed by the thread that processes the additional
	; cell. We insure this by incrementing the eop buffer size by 1
	; in the LLC encap microblock if an additional cell is needed
	; for the packet. Then, the condition below will not hold true
	; since 1 more byte is remaining in the eop buffer and the buffer
	; will not be freed below. Incrementing the eop buffer size by 1
	; does not affect any functionality.
	alu[--, txc_next_buf_size, -, cell_payload_in_next_buf]
	bne[read_next_payload#]

	; Store next buffer handle in free_next_buf in LW's for
	; the dl_buf_free() macro.
	alu[free_next_buf, Ox80000000, OR, *l$index0[3], >>2]

read_next_payload#:

	; Add next_buf_rd_sig and next_buf_rd_sig+1 to sigmask_b.
	alu[sigmask_b, sigmask_b, OR, 0x3, <<(&next_buf_rd_sig)]

	.reg next_payload_addr

	; Compute absolute DRAM address of start of payload in next buffer.
	dl_buf_get_data_from_meta(next_payload_addr, *l$index0[3])

	; Compute cell_payload_in_next_buf in quad-words
	; and absolute d-xfer-read address.
	_aal5_tx_compute_next_dram_indirect_read_format(
		next_payload_reg_base_lw,
		cell_payload_in_curr_buf,
		cell_payload_in_next_buf)

	; Read payload from next buffer.
	dram[read, --, next_payload_addr, 0, max_7], indirect_ref, \
		sig_done[next_buf_rd_sig]

read_curr_payload#:

	; Compute cell_payload_in_curr_buf in quad-words
	; and absolute d-xfer-read address.
	_aal5_tx_compute_curr_dram_indirect_read_format( \
		cell_payload_in_curr_buf)

	; Read payload from current buffer.
	dram[read, $$payload[0], *l$index0[1], 0, max_7], indirect_ref, \
		sig_done[curr_buf_rd_sig]

	; Update TXC1.

	; lm:txc_bytes_rem_in_curr_buf -= cell_payload_in_curr_buf
	alu[*l$index0[0], *l$index0[0], -, cell_payload_in_curr_buf]

	; lm:txc_curr_payload_addr += cell_payload_in_curr_buf
	alu[*l$index0[1], *l$index0[1], +, cell_payload_in_curr_buf]

	; ctx_arb on sigmask_a.
	ctx_arb[--], defer[2], br[end#]

	local_csr_wr[active_ctx_wakeup_events, sigmask_a]	

	; Init sigmask_a with prev_thread_done_sig for use the next time.
	alu[sigmask_a, --, B, 1, <<(&prev_thread_done_sig)]

addn_cell#:

	; if (end_of_payload == 1), free buffer $buf_hdl_lw.
	; This is the additional cell dequeue request. The
	; associated buffer is freed in this iteration.
	alu[free_curr_buf, Ox80000000, OR, $buf_hdl_lw]	

	; Store packet_size in a gpr.
	ld_field_w_clr[txc_packet_size, 0011, *l$index0[6], >>16]

	; Store lm:txc_freelist in a gpr for use in pipestage 2.
	alu[freelist, 0xF, AND, *l$index0[5], >>24]

	; Set gpr's for additional cell path.
	immed[cell_payload, 0]
	immed[cell_payload_in_curr_buf, 0]
	immed[cell_payload_in_next_buf, 0]

	ctx_arb[prev_thread_done_sig], defer[2]

	immed[curr_payload_addr_offset, 0]

	; Store output port in gpr to use in pipestage 2.
	ld_field_w_clr[outport, 0011, *l$index0[0], >>16]

end#:
	.io_completed meta_rd_sig
	.io_completed prev_thread_done_sig	

.end
#endm


//------------------------------------------------------------------
// _aal5_tx_tbuf_full()
//
//    Description: Stalls until TBUF elements become available.
//
//    Parameters: None.
//
//------------------------------------------------------------------
#macro _aal5_tx_tbuf_full()
.begin

	.reg cells_pending, cells_tx_by_fabric_masked

	xbuf_alloc($cells_tx_by_fabric, 1, read)

check_tx_seq#:

	; Read number of cells actually transmitted from this partition.
	msf[read, $cells_tx_by_fabric[0], *l$index1[TX_SEQUENCE_INDEX], \
		partition_number, 1], sig_done[tx_sequence_sig]

waiting_for_tx_seq_rd_sig#:
	br_!signal[tx_sequence_sig, waiting_for_tx_seq_rd_sig#]	

	// Mask-out the highest bit of the sequence word (sequence number is set on lowest 8 bits)
	alu[cells_tx_by_fabric_masked, $cells_tx_by_fabric[0], AND, TX_SEQUENCE_MASK]

	// Check if we have a wrap-round case for the cells transmit
	.if (cells_tx_by_code >= cells_tx_by_fabric_masked)
		alu[cells_pending, cells_tx_by_code, -, cells_tx_by_fabric_masked]
	.else
        .begin
			.reg cells_tx_by_code_corrected

			alu[cells_tx_by_code_corrected, cells_tx_by_code, +, TBUF_ELE_NUM]
			alu[cells_pending, cells_tx_by_code_corrected, -, cells_tx_by_fabric_masked]
		.end						
	.endif

	.if (cells_pending >= TBUF_THRESHOLD)
		br[check_tx_seq#]
	.endif

	xbuf_free($cells_tx_by_fabric)

.end
#endm


//------------------------------------------------------------------
// _aal5_inc_counter()
//
//    Description: Increments the specified counter in SRAM.
//
//    Parameters: None.
//        Inputs:
//        COUNTER_TYPE:
//            Name of the counter to increment.
//        in_index:
//            Index at which the counter resides within COUNTER_TYPE.
//
//------------------------------------------------------------------
#ifdef COUNTERS
#macro _aal5_inc_counter(COUNTER_TYPE, in_index)
.begin

	.reg index, baseaddr
    .set baseaddr

#if streq ('COUNTER_TYPE','CELLS_PER_OUTPORT')
    immed32[baseaddr, AAL5_TX_CELLS_PER_OUTPORT_COUNTER_SRAM_BASE]
#endif

#if streq ('COUNTER_TYPE','CELLS_PER_VCQ')
    immed32[baseaddr, AAL5_TX_CELLS_PER_VCQ_COUNTER_SRAM_BASE]
#endif

#if streq ('COUNTER_TYPE','PKTS_PER_OUTPORT')
    immed32[baseaddr, AAL5_TX_PKTS_PER_OUTPORT_COUNTER_SRAM_BASE]
#endif

#if streq ('COUNTER_TYPE','PKTS_PER_VCQ')
    immed32[baseaddr, AAL5_TX_PKTS_PER_VCQ_COUNTER_SRAM_BASE]
#endif

	; Multiply in_index by 4 to get the SRAM byte address
	; that stores this counter.
	; Add the counter base to get the absolute SRAM address.
	alu_shf_left(index, baseaddr, +, in_index, 2)

	sram[incr, --, index, 0]

.end
#endm
#endif // COUNTERS

#endif // _AAL5_TX_UTIL_COMMON_UC_