/////////////////////////////////////////////////////////////////////////////////////
//
//
//                  I N T E L   P R O P R I E T A R Y
//
//     COPYRIGHT [c]  2002 BY  INTEL  CORPORATION.  ALL RIGHTS
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT
//     THE PRIOR WRITTEN PERMISSION OF :
//
//                        INTEL  CORPORATION
//
//                     2200 MISSION COLLEGE BLVD
//
//               SANTA  CLARA,  CALIFORNIA  95052-8119
//
/////////////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: csix_tx.c
//
//      Abstract: 
//				 A context pipe-stage microblock which receives a transmit request
//               from the Queue Manager, prepares the CSIX base header,CSIX extension
//               header, and additional Traffic Manager header (TM header, in
//               our case, is the info which ingress IXP2xxx wants to
//               pass to the egress IXP2xxx) and sends out CSIX cframes.
//
//
//      History:
//
//      Date            Comment                  				       By
//      ---------------------------------------------------------------------
//
//      07/11/2002      Created from David Chou's microcode version   Uday Naik
//
/////////////////////////////////////////////////////////////////////////////////////
//
// Include files
//
/////////////////////////////////////////////////////////////////////////////////////

#include <ixp.h>

#define extern 

// global definition of common constants which are shared by all microblocks

#include "dl_system.h"

// header files used in the project 

#include "csix_tx_c.h"
#include "csix_tx_util.c"
															
/////////////////////////////////////////////////////////////////////////////////////
//
// Packet processing entry point for csix microblock
//
/////////////////////////////////////////////////////////////////////////////////////

void CsixTx(void)
{

	dl_buf_handle_t								prevBufferHandle	    = {0};
	uint32_t		 							cellCountRemaining    	= 0;
	tcw_t										tcw 				    = {0};
	SIGNAL_MASK									sigmask_phase1Default 	= sigmask_phase1;
	SIGNAL_MASK									sigmask_phase2Default 	= sigmask_phase2;
	tbuf_t										tbuf 				    = {0};
	SIGNAL										sig_txRequestInit;	
	__declspec(sram_write_reg)	tcw_t			tcwWriteXfer;
	__declspec(sram_write_reg)  prepend_t		prependWriteXfer;
	__declspec(sram_read_reg) 	tx_request_t  	txRequest;

	// Kick start with a null request by reading zeros into txRequest
	// This is done by reading an SRAM location with all 0's

	// Read the meta data from SRAM.

	sram_read((__declspec(sram_read_reg) void*) &txRequest, 
			  (volatile __declspec(sram) void*) SRAM_ZERO_BLOCK, sizeof(txRequest) / 4
			  , ctx_swap, &sig_txRequestInit);
				 

	// The first time through the loop, wait for next thread signal only 

	wait_for_all(&sig_nextThread1);

	while (1) 
	{

		//////////////////////////////////////////////////////////////////////////////////
		//
		// This is the beginning of phase 1. Here we validate the tbuf for request n-1,
		// setup the DRAM operation for request n and do the scratch read for request n+1
		//
		//////////////////////////////////////////////////////////////////////////////////

		// set up default signal masks for phase 1 and 2 

		sigmask_phase1 = sigmask_phase1Default;
		sigmask_phase2 = sigmask_phase2Default;

		// signal the next thread to enter phase 1. Need to make sure that we dont
		// swap out until phase 1 is actually complete

		CsixTxSignalNext(__signal_number(&sig_nextThread1));

		// check if the packet n-1 is NULL. If it is, skip the transmit of the tbuf

		if (prevBufferHandle.value != 0)  
		{
		
			// copy the transmit control word to transfer registers

			tcwWriteXfer = tcw;

			// transmit the tbuf for packet n-1. This macro also checks if the 
			// buffer handle is done and if so it frees it. 
			
			CsixTxTransmitTbuf( 
								&tcwWriteXfer, 
								tbuf.txValidateAddress,   
								prevBufferHandle, 
								cellCountRemaining, 
								&sig_tcwWrite
							  );		
		}
		else
		{
			// clear the signal bit in sigmask_phase1

			CsixTxClearSignal(sigmask_phase1, __signal_number(&sig_tcwWrite));

		} // end if previous buffer handle != 0

		// check if the packet n is NULL. If it is, skip the dram to tbuf 

		if (txRequest.value)
		{

			// allocate a new tbuf 

			tbuf = CsixTxGetNextTbuf();

			// copy the buffer handle in tx request  to the prev_buffer_handle for 
			// next iteration and set up cell count reamining to 0 by default 
	
			prevBufferHandle   =  txRequest.bufHandle; 
			cellCountRemaining =  0;

			// Check if this is a packet with SOP/EOP bit set and can fit into one 
			// cframe (0xc0 in the MSB of txRequest1)

			if (txRequest.bufHandle.sop && txRequest.bufHandle.eop &&
				(txRequest.bufHandle.seg_count == 0))
			{

				// copy data from dram to tbuf for packet n. The function return a tcw 
				// for later use (13). tx_request1 contains the buffer handle

				tcw = CsixTxMoveSopEopDramToTbuf( 
												  txRequest, 
												  txRequest.bufHandle, 
												  tbuf, 
												  &prependWriteXfer, 
												  &sig_dramToTbuf, 
												  &sig_tmhWrite
											    );
											  
			}  
			else
			{
			
				// get the cell count remaning for the buffer 

				cellCountRemaining = txRequest.bufHandle.seg_count;
				
				// check if it is an sop buffer

				if (txRequest.bufHandle.sop)
				{

					// copy data from dram to tbuf for packet n. The function return a tcw 
					// for later use.  txRequest1 contains the buffer handle 
	
					tcw = CsixTxMoveSopDramToTbuf(
													txRequest, 
												    txRequest.bufHandle, 
												    tbuf, 
												    &prependWriteXfer, 
												    &sig_dramToTbuf, 
												    &sig_tmhWrite
												  );					
				}
				else
				{

					// copy data from dram to tbuf for packet n. The function return a tcw 
					// for later use.  tx_request1 contains the buffer handle 
	
					tcw = CsixTxMoveMopEopDramToTbuf(
														txRequest, 
													    txRequest.bufHandle, 
													    tbuf, 
												 	    &prependWriteXfer, 
												 	    &sig_dramToTbuf, 
												   	    &sig_tmhWrite
											   	    );
				
				}  // end if sop buffer

			}	// end if min packet

		}
		else
		{
			 // No data has been moved to DRAM 

			// clear the signal bit in sigmask_phase2 for dram to tbuf

			CsixTxClearSignal(sigmask_phase2, __signal_number(&sig_dramToTbuf.even));
		
			// clear the signal bit in sigmask_phase2 for dram to tbuf + 1

			CsixTxClearSignal(sigmask_phase2, __signal_number(&sig_dramToTbuf.odd));
	
			// clear the signal bit in sigmask_phase1 for  

			CsixTxClearSignal(sigmask_phase1, __signal_number(&sig_tmhWrite));

			// copy null to the prev_buffer_handle 
	
			prevBufferHandle.value  = 0;
		
		}	// end if buffer handle is not null

		// read tranmsit request for packet n+1 

		CsixTxReadTransmitRequest(&txRequest, &sig_scratchRead);

		// wait for read transmit request and transmit tbuf to finish and signal from
		// next thread 

		wait_for_all(sigmask_phase1);

		// indicate to the compiler that the signals in sigmask_phase 1 may be reused

		__implicit_read(&sig_scratchRead);
		__implicit_read(&sig_tcwWrite);
		__implicit_read(&sig_tmhWrite);
		__implicit_read(&sig_nextThread2);
		__implicit_read(&prependWriteXfer);
		__implicit_read((void *)&tcwWriteXfer);

		///////////////////////////////////////////////////////////////////////////////////
		//
		// Start of Phase 2
		//
		///////////////////////////////////////////////////////////////////////////////////

		// signal the next thread to enter phase 2 

		CsixTxSignalNext(__signal_number(&sig_nextThread2));

		// check for a null request in the scratch ring 

		if (txRequest.value) 
		{

			// Check if this is a packet with SOP/EOP bit set and can fit into one 
			// cframe 

			if (txRequest.bufHandle.sop && txRequest.bufHandle.eop &&
				(txRequest.bufHandle.seg_count == 0))	
			{

				// read the meta data and context from SRAM for packet n+1 

				Dl_MetaLoadCache(txRequest.bufHandle, &sig_metaRead, 7);

				// wait for dram to tbuf for packet n and read of meta for packet n+1 to finish 
				// and signal from next thread 

				wait_for_all(sigmask_phase2);

			}
			else
			{

				// Do a CAM lookup, get the txc context and meta data. This will also 
				// do a ctx arb to finish phase 2

				CsixTxGetContext(txRequest, CSIX_TX_CTX_SRAM_BASE);

			}  // end if min packet
		}
		else
		{

			// clear the signal bit in sigmask_phase2

			CsixTxClearSignal( sigmask_phase2, __signal_number(&sig_metaRead));
	
			// wait for dram to tbuf for packet n 

			wait_for_all(sigmask_phase2);
		
		}	// end if txRequest.value

		// indicate to the compiler that the signals for phase2 are done with 

		__implicit_read(&sig_metaRead);
		__implicit_read(&sig_dramToTbuf);
   		__implicit_read(&sig_nextThread1);

	} // end while (1)
	
}


/////////////////////////////////////////////////////////////////////////////////////

// The main code starts here

void main(void)
{

	// initialize the microblock

	CsixTx_Init();

	// call the packet processing block

	CsixTx();

}

/////////////////////////////////////////////////////////////////////////////////////

