/*************************************************************************************
**
**
**                  I N T E L   P R O P R I E T A R Y
**
**     COPYRIGHT [c]  2002 BY  INTEL  CORPORATION.  ALL RIGHTS
**     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY
**     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A
**     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER
**     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,
**     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT
**     THE PRIOR WRITTEN PERMISSION OF :
**
**                        INTEL  CORPORATION
**
**                     2200 MISSION COLLEGE BLVD
**
**               SANTA  CLARA,  CALIFORNIA  95052-8119
**
*************************************************************************************
**
**
**      File Name: csix_tx_c.h
**
**      Purpose: header file csix transmit block
**
**		
*************************************************************************************/

#ifndef _CSIX_TX_C_H
#define _CSIX_TX_C_H

/*************************************************************************************/

#include <ixp.h>
#include <ixp_lib.h>
#include <dl_buf.h>
#include <dl_meta.h>
#include "hardware.h"

/*************************************************************************************/

/*
 *	ISO C99: 7.18 Integer types <stdint.h> : For now define them here
 */

#ifndef _STDINT_H
#define _STDINT_H   1

/* Exact integral types.  */
/* Signed.  */

typedef signed char     int8_t;
typedef short int       int16_t;
typedef int             int32_t;
typedef long long int   int64_t;

/* Unsigned.  */

typedef unsigned char           uint8_t;
typedef unsigned short int      uint16_t;
typedef unsigned int            uint32_t;
typedef unsigned long long int  uint64_t;

#endif /* _STDINT_H */

#ifndef INLINE
#define INLINE __forceinline
#endif


/*************************************************************************************/

/* specify that the meta data is cached in dram read transfer registers */

#define DL_META_DATA_IN_READ_DXFER 		

/* specify size of the meta data cache as 7 long words  */

#define META_CACHE_SIZE		7

#include "dl_buf.c"
#include "dl_meta.c"

/*************************************************************************************/

extern void CsixTx_Init();

/*************************************************************************************

  Signal declaration 

*************************************************************************************/

/* These signals are local to the thread */

SIGNAL 			sig_scratchRead; 		/* signal for scratch ring read */
SIGNAL 			sig_metaRead; 			/* signal for reading meta data from sram */
SIGNAL 			sig_txcRead;; 			/* signal for reading in txc */
SIGNAL 			sig_txcWrite; 			/* signal for writing out txc */
SIGNAL_PAIR 	sig_dramToTbuf; 		/* signal for writing from dram to tbuf */
SIGNAL 			sig_tcwWrite; 			/* signal for writing transmit control word */
SIGNAL 			sig_tmhWrite; 			/* signal for writing TM header to tbuf */

/* These are for inter-thread signaling and must be global */

SIGNAL		sig_nextThread1; 		/* signal for phase 1 */
SIGNAL		sig_nextThread2; 		/* signal for phase 2 */

		
/* 
  This is the inter-ME signal generated by the ME whcih initializes common 
  resources, such as scratch ring, CSIX TX ME will wait for this signal at 
  end of it's init to make sure all common resource are available 
*/

SIGNAL			sig_initComplete;		

/* Signal masks for phase 1 and 2 */

SIGNAL_MASK		sigmask_phase1;
SIGNAL_MASK		sigmask_phase2;

/*************************************************************************************
**
**		Clear a signal from a signal mask 
**
*************************************************************************************/

#define CsixTxClearSignal(sigmask, signalNumber) (sigmask &= ~(1 << signalNumber))


/************************************************************************************
** 	
**		Set a signal in a signal mask 
**
*************************************************************************************/

#define CsixTxSetSignal(sigmask, signalNumber) (sigmask |= (1 << signalNumber))

/*************************************************************************************

 format of the transmit request on the scratch ring (2 words )

 *************************************************************************************/

typedef __declspec(packed)	struct __tx_request 
{
	union 
	{
		struct
		{
			uint32_t	valid			:1;			/* valid bit */
			uint32_t	reserved		:21;		/* reserved  */
			uint32_t 	egressBlade		:6;			/* egress blade id */
			uint32_t	class			:4;			/* class id */
		};

		uint32_t value;
	};
			
	dl_buf_handle_t	bufHandle;						/* buffer handle */
	

} tx_request_t;

/*************************************************************************************

 format of the transmit context (8 words) 

 *************************************************************************************/

typedef __declspec(packed) struct __txc 
{
	union 
	{
		struct 
		{
			uint32_t		reserved		:15;	/* reserved					*/
			uint32_t		valid			:1;		/* if set txc is valid		*/
			uint32_t		bytesSent		:16;	/* bytes sent for current buffer */
			uint32_t		alignOffset		:8;		/* offset for quad word align */
			uint32_t 		seqNumber		:16;	/* sequence number 			*/
			uint32_t 		eopFlag			:1;		/* EOP flag					*/
			uint32_t		freeListId		:7;		/* reserved 				*/
			uint32_t		nextHopId		:16;	/* next hop id 				*/
			uint32_t		classId			:16;	/* class id 				*/
			uint32_t		dramToTbufBytes	:16;	/* bytes transferred from DRAM to TBUF */
			uint32_t		offset			:16;	/* offset into buffer		*/
			uint32_t		reserved2[3];			/* reserved 				*/
			dl_buf_handle_t	bufHandle;				/* last buffer handle for this queue */
		};

		uint32_t value[8];
	};

} txc_t;

/***************************************************************************************

 Structure with relevant information on tbuf ( 2 words )

 **************************************************************************************/

typedef struct __tbuf_struct 
{

	uint32_t	address;							/* address for accessing tbuf    */
	uint32_t	txValidateAddress;					/* address for validating tbuf   */
	 
} tbuf_t;

/**************************************************************************************

 CSIX extension header (1 word) 

 **************************************************************************************/

typedef __declspec(packed) struct __csix_ext_header
{
	union
	{
		struct
		{
			uint32_t	class			:8;			/* class id 			*/
			uint32_t	private			:2;			/* CSIX private bits	*/
			uint32_t 	csixReserved	:10;		/* CSIX reserved bits	*/
			uint32_t	egressBlade		:12;		/* egress blade			*/	 
		};

		uint32_t value;
	};

} csix_ext_header_t;

/****************************************************************************************

 CSIX transmit control word (1 word) 

 ****************************************************************************************/

typedef __declspec(packed) struct __tcw 
{

	union 
	{
		struct 
		{
			uint32_t	payloadLength	:8;		/* length of the payload 			*/
			uint32_t	prependOffset	:3;		/* offset of the prepend header		*/
			uint32_t	prependLength	:5;		/* length of the prepend header		*/
			uint32_t	payloadOffset	:3;		/* offset of the payload header		*/
			uint32_t	reserved2		:1;		/* reserved							*/
			uint32_t	skip			:1;		/* Allows software to alloc tbuf but skip it */
			uint32_t	reserved3		:1;		/* reserved							*/
			uint32_t	csixReserved	:1;		/* CSIX reserved bit - goes into CSIX header */
			uint32_t	private			:1;		/* CSIX private bit					*/
			uint32_t	reserved		:4;		/* reserved bit						*/
			uint32_t	type			:4;		/* CSIX type field 					*/
		};

		uint32_t value;
	};

	csix_ext_header_t	extensionHeader;		/* CSIX extension header			*/

} tcw_t;

/*****************************************************************************************

 CSIX per cframe prepend (2 words)

 *****************************************************************************************/

typedef __declspec(packed) struct __cframe_prepend
{

	union
	{
		struct
		{
			uint32_t	alignOffset		:8;		/* quad word alignment offset 	*/
			uint32_t	sequenceNumber	:16;	/* sequence number of frame		*/
			uint32_t	eop				:1;		/* if set, bit is eop frame		*/
			uint32_t	sop				:1;		/* if set, bit is sop frame		*/
			uint32_t	ingressSrcBlade	:6;		/* ingress blade id #			*/
		};

		uint32_t value;
	};

	uint32_t	nextHopId	:16;				/* next hop id 					*/
	uint32_t 	classId		:16;				/* class id 					*/

} cframe_prepend_t;

/****************************************************************************************

 CSIX per packet prepend (2 words) 

 ****************************************************************************************/

typedef __declspec(packed) struct __packet_prepend
{

	uint32_t	inputPort			:16;	/* input port		*/
	uint32_t	outputPort			:16;	/* output port		*/
	uint32_t 	flowId;						/* flow id 			*/

} packet_prepend_t;

/*****************************************************************************************

 CSIX min packet prepend 

*****************************************************************************************/

typedef struct __prepend 
{

	cframe_prepend_t	cframePrepend;
	packet_prepend_t	packetPrepend;

} prepend_t;

/*****************************************************************************************
 
 Globals

*****************************************************************************************/

/* shared across threads storing next free tbuf to be used  */

extern __declspec(shared) 	uint32_t					globalCurrentTbufElementNumber;

/* An array of 16 transmit contexts in local memory  */

extern __declspec(local_mem shared aligned(16)) txc_t 	globalTxc[16];

/* A pointer to the current transmit context being processed by a thread */

extern __declspec(local_mem aligned(16))  		txc_t* 	pGlobalCurrentTxc;

/***************************************************************************************

 define the first thread for CSIX TX 

****************************************************************************************/

#define CSIX_TX_FIRST_THREAD			0x0

/*************************************************************************************

 Following definition enable CSIX TX waiting for an inter-ME signal 
 generated by the ME whcih initializes common resources, such as scratch ring,
 near the end of thread 0 of CSIX TX initialization to make sure all common 
 resource are available 

*************************************************************************************/
 
#define WAIT_FOR_COMMON_RESOURCE_INITIALIZATION 	

/*************************************************************************************/

/* MSF_TX_CNTL */

#define CSIX_TX_CTL_LW 	0x10410004		/* 0001 0000 0100 0001 Tx enable, CSIX,  
										   duplex Rx_CBus_Mode					
										   0000 0000 0000 0100 32 bits CWORD, 	
										   CSIX idle
 										 */


/* 
  initial first transmit control word for first cell in one packet (includes 
  info for CSIX base header) 
 */

#define TCW0_FCIP_BASE_LW	0x00100001	/* TM playload length = 0 bytes (will 
										   be overwritten in code), prepend 
										   length = 16 bytes, prepend offset=0,
										   payload offset = 0 (will be overwritten
										   in code), cframe type = 1 (unicast), 
										   other fields are zero
										 */


/* 
   initial first transmit control word for the cell which is not the first cell
   in one packet (includes info for CSIX base header)
 */

#define TCW0_NOT_FCIP_BASE_LW	0x00080001	/* TM playload length = 0 bytes (will 
											   be overwritten in code), prepend 
											   length = 8 bytes, prepend offset=0,
											   payload offset=0, cframe type = 1 
											   (unicast), other fields are zero
											 */

/* 
   for tbuf elment with is 128 bytes long, we have 48 tbuf elements for data
  (0 to 47) and 16 tbuf elements fo control 
 */

#define MAX_AVAIL_DATA_TBUF		0x30
#define FIRST_AVAIL_DATA_TBUF	0
  
/* 
  definition of size of fully loaded cframe TM payload in bytes
  it will be 112 = 128 (cframe payload size) - 16 (TM header size)
  for the first cell in one packet and
  120 = 128 (cframe payload size) - 8 (TM header size)
  for cell except the first cell in one packet = 120
 */

#define FULL_TM_PAYLOAD_SIZE_FCIP		112		
#define FULL_TM_PAYLOAD_SIZE_NOT_FCIP	120		


/*
   definition of the base bits pattern of indir_ref used in instruction which 
   moves packet data from dram to tbuf
 */

#define DRAM_INDIREF_BASE_LW	0x02000010	/* overrid bit of Ref_Cnt field (bit
											   25) override bit of RBUF*TBUF 
											   Byte addr field (bit 4) and must 
											   always be set
											 */

/* definition of CAM state */

#define CAM_STATE_INIT		0    	
#define CAM_STATE_VALID		1		   
									 

/* 
   The 7th long word of transmit context in local memory keep the last 
   dl_buffer_handle for this queue, it will be initialized to an address
   which is never possible be used as dl_buffer_handle at the initialization 
   to make the initial 16 CAM entries dummy. This means, any CAM hit for these
   initial 16 CAM entries still requires the code to read the meta data from 
   sram as the result of the comparison of it's actual dl_buffer_handle to 
   TXC[7] in local memory. Here we set it to be 0xFF, since dl_buffer_handle
   should be multiles of 32. 
   When the last cframe of one buffer is processed, we also set this TXC[7] 
   to 0xF3, so the code will prevent one new buffer in later runs to use stale
   meta data in local memory, in the case that the transmit context for this 
   queue stays in local memory and the new buffer uses the same 
   dl_buffer_handle as the previous buffer for this queue (the chance of this
   condition to happen should be very low).
*/

#define INVALID_HANDLE			0xFF

/* 
   since each TXC in sram has 32 bytes, so queue number left shifting 5 bits 
   will give the offset of this TXC entry to sram_txc_base
*/

#define SRAM_TXC_SHIFT	5

/*************************************************************************************/
		
#endif /* _CSIX_TX_C_H */

/*************************************************************************************/