/////////////////////////////////////////////////////////////////////////////////////
//
//
//                  I N T E L   P R O P R I E T A R Y
//
//     COPYRIGHT [c]  2002 BY  INTEL  CORPORATION.  ALL RIGHTS
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT
//     THE PRIOR WRITTEN PERMISSION OF :
//
//                        INTEL  CORPORATION
//
//                     2200 MISSION COLLEGE BLVD
//
//               SANTA  CLARA,  CALIFORNIA  95052-8119
//
/////////////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: csix_tx_util.c
//
//      Abstract: Utility functions for csix transmit 
//
//
//
/////////////////////////////////////////////////////////////////////////////////////

#include "csix_tx_c.h"

/////////////////////////////////////////////////////////////////////////////////////
// 
// CsixTxReadTransmitRequest()
//
// Description:
// 	
//		Read the transmit request from a scratch ring
//
// Inputs: 
//
//		txRequest:		2 xfer registers that will contain the tx request
//						when the I/O completes
//
//		sig_txRequest:	signal number to wait on
//
/////////////////////////////////////////////////////////////////////////////////////

INLINE void CsixTxReadTransmitRequest 
( 
	__declspec(sram_read_reg) 	tx_request_t *txRequest,
	SIGNAL 									 *sig_txRequest
)
{

	// get tx request - 2 words from scratch ring 

	scratch_get_ring(
						(__declspec(sram_read_reg) void*) txRequest, 
						(volatile __declspec(scratch) void*) (CSIX_TX_RING_IN << 2), 
						2, 
						sig_done, 
						sig_txRequest
					);
}


/////////////////////////////////////////////////////////////////////////////////////
// 
// CsixTxGetNextTbuf()
//
// Description:
// 	
//		Allocate a tbuf element. Instead of returning the element number we return
//		two values which are actually useful - the element address and the address 
//		for validating the tbuf. This allows us to optimize the wrap around case 
//		better.
//
// Outputs: 
//
//		tbuf structure 
// 
//
/////////////////////////////////////////////////////////////////////////////////////

INLINE tbuf_t CsixTxGetNextTbuf(void)
{

	tbuf_t		tbuf;
	uint32_t	tbufElementNumber;

	// get the current tbuf element 

	tbufElementNumber = globalCurrentTbufElementNumber;

	// increment the current tbuf element number

	if (globalCurrentTbufElementNumber < (MAX_AVAIL_DATA_TBUF - 1))
		globalCurrentTbufElementNumber++;
	else
		globalCurrentTbufElementNumber = 0;

	// store the byte address of the TBUF assuming 128 byte TBUFs

	tbuf.address = RBUF_TBUF_BASE + (tbufElementNumber << 7);

	// store the address for validating the TBUF 

#ifdef IXP2800

	tbuf.txValidateAddress = TBUF_ELEMENT_CTRL_VALID	+ (tbufElementNumber << 4);

#else

    tbuf.txValidateAddress = TBUF_ELEMENT_CTRL_VALID	+ (tbufElementNumber << 3);

#endif

	return tbuf;

}


/////////////////////////////////////////////////////////////////////////////////////
// 
// CsixTxSignalNext()
//
// Description:
// 	
//		signal the next thread
//
// Outputs: 		
//
// Inputs:  
//
//
//		signalNumber	signal number
// 
//
/////////////////////////////////////////////////////////////////////////////////////


INLINE void CsixTxSignalNext(uint32_t	signalNumber)
{

	// signal number goes into bits 3..6 and bit 7 is set to signal the next context.

	local_csr_write(local_csr_same_me_signal, (0x80 | (signalNumber << 3)));

}

/////////////////////////////////////////////////////////////////////////////////////
// 
// CsixTxTransmitTbuf()
//
// Description:
// 	
//		Validate a TBUF for transmit. Also if the cell count has gone to 0
//		then free the buffer  
//
// Outputs: 
//
// Inputs:  
//		
//		tcw							2 xfer registers that hold the transmit control word
//		addrOfTxControl:			address to validate tbuf
//		bufferHandle:				buffer handle
//		cellCountRemained:			cell count remaining
//		sig_tbufValidate:			signal number
// 
//
/////////////////////////////////////////////////////////////////////////////////////

INLINE  void CsixTxTransmitTbuf
(
	__declspec(sram_write_reg) 	tcw_t* 			tcw,
	uint32_t									addrOfTxControl,
	dl_buf_handle_t 							bufferHandle, 
	uint32_t									cellCountRemained,
	SIGNAL*		 								sig_tbufValidate
)

{

	// check if the cell count remaining for the buffer is 0

	if (cellCountRemained == 0)
		Dl_BufDrop(bufferHandle);

 	// write transmit contrl word (TCW) into MSF. write to TBUF_ELEMENT_CTRL_V_# 
	// corresponding the tbuf_element to initiate transmission.

	msf_write(
				(__declspec(sram_write_reg) void*) tcw, 
				(volatile void *) addrOfTxControl, 
				2, 
				sig_done, 
				sig_tbufValidate
			 );

}

/////////////////////////////////////////////////////////////////////////////////////
//	
//	CsixTxGetTransmitControlWord[]
//
// Description:
// 	
//		Get transmit control word from offset, length and transmit request 
//
// Outputs: 
//
//		tcw					transmit control word
//
// Inputs:  
//
//		tcwBase				base for transmit control word
//		offset				offset into quadword to start transmit
//		payloadLength		payload length
//	    txRequest			transmit request
//
/////////////////////////////////////////////////////////////////////////////////////

INLINE tcw_t CsixTxGetTransmitControlWord
(
	uint32_t		tcwBase, 
	uint32_t		offset,
	uint32_t		payloadLength,
	tx_request_t	txRequest
)
{

	tcw_t		tcw;

	// set the class and egress blade 
 
 	tcw.extensionHeader.value 			= 0;	/* reset extension header */ 
 	tcw.extensionHeader.class 			= txRequest.class; 
	tcw.extensionHeader.egressBlade 	= txRequest.egressBlade;

	// fill in the byte offset (0 to 7) to start transmit (13..15 of tcw)

	tcw.value 			= tcwBase;
	tcw.payloadOffset   = offset;
	tcw.payloadLength   = payloadLength;

	return tcw;

}

/////////////////////////////////////////////////////////////////////////////////////
// 
// CsixTxMoveSopEopDramToTbuf()
//
// Description:
// 	
//		For a single cell packet (SOP/EOP both set) this macro sets up the dram
//		to tbuf transfer. It also writes the 
//
// Outputs: 
//
//		tcw					transmit control word
//
// Inputs:  
//
//		txRequest			transmit request
//		bufferHandle		buffer handle
//		tbuf				tbuf structure
//		prependXfer			prepend write transfer registers
//		sig_dram			dram signal
//		sig_msfWrite		signal to wait on 
//
/////////////////////////////////////////////////////////////////////////////////////

INLINE tcw_t  CsixTxMoveSopEopDramToTbuf
(
	tx_request_t				txRequest, 
	dl_buf_handle_t				bufferHandle,
	tbuf_t						tbuf, 
	__declspec(sram_write_reg) 	prepend_t *prependXfer,
	SIGNAL_PAIR*				sig_dram,
	SIGNAL*						sig_msfWrite
)
{

	
	uint32_t				dramAddress; 	//	dram address at which data starts
	uint32_t				payloadLength; 	//  payload length to be transmitted
 	dram_rbuf_tbuf_ind_t	indirRef;	 	//	used for indirect ref calculation
	prepend_t				prepend;		//  used to create prepend header in gprs

  	// Get the dram address at which the data starts 

	dramAddress = Dl_BufGetData(bufferHandle) + dlMeta.offset;

	// Copy buffer size from meta data into payload length. Round off to nearest 
	// quad word boundary

	payloadLength = dlMeta.bufferSize + (dlMeta.offset & 0x7);
    
	// Write the tbuf address into the indirect ref token. Note since it is SOP/EOP 
	// buffer we have a 16 byte prepend 

	indirRef.value = DRAM_INDIREF_BASE_LW;
	indirRef.buf_addr = tbuf.address + 16;

    // Write the ref_cnt into the indirect token. 
	// Since ref count 0 will fetch 8 bytes, and ref count 1 will fetch
	// 16 bytes......, and ref count 15 will fetch 128 bytes, 
	// the right way to do it is followinhg:
	//    a) decrease length by one
	//    b) then right shift the result by 3

	indirRef.ref_count =  (payloadLength - 1) >> 3;
	 
    // Write to the tbuf. Note that how many bytes of packet payload are copied 
	// to tbuf from dram is controlled by ref cnt field in indirRef, following 14 is
	// dummy	

	dram_tbuf_write_ind((volatile void __declspec (dram)*) dramAddress, 14, indirRef, \
						sig_done, sig_dram);

	// initialize cframe prepend header 

	prepend.cframePrepend.ingressSrcBlade 	= THIS_BLADE_ID;
	prepend.cframePrepend.alignOffset		 = dlMeta.offset & 0x7;
	prepend.cframePrepend.sop				 = 1;
	prepend.cframePrepend.eop				 = 1;
	prepend.cframePrepend.sequenceNumber 	 = 0;
	prepend.cframePrepend.nextHopId			 = dlMeta.nextHopId;
	prepend.cframePrepend.classId			 = dlMeta.classId;

	// initialize per packet prepend header 

	prepend.packetPrepend.inputPort 		= dlMeta.inputPort;
	prepend.packetPrepend.outputPort 		= dlMeta.outputPort;
	prepend.packetPrepend.flowId 			= dlMeta.flowId;

	// copy the prepend header into transfer registers 

	*prependXfer = prepend;

    // write 4 words of prepend header to tbuf using msf write 

	msf_write (prependXfer, (volatile void*) tbuf.address, 4, sig_done, sig_msfWrite);

	// Compute the transmit control word for later use with the payload length

	return CsixTxGetTransmitControlWord(TCW0_FCIP_BASE_LW, 0, \
												payloadLength, txRequest);


}

/////////////////////////////////////////////////////////////////////////////////////
// 
// CsixTxSwapInContext
//
// Description:
// 	
//		Read transmit context from SRAM 
//
// Outputs: 	
//
// Inputs:  
//
//		txc					xfer registers to read in context
//		txRequest			transmit request
//		sramTxcBase			base for sram contexts
//		sig_txcRead			signal to wait on
// 
//
/////////////////////////////////////////////////////////////////////////////////////

INLINE void CsixTxSwapInContext
(
	 	__declspec(sram_read_reg) 	uint32_t	*txcIn,
		uint32_t								queueNumber, 
		uint32_t								sramTxcBase, 
		SIGNAL*									sig_txcRead
)
{

	uint32_t	sramOffset;

	// Each context is 32 bytes. So shift by 5.

	sramOffset = sramTxcBase + (queueNumber << SRAM_TXC_SHIFT);

	// Issue the read 

	sram_read(txcIn, (volatile void __declspec (sram)*) sramOffset, 4, sig_done, 
			  sig_txcRead);

}

/////////////////////////////////////////////////////////////////////////////////////
// 
// CsixTxSwapOutContext
//
// Description:
// 	
//		swap out context into sram
//
// Inputs:  
//
//		pCurrentTxc			pointer in local memory to current context
//		txcOut				xfer registers to read out context
//		camEntry			transmit request
//		sramTxcBase			base for sram contexts
//		sig_txcWrite		signal to wait on
//		
/////////////////////////////////////////////////////////////////////////////////////

INLINE void CsixTxSwapOutContext
(
	__declspec(local_mem aligned(16)) 	txc_t   *pCurrentTxc,
	__declspec(sram_write_reg) 	uint32_t		*txcOut, 
	uint32_t									camEntry, 
	uint32_t									sramTxcBase, 
	SIGNAL*										sig_txcWrite
)
{

	uint32_t	sramOffset;

	// move LRU txc from LM to TX register
		
	txcOut[0] = pCurrentTxc->value[0];
	txcOut[1] = pCurrentTxc->value[1];
	txcOut[2] = pCurrentTxc->value[2];
	txcOut[3] = pCurrentTxc->value[3];

	// Compute txc addr in SRAM of LRU TXC. The cam entry contains the queue #
	// Each context is 32 bytes so shift by SRAM_TXC_SHIFT which is 5

	sramOffset  = sramTxcBase+ (cam_read_tag(camEntry) << SRAM_TXC_SHIFT); 

	// swap out the LRU context 
 
 	sram_write(txcOut, (volatile void __declspec (sram)*) sramOffset, \
						4, sig_done, sig_txcWrite);

}

////////////////////////////////////////////////////////////////////////////////////
//
// CsixTxGetContext()
//
// Description:
// 	
//		swap out context into sram
//
// Outputs: 
//
// Inputs:  
//
//		txRequest		transmit request
//		sramTxcBase		base for sram contexts
//
/////////////////////////////////////////////////////////////////////////////////////
 
INLINE void CsixTxGetContext
( 
	tx_request_t	txRequest,
	uint32_t	 	sramTxcBase
)
{

	uint32_t								entryNum;
	uint32_t								cellCountRemained;
	uint32_t								tmp;
	cam_lookup_t							camResult;
	__declspec(sram_write_reg)	uint32_t	txcOutXfer[4];
	__declspec(sram_read_reg)	uint32_t	txcInXfer[4];

	// Lookup up the queue # in the CAM 

	camResult = cam_lookup(txRequest.value);

	// Get the CAM entry # 

	entryNum = camResult.entry_num;

	// Get the appropriate context in local memory

	pGlobalCurrentTxc = (__declspec(local_mem aligned(16)) txc_t*) &(globalTxc[entryNum]);

	// check if it is a miss

	if (camResult.hit == 0)
	{

		// read the meta data from SRAM for the context - only 7 words

		Dl_MetaLoadCache(txRequest.bufHandle, &sig_metaRead, 7);

		// swap out context 

		CsixTxSwapOutContext( pGlobalCurrentTxc, txcOutXfer, camResult.entry_num, 
									sramTxcBase, &sig_txcWrite);

		// swap in new context  (2)

		CsixTxSwapInContext( txcInXfer, camResult.entry_num, sramTxcBase, 
								  &sig_txcRead);

		// set the bit in the bitmask for txc read signal 

		CsixTxSetSignal(sigmask_phase2, __signal_number(&sig_txcRead));
	

		// set the bit in the bitmask for the txc write signal

		CsixTxSetSignal(sigmask_phase2, __signal_number(&sig_txcWrite));

		// update the buffer handle in the transmit context

		pGlobalCurrentTxc->bufHandle.value = txRequest.bufHandle.lw_offset;

		// update the CAM with state CAM_STATE_VALID, this CAM state is not used now, 
		// but may be used later

		cam_write(entryNum, txRequest.value, CAM_STATE_VALID);

		// wait for dram to tbuf for packet n and read of meta and context for packet 
		// n+1 to finish and signal from next thread (5)

		wait_for_all(sigmask_phase2);

		// copy in the transmit context 

		pGlobalCurrentTxc->value[0] = txcInXfer[0];
		pGlobalCurrentTxc->value[1] = txcInXfer[1];
		pGlobalCurrentTxc->value[2] = txcInXfer[2];
		pGlobalCurrentTxc->value[3] = txcInXfer[3];
		
	}
	else 
	{

		// Its a CAM hit

		// Check if meta data read is needed 

		if ((txRequest.bufHandle.lw_offset) != pGlobalCurrentTxc->value[7])
		{

			// read the meta data from SRAM 

			Dl_MetaLoadCache(txRequest.bufHandle, &sig_metaRead, 7);

			pGlobalCurrentTxc->value[7] = txRequest.bufHandle.lw_offset;

			// wait for dram to tbuf for packet n and read of meta and context for packet 
			// n+1 to finish and signal from next thread (5)

			wait_for_all(sigmask_phase2);
		
		}
		else
		{
  			// clear the signal for reading meta data 

			CsixTxClearSignal(sigmask_phase2, __signal_number(&sig_metaRead));
			
			// wait for phase 2 to end 

			wait_for_all(sigmask_phase2);
		}
	}	

}

/////////////////////////////////////////////////////////////////////////////////////
// 
// CsixTxMoveSopDramToTbuf
//
// Description:
// 	
//		For a sop cell (SOP bit only is set) this macro sets up the dram
//		to tbuf transfer
//
// Outputs: 
//
//		tcw					transmit control word 			
//
// Inputs:  
//		
//		txRequest			transmit request
//		bufferHandle		buffer handle
//		tbuf				structure with adress of tbuf
//		prependXfer			write transfer registers 
//		sig_dram			dram signal pair to wait on 
//		sig_msfWrite		msf write signal to wait on
//
/////////////////////////////////////////////////////////////////////////////////////

INLINE tcw_t CsixTxMoveSopDramToTbuf
( 
	tx_request_t							txRequest, 
	dl_buf_handle_t							bufferHandle,
	tbuf_t									tbuf, 
	__declspec(sram_write_reg) 	prepend_t   *prependXfer,
	SIGNAL_PAIR*							sig_dram,
	SIGNAL*									sig_msfWrite
) 

{
	uint32_t				offset;			// offset into DRAM buffer	
	uint32_t				dramAddress;	// dram address at which data starts
	dram_rbuf_tbuf_ind_t	indirRef;	 	//	used for indirect ref calculation
	uint32_t				payloadLength;	// payload length to be transmitted
	prepend_t				prepend;		// CSIX Prepend structure

	// set payload length to the default payload for SOP cell (112 bytes)

	payloadLength = FULL_TM_PAYLOAD_SIZE_FCIP;

	// Get offset in buffer of the data and store it into local memory 

	offset = dlMeta.offset;
	pGlobalCurrentTxc->offset = offset;

	// Get the dram address for the buffer

	dramAddress = Dl_BufGetData(bufferHandle) + dlMeta.offset;

	// set buffer descriptor valid bit 

	pGlobalCurrentTxc->valid = 1;

	// bytes sent 

	pGlobalCurrentTxc->bytesSent = payloadLength;

	// total bytes to be transfered from dram to tbuf for this buffer

	pGlobalCurrentTxc->dramToTbufBytes = dlMeta.bufferSize + (offset & 0x7);

	// sequence number for this packet

	pGlobalCurrentTxc->seqNumber = 0;
	
	// Alignment offset 

	pGlobalCurrentTxc->alignOffset = offset & 0x7;

	// Next hop id 

	pGlobalCurrentTxc->nextHopId = dlMeta.nextHopId;

	// Write the tbuf address into the indirect ref token. Note since it is SOP/EOP 
	// buffer we have a 16 byte prepend 

	indirRef.value = DRAM_INDIREF_BASE_LW;
	indirRef.buf_addr = tbuf.address + 16;

    // Write the ref_cnt into the indirect token. 
	// Since ref count 0 will fetch 8 bytes, and ref count 1 will fetch
	// 16 bytes......, and ref count 15 will fetch 128 bytes, 
	// the right way to do it is followinhg:
	//    a) decrease length by one
	//    b) then right shift the result by 3

	indirRef.ref_count =  (payloadLength - 1) >> 3;
	 
    // Write to the tbuf. Note that how many bytes of packet payload are copied 
	// to tbuf from dram is controlled by ref cnt field in indirRef, following 14 is
	// dummy	

	dram_tbuf_write_ind((volatile void __declspec (dram)*) dramAddress, 14, indirRef, \
						sig_done, sig_dram);

  	// initialize cframe prepend header 

	prepend.cframePrepend.ingressSrcBlade 	 = THIS_BLADE_ID;
	prepend.cframePrepend.alignOffset		 = offset & 0x7;
	prepend.cframePrepend.sop				 = 1;
	prepend.cframePrepend.eop				 = 0;
	prepend.cframePrepend.sequenceNumber 	 = 0;
	prepend.cframePrepend.nextHopId			 = dlMeta.nextHopId;
	prepend.cframePrepend.classId			 = dlMeta.classId;

	// initialize per packet prepend header 

	prepend.packetPrepend.inputPort 		= dlMeta.inputPort;
	prepend.packetPrepend.outputPort 		= dlMeta.outputPort;
	prepend.packetPrepend.flowId 			= dlMeta.flowId;

	// copy the prepend header into transfer registers 

	*prependXfer = prepend;

    // write 4 words of prepend header to tbuf using msf write 

	msf_write (prependXfer, (volatile void*) tbuf.address, 4, sig_done, sig_msfWrite);

	// Compute the transmit control word for later use with the payload length

	return CsixTxGetTransmitControlWord(TCW0_FCIP_BASE_LW, 0, \
												payloadLength, txRequest);
  
}

/////////////////////////////////////////////////////////////////////////////////////
// 
// CsixTxMoveMopEopDramToTbuf[]
//
// Description:
// 	
//		For a mop or eop only cell handle the dram to tbuf xfer
//
// Outputs: 
//	
//		tcw						transmit control word
//
// Inputs:  
//
//		txRequest			transmit request
//		bufferHandle		buffer handle
//		tbuf				structure with adress of tbuf
//		prependXfer			write transfer registers 
//		sig_dram			dram signal pair to wait on 
//		sig_msfWrite		msf write signal to wait on
//		
//
/////////////////////////////////////////////////////////////////////////////////////

INLINE tcw_t CsixTxMoveMopEopDramToTbuf
(
	tx_request_t							txRequest, 
	dl_buf_handle_t							bufferHandle,
	tbuf_t									tbuf, 
	__declspec(sram_write_reg) 	prepend_t   *prependXfer,
	SIGNAL_PAIR*							sig_dram,
	SIGNAL*									sig_msfWrite
)
{

	uint32_t				offset;				// offset into DRAM buffer	
	uint32_t				dramAddress;		// dram address at which data starts
	dram_rbuf_tbuf_ind_t 	indirRef;	 		//	used for indirect ref calculation
	uint32_t				payloadLength;		// payload length to be transmitted
	prepend_t				prepend;			// CSIX Prepend structure
	uint32_t				cellCountRemained;  // cells left 
	uint32_t				bytesSent;			// bytes sent from this buffer

	// set payload length to length for a non sop cell (120) 

	payloadLength = FULL_TM_PAYLOAD_SIZE_NOT_FCIP;

	// get remaining cell count from buffer handle 

  	cellCountRemained = bufferHandle.seg_count;

	// check if the buffer descriptor is valid by checking the valid bit

	if (pGlobalCurrentTxc->valid == 0)
	{

		// Previous buffer is done, so this is a new buffer and this is
		// not the first cell of a packet

		// Set the buffer descriptor valid bit 

		pGlobalCurrentTxc->valid = 1;

		// Set the bytes sent to zero 

		pGlobalCurrentTxc->bytesSent = 0;

		// Get the offset from the meta data and store it in the context 

		pGlobalCurrentTxc->offset = dlMeta.offset;

		// Get the size of bytes in buffer and store it in the dram_to_tbuf_bytes field

		pGlobalCurrentTxc->dramToTbufBytes = dlMeta.bufferSize;

	}

	// store bytes_sent for this buffer

	bytesSent = pGlobalCurrentTxc->bytesSent;

	// increment bytes sent for this buffer 

	pGlobalCurrentTxc->bytesSent += payloadLength;

	 // check if this is the last cell of the buffer

	if (cellCountRemained == 0)
	{
	 
		// reset buffer handle 

		pGlobalCurrentTxc->bufHandle.value = INVALID_HANDLE;

		// set valid bit to 0

		pGlobalCurrentTxc->valid = 0;

		// calculate actual payload length

		payloadLength =  pGlobalCurrentTxc->dramToTbufBytes - bytesSent; 

		// check if the last cell of the packet 

		if (bufferHandle.eop)
			pGlobalCurrentTxc->eopFlag = 1;

	}

 	// Get the dram base address for the buffer

    dramAddress = Dl_BufGetData(bufferHandle)  + pGlobalCurrentTxc->offset + bytesSent; 

	// Write the tbuf address into the indirect ref token. Note since it is MOP/EOP 
	// buffer we have a 8 byte prepend 

	indirRef.value = DRAM_INDIREF_BASE_LW;
	indirRef.buf_addr = tbuf.address + 8;

    // Write the ref_cnt into the indirect token. 
	// Since ref count 0 will fetch 8 bytes, and ref count 1 will fetch
	// 16 bytes......, and ref count 15 will fetch 128 bytes, 
	// the right way to do it is followinhg:
	//    a) decrease length by one
	//    b) then right shift the result by 3

	indirRef.ref_count =  (payloadLength - 1) >> 3;
	 
    // Write to the tbuf. Note that how many bytes of packet payload are copied 
	// to tbuf from dram is controlled by ref cnt field in indirRef, following 14 is
	// dummy	

	dram_tbuf_write_ind((volatile void __declspec (dram)*) dramAddress, 15, indirRef, \
						sig_done, sig_dram);

	// increment the seq number 

	pGlobalCurrentTxc->seqNumber++;

	// Create the per cframe prepend

	prepend.cframePrepend.value				 = 0; /* reset first LW */
	prepend.cframePrepend.ingressSrcBlade 	 = THIS_BLADE_ID;
	prepend.cframePrepend.eop				 = pGlobalCurrentTxc->eopFlag;
	prepend.cframePrepend.sequenceNumber 	 = pGlobalCurrentTxc->seqNumber;
	prepend.cframePrepend.nextHopId			 = pGlobalCurrentTxc->nextHopId;
	prepend.cframePrepend.classId			 = pGlobalCurrentTxc->classId;

	// copy the prepend header into transfer registers 

	prependXfer->cframePrepend = prepend.cframePrepend;

    // write 2 words of prepend header to tbuf using msf write 

	msf_write (prependXfer, (volatile void*) tbuf.address, 2, sig_done, sig_msfWrite);

	// Compute the transmit control word for later use with the payload length

	return CsixTxGetTransmitControlWord(TCW0_NOT_FCIP_BASE_LW, 0, \
												payloadLength, txRequest);	
												
}					

/////////////////////////////////////////////////////////////////////////////////////

// end __CSIX_TX_UTIL_UC__

/////////////////////////////////////////////////////////////////////////////////////

