/////////////////////////////////////////////////////////////////////////////////////
//
//
//                  I N T E L   P R O P R I E T A R Y
//
//     COPYRIGHT [c]  2002 BY  INTEL  CORPORATION.  ALL RIGHTS
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT
//     THE PRIOR WRITTEN PERMISSION OF :
//
//                        INTEL  CORPORATION
//
//                     2200 MISSION COLLEGE BLVD
//
//               SANTA  CLARA,  CALIFORNIA  95052-8119
//
/////////////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: csix_tx.uc
//
//      Purpose: a context pipe-stage microblock which receives tansmit request
//               from Queue Manager, prepare CSIX base header, CSIX extension
//               header, and additional Traffic Manager header (TM header, in
//               our case, this is the info which ingress Sausalito wants to
//               pass to the egress Sausalito, which includes destination L2
//               port id, next hop, sequence number, and flags of SOP & EOP
//               now), then send CSIX payload (TM header and payload from packet
//               buffer) and CSIX headers through MSF Transmit State Machine.
//
//
//      History:
//
//      Date            Comment                         				By
//      ---------------------------------------------------------------------------
//
//      11/21/2001      Created                         			David Chou
//
//		3/18/02			Modified to do packet interleaving			Uday Naik	
//						and adhere to coding convention
//
/////////////////////////////////////////////////////////////////////////////////////

#ifndef __CSIX_TX_UC__
#define __CSIX_TX_UC__

/////////////////////////////////////////////////////////////////////////////////////
//
// The CSIX transmit microblock for Sausalito is a context pipe-stage that
// runs on a single microengine. It receives transmit requests from the Queue
// Manager. 
// 
// With each transmit request, the microblock prepares a Transmit
// Control word (which include a CSIX base header and extension header info),
// and prepares an additional Traffic Manager header (TM header). The TM
// header is the info which ingress Sausalito wants to pass to the egress
// Sausalito, and includes the destination L2 port id, next hop, cell sequence
// number in one packet, and flags of SOP & EOP now). 
//
// The microblock moves the CSIX payload (TM header and payload from packet buffer) 
// into a TBUF element, and then sends this CSIX Frame to the fabric through the 
// MSF Transmit State Machine by writing the TCW to the corresponding
// TBUF_ELEMENT_CTRL_V_# address.  
//
// Every transmit request has an associated packet, which is being segmented into 
// cframes. The transmit context descripter (TXC which includes associated 
// segmentation state for the packet and the packet meta data) is cached in local 
// memory and is looked up using the CAM to speedup the processing.  If an entire 
// buffer for a packet has been transmitted, then the buffer is freed to the buffer 
// freelist.
//
// Assumptions:
//	1) Xscale core initialized transmit context (TXC) to all zero in SRAM
//	2) Queue Manager (QM) created transmit request in scratch ring
//  3) TBUF element size is 128 bytes
//  4) TM header size for the first cell in one packet is 16 bytes
//  5) TM header size for the cell which is not the first cell in one 
//     packet is 8 bytes
//  6) fabric ingress blade number is imported from Xscale core
//  7) TXC in local memory starts at local memory address 0, each TXC occupies
//     32 bytes in sram, but it occupies 64 bytes in local memoy to match 
//     cam_looup[.....], lm_addr0[0] operation
//  8) In the first longword of a valid tx request issued by QM, valid bit is
//     always set, and bit 30 to bit 16 are always zero, so we can use this
//     longword to do CAM lookup based on the queue number
//
/////////////////////////////////////////////////////////////////////////////////////

// following definition used to activate global counters update in sram
//#define COUNTERS

// include stdmac.uc in IXPblocks Portable library

#include <stdmac.uc>

// include local memory macros in IXPblocks 

#include <localmem.uc>

// global definition of common constants which are shared by all microblocks

#include "dl_system.h"

// definition of hardware register address 

#include <hardware.h>

// definition of constants used

#include "csix_tx.h"

// definition of macros used

#include "csix_tx_util.uc"

// initialization code for CSIX TX

#include "csix_tx_init.uc"

// dispatch loop macros

#include "dispatch_loop.uc"
										
// The following counters are stored in absolute registers. These counters are
// primarily used in debugging.
#ifdef	_DEBUG_COUNTERS_

// declare some absolute registers
.reg	@csix_tx_num_cells_txed			// number of C-Frames transmitted
.reg	@csix_tx_num_buf_freed			// number of buffers freed
.reg	@csix_tx_num_tx_req				// number of valid transmit requests

#endif
				
/////////////////////////////////////////////////////////////////////////////////////
//
// Packet processing entry point for csix microblock
//
// Size:	min packet is 74 instructions
//
/////////////////////////////////////////////////////////////////////////////////////

#macro csix_tx[]

.begin

	.reg		addr_of_tbuf			; address of tbuf
	.reg		addr_of_tx_control		; address of TBUF_ELEMENT_CTRL_V_#
	.reg		prev_buffer_handle		; buffer handle for packet n-1
	.reg		tcw						; transmit control word
	.reg		ex_header				; extension header
	.reg		sigmask_phase1			; signal mask for phase 1
	.reg		sigmask_phase2			; signal mask for phase 2
	.reg		tx_request_ring			; scratch ring for tx request 
	.reg		tmp						; temporary register
	.reg		tcw_fcip_base			; base for tcw for first cell of packet
	.reg		tcw_nfcip_base			; base for tcw for not first cell of packet
	.reg		indiref_base			; base for indirect ref of dram 
	.reg		s_txc_base				; base for tx context in sram
	.reg		cell_count_remaining	; cell count remaining in buffer
	.sig		zero_read_signal		; signal to read in zero's into tx_request
	.reg		lm_global_base			; base in local memory for globals 
	.reg		align_offset			; offset into quad word for alignment
	.reg		sigmask2_default 		; default value for signal mask for phase 2
	.reg		tx_counter_base			; base address for counters
#ifdef COUNTERS
.reg $byte_cnt							; sram xfer register to keep value to add  
										; to bytes transmitted counter in sram
#endif //#ifdef	COUNTERS	


#if IS_IXPTYPE(__IXP28XX)  //NEW_28XX_CODE

    // allocate NN registers

    .reg nn_tx_request0, nn_tx_request1

	#define tx_request0 nn_tx_request0
    #define tx_request1 nn_tx_request1

#else

    // allocate 2 read xfer registers for reading the transmit request

	xbuf_alloc[$tx_request, 2, read]

	#define tx_request0 $tx_request[0]
    #define tx_request1 $tx_request[1]

#endif
	
	
	// allocate 6 read xfer registers for reading in the meta data 

	xbuf_alloc[$$dl_meta, 7, read]
	
	// For the initial run let the assembler think the meta data is loaded 

	.set $$dl_meta[0] $$dl_meta[1] $$dl_meta[2] $$dl_meta[3] $$dl_meta[4] $$dl_meta[5] $$dl_meta[6]

	// initialize local memory index 1 to the base for globals 

	immed32[lm_global_base, GLOBAL_LM_BASE]

	// 3 cycles for this to take effect

	localmem_set_address[0, lm_global_base, LM_HANDLE_1]

	nop
	nop
	nop

	// set previous buffer handle and cell count remaining to 0

	alu[prev_buffer_handle, --, B, 0] 
	alu[cell_count_remaining, --, B, 0]

	// To get rid of warnings

	alu[tcw, --, B, 0]
	alu[ex_header, --, B, 0]
	alu[addr_of_tx_control, --, B, 0]

	// kick start with a null request by reading 0 into tx_request0

#if IS_IXPTYPE(__IXP28XX)  //NEW_28XX_CODE
    
	move[tx_request0, 0]
	move[tx_request1, 0]

#else

	immed32[tmp, SRAM_ZERO_BLOCK]
	sram[read, tx_request0, tmp, 0, 2] , ctx_swap[zero_read_signal]

#endif
	// gprs initialized so they may save some cycles 

	_csix_tx_init_gprs[ tcw_fcip_base, tcw_nfcip_base, tx_request_ring, tx_counter_base,\
			   s_txc_base, indiref_base, sigmask2_default ]

	// The first time through the loop, wait for next thread signal only 

	ctx_arb[next_thread_signal_1]


phase1#:

	// At this point the meta data i/o is finished and the signals for dram and
	// sram have been received. let the assembler know

	.io_completed $$dl_meta[0] $$dl_meta[1] $$dl_meta[2] $$dl_meta[3] $$dl_meta[4] $$dl_meta[5] $$dl_meta[6]
	.io_completed dram_to_tbuf_signal meta_read_signal
#ifdef COUNTERS
	.set $byte_cnt
#endif //COUNTERS


	// set up default signal masks for phase 1 and 2 

	alu[sigmask_phase1, --, B, SIG_MASK_PHASE1]
	alu[sigmask_phase2, --, B, sigmask2_default]

	// signal the next thread to enter phase 1 (3) 

	_csix_tx_signal_next[_NEXT_THREAD_SIGNAL1]

transmit_tbuf_setup#:

	// check if the packet n-1 is NULL. If it is, skip the transmit of the tbuf

	alu[--, prev_buffer_handle, - , 0x0]
	beq [no_transmit#]

	// allocate two write xfer registers for the transmit control 
	
	xbuf_alloc[$tcw, 2, write]	; This line doesn't produce any code.

	// transmit the tbuf for packet n-1. This macro also checks if the 
	// buffer handle is done and if so it frees it. (8)

	_csix_tx_transmit_tbuf[ $tcw, addr_of_tx_control, tcw, ex_header, \
							prev_buffer_handle, cell_count_remaining, \
							tcw_write_signal]

#ifdef COUNTERS
	_csix_tx_update_counters($byte_cnt, tcw, ex_header, prev_buffer_handle, \
								tx_counter_base, counter_update_signal)
	// add update_counter_signal to sigmask_phase1
	_csix_tx_set_signal[sigmask_phase1, _COUNTER_UPDATE_SIGNAL]
#endif // COUNTERS

transmit_tbuf_setup_done#:

	// check if the packet n is NULL. If it is, skip the dram to tbuf 

	alu[--, tx_request0, -, 0]
	beq[no_dram_to_tbuf#]

	_csix_tx_debug_incr_counter(@csix_tx_rcvd_cells)

	// get the CSIX extension header from the tx request before we overwrite it
	// in phase 2 (3)

	_csix_tx_get_extension_header[ ex_header, tx_request0]

	// allocate a new tbuf (9)

	_csix_tx_get_next_tbuf[addr_of_tbuf, addr_of_tx_control]

	// allocate 4 write xfer registers for writing the prepend to the tbuf

	xbuf_alloc[$prepend, 4, write]

	// Check if this is a packet with SOP/EOP bit set and can fit into one 
	// cframe (0xc0 in the MSB of $tx_request_lw1), (1)

	br!=byte[tx_request1, 3, 0xC0, dram_multi_cframe_packet#] , defer [2]

		// copy the $tx_request[1] to the prev_buffer_handle for next iteration
		// and set up cell count reamining to 0 by default 
	
		alu[prev_buffer_handle, --, B, tx_request1] 
		alu[cell_count_remaining, --, B, 0]

	// copy data from dram to tbuf for packet n. The function return a tcw 
	// for later use (13). tx_request1 contains the buffer handle



	_csix_tx_move_sop_eop_dram_to_tbuf[ tcw, align_offset, tcw_fcip_base, \
										tx_request1, addr_of_tbuf, \
										dram_to_tbuf_signal ]

	// write tm header to tbuf for packet n (10)

	_csix_tx_write_tm_header_to_tbuf[   $prepend, align_offset, addr_of_tbuf, \
										tmh_write_signal , SOP_EOP ]
	
dram_move_done#:

	// read tranmsit request for packet n+1 (1) 

	_csix_tx_read_transmit_request[tx_request, tx_request_ring, scratch_read_signal]

	// wait for read transmit request and transmit tbuf to finish and signal from
	// next thread (4)

	local_csr_wr[active_ctx_wakeup_events, sigmask_phase1]		
	ctx_arb[--] 

	// At this point io is complete for $prepend, $tcw and $tx_request. 
	// let the assembler know
#if IS_IXPTYPE(__IXP28XX)  //NEW_28XX_CODE

	.io_completed 	$prepend[0] $prepend[1] $prepend[2] $prepend[3] \
					$tcw[0] $tcw[1] 
#else

	.io_completed 	$prepend[0] $prepend[1] $prepend[2] $prepend[3] \
					$tcw[0] $tcw[1] $tx_request[0] $tx_request[1]

#endif

	// At this point we are done with the signals from phase 1, let the assembler 
	// know

	.io_completed scratch_read_signal tcw_write_signal tmh_write_signal

#ifdef COUNTERS
	.io_completed $byte_cnt
	.io_completed counter_update_signal
#endif //COUNTERS

	// dont need the prepend registers any more

	xbuf_free[$prepend]

	// deallocate the transfer registers

	xbuf_free[$tcw]

phase2#:

	// signal the next thread to enter phase 2 (3)

	_csix_tx_signal_next[_NEXT_THREAD_SIGNAL2]

	// check for a null request in the scratch ring (2)

	alu[--, tx_request0, -, 0x0]
	beq[no_valid_tx_request#]


	// we have a valid 	tx request, increment debug counter
	_csix_tx_debug_incr_counter[@csix_tx_num_tx_req]
	
	// Check if this is a packet with SOP/EOP bit set and can fit into one 
	// cframe (0xc0 in the MSB of $tx_request_lw1), (1)

	br!=byte[tx_request1, 3, 0xC0, sram_multi_cframe_packet#]

	// read the meta data and context from SRAM for packet n+1 (3). Buffer
	// handle is in $tx_request[1]. We read in 6 long words starting with the
	// second long word of the meta data. The first long word contains the
	// next pointer which is not needed

	dl_meta_load_cache[tx_request1, dl_meta, meta_read_signal, 1, 6]

	// wait for dram to tbuf for packet n and read of meta for packet n+1 to finish 
	// and signal from next thread (5)

	local_csr_wr[active_ctx_wakeup_events, sigmask_phase2]		
	ctx_arb[--] , br [phase1#]

no_valid_tx_request#:

	// clear the signal bit in sig_mask_phase2

	_csix_tx_clear_signal[ sigmask_phase2, _META_READ_SIGNAL]
	
	// wait for dram to tbuf for packet n 

	local_csr_wr[active_ctx_wakeup_events, sigmask_phase2]		
	ctx_arb[--] , br [phase1#]
	

no_transmit#:

	// clear the signal bit in sig_mask_phase1

	_csix_tx_clear_signal[sigmask_phase1, _TCW_WRITE_SIGNAL]


	// jump to point where the tbuf transmit is over 

	br[transmit_tbuf_setup_done#]

no_dram_to_tbuf#:

	// clear the signal bit in sig_mask_phase2 for dram to tbuf

	_csix_tx_clear_signal[sigmask_phase2, _DRAM_TO_TBUF_SIGNAL_1]

	// clear the signal bit in sig_mask_phase2 for dram to tbuf + 1

	_csix_tx_clear_signal[sigmask_phase2, _DRAM_TO_TBUF_SIGNAL_2]

	// clear the signal bit in sig_mask_phase1 for  

	_csix_tx_clear_signal[sigmask_phase1, _TMH_WRITE_SIGNAL]

	// copy null to the prev_buffer_handle 
	
	alu[prev_buffer_handle, --, B, 0] 

	// jump to where the dram move is done

	br[dram_move_done#]

	
dram_multi_cframe_packet#:

	 br_bclr[tx_request1, SOP_BIT_LOC, not_sop_cell#]

	// get remaining cell count from $tx_request[1] for later use

	alu_shf[ cell_count_remaining, CELL_COUNT_REMAINED_MASK, AND, \
	   		 tx_request1, >>CELL_COUNT_LOC ]

	// copy data from dram to tbuf for packet n. The function return a tcw 
	// for later use.  tx_request1 contains the buffer handle (22)
	

	_csix_tx_move_sop_dram_to_tbuf[ tcw, tcw_fcip_base, tx_request1, addr_of_tbuf, \
									dram_to_tbuf_signal]

	// write tm header to tbuf for packet n (10)

	_csix_tx_write_tm_header_to_tbuf[   $prepend, 0, addr_of_tbuf, \
										tmh_write_signal, SOP_ONLY ]

	// jump to where the DRAM move is done

	br[dram_move_done#]

not_sop_cell#:

	// copy data from dram to tbuf for packet n. The function return a tcw 
	// for later use. tx_request1 contains the buffer handle

	_csix_tx_move_mop_eop_dram_to_tbuf[ cell_count_remaining, tcw, tcw_nfcip_base, \
										tx_request1, addr_of_tbuf, \
										dram_to_tbuf_signal ]

	// write tm header to tbuf for packet n (10)

	_csix_tx_write_tm_header_to_tbuf[   $prepend, 0, addr_of_tbuf, \
										tmh_write_signal, MOP_EOP ]

	// jump to where the DRAM move is done

	br[dram_move_done#]


sram_multi_cframe_packet#:

	// Do a CAM lookup, get the txc context and meta data. This will also 
	// do a ctx arb to finish phase 2

	_csix_tx_get_context[ tx_request0, tx_request1, s_txc_base, sigmask_phase2]

	// go to the start of phase 1

	br[phase1#]

.end

#endm


/////////////////////////////////////////////////////////////////////////////////////

// The main code starts here

main#:

	#ifdef _DEBUG_COUNTERS_
	.reg @csix_tx_rcvd_cells
	immed[@csix_tx_rcvd_cells, 0]
	#endif // _DEBUG_COUNTERS_

	// initialize the microblock

	csix_tx_init[]

	// call the packet processing block

	csix_tx[]

/////////////////////////////////////////////////////////////////////////////////////

#endif 		// __CSIX_TX_UC__

/////////////////////////////////////////////////////////////////////////////////////


      

