/////////////////////////////////////////////////////////////////////////////////////
//
//
//                  I N T E L   P R O P R I E T A R Y
//
//     COPYRIGHT [c]  2002 BY  INTEL  CORPORATION.  ALL RIGHTS
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT
//     THE PRIOR WRITTEN PERMISSION OF :
//
//                        INTEL  CORPORATION
//
//                     2200 MISSION COLLEGE BLVD
//
//               SANTA  CLARA,  CALIFORNIA  95052-8119
//
/////////////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: csix_tx1.uc
//
//      Purpose: a context pipe-stage microblock which receives tansmit request
//               from Queue Manager, prepare CSIX base header, CSIX extension
//               header, and additional Traffic Manager header (TM header, in
//               our case, this is the info which ingress Castine wants to
//               pass to the egress Castine, which includes destination L2
//               port id, next hop, sequence number, and flags of SOP & EOP
//               now). It then sends request to the next ME that sends CSIX payload 
//				 (TM header and payload from packet
//               buffer) and CSIX headers through MSF Transmit State Machine.
//
//
/////////////////////////////////////////////////////////////////////////////////////

#ifndef __CSIX_TX1_UC__
#define __CSIX_TX1_UC__

/////////////////////////////////////////////////////////////////////////////////////
//
// The CSIX transmit microblock for Castine is a context pipe-stage that
// runs on a two  microengines in a context pipeline. It receives transmit requests 
// from the Queue Manager. 
// 
// With each transmit request, the microblock prepares a Transmit
// Control word (which include a CSIX base header and extension header info),
// and prepares an additional Traffic Manager header (TM header). The TM
// header is the info which ingress Castine wants to pass to the egress
// Castine, and includes the destination L2 port id, next hop, cell sequence
// number in one packet, and flags of SOP & EOP now). 
//
// The microblock moves the CSIX payload (TM header and payload from packet buffer) 
// into a TBUF element, and then sends this CSIX Frame to the fabric through the 
// MSF Transmit State Machine by writing the TCW to the corresponding
// TBUF_ELEMENT_CTRL_V_# address.  
//
// Every transmit request has an associated packet, which is being segmented into 
// cframes. The transmit context descripter (TXC which includes associated 
// segmentation state for the packet and the packet meta data) is cached in local 
// memory and is looked up using the CAM to speedup the processing.  If an entire 
// buffer for a packet has been transmitted, then the buffer is freed to the buffer 
// freelist.
//
// Assumptions:
//	1) Xscale core initialized transmit context (TXC) to all zero in SRAM
//	2) Queue Manager (QM) created transmit request in scratch ring
//  3) TBUF element size is 128 bytes
//  4) TM header size for the first cell in one packet is 16 bytes
//  5) TM header size for the cell which is not the first cell in one 
//     packet is 8 bytes
//  6) fabric ingress blade number is imported from Xscale core
//  7) TXC in local memory starts at local memory address 0, each TXC occupies
//     32 bytes in sram, but it occupies 64 bytes in local memoy to match 
//     cam_looup[.....], lm_addr0[0] operation
//  8) In the first longword of a valid tx request issued by QM, valid bit is
//     always set, and bit 30 to bit 16 are always zero, so we can use this
//     longword to do CAM lookup based on the queue number
//
/////////////////////////////////////////////////////////////////////////////////////

// NN ring msg format sent to next ME
// lw0 - tx_request0
// lw1 - dram_addr
// lw2 - 0:15 payload_len , 16 (0 is SOP_EOP), 17 ( 1 if MOP_EOP) 24:31 - cell_count_remaining
// lw3-6 - prepend0-3

// following definition selects IXP24000 definition in hardware.h


// include stdmac.uc in IXPblocks Portable library

#include <stdmac.uc>

// include local memory macros in IXPblocks 

#include <localmem.uc>

// global definition of common constants which are shared by all microblocks

#include "dl_system.h"

// definition of hardware register address 

#include <hardware.h>


// definition of constants used

#include "csix_tx.h"

// definition of macros used

#include "csix_tx_util1.uc"

// initialization code for CSIX TX

#include "csix_tx_init1.uc"

// dispatch loop macros

#include "dispatch_loop.uc"

#ifdef	_DEBUG_COUNTERS_
// The following counters are stored in absolute registers. These counters are
// primarily used in debugging.
.reg	@csix_tx_rcvd_cells				
.reg	@csix_tx_num_tx_req				// number of valid transmit requests
#endif
														
/////////////////////////////////////////////////////////////////////////////////////
//
// Packet processing entry point for csix microblock
//
//
/////////////////////////////////////////////////////////////////////////////////////

#macro csix_tx1[]

.begin

	.reg		addr_of_tbuf			; address of tbuf
	.reg		addr_of_tx_control		; address of TBUF_ELEMENT_CTRL_V_#
	.reg		prev_buffer_handle		; buffer handle for packet n-1
	.reg		tcw						; transmit control word
	.reg		ex_header				; extension header
	.reg		sigmask_phase1			; signal mask for phase 1
	.reg		sigmask_phase2			; signal mask for phase 2
	.reg		tmp						; temporary register
	.reg		tcw_fcip_base			; base for tcw for first cell of packet
	.reg		tcw_nfcip_base			; base for tcw for not first cell of packet
	.reg		indiref_base			; base for indirect ref of dram 
	.reg		s_txc_base				; base for tx context in sram
	.reg		cell_count_remaining	; cell count remaining in buffer
	.sig		zero_read_signal		; signal to read in zero's into tx_request
	.reg		lm_global_base			; base in local memory for globals 
	.reg		align_offset			; offset into quad word for alignment
	.reg		sigmask2_default 		; default value for signal mask for phase 2
	.reg		next_thread_signal_value; used to signal next thread in one instruction
	
	// initialize next thread signal value

	alu[tmp, --, B, 0x80]
	alu_shf[next_thread_signal_value, tmp, OR, &next_thread_signal_1, <<3]	

	// initialize local memory index 1 to the base for globals 

	immed32[lm_global_base, GLOBAL_LM_BASE]

	// 3 cycles for this to take effect

	localmem_set_address[0, lm_global_base, LM_HANDLE_1]

	nop
	nop
	nop

	// allocate 2  registers for reading the transmit request

	.reg tx_request0 tx_request1
	
	// allocate 6 read xfer registers for reading in the meta data 

	xbuf_alloc[$$dl_meta, 7, read]
	
	// For the initial run let the assembler think the meta data is loaded 

	.set $$dl_meta[0] $$dl_meta[1] $$dl_meta[2] $$dl_meta[3] $$dl_meta[4] $$dl_meta[5] $$dl_meta[6]


	// set previous buffer handle and cell count remaining to 0

	alu[prev_buffer_handle, --, B, 0] 
	alu[cell_count_remaining, --, B, 0]

	// To get rid of warnings

	alu[tcw, --, B, 0]
	alu[ex_header, --, B, 0]
	alu[addr_of_tx_control, --, B, 0]

	immed[tx_request0, 0]
	immed[tx_request1, 0]

	// gprs initialized so they may save some cycles 

	_csix_tx_init_gprs1[ tcw_fcip_base, tcw_nfcip_base,\
			   s_txc_base, indiref_base, sigmask2_default ]

	// The first time through the loop, wait for next thread signal only 

	ctx_arb[next_thread_signal_1]

	// signal the next thread to enter the critical section 

	_csix_tx_signal_next[next_thread_signal_value]

	// At this point the meta data i/o is finished and the signals for dram and
	// sram have been received. let the assembler know

	.io_completed $$dl_meta[0] $$dl_meta[1] $$dl_meta[2] $$dl_meta[3] $$dl_meta[4] $$dl_meta[5] $$dl_meta[6]
	.io_completed dram_to_tbuf_signal meta_read_signal

	// allocate 4  registers for writing the prepend to the tbuf

	.reg prepend0 prepend1 prepend2 prepend3
	.reg dram_addr payload_length
	
read_next_nn_request#:

	// read tranmsit request for packet n+1 (3) 

	_csix_tx_read_transmit_request_nn[tx_request, no_valid_tx_request#]	

	// we have a valid 	tx request, increment debug counter

	_csix_tx_debug_incr_counter[@csix_tx_num_tx_req]
	_csix_tx_debug_incr_counter(@csix_tx_rcvd_cells)

	// Check if this is a packet with SOP/EOP bit set and can fit into one 
	// cframe (0xc0 in the MSB of $tx_request_lw1), (1). The defer (1) is 
	// because cam lookup cant follow br=byte unless it is in the defer slot

	br=byte[tx_request1, 3, 0xC0, single_cframe_packet#], defer[1]

	// Multi-cframe case is actually the worst case path - 111 bytes where
	// the packet splits into 2 tbufs. 

	// Do a CAM lookup, get the txc context and meta data. This will also 
	// do a ctx arb to finish phase 2

	_csix_tx_get_context[ tx_request, s_txc_base, sigmask_phase2]

	// signal the next thread to enter the critical section. 

	_csix_tx_signal_next[next_thread_signal_value]

	// Check for SOP cell 
	
	br_bclr[tx_request1, SOP_BIT_LOC, not_sop_cell#]

	// get remaining cell count from tx_request1 for later use

	alu_shf[ cell_count_remaining, CELL_COUNT_REMAINED_MASK, AND, \
	   		 tx_request1, >>CELL_COUNT_LOC ]

	// prepare to copy data from dram to tbuf for packet n. The function return a tcw 
	// for later use.  tx_request1 contains the buffer handle (22)
	
	_csix_tx_prepare_to_move_sop_dram_to_tbuf[ dram_addr, payload_length, tx_request1]

	// prepare to write tm header to tbuf for packet n (10)

	_csix_tx_prepare_to_write_tm_header_to_tbuf[   prepend, 0,  SOP_ONLY ]

	// write the request to the next ME for further processing (10)

	_csix_tx_write_nn_request[tx_request0, tx_request1, dram_addr, payload_length, SOP_ONLY, cell_count_remaining, prepend]

	// jump to reading next packet 

	br[read_next_nn_request#]

not_sop_cell#:

	// prepare to copy data from dram to tbuf for packet n. The function return a tcw 
	// for later use. tx_request1 contains the buffer handle (16)

	_csix_tx_prepare_to_move_mop_eop_dram_to_tbuf[ cell_count_remaining, dram_addr, payload_length, \
											tx_request1]

	// prepare to write tm header to tbuf for packet n (10)

	_csix_tx_prepare_to_write_tm_header_to_tbuf[   prepend, 0,  MOP_EOP ]

	// write the request to the next ME for further processing	(10)

	_csix_tx_write_nn_request[tx_request0, tx_request1, dram_addr, payload_length, MOP_EOP, cell_count_remaining, prepend]

	// jump to reading the next request 

	br[read_next_nn_request#]

single_cframe_packet#:

	// This case actually only takes 45 instructions. So moved the other
	// case into the no-penalty side of the branch. 

	// Read the meta data and context from SRAM for packet n+1 (3). Buffer
	// handle is in $tx_request1. We read in 6 long words starting with the
	// second long word of the meta data. The first long word contains the
	// next pointer which is not needed

	dl_meta_load_cache[tx_request1, dl_meta, meta_read_signal, 1, 6]
	
	// wait for read of meta to finish and signal from next thread
		
	ctx_arb[meta_read_signal, next_thread_signal_1] 

	// signal the next thread to enter the critical section 

	_csix_tx_signal_next[next_thread_signal_value]

	// prepare to copy data from dram to tbuf for packet n. The function return a tcw 
	// for later use. It also returns dram_addr to pass to 2nd ME.
	// tx_request1 contains the buffer handle (7)

	_csix_tx_prepare_to_move_sop_eop_dram_to_tbuf[ dram_addr, payload_length, align_offset, \
											tx_request1 ]

	// prepare to write tm header to tbuf for packet n (10) -- it returns prepend header to be sent to next ME

	_csix_tx_prepare_to_write_tm_header_to_tbuf[  prepend, align_offset, SOP_EOP ]

	// write the request to next ME that will perform dram t0 tbuf write and 
	// tbuf validate (9)

	_csix_tx_write_nn_request[tx_request0, tx_request1, dram_addr, payload_length, \
								SOP_EOP, cell_count_remaining, prepend]

	// jump to reading next packet 

	br[read_next_nn_request#]

no_valid_tx_request#:
	
	// wait for the inter thread signal only 

	ctx_arb[next_thread_signal_1] 

	// signal the next thread to enter the critical section 

	_csix_tx_signal_next[next_thread_signal_value]

	// read the next request 

	br[read_next_nn_request#]

.end

#endm


/////////////////////////////////////////////////////////////////////////////////////

// The main code starts here

main#:

	// initialize the microblock

	csix_tx_init1[]

	// call the packet processing block

	csix_tx1[]

/////////////////////////////////////////////////////////////////////////////////////

#endif 		// __CSIX_TX1_UC__

/////////////////////////////////////////////////////////////////////////////////////


      

