/////////////////////////////////////////////////////////////////////////////////////
//                                                                      
//                   I N T E L   P R O P R I E T A R Y                   
//                                                                       
//      COPYRIGHT (c)  2002 BY  INTEL  CORPORATION.  ALL RIGHTS          
//      RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//      BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//      RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//      LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//      MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//      THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                       
//                         INTEL  CORPORATION                            
//                                                                      
//                      2200 MISSION COLLEGE BLVD                        
//                                                                       
//                SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                       
//
/////////////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: csix_tx_init.uc
//
//      Purpose: Initialization code for CSIX TX
//
//      History:
//
//      Date            Comment                  				       By
//      ---------------------------------------------------------------------
//
//      01/11/2002      Created                  				       David Chou
//
//		03/25/02		modified for packet interleaving
//						and coding convention						   Uday Naik
//
/////////////////////////////////////////////////////////////////////////////////////

#ifndef __CSIX_TX_INIT1_UC__
#define __CSIX_TX_INIT1_UC__


/////////////////////////////////////////////////////////////////////////////////////
//
// Following definition enable CSIX TX waiting for an inter-ME signal 
// generated by the ME whcih initializes common resources, such as scratch ring,
// near the end of thread 0 of CSIX TX initialization to make sure all common 
// resource are available 
//
/////////////////////////////////////////////////////////////////////////////////////

 
#define WAIT_FOR_COMMON_RESOURCE_INITIALIZATION 	




/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_init_globals
//
// Description:
// 	
//		Initialize the globals stored in local memory
//
// Outputs: 
//
//
// Inputs:  
//
//		
// Constants
//
//		None
//
// Size: 
//
//		14 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_init_globals[]

.begin
	
	.reg	lm_global_base

	// initialize local memory index 1 to the base for globals 

	immed32[lm_global_base, GLOBAL_LM_BASE]

	// 3 cycles for this to take effect

	localmem_set_address[0, lm_global_base, LM_HANDLE_1]

	nop
	nop
	nop

	// available tbuf element initialized to 0

	move[*l$index1[GLOBAL_AVAIL_TBUF_ELEMENT_INDEX], 0]	
   
	// initialized to 128 to help in signal computation

	move[*l$index1[GLOBAL_NUMBER_128_INDEX], 0x80]

	// mask to set the bd valid bit = 0 and bytes already sent in this buffer to 0

    move[*l$index1[GLOBAL_MASK_BD_BYTES_SENT_INDEX], MASK_OUT_BD_BYTES_SENT_LW]

	// starting address of TBUF_ELEMENT_CTRL_V_# 

   	move[*l$index1[GLOBAL_CTW_VD_BASE_INDEX], TBUF_ELEMENT_CTRL_VALID]

 	// starting address of TBUF 

    move[*l$index1[GLOBAL_TBUF_BASE_INDEX], RBUF_TBUF_BASE] 
	
.end

#endm

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_init_cam
//
// Description:
// 	
//		Initialize the cam
//
//
// Outputs: 
//
//
// Inputs:  
//
//		
// Constants
//
//		None
//
// Size: 
//
//		11 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_init_cam[]

.begin

	.reg 		entry
	.reg 		tag
	.reg 		tmp
	.reg 		val

 	// CAM initialization

	cam_clear

	// Since the CSIX TX just run on one ME, It will be more efficient to use CAM 
	// without reference count. The simplest way is to swap out LRU, when a miss
	// happened. In order to keep valid trnasmit context (TXC) in sram, when first
	// 16 swapout happened, we need to initialize CAM and LM entry to make the 
	// first 16 swapout not generating invalid TXC in sram. Initialize CAM entries 
	// to first 16 queue number and corresponding 64 LM dword to 0


	// initialize CAM entries

    immed[entry, 0x0]

	// initialize CAM entries corresponds to queue 0 to 15

	immed[tag, 0x0]
	immed_w1[tag, 0x8000]

init_cam#:

    // These 16 initial CAM entries are dummy, and we update CAM state with 
	// CAM_STATE_INIT 

	cam_write[entry, tag, CAM_STATE_INIT]

    alu[entry, entry, +, 1] 
	alu[tag, tag, +, 1]

	br!=byte[entry,0, 0x10, init_cam#] 

.end

#endm

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_init_local_memory
//
// Description:
// 	
//		Initialize the local memory for contexts
//
// Outputs: 
//
//
// Inputs:  
//
//		
// Constants
//
//		None
//
// Size: 
//
//		14 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_init_local_memory[]

.begin

	.reg 		i
	.reg		j
	.reg		lm_txc_base

	// initialize local memory index 0 to the base for transmit contexts

	immed32[lm_txc_base, TXC_LM_BASE]

	// 3 cycles for this to take effect

	localmem_set_address[0, lm_txc_base, LM_HANDLE_0]

	nop
	nop
	nop

	move[i,0]

	// Initialize 16 contexts

	.while (i < 16)	
		 
		move[j, 0]
		
		// initialize first 7 longwords of a context to 0

		.while (j < 7)

			alu[*l$index0++, --, B, 0]
			alu[j, j, +, 1]

		.endw

		// set the eigth entry in the context to a invalid buffer handle
		// so that meta data will be read in even if we get a cam hit

		alu[*l$index0++, --, B, IMPOSSIBLE_DL_HDL_ADDR]
		alu[i, i, +, 1]

		move[j, 0]
		
		// initialize next 8 longwords  of a context to 0

		.while (j < 8)

			alu[*l$index0++, --, B, 0]
			alu[j, j, +, 1]

		.endw

	.endw

.end

#endm


/////////////////////////////////////////////////////////////////////////////////////
// 
// csix_tx_init[]
//
// Description:
// 	
//		Initialization of the microblock
//
// Outputs: 
//
//
// Inputs:  
//
//		
// Constants
//
//		None
//
// Size: 
//
//		14 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro csix_tx_init1[]

.begin

	.reg 	sig
	
	// For thread 0 we will do some initialization. every one else waits

	br!=ctx[0, EXIT#]

ctx0_init#:
	
	// init NN ring
	.reg ctx_enable
	local_csr_rd[ctx_enables]
	immed[ctx_enable, 0]
	alu_shf[ctx_enable, ctx_enable, AND~, 3, <<18]
	alu_shf[ctx_enable, ctx_enable, OR, 1, <<18]
	local_csr_wr[ctx_enables, ctx_enable]
	local_csr_wr[nn_put, 0]
	local_csr_wr[nn_get, 0]	
	alu[--, --, b, 0]
	alu[--, --, b, 0]
	alu[--, --, b, 0]

    // Initialize MSF_TX_CNTL register

	//_csix_tx_msf_tx_control_init[]

    // start to initialize global GPRs

	_csix_tx_init_globals[]

  	// Initialize the cam 

	_csix_tx_init_cam[]

	// initialize the local memory 

	_csix_tx_init_local_memory[]

	// initialize the debug counters

	_csix_tx_debug_init_counters[]
    

	// following inter-ME signal common_init_complete_sig_num is generated by 
	// the ME whcih initializes common resources, such as scratch ring, CSIX TX
	// ME will wait for this signal to make sure all common resource are 
	// available, when it start CSIX TX processing

#ifdef WAIT_FOR_COMMON_RESOURCE_INITIALIZATION 

	ctx_arb[init_complete_signal]

#endif 


    // For kick start thread 0 will send itself an initial signal 

	immed[sig, _NEXT_THREAD_SIGNAL1]		; prepare value for signaling
    alu_shf[sig, 0x0, OR, sig, <<3]         ; value to signal thread 0
    local_csr_wr[SAME_ME_SIGNAL, sig]   	; signal thread 0

	
EXIT#:


.end

#endm  

/////////////////////////////////////////////////////////////////////////////////////

#endif 		//  __CSIX_TX_INIT1_UC__

/////////////////////////////////////////////////////////////////////////////////////
