/////////////////////////////////////////////////////////////////////////////////////
//
//
//                  I N T E L   P R O P R I E T A R Y
//
//     COPYRIGHT [c]  2002 BY  INTEL  CORPORATION.  ALL RIGHTS
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT
//     THE PRIOR WRITTEN PERMISSION OF :
//
//                        INTEL  CORPORATION
//
//                     2200 MISSION COLLEGE BLVD
//
//               SANTA  CLARA,  CALIFORNIA  95052-8119
//
/////////////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: csix_tx_util.uc
//
//      Purpose: Utility macros for csix transmit 
//
//
//      History:
//
//      Date            Comment                         				By
//      ---------------------------------------------------------------------------
//
//      11/21/2001      Created                         			David Chou
//
//		3/18/02			Modified to do packet interleaving			Uday Naik	
//						and adhere to coding convention
//
/////////////////////////////////////////////////////////////////////////////////////

#ifndef __CSIX_TX_UTIL_UC__
#define __CSIX_TX_UTIL_UC__

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_read_transmit_request[]
//
// Description:
// 	
//		Read the transmit request from a scratch ring
//
// Outputs: 
//
//		$tx_request:		2 xfer registers that will contain the tx request
//							when the I/O completes
//
// Inputs:  
//
//		ring:				scratch ring number to read from
//		signal_number:		signal number to wait on
//
// Constants
//
//		None
//
// Size: 
//
//		1 instruction
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_read_transmit_request[tx_request, ring, signal_number]

.begin

#if IS_IXPTYPE(__IXP28XX)  //NEW_28XX_CODE

	alu[tx_request/**/0, --, b, 0]
	br_inp_state[nn_empty, next_ngbr_empty#]
		read_from_nn#:
			alu[tx_request/**/0, --, b, *n$index++]
			alu[tx_request/**/1, --, b, *n$index++]
	next_ngbr_empty#:

#else
	// get tx request - 2 words from scratch ring 

	scratch[get, tx_request/**/0, ring, 0, 2], sig_done[signal_number]

#endif

.end

#endm

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_get_next_tbuf[]
//
// Description:
// 	
//		Allocate a tbuf element. Instead of returning the element number we return
//		two values which are actually useful - the element address and the address 
//		for validating the tbuf. This allows us to optimize the wrap around case 
//		better.
//
// Outputs: 
//
//		addr_of_tbuf:		Address of the TBUF for data copy
//		addr_of_tx_control:	Addr for validating the tbuf
//
// Constants
//
//		None
//
// Size: 
//
//		9 instructions in worst case 
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_get_next_tbuf[addr_of_tbuf, addr_of_tx_control]

.begin	

	.reg	tbuf_element
	.reg	tmp
	
	// get the current tbuf element
  
	alu[tbuf_element, --, B, *l$index1[GLOBAL_AVAIL_TBUF_ELEMENT_INDEX]]

	// increase @avail_tbuf_element by 1

	alu[*l$index1[GLOBAL_AVAIL_TBUF_ELEMENT_INDEX], \
						*l$index1[GLOBAL_AVAIL_TBUF_ELEMENT_INDEX], +, 1] 

	// Check if @avail_tbuf equal to MAX_AVAIL_DATA_TBUF

	alu[--, *l$index1[GLOBAL_AVAIL_TBUF_ELEMENT_INDEX], -, MAX_AVAIL_DATA_TBUF]

	blt[EXIT#] , defer [3]
		
		// compute the address of the tbuf 

		alu_shf[tmp, --, B, tbuf_element, <<7]
		alu[addr_of_tbuf, tmp, +, *l$index1[GLOBAL_TBUF_BASE_INDEX]]

		// start the compute of the address for tx control 

#ifdef IXP2800
		alu_shf[tmp, --, B, tbuf_element, <<4]
#else
		alu_shf[tmp, --, B, tbuf_element, <<3]
#endif

	// if we get here, we need to reset to the first element. 

	alu[*l$index1[GLOBAL_AVAIL_TBUF_ELEMENT_INDEX] , --, B, FIRST_AVAIL_DATA_TBUF]

EXIT#:

	// finish the compute of the address of the tx control 

	alu[addr_of_tx_control, tmp, +, *l$index1[GLOBAL_CTW_VD_BASE_INDEX]]

.end

#endm

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_signal_next[]
//
// Description:
// 	
//		signal the next thread
//
// Outputs: 		
//
// Inputs:  
//
// Constants
//
//		SIG_NUM		signal number
//
// Size: 
//
//		2 instruction
// 
//
/////////////////////////////////////////////////////////////////////////////////////


#macro _csix_tx_signal_next[SIG_NUM]

.begin

	.reg temp

	// signal number goes into bits 3..6 and bit 7 is set to signal the next context.
	// We store 0x80 in local memory to optimize

	alu_shf[temp, *l$index1[GLOBAL_NUMBER_128_INDEX], OR, SIG_NUM, <<3]	
	local_csr_wr[SAME_ME_SIGNAL, temp]		

.end

#endm

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_clear_signal[]
//
// Description:
// 	
//		Clear a signal from a signal mask 
//
// Outputs: 
//
//
// Inputs:  
//
//		sigmask:			signal mask
//		sig:				signal number to clear
//
// Constants
//
//		None
//
// Size: 
//
//		1-3 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_clear_signal[sigmask, sig]

.begin

	.reg tmp

#if ( isnum(sig) )
	
	#define_eval SIG_SHIFT sig
	alu_shf[sigmask, sigmask, AND~, 1, <<SIG_SHIFT ]
	#undef SIG_SHIFT

#else

	alu[tmp, &sig, -, 1]
	alu[--, tmp, OR, 0]										
	alu_shf[sigmask, sigmask, AND~, 1, <<indirect]

#endif

.end

#endm

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_set_signal[]
//
// Description:
// 	
//		Set a signal in a signal mask 
//
// Outputs: 
//
//
// Inputs:  
//
//		sigmask:			signal mask
//		sig:				signal number to set
//
// Constants
//
//		None
//
// Size: 
//
//		1-3 instructions
// 
//
////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_set_signal[sigmask, sig]

.begin

	.reg tmp

#if ( isnum(sig) )

	#define_eval SIG_SHIFT sig
	alu_shf[sigmask, sigmask, OR, 1, <<SIG_SHIFT]
	#undef SIG_SHIFT

#else

	alu[tmp, &sig, -, 1]
	alu[--, tmp, OR, 0]										
	alu_shf[sigmask, sigmask, OR, 1, <<indirect]


#endif

.end

#endm

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_transmit_tbuf[]
//
// Description:
// 	
//		Validate a TBUF for transmit. Also if the cell count has gone to 0
//		then free the buffer  
//
// Outputs: 
//
// Inputs:  
//
//		$tcw						xfer register
//		addr_of_tx_control:			address to validate tbuf
//		tcw:						transmit control word
//		ex_header:					extension header
//		buffer_handle:				buffer handle
//		cell_count_remained:		cell count remaining
//		signal_number:				signal number
//		
// Constants
//
//		None
//
// Size: 
//
//		8 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_transmit_tbuf[ $tcw, addr_of_tx_control, tcw, ex_header, \
								buffer_handle, cell_count_remained, signal_number ]

.begin

	// check if the cell count remaining for the buffer is 0

	br!=byte[cell_count_remained, 0, 0x00, exit#] , defer [2]

		// copy the element control word and the extension header into write
		// transfer registers

		alu[$tcw[0],	--, B, tcw]			; element control word
		alu[$tcw[1],	--, B, ex_header]	; element CSIX extension header

	// free the buffer handle 

	dl_buf_free[buffer_handle, BUF_FREE_LIST0]

	// increment debug counter for buffers freed
	_csix_tx_debug_incr_counter[@csix_tx_num_buf_freed]

exit#:

 	// write transmit contrl word (TCW) into MSF. write to TBUF_ELEMENT_CTRL_V_# 
	// corresponding the tbuf_element to initiate transmission.

	msf[write, $tcw[0], addr_of_tx_control, 0, 2], sig_done[signal_number]

	// increment debug counters for c-frames transmitted.
	_csix_tx_debug_incr_counter[@csix_tx_num_cells_txed]

.end

#endm

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_get_extension_header[]
//
// Description:
// 	
//		Compute the extension header from the transmit request word 0
//
// Outputs: 
//
//		ex_header					Extension header returned
//
// Inputs:  
//
//		tx_request0					transmit request 0
//		
// Constants
//
//		None
//
// Size: 
//
//		3 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_get_extension_header[ex_header, tx_request0]

.begin

	.reg 	class_id
	
	// tx_request0 contains the class id (0..3), the destination blade on the
	// fabric (4..9) and the valid bit in bit 31

	// class id is the first byte in the extension header for unicast cframe, it
	// is in bit 32 to bit 24, and  the destination blade is in bit 11 to bit 0


 	// shift by 4 bits and then and with 0x3f to get 6 bits of destination blade

    alu_shf[ex_header, DESTINATION_FIELD_MASK, AND, tx_request0, >>DEST_FIELD_SHIFT]	

	// get the class id which is in bits 0..3. And with 0xf

	alu[class_id, CLASS_FIELD_MASK, AND, tx_request0]

	// fill class field (bit 24 to 31) in extension header

    alu_shf[ex_header, ex_header, OR, class_id, <<CLASS_FIELD_SHIFT]


.end

#endm

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_move_sop_eop_dram_to_tbuf[]
//
// Description:
// 	
//		For a single cell packet (SOP/EOP both set) this macro sets up the dram
//		to tbuf transfer
//
// Outputs: 
//
//		tcw
// 		align_offset		offset into quadword for start of data
//
// Inputs:  
//
//		tcw_base			transmit control word base
//		buffer_handle		buffer handle
//		addr_of_tbuf		adress of tbuf
//		signal_number		signal to wait on 
//		
// Constants
//
//		None
//
// Size: 
//
//		13 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_move_sop_eop_dram_to_tbuf[  tcw, align_offset, tcw_base, \
											buffer_handle, addr_of_tbuf, \
											signal_number ]

.begin
	
	.reg	dram_addr			; dram address at which data starts
	.reg	buffer_offset		; offset into buffer at which payload starts
	.reg	tmp_addr_of_tbuf	; used for indirect ref calculation
	.reg	indir_ref			; used for indirect ref calculation
	.reg	payload_length		; payload length to be transmitted

  	// Get the dram base address for the buffer

    dl_buf_get_data[dram_addr, buffer_handle]

	// Get offset in buffer of the data 

	dl_meta_get_offset[buffer_offset]

	// get dram read address for buffer

	alu[dram_addr, dram_addr, +, buffer_offset] 

	// copy buffer size from meta data into payload length 

	dl_meta_get_buffer_size[payload_length]

	// compute offset into quadword for buffer offset

	alu[align_offset, buffer_offset, AND, 0x7]

	// increment the payload length by offset to round off to quad word boundary

	alu[payload_length, payload_length, + , align_offset]

	// Compute the transmit control word for later use with the payload length

	_csix_tx_get_transmit_control_word[tcw, tcw_base, 0, payload_length]
    
	// prepare adrress of tbuf elment for cell data as indirect token for
    // following dram[tbuf_write....] intsruction

	alu[tmp_addr_of_tbuf, addr_of_tbuf, +, 16]
	alu[indir_ref, indiref_base, OR, tmp_addr_of_tbuf, <<5]

    // since ref count 0 will fetch 8 bytes, and ref count 1 will fetch
	// 16 bytes......, and ref count 15 will fetch 128 bytes, 
	// the right way to do it is followinhg:
	//    a) decrease length by one
	//    b) then right shift the result by 3
	// We will do step a first, then combine the right shift in the shift into
	// the ref count in indir_ref.  In this way, we make sure that we fetch 
	// enough and closet to 8 byte alignment of packet data	
	
	alu[payload_length, payload_length, -, 0x01]
  	alu_shf[payload_length, --, B, payload_length, >>3]

    // left shift 21 bit to ref_cnt location

    alu[--, indir_ref, OR, payload_length, <<21] 
			 
    // how many bytes of packet payload are copied to tbuf from dram is 
	// controlled by ref cnt field in indir_ref, following max_14 is
	// dummy	 

    dram[tbuf_wr, --, dram_addr, 0, max_14], indirect_ref, sig_done[signal_number]

.end

#endm

/////////////////////////////////////////////////////////////////////////////////////
//	
//	_csix_tx_write_tm_header_to_tbuf[]
//
// Description:
// 	
//		Write the TM header to a tbuf
//
// Outputs: 
//
//		
//
// Inputs:  
//
//		$prepend			packet prepend xfer registers
// 		align_offset		offset into quadword for start of packet data
//		addr_of_tbuf		address of tbuf
//		signal_number		signal number to wait on
//		
// Constants
//
//		SOP_EOP_FLAG		indicates cell type
//
// Size: 
//
//		9 instructions 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_write_tm_header_to_tbuf[ $prepend, align_offset, addr_of_tbuf, \
										 signal_number, SOP_EOP_FLAG]

.begin
 
	.reg	prepend_header0
	.reg	tmp
	.reg	increment
	.reg	header_type
	.reg	output_port
	.reg	flow_id
	.reg	color
	.reg	nexthop_id
	.reg	class_id

	// ingress blade number should be imported from xscale core, since there
	// is no xscale now, we use THIS_BLADE_ID defined in system.h in 
	// dispatch_loop 

#if (SOP_EOP_FLAG != MOP_EOP)

	// This is the code for sop/eop or sop only cells. In this case we 
	// need a 2 word packet prepend and a 2 word frame prepend

#if (SOP_EOP_FLAG == SOP_EOP)

	// initialize prepend header word 0 with alignment offset and blade id 

	alu_shf[prepend_header0, THIS_BLADE_ID, OR, align_offset, <<24]

	// set the EOP, SOP flag here and copy into write transfer registers

	alu[$prepend[0], prepend_header0, OR, 0xC0]

#else

	// initialize prepend header word 0 with alignment offset and blade id 

	alu[prepend_header0, THIS_BLADE_ID, OR, *l$index0[TXC_SEQ_NUMBER_INDEX]]

	// set the SOP flag only 

	alu[$prepend[0], prepend_header0, OR, 0x40]

#endif

	// copy next hop id in from meta data 

    dl_meta_get_nexthop_id[nexthop_id]

	// copy class id from meta data 

    dl_meta_get_class_id[class_id]	
	
	// copy these into the prepend header 
		
	alu_shf[$prepend[1], class_id, or, nexthop_id, <<16]	

	// get header type from the meta data 

	dl_meta_get_header_type[header_type]

	// get output port from the meta data

	dl_meta_get_output_port[output_port]

	// copy output port and header type into prepend3

	alu[$prepend[2], output_port, OR, header_type, <<16]

	// get the flow id from the meta data

	dl_meta_get_flow_id[flow_id]

	// get the color id from the meta data 

	dl_meta_get_color[color]

	// copy flow id and color into prepend3

	alu[$prepend[3], flow_id, OR, color, <<24]

    // write TM header to tbuf using msf write 

	msf[write, $prepend[0], addr_of_tbuf, 0, 4], sig_done[signal_number]

#else

	// This is the case for MOP cells and EOP cells. We only need a 2 word
	// frame prepend in these cases

	// increment the seq number by 256 since we are storing in bit 8..24

	alu[increment, --, B, 1, <<8]

   	alu[*l$index0[TXC_SEQ_NUMBER_INDEX], *l$index0[TXC_SEQ_NUMBER_INDEX], \
		+, increment]
		
	// copy the port and sequence number into prepend0 from local memory. If this
	// is an EOP cell, that will also be set in the local memory field 

	ld_field_w_clr[tmp, 0111, *l$index0[TXC_SEQ_NUMBER_INDEX]]	
	alu[$prepend[0], THIS_BLADE_ID, OR, tmp]	

	// copy the prepend into transfer registers
	
    alu[$prepend[1],	--, B, *l$index0[TXC_NEXTHOP_CLASS_ID_INDEX]]		

	// write the prepend to the tbuf 

    msf[write, $prepend[0], addr_of_tbuf, 0, 2], sig_done[signal_number]

#endif

.end

#endm

/////////////////////////////////////////////////////////////////////////////////////
//	
//	_csix_tx_get_transmit_control_word[]
//
// Description:
// 	
//		Get transmit control word from offset and length
//
// Outputs: 
//
//		tcw					transmit control word
//
// Inputs:  
//
//		tcw_base			base for transmit control word
//		offset				offset into quadword to start transmit
//		payload_length		payload length
//		
// Constants
//
//
//
// Size: 
//
//		2 instructions 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_get_transmit_control_word[ tcw, tcw_base, offset, payload_length]

.begin

	// fill in the byte offset (0 to 7) to start transmit (13..15 of tcw)

	alu_shf[tcw, tcw_base, or, offset, <<13]

	// fill in the payload length into the transmit control word
	
	ld_field[tcw, 1000, payload_length, <<24] 


.end

#endm


/////////////////////////////////////////////////////////////////////////////////////
//
// _csix_tx_init_gprs
//
//	Description: initialize context relative GPR for specific thread
//
//	Size:
//
//		5 instructions
//
/////////////////////////////////////////////////////////////////////////////////////
 
#macro _csix_tx_init_gprs[ v_tcw0_fcip_base, v_tcw0_nfcip, scratch_ring, s_counter_base,\
							 s_txc_base, v_indiref_base, sigmask_phase2_default ]

.begin

	// base for the first transmit control word for the first cell of one packet 

	move[v_tcw0_fcip_base, TCW0_FCIP_BASE_LW]

	// base for the first transmit control word for cells which is not the first 
	// cell of one packet 

 	move[v_tcw0_nfcip, TCW0_NOT_FCIP_BASE_LW]

#if !IS_IXPTYPE(__IXP28XX)  //NEW_28XX_CODE
 	// scratch ring which QM put tx requests
    // scratch instruction requires context relative GPR
	// Scratch for communicating between Queue Manager and CSIX Transmit Blocks
	// QM_TO_CSIX_TX_SCR_RING is defined in the common header file "system.h"
	// in ..\dispatch_loop

	alu_shf[scratch_ring, --, b,  QM_TO_CSIX_TX_SCR_RING, <<2]
 
	// get the counter base for CSIX Transmit

	immed32[s_counter_base, CSIX_TX_COUNTERS_SRAM_BASE]
 
#endif

	// following s_txc_base needed to be context relative
    // Base address of CSIX Transmit Contexts in SRAM (CSIX_TX_CTX_SRAM_BASE)
	// is defined in the common header file "system.h" in ..\dispatch_loop

	immed32[s_txc_base, CSIX_TX_CTX_SRAM_BASE]

	// base to calculate the address field in indirect ref of 
	// dram[tbuf_wr....] instruction

	move[v_indiref_base, DRAM_INDIREF_BASE_LW]

	// default phase 2 signal mask at start of each iteration of loop 

	move[sigmask_phase2_default, SIG_MASK_PHASE2]

.end

#endm 

////////////////////////////////////////////////////////////////////////////////////
//
// _csix_tx_get_context[]
//
// Description:
// 	
//		swap out context into sram
//
// Outputs: 
//
//		
//
// Inputs:  
//
//		$tx_request			transmit request
//		sram_txc_base		base for sram contexts
//		sigmask				signal mask
//		
// Constants
//
//		None
//
// Size: 
//
//		26 instructions
//	
//
/////////////////////////////////////////////////////////////////////////////////////
 
#macro _csix_tx_get_context[ tx_request0, tx_request1, sram_txc_base, sigmask]

.begin

	.reg   	cam_result
	.reg	cam_entry
	.reg	cell_count_remained
	.reg	tmp

	// check if the q# in CAM, if found load the address into lm_addr0

	cam_lookup[cam_result, tx_request0], lm_addr0[0]

	// get the actual cam entry 

	alu_shf[cam_entry, CAM_ENTRY_MASK, AND, cam_result, >>CAM_ENTRY_LOC]

    // "hit" is not in the worst path case, so can take penalty for branch out

	br_bset[cam_result, 7, CAM_HIT#]

	// read the meta data from SRAM for the context (3)

	.io_completed	meta_read_signal

	dl_meta_load_cache[tx_request1, $$dl_meta, meta_read_signal, 1, 6]

	// allocate xfer registers for reading context

	xbuf_alloc[$$txc_in, 4, read]
	
	// allocate xfer registers for writing context

	xbuf_alloc[$txc_out, 4, write]

	// swap out context (7)

	_csix_tx_swap_out_context[ $txc_out, cam_entry, sram_txc_base, txc_write_signal]

	// swap in new context  (2)

	_csix_tx_swap_in_context[ $$txc_in , tx_request0, sram_txc_base, txc_read_signal]

	// set the bit in the bitmask for txc read signal 

	_csix_tx_set_signal[sigmask, _TXC_READ_SIGNAL]

	// set the bit in the bitmask for the txc write signal

	_csix_tx_set_signal[sigmask, _TXC_WRITE_SIGNAL]

	// wait for dram to tbuf for packet n and read of meta and context for packet 
	// n+1 to finish and signal from next thread (5)

	local_csr_wr[active_ctx_wakeup_events, sigmask]		
	ctx_arb[--] , defer[2]
		
		// Update *l$index0[7] with current buffer handle for later comparison in 
		// other threads coresponding to the same queue number to determine whether 
		// a new buffer arrives, if a new buffer arrives, then we need to read meta
		// data to update TXC in phase 2. 

   		 alu_shf[*l$index0[7], tx_request1, AND~, 0xff, <<24] 

		// update the CAM with state CAM_STATE_VALID, this CAM state is not used now, 
		// but may be used later

		cam_write[cam_entry, tx_request0, CAM_STATE_VALID]

	// At this point the context is read in/written out. let the assembler know
 
	.io_completed $$txc_in[0] $$txc_in[1] 	$$txc_in[2] 	$$txc_in[3]
	.io_completed $txc_out[0] $txc_out[1] 	$txc_out[2] 	$txc_out[3]

	// At this point we are done with the signals from phase 2, let the assembler 
	// know

	.io_completed  txc_read_signal txc_write_signal

	// copy in the transmit context 

	alu[*l$index0[0], --, B, $$txc_in[0]]
	alu[*l$index0[1], --, B, $$txc_in[1]]

	br[EXIT#] , defer[2]

		alu[*l$index0[2], --, B, $$txc_in[2]]
		alu[*l$index0[3], --, B, $$txc_in[3]]

	// done with the xbuf's for txc -- only an assembler directive - no instruction
	// is generated for the below

	xbuf_free[$$txc_in]
	xbuf_free[$txc_out]

CAM_HIT#:


 	// See if we already have the meta data 

	alu_shf[tmp, tx_request1, AND~, 0xff, <<24]
	alu[--, tmp, -, *l$index0[7]]

	// not equal - new buffer arrives, need to get meta data

	beq[cam_hit_noneed_get_meta_data#]
	
	// following label cam_hit_need_get_meta_data#: is not referenced, just put here
	// for readability

cam_hit_need_get_meta_data#:  
	
	// read the meta data from SRAM 

	.io_completed meta_read_signal

	dl_meta_load_cache[tx_request1, $$dl_meta, meta_read_signal, 1, 6]

	// wait for dram to tbuf for packet n and read of meta and context for packet 
	// n+1 to finish and signal from next thread (5)

	local_csr_wr[active_ctx_wakeup_events, sigmask]		
	ctx_arb[--] , br [EXIT#] , defer[1]
		
		// update *l$index0[7] with current buffer handle for later comparison in 
		// other threads coresponding to the same queue number to determine whether 
		// a new buffer arrives, if a new buffer arrives, then we need to read meta
		// data here to update TXC in phase 2  

   		 alu_shf[*l$index0[7], tx_request1, AND~, 0xff, <<24] 

cam_hit_noneed_get_meta_data#:

  	// clear the signal for reading meta data 

	_csix_tx_clear_signal[sigmask, _META_READ_SIGNAL]

	// wait for dram to tbuf for packet n and signal from next thread (5)

	local_csr_wr[active_ctx_wakeup_events, sigmask]		
	ctx_arb[--] 

EXIT#:

.end
#endm

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_swap_out_context[]
//
// Description:
// 	
//		swap out context into sram
//
// Outputs: 
//
//		
//
// Inputs:  
//
//		$txc_out			xfer registers to read out context
//		cam_entry			transmit request
//		sram_txc_base		base for sram contexts
//		signal_number		signal to wait on
//		
// Constants
//
//		None
//
// Size: 
//
//		7 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_swap_out_context[ $txc_out, cam_entry, sram_txc_base, signal_number]

.begin

	.reg 	lru_tag
	.reg	lru_offset

 	// Since CSIX TX just run on one ME, It will be more efficient to use CAM 
    // without reference count. The simplest way is to swap out LRU, when a 
	// miss happened. We already initialized CAM entries to first 16 queue 
	// number and corresponding 64 LM dword to 0, so that the first 16 swapout 
	// will not not generate invalid TXC in sram. 
     
    // move LRU txc from LM to TX registers

	alu[$txc_out[0], --, B, *l$index0[0]]
	alu[$txc_out[1], --, B, *l$index0[1]]
	alu[$txc_out[2], --, B, *l$index0[2]]
	alu[$txc_out[3], --, B, *l$index0[3]]

	// compute txc addr in SRAM of LRU TXC

	cam_read_tag[lru_tag, cam_entry]

	// lru_tag has valid bit in bit 31, this bit should disappear after 
	// <<SRAM_TXC_SHIFT (5),

	alu_shf[lru_offset, --, B, lru_tag, <<SRAM_TXC_SHIFT]

	// swap out LRU TXC. No need to swap out LRU BD, since the QM will 
	// maintain the updated BD
 
 	sram[write, $txc_out[0], sram_txc_base, lru_offset, 4], sig_done[signal_number]

.end

#endm


/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_swap_in_context[]
//
// Description:
// 	
//		Read transmit context from SRAM using q#
//
// Outputs: 
//
//		
//
// Inputs:  
//
//		$$txc_in			xfer registers to read in context
//		$tx_request			transmit request
//		sram_txc_base		base for sram contexts
//		signal_number		signal to wait on
//		
// Constants
//
//		None
//
// Size: 
//
//		2 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_swap_in_context[ $$txc_in, tx_request, sram_txc_base, signal_number]

.begin

	.reg	txc_offset

	// $tx_request[0] has valid bit in bit 31, this bit should disappear after 
	// <<SRAM_TXC_SHIFT (5), so we will get right new_txc_offset

	alu_shf[txc_offset, --, B, tx_request, <<SRAM_TXC_SHIFT]
	sram[read, $$txc_in[0], sram_txc_base, txc_offset, 4], sig_done[signal_number]

.end

#endm

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_move_sop_dram_to_tbuf[]
//
// Description:
// 	
//		For a sop cell (SOP bit only is set) this macro sets up the dram
//		to tbuf transfer
//
// Outputs: 
//
//		tcw					transmit control word 			
//
// Inputs:  
//
//		tcw_base			transmit control word base
//		buffer_handle		buffer handle
//		addr_of_tbuf		adress of tbuf
//		signal_number		signal to wait on 
//		
// Constants
//
//		None
//
// Size: 
//
//		22 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_move_sop_dram_to_tbuf [ tcw, tcw_base, buffer_handle, \
										addr_of_tbuf, signal_number ]

.begin

	.reg	align_offset			; offset into quad word at which data starts
	.reg	dram_addr				; dram address at which data starts
	.reg	tmp_addr_of_tbuf		; used for indirect ref calculation
	.reg	indir_ref				; used for indirect ref calculation
	.reg	payload_length			; payload length to be transmitted
	.reg	actual_payload_length	; actual payload length transmitted 
	.reg	buf_size				; buffer size for this buffer

	// set payload length to the default payload for SOP cell (112 bytes)

	immed[payload_length, FULL_TM_PAYLOAD_SIZE_FCIP]

	// Get the dram base address for the buffer

    dl_buf_get_data[dram_addr, buffer_handle]

	// Get offset in buffer of the data and store it into local memory 

	dl_meta_get_offset[*l$index0[TXC_OFFSET_INDEX]] 

	// get dram read address for buffer

	alu[dram_addr, dram_addr, +, *l$index0[TXC_OFFSET_INDEX]] 

	// Compute the transmit control word for later use 

	_csix_tx_get_transmit_control_word[tcw, tcw_base, 0, payload_length]

	// set buffer descriptor valid bit = 1, bytes_sent_already = 0 

	ld_field_w_clr[ *l$index0[TXC_BD_VALID_INDEX], 0100, VALID_BD_BYTE, \
					<<VALID_BD_BIT_LOC ]

	// compute offset into quadword for buffer offset

	alu[align_offset, *l$index0[TXC_OFFSET_INDEX], AND, 0x7]

	// store in local memory total bytes to be transferred from dram to tbuf for 
	// this buffer. This is the actual buffer size aligned to nearest quadword 

	dl_meta_get_buffer_size[buf_size]
	alu[buf_size, buf_size, +, align_offset]

	alu[*l$index0[TXC_DRAM_TO_TBUF_SIZE_INDEX], \
		*l$index0[TXC_DRAM_TO_TBUF_SIZE_INDEX], OR, buf_size, <<16]		
										
	// set align offset, set seq#  to zero, only in first buffer of each packet
	// has this data 

    alu[*l$index0[TXC_SEQ_NUMBER_INDEX], --, b, align_offset, <<24]
	
	// get next hop ip id and class_id	
.begin 
.reg v_nexthop_id v_classid
    dl_meta_get_nexthop_id[v_nexthop_id]
    dl_meta_get_class_id[v_classid]		
	alu_shf[*l$index0[TXC_NEXTHOP_CLASS_ID_INDEX], v_classid, or, v_nexthop_id, <<16]	
.end //v_nexthop_id v_classid

    // Update bytes already sent field in local memory for next cframe use
	// in the same buffer by the fully loaded payload size for current cell (112
	// for first cell in a packet, and 120 for others),

	alu[*l$index0[TXC_BYTES_SENT_INDEX], *l$index0[TXC_BYTES_SENT_INDEX], +, \
		 payload_length]
    
	// prepare adrress of tbuf elment for cell data as indirect token for
    // following dram[tbuf_write....] intsruction

	alu[tmp_addr_of_tbuf, addr_of_tbuf, +, 16]
	alu[indir_ref, indiref_base, OR, tmp_addr_of_tbuf, <<5]

    // since ref count 0 will fetch 8 bytes, and ref count 1 will fetch
	// 16 bytes......, and ref count 15 will fetch 128 bytes, 
	// the right way to do it is followinhg:
	//    a) decrease length by one
	//    b) then right shift the result by 3
	// We will do step a) first, then combine the right shift in the shift into
	// the ref count in indir_ref.  In this way, we make sure that we fetch 
	// enough and closet to 8 byte alignment of packet data	
	
	alu[payload_length, payload_length, -, 0x01]
  	alu_shf[payload_length, --, B, payload_length, >>3]

    // left shift 21 bit to ref_cnt location

    alu[--, indir_ref, OR, payload_length, <<21] 
			 
    // how many bytes of packet payload are copied to tbuf from dram is 
	// controlled by ref cnt field in indir_ref, following max_15 is
	// dummy	 

    dram[tbuf_wr, --, dram_addr, 0, max_15], indirect_ref, sig_done[signal_number]

.end

#endm 


/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_move_mop_eop_dram_to_tbuf[]
//
// Description:
// 	
//		For a mop or eop only cell handle the dram to tbuf xfer
//
// Outputs: 
//
//		cell_count_remained		cell count remaining	
//		tcw						transmit control word
//
// Inputs:  
//
//		tcw_base				transmit control word base
//		buffer_handle			buffer handle
//		addr_of_tbuf			adress of tbuf
//		signal_number			signal to wait on 
//		
// Constants
//
//		None
//
// Size: 
//
//		27 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_move_mop_eop_dram_to_tbuf [ cell_count_remained, tcw, tcw_base, \
										    buffer_handle, addr_of_tbuf, \
										    signal_number ]
										

.begin

	.reg	total_payload_size		; total payload size 
	.reg	bytes_sent_already		; bytes already sent in the buffer
	.reg	dram_addr				; dram address at which data starts
	.reg	tmp_addr_of_tbuf		; used for indirect ref calculation
	.reg	indir_ref				; used for indirect ref calculation
	.reg	payload_length			; payload length to be transmitted
	.reg	buffer_offset			; offset into buffer
	.reg	buf_size				; size of the buffer

	// check if the buffer descriptor is valid by checking the valid bit

	br_bset[*l$index0[TXC_BD_VALID_INDEX], VALID_BD_BIT_LOC, bd_valid#] , defer [2]

		// set payload length to length for a non sop cell (120) 

		immed[payload_length, FULL_TM_PAYLOAD_SIZE_NOT_FCIP]

		// get remaining cell count from $tx_request[1] for later use

  	    alu_shf[ cell_count_remained, CELL_COUNT_REMAINED_MASK, AND, buffer_handle,\
			 	 >>CELL_COUNT_LOC]

	// BD bit not valid, previous buffer done, so this is a new buffer and this is
	// not the first cell of a packet

	// set the bd valid bit = 1, bytes already sent = 0,  since this is a new 
	// buffer, so need to clear bytes_sent_already.

	alu_shf[*l$index0[TXC_BD_VALID_INDEX], --, B, 1, <<VALID_BD_BIT_LOC]

	// copy in the buffer size and offset. offset is always 0 or a multiple of 8

	// Get offset in buffer of the data and store it into local memory 

	dl_meta_get_offset[*l$index0[TXC_OFFSET_INDEX]] 
	
	// Get the buffer size and store that also in local memory

	dl_meta_get_buffer_size[buf_size]

	alu[*l$index0[TXC_DRAM_TO_TBUF_SIZE_INDEX], \
		*l$index0[TXC_DRAM_TO_TBUF_SIZE_INDEX], OR, buf_size, <<16]		
										

bd_valid#:

	// get bytes already sent in this buffer, will be used to calculate dram addr
	// to fetch payload data and payload length in the last cell of a buffer

	ld_field_w_clr[bytes_sent_already, 0011, *l$index0]

    // update bytes already sent field in local memory for next cframe use
	// in the same buffer by the fully loaded payload size for current cell (112
	// for first cell in a packet, and 120 for others, i.e., tm_payload_len),
	// it is not necessary to check whether this is the last cell in one buffer,
	// if it is the case, this field will be reset to zero later

	alu[*l$index0[TXC_BYTES_SENT_INDEX], *l$index0[TXC_BYTES_SENT_INDEX], +, \
					payload_length]

checkeob#:  

	// check if this is the last cell in this buffer

	br!=byte[cell_count_remained, 0, 0x00, finish#]

	// end of buffer, set TXC[7] to 0xFF, so the code will prevent one new buffer
	// in later runs to use stale meta data in local memory, in the case that the 
	// transmit context for this queue stays in local memory and the new buffer 
	// uses the same dl_buffer_handle as the previous buffer for this queue (the 
	// chance of this condition to happen should be very low)

	alu[*l$index0[TXC_BUFFER_HANDLE_INDEX], --, B, IMPOSSIBLE_DL_HDL_ADDR]

	// check if this is the the last cell in this packet (if this is the last
	// cell in this buffer and EOP flag (bit 31)set in tx request second long word
	// also perform things general to last cell of one buffer in the defer slots
	// of following br_bclr to save 3 instruction cycles

	br_bclr[buffer_handle, EOP_BIT_LOC, finish#], defer[3]

		// it is end of current buffer, set the bd valid bit = 0 and bytes already
		// sent in this buffer to 0

		alu[*l$index0[TXC_BD_VALID_INDEX], --, B, 0]
  
		// this is the last cell in current buffer, it's payload may not be exact
		// FULL_TM_PAYLOAD_SIZE, so we need to calculate here from *l$index[3]
		// and bytes_sent_already
  
		ld_field_w_clr[total_payload_size, 0011, *l$index0[TXC_OFFSET_INDEX], >>16]
		alu[payload_length, total_payload_size, -, bytes_sent_already]   

	// set the EOP flag here

	alu[*l$index0[TXC_SEQ_NUMBER_INDEX], *l$index0[TXC_SEQ_NUMBER_INDEX], OR, 0x80]

finish#:

 	// Get the dram base address for the buffer

    dl_buf_get_data[dram_addr, buffer_handle]

	// get dram read address for buffer by adding offset. We still need to add 
	// bytes_sent_already which we will do when do the dram[] operation

	ld_field_w_clr[buffer_offset, 0011, *l$index0[TXC_OFFSET_INDEX]]
	alu[dram_addr, dram_addr, +, buffer_offset] 

	// Compute the transmit control word for later use 

	_csix_tx_get_transmit_control_word[tcw, tcw_base, 0, payload_length]
											  
	// prepare adrress of tbuf elment for cell data as indirect token for
    // following dram[tbuf_write....] intsruction

	alu[tmp_addr_of_tbuf, addr_of_tbuf, +, 8]
	alu[indir_ref, indiref_base, OR, tmp_addr_of_tbuf, <<5]

    // since ref count 0 will fetch 8 bytes, and ref count 1 will fetch
	// 16 bytes......, and ref count 15 will fetch 128 bytes, 
	// the right way to do it is followinhg:
	//    a) decrease length by one
	//    b) then right shift the result by 3
	// We will do step a) first, then combine the right shift in the shift into
	// the ref count in indir_ref.  In this way, we make sure that we fetch 
	// enough and closet to 8 byte alignment of packet data	

	alu[payload_length, payload_length, -, 0x01]
  	alu_shf[payload_length, --, B, payload_length, >>3]

    // left shift 21 bit to ref_cnt location

    alu[--, indir_ref, OR, payload_length, <<21] 
			 
    // how many bytes of packet payload are copied to tbuf from dram is 
	// controlled by ref cnt field in indir_ref, following max_14 is
	// dummy	 

    dram[tbuf_wr, --, dram_addr, bytes_sent_already, max_15], indirect_ref, \
															sig_done[signal_number]

.end

#endm


#ifdef COUNTERS

///////////////////////////////////////////////////////////////////////////////
// _csix_tx_update_counters()
//
// Description: update CSIX TX counters in sram
//
// Outputs:							
//		out_bytecnt			sram xfer register to keep the value to add to 
// 							counters in sram
//
// Inputs:  
// 		in_tcw:				transmit control word which has payload length 
//                          and prepend length
//		in_ex_header		cframe extension header which has class_id
//		in_buffer_handle	buffer handle which has EOP bit and cell count
//							remained
//      in_tx_request0		tx_request0 which has voq in lower 10 bits
//		in_sram_add_sig:	signal for sram add operation
//
// Constants:
//							None 
//
// Size: 
//		15 instructions. (Worst case cycle count)
//
///////////////////////////////////////////////////////////////////////////////
#macro _csix_tx_update_counters(out_bytecnt, in_tcw, in_ex_header, \
						in_buffer_handle, in_counter_base, in_sram_add_sig)
.begin
.reg voq bytes_txed paylo cntr_addr tmpxx
	alu[tmpxx, DESTINATION_FIELD_MASK, AND, in_ex_header]	; destination blade
	// get class_id from in_ex_header
	alu_shf[voq, CLASS_FIELD_MASK, AND, in_ex_header, >>CLASS_FIELD_SHIFT]
	alu_shf[voq, voq, OR, tmpxx, <<DEST_FIELD_SHIFT]
	alu_shf[cntr_addr, in_counter_base, OR, voq, <<4]	
	sram[incr, --, cntr_addr, CFRAME_TRANSMITTED_CNT_IDX]	; increment cframe
															; transmitted
	alu_shf[bytes_txed, --, b, in_tcw, >>TCWO_PREPEND_LEN_LSB_LOC] ; prepend

	alu_shf[tmpxx, EOP_CELL_COUNT_MASK, AND, in_buffer_handle, >>CELL_COUNT_LOC]
    br!=byte[tmpxx, 0, EOP_EOP_CELL_COUNT, update_bytes_txed#], defer[3]
		alu[bytes_txed, bytes_txed, and, TCW0_PREPEND_MASK]	; mask out
		alu_shf[paylo, --, b, in_tcw, >>TCW0_PAYLOAD_LEN_LSB_LOC] ; payload
		alu[out_bytecnt, bytes_txed, +, paylo]

	sram[incr, --, cntr_addr, PKT_TRANSMITTED_CNT_IDX]	; increment pkts 
														; transmitted
update_bytes_txed#:
	sram[add, out_bytecnt, cntr_addr, BYTE_TRANSMITTED_CNT_IDX],\
	     sig_done[in_sram_add_sig]			 	
.end //		 
#endm // end of macro _csix_tx_update_counters()
	
#endif // COUNTERS


/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_debug_init_counters[]
//
// Description:
// 	
//		Initialize the debug counters in absolute registers to zeros.
//
// Outputs: 
//
//		 			
//
// Inputs:  
//
//		
//		
// Constants
//
//		None
//
// Size: 
//
//		3 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////
#macro	_csix_tx_debug_init_counters[]

#ifdef	_DEBUG_COUNTERS_

	immed[@csix_tx_num_cells_txed,0x0]	
	immed[@csix_tx_num_buf_freed,0x0]	
	immed[@csix_tx_num_tx_req,0x0]	

#endif
#endm

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_debug_incr_counter[]
//
// Description:
// 	
//		Increment the specified counter by 1. This macro is enabled when the build
//	switch _DEBUG_COUNTERS_ is enabled.
//
// Outputs: 
//
//		 			
//
// Inputs:  
//		counter		the debug counter to increment
//		
//		
// Constants
//
//		None
//
// Size: 
//
//		1 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////
#macro	_csix_tx_debug_incr_counter[counter]

#ifdef	_DEBUG_COUNTERS_

	alu[counter,counter,+,1]

#endif
#endm

/////////////////////////////////////////////////////////////////////////////////////

#endif 		// __CSIX_TX_UTIL_UC__

/////////////////////////////////////////////////////////////////////////////////////

