/////////////////////////////////////////////////////////////////////////////////////
//
//
//                  I N T E L   P R O P R I E T A R Y
//
//     COPYRIGHT [c]  2002 BY  INTEL  CORPORATION.  ALL RIGHTS
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT
//     THE PRIOR WRITTEN PERMISSION OF :
//
//                        INTEL  CORPORATION
//
//                     2200 MISSION COLLEGE BLVD
//
//               SANTA  CLARA,  CALIFORNIA  95052-8119
//
/////////////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: csix_tx_util.uc
//
//      Purpose: Utility macros for csix transmit 
//
//
//      History:
//
//      Date            Comment                         				By
//      ---------------------------------------------------------------------------
//
//      11/21/2001      Created                         			David Chou
//
//		3/18/02			Modified to do packet interleaving			Uday Naik	
//						and adhere to coding convention
//
/////////////////////////////////////////////////////////////////////////////////////

#ifndef __CSIX_TX_UTIL1_UC__
#define __CSIX_TX_UTIL1_UC__

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_read_transmit_request[]
//
// Description:
// 	
//		Read the transmit request from a scratch ring
//
// Outputs: 
//
//		tx_request:		2 xfer registers that will contain the tx request
//							when the I/O completes
//
// Inputs:  
//
//		ring:				scratch ring number to read from
//		signal_number:		signal number to wait on
//
// Constants
//
//		None
//
// Size: 
//
//		1 instruction
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_read_transmit_request[tx_request, ring, signal_number]

.begin

	// get tx request - 2 words from scratch ring 

	scratch[get, tx_request0, ring, 0, 2], sig_done[signal_number]

.end

#endm



#macro _csix_tx_read_transmit_request_nn[tx_request, no_valid_tx_request_label]

.begin	

	br_inp_state[nn_empty, no_valid_tx_request_label]

	alu[tx_request/**/0, --, b, *n$index++]
	alu[tx_request/**/1, --, b, *n$index++]

.end

#endm


/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_signal_next[]
//
// Description:
// 	
//		signal the next thread
//
// Outputs: 		
//
// Inputs:  
//
// Constants
//
//		SIG_NUM		signal number
//
// Size: 
//
//		2 instruction
// 
//
/////////////////////////////////////////////////////////////////////////////////////


#macro _csix_tx_signal_next[next_thread_signal_gpr]

.begin

	.reg temp

	// signal number goes into bits 3..6 and bit 7 is set to signal the next context.
	// We store 0x80 in local memory to optimize

	local_csr_wr[SAME_ME_SIGNAL, next_thread_signal_gpr]		

.end

#endm

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_clear_signal[]
//
// Description:
// 	
//		Clear a signal from a signal mask 
//
// Outputs: 
//
//
// Inputs:  
//
//		sigmask:			signal mask
//		sig:				signal number to clear
//
// Constants
//
//		None
//
// Size: 
//
//		1-3 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_clear_signal[sigmask, sig]

.begin

	.reg tmp

#if ( isnum(sig) )
	
	#define_eval SIG_SHIFT sig
	alu_shf[sigmask, sigmask, AND~, 1, <<SIG_SHIFT ]
	#undef SIG_SHIFT

#else

	alu[tmp, &sig, -, 1]
	alu[--, tmp, OR, 0]										
	alu_shf[sigmask, sigmask, AND~, 1, <<indirect]

#endif

.end

#endm

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_set_signal[]
//
// Description:
// 	
//		Set a signal in a signal mask 
//
// Outputs: 
//
//
// Inputs:  
//
//		sigmask:			signal mask
//		sig:				signal number to set
//
// Constants
//
//		None
//
// Size: 
//
//		1-3 instructions
// 
//
////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_set_signal[sigmask, sig]

.begin

	.reg tmp

#if ( isnum(sig) )

	#define_eval SIG_SHIFT sig
	alu_shf[sigmask, sigmask, OR, 1, <<SIG_SHIFT]
	#undef SIG_SHIFT

#else

	alu[tmp, &sig, -, 1]
	alu[--, tmp, OR, 0]										
	alu_shf[sigmask, sigmask, OR, 1, <<indirect]


#endif

.end

#endm


/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_prepare_to_move_sop_eop_dram_to_tbuf[]
//
// Description:
// 	
//		For a single cell packet (SOP/EOP both set) this macro sets up the dram
//		to tbuf transfer
//
// Outputs: 
//
//		dram_addr			address to read packet data from dram
//		payload_length		payload length to read
// 		align_offset		offset into quadword for start of data
//
// Inputs:  
//
//		buffer_handle		buffer handle
//		
// Constants
//
//		None
//
// Size: 
//
//		7 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_prepare_to_move_sop_eop_dram_to_tbuf[  dram_addr, payload_length, align_offset, \
											buffer_handle ]

.begin
	
	.reg	buffer_offset		; offset into buffer at which payload starts
	.reg	tmp_addr_of_tbuf	; used for indirect ref calculation
	.reg	indir_ref			; used for indirect ref calculation

  	// Get the dram base address for the buffer

    dl_buf_get_data[dram_addr, buffer_handle]

	// Get offset in buffer of the data 

	dl_meta_get_offset[buffer_offset]

	// get dram read address for buffer

	alu[dram_addr, dram_addr, +, buffer_offset] 

	// copy buffer size from meta data into payload length 

	dl_meta_get_buffer_size[payload_length]

	// compute offset into quadword for buffer offset

	alu[align_offset, buffer_offset, AND, 0x7]

	// increment the payload length by offset to round off to quad word boundary

	alu[payload_length, payload_length, + , align_offset]


.end

#endm

/////////////////////////////////////////////////////////////////////////////////////
//	
//	_csix_tx_prepare_to_write_tm_header_to_tbuf[]
//
// Description:
// 	
//		Write the TM header to a tbuf
//
// Outputs: 
//		prepend			packet prepend header registers (prepend0 - prepend3)
//		
//
// Inputs:  
//
// 		align_offset		offset into quadword for start of packet data
//		signal_number		signal number to wait on
//		
// Constants
//
//		SOP_EOP_FLAG		indicates cell type
//
// Size: 
//
//		9 instructions for SOP/EOP and SOP case and 5 instructions for MOP case
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_prepare_to_write_tm_header_to_tbuf[ prepend, align_offset,  SOP_EOP_FLAG]

.begin
 
	.reg	prepend_header0
	.reg	tmp
	.reg	increment
	.reg	header_type
	.reg	output_port
	.reg	flow_id
	.reg	color
	.reg	nexthop_id
	.reg	class_id

	// ingress blade number should be imported from xscale core, since there
	// is no xscale now, we use THIS_BLADE_ID defined in system.h in 
	// dispatch_loop 

#if (SOP_EOP_FLAG != MOP_EOP)

	// This is the code for sop/eop or sop only cells. In this case we 
	// need a 2 word packet prepend and a 2 word frame prepend

#if (SOP_EOP_FLAG == SOP_EOP)

	// initialize prepend header word 0 with alignment offset and blade id 

	alu_shf[prepend_header0, THIS_BLADE_ID, OR, align_offset, <<24]

	// set the EOP, SOP flag here and copy into write transfer registers

	alu[prepend/**/0, prepend_header0, OR, 0xC0]

#else

	// initialize prepend header word 0 with alignment offset and blade id 

	alu[prepend_header0, THIS_BLADE_ID, OR, *l$index0[TXC_SEQ_NUMBER_INDEX]]

	// set the SOP flag only 

	alu[prepend/**/0, prepend_header0, OR, 0x40]

#endif

	// copy next hop id in from meta data 

    dl_meta_get_nexthop_id[nexthop_id]

	// copy class id from meta data 

    dl_meta_get_class_id[class_id]	
	
	// copy these into the prepend header 
		
	alu_shf[prepend/**/1, class_id, or, nexthop_id, <<16]	

	// get header type from the meta data 

	dl_meta_get_header_type[header_type]

	// get output port from the meta data

	dl_meta_get_output_port[output_port]

	// copy output port and header type into prepend3

	alu[prepend/**/2, output_port, OR, header_type, <<16]

	// get the flow id from the meta data

	dl_meta_get_flow_id[flow_id]

	// get the color id from the meta data 

	dl_meta_get_color[color]

	// copy flow id and color into prepend3

	alu[prepend/**/3, flow_id, OR, color, <<24]

    // write TM header to tbuf using msf write 

	//msf[write, prepend/**/0, addr_of_tbuf, 0, 4], sig_done[signal_number]

#else

	// This is the case for MOP cells and EOP cells. We only need a 2 word
	// frame prepend in these cases (5 instructi

	// increment the seq number by 256 since we are storing in bit 8..24

	alu[increment, --, B, 1, <<8]

   	alu[*l$index0[TXC_SEQ_NUMBER_INDEX], *l$index0[TXC_SEQ_NUMBER_INDEX], \
		+, increment]
		
	// copy the port and sequence number into prepend0 from local memory. If this
	// is an EOP cell, that will also be set in the local memory field 

	ld_field_w_clr[tmp, 0111, *l$index0[TXC_SEQ_NUMBER_INDEX]]	
	alu[prepend/**/0, THIS_BLADE_ID, OR, tmp]	

	// copy the prepend into transfer registers
	
    alu[prepend/**/1,	--, B, *l$index0[TXC_NEXTHOP_CLASS_ID_INDEX]]		


#endif

.end

#endm



/////////////////////////////////////////////////////////////////////////////////////
//
// _csix_tx_init_gprs
//
//	Description: initialize context relative GPR for specific thread
//
//	Size:
//
//		5 instructions
//
/////////////////////////////////////////////////////////////////////////////////////
 
#macro _csix_tx_init_gprs1[ v_tcw0_fcip_base, v_tcw0_nfcip,\
							 s_txc_base, v_indiref_base, sigmask_phase2_default ]

.begin

	// base for the first transmit control word for the first cell of one packet 

	move[v_tcw0_fcip_base, TCW0_FCIP_BASE_LW]

	// base for the first transmit control word for cells which is not the first 
	// cell of one packet 

 	move[v_tcw0_nfcip, TCW0_NOT_FCIP_BASE_LW]

	// following s_txc_base needed to be context relative
    // Base address of CSIX Transmit Contexts in SRAM (CSIX_TX_CTX_SRAM_BASE)
	// is defined in the common header file "system.h" in ..\dispatch_loop

	immed32[s_txc_base, CSIX_TX_CTX_SRAM_BASE]
 
	// base to calculate the address field in indirect ref of 
	// dram[tbuf_wr....] instruction

	move[v_indiref_base, DRAM_INDIREF_BASE_LW]

	// default phase 2 signal mask at start of each iteration of loop 

	move[sigmask_phase2_default, SIG_MASK_PHASE2]

.end

#endm 

////////////////////////////////////////////////////////////////////////////////////
//
// _csix_tx_get_context[]
//
// Description:
// 	
//		swap out context into sram
//
// Outputs: 
//
//		
//
// Inputs:  
//
//		tx_request			transmit request
//		sram_txc_base		base for sram contexts
//		sigmask				signal mask
//		
// Constants
//
//		None
//
// Size: 
//
//		26 instructions
//	
//
/////////////////////////////////////////////////////////////////////////////////////
 
#macro _csix_tx_get_context[ tx_request, sram_txc_base, sigmask]

.begin

	.reg   	cam_result
	.reg	cam_entry
	.reg	cell_count_remained
	.reg	tmp

	// check if the q# in CAM, if found load the address into lm_addr0

	cam_lookup[cam_result, tx_request0], lm_addr0[0]

	// get the actual cam entry 

	alu_shf[cam_entry, CAM_ENTRY_MASK, AND, cam_result, >>CAM_ENTRY_LOC]

    // "hit" is not in the worst path case, so can take penalty for branch out

	br_bset[cam_result, 7, CAM_HIT#]

	// read the meta data from SRAM for the context (3)

	.io_completed	meta_read_signal

	dl_meta_load_cache[tx_request1, $$dl_meta, meta_read_signal, 1, 6]

	// allocate xfer registers for reading context

	xbuf_alloc[$$txc_in, 4, read]
	
	// allocate xfer registers for writing context

	xbuf_alloc[$txc_out, 4, write]

	// swap out context (7)

	_csix_tx_swap_out_context[ $txc_out, cam_entry, sram_txc_base, txc_write_signal]

	// swap in new context  (2)

	_csix_tx_swap_in_context[ $$txc_in , tx_request, sram_txc_base, txc_read_signal]

	// wait for read of meta and context swap in/out and next thread signals

	ctx_arb[next_thread_signal_1, txc_read_signal, txc_write_signal, \
			meta_read_signal] , defer[2]
		
		// Update *l$index0[7] with current buffer handle for later comparison in 
		// other threads coresponding to the same queue number to determine whether 
		// a new buffer arrives, if a new buffer arrives, then we need to read meta
		// data to update TXC in phase 2. 

   		 alu_shf[*l$index0[7], tx_request1, AND~, 0xff, <<24] 

		// update the CAM with state CAM_STATE_VALID, this CAM state is not used now, 
		// but may be used later

		cam_write[cam_entry, tx_request0, CAM_STATE_VALID]

	// At this point the context is read in/written out. let the assembler know
 
	.io_completed $$txc_in[0] $$txc_in[1] 	$$txc_in[2] 	$$txc_in[3]
	.io_completed $txc_out[0] $txc_out[1] 	$txc_out[2] 	$txc_out[3]

	// At this point we are done with the signals from phase 2, let the assembler 
	// know

	.io_completed  txc_read_signal txc_write_signal

	// copy in the transmit context 

	alu[*l$index0[0], --, B, $$txc_in[0]]
	alu[*l$index0[1], --, B, $$txc_in[1]]

	br[EXIT#] , defer[2]

		alu[*l$index0[2], --, B, $$txc_in[2]]
		alu[*l$index0[3], --, B, $$txc_in[3]]

	// done with the xbuf's for txc -- only an assembler directive - no instruction
	// is generated for the below

	xbuf_free[$$txc_in]
	xbuf_free[$txc_out]

CAM_HIT#:


 	// See if we already have the meta data 

	alu_shf[tmp, tx_request1, AND~, 0xff, <<24]
	alu[--, tmp, -, *l$index0[7]]

	// not equal - new buffer arrives, need to get meta data

	beq[cam_hit_noneed_get_meta_data#]
	
	// following label cam_hit_need_get_meta_data#: is not referenced, just put here
	// for readability

cam_hit_need_get_meta_data#:  
	
	// read the meta data from SRAM 

	.io_completed meta_read_signal

	dl_meta_load_cache[tx_request1, $$dl_meta, meta_read_signal, 1, 6]

	// wait for meta read to complete 

	ctx_arb[meta_read_signal, next_thread_signal_1], br[EXIT#], defer[1]
		
		// update *l$index0[7] with current buffer handle for later comparison in 
		// other threads coresponding to the same queue number to determine whether 
		// a new buffer arrives, if a new buffer arrives, then we need to read meta
		// data here to update TXC in phase 2  

   		 alu_shf[*l$index0[7], tx_request1, AND~, 0xff, <<24] 


cam_hit_noneed_get_meta_data#:

	// wait for signal from next thread
	
	ctx_arb[next_thread_signal_1] 


EXIT#:

.end
#endm

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_swap_out_context[]
//
// Description:
// 	
//		swap out context into sram
//
// Outputs: 
//
//		
//
// Inputs:  
//
//		$txc_out			xfer registers to read out context
//		cam_entry			transmit request
//		sram_txc_base		base for sram contexts
//		signal_number		signal to wait on
//		
// Constants
//
//		None
//
// Size: 
//
//		7 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_swap_out_context[ $txc_out, cam_entry, sram_txc_base, signal_number]

.begin

	.reg 	lru_tag
	.reg	lru_offset

 	// Since CSIX TX just run on one ME, It will be more efficient to use CAM 
    // without reference count. The simplest way is to swap out LRU, when a 
	// miss happened. We already initialized CAM entries to first 16 queue 
	// number and corresponding 64 LM dword to 0, so that the first 16 swapout 
	// will not not generate invalid TXC in sram. 
     
    // move LRU txc from LM to TX registers

	alu[$txc_out[0], --, B, *l$index0[0]]
	alu[$txc_out[1], --, B, *l$index0[1]]
	alu[$txc_out[2], --, B, *l$index0[2]]
	alu[$txc_out[3], --, B, *l$index0[3]]

	// compute txc addr in SRAM of LRU TXC

	cam_read_tag[lru_tag, cam_entry]

	// lru_tag has valid bit in bit 31, this bit should disappear after 
	// <<SRAM_TXC_SHIFT (5),

	alu_shf[lru_offset, --, B, lru_tag, <<SRAM_TXC_SHIFT]

	// swap out LRU TXC. No need to swap out LRU BD, since the QM will 
	// maintain the updated BD
 
 	sram[write, $txc_out[0], sram_txc_base, lru_offset, 4], sig_done[signal_number]

.end

#endm


/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_swap_in_context[]
//
// Description:
// 	
//		Read transmit context from SRAM using q#
//
// Outputs: 
//
//		
//
// Inputs:  
//
//		$$txc_in			xfer registers to read in context
//		tx_request			transmit request
//		sram_txc_base		base for sram contexts
//		signal_number		signal to wait on
//		
// Constants
//
//		None
//
// Size: 
//
//		2 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_swap_in_context[ $$txc_in, tx_request, sram_txc_base, signal_number]

.begin

	.reg	txc_offset

	// tx_request0 has valid bit in bit 31, this bit should disappear after 
	// <<SRAM_TXC_SHIFT (5), so we will get right new_txc_offset

	alu_shf[txc_offset, --, B, tx_request0, <<SRAM_TXC_SHIFT]
	sram[read, $$txc_in[0], sram_txc_base, txc_offset, 4], sig_done[signal_number]

.end

#endm

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_prepare_to_move_sop_dram_to_tbuf[]
//
// Description:
// 	
//		For a sop cell (SOP bit only is set) this macro sets up the dram
//		to tbuf transfer
//
// Outputs: 
//
//		dram_addr			address to read packet data from dram
//		payload_length		payload length to read			
//
// Inputs:  
//
//		buffer_handle		buffer handle
//		
// Constants
//
//		None
//
// Size: 
//
//		14 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_prepare_to_move_sop_dram_to_tbuf [ dram_addr, payload_length, \
											buffer_handle ] 

.begin

	.reg	align_offset			; offset into quad word at which data starts
	//.reg	dram_addr				; dram address at which data starts
	.reg	tmp_addr_of_tbuf		; used for indirect ref calculation
	.reg	indir_ref				; used for indirect ref calculation
	//.reg	payload_length			; payload length to be transmitted
	.reg	actual_payload_length	; actual payload length transmitted 
	.reg	buf_size				; buffer size for this buffer

	// set payload length to the default payload for SOP cell (112 bytes)

	immed[payload_length, FULL_TM_PAYLOAD_SIZE_FCIP]

	// Get the dram base address for the buffer

    dl_buf_get_data[dram_addr, buffer_handle]

	// Get offset in buffer of the data and store it into local memory 

	dl_meta_get_offset[*l$index0[TXC_OFFSET_INDEX]] 

	// get dram read address for buffer

	alu[dram_addr, dram_addr, +, *l$index0[TXC_OFFSET_INDEX]] 

	// Compute the transmit control word for later use 

	//_csix_tx_get_transmit_control_word[tcw, tcw_base, 0, payload_length]

	// set buffer descriptor valid bit = 1, bytes_sent_already = 0 

	ld_field_w_clr[ *l$index0[TXC_BD_VALID_INDEX], 0100, VALID_BD_BYTE, \
					<<VALID_BD_BIT_LOC ]

	// compute offset into quadword for buffer offset

	alu[align_offset, *l$index0[TXC_OFFSET_INDEX], AND, 0x7]

	// store in local memory total bytes to be transferred from dram to tbuf for 
	// this buffer. This is the actual buffer size aligned to nearest quadword 

	dl_meta_get_buffer_size[buf_size]
	alu[buf_size, buf_size, +, align_offset]

	alu[*l$index0[TXC_DRAM_TO_TBUF_SIZE_INDEX], \
		*l$index0[TXC_DRAM_TO_TBUF_SIZE_INDEX], OR, buf_size, <<16]		
										
	// set align offset, set seq#  to zero, only in first buffer of each packet
	// has this data 

    alu[*l$index0[TXC_SEQ_NUMBER_INDEX], --, b, align_offset, <<24]
	
	// get next hop ip id and class_id	
.begin 
.reg v_nexthop_id v_classid
    dl_meta_get_nexthop_id[v_nexthop_id]
    dl_meta_get_class_id[v_classid]		
	alu_shf[*l$index0[TXC_NEXTHOP_CLASS_ID_INDEX], v_classid, or, v_nexthop_id, <<16]	
.end //v_nexthop_id v_classid

    // Update bytes already sent field in local memory for next cframe use
	// in the same buffer by the fully loaded payload size for current cell (112
	// for first cell in a packet, and 120 for others),

	alu[*l$index0[TXC_BYTES_SENT_INDEX], *l$index0[TXC_BYTES_SENT_INDEX], +, \
		 payload_length]
    
.end

#endm 


/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_prepare_to_move_mop_eop_dram_to_tbuf[]
//
// Description:
// 	
//		For a mop or eop only cell handle the dram to tbuf xfer
//
// Outputs: 
//
//		cell_count_remained		cell count remaining	
//		dram_addr			address to read packet data from dram
//		payload_length		payload length to read
//
// Inputs:  
//
//		buffer_handle			buffer handle
//		
// Constants
//
//		None
//
// Size: 
//
//		20 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_prepare_to_move_mop_eop_dram_to_tbuf [ cell_count_remained, dram_addr, payload_length,  \
											buffer_handle ]
										

.begin

	.reg	total_payload_size		; total payload size 
	.reg	bytes_sent_already		; bytes already sent in the buffer
	.reg	tmp_addr_of_tbuf		; used for indirect ref calculation
	.reg	indir_ref				; used for indirect ref calculation
	.reg	buffer_offset			; offset into buffer
	.reg	buf_size				; size of the buffer

	// check if the buffer descriptor is valid by checking the valid bit

	br_bset[*l$index0[TXC_BD_VALID_INDEX], VALID_BD_BIT_LOC, bd_valid#] , defer [2]

		// set payload length to length for a non sop cell (120) 

		immed[payload_length, FULL_TM_PAYLOAD_SIZE_NOT_FCIP]

		// get remaining cell count from tx_request1 for later use

  	    alu_shf[ cell_count_remained, CELL_COUNT_REMAINED_MASK, AND, buffer_handle,\
			 	 >>CELL_COUNT_LOC]

	// BD bit not valid, previous buffer done, so this is a new buffer and this is
	// not the first cell of a packet

	// set the bd valid bit = 1, bytes already sent = 0,  since this is a new 
	// buffer, so need to clear bytes_sent_already.

	alu_shf[*l$index0[TXC_BD_VALID_INDEX], --, B, 1, <<VALID_BD_BIT_LOC]

	// copy in the buffer size and offset. offset is always 0 or a multiple of 8

	// Get offset in buffer of the data and store it into local memory 

	dl_meta_get_offset[*l$index0[TXC_OFFSET_INDEX]] 
	
	// Get the buffer size and store that also in local memory

	dl_meta_get_buffer_size[buf_size]

	alu[*l$index0[TXC_DRAM_TO_TBUF_SIZE_INDEX], \
		*l$index0[TXC_DRAM_TO_TBUF_SIZE_INDEX], OR, buf_size, <<16]		
										

bd_valid#:

	// get bytes already sent in this buffer, will be used to calculate dram addr
	// to fetch payload data and payload length in the last cell of a buffer

	ld_field_w_clr[bytes_sent_already, 0011, *l$index0]

    // update bytes already sent field in local memory for next cframe use
	// in the same buffer by the fully loaded payload size for current cell (112
	// for first cell in a packet, and 120 for others, i.e., tm_payload_len),
	// it is not necessary to check whether this is the last cell in one buffer,
	// if it is the case, this field will be reset to zero later

	alu[*l$index0[TXC_BYTES_SENT_INDEX], *l$index0[TXC_BYTES_SENT_INDEX], +, \
					payload_length]

checkeob#:  

	// check if this is the last cell in this buffer

	br!=byte[cell_count_remained, 0, 0x00, finish#]

	// end of buffer, set TXC[7] to 0xFF, so the code will prevent one new buffer
	// in later runs to use stale meta data in local memory, in the case that the 
	// transmit context for this queue stays in local memory and the new buffer 
	// uses the same dl_buffer_handle as the previous buffer for this queue (the 
	// chance of this condition to happen should be very low)

	alu[*l$index0[TXC_BUFFER_HANDLE_INDEX], --, B, IMPOSSIBLE_DL_HDL_ADDR]

	// check if this is the the last cell in this packet (if this is the last
	// cell in this buffer and EOP flag (bit 31)set in tx request second long word
	// also perform things general to last cell of one buffer in the defer slots
	// of following br_bclr to save 3 instruction cycles

	br_bclr[buffer_handle, EOP_BIT_LOC, finish#], defer[3]

		// it is end of current buffer, set the bd valid bit = 0 and bytes already
		// sent in this buffer to 0

		alu[*l$index0[TXC_BD_VALID_INDEX], --, B, 0]
  
		// this is the last cell in current buffer, it's payload may not be exact
		// FULL_TM_PAYLOAD_SIZE, so we need to calculate here from *l$index[3]
		// and bytes_sent_already
  
		ld_field_w_clr[total_payload_size, 0011, *l$index0[TXC_OFFSET_INDEX], >>16]
		alu[payload_length, total_payload_size, -, bytes_sent_already]   

	// set the EOP flag here

	alu[*l$index0[TXC_SEQ_NUMBER_INDEX], *l$index0[TXC_SEQ_NUMBER_INDEX], OR, 0x80]

finish#:

 	// Get the dram base address for the buffer

    dl_buf_get_data[dram_addr, buffer_handle]

	// get dram read address for buffer by adding offset. We still need to add 
	// bytes_sent_already which we will do when do the dram[] operation

	ld_field_w_clr[buffer_offset, 0011, *l$index0[TXC_OFFSET_INDEX]]
	alu[dram_addr, dram_addr, +, buffer_offset] 

	alu[dram_addr, dram_addr, +, bytes_sent_already]

.end

#endm

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_write_nn_request[]
//
// Description:
// 	
//		Writes nn request to the 2nd ME
//
// Outputs: 
//		none
//
// Inputs:  
//
//		tx_request0					tx_request0 read from NN ring from QM		
//		tx_request1					tx_request1 read from NN ring from QM
//		dram_addr					dram_addr of packet to start transmitting
//		payload_length				payload length to transmit
//		SOP_EOP flag				flags whether cell is SOP or EOP
//		cell_count_remaining		cell count remaining in the packet
//		prepend						4 registers of prepend header
//		
// Constants
//
//		None
//
// Size: 
//
// 
//
/////////////////////////////////////////////////////////////////////////////////////
// writing to NN ring
#macro _csix_tx_write_nn_request[tx_request0, tx_request1, dram_addr, payload_length, SOP_EOP_FLAG, cell_count_remaining, prepend]
// NN ring msg format
// lw0 - tx_request0
// lw1 - dram_addr
// lw2 - 0:15 payload_len , 16 (0 is SOP_EOP), 17 ( 1 if MOP_EOP) 24:31 - cell_count_remaining
// lw3-6 - prepend0-3
writing_to_NN_ring#:
	br_inp_state[NN_FULL, writing_to_NN_ring#]

#if (SOP_EOP_FLAG == SOP_EOP)
	// write request in case of min packet
	// leave sop and eop bits as zero
	alu[*n$index++, --, b, tx_request0]
	alu[*n$index++, --, b, tx_request1]
	alu[*n$index++, --, b, dram_addr]
	alu[*n$index++, --, b, payload_length]
	alu[*n$index++, --, b, prepend/**/0]
	alu[*n$index++, --, b, prepend/**/1]
	alu[*n$index++, --, b, prepend/**/2]
	alu[*n$index++, --, b, prepend/**/3]
#else 
#if (SOP_EOP_FLAG != MOP_EOP)
	alu[*n$index++, --, b, tx_request0]
	alu[*n$index++, --, b, tx_request1]
	alu[*n$index++, --, b, dram_addr]
	.begin
	.reg temp
	alu[temp, payload_length, or, 1, <<16]
	alu[*n$index++, temp, or, cell_count_remaining, <<24]
	.end
	alu[*n$index++, --, b, prepend/**/0]
	alu[*n$index++, --, b, prepend/**/1]
	alu[*n$index++, --, b, prepend/**/2]
	alu[*n$index++, --, b, prepend/**/3]
#else
	// MOP_EOP case
	alu[*n$index++, --, b, tx_request0]
	alu[*n$index++, --, b, tx_request1]
	alu[*n$index++, --, b, dram_addr]
	.begin
	.reg temp
	alu[temp, payload_length, or, 1, <<16]
	alu[temp, payload_length, or, 1, <<17]
	alu[*n$index++, temp, or, cell_count_remaining, <<24]
	.end
	alu[*n$index++, --, b, prepend/**/0]
	alu[*n$index++, --, b, prepend/**/1]
	alu[*n$index++, --, b, 0]
	alu[*n$index++, --, b, 0]
#endif
#endif

#endm


/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_debug_init_counters[]
//
// Description:
// 	
//		Initialize the debug counters in absolute registers to zeros.
//
// Outputs: 
//
//		 			
//
// Inputs:  
//
//		
//		
// Constants
//
//		None
//
// Size: 
//
//		3 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////
#macro	_csix_tx_debug_init_counters[]

#ifdef	_DEBUG_COUNTERS_

	immed[@csix_tx_rcvd_cells,0x0]	
	immed[@csix_tx_num_tx_req,0x0]	

#endif
#endm

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_debug_incr_counter[]
//
// Description:
// 	
//		Increment the specified counter by 1. This macro is enabled when the build
//	switch _DEBUG_COUNTERS_ is enabled.
//
// Outputs: 
//
//		 			
//
// Inputs:  
//		counter		the debug counter to increment
//		
//		
// Constants
//
//		None
//
// Size: 
//
//		1 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////
#macro	_csix_tx_debug_incr_counter[counter]

#ifdef	_DEBUG_COUNTERS_

	alu[counter,counter,+,1]

#endif
#endm

/////////////////////////////////////////////////////////////////////////////////////

#endif 		// __CSIX_TX_UTIL1_UC__

/////////////////////////////////////////////////////////////////////////////////////

