/////////////////////////////////////////////////////////////////////////////////////
//
//
//                  I N T E L   P R O P R I E T A R Y
//
//     COPYRIGHT [c]  2002 BY  INTEL  CORPORATION.  ALL RIGHTS
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT
//     THE PRIOR WRITTEN PERMISSION OF :
//
//                        INTEL  CORPORATION
//
//                     2200 MISSION COLLEGE BLVD
//
//               SANTA  CLARA,  CALIFORNIA  95052-8119
//
/////////////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: csix_tx_util.uc
//
//      Purpose: Utility macros for csix transmit 
//
//
//      History:
//
//      Date            Comment                         				By
//      ---------------------------------------------------------------------------
//
//      11/21/2001      Created                         			David Chou
//
//		3/18/02			Modified to do packet interleaving			Uday Naik	
//						and adhere to coding convention
//
/////////////////////////////////////////////////////////////////////////////////////

#ifndef __CSIX_TX_UTIL2_UC__
#define __CSIX_TX_UTIL2_UC__


/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_get_next_tbuf[]
//
// Description:
// 	
//		Allocate a tbuf element. Instead of returning the element number we return
//		two values which are actually useful - the element address and the address 
//		for validating the tbuf. This allows us to optimize the wrap around case 
//		better.
//
// Outputs: 
//
//		addr_of_tbuf:		Address of the TBUF for data copy
//		addr_of_tx_control:	Addr for validating the tbuf
//
// Constants
//
//		None
//
// Size: 
//
//		9 instructions in worst case 
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_get_next_tbuf[addr_of_tbuf, addr_of_tx_control]

.begin	

	.reg	tbuf_element
	.reg	tmp
	
	// get the current tbuf element
  
	alu[tbuf_element, --, B, *l$index1[GLOBAL_AVAIL_TBUF_ELEMENT_INDEX]]

	// increase @avail_tbuf_element by 1

	alu[*l$index1[GLOBAL_AVAIL_TBUF_ELEMENT_INDEX], \
						*l$index1[GLOBAL_AVAIL_TBUF_ELEMENT_INDEX], +, 1] 

	// Check if @avail_tbuf equal to MAX_AVAIL_DATA_TBUF

	alu[--, *l$index1[GLOBAL_AVAIL_TBUF_ELEMENT_INDEX], -, MAX_AVAIL_DATA_TBUF]

	blt[EXIT#] , defer [3]
		
		// compute the address of the tbuf 

		alu_shf[tmp, --, B, tbuf_element, <<7]
		alu[addr_of_tbuf, tmp, +, *l$index1[GLOBAL_TBUF_BASE_INDEX]]

		// start the compute of the address for tx control 

#if IS_IXPTYPE(__IXP28XX)
		alu_shf[tmp, --, B, tbuf_element, <<4]
#else
		alu_shf[tmp, --, B, tbuf_element, <<3]
#endif

	// if we get here, we need to reset to the first element. 

	alu[*l$index1[GLOBAL_AVAIL_TBUF_ELEMENT_INDEX] , --, B, FIRST_AVAIL_DATA_TBUF]

EXIT#:

	// finish the compute of the address of the tx control 

	alu[addr_of_tx_control, tmp, +, *l$index1[GLOBAL_CTW_VD_BASE_INDEX]]

.end

#endm


/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_transmit_tbuf[]
//
// Description:
// 	
//		Validate a TBUF for transmit. Also if the cell count has gone to 0
//		then free the buffer  
//
// Outputs: 
//
// Inputs:  
//
//		$tcw						xfer register
//		addr_of_tx_control:			address to validate tbuf
//		tcw:						transmit control word
//		ex_header:					extension header
//		buffer_handle:				buffer handle
//		cell_count_remained:		cell count remaining
//		signal_number:				signal number
//		
// Constants
//
//		None
//
// Size: 
//
//		8 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_transmit_tbuf[ $tcw, addr_of_tx_control, tcw, ex_header, \
								buffer_handle, cell_count_remained, signal_number ]

.begin

	// check if the cell count remaining for the buffer is 0

	br!=byte[cell_count_remained, 0, 0x00, exit#] , defer [2]

		// copy the element control word and the extension header into write
		// transfer registers

		alu[$tcw[0],	--, B, tcw]			; element control word
		alu[$tcw[1],	--, B, ex_header]	; element CSIX extension header

	// increment debug counter for buffers freed
	_csix_tx_debug_incr_counter[@csix_tx_num_buf_freed]

	// free the buffer handle 

writing_to_nn_ring#:

	br_inp_state[NN_FULL, default_buf_free#]
	
	alu[*n$index++, buffer_handle, or, 0x3, <<30]

	br[exit#]

default_buf_free#:

	#ifndef FREELIST_MANAGER

	// increment debug counter for buffers freed to sram
	_csix_tx_debug_incr_counter[@csix_tx_num_def_buf_freed]

	dl_buf_free[buffer_handle, BUF_FREE_LIST0]

	#else

	br[writing_to_nn_ring#]

	#endif

exit#:

 	// write transmit contrl word (TCW) into MSF. write to TBUF_ELEMENT_CTRL_V_# 
	// corresponding the tbuf_element to initiate transmission.

	msf[write, $tcw[0], addr_of_tx_control, 0, 2], sig_done[signal_number]

	// increment debug counters for c-frames transmitted.
	_csix_tx_debug_incr_counter[@csix_tx_num_cells_txed]

.end

#endm

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_get_extension_header[]
//
// Description:
// 	
//		Compute the extension header from the transmit request word 0
//
// Outputs: 
//
//		ex_header					Extension header returned
//
// Inputs:  
//
//		tx_request0					transmit request 0
//		
// Constants
//
//		None
//
// Size: 
//
//		3 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_get_extension_header[ex_header, tx_request0]

.begin

	.reg 	class_id
	
	// tx_request0 contains the class id (0..1), the destination blade on the
	// fabric (2..7)

	// class id is the first byte in the extension header for unicast cframe, it
	// is in bit 32 to bit 24, and  the destination blade is in bit 11 to bit 0


 	// shift by 2 bits and then and with 0x3f to get 6 bits of destination blade

	alu_shf[ex_header, DESTINATION_FIELD_MASK, AND, tx_request0, >>DEST_FIELD_SHIFT]	

	#ifndef _HARDWARE_DEBUG_

	// get the class id which is in bits 0..1. And with 0x3

	alu[class_id, CLASS_FIELD_MASK, AND, tx_request0]

	#else

	// when debugging on hardware, we assign destination blade number
	// as the class id. this allows us to test for multiple queues
	// on the egress blade even in non diffserv applications.

	alu_shf[class_id, DESTINATION_FIELD_MASK, AND, tx_request0, >>DEST_FIELD_SHIFT]	

	#endif

	// fill class field (bit 24 to 31) in extension header

    alu_shf[ex_header, ex_header, OR, class_id, <<CLASS_FIELD_SHIFT]


.end

#endm

#ifdef	MULTICAST_SUPPORT_ENABLE

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_cframe_header[]
//
// Description:
// 	
//		Compute the extension header from the transmit request word 0. Also update the
//		base header if it is a multicast packet.
//
// Outputs: 
//
//		ex_header					Extension header returned
//		tcw							Transmit Control Word is modified for
//									Multicast packets
//
// Inputs:  
//
//		tx_request0					transmit request 0
//		
// Constants
//
//		None
//
// Size: 
//
//		5 instructions for unicast packet.
//		8 instructions for multicast packet.
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_cframe_header[tcw, ex_header, tx_request0]

.begin

	.reg 	class_id

	// first check if this is for multicast packet

	alu[--, tx_request0, -, mcast_reserved_voq]

	bne[exit#], defer[3]
	
	#ifndef _HARDWARE_DEBUG_

	// get the class id which is in bits 0..1. And with 0x3

	alu[class_id, CLASS_FIELD_MASK, AND, tx_request0]

	#else

	// when debugging on hardware, we assign destination blade number
	// as the class id. this allows us to test for multiple queues
	// on the egress blade even in non diffserv applications.

	alu_shf[class_id, DESTINATION_FIELD_MASK, AND, tx_request0, >>DEST_FIELD_SHIFT]	

	#endif

	// normal packet

	// tx_request0 contains the class id (0..1), the destination blade on the
	// fabric (2..7)

	// class id is the first byte in the extension header for unicast cframe, it
	// is in bit 32 to bit 24, and  the destination blade is in bit 11 to bit 0


 	// shift by 2 bits and then and with 0x3f to get 6 bits of destination blade

	alu_shf[ex_header, DESTINATION_FIELD_MASK, AND, tx_request0, >>DEST_FIELD_SHIFT]	

	// fill class field (bit 24 to 31) in extension header

    alu_shf[ex_header, ex_header, OR, class_id, <<CLASS_FIELD_SHIFT]

	
	// we optimized critical path, which is unicast packets. For multicast
	// we will overwrite some fields as needed.

mcast_pkt#:

	// update TCW for mcast packet. Reset the type to '3' (Mcast ID frame type)

	alu[tcw, tcw, AND~, 0xf]

	alu[tcw, tcw, OR, 0x3]

	// overwrite the ex_header: set NHID as destination address

	alu_shf[ex_header, mcast_dst_id, OR, class_id, <<CLASS_FIELD_SHIFT]


exit#:

.end

#endm

#endif	// MULTICAST_SUPPORT_ENABLE


/////////////////////////////////////////////////////////////////////////////////////
//	
//	_csix_tx_get_transmit_control_word[]
//
// Description:
// 	
//		Get transmit control word from offset and length
//
// Outputs: 
//
//		tcw					transmit control word
//
// Inputs:  
//
//		tcw_base			base for transmit control word
//		offset				offset into quadword to start transmit
//		payload_length		payload length
//		
// Constants
//
//
//
// Size: 
//
//		2 instructions 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_get_transmit_control_word[ tcw, tcw_base, offset, payload_length]

.begin

	// fill in the byte offset (0 to 7) to start transmit (13..15 of tcw)

	alu_shf[tcw, tcw_base, or, offset, <<13]

	// fill in the payload length into the transmit control word
	
	ld_field[tcw, 1000, payload_length, <<24] 


.end

#endm


/////////////////////////////////////////////////////////////////////////////////////
//
// _csix_tx_init_gprs
//
//	Description: initialize context relative GPR for specific thread
//
//	Size:
//
//		6 instructions
//
/////////////////////////////////////////////////////////////////////////////////////
 
#macro _csix_tx_init_gprs2()


.begin

	// base for the first transmit control word for the first cell of one packet 

	move[tcw_fcip_base, TCW0_FCIP_BASE_LW]

	// base for the first transmit control word for cells which is not the first 
	// cell of one packet 

 	move[tcw_nfcip_base, TCW0_NOT_FCIP_BASE_LW]

 	
	// base to calculate the address field in indirect ref of 
	// dram[tbuf_wr....] instruction

	move[indiref_base, DRAM_INDIREF_BASE_LW]


#ifdef	MULTICAST_SUPPORT_ENABLE

	// initialize mutlicast VOQ number

	alu_shf[mcast_reserved_voq, --, B, MCAST_RESERVED_BLADE_ID, <<2]

#endif

	
.end


#endm 


/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_write_nn_request[]
//
// Description:
// 	
//		Writes nn request to the 2nd ME
//
// Outputs: 
//		tx_request0					tx_request0 		
//		tx_request1					tx_request1
//		dram_addr					dram_addr of packet to start transmitting
//		payload_length				payload length to transmit
//		sop_eop_reg					flags whether cell is SOP or EOP
//		cell_count_remaining		cell count remaining in the packet
//		prepend						4 registers of prepend header
//
// Inputs:  
//
//		none
//		
// Constants
//
//		None
//
// Size: 
//
//		
/////////////////////////////////////////////////////////////////////////////////////
#macro read_nn_request(tx_request0, tx_request1, dram_addr, payload_length, sop_eop_reg, cell_count_remaining, prepend)
	alu[tx_request0, --, b, *n$index++]
	alu[tx_request1, --, b, *n$index++]
	alu[dram_addr, --, b, *n$index++]
	ld_field_w_clr[payload_length, 0011, *n$index]
	ld_field_w_clr[sop_eop_reg, 0001, *n$index, >>16]
	ld_field_w_clr[cell_count_remaining, 0001, *n$index++, >>24]
	// now four more
	wait_for_next_four#:
	br_inp_state[nn_empty, wait_for_next_four#]

	alu[prepend[0], --, b, *n$index++]

#ifdef	MULTICAST_SUPPORT_ENABLE
	
	alu_shf[mcast_dst_id, --, B, *n$index, >>16]

#endif

	alu[prepend[1], --, b, *n$index++]
	alu[prepend[2], --, b, *n$index++]
	alu[prepend[3], --, b, *n$index++]

#endm



/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_clear_signal[]
//
// Description:
// 	
//		Clear a signal from a signal mask 
//
// Outputs: 
//
//
// Inputs:  
//
//		sigmask:			signal mask
//		sig:				signal number to clear
//
// Constants
//
//		None
//
// Size: 
//
//		1-3 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_clear_signal[sigmask, sig]

.begin

	.reg tmp

#if ( isnum(sig) )
	
	#define_eval SIG_SHIFT sig
	alu_shf[sigmask, sigmask, AND~, 1, <<sig ]
	#undef SIG_SHIFT


#else

	alu_shf[sigmask, sigmask, AND~, 1, <<(&sig)]

#endif

.end

#endm

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_set_signal[]
//
// Description:
// 	
//		Set a signal in a signal mask 
//
// Outputs: 
//
//
// Inputs:  
//
//		sigmask:			signal mask
//		sig:				signal number to set
//
// Constants
//
//		None
//
// Size: 
//
//		1-3 instructions
// 
//
////////////////////////////////////////////////////////////////////////////////////

#macro _csix_tx_set_signal[sigmask, sig]

.begin

	.reg tmp

#if ( isnum(sig) )

	#define_eval SIG_SHIFT sig
	alu_shf[sigmask, sigmask, OR, 1, <<sig]
	#undef SIG_SHIFT

#else

	alu_shf[sigmask, sigmask, OR, 1, <<(&sig)]


#endif

.end

#endm


/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_debug_init_counters[]
//
// Description:
// 	
//		Initialize the debug counters in absolute registers to zeros.
//
// Outputs: 
//
//		 			
//
// Inputs:  
//
//		
//		
// Constants
//
//		None
//
// Size: 
//
//		3 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////
#macro	_csix_tx_debug_init_counters[]

#ifdef	_DEBUG_COUNTERS_

	immed[@csix_tx_num_cells_txed,0x0]	
	immed[@csix_tx_num_buf_freed,0x0]	
	immed[@csix_tx_num_def_buf_freed, 0x0]

#endif
#endm

/////////////////////////////////////////////////////////////////////////////////////
// 
// _csix_tx_debug_incr_counter[]
//
// Description:
// 	
//		Increment the specified counter by 1. This macro is enabled when the build
//	switch _DEBUG_COUNTERS_ is enabled.
//
// Outputs: 
//
//		 			
//
// Inputs:  
//		counter		the debug counter to increment
//		
//		
// Constants
//
//		None
//
// Size: 
//
//		1 instructions
// 
//
/////////////////////////////////////////////////////////////////////////////////////
#macro	_csix_tx_debug_incr_counter[counter]

#ifdef	_DEBUG_COUNTERS_

	alu[counter,counter,+,1]

#endif
#endm


/////////////////////////////////////////////////////////////////////////////////////

#endif 		// __CSIX_TX_UTIL2_UC__

/////////////////////////////////////////////////////////////////////////////////////

