////////////////////////////////////////////////////////////////////////////////
//                                                                     
//                  I N T E L   P R O P R I E T A R Y                   
//                                                                      
//     COPYRIGHT (c)  2001-2002 BY  INTEL  CORPORATION.  ALL RIGHTS          
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT   
//     THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                      
//                        INTEL  CORPORATION                            
//                                                                     
//                     2200 MISSION COLLEGE BLVD                        
//                                                                      
//               SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                      
////////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: mphy16_tx.uc
//
//      Purpose: MPHY_16 PACKET (POS/Ethernet) transmitter for Egress IXP2400
//
//      History:
//
//
//      Date            Comment                         		By
//      -----------------------------------------------------------------------	
//
//      03/08/2002      Created									David Chou
//
//
///////////////////////////////////////////////////////////////////////////////


#ifndef	__MPHY16_TX_UC__
#define	__MPHY16_TX_UC__

///////////////////////////////////////////////////////////////////////////////
// 
// This MPHY_16 PACKET transmitter (TX) runs on POS-PHY MPHY-16 transmit mode  
// on the egress IXP2400. The basic unit of transmission is an mpacket, the 
// size of the mpacket is the tbuf element size which can be configured to 64, 
// or 128, or 256 bytes in IXP2400. PACKRET TX receives a transmit request for 
// a packet. If the size of the packet is too large to fit into one mpacket,
// PACKET TX segments this IP packet into several mpackets. This PACKET TX 
// supports up to 16 ports. 
//
// In IXP2400, the Tx_MPHY_status MSF register provides the status of PHY 
// interface for the software to do the transimission flow control to prevent 
// the head of line blocking issue. There are two bits (Tx_Pending[x] and 
// Tx_Status[x]) for each port (x) in Tx_MPHY_Status register. When 
// Tx_Pending[x]=1, it indicates that there are pending transmit activity for 
// that port, then Tx_Status[x] is indeterminant. When Tx_Pending[x]=0 and 
// Tx_Status[x]=1, the port is ready to accept at least "n" mpackets of data.
// This size of "n" depends on the characteristics of the port PHY and size of 
// the tbuf element. Also maximum 16 (QUEUE_DEPTH) packets are queued in the 
// local memory of PACKET TX ME to respond to the change of Tx_MPHY_Status 
// immediately.  Since one ME is not fast enough to finish the tasks for all 
// ports and that certain packets are queued in the local memory, the transmit
// context (TXC) is too huge to be moved between two Packet TX MEs through SRAM, 
// two ME are used in this solution, they are identified as 
// MPHY16_PACKET_TX_FIRST_ME and MPHY16_PACKET_TX_SECOND_ME in the code and 
// defined in "Preprocessor definitions" of "Assembler" of "Build Settings" of
// the project.
//
// Each ME controls up to 8 ports as specified by PORTS_CTRLED_BY_THIS_ME for
// each ME in the code. MPHY16_PACKET_TX_FIRST_ME controls port 0 to port 7, 
// MPHY16_PACKET_TX_SECOND_ME controls port 8 to 15. These two ME runs 
// independently, but the threads inside each ME uses folding mechanism to run
// in strict order of three phases. TBUF elements are shared by both MEs, the 
// next available TBUF element is kept in one longword in the scratch memory, 
// the scratch memory atomic operation SCRACTH[test_and_incr..] is used to get  
// and increment the next available TBUF element.   
//
// In order to avoid overflooding the packet queue in the local memory, this
// microblock reads scratch ring to get tx request for a specific port only 
// when there are spaces in the packet queue for that port. It also keeps 
// account of how many packets transmitted for each port, and update pkts_txd 
// count in Egress Scheduler through a reflect write when it finishs the 
// transmission of one packet, so Egress Scheduler will not overflood it's 
// downstream. 
//
// Following are the assumptions used to develop this Packet TX microblock:
// 1. The meta data at SRAM and the payload at DRAM are correct, Packet TX will 
//    not do any sanity check in the size and order of the buffers in any 
//    packet.
// 2. The payload in DRAM will be contiguous from the offset to the end of a 
//    buffer in a packet.
// 3. The offset for all buffers in a packet except SOP buffer is 0.
//
// There are three issues need to be addressed in the Packet TX microblock. The
// first two are imposed by IXP2400 hardware, the third one keeps this Packet
// TX microblock capable to handle the packets received in diffrent receiving 
// interface in (POS, ATM, Ethernet) in ingress microblocks. They are
// 1. DRAM[tbuf_wr,...] instruction, which is used to move packet data from 
//    DRAM to TBUF, always access DRAM and TBUF on 8-bytes word boundary.
// 2. The sum of the prepend length and payload length of a mpacket must be an
//    integral multiple of the bus width, except for a EOP mpacket. In MPHY-16 
//    mode, the bus width is 32 bits (4 bytes).
// 3. The payload data offset in the first buffer of a packet can be any value. 
// 
// In order to address the abovementioned three issues, Packet TX transmits 
// the maximum allowable payload bytes for every non-eop mpacket by making sure 
// that the sum of the prepend length and payload length of a non-eop mpacket
// is an integral multiple of 4 bytes. If necessary, maximum of 3 bytes at the 
// end of one buffer is saved in the local memory and transmit with the 
// payload of the next buffer of the same packet in the next mpacket.
//
// This Packet TX microblock keeps active buffer (which the microblock is 
// processing) descriptor info and secondary buffer (buffer descriptor for the 
// next buffer after the active buffer in the packet, it is in the software 
// next pointer field of meta data of the active buffer) descriptor info in 
// local memory, this microblock will read the necessary info (buffer size, 
// data offset, freelist, and the software next pointer) of the secondary 
// buffer from sram in parallel with it's processing of the active buffer. When
// the processing of the active buffer finishes, the secondary buffer will 
// become the active buffer, and continue until all buffers in the packet are 
// processed.
//
///////////////////////////////////////////////////////////////////////////////


// Following definition enable Packet TX waiting for an inter-ME signal 
// generated by the ME whcih initializes common resources, such as scratch 
// ring, near the end of thread 0 of Packet TX initialization to make sure all 
// common resource are available. This option should be enabled when this 
// Packet TX microblock is integrated with other egress microblocks.  
#define WAIT_FOR_COMMON_RESOURCE_INITIALIZATION 	

// Following definition runs this Packet TX microblock in SPHY POS mode, it is 
// used for debug before MPHY_16 model is ready in Workbench, should be disabled
// when MPHY_16 model is ready in Workbench.
//#define SIMULATE_MPHY_16_WITH_SPHY	

// Following definition used in the stress test of unit test of this Packet TX 
// microblock, because we don't really allocate buffer in stress test, 
// keeping freeing buffer will change meta data in SRAM and stop the stress
// test. This option should be disabled when this Packet TX microblock is 
// integrated with other egress microblocks. This definition will be enabled 
// through "Processor definifitions" of Assembler in Build Settings of the
// unit test project.  
//#define SKIP_FREE_BD 


// Following definition used in debugging Tx_pending locked in 1 problem, should be
// disabled in normal operation
//#define DEBUG_TX_PENDING_LOCKED

// Following definition used for dbugging purpose, should be disabled for
// normal operation.
// #define DEBUG_FLAG	

// Following definition used to avoid packet generation blocked due to 
// tx_pending bit locked to 1 in two ME running cases
//#define MPHY_STAT_ALWAYS_READY

// Following definition used for dbugging purpose, should be disabled for
// normal operation.
//#define CHECK_DISCARD

// Following definition used for gloable counters, it is turned off in 
// release 3
//#define COUNTERS

// Following definition will adjust mpkt length to avoid SKIP bit set in the 
// last mpkt of non-EOP buffer due to the hardware limitation that the sum of
// payload length and prepend length must be mulitple of deviec bus width
// This is necessary for MPHY-16 mode in IXP2400 chip rev. A0 and A1 due to 
// one MSF hardware bug (HW-bug#1294), more instructions are required to do 
// this, but fortunately, this doesn't happen in worst case path 
#define AVOID_USING_SKIP_BIT_IN_TCW0

// following definition selects sram transfer registers to be used with 
// dl_meta.uc dispatch loop
#define DL_META_DATA_IN_SXFER

//	Include Files

// include stdmac.uc in IXPblocks Portable library

#include <stdmac.uc>

// include localmem.uc in IXPblocks Portable library for read/write local
// memory

#include <localmem.uc>

// include header file with the global definition of common constants which
//  are shared by all microblocks

#include "dl_system.h"

// include the Packet TX header file with microblock specific definitions  

#include "mphy16_tx.h" 

// include local utilization macros within this microblock

#include "mphy16_tx_util.uc"

// include initializatiuon macros within this microblock

#include "mphy16_tx_init.uc"

// include global dispatch loop macros
#include "dl_meta.uc"

// include global dispatch loop macros
#include "dl_buf.uc"

/////////////////////////////////////////////////////////////////////////////
//	Global Registers
//  Absolute GPR variables:
.reg @qstat_vector				; queue status vector for ports, each port 
								; occupies one bit, when set, it means that
								; there is at least one packet queued for 
								; that port
.reg @rr_vector					; round robin mask
.reg @txc_left_paylo_vector		; mask to stop mpacket be transmitted for  
								; specific port when the leftover payload 
								; from previous buffer is not ready in local
								; memory 
.reg @mphy_stat_vector			; flow control vector on port's Tx_MPHY_Status
.reg @txc_rd_snd_bd_vector		; mask to stop mpacket be transmitted for  
								; specific port when the secondary buffer data
								; is not ready in local memory 
.reg @busy_mask					; mask for ports which is not ready due to 
								; Tx_MPHY_Status
#ifdef DEBUG_TX_PENDING_LOCKED
.reg  @tbuf_no_addr
#endif //#ifdef DEBUG_TX_PENDING_LOCKED

#ifdef UNIT_TEST
	.reg	@buf_freed
#endif // UNIT_TEST

#ifdef CHECK_DISCARD
.reg	@pkts_discarded
#endif	//CHECK_DISCARD
// following absolute GPR variables are preassigned in initialization to save 
// instruction cycles in Packet TX process
.reg @tcw_vd_base				; base to calculate TBUF_ELEMENT_CTRL_V_# address
.reg @qpkt_incr					; increment to fisr longword of queue status for 
								; each port to increase it's counts of packets in
								; queue by 1 
.reg @tail_incr					; increment to fisr longword of queue status for 
								; each port to increase it's queue's tail offset
								; by queue entry size
#ifdef COUNTERS
.reg @cntr_base					; start address of Packet TX counters in sram
								; initialized to PACKET_TCX_COUNTER_BASE
.reg $byte_cnt					; sram xfer register to keep value to add to 
								; packets transmitted counter in sram		
#endif //#ifdef	COUNTERS

// context relative GPR variables:
.reg tbuf_ele_1					; tbuf element number for mpacket in current thraed
.reg tcw0_1						; first transmit control word 
.reg exe_stat_flag				; thread execution status flags 
.reg deq_port					; port for the mpacket handled by this thread
.reg deq_port_bit				; bit pattern of deq_port = (1 << deq_port)
.reg rmnd_paylo					; remained payload size for the active buffer
.reg offset_rpaylo				; offset of the remained payload for the active
								; buffer
.reg sig_mask_1					; wakeup events for a thread to wakeup in phase 1
.reg sig_mask_2					; wakeup events for a thread to wakeup in phase 2

.reg enq_port					; enqueue port number
.reg sigmask_phs1_default 		; default value for signal mask for phase 1
.reg sigmask_phs2_default 		; default value for signal mask for phase 2


// following context relative GPR variables are preassigned in initialization 
// to save instruction cycles in Packet TX process
.reg tbuf_base   				; base to calculate tbuf address
.reg sring_tr           		; scratch ring which QM put tx requests
.reg sig1_next_context_gpr		; value to write to SAME_ME_SIGNAL csr to wake up
								; next thread
.reg sig2_next_context_gpr		; value to write to SAME_ME_SIGNAL csr to wake up
								; next thread
.reg sig3_next_context_gpr		; value to write to SAME_ME_SIGNAL csr to wake up
								; next thread

.reg indiref_base				; base to calculate the address field in indirect ref
								; of dram[tbuf_wr....] instruction
.reg tbuf_scr_addr				; scratch memory address where the next available tbuf
								; element and kept and incremented by scratch atomic
								; operation
.reg queue_stat_base			; starting local memory address for queue status 
								; info
.reg tx_mphy_stat_addr			; address of MSF Tx_MPHY_Staus register
.reg tx_seq_0_addr				; address of MSF Tx_Sequence_0 register

// context relative sram xfer registers
// following context relative sram registers are used to keep values gotten in 
// I/O intructions
.reg $tx_request_lw0			; tx request from scratch ring 
.reg $mphy_stat					; status of PHY from Tx_MPHY_Status MSF 
								; register
.reg $avail_tbuf_ele			; next available tbuf element from scratch 
								; atomic operation
.reg $txed_port					; packets transmitted for specific port, used
								; reflect write
.reg $mpkts_sent0				; mpackets actually transmitted in 
								; Tx_Sequence_0	MSF register
///////////////////////////////////////////////////////////////////////////////

//	Global Signals
.sig sig_scratch_access_1 			; signal for scratch ring read
.sig sig_scratch_access_2 			; signal for scratch memory read
.sig volatile sig1_next_context		; signal to wakeup next thread, declared as
									; volatile to eliminate assembling warning
.sig volatile sig2_next_context		; signal to wakeup next thread, declared as
									; volatile to eliminate assembling warning
.sig volatile sig3_next_context		; signal to wakeup next thread, declared as
									; volatile to eliminate assembling warning
.sig sig_sram_read_1 				; signal 1 for sram read
#ifdef DL_SPLIT_META_DATA
.sig sig_sram_read_11 				; signal 11 for sram read
#endif
.sig sig_msf_access_1				; signal 1 for msf access (read or write)
.sig sig_msf_access_2				; signal 2 for msf access (read or write)
.sig sig_msf_access_3				; signal 3 for msf access (read or write)
.sig sig_reflect_write 				; signal for reflect write
.sig sig_dram_access_1				; signal 1 for dram access (dram[tbuf_wr..])


#ifdef DEBUG_TX_PENDING_LOCKED
.sig sig_sram_write_1 				; signal 1 for sram write
.addr sig_sram_write_1      	SRAM_WRITE_1_SIG 		; manual signal allocation
#endif //#ifdef DEBUG_TX_PENDING_LOCKED

#ifdef COUNTERS
.sig sig_counter_update									; signal for update counter
.addr sig_counter_update		SCRTACH_ACCESS_1_SIG	; manual signal allocation
														; reuse this signal
														; there is no conflict	
#endif //#ifdef COUNTERS

//////////////////////////////////////////////////////////////////////////////
// mphy16_tx[]
//
// Description: inifinite loop for Packet TX process 
//                
// Outputs: 
//
//								None
//
// Inputs:  
//
//								None
//
// Constants
//
//								None

//
// Size: 
//
//		153 instructions for worst case (mini packet)
//
/////////////////////////////////////////////////////////////////////////////////////

#macro mphy16_tx[]

// following start# is the beginning of infinite packet transmit processing 
// loop
	// allocate 3 read xfer registers for reading in the meta data 
	xbuf_alloc[$dl_meta, 3, read]	

start#:

// beginning of phase 1
	// wake up next thread in beginning of phase 1 to reduce wakeup latency
	local_csr_wr[SAME_ME_SIGNAL, sig1_next_context_gpr] 

	// reset exe_stat_falg 
	alu[exe_stat_flag, --, b, 0]					; reset flags
	
	// since 8 ports share the same scratch ring to get tx request for them, 
	// we don't get this coming tx request is for which port, so we can't 
	// whether to get tx request or not based on the queue condition of
	// each port 
	// get tx request from scracth ring
	scratch[get, $tx_request_lw0, sring_tr, 0, 1], sig_done[sig_scratch_access_1]

update_mphy_stat#:
	// update mphy_stat_vector, busy_mask, with updated Tx_MPHY_Status
	_mphy16_tx_adjust_mphy_fc_mask[]
 
check_mpkt_ready_for_tx#:
.begin
.reg flowcntl_vector
.reg final_vector
.reg minus_two
.reg qentry_lm_base
	// check whether any port is ready to transmit
	alu[flowcntl_vector, @qstat_vector, and, @mphy_stat_vector]
	alu[flowcntl_vector, flowcntl_vector, and, @txc_left_paylo_vector]
	alu[flowcntl_vector, flowcntl_vector, and, @txc_rd_snd_bd_vector]
	alu[final_vector, flowcntl_vector, and, @rr_vector]

	ffs[deq_port, final_vector]
	// find port, not worst case, use defer[1] to make it equal to worst case 
	bne[do_deq#], defer[1]
	// use defer slots to initialize minus_two (0xFFFFFFFE) for use 
	// in following	_mphy16_tx_update_rond_robin_vecror[deq_port, minus_two] 
    	alu[minus_two, --, ~b, 1]	; set minus to 0xFFFFFFFE

	ffs[deq_port, flowcntl_vector]
	// no port ready, not worst case, can take branchout penalty 
	beq[no_port_ready_for_tx#]

	// start to prepare mpacket to transmit
do_deq#:

	// now we need one tbuf element for transmission 
	// get next available tbuf element  from scratch memory by atomic operation, 
	// also increase it in scractch memory
	scratch[test_and_incr, $avail_tbuf_ele, tbuf_scr_addr, 0], \
			sig_done[sig_scratch_access_2]

	// get mpackets actually sent out of the tbuf for use in phase 2 to avoid 
	// overwritting existing mpackets which have not been transmitted out yst
	msf[read, $mpkts_sent0, tx_seq_0_addr, 0, 1], sig_done[sig_msf_access_3]	

	//	updating round robin vecror (@rr_vector)
	_mphy16_tx_update_rond_robin_vecror[deq_port, minus_two]

	// set local memory index 1 to deq_port's queue status info, also initialize
	// qentry_lm_base for setting local memory index 0 
	_mphy16_tx_set_lmindex1_for_deq_port_stat[qentry_lm_base, deq_port]

	// set local memory index 0 to queue head of the deq_port 
	// initialize tcw0 for SOP_EOP mpacket (worst case) with deq_port in channel 
	// field
	_mphy16_tx_set_lmindex0_for_deq_entry[tcw0_1, qentry_lm_base, deq_port]

.end // flowcntl_vector final_vector minus_two qentry_lm_base


.begin 
.reg tmp
	alu[tmp, --, b, *l$index0[ABD_0_OFFSET]]
	// not SOP mpacket,  not worst case, can take branchout penatly
    br_bclr[tmp, SOP_BIT_LOC, not_sop_mpacket#]
.end

.begin 
.reg tmp
.reg offset_mod_8
.reg addr_of_tbuf
.reg eob_bd
.reg dram_addr
.reg indir_ref
.reg addr_of_tx_control
.reg $w0
.reg $w1
.xfer_order $w0 $w1
// .reg eob_flist 	; not used now

check_whether_sop_eop_mpacket#:
	alu_shf[rmnd_paylo, --, b, *l$index0[ABD_1_OFFSET], >>PAYLO_RMND_LOC]
    ld_field_w_clr[offset_rpaylo, 0011, *l$index0[ABD_1_OFFSET]]
	alu[offset_mod_8, offset_rpaylo, and, CONST_07]
    alu[tmp, TBUF_ELE_SIZE, -, offset_mod_8]
	alu[--, rmnd_paylo, -, tmp]
	// not sop_eop packet, can take branchout penalty
	bgt[sop_only_mpacket#]

	// it is sop_eop packet
 	// before _advance_queue_head, need to save eob_bd and eob_flist
	// for free buffer in phase 3
    ld_field_w_clr[eob_bd, 0111, *l$index0[ABD_0_OFFSET]]
	// flist not used now
	//alu_shf[eob_flist, FREE_LIST_MASK, and, *l$index0[ABD_0_OFFSET], \
	//		>>FLIST_QUEUE_LOC] 

	// advanced queue head, so next packet in the queue of this port will be 
	// processed when this port is selected again
	_mphy16_tx_advance_queue_head[deq_port_bit, deq_port]
	// decrease the allowed_mpkts of the port by one, if it become zero, 
	// remove this port from @mohy_stat_fc_vector
	_mphy16_tx_decrement_allowed_mpkts_in_this_port[deq_port_bit]
	// notify scheduler that one packet has been transmitted from this port 
	_mphy16_tx_notify_scheduler[deq_port, $txed_port]

	// swap out to wait scrtach access, reflect write, msf read to complete 
	// and my turn
	ctx_arb[sig_scratch_access_1, sig_scratch_access_2, sig_reflect_write, \
			sig_msf_access_3, sig2_next_context], and, defer[2]

		alu[sig_mask_2, --, b, 0]			; reset sig_mask_2
		// workaround for HW-bug#1249, copy channel # into tcw1 (reserved in PRM)
		alu_shf[$w1, --, b, deq_port]
// end of of phase 1 with sop_eop mpkt ready for tx	


// beginning of phase 2 with sop_eop mpkt ready for tx	
	// wake up next thread in beginning of phase 2 to reduce wakeup latency
	local_csr_wr[SAME_ME_SIGNAL, sig2_next_context_gpr] 


	// initialize the sig_mask_2 for sop_eop_mpkt case 
	alu[sig_mask_2, --, b, sigmask_phs2_default]		; initialize to default

	// $avail_tbuf_ele read in phase 1 is ready now, get next available 
	// tbuf element (tbuf_ele) and corresponding address (addr_of_tbuf) from it
	_mphy16_tx_get_tbuf_ele_and_addr[tbuf_ele_1, addr_of_tbuf, \
									$avail_tbuf_ele, tbuf_base]

	// check tbuf full to avoid overwritting tbuf [10]
	_mphy16_tx_check_tbuf_full(tbuf_ele_1, sop_eop_tbuf_full#)

sop_eop_ready_to_transmit#: 

// following .set used to eliminate assembling warning
//.set addr_of_tbuf eob_bd offset_mod_8 $w0 $w1
	// dram access takes longer latency, so do it first
	// move payload to tbuf
	_mphy16_tx_move_sop_paylo_to_tbuf[indir_ref, dram_addr, eob_bd, addr_of_tbuf, indiref_base, offset_rpaylo, rmnd_paylo, offset_mod_8, sig_dram_access_1] 

 	// set payload_length and payload_offset fields in tcw0_1
	_mphy16_tx_set_tcw0_with_paylo_len_offset[tcw0_1, rmnd_paylo, offset_mod_8]

	
	// now, check whether a valid tx request gotten in the scratch read from the
	// scratch ring in phase 1. If there is a valid tx request, enqueu that 
	// packet to the queue of that port 
.begin 
.reg pkts_in_queue
.reg qentry_lm_base
// following .set used to eliminate assembling warning
.set $dl_meta[0] $dl_meta[1] $dl_meta[2] enq_port
	// check whether there is null requests in scratch ring
	alu[--, $tx_request_lw0, -, 0x0]		; valid tx request
	// no valid tx request, not worst case, can take branchout penalty
	beq[sop_eop_no_valid_tx_request#]		; no valid tx request
	
	// set sop meta read flag in exe_stst_flag 
	alu_shf[exe_stat_flag, exe_stat_flag, or, 1, <<READ_SOP_BD_FLAG_BIT] 

	// set local memory index 1 to port queue status info, so ppkts_in_quue
	// can be gotten to determine whether it is necessary to discard this packet
	// due to the queue full
	_mphy16_tx_set_lmindex1_for_enqueue_sop[qentry_lm_base, enq_port, $tx_request_lw0]

    // read SOP meta data, read earlier to reduce latency 
	// read meta data from SRAM by using the BD pointer in tx request
 	_mphy16_tx_read_meta_data_from_sram($dl_meta, sig_mask_2, $tx_request_lw0) 

	// check if the queue is full already, theorecically this should not happen,
	// Scheduler wll control the number of packets queued to QM
	alu_shf[pkts_in_queue, PKTS_IN_Q_MASK, and, *l$index1[Q_STAT_OFFSET], >>PKTS_IN_QUEUE_SHFT]	
	alu[--, QUEUE_DEPTH, -, pkts_in_queue]	

	// queue full, not worst case, can take branchout penalty
	beq[sop_eop_discard_this_pkt#]	; queue for this port full, discard this packet

	// following stuff can be done in phase 3, but put here for consuming sram
	// access latency, set local memory index 0 to queue entry at queue tail of 
	// that port and enqueue this packet 
	_mphy16_tx_set_lmindex0_and_enqueue_sop[qentry_lm_base, $tx_request_lw0]
.end // pkts_in_queue qentry_lm_base

sop_eop_no_valid_tx_request#:
	// end of phase 2
	// swap out to wait for dram and msf acess complete, and wait for my turn
	// also calculate address of TBUF_ELEMENT_CONTROL_V# for tbuf_element for
	// later use to save 2 instruction cycles in ctx_arb defer slots
	// swap out to wait signals specified in sig_mask_2
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask_2]	; set active_ctx_wakeup_events
														; csr
	ctx_arb[--]	, defer[2]								; ctx_arb on sig_mask_2
#if IS_IXPTYPE(__IXP28XX)
		alu_shf[tmp, --, B, tbuf_ele_1, <<ELE_2_TX_CNTRL_ADDR]
#else // #ifdef IXP2800
		alu_shf[tmp, --, B, tbuf_ele_1, <<3]
#endif // #ifdef IXP2800
		alu[addr_of_tx_control, tmp, +, @tcw_vd_base]
	
	// following .io_complete is used to eliminate assembling warning due to the
	// usage of "ACTIVE_CTX_WAKEUP_EVENTS" csr
	.io_completed sig3_next_context
	.io_completed sig_dram_access_1
	.io_completed sig_msf_access_1
	.io_completed sig_sram_read_1
#ifdef DL_SPLIT_META_DATA
	.io_completed sig_sram_read_11
#endif
#ifdef DEBUG_TX_PENDING_LOCKED
	.io_completed sig_sram_write_1 	
#endif //#ifdef DEBUG_TX_PENDING_LOCKED

// end of of phase 2 in sop_eop mpacket path	


// beginning of phase 3	in sop_eop mpacket path
	// wake up next thread in the very beginning of phase 3 to reduce wakeup 
	// latency
	local_csr_wr[SAME_ME_SIGNAL, sig3_next_context_gpr]

	// get Tx_MPHY_Status by msf read for use in phase 1
	msf[read, $mphy_stat, tx_mphy_stat_addr,  0, 1], sig_done[sig_msf_access_2]	

    // write transmit contrl word (TCW) to TBUF_ELEMENT_CTRL_V_# corresponding
	// the tbuf_element to initiate transmission
    // Note that we have to adjust transmit port number to base assigned to Packet TX
	alu[$w0, tcw0_1, +, PACKET_TX_PORT_BASE] ; element trnsmit control word
	msf[write, $w0, addr_of_tx_control, 0, 2], sig_done[sig_msf_access_1]

#ifdef COUNTERS
	// update bytes tranmitted 	
	_mphy16_tx_update_bytes_transmitted[tcw0_1, sig_counter_update, $byte_cnt]			
#endif //#ifdef	COUNTERS

	_mphy16_tx_free_buffer(eob_bd)			; free buffer

	// if didn't read meta data in phase 2, just branch out
	br_bclr[exe_stat_flag, READ_SOP_BD_FLAG_BIT, sop_eop_cont_1#] 

	// medat data for sop is ready, move them to local memory
	_mphy16_tx_save_sop_meta_to_lm[enq_port] 

sop_eop_cont_1#:
#ifndef COUNTERS
    // swap out to wait for  msf access to complete and my turn
	ctx_arb[sig_msf_access_1, sig_msf_access_2, sig1_next_context], br[start#]
#else // #ifndef COUNTERS
    // swap out to wait for  msf access and sram access to complete and my turn
	ctx_arb[sig_msf_access_1, sig_msf_access_2, sig_counter_update, \
			sig1_next_context], br[start#]
#endif //#ifndef COUNTERS
// end of phase 3 in the sop_eop mpacket path

sop_eop_discard_this_pkt#:
 	// discard tx request, like no vaid tx request
	// local memory index 1 already set, but it is not used, and will be 
	// overwritten later  
	alu_shf[exe_stat_flag, exe_stat_flag, and~, 1, <<READ_SOP_BD_FLAG_BIT] 	
#ifdef CHECK_DISCARD
	alu[@pkts_discarded, @pkts_discarded, +, 1]
#endif	//CHECK_DISCARD
#ifdef COUNTERS
	_mphy16_tx_update_pkts_dropped[deq_port]	; update packets discarded					
#endif //#ifdef	COUNTERS
	br[sop_eop_no_valid_tx_request#]

///////////////////////////////////////////////////////////////////////////////
// branch taken to wait to make sure tbuf elements not overwritten before they
// have been transmitted in eop_sop mpacket case, otherwise it also causes
// tx_pending bits in Tx_MPHY_Status msf register locked in 1, in addition to 
// screwing up previos mpackets		
///////////////////////////////////////////////////////////////////////////////
sop_eop_tbuf_full#:	
	_mphy16_tx_wait_till_tbuf_not_full(tbuf_ele_1, sop_eop_ready_to_transmit#)	
.end  // tmp offset_mod_8 addr_of_tbuf eob_bd dram_addr indir_ref 
      // addr_of_tx_control $w0 $w1 eob_flist

// end of sop_eop mpacket processing code

///////////////////////////////////////////////////////////////////////////////


// Following hadling branch out cases which is not sop_eop mpacket
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// beginning of no port ready to transmit code 
// no port ready to transmit, check whether valid tx request received
no_port_ready_for_tx#:

	// swap out to wait scratch read Tx request to complete and my turn
	ctx_arb[sig_scratch_access_1, sig2_next_context], and, defer[2]
		alu[sig_mask_2, --, b, 0]		; reset sig_mask_2
		nop
// end of phase 1 of no port ready to transmit

// beginning of phase 2 of no port ready to transmit 
	// wake up next thread in beginning of phase 2 to reduce wakeup latency
	local_csr_wr[SAME_ME_SIGNAL, sig2_next_context_gpr]

	_mphy16_tx_set_signal[sig_mask_2, sig3_next_context] ; initialized to sig3_next_context

	// now, check whether a valid tx request gotten in the scratch read from the
	// scratch ring in phase 1. If there is a valid tx request, enqueu that 
	// packet to the queue of that port 
.begin 
.reg pkts_in_queue
.reg qentry_lm_base
// following .set used to eliminate assembling warning
.set $dl_meta[0] $dl_meta[1] $dl_meta[2] enq_port
	alu[--, $tx_request_lw0, -, 0x0]
	beq[no_port_rdy_no_tx_request#]

	// set sop meta read flag in exe_stst_flag 
	alu_shf[exe_stat_flag, exe_stat_flag, or, 1, <<READ_SOP_BD_FLAG_BIT]  ; set flag

	// set local memory index 1 to port queue status info, so ppkts_in_quue
	// can be gotten to determine whether it is necessary to discard this packe
	// due to the queue full
	_mphy16_tx_set_lmindex1_for_enqueue_sop[qentry_lm_base, enq_port, $tx_request_lw0]

    // read SOP meta data, read earlier to reduce latency 
	// read meta data from SRAM by using the BD pointer in tx request
 	_mphy16_tx_read_meta_data_from_sram($dl_meta, sig_mask_2, $tx_request_lw0) 

	// check if the queue is full already, theorecically this should not happen,
	// Scheduler wll control the packets queued to QM
	alu_shf[pkts_in_queue, PKTS_IN_Q_MASK, and, *l$index1[Q_STAT_OFFSET], >>PKTS_IN_QUEUE_SHFT]	
	alu[--, QUEUE_DEPTH, -, pkts_in_queue]	

	beq[no_port_rdy_discard_this_pkt#]		; queue for this port full, discard 
											; this packet
 	
	// set local memory index 0 to queue tail of this port and enqueue this packet 
	_mphy16_tx_set_lmindex0_and_enqueue_sop[qentry_lm_base, $tx_request_lw0]
.end // pkts_in_queue qentry_lm_base
no_port_rdy_no_tx_request#:
	// swap out to wait signals specified in sig_mask_2
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask_2]	; set active_ctx_wakeup_events
														; csr
	ctx_arb[--]											; ctx_arb on sig_mask_2
	// following .io_complete is used to eliminate assembling warning due to the
	// usage of "ACTIVE_CTX_WAKEUP_EVENTS" csr
	.io_completed sig3_next_context
	.io_completed sig_sram_read_1
#ifdef DL_SPLIT_META_DATA
	.io_completed sig_sram_read_11
#endif
// end of phase 2 of no port ready to transmit

// beginning of phase 3 of no port ready to transmit
	// wake up next thread in beginning of phase 3 to reduce wakeup latency
	local_csr_wr[SAME_ME_SIGNAL, sig3_next_context_gpr]

	// get Tx_MPHY_Status by msf read for use in phase 1
	msf[read, $mphy_stat, tx_mphy_stat_addr,  0, 1], sig_done[sig_msf_access_1]	

	// if didn't read meta data in phase 2, just branch out
	br_bclr[exe_stat_flag, READ_SOP_BD_FLAG_BIT, no_port_rdy_cont_1#] 

	// meda data for sop is ready, move them to local memory
	_mphy16_tx_save_sop_meta_to_lm[enq_port] 

no_port_rdy_cont_1#:
	// swap out to waitfor msf read to complete and  my turn
	ctx_arb[sig_msf_access_1, sig1_next_context], and, br[start#]
// end of phase 3 of no port ready to transmit

no_port_rdy_discard_this_pkt#:
	// discard tx request, like no tx request
	// local memory index 1 already set, but it is not used, and will be 
	// overwritten later  
	alu_shf[exe_stat_flag, exe_stat_flag, and~, 1, <<READ_SOP_BD_FLAG_BIT]
#ifdef CHECK_DISCARD
	alu[@pkts_discarded, @pkts_discarded, +, 1]
#endif	//CHECK_DISCARD	
#ifdef COUNTERS
	_mphy16_tx_update_pkts_dropped[deq_port]	; update packets discarded					
#endif //#ifdef	COUNTERS														
	br[no_port_rdy_no_tx_request#]
// end of no port ready for transmission code
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// beginning of sop_only mpacket processing code
// sop_only_mpacket path in middle of phase 1
// when code comes to here, rmnd_paylo and offset_rpaylo alraedy set properly 
sop_only_mpacket#:

	// setup basic sig_mask_1 for sig2_next_context, sig_scratch_access_1, 
	// sig_scratch_access_2, and sig_msf_access3
	alu[sig_mask_1, --, b, sigmask_phs1_default ]	; set to sigmask_phs1_default 	
	// clear sig_reflect_write from sig_mask_1 first, since this is sop_only
	// mpacket, not EOP mpacket
	_mphy16_tx_clear_signal(sig_mask_1, sig_reflect_write)

.begin 
.reg tmp
.reg sop_bd
.reg sop_only_paylo
.reg offset_mod_8
.reg addr_of_tbuf
.reg dram_addr
.reg indir_ref
.reg addr_of_tx_control
//.reg deq_port_bit
.reg $w0
.reg $w1
.xfer_order $w0 $w1
	// prepare the deq_port_bit (which is equal to 1 <<deq_port) for later use
	alu[--, deq_port, or, 0]
	alu_shf[deq_port_bit, --, b, 1, <<indirect]		; indirect shift, get deq_port_bit

	// reset p_sop_flag in queue entry
	alu_shf[*l$index0[ABD_0_OFFSET], *l$index0[ABD_0_OFFSET], and~, 1, <<SOP_BIT_LOC]

// following .set used to eliminate assembling warning
.set $dl_meta[0] $dl_meta[1] $dl_meta[2] enq_port  
	// check whether to read secondary BD 
	alu[tmp, --, b, *l$index0[ABD_0_OFFSET]]
    br_bclr[tmp, RD_NBD_BIT_LOC, sop_only_rd_sbd_done#]


    // read secondary buffer meta data from SRAM 
	_mphy16_tx_read_meta_data_from_sram($dl_meta, sig_mask_1, \
 														*l$index0[SBD_0_OFFSET]) 
	// reset read secobary buffer meta data bit 
	alu_shf[*l$index0[ABD_0_OFFSET], *l$index0[ABD_0_OFFSET], and~, 1, <<RD_NBD_BIT_LOC]
	// set read secondary BD flag in exe_stat_flag for use in phase 2
	alu_shf[exe_stat_flag, exe_stat_flag, or, 1, <<READ_SND_BD_FLAG_BIT]	; set flag
	// check whether this port needed to be removed from @txc_rd_snd_bd_vector to
	// prvent any following threads in the same run to access the secondary buffer meta  
	// data (which is not ready yet in this phase, but will be ready in next phase) when
	// it becomes active buffer
	// calculate the max bytes to consume before secondary buffer meta data need to be
	// ready
	alu_shf[tmp, --, b, TBUF_ELE_SIZE, <<ME_THREAD_NUM_SHFT] ; max bytes required
	alu[--, rmnd_paylo, -, tmp]			; compare with how much left
	bgt[sop_only_rd_sbd_done#]
	// remove this port from @txc_rd_snd_bd_vector to prvent any following
	// threads in the same run to access the secondary buffer data (which 
	// is not ready yet) when it becomes active buffer, put this port back
	// when secondary meta data saved in phase 2.
	alu[@txc_rd_snd_bd_vector, @txc_rd_snd_bd_vector, and~, deq_port_bit]

sop_only_rd_sbd_done#:
  	// get sop_bd, offset_rpaylo, and offset_mod_8 for use in phase 2
    ld_field_w_clr[sop_bd, 0111, *l$index0[ABD_0_OFFSET]]
	alu[offset_mod_8, offset_rpaylo, and, CONST_07]
	
	// now consider the restriction of the limitation that the sum of the prepend
	// length and payload length must be multiple of 4
	alu[tmp, TBUF_ELE_SIZE, -, offset_mod_8]
	alu[sop_only_paylo, tmp, and, CONST_FC]	
	
	// update ab_paylo_rmnd and ab_offset_rpaylo in queue entry with 
	// sop_only_paylo
	_mphy16_tx_update_ab_paylo_rmnd_and_offset_rpaylo[sop_only_paylo]

 	// decrease the allowed_mpkts of the port by one, if it become zero, 
	// remove this port from @mohy_stat_fc_vector
	_mphy16_tx_decrement_allowed_mpkts_in_this_port[deq_port_bit]
 
	// swap out to wait signals specified in sig_mask_1
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask_1] 	; set csr
	ctx_arb[--]	, defer[2]					
		//use defer slot to initialize the basic sig_mask_2
		// initialize the basic sig_mask_2 for sop_only_mpkt case 
		alu[sig_mask_2, --, b, sigmask_phs2_default] ; initialize to sigmask_phs2_default
		// workaround for HW-bug#1249, copy channel # into tcw1 (reserved in PRM)
		alu_shf[$w1, --, b, deq_port]
	// following .io_complete is used to eliminate assembling warning due to the
	// usage of "ACTIVE_CTX_WAKEUP_EVENTS" csr
	.io_completed sig2_next_context
	.io_completed sig_scratch_access_1
	.io_completed sig_scratch_access_2
	.io_completed sig_msf_access_3
	.io_completed sig_sram_read_1
#ifdef DL_SPLIT_META_DATA
	.io_completed sig_sram_read_11
#endif
// end of of phase 1 with sop_only mpkt ready for tx


// beginning of phase 2 with sop_only mpkt ready for tx	
	// wake up next thread in beginning of phase 2 to reduce wakeup latency
	local_csr_wr[SAME_ME_SIGNAL, sig2_next_context_gpr] 


	// $avail_tbuf_ele read in phase 1 is ready now, get next available 
	// tbuf element (tbuf_ele) and corresponding address (addr_of_tbuf) from it
	_mphy16_tx_get_tbuf_ele_and_addr[tbuf_ele_1, addr_of_tbuf, \
									$avail_tbuf_ele, tbuf_base]
	// check tbuf full to avoid overwritting tbuf [10]
	_mphy16_tx_check_tbuf_full[tbuf_ele_1, sop_only_tbuf_full#]

sop_only_ready_to_transmit#: 

// following .set used to eliminate assembling warning
// .set addr_of_tbuf sop_bd sop_only_paylo offset_mod_8 $w0 $w1

	// dram access takes longer latency, so do it first
	// move payload to tbuf
	_mphy16_tx_move_sop_paylo_to_tbuf[indir_ref, dram_addr, sop_bd, addr_of_tbuf, \
			indiref_base, offset_rpaylo, sop_only_paylo, offset_mod_8, \
			sig_dram_access_1] 

	// setup tcw0_1 base for sop_only case
	move[tcw0_1, TCW0_SOP_ONLY_BASE_LW]			; tcw0 base for SOP_ONLY mpacket
	// set payload_length, payload_offset, and channel fields in tcw0
	_mphy16_tx_set_tcw0_with_paylo_len_offset_chnl[tcw0_1, sop_only_paylo, \
												offset_mod_8, deq_port]

	// check whether we need to save secondary buffer meta data 
	// if didn't read meta data in phase 2, just branch out
	br_bclr[exe_stat_flag, READ_SND_BD_FLAG_BIT, sop_only_save_sbd_meta_done#]
	 
	// medat data for secondary buffer is ready, move them to local memory
	_mphy16_tx_save_sbd_meta_to_lm[] 
	// this port was removed from @txc_rd_snd_bd_vector to prvent any following
	// threads in the same run to access the secondary buffer data when it 
	// becomes active buffer, now secondary meta data saved, put this port back
	alu_shf[@txc_rd_snd_bd_vector, @txc_rd_snd_bd_vector, or, deq_port_bit]

sop_only_save_sbd_meta_done#:

	// now, check whether a valid tx request gotten in the scratch read from the
	// scratch ring in phase 1. If there is a valid tx request, enqueu that 
	// packet to the queue of that port 
.begin 
.reg pkts_in_queue
.reg qentry_lm_base
// following .set used to eliminate assembling warning
.set $dl_meta[0] $dl_meta[1] $dl_meta[2]   
	// check whether there is null requests in scratch ring
	alu[--, $tx_request_lw0, -, 0x0]		; valid tx request
	// no valid tx request, not worst case, can take branchout penalty
	beq[sop_only_no_valid_tx_request#]		; no valid tx request
	
	// set sop meta read flag in exe_stst_flag 
	alu_shf[exe_stat_flag, exe_stat_flag, or, 1, <<READ_SOP_BD_FLAG_BIT] 

	// set local memory index 1 to port queue status info, so ppkts_in_quue
	// can be gotten to determine whether it is necessary to discard this packet
	// due to the queue full
	_mphy16_tx_set_lmindex1_for_enqueue_sop[qentry_lm_base, enq_port, \
											$tx_request_lw0]

    // read SOP meta data, read earlier to reduce latency 
	// read meta data from SRAM by using the BD pointer in tx request
 	_mphy16_tx_read_meta_data_from_sram($dl_meta, sig_mask_2, $tx_request_lw0) 
	 
 
	// check if the queue is full already, theorecically this should not happen,
	// Scheduler wll control the number of packets queued to QM
	alu_shf[pkts_in_queue, PKTS_IN_Q_MASK, and, *l$index1[Q_STAT_OFFSET], >>PKTS_IN_QUEUE_SHFT]	
	alu[--, QUEUE_DEPTH, -, pkts_in_queue]	

	// queue full, not worst case, can take branchout penalty
	beq[sop_only_discard_this_pkt#]	; queue for this port full, discard this packet

	// following stuff can be done in phase 3, but put here for consuming sram
	// access latency, set local memory index 0 to queue entry at queue tail of 
	// that port and enqueue this packet 
	_mphy16_tx_set_lmindex0_and_enqueue_sop[qentry_lm_base, $tx_request_lw0]
.end // pkts_in_queue qentry_lm_base

sop_only_no_valid_tx_request#:
	// end of phase 2
	// swap out to wait for dram and msf acess complete, and wait for my turn
	// also calculate address of TBUF_ELEMENT_CONTROL_V# for tbuf_element for
	// later use to save 2 instruction cycles in ctx_arb defer slots
	// swap out to wait signals specified in sig_mask_2
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask_2]	; set active_ctx_wakeup_events
														; csr
	ctx_arb[--]	, defer[2]								; ctx_arb on sig_mask_2
#if IS_IXPTYPE(__IXP28XX)
		alu_shf[tmp, --, B, tbuf_ele_1, <<ELE_2_TX_CNTRL_ADDR]
#else // #ifdef IXP2800
		alu_shf[tmp, --, B, tbuf_ele_1, <<3]
#endif // #ifdef IXP2800
		alu[addr_of_tx_control, tmp, +, @tcw_vd_base]
	
	// following .io_complete is used to eliminate assembling warning due to the
	// usage of "ACTIVE_CTX_WAKEUP_EVENTS" csr
	.io_completed sig3_next_context
	.io_completed sig_dram_access_1
	.io_completed sig_msf_access_1
	.io_completed sig_sram_read_1
#ifdef DL_SPLIT_META_DATA
	.io_completed sig_sram_read_11
#endif
#ifdef DEBUG_TX_PENDING_LOCKED
	.io_completed sig_sram_write_1 	
#endif //#ifdef DEBUG_TX_PENDING_LOCKED

// end of of phase 2 in sop_noly mpacket path	


// beginning of phase 3	in sop_only mpacket path
	// wake up next thread in the very beginning of phase 3 to reduce wakeup 
	// latency
	local_csr_wr[SAME_ME_SIGNAL, sig3_next_context_gpr]

	// get Tx_MPHY_Status by msf read for use in phase 1
	msf[read, $mphy_stat, tx_mphy_stat_addr,  0, 1], sig_done[sig_msf_access_2]	

    // write transmit contrl word (TCW) to TBUF_ELEMENT_CTRL_V_# corresponding
	// the tbuf_element to initiate transmission
    // Note that we have to adjust transmit port number to base assigned to Packet TX
	alu[$w0, tcw0_1, +, PACKET_TX_PORT_BASE] ; element trnsmit control word
	msf[write, $w0, addr_of_tx_control, 0, 2], sig_done[sig_msf_access_1]

#ifdef COUNTERS
	// update bytes tranmitted 	
	_mphy16_tx_update_bytes_transmitted[tcw0_1, sig_counter_update, $byte_cnt]					
#endif //#ifdef	COUNTERS

	// if didn't read meta data in phase 2, just branch out
	br_bclr[exe_stat_flag, READ_SOP_BD_FLAG_BIT, sop_only_cont_1#] 

	// medat data for sop is ready, move them to local memory
	_mphy16_tx_save_sop_meta_to_lm[enq_port] 

sop_only_cont_1#:
#ifndef COUNTERS
    // swap out to wait for  msf access to complete and my turn
	ctx_arb[sig_msf_access_1, sig_msf_access_2, sig1_next_context], br[start#]
#else // #ifndef COUNTERS
    // swap out to wait for  msf access and sram access to complete and my turn
	ctx_arb[sig_msf_access_1, sig_msf_access_2, sig_counter_update, \
			sig1_next_context], br[start#]
#endif //#ifndef COUNTERS
// end of phase 3 in the sop_only mpacket path

sop_only_discard_this_pkt#:
 	// discard tx request, like no vaid tx request
	// local memory index 1 already set, but it is not used, and will be 
	// overwritten later  
	alu_shf[exe_stat_flag, exe_stat_flag, and~, 1, <<READ_SOP_BD_FLAG_BIT] 
#ifdef CHECK_DISCARD
	alu[@pkts_discarded, @pkts_discarded, +, 1]
#endif	//CHECK_DISCARD	
#ifdef COUNTERS
	_mphy16_tx_update_pkts_dropped[deq_port]	; update packets discarded					
#endif //#ifdef	COUNTERS
	br[sop_only_no_valid_tx_request#]

///////////////////////////////////////////////////////////////////////////////
// branch taken to wait to make sure tbuf elements not overwritten before they
// have been transmitted in sop_only mpacket case, otherwise it also causes
// tx_pending bits in Tx_MPHY_Status msf register locked in 1, in addition to 
// screwing up previos mpackets		
///////////////////////////////////////////////////////////////////////////////
sop_only_tbuf_full#:
	_mphy16_tx_wait_till_tbuf_not_full(tbuf_ele_1, sop_only_ready_to_transmit#)	
		
.end  // tmp sop_bd sop_only_paylo offset_mod_8 addr_of_tbuf dram_addr 
	  // indir_ref sop_only_paylo addr_of_tx_control deq_port_bit $w0 $w1 

// end of sop_only mpacket processing code
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// beginning of not_sop_mpacket processing code
// now still in phase 1
not_sop_mpacket#:

	// setup basic sig_mask_1 for sig2_next_context, sig_scratch_access_1, 
	// sig_scratch_access_2, and sig_msf_access3
	alu[sig_mask_1, --, b, sigmask_phs1_default] ; initialized to sigmask_phs1_default
	// clear the sig_reflect_write from sig_mask_1 first, since most mpacket is
	// not EOP mpacket
	_mphy16_tx_clear_signal(sig_mask_1, sig_reflect_write)

	// initialize tcw0_1 by setting it's channel field to deq_port
	alu[tcw0_1, deq_port, and, CONST_0F]					; set channel of tcw0_1

.begin 
.reg tmp
.reg ab_bd
//.reg ab_flist		; not used now
.reg addr_of_tbuf
.reg dram_addr
.reg indir_ref
.reg addr_of_tx_control
.reg lov_to_next_offset
.reg $w0
.reg $w1
.xfer_order $w0 $w1
.reg $$dw0
.reg $$dw1
.reg $$dw2
.reg $$dw3
.xfer_order $$dw0 $$dw1 $$dw2 $$dw3
	// prepare the deq_port_bit (which is equal to 1 <<deq_port) for later use
	alu[--, deq_port, or, 0]
	alu_shf[deq_port_bit, --, b, 1, <<indirect]		; indirect shift, get deq_port_bit
 	
// following .set used to eliminate assembling warning
.set $dl_meta[0] $dl_meta[1] $dl_meta[2] 
.set $w0 $w1
.set $$dw0 $$dw1 $$dw2 $$dw3
.set lov_to_next_offset

	alu_shf[rmnd_paylo, --, b, *l$index0[ABD_1_OFFSET], >>PAYLO_RMND_LOC]
    ld_field_w_clr[offset_rpaylo, 0011, *l$index0[ABD_1_OFFSET]]

	// check whether to read secondary BD 
	alu[tmp, --, b, *l$index0[ABD_0_OFFSET]]
    br_bclr[tmp, RD_NBD_BIT_LOC, not_sop_sbd_rd_done#]

.begin 
.reg offset_mod_8
    // read secondary buffer meta data from SRAM 
	_mphy16_tx_read_meta_data_from_sram($dl_meta, sig_mask_1, \
													*l$index0[SBD_0_OFFSET]) 
	// reset need to read secondary buffer meta data flag 
	alu_shf[*l$index0[ABD_0_OFFSET], *l$index0[ABD_0_OFFSET], and~, 1, \
			<<RD_NBD_BIT_LOC]
	// set read secondary BD flag in exe_stat_flag for use in phase 2
	alu_shf[exe_stat_flag, exe_stat_flag, or, 1, <<READ_SND_BD_FLAG_BIT]
	// check whether this port needed to be removed from @txc_rd_snd_bd_vector to
	// prvent any following threads in the same run to access the secondary buffer
	// meta data (which is not ready yet in this phase, but will be ready in next 
	// phase) when it becomes active buffer
	// calculate the max bytes to consume before secondary buffer meta data need 
	// to be ready
	alu_shf[tmp, --, b, TBUF_ELE_SIZE, <<ME_THREAD_NUM_SHFT] ; max bytes required
	alu[--, rmnd_paylo, -, tmp]			; compare with how much left
	bgt[not_sop_sbd_rd_done#]
	// remove this port from @txc_rd_snd_bd_vector to prvent any following
	// threads in the same run to access the secondary buffer data (which 
	// is not ready yet) when it becomes active buffer, put this port back
	// when secondary meta data saved in phase 2. 
	alu[@txc_rd_snd_bd_vector, @txc_rd_snd_bd_vector, and~, deq_port_bit]

not_sop_sbd_rd_done#:
  	// get ab_bd, rmnd_paylo, offset_rpaylo, offset_mod_8, ab_flist for active 
	// buffer
    ld_field_w_clr[ab_bd, 0111, *l$index0[ABD_0_OFFSET]] ; get active buffer descriptor
	alu[offset_mod_8, offset_rpaylo, and, CONST_07]
	// flist not used now
 	//alu_shf[ab_flist, FREE_LIST_MASK, and, *l$index0[ABD_0_OFFSET], \
	//		>>FLIST_QUEUE_LOC] 
.begin
.reg lpb_paylo
	// check whether there is some leftover bytes (maximum 8 bytes) from 
	// previous bufferwhich need to be moved to tbuf at the beginning
	// of this mpacket
	alu[tmp, --, b, *l$index1[LP_STAT_OFFSET]]
    br_bclr[tmp, LP_FLAG_LOC, not_sop_no_leftover#]
	// yes, there is leftover from previous buffer in the same packet
	// move leftover bytes to $w0 for latter use, also clear that bit,
	// so later threads in the same run will not repeat the  same thing
    alu_shf[*l$index1[LP_STAT_OFFSET], *l$index1[LP_STAT_OFFSET], and~, 1, \
			<<LP_FLAG_LOC]		; reset the flag
	alu[$w0, --, b, *l$index1[LP_LW0_OFFSET]] ; get leftover from previous buffer
	alu[$w1, --, b, 0]						  ; dummy
	// payload leftover will be put into prepend to eliminate any alignment problem
	// also before saved to local memory, byte alignment is done on leftover bytes
	// so make sure the offset of those leftover bytes is 0
	ld_field_w_clr[lpb_paylo, 0001, *l$index1[LP_STAT_OFFSET]]
	alu_shf[tcw0_1, tcw0_1, or, lpb_paylo, <<TCWO_PREPEND_LEN_LSB_LOC]
	alu_shf[exe_stat_flag, exe_stat_flag, or, 1, \
			<<LEFTOV_FROM_PB_FLAG_BIT]				;set flag

	// check whether current mpacket is eop packet
	// first check whether this is a last buffer in a packet
    ld_field_w_clr[tmp, 0111, *l$index0[SBD_0_OFFSET]]
	alu[--, tmp, -, IX_NULL]
	bne[lp_not_last_buffer#]
	// check whether all remianed payload can fit into one tbuf, leftover already
	// occupies 8 bytes in tbuf 
	alu[tmp, offset_mod_8, +, rmnd_paylo]
	alu[--, TBUF_SIZE_MINUS_8, -, tmp]
	blt[lp_last_bp_not_eop_mpkt#]
	// it is EOP case, set EOP bit in tcw0_1
	alu_shf[tcw0_1, tcw0_1, or, 1, <<TCW0_EOP_BIT_SHFT]
	// set payload_length and payload_offset of tcw0_1 with rmnd_paylo and
	// offset_mod_8
	_mphy16_tx_set_tcw0_with_paylo_len_offset[tcw0_1, rmnd_paylo, offset_mod_8]
	// eop mpacket is eob mpacket too, need to free buffer at phase 3
	alu_shf[exe_stat_flag, exe_stat_flag, or, 1, <<EOB_MPKT_FLAG_BIT]
	br[not_sop_need_to_advance_queue_head#] 
	
	// following case need to consider multiple of 4 restriction
lp_last_bp_not_eop_mpkt#:
	// calculate what max can be transmit from this buffer
	// first total max payload (include lp_paylo) must be multiple of 4
.begin 
.reg alw_abd_paylo_with_m4r	; maximum allowable payload from active buffer
							; with consideration of multiple of 4 bytes

	alu[tmp, TBUF_SIZE_MINUS_8, -, offset_mod_8]
	alu[tmp, tmp, +, lpb_paylo]	;  now offset_mod_8 ahould be 0
	alu[tmp, tmp, and, CONST_FC]	
	alu[alw_abd_paylo_with_m4r, tmp, -, lpb_paylo]
	// update ab_paylo_rmnd and ab_offset_rpaylo in queue entry with 
	// alw_abd_paylo_with_m4r
	_mphy16_tx_update_ab_paylo_rmnd_and_offset_rpaylo[alw_abd_paylo_with_m4r]

	// set payload_length and payload_offset of tcw0_1 with alw_abd_paylo_with_m4r
	// and offset_mod_8
	_mphy16_tx_set_tcw0_with_paylo_len_offset[tcw0_1, alw_abd_paylo_with_m4r, \
											offset_mod_8]
	
	br[not_sop_maintain_allowed_mpkts#]
.end //	 alw_abd_paylo_with_m4r
lp_not_last_buffer#:
	// since not last buffer in a packet , so only check whether it is a eob 
	// mpacket	
	// which rmnd_paylo we can say it is eob mpacket
	// calculate what max can be transmit from this buffer
	// first total max payload (include lp_paylo) must be multiple of 4
.begin
.reg alw_abd_paylo_with_m4r	; maximum allowable payload from active buffer
							; with consideration of multiple of 4 bytes
	alu[tmp, TBUF_SIZE_MINUS_8, -, offset_mod_8]	;  now offset_mod_8 ahould be 0
	alu[tmp, tmp, +, lpb_paylo]	
	alu[tmp, tmp, and, CONST_FC]
	alu[alw_abd_paylo_with_m4r, tmp, -, lpb_paylo]

	// if rmnd_paylo is less than this max allowable bytes, of cours, we are in 
	// eob 
	alu[--, alw_abd_paylo_with_m4r, -, rmnd_paylo]
	blt[lp_not_last_bp_not_eob#]

	// it is eob, set eob bit
	alu_shf[exe_stat_flag, exe_stat_flag, or, 1, <<EOB_MPKT_FLAG_BIT]

.begin
.reg need_to_transmit
.reg allow_to_transmit
	alu[need_to_transmit, lpb_paylo, +, rmnd_paylo]
	alu[allow_to_transmit, need_to_transmit, and, CONST_FC]

	// set payload_length and payload_offset of tcw0_1 with allow_to_transmit
	// and offset_mod_8
	_mphy16_tx_set_tcw0_with_paylo_len_offset[tcw0_1, allow_to_transmit, offset_mod_8]

	// check whether there are some leftover to raed to $$dw0, $$dw1, $$dw2, and
	// $$dw3
	// check whether we have some leftover to next buffer 
	alu[--, need_to_transmit, -, allow_to_transmit]		
	beq[lp_not_last_buffer_no_need_to_save_leftover#]	; it meets multiple of 4 criteria
	// set flag in exe_stat_flag for later use 
	alu_shf[exe_stat_flag, exe_stat_flag, or, 1, <<SAVE_LEFTOV_TO_LM_FLAG_BIT]
	// get true active buffer BD from ab_bd  
	alu_shf[tmp, --, b, ab_bd, <<TRUE_DL_HDL_SHIFT]
    // use following dl_buf_get_data_from_meta macro from dispatch_loop.uc to
	// get dram address at the beginning of the buffer from true buffer handle
	// now it takes one instruction
    dl_buf_get_data_from_meta[dram_addr, tmp]	; get beginning dram addr for buffer  
	
	alu[tmp, offset_rpaylo, +, allow_to_transmit]
	alu[dram_addr, dram_addr, +, tmp]	
	// in some case, the spome of offset_mod_8 + bytes_leftvoer will over 8 bytes 
	// (for example, offset_mod_8=7, and bytes_leftover is 3), so inorder to 
	// cover all cases, read 16 bytes here 
	// add sig_dram_access_1 to sig_mask_1
	dram[read, $$dw0, dram_addr, 0, 2], sig_done[sig_dram_access_1]	
	// add sig_dram_access_1 to sig_mask_1
	_mphy16_tx_set_signal[sig_mask_1, sig_dram_access_1]	; add sig_dram_access_1
	// since we use sig_mask_1 and ACTIVE_CTX_WAKEUP_EVENTS csr, we need to add to 
	// wait for the push signal of dram access
	_mphy16_tx_set_push_signal[sig_mask_1, sig_dram_access_1]	; add push signal

	// now saving the leftover to next buffer offset to LM should be safe, because 
	// lp_flag not set yet, also this port will be removed from 
	// @txc_left_paylo_vector to make sure any thread to access this port before
	// those $$dw0, $$dw1, $$dw2, and $$dw3 is gotten and saved to LM
	alu[@txc_left_paylo_vector, @txc_left_paylo_vector, and~, deq_port_bit]
	alu[lov_to_next_offset, tmp, and, CONST_07]	; 8 bytes boudary due to dram[read..] 
	alu[tmp, rmnd_paylo, -, allow_to_transmit]
	ld_field[*l$index1[LP_STAT_OFFSET], 0001, tmp]	; save leftover length
.end // need_to_transmit allow_to_transmit 
lp_not_last_buffer_no_need_to_save_leftover#:
	_mphy16_tx_secondary_buffer_become_active_buffer[]
	br[not_sop_maintain_allowed_mpkts#]	
lp_not_last_bp_not_eob#:
	
	// update ab_paylo_rmnd and ab_offset_rpaylo in queue entry with 
	// alw_abd_paylo_with_m4r
	_mphy16_tx_update_ab_paylo_rmnd_and_offset_rpaylo[alw_abd_paylo_with_m4r]
	// set payload_length and payload_offset of tcw0_1 with alw_abd_paylo_with_m4r
	// and offset_mod_8
	_mphy16_tx_set_tcw0_with_paylo_len_offset[tcw0_1, alw_abd_paylo_with_m4r, \
											offset_mod_8]
	br[not_sop_maintain_allowed_mpkts#]
.end //	 alw_abd_paylo_with_m4r
.end //  lpb_paylo, lp_offset

// no leftover from previous buffer
 not_sop_no_leftover#:

	// check whether current mpacket is eop packet
	// first check whether this is a last buffer in a packet
    ld_field_w_clr[tmp, 0111, *l$index0[SBD_0_OFFSET]]
	alu[--, tmp, -, IX_NULL]
	bne[no_lp_not_last_buffer#]
	// check whether all remianed payload can fit into one tbuf, leftover already
	// occupies 8 bytes 
	alu[tmp, offset_mod_8, +, rmnd_paylo]

	alu[--, TBUF_ELE_SIZE, -, tmp]
	blt[no_lp_last_bp_not_eop_mpkt#]
	// it is EOP case, set EOP bit in tcw0_1
	alu_shf[tcw0_1, tcw0_1, or, 1, <<TCW0_EOP_BIT_SHFT]
	// set payload_length and payload_offset of tcw0_1 with rmnd_paylo
	// and offset_mod_8
	_mphy16_tx_set_tcw0_with_paylo_len_offset[tcw0_1, rmnd_paylo, offset_mod_8]
	alu_shf[exe_stat_flag, exe_stat_flag, or, 1, <<EOB_MPKT_FLAG_BIT]
	br[not_sop_need_to_advance_queue_head#] 
	
	// following case need to consider multiple of 4 restriction
no_lp_last_bp_not_eop_mpkt#:
	// calculate what max can be transmit from this buffer
	// first total max payload (include lp_paylo) must be multiple of 4
.begin 
.reg alw_abd_paylo_with_m4r	; maximum allowable payload from active buffer
							; with consideration of multiple of 4 bytes
	alu[tmp, TBUF_ELE_SIZE, -, offset_mod_8]
	alu[alw_abd_paylo_with_m4r, tmp, and, CONST_FC]

	// update ab_paylo_rmnd and ab_offset_rpaylo in queue entry with 
	// alw_abd_paylo_with_m4r
	_mphy16_tx_update_ab_paylo_rmnd_and_offset_rpaylo[alw_abd_paylo_with_m4r]
	// set payload_length and payload_offset of tcw0_1 with alw_abd_paylo_with_m4r
	// and offset_mod_8
	_mphy16_tx_set_tcw0_with_paylo_len_offset[tcw0_1, alw_abd_paylo_with_m4r, \
											offset_mod_8]
	
	br[not_sop_maintain_allowed_mpkts#]
.end //	 alw_abd_paylo_with_m4r
no_lp_not_last_buffer#:
	// check whether is EOB
	// which rmnd_paylo we can say it is eob
	// calculate what max can be transmit from this buffer
	// first total max payload (include lp_paylo) must be multiple of 4
.begin
.reg alw_abd_paylo_with_m4r	; maximum allowable payload from active buffer
							; with consideration of multiple of 4 bytes
	alu[tmp, TBUF_ELE_SIZE, -, offset_mod_8]
	alu[alw_abd_paylo_with_m4r, tmp, and, CONST_FC]

	// if rmnd_paylo is less than this max allowable bytes, of cours, we are in 
	// eob 
	alu[--, alw_abd_paylo_with_m4r, -, rmnd_paylo]
	blt[no_lp_not_last_bp_not_eob#]

	// it is eob, set eob bit
	alu_shf[exe_stat_flag, exe_stat_flag, or, 1, <<EOB_MPKT_FLAG_BIT]
.begin
.reg need_to_transmit   
.reg allow_to_transmit 
	alu[need_to_transmit, --, b, rmnd_paylo]
	alu[allow_to_transmit, need_to_transmit, and, CONST_FC]

	// check need_to_transmit less than 4
	alu[--, need_to_transmit, -, 4]
	bgt[no_lp_not_last_buffer_something_to_transmit#]
	// if it is the case, then set skip bit in tcw0_1
	alu_shf[tcw0_1, tcw0_1, or, 1, <<TCW0_SKIP_BIT_SHFT]	; set skip bit
	// set flag in exe_stat_flag for later use 
	alu_shf[exe_stat_flag, exe_stat_flag, or, 1, <<SKIP_TRANSMIT_FLAG_BIT]
no_lp_not_last_buffer_something_to_transmit#:
	// set payload_length and payload_offset of tcw0_1 with allow_to_transmit
	// and offset_mod_8
	_mphy16_tx_set_tcw0_with_paylo_len_offset[tcw0_1, allow_to_transmit, offset_mod_8]
	// check whether there are some leftover to raed to $$dw0, $$dw1, $$dw2, and 
	// $$dw3
	alu[--, need_to_transmit, -, allow_to_transmit]		
	beq[no_lp_not_last_buffer_no_need_to_save_leftover#] ; it meets multiple of 4 criteria
	// set flag in exe_stat_flag for later use 
	alu_shf[exe_stat_flag, exe_stat_flag, or, 1, <<SAVE_LEFTOV_TO_LM_FLAG_BIT]
	// get true active buffer BD from ab_bd  
	alu_shf[tmp, --, b, ab_bd, <<TRUE_DL_HDL_SHIFT]
    // use following dl_buf_get_data_from_meta macro from dispatch_loop.uc to
	// get dram address at the beginning of the buffer from true buffer handle
	// now it takes one instruction
    dl_buf_get_data_from_meta[dram_addr, tmp]
	
	alu[tmp, offset_rpaylo, +, allow_to_transmit]
	alu[dram_addr, dram_addr, +, tmp]
	// in some case, the spome of offset_mod_8 + bytes_leftvoer will over 8 bytes 
	// (for example, offset_mod_8=7, and bytes_leftover is 3), so inorder to 
	// cover all cases, read 16 bytes here 
	dram[read, $$dw0, dram_addr, 0, 2], sig_done[sig_dram_access_1]
	// add sig_dram_access_1 to sig_mask_1
	_mphy16_tx_set_signal[sig_mask_1, sig_dram_access_1]	; add sig_dram_access_1
	// since we use sig_mask_1 and ACTIVE_CTX_WAKEUP_EVENTS csr, we need to add to 
	// wait for the push signal of dram access
	_mphy16_tx_set_push_signal[sig_mask_1, sig_dram_access_1]	; add push signal
	// now saving the leftover to next buffer offset to LM should be safe, because 
	// lp_flag not set yet, also this port will be removed from 
	// @txc_left_paylo_vector to make sure any thread to access this port before
	// those $$dw0, $$dw1, $$dw2, and $$dw3 is gotten and saved to LM
	alu[@txc_left_paylo_vector, @txc_left_paylo_vector, and~, deq_port_bit]
	alu[lov_to_next_offset, tmp, and, CONST_07]	; 8 bytes boudary due ro dram[read..] 
	alu[tmp, rmnd_paylo, -, allow_to_transmit]
	ld_field[*l$index1[LP_STAT_OFFSET], 0001, tmp]	; save leftover length
.end // allow_to_transmit need_to_transmit
no_lp_not_last_buffer_no_need_to_save_leftover#:
	_mphy16_tx_secondary_buffer_become_active_buffer[]
	br[not_sop_maintain_allowed_mpkts#]	
no_lp_not_last_bp_not_eob#:	
	// update ab_paylo_rmnd and ab_offset_rpaylo in queue entry with 
	// alw_abd_paylo_with_m4r
	_mphy16_tx_update_ab_paylo_rmnd_and_offset_rpaylo[alw_abd_paylo_with_m4r]
	// set payload_length and payload_offset of tcw0_1 with alw_abd_paylo_with_m4r
	// and offset_mod_8
	_mphy16_tx_set_tcw0_with_paylo_len_offset[tcw0_1, alw_abd_paylo_with_m4r, \
											offset_mod_8]
.end //	 alw_abd_paylo_with_m4r
	br[not_sop_maintain_allowed_mpkts#]	   

	// advance queue header for EOP mpacket 
not_sop_need_to_advance_queue_head#:
	// even the deq_port_bit wil be calculated in _mphy16_tx_advance_queue_head[] 
	// macro gaian, but it is the same as original deq_port_bit 
	_mphy16_tx_advance_queue_head[deq_port_bit, deq_port]
	// notify scheduler that one packet has been transmitted from this port 
	_mphy16_tx_notify_scheduler[deq_port, $txed_port]
	_mphy16_tx_set_signal[sig_mask_1, sig_reflect_write]	; add sig_reflect_write

not_sop_maintain_allowed_mpkts#:
	// if one mpkt will be transmitted to the port, then need to decrease
	// allowed_mpkts
	br_bset[exe_stat_flag, SKIP_TRANSMIT_FLAG_BIT, not_sop_maintain_al_mpkts_done#]
	// decrease the allowed_mpkts of the port by one, if it become zero, 
	// remove this port from @mohy_stat_fc_vector
	_mphy16_tx_decrement_allowed_mpkts_in_this_port[deq_port_bit]
not_sop_maintain_al_mpkts_done#:

	// swap out to wait signals specified in sig_mask_1
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask_1]	; set csr													; csr
	ctx_arb[--]	, defer[2]								; ctx_arb on sig_mask_1
	//use defer slot to initialize the sig_mask_2 for phase 2 for not_sop_mpkt 
		// initialized sig_mask_2 to sig3_next_context
		alu[sig_mask_2, --, b, 0]		; reset
		_mphy16_tx_set_signal[sig_mask_2, sig3_next_context]	; add sig3_next_context

	// following .io_complete is used to eliminate assembling warning due to the
	// usage of "ACTIVE_CTX_WAKEUP_EVENTS" csr
	.io_completed sig2_next_context
	.io_completed sig_scratch_access_1
	.io_completed sig_scratch_access_2
	.io_completed sig_msf_access_3
	.io_completed sig_sram_read_1
#ifdef DL_SPLIT_META_DATA
	.io_completed sig_sram_read_11
#endif
	.io_completed sig_dram_access_1
	.io_completed sig_reflect_write
.end // offset_mod_8

// end of of phase 1 with not_sop mpkt ready for tx

// beginning of phase 2 with not_sop mpkt ready for tx	
	// wake up next thread in beginning of phase 2 to reduce wakeup latency
	local_csr_wr[SAME_ME_SIGNAL, sig2_next_context_gpr] 

	// $avail_tbuf_ele read in phase 1 is ready now, get next available 
	// tbuf element (tbuf_ele) and corresponding address (addr_of_tbuf) from it
	_mphy16_tx_get_tbuf_ele_and_addr[tbuf_ele_1, addr_of_tbuf, \
									$avail_tbuf_ele, tbuf_base]

	// check tbuf full to avoid overwritting tbuf [10]
	_mphy16_tx_check_tbuf_full(tbuf_ele_1, not_sop_tbuf_full#)

not_sop_ready_to_transmit#: 

// following .set used to eliminate assembling warning
//.set addr_of_tbuf ab_bd lov_to_next_offset $w0 $w1 $$dw0 $$dw1 $$dw2

	// check whether need to move leftover ($w0) to the beginning of tbuf
	br_bclr[exe_stat_flag, LEFTOV_FROM_PB_FLAG_BIT, ph2_move_lp_to_tbuf_done#]
	msf[write, $w0, addr_of_tbuf, 0, 1], sig_done[sig_msf_access_1]
	_mphy16_tx_set_signal[sig_mask_2, sig_msf_access_1 ]	; add sig_msf_access_1 

	alu[addr_of_tbuf, addr_of_tbuf, +, 8]
ph2_move_lp_to_tbuf_done#:

	// check whether need to move payload from active buffer to tbuf
	br_bset[exe_stat_flag, SKIP_TRANSMIT_FLAG_BIT, ph2_move_paylo_done#]
	// payload offset and payload length can be retrived from tcw0_1  
	_mphy16_tx_move_not_sop_paylo_to_tbuf[indir_ref, dram_addr, ab_bd, addr_of_tbuf, \
							indiref_base, offset_rpaylo, tcw0_1, sig_dram_access_1] 
	// add sig_dram_access_1 to sig_mask_2
	_mphy16_tx_set_signal[sig_mask_2, sig_dram_access_1]	; add sig_dram_access_1
	// since we use sig_mask_2 and ACTIVE_CTX_WAKEUP_EVENTS csr, we need to add to 
	// wait for the push signal of dram access
	_mphy16_tx_set_push_signal[sig_mask_2, sig_dram_access_1]	; add push signal
ph2_move_paylo_done#:

	// check whether we have leftover which needed to be to saved in LM
	br_bclr[exe_stat_flag, SAVE_LEFTOV_TO_LM_FLAG_BIT, ph2_save_leftov_to_lm_done#]
	// $$dw0, $$dw1, $$dw2, and $$dw3 ready now, do necessary bytes alignment with
	// lov_to_next_offset, so the leftover will be realigned in first two longword
	// without any offset to the first byte of the first long word
.begin 
.reg r0
	_mphy16_tx_leftover_bytes_alignment[lov_to_next_offset, $$dw0, $$dw1, $$dw2, \
										$$dw3, r0]
	alu[*l$index1[LP_LW0_OFFSET], --, b, r0]	; save leftover to local memory
.end // r0	
	// set flag in queue status
    alu_shf[*l$index1[LP_STAT_OFFSET], *l$index1[LP_STAT_OFFSET], or, 1, \
			<<LP_FLAG_LOC]											; set flag 
   	// data saved, port ready to be transmitted again, put this port back to 
	// @txc_left_paylo_vector 
	alu[@txc_left_paylo_vector, @txc_left_paylo_vector, or, deq_port_bit]
ph2_save_leftov_to_lm_done#:

	// check whether we need to save secondary buffer meta data 
	// if didn't read meta data in phase 2, just branch out
	br_bclr[exe_stat_flag, READ_SND_BD_FLAG_BIT, not_sop_save_sbd_meta_done#]

	// medat data for secondary buffer is ready, move them to local memory
	_mphy16_tx_save_sbd_meta_to_lm[] 
	// this port was removed from @txc_rd_snd_bd_vector to prvent any following
	// threads in the same run to access the secondary buffer data when it 
	// becomes active buffer, now secondary meta data saved, put this port back
	alu_shf[@txc_rd_snd_bd_vector, @txc_rd_snd_bd_vector, or, deq_port_bit]
not_sop_save_sbd_meta_done#:

	// now, check whether a valid tx request gotten in the scratch read from the
	// scratch ring in phase 1. If there is a valid tx request, enqueu that 
	// packet to the queue of that port 
.begin 
.reg pkts_in_queue
.reg qentry_lm_base
// following .set used to eliminate assembling warning
.set $dl_meta[0] $dl_meta[1] $dl_meta[2] enq_port   
	// check whether there is null requests in scratch ring
	alu[--, $tx_request_lw0, -, 0x0]		; valid tx request
	// no valid tx request, not worst case, can take branchout penalty
	beq[not_sop_no_valid_tx_request#]		; no valid tx request
	
	// set sop meta read flag in exe_stst_flag 
	alu_shf[exe_stat_flag, exe_stat_flag, or, 1, <<READ_SOP_BD_FLAG_BIT] 

	// set local memory index 1 to port queue status info, so ppkts_in_quue
	// can be gotten to determine whether it is necessary to discard this packet
	// due to the queue full
	_mphy16_tx_set_lmindex1_for_enqueue_sop[qentry_lm_base, enq_port, \
											$tx_request_lw0]

    // read SOP meta data, read earlier to reduce latency 
	// read meta data from SRAM by using the BD pointer in tx request
  	_mphy16_tx_read_meta_data_from_sram($dl_meta, sig_mask_2, $tx_request_lw0) 
 
	// check if the queue is full already, theorecically this should not happen,
	// Scheduler wll control the number of packets queued to QM
	alu_shf[pkts_in_queue, PKTS_IN_Q_MASK, and, *l$index1[Q_STAT_OFFSET], \
			>>PKTS_IN_QUEUE_SHFT]
	alu[--, QUEUE_DEPTH, -, pkts_in_queue]	

	// queue full, not worst case, can take branchout penalty
	beq[not_sop_discard_this_pkt#]	; queue for this port full, discard this packet

	// following stuff can be done in phase 3, but put here for consuming sram
	// access latency, set local memory index 0 to queue entry at queue tail of 
	// that port and enqueue this packet 
	_mphy16_tx_set_lmindex0_and_enqueue_sop[qentry_lm_base, $tx_request_lw0]
.end // pkts_in_queue qentry_lm_base

not_sop_no_valid_tx_request#:
	// end of phase 2
	// swap out to wait for dram and msf acess complete, and wait for my turn
	// also calculate address of TBUF_ELEMENT_CONTROL_V# for tbuf_element for
	// later use to save 2 instruction cycles in ctx_arb defer slots
	// swap out to wait signals specified in sig_mask_2
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask_2]	; set csr
	ctx_arb[--]	, defer[2]								; ctx_arb on sig_mask_2
#if IS_IXPTYPE(__IXP28XX)
		alu_shf[tmp, --, B, tbuf_ele_1, <<ELE_2_TX_CNTRL_ADDR]
#else // #ifdef IXP2800
		alu_shf[tmp, --, B, tbuf_ele_1, <<3]
#endif // #ifdef IXP2800
		alu[addr_of_tx_control, tmp, +, @tcw_vd_base]
	
	// following .io_complete is used to eliminate assembling warning due to the
	// usage of "ACTIVE_CTX_WAKEUP_EVENTS" csr
	.io_completed sig3_next_context
	.io_completed sig_dram_access_1
	.io_completed sig_msf_access_1
	.io_completed sig_sram_read_1
#ifdef DL_SPLIT_META_DATA
	.io_completed sig_sram_read_11
#endif
#ifdef DEBUG_TX_PENDING_LOCKED
	.io_completed sig_sram_write_1 	
#endif //#ifdef DEBUG_TX_PENDING_LOCKED

// end of of phase 2 in sop_noly mpacket path	


// beginning of phase 3	in not_sop mpacket path
	// wake up next thread in the very beginning of phase 3 to reduce wakeup 
	// latency
	local_csr_wr[SAME_ME_SIGNAL, sig3_next_context_gpr]

	// get Tx_MPHY_Status by msf read for use in phase 1
	msf[read, $mphy_stat, tx_mphy_stat_addr,  0, 1], sig_done[sig_msf_access_2]	

    // write transmit contrl word (TCW) to TBUF_ELEMENT_CTRL_V_# corresponding
	// the tbuf_element to initiate transmission
    // Note that we have to adjust transmit port number to base assigned to Packet TX
	alu[$w0, tcw0_1, +, PACKET_TX_PORT_BASE] ; element trnsmit control word
	// workaround for HW-bug#1249, copy channel # into tcw1 (reserved in PRM)
	alu_shf[$w1, --, b, deq_port]
	msf[write, $w0, addr_of_tx_control, 0, 2], sig_done[sig_msf_access_1]

#ifdef COUNTERS
	// update bytes tranmitted 	
	_mphy16_tx_update_bytes_transmitted[tcw0_1, sig_counter_update, $byte_cnt]			
#endif //#ifdef	COUNTERS

	// check whether need tio free buffer
	br_bclr[exe_stat_flag, EOB_MPKT_FLAG_BIT, not_sop_free_buffer_done#] 

	// free buffer handle  
	_mphy16_tx_free_buffer(ab_bd)			; free buffer

not_sop_free_buffer_done#:


	// if didn't read meta data in phase 2, just branch out
	br_bclr[exe_stat_flag, READ_SOP_BD_FLAG_BIT, not_sop_save_sop_meta_done#] 
	// medat data for sop is ready, move them to local memory
	_mphy16_tx_save_sop_meta_to_lm[enq_port] 
not_sop_save_sop_meta_done#:
#ifndef COUNTERS
    // swap out to wait for  msf access to complete and my turn
	ctx_arb[sig_msf_access_1, sig_msf_access_2, sig1_next_context], br[start#]
#else // #ifndef COUNTERS
    // swap out to wait for  msf access and sram access to complete and my turn
	ctx_arb[sig_msf_access_1, sig_msf_access_2, sig_counter_update, \
			sig1_next_context], br[start#]
#endif //#ifndef COUNTERS
// end of phase 3 in the not_sop mpacket path


not_sop_discard_this_pkt#:
 	// discard tx request, like no vaid tx request
	// local memory index 1 already set, but it is not used, and will be 
	// overwritten later  
	alu_shf[exe_stat_flag, exe_stat_flag, and~, 1, <<READ_SOP_BD_FLAG_BIT]
#ifdef CHECK_DISCARD
	alu[@pkts_discarded, @pkts_discarded, +, 1]
#endif	//CHECK_DISCARD	 
#ifdef COUNTERS
	_mphy16_tx_update_pkts_dropped[deq_port]	; update packets discarded					
#endif //#ifdef	COUNTERS	
	br[not_sop_no_valid_tx_request#]

///////////////////////////////////////////////////////////////////////////////
// branch taken to wait to make sure tbuf elements not overwritten before they
// have been transmitted in not_sop mpacket case, otherwise it also causes
// tx_pending bits in Tx_MPHY_Status msf register locked in 1, in addition to 
// screwing up previos mpackets		
///////////////////////////////////////////////////////////////////////////////
not_sop_tbuf_full#:	

	_mphy16_tx_wait_till_tbuf_not_full(tbuf_ele_1, not_sop_ready_to_transmit#)	

.end  // tmp offset_mod_8 addr_of_tbuf dram_addr indir_ref 
      // addr_of_tx_control deq_port_bit $w0 $w1 $$dw0 $$Dw1 $$dw2 $$dw3
// end of not_sop mpacket processing
///////////////////////////////////////////////////////////////////////////////

#endm // end of macro mphy16_tx[]
///////////////////////////////////////////////////////////////////////////////
	     
														
///////////////////////////////////////////////////////////////////////////////
//
//	Main entry point -- Code begins execution here.
//
///////////////////////////////////////////////////////////////////////////////

main#:

	// initialize the microblock
    mphy16_tx_init[]

	// packet processing block
	mphy16_tx[]

///////////////////////////////////////////////////////////////////////////////

#endif	// __MPHY16_TX_UC__



