///////////////////////////////////////////////////////////////////////////////
//                                                                     
//                  I N T E L   P R O P R I E T A R Y                   
//                                                                      
//     COPYRIGHT (c)  2001-2003 BY  INTEL  CORPORATION.  ALL RIGHTS          
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//     THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                      
//                        INTEL  CORPORATION                            
//                                                                     
//                     2200 MISSION COLLEGE BLVD                        
//                                                                      
//               SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                      
///////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: pkt_tx_16p.uc
//
//      Purpose: Mulitport (up to 16 ports) PACKET transmitter for Egress 
//               IXP24XX/IXP28XX
//
///////////////////////////////////////////////////////////////////////////////

#ifndef	__PKT_TX_P16_UC__
#define	__PKT_TX_P16_UC__

///////////////////////////////////////////////////////////////////////////////
// 
// This PACKET transmitter (TX) microblock can support up to 16 ports in
// IXP24XX/IXP28XX with 128 byte tbuf element.  The basic unit of transmission 
// is an mpacket, the maximum size of an mpacket is the size of a tbuf element. 
// PACKRET TX receives a transmit request for each packet. If the size of the
// packet is too large to fit into one mpacket, PACKET TX segments this packet 
// into several mpackets. 
//
// In IXP24XX/IXP28XX, the Tx_MPHY_Status/Tx_Multiple_Port_Status_0 MSF 
// register provides the Tx FIFO status of port 0 to port 15 for the software 
// to do the transimission flow control to prevent packet overwritting on the 
// ports. IXP24XX uses SPI-3 interface, there are two bits (Tx_Pending[x] and 
// Tx_Status[x]) for each port (x) in Tx_MPHY_Status register. When 
// Tx_Pending[x]=1, it indicates that there are pending transmit activity for 
// that port, then Tx_Status[x] is indeterminant. When Tx_Pending[x]=0 and 
// Tx_Status[x]=1, the port is ready to accept at least "n" mpackets of data.
// This size of "n" depends on the characteristics of the port PHY and size of 
// the tbuf element. Based on SPI-4, IXP28XX also have two bits assigned to 
// ecah port too, the case with both MSB and LSB as 1 is reserved for framing
// or to indicate a disabled status link, the case with MSB as 1 and LSB as 0
// is "STATISFIED" status which indicates that the corresponding port's FIFO 
// is almost full; the case with MSB as 0 and LSB as 1 is "HUNGRY" status 
// which means that up to MaxBurst2 16-byte blocks or the remainder of what 
// was previously granted (whichever is greater) may be sent to the 
// corresponding port until the next status update; the case with both MSB and 
// LSB as 0 is "STARVING" status which indicates that buffer underflow is 
// imminent in the corresponding PHY port and transfers for up to MaxBurst1 
// 16-byte blocks may be sent to the corresponding port until next status 
// update. The values of MaxBurst1 and MaxBurst2 are the characteristics of 
// the interface PHY and usually are programable. 
//
// Maximum 16 (QUEUE_DEPTH) packets are queued in the local memory of Packet 
// Tx ME for each port to respond to the change of 
// TX_MPHY_Status/Tx_Multiple_Port_Status_0 immediately.  Since one ME is not
// fast enough to finish the tasks for all ports, two ME are used in this 
// solution, they are identified as PACKET_TX_FIRST_ME and PACKET_TX_SECOND_ME
// in the code and  defined in "Preprocessor definitions" of "Assembler" of 
// "Build Settings" of the project.
//
// Each ME controls up to 8 ports as specified by 
// PORTS_CTRLED_BY_FIRST_PKT_TX_ME for the first Packet TX ME and 
// PORTS_CTRLED_BY_SECOND_PKT_TX_ME for the second Packet ME. These 
// PORTS_CTRLED_BY_FIRST_PKT_TX_ME and PORTS_CTRLED_BY_SECOND_PKT_TX_ME have 
// been predefined for following TX_PHY_MODE: MPHY_16 (IXP24XX with 16 ports),
// SPI_4_10PORTS (IXP28XX with 10 ports), and SPI_4_16PORTS (IXP28XX with 16 
// ports) in the code, when you build a project, if you select one of the 
// above-mentioned TX_PHY_MODE, you don't need to define
 // PORTS_CTRLED_BY_FIRST_PKT_TX_ME and PORTS_CTRLED_BY_SECOND_PKT_TX_ME.
// These two ME runs independently,  but the threads inside each ME uses 
// folding mechanism to run in strict order of two phases. TBUF elements are 
// shared by both MEs, the  next available TBUF element is kept in one longword
// in one scratch memory, the scratch memory atomic operation is used to get  
// the next available TBUF element.   
//
// The microblock keeps account of how many packets transmitted for each port, 
// and update pkts_txd count in Egress Scheduler through a reflect write when 
// it finishs the transmission of one packet, so Egress Scheduler will not 
// overflood it's downstream microblocks. 
//
// Following are the assumptions used to develop this Packet TX microblock:
// 1. The meta data at SRAM and the payload at DRAM are correct, Packet TX will 
//    not do any sanity check in the size and order of the buffers in any 
//    packet.
// 2. The payload in DRAM will be contiguous from the offset to the end of a 
//    buffer in a packet.
// 3. The offset for all buffers in a packet except SOP buffer is 0.
//
// There are three issues need to be addressed in the Packet TX microblock. The
// first two are imposed by IXP24XX/IXP28XX hardware, the third one keeps this 
// Packet TX microblock capable to handle the packets received in diffrent 
// receiving interface (POS, ATM, Ethernet) in ingress microblocks. They are
// 1. DRAM[tbuf_wr,...] instruction, which is used to move packet data from 
//    DRAM to TBUF, always access DRAM and TBUF on 8-bytes word boundary.
// 2. SPI-3 requires that all data transfer except the last fragment (with EOP)
//    be multiple of device bus width (in MPHY_16 mode, this is 4 bytes). SPI-4 
//    requires that all data transfer except the last fragment (with EOP) be
//    multiple of 16 byte block.
// 3. The payload data offset in the first buffer of a packet can be any value. 
// 
// In order to address the abovementioned three issues, Packet TX transmits 
// the maximum allowable payload bytes for every non-eop mpacket by making sure 
// that the sum of the prepend length and payload length of a non-eop mpacket
// is an integral multiple of 4/16 bytes for IXP24XX/IXP28XX. If necessary, 
// maximum of 3/15 bytes at the  end of one buffer is saved in the local memory 
// and transmit with the payload of the next buffer of the same packet in the 
// next mpacket.
//
// This Packet TX microblock keeps current buffer (which the microblock is 
// processing) descriptor info and the next buffer (buffer descriptor for the 
// next buffer after the current buffer in the packet, it is in the software 
// next pointer field of meta data of the current buffer) descriptor info in 
// local memory, this microblock will read the necessary info (buffer size, 
// data offset, freelist, and the software next pointer) of the next buffer 
// from sram in parallel with it's processing of the current buffer. When
// the processing of the current buffer finishes, the next buffer will become 
// the current buffer, and continue until all buffers in the packet are 
// processed.
//
// Mandatory assembling options for this Packet Transmit microblock:
//
// A. THIS_ME:
//		in the build option of any ME of this Packet TX microblock, either 
//		"THIS_ME=PACKET_TX_FIRST_ME" or "THIS_ME=PACKET_TX_SECOND_ME" must be
//		specified to identify which ports that ME serves. PACKET_TX_FIRST_ME
//		is in charge of the initialization of the shared resources (such as MSF
//		interface and scratch memory holding next available TBUF element), so 
//		if only one ME is used, "THIS_ME=PACKET_TX_FIRST_ME" must be specified
//		for that ME
//
// B. SCHEDULER_ME or DISABLE_TX2SCHED_FEEDBACK:
//		when the feedback of how many packets transmitted for each port is 
//		required by Egress Scheduler, the ME number of Egress Scheduler must be 
//		specified through "SCHEDULER_ME" option, such as "SCHEDULER_ME=0x13" 
//		when Egress Scheduler ME is in ME 3 of cluster 1. If the feedback is
//      not required, then "DISABLE_TX2SCHED_FEEDBACK" must be specified, and
//		"SCHEDULER_ME" option is not required, the optional assembling option 
//		"CHECK_PORT_QUEUE_OVERFLOW" (see following) can be used to avoid 
//		overflooding packet queue information in the local memory at this case
//
//
// Optional assembling options available for this Packet Transmit microblock:
//
// A. ADD_L2_HEADER: 
//		this option enables this microblock to add layer 2 header, it combines
//      with "POS_TX" option to add PPP header before payload; it combines  
// 		with "ETHERNET_TX" option to add Ethernet header before payload
//
// B. POS_TX:
//		this option combines with "ADD_L2_HEADER" option enable this microblock
//		to add 2 bytes PPP header before payload 
//
// C. ETHERNET_TX:
//		this option combines with "ADD_L2_HEADER" option enable this microblock
//		to add 14 bytes Ethernet header before payload
//
// D. _DEBUG_COUNTERS_:
//		this option enable this microblock to accumulate folowing three 
//		statitistics in absolute GPRs to provide information for debugging:
//			a. number of tx requests this microblock received
//			b. number of packets this microblock tranmitted 
//			c. number of buffers this microblock freed
//
// E. UNIT_TEST:
//		this option can be used in the unit tests where the buffer is reused,
//		so some dummy instruction is executed instead of actually freeing the
//		buffer
//  
// F. CHECK_PORT_QUEUE_OVERFLOW:
// 		this option enales this microblock to discard extra tx_requests which
//		will overflood the packet queue in the local memory, it should be used
//		when the egress scheduler is not used in the pipeline. this option is 
//		not required when the egress scheduler, which controls in  egress how 
//		many tx_request sent for a specific port to this microblock, is used  
//
// G. STRIP_4BYTES_CRC:
//		this option enables this microblock to strip last 4 bytes CRC from 
//		payload, it can be used when the Ethernet media chip doesn't strip 
//		four bytes CRC when it passes the packet to IXP, even it blocks out 
//		packets with CRC error
//
//
///////////////////////////////////////////////////////////////////////////////

// following definition selects sram transfer registers to be used with 
// dl_meta.uc dispatch loop

#define DL_META_DATA_IN_SXFER

//	Include Files

// include stdmac.uc in IXPblocks Portable library

#include <stdmac.uc>

// include localmem.uc in IXPblocks Portable library for read/write local
// memory

#include <localmem.uc>

// include header file with the global definition of common constants which
//  are shared by all microblocks

#include <dl_system.h>

// include global dispatch loop macros

#include <dispatch_loop.uc>

#include <hardware.h>

// include the Packet TX header file with microblock specific definitions  

#include <pkt_tx_16p.h> 

// include initializatiuon macros within this microblock

#include <pkt_tx_16p_init.uc>

// include local utilization macros within this microblock

#include <pkt_tx_16p_util.uc>

///////////////////////////////////////////////////////////////////////////////
//
//	Global Registers
//  Absolute GPR variables:

.reg	@ep						; Empty ports. Each port occupies 1 bit, when
								; set, it means that there is no packets 
								; queued for this port, initialized to 0xFFFF
.reg	@bfcp					; Ports blocked by flow control. Each port 
								; occupies 1 bit, when set, it means that this
								; port is blocked by PHY status, waiting port 
								; ready during initialization
.reg	@vap					; Virtual available ports
.reg	@rp						; Ports ready to process packet in tx request
								; immediately
.reg	@busy_port				; mask for ports which is not ready due to
 								; PHY status
.reg	@vpp					; Virtual port queue to pick next 
								; mpackets/mpacket for next iteration.

.reg	@blocked_by_read_next_buf ; mask to stop mpacket be transmitted for  
								; specific port when the next buffer meta data
								; is not ready in local memory 

.reg	@blocked_by_read_leftover ; mask to stop mpacket be transmitted for  
								; specific port when leftover bytes is not 
								; not ready in local memory 
.reg 	@port_mask				; ports serviced by this ME
.reg 	@default_sigmask_1		; default signal mask for phase 1
.reg 	@default_sigmask_2		; default signal mask for phase 2

.reg    @rr_vector				; round-robin vector

// following absolute GPR variables are preassigned in initialization to save 
// instruction cycles in Packet TX process

.reg 	@fc_check_pattern		; pattern to check port status in IXP28XX, 
								; initialized to 0xAAAAAAAA, i.e. 
								; 1010101010101010

#ifdef	_DEBUG_COUNTERS_

//---------------------------------------------------------------------
// Declare some counters to ease debugging.
//---------------------------------------------------------------------
.reg @pkt_tx_16p_tx_request_rxed	// number of tx request received
.reg @pkt_tx_16p_pkt_txed			// number of packets tranmitted 
.reg @pkt_tx_16p_buffer_freed		// number of buffer freed 
.reg @pkt_tx_16p_mpkts_txed			// number of m-packets transmitted

#endif	// #ifdef	_DEBUG_COUNTERS_

// context relative GPR variables:

.reg	exe_stat_flag			; thread execution status flags 
.reg	thread_specific_flag	; thread specific flags
								; buffer
.reg	sig_mask_1				; signal mask in phase 1
.reg	sig_mask_2				; signal mask in phase 2

// following context relative GPR variables are preassigned in initialization 
// to save instruction cycles in Packet TX process

.reg	sig_next_context_gpr	; value to write to SAME_ME_SIGNAL csr to wake 
								; up next thread

.reg	tbuf_scr_addr			; scratch memory address where the next 
								; available tbuf element kept and incremented
								; by scratch atomic operation
.reg	queue_stat_base			; starting local memory address for queue 
								; status info
.reg	tx_mphy_stat_addr		; address of MSF TX port status register
.reg	tx_seq_addr				; address of MSF Tx Sequence register

.reg	cap_reflect_addr_base	; base to use cap calculating address mode to 
								; do reflect write

// context relative sram xfer registers
// following context relative sram registers are used to keep values gotten in 
// I/O intructions

.reg $tx_request				; tx request in sram xfer reg
.reg $mphy_fc_stat				; status of PHY Tx FIFO status, read from
								; from Tx_MPHY_Status MSF (IXP24XX) or 
								; Tx_Multiple_Port_Status (IXP28XX) msf reg

.reg $mpkts_sent				; mpackets actually transmitted, read from
								; Tx_Sequence_0	MSF register
 
///////////////////////////////////////////////////////////////////////////////
//
// pkt_tx_16p()
//
// Description: 16 ports packet tx dispatch loop
//
// Outputs:							
//		None 
//
// Inputs:
//		None 
//
// Constants:
//		None 
//
//
// Size:
//		  instructions
//
///////////////////////////////////////////////////////////////////////////////

#macro pkt_tx_16p()

.begin

	.reg	tbuf_ele_1		; tbuf element for first mpacket beging processed
	.reg	tcw0_1			; first transmit control word for first mpacket 
	.reg	tcw0_2			; first transmit control word for second mpacket 
	.reg	tcw0_3			; first transmit control word for third mpacket 
	.reg	buf_handle		; buffer handle 
	.reg 	mpacket_number	; number of mpackets being processed
	.reg 	tx_port			; port to transmit the mpacket/mpackets
	.reg	$avail_tbuf_ele_1	; next available tbuf element from scratch 
								; atomic operation
	.reg	v_select_port	; port selected from virtual queue for possible 
							; transmission in next iteration
		
	.reg    $w6
	.reg	$w7
	.xfer_order $w6 $w7

	// following .set used to eliminate assembling warning

	.set tcw0_1 buf_handle $tx_request tbuf_ele_1 $avail_tbuf_ele_1
	.set tcw0_2 
	.set tcw0_3
	.set v_select_port
	.set tx_port
	 
	// allocate 5 read xfer registers for reading in the meta data 

	xbuf_alloc($dl_meta, 5, read)	

	// following .set used to eliminate assembling warning

	.set $dl_meta0 $dl_meta1 $dl_meta2 $dl_meta3 $dl_meta4

	// initialialized tcw1_0 to save instruction cycle in IXP28XX

	alu[$w7, --, B, 0]

	// initialized for no packet available from virtual queue or tx request
	
	alu[exe_stat_flag, --, B, 0]
	alu[mpacket_number, --, B, 0]

	// no mpacket available to transmit in the beginning

	br[picked_from_virtual_queue_or_picked_none_path#]		

	// critical path

picked_from_tx_request_path#:

// When a valid tx_request is received by this thread from the previous 
// iteration, if there is no packet queued in the local memory for the port
// of the tx_request and that port is not flow-controlled, this packet can be
// transmitted immediately without queued into local memeory, so a lot of 
// instruction cycles is saved. The design (including the concept of virtual
// queue) assures that all critical packets will always be transmitted in this 
// picked_from_tx_request path. When this iteration (phase 1 and 2) finishes,
// depending on whether there is a packet ailbale for transmission for next 
// iteration from a valid tx_request received in this iteration, it will 
// branch to picked_from_tx_request_path#,  or 
// picked_from_virtual_queue_or_picked_none_path# for next iterarion.

	_picked_from_tx_request(picked_from_tx_request_path#, \
						picked_from_virtual_queue_or_picked_none_path#)

	// non-critical path

picked_from_virtual_queue_or_picked_none_path#:

// When a packet picked from virtual queue for this thread in the previous 
// iteration, if the packet picked is still available, this packet will be
// be transmitted. The packet picked may be not availbale any more for one
// of the following reason:
//				 a. the packet has been transmitted in previous iteration 
//      	     b. the port is blocked due to flow control
// In the case the packet is not avaible, other available packet on the local
// memory queue from other potrs will be selected to transmit on a round-roin
// ase. If there is no packet from vitual queue or valid tx_request is picked
// for this thread from the previous iteration, any available packet on the 
// local memory queue will be selected to transmit on a round-roin base. 
// When this iteration (phase 1 and 2) finishes, depending on whether there 
// is a packet availbale for transmission for next iteration from a valid 
// tx_request received in this iteration, it will  branch to 
// picked_from_tx_request_path#,  or 
// picked_from_virtual_queue_or_picked_none_path# for next iterarion

	_picked_from_virtual_queue_or_picked_none(picked_from_tx_request_path#,
			picked_from_virtual_queue_or_picked_none_path#)

.end

#endm	// pkt_tx_16p()


///////////////////////////////////////////////////////////////////////////////
//
// _picked_from_tx_request()
//
// Description: Process packet transmit when there is a packet from tx_request
//              is selected for this thread from previous iteration
//
// Outputs:							
//		None 
//
// Inputs:
//		None 
//
// Constants:
//		None 
//
//	Labels:
//		PICKED_FROM_TX_REQUEST_PATH:	symbolic address to go when a packet
//										from tx_request selected for this
//                                      thread for next iteration
//		PICKED_FROM_VIRTUAL_QUEUE_OR_PICKED_NONE_PATH:	symbolic address to go
//										when a packet from virtual queue or 
//										none either from virtual queue nor 
//										valid tx_request is selected for this 
//										thread for next iteration 
//
// Size:
//		  instructions
//
///////////////////////////////////////////////////////////////////////////////

#macro _picked_from_tx_request(PICKED_FROM_TX_REQUEST_PATH, \
			PICKED_FROM_VIRTUAL_QUEUE_OR_PICKED_NONE_PATH)

.begin

	.reg 	offset
	.reg    offset_mod_8
	.reg 	tx_port_bit		; bit for the port to transmit the mpacket/mpackets
	.reg	mpkt_1_size
	.reg	mpkt_2_size
	.reg 	mpkt_3_size
	.reg	leftover_len
	.reg	leftover_offset

	.reg	$txed_port				; packets transmitted for specific port, used
									; reflect write
	.reg    $w0
	.reg	$w1
	.reg    $w2
	.reg	$w3
	.xfer_order	$w0 $w1 $w2 $w3

	.reg    $w4
	.reg	$w5
	.xfer_order $w4 $w5

	.reg	$$dw0
	.reg	$$dw1
	.reg	$$dw2
	.reg	$$dw3
	.reg	$$dw4
	.reg	$$dw5
	.xfer_order $$dw0 $$dw1 $$dw2 $$dw3 $$dw4 $$dw5
	.set $$dw0 $$dw1 $$dw2 $$dw3 $$dw4 $$dw5

	#ifdef ADD_L2_HEADER
	.reg nexthop_id
	#ifdef ETHERNET_TX
	.reg $l2_entry_lw0 $l2_entry_lw1 $l2_entry_lw2 $l2_entry_lw3
	.xfer_order $l2_entry_lw0 $l2_entry_lw1 $l2_entry_lw2 $l2_entry_lw3	
	#endif // ETHERNET_TX 
	#endif //#ifdef ADD_L2_HEADER

	// following .set used to eliminate assembling warning

	.set offset offset_mod_8 tx_port_bit
	.set mpkt_1_size mpkt_2_size mpkt_3_size
	.set leftover_len leftover_offset

	#ifdef ADD_L2_HEADER
	.set nexthop_id
	#ifdef ETHERNET_TX
	.set $l2_entry_lw0 $l2_entry_lw1 $l2_entry_lw2 $l2_entry_lw3
	#endif // ETHERNET_TX 
	#endif //#ifdef ADD_L2_HEADER

phase_1#:

	.begin
	.sig sig_mpkt_1_tcw			; signal for validating tcw of first mpacket
	.sig sig_mpkt_2_tcw			; signal for validating tcw of second mpacket
	.sig sig_mpkt_3_tcw			; signal for validating tcw of third mpacket
	#ifndef DISABLE_TX2SCHED_FEEDBACK
	.sig sig_reflect_write 			; signal for notifying scheduler
	#endif	// #ifndef DISABLE_TX2SCHED_FEEDBACK
	#ifdef ADD_L2_HEADER
	#ifdef ETHERNET_TX
	.sig sig_l2_table
	#endif // ETHERNET_TX 
	#endif //#ifdef ADD_L2_HEADER

	_pkt_tx_16p_wakeup_next_thread(sig_next_context_gpr)		; (1)
	alu[sig_mask_1, --, b, @default_sigmask_1]	; initialize sig_mask_1

	// process what left by previous iteration 
	// wriet tcw for one mpacket case in previous iteration, this is the 
	// critical case, otherwise branch to not_critical_tcw_write#

	_pkt_tx_16p_write_critcal_tcw_for_prev_iter(sig_mask_1, mpacket_number, \
							tx_port, not_critial_tcw_write#) 	; (7)
write_tcws_for_prev_iter_done#:

	br_bclr[exe_stat_flag, ESF_FREE_BUF_BIT, free_buffer_done#]

	_pkt_tx_16p_free_buffer(buf_handle)							; (4)

free_buffer_done#:

	// read tx sequencer for check tbuf full status in phase 2 (1)

	_pkt_tx_16p_read_tx_sequencer($mpkts_sent, sig_mask_1, sig_msf_access_2)							

	// start to process the packet picked from tx_request
	// get tx_port and tx_port_bit from tx_request

	_pkt_tx_16p_get_tx_port(tx_port, tx_port_bit, $tx_request)	; (3)
	alu[--, @rp, AND, tx_port_bit]
	beq[packet_from_tx_request_not_ready#]		; either packet in queue
													; or flow controlled for 
													; tx_port

	// process tx request for critical case (sop-eop mpacket) in phase 1, 
	// including setting lm index 1 and updating flow control to assure not 
	// overflooding PHY, branch out for other cases to reduce instruction for
	// critical case (39 for ixp2400 A0/A1 / 33 for others)

	_pkt_tx_16p_process_critcal_tx_request_phase_1(mpacket_number, buf_handle, \
				offset, offset_mod_8, tx_port, tx_port_bit, $tx_request, \
				not_sop_eop_mpacket_phase_1#)						
process_tx_request_phase_1_done#:

	// pick packet for next iteration

	ffs[v_select_port, @vpp]
	bne[tx_request_packet_from_virtual_queue_available#]

	// get tx request from scratch ring (1) 

	_pkt_tx_16p_get_tx_request($tx_request, lm1_tx_req_ring, \
												sig_scratch_access_1]

tx_request_get_tx_request_done#:

	_pkt_tx_16p_read_msf_flow_control($mphy_fc_stat, tx_mphy_stat_addr, \
						sig_msf_access_1)	; (1)
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask_1]
	// following .io_complete is used to eliminate assembling warning due to
	// the usage of "ACTIVE_CTX_WAKEUP_EVENTS" csr

	.io_completed sig_sram_access_1
	.io_completed sig_next_context
	.io_completed sig_scratch_access_1
	.io_completed sig_scratch_access_2
	.io_completed sig_msf_access_1
	.io_completed sig_msf_access_2
	.io_completed sig_mpkt_1_tcw
	.io_completed sig_mpkt_2_tcw
	.io_completed sig_mpkt_3_tcw
	.io_completed sig_dram_access_1
	#ifndef DISABLE_TX2SCHED_FEEDBACK
	.io_completed sig_reflect_write
	#endif	// #ifndef DISABLE_TX2SCHED_FEEDBACK
	#ifdef ADD_L2_HEADER
	#ifdef ETHERNET_TX
	.io_completed sig_l2_table
	#endif // ETHERNET_TX 
	#endif //#ifdef ADD_L2_HEADER
	ctx_arb[--], br[tx_request_phase_2#]
	
	
	// non-critical path 

not_critial_tcw_write#:
	_pkt_tx_16p_write_non_critcal_tcws_for_prev_iter(sig_mask_1, \
											mpacket_number, tx_port) 
	br[write_tcws_for_prev_iter_done#]

	// process tx request for not sop-eop mpacket cases in phase 1, 

not_sop_eop_mpacket_phase_1#:
	_pkt_tx_16p_process_non_critcal_tx_request_phase_1(mpacket_number, \
		buf_handle, offset, offset_mod_8, tx_port, tx_port_bit)
	br[process_tx_request_phase_1_done#]
	
tx_request_packet_from_virtual_queue_available#:

	//set local memory index 0 to port queue status of v_select_port

	_pkt_tx_16p_set_lm_index0_to_port_status_info(v_select_port)	; (3)

	nop
	// set ESF_PICK_VIRTUAL_QUEUE_BIT in exe_stat_flag
	alu[exe_stat_flag, exe_stat_flag, OR, 1, <<ESF_PICK_VIRTUAL_QUEUE_BIT]
	
	// no tx_request_read
	
	_pkt_tx_16p_clear_signal(sig_mask_1, sig_scratch_access_1)	

	// update virtual_queue_count for v_select_port and related global flags
	 
	_pkt_tx_16p_decrement_virtual_queue_count_lm0(v_select_port)

	br[tx_request_get_tx_request_done#]

packet_from_tx_request_not_ready#:
  
  	// set local memory index 1 to port status

	_pkt_tx_16p_set_lm_index_1_to_port_status_info(tx_port) ; (3)
	nop
	alu[exe_stat_flag, --, B, 0]							; reset
	alu[mpacket_number, --, B, 0]							; reset

 #ifdef CHECK_PORT_QUEUE_OVERFLOW
	_pkt_tx_16p_check_discard(save_sop_to_lm_done#)			; (6)
#endif	// #ifdef CHECK_PORT_QUEUE_OVERFLOW

	// set local memory index 0 to tail of the queue
	_pkt_tx_16p_set_lm_index_0_to_queue_tail(tx_port)		; (6)


	// save this packet to loacl memory
	 
	_pkt_tx_16p_save_sop_to_lm($tx_request)

	// update related global flags

	_pkt_tx_16p_set_port_not_empty(tx_port_bit)

#ifdef CHECK_PORT_QUEUE_OVERFLOW
save_sop_to_lm_done#:
#endif	// #ifdef CHECK_PORT_QUEUE_OVERFLOW

	// find any packet in local memory queue available to tx 

	_pkt_tx_16p_find_port_to_tx_from_lm_queue(tx_port, tx_port_bit, \
												no_port_availbale_to_tx#)

	_pkt_tx_16p_process_pkt_in_lm_queue_phase_1(mpacket_number, buf_handle, \
				offset, offset_mod_8, tx_port, tx_port_bit)

	br[process_pkt_in_lm_queue_phase_1_done#]

no_port_availbale_to_tx#:

	// no tbuf_ele allocation

	_pkt_tx_16p_clear_signal(sig_mask_1, sig_scratch_access_2)	

process_pkt_in_lm_queue_phase_1_done#:

	// pick packet for next iteration

	ffs[v_select_port, @vpp]
	bne[lm_queue_packet_from_virtual_queue_available#]

	// get tx request from scratch ring (1) 

	_pkt_tx_16p_get_tx_request($tx_request, lm1_tx_req_ring, \
												sig_scratch_access_1]

	br[lm_queue_get_tx_request_done#]


lm_queue_packet_from_virtual_queue_available#:

	//set local memory index 0 to port queue status of v_select_port

	_pkt_tx_16p_set_lm_index0_to_port_status_info(v_select_port)	; (3)

	nop
	// set ESF_PICK_VIRTUAL_QUEUE_BIT in exe_stat_flag
	alu[exe_stat_flag, exe_stat_flag, OR, 1, <<ESF_PICK_VIRTUAL_QUEUE_BIT]
	
	// no tx_request_read
	
	_pkt_tx_16p_clear_signal(sig_mask_1, sig_scratch_access_1)	

	// update virtual_queue_count for v_select_port and related global flags
	 
	_pkt_tx_16p_decrement_virtual_queue_count_lm0(v_select_port)


lm_queue_get_tx_request_done#:

	_pkt_tx_16p_read_msf_flow_control($mphy_fc_stat, tx_mphy_stat_addr, \
						sig_msf_access_1)	; (1)
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask_1]
	// following .io_complete is used to eliminate assembling warning due to
	// the usage of "ACTIVE_CTX_WAKEUP_EVENTS" csr

	.io_completed sig_sram_access_1
	.io_completed sig_next_context
	.io_completed sig_scratch_access_1
	.io_completed sig_scratch_access_2
	.io_completed sig_msf_access_1
	.io_completed sig_msf_access_2
	.io_completed sig_mpkt_1_tcw
	.io_completed sig_mpkt_2_tcw
	.io_completed sig_mpkt_3_tcw
	.io_completed sig_dram_access_1
	#ifndef DISABLE_TX2SCHED_FEEDBACK
	.io_completed sig_reflect_write
	#endif	// #ifndef DISABLE_TX2SCHED_FEEDBACK
	#ifdef ADD_L2_HEADER
	#ifdef ETHERNET_TX
	.io_completed sig_l2_table
	#endif // ETHERNET_TX 
	#endif //#ifdef ADD_L2_HEADER
	ctx_arb[--], br[lm_queue_phase_2#]

.end

tx_request_phase_2#:

	.begin
	.sig sig_dram_access_2		; signal 2 for dram access (dram[tbuf_wr..])
	.sig sig_dram_access_3		; signal 3 for dram access (dram[tbuf_wr..])

	#ifdef ADD_L2_HEADER
	.sig sig_l2_header			; signal for l2 table entry read
	#ifdef POS_TX
		.reg	$prepend_0
	#endif	// #ifdef POS_TX
	#endif	// #ifdef ADD_L2_HEADER

	_pkt_tx_16p_wakeup_next_thread(sig_next_context_gpr)
	//initialize the sig_mask_2 for phase 2 for sop_eop_mpkt 
	alu[sig_mask_2, --, b, @default_sigmask_2]	
	
	// tbuf already allocated, check tbuf not full and move payload from dram
	// to tbuf, prepare transmit control word for critical case (sop-eop 
	// mpacket), branch out for other cases to save instruction cycle (24)

	_pkt_tx_16p_process_critical_tx_request_phase_2(sig_mask_2, \
			mpacket_number, offset, offset_mod_8, tx_port, tx_port_bit, \
			buf_handle, not_sop_eop_mpacket_phase_2#) 				
process_tx_request_phase_2_done#:

	br_bset[exe_stat_flag, ESF_PICK_VIRTUAL_QUEUE_BIT, \
							tx_request_picked_from_virtual_queue_for_next_iter#]
	
	// check whether there is null tx requests in scratch ring 

	alu[--, $tx_request, -, 0x0]					; valid tx request
	// no valid tx request
	beq[tx_request_none_picked_for_next_iter#]		
			
	_pkt_tx_16p_read_meta_data($dl_meta, sig_mask_2, sig_sram_access_1, \
					$tx_request) 					; (4)

	_pkt_tx_16p_debug_incr_counter(@pkt_tx_16p_tx_request_rxed)


tx_request_phase_2_done#:

	// update flow control info based on port status (20 (IXP24XX) / 26 
	// (IXP28XX) for thread 7, 4 for other thread)

	_pkt_tx_16p_update_flow_control_from_status()  
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask_2]

	// following .io_complete is used to eliminate assembling warning due to
	// the usage of "ACTIVE_CTX_WAKEUP_EVENTS" csr

	.io_completed sig_next_context
	.io_completed sig_dram_access_1
	.io_completed sig_dram_access_2
	.io_completed sig_dram_access_3
	.io_completed sig_msf_access_2
	.io_completed sig_sram_access_1
	.io_completed sig_msf_access_1
	#ifdef ADD_L2_HEADER
	.io_completed sig_l2_header
	#endif	// #ifdef ADD_L2_HEADER
	ctx_arb[--], br[PICKED_FROM_TX_REQUEST_PATH]


	// non-critical path

	// move payload from dram to tbuf, prepare transmit control word for not 
	// sop-eop mpacket cases

not_sop_eop_mpacket_phase_2#:
	_pkt_tx_16p_process_non_critical_tx_request_phase_2(sig_mask_2, \
			mpacket_number, offset, offset_mod_8, tx_port, tx_port_bit, \
			buf_handle) 				
	br[process_tx_request_phase_2_done#]


tx_request_picked_from_virtual_queue_for_next_iter#:
tx_request_none_picked_for_next_iter#:

	_pkt_tx_16p_clear_signal(sig_mask_2, sig_sram_access_1)	

	// update flow control info based on port status (20 (IXP24XX) / 26 
	// (IXP28XX) for thread 7, 4 for other thread)

	_pkt_tx_16p_update_flow_control_from_status()  

	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask_2]

	// following .io_complete is used to eliminate assembling warning due to 
	// the usage of "ACTIVE_CTX_WAKEUP_EVENTS" csr

	.io_completed sig_next_context
	.io_completed sig_dram_access_1
	.io_completed sig_dram_access_2
	.io_completed sig_dram_access_3
	.io_completed sig_msf_access_2
	.io_completed sig_sram_access_1
	.io_completed sig_msf_access_1
	#ifdef ADD_L2_HEADER
	.io_completed sig_l2_header
	#endif	// #ifdef ADD_L2_HEADER
	ctx_arb[--], br[PICKED_FROM_VIRTUAL_QUEUE_OR_PICKED_NONE_PATH]

	.end


lm_queue_phase_2#:

	.begin
	.sig sig_dram_access_2		; signal 2 for dram access (dram[tbuf_wr..])
	.sig sig_dram_access_3		; signal 3 for dram access (dram[tbuf_wr..])

	#ifdef ADD_L2_HEADER
	.sig sig_l2_header			; signal for l2 table entry read
	#ifdef POS_TX
	.reg	$prepend_0
	#endif	// #ifdef POS_TX
	#endif	// #ifdef ADD_L2_HEADER

	_pkt_tx_16p_wakeup_next_thread(sig_next_context_gpr)
	//initialize the sig_mask_2 for phase 2 for sop_eop_mpkt 
	alu[sig_mask_2, --, b, @default_sigmask_2]	
	
	// tbuf already allocated, check tbuf not full and move payload from dram
	// to tbuf, prepare transmit control word
	// (for 1 mpacket:  29 for ixp2400 A0/A1,  23 for others)

	_pkt_tx_16p_process_pkt_in_lm_queue_phase_2(sig_mask_2, mpacket_number, \
				offset, offset_mod_8, tx_port, tx_port_bit, buf_handle) 						;

	br_bset[exe_stat_flag, ESF_PICK_VIRTUAL_QUEUE_BIT, \
						lm_queue_picked_from_virtual_queue_for_next_iter#]
	
	// check whether there is null tx requests in scratch ring 

	alu[--, $tx_request, -, 0x0]					; valid tx request
	// no valid tx request
	beq[lm_queue_none_picked_for_next_iter#]		
			
	_pkt_tx_16p_read_meta_data($dl_meta, sig_mask_2, sig_sram_access_1, \
					$tx_request) 					; (4)

	_pkt_tx_16p_debug_incr_counter(@pkt_tx_16p_tx_request_rxed)
	
lm_queu_phase_2_done#:

	// update flow control info based on port status (20 (IXP24XX) / 26 
	// (IXP28XX) for thread 7, 4 for other thread)

	_pkt_tx_16p_update_flow_control_from_status()  
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask_2]

	// following .io_complete is used to eliminate assembling warning due to 
	// the usage of "ACTIVE_CTX_WAKEUP_EVENTS" csr

	.io_completed sig_next_context
	.io_completed sig_dram_access_1
	.io_completed sig_dram_access_2
	.io_completed sig_dram_access_3
	.io_completed sig_msf_access_2
	.io_completed sig_sram_access_1
	.io_completed sig_msf_access_1
	#ifdef ADD_L2_HEADER
	.io_completed sig_l2_header
	#endif	// #ifdef ADD_L2_HEADER
	ctx_arb[--], br[PICKED_FROM_TX_REQUEST_PATH]


	// non-critical path


lm_queue_picked_from_virtual_queue_for_next_iter#:
lm_queue_none_picked_for_next_iter#:

	//  clear sig_sram_access_1

	_pkt_tx_16p_clear_signal(sig_mask_2, sig_sram_access_1)	

	// update flow control info based on port status (20 (IXP24XX) / 26 
	// (IXP28XX) for thread 7, 4 for other thread)

	_pkt_tx_16p_update_flow_control_from_status()  

	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask_2]

	// following .io_complete is used to eliminate assembling warning due to 
	// the usage of "ACTIVE_CTX_WAKEUP_EVENTS" csr

	.io_completed sig_next_context
	.io_completed sig_dram_access_1
	.io_completed sig_dram_access_2
	.io_completed sig_dram_access_3
	.io_completed sig_msf_access_2
	.io_completed sig_sram_access_1
	.io_completed sig_msf_access_1
	#ifdef ADD_L2_HEADER
	.io_completed sig_l2_header
	#endif	// #ifdef ADD_L2_HEADER
	ctx_arb[--], br[PICKED_FROM_VIRTUAL_QUEUE_OR_PICKED_NONE_PATH]

	.end

.end

#endm	// _picked_from_tx_request()


///////////////////////////////////////////////////////////////////////////////
//
// _picked_from_virtual_queue()
//
// Description: Process packet transmit when there is a packet from vitual 
//				queue is selected for this thread from previous iteration
//

///////////////////////////////////////////////////////////////////////////////
//
// _picked_from_virtual_queue_or_picked_none()
//
// Description: Process packet transmit when there is a packet from vitual 
//				queue is selected for this thread from previous iteration or
//				there is no packet either from vitual queue or tx_reuest is 
//				selected for this thread from previous iteration (depends on 
//				"ESF_PICK_VIRTUAL_QUEUE_BIT" bit in exe_stat_flag) 
//
// Outputs:							
//		None 
//
// Inputs:
//		None 
//
// Constants:
//		None 
//
//	Labels:
//		PICKED_FROM_TX_REQUEST_PATH:	symbolic address to go when a packet
//										from tx_request selected for this
//                                      thread for next iteration
//		PICKED_FROM_VIRTUAL_QUEUE_OR_PICKED_NONE_PATH:	symbolic address to go
//										when a packet from virtual queue or 
//										none either from virtual queue nor 
//										valid tx_request is selected for this 
//										thread for next iteration 
//
// Size:
//		  instructions
//
///////////////////////////////////////////////////////////////////////////////

#macro _picked_from_virtual_queue_or_picked_none(PICKED_FROM_TX_REQUEST_PATH, \
			PICKED_FROM_VIRTUAL_QUEUE_OR_PICKED_NONE_PATH)
.begin

	.reg 	offset
	.reg	offset_mod_8
	.reg 	tx_port_bit		; bit for the port to transmit the mpacket/mpackets
	.reg	mpkt_1_size
	.reg	mpkt_2_size
	.reg 	mpkt_3_size
	.reg	leftover_len
	.reg	leftover_offset
	.reg	$txed_port				; packets transmitted for specific port, used
									; reflect write
	.reg    $w0
	.reg	$w1
	.reg    $w2
	.reg	$w3
	.xfer_order	$w0 $w1 $w2 $w3
	.reg    $w4
	.reg	$w5
	.xfer_order $w4 $w5

	.reg	$$dw0
	.reg	$$dw1
	.reg	$$dw2
	.reg	$$dw3
	.reg	$$dw4
	.reg	$$dw5
.xfer_order $$dw0 $$dw1 $$dw2 $$dw3 $$dw4 $$dw5
.set $$dw0 $$dw1 $$dw2 $$dw3 $$dw4 $$dw5

//	.set tx_port_bit

	#ifdef ADD_L2_HEADER
	.reg nexthop_id
	#ifdef ETHERNET_TX
	.reg $l2_entry_lw0 $l2_entry_lw1 $l2_entry_lw2 $l2_entry_lw3
	.xfer_order $l2_entry_lw0 $l2_entry_lw1 $l2_entry_lw2 $l2_entry_lw3	
	#endif // ETHERNET_TX 
	#endif //#ifdef ADD_L2_HEADER

	// following .set used to eliminate assembling warning

	.set offset offset_mod_8 tx_port_bit
	.set mpkt_1_size mpkt_2_size mpkt_3_size
	#ifdef ADD_L2_HEADER
	.set nexthop_id
	#ifdef ETHERNET_TX	
	.set $l2_entry_lw0 $l2_entry_lw1 $l2_entry_lw2 $l2_entry_lw3
	#endif // ETHERNET_TX 
	#endif //#ifdef ADD_L2_HEADER
	.set leftover_len leftover_offset

phase_1_start#:

	.begin
	.sig sig_mpkt_1_tcw			; signal for validating tcw of first mpacket
	.sig sig_mpkt_2_tcw			; signal for validating tcw of second mpacket
	.sig sig_mpkt_3_tcw			; signal for validating tcw of third mpacket
	#ifndef DISABLE_TX2SCHED_FEEDBACK
	.sig sig_reflect_write 				; signal for notifying scheduler
	#endif	// #ifndef DISABLE_TX2SCHED_FEEDBACK
	#ifdef ADD_L2_HEADER
	#ifdef ETHERNET_TX
	.sig sig_l2_table
	#endif // ETHERNET_TX 
	#endif //#ifdef ADD_L2_HEADER
	.reg	tmp

	_pkt_tx_16p_wakeup_next_thread(sig_next_context_gpr)
	alu[sig_mask_1, --, b, @default_sigmask_1]	; initialize sig_mask_1

	// process what left by previous iteration 

	_pkt_tx_16p_write_tcws_for_prev_iter(sig_mask_1, mpacket_number, \
							tx_port)			; (10 for one mpacket)

	br_bclr[exe_stat_flag, ESF_FREE_BUF_BIT, free_buffer_done#]

	_pkt_tx_16p_free_buffer(buf_handle)			; (4)

free_buffer_done#:

	// read tx sequencer for check tbuf full status in phase 2 (1)

	_pkt_tx_16p_read_tx_sequencer($mpkts_sent, sig_mask_1, sig_msf_access_2)							

	br_bclr[exe_stat_flag, ESF_PICK_VIRTUAL_QUEUE_BIT, \
				not_virtual_queue_selected_in_prev_iter#]

	// check whethetr the virtual select port (v_select_port) still available 
	// for transmission
	
	alu[--, v_select_port, or, 0]
	alu_shf[tmp, --, b, 1, <<indirect]			; v_select_port_bit

	alu[tmp, tmp, AND~, @ep]
	alu[tmp, tmp, AND~, @bfcp]
	alu[tmp, tmp, AND~ , @blocked_by_read_leftover]								
	alu[tmp, tmp, AND~ , @blocked_by_read_next_buf]			
	beq[v_select_port_not_available_now#]

	alu[tx_port, --, B, v_select_port]
	alu[tx_port_bit, --, B, tmp]
	alu[mpacket_number, --, B, 0]
	alu[exe_stat_flag, --, B, 0]
	br[one_port_from_lm_queue_found_to_be_available_to_tx#]

v_select_port_not_available_now#:
	
	// since v_select_port not available, maintain virtual count for
	// v_select_port

	_pkt_tx_16p_set_lm_index0_to_port_status_info(v_select_port)	; (3)
	nop
	nop
	nop
	_pkt_tx_16p_increment_virtual_queue_count_lm0(v_select_port)

not_virtual_queue_selected_in_prev_iter#:
	alu[mpacket_number, --, B, 0]
	alu[exe_stat_flag, --, B, 0]

try_to_find_a_port_from_lm_queue_to_tx#:	

	_pkt_tx_16p_find_port_to_tx_from_lm_queue(tx_port, tx_port_bit, \
												no_port_availbale_to_tx#)
one_port_from_lm_queue_found_to_be_available_to_tx#:

	_pkt_tx_16p_process_pkt_in_lm_queue_phase_1(mpacket_number, buf_handle, \
				offset, offset_mod_8, tx_port, tx_port_bit)

	br[process_pkt_in_lm_queue_phase_1_done#]

no_port_availbale_to_tx#:

	// clear signal for allocate $avail_tbuf_ele_1 first, will added later if
	// required
	_pkt_tx_16p_clear_signal(sig_mask_1, sig_scratch_access_2)	; clear sig_scratch_access_2

process_pkt_in_lm_queue_phase_1_done#:

	// pick packet for next iteration

	ffs[v_select_port, @vpp]
	beq[pick_from_virtual_queue_done#]
	//set local memory index 0 to port queue status of v_select_port

	_pkt_tx_16p_set_lm_index0_to_port_status_info(v_select_port)	; (3)

	nop
	// set ESF_PICK_VIRTUAL_QUEUE_BIT in exe_stat_flag
	alu[exe_stat_flag, exe_stat_flag, OR, 1, <<ESF_PICK_VIRTUAL_QUEUE_BIT]
	
	// no tx_request_read
	
	_pkt_tx_16p_clear_signal(sig_mask_1, sig_scratch_access_1)	

	// update virtual_queue_count for v_select_port and related global flags
	 
	_pkt_tx_16p_decrement_virtual_queue_count_lm0(v_select_port)

	br[phase_1_done#]

pick_from_virtual_queue_done#:

	// get tx request from scratch ring  (1)

	_pkt_tx_16p_get_tx_request($tx_request, lm1_tx_req_ring, \
												sig_scratch_access_1]

phase_1_done#:

	_pkt_tx_16p_read_msf_flow_control($mphy_fc_stat, tx_mphy_stat_addr, \
						sig_msf_access_1)	; (1)

	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask_1]
	ctx_arb[--]	

	// following .io_complete is used to eliminate assembling warning due to
	// the usage of "ACTIVE_CTX_WAKEUP_EVENTS" csr

	.io_completed sig_next_context
	.io_completed sig_scratch_access_1
	.io_completed sig_scratch_access_2
	.io_completed sig_msf_access_1
	.io_completed sig_msf_access_2
	.io_completed sig_mpkt_1_tcw
	.io_completed sig_mpkt_2_tcw
	.io_completed sig_mpkt_3_tcw
	.io_completed sig_sram_access_1
	.io_completed sig_dram_access_1
	#ifndef DISABLE_TX2SCHED_FEEDBACK
	.io_completed sig_reflect_write 			
	#endif	// #ifndef DISABLE_TX2SCHED_FEEDBACK
	#ifdef ADD_L2_HEADER
	#ifdef ETHERNET_TX
	.io_completed sig_l2_table
	#endif // ETHERNET_TX 
	#endif //#ifdef ADD_L2_HEADER

	.end

_picked_from_virtual_queue_or_picked_none_phase_2_start#:

	.begin
	.sig sig_dram_access_2			; signal 2 for dram access (dram[tbuf_wr..])
	.sig sig_dram_access_3			; signal 3 for dram access (dram[tbuf_wr..])

	#ifdef ADD_L2_HEADER
	.sig sig_l2_header			; signal 6 for msf access (read or write)
	#ifdef POS_TX
	.reg	$prepend_0
	#endif	// #ifdef POS_TX
	#endif	// #ifdef ADD_L2_HEADER

	_pkt_tx_16p_wakeup_next_thread(sig_next_context_gpr)

	//initialize the sig_mask_2 for phase 2 for sop_eop_mpkt 

	alu[sig_mask_2, --, b, @default_sigmask_2]	

	// (for 1 mpacket:  29 for ixp2400 A0/A1,  23 for others)


	_pkt_tx_16p_process_pkt_in_lm_queue_phase_2(sig_mask_2, mpacket_number, \
					 offset, offset_mod_8, tx_port, tx_port_bit, buf_handle)		

	br_bset[exe_stat_flag, ESF_PICK_VIRTUAL_QUEUE_BIT, \
							picked_from_virtual_queue_for_next_iter#]

	alu[--, $tx_request, -, 0x0]		; valid tx request

	// no valid tx request

	beq[none_picked_for_next_iter#]		

	_pkt_tx_16p_read_meta_data($dl_meta, sig_mask_2, sig_sram_access_1, \
					$tx_request) 					; (4)

	_pkt_tx_16p_debug_incr_counter(@pkt_tx_16p_tx_request_rxed)

phase_2_done#:

	// update flow control info based on port status (20 (IXP24XX) / 26 
	// (IXP28XX) for thread 7, 4 for other thread)

	_pkt_tx_16p_update_flow_control_from_status()  
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask_2]

	// following .io_complete is used to eliminate assembling warning due to 
	// the usage of "ACTIVE_CTX_WAKEUP_EVENTS" csr

	.io_completed sig_next_context
	.io_completed sig_dram_access_1
	.io_completed sig_dram_access_2
	.io_completed sig_dram_access_3
	.io_completed sig_msf_access_2
	.io_completed sig_sram_access_1
	.io_completed sig_msf_access_1
	#ifdef ADD_L2_HEADER
	.io_completed sig_l2_header
	#endif	// #ifdef ADD_L2_HEADER
	ctx_arb[--], br[PICKED_FROM_TX_REQUEST_PATH]

picked_from_virtual_queue_for_next_iter#:
none_picked_for_next_iter#:

	_pkt_tx_16p_clear_signal(sig_mask_2, sig_sram_access_1)	

	// update flow control info based on port status (20 (IXP24XX) / 26 
	// (IXP28XX) for thread 7, 4 for other thread)

	_pkt_tx_16p_update_flow_control_from_status()  
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask_2]

	// following .io_complete is used to eliminate assembling warning due to
	// the usage of "ACTIVE_CTX_WAKEUP_EVENTS" csr
	.io_completed sig_next_context
	.io_completed sig_dram_access_1
	.io_completed sig_dram_access_2
	.io_completed sig_dram_access_3
	.io_completed sig_msf_access_2
	.io_completed sig_sram_access_1
	.io_completed sig_msf_access_1
	#ifdef ADD_L2_HEADER
	.io_completed sig_l2_header
	#endif	// #ifdef ADD_L2_HEADER
	ctx_arb[--], br[PICKED_FROM_VIRTUAL_QUEUE_OR_PICKED_NONE_PATH]

	.end
	 
.end	

#endm	// _picked_from_virtual_queue_or_picked_none()
														
///////////////////////////////////////////////////////////////////////////////
//
//	Main entry point -- Code begins execution here.
//
///////////////////////////////////////////////////////////////////////////////

main#:

	// initialize the microblock

    pkt_tx_16p_init()

	// packet processing block

	pkt_tx_16p()

///////////////////////////////////////////////////////////////////////////////

#endif	// __PKT_TX_P16_UC__



