/*******************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 1998-2002 By Intel Corporation.  All rights reserved.  
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *******************************************************************************/

/**
 *******************************************************************************
 * @ingroup Packet Tx
 * File:  packet_txx.h
 *
 * @description:
 *    This is the SPHY and NPHY4 transmit code header file.  This file 
 *    defines data type and constants which will be used in the SPHY and 
 *	  MPHY_4 tramiting microengine C code.
 *******************************************************************************/

/**
 *******************************************************************************
 * Contents:
 *    Definitions of data types 
 *		exe_stat_flag_t
 *		input_tx_req_t
 *		q_entry_t	
 *		q_port_t
 *		q_status_t
 *		in_parameter_1_t
 *
 *    Definitions of the constants used in the packet Tx module
 *
 *	  Prototypes of the internal functions used in the Packet Tx Module
 *******************************************************************************/

#ifndef	__PACKETTX_H__
#define	__PACKETTX_H__

#include <ixp.h>
#include <definitions.h>
#include <dl_system.h>
#include <sig_functions.h>
#include <hardware.h>
#include <dl_buf.h>
#include <ixp_lib.h>

/*****************************************************************************
 * Abstract 
 *    Thread excution status flag
 *
 * Purpose/Intended Use:
 *    This data stucture will be used to store the thread execution staus 
 *    information.
 *****************************************************************************/
typedef struct
{
 union
 {
  struct
  {
    unsigned int myRdRingFlag;							/* Reserved bits.	 */
  };
  unsigned int value;
 };
} thread_flag_t;


/*****************************************************************************
 * Abstract 
 *    Input Tx request data structure
 *
 * Purpose/Intended Use:
 *    This data stucture will be used to store Input TX request data which is 
 *    from scratch ring read.
 *****************************************************************************/
typedef struct
{
 union
 {
  struct
  {
    unsigned int validBit				: 1;
    unsigned int reserved				: 3;
    unsigned int portNum				: 4;
    unsigned int sopMetaHndl			: 24;
  };
  unsigned int value;
 };
} input_tx_req_t;


/*****************************************************************************
 * Abstract 
 *    Queue entry for each port
 *
 * Purpose/Intended Use:
 *    This data stucture will be used to store queue entries for each port. 
 *****************************************************************************/
typedef __declspec(local_mem shared aligned(16)) struct
{
 union
 {
  struct
  {
    unsigned int mpktFlag;					/*1 mpkt ready to be transmit.*/
    unsigned int globaleRdRingFlag;			/*1 thread is reading ring.   */
  };
  unsigned int value[14];
 };
} q_entry_t;

/*****************************************************************************
 * Abstract 
 *    Queue data structure for each port
 *
 * Purpose/Intended Use:
 *    This data stucture will be used to store queue for each port. 
 *****************************************************************************/
typedef struct
{
	q_entry_t q_entry;	/* Each port queue has up to 16 enties */
} q_port_t;


/*****************************************************************************
 * Abstract 
 *    Input parameter data stucture 
 *
 * Purpose/Intended Use:
 *    This data stucture will be used to a group of the parameters to the 
 *	  move_not_sop_to_tbuf function.
 *****************************************************************************/

typedef struct
{
	unsigned int inBufDesc;			/* Input buffer descripter.		 */
	unsigned int inAddrTbuf;	/* Input tbuffer element address.*/
	unsigned int inOffset;		/* Input dram offset. 			 */
	unsigned int inTcw0;		/* Input control word. 			 */
	SIGNAL_PAIR *pInSigDram		/* Input signal address. 		 */
} in_parameter_1_t;
	

#if	(TX_PHY_MODE != SPHY_1_32)\
      && (TX_PHY_MODE != SPHY_1_16_2_8) \
	  && (TX_PHY_MODE != SPHY_2_16)\
	  && (TX_PHY_MODE != SPHY_4_8)\
	  && (TX_PHY_MODE != MPHY_4)\
	  && (TX_PHY_MODE != MPHY_16)
#undef	TX_PHY_MODE
#define	TX_PHY_MODE		SPHY_4_8	/* This is the default TX PHY mode*/
#endif

#define TOTAL_TBUF_SIZE 				8192
#define TOTOL_THREADS_PER_ME			8
#define TBUF_ELE_SIZE					128
#define TBUF_ADDR_SHF					7
#define TBUF_ELE_SIZE_MINUS_8			(TBUF_ELE_SIZE - 8)
#define TWO_TBUF_ELE_SIZE_MINUS_16 		((TBUF_ELE_SIZE * 2) - 16)

#if (TX_PHY_MODE == SPHY_1_32)
#define TBUF_WAIT_THRESHOLD        		32
#define TBUF_ELE_NUM_PER_SEGMENT		(TOTAL_TBUF_SIZE / TBUF_ELE_SIZE)
#endif


#if (TX_PHY_MODE == SPHY_4_8)
#define TBUF_WAIT_THRESHOLD        		8
#define TBUF_ELE_NUM_PER_SEGMENT		(TOTAL_TBUF_SIZE / (TBUF_ELE_SIZE * 4))
#endif

#if (TX_PHY_MODE == MPHY_4)
#define TBUF_WAIT_THRESHOLD        		8
#define TBUF_ELE_NUM_PER_SEGMENT		(TOTAL_TBUF_SIZE / (TBUF_ELE_SIZE * 4))
#endif

#if (TX_PHY_MODE == SPHY_1_32)
#define BUS_WIDTH_IN_BYTES				4
#endif

#if (TX_PHY_MODE == MPHY_4)
#define BUS_WIDTH_IN_BYTES				4
#endif


#define POS_TX_UP_CTL_WD				0x001D
#define ATM_TX_UP_CTL_WD				0x000C
#define MPHY_16_POS_TX_FIRST_ME			6
#define MPHY_16_POS_TX_SECOND_ME		7

#define	PORTS_CTRLED_BY_FIRST_POS_TX_ME		0x00FF
#define	PORTS_CTRLED_BY_SECOND_POS_TX_ME	0xFF00


#if (TX_PHY_MODE == SPHY_1_32)
#define TBUF_ELE_BITS_MASK 				((TOTAL_TBUF_SIZE / TBUF_ELE_SIZE) - 1)
#endif

#if (TX_PHY_MODE == SPHY_4_8)
#define TBUF_ELE_BITS_MASK 				(((TOTAL_TBUF_SIZE / 4) / TBUF_ELE_SIZE) - 1)
#endif 

#if (TX_PHY_MODE == MPHY_4)
#define TBUF_ELE_BITS_MASK 				(((TOTAL_TBUF_SIZE / 4) / TBUF_ELE_SIZE) - 1)
#endif



#define QUEUE_ENTRY_BASE				0
#define QUEUE_ENTRY_SIZE				0x10		
#define QUEUE_DEPTH						(0x100 / QUEUE_ENTRY_SIZE) 
#define CONST_07						0x07	
#define TCW0_EOP_BIT_SHFT				8
#define TCW0_SKIP_BIT_SHFT				11
#define DRAM_INDIREF_BASE_LW			0x02000010

#if (TX_PHY_MODE == SPHY_1_32)
#define ME_THREAD_NUM_SHFT				3
#endif

#if (TX_PHY_MODE == SPHY_4_8)
#define ME_THREAD_NUM_SHFT				1
#endif

#if (TX_PHY_MODE == MPHY_4)
#define ME_THREAD_NUM_SHFT	1
#endif


#define PKTS_IN_Q_MASK					0xFF

#ifdef COUNTERS
#define	PKT_TRANSMITTED_CNT_IDX			0x00
#define	PKT_DROPPED_CNT_IDX				0x04
#define	BYTE_TRANSMITTED_CNT_IDX		0x08
#define RESERVED_CNT_IDX				0x0C
#define PORT_MASK_IN_TCW            	0x0F
#endif

#define	TCW0_PAYLOAD_LEN_LSB_LOC		24
#define TCW0_PAYLOAD_OFFSET_LSB_LOC		13
#define TCWO_PREPEND_LEN_LSB_LOC		16
#define TCWO_PREPEND_OFFSET_LSB_LOC		21
#define TCW0_PAYLOAD_OFFSET_MASK		0x07

#ifdef DEBUG_TX_PENDING_LOCKED
#define SRAM_WRITE_1_SIG	11
#define FIRST_ME_TBUF_NO_SRAM_BASE		0x100000
#define SECOND_ME_TBUF_NO_SRAM_BASE		0x200000
#endif



#define PORT_GLOBAL_INFO_BASE			2048
#define PORT_GLOBAL_OFFSET_SHF			7
#define TCW_VD_BASE_SHF					7
#define TBUF_BASE_SHF					2048
#define TX_SEQUENCE_ADDR_SHF			2

#ifdef IXP2800
#define PACKETP_TX_SPHY_1X32_CTL_LW		0x20000454
#define ELE_2_TX_CNTRL_ADDR				4		
#else 
#define PACKET_TX_SPHY_1X32_CTL_LW		0x10000004
#endif


#ifdef IXP2800
#define PACKET_TX_SPHY_4X8_CTL_LW		0x20000454
#define ELE_2_TX_CNTRL_ADDR				4
#else 
#define PACKET_TX_SPHY_4X8_CTL_LW		0xFF200004
#endif

#ifdef IXP2800
#define PACKET_TX_MPHY4_CTL_LW	0x20000454
#define ELE_2_TX_CNTRL_ADDR	4
#else
#define PACKET_TX_MPHY4_CTL_LW	0xFF080004
#endif


#define PORT_QUEUE_INFO_BASE				0
#define PORT_0 	0
#define PORT_1	1
#define PORT_2 	2
#define PORT_3	3

#if (TX_PHY_MODE == SPHY_1_32)
#define	SCRATCH_RING_THREAD_0		PACKET_TX_IN_0
#define	SCRATCH_RING_THREAD_1		PACKET_TX_IN_0
#define	SCRATCH_RING_THREAD_2		PACKET_TX_IN_0
#define	SCRATCH_RING_THREAD_3		PACKET_TX_IN_0
#define	SCRATCH_RING_THREAD_4		PACKET_TX_IN_0
#define	SCRATCH_RING_THREAD_5		PACKET_TX_IN_0
#define	SCRATCH_RING_THREAD_6		PACKET_TX_IN_0
#define	SCRATCH_RING_THREAD_7		PACKET_TX_IN_0
#define PORT_HANDLED_BY_THREAD_0	PORT_0
#define PORT_HANDLED_BY_THREAD_1	PORT_0
#define PORT_HANDLED_BY_THREAD_2	PORT_0
#define PORT_HANDLED_BY_THREAD_3	PORT_0
#define PORT_HANDLED_BY_THREAD_4	PORT_0
#define PORT_HANDLED_BY_THREAD_5	PORT_0
#define PORT_HANDLED_BY_THREAD_6	PORT_0
#define PORT_HANDLED_BY_THREAD_7	PORT_0
#endif

#if (TX_PHY_MODE == SPHY_4_8)
#define	SCRATCH_RING_THREAD_0		PACKET_TX_IN_0
#define	SCRATCH_RING_THREAD_1		PACKET_TX_IN_0
#define	SCRATCH_RING_THREAD_2		PACKET_TX_IN_1
#define	SCRATCH_RING_THREAD_3		PACKET_TX_IN_1
#define	SCRATCH_RING_THREAD_4		PACKET_TX_IN_2
#define	SCRATCH_RING_THREAD_5		PACKET_TX_IN_2
#define	SCRATCH_RING_THREAD_6		PACKET_TX_IN_3
#define	SCRATCH_RING_THREAD_7		PACKET_TX_IN_3
#define PORT_HANDLED_BY_THREAD_0	PORT_0
#define PORT_HANDLED_BY_THREAD_1	PORT_0
#define PORT_HANDLED_BY_THREAD_2	PORT_1
#define PORT_HANDLED_BY_THREAD_3	PORT_1
#define PORT_HANDLED_BY_THREAD_4	PORT_2
#define PORT_HANDLED_BY_THREAD_5	PORT_2
#define PORT_HANDLED_BY_THREAD_6	PORT_3
#define PORT_HANDLED_BY_THREAD_7	PORT_3
#endif

#if (TX_PHY_MODE == MPHY_4)
#define	SCRATCH_RING_THREAD_0		PACKET_TX_IN_0
#define	SCRATCH_RING_THREAD_1		PACKET_TX_IN_0
#define	SCRATCH_RING_THREAD_2		PACKET_TX_IN_1
#define	SCRATCH_RING_THREAD_3		PACKET_TX_IN_1
#define	SCRATCH_RING_THREAD_4		PACKET_TX_IN_2
#define	SCRATCH_RING_THREAD_5		PACKET_TX_IN_2
#define	SCRATCH_RING_THREAD_6		PACKET_TX_IN_3
#define	SCRATCH_RING_THREAD_7		PACKET_TX_IN_3
#define PORT_HANDLED_BY_THREAD_0	PORT_0
#define PORT_HANDLED_BY_THREAD_1	PORT_0
#define PORT_HANDLED_BY_THREAD_2	PORT_1
#define PORT_HANDLED_BY_THREAD_3	PORT_1
#define PORT_HANDLED_BY_THREAD_4	PORT_2
#define PORT_HANDLED_BY_THREAD_5	PORT_2
#define PORT_HANDLED_BY_THREAD_6	PORT_3
#define PORT_HANDLED_BY_THREAD_7	PORT_3
#endif

#define TCW0_SOP_EOP_BASE_W0		0x0300
#define TCW0_SOP_EOP_BASE_W1 		0x0000
#define TOTAOL_QUEUE_ENTRY_SIZE_PER_PORT	(QUEUE_ENTRY_SIZE * QUEUE_DEPTH)
#define PPP_IP_PROTOCOL				0x0021
#define IP_PROTOCOL_SHFT			16

#if (TX_PHY_MODE == SPHY_1_32)
#define DEVICE_BUS_WIDTH_RESTRICTION_MASK	0xFC
#endif

#if (TX_PHY_MODE == SPHY_4_8)
#define DEVICE_BUS_WIDTH_RESTRICTION_MASK	0xFF
#endif

#if (TX_PHY_MODE == MPHY_4)
#define DEVICE_BUS_WIDTH_RESTRICTION_MASK	0xFC
#endif

#define TCW0_SOP_BIT_LOC					9
#define TCW0_EOP_BIT_LOC					8
#define PE_SPHY_TX_FLAG_CLEAR				0
#define PE_SPHY_TX_FLAG_SET					1


#if (TX_PHY_MODE == SPHY_1_32)
#define SKIP_GET_TXR_THRESHOLD				(QUEUE_DEPTH - 8)
#endif

#if (TX_PHY_MODE == SPHY_4_8)
#define SKIP_GET_TXR_THRESHOLD				(QUEUE_DEPTH - 2)
#endif 

#if (TX_PHY_MODE == MPHY_4)
#define SKIP_GET_TXR_THRESHOLD				(QUEUE_DEPTH - 2)
#endif

#ifndef ADD_PPP_HEADER
#define TCW0_SOP_EOP_BASE_W1 				0x0000
#else
#define TCW0_SOP_EOP_BASE_W1 				0x0002
#define PPP_IP_PROTOCOL						0x0021
#define IP_PROTOCOL_SHFT					16
#define TBUF_ELE_SIZE_MINUS_18 				(TBUF_ELE_SIZE - 18)
#define TCWO_W1_SHF	16
#define TWO_TBUF_ELE_SIZE_MINUS_26 			((TBUF_ELE_SIZE * 2) - 26)
#define PPP_HEADER_SIZE	2
#endif

#define ME_INIT_SIGNAL_C	10
	extern exe_stat_flag_t exeStatFlag;

#if (TX_PHY_MODE == SPHY_1_32)
	extern __declspec(local_mem shared) q_port_t qPort[1];
	extern __declspec(local_mem shared) q_status_t qStatus[1];
#endif

#if ((TX_PHY_MODE == SPHY_4_8) || (TX_PHY_MODE == MPHY_4))
	extern __declspec(local_mem shared) q_port_t qPort[4];
	extern __declspec(local_mem shared) q_status_t	qStatus[4];
#endif

	extern SIGNAL sig_next1;
	extern SIGNAL sig_next2;
	extern SIGNAL sig_next3;
	extern SIGNAL sig_sramRead1;
	extern SIGNAL sig_msfAccess1;
	extern SIGNAL sig_msfAccess2;
	extern SIGNAL sig_msfAccess3;
	extern SIGNAL sig_reflectWrite;
	extern SIGNAL sig_commonInitComplete;
	extern SIGNAL_PAIR sig_dramAccess1;
	extern SIGNAL_PAIR sig_dramAccess2;
	extern SIGNAL_MASK sigmask_phs1Default;
	extern SIGNAL_MASK sigmask_phs2Default;
	extern same_me_signal_t ntSig1Gpr;
	extern same_me_signal_t ntSig2Gpr;
	extern same_me_signal_t ntSig3Gpr;
	extern __declspec(gp_reg) int port;
	extern __declspec(sram_read_reg) input_tx_req_t txReq;
	extern __declspec(sram_read_reg) int mpktsSent[2];
	extern unsigned int scrRingTr, addrTxSeq;
	extern unsigned int portEntryTurnaround;


#ifdef MICROC_INTEGRATION

#if (TX_PHY_MODE == SPHY_1_32)
	extern __declspec(remote dram_read_reg) int transmit_count;
#else
	extern __declspec(remote dram_read_reg) int transmit_count[16];
#endif

#else
	extern __declspec(remote dram_read_reg) int txd_p0;
	extern __declspec(remote dram_read_reg) int txd_p1;
	extern __declspec(remote dram_read_reg) int txd_p2;
	extern __declspec(remote dram_read_reg) int txd_p3;
#endif

extern void PacketTxx_Init(void);
extern void PacketTxx_RdMeta(
							input_tx_req_t *pRing,	/* IN:ring req	pointer*/
							SIGNAL *ioSig);			/* IN: Signal pointer*/

#endif	/*	__PACKETTX_H__*/
