/*******************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 1998-2002 By Intel Corporation.  All rights reserved.  
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *******************************************************************************/
/*****************************************************************************
 *
 * File:  PacketTxx_util.c
 *
 * Abstract:
 *      This is the SPHY and NPHY4 transmit code utility file.  This file 
 *      defines the utility functions which will be used in the SPHY and MPHY_4
 *		tramiting microengine C code..
 * Contents:
 *        Definitions of utility functions.
 *****************************************************************************/

#include <sphy_mphy4_tx.h>
#include <dl_meta.c>

/*****************************************************************************
 * Abstract:
 *  Clear a signal from a signal mask 
 *
 * Side Effects:
 *    None
 *
 * Assumptions:
 * 	  None
 *****************************************************************************/
INLINE  int 							/* OUT: the signal mask				 */
PacketTx_ClearSignal(			
				int io_sigmask,			/* IN: signal mask 					 */
				SIGNAL * pinSig)		/* IN: pointer to a signal			 */
{
	return(io_sigmask & ~(1 << __signal_number(pinSig)));
} /*packetTx_ClearSignal()*/

/*****************************************************************************
 * Abstract:
 *  Set a signal in a signal mask 
 *
 * Side Effects:
 *    None
 *
 * Assumptions:
 * 	  None
 *****************************************************************************/
INLINE  int							
PacketTx_SetSignal(						/* OUT: Signal mask	     			 */
			int io_sigmask, 			/* IN:	Signals mask				 */
			SIGNAL *pinSig)				/* IN:  Pointer to a signal			 */
{
	return(io_sigmask | ( 1 << __signal_number(pinSig)));
} /*packetTx_SetSignal()*/

/*****************************************************************************
 * Abstract:
 *    set  16 bits MSF registers
 *
 * Side Effects:
 *    None
 *
 * Assumptions:
 * 	  None
 *****************************************************************************/
INLINE  void							/* No return value					 */ 
PacketTx_Set16BitMsfReg(
			int inMsfRegAddr, 			/* IN: MSF register address			 */
			int inMsfRegVal)			/* IN: The value written to MSF		 */ 
{
	msf_fast_write((void *)(inMsfRegAddr | (inMsfRegVal << 16)));
} /*PacketTx_Set16BitMsfReg()*/

/*****************************************************************************
 * Abstract:
 *	Initialize the common globals in context relative gprs to save 
 *  instruction cycles.
 *
 * Side Effects:
 *    None
 *
 * Assumptions:
 * 	  None
 *****************************************************************************/
INLINE void										/* No return value 			 */
PacketTx_InitGprs()
{
#ifndef DISABLE_TX2SCHED_FEEDBACK
	sigmask_phs1Default =  __signals(&sig_next2, &sig_reflectWrite, &sig_msfAccess3);
#else
	sigmask_phs1Default =  __signals(&sig_next2, &sig_msfAccess3);
#endif

#ifdef ADD_PPP_HEADER
	sigmask_phs2Default = __signals(&sig_next3, &sig_dramAccess1, &sig_msfAccess1);
#else
	sigmask_phs2Default = __signals(&sig_next3, &sig_dramAccess1);
#endif

} /*PacketTx_init_gprs()*/


/*****************************************************************************
 * Abstract:
 *  set local memory index 0 to the tail of the queue entry for the port and 
 *  put the packet in tx request in queue tail.
 *
 * Side Effects:
 *    None
 *
 * Assumptions:
 * 	  None
 *****************************************************************************/
INLINE q_entry_t *									/* OUT: q_entry pointer	 */
PacketTx_UpdateQueueTail(
						 q_status_t *pStatus,       /* IN/OUT: Status pointer*/
						 int inTurnaround)	    	/* IN: Turn around value */
{
	q_entry_t *pTmp;
	pTmp = &(qPort[port].q_entry[(pStatus->tailOffset >> 4)]);
	if(pStatus->tailOffset == inTurnaround)
		pStatus->tailOffset = pStatus->qEntryBase;
	else
		pStatus->tailOffset = pStatus->tailOffset + QUEUE_ENTRY_SIZE;
	return(pTmp);
} /* PacketTx_SetLmindex0ToQueueTail()*/

/*****************************************************************************
 * Abstract:
 *   advance queue head after the last mpcket of one packet is sent to the port.
 *
 * Side Effects:
 *    None
 *
 * Assumptions:
 * 	  None
 *****************************************************************************/
INLINE  void										/*	No return value		 */
PacketTx_AdvanceQueueHead(
							q_status_t *pStatus,    /* IN/OUT: Status pointer*/
							int inTurnaround)		/* IN: Turnaround value. */
{

	if(pStatus->headOffset == inTurnaround)
	{
		pStatus->headOffset = pStatus->qEntryBase;
	}
	else
	{
		pStatus->headOffset += QUEUE_ENTRY_SIZE;
	}
 	pStatus->pktsInQ -= 1;
	pStatus->pktsTxed += 1;

#ifdef COUNTERS
	/*Need implement counter code here*/
#endif
} /* PacketTx_AdvanceQueueHead()*/

/*****************************************************************************
 * Abstract:
 *   notify scheduler that one packet has been transmitted from this port.
 *
 * Side Effects:
 *    None
 *
 * Assumptions:
 * 	  None
 ******************************************************************************/
INLINE void											/* No return value        */
PacketTx_NotifyScheduler(
							q_status_t *pStatus,	/* IN/OUT: Status pointer */ 
							int inPort)				/* IN: port			  	  */
{
	__declspec(sram_write_reg) int outTxdPort;

	outTxdPort = pStatus->pktsTxed;

#ifdef MICROC_INTEGRATION
#if (TX_PHY_MODE == SPHY_1_32)
			reflect_write_D(&outTxdPort,
							SCHEDULER_ME,
							&transmit_count,
							0,
							1,
					  		sig_initiator,
							sig_done,
							&sig_reflectWrite);
#else
	switch(inPort)
	{
		case 0:
			reflect_write_D(&outTxdPort,
							SCHEDULER_ME,
							&transmit_count[0],
							0,
							1,
					  		sig_initiator,
							sig_done,
							&sig_reflectWrite);
			break;

		case 1:
			reflect_write_D(&outTxdPort,
							SCHEDULER_ME,
							&transmit_count[1],
							0,
							1,
					  		sig_initiator,
							sig_done,
							&sig_reflectWrite);
			break;

		case 2:
			reflect_write_D(&outTxdPort,
							SCHEDULER_ME,
							&transmit_count[2],
							0,
							1,
					  		sig_initiator,
							sig_done,
							&sig_reflectWrite);
			break;

		case 3:
			reflect_write_D(&outTxdPort,
							SCHEDULER_ME,
							&transmit_count[3],
							0,
							1,
					  		sig_initiator,
							sig_done,
							&sig_reflectWrite);
			break;
	}
#endif

#else
	switch(inPort)
	{
		case 0:
			reflect_write_D(&outTxdPort,
							SCHEDULER_ME,
							&txd_p0,
							0,
							1,
					  		sig_initiator,
							sig_done,
							&sig_reflectWrite);
			break;

		case 1:
			reflect_write_D(&outTxdPort,
							SCHEDULER_ME,
							&txd_p1,
							0,
							1,
					  		sig_initiator,
							sig_done,
							&sig_reflectWrite);
			break;

		case 2:
			reflect_write_D(&outTxdPort,
							SCHEDULER_ME,
							&txd_p2,
							0,
							1,
					  		sig_initiator,
							sig_done,
							&sig_reflectWrite);
			break;

		case 3:
			reflect_write_D(&outTxdPort,
							SCHEDULER_ME,
							&txd_p3,
							0,
							1,
					  		sig_initiator,
							sig_done,
							&sig_reflectWrite);
			break;
	}
#endif

} /* PacketTx_NotifyScheduler()*/

/*****************************************************************************
 * Abstract:
 * 	get next available tbuf element (out_tbufele) and corresponding
 *	address (out_tbuf_addr) from data in local memory.
 *
 * Side Effects:
 *    None
 *
 * Assumptions:
 * 	  None
 *****************************************************************************/
INLINE  void										/* No return value		 */
PacketTx_GetTbufEleAndAddr(
							q_status_t *pStatus,	/*IN/OUT: Status pointer */
							int *outTbufEle,	/* OUT: TBuf element number  */
							int *outTbufAddr)	/* OUT: TBuf element address */ 
{

 	*outTbufEle = pStatus->nextAvailTbufEle;
	pStatus->nextAvailTbufEle = (pStatus->nextAvailTbufEle + 1)
									& TBUF_ELE_BITS_MASK;
	*outTbufAddr = (*outTbufEle << TBUF_ADDR_SHF) + pStatus->tbufBase;

} /* PacketTx_GetTbufEleAndAddr()*/

/*****************************************************************************
 * Abstract:
 *	move payload of sop or sop_eop mpacket to tbuf
 *
 * Side Effects:
 *    None
 *
 * Assumptions:
 * 	  None
 *****************************************************************************/
INLINE  void							/* No return value.					 */
PacketTxx_LoadDataToTbuf(
				int inBufDesc,			/* IN: Buffer descriptor			 */
				int inAddrTbuf,			/* IN: TBUffer element address		 */
			 	int inOffset,			/* IN: Data offset					 */
				int inDataLen,			/* IN: Data length				 */
				int inOffsetMod8,		/* IN: Offset node 8				 */
				SIGNAL_PAIR *inSigDram) /* IN: Pointer to signal pair		 */
{
	int tmpx;
	dram_rbuf_tbuf_ind_t indir;
	int dramAddr;

	indir.value = DRAM_INDIREF_BASE_LW;
#ifndef ADD_PPP_HEADER
	indir.buf_addr = inAddrTbuf;
#else
	indir.buf_addr = inAddrTbuf + 8;
#endif

	tmpx = inBufDesc << 2;
	dramAddr = dl_buf_get_data_from_meta(tmpx) + inOffset;
	tmpx = ((inOffsetMod8 + inDataLen -1) >> 3);
	indir.ref_count = tmpx; 
    dram_tbuf_write_ind((__declspec(dram) int *) dramAddr, 16, indir,\
						 sig_done, inSigDram);

} /* PacketTx_MoveSopPayloToTbuf()*/

/*****************************************************************************
 * Abstract:
 *	set payload_length, payload_offset, and channel fields in trnsmit control word.
 *
 * Side Effects:
 *    None
 *
 * Assumptions:
 * 	  None
 *****************************************************************************/
INLINE  int										/* OUT: New control word	 */
PacketTx_SetTcw0WithPayloLenOffsetChnl(
			int ioTcw,							/* IN: Original control word */
			int inPayloLen,						/* IN: Payload length		 */
			int inPayloOffset,					/* IN: Payload offset		 */
			int inChnl)							/* IN: Channel number		 */	
{
	return (ioTcw | (inPayloLen << TCW0_PAYLOAD_LEN_LSB_LOC)
				  | (inPayloOffset << TCW0_PAYLOAD_OFFSET_LSB_LOC)
				  | inChnl);
} /* PacketTx_SetTcw0WithPayloLenOffsetChnl()*/

/*****************************************************************************
 * Abstract:
 *	set payload_length, payload_offset in 
 *  trnsmit control word
 * Side Effects:
 *    None
 *
 * Assumptions:
 * 	  None
 *****************************************************************************/
INLINE  int									/* OUT: New control word		 */
PacketTx_SetTcw0WithPayloLenOffset(		
			int ioTcw,						/* IN: Original control word     */
			int inPayloLen,					/* IN: Payload length			 */
			int inPayloOffset)				/* IN: Payload offset			 */
{
	return( ioTcw | (inPayloLen << TCW0_PAYLOAD_LEN_LSB_LOC)
	         | (inPayloOffset << TCW0_PAYLOAD_OFFSET_LSB_LOC));
} /* PacketTx_SetTcw0WithPayloLenOffset()*/

/*****************************************************************************
 * Abstract:
 *	save sop meta data to local memory
 * Side Effects:
 *    None
 *
 * Assumptions:
 * 	  None
 *****************************************************************************/
INLINE  void										/* OUT: No return value  */
PacketTx_SaveSopMeta(
							q_entry_t *pEntry,		/* IN/OUT: entry pointer */
							q_status_t *pStatus,	/* OUT: Status pointer   */
							int inTxrLw0)			/* IN: input TX word	 */
{
	pEntry->secndBufDesc = dl_meta_get_buffer_next();
	pEntry->currBufDesc = inTxrLw0;
	pEntry->sopFlag = 1; 
	pEntry->payloadOffset = dl_meta.offset;

	if(pEntry->secndBufDesc != IX_NULL)
	{
		pEntry->rdNxtMetaFlag = 1;
	}

	pEntry->payloadLen = dl_meta.buffer_size;
	pStatus->pktsInQ++;

} /* PacketTx_SaveSopMeta()*/

/*****************************************************************************
 * Abstract:
 *  read the second and eighth longword of meta data from sram
 *
 * Side Effects:
 *    None
 *
 * Assumptions:
 * 	  None
 *****************************************************************************/
INLINE void 							/* OUT: New signal mask				 */
PacketTxx_LoadMeta(
		input_tx_req_t *pRing,			/* IN: scratch ring request	pointer	 */
		SIGNAL *ioSig)					/* IN: Signal pointer				 */
{

	dl_buf_handle_t handle;

	/*
	 *read meta data by dispatch_loop macro
	 */
	handle.value = scrRing->sopMetaHndl;
	dl_meta_load_cache(handle, ioSig, 3);
} /* PacketTx_ReadMetaDataFromSram()*/

/*****************************************************************************
 * Abstract:
 *  	Check whether tbuf is full to avoid overwritting tbuf, also utilizes
 *		defer slots to compose tcw0 for sop_eop mpkt
 *
 * return value:
 *		1: The buffer is full
 *		0: the buffer is not full
 * Side Effects:
 *    None
 *
 * Assumptions:
 * 	  None
 *****************************************************************************/
INLINE  int							/* OUT: Buffer fullness					 */
PacketTx_CheckTbufFullSopEop(	
		int * ioTcw0,				/* IN: Control word						 */
		int inPaylo,				/* IN: Payload length					 */
		int inOffsetMod8,			/* IN: Offset mode 8					 */
		int inTbufEle)				/* IN: Tbuffer lelement					 */
{

#ifndef SKIP_TBUF_FULL_CHECK_FOR_SOP_EOP_MPACKET

	int resultReg, tmpTbuf;

	resultReg = TBUF_ELE_BITS_MASK & mpktsSent[0];
	tmpTbuf = inTbufEle;
	
	/*
	 * set payload_length and payload_offset fields in io_tcw0 with in_paylo 
	 * and in_ofst_mod8
	 */
	*ioTcw0 = *ioTcw0 | ( inPaylo << TCW0_PAYLOAD_LEN_LSB_LOC)
				| (inOffsetMod8 << TCW0_PAYLOAD_OFFSET_LSB_LOC);
	if(inTbufEle < resultReg)
	{		
		tmpTbuf += TBUF_ELE_NUM_PER_SEGMENT; /*Wrap around, add tbuf element number*/
	}

	resultReg = tmpTbuf - resultReg; 	     /* Get difference*/

	if(resultReg >= TBUF_WAIT_THRESHOLD)	 /* Compare with threshold*/
	{
		return 1;
	}
	else
		return 0;
#else

	/*
	 * set payload_length and payload_offset fields in io_tcw0 with in_paylo 
	 * and in_ofst_mod8
	 */
	*ioTcw0 = *ioTcw0 | (inPaylo << TCW0_PAYLOAD_LEN_LSB_LOC)\
				| (inOffsetMod8 << TCW0_PAYLOAD_OFFSET_LSB_LOC);
	return 0;

#endif
	 
} /* PacketTx_CheckTbufFullSopEop()*/

/*****************************************************************************
 * Abstract:
 *  	Wait till tbuf not full to avoid overwritting
 *	
 *
 * Side Effects:
 *    None
 *
 * Assumptions:
 * 	  None
 *****************************************************************************/
INLINE  void								/* OUT: No return value			 */
PacketTx_WaitTillTbufNotFull( 
							int inTbufEle ) /* IN: Tbuffer element		 	 */
{
	unsigned int result;
	unsigned int tmpTbuf;

	while (1)
	{
		/*
		 *read the actual number of mpkts out of tbuf
		 */
		msf_read(&mpktsSent, (void *)addrTxSeq, 1, sig_done, &sig_msfAccess3);
		result = 0;
		while( signal_test(&sig_msfAccess3) == 0)
			result = 0;
		result = TBUF_ELE_BITS_MASK & mpktsSent[0];
		tmpTbuf = inTbufEle;		
		if(inTbufEle <  result)
		{
			tmpTbuf = tmpTbuf + TBUF_ELE_NUM_PER_SEGMENT;
		}
		result = tmpTbuf - result;
		if( result < TBUF_WAIT_THRESHOLD)
			return;
	}
} /* PacketTx_WaitTillTbufNotFull() */

/*****************************************************************************
 * Abstract:
 *  	Check whether tbuf is full to avoid overwritting tbuf
 *	
 *
 * Side Effects:
 *    None
 *
 * Return:
 *		0-- bufer not full
 *		1-- buffer is full
 * Assumptions:
 * 	  None
 *****************************************************************************/
INLINE int									   /* OUT: Tnbuffer fullness	 */
PacketTx_CheckTbufFullNotSopEop(
								int inTbufEle) /* IN: Tbuffer lelment		 */
{
	unsigned int resultReg;
	unsigned int tmpTbuf;

	resultReg = TBUF_ELE_BITS_MASK & mpktsSent[0];
	tmpTbuf = inTbufEle;

	if(inTbufEle < resultReg)
	{
		tmpTbuf = tmpTbuf + TBUF_ELE_NUM_PER_SEGMENT; /* wrap around*/
	}

	resultReg = tmpTbuf - resultReg; 	/*compare with totals in tbuf*/

	if(resultReg < TBUF_WAIT_THRESHOLD)	/* compare with threshold */
	{
		return 0;
	}
	else
	{
		return 1;
	}

} /* PacketTx_CheckTbufFullNotSopEop()*/

/*****************************************************************************
 * Abstract:
 *  	update ab_paylo_rmnd and ab_offset_rpaylo in queue entry 
 *	
 *
 * Side Effects:
 *    None
 *
 * Assumptions:
 * 	  None
 ******************************************************************************/
INLINE  void										/* OUT: no return value   */
PacketTx_UpdateAbPayloRmndAndOffsetRpaylo(
								q_entry_t *pEntry,  /* IN/OUT: Entry pointer  */
								int ioMpktPayloLen)	/* IN: Payload lenth	  */	
{
	int nextMpktLen, tmpx;

#ifndef AVOID_USING_SKIP_BIT_IN_TCW0

	pEntry->payloadOffset += ioMpktPayloLen;
	pEntry->payloadLen -= ioMpktPayloLen;

#else

	/*
	 * adjust current mpkt size to make sure next mpkt in this buffer always have
	 * more than 4 bytes payload, so SKIT bit never set in TCW0, this only happens
	 * in SPHY_1x32 mode
	 */  
	tmpx = pEntry->payloadLen;
	nextMpktLen = tmpx - ioMpktPayloLen;
	if(nextMpktLen < BUS_WIDTH_IN_BYTES]
	{
		ioMpktPayloLen -= BUS_WIDTH_IN_BYTES;
	}
	pEntry->payloadOffset += ioMpktPayloLen;
	pEntry->payloadLen -= ioMpktPayloLen;

#endif
} /* PacketTx_UpdateAbPayloRmndAndOffsetRpaylo()*/

/*****************************************************************************
 * Abstract:
 *  	save secodary buffer data to local memory
 *	
 *
 * Side Effects:
 *    None
 *
 * Assumptions:
 * 	  None
 ******************************************************************************/
INLINE void										/* OUT: No return value		  */
PacketTx_SaveSbdMeta(	
					q_entry_t *pEntry,			/* OUT: Entry pointer     	  */
					q_status_t *pStatus)		/* OUT: Status pointer 	      */ 
{
	pStatus->secndBufDesc = (dl_meta_get_buffer_next() & 0xFFFFFF);
	pEntry->secndBufSize = dl_meta.buffer_size;
	if(pStatus->secndBufDesc != IX_NULL)
	{
		pEntry->rdScndMetaFlag = 1;
	}
} /* PacketTx_SaveSbdMeta()*/

/*****************************************************************************
 * Abstract:
 *  	 replace active buffer data with secondary buffer data in  
 *       queue entry
 *	
 *
 * Side Effects:
 *    None
 *
 * Assumptions:
 * 	  None
 ******************************************************************************/
INLINE  void									/* OUT: No return value		  */
PacketTx_SecondaryBufferBecomeActiveBuffer(
									q_entry_t *pEntry, /* OUT: Entry pointer  */
									q_status_t *pStatus)/* OUT: Status pointer*/
{
	pEntry->value[0] = pEntry->value[2];
	pEntry->value[1] = pEntry->value[3];
	pEntry->value[2] = pStatus->secndBufDesc;
	pEntry->value[3] = 0;
	pStatus->secndBufDesc = IX_NULL;
} /* PacketTx_SecondaryBufferBecomeActiveBuffer()*/

/*****************************************************************************
 * Abstract:
 *  	 move payload of not_sop mpacket to tbuf 
 *	
 *
 * Side Effects:
 *    None
 *
 * Assumptions:
 * 	  None
 ******************************************************************************/
INLINE  void									/* OUT: No return value		  */
PacketTx_MoveNotSopPayloToTbuf(
	in_parameter_1_t pPara)						/* IN: parametr pointer       */
{
	dram_rbuf_tbuf_ind_t indir;
	int tmpx;
	int dramAddr;

	indir.value = DRAM_INDIREF_BASE_LW;
	indir.buf_addr = pPara.inAddrTbuf;
	indir.ref_count =(((pPara.inTcw0 >> TCW0_PAYLOAD_LEN_LSB_LOC) +
					 ((pPara.inTcw0 >> TCW0_PAYLOAD_OFFSET_LSB_LOC) &
					 TCW0_PAYLOAD_OFFSET_MASK) - 1) >> 3);
	tmpx = pPara.inBufDesc << 2;
    dramAddr = dl_buf_get_data_from_meta(tmpx);
    dram_tbuf_write_ind((__declspec(dram) int *) (dramAddr + pPara.inOffset),
						 16, indir, sig_done, pPara.pInSigDram);
} /* PacketTx_MoveNotSopPayloToTbuf()*/

/*****************************************************************************
 * Abstract:
 * do necessary bytes alignment with offset, so the leftover 
 *              will be realigned in first longword without any offset 
 *              to the first byte of the long word
 *	
 *
 * Side Effects:
 *    None
 *
 * Assumptions:
 * 	  None
 ******************************************************************************/
INLINE long int											/* OUT: Aligned long word */
PacketTx_LeftOverBytesAlignment(
				int inOffset,							/* IN: Offset			  */
				int inLw[4])							/* IN: Data array		  */
{
	int offset, tmp;

	if( inOffset == 4)
	{
		return(inLw[1]);
	}
	else
	{
		if(inOffset > 4)
		{
			offset = inOffset - 4;
			byte_align_block_be(2, &tmp, &inLw[1], offset);
		}
		else
		{
			byte_align_block_be(2, &tmp, &inLw[0], offset);
		}
	}
	return(tmp);
} /* PacketTx_LeftOverBytesAlignment()*/

/*****************************************************************************
 * Abstract:
 * free buffer handle by dispatch loop macro dl_buf_free
 *	
 *
 * Side Effects:
 *    None
 *
 * Assumptions:
 * 	  None
 ******************************************************************************/
INLINE void											/* OUT: No return value   */
PacketTx_FreeBuffer( int inBufHandle)				/* IN: Buffer hadle		  */
{

	dl_buf_handle_t handle;
			
#ifndef SKIP_FREE_BD

	/*
	 *use dl_buf_free in dl_buf.uc in dispatch_loop to free buffer handle
	 */
	handle.value = inBufHandle;
	dl_buf_free(handle, BUF_FREE_LIST0, BUF_POOL);

#endif

} /* PacketTx_FreeBuffer()*/