////////////////////////////////////////////////////////////////////////////////
//                                                                     
//                  I N T E L   P R O P R I E T A R Y                   
//                                                                      
//     COPYRIGHT (c)  2001-2002 BY  INTEL  CORPORATION.  ALL RIGHTS          
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//     THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                      
//                        INTEL  CORPORATION                            
//                                                                     
//                     2200 MISSION COLLEGE BLVD                        
//                                                                      
//               SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                      
////////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: sphy_mphy4_tx.uc
//
//      Purpose: 1st TX ME in Packet transmitter for Egress IXP2800
//
///////////////////////////////////////////////////////////////////////////////


#ifndef	__SPHY_MPHY4_TX1_UC__
#define	__SPHY_MPHY4_TX1_UC__

// 
// This SPHY/MPHY4 packet (POS/Ethernet) transmitter (TX) runs on POS-PHY 
// SPHY/MPHY4 trnsmit mode on the egress IXP2xxx. POS SPHY mode can be 
// configured as 1x32 This microblock supports only SPHY_1x32(one OC192)
// , which can be selected through "#define TX_PHY_MODE 
// SPHY_1_32",  in dl_system.h of the project. This microblock runs on 2 MEs.
// 
// In the egress pipeline, there is L2-encap block which added approriate 
// PPP or Ethernet header before tx request is sent to this microblock, 
// so this microblock can handle both POS and Ethernet transmission. For
// POS, the min packet (49 bytes long) is the critical case. 
//
// In order to avoid overflooding the packet queue in the local memory, this
// microblock reads NN ring to get tx request for a specific port only 
// when there are spaces in the packet queue for that port. It also keeps 
// account of how many packets transmitted for each port, and update pkts_txd 
// count in Egress Scheduler through a reflect write when it finishs the 
// transmission of one packet, so Egress Scheduler will not overflood it's 
// downstream. 
//
// Following are the assumptions used to develop this microblock:
// 1. The meta data at SRAM and the L2 header and payload at DRAM are correct,
//    this microblock will not do any sanity check in the size and order of 
//    the buffers in any packet.
// 2. The L2 header and payload in DRAM will be contiguous from the offset to 
//    the end of a buffer in a packet.
// 3. The offset for all buffers in a packet except SOP buffer is 0.
//
// There are three issues need to be addressed in this microblock. The
// first two are imposed by IXP2xxx hardware, the third one keeps this 
// microblock capable to handle the packets received in diffrent receiving 
// interface in (POS, ATM, Ethernet) in ingress microblocks. They are
// 1. DRAM[tbuf_wr,...] instruction, which is used to move packet data from 
//    DRAM to TBUF, always access DRAM and TBUF on 8-bytes word boundary.
// 2. The sum of the prepend length and payload length of a mpacket must be an
//    integral multiple of the bus width, except for a EOP mpacket. In SPHY
//    1X32 and MPHY4 mode, the bus width is 32 bits (4 bytes).
// 3. The offset of L2 header and  payload data offset in the first buffer of 
//    a packet can be any value. 
// 
// In order to address the abovementioned three issues, POS TX transmits the
// maximum allowable payload bytes for every non-eop mpacket by making sure 
// that the sum of the prepend length and payload length of a non-eop mpacket
// is an integral multiple of device bus width. If necessary, maximum of 3 
// bytes at the  end of one buffer is saved in the local memory and transmit 
// with the payload of the next buffer of the same packet in the next mpacket.
//
// This microblock keeps active buffer (which the microblock is 
// processing) descriptor info and secondary buffer (buffer descriptor for the 
// next buffer after the active buffer in the packet, it is in the software 
// next pointer field of meta data of the active buffer) descriptor info in 
// local memory, this microblock will read the necessary info (buffer size, 
// data offset, freelist, and the software next pointer) of the secondary 
// buffer from sram in parallel with it's processing of the active buffer. When
// the processing of the active buffer finishes, the secondary buffer will 
// become the active buffer, and continue until all buffers in the packet are 
// processed.
//  
// The packet Tx works on two or three MEs configurations using context 
// pipeline. If the minimum packet size instruction cycle budget is tight and
// the media port TX FIFO transient is relative large, then thee MEs 
// configuration is recommended, otherwise two MEs configuration is 
// recommended. In both two MEs and three MEs configurations, the first ME does
// some processing on packets and sends request to next ME about each mpacket 
// to transmit. In the two MEs configuration, the second ME moves prepend and 
// payload to tbuf and validate the tbuf, and can perform flow control functon
// based on media port TX FIFO status to prevent port TXFIFO overwritting 
// optionally. In the three MEs, the second ME moves prepend and payload to 
// tbuf, and the third ME validates the tbuf, and can perform flow control 
// functon based on media port TX FIFO status to prevent port TXFIFO 
// overwritting optionally. On higher level, all packet-level processing is 
// done on first ME and all mpacket-level in the second or second/third ME. 
/*
1st ME:
	
while (1)
	{
		read request (packet p1) from NN ring from QM.
		Issue SRAM read for metadata of the p1.
		Find first request in LM queue (packet p2) and read the request
		Determine whether the packet is SOP, EOP, none of these or a min packet and call corresponding macro.
		Sends scheduler feedback about packet transmit when reaches EOP
		Calculate DRAM address to read the payload
		Send the request to next ME using NN to transmit a mpacket of p2

		Wait for metadata read signal and int thd signal
		
		Send int thd signal
		Store the metadata read to the tail of LM queue
	}

2nd ME (in two MEs configuration):

while (1)
	{
		for (i=0; I < max_phase; i++)
			{
				phase# i:
					issue tbuf validate of mpacket read in previous iteration and generate signal tbuf_validate_i
					wait for signals tbuf_validate_i, dram_(i+1), prepend_(i+1)
					read a mpacket transmit request from NN ring
					grab a tbuf for the mapacket
					issue dram to tbuf transfer request and generate signal dram_i
					perform flow control function based on port TXFIFO status optionally
					add prepend (l2 header, SPI-4 prepend or leftover) and generate signal prepend_i
					branch to phase_(i+1);
			}
			// here (i+1) means (i+1) mod max_phase;

	}

		
2nd ME (in three MEs configuration):

while (1)
	{
		for (i=0; I < max_phase; i++)
			{
				phase# i:
					issue request for mpacket read in previous iteration to next ME through NN ring
					wait for signals dram_(i+1), prepend_(i+1)
					read a mpacket transmit request from NN ring
					grab a tbuf for the mapacket					
					issue dram to tbuf transfer request and generate signal	dram_i
					add prepend (l2 header, SPI-4 prepend or leftover) and generate signal prepend_i
					branch to phase_(i+1);
			}
			// here (i+1) means (i+1) mod max_phase;

	}


3rd ME (only in three MEs configuration):

while (1)
	{
		for (i=1; I < 3; i++)
			{
				phase# i:
					read mpacket request from NN ring 
					perform flow control function based on port TXFIFO status optionally
					issue tbuf validate of mpacket just read from NN ring
					branch to phase_(i+1);
			}

	}

The interface between 1st ME and 2nd ME is

lw0: meta_data header ptr (used to free a buffer, if buffer is transmitted completeely)
lw1: 0-29 	offset of payload to be transmitted
30	bit clear means the mpacket is eop,
31	bit clear means the mpacket is sop
lw2: 	payload size to be transmitted
 
In case of non-sop mpacket, we have 3 or 6 more LWs on the NN ring, 
The first three more LWs are:

lw3:	leftover_lw0, first longword of leftover from the previous buffer has 
		to be prepended to the current buffer
lw4:	exe_stat_flag  includes information about various condition flags
lw5:	partially created transmit control word 

If LEFTOV_FROM_PB_FLAG_BIT is set in exe_stat_flag, there will be three more 
LWs:

lw6:	leftover_lw1, second longword of leftover from the previous buffer has 
		to be prepended to the current buffer
lw7:	leftover_lw2, third longword of leftover from the previous buffer has 
		to be prepended to the current buffer
lw8:	leftover_lw3, forth longword of leftover from the previous buffer has 
		to be prepended to the current buffer

The interface between 2nd ME and 3rd ME is

lw0: meta_data header ptr (used to free a buffer, if buffer is transmitted completeely)
lw1: transmit control word 0 for this mpacket
lw2: bit 30 - EOB flag: when set, this buffer is completely transmitted, should be freed
	 bit 15 to 0: tbuf element number assigned to this mpacket in 2nd ME

*/


///////////////////////////////////////////////////////////////////////////////

// Following definition enable this microblock waiting for an inter-ME signal 
// generated by the ME whcih initializes common resources, such as scratch ring,
// near the end of thread 0 of this microblockinitialization to make sure all 
// common resource are available. This option should be enabled when this POS TX 
// microblock is integrated with other egress microblocks.  
#define WAIT_FOR_COMMON_RESOURCE_INITIALIZATION 	


// following definition selects sram transfer registers to be used with 
// dl_meta.uc dispatch loop
#define DL_META_DATA_IN_SXFER

// Following are the optional assembling options for this TX ME:

// ADD_L2_HEADER: enable this microblock to add layer 2 header, it combines 
// with POS_TX option to add PPP header before payload, it combines with
// ETHERNET_TX option to add Ethernet header before payload

// POS_TX: combines with ADD_L2_HEADER option enable this microblock to add
// PPP header before payload

// ADD_PPP_ADDR_AND_CNTRL: combines with ADD_L2_HEADER and POS_TX options 
// enable this microblock add  PPP address and control fields with PPP header 
// before payload

// ETHERNET_TX: combines with ADD_L2_HEADER option enable this microblock to
// add Ethernet header before payload

// DISABLE_TX2SCHED_FEEDBACK: disable the feedback to SCHEDULER microblock. 
// this microblock will update a remote dram transfer register which keeps
// account of the packets transmitted by this TX ME to prevent overflooding 
// the local memory packet queue in this TX ME by default. When there is no 
// need to feedback (e.g., no SCHEDULER in the pipeline), this option can be
// set to save three instruction for each packet

// SCHEDULER_ME: when "DISABLE_TX2SCHED_FEEDBACK" option is not set, this 
// TX ME need the ME location of SCHEDULER for it to do feedback through
// a reflector write, e.g. SCHEDULER_ME=0x06

// DISABLE_SNDB_RD_LOCK_CHECK: When ((BUFFER_SIZE - MAX_OFFSET) / 
// TBUF_ELEMENT_SIZE) is greater than 8 (total threads per ME), then it is
// sure that the info of the next buffer will be available before it is 
// needed, then this option can be set to save one instruction per iteration

// DISABLE_QUEUE_FULL_CHECK: This TX ME check how many packets in it's local
// memory packet queue and stop to get tx request from NN ring, when the packet
// in queue reaches a threshold to prevent packet queue overflow by default. If
// there is other mechanism can prevent packet queue overflow in place, e.g.,
// MAX_IN_FLIGHT in SCHEDULER, then this option can be set to save two 
// instructions for each iteration.


// include stdmac.uc in IXPblocks Portable library

#include <stdmac.uc>

// include local memory macros in IXPblocks 

#include <localmem.uc>

// global definition of common constants which are shared by all microblocks

#include "dl_system.h"

// definition of hardware register address 

#include <hardware.h>

// definition of PHY operation mode and misc.

#include <definitions.h>

// definition of constants used

#include "sphy_mphy4_tx.h"

// definition of utilization macros used

#include "sphy_mphy4_tx_util.uc"

// initialization code for POS/Ethernet TX

#include "sphy_mphy4_tx_init.uc"

// definition of common utilization macros used

#include "sphy_mphy4_tx_common_util.uc"

// dispatch loop macros

#include "dispatch_loop.uc"


.reg addr_tx_seq		; MSF register address to read mpkt txed out of tbuf
// context relative GPR variables:

// following context relative GPR variables are preassigned in initialization 
// to save instrction cycles in POS TX process
.reg sring_tr           		; scratch ring which QM put tx requests
.reg sig1_next_context_gpr		; value to write to SAME_ME_SIGNAL csr to wake up
								; next thread
.reg sig3_next_context_gpr		; value to write to SAME_ME_SIGNAL csr to wake up
								; next thread
.reg port_entry_turnaround		; packet queue head/tail offset value to reset 
								; head/tail offset
.reg $txed_port					; xfer register to keep packet txed to notify
								; egress scheduler by reflect write	


///////////////////////////////////////////////////////////////////////////////

//	Global Signals
.sig sig_scratch_access_1 			; signal for scratch ring read
// following three signals need to be declared as volatile
.sig volatile sig1_next_context		; signal to wakeup next thread, declared as
									; volatile to eliminate assembling warning

.sig volatile sig3_next_context		; signal to wakeup next thread, declared as
									; volatile to eliminate assembling warning

.sig volatile sig_reflector_write		; signal for reflector write to feedbeck 
									; to scheduler
									; volatile to eliminate assembling warning
.sig sig_sram_read_1 				; signal 1 for sram read
.sig sig_sram_read_2 				; signal 2 for sram read
.sig sig_msf_access_1				; signal 1 for msf access (read or write)

.sig sig_dram_access_1				; signal 1 for dram access (dram[tbuf_wr..])
.sig sig_dram_access_2				; signal 2 for dram access (dram[tbuf_wr..])
.sig volatile common_init_complete_sig_num		; this is the inter-ME signal 
										; generated by the ME whcih initializes
										; common resources, such as scratch ring,
										; POS TX MEs will wait for this signal at
										; end of their initialization to make 
										; sure all common resource are available 
.sig sig_l2_table_entry_read			; signal for reading l2 table entry

//	Manual signal allocation
.addr common_init_complete_sig_num	ME_INIT_SIGNAL	; manual signal allocation,
													; ME_INIT_SIGNAL defined as
													; in dl_system.h  

#ifdef DEBUG_TX_PENDING_LOCKED
.sig sig_sram_write_1 				; signal 1 for sram write
#endif //#ifdef DEBUG_TX_PENDING_LOCKED

#ifdef COUNTERS
.sig sig_counter_update									; signal for update counter
#endif //#ifdef COUNTERS		

#ifdef _DEBUG_COUNTERS_
.reg @packets_received 
.reg @requests_sent

alu[@packets_received, --, b, 0]
alu[@requests_sent, --, b, 0]

#endif	// #ifdef _DEBUG_COUNTERS_


///////////////////////////////////////////////////////////////////////////////
// _sphy_mphy4_tx_sop_eop_mpacket()
//
// Description: handle the packet which can be fitted into one mpacket 
//				dequeue one packet
//
// Outputs:					
//							None
//
// Inputs: 
//		in_payload			payload length
//      in_offset			payload offset
//		in_pkt_type			packet type [Needed only for PPP]
//		in_offset_mod_8		remaider of offset divided by 8
//		PHASE1_START_LABEL	symbolic address for start of inifinite TX 
//							processing loop
//		META_DATA_READ		a value of 1 means a packet was read from NN ring
//							and this macro has to wait for metadata read
//
// Constants:
//						 	None 
//
//		
//
///////////////////////////////////////////////////////////////////////////////
#macro _sphy_mphy4_tx_sop_eop_mpacket(in_payload, in_offset, in_pkt_type, \
		in_offset_mod_8, PHASE1_START_LABEL, META_DATA_READ)
	.begin
	.reg eob_bd
		
		// advanced queue head, so next packet in the queue of this port will be 
		// processed when this port is selected again
		_sphy_mphy4_tx_advance_queue_head(port_entry_turnaround)	

	//end of phase 1  -- phase 1 does not have any ctx_arb in the end

	sop_eop_mpkt_phase2#:

	// beginning of phase 2

	.begin 
	sop_eop_end_of_phase_2#:
		// end of phase 2
		// swap out to wait signals specified in sig_mask_2

		// wait for meta data read only if meta data is read
	
	#ifdef DISABLE_TX2SCHED_FEEDBACK
	
	#ifndef ADD_L2_HEADER
		#if (META_DATA_READ == 1)
				ctx_arb[sig_sram_read_1, sig3_next_context]
		#else
				ctx_arb[ sig3_next_context]
		#endif
	#else	// ADD_L2_HEADER
		#ifdef POS_TX
			#if (META_DATA_READ == 1)
				ctx_arb[sig_sram_read_1, sig3_next_context]
			#else
				ctx_arb[ sig3_next_context]
			#endif
		#endif	// #ifdef POS_TX

		#ifdef ETHERNET_TX
			#if (META_DATA_READ == 1)
				ctx_arb[sig_sram_read_1, sig_l2_table_entry_read, \
													sig3_next_context]
			#else
				ctx_arb[sig_l2_table_entry_read, sig3_next_context]
			#endif
		#endif	// #ifdef ETHERNET_TX
	#endif // #ifdef ADD_L2_HEADER

	#else	// TX2SCHED_FEEDBACK

		// wait for meta data read only if meta data is read
	#ifndef ADD_L2_HEADER
		#if (META_DATA_READ == 1)
				ctx_arb[sig_sram_read_1, sig_reflector_write, sig3_next_context]
		#else
				ctx_arb[sig_reflector_write, sig3_next_context]
		#endif
	#else	// ADD_L2_HEADER
		#ifdef POS_TX
			#if (META_DATA_READ == 1)
				ctx_arb[sig_sram_read_1, sig_reflector_write, sig3_next_context]
			#else
				ctx_arb[sig_reflector_write, sig3_next_context]
			#endif
		#endif	// #ifdef POS_TX

		#ifdef ETHERNET_TX
			#if (META_DATA_READ == 1)
				ctx_arb[sig_sram_read_1, sig_l2_table_entry_read, \
										sig_reflector_write, sig3_next_context]
			#else
				ctx_arb[sig_l2_table_entry_read, sig_reflector_write, sig3_next_context]
			#endif
		#endif	// #ifdef ETHERNET_TX
	#endif // #ifdef ADD_L2_HEADER
	#endif	// #ifdef DISABLE_TX2SCHED_FEEDBACK

		sop_eop_mpkt_phase3#:
	
		// beginning of phase 3	in sop_eop mpacket path
		// wake up next thread in the very beginning of phase 3 to reduce wakeup 
		// latency
		local_csr_wr[SAME_ME_SIGNAL, sig3_next_context_gpr]

	
		// write the tx request to the next ME

			waiting_for_nn_to_become_not_full#:
			br_inp_state[nn_full, waiting_for_nn_to_become_not_full#]
		
			_sphy_mphy4_tx_debug_incr_counter(@requests_sent)

			alu[*n$index++, *l$index0[ABD_0_OFFSET], and~, 0xf, <<24] 	; meta data header ptr
			#if (META_DATA_READ == 1)
				// set the lm memory to tail pointer to store the last packet
				local_csr_wr[active_lm_addr_0, current_tail]
			#endif
		#ifndef ADD_L2_HEADER
			alu[*n$index++, --, b, in_offset]		; bit 30 = 0 means eop, bit 31 = 0 means sop
		#else	// ADD_L2_HEADER
			#ifdef POS_TX
				alu[*n$index++, --, b, in_offset]		; bit 30 = 0 means eop, bit 31 = 0 means sop
			#endif	// #ifdef POS_TX
			#ifdef ETHERNET_TX
			.begin
				.reg tmp
				alu[tmp, --, b, in_offset]
				br=byte[l2_header_exist, 0, L2_HEADER_IN_PAYLOAD, \
													l2_header_in_paylo#]
				alu[tmp, tmp, or, 1, <<29]		; bit 29 = 1 means l2 entry passed to next NE
			l2_header_in_paylo#:
				alu[*n$index++, --, b, tmp]		; bit 30 = 0 means eop, bit 31 = 0 means sop
			.end
			#endif	// #ifdef ETHERNET_TX
		#endif // #ifndef ADD_L2_HEADER
		#ifdef POS_TX
			#ifdef DYNAMIC_PPP_ENCAP
				alu_shf[*n$index++, in_payload, OR, in_pkt_type, <<16]	;Dynamic PPP Encap
			#else
				alu[*n$index++, --, b, in_payload]
			#endif
		#else
			alu[*n$index++, --, b, in_payload]
		#endif

		#ifdef ADD_L2_HEADER
			#ifdef ETHERNET_TX
			
			// pass l2 entry to next ME
			
			_sphy_mphy4_tx_pass_l2_entry_to_next_me(l2_header_exist)

			#endif	// #ifdef ETHERNET_TX
		#endif // #ifdef ADD_L2_HEADER

	.end 	
	
	// if meta data was read, store it in LM queue

	#if (META_DATA_READ == 1)

			// meda data for sop is ready, move them to local memory
			_sphy_mphy4_tx_save_sop_meta_to_lm(tx_request_lw0) 
	#endif

		br[PHASE1_START_LABEL]

	.end 

	// end of sop_eop mpacket processing code
#endm // end of #macro _sphy_mphy4_tx_sop_eop_mpacket()

///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// _sphy_mphy4_tx_get_tx_request_for_no_tx()
//
// Description: handle no port is ready to transmit case, check and handle 
//				valid tx request
//
// Outputs:					
//							None
//
// Inputs:  
//							None
//
// Constants:
//						 	None 
//
//	Labels:
//		PHASE1_START_LABEL	symbolic address for start of inifinite TX 
//							processing loop
/////////////////////////////////////////////////////////////////////////////////
#macro _sphy_mphy4_tx_get_tx_request_for_no_tx(PHASE1_START_LABEL)
	
	.begin
	
	// set the lm memory to tail pointer to store the last packet		
	local_csr_wr[active_lm_addr_0, current_tail]
		
		// end of phase 1
		// swap out to wait signals specified in sig_mask_2
		local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask_2]	; set active_ctx_wakeup_events
															; csr
		ctx_arb[--]											; ctx_arb on sig_mask
		// following .io_complete is used to eliminate assembling warning due to the
		// usage of "ACTIVE_CTX_WAKEUP_EVENTS" csr
		.io_completed sig3_next_context
		.io_completed sig_sram_read_1
		#ifndef DISABLE_TX2SCHED_FEEDBACK
		.io_completed sig_reflector_write
		#endif	// #ifndef DISABLE_TX2SCHED_FEEDBACK

		// end of phase 2 of no port ready to transmit
	
		// beginning of phase 3 of no port ready to transmit
		// wake up next thread in beginning of phase 3 to reduce wakeup latency

		local_csr_wr[SAME_ME_SIGNAL, sig3_next_context_gpr]

		// set $w1 to deq_port for next run is sop_eop case (POS min packet 
		// case) to save one instruction in that critical case
		// workaround for HW-bug#1249, copy channel # into tcw1 (reserved in PRM)
		alu[$w1, --, B, deq_port]	


		// if didn't read meta data in phase 2, just branch out
		br_bclr[exe_stat_flag, READ_SOP_BD_FLAG_BIT, no_port_rdy_cont_1#] 

		// meda data for sop is ready, move them to local memory
		_sphy_mphy4_tx_save_sop_meta_to_lm(tx_request_lw0) 

	no_port_rdy_cont_1#:
		// swap out to wait for my turn
		br[PHASE1_START_LABEL]
		// end of phase 3 of no port ready to transmit

		
	.end // sig_mask_2
	// end of not ready for transmission code
#endm // end of macro _sphy_mphy4_tx_get_tx_request_for_no_tx()


///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// _sphy_mphy4_tx_sop_only_non_critical_mpkt()
//
// Description: set local memory index 0 to the head of the queue entry to 
//				dequeue one packet
//
// Outputs:					
//							None
//
// Inputs:  
//		in_payload			payload length
//      in_offset			payload offset
//		in_pkt_type			needed only for PPP
//		in_offset_mod_8		remaider of offset divided by 8
//
// Constants:
//						 	None 
//
//	Labels:
//		PHASE1_START_LABEL	symbolic address for start of inifinite TX 
//							processing loop
/////////////////////////////////////////////////////////////////////////////////
#macro _sphy_mphy4_tx_sop_only_non_critical_mpkt(in_payload, in_offset, in_pkt_type, \
		in_offset_mod_8, PHASE1_START_LABEL)

	// beginning of sop_only_non_critical mpacket processing code
	// sop_only_non_critical_mpacket path in middle of phase 1
	 .begin
	.reg  sop_bd rmnd_ofset dram_addr indir_ref enq_port sop_paylo tmp

		// reset p_sop_flag in queue entry
		alu_shf[*l$index0[ABD_0_OFFSET], *l$index0[ABD_0_OFFSET], and~, 1, \
																<<SOP_BIT_LOC]

		// following .set used to eliminate assembling warning
		.set enq_port   

		// check whether to read secondary BD 
		xbuf_alloc[$dl_meta_sec, 3, read]

		.set $dl_meta_sec[0] $dl_meta_sec[1] $dl_meta_sec[2]		

		alu[tmp, --, b, *l$index0[ABD_0_OFFSET]]
	    br_bclr[tmp, RD_NBD_BIT_LOC, sop_only_non_critical_rd_sbd_done#]
		

		 // read secondary buffer meta data from SRAM 
		_sphy_mphy4_tx_read_sb_meta_data_from_sram($dl_meta_sec, sig_mask_2, \
															*l$index0[SBD_0_OFFSET]) 
		// reset read secobary buffer meta data bit 
		alu_shf[*l$index0[ABD_0_OFFSET], *l$index0[ABD_0_OFFSET], and~, 1, \
																<<RD_NBD_BIT_LOC]
		// set read secondary BD flag in exe_stat_flag for use in phase 2
		alu_shf[exe_stat_flag, exe_stat_flag, or, 1, <<READ_SND_BD_FLAG_BIT]	; set flag

		// check whether this port should be blocked from transmission for any following
		// threads which serve the same in the same run to prvent them accessing  the 
		// secondary buffer meta data (which is not ready yet in this phase, but will be 
		// ready in next phase) when they become active buffer
		// calculate the max bytes to consume before secondary buffer meta data need to be
		// ready
		alu_shf[tmp, --, b, TBUF_ELE_SIZE, <<ME_THREAD_NUM_SHFT] ; max bytes required
		alu[--, in_payload, -, tmp]			; compare with how much left
		bgt[sop_only_non_critical_rd_sbd_done#]
		// set *l$index1[GLOBAL_NOT_LOCK_SNDB_RD_FLAG_INDEX] to 0 to prvent any threads
		// handling the same port in the same run to access the secondary buffer data 
		// (which is not ready yet) when it becomes active buffer, set 
		// put *l$index1[GLOBAL_NOT_LOCK_SNDB_RD_FLAG_INDEX] to 1
		// when secondary meta data saved in phase 2.
		alu[*l$index1[GLOBAL_NOT_LOCK_SNDB_RD_FLAG_INDEX], --, B, 0]

	sop_only_non_critical_rd_sbd_done#:
	  	// get sop_bd, offset_rpaylo, and offset_mod_8 for use in phase 2
	    ld_field_w_clr[sop_bd, 0111, *l$index0[ABD_0_OFFSET]]
	
		// now consider the restriction of the limitation that the sum of the prepend
		// length and payload length must be multiple of device bus width
	#ifndef ADD_L2_HEADER
		#ifdef SPI4_PREPEND
		.begin
		.reg allow_total
			alu[sop_paylo, SOP_EOP_MAX_PAYLOAD, -, offset_mod_8]
			alu[allow_total, sop_paylo, +, SPI4_PREPEND_HEADER_SIZE]
			alu[allow_total, allow_total, and, DEVICE_BUS_WIDTH_RESTRICTION_MASK]
			alu[sop_paylo, allow_total, -, SPI4_PREPEND_HEADER_SIZE]
		.end // allow_total
		#else
			alu[sop_paylo, TBUF_ELE_SIZE, -, offset_mod_8]
			alu[sop_paylo, sop_paylo, and, DEVICE_BUS_WIDTH_RESTRICTION_MASK]
		#endif
	#else // ADD_L2_HEADER
	#ifdef POS_TX
	.begin
	.reg allow_total
		alu[sop_paylo, SOP_EOP_MAX_PAYLOAD, -, offset_mod_8]
		alu[allow_total, sop_paylo, +, PPP_HEADER_SIZE]
		alu[allow_total, allow_total, and, DEVICE_BUS_WIDTH_RESTRICTION_MASK]
		alu[sop_paylo, allow_total, -, PPP_HEADER_SIZE]
	.end //allow_total
	#endif	// #ifdef POS_TX
	#ifdef ETHERNET_TX
	.begin
	.reg allow_total
		br=byte[l2_header_exist, 0, L2_HEADER_IN_PAYLOAD, l2_in_paylo#]
		alu[sop_paylo, SOP_EOP_MAX_PAYLOAD, -, offset_mod_8]
		alu[allow_total, sop_paylo, +, ETHERNET_HEADER_SIZE]
		alu[allow_total, allow_total, and, DEVICE_BUS_WIDTH_RESTRICTION_MASK]
		alu[sop_paylo, allow_total, -, ETHERNET_HEADER_SIZE]
		br[sop_paylo_size_done#]
	l2_in_paylo#:
		alu[sop_paylo, TBUF_ELE_SIZE, -, offset_mod_8]
		alu[sop_paylo, sop_paylo, and, DEVICE_BUS_WIDTH_RESTRICTION_MASK]
	sop_paylo_size_done#:
	.end //allow_total
	#endif	// #ifdef ETHERNET_TX
	#endif //#ifndef ADD_L2_HEADER

	
		// update ab_paylo_rmnd and ab_offset_rpaylo in queue entry with 
		// sop_only_non_critical_paylo [3]
		_sphy_mphy4_tx_update_ab_paylo_rmnd_and_offset_rpaylo(sop_paylo)

	#ifdef ADD_L2_HEADER
		#ifdef ETHERNET_TX

		// add signal for l2 table entry read

		_sphy_mphy4_tx_set_signal[sig_mask_2, sig_l2_table_entry_read]
		#endif	// #ifdef ETHERNET_TX
	#endif // #ifdef ADD_L2_HEADER

	 	// swap out to wait signals specified in sig_mask_1
		local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask_2] 	; set csr
		ctx_arb[--]	
		// following .io_complete is used to eliminate assembling warning due to the
		// usage of "ACTIVE_CTX_WAKEUP_EVENTS" csr
		.io_completed sig3_next_context
		.io_completed sig_sram_read_1
		.io_completed sig_sram_read_2
		.io_completed sig_l2_table_entry_read
		#ifndef DISABLE_TX2SCHED_FEEDBACK
		.io_completed sig_reflector_write
		#endif	// #ifndef DISABLE_TX2SCHED_FEEDBACK

		// end of of phase 1 with sop_only non critical mpkt ready for tx


		// beginning of phase 2 with sop_only mpkt ready for tx	
		// wake up next thread in beginning of phase 2 to reduce wakeup latency
		local_csr_wr[SAME_ME_SIGNAL, sig3_next_context_gpr] 

	sop_only_non_critical_ready_to_transmit#:

		// write the NN request

		waiting_for_nn_to_become_not_full#:
			br_inp_state[nn_full, waiting_for_nn_to_become_not_full#]

			_sphy_mphy4_tx_debug_incr_counter(@requests_sent)
		
			alu[*n$index++, --, b, sop_bd]			; meta data header ptr


		#ifndef ADD_L2_HEADER
			alu[*n$index++, in_offset, or, 1, <<30]			; bit 30 = 0 means eop, bit 31 = 0 means sop
		#else	// ADD_L2_HEADER
			#ifdef POS_TX
				alu[*n$index++, in_offset, or, 1, <<30]			; bit 30 = 0 means eop, bit 31 = 0 means sop
			#endif	// #ifdef POS_TX
			#ifdef ETHERNET_TX
			.begin
				.reg tmp
				alu[tmp, in_offset, or, 1, <<30]
				br=byte[l2_header_exist, 0, L2_HEADER_IN_PAYLOAD, \
													l2_header_in_paylo#]
				alu[tmp, tmp, or, 1, <<29]		; bit 29 = 1 means l2 entry passed to next NE
			l2_header_in_paylo#:
				alu[*n$index++, --, b, tmp]		; bit 30 = 0 means eop, bit 31 = 0 means sop
			.end
			#endif	// #ifdef ETHERNET_TX
		#endif // #ifndef ADD_L2_HEADER
		#ifdef POS_TX
			#ifdef DYNAMIC_PPP_ENCAP
				alu_shf[*n$index++, sop_paylo, OR, in_pkt_type, <<16] ;Dynamic PPP Encap	
			#else
				alu[*n$index++, --, b, sop_paylo]
			#endif
		#else
			alu[*n$index++, --, b, sop_paylo]
		#endif

		#ifdef ADD_L2_HEADER
			#ifdef ETHERNET_TX
			
			// pass l2 entry to next ME
			
			_sphy_mphy4_tx_pass_l2_entry_to_next_me(l2_header_exist)

			#endif	// #ifdef ETHERNET_TX
		#endif // #ifdef ADD_L2_HEADER
	
			// check whether we need to save secondary buffer meta data 
		// if didn't read meta data in phase 1, just branch out
		br_bclr[exe_stat_flag, READ_SND_BD_FLAG_BIT, sop_only_non_critical_save_sbd_meta_done#]
	 
		// medat data for secondary buffer is ready, move them to local memory
		_sphy_mphy4_tx_save_sbd_meta_to_lm($dl_meta_sec) 
		// this port was blocked to prvent any following threads (which serves 
		// this port) in the same run to access the secondary buffer data when it 
		// becomes active buffer, now secondary meta data saved, put this port back
		alu[*l$index1[GLOBAL_NOT_LOCK_SNDB_RD_FLAG_INDEX], --, B, 1]	; unblock

	sop_only_non_critical_save_sbd_meta_done#:

		// set the lm_addr to tail of current LM queue to add the packet read from NN ring
		local_csr_wr[active_lm_addr_0, current_tail]
		nop					; wait local memory index 0 to settle down
		nop					; wait local memory index 0 to settle down

	sop_only_non_critical_end_of_phase_2#:
		// end of phase 2
	
	
		// if didn't read meta data in phase 2, just branch out
		br_bclr[exe_stat_flag, READ_SOP_BD_FLAG_BIT, sop_only_non_critical_cont_1#] 

		// medat data for sop is ready, move them to local memory
		_sphy_mphy4_tx_save_sop_meta_to_lm(tx_request_lw0) 


	sop_only_non_critical_cont_1#:

				br[PHASE1_START_LABEL]

	// end of phase 3 in the sop_only_non_critical mpacket path


	xbuf_free($dl_meta_sec)
	.end // sig_mask_2 sop_bd rmnd_ofset dram_addr indir_ref enq_port sop_paylo tmp

	// end of sop_only_non_critical mpacket processing code
#endm // end of #macro _sphy_mphy4_tx_sop_only_non_critical_mpkt()

///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// _sphy_mphy4_tx_one_buf_pkt()
//
// Description: handle up to 3 mpackets of a packet which just have one buffer
//				to optimize the performance
//
// Outputs:					
//							None
//
// Inputs:  
//		in_payload			payload length
//      in_offset			payload offset
//		in_pkt_type			needed only for PPP
//		in_offset_mod_8		remaider of offset divided by 8
//
// Constants:
//						 	None 
//
//	Labels:
//		PHASE1_START_LABEL	symbolic address for start of inifinite TX 
//							processing loop
/////////////////////////////////////////////////////////////////////////////////
#macro _sphy_mphy4_tx_one_buf_pkt(in_payload, in_offset, in_pkt_type, \
		in_offset_mod_8, PHASE1_START_LABEL)

	// beginning of one_buf_pkt mpacket processing code
	// one_buf_pkt_mpacket path in middle of phase 1
	 .begin
	.reg sop_bd sop_paylo snd_mpkt_paylo thrd_mpkt_paylo mpkt_numb
	.reg leftover_w0 
	.set leftover_w0 thrd_mpkt_paylo

	  	// get sop_bd 
	    ld_field_w_clr[sop_bd, 0111, *l$index0[ABD_0_OFFSET]]
	
		// now consider the restriction of the limitation that the sum of the prepend
		// length and payload length must be multiple of device bus width
	#ifndef ADD_L2_HEADER
		#ifdef SPI4_PREPEND
		.begin
		.reg allow_total
			alu[sop_paylo, SOP_EOP_MAX_PAYLOAD, -, offset_mod_8]
			alu[allow_total, sop_paylo, +, SPI4_PREPEND_HEADER_SIZE]
			alu[allow_total, allow_total, and, DEVICE_BUS_WIDTH_RESTRICTION_MASK]
			alu[sop_paylo, allow_total, -, SPI4_PREPEND_HEADER_SIZE]
		.end // allow_total
		#else
			alu[sop_paylo, TBUF_ELE_SIZE, -, offset_mod_8]
			alu[sop_paylo, sop_paylo, and, DEVICE_BUS_WIDTH_RESTRICTION_MASK]
		#endif
	#else // ADD_L2_HEADER
	#ifdef POS_TX
	.begin
	.reg allow_total
		alu[sop_paylo, SOP_EOP_MAX_PAYLOAD, -, offset_mod_8]
		alu[allow_total, sop_paylo, +, PPP_HEADER_SIZE]
		alu[allow_total, allow_total, and, DEVICE_BUS_WIDTH_RESTRICTION_MASK]
		alu[sop_paylo, allow_total, -, PPP_HEADER_SIZE]
	.end //allow_total
	#endif	// #ifdef POS_TX
	#ifdef ETHERNET_TX
	.begin
	.reg allow_total
		br=byte[l2_header_exist, 0, L2_HEADER_IN_PAYLOAD, l2_in_paylo#]
		alu[sop_paylo, SOP_EOP_MAX_PAYLOAD, -, offset_mod_8]
		alu[allow_total, sop_paylo, +, ETHERNET_HEADER_SIZE]
		alu[allow_total, allow_total, and, DEVICE_BUS_WIDTH_RESTRICTION_MASK]
		alu[sop_paylo, allow_total, -, ETHERNET_HEADER_SIZE]
		br[sop_paylo_size_done#]
	l2_in_paylo#:
		alu[sop_paylo, TBUF_ELE_SIZE_MINUS_16, -, offset_mod_8]
		alu[sop_paylo, sop_paylo, and, DEVICE_BUS_WIDTH_RESTRICTION_MASK]
	sop_paylo_size_done#:
	.end //allow_total
	#endif	// #ifdef ETHERNET_TX
	#endif //#ifndef ADD_L2_HEADER
	

	#ifdef ADD_L2_HEADER
		#ifdef ETHERNET_TX

		// add signal for l2 table entry read

		_sphy_mphy4_tx_set_signal[sig_mask_2, sig_l2_table_entry_read]
		#endif	// #ifdef ETHERNET_TX
	#endif // #ifdef ADD_L2_HEADER


		// decide up to two next mpacket size in this one buffer packet, in 
		// the case this packet can be fitted into two or three mpackets, 
		// this macro will advance queue head, otherwise it updates the payload
		// size and offset of the remaining payload 
				
		_sphy_mphy4_decide_next_two_mpkt_paylo(sig_mask_2, mpkt_numb, \
			snd_mpkt_paylo, thrd_mpkt_paylo, in_payload, in_offset, sop_paylo)

	 	// swap out to wait signals specified in sig_mask_1
		local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask_2] 	; set csr
		ctx_arb[--]	
		// following .io_complete is used to eliminate assembling warning due to the
		// usage of "ACTIVE_CTX_WAKEUP_EVENTS" csr
		.io_completed sig3_next_context
		.io_completed sig_sram_read_1
		.io_completed sig_sram_read_2
		.io_completed sig_l2_table_entry_read
		#ifndef DISABLE_TX2SCHED_FEEDBACK
		.io_completed sig_reflector_write
		#endif	// #ifndef DISABLE_TX2SCHED_FEEDBACK

		// end of of phase 1 with sop_only non critical mpkt ready for tx


		// beginning of phase 2 with sop_only mpkt ready for tx	
		// wake up next thread in beginning of phase 2 to reduce wakeup latency
		local_csr_wr[SAME_ME_SIGNAL, sig3_next_context_gpr] 

	one_buf_pkt_ready_to_transmit#:


		// write the NN request

		waiting_for_nn_to_become_not_full#:
			br_inp_state[nn_full, waiting_for_nn_to_become_not_full#]

			_sphy_mphy4_tx_debug_incr_counter(@requests_sent)
		
			alu[*n$index++, --, b, sop_bd]			; meta data header ptr


		#ifndef ADD_L2_HEADER
			alu[*n$index++, in_offset, or, 1, <<30]			; bit 30 = 0 means eop, bit 31 = 0 means sop
		#else	// ADD_L2_HEADER
			#ifdef POS_TX
				alu[*n$index++, in_offset, or, 1, <<30]			; bit 30 = 0 means eop, bit 31 = 0 means sop
			#endif	// #ifdef POS_TX
			#ifdef ETHERNET_TX
			.begin
				.reg tmp
				alu[tmp, in_offset, or, 1, <<30]
				br=byte[l2_header_exist, 0, L2_HEADER_IN_PAYLOAD, \
													l2_header_in_paylo#]
				alu[tmp, tmp, or, 1, <<29]		; bit 29 = 1 means l2 entry passed to next NE
			l2_header_in_paylo#:
				alu[*n$index++, --, b, tmp]		; bit 30 = 0 means eop, bit 31 = 0 means sop
			.end
			#endif	// #ifdef ETHERNET_TX
		#endif // #ifndef ADD_L2_HEADER
		#ifdef POS_TX
			#ifdef DYNAMIC_PPP_ENCAP
				alu_shf[*n$index++, sop_paylo, OR, in_pkt_type, <<16] ;Dynamic PPP Encap	
			#else
				alu[*n$index++, --, b, sop_paylo]
			#endif
		#else
			alu[*n$index++, --, b, sop_paylo]
		#endif

		#ifdef ADD_L2_HEADER
			#ifdef ETHERNET_TX
			
			// pass l2 entry to next ME
			
			_sphy_mphy4_tx_pass_l2_entry_to_next_me(l2_header_exist)

			#endif	// #ifdef ETHERNET_TX
		#endif // #ifdef ADD_L2_HEADER

		.begin
		.reg sop_eop_to_write tcw0_1 mpkt_offset mpkt_offset_mod_8

		// pass second mpacket info to TX2, only 6 lw, because there is no 
		// leftover

			alu[sop_eop_to_write, --, b, 3, <<30] ; assume not sop-eop mpacket		
			alu[mpkt_offset, in_offset, +, sop_paylo]
			alu[mpkt_offset_mod_8, mpkt_offset, and, CONST_07]
			alu[tcw0_1, --, b, 0]					; assume not eop mpacket 
			_sphy_mphy4_tx_set_tcw0_with_paylo_len_offset(tcw0_1, \
									snd_mpkt_paylo, mpkt_offset_mod_8)
			
			// clear EOB_MPKT_FLAG_BIT first, and set when necessary later

			alu[exe_stat_flag, exe_stat_flag, and~, 1, <<EOB_MPKT_FLAG_BIT]

			br!=byte[mpkt_numb, 0, 2, mpkt_2_continue#]

			// it is eop mpacket set EOP bit in tcw0_1, and EOB_MPKT_FLAG_BIT
			// exe_stat_flag accordngly

			alu_shf[tcw0_1, tcw0_1, or, 1, <<TCW0_EOP_BIT_SHFT]
			alu[exe_stat_flag, exe_stat_flag, or, 1, <<EOB_MPKT_FLAG_BIT]

		mpkt_2_continue#:

			_sphy_mphy4_tx_debug_incr_counter(@requests_sent)
			
			alu[*n$index++, --, b, sop_bd]			; meta data header ptr
			alu[*n$index++, --, b, sop_eop_to_write]
			alu[*n$index++, --, b, mpkt_offset]
			
			// first 3 more longwords, definitely no leftover 
			alu[*n$index++, --, b, leftover_w0]
			alu[*n$index++, --, b, exe_stat_flag]
			alu[*n$index++, --, b, tcw0_1]

			br=byte[mpkt_numb, 0, 2, tx_request_to_tx2_done#]

		// pass third mpacket info to TX2, only 6 lw, because there is no 
		// leftover

			alu[sop_eop_to_write, --, b, 3, <<30] ; assume not sop-eop mpacket
			alu[mpkt_offset, in_offset, +, sop_paylo]
			alu[mpkt_offset, mpkt_offset, +, snd_mpkt_paylo]

			alu[mpkt_offset_mod_8, mpkt_offset, and, CONST_07]
			alu[tcw0_1, --, b, 0]					; assume not eop mpacket 
			_sphy_mphy4_tx_set_tcw0_with_paylo_len_offset(tcw0_1, \
								thrd_mpkt_paylo, mpkt_offset_mod_8)
			br!=byte[mpkt_numb, 0, 3, mpkt_3_continue#]

			// it is eop mpacket set EOP bit in tcw0_1, and EOB_MPKT_FLAG_BIT
			// exe_stat_flag accordngly

			alu_shf[tcw0_1, tcw0_1, or, 1, <<TCW0_EOP_BIT_SHFT]
			alu[exe_stat_flag, exe_stat_flag, or, 1, <<EOB_MPKT_FLAG_BIT]
			
		mpkt_3_continue#:
			_sphy_mphy4_tx_debug_incr_counter(@requests_sent)

		waiting_for_nn_to_become_not_full_1#:
			br_inp_state[nn_full, waiting_for_nn_to_become_not_full_1#]
			
			alu[*n$index++, --, b, sop_bd]		; meta data header ptr
			alu[*n$index++, --, b, sop_eop_to_write]	
			alu[*n$index++, --, b, mpkt_offset]
			
			// first 3 more longwords, definitely no leftover 
			alu[*n$index++, --, b, leftover_w0]
			alu[*n$index++, --, b, exe_stat_flag]
			alu[*n$index++, --, b, tcw0_1]

	tx_request_to_tx2_done#:
		.end	// sop_eop_to_write tcw0_1 mpkt_offset mpkt_offset_mod_8
		
		// set the lm_addr to tail of current LM queue to add the packet read from NN ring
		local_csr_wr[active_lm_addr_0, current_tail]
        nop					; wait local memory index 0 to settle down
		nop					; wait local memory index 0 to settle down

		// if didn't read meta data in phase 2, just branch out
		br_bclr[exe_stat_flag, READ_SOP_BD_FLAG_BIT, one_buf_pkt_cont_1#] 

		// medat data for sop is ready, move them to local memory
		_sphy_mphy4_tx_save_sop_meta_to_lm(tx_request_lw0) 


	one_buf_pkt_cont_1#:

				br[PHASE1_START_LABEL]

	// end of phase 3 in the one_buf_pkt mpacket path


	.end // sop_bd sop_paylo snd_mpkt_paylo thrd_mpkt_paylo mpkt_numb leftover_w0

#endm // end of #macro _sphy_mphy4_tx_one_buf_pkt()

///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// _sphy_mphy4_tx_not_sop_mpkt()
//
// Description: handle mpacket which is not the first mpacket of a 
//				multiple-mpacket packet
//
// Outputs:					
//							None
//
// Inputs:  
//						 	None 
//
// Constants:
//						 	None 
//
//	Labels:
//		PHASE1_START_LABEL	symbolic address for start of inifinite TX 
//							processing loop
//
///////////////////////////////////////////////////////////////////////////////

#macro _sphy_mphy4_tx_not_sop_mpkt(PHASE1_START_LABEL)
	///////////////////////////////////////////////////////////////////////////////
	// beginning of not_sop_mpacket processing code
	// now still in phase 1
	not_sop_mpkt#:
	.begin 
	.reg ab_bd dram_addr indir_ref lov_to_next_offset rmnd_paylo offset_rpaylo tmp
	.reg max_size_except_leftover

	//.reg sig_mask_2 
	.reg $$dw0 $$dw1 $$dw2 $$dw3 $$dw4 $$dw5
	.xfer_order $$dw0 $$dw1 $$dw2 $$dw3 $$dw4 $$dw5

		alu[tcw0_1, --, B, deq_port]	; initialize tcw0_1 to deq_port

	.reg sop_eop_to_write
	alu[sop_eop_to_write, --, b, 3, <<30]			; not sop and assumed not eop by default

	// following .set used to eliminate assembling warning
	//.set $dl_meta[0] $dl_meta[1] $dl_meta[2] 
	.set $w0 $w1
	.set $$dw0 $$dw1 $$dw2 $$dw3 $$dw4 $$dw5
	.set lov_to_next_offset
	.reg leftover_w0 leftover_w1 leftover_w2 leftover_w3
	.set leftover_w0 leftover_w1 leftover_w2 leftover_w3

		alu_shf[rmnd_paylo, --, b, *l$index0[ABD_1_OFFSET], >>PAYLO_RMND_LOC]
	    ld_field_w_clr[offset_rpaylo, 0011, *l$index0[ABD_1_OFFSET]]

		xbuf_alloc[$dl_meta_sec, 3, read]

		.set $dl_meta_sec[0] $dl_meta_sec[1] $dl_meta_sec[2]

		// check whether to read secondary BD 
		alu[tmp, --, b, *l$index0[ABD_0_OFFSET]]
	    br_bclr[tmp, RD_NBD_BIT_LOC, not_sop_sbd_rd_done#]

		 // read secondary buffer meta data from SRAM 
		_sphy_mphy4_tx_read_sb_meta_data_from_sram($dl_meta_sec, sig_mask_2, \
														*l$index0[SBD_0_OFFSET]) 

		// reset need to read secondary buffer meta data flag 
		alu_shf[*l$index0[ABD_0_OFFSET], *l$index0[ABD_0_OFFSET], and~, 1, \
				<<RD_NBD_BIT_LOC]
		// set read secondary BD flag in exe_stat_flag for use in phase 2
		alu_shf[exe_stat_flag, exe_stat_flag, or, 1, <<READ_SND_BD_FLAG_BIT]

		// check whether this port should be blocked from transmission for any 
		// following threads which serve the same in the same run to prvent them 
		// accessing  the secondary buffer meta data (which is not ready yet in this
		// phase, but will be ready in next phase) when they become active buffer
		// calculate the max bytes to consume before secondary buffer meta data need
		// to be ready

		alu_shf[tmp, --, b, TBUF_ELE_SIZE, <<ME_THREAD_NUM_SHFT] ; max bytes required
		alu[--, rmnd_paylo, -, tmp]			; compare with how much left
		bgt[not_sop_sbd_rd_done#]
		// set *l$index1[GLOBAL_NOT_LOCK_SNDB_RD_FLAG_INDEX] to 0 to prvent any threads
		// handling the same port in the same run to access the secondary buffer data 
		// (which is not ready yet) when it becomes active buffer, set 
		// put *l$index1[GLOBAL_NOT_LOCK_SNDB_RD_FLAG_INDEX] to 1
		// when secondary meta data saved in phase 2.
		alu[*l$index1[GLOBAL_NOT_LOCK_SNDB_RD_FLAG_INDEX], --, B, 0]

	not_sop_sbd_rd_done#:
	  	// get ab_bd, rmnd_paylo, offset_rpaylo, offset_mod_8, ab_flist for active 
		// buffer
	    ld_field_w_clr[ab_bd, 0111, *l$index0[ABD_0_OFFSET]] ; get active buffer descriptor
		alu[offset_mod_8, offset_rpaylo, and, CONST_07]
	.begin
	.reg lpb_paylo
		// check whether there is some leftover bytes (maximum 8 bytes) from 
		// previous bufferwhich need to be moved to tbuf at the beginning
		// of this mpacket
		alu[tmp, --, B, *l$index1[GLOBAL_LEFTOVER_FLAG_LEN_INDEX]]
		alu[tmp, --, B, tmp, >>31]				; MSB 
		alu[--, tmp, -, PE_SPHY_TX_FLAG_CLEAR]
	    beq[not_sop_no_leftover#]
		// yes, there is leftover from previous buffer in the same packet
		// move leftover bytes to leftover_w0... for latter use, also clear that
		// bit, so later threads in the same run will not repeat the  same thing
		alu[tmp, --, B, *l$index1[GLOBAL_LEFTOVER_FLAG_LEN_INDEX]]
	    alu[tmp, tmp, AND~, PE_SPHY_TX_FLAG_SET, <<31] 	; reset
	    alu[*l$index1[GLOBAL_LEFTOVER_FLAG_LEN_INDEX], --, B, tmp]	 ; reset

		alu[leftover_w0, --, b, *l$index1[GLOBAL_LEFTOVER_LW0_INDEX]] 
		alu[leftover_w1, --, b, *l$index1[GLOBAL_LEFTOVER_LW1_INDEX]] 
		alu[leftover_w2, --, b, *l$index1[GLOBAL_LEFTOVER_LW2_INDEX]] 
		alu[leftover_w3, --, b, *l$index1[GLOBAL_LEFTOVER_LW3_INDEX]] 
		alu[$w1, --, b, 0]						  ; dummy
		// payload leftover will be put into prepend to eliminate any alignment problem
		// also before saved to local memory, byte alignment is done on leftover bytes
		// so make sure the offset of those leftover bytes is 0
		alu[tmp, --, B, *l$index1[GLOBAL_LEFTOVER_FLAG_LEN_INDEX]]
		ld_field_w_clr[lpb_paylo, 0001, tmp]
		alu_shf[tcw0_1, tcw0_1, or, lpb_paylo, <<TCWO_PREPEND_LEN_LSB_LOC]
		alu[max_size_except_leftover, --, B, TBUF_ELE_SIZE_MINUS_8]
		alu[--, lpb_paylo, -, 8]
		ble[max_size_except_leftover_decision_done#]
		alu[max_size_except_leftover, --, B, TBUF_ELE_SIZE_MINUS_16]
		max_size_except_leftover_decision_done#:
		alu_shf[exe_stat_flag, exe_stat_flag, or, 1, \
				<<LEFTOV_FROM_PB_FLAG_BIT]				;set flag

		// check whether current mpacket is eop packet
		// first check whether this is a last buffer in a packet
	    ld_field_w_clr[tmp, 0111, *l$index0[SBD_0_OFFSET]]
		alu[--, tmp, -, IX_NULL]
		bne[lp_not_last_buffer#]
		// check whether all remianed payload can fit into one tbuf, leftover already
		// occupies 8 bytes in tbuf 
		alu[tmp, offset_mod_8, +, rmnd_paylo]
		alu[--, max_size_except_leftover, -, tmp]
		blt[lp_last_bp_not_eop_mpkt#]
		// it is EOP case, set EOP bit in tcw0_1
		alu_shf[tcw0_1, tcw0_1, or, 1, <<TCW0_EOP_BIT_SHFT]
		// set payload_length and payload_offset of tcw0_1 with rmnd_paylo and
		// offset_mod_8
		_sphy_mphy4_tx_set_tcw0_with_paylo_len_offset(tcw0_1, rmnd_paylo, offset_mod_8)
		// eop mpacket is eob mpacket too, need to free buffer at phase 3
		alu_shf[exe_stat_flag, exe_stat_flag, or, 1, <<EOB_MPKT_FLAG_BIT]
		br[not_sop_need_to_advance_queue_head#] 
	
		// following case need to consider device bus width restriction
	lp_last_bp_not_eop_mpkt#:
		// calculate what max can be transmit from this buffer
		// first total max payload (include lp_paylo) must be multiple of 
		// device bus width
	.begin 
	.reg alw_mpkt_paylo	; maximum allowable mpkt payload from active buffer
						; with consideration of multiple of device bus width

		alu[tmp, max_size_except_leftover, -, offset_mod_8]
		alu[tmp, tmp, +, lpb_paylo]	;  now offset_mod_8 ahould be 0
		alu[tmp, tmp, and, DEVICE_BUS_WIDTH_RESTRICTION_MASK]	
		alu[alw_mpkt_paylo, tmp, -, lpb_paylo]

		// update ab_paylo_rmnd and ab_offset_rpaylo in queue entry with 
		// alw_mpkt_paylo
		_sphy_mphy4_tx_update_ab_paylo_rmnd_and_offset_rpaylo(alw_mpkt_paylo)
		// set payload_length and payload_offset of tcw0_1 with alw_mpkt_paylo
		// and offset_mod_8
		_sphy_mphy4_tx_set_tcw0_with_paylo_len_offset(tcw0_1, alw_mpkt_paylo, \
												offset_mod_8)	
		br[not_sop_end_of_phase1#]
	.end //	 alw_mpkt_paylo
	lp_not_last_buffer#:
		// since not last buffer in a packet , so only check whether it is a eob 
		// mpacket	
		// which rmnd_paylo we can say it is eob mpacket
		// calculate what max can be transmit from this buffer
		// first total max payload (include lp_paylo) must be multiple of device 
		// bus width
	.begin
	.reg alw_mpkt_paylo	; maximum allowable payload from active buffer
						; with consideration of multiple of device bus width
		alu[tmp, max_size_except_leftover, -, offset_mod_8]	;  now offset_mod_8 ahould be 0
		alu[tmp, tmp, +, lpb_paylo]	
		alu[tmp, tmp, and, DEVICE_BUS_WIDTH_RESTRICTION_MASK]
		alu[alw_mpkt_paylo, tmp, -, lpb_paylo]

		// if rmnd_paylo is less than this max allowable bytes, of cours, we are in 
		// eob 
		alu[--, alw_mpkt_paylo, -, rmnd_paylo]
		blt[lp_not_last_bp_not_eob#]

		// it is eob, set eob bit
		alu_shf[exe_stat_flag, exe_stat_flag, or, 1, <<EOB_MPKT_FLAG_BIT]

	.begin
	.reg need_to_transmit
	.reg allow_to_transmit
	.reg tmp1
		alu[need_to_transmit, lpb_paylo, +, rmnd_paylo]
		alu[allow_to_transmit, need_to_transmit, and, DEVICE_BUS_WIDTH_RESTRICTION_MASK]
		alu[allow_to_transmit, need_to_transmit, -, lpb_paylo]

		// set payload_length and payload_offset of tcw0_1 with allow_to_transmit
		// and offset_mod_8
		_sphy_mphy4_tx_set_tcw0_with_paylo_len_offset(tcw0_1, allow_to_transmit, \
										offset_mod_8)

		// check whether there are some leftover to raed to $$dw0, $$dw1, 
		// $$dw2, $$dw3, $$dw4, and $$dw5
		// check whether we have some leftover to next buffer 
		alu[--, rmnd_paylo, -, allow_to_transmit]		
		beq[lp_not_last_buffer_no_need_to_save_leftover#]	; it meets multiple of 4 criteria
		// set flag in exe_stat_flag for later use 
		alu_shf[exe_stat_flag, exe_stat_flag, or, 1, <<SAVE_LEFTOV_TO_LM_FLAG_BIT]
		// get true active buffer BD from ab_bd  
		alu_shf[tmp, --, b, ab_bd, <<2]
	    // use following dl_buf_get_data_from_meta macro from dispatch_loop.uc to
		// get dram address at the beginning of the buffer from true_dl_buf_hdl
		// now it takes one instruction
	    dl_buf_get_data_from_meta(dram_addr, tmp)	; get beginning dram addr for buffer  
	
		alu[tmp, offset_rpaylo, +, allow_to_transmit]
		alu[dram_addr, dram_addr, +, tmp]	
		// in some case, the sum of offset_mod_8 + bytes_leftvoer will over 16 bytes 
		// (for example, offset_mod_8=7, and bytes_leftover is 15), so in order to 
		// cover all cases, read 24 bytes here 
		// add sig_dram_access_1 to sig_mask
		dram[read, $$dw0, dram_addr, 0, 3], sig_done[sig_dram_access_1]	
		_sphy_mphy4_tx_set_signal(sig_mask_2, sig_dram_access_1)	; add sig_dram_access_1
		// since we use sig_mask and ACTIVE_CTX_WAKEUP_EVENTS csr, we need to add to 
		// wait for the push signal of dram access
		_sphy_mphy4_tx_set_push_signal(sig_mask_2, sig_dram_access_1)	; add push signal

		// this port should be blocked from transmission before those $$dw0, $$dw1, 
		//$$dw2, $$dw3, $$dw4, and $$dw5 are gotten and saved to LM
		alu[*l$index1[GLOBAL_NOT_LOCK_LEFTOVER_FLAG_INDEX], --, B, 0]	; block this port
																; until leftover ready
		alu[lov_to_next_offset, tmp, and, CONST_07]	; 8 bytes boudary due to dram[read..] 
		alu[tmp, rmnd_paylo, -, allow_to_transmit]
		alu[tmp1, --, B, *l$index1[GLOBAL_LEFTOVER_FLAG_LEN_INDEX]]
		ld_field[tmp1, 0001, tmp]	; save leftover length
		alu[*l$index1[GLOBAL_LEFTOVER_FLAG_LEN_INDEX], --, B, tmp1]	; save leftover length
	.end // need_to_transmit allow_to_transmit tmp1
	lp_not_last_buffer_no_need_to_save_leftover#:
		_sphy_mphy4_tx_secondary_buffer_become_active_buffer()
		br[not_sop_end_of_phase1#]	

	lp_not_last_bp_not_eob#:	
		// update ab_paylo_rmnd and ab_offset_rpaylo in queue entry with 
		// alw_mpkt_paylo
		_sphy_mphy4_tx_update_ab_paylo_rmnd_and_offset_rpaylo(alw_mpkt_paylo)
		// set payload_length and payload_offset of tcw0_1 with alw_mpkt_paylo
		// and offset_mod_8
		_sphy_mphy4_tx_set_tcw0_with_paylo_len_offset(tcw0_1, alw_mpkt_paylo, \
												offset_mod_8)
		br[not_sop_end_of_phase1#]
	.end //	 alw_mpkt_paylo
	.end //  lpb_paylo, lp_offset

	// no leftover from previous buffer
	 not_sop_no_leftover#:

		// check whether current mpacket is eop packet
		// first check whether this is a last buffer in a packet
	    ld_field_w_clr[tmp, 0111, *l$index0[SBD_0_OFFSET]]
		alu[--, tmp, -, IX_NULL]
		bne[no_lp_not_last_buffer#]
		// check whether all remianed payload can fit into one tbuf, leftover already
		// occupies 8 bytes 
		alu[tmp, offset_mod_8, +, rmnd_paylo]

		alu[--, TBUF_ELE_SIZE, -, tmp]
		blt[no_lp_last_bp_not_eop_mpkt#]
		// it is EOP case, set EOP bit in tcw0_1
		alu_shf[tcw0_1, tcw0_1, or, 1, <<TCW0_EOP_BIT_SHFT]
		// set payload_length and payload_offset of tcw0_1 with rmnd_paylo
		// and offset_mod_8
		_sphy_mphy4_tx_set_tcw0_with_paylo_len_offset(tcw0_1, rmnd_paylo, offset_mod_8)
		alu_shf[exe_stat_flag, exe_stat_flag, or, 1, <<EOB_MPKT_FLAG_BIT]
		br[not_sop_need_to_advance_queue_head#] 
	
		// following case need to consider device bus width restriction
	no_lp_last_bp_not_eop_mpkt#:
		// calculate what max can be transmit from this buffer
		// first total max payload (include lp_paylo) must be multiple of device bus
		// width
	.begin 
	.reg alw_mpkt_paylo		; maximum allowable payload from active buffer
							; with consideration of multiple of device bus width
		alu[tmp, TBUF_ELE_SIZE, -, offset_mod_8]
		alu[alw_mpkt_paylo, tmp, and, DEVICE_BUS_WIDTH_RESTRICTION_MASK]
		// update ab_paylo_rmnd and ab_offset_rpaylo in queue entry with 
		// alw_mpkt_paylo
		_sphy_mphy4_tx_update_ab_paylo_rmnd_and_offset_rpaylo(alw_mpkt_paylo)
		// set payload_length and payload_offset of tcw0_1 with alw_mpkt_paylo
		// and offset_mod_8
		_sphy_mphy4_tx_set_tcw0_with_paylo_len_offset(tcw0_1, alw_mpkt_paylo, \
												offset_mod_8)	
		br[not_sop_end_of_phase1#]
	.end //	 alw_mpkt_paylo
	no_lp_not_last_buffer#:
		// check whether is EOB
		// which rmnd_paylo we can say it is eob
		// calculate what max can be transmit from this buffer
		// first total max payload (include lp_paylo) must be multiple of device 
		// bus width
	.begin
	.reg alw_mpkt_paylo	; maximum allowable payload from active buffer
						; with consideration of multiple of device bus width
		alu[tmp, TBUF_ELE_SIZE, -, offset_mod_8]
		alu[alw_mpkt_paylo, tmp, and, DEVICE_BUS_WIDTH_RESTRICTION_MASK]

		// if rmnd_paylo is less than this max allowable bytes, of cours, we are in 
		// eob 
		alu[--, alw_mpkt_paylo, -, rmnd_paylo]
		blt[no_lp_not_last_bp_not_eob#]

		// it is eob, set eob bit
		alu_shf[exe_stat_flag, exe_stat_flag, or, 1, <<EOB_MPKT_FLAG_BIT]
	.begin
	.reg allow_to_transmit 
	.reg tmp1
		alu[allow_to_transmit, rmnd_paylo, and, DEVICE_BUS_WIDTH_RESTRICTION_MASK]

		// check whether we have something to transmit
		alu[--, allow_to_transmit, -, 0]
		bgt[no_lp_not_last_buffer_something_to_transmit#]
		// if it is the case, then set skip bit in tcw0_1
		alu_shf[tcw0_1, tcw0_1, or, 1, <<TCW0_SKIP_BIT_SHFT]	; set skip bit
		// set flag in exe_stat_flag for later use 
		alu_shf[exe_stat_flag, exe_stat_flag, or, 1, <<SKIP_TRANSMIT_FLAG_BIT]
	no_lp_not_last_buffer_something_to_transmit#:
		// set payload_length and payload_offset of tcw0_1 with allow_to_transmit
		// and offset_mod_8
		_sphy_mphy4_tx_set_tcw0_with_paylo_len_offset(tcw0_1, allow_to_transmit, \
											offset_mod_8)
		// check whether there are some leftover to raed to $$dw0, $$dw1, 
		// $$dw2, $$dw3, $$dw4, and $$dw5
		alu[--, rmnd_paylo, -, allow_to_transmit]		
		beq[no_lp_not_last_buffer_no_need_to_save_leftover#] ; it meets multiple of device 
															 ; bus width criteria
		// set flag in exe_stat_flag for later use 
		alu_shf[exe_stat_flag, exe_stat_flag, or, 1, <<SAVE_LEFTOV_TO_LM_FLAG_BIT]
		// get true active buffer BD from ab_bd  
		alu_shf[tmp, --, b, ab_bd, <<2]
	    // use following dl_buf_get_data_from_meta macro from dispatch_loop.uc to
		// get dram address at the beginning of the buffer from true_dl_buf_hdl
		// now it takes one instruction
	    dl_buf_get_data_from_meta(dram_addr, tmp)
	
		alu[tmp, offset_rpaylo, +, allow_to_transmit]
		alu[dram_addr, dram_addr, +, tmp]
		// in some case, the sum of offset_mod_8 + bytes_leftvoer will over 16 bytes 
		// (for example, offset_mod_8=7, and bytes_leftover is 15), so in order to 
		// cover all cases, read 24 bytes here 
		// add sig_dram_access_1 to sig_mask
		dram[read, $$dw0, dram_addr, 0, 3], sig_done[sig_dram_access_1]	
		// add sig_dram_access_1 to sig_mask_1
		_sphy_mphy4_tx_set_signal(sig_mask_2, sig_dram_access_1)	; add sig_dram_access_1
		// since we use sig_mask and ACTIVE_CTX_WAKEUP_EVENTS csr, we need to add to 
		// wait for the push signal of dram access
		_sphy_mphy4_tx_set_push_signal(sig_mask_2, sig_dram_access_1)	; add push signal
		// this port should be blocked from transmission before those $$dw0, 
		// $$dw1, $$dw2, $$dw3, $$dw4, and $$dw5  is gotten and saved to LM
		alu[*l$index1[GLOBAL_NOT_LOCK_LEFTOVER_FLAG_INDEX], --, B, 0]	; block this port
																; until leftover ready
		alu[lov_to_next_offset, tmp, and, CONST_07]	; 8 bytes boudary due ro dram[read..] 
		alu[tmp, rmnd_paylo, -, allow_to_transmit]
		alu[tmp1, --, B, *l$index1[GLOBAL_LEFTOVER_FLAG_LEN_INDEX]]
		ld_field[tmp1, 0001, tmp]	; save leftover length
		alu[*l$index1[GLOBAL_LEFTOVER_FLAG_LEN_INDEX], --, B, tmp1]	; save leftover length
	.end // allow_to_transmit tmp1
	no_lp_not_last_buffer_no_need_to_save_leftover#:
		_sphy_mphy4_tx_secondary_buffer_become_active_buffer()
		br[not_sop_end_of_phase1#]	
	no_lp_not_last_bp_not_eob#:	
		// update ab_paylo_rmnd and ab_offset_rpaylo in queue entry with 
		// alw_mpkt_paylo
		_sphy_mphy4_tx_update_ab_paylo_rmnd_and_offset_rpaylo(alw_mpkt_paylo)
		// set payload_length and payload_offset of tcw0_1 with alw_mpkt_paylo
		// and offset_mod_8
		_sphy_mphy4_tx_set_tcw0_with_paylo_len_offset(tcw0_1, alw_mpkt_paylo, \
												offset_mod_8)
	.end //	 alw_mpkt_paylo
		br[not_sop_end_of_phase1#]	   

		// advance queue header for EOP mpacket 
	not_sop_need_to_advance_queue_head#:

		// even the deq_port_bit wil be calculated in _sphy_mphy4_tx_advance_queue_head[] 
		// macro gaian, but it is the same as original deq_port_bit 
		_sphy_mphy4_tx_advance_queue_head(port_entry_turnaround)
	
		#ifndef DISABLE_TX2SCHED_FEEDBACK

		// add signal for reflect write

		_sphy_mphy4_tx_set_signal[sig_mask_2, sig_reflector_write]
		#endif	// #ifndef DISABLE_TX2SCHED_FEEDBACK

	not_sop_end_of_phase1#:
		
		// swap out to wait signals specified in sig_mask
		local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask_2]	; set csr													; csr
		ctx_arb[--]

		// following .io_complete is used to eliminate assembling warning due to the
		// usage of "ACTIVE_CTX_WAKEUP_EVENTS" csr
		.io_completed sig3_next_context
		.io_completed sig_sram_read_1
		.io_completed sig_sram_read_2
		.io_completed sig_dram_access_1
		.io_completed sig_l2_table_entry_read
		#ifndef DISABLE_TX2SCHED_FEEDBACK
		.io_completed sig_reflector_write
		#endif	// #ifndef DISABLE_TX2SCHED_FEEDBACK

	not_sop_ready_to_transmit#:

		// write the NN request

		waiting_for_nn_to_become_not_full#:
			br_inp_state[nn_full, waiting_for_nn_to_become_not_full#]

			_sphy_mphy4_tx_debug_incr_counter(@requests_sent)
			
			alu[*n$index++, --, b, ab_bd]			; meta data header ptr
			alu[*n$index++, sop_eop_to_write, or, 0]			; bit 30 = 0 means eop, bit 31 = 0 means sop
			alu[*n$index++, --, b, offset_rpaylo]
			

			// first 3 more longwords 
			alu[*n$index++, --, b, leftover_w0]
			alu[*n$index++, --, b, exe_stat_flag]
			alu[*n$index++, --, b, tcw0_1]

			// check LEFTOV_FROM_PB_FLAG_BIT in exe_stat_flag, pass another 
			// three longword leftover only when this bit set

			br_bset[exe_stat_flag, LEFTOV_FROM_PB_FLAG_BIT, \
												not_sop_pass_three_more_lw#]

		not_sop_leftover_pass_done#:
			
		// check whether we have leftover which needed to be to saved in LM
		br_bclr[exe_stat_flag, SAVE_LEFTOV_TO_LM_FLAG_BIT, ph2_save_leftov_to_lm_done#]

		// $$dw0, $$dw1, $$dw2, $$dw3, $$dw4, and $$dw5 ready now, do necessary
		// bytes alignment with lov_to_next_offset, so the leftover will be 
		// realigned and saved into local memory

		_sphy_mphy4_tx_process_and_save_leftover(lov_to_next_offset)

		// set flag in queue status
	    alu[tmp, --, B, *l$index1[GLOBAL_LEFTOVER_FLAG_LEN_INDEX]]
		alu[tmp, tmp, OR, PE_SPHY_TX_FLAG_SET, <<31] ; set
	    alu[*l$index1[GLOBAL_LEFTOVER_FLAG_LEN_INDEX], --, B, tmp]	; set 
	   	// data saved, port ready to be transmitted again, put this port back
		alu[*l$index1[GLOBAL_NOT_LOCK_LEFTOVER_FLAG_INDEX], --, B, 1]	; unblock
	ph2_save_leftov_to_lm_done#:

		// check whether we need to save secondary buffer meta data 
		// if didn't read meta data in phase 2, just branch out
		br_bclr[exe_stat_flag, READ_SND_BD_FLAG_BIT, not_sop_save_sbd_meta_done#]

		// medat data for secondary buffer is ready, move them to local memory
		_sphy_mphy4_tx_save_sbd_meta_to_lm($dl_meta_sec) 
		// this port was blocked to prvent any following threads (which serves 
		// this port) in the same run to access the secondary buffer data when it 
		// becomes active buffer, now secondary meta data saved, put this port back
		alu[*l$index1[GLOBAL_NOT_LOCK_SNDB_RD_FLAG_INDEX], --, B, 1]	; unblock
	not_sop_save_sbd_meta_done#:

	local_csr_wr[active_lm_addr_0, current_tail]
		// beginning of phase 2 with not_sop mpkt ready for tx	
		// wake up next thread in beginning of phase 2 to reduce wakeup latency
		local_csr_wr[SAME_ME_SIGNAL, sig3_next_context_gpr] 
		nop

		// if didn't read meta data in phase 2, just branch out
		br_bclr[exe_stat_flag, READ_SOP_BD_FLAG_BIT, not_sop_no_tx_request_read#] 

		// medat data for sop is ready, move them to local memory
		_sphy_mphy4_tx_save_sop_meta_to_lm(tx_request_lw0) 


	not_sop_no_tx_request_read#:
	not_sop_no_valid_tx_request#:

	not_sop_free_buffer_done#:


	not_sop_save_sop_meta_done#:

			br[PHASE1_START_LABEL]

	not_sop_pass_three_more_lw#:

			// pass three more longword leftover
			
			alu[*n$index++, --, b, leftover_w1]
			alu[*n$index++, --, b, leftover_w2]
			alu[*n$index++, --, b, leftover_w3]
	br[not_sop_leftover_pass_done#]	

	.end	// ab_bd dram_addr indir_ref lov_to_next_offset  rmnd_paylo offset_rpaylo
			// $$dw0 $$Dw1 $$dw2 $$dw3 $$dw4 $$dw5
	// end of not_sop mpacket processing
#endm // end of #macro _sphy_mphy4_tx_not_sop_mpkt()


///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// find_first_request_and_process()
//
// Description: looks at the LM queue and find the head of LM queue, if available
//				and calls the appropriate macro to transmit a mpacket from it
//
//
// Outputs:					
//							None
//
// Inputs:  
//						 	META_DATA_READ:	a value of 1 means a request is read
//											from NN ring and meta data read is issued 
//
// Constants:
//						 	None 
//
//
///////////////////////////////////////////////////////////////////////////////
#macro find_first_request_and_process(META_DATA_READ)
		
	#ifndef DISABLE_SNDB_RD_LOCK_CHECK
	.begin
	.reg tmp	
		// check if LM queue is locked
	    alu[tmp, --, B, *l$index1[GLOBAL_NOT_LOCK_SNDB_RD_FLAG_INDEX]]
		// check if LM queue is empty
	    alu[--, *l$index1[GLOBAL_PKTS_IN_QUEUE_INDEX], -, 0] 
		ble[not_ready_to_tx#]	
		
		// check if LM queue is locked
	    alu[tmp, tmp, AND, *l$index1[GLOBAL_NOT_LOCK_LEFTOVER_FLAG_INDEX]] 
		beq[not_ready_to_tx#]
	.end
	#else	// DISABLE_SNDB_RD_LOCK_CHECK
	    alu[--, *l$index1[GLOBAL_PKTS_IN_QUEUE_INDEX], -, 0] 
		ble[not_ready_to_tx#]	

		// check waiting for leftover reding finish?
	    alu[--, *l$index1[GLOBAL_NOT_LOCK_LEFTOVER_FLAG_INDEX], AND, 1] 
		beq[not_ready_to_tx#]

	#endif	// #ifndef DISABLE_SNDB_RD_LOCK_CHECK
	
		
	.begin 
	.reg tmp
		// check if SOP
		alu[tmp, --, b, *l$index0[ABD_0_OFFSET]]
		// not SOP mpacket,  not worst case, can take branchout penatly
	    br_bclr[tmp, SOP_BIT_LOC, not_sop_mpacket#]
	.end

	.begin 
	.reg offset_mod_8
	.reg payload_len
	.reg payload_offset

	#ifdef ADD_L2_HEADER
		#ifdef ETHERNET_TX
		.begin
		.reg nexthop_id l2_table_base tmp
			alu[nexthop_id, --, b, *l$index0[SBD_1_OFFSET]]
			move(l2_table_base, L2_TABLE_SRAM_BASE)

			// read it even L2 header exists to reduce instruction to check 

			_sphy_mphy4_tx_read_l2_table_entry(l2_table_base, nexthop_id, \
													sig_l2_table_entry_read)

			alu[l2_header_exist, --, b, L2_HEADER_IN_PAYLOAD]	
			move(tmp, L2_HEADER_ALREADY_EXIST_ID)
			alu[--, nexthop_id, -, tmp]
			beq[check_done#]
			alu[l2_header_exist, --, b, L2_HEADER_NOT_IN_PAYLOAD]	
		check_done#:
		.end 
		#endif	// #ifdef ETHERNET_TX
	#endif // #ifdef ADD_L2_HEADER


	check_whether_sop_eop_mpacket#:
		// check if packet is min packet
		alu_shf[payload_len, --, b, *l$index0[ABD_1_OFFSET], >>PAYLO_RMND_LOC]
	    ld_field_w_clr[payload_offset, 0011, *l$index0[ABD_1_OFFSET]]
		alu[offset_mod_8, payload_offset, and, CONST_07]
	.begin
	.reg tmp
	// calculate payload length
	#ifndef ADD_L2_HEADER
		#ifdef SPI4_PREPEND
			alu[tmp, SOP_EOP_MAX_PAYLOAD, -, offset_mod_8]
		#else
	   		alu[tmp, TBUF_ELE_SIZE, -, offset_mod_8]
		#endif
	#else // ADD_L2_HEADER 
	#ifdef POS_TX
	   alu[tmp, SOP_EOP_MAX_PAYLOAD, -, offset_mod_8]
	#endif	// #ifdef POS_TX
	#ifdef ETHERNET_TX
		br=byte[l2_header_exist, 0, L2_HEADER_IN_PAYLOAD, l2_in_paylo#]
		alu[tmp, SOP_EOP_MAX_PAYLOAD, -, offset_mod_8]
		br[continue_for_size_decision#]
	l2_in_paylo#:
	   alu[tmp, TBUF_ELE_SIZE, -, offset_mod_8]
	continue_for_size_decision#:
	#endif	// #ifdef ETHERNET_TX
	#endif // #ifndef ADD_L2_HEADER 
		alu[--, payload_len, -, tmp]
	.end //tmp
		// not sop_eop packet, can take branchout penalty

		bgt[sop_mpkt#]

		// min packet case

		_sphy_mphy4_tx_sop_eop_mpacket(payload_len, payload_offset, *l$index0[ABD_PKT_TYPE_OFFSET], \
									offset_mod_8, phase1#, META_DATA_READ)

	sop_mpkt#:
	alu[--, *l$index0[SBD_0_OFFSET], -, IX_NULL]	; is null (0xFF)?
	bne[sop_only_non_critical_mpkt#]

		alu[sig_mask_2, --, b, sigmask_phs2_default]			; initialize sig_mask_2
		#if (META_DATA_READ == 0)
			_sphy_mphy4_tx_clear_signal(sig_mask_2, sig_sram_read_1)	; clear sig_sram_read_1
			alu[exe_stat_flag, --, b, 0]		; reset exe_stat_flag
		#else			
			alu[exe_stat_flag, --, b, 1, <<READ_SOP_BD_FLAG_BIT]		; set exe_stat_flag
		#endif
	
	// handle up to 3 mpackets from sop for a packet with only one buffer to
	// optimize the perfomance

	_sphy_mphy4_tx_one_buf_pkt(payload_len, payload_offset, *l$index0[ABD_PKT_TYPE_OFFSET], \
								offset_mod_8, phase1#)

	
	sop_only_non_critical_mpkt#:
		// we use sigmasks for non-min packet case to find out whether we have to wait for
		// meta data read of not -- code below sets sigmask depending on META_DATA_READ
		// We do not do this in min packet to save cycles used to manipulate sigmask
		// Instead we use up some code space to have different code for all possible cases
		// where we wait for different set of signals
		alu[sig_mask_2, --, b, sigmask_phs2_default]			; initialize sig_mask_2
		#if (META_DATA_READ == 0)
			_sphy_mphy4_tx_clear_signal(sig_mask_2, sig_sram_read_1)	; clear sig_sram_read_1
			alu[exe_stat_flag, --, b, 0]		; reset exe_stat_flag
		#else			
			alu[exe_stat_flag, --, b, 1, <<READ_SOP_BD_FLAG_BIT]		; set exe_stat_flag
		#endif

		// handle sop only mpacket for non-critical packet which has payload spanning
		// over one buffer
 
		_sphy_mphy4_tx_sop_only_non_critical_mpkt(payload_len, payload_offset, *l$index0[ABD_PKT_TYPE_OFFSET], \
									offset_mod_8, phase1#)


	not_sop_mpacket#:
		// case when mpacket not sop. again using sigmask
		alu[sig_mask_2, --, b, sigmask_phs2_default]			; initialize sig_mask_2
		#if (META_DATA_READ == 0)
			_sphy_mphy4_tx_clear_signal(sig_mask_2, sig_sram_read_1)	; clear sig_sram_read_1
			alu[exe_stat_flag, --, b, 0]		; reset exe_stat_flag
		#else			
			alu[exe_stat_flag, --, b, 1, <<READ_SOP_BD_FLAG_BIT]		; set exe_stat_flag
		#endif

		_sphy_mphy4_tx_not_sop_mpkt(phase1#)


	.end // payload_len payload_offset offset_mod_8 

	not_ready_to_tx#:
		// could not get head of LM queue
		// sets the sigmask using META_DATA_READ
		alu[sig_mask_2, --, b, sigmask_phs2_default]			; initialize sig_mask_2
		#if (META_DATA_READ == 0)
			_sphy_mphy4_tx_clear_signal(sig_mask_2, sig_sram_read_1)	; clear sig_sram_read_1
			alu[exe_stat_flag, --, b, 0]		; reset exe_stat_flag
		#else			
			alu[exe_stat_flag, --, b, 1, <<READ_SOP_BD_FLAG_BIT]		; set exe_stat_flag
		#endif		

		_sphy_mphy4_tx_get_tx_request_for_no_tx(phase1#)

	
#endm



///////////////////////////////////////////////////////////////////////////////
// get_request_from_nn_ring()
//
// Description: reads request from the nn ring and issue meta data read
//
//
// Outputs:					
//							tx_request_lw0: 	packet handle read from nn ring
//
// Inputs:  
//						 	no_request_path:	label to jump in case when nn 
//												ring is empty
//
// Constants:
//						 	None 
//
//
///////////////////////////////////////////////////////////////////////////////
#macro get_request_from_nn_ring(tx_request_lw0, no_request_path)

		#ifndef DISABLE_QUEUE_FULL_CHECK
		// find if LM queue is full
		alu[--, *l$index1[GLOBAL_PKTS_IN_QUEUE_INDEX], -, SKIP_GET_TXR_THRESHOLD]
		// queue full, so don't get tx request, keep tx requests in NN ring
		bgt[no_request_path]
		#endif 	// #ifndef DISABLE_QUEUE_FULL_CHECK
			
			
	
		// get tx request from NN ring
		br_inp_state[nn_empty, no_request_path]

		_sphy_mphy4_tx_debug_incr_counter(@packets_received)

		alu[tx_request_lw0, --, b, *n$index++]		; read the incoming request

		_sphy_mphy4_tx_read_meta_data_from_sram($dl_meta, tx_request_lw0) 

		// set local memory index 0 to queue entry at queue tail of 
		// that port and enqueue this packet 
		_sphy_mphy4_tx_move_tail_ahead(current_tail, port_entry_turnaround)


#endm



											
///////////////////////////////////////////////////////////////////////////////
// sphy_mphy4_tx()
//
// Description:		Packet (POS/Ethernet) transmission
//
// Outputs:					
//					None
//
// Inputs:  
//					None
//
// Constants:
//					None 
//
//	Size:
//		 			?? instructions for worst case
//
///////////////////////////////////////////////////////////////////////////////

#macro sphy_mphy4_tx()

	.begin 
	.reg tcw0_1					; mpkt 1 transmit control word 0
	.reg sigmask_phs2_default 	; default value for signal mask for phase 2
	.reg exe_stat_flag			; thread execution status flags 
	.reg deq_port
	.reg indiref_base				; base to calculate the address field in indirect ref
	.reg $w0 $w1
	.xfer_order $w0 $w1
	.reg tx_request_lw0 current_tail
	.reg sig_mask_2
	#ifdef ADD_L2_HEADER
		#ifdef ETHERNET_TX
			.reg l2_header_exist
			.reg $l2_entry_lw0 $l2_entry_lw1 $l2_entry_lw2 $l2_entry_lw3
			.xfer_order $l2_entry_lw0 $l2_entry_lw1 $l2_entry_lw2 $l2_entry_lw3
		#endif	// #ifdef ETHERNET_TX
	#endif // #ifdef ADD_L2_HEADER
	
	// set this registers to avoid warnings
	.set current_tail
	.set tx_request_lw0
									

		// allocate 5 read xfer registers for reading in the meta data 
		xbuf_alloc[$dl_meta, 5, read]
		.set $dl_meta[0] $dl_meta[1] $dl_meta[2] $dl_meta[3] $dl_meta[4]


		#ifdef ADD_L2_HEADER
			#ifdef ETHERNET_TX
			.set $l2_entry_lw0 $l2_entry_lw1 $l2_entry_lw2 $l2_entry_lw3
			#endif	// #ifdef ETHERNET_TX
		#endif // #ifdef ADD_L2_HEADER

		// gprs initialized so they may save some cycles 

		_sphy_mphy4_tx_init_gprs(indiref_base,  \
				sigmask_phs2_default)

		// The first time through the loop, wait for next thread signal only 

		// wait for my turn
		ctx_arb[sig1_next_context]

		// initialize deq_port to port number handled by this thread
		alu[deq_port, --, B, 0]

		// initialize $w1 to deq_port, this $w1 should not changed or set back 
		// to deq_port at the beginning of following phase_1#:, s can save
		// one instruction cycle in sop_eop (POS min packet case)
		// workaround for HW-bug#1249, copy channel # into tcw1 (reserved in PRM)
		alu[$w1, --, B, deq_port]	


		// wake up next thread in beginning of phase 1 to reduce wakeup latency
		local_csr_wr[SAME_ME_SIGNAL, sig1_next_context_gpr] 
	

	phase1#:

		// set local memory index 0 to queue head for SPHY 
		
		_sphy_mphy4_tx_set_lmindex0_to_queue_head()	

		get_request_from_nn_ring(tx_request_lw0, no_request_path#)

		// got a request

		find_first_request_and_process(1) ; argument value of 1 means that we got a request from nn ring and hence, read meta data

		no_request_path#:

		// no request
		
		find_first_request_and_process(0); ; argument value of 0 means that we dit not get a request from nn ring and hence, read meta data





.end 
#endm // end of #macro sphy_mphy4_tx()

///////////////////////////////////////////////////////////////////////////////
// The main code starts here

main#:
		// initialize the microblock
	#if(TX_PHY_MODE == SPHY_1_32)
		sphy_mphy4_tx_init(port_entry_turnaround, SPHY_1_32)
	#endif // #if(TX_PHY_MODE == SPHY_1_32)


		// call the packet processing block
		sphy_mphy4_tx()

/////////////////////////////////////////////////////////////////////////////////////

#endif 		// __SPHY_MPHY4_TX1_UC__

/////////////////////////////////////////////////////////////////////////////////////


      

