////////////////////////////////////////////////////////////////////////////////
//                                                                     
//                  I N T E L   P R O P R I E T A R Y                   
//                                                                      
//     COPYRIGHT (c)  2001-2002 BY  INTEL  CORPORATION.  ALL RIGHTS          
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//     THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                      
//                        INTEL  CORPORATION                            
//                                                                     
//                     2200 MISSION COLLEGE BLVD                        
//                                                                      
//               SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                      
////////////////////////////////////////////////////////////////////////////////
//
//
//      File Name: sphy_mphy4_tx2.uc
//
//      Purpose: 2nd TX ME in Packet transmitter for Egress IXP2800
//
///////////////////////////////////////////////////////////////////////////////


#ifndef	__SPHY_MPHY4_TX2_UC__
#define	__SPHY_MPHY4_TX2_UC__

// This is 2nd ME in Packet TX microblock
// High-level view of what all MEs do is as follows
/*
1st ME:
	
while (1)
	{
		read request (packet p1) from NN ring from QM.
		Issue SRAM read for metadata of the p1.
		Find first request in LM queue (packet p2) and read the request
		Determine whether the packet is SOP, EOP, none of these or a min packet and call corresponding macro.
		Sends scheduler feedback about packet transmit when reaches EOP
		Calculate DRAM address to read the payload
		Send the request to next ME using NN to transmit a mpacket of p2

		Wait for metadata read signal and int thd signal
		
		Send int thd signal
		Store the metadata read to the tail of LM queue
	}

2nd ME (in two MEs configuration):

while (1)
	{
		for (i=0; I < max_phase; i++)
			{
				phase# i:
					issue tbuf validate of mpacket read in previous iteration and generate signal tbuf_validate_i
					wait for signals tbuf_validate_i, dram_(i+1), prepend_(i+1)
					read a mpacket transmit request from NN ring
					grab a tbuf for the mapacket
					issue dram to tbuf transfer request and generate signal dram_i
					perform flow control function based on port TXFIFO status optionally
					add prepend (l2 header, SPI-4 prepend or leftover) and generate signal prepend_i
					branch to phase_(i+1);
			}
			// here (i+1) means (i+1) mod max_phase;

	}

		
2nd ME (in three MEs configuration):

while (1)
	{
		for (i=0; I < max_phase; i++)
			{
				phase# i:
					issue request for mpacket read in previous iteration to next ME through NN ring
					wait for signals dram_(i+1), prepend_(i+1)
					read a mpacket transmit request from NN ring
					grab a tbuf for the mapacket					
					issue dram to tbuf transfer request and generate signal	dram_i
					add prepend (l2 header, SPI-4 prepend or leftover) and generate signal prepend_i
					branch to phase_(i+1);
			}
			// here (i+1) means (i+1) mod max_phase;

	}


3rd ME (only in three MEs configuration):

while (1)
	{
		for (i=1; I < 3; i++)
			{
				phase# i:
					read mpacket request from NN ring 
					perform flow control function based on port TXFIFO status optionally
					issue tbuf validate of mpacket just read from NN ring
					branch to phase_(i+1);
			}

	}

The interface between 1st ME and 2nd ME is

lw0: meta_data header ptr (used to free a buffer, if buffer is transmitted completeely)
lw1: 0-29 	offset of payload to be transmitted
30	bit clear means the mpacket is eop,
31	bit clear means the mpacket is sop
lw2: 	payload size to be transmitted
 
In case of non-sop mpacket, we have 3 or 6 more LWs on the NN ring, 
The first three more LWs are:

lw3:	leftover_lw0, first longword of leftover from the previous buffer has 
		to be prepended to the current buffer
lw4:	exe_stat_flag  includes information about various condition flags
lw5:	partially created transmit control word 

If LEFTOV_FROM_PB_FLAG_BIT is set in exe_stat_flag, there will be three more 
LWs:

lw6:	leftover_lw1, second longword of leftover from the previous buffer has 
		to be prepended to the current buffer
lw7:	leftover_lw2, third longword of leftover from the previous buffer has 
		to be prepended to the current buffer
lw8:	leftover_lw3, forth longword of leftover from the previous buffer has 
		to be prepended to the current buffer

The interface between 2nd ME and 3rd ME is

lw0: meta_data header ptr (used to free a buffer, if buffer is transmitted completeely)
lw1: transmit control word 0 for this mpacket
lw2: bit 30 - EOB flag: when set, this buffer is completely transmitted, should be freed
	 bit 15 to 0: tbuf element number assigned to this mpacket in 2nd ME

*/


// following definition selects sram transfer registers to be used with 
// dl_meta.uc dispatch loop
#define DL_META_DATA_IN_SXFER

// Following are the optional assembling options for this TX ME:

// ADD_L2_HEADER: enable this microblock to add layer 2 header, it combines 
// with POS_TX option to add PPP header before payload, it combines with
// ETHERNET_TX option to add Ethernet header before payload

// POS_TX: combines with ADD_L2_HEADER option enable this microblock to add
// PPP header before payload

// ADD_PPP_ADDR_AND_CNTRL: combines with ADD_L2_HEADER and POS_TX options 
// enable this microblock add  PPP address and control fields with PPP header 
// before payload

// ETHERNET_TX: combines with ADD_L2_HEADER option enable this microblock to
// add Ethernet header before payload

// VALIDATE_TBUF_IN_THIRD_TX_ME: This TX ME validates tbuf element by default 
// (two TX ME configuration), when this option is set, this TX ME doesn't 
// validate tbuf element, it passes the necessary info to next ME through NN
// ring, and that TX ME validate tbuf element (3 TX ME configuration)

// CHECK_PORT_TX_FIFO_STATUS: enable this TX ME to perform flow control based
// on meia port TX FIFO status to prevent overwritting of port TX FIFO, this 
// option in this TX ME only "VALIDATE_TBUF_IN_THIRD_TX_ME" option not set. 
// When "VALIDATE_TBUF_IN_THIRD_TX_ME" option is set, the optional flow control
// function will take place in 3rd TX ME when this option is set in 3rd TX ME  


// include stdmac.uc in IXPblocks Portable library

#include <stdmac.uc>

// include local memory macros in IXPblocks 

#include <localmem.uc>

// global definition of common constants which are shared by all microblocks

#include "dl_system.h"

// definition of hardware register address 

#include <hardware.h>

// definition of PHY operation mode and misc.

#include <definitions.h>

// definition of constants used

#include "sphy_mphy4_tx.h"

// definition of utilization macros used

#include "sphy_mphy4_tx_util2.uc"

// initialization code for POS/Ethernet TX

#include "sphy_mphy4_tx_init2.uc"

// definition of common utilization macros used

#include "sphy_mphy4_tx_common_util.uc"

// dispatch loop macros

#include "dispatch_loop.uc"

#ifndef VALIDATE_TBUF_IN_THIRD_TX_ME
#ifdef CHECK_PORT_TX_FIFO_STATUS
	.reg @fc_credit
	.reg addr_port_stat
	.if (ctx() == 0)
	.reg $port_stat
	.endif	// .if (ctx() == 0)	
	.reg ind_ref_reg
	.sig volatile sig_msf_read
#endif  // #ifdef CHECK_PORT_TX_FIFO_STATUS
#endif	// #ifndef VALIDATE_TBUF_IN_THIRD_TX_ME

#ifdef COUNTERS
.reg @cntr_base					; start address of Packet TX counters in sram
								; initialized to PACKET_TCX_COUNTER_BASE
.reg $byte_cnt					; sram xfer register to keep value to add to 
								; packets transmitted counter in sram		
#endif //#ifdef	COUNTERS

#ifdef _DEBUG_COUNTERS_
.reg @request_received
.reg @tbufs_transmitted
.reg @buffers_transmitted
#ifndef VALIDATE_TBUF_IN_THIRD_TX_ME
.reg @default_buffers_freed
alu[@default_buffers_freed, --, b, 0]
#ifdef CHECK_PORT_TX_FIFO_STATUS
.reg @is_satisfied
.reg @is_hungry
.reg @is_starving
alu[@is_satisfied, --, b, 0]
alu[@is_hungry, --, b, 0]
alu[@is_starving, --, b, 0]
#endif  // #ifdef CHECK_PORT_TX_FIFO_STATUS
#endif	// #ifndef VALIDATE_TBUF_IN_THIRD_TX_ME

alu[@request_received, --, b, 0]
alu[@tbufs_transmitted, --, b, 0]
alu[@buffers_transmitted, --, b, 0]
#endif	// #ifdef _DEBUG_COUNTERS_

#ifdef PORT_STATUS_COUNTER
	.reg @port_not_ready_numb
	alu[@port_not_ready_numb, --, b, 0]
#endif	// #ifdef PORT_STATUS_COUNTER

#ifdef UNIT_TEST
.reg @buf_freed			; buffer freed for debuging
#endif //UNIT_TEST

.reg addr_tx_seq		; MSF register address to read mpkt txed out of tbuf
// context relative GPR variables:

								; head/tail offset
.reg $txed_port					; xfer register to keep packet txed to notify
								; egress scheduler by reflect write							

#ifdef ADD_L2_HEADER
	#ifdef POS_TX
		.reg ppp_header_base mpls_pkt_type
		move(ppp_header_base, PPP_IP_PROTOCOL)
		immed32[mpls_pkt_type, 0x0281]
	#endif	// #ifdef POS_TX
#endif	// #ifdef ADD_L2_HEADER

///////////////////////////////////////////////////////////////////////////////

//	Global Signals
.sig sig_scratch_access_1 			; signal for scratch ring read
.sig sig_sram_read_1 				; signal 1 for sram read
.sig sig_msf_access_1				; signal 1 for msf access (read or write)
.sig sig_msf_access_2				; signal 2 for msf access (read or write)
.sig sig_msf_access_3				; signal 3 for msf access (read or write)
.sig sig_dram_access_1				; signal 1 for dram access (dram[tbuf_wr..])
.sig sig_dram_access_2				; signal 2 for dram access (dram[tbuf_wr..])
.sig volatile common_init_complete_sig_num		; this is the inter-ME signal 
										; generated by the ME whcih initializes
										; common resources, such as scratch ring,
										; POS TX MEs will wait for this signal at
										; end of their initialization to make 
										; sure all common resource are available 

//	Manual signal allocation
.addr common_init_complete_sig_num	ME_INIT_SIGNAL	; manual signal allocation,
													; ME_INIT_SIGNAL defined as
													; in dl_system.h  

#ifdef DEBUG_TX_PENDING_LOCKED
.sig sig_sram_write_1 				; signal 1 for sram write
#endif //#ifdef DEBUG_TX_PENDING_LOCKED

#ifdef COUNTERS
.sig sig_counter_update									; signal for update counter
#endif //#ifdef COUNTERS		



///////////////////////////////////////////////////////////////////////////////

											
///////////////////////////////////////////////////////////////////////////////
// sphy_mphy4_tx2()
//
// Description:		Packet (POS/Ethernet) transmission
//
// Outputs:					
//					None
//
// Inputs:  
//					None
//
// Constants:
//					None 
//
//	Size:
//		 			?? instructions for worst case
//
///////////////////////////////////////////////////////////////////////////////

#macro sphy_mphy4_tx2()

.begin 

	// ME2 starts here
	// the steps are
	// read from NN
	// grab a tbuf - no ctx_arb between tbuf grabbing and nn read. this ensures ordering, 
	//				 hence, no other ordering required
	// check if sop, mop or eop
	// issue dram writes
	// validate tbuf

	#ifdef SPI4_PREPEND
		#define INTERLEAVING_PHASES 3	// should be 2 or 3
		#if (INTERLEAVING_PHASES < 2)
			#warning "For spi4-prepend interleaving phases has to be >= 2. The value is reset to 2."
			#undef 	INTERLEAVING_PHASES
			#define INTERLEAVING_PHASES 2
		#endif
	#else
		#define INTERLEAVING_PHASES 1	// should be greater than or equal to 1 -- tested for 1, 2 and 3
	#endif

	// the code works in several phases. The dram signal generated by phase i is waited 
	// on phase (i-1) mod n. This gives the dram access sufficient time to complete. The allowed
	// latency for dram access = cycle budget * 8 * number of interleaving phases.

	.reg @packet_payload
	.reg packet_type
	.reg payload_offset
	.reg tcw_sop_eop_base		; gpr keeps tcw base for sop-eop mpacket to 
								; save instruction cycle 

	#ifdef ADD_L2_HEADER
		#ifdef ETHERNET_TX
			.reg l2_header_passed
		#endif	// #ifdef ETHERNET_TX
	#endif // #ifdef ADD_L2_HEADER

	
	#ifndef VALIDATE_TBUF_IN_THIRD_TX_ME
	.sig sig_tbuf_validate
	#endif	// #ifndef VALIDATE_TBUF_IN_THIRD_TX_ME
	.sig msf_seq_read

	#ifdef COUNTERS
		.sig sig_counter_update
	#endif	

	#ifdef SPI4_PREPEND
		.sig sig_prepend_write		// for MSF SPI4 prepend write
		.sig sig_sram_meta_read		// for SRAM meta-data read

		// allocate for SPI4 prepend header information (read bank)
		.reg read $$prepend_meta0 $$prepend_meta1 $$prepend_meta2
		.reg read $$prepend_meta3 $$prepend_meta4 $$prepend_meta5
		.reg read $$prepend_meta6
		.xfer_order $$prepend_meta0 $$prepend_meta1 $$prepend_meta2 \
					$$prepend_meta3 $$prepend_meta4 $$prepend_meta5 $$prepend_meta6
		.set 		$$prepend_meta0 $$prepend_meta1 $$prepend_meta2 \
					$$prepend_meta3 $$prepend_meta4 $$prepend_meta5 $$prepend_meta6

		// allocate for SPI4 prepend header information (write bank)
		.reg write $msf_prepend_lw0
		.reg write $msf_prepend_lw1
		.reg write $msf_prepend_lw2
		.reg write $msf_prepend_lw3
		.xfer_order $msf_prepend_lw0 $msf_prepend_lw1 \
					$msf_prepend_lw2 $msf_prepend_lw3

		.set 		$msf_prepend_lw0 $msf_prepend_lw1 \
					$msf_prepend_lw2 $msf_prepend_lw3

	#endif
	
	// initialize tcw_sop_eop_base to save instruction cycle later
	immed[tcw_sop_eop_base, TCW0_SOP_EOP_BASE_W0]
	immed_w1[tcw_sop_eop_base, TCW0_SOP_EOP_BASE_W1]

	// the following code runs for all phases.. i denotes to phase #
	#define_eval i 0
	#while ( i < INTERLEAVING_PHASES)
		
		.sig sig_dram_to_tbuf_/**/i

		#ifndef SPI4_PREPEND
			.sig sig_prepend_write_/**/i
		#endif

		#ifdef SPI4_PREPEND
			.reg addr_of_tbuf/**/i
			immed32[addr_of_tbuf/**/i, 0]
		#endif


		.reg meta_data_ptr_/**/i
		.reg tcw_/**/i
		.reg sop_eop_/**/i
		#ifndef VALIDATE_TBUF_IN_THIRD_TX_ME
		.reg addr_of_tx_control_/**/i
		immed32[addr_of_tx_control_/**/i, 0]
		#else	// VALIDATE_TBUF_IN_THIRD_TX_ME
		.reg flags_and_tbuf_element_/**/i
		immed32[flags_and_tbuf_element_/**/i, 0]
		#endif	// #ifndef VALIDATE_TBUF_IN_THIRD_TX_ME

		immed32[meta_data_ptr_/**/i, 0]	
		immed32[tcw_/**/i, 0]	
		immed32[sop_eop_/**/i, 0]

		.reg write $msf_prepend_lw0_/**/i
		.reg write $msf_prepend_lw1_/**/i
		.reg write $msf_prepend_lw2_/**/i
		.reg write $msf_prepend_lw3_/**/i
		.xfer_order $msf_prepend_lw0_/**/i $msf_prepend_lw1_/**/i \
					$msf_prepend_lw2_/**/i $msf_prepend_lw3_/**/i

		.set $msf_prepend_lw0_/**/i $msf_prepend_lw1_/**/i \
					$msf_prepend_lw2_/**/i $msf_prepend_lw3_/**/i

	#define_eval i (i + 1)
	#endloop

	
	.reg exe_stat_flag			; thread execution status flags 
	.reg leftover_bytes			; size of leftover in byte
	.reg deq_port
	.reg indiref_base				; base to calculate the address field in indirect ref
	.reg $w0 $w1
	.xfer_order $w0 $w1

	.reg tbuf_element
	.reg offset_mod_8
	.reg indir_ref dram_addr

#ifndef SPI4_PREPEND
	.reg addr_of_tbuf
#endif


	// registers declaration for flow control
	// ok_to_schedule_cnt is number of mpackets that be scheduled before reading the calendar seq num again
	// old_seq_read is the calendar seq num when last read
	// read_tx_seq_mutex_flag is to ensure two threads are not simulataneously reading calendar seq num
	// msf_read_timestamp_current is a sequence number to avoid updating old_seq_read with stale value
	// every update to old_seq_read is marked with a timestamp and any subsequent update attempts
	// with older timestamp will be thrown away
	.reg global @ok_to_schedule_cnt @old_seq_read @read_tx_seq_mutex_flag @msf_read_timestamp_current 
	.reg @msf_read_timestamp_old
	.reg thd_timestamp
	// init of flow control registers
	move(@read_tx_seq_mutex_flag, 0)
	move(@ok_to_schedule_cnt, TBUF_MAX_VALID_ELEMENTS);
	move(@old_seq_read, 0)
	move(@msf_read_timestamp_current, 0)
	move(@msf_read_timestamp_old, @msf_read_timestamp_current)
	alu[thd_timestamp, --, b, 0]




		// allocate 1 read xfer registers for reading mpackets actually transmitted
		xbuf_alloc[$mpkts_sent, 1, read]

		// gprs initialized so they may save some cycles 

		_sphy_mphy4_tx_init_gprs(indiref_base, sigmask_phs1_default, \
				sigmask_phs2_default)


		// initialize deq_port to port number handled by this thread
		alu[deq_port, --, B, 0]

		// initialize $w1 to deq_port, this $w1 should not changed or set back 
		// to deq_port at the beginning of following phase_1#:, s can save
		// one instruction cycle in sop_eop (POS min packet case)
		// workaround for HW-bug#1249, copy channel # into tcw1 (reserved in PRM)
		alu[$w1, --, B, deq_port]	

	// create default SIGMASKS for each phase
	#define_eval i 0
	#while ( i < INTERLEAVING_PHASES)

		#define_eval next_i ( ( i + 1) % INTERLEAVING_PHASES)
		#define_eval prev_i ( ( i + INTERLEAVING_PHASES - 1) % INTERLEAVING_PHASES)

	

		.reg default_sigmask_/**/i signal_mask_/**/i
		alu[default_sigmask_/**/i, --, b, 0]
		
		// ith phase in next iteration will wait for signal generated by (i+1)th phase in this iteration
		_sphy_mphy4_tx_set_signal[default_sigmask_/**/i, sig_dram_to_tbuf_/**/next_i]
		_sphy_mphy4_tx_set_signal[default_sigmask_/**/i, sig_dram_to_tbuf_/**/next_i + 1] 
		
		#ifdef ADD_L2_HEADER
			#ifdef POS_TX
			_sphy_mphy4_tx_set_signal[default_sigmask_/**/i, sig_prepend_write_/**/next_i]
			#endif	// #ifdef POS_TX
			#ifdef ETHERNET_TX	
			_sphy_mphy4_tx_set_signal[default_sigmask_/**/i, sig_prepend_write_/**/next_i]
			#endif	// #ifdef ETHERNET_TX
		#else
			#ifdef SPI4_PREPEND
				_sphy_mphy4_tx_set_signal[default_sigmask_/**/i, sig_sram_meta_read]
				_sphy_mphy4_tx_set_signal[default_sigmask_/**/i, sig_prepend_write]
			#endif
		#endif	// #ifdef ADD_L2_HEADER
		
		#ifndef VALIDATE_TBUF_IN_THIRD_TX_ME
		_sphy_mphy4_tx_set_signal[default_sigmask_/**/i, sig_tbuf_validate]
		#endif	// #ifndef VALIDATE_TBUF_IN_THIRD_TX_ME		
		_sphy_mphy4_tx_set_signal[default_sigmask_/**/i, msf_seq_read]
		
		#ifdef COUNTERS
			_sphy_mphy4_tx_set_signal[default_sigmask_/**/i, sig_counter_update]
		#endif
		
		// initialize sigmasks

		alu[signal_mask_/**/i, --, b, default_sigmask_/**/i]
	
		// clear some signals for the first iteration -- as all prev memory accesses have not been issued
		_sphy_mphy4_tx_clear_signal[signal_mask_/**/i, sig_dram_to_tbuf_/**/next_i]
		_sphy_mphy4_tx_clear_signal[signal_mask_/**/i, sig_dram_to_tbuf_/**/next_i + 1]
		#ifdef ADD_L2_HEADER
			#ifdef POS_TX
			_sphy_mphy4_tx_clear_signal[signal_mask_/**/i, sig_prepend_write_/**/next_i]
			#endif	// #ifdef POS_TX
			#ifdef ETHERNET_TX	
			_sphy_mphy4_tx_clear_signal[signal_mask_/**/i, sig_prepend_write_/**/next_i]
			#endif	// #ifdef ETHERNET_TX
		#else
			#ifdef SPI4_PREPEND
				_sphy_mphy4_tx_clear_signal[signal_mask_/**/i, sig_sram_meta_read]
			#endif
		#endif	// #ifdef ADD_L2_HEADER

	#define_eval i (i + 1)
	#endloop // end of loop

	#define_eval last (INTERLEAVING_PHASES -1)
	alu[signal_mask_/**/last, --, b, 0] ; init case -- no signals to wait for the first time
	
	// issue to first read for msg seq num
	// issue read for calendar seq num
	msf[read, $mpkts_sent[0], addr_tx_seq, 0, 1], sig_done[msf_seq_read]
	// note the timestamp of reading the calendar seq num and increment
	// the globally available timestamp
	alu[thd_timestamp, @msf_read_timestamp_current, +, 1]
	alu[@msf_read_timestamp_current, @msf_read_timestamp_current, +, 1]	
	_sphy_mphy4_tx_set_signal[signal_mask_/**/last, msf_seq_read]
	
		
	
	// waiting for signal mask for first time
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, signal_mask_/**/last]

	ctx_arb[--] 
	
	
	// the main code -- for each phase starts here

	#define_eval i 0

	#while (i < INTERLEAVING_PHASES)

	// calculate next and prev phases
	#define_eval next_i ( ( i + 1) % INTERLEAVING_PHASES)
	#define_eval prev_i ( ( i + INTERLEAVING_PHASES - 1) % INTERLEAVING_PHASES)

	#ifdef SPI4_PREPEND
		#define_eval prev_prev_i ( ( prev_i + INTERLEAVING_PHASES - 1) % INTERLEAVING_PHASES)
	#else
		#define_eval prev_prev_i i
	#endif

	phase/**/i/**/#:

		
		#ifdef COUNTERS
		.io_completed sig_counter_update
		#endif

		#ifndef VALIDATE_TBUF_IN_THIRD_TX_ME
		.io_completed sig_tbuf_validate
		#endif	// #ifndef VALIDATE_TBUF_IN_THIRD_TX_ME
		.io_completed sig_dram_to_tbuf_/**/i	
		.io_completed msf_seq_read

		#ifdef SPI4_PREPEND
		.io_completed sig_sram_meta_read
		.io_completed sig_prepend_write
		#else
		.io_completed sig_prepend_write_/**/i
		#endif


		// initialize sigmask
		alu[signal_mask_/**/prev_i, --, b, default_sigmask_/**/prev_i]


		#ifndef VALIDATE_TBUF_IN_THIRD_TX_ME
		#ifdef CHECK_PORT_TX_FIFO_STATUS
	
		// get port status into $port_stat in thread 0 for all thread to share

		alu[--, ind_ref_reg, or, 0]
		msf[read, $port_stat, addr_port_stat, 0, 1], indirect_ref, sig_done[sig_msf_read]
	
		#endif  // #ifdef CHECK_PORT_TX_FIFO_STATUS
		#endif	// #ifndef VALIDATE_TBUF_IN_THIRD_TX_ME

		alu[--, --, b, meta_data_ptr_/**/prev_prev_i]

		#ifndef VALIDATE_TBUF_IN_THIRD_TX_ME
		beq[no_tbuf_to_validate/**/i/**/#]

		#ifndef VALIDATE_TBUF_IN_THIRD_TX_ME
		#ifdef CHECK_PORT_TX_FIFO_STATUS
	
		// before validating tbuf, check whether TX FIFO of the port ready to
		// receive another mpacket

		alu[@fc_credit, @fc_credit, -, 1]
		ble[port_tx_fifo_full_phase_/**i**/#]
		port_tx_fifo_check_done_phase_/**i**/#:
	
		#endif  // #ifdef CHECK_PORT_TX_FIFO_STATUS
		#endif	// #ifndef VALIDATE_TBUF_IN_THIRD_TX_ME

		// validate previous tbuf element

		alu[$w0, --, b, tcw_/**/prev_prev_i]		; element trnsmit control word

		// write transmit contrl word (TCW) to TBUF_ELEMENT_CTRL_V_# corresponding
		// the tbuf_element to initiate transmission
		msf[write, $w0, addr_of_tx_control_/**/prev_prev_i, 0, 2], sig_done[sig_tbuf_validate]
		_sphy_mphy4_tx_debug_incr_counter(@tbufs_transmitted)	

		#ifndef SPI4_PREPEND
			alu[--, sop_eop_/**/i, and, 1]
			bne[no_buffer_free/**/i/**/#]

			_sphy_mphy4_tx_free_buffer(meta_data_ptr_/**/i, buf_free_label/**/i)	; free buffer

			buffer_free_continue/**/i/**/#:

			_sphy_mphy4_tx_debug_incr_counter(@buffers_transmitted)

			no_buffer_free/**/i/**/#:
		#endif

		
		tbuf_validate_done_/**/i/**/#:

		#else	// VALIDATE_TBUF_IN_THIRD_TX_ME
		
		// pass previous mpacket to next TX ME to validatde
		
		beq[pass_mpacket_to_next_me_done_/**/i/**/#]
		
		// pass necessary info to third TX ME to validate tbuf and free
		// buffer if this is an end of buffer mpacket
				
		sop_eop_waiting_for_nn_to_become_not_full_/**/i#:
		br_inp_state[nn_full, sop_eop_waiting_for_nn_to_become_not_full_/**/i#]
	
			alu[*n$index++, --, b, meta_data_ptr_/**/prev_prev_i]	; meta data header ptr
			alu[*n$index++, --, b, tcw_/**/prev_prev_i]			; tcw
			alu[*n$index++, --, b, flags_and_tbuf_element_/**/prev_prev_i]
			_sphy_mphy4_tx_debug_incr_counter(@tbufs_transmitted)	

		pass_mpacket_to_next_me_done_/**/i/**/#:
		#endif	// #ifndef VALIDATE_TBUF_IN_THIRD_TX_ME


		#ifdef COUNTERS
			// update counters in sram (10) 	
			_sphy_mphy4_tx_update_counters_sop_eop($byte_cnt, tcw_/**/prev_prev_i, \
															sig_counter_update)			
		#endif //#ifdef	COUNTERS


		// Writing prepend header information to TBUF for SOP and SEP cases
		#ifdef SPI4_PREPEND

			// if there was incoming request, there will be TBUF validation
			alu[--, --, B, meta_data_ptr_/**/prev_i]
			beq[spi4_prepend_write_clear_signal_/**/i#]

			alu[--, sop_eop_/**/prev_i, and, 1]
			bne[no_buffer_free/**/i/**/#]
			
     		_sphy_mphy4_tx_free_buffer(meta_data_ptr_/**/prev_i, buf_free_label/**/i)	; free buffer

			buffer_free_continue/**/i/**/#:

			_sphy_mphy4_tx_debug_incr_counter(@buffers_transmitted)

			no_buffer_free/**/i/**/#:

			// write prepend only for sop and sep cases
			br_bclr[sop_eop_/**/prev_i, 1, spi4_prepend_write_start_/**/i#]

			spi4_prepend_write_clear_signal_/**/i#:

			// clear this signal if prepend is not written
			_sphy_mphy4_tx_clear_signal[signal_mask_/**/i, sig_prepend_write]
			br[spi4_prepend_write_done_/**/i#]
						
			spi4_prepend_write_start_/**/i#:

			move($msf_prepend_lw0, $$prepend_meta[3])
			move($msf_prepend_lw1, $$prepend_meta[4])
			move($msf_prepend_lw2, $$prepend_meta[5])
			move($msf_prepend_lw3, $$prepend_meta[6])
			msf[write, $msf_prepend_lw0, addr_of_tbuf/**/prev_i, 0, 4], sig_done[sig_prepend_write]	

			spi4_prepend_write_done_/**/i#:

		#endif	// SPI4_PREPEND

		#if IS_IXPTYPE(__IXP28XX)
			// we have a request here
			check_if_tbuf_full_/**/i/**/#:
			
			// check if ok to schedule
			alu[@ok_to_schedule_cnt, @ok_to_schedule_cnt, -, 1]
			ble[tbuf_full/**/i/**/#]

			tbuf_full_check_done_/**/i/**/#:
		#endif
			

		// now read new request and process it

		br_inp_state[nn_empty, no_new_request/**/i/**/#]
		
		// issue read for calendar seq num
		msf[read, $mpkts_sent[0], addr_tx_seq, 0, 1], sig_done[msf_seq_read]
		// note the timestamp of reading the calendar seq num and increment
		// the globally available timestamp
		alu[thd_timestamp, @msf_read_timestamp_current, +, 1]
		alu[@msf_read_timestamp_current, @msf_read_timestamp_current, +, 1]	

		_sphy_mphy4_tx_debug_incr_counter(@request_received)
			
		// read rest of words from NN ring
		alu[meta_data_ptr_/**/i, --, b, *n$index++]
		alu[sop_eop_/**/i, --, b, *n$index, >>30]

		#ifndef ADD_L2_HEADER	

		// check whether eop, sop, both (min packet) or none
		bne[not_sop_eop_/**/i/**/#], defer[2]
			alu[payload_offset, *n$index++, and~, FLAG_BITS_FROM_MSB, <<29]
			alu[@packet_payload, --, b, *n$index++]	

		#else	// ADD_L2_HEADER	


			#ifdef POS_TX
			// check whether eop, sop, both (min packet) or none
			#ifdef DYNAMIC_PPP_ENCAP
				bne[not_sop_eop_/**/i/**/#], defer[3]
			#else
				bne[not_sop_eop_/**/i/**/#], defer[2]
			#endif
				alu[payload_offset, *n$index++, and~, FLAG_BITS_FROM_MSB, <<29]
				#ifdef DYNAMIC_PPP_ENCAP
					alu[@packet_payload, 0, +16, *n$index]	 ;Dynamic PPP Encap
					alu_shf[packet_type, --, B, *n$index++, >>16]	 ;Dynamic PPP Encap  */
					// MPLS encoding value is 0x8, i.e. bit 3 is set.
					// These 2 lines of code work with the assumption that ipv4 and ipv6
					// header types are 0x21 and 0x57 respectively. As such, both ipv4 and
					// ipv6 header types have bit 3 clear.
					br_bclr[packet_type, 3, not_mpls_type_sep#]
					alu[packet_type, --, B, mpls_pkt_type]

not_mpls_type_sep#:
				#else
					alu[@packet_payload, --, b, *n$index++]	
				#endif
			#endif	// #ifdef POS_TX

			#ifdef ETHERNET_TX
				alu_shf[l2_header_passed, 1, and, *n$index, >>29]
				alu[payload_offset, *n$index++, and~, FLAG_BITS_FROM_MSB, <<29]
				alu[@packet_payload, --, b, *n$index++]
				br=byte[l2_header_passed, 0, 0, get_l2_table_entry_done_/**/i#]	
	
			// get l2 table entry from  NN ring
		
			wait_for_nn_to_become_available_1/**/i/**/#:
			br_inp_state[nn_empty, wait_for_nn_to_become_available_1/**/i/**/#]

					alu[$msf_prepend_lw0_/**/i, --, b, *n$index++]	
					alu[$msf_prepend_lw1_/**/i, --, b, *n$index++]	
					alu[$msf_prepend_lw2_/**/i, --, b, *n$index++]	
			wait_for_nn_to_become_available_2/**/i/**/#:
			br_inp_state[nn_empty, wait_for_nn_to_become_available_2/**/i/**/#]
					alu[$msf_prepend_lw3_/**/i, --, b, *n$index++]

					// read two extra two longword (duplicate of the fouth 
					// longword of the l2 table entry), because this ME NN
					// ring become empty when valid element on it less than
					// or equal to 2 based on it's initialization 
							
					alu[$msf_prepend_lw3_/**/i, --, b, *n$index++]
					alu[$msf_prepend_lw3_/**/i, --, b, *n$index++]
				get_l2_table_entry_done_/**/i#:
				// check whether eop, sop, both (min packet) or none
				alu[--, sop_eop_/**/i, -, 0]
				bne[not_sop_eop_/**/i/**/#]
			#endif	// #ifdef ETHERNET_TX
		
		#endif	// #ifndef ADD_L2_HEADER
			
		// sop_eop code
		alu[tcw_/**/i, --, B, tcw_sop_eop_base]			; initialize
		#ifdef ADD_L2_HEADER
			#ifdef ETHERNET_TX
				br=byte[l2_header_passed, 0, 1, no_need_to_change_tcw_/**/i#]	
				immed_w1[tcw_/**/i, 0]		; l2 header in payload
				_sphy_mphy4_tx_clear_signal[signal_mask_/**/i, sig_prepend_write_/**/i]
			no_need_to_change_tcw_/**/i#:
			#endif	// #ifdef ETHERNET_TX
		#endif // #ifdef ADD_L2_HEADER

		// grab a tbuf element
		
		alu[tbuf_element, --, b, *l$index1[GLOBAL_AVAIL_TBUF_ELEMENT_INDEX]]

		_sphy_mphy4_tx_update_next_availble_tbuf_ele()

		// now get this tbuf_ele address (addr_of_tbuf_1)
		#ifdef SPI4_PREPEND
			_sphy_mphy4_tx_get_tbuf_addr(addr_of_tbuf/**/i, tbuf_element)
		#else
			_sphy_mphy4_tx_get_tbuf_addr(addr_of_tbuf, tbuf_element)
		#endif

		_sphy_mphy4_tx_update_tcw_sop_eop(tcw_/**/i, tbuf_element, @packet_payload)

		#ifndef VALIDATE_TBUF_IN_THIRD_TX_ME		
		.begin
		.reg tmp
		alu_shf[tmp, --, B, tbuf_element, <<TX_CNTRL_ADDR_SHFT]
		alu[addr_of_tx_control_/**/i, tmp, +, *l$index1[GLOBAL_CTW_VD_BASE_INDEX]]
		.end
		#endif	// #ifndef VALIDATE_TBUF_IN_THIRD_TX_ME

		sop_eop_ready_to_transmit/**/i/**/#:
			// dram access takes longer latency, so do it first
			// move payload to tbuf [9]
	
			alu[offset_mod_8, payload_offset, and, 0x7]

		#ifdef SPI4_PREPEND
			_sphy_mphy4_tx_move_sop_paylo_to_tbuf(indir_ref, dram_addr, meta_data_ptr_/**/i, \
						addr_of_tbuf/**/i, indiref_base, payload_offset, @packet_payload, \
						offset_mod_8, sig_dram_to_tbuf_/**/i) 
		#else
			_sphy_mphy4_tx_move_sop_paylo_to_tbuf(indir_ref, dram_addr, meta_data_ptr_/**/i, \
						addr_of_tbuf, indiref_base, payload_offset, @packet_payload, \
						offset_mod_8, sig_dram_to_tbuf_/**/i)
		#endif


		#ifdef ADD_L2_HEADER
			#ifdef POS_TX
	    	// write  PPP header for SOP mpacket
			#ifdef DYNAMIC_PPP_ENCAP
				#ifndef ADD_PPP_ADDR_AND_CNTRL
					alu_shf[$msf_prepend_lw0_/**/i, --, b, packet_type, \
										<<IP_PROTOCOL_SHFT] ;Dynamic PPP Encap
				#else	//#ifndef ADD_PPP_ADDR_AND_CNTRL
					alu_shf[$msf_prepend_lw0_/**/i, ppp_header_base, or, \
							packet_type, <<IP_PROTOCOL_SHFT] ;Dynamic PPP Encap
				#endif	// #ifndef ADD_PPP_ADDR_AND_CNTRL
			#else	//  not DYNAMIC_PPP_ENCAP
				alu[$msf_prepend_lw0_/**/i, --, b, ppp_header_base]
			#endif	// #ifdef DYNAMIC_PPP_ENCAP

		    // write PPP header to tbuf
			msf[write, $msf_prepend_lw0_/**/i, addr_of_tbuf, 0, 1], \
											sig_done[sig_prepend_write_/**/i]
			#endif	// #ifdef POS_TX

			#ifdef ETHERNET_TX	
				br!=byte[l2_header_passed, 0, 1, l2_header_write_done_/**/i#]	
				msf[write, $msf_prepend_lw0_/**/i, addr_of_tbuf, 0, 4], sig_done[sig_prepend_write_/**/i]
				l2_header_write_done_/**/i#:
			#endif	// #ifdef ETHERNET_TX
		#endif //ADD_L2_HEADER

		// Read 4 LWs from meta-data for SPI4 prepend header, starting from LW3
		// The call is optimized for min packet cases. Correct macro should be:
		//	dl_meta_load_cache(meta_data_ptr_/**/i, $$prepend_meta, sig_sram_meta_read, 3, 4)
		#ifdef SPI4_PREPEND
		.begin
		.reg sram_offset
			ld_field_w_clr[sram_offset, 1000, meta_data_ptr_/**/i]
			ld_field[sram_offset, 0111, meta_data_ptr_/**/i, <<2]
			sram[read, $$prepend_meta3, 12, sram_offset, 4], sig_done[sig_sram_meta_read]					
		.end
		#endif	//ADD_L2_HEADER

		alu_shf[tcw_/**/i, tcw_/**/i, OR, offset_mod_8, <<TCW0_PAYLOAD_OFFSET_LSB_LOC]
	
		#ifdef VALIDATE_TBUF_IN_THIRD_TX_ME
		alu[flags_and_tbuf_element_/**/i, tbuf_element, or, 1, <<FREE_BUFFER_FLAG_BIT]
		#endif	// #ifdef VALIDATE_TBUF_IN_THIRD_TX_ME

		phase/**/i/**/wait_signals#:	

		// waiting for signal mask of phase i
	    local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, signal_mask_/**/i]

		#ifdef SPI4_PREPEND
			ctx_arb[--], br[phase/**/next_i/**/_sop_eop#]
		#else
			ctx_arb[--], br[phase/**/next_i/**/#]  
		#endif

		// unrolling codes to optimize for sop_eop cases:
		//		Several checks/copies are removed since sop_eop is assumed
		#ifdef SPI4_PREPEND

			phase/**/i/**/_sop_eop#:

				#ifdef COUNTERS
				.io_completed sig_counter_update
				#endif

				#ifndef VALIDATE_TBUF_IN_THIRD_TX_ME
				.io_completed sig_tbuf_validate
				#endif	// #ifndef VALIDATE_TBUF_IN_THIRD_TX_ME
				.io_completed sig_dram_to_tbuf_/**/i
				.io_completed sig_sram_meta_read
				.io_completed sig_prepend_write	
				.io_completed msf_seq_read

				// initialize sigmask
				alu[signal_mask_/**/prev_i, --, b, default_sigmask_/**/prev_i]		

				alu[--, --, b, meta_data_ptr_/**/prev_prev_i]
				beq[no_tbuf_to_validate/**/i/**/#]
		
				#ifndef VALIDATE_TBUF_IN_THIRD_TX_ME

					#ifdef CHECK_PORT_TX_FIFO_STATUS
	
					// before validating tbuf, check whether TX FIFO of the port ready to
					// receive another mpacket

						alu[@fc_credit, @fc_credit, -, 1]
						ble[port_tx_fifo_full_phase_sop_eop_/**i**/#]
						port_tx_fifo_check_done_phase_sop_eop_/**i**/#:
	
					#endif  // #ifdef CHECK_PORT_TX_FIFO_STATUS

				alu[$w0, --, b, tcw_/**/prev_prev_i]	; element trnsmit control word

				// write transmit contrl word (TCW) to TBUF_ELEMENT_CTRL_V_# corresponding
				// the tbuf_element to initiate transmission
				msf[write, $w0, addr_of_tx_control_/**/prev_prev_i, 0, 2], sig_done[sig_tbuf_validate]
				_sphy_mphy4_tx_debug_incr_counter(@tbufs_transmitted)

				#else	// VALIDATE_TBUF_IN_THIRD_TX_ME

				// pass previous mpacket to next TX ME to validatde
		
				beq[pass_mpacket_to_next_me_done_sop_eop_/**/i/**/#]
		
				// pass necessary info to third TX ME to validate tbuf and free
				// buffer if this is an end of buffer mpacket
				
				sop_eop_waiting_for_nn_to_become_not_full_sop_eop_/**/i#:
				br_inp_state[nn_full, \
						sop_eop_waiting_for_nn_to_become_not_full_sop_eop_/**/i#]
	
				alu[*n$index++, --, b, meta_data_ptr_/**/prev_prev_i]	; meta data header ptr
				alu[*n$index++, --, b, tcw_/**/prev_prev_i]			; tcw
				alu[*n$index++, --, b, flags_and_tbuf_element_/**/prev_prev_i]
				_sphy_mphy4_tx_debug_incr_counter(@tbufs_transmitted)

				pass_mpacket_to_next_me_done_sop_eop_/**/i/**/#:		

				#endif	// #ifndef VALIDATE_TBUF_IN_THIRD_TX_ME


				#ifdef COUNTERS
					// update counters in sram (10)
					_sphy_mphy4_tx_update_counters_sop_eop($byte_cnt, tcw_/**/prev_prev_i, \
																sig_counter_update)
				#endif //#ifdef	COUNTERS	
			
				_sphy_mphy4_tx_free_buffer(meta_data_ptr_/**/prev_i, buf_free_label_sop_eop/**/i)

			buffer_free_continue/**/i/**/_sop_eop#:

				_sphy_mphy4_tx_debug_incr_counter(@buffers_transmitted)

				// Writing prepend header information to TBUF for SOP and SEP cases
				move($msf_prepend_lw0, $$prepend_meta[3])
				move($msf_prepend_lw1, $$prepend_meta[4])
				move($msf_prepend_lw2, $$prepend_meta[5])
				move($msf_prepend_lw3, $$prepend_meta[6])
				msf[write, $msf_prepend_lw0, addr_of_tbuf/**/prev_i, 0, 4], sig_done[sig_prepend_write]	

			// now read new request and process it

			#if IS_IXPTYPE(__IXP28XX)
			
					// check if ok to schedule
					alu[@ok_to_schedule_cnt, @ok_to_schedule_cnt, -, 1]
					ble[tbuf_full/**/i/**/#]

			#endif

			br_inp_state[nn_empty, no_new_request/**/i/**/#]
		
			// issue read for calendar seq num
			msf[read, $mpkts_sent[0], addr_tx_seq, 0, 1], sig_done[msf_seq_read]
			// note the timestamp of reading the calendar seq num and increment
			// the globally available timestamp
			alu[thd_timestamp, @msf_read_timestamp_current, +, 1]
			alu[@msf_read_timestamp_current, @msf_read_timestamp_current, +, 1]	
			_sphy_mphy4_tx_debug_incr_counter(@request_received)
			
			// read rest of words from NN ring
			alu[meta_data_ptr_/**/i, --, b, *n$index++]
			alu[sop_eop_/**/i, --, b, *n$index, >>30]

			// check whether eop, sop, both (min packet) or none
			bne[not_sop_eop_/**/i/**/#], defer[2]
				alu[payload_offset, *n$index++, and~, FLAG_BITS_FROM_MSB, <<29]
				alu[@packet_payload, --, b, *n$index++]	
		
			// sop_eop code
			immed[tcw_/**/i, TCW0_SOP_EOP_BASE_W0]
			immed_w1[tcw_/**/i, TCW0_SOP_EOP_BASE_W1]

			// grab a tbuf element
	
			alu[tbuf_element, --, b, *l$index1[GLOBAL_AVAIL_TBUF_ELEMENT_INDEX]]

			_sphy_mphy4_tx_update_next_availble_tbuf_ele()

			// now get this tbuf_ele address (addr_of_tbuf_1) 
			_sphy_mphy4_tx_get_tbuf_addr(addr_of_tbuf/**/i, tbuf_element)

			// check tbuf full to avoid overwritting tbuf [9]

			_sphy_mphy4_tx_update_tcw_sop_eop(tcw_/**/i, tbuf_element, \
															@packet_payload)
		
			#ifndef VALIDATE_TBUF_IN_THIRD_TX_ME		
			.begin
			.reg tmp
			alu_shf[tmp, --, B, tbuf_element, <<TX_CNTRL_ADDR_SHFT]
			alu[addr_of_tx_control_/**/i, tmp, +, *l$index1[GLOBAL_CTW_VD_BASE_INDEX]]
			.end
			#endif	// #ifndef VALIDATE_TBUF_IN_THIRD_TX_ME
	
			// dram access takes longer latency, so do it first
			// move payload to tbuf [9]

			alu[offset_mod_8, payload_offset, and, 0x7]

			_sphy_mphy4_tx_move_sop_paylo_to_tbuf(indir_ref, dram_addr, meta_data_ptr_/**/i, \
						addr_of_tbuf/**/i, indiref_base, payload_offset, @packet_payload, \
						offset_mod_8, sig_dram_to_tbuf_/**/i) 

			// Read 4 LWs from meta-data for SPI4 prepend header, starting from LW3
			.begin
			.reg sram_offset
				ld_field_w_clr[sram_offset, 1000, meta_data_ptr_/**/i]
				ld_field[sram_offset, 0111, meta_data_ptr_/**/i, <<2]
				sram[read, $$prepend_meta3, 12, sram_offset, 4], sig_done[sig_sram_meta_read]					
			.end

			alu_shf[tcw_/**/i, tcw_/**/i, OR, offset_mod_8, <<TCW0_PAYLOAD_OFFSET_LSB_LOC]

			#ifdef VALIDATE_TBUF_IN_THIRD_TX_ME
			alu[flags_and_tbuf_element_/**/i, tbuf_element, or, 1, <<FREE_BUFFER_FLAG_BIT]
			#endif	// #ifdef VALIDATE_TBUF_IN_THIRD_TX_ME

			phase/**/i/**/_sop_eop_wait_signals#:		

			// waiting for signal mask of phase i
		    local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, signal_mask_/**/i]

			ctx_arb[--], br[phase/**/next_i/**/_sop_eop#]

			// free buffer by dl_buf_free

			buf_free_label_sop_eop/**/i#:
			dl_buf_free(meta_data_ptr_/**/i, BUF_FREE_LIST0)
			#ifndef VALIDATE_TBUF_IN_THIRD_TX_ME
			_sphy_mphy4_tx_debug_incr_counter(@default_buffers_freed)
			#endif	// #ifndef VALIDATE_TBUF_IN_THIRD_TX_ME
			br[buffer_free_continue/**/i/**/_sop_eop#]

		#endif	// end of SPI4_PREPEND phase/**/i/**/_sop_eop#
	
		#ifndef VALIDATE_TBUF_IN_THIRD_TX_ME

		// free buffer by dl_buf_free

		buf_free_label/**/i#:
			dl_buf_free(meta_data_ptr_/**/i, BUF_FREE_LIST0)
			_sphy_mphy4_tx_debug_incr_counter(@default_buffers_freed)

			br[buffer_free_continue/**/i/**/#]
		#endif	// #ifndef VALIDATE_TBUF_IN_THIRD_TX_ME

		#define_eval i ( i + 1)
		#endloop
		
	
	// non-critical branches
	// the code for non-critical path for each phase is defined below
	
	#define_eval i 0

	#while (i < INTERLEAVING_PHASES)

	#define_eval next_i ( ( i + 1) % INTERLEAVING_PHASES)
	#define_eval prev_i ( ( i + INTERLEAVING_PHASES - 1) % INTERLEAVING_PHASES)


		#ifndef VALIDATE_TBUF_IN_THIRD_TX_ME
		no_tbuf_to_validate/**/i/**/#:
			#ifndef VALIDATE_TBUF_IN_THIRD_TX_ME
			// clear the signal of tbuf_validate
			_sphy_mphy4_tx_clear_signal[signal_mask_/**/i, sig_tbuf_validate]


			#ifndef VALIDATE_TBUF_IN_THIRD_TX_ME
			#ifdef CHECK_PORT_TX_FIFO_STATUS
			_sphy_mphy4_tx_update_fc_credit_for_no_tbuf_validation(@fc_credit,
							*$index)	
			#endif  // #ifdef CHECK_PORT_TX_FIFO_STATUS
			#endif	// #ifndef VALIDATE_TBUF_IN_THIRD_TX_ME



			#endif	// #ifndef VALIDATE_TBUF_IN_THIRD_TX_ME

			#ifdef COUNTERS
				_sphy_mphy4_tx_clear_signal[signal_mask_/**/i, sig_counter_update]
			#endif		
			br[tbuf_validate_done_/**/i/**/#]
		#endif	// #ifndef VALIDATE_TBUF_IN_THIRD_TX_ME

		
		no_new_request/**/i/**/#:
			
			_sphy_mphy4_tx_clear_signal[signal_mask_/**/i, msf_seq_read]
			
			#if IS_IXPTYPE(__IXP28XX)
			// since did not schedule anything, so bump up ok_to_schedule_cnt
			alu[@ok_to_schedule_cnt, @ok_to_schedule_cnt, +, 1]

			#endif
			
			// clear the signals for dram access and prepend write in appropriate bitmasks
			_sphy_mphy4_tx_clear_signal[signal_mask_/**/prev_i, sig_dram_to_tbuf_/**/i]
			_sphy_mphy4_tx_clear_signal[signal_mask_/**/prev_i, sig_dram_to_tbuf_/**/i + 1]
			#ifdef ADD_L2_HEADER
				_sphy_mphy4_tx_clear_signal[signal_mask_/**/prev_i, sig_prepend_write_/**/i]
			#else
				#ifdef SPI4_PREPEND
					_sphy_mphy4_tx_clear_signal[signal_mask_/**/i, sig_sram_meta_read]
				#endif
			#endif	// ADD_L2_HEADER
	
			alu[meta_data_ptr_/**/i, --, b, 0]					; clear the meta data ptr
			
			// waiting for signal mask of phase i
	    	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, signal_mask_/**/i]

			ctx_arb[--], br[phase/**/next_i/**/#]  
			
			

		tbuf_full/**/i/**/#:	
			// read calendar seq num and update ok to schedule
			// if it will be still not ok to schedule, then wait in a loop
			#if (!IS_IXPTYPE(__IXP28XX))
			_sphy_mphy4_tx_sop_eop_wait_till_tbuf_not_full(tbuf_element, \
														tbuf_full_check_done_/**/i/**/#)
			#else
				update_ok_to_schedule(@ok_to_schedule_cnt, $mpkts_sent[0], @old_seq_read, \
								thd_timestamp, @msf_read_timestamp_old)
				alu[--, --, b, @ok_to_schedule_cnt]
				bgt[tbuf_full_check_done_/**/i/**/#]
				prevent_tbuf_overwrite(	@ok_to_schedule_cnt,
										addr_tx_seq,
										TBUF_MAX_VALID_ELEMENTS, @old_seq_read, tbuf_full_check_done_/**/i/**/#, \
										thd_timestamp, @msf_read_timestamp_current, @msf_read_timestamp_old)	
			#endif			

		not_sop_eop_/**/i/**/#:


			// check whether sop
			alu[--, sop_eop_/**/i, and, 1, <<1]
			bne[not_sop_/**/i/**/#]

			// sop only case	
			// grab a tbuf element
		
			alu[tbuf_element, --, b, *l$index1[GLOBAL_AVAIL_TBUF_ELEMENT_INDEX]]

			_sphy_mphy4_tx_update_next_availble_tbuf_ele()

			// now get this tbuf_ele address (addr_of_tbuf_1) 
			#ifdef SPI4_PREPEND 
				_sphy_mphy4_tx_get_tbuf_addr(addr_of_tbuf/**/i, tbuf_element)
			#else
				_sphy_mphy4_tx_get_tbuf_addr(addr_of_tbuf, tbuf_element)
			#endif

		sop_only_ready_to_transmit/**/i/**/#:
			// dram access takes longer latency, so do it first
			// move payload to tbuf [9]
			alu[offset_mod_8, payload_offset, and, 0x7]

			#ifdef SPI4_PREPEND
				_sphy_mphy4_tx_move_sop_paylo_to_tbuf(indir_ref, dram_addr, meta_data_ptr_/**/i, \
							addr_of_tbuf/**/i, indiref_base, payload_offset, @packet_payload, \
							offset_mod_8, sig_dram_to_tbuf_/**/i) 				
			#else
				_sphy_mphy4_tx_move_sop_paylo_to_tbuf(indir_ref, dram_addr, meta_data_ptr_/**/i, \
							addr_of_tbuf, indiref_base, payload_offset, @packet_payload, \
							offset_mod_8, sig_dram_to_tbuf_/**/i)
			#endif
				

			#ifndef ADD_L2_HEADER
			   	// setup tcw0_1 base for sop_only case
				alu_shf[tcw_/**/i, --, B, 1, <<TCW0_SOP_BIT_LOC]	; tcw0 base for sop only mpacket
				#ifdef SPI4_PREPEND
					immed_w1[tcw_/**/i, TCW0_SOP_EOP_BASE_W1]
				#endif

			#else // ADD_L2_HEADER
			#ifdef POS_TX
	    	// write  PPP header for SOP mpacket
				#ifdef DYNAMIC_PPP_ENCAP
					// MPLS encoding value is 0x8, i.e. bit 3 is set.
					// These 2 lines of code work with the assumption that ipv4 and ipv6
					// header types are 0x21 and 0x57 respectively. As such, both ipv4 and
					// ipv6 header types have bit 3 clear.
					br_bclr[packet_type, 3, not_mpls_type_sop#]
					alu[packet_type, --, B, mpls_pkt_type]

not_mpls_type_sop#:

					#ifndef ADD_PPP_ADDR_AND_CNTRL
						alu_shf[$msf_prepend_lw0_/**/i, --, b, packet_type, \
										<<IP_PROTOCOL_SHFT] ;Dynamic PPP Encap
					#else	//#ifndef ADD_PPP_ADDR_AND_CNTRL
						alu_shf[$msf_prepend_lw0_/**/i, ppp_header_base, or, \
							packet_type, <<IP_PROTOCOL_SHFT] ;Dynamic PPP Encap
					#endif	// #ifndef ADD_PPP_ADDR_AND_CNTRL
				#else	//  not DYNAMIC_PPP_ENCAP
					alu[$msf_prepend_lw0_/**/i, --, b, ppp_header_base]
				#endif	// #ifdef DYNAMIC_PPP_ENCAP

			    // write PPP header to tbuf
				msf[write, $msf_prepend_lw0_/**/i, addr_of_tbuf, 0, 1], sig_done[sig_prepend_write_/**/i]
			   	// setup tcw0_1 base for sop_only case
				alu_shf[tcw_/**/i, --, B, 1, <<TCW0_SOP_BIT_LOC]	; tcw0 base for sop only mpacket
				ld_field[tcw_/**/i, 1100, TCW0_SOP_EOP_BASE_W1, <<TCWO_W1_SHF]
				#endif	// #ifdef POS_TX

				#ifdef ETHERNET_TX	
			   	// setup tcw0_1 base for sop_only case
				alu_shf[tcw_/**/i, --, B, 1, <<TCW0_SOP_BIT_LOC]	; tcw0 base for sop only mpacket
				br!=byte[l2_header_passed, 0, 1,l2_header_in_paylo_/**/i#]	
				msf[write, $msf_prepend_lw0_/**/i, addr_of_tbuf, 0, 4], sig_done[sig_prepend_write_/**/i]
				ld_field[tcw_/**/i, 1100, TCW0_SOP_EOP_BASE_W1, <<TCWO_W1_SHF]
				br[sop_l2_header_write_done_/**/i#]
				l2_header_in_paylo_/**/i#:
				_sphy_mphy4_tx_clear_signal[signal_mask_/**/i, sig_prepend_write_/**/i]
				sop_l2_header_write_done_/**/i#:
				#endif	// #ifdef ETHERNET_TX
			#endif //#ifndef ADD_L2_HEADER
			// set payload_length, payload_offset, and channel fields in tcw0
			_sphy_mphy4_tx_set_tcw0_with_paylo_len_offset_chnl(tcw_/**/i, @packet_payload, \
														offset_mod_8, deq_port)

			#ifndef VALIDATE_TBUF_IN_THIRD_TX_ME		
			.begin
			.reg tmp
			alu_shf[tmp, --, B, tbuf_element, <<TX_CNTRL_ADDR_SHFT]
			alu[addr_of_tx_control_/**/i, tmp, +, *l$index1[GLOBAL_CTW_VD_BASE_INDEX]]
			.end
			#endif	// #ifndef VALIDATE_TBUF_IN_THIRD_TX_ME


			// Read 4 LWs from meta-data for SPI4 prepend, starting from LW3
			#ifdef SPI4_PREPEND
				dl_meta_load_cache(meta_data_ptr_/**/i, $$prepend_meta, sig_sram_meta_read, 3, 4)
			#endif	//ADD_L2_HEADER

			#ifdef VALIDATE_TBUF_IN_THIRD_TX_ME
			alu[flags_and_tbuf_element_/**/i, --, b, tbuf_element]	; no need to free buffer
			#endif	// #ifdef VALIDATE_TBUF_IN_THIRD_TX_ME

			// waiting for signal mask of phase i
    		local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, signal_mask_/**/i]

			ctx_arb[--], br[phase/**/next_i/**/#] 
		
		not_sop_/**/i/**/#:				
	
			// read 3 more lws first, then read another 3 lws  leftover
			// only whwn EFTOV_FROM_PB_FLAG_BIT in exe_stat_flag is set
			  
		wait_for_nn_to_become_non_empty_/**/i/**/#:
			br_inp_state[nn_empty, wait_for_nn_to_become_non_empty_/**/i/**/#]

			alu[$msf_prepend_lw0_/**/i, --, b, *n$index++]	; leftover lw0
			alu[exe_stat_flag, --, b, *n$index++]
			alu[tcw_/**/i, --, b, *n$index++]

			br_bset[exe_stat_flag, LEFTOV_FROM_PB_FLAG_BIT, \
										not_sop_read_three_more_lw_/**/i/**/#]

		not_sop_all_passover_read_done_/**/i/**/#:

			// grab a tbuf element
	
			alu[tbuf_element, --, b, *l$index1[GLOBAL_AVAIL_TBUF_ELEMENT_INDEX]]

			_sphy_mphy4_tx_update_next_availble_tbuf_ele()

			// now get this tbuf_ele address (addr_of_tbuf_1)
			#ifdef SPI4_PREPEND 
				_sphy_mphy4_tx_get_tbuf_addr(addr_of_tbuf/**/i, tbuf_element)
			#else
				_sphy_mphy4_tx_get_tbuf_addr(addr_of_tbuf, tbuf_element)
			#endif

			
		not_sop_ready_to_transmit/**/i/**/#:
				
				
			#ifdef ADD_L2_HEADER
				br_bclr[exe_stat_flag, LEFTOV_FROM_PB_FLAG_BIT, not_sop_no_prepend_write_/**/i/**/#]
			#else	// not ADD_L2_HEADER
				#ifdef SPI4_PREPEND
					br_bclr[exe_stat_flag, LEFTOV_FROM_PB_FLAG_BIT, not_sop_spi4_no_prepend_write_/**/i/**/#]
				#else
					br_bclr[exe_stat_flag, LEFTOV_FROM_PB_FLAG_BIT, not_sop_prepend_write_done_/**/i/**/#]
				#endif		
			#endif	// #ifdef ADD_L2_HEADER	

			alu_shf[leftover_bytes, 0x1F, AND, tcw_/**/i, >>TCWO_PREPEND_LEN_LSB_LOC]

			alu[--, leftover_bytes, -, 8]
			bgt[leftover_more_than_8_bytes_/**/i/**/#]

			#ifdef SPI4_PREPEND
				// reuse sig_sram_meta_read signal since this signal is not used for SPI4 header prepend
				msf[write, $msf_prepend_lw0, addr_of_tbuf/**/i, 0, 2], sig_done[sig_sram_meta_read]
				alu[addr_of_tbuf/**/i, addr_of_tbuf/**/i, +, 8]
			#else
				msf[write, $msf_prepend_lw0_/**/i, addr_of_tbuf, 0, 2], sig_done[sig_prepend_write_/**/i]
				alu[addr_of_tbuf, addr_of_tbuf, +, 8]
			#endif
			
			br[leftover_write_done_/**/i/**/#]

		leftover_more_than_8_bytes_/**/i/**/#:

			#ifdef SPI4_PREPEND
				// reuse sig_sram_meta_read signal since this signal is not used for SPI4 header prepend
				msf[write, $msf_prepend_lw0, addr_of_tbuf/**/i, 0, 4], sig_done[sig_sram_meta_read]
				alu[addr_of_tbuf/**/i, addr_of_tbuf/**/i, +, 16]
			#else
				msf[write, $msf_prepend_lw0_/**/i, addr_of_tbuf, 0, 4], sig_done[sig_prepend_write_/**/i]
				alu[addr_of_tbuf, addr_of_tbuf, +, 16]
			#endif

		leftover_write_done_/**/i/**/#:

			#ifndef ADD_L2_HEADER
				#ifndef SPI4_PREPEND
					_sphy_mphy4_tx_set_signal[signal_mask_/**/prev_i, sig_prepend_write_/**/i]
				#endif
			#endif	// #ifndef ADD_L2_HEADER

		not_sop_prepend_write_done_/**/i/**/#:

			// check whether need to move payload from active buffer to tbuf
			br_bset[exe_stat_flag, SKIP_TRANSMIT_FLAG_BIT, not_sop_no_dram2tbuf_/**/i/**/#]
			// payload offset and payload length can be retrived from tcw0_1
				
			alu[payload_offset, --, b, @packet_payload]

			#ifdef SPI4_PREPEND
				_sphy_mphy4_tx_move_not_sop_paylo_to_tbuf(indir_ref, dram_addr, meta_data_ptr_/**/i, \
					addr_of_tbuf/**/i, indiref_base, payload_offset, tcw_/**/i, sig_dram_to_tbuf_/**/i)
			#else
				_sphy_mphy4_tx_move_not_sop_paylo_to_tbuf(indir_ref, dram_addr, meta_data_ptr_/**/i, \
					addr_of_tbuf, indiref_base, payload_offset, tcw_/**/i, sig_dram_to_tbuf_/**/i)
			#endif

		not_sop_dram2tbuf_done_/**/i/**/#:
			
			// take care of End of buffer
			// copy EOB bit into EOP bit because we will free buffer when end of buffer
			// br_bclr[exe_stat_flag, EOB_MPKT_FLAG_BIT, not_sop_free_buffer_done#]
			alu[sop_eop_/**/i, 1, and~, exe_stat_flag, >>EOB_MPKT_FLAG_BIT]
			//alu[sop_eop_/**/i, --, b, exe_stat_flag, <<30]

			#ifdef SPI4_PREPEND
				// preserve SOP bit for next iteration
				alu[sop_eop_/**/i, sop_eop_/**/i, or, 2]
			#endif

			#ifndef VALIDATE_TBUF_IN_THIRD_TX_ME

			.begin
			.reg tmp
			alu_shf[tmp, --, B, tbuf_element, <<TX_CNTRL_ADDR_SHFT]		
			alu[addr_of_tx_control_/**/i, tmp, +, \
										*l$index1[GLOBAL_CTW_VD_BASE_INDEX]]
			.end

			#else	// VALIDATE_TBUF_IN_THIRD_TX_ME

			alu[flags_and_tbuf_element_/**/i, --, b, tbuf_element]
			br_bclr[exe_stat_flag, EOB_MPKT_FLAG_BIT, \
											free_buffer_bit_set_done_/**/i#]
			alu[flags_and_tbuf_element_/**/i, flags_and_tbuf_element_/**/i, \
								or, 1, <<FREE_BUFFER_FLAG_BIT]		; set flag
			free_buffer_bit_set_done_/**/i#:
			#endif	// #ifndef VALIDATE_TBUF_IN_THIRD_TX_ME
						
			// waiting for signal mask of phase i
			local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, signal_mask_/**/i]

			ctx_arb[--], br[phase/**/next_i/**/#]

			#if (!IS_IXPTYPE(__IXP28XX))
			sop_only_tbuf_full/**/i/**/#:
					_sphy_mphy4_tx_wait_till_tbuf_not_full(tbuf_element, \
															sop_only_ready_to_transmit/**/i/**/#)
		
			not_sop_tbuf_full/**/i/**/#:
					_sphy_mphy4_tx_wait_till_tbuf_not_full(tbuf_element, \
															not_sop_ready_to_transmit/**/i/**/#)
			sop_eop_tbuf_full/**/i/**/#:
					_sphy_mphy4_tx_wait_till_tbuf_not_full(tbuf_element, \
															sop_eop_ready_to_transmit/**/i/**/#)
			#endif

			#ifdef 	ADD_L2_HEADER											
			not_sop_no_prepend_write_/**/i/**/#:		
					_sphy_mphy4_tx_clear_signal[signal_mask_/**/prev_i, sig_prepend_write_/**/i]
					br[not_sop_prepend_write_done_/**/i/**/#]
			#else
				#ifdef SPI4_PREPEND
				not_sop_spi4_no_prepend_write_/**/i/**/#:
					_sphy_mphy4_tx_clear_signal[signal_mask_/**/i, sig_sram_meta_read]
					br[not_sop_prepend_write_done_/**/i/**/#]
				#endif
			#endif	// #ifdef 	ADD_L2_HEADER

			not_sop_no_dram2tbuf_/**/i/**/#:
					_sphy_mphy4_tx_clear_signal[signal_mask_/**/prev_i, sig_dram_to_tbuf_/**/i]
					_sphy_mphy4_tx_clear_signal[signal_mask_/**/prev_i, sig_dram_to_tbuf_/**/i + 1]
					br[not_sop_dram2tbuf_done_/**/i/**/#]


			not_sop_read_three_more_lw_/**/i/**/#:
				wait_for_nn_to_become_non_empty3_/**/i/**/#:
					br_inp_state[nn_empty, wait_for_nn_to_become_non_empty3_/**/i/**/#]

					alu[$msf_prepend_lw1_/**/i, --, b, *n$index++]	; leftover lw1
					alu[$msf_prepend_lw2_/**/i, --, b, *n$index++]	; leftover lw2
					alu[$msf_prepend_lw3_/**/i, --, b, *n$index++]	; leftover lw3

					br[not_sop_all_passover_read_done_/**/i/**/#]

			#ifndef VALIDATE_TBUF_IN_THIRD_TX_ME
			#ifdef CHECK_PORT_TX_FIFO_STATUS
		
				// when @fc_credit become 0 

				port_tx_fifo_full_phase_/**i**/#:
					_sphy_mphy4_tx_wait_port_tx_fifo_ready(@fc_credit,
						*$index, port_tx_fifo_check_done_phase_/**i**/#)

				#ifdef SPI4_PREPEND
					port_tx_fifo_full_phase_sop_eop_/**i**/#:
						_sphy_mphy4_tx_wait_port_tx_fifo_ready(@fc_credit,
							*$index, port_tx_fifo_check_done_phase_sop_eop_/**i**/#)

				#endif	// #ifdef SPI4_PREPEND
			#endif  // #ifdef CHECK_PORT_TX_FIFO_STATUS
			#endif	// #ifndef VALIDATE_TBUF_IN_THIRD_TX_ME

			//undefine_signals

	#define_eval i ( i + 1)
	#endloop

	


.end 
#endm // end of #macro sphy_mphy4_tx()

///////////////////////////////////////////////////////////////////////////////
// The main code starts here

main#:
		// initialize the microblock
	#if(TX_PHY_MODE == SPHY_1_32)
		sphy_mphy4_tx_init(port_entry_turnaround, SPHY_1_32)
	#endif // #if(TX_PHY_MODE == SPHY_1_32)

	// call the packet processing block
	sphy_mphy4_tx2()

/////////////////////////////////////////////////////////////////////////////////////

#endif 		// __SPHY_MPHY4_TX2_UC__

/////////////////////////////////////////////////////////////////////////////////////


      

