/*****************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 1998-2003 By Intel Corporation.  All rights reserved.  
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *****************************************************************************/

#include <assert.h>
#include <string.h>

#include "pdk_ns_helper.h"
#include "npf_common.h"
#include "fppapi.h"
#include "ds.h"
/******************************************************************************/
static int 		dsInitDone = 0;
static uint64_t 	dpeCount = 1;
PilCriticalSection 	dsLock;

static NPF_USERCONTEXT 	dsContext;
static FPPI_CBHANDLE 	dpeOperationHandle;

static DList   		UserList;
static DList   		DpeList;
static  DList  		RequestList;
/******************************************************************************/
#define DS_LOCK() 	\
			DS_TRACE("\nLOCK:%s(%d)\n", __FUNCTION__, __LINE__);\
			PIL_EnterCriticalSection( &dsLock )
#define DS_UNLOCK() 	\
			DS_TRACE("\nUNLOCK:%s(%d)\n", __FUNCTION__, __LINE__);\
			PIL_LeaveCriticalSection( &dsLock )

#define CHK_INIT() \
    if ( !dsInitDone ) {\
        return NPF_COMPONENT_UNINITIALIZED;\
    }

#define CHECK_DPE(info, handle)\
    if((info = (DS_DpeInfo_t *)npf_list_searchByKey(&DpeList, &handle)) == NULL)\
    {\
	DS_ERROR("NPF_DS_E_NO_SUCH_DPE(%d)", handle);\
	req->error = NPF_DS_E_NO_SUCH_DPE;\
	elem.dpeHandle = handle;\
	req->temp = &elem;\
        dsCallback(req);\
        DS_UNLOCK();\
	return(NPF_E_UNKNOWN);\
    }\
    DS_TRACE("DPE handle (%d):	Exists", handle);\
/******************************************************************************/
static NPF_boolean_t isCallbackReady(DiffServRequest_t *resp, NPF_DS_AsyncResponse_t *async);
NPF_error_t IfGetFEid(NPF_IfHandle_t ifHandle, uint32_t *no, FPPI_FEID *feId);
extern NPF_error_t IfGetPortAttribs(NPF_IfHandle_t , int *, NPF_IfType_t *, NPF_uint16_t *);
extern int cm_getDataByHandle(NPF_HANDLE handle, NPF_DATA *ppdata);
extern NPF_RET pdk_ns_getNSRootHandle (NPF_NSHDL *root_ns_hdl);
static DiffServCorrelator_t *newDsCorrelator(DiffServRequest_t *req);
NPF_DS_AsyncResponse_t *consolidateFEresponse(DiffServRequest_t *resp, int *count);

/********************************************************************************
* Function:	QOSRegister							*
* Description:	This is implementation of NPF_DS_Register. Check user already	*
*		Registered if not store user context in UserList. Otherwise	*
*		return an error NPF_E_CALLBACK_ALREADY_REGISTERED.		* 
*		If callback is NULL, return NPF_E_BAD_CALLBACK_FUNCTION		*
********************************************************************************/
NPF_error_t QOSRegister(NPF_userContext_t      context,
                    NPF_DS_CallbackFunc_t   callback,
                    NPF_callbackHandle_t   *cbhandle)
{
    NPF_callbackHandle_t tmpHandle;
    DSuserInfo_t           *user;
    CHK_INIT();
    DS_FUNCTION();
    DS_LOCK();
    user = (DSuserInfo_t *)PIL_AllocateMemory(sizeof(DSuserInfo_t),PIL_MEM_ZERO_INIT);
    if(user == NULL)
    {
        DS_UNLOCK();
        return NPF_E_UNKNOWN;
    }
    user->context = context;
    user->callback = callback;
    if(npf_list_searchByKey(&UserList, user) != NULL)
    {
        DS_UNLOCK();
        return NPF_E_CALLBACK_ALREADY_REGISTERED;
    }
    if (pdk_callback_register( NPF_CAT_DS, (NPF_USERCONTEXT)context,
            (NPF_GENERAL_CBFUNC)callback,
            (NPF_CBHANDLE *)&tmpHandle ) != NPF_SUCCESS)
    {
        PIL_FreeMemory(user);
        DS_UNLOCK();
        return NPF_E_BAD_CALLBACK_FUNCTION;
    }
    *cbhandle = tmpHandle;
    user->handle = tmpHandle;
    npf_list_pushBack(&UserList, user);
    DS_UNLOCK();
    return( NPF_NO_ERROR );
}

/********************************************************************************
* Function:	QOSDeregister							*
* Description:	This is implementation NPF_DS_Deregister.			*
*		Check user registered before, if not return 			*
*		NPF_E_BAD_CALLBACK_HANDLE. Remove user context stored during 	*
*		NPF_DS_register							*
********************************************************************************/
NPF_error_t QOSDeregister(NPF_callbackHandle_t   cbhandle)
{
    DSuserInfo_t           key;
    CHK_INIT();
    DS_FUNCTION();
    DS_LOCK();
    key.context = 0;
    key.callback = NULL;
    key.handle = cbhandle;
    if (pdk_callback_deregister( cbhandle ) != NPF_SUCCESS)
    {
        DS_UNLOCK();
        return( NPF_E_BAD_CALLBACK_HANDLE);
    }
    if ((npf_list_removeByKeyFree(&UserList, &key)) == -1)
    {
        DS_UNLOCK();
        return NPF_E_BAD_CALLBACK_HANDLE;
    }
    DS_UNLOCK();
    return( NPF_SUCCESS );
}

/*********************************************************************************
* Function:	DpeCreate							 *
* Description:	This is inplementation of NPF_DS_DpeCreate.			 *
*		Create new DS_DpeInfo_t object. Find attributes of interface with*
*		handle "ifHandle". Find the FE(s) associated with the given	 *
*		interface handle. Send FPPAPI request to all these FE(s)	 *
*		dpeCount is used as DPE handle, which is incremented for each    *
*		DPE created. DS_DpeInfo_t object created here will be deleted	 *
*		on NPF_DS_DpeDelete success. Save DS_DpeInfo_t object in DpeList *
*		When error encountered, generate internal async response through *
*		dsCallback function.
*********************************************************************************/
NPF_error_t DpeCreate(NPF_IfHandle_t ifHandle, NPF_DS_DpeDirection_t dir, NPF_DS_Dpe_t *dpe, DiffServRequest_t *req)
{
    DiffServCorrelator_t *corr;
    FPPI_DS_Dpe_t	elem;
    DS_DpeInfo_t	*info;
    NPF_IfType_t 	portType;
    NPF_uint16_t	mtu;

    DS_FUNCTION();
    DS_LOCK();
    if((corr = newDsCorrelator(req)) == NULL)
    {
	req->error = NPF_DS_E_RESOURCE;
	req->temp = &elem;
        dsCallback(req);
        DS_UNLOCK();
	return(NPF_E_UNKNOWN);
    }
    if(dpe == NULL)
    {
	req->error = NPF_DS_E_INVALID_DPE_PARAMETER;
	req->temp = &elem;
        dsCallback(req);
        DS_UNLOCK();
	return(NPF_E_UNKNOWN);
    }
    info = (DS_DpeInfo_t *)PIL_AllocateMemory(sizeof(DS_DpeInfo_t), PIL_MEM_ZERO_INIT);
    if(info == NULL)
    {
	req->error = NPF_DS_E_RESOURCE;
	req->temp = &elem;
        dsCallback(req);
        DS_UNLOCK();
	return(NPF_E_UNKNOWN);
    }
    if(IfGetPortAttribs(ifHandle, &elem.port, &portType, &mtu) != NPF_SUCCESS)
    {
	PIL_FreeMemory(info);
	req->error = NPF_DS_E_INVALID_INTERFACE;
	req->temp = &elem;
        dsCallback(req);
        DS_UNLOCK();
	return(NPF_E_UNKNOWN);
    }
    if(IfGetFEid(ifHandle, &info->fec, info->feid) != NPF_SUCCESS)
    {
	PIL_FreeMemory(info);
	req->error = NPF_DS_E_INVALID_INTERFACE;
	req->temp = &elem;
        dsCallback(req);
        DS_UNLOCK();
	return(NPF_E_UNKNOWN);
    }

    info->type = dpe->type;
    info->dpeHandle = dpeCount;
    info->dir = dir;
    info->port = elem.port;

    elem.dir = dir;
    elem.dpeHandle = dpeCount;
    elem.op = NPF_DS_DPE_CREATE;
    elem.prevDpeHandle = req->no; /* Used in FP, to handle DSCP classifier */

    memcpy(&elem.u.dpe, dpe, sizeof(NPF_DS_Dpe_t));
    DS_TRACE("dpeHandle: %d", elem.dpeHandle);
    DsFPPAPIcall(info, &elem, NPF_DS_DPE_CREATE, corr, req);
    dpeCount++;
    npf_list_pushBack(&DpeList, info); /* The memory allocated will be deleted in on DpeDelete Success */
    DS_UNLOCK();
    return(NPF_NO_ERROR);
}

/********************************************************************************
* Function:	DpeAddAssociation						*
* Description:	This is implementation of NPF_DS_DpeAddAssociation.		*
*		Check DPE handles in dpeSet are valid. If not generate internal *
*		callback. Send NPF_DS_DPE_ADD_ASSOC request to FE(s)		*
********************************************************************************/
NPF_error_t DpeAddAssociation(NPF_DS_DpeSetHandle_t *dpeSet, DiffServRequest_t *req)
{
    DiffServCorrelator_t *corr;
    FPPI_DS_Dpe_t	elem;
    DS_DpeInfo_t	*dpe;
    DS_DpeInfo_t	*prevDpe;
    DS_FUNCTION();
    DS_LOCK();
    if((corr = newDsCorrelator(req)) == NULL)
    {
	req->error = NPF_DS_E_RESOURCE;
	req->temp = &elem;
        dsCallback(req);
        DS_UNLOCK();
	return(NPF_E_UNKNOWN);
    }
    CHECK_DPE(dpe, dpeSet->dpeHandle);
    CHECK_DPE(prevDpe, dpeSet->prevDpeHandle);
    if(dpe->port != prevDpe->port)
    {
	req->error = NPF_DS_E_DPE_ADD_ASSOC_FAILED;
	req->temp = &elem;
        dsCallback(req);
        DS_UNLOCK();
	return(NPF_E_UNKNOWN);
    }
    DS_TRACE("DPE handle:	%d, prev DPE handle:	%d", dpeSet->dpeHandle, dpeSet->prevDpeHandle);
    elem.prevDpeHandle = dpeSet->prevDpeHandle;
    DsFPPAPIcall(dpe, &elem, NPF_DS_DPE_ADD_ASSOC, corr, req);
    DS_UNLOCK();
    return(NPF_NO_ERROR);
}

/********************************************************************************
* Function:	DpeDelAssociation						*
* Description:	This is implementation of NPF_DS_DpeDelAssociation.		*
*		Check DPE handles in dpeSet are valid. If not generate internal *
*		callback. Send NPF_DS_DPE_DELETE_ASSOC request to FE(s)		*
********************************************************************************/
NPF_error_t DpeDelAssociation(NPF_DS_DpeSetHandle_t *dpeSet, DiffServRequest_t *req)
{
    DiffServCorrelator_t *corr;
    FPPI_DS_Dpe_t	elem;
    DS_DpeInfo_t	*dpe;
    DS_DpeInfo_t	*prevDpe;
    DS_FUNCTION();
    DS_LOCK();
    if((corr = newDsCorrelator(req)) == NULL)
    {
	req->error = NPF_DS_E_RESOURCE;
	req->temp = &elem;
        dsCallback(req);
        DS_UNLOCK();
	return(NPF_E_UNKNOWN);
    }
    CHECK_DPE(dpe, dpeSet->dpeHandle);
    CHECK_DPE(prevDpe, dpeSet->prevDpeHandle);
    DS_TRACE("DPE handle:	%d, prev DPE handle:	%d", dpeSet->dpeHandle, dpeSet->prevDpeHandle);
    elem.prevDpeHandle = dpeSet->prevDpeHandle;
    DsFPPAPIcall(dpe, &elem, NPF_DS_DPE_DELETE_ASSOC, corr, req);
    DS_UNLOCK();
    return(NPF_NO_ERROR);
}

/********************************************************************************
* Function:	DpeDelete							*
* Description:	This is implementation of NPF_DS_DpeDelete.			*
*		Check DPE handle "dpeHandle" is valid. If not generate internal *
*		callback. Send NPF_DS_DPE_DELETE request to FE(s)		*
********************************************************************************/
NPF_error_t DpeDelete(NPF_DS_DpeHandle_t dpeHandle, DiffServRequest_t *req)
{
    DiffServCorrelator_t *corr;
    FPPI_DS_Dpe_t	elem;
    DS_DpeInfo_t	*info;
    DS_FUNCTION();
    DS_LOCK();
    if((corr = newDsCorrelator(req)) == NULL)
    {
	req->error = NPF_DS_E_RESOURCE;
	req->temp = &elem;
        dsCallback(req);
        DS_UNLOCK();
	return(NPF_E_UNKNOWN);
    }
    CHECK_DPE(info, dpeHandle);
    DS_TRACE("DPE handle:	%d", dpeCount);
    DsFPPAPIcall(info, &elem, NPF_DS_DPE_DELETE, corr, req);
    DS_UNLOCK();
    return(NPF_NO_ERROR);
}

/********************************************************************************
* Function:	DpeQuery							*
* Description:	This is implementation of NPF_DS_DpeQuery.			*
*		Check DPE handle "dpeHandle" is valid. If not generate internal *
*		callback. Send NPF_DS_DPE_QUERY request to FE(s)		*
********************************************************************************/
NPF_error_t DpeQuery(NPF_DS_DpeHandle_t dpeHandle, DiffServRequest_t *req)
{
    DiffServCorrelator_t *corr;
    FPPI_DS_Dpe_t	elem;
    DS_DpeInfo_t	*info;
    DS_FUNCTION();
    DS_LOCK();
    if((corr = newDsCorrelator(req)) == NULL)
    {
	req->error = NPF_DS_E_RESOURCE;
	req->temp = &elem;
        dsCallback(req);
        DS_UNLOCK();
	return(NPF_E_UNKNOWN);
    }
    CHECK_DPE(info, dpeHandle);
    DS_TRACE("DPE handle:	%d", dpeCount);
    DsFPPAPIcall(info, &elem, NPF_DS_DPE_QUERY, corr, req);
    DS_UNLOCK();
    return(NPF_NO_ERROR);
}

/********************************************************************************
* Function:	DpeModify							*
* Description:	This is implementation of NPF_DS_DpeModify.			*
*		Check DPE handle "dpeHandle" is valid. If not generate internal *
*		callback. Send NPF_DS_DPE_MODIFY request to FE(s)		*
********************************************************************************/
NPF_error_t DpeModify(NPF_DS_DpeHandle_t dpeHandle, NPF_DS_Dpe_t *dpe, DiffServRequest_t *req)
{
    DiffServCorrelator_t *corr;
    FPPI_DS_Dpe_t	elem;
    DS_DpeInfo_t	*info;
    DS_FUNCTION();
    DS_LOCK();
    if((corr = newDsCorrelator(req)) == NULL)
    {
	req->error = NPF_DS_E_RESOURCE;
	req->temp = &elem;
        dsCallback(req);
        DS_UNLOCK();
	return(NPF_E_UNKNOWN);
    }
    CHECK_DPE(info, dpeHandle);
    memcpy(&elem.u.dpe, dpe, sizeof(NPF_DS_Dpe_t));
    DsFPPAPIcall(info, &elem, NPF_DS_DPE_MODIFY, corr, req);
    DS_UNLOCK();
    return(NPF_NO_ERROR);
}

/********************************************************************************
* Function:	DpeStatistics							*
* Description:	This is implementation of NPF_DS_DpeStatistics.			*
*		Check DPE handle "dpeHandle" is valid. If not generate internal *
*		callback. Send NPF_DS_DPE_STATISTICS request to FE(s)		*
********************************************************************************/
NPF_error_t DpeStatistics(NPF_DS_DpeHandle_t dpeHandle, DiffServRequest_t *req)
{
    DiffServCorrelator_t *corr;
    FPPI_DS_Dpe_t	elem;
    DS_DpeInfo_t	*info;
    DS_FUNCTION();
    DS_LOCK();
    if((corr = newDsCorrelator(req)) == NULL)
    {
	req->error = NPF_DS_E_RESOURCE;
	req->temp = &elem;
        dsCallback(req);
        DS_UNLOCK();
	return(NPF_E_UNKNOWN);
    }
    CHECK_DPE(info, dpeHandle);
    DS_TRACE("DPE handle:	%d", dpeCount);
    DsFPPAPIcall(info, &elem, NPF_DS_DPE_STATISTICS, corr, req);
    DS_UNLOCK();
    return(NPF_NO_ERROR);
}

/********************************************************************************
* Function:	dsCallback							*
* Description:	This is helper function to construct Async response.		*
*		Each time this function is called, new Async response is created*
*		and stored in Async field of FE responded. If response is not 	*
*		from FE (Internal Error) Async response is considered for all	*
*		FE(s). FE response data is copied to Async response based on 	*
*		callback type. When all FE(s) responded, a single Async response*
*		is constructed by using individual async responses of one FE    *
*		and async responses of all other FE(s) are deleted. After	*
*		user callback invoked request object is removed from RequestList*
********************************************************************************/
void dsCallback(DiffServRequest_t *resp)
{
    int 		   count = 0;
    NPF_DS_AsyncResponse_t *async;
    FPPI_DS_Dpe_t	    *dpeInfo;
    NPF_boolean_t	    invokeCallback = NPF_FALSE;

    DS_FUNCTION();
    async = (NPF_DS_AsyncResponse_t *)PIL_AllocateMemory(sizeof(NPF_DS_AsyncResponse_t), 
							PIL_MEM_ZERO_INIT);
    if(async == NULL) 
    {
	DS_ERROR("async == NULL");
	return;
    }
    if(resp == NULL || resp->temp == NULL)
    {
	DS_ERROR("resp == NULL || resp->temp == NULL");
	return;
    }
    if(resp->error != NPF_NO_ERROR) resp->data.allOK = NPF_FALSE;
    dpeInfo = (FPPI_DS_Dpe_t *)resp->temp;
    async->dpeHandle = dpeInfo->dpeHandle;
    async->error = resp->error;
    switch(resp->type)
    {
	case NPF_DS_DPE_CREATE:
	{
	    async->dpeHandle = dpeInfo->dpeHandle;
	}
	break;
	case NPF_DS_DPE_ADD_ASSOC:
	{
	    async->dpeHandle = dpeInfo->dpeHandle;
	    async->prevDpeHandle = dpeInfo->prevDpeHandle;
	}
	break;
	case NPF_DS_DPE_MODIFY:
	{
	}
	break;
	case NPF_DS_DPE_DELETE:
	{
	    async->dpeHandle = dpeInfo->dpeHandle;
	    if(resp->error == NPF_NO_ERROR)
	    {
                npf_list_removeByKeyFree(&DpeList, &dpeInfo->dpeHandle);
	    }
	}
	break;
	case NPF_DS_DPE_QUERY:
	{
	    async->dpeHandle = dpeInfo->dpeHandle;
	    memcpy(&async->u.dpe, &dpeInfo->u.dpe, sizeof(NPF_DS_Dpe_t));
	}
	break;
	case NPF_DS_DPE_STATISTICS:
	{
	    async->dpeHandle = dpeInfo->dpeHandle;
	    memcpy(&async->u.statsResp, &dpeInfo->u.stats, sizeof(NPF_DS_StatsResp_t));
	}
	break;
	case NPF_DS_DPE_DELETE_ASSOC:
	{
	    async->dpeHandle = dpeInfo->dpeHandle;
	    async->prevDpeHandle = dpeInfo->prevDpeHandle;
	}
	break;
    }
    invokeCallback = isCallbackReady(resp, async);
    DS_TRACE("invokeCallback:	%d", invokeCallback);
    if(invokeCallback)
    {
	NPF_DS_AsyncResponse_t *asyncResp = NULL;
	uint32_t		i;
        asyncResp = consolidateFEresponse(resp, &count);
	if(asyncResp == NULL)
	{
	    DS_ERROR("asyncResp == NULL");
	    return;
	}
        resp->data.nResp = count;
        resp->data.resp = asyncResp;
	if(resp->report == NPF_REPORT_ALL || (resp->report == NPF_REPORT_ERRORS && resp->data.allOK != NPF_TRUE))
	{
	    DS_TRACE("Calling : cbthread_pushDSAPICB :: Func: %X, Corr: %X, Resp: %X", 
						resp->callback, resp->corr, resp->data.resp);
	    cbthread_pushDSAPICB(resp->callback, resp->context, resp->corr, &resp->data);
	    DS_TRACE("cbthread_pushDSAPICB :	Returned !!!!!");
	}
	else
	{
	    DS_TRACE("Freeing asyncResp : %X", asyncResp);
	    if(asyncResp != NULL)
                PIL_FreeMemory(asyncResp);
	}
	/* Free memory allocated for individual responses from all FEs */
	for(count = 0; count < resp->fec; count ++)
 	{
	    DS_TRACE("Freeing response of FE(%d)", resp->reqResp[count].id);
	    for(i = 0; i < resp->reqResp[count].resps; i ++)
	    {
	        if(resp->reqResp[count].async[i] != NULL)
	        {
	    	     DS_TRACE("Freeing response (%d) of FE(%d)", i, resp->reqResp[count].id);
                     PIL_FreeMemory(resp->reqResp[count].async[i]);
	  	}
	    }
	}	
	PIL_FreeMemory(resp->reqResp);
	DS_TRACE("Freeing Request State : %X", resp);
	PIL_FreeMemory(resp);
    }
}

/********************************************************************************
* Function:	newQosRequest							*
* Description:	This is helper function used to correlate user request and FE	*
*		response. New request object is created and all active FE(s) are*
*		determined. Sufficient memory is allocated to hold FE		*
*		response(s). All user information like callback function, 	*
*		correlator and error reporting are stored in request object.	*
********************************************************************************/
NPF_error_t newQosRequest(NPF_uint32_t no,
                        NPF_IfCallbackType_t type,
                        NPF_callbackHandle_t cbh,
                        NPF_correlator_t     corr,
                        NPF_errorReporting_t report,
                        DiffServRequest_t            **req)
{
    DiffServRequest_t           *request;
    NPF_NSHDL 		sys_ns_hdl, fe_ns_hdl;
    NPF_HANDLE 		sys_npf_hdl, fe_npf_hdl;
    NODE_ITERATOR 	fe_itr;
    PDK_FE_t* 		pdk_fe;
    uint32_t		fec;

    DS_FUNCTION();
    DS_LOCK();
    if(no < 1)
    {
        DS_UNLOCK();
        DS_ERROR("Requests ( %d ) > %d ***", no, DS_MAX_REQS);
        return(NPF_E_UNKNOWN);
    }
    if(no > DS_MAX_REQS)
    {
        DS_UNLOCK();
        DS_ERROR("Requests ( %d ) > %d ***", no, DS_MAX_REQS);
        return(NPF_E_UNKNOWN);
    }

    request = (DiffServRequest_t *)PIL_AllocateMemory(sizeof(DiffServRequest_t),PIL_MEM_ZERO_INIT);
    if(request == NULL)
    {
        DS_UNLOCK();
        DS_ERROR("request == NULL, returning NPF_E_UNKNOWN");
        return NPF_E_UNKNOWN;
    }

    if (pdk_callback_isCBExist(cbh) == NPF_INVALID_PARAMETERS)
    {
        PIL_FreeMemory(request);
        DS_UNLOCK();
        DS_ERROR("***** No Callback found *****");
        return (NPF_E_BAD_CALLBACK_HANDLE);
    }

    /* Need to traverse Namespace, get all FEs; for each FE that is still ON */
    if (NPF_SUCCESS != pdk_ns_getNSRootHandle (&sys_ns_hdl))
    {
        DS_UNLOCK();
        PIL_FreeMemory(request);
        return(NPF_E_UNKNOWN);
    }
    if (NPF_SUCCESS != npf_ns_getDataHandle(sys_ns_hdl, &sys_npf_hdl))
    {
        DS_UNLOCK();
        PIL_FreeMemory(request);
        npf_ns_close (sys_ns_hdl);
        return(NPF_E_UNKNOWN);
    }
    /* use NS helper fn. to get an iterator to all FEs in the
       system */
    if (NPF_SUCCESS != pdk_ns_getChildItrByType(sys_npf_hdl,NPF_FE,&fe_itr))
    {
        DS_UNLOCK();
        PIL_FreeMemory(request);
        npf_ns_close(sys_ns_hdl);
        return(NPF_E_UNKNOWN);
    }
    npf_ns_close (sys_ns_hdl);

    /* iterate through all FEs using this iterator */
    fec = npf_ns_count (fe_itr);	//Nof of FEs
    request->reqResp = (DS_ReqResp_t *)PIL_AllocateMemory(fec *sizeof(DS_ReqResp_t),
									PIL_MEM_ZERO_INIT);
    if(request->reqResp == NULL)
    {
        DS_UNLOCK();
        PIL_FreeMemory(request);
        DS_UNLOCK();
        return NPF_E_UNKNOWN;
    }
    fe_ns_hdl = npf_ns_first (fe_itr);
    do 
    {
        if (NPF_SUCCESS == npf_ns_getDataHandle (fe_ns_hdl, &fe_npf_hdl)) 
	{
            if (cm_getDataByHandle(fe_npf_hdl, (void **)&pdk_fe) == -1) 
	    {
                npf_ns_close (fe_ns_hdl);
                continue;
            }
            /* check if the handle is FE handle */
            if (((PDK_SHIM_t *)pdk_fe)->type != NPF_FE) 
	    {
                npf_ns_close (fe_ns_hdl);
                continue;
            }
            if (pdk_fe->attribs.status == NPF_FE_UP) 
	    {
		request->reqResp[request->fec].id = pdk_fe->id;
		request->reqResp[request->fec].reqs = no;
		request->fec = request->fec + 1;
            } 
        } 
        npf_ns_close (fe_ns_hdl);
    } while ((fe_ns_hdl = npf_ns_next (fe_itr)) != 0);
    npf_ns_done(fe_itr);

    request->callback = (NPF_DS_CallbackFunc_t) pdk_callback_func((NPF_CBHANDLE)cbh,
                                (NPF_USERCONTEXT *)&request->context);
    if(request->callback == NULL)
    {
        PIL_FreeMemory(request->reqResp);
        PIL_FreeMemory(request);
        DS_UNLOCK();
        DS_ERROR("***** No Callback found *****");
        return (NPF_E_BAD_CALLBACK_FUNCTION);
    }
    DS_INFO("no= %d, type= %d, handle=%d, corr= %d, report= %d,", no, type, cbh, corr, report);
    request->no = no;
    request->type = type;
    request->cbh = cbh;
    request->corr = corr;
    request->report = report;
    request->data.type = type;
    request->data.allOK = NPF_TRUE;

    /*npf_list_pushBack(&RequestList, request);*/
    DS_TRACE("Request State : %X", request);
    *req = request;
    DS_UNLOCK();
    return NPF_NO_ERROR;
}

/********************************************************************************
* Function:	checkRegister							*
* Description:	Compare function for UserList					*
*********************************************************************************/
static int checkRegister(const void* data, const void* key)
{
    DSuserInfo_t *user = (DSuserInfo_t *)data;
    DSuserInfo_t *ky = (DSuserInfo_t *)key;
    if(ky->context == 0 && ky->callback == NULL && ky->handle == user->handle)
        /* Deregister case */
        return 1;
    else if(user->context == ky->context && user->callback == ky->callback)
        return 1;
    else
        return 0;
}

/********************************************************************************
* Function:	checkDpe							*
* Description:	Compare function for DpeList					*
*********************************************************************************/
static int checkDpe(const void* data, const void* key)
{
    if(((DS_DpeInfo_t *)data)->dpeHandle == *((NPF_DS_DpeHandle_t *)key))
	return(1);
    return(0);
}

/********************************************************************************
* Function:	checkRequest							*
* Description:	Compare function for requestList				*
*********************************************************************************/
static int checkRequest (const void* data, const void* ky)
{
    DiffServRequest_t   *request;
    DiffServRequest_t   *key;
    DS_FUNCTION();
    request = (DiffServRequest_t *)data;
    key = (DiffServRequest_t *)ky;
    if(request->cbh == key->cbh)
    {
        TRACE("Request Found ....");
        return(NPF_TRUE);
    }
    return (NPF_FALSE);
}

/********************************************************************************
* Function:	pdk_ds_init							*
* Description:	DiffServ module initialization routine, called by pdk_init	*
*		UserList (used to store Registers), RequestList(used to store   *
*		user requests) and DpeList(used to store DPE information) are	*
*		initialized here. Registers with FPPI with callback function	*
*		"dpeOperationCallback". Mark DiffServ module is initialized	*
*********************************************************************************/
NPF_RET pdk_ds_init()
{
    if (dsInitDone == 1)
        return NPF_COMPONENT_REINITIALIZED;
    PIL_InitializeCriticalSection(&dsLock);
    DS_LOCK();
    npf_list_init(&UserList, PIL_FreeMemory);
    UserList.compare = &checkRegister;
    npf_list_init(&DpeList, PIL_FreeMemory);
    DpeList.compare = &checkDpe;
    npf_list_init(&RequestList, PIL_FreeMemory);
    RequestList.compare = &checkRequest;
    if (FPPAPI_register_callback(dpeOperationCallback,
        &dpeOperationHandle, FPPI_DS_DpeOperation,
        dsContext ) != FPPI_SUCCESS)
    {
	return (NPF_FAIL);
    }
    dsInitDone = 1;
    DS_UNLOCK();
    return(NPF_SUCCESS);
}

/********************************************************************************
* Function:	pdk_ds_shutdown							*
* Description:	DiffServ module shutdown routine, called by pdk_shutdown	*
*		Destroy all Lists. Deregisters with FPPI for callback function	*
*		"dpeOperationCallback". Mark DiffServ module is not initialized	*
*********************************************************************************/
NPF_RET pdk_ds_shutdown()
{
    DS_LOCK();
    npf_list_destroy(&UserList);
    npf_list_destroy(&DpeList);
    npf_list_destroy(&RequestList);
    FPPAPI_deregister_callback(dpeOperationHandle);
    dsInitDone = 0;
    DS_UNLOCK();
    PIL_DestroyCriticalSection(&dsLock);
    return(NPF_SUCCESS);
}

/********************************************************************************
* Function:	dpeOperationCallback						*
* Description:	Callback function for FPPI.					*
*		Update request object with FE response.				*
*		Mark request object such that response is from FE.		*
*		invoke "dsCallback"
********************************************************************************/
void dpeOperationCallback(
        FPPI_FEID feid,
        FPPI_CORRELATOR fpcorrelator,
        FPPI_CONTEXT fpcontext,
        void *response )
{
    FPPI_DS_Dpe_t      	  *dpeInfo;
    DiffServCorrelator_t  *corr;
    DiffServRequest_t     *resp;
    DS_FUNCTION();
    DS_TRACE("FE: %d, corr: %X", feid, fpcorrelator);
    DS_LOCK();
    corr = (DiffServCorrelator_t *)fpcorrelator;
    if(corr == NULL) return;
    resp = corr->req;
    resp->feResp = 1;
    resp->id = feid;
    dpeInfo = (FPPI_DS_Dpe_t *)response;
    resp->error = dpeInfo->error;
    resp->type = dpeInfo->op;
    resp->temp = dpeInfo;
    dsCallback(resp);
    PIL_FreeMemory(corr);
    DS_UNLOCK();
}

/********************************************************************************
* Function:	getDpeType							*
* Description:	Helper function to find DPE type through DPE handle		*
********************************************************************************/
NPF_DS_DpeType_t getDpeType(NPF_DS_DpeHandle_t dpeHandle)
{
    DS_DpeInfo_t		*info;
    if((info = npf_list_searchByKey(&DpeList, (void *)&dpeHandle)) == NULL) return(-1);
    return(info->type);
}

/********************************************************************************
* Function:	isCallbackReady							*
* Description:	Helper function to find user callback can be made.		*
*		For internal Async responses, update async filed of all FE(s)	*
*		If response is from FE, update async field of FE responded, and	*
*		increment response count for the FE. When all FE(s) reqest num  *
*		equals response count, return TRUE, elese return FALSE.		*
********************************************************************************/
static NPF_boolean_t isCallbackReady(DiffServRequest_t *resp, NPF_DS_AsyncResponse_t *async)
{
    int 	    fec;
    DS_ReqResp_t    *feResp;
    NPF_boolean_t   allFeResponded = NPF_FALSE;
    DS_FUNCTION();

    if(!resp->feResp)
    {
	/*
	 * Internally Generated Response
	 * Update responses for all FEs
	*/
	DS_TRACE("Internally Generated Response");
        for(fec = 0; fec < resp->fec; fec++)
        {
	    feResp = &resp->reqResp[fec];
	    if(!feResp) continue;
	    feResp->async[feResp->resps++] = async;
	}
    }
    for(fec = 0; fec < resp->fec; fec++)
    {
	feResp = &resp->reqResp[fec];
	if(!feResp) continue;
	if(feResp->used != 1) /* Request not sent to this FE */
	{
	    DS_TRACE("Request not sent to this FE(%d)", feResp->id);
	}
	if(resp->id == feResp->id)
	{
	    DS_TRACE("Updating response for FE(%d)", feResp->id);
	    feResp->error = async->error;
	    feResp->async[feResp->resps++] = async;
	}
	if(feResp->reqs == feResp->resps) 
	{
	    DS_TRACE("FE(%d) Done !!!!!!!!", feResp->id);
	    feResp->done = NPF_TRUE;
	}
    }
    /*
     * Check All FE(s) responded
    */
    for(fec = 0; fec < resp->fec; fec++)
    {
	feResp = &resp->reqResp[fec];
	if(!feResp) continue;
	if(!feResp->done)
	{
	    DS_TRACE("FE(%d) pending .......", feResp->id);
	}
	allFeResponded = feResp->done;
    }
    return(allFeResponded);
}

/********************************************************************************
* Function:	newDsCorrelator							*
* Description:	Helper function, which returns new correlator			*
*		request object is stored in correlator.				*
********************************************************************************/
static DiffServCorrelator_t *newDsCorrelator(DiffServRequest_t *req)
{
    DiffServCorrelator_t	*tmp = NULL;
    tmp = (DiffServCorrelator_t *)PIL_AllocateMemory(sizeof(DiffServCorrelator_t), PIL_MEM_ZERO_INIT);
    if(tmp != NULL)
    {
	tmp->req = req;
	return tmp;
    }
    return(tmp);
}

/********************************************************************************
* Function:	DsFPPAPIcall							*
* Description:	Helper function, which calls FPPAPI				*
********************************************************************************/
void DsFPPAPIcall(DS_DpeInfo_t *info,
                  FPPI_DS_Dpe_t *elem,
		  NPF_DS_CallbackType_t op,
                  void          *corr,
                  DiffServRequest_t *req)
{
    int         i;
    int         j;
    uint32_t    fec;

    DS_FUNCTION();
    elem->type = info->type;
    elem->op = op;
    elem->dpeHandle = info->dpeHandle;
    DS_TRACE("DpeHandle: %d, DpeType: %d, DpeOp: %d", elem->dpeHandle, elem->type, elem->op);
    /*
     *Following two loops used to mark FE(s) in request to send FPPI request
    */
    for(i = 0; i < info->fec; i++) /* Loop through no of FEs associated with the interface */
    {
        for(j = 0; j < req->fec; j++) /* Loop through no of active FEs in the system */
        {
            if(info->feid[i] == req->reqResp[j].id)
            {
                req->reqResp[j].used = 1;
            }
        }
    }
    for(fec = 0; fec < req->fec; fec++)
    {
        if(req->reqResp[fec].used == 1)
        {
    	    DS_TRACE("FE: %d, corr: %X", req->reqResp[fec].id, corr);
            if(FPPAPI_QOS_DpeOperation(req->reqResp[fec].id,
                                    (FPPI_CORRELATOR)corr,
                                    dpeOperationHandle,
                                    elem) != FPPI_SUCCESS)
            {
                req->error = NPF_DS_E_UNKNOWN;
                dsCallback(req);
            }
        }
    }
}

/********************************************************************************
* Function:	consolidateFEresponse						*
* Description:	Helper function, which consolidates the FE(s) responses		*
*		If All FE(s) responded success or failure consider response(s)  *
*		first FE to invoke user callback. When some FE responded success*
*		and some failed, take response of first FE responded success.   *
*		Generate internal event to indicate failed FE(s)		*
* TO DO:	Generate internal Event to indicate failed FE(s)	
********************************************************************************/
NPF_DS_AsyncResponse_t *consolidateFEresponse(DiffServRequest_t *resp, int *count)
{
    NPF_DS_AsyncResponse_t *asyncResp = NULL;
    DS_ReqResp_t	   *feResp = NULL;
    uint32_t		    pfec = 0;
    DS_ReqResp_t	   *passFE[MAXFENUM];
    uint32_t		    ffec = 0;
    DS_ReqResp_t	   *failFE[MAXFENUM];
    int 		    i;
    DS_FUNCTION();

    /*
     * Go through all FE responses
     * If All FE respnse is same (Success/Failure) take response first FE
     * If Some passed and some failed, take response of first Success FE
     * Generate Failure Event(s) for other FE(s)
    */
    for(i = 0; i < resp->fec; i++)
    {
	feResp = &resp->reqResp[i];
	if(feResp == NULL) continue;
	//if(!feResp->used) continue;
	if(!feResp->error)
	{
	    DS_TRACE("FE(%d) responded SUCCESS..........", feResp->id);
	    passFE[pfec++] = feResp;
	}
	else
	{
	    DS_TRACE("FE(%d) responded XXXXXX FAILED XXXX", feResp->id);
	    failFE[ffec++] = feResp;
	}
    }
    if(pfec == resp->fec)
    {
	DS_TRACE("All FE responded Success ^^^^^^^^^^^^^^^");
	feResp = passFE[0];
    }
    else if(ffec == resp->fec)
    {
	DS_TRACE("All FE responded Failure VVVVVVVVV");
	feResp = failFE[0];
    }
    else
    {
	DS_TRACE("%d FE respondes Success, %d FE responded Failure", pfec, ffec);
	/*
	 * Send Async response of passed FE
	 * raise Event for all Failed FE(s)
	 */
	feResp = passFE[0];
    }

    asyncResp = (NPF_DS_AsyncResponse_t *)PIL_AllocateMemory(
		feResp->resps*sizeof(NPF_DS_AsyncResponse_t), 
		PIL_MEM_ZERO_INIT);
    if(asyncResp == NULL)
    {
	DS_ERROR("asyncResp == NULL");
	return(NULL);
    }

    for(i = 0; i < feResp->resps; i ++)
    {
        if(feResp->async[i] != NULL)
        {
             memcpy(&asyncResp[i], feResp->async[i], sizeof(NPF_DS_AsyncResponse_t));
        }
    }	
    DS_TRACE("asyncResp(%d) : %X", feResp->resps, asyncResp);
    *count = feResp->resps;
    return(asyncResp);
}
