/*****************************************************************************
 *                  INTEL CORPORATION PROPRIETARY INFORMATION
 *  This software is supplied under the terms of a license agreement or
 *  nondisclosure agreement with Intel Corporation and may not be copied
 *  or disclosed except in accordance with the terms of that agreement.
 *  Copyright (c) 2001 Intel Corporation. All Rights Reserved.
 *
 *$Source: 
 *$Revision: 
 *$Date: 
 * Description:
 *     This file contains the DiffServ API implementation.
 *
 *****************************************************************************/

#include <assert.h>
#include <string.h>

#include "list.h"
#include "npf_common.h"
#include "pil.h"
#include "log.h"
#include "fpp_common.h"
#include "pdk_ns_helper.h"
#include "npf_ipv4_types.h"
#include "npf_ds.h"
#include "ds.h"

#define REQUEST(no, type, cbh, corr, report, request) \
    if((ret = newQosRequest(no, type, cbh, corr, report, &request)) != NPF_NO_ERROR)\
    {\
        DS_ERROR("%s", "newRequest != NPF_NO_ERROR");\
        return(ret);\
    }\


/*****************************************************************************
* Name: NPF_DS_Register
* Description:
*               This function is used by an application to register its completion
*               callback function for receiving asynchronous responses related to
*               API function calls. The application may register multiple callback
*               functions using this function. The callback function is identified
*               by the pair of "context" and "callback", and for each individual
*               pair, a unique callback will be assigned for future reference.
*               Since the callback function is identified by both "context" and
*               "callback", duplicate registration of the same callback function
*               with different "context" is allowed. Also, the same "context" can
*               be shared among different callback functions. Duplicate registration of
*               the same "context" and "callback" pair has no effect, will output
*               a handle that is already assigned to the pair, and will return
*               NPF_IF_E_ALREADY_REGISTERED.
*               Note: NPF_DS_Register() is a synchronous function and has no completion
*               callback associated with it.
* Input Parameters:
*       context: A context item for uniquely identifying the context of the application
*                       registering the completion callback function. The exact value will
*                       be provided back to the registered completion callback function as
*                       its first parameter when it is called.  Application can assign any
*                       value to the "context" and the value is completely opaque to the
*                       API implementation.
*       callback: Pointer to the completion callback function to be registered.
* Output Parameters:
*       cbhandle: A unique identifier assigned for the registered "context" and
*       "callback" pair. This handle will be used by the application to specify which
*                       callback to be called when invoking asynchronous API functions.
*                       It will also be used when de-registering the "context" and
*                       "callback" pair.
* Return Codes:
*               NPF_NO_ERROR: The registration completed successfully.
*               NPF_E_BAD_CALLBACK_DS_FUNCTION: "callback" is NULL.
*               NPF_E_CALLBACK_ALREADY_REGISTERED: No new registration was made since the
*                       "context" and "callback" pair was already registered.
*******************************************************************************/

NPF_error_t NPF_DS_Register (
            NPF_IN NPF_userContext_t      context,
            NPF_IN NPF_DS_CallbackFunc_t  callback,
            NPF_OUT NPF_callbackHandle_t   *cbhandle)
{
    return(QOSRegister(context, callback, cbhandle));
}

/*****************************************************************************
* Name: NPF_DS_Deregister
* Description: This function is used by an application to de-register a pair
*               of user context and callback function. After the Deregister function
*               returns, no more function calls can be made using the deregistered
*               callback handle.
* Input Parameters:
*               cbhandle: The unique identifier representing the pair of
*               user context and callback function to be de-registered.
* Output Parameters:
*               None.
* Return Codes:
*               NPF_NO_ERROR: The de-registration completed successfully.
*               NPF_BAD_CALLBACK_HANDLE: The API implementation does not
*                                               recognize the callback handle.
*                                               There is no effect to the registered
*                                               callback functions.
*******************************************************************************/
NPF_error_t NPF_DS_Deregister(
    NPF_IN NPF_callbackHandle_t cbhandle)
{
    return(QOSDeregister(cbhandle));
}

/*****************************************************************************
* Name: NPF_DS_NPF_DS_DpeCreate
* Description: 
* Input Parameters:
*               
*               
* Output Parameters:
*               None.
* Return Codes:
*******************************************************************************/

NPF_error_t NPF_DS_DpeCreate(
            NPF_IN    NPF_callbackHandle_t      cbhandle,
            NPF_IN    NPF_correlator_t          correlator,
            NPF_IN    NPF_errorReporting_t      errorReporting,
            NPF_IN    NPF_IfHandle_t            interface,
            NPF_IN    NPF_DS_DpeDirection_t     dir,
            NPF_IN    NPF_uint32_t              no,
            NPF_IN    NPF_DS_Dpe_t             *dpe)
{
    uint32_t		count;
    DiffServRequest_t           *request = NULL;
    NPF_error_t         ret = 0;
    REQUEST(no, NPF_DS_DPE_CREATE, cbhandle, correlator, errorReporting, request)
    for(count = 0; count < no; count ++)
    {
        DpeCreate(interface, dir, &dpe[count], request);
    }
    return(NPF_NO_ERROR);
}

/*****************************************************************************
* Name: NPF_DS_DpeAddAssociation
* Description: 
* Input Parameters:
*               
*               
* Output Parameters:
*               None.
* Return Codes:
*******************************************************************************/

NPF_error_t NPF_DS_DpeAddAssociation(
       NPF_IN    NPF_callbackHandle_t      cbhandle,
       NPF_IN    NPF_correlator_t          correlator,
       NPF_IN    NPF_errorReporting_t      errorReporting,
       NPF_IN    NPF_uint32_t              noOfEntries,
       NPF_IN    NPF_DS_DpeSetHandle_t     *dpeSetHandle)
{
    uint32_t		count;
    DiffServRequest_t           *request = NULL;
    NPF_error_t         ret = 0;
    REQUEST(noOfEntries, NPF_DS_DPE_ADD_ASSOC, cbhandle, correlator, errorReporting, request)
    for(count = 0; count < noOfEntries; count ++)
    {
        DpeAddAssociation(&dpeSetHandle[count], request);
    }
    return(NPF_NO_ERROR);
}

/*****************************************************************************
* Name: NPF_DS_DpeDelAssociation
* Description: 
* Input Parameters:
*               
*               
* Output Parameters:
*               None.
* Return Codes:
*******************************************************************************/

NPF_error_t NPF_DS_DpeDelAssociation(
       NPF_IN    NPF_callbackHandle_t      cbhandle,
       NPF_IN    NPF_correlator_t          correlator,
       NPF_IN    NPF_errorReporting_t      errorReporting,
       NPF_IN    NPF_uint32_t              noOfEntries,
       NPF_IN    NPF_DS_DpeSetHandle_t     *dpeSetHandle)
{
    uint32_t		count;
    DiffServRequest_t           *request = NULL;
    NPF_error_t         ret = 0;

    REQUEST(noOfEntries, NPF_DS_DPE_DELETE_ASSOC, cbhandle, correlator, errorReporting, request)
    for(count = 0; count < noOfEntries; count ++)
    {
        DpeDelAssociation(&dpeSetHandle[count], request);
    }
    return(NPF_NO_ERROR);
}

/*****************************************************************************
* Name: NPF_DS_NPF_DS_DpeModify
* Description: 
* Input Parameters:
*               
*               
* Output Parameters:
*               None.
* Return Codes:
*******************************************************************************/

NPF_error_t NPF_DS_DpeModify(
       NPF_IN    NPF_callbackHandle_t      cbhandle,
       NPF_IN    NPF_correlator_t          correlator,
       NPF_IN    NPF_errorReporting_t      errorReporting,
       NPF_IN    NPF_uint32_t              noOfEntries,
       NPF_IN    NPF_DS_DpeHandle_t        *dpeHandle,
       NPF_IN    NPF_DS_Dpe_t              *dpe)
{
    uint32_t		count;
    DiffServRequest_t           *request = NULL;
    NPF_error_t         ret = 0;

    REQUEST(noOfEntries, NPF_DS_DPE_MODIFY, cbhandle, correlator, errorReporting, request)
    for(count = 0; count < noOfEntries; count ++)
    {
        DpeModify(dpeHandle[count], dpe, request);
    }
    return(NPF_NO_ERROR);
}

/*****************************************************************************
* Name: NPF_DS_DpeQueryAttributes
* Description: 
* Input Parameters:
*               
*               
* Output Parameters:
*               None.
* Return Codes:
*******************************************************************************/
NPF_error_t NPF_DS_DpeQueryAttributes(
       NPF_IN    NPF_callbackHandle_t      cbhandle,
       NPF_IN    NPF_correlator_t          correlator,
       NPF_IN    NPF_errorReporting_t      errorReporting,
       NPF_IN    NPF_uint32_t              nHandles,
       NPF_IN    NPF_DS_DpeHandle_t        *dpeHandle)
{
    uint32_t		count;
    DiffServRequest_t           *request = NULL;
    NPF_error_t         ret = 0;

    REQUEST(nHandles, NPF_DS_DPE_QUERY, cbhandle, correlator, errorReporting, request)
    for(count = 0; count < nHandles; count ++)
    {
        DpeQuery(dpeHandle[count], request);
    }
    return(NPF_NO_ERROR);
}

/*****************************************************************************
* Name: NPF_DS_NPF_DS_DpeStats
* Description: 
* Input Parameters:
*               
*               
* Output Parameters:
*               None.
* Return Codes:
*******************************************************************************/
NPF_error_t NPF_DS_DpeQueryStats(
       NPF_IN    NPF_callbackHandle_t      cbhandle,
       NPF_IN    NPF_correlator_t          correlator,
       NPF_IN    NPF_errorReporting_t      errorReporting,
       NPF_IN    NPF_uint32_t              nHandles,
       NPF_IN    NPF_DS_DpeHandle_t        *dpeHandle)
{
    uint32_t		count;
    DiffServRequest_t           *request = NULL;
    NPF_error_t         ret = 0;

    REQUEST(nHandles, NPF_DS_DPE_STATISTICS, cbhandle, correlator, errorReporting, request)
    for(count = 0; count < nHandles; count ++)
    {
        DpeStatistics(dpeHandle[count], request);
    }
    return(NPF_NO_ERROR);
}

/*****************************************************************************
* Name: NPF_DS_NPF_DS_DpeDelete
* Description: 
* Input Parameters:
*               
*               
* Output Parameters:
*               None.
* Return Codes:
*******************************************************************************/
NPF_error_t NPF_DS_DpeDelete(
       NPF_IN    NPF_callbackHandle_t      cbhandle,
       NPF_IN    NPF_correlator_t          correlator,
       NPF_IN    NPF_errorReporting_t      errorReporting,
       NPF_IN    NPF_uint32_t              nHandles,
       NPF_IN    NPF_DS_DpeHandle_t        *dpeHandle)
{
    uint32_t		count;
    DiffServRequest_t           *request = NULL;
    NPF_error_t         ret = 0;

    REQUEST(nHandles, NPF_DS_DPE_DELETE, cbhandle, correlator, errorReporting, request)
    for(count = 0; count < nHandles; count ++)
    {
        DpeDelete(dpeHandle[count], request);
    }
    return(NPF_NO_ERROR);
}
