/*****************************************************************************
 *                  INTEL CORPORATION PROPRIETARY INFORMATION
 *  This software is supplied under the terms of a license agreement or
 *  nondisclosure agreement with Intel Corporation and may not be copied
 *  or disclosed except in accordance with the terms of that agreement.
 *  Copyright (c) 2001 Intel Corporation. All Rights Reserved.
 *
 *$Source: /cvs/constellations/ixa_3.0/src/cp_pdk/ControlPlane/APIImplementations/IM/src/npf_if_api.c,v $
 *$Revision: 1.6 $
 *$Date: 2003/11/17 16:57:10 $
 * Description:
 *     This file contains the interface management NPF api implementation.
 *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "npf_common.h"
#include "npf_ipv4_types.h"
#include "npf_cm_types.h"
#include "pdk_callback.h"
#include "pdk_cm.h"
#include "pdk_macros.h"
#include "pdk_ns_helper.h"
#include "pdkmgr.h"
#include "pil.h"
#include "fpp_common.h"
#include "fppapi.h"
#include "list.h"
#include "npf_atm_types.h"
#include "if.h"

#define REQUEST(n_if, type, cbh, corr, report, request) \
    if((ret = newRequest(n_if, type, cbh, corr, report, &request)) != NPF_NO_ERROR)\
    {\
        IF_ERROR("%s", "newRequest != NPF_NO_ERROR");\
        return(ret);\
    }\

/*****************************************************************************
*    Name: NPF_IfRegister
*    Description:
*		This function is used by an application to register its completion 
*		callback function for receiving asynchronous responses related to 
*		API function calls. The application may register multiple callback
*		functions using this function. The callback function is identified 
*		by the pair of userContext and ifCallbackFunc, and for each individual 
*		pair, a unique ifCallbackHandle will be assigned for future reference. 
*		Since the callback function is identified by both userContext and 
*		ifCallbackFunc, duplicate registration of the same callback function 
*		with different userContext is allowed. Also, the same userContext can 
*		be shared among different callback functions. Duplicate registration of 
*		the same userContext and ifCallbackFunc pair has no effect, will output 
*		a handle that is already assigned to the pair, and will return 
*		NPF_IF_E_ALREADY_REGISTERED.
*		Note: NPF_IfRegister() is a synchronous function and has no completion 
*		callback associated with it.
* Input Parameters:
*	userContext: A context item for uniquely identifying the context of the application
*			registering the completion callback function. The exact value will 
*			be provided back to the registered completion callback function as 
*			its first parameter when it is called.  Application can assign any 
*			value to the userContext and the value is completely opaque to the 
*			API implementation.
*	ifCallbackFunc: Pointer to the completion callback function to be registered.
* Output Parameters:
*	ifCallbackHandle: A unique identifier assigned for the registered userContext and
*	ifCallbackFunc pair. This handle will be used by the application to specify which 
*			callback to be called when invoking asynchronous API functions. 
*			It will also be used when de-registering the userContext and 
*			ifCallbackFunc pair.
* Return Codes:
*		NPF_NO_ERROR: The registration completed successfully.
*		NPF_IF_E_BAD_CALLBACK_FUNCTION: ifCallbackFunc is NULL.
*		NPF_IF_E_ALREADY_REGISTERED: No new registration was made since the
*			userContext and ifCallbackFunc pair was already registered.
*******************************************************************************/
NPF_error_t NPF_IfRegister (
            NPF_IN NPF_userContext_t      context,
            NPF_IN NPF_IfCallbackFunc_t   callback,
            NPF_OUT NPF_callbackHandle_t   *cbhandle)
{
    IF_FUNCTION();
    return(IfRegister(context, callback, cbhandle));
}


/*****************************************************************************
*    Name: NPF_IfDeregister
*    Description: This function is used by an application to de-register a pair 
*		of user context and callback function. After the Deregister function 
*		returns, no more function calls can be made using the deregistered 
*		callback handle.
* Input Parameters:
*		ifCallbackHandle: The unique identifier representing the pair of 
*		user context and callback function to be de-registered.
* Output Parameters:
*		None.
* Return Codes:
*		NPF_NO_ERROR: The de-registration completed successfully.
*		NPF_IF_E_BAD_CALLBACK_HANDLE: The API implementation does not 
*						recognize the callback handle. 
*						There is no effect to the registered 
*						callback functions.
*******************************************************************************/
NPF_error_t NPF_IfDeregister(
    NPF_IN NPF_callbackHandle_t cbhandle)
{
    IF_FUNCTION();
    return(IfDeregister(cbhandle));
}

/*****************************************************************************
*    Name: NPF_IfEventRegister
*    Description: This function is used by an application to register its event 
*		handler function for receiving asynchronous event notifications 
*		from this API. Application may register multiple handler functions 
*		using this function. The event handler function is identified by the 
*		pair of userContext and ifEventHandlerFunc, and for each individual 
*		pair, a unique ifEventHandlerHandle will be assigned for future reference. 
*		Since the event handler function is identified by both userContext and
*		ifEventHandlerFunc, duplicate registration of same event handler function 
*		with different userContext is allowed. Also, same userContext can be shared 
*		among different event handler functions. Duplicate registration of the same 
*		userContext and ifEventHandlerFunc pair has no effect, and will output a handle 
*		that is already assigned to the pair, and will return NPF_IF_E_ALREADY_REGISTERED.
*		Notes: Besides registering a handler function, this call enables events. 
*		The handler function could be called at any time following the invocation 
*		of IfEventRegister(). NPF_IfEventRegister() is a synchronous function and 
*		has no completion callback associated with it.
* Input Parameters:
*		userContext: A context item for uniquely identifying the context of the 
*				application registering the event handler function. 
*				The exact value will be provided back to the registered 
*				event handler function as its 1 st parameter when it is called. 
*				Application can assign any value to the userContext and the value 
*				is completely opaque to the API implementation.
*		ifEventHandlerFunc: Pointer to the event handler function to be registered.
* Output Parameters:
*		ifEventHandlerHandle: A unique identifier assigned for the registered
*			userContext and ifEventHandlerFunc pair. This handle will be used by the
*			application de-registering the userContext and ifEventHandlerFunc pair.
* Return Codes:
*		NPF_NO_ERROR: The registration completed successfully.
*		NPF_IF_E_BAD_CALLBACK_HANDLE: ifEventHandlerFunc is NULL or not recognized.
*		NPF_IF_E_ALREADY_REGISTERED: No new registration was made since the
*				userContext and ifEventHandlerFunc pair was already registered.
*******************************************************************************/
NPF_error_t NPF_IfEventRegister(
                NPF_userContext_t 	 	userContext,
                NPF_IfEventHandlerFunc_t 	ifEventHandlerFunc,
                NPF_IfEventHandlerHandle_t 	*ifEventHandlerHandle)
{
    IF_FUNCTION();
    return (IfEventRegister(userContext, ifEventHandlerFunc, ifEventHandlerHandle));
}

/*****************************************************************************
*    Name: NPF_IfEventDeregister
*    Description: This function is used by an application to de-register a pair 
*		of user context and event handler function.
* Input Parameters:
*		ifEventHandlerHandle: The unique identifier representing the pair 
*				of user context and event handler function to be de-registered.
* Output Parameters:
*		None.
* Return Codes:
*		NPF_NO_ERROR: The de-registration completed successfully.
*		NPF_E_BAD_CALLBACK_HANDLE: The API implementation does not recognize the
*			event handler handle. There is no effect to the registered event 
*			handler functions.
*******************************************************************************/
NPF_error_t NPF_IfEventDeregister(NPF_IfEventHandlerHandle_t cbhandle)
{
    IF_FUNCTION();
    return (IfEventDeregister(cbhandle));
}

/*****************************************************************************
* Name: NPF_IfCreate
*
* Description:
* 				Function to Create one or more Interfaces
*				This function creates one or more interfaces of a 
*				given type, including "typeless" (type unknown). 
*				Interfaces created by this function are in the 
*				Administratively Disabled
* Input Parameters:
*	if_cbHandle - 		the registered callback handle.
*	if_cbCorrelator -	the application's context for this call.
*	if_errorReporting - 	the desired callback.
*	n_if - 			number of interfaces to create.
*	if_Type -		the interface type: NPF_IF_TYPE_LAN, NPF_IF_TYPE_IPv4,
*				NPF_IF_TYPE_ATM, NPF_IF_TYPE_POS, or NPF_IF_TYPE_UNK. 
*				All interfaces created 	by one function invocation are 
*				of the same type.
* Output Parameters:
*				None
* Asynchronous Error Codes:
*	 			NPF_NO_ERROR : operation successful.
*				NPF_IF_E_INVALID_PARAM: operation failed, interface not created.
* Asynchronous response:
*				A total of n_if asynchronous responses (NPF_IfAsyncResponse_t) 
*				will be passed to the callback 	function, in one or more 
*				invocations. Each response contains the new interface handle or 
*				a possible error code. The union in the callback response 
*				structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfCreate(
                NPF_IN NPF_callbackHandle_t if_cbHandle,
                NPF_IN NPF_correlator_t if_cbCorrelator,
                NPF_IN NPF_errorReporting_t if_errorReporting,
                NPF_IN NPF_uint32_t n_if,
                NPF_IN NPF_IfType_t if_Type)
{
    request_t		*request = NULL;
    NPF_uint32_t	count = 0;
    NPF_error_t		ret = 0;

    IF_FUNCTION();
    REQUEST(n_if, NPF_IF_CREATE, if_cbHandle, 
		if_cbCorrelator, if_errorReporting, request)
    for(count = 0; count < n_if; count++)
    {
	IfCreate(if_Type, request);
    }
    return(NPF_NO_ERROR);
}

/*****************************************************************************
* Name: NPF_IfDelete
* Description:	Function to Delete one or more Interfaces
* Input Parameters:
* 		if_cbHandle - 		the registered callback handle.
* 		if_cbCorrelator - 	the applications context for this call.
* 		if_errorReporting - 	the desired callback.
* 		n_handles - 		the number of interfaces to delete.
* 		if_HandleArray - 	pointer to an array of handles of the interfaces to be deleted.
* Parameters
*		None
* Asynchronous Error Codes:
* 		NPF_NO_ERROR: Operation successful.
* 		NPF_IF_E_INVALID_PARAM: Interface not deleted.
* Asynchronous response :
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) 
*		will be passed to the callback function, in one or more invocations. 
*		Each response contains the handle of the deleted interface,
*		or a possible error code. The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfDelete(
                NPF_IN NPF_callbackHandle_t if_cbHandle,
                NPF_IN NPF_correlator_t if_cbCorrelator,
                NPF_IN NPF_errorReporting_t if_errorReporting,
                NPF_IN NPF_uint32_t n_handles,
                NPF_IN NPF_IfHandle_t *if_HandleArray)
{
    NPF_uint32_t	count;
    request_t		*request;
    NPF_error_t		ret;

    REQUEST(n_handles, NPF_IF_DELETE, if_cbHandle, 
		if_cbCorrelator, if_errorReporting, request)
    for(count = 0; count < n_handles; count++)
    {
	IfDelete(if_HandleArray[count], request);
    }
    return NPF_NO_ERROR;
}


/*****************************************************************************
* Name: NPF_IfBind
*
* Description:	Function to Bind one or more pairs of Interfaces
*		This function binds one or more pairs of interfaces in parent-child 
*		relationships. Each binding associates two interfaces with each other, 
*		one as parent, and one as child. Multiple bindings can be made in a single
*		call. An interface can have multiple parents; it can also have multiple 
*		children. Such relationships are indicated by multiple one-to-one binding 
*		entries, since a single many-to-one binding entry is not supported. 
*		An interface can be at the same time the parent of one and the child of 
*		another. An implementation SHOULD return an error if cycles occur 
*		(e.g. an interface is the child of one of its own children) can have, 
*		or restrict the depth of the hierarchy.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the application's context for this call.
*		if_errorReporting: the desired callback.
*		if_nbinds: number of bindings in the array.
*		if_bindArray: pointer to an array of interface handle parent/child bindings.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_CHILD_HANDLE: Child Handle is null or invalid; no binding done.
*		NPF_IF_E_INVALID_PARENT_HANDLE: Parent Handle is null or invalid; no binding done.
*		NPF_IF_E_INVALID_PARAM; Binding failed.No binding done.
* Asynchronous response:
*		A total of n_binds asynchronous responses (NPF_IfAsyncResponse_t) 
*		will be passed to the callback function, in one or more invocations. 
*		Each response contains the parent interface handle and a possible error code. 
*		The particular binding to which the response code pertains is identified 
*		in the callback by the two handles: the parent handle is in the usual 
*		ifHandle position, and the child handle is in the union part of 
*		the callback structure.
*******************************************************************************/
NPF_error_t NPF_IfBind(
                NPF_IN NPF_callbackHandle_t if_cbHandle,
                NPF_IN NPF_correlator_t if_cbCorrelator,
                NPF_IN NPF_errorReporting_t if_errorReporting,
                NPF_IN NPF_uint32_t if_nbinds,
                NPF_IN NPF_IfBinding_t *if_bindArray)
{
    NPF_uint32_t	count;
    request_t		*request;
    NPF_error_t		ret;
    REQUEST(if_nbinds, NPF_IF_BIND, if_cbHandle, 
		if_cbCorrelator, if_errorReporting, request)
    for(count = 0; count < if_nbinds; count++)
    {
	NPF_IfBinding_t  *ifBind = &if_bindArray[count];
	IfBind(ifBind, request);
    }
    return NPF_NO_ERROR;
}


/*****************************************************************************
* Name: NPF_IfGenericStatsGet
*
* Description: This function returns, via a callback, a pointer to a generic 
*		interface statistics structure containing the current counter 
*		values for one or more indicated interfaces.
* Input Parameters
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the applications context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces to get statistics for.
*		if_HandleArray: pointer to an array of interface handles.
* Output Parameters:
*		None
* Asynchronous Error Codes
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) 
*		will be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle or a possible
*		error code. If the error code indicates success, the union in the 
*		callback response structure contains a pointer to the 
*		NPF_IfStatistics_t structure for that interface.
*******************************************************************************/
NPF_error_t NPF_IfGenericStatsGet(
                NPF_IN NPF_callbackHandle_t if_cbHandle,
                NPF_IN NPF_correlator_t if_cbCorrelator,
                NPF_IN NPF_errorReporting_t if_errorReporting,
                NPF_IN NPF_uint32_t n_handles,
                NPF_IN NPF_IfHandle_t *if_HandleArray)
{
    NPF_uint32_t	count;
    request_t		*request;
    NPF_error_t		ret;

    REQUEST(n_handles, NPF_IF_STATS_GET, if_cbHandle, 
		if_cbCorrelator, if_errorReporting, request)
    for(count = 0; count < n_handles; count++)
    {
	IfGenericStatsGet(if_HandleArray[count], request);
    }
    return NPF_NO_ERROR;
}


/*****************************************************************************
* Name: NPF_IfAttrSet
*
* Description:
*		This function sets all the attributes of one or more interfaces, 
*		from the contents of an array of structures passed by the caller, 
*		as defined in NPF_IfGeneric_t. Ownership of the structure memory 
*		remains with the caller (the API implementation must copy all 
*		needed contents before returning). Any single attribute can be set 
*		with its own function call; this function is included as a way to set 
*		multiple attributes atomically and efficiently. 
*		Note: the number of NPF_IfGeneric_t structures and the number of
*		interface handles in the two arrays must be the same, equal to the 
*		n_handles argument. This function sets a different set of attributes 
*		for each named interface.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the applications context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces to set attributes for.
*		if_HandleArray: pointer to an array of interface handles.
*		if_StructArray: pointer to a structurean array of structures containing 
*				the new interface attributes.
* Output Parameters:
*		None
* Asynchronous Error Codes
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_ATTRIBUTE: An attribute (other than those mentioned below) 
*		was invalid.
* Generic Interface Errors:
*		NPF_IF_E_INVALID_HANDLE: if_Handle is null or invalid.
*		NPF_IF_E_INVALID_SPEED: Invalid interface speed parameter.
*		NPF_IF_E_INVALID_IF_TYPE: Invalid interface type code.
*		NPF_IF_E_INVALID_ADMIN_STATUS: Invalid administrative status code.
* IPv4 Interface Errors:
*		NPF_IF_E_INVALID_IPADDR: Invalid IP address
*		NPF_IF_E_INVALID_NETPLEN: Invalid network prefix length
*		NPF_IF_E_INVALID_MTU: MTU value is invalid.
*		NPF_IF_E_INVALID_FIB_HANDLE: FIB handle is invalid.
* LAN Interface Errors:
*		NPF_IF_E_INVALID_PORT: Port specification is invalid.
*		NPF_IF_E_INVALID_PROMISC_MODE: Promiscuous Mode code is invalid.
*		NPF_IF_E_INVALID_LAN_SPEED: LAN speed code is invalid.
* ATM and POS Interface Errors:
*		NPF_IF_E_INVALID_PORT: Port specification is invalid.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) will be 
*		passed to the callback function, in one or more invocations. Each response 
*		contains an interface handle and a success code or a possible error code for 
*		that interface. The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfAttrSet(
                 NPF_IN NPF_callbackHandle_t if_cbHandle,
                 NPF_IN NPF_correlator_t if_cbCorrelator,
                 NPF_IN NPF_errorReporting_t if_errorReporting,
                 NPF_IN NPF_uint32_t n_handles,
                 NPF_IN NPF_IfHandle_t *if_HandleArray,
                 NPF_IN NPF_IfGeneric_t *if_StructArray)

{
    uint32_t		count;
    request_t		*request;
    NPF_error_t		ret;

    IF_FUNCTION();
    REQUEST(n_handles, NPF_IF_ATTR_SET, if_cbHandle, 
		if_cbCorrelator, if_errorReporting, request)
    for(count = 0; count < n_handles; count ++)
    {
	 IfAttrSet(if_HandleArray[count], &if_StructArray[count], request);
    }
    return NPF_NO_ERROR;
}


/*****************************************************************************
* Name: NPF_IfCreateAndSet
*
* Description: This function simultaneously creates and sets all the attributes 
*		of one or more interfaces, from the contents of an array of structures 
*		passed by the caller (NPF_IfGeneric_t). Each interface is created
*		with a different set of attributes. Ownership of the structure memory 
*		remains with the caller (the API implementation must copy all contents 
*		before returning).
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the applications context for this call.
*		if_errorReporting: the desired callback.
*		n_if: the number of interfaces to set attributes for.
*		if_StructArray: pointer to an array of structures containing the new 
*		interface attributes.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_ATTRIBUTE: An attribute (other than those mentioned below) 
*		was invalid.
* Generic Interface Errors:
*		NPF_IF_E_INVALID_HANDLE : if_Handle is null or invalid.
*		NPF_IF_E_INVALID_SPEED: Invalid interface speed parameter.
*		NPF_IF_E_INVALID_IF_TYPE: Invalid interface type code.
*		NPF_IF_E_INVALID_ADMIN_STATUS: Invalid administrative status code.
* IPv4 Interface Errors:
*		NPF_IF_E_INVALID_IPADDR: Invalid IP address
*		NPF_IF_E_INVALID_NETPLEN: Invalid network prefix length
*		NPF_IF_E_INVALID_MTU: MTU value is invalid.
*		NPF_IF_E_INVALID_FIB_HANDLE: FIB handle is invalid.
* LAN Interface Errors:
*		NPF_IF_E_INVALID_PORT: Port specification is invalid.
*		NPF_IF_E_INVALID_PROMISC_MODE: Promiscuous Mode code is invalid.
*		NPF_IF_E_INVALID_LAN_SPEED: LAN speed code is invalid.
* ATM and POS Interface Errors:
*		NPF_IF_E_INVALID_PORT: Port specification is invalid.
* Asynchronous response:
* 		A total of n_if asynchronous responses (NPF_IfAsyncResponse_t) will be 
*		passed to the callback function, in one or more invocations. 
*		Each response contains the new interface handle and a success code
*		or a possible error code if an interface could not be created or any 
*		attributes could not be set. Responses are linked to interface attributes 
*		in the following way: for each response, the union in the response
*		structure contains the corresponding index of the if_StructArray element 
*		that contained its attributes. For example, the response for the first array 
*		element will include an Interface Handle and an arrayIndex value of zero; 
*		the response for the tenth array element an arrayIndex of 9, and so on.
*******************************************************************************/
NPF_error_t NPF_IfCreateAndSet(
                 NPF_IN NPF_callbackHandle_t if_cbHandle,
                 NPF_IN NPF_correlator_t if_cbCorrelator,
                 NPF_IN NPF_errorReporting_t if_errorReporting,
                 NPF_IN NPF_uint32_t n_if,
                 NPF_IN NPF_IfGeneric_t *if_StructArray)

{
    uint32_t		count;
    request_t		*request;
    NPF_error_t		ret;

    IF_FUNCTION();
    REQUEST(n_if, NPF_IF_CREATE_AND_SET, if_cbHandle, 
		if_cbCorrelator, if_errorReporting, request)
    for(count = 0; count < n_if; count ++)
    {
	 IfCreateAndSet(&if_StructArray[count], request);
    }
    return NPF_NO_ERROR;
}


/*****************************************************************************
* Name: NPF_IfEnable
*
* Description: This function administratively enables one or more interfaces: 
*		if the interface is operationally ready, it can now send and receive packets.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the applications context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces to enable.
*		if_HandleArray: pointer to an array of interface handles.
* Output Parameters:
*	None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) 
*		will be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success
*		code or a possible error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfEnable(
                 NPF_IN NPF_callbackHandle_t if_cbHandle,
                 NPF_IN NPF_correlator_t if_cbCorrelator,
                 NPF_IN NPF_errorReporting_t if_errorReporting,
                 NPF_IN NPF_uint32_t n_handles,
                 NPF_IN NPF_IfHandle_t *if_HandleArray)
{
    NPF_uint32_t	count;
    request_t		*request;
    NPF_error_t		ret;
    REQUEST(n_handles, NPF_IF_ENABLE, if_cbHandle, 
		if_cbCorrelator, if_errorReporting, request)
    for(count = 0; count < n_handles; count++)
    {
	IfEnable(if_HandleArray[count], request);
    }
    return NPF_NO_ERROR;
}


/*****************************************************************************
* Name: NPF_IfDisable
*
* Description: This function disables one or more interfaces, administratively 
*		(but not operationally). Once disabled, it can no longer send or ]
*		receive packets.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the applications context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces to disable.
*		if_HandleArray: pointer to an array of interface handles.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: an if_Handle is null or invalid.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) 
*		will be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a 
*		possible error code for that interface. The union in the callback 
*		response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfDisable(
                 NPF_IN NPF_callbackHandle_t if_cbHandle,
                 NPF_IN NPF_correlator_t if_cbCorrelator,
                 NPF_IN NPF_errorReporting_t if_errorReporting,
                 NPF_IN NPF_uint32_t n_handles,
                 NPF_IN NPF_IfHandle_t *if_HandleArray)
{
    NPF_uint32_t	count;
    request_t		*request;
    NPF_error_t		ret;
    REQUEST(n_handles, NPF_IF_DISABLE, if_cbHandle, 
		if_cbCorrelator, if_errorReporting, request)
    for(count = 0; count < n_handles; count++)
    {
	IfDisable(if_HandleArray[count], request);
    }
    return NPF_NO_ERROR;
}


/*****************************************************************************
* Name: NPF_IfOperStatusGet
*
* Descriiptior: This function returns the operational status of one or more interfaces.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the application's context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces to query.
*		if_HandleArray: pointer to an array of interface handles.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) will 
*		be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a possible 
*		error code for that interface. If the code indicates success, 
*		the union in the callback response structure contains the operational status 
*		of the interface.
*******************************************************************************/
NPF_error_t NPF_IfOperStatusGet(
                 NPF_IN NPF_callbackHandle_t if_cbHandle,
                 NPF_IN NPF_correlator_t if_cbCorrelator,
                 NPF_IN NPF_errorReporting_t if_errorReporting,
                 NPF_IN NPF_uint32_t n_handles,
                 NPF_IN NPF_IfHandle_t *if_HandleArray)

{
    NPF_uint32_t	count;
    request_t		*request;
    NPF_error_t		ret;

    REQUEST(n_handles, NPF_IF_OPER_STATUS_GET, if_cbHandle, 
		if_cbCorrelator, if_errorReporting, request)
    for(count = 0; count < n_handles; count++)
    {
	NPF_IfHandle_t  ifHandle = if_HandleArray[count];
	IfOperStatusGet(ifHandle, request);
    }
    return NPF_NO_ERROR;
}


/*****************************************************************************
* Name: NPF_IfLAN_SrcAddrGet
*
* Description: This function returns, via a callback, the Source MAC address of 
*		a LAN interface. The Source MAC address is the address to be sent 
*		by default as Source Address in packets sent by the interface, and the
*		default Destination Address to "listen" for in received packets.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the application's context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces to get the MAC address for.
*		if_HandleArray: pointer to an array of interface handles.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: an if_Handle is null or invalid, or is not a LAN interface.
*		NPF_IF_E_NO_SRC_ADDRESS: No source address is present in this interface.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) 
*		will be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a possible 
*		error code for that interface. If the code indicates success, 
*		the union in the callback response structure contains the Source MAC address 
*		of the interface.
*******************************************************************************/
NPF_error_t NPF_IfLAN_SrcAddrGet(
                 NPF_IN NPF_callbackHandle_t if_cbHandle,
                 NPF_IN NPF_correlator_t if_cbCorrelator,
                 NPF_IN NPF_errorReporting_t if_errorReporting,
                 NPF_IN NPF_uint32_t n_handles,
                 NPF_IN NPF_IfHandle_t *if_HandleArray)
{
    return NPF_E_UNKNOWN;;
}


/*****************************************************************************
* Name: NPF_IfLAN_SrcAddrSet
*
* Description: This function sets the Source MAC address of one or more LAN interfaces. 
*		If more than one LAN interface is specified, each receives its MAC 
*		address from a different element of the source address array.
*		The Source MAC address is the address to be sent by default as Source 
*		Address in packets sent by the interface, and the default Destination 
*		Address to "listen" for in received packets.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the applications context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces to set the MAC addresses for.
*		if_HandleArray: pointer to an array of interface handles.
*		if_LAN_SrcAddrArray: pointer to an array of MAC addresses containing 
*					one address for each interface handle.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: an if_Handle is null or invalid, or is not a LAN interface.
*		NPF_IF_E_INVALID_MAC_ADDR: Source MAC address is invalid.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) 
*		will be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a possible 
*		error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfLAN_SrcAddrSet(
                 NPF_IN NPF_callbackHandle_t if_cbHandle,
                 NPF_IN NPF_correlator_t if_cbCorrelator,
                 NPF_IN NPF_errorReporting_t if_errorReporting,
                 NPF_IN NPF_uint32_t n_handles,
                 NPF_IN NPF_IfHandle_t *if_HandleArray,
                 NPF_IN NPF_MAC_Address_t *if_LAN_SrcAddrArray)

{
    return NPF_E_UNKNOWN;;
}


/*****************************************************************************
* Name: NPF_IfLAN_MAC_RcvAddrListSet
*
* Description: This function sets a list of receive MAC addresses 
*		(represented as an array) for one or more interfaces.
*		The given list replaces any list already present in the interface. 
*		If the if_n_MAC_Addrs parameter is zero, the entire list is deleted. 
*		The same address list is applied to all indicated interfaces.
*		Note: it is not necessary to include the Source MAC Address in an interface's 
*		list of receive MAC addresses; LAN interfaces should receive on this address 
*		by default, and the list of receive MAC addresses is in addition to 
*		the source address.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the application's context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces to set the address list for.
*		if_HandleArray: pointer to an array of interface handles.
*		if_n_MAC_Addrs: number of MAC addresses in the array
*		if_LAN_AddrArray: pointer to the array of receive MAC addresses to set on all the
*		interfaces.
* Output Parameters
*		None
* Asynchronous Error Codes
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: an if_Handle is null or invalid, or is not a LAN interface.
*		NPF_IF_E_INVALID_MAC_ADDR: One or more of the MAC addresses is invalid for
*		receiving.
* Asynchronous response
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) 
		will be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a possible 
*		error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfLAN_MAC_RcvAddrListSet(
                 NPF_IN NPF_callbackHandle_t if_cbHandle,
                 NPF_IN NPF_correlator_t if_cbCorrelator,
                 NPF_IN NPF_errorReporting_t if_errorReporting,
                 NPF_IN NPF_uint32_t n_handles,
                 NPF_IN NPF_IfHandle_t *if_HandleArray,
                 NPF_IN NPF_uint32_t if_n_MAC_Addrs,
                 NPF_IN NPF_MAC_Address_t *if_LAN_AddrArray)
{
    return NPF_E_UNKNOWN;;
}


/*****************************************************************************
* Name:  NPF_IfLAN_MAC_RcvAddrListAdd
*
* Description: This function adds addresses to a list of receive MAC addresses 
*		(represented as an array) in thefor one or more interfaces. 
*		The same list of addresses is added to each interface.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the applications context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces to add the addresses to.
*		if_HandleArray: the handle of the interface.
*		if_n_MAC_Addrs: number of MAC addresses in the array
*		if_LAN_AddrArray: pointer to an array of receive MAC addresses to add 
*					to all the indicated interfaces.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful
*		NPF_IF_E_INVALID_HANDLE: an if_Handle is null or invalid, 
*					or not an LAN interface.
*		NPF_IF_E_INVALID_MAC_ADDR: One or more of the MAC addresses 
*					is invalid for receiving.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) 
*		will be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a possible 
*		error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfLAN_MAC_RcvAddrListAdd(
                 NPF_IN NPF_callbackHandle_t if_cbHandle,
                 NPF_IN NPF_correlator_t if_cbCorrelator,
                 NPF_IN NPF_errorReporting_t if_errorReporting,
                 NPF_IN NPF_uint32_t n_handles,
                 NPF_IN NPF_IfHandle_t *if_HandleArray,
                 NPF_IN NPF_uint32_t if_n_MAC_Addrs,
                 NPF_IN NPF_MAC_Address_t *if_LAN_AddrArray)
{
    return NPF_E_UNKNOWN;;
}


/*****************************************************************************
* Name: NPF_IfLAN_PromiscSet
*
* Description: This function puts one or more LAN interfaces in "promiscuous" mode 
*		(i.e. receives all packets, regardless of the destination MAC address).
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the applications context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces in the if_HandleArray array.
*		if_HandleArray: pointer to an array of interface handles.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: an if_Handle is null or invalid, or is not a LAN interface.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t)
*		 will be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a possible 
*		error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfLAN_PromiscSet(
                 NPF_IN NPF_callbackHandle_t if_cbHandle,
                 NPF_IN NPF_correlator_t if_cbCorrelator,
                 NPF_IN NPF_errorReporting_t if_errorReporting,
                 NPF_IN NPF_uint32_t n_handles,
                 NPF_IN NPF_IfHandle_t *if_HandleArray)
{
    NPF_uint32_t	count;
    request_t		*request;
    NPF_error_t		ret;

    REQUEST(n_handles, NPF_IF_LAN_PROMISC_SET, if_cbHandle, 
		if_cbCorrelator, if_errorReporting, request)
    for(count = 0; count < n_handles; count++)
    {
	IfPromiscSet(if_HandleArray[count], request);
    }
    return NPF_NO_ERROR;
}


/*****************************************************************************
* Name:  NPF_IfLAN_PromiscClear
*
* Description:  This function reverses the effect of NPF_IfLAN_PromiscSet() 
*		on one or more interfaces.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the applications context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces in the if_Handle array.
*		if_HandleArray: pointer to an array of interface handles.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: an if_Handle is null or invalid, or is not a LAN interface.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) 
*		will be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a possible 
*		error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfLAN_PromiscClear(
                 NPF_IN NPF_callbackHandle_t if_cbHandle,
                 NPF_IN NPF_correlator_t if_cbCorrelator,
                 NPF_IN NPF_errorReporting_t if_errorReporting,
                 NPF_IN NPF_uint32_t n_handles,
                 NPF_IN NPF_IfHandle_t *if_HandleArray)
{
    NPF_uint32_t	count;
    request_t		*request;
    NPF_error_t		ret;

    REQUEST(n_handles, NPF_IF_LAN_PROMISC_CLEAR, if_cbHandle, 
		if_cbCorrelator, if_errorReporting, request)
    for(count = 0; count < n_handles; count++)
    {
	NPF_IfHandle_t  ifHandle = if_HandleArray[count];
	IfPromiscClear(ifHandle, request);
    }
    return NPF_NO_ERROR;
}

/*****************************************************************************
* Name: NPF_IfLAN_FullDuplexSet
*
* Description: This function sets Full Duplex Mode on one or more LAN interfaces. 
*		It is meaningful on 10/100 megabit Ethernet interfaces, 
*		where this option is selectable. Implementation of this function is optional.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the applications context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces in the if_HandleArray array.
*		if_HandleArray: pointer to an array of interface handles.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid, or is not a LAN interface.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) 
*		will be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a possible 
*		error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfLAN_FullDuplexSet(
                 NPF_IN NPF_callbackHandle_t if_cbHandle,
                 NPF_IN NPF_correlator_t if_cbCorrelator,
                 NPF_IN NPF_errorReporting_t if_errorReporting,
                 NPF_IN NPF_uint32_t n_handles,
                 NPF_IN NPF_IfHandle_t *if_HandleArray)
{
    return NPF_E_UNKNOWN;;
}


/*****************************************************************************
* Name: NPF_IfLAN_FullDuplexClear
*
* Description:  This function reverses the effect of NPF_IfLAN_FullDuplexSet() 
*		on one or more LAN interfaces. It is meaningful on 10/100 megabit 
*		Ethernet interfaces, where this option is selectable. 
*		Implementation of this function is optional.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the applications context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces in the if_HandleArray array.
*		if_HandleArray: pointer to an array of interface handles.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid, or is not a LAN interface.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) will 
*		be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a 
*		possible error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfLAN_FullDuplexClear(
                 NPF_IN NPF_callbackHandle_t if_cbHandle,
                 NPF_IN NPF_correlator_t if_cbCorrelator,
                 NPF_IN NPF_errorReporting_t if_errorReporting,
                 NPF_IN NPF_uint32_t n_handles,
                 NPF_IN NPF_IfHandle_t *if_HandleArray)
{
    return NPF_E_UNKNOWN;;
}


/*****************************************************************************
* Name: NPF_IfLAN_SpeedSet
*
* Description: This function sets the speed (10 or 100 megabit/second) or Autosense 
*		on one or more LAN interfaces. It is meaningful on 10/100 megabit Ethernet 
*		interfaces, where this option is selectable. 
*		Implementation of this function is optional.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the applications context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces in the if_HandleArray array.
*		if_HandleArray: pointer to an array of interface handles.
*		speedArray: pointer to an array of LAN Speed codes, 
*			one for each interface in the if_HandleArray array.
* Output Parameters:
*		None
Asynchronous Error Codes
*		NPF_NO_ERROR : Operation successful.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid, or is not a LAN interface.
*		NPF_IF_E_INVALID_SPEED: One of the codes in the speedArray array is not valid.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) will 
*		be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a 
*		possible error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfLAN_SpeedSet(
                 NPF_IN NPF_callbackHandle_t if_cbHandle,
                 NPF_IN NPF_correlator_t if_cbCorrelator,
                 NPF_IN NPF_errorReporting_t if_errorReporting,
                 NPF_IN NPF_uint32_t n_handles,
                 NPF_IN NPF_IfHandle_t *if_HandleArray,
                 NPF_IN NPF_IfLAN_Speed_t *speedArray)
{
    uint32_t            count;
    request_t           *request;
    NPF_error_t         ret;

    IF_FUNCTION();
    REQUEST(n_handles, NPF_IF_LAN_SPEED_SET, if_cbHandle,
                if_cbCorrelator, if_errorReporting, request)
    for(count = 0; count < n_handles; count++)
    {
        IfSpeedSet(if_HandleArray[count], speedArray[count], request);
    }
    return NPF_NO_ERROR;
}


/*****************************************************************************
* Name: NPF_IfLAN_FlowControlTxEnable
*
* Description: This function enables sending flow control messages on one or more 
*		LAN interfaces. It is meaningful on 10/100 megabit interfaces in 
*		full duplex mode, or on gigabit Ethernet interfaces. 
*		Implementation of this function is optional.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the applications context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces in the if_HandleArray array.
*		if_HandleArray: pointer to an array of interface handles.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid, or is not a LAN interface.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) will 
*		be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a 
*		possible error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfLAN_FlowControlTxEnable(
                 NPF_IN NPF_callbackHandle_t if_cbHandle,
                 NPF_IN NPF_correlator_t if_cbCorrelator,
                 NPF_IN NPF_errorReporting_t if_errorReporting,
                 NPF_IN NPF_uint32_t n_handles,
                 NPF_IN NPF_IfHandle_t *if_HandleArray)
{
    return NPF_E_UNKNOWN;;
}


/*****************************************************************************
* Name: NPF_IfLAN_FlowControlTxDisable
*
* Description: This function disables the sending of flow control messages on one or 
*		more LAN interfaces. It is meaningful on 10/100 megabit interfaces in 
*		full duplex mode, or on gigabit Ethernet interfaces.
*		Implementation of this function is optional.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the applications context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces in the if_HandleArray array.
*		if_HandleArray: pointer to an array of interface handles.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid, or is not a LAN interface.
*******************************************************************************/
NPF_error_t NPF_IfLAN_FlowControlTxDisable(
                 NPF_IN NPF_callbackHandle_t if_cbHandle,
                 NPF_IN NPF_correlator_t if_cbCorrelator,
                 NPF_IN NPF_errorReporting_t if_errorReporting,
                 NPF_IN NPF_uint32_t n_handles,
                 NPF_IN NPF_IfHandle_t *if_HandleArray)
{
    return NPF_E_UNKNOWN;;
}


/*****************************************************************************
* Name: NPF_IfLAN_FlowControlRxEnable
*
* Description: This function enables responding to received flow control messages on 
*		one or more LAN interfaces. It is meaningful on 10/100 megabit interfaces 
*		in full duplex mode, or on gigabit Ethernet interfaces.
*		Implementation of this function is optional.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the applications context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces in the if_HandleArray array.
*		if_HandleArray: pointer to an array of interface handles.
* Output Parameters:
None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid, or is not a LAN interface.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) will 
*		be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a 
*		possible error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfLAN_FlowControlRxEnable(
                 NPF_IN NPF_callbackHandle_t if_cbHandle,
                 NPF_IN NPF_correlator_t if_cbCorrelator,
                 NPF_IN NPF_errorReporting_t if_errorReporting,
                 NPF_IN NPF_uint32_t n_handles,
                 NPF_IN NPF_IfHandle_t *if_HandleArray)
{
    return NPF_E_UNKNOWN;;
}


/*****************************************************************************
* Name: NPF_IfLAN_FlowControlRxDisable
*
* Description: This function disables responding to received flow control messages 
*		on one or more LAN interfaces. It is meaningful on 10/100 megabit 
*		interfaces in full duplex mode, or on gigabit Ethernet interfaces.
*		Implementation of this function is optional.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the applications context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces in the if_HandleArray array.
*		if_HandleArray: pointer to an array of interface handles.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid, or is not a LAN interface.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) will 
*		be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a 
*		possible error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfLAN_FlowControlRxDisable(
                 NPF_IN NPF_callbackHandle_t if_cbHandle,
                 NPF_IN NPF_correlator_t if_cbCorrelator,
                 NPF_IN NPF_errorReporting_t if_errorReporting,
                 NPF_IN NPF_uint32_t n_handles,
                 NPF_IN NPF_IfHandle_t *if_HandleArray)
{
    return NPF_E_UNKNOWN;;
}


/*****************************************************************************
* Name: NPF_IfIPv4AddrSet
*
* Description: This function sets the Primary IP address and network prefix length 
*		of one or more IPv4 interfaces. The if_Handle and if_IPv4Addr arrays 
*		must both contain the same number of entries, equal to the value of
*		n_handles. The address of each interface is set from a different element 
*		of the address array.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the applications context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces to set the IP address for.
*		if_HandleArray: pointer to an array of interface handles.
*		if_IPv4AddrArray: pointer to an array of IPv4 address/length structures.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid, or 
*					is not an IPv4 interface.
*		NPF_IF_E_INVALID_IPADDR: IP address is not a valid unicast address.
*		NPF_IF_E_INVALID_NETPLEN: Invalid network prefix length.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) will 
*		be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a 
*		possible error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfIPv4AddrSet(
                 NPF_IN NPF_callbackHandle_t if_cbHandle,
                 NPF_IN NPF_correlator_t if_cbCorrelator,
                 NPF_IN NPF_errorReporting_t if_errorReporting,
                 NPF_IN NPF_uint32_t n_handles,
                 NPF_IN NPF_IfHandle_t *if_HandleArray,
                 NPF_IN NPF_IPv4NetAddr_t *if_IPv4AddrArray)
{
    uint32_t            count;
    request_t           *request;
    NPF_error_t         ret;

    IF_FUNCTION();
    REQUEST(n_handles, NPF_IF_IPV4ADDR_SET, if_cbHandle,
                if_cbCorrelator, if_errorReporting, request)
    for(count = 0; count < n_handles; count++)
    {
        IfAddrSet(if_HandleArray[count], &if_IPv4AddrArray[count], request);
    }
    return NPF_NO_ERROR;
}


/*****************************************************************************
* Name: NPF_IfIPv4MTUSet
*
* Description: This function sets the IPv4 Maximum Transmission Unit (MTU) of 
*		one or more IPv4 interfaces. The if_HandleArray and if_IPv4MTU_Array 
*		arrays must both contain the same number of entries, equal to the value 
*		of n_handles. The MTU of each interface is set from a different element 
*		of the MTU array.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the applications context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces to set the IP address for.
*		if_HandleArray: the handle of each interface.
*		if_IPv4MTU_Array: the corresponding MTU values to be set.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid, 
*					or is not an IPv4 interface.
*		NPF_IF_E_INVALID_MTU: MTU value is invalid.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) will 
*		be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a 
*		possible error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfIPv4MTUSet(
                 NPF_IN NPF_callbackHandle_t if_cbHandle,
                 NPF_IN NPF_correlator_t if_cbCorrelator,
                 NPF_IN NPF_errorReporting_t if_errorReporting,
                 NPF_IN NPF_uint32_t n_handles,
                 NPF_IN NPF_IfHandle_t *if_HandleArray,
                 NPF_IN NPF_uint16_t *if_IPv4MTUArray)
{
    uint32_t       	count;
    request_t		*request;
    NPF_error_t    	ret;

    IF_FUNCTION();
    REQUEST(n_handles, NPF_IF_IPV4MTU_SET, if_cbHandle, 
		if_cbCorrelator, if_errorReporting, request)
    for(count = 0; count < n_handles; count++)
    {
	IfMTUSet(if_HandleArray[count], if_IPv4MTUArray[count], request);
    }
    return NPF_NO_ERROR;
}



/*****************************************************************************
* Name: NPF_IfIPv4FIBSet
*
* Description: This function associates an IPv4 Forwarding Table (FIB) with one 
*		or more IPv4 interfaces. The new FIB handle becomes the if_FIB_Handle 
*		attribute of the interface. A single FIB is associated with all the
*		interfaces in the if_HandleArray array.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the applications context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces to set the FIB for.
*		if_HandleArray: pointer to an array of interface handles.
*		if_FIB_Handle: the new FIB handle.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation complete.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid, 
*					or is not an IPv4 interface.
*		NPF_IF_E_INVALID_FIB_HANDLE: Invalid FIB handle.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) will 
*		be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a 
*		possible error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfIPv4FIBSet(
                 NPF_IN NPF_callbackHandle_t if_cbHandle,
                 NPF_IN NPF_correlator_t if_cbCorrelator,
                 NPF_IN NPF_errorReporting_t if_errorReporting,
                 NPF_IN NPF_uint32_t n_handles,
                 NPF_IN NPF_IfHandle_t *if_HandleArray,
                 NPF_IPv4UC_FwdTableHandle_t  if_FIB_Handle)
{
    uint32_t            count;
    request_t           *request;
    NPF_error_t         ret;

    IF_FUNCTION();
    REQUEST(n_handles,NPF_IF_IPV4FIB_SET, if_cbHandle,
                if_cbCorrelator, if_errorReporting, request)
    for(count = 0; count < n_handles; count++)
    {
        IfIPv4FibSet(if_HandleArray[count], if_FIB_Handle, request);
    }
    return NPF_NO_ERROR;
}

#ifdef _INCLUDE_IPV6

/*****************************************************************************
* Name: NPF_IfIPv6AddrAdd
*
* Description: This function sets the IP address and network prefix length 
*		of one or more IPv6 interfaces. The if_Handle and if_IPv6Addr arrays 
*		must both contain the same number of entries, equal to the value of
*		n_handles. The address of each interface is set from a different element 
*		of the address array.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the application's context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces to set the IP address for.
*		if_HandleArray: pointer to an array of interface handles.
*		if_IPv6AddrArray: pointer to an array of IPv6 address/length structures.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid, or 
*					is not an IPv6 interface.
*		NPF_IF_E_INVALID_IPADDR: IP address is not a valid unicast address.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) will 
*		be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a 
*		possible error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfIPv6AddrAdd(
                 NPF_IN NPF_callbackHandle_t if_cbHandle,
                 NPF_IN NPF_correlator_t if_cbCorrelator,
                 NPF_IN NPF_errorReporting_t if_errorReporting,
                 NPF_IN NPF_uint32_t n_handles,
                 NPF_IN NPF_IfHandle_t *if_HandleArray,
                 NPF_IN NPF_IPv6NetAddr_t *if_IPv6AddrArray)
{
    uint32_t            count;
    request_t           *request;
    NPF_error_t         ret;

    IF_FUNCTION();
    REQUEST(n_handles, NPF_IF_IPV6_ADDR_ADD, if_cbHandle,
                if_cbCorrelator, if_errorReporting, request)
    for(count = 0; count < n_handles; count++)
    {
        IfIPv6AddrAdd(if_HandleArray[count], &if_IPv6AddrArray[count], request);
    }
    return NPF_NO_ERROR;
}

/*****************************************************************************
* Name: NPF_IfIPv6AddrDelete
*
* Description: This function deletes the IP address and network prefix length 
*		of one or more IPv6 interfaces. The if_Handle and if_IPv6Addr arrays 
*		must both contain the same number of entries, equal to the value of
*		n_handles. The address of each interface is set from a different element 
*		of the address array.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the application's context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces to set the IP address for.
*		if_HandleArray: pointer to an array of interface handles.
*		if_IPv6AddrArray: pointer to an array of IPv6 address/length structures.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid, or 
*					is not an IPv6 interface.
*		NPF_IF_E_INVALID_IPADDR: IP address is not a valid unicast address.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) will 
*		be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a 
*		possible error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfIPv6AddrDelete(
                NPF_IN NPF_callbackHandle_t if_cbHandle,
                NPF_IN NPF_correlator_t     if_cbCorrelator,
                NPF_IN NPF_errorReporting_t if_errorReporting,
                NPF_IN NPF_uint32_t         n_handles,
                NPF_IN NPF_IfHandle_t       *if_HandleArray,
                NPF_IN NPF_IPv6NetAddr_t    *if_IPv6AddrArray)
{
    uint32_t            count;
    request_t           *request;
    NPF_error_t         ret;

    IF_FUNCTION();
    REQUEST(n_handles, NPF_IF_IPV6_ADDR_DELETE, if_cbHandle,
                if_cbCorrelator, if_errorReporting, request)
    for(count = 0; count < n_handles; count++)
    {
        IfIPv6AddrDelete(if_HandleArray[count], &if_IPv6AddrArray[count], request);
    }
    return NPF_NO_ERROR;
}

#endif /* IPV6 */

/*****************************************************************************
* Name: NPF_IfIPv6UCZoneSet
*
* Description: Set an IPv6 Interface's Scope Specific Zones address
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the application's context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces to set the IP address for.
*		if_HandleArray: pointer to an array of interface handles.
*		if_scopeArray: 
*		if_zoneArray: 
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid, or 
*					is not an IPv6 interface.
*		NPF_IF_E_INVALID_IPADDR: IP address is not a valid 
*		NPF_IF_E_INVALID_SCOPEVALUE: scope is not valid 
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) will 
*		be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a 
*		possible error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/

NPF_error_t NPF_IfIPv6UCZoneSet(
                NPF_IN NPF_callbackHandle_t if_cbHandle,
                NPF_IN NPF_correlator_t     if_cbCorrelator,
                NPF_IN NPF_errorReporting_t if_errorReporting,
                NPF_IN NPF_uint32_t         n_handles,
                NPF_IN NPF_IfHandle_t       *if_HandleArray,
                NPF_IN NPF_uint8_t          *if_scopeArray,
                NPF_IN NPF_uint32_t         *if_zoneArray)
{
    IF_FUNCTION();
    return NPF_E_UNKNOWN;
}

/*****************************************************************************
* Name: NPF_IfIPv6MTUSet
*
* Description: This function sets the IPv6 Maximum Transmission Unit (MTU) of 
*		one or more IPv6 interfaces. The if_HandleArray and if_IPv6MTU_Array 
*		arrays must both contain the same number of entries, equal to the value 
*		of n_handles. The MTU of each interface is set from a different element 
*		of the MTU array.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the applications context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces to set the IP address for.
*		if_HandleArray: the handle of each interface.
*		if_IPv6MTU_Array: the corresponding MTU values to be set.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid, 
*					or is not an IPv6 interface.
*		NPF_IF_E_INVALID_MTU: MTU value is invalid.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) will 
*		be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a 
*		possible error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfIPv6MTUSet(
                NPF_IN NPF_callbackHandle_t if_cbHandle,
                NPF_IN NPF_correlator_t     if_cbCorrelator,
                NPF_IN NPF_errorReporting_t if_errorReporting,
                NPF_IN NPF_uint32_t         n_handles,
                NPF_IN NPF_IfHandle_t       *if_HandleArray,
                NPF_IN NPF_uint16_t         *if_IPv6MTU_Array)
{
    uint32_t       	count;
    request_t		*request;
    NPF_error_t    	ret;

    IF_FUNCTION();
    REQUEST(n_handles, NPF_IF_IPV4MTU_SET, if_cbHandle, 
		if_cbCorrelator, if_errorReporting, request)
    for(count = 0; count < n_handles; count++)
    {
	IfMTUSet(if_HandleArray[count], if_IPv6MTU_Array[count], request);
    }
    return NPF_NO_ERROR;
}

/*****************************************************************************
* Name: NPF_IfIPv6FIBSet
*
* Description: This function associates an IPv6 Forwarding Table (FIB) with one 
*		or more IPv6 interfaces. The new FIB handle becomes the if_FIB_Handle 
*		attribute of the interface. A single FIB is associated with all the
*		interfaces in the if_HandleArray array.
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the applications context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces to set the FIB for.
*		if_HandleArray: pointer to an array of interface handles.
*		if_FIB_Handle: the new FIB handle.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation complete.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid, 
*					or is not an IPv6 interface.
*		NPF_IF_E_INVALID_FIB_HANDLE: Invalid FIB handle.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) will 
*		be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a 
*		possible error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfIPv6FIBSet(
                NPF_IN NPF_callbackHandle_t if_cbHandle,
                NPF_IN NPF_correlator_t     if_cbCorrelator,
                NPF_IN NPF_errorReporting_t if_errorReporting,
                NPF_IN NPF_uint32_t         n_handles,
                NPF_IN NPF_IfHandle_t       *if_HandleArray,
                NPF_IN NPF_IPv6PrefixTableHandle_t if_FIB_Handle)
{
    IF_FUNCTION();
    return NPF_E_UNKNOWN;
}

/*****************************************************************************
* Name: NPF_IfIPv6inv4DstAddrSet
*
* Description: Set IPv6 in IPv4 Tunnel Destination Address
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the application's context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces to set the IP address for.
*		if_HandleArray: pointer to an array of interface handles.
*		if_IPv4AddressArray: pointer to an array of IPv4 address/length structures.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid
*		NPF_IF_E_INVALID_IPADDR: IP address is not a valid address.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) will 
*		be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a 
*		possible error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfIPv6inv4DstAddrSet(
                NPF_IN NPF_callbackHandle_t if_cbHandle,
                NPF_IN NPF_correlator_t     if_cbCorrelator,
                NPF_IN NPF_errorReporting_t if_errorReporting,
                NPF_IN NPF_uint32_t         n_handles,
                NPF_IN NPF_IfHandle_t       *if_HandleArray,
                NPF_IN NPF_IPv4Address_t    *if_IPv4AddressArray)
{
    IF_FUNCTION();
    return NPF_E_UNKNOWN;
}

/*****************************************************************************
* Name: NPF_IfIPv6inv4SrcAddrSet
*
* Description: Set IPv6 in IPv4 Tunnel Source Address
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the application's context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces to set the IP address for.
*		if_HandleArray: pointer to an array of interface handles.
*		if_IPv4AddressArray: pointer to an array of IPv4 address/length structures.
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid
*		NPF_IF_E_INVALID_IPADDR: IP address is not a valid address.
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) will 
*		be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a 
*		possible error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfIPv6inv4SrcAddrSet(
                NPF_IN NPF_callbackHandle_t if_cbHandle,
                NPF_IN NPF_correlator_t     if_cbCorrelator,
                NPF_IN NPF_errorReporting_t if_errorReporting,
                NPF_IN NPF_uint32_t         n_handles,
                NPF_IN NPF_IfHandle_t       *if_HandleArray,
                NPF_IN NPF_IPv4Address_t    *if_IPv4AddressArray)
{
    IF_FUNCTION();
    return NPF_E_UNKNOWN;
}

/*****************************************************************************
* Name: NPF_IfIPv6inv4TTLSet
*
* Description: Set IPv6 in IPv4 Tunnel Time To Live
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the application's context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces to set the IP address for.
*		if_HandleArray: pointer to an array of interface handles.
*		if_IPv4TTLArray: pointer to an array of TTL values
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid
*		NPF_IF_E_INVALID_TTL: if TTL is not valid
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) will 
*		be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a 
*		possible error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfIPv6inv4TTLSet(
                NPF_IN NPF_callbackHandle_t if_cbHandle,
                NPF_IN NPF_correlator_t     if_cbCorrelator,
                NPF_IN NPF_errorReporting_t if_errorReporting,
                NPF_IN NPF_uint32_t         n_handles,
                NPF_IN NPF_IfHandle_t       *if_HandleArray,
                NPF_IN NPF_uint8_t          *if_IPv4TTLArray)
{
    IF_FUNCTION();
    return NPF_E_UNKNOWN;
}

/*****************************************************************************
* Name: NPF_IfIPv6inv4DSCPSet
*
* Description: Set IPv6 in IPv4 Tunnel DSCP
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the application's context for this call.
*		if_errorReporting: the desired callback.
*		n_handles: the number of interfaces to set the IP address for.
*		if_HandleArray: pointer to an array of interface handles.
*		if_IPv4DSCPArray: pointer to an array of DSCP values
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid
*		NPF_IF_E_INVALID_DSCP: if DSCP is not valid
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) will 
*		be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a 
*		possible error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfIPv6inv4DSCPSet(
                NPF_IN NPF_callbackHandle_t if_cbHandle,
                NPF_IN NPF_correlator_t     if_cbCorrelator,
                NPF_IN NPF_errorReporting_t if_errorReporting,
                NPF_IN NPF_uint32_t         n_handles,
                NPF_IN NPF_IfHandle_t       *if_HandleArray,
                NPF_IN NPF_uint8_t          *if_IPv4DSCPArray)
{
    IF_FUNCTION();
    return NPF_E_UNKNOWN;
}

/*****************************************************************************
* Name: NPF_IfIICMPv6RateLimitSet
*
* Description: Set ICMPIv6 Rate Limiting Parameters
* Input Parameters:
*		if_cbHandle: the registered callback handle.
*		if_cbCorrelator: the application's context for this call.
*		if_errorReporting: the desired callback.
*		if_rateLimitType: rate limit type
*		if_rateLimitValue: pointer to an of rate limit
* Output Parameters:
*		None
* Asynchronous Error Codes:
*		NPF_NO_ERROR: Operation successful.
*		NPF_IF_E_INVALID_HANDLE: An if_Handle is null or invalid
* Asynchronous response:
*		A total of n_handles asynchronous responses (NPF_IfAsyncResponse_t) will 
*		be passed to the callback function, in one or more invocations. 
*		Each response contains an interface handle and a success code or a 
*		possible error code for that interface. 
*		The union in the callback response structure is unused.
*******************************************************************************/
NPF_error_t NPF_IfIICMPv6RateLimitSet(
                NPF_IN NPF_callbackHandle_t      if_cbHandle,
                NPF_IN NPF_correlator_t          if_cbCorrelator,
                NPF_IN NPF_errorReporting_t      if_errorReporting,
                NPF_IN NPF_ICMPv6RateLimitType_t if_rateLimitType,
                NPF_IN NPF_uint32_t              *if_rateLimitValue)
{
    IF_FUNCTION();
    return NPF_E_UNKNOWN;
}
