/*****************************************************************************
*                  INTEL CORPORATION PROPRIETARY INFORMATION
*  This software is supplied under the terms of a license agreement or
*  nondisclosure agreement with Intel Corporation and may not be copied
*  or disclosed except in accordance with the terms of that agreement.
*  Copyright (c) 2001-3 Intel Corporation. All Rights Reserved.
*****************************************************************************/


/*****************************************************************************
* $Source: /cvs/constellations/ixa_3.0/src/cp_pdk/ControlPlane/APIImplementations/IPv4/src/npf_ipv4_api.c,v $
* $Revision: 1.8 $
* $Date: 2003/11/17 16:57:11 $
*
* Description:
*     Implementation of the IPv4/ARP component. Contains implementations of
*     functions defined in npf_ipv4_api.h as well as helper functions to
*     support those API calls.
*
*****************************************************************************/

/* Standard C Libraries */
#include <string.h>
#include <assert.h>
#include <stdio.h>

/* FPP libraries */
#include "fpp_common.h"
#include "fppapi.h"

/* Helper libraries */
#include "log.h"
#include "list.h"
#include "pil.h"

/* Helper PDK Framework header files */
#include "pdk_callback.h"
#include "pdk_ns_helper.h"
#include "npf_common.h"
#include "pdk_cm.h"

#include "npf_ns_api.h"

/*** Files written as a part of IPv4 Component ****/
/* External NPF IPV4 API Files */
#include "npf_ipv4_types.h"
#include "npf_ipv4_api.h"
/* Internal IPV4 Files */
#include "ipv4_types.h"
#include "ipv4_macros.h"
#include "ipv4_modules.h"
#include "ipv4_req_handler.h"

/* Include other Modules that are to be called */
#include "pdkmgr.h"
#include "pdk_l2idm.h"


/*****************************************************************************
*   Define Main Data Structure
*****************************************************************************/

IPV4Mgr_t	ipv4Mgr;


/*****************************************************************************
*
*   Component initialization and shutdown
*
*****************************************************************************/

/*****************************************************************************
*    Name: pdk_ipv4_init
*   
*    Description:
*        Initialization routine called by the PDK Manager. Initializes
*        internal data structures such as lists, locks and counters.
*        
*    Parameters:
*        None
*        
*    Preconditions:
*        ipv4Mgr.initialized == 0
*    
*    Postconditions:
*        internal data structures are initialized
*        ipv4Mgr.initialized == 1
*    
*    Returns:
*        NPF_SUCCESS if successful error reported otherwise
*    
*/   
NPF_RET pdk_ipv4_init()
{    
    NPF_RET             retVal;
    
    IPV4_DBG_TRACE();
    
        /* IPV4 Module Initializes L2Id Manager */ 
    if((retVal=l2idm_init()) != NPF_SUCCESS)
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
                         "L2 Id Manager Initialization failed");
        return retVal;
    }
    /* Initialize IPV4 Module */
    if((retVal=ipv4_InitIPV4Mgr()) != NPF_SUCCESS)
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
                         "IPV4 Manager Initialization failed");
        return retVal;
    }

    npf_logger_Write(LOG_IPV4, LOG_TRACE, "IPv4 module initialised");
    
    RETURN( NPF_SUCCESS );
}

/*****************************************************************************
*    Name: pdk_ipv4_shutdown
*   
*    Description:
*        Cleans up internal data structures.
*        
*    Parameters:
*        None
*        
*    Preconditions:
*        ipv4Mgr.initialized == 1
*    
*    Postconditions:
*        ipv4Mgr.initialized == 0
*        internal data structures cleaned up
*    
*    Returns:
*        NPF_SUCCESS if successful
*        NPF_COMPONENT_UNINITIALIZED if the component is not initialized
*    
*/   
NPF_RET pdk_ipv4_shutdown( void ) 
{
    NPF_RET retVal;

    IPV4_DBG_TRACE();
    CHK_INIT();
     
    /* Shutdown L2Id Manager */
    if((retVal=l2idm_shutdown()) != NPF_SUCCESS)
    {
        npf_logger_Write(LOG_IPV4, LOG_TRACE,
                         "L2 Id Manager shutdown failed");
        return(retVal);
    }

    /* Shutdown the IPV4 CP Manager */   
    if((retVal=ipv4_ShutdownIPV4Mgr()) != NPF_SUCCESS)
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
                         "IPV4 CP Manager shutdown failed");
        return(retVal);
    }
      
    npf_logger_Write(LOG_IPV4, LOG_TRACE, "IPv4 module shut down complete");

    RETURN( NPF_SUCCESS );
}





/*****************************************************************************
*
*   External NPF IPV4 API - Implementation
*
****************************************************************************/


/*****************************************************************************
*    Name: NPF_IPv4UC_Register   
*    Description:
*       Registers user with the IPv4 . This method needs to return
*       a callback specific to a user, to allow all other calls to pass this 
*       handle in all those calls which are within the scope of this user.
*    Parameters:
*        IN:  context - user context
*             response_callback - callback function
*             callbackHandle - user callback handle
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
NPF_error_t NPF_IPv4UC_Register(
    	      NPF_IN NPF_userContext_t userContext, 
              NPF_IN NPF_IPv4UC_CallbackFunc_t callbackFunc,
              NPF_OUT NPF_callbackHandle_t *callbackHandle
	      )
{
    IPV4UserInfo_t *userInfo = NULL;
    NPF_error_t retVal;
    
    IPV4_DBG_TRACE();
    CHK_INIT();
   
    /* Initializing the User specific information */ 
    if ((retVal= ipv4_InitUserInfo(userContext,callbackFunc,&userInfo)) 
		                   != NPF_SUCCESS)
    { 
        npf_logger_Write(LOG_IPV4, LOG_ERR,
		"Unable to Initialze the user Context ");
        RETURN(retVal);
    }

    /* Hive it off the User List in IPV4Mgr_t */
    if((retVal = ipv4_AddInUserListToMgr(userInfo)) != NPF_SUCCESS) 
    {  
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"Unable to Add the User Context in the List ");
       /* Remove the memory for the UserInfo */
       ipv4_RemoveUserInfo(userInfo);
       RETURN(retVal);
    }
   
    /* Give back the callback Handle */
    *callbackHandle = userInfo->cbHndl; 

    RETURN(NPF_SUCCESS);
}

/*****************************************************************************
*    Name: NPF_IPv4UC_Deregister
*   
*    Description:
*        Deregister the callback handle. No calls should be made after 
*        this call.
*    Parameters:
*        IN:  callbackHandle - callback handle
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
NPF_error_t NPF_IPv4UC_Deregister(
	NPF_IN NPF_callbackHandle_t callbackHandle) 
{
    
    NPF_error_t retVal;
    IPV4_DBG_TRACE();
    CHK_INIT();
    
    /* Search UserInfo based on cbHandle and Delete it */
   if ((retVal = ipv4_DelInUserListFrmMgr(callbackHandle)) != NPF_SUCCESS)
   {
      npf_logger_Write(LOG_IPV4, LOG_ERR,
			"Unable to delete User From Mgr. ");
   }
     
    RETURN(retVal);
}

/*****************************************************************************
*    Name: NPF_IPv4UC_EventRegister
*   
*    Description:
*        Register with the IPv4 for event callbacks.
     Parameters:
*        IN:  
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/    
NPF_error_t NPF_IPv4UC_EventRegister(
	NPF_IN NPF_userContext_t userContext,
	NPF_IN NPF_IPv4UC_EventCallFunc_t eventCallFunc,
	NPF_OUT NPF_callbackHandle_t *eventCallHandle
	) 
{   
    IPV4_DBG_TRACE();
    CHK_INIT();
    RETURN(NPF_IPV4UC_FUNCTION_NOT_SUPPORTED);
}

/*****************************************************************************
*    Name: NPF_IPv4UC_EventDeregister
*   
*    Description:
*        Deregister the event callback handle. No calls
*        should be made after this call.
*        
*    Parameters:
*        IN:  callbackHandle - callback handle
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
NPF_error_t NPF_IPv4UC_EventDeregister(
	NPF_IN NPF_callbackHandle_t callbackHandle
	) 
{
    IPV4_DBG_TRACE();
    CHK_INIT();
    RETURN(NPF_IPV4UC_FUNCTION_NOT_SUPPORTED);
}

/*****************************************************************************
*    Name: NPF_IPv4UC_GetSupported
*   
*    Description:
*          Returns supported mode 
*        
*    Parameters:
*    Preconditions:
*    
*    Postconditions:
*    Returns:
*         Supported Mode
*    
*/
NPF_IPv4UC_SupportedMode_t NPF_IPv4UC_GetSupportedModes()
{
   NPF_IPv4UC_SupportedMode_t  supMode = NPF_IPV4UC_BOTH_SUPPORTED;
   IPV4_DBG_TRACE();
   CHK_INIT();
   /* Both discrete and Unified are supported */ 
   RETURN(supMode);
} /* NPF_IPv4UC_GetSupportedModes */


/*****************************************************************************
*    Name: NPF_IPv4UC_GetSupported
*   
*    Description:
*          Returns Preferred mode 
*        
*    Parameters:
*    Preconditions:
*    
*    Postconditions:
*    Returns:
*         Preferred Mode
*    
*/
NPF_IPv4UC_PreferredMode_t NPF_IPv4UC_GetPreferredMode()
{
   NPF_IPv4UC_PreferredMode_t prefMode = NPF_IPV4UC_DISCRETE_PREFERRED;
   IPV4_DBG_TRACE();
   CHK_INIT();
   /* Discrete is preferred */ 
   RETURN(prefMode);
} /* NPF_IPv4UC_GetPreferredModes */


/*****************************************************************************
*    Name: NPF_IPv4UC_PrefixTableHandleCreate
*   
*    Description:
*       Create a prefix table handle. Necessary to make
*       calls to add/delete prefixes
*        
*    Parameters:
*        IN: callbackHandle - callback handle
*            correlator - correlator
*            error - report all/none/errors
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
NPF_error_t NPF_IPv4UC_PrefixTableHandleCreate(
	NPF_IN NPF_callbackHandle_t callbackHandle, 
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting
	)
{
    NPF_error_t        retVal;
    IPV4UsrReqInfo_t   *usrReq = NULL;
    IPV4_DBG_TRACE();
    CHK_INIT();
    
     errorReporting = NPF_REPORT_ALL;  /* Overide the original value*/
     retVal = NPF_SUCCESS;
     
     /* Initialize user request from parameters.  */ 
     if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
				     IPV4_TABLE_TYPE_PREFIX, IPV4_TRANS_DISC,
				     NPF_IPV4UC_PREFIX_TABLE_HANDLE_CREATE,
				     callbackHandle,0,&usrReq))
		          != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request could not be initialized");
	      goto InvokeCB;
     }
     /* Invoke User Request Handler : Which processes this request */ 
     if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
     {
         npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request Handler encountered error ");
	 goto InvokeCB;
     }
     
     /* User request is processed; We don't need to call FPP APIs in this
      * case. Just return Here */
InvokeCB:
    /* Inovoke callback to the User */
    ipv4_InvokeAppCallback(usrReq, NULL,
                      NPF_IPV4UC_PREFIX_TABLE_HANDLE_CREATE,retVal);

    /* Callback is given to User. Now, cleanup the context memory */ 
    ipv4_DeleteUsrReq(usrReq);

    RETURN(retVal);
}

/*****************************************************************************
*    Name: NPF_IPv4UC_NextHopTableHandleCreate
*   
*    Description:
*       Create a next hop table handle. Necessary to make
*       calls to add/delete next hops.           
*        
*    Parameters:
*        IN: callbackHandle - callback handle
*            correlator - correlator
*            error - report all/none/errors
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
NPF_error_t NPF_IPv4UC_NextHopTableHandleCreate( 
	NPF_IN NPF_callbackHandle_t callbackHandle,
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting)
{
    NPF_error_t     retVal;
    IPV4UsrReqInfo_t   *usrReq = NULL;
    IPV4_DBG_TRACE();
    CHK_INIT();
    
     errorReporting = NPF_REPORT_ALL;  /* Overide the original value*/
     retVal = NPF_SUCCESS;
     
     /* Initialize user request from parameters */ 
     if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
				     IPV4_TABLE_TYPE_NHOP, IPV4_TRANS_DISC,
				     NPF_IPV4UC_NEXT_HOP_TABLE_HANDLE_CREATE,
				     callbackHandle,0,&usrReq))
		          != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request could not be initialized");
	goto InvokeCB;
     }
     /* Invoke User Request Handler : Which processes this request */ 
     if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request Handler encountered error ");
	goto InvokeCB;
     }
     
     /* User request is processed; We don't need to call FPP APIs in this
      * case. Just return Here */
InvokeCB:
    /* Inovoke callback to the User */
    ipv4_InvokeAppCallback(usrReq, NULL,
                      NPF_IPV4UC_NEXT_HOP_TABLE_HANDLE_CREATE,retVal);
    /* Callback is given to User. Now, cleanup the context memory */ 
    ipv4_DeleteUsrReq(usrReq);

    RETURN(retVal);
}

/*****************************************************************************
*    Name: NPF_IPv4UC_PrefixNextHopTableBind
*   
*    Description:
*         Bind a prefix and a next hop table 
*    Parameters:
*                NPF_IN callbackHandle: Callback Handle
*                NPF_IN correlator: Uniquely identifies this fn. call
*                NPF_IN errorReporting: Defines the verbosity of Callbacks 
*                NPF_IN prefixTableHandle: Prefix Table Handle
*                NPF_IN nextHopTableHandle: Next Hop Table Handle
*        
*    Preconditions:
*            Both Prefix and Next-Hop Tables should have been created successfully
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/

NPF_error_t NPF_IPv4UC_PrefixNextHopTableBind(
                NPF_IN NPF_callbackHandle_t callbackHandle,
                NPF_IN NPF_correlator_t correlator,
                NPF_IN NPF_errorReporting_t errorReporting, 
                NPF_IN NPF_IPv4UC_PrefixTableHandle_t prefixTableHandle,
                NPF_IN NPF_IPv4UC_NextHopTableHandle_t nextHopTableHandle)
{
    NPF_error_t     retVal = NPF_NO_ERROR;
    IPV4UsrReqInfo_t   *usrReq = NULL;
    IPV4_DBG_TRACE();
    CHK_INIT();
   
   
    if (!prefixTableHandle || !nextHopTableHandle)
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
	                 "Invalid parameters; Table Handles are not correct ");
        retVal = NPF_INVALID_PARAMETERS;
        goto InvokeCB;
    }

    /* Initialize user request from parameters */ 
    if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
				     IPV4_TABLE_TYPE_PREFIX, IPV4_TRANS_DISC,
				    NPF_IPV4UC_PREFIX_NEXT_HOP_TABLE_BIND,
				     callbackHandle,prefixTableHandle,&usrReq))
		          != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request could not be initialized");
        goto InvokeCB;
     }

     /* Do some special processing for this request  ONLY. We need to fill
        in the bound table handle in request */ 
      usrReq->bndTblHandle = nextHopTableHandle;
 
     /* Invoke User Request Handler : Which processes this request */ 
     if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request Handler encountered error ");
        goto InvokeCB;
     }

     /* User request is processed; We don't need to call FPP APIs in this
      * case. Just return Here */

InvokeCB:
    /* Inovoke callback to the User */
    ipv4_InvokeAppCallback(usrReq, NULL,
                     NPF_IPV4UC_PREFIX_NEXT_HOP_TABLE_BIND,retVal);
    /* Callback is given to User. Now, cleanup the context memory */ 
    ipv4_DeleteUsrReq(usrReq);
    RETURN(retVal);
} /* NPF_IPv4UC_PrefixNextHopTableBind */


/*****************************************************************************
*    Name: NPF_IPv4UC_AddResTableHandleCreate
*   
*    Description:
*       Create a ARP table handle. Necessary to make
*       calls to add/delete l2 calls.           
*        
*    Parameters:
*        IN: callbackHandle - callback handle
*            correlator - correlator
*            error - report all/none/errors
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
NPF_error_t NPF_IPv4UC_AddResTableHandleCreate(
	NPF_IN NPF_callbackHandle_t callbackHandle,
	NPF_IN NPF_correlator_t correlator, 
	NPF_IN NPF_errorReporting_t errorReporting)
{
    NPF_error_t     retVal;
    IPV4UsrReqInfo_t   *usrReq = NULL;
    IPV4_DBG_TRACE();
    CHK_INIT();
    
     errorReporting = NPF_REPORT_ALL;  /* Overide the original value*/
     retVal = NPF_SUCCESS;
     
     /* Initialize user request from parameters */ 
     if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
				     IPV4_TABLE_TYPE_ADDRES, IPV4_TRANS_DISC,
				     NPF_IPV4UC_ADDRESS_RES_TABLE_HANDLE_CREATE ,
				     callbackHandle,0,&usrReq))
		          != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request could not be initialized");
	     goto InvokeCB;
     }
     /* Invoke User Request Handler : Which processes this request */ 
     if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request Handler encountered error ");
	     goto InvokeCB;
     }
     
     /* User request is processed; We don't need to call FPP APIs in this
      * case. Just return Here */

InvokeCB:
    /* Inovoke callback to the User */
    ipv4_InvokeAppCallback(usrReq, NULL,
                      NPF_IPV4UC_ADDRESS_RES_TABLE_HANDLE_CREATE,retVal);
    /* Callback is given to User. Now, cleanup the context memory */ 
    ipv4_DeleteUsrReq(usrReq);
    
    RETURN(retVal);
}

/*****************************************************************************
*    Name: NPF_IPv4UC_PrefixTableHandleDelete
*   
*    Description:
*        Delete the prefix table handle
*        
*    Parameters:
*        IN: callbackHandle - callback handle
*            correlator - correlator
*            error - report all/none/errors
*            tableHandle - prefix table handle to be deleted
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
NPF_error_t NPF_IPv4UC_PrefixTableHandleDelete(
	NPF_IN NPF_callbackHandle_t callbackHandle, 
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting,
	NPF_IN NPF_IPv4UC_PrefixTableHandle_t tableHandle)
{
    NPF_error_t     retVal;
    IPV4UsrReqInfo_t   *usrReq = NULL;
    IPV4_DBG_TRACE();
    CHK_INIT();
    
    retVal = NPF_SUCCESS;
     
    /* Basic check for the data that is to be sent to FP */  
    if ( !tableHandle)
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
		                    	"Invalid parameters");
        retVal = NPF_INVALID_PARAMETERS;
        goto InvokeCB;
    }
    
    /* Initialize user request from parameters */ 
    if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
				     IPV4_TABLE_TYPE_PREFIX, IPV4_TRANS_DISC,
				     NPF_IPV4UC_PREFIX_TABLE_HANDLE_DELETE ,
				     callbackHandle,tableHandle,&usrReq))
		          != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request could not be initialized");
	     goto InvokeCB;
     }
     /* Invoke User Request Handler : Which processes this request */ 
     if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request Handler encountered error ");
	goto InvokeCB;
     }
     
     /* User request is processed; We don't need to call FPP APIs in this
      * case. Just return Here */
InvokeCB:
    /* Invoke callback to the User */
    ipv4_InvokeAppCallback(usrReq, NULL,
                      NPF_IPV4UC_PREFIX_TABLE_HANDLE_DELETE,retVal);
    /* Callback is given to User. */ 
    ipv4_DeleteUsrReq(usrReq);

    RETURN(retVal);
}

/*****************************************************************************
*    Name: NPF_IPv4UC_NextHopTableHandleDelete
*   
*    Description:
*        Delete the next hop table handle
*        
*    Parameters:
*        IN: callbackHandle - callback handle
*            correlator - correlator
*            error - report all/none/errors
*            tableHandle - next hop table handle to be deleted        
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
NPF_error_t NPF_IPv4UC_NextHopTableHandleDelete( 
	NPF_IN NPF_callbackHandle_t callbackHandle,
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting,
	NPF_IN NPF_IPv4UC_NextHopTableHandle_t tableHandle)
{
    NPF_error_t     retVal;
    IPV4UsrReqInfo_t   *usrReq = NULL;
    IPV4_DBG_TRACE();
    CHK_INIT();
    
    retVal = NPF_SUCCESS;
     
    /* Basic check for the data that is to be sent to FP */  
    if ( !tableHandle)
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
		                    	"Invalid parameters");
        retVal = NPF_INVALID_PARAMETERS;
        goto InvokeCB;
    }
    /* Initialize user request from parameters */ 
    if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
				     IPV4_TABLE_TYPE_NHOP, IPV4_TRANS_DISC,
				     NPF_IPV4UC_NEXT_HOP_TABLE_HANDLE_DELETE ,
				     callbackHandle,tableHandle,&usrReq))
		          != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request could not be initialized");
	     goto InvokeCB;
     }
     /* Invoke User Request Handler : Which processes this request */ 
     if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request Handler encountered error ");
	     goto InvokeCB;
     }
     
     /* User request is processed; We don't need to call FPP APIs in this
      * case. Just return Here */
InvokeCB:
    /* Invoke callback to the User */
    ipv4_InvokeAppCallback(usrReq, NULL,
                      NPF_IPV4UC_NEXT_HOP_TABLE_HANDLE_DELETE,retVal);
    /* Callback is given to User. Now, cleanup the context memory */ 
    ipv4_DeleteUsrReq(usrReq);

    RETURN(retVal);
}

/*****************************************************************************
*    Name: NPF_IPv4UC_AddResTableHandleDelete
*   
*    Description:
*        Delete the address resolution table handle
*        
*    Parameters:
*        IN: callbackHandle - callback handle
*            correlator - correlator
*            error - report all/none/errors
*            tableHandle - address resolution table handle to be deleted        
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
NPF_error_t NPF_IPv4UC_AddResTableHandleDelete(
	NPF_IN NPF_callbackHandle_t callbackHandle,
	NPF_IN NPF_correlator_t correlator, 
	NPF_IN NPF_errorReporting_t errorReporting,
	NPF_IN NPF_IPv4UC_AddResTableHandle_t tableHandle)
{
    NPF_error_t     retVal;
    IPV4UsrReqInfo_t   *usrReq = NULL;
    IPV4_DBG_TRACE();
    CHK_INIT();
    
    retVal = NPF_SUCCESS;
     
    /* Basic check for the data that is to be sent to FP */  
    if ( !tableHandle)
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
		                    	"Invalid parameters");
        retVal = NPF_INVALID_PARAMETERS;
        goto InvokeCB;
    }
    /* Initialize user request from parameters */ 
    if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
				     IPV4_TABLE_TYPE_ADDRES, IPV4_TRANS_DISC,
				     NPF_IPV4UC_ADDRESS_RES_TABLE_HANDLE_DELETE ,
				     callbackHandle,tableHandle,&usrReq))
		          != NPF_SUCCESS)
     {
         npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request could not be initialized");
	 goto InvokeCB;
     }
     /* Invoke User Request Handler : Which processes this request */ 
     if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request Handler encountered error ");
	goto InvokeCB;
     }
     
     /* User request is processed; We don't need to call FPP APIs in this
      * case. Just return Here */
InvokeCB:
    /* Invoke callback to the User */
    ipv4_InvokeAppCallback(usrReq, NULL,
                      NPF_IPV4UC_ADDRESS_RES_TABLE_HANDLE_DELETE,retVal);
    /* Callback is given to User. Now, cleanup the context memory */ 
    ipv4_DeleteUsrReq(usrReq);
    RETURN(retVal);
}


/* >>>>>>>>>>>>>>>>>> Following requests are not valid >>>>>>>>>>>>>>>>>> */

/*****************************************************************************
*    Name: NPF_IPv4UC_NextHopTableAttributeQuery
*   
*    Description:
*        Query for next hop attribute - Not implemented because
*        it is not supported by the forwarding plane.
*        
*    Parameters:
*        IN:  callbackHandle - callback handle
*             correlator - correlator
*             error - error reporting = no CB, only error CB or CB
*             tableHandle - handle to next hop table
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
NPF_error_t NPF_IPv4UC_NextHopTableAttributeQuery(
	NPF_IN NPF_callbackHandle_t callbackHandle, 
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting,
	NPF_IN NPF_IPv4UC_NextHopTableHandle_t tableHandle)
{
    IPV4_DBG_TRACE();
    CHK_INIT();
    RETURN(NPF_IPV4UC_FUNCTION_NOT_SUPPORTED);
}


/*****************************************************************************
*    Name: NPF_IPv4UC_PrefixTableAttributeQuery
*   
*    Description:
*        Query for an attribute - Not Implemented yet as it is not
*        supported by the forwarding plane
*        
*    Parameters:
*        IN:  callbackHandle - callback handle
*             correlator - correlator
*             error - error reporting = no CB, only error CB or CB
*             tableHandle - handle to prefix table
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
NPF_error_t NPF_IPv4UC_PrefixTableAttributeQuery(
	NPF_IN NPF_callbackHandle_t callbackHandle,
	NPF_IN NPF_correlator_t correlator, 
	NPF_IN NPF_errorReporting_t errorReporting,
	NPF_IN NPF_IPv4UC_PrefixTableHandle_t tableHandle)
{
    IPV4_DBG_TRACE();
    CHK_INIT();
    RETURN(NPF_IPV4UC_FUNCTION_NOT_SUPPORTED);
}



/*****************************************************************************
*    Name: NPF_IPv4UC_AddResTableAttributeQuery
*   
*    Description:
*        Query for an attribute         
*    Parameters:
*        IN:  cbhandle - callback handle
*             correlator - correlator
*             error - error reporting = no CB, only error CB or CB
*             tableHandle - handle to prefix table
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/

NPF_error_t NPF_IPv4UC_AddResTableAttributeQuery(
	NPF_IN NPF_callbackHandle_t callbackHandle,
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting,
	NPF_IN NPF_IPv4UC_AddResTableHandle_t tableHandle)
{
    IPV4_DBG_TRACE();
    CHK_INIT();
    RETURN(NPF_IPV4UC_FUNCTION_NOT_SUPPORTED);
}


/* >>>>>>>>>>>>>>>>>> Following requests are sent to FP >>>>>>>>>>>>>>>>>> */


/*****************************************************************************
*    Name: NPF_IPv4UC_PrefixEntryAdd
*   
*    Description:
*        Add the Prefix table entries as denoted by the array
*        
*    Parameters:
*        IN:  callbackHandle - callback handle
*             correlator - correlator
*             error - error reporting = no CB, only error CB or CB
*             tableHandle - handle to prefix table
*             numEntries - number of entries to be added
*             entryArray - array of entries to be added       
*       
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
extern NPF_error_t NPF_IPv4UC_PrefixEntryAdd( 
	NPF_IN NPF_callbackHandle_t callbackHandle,
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting,
	NPF_IN NPF_IPv4UC_PrefixTableHandle_t tableHandle,
	NPF_IN NPF_uint32_t numEntries, 
	NPF_IN NPF_IPv4UC_Prefix_t *prefixArray,
	NPF_IN NPF_uint32_t *nextHopIdArray)
{
    NPF_error_t     retVal = NPF_NO_ERROR;
    IPV4UsrReqInfo_t   *usrReq = NULL;
    IPV4TransFPInfo_t  *transFPInfo = NULL;
    IPV4ReqToFEInfo_t  *reqToFE=NULL;
    uint32_t           i;
    FPPI_UnicastPrefixUpdateInfo_t *prefixData = NULL; /* Data to be passed to FP */

    IPV4_DBG_TRACE();
    CHK_INIT();
    
    
    /* Basic check for the data that is to be sent to FP */  
    if ( !tableHandle || !numEntries || !prefixArray || !nextHopIdArray )
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
		                    	"Invalid parameters");
        RETURN(NPF_INVALID_PARAMETERS);
    }

#if 0  /* For Array Support */
    /* Basic check for single entry : Beta Only */ 
    if(numEntries > 1)
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
		                    	"More than one entry not supported !");
        RETURN(NPF_INVALID_PARAMETERS);
    }
#endif

    /* Initialize user request from parameters */ 
    if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
				     IPV4_TABLE_TYPE_PREFIX, IPV4_TRANS_DISC,
				     NPF_IPV4UC_PREFIX_ENTRY_ADD,
				     callbackHandle,tableHandle,&usrReq))
		          != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request could not be initialized");
        retVal = NPF_E_UNKNOWN;
        goto CleanUp;
     }
     
     /* Invoke User Request Handler : Which processes this request */ 
     if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request Handler encountered error ");
        retVal = NPF_E_UNKNOWN;
        goto CleanUp;
     }
     
     /* User request is processed; We need to call FPP API here to send
      * the request */

      assert(usrReq->transInfo); /* Safety check */
     
      /* We can extract the FP info now */
      transFPInfo = &usrReq->transInfo->transFpInfo; 
      
      /* Allocate FP specific Prefix Block */
      if ((retVal = ipv4_CreateFpPfxDataBlock(numEntries,prefixArray,
                    nextHopIdArray,&prefixData)) != NPF_SUCCESS)
      {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			        "Unable to create FP Prefix Data Block");
        retVal = NPF_E_UNKNOWN;
        goto CleanUp;
      }


      /* Send the request to FP */
      for(i = 0; i < transFPInfo->numFEToSend ; i++)
      {
         reqToFE = transFPInfo->reqFEPtrArray[i];
         if(reqToFE == NULL)
         {
             npf_logger_Write(LOG_IPV4, LOG_ERR, 
			           "Request to FE is NULL");
             retVal = NPF_E_UNKNOWN;
             goto CleanUp;
         }
         
         npf_logger_Write(LOG_IPV4, LOG_TRACE, 
             "Req to FE [FEId=%d numEntries=%d] sent;",
             reqToFE->feId, numEntries);

         if((retVal = FPPAPI_ipv4_unicastAddPrefix(
                      reqToFE->feId, 
	              reqToFE->fppi_correlator, 
         (FPPI_CBHANDLE)ipv4Mgr.fppInfo.cbInfoArray[IPV4_FP_CB_ADD_PFX].cbHndl, 
	              numEntries,
		       prefixData))
			   != NPF_SUCCESS)
	      {
                npf_logger_Write(LOG_IPV4, LOG_ERR, 
			         "Unable to process request to FPP");
                retVal = NPF_E_UNKNOWN;
                goto CleanUp;
            
	      }
         npf_logger_Write(LOG_IPV4, LOG_TRACE, 
             "\n *** Waiting For callback *** \n");
     }
     /* Clear the memory created for FP request */
     if(prefixData)
        ipv4_DelFpPfxDataBlock(prefixData);
     RETURN(NPF_SUCCESS); 
CleanUp:
/* Clear the memory created for FP request */
ipv4_DelFpPfxDataBlock(prefixData);
/* Clear the request context memory */
ipv4_DeleteUsrReq(usrReq);
RETURN(retVal); 
}

/*****************************************************************************
*    Name: NPF_IPv4UC_PrefixEntryDelete
*   
*    Description:
*        Delete the Prefix table entries as denoted by the array
*        
*    Parameters:
*        IN:  callbackHandle - callback handle
*             correlator - correlator
*             error - error reporting = no CB, only error CB or CB
*             tableHandle - handle to prefix table
*             numEntries - number of entries to be deleted
*             entryArray - array of entries to be deleted
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
NPF_error_t NPF_IPv4UC_PrefixEntryDelete(
	NPF_IN NPF_callbackHandle_t callbackHandle, 
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting,
	NPF_IN NPF_IPv4UC_PrefixTableHandle_t tableHandle,
	NPF_IN NPF_uint32_t numEntries,
	NPF_IN NPF_IPv4UC_Prefix_t *prefixArray) 
{
    NPF_error_t     retVal = NPF_NO_ERROR;
    IPV4UsrReqInfo_t   *usrReq = NULL;
    IPV4TransFPInfo_t  *transFPInfo = NULL;
    IPV4ReqToFEInfo_t  *reqToFE=NULL;
    uint32_t           i;
    IPV4_DBG_TRACE();
    CHK_INIT();
    
   
    /* Basic check for the data that is to be sent to FP */  
    if ( !tableHandle || !numEntries || !prefixArray )
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"Invalid parameters");
        RETURN(NPF_INVALID_PARAMETERS);
    }

#if 0 /* For Array Support */
    /* Basic check for single entry : Beta Only */ 
    if(numEntries > 1)
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
		                    	"More than one entry not supported !");
        RETURN(NPF_INVALID_PARAMETERS);
    }
#endif

    /* Initialize user request from parameters */ 
    if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
				     IPV4_TABLE_TYPE_PREFIX, IPV4_TRANS_DISC,
				     NPF_IPV4UC_PREFIX_ENTRY_DELETE,
				     callbackHandle,tableHandle,&usrReq))
		          != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request could not be initialized");
        ipv4_DeleteUsrReq(usrReq);
        RETURN(NPF_E_UNKNOWN);
     }
     
     /* Invoke User Request Handler : Which processes this request. All the
      * data-struc. linkages are done in this routine. */ 
     if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request Handler encountered error ");
         
        ipv4_DeleteUsrReq(usrReq);
        RETURN(NPF_E_UNKNOWN);
     }
     
     /* User request is processed; We need to call FPP API here to send
      * the request */
      
     assert(usrReq->transInfo); /* Safety check */

     /* We can extract the FP info now */
     transFPInfo = &usrReq->transInfo->transFpInfo; 

     /* Send the request to FP */
     for(i = 0; i < transFPInfo->numFEToSend ; i++)
       {
           reqToFE = transFPInfo->reqFEPtrArray[i];
           if(reqToFE == NULL)
           {
               npf_logger_Write(LOG_IPV4, LOG_ERR, 
			           "Request to FE is NULL");
               ipv4_DeleteUsrReq(usrReq);
               RETURN(NPF_E_UNKNOWN);
           }
           npf_logger_Write(LOG_IPV4, LOG_TRACE, 
               "Req to FE [FEId=%d numEntries=%d] sent;",
               reqToFE->feId, numEntries);

           if((retVal = FPPAPI_ipv4_unicastDelPrefix(
                reqToFE->feId, 
		          reqToFE->fppi_correlator, 
         (FPPI_CBHANDLE)ipv4Mgr.fppInfo.cbInfoArray[IPV4_FP_CB_DEL_PFX].cbHndl, 
		          numEntries,
                (NPF_IPv4UC_Prefix_t*)prefixArray))
			    != NPF_SUCCESS)
           {
               npf_logger_Write(LOG_IPV4, LOG_ERR, 
			           "Unable to process request to FPP");
              ipv4_DeleteUsrReq(usrReq);
              RETURN(NPF_E_UNKNOWN);
           }
         npf_logger_Write(LOG_IPV4, LOG_TRACE, 
             "*** Waiting For callback ***");
       }
    RETURN(retVal); 
}

/*****************************************************************************
*    Name: NPF_IPv4UC_PrefixTableFlush
*   
*    Description:
*        Flush all the Prefix table entries
*        
*    Parameters:
*        IN:  callbackHandle - callback handle
*             correlator - correlator
*             error - error reporting = no CB, only error CB or CB
*             tableHandle - handle to prefix table
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
NPF_error_t NPF_IPv4UC_PrefixTableFlush(
	NPF_IN NPF_callbackHandle_t callbackHandle,
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting, 
	NPF_IN NPF_IPv4UC_PrefixTableHandle_t tableHandle)
{
    NPF_error_t     retVal = NPF_NO_ERROR;
    IPV4TransFPInfo_t  *transFPInfo = NULL;
    IPV4ReqToFEInfo_t  *reqToFE=NULL;
    IPV4UsrReqInfo_t   *usrReq = NULL;
    uint32_t           i;
    IPV4_DBG_TRACE();
    CHK_INIT();
    
    
    /* Basic check for the data that is to be sent to FP */  
    if ( !tableHandle)
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"Invalid parameters");
        retVal = NPF_INVALID_PARAMETERS;
        RETURN(NPF_E_UNKNOWN);
    }
 
    /* Initialize user request from parameters */ 
    if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
				     IPV4_TABLE_TYPE_PREFIX, IPV4_TRANS_DISC,
	                             NPF_IPV4UC_PREFIX_TABLE_FLUSH,
				     callbackHandle,tableHandle,&usrReq))
		          != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request could not be initialized");
        ipv4_DeleteUsrReq(usrReq);
        RETURN(NPF_E_UNKNOWN);
     }
     
     /* Invoke User Request Handler : Which processes this request. All the
      * data-struc. linkages are done in this routine. */ 
     if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request Handler encountered error ");
        ipv4_DeleteUsrReq(usrReq);
        RETURN(NPF_E_UNKNOWN);
     }
     
     /* User request is processed; We need to call FPP API here to send
      * the request */
      
     assert(usrReq->transInfo); /* Safety check */

     /* We can extract the FP info now */
     transFPInfo = &usrReq->transInfo->transFpInfo; 

     /* Send the request to FP */
     for(i = 0; i < transFPInfo->numFEToSend ; i++)
       {
           reqToFE = transFPInfo->reqFEPtrArray[i];
           if(reqToFE == NULL)
           {
               npf_logger_Write(LOG_IPV4, LOG_ERR, 
			           "Request to FE is NULL");
               ipv4_DeleteUsrReq(usrReq);
               RETURN(NPF_E_UNKNOWN);
           }
           npf_logger_Write(LOG_IPV4, LOG_TRACE, 
               "Req to FE [FEId=%d] sent;",
               reqToFE->feId);

           if((retVal = FPPAPI_ipv4_unicastFlushPrefix(
                reqToFE->feId, 
		          reqToFE->fppi_correlator, 
      (FPPI_CBHANDLE)ipv4Mgr.fppInfo.cbInfoArray[IPV4_FP_CB_FLUSH_PFX].cbHndl)) 
			          != NPF_SUCCESS)
           {
               npf_logger_Write(LOG_IPV4, LOG_ERR, 
			                "Unable to process request to FPP");
               ipv4_DeleteUsrReq(usrReq);
               RETURN(NPF_E_UNKNOWN);
           }
         npf_logger_Write(LOG_IPV4, LOG_TRACE, 
             "*** Waiting For callback ***");
       }
    RETURN(retVal); 
}

/*****************************************************************************
*    Name: NPF_IPv4UC_PrefixEntryQuery
*   
*    Description:
*        Query for an attribute for an entry         
*    Parameters:
*        IN:  cbhandle - callback handle
*             correlator - correlator
*             error - error reporting = no CB, only error CB or CB
*             tableHandle - handle to table
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/

NPF_error_t NPF_IPv4UC_PrefixEntryQuery(
	NPF_IN NPF_callbackHandle_t callbackHandle, 
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting,
	NPF_IN NPF_IPv4UC_PrefixTableHandle_t tableHandle,
	NPF_IN NPF_uint32_t numEntries,
	NPF_IN NPF_IPv4UC_Prefix_t *prefixArray)
{
    NPF_error_t     retVal = NPF_NO_ERROR;
    IPV4UsrReqInfo_t   *usrReq = NULL;
    IPV4TransFPInfo_t  *transFPInfo = NULL;
    IPV4ReqToFEInfo_t  *reqToFE=NULL;
    uint32_t           i;
    IPV4_DBG_TRACE();
    CHK_INIT();
    
    
    /* Basic check for the data that is to be sent to FP */  
    if ( !tableHandle || !numEntries || !prefixArray )
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"Invalid parameters");
        ipv4_DeleteUsrReq(usrReq);
        RETURN(NPF_INVALID_PARAMETERS);
    }
    /* Basic check for single entry : Beta Only */ 
    if(numEntries > 1)
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
		                    	"More than one entry not supported !");
        RETURN(NPF_INVALID_PARAMETERS);
    }
    
    /* Initialize user request from parameters */ 
    if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
				     IPV4_TABLE_TYPE_PREFIX, IPV4_TRANS_DISC,
	                             NPF_IPV4UC_PREFIX_ENTRY_QUERY,
				     callbackHandle,tableHandle,&usrReq))
		          != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request could not be initialized");
        ipv4_DeleteUsrReq(usrReq);
        RETURN(NPF_E_UNKNOWN);
     }
     
     /* Invoke User Request Handler : Which processes this request. All the
      * data-struc. linkages are done in this routine. */ 
     if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request Handler encountered error ");
        ipv4_DeleteUsrReq(usrReq);
        RETURN(NPF_E_UNKNOWN);
     }
     
     /* User request is processed; We need to call FPP API here to send
      * the request */
      
     assert(usrReq->transInfo); /* Safety check */

     /* We can extract the FP info now */
     transFPInfo = &usrReq->transInfo->transFpInfo; 

     /* Send the request to FP */
     for(i = 0; i < transFPInfo->numFEToSend ; i++)
       {
           reqToFE = transFPInfo->reqFEPtrArray[i];
           if(reqToFE == NULL)
           {
               npf_logger_Write(LOG_IPV4, LOG_ERR, 
			           "Request to FE is NULL");
               ipv4_DeleteUsrReq(usrReq);
               RETURN(NPF_E_UNKNOWN);
           }
           npf_logger_Write(LOG_IPV4, LOG_TRACE, 
               "Req to FE [FEId=%d numEntries=%d] sent;",
               reqToFE->feId, numEntries);

           if((retVal = FPPAPI_ipv4_unicastQueryPrefix(
                          reqToFE->feId, 
		          reqToFE->fppi_correlator, 
      (FPPI_CBHANDLE)ipv4Mgr.fppInfo.cbInfoArray[IPV4_FP_CB_QUERY_PFX].cbHndl, 
                          numEntries, 
                          prefixArray))
			    != NPF_SUCCESS)
           {
               npf_logger_Write(LOG_IPV4, LOG_ERR, 
			              "Unable to process request to FPP");
              ipv4_DeleteUsrReq(usrReq);
              RETURN(NPF_E_UNKNOWN);
           }
         npf_logger_Write(LOG_IPV4, LOG_TRACE, 
             "*** Waiting For callback ***");
       }
    RETURN(retVal); 
}



/*****************************************************************************
*    Name: NPF_IPv4UC_NextHopEntryAdd
*   
*    Description:
*        Add the next hop entry as specified by nextHopArray
*        
*    Parameters:
*        IN:  callbackHandle - callback handle
*             correlator - correlator
*             error - error reporting = no CB, only error CB or CB
*             tableHandle - handle to next hop table
*             entryArray - entry to be added
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
NPF_error_t NPF_IPv4UC_NextHopEntryAdd(
	NPF_IN NPF_callbackHandle_t callbackHandle,
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting,
	NPF_IN NPF_IPv4UC_NextHopTableHandle_t tableHandle, 
	NPF_IN NPF_uint32_t numEntries,
	NPF_IN NPF_uint32_t *nextHopIdArray,
	NPF_IN NPF_IPv4UC_NextHopArray_t *nextHopArrays)
{
    NPF_error_t     retVal = NPF_NO_ERROR;
    IPV4UsrReqInfo_t   *usrReq = NULL;
    IPV4TransFPInfo_t  *transFPInfo = NULL;
    IPV4ReqToFEInfo_t  *reqToFE=NULL;
    uint32_t           i;
    FPPI_UnicastNextHopUpdateInfo_t *fpNhopBlock = NULL;

    IPV4_DBG_TRACE();
    CHK_INIT();
    
    
    /* Basic check for the data that is to be sent to FP */  
    if ( !tableHandle || !numEntries || !nextHopArrays )
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"Invalid parameters");
        RETURN(NPF_INVALID_PARAMETERS);
    }

#if 0 /* For Array Support */
    /* Basic check for single entry : Beta Only */ 
    if(numEntries > 1)
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
		                    	"More than one entry not supported !");
        RETURN(NPF_INVALID_PARAMETERS);
    }
#endif
    
    /* Initialize user request from parameters */ 
    if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
				     IPV4_TABLE_TYPE_NHOP, IPV4_TRANS_DISC,
	                             NPF_IPV4UC_NEXT_HOP_ENTRY_ADD,
				     callbackHandle,tableHandle,&usrReq))
		          != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request could not be initialized");
        retVal = NPF_E_UNKNOWN;
        goto CleanUp; 
     }
      
     /* Invoke User Request Handler : Which processes this request */ 
     if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request Handler encountered error ");
        retVal = NPF_E_UNKNOWN;
        goto CleanUp; 
     }
 
     /* Store this context for L2Id generation logic */ 
     if(( retVal = ipv4_AddNhopCntxtToUsrReq(usrReq, numEntries,nextHopArrays,
				    nextHopIdArray)) != NPF_SUCCESS)
     {
         npf_logger_Write(LOG_IPV4, LOG_ERR,
			"Next-Hop Context couldn't be stored ");
         
         retVal = NPF_E_UNKNOWN;
         goto CleanUp; 
     }
     /* Allocate memory for FP Nhop Block in case of delete. */ 
     if((retVal = ipv4_AllocateFPNhopAddBlock(numEntries, &fpNhopBlock))  != NPF_SUCCESS)
     {
          npf_logger_Write(LOG_IPV4, LOG_ERR,
			     "Unable to Process User req to FP");
          retVal = NPF_E_UNKNOWN;
          goto CleanUp; 
     }
     
     /*Process the Nhop entry  to send to FP (NH Table + L2 Table). 
      * This also updates the list of FEs that this table is associated with */
     if(( retVal = ipv4_PrcUserReqToFPForNhopAdd(usrReq,numEntries,
				    nextHopIdArray,nextHopArrays,fpNhopBlock))
 		    != NPF_SUCCESS)
     {
         npf_logger_Write(LOG_IPV4, LOG_ERR,
			"Unable to Process User req to FP");
        retVal = NPF_E_UNKNOWN;
        goto CleanUp; 
     }
 
      
     /* We can extract the FP info now */
     transFPInfo = &usrReq->transInfo->transFpInfo; 

     /* Send the request to FP */
     for(i = 0; i < transFPInfo->numFEToSend ; i++)
     {
         reqToFE = transFPInfo->reqFEPtrArray[i];
         if(reqToFE == NULL)
         {
             npf_logger_Write(LOG_IPV4, LOG_ERR, 
			           "Request to FE is NULL");
             retVal = NPF_E_UNKNOWN;
             goto CleanUp; 
         }
           npf_logger_Write(LOG_IPV4, LOG_TRACE, 
               "Req to FE [FEId=%d numEntries=%d] sent;",
               reqToFE->feId, numEntries);
         
         if((retVal = FPPAPI_ipv4_unicastAddNextHop(
              reqToFE->feId, 
              reqToFE->fppi_correlator, 
        (FPPI_CBHANDLE)ipv4Mgr.fppInfo.cbInfoArray[IPV4_FP_CB_ADD_NHOP].cbHndl,
              numEntries,
	           fpNhopBlock ))
	             != NPF_SUCCESS)
         {
             npf_logger_Write(LOG_IPV4, LOG_ERR, 
			           "Unable to process request to FPP");
             retVal = NPF_E_UNKNOWN;
             goto CleanUp; 
         }
         npf_logger_Write(LOG_IPV4, LOG_TRACE, 
             "*** Waiting For callback ***");
     }
    if(fpNhopBlock)
       ipv4_FreeFPNhopAddBlock(numEntries,fpNhopBlock);
    RETURN(NPF_SUCCESS); 


CleanUp:
   if(fpNhopBlock)
      ipv4_FreeFPNhopAddBlock(numEntries, fpNhopBlock);
   if(usrReq)
      ipv4_DeleteUsrReq(usrReq);
   RETURN(retVal);

}

/*****************************************************************************
*    Name: NPF_IPv4UC_NextHopEntryDelete
*   
*    Description:
*        Delete the next hop entry as specified by nextHopArray
*
*    Parameters:
*        IN:  callbackHandle - callback handle
*             correlator - correlator
*             error - error reporting = no CB, only error CB or CB
*             tableHandle - handle to next hop table
*             entryArray - entry to be deleted
*        
*    Preconditions:
*            
*    Postconditions:
*        
*    Returns:
*        Standard error value
*    
*/
NPF_error_t NPF_IPv4UC_NextHopEntryDelete( 
	NPF_IN NPF_callbackHandle_t callbackHandle,
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting,
	NPF_IN NPF_IPv4UC_NextHopTableHandle_t tableHandle,
	NPF_IN NPF_uint32_t numEntries, 
	NPF_IN NPF_uint32_t *nextHopIdArray)
{
    NPF_error_t     retVal = NPF_NO_ERROR;
    IPV4UsrReqInfo_t   *usrReq = NULL;
    IPV4TransFPInfo_t  *transFPInfo = NULL;
    IPV4ReqToFEInfo_t  *reqToFE=NULL;
    uint32_t           i;
    FPPI_UnicastNextHopDeleteInfo_t   *fpNhopDelBlock = NULL;
    
    IPV4_DBG_TRACE();
    CHK_INIT();
    
    
    /* Basic check for the data that is to be sent to FP */  
    if ( !tableHandle || !numEntries || !nextHopIdArray)
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			           "Invalid parameters");
        RETURN(NPF_INVALID_PARAMETERS);
    }
    
    /* Initialize user request from parameters */ 
    if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
				     IPV4_TABLE_TYPE_NHOP, IPV4_TRANS_DISC,
	              NPF_IPV4UC_NEXT_HOP_ENTRY_DELETE,
				     callbackHandle,tableHandle,&usrReq))
		          != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request could not be initialized");
        retVal = NPF_E_UNKNOWN;
        goto CleanUp;     
     }
     
     /* Invoke User Request Handler : Which processes this request. All the
      * data-struc. linkages are done in this routine. */ 
     if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
     {
         npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request Handler encountered error ");
         retVal = NPF_E_UNKNOWN;
         goto CleanUp;     
     }
     
     assert(usrReq->transInfo); /* Safety check */

     /* Allocate memory for FP Nhop Del Block */ 
     if((retVal = ipv4_AllocateFPNhopDelBlock(numEntries,
                              &fpNhopDelBlock))  != NPF_SUCCESS)
     {
          npf_logger_Write(LOG_IPV4, LOG_ERR,
			"Unable to Process User req (Delete Nhop) to FP");
          retVal = NPF_E_UNKNOWN;
          goto CleanUp;     
     }

     /*Process the Nhop entry  to send to FP (NH Table + L2 Table). 
      * This also updates the list of FEs that this table is associated with */
     if(( retVal = ipv4_PrcUserReqToFPForNhopDel(usrReq,numEntries,
				    nextHopIdArray,fpNhopDelBlock))
 		    != NPF_SUCCESS)
     {
         npf_logger_Write(LOG_IPV4, LOG_ERR,
			"Unable to Process User req to FP");
         retVal = NPF_E_UNKNOWN;
         goto CleanUp;     
     }
 
         
     /* We can extract the FP info now */
     transFPInfo = &usrReq->transInfo->transFpInfo; 
     /* Send the request to FP */
     for(i = 0; i < transFPInfo->numFEToSend ; i++)
     {
         reqToFE = transFPInfo->reqFEPtrArray[i];
         if(reqToFE == NULL)
         {
             npf_logger_Write(LOG_IPV4, LOG_ERR, 
			           "Request to FE is NULL");
             retVal = NPF_E_UNKNOWN;
             goto CleanUp; 
         }
         npf_logger_Write(LOG_IPV4, LOG_TRACE, 
              "Req to FE [FEId=%d numEntries=%d] sent;",
               reqToFE->feId, numEntries);

         fpNhopDelBlock->nhInfo.feid = reqToFE->feId;/* Override FeId */
         
         if((retVal = FPPAPI_ipv4_unicastDelNextHop(
              reqToFE->feId, 
              reqToFE->fppi_correlator, 
       (FPPI_CBHANDLE)ipv4Mgr.fppInfo.cbInfoArray[IPV4_FP_CB_DEL_NHOP].cbHndl, 
              numEntries,
	           fpNhopDelBlock ))
	             != NPF_SUCCESS)
         {
             npf_logger_Write(LOG_IPV4, LOG_ERR, 
			        "Unable to process request to FPP");
             retVal = NPF_E_UNKNOWN;
             goto CleanUp;     
         }
         npf_logger_Write(LOG_IPV4, LOG_TRACE, 
             "*** Waiting For callback ***");
     }

    if(fpNhopDelBlock)
       ipv4_FreeFPNhopDelBlock(fpNhopDelBlock);
    RETURN(NPF_SUCCESS); 
CleanUp:
    if(usrReq) 
       ipv4_DeleteUsrReq(usrReq);
    if(fpNhopDelBlock)
       ipv4_FreeFPNhopDelBlock(fpNhopDelBlock);
    RETURN(retVal); 
}

/*****************************************************************************
*    Name: NPF_IPv4UC_NextHopTableFlush
*   
*    Description:
*       
*        
*    Parameters:
*        IN:  callbackHandle - callback handle
*             correlator - correlator
*             error - error reporting = no CB, only error CB or CB
*             tableHandle - handle to next hop table
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
NPF_error_t NPF_IPv4UC_NextHopTableFlush(
	NPF_IN NPF_callbackHandle_t callbackHandle,
	NPF_IN NPF_correlator_t correlator, 
	NPF_IN NPF_errorReporting_t errorReporting,
	NPF_IN NPF_IPv4UC_NextHopTableHandle_t tableHandle)
{

    NPF_error_t     retVal;
    IPV4UsrReqInfo_t   *usrReq = NULL;
    IPV4TransFPInfo_t  *transFPInfo = NULL;
    IPV4ReqToFEInfo_t  *reqToFE=NULL;
    uint32_t           i;
	 NPF_IN NPF_uint32_t numEntries=0; 
    FPPI_UnicastNextHopDeleteInfo_t   *fpNhopFlushBlock = NULL;
    
    IPV4_DBG_TRACE();
    CHK_INIT();
    
    retVal = NPF_SUCCESS;
   
    /* Basic check for the data that is to be sent to FP */  
    if ( !tableHandle)
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"Invalid parameters");
        retVal = NPF_INVALID_PARAMETERS;
        RETURN(NPF_E_UNKNOWN);
    }
    
    /* Initialize user request from parameters */ 
    if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
				     IPV4_TABLE_TYPE_NHOP, IPV4_TRANS_DISC,
	                             NPF_IPV4UC_NEXT_HOP_ENTRY_DELETE,
				     callbackHandle,tableHandle,&usrReq))
		          != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
		     	"User Request could not be initialized");
        goto CleanUp;     
     }
 
     /* Invoke User Request Handler : Which processes this request.
        All the data-struc. linkages are done in this routine. */ 
     if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request Handler encountered error ");
         goto CleanUp;     
     }
     
     assert(usrReq->transInfo); /* Safety check */
          /* We can extract the FP info now */
     transFPInfo = &usrReq->transInfo->transFpInfo; 
     /* Send the request to FP */
     for(i = 0; i < transFPInfo->numFEToSend ; i++)
     {
         reqToFE = transFPInfo->reqFEPtrArray[i];
         if(reqToFE == NULL)
         {
             npf_logger_Write(LOG_IPV4, LOG_ERR, 
			           "Request to FE is NULL");
             retVal = NPF_E_UNKNOWN;
             goto CleanUp;     
         }
         
         npf_logger_Write(LOG_IPV4, LOG_TRACE, 
            "Req to FE [FEId=%d] sent",
             reqToFE->feId);
         
         /* Do the relevant processing to create fpNhopFlushBlock */ 
         if((retVal = ipv4_PrcUserReqToFPForNhopFlush(
                  (FPPI_FEID)reqToFE->feId, usrReq, &numEntries,
                  &fpNhopFlushBlock)) != NPF_SUCCESS)
         {
            npf_logger_Write(LOG_IPV4, LOG_ERR,
			   "Unable to process request to FP ( Nhop Flush Operation )");
            goto CleanUp;     
         }
      
         if((retVal = FPPAPI_ipv4_unicastFlushNextHop(
              reqToFE->feId, 
              reqToFE->fppi_correlator, 
    (FPPI_CBHANDLE)ipv4Mgr.fppInfo.cbInfoArray[IPV4_FP_CB_FLUSH_NHOP].cbHndl,
              numEntries,
              fpNhopFlushBlock))
	           != NPF_SUCCESS)
         {
             npf_logger_Write(LOG_IPV4, LOG_ERR, 
			             "Unable to process request to FPP");
             goto CleanUp;     
         }
         if(fpNhopFlushBlock)
            ipv4_FreeFPNhopDelBlock(fpNhopFlushBlock);
         
         npf_logger_Write(LOG_IPV4, LOG_TRACE, 
             "*** Waiting For callback ***");
     }

    RETURN(NPF_SUCCESS); 
CleanUp:
    if(usrReq) 
       ipv4_DeleteUsrReq(usrReq);
    if(fpNhopFlushBlock)
       ipv4_FreeFPNhopDelBlock(fpNhopFlushBlock);
    RETURN(retVal); 
}

/*****************************************************************************
*    Name: NPF_IPv4UC_NextHopEntryQuery
*   
*    Description:
*        Query for an attribute for an entry         
*    Parameters:
*        IN:  cbhandle - callback handle
*             correlator - correlator
*             error - error reporting = no CB, only error CB or CB
*             tableHandle - handle to table
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/


NPF_error_t NPF_IPv4UC_NextHopEntryQuery( 
	NPF_IN NPF_callbackHandle_t callbackHandle,
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting,
	NPF_IN NPF_IPv4UC_NextHopTableHandle_t tableHandle,
	NPF_IN NPF_uint32_t numEntries, 
	NPF_IN NPF_uint32_t *nextHopIdArray)
{  
    NPF_error_t     retVal;
    IPV4UsrReqInfo_t   *usrReq = NULL;
    IPV4TransFPInfo_t  *transFPInfo = NULL;
    IPV4ReqToFEInfo_t  *reqToFE=NULL;
    uint32_t           i;
    IPV4_DBG_TRACE();
    CHK_INIT();
    
    retVal = NPF_SUCCESS;
    
    /* Basic check for the data that is to be sent to FP */  
    if ( !tableHandle || !numEntries || !nextHopIdArray)
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"Invalid parameters");
        RETURN(NPF_INVALID_PARAMETERS);
    }
    /* Basic check for single entry : Beta Only */ 
    if(numEntries > 1)
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
		                    	"More than one entry not supported !");
        RETURN(NPF_INVALID_PARAMETERS);
    }
    
    /* Initialize user request from parameters */ 
    if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
				     IPV4_TABLE_TYPE_NHOP, IPV4_TRANS_DISC,
	                             NPF_IPV4UC_NEXT_HOP_ENTRY_QUERY,
				     callbackHandle,tableHandle,&usrReq))
		          != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request could not be initialized");
         
        ipv4_DeleteUsrReq(usrReq);
        RETURN(NPF_E_UNKNOWN);
     }
     
     /* Invoke User Request Handler : Which processes this request. All the
      * data-struc. linkages are done in this routine. */ 
     if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request Handler encountered error ");
         
        ipv4_DeleteUsrReq(usrReq);
        RETURN(NPF_E_UNKNOWN);
     }
     
     /* User request is processed; We need to call FPP API here to send
      * the request */
      
     assert(usrReq->transInfo); /* Safety check */

     /* We can extract the FP info now */
     transFPInfo = &usrReq->transInfo->transFpInfo; 

     /* Send the request to FP */
     for(i = 0; i < transFPInfo->numFEToSend ; i++)
       {
           reqToFE = transFPInfo->reqFEPtrArray[i];
           if(reqToFE == NULL)
           {
             npf_logger_Write(LOG_IPV4, LOG_ERR, 
			           "Request to FE is NULL");
             ipv4_DeleteUsrReq(usrReq);
             RETURN(NPF_E_UNKNOWN);
           }
           npf_logger_Write(LOG_IPV4, LOG_TRACE, 
               "Req to FE [FEId=%d numEntries=%d] sent;",
               reqToFE->feId, numEntries);

           if((retVal = FPPAPI_ipv4_unicastQueryNextHop(
                reqToFE->feId, 
		          reqToFE->fppi_correlator, 
     (FPPI_CBHANDLE)ipv4Mgr.fppInfo.cbInfoArray[IPV4_FP_CB_QUERY_NHOP].cbHndl, 
		          numEntries,
                (NPF_uint32_t*) nextHopIdArray))
			    != NPF_SUCCESS)
           {
               npf_logger_Write(LOG_IPV4, LOG_ERR, 
			             "Unable to process request to FPP");
               
              ipv4_DeleteUsrReq(usrReq);
              RETURN(NPF_E_UNKNOWN);
           }
         npf_logger_Write(LOG_IPV4, LOG_TRACE, 
             "*** Waiting For callback ***");
       }
    RETURN(NPF_NO_ERROR); 
}



/*****************************************************************************
*    Name: NPF_IPv4UC_AddResEntryAdd
*   
*    Description: Add an address resolution entry
*       
*        
*    Parameters:
*        IN:  callbackHandle - callback handle
*             correlator - correlator
*             error - error reporting = no CB, only error CB or CB
*             tableHandle - handle to address resolution table
*             entry - address resolution entry to be added
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
extern NPF_error_t NPF_IPv4UC_AddResEntryAdd(
	NPF_IN NPF_callbackHandle_t callbackHandle, 
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting,
	NPF_IN NPF_IPv4UC_AddResTableHandle_t tableHandle,
	NPF_IN NPF_uint32_t numEntries,
	NPF_IN NPF_IPv4UC_AddResEntry_t *entryArray) 
{
    NPF_error_t     retVal;
    IPV4UsrReqInfo_t   *usrReq = NULL;
    IPV4TransFPInfo_t  *transFPInfo = NULL;
    IPV4ReqToFEInfo_t  *reqToFE=NULL;
    FPPI_IPv4UniNHResolEntry_t  *fpAddResBlock = NULL;
    uint32_t           i;
    IPV4_DBG_TRACE();
    CHK_INIT();
    
    /* Basic check for the data that is to be sent to FP */  
    if ( !tableHandle || !numEntries || !entryArray )
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"Invalid parameters");
        retVal = NPF_INVALID_PARAMETERS;
        goto CleanUp;
    }
    /* Basic check for single entry : Beta Only */ 
    if(numEntries > 1)
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
		                    	"More than one entry not supported !");
        RETURN(NPF_INVALID_PARAMETERS);
    }
    
    /* Initialize user request from parameters */ 
    if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
				     IPV4_TABLE_TYPE_ADDRES, IPV4_TRANS_DISC,
	                             NPF_IPV4UC_ADDRESS_RES_ENTRY_ADD,
				     callbackHandle,tableHandle,&usrReq))
		          != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request could not be initialized");
         retVal = NPF_E_UNKNOWN;
         goto CleanUp;
     }
     /* Invoke User Request Handler : Which processes this request. All the
      * data-struc. linkages are done in this routine. */ 
     if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request Handler encountered error ");
         retVal = NPF_E_UNKNOWN;
         goto CleanUp;
     }
     
     /* User request is processed; We need to call FPP API here to send
      * the request */
      
     assert(usrReq->transInfo); /* Safety check */

    /* Allocate memory for the Next-Hop Block to be sent to FP through FPPAPI */ 
     fpAddResBlock = (FPPI_IPv4UniNHResolEntry_t *) PIL_AllocateMemory(
                     (sizeof(FPPI_IPv4UniNHResolEntry_t)*numEntries), PIL_MEM_ZERO_INIT);
     
     if(fpAddResBlock == NULL)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request Handler encountered error ");
         retVal = NPF_E_UNKNOWN;
         goto CleanUp;
     }

     /*Process the L2 entry  to send to FP (NH Table + L2 Table). 
      * This also updates the list of FEs that this table is associated with */
     if(( retVal = ipv4_PrcUserReqToFPForArpAdd(usrReq,numEntries,
				    entryArray,fpAddResBlock))
 		    != NPF_SUCCESS)
     {
         npf_logger_Write(LOG_IPV4, LOG_ERR,
			"Unable to Process User req to FP");
         retVal = NPF_E_UNKNOWN;
         goto CleanUp;
     }
 
     /* We can extract the FP info now */
     transFPInfo = &usrReq->transInfo->transFpInfo; 

     /* Send the request to FP */
     for(i = 0; i < transFPInfo->numFEToSend ; i++)
     {
         reqToFE = transFPInfo->reqFEPtrArray[i];
         if(reqToFE == NULL)
         {
             npf_logger_Write(LOG_IPV4, LOG_ERR, 
			           "Request to FE is NULL");
             retVal = NPF_E_UNKNOWN;
             goto CleanUp;
         }
           npf_logger_Write(LOG_IPV4, LOG_TRACE, 
               "Req to FE [FEId=%d numEntries=%d] sent;",
               reqToFE->feId, numEntries);

         if((retVal = FPPAPI_ipv4_unicastToL2Add(
              reqToFE->feId, 
              reqToFE->fppi_correlator, 
        (FPPI_CBHANDLE)ipv4Mgr.fppInfo.cbInfoArray[IPV4_FP_CB_ADD_L2].cbHndl, 
              numEntries,
	           fpAddResBlock ))
	             != NPF_SUCCESS)
         {
             npf_logger_Write(LOG_IPV4, LOG_ERR, 
			           "Unable to process request to FPP");
             retVal = NPF_E_UNKNOWN;
             goto CleanUp;
         }
         npf_logger_Write(LOG_IPV4, LOG_TRACE, 
             "*** Waiting For callback ***");
     }
    /* We are going to free UsrReq Context in Calllback */ 
    if(fpAddResBlock)
       PIL_FreeMemory(fpAddResBlock); 
     RETURN(NPF_NO_ERROR);

CleanUp:
    if(fpAddResBlock) 
       PIL_FreeMemory(fpAddResBlock); 
    ipv4_DeleteUsrReq(usrReq);
    RETURN(retVal); 
}


/*****************************************************************************
*    Name: NPF_IPv4UC_AddResEntryDelete
*   
*    Description: Delete an address resolution entry
*       
*        
*    Parameters:
*        IN:  callbackHandle - callback handle
*             correlator - correlator
*             error - error reporting = no CB, only error CB or CB
*             tableHandle - handle to address resolution table
*             entry - address resolution entry to be deleted
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
NPF_error_t NPF_IPv4UC_AddResEntryDelete(
	NPF_IN NPF_callbackHandle_t callbackHandle,
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting, 
	NPF_IN NPF_IPv4UC_AddResTableHandle_t tableHandle,
	NPF_IN NPF_uint32_t numEntries,
	NPF_IN NPF_IPv4UC_AddResKey_t *entryArray)
{
    NPF_error_t     retVal;
    IPV4UsrReqInfo_t   *usrReq = NULL;
    IPV4TransFPInfo_t  *transFPInfo = NULL;
    IPV4ReqToFEInfo_t  *reqToFE=NULL;
    FPPI_IPv4UniNHResolEntry_t  *fpAddResBlock = NULL;
    uint32_t           i;
    IPV4_DBG_TRACE();
    CHK_INIT();
    
    /* Basic check for the data that is to be sent to FP */  
    if ( !tableHandle || !numEntries || !entryArray )
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"Invalid parameters");
        retVal = NPF_INVALID_PARAMETERS;
        goto CleanUp;
    }
    /* Basic check for single entry : Beta Only */ 
    if(numEntries > 1)
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
		                    	"More than one entry not supported !");
        RETURN(NPF_INVALID_PARAMETERS);
    }
    
    /* Initialize user request from parameters */ 
    if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
				     IPV4_TABLE_TYPE_ADDRES, IPV4_TRANS_DISC,
	                             NPF_IPV4UC_ADDRESS_RES_ENTRY_ADD,
				     callbackHandle,tableHandle,&usrReq))
		          != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request could not be initialized");
        retVal = NPF_E_UNKNOWN;
        goto CleanUp;
     }
     /* Invoke User Request Handler : Which processes this request. All the
      * data-struc. linkages are done in this routine. */ 
     if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request Handler encountered error ");
        retVal = NPF_E_UNKNOWN;
        goto CleanUp;
     }
     
     assert(usrReq->transInfo); /* Safety check */

    /* Allocate memory for the Next-Hop Block to be sent to FP through FPPAPI */ 
     fpAddResBlock = (FPPI_IPv4UniNHResolEntry_t*) PIL_AllocateMemory(
                     (sizeof(FPPI_IPv4UniNHResolEntry_t)*numEntries), PIL_MEM_ZERO_INIT);
     
     if(fpAddResBlock == NULL)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request Handler encountered error ");
        retVal = NPF_E_UNKNOWN;
        goto CleanUp;
     }

     /*Process the L2 entry  to send to FP (NH Table + L2 Table). 
      * This also updates the list of FEs that this table is associated with */
     if(( retVal = ipv4_PrcUserReqToFPForArpDel(usrReq,numEntries,
				    entryArray,fpAddResBlock))
 		    != NPF_SUCCESS)
     {
         npf_logger_Write(LOG_IPV4, LOG_ERR,
			"Unable to Process User req to FP");
        retVal = NPF_E_UNKNOWN;
        goto CleanUp;
     }
 
          
     /* We can extract the FP info now */
     transFPInfo = &usrReq->transInfo->transFpInfo; 

     /* Send the request to FP */
     for(i = 0; i < transFPInfo->numFEToSend ; i++)
     {
         reqToFE = transFPInfo->reqFEPtrArray[i];
         if(reqToFE == NULL)
         {
             npf_logger_Write(LOG_IPV4, LOG_ERR, 
			           "Request to FE is NULL");
             retVal = NPF_E_UNKNOWN;
             goto CleanUp;
         }
           npf_logger_Write(LOG_IPV4, LOG_TRACE, 
               "Req to FE [FEId=%d numEntries=%d] sent;",
               reqToFE->feId, numEntries);

         if((retVal = FPPAPI_ipv4_unicastToL2Del(
              reqToFE->feId, 
              reqToFE->fppi_correlator, 
        (FPPI_CBHANDLE)ipv4Mgr.fppInfo.cbInfoArray[IPV4_FP_CB_DEL_L2].cbHndl, 
              numEntries,
	           fpAddResBlock ))
	             != NPF_SUCCESS)
         {
             npf_logger_Write(LOG_IPV4, LOG_ERR, 
			           "Unable to process request to FPP");
             retVal = NPF_E_UNKNOWN;
             goto CleanUp;
         }
         npf_logger_Write(LOG_IPV4, LOG_TRACE, 
             "*** Waiting For callback ***");
     }
    /* We are going to free UsrReq Context in Calllback */ 
    if(fpAddResBlock)
       PIL_FreeMemory(fpAddResBlock); 
     RETURN(NPF_NO_ERROR);

CleanUp:
    if(fpAddResBlock) 
       PIL_FreeMemory(fpAddResBlock); 
    ipv4_DeleteUsrReq(usrReq);
    RETURN(retVal); 
}

/*****************************************************************************
*    Name: NPF_IPv4UC_AddResTableFlush
*   
*    Description:
*       
*        
*    Parameters:
*        IN:  callbackHandle - callback handle
*             correlator - correlator
*             error - error reporting = no CB, only error CB or CB
*             tableHandle - handle to ARP table
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/

NPF_error_t NPF_IPv4UC_AddResTableFlush( 
	NPF_IN NPF_callbackHandle_t callbackHandle,
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting,
	NPF_IN NPF_IPv4UC_AddResTableHandle_t tableHandle)
{
    RETURN(NPF_IPV4UC_FUNCTION_NOT_SUPPORTED);
}

/*****************************************************************************
*    Name: NPF_IPv4UC_AddResEntryQuery
*   
*    Description:
*        Query for an attribute for an entry         
*    Parameters:
*        IN:  callbackHandle - callback handle
*             correlator - correlator
*             error - error reporting = no CB, only error CB or CB
*             tableHandle - handle to prefix table
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/


NPF_error_t NPF_IPv4UC_AddResEntryQuery(
	NPF_IN NPF_callbackHandle_t callbackHandle,
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting, 
	NPF_IN NPF_IPv4UC_AddResTableHandle_t tableHandle,
	NPF_IN NPF_uint32_t numEntries,
	NPF_IN NPF_IPv4UC_AddResKey_t *entryArray)
{
    IPV4_DBG_TRACE();
    CHK_INIT();
      
    RETURN(NPF_IPV4UC_FUNCTION_NOT_SUPPORTED);
}

/*****************************************************************************
*    Name: NPF_IPv4UC_FibTableHandleCreate
*   
*    Description:
*       Create a Fib table handle. Necessary to make
*       calls to add/delete prefixes/nexthops
*        
*    Parameters:
*        IN: callbackHandle - callback handle
*            correlator - correlator
*            error - report all/none/errors
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
#ifdef IPV4_UNIF_SUPPORT
NPF_error_t NPF_IPv4UC_FibTableHandleCreate(
	NPF_IN NPF_callbackHandle_t callbackHandle, 
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting
	)
{
    NPF_error_t        retVal;
    IPV4UsrReqInfo_t   *usrReq   = NULL;
    IPV4UserInfo_t     *userInfo = NULL;

    IPV4_DBG_TRACE();
    CHK_INIT();
    
     errorReporting = NPF_REPORT_ALL;  /* Overide the original value*/
     retVal = NPF_SUCCESS;
     
     /* Initialize user request from parameters.  */ 
     if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
				     IPV4_TABLE_TYPE_FIB, IPV4_TRANS_UNIF,
				     NPF_IPV4UC_FIB_TABLE_HANDLE_CREATE,
				     callbackHandle,0,&usrReq))
		          != NPF_SUCCESS)

     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request could not be initialized");
	     goto InvokeCB;
     }

     /* Initialise the user so that he can use only Unified APIs */
     if ((retVal = ipv4_GetUserByHandleFrmMgr(usrReq->callbackHandle, &userInfo)) != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR, "Unable to Get the User Info");
        goto InvokeCB;
     }

     userInfo->usrType = IPV4_USR_UNIF;
     /* Invoke User Request Handler : Which processes this request */ 
     if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request Handler encountered error ");
	     goto InvokeCB;
     }
    
     RETURN(NPF_SUCCESS); 
     /* User request is processed; We don't need to call FPP APIs in this
      * case. Just return Here */
InvokeCB:
    /* Inovoke callback to the User */
    ipv4_InvokeAppCallback(usrReq, NULL,
                      NPF_IPV4UC_FIB_TABLE_HANDLE_CREATE,retVal);

    /* Callback is given to User. Now, cleanup the transaction memory */ 
    if(usrReq != NULL && usrReq->transInfo != NULL)
       ipv4_DeleteTransInfo(usrReq->transInfo);

    RETURN(NPF_NO_ERROR);
}

/*****************************************************************************
*    Name: NPF_IPv4UC_FibTableHandleDelete
*  
*    Description:
*        Delete the FIB table handle
*
*    Parameters:
*        IN: callbackHandle - callback handle
*            correlator - correlator
*            error - report all/none/errors
*            tableHandle - FIB table handle to be deleted
*
*    Preconditions:
*
*
*    Postconditions:
*
*
*    Returns:
*        Standard error value
*
*/
NPF_error_t NPF_IPv4UC_FibTableHandleDelete(
	NPF_IN NPF_callbackHandle_t callbackHandle, 
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting,
   NPF_IN NPF_IPv4UC_FibTableHandle_t tableHandle
	)
{
    NPF_error_t        retVal;
    IPV4UsrReqInfo_t   *usrReq = NULL;
    IPV4_DBG_TRACE();
    CHK_INIT();
    
     errorReporting = NPF_REPORT_ALL;  /* Overide the original value*/
     retVal = NPF_SUCCESS;
     
     /* Initialize user request from parameters.  */ 
     if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
				     IPV4_TABLE_TYPE_FIB, IPV4_TRANS_UNIF,
				     NPF_IPV4UC_FIB_TABLE_HANDLE_DELETE,
				     callbackHandle, tableHandle,&usrReq))
		          != NPF_SUCCESS)

     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request could not be initialized");
	     goto InvokeCB;
     }

     /* Invoke User Request Handler : Which processes this request */ 
     if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request Handler encountered error ");
	     goto InvokeCB;
     }
    
     RETURN(NPF_SUCCESS); 
     /* User request is processed; We don't need to call FPP APIs in this
      * case. Just return Here */
InvokeCB:
    /* Inovoke callback to the User */
    ipv4_InvokeAppCallback(usrReq, NULL,
                      NPF_IPV4UC_FIB_TABLE_HANDLE_DELETE,retVal);

    /* Callback is given to User. Now, cleanup the transaction memory */ 
    if(usrReq != NULL && usrReq->transInfo != NULL)
       ipv4_DeleteTransInfo(usrReq->transInfo);

    RETURN(NPF_NO_ERROR);
}

/*****************************************************************************
*    Name: NPF_IPv4UC_PrefixTableFlush
*   
*    Description:
*        Flush all the Prefix table entries
*        
*    Parameters:
*        IN:  callbackHandle - callback handle
*             correlator - correlator
*             error - error reporting = no CB, only error CB or CB
*             tableHandle - handle to prefix table
*        
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
NPF_error_t NPF_IPv4UC_FibTableFlush(
	NPF_IN NPF_callbackHandle_t callbackHandle,
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting, 
	NPF_IN NPF_IPv4UC_FibTableHandle_t tableHandle)
{
    NPF_error_t     retVal;
    IPV4UsrReqInfo_t   *usrReq = NULL;

    IPV4_DBG_TRACE();
    CHK_INIT();
    
    retVal = NPF_SUCCESS;
    
    /* This is true only for Beta */ 
    npf_logger_Write(LOG_IPV4, LOG_ERR,
		     	"Fib Table Flush Not Supported");
    RETURN(NPF_IPV4UC_FUNCTION_NOT_SUPPORTED);
    
    /* Basic check for the data that is to be sent to FP */  
    if ( !tableHandle)
    {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"Invalid parameters");
        retVal = NPF_INVALID_PARAMETERS;
        RETURN(NPF_E_UNKNOWN);
    }
 
    /* Initialize user request from parameters */ 
    if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
				     IPV4_TABLE_TYPE_FIB, IPV4_TRANS_UNIF,
	                             NPF_IPV4UC_FIB_TABLE_FLUSH,
				     callbackHandle,tableHandle,&usrReq))
		          != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request could not be initialized");
        if(usrReq != NULL && usrReq->transInfo != NULL)
           ipv4_DeleteTransInfo(usrReq->transInfo);
        RETURN(NPF_E_UNKNOWN);
     }
     
     /* Invoke User Request Handler : Which processes this request. All the
      * data-struc. linkages are done in this routine. */ 
     if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
     {
        npf_logger_Write(LOG_IPV4, LOG_ERR,
			"User Request Handler encountered error ");
        if(usrReq != NULL && usrReq->transInfo != NULL)
           ipv4_DeleteTransInfo(usrReq->transInfo);
        RETURN(NPF_E_UNKNOWN);
     }
     
     /* User request is processed; We need to call FPP API here to send
      * the request */
      
     assert(usrReq->transInfo); /* Safety check */

     RETURN(NPF_NO_ERROR);
}

/*****************************************************************************
*    Name: NPF_IPv4UC_FibEntryAdd
*   
*    Description:
*        Add the Fib table entries as denoted by the array
*        
*    Parameters:
*        IN:  callbackHandle - callback handle
*             correlator - correlator
*             error - error reporting = no CB, only error CB or CB
*             tableHandle - handle to prefix table
*             numEntries - number of entries to be added
*             prefixArray - array of prefixes to be added       
*             nextHopArray - array of next hops to be added       
*       
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
extern NPF_error_t NPF_IPv4UC_FibEntryAdd( 
	NPF_IN NPF_callbackHandle_t callbackHandle,
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting,
	NPF_IN NPF_IPv4UC_PrefixTableHandle_t tableHandle,
	NPF_IN NPF_uint32_t numEntries, 
	NPF_IN NPF_IPv4UC_Prefix_t *prefixArray,
	NPF_IN NPF_IPv4UC_NextHopArray_t *nextHopArray)
{
   NPF_error_t         retVal;
   uint32_t            i;
   NPF_uint32_t*       generatedNHId;
   IPV4UsrReqInfo_t   *usrReq = NULL;

   IPV4_DBG_TRACE();
   CHK_INIT();

   retVal = NPF_SUCCESS;

   /* Basic check for the data that is to be sent to FP */  
   if ( !tableHandle || !numEntries || !prefixArray || !nextHopArray )
   {
      npf_logger_Write(LOG_IPV4, LOG_ERR,
            "Invalid parameters");
      retVal = NPF_INVALID_PARAMETERS;
      RETURN(NPF_E_UNKNOWN);
   }
   
   /* Basic check for single entry : Beta Only */ 
   if(numEntries > 1)
   {
       npf_logger_Write(LOG_IPV4, LOG_ERR,
		                    	"More than one entry not supported !");
       RETURN(NPF_INVALID_PARAMETERS);
   }

   /* Initialize user request from parameters */ 
   if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
               IPV4_TABLE_TYPE_FIB, IPV4_TRANS_UNIF,
               NPF_IPV4UC_FIB_ENTRY_ADD,
               callbackHandle,tableHandle,&usrReq))
         != NPF_SUCCESS)
   {
      npf_logger_Write(LOG_IPV4, LOG_ERR,
            "User Request could not be initialized");
      if(usrReq != NULL && usrReq->transInfo != NULL)
         ipv4_DeleteTransInfo(usrReq->transInfo);
      RETURN(NPF_E_UNKNOWN);
   }

   /* Invoke User Request Handler : Which processes this request */ 
   if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
   {
      npf_logger_Write(LOG_IPV4, LOG_ERR,
            "User Request Handler encountered error ");
      if(usrReq != NULL && usrReq->transInfo != NULL)
         ipv4_DeleteTransInfo(usrReq->transInfo);
      RETURN(NPF_E_UNKNOWN);
   }

   generatedNHId = (NPF_uint32_t *)PIL_AllocateMemory(sizeof(NPF_uint32_t)*numEntries, PIL_MEM_ZERO_INIT);
   if (!generatedNHId)
   {
      npf_logger_Write(LOG_IPV4, LOG_ERR,
         "Could not allocate memory ");
      RETURN(NPF_E_UNKNOWN);
   }
   for (i = 0; i < numEntries; i++)
      ipv4_GetNhIdFrmUsr(usrReq->transInfo->usrInfo, &generatedNHId[i]);

   if(( retVal = ipv4_AddNhopCntxtToUsrReq(usrReq, numEntries,nextHopArray,
       generatedNHId)) != NPF_SUCCESS)
   {
      npf_logger_Write(LOG_IPV4, LOG_ERR,
         "Next-Hop Context couldn't be stored ");
      if(usrReq != NULL && usrReq->transInfo != NULL)
         ipv4_DeleteTransInfo(usrReq->transInfo);
      PIL_FreeMemory(generatedNHId);
      RETURN(NPF_E_UNKNOWN);
   }
   if(( retVal = ipv4_AddPrefixCntxtToUsrReq(usrReq, numEntries,
       prefixArray)) != NPF_SUCCESS)
   {
      npf_logger_Write(LOG_IPV4, LOG_ERR,
         "Prefix Context couldn't be stored ");
      if(usrReq != NULL && usrReq->transInfo != NULL)
         ipv4_DeleteTransInfo(usrReq->transInfo);
      PIL_FreeMemory(generatedNHId);
      RETURN(NPF_E_UNKNOWN);
   }
   PIL_FreeMemory(generatedNHId);
   usrReq->transInfo->mapTransInfo->unifReqState = UNIF_ADD_QUERY_PROGRESS; 

   if ((retVal = NPF_IPv4UC_PrefixEntryQuery (callbackHandle, 
           usrReq->transInfo->int_correlator, 
           errorReporting, usrReq->transInfo->usrInfo->tableArray[IPV4_TABLE_TYPE_PREFIX].intHandle, 
           numEntries, prefixArray)) != NPF_SUCCESS)
   {
      npf_logger_Write(LOG_IPV4, LOG_ERR,
         "Error in prefix query");
      if(usrReq != NULL && usrReq->transInfo != NULL)
         ipv4_DeleteTransInfo(usrReq->transInfo);
      RETURN(NPF_E_UNKNOWN);
   }
   RETURN(NPF_NO_ERROR); 
}

/*****************************************************************************
*    Name: NPF_IPv4UC_FibEntryDelete
*   
*    Description:
*        Delete the Fib table entries as denoted by the array
*        
*    Parameters:
*        IN:  callbackHandle - callback handle
*             correlator - correlator
*             error - error reporting = no CB, only error CB or CB
*             tableHandle - handle to prefix table
*             numEntries - number of entries to be added
*             prefixArray - array of prefixes to be added       
*       
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
extern NPF_error_t NPF_IPv4UC_FibEntryDelete( 
	NPF_IN NPF_callbackHandle_t callbackHandle,
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting,
	NPF_IN NPF_IPv4UC_PrefixTableHandle_t tableHandle,
	NPF_IN NPF_uint32_t numEntries, 
	NPF_IN NPF_IPv4UC_Prefix_t *prefixArray)
{
   NPF_error_t         retVal;
   IPV4UsrReqInfo_t   *usrReq = NULL;

   IPV4_DBG_TRACE();
   CHK_INIT();

   retVal = NPF_SUCCESS;

   /* Basic check for the data that is to be sent to FP */  
   if ( !tableHandle || !numEntries || !prefixArray )
   {
      npf_logger_Write(LOG_IPV4, LOG_ERR,
            "Invalid parameters");
      retVal = NPF_INVALID_PARAMETERS;
      RETURN(NPF_E_UNKNOWN);
   }
   /* Basic check for single entry : Beta Only */ 
   if(numEntries > 1)
   {
       npf_logger_Write(LOG_IPV4, LOG_ERR,
		                    	"More than one entry not supported !");
       RETURN(NPF_INVALID_PARAMETERS);
   }

   /* Initialize user request from parameters */ 
   if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
               IPV4_TABLE_TYPE_FIB, IPV4_TRANS_UNIF,
               NPF_IPV4UC_FIB_ENTRY_DELETE,
               callbackHandle,tableHandle,&usrReq))
         != NPF_SUCCESS)
   {
      npf_logger_Write(LOG_IPV4, LOG_ERR,
            "User Request could not be initialized");
      if(usrReq != NULL && usrReq->transInfo != NULL)
         ipv4_DeleteTransInfo(usrReq->transInfo);
      RETURN(NPF_E_UNKNOWN);
   }

   /* Invoke User Request Handler : Which processes this request */ 
   if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
   {
      npf_logger_Write(LOG_IPV4, LOG_ERR,
            "User Request Handler encountered error ");
      if(usrReq != NULL && usrReq->transInfo != NULL)
         ipv4_DeleteTransInfo(usrReq->transInfo);
      RETURN(NPF_E_UNKNOWN);
   }

   if(( retVal = ipv4_AddPrefixCntxtToUsrReq(usrReq, numEntries,
       prefixArray)) != NPF_SUCCESS)
   {
      npf_logger_Write(LOG_IPV4, LOG_ERR,
         "Prefix Context couldn't be stored ");
      if(usrReq != NULL && usrReq->transInfo != NULL)
         ipv4_DeleteTransInfo(usrReq->transInfo);
      RETURN(NPF_E_UNKNOWN);
   }

   usrReq->transInfo->mapTransInfo->unifReqState = UNIF_DEL_QUERY_PROGRESS; 

   if ((retVal = NPF_IPv4UC_PrefixEntryQuery (callbackHandle, 
           usrReq->transInfo->int_correlator, 
           errorReporting, 
           usrReq->transInfo->usrInfo->tableArray[IPV4_TABLE_TYPE_PREFIX].intHandle, 
           numEntries, prefixArray)) != NPF_SUCCESS)
   {
      npf_logger_Write(LOG_IPV4, LOG_ERR,
         "Error in prefix query");
      if(usrReq != NULL && usrReq->transInfo != NULL)
         ipv4_DeleteTransInfo(usrReq->transInfo);
      RETURN(NPF_E_UNKNOWN);
   }
   RETURN(NPF_NO_ERROR); 
}

/*****************************************************************************
*    Name: NPF_IPv4UC_FibEntryQuery
*   
*    Description:
*        Delete the Fib table entries as denoted by the array
*        
*    Parameters:
*        IN:  callbackHandle - callback handle
*             correlator - correlator
*             error - error reporting = no CB, only error CB or CB
*             tableHandle - handle to prefix table
*             numEntries - number of entries to be added
*             prefixArray - array of prefixes to be added       
*       
*    Preconditions:
*        
*    
*    Postconditions:
*        
*    
*    Returns:
*        Standard error value
*    
*/
extern NPF_error_t NPF_IPv4UC_FibEntryQuery( 
	NPF_IN NPF_callbackHandle_t callbackHandle,
	NPF_IN NPF_correlator_t correlator,
	NPF_IN NPF_errorReporting_t errorReporting,
	NPF_IN NPF_IPv4UC_PrefixTableHandle_t tableHandle,
	NPF_IN NPF_uint32_t numEntries, 
	NPF_IN NPF_IPv4UC_Prefix_t *prefixArray)
{
   NPF_error_t         retVal;
   IPV4UsrReqInfo_t   *usrReq = NULL;

   IPV4_DBG_TRACE();
   CHK_INIT();

   retVal = NPF_SUCCESS;

   /* Basic check for the data that is to be sent to FP */  
   if ( !tableHandle || !numEntries || !prefixArray )
   {
      npf_logger_Write(LOG_IPV4, LOG_ERR,
            "Invalid parameters");
      retVal = NPF_INVALID_PARAMETERS;
      RETURN(NPF_E_UNKNOWN);
   }
   /* Basic check for single entry : Beta Only */ 
   if(numEntries > 1)
   {
       npf_logger_Write(LOG_IPV4, LOG_ERR,
		                    	"More than one entry not supported !");
       RETURN(NPF_INVALID_PARAMETERS);
   }

   /* Initialize user request from parameters */ 
   if((retVal= ipv4_InitReqFrmUsr(correlator,errorReporting,
               IPV4_TABLE_TYPE_FIB, IPV4_TRANS_UNIF,
               NPF_IPV4UC_FIB_ENTRY_QUERY,
               callbackHandle,tableHandle,&usrReq))
         != NPF_SUCCESS)
   {
      npf_logger_Write(LOG_IPV4, LOG_ERR,
            "User Request could not be initialized");
      if(usrReq != NULL && usrReq->transInfo != NULL)
         ipv4_DeleteTransInfo(usrReq->transInfo);
      RETURN(NPF_E_UNKNOWN);
   }

   /* Invoke User Request Handler : Which processes this request */ 
   if ((retVal = ipv4_InvokeUserReqHandler(usrReq)) != NPF_SUCCESS)
   {
      npf_logger_Write(LOG_IPV4, LOG_ERR,
            "User Request Handler encountered error ");
      if(usrReq != NULL && usrReq->transInfo != NULL)
         ipv4_DeleteTransInfo(usrReq->transInfo);
      RETURN(NPF_E_UNKNOWN);
   }

   if(( retVal = ipv4_AddPrefixCntxtToUsrReq(usrReq, numEntries,
       prefixArray)) != NPF_SUCCESS)
   {
      npf_logger_Write(LOG_IPV4, LOG_ERR,
         "Prefix Context couldn't be stored ");
      if(usrReq != NULL && usrReq->transInfo != NULL)
         ipv4_DeleteTransInfo(usrReq->transInfo);
      RETURN(NPF_E_UNKNOWN);
   }

   usrReq->transInfo->mapTransInfo->unifReqState = UNIF_FIB_QUERY_PREFIX_PROGRESS; 

   if ((retVal = NPF_IPv4UC_PrefixEntryQuery (callbackHandle, 
           usrReq->transInfo->int_correlator, 
           errorReporting, 
           usrReq->transInfo->usrInfo->tableArray[IPV4_TABLE_TYPE_PREFIX].intHandle, 
           numEntries, prefixArray)) != NPF_SUCCESS)
   {
      npf_logger_Write(LOG_IPV4, LOG_ERR,
         "Error in prefix query");
      if(usrReq != NULL && usrReq->transInfo != NULL)
         ipv4_DeleteTransInfo(usrReq->transInfo);
      RETURN(NPF_E_UNKNOWN);
   }
   RETURN(NPF_NO_ERROR); 
}
#endif /* D_IPV4_UNIF_SUPPORT */

