//////////////////////////////////////////////////////////////////////////////
//           INTEL CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or
//  nondisclosure agreement with Intel Corporation and may not be copied
//  or disclosed except in accordance with the terms of that agreement.
//      Copyright (c) 1999-2000 Intel Corporation.  All Rights Reserved.
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////

/********************************************************************************
 * $Source: /cvs/constellations/ixa_3.0/src/cp_pdk/ControlPlane/CoreBlocks/BindingDiscovery/src/bind_cap_disc.c,v $
 * $Revision: 1.6 $
 * $Date: 2003/11/17 16:57:11 $
 * 
 * bind_cap_disc.c
 * ---------------
 * This is the Binding and Capability Discovery module. It registers callbacks
 * with the FP Plug in API to receive notifications when a FE binds and reports
 * its (and its ports') capabilities. It then interacts with the NPF Namespace
 * module and populates it with the new nodes corresponding to the new FE and its
 * ports. It also interacts with the Config & Mgmt (CM) module to populate the 
 * FE & port attributes and capabilities of the new FE and port objects.
 *
 * Assumptions:
 *   1. It is initialized by the PDK manager
 *   2. This must be initialized AFTER FP Plug in API and BEFORE CM module.
 *   3. In the Namespace, there is only one instance of the System. That is,
 *      /System/0 is the root node for the entire instance of the Namespace
 *******************************************************************************/


#include <assert.h>
#include <stdio.h>
#include <string.h>
#include "stdlib.h"
#include "pil.h"
#include "pdk_callback.h"
#include "log.h"
#include "npf_ns_api.h"
#include "fppCapabilityTypes.h"
#include "pdk_cm.h"
#include "pdk_cm_types.h"
#include "pdk_bd.h"
#include "pdk_ns_helper.h"
#include "pdk_macros.h"
#include "hwaddr.h"

/* 
   Globals for Binding & Discovery
*/

static uint32_t      g_bd_initialized = FALSE;
static FPPI_CONTEXT  g_feBind_Event_context;
static FPPI_CONTEXT  g_fePortCap_Event_context;
static FPPI_CBHANDLE g_feBind_Event_handle; 
static FPPI_CBHANDLE g_fePortCap_Event_handle;
static char          g_sys_path[NPF_NS_MAXPATH]; /* Root Namespace path */
static int           g_fe_instance = 0; /* for tracking FE instances */

/* Macros for debugging */

#ifdef BD_DEBUG
#define PDBG_PRINT(fmt, args...) printf(fmt,## args)
#define PDBG_TRACE()             DBG_TRACE () /* macro */
#define PRETURN(val)             RETURN(val) /* macro */
#else
#define PDBG_PRINT(fmt, args...)  /* nothing */
#define PDBG_TRACE()              /* nothing */
#define PRETURN(val)              return val /* return normally */
#endif

static int bd_checkFE (FPPI_FEID feid, void* response, NPF_HANDLE* fe);
void       on_FEBind_Event( FPPI_FEID fe_id, FPPI_CORRELATOR cb_correlator,
                FPPI_CONTEXT cb_context, void* response_data );
void       on_FEPortCap_Event( FPPI_FEID fe_id, FPPI_CORRELATOR cb_correlator,
                   FPPI_CONTEXT cb_context, void* response_data );

/****************************************************************************
 * Function: bd_init()                                                      *
 * -------------------                                                      *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: Invoked by PDK manager during the initialization of the     *
 *              Control plane of the PDK.                                   *
 * Assumption:  Assumes that FPP module has already been initialized - this *
 *              is required for registering different callbacks             *
 *                                                                          *
 * Return: 0 on success, error values from registration functions           *
 *                                                                          *
 ****************************************************************************/

NPF_RET bd_init()
{
    int err;

    PDBG_TRACE();

    /* Contexts will be the event we are registering for. 
       This way we can easily check if the FPP callback has the correct 
       context
    */
    g_feBind_Event_context = (void*) FPPI_FEBind_Event;
    g_fePortCap_Event_context = (void*) FPPI_FECapability_Event;
    
    /* Register FE Bind event callback with FPP API */
    err = FPPAPI_register_callback(&on_FEBind_Event,
                   &g_feBind_Event_handle,
                   FPPI_FEBind_Event,
                   g_feBind_Event_context);

    /* Register Port Capability callback with FPP API */
    err = FPPAPI_register_callback(&on_FEPortCap_Event,
                   &g_fePortCap_Event_handle,
                   FPPI_FECapability_Event,
                   g_fePortCap_Event_context);

    /* initialize the Namespace path string */
    memset(g_sys_path, 0, NPF_NS_MAXPATH);

    g_bd_initialized = TRUE;

    PRETURN (err);
}

/****************************************************************************
 * Function: bd_shutdown()                                                  *
 * -----------------------                                                  *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: Invoked by PDK manager during the shutdown of the Control   *
 *              plane of the PDK.                                           *
 * Assumption:  Assumes that FPP module has not yet been shutdown - this    *
 *              is required for deregistering different callbacks           *
 *                                                                          *
 * Return: 0 on success, error values from deregistration functions         *
 *                                                                          *
 ****************************************************************************/

NPF_RET bd_shutdown()
{
    int err;

    PDBG_TRACE();

    /* De-register FE Bind event callback with FPP API */
    err = FPPAPI_deregister_callback( g_feBind_Event_handle );

    /* De-register Port Capability callback with FPP API */
    err = FPPAPI_deregister_callback( g_fePortCap_Event_handle );

    g_bd_initialized = FALSE;

    PRETURN(err);
}

/****************************************************************************
 * Function: bd_feBind_register_callback()                                  *
 * ---------------------------------------                                  *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: This function MUST be invoked ONLY by internal modules of   *
 *              the PDK to register for FE bind events. External apps MUST  *
 *              use the CM Event Registration API.                          *
 *                                                                          *
 * Return: 0 on success, error values from internal registration fns        *
 *                                                                          *
 ****************************************************************************/

NPF_RET bd_feBind_register_callback( 
                   /* IN: */ NPF_USERCONTEXT context, 
               /* IN: */ BIND_CALLBACK callback,
               /* OUT:*/ NPF_CBHANDLE* cb_handle )
{
    /* 
       invoke the PDK callback library routine to store the callback
       this set of routines maintains any kind of callback function
       it requires a context - we will use the event type as the
       context
    */
    PDBG_TRACE();
    PRETURN (pdk_callback_register( PDK_BD_EVT_FEBIND, 
                   (NPF_USERCONTEXT) context,
                   (NPF_GENERAL_CBFUNC) callback,
                   cb_handle));
}

/****************************************************************************
 * Function: bd_feBind_deregister_callback()                                *
 * ---------------------------------------                                  *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: This function MUST be invoked ONLY by internal modules of   *
 *              the PDK to deregister for FE bind events.                   *
 *                                                                          *
 * Return: 0 on success, error values from internal deregistration fns      *
 *                                                                          *
 ****************************************************************************/


NPF_RET bd_feBind_deregister_callback( NPF_CBHANDLE cb_handle )
{
    /* 
       invoke the PDK callback library routine to deregister the
       callback function
    */
    PDBG_TRACE();
    PRETURN (pdk_callback_deregister( cb_handle ));
}

/****************************************************************************
 * Function: on_FEBind_Event()                                              *
 * ---------------------------                                              *
 * Parameters: FPPI_FEID - ID of the FE that sent a BIND request            *
 *             FPPI_CORRELATOR - ignore this, not required                  *
 *             FPPI_CONTEXT - context in which callback was registered      *
 *             void* - void parameter to specific response data             *
 *                                                                          *
 * Description: This fn is invoked when a FE binds. No capabilities are     *
 *              reported now. FE and individual port capabilities are       *
 *              reported in the FEPortCap event.                            *
 *              This function has to respond with a BIND_REPLY to the FE    *
 *              that just sent a BIND request.                              *
 * Return: none                                                             *   
 *                                                                          *
 ****************************************************************************/

void on_FEBind_Event( FPPI_FEID fe_id, FPPI_CORRELATOR cb_correlator,
              FPPI_CONTEXT cb_context, void* response_data )
{
    /* ignore the correlator, since this is not an API call
       check the context; if this is not the same as the context we 
       registered the callback in, simply return. 
    */

    // Warning: this function is called from FPibNotify::Report, which should not happen!
    PDBG_TRACE();
    if( cb_context != g_feBind_Event_context )
    PRETURN (/*void*/);

    PDBG_PRINT ("In B&D: FE ID = %d \n", fe_id);

    /* send a BIND REPLY to the FE. Nothing else needs to be done now. The FE
       will subsequently send its capabilities.
    */
    FPPAPI_feBind_Reply(fe_id, FPPI_SUCCESSFUL);

    PRETURN (/*void*/);
}

/****************************************************************************
 * Function: on_FEPortCap_Event()                                           *
 * ------------------------------                                           *
 * Parameters: FPPI_FEID - ID of the FE that is sending its capabilities    *
 *             FPPI_CORRELATOR - ignore this, not required                  *
 *             FPPI_CONTEXT - context in which callback was registered      *
 *             void* - void parameter to specific response data             *
 *                                                                          *
 * Description: This fn is invoked when a FE reports its capabilities       *
 *              This module MUST be the first module that handles this event*
 *              This function creates FE and its individual PORT nodes in   *
 *              the Namespace, and initializes their respective attributes  *
 *              learnt from the Bind event. Only after this is successfully *
 *              completed should other modules within the PDK (and external *
 *              applications) know about the new FE that has bound to the CE*
 *                                                                          *
 * Return: none                                                             *   
 *                                                                          *
 ****************************************************************************/

void on_FEPortCap_Event( FPPI_FEID fe_id, FPPI_CORRELATOR cb_correlator,
             FPPI_CONTEXT cb_context, void* response_data )
{
    /**********************************************************
     *                 Local variables                        *
     **********************************************************/

    /*
      for creating FE and PORT nodes in the Namespace
       and initializing their data in CM
    */
    int                      npf_ret, i;
    char                     fe_path[NPF_NS_MAXPATH], port_path[NPF_NS_MAXPATH];
    int                      num_ports = 0, prev_fe_ports = 0, check_fe = 0;
    NPF_NSHDL                ns_instance_handle, ns_type_handle, port_ns_handle;
    NPF_HANDLE               fe_npf_handle, port_npf_handle;
    PDK_FE_ATTRIBS           pdk_fe_attribs;
    PDK_FE_CAPS              pdk_fe_caps;
    PDK_PORT_ATTRIBS         pdk_port_attribs;
    PDK_PORT_CAPS            pdk_port_caps;
    FPPI_FE_Caps*            fe_caps;
    FPPI_Port_Caps*          port_caps;
    struct sockaddr          mac_sock;

    int                      fe_instance;

    /* for creating the Root Path for all FEs */
    NPF_NSHDL                sys_hdl;
    NODE_ITERATOR            port_itr;

    /* for keeping track of the port instances */
    int                port_instance = 0; 
    
     /* for invoking registered callbacks */
    NPF_CBHANDLE       cb_hdl, cb_hdl_next;
    NPF_USERCONTEXT    context;
    BIND_CALLBACK      client_cb_func;

    /**********************************************************
     *                 Function                               *
     **********************************************************/

    /* ignore the correlator, since this is not an API call
       check the context; if this is not the same as the context we 
       registered the callback in, simply return. */
    PDBG_TRACE();
    if( cb_context != g_fePortCap_Event_context )
    PRETURN (/*void*/);

    PDBG_PRINT ("In B&D cap event: FE ID = %d \n", fe_id);

    /* extract the FE attributes and capabilities */
    fe_caps = (FPPI_FE_Caps*) response_data;
    
    /* Modify the fename, first occurence digit of input fename, myFE01->1 */
    if (strpbrk(fe_caps->fe_name, "0123456789") == NULL) {
        PDBG_PRINT("FE Name must contain digits 0 ~ 99\n");
        PRETURN (/*void*/);
    }
    strcpy(fe_caps->fe_name, strpbrk(fe_caps->fe_name, "0123456789"));

    /* this is to cut the alphabet after the digit */
    sprintf(fe_caps->fe_name, "%d", atoi(fe_caps->fe_name));
    /* FE path is named by the fename, which is a input number */
    fe_instance = atoi(fe_caps->fe_name);
    
    /* Get the Root node for all FEs - this will be the path of the 
       first instance node of TYPE SYSTEM in the NS hierarchy

       Do this only the first time. Subsequent calls to this function will
       have the path set. */
    if(strlen(g_sys_path) == 0)
    {
        npf_ret = pdk_ns_getNSRootHandle (&sys_hdl);
        if (NPF_SUCCESS != npf_ret) {
            /* in case of the colocated */
            PDBG_PRINT ("Unable to get NS root - Creating now\n");
            npf_ns_create("/NPFSys", NPF_NS_TYPENODE, NPF_SYSTEM, 0, &sys_hdl);
            npf_ns_close(sys_hdl);
            npf_ns_create("/NPFSys/0", NPF_NS_INSTNODE, NPF_SYSTEM, 0, &sys_hdl);
        }
        npf_ret = npf_ns_getPath(sys_hdl, g_sys_path);
        if(npf_ret != NPF_SUCCESS) {
            npf_ns_close(sys_hdl);
            PDBG_PRINT ("Unable to get System path!\n");
            PRETURN (/*void*/);
        }
        /* close the NS handle of the node */
        npf_ns_close(sys_hdl);
        strcat(g_sys_path,"/");
        PDBG_PRINT ("System path = %s\n", g_sys_path);
    } /* end (get system path)  */
    
    /*
      1. create a node in the NPF namespace for the new FE
      2. Get a NPF_HANDLE to the FE object created
      3. Set the FE properties learnt from the BIND event
    */
    /* initialize the Namespace path string */
    memset(fe_path, 0, NPF_NS_MAXPATH);
    strcpy(fe_path, g_sys_path);
    
    /* first create the TYPE node for the FE object in the Namespace */
    strcat(fe_path, "FE");
    /* check if FE type node exists */
    npf_ret = npf_ns_open(fe_path, &ns_type_handle);
    if(npf_ret != NPF_SUCCESS) {
    npf_ret = npf_ns_create(fe_path, NPF_NS_TYPENODE,
                NPF_FE, 0, &ns_type_handle);
    if(npf_ret != NPF_SUCCESS) {
        // do error handling here
        PDBG_PRINT ("Error creating FE type node\n");
        PRETURN (/*void*/);
    }
    PDBG_PRINT ("new FE type node handle = %d\n", 
            ns_type_handle);
    }
    npf_ns_close(ns_type_handle);
    
    /* 
       check if this FE had bound earlier. In that case, the NS node will
       still be present in an inactive state. This just needs to be activated
       the fe_npf_handle is returned w/o open status, so don't need close it
       but better fix it to open status in bd_checkFE and close it by caller
    */
    check_fe = bd_checkFE (fe_id, response_data, &fe_npf_handle);
    PDBG_PRINT  ("Check for FE returned = %d\n", check_fe);

    if (check_fe == -1) {
        PDBG_PRINT ("Error duplicated FE name\n");
        PRETURN (/*void*/);
    }

    if (check_fe == 1)
        PDBG_PRINT ("FE NPF Handle = %d\n", fe_npf_handle);
    if (check_fe != 1)
    {
        /* next create the INSTANCE node if it is a new FE */
        memset(fe_path, 0, NPF_NS_MAXPATH);
        sprintf(fe_path, "%sFE/%d", g_sys_path, fe_instance);
        
        npf_ret = npf_ns_create(fe_path, NPF_NS_INSTNODE,
                    NPF_FE, 0, &ns_instance_handle);
        if(npf_ret != NPF_SUCCESS) {
            PDBG_PRINT ("Error creating new FE node: %d \n", npf_ret );
            PRETURN (/*void*/);
        }
        
        PDBG_PRINT ("new FE instance handle = %d\n", ns_instance_handle);
        /* now get a handle to the FE object created */
        npf_ret = npf_ns_getDataHandle( ns_instance_handle, &fe_npf_handle);
        if(npf_ret != NPF_SUCCESS) {
            PDBG_PRINT ("unable to get FE instance hdl\n");
            npf_ns_delete(fe_path);
            npf_ns_close(ns_instance_handle);
            PRETURN (/*void*/);
        }
        PDBG_PRINT ("new FE NPF handle = %d\n", fe_npf_handle);
    } /* end if (FE already existed from previous bind) */
    
    /* if this is a FE from a previous bind, we already have the FE's 
       NPF_HANDLE */
    memset(&pdk_fe_attribs, 0, sizeof(PDK_FE_ATTRIBS));
    
    strcpy(pdk_fe_attribs.name, fe_caps->fe_name);
    pdk_fe_attribs.id = fe_id;
    pdk_fe_attribs.ipaddr.s_addr = fe_caps->fe_addr;
    pdk_fe_attribs.num_ports = fe_caps->ports_caps.port_count;
    pdk_fe_attribs.status = NPF_FE_UP;

    /* FIX hack for putting the caps into attribs since no npf_fe_get_caps */
    pdk_fe_attribs.max_arp_table = 1;
    pdk_fe_attribs.max_route_table = 1;
    pdk_fe_attribs.max_prefix_table = 1;
    pdk_fe_attribs.max_nexthop_table = 1;
    pdk_fe_attribs.max_ftn_table = 1;
    pdk_fe_attribs.max_ilm_table = 1;
    /* above two lines should be put to pdk_fe_caps in later RELEASE */

    PDBG_PRINT ("In B&D: FE NAME = %s \n NUM PORTS = %d\n", 
        fe_caps->fe_name,
        fe_caps->ports_caps.port_count);

    /* set the FE properties learnt from the BIND event */
    cm_fe_set_attribs(fe_npf_handle, &pdk_fe_attribs);

    /* CAPABILITIES */

    /* not report from FE so just make it 1 for now */
    pdk_fe_caps.max_arp_table = 1;
    pdk_fe_caps.max_route_table = 1;
    pdk_fe_caps.max_prefix_table = 1;
    pdk_fe_caps.max_nexthop_table = 1;
    pdk_fe_caps.max_ftn_table = 1;
    pdk_fe_caps.max_ilm_table = 1;

    /* support for IPv4, IPv6 or BOTH routing table */
    fe_caps->caps_bitmap = 1;

    switch (fe_caps->caps_bitmap) {
    case 1:
        /* IPv4 - bit 1 */
        pdk_fe_caps.route_table_type = NPF_ROUTER_IPv4;
        break;
    case 2:
        /* IPv6 - bit 2 */
        pdk_fe_caps.route_table_type = NPF_ROUTER_IPv6;
        break;
    case 3:
        /* BOTH - bit 1 & 2, regardless of architecture it should be 3 */
        pdk_fe_caps.route_table_type = NPF_ROUTER_BOTH;
	break;
    default:
        assert(0);
    }

    /* set the FE capabilities learnt from the BIND event */
    cm_fe_set_caps(fe_npf_handle, &pdk_fe_caps);
    
    npf_ns_close(ns_instance_handle);

    /* now the PORT nodes - we have to create them if this was not a FE
       from a previous bind. If it was, then the PORT nodes are already
       there in Namespace */
    if (check_fe != 1)
    {
        memset(port_path, 0, NPF_NS_MAXPATH);
        sprintf(port_path, "%sFE/%d/Port", g_sys_path, fe_instance);

        /* check if PORT TYPE node exists, if not, create it */
        npf_ret = npf_ns_open (port_path, &ns_type_handle);
        if (npf_ret != NPF_SUCCESS) { // do this only when the error is invalid input!
        npf_ret = npf_ns_create(port_path, NPF_NS_TYPENODE,
                    NPF_PORT, 0, &ns_type_handle);
        if(npf_ret != NPF_SUCCESS) {
            PDBG_PRINT ("Error creating PORT Type node %d\n",
                npf_ret);
            PRETURN (/*void*/);
        }
        }
        /* close the port type node */
        npf_ns_close(ns_type_handle);
        PDBG_PRINT ("new Port Type handle = %d\n", 
            ns_type_handle);
        num_ports = pdk_fe_attribs.num_ports;
    }
    else /* FE node (and its ports) are already there in NS from prev bind */
    {
        /* we now have to read the FE's directory to get the
           handles to all previously created port nodes - what a
           pain!
           Making use of the NS helper routines -
           Get an iterator to the first child of type PORT of this FE
        */
        npf_ret = pdk_ns_getChildItrByType (fe_npf_handle,
                        NPF_PORT,
                        &port_itr);
        if (npf_ret != NPF_SUCCESS) {
        PDBG_PRINT ("Unable to get port iterator for old nodes\n");
        PRETURN (/*void*/);
        }
        prev_fe_ports = npf_ns_count (port_itr);
        PDBG_PRINT ("Number of ports reported previously by FE = %d\n",
            prev_fe_ports);
        num_ports = prev_fe_ports;
    }
    /*
      read the first FPPI_Port_Caps structure from the response
      this pointer will be used to step through the FPPI_FE_Caps
      structure for each port
    */
    port_caps = fe_caps->ports_caps.port_caps_list;
    assert(num_ports <= 1024); // for IXP only
    for(i=0; i<num_ports; i++)
    {
        if (check_fe != 1)
        {
            /* next create the INSTANCE node for the new FE if new FE */
            memset(port_path, 0, NPF_NS_MAXPATH);
            sprintf(port_path, "%sFE/%d/Port/%d", g_sys_path,
                fe_instance, port_instance);
            
            port_instance++;
            npf_ret = npf_ns_create(port_path, NPF_NS_INSTNODE,
                        NPF_PORT, 0, &port_ns_handle);
            if(npf_ret != NPF_SUCCESS) {
            PDBG_PRINT ("Unable to create PORT instance: %d\n",
                    npf_ret);
            PRETURN (/*void*/);
            }
            PDBG_PRINT ("Port Path = %s \n new Port Instance handle = %d\n", 
                port_path, port_ns_handle);
    
            /* now get a handle to the PORT object created */
            npf_ret = npf_ns_getDataHandle(port_ns_handle, 
                           &port_npf_handle);
            if(npf_ret != NPF_SUCCESS) {
            PDBG_PRINT ("unable to get port data hdl\n");
            PDBG_PRINT ("deleting this port instance\n");
            npf_ns_delete(port_path);
            npf_ns_close(port_ns_handle);
            continue;
            }
        } /* end if (NEW PORT node)*/
        else
        {
            /* We already have an iterator to the port nodes if this
               FE was from a previous bind */
            if (i == 0)
            port_ns_handle = npf_ns_first (port_itr);
            else
            port_ns_handle = npf_ns_next (port_itr);
            PDBG_PRINT ("Port Instance handle = %d\n", 
                port_ns_handle);
            /* now get a handle to the PORT object created */
            npf_ret = npf_ns_getDataHandle(port_ns_handle, 
                           &port_npf_handle);
            if(npf_ret != NPF_SUCCESS) {
            PDBG_PRINT ("unable to get port data hdl\n");
            npf_ns_close(port_ns_handle);
            continue;
            }
        }

        /* get the port properties and capabilities */
        memset(&pdk_port_attribs, 0, sizeof(PDK_PORT_ATTRIBS));
        memset(&pdk_port_caps, 0, sizeof(PDK_PORT_CAPS));
        
        pdk_port_attribs.id = port_caps->portid;
        pdk_port_attribs.type = port_caps->port_type;
        npf_hwaddr_char2saddr (port_caps->mac_addr, &mac_sock);

        memcpy(pdk_port_attribs.macaddr,mac_sock.sa_data, 
           MACADDRLEN);
        pdk_port_attribs.maxTxRate = port_caps->link_speed;
        /* current linkspeed should be same as max when first bind */
        pdk_port_attribs.curTxRate = port_caps->link_speed;
        pdk_port_attribs.MTU = port_caps->mtu;
        pdk_port_attribs.status = 0; /* port is down and will up after setip */
        pdk_port_attribs.linkstatus = 0; /* FE fire link up after boot */

	pdk_port_caps.atm_port_caps = port_caps->atmCaps;
        
        /* TO be done later - FIll in PDK_PORT_CAPS also */
        cm_port_set_attribs(port_npf_handle, &pdk_port_attribs);
        cm_port_set_caps(port_npf_handle, &pdk_port_caps);
        /* get to the next PORT_CAPS struct */
        port_caps += 1;
        /* close namespace handle of port */
        npf_ns_close(port_ns_handle);
    } /* end (for all ports in FE) */
    
    /* close the FE instance handle now */
    npf_ns_close(ns_instance_handle);
    if (check_fe == 1)
    npf_ns_done (port_itr);

    /* Send a REPLY to the FE saying that the FECAPABILITY event was
       processed successfully by the CE
    */
    npf_ret = FPPAPI_feCapability_Reply (fe_id);
    if (npf_ret != NPF_SUCCESS)
    PDBG_PRINT ("Error sending feCapability Reply to the FE\n");

    PDBG_PRINT ("Sent FECAPABILITY reply to the FE\n");
    
    /* increment the SYSTEM.num_FEs before CBFunc, when unbind C&M decrement */
    if(cm_sys_incr_numfes() != NPF_SUCCESS)
        assert(0);

    /* 
       Now invoke all callbacks registered by internal modules to inform them
       about the FE bind event 
    */
    
    PDBG_PRINT ("Invoking registered callbacks\n");
    
    client_cb_func = (BIND_CALLBACK) pdk_callback_firstFunc(PDK_BD_EVT_FEBIND,
                                &context,
                                &cb_hdl_next);
    /* 
       if the first function registered was NULL, this will be a problem
       suggestion: fix  this at registration time - cant afford to handle
       this kind of problem here.
    */
    if (client_cb_func == NULL)
    PRETURN (/*void*/);
    do
    {
        cb_hdl = cb_hdl_next;
        client_cb_func(fe_id, context);
    } while ((client_cb_func = (BIND_CALLBACK) 
          pdk_callback_nextFunc(cb_hdl,
                    &context,
                    &cb_hdl_next)) != NULL);
    /* no use after using input a FE path, like myFE23 will be /Sys/0/FE/23 */
    g_fe_instance++;

    PRETURN (/*void*/);
}

/* return fe_npf_hdl without open it, caller doesn't need to close it 
 * return 1 if found and inactive, 0 if not found, -1 if found but active */
static int bd_checkFE (FPPI_FEID fe_id, void* response_data, NPF_HANDLE* fe_npf_hdl)
{
    NPF_RET        ret;
    char           fe_path[NPF_NS_MAXPATH];
    FPPI_FE_Caps*  fe_caps;
    NPF_FE_ATTRIBS fe_attribs;
    NPF_NSHDL      fe_ns_hdl;
    NODE_ITERATOR  fe_itr;
    int            fe_present = 0;

    PDBG_TRACE();

    fe_caps = (FPPI_FE_Caps*) response_data;

    /*
      now we have to traverse the namespace looking for each FE instance
      and check its name for this FE
    */
  
    /* open the root path - sys_path should be /NPFSys/0/ */ 
    memset(fe_path, 0, NPF_NS_MAXPATH);
    /* open the FE Type Node */
    sprintf(fe_path, "%sFE", g_sys_path);
    
    /* open the fe type node */
    ret = npf_ns_open(fe_path, &fe_ns_hdl);
    if(ret != NPF_SUCCESS) {
        PDBG_PRINT ("Unable to open FE node in NS\n");
        PRETURN (fe_present);
    }
    
    /* read its directory, that is the FE INSTANCE Nodes */
    ret = npf_ns_readDir(fe_path, &fe_itr);
    if(ret != NPF_SUCCESS) {
        npf_ns_close(fe_ns_hdl);
        PDBG_PRINT ("No exisiting FEs: Unable to open FE Directory in NS\n");
        PRETURN (fe_present);
    }
    /* close the fe type node */
    npf_ns_close(fe_ns_hdl);
    fe_ns_hdl = npf_ns_first(fe_itr);
    do 
    {
        /* get this FE's data handle */
        ret = npf_ns_getDataHandle(fe_ns_hdl, fe_npf_hdl);
        if (ret != NPF_SUCCESS) {
            PDBG_PRINT ("Unable to get FE handle \n");
            npf_ns_close(fe_ns_hdl);
            PRETURN (fe_present);
        }
        ret = cm_fe_get_attribs(*fe_npf_hdl, &fe_attribs);
        /* FE's name is stored in NPF_FE_ATTRIBS after bound */
        if(strcmp(fe_attribs.name, fe_caps->fe_name) == 0) {
            /* if the FE is on, new FE has name collision */
            PDBG_PRINT ("Found FE from previous bind!!\n");
            if (fe_attribs.status == NPF_FE_UP) {
                fe_present = -1;
                PDBG_PRINT ("A FE with same name is active!!\n");
            } else
                fe_present = 1;
            npf_ns_close(fe_ns_hdl);
            break;
        }
        npf_ns_close(fe_ns_hdl);
        fe_ns_hdl = npf_ns_next(fe_itr);
    } while ( fe_ns_hdl != (NPF_NSHDL) NULL);
    npf_ns_done(fe_itr);

    PRETURN (fe_present);
}

