/****************************************************************************
 *             INTEL CORPORATION PROPRIETARY INFORMATION
 *  This software is supplied under the terms of a license agreement or
 *  nondisclosure agreement with Intel Corporation and may not be copied
 *  or disclosed except in accordance with the terms of that agreement.
 *      Copyright (c) 2002 Intel Corporation. All Rights Reserved.
 *
 * $Source: /cvs/constellations/ixa_3.0/src/cp_pdk/ControlPlane/CoreBlocks/L2IDManager/src/l2idmgr.c,v $
 * $Revision: 1.8 $
 * $Date: 2003/11/17 16:57:12 $
 *
 * Description:
 *    Manage L2ID lists per FEID and NHID.
 *
 ****************************************************************************/

#include <string.h>
#include "list.h"
#include "log.h"
#include "fpp_common.h"
#include "npf_common.h"
#include "fppCapabilityTypes.h"
#include "pdk_l2idm.h"

typedef enum
{
    l2idm_forNHTable,
    l2idm_forL2Table
} L2IDM_FOR_TABLE_TYPE;

typedef enum
{
    l2idm_actionCommit,
    l2idm_actionDelete
} L2IDM_ACTION_TYPE;

typedef enum
{
    l2idm_normal        = 0,
    l2idm_normal_secure = 1,
    l2idm_pending       = 2,
    l2idm_old           = 3,
    l2idm_delete        = 4
} L2IDM_ENTRY_STATUS_TYPE;

typedef enum
{
    l2idm_ipv4,
    l2idm_ipv6
} L2IDM_IPADDR_TYPE;

typedef struct 
{
    L2IDM_IPADDR_TYPE  ipType;
    void*              ipAddrPtr;
} L2IDM_IPADDR_REF;

typedef struct
{
    L2IDM_IPADDR_TYPE  ipType;
    union {
        NPF_IPv4Address_t  ipv4Addr;
        NPF_IPv6Address_t  ipv6Addr;
    } addr;
} L2IDM_IPADDR;

/* The object that contains the master lists for NH and L2 tables
   with sync objects. */
typedef struct L2IDM_MASTER_LIST_tag
{
    DList list;
    PilCriticalSection listLock;
} L2IDM_MASTER_LIST;

/* Struct per blade (FE). New obj instantiated when module receives
   an feid that it does not yet contain already. Contains list
   of ptrs to structs below. */
typedef struct L2IDM_FEID_LIST_ENTRY_tag
{
    FPPI_FEID  feid;
    DList      nhidList;
    DList      l2idList;
    uint32_t   l2idIncrementor;
} L2IDM_FEID_LIST_ENTRY;

/* Struct per nhid for a specified feid. New list instantiated when
   L2id mgr rcvs a new nhid. Contains list of NHAddrs and
   corresponding L2ids. */
typedef struct L2IDM_NHID_LIST_ENTRY_tag
{
    FPPI_NHID  nhid;
    DList      l2idList;
} L2IDM_NHID_LIST_ENTRY;

/* List entry per feid (and possibly nhid) for NH */
typedef struct L2IDM_L2ID_LIST_ENTRY_tag
{
    L2IDM_IPADDR            nhAddr;
    FPPI_L2ID               l2id;
    L2IDM_ENTRY_STATUS_TYPE status;
} L2IDM_L2ID_LIST_ENTRY;

#define UNINITIALIZED       0
#define INITIALIZED         1

#define L2ID_OK             0
#define L2ID_NO_FEID_ENTRY  1
#define L2ID_NO_NHID_ENTRY  2

/* Set the address size constants for IPv4 and IPv6 */
static const long IPV4_ADDR_SZ    = sizeof(NPF_IPv4Address_t);
static const long IPV6_ADDR_SZ    = sizeof(NPF_IPv6Address_t);

/* The global list that contains one sublist per feid */
static L2IDM_MASTER_LIST g_l2idMgrMasterList;

/* Flag to indicate whether module is initialized or not */
static uint32_t g_l2idmState = UNINITIALIZED;

/*****************************************************************************
* Internal method prototypes
*****************************************************************************/

//void CleanupAList(DList* list);

NPF_RET UpdateForNextHopTable_Core(
    FPPI_FEID*          feids,
    FPPI_NHID           nhid,
    uint32_t            countIn,
    L2IDM_IPADDR_REF*   nhAddrs,
    uint32_t*           numOldL2ids,
    FPPI_FEandL2IDs**   oldL2idList,
    uint32_t*           numNewL2ids,
    FPPI_FEandL2IDs**   newL2idList);

NPF_RET UpdateForL2Table_Core(
    FPPI_FEID           feid,
    uint32_t            numNHAddrs,
    L2IDM_IPADDR_REF*   nhAddrs,
    uint32_t*           numNewL2ids,
    FPPI_L2ID**         newL2idList);

NPF_RET CommitForNextHopTable_Core(
    FPPI_NHID           nhid,
    uint32_t            numNHAddrs,
    L2IDM_IPADDR_REF*   nhAddrs);

NPF_RET CommitForL2Table_Core(
    FPPI_FEID           feid,
    uint32_t            numNHAddrs,
    L2IDM_IPADDR_REF*   nhAddrs);

NPF_RET GetForNextHopTable_Core(
    FPPI_NHID           nhid,
    uint32_t            numNHAddrs,
    L2IDM_IPADDR_REF*   nhAddrs,
    uint32_t*           numL2ids,
    FPPI_FEandL2IDs**   l2idList);

NPF_RET GetForL2Table_Core(
    FPPI_FEID           feid,
    uint32_t            numNHAddrs,
    L2IDM_IPADDR_REF*   nhAddrs,
    uint32_t*           numL2ids,
    FPPI_L2ID**         l2idList);

NPF_RET DeleteForNextHopTable_Core(
    FPPI_NHID           nhid,
    uint32_t            numNHAddrs,
    L2IDM_IPADDR_REF*   nhAddrs);

NPF_RET DeleteForL2Table_Core(
    FPPI_FEID           feid,
    uint32_t            numNHAddrs,
    L2IDM_IPADDR_REF*   nhAddrs);

static NPF_RET UpdateL2IDTableForNH(
    L2IDM_IPADDR_TYPE       addrType,
    void*                   nhAddr,
    L2IDM_FEID_LIST_ENTRY*  fiedEntry,
    DList*                  l2Tbl,
    BOOL*                   isNewEntry,
    FPPI_L2ID*              newL2ID);

static NPF_RET UpdateSingleEntry(
    FPPI_FEID             feid,
    FPPI_NHID             nhid,
    L2IDM_IPADDR_TYPE     addrType,
    void*                 nhAddr,
    BOOL*                 isNewEntry,
    uint32_t*             newL2ID);

static NPF_RET HandleOldEntries(
    FPPI_NHID          nhid,
    uint32_t           countIn,
    FPPI_FEID*         feids,
    uint32_t*          numOldL2ids,
    FPPI_FEandL2IDs**  oldL2idList);

static NPF_RET CommitOrDeleteCore(
    L2IDM_FOR_TABLE_TYPE    tblType,
    L2IDM_ACTION_TYPE       action,
    FPPI_FEID               feid,
    FPPI_NHID               nhid,
    uint32_t                numNHAddrs,
    L2IDM_IPADDR_REF*       nhAddrs,
    BOOL*                   isDeleted);

static uint32_t GetTableBasedOnFEIDandNHID(
    L2IDM_FOR_TABLE_TYPE     tblType,
    FPPI_FEID                feid,
    FPPI_NHID                nhid,
    DList**                  table,
    L2IDM_FEID_LIST_ENTRY**  feidEntry);

static NPF_RET CreateAndInitNHIDObj(
    FPPI_NHID  nhid,
    L2IDM_NHID_LIST_ENTRY** nhidEntry);

static NPF_RET AddNHAddrInfoToList(
    L2IDM_FEID_LIST_ENTRY* fiedEntry,
    DList*                 list,
    uint32_t               numNHAddrs,
    L2IDM_IPADDR_TYPE      addrType,
    void*                  nhAddrs,
    uint32_t*              numNewL2ids,
    FPPI_L2ID**            newL2idList);

static NPF_RET UpdateL2IDTableForL2(
    uint32_t                numNHAddrs,
    L2IDM_IPADDR_REF*       nhAddrs,
    L2IDM_FEID_LIST_ENTRY*  fiedEntry,
    DList*                  l2Tbl,
    uint32_t*               numNewL2ids,
    FPPI_L2ID**             newL2idList);

static NPF_RET CreateAndFillNewFEIDList(
    L2IDM_FOR_TABLE_TYPE    tblType,
    FPPI_FEID               feid,
    FPPI_NHID               nhid,
    uint32_t                numNHAddrs,
    L2IDM_IPADDR_TYPE       addrType,
    void*                   nhAddr,
    uint32_t*               numNewL2ids,
    FPPI_L2ID**             newL2idList);

static void ScanForNhAddrInL2List(
    BOOL                    firstPass,
    L2IDM_L2ID_LIST_ENTRY*  entry,
    uint32_t                numNHAddrs,
    L2IDM_IPADDR_REF*       nhAddrs,
    uint32_t*               newOutListIncr,
    uint32_t*               tmpFlags,
    uint32_t*               numNewL2ids,
    FPPI_L2ID*              outNewL2id);

static void GetNhAddrFromL2List(
    BOOL                    firstPass,
    FPPI_FEID               feid,
    L2IDM_L2ID_LIST_ENTRY*  entry,
    uint32_t                numNHAddrs,
    L2IDM_IPADDR_REF*       nhAddrs,
    uint32_t*               outCnt,
    FPPI_FEandL2IDs**       l2idList);

static NPF_RET CommitOrDeleteEntriesInL2IDTable(
    L2IDM_FOR_TABLE_TYPE  tblType,
    L2IDM_ACTION_TYPE     action,
    uint32_t              numNHAddrs,
    L2IDM_IPADDR_REF*     nhAddrs,
    DList*                l2Tbl);

static NPF_RET RemoveFlaggedListEntries(
    DList* l2Tbl,
    BOOL*  entryDeleted);

static NPF_RET RetrieveL2IDs(
    uint32_t            numNHAddrs,
    L2IDM_IPADDR_REF*   nhAddrs,
    DList*              l2Tbl,
    uint32_t*           numL2ids,
    FPPI_L2ID**         l2idList);

static void CheckDeleteEntryFromMasterList(
    FPPI_FEID              feid,
    L2IDM_FEID_LIST_ENTRY* fiedEntry,
    BOOL*                  isDeleted);

/****************************************************************************
 * Function: compareFEIDEntry()                                             *
 * ---------------------------                                              *
 * Parameters: data - the data element in the master list                   *
 *             key  - feid                                                  *
 *                                                                          *
 * Description: Compare the nhid to find list entry                         *
 *                                                                          *
 * Return: 1 if comparison is successful , 0 if comparison fails            *
 *                                                                          *
 ****************************************************************************/

static int compareFEIDEntry(const void* data, const void* key)
{
    return ( ((L2IDM_FEID_LIST_ENTRY*)data)->feid == *((FPPI_FEID*)key) );
}

/****************************************************************************
 * Function: compareNHIDEntry()                                             *
 * ---------------------------                                              *
 * Parameters: data - the data element in the FEID entry                    *
 *             key  - nhid                                                  *
 *                                                                          *
 * Description: Compare the nhid to find list entry                         *
 *                                                                          *
 * Return: 1 if comparison is successful , 0 if comparison fails            *
 *                                                                          *
 ****************************************************************************/

static int compareNHIDEntry(const void* data, const void* key)
{
    return ( ((L2IDM_NHID_LIST_ENTRY*)data)->nhid == *((FPPI_NHID*)key) );
}

/*****************************************************************************
*
*   Function: l2idm_init()
*
*   Parameters: none
*
*   Description: initialize the label manager. To be called by PDK manager
*
*   Return: NPF_SUCCESS
*
*****************************************************************************/

NPF_RET l2idm_init()
{
    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    if(g_l2idmState == INITIALIZED)
    {
        return NPF_SUCCESS;
    }         

    npf_list_init(&(g_l2idMgrMasterList.list), PIL_FreeMemory);
    g_l2idMgrMasterList.list.compare = &compareFEIDEntry;
    PIL_InitializeCriticalSection(&(g_l2idMgrMasterList.listLock));
    g_l2idmState = INITIALIZED;    

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");

    return NPF_SUCCESS;

} /* l2idm_init */

/*****************************************************************************
*   Function: l2idm_shutdown()
*
*   Parameters: none
*
*   Description: shutdown the L2ID manager
*
*   Return:  NPF_SUCCESS or error code
*
*****************************************************************************/

NPF_RET l2idm_shutdown()
{
    DListItr* itr1 = NULL;
    DListItr* itr2 = NULL;
    L2IDM_FEID_LIST_ENTRY* p1Entry;
    L2IDM_NHID_LIST_ENTRY* p2Entry;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    if(g_l2idmState != INITIALIZED)
    {
        return NPF_COMPONENT_UNINITIALIZED;
    }    
    
    PIL_EnterCriticalSection(&g_l2idMgrMasterList.listLock);

    if ( (itr1 = npf_list_itrCreate(&g_l2idMgrMasterList.list)) == NULL ||
        npf_list_itrFirst(itr1) != 0)
    {
        /* List is probably empty */
        goto LBL_END;
    }

    do
    {
        /* For each FEID, remove embedded list data and destroy the list */
        if ((p1Entry = (L2IDM_FEID_LIST_ENTRY*)npf_list_itrGetData(itr1)) ==
            NULL)
        {
            /* No data or itr messed up */
            npf_logger_write(LOG_IPV4, LOG_INFO,
                "No data in this list entry.\n");
            continue;
        }
        
        /* Destroy all L2ID entries for this FEID then destroy
           the list. */
        npf_list_destroy(&p1Entry->l2idList);
        
        /* Now delete the NHID list and all embedded entries */            
        if ( (itr2 = npf_list_itrCreate(&p1Entry->nhidList)) == NULL)
        {
            /* List is probably empty */
            continue;
        }

        if (npf_list_itrFirst(itr2) == -1)
        {
            /* Destroy the embedded list */
            npf_list_destroy(&p1Entry->nhidList);

            /* Remove this entry from higher level list */
            npf_list_itrRemoveFree(itr1);

            /* Continue on to next master list entry (next FEID) */
            continue;
        }

        do
        {
            /* For each NHID entry in list destroy embedded data */
            if ((p2Entry = (L2IDM_NHID_LIST_ENTRY*)
                npf_list_itrGetData(itr2)) == NULL)
            {
                /* No data or itr messed up */
                npf_logger_write(LOG_IPV4, LOG_INFO,
                    "No data in this list entry.\n");
                continue;
            }
        
            /* Destroy all L2ID entries for this NHID and FEID then
               destroy the list. */
            npf_list_destroy(&p2Entry->l2idList);

            /* Remove this entry from higher level list */
            npf_list_itrRemoveFree(itr2);

        } while( npf_list_itrNext(itr2) != -1);

        /* Destroy the list */
        npf_list_destroy(&p1Entry->nhidList);
        /* Remove this entry from higher level list */
        npf_list_itrRemoveFree(itr1);

    } while (npf_list_itrNext(itr1) != -1);

LBL_END:

    if (itr1 != NULL)
    {
        npf_list_itrDelete(itr1);
    }
    if (itr2 != NULL)
    {
        npf_list_itrDelete(itr2);
    }

    npf_list_destroy(&g_l2idMgrMasterList.list); 
    
    PIL_LeaveCriticalSection(&g_l2idMgrMasterList.listLock);    
    PIL_DestroyCriticalSection(&g_l2idMgrMasterList.listLock); 
    g_l2idmState = UNINITIALIZED; 

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");

    return NPF_SUCCESS;
}

/*****************************************************************************
*   Function: l2idm_FlushNHIDList()
*
*   Parameters: none
*
*   Description: Delete all NHID entries for all FEID entries in master
*                list. Don't destroy the NHID lists, just empty them and
*                do not touch the top level L2ID lists in the FEID entry
*                (those that are peer with the NHID list).
*
*   Return:  NPF_SUCCESS or error code
*
*****************************************************************************/

NPF_RET l2idm_FlushNHIDList()
{
    DListItr* itr1 = NULL;
    DListItr* itr2 = NULL;
    L2IDM_FEID_LIST_ENTRY* p1Entry;
    L2IDM_NHID_LIST_ENTRY* p2Entry;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    if(g_l2idmState != INITIALIZED)
    {
        return NPF_COMPONENT_UNINITIALIZED;
    }    
    
    PIL_EnterCriticalSection(&g_l2idMgrMasterList.listLock);

    if ( (itr1 = npf_list_itrCreate(&g_l2idMgrMasterList.list)) == NULL ||
        npf_list_itrFirst(itr1) != 0)
    {
        /* List is probably empty */
        goto LBL_END;
    }

    /* For each FEID entry, remove all NHID list entries */
    do
    {
        /* Get the next FEID entry in master list */
        if ((p1Entry = (L2IDM_FEID_LIST_ENTRY*)npf_list_itrGetData(itr1)) ==
            NULL)
        {
            /* No data or itr messed up */
            npf_logger_write(LOG_IPV4, LOG_INFO,
                "No data in this list entry.\n");
            continue;
        }
        
        /* Now empty the NHID list and free all embedded data */            
        if ( (itr2 = npf_list_itrCreate(&p1Entry->nhidList)) == NULL)
        {
            /* List is already empty so continue to next FEID entry */
            continue;
        }

        if (npf_list_itrFirst(itr2) == -1)
        {
            /* List is already empty so continue to next FEID entry */
            continue;
        }

        do
        {
            /* For each NHID entry in list destroy embedded data */
            if ((p2Entry = (L2IDM_NHID_LIST_ENTRY*)
                npf_list_itrGetData(itr2)) == NULL)
            {
                /* No data or itr messed up */
                npf_logger_write(LOG_IPV4, LOG_INFO,
                    "No data in this list entry.\n");
                continue;
            }
        
            /* Destroy all L2ID entries for this NHID and FEID then
               destroy the list. */
            npf_list_destroy(&p2Entry->l2idList);

            /* Remove this entry from higher level list */
            npf_list_itrRemoveFree(itr2);

        } while( npf_list_itrNext(itr2) != -1);

    } while (npf_list_itrNext(itr1) != -1);

LBL_END:

    if (itr2 != NULL)
    {
        npf_list_itrDelete(itr2);
    }
    if (itr1 != NULL)
    {
        npf_list_itrDelete(itr1);
    }

    PIL_LeaveCriticalSection(&g_l2idMgrMasterList.listLock);    

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");

    return NPF_SUCCESS;
}

/*****************************************************************************
*   Function: l2idm_GetNumOfL2Ids()
*
*   Parameters: none
*
*   Description: Get Number of L2Ids in the L2ID manager
*
*   Return:  NPF_SUCCESS or error code
*
*****************************************************************************/
NPF_RET l2idm_GetNumOfL2Ids(FPPI_FEID feId, uint32_t *numL2Ids)
{
    NPF_RET retVal  = NPF_SUCCESS;
    DListItr* itr1 = NULL;
    L2IDM_FEID_LIST_ENTRY* feEntry = NULL;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    if(g_l2idmState != INITIALIZED)
    {
        return NPF_COMPONENT_UNINITIALIZED;
    }    
    
    PIL_EnterCriticalSection(&g_l2idMgrMasterList.listLock);

    if ( (itr1 = npf_list_itrCreate(&g_l2idMgrMasterList.list)) == NULL ||
        npf_list_itrFirst(itr1) != 0)
    {
        /* List is probably empty */
        goto LBL_END;
    }

    do
    {
       /* For each FEID*/
       if((feEntry = (L2IDM_FEID_LIST_ENTRY*)npf_list_itrGetData(itr1))
              == NULL)
        {
            /* No data or itr messed up */
            npf_logger_write(LOG_IPV4, LOG_INFO,
                "No data in this list entry.\n");
            continue;
        }
        if( feId == feEntry->feid)
        {
           /* Now get number from L2Id list and all embedded entries */
           *numL2Ids = npf_list_size(&feEntry->nhidList);
           npf_logger_write(LOG_IPV4, LOG_INFO,
                 "Total L2Ids = %d\n",*numL2Ids);
           goto LBL_END; 
        }
    } while (npf_list_itrNext(itr1) != -1);
    
LBL_END:

    if (itr1 != NULL)
    {
        npf_list_itrDelete(itr1);
    }
    PIL_LeaveCriticalSection(&g_l2idMgrMasterList.listLock);    
    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");
    return retVal;
}

NPF_RET l2idm_GetL2IdList(FPPI_FEID feId, uint32_t numEntries,
                          FPPI_L2ID *l2IdArray)
{
    NPF_RET retVal  = NPF_SUCCESS;
    DListItr* itr1 = NULL;
    DListItr* itr2 = NULL;
    DListItr* itr3 = NULL;
    L2IDM_FEID_LIST_ENTRY* feEntry = NULL;
    L2IDM_NHID_LIST_ENTRY* nhEntry = NULL;
    L2IDM_L2ID_LIST_ENTRY* l2Entry = NULL;
    uint32_t i=0, numL2 =0;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");
    
    if(l2IdArray == NULL)
      return NPF_E_UNKNOWN;
     
    if(g_l2idmState != INITIALIZED)
    {
        return NPF_COMPONENT_UNINITIALIZED;
    }    
    
    PIL_EnterCriticalSection(&g_l2idMgrMasterList.listLock);

    if ( (itr1 = npf_list_itrCreate(&g_l2idMgrMasterList.list)) == NULL ||
        npf_list_itrFirst(itr1) != 0)
    {
        /* List is probably empty */
        goto LBL_END;
    }

    do
    {
        /* For each FEID, remove embedded list data and destroy the list */
        if ((feEntry = (L2IDM_FEID_LIST_ENTRY*)npf_list_itrGetData(itr1)
            ) == NULL)
        {
            /* No data or itr messed up */
            npf_logger_write(LOG_IPV4, LOG_INFO,
                "No data in this list entry.\n");
            continue;
        }
        if( feEntry->feid != feId)
        {
            /* feid doesn't match */
            npf_logger_write(LOG_IPV4, LOG_INFO,
                "No data available for FeId %d .\n", feId);
            continue;
        }

        numL2 = npf_list_size(&feEntry->nhidList);
        /* Safe check */
        if(numL2 != numEntries)
        {
           npf_logger_write(LOG_IPV4, LOG_ERR,
                    "Invalid number of entries received .\n");
           retVal = NPF_E_UNKNOWN;
           goto LBL_END;
        }
        /* Get the iterator for nhid list  */ 
        if ( (itr2 = npf_list_itrCreate(&feEntry->nhidList))== NULL)
        {
            /* List is probably empty */
            continue;
        }
        if (npf_list_itrFirst(itr2) == -1)
        {
            npf_list_itrRemoveFree(itr1);
            /* Continue on to next master list entry (next FEID) */
            continue;
        }

        do
        {
            /* For each L2Id entry in list*/
            if ((nhEntry = (L2IDM_NHID_LIST_ENTRY*)
                npf_list_itrGetData(itr2)) == NULL)
            {
                /* No data or itr messed up */
                npf_logger_write(LOG_IPV4, LOG_INFO,
                    "No data in this list entry.\n");
                continue;
            }
            
            if ( (itr3 = npf_list_itrCreate(&nhEntry->l2idList))== NULL)
            {
                /* List is probably empty */
                continue;
            }
            if (npf_list_itrFirst(itr3) == -1)
            {
                npf_list_itrRemoveFree(itr1);
                /* Continue on to next master list entry (next FEID) */
                continue;
            }
            /* For each L2Id entry in list*/
            do
            {
               if ((l2Entry = (L2IDM_L2ID_LIST_ENTRY*)
                      npf_list_itrGetData(itr3)) == NULL)
               {
                   /* No data or itr messed up */
                   npf_logger_write(LOG_IPV4, LOG_INFO,
                    "No data in this list entry.\n");
                   continue;
               }

               l2IdArray[i] = l2Entry->l2id; 
               i++;    
            } while( npf_list_itrNext(itr3) != -1);
        } while( npf_list_itrNext(itr2) != -1);
    } while (npf_list_itrNext(itr1) != -1);

LBL_END:

    if (itr1 != NULL)
    {
        npf_list_itrDelete(itr1);
    }
    if (itr2 != NULL)
    {
        npf_list_itrDelete(itr2);
    }

    PIL_LeaveCriticalSection(&g_l2idMgrMasterList.listLock);    
    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");
    return retVal;
}

/*****************************************************************************
*   Function: l2idm_UpdateForNextHopTable()
*
*   Parameters:
*       nhid        - next hop ID specified by app
*       countIn     - num elements in next two params
*       nhAddrs     - array of next hop addrs to add
*       feids       - forwarding element ID list (same as blade ID)
*       numOldL2ids - [out] num old L2IDs added
*       oldL2idList - [out] list of old L2IDs added
*       numNewL2ids - [out] num new L2IDs added
*       newL2idList - [out] list of new L2IDs added
*
*   Description:
*       Recieves a list of <fied, NHaddrs> pairs for the specified nhid.
*       Compares these with those in internal L2 table and:
*         1) if <fied, NHAddr> is not in internal table, adds new entry,
*            assigning new L2IDs for each. Sets pending flag to TRUE.
*            Adds all new L2IDs to new L2ID out list.
*         2) if there are L2IDs in the internal table that are not in the
*            passed in list these L2IDs are first marked as "old". Then,
*            after all input entries have been checked against the
*            internal table, the internal table is scanned and all entries
*            marked as old are added to the old list. At the same time
*            the old marking is removed.
*
*       If a table does not yet exist for the specified FEID and NHID a
*       new one is created and added to the appropriate entry.
*       Memory for out lists is allocated within this method and must be
*       freed by the caller!
*
*   Return:  NPF_SUCCESS or error code
*
*****************************************************************************/

NPF_RET l2idm_UpdateForNextHopTable(
    FPPI_FEID*          feids,
    FPPI_NHID           nhid,
    uint32_t            countIn,
    NPF_IPv4Address_t*  nhAddrs,
    uint32_t*           numOldL2ids,
    FPPI_FEandL2IDs**   oldL2idList,
    uint32_t*           numNewL2ids,
    FPPI_FEandL2IDs**   newL2idList)
{
    L2IDM_IPADDR_REF ipAddr;

    ipAddr.ipType     = l2idm_ipv4;
    ipAddr.ipAddrPtr  = nhAddrs;

    return UpdateForNextHopTable_Core(
        feids,
        nhid,
        countIn,
        &ipAddr,
        numOldL2ids,
        oldL2idList,
        numNewL2ids,
        newL2idList);
}

NPF_RET l2idm_ipv6_UpdateForNextHopTable(
    FPPI_FEID*          feids,
    FPPI_NHID           nhid,
    uint32_t            countIn,
    NPF_IPv6Address_t*  nhAddrs,
    uint32_t*           numOldL2ids,
    FPPI_FEandL2IDs**   oldL2idList,
    uint32_t*           numNewL2ids,
    FPPI_FEandL2IDs**   newL2idList)
{
    L2IDM_IPADDR_REF ipAddr;

    ipAddr.ipType     = l2idm_ipv6;
    ipAddr.ipAddrPtr  = nhAddrs;

    return UpdateForNextHopTable_Core(
        feids,
        nhid,
        countIn,
        &ipAddr,
        numOldL2ids,
        oldL2idList,
        numNewL2ids,
        newL2idList);
}

NPF_RET UpdateForNextHopTable_Core(
    FPPI_FEID*          feids,
    FPPI_NHID           nhid,
    uint32_t            countIn,
    L2IDM_IPADDR_REF*   nhAddrs,
    uint32_t*           numOldL2ids,
    FPPI_FEandL2IDs**   oldL2idList,
    uint32_t*           numNewL2ids,
    FPPI_FEandL2IDs**   newL2idList)
{
    NPF_RET retVal  = NPF_SUCCESS;
    BOOL isNewEntry;
    uint32_t i;
    FPPI_L2ID newL2ID;
    BOOL isLocked = FALSE;
    size_t             addrSz;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    /* Init out params */
    *numNewL2ids = 0;
    *newL2idList = NULL;
    *numOldL2ids = 0;
    *oldL2idList = NULL;

    /* Check in params */
    if (countIn == 0 || nhAddrs->ipAddrPtr == NULL)
    {
        retVal = NPF_INVALID_PARAMETERS;
        npf_logger_write(LOG_IPV4, LOG_ERR,
            "Invalid parms into L2IDmgr - exiting with status %d.\n",
             retVal);
        goto LBL_END;
    }

    /* Allocate ample memory for out list of new structs */
    if ((*newL2idList = (FPPI_FEandL2IDs*)PIL_AllocateMemory(
        sizeof(FPPI_FEandL2IDs) * countIn, PIL_MEM_ZERO_INIT)) == NULL)
    {
        npf_logger_write(LOG_IPV4, LOG_ERR, "Could not allocate memory.\n");
        retVal = NPF_OUT_OF_MEMORY;
        goto LBL_END;
    }

    /* Set address size */
    addrSz = nhAddrs->ipType == l2idm_ipv4 ? IPV4_ADDR_SZ : IPV6_ADDR_SZ;

    /* Lock the list while searching, updating. */
    PIL_EnterCriticalSection(&g_l2idMgrMasterList.listLock);
    isLocked = TRUE;

    /* Loop over all FEIDs and NHAddrs supplied by client */
    for (i = 0; i < countIn; i++)
    {
        isNewEntry = FALSE;

        if ((retVal = UpdateSingleEntry(
            feids[i],
            nhid,
            nhAddrs->ipType,
            ((void*)nhAddrs->ipAddrPtr) + (i * addrSz),
            &isNewEntry,
            &newL2ID)) != NPF_SUCCESS)
        {
            /* On error break out of loop and handle old entries */
            break;
        }

        /* If this entry is a new one put it in new out list */
        if (isNewEntry)
        {
            (*newL2idList)[i].feid = feids[i];
            (*newL2idList)[i].l2id = newL2ID;

            /* Increment count */
            (*numNewL2ids)++;
        }

    } /* for (i = 0; i < countIn; i++) */

    /* Now loop through list elements with specified NHID and FEID
       and count number of entries marked as old. At the same time
       remark those with normal secure to just normal. Then return
       in out list all old L2IDs and FEIDs. */
    retVal = HandleOldEntries(
        nhid, countIn, feids, numOldL2ids, oldL2idList);

LBL_END:

    if (isLocked)
    {
        PIL_LeaveCriticalSection(&g_l2idMgrMasterList.listLock);
    }

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/*****************************************************************************
*   Function: l2idm_UpdateForL2Table()
*
*   Parameters:
*       feid        - forwarding element ID (same as blade ID)
*       numNHAddrs  - num next hop addrs to add
*       nhAddrs     - array of next hop addrs to add
*       numNewL2ids - [out] num new L2IDs added
*       newL2idList - [out] list of new L2IDs added
*
*   Description:
*       Recieves a list of NHaddrs for a specified feid. Compares
*       NHAddrs with those in internal L2 table and:
*         1) if NHAddr is not in internal table, adds new NHaddrs,
*            assigning new L2IDs for each. Sets pending flag to TRUE.
*            Adds all new L2IDs to new L2ID out list.
*         2) if there are L2IDs in the internal table that are not in the
*            passed in list these L2IDs are put into the old L2ID out list.
*
*       If a table does not yet exist for the specified FEID and NHID a
*       new one is created nad added to the appropriate entry.
*       Memory for out lists is allocated within this method and must be
*       freed by the caller!
*
*   Return:  NPF_SUCCESS or error code
*
*****************************************************************************/

NPF_RET l2idm_UpdateForL2Table(
    FPPI_FEID           feid,
    uint32_t            numNHAddrs,
    NPF_IPv4Address_t*  nhAddrs,
    uint32_t*           numNewL2ids,
    FPPI_L2ID**         newL2idList)
{
    L2IDM_IPADDR_REF ipAddr;

    ipAddr.ipType     = l2idm_ipv4;
    ipAddr.ipAddrPtr  = nhAddrs;

    return UpdateForL2Table_Core(
        feid,
        numNHAddrs,
        &ipAddr,
        numNewL2ids,
        newL2idList);
}

NPF_RET l2idm_ipv6_UpdateForL2Table(
    FPPI_FEID           feid,
    uint32_t            numNHAddrs,
    NPF_IPv6Address_t*  nhAddrs,
    uint32_t*           numNewL2ids,
    FPPI_L2ID**         newL2idList)
{
    L2IDM_IPADDR_REF ipAddr;

    ipAddr.ipType     = l2idm_ipv6;
    ipAddr.ipAddrPtr  = nhAddrs;

    return UpdateForL2Table_Core(
        feid,
        numNHAddrs,
        &ipAddr,
        numNewL2ids,
        newL2idList);
}

NPF_RET UpdateForL2Table_Core(
    FPPI_FEID           feid,
    uint32_t            numNHAddrs,
    L2IDM_IPADDR_REF*   nhAddrs,
    uint32_t*           numNewL2ids,
    FPPI_L2ID**         newL2idList)
{
    NPF_RET retVal = NPF_SUCCESS;
    DList* table   = NULL;
    uint32_t tableRet = 0;
    L2IDM_FEID_LIST_ENTRY* fiedEntry = NULL;
    BOOL isLocked = FALSE;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    /* Init out params */
    *numNewL2ids = 0;
    *newL2idList = NULL;

    /* Check in params */
    if (numNHAddrs == 0 || nhAddrs->ipAddrPtr == NULL)
    {
        retVal = NPF_INVALID_PARAMETERS;
        npf_logger_write(LOG_IPV4, LOG_ERR,
            "Invalid parms into L2IDmgr - exiting with status %d.\n",
             retVal);
        goto LBL_END;
    }

    /* Lock the list while searching, updating. */
    PIL_EnterCriticalSection(&g_l2idMgrMasterList.listLock);
    isLocked = TRUE;

    /* Locate table with specified FEID and NHID if it exists */
    tableRet = GetTableBasedOnFEIDandNHID(l2idm_forL2Table, feid, 0,
        &table, &fiedEntry);

    switch (tableRet)
    {
    case L2ID_OK:

        /* Found NHID list - compare entries with those in passed in
           NHaddrs list. Fill new and old L2ID out lists. */
        UpdateL2IDTableForL2(numNHAddrs, nhAddrs,
            fiedEntry, table, numNewL2ids, newL2idList);

        break;
        
    case L2ID_NO_FEID_ENTRY:

        /* Create a new master table entry for this new FEID and insert
           into the master global list. The old L2ID out list will
           be empty. */
        retVal = CreateAndFillNewFEIDList(l2idm_forL2Table, feid,
           0, numNHAddrs, nhAddrs->ipType, nhAddrs, numNewL2ids, newL2idList);

        break;
        
    } /* switch (tableRet) */

LBL_END:

    if (isLocked)
    {
        PIL_LeaveCriticalSection(&g_l2idMgrMasterList.listLock);    
    }

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/*****************************************************************************
*   Function: l2idm_CommitForNextHopTable()
*
*   Parameters:
*       nhid        - next hop ID specified by app
*       numNHAddrs  - num next hop addrs to add
*       nhAddrs     - array of next hop addrs to add
*
*   Description:
*       This method is invoked when a callback is received from FP that
*       a previous NHAdd method was successful. This method must locate
*       the passed in NHAddrs with the specified NHID in the internal
*       L2ID table and unset the pending flags. It must then delete all
*       entries in internal table that do not exist in passed in NHAddr
*       table. All feid records must be checked so a loop is made here.
*       Pass on to core method for processing.
*
*   Return:  NPF_SUCCESS or error code
*
*****************************************************************************/

NPF_RET l2idm_CommitForNextHopTable(
    FPPI_NHID           nhid,
    uint32_t            numNHAddrs,
    NPF_IPv4Address_t*  nhAddrs)
{
    L2IDM_IPADDR_REF ipAddr;
/*    L2IDM_IPADDR_TYPE  ipType;
    union {
        NPF_IPv4Address_t  ipv4Addr;
        NPF_IPv6Address_t  ipv6Addr;
    } addr;
} L2IDM_IPADDR;
*/
    ipAddr.ipType     = l2idm_ipv4;
    ipAddr.ipAddrPtr  = nhAddrs;

    return CommitForNextHopTable_Core(
        nhid,
        numNHAddrs,
        &ipAddr);
}

NPF_RET l2idm_ipv6_CommitForNextHopTable(
    FPPI_NHID           nhid,
    uint32_t            numNHAddrs,
    NPF_IPv6Address_t*  nhAddrs)
{
    L2IDM_IPADDR_REF ipAddr;

    ipAddr.ipType     = l2idm_ipv6;
    ipAddr.ipAddrPtr  = nhAddrs;

    return CommitForNextHopTable_Core(
        nhid,
        numNHAddrs,
        &ipAddr);
}

NPF_RET CommitForNextHopTable_Core(
    FPPI_NHID           nhid,
    uint32_t            numNHAddrs,
    L2IDM_IPADDR_REF*   nhAddrs)
{
    NPF_RET retVal      = NPF_SUCCESS;
    DListItr * itr      = NULL;
    L2IDM_FEID_LIST_ENTRY* p1Entry;
    BOOL continueLoop   = TRUE;
    BOOL isNodeDeleted  = FALSE;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    /* Lock the list while searching, updating. */
    PIL_EnterCriticalSection(&g_l2idMgrMasterList.listLock);

    while (continueLoop)
    {
        isNodeDeleted = FALSE;

        if ( (itr = npf_list_itrCreate(&g_l2idMgrMasterList.list)) == NULL ||
            npf_list_itrFirst(itr) != 0)
        {
            /* List is probably empty */
            retVal = NPF_FAIL;
            npf_logger_write(LOG_IPV4, LOG_ERR,
                "Error - list is empty for commit.\n");
            goto LBL_END;
        }

        do {
            if ((p1Entry = (L2IDM_FEID_LIST_ENTRY*)
                npf_list_itrGetData(itr)) == NULL)
            {
                /* No data or itr messed up */
                retVal = L2ID_NO_FEID_ENTRY;
                npf_logger_write(LOG_IPV4, LOG_INFO,
                    "No data in this list entry.\n");
                continue;
            }

            retVal = CommitOrDeleteCore(
                l2idm_forNHTable,
                l2idm_actionCommit,
                p1Entry->feid,
                nhid,
                numNHAddrs,
                nhAddrs,
                &isNodeDeleted);

            if (retVal != NPF_SUCCESS)
            {
                continueLoop = FALSE;
                break;
            }

            if (isNodeDeleted)
            {
                /* Must destroy iterator and recreate */
                break;
            }

        } while( npf_list_itrNext(itr) != -1);

        if (!isNodeDeleted)
        {
            continueLoop = FALSE;
        }
        else
            break;

    } /* while (continueLoop) */

LBL_END:

    /* Destroy the list iterator */
    if (itr != NULL)
    {
        npf_list_itrDelete(itr);
    }

    PIL_LeaveCriticalSection(&g_l2idMgrMasterList.listLock);    

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/*****************************************************************************
*   Function: l2idm_CommitForL2Table()
*
*   Parameters:
*       feid        - forwarding element ID (same as blade ID)
*       numNHAddrs  - num next hop addrs to add
*       nhAddrs     - array of next hop addrs to add
*
*   Description:
*       Callback received from FP that a previous ARPAdd method was
*       successful. This method must locate the passed in NHAddrs in the
*       specified FEID internal L2ID table and unset the pending flags.
*       It must then delete all entries in internal table that do exist
*       in passed in NHAddr table. Pass on to core method for processing.
*
*   Return:  NPF_SUCCESS or error code
*
*****************************************************************************/

NPF_RET l2idm_CommitForL2Table(
    FPPI_FEID           feid,
    uint32_t            numNHAddrs,
    NPF_IPv4Address_t*  nhAddrs)
{
    L2IDM_IPADDR_REF ipAddr;

    ipAddr.ipType     = l2idm_ipv4;
    ipAddr.ipAddrPtr  = nhAddrs;

    return CommitForL2Table_Core(
        feid,
        numNHAddrs,
        &ipAddr);
}

NPF_RET l2idm_ipv6_CommitForL2Table(
    FPPI_FEID           feid,
    uint32_t            numNHAddrs,
    NPF_IPv6Address_t*  nhAddrs)
{
    L2IDM_IPADDR_REF ipAddr;

    ipAddr.ipType     = l2idm_ipv6;
    ipAddr.ipAddrPtr  = nhAddrs;

    return CommitForL2Table_Core(
        feid,
        numNHAddrs,
        &ipAddr);
}

NPF_RET CommitForL2Table_Core(
    FPPI_FEID           feid,
    uint32_t            numNHAddrs,
    L2IDM_IPADDR_REF*   nhAddrs)

{
    NPF_RET retVal = NPF_SUCCESS;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    /* Lock the list while searching, updating. */
    PIL_EnterCriticalSection(&g_l2idMgrMasterList.listLock);

    retVal = CommitOrDeleteCore(
        l2idm_forL2Table,
        l2idm_actionCommit,
        feid,
        0,
        numNHAddrs,
        nhAddrs,
        NULL);

    PIL_LeaveCriticalSection(&g_l2idMgrMasterList.listLock);    

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/*****************************************************************************
*   Function: l2idm_GetForNextHopTable()
*
*   Parameters:
*       feid        - forwarding element ID (same as blade ID)
*       nhid        - next hop ID specified by app
*       numNHAddrs  - num next hop addrs to add
*       nhAddrs     - array of next hop addrs to add
*       numL2ids    - [out] num L2IDs returned
*       l2idList    - [out] list of L2IDs that correspond with nhAddrs
*
*   Description:
*       Locate and return the L2IDs that correspond with the NHAddrs
*       passed in. Do not alter the internal L2ID tables. Must loop
*       through all feid entries in internal L2 table.
*
*   Return:  NPF_SUCCESS or error code
*
*****************************************************************************/

NPF_RET l2idm_GetForNextHopTable(
    FPPI_NHID           nhid,
    uint32_t            numNHAddrs,
    NPF_IPv4Address_t*  nhAddrs,
    uint32_t*           numL2ids,
    FPPI_FEandL2IDs**   l2idList)
{
    L2IDM_IPADDR_REF ipAddr;

    ipAddr.ipType     = l2idm_ipv4;
    ipAddr.ipAddrPtr  = nhAddrs;

    return GetForNextHopTable_Core(
        nhid,
        numNHAddrs,
        &ipAddr,
        numL2ids,
        l2idList);
}

NPF_RET l2idm_ipv6_GetForNextHopTable(
    FPPI_NHID           nhid,
    uint32_t            numNHAddrs,
    NPF_IPv6Address_t*  nhAddrs,
    uint32_t*           numL2ids,
    FPPI_FEandL2IDs**   l2idList)
{
    L2IDM_IPADDR_REF ipAddr;

    ipAddr.ipType     = l2idm_ipv6;
    ipAddr.ipAddrPtr  = nhAddrs;

    return GetForNextHopTable_Core(
        nhid,
        numNHAddrs,
        &ipAddr,
        numL2ids,
        l2idList);
}

NPF_RET GetForNextHopTable_Core(
    FPPI_NHID           nhid,
    uint32_t            numNHAddrs,
    L2IDM_IPADDR_REF*   nhAddrs,
    uint32_t*           numL2ids,
    FPPI_FEandL2IDs**   l2idList)
{
    NPF_RET retVal                   = NPF_SUCCESS;
    DListItr* itr1                   = NULL;
    DListItr* itr2                   = NULL;
    DList* table                     = NULL;
    L2IDM_FEID_LIST_ENTRY* feidEntry = NULL;
    L2IDM_L2ID_LIST_ENTRY* l2idEntry = NULL;
    BOOL firstPass                   = TRUE;
    uint32_t outCnt;
    uint32_t i;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    PIL_EnterCriticalSection(&g_l2idMgrMasterList.listLock);

    /* Loop through internal list entries with specified NHID and FEID
       twice. The first time count the number matching entries and allocate
       memory. The second time copy the L2IDs and FEIDs into the 
       out list. */
    for (i = 0; i < 2; i++)
    {
        outCnt = 0;

        /* Loop over each FEID l2idEntry in master internal L2ID list */
        if ( (itr1 = npf_list_itrCreate(&g_l2idMgrMasterList.list)) ==
            NULL || npf_list_itrFirst(itr1) != 0)
        {
            retVal = NPF_FAIL;
            npf_logger_write(LOG_IPV4, LOG_ERR,
                "Error creating list itr - L2ID list is empty.\n");
            goto LBL_END;
        }

        do {
            if ((feidEntry = (L2IDM_FEID_LIST_ENTRY*)
                npf_list_itrGetData(itr1)) == NULL)
            {
                /* No data or itr messed up */
                npf_logger_write(LOG_IPV4, LOG_INFO,
                    "No data in this list feidEntry.\n");
                continue;
            }

            /* Locate table with specified FEID and NHID if it exists */
            if (GetTableBasedOnFEIDandNHID(l2idm_forNHTable,
                feidEntry->feid, nhid, &table, &feidEntry) != L2ID_OK)
            {
                retVal = NPF_FAIL;
                npf_logger_write(LOG_IPV4, LOG_ERR,
                    "Table does not exist.\n");
                goto LBL_END;
            }

            if ( (itr2 = npf_list_itrCreate(table)) == NULL ||
                npf_list_itrFirst(itr2) != 0)
            {
                retVal = NPF_FAIL;
                npf_logger_write(LOG_IPV4, LOG_ERR,
                    "Error creating list itr - L2ID list is empty.\n");
                goto LBL_END;
            }

            /* Now loop over all L2ID entries */
            do {
                if ((l2idEntry = (L2IDM_L2ID_LIST_ENTRY*)
                    npf_list_itrGetData(itr2)) == NULL)
                {
                    /* No data or itr messed up */
                    npf_logger_write(LOG_IPV4, LOG_INFO,
                        "No data in this list l2idEntry.\n");
                    continue;
                }

                /* For each l2idEntry, loop through passed in NHAddr list
                   to count matches in first loop and to copy l2ids in
                   second loop. Only consider non-pending entries. */
                if (l2idEntry->status == l2idm_normal)
                {
                    /* In first pass just get num NHAddrs so can
                       allocate memory. In second pass copy data. */
                    GetNhAddrFromL2List(firstPass, feidEntry->feid,
                        l2idEntry, numNHAddrs, nhAddrs, &outCnt, l2idList);
                }
    
            } while( npf_list_itrNext(itr2) != -1);
    
            /* Destroy the list iterator */
            if (itr2 != NULL)
            {
                npf_list_itrDelete(itr2);
            }
    
        } while( npf_list_itrNext(itr1) != -1);

        /* Destroy the list iterator */
        if (itr1 != NULL)
        {
            npf_list_itrDelete(itr1);
        }
    
        if (i == 0 && outCnt > 0)
        {
            /* Allocate memory for out list of old structs */
            if ((*l2idList = (FPPI_FEandL2IDs*)PIL_AllocateMemory(
                sizeof(FPPI_FEandL2IDs) * outCnt, PIL_MEM_ZERO_INIT)) ==
                NULL)
            {
                npf_logger_write(LOG_IPV4, LOG_ERR,
                    "Could not allocate memory.\n");
                retVal = NPF_OUT_OF_MEMORY;
                goto LBL_END;
            }
        } /* if (i == 0) */

        firstPass = FALSE;

    } /* for (i = 0; i < 2; i++) */

    /* Set out param */
    *numL2ids = outCnt;

LBL_END:

    PIL_LeaveCriticalSection(&g_l2idMgrMasterList.listLock);    

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");

    return retVal;

}

/*****************************************************************************
*   Function: l2idm_GetForL2Table()
*
*   Parameters:
*       feid        - forwarding element ID (same as blade ID)
*       numNHAddrs  - num next hop addrs to add
*       nhAddrs     - array of next hop addrs to add
*       numL2ids    - [out] num L2IDs returned
*       l2idList    - [out] list of L2IDs that correspond with nhAddrs
*
*   Description:
*       Locate and return the L2IDs that correspond with the NHAddrs
*       passed in. Do not alter the internal L2ID tables. Do not get if
*       pending flag is set.
*
*   Return:  NPF_SUCCESS or error code
*
*****************************************************************************/

NPF_RET l2idm_GetForL2Table(
    FPPI_FEID           feid,
    uint32_t            numNHAddrs,
    NPF_IPv4Address_t*  nhAddrs,
    uint32_t*           numL2ids,
    FPPI_L2ID**         l2idList)
{
    L2IDM_IPADDR_REF ipAddr;

    ipAddr.ipType     = l2idm_ipv4;
    ipAddr.ipAddrPtr  = nhAddrs;

    return GetForL2Table_Core(
    feid,
    numNHAddrs,
    &ipAddr,
    numL2ids,
    l2idList);
}

NPF_RET l2idm_ipv6_GetForL2Table(
    FPPI_FEID           feid,
    uint32_t            numNHAddrs,
    NPF_IPv6Address_t*  nhAddrs,
    uint32_t*           numL2ids,
    FPPI_L2ID**         l2idList)
{
    L2IDM_IPADDR_REF ipAddr;

    ipAddr.ipType     = l2idm_ipv6;
    ipAddr.ipAddrPtr  = nhAddrs;

    return GetForL2Table_Core(
    feid,
    numNHAddrs,
    &ipAddr,
    numL2ids,
    l2idList);
}


NPF_RET GetForL2Table_Core(
    FPPI_FEID           feid,
    uint32_t            numNHAddrs,
    L2IDM_IPADDR_REF*   nhAddrs,
    uint32_t*           numL2ids,
    FPPI_L2ID**         l2idList)
{
    NPF_RET retVal = NPF_SUCCESS;
    DList* table   = NULL;
    L2IDM_FEID_LIST_ENTRY* fiedEntry = NULL;
    BOOL isLocked = FALSE;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    /* Check in params */
    if (numNHAddrs == 0 || nhAddrs->ipAddrPtr == NULL)
    {
        retVal = NPF_INVALID_PARAMETERS;
        npf_logger_write(LOG_IPV4, LOG_ERR,
            "Invalid parms into L2IDmgr - exiting with status %d.\n",
             retVal);
        goto LBL_END;
    }

    /* Init out params */
    *numL2ids = 0;
    *l2idList = NULL;

    /* Lock the list while searching, updating. */
    PIL_EnterCriticalSection(&g_l2idMgrMasterList.listLock);
    isLocked = TRUE;

    /* Locate table with specified FEID and NHID if it exists */
    if (GetTableBasedOnFEIDandNHID(l2idm_forL2Table, feid, 0,
        &table, &fiedEntry) != L2ID_OK)
    {
        retVal = NPF_FAIL;
        npf_logger_write(LOG_IPV4, LOG_ERR,
            "Table does not exist.\n");
        goto LBL_END;
    }

    /* Step through list and return L2ID entries found in passed in
       NHAddrs list. */
    retVal = RetrieveL2IDs(numNHAddrs, nhAddrs, table, numL2ids, l2idList);

LBL_END:

    if (isLocked)
    {
        PIL_LeaveCriticalSection(&g_l2idMgrMasterList.listLock);    
    }

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/*****************************************************************************
*   Function: l2idm_DeleteForNextHopTable()
*
*   Parameters:
*       nhid        - next hop ID specified by app
*       numNHAddrs  - num next hop addrs to add
*       nhAddrs     - array of next hop addrs to add
*
*   Description:
*       Locate and delete the NHAddr/L2ID entries in the internal L2ID
*       table that corresponds to the specifed FEID and NHID that match
*       NHAddrs in the passed in list. Do not delete if pending flag is set.
*
*   Return:  NPF_SUCCESS or error code
*
*****************************************************************************/

NPF_RET l2idm_DeleteForNextHopTable(
    FPPI_NHID           nhid,
    uint32_t            numNHAddrs,
    NPF_IPv4Address_t*  nhAddrs)
{
    L2IDM_IPADDR_REF ipAddr;

    ipAddr.ipType     = l2idm_ipv4;
    ipAddr.ipAddrPtr  = nhAddrs;

    return DeleteForNextHopTable_Core(
        nhid,
        numNHAddrs,
        &ipAddr);
}

NPF_RET l2idm_ipv6_DeleteForNextHopTable(
    FPPI_NHID           nhid,
    uint32_t            numNHAddrs,
    NPF_IPv6Address_t*  nhAddrs)
{
    L2IDM_IPADDR_REF ipAddr;

    ipAddr.ipType     = l2idm_ipv6;
    ipAddr.ipAddrPtr  = nhAddrs;

    return DeleteForNextHopTable_Core(
        nhid,
        numNHAddrs,
        &ipAddr);
}

NPF_RET DeleteForNextHopTable_Core(
    FPPI_NHID           nhid,
    uint32_t            numNHAddrs,
    L2IDM_IPADDR_REF*   nhAddrs)
{
    NPF_RET retVal      = NPF_SUCCESS;
    DListItr * itr      = NULL;
    L2IDM_FEID_LIST_ENTRY* p1Entry;
    BOOL continueLoop   = TRUE;
    BOOL isNodeDeleted;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    /* Lock the list while searching, updating. */
    PIL_EnterCriticalSection(&g_l2idMgrMasterList.listLock);

    while (continueLoop)
    {
        isNodeDeleted = FALSE;

        if ( (itr = npf_list_itrCreate(&g_l2idMgrMasterList.list)) == NULL ||
            npf_list_itrFirst(itr) != 0)
        {
            /* List is probably empty */
            retVal = NPF_FAIL;
            npf_logger_write(LOG_IPV4, LOG_ERR,
                "Error - list is empty for commit.\n");
            goto LBL_END;
        }

        do {
            if ((p1Entry = (L2IDM_FEID_LIST_ENTRY*)
                npf_list_itrGetData(itr)) == NULL)
            {
                /* No data or itr messed up */
                retVal = L2ID_NO_FEID_ENTRY;
                npf_logger_write(LOG_IPV4, LOG_INFO,
                    "No data in this list entry.\n");
                continue;
            }

            retVal = CommitOrDeleteCore(
                l2idm_forNHTable,
                l2idm_actionDelete,
                p1Entry->feid,
                nhid,
                numNHAddrs,
                nhAddrs,
                &isNodeDeleted);

            if (retVal != NPF_SUCCESS)
            {
                continueLoop = FALSE;
                break;
            }

            if (isNodeDeleted)
            {
                /* Must destroy iterator and recreate */
                break;
            }

        } while( npf_list_itrNext(itr) != -1);

        if (!isNodeDeleted)
        {
            continueLoop = FALSE;
        }
        else
            break;

    } /* while (continueLoop) */

LBL_END:

    /* Destroy the list iterator */
    if (itr != NULL)
    {
        npf_list_itrDelete(itr);
    }

    PIL_LeaveCriticalSection(&g_l2idMgrMasterList.listLock);    

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/*****************************************************************************
*   Function: l2idm_DeleteForL2Table()
*
*   Parameters:
*       feid        - forwarding element ID (same as blade ID)
*       numNHAddrs  - num next hop addrs to add
*       nhAddrs     - array of next hop addrs to add
*
*   Description:
*       Locate and delete the NHAddr/L2ID entries in the internal L2ID
*       table that corresponds to the specifed FEID that match NHAddrs
*       in the passed in list. Do not delete if pending flag is set.
*
*   Return:  NPF_SUCCESS or error code
*
*****************************************************************************/

NPF_RET l2idm_DeleteForL2Table(
    FPPI_FEID           feid,
    uint32_t            numNHAddrs,
    NPF_IPv4Address_t*  nhAddrs)
{
    L2IDM_IPADDR_REF ipAddr;

    ipAddr.ipType     = l2idm_ipv4;
    ipAddr.ipAddrPtr  = nhAddrs;

    return DeleteForL2Table_Core(
        feid,
        numNHAddrs,
        &ipAddr);
}

NPF_RET l2idm_ipv6_DeleteForL2Table(
    FPPI_FEID           feid,
    uint32_t            numNHAddrs,
    NPF_IPv6Address_t*  nhAddrs)
{
    L2IDM_IPADDR_REF ipAddr;

    ipAddr.ipType     = l2idm_ipv6;
    ipAddr.ipAddrPtr  = nhAddrs;

    return DeleteForL2Table_Core(
        feid,
        numNHAddrs,
        &ipAddr);
}

NPF_RET DeleteForL2Table_Core(
    FPPI_FEID           feid,
    uint32_t            numNHAddrs,
    L2IDM_IPADDR_REF*   nhAddrs)
{
    NPF_RET retVal = NPF_SUCCESS;
    BOOL temp;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    /* Lock the list while searching, updating. */
    PIL_EnterCriticalSection(&g_l2idMgrMasterList.listLock);

    retVal = CommitOrDeleteCore(
        l2idm_forL2Table,
        l2idm_actionDelete,
        feid,
        0,
        numNHAddrs,
        nhAddrs,
        &temp);

    PIL_LeaveCriticalSection(&g_l2idMgrMasterList.listLock);    

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/*****************************************************************************
*   Function: HandleOldEntries()
*
*   Parameters:
*       nhid        - next hop ID specified by app
*       countIn     - num elements in feids list
*       feids       - forwarding element ID (same as blade ID)
*       numOldL2ids - [out] num old L2IDs added
*       oldL2idList - [out] list of old L2IDs added
*
*   Description:
*       Locate appropriate internal L2 table. Scan table NHAddr entries
*       with the pending flag set that match those in the passed in
*       NHAddr list. If found, unset the pending flag.
*
*   Return:  NPF_SUCCESS 
*
*****************************************************************************/

NPF_RET HandleOldEntries(
    FPPI_NHID          nhid,
    uint32_t           countIn,
    FPPI_FEID*         feids,
    uint32_t*          numOldL2ids,
    FPPI_FEandL2IDs**  oldL2idList)
{
    NPF_RET retVal                   = NPF_SUCCESS;
    DListItr* itr                    = NULL;
    DList* table                     = NULL;
    L2IDM_FEID_LIST_ENTRY* fiedEntry = NULL;
    L2IDM_L2ID_LIST_ENTRY* entry     = NULL;
    uint32_t outCnt;
    uint32_t i, j;

    PIL_EnterCriticalSection(&g_l2idMgrMasterList.listLock);

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    /* Loop through internal list entries with specified NHID and FEID
       twice. The first time count the number of old entries and allocate
       memory. The second time copy the L2IDs and FEIDs into the new
       out list. */
    for (i = 0; i < 2; i++)
    {
        outCnt = 0;

        /* Now loop over all input <feid,l2id> pairs */
        for (j = 0; j < countIn; j++)
        {
            /* Locate table with specified FEID and NHID if it exists */
            if (GetTableBasedOnFEIDandNHID(l2idm_forNHTable, feids[j], nhid,
                &table, &fiedEntry) != L2ID_OK)
            {
                retVal = NPF_FAIL;
                npf_logger_write(LOG_IPV4, LOG_ERR,
                    "Table does not exist.\n");
                goto LBL_END;
            }
    
            if ( (itr = npf_list_itrCreate(table)) == NULL ||
                npf_list_itrFirst(itr) != 0)
            {
                retVal = NPF_FAIL;
                npf_logger_write(LOG_IPV4, LOG_ERR,
                    "Error creating list itr - L2ID list is empty.\n");
                goto LBL_END;
            }

            do {
                if ((entry = (L2IDM_L2ID_LIST_ENTRY*)
                    npf_list_itrGetData(itr)) == NULL)
                {
                    /* No data or itr messed up */
                    npf_logger_write(LOG_IPV4, LOG_INFO,
                        "No data in this list entry.\n");
                    continue;
                }

                /* If old entry, increment cntr or copy data to out list. */
                if (entry->status == l2idm_old)
                {
                    if (i == 0)
                    {
                        /* First pass so increment counter */
                        outCnt++;
                    }
                    else {
                        /* Copy L2ID and FEID to out list */
                        (*oldL2idList)[outCnt].l2id = entry->l2id;
                        (*oldL2idList)[outCnt].feid = feids[j];

                        (*numOldL2ids)++;
                        outCnt++;
    
                    } /* if (i == 0) */
                } /* if (entry->status != l2idm_pending) */
                else if (entry->status == l2idm_normal_secure)
                {
                    entry->status = l2idm_normal;
                }
    
            } while( npf_list_itrNext(itr) != -1);
    
            /* Destroy the list iterator */
            if (itr != NULL)
            {
                npf_list_itrDelete(itr);
            }
    
        } /* for (j = 0; j < countIn; j++) */

        if (i == 0 && outCnt > 0)
        {
            /* Allocate memory for out list of old structs */
            if ((*oldL2idList = (FPPI_FEandL2IDs*)PIL_AllocateMemory(
                sizeof(FPPI_FEandL2IDs) * outCnt, PIL_MEM_ZERO_INIT)) ==
                NULL)
            {
                npf_logger_write(LOG_IPV4, LOG_ERR,
                    "Could not allocate memory.\n");
                retVal = NPF_OUT_OF_MEMORY;
                goto LBL_END;
            }
        } /* if (i == 0) */

    } /* for (i = 0; i < 2; i++) */

LBL_END:

    PIL_LeaveCriticalSection(&g_l2idMgrMasterList.listLock);

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

NPF_RET FlushNextHops()
{
    return NPF_SUCCESS;
}
/*****************************************************************************
*   Function: CommitOrDeleteCore()
*
*   Parameters:
*       tblType    - update is for NH or L2 table on data plane
*       action      - identify if this is a commit or delete
*       feid        - forwarding element ID (same as blade ID)
*       nhid        - next hop ID specified by app
*       numNHAddrs  - num next hop addrs to add
*       nhAddrs     - array of next hop addrs to add
*
*   Description:
*       Locate appropriate internal L2 table. Scan table NHAddr entries
*       and: for commit set the pending flag to normal if NHAddr matches
*       with passed in NHAddr list; for delete remove the entry from the
*       list (ignore if pending flag set).
*
*   Return:  NPF_SUCCESS 
*
*****************************************************************************/

NPF_RET CommitOrDeleteCore(
    L2IDM_FOR_TABLE_TYPE    tblType,
    L2IDM_ACTION_TYPE       action,
    FPPI_FEID               feid,
    FPPI_NHID               nhid,
    uint32_t                numNHAddrs,
    L2IDM_IPADDR_REF*       nhAddrs,
    BOOL*                   isDeleted)
{
    NPF_RET retVal = NPF_SUCCESS;
    DList* table   = NULL;
    L2IDM_FEID_LIST_ENTRY* fiedEntry = NULL;
    BOOL isLocked = FALSE;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    /* Check in params */
    if (numNHAddrs == 0 || nhAddrs->ipAddrPtr == NULL)
    {
/*        if(action == l2idm_actionAdd)
        {
            retVal = FlushNextHops();
            goto LBL_END;
        }
*/     
        retVal = NPF_INVALID_PARAMETERS;
        npf_logger_write(LOG_IPV4, LOG_ERR,
            "Invalid parameters to L2IDmgr - exiting with status %d.\n",
             retVal);
        goto LBL_END;
    }

    /* Lock the list while searching, updating. */
    PIL_EnterCriticalSection(&g_l2idMgrMasterList.listLock);
    isLocked = TRUE;

    /* Locate table with specified FEID and NHID if it exists */
    if (GetTableBasedOnFEIDandNHID(tblType, feid, nhid,
        &table, &fiedEntry) != L2ID_OK)
    {
        retVal = NPF_FAIL;
        npf_logger_write(LOG_IPV4, LOG_ERR,
            "Table does not exist.\n");
        goto LBL_END;
    }

    /* Step through list and remove pending flag on all entries that
       also exist in passed in nhAddrs list. */
    if ( (retVal = CommitOrDeleteEntriesInL2IDTable(
        tblType, action, numNHAddrs, nhAddrs, table)) != NPF_SUCCESS)
    {
        goto LBL_END;
    }

    /* The only time the parent lists don't need to be checked
       if they are empty is if the action is commit for L2 table.
       (Note that commit for NH table may delete all entries in a
       NHID table and therefore require removal.) In this case no
       list items can be deleted so a list empty check would be
       frivolous. */

    if (tblType == l2idm_forNHTable && action == l2idm_actionDelete)
    {
        /* First check if table is empty */
        if (npf_list_size(table) == 0)
        {
            /* L2 table is empty - now check who parent is */
            if (tblType == l2idm_forNHTable)
            {
                /* This is the only list in the entry so destroy list */
                npf_list_destroy(table);

                /* Remove entry from parent */
                npf_list_removeByKeyFree(&fiedEntry->nhidList,
                    (void *)&nhid);
            }

            /* Now check to see if parent should be deleted from
               master list */
            CheckDeleteEntryFromMasterList(feid, fiedEntry, isDeleted);

        } /* if (npf_list_size(table) == 0) */

    } /* if (action != l2idm_actionCommit || ... */

LBL_END:

    if (isLocked)
    {
        PIL_LeaveCriticalSection(&g_l2idMgrMasterList.listLock);    
    }

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/*****************************************************************************
*   Function: GetNhAddrFromL2List()
*
*   Parameters:
*       firstPass   - flag indicating if this is first or second invocation
*       feid        - forwarding element ID (same as blade ID)
*       entry       - internal L2ID table entry
*       numNHAddrs  - num next hop addrs to add
*       nhAddrs     - array of next hop addrs to add
*       outCnt      - [out] count of matches found, array incrementer
*       l2idList    - [out] list of <feid,l2id> pairs
*
*   Description:
*       Search through passed in newNHAddr list for match with passed in
*       L2ID list entry. If found, increment counter if first pass or put
*       <feid,l2id> pair into out list of second pass.
*
*   Return: none
*
*****************************************************************************/

void GetNhAddrFromL2List(
    BOOL                    firstPass,
    FPPI_FEID               feid,
    L2IDM_L2ID_LIST_ENTRY*  entry,
    uint32_t                numNHAddrs,
    L2IDM_IPADDR_REF*       nhAddrs,
    uint32_t*               outCnt,
    FPPI_FEandL2IDs**       l2idList)
{
    uint32_t i;
    BOOL found = FALSE;
    size_t addrSz;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    /* Set address size */
    addrSz = nhAddrs->ipType == l2idm_ipv4 ? IPV4_ADDR_SZ : IPV6_ADDR_SZ;

    for (i = 0; i < numNHAddrs; i++)
    {
        if (nhAddrs->ipType == entry->nhAddr.ipType) 
        {

            found = TRUE;
            memcpy( (((void*)nhAddrs->ipAddrPtr) + (i * addrSz)),
                     &(entry->nhAddr.addr), addrSz); 

            if (firstPass)
            {
                /* Just increment new NHaddr counter */
                (*outCnt)++;
            }
            else {
                /* This NHAddr already exists in L2 table, put the
                   corresponding L2ID in the new L2ID out list. */
                (*l2idList)[*outCnt].l2id = entry->l2id;
                (*l2idList)[*outCnt].feid = feid;
                (*outCnt)++;
            }

            /* Break out of loop since it has been found */
            break;

        } /* if (nhAddrs->ipType == entry->nhAddr.ipType && ... */

    } /* for (i = 0; i < numNHAddrs; i++) */

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");
}

/*****************************************************************************
*   Function: CleanupAList()
*
*   Parameters: list  - ptr to DList
*
*   Description:
*       Shutdown the passed in list by removing and freeing any list
*       entries then destroying the list.
*
*   Return: none
*
*****************************************************************************/
#if 0
/* The npf_list_destroy() method does this for us */
void CleanupAList(DList* list)
{
    DListItr * itr = NULL;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    if ( (itr = npf_list_itrCreate(list)) == NULL)
    {
        /* List is probably empty */
        goto LBL_END;
    }

    if(npf_list_itrFirst(itr) != -1)
    {
        do {
            /* Free each entry in the list */
            npf_list_itrRemoveFree(itr);
        } while( npf_list_itrNext(itr) != -1);
    }

    /* Destroy the list iterator */
    iif (itr != NULL)
    {
        npf_list_itrDelete(itr);
    }

LBL_END:

    /* Destroy the list */
    npf_list_destroy(list);

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");
}
#endif
/*****************************************************************************
*   Function: GetTableBasedOnFEIDandNHID()
*
*   Parameters:
*       tblType  - update is for NH or L2 table on data plane
*       feid      - forwarding element ID (same as blade ID)
*       nhid      - next hop ID specified by app
*       table     - [out] ptr to ptr to specific list
*       feidEntry - [out] ptr to ptr to feid table
*
*   Description:
*       Find the internal table for the specified feid and nhid. If the
*       table does not exist, the out table param is NULL and the
*       appropriate error code is returned indicating which table is
*       missing.
*
*   Return: L2ID_OK, L2ID_NO_FEID_ENTRY or L2ID_NO_NHID_ENTRY
*
*****************************************************************************/

uint32_t GetTableBasedOnFEIDandNHID(
    L2IDM_FOR_TABLE_TYPE     tblType,
    FPPI_FEID                feid,
    FPPI_NHID                nhid,
    DList**                  table,
    L2IDM_FEID_LIST_ENTRY**  feidEntry)
{
    uint32_t retVal     = L2ID_OK;
    DListItr* itr1      = NULL;
    DListItr* itr2      = NULL;
    BOOL foundFeidTbl   = FALSE;
    BOOL foundNhidTbl   = FALSE;
    L2IDM_FEID_LIST_ENTRY* p1Entry;
    L2IDM_NHID_LIST_ENTRY* p2Entry;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    /* Set out param */
    *table = NULL;

    if ( (itr1 = npf_list_itrCreate(&g_l2idMgrMasterList.list)) == NULL ||
        npf_list_itrFirst(itr1) != 0)
    {
        /* List is probably empty */
        retVal = L2ID_NO_FEID_ENTRY;
        goto LBL_END;
    }

    do {

        if ((p1Entry = (L2IDM_FEID_LIST_ENTRY*)
            npf_list_itrGetData(itr1)) == NULL)
        {
            /* No data or itr messed up */
            retVal = L2ID_NO_FEID_ENTRY;
            npf_logger_write(LOG_IPV4, LOG_INFO,
                "No data in this list entry.\n");
            continue;
        }

        if (p1Entry->feid == feid)
        {
            /* Set out params to this FEID table */
            *table       = &(p1Entry->l2idList);
            *feidEntry   = p1Entry;
            foundFeidTbl = TRUE;

            if (tblType == l2idm_forNHTable)
            {
                /* Now look in the contained NHID entries for a match
                   with the passed in NHID. */
                if ( (itr2 = npf_list_itrCreate(&p1Entry->nhidList)) == NULL)
                {
                    /* List is probably empty */
                    retVal = L2ID_NO_NHID_ENTRY;
                    goto LBL_END;
                }

                if (npf_list_itrFirst(itr2) != 0)
                {
                    /* List is probably empty */
                    retVal = L2ID_NO_NHID_ENTRY;
                    npf_list_itrDelete(itr2);
                    goto LBL_END;
                }

//                if (npf_list_itrFirst(itr2) != 0)

                do {

                    if ((p2Entry = (L2IDM_NHID_LIST_ENTRY *)
                        npf_list_itrGetData(itr2)) == NULL)
                    {
                        /* No data or itr messed up */
                        npf_logger_write(LOG_IPV4, LOG_INFO,
                            "No data in this list entry.\n");
                        continue;
                    }

                    if (p2Entry->nhid == nhid)
                    {
                        /* Set out param to this NHID table */
                        *table = &(p2Entry->l2idList);
                        foundNhidTbl = TRUE;
                    }
                } while( foundNhidTbl == FALSE && npf_list_itrNext(itr2)
                      != -1);

                /* Destroy the list iterator */
                if (itr2 != NULL)
                {
                    npf_list_itrDelete(itr2);
                }

                if (foundNhidTbl == FALSE)
                {
                    retVal = L2ID_NO_NHID_ENTRY;
                }
            } /* if (tblType == l2idm_forNHTable) */

        } /* if (p1Entry->feid == feid) */

    } while( foundFeidTbl == FALSE && npf_list_itrNext(itr1) != -1);

    if (foundFeidTbl == FALSE)
    {
        retVal = L2ID_NO_FEID_ENTRY;
    }

LBL_END:

    /* Destroy the list iterator */
    if (itr1 != NULL)
    {
        npf_list_itrDelete(itr1);
    }

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/*****************************************************************************
*   Function: CreateAndFillNewFEIDList()
*
*   Parameters:
*       tblType    - update is for NH or L2 table on data plane
*       feid        - forwarding element ID (same as blade ID)
*       nhid        - next hop ID specified by app
*       numNHAddrs  - num next hop addrs to add
*       nhAddrs     - array of next hop addrs to add
*       numNewL2ids - [out] num new L2IDs added
*       newL2idList - [out] list of new L2IDs added
*
*   Description:
*       Create and fill new FEID structure. Create and fill new NHID
*       structure if tblType is l2idm_forNHTable. Then create new L2ID
*       structure for each new NHaddr to be added. Fill out params
*       accordingly.
*
*   Return: NPF_SUCCESS or error code
*
*****************************************************************************/

NPF_RET CreateAndFillNewFEIDList(
    L2IDM_FOR_TABLE_TYPE    tblType,
    FPPI_FEID               feid,
    FPPI_NHID               nhid,
    uint32_t                numNHAddrs,
    L2IDM_IPADDR_TYPE       addrType,
    void*                   nhAddrs,
    uint32_t*               numNewL2ids,
    FPPI_L2ID**             newL2idList)
{
    NPF_RET retVal = NPF_SUCCESS;
    L2IDM_FEID_LIST_ENTRY* fiedEntry = NULL;
    L2IDM_NHID_LIST_ENTRY* nhidEntry = NULL;
    DList* tmpList;
    BOOL isLocked                    = FALSE;
    BOOL addedNHEntry                = FALSE;
    BOOL addedFEEntry                = FALSE;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    /* Allocate new feid structure */
    if ((fiedEntry = (L2IDM_FEID_LIST_ENTRY*)PIL_AllocateMemory(
        sizeof(L2IDM_FEID_LIST_ENTRY), PIL_MEM_ZERO_INIT)) == NULL)
    {
        npf_logger_write(LOG_IPV4, LOG_ERR, "Could not allocate memory.\n");
        retVal = NPF_OUT_OF_MEMORY;
        goto LBL_END;
    }

    /* Init structure elements */
    fiedEntry->feid            = feid;
    fiedEntry->l2idIncrementor = 1;
    npf_list_init(&(fiedEntry->nhidList), PIL_FreeMemory);
    fiedEntry->nhidList.compare = &compareNHIDEntry;
    npf_list_init(&(fiedEntry->l2idList), PIL_FreeMemory);

    if (tblType == l2idm_forNHTable)
    {
        if ( (retVal = CreateAndInitNHIDObj(nhid, &nhidEntry)) !=
            NPF_SUCCESS)
        {
            goto LBL_END;
        }

        /* Add the NHID entry to the list in the FEID entry */
        if (npf_list_pushFront(&(fiedEntry->nhidList), nhidEntry) != 0)
        {
            retVal = NPF_FAIL;
            npf_logger_write(LOG_IPV4, LOG_ERR,
                "Error pushing onto list; ret = %d.\n", retVal);
            goto LBL_END;
        }

        addedNHEntry = TRUE;

        tmpList = &(nhidEntry->l2idList);
    }
    else {
        tmpList = &(fiedEntry->l2idList);
    }

    /* Create a new L2ID list entry for each new NHaddr and insert
       into appropriate list. */
    if ( (retVal = AddNHAddrInfoToList(fiedEntry, tmpList, numNHAddrs,
        addrType, nhAddrs, numNewL2ids, newL2idList)) != NPF_SUCCESS)
    {
        goto LBL_END;
    }

    PIL_EnterCriticalSection(&g_l2idMgrMasterList.listLock);
    isLocked = TRUE;

    /* Now insert the FEID entry into the master global list */
    if (npf_list_pushFront(&(g_l2idMgrMasterList.list), fiedEntry) != 0)
    {
        retVal = NPF_FAIL;
        npf_logger_write(LOG_IPV4, LOG_ERR,
            "Error pushing onto list; ret = %d.\n", retVal);
    }

    addedFEEntry = TRUE;

LBL_END:

    if (retVal != NPF_SUCCESS)
    {
        if (addedNHEntry)
        {
            npf_list_popFrontFree(&(fiedEntry->nhidList));
        }
        else if (nhidEntry != NULL)
        {
            /* Not in list but mem is allocated */
            PIL_FreeMemory(nhidEntry);
        }

        npf_list_destroy(&(fiedEntry->nhidList));
        npf_list_destroy(&(fiedEntry->l2idList));

        if (addedFEEntry)
        {
            npf_list_popFrontFree(&(g_l2idMgrMasterList.list));
        }
        else if (fiedEntry != NULL)
        {
            /* Not in list but mem is allocated */
            PIL_FreeMemory(fiedEntry);
        }
    }

    if (isLocked)
    {
        PIL_LeaveCriticalSection(&g_l2idMgrMasterList.listLock);    
    }

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/*****************************************************************************
*   Function: CreateAndInitNHIDObj()
*
*   Parameters:
*       nhid      - next hop ID specified by app
*       nhidEntry - [out] ptr to new object
*
*   Description:
*       Create and init new NHID structure.
*
*   Return: NPF_SUCCESS or error code
*
*****************************************************************************/

NPF_RET CreateAndInitNHIDObj(
    FPPI_NHID  nhid,
    L2IDM_NHID_LIST_ENTRY** nhidEntry)
{
    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    /* Allocate new nhid structure */
    if ((*nhidEntry = (L2IDM_NHID_LIST_ENTRY*)PIL_AllocateMemory(
        sizeof(L2IDM_NHID_LIST_ENTRY), PIL_MEM_ZERO_INIT)) == NULL)
    {
        npf_logger_write(LOG_IPV4, LOG_ERR, "Could not allocate memory.\n");
        return NPF_OUT_OF_MEMORY;
    }

    /* Init structure elements */
    (*nhidEntry)->nhid  = nhid;
    npf_list_init(&((*nhidEntry)->l2idList), PIL_FreeMemory);

    npf_logger_write(LOG_IPV4, LOG_TRACE, "Leaving method.\n");

    return NPF_SUCCESS;
}

/*****************************************************************************
*   Function: AddNHAddrInfoToList()
*
*   Parameters:
*       fiedEntry   - ptr to FEID obj to access l2idIncrementor
*       list        - ptr to list in which to add new L2ID entries
*       numNHAddrs  - num next hop addrs to add
*       nhAddrs     - array of next hop addrs to add
*       numNewL2ids - [out] num new L2IDs added
*       newL2idList - [out] list of new L2IDs added
*
*   Description:
*       Create and fill new L2ID structures, one per nhAddrs. Put new
*       objects into passed in list (FEID obj list for L2 table add or NHID
*       obj list for NH table add). Allocate memory for newL2idList out
*       param and fill. On error, all entries that were processed correctly
*       are still returned to caller.
*
*   Return: NPF_SUCCESS or error code
*
*****************************************************************************/

NPF_RET AddNHAddrInfoToList(
    L2IDM_FEID_LIST_ENTRY*  fiedEntry,
    DList*                  list,
    uint32_t                numNHAddrs,
    L2IDM_IPADDR_TYPE       addrType,
    void*                   nhAddrs,
    uint32_t*               numNewL2ids,
    FPPI_L2ID**             newL2idList)
{
    NPF_RET                 retVal = NPF_SUCCESS;
    uint32_t                i;
    L2IDM_L2ID_LIST_ENTRY*  l2idEntry;
    size_t                  addrSz;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    *numNewL2ids = 0;
    *newL2idList = NULL;

    /* Set address size */
    addrSz = addrType == l2idm_ipv4 ? IPV4_ADDR_SZ : IPV6_ADDR_SZ;

    /* Allocate memory for new L2ID out list */
    if ((*newL2idList = (FPPI_L2ID*)PIL_AllocateMemory(
        sizeof(FPPI_L2ID) * numNHAddrs, PIL_MEM_ZERO_INIT)) == NULL)
    {
        npf_logger_write(LOG_IPV4, LOG_ERR,
            "Could not allocate memory.\n");
        retVal = NPF_OUT_OF_MEMORY;
        goto LBL_END;
    }

    for (i = 0; i < numNHAddrs; i++)
    {
        /* Allocate new L2ID list entry */
        if ((l2idEntry = (L2IDM_L2ID_LIST_ENTRY*)PIL_AllocateMemory(
            sizeof(L2IDM_L2ID_LIST_ENTRY), PIL_MEM_ZERO_INIT)) == NULL)
        {
            npf_logger_write(LOG_IPV4, LOG_ERR,
                "Could not allocate memory.\n");
            retVal = NPF_OUT_OF_MEMORY;
            goto LBL_END;
        }

        /* Fill entry values */
        l2idEntry->l2id     = (fiedEntry->l2idIncrementor)++;
        l2idEntry->status   = l2idm_pending;
        l2idEntry->nhAddr.ipType = addrType;
        if(nhAddrs)
        {
            memcpy(&(l2idEntry->nhAddr.addr),
               &((NPF_IPv4Address_t*)nhAddrs)[i], addrSz);
        }
        /* Insert entry into passed in list */
        if (npf_list_pushFront(list, l2idEntry) != 0)
        {
            retVal = NPF_FAIL;
            npf_logger_write(LOG_IPV4, LOG_ERR,
                "Error pushing onto list; ret = %d.\n", retVal);
            goto LBL_END;
        }

        /* Add entry into L2ID out list */
        (*newL2idList)[i] = l2idEntry->l2id;

        /* Increment out param */
        (*numNewL2ids)++;

    } /* for (i = 0; i < numNHAddrs; i++) */

LBL_END:

    if (*newL2idList != NULL && numNewL2ids == 0)
    {
        /* The out list is empty so delete and reset out ptr to NULL */
        PIL_FreeMemory(*newL2idList);
        *newL2idList = NULL;
    }

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/*****************************************************************************
*   Function: UpdateL2IDTableForL2()
*
*   Parameters:
*       numNHAddrs  - num next hop addrs to add
*       nhAddrs     - array of next hop addrs to add
*       fiedEntry   - ptr to master list entry for this FEID
*       l2Tbl       - ptr to low level L2ID table to edit
*       numNewL2ids - [out] num new L2IDs added
*       newL2idList - [out] list of new L2IDs added
*
*   Description:
*       Compare the new NHAddrs with those in the internal L2ID list. Add
*       new ones that are not already in internal list and set the pending
*       flag. Return all L2IDs of those NHaddrs that are now in list that
*       do not have the pending flag set. Return list of old L2IDs that
*       are in internal list but not passed in nhAddrs list.
*
*   Return: none
*
*****************************************************************************/

NPF_RET UpdateL2IDTableForL2(
    uint32_t                numNHAddrs,
    L2IDM_IPADDR_REF*       nhAddrs,
    L2IDM_FEID_LIST_ENTRY*  fiedEntry,
    DList*                  l2Tbl,
    uint32_t*               numNewL2ids,
    FPPI_L2ID**             newL2idList)
{
    NPF_RET retVal        = NPF_SUCCESS;
    uint32_t* tmpFlags    = NULL;
    DListItr * itr        = NULL;
    FPPI_L2ID* tmpNewList = NULL;
    uint32_t i, j;
    L2IDM_L2ID_LIST_ENTRY* entry;
    BOOL firstPass = TRUE;
    uint32_t newOutListIncr;
    size_t addrSz;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    /* Allocate memory for temporary array */
    if ((tmpFlags = (uint32_t*)PIL_AllocateMemory(
        sizeof(uint32_t) * numNHAddrs, PIL_MEM_ZERO_INIT)) == NULL)
    {
        npf_logger_write(LOG_IPV4, LOG_ERR,
            "Could not allocate memory.\n");
        retVal = NPF_OUT_OF_MEMORY;
        goto LBL_END;
    }

    /* Set address size */
    addrSz = nhAddrs->ipType == l2idm_ipv4 ? IPV4_ADDR_SZ : IPV6_ADDR_SZ;

    for (i = 0; i < 2; i++)
    {
        /* Reset incrementor */
        newOutListIncr = 0;

        /* Loop through list of current NHaddrs in internal table */
        if ( (itr = npf_list_itrCreate(l2Tbl)) != NULL &&
           npf_list_itrFirst(itr) == 0)
        {
            /* Loop through all entries in internal L2ID list and compare
               with all entries in passed in NHAddr list. */
            do {
                if ((entry = (L2IDM_L2ID_LIST_ENTRY*)
                    npf_list_itrGetData(itr)) == NULL)
                {
                    /* No data or itr messed up */
                    npf_logger_write(LOG_IPV4, LOG_INFO,
                        "No data in this list entry.\n");
                    continue;
                }

                /* Do not consider pending entries since they don't for
                   sure exist yet in data plane tables. */
                if (entry->status == l2idm_normal)
                {
                    /* In first pass just get num new and old NHAddrs so can
                       allocate memory. In second pass copy data. */
                    ScanForNhAddrInL2List(firstPass, entry, numNHAddrs,
                        nhAddrs, &newOutListIncr, tmpFlags, numNewL2ids,
                        tmpNewList);
                }

            } while( npf_list_itrNext(itr) != -1);

            /* Destroy the list iterator */
            if (itr != NULL)
            {
                npf_list_itrDelete(itr);
            }
        } /* if ( (itr = npf_list_itrCreate(l2Tbl)) != NULL && ... */

        if (firstPass == TRUE)
        {
            /* Count number of new NHAddrs not marked and add to new list.
               These are new ones to be added to internal table. */
            for (j = 0; j < numNHAddrs; j++)
            {
                if (tmpFlags[j] == 0)
                {
                    (*numNewL2ids)++;
                }
            }

            /* Allocate memory for out lists */
            if (*(numNewL2ids) > 0 &&
               (*newL2idList = tmpNewList = (FPPI_L2ID*)
                PIL_AllocateMemory(sizeof(FPPI_L2ID) * *(numNewL2ids),
                PIL_MEM_ZERO_INIT)) == NULL)
            {
                npf_logger_write(LOG_IPV4, LOG_ERR,
                    "Could not allocate memory.\n");
                retVal = NPF_OUT_OF_MEMORY;
                goto LBL_END;
            }

        } /* if (firstPass == TRUE) */

        firstPass = FALSE;

    } /* for (i = 0; i < 2; i++) */

    /* Scan through flag list and add new L2ID entries for each
       flag entry that is not marked. These are new ones to be
       added to internal table. */
    for (j = 0; j < numNHAddrs; j++)
    {
        if (tmpFlags[j] == 0)
        {
            /* Allocate memory for new L2ID object */
            if ((entry = (L2IDM_L2ID_LIST_ENTRY*)PIL_AllocateMemory(
                sizeof(L2IDM_L2ID_LIST_ENTRY), PIL_MEM_ZERO_INIT)) == NULL)
            {
                npf_logger_write(LOG_IPV4, LOG_ERR,
                    "Could not allocate memory.\n");
                retVal = NPF_OUT_OF_MEMORY;
                goto LBL_END;
            }

            /* Set elements */
            entry->l2id             = (fiedEntry->l2idIncrementor)++;
            entry->nhAddr.ipType    = nhAddrs->ipType;
            entry->status           = l2idm_pending;
            memcpy(&(entry->nhAddr.addr),
                   ((void*)nhAddrs->ipAddrPtr) + (j * addrSz), addrSz);

            /* Stuff entry into internal L2ID table */
            if (npf_list_pushFront(l2Tbl, entry) != 0)
            {
                retVal = NPF_FAIL;
                npf_logger_write(LOG_IPV4, LOG_ERR,
                    "Error pushing onto list; ret = %d.\n", retVal);
                PIL_FreeMemory(entry);
                goto LBL_END;
            }

            /* Add L2ID into new L2ID out list */
            (*newL2idList)[newOutListIncr++] = entry->l2id;

        } /* if (tmpFlags[j] == 0) */

    } /* for (j = 0; j < numNHAddrs; j++) */

LBL_END:

    if (tmpFlags != NULL)
    {
        PIL_FreeMemory(tmpFlags);
    }

    if (retVal != NPF_SUCCESS)
    {
        /* Free all memory on error */
        if (*newL2idList != NULL)
        {
            PIL_FreeMemory(*newL2idList);
        }
    }

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/*****************************************************************************
*   Function: ScanForNhAddrInL2List()
*
*   Parameters:
*       firstPass   - flag indicating if this is first or second invocation
*       entry       - internal L2ID table entry
*       numNHAddrs  - num next hop addrs to add
*       nhAddrs     - array of next hop addrs to add
*       newOutListIncr - [in/out] counter for outNewL2id
*       tmpFlags    - [out] array of flags corresponding to nhAddrs
*       numNewL2ids - [out] num new L2IDs added
*       newL2idList - [out] list of new L2IDs added
*
*   Description:
*       Search through newNHAddr list for this NHAddr entry in internal
*       L2ID table. If found, increment counter if first pass or put L2ID
*       in L2id out list.
*
*   Return: none
*
*****************************************************************************/

void ScanForNhAddrInL2List(
    BOOL                    firstPass,
    L2IDM_L2ID_LIST_ENTRY*  entry,
    uint32_t                numNHAddrs,
    L2IDM_IPADDR_REF*       nhAddrs,
    uint32_t*               newOutListIncr,
    uint32_t*               tmpFlags,
    uint32_t*               numNewL2ids,
    FPPI_L2ID*              outNewL2id)
{
    uint32_t i;
    BOOL found = FALSE;
    size_t addrSz;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    /* Set address size */
    addrSz = nhAddrs->ipType == l2idm_ipv4 ? IPV4_ADDR_SZ : IPV6_ADDR_SZ;

    for (i = 0; i < numNHAddrs; i++)
    {
        if (nhAddrs->ipType == entry->nhAddr.ipType &&
            (memcmp( ((void*)nhAddrs->ipAddrPtr) + (i * addrSz),
                     &(entry->nhAddr.addr), addrSz) == 0))
        {
            found = TRUE;

            if (firstPass)
            {
                /* Just increment new NHaddr counter */
                (*numNewL2ids)++;

                /* Mark this entry in NHaddr inlist so that we know
                   to skip it later. */
                tmpFlags[i] = 1;
            }
            else {
                /* This NHAddr already exists in L2 table, put the
                   corresponding L2ID in the new L2ID out list. */
                outNewL2id[(*newOutListIncr)++] = entry->l2id;
            }

            /* Break out of loop since it has been found */
            break;

        } /* nhAddrs->ipType == entry->nhAddr.ipType && ... */

    } /* for (i = 0; i < numNHAddrs; i++) */

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");
}

/*****************************************************************************
*   Function: CommitOrDeleteEntriesInL2IDTable()
*
*   Parameters:
*       tblType    - update is for NH or L2 table on data plane
*       action      - identify if this is a commit or delete
*       numNHAddrs  - num next hop addrs to add
*       nhAddrs     - array of next hop addrs to add
*       l2Tbl       - ptr to low level L2ID table to edit
*
*   Description:
*       For each entry in internal L2 table that has the pending flag
*       set, see if the entry also exists in the NHaddr table. If so,
*       perform one of the following steps dependent on the action flag:
*          1) action == commit: unset the pending flag
*          2) action == delete: delete the entry from the table
*
*       If the entry is not found in the passed in NHAddr table,
*          1) action == commit: delete the entry from the table
*          2) action == delete: log warning message and return success
*
*       The internal L2DI table has the potential to store obseleted entries
*       if not used correctly. When entries are added for either L2 or NH
*       table using update, a new entry is created in the internal table
*       with the pending flag set. If the user chooses to never commit
*       those new entries they will remain in the list indefinetely.
*
*   Return:  NPF_SUCCESS or error code
*
*****************************************************************************/

NPF_RET CommitOrDeleteEntriesInL2IDTable(
    L2IDM_FOR_TABLE_TYPE  tblType,
    L2IDM_ACTION_TYPE     action,
    uint32_t              numNHAddrs,
    L2IDM_IPADDR_REF*     nhAddrs,
    DList*                l2Tbl)
{
    NPF_RET retVal      = NPF_SUCCESS;
    DListItr * itr      = NULL;
    BOOL found          = FALSE;
    uint32_t i;
    L2IDM_L2ID_LIST_ENTRY* entry;
    BOOL isEntryDeleted = FALSE;
    size_t addrSz;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    /* Set address size */
    addrSz = nhAddrs->ipType == l2idm_ipv4 ? IPV4_ADDR_SZ : IPV6_ADDR_SZ;

    /* Loop through list of current NHaddrs in internal table */
    if ( (itr = npf_list_itrCreate(l2Tbl)) == NULL ||
        npf_list_itrFirst(itr) != 0)
    {
        /* List is probably empty */
        retVal = NPF_FAIL;
        npf_logger_write(LOG_IPV4, LOG_ERR,
            "Error creating list iterator - list is probably empty.\n");
        goto LBL_END;
    }

    /* Loop through all entries in internal L2ID list and compare
       with all entries in passed in NHAddr list. */
    do {
        if ((entry = (L2IDM_L2ID_LIST_ENTRY*)npf_list_itrGetData(itr)) ==
            NULL)
        {
            /* No data or itr messed up */
            npf_logger_write(LOG_IPV4, LOG_INFO,
                "No data in this list entry.\n");
            continue;
        }

        /* Loop through NHAddr list looking for a match */
        for (i = 0; i < numNHAddrs; i++)
        {
            if (nhAddrs->ipType == entry->nhAddr.ipType && entry->nhAddr.ipType == l2idm_ipv4)
            {
                if(memcmp( ((NPF_IPv4Address_t *)nhAddrs->ipAddrPtr),
                           &(entry->nhAddr.addr.ipv4Addr), addrSz) == 0)
                {
                
                    found = TRUE;

                    if (action == l2idm_actionCommit &&
                        entry->status == l2idm_pending)
                    {
                        /* Unset pending flag to make it a
                           legitimate entry. */
                        entry->status = l2idm_normal;
                        break;
                    }
                    else if (action == l2idm_actionDelete &&
                             entry->status == l2idm_normal)
                    {
                        /* Mark this entry for deletion */
                        entry->status = l2idm_delete;
                        isEntryDeleted = TRUE;
                        break;
                    }
                } /* if (nhAddrs->ipType == entry->nhAddr.ipType && ... */
            }
#if 0
            if (nhAddrs->ipType == entry->nhAddr.ipType == l2idm_ipv6)
            {
                if(memcmp( ((NPF_IPv6Address_t *)nhAddrs->ipAddrPtr),
                           &(entry->nhAddr.addr.ipv6Addr), addrSz) == 0)
                {
                    found = TRUE;

                    if (action == l2idm_actionCommit &&
                        entry->status == l2idm_pending)
                    {
                        /* Unset pending flag to make it a
                           legitimate entry. */
                        entry->status = l2idm_normal;
                        break;
                    }
                    else if (action == l2idm_actionDelete &&
                             entry->status == l2idm_normal)
                    {
                        /* Mark this entry for deletion */
                        entry->status = l2idm_delete;
                        isEntryDeleted = TRUE;
                        break;
                    }                    
                }
            }
#endif
        } /* for (i = 0; i < numNHAddrs; i++) */
    
        if (!found)
        {
            /* Entry was not found in passed in NHAddr list.
               If this is a commit for NH table remove the entry
               from the internal L2ID table. If this is a commit
               for L2 table just leave it alone. User can use
               delete to remove it if desired. */
            if (action == l2idm_actionCommit &&
                tblType == l2idm_forNHTable)
            {
                /* NHAddr was not found in passed in list so it must
                   be marked for deletion from internal L2ID table. */
                entry->status = l2idm_delete;
                isEntryDeleted = TRUE;
            }
            else if (action == l2idm_actionDelete)
            {
                /* This should not happen but is user dependent.
                   Post logger msg and return without error. */
                npf_logger_write(LOG_IPV4, LOG_INFO,
                    "List entry not found for L2ID deletion.\n");
            }
        } /* if (!found) */

    } while( npf_list_itrNext(itr) != -1);

LBL_END:

    /* Destroy the list iterator */
    if (itr != NULL)
    {
        npf_list_itrDelete(itr);
    }

    /* Remove list entries marked for deletion */
    while (isEntryDeleted == TRUE)
    {
        RemoveFlaggedListEntries(l2Tbl, &isEntryDeleted);
    }

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/*****************************************************************************
*   Function: RemoveFlaggedListEntries()
*
*   Parameters:
*       l2Tbl        - ptr to low level L2ID table
*       entryDeleted - out flag indicating whether a list entry was deleted
*
*   Description:
*       Loop through passed in list and remove entries marked for deletion.
*       If an entry is deleted, the internal list size is decremented by
*       one. This has the effect of ending the loop one entry early for
*       each entry deleted. To compensate for this, this method must be
*       invoked in a loop until the out param is FALSE, indicating that
*       there are no more entries marked for deletion.
*
*   Return:  TRUE if an entry was deleted and FALSE otherwise.
*
*****************************************************************************/

NPF_RET RemoveFlaggedListEntries(
    DList* l2Tbl,
    BOOL*  entryDeleted)
{
    NPF_RET retVal               = NPF_SUCCESS;
    DListItr * itr               = NULL;
    L2IDM_L2ID_LIST_ENTRY* entry = NULL;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    /* Set out param */
    *entryDeleted = FALSE;

    /* Loop through list of current NHaddrs in internal table */
    if ( (itr = npf_list_itrCreate(l2Tbl)) == NULL ||
        npf_list_itrFirst(itr) != 0)
    {
        /* List is probably empty. This is expected when user
           deletes last entry in list. */
        npf_logger_write(LOG_IPV4, LOG_INFO,
            "OK - list is now empty.\n");
        goto LBL_END;
    }

    /* Loop through all entries in L2ID list and delete those marked
       for deletion. */
    do {
        if ((entry = (L2IDM_L2ID_LIST_ENTRY*)npf_list_itrGetData(itr)) ==
            NULL)
        {
            /* No data or itr messed up */
            npf_logger_write(LOG_IPV4, LOG_INFO,
                "No data in this list entry.\n");
            continue;
        }

        if (entry->status == l2idm_delete)
        {
            if (npf_list_itrRemoveFree(itr) != 0)
            {
                retVal = NPF_FAIL;
                npf_logger_write(LOG_IPV4, LOG_ERR,
                    "Error deleting list entry.\n");
                goto LBL_END;
            }

            *entryDeleted = TRUE;
        }
    } while( npf_list_itrNext(itr) != -1);

LBL_END:

    /* Destroy the list iterator */
    if (itr != NULL)
    {
        npf_list_itrDelete(itr);
    }

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/*****************************************************************************
*   Function: RetrieveL2IDs()
*
*   Parameters:
*       numNHAddrs  - num next hop addrs to add
*       nhAddrs     - array of next hop addrs to add
*       l2Tbl       - ptr to low level L2ID table
*       numL2ids    - [out] num L2IDs returned
*       l2idList    - [out] list of L2IDs that correspond with nhAddrs
*
*   Description:
*       For each entry in internal L2 table that has a match in the
*       passed in NHAddr list, put the associated L2ID into the out list.
*       Do not return L2IDs for entries that are pending. They may not
*       exist in data plane yet and may never exist there if update was
*       not successful.
*
*   Return:  NPF_SUCCESS or error code
*
*****************************************************************************/

NPF_RET RetrieveL2IDs(
    uint32_t            numNHAddrs,
    L2IDM_IPADDR_REF*   nhAddrs,
    DList*              l2Tbl,
    uint32_t*           numL2ids,
    FPPI_L2ID**         l2idList)
{
    NPF_RET retVal           = NPF_SUCCESS;
    DListItr * itr           = NULL;
    uint32_t l2idIncrementor = 0;
    uint32_t i;
    L2IDM_L2ID_LIST_ENTRY* entry;
    size_t addrSz;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    /* Set address size */
    addrSz = nhAddrs->ipType == l2idm_ipv4 ? IPV4_ADDR_SZ : IPV6_ADDR_SZ;

    if ( (itr = npf_list_itrCreate(l2Tbl)) == NULL ||
        npf_list_itrFirst(itr) != 0)
    {
        retVal = NPF_FAIL;
        npf_logger_write(LOG_IPV4, LOG_ERR,
            "Error creating list itr - L2ID list is probably empty.\n");
        goto LBL_END;
    }

    /* Allocate memory for the out L2 list, assuming that there
       will be one entry per NHAddr entry. If there are fewer,
       the out numL2IDs param will reflect this. There can not be
       more. */
    if ((*l2idList = (FPPI_L2ID*)PIL_AllocateMemory(
        sizeof(FPPI_L2ID) * numNHAddrs, PIL_MEM_ZERO_INIT)) == NULL)
    {
        npf_logger_write(LOG_IPV4, LOG_ERR,
            "Could not allocate memory.\n");
        retVal = NPF_OUT_OF_MEMORY;
        goto LBL_END;
    }

    /* Loop through all entries in internal L2ID list and compare
       with all entries in passed in NHAddr list. */
    do {
        if ((entry = (L2IDM_L2ID_LIST_ENTRY*)
            npf_list_itrGetData(itr)) == NULL)
        {
            /* No data or itr messed up */
            npf_logger_write(LOG_IPV4, LOG_INFO,
                "No data in this list entry.\n");
            continue;
        }

        /* Only consider entries that are not pending */
        if (entry->status == l2idm_normal)
        {
            /* Loop through NHAddr list looking for a match */
            for (i = 0; i < numNHAddrs; i++)
            {
                /* Check for NHAddr match */
                if (nhAddrs->ipType ==
                    entry->nhAddr.ipType && nhAddrs->ipType == l2idm_ipv4)
                {
                    if(memcmp( ((NPF_IPv4Address_t *)nhAddrs->ipAddrPtr),
                               &(entry->nhAddr.addr.ipv4Addr), addrSz) == 0)
                {
                    /* Error check to make sure array is not overrun */
                    if (l2idIncrementor >= numNHAddrs)
                    {
                        /* This should never happen. Means that there are
                           duplicate NHAddrs in an L2ID table. Need to
                           set error and exit to troubleshoot problem. */
                        npf_logger_write(LOG_IPV4, LOG_ERR,
                            "Duplicate NHAddr in internal L2ID table.\n");
                        retVal = NPF_FAIL;
                        goto LBL_END;
                    }

                    /* Put corresponding L2ID into out list */
                    (*l2idList)[l2idIncrementor++] = entry->l2id;

                    /* Increment counter out param */
                    (*numL2ids)++;

                    /* Entry found so break from loop to go on to next */
                    break;
                }
                }
                else if (nhAddrs->ipType ==
                         entry->nhAddr.ipType && nhAddrs->ipType == l2idm_ipv6)
                {
                    if(memcmp( ((NPF_IPv6Address_t *)nhAddrs->ipAddrPtr),
                               &(entry->nhAddr.addr.ipv6Addr), addrSz) == 0)
                {
                    /* Error check to make sure array is not overrun */
                    if (l2idIncrementor >= numNHAddrs)
                    {
                        /* This should never happen. Means that there are
                           duplicate NHAddrs in an L2ID table. Need to
                           set error and exit to troubleshoot problem. */
                        npf_logger_write(LOG_IPV4, LOG_ERR,
                            "Duplicate NHAddr in internal L2ID table.\n");
                        retVal = NPF_FAIL;
                        goto LBL_END;
                    }

                    /* Put corresponding L2ID into out list */
                    (*l2idList)[l2idIncrementor++] = entry->l2id;

                    /* Increment counter out param */
                    (*numL2ids)++;

                    /* Entry found so break from loop to go on to next */
                    break;
                }
                } /* if (nhAddrs->ipType == entry->nhAddr.ipType && ... */
            } /* for (i = 0; i < numNHAddrs; i++) */
        } /* if (entry->status == l2idm_normal) */

    } while( npf_list_itrNext(itr) != -1);

LBL_END:

    /* Destroy the list iterator */
    if (itr != NULL)
    {
        npf_list_itrDelete(itr);
    }

    if (retVal != NPF_SUCCESS)
    {
        /* Free all memory on error */
        if (*l2idList != NULL)
        {
            PIL_FreeMemory(*l2idList);
        }
    }

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/*****************************************************************************
*   Function: CheckDeleteEntryFromMasterList()
*
*   Parameters:
*       fiedEntry - ptr to an entry in the master L2ID table
*
*   Description:
*       Check if both contained lists in this master list entry are empty.
*       If they are, destroy the lists and remove this entry from the
*       master list.
*
*   Return: none
*
*****************************************************************************/

void CheckDeleteEntryFromMasterList(
    FPPI_FEID              feid,
    L2IDM_FEID_LIST_ENTRY* fiedEntry,
    BOOL*                  isDeleted)
{
    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    /* There are two lists in the parent. Only destroy
       entry if both lists are empty. */
    if (npf_list_size(&fiedEntry->nhidList) == 0 &&
        npf_list_size(&fiedEntry->l2idList) == 0)
    {
        /* Destroy the tables */
        npf_list_destroy(&fiedEntry->nhidList);
        npf_list_destroy(&fiedEntry->l2idList);

        PIL_EnterCriticalSection(&g_l2idMgrMasterList.listLock);

        /* Now delete the FEID entry from master list */
        npf_list_removeByKeyFree(&g_l2idMgrMasterList.list,
            (void *)&feid);

        PIL_LeaveCriticalSection(&g_l2idMgrMasterList.listLock);    

        /* Set out param to indicate master list entry was deleted */
        *isDeleted = TRUE;
    }

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");
}

/*****************************************************************************
*   Function: UpdateSingleEntry()
*
*   Parameters:
*       feid        - forwarding element ID (same as blade ID)
*       nhid        - next hop ID specified by app
*       nhAddr      - next hop addr to add
*       isNewEntry  - [out] flag to indicate if this nhAddr is new to list
*       newL2idList - [out] new L2ID assigned to new nhAddr
*
*   Description:
*       Handle updating of internal list with the passed in nhAddr entry
*       with specified feid and nhid.
*
*   Return: none
*
*****************************************************************************/

NPF_RET UpdateSingleEntry(
    FPPI_FEID             feid,
    FPPI_NHID             nhid,
    L2IDM_IPADDR_TYPE     addrType,
    void*                 nhAddr,
    BOOL*                 isNewEntry,
    uint32_t*             newL2ID)
{
    NPF_RET retVal                   = NPF_SUCCESS;
    uint32_t tableRet                = 0;
    DList* table                     = NULL;
    L2IDM_FEID_LIST_ENTRY* fiedEntry = NULL;
    L2IDM_NHID_LIST_ENTRY* nhidEntry = NULL;
    uint32_t numNewL2ids             = 0;
    FPPI_L2ID* newL2idList           = NULL;
    BOOL addedToNHList               = FALSE;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    /* Locate table with specified FEID and NHID if it exists */
    tableRet = GetTableBasedOnFEIDandNHID(l2idm_forNHTable, feid, nhid,
        &table, &fiedEntry);

    switch (tableRet)
    {
    case L2ID_OK:

        /* Found NHID list - compare entries with those in passed in
           NHaddrs list. Fill new and old L2ID out lists. */
        UpdateL2IDTableForNH(addrType, nhAddr, fiedEntry, table,
            isNewEntry, newL2ID);

        break;
        
    case L2ID_NO_FEID_ENTRY:

        /* Create a new master table entry for this new FEID and insert
           into the master global list. Fill in new L2ID entry. */
        retVal = CreateAndFillNewFEIDList(l2idm_forNHTable, feid, nhid,
            1, addrType, nhAddr, &numNewL2ids, &newL2idList);

        break;
        
    case L2ID_NO_NHID_ENTRY:

        /* Create a new table entry in the appropriate FEID table
           for this new NHID. The old L2ID out list will be empty. */
        if ( (retVal = CreateAndInitNHIDObj(nhid, &nhidEntry)) !=
            NPF_SUCCESS)
        {
            goto LBL_END;
        }

        /* Add the NHID entry to the list in the FEID entry */
        if (npf_list_pushFront(&(fiedEntry->nhidList), nhidEntry) != 0)
        {
            retVal = NPF_FAIL;
            npf_logger_write(LOG_IPV4, LOG_ERR,
                "Error pushing onto list; ret = %d.\n", retVal);
            goto LBL_END;
        }

        addedToNHList = TRUE;

        /* Create a new L2ID list entry for each new NHaddr and insert
           into appropriate list. Fill new L2ID out list. */
        if ( (retVal = AddNHAddrInfoToList(fiedEntry,
            &(nhidEntry->l2idList), 1, addrType, nhAddr, &numNewL2ids,
            &newL2idList)) != NPF_SUCCESS)
        {
            goto LBL_END;
        }

        break;
    } /* switch (tableRet) */

    if (tableRet == L2ID_NO_FEID_ENTRY || tableRet == L2ID_NO_NHID_ENTRY)
    {
        /* Transfer new data to out params */
        *isNewEntry = TRUE;
        *newL2ID    = *newL2idList;

        /* Free memory */
        PIL_FreeMemory(newL2idList);
    }

LBL_END:

    if (retVal != NPF_SUCCESS && addedToNHList)
    {
        /* Pop from front of list and free data */
        npf_list_popFrontFree(&(fiedEntry->nhidList));
    }

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/*****************************************************************************
*   Function: UpdateL2IDTableForNH()
*
*   Parameters:
*       nhAddr      - next hop addr to add
*       fiedEntry   - ptr to master list entry for this FEID
*       l2Tbl       - ptr to low level L2ID table to edit
*       isNewEntry  - [out] flag to indicate if this nhAddr is new to list
*       newL2idList - [out] new L2ID assigned to new nhAddr
*
*   Description:
*       Loop through passed in list and look for passed in NHAddr entry.
*       Ignore all entries marked as pending but not those marked as old.
*       If found, don't need to do anything except if the status flag
*       for the entry is set to old reset it to normal. If not found, add
*       a new entry and set the status flag to pending (awaiting commit).
*       Set the out flag to true and place the new L2ID into out param.
*       For all internal entries that do not match passed in NHAddr and
*       that are not marked as normal pending, set the status flag to old.
*
*   Return: none
*
*****************************************************************************/

NPF_RET UpdateL2IDTableForNH(
    L2IDM_IPADDR_TYPE       addrType,
    void*                   nhAddr,
    L2IDM_FEID_LIST_ENTRY*  fiedEntry,
    DList*                  l2Tbl,
    BOOL*                   isNewEntry,
    FPPI_L2ID*              newL2ID)
{
    NPF_RET retVal               = NPF_SUCCESS;
    DListItr * itr               = NULL;
    L2IDM_L2ID_LIST_ENTRY* entry = NULL;
    BOOL isFound                 = FALSE;
    size_t addrSz;

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Entering method.\n");

    /* Set address size */
    addrSz = addrType == l2idm_ipv4 ? IPV4_ADDR_SZ : IPV6_ADDR_SZ;

    /* Loop through list of current NHaddrs in internal table */
    if ( (itr = npf_list_itrCreate(l2Tbl)) != NULL &&
       npf_list_itrFirst(itr) == 0)
    {
        /* Loop through all entries in internal L2ID list and compare
           with all entries in passed in NHAddr list. */
        do {
            if ((entry = (L2IDM_L2ID_LIST_ENTRY*)
                npf_list_itrGetData(itr)) == NULL)
            {
                /* No data or itr messed up */
                npf_logger_write(LOG_IPV4, LOG_INFO,
                    "No data in this list entry.\n");
                continue;
            }

            /* Do not consider pending entries since they don't for
               sure exist yet in data plane tables. */
            if (entry->status != l2idm_pending)
            {
                if (entry->nhAddr.ipType == addrType &&
                    (memcmp(&(entry->nhAddr.addr), nhAddr, addrSz) == 0))
                {
                    /* Entry is found so make sure it is given
                       normal secure status so not marked as old
                       during next pass of this same top level
                       invocation. */
                    entry->status = l2idm_normal_secure;

                    /* Mark entry as found and break out of loop */
                    isFound = TRUE;
                    break;

                }
                else {
                    if (entry->status != l2idm_normal_secure)
                    {
                        /* Entry not found during this top level method
                           invocation so set to old status. */
                        entry->status = l2idm_old;
                    }
                } /* if (memcmp(entry->nhAddr, .. */
            } /* if (entry->status != l2idm_pending) */

        } while( npf_list_itrNext(itr) != -1);

        /* Destroy the list iterator */
        if (itr != NULL)
        {
            npf_list_itrDelete(itr);
        }
    } /* if ( (itr = npf_list_itrCreate(l2Tbl)) != NULL && ... */

    if (!isFound)
    {
        /* NHaddr not found in internal table so add entry and mark
           as pending. */

        /* Allocate memory for new L2ID object */
        if ((entry = (L2IDM_L2ID_LIST_ENTRY*)PIL_AllocateMemory(
            sizeof(L2IDM_L2ID_LIST_ENTRY), PIL_MEM_ZERO_INIT)) == NULL)
        {
            npf_logger_write(LOG_IPV4, LOG_ERR,
                "Could not allocate memory.\n");
            retVal = NPF_OUT_OF_MEMORY;
            goto LBL_END;
        }

        /* Set elements */
        entry->l2id          = (fiedEntry->l2idIncrementor)++;
        entry->nhAddr.ipType = addrType;
        memcpy(&(entry->nhAddr.addr), nhAddr, addrSz);
        entry->status        = l2idm_pending;

        /* Stuff entry into internal L2ID table */
        if (npf_list_pushFront(l2Tbl, entry) != 0)
        {
            retVal = NPF_FAIL;
            npf_logger_write(LOG_IPV4, LOG_ERR,
                "Error pushing onto list; ret = %d.\n", retVal);
            PIL_FreeMemory(entry);
            goto LBL_END;
        }

        /* Set out params */
        *isNewEntry = TRUE;
        *newL2ID    = entry->l2id;

    } /* if (!isFound) */

LBL_END:

    npf_logger_write(LOG_IPV4, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/*****************************************************************************
*
*   The remainder of the file is for testing and debugging.
*
*****************************************************************************/

#ifdef _GWGTEST_

#include <stdio.h>

void DumpL2IDTable(
    DList* l2idTbl);

/*****************************************************************************
*   Function: DumpInternalList
*
*   Parameters: none
*
*   Description:
*       Dump the entire contents of the internal global L2ID list. Used
*       for testing and debugging.
*
*   Return: none
*
*****************************************************************************/

void DumpInternalList()
{
    DListItr* itr1      = NULL;
    DListItr* itr2      = NULL;
    L2IDM_FEID_LIST_ENTRY* p1Entry;
    L2IDM_NHID_LIST_ENTRY* p2Entry;
    int32_t sz1, sz2;

    printf("\n****************** Master list Dump ******************\n");
    printf("Size of master list = %d\n\n",
        npf_list_size(&g_l2idMgrMasterList.list));

    PIL_EnterCriticalSection(&g_l2idMgrMasterList.listLock);

    if ( (itr1 = npf_list_itrCreate(&g_l2idMgrMasterList.list)) == NULL ||
        npf_list_itrFirst(itr1) != 0)
    {
        /* List is probably empty */
        goto LBL_END;
    }

    do {

        if ((p1Entry = (L2IDM_FEID_LIST_ENTRY*)
            npf_list_itrGetData(itr1)) == NULL)
        {
            /* No data or itr messed up */
            npf_logger_write(LOG_IPV4, LOG_INFO,
                "No data in this list entry.\n");
            continue;
        }

        /* Dump FEID entry */
        printf("Feid = %d\n", p1Entry->feid);
        printf("l2idIncrementor = %d\n", p1Entry->l2idIncrementor);
        printf("Size of nhidList = %d\n", npf_list_size(&p1Entry->nhidList));
        printf("Size of l2idList = %d\n",
            (sz2 = npf_list_size(&p1Entry->l2idList)));

        /* Now look in the contained NHID entries */
        if ( (itr2 = npf_list_itrCreate(&p1Entry->nhidList)) != NULL &&
            npf_list_itrFirst(itr2) == 0)
        {
            do {

                if ((p2Entry = (L2IDM_NHID_LIST_ENTRY *)
                    npf_list_itrGetData(itr2)) == NULL)
                {
                    /* No data or itr messed up */
                    npf_logger_write(LOG_IPV4, LOG_INFO,
                        "No data in this list entry.\n");
                    continue;
                }

                /* Dump contents of NHID entry */
                printf("   nhid = %d\n", p2Entry->nhid);
                printf("   Size of NHID l2idList = %d\n",
                    (sz1 = npf_list_size(&p2Entry->l2idList)));

                if (sz1 > 0)
                {
                    DumpL2IDTable(&p2Entry->l2idList);
                }

            } while(npf_list_itrNext(itr2) != -1);

            /* Destroy the list iterator */
            if (itr2 != NULL)
            {
                npf_list_itrDelete(itr2);
            }
        } /* if ( (itr2 = npf_list_itrCreate(&p1Entry->nhidList)) != ... */

        /* Now dump the contained L2ID entries */
        if (sz2 > 0)
        {
            printf("   FEID l2idList entries:\n\n");
            DumpL2IDTable(&p1Entry->l2idList);
        }

    } while(npf_list_itrNext(itr1) != -1);

LBL_END:

    PIL_LeaveCriticalSection(&g_l2idMgrMasterList.listLock);    

    /* Destroy the list iterator */
    if (itr1 != NULL)
    {
        npf_list_itrDelete(itr1);
    }
}

/*****************************************************************************
*   Function: DumpL2IDTable
*
*   Parameters: 
*       l2idTbl  - ptr to L2ID table
*
*   Description:
*       Dump the contents of the passed in L2ID table. Used for testing
*       and debugging.
*
*   Return: none
*
*****************************************************************************/

void DumpL2IDTable(
    DList* l2idTbl)
{
    DListItr* itr2                 = NULL;
    L2IDM_L2ID_LIST_ENTRY* p2Entry = NULL;

    if ( (itr2 = npf_list_itrCreate(l2idTbl)) == NULL ||
        npf_list_itrFirst(itr2) != 0)
    {
        /* List is probably empty */
        goto LBL_END;
    }

    do {

        if ((p2Entry = (L2IDM_L2ID_LIST_ENTRY*)
            npf_list_itrGetData(itr2)) == NULL)
        {
            /* No data or itr messed up */
            npf_logger_write(LOG_IPV4, LOG_INFO,
                "No data in this list entry.\n");
            continue;
        }

        /* Dump contents of NHID entry */
        printf("      nhAddr Type = %d\n", p2Entry->nhAddr.ipType);
        switch (p2Entry->nhAddr.ipType)
        {
        case l2idm_ipv4:
            printf("      nhAddr      = %d\n", p2Entry->nhAddr.addr.ipv4Addr);
            break;
        case l2idm_ipv6:
            printf("      nhAddr      = %d.%d.%d.%d\n",
                (p2Entry->nhAddr.addr.ipv6Addr)[0],
                (p2Entry->nhAddr.addr.ipv6Addr)[1],
                (p2Entry->nhAddr.addr.ipv6Addr)[2],
                (p2Entry->nhAddr.addr.ipv6Addr)[3]);
            break;
        }
        printf("      l2id        = %d\n", p2Entry->l2id);
        printf("      status      = %d\n\n", p2Entry->status);

    } while(npf_list_itrNext(itr2) != -1);

    /* Destroy the list iterator */
    if (itr2 != NULL)
    {
        npf_list_itrDelete(itr2);
    }

LBL_END:

}
#endif /* _GWGTEST_ */
