/*****************************************************************************
*                  INTEL CORPORATION PROPRIETARY INFORMATION
*  This software is supplied under the terms of a license agreement or
*  nondisclosure agreement with Intel Corporation and may not be copied
*  or disclosed except in accordance with the terms of that agreement.
*  Copyright (c) 2001 Intel Corporation. All Rights Reserved.
*****************************************************************************/


/*****************************************************************************
* $Source: /CODEBASE/ControlPlane/CoreBLocks/NameSpace/src/namespace.c 
* $Revision: 1.6 $
* $Date: 2003/11/17 16:57:13 $
*
* Description:
*    The implementation of NPF PDK namespace. As a convention, the external
*    NPF API functions are named as npf_ns_xxx, the functions exposed only
*    to other PDK components are ns_xxx, while the rest of them are internal
*    functions used only in the implementation of the above functions.
*
*    All the public NPF API functions deployes the g_nsLock to protect the
*    namespace hierachy. Since the private functions are called from the
*    protected publick functions, no extra lock is needed for the private
*    functions.  
*
* Assumptions:
*     1. PDK Manager initializes and shutdown namespace
*     2. System data like FE and ports are maintained by other PDK modules
*****************************************************************************/
 
#include "ns.h"

#ifdef NSREFCOUNT

#include <assert.h>

typedef struct {
    char *funcName;
    uint32_t refCount;
} FuncNode_t;

typedef struct {
    NPF_NSHDL node;
    char *nodeName;
    DList funcList;
} RefNode_t;

static DList g_refNodeList;

static void DumpFuncs( DList *funcList ) {
    DListItr *itr;
    FuncNode_t *funcNode = NULL;

    if ( (itr = npf_list_itrCreate( funcList )) != NULL ) {
        npf_list_itrFirst( itr );
        do {
            funcNode = npf_list_itrGetData( itr );
            assert( funcNode );
            printf( "%d references for function %s\n",
                    funcNode->refCount, funcNode->funcName );
        } while ( npf_list_itrNext( itr ) != -1 );
        npf_list_itrDelete( itr );
    }
}

static void DumpRefCount() {
    DListItr *itr;
    RefNode_t *refNode = NULL;

    printf( "***************************************\n" );
    printf( "Namespace handle leak report\n" );

    if ( (itr = npf_list_itrCreate( &g_refNodeList )) != NULL ) {
        npf_list_itrFirst( itr );
        do {
            refNode = npf_list_itrGetData( itr );
            assert( refNode );
            printf( "\nOutstanding references for 0x%x, %s\n",
                    refNode->node, refNode->nodeName );
            DumpFuncs( &(refNode->funcList) );
        } while ( npf_list_itrNext( itr ) != -1 );
        npf_list_itrDelete( itr );
    }
    printf( "***************************************\n" );
}

static void IncFuncCount( char *func, DList *funcList ) {
    DListItr *itr;
    FuncNode_t *funcNode = NULL;
    uint8_t found = 0;

    if ( (itr = npf_list_itrCreate( funcList )) != NULL ) {
        npf_list_itrFirst( itr );
        do {
            funcNode = npf_list_itrGetData( itr );
            assert( funcNode );
            if ( strcmp( func, funcNode->funcName ) == 0 ) {
                found = 1;
            }
        } while ( npf_list_itrNext( itr ) != -1 && !found );
        npf_list_itrDelete( itr );
    }
    if ( !found ) {
        funcNode = (FuncNode_t *)malloc( sizeof( FuncNode_t ) );
        funcNode->funcName = strdup( func );
        funcNode->refCount = 0;
        npf_list_pushBack( funcList, funcNode );
    }
    funcNode->refCount++;
}

static void IncNodeCount( NPF_NSHDL node, char *func, int linenumber ) {
    DListItr *itr;
    RefNode_t *refNode = NULL;
    char tempName[NPF_NS_MAXPATH];
    uint8_t found = 0;

    if (npf_ns_getPath (node, tempName) != NPF_SUCCESS)
        return;
    printf ("%s increment in %s:%d\n", tempName, func, linenumber);

    if ( (itr = npf_list_itrCreate( &g_refNodeList )) != NULL ) {
        npf_list_itrFirst( itr );
        do {
            refNode = npf_list_itrGetData( itr );
            assert( refNode );
            if ( refNode->node == node ) {
                found = 1;
            }
        } while ( npf_list_itrNext( itr ) != -1 && !found );
        npf_list_itrDelete( itr );
    }
    if ( !found ) {
        if ( npf_ns_getPath( node, tempName ) != NPF_SUCCESS ) {
            return;
        }
        refNode = (RefNode_t *)malloc( sizeof( RefNode_t ) );
        refNode->node = node;
        refNode->nodeName = strdup( tempName );
        npf_list_init( &(refNode->funcList), free );
        npf_list_pushBack( &g_refNodeList, refNode );
    }
    IncFuncCount( func, &(refNode->funcList) );
}

static void DecFuncCount( char *func, DList *funcList ) {
    DListItr *itr;
    FuncNode_t *funcNode = NULL;
    uint8_t found = 0;

    if ( (itr = npf_list_itrCreate( funcList )) != NULL ) {
        npf_list_itrFirst( itr );
        do {
            funcNode = npf_list_itrGetData( itr );
            assert( funcNode );
            if ( strcmp( func, funcNode->funcName ) == 0 ) {
                found = 1;
            }
        } while ( npf_list_itrNext( itr ) != -1 && !found );
        npf_list_itrDelete( itr );
    }
    if ( !found ) {
        funcNode = (FuncNode_t *)malloc( sizeof( FuncNode_t ) );
        funcNode->funcName = strdup( func );
        funcNode->refCount = 0;
        npf_list_pushBack( funcList, funcNode );
    }
    funcNode->refCount--;
}

static void DecNodeCount( NPF_NSHDL node, char *func, int linenumber ) {
    DListItr *itr;
    RefNode_t *refNode = NULL;
    char tempName[NPF_NS_MAXPATH];
    uint8_t found = 0;

    if (npf_ns_getPath (node, tempName) != NPF_SUCCESS)
        return;
    printf ("%s decrement in %s:%d\n", tempName, func, linenumber);

    if ( (itr = npf_list_itrCreate( &g_refNodeList )) != NULL ) {
        npf_list_itrFirst( itr );
        do {
            refNode = npf_list_itrGetData( itr );
            assert( refNode );
            if ( refNode->node == node ) {
                found = 1;
            }
        } while ( npf_list_itrNext( itr ) != -1 && !found );
        npf_list_itrDelete( itr );
    }
    if ( !found ) {
        if ( npf_ns_getPath( node, tempName ) != NPF_SUCCESS ) {
            return;
        }
        refNode = (RefNode_t *)malloc( sizeof( RefNode_t ) );
        refNode->node = node;
        refNode->nodeName = strdup( tempName );
        npf_list_init( &(refNode->funcList), free );
        npf_list_pushBack( &g_refNodeList, refNode );
    }
    DecFuncCount( func, &(refNode->funcList) );
}

static void RefCountShutdown() 
{
    RefNode_t * refNode;

    while(npf_list_popBack(&g_refNodeList, (void **)&refNode) != -1)
    {
        if(!refNode) continue;
        free(refNode->nodeName);
        /* this function frees all the remaining func nodes */
        npf_list_destroy(&(refNode->funcList));
        free(refNode);
    }
} /* RefCountShutdown */

#endif


/* internal functions, not exposed outside namespace */
    static NPF_NS_NODE * GetNode(char * pathStr);
    static char * GetPath(NPF_NS_NODE * node);
    static Bool NoRef(NPF_NS_NODE * node);
    static void RemoveNode(NPF_NS_NODE * node);
    static Bool foundNsHdl(NPF_NSHDL hNode, 
                           NPF_NS_NODE * node);
    static Bool isValid(NPF_NSHDL hNode);
    static NPF_RET CreateNode(char *         pathStr,
                              NPF_NSNODETYPE nodeType,
                              NPF_DATATYPE   dataType,
                              NPF_NSHDL      hInstance,
                              NPF_NS_NODE ** node); 


/* global variables for namespace */ 
    static NPF_NS_NODE * g_nsRoot;  /* the namespace root */
    static RWLOCK        g_nsLock;  /* the ns global lock */
    static STATE         g_nsState; /* the ns state       */   
    static DList         g_dirTab;  /* the table of outstanding dir iterator*/
    static DList         g_assoTab; /* the table of outstanding associate 
                                       iterator*/

/*****************************************************************************
*
*   Function: GetNode()
*
*   Parameters:
*        pathStr - a null-terminated string for the path
*
*   Description: Get reference to the node with the path
*
*   Return:  a pointer to the node. NULL pointer means the node doesn't exist
*
*****************************************************************************/
NPF_NS_NODE * GetNode(char * pathStr)
{
    char *          input;
    char *          str;
    char            name[NPF_NS_MAXNAME];
    DList *         nodes;
    DList           nsRooter; /* list to contain g_nsRoot only */
    DListItr *      itr;
    Bool            found;
    NPF_NS_NODE *   tempRoot;
    NPF_NS_NODE *   node;
 
 
    if((!pathStr) || (strchr(pathStr, NPF_NS_NAMEDELIM) != pathStr)
                  || (strlen(pathStr) > NPF_NS_MAXPATH))
    {
        return NULL;
    }

    /* pathStr is "/" */ 
    if(strlen(pathStr) == 1)
    {
        return g_nsRoot;
    }

    /* put the g_nsRoot into a DList */
    npf_list_init(&nsRooter, NULL);
    npf_list_pushFront(&nsRooter, (void *)g_nsRoot);
    nodes = &nsRooter;

    /* parse the path string and travers the path */
    input = pathStr;
    str = strchr(++input, NPF_NS_NAMEDELIM);
    while(str)
    {
        memset(name, 0, NPF_NS_MAXNAME);
        if(str-input > NPF_NS_MAXNAME) 
        {
            npf_list_popFront(&nsRooter, (void *)&tempRoot);
            npf_list_destroy(&nsRooter);
            return NULL;
        }
        strncpy(name, input, str-input);
 
        found = FALSE;
        if((itr = npf_list_itrCreate(nodes)) != NULL)
        {
            npf_list_itrFirst(itr);
            do {
                node = (NPF_NS_NODE *)npf_list_itrGetData(itr);
                if(!node) continue;
                if(!strcmp(node->entry.name, name))
                {
                     nodes = &(node->children);
                     found = TRUE;
                     break;
                }
            } while( npf_list_itrNext(itr) != -1);
            npf_list_itrDelete(itr);
        }
        if(!found)
        {
            npf_list_popFront(&nsRooter, (void *)&tempRoot);
            npf_list_destroy(&nsRooter);
            return NULL;
        }

        input = str;
        str = strchr(++input, NPF_NS_NAMEDELIM);
    } /* while */
 

    /* deal with the last node on the path */ 
    memset(name, 0, NPF_NS_MAXNAME);
    strcpy(name, input);            
    if((itr = npf_list_itrCreate(nodes)) == NULL)
    {
        npf_list_popFront(&nsRooter, (void *)&tempRoot);
        npf_list_destroy(&nsRooter); 
        return NULL;
    }
    npf_list_itrFirst(itr);
    do {
        node = (NPF_NS_NODE *)npf_list_itrGetData(itr);
        if(!strcmp(node->entry.name, name))
        {
            npf_list_itrDelete(itr);
            npf_list_popFront(&nsRooter, (void *)&tempRoot);
            npf_list_destroy(&nsRooter);
            return node;
        }
    } while( npf_list_itrNext(itr) != -1);

    /* clean up side effect */
    npf_list_itrDelete(itr);
    npf_list_popFront(&nsRooter, (void *)&tempRoot);
    npf_list_destroy(&nsRooter);

    return NULL;

} /* GetNode */


/*****************************************************************************
*
*   Function: GetPath()
*
*   Parameters:
*        node - a pointer to a namespace node
*
*   Description: Get the path string of the node
*
*   Return:  a pointer to the path string. It should be freed by caller.
*
*****************************************************************************/
char * GetPath(NPF_NS_NODE * node)
{
    char * path;
    char tempPath[NPF_NS_MAXPATH];
    NPF_NS_NODE * parent;


    if(!node) return NULL;
    path = (char *)PIL_AllocateMemory(NPF_NS_MAXPATH, PIL_MEM_ZERO_INIT);
    if(!path) return NULL;

    /* traverse the path and make the path string */
    memset(path, 0, NPF_NS_MAXPATH);
    path[0] = NPF_NS_NAMEDELIM;
    strcpy(path+1, node->entry.name);    
    parent = node->pParent;    
    while(parent)
    {
        memset(tempPath, 0, NPF_NS_MAXPATH); 
        strcpy(tempPath, path);

        memset(path, 0, NPF_NS_MAXPATH);
        path[0] = NPF_NS_NAMEDELIM;
        strcpy(path+1, parent->entry.name);     
        strcat(path, tempPath);
        
        parent = parent->pParent; 
    }

    return path; /* to be freed by caller */
}

/*****************************************************************************
*
*   Function: ns_init()
*
*   Parameters: none
*
*   Description: initialize the namespace. To be called by PDK manager
*
*   Return: NPF_OSAS_FAIL if underlying OS support fails
*           NPF_SUCCESS if successful
*
*****************************************************************************/
NPF_RET ns_init()
{

    if(g_nsState == INITIALIZED)
    {
        return NPF_SUCCESS;
    }

    g_nsRoot = NULL;
    if(init(&g_nsLock))
    {
#ifdef NSREFCOUNT
    npf_list_init(&g_refNodeList, free);
#endif

        npf_list_init(&g_dirTab, PIL_FreeMemory);  
        npf_list_init(&g_assoTab, PIL_FreeMemory);  
        g_nsState = INITIALIZED;
        return NPF_SUCCESS;
    }

    return NPF_OSAS_FAIL;
} /* ns_init */


/*****************************************************************************
*
*   Function: CleanNewNode()
*
*   Parameters: result - return value
*
*   Description: help macro to clean up new node and return back to
*        caller. Used in CreateNode().
*
*****************************************************************************/ 
#define CleanNewNode(result) \
        PIL_FreeMemory(newNode->entry.name); \
        PIL_FreeMemory(newNode); \
        return result;


/*****************************************************************************
*   Function: CreateNode()
*
*   Parameters:
*       pathStr - null-terminated path string of the node to be created
*       nodeType - type of the node to be created: type, instance, or alias
*       dataType - data type to be refed by this node
*       hInstance - namepsace handle to the instance node for alias
*       node - the output namespace node created
*
*   Description: an internal function to create a namespace node. This
*        function is called by the NPF API functions npf_ns_create and
*        npf_ns_alias. The ns state checking and locking are done in the 
*        calling API functions. 
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_NS_NODE_EXISTS if a node already exists at the path
*           NPF_OUT_OF_MEMORY if the system is out of memory
*           NPF_SUCCESS if successful
*
*****************************************************************************/
NPF_RET CreateNode(char *         pathStr,
                   NPF_NSNODETYPE nodeType,
                   NPF_DATATYPE   dataType,
                   NPF_NSHDL      hInstance,
                   NPF_NS_NODE ** node)
{
    char          name[NPF_NS_MAXNAME];
    NPF_NS_NODE * newNode;
    NPF_RET       result;
    NPF_HANDLE    hData;
    char          parentPath[NPF_NS_MAXPATH];
    char *        delim; /* delimiter '/' */
    NPF_NS_NODE * parentNode;


    if((!pathStr) || (strchr(pathStr, NPF_NS_NAMEDELIM) != pathStr)
                  || (strlen(pathStr) > NPF_NS_MAXPATH)) 
    {
        return NPF_INVALID_PARAMETERS;
    }

    /* Check if the node already exists */
    if(GetNode(pathStr)) 
    {
        return NPF_NS_NODE_EXISTS;
    }

    /* walk to the parent node */
    delim = (char *)strrchr(pathStr, NPF_NS_NAMEDELIM);
    if(!delim) 
    {
        return NPF_INVALID_PARAMETERS;
    }

    if(delim == pathStr) /* want to creat the g_nsRoot */
    {
        if(g_nsRoot) /* g_nsRoot already exits! */
        {
            return NPF_INVALID_PARAMETERS;
        }
        parentNode = NULL;
    }
    else
    {
        memset(parentPath, 0, NPF_NS_MAXPATH); 
        strncpy(parentPath, pathStr, delim-pathStr);
        parentNode = GetNode(parentPath);
        if(!parentNode)
        {
            return NPF_INVALID_PARAMETERS;
        }
    }

    /* create the new node */ 
    memset(name, 0, NPF_NS_MAXNAME);
    strcpy(name, ++delim);   
    newNode = (NPF_NS_NODE *)PIL_AllocateMemory(sizeof(NPF_NS_NODE), 
                                                 PIL_MEM_ZERO_INIT);  
    if(!newNode) 
    {
        return NPF_OUT_OF_MEMORY;
    }
    newNode->entry.name = (char *)PIL_AllocateMemory((strlen(name)+1), 
                                                 PIL_MEM_ZERO_INIT);
    if(!newNode->entry.name)
    {
        PIL_FreeMemory(newNode);
        return NPF_OUT_OF_MEMORY;
    }
    strcpy(newNode->entry.name, name);
    newNode->entry.nodeType = nodeType;
    newNode->entry.dataType = dataType;
    switch(nodeType)
    {
        case NPF_NS_TYPENODE:
            if(parentNode && parentNode->entry.nodeType == NPF_NS_TYPENODE)
            {
                CleanNewNode(NPF_INVALID_PARAMETERS);
            }
            break;

        case NPF_NS_INSTNODE:
            if(parentNode && parentNode->entry.nodeType == NPF_NS_INSTNODE)
            {
                CleanNewNode(NPF_INVALID_PARAMETERS);  
            }

            /* talk to CM to creat hData */
            result = cm_newDataByType(dataType, (NPF_NSHDL)newNode,  &hData);
            if(result != NPF_SUCCESS)
            {
                CleanNewNode(result); 
            }
            newNode->entry.entryAttrib.hData = hData;
            break;

        case NPF_NS_ALIASNODE:
            if(!isValid(hInstance) ||
               ((NPF_NS_NODE *)hInstance)->entry.nodeType == NPF_NS_ALIASNODE)
            {
                CleanNewNode(NPF_INVALID_PARAMETERS);
            }
            
            newNode->entry.entryAttrib.hInstance = hInstance; 
            break;

        default:
            CleanNewNode(NPF_INVALID_PARAMETERS);

    } /* switch */

    /* npf_list_init never fail */
    npf_list_init(&newNode->children, PIL_FreeMemory);
    npf_list_init(&newNode->associates, PIL_FreeMemory); 

    /* add the new node into parent's children or record it as the g_nsRoot */
    if(parentNode)
    {
        npf_list_pushBack(&parentNode->children, newNode);
    }
    else
    {
         g_nsRoot = newNode; 
    }
    
    newNode->pParent = parentNode;
    *node = newNode;    
   
    return NPF_SUCCESS;   

} /* CreateNode */


/*****************************************************************************
*
*   Function: CheckInit()
*
*   Parameters: none
*
*   Description: help macro to check if the namespace is initialized.
*       Used by external namespcae API functions.
*
*****************************************************************************/  
#define CheckInit() \
    if(g_nsState != INITIALIZED) { \
        return NPF_COMPONENT_UNINITIALIZED; \
    } 


/*****************************************************************************
*   Function: npf_ns_create()
*
*   Parameters: 
*       pathStr - null-terminated path string of the node to be created
*       nodeType - type of the node to be created: type, instance, or alias
*       dataType - data type to be refed by this node
*       hInstance - namepsace handle to the instance node for alias
*       hNode - the output namespace handle to the node created
*
*   Description: create a namespace node in open mode. The caller must
*        close the handle after using it.
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_NS_NODE_EXISTS if a node already exists at the path 
*           NPF_OUT_OF_MEMORY if the system is out of memory
*           NPF_COMPONENT_UNINITIALIZED if namespace is not initialized  
*           NPF_SUCCESS if successful
*
*****************************************************************************/
#ifdef NSREFCOUNT
NPF_RET npf_ns_createWrapper(char *         pathStr,
                      NPF_NSNODETYPE nodeType,
                      NPF_DATATYPE   dataType,
                      NPF_NSHDL      hInstance,
                      NPF_NSHDL *    hNode,
                      char      *    caller,
                      int            linenumber    ) {
    NPF_RET ret;
    ret = npf_ns_createD( pathStr, nodeType, dataType, hInstance, hNode );
    if ( ret == NPF_SUCCESS ) {
        IncNodeCount( *hNode, caller, linenumber );
    }
    return ret;
}
NPF_RET npf_ns_createD(char *         pathStr,
                       NPF_NSNODETYPE nodeType,
                       NPF_DATATYPE   dataType,
                       NPF_NSHDL      hInstance,
                       NPF_NSHDL *    hNode) {
#else

NPF_RET npf_ns_create(char *         pathStr,
                      NPF_NSNODETYPE nodeType,
                      NPF_DATATYPE   dataType,
                      NPF_NSHDL      hInstance,
                      NPF_NSHDL *    hNode)  
{
#endif
    NPF_RET result;
    NPF_NS_NODE * node;


    CheckInit();

    acquire(&g_nsLock, RW_WRITE);
    result = CreateNode(pathStr, nodeType, dataType, 
                        hInstance, &node);
    if(result == NPF_SUCCESS)
    {
        /* new node is created in open mode */ 
        node->entry.refCount = 1; 
        *hNode = (NPF_NSHDL)node;
    } else 
        *hNode = 0x0;

    release(&g_nsLock);

    return result; 

} /* npf_ns_create */


/*****************************************************************************
*
*   Function: npf_ns_alias()
*
*   Parameters:
*       canonName - null-terminated canonical path string
*       aliasName - null-terminated alias path string
*
*   Description: create an alias of a namespace node
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_NS_NODE_EXISTS if the alias exists
*           NPF_OUT_OF_MEMORY the system is out of memory
*           NPF_COMPONENT_UNINITIALIZED if namespace is not initialized  
*           NPF_SUCCESS if successful
*
*****************************************************************************/
NPF_RET npf_ns_alias(char * canonName,
                     char * aliasName)
{
    NPF_NS_NODE * node;
    NPF_NS_NODE * alias;
    NPF_RET   result;


    CheckInit();

    acquire(&g_nsLock, RW_WRITE);

    node = GetNode(canonName);
    if(!node) 
    {
        release(&g_nsLock);
        return NPF_INVALID_PARAMETERS;
    }

    result =  CreateNode(aliasName,
                         NPF_NS_ALIASNODE,
                         node->entry.dataType,
                         (NPF_NSHDL)node,
                         &alias);

    release(&g_nsLock);

    return result;

} /* npf_ns_alias */


/*****************************************************************************
*
*   Function: NoRef()
*
*   Parameters:
*       node - a namespace node
*
*   Description: check if there is reference to the node and its descedants.
*        This check is needed when a node is to be deleted.
*
*   Return: TRUE if there is no reference, FALSE otherwise
*
*****************************************************************************/
Bool NoRef(NPF_NS_NODE * node)
{
    NPF_NS_NODE * child;
    DListItr *    itr;

    if(!node) 
    {
        return TRUE;
    }

    if ((itr = npf_list_itrCreate(&(node->children))) != NULL) 
    {
        npf_list_itrFirst(itr);
        do {
            child = (NPF_NS_NODE *)npf_list_itrGetData(itr);
            /* recursing to children */
            if(!NoRef(child)) 
            {
                npf_list_itrDelete(itr); 
                return FALSE;
            }
        } while( npf_list_itrNext(itr) != -1);
        npf_list_itrDelete(itr);    
    }

    switch (node->entry.nodeType)
    {
        case NPF_NS_TYPENODE:
            break;

        case NPF_NS_INSTNODE:
            if (cm_delDataCheck(node->entry.entryAttrib.hData) != NPF_SUCCESS)
            return FALSE;

        case NPF_NS_ALIASNODE:
            break;

        default:
    }

//    return (node->entry.refCount == 0);
    return TRUE; /* don't check reference count */

} /* NoRef */


/*****************************************************************************
*
*   Function: RemoveNode()
*
*   Parameters:
*       node - a namespace node to be removed
*
*   Description: remove a namespace node. This function must be used with
*        the NoRef() check. A node can be removed if there is no reference 
*        to it and its descedants. If the node is removed, all of its 
*        descedants are also removed.
*
*   Return: none
*
*****************************************************************************/
void RemoveNode(NPF_NS_NODE * node)
{
    NPF_NS_NODE * child;
    NPF_NS_NODE * oldNode;
    NPF_NS_NODE * sibling;
    NPF_NS_NODE * parent;
    NPF_NS_NODE * anode;
    DListItr *    itr;
 

    if(!node) return;

    while(npf_list_popBack(&(node->children), (void **)&child) != -1)
    {
        /* recursion to the child */
        RemoveNode(child); 
    } // while


    /* clean all associates */
    while(npf_list_popBack(&node->associates, (void **)&anode) != -1)
    {
        /* remove node from anode's associates */
		if(!anode) continue;
        if((itr = npf_list_itrCreate(&anode->associates)) != NULL)
        {
            if(npf_list_itrFirst(itr) == 0)
            {
                do
                {
                    sibling = (NPF_NS_NODE *)npf_list_itrGetData(itr);
                    if(sibling == node)
                    {
                        npf_list_itrRemove(itr, (void **)&oldNode);
                        break;
                    }
                } while( npf_list_itrNext(itr) != -1);
            }
            npf_list_itrDelete(itr);
        }            
    }

    parent = node->pParent;
    if(!parent) /* this is the root */
    {
        g_nsRoot = NULL;
    }
    else /* not the root */
    {
        /* remove from parent's children list  */
        if((itr = npf_list_itrCreate(&parent->children)) != NULL)
        {
            if (npf_list_itrFirst(itr) == 0)
            {
               do {
                  sibling = (NPF_NS_NODE *)npf_list_itrGetData(itr);
                  if(sibling == node)
                  {
                     npf_list_itrRemove(itr, (void **)&oldNode);
                     break;
                  }
               } while( npf_list_itrNext(itr) != -1);
            }
            npf_list_itrDelete(itr);
        }                       
    } /* else */


    if (node->entry.nodeType == NPF_NS_INSTNODE)
    {
        cm_delDataByHandle(node->entry.entryAttrib.hData);  
    }
    PIL_FreeMemory(node->entry.name);
    PIL_FreeMemory(node);

    return;

} /* RemoveNode */ 

/*****************************************************************************
*
*   Function: npf_ns_delete()
*
*   Parameters:
*       name - null-terminated path string of a namespace node to be removed
*
*   Description: remove a namespace node at the given path. This node can 
*       be removed if there is no reference to it or its descedants. 
*       If the node is removed, all of its descedants are also removed.
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_NS_NODE_IN_USE if the node is in use
*           NPF_COMPONENT_UNINITIALIZED if namespace is not initialized  
*           NPF_SUCCESS if successful 
*
*****************************************************************************/
NPF_RET npf_ns_delete(char * name)
{
    NPF_NS_NODE * node;


    CheckInit();

    acquire(&g_nsLock, RW_WRITE);
    node = GetNode(name);
    if(!node) 
    {
        release(&g_nsLock); 
        return NPF_INVALID_PARAMETERS;
    }


    /* the node can be removed iff no ref to it and its children */
    if(!NoRef(node)) 
    {
        release(&g_nsLock); 
        return NPF_NS_NODE_IN_USE;
    }

    /* remove this node and its children in DFS order */
    RemoveNode(node);

    release(&g_nsLock);
    return NPF_SUCCESS; 

} /* npf_ns_delete */


/*****************************************************************************
*
*   Function: npf_ns_rename()
*
*   Parameters:
*       oldName - null-terminated old path string of a namespace node
*       newName - null-terminated new path string of the namespace node
*
*   Description: rename a namespace node, which causes moving a namespace
*        node from one location on the hierachy to another location 
*
*   Return: NPF_NS_NODE_EXISTS if a node already exists 
*           NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_OUT_OF_MEMORY if the system is out of memory 
*           NPF_COMPONENT_UNINITIALIZED if namespace is not initialized  
*           NPF_SUCCESS if successful
*
*****************************************************************************/
NPF_RET npf_ns_rename(char * oldName,
                      char * newName)
{
    NPF_NS_NODE * node;
    NPF_NS_NODE * oldNode;
    NPF_NS_NODE * sibling;
    NPF_NS_NODE * newParent;
    NPF_NS_NODE * oldParent;
    char *        delim;  
    char *        nodename;
    char          newParentPath[NPF_NS_MAXPATH];
    DListItr *    itr; 


    CheckInit();

    acquire(&g_nsLock, RW_WRITE);

    /* make sure the new name is not currently in use */
    node = GetNode(newName);
    if(node)
    {
        release(&g_nsLock);
        return NPF_NS_NODE_EXISTS;
    }    

    /* get the old node */
    node = GetNode(oldName);
    if(!node) 
    {
        release(&g_nsLock);   
        return NPF_INVALID_PARAMETERS;
    }

    /* find its new parent */
    delim = (char *)strrchr(newName, NPF_NS_NAMEDELIM);
    if(!delim)
    {
        release(&g_nsLock);
        return NPF_INVALID_PARAMETERS;
    }
    memset(newParentPath, 0, NPF_NS_MAXPATH);
    strncpy(newParentPath, newName, delim-newName);
    newParent = GetNode(newParentPath);
    if(!newParent)
    {
        release(&g_nsLock);
        return NPF_INVALID_PARAMETERS;
    }   

    delim++;

    /* prepare the new node name */
    nodename = (char *)PIL_AllocateMemory(((strlen(delim))+1), 
                                          PIL_MEM_ZERO_INIT);
    if(!nodename)
    {
        release(&g_nsLock);
        return NPF_OUT_OF_MEMORY;
    }   
    strcpy(nodename, delim);

    /* move the node out of its old parent's children */
    oldParent = node->pParent;
    if((itr = npf_list_itrCreate(&oldParent->children)) != NULL)
    {
        npf_list_itrFirst(itr); 
        do {
            sibling = (NPF_NS_NODE *)npf_list_itrGetData(itr);
            if(sibling == node)
            {
                npf_list_itrRemove(itr, (void **)&oldNode);
                break;
        }
        } while( npf_list_itrNext(itr) != -1);
        npf_list_itrDelete(itr);      
    }

    /* update name and add the node to its new parent's children list */
    PIL_FreeMemory(node->entry.name);
    node->entry.name = nodename;
    node->pParent = newParent;
    npf_list_pushBack(&newParent->children, node);

    
    release(&g_nsLock);               
    return NPF_SUCCESS; 

} /* npf_ns_rename */


/*****************************************************************************
*
*   Function: npf_ns_open()
*
*   Parameters:
*       name - null-terminated path string of a namespace node
*       hNode - output namespace handle to the node 
*
*   Description: open a namespace  node, which causing the ref count to 
*        the node increamented by 1. The caller must close it after using it.
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid 
*           NPF_COMPONENT_UNINITIALIZED if namespace is not initialized  
*           NPF_SUCCESS if successful    
*
*****************************************************************************/
#ifdef NSREFCOUNT
NPF_RET npf_ns_openWrapper( char *name, NPF_NSHDL *hNode, char *caller, int linenumber ) {
    NPF_RET ret;
    ret = npf_ns_openD( name, hNode );
    if ( ret == NPF_SUCCESS ) {
        IncNodeCount( *hNode, caller, linenumber );
    }
    return ret;
}
NPF_RET npf_ns_openD( char *name, NPF_NSHDL *hNode ) {
#else
NPF_RET npf_ns_open(char *      name,
                    NPF_NSHDL * hNode)
{
#endif
    NPF_NS_NODE * node;

    CheckInit();

    acquire(&g_nsLock, RW_WRITE);
    node = GetNode(name);
    if(node)
    {
        /* increament the ref count */
        node->entry.refCount++;
        *hNode = (NPF_NSHDL)node;
        release(&g_nsLock);
        return NPF_SUCCESS;    
    }

    release(&g_nsLock);
    return NPF_INVALID_PARAMETERS;

} /* npf_ns_open */


/*****************************************************************************
*
*   Function: npf_ns_getPath()
*
*   Parameters:
*       hNode - namespace handle to a node
*       path - null-terminated path string, allocated by caller
*
*   Description: Get the path string of a namespace node from an open handle. 
*        The caller must allocate the memory of size NPF_NS_MAXPATH for 
*        path string before calling this function.
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIALIZED if namespace is not initialized
*           NPF_OUT_OF_MEMORY if the system run out of memory
*           NPF_SUCCESS if successful
*
*****************************************************************************/
NPF_RET npf_ns_getPath(NPF_NSHDL hNode,
                       char * path)
{
    char * pathstr;
    NPF_NS_NODE * node;

    CheckInit();
 
    acquire(&g_nsLock, RW_READ);
 
    if(!isValid(hNode))
    {
        release(&g_nsLock);
        return NPF_INVALID_PARAMETERS;
    }
 
    node  = (NPF_NS_NODE *)hNode;   
    pathstr = GetPath(node);
    if(!pathstr)
    {
        release(&g_nsLock);
        return NPF_OUT_OF_MEMORY;
    }

    strcpy(path, pathstr);
    PIL_FreeMemory(pathstr); // allocated in GetPath()

    release(&g_nsLock);
    return NPF_SUCCESS;

} /* npf_ns_getPath */  


/*****************************************************************************
*
*   Function: foundNsHdl()
*
*   Parameters:
*       hNode - a namespace handle
*       node - a namespace node
*
*   Description: Check if the input handle is a handle to a node found in
*        the subtree g_nsRooted at the input node
*
*   Return: TRUE is found, FALSE if not
*
*****************************************************************************/
Bool foundNsHdl(NPF_NSHDL hNode, NPF_NS_NODE * node)
{
    DListItr *    itr;
    NPF_NS_NODE * child; 

    if(!node) return FALSE;
    if(hNode == (NPF_NSHDL)node) return TRUE;

    if((itr = npf_list_itrCreate(&(node->children))) == NULL) 
    {
        return FALSE;  
    }

    if(npf_list_itrFirst(itr) != -1)
    {
        do {
            child = (NPF_NS_NODE *)npf_list_itrGetData(itr);
            /* recursing to children */
            if(foundNsHdl(hNode, child)) 
            {
                npf_list_itrDelete(itr);
                return TRUE;
            }
        } while( npf_list_itrNext(itr) != -1);
    }
    npf_list_itrDelete(itr);   

    return FALSE;

} /* foundNsHdl */


/*****************************************************************************
*
*   Function: isValid()
*
*   Parameters:
*       hNode - a namespace handle
*
*   Description: Check if the input handle is valid
*
*   Return: TRUE is valid, FALSE if not
*
*****************************************************************************/
Bool isValid(NPF_NSHDL hNode)
{
    return foundNsHdl(hNode, g_nsRoot);
}


/*****************************************************************************
*
*   Function: npf_ns_getDataType()
*
*   Parameters:
*       hNode - a namespace handle
*       dataType - output data type of the node
*
*   Description: get the data type referenced by the node from an open
*        handle.
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIALIZED if namespace is not initialized  
*           NPF_SUCCESS if successful 
*
*****************************************************************************/
NPF_RET npf_ns_getDataType(NPF_NSHDL      hNode,
                           NPF_DATATYPE * dataType)
{
    NPF_NS_NODE * node;

    CheckInit();

    acquire(&g_nsLock, RW_READ); 

    if(!isValid(hNode)) 
    {
        release(&g_nsLock);
        return NPF_INVALID_PARAMETERS;
    }

    node  = (NPF_NS_NODE *)hNode;
    *dataType = node->entry.dataType;
    release(&g_nsLock);

    return NPF_SUCCESS; 

} /* npf_ns_getDataType */


/*****************************************************************************
*
*   Function: npf_ns_getDataHandle()
*
*   Parameters:
*       hNode - a open namespace handle to an instance node
*       hData - output data handle
*
*   Description: get the data handle referenced by the node
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIALIZED if namespace is not initialized  
*           NPF_SUCCESS if successful
*
*****************************************************************************/
NPF_RET npf_ns_getDataHandle(NPF_NSHDL    hNode,
                             NPF_HANDLE * hData)
{
    NPF_NS_NODE * node;

    CheckInit();

    acquire(&g_nsLock, RW_READ); 

    node = (NPF_NS_NODE *)hNode; 
    if((!isValid(hNode)) || (node->entry.nodeType != NPF_NS_INSTNODE)) 
    {
        release(&g_nsLock); 
        return NPF_INVALID_PARAMETERS;
    }

    *hData = node->entry.entryAttrib.hData; 
    release(&g_nsLock);

    return NPF_SUCCESS;

} /* npf_ns_getDataHandle */


/*****************************************************************************
*
*   Function: npf_ns_getParentNode()
*
*   Parameters:
*       hNode - an open namespace handle to an instance node
*       hParent - output open handle to the parent node
*
*   Description: Get an open handle to the parent node in the namespace 
*        hierarchy. The caller must close this handle after using it.
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIALIZED if namespace is not initialized
*           NPF_SUCCESS if successful
*
*****************************************************************************/
#ifdef NSREFCOUNT
NPF_RET npf_ns_getParentNodeWrapper(NPF_NSHDL   hNode,
                             NPF_NSHDL * hParent,
                             char *caller, int linenumber ) {
    NPF_RET ret;
    ret = npf_ns_getParentNodeD( hNode, hParent );
    if ( ret == NPF_SUCCESS ) {
        IncNodeCount( *hParent, caller, linenumber );
    }
    return ret;
}
NPF_RET npf_ns_getParentNodeD(NPF_NSHDL   hNode,
                             NPF_NSHDL * hParent) {
#else
NPF_RET npf_ns_getParentNode(NPF_NSHDL   hNode,
                             NPF_NSHDL * hParent)
{
#endif
    NPF_NS_NODE * node;
    NPF_NS_NODE * parent;

    CheckInit(); 
 
    acquire(&g_nsLock, RW_READ);
 
    node = (NPF_NS_NODE *)hNode;
    if(!isValid(hNode))
    {
        release(&g_nsLock);
        return NPF_INVALID_PARAMETERS;
    }
 
    parent = node->pParent;
    if(parent)
    {
        parent->entry.refCount++;
    }
    *hParent = (NPF_NSHDL)parent;
    release(&g_nsLock);

    return NPF_SUCCESS;
 
} /* npf_ns_getParentNode */ 


/*****************************************************************************
*
*   Function: npf_ns_getInstParent()
*
*   Parameters:
*       hNode - an open namespace handle to an instance node
*       hInstParent - output handle to the instant parent node
*
*   Description: For an instant node, get an open handle to the instant 
*        parent node, which is its grand parent in the namespace hierarchy.
*        The caller must close this handle after using it.
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIALIZED if namespace is not initialized
*           NPF_SUCCESS if successful
*
*****************************************************************************/
#ifdef NSREFCOUNT
NPF_RET npf_ns_getInstParentWrapper(NPF_NSHDL   hNode,
                             NPF_NSHDL * hInstParent,
                             char *caller, int linenumber ) {
    NPF_RET ret;
    ret = npf_ns_getInstParentD( hNode, hInstParent );
    if ( ret == NPF_SUCCESS ) {
        IncNodeCount( *hInstParent, caller, linenumber );
    }
    return ret;
}
NPF_RET npf_ns_getInstParentD(NPF_NSHDL   hNode,
                             NPF_NSHDL * hInstParent) {
#else
NPF_RET npf_ns_getInstParent(NPF_NSHDL   hNode,
                             NPF_NSHDL * hInstParent)
{
#endif
    NPF_NS_NODE * node;
    NPF_NS_NODE * parent;
 
    CheckInit();
 
    acquire(&g_nsLock, RW_READ);
 
    node = (NPF_NS_NODE *)hNode;
    if((!isValid(hNode)) || (node->entry.nodeType != NPF_NS_INSTNODE) 
                         || (node->pParent == NULL))
    {
        release(&g_nsLock);
        return NPF_INVALID_PARAMETERS;
    }
 
    parent = node->pParent->pParent;
    if(parent)
    {
        parent->entry.refCount++;
    }
    *hInstParent = (NPF_NSHDL)node->pParent->pParent;
    release(&g_nsLock);

    return NPF_SUCCESS;
 
} /* npf_ns_getInstParent */    


/*****************************************************************************
*
*   Function: npf_ns_readDir()
*
*   Parameters:
*       name - the null-terminated path string of a node
*       dirItr - node iterator for the directory
*
*   Description: read the directory part of the node by getting a
*        directory iterator. The caller must call npf_ns_done() after
*        using this iterator.
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIALIZED if namespace is not initialized  
*           NPF_OUT_OF_MEMORY if the system is out of memory
*           NPF_SUCCESS if successful
*
*****************************************************************************/
NPF_RET npf_ns_readDir(char *    name,
                       NODE_ITERATOR * dirItr)
{
    NPF_NS_NODE * node; 
    DListItr *    itr; 

    CheckInit();

    acquire(&g_nsLock, RW_READ); 
    node = GetNode(name); 
    if(node)
    {
        /* form the child iterator of the node */ 
        itr = npf_list_itrCreate(&node->children);
        /* check if the itr creation fails or addtion to g_dirTab fails */
        if((npf_list_size(&node->children) && !itr)  
           || (npf_list_pushFront(&g_dirTab, itr) == -1)) 
        {
            npf_list_itrDelete(itr);
            release(&g_nsLock); 
            return NPF_OUT_OF_MEMORY;
        }

        *dirItr = (NODE_ITERATOR)itr;
        release(&g_nsLock);

        return NPF_SUCCESS; 
    }

    release(&g_nsLock);
    return NPF_INVALID_PARAMETERS; 

} /* npf_ns_readDir */


/*****************************************************************************
*
*   Function: npf_ns_addAssociate()
*
*   Parameters:
*       hNode - an open namespace handle to a node
*       associate - a namespace node to be associated
*
*   Description: add an associate to the node
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIALIZED if namespace is not initialized
*           NPF_SUCCESS if successful
*
*****************************************************************************/
NPF_RET npf_ns_addAssociate(NPF_NSHDL hNode,
                            NPF_NSHDL associate)
{
    NPF_NS_NODE * node;
    NPF_NS_NODE * anode; // associated node
    NPF_NS_NODE * fellow;
    DListItr *    itr; 
    Bool          found;
 
    CheckInit();
 
    acquire(&g_nsLock, RW_WRITE);
 
    if(!isValid(hNode) || !isValid(associate))
    {
        release(&g_nsLock);
        return NPF_INVALID_PARAMETERS;
    }

    node  = (NPF_NS_NODE *)hNode; 
    anode = (NPF_NS_NODE *)associate; 

    /* check if anode has already associated */
    found = FALSE;
    if((itr = npf_list_itrCreate(&node->associates)) != NULL)
    {
        npf_list_itrFirst(itr);
        do {
            fellow = (NPF_NS_NODE *)npf_list_itrGetData(itr);
            if(fellow == anode)
            {
                found = TRUE;
                break;
            }
        } while( npf_list_itrNext(itr) != -1);
        npf_list_itrDelete(itr);
    }         

    if(!found)
    {
        npf_list_pushBack(&node->associates, anode);
    }

    release(&g_nsLock);

    return NPF_SUCCESS; 

} /* npf_ns_addAssociate */


/*****************************************************************************
*
*   Function: npf_ns_delAssociate()
*
*   Parameters:
*       hNode - an open namespace handle to a node
*       associate - handle to the accociated node
*
*   Description: delete an associate from the node
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIALIZED if namespace is not initialized
*           NPF_SUCCESS if successful
*
*****************************************************************************/
NPF_RET npf_ns_delAssociate(NPF_NSHDL hNode,
                            NPF_NSHDL associate)
{
    NPF_NS_NODE * node;
    NPF_NS_NODE * anode; // associated node
    NPF_NS_NODE * fellow;
    DListItr *    itr; 
    Bool          found;

    CheckInit(); 
 
    acquire(&g_nsLock, RW_WRITE);
 
    if(!isValid(hNode) || !isValid(associate))
    {
        release(&g_nsLock);
        return NPF_INVALID_PARAMETERS;
    }
 
    node  = (NPF_NS_NODE *)hNode;
    anode = (NPF_NS_NODE *)associate; 

    found = FALSE;
    if((itr = npf_list_itrCreate(&node->associates)) != NULL)
    {
        npf_list_itrFirst(itr);
        do {
            fellow = (NPF_NS_NODE *)npf_list_itrGetData(itr);
            if(fellow == anode)
            {
                found = TRUE;
                npf_list_itrRemove(itr, (void **)&anode);
                break;
            }
        } while( npf_list_itrNext(itr) != -1);
        npf_list_itrDelete(itr);     
    }

    release(&g_nsLock);

    if(found)
    {
        return NPF_SUCCESS; 
    }

    return NPF_INVALID_PARAMETERS;

} /* npf_ns_delAssociate */


/*****************************************************************************
*
*   Function: npf_ns_enumAssociate()
*
*   Parameters:
*       hNode - an open namespace handle to a node
*       associateItr - the output iterator of associates
*
*   Description: enumerate the associates of the node. If the node has no
*       associate, the output associateItr will be NULL.
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIALIZED if namespace is not initialized
*           NPF_OUT_OF_MEMORY if the system is out of memory
*           NPF_SUCCESS if successful
*
*****************************************************************************/
NPF_RET npf_ns_enumAssociate(NPF_NSHDL hNode,
                             NODE_ITERATOR * associateItr)
{
    NPF_NS_NODE * node;
    DListItr *    itr; 

    CheckInit();

    acquire(&g_nsLock, RW_READ);
 
    if(!isValid(hNode))
    {
        release(&g_nsLock);
        return NPF_INVALID_PARAMETERS;
    }
    node  = (NPF_NS_NODE *)hNode;  

    /* form the associate iterator of the node */
    if(npf_list_size(&node->associates) == 0)
    {
        *associateItr = 0x0;
        release(&g_nsLock);
        return NPF_SUCCESS;
    }
 
    if((itr = npf_list_itrCreate(&node->associates) ) == NULL)
    {
        release(&g_nsLock);
        /* refer to list.c for why */
        return NPF_OUT_OF_MEMORY;
    }
 
    /* this doesn't fail since itr != NULL */
    npf_list_pushFront(&g_assoTab, itr);  

    *associateItr = (NODE_ITERATOR)itr;  

    release(&g_nsLock);
    return NPF_SUCCESS; 
} /* npf_ns_enumAssociate */ 


/*****************************************************************************
*
*   Function: npf_ns_first()
*
*   Parameters:
*       dirItr - a directory handle
*
*   Description: Get an open namespec handle to the first child in the 
*        directory. The caller must close this handle after using it.
*
*   Return: The handle to the first child, or NULL if none.
*
*****************************************************************************/
#ifdef NSREFCOUNT
NPF_NSHDL npf_ns_firstWrapper(NODE_ITERATOR dirItr, char *caller, int linenumber) {
    NPF_NSHDL ret;
    ret = npf_ns_firstD( dirItr );
    if ( ret ) {
        IncNodeCount( ret, caller, linenumber);
    }
    return ret;
}
NPF_NSHDL npf_ns_firstD(NODE_ITERATOR dirItr) {
#else
NPF_NSHDL npf_ns_first(NODE_ITERATOR dirItr)
{
#endif

    NPF_NS_NODE * node;
    DListItr * itr = (DListItr *)dirItr;


    if(g_nsState != INITIALIZED)
    {
        return 0;
    }  

    if(npf_list_itrFirst(itr) == -1)
    {
        return 0;
    }
    else
    {
        node = npf_list_itrGetData(itr);
        if(node) 
        {
            node->entry.refCount++;
        }
            
        return (NPF_NSHDL)node;
    }
}


/*****************************************************************************
*
*   Function: npf_ns_next()
*
*   Parameters:
*       nodeItr - a node iterator
*
*   Description: Get an open namespace handle to the next child in the 
*        directory. The caller must close this handle after using it. 
*
*   Return: The handle to the next child, or NULL if none.
*
*****************************************************************************/
#ifdef NSREFCOUNT
NPF_NSHDL npf_ns_nextWrapper(NODE_ITERATOR nodeItr, char *caller, int linenumber) {
    NPF_NSHDL ret;
    ret = npf_ns_nextD( nodeItr );
    if ( ret ) {
        IncNodeCount( ret, caller, linenumber );
    }
    return ret;
}
NPF_NSHDL npf_ns_nextD(NODE_ITERATOR nodeItr) {
#else
NPF_NSHDL npf_ns_next(NODE_ITERATOR nodeItr)
{
#endif
    NPF_NS_NODE * node; 
    DListItr * itr = (DListItr *)nodeItr;


    if(g_nsState != INITIALIZED)
    {
        return 0;
    }  

    if(npf_list_itrNext(itr) == -1)
    {
        return 0;
    }
    else
    {
        node = npf_list_itrGetData(itr);
        if(node)
        {
            node->entry.refCount++;
        } 
        return (NPF_NSHDL)node;
    }
}


/*****************************************************************************
*
*   Function: npf_ns_prev()
*
*   Parameters:
*       nodeItr - a node iterator
*
*   Description: Get an open namespace handle to the previous child in 
*        the directory. The caller must close this handle after using it. 
*
*   Return: The handle to the previous child, or NULL if none.
*
*****************************************************************************/
#ifdef NSREFCOUNT
NPF_NSHDL npf_ns_prevWrapper(NODE_ITERATOR nodeItr, char *caller, int linenumber) {
    NPF_NSHDL ret;
    ret = npf_ns_prevD( nodeItr );
    if ( ret ) {
        IncNodeCount( ret, caller, linenumber );
    }
    return ret;
}
NPF_NSHDL npf_ns_prevD(NODE_ITERATOR nodeItr) {
#else
NPF_NSHDL npf_ns_prev(NODE_ITERATOR nodeItr)
{
#endif
    NPF_NS_NODE * node;
    DListItr * itr = (DListItr *)nodeItr;

    if(g_nsState != INITIALIZED)
    {
        return 0;
    }  

    if(npf_list_itrPrev(itr) == -1)
    {
        return 0;
    }
    else
    {
        node = npf_list_itrGetData(itr);
        if(node)
        {
            node->entry.refCount++;
        }
        return (NPF_NSHDL)node; 
    }   
}


/*****************************************************************************
*
*   Function: npf_ns_count()
*
*   Parameters:
*       nodeItr - a node iterator
*
*   Description: Get the count of children in the directory.
*
*   Return: The count if succesful, -1 otherwise.
*
*****************************************************************************/
int npf_ns_count(NODE_ITERATOR nodeItr)
{

    DListItr * itr = (DListItr *)nodeItr;

    if (!nodeItr) return 0;

    if(g_nsState != INITIALIZED)
    {
        return -1;
    }  

    return npf_list_itrListSize(itr);
}


/*****************************************************************************
*
*   Function: npf_ns_done()
*
*   Parameters:
*       nodeItr - a node iterator
*
*   Description: Done with this directory iterator. This must be called
*        after the client finishes using the directory iterator.
*
*   Return: none
*
*****************************************************************************/
void npf_ns_done(NODE_ITERATOR nodeItr)
{
    DListItr * itr = (DListItr *)nodeItr;
    DListItr * tabItr;
    DListItr * runItr;
    Bool found = FALSE;

    if(g_nsState != INITIALIZED)
    {
        return;
    }

    if((tabItr = npf_list_itrCreate(&g_dirTab)) != NULL)
    {
        npf_list_itrFirst(tabItr);
        do {
            runItr = (DListItr *)npf_list_itrGetData(tabItr);
            if(runItr == itr)
            {
                npf_list_itrRemove(tabItr, (void **)&runItr);
                found = TRUE;
                break;
            }
        } while( npf_list_itrNext(tabItr) != -1);
        npf_list_itrDelete(tabItr);
    }
 
    if(!found)
    {
        if((tabItr = npf_list_itrCreate(&g_assoTab)) != NULL)
        {
            npf_list_itrFirst(tabItr);
            do {
                runItr = (DListItr *)npf_list_itrGetData(tabItr);
                if(runItr == itr)
                {
                    npf_list_itrRemove(tabItr, (void **)&runItr);
                    break;
                }
            } while( npf_list_itrNext(tabItr) != -1);
            npf_list_itrDelete(tabItr);
        }
    }

    npf_list_itrDelete(itr);

} /* npf_ns_done */  


/*****************************************************************************
*
*   Function: npf_ns_close()
*
*   Parameters:
*       hNode - a namespace node handle
*
*   Description: Close this handle to the namespace node. This causes
*        the internal ref count decreamented by 1.
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIALIZED if namespace is not initialized  
*           NPF_SUCCESS if successful
*
*****************************************************************************/
#ifdef NSREFCOUNT
NPF_RET npf_ns_closeWrapper(NPF_NSHDL hNode, char *caller, int linenumber) {
    NPF_RET ret;
    ret = npf_ns_closeD( hNode );
    if ( ret == NPF_SUCCESS ) {
        DecNodeCount( hNode, caller, linenumber );
    }
    return ret;
}
NPF_RET npf_ns_closeD(NPF_NSHDL hNode) {
#else
NPF_RET npf_ns_close(NPF_NSHDL hNode)
{
#endif
    NPF_NS_NODE * node;

    CheckInit();

    acquire(&g_nsLock, RW_WRITE); 
    if(!isValid(hNode)) 
    {
        release(&g_nsLock);
        return NPF_INVALID_PARAMETERS;
    }

    node = (NPF_NS_NODE *)hNode;
    if(node->entry.refCount == 0)
    {
        release(&g_nsLock);
        return NPF_SUCCESS;
//        return NPF_INVALID_PARAMETERS; // Fix for bug
    }        
    node->entry.refCount--;
    release(&g_nsLock);
    return NPF_SUCCESS;
 
} /* npf_ns_close */


/*****************************************************************************
*
*   Function: npf_ns_getCanonName()
*
*   Parameters:
*       nameStr - null-terminated path string to a node
*       canonName - null-terminated canonical path string, allocated by
*                   caller     
*
*   Description: Get the canonical path string of a namespace node, typically,
*        an alias node. The caller must allocate the memory of size
*        NPF_NS_MAXPATH for canonName string before calling this function.
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIALIZED if namespace is not initialized  
*           NPF_SUCCESS if successful
*
*****************************************************************************/
NPF_RET npf_ns_getCanonName(char * nameStr,
                             char * canonName)
{
    NPF_NS_NODE * node; 
    char * path;

    CheckInit();

    acquire(&g_nsLock, RW_READ); 
    node = GetNode(nameStr);
    if(node)
    {
        if(node->entry.nodeType == NPF_NS_ALIASNODE) 
        {
            node = (NPF_NS_NODE *)node->entry.entryAttrib.hInstance;
            path = GetPath(node);
            if(path) 
            {
                strcpy(canonName, path);
                PIL_FreeMemory(path);
                release(&g_nsLock);
                return NPF_SUCCESS; 
            }
        }
        else
        {
            strcpy(canonName, nameStr);
            release(&g_nsLock);
            return NPF_SUCCESS;
        }
    }       

    release(&g_nsLock);
    return NPF_INVALID_PARAMETERS;

} /* npf_ns_getCanonName */ 


/*****************************************************************************
*
*   Function: ns_shutdown()
*
*   Parameters: none
*
*   Description: Shutdown namespace, called by PDK manager
*
*   Return:
*           NPF_COMPONENT_UNINITIALIZED if namespace is not initialized
*           NPF_SUCCESS if successful 
*
*****************************************************************************/
NPF_RET ns_shutdown()
{
//DListItr * itr;
DListItr * nodeItr;

    CheckInit();

    acquire(&g_nsLock, RW_WRITE);   

#ifdef NSREFCOUNT
    DumpRefCount();
    RefCountShutdown();
#endif

    /* clean up dangling dir iterators from the outstanding dir table */
    while(npf_list_popBack(&g_dirTab, (void **)&nodeItr) != -1)
        npf_list_itrDelete(nodeItr);

#if 0
    if((itr = npf_list_itrCreate(&g_dirTab)) != NULL)
    {
        npf_list_itrFirst(itr);
        do {
            nodeItr = (DListItr *)npf_list_itrGetData(itr);
            npf_list_itrDelete(nodeItr);
        } while( npf_list_itrNext(itr) != -1);
        npf_list_itrDelete(itr);
    }
#endif

    npf_list_destroy(&g_dirTab); 

    /* clean up dangling associate iterators from the outstanding asso table*/
    while(npf_list_popBack(&g_assoTab, (void **)&nodeItr) != -1)
        npf_list_itrDelete(nodeItr);

#if 0
    if((itr = npf_list_itrCreate(&g_assoTab)) != NULL)
    {
        npf_list_itrFirst(itr);
        do {
            nodeItr = (DListItr *)npf_list_itrGetData(itr);
            npf_list_itrDelete(nodeItr);
        } while( npf_list_itrNext(itr) != -1);
        npf_list_itrDelete(itr);
    }
#endif

    npf_list_destroy(&g_assoTab);
    
    RemoveNode(g_nsRoot);
    g_nsState = UNINITIALIZED;
    release(&g_nsLock);
    destroy(&g_nsLock);

    return NPF_SUCCESS; 

} /* ns_shutdown */
