//////////////////////////////////////////////////////////////////////////////
//           INTEL CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or
//  nondisclosure agreement with Intel Corporation and may not be copied
//  or disclosed except in accordance with the terms of that agreement.
//      Copyright (c) 1999-2000 Intel Corporation.  All Rights Reserved.
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////

/****************************************************************************
 * $Source: /CODEBASE/ControlPlane/CoreBlocks/NameSpace/src/ns_helper.c
 * $Revision: 1.6 $
 * $Date: 2003/11/17 16:57:13 $
 * 
 * ns_helper.c
 * 
 * Set of Namespace helper routines
 ****************************************************************************/

#include "pdk_ns_helper.h"
#include "pdk_macros.h"
#include "log.h"
#include <assert.h>
#include <string.h>

#ifdef NSHLPR_DEBUG
#define PDBG_PRINT(fmt, args...) printf(fmt,## args)
#define PDBG_TRACE() DBG_TRACE () /* macro */
#define PRETURN(val) RETURN(val) /* macro */
#else
#define PDBG_PRINT(fmt, args...) /* nothing */
#define PDBG_TRACE() /* nothing */
#define PRETURN(val) return (val) /* return normally */
#endif

/* Internal structure for mapping an interface to FE(s) and port(s) */

typedef struct _pdk_ns_feport_ 
{
    NPF_HANDLE fe_hdl;
    FPPI_FEID  fe_id;
    uint32_t   num_ports;
    DList      ports;
} pdk_fe_ports_t;

static NPF_RET assoc_FEPorts(NODE_ITERATOR port_itr, 
                             DList* fes,
                             int* fe_count, 
                             int* port_count );


/****************************************************************************
 * Function: get_fe_handle()                                                *
 * -------------------------                                                *
 * Parameters: NPF_HANDLE - handle to interface object                      *
 *                                                                          *
 * Description: Internal function to determine the FE that the specified    *
 *              Interface belongs to. This function has to traverse the NS  *
 *              to follow the Associates of the Interface object. The       *
 *              Associate of an Interface must be a PORT object. The parent *
 *              of the Port node must be the FE the interface belongs to    *
 *                                                                          *
 * Return: NPF_HANDLE of the FE this Interface belongs to                   *
 *                                                                          *
 * Assumtions: An interface belongs to only one FE. Code needs to be        *
 *             modified if this is not the case.                            *
 *                                                                          *
 ****************************************************************************/

/* Return fe_npf_hdl in open status, caller should close it */

NPF_RET pdk_ns_getFE_from_If(NPF_HANDLE intf_npf_hdl, 
                             NPF_HANDLE* fe_npf_hdl )
{
    NPF_NSHDL     intf_ns_hdl, fe_ns_hdl, port_ns_hdl;
    NPF_RET       npf_ret;
    NODE_ITERATOR intf_itr;
    int num_ports = 0;

    PDBG_TRACE();
    /*
      - have to get the FE ID of the FE this interface
        belongs to
      - using the interface NPF_HANDLE, get the NSHDL
      - enumerate the associates of the interface
      - the associate must be a port object
      - get the port's parent (must be a FE)
      - using the FE handle, get its ID
    */

    /* get NS_HDL of INTERFACE using the NPF_HANDLE */
    npf_ret = cm_getNSHDLByHandle(intf_npf_hdl, &intf_ns_hdl);
    if(npf_ret != NPF_SUCCESS) 
    {
        PDBG_PRINT ("Could not get NS hdl from C&M \n");
        PRETURN (npf_ret);
    }
    
    /* read the associates - get an iterator */
    npf_ret = npf_ns_enumAssociate(intf_ns_hdl, &intf_itr);
    if(npf_ret != NPF_SUCCESS) 
    {
        PDBG_PRINT ("Could not enumerate Intf associates \n");
        PRETURN (npf_ret);
    }
    num_ports = npf_ns_count(intf_itr);
    if(num_ports > 0) 
    {
        port_ns_hdl = npf_ns_first(intf_itr);
        npf_ret = npf_ns_getInstParent(port_ns_hdl, &fe_ns_hdl);
        /* check npf_ret */
        npf_ret = npf_ns_getDataHandle(fe_ns_hdl, fe_npf_hdl);
        /* check npf_ret */
        npf_ns_close(port_ns_hdl);
    }
    npf_ns_done(intf_itr);

    PDBG_PRINT ("Got FE handle \n");
    PRETURN (npf_ret);
}

/****************************************************************************
 * Function: pdk_ns_getChildByType()                                        *
 * ------------------------                                                 *
 * Parameters: root_npf_hdl - handle to the root object
 *             child_type   - type of the child we are looking for
 *  Out Param: Child Handle + Child Path  
 *                                                                          *
 * Description: This is an internal/external func which is used to iterate  *
 *              to find Child By Type                                       *
 *                                                                          *
 * Return: NPF_RET indicating success of the function                       *
 *                                                                          *
 ****************************************************************************/

NPF_RET pdk_ns_getChildByType (NPF_HANDLE root_npf_hdl,
                               NPF_DATATYPE child_type,
                               char         *node_path,
                               NPF_NSHDL    *child_hdl 
                               )
{
    NPF_RET       npf_ret;
    NPF_NSHDL     root_ns_hdl, node_hdl;
    NPF_DATATYPE  node_type;
    NODE_ITERATOR node_itr;
    int           num_children = 0, child_found = 0;

    PDBG_TRACE();

    /* get NS_HDL of ROOT using the NPF_HANDLE */
    npf_ret = cm_getNSHDLByHandle( root_npf_hdl, &root_ns_hdl );
    if(npf_ret != NPF_SUCCESS) 
    {
        PDBG_PRINT ("Could not get NS hdl of Root node from CM\n");
        PRETURN (npf_ret);
    }

    /* get the path of the NS node */
    npf_ret = npf_ns_getPath( root_ns_hdl, node_path );
    if(npf_ret != NPF_SUCCESS) 
    {
        PDBG_PRINT ("Could not get node path from NS\n");
        PRETURN (npf_ret);
    }

    /* use the path to read the children of this node */
    npf_ret = npf_ns_readDir( node_path, &node_itr );
    if(npf_ret != NPF_SUCCESS) 
    {
        PDBG_PRINT ("Could not read node directory\n");
        PRETURN (npf_ret);
    }

    /* get the first node (child node) of ROOT */
    node_hdl = npf_ns_first(node_itr);
    num_children = npf_ns_count(node_itr);

    /* check if this matches the type passed in
       Possible types - PORT, ARP_TABLE, ROUTER, INTERFACE, etc.
       if not., search through all the children
    */
    PDBG_PRINT ("Checking the child's type...\n");

    npf_ret = npf_ns_getDataType( node_hdl, &node_type );
    if ( node_type == child_type ) 
    {
        /* get the path of the NS node */
        npf_ns_getPath( node_hdl, node_path );
        npf_ns_close (node_hdl);
        child_found = 1;
    }
    else
    {
        while (node_hdl != (NPF_NSHDL) NULL) 
        {
            npf_ns_close (node_hdl);
            node_hdl = npf_ns_next(node_itr);
            npf_ret = npf_ns_getDataType( node_hdl, &node_type );
            if ( node_type == child_type ) 
            {
                PDBG_PRINT ("Found child with matching TYPE\n");
                child_found = 1;
                break;
            } /* end if (node type) */
        } /* end while (more children) */
    } /* end if (type not matched) */

    npf_ns_done (node_itr);
    if (child_found == 1) 
    {
        *child_hdl = node_hdl;
        PRETURN (NPF_SUCCESS);
    }
    PDBG_PRINT ("Child with the relevant type not found !! \n");
    PRETURN (NPF_NS_NODE_NOTFOUND); /* node not found */
}

/****************************************************************************
 * Function: pdk_ns_getChildItrByType                                       *
 * ------------------------                                                 *
 * Parameters: NPF_HANDLE - handle to FE object                             *
 *  Out Param: NODE_ITERATOR: Iterator to enumerate the ports of this FE    *
 *                                                                          *
 * Description: Internal function to iterate through the ports of a FE      *
 *              Given the NPF_HANDLE of a FE, this function returns an      *
 *              enumerator to its children (port nodes). This fn. is used   *
 *              by other functions of this module.                          *
 *                                                                          *
 * Return: NPF_RET indicating success of the function                       *
 *                                                                          *
 ****************************************************************************/

NPF_RET pdk_ns_getChildItrByType (NPF_HANDLE root_npf_hdl,
                                  NPF_DATATYPE child_type,
                                  NODE_ITERATOR* child_itr )
{
    NPF_RET       npf_ret;
    NPF_NSHDL     node_hdl;
    char          node_path[NPF_NS_MAXPATH];

    PDBG_TRACE();
    /* Get the node by the TYPE */
    if((npf_ret = pdk_ns_getChildByType(root_npf_hdl, child_type,node_path,
                            &node_hdl)) != NPF_SUCCESS)
    {
       PDBG_PRINT ("Could not get the child\n");
       PRETURN (NPF_NS_NODE_NOTFOUND); /* node not found */
    }
    PDBG_PRINT ("Got the Child at the path %s \n", node_path);

    /* now get the path of this child TYPE node & read its dir */
    npf_ret = npf_ns_getPath( node_hdl, node_path );
    npf_ns_close ( node_hdl );
    if(npf_ret != NPF_SUCCESS) 
    {
         PDBG_PRINT ("Could not read path of child\n");
         PRETURN (npf_ret);
    }

    /* use the path to read the children of this node */
    npf_ret = npf_ns_readDir( node_path, child_itr );
    if(npf_ret != NPF_SUCCESS) 
    {
        PDBG_PRINT ("Could not read directory\n");
        PRETURN (npf_ret);
    }

    PRETURN (NPF_SUCCESS);
}


/****************************************************************************
 * Function: get_fe_interfaces()                                            *
 * -----------------------------                                            *
 * Parameters: NPF_HANDLE - handle to FE object                             *
 *  Out Param: *num_intfs - number of interfaces of this FE                 *
 *  Out Param: NPF_HANDLE* - (Ptr to) array of NPF_HANDLEs to INTERFACEs    *
 *                          that belong to this FE                          *
 *                                                                          *
 * Description: Internal function to determine the interfaces that belong   *
 *              to this FE. Given the NPF_HANDLE of a FE, this function     *
 *              traverses the Namespace for each PORT of the FE, to find    *
 *              out the interfaces they correspond to.                      *
 *                                                                          *
 * Return: NPF_RET indicating success of the function                       *
 *                                                                          *
 * Assumtions: An interface belongs to only one FE. Code needs to be        *
 *             modified if this is not the case.                            *
 *                                                                          *
 ****************************************************************************/

NPF_RET pdk_getFE_Intfs (NPF_HANDLE fe_npf_hdl,
                         int* num_intfs,
                         NPF_HANDLE* fe_intfs )
{
    NPF_RET       npf_ret;
    NPF_NSHDL     port_hdl, intf_ns_hdl;
    NPF_HANDLE    intf_hdl;
    NODE_ITERATOR port_itr, intf_itr;
    int           num_ports=0, i = 0, intfs_count=0, j = 0;

    PDBG_TRACE();
    /*
      use internal fn to read the directory of the FE (children)
      get_port_itr() returns an iterator to the list of all
      PORT (instances) nodes of the FE
    */
    npf_ret = pdk_ns_getChildItrByType (fe_npf_hdl, NPF_PORT, &port_itr);
    if(npf_ret != NPF_SUCCESS) 
    {
        PDBG_PRINT ("Could not get port instance from FE\n");
        PRETURN (npf_ret);
    }

    num_ports = npf_ns_count(port_itr);
    PDBG_PRINT ("Number of ports = %d \n", num_ports);   

    port_hdl = npf_ns_first(port_itr);

    /* first we have to find out the total number of interfaces
       since we need to allocate that much memory

       Note: We need to traverse through all the ports, finding
       out the number of interfaces it has. 
    */
    while (++i < num_ports) 
    {
        /* read the associates - get an iterator */
        npf_ret = npf_ns_enumAssociate(port_hdl, &intf_itr);
        if(npf_ret != NPF_SUCCESS) 
        {
            PDBG_PRINT ("Could not enumerate port's associates\n");
            PRETURN (npf_ret);
        }

        *num_intfs = npf_ns_count(intf_itr);
        intfs_count += *num_intfs;
        npf_ns_done(intf_itr);
    }
    PDBG_PRINT ("Total number of interfaces = %d\n", intfs_count);
                       
    /* allocate memory for NUM_PORTS * NUM_INTFS */
    fe_intfs = (NPF_HANDLE *) PIL_AllocateMemory (num_ports* intfs_count * 
                      sizeof(NPF_HANDLE), PIL_MEM_ZERO_INIT);
    if (fe_intfs == NULL)
    PRETURN (NPF_OUT_OF_MEMORY);

    /*
      Now loop through each port again, filling in the 
      interface handles....Any better way to do this?? 

      for each port of the FE
          read the associates structure of the port
      for each associate of the port
          get the NS_HDL of the INTERFACE
          get the NPF_HANDLE of the INTERFACE
    */
    i = 0;
    while (++i < num_ports) 
    {
        /* read the associates - get an iterator */
        npf_ret = npf_ns_enumAssociate(port_hdl, &intf_itr);
        if(npf_ret != NPF_SUCCESS) 
        {
            PDBG_PRINT ("Could not enumerate port's associates\n");
            PRETURN (npf_ret);
        }
        *num_intfs = npf_ns_count(intf_itr);
        intf_ns_hdl = npf_ns_first(intf_itr);
    
        while(++j<*num_intfs) 
        {
            PDBG_PRINT ("Getting interface data\n");
            npf_ret = npf_ns_getDataHandle(intf_ns_hdl, &intf_hdl); 
            if(npf_ret != NPF_SUCCESS) 
            {
                PDBG_PRINT ("Could not get data handle for interface node\n");
                PRETURN (npf_ret);
            }

            fe_intfs[intfs_count++] = intf_hdl;
            npf_ns_close(intf_ns_hdl);
            intf_ns_hdl = npf_ns_next(intf_itr);
        } /* end while (intfs) */

        npf_ns_done(intf_itr);
        port_hdl = npf_ns_next(port_itr);
    } /* end while (ports) */

    npf_ns_done(port_itr);
    PDBG_PRINT ("Done reading all interfaces\n");
    PRETURN (npf_ret);
}


NPF_RET pdk_ns_getPort_from_ID (uint32_t fe_id, 
                                FPPI_PortID port_id,
                                NPF_HANDLE* port_npf_hdl )
{
    NPF_HANDLE fe_npf_hdl;
    NPF_NSHDL fe_ns_hdl, port_ns_hdl;
    NODE_ITERATOR port_itr;
    int num_ports = 0, i = 0;
    uint32_t prt_id;
    NPF_RET ret;

    PDBG_TRACE();
    /* 
       from the FE ID, get all the ports that belong to the FE
       For each port, see if the port's ID matches the port id we recvd.
    */
    cm_getFEHandleById (fe_id, &fe_npf_hdl);
    PDBG_PRINT("Got FE's NPF handle::: FE id = %d \t FE NPF HDL = %d\n",
                fe_id, fe_npf_hdl);
    ret = cm_getNSHDLByHandle(fe_npf_hdl, &fe_ns_hdl);
    if(ret != NPF_SUCCESS) 
    {
        PDBG_PRINT ("Cannot get FE's NS handle\n" );
        /* cannot insert packet */
        PRETURN (ret);
    }

    /*
      use internal fn to read the directory of the FE (children)
      get_port_itr() returns an iterator to the list of all
      PORT (instances) nodes of the FE
    */
    ret = pdk_ns_getChildItrByType (fe_npf_hdl, NPF_PORT, &port_itr);
    if(ret != NPF_SUCCESS) 
    {
        PRETURN (ret);
    }

    num_ports = npf_ns_count(port_itr);
    port_ns_hdl = npf_ns_first(port_itr);
    /*
      for each port of the FE
          get the port's NPF handle
      access the PORT structure in C&M
      check if port id matches
    */
    while (++i < num_ports) 
    {
        ret = npf_ns_getDataHandle(port_ns_hdl, port_npf_hdl);
        /* the above port_npf_hdl should be &port_npf_hdl! */
        if(ret != NPF_SUCCESS) 
        {
            npf_ns_close(port_ns_hdl);
            npf_ns_done(port_itr);
            PRETURN (ret);
        }

        /* check the port's ID */
        ret = cm_port_get_portid (*port_npf_hdl, &prt_id);
        if(ret != NPF_SUCCESS) 
        {
            npf_ns_close(port_ns_hdl);
            npf_ns_done(port_itr);
            PRETURN (ret);
        }

        if (prt_id == (uint32_t) (port_id)) 
        {
            npf_ns_close (port_ns_hdl);
            npf_ns_done(port_itr);
            PRETURN (  NPF_SUCCESS );
        }
        port_ns_hdl = npf_ns_next(port_itr);
    } /* end while (intfs) */

    PRETURN (NPF_NS_NODE_NOTFOUND);
}


/* From Interface Handle, get all FEs it belongs to
   also all ports on each of the FEs it corresponds to

   Memory will be allocated as required for the pdk_FEPortList_t
   structure

   Memory has to be freed by caller as desired using PIL_FreeMemory

   if_npf_hdl is open and pdk_FEPortList_t.pdk_port_t[].port_hdl
   returns close status, also pdk_FEPortList_t.fe_hdl is close too
*/

NPF_RET pdk_ns_getFEPorts_from_IF(NPF_HANDLE         if_npf_hdl,
                                  int*               num_fes,
                                  pdk_FEPortList_t** fe_ports )
{
    NPF_RET           ret = NPF_FAIL;
    NPF_NSHDL         if_ns_hdl;
    NODE_ITERATOR     port_itr;
    pdk_port_t*       port = NULL;
    pdk_fe_ports_t*   fe = NULL;
    DList             fes;
    DListItr*         itr, *portlist_itr;
    int               fe_count = 0, num_fe = 0, port_count = 0;
    NPF_DATA          data;
    pdk_FEPortList_t* temp_fe_ports = NULL;

    PDBG_TRACE ();

    *num_fes = 0;

    /* initialize the list of FEs */
    npf_list_init(&fes, PIL_FreeMemory);

    /* first, get all associations of this interface */
    ret = cm_getNSHDLByHandle(if_npf_hdl, &if_ns_hdl);
    if(ret != NPF_SUCCESS) 
    {
        PDBG_PRINT ("Could not get IF NS hdl from C&M \n");
        PRETURN (ret);
    }
    
    /* read the associates - get an iterator */
    ret = npf_ns_enumAssociate(if_ns_hdl, &port_itr);
    if(ret != NPF_SUCCESS) 
    {
        PDBG_PRINT ("Could not enumerate Intf associates \n");
        PRETURN (ret);
    }

    PDBG_PRINT ("Associating all ports and FEs now\n");
    ret = assoc_FEPorts (port_itr, &fes, &fe_count, &port_count);
    PDBG_PRINT("assoc_FEPorts() returned::: Num fes = %d \t Num ports = %d\n", 
                fe_count, port_count);

    /* now, transfer all this info to the struct we have to return */
    if(fe_count)
    {
       temp_fe_ports = (pdk_FEPortList_t*) PIL_AllocateMemory (
                                fe_count * sizeof (pdk_FEPortList_t),
                                PIL_MEM_ZERO_INIT);
       assert(temp_fe_ports); 
       num_fe = 0;

       /* iterate through DList of FEs we just built & populate the 
       structure we have to return
        */
       port_count = 0;
       PDBG_PRINT ("Transferring all info into return data structure\n");
       itr = npf_list_itrCreate (&fes);
       if(!itr)
      	  return NPF_FAIL;
       if (npf_list_itrFirst (itr) != -1) 
       {
           do 
           {
               PDBG_PRINT ("For FE Number %d\n", num_fe);

               fe = (pdk_fe_ports_t*) npf_list_itrGetData(itr);
               temp_fe_ports[num_fe].fe_hdl = fe->fe_hdl;
               temp_fe_ports[num_fe].fe_id = fe->fe_id;
               temp_fe_ports[num_fe].num_ports = fe->num_ports;

               if(fe->num_ports)
               {
                  temp_fe_ports[num_fe].ports = (pdk_port_t*) 
                       PIL_AllocateMemory ( fe->num_ports*sizeof(pdk_port_t), 
                           PIL_MEM_ZERO_INIT);
                  assert (temp_fe_ports[num_fe].ports);

                  portlist_itr = npf_list_itrCreate (&(fe->ports));
                  assert (portlist_itr);
                  port_count = 0;
                  if (npf_list_itrFirst (portlist_itr) != -1) 
                  {
                   do {
                       PDBG_PRINT ("\t Creating port number %d\n", port_count);
                       port = (pdk_port_t*) npf_list_itrGetData(portlist_itr);
                       temp_fe_ports[num_fe].ports[port_count].port_id = port->port_id;
                       temp_fe_ports[num_fe].ports[port_count++].port_hdl = port->port_hdl;
                    } while (npf_list_itrNext(portlist_itr) != -1);
                 }/* if (iterating through ports of FE */
                 npf_list_itrDelete(portlist_itr);
              }
              num_fe++;
           } while (npf_list_itrNext(itr) != -1);
        }/* if (iterating through the list of FEs */
      npf_list_itrDelete(itr);
    } /* fe_count */
    
    *num_fes = num_fe;
    PDBG_PRINT ("Copied into return structure:: Num of FEs = %d\n", *num_fes);
    PDBG_PRINT ("Now freeing temporary lists\n");
    /* now free up all temporary storage we allocated */

    while (npf_list_popFront(&fes, (void**)&data) != -1) 
    {
        fe = (pdk_fe_ports_t*) data;
        while (npf_list_popFront (&(fe->ports), (void**)&data) != -1)
        {
            port = (pdk_port_t*) data;
            PIL_FreeMemory (port);
        }
        PIL_FreeMemory (fe);
    }
    npf_list_destroy (&fes);
    
    *fe_ports = temp_fe_ports;
    PRETURN (NPF_SUCCESS);
}

/* fes.fehdl and fes.port.porthdl is return w/o open status */
static NPF_RET assoc_FEPorts (NODE_ITERATOR port_itr, 
                              DList* fes,
                              int* fe_count, 
                              int* port_count )
{
    NPF_RET         ret;
    NPF_DATATYPE    type;
    NPF_NSHDL       port_ns_hdl, fe_ns_hdl;
    NPF_HANDLE      port_npf_hdl, fe_npf_hdl;
    pdk_port_t*     port = NULL;
    int             num_ports = 0, num_fes = 0, fe_found = 0;
    uint32_t        fe_id, port_id;
    pdk_fe_ports_t* fe;
    DListItr*       felist_itr;
    
    PDBG_TRACE ();

    num_ports = npf_ns_count(port_itr);
    *port_count = num_ports;
    PDBG_PRINT ("# of ports associated with this iterator = %d\n",
        num_ports);
    /* assert (num_ports);*/
    if(!num_ports)
        return(NPF_FAIL);
    port_ns_hdl = npf_ns_first(port_itr);
    PDBG_PRINT ("Now iterating over each associate \n");

    do 
    {
        PDBG_PRINT ("Iteration num = %d\n", i++);
        /* get the data handle for this port */
        ret = npf_ns_getDataHandle (port_ns_hdl, &port_npf_hdl);
        if (NPF_SUCCESS != ret) 
        {
            PDBG_PRINT ("Could not get port NPF hdl from NS\n");
            npf_ns_close (port_ns_hdl);
            npf_ns_done (port_itr);
            PRETURN (ret);
        }

        cm_port_get_portid(port_npf_hdl, &port_id);
        PDBG_PRINT ("Port id  = %d\n", port_id);
        /* get this port's parent (FE)*/
        ret = npf_ns_getInstParent(port_ns_hdl, &fe_ns_hdl);
        /* check this ret */
	npf_ns_getDataType(fe_ns_hdl, &type);
	if(type != NPF_FE)
	{
	    continue;
	}
        port = (pdk_port_t*)PIL_AllocateMemory(
                                sizeof(pdk_port_t),
                                PIL_MEM_ZERO_INIT);
        assert(port); /* no assert, please! */
        port->port_id = port_id;
        port->port_hdl = port_npf_hdl;

        ret = npf_ns_getDataHandle(fe_ns_hdl, &fe_npf_hdl);
        if (NPF_SUCCESS != ret) 
        {
            npf_ns_close(fe_ns_hdl);
            PDBG_PRINT ("Unable to obtain FE handles from NS\n");
            PRETURN (ret);
        }

        /* close fe handle got from ns_getInstParent */
        npf_ns_close(fe_ns_hdl);
        cm_fe_get_feid(fe_npf_hdl, &fe_id);
        PDBG_PRINT ("Port %d belongs to FE ID %d\n", port_id, fe_id);

        if (npf_list_size (fes) == 0) 
        {
            PDBG_PRINT ("No FEs found yet for this interface\n");
            /* no FEs yet, so add a new one */
            fe = (pdk_fe_ports_t*) PIL_AllocateMemory (sizeof(pdk_fe_ports_t),
                               PIL_MEM_ZERO_INIT);
            assert (fe); /* change assert to error checking! */
            fe->fe_hdl = fe_npf_hdl;
            fe->fe_id = fe_id;
            fe->num_ports = 1;
            npf_list_init(&fe->ports, PIL_FreeMemory);
            npf_list_pushBack(&(fe->ports), port);
            npf_list_pushBack(fes, fe);
            num_fes++;
            *fe_count = num_fes;
            PDBG_PRINT("Created new Fe and added this port to the FE\n");
        }
        else
        {
            PDBG_PRINT("Checking if this FE exists in FE list\n");
            /* first see if this FE is already in the list */
            felist_itr = npf_list_itrCreate (fes);
            assert (felist_itr);
            if (npf_list_itrFirst (felist_itr) != -1) 
            {
                do {
                    fe = (pdk_fe_ports_t*) npf_list_itrGetData (felist_itr);
                    if (fe->fe_id == fe_id) 
                    {
                        fe_found = 1;
                        PDBG_PRINT ("Found FE with FE id = %d\n", fe_id);
                        fe->num_ports++;
                        npf_list_pushBack (&(fe->ports), port);
                        PDBG_PRINT ("Added port to the found FE\n");
                
                        npf_list_itrDelete (felist_itr);
                        break;
                    }/* end (if FE FOUND) */
                } while (npf_list_itrNext(felist_itr) != -1);
                /* 
                   if we could not find the FE, then this is a new FE, so add this
                */
                if (!fe_found) 
                {
                    PDBG_PRINT ("Could not find FE. Creating new FE and adding port\n");
                    fe = (pdk_fe_ports_t*) PIL_AllocateMemory (sizeof(pdk_fe_ports_t),
                            PIL_MEM_ZERO_INIT);
                    assert (fe); /* replace assert by error checking */
                    fe->fe_id = fe_id;
                    fe->fe_hdl = fe_npf_hdl;
                    fe->num_ports = 1;
                    npf_list_init(&fe->ports, PIL_FreeMemory);

                    /* add the new port to the FE */
                    npf_list_pushBack(&(fe->ports), port);
                    fe->num_ports++;
                    num_fes++; /* new FE */
                    *fe_count = num_fes;
                    npf_list_pushBack (fes, fe);
                } /* end if(FE was not found) */
            } /* end if (iterating thru the FE list */
            npf_list_itrDelete (felist_itr);
        }        
        /* END CODE PASTE from DEBUGGING */
 
        PDBG_PRINT ("After adding port, FE count = %d\n", *fe_count);

        /* Closing the handle in NS */
        npf_ns_close (port_ns_hdl);
    } while ((port_ns_hdl = npf_ns_next(port_itr)) != 0); 
    /* end (for all port associates) */
    /* Should let the caller close this
       npf_ns_close (port_ns_hdl); */
    npf_ns_done (port_itr);
    
    PRETURN (NPF_SUCCESS);
}

/* Caller has to close the handle after using it */
NPF_RET pdk_ns_getNSRootHandle (NPF_NSHDL *root_ns_hdl)
{
    NPF_NSHDL ns_root;
    NODE_ITERATOR node_itr;
    char path[NPF_NS_MAXPATH];
    NPF_RET npf_ret;

    PDBG_TRACE();
    strcpy(path, "/");

    // get the root path first
    npf_ret = npf_ns_open(path, &ns_root);
    if(npf_ret != NPF_SUCCESS) 
    {
        PDBG_PRINT ("Unable to open NS root!");
        PRETURN (NPF_NS_NODE_NOTFOUND);
    }

    npf_ret = npf_ns_getPath(ns_root, path);
    if(npf_ret != NPF_SUCCESS) 
    {
        PDBG_PRINT ("Unable to get root path!");
        PRETURN (NPF_NS_NODE_NOTFOUND);
    }
    /* 
       read the root's directory - essentially get the path of the
       first instance of the system
    */
    npf_ret = npf_ns_readDir(path, &node_itr);
    if(npf_ret != NPF_SUCCESS) 
    {
        PDBG_PRINT ("Unable to get root path!");
        PRETURN (NPF_NS_NODE_NOTFOUND);
    }

    *root_ns_hdl = npf_ns_first(node_itr);
    if (*root_ns_hdl != (NPF_NSHDL)NULL)
    {
        PDBG_PRINT ("Found root node\n");
    }

    npf_ns_done(node_itr);
    PRETURN (NPF_SUCCESS);
}

/* Get all Active FEs in system */

NPF_RET pdk_ns_get_FEs(int         *numFEs,
                       FPPI_FEID  feArray[])
{
    NPF_NSHDL       sys_ns_hdl, fe_ns_hdl;
    NPF_HANDLE      sys_npf_hdl, fe_npf_hdl;
    NODE_ITERATOR   feItr;
    int             feCount = 0;

    if (cm_sys_get_numfes(numFEs) != NPF_SUCCESS)
    {
        return NPF_FAIL;
    }
  
    if (pdk_ns_getNSRootHandle(&sys_ns_hdl) != NPF_SUCCESS)
    {
       PDBG_PRINT("Could not get System Root NS Handle \n");
       return NPF_FAIL;
    }
  
    if (npf_ns_getDataHandle(sys_ns_hdl, &sys_npf_hdl) != NPF_SUCCESS)
    {
        PDBG_PRINT("Could not get System Root Data Handle \n");
        npf_ns_close(sys_ns_hdl);
        return NPF_FAIL;
     }
  
     if(pdk_ns_getChildItrByType(sys_npf_hdl, NPF_FE, &feItr) != NPF_SUCCESS)
     {
        npf_ns_close(sys_ns_hdl);
        return NPF_FAIL;
     }
  
     // In failure handle returned is 0
     fe_ns_hdl = npf_ns_first(feItr);
     
     while(fe_ns_hdl)
     {
        if(npf_ns_getDataHandle(fe_ns_hdl, &fe_npf_hdl) == NPF_SUCCESS)
        {
           if (cm_fe_get_feid(fe_npf_hdl, &feArray[feCount]) == -1)
           {
              PDBG_PRINT("Could not get System Root Data Handle \n");
              return NPF_FAIL;
           }
        }
        fe_ns_hdl = npf_ns_next(feItr);
        feCount++;
     }
     
     return NPF_SUCCESS;
}
