//////////////////////////////////////////////////////////////////////////////
//           INTEL CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or
//  nondisclosure agreement with Intel Corporation and may not be copied
//  or disclosed except in accordance with the terms of that agreement.
//      Copyright (c) 1999-2000 Intel Corporation.  All Rights Reserved.
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////

/******************************************************************************
 * $Source: /cvs/constellations/ixa_3.0/src/cp_pdk/ControlPlane/CoreBlocks/PacketHandler/src/ph_tun.c,v $
 * $Revision: 1.6 $:                                                        
 * $Date: 2003/11/17 16:57:13 $:                                            
 * 
 * pkt_hdlr.c
 * -----------------                                                        
 * This is the CE Packet Handler module. It is responsible for delivering
 * tunneled data packets to the corresponding virtual interface, and listening
 * for packets transmitted on any virtual interface and subsuquently tunneling
 * them to the correct FE for transmission on the correct egress interface. 
 * 
 * Assumptions:
 *   1. It is initialized by the PDK manager
 *   2. It registers callbacks with the VIDD COntroller module
 *   3. The Virtual Interfaces are accessible as normal files - open(), close(),
 *      read(), write(), etc.
 *   4. OS dependencies-this module makes file I/O calls (open, close), standard
 *      I/O calls like select(), etc. to interact with the virtual interfaces
 ******************************************************************************/

#include "pil.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#ifndef _VXWORKS_
#include <sys/time.h>
#endif
#ifdef PH_TIMING
#include <sys/poll.h>
#endif
#include <assert.h>

#include "pdk_macros.h"
#include "log.h"
#include "npf_ns_types.h"
#include "npf_ns_api.h"
#include "pkt_hdlr.h"
#include "pdk_cm.h"
#include "fppapi.h"
#include "pdkmgr.h"

/* GLobals variables */

/* context for VI create/delete event */
static NPF_USERCONTEXT   g_vi_Event_context; 
 /* Callback handle for VI Create/delete events*/
static NPF_CBHANDLE      g_vi_Event_handle;
/* context for FPPI_Packet Recv event */
static NPF_USERCONTEXT   g_ph_PktRecv_context; 
 /* Callback handle for FPPI_Packet Recv event */
static NPF_CBHANDLE      g_ph_PktRecv_handle;
/* context for FPPI_SendPacket call */
static NPF_USERCONTEXT   g_ph_PktSend_context; 
 /* Callback handle for FPPI_SendPacket call */
static NPF_CBHANDLE      g_ph_PktSend_handle;

/* global variable to signal thread completion */
static uint32_t g_continue = 1;

/* list of virtual interfaces to perform packet I/O on */
static DList g_vi_List;
static PilCriticalSection g_ph_mutex;

/* global var for component initialization */
static uint32_t g_initialized = 0;

/* Packet Listen thread */
static PilThread g_ph_thread;

#define PH_THRD_STACK_SIZE 1024 /* stack size */
#define PH_THRD_NAME       "Packet Listen Thread" /* name of thread */
#define PH_THRD_PRIORITY   4 /* medium priority (1-10) */

#define MAX_NUM_INTFS      40 /* total number of virtual interfaces */
#define MAX_PKT_LEN        1520 /* just over the MTU of an Eth interface */
#define MAXFD(a,b)         ( (a)>(b) ? (a):(b))

/* Macros for debugging. Enabled from Makefile */

#ifdef PKT_HDLR_DEBUG
#define PPDBG_PRINT(fmt, args...) printf(fmt,## args)
#define PPDBG_TRACE() DBG_TRACE() /* macro */
#endif

#ifdef PH_DEBUG
#define PDBG_PRINT(fmt, args...) printf(fmt,## args)
#define PDBG_TRACE() DBG_TRACE() /* macro */
#define return(val) return(val) 
#else
#define PDBG_PRINT(fmt, args...) /* nothing */
#define PDBG_TRACE() /* nothing */
#define return(val) \
   { \
     return val;\
    } /* return normally */
#endif

/* Tap file descriptor array, static set all 0 */
static int fd_list[MAX_NUM_INTFS];
static int num_fds, max_fd;
static fd_set fdset;

/*****************************************************************************
* Name: CHK_INIT
*
* Description:
*     Utility macro used to check if the 
*     component is properly initialized.
*
*/
#define CHK_INIT() \
    if ( !g_initialized ) {\
        return NPF_COMPONENT_UNINITIALIZED;\
    }

/* functions that PH registers as callbacks*/
void on_vi_Event ( NPF_USERCONTEXT context, PDK_VIDD_EVT_CBTYPE cbtype,
           NPF_DATA* response );
void on_PktRecv_Event ( FPPI_FEID fe_id, FPPI_CORRELATOR cb_correlator,
            FPPI_CONTEXT cb_context, void* response_data );
void on_PktSend_CB ( FPPI_FEID fe_id, FPPI_CORRELATOR cb_correlator,
             FPPI_CONTEXT cb_context, void* response_data );

/* Internal functions */
static void     ph_print_pdk_vi ( PDK_VIC_t* vi );
static void     ph_print_vi ( ph_vif_t* vi );
static void     ph_print_packets ( npf_PacketList_t* pkt_list );
static uint32_t ph_pkt_listen ( void* );
static void     ph_write_packet ( FPPI_FEID fe_id, FPPI_PortID port_id,
                                  int len, char* buffer );

/****************************************************************************
 * Function: ph_init()                                                      *
 * --------------------                                                     *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: Invoked by PDK manager during the initialization of the     *
 *              Control plane of the PDK.                                   *
 * Assumption:  Assumes that VIC module has already been initialized - this *
 *              is required for registering different callbacks             *
 *                                                                          *
 * Return: 0 on success, error values from registration functions           *
 *                                                                          *
 ****************************************************************************/

NPF_RET ph_init()
{
    NPF_RET ret;
    int     i;

    PDBG_TRACE();

    if (g_initialized) {
        return (1);
    }

    /* must init before pthread created */
    g_initialized = 1;

    PIL_InitializeCriticalSection(&g_ph_mutex);

    g_vi_Event_context = (void*) PDK_VIDD_EVT_VI_EVENT;
    g_ph_PktRecv_context = (void*) FPPI_PacketIO_Event;
    
    /* register callback with FPPI for processing tunneled packets */

    /* Register Pkt Recv callback with FPP API */
    ret = FPPAPI_register_callback( &on_PktRecv_Event,
                    &g_ph_PktRecv_handle,
                    FPPI_PacketIO_Event,
                    g_ph_PktRecv_context );

    /* Register Pkt Send callback with FPP API */
    ret = FPPAPI_register_callback( &on_PktSend_CB,
                    &g_ph_PktSend_handle,
                    FPPI_PacketIO_Event,
                    g_ph_PktSend_context );

#ifdef _LINUX_
    /* currently only have a vic for LINUX */
    /* register for VI events category with VIDD Controller */
    ret = vic_vi_event_register_callback ( g_vi_Event_context,
                       (NPF_GENERAL_CBFUNC) &on_vi_Event,
                       &g_vi_Event_handle );
#endif /* _LINUX_ */

    /* initialize list of virtual interfaces */
    npf_list_init (&g_vi_List, PIL_FreeMemory);
    for (i=0; i<MAX_NUM_INTFS; i++)
        fd_list[i] = 0;
    num_fds = 0;
    max_fd  = 0;

    /* initialize the file desriptor set */
    FD_ZERO (&fdset);
    
    /* initialize the Pkt Listen thread */
    g_ph_thread = PIL_CreateThread ( &ph_pkt_listen, /* thread fn. */
                       PH_THRD_STACK_SIZE, /*stack size */
                       PH_THRD_NAME, /* name of thread */
                       PH_THRD_PRIORITY, /* priority */
                       NULL, /* arguments to thread */  
                       0 ); /* options */ 

    return (ret);
}

/****************************************************************************
 * Function: ph_shutdown()                                                  *
 * ------------------------                                                 *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: Invoked by PDK manager during the shutdown of the           *
 *              Control plane of the PDK.                                   *
 * Assumption:  Assumes that VIC module is not yet shutdown - this          *
 *              is required for deregistering different callbacks           *
 *                                                                          *
 * Return: 0 on success, error values from deregistration functions         *
 *                                                                          *
 ****************************************************************************/

NPF_RET ph_shutdown()
{
    NPF_RET ret;
    NPF_DATA data;
    ph_vif_t* p_vif;

    PDBG_TRACE();
    CHK_INIT();
    g_initialized = 0;

    PIL_EnterCriticalSection (&g_ph_mutex);

    /* signal thread to stop */
    g_continue = 0;
    
    PIL_LeaveCriticalSection (&g_ph_mutex);
    if (PIL_TIMEOUT == PIL_WaitForThread(g_ph_thread, EXITTIMEOUT))
          PIL_TerminateThread(g_ph_thread);
    PIL_DestroyThread(g_ph_thread, NULL);

#ifdef _LINUX_
    /* deregister from VI events */
    ret = vic_vi_event_deregister_callback (g_vi_Event_handle);
#endif /* _LINUX */

    /* deregister from FPPI events */
    ret = FPPAPI_deregister_callback (g_ph_PktRecv_handle);

    /* clean up local data structures */
    /* iterate through list & deallocate all objects */
    PIL_EnterCriticalSection (&g_ph_mutex);
    while (npf_list_popFront(&g_vi_List, (void**)&data) != -1) 
    {
        p_vif = (ph_vif_t*) data;
        PIL_FreeMemory(p_vif->feport_list->ports);
        PIL_FreeMemory(p_vif->feport_list);
        PIL_FreeMemory(p_vif);
    }
    npf_list_destroy (&g_vi_List);
    PIL_LeaveCriticalSection (&g_ph_mutex);
    PIL_DestroyCriticalSection (&g_ph_mutex);

    return (ret);
}

void on_vi_Event ( NPF_USERCONTEXT context,
           PDK_VIDD_EVT_CBTYPE cbtype,
           NPF_DATA* response )
{
    int i;
    PDK_VIC_t* pdk_vi;
    ph_vif_t* p_vif;
    DListItr* itr;
    
    PDBG_TRACE ();

    if (!g_initialized)
	return (/*void*/);

    if (context != (void*) PDK_VIDD_EVT_VI_EVENT)
    return (/*void*/);

    /* Update local data structure depending on whether a VI was
       created or deleted */
    pdk_vi = (PDK_VIC_t*) response;
    ph_print_pdk_vi (pdk_vi);

    if (cbtype == PDK_VIDD_EVT_VI_CREATE) {

    PDBG_PRINT ("Recvd VI Create event from VIC\n");

    p_vif = (ph_vif_t *) PIL_AllocateMemory (sizeof(ph_vif_t), 
                         PIL_MEM_ZERO_INIT);
    if (p_vif == NULL) {
        PDBG_PRINT ("Pkt Hdlr: Out of memory\n");
        return (/*void*/);
    }
    //PDBG_PRINT ("Copying data from callback\n");

    strcpy( p_vif->vi_name, pdk_vi->name );
    p_vif->intf_hdl = pdk_vi->intf_hdl;
    p_vif->ipaddr   = pdk_vi->ipaddr;
    p_vif->mask     = pdk_vi->mask;
    p_vif->isValid  = TRUE;
    p_vif->vi_fd    = pdk_vi->fd;
    memcpy(p_vif->hwaddr, pdk_vi->hwaddr, ETH_ALEN);
    pdk_ns_getFEPorts_from_IF (p_vif->intf_hdl,
                  &(p_vif->num_fes),
                  &p_vif->feport_list);

    /* print out the ports per FE */
    ph_print_vi (p_vif);

    PIL_EnterCriticalSection (&g_ph_mutex);

    num_fds++;

    if (p_vif->vi_fd > max_fd)
        max_fd = p_vif->vi_fd;

    for (i=0; i<MAX_NUM_INTFS; i++)
        if (fd_list[i] == 0) {
            fd_list[i] = p_vif->vi_fd;
            /* turn on the bits in the fdset struct for select */
            FD_SET(fd_list[i], &fdset);
            break;
        }
    if (i == MAX_NUM_INTFS) {
        PDBG_PRINT("Too many virtual devices\n");
        return (/* void */);
    }

    /* add the new virtual interface to the list of VIs 
       packet handler thread has to interact with */
    npf_list_pushBack(&g_vi_List, p_vif);
    PIL_LeaveCriticalSection (&g_ph_mutex);

    PDBG_PRINT ("Added new VI to list \n");
    } /* end if VI CREATE event */

    else if (cbtype == PDK_VIDD_EVT_VI_DELETE) {

    PDBG_PRINT ("Recvd VI Create event from VIC\n");
    /* search through the list for this VI and set
       the flag to INVALID  */
    p_vif = NULL;
    /* Now we have to wait till the pkt listen thread
       finishes one iteration, thereby closing all
       open descriptors. Only then should we return
       back to VIC
    */
    PIL_EnterCriticalSection (&g_ph_mutex);

    /* Delete the VI struct maintained by Pkt Handler
       by Pkt Handler */
    if ((itr = npf_list_itrCreate(&g_vi_List)) == NULL)
        return (/*void*/);

    if (npf_list_itrFirst(itr) != -1) {
	do {
	    p_vif = (ph_vif_t*) npf_list_itrGetData(itr);
	    if ( strcmp(p_vif->vi_name, pdk_vi->name) == 0) {
		PDBG_PRINT ("Found match for VI\n");

                /* remove this fd from fd_list and fdset */
                for (i=0; i<MAX_NUM_INTFS; i++)
                    if (fd_list[i] == p_vif->vi_fd) {
                        fd_list[i] = 0;
                        FD_CLR(p_vif->vi_fd, &fdset);
                        break;
                    }

                /* reset the max_fd if needed */
                if (p_vif->vi_fd == max_fd) {
                    max_fd = 0;
                    for (i=0; i<MAX_NUM_INTFS; i++)
                        if (fd_list[i] > max_fd)
                            max_fd = fd_list[i];
                }

		p_vif->isValid = FALSE;
                PIL_FreeMemory(p_vif->feport_list->ports);
                PIL_FreeMemory(p_vif->feport_list);
		npf_list_itrRemoveFree (itr);
                num_fds--;
		break;
	    }
	} while (npf_list_itrNext(itr) != -1);
    }
    npf_list_itrDelete(itr);
    PIL_LeaveCriticalSection (&g_ph_mutex);
    }  /* end if VI DELETE event */
    return (/*void*/);
}

/****************************************************************************
 * Function: ph_pkt_listen()                                                *
 * ------------------------                                                 *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: Thread started during initialization of the Packet Handler  *
 *              Responsible for listening for packets transmitted on all    *
 *              virtual interface devices. On recving a packet from the     *
 *              device, this function tunnels it through the FP Plugin API  *
 *                                                                          *
 * Assumptions:                                                             *
 *                                                                          *
 * Return: None                                                             *
 *                                                                          *
 ****************************************************************************/
static uint32_t ph_pkt_listen ( void* arg )
{
    int             nselect=0, nread =0, found, checked_fds;
    int             i = 0, j = 0, k = 0;
    char            buffer[MAX_PKT_LEN];
    struct timeval  timeout; /* for performing select() */
    uint32_t        fe_id, port_id, num_ports = 0, pkt_count = 0;
    FPPI_CORRELATOR pkt_snd_corr = 0;
    FPPI_RET        ret;
    npf_Packet_t    packet;
    DListItr*       itr;
    ph_vif_t*       p_vif;
    fd_set          rset;
#ifdef PH_TIMING
    struct timeval  strt_time, end_time;
    float           total_time;
#endif

    PDBG_TRACE ();
    CHK_INIT ();

    timeout.tv_sec = 5;
    timeout.tv_usec = 0; /* timeout for select() - we dont wait for
                any device to become readable */
    while(g_continue) {

    /* we have to yield if the PDK thread wants to execute something in 
       the Pkt Handler - creating/deleting a VI, etc. */

    /* Do not remove this line - since we remove the mutex should be ok */
    poll(0, 0, 5);
    /* sched_yield(); */

#ifdef PH_TIMING
    gettimeofday (&strt_time, 0);
#endif

      /* Try not to lock here but later to speed up the 
       * critical path. But may cause trouble when pull off/on
       * the link while packet are still transmitting */
//    PIL_EnterCriticalSection (&g_ph_mutex);

    if (num_fds <= 0) {
//	PIL_LeaveCriticalSection (&g_ph_mutex);
        continue;
    }

    PDBG_PRINT ("Number of devices = %d \t MAX fd = %d \n",
            num_fds, max_fd);

    /* Everytime needs to re-copy the rset from global fdset */
    memcpy(&rset, &fdset, sizeof(fd_set));
    
    /* now we have to select on all these devices */
    nselect = select (max_fd+1, &rset, NULL, NULL, &timeout);
//    nselect = select (max_fd+1, &rset, NULL, NULL, NULL);

    PDBG_PRINT ("Select return value = %d \n", nselect);

    if (nselect <= 0) {
//        PIL_LeaveCriticalSection (&g_ph_mutex);
        continue;
    }

    checked_fds = 0;
    /* some devices are readable */
    for(i=0; checked_fds<num_fds && i<MAX_NUM_INTFS; i++) {

        if (fd_list[i] == 0)
            continue;

        checked_fds++;

        if (!FD_ISSET(fd_list[i], &rset))
            continue;

        nread = read (fd_list[i], buffer, MAX_PKT_LEN);

        if (nread <= 0) {
            PDBG_PRINT ("Error from device %d: %s\n",
                    nread, strerror (nread));
            continue;
        }

        PDBG_PRINT ("Read %d bytes from device %d\n",
            nread, fd_list[i]);

        /* map the FD to the vi device */
        PIL_EnterCriticalSection (&g_ph_mutex);
        itr = npf_list_itrCreate(&g_vi_List);
        assert (itr);
        if (npf_list_itrFirst(itr) == -1) {
            npf_list_itrDelete(itr);
            PIL_LeaveCriticalSection (&g_ph_mutex);
            continue;
        }

        found = 0;
        do {
            p_vif = (ph_vif_t*) npf_list_itrGetData(itr);
            /* check if the FD matches */
            if (p_vif->vi_fd == fd_list[i]) {
                found = 1;
                break;
            }
        } while (npf_list_itrNext(itr) != -1);
        npf_list_itrDelete(itr);
        PIL_LeaveCriticalSection (&g_ph_mutex);

        if (!found)
            continue;

        PDBG_PRINT ("Device %d (%s) read returned %d bytes\n",
            fd_list[i], p_vif->vi_name, nread);

        for (j = 0; j < p_vif->num_fes; j++) {
            fe_id = p_vif->feport_list[j].fe_id;
            num_ports = 
                p_vif->feport_list[j].num_ports;
            for( k = 0; k < num_ports; k++ ) {
                port_id = 
                    p_vif->feport_list[j].ports[k].port_id;

                packet.length = nread;
                packet.portid = 
                    (FPPI_PortID) port_id;
                packet.buffer = buffer;
                pkt_count = 1;
                /* Now send the packet */
               
                PDBG_PRINT ( "Sending packet of "
                         "length %d out of "
                         "port ID %d on FE ID "
                         "%d\n",
                         nread, port_id, 
                         fe_id );

#ifdef PH_TIMING
                gettimeofday (&end_time, 0);
                total_time = ((end_time.tv_sec - strt_time.tv_sec)*1000000
                          + (end_time.tv_usec - strt_time.tv_usec));
                printf ("Time to set up dev fds, read packet from"
                    "device = %f\n", total_time);

                gettimeofday (&strt_time, 0);
#endif

                ret =  FPPAPI_SendPacket (fe_id, pkt_snd_corr,
                        g_ph_PktSend_handle,
                        &packet, pkt_count);
                PDBG_PRINT ("Sent pkt\n");
#ifdef PH_TIMING
                gettimeofday (&end_time, 0);
                total_time = ((end_time.tv_sec - strt_time.tv_sec)*
                    1000000+(end_time.tv_usec - strt_time.tv_usec));
                printf ("Time to send packet (FPPI) = %f\n",total_time);
#endif

                if (ret != FPPI_SUCCESS)
                    PDBG_PRINT ("Error sending packet"
                       "Err = %d\n",
                        ret);
            } /* end (for all ports of this FE) */
        }/* end (for all FEs the VI belongs to */
    } /* for each fd */
    
//    PIL_LeaveCriticalSection (&g_ph_mutex);
    } /* end while (1) */
    PIL_ExitThread(0);
    return ((uint32_t)0);    
}


/****************************************************************************
 * Function: on_PktRecv()                                                   *
 * ----------------------                                                   *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: Callback function registered with FP Plug in API. Invoked   *
 *              when FPPI recvs a tunneled data packet. This function will  *
 *              inject the packet into the correct virtual interface device *
 *              The VI device delivers it into the IP stack subsequently.   *
 *                                                                          *
 * Assumptions: Packet recvd is a complete IP packet and is recvd as a flat *
 *              unsigned char buffer.                                       *
 *                                                                          *
 * Return: None                                                             *
 *                                                                          *
 ****************************************************************************/
void on_PktSend_CB ( FPPI_FEID fe_id, 
             FPPI_CORRELATOR cb_correlator,
             FPPI_CONTEXT cb_context, 
             void* response_data )
{
    PDBG_TRACE();
    if (!g_initialized)
	return (/*void*/);    
    return (/*void*/);
}

/****************************************************************************
 * Function: on_PktRecv()                                                   *
 * ----------------------                                                   *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: Callback function registered with FP Plug in API. Invoked   *
 *              when FPPI recvs a tunneled data packet. This function will  *
 *              inject the packet into the correct virtual interface device *
 *              The VI device delivers it into the IP stack subsequently.   *
 *                                                                          *
 * Assumptions: Packet recvd is a complete IP packet and is recvd as a flat *
 *              unsigned char buffer.                                       *
 *                                                                          *
 * Return: None                                                             *
 *                                                                          *
 ****************************************************************************/
void on_PktRecv_Event ( FPPI_FEID fe_id, 
            FPPI_CORRELATOR cb_correlator,
            FPPI_CONTEXT cb_context, 
            void* response_data )
{
    int               i=0, buf_len = 0, port_id = 0;
    npf_Packet_t*     pkt;
    npf_PacketList_t* fppi_pkt_list;
#ifdef PH_TIMING
    struct timeval    strt_time, end_time;
    float             total_time;
#endif

    PDBG_TRACE ();
    if (!g_initialized)
	return (/*void*/);

#ifdef PH_TIMING
    gettimeofday (&end_time, 0);
#endif
    fppi_pkt_list = (npf_PacketList_t*) response_data;
    PDBG_PRINT ("Recvd packets from FE %d \n", fe_id);
    ph_print_packets (fppi_pkt_list);

    /* check the context first */
    if (cb_context != g_ph_PktRecv_context)
    return (/*void*/);

    PIL_EnterCriticalSection (&g_ph_mutex);

    /* response data will be a list of packets recv. this is of type
       npf_Packetlist */
    pkt = fppi_pkt_list->packet_list;
    for(i=0; i<fppi_pkt_list->packet_count; i++)
    {
        /* get the packet */
        buf_len = pkt->length;
        port_id = pkt->portid;
        
        /* write this packet into VI device */
        ph_write_packet (fe_id, port_id, pkt->length, pkt->buffer);
        pkt += 1; 
    }
    PIL_LeaveCriticalSection (&g_ph_mutex);
#ifdef PH_TIMING
    gettimeofday (&end_time, 0);
    total_time = ((end_time.tv_sec - strt_time.tv_sec) * 1000000 +
		  (end_time.tv_usec - strt_time.tv_usec));
    printf ("Time to write acket into device = %f\n", total_time);
#endif
    return (/*void*/);
}

/****************************************************************************
 * Function: ph_write_packet()                                              *
 * ---------------------------                                              *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: Function to inject a packet into a virtual interface device *
 *              Invoked by on_PktRecv() when Pkt Hdlr recvs a tunneled pkt  *
 *              from FP Plug in.                                            *
 *                                                                          *
 * Parameters: FE ID - fe the packet was recvd from                         *
 *             Port id - port on FE on which packet was recvd on            *
 *             Buffer - the packet itself                                   *
 *                                                                          *
 * Assumptions: Packet recvd is a complete IP packet and is recvd as a flat *
 *              unsigned char buffer.                                       *
 *                                                                          *
 * Return: None                                                             *
 *                                                                          *
 ****************************************************************************/
static void ph_write_packet (FPPI_FEID fe_id, FPPI_PortID port_id,
                             int len, char* buffer)
{
    NPF_RET ret;
    NPF_HANDLE port_npf_hdl; /* handle corresponding to port_id */
    NPF_HANDLE if_npf_hdls[MAXIFOWNPORT]; /* corresponding interface handles */
    int num_intfs = 0, i = 0, nwrite = 0;
    DListItr *itr;
    ph_vif_t *p_vif;
    int found;

    PDBG_TRACE ();
    ret = pdk_ns_getPort_from_ID ( (uint32_t)fe_id, 
                   port_id,
                   &port_npf_hdl );
    if (ret != NPF_SUCCESS)
        return (/*void*/);

    /* use internal CM function to get a list of all interfaces
       that correspond to this port */
    cm_getIFByPort (port_npf_hdl, &num_intfs, if_npf_hdls);
    
    for(i=0; i<num_intfs; i++) {

        PIL_LeaveCriticalSection (&g_ph_mutex);
        /* search the g_vi_List for same interface handle */
        if ((itr = npf_list_itrCreate(&g_vi_List)) == NULL) {
            PIL_LeaveCriticalSection (&g_ph_mutex);
            continue;
        }

        if (npf_list_itrFirst(itr) == -1) {
            npf_list_itrDelete(itr);
            PIL_LeaveCriticalSection (&g_ph_mutex);
            continue;
        }

        found = 0;
        do {
            p_vif = (ph_vif_t*) npf_list_itrGetData(itr);
            if (p_vif->intf_hdl == if_npf_hdls[i]) {
                found = 1;
                break;
            }
        } while (npf_list_itrNext(itr) != -1);
        npf_list_itrDelete(itr);
        PIL_LeaveCriticalSection (&g_ph_mutex);

        if (!found)
            continue;

        PDBG_PRINT ("VI name = %s\n", p_vif->vi_name);
        if (strlen (p_vif->vi_name) > 0) {
            memcpy(buffer, p_vif->hwaddr, ETH_ALEN);
            nwrite = write (p_vif->vi_fd, buffer, len);
            PDBG_PRINT ("Written %d bytes into device %s\n", 
                nwrite, p_vif->vi_name);
            if (nwrite <= 0) {
            PDBG_PRINT ("Error writing packet to device : %s\n",
                    strerror (nwrite));
            } /* end if nwrite < 0 */
        } /* end if VI Name */
    } /* end for loop */
    return (/*void*/);
}

/* helper functions to print VI data */
static void ph_print_vi ( ph_vif_t* vi )
{
    PDBG_TRACE();
    PDBG_PRINT ("VI Name = %s \n", vi->vi_name);
    PDBG_PRINT ("VI Intf = %d \n", vi->intf_hdl);
    PDBG_PRINT ("VI IP   = 0x%x \n", vi->ipaddr);
    PDBG_PRINT ("VI mask = 0x%x \n", vi->mask);
    PDBG_PRINT ("VI Valid= %d \n", vi->isValid);
    PDBG_PRINT ("VI fd   = %d \n", vi->vi_fd);
    return(/*void*/);
}

static void ph_print_pdk_vi ( PDK_VIC_t* vi )
{
    PDBG_TRACE();
    PDBG_PRINT ("VI Name = %s \n", vi->name);
    PDBG_PRINT ("VI Intf = %d \n", vi->intf_hdl);
    PDBG_PRINT ("VI FE   = %d \n", vi->fe_hdl);
    PDBG_PRINT ("VI IP   = 0x%x \n", vi->ipaddr);
    PDBG_PRINT ("VI mask = 0x%x \n", vi->mask);
    return(/*void*/);
}

static void ph_print_packets ( npf_PacketList_t* pkt_list )
{
    int i;
    npf_Packet_t* pkt;

    PDBG_TRACE();
    PDBG_PRINT ("Number of packets = %d \n", pkt_list->packet_count);
    pkt = pkt_list->packet_list;
    for(i=0; i<pkt_list->packet_count;i++) {
    PDBG_PRINT ("Pkt Num %d \n \t Pkt len = %d \n \t Port id = %d \n",
            i, pkt->length, pkt->portid);
    pkt += 1;
    }
    return(/*void*/);
}
