/*****************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 1998-2003 By Intel Corporation.  All rights reserved.  
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *****************************************************************************/

#include "pil.h"

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <string.h>

#include <ioLib.h>
#include <sockLib.h>
#include <pipeDrv.h>
#include <netBufLib.h>

#include "fpp_common.h"
#include "fppapi.h"
#include "log.h"

#include "vxworks_vidd.h"

#define PH_IP_PROTOCOL 105
#define MAX_MESSAGE_LEN 1500

/* globals */
static int g_ph_socketd;
static int g_shutdownPipe[2];

#define SHUTDOWN_PIPE_NAME "/pipe/packethandler"
#define SHUTDOWN_PIPE_MSG "exit"

static PilCriticalSection g_ph_lock;
static PilThread g_ph_RcvrThread;

static NET_POOL_ID g_pdk_vidd_netPoolID;

typedef struct ph_header_t
{
    FPPI_PortID portid;
    uint32_t    length;
} ph_header;

typedef struct ph_feaddr_t{
    FPPI_FEID           feid;
    struct sockaddr_in  addr;
} ph_feaddr;



/* transport plugin callback stuff */
void ph_fe_bind(
            FPPI_FEID feid, 
            FPPI_CORRELATOR cb_correlator,
            FPPI_CONTEXT cb_context, 
            void* response_data
            );
void ph_fe_unbind(
            FPPI_FEID feid, 
            FPPI_CORRELATOR cb_correlator,
            FPPI_CONTEXT cb_context, 
            void* response_data
            );
static NPF_CBHANDLE      g_ph_fe_bind_handle;
static NPF_CBHANDLE      g_ph_fe_unbind_handle;



/* hash stuff */
/* magic hash size, prime so hopefully that will help */
#define HASH_TABLE_LEN 17

ph_feaddr g_ph_fe_hashtable[HASH_TABLE_LEN];
int g_ph_fe_hashkey_addr[HASH_TABLE_LEN];

#define HASH_FUNC_FEID(a) (((uint32_t)(a))%HASH_TABLE_LEN)
#define HASH_LOOKUP_BY_FEID(a) (g_ph_fe_hashtable + (HASH_FUNC_FEID(a)))

#define HASH_FUNC_ADDR(a) (((uint32_t)(a))%HASH_TABLE_LEN)
#define HASH_LOOKUP_BY_ADDR(a) (g_ph_fe_hashtable + g_ph_fe_hashkey_addr[(HASH_FUNC_ADDR(a))])


int ph_convertToMblk(char* data, int len, M_BLK_ID * pMblk)
{
    /**
     * Get an mBlk-clMblk-cluster construct (VxWorks buffer) from the memory pool.
     * If we cannot get a construct, return -1.
     */
    if (!g_pdk_vidd_netPoolID)
    {
        pdk_vidd_getNetPool(&g_pdk_vidd_netPoolID);
    }
    if (!g_pdk_vidd_netPoolID)
        return -1;

    *pMblk = netTupleGet(g_pdk_vidd_netPoolID, (int)len, M_DONTWAIT, MT_DATA, FALSE);
    if(*pMblk == NULL)
        return -1;

    /* IP header is not double-word aligned, must copy data */
    bcopy(data,(*pMblk)->mBlkHdr.mData, len);

    (*pMblk)->mBlkHdr.mFlags |= M_PKTHDR;
    (*pMblk)->mBlkHdr.mLen = len;
    (*pMblk)->mBlkPktHdr.len = len;

    return 0;
}

uint32_t ph_receiver_thread(void* a)
{
    char buf[MAX_MESSAGE_LEN + 20];
    int sd = (int)a;
    struct sockaddr_in from;
    int len;
    int ret;
    struct ip *iph;
    ph_header *ph;
    ph_feaddr *pf;
    M_BLK_ID mBlk;

    fd_set rfds;
    int maxfd, retval;

    maxfd = sd;
    if (g_shutdownPipe[0] > sd) maxfd = g_shutdownPipe[0];
    maxfd++;

    while(1)
    {
        FD_ZERO(&rfds);
        FD_SET(g_shutdownPipe[0],&rfds);
        FD_SET(sd,&rfds);

        retval = select(maxfd, &rfds, NULL, NULL, NULL);

        if (retval < 0)
        {
            npf_logger_write(LOG_PH, LOG_ERR, "Error returned from select\n");
            return -1;
        }
        if (FD_ISSET(sd,&rfds))
        {
            ret = recvfrom(sd,buf,sizeof(buf),0,(struct sockaddr*)&from,&len);
            if (ret < 0)
            {
                npf_logger_write(LOG_PH, LOG_ERR, "Error from recvfrom %d",errno);
                return 0;
            }
            iph = (struct ip*)buf;
    
            npf_logger_write(LOG_PH, LOG_TRACE,
                "protocol %x, saddr %x, daddr %x, len %d\n",
                iph->ip_p,iph->ip_src,iph->ip_dst,ntohl(iph->ip_len));

            ph = (ph_header*)(buf + sizeof(struct ip));

            PIL_EnterCriticalSection(&g_ph_lock);

            pf = HASH_LOOKUP_BY_ADDR(iph->ip_src.s_addr);

            npf_logger_write(LOG_PH, LOG_TRACE,
                "portid %d, length %d\n",
                ntohl(ph->portid),ntohl(ph->length));

            if (0 == 
                ph_convertToMblk((char*)((int)ph + sizeof(ph_header)),ntohl(ph->length),&mBlk))
            {
                pdk_vidd_receive_pkt(mBlk,ntohl(ph->portid),pf->feid);
            }

            PIL_LeaveCriticalSection(&g_ph_lock);
        }
        if (FD_ISSET(g_shutdownPipe[0],&rfds))
        {
            npf_logger_write(LOG_PH, LOG_TRACE, "Got a shutdown event...exiting\n");
            return 0;
        }

    }
}

NPF_RET ph_init()
{
    NPF_RET ret;

    PIL_InitializeCriticalSection(&g_ph_lock);

    g_ph_socketd = socket(AF_INET,SOCK_RAW,PH_IP_PROTOCOL);
    if (g_ph_socketd < 0)
    {
        perror("Error creating socket\n");
        return ret;
    }

    pipeDrv();
    pipeDevCreate(SHUTDOWN_PIPE_NAME, 10, 5);

    g_shutdownPipe[0] = g_shutdownPipe[1] =
        open(SHUTDOWN_PIPE_NAME, O_RDWR, 0666);

    g_ph_RcvrThread = PIL_CreateThread(ph_receiver_thread,0,"ph_receiver_thread",0,(void*)g_ph_socketd,0);

    memset(g_ph_fe_hashtable,0,sizeof(g_ph_fe_hashtable));

    ret = FPPAPI_register_callback( ph_fe_bind,
                    &g_ph_fe_bind_handle,
                    FPPI_FEBind_Event,
                    NULL );

    ret = FPPAPI_register_callback( ph_fe_unbind,
                    &g_ph_fe_unbind_handle,
                    FPPI_FEUnBind_Event,
                    NULL );


    pdk_vidd_getNetPool(&g_pdk_vidd_netPoolID);
    return ret;
}

void shutdownReceiveThread()
{
    write( g_shutdownPipe[1], SHUTDOWN_PIPE_MSG, strlen(SHUTDOWN_PIPE_MSG)+1 );
    PIL_WaitForThread(g_ph_RcvrThread,PIL_INFINITE_TIMEOUT);
    PIL_DestroyThread (g_ph_RcvrThread, NULL);

    close(g_shutdownPipe[0]);
    close(g_shutdownPipe[1]);
    return;
}

NPF_RET ph_shutdown()
{
    NPF_RET ret;

    shutdownReceiveThread();

    ret = FPPAPI_deregister_callback (g_ph_fe_bind_handle);
    ret = FPPAPI_deregister_callback (g_ph_fe_unbind_handle);

    PIL_DestroyCriticalSection(&g_ph_lock);
    return ret;
}

void ph_fe_bind(
            FPPI_FEID feid, 
            FPPI_CORRELATOR cb_correlator,
            FPPI_CONTEXT cb_context, 
            void* response_data
            )
{
    int size;
    ph_feaddr* fe;

    PIL_EnterCriticalSection(&g_ph_lock);

    fe = HASH_LOOKUP_BY_FEID(feid);

    fe->feid = feid;

    size = sizeof(fe->addr);
    getpeername((int)response_data,(struct sockaddr*)&(fe->addr),&size);

    g_ph_fe_hashkey_addr[HASH_FUNC_ADDR(fe->addr.sin_addr.s_addr)] = HASH_FUNC_FEID(feid);

    PIL_LeaveCriticalSection(&g_ph_lock);
}

void ph_fe_unbind(
            FPPI_FEID feid, 
            FPPI_CORRELATOR cb_correlator,
            FPPI_CONTEXT cb_context, 
            void* response_data
            )
{
    ph_feaddr* fe;

    PIL_EnterCriticalSection(&g_ph_lock);

    fe = HASH_LOOKUP_BY_FEID(feid);

    memset(fe,0,sizeof(ph_feaddr));

    PIL_LeaveCriticalSection(&g_ph_lock);
}

void ph_write_packet(FPPI_FEID feid, FPPI_PortID portid, M_BLK_ID pMblk)
{
    char buf[MAX_MESSAGE_LEN + 20];
    int len;
    ph_header* ph;

    ph_feaddr* fe;

    PIL_EnterCriticalSection(&g_ph_lock);

    fe = HASH_LOOKUP_BY_FEID(feid);
    if (!fe->addr.sin_addr.s_addr)
    {
        npf_logger_write(LOG_PH, LOG_ERR, "Unknown feid %d\n",feid);
        return;
    }

    npf_logger_write(LOG_PH, LOG_TRACE, "sending to %x\n",fe->addr.sin_addr);

    len = netMblkToBufCopy(pMblk, buf+sizeof(ph_header), NULL);

    ph = (ph_header*)buf;
    ph->portid = htonl(portid);
    ph->length = htonl(len);

    sendto(g_ph_socketd,
            buf,len+sizeof(ph_header),
            0,
            (struct sockaddr*)&(fe->addr),
            sizeof(struct sockaddr_in));

    if (pMblk != NULL)
    {
       netMblkClChainFree(pMblk);
    }

    PIL_LeaveCriticalSection(&g_ph_lock);
}

