/*****************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 1998-2003 By Intel Corporation.  All rights reserved.  
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *****************************************************************************/

#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#ifdef __linux
#include <stdint.h>
#include <linux/if_tun.h>
#include <linux/if_ether.h>
#endif

#include <net/if.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#include "pdk_vic.h"

#define MAXBUFSIZE 256
struct in6_ifreq {
    struct in6_addr ifr6_addr;
    int             ifr6_prefixlen;
    int             ifr6_ifindex;
};

static int set_flag(int skfd, char *ifname, short flag) {

    struct ifreq ifr;

    /* retrieve the current flag */
    memset(&ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
    if (ioctl(skfd, SIOCGIFFLAGS, &ifr) < 0) {
        printf("%s: could not retrieve flags: %s\n", ifname, strerror(errno));
        close(skfd);
        return (-1);
    }

    /* set  the flag */
    strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
    ifr.ifr_flags |= flag;
    if (ioctl(skfd, SIOCSIFFLAGS, &ifr) < 0) {
        printf("%s: could not set flags: %s\n", ifname, strerror(errno));
        close(skfd);
        return (-1);
    }
    return 0;
}

static int clr_flag(int skfd, char *ifname, short flag) {

    struct ifreq ifr;

    /* retrieve the current flag */
    memset(&ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
    if (ioctl(skfd, SIOCGIFFLAGS, &ifr) < 0) {
        printf("%s: could not retrieve flags: %s\n", ifname, strerror(errno));
        close(skfd);
        return (-1);
    }

    /* remove the flag */
    strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
    ifr.ifr_flags &= flag;
    if (ioctl(skfd, SIOCSIFFLAGS, &ifr) < 0) {
        printf("%s: could not set flags: %s\n", ifname, strerror(errno));
        close(skfd);
        return (-1);
    }
    return 0;

}

/* Allocate Ether TAP device, returns opened fd. 
 * Stores dev name in the first arg(must be large enough). */

int tap_open(char *dev, char *hwaddr)
{
#ifdef __linux
    struct ifreq ifr;
    int fd, err;

    if ( (fd = open("/dev/net/tun", O_RDWR)) < 0 ) {
       printf("error open /dev/net/tun\n");
       return -1;
    }

    memset(&ifr, 0, sizeof(ifr));
    ifr.ifr_flags = IFF_TAP | IFF_NO_PI;
    if ( *dev )
       strncpy(ifr.ifr_name, dev, IFNAMSIZ);

    if ( (err = ioctl(fd, TUNSETIFF, (void *) &ifr)) < 0 ) {
       close(fd);
       return err;
    }

#if 0
    /* Disable Checksum, Not sure if this is necessary */
    if ( (err = ioctl(fd, TUNSETNOCSUM, 0)) < 0) {
        close(fd);
        return err;
    }
#endif

    strcpy(dev, ifr.ifr_name);

    {
        /* open socket to set NPARP and PROMISC */
        int skfd, err;
        /* set_mac_address() is not supported by tun.c right now
        struct sockaddr sa; */

        skfd = socket(AF_INET, SOCK_DGRAM, 0);
        if (skfd < 0) {
            perror("socket");
            exit(1);
        }
        if ((err = set_flag(skfd, dev, (IFF_NOARP | IFF_PROMISC))) < 0) {
            close(skfd);
            close(fd);
            return (-1);
        }

#if 0 // set_mac_address() is not supported by tun.c right now
        /* fill in the ifr_hwaddr struct netmask */
        sa.sa_family = AF_INET;
        sa.sa_data[0] = 0x01;
        sa.sa_data[1] = 0x02;
        sa.sa_data[2] = 0x03;
        sa.sa_data[3] = 0x04;
        sa.sa_data[4] = 0x05;
        sa.sa_data[5] = 0x06;
        memcpy((char *)&ifr.ifr_hwaddr, (char *)&sa, sizeof(struct sockaddr));
        /* make the ioctl now */
        if ((err = ioctl(skfd, SIOCSIFHWADDR, &ifr)) < 0) {
            printf("%s: could not set hwaddr: %s\n", dev, strerror(errno));
            close(skfd);
            close(fd);
            return (-1);
        }
#endif

        /* retrieve the hwaddress */
        memset(&ifr, 0, sizeof(ifr));
        strncpy(ifr.ifr_name, dev, IFNAMSIZ);
        if ((err = ioctl(skfd, SIOCGIFHWADDR, &ifr)) < 0) {
            printf("%s: could not retrieve hwaddr: %s\n",dev,strerror(errno));
            close(skfd);
            close(fd);
            return (-1);
        }

        memcpy(hwaddr, ifr.ifr_hwaddr.sa_data, ETH_ALEN);
        
        close(skfd);
    }
    return fd;
#else // vxworks
    return -1;
#endif

} 

int tap_close(int fd, char *dev)
{
    return close(fd);
}

/* Write frames to TAP device */
int tap_write(int fd, char *buf, int len)
{
    return write(fd, buf, len);
}

/* Read frames from TAP device */
int tap_read(int fd, char *buf, int len)
{
    return read(fd, buf, len);
}

/* ifup the interface */
int tap_ifup(int fd, char *dev)
{
    int skfd, err;
    /* open a socket */
    skfd = socket(AF_INET, SOCK_DGRAM, 0);
    if (skfd < 0) {
        perror("socket");
        exit(1);
    }
    if ((err = set_flag(skfd, dev, IFF_UP)) < 0) {
        close(skfd);
        return (-1);
    }
    close(skfd);
    return 0;

}

/* ifdown the interface */
int tap_ifdown(int fd, char *dev)
{
    int skfd, err;
    /* open a socket */
    skfd = socket(AF_INET, SOCK_DGRAM, 0);
    if (skfd < 0) {
        perror("socket");
        exit(1);
    }
    if ((err = clr_flag(skfd, dev, ~IFF_UP)) < 0) {
        close(skfd);
        return (-1);
    }
    close(skfd);
    return 0;
}

/* ifconfig the interface, ipaddr and mask is host order */
int tap_ipconfig(int fd, char *dev, uint32_t ipaddr, uint32_t mask)
{
#ifdef __linux
    struct ifreq ifr;
    struct sockaddr_in sin;
    int skfd, err;

    /* convert to network order for ioctl */
    ipaddr = htonl(ipaddr);
    mask = htonl(mask);

    memset(&ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, dev, IFNAMSIZ);

    /* open a socket */
    skfd = socket(AF_INET, SOCK_DGRAM, 0);
    if (skfd < 0) {
        perror("socket");
        exit(1);
    }

    /* fill in the sockaddr_in struct ipaddress */
    sin.sin_family = AF_INET;
    sin.sin_port = 0;
    sin.sin_addr.s_addr = ipaddr;
    memcpy((char *)&ifr.ifr_addr, (char *)&sin, sizeof(struct sockaddr));
    /* make the ioctl now */
    if ((err = ioctl(skfd, SIOCSIFADDR, &ifr)) < 0) {
        printf("%s: could not set address: %s\n", dev, strerror(errno));
        goto error;
    }

    /* fill in the sockaddr_in struct netmask */
    sin.sin_addr.s_addr = mask;
    memcpy((char *)&ifr.ifr_netmask, (char *)&sin, sizeof(struct sockaddr));
    /* make the ioctl now */
    if ((err = ioctl(skfd, SIOCSIFNETMASK, &ifr)) < 0) {
        printf("%s: could not set netmask: %s\n", dev, strerror(errno));
        goto error;
    }

    /* fill in the sockaddr_in struct broadcast */
    sin.sin_addr.s_addr = ipaddr | ~(mask);
    memcpy((char *)&ifr.ifr_broadaddr, (char *)&sin, sizeof(struct sockaddr));
    /* make the ioctl now */
    if ((err = ioctl(skfd, SIOCSIFBRDADDR, &ifr)) < 0) {
        printf("%s: could not set broadcast: %s\n", dev, strerror(errno));
        goto error;
    }

    if ((err = set_flag(skfd, dev, (IFF_UP|IFF_RUNNING)) < 0))
        goto error;

    close(skfd);
    return 0;

error:
    close(skfd);
    return -1;
#else //vxworks
    return -1;
#endif
}

/* ifconfig the MTU */
int tap_mtuconfig(int fd, char *dev, uint32_t mtu)
{
    struct ifreq ifr;
    int skfd, err;

    memset(&ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, dev, IFNAMSIZ);

    /* open a socket */
    skfd = socket(AF_INET, SOCK_DGRAM, 0);
    if (skfd < 0) {
        perror("socket");
        exit(1);
    }

    /* fill in the mtu */
    ifr.ifr_mtu = mtu;

    /* make the ioctl now */
    if ((err = ioctl(skfd, SIOCSIFMTU, &ifr)) < 0) {
        printf("%s: could not set mtu: %s\n", dev, strerror(errno));
        goto error;
    }

    close(skfd);
    return 0;

error:
    close(skfd);
    return -1;
}

int tap_ipconfig6(int fd, char *dev, IPv6A *ipv6addr, int prefix_len)
{
#ifdef __linux
    struct ifreq ifr;
    struct in6_ifreq ifr6;
    int    skfd, err;
    char   buf[MAXBUFSIZE];

    /* For now, we use the 'system' call to set the IPv6 address.
     * Need to change this to ioctl (similar to how it is done
     * in tap_ipconfig above) at some future date.*/

    /* open a socket */
    skfd = socket(AF_INET6, SOCK_DGRAM, 0);
    if (skfd < 0) 
    {
        perror("socket");
        printf("error creating socket\n");
	return -1;
        /*exit(1);*/
    }

    if ((err = set_flag(skfd, dev, (IFF_UP|IFF_RUNNING)) < 0))
        goto error;

    strcpy(ifr.ifr_name, dev);
    if(ioctl(skfd, SIOCGIFINDEX, &ifr) < 0) {
        perror("Error getting IF Index\n");
        goto error;
    }
   
    inet_ntop(AF_INET6, ipv6addr, buf, MAXBUFSIZE);
    printf("Configuring Virtual interface with address: %s\n", buf);

    memcpy((char *) &ifr6.ifr6_addr, (char *)ipv6addr, sizeof(struct in6_addr));
    ifr6.ifr6_ifindex = ifr.ifr_ifindex;
    ifr6.ifr6_prefixlen = prefix_len;

    if(ioctl(skfd, SIOCSIFADDR, &ifr6) < 0) {
        perror("Error Setting IPv6 address\n");
        goto error;
    }
   
    close(skfd);
    printf("<-- leaving tap_ipconfig6\n");
    return 0;

error:
    close(skfd);
    printf("<-- leaving tap_ipconfig6\n");
    return -1;
#else //vxworks
    return -1;
#endif
}


/*************IPv6 addition**************/
/*This function deletes the randomly assigned link local address
 * and puts in a predetermined link local address based on the
 * port number*/
int setlladdr(PDK_VIC_t *vic)
{
    struct ifreq ifr;
    struct in6_ifreq ifr6;
    int skfd;
    char buf[256];

    /*open a socket*/
    skfd = socket(AF_INET6, SOCK_DGRAM, 0);
    if (skfd < 0) {
        perror("socket");
        printf("error creating socket\n");
        return -1;
    }

    /*Get the interface index*/
    strcpy(ifr.ifr_name, vic->name);
    if(ioctl(skfd, SIOCGIFINDEX, &ifr) < 0) {
        perror("Error getting if index");
        goto error;
    }	
    /*first add the new link local address*/
    sprintf(buf, "fe80::201:2ff:fe03:4%02d", vic->port);
    printf("Adding the link local address: %s\n", buf);

    if(inet_pton(AF_INET6, buf, &ifr6.ifr6_addr)<= 0) {
        perror("Error doing inet_pton");
        goto error;
    }

    ifr6.ifr6_prefixlen = 10;
    ifr6.ifr6_ifindex = ifr.ifr_ifindex;

    if(ioctl(skfd, SIOCSIFADDR, &ifr6)<0) {
        perror("Error setting IPv6 address");
        goto error;
    }

    /* Now we delete the old link local address.
     * The LL address is based on the randomly generated
     * hardware address. The following seems to always
     * be the case with the tun device driver. It may not
     * be the case for other virtual device drivers*/

    sprintf(buf, "fe80::2ff:%02xff:fe%02x:%02x%02x", (vic->hwaddr[2] & 0377),
            (vic->hwaddr[3] & 0377), (vic->hwaddr[4] & 0377), 
            (vic->hwaddr[5] & 0377));
    printf("Deleting IPv6 address: %s\n", buf);

    if(inet_pton(AF_INET6, buf, &ifr6.ifr6_addr)<= 0) {
        perror("Error doing inet_pton");
        goto error;
    }
	
    ifr6.ifr6_prefixlen = 10;

    if(ioctl(skfd, SIOCDIFADDR, &ifr6)<0) {
        perror("Error deleting IPv6 address");
        goto error;
    }

    close(skfd);
    return 0;

error:
    close(skfd);
    return -1;
}
