//////////////////////////////////////////////////////////////////////////////
//           INTEL CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or
//  nondisclosure agreement with Intel Corporation and may not be copied
//  or disclosed except in accordance with the terms of that agreement.
//      Copyright (c) 1999-2000 Intel Corporation.  All Rights Reserved.
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////

/*******************************************************************************
 * $Source: /cvs/constellations/ixa_3.0/src/cp_pdk/ControlPlane/CoreBlocks/VIDDController/src/tun_vidd.c,v $
 * $Revision: 1.6 $
 * $Date: 2003/11/17 16:57:14 $
 * 
 * tun_vidd.c
 * -----------------                                                        
 * This is the main VIDD controller. It initializes, creates, configures,   
 * maintains, and deletes Virtual interfaces. It interfaces with the other  
 * components of the NPF PDK - mainly Configuration & Mgmt module for       
 * receiving notifications of INTERFACE events(DOWN, UP, IP_ADDR_CHANGE..)  
 * 
 * Assumptions:
 *   1. It is initialized by the PDK manager
 *   2. C&M module must be initialized before this can be initialized
 *   3. The Virtual Interface Device Driver (VIDD) is already registered in the
 *      Linux kernel
 *   4. OS dependencies - this module makes file I/O calls (open, close), 
 *      standard socket calls and ioctl (I/O control) calls to interact with
 *      VIDD
 ******************************************************************************/

#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <netinet/in.h>

/* NPF PDK include files for interacting with CM, etc. */
#include "npf_ns_api.h"
#include "npf_cm_types.h"
#include "pdk_cm.h"
#include "pdk_ns_helper.h"
#include "pdk_macros.h"
#include "pdk_ph.h"
#include "log.h"

#ifdef _VXWORKS_
#include "vxworks_vidd.h"
#else
#include "tun_vidd.h"
#endif

/* GLobals variables */

/* context for INTERFACE/PORT events */
static NPF_USERCONTEXT   ifEvent_context;
static NPF_USERCONTEXT   portEvent_context;

/* Callback handle for INTERFACE/PORT events*/
static NPF_CBHANDLE      ifEvent_handle;
static NPF_CBHANDLE      portEvent_handle;

static uint32_t g_initialized = 0;

#define MAX_NUM_INTFS 20  /* per FE */
#define MAX_NUM_VIFS  100 /* per CE */

/* total number of virtual interfaces created */ 
static DList VIFList;

/* Internal error codes (from ioctls, etc.) */
extern int errno;

/* the path of the main Virtual interface character driver */
#define VID_PATH "/dev/net/tun"

#ifdef VIC_DEBUG
#define PDBG_PRINT(fmt, args...) npf_logger_write(LOG_VIC, LOG_ERR, fmt,## args)
#define PDBG_TRACE()             DBG_TRACE () /* macro */
#else
#define PDBG_PRINT(fmt, args...)  /* nothing */
#define PDBG_TRACE()              /* nothing */
#endif

extern int vidd_close(PDK_VIC_t *vif);
extern int vidd_open(PDK_VIC_t *vif);
extern int vidd_ifconfig(PDK_VIC_t *vif, uint32_t ipaddr, uint32_t mask);
extern int vidd_mtuconfig(PDK_VIC_t *vif, int mtu);
extern int vidd_ifup(PDK_VIC_t *vif);
extern int vidd_ifdown(PDK_VIC_t *vif);

/*****************************************************************************
* Name: CHK_INIT
*
* Description:
*     Utility macro used to check if the 
*     component is properly initialized.
*
*/
#define CHK_INIT() \
    if ( !g_initialized ) {\
        return NPF_COMPONENT_UNINITIALIZED;\
    }

/* callback functions registered with CM */
void on_IntfEvent( NPF_USERCONTEXT context,
		   NPF_HANDLE      if_npf_handle,
		   NPF_EVENT       if_event,
		   NPF_DATA        response_data);

void on_PortEvent( NPF_USERCONTEXT context,
		   NPF_HANDLE      port_npf_handle,
		   NPF_EVENT       port_event,
		   NPF_DATA        response_data);

/****************************************************************
 * Internal functions for managing virtual interfaces           *
 ****************************************************************/
#ifdef _LINUX_
static int vidd_shutdown();
#endif
static int searchVI(char *name, PDK_VIC_t **vip);
static int removeFreeVI(PDK_VIC_t *vip);


/****************************************************************************
 * Function: vic_init()                                                     *
 * --------------------                                                     *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: Invoked by PDK manager during the initialization of the     *
 *              Control plane of the PDK.                                   *
 * Assumption:  Assumes that C&M module has already been initialized - this *
 *              is required for registering different callbacks             *
 *                                                                          *
 * Return: 0 on success, error values from registration functions           *
 *                                                                          *
 * Note: this function will be extended to register the vidd module with the*
 *       kernel from here itself instead of a command line script           *
 ****************************************************************************/

NPF_RET vic_init()
{
    NPF_RET ret;

    if (g_initialized) {
        return (1);
    }

    npf_list_init(&VIFList, PIL_FreeMemory);

    ifEvent_context =  (void*) PDK_EVT_IF;
    
    /* register for INTERFACE event category */
    ret = pdk_if_event_register(ifEvent_context,
				(NPF_GENERAL_CBFUNC)&on_IntfEvent,
				&ifEvent_handle);

    portEvent_context =  (void*) PDK_EVT_PORT;

    /* register for INTERFACE event category */
    ret = pdk_port_event_register(portEvent_context,
                                  (NPF_GENERAL_CBFUNC)&on_PortEvent,
                                  &portEvent_handle);

    g_initialized = 1;

    /* need init after g_init */
#ifdef _VXWORKS_
    ph_init ();
#endif

    return (ret);
}

/****************************************************************************
 * Function: vic_shutdown()                                                 *
 * ------------------------                                                 *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: Invoked by PDK manager during the shutdown of the           *
 *              Control plane of the PDK.                                   *
 * Assumption:  Assumes that C&M module is not yet shutdown - this          *
 *              is required for deregistering different callbacks           *
 *                                                                          *
 * Return: 0 on success, error values from deregistration functions         *
 *                                                                          *
 * Note: this function will be extended to remove the vidd module from the  *
 *       kernel from here itself instead of a command line script           *
 ****************************************************************************/

NPF_RET vic_shutdown()
{
    NPF_RET ret;

#ifdef _VXWORKS_
    ph_shutdown ();
#endif
    /* deregister for INTERFACE/PORT event category */
    ret = pdk_if_event_deregister(ifEvent_handle);
    ret = pdk_port_event_deregister(portEvent_handle);

#ifdef _LINUX_
    vidd_shutdown();
#endif
    /* something different for VxWorks */

    npf_list_destroy(&VIFList);
    
    return (ret);
}

#ifdef _LINUX_
/****************************************************************************
 * Function: vidd_shutdown()                                                *
 * -------------------------                                                *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: Invoked by VIDD controller for removing the VIDD module     *
 *              (character driver).                                         *
 *                                                                          *
 * Return: 0 on success, error values from deregistration functions         *
 *                                                                          *
 * Note: this function will be extended to rmmod the vidd module from the   *
 *       kernel from here itself instead of a command line script           *
 ****************************************************************************/

static int vidd_shutdown()
{
    NPF_DATA  data;
    PDK_VIC_t *vif;
    
    CHK_INIT();
    g_initialized = 0;

    while (npf_list_popFront(&VIFList, (void**)&data) != -1)
    {
        vif = (PDK_VIC_t *)data;
#ifdef _VXWORKS_
        pdk_vif_delete(vif);
#else
        vidd_close(vif);
#endif
        PIL_FreeMemory(vif);
    }

    return (0);
}
#endif

/****************************************************************************
 * Function: vic_VI_Event_register_callback()                               *
 * ---------------------------------------                                  *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: This function MUST be invoked ONLY by internal modules of   *
 *              the PDK to register for VI Create/Delete events.            *
 *                                                                          *
 * Return: 0 on success, error values from internal registration fns        *
 *                                                                          *
 ****************************************************************************/

NPF_RET vic_vi_event_register_callback( 
				       /* IN: */ NPF_USERCONTEXT context, 
				       /* IN: */ VIEVENT_CALLBACK callback,
				       /* OUT:*/ NPF_CBHANDLE* cb_handle )
{
    /* 
       invoke the PDK callback library routine to store the callback
       this set of routines maintains any kind of callback function
       it requires a context - we will use the event type as the
       context
    */
    CHK_INIT ();
    return (pdk_callback_register( PDK_VIDD_EVT_VI_EVENT, 
				   (NPF_USERCONTEXT) context,
				   (NPF_GENERAL_CBFUNC) callback,
				   cb_handle));
}

/****************************************************************************
 * Function: vic_VIEvent_deregister_callback()                              *
 * ---------------------------------------                                  *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: This function MUST be invoked ONLY by internal modules of   *
 *              the PDK to deregister for VI Create/Delete events.          *
 *                                                                          *
 * Return: 0 on success, error values from internal deregistration fns      *
 *                                                                          *
 ****************************************************************************/

NPF_RET vic_vi_event_deregister_callback( NPF_CBHANDLE cb_handle )
{
    /* 
       invoke the PDK callback library routine to deregister the
       callback function
    */
    CHK_INIT ();
    return (pdk_callback_deregister( cb_handle ));
}

/****************************************************************************
 * Function: on_IntfEvent()                                                 *
 * ------------------------                                                 *
 * Parameters: NPF_USERCONTEXT - passed in during registration              *
 *             NPF_HANDLE - handle to interface that generated the event    *
 *             NPF_EVENT - actual event that occured                        *
 *             NPF_DATA - event-specific response data                      *
 *                                                                          *
 * Description: Invoked by C&M when it generates an INTERFACE event         *
 *              On a NPF_IF_IPADDR_CHANGE event, this function checks to    *
 *              see if a corresponding VI has been created, else it         *
 *              registers a virtual interface corresponding to the actual   *
 *              interface with the kernel (via a IOCTL call to VIDD)        *
 *                                                                          *
 * Return: none                                                             *
 *                                                                          *
 ****************************************************************************/
void on_IntfEvent( NPF_USERCONTEXT context,
		   NPF_HANDLE      if_npf_handle,
		   NPF_EVENT       if_event,
		   NPF_DATA        response_data)
{
    NPF_INTERFACE_t   *intf;
    NPF_FE_ATTRIBS    fe_attribs;
    NPF_HANDLE        fe_npf_hdl;
    PDK_VIC_t         *vip, tmp_vip;
    uint32_t          fe_id;
    char              vif_name[IFNAME_LEN];
    char              vif_name6[IFNAME_LEN];

    #ifdef _VXWORKS_
    /* for invoking registered callbacks */
    NPF_CBHANDLE       cb_hdl, cb_hdl_next;
    NPF_USERCONTEXT    cb_context;
    VIEVENT_CALLBACK   client_cb_func;
    #endif
    int                err, num_ports = 0;

    NPF_RET            npf_ret;
    NPF_HANDLE         port_npf_hdl;
    NPF_NSHDL          if_ns_hdl, port_ns_hdl;
    NODE_ITERATOR      port_itr;
    uint32_t           port_id;
    NPF_PORT_ATTRIBS   port_attrs;

    if ( !g_initialized )
	    return;

    /* check the context first */
    if (context != (void*) PDK_EVT_IF)
	   return;

    memset(vif_name, 0, IFNAME_LEN);
    memset(vif_name6, 0, IFNAME_LEN);

    intf = (NPF_INTERFACE_t*) response_data;

    /*
      - get the interface properties from CM
      - check if the virtual interface corr. to this
        has been created. if not, create it.
      - if already created, just do an IOCTL to configure it
      - need to get the FE ID of the FE this interface
        belongs to
    */
    PDBG_PRINT("**************************************\n");
    if (if_event == NPF_EVENT_IF_UP)
        PDBG_PRINT("Got %s in vidd_controller\n", "IF_UP");
    else if (if_event == NPF_EVENT_IF_DOWN)
        PDBG_PRINT("Got %s in vidd_controller\n", "IF_DOWN");
    else if (if_event == NPF_EVENT_IF_DELETE)
        PDBG_PRINT("Got %s in vidd_controller\n", "IF_DELETE");
    else if (if_event == NPF_EVENT_IF_IPADDR_CHANGE)
        PDBG_PRINT("Got IF_ADDR_CHANGE in vidd_controller\n");
    else if (if_event == NPF_EVENT_IF_IPFORWARDING_CHANGE) {
        return ;
    } else
        PDBG_PRINT("WRONG IF TYPE\n");
    PDBG_PRINT("name %s, desc %s, status %d, index %d\n",
              intf->name, intf->desc, intf->status, intf->index);
    /****************IPv6 addition ****************/
    if(intf->vifname)
        PDBG_PRINT("vifname: %s, ", intf->vifname);
    if(intf->vifname6)
        PDBG_PRINT("vifname6: %s, ", intf->vifname6);

    PDBG_PRINT("ipaddr %.2X, submask %.2X, ipfwd %d\n",
    intf->ipattribs.addr_u.ipaddr, intf->ipattribs.mask_u.submask, intf->ipattribs.ipfwding);
    /****************IPv6 addition ****************/
    PDBG_PRINT("**************************************\n");

    /* can't validate ipaddr here cuz when IFDOWN time the
     * ipaddr is set to be 0 */

    switch (if_event) {

    case NPF_EVENT_IF_UP:
    {
        PDBG_PRINT ("INTERFACE UP EVENT recvd\n");
        if((strcmp(intf->vifname,"") == 0) && (strcmp(intf->vifname6,"") == 0)) {

            /* The purpose of getting IF-to-Port association is to get
             * the FE id and port id to assign on the VIx, and
             * ONLY EVENT_IF_UP has the chance that intf->vifname is
             * not defined, IP_CHANGE, DOWN, DELETE all has vif created
             * Can get vi? directly from intf->vifname */

            /* if vifname not exist, need check association to get vi? */

            if (pdk_ns_getFE_from_If(if_npf_handle, &fe_npf_hdl)!=NPF_SUCCESS)
                return ;
            if (cm_fe_get_attribs(fe_npf_hdl, &fe_attribs) != NPF_SUCCESS) {
                cm_handle_close(fe_npf_hdl);
                return ;
            }
            if (cm_fe_get_feid(fe_npf_hdl, &fe_id) != NPF_SUCCESS) {
                cm_handle_close(fe_npf_hdl);
                return ;
            }
            cm_handle_close(fe_npf_hdl);

            /* get the first associate of the interface. this has to 
             * be a port. */
    
            /* get NS_HDL of INTERFACE using the NPF_HANDLE */
            npf_ret = cm_getNSHDLByHandle(if_npf_handle, &if_ns_hdl);
            if(npf_ret != NPF_SUCCESS) {
                PDBG_PRINT ("Could not get NS hdl from C&M \n");
                return ;
            }

            /* read the associates - get an iterator */
            npf_ret = npf_ns_enumAssociate(if_ns_hdl, &port_itr);
            if(npf_ret != NPF_SUCCESS) {
                PDBG_PRINT ("Could not enumerate Intf associates \n");
                return ;
            }
            num_ports = npf_ns_count (port_itr);

            /* Q: if npf_ns_enumAssociate return NPF_SUCCESS when no asso.? */
            if (num_ports==0)
                return ;

            if(num_ports > 0) {
                port_ns_hdl = npf_ns_first (port_itr);
                npf_ret = npf_ns_getDataHandle (port_ns_hdl, &port_npf_hdl);
                if(npf_ret != NPF_SUCCESS) {
                    PDBG_PRINT ("Could not get port NPF handle \n");
                    return ;
                }
                cm_port_get_portid(port_npf_hdl, &port_id);
                PDBG_PRINT ("Port id  = %d\n", port_id);
                cm_port_get_attribs(port_npf_hdl, &port_attrs);
            }
            npf_ns_done (port_itr);
            npf_ns_close (port_ns_hdl);
            /* Don't need to close it, it is closed in the cmcallback
             *     npf_ns_close (if_ns_hdl); */

            /* create virtual interface object */
            vip = (PDK_VIC_t *)PIL_AllocateMemory(
            sizeof(PDK_VIC_t), PIL_MEM_ZERO_INIT);
            vip->fe_id = fe_id;
            vip->fe_hdl = fe_npf_hdl;
            vip->fe_ipaddr = fe_attribs.ipaddr.s_addr;
            vip->intf_hdl = if_npf_handle;
            vip->port = port_id;
            vip->mtu = port_attrs.MTU;
            vip->linkspeed = port_attrs.curTxRate;
            vip->type = port_attrs.type;

#ifdef IPv4

#ifdef _VXWORKS_
            vip->ipaddr = htonl(intf->ipattribs.addr_u.ipaddr);
#else
            vip->ipaddr = htonl(intf->ipattribs.addr_u.ipaddr);
#endif
            vip->mask = htonl(intf->ipattribs.mask_u.submask);
	    /* Don't create VIF if ipv4 address is 0 */
	    if (vip->ipaddr == 0) {
                PIL_FreeMemory(vip);
                return ;
            }
            /* create new VIF name */
            sprintf(vif_name, "vi%s-%d", fe_attribs.name, (int)port_id);
            strcpy(vip->name, vif_name);

#ifdef _VXWORKS_
            vip->fd = pdk_vif_create(vip, vip->hwaddr);
#else
            /* open the tap and assign ipaddress */ 
            vidd_open(vip);
#endif
#else
#ifdef _LINUX_
            /***********IPv6 addition*****************/
            memcpy(&vip->ipv6addr, &(intf->ipattribs.addr_u.ipv6addr), sizeof(IPv6A));
	    /* Don't create VIF if ipv6 address is 0 */
	    if (vip->ipv6addr == 0) {
                PIL_FreeMemory(vip);
                return ;
            }
            vip->prefix_len = PrefixLen(intf->ipattribs.mask_u.ipv6submask);

            inet_ntop(AF_INET6,  &(intf->ipattribs.addr_u.ipv6addr), buf, 256);
            PDBG_PRINT("The IPv6 address received: %s\n", buf);

            inet_ntop(AF_INET6,  &(vip->ipv6addr), buf, 256);
            PDBG_PRINT("The IPv6 address stored: %s\n", buf);
            /***********IPv6 addition*****************/

            /* create new VIF name */
            sprintf(vif_name6, "vi%d", port_id);
            strcpy(vip->name6, vif_name6);
            /* open the tap and assign ipaddress */ 
            vidd_open(vip->hwaddr);
#endif
#endif

#ifdef _VXWORKS_
            /* if the tap_open fails, means it might already have the 
             * interface device there */
            if (vip->fd < 0) {
                PIL_FreeMemory(vip);
                PDBG_PRINT("Can't open vi %s, already exist?\n", vip->name);
                return;
            }
#endif

            npf_list_pushBack(&VIFList, vip);

            PDBG_PRINT("MAC Address %02X:%02X:%02X:%02X:%02X:%02X\n",
                (vip->hwaddr[0] & 0377), (vip->hwaddr[1] & 0377),
                (vip->hwaddr[2] & 0377), (vip->hwaddr[3] & 0377),
                (vip->hwaddr[4] & 0377), (vip->hwaddr[5] & 0377));

#ifdef IPv4
            /* store the new VI name for the interface */
            cm_if_set_vifname(if_npf_handle, vif_name);
#else
#ifdef _LINUX_
            /***********IPv6 addition*****************/
            vidd_ifconfig6(vip, &(vip->ipv6addr), 
                            vip->prefix_len);

            /* Now delete the randomly assigned link local address
            * and assign a predetermined address*/
            vidd_setlladdr(vip);

            /* store the new VI name for the interface */
            cm_if_set_vifname(if_npf_handle, vif_name6);
            /***********IPv6 addition*****************/
#endif
#endif

#ifdef _LINUX_

            err = vidd_ifconfig(vip, vip->ipaddr, vip->mask);
            if(err < 0) {
                PDBG_PRINT ("error up virtual interface %s: %s \n", 
                            vip->name, strerror(errno));
                return ;
            }
            err = vidd_ifup(vip);
            if(err < 0) {
                PDBG_PRINT ("error up virtual interface %s: %s \n", 
                            vip->name, strerror(errno));
                return ;
            }

#else // _VXWORKS_
            /* linux doesn't care about this since vidd and ph are
             * one and the same.
             * does vxworks still care?  ph doesn't, but does vidd?
             */
            /* notify Packet Handler through callback */
            memcpy(&tmp_vip, vip, sizeof(PDK_VIC_t));

            /* now call back to packet handler since
               it can open this device to listen for packets */
            client_cb_func = (VIEVENT_CALLBACK) 
            pdk_callback_firstFunc(PDK_VIDD_EVT_VI_EVENT,
                                   &cb_context,
                                   &cb_hdl_next);
            if (client_cb_func != NULL) {
                do {
                    cb_hdl = cb_hdl_next;
                    client_cb_func (cb_context, 
                                    PDK_VIDD_EVT_VI_CREATE, 
                                    (NPF_DATA*) &tmp_vip);
                } while ((client_cb_func = (VIEVENT_CALLBACK) 
                    pdk_callback_nextFunc(cb_hdl, &cb_context,
                    &cb_hdl_next)) != NULL);
            } /* end of callbacks */
#endif

        } else {
            /* we have to set the VI to be UP */
            PDBG_PRINT ("INTERFACE UP EVENT recvd for %s\n", intf->vifname);
            searchVI(intf->vifname, &vip);
#ifdef IPv4
#ifdef _VXWORKS_
	    err = pdk_vif_setflags(vip, IF_STATUS_UP);
#else
            err = vidd_ifup(vip);
#endif
            if(err < 0) {
                PDBG_PRINT ("error up virtual interface %s: %s \n", 
                            vip->name, strerror(errno));
                return ;
            }
#else

            /***********IPv6 addition*****************/
#ifdef _LINUX_
            vidd_ifconfig6(vip, &(vip->ipv6addr), 
					      vip->prefix_len);

            /* Now delete the randomly assigned link local address * and assign a predetermined address*/
            vidd_setlladdr(vip);
#endif
            /***********IPv6 addition*****************/
#endif

            PDBG_PRINT ("Setting virtual interface %s UP \n", vip->name);
        }
        break;
    }
    case NPF_EVENT_IF_IPADDR_CHANGE:
    {
#ifdef IPv4
        if (intf->vifname) {
            searchVI(intf->vifname, &vip);
            if (strlen (vip->name) > 0) {
                PDBG_PRINT ("Recvd NPF_IF_IPADDR_CHANGE EVENT for %s\n", 
                        vip->name);
#ifdef _VXWORKS_
		pdk_vif_ipaddr_update(vip, intf->ipattribs.addr_u.ipaddr, 
				      intf->ipattribs.mask_u.submask);
#else
                vidd_ifconfig(vip,
                         intf->ipattribs.addr_u.ipaddr, htonl(intf->ipattribs.mask_u.submask));
#endif
            }
        }

#else

#ifdef _LINUX_
        if(intf->vifname6)
        {
            searchVI(intf->vifname6, &vip);
            if (strlen (vip->name) > 0)
            {
                PDBG_PRINT ("Recvd NPF_IF_IPADDR_CHANGE EVENT for %s\n",
                            vip->name6);
                vidd_ifconfig6(vip,
                         &(intf->ipattribs.addr_u.ipv6addr), 
                PrefixLen(intf->ipattribs.mask_u.ipv6submask));
            }
        }
#endif
#endif
        break;
    }
    case NPF_EVENT_IF_DOWN:
    {
        if((strcmp(intf->vifname,"") == 0) && (strcmp(intf->vifname6,"") == 0)) 
           break;

#ifdef IPv4
        if (intf->vifname) {
            if (searchVI(intf->vifname, &vip) == -1)
                break;
            if (strlen (vip->name) > 0) {
                /* we have to set the VI to be DOWN */
                PDBG_PRINT ("Recvd NPF_IF_DOWN EVENT for %s\n", vip->name);
#ifdef _VXWORKS_
		pdk_vif_setflags(vip, IF_STATUS_DOWN);
#else
                vidd_ifdown(vip);
#endif
            }
        }

#else

        PDBG_PRINT("v4 If down processed\n");

#ifdef _LINUX_
        if(intf->vifname6)
        {
            searchVI(intf->vifname6, &vip);
            if ((vip) && (strlen (vip->name6) > 0) )
            {
                /* we have to set the VI to be DOWN */
                PDBG_PRINT ("Recvd NPF_IF_DOWN EVENT for %s\n", vip->name6);
                vidd_ifdown(vip);
            }
        }
#endif

#endif
        break;
    }
    case NPF_EVENT_IF_DELETE:
    {
#ifdef IPv4
        if (strlen (intf->vifname) > 0) {
            searchVI(intf->vifname, &vip);
            PDBG_PRINT ("Recvd NPF_IF_DELETE EVENT for %s\n", vip->name);
            memcpy(&tmp_vip, vip, sizeof(PDK_VIC_t));
		    
#ifdef _VXWORKS_
            /* first call back to packet handler since
               it will have all these devices open */
            client_cb_func = (VIEVENT_CALLBACK) 
            pdk_callback_firstFunc(PDK_VIDD_EVT_VI_EVENT,
                                   &cb_context,
                                   &cb_hdl_next);
            if (client_cb_func != NULL) {
                do {
                    cb_hdl = cb_hdl_next;
                    client_cb_func (context, 
                                    PDK_VIDD_EVT_VI_DELETE, 
                                    (NPF_DATA*) &tmp_vip);
                } while ((client_cb_func = (VIEVENT_CALLBACK) 
                pdk_callback_nextFunc(cb_hdl, &cb_context,
                                      &cb_hdl_next)) != NULL);
            } /* end of callbacks */
#endif
		    
            /* now we can delete the virtual interface */
#ifdef _VXWORKS_
	    pdk_vif_delete(vip);
#else
            vidd_close(vip);
#endif
            /* remove the VI name for the interface since it is down */
            strcpy(vif_name, "");
            cm_if_set_vifname(if_npf_handle, vif_name);

            /* remove from the VIList */
            removeFreeVI(vip);
        }
#else
#endif
        break;
    }
    default:
        break;
    }
    return ;
}

/****************************************************************************
 * Function: on_PortEvent()                                                 *
 * ------------------------                                                 *
 * Parameters: NPF_USERCONTEXT - passed in during registration              *
 *             NPF_HANDLE - handle to port that generated the event         *
 *             NPF_EVENT - actual event that occured                        *
 *             NPF_DATA - event-specific response data                      *
 *                                                                          *
 * Description: Invoked by C&M when it generates an PORT event like         *
 *              PORT UP/DOWN/MTU/LINKSPEED but ignore UP/DOWN because it's  *
 *              taking care by the IF UP/DOWN. Only see MTU/LINKSPEED to    *
 *              change the corresponding port setting by calling tun_dev    *
 *                                                                          *
 * Return: none                                                             *
 *                                                                          *
 ****************************************************************************/
void on_PortEvent( NPF_USERCONTEXT context,
		   NPF_HANDLE      port_npf_handle,
		   NPF_EVENT       port_event,
		   NPF_DATA        response_data)
{
    int             numif;
    NPF_HANDLE      iflist[MAXIFOWNPORT];
    NPF_INTERFACE_t intf;
    NPF_PORT_t      *port;
    PDK_VIC_t       *vip;

    if ( !g_initialized )
        return ;

    /* check the context first */
    if (context != (void*) PDK_EVT_PORT)
        return ;

    port = (NPF_PORT_t*) response_data;

    PDBG_PRINT("**************************************\n");

    switch (port_event) {
    case NPF_EVENT_PORT_MTU_CHANGE:
        PDBG_PRINT("Got %s in tun_vidd\n", "MTU_CHANGE");
        break;
    case NPF_EVENT_PORT_LINKSPEED_CHANGE:
        PDBG_PRINT("Got %s in tun_vidd\n", "LINKSPEED_CHANGE");
        break;
    case NPF_EVENT_PORT_LINKUP:
        PDBG_PRINT("Got PORT_LINKUP in tun_vidd, IGNORE\n");
        return ;
        break;
    case NPF_EVENT_PORT_LINKDOWN:
        PDBG_PRINT("Got PORT_LINKDOWN in tun_vidd, IGNORE\n");
        return ;
        break;
    default:
        PDBG_PRINT("WRONG IF TYPE\n");
        return ;
    }

    PDBG_PRINT("**************************************\n");

    numif = 0;
    if (cm_getIFByPort(port_npf_handle, &numif, iflist) != NPF_SUCCESS)
        return ;

    /* Note: Don't support one to many mapping */
    if (numif != 1) {
        PDBG_PRINT("Does not support ont to many mapping");
        return ;
    }

    if (cm_if_get_attribs(iflist[0], &intf) != NPF_SUCCESS)
        return ;

    if (strlen(intf.vifname) == 0)
        return ;
    
    if (searchVI(intf.vifname, &vip) == -1)
        return ;

    switch(port_event) {
    case NPF_EVENT_PORT_MTU_CHANGE:
        PDBG_PRINT ("Recvd PORT MTU EVENT for %s\n", vip->name);
#ifdef _VXWORKS_
	pdk_vif_mtu_config(vip, port->attribs.MTU);
#else
        vidd_mtuconfig(vip, port->attribs.MTU);
#endif
        break;
    case NPF_EVENT_PORT_LINKSPEED_CHANGE:
        PDBG_PRINT ("Recvd PORT LINKSPEED EVENT for %s, DO NOTHING\n", vip->name);
        break;
    default:
    }

    return;
}

static int searchVI(char *name, PDK_VIC_t **vip)
{

    int      found = 0;
    DListItr *itr;

    if (name == NULL) {
        *vip = NULL;
        return -1;
    }
    
    itr = npf_list_itrCreate(&VIFList);
    if (npf_list_itrFirst(itr) != -1) {
        do {
            *vip = (PDK_VIC_t *) npf_list_itrGetData(itr);
            if ((*vip)->name && strcmp((*vip)->name, name) == 0) {
                found = 1;
                break;
            }
            else if ((*vip)->name6 && strcmp((*vip)->name6, name) == 0) {
                found = 1;
                break;
            }
        } while (npf_list_itrNext(itr) != -1);
    }
    npf_list_itrDelete(itr);

    if (!found) {
        *vip = NULL;
        return -1;
    }

    return 0;

}

static int removeFreeVI(PDK_VIC_t *vip)
{

    int       found = 0;
    PDK_VIC_t *tmp_vip;
    DListItr  *itr;

    itr = npf_list_itrCreate(&VIFList);
    if (npf_list_itrFirst(itr) != -1) {
        do {
            tmp_vip = (PDK_VIC_t *) npf_list_itrGetData(itr);
            if (tmp_vip == vip) {
                npf_list_itrRemoveFree(itr);
                found = 1;
                break;
            }
        } while (npf_list_itrNext(itr) != -1);
    }
    npf_list_itrDelete(itr);

    if (!found)
        return -1;
    else
        return 0;

}
