//////////////////////////////////////////////////////////////////////////////
//           INTEL CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or
//  nondisclosure agreement with Intel Corporation and may not be copied
//  or disclosed except in accordance with the terms of that agreement.
//      Copyright (c) 1999-2000 Intel Corporation.  All Rights Reserved.
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////

/********************************************************************************
 * $Source: CODEBASE/ControlPlane/CoreBlocks/VIDDController/src/vidd_controller.c
 * $Revision: 1.6 $:                                                        
 * $Date: 2003/11/17 16:57:14 $:                                            
 * 
 * vidd_controller.c                                                        
 * -----------------                                                        
 * This is the main VIDD controller. It initializes, creates, configures,   
 * maintains, and deletes Virtual interfaces. It interfaces with the other  
 * components of the NPF PDK - mainly Configuration & Mgmt module for       
 * receiving notifications of INTERFACE events(DOWN, UP, IP_ADDR_CHANGE..)  
 * 
 * Assumptions:
 *   1. It is initialized by the PDK manager
 *   2. C&M module must be initialized before this can be initialized
 *   3. The Virtual Interface Device Driver (VIDD) is already registered in the
 *      Linux kernel
 *   4. OS dependencies - this module makes file I/O calls (open, close), standard
 *      socket calls and ioctl (I/O control) calls to interact with VIDD
 *********************************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h> /* ioctl() */
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/socket.h> /* socket(), inet_aton(), etc. */
#include <netinet/in.h> /* inet_aton(), etc. */
#include <netinet/ip.h>
#include <arpa/inet.h> /* inet_aton(), etc. */
#include <net/if.h> /* struct ifreq */
#include <net/if_arp.h> /* struct ifreq */
#include <linux/types.h>
#include <net/route.h>

#include "npf_ns_api.h"
#include "npf_cm_types.h"
#include "pdk_cm.h"
#include "pdk_ns_helper.h"
#include "pdk_macros.h"
#include "pdk_vic.h"
#include "vip_tunnel.h"
#include "log.h"

#ifdef VIC_DEBUG
#include "log.h"
#define PDBG_PRINT(fmt, args...) npf_logger_write(LOG_VIC, LOG_ERR, fmt,## args)
#else
#define PDBG_PRINT(fmt, args...)  /* nothing */
#endif


#define GET_SOCKET(skfd) \
    skfd = socket(AF_INET,SOCK_DGRAM,0); \
    if (skfd < 0) \
    { \
	    PDBG_PRINT ("cannot open socket:%s", strerror(errno)); \
	    return (-1); \
    }

#define GET_SOCKET6(skfd) \
    skfd = socket(AF_INET6,SOCK_DGRAM,0); \
    if (skfd < 0) \
    { \
	    PDBG_PRINT ("cannot open socket:%s", strerror(errno)); \
	    return (-1); \
    }

#define CLOSE_SOCKET(skfd) \
    close(skfd)

#define SOCK_IOCTL(skfd, cmd, ifr) \
    if (ioctl(skfd,cmd,ifr) < 0) \
    { \
        PDBG_PRINT ("%s error %s", #cmd, strerror(errno)); \
        close(skfd); \
        return (-1); \
    }


/* Set a certain interface flag. */
static int set_flag(int skfd, char *ifname, short flag)
{
    struct ifreq ifr;
    
    strcpy(ifr.ifr_name, ifname);
    SOCK_IOCTL(skfd, SIOCGIFFLAGS, &ifr);

    strcpy(ifr.ifr_name, ifname);
    ifr.ifr_flags |= flag;
    SOCK_IOCTL(skfd, SIOCSIFFLAGS, &ifr);

    return (0);
}

/* Set a certain interface flag. */
static int reset_flag(int skfd, char *ifname, short flag)
{
    struct ifreq ifr;
    
    strcpy(ifr.ifr_name, ifname);
    SOCK_IOCTL(skfd, SIOCGIFFLAGS, &ifr);

    strcpy(ifr.ifr_name, ifname);
    ifr.ifr_flags &= ~flag;
    SOCK_IOCTL(skfd, SIOCSIFFLAGS, &ifr);

    return (0);
}

int vidd_open(PDK_VIC_t *vif)
{
    int skfd;
    struct ifreq ifr;

    struct vip_tunnel_parm vtp;

    memset(&ifr,0,sizeof(ifr));
    memset(&vtp,0,sizeof(vtp));

    vtp.iph.version = 4;
    vtp.iph.ihl = 5;
#ifndef IP_DF
#define IP_DF       0x4000      /* Flag: "Don't Fragment"   */
#endif
    vtp.iph.frag_off = htons(IP_DF);
    vtp.iph.protocol = IPPROTO_VIP;
    vtp.iph.saddr = INADDR_ANY;//vif->ipaddr;
    vtp.iph.daddr = vif->fe_ipaddr;
    vtp.portid = vif->port;
    strncpy(vtp.name,vif->name,IFNAMSIZ-1);

    strcpy(ifr.ifr_name,"vi0");
    ifr.ifr_ifru.ifru_data = (void*)&vtp;

    GET_SOCKET(skfd);

    SOCK_IOCTL(skfd,SIOCADDTUNNEL,&ifr);

    CLOSE_SOCKET(skfd);

    return 0;
}

int vidd_close(PDK_VIC_t *vif)
{
    int skfd;
    struct ifreq ifr;
    struct vip_tunnel_parm vtp;

    memset(&ifr,0,sizeof(ifr));
    memset(&vtp,0,sizeof(vtp));

    strncpy(vtp.name,vif->name,IFNAMSIZ-1);

    strcpy(ifr.ifr_name,"vi0");
    ifr.ifr_ifru.ifru_data = (void*)&vtp;

    GET_SOCKET(skfd);

    SOCK_IOCTL(skfd,SIOCDELTUNNEL,&ifr);

    CLOSE_SOCKET(skfd);

    return 0;
}

int vidd_ifup(PDK_VIC_t *vif)
{
    int skfd;
    int ret;
    GET_SOCKET(skfd);
    ret = set_flag(skfd,vif->name,(IFF_UP|IFF_RUNNING|IFF_MULTICAST));
    CLOSE_SOCKET(skfd);
    return ret;
}

int vidd_ifdown(PDK_VIC_t *vif)
{
    int skfd;
    int ret;
    GET_SOCKET(skfd);
    ret = reset_flag(skfd,vif->name,(IFF_UP|IFF_RUNNING|IFF_MULTICAST));
    CLOSE_SOCKET(skfd);
    return ret;
}

int vidd_ifconfig(PDK_VIC_t *vif, uint32_t ipaddr, uint32_t mask)
{
    int skfd;
    struct ifreq ifr;
    struct sockaddr_in sin;

    memset(&ifr,0,sizeof(ifr));
    memset(&sin,0,sizeof(sin));

    strncpy(ifr.ifr_name,vif->name,IFNAMSIZ-1);
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = ipaddr;
    memcpy(&(ifr.ifr_addr),&(sin),sizeof(struct sockaddr));

    GET_SOCKET(skfd);

    SOCK_IOCTL(skfd,SIOCSIFADDR,&ifr);

    sin.sin_addr.s_addr = mask;
    memcpy(&(ifr.ifr_netmask),&(sin),sizeof(struct sockaddr));

    SOCK_IOCTL(skfd,SIOCSIFNETMASK,&ifr);

    sin.sin_addr.s_addr = ipaddr | ~(mask);
    memcpy(&(ifr.ifr_broadaddr),&(sin),sizeof(struct sockaddr));

    SOCK_IOCTL(skfd,SIOCSIFBRDADDR,&ifr);

    CLOSE_SOCKET(skfd);

    return 0;
}

struct in6_ifreq {
    struct in6_addr ifr6_addr;
    int             ifr6_prefixlen;
    int             ifr6_ifindex;
};

int vidd_ifconfig6(PDK_VIC_t *vif, IPv6A *ipv6addr, int prefix_len)
{
    struct ifreq ifr;
    struct in6_ifreq ifr6;
    int    skfd, err;
    char   buf[256];

    /* For now, we use the 'system' call to set the IPv6 address.
     * Need to change this to ioctl (similar to how it is done
     * in tap_ipconfig above) at some future date.*/

    /* open a socket */
    GET_SOCKET6(skfd);

    if ((err = set_flag(skfd, vif->name, (IFF_UP|IFF_RUNNING)) < 0))
        goto error;

    strcpy(ifr.ifr_name, vif->name);
    if(ioctl(skfd, SIOCGIFINDEX, &ifr) < 0) {
        perror("Error getting IF Index\n");
        goto error;
    }
   
    inet_ntop(AF_INET6, ipv6addr, buf, 256);

    memcpy((char *) &ifr6.ifr6_addr, (char *)ipv6addr, sizeof(struct in6_addr));
    ifr6.ifr6_ifindex = ifr.ifr_ifindex;
    ifr6.ifr6_prefixlen = prefix_len;

    SOCK_IOCTL(skfd, SIOCSIFADDR, &ifr6);
   
error:
    CLOSE_SOCKET(skfd);

    return err;
}

int vidd_mtuconfig(PDK_VIC_t *vif, int mtu)
{
    int skfd;
    struct ifreq ifr;

    memset(&ifr,0,sizeof(ifr));

    strncpy(ifr.ifr_name,vif->name,IFNAMSIZ-1);

    ifr.ifr_mtu = mtu;

    GET_SOCKET(skfd);

    SOCK_IOCTL(skfd,SIOCSIFMTU,&ifr);

    CLOSE_SOCKET(skfd);

    return 0;
}

/*************IPv6 addition**************/
/*This function deletes the randomly assigned link local address
 * and puts in a predetermined link local address based on the
 * port number*/
int vidd_setlladdr(PDK_VIC_t *vic)
{
    struct ifreq ifr;
    struct in6_ifreq ifr6;
    int skfd;
    char buf[256];

    /*open a socket*/
    GET_SOCKET6(skfd);

    /*Get the interface index*/
    strcpy(ifr.ifr_name, vic->name);
    SOCK_IOCTL(skfd, SIOCGIFINDEX, &ifr);

    /*first add the new link local address*/
    sprintf(buf, "fe80::201:2ff:fe03:4%02d", vic->port);

    if(inet_pton(AF_INET6, buf, &ifr6.ifr6_addr)<= 0) {
        perror("Error doing inet_pton");
        goto error;
    }

    ifr6.ifr6_prefixlen = 10;
    ifr6.ifr6_ifindex = ifr.ifr_ifindex;

    SOCK_IOCTL(skfd, SIOCSIFADDR, &ifr6);

    /* Now we delete the old link local address.
     * The LL address is based on the randomly generated
     * hardware address. The following seems to always
     * be the case with the tun device driver. It may not
     * be the case for other virtual device drivers*/

    sprintf(buf, "fe80::2ff:%02xff:fe%02x:%02x%02x", (vic->hwaddr[2] & 0377),
            (vic->hwaddr[3] & 0377), (vic->hwaddr[4] & 0377), 
            (vic->hwaddr[5] & 0377));

    if(inet_pton(AF_INET6, buf, &ifr6.ifr6_addr)<= 0) {
        perror("Error doing inet_pton");
        goto error;
    }
	
    ifr6.ifr6_prefixlen = 10;

    SOCK_IOCTL(skfd, SIOCDIFADDR, &ifr6);

    CLOSE_SOCKET(skfd);
    return 0;

error:
    CLOSE_SOCKET(skfd);
    return -1;
}
