/****************************************************************************
 *             INTEL CORPORATION PROPRIETARY INFORMATION
 *  This software is supplied under the terms of a license agreement or
 *  nondisclosure agreement with Intel Corporation and may not be copied
 *  or disclosed except in accordance with the terms of that agreement.
 *      Copyright (c) 2001 Intel Corporation. All Rights Reserved.
 *
 * $Source: /cvs/constellations/ixa_3.0/src/cp_pdk/ControlPlane/DataRelayShim/src/fppcommon.c,v $
 * $Revision: 1.6 $
 * $Date: 2003/11/17 16:57:15 $
 *
 * Description:
 * Contains init and shutdown methods, and some helper methods called
 * from fppapi.cc.
 *
 ****************************************************************************/

#include "log.h"
#include "msgthreads.h"
#include "drspriv.h"
#include "pdk_callback.h"
#include "npf_ds_types.h"
#include "FPModule.h"      /* For FP start method */
#include "fpp_common.h"

/********************************* Globals *********************************/

int Fppi_Initialized     = 0;
int Fppi_Started         = 0;

extern FPPI_FEID_MAP g_feInfo;
extern PilCriticalSection g_beFlagslock;
PilCriticalSection fppi_lock;
DList methodInfoList;

static QUEUE_SERVICE_PARAMS g_cpTofpThrdParms;
QUEUE_SERVICE_PARAMS g_fpTocpThrdParms;

#define DRS_THREAD_SHUTDOWN_TIMEOUT    15000
#define DRS_THREAD_STARTUP_TIMEOUT     15000

#define DRS_CP_TO_FP_THREAD_NAME       "CPtoFPSrvcThread"
#define DRS_FP_TO_CP_THREAD_NAME       "FPtoCPSrvcThread"

/****************************************************************************
 Declare local methods
 ***************************************************************************/

static FPPI_RET fppapi_MsgToFPCore(
    FPPI_FEID             feid,
    FPPI_NHID             nhid,
    FPPI_PortID           portid,
    uint32_t  dir,
    uint32_t      tableid,
    uint32_t   instid,
    FPPI_CORRELATOR       correlator,
    FPPI_CBHANDLE         cbhandle,
    NPF_CBCAT             cbtype,
    void*                 policy_list,
    uint32_t              policy_count,
    uint32_t      nexttableid,
    uint32_t   nextinstid);

/****************************************************************************
 * Function: compareCBInfo()                                                *
 * ---------------------------                                              *
 * Parameters: data - the data element in the CBInfo list                   *
 *             key  - the key corresponding to the data element which is    *
 *                    being searched in the list                            *
 *                                                                          *
 * Description: This fn is used to compare the data mem addresses in the    *
 *              METHOD_INFO list.                                           *
 *                                                                          *
 * Return: 1 if comparison is successful , 0 if comparison fails            *
 *                                                                          *
 ****************************************************************************/

static int compareCBInfo (const void* data, const void* key)
{
    return ( data == key );
}

/****************************************************************************
 * Function: FPPAPI_register_callback()                                     *
 * ---------------------------                                              *
 * Parameters: fppapi_callback - the callback function being registered     *
 *             FPPI_CBHANDLE - unique ID generated by the Plugin, used to   *
 *                identify the callback function                            *
 *             NPF_CBCAT - the callback type for which the fn is registered *
 *             FPPI_CONTEXT - value passed by client which should be        *
 *                returned when callback is invoked                         *
 *                                                                          *
 * Description: This is the generic function used by the client             *
 *              to register a callback function for a particular event/     *
 *              call identified by the callback type.                       *
 *                                                                          *
 * Return: FPPI_SUCCESS, FPPI_FAIL                                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET FPPAPI_register_callback(
    fppapi_callback  callback,
    FPPI_CBHANDLE*   cbhandle,
    NPF_CBCAT        cbtype,
    FPPI_CONTEXT     context) 
{
    npf_logger_write(LOG_DRS, LOG_TRACE,
        "Entering method with cbhandle: %d.\n", cbhandle);

    if (Fppi_Initialized == 0)
        return FPPI_NOT_INITIALIZED;

    if (pdk_callback_register(cbtype, context,
        (NPF_GENERAL_CBFUNC)callback, cbhandle) != NPF_SUCCESS)
        return FPPI_FAIL;

    npf_logger_write(LOG_DRS, LOG_TRACE,
        "Leaving method.\n");

    return FPPI_SUCCESS;
}

/****************************************************************************
 * Function: FPPAPI_deregister_callback()                                   *
 * ---------------------------                                              *
 * Parameters: FPPI_CBHANDLE - unique ID used to identify the callback fn   *
 *                                                                          *
 * Description: This is the generic function used by the client             *
 *              to de-register a callback function.                         *
 *                                                                          *
 * Return: FPPI_SUCCESS, FPPI_FAIL                                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET FPPAPI_deregister_callback(
    FPPI_CBHANDLE    cbhandle)
{
    npf_logger_write(LOG_DRS, LOG_TRACE,
        "Entering method with cbhandle: %d.\n", cbhandle);

    if (Fppi_Initialized == 0)
        return FPPI_NOT_INITIALIZED;

    if (pdk_callback_deregister(cbhandle) != NPF_SUCCESS)
        return FPPI_FAIL;

    npf_logger_write(LOG_DRS, LOG_TRACE,
        "Leaving method.\n");

    return FPPI_SUCCESS;
}

/****************************************************************************
 * Function: FPPAPI_Initialize()                                            *
 * ---------------------------                                              *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: This function is used to initialize the Transport Plugin    *
 *              module in the control plane. It initializes the different   *
 *              lists used by the plugin module.                            *
 *                                                                          *
 * Return: FPPI_SUCCESS, FPPI_FAIL                                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET FPPAPI_Initialize (void)
{
    FPPI_RET retVal;

    npf_logger_write(LOG_DRS, LOG_TRACE,
        "Entering method.\n");

    if (Fppi_Initialized == 1)
        return FPPI_FAIL;

    PIL_InitializeCriticalSection(&g_beFlagslock);

    PIL_InitializeCriticalSection(&fppi_lock);
    PIL_EnterCriticalSection(&fppi_lock);

    npf_list_init(&methodInfoList, PIL_FreeMemory);
    methodInfoList.compare = &compareCBInfo;

    /* Create event queue package */
    if ((retVal = InitThreadObj(&g_cpTofpThrdParms)) != FPPI_SUCCESS)
    {
        npf_logger_write(LOG_DRS, LOG_ERR,
            "CP to FP Msg queue creation failed: err=%d\n", retVal);
        goto DRS_LBL_1;
    }
    else {
        npf_logger_write(LOG_DRS, LOG_INFO,
            "CP to FP Msg queues initialized.\n");
    }

    if ((retVal = InitThreadObj(&g_fpTocpThrdParms)) != FPPI_SUCCESS)
    {
        npf_logger_write(LOG_DRS, LOG_ERR,
            "FP to CP Msg queue creation failed: err=%d\n", retVal);

        /* Shutdown previous thread */
        retVal = DestroyThreadObj(&g_cpTofpThrdParms);
    }
    else {
        Fppi_Initialized = 1;
        npf_logger_write(LOG_DRS, LOG_INFO,
            "FP to CP Msg queues initialized.\n");
    }

DRS_LBL_1:

    PIL_LeaveCriticalSection(&fppi_lock);

    npf_logger_write(LOG_DRS, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/****************************************************************************
 * Function: FPPAPI_Start()                                                 *
 * ---------------------------                                              *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: This function is used to start the Data relay shim. The     *
 *              message service threads are started and then method waits   *
 *              for the threads to return a signal that the threads have    *
 *              indeed started. This is done to ensure the DRS is           *
 *              completely started and ready for action before the FP is    *
 *              started. When the FP starts it sends a cap report to the    *
 *              CP so the DRS must be already fully operational.            *
 *                                                                          *
 * Return: FPPI_SUCCESS, FPPI_FAIL                                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET FPPAPI_Start (void)
{
    FPPI_RET retVal = FPPI_SUCCESS;
    char* coloFEName = "ColoFE_1";
    char* coloHostID = "ColoCP";
    int32_t fpRet;

    npf_logger_write(LOG_DRS, LOG_TRACE,
        "Entering method.\n");

    if ((Fppi_Initialized == 0) || (Fppi_Started == 1))
    {
        retVal = FPPI_NOT_INITIALIZED;
        npf_logger_write(LOG_DRS, LOG_ERR,
            "Already init or started: err=%d\n", retVal);
        goto DRS_LBL_1;
    }

    /* Spawn the msg queue threads that wait for and act on messages
       placed in the CPtoFP queue by the CP and the FPtoCP queue by the FP. */
    if ((retVal = CreateDRSThread(
        &g_cpTofpThrdParms,
        DRS_CP_TO_FP_THREAD_NAME)) != FPPI_SUCCESS)
    {
        npf_logger_write(LOG_DRS, LOG_ERR,
            "CP to FP service thread creation failed: err=%d\n", retVal);
        goto DRS_LBL_1;
    }

    if ((retVal = CreateDRSThread(
        &g_fpTocpThrdParms,
        DRS_FP_TO_CP_THREAD_NAME)) != FPPI_SUCCESS)
    {
        npf_logger_write(LOG_DRS, LOG_ERR,
            "FP to CP Msg service thread creation failed: err=%d\n", retVal);
        goto DRS_LBL_1;
    }

    /* Stay here until we receive signals that both threads have started */
    if (PIL_AcquireSemaphore(g_cpTofpThrdParms.thrdUpNotifier,
        DRS_THREAD_STARTUP_TIMEOUT) != PIL_OK)
    {
        retVal = FPPI_FAIL;
        npf_logger_write(LOG_DRS, LOG_ERR,
            "DRSThread - Error waiting for event: err=%d\n", retVal);
        goto DRS_LBL_1;
    }

    if (PIL_AcquireSemaphore(g_fpTocpThrdParms.thrdUpNotifier,
        DRS_THREAD_STARTUP_TIMEOUT) != PIL_OK)
    {
        retVal = FPPI_FAIL;
        npf_logger_write(LOG_DRS, LOG_ERR,
            "DRSThread - Error waiting for event: err=%d\n", retVal);
        goto DRS_LBL_1;
    }

    /* OK - start successful so can set flag */
    Fppi_Started = 1;

    /* Now that the CP and DRS are up and running, must init and
        start the FP. 
        The prototype for the FP main method is:
             int fp_main (char *feid, char *hostid)
        The feid passed into fp_main is a string that the FP refers to
        as FEname. The FP passes it into BENDAPI_bindRequest (in DRS)
        where it is stored and mapped into the feid created there
        (uint32_t). The feid is passed between CP and TPI. */

    if ((fpRet = fp_main (coloFEName, coloHostID)) != 0)
    {
        retVal = FPPI_FAIL;
        npf_logger_write(LOG_DRS, LOG_ERR,
            "Startup of FP failed: err=%d\n", fpRet);
    }

DRS_LBL_1:

    if (retVal != FPPI_SUCCESS)
    {
        /* Shutdown the service threads */
        retVal = ShutdownDRSThread(&g_cpTofpThrdParms);
        retVal = ShutdownDRSThread(&g_fpTocpThrdParms);

        Fppi_Started = 0;
    }

    npf_logger_write(LOG_DRS, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/****************************************************************************
 * Function: FPPAPI_Stop()                                                  *
 * ---------------------------                                              *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: This function is used to stop the DRS in terms of ending    *
 *              the message servicing threads.                              *
 *                                                                          *
 * Return: FPPI_SUCCESS, FPPI_FAIL                                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET FPPAPI_Stop (void)
{
    FPPI_RET retVal = FPPI_SUCCESS;
    FPPI_Status status = FPPI_SUCCESSFUL;

    npf_logger_write(LOG_DRS, LOG_TRACE,
        "Entering method.\n");

    if ((Fppi_Initialized == 0) || (Fppi_Started == 0))
    {
        retVal = FPPI_NOT_INITIALIZED;
        npf_logger_write(LOG_DRS, LOG_ERR,
            "FPPAPI_Stop - DRS is not init or started: err=%d\n", retVal);
        goto DRS_LBL_1;
    }

    /* Send UnBind Event to all FEs */
    retVal = fppapi_MsgToFP(
        g_feInfo.feid, 0, NULL, 0, FPPI_CEUnBind_Event, &status, 1);

    npf_logger_write(LOG_DRS, LOG_TRACE,
        "Now shutdown the CP to FP thread.\n");

    /* Shutdown the service threads */
    retVal = ShutdownDRSThread(&g_cpTofpThrdParms);

    npf_logger_write(LOG_DRS, LOG_TRACE,
        "Now shutdown the FP to CP thread.\n");

    retVal = ShutdownDRSThread(&g_fpTocpThrdParms);

    /* DRS is stopped so reset flag */
    Fppi_Started = 0;

DRS_LBL_1:

    npf_logger_write(LOG_DRS, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/****************************************************************************
 * Function: ShutdownDRSThread()                                            *
 * ---------------------------                                              *
 * Parameters: srvcThrdParms - struct ptr containing relevant thread params *
 *                                                                          *
 * Description: Shutdown a msg thread. Send a shutdown msg then wait for    *
 *              it to expire. If it does not expire, force termination.     *
 *                                                                          *
 * Return: FPPI_SUCCESS, FPPI_FAIL                                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET ShutdownDRSThread(
    LPQUEUE_SERVICE_PARAMS srvcThrdParms)
{
    FPPI_RET retVal             = FPPI_SUCCESS;
    LPDRS_PARAM_TYPE shtdnParms = NULL;
    PilThread thrdHndl          = NULL;
    BOOL isCSLocked             = FALSE;

    npf_logger_write(LOG_DRS, LOG_TRACE,
        "Entering method.\n");

    /* Lock CS to protect param data */
    PIL_EnterCriticalSection(&(srvcThrdParms->queSyncObj));
    isCSLocked = TRUE;

    /* Don't do anything if thread is not started */
    if (srvcThrdParms->thrdHandle == NULL)
    {
        npf_logger_write(LOG_DRS, LOG_INFO,
            "Bypassing DRS service thread shutdown - not started.\n");
        goto DRS_LBL_1;
    }

    /* Create shutdown message for msg boxe */
    shtdnParms = (LPDRS_PARAM_TYPE)PIL_AllocateMemory
        (sizeof(DRS_PARAM_TYPE), PIL_MEM_ZERO_INIT);

    if (shtdnParms == NULL)
    {
        retVal = FPPI_OUT_OF_MEMORY;
        npf_logger_write(LOG_DRS, LOG_ERR, "Could not allocate memory.\n");
        goto DRS_LBL_1;
    }

    /* Set the event type */
    shtdnParms->evType = DRS_EVENT_SHUTDOWN;

    /* Insert shutdown msg into inbox msg queue */
    npf_list_pushBack(&(srvcThrdParms->msgQueue), (void *)shtdnParms);

    /* Now notify threads that a message is pending in its queue */
    if (PIL_ReleaseSemaphore(srvcThrdParms->newMsgNotifier) != PIL_OK)
    {
        npf_logger_write(LOG_DRS, LOG_ERR,
            "Error setting event object.\n");
        retVal = FPPI_FAIL;
        goto DRS_LBL_1;
    }

    /* Set thread handles in local vars while lock is on */
    thrdHndl = srvcThrdParms->thrdHandle;

    /* Unlock the msg queues so they can be serviced by the threads */
    PIL_LeaveCriticalSection(&(srvcThrdParms->queSyncObj));
    isCSLocked = FALSE;

    /* Wait for threads to join (expire). If they do not quit
       within the specified timeout, proceed with ungraceful thread
       termination. */
    if (PIL_WaitForThread(thrdHndl, DRS_THREAD_SHUTDOWN_TIMEOUT) != PIL_OK)
    {
        npf_logger_write(LOG_DRS, LOG_ERR,
            "Thread wait call failed.\n");

        /* Terminate thread ungracefully since can't do it gracefully */
        PIL_TerminateThread(thrdHndl);

        npf_logger_write(LOG_DRS, LOG_ERR,
            "Called Terminate Thread on DRS msg thread.\n");
        retVal = FPPI_FAIL;
    }

    /* Cleanup thread resources */
    PIL_DestroyThread(thrdHndl, NULL);

DRS_LBL_1:

    if (isCSLocked == TRUE)
    {
        /* Unlock the msg queues so they can be serviced by the threads */
        PIL_LeaveCriticalSection(&(srvcThrdParms->queSyncObj));
    }

    npf_logger_write(LOG_DRS, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/****************************************************************************
 * Function: FPPAPI_Shutdown()                                              *
 * ---------------------------                                              *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: This function is used to de-initialize the DRS module.      *
 *              It destroys resources initialized by the module.            *
 *                                                                          *
 * Return: FPPI_SUCCESS, FPPI_FAIL                                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET FPPAPI_Shutdown (void)
{
    FPPI_RET retVal;

    npf_logger_write(LOG_DRS, LOG_TRACE,
        "Entering method.\n");

    if (Fppi_Initialized == 0)
    {
        return FPPI_NOT_INITIALIZED;
    }

    /* Destroy all allocated resources */
    if ((retVal = DestroyThreadObj(&g_cpTofpThrdParms)) != FPPI_SUCCESS)
    {
        npf_logger_write(LOG_DRS, LOG_ERR,
            "CP to FP Msg queue destruction failed: err=%d\n", retVal);
    }
    if ((retVal = DestroyThreadObj(&g_fpTocpThrdParms)) != FPPI_SUCCESS)
    {
        npf_logger_write(LOG_DRS, LOG_ERR,
            "FP to CP Msg queue destruction failed: err=%d\n", retVal);
    }

    Fppi_Initialized = 0;

    PIL_DestroyCriticalSection(&fppi_lock);
    PIL_DestroyCriticalSection(&g_beFlagslock);

    npf_logger_write(LOG_DRS, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/****************************************************************************
 * Function: FPPAPI_feBind_Reply()                                          *
 * ---------------------------                                              *
 * Parameters: feid    - unique ID for a FE which is generated by TPlugin   *
 *             retVal  - defines whether to accept the FE or not            *
 *                                                                          *
 * Description: This function is used to respond to a FE bind event.        *
 *              The client can decide to accept or reject the FE.           *
 *                                                                          *
 * Return: FPPI_SUCCESS, FPPI_FAIL                                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET FPPAPI_feBind_Reply(
    FPPI_FEID    feid,
    FPPI_Status  status)
{
    FPPI_RET retVal = FPPI_SUCCESS;
    NPF_CBCAT cbType;

    npf_logger_write(LOG_DRS, LOG_TRACE,
        "Entering method.\n");

    if ((Fppi_Initialized == 0) || (Fppi_Started == 0))
        return FPPI_NOT_INITIALIZED;

    if (status == FPPI_SUCCESSFUL)
    {
        cbType = FPPI_FEBind_Event;
    }
    else if (status == FPPI_FAILURE)
    {
        cbType = FPPI_CEUnBind_Event;
    }
    else {
        retVal = FPPI_INVALID_PARAMETERS;
        npf_logger_write(LOG_DRS, LOG_ERR, "Invalid status\n");
        goto DRS_LBL_1;
    }

    retVal = fppapi_MsgToFP(
        feid, 0, NULL, 0, cbType, &status, 1);

DRS_LBL_1:

    npf_logger_write(LOG_DRS, LOG_TRACE,
        "Leaving method.\n");

    return retVal;
}

/****************************************************************************
 * Function: FPPAPI_feCapability_Reply()                                    *
 * ---------------------------                                              *
 * Parameters: feid - unique ID for a FE which is generated by TPlugin      *
 *                                                                          *
 * Description: This function is called after the FE sends its Capabilities *
 *              to the CE. This is used to notify the FE that the CE is     *
 *              ready to receive async events from the FE.                  *
 *                                                                          *
 * Return: FPPI_SUCCESS, FPPI_FAIL                                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET FPPAPI_feCapability_Reply(
        FPPI_FEID       feid)
{
    npf_logger_write(LOG_DRS, LOG_TRACE,
        "Entering and leaving method.\n");

    return fppapi_MsgToFP(
        feid, 0, NULL, 0, FPPI_CEReady_Event, NULL, 0);
}

/****************************************************************************
 * Function: fppapi_MsgToFP()                                               *
 * ---------------------------                                              *
 * Parameters: feid         - unique ID for a FE generated by TPlugin       *
 *             nhid         - unique ID for a dest network address          *
 *             correlator   - used to distinguish multiple invocations      *
 *             cbhandle     - unique ID used to identify the callback fn    *
 *             cbtype       - enum that defines the CB type                 *
 *             policy_list  - array of policy entries to be added           *
 *             policy_count - the number of entries in the array            *
 *                                                                          *
 * Description: Map input params and call the core method to add a policy.  *
 *                                                                          *
 * Return: FPPI_SUCCESS, FPPI_FAIL                                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET fppapi_MsgToFP(
    FPPI_FEID        feid,
    FPPI_NHID        nhid,
    FPPI_CORRELATOR  correlator,
    FPPI_CBHANDLE    cbhandle,
    NPF_CBCAT        cbtype,
    void             *policy_list,
    uint32_t         policy_count)
{
    return fppapi_MsgToFPCore(
        feid,
        nhid,
        0,
        0,
        0,
        0,
        correlator,
        cbhandle,
        cbtype,
        policy_list,
        policy_count,
        0,
        0);
}

/****************************************************************************
 * Function: fppapi_MsgToFPCore()                                           *
 * ---------------------------                                              *
 * Parameters: feid        - unique ID for a FE generated by TPlugin        *
 *             nhid        - user ID used to identify a prefix (n/w addr)   *
 *             portid      - port identifier                                *
 *             dir         - data path direction (ingress, egress, both)    *
 *             tableid     - FE table identifier                            *
 *             instid      - FE table entry instance identifier             *
 *             correlator  - used to distinguish multiple invocations       *
 *             cbhandle    - unique ID used to identify the callback fn     *
 *             cbtype      - enum that defines the CB type                  *
 *             policy_list - pointer to QoS policy list to encapsulate      *
 *             policy_count- number of policies in list                     *
 *             nexttableid - FE table identifier for next stage             *
 *             nextinstid  - next stage FE table entry instance identifier  *
 *                                                                          *
 * Description: This is a common internal fn used by the other FPPAPI fns   *
 *              to add policies to a FE. The policies depend on the caller  *
 *              function and defined by the cbtype parameter. This fn takes *
 *              care of converting the FPPI struct to the COPS PIB policy   *
 *              and then sending a COPS decision message to the FE.         *
 *                                                                          *
 * Return: FPPI_SUCCESS, FPPI_FAIL                                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET fppapi_MsgToFPCore(
    FPPI_FEID             feid,
    FPPI_NHID             nhid,
    FPPI_PortID           portid,
    uint32_t  dir,
    uint32_t      tableid,
    uint32_t   instid,
    FPPI_CORRELATOR       correlator,
    FPPI_CBHANDLE         cbhandle,
    NPF_CBCAT             cbtype,
    void*                 policy_list,
    uint32_t              policy_count,
    uint32_t      nexttableid,
    uint32_t   nextinstid)
{
    FPPI_RET retVal                = FPPI_SUCCESS;
    METHOD_INFO* methInfo          = NULL;
    LPDRS_GENERAL_EVENT_MSG mspPtr = NULL;
    void* outObj                   = NULL;
    BOOL isMsgInQueue              = FALSE;

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Entered method (thread # %d).\n",
        PIL_GetCurrentThreadId());

    if ((Fppi_Initialized == 0) || (Fppi_Started == 0))
    {
        return FPPI_NOT_INITIALIZED;
    }

    /* Store the cb handle and correlator from the CP for msgs that
       expect a callback from the FP. */
    if (cbtype != FPPI_CEReady_Event &&
        cbtype != FPPI_FEBind_Event  &&
        cbtype != FPPI_CEUnBind_Event)
    {
        /* These msgs get a callback from FP to CP so store cb handle
           and correlator */

        npf_logger_write(LOG_DRS, LOG_INFO,
            "Callback data needs to go into list (thread # %d).\n",
            PIL_GetCurrentThreadId());

        /* Check cbhandle in CB List */
        if (pdk_callback_isCBExist(cbhandle) == NPF_INVALID_PARAMETERS)
        {
            retVal = FPPI_INVALID_PARAMETERS;
            goto DRS_LBL_1;
        }

        // Cache cbhandle, correlator in METHOD_INFO List
        if ((methInfo = (METHOD_INFO *)PIL_AllocateMemory
            (sizeof(METHOD_INFO), PIL_MEM_ZERO_INIT)) == NULL)
        {
            retVal = FPPI_OUT_OF_MEMORY;
            npf_logger_write(LOG_DRS, LOG_ERR,
                "Could not allocate memory.\n");
            goto DRS_LBL_1;
        }

        methInfo->cbhandle   = cbhandle;
        methInfo->correlator = correlator;

        npf_logger_write(LOG_DRS, LOG_INFO,
            "Stuff callback data into list (thread # %d).\n",
            PIL_GetCurrentThreadId());

        PIL_EnterCriticalSection(&fppi_lock);
        npf_list_pushBack(&methodInfoList, (void *)methInfo);
        PIL_LeaveCriticalSection(&fppi_lock);

    } // if (cbtype != FPPI_CEReady_Event)

    /* Copy data destined for FP into local DRS memory into the
       struct expected by the FP. */
    if ((retVal = CopyAndPack(cbtype, portid, nhid, dir, tableid, instid,
        nexttableid, nextinstid, policy_count, policy_list, &outObj)) !=
        FPPI_SUCCESS)
    {
        goto DRS_LBL_1;
    }

    /* Allocate memory for mailbox msg */
    if ((mspPtr = (LPDRS_GENERAL_EVENT_MSG)PIL_AllocateMemory
        (sizeof(DRS_GENERAL_EVENT_MSG), PIL_MEM_ZERO_INIT)) == NULL)
    {
        retVal = FPPI_OUT_OF_MEMORY;
        npf_logger_write(LOG_DRS, LOG_ERR, "Could not allocate memory.\n");
        goto DRS_LBL_1;
    }

    /* Fill msg object with data */
    mspPtr->evType     = DRS_MSG_TO_FP;
    mspPtr->cbtype     = cbtype;
    mspPtr->correlator = methInfo;
    mspPtr->outObj     = outObj;

    /* Insert data into inbox msg queue as a msg */
    PIL_EnterCriticalSection(&(g_cpTofpThrdParms.queSyncObj));

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Stuff msg into CP to FP queue (thread # %d).\n",
        retVal, PIL_GetCurrentThreadId());

    npf_list_pushBack(&(g_cpTofpThrdParms.msgQueue), (void *)mspPtr);
    isMsgInQueue = TRUE;

    /* Now notify threads that a message is pending in its queue */
    if (PIL_ReleaseSemaphore(g_cpTofpThrdParms.newMsgNotifier) != PIL_OK)
    {
        npf_logger_write(LOG_DRS, LOG_ERR,
            "Error setting event object.\n");
    }

    /* Unlock the msg queues so they can be serviced by the threads */
    PIL_LeaveCriticalSection(&(g_cpTofpThrdParms.queSyncObj));

DRS_LBL_1:

    if (retVal != FPPI_SUCCESS)
    {
        if (!isMsgInQueue)
        {
            /* Perform full cleanup since msg not put into queue.
               Otherwise leave all data alone since it will be referenced
               when the service thread pops msg from queue. */
            PIL_EnterCriticalSection(&fppi_lock);
            npf_list_removeByKeyFree(&methodInfoList, (void *)methInfo);
            PIL_LeaveCriticalSection(&fppi_lock);

            if (mspPtr != NULL)
            {
                /* Free local mem on error */
                PIL_FreeMemory(mspPtr);
            }

            if (outObj != NULL)
            {
                /* Free local mem on error */
                PIL_FreeMemory(outObj);
            }
        }
    }

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Leaving method with status %d (thread # %d).\n",
        retVal, PIL_GetCurrentThreadId());

    return retVal;
}

