/****************************************************************************
 *             INTEL CORPORATION PROPRIETARY INFORMATION
 *  This software is supplied under the terms of a license agreement or
 *  nondisclosure agreement with Intel Corporation and may not be copied
 *  or disclosed except in accordance with the terms of that agreement.
 *      Copyright (c) 2001 Intel Corporation. All Rights Reserved.
 *
 * $Source: /cvs/constellations/ixa_3.0/src/cp_pdk/ControlPlane/DataRelayShim/src/msgthreads.c,v $
 * $Revision: 1.6 $
 * $Date: 2003/11/17 16:57:15 $
 *
 * Description:
 *      Implements method to create event loops in new threads. The thread
 *      method is also contained here. The event threads wait for and
 *      processes messages from the CP and FP and then call the appropriate
 *      methods in the FP or CP.
 *
 ****************************************************************************/

#include "list.h"
#include "log.h"
#include "fpp_common.h"
#include "msgthreads.h"
#include "bepriv.h"

/********************************* Externs *********************************/

extern PilCriticalSection fppi_lock;
extern DList methodInfoList;

/***************************** Internal Methods *****************************/

/* Forward declaration of event loop thread function */
static void DRSThread(
    void* argPtr);

static FPPI_RET MakeCBIntoCP(
    LPDRS_GENERAL_EVENT_MSG msgPtr);

/****************************************************************************
 * Function: InitThreadObj()                                                *
 * ---------------------------                                              *
 * Parameters: drsThrdParms  - ptr to thread data object                    *
 *                                                                          *
 * Description: Initializes msg queue packages by creating for each: a msg  *
 *              queue, a sync object, and an event obj.                     *
 *                                                                          *
 * Return: FPPI_SUCCESS or an error code on error.                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET InitThreadObj(
    LPQUEUE_SERVICE_PARAMS drsThrdParms)
{
    FPPI_RET retVal = FPPI_FAIL;

    npf_logger_write(LOG_DRS, LOG_INFO, "Entered InitThreadObj.\n");

    /* Clear out params */
    drsThrdParms->thrdHandle       = NULL;
    drsThrdParms->newMsgNotifier   = NULL;
    drsThrdParms->thrdUpNotifier   = NULL;
    drsThrdParms->retVal           = FPPI_SUCCESS;

    /* Init the queue sync object */
    PIL_InitializeCriticalSection(&(drsThrdParms->queSyncObj));

    /* Lock for the rest of init process just in case */
    PIL_EnterCriticalSection(&(drsThrdParms->queSyncObj));

    /* Init the event queues */
    npf_list_init(&(drsThrdParms->msgQueue), PIL_FreeMemory);

    /* Create the new event notifier objects as binary semaphores
       initially set to unsignalled (empty). */
    if ((drsThrdParms->newMsgNotifier = PIL_CreateBinarySemaphore(
        FALSE)) == NULL)
    {
        npf_logger_write(LOG_DRS, LOG_ERR,
            "InitThreadObj - Error creating event obj.\n");
        goto DRS_LBL_1;
    }

    /* Create the thread up events. These are signaled by the threads
       upon startup to the master thread. The master thread may then
       proceed with system init and startup (FP). */
    if ((drsThrdParms->thrdUpNotifier = PIL_CreateBinarySemaphore(
        FALSE)) == NULL)
    {
        npf_logger_write(LOG_DRS, LOG_ERR,
            "InitThreadObj - Error creating event obj.\n");
        goto DRS_LBL_1;
    }

    /* Initializaton successful - skip to end */
    retVal = FPPI_SUCCESS;

DRS_LBL_1:

    /* Unlock structures */
    PIL_LeaveCriticalSection(&(drsThrdParms->queSyncObj));

    if (retVal != FPPI_SUCCESS)
    {
        /* There was an error so destroy all event queue components */
        DestroyThreadObj(drsThrdParms);
    }

    npf_logger_write(LOG_DRS, LOG_INFO,
        "InitThreadObj - Exiting with status %d.\n", retVal);

    return retVal;
}

/****************************************************************************
 * Function: DestroyThreadObj()                                             *
 * ---------------------------                                              *
 * Parameters: drsThrdParms  - ptr to thread data object                    *
 *                                                                          *
 * Description: Destroys all components of the in and out msg queue pkgs.   *
 *                                                                          *
 * Return: FPPI_SUCCESS or an error code on error.                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET DestroyThreadObj(
    LPQUEUE_SERVICE_PARAMS drsThrdParms)
{
    FPPI_RET retVal = FPPI_SUCCESS;

    npf_logger_write(LOG_DRS, LOG_INFO, "Entered DestroyThreadObj.\n");

    /*******************************/
    /* Destroy the in msg package  */

    /* Lock the sync object (if it exists) while destroying contents */
    PIL_EnterCriticalSection(&(drsThrdParms->queSyncObj));

    /* Destroy in event queue and all remaining events in queue */
    npf_list_destroy(&(drsThrdParms->msgQueue));

    /* Destroy the event notifier object */
    if (drsThrdParms->newMsgNotifier != NULL)
    {
        if (PIL_DestroySemaphore(drsThrdParms->newMsgNotifier) != PIL_OK)
        {
            npf_logger_write(LOG_DRS, LOG_ERR,
                "InitThreadObj - Error destroying event object.\n");
            retVal = FPPI_FAIL;
        }
    }

    /* Destroy the thread up event notifier object */
    if (drsThrdParms->thrdUpNotifier != NULL)
    {
        if (PIL_DestroySemaphore(drsThrdParms->thrdUpNotifier) != PIL_OK)
        {
            npf_logger_write(LOG_DRS, LOG_ERR,
                "InitThreadObj - Error destroying event object.\n");
            retVal = FPPI_FAIL;
        }
    }

    /* Unlock the sync object */
    PIL_LeaveCriticalSection(&(drsThrdParms->queSyncObj));

    /* Destroy the queue sync object */
    PIL_DestroyCriticalSection(&(drsThrdParms->queSyncObj));

    npf_logger_write(LOG_DRS, LOG_INFO,
        "DestroyThreadObj - Exiting with status %d.\n", retVal);

    return retVal;
}

/****************************************************************************
 * Function: CreateDRSThread()                                              *
 * ---------------------------                                              *
 * Parameters: drsThrdParms  - ptr to thread data object                    *
 *             threadName    - string used for thread name                  *
 *                                                                          *
 * Description: Create the event loop threads. Pass in relevant parameters. *
 *              The threads execute until a shutdown event is put into the  *
 *              queues and the threads get around to servicing it.          *
 *                                                                          *
 * Return: FPPI_SUCCESS or an error code on error.                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET CreateDRSThread(
    LPQUEUE_SERVICE_PARAMS drsThrdParms,
    char* threadName)
{
    FPPI_RET retVal        = FPPI_SUCCESS;
    PilThread threadHndl   = 0;
    PilThreadRoutine fnPtr = (PilThreadRoutine)DRSThread;

    npf_logger_write(LOG_DRS, LOG_INFO,
            "Entered CreateDRSThread.\n");

    /* Start the thread */
    if ((threadHndl = PIL_CreateThread(
        fnPtr,
        2048,
        threadName,
        4,                            /* priority from 1-8 inclusive */
        (void *)drsThrdParms,
        0)) == NULL)
    {
        npf_logger_write(LOG_DRS, LOG_ERR,
            "CreateDRSThread - Error starting thread.\n");
        retVal = FPPI_FAIL;
        goto DRS_LBL_1;
    }
    else {

        PIL_EnterCriticalSection(&(drsThrdParms->queSyncObj));

        drsThrdParms->thrdHandle = threadHndl;

        PIL_LeaveCriticalSection(&(drsThrdParms->queSyncObj));

        npf_logger_write(LOG_DRS, LOG_INFO,
            "CreateDRSThread - DRS msg thread started.\n");
    }

DRS_LBL_1:

    npf_logger_write(LOG_DRS, LOG_INFO,
        "CreateDRSThread - Exiting with status %d.\n", retVal);

    return retVal;
}

/****************************************************************************
 * Function: DRSThread()                                                    *
 * ---------------------------                                              *
 * Parameters: argPtr - ptr to struct containing necessary data             *
 *                                                                          *
 * Description: Thread that pops events off the msg queue and services      *
 *              them. Events are popped first in first out and deleted.     *
 *                                                                          *
 * Return: FPPI_SUCCESS or an error code on error.                          *
 *                                                                          *
 ****************************************************************************/

void DRSThread(
    void* argPtr)
{
    FPPI_RET retVal                    = FPPI_SUCCESS;
    BOOL loopContin                    = TRUE;
    LPQUEUE_SERVICE_PARAMS thrdParams  = (LPQUEUE_SERVICE_PARAMS)argPtr;
    LPDRS_GENERAL_EVENT_MSG eventObj   = NULL;
    BOOL isObjLocked                   = FALSE;

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Entered method (thread # %d).\n", PIL_GetCurrentThreadId());

    PIL_EnterCriticalSection(&(thrdParams->queSyncObj));
    isObjLocked = TRUE;

    /* Notify the master thread that this thread has started and
       entered the wait cycle. */
    if (PIL_ReleaseSemaphore(thrdParams->thrdUpNotifier) != PIL_OK)
    {
        npf_logger_write(LOG_DRS, LOG_ERR,
            "DRSThread - Error setting event object (thread # %d).\n",
             PIL_GetCurrentThreadId());
        retVal = FPPI_FAIL;
        goto DRS_LBL_1;
    }

    PIL_LeaveCriticalSection(&(thrdParams->queSyncObj));
    isObjLocked = FALSE;

    while (loopContin)
    {
        npf_logger_write(LOG_DRS, LOG_INFO,
            "DRSThread - waiting for event (thread # %d).\n",
            PIL_GetCurrentThreadId());

        /* Wait for an event to arrive */
        if (PIL_AcquireSemaphore(
            thrdParams->newMsgNotifier,
            PIL_INFINITE_TIMEOUT)
            != PIL_OK)
        {
            npf_logger_write(LOG_DRS, LOG_ERR,
                "DRSThread - Error waiting for event (thread # %d).\n",
                PIL_GetCurrentThreadId());

            /* Exit loop */
            break;
        }

//        sleep(2);
        
        npf_logger_write(LOG_DRS, LOG_INFO,
            "DRSThread - event arrived, wait ended (thread # %d).\n",
            PIL_GetCurrentThreadId());

        /* Lock the event queue to access events. May want to eventually
           copy event from queue into local mem so can unlock queue right
           away instead of holding for entire callback into FP/CP. */
        PIL_EnterCriticalSection(&(thrdParams->queSyncObj));
        isObjLocked = TRUE;

        /* Event has arrived. Enter loop to service all events
           in queue until queue is empty. */
        while (npf_list_popFront(&(thrdParams->msgQueue),
            (void**)(&eventObj)) != -1)
        {
            if (isObjLocked)
            {
                /* Leave CS now to free access. List entry has been popped. */
                PIL_LeaveCriticalSection(&(thrdParams->queSyncObj));
                isObjLocked = FALSE;
            }

            npf_logger_write(LOG_DRS, LOG_TRACE,
                "An entry has been popped from list: type=%d, cbtype=%d" \
                " (thread # %d).\n", eventObj->evType, eventObj->cbtype,
                PIL_GetCurrentThreadId());

            /* Act on the event by calling appropriate handler method */
            switch (eventObj->evType)
            {
            case DRS_MSG_TO_FP:

                /* Call appropriate callback method into FP */
                if ((retVal = MakeCBIntoFP(eventObj)) != FPPI_SUCCESS)
                {
                    npf_logger_write(LOG_DRS, LOG_ERR,
                        "Error making callback into FP (thread # %d).\n",
                        PIL_GetCurrentThreadId());
                }

                break;

            case DRS_EVENT_MSG_TO_CP:
            case DRS_RESPONSE_MSG_TO_CP:

                /* Call appropriate callback method into CP */
                if ((retVal = MakeCBIntoCP(eventObj)) != FPPI_SUCCESS)
                {
                    npf_logger_write(LOG_DRS, LOG_ERR,
                        "Error making callback into CP (thread # %d).\n",
                        PIL_GetCurrentThreadId());
                }

                break;

            case DRS_EVENT_SHUTDOWN:
                loopContin = FALSE;

                /* Break out of local switch */
                break;

            default:
                npf_logger_write(LOG_DRS, LOG_ERR,
                    "DRSThread - Unrecognized event (thread # %d).\n",
                    PIL_GetCurrentThreadId());
                break;

            } /* switch (eventObj->evType) */

            /* Free DRS mem (allocated in method that placed this msg
               in the mailbox). All memory for embedded msg data was
               allocated as a single block. */
            if (eventObj->evType != DRS_EVENT_SHUTDOWN)
            {
                /* Free embedded memory */
                if(eventObj->outObj != NULL)
                   PIL_FreeMemory(eventObj->outObj);
            }

            /* Free msg object */
            PIL_FreeMemory(eventObj);

            if (retVal != FPPI_SUCCESS)
            {
                /* This is where execution may break out of the while
                   loop and shutdown the thread if an error is serious
                   enough. At the time of this writing, an error would
                   occur while making a callback into the FP or CP.
                   In these cases, the queue service thread should
                   normally not break, but instead log an error message
                   and continue servicing the queue. */
            }

            if (loopContin == FALSE)
            {
                /* Break out of local while loop even if there are
                   more msgs in queue. Quit now! */
                break;
            }

        } /* while (npf_list_popBack(&(thrdParams->msgQueue), ... */

        /* Set the return value in case another thread cares */
        thrdParams->retVal = retVal;

        /* Unlock the event queue */
        if (isObjLocked)
        {
            PIL_LeaveCriticalSection(&(thrdParams->queSyncObj));
            isObjLocked = FALSE;
        }

    } /* while (loopContin) */

DRS_LBL_1:

    /* Unlock the event queue */
    if (isObjLocked)
    {
        PIL_LeaveCriticalSection(&(thrdParams->queSyncObj));
        isObjLocked = FALSE;
    }

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Exiting method with status %d (thread # %d).\n",
         retVal, PIL_GetCurrentThreadId());

    return;
}

/****************************************************************************
 * Function: MakeCBIntoCP()                                                 *
 * ---------------------------                                              *
 * Parameters: cbtype      - enum that defines the CB type                  *
 *             correlator  - used to distinguish multiple invocations       *
 *             outObj      - data to pass to the CP                         *
 *                                                                          *
 * Description: The callback methods that the CP has registered with the    *
 *              FP are invoked in this method. They are invoked             *
 *              sequentially in a loop.                                     *
 *                                                                          *
 *              This method must be compiled under the CP because it        *
 *              contains pdk_callback... calls. This is an artifact of      *
 *              how the callback list is implemented in the colocated       *
 *              PDK. Separate lists are maintained for the CP and the       *
 *              FP, and are identified using #defines during the            *
 *              compilation process.                                        *
 *                                                                          *
 * Return: FPPI_SUCCESS, FPPI_FAIL                                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET MakeCBIntoCP(
    LPDRS_GENERAL_EVENT_MSG msgPtr)
{
    FPPI_RET retVal = FPPI_SUCCESS;
    FPPI_CONTEXT context;
    FPPI_CBHANDLE cbid     = 0;
    FPPI_CBHANDLE nextcbid = 0;
    fppapi_callback cb     = NULL;
    METHOD_INFO* methInfo  = NULL;
    FPPI_CORRELATOR cpCorr = NULL;

    npf_logger_write(LOG_DRS, LOG_TRACE,
        "Entered method with event type %d (thread # %d).\n",
        msgPtr->evType, PIL_GetCurrentThreadId());

    switch (msgPtr->evType)
    {
    case DRS_EVENT_MSG_TO_CP:

        /* First handle the events - msgs coming from FP that are not
           in response to an earlier msg from the CP. */

        /* Retrieve the first CB for this event */
        if ((cb = (fppapi_callback)pdk_callback_firstFunc(
            msgPtr->cbtype, &context, &nextcbid)) == NULL)
        {
            npf_logger_write(LOG_DRS, LOG_INFO,
                "No callbacks registered for this event.\n");
            goto DRS_LBL_1;
        }

        /* Make all registered callbacks in a loop */
        do {
            cbid = nextcbid;

            npf_logger_write(LOG_DRS, LOG_TRACE,
                "Make event CB into CP (thread # %d).\n",
                msgPtr->evType, PIL_GetCurrentThreadId());

            /* Make the callback into CP */
            cb(msgPtr->feid, cpCorr, context, msgPtr->outObj);

        } while ((cb = (fppapi_callback)pdk_callback_nextFunc(
            cbid, &context, &nextcbid)) != NULL);

        break;

    case DRS_RESPONSE_MSG_TO_CP:

        /* Now handle msgs that are in response to an earlier msg from
           the CP. Base callback on the cbid passed in originally from CP.
           Must also pass back correlator passed in by original CP call. */

        /* Make sure there is a correlator passed up from FP */
        if (msgPtr->correlator == NULL)
        {
            npf_logger_write(LOG_DRS, LOG_ERR,
                "Mandatory correlator not available.\n");
            retVal = FPPI_FAIL;
            goto DRS_LBL_1;
        }

        /* Get the CP correlator and cbhandle from the methodInfoList */
        PIL_EnterCriticalSection(&fppi_lock);
        npf_list_removeByKey(&methodInfoList,
            msgPtr->correlator, (void**)&methInfo);
        PIL_LeaveCriticalSection(&fppi_lock);

        if (methInfo == NULL)
        {
            npf_logger_write(LOG_DRS, LOG_ERR,
                "Could not find DRS correlator.\n");
            retVal = FPPI_FAIL;
            goto DRS_LBL_1;
        }
        else {
            /* Set correlator used by CP and callback ID */
            cbid   = methInfo->cbhandle;
            cpCorr = methInfo->correlator;

            /* Now can free the methinfo object */
            PIL_FreeMemory(methInfo);
        }

        if ((cb = (fppapi_callback)pdk_callback_func(cbid, &context))
            == NULL)
        {
            npf_logger_write(LOG_DRS, LOG_INFO,
                "No callbacks registered for this event.\n");
            goto DRS_LBL_1;
        }

        npf_logger_write(LOG_DRS, LOG_TRACE,
            "Make response CB into CP (thread # %d).\n",
            msgPtr->evType, PIL_GetCurrentThreadId());

        /* Make the callback into the CP */
        cb(msgPtr->feid, cpCorr, context, msgPtr->outObj);

        break;

    default:
        npf_logger_write(LOG_DRS, LOG_ERR,
            "Not valid CP CB msg type %d \n", msgPtr->evType);
    }

DRS_LBL_1:

    npf_logger_write(LOG_DRS, LOG_TRACE,
        "Leaving method with status %d.\n", retVal);

    return retVal;
}

