/****************************************************************************
 *             INTEL CORPORATION PROPRIETARY INFORMATION
 *  This software is supplied under the terms of a license agreement or
 *  nondisclosure agreement with Intel Corporation and may not be copied
 *  or disclosed except in accordance with the terms of that agreement.
 *      Copyright (c) 2003 Intel Corporation. All Rights Reserved.
 *
 * $Source: /cvs/constellations/ixa_3.0/src/cp_pdk/ControlPlane/TransportPlugin/src/fppcommon.c,v $
 * $Revision: 1.6 $
 * $Date: 2003/11/17 16:57:17 $
 *
 * Description:
 * Contains some compare methods, init and shutdown methods, and some
 * helper methods called from fppapi.cc.
 *
 ****************************************************************************/

#include "stdio.h"
#include "string.h"
#ifdef FPPI_TIMING
#include "sys/time.h"
#endif
#include "unistd.h"
#include "errno.h"

#include "fppapi.h"
#include "fppapi_types.h"

#include "log.h"
#include "fpppriv.h"
#include "fppencap_forces.h"

#include "interconnect_api.h"

/********************************* Globals *********************************/

int Fppi_Initialized = 0; /* Has TP been inited? */
int Fppi_Started = 0;     /* Has TP been started? */

PilCriticalSection  fppi_lock; /* critical section to protect globals */

DList    FEList; /* list of bound FE's */

static uint32_t g_fpp_correlator = 0; /* each fpp command needs unique correlator, this global is incremented and used for each command */

#ifdef FPPI_TIMING
struct timeval report_strt;
#endif

#define INTERCONNECT_BUFFER_COUNT 6

/****************************************************************************
 Declare local methods
 ***************************************************************************/
int fpp_felist_compare(const void* data, const void* key)
{
	return (((FEInfo)data)->feid == (FPPI_FEID)key);
}

FPPI_RET internal_feCapability_Request(FEInfo feinfo);
FPPI_RET internal_feUnbindRequest(FEInfo feinfo);

/****************************************************************************
 * Function: FPPAPI_register_callback()
 * ---------------------------
 * Parameters: callback - the callback function being registered
 *             cbhandle - unique ID generated by the Plugin, used to
 *                        identify the callback function
 *             cbtype   - the callback type for which the fn is registered
 *             context  - value passed by client which will be
 *                        returned when callback is invoked
 *
 * Description: This is the generic function used by the client
 *              to register a callback function for a particular event
 *              call identified by the callback type.
 *
 * Return: FPPI_SUCCESS, FPPI_FAIL
 *
 ****************************************************************************/

FPPI_RET FPPAPI_register_callback(
    fppapi_callback  callback,
    FPPI_CBHANDLE*   cbhandle,
    NPF_CBCAT        cbtype,
    FPPI_CONTEXT     context) 
{
    if (Fppi_Initialized == 0)
    return FPPI_NOT_INITIALIZED;

    if ( pdk_callback_register( cbtype, context, (NPF_GENERAL_CBFUNC)callback, cbhandle) != NPF_SUCCESS)
    {
        return FPPI_FAIL;
    }
    return FPPI_SUCCESS;
}

/****************************************************************************
 * Function: FPPAPI_deregister_callback()
 * ---------------------------
 * Parameters: cbhandle - unique ID used to identify the callback fn
 *
 * Description: This is the generic function used by the client
 *              to de-register a callback function.
 *
 * Return: FPPI_SUCCESS, FPPI_FAIL
 *
 ****************************************************************************/

FPPI_RET FPPAPI_deregister_callback(
    FPPI_CBHANDLE    cbhandle)
{
    if (Fppi_Initialized == 0)
    return FPPI_NOT_INITIALIZED;

    if (pdk_callback_deregister(cbhandle) != NPF_SUCCESS)
    {
        return FPPI_FAIL;
    }
    return FPPI_SUCCESS;
}

/****************************************************************************
 * Function: FPPAPI_Initialize()                                            *
 * ---------------------------                                              *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: This function is used to initialize the Transport Plugin    *
 *              module in the control plane. It initializes the different   *
 *              lists used by the plugin module.                            *
 *                                                                          *
 * Return: FPPI_SUCCESS, FPPI_FAIL                                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET FPPAPI_Initialize (void)
{
    if (Fppi_Initialized == 1)
        return FPPI_FAIL;

    PIL_InitializeCriticalSection(&fppi_lock);
    PIL_EnterCriticalSection(&fppi_lock);

    npf_list_init(&FEList, PIL_FreeMemory);
    FEList.compare = fpp_felist_compare;

    Fppi_Initialized = 1;

    PIL_LeaveCriticalSection(&fppi_lock);
    return FPPI_SUCCESS;
}

/****************************************************************************
 * Function: FPPAPI_Start()                                                 *
 * ---------------------------                                              *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: This function is used to start the Transport Plugin in      *
 *              terms of initializing the Interconnect layer. After this    *
 *              call the Plugin can receive connections from FEs.           *
 *                                                                          *
 * Return: FPPI_SUCCESS, FPPI_FAIL                                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET FPPAPI_Start (void)
{
    FPPI_RET ret;
    CBServerDatagramFuncs serverparams;

    if ((Fppi_Initialized == 0) || (Fppi_Started == 1))
	    return FPPI_NOT_INITIALIZED;

    PIL_EnterCriticalSection(&fppi_lock);

    pb_PlBufferInit(INTERCONNECT_BUFFER_COUNT);

    serverparams.datagram_receive = ForCES_DatagramReceive;
    serverparams.datagram_exception = ForCES_DatagramException;

    if ( InitializeDatagramServer(&serverparams) != PL_SUCCESS )
    {
        ret = FPPI_FAIL;
    }
    else
    {
        Fppi_Started = 1;
        ret = FPPI_SUCCESS;
    }

    PIL_LeaveCriticalSection(&fppi_lock);
    return ret;
}

/****************************************************************************
 * Function: FPPAPI_Shutdown()
 * ---------------------------
 * Parameters: none
 *
 * Description: This function is used to de-initialize the Transport Plugin
 *              module. It destroys the lists initialized by the module.
 *              It must be called after initialize and after stop.
 *
 * Return: FPPI_SUCCESS, FPPI_FAIL
 *
 ****************************************************************************/

FPPI_RET FPPAPI_Shutdown (void)
{
    if (Fppi_Initialized == 0)
        return FPPI_NOT_INITIALIZED;

    PIL_EnterCriticalSection(&fppi_lock);

    Fppi_Initialized = 0;

    pb_PlBufferDeInit();
    DeInitializeDatagram();

    npf_list_destroy(&FEList);

    PIL_LeaveCriticalSection(&fppi_lock);
    PIL_DestroyCriticalSection(&fppi_lock);

    return FPPI_SUCCESS;
}

/****************************************************************************
 * Function: FPPAPI_Stop()                                                  *
 * ---------------------------                                              *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: This function is used to stop the Transport Plugin in       *
 *              terms of de-initializing the Interconnect layer. After this *
 *              call the Plugin deletes all connections from FEs.           *
 *                                                                          *
 * Return: FPPI_SUCCESS, FPPI_FAIL                                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET FPPAPI_Stop (void)
{
    DListItr *itr;
    FEInfo feinfo;

    if ((Fppi_Initialized == 0) || (Fppi_Started == 0))
	    return FPPI_NOT_INITIALIZED;

    PIL_EnterCriticalSection(&fppi_lock);

    Fppi_Started = 0;

    // send an unbind to all bound FEs

    feinfo = NULL;

    if ((itr = npf_list_itrCreate(&FEList)) != NULL)
    {
        npf_list_itrFirst(itr);
        do {
            feinfo = (FEInfo)npf_list_itrGetData(itr);
            internal_feUnbindRequest(feinfo);

            /* will be deleted on unbind callback */

        } while (npf_list_itrNext(itr) != -1);
        npf_list_itrDelete(itr);
    }

    PIL_LeaveCriticalSection(&fppi_lock);
    return FPPI_SUCCESS;
}


/****************************************************************************
 * Function: internal_feCapability_Request()
 * ---------------------------
 * Parameters: feinfo  - handle to FE that capabiltiy request is
 *                       being sent to
 *
 * Description: This function is an internal function used to send
 *              a capability request to an FE
 *
 * Return: FPPI_SUCCESS, FPPI_FAIL
 *
 ****************************************************************************/
FPPI_RET internal_feCapability_Request(FEInfo feinfo)
{
    ForCESMessage msg;

    msg = newForCESMessage(feinfo,Capability_Request,0,0);

    if (!msg) return FPPI_FAIL;

    ForCESMessageSend(feinfo,msg,0);

    return FPPI_SUCCESS;
}

/****************************************************************************
 * Function: FPPAPI_feBind_Reply()                                          *
 * ---------------------------                                              *
 * Parameters: feid    - unique ID for a FE which is generated by TPlugin   *
 *             status  - defines whether to accept the FE or not            *
 *                                                                          *
 * Description: This function is used to respond to a FE bind event.        *
 *              The client can decide to accept or reject the FE.           *
 *                                                                          *
 * Return: FPPI_SUCCESS, FPPI_FAIL                                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET FPPAPI_feBind_Reply(
    FPPI_FEID    feid,
    FPPI_Status  status)
{
    FEInfo feinfo;
    ForCESMessage msg;
    forces_bind_resp_t resp;

    FPPI_RET retstatus = FPPI_SUCCESS;

    npf_logger_write (LOG_TPI, LOG_INFO,
        "FPPAPI_feBind_Reply enter\n");

    if ((Fppi_Initialized == 0) || (Fppi_Started == 0))
        return FPPI_NOT_INITIALIZED;

    feinfo = FEInfoFind(feid);

    if (!feinfo) return FPPI_INVALID_PARAMETERS;

    PIL_EnterCriticalSection(&fppi_lock);

    msg = newForCESMessage(feinfo,FE_Bind_Response,0,sizeof(forces_bind_resp_t));
    if (!msg)
    {
	    retstatus = FPPI_OUT_OF_MEMORY;
	    goto TPI_LBL;
    }

    if (status == FPPI_SUCCESSFUL)
    {
	    resp.bind_result = 1; /* set generic data to successful flag */
    }
    else if (status == FPPI_FAILURE)
    {
	    resp.bind_result = 0; /* set generic data to failure flag */
    } else
    {
	    retstatus = FPPI_INVALID_PARAMETERS;
	    goto TPI_LBL;
    }

    encapsulate_forces_bind_resp_t(&resp,(char*)&(msg->data));
    ForCESMessageSend(feinfo,msg,sizeof(forces_bind_resp_t));

    internal_feCapability_Request(feinfo);

TPI_LBL:
    PIL_LeaveCriticalSection(&fppi_lock);
    return retstatus;
}

/****************************************************************************
 * Function: FPPAPI_feCapability_Reply()                                    *
 * ---------------------------                                              *
 * Parameters: feid - unique ID for a FE which is generated by TPlugin      *
 *                                                                          *
 * Description: This function is called after the FE sends its Capabilities *
 *              to the CE. This is used to notify the FE that the CE is     *
 *              ready to receive async events from the FE.                  *
 *                                                                          *
 * Return: FPPI_SUCCESS, FPPI_FAIL                                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET FPPAPI_feCapability_Reply(
        FPPI_FEID       feid)
{
    return FPPI_SUCCESS;
}

/****************************************************************************
 * Function: internal_feUnbindRequest()
 * ---------------------------
 * Parameters: feinfo  - handle to FE that unbind request is
 *                       being sent to
 *
 * Description: This function is an internal function used to send
 *              an unbind request to an FE
 *
 * Return: FPPI_SUCCESS, FPPI_FAIL
 *
 ****************************************************************************/
FPPI_RET internal_feUnbindRequest(
                FEInfo feinfo
                )
{
    ForCESMessage msg;
    forces_unbind_t unbind;

    msg = newForCESMessage(feinfo,Unbind,0,sizeof(forces_unbind_t));
    if (!msg)
        return FPPI_OUT_OF_MEMORY;

    /* yes, these should be #defs or something other than
       magic numbers.  but they should go in a forces header somewhere */
    unbind.reason = 0; /* optional */
    unbind.fe_behavior = 2; /* 1==continue, 2==stop */
    encapsulate_forces_unbind_t(&unbind,(char*)&(msg->data));
    ForCESMessageSend(feinfo,msg,sizeof(forces_unbind_t));

    return FPPI_SUCCESS;
}

/****************************************************************************
 * Function: FPPAPI_feUnbindRequest()
 * ---------------------------
 * Parameters: feid  - id of FE that unbind request is
 *                     being sent to
 *
 * Description: This function is the external function used to send
 *              an unbind request to an FE
 *
 * Return: FPPI_SUCCESS, FPPI_FAIL
 *
 ****************************************************************************/
FPPI_RET FPPAPI_feUnbind_Request(
                FPPI_FEID feid
                )
{
    FEInfo feinfo;
    FPPI_RET retstatus = FPPI_SUCCESS;

    if ((Fppi_Initialized == 0) || (Fppi_Started == 0))
        return FPPI_NOT_INITIALIZED;

    PIL_EnterCriticalSection(&fppi_lock);

    feinfo = FEInfoFind(feid);

    if (!feinfo)
    {
        PIL_LeaveCriticalSection(&fppi_lock);
        return FPPI_INVALID_PARAMETERS;
    }

    retstatus =  internal_feUnbindRequest(feinfo);

    PIL_LeaveCriticalSection(&fppi_lock);

    return retstatus;
}

/*
 * default handler is for all those commands/configs/querys that 
 * we dont care about any data responses, just the status
 */
void default_handler(
                FPPI_FEID       feid,
                uint32_t        cmd_type,
                uint32_t        block_type,
                FPPI_CBHANDLE   cbhandle,
                FPPI_CORRELATOR correlator,
                uint32_t        global_result,
                char*           data)
{
    FPPI_Status* status;
    FPPI_CONTEXT    cbcontext;
    fppapi_callback cb;

    status = (FPPI_Status*) PIL_AllocateMemory(sizeof(FPPI_Status),0);

    if (!status) return;

    *status = (FPPI_Status)global_result;

    cb = (fppapi_callback)pdk_callback_func(cbhandle,&cbcontext);

    if (!cb) 
    {
       PIL_FreeMemory(status);
       return;
    }

    cb(feid, correlator, cbcontext, status);

    PIL_FreeMemory(status);

    return;
}

/****************************************************************************
 * Function: newFPPCommand()
 * ---------------------------
 * Parameters: feid       - id of FE command will be sent to
 *             correlator - a correlator that the caller will get returned
 *                          when the response to this command comes back
 *             cbhandle   - callback to invoke when response comes back
 *             handler    - TP handler to decap return values etc. from
 *                          response to this command
 *             cmd_type   - forces command type, add, del, etc.
 *             block_type - which block to send command to, ipv4, ds, etc.
 *             size       - how big a buffer the caller requires to encap
 *                          parameters, data, etc.
 *
 * Description: This function creates a new forces command to be sent
 *              to the FE.  It also stores state required to make callbacks
 *              and the like when the response is sent back from the FE.
 *
 * Return: a valid FPPCommand or NULL
 *
 ****************************************************************************/
FPPCommand newFPPCommand(
                FPPI_FEID           feid,
                FPPI_CORRELATOR     correlator,
                FPPI_CBHANDLE       cbhandle,
                FPPCommandHandler   handler,
                uint32_t            cmd_type,
                uint32_t            block_type,
                size_t              size
                )
{
    FPPCommand fppcmd;
    forces_command_t forcescmd;
    FEInfo feinfo;

    feinfo = FEInfoFind(feid);
    if (!feinfo) return NULL;

    if (!FEInfoIsBlockTypeSupported(feinfo,block_type)) return NULL;

    fppcmd = (FPPCommand)PIL_AllocateMemory(sizeof(struct FPPCommand_t),0);
    if (!fppcmd) return NULL;

    fppcmd->feinfo = feinfo;

    PIL_EnterCriticalSection(&fppi_lock);

    fppcmd->msg = newForCESMessage(
                    feinfo,
                    Config_Query_Request,
                    ++g_fpp_correlator,
                    size + sizeof(forces_command_t));

    if (!fppcmd->msg)
    {
        PIL_LeaveCriticalSection(&fppi_lock);
        deleteFPPCommand(fppcmd);
        return NULL;
    }

    /* fill in forces_command_t */
    forcescmd.cmd_type = cmd_type;
    forcescmd.block_type = block_type;
    forcescmd.block_handle = 0;
    forcescmd.length = size;
    fppcmd->buffer = encapsulate_forces_command_t(&forcescmd,(char*)&(fppcmd->msg->data));

    fppcmd->size = size + sizeof(forces_command_t);
    fppcmd->forces_correlator = g_fpp_correlator;
    fppcmd->fppi_correlator = correlator;
    fppcmd->fppi_cbhandle = cbhandle;
    fppcmd->handler = handler;
    fppcmd->cmd_type = cmd_type;
    fppcmd->block_type = block_type;

    PIL_LeaveCriticalSection(&fppi_lock);
    return fppcmd;
}

/****************************************************************************
 * Function: deleteFPPCommand()
 * ---------------------------
 * Parameters: fppcmd - command to delete
 *
 * Description: This function deletes a forces command and frees the memory
 *
 * Return: nothing
 *
 ****************************************************************************/
void deleteFPPCommand(FPPCommand fppcmd)
{
    if (!fppcmd) return;

    deleteForCESMessage(fppcmd->msg);

    PIL_FreeMemory(fppcmd);
}

/****************************************************************************
 * Function: FPPCommandSend()
 * ---------------------------
 * Parameters: fppcmd - command to send
 *
 * Description: This function sends a forces command off to the FE
 *
 * Return: FPPI_SUCCESS, FPPI_FAIL
 *
 ****************************************************************************/
FPPI_RET FPPCommandSend(FPPCommand fppcmd)
{
    if (!fppcmd) return FPPI_FAIL;

    FEInfoAddOutstandingCommand(fppcmd->feinfo,fppcmd);

    ForCESMessageSend(fppcmd->feinfo,fppcmd->msg,fppcmd->size);

    return FPPI_SUCCESS;
}
