/****************************************************************************
 *             INTEL CORPORATION PROPRIETARY INFORMATION
 *  This software is supplied under the terms of a license agreement or
 *  nondisclosure agreement with Intel Corporation and may not be copied
 *  or disclosed except in accordance with the terms of that agreement.
 *      Copyright (c) 2001 Intel Corporation. All Rights Reserved.
 *
 * $Source: /cvs/constellations/ixa_3.0/src/cp_pdk/ForwardingPlane/DataRelayShim/src/backendapi.c,v $
 * $Revision: 1.6 $
 * $Date: 2003/11/17 16:57:19 $
 *
 * Description:
 * This file implements the Backend API defined in backendapi.h, it also 
 * implements the CbPibResponse() callback interface exposed by the ForCES PIB 
 * layer. In addition, it contains the impl. of some of the compare functions 
 * used by the ReportList.
 *
 ****************************************************************************/

#include "string.h"
#include "log.h"
#include "pil.h"
#include "fpp_common.h"
#include "pdk_callback.h"
#include "fppCapabilityTypes.h"
#include "msgthreads.h"
#include "bepriv.h"

/********************************* Externs *********************************/

extern PilCriticalSection fppi_lock;
extern DList methodInfoList;

/********************************* Globals *********************************/

/* Declare a crit section to protect globals */
PilCriticalSection g_beFlagslock;
static int32_t Bend_Initialized = 0;

/* Externed and set to non-zero in CP after FP calls BENDAPI_CapsReqst.
   This inturn sends msg through DRS (registered CB) to CP which calls
   FPPAPI_FECaps_Reply. This sets Bend_Ready to non-zero before msg is
   sent to FP. */
int32_t Bend_Ready              = 0;
FPPI_FEID_MAP g_feInfo;

/****************************************************************************
 * Function: BENDAPI_register_callback()                                    *
 * ---------------------------                                              *
 * Parameters: callback - the callback function being registered            *
 *             cbhandle - unique ID generated by the Plugin, used to        *
 *                        identify the callback function                    *
 *             cbtype   - the callback type for which the fn is registered  *
 *             context  - value passed by client which should be returned   *
 *                        returned when callback is invoked                 *
 *                                                                          *
 * Description: This is the generic function used by the client             *
 *              to register a callback function for a particular event/     *
 *              call identified by the callback type.                       *
 *                                                                          *
 * Return: FPPI_SUCCESS, FPPI_FAIL                                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET BENDAPI_register_callback(
    backendapi_callback  callback,
    FPPI_CBHANDLE        *cbhandle,
    NPF_CBCAT            cbtype,
    FPPI_CONTEXT         context)
{
    FPPI_RET retVal = FPPI_SUCCESS;

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Entered method (thread # %d).\n", PIL_GetCurrentThreadId());

    PIL_EnterCriticalSection(&g_beFlagslock);

    if (Bend_Initialized == 0)
    {
        retVal = FPPI_NOT_INITIALIZED;
        PIL_LeaveCriticalSection(&g_beFlagslock);
        goto DRS_LBL_1;
    }

    PIL_LeaveCriticalSection(&g_beFlagslock);

    if (pdk_callback_register(cbtype, context,
        (NPF_GENERAL_CBFUNC)callback, cbhandle) != NPF_SUCCESS)
    {
        npf_logger_write(LOG_TPI, LOG_ERR, "register failed \n");
        retVal = FPPI_FAIL;
    }

DRS_LBL_1:

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Leaving method (thread # %d).\n", PIL_GetCurrentThreadId());

    return retVal;
}

/****************************************************************************
 * Function: BENDAPI_deregister_callback()                                  *
 * ---------------------------                                              *
 * Parameters: cbhandle  - unique ID used to identify the callback fn       *
 *                                                                          *
 * Description: This is the generic function used by the client             *
 *              to de-register a callback function.                         *
 *                                                                          *
 * Return: FPPI_SUCCESS, FPPI_FAIL                                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET BENDAPI_deregister_callback(
    FPPI_CBHANDLE    cbhandle)
{
    FPPI_RET retVal = FPPI_SUCCESS;

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Entered method (thread # %d).\n", PIL_GetCurrentThreadId());

    PIL_EnterCriticalSection(&g_beFlagslock);

    if (Bend_Initialized == 0)
    {
        retVal = FPPI_NOT_INITIALIZED;
        PIL_LeaveCriticalSection(&g_beFlagslock);
        goto DRS_LBL_1;
    }

    PIL_LeaveCriticalSection(&g_beFlagslock);

    if (pdk_callback_deregister(cbhandle) != NPF_SUCCESS)
    {
        retVal = FPPI_FAIL;
    }

DRS_LBL_1:

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Leaving method (thread # %d).\n", PIL_GetCurrentThreadId());

    return retVal;
}

/****************************************************************************
 * Function: BENDAPI_Initialize()                                           *
 * ---------------------------                                              *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: This function is used to initialize the Transport Plugin    *
 *              module in the forwarding plane. It initializes the different*
 *              lists used by the plugin module.                            *
 *                                                                          *
 * Return: FPPI_SUCCESS, FPPI_FAIL                                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET BENDAPI_Initialize (void)
{
    FPPI_RET retVal = FPPI_SUCCESS;

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Entered method (thread # %d).\n", PIL_GetCurrentThreadId());

    PIL_EnterCriticalSection(&g_beFlagslock);

    if (Bend_Initialized == 1)
    {
        retVal = FPPI_FAIL;
    }
    else {
        Bend_Initialized = 1;
    }

    g_feInfo.isSet = FALSE;

    PIL_LeaveCriticalSection(&g_beFlagslock);

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Leaving method (thread # %d).\n", PIL_GetCurrentThreadId());

    return retVal;
}

/****************************************************************************
 * Function: BENDAPI_Shutdown()                                             *
 * ---------------------------                                              *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: This function is used to de-initialize the Transport Plugin *
 *              module. It destroys the lists initialized by the module.    *
 *                                                                          *
 * Return: FPPI_SUCCESS, FPPI_FAIL                                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET BENDAPI_Shutdown (void)
{
    FPPI_RET retVal = FPPI_SUCCESS;

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Entered method (thread # %d).\n", PIL_GetCurrentThreadId());

    PIL_EnterCriticalSection(&g_beFlagslock);

    if (Bend_Initialized == 0)
    {
        retVal = FPPI_NOT_INITIALIZED;
    }
    else {
        Bend_Initialized = 0;
    }

    g_feInfo.isSet = FALSE;

    PIL_LeaveCriticalSection(&g_beFlagslock);

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Leaving method (thread # %d).\n", PIL_GetCurrentThreadId());

    return retVal;
}

/****************************************************************************
 * Function: BENDAPI_bindRequest()                                          *
 * ---------------------------                                              *
 * Parameters: ipaddr    - the IP Address of the Control Element            *
 *             fename    - the string ID of the Control Element             *
 *             cbhandle  - the ID used to identify the callback fn          *
 *                                                                          *
 * Description: This fn is used to send a Bind Request to the CP. An object *
 *              is created to hold the association of identification info   *
 *              passed in from the FE and a locally created FE ID passed    *
 *              in to the CP.
 *                                                                          *
 * Return: FPPI_SUCCESS on success, otherwise error code.                   *
 *                                                                          *
 ****************************************************************************/

FPPI_RET BENDAPI_bindRequest(
    char*            ipaddr,
    char*            fename,
    FPPI_FEID        fpFeid,
    FPPI_CBHANDLE    cbhandle)
{
    FPPI_RET retVal  = FPPI_SUCCESS;
    NPF_CBCAT cbType = FPPI_FEBind_Event;

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Entered method (thread # %d).\n", PIL_GetCurrentThreadId());

    PIL_EnterCriticalSection(&g_beFlagslock);

    if (Bend_Initialized == 0)
    {
        retVal = FPPI_NOT_INITIALIZED;
        PIL_LeaveCriticalSection(&g_beFlagslock);
        goto DRS_LBL_1;
    }

    /* Bind successful so set global object with bind data */
    g_feInfo.feid = fpFeid;
    strcpy(g_feInfo.feName, fename);
    strcpy(g_feInfo.feIPAddr, ipaddr);
    g_feInfo.isSet = TRUE;

    PIL_LeaveCriticalSection(&g_beFlagslock);

    /* Put msg into mbox */
    if ((retVal = StuffMsgIntoFPtoCPQueue(
        drs_FPEvent, cbType, fpFeid, NULL, NULL)) != FPPI_SUCCESS)
    {
        goto DRS_LBL_1;
    }

DRS_LBL_1:

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Leaving method (thread # %d).\n", PIL_GetCurrentThreadId());

    return retVal;
}

/****************************************************************************
 * Function: BENDAPI_unbindRequest()                                        *
 * ---------------------------                                              *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: This fn is used to send a UnBind Request to the Control Pl. *
 *                                                                          *
 * Return: FPPI_SUCCESS on success, otherwise error code.                   *
 *                                                                          *
 ****************************************************************************/

FPPI_RET BENDAPI_unbindRequest(void)
{
    FPPI_RET retVal  = FPPI_SUCCESS;
    NPF_CBCAT cbType = FPPI_FEUnBind_Event;
    FPPI_FEID feid;

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Entered method (thread # %d).\n", PIL_GetCurrentThreadId());

    PIL_EnterCriticalSection(&g_beFlagslock);

    if (Bend_Initialized == 0 || g_feInfo.isSet == FALSE)
    {
        retVal = FPPI_NOT_INITIALIZED;
        PIL_LeaveCriticalSection(&g_beFlagslock);
        goto DRS_LBL_1;
    }

    feid = g_feInfo.feid;

    PIL_LeaveCriticalSection(&g_beFlagslock);

    /* Put msg into mbox */
    if ((retVal = StuffMsgIntoFPtoCPQueue(
        drs_FPEvent, cbType, feid, NULL, NULL)) != FPPI_SUCCESS)
    {
        goto DRS_LBL_1;
    }

    PIL_EnterCriticalSection(&g_beFlagslock);

    g_feInfo.isSet = FALSE;

    PIL_LeaveCriticalSection(&g_beFlagslock);

DRS_LBL_1:

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Leaving method (thread # %d).\n", PIL_GetCurrentThreadId());

    return retVal;
}

/****************************************************************************
 * Function: BENDAPI_capabilityRequest()                                    *
 * ---------------------------                                              *
 * Parameters: fe_caps   - FE Capability struct                             *
 *             cbhandle  - the ID used to identify the callback fn          *
 *                                                                          *
 * Description: This fn is used to send FE Capabilities to the CP           *
 *                                                                          *
 * Return: FPPI_SUCCESS on success, otherwise error code.                   *
 *                                                                          *
 ****************************************************************************/

FPPI_RET BENDAPI_capabilityRequest(
    FPPI_FE_Caps*   fe_caps,
    FPPI_CBHANDLE   cbhandle)    /* Not used */
{
    FPPI_RET retVal  = FPPI_SUCCESS;
    NPF_CBCAT cbType = FPPI_FECapability_Event;
    void* outObj = NULL;
    FPPI_FEID feid;

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Entered method (thread # %d).\n", PIL_GetCurrentThreadId());

    PIL_EnterCriticalSection(&g_beFlagslock);

    if (Bend_Initialized == 0 || g_feInfo.isSet == FALSE)
    {
        retVal = FPPI_NOT_INITIALIZED;
        PIL_LeaveCriticalSection(&g_beFlagslock);
        goto DRS_LBL_1;
    }

    feid = g_feInfo.feid;

    PIL_LeaveCriticalSection(&g_beFlagslock);

    /* Copy data destined to CP into local DRS memory */
    if ((retVal = CopyAndPackFPEvent(cbType, (void*)fe_caps, 1, &outObj)) !=
        FPPI_SUCCESS)
    {
        goto DRS_LBL_1;
    }

    /* Put msg into mbox */
    if ((retVal = StuffMsgIntoFPtoCPQueue(
        drs_FPEvent, cbType, feid, NULL, outObj)) != FPPI_SUCCESS)
    {
        goto DRS_LBL_1;
    }

DRS_LBL_1:

    if (retVal != FPPI_SUCCESS)
    {
        if (outObj != NULL)
        {
            /* Free local mem on error */
            PIL_FreeMemory(outObj);
            outObj = NULL;
        }
    }

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Exiting method with status %d (thread # %d).\n",
         retVal, PIL_GetCurrentThreadId());

    return retVal;
}

/****************************************************************************
 * Function: BENDAPI_sendPacket()                                           *
 * ---------------------------                                              *
 * Parameters: packet_list - array of npf_Packet_t structs                  *
 *             packet_count - the number of entries in the array            *
 *                                                                          *
 * Description: This fn is used to send control pkts to the control plane.  *
 *                                                                          *
 * Return: FPPI_SUCCESS on success, otherwise error code.                   *
 *                                                                          *
 ****************************************************************************/

FPPI_RET BENDAPI_sendPacket(
    npf_Packet_t*  packet_list,
    uint32_t       packet_count)
{
    FPPI_RET retVal  = FPPI_SUCCESS;
    NPF_CBCAT cbType = FPPI_PacketIO_Event;
    void* outObj = NULL;
    FPPI_FEID feid;

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Entered method (thread # %d).\n", PIL_GetCurrentThreadId());

    PIL_EnterCriticalSection(&g_beFlagslock);

    if ((Bend_Initialized == 0) || (Bend_Ready == 0))
    {
        retVal = FPPI_NOT_INITIALIZED;
        PIL_LeaveCriticalSection(&g_beFlagslock);
        goto DRS_LBL_1;
    }

    feid = g_feInfo.feid;

    PIL_LeaveCriticalSection(&g_beFlagslock);

    /* Copy data destined to CP into local DRS memory */
    if ((retVal = CopyAndPackFPEvent(
        cbType, (void*)packet_list, packet_count, &outObj)) !=
        FPPI_SUCCESS)
    {
        goto DRS_LBL_1;
    }

    /* Put msg into mbox */
    if ((retVal = StuffMsgIntoFPtoCPQueue(
        drs_FPEvent, cbType, feid, NULL, outObj)) != FPPI_SUCCESS)
    {
        goto DRS_LBL_1;
    }

DRS_LBL_1:

    if (retVal != FPPI_SUCCESS)
    {
        if (outObj != NULL)
        {
            /* Free local mem on error */
            PIL_FreeMemory(outObj);
            outObj = NULL;
        }
    }

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Leaving method (thread # %d).\n", PIL_GetCurrentThreadId());

    return retVal;
}

/****************************************************************************
 * Function: BENDAPI_sendEvent()                                            *
 * ---------------------------                                              *
 * Parameters: event_type - the event type defined in pdk_callback.h        *
 *             event_data - the struct corresponding to the event           *
 *                                                                          *
 * Description: This fn is used to send async Events to the Control Plane   *
 *                                                                          *
 * Return: FPPI_SUCCESS on success, otherwise error code.                   *
 *                                                                          *
 ****************************************************************************/

FPPI_RET BENDAPI_sendEvent(
        NPF_CBCAT       event_type,
        void*           event_data)
{
    FPPI_RET retVal  = FPPI_SUCCESS;
    void* outObj = NULL;
    FPPI_FEID feid;

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Entered method (thread # %d).\n", PIL_GetCurrentThreadId());

    PIL_EnterCriticalSection(&g_beFlagslock);

    if ((Bend_Initialized == 0) || (Bend_Ready == 0))
    {
        retVal = FPPI_NOT_INITIALIZED;
        PIL_LeaveCriticalSection(&g_beFlagslock);
        goto DRS_LBL_1;
    }

    feid = g_feInfo.feid;

    PIL_LeaveCriticalSection(&g_beFlagslock);

    /* Copy data destined to CP into local DRS memory */
    if ((retVal = CopyAndPackFPEvent(
        event_type, event_data, 1, &outObj)) !=
        FPPI_SUCCESS)
    {
        goto DRS_LBL_1;
    }

    /* Put msg into mbox */
    if ((retVal = StuffMsgIntoFPtoCPQueue(
        drs_FPEvent, event_type, feid, NULL, outObj)) != FPPI_SUCCESS)
    {
        goto DRS_LBL_1;
    }

DRS_LBL_1:

    if (retVal != FPPI_SUCCESS)
    {
        if (outObj != NULL)
        {
            /* Free local mem on error */
            PIL_FreeMemory(outObj);
            outObj = NULL;
        }
    }

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Leaving method (thread # %d).\n", PIL_GetCurrentThreadId());

    return retVal;
}

/****************************************************************************
 * Function: BENDAPI_Report_Status()                                        *
 * ---------------------------                                              *
 * Parameters: status       - the status of the call (success or failure)   *
 *             cbtype       - the callback type for which the fn is         *
 *                            registered.                                   *
 *             cbcorrelator - the unique ID generated by the Plugin which   *
 *                            is passed to the client during callback       *
 *                            invocation.                                   *
 *             response_data - the struct containing any response data      *
 *                                                                          *
 * Description: This fn is used to notify the Control Plane about the       *
 *              status of a call which was sent by the CP (decision msg)    *
 *              to be executed on the Forwarding Plane. The message may     *
 *              contain statistics or other data to be passed to the CP     *
 *              in addition to the status code.                             *
 *                                                                          *
 * Return: FPPI_SUCCESS on success, otherwise error code.                   *
 *                                                                          *
 ****************************************************************************/

FPPI_RET BENDAPI_Report_Status(
    FPPI_Status       status,
    NPF_CBCAT         cbtype,
    FPPI_CBCORRELATOR cbcorrelator,
    void              *response_data)
{
    FPPI_RET retVal  = FPPI_SUCCESS;
    void* outObj     = NULL;
    FPPI_FEID feid;

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Entered method (thread # %d).\n", PIL_GetCurrentThreadId());

    PIL_EnterCriticalSection(&g_beFlagslock);

    if ((Bend_Initialized == 0) || (Bend_Ready == 0))
    {
        retVal = FPPI_NOT_INITIALIZED;
        PIL_LeaveCriticalSection(&g_beFlagslock);
        goto DRS_LBL_1;
    }

    feid = g_feInfo.feid;

    PIL_LeaveCriticalSection(&g_beFlagslock);

    /* Log status from FP */
    if (status == FPPI_SUCCESSFUL)
    {
        npf_logger_write(LOG_TPI, LOG_INFO, "Success report from FP.\n");
    }
    else if (status == FPPI_FAILURE) {
        npf_logger_write(LOG_TPI, LOG_INFO, "Failure report from FP.\n");
    }
    else {
        npf_logger_write(LOG_TPI, LOG_ERR, "Error Unknown Status !! \n");
    }
    /* Copy data destined to CP into local DRS memory */
    if ((retVal = CopyAndPackResponse(
        status, cbtype, response_data, &outObj)) != FPPI_SUCCESS)
    {
        goto DRS_LBL_1;
    }

    /* Put msg into mbox */
    if ((retVal = StuffMsgIntoFPtoCPQueue(
        drs_FPResponse, cbtype, feid,
        (void*)cbcorrelator, outObj)) != FPPI_SUCCESS)
    {
        goto DRS_LBL_1;
    }

DRS_LBL_1:

    if (retVal != FPPI_SUCCESS)
    {
        if (outObj != NULL)
        {
            /* Free local mem on error */
            PIL_FreeMemory(outObj);
            outObj = NULL;
        }
    }

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Leaving method (thread # %d).\n", PIL_GetCurrentThreadId());

    return retVal;
}

/****************************************************************************
 * Function: MakeCBIntoFP()                                                 *
 * ---------------------------                                              *
 * Parameters: cbtype      - enum that defines the CB type                  *
 *             correlator  - used to distinguish multiple invocations       *
 *             outObj      - data to pass to the FP                         *
 *                                                                          *
 * Description: The callback methods that the FP has registered with the    *
 *              CP are invoked in this method. They are invoked             *
 *              sequentially in a loop.                                     *
 *                                                                          *
 *              This method must be compiled under the FP because it        *
 *              contains pdk_callback... calls. This is an artifact of      *
 *              how the callback list is implemented in the colocated       *
 *              PDK. Separate lists are maintained for the CP and the       *
 *              FP, and are identified using #defines during the            *
 *              compilation process (TEAM = CP or FP).                      *
 *                                                                          *
 * Return: FPPI_SUCCESS, FPPI_FAIL                                          *
 *                                                                          *
 ****************************************************************************/

FPPI_RET MakeCBIntoFP(
    LPDRS_GENERAL_EVENT_MSG msgPtr)
{
    FPPI_RET retVal = FPPI_SUCCESS;
    FPPI_CONTEXT context;
    FPPI_CBHANDLE cbid, nextcbid;
    backendapi_callback cb;

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Entered method (thread # %d).\n",
        PIL_GetCurrentThreadId());

    if (msgPtr->cbtype != FPPI_INVALID_CBTYPE)
    {
        /* Invoke all registered CBs sequentially in a loop */
        if ((cb = (backendapi_callback)pdk_callback_firstFunc(
            msgPtr->cbtype, &context, &nextcbid)) == NULL)
        {
            npf_logger_write(LOG_DRS, LOG_INFO,
                "No FP callbacks to be made for cbtype = %d \n",
                msgPtr->cbtype);
            goto DRS_LBL_1;
        }

        do
        {
            npf_logger_write(LOG_DRS, LOG_INFO,
                "Making callback into FP of cbtype %d (thread # %d).\n",
                msgPtr->cbtype, PIL_GetCurrentThreadId());

            cbid = nextcbid;

            /* Call into FP now */
            cb((FPPI_CBCORRELATOR)msgPtr->correlator, context, msgPtr->outObj);

        } while ((cb = (backendapi_callback)pdk_callback_nextFunc(
            cbid, &context, &nextcbid)) != NULL);

    } /* if (cbtype != FPPI_INVALID_CBTYPE) */

DRS_LBL_1:

    /* Passed in memory is freed by caller of this method */

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Leaving method with status %d (thread # %d).\n",
        retVal, PIL_GetCurrentThreadId());

    return retVal;
}


/* ATM */

#ifdef _INCLUDE_ATM
/****************************************************************************
 * Function: BENDAPI_sendATMEvent()                                         *
 * ---------------------------                                              *
 * Parameters: event_type - the event type defined in pdk_callback.h        *
 *             event_data - the struct corresponding to the event           *
 *                                                                          *
 * Description: This fn is used to send async Events to the Control Plane   *
 *                                                                          *
 * Return: FPPI_SUCCESS on success, otherwise error code.                   *
 *                                                                          *
 ****************************************************************************/

FPPI_RET BENDAPI_sendATMEvent(
        NPF_CBCAT       event_type,
        void*           event_data)
{
    FPPI_RET retVal  = FPPI_SUCCESS;
    void* outObj = NULL;
    FPPI_FEID feid;

    PIL_EnterCriticalSection(&g_beFlagslock);

    if ((Bend_Initialized == 0) || (Bend_Ready == 0))
    {
        retVal = FPPI_NOT_INITIALIZED;
        PIL_LeaveCriticalSection(&g_beFlagslock);
        goto DRS_LBL_1;
    }
    feid = g_feInfo.feid;

    PIL_LeaveCriticalSection(&g_beFlagslock);

    /* Copy data destined to CP into local DRS memory */
    if ((retVal = CopyAndPackFPEvent_atm(
        event_type, event_data, 1, &outObj)) !=
        FPPI_SUCCESS)
    {
        goto DRS_LBL_1;
    }
    /* Put msg into mbox */
    if ((retVal = StuffMsgIntoFPtoCPQueue(
        drs_FPEvent, event_type, feid, NULL, outObj)) != FPPI_SUCCESS)
    {
        goto DRS_LBL_1;
    }
DRS_LBL_1:
    if (retVal != FPPI_SUCCESS)
    {
        if (outObj != NULL)
        {
            /* Free local mem on error */
            PIL_FreeMemory(outObj);
            outObj = NULL;
        }
    }
    return retVal;
}

/****************************************************************************
 * Function: BENDAPI_Report_ATMStatus()                                     *
 * ---------------------------                                              *
 * Parameters: status       - the status of the call (success or failure)   *
 *             cbtype       - the callback type for which the fn is         *
 *                            registered.                                   *
 *             cbcorrelator - the unique ID generated by the Plugin which   *
 *                            is passed to the client during callback       *
 *                            invocation.                                   *
 *             response_data - the struct containing any response data      *
 *                                                                          *
 * Description: This fn is used to notify the Control Plane about the       *
 *              status of a call which was sent by the CP (decision msg)    *
 *              to be executed on the Forwarding Plane. The message may     *
 *              contain statistics or other data to be passed to the CP     *
 *              in addition to the status code.                             *
 *                                                                          *
 * Return: FPPI_SUCCESS on success, otherwise error code.                   *
 *                                                                          *
 ****************************************************************************/


FPPI_RET BENDAPI_Report_ATMStatus(
    FPPI_Status       status,
    NPF_CBCAT         cbtype,
    FPPI_CBCORRELATOR cbcorrelator,
    void              *response_data)
{
    FPPI_RET retVal  = FPPI_SUCCESS;
    void* outObj     = NULL;
    FPPI_FEID feid;

    PIL_EnterCriticalSection(&g_beFlagslock);

    if ((Bend_Initialized == 0) || (Bend_Ready == 0))

    {
        retVal = FPPI_NOT_INITIALIZED;
        PIL_LeaveCriticalSection(&g_beFlagslock);
        goto DRS_LBL_1;
    }
    feid = g_feInfo.feid;

    PIL_LeaveCriticalSection(&g_beFlagslock);

    /* Log status from FP */
    if (status == FPPI_SUCCESSFUL)
    {
        npf_logger_write(LOG_TPI, LOG_INFO, "Success report from FP.\n");
    }
    else if (status == FPPI_FAILURE) 
    {
        npf_logger_write(LOG_TPI, LOG_INFO, "Failure report from FP.\n");
    }
    else 
    {
        npf_logger_write(LOG_TPI, LOG_ERR, "Error Unknown Status !! \n");
    }
    /* Copy data destined to CP into local DRS memory */
    if ((retVal = CopyAndPackResponse_atm(
        status, cbtype, response_data, &outObj)) != FPPI_SUCCESS)
    {
        goto DRS_LBL_1;
    }
    /* Put msg into mbox */
    if ((retVal = StuffMsgIntoFPtoCPQueue(
        drs_FPResponse, cbtype, feid,
        (void*)cbcorrelator, outObj)) != FPPI_SUCCESS)
    {
        goto DRS_LBL_1;
    }
DRS_LBL_1:
    if (retVal != FPPI_SUCCESS)
    {
        if (outObj != NULL)
        {
            /* Free local mem on error */
            PIL_FreeMemory(outObj);
            outObj = NULL;
        }
    }
    return retVal;
}

#endif /* _INCLUDE_ATM */


