/****************************************************************************
 *             INTEL CORPORATION PROPRIETARY INFORMATION
 *  This software is supplied under the terms of a license agreement or
 *  nondisclosure agreement with Intel Corporation and may not be copied
 *  or disclosed except in accordance with the terms of that agreement.
 *      Copyright (c) 2001 Intel Corporation. All Rights Reserved.
 *
 * $Source: /cvs/constellations/ixa_3.0/src/cp_pdk/ForwardingPlane/DataRelayShim/src/bepackinfo.c,v $
 * $Revision: 1.6 $
 * $Date: 2003/11/17 16:57:19 $
 *
 * Description:
 *     Helper functions for copying FP data into structures for CP and
 *     putting the structures into the message queue for the CP.
 *
 ****************************************************************************/

#include "string.h"
#include "fpp_common.h"
#include "fppCapabilityTypes.h"
#include "log.h"
#include "msgthreads.h"
#include "bepriv.h"

#ifdef _INCLUDE_MPLS
#include "fppapi_mpls_types.h"
#endif

#ifdef _INCLUDE_ATM
#include "pdk_cm_types.h"
#include "pdk_callback_types.h"
#include "npf_atm_types.h"
#include "fppapi_atm_types.h"
#endif /* _INLCUDE_ATM_ */



/* Must have access to same object that the service thread uses to
   service the queue. This is defined in the Control Plane DRS. */
extern QUEUE_SERVICE_PARAMS g_fpTocpThrdParms;

/****************************************************************************
 * Function: CopyAndPackFPEvent()                                           *
 * ---------------------------                                              *
 * Parameters: cbType  - type of message                                    *
 *             inObj   - ptr to object to copy and pack                     *
 *             count   - number of objects in inObj                         *
 *             outObj  - ptr to new copy of object                          *
 *                                                                          *
 * Description: This method handles event msgs from FP. They are not        *
 *              responses to msgs originally sent from CP.                  *
 *              Data is copied from FP memory into DRS local mem.           *
 *                                                                          *
 * Return: FPPI_SUCCESS on success, otherwise error code.                   *
 *                                                                          *
 ****************************************************************************/

FPPI_RET CopyAndPackFPEvent(
    NPF_CBCAT cbType,
    void*     inObj,
    uint32_t  count,
    void**    outObj)
{
    FPPI_RET retVal = FPPI_SUCCESS;
    uint32_t dataSz;
    uint32_t capsListSz;
    uint32_t i;
    char* tmpPtr;

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Entered method (thread # %d).\n", PIL_GetCurrentThreadId());

    switch (cbType)
    {
    case FPPI_FECapability_Event:
        {
        FPPI_FE_Caps* feCapsOut = NULL;
        FPPI_FE_Caps* feCapsIn = (FPPI_FE_Caps*)inObj;

        capsListSz = sizeof(FPPI_Port_Caps) *
            feCapsIn->ports_caps.port_count;
        dataSz     = capsListSz;
    
        // Allocate memory for FP struct and embedded obj 
        if ((feCapsOut = (FPPI_FE_Caps*)PIL_AllocateMemory(
            sizeof(FPPI_FE_Caps) + dataSz, PIL_MEM_ZERO_INIT)) == NULL)
        {
            retVal = FPPI_OUT_OF_MEMORY;
            goto DRS_LBL_1;
        }

        // Copy data into new local struct 
        memcpy(feCapsOut->fe_name, feCapsIn->fe_name, FENAMELEN);
        feCapsOut->ports_caps.port_count = feCapsIn->ports_caps.port_count;

        /* Set temp ptr into out object just past main struct */
        tmpPtr = (char*)feCapsOut + sizeof(FPPI_FE_Caps);

        if (feCapsIn->ports_caps.port_count > 0)
        {
            /* All the entries of the port caps list must be stored back
               to back in memory for this block copy to be successful. */
            memcpy(tmpPtr, feCapsIn->ports_caps.port_caps_list, capsListSz);
            feCapsOut->ports_caps.port_caps_list = (FPPI_Port_Caps*)tmpPtr;
        }
        else {
            feCapsOut->ports_caps.port_caps_list = NULL;
        }

        /* Set ptr into out object just past caps list entries */
        tmpPtr += capsListSz;
        /* Set out param */ 
        *outObj = feCapsOut;
        }
        break;

    case FPPI_PacketIO_Event:
        {
        uint32_t pktsSz;
        uint32_t pktObjSz;
        char* pktDataPtr;
        npf_Packet_t* pktObjPtr;
        npf_PacketList_t* list;
        npf_Packet_t* inPkts = (npf_Packet_t*)inObj;

        pktObjSz = sizeof(npf_Packet_t);
        pktsSz = dataSz = pktObjSz * count;

        /* Add size of embedded pkt data */
        for (i = 0; i < count; i++)
        {
            dataSz += inPkts[i].length;
        }

        /* Allocate memory for FP struct and embedded objs */
        if ((list = (npf_PacketList_t*)PIL_AllocateMemory(
            sizeof(npf_PacketList_t) + dataSz, PIL_MEM_ZERO_INIT)) == NULL)
        {
            retVal = FPPI_OUT_OF_MEMORY;
            goto DRS_LBL_1;
        }

        /* Copy list object elements */
        list->packet_count = count;
        if (inPkts != NULL)
        {
            list->packet_list = pktObjPtr = (npf_Packet_t*)(list + 1);
        }
        else {
            list->packet_list = NULL;
        }

        /* Init ptr for copying packet data just beyond the packet objects */
        pktDataPtr = ((char*)pktObjPtr) + pktsSz;

        for (i = 0; i < count; i++)
        {
            /* Copy the packet object */
            memcpy(pktObjPtr, &(inPkts[i]), pktObjSz);

            if (inPkts[i].buffer != NULL)
            {
                /* Copy the packet data */
                memcpy(pktDataPtr, inPkts[i].buffer, inPkts[i].length);

                /* Set the new data ptr in this pkt */
                pktObjPtr->buffer = pktDataPtr;
            }
            else {
                pktObjPtr->buffer = NULL;
            }

            /* Increment pkt obj ptr and pkt data ptr for next iteration */
            pktObjPtr++;
            pktDataPtr += inPkts[i].length;
        }

        /* Set out param */
        *outObj = list;
        }
        break;

    case FPPI_LinkUp_Event:
    case FPPI_LinkDown_Event:
        {
        FPPI_PortID* outPort;
        FPPI_PortID* inPort = (FPPI_PortID*)inObj;

        dataSz = sizeof(FPPI_PortID);
    
        // Allocate memory for FP struct and embedded obj 
        if ((outPort = (FPPI_PortID*)PIL_AllocateMemory(
            dataSz, PIL_MEM_ZERO_INIT)) == NULL)
        {
            retVal = FPPI_OUT_OF_MEMORY;
            goto DRS_LBL_1;
        }

        *outPort = *inPort;

        /* Set out param */ 
        *outObj = outPort;
        }
        break;

    case FPPI_Directly_Connected_Host_Event:
        {
        npf_fpm_ipv4_exception_t* outEvent;
        npf_fpm_ipv4_exception_t* inEvent = (npf_fpm_ipv4_exception_t*)inObj;

        dataSz = sizeof(npf_fpm_ipv4_exception_t);
    
        // Allocate memory for FP struct and embedded obj 
        if ((outEvent = (npf_fpm_ipv4_exception_t*)PIL_AllocateMemory( dataSz, PIL_MEM_ZERO_INIT)) == NULL)
        {
            retVal = FPPI_OUT_OF_MEMORY;
            goto DRS_LBL_1;
        }
	memcpy(outEvent, inEvent, dataSz);
        /* Set out param */ 
        *outObj = outEvent;
        }
        break;

#if 0
    case FPPI_ARP_AddEntry_Event:
    case FPPI_ARP_DelEntry_Event:
        {
        npf_ipv4_ARPEventData_t* inArp = (npf_ipv4_ARPEventData_t*)inObj;
        npf_ipv4_ARPEventData_t* outArp;

        dataSz = sizeof(npf_ipv4_ARPEventData_t);
    
        // Allocate memory for FP struct and embedded obj 
        if ((outArp = (npf_ipv4_ARPEventData_t*)PIL_AllocateMemory(
            dataSz, PIL_MEM_ZERO_INIT)) == NULL)
        {
            retVal = FPPI_OUT_OF_MEMORY;
            goto DRS_LBL_1;
        }

        *outArp = *inArp;

        /* Set out param */ 
        *outObj = outArp;
        }
        break;
#endif

    default:
        npf_logger_write(LOG_DRS, LOG_ERR, "Not a valid CB type.\n");
    }

DRS_LBL_1:

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Exiting method with status %d (thread # %d).\n",
         retVal, PIL_GetCurrentThreadId());

    return retVal;
}

/****************************************************************************
 * Function: CopyAndPackResponse()                                          *
 * ---------------------------                                              *
 * Parameters: status  - the status of the call (success or failure)        *
 *             cbType  - type of message                                    *
 *             inObj   - ptr to object to copy and pack                     *
 *             outObj  - ptr to new copy of object                          *
 *                                                                          *
 * Description: This method handles response msgs from FP. They are         *
 *              responses to msgs originally sent from CP. Some of these    *
 *              response msgs contain data (e.g. statistics request) and    *
 *              others contain only a status (success or fail) indicating   *
 *              whether the original msg sent from CP succeeded on the FP.  *
 *              In the latter case the inObj param is NULL.                 *
 *                                                                          *
 *              Data is copied from FP memory into DRS local mem. The data  *
 *              is passed to the CP then freed by the DRS when the method   *
 *              returns.                                                    *
 *                                                                          *
 * Return: FPPI_SUCCESS on success, otherwise error code.                   *
 *                                                                          *
 ****************************************************************************/

FPPI_RET CopyAndPackResponse(
    FPPI_Status  status,
    NPF_CBCAT    cbType,
    void*        inObj,
    void**       outObj)
{
    FPPI_RET retVal = FPPI_SUCCESS;
    uint32_t dataSz;
    uint32_t embedSz;
    FPPI_Status* statPtr;

    /* Set out param */
    *outObj = NULL;

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Entered method (thread # %d).\n", PIL_GetCurrentThreadId());

    switch (cbType)
    {
    case FPPI_IPv4_UniFlushNextHop:
    case FPPI_IPv4_UniFlushPrefix:
         if (inObj == NULL)
         {
            /* No data to copy and pack so break. Out obj is set
               to NULL at top of method. */
            break;
         }
         {
           NPF_IPv4UC_CallbackData_t *inList = 
                              (NPF_IPv4UC_CallbackData_t *)inObj;
           NPF_IPv4UC_CallbackData_t *outStatus;
           dataSz = sizeof(NPF_IPv4UC_CallbackData_t);
           embedSz = sizeof(NPF_IPv4UC_AsyncResponse_t) * inList->numResp;
           /* Allocate memory for FP struct and embedded obj */
           if ((outStatus = (NPF_IPv4UC_CallbackData_t*)PIL_AllocateMemory(
            dataSz + embedSz, PIL_MEM_ZERO_INIT)) == NULL)
           {
              retVal = FPPI_OUT_OF_MEMORY;
              goto DRS_LBL_1;
           }
           outStatus->numResp = inList->numResp;
           outStatus->allOK = inList->allOK; 
           outStatus->type = inList->type;
           if (inList->resp != NULL)
           {
            /* Copy data to out struct */
               outStatus->resp = (NPF_IPv4UC_AsyncResponse_t*)(outStatus + 1);
               memcpy(outStatus->resp, inList->resp, embedSz);
           }
           else {
              outStatus->resp = NULL;
           }

           *outObj = outStatus;
         }
         break;

/* =====>> To be deprecated soon and replaced with those below */
    case FPPI_IPv4_AddRoute:
    case FPPI_IPv4_DelRoute:
    case FPPI_ARP_AddEntry:
    case FPPI_ARP_DelEntry:
/* <<===== */

    case FPPI_IPv4_UniFlushToL2:

#ifdef _INCLUDE_IPV6TM

    case FPPI_IPv6TM_EnableAuto:
    case FPPI_IPv6TM_DisableConfig:
    case FPPI_IPv6TM_DisableAuto:
    case FPPI_IPv6TM_Enable6to4:
    case FPPI_IPv6TM_Disable6to4:
    case FPPI_IPv6TM_EnableConfig:
#endif

#ifdef _INCLUDE_IPV6
    case FPPI_IPv6_FlushNextHop:
    case FPPI_IPv6_FlushPrefix:
#endif

    case FPPI_L2_SetProperties:
    case FPPI_L3_SetProperties:
    case FPPI_L3_DelProperties:
        /* These responses do not return any data more than a status
           variable from FP - success or failure. */

        /* Allocate memory for status object */
        if ((statPtr = (FPPI_Status*)PIL_AllocateMemory(
            sizeof(FPPI_Status), PIL_MEM_ZERO_INIT)) == NULL)
        {
            retVal = FPPI_OUT_OF_MEMORY;
            goto DRS_LBL_1;
        }

        *statPtr = status;

        /* Set out param */ 
        *outObj = statPtr;

        break;

    case FPPI_IPv4_UniAddNextHop:
    case FPPI_IPv4_UniDelNextHop:
    case FPPI_IPv4_UniAddPrefix:
    case FPPI_IPv4_UniUpdatePrefix:
    case FPPI_IPv4_UniDelPrefix:
    case FPPI_IPv4_UniAddToL2:
    case FPPI_IPv4_UniDelToL2:
        if (inObj == NULL)
        {
            /* No data to copy and pack so break. Out obj is set
               to NULL at top of method. */
            break;
        }

        {
        /* FP returns a structure indicating status for each entry */
        npf_IPv4EntryStatusList* inList = (npf_IPv4EntryStatusList*)inObj;
        npf_IPv4EntryStatusList* outStatus;

        dataSz = sizeof(npf_IPv4EntryStatusList);
        embedSz = sizeof(npf_IPv4_Status_t) * inList->numEntries;
    
        /* Allocate memory for FP struct and embedded obj */
        if ((outStatus = (npf_IPv4EntryStatusList*)PIL_AllocateMemory(
            dataSz + embedSz, PIL_MEM_ZERO_INIT)) == NULL)
        {
            retVal = FPPI_OUT_OF_MEMORY;
            goto DRS_LBL_1;
        }

        outStatus->numEntries = inList->numEntries;

        /* Copy data to out struct */
        if (inList->entryList != NULL)
        {
            outStatus->entryList = (npf_IPv4_Status_t*)(outStatus + 1);
            memcpy(outStatus->entryList, inList->entryList, embedSz);
        }
        else {
            outStatus->entryList = NULL;
        }

        /* Set out param */ 
        *outObj = outStatus;

        }
        break;

#ifdef _INCLUDE_IPV6
    case FPPI_IPv6_AddNextHop:
    case FPPI_IPv6_DelNextHop:
    case FPPI_IPv6_AddPrefix:
    case FPPI_IPv6_DelPrefix:
    case FPPI_IPv6_AddToL2:
    case FPPI_IPv6_DelToL2:
        if (inObj == NULL)
        {
            /* No data to copy and pack so break. Out obj is set
               to NULL at top of method. */
            break;
        }

        {
        /* FP returns a structure indicating status for each entry */
        npf_IPv6EntryStatusList* inList = (npf_IPv6EntryStatusList*)inObj;
        npf_IPv6EntryStatusList* outStatus;

        dataSz = sizeof(npf_IPv6EntryStatusList);
        embedSz = sizeof(npf_IPv6_Status_t) * inList->numEntries;

        /* Allocate memory for FP struct and embedded obj */
        if ((outStatus = (npf_IPv6EntryStatusList*)PIL_AllocateMemory(
            dataSz + embedSz, PIL_MEM_ZERO_INIT)) == NULL)
        {
            retVal = FPPI_OUT_OF_MEMORY;
            goto DRS_LBL_1;
        }

        outStatus->numEntries = inList->numEntries;

        if (inList->entryList != NULL)
        {
            /* Copy data to out struct */
            outStatus->entryList = (npf_IPv6_Status_t*)(outStatus + 1);
            memcpy(outStatus->entryList, inList->entryList, embedSz);
        }
        else {
            outStatus->entryList = NULL;
        }

        /* Set out param */
        *outObj = outStatus;

        }
        break;
#endif

    case FPPI_IPv4_UniQueryPrefix:
    case FPPI_IPv4_UniQueryNextHop:
        if (inObj == NULL)
        {
            /* No data to copy and pack so break. Out obj is set
               to NULL at top of method. */
            break;
        }
        {
           NPF_IPv4UC_CallbackData_t *inList = 
                              (NPF_IPv4UC_CallbackData_t *)inObj;
           
           NPF_IPv4UC_CallbackData_t *outStatus;
           dataSz = sizeof(NPF_IPv4UC_CallbackData_t);
           embedSz = sizeof(NPF_IPv4UC_AsyncResponse_t) * inList->numResp;
           /* Allocate memory for FP struct and embedded obj */
           if ((outStatus = (NPF_IPv4UC_CallbackData_t*)PIL_AllocateMemory(
            dataSz + embedSz, PIL_MEM_ZERO_INIT)) == NULL)
           {
              retVal = FPPI_OUT_OF_MEMORY;
              goto DRS_LBL_1;
           }

           outStatus->numResp = inList->numResp;
           outStatus->allOK = inList->allOK; 
           outStatus->type = inList->type;
           if (inList->resp != NULL)
           {
            /* Copy data to out struct */
               outStatus->resp = (NPF_IPv4UC_AsyncResponse_t*)(outStatus + 1);
               memcpy(outStatus->resp, inList->resp, embedSz);
           }
           else {
              outStatus->resp = NULL;
           }

           *outObj = outStatus;
        }
   
        /* Not implemented yet - waiting for CP and FP struct defs */
        break;

#if 0
    case FPPI_ARP_QueryEntry:
        if (inObj == NULL)
        {
            /* No data to copy and pack so break. Out obj is set
               to NULL at top of method. */
            break;
        }

        {
        npf_ipv4_ARPList_t* inArpQ = (npf_ipv4_ARPList_t*)inObj;
        npf_ipv4_ARPList_t* outArpQ;

        dataSz = sizeof(npf_ipv4_ARPList_t);
        embedSz = sizeof(npf_ipv4_ARPEntry_t) * inArpQ->arp_count;
    
        /* Allocate memory for FP struct and embedded obj */
        if ((outArpQ = (npf_ipv4_ARPList_t*)PIL_AllocateMemory(
            dataSz + embedSz, PIL_MEM_ZERO_INIT)) == NULL)
        {
            retVal = FPPI_OUT_OF_MEMORY;
            goto DRS_LBL_1;
        }

        outArpQ->arp_count = inArpQ->arp_count;
        outArpQ->arp_list  = NULL;

        if (inArpQ->arp_list != NULL)
        {
            /* Copy data to out struct */
            outArpQ->arp_list = (npf_ipv4_ARPEntry_t*)(outArpQ + 1);
            memcpy(outArpQ->arp_list, inArpQ->arp_list, embedSz);
        }

        /* Set out param */ 
        *outObj = outArpQ;

        }
        break;
#endif

    case FPPI_L2_GetStatistics:
        if (inObj == NULL)
        {
            /* No data to copy and pack so break. Out obj is set
               to NULL at top of method. */
            break;
        }

        {
        npf_L2StatsList_t* inArp = (npf_L2StatsList_t*)inObj;
        npf_L2StatsList_t* outL2Stats;

        dataSz = sizeof(npf_L2StatsList_t);
        embedSz = sizeof(npf_L2StatsEntry_t) * inArp->port_count;
    
        /* Allocate memory for FP struct and embedded obj */
        if ((outL2Stats = (npf_L2StatsList_t*)PIL_AllocateMemory(
            dataSz + embedSz, PIL_MEM_ZERO_INIT)) == NULL)
        {
            retVal = FPPI_OUT_OF_MEMORY;
            goto DRS_LBL_1;
        }

        outL2Stats->port_count = inArp->port_count;
        outL2Stats->stats_list = NULL;

        if (inArp->stats_list != NULL)
        {
            /* Copy data to out struct */
            outL2Stats->stats_list = (npf_L2StatsEntry_t*)(outL2Stats + 1);
            memcpy(outL2Stats->stats_list, inArp->stats_list, embedSz);
        }

        /* Set out param */ 
        *outObj = outL2Stats;

        }
        break;

    case FPPI_L3_GetStatistics:
        if (inObj == NULL)
        {
            /* No data to copy and pack so break. Out obj is set
               to NULL at top of method. */
            break;
        }

        {
        npf_L3StatsList_t* inL3 = (npf_L3StatsList_t*)inObj;
        npf_L3StatsList_t* outL3Stats;

        dataSz = sizeof(npf_L3StatsList_t);
        embedSz = sizeof(npf_L3StatsEntry_t) * inL3->port_count;
    
        /* Allocate memory for FP struct and embedded obj */
        if ((outL3Stats = (npf_L3StatsList_t*)PIL_AllocateMemory(
            dataSz + embedSz, PIL_MEM_ZERO_INIT)) == NULL)
        {
            retVal = FPPI_OUT_OF_MEMORY;
            goto DRS_LBL_1;
        }

        outL3Stats->port_count = inL3->port_count;
        outL3Stats->stats_list = NULL;

        if (inL3->stats_list != NULL)
        {
            /* Copy data to out struct */
            outL3Stats->stats_list = (npf_L3StatsEntry_t*)(outL3Stats + 1);
            memcpy(outL3Stats->stats_list, inL3->stats_list, embedSz);
        }

        /* Set out param */ 
        *outObj = outL3Stats;

        }
        break;

    case FPPI_ICMP_GetStatistics:
        if (inObj == NULL)
        {
            /* No data to copy and pack so break. Out obj is set
               to NULL at top of method. */
            break;
        }

        {
        npf_ICMPStatsList_t* inICMP = (npf_ICMPStatsList_t*)inObj;
        npf_ICMPStatsList_t* outICMP;

        dataSz = sizeof(npf_ICMPStatsList_t);
        embedSz = sizeof(npf_ICMPStatsEntry_t) * inICMP->port_count;
    
        /* Allocate memory for FP struct and embedded obj */
        if ((outICMP = (npf_ICMPStatsList_t*)PIL_AllocateMemory(
            dataSz + embedSz, PIL_MEM_ZERO_INIT)) == NULL)
        {
            retVal = FPPI_OUT_OF_MEMORY;
            goto DRS_LBL_1;
        }

        outICMP->port_count = inICMP->port_count;
        outICMP->stats_list = NULL;

        if (inICMP->stats_list != NULL)
        {
            /* Copy data to out struct */
            outICMP->stats_list = (npf_ICMPStatsEntry_t*)(outICMP + 1);
            memcpy(outICMP->stats_list, inICMP->stats_list, embedSz);
        }

        /* Set out param */ 
        *outObj = outICMP;

        }
        break;

/* MPLS */
#ifdef _INCLUDE_MPLS

    case FPPI_CAT_MPLS:
        if (inObj == NULL)
        {
            /* No data to copy and pack so break. Out obj is set
               to NULL at top of method. */
            break;
        }

        {
           FPPI_MPLS_Resp_t *inMplsPayload = (FPPI_MPLS_Resp_t *)inObj;
           FPPI_MPLS_Resp_t *outMplsPayload;

           dataSz = sizeof(FPPI_MPLS_Resp_t);
    
           /* Allocate memory for FP struct and embedded obj */
           if ((outMplsPayload = (FPPI_MPLS_Resp_t *)PIL_AllocateMemory(
                                  dataSz, PIL_MEM_ZERO_INIT)) == NULL)
           {
               retVal = FPPI_OUT_OF_MEMORY;
               goto DRS_LBL_1;
           }

           /* Copy data to out struct */
           *outMplsPayload = *inMplsPayload;

           /* Set out param */ 
           *outObj = outMplsPayload;

        }
        break;

#endif /* MPLS */

        /* DS */
#ifdef _INCLUDE_DIFFSERV
        case FPPI_DS_DpeOperation:
        {
	    FPPI_DS_Dpe_t      *in;
	    FPPI_DS_Dpe_t       *dpeInfo;
	    uint32_t            infoSize;
	    infoSize = sizeof(FPPI_DS_Dpe_t);
	    in = (FPPI_DS_Dpe_t *)inObj;
	    dpeInfo = (FPPI_DS_Dpe_t *)PIL_AllocateMemory(infoSize, PIL_MEM_ZERO_INIT);
	    memcpy(dpeInfo, in, sizeof(FPPI_DS_Dpe_t));
	    *outObj = dpeInfo;
        }
        break;
#endif

       default:
          break;
    }

DRS_LBL_1:

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Exiting method with status %d (thread # %d).\n",
         retVal, PIL_GetCurrentThreadId());

    return retVal;
}

/****************************************************************************
 * Function: StuffMsgIntoFPtoCPQueue()                                            *
 * ---------------------------                                              *
 * Parameters: cbType     - type of message                                 *
 *             feid       - unique ID for specific FE                       *
 *             correlator - app method invocation instance id               *
 *             outObj     - ptr to data to stuff into queue                 *
 *                                                                          *
 * Description: Place passed in data into msg structure and then into       *
 *              queue.
 *                                                                          *
 * Return: FPPI_SUCCESS on success, otherwise error code.                   *
 *                                                                          *
 ****************************************************************************/

FPPI_RET StuffMsgIntoFPtoCPQueue(
    drs_FP_Msg_Type  msgType,
    NPF_CBCAT        cbType,
    FPPI_FEID        feid,
    FPPI_CORRELATOR  correlator,
    void*            outObj)
{
    FPPI_RET retVal                = FPPI_SUCCESS;
    LPDRS_GENERAL_EVENT_MSG mspPtr = NULL;
    BOOL isMsgInQueue              = FALSE;

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Entered method with type %d (thread # %d).\n",
        cbType, PIL_GetCurrentThreadId());

    /* Allocate memory for mailbox msg */
    if ((mspPtr = (LPDRS_GENERAL_EVENT_MSG)PIL_AllocateMemory
        (sizeof(DRS_GENERAL_EVENT_MSG), PIL_MEM_ZERO_INIT)) == NULL)
    {
        retVal = FPPI_OUT_OF_MEMORY;
        npf_logger_write(LOG_DRS, LOG_ERR, "Could not allocate memory.\n");
        goto DRS_LBL_1;
    }

    /* Fill msg object with data */
    mspPtr->evType = msgType == drs_FPEvent ? DRS_EVENT_MSG_TO_CP :
        DRS_RESPONSE_MSG_TO_CP;
    mspPtr->cbtype     = cbType;
    mspPtr->feid       = feid;
    mspPtr->correlator = correlator;
    mspPtr->outObj     = outObj;

    /* Lock queue package */
    PIL_EnterCriticalSection(&(g_fpTocpThrdParms.queSyncObj));

    /* Insert data into CP bound msg queue as a msg */
    npf_list_pushBack(&(g_fpTocpThrdParms.msgQueue), (void *)mspPtr);
    isMsgInQueue = TRUE;

    /* Now notify threads that a message is pending in its queue */
    if (PIL_ReleaseSemaphore(g_fpTocpThrdParms.newMsgNotifier) != PIL_OK)
    {
        npf_logger_write(LOG_DRS, LOG_ERR,
            "FPPAPI_Stop - Error setting event object.\n");
    }

    /* Unlock the msg queues so they can be serviced by the threads */
    PIL_LeaveCriticalSection(&(g_fpTocpThrdParms.queSyncObj));

DRS_LBL_1:

    if (retVal != FPPI_SUCCESS)
    {
        if (!isMsgInQueue)
        {
            /* Perform cleanup since msg not put into queue.
               Otherwise leave all data alone since it will be referenced
               when the service thread pops the msg from queue. */

            if (mspPtr != NULL)
            {
                /* Free local mem on error */
                PIL_FreeMemory(mspPtr);
                mspPtr = NULL;
            }
        } /* if (!isMsgInQueue) */
    } /* if (retVal != FPPI_SUCCESS) */

    npf_logger_write(LOG_DRS, LOG_INFO,
        "Exiting method with type %d; status %d (thread # %d).\n",
         cbType, retVal, PIL_GetCurrentThreadId());

    return retVal;
}


/* ATM */
#ifdef _INCLUDE_ATM
/****************************************************************************
 * Function: CopyAndPackFPEvent_atm()                                       *
 * ---------------------------                                              *
 * Parameters: cbType  - type of message                                    *
 *             inObj   - ptr to object to copy and pack                     *
 *             outObj  - ptr to new copy of object                          *
 *                                                                          *
 * Description: This method handles event msgs from FP. They are not        *
 *              responses to msgs originally sent from CP.                  *
 *              Data is copied from FP memory into DRS local mem.           *
 *                                                                          *
 * Return: FPPI_SUCCESS on success, otherwise error code.                   *
 *                                                                          *
 ****************************************************************************/
FPPI_RET CopyAndPackFPEvent_atm(
    NPF_CBCAT cbType,
    void*     inObj,
    uint32_t  count,
    void**    outObj)
{
	switch(cbType)
	{
		case FPPI_ATM_OAM_AisAlarm_Event:
		case FPPI_ATM_OAM_RdiAlarm_Event:
		case FPPI_ATM_OAM_FpmStats_Event:
		case FPPI_ATM_OAM_BrStats_Event:
		case FPPI_ATM_OAM_Lb_Event:
		{
			NPF_ATM_EventData_t *buffer;

			if ((buffer = (NPF_ATM_EventData_t *)
					PIL_AllocateMemory(sizeof(NPF_ATM_EventData_t), 
							PIL_MEM_ZERO_INIT)) == NULL)
			{
					return FPPI_OUT_OF_MEMORY;
			}
			memcpy(buffer, inObj, sizeof(NPF_ATM_EventData_t));
			*outObj = buffer;
		}
		break;

		default:
			npf_logger_write(LOG_DRS, LOG_ERR, "Not a valid CB type.\n");
			break;
		} /* end of cbType */

		return FPPI_SUCCESS;

}





/****************************************************************************
 * Function: CopyAndPackResponse_atm()                                      *
 * ---------------------------                                              *
 * Parameters: status  - the status of the call (success or failure)        *
 *             cbType  - type of message                                    *
 *             inObj   - ptr to object to copy and pack                     *
 *             outObj  - ptr to new copy of object                          *
 *                                                                          *
 * Description: This method handles response msgs from FP. They are         *
 *              responses to msgs originally sent from CP. Some of these    *
 *              response msgs contain data (e.g. statistics request) and    *
 *              others contain only a status (success or fail) indicating   *
 *              whether the original msg sent from CP succeeded on the FP.  *
 *              In the latter case the inObj param is NULL.                 *
 *                                                                          *
 *              Data is copied from FP memory into DRS local mem. The data  *
 *              is passed to the CP then freed by the DRS when the method   *
 *              returns.                                                    *
 *                                                                          *
 * Return: FPPI_SUCCESS on success, otherwise error code.                   *
 *                                                                          *
 ****************************************************************************/

FPPI_RET CopyAndPackResponse_atm(
  FPPI_Status  status,
    NPF_CBCAT    cbType,
    void*        inObj,
    void**       outObj)
{
	FPPI_RET retVal = FPPI_SUCCESS;

	/* Set out param */
	*outObj = NULL;

	switch (cbType)
	{
		case FPPI_ATM_IfAttrSet:        
		case FPPI_ATM_IfDelete:        
		case FPPI_ATM_IfEnable:        
		case FPPI_ATM_IfDisable:        

		case FPPI_ATM_VccSet:        
		case FPPI_ATM_VccBind:        
		case FPPI_ATM_VccDelete:        
		case FPPI_ATM_VccEnable:        
		case FPPI_ATM_VccDisable:
		case FPPI_ATM_VccOperStatusGet:
               
		case FPPI_ATM_AAL2_ChannelSet:        
		case FPPI_ATM_AAL2_ChannelDel:        
		case FPPI_ATM_ConnectionXcSet:        
		case FPPI_ATM_ConnectionXcDel:
		case FPPI_ATM_OAM_CP_Set:
		case FPPI_ATM_OAM_CC_Set:
		case FPPI_ATM_OAM_LB_Set:
		case FPPI_ATM_OAM_Monitor:
		case FPPI_ATM_OAM_AlarmSet:
		case FPPI_ATM_OAM_AlarmClear:
		case FPPI_ATM_OAM_PM_Set:
			{
				IfATM_FpResponse_t *userResponse = NULL;
				IfATM_FpResponse_t *outConn = NULL;
				userResponse = (IfATM_FpResponse_t *)inObj;

				if ((outConn = (IfATM_FpResponse_t *)PIL_AllocateMemory(
									sizeof(IfATM_FpResponse_t),
									PIL_MEM_ZERO_INIT)) == NULL)
				{
					retVal = FPPI_OUT_OF_MEMORY;
					return retVal;
				}
				memcpy(outConn, userResponse, sizeof(IfATM_FpResponse_t));
				*outObj = outConn;
			}
			break;
		case FPPI_ATM_VccStatsGet:
			{
				IfATM_VccStats_t *userResponse = NULL;
				IfATM_VccStats_t *outConn = NULL;
				userResponse = (IfATM_VccStats_t *)inObj;

				if ((outConn = (IfATM_VccStats_t *)PIL_AllocateMemory(
									sizeof(IfATM_VccStats_t),
									PIL_MEM_ZERO_INIT)) == NULL)
				{
					retVal = FPPI_OUT_OF_MEMORY;
					return retVal;
				}
				memcpy(outConn, userResponse, sizeof(IfATM_VccStats_t));
				*outObj = outConn;
			}
			break;
		case FPPI_ATM_IfGenStatsGet:
			{
				IfATM_IfStatistics_t *userResponse = NULL;
				IfATM_IfStatistics_t *outConn = NULL;
				userResponse = (IfATM_IfStatistics_t *)inObj;

				if ((outConn = (IfATM_IfStatistics_t *)PIL_AllocateMemory(
									sizeof(IfATM_IfStatistics_t),
									PIL_MEM_ZERO_INIT)) == NULL)
				{
					retVal = FPPI_OUT_OF_MEMORY;
					return retVal;
				}
				memcpy(outConn, userResponse, sizeof(IfATM_IfStatistics_t));
				*outObj = outConn;
			}
			break;
      default:
	      break;

      }/* end of switch */


	return retVal;
}/* endof of CopyAndPackResponse_atm */

#endif /* _INLCUDE_ATM_ */



