/*****************************************************************************
*                  INTEL CORPORATION PROPRIETARY INFORMATION
*  This software is supplied under the terms of a license agreement or
*  nondisclosure agreement with Intel Corporation and may not be copied
*  or disclosed except in accordance with the terms of that agreement.
*  Copyright (c) 2001 Intel Corporation. All Rights Reserved.
*****************************************************************************/
#ifndef _VXWORKS_
#include <stdint.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include "FPModule.h"
#include "FidMgr.h"
#include "fp_common.h"
#include "ipv4_mgr_impl.h"
#include "pil.h"
#include "log.h"

#define IX_SUCCESS 0
typedef uint32_t	ix_error;
typedef uint32_t	ix_uint32;

typedef struct ix_s_l2_addr
{
	uint8_t etherDstAddr[6]; /* Ethernet address in network byte order */
} ix_l2_addr;


extern router_config g_current_router_config;
extern uint32_t mask_to_prefixLength(unsigned int mask);
static void ipv4_display_cb_type(uint32_t cbType);

static	DList		g_ipv4_pendinglist;
static uint32_t g_mybladeID;
static int bits_to_mask( int prefixLength) ;
static uint32_t convert_to_npf(ix_error error);

/*****************************************************************************
*
*						Callback prototypes
*
*****************************************************************************/

static ix_error ipv4_cb(ix_error arg_Result,void* arg_pContext);

int ipv4_InitFPEngine()
{
	npf_logger_write(LOG_FP_IPV4, LOG_TRACE, "Entering ipv4_InitFPEngine");
    /*initialize a linked list to hold the Fid/Qid*/
	npf_list_init(&g_ipv4_pendinglist, PIL_FreeMemory);
	return 0;
}

int ipv4_DestroyFPEngine()
{
	/*free resources for any pending u_c*/
	ipv4_free_resources();

	/*Destroy List*/
	npf_list_destroy(&g_ipv4_pendinglist);
	return 0;
}

/*****************************************************************************
*
*
*   Function:   ipv4_add_prefix()
*
*   Parameters: 
*               prefixes:
*				t_c:
*               interface: interface number
*				routerType: type of route: 0->IPV4, 1->MPLS
*
*   Description: 
*				This function requests the IXASDK adding prefixes.
*				It provides a callback function to the IXASDK to call back with 
*				the status on adding prefixes.           
*
*   Return: none
*
*****************************************************************************/
void ipv4_add_prefix(npf_ipv4_UnicastPrefixList_t *prefixes, 
					 ipv4_TranslatorContext_t *t_c)
{
	int32_t iRet = 0;
	uint32_t i = 0;
	ipv4_uc_info *uc_info = NULL;
	ipv4_UserContext_t *u_c = NULL;
	uint32_t *reqComp_Indctr = NULL;

	npf_logger_write(LOG_FP_IPV4, LOG_TRACE, "Entering  ipv4_add_prefix");
		
	reqComp_Indctr = (uint32_t *)PIL_AllocateMemory(
		sizeof(uint32_t), PIL_MEM_ZERO_INIT);
	if(!reqComp_Indctr)
	{
		npf_logger_write(LOG_FP_IPV4, LOG_ERR, 
			"Mem allocation for reqComp_Indctr failed");
		iRet = -1;
		goto finish;
	}
	
	/*set the request completion indicator counter*/
	*reqComp_Indctr = prefixes->numEntries;

	for (i = 0; i < prefixes->numEntries; i++)
	{
		u_c = (ipv4_UserContext_t *)PIL_AllocateMemory(
		sizeof(ipv4_UserContext_t), PIL_MEM_ZERO_INIT);
		if(!u_c)
		{
			npf_logger_write(LOG_FP_IPV4, LOG_ERR, "Mem allocation for u_c failed");
			iRet = -1;
			(*reqComp_Indctr)--;
			t_c->status[i].retStatus = NPF_E_UNKNOWN;
			t_c->overall_status = NPF_E_UNKNOWN;
			goto finish;
		}	

		u_c->reqComp_Indctr = reqComp_Indctr;
		u_c->t_c = t_c;
		/*set the index to 'i' aligning with the same offset as the addr
		when the callback comes, we use index i to set the error code
		for the respective down call addr*/
		u_c->index = i;
		
		/*Call into the IXASDK*/		
		
		uc_info = (ipv4_uc_info *)PIL_AllocateMemory(
		sizeof(ipv4_uc_info ), PIL_MEM_ZERO_INIT);
		if(!uc_info)
		{
			npf_logger_write(LOG_FP_IPV4, LOG_ERR, 
				"Mem allocation failed for uc_info");
			iRet = -1;
			(*reqComp_Indctr)--;
			t_c->status[i].retStatus = NPF_E_UNKNOWN;
			t_c->overall_status = NPF_E_UNKNOWN;
			goto finish;
		}
		uc_info->type = FPPI_IPv4_UniAddPrefix;
		uc_info->u_c = (void *)u_c;
		/*insert this u_c info into the u_c pending list*/
		npf_list_pushBack(&g_ipv4_pendinglist, uc_info);
	}

finish:
	if(iRet == -1)
	{
		/*Note that the u_c's and the related resources they point to 
		will be anyhow freed during shut down*/
		if(ipv4_is_Valid(u_c))
			PIL_FreeMemory(u_c);

		if(!reqComp_Indctr || *reqComp_Indctr == 0)
		{
			/*inform the Translator component about failure via the up-call*/
			fpm_ipv4_mgr_upcall(iRet, t_c);
			if(reqComp_Indctr)
				PIL_FreeMemory(reqComp_Indctr);
		}
	}

	npf_logger_write(LOG_FP_IPV4, LOG_TRACE, "Leaving  ipv4_add_prefix");
}


/*****************************************************************************
*
*
*   Function:   ipv4_delete_prefix()
*
*   Parameters: 
*               prefixes:
*				t_c:
*               interface: interface number
*				routerType: type of route: 0->IPV4, 1->MPLS
*
*   Description: 
*				This function requests the IXASDK deleting prefixes.
*				It provides a callback function to the IXASDK to call back with 
*				the status on deleting prefixes.           
*
*   Return: none
*
*****************************************************************************/
void ipv4_delete_prefix(npf_ipv4_UnicastPrefixList_t *prefixes, 
					 ipv4_TranslatorContext_t *t_c)
{
	int32_t iRet = 0;
	uint32_t i = 0;
	ipv4_uc_info *uc_info = NULL;
	ipv4_UserContext_t *u_c = NULL;
	uint32_t *reqComp_Indctr = NULL;

	npf_logger_write(LOG_FP_IPV4, LOG_TRACE, "Entering  ipv4_delete_prefix");	
		
	reqComp_Indctr = (uint32_t *)PIL_AllocateMemory(
		sizeof(uint32_t), PIL_MEM_ZERO_INIT);
	if(!reqComp_Indctr)
	{
		npf_logger_write(LOG_FP_IPV4, LOG_ERR, 
			"Mem allocation for reqComp_Indctr failed");
		iRet = -1;
		(*reqComp_Indctr)--;
		t_c->status[i].retStatus = NPF_E_UNKNOWN;
		t_c->overall_status = NPF_E_UNKNOWN;
		goto finish;
	}
	
	/*set the request completion indicator counter*/
	*reqComp_Indctr = prefixes->numEntries;

	for (i = 0; i < prefixes->numEntries; i++)
	{
		
		u_c = (ipv4_UserContext_t *)PIL_AllocateMemory(
			sizeof(ipv4_UserContext_t), PIL_MEM_ZERO_INIT);
		if(!u_c)
		{
			npf_logger_write(LOG_FP_IPV4, LOG_ERR, "Mem allocation for u_c failed");
			iRet = -1;
			goto finish;
		}

		u_c->reqComp_Indctr = reqComp_Indctr;
		u_c->t_c = t_c;
		/*set the index to 'i' aligning with the same offset as the addr
		when the callback comes, we use index i to set the error code
		for the respective down call addr*/
		u_c->index = i; 
		
		/*Call into the IXASDK*/

		uc_info = (ipv4_uc_info *)PIL_AllocateMemory(
		sizeof(ipv4_uc_info ), PIL_MEM_ZERO_INIT);
		if(!uc_info)
		{
			npf_logger_write(LOG_FP_IPV4, LOG_ERR, 
				"Mem allocation failed for uc_info");
			iRet = -1;
			(*reqComp_Indctr)--;
			t_c->status[i].retStatus = NPF_E_UNKNOWN;
			t_c->overall_status = NPF_E_UNKNOWN;
			goto finish;
		}
		uc_info->type = FPPI_IPv4_UniDelPrefix;
		uc_info->u_c = (void *)u_c;
		/*insert this u_c info into the u_c pending list*/
		npf_list_pushBack(&g_ipv4_pendinglist, uc_info);
	}

finish:
	if(iRet == -1)
	{
		/*Note that the u_c's and the related resources they point to 
		will be anyhow freed during shut down*/
		if(ipv4_is_Valid(u_c))
			PIL_FreeMemory(u_c);

		if(!reqComp_Indctr || *reqComp_Indctr == 0)
		{
			/*inform the Translator component about failure via the up-call*/
			fpm_ipv4_mgr_upcall(iRet, t_c);
			if(reqComp_Indctr)
				PIL_FreeMemory(reqComp_Indctr);
		}
	}

	npf_logger_write(LOG_FP_IPV4, LOG_TRACE, "Leaving  ipv4_delete_prefix");
}

/*****************************************************************************
*
*   Function:ipv4_flush_prefixes()
*
*   Parameters: 
*
*
*   Description: 
*				This function purges the prefixes
*
*   Return: none
*
*
*****************************************************************************/
void ipv4_flush_prefixes(ipv4_TranslatorContext_t* t_c)
{    
	int32_t iRet = 0;
	
	npf_logger_write(LOG_FP_CM, LOG_TRACE, "Entering  ipv4_flush_prifxes");
	
	/*Call into the IXASDK*/
	fpm_ipv4_mgr_upcall(iRet, t_c);
	npf_logger_write(LOG_FP_CM, LOG_TRACE, "Leaving cm_set_L3_attrib");
}


/*****************************************************************************
*
*
*   Function:   ipv4_update_prefix()
*
*   Parameters: 
*               prefixes:
*				t_c:
*               interface: interface number
*				routerType: type of route: 0->IPV4, 1->MPLS
*
*   Description: 
*				This function requests the IXASDK updating prefixes.
*				It provides a callback function to the IXASDK to call back with 
*				the status on updating prefixes.          
*
*   Return: none
*
*****************************************************************************/
void ipv4_update_prefix(npf_ipv4_UnicastPrefixList_t *prefixes, 
					 ipv4_TranslatorContext_t *t_c)
{
	int32_t iRet = 0;
	uint32_t i = 0;
	ipv4_uc_info *uc_info = NULL;
	ipv4_UserContext_t *u_c = NULL;
	uint32_t *reqComp_Indctr = NULL;

	npf_logger_write(LOG_FP_IPV4, LOG_TRACE, "Entering ipv4_update_prefix");
	
	reqComp_Indctr = (uint32_t *)PIL_AllocateMemory(
		sizeof(uint32_t), PIL_MEM_ZERO_INIT);
	if(!reqComp_Indctr)
	{
		npf_logger_write(LOG_FP_IPV4, LOG_ERR, 
			"Mem allocation for reqComp_Indctr failed");
		iRet = -1;
		goto finish;
	}
	
	/*set the request completion indicator counter*/
	*reqComp_Indctr = prefixes->numEntries;

	for (i = 0; i < prefixes->numEntries; i++)
	{
		u_c = (ipv4_UserContext_t *)PIL_AllocateMemory(
		sizeof(ipv4_UserContext_t),PIL_MEM_ZERO_INIT);
		if(!u_c)
		{
			npf_logger_write(LOG_FP_IPV4, LOG_ERR, "Mem allocation for u_c failed");
			iRet = -1;
			(*reqComp_Indctr)--;
			t_c->status[i].retStatus = NPF_E_UNKNOWN;
			t_c->overall_status = NPF_E_UNKNOWN;
			goto finish;
		}	
		u_c->reqComp_Indctr = reqComp_Indctr;
		u_c->t_c = t_c;	
		/*set the index to 'i' aligning with the same offset as the addr
		when the callback comes, we use index i to set the error code
		for the respective down call addr*/
		u_c->index = i; 
		
		/*Call into the IXASDK*/
	
		uc_info = (ipv4_uc_info *)PIL_AllocateMemory(
		sizeof(ipv4_uc_info ), PIL_MEM_ZERO_INIT);
		if(!uc_info)
		{
			npf_logger_write(LOG_FP_IPV4, LOG_ERR, 
				"Mem allocation failed for uc_info");
			iRet = -1;
			(*reqComp_Indctr)--;
			t_c->status[i].retStatus = NPF_E_UNKNOWN;
			t_c->overall_status = NPF_E_UNKNOWN;
			goto finish;
		}
		uc_info->type = FPPI_IPv4_UniUpdatePrefix;
		uc_info->u_c = (void *)u_c;
		/*insert this u_c info into the u_c pending list*/
		npf_list_pushBack(&g_ipv4_pendinglist, uc_info);
	}

finish:
	if(iRet == -1)
	{
		/*Note that the u_c's and the related resources they point to 
		will be anyhow freed during shut down*/
		if(ipv4_is_Valid(u_c))
			PIL_FreeMemory(u_c);

		if(!reqComp_Indctr || *reqComp_Indctr == 0)
		{
			/*inform the Translator component about failure via the up-call*/
			fpm_ipv4_mgr_upcall(iRet, t_c);
			if(reqComp_Indctr)
				PIL_FreeMemory(reqComp_Indctr);
		}
	}

	npf_logger_write(LOG_FP_IPV4, LOG_TRACE, "Leaving ipv4_update_prefix");
}


/*****************************************************************************
*
*
*   Function:   ipv4_add_next_hop()
*
*   Parameters: 
*               nexthops:
*				t_c:
*
*   Description: 
*				This function requests the IXASDK adding nextops.
*				It provides a callback function to the IXASDK to call back with 
*				the status on adding prefixes.           
*
*   Return: none
*
*****************************************************************************/
void ipv4_add_next_hop(FPPI_UnicastNextHopUpdateInfo_t *nexthops, 
					 ipv4_TranslatorContext_t *t_c)
{
	int32_t iRet = 0;
	uint32_t i = 0, j = 0;
	ipv4_uc_info *uc_info = NULL;
	ipv4_UserContext_t *u_c = NULL;
	uint32_t *reqComp_Indctr = NULL;

	npf_logger_write(LOG_FP_IPV4, LOG_TRACE, "Entering ipv4_add_next_hop");
	if(nexthops->numNHEntries > 1 || nexthops->numOldL2IDEntries > 1)
	{
		npf_logger_write(LOG_FP_IPV4, LOG_ERR, 
		"IXASDK30 does not support ECMP numEntries:%d, numOldL2IDEntries:%d",
			nexthops->numNHEntries, nexthops->numOldL2IDEntries);
		iRet = -1;
		goto finish;
	}
	reqComp_Indctr = (uint32_t *)PIL_AllocateMemory(
		sizeof(uint32_t), PIL_MEM_ZERO_INIT);
	if(!reqComp_Indctr)
	{
		npf_logger_write(LOG_FP_IPV4, LOG_ERR, 
			"Mem allocation for reqComp_Indctr failed");
		iRet = -1;
		goto finish;
	}
	
	/*set the request completion indicator counter, each add next hop is 
	completed when it gets a callback from ipv4 cc for adding to NH 
	table, and from  fp cc for adding to L2 table*/
	*reqComp_Indctr = nexthops->numNHEntries;

	for (i = 0; i < nexthops->numNHEntries; i++)
	{
		u_c = (ipv4_UserContext_t *)PIL_AllocateMemory(
		sizeof(ipv4_UserContext_t), PIL_MEM_ZERO_INIT);
		if(!u_c)
		{
			npf_logger_write(LOG_FP_IPV4, LOG_ERR, "Mem allocation for u_c failed");
			iRet = -1;
			(*reqComp_Indctr)--;
			t_c->status[i].retStatus = NPF_E_UNKNOWN;
			t_c->overall_status = NPF_E_UNKNOWN;
			goto finish;
		}	
		u_c->reqComp_Indctr = reqComp_Indctr;
		u_c->t_c = t_c;
		/*set the index to 'i' aligning with the same offset as the addr
		when the callback comes, we use index i to set the error code
		for the respective down call addr*/
		u_c->index = i; 
		
		uc_info = (ipv4_uc_info *)PIL_AllocateMemory(
		sizeof(ipv4_uc_info ), PIL_MEM_ZERO_INIT);
		if(!uc_info)
		{
			npf_logger_write(LOG_FP_IPV4, LOG_ERR, 
				"Mem allocation failed for uc_info");
			iRet = -1;
			(*reqComp_Indctr)--;
			t_c->status[i].retStatus = NPF_E_UNKNOWN;
			t_c->overall_status = NPF_E_UNKNOWN;
			goto finish;
		}
		uc_info->type = FPPI_IPv4_UniAddNextHop;
		uc_info->u_c = (void *)u_c;
		/*insert this u_c info into the u_c pending list*/
		npf_list_pushBack(&g_ipv4_pendinglist, uc_info);		

		/*populate the nh_info struct for IXASDK30
		This information will be populated on all the 
		blades*/		

		/*Call into the IXASDK*/		
		if(nexthops->nhInfo[i].feid == g_mybladeID 
			/*&& g_intf_type == ETHERNET_TYPE*/)
		{
			/*add an L2 entry into the L2 table via FP CC*/
			
		}
		if(nexthops->numOldL2IDEntries > 0)
		{
			/*this is an update to existing nhid*/
			/*add an entry into the NH Table on IPv4 CC*/
							
		}
		else
		{
			/*add an entry into the NH Table on IPv4 CC*/	
		}		
	}	
	if(nexthops->numOldL2IDEntries > 0)
	{
		/*delete the old L2 entries on my blade*/
		for(j = 0; j < nexthops->numOldL2IDEntries; ++j)
		{			
			if(nexthops->oldFEandL2IDs[j].feid == g_mybladeID
				/*&& g_intf_type == ETHERNET_TYPE*/)
			{
				/*delete an L2 entry into the L2 table via FP CC*/
				
			}
		}		
	}

finish:
	if(iRet == -1)
	{
		/*Note that the u_c's and the related resources they point to 
		will be anyhow freed during shut down*/
		if(ipv4_is_Valid(u_c))
			PIL_FreeMemory(u_c);

		if(!reqComp_Indctr || *reqComp_Indctr == 0)
		{
			/*inform the Translator component about failure via the up-call*/
			fpm_ipv4_mgr_upcall(iRet, t_c);
			if(reqComp_Indctr)
				PIL_FreeMemory(reqComp_Indctr);
		}
	}

	npf_logger_write(LOG_FP_IPV4, LOG_TRACE, "Leaving ipv4_add_next_hop");
}
/*****************************************************************************
*
*
*   Function:   ipv4_delete_next_hop()
*
*   Parameters: 
*               prefixes:
*				t_c:
*
*   Description: 
*				This function requests the IXASDK deleting prefixes.
*				It provides a callback function to the IXASDK to call back with 
*				the status on deleting nexthops.           
*
*   Return: none
*
*****************************************************************************/
void ipv4_delete_next_hop(npf_ipv4_UnicastNextHopList_t *nexthops, 
					 ipv4_TranslatorContext_t *t_c)
{
	int32_t iRet = 0;
	uint32_t i = 0;
	ipv4_uc_info *uc_info = NULL;
	ipv4_UserContext_t *u_c = NULL;
	uint32_t *reqComp_Indctr = NULL;

	npf_logger_write(LOG_FP_IPV4, LOG_TRACE, "Entering  ipv4_delete_next_hop");	
	if(nexthops->numEntries > 1)
	{
		npf_logger_write(LOG_FP_IPV4, LOG_ERR, 
		"IXASDK30 does not support ECMP numEntries:%d",nexthops->numEntries);
		iRet = -1;
		goto finish;
	}	
	reqComp_Indctr = (uint32_t *)PIL_AllocateMemory(
		sizeof(uint32_t), PIL_MEM_ZERO_INIT);
	if(!reqComp_Indctr)
	{
		npf_logger_write(LOG_FP_IPV4, LOG_ERR, 
			"Mem allocation for reqComp_Indctr failed");
		iRet = -1;
		goto finish;
	}
	
	/*set the request completion indicator counter*/
	*reqComp_Indctr = nexthops->numEntries;

	for (i = 0; i < nexthops->numEntries; i++)
	{
		
		u_c = (ipv4_UserContext_t *)PIL_AllocateMemory(
			sizeof(ipv4_UserContext_t), PIL_MEM_ZERO_INIT);
		if(!u_c)
		{
			npf_logger_write(LOG_FP_IPV4, LOG_ERR, "Mem allocation for u_c failed");
			iRet = -1;
			(*reqComp_Indctr)--;
			t_c->status[i].retStatus = NPF_E_UNKNOWN;
			t_c->overall_status = NPF_E_UNKNOWN;
			goto finish;
		}

		u_c->reqComp_Indctr = reqComp_Indctr;
		u_c->t_c = t_c;
		/*set the index to 'i' aligning with the same offset as the addr
		when the callback comes, we use index i to set the error code
		for the respective down call addr*/
		u_c->index = i; 
		
		/*Call into the IXASDK*/

		uc_info = (ipv4_uc_info *)PIL_AllocateMemory(
		sizeof(ipv4_uc_info ), PIL_MEM_ZERO_INIT);
		if(!uc_info)
		{
			npf_logger_write(LOG_FP_IPV4, LOG_ERR, 
				"Mem allocation failed for uc_info");
			iRet = -1;
			(*reqComp_Indctr)--;
			t_c->status[i].retStatus = NPF_E_UNKNOWN;
			t_c->overall_status = NPF_E_UNKNOWN;
			goto finish;
		}
		uc_info->type = FPPI_IPv4_UniDelNextHop;
		uc_info->u_c = (void *)u_c;
		/*insert this u_c info into the u_c pending list*/
		npf_list_pushBack(&g_ipv4_pendinglist, uc_info);
	}

finish:
	if(iRet == -1)
	{
		/*Note that the u_c's and the related resources they point to 
		will be anyhow freed during shut down*/
		if(ipv4_is_Valid(u_c))
			PIL_FreeMemory(u_c);

		if(!reqComp_Indctr || *reqComp_Indctr == 0)
		{
			/*inform the Translator component about failure via the up-call*/
			fpm_ipv4_mgr_upcall(iRet, t_c);
			if(reqComp_Indctr)
				PIL_FreeMemory(reqComp_Indctr);
		}
	}

	npf_logger_write(LOG_FP_IPV4, LOG_TRACE, "Leaving  ipv4_delete_next_hop");
}

/*****************************************************************************
*
*   Function:ipv4_flush_next_hops()
*
*   Parameters: 
*
*
*   Description: 
*				This function purges all routes and next hops
*
*   Return: none
*
*
*****************************************************************************/
void ipv4_flush_next_hops(npf_ipv4_UnicastNextHopList_t *nexthops,
                          ipv4_TranslatorContext_t* t_c)
{    
	int32_t iRet = 0;
	
	npf_logger_write(LOG_FP_CM, LOG_TRACE, "Entering  ipv4_flush_next_hops");
	
	/*Call into the IXASDK*/
	
	fpm_ipv4_mgr_upcall(iRet, t_c);
	npf_logger_write(LOG_FP_CM, LOG_TRACE, "Leaving ipv4_flush_next_hops");
}

/*****************************************************************************
*
*
*   Function:   ipv4_add_arp()
*
*   Parameters: 
*               prefixes:
*				t_c:
*
*   Description: 
*				This function requests the IXASDK adding arp eintries.
*				It provides a callback function to the IXASDK to call back with 
*				the status on adding arp entry.           
*
*   Return: none
*
*****************************************************************************/
void ipv4_add_arp(npf_ipv4_UnicastToL2List_t *arpentry, 
					 ipv4_TranslatorContext_t *t_c)
{
	int32_t iRet = 0;
	uint32_t i = 0;
	ipv4_uc_info *uc_info = NULL;
	ipv4_UserContext_t *u_c = NULL;
	uint32_t *reqComp_Indctr = NULL;
	ix_uint32  arg_L2Id;
	ix_uint32  arg_NextHopIP;
	ix_uint32  arg_OutputPort; 
	ix_l2_addr arg_L2Addr;

	npf_logger_write(LOG_FP_IPV4, LOG_TRACE, "Entering ipv4_add_arp");
	
	reqComp_Indctr = (uint32_t *)PIL_AllocateMemory(
		sizeof(uint32_t), PIL_MEM_ZERO_INIT);
	if(!reqComp_Indctr)
	{
		npf_logger_write(LOG_FP_IPV4, LOG_ERR, 
			"Mem allocation for reqComp_Indctr failed");
		iRet = -1;
		goto finish;
	}
	
	/*set the request completion indicator counter*/
	*reqComp_Indctr = arpentry->numEntries;

	for (i = 0; i < arpentry->numEntries; i++)
	{
		u_c = (ipv4_UserContext_t *)PIL_AllocateMemory(
		sizeof(ipv4_UserContext_t),PIL_MEM_ZERO_INIT);
		if(!u_c)
		{
			npf_logger_write(LOG_FP_IPV4, LOG_ERR, "Mem allocation for u_c failed");
			iRet = -1;
			(*reqComp_Indctr)--;
			t_c->status[i].retStatus = NPF_E_UNKNOWN;
			t_c->overall_status = NPF_E_UNKNOWN;
			goto finish;
		}	
		u_c->reqComp_Indctr = reqComp_Indctr;
		u_c->t_c = t_c;		
		/*set the index to 'i' aligning with the same offset as the addr
		when the callback comes, we use index i to set the error code
		for the respective down call addr*/
		u_c->index = i; 

		uc_info = (ipv4_uc_info *)PIL_AllocateMemory(
		sizeof(ipv4_uc_info ), PIL_MEM_ZERO_INIT);
		if(!uc_info)
		{
			npf_logger_write(LOG_FP_IPV4, LOG_ERR, 
				"Mem allocation failed for uc_info");
			iRet = -1;
			(*reqComp_Indctr)--;
			t_c->status[i].retStatus = NPF_E_UNKNOWN;
			t_c->overall_status = NPF_E_UNKNOWN;
			goto finish;
		}
		uc_info->type = FPPI_IPv4_UniAddToL2;
		uc_info->u_c = (void *)u_c;
		/*insert this u_c info into the u_c pending list*/
		npf_list_pushBack(&g_ipv4_pendinglist, uc_info);

		/*Call into the IXASDK*/
		arg_L2Id = arpentry->toL2Info[i].l2id;
		switch(arpentry->toL2Info[i].l2Entry.mediaAddress.type)
		{
		case NPF_MAC_ADDRESS:
			npf_logger_write(LOG_FP_IPV4, LOG_INFO, 
				"MAC Address type");			
			memcpy(arg_L2Addr.etherDstAddr, 
				arpentry->toL2Info[i].l2Entry.mediaAddress.u.MAC_address, 6);
			
			//memcpy(u_c->u.mac, arg_L2Addr.etherDstAddr, 6);

			npf_logger_write(LOG_FP_IPV4, LOG_INFO, 
				"arg_L2Addr: 0x%x::%x::%x::%x::%x::%x", 
				arg_L2Addr.etherDstAddr[0],arg_L2Addr.etherDstAddr[1],
				arg_L2Addr.etherDstAddr[2],arg_L2Addr.etherDstAddr[3],
				arg_L2Addr.etherDstAddr[4],arg_L2Addr.etherDstAddr[5]);

			break;
		case NPF_ATM_VC:
			npf_logger_write(LOG_FP_IPV4, LOG_INFO, 
				"ATM Not currently supported");
			iRet = -1;
			goto finish;
		}
		switch(arpentry->toL2Info[i].l2Entry.interfaceHandle.type)
		{
		case NPF_IF_TYPE_POS:
				npf_logger_write(LOG_FP_IPV4, LOG_INFO, 
				"POS Interface type");
			arg_OutputPort = 
				arpentry->toL2Info[i].l2Entry.interfaceHandle.u.POS_Attr.port;
			npf_logger_write(LOG_FP_IPV4, LOG_INFO, 
				"arg_OutputPort:%d", arg_OutputPort);
			break;
		case NPF_IF_TYPE_IPV4:
			npf_logger_write(LOG_FP_IPV4, LOG_INFO, 
				"NPF_IF_TYPE_IPV4 Interface type");
			break;
		case NPF_IF_TYPE_LAN:
			npf_logger_write(LOG_FP_IPV4, LOG_INFO, 
				"NPF_IF_TYPE_LAN Interface type");
			arg_OutputPort = 
				arpentry->toL2Info[i].l2Entry.interfaceHandle.u.LAN_Attr.port;
			npf_logger_write(LOG_FP_IPV4, LOG_INFO, 
				"arg_OutputPort:%d", arg_OutputPort);
			break;
		case NPF_IF_TYPE_ATM:
			npf_logger_write(LOG_FP_IPV4, LOG_INFO, 
				"ATM interface Not currently supported");
			iRet = -1;
			goto finish;
		}

		arg_NextHopIP = arpentry->toL2Info[i].l2Entry.IP_Address;		
	}

finish:
	if(iRet == -1)
	{
		/*Note that the u_c's and the related resources they point to 
		will be anyhow freed during shut down*/
		if(ipv4_is_Valid(u_c))
			PIL_FreeMemory(u_c);

		if(!reqComp_Indctr || *reqComp_Indctr == 0)
		{
			/*inform the Translator component about failure via the up-call*/
			fpm_ipv4_mgr_upcall(iRet, t_c);
			if(reqComp_Indctr)
				PIL_FreeMemory(reqComp_Indctr);
		}
	}

	npf_logger_write(LOG_FP_IPV4, LOG_TRACE, "Leaving ipv4_add_arp");
}

/*****************************************************************************
*
*
*   Function:   ipv4_delete_arp()
*
*   Parameters: 
*               prefixes:
*				t_c:
*
*   Description: 
*				This function requests the IXASDK deleting arp.
*				It provides a callback function to the IXASDK to call back with 
*				the status on deleting arp.           
*
*   Return: none
*
*****************************************************************************/
void ipv4_delete_arp(npf_ipv4_UnicastToL2List_t *arpentry, 
					 ipv4_TranslatorContext_t *t_c)
{
	int32_t iRet = 0;
	uint32_t i = 0;
	ipv4_uc_info *uc_info = NULL;
	ipv4_UserContext_t *u_c = NULL;
	uint32_t *reqComp_Indctr = NULL;
	ix_uint32  arg_L2Id;
	ix_uint32  arg_NextHopIP;
	ix_uint32  arg_OutputPort; 
	ix_l2_addr arg_L2Addr;

	npf_logger_write(LOG_FP_IPV4, LOG_TRACE, "Entering ipv4_delete_arp");
	
	reqComp_Indctr = (uint32_t *)PIL_AllocateMemory(
		sizeof(uint32_t), PIL_MEM_ZERO_INIT);
	if(!reqComp_Indctr)
	{
		npf_logger_write(LOG_FP_IPV4, LOG_ERR, 
			"Mem allocation for reqComp_Indctr failed");
		iRet = -1;
		goto finish;
	}
	
	/*set the request completion indicator counter*/
	*reqComp_Indctr = arpentry->numEntries;

	for (i = 0; i < arpentry->numEntries; i++)
	{
		u_c = (ipv4_UserContext_t *)PIL_AllocateMemory(
		sizeof(ipv4_UserContext_t),PIL_MEM_ZERO_INIT);
		if(!u_c)
		{
			npf_logger_write(LOG_FP_IPV4, LOG_ERR, "Mem allocation for u_c failed");
			iRet = -1;
			(*reqComp_Indctr)--;
			t_c->status[i].retStatus = NPF_E_UNKNOWN;
			t_c->overall_status = NPF_E_UNKNOWN;
			goto finish;
		}	
		u_c->reqComp_Indctr = reqComp_Indctr;
		u_c->t_c = t_c;		
		/*set the index to 'i' aligning with the same offset as the addr
		when the callback comes, we use index i to set the error code
		for the respective down call addr*/
		u_c->index = i; 

		uc_info = (ipv4_uc_info *)PIL_AllocateMemory(
		sizeof(ipv4_uc_info ), PIL_MEM_ZERO_INIT);
		if(!uc_info)
		{
			npf_logger_write(LOG_FP_IPV4, LOG_ERR, 
				"Mem allocation failed for uc_info");
			iRet = -1;
			(*reqComp_Indctr)--;
			t_c->status[i].retStatus = NPF_E_UNKNOWN;
			t_c->overall_status = NPF_E_UNKNOWN;
			goto finish;
		}
		uc_info->type = FPPI_IPv4_UniDelToL2;
		uc_info->u_c = (void *)u_c;
		/*insert this u_c info into the u_c pending list*/
		npf_list_pushBack(&g_ipv4_pendinglist, uc_info);

		/*Call into the IXASDK*/
		arg_L2Id = arpentry->toL2Info[i].l2id;
		switch(arpentry->toL2Info[i].l2Entry.mediaAddress.type)
		{
		case NPF_MAC_ADDRESS:
			npf_logger_write(LOG_FP_IPV4, LOG_INFO, 
				"MAC Address type");
			memcpy(arg_L2Addr.etherDstAddr, 
				arpentry->toL2Info[i].l2Entry.mediaAddress.u.MAC_address, 6);

			//memcpy(u_c->u.mac, arg_L2Addr.etherDstAddr, 6);

			npf_logger_write(LOG_FP_IPV4, LOG_INFO, 
				"arg_L2Addr: 0x%x::%x::%x::%x::%x::%x", 
				arg_L2Addr.etherDstAddr[0],arg_L2Addr.etherDstAddr[1],
				arg_L2Addr.etherDstAddr[2],arg_L2Addr.etherDstAddr[3],
				arg_L2Addr.etherDstAddr[4],arg_L2Addr.etherDstAddr[5]);
			break;
		case NPF_ATM_VC:
			npf_logger_write(LOG_FP_IPV4, LOG_INFO, 
				"ATM Not currently supported");
			iRet = -1;
			goto finish;
		}
		switch(arpentry->toL2Info[i].l2Entry.interfaceHandle.type)
		{
		case NPF_IF_TYPE_POS:
				npf_logger_write(LOG_FP_IPV4, LOG_INFO, 
				"POS Interface type");
			arg_OutputPort = 
				arpentry->toL2Info[i].l2Entry.interfaceHandle.u.POS_Attr.port;
			npf_logger_write(LOG_FP_IPV4, LOG_INFO, 
				"arg_OutputPort:%d", arg_OutputPort);
			break;
		case NPF_IF_TYPE_IPV4:
			npf_logger_write(LOG_FP_IPV4, LOG_INFO, 
				"NPF_IF_TYPE_IPV4 Interface type");
			break;
		case NPF_IF_TYPE_LAN:
			npf_logger_write(LOG_FP_IPV4, LOG_INFO, 
				"NPF_IF_TYPE_LAN Interface type");
			arg_OutputPort = 
				arpentry->toL2Info[i].l2Entry.interfaceHandle.u.LAN_Attr.port;
			npf_logger_write(LOG_FP_IPV4, LOG_INFO, 
				"arg_OutputPort:%d", arg_OutputPort);
			break;
		case NPF_IF_TYPE_ATM:
			npf_logger_write(LOG_FP_IPV4, LOG_INFO, 
				"ATM interface Not currently supported");
			iRet = -1;
			goto finish;
		}

		arg_NextHopIP = arpentry->toL2Info[i].l2Entry.IP_Address;		
	}

finish:
	if(iRet == -1)
	{
		/*Note that the u_c's and the related resources they point to 
		will be anyhow freed during shut down*/
		if(ipv4_is_Valid(u_c))
			PIL_FreeMemory(u_c);

		if(!reqComp_Indctr || *reqComp_Indctr == 0)
		{
			/*inform the Translator component about failure via the up-call*/
			fpm_ipv4_mgr_upcall(iRet, t_c);
			if(reqComp_Indctr)
				PIL_FreeMemory(reqComp_Indctr);
		}
	}

	npf_logger_write(LOG_FP_IPV4, LOG_TRACE, "Leaving ipv4_delete_arp");
}

/*****************************************************************************
*
*   Function:ipv4_flush_arps()
*
*   Parameters: 
*
*
*   Description: 
*				This function purges all routes and next hops
*
*   Return: none
*
*
*****************************************************************************/
void ipv4_flush_arps(ipv4_TranslatorContext_t* t_c)
{    
	int32_t iRet = 0;
	
	npf_logger_write(LOG_FP_CM, LOG_TRACE, "Entering  ipv4_flush_arps");
	
	/*Call into the IXASDK*/
	  
	fpm_ipv4_mgr_upcall(iRet, t_c);
	npf_logger_write(LOG_FP_CM, LOG_TRACE, "Leaving ipv4_flush_arps");
}

/*****************************************************************************
******************************************************************************
*							Callback Function Implementations
******************************************************************************
*****************************************************************************/


/*****************************************************************************
*
*   Function:   ipv4_cb()
*
*   Parameters: 
*			arg_Result:error code returned from the IXASDK messaging operation.
*			arg_pContext:pointer to user context		
*
*   Description:Callback function that contains the status information 
*				for a specific add/delete prefix/nh/arp. 
*				Once all the required status is available the translator 
*				fpm_ipv4_mgr_upcall() is invoked.
*
*   Return: 0 on success, -1 on failure
*****************************************************************************/
ix_error ipv4_cb(ix_error arg_Result,void* arg_pContext)
{
	int32_t iRet = 0;
	ipv4_UserContext_t * u_c = NULL;
	ipv4_TranslatorContext_t * t_c = NULL;
	
	if(!arg_pContext)
	{
		npf_logger_write(LOG_FP_IPV4, LOG_ERR, "Invalid pContext");
		iRet = -1;
		goto finish;
	}
	/*validate the u_c*/
	if(!ipv4_is_Valid(arg_pContext))
	{
		iRet =  -1;
		npf_logger_write(LOG_FP_IPV4, LOG_ERR, "invalid u_c:0x%x", 
			arg_pContext);
		goto finish;
	}

	u_c = (ipv4_UserContext_t *)arg_pContext;
	t_c = (ipv4_TranslatorContext_t *)u_c->t_c;
	if(!t_c)
	{
		npf_logger_write(LOG_FP_IPV4, LOG_ERR, 
			"t_c is invalid");
		iRet =  -1;
		goto finish;
	}
	
	ipv4_display_cb_type(t_c->cbType);
	npf_logger_write(LOG_FP_IPV4, LOG_INFO, "arg_Result:%d", arg_Result);

	if(*(u_c->reqComp_Indctr) != 0)
	{
		/*decrement the request completion indicator counter*/
		(*(u_c->reqComp_Indctr))--;

		/*set the error code*/		
		t_c->status[u_c->index].retStatus = convert_to_npf(arg_Result);
		if(t_c->status[u_c->index].retStatus != NPF_NO_ERROR)
			t_c->overall_status = NPF_E_UNKNOWN;
	}
	if(*(u_c->reqComp_Indctr) == 0)
	{
		/*at this point we have got all the information for the 
		translator request. We are ready to call into the translator 
		upcall with the received information...also, free the 
		reqComp_Indctr*/	
		if(t_c->overall_status == NPF_NO_ERROR)
			fpm_ipv4_mgr_upcall(0, t_c);
		else
			fpm_ipv4_mgr_upcall(-1, t_c);

		/*free the user context resources for the the related down-call*/
		PIL_FreeMemory(u_c->reqComp_Indctr);
	}

	/*free the u_c for this callback*/
	PIL_FreeMemory(u_c);

finish:/*only if something is seriously wrong. Note that the u_c's and the 
	   related resources they point to will be anyhow freed during shut down*/

	if(iRet == -1)
		fpm_ipv4_mgr_upcall(iRet, t_c);
	return iRet;
}

/*****************************************************************************
*   Function: bits_to_mask()
*
*   Parameters: prefixLength: prefix length
*
*   Description: Gets the mask out of prefix length
*
*   return : Mask formed out of the pefixLength
*****************************************************************************/
int bits_to_mask( int prefixLength) 
{
    int temp = 1;
    int i = 0;

    for ( i = 0; i < prefixLength; i++ ) 
	{
        temp = temp << 1;
        temp = temp | 1;
    }
    temp = temp << ( 32 - prefixLength );

    printf( "0x%x\n", temp );
    return temp;
}

/*****************************************************************************
*   Function: convert_to_npf()
*
*   Parameters: error: ixasdk3.0 error code
*
*   Description: converst the ixa error to npf error
*
*   return : npf error
*****************************************************************************/
uint32_t convert_to_npf(ix_error error)
{
	uint32_t npf_error;
	npf_logger_write(LOG_FP_IPV4, LOG_INFO, "error code: %d", error);
		
	return NPF_NO_ERROR;
}

/*****************************************************************************
*
*   Function:   ipv4_is_Valid()
*
*   Parameters: 
*			u_c: Pointer to u_c
*
*   Description: 
*				Searches for the u_c in the pending u_c list, pops out the
*				entry if the match is found 
*
*   Return: 1 for if match found, 0 if match is not found
*****************************************************************************/
uint32_t ipv4_is_Valid(void *u_c)
{
	ipv4_uc_info	*uc_info = NULL;
	DListItr *itr;
	int found = 0;

	if(!u_c)
	{
		npf_logger_write(LOG_FP_IPV4, LOG_ERR, "Invalid u_c");
		return 0;
	}

	npf_logger_write(LOG_FP_IPV4, LOG_TRACE,
		"Entering cm_is_Valid with u_c:0x%x", u_c);
	
	/*walk through the list	to look for a match*/
	itr = npf_list_itrCreate(&g_ipv4_pendinglist);
	if (npf_list_itrFirst(itr) != -1) 
	{
        do {
				uc_info = (ipv4_uc_info *)npf_list_itrGetData(itr);
				if(uc_info->u_c == u_c)
				{
					found = 1;
					npf_logger_write(LOG_FP_IPV4, LOG_TRACE,
						"Found a match");
					npf_list_itrRemoveFree(itr);
				}
			} while (!found && npf_list_itrNext(itr) != -1);
    }	
	npf_list_itrDelete(itr);	
	npf_logger_write(LOG_FP_IPV4, LOG_TRACE, "Leaving ipv4_is_Valid");
	
	if(found)
		return 1;
	return 0;

}

/*****************************************************************************
*
*   Function:   ipv4_free_resources()
*
*   Parameters: 
*			none
*
*   Description: 
*				Walks thro' the u_c pending list and frees up the resources
				and appropriately calls into the translator upcall
*
*   Return: none
*****************************************************************************/
void ipv4_free_resources()
{
	ipv4_uc_info	*uc_info = NULL;
	DListItr		*itr;
	uint32_t		valid_type = 1;

	npf_logger_write(LOG_FP_IPV4, LOG_TRACE,
		"Entering ipv4_free_resources with");
	
	/*walk through the list	to look for a match*/
	itr = npf_list_itrCreate(&g_ipv4_pendinglist);
	if (npf_list_itrFirst(itr) != -1) 
	{
        do {
				uc_info = (ipv4_uc_info *)npf_list_itrGetData(itr);
				valid_type = 1;

				if(uc_info)
				{
					switch(uc_info->type)
					{
						case FPPI_IPv4_UniAddPrefix:
							npf_logger_write(LOG_FP_IPV4, LOG_INFO,
								"Freeing FPPI_IPv4_UniAddPrefix related resources");
							break;
						case FPPI_IPv4_UniDelPrefix:
							npf_logger_write(LOG_FP_IPV4, LOG_INFO,
							"Freeing FPPI_IPv4_UniDelPrefix related resources");
							break;
						case FPPI_IPv4_UniUpdatePrefix:
							npf_logger_write(LOG_FP_IPV4, LOG_INFO,
							"Freeing FPPI_IPv4_UniUpdatePrefix related resources");
							break;
						case FPPI_IPv4_UniAddNextHop:
							npf_logger_write(LOG_FP_IPV4, LOG_INFO,
							"Freeing FPPI_IPv4_UniAddNextHop related resources");
							break;
						case FPPI_IPv4_UniDelNextHop:
							npf_logger_write(LOG_FP_IPV4, LOG_INFO,
							"Freeing FPPI_IPv4_UniDelNextHop related resources");
							break;
						case FPPI_IPv4_UniAddToL2:
							npf_logger_write(LOG_FP_IPV4, LOG_INFO,
							"Freeing FPPI_IPv4_UniAddToL2 related resources");
							break;
						case FPPI_IPv4_UniDelToL2:
							npf_logger_write(LOG_FP_IPV4, LOG_INFO,
							"Freeing FPPI_IPv4_UniDelToL2 related resources");
							break;
						default:
							npf_logger_write(LOG_FP_IPV4, LOG_ERR,
								"Unknown Type");
							valid_type = 0;
					}
					if(valid_type && uc_info->u_c)
					{
						/*call into the translator*/
						fpm_ipv4_mgr_upcall( -1,
							((ipv4_UserContext_t *)(uc_info->u_c))->t_c
							);
						/*free the related resources*/
						if(((ipv4_UserContext_t *)
							(uc_info->u_c))->reqComp_Indctr)
						{
							PIL_FreeMemory(
								((ipv4_UserContext_t *)
								(uc_info->u_c))->reqComp_Indctr
								);
							((ipv4_UserContext_t *)
								(uc_info->u_c))->reqComp_Indctr = NULL;
						}
						PIL_FreeMemory(uc_info->u_c);							
					}
					npf_list_itrRemoveFree(itr);
				}
			} while (npf_list_itrNext(itr) != -1);
    }	
	npf_list_itrDelete(itr);	
	npf_logger_write(LOG_FP_IPV4, LOG_TRACE, "Leaving ipv4_free_resources");	
}
/*****************************************************************************
*
*   Function:   ipv4_display_cb_type()
*
*   Parameters: 
*			cbType: callback type.
*
*   Description: Dsiplays the the name of the callback type
*
*   Return: 0 on success, -1 on failure
*****************************************************************************/
void ipv4_display_cb_type(uint32_t cbType)
{
	switch(cbType)
	{
	case FPPI_IPv4_UniAddPrefix:
		npf_logger_write(LOG_FP_IPV4, LOG_INFO,
			"This callback is for request to add Prefix");
		break;
	case FPPI_IPv4_UniDelPrefix:
		npf_logger_write(LOG_FP_IPV4, LOG_INFO,
			"This callback is for request to delete Prefix");
		break;
	case FPPI_IPv4_UniUpdatePrefix:
		npf_logger_write(LOG_FP_IPV4, LOG_INFO,
			"This callback is for request to update Prefix");
		break;
	case FPPI_IPv4_UniAddNextHop:
		npf_logger_write(LOG_FP_IPV4, LOG_INFO,
			"This callback is for request to add nexthop");
		break;
	case FPPI_IPv4_UniDelNextHop:
		npf_logger_write(LOG_FP_IPV4, LOG_INFO,
			"This callback is for request to delete nexthop");
		break;
	case FPPI_IPv4_UniAddToL2:
		npf_logger_write(LOG_FP_IPV4, LOG_INFO,
			"This callback is for request to add arp");
		break;
	case FPPI_IPv4_UniDelToL2:
		npf_logger_write(LOG_FP_IPV4, LOG_INFO,
			"This callback is for request to delete arp");
		break;
	default:
		npf_logger_write(LOG_FP_IPV4, LOG_ERR,
			"Invalid callback Type");
	}
}

/*****************************************************************************
*
*   Function: GetICMPStats() 
*
*   Parameters: portid: ID of the Port 
*
*   Description: This function gets the ICMP Statistics for a port 
*             
*
*   Return: icmpstatistics structure  
*           
*****************************************************************************/

icmpstatistics GetICMPStats (uint32_t portid)
{
    long    result;
    long	value = 0;
    icmpstatistics stat;

    PDBG_TRACE ();
    ECK (stub_l3_statsGetStatistics(&l3Iface, &result, &value, portid, 12));
    stat.msgs = value;
    value = 0;
    ECK (stub_l3_statsGetStatistics(&l3Iface, &result, &value, portid, 13));
    stat.errors = value;
    value = 0;
    ECK (stub_l3_statsGetStatistics(&l3Iface, &result, &value, portid, 14));
    stat.dstunreachable = value;
    value = 0;
    ECK (stub_l3_statsGetStatistics(&l3Iface, &result, &value, portid, 15));
    stat.timeexcd = value;
    value = 0;
    ECK (stub_l3_statsGetStatistics(&l3Iface, &result, &value, portid, 16));
    stat.paramprobs = value;
    
    PRETURN (stat);
}

/*****************************************************************************
*   Function: QueryforMACAddr() 
*
*   Parameters: ipaddr: IP Address for which MAC Addr needs to be queried 
*
*   Description: This function queries the L3 Fwdr ACE for the ARP Entry 
*             
*
*   Return: ARPQreturn structure which has the MAC Addr entry and the result
*            result = -1, if call fails, result = 0, if call succeeds
*****************************************************************************/

ARPQreturn QueryforMACAddr (uint32_t ipaddr)
{
    long       result;
    tMac       macaddr;
    ARPQreturn val;

    PDBG_TRACE ();
    //ECK (stub_EthEncap_arpTableLookupAddress (&Arpace, &result, &macaddr, ipaddr));
	//stub_ethencap_lookupEntry(&Arpace, &result, &macaddr, ipaddr);
    //val.value =  result;
    //memcpy (val.hwaddr, macaddr.mac, sizeof(macaddr.mac));

    PRETURN (val);
}

#endif
