/****************************************************************************
 *             INTEL CORPORATION PROPRIETARY INFORMATION
 *  This software is supplied under the terms of a license agreement or
 *  nondisclosure agreement with Intel Corporation and may not be copied
 *  or disclosed except in accordance with the terms of that agreement.
 *      Copyright (c) 2001 Intel Corporation. All Rights Reserved.
 *
 * $Source: /cvs/constellations/ixa_3.0/src/cp_pdk/ForwardingPlane/FPModule/IXASDKStubs/src/ixasdk_stub_engine.c,v $
 * $Revision: 1.6.2.1 $
 * $Date: 2004/03/17 09:52:07 $
 *
 * Description:
 *      Implements method to create event loops in new threads. The thread
 *      method is also contained here. The event threads wait for and
 *      processes messages from the CP and FP and then call the appropriate
 *      methods in the FP or CP.
 *
 ****************************************************************************/

#include <time.h>
#include "list.h"
#include "log.h"

#include "cci/cc.h"
#include "ix_cc.h"
#include "cc/ix_cc_rtmv4.h"
#include "cc/ix_cc_ipv4.h"
#if 0 /* Future Release support */
#include "cc/ix_cc_ipv6.h"
#include "cc/ix_cc_v6v4.h"
#endif
#ifdef _INCLUDE_ATM
#include "cc/ix_cc_atm.h"
#endif



#include "cc/ix_cc_stkdrv_common.h"
#include "cc/ix_cc_stkdrv_pktclass.h"
#include "cc/ix_cc_stkdrv.h"
#ifdef _INCLUDE_DIFFSERV
#include "cc/ix_cc_pos_rx.h"
#endif
#include "cc/ix_cc_fpm_core.h"
#include "cc/ix_cc_eth_rx.h"

#include "ixasdk_stubs_pri.h"
#include "ixasdk_methodmap.h"


/* Global object externed in other files */
QUEUE_SERVICE_PARAMS g_stubsParms;

static BOOL g_isSpecificModuleInit = FALSE;


#define STUBS_THREAD_SHUTDOWN_WAIT_DURATION    15000

/****************************************************************************
 Declare local methods
****************************************************************************/

/* Forward declaration of event loop thread function */
void IXAStubsSrvcThread(
    void* argPtr);

NPF_RET SpecificStubInit(
    QUEUE_SERVICE_PARAMS* srvcParms);

NPF_RET SpecificStubShutdown(
    QUEUE_SERVICE_PARAMS* srvcParms);

NPF_RET MakeCBintoFP(
    LPSTUBS_GENERAL_EVENT_MSG eventObj);


/****************************************************************************
 * Function: ixasdk_stubs_Init()                                            *
 * ---------------------------                                              *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: General init method that the PDK calls at system            *
 *              startup. Calls internal init methods for each stub module.  *
 *              This way, a separate list, thread, etc. can                 *
 *              be initialized for each API module ( e.g. IPv4, IPv6,       *
 *              GTP, etc.) Each module init creates list and event          *
 *              notifier then starts thread.                                *
 *                                                                          *
 * Return: NPF_SUCCESS or error code                                        *
 *                                                                          *
 ****************************************************************************/

NPF_RET ixasdk_stubs_Init()
{
    NPF_RET retVal = NPF_SUCCESS;

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Entering ixasdk_stubs_Init.\n");
    
    /* Call local inits per module if necessary. For example,
       if the IXA SDK implements each core component in its own
       thread, then there should be a list and thread for each
       core component stubbed out. In this case there needs to
       be an init for each, i.e. IPv4StubsInit(), etc. For 1.1,o
       there is only one thread that all the core components
       use so just need one init here. */
    retVal = SpecificStubInit(&g_stubsParms);

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Leaving ixasdk_stubs_Init.\n");
    
    return retVal;
}

/****************************************************************************
 * Function: ixasdk_stubs_Shutdown()                                        *
 * ---------------------------                                              *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: General deinit method that the PDK calls at system          *
 *              shutdown. Calls internal init methods for each stub module. *
 *              This way, a separate list, thread, etc. can                 *
 *              be initialized for each API module ( e.g. IPv4, IPv6,       *
 *              GTP, etc.)                                                  *
 *                                                                          *
 * Return: NPF_SUCCESS or error code                                        *
 *                                                                          *
 ****************************************************************************/

NPF_RET ixasdk_stubs_Shutdown()
{
    NPF_RET retVal = NPF_SUCCESS;

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Entering ixasdk_stubs_Shutdown.\n");
    
    /* Call specific shutdown method for module */
    retVal = SpecificStubShutdown(&g_stubsParms);

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Leaving ixasdk_stubs_Shutdown.\n");
    
    return retVal;
}

/****************************************************************************
 * Function: SpecificStubInit()                                             *
 * ---------------------------                                              *
 * Parameters: srvcParms   -  ptr to structure passed to service thread.    *
 *                                                                          *
 * Description: Init the module structure.                                  *
 *                                                                          *
 * Return: NPF_SUCCESS or error code                                        *
 *                                                                          *
 ****************************************************************************/

NPF_RET SpecificStubInit(
    QUEUE_SERVICE_PARAMS* srvcParms)
{
    NPF_RET retVal = NPF_SUCCESS;
    PilThread threadHndl   = 0;
    PilThreadRoutine fnPtr = (PilThreadRoutine)IXAStubsSrvcThread;

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Entering SpecificStubInit.\n");
    
    if (g_isSpecificModuleInit)
    {
        retVal = NPF_FAIL;
        npf_logger_write(LOG_IXA_STUBS, LOG_ERR,
            "Stub module already initialized.\n");
        goto STUBS_LBL_1;
    }

    /* Clear out params */
    srvcParms->thrdHandle       = NULL;
    srvcParms->newMsgNotifier   = NULL;
    srvcParms->thrdUpNotifier   = NULL;
    srvcParms->retVal           = NPF_SUCCESS;

    /* Init the queue sync object */
    PIL_InitializeCriticalSection(&(srvcParms->queSyncObj));

    /* Init the event queues */
    npf_list_init(&(srvcParms->msgQueue), PIL_FreeMemory);

    /* Create the new event notifier object as a binary semaphore
       that is initially unsignalled (empty). */
    if ((srvcParms->newMsgNotifier = PIL_CreateBinarySemaphore(
        FALSE)) == NULL)
    {
        retVal = NPF_FAIL;
        npf_logger_write(LOG_IXA_STUBS, LOG_ERR,
            "Error creating event obj.\n");
        goto STUBS_LBL_1;
    }

    /* Create the thread up events. These are signaled by the threads
       upon startup to the master thread. The master thread may then
       proceed with system init and startup (FP). */
    if ((srvcParms->thrdUpNotifier = PIL_CreateBinarySemaphore(
        FALSE)) == NULL)
    {
        retVal = NPF_FAIL;
        npf_logger_write(LOG_IXA_STUBS, LOG_ERR,
            "Error creating event obj.\n");
        goto STUBS_LBL_1;
    }

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Creating thread to service IXA stubs invocations from FP.\n");
    
    /* Start the inbox thread */
    if ((threadHndl = PIL_CreateThread(
        fnPtr,
        2048,
        "IXAStubsSrvcThread",
        4,                            /* priority from 1-8 inclusive */
        (void *)srvcParms,
        0)) == NULL)
    {
        npf_logger_write(LOG_IXA_STUBS, LOG_ERR,
            "Error starting thread.\n");
        retVal = NPF_FAIL;
        goto STUBS_LBL_1;
    }
    else {

        PIL_EnterCriticalSection(&(srvcParms->queSyncObj));

        srvcParms->thrdHandle = threadHndl;

        PIL_LeaveCriticalSection(&(srvcParms->queSyncObj));

        npf_logger_write(LOG_IXA_STUBS, LOG_INFO,
            "Create IXA stubs msg thread invocation completed successfully.\n");
    }

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Now wait for msg thread to set event indicating that it is about.\n
         to enter wait loop\n.");

    /* Stay here until we receive a signal that the thread has started */
    if (PIL_AcquireSemaphore(srvcParms->thrdUpNotifier,
        PIL_INFINITE_TIMEOUT) != PIL_OK)
    {
        retVal = NPF_FAIL;
        npf_logger_write(LOG_IXA_STUBS, LOG_ERR,
            "Error waiting for event: err=%d\n", retVal);
        goto STUBS_LBL_1;
    }

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "IXA stubs msg thread is now completely initialized and waiting.\n");

    /* Set flag that this module is initialized */
    g_isSpecificModuleInit = TRUE;

STUBS_LBL_1:

    if (retVal != NPF_SUCCESS)
    {
        SpecificStubShutdown(srvcParms);
    }

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Leaving SpecificStubInit.\n");
    
    return retVal;
}

/****************************************************************************
 * Function: SpecificStubShutdown()                                         *
 * ---------------------------                                              *
 * Parameters: srvcParms   -  ptr to structure passed to service thread.    *
 *                                                                          *
 * Description: Shutdown the module structure.                              *
 *                                                                          *
 * Return: NPF_SUCCESS or error code                                        *
 *                                                                          *
 ****************************************************************************/

NPF_RET SpecificStubShutdown(
    QUEUE_SERVICE_PARAMS* srvcParms)
{
    NPF_RET retVal = NPF_SUCCESS;
    LPSTUBS_PARAM_TYPE shtdnParms = NULL;
    uint32_t timeOut = STUBS_THREAD_SHUTDOWN_WAIT_DURATION;
    PilThread thrdHndl = NULL;
    BOOL isObjLocked = FALSE;

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Entered SpecificStubShutdown.\n");
    
    if (!g_isSpecificModuleInit)
    {
        retVal = NPF_FAIL;
        npf_logger_write(LOG_IXA_STUBS, LOG_ERR,
            "Stub module not initialized.\n");
        goto STUBS_LBL_1;
    }

    /* Lock the sync object (if it exists) while destroying contents */
    if (srvcParms->queSyncObj != NULL)
    {
        PIL_EnterCriticalSection(&(srvcParms->queSyncObj));
        isObjLocked = TRUE;
    }

    /* Don't do anything if thread is not started */
    if (srvcParms->thrdHandle == NULL)
    {
        npf_logger_write(LOG_IXA_STUBS, LOG_INFO,
            "Bypassing STUBS service thread shutdown - not started.\n");
        goto STUBS_LBL_1;
    }

    /* Allocate memory for param */
    shtdnParms = (LPSTUBS_PARAM_TYPE)PIL_AllocateMemory
        (sizeof(STUBS_PARAM_TYPE), PIL_MEM_ZERO_INIT);
    
    if (shtdnParms == NULL)
    {
        retVal = NPF_OUT_OF_MEMORY;
        npf_logger_write(LOG_IXA_STUBS, LOG_ERR,
            "Could not allocate memory.\n");
        goto STUBS_LBL_1;
    }

    /* Set the event type */
    shtdnParms->evType = STUBS_EVENT_SHUTDOWN;

    /* Insert shutdown msg into inbox msg queue */
    npf_list_pushBack(&(srvcParms->msgQueue), (void *)shtdnParms);
    
    /* Now notify threads that a message is pending in its queue */
    if (PIL_ReleaseSemaphore(srvcParms->newMsgNotifier) != PIL_OK)
    {
        npf_logger_write(LOG_IXA_STUBS, LOG_ERR,
            "Error setting event object.\n");
        retVal = NPF_FAIL;
        goto STUBS_LBL_1;
    }

    /* Set thread handles in local vars while lock is on */
    thrdHndl = srvcParms->thrdHandle;
    
    /* Unlock the msg queues so they can be serviced by the threads */
    if (isObjLocked)
    {
        PIL_LeaveCriticalSection(&(srvcParms->queSyncObj));
    }
    
    /* Wait for threads to join (expire). If they do not quit
       within the specified timeout, proceed with ungraceful thread
       termination. */
    if (PIL_WaitForThread(thrdHndl, timeOut) != PIL_OK)
    {
        npf_logger_write(LOG_IXA_STUBS, LOG_ERR,
            "Thread wait call failed - now call thread terminate.\n");

        /* Terminate thread ungracefully since can't do it gracefully */
        PIL_TerminateThread(thrdHndl);

        npf_logger_write(LOG_IXA_STUBS, LOG_ERR,
            "Called Terminate Thread on STUBS msg thread.\n");
        retVal = NPF_FAIL;
    }

    /* Lock the sync object again while destroying more contents */
    if (srvcParms->queSyncObj != NULL && !isObjLocked)
    {
        PIL_EnterCriticalSection(&(srvcParms->queSyncObj));
    }

    /* Cleanup thread resources */
    PIL_DestroyThread(thrdHndl, NULL);

    /* Destroy event queue and all remaining events in queue */
    npf_list_destroy(&(srvcParms->msgQueue));

    /* Destroy the event notifier object */
    if (srvcParms->newMsgNotifier != NULL &&
        PIL_DestroySemaphore(srvcParms->newMsgNotifier) != PIL_OK)
    {
        npf_logger_write(LOG_IXA_STUBS, LOG_ERR,
            "Error destroying event object.\n");
        retVal = NPF_FAIL;
    }

    /* Destroy the thread up event notifier object */
    if (srvcParms->thrdUpNotifier != NULL &&
        PIL_DestroySemaphore(srvcParms->thrdUpNotifier) != PIL_OK)
    {
        npf_logger_write(LOG_IXA_STUBS, LOG_ERR,
            "Error destroying event object.\n");
        retVal = NPF_FAIL;
    }

    /* Unlock the sync object */
    if (isObjLocked)
    {
        PIL_LeaveCriticalSection(&(srvcParms->queSyncObj));
    }

    /* Destroy the queue sync object */
    PIL_DestroyCriticalSection(&(srvcParms->queSyncObj));

STUBS_LBL_1:

    if (srvcParms->queSyncObj != NULL && isObjLocked)
    {
        PIL_LeaveCriticalSection(&(srvcParms->queSyncObj));
    }

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Leaving SpecificStubShutdown.\n");
    
    return retVal;
}

/****************************************************************************
 * Function: IXAStubsSrvcThread()                                              *
 * ---------------------------                                              *
 * Parameters: argPtr - ptr to struct containing necessary data             *
 *                                                                          *
 * Description: This thread method waits on an event to fire, then          *
 *              services each event sequentially from a queue until queue   *
 *              is empty. Events are popped first in first out.             *
 *                                                                          *
 * Return: NPF_SUCCESS or an error code on error.                           *
 *                                                                          *
 ****************************************************************************/

void IXAStubsSrvcThread(
    void* argPtr)
{
    NPF_RET retVal                     = NPF_SUCCESS;
    BOOL loopContin                    = TRUE;
    LPQUEUE_SERVICE_PARAMS thrdParams  = (LPQUEUE_SERVICE_PARAMS)argPtr;
    LPSTUBS_GENERAL_EVENT_MSG eventObj = NULL;
    BOOL isObjLocked                   = FALSE;

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Entered method.\n");

    PIL_EnterCriticalSection(&(thrdParams->queSyncObj));
    isObjLocked = TRUE;

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Set event to notify caller IXA stub thread started.\n");

    /* Notify the master thread that this thread has started and
       entered the wait cycle. */
    if (PIL_ReleaseSemaphore(thrdParams->thrdUpNotifier) != PIL_OK)
    {
        npf_logger_write(LOG_IXA_STUBS, LOG_ERR,
            "Error setting event object.\n");
        retVal = NPF_FAIL;
        goto STUBS_LBL_1;
    }

    PIL_LeaveCriticalSection(&(thrdParams->queSyncObj));
    isObjLocked = FALSE;

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Enter loop and wait for IXA stub invocations from FP.\n");

    /* Enter queue servicing loop and remain here for life of program */
    while (loopContin)
    {
        npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
            "Waiting for event.\n");

        /* Wait for an event to arrive */
        if (PIL_AcquireSemaphore(
            thrdParams->newMsgNotifier,
            PIL_INFINITE_TIMEOUT)
            != PIL_OK)
        {
            npf_logger_write(LOG_IXA_STUBS, LOG_ERR,
                "Error waiting for event.\n");

            /* Exit loop */
            break;
        }

        npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
            "Event arrived, wait ended (thread # %d).\n",
             PIL_GetCurrentThreadId());

        /* Lock the event queue to access events. May want to eventually
           copy event from queue into local mem so can unlock queue right
           away instead of holding for entire callback into FP/CP. */
        PIL_EnterCriticalSection(&(thrdParams->queSyncObj));
        isObjLocked = TRUE;

        npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
            "Pop event from back of queue (thread # %d).\n",
             PIL_GetCurrentThreadId());

        /* Event has arrived. Enter loop to service all events
           in queue until queue is empty. */
        while (npf_list_popFront(&(thrdParams->msgQueue),
            (void **)(&eventObj)) != -1)
        {
            if (isObjLocked)
            {
                /* Leave CS now to free access. List entry has been popped. */
                PIL_LeaveCriticalSection(&(thrdParams->queSyncObj));
                isObjLocked = FALSE;
            }

            npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
                "Popped list entry; type: %d (thread # %d).\n",
                eventObj->evType, PIL_GetCurrentThreadId());

            /* Act on the event by calling appropriate handler method */
            if (eventObj->evType == STUBS_EVENT_SHUTDOWN)
            {
                npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
                    "Shutdown event received in IXA stubs thread.\n");

                loopContin = FALSE;
            }
            else {
                /* Call appropriate callback method into CP */
                if ((retVal = MakeCBintoFP(eventObj)) != NPF_SUCCESS)
                {
                    /* Don't do anything at this point */
                }
            }

            /* Free msg object */
            PIL_FreeMemory(eventObj);

            if (retVal != NPF_SUCCESS)
            {
                /* This is where execution may break out of the while
                   loop and shutdown the thread if an error is serious
                   enough. At the time of this writing, an error would
                   occur while making a callback into the FP or CP.
                   In these cases, the queue service thread should
                   normally not break, but instead log an error message
                   and continue servicing the queue. */
            }

            if (loopContin == FALSE)
            {
                /* Break out of local while loop even if there are
                   more msgs in queue. Quit now! */
                break;
            }

        } /* while npf_list_popBack(&(thrdParams->msgQueue), ... */

        npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
            "No more events in queue (thread # %d).\n",
            PIL_GetCurrentThreadId());

        /* Set the return value in case another thread cares */
        thrdParams->retVal = retVal;

        /* Unlock the event queue */
        if (isObjLocked)
        {
            PIL_LeaveCriticalSection(&(thrdParams->queSyncObj));
            isObjLocked = FALSE;
        }

    } /* while (loopContin) */

STUBS_LBL_1:

    /* Unlock the event queue */
    if (isObjLocked)
    {
        PIL_LeaveCriticalSection(&(thrdParams->queSyncObj));
        isObjLocked = FALSE;
    }

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Exiting IXAStubsSrvcThread with status %d.\n", retVal);

    return;
}

/****************************************************************************
 * Function: MakeCBintoFP()                                                 *
 * ---------------------------                                              *
 * Parameters: eventObj  - ptr to struct containing necessary data          *
 *                                                                          *
 * Description: Invoke the callback method after casting to method specific *
 *              type.
 *                                                                          *
 * Return: NPF_SUCCESS or an error code on error.                           *
 *                                                                          *
 ****************************************************************************/

NPF_RET MakeCBintoFP(
    LPSTUBS_GENERAL_EVENT_MSG eventObj)
{
    NPF_RET retVal  = NPF_SUCCESS;

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Entered method.\n");

    switch (eventObj->evType)
    {
    case STUBS_IPv4_ROUTE_OP:
        npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
            "Calling STUBS_IPv4_ROUTE_OP callback into FP.\n");

        ((ix_cc_ipv4_cb_route_op)(eventObj->cbMethod))(
            eventObj->ixReturn, eventObj->context);
        break;

    case STUBS_IPv4_ROUTE_DUMP:
        {
        ix_cc_ipv4_dump_data* parms;

        npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
            "Calling STUBS_IPv4_ROUTE_DUMP callback into FP.\n");

        /* Allocate memory for param */
        parms = (ix_cc_ipv4_dump_data*)PIL_AllocateMemory
            (sizeof(ix_cc_ipv4_dump_data), PIL_MEM_ZERO_INIT);

        if (parms == NULL)
        {
            retVal = NPF_OUT_OF_MEMORY;
            npf_logger_write(LOG_IXA_STUBS, LOG_ERR,
                "Could not allocate memory.\n");
            goto STUBS_LBL_1;
        }

        ((ix_cc_ipv4_cb_dump_data)(eventObj->cbMethod))(
            eventObj->ixReturn, eventObj->context, parms);

        /* Now free local memory */
        PIL_FreeMemory(parms);
        }

        break;

    case STUBS_IPv4_NEXT_HOP:
        {
        ix_cc_rtmv4_next_hop_info* parms;

        npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
            "Calling STUBS_IPv4_NEXT_HOP callback into FP.\n");

        /* Allocate memory for param */
        parms = (ix_cc_rtmv4_next_hop_info*)PIL_AllocateMemory
            (sizeof(ix_cc_rtmv4_next_hop_info), PIL_MEM_ZERO_INIT);

        if (parms == NULL)
        {
            retVal = NPF_OUT_OF_MEMORY;
            npf_logger_write(LOG_IXA_STUBS, LOG_ERR,
                "Could not allocate memory.\n");
            goto STUBS_LBL_1;
        }

        ((ix_cc_ipv4_cb_get_next_hop)(eventObj->cbMethod))(
            eventObj->ixReturn, eventObj->context, parms);

        /* Now free local memory */
        PIL_FreeMemory(parms);
        }

        break;

    case STUBS_IPv4_SLEEP_TIME:
        {
        ix_uint16 parms = 1;

        npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
            "Calling STUBS_IPv4_SLEEP_TIME callback into FP.\n");

        ((ix_cc_ipv4_cb_get_sleep_time)(eventObj->cbMethod))(
            eventObj->ixReturn, eventObj->context, parms);

        }
        break;

    case STUBS_IPv4_QUEUE_DEPTH:
        {
        ix_uint16 parms = 1;

        npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
            "Calling STUBS_IPv4_QUEUE_DEPTH callback into FP.\n");

        ((ix_cc_ipv4_cb_get_queue_depth)(eventObj->cbMethod))(
            eventObj->ixReturn, eventObj->context, parms);

        }
        break;

    case STUBS_STKDRV_NUM_PORTS:
        {
        ix_uint32 parms = 2;

        npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
            "Calling STUBS_STKDRV_NUM_PORTS callback into FP.\n");

        ((ix_cc_stkdrv_cb_get_num_ports)(eventObj->cbMethod))(
            eventObj->ixReturn, eventObj->context, &parms);

        }
        break;

    case STUBS_STKDRV_GET_PROPERTIES:
        {
        ix_cc_properties* parms;
	uint32_t num_ports = 2;

        npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
            "Calling STUBS_STKDRV_GET_PROPERTIES callback into FP.\n");

        /* Allocate memory for param */
        parms = (ix_cc_properties*)PIL_AllocateMemory
            (num_ports * sizeof(ix_cc_properties), PIL_MEM_ZERO_INIT);
#ifdef SQESTUBS
        parms->media_type=1;
#endif /* SQESTUBS */

        if (parms == NULL)
        {
            retVal = NPF_OUT_OF_MEMORY;
            npf_logger_write(LOG_IXA_STUBS, LOG_ERR,
                "Could not allocate memory.\n");
            goto STUBS_LBL_1;
        }
	parms[0].blade_id = 1;
	parms[0].port_id = 0;
	parms[0].media_type = 0;
	parms[0].link_speed = 1000;
	parms[0].mtu = 9216;
	parms[1].blade_id = 1;
	parms[1].port_id = 1;
	parms[1].media_type = 0;
	parms[1].link_speed = 1000;
	parms[1].mtu = 9216;
        ((ix_cc_stkdrv_cb_get_property)(eventObj->cbMethod))(
            eventObj->ixReturn, eventObj->context, parms);

        /* Now free local memory */
        PIL_FreeMemory(parms);
        }

        break;

    case STUBS_ARP_OP:
        npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
            "Calling STUBS_ARP_OP callback into FP.\n");

        ((ix_cc_fpm_arp_op_cb)(eventObj->cbMethod))(
            eventObj->ixReturn, eventObj->context);
        break;
#ifdef _INCLUDE_ATM
    case STUBS_L2_ATM_POS_TX_STATS:
        {
        ix_uint64 parms = 1;

        npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
            "Calling STUBS_L2_ATM_POS_STATS callback into FP.\n");

        ((ix_cc_fpm_atm_pos_tx_stats_op_cb)(eventObj->cbMethod))(
            eventObj->ixReturn, eventObj->context, &parms);
        }
        break;
#endif /* _INCLDUE_ATM*/
    case STUBS_L2_ETH_TX_STATS:
        {
        ix_uint64 parms = 1;

        npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
            "Calling STUBS_L2_ETH_STATS callback into FP.\n");

        ((ix_cc_fpm_eth_tx_stats_op_cb)(eventObj->cbMethod))(
            eventObj->ixReturn, eventObj->context, &parms);
        }
        break;

#ifdef _INCLUDE_DIFFSERV
    case STUBS_POS_RX_STATS:
        {
        ix_uint64 parms = 1;

        npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
            "Calling STUBS_L2_ATM_POS_STATS callback into FP.\n");

        ((ix_cc_pos_rx_cb_get_statistics_info)(eventObj->cbMethod))(
            eventObj->ixReturn, eventObj->context, &parms);

        }
        break;
#endif
    case STUBS_ETH_RX_STATS:
        {
        ix_uint64 parms = 1;
	ix_uint32 len = 1;

        npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
            "Calling STUBS_ETH_RX_STATS callback into FP.\n");

        ((ix_cc_eth_rx_cb_get_statistics_info)(eventObj->cbMethod))(
            eventObj->ixReturn, eventObj->context, &parms, len);

        }
        break;

#if 0 /* Future Release support */
    case STUBS_IPv6_ROUTE_OP:
        npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
            "Calling STUBS_IPv6_ROUTE_OP callback into FP.\n");

        ((ix_cc_ipv6_cb_route_op)(eventObj->cbMethod))(
            eventObj->ixReturn, eventObj->context);
        break;

    case STUBS_IPV6TM_ADD_OP:
        npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
            "Calling STUBS_IPV6TM_ADD_OP callback into FP.\n");

        ((ix_cc_v6v4_cb_add_tunnel)(eventObj->cbMethod))(
            eventObj->ixReturn, eventObj->context, 5);
        break;

    case STUBS_IPV6TM_DEL_OP:
        npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
            "Calling STUBS_IPV6TM_DEL_OP callback into FP.\n");

        ((ix_cc_v6v4_cb)(eventObj->cbMethod))(
            eventObj->ixReturn, eventObj->context);
        break;

#endif /* Future Release support */

#ifdef _INCLUDE_ATM
    case STUBS_ATM_GET_LINK_ATTRIBUTE:
        {
            ix_atm_port_attribs *parms;


            /* Allocate memory for param */
            parms = (ix_atm_port_attribs*)PIL_AllocateMemory
                (sizeof(ix_atm_port_attribs), PIL_MEM_ZERO_INIT);
    
            npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
                "Calling STUBS_ATM_GET_LINK_ATTRIBUTE callback into FP.\n");
    
            if (parms == NULL)
            {
                retVal = NPF_OUT_OF_MEMORY;
                npf_logger_write(LOG_IXA_STUBS, LOG_ERR,
                    "Could not allocate memory.\n");
                goto STUBS_LBL_1;
            }
    
            ((ix_cc_atm_link_cb)(eventObj->cbMethod))(
                eventObj->ixReturn, eventObj->context,parms);

            /* Now free local memory */
            PIL_FreeMemory(parms);
            break;
        }

        case STUBS_ATM_SET_LINK_ATTRIBUTE:
        {
            ix_atm_port_attribs *parms;


            /* Allocate memory for param */
            parms = (ix_atm_port_attribs*)PIL_AllocateMemory
                (sizeof(ix_atm_port_attribs), PIL_MEM_ZERO_INIT);
   
            npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
                "Calling STUBS_ATM_SET_LINK_ATTRIBUTE callback into FP.\n");
   
            if (parms == NULL)
            {
                retVal = NPF_OUT_OF_MEMORY;
                npf_logger_write(LOG_IXA_STUBS, LOG_ERR,
                    "Could not allocate memory.\n");
                goto STUBS_LBL_1;
            }
   
            ((ix_cc_atm_link_cb)(eventObj->cbMethod))(
                eventObj->ixReturn, eventObj->context,parms);

             /* Now free local memory */
             PIL_FreeMemory(parms);

            break;
        }


        case STUBS_ATM_GET_CONN_STATISTICS:
        {
            ix_atm_statistics *parms;


            /* Allocate memory for param */
            parms = (ix_atm_statistics*)PIL_AllocateMemory
                (sizeof(ix_atm_statistics), PIL_MEM_ZERO_INIT);
  
            npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
                "Calling STUBS_ATM_GET_CONN_STATISTICS callback into FP.\n");

            if (parms == NULL)
            {
                retVal = NPF_OUT_OF_MEMORY;
                npf_logger_write(LOG_IXA_STUBS, LOG_ERR,
                    "Could not allocate memory.\n");
                goto STUBS_LBL_1;
            }
  
            ((ix_cc_atm_conn_stats_cb)(eventObj->cbMethod))(
                eventObj->ixReturn, eventObj->context,parms);

             /* Now free local memory */
             PIL_FreeMemory(parms);

            break;
        }

        case STUBS_ATM_SET_CONN_STATISTICS_IND:
        {
            npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
             "Calling STUBS_ATM_SET_CONN_STATISTICS_IND callback into FP.\n");
            ((ix_cc_atm_cb)(eventObj->cbMethod))(
                eventObj->ixReturn, eventObj->context);
            break;
        }

        case STUBS_ATM_SET_CONN_TRAFFIC_IND:
        {
            npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
                "Calling STUBS_ATM_SET_CONN_TRAFFIC_IND callback into FP.\n");
            ((ix_cc_atm_cb)(eventObj->cbMethod))(
                eventObj->ixReturn, eventObj->context);
            break;
        }

        case STUBS_ATM_GET_OAM_STATISTICS:
        {
            npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
                "Calling STUBS_ATM_GET_OAM_STATISTICS callback into FP.\n");
            ((ix_cc_atm_oam_stats_cb)(eventObj->cbMethod))(
                eventObj->ixReturn, eventObj->context);
            break;
        }

        case STUBS_ATM_SET_OAM_CC:
        {
            npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
                "Calling STUBS_ATM_SET_OAM_CC callback into FP.\n");
            ((ix_cc_atm_oam_cb)(eventObj->cbMethod))(
                eventObj->ixReturn, eventObj->context);
            break;
        }

        case STUBS_ATM_GET_OAM_PROPERTIES:
        {
            ix_atm_oam_setup_req *parms;


            /* Allocate memory for param */
            parms = (ix_atm_oam_setup_req*)PIL_AllocateMemory
                (sizeof(ix_atm_oam_setup_req), PIL_MEM_ZERO_INIT);
 
            npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
                "Calling STUBS_ATM_GET_CONN_STATISTICS callback into FP.\n");
 
            if (parms == NULL)
            {
                retVal = NPF_OUT_OF_MEMORY;
                npf_logger_write(LOG_IXA_STUBS, LOG_ERR,
                    "Could not allocate memory.\n");
                goto STUBS_LBL_1;
            }
 
            ((ix_cc_atm_oam_prop_cb)(eventObj->cbMethod))(
                eventObj->ixReturn, eventObj->context,parms);

             /* Now free local memory */
             PIL_FreeMemory(parms);

            break;
        }

        case STUBS_ATM_CONNECTION_SETUP:
        case STUBS_ATM_CONNECTION_DEL:
        case STUBS_ATM_SET_AAL5_MUX:
        case STUBS_ATM_SET_AAL2_CHANNEL:
        case STUBS_ATM_DEL_AAL2_CHANNEL:
        case STUBS_ATM_SET_XC_CONNECTION:
        case STUBS_ATM_DEL_XC_CONNECTION:
        case STUBS_ATM_DATA_REQ:
        {
            npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
                "Calling %d callback into FP.\n",eventObj->evType);
            ((ix_cc_atm_cb)(eventObj->cbMethod))(
                eventObj->ixReturn, eventObj->context);
            break;
        }
#endif

    default:
        npf_logger_write(LOG_IXA_STUBS, LOG_ERR,
            "Error - unidentified IXA SDK method type\n");
    } /* switch (eventObj->evType) */

STUBS_LBL_1:

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Exiting MakeCBintoFP with status %d.\n", retVal);

    return retVal;
}

/****************************************************************************
 * Function: InsertATMMethodIntoList()                                         *
 * ---------------------------                                              *
 * Parameters: methodID         - unique ID for specific method invocation  *
 *             eventType        - ID for classification of method           *
 *             arg_Callback     - callback method into FP                   *
 *             arg_pUserContext - user context to be sent back to FP        *
 *                                                                          *
 * Description: Pack relevant data into object then put into queue.         *
 *              Trigger event so thread can service queue. Introduce a      *
 *              variable delay before inserting into queue to simulate      *
 *              random callback times into FP.                              *
 *                                                                          *
 * Return: FPPI_SUCCESS or error code                                       *
 *                                                                          *
 ****************************************************************************/

NPF_RET InsertATMMethodIntoList(
    uint32_t methodID,
    uint32_t eventType,
    void*    arg_Callback,
    void*    arg_pUserContext,
    ix_error ixRet)
{
    NPF_RET retVal = NPF_SUCCESS;
    LPSTUBS_GENERAL_EVENT_MSG mspPtr;

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Entered method with type: %s.\n",
        StubsMethodMap(methodID));

    /* Allocate memory for msg structure */
    if ((mspPtr = (LPSTUBS_GENERAL_EVENT_MSG)PIL_AllocateMemory
        (sizeof(STUBS_GENERAL_EVENT_MSG), PIL_MEM_ZERO_INIT)) == NULL)
    {
        retVal = FPPI_OUT_OF_MEMORY;
        npf_logger_write(LOG_DRS, LOG_ERR, "Could not allocate memory.\n");
        goto STUBS_LBL_1;
    }

    /* Set return value 
    SetIXAReturnObject(&ixRet); */

    /* Fill msg object with data */
    mspPtr->evType    = eventType;
    mspPtr->cbMethod  = arg_Callback;
    mspPtr->context   = arg_pUserContext;
    mspPtr->ixReturn  = ixRet;

    /* Delay a variable amount of time before insertion into queue
       to simulate integrated operation */
    DelayRandomTime();

    /* Lock the queue */
    PIL_EnterCriticalSection(&(g_stubsParms.queSyncObj));

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Insert method into msg queue.\n");

    /* Insert the object into the msg queue */
    npf_list_pushBack(&(g_stubsParms.msgQueue), (void *)mspPtr);

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Set event - a new event is pending.\n");

    /* Now notify thread that a message is pending in its queue */
    if (PIL_ReleaseSemaphore(g_stubsParms.newMsgNotifier) != PIL_OK)
    {
        npf_logger_write(LOG_DRS, LOG_ERR,
            "Error setting event object.\n");
    }

    /* Unlock the msg queues so they can be serviced by the threads */
    PIL_LeaveCriticalSection(&(g_stubsParms.queSyncObj));

STUBS_LBL_1:

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Exiting method of type %s with status %d.\n",
        StubsMethodMap(methodID), retVal);

    return retVal;
}



/****************************************************************************
 * Function: InsertMethodIntoList()                                         *
 * ---------------------------                                              *
 * Parameters: methodID         - unique ID for specific method invocation  *
 *             eventType        - ID for classification of method           *
 *             arg_Callback     - callback method into FP                   *
 *             arg_pUserContext - user context to be sent back to FP        *
 *                                                                          *
 * Description: Pack relevant data into object then put into queue.         *
 *              Trigger event so thread can service queue. Introduce a      *
 *              variable delay before inserting into queue to simulate      *
 *              random callback times into FP.                              *
 *                                                                          *
 * Return: FPPI_SUCCESS or error code                                       *
 *                                                                          *
 ****************************************************************************/

NPF_RET InsertMethodIntoList(
    uint32_t methodID,
    uint32_t eventType,
    void*    arg_Callback,
    void*    arg_pUserContext)
{
    NPF_RET retVal = NPF_SUCCESS;
    LPSTUBS_GENERAL_EVENT_MSG mspPtr;
    ix_error ixRet;

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Entered method with type: %s.\n",
        StubsMethodMap(methodID));

    /* Allocate memory for msg structure */
    if ((mspPtr = (LPSTUBS_GENERAL_EVENT_MSG)PIL_AllocateMemory
        (sizeof(STUBS_GENERAL_EVENT_MSG), PIL_MEM_ZERO_INIT)) == NULL)
    {
        retVal = FPPI_OUT_OF_MEMORY;
        npf_logger_write(LOG_DRS, LOG_ERR, "Could not allocate memory.\n");
        goto STUBS_LBL_1;
    }

    /* Set return value */
    SetIXAReturnObject(&ixRet);

    /* Fill msg object with data */
    mspPtr->evType    = eventType;
    mspPtr->cbMethod  = arg_Callback;
    mspPtr->context   = arg_pUserContext;
    mspPtr->ixReturn  = ixRet;

    /* Delay a variable amount of time before insertion into queue
       to simulate integrated operation */
    DelayRandomTime();

    /* Lock the queue */
    PIL_EnterCriticalSection(&(g_stubsParms.queSyncObj));

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Insert method into msg queue.\n");

    /* Insert the object into the msg queue */
    npf_list_pushBack(&(g_stubsParms.msgQueue), (void *)mspPtr);

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Set event - a new event is pending.\n");

    /* Now notify thread that a message is pending in its queue */
    if (PIL_ReleaseSemaphore(g_stubsParms.newMsgNotifier) != PIL_OK)
    {
        npf_logger_write(LOG_DRS, LOG_ERR,
            "Error setting event object.\n");
    }

    /* Unlock the msg queues so they can be serviced by the threads */
    PIL_LeaveCriticalSection(&(g_stubsParms.queSyncObj));

STUBS_LBL_1:

    npf_logger_write(LOG_IXA_STUBS, LOG_TRACE,
        "Exiting method of type %s with status %d.\n",
        StubsMethodMap(methodID), retVal);

    return retVal;
}

/****************************************************************************
 * Function: SetIXAReturnObject()                                           *
 * ---------------------------                                              *
 * Parameters: retObj - ptr to return code                                  *
 *                                                                          *
 * Description: Set the return code based on some randomization.            *
 *                                                                          *
 * Return: none                                                             *
 *                                                                          *
 ****************************************************************************/

void SetIXAReturnObject(
    ix_error* retObj)
{
#if 0  /* Turn random error off for now and always return success */
    uint32_t minLimit = 1;
    uint32_t maxLimit = 50;
#endif

    /* Init return code to success */
    *retObj = (ix_error)0;

#if 0  /* Turn random error off for now and always return success */
    srand(time(NULL));

    if ( (minLimit + rand() % maxLimit) == 25)
    {
        /* Set to error */
        *retObj = (ix_error)1;
    }
#endif
}

/****************************************************************************
 * Function: DelayRandomTime()                                              *
 * ---------------------------                                              *
 * Parameters: none                                                         *
 *                                                                          *
 * Description: Delay a random amount of time before returning.             *
 *                                                                          *
 * Return: none                                                             *
 *                                                                          *
 ****************************************************************************/

void DelayRandomTime()
{
    uint32_t x;
    const uint32_t minTime  = 1;
    const uint32_t maxTime  = 2000;

    srand(time(NULL));

    x = minTime + rand() % maxTime;

    PIL_DelayThread(x);
}
