/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 5
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      PH Module Core Component
 *
 * = FILENAME
 *      pkthndlr_linux_kernel_module.c
 *
 * = DESCRIPTION
 *      The file defines the call-back functions to send/receive the packets from
 *      kernel and fpm. This also contains the functions used for encoding and decoding
 *      of extra ip-header and meta header. 
 *
 * = AUTHOR
 *       Sankaralingam Thirunavukkarasu 
 *       sankaralingam.thirunavukkarasu@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      04/09/2002 (MM/DD/YYYY)
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */

#include "ix_ph.h"

extern ix_error ix_cc_fpm_sync_send_packet ( ix_uint32 arg_PortID, char* arg_pBuffer, 
		                       ix_uint32 arg_PacketLen);


/* global variable to store the remote ip-address */
ix_uint32    gDest_Addr=0;

/**
 * NAME:  ix_ph_set_cp_address
 *
 * DESCRIPTION: This is the function for setting the destination address 
 *					  
 * @Param:  - arg_address - destination address 
 *
 * @Return: IX_SUCCESS
 *
 */
ix_error ix_ph_set_cp_address ( ix_uint32 arg_address )
{
       /* assign the address to the global variable */
       gDest_Addr = arg_address;
#ifdef IX_DEBUG
       ix_ossl_message_log( "PH CORE: ix_ph_set_cp_address called \n"); 
#endif
       return IX_SUCCESS;
}

/**
 * NAME:  ix_ph_receive_packet_fpm
 *
 * DESCRIPTION: This is the call-back function registered with fpm to receive the
 *              incoming packets 
 *					  
 * @Param:  - arg_bladeid - blade-id 
 * @Param:  - arg_portid - port-id 
 * @Param:  - arg_len - length of the buffer 
 * @Param:  - arg_buffer - the packet buffer 
 *
 * @Return: IX_SUCCESS / IX_FAILURE
 *
 */
ix_error ix_ph_receive_packet_fpm (ix_uint32 arg_bladeid, ix_uint32 arg_portid, 
					ix_uint32 arg_len, char *arg_buffer )
{
	/* add meta header and another ip-header and send the packet to cp
	 * by calling this function */
#ifdef IX_DEBUG
         ix_ossl_message_log(  "ix_ph_receive_packet_fpm called !!!\n");
#endif
         return ix_ph_send_packet_cp( arg_bladeid, arg_portid, arg_len, arg_buffer );
}

/**
 * NAME:  ix_ph_receive_packet_cp
 *
 * DESCRIPTION: This is the call-back function registered with kernel to receive the
 *              incoming packets from the remote cc 
 *					  
 * @Param:  - sk_buff - the kernel buffer in which the packet is received 
 *
 * @Return: IX_SUCCESS / IX_FAILURE
 *
 */
ix_error ix_ph_receive_packet_cp ( struct sk_buff *sk_buff )
{
	/* remove the ip-header and meta header and send the packet to fpm */
#ifdef IX_DEBUG
        ix_ossl_message_log(  "ix_ph_receive_packet_cp called !!!\n");
#endif
	return ix_ph_send_packet_fpm ( sk_buff );
}

/**
 * NAME:  ix_ph_send_packet_fpm
 *
 * DESCRIPTION: This is the function for removing the ip and meta headers for the
 *              incoming packets from remote cc and send the ipv4 packet to fpm
 *					  
 * @Param:  - arg_sk_buff - the packet buffer 
 *
 * @Return: IX_SUCCESS / IX_FAILURE
 *
 */
ix_error ix_ph_send_packet_fpm ( struct sk_buff *arg_skb_buff )
{
	meta_header *metadata;
	ix_uint32 length = 0;
	ix_uint32 portid = 0;
	struct iphdr *ipheader;

#ifdef IX_DEBUG
      	ix_ossl_message_log( "PH CORE: ix_ph_send_packet_fpm called \n"); 
#endif

	ipheader = arg_skb_buff->nh.iph;

	/* remove the meta-header */
        skb_pull (arg_skb_buff, sizeof(meta_header));

	metadata = (meta_header*)(ipheader+1);

        portid = metadata->portid;
        length = metadata->length;


#ifdef IX_DEBUG
      	ix_ossl_message_log( "PH CORE: Port-Id  = %d\n", ntohl(portid)); 
      	ix_ossl_message_log( "PH CORE: Length  = %d\n", ntohl(length)); 
#endif

	/* send the packet to fpm */
        ix_cc_fpm_sync_send_packet(ntohl(portid), 
                                 arg_skb_buff->data, 
                                 ntohl(length)); 
        kfree_skb (arg_skb_buff);
	return IX_SUCCESS;
}

/* Need this wrapper because NF_HOOK takes the function address */
static inline int do_ip_send(struct sk_buff *skb)
{
#ifdef IX_DEBUG
      	ix_ossl_message_log( "PH CORE: do_ip_send() \n"); 
#endif
	return ip_send(skb);
}

/**
 * NAME:  ix_ph_send_packet_cp
 *
 * DESCRIPTION: This is the function for encoding the Ip-header and meta header to all 
 *              incoming packets from fpm and send it to remote cc 
 *					  
 * @Param:  - arg_bladeid - blade-id 
 * @Param:  - arg_portid - port-id 
 * @Param:  - arg_len - length of the buffer 
 * @Param:  - arg_buffer - the packet buffer 
 *
 * @Return: IX_SUCCESS / IX_FAILURE
 *
 */
ix_error ix_ph_send_packet_cp ( ix_uint32 arg_bladeid, ix_uint32 arg_portid, ix_uint32 arg_len,
		char *arg_buffer )
{
	struct sk_buff *skb, *new_skb;
	struct iphdr *ipheader1;
	meta_header *metadata; 
	struct rtable *rt; 
        struct iphdr *iph;
	int headroom;
	struct net_device_stats vstats;
	struct net_device_stats *stats = &vstats;
	struct rt_key key = {dst:gDest_Addr, src:INADDR_ANY,tos:0, oif:0} ;

#ifdef IX_DEBUG
	char *ch;
	int i;
#endif

	if (ip_route_output_key(&rt, &key)) {
#ifdef IX_DEBUG
      	    ix_ossl_message_log( "PH CORE: ip_route_output Failed \n"); 
#endif
	    return 1;
	}

	headroom = ((rt->u.dst.dev->hard_header_len+15)&~15) + sizeof(struct iphdr);

	skb = alloc_skb ( arg_len , GFP_ATOMIC);

	if ( skb == NULL )
	{
#ifdef IX_DEBUG
      	    ix_ossl_message_log( "PH CORE: skb memory allocation Failed \n"); 
#endif
	    return 1;
	}

	skb_put(skb,arg_len);
	memcpy (skb->data,arg_buffer,arg_len); 
        skb->nh.iph = (struct iphdr*)(skb->data);

#ifdef IX_DEBUG
      	    ix_ossl_message_log( "PH CORE: headroom = %d \n", headroom); 
      	    ix_ossl_message_log( "PH CORE: skb->head = %d \n", skb->head); 
      	    ix_ossl_message_log( "PH CORE: skb->data = %u \n", skb->data); 
      	    ix_ossl_message_log( "PH CORE: skb->end = %u \n", skb->end); 
      	    ix_ossl_message_log( "PH CORE: skb->tail = %u \n", skb->tail); 
      	    ix_ossl_message_log( "PH CORE: skb->raw = %u \n", skb->nh.raw); 
      	    ix_ossl_message_log( "PH CORE: arg_len = %u \n", arg_len);
      	    ix_ossl_message_log( "PH CORE: skb->data_len = %u \n", skb->data_len);
      	    ix_ossl_message_log( "PH CORE: skb->len = %u \n", skb->len); 
#endif

        new_skb = skb_realloc_headroom(skb, headroom);

        if (skb->sk)
	         skb_set_owner_w(new_skb, skb->sk);
	dev_kfree_skb(skb);
	skb = new_skb;

        skb->nh.raw = skb_push(skb,
                           sizeof(struct iphdr) + sizeof(meta_header));


	skb->dst = &rt->u.dst;

	ipheader1 = skb->nh.iph;
        iph = skb->nh.iph;
        iph->version = 4;
	iph->ihl = 5;
	iph->frag_off = 0;
	iph->protocol = IP_PROTO_VIP;
	iph->tos = 0;
	iph->daddr = rt->rt_dst;
	iph->saddr = rt->rt_src;
	iph->ttl = ipheader1->ttl;

	metadata = (meta_header *) (iph + 1);
	metadata->portid = htonl(arg_portid);
	metadata->length = htonl(skb->len - sizeof(meta_header) - sizeof(struct iphdr));

#ifdef IX_DEBUG
      	    ix_ossl_message_log( "PH CORE: skb->head = %u \n", skb->head); 
      	    ix_ossl_message_log( "PH CORE: skb->data = %u \n", skb->data); 
      	    ix_ossl_message_log( "PH CORE: skb->end = %u \n", skb->end); 
      	    ix_ossl_message_log( "PH CORE: skb->tail = %u \n", skb->tail); 
      	    ix_ossl_message_log( "PH CORE: skb->raw = %u \n", skb->nh.raw); 
      	    ix_ossl_message_log( "PH CORE: arg_len = %u \n", arg_len);
      	    ix_ossl_message_log( "PH CORE: skb->data_len = %u \n", skb->data_len);
      	    ix_ossl_message_log( "PH CORE: skb->len = %u \n", skb->len); 

	ch = (char*)skb->data;
	ix_ossl_message_log( "The pkt sent  ....\n");
        for (i=0; i<( sizeof(struct iphdr) + sizeof(meta_header) + arg_len); i++)
	{
		ix_ossl_message_log( "%x ,",*ch++);
	}
	ix_ossl_message_log( "The pkt sent to %x  ....\n", rt->rt_dst);
	ix_ossl_message_log( "The pkt src to %x  ....\n", rt->rt_src);

#endif
	/* transmit the packet */ 
	IPTUNNEL_XMIT();
	return IX_SUCCESS;
}
 
