/*****************************************************************************
*                  INTEL CORPORATION PROPRIETARY INFORMATION
*  This software is supplied under the terms of a license agreement or
*  nondisclosure agreement with Intel Corporation and may not be copied
*  or disclosed except in accordance with the terms of that agreement.
*  Copyright (c) 2001 Intel Corporation. All Rights Reserved.
*****************************************************************************/

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <assert.h>
#include <errno.h>

#include "fp_common.h"
#include "pil.h"
#include "rwlock.h"
#include "npf_ds_types.h"
#include "npf_mpls_types.h"
#include "FPModule.h"

#include "dpe_mgr.h"
#include "qos_mgr.h"
#include "qos_mgr_impl.h"
#include "FidMgr.h"
#include "qos_flow_events.h"

/* Current router configuration */
extern router_config    g_current_router_config;

/* System specific constants - initialized by platform specific FP Manager */

extern uint32_t g_max_flow_id;
extern uint32_t g_max_queue_id;
extern uint32_t g_max_num_ports;

/* these tables will be allocated after querying the max 
   supported Flow ID and Queue ID on the system 
*/
static p_dpe_clfr_t*      g_dpe_clfr_tbl; 
static p_dpe_marker_t*    g_dpe_marker_tbl; 
static p_dpe_meter_t*     g_dpe_meter_tbl; 
static p_dpe_dropper_t*   g_dpe_dropper_tbl; 
static p_dpe_queue_t*     g_dpe_queue_tbl; 
static p_dpe_scheduler_t* g_dpe_scheduler_tbl; 

static dpe_flow_tuple_t   g_best_effort_flow_tuple;

/* Global lock for DPE Manager
*/
static PilCriticalSection g_dpe_global_lock;

#define DPE_LOCK() PIL_EnterCriticalSection( &g_dpe_global_lock )
#define DPE_UNLOCK() PIL_LeaveCriticalSection( &g_dpe_global_lock )

/*****************************************************************************
*
*   Function:   fpm_dpe_mgr_shutdown()
*
*   Parameters: None 
*
*   Description: This function initializes the DPE Manager. All the DPE tables
*                are allocated memory.
*                For best effort traffic, best effort flow id and queue id are
*                allocated & the data path is setup for this,
*
*   Return: 0 on success, -1 on error
*****************************************************************************/

int fpm_dpe_mgr_init()
{
    int errtype = 0, i;

    PDBG_TRACE ();
    PDBG_PRINT ("Max Flow id  = %d\n Max Queue Id = %d\n Num ports = %d\n", 
		g_max_flow_id, g_max_queue_id, g_max_num_ports);

    PIL_InitializeCriticalSection (&g_dpe_global_lock);
    /* Now initialize the DPE tables - We need to allocate only pointers
       for each table. Entries for each table are allocated only when
       required.
     */
    for (i = 0; i < g_max_flow_id; i++) {
	g_dpe_clfr_tbl = (p_dpe_clfr_t*) 
	    PIL_AllocateMemory (g_max_flow_id *
				sizeof (p_dpe_clfr_t),
				PIL_MEM_ZERO_INIT);
	g_dpe_marker_tbl = (p_dpe_marker_t*) 
	    PIL_AllocateMemory (g_max_flow_id *
				sizeof (p_dpe_marker_t),
				PIL_MEM_ZERO_INIT);
	g_dpe_meter_tbl = (p_dpe_meter_t*) 
	    PIL_AllocateMemory (g_max_flow_id *
				sizeof (p_dpe_meter_t),
				PIL_MEM_ZERO_INIT);
	g_dpe_dropper_tbl = (p_dpe_dropper_t*) PIL_AllocateMemory 
	    (g_max_flow_id *
	     sizeof (p_dpe_dropper_t),
	     PIL_MEM_ZERO_INIT);
	g_dpe_queue_tbl = (p_dpe_queue_t*) PIL_AllocateMemory 
	    (g_max_flow_id *
	     sizeof (p_dpe_queue_t),
	     PIL_MEM_ZERO_INIT);
    } /* end (create memory for all tables of size MAX FLOW ID) */

    g_dpe_scheduler_tbl = (p_dpe_scheduler_t*) PIL_AllocateMemory 
	(g_max_queue_id * sizeof (p_dpe_scheduler_t),
	 PIL_MEM_ZERO_INIT);

    if ((!g_dpe_clfr_tbl) || (!g_dpe_marker_tbl) || (!g_dpe_meter_tbl) || 
	(!g_dpe_dropper_tbl) || (!g_dpe_queue_tbl) || (!g_dpe_scheduler_tbl)) {
	PDBG_PRINT ("Unable to allocate memory for DPE Table(s)!\n");
	if (g_dpe_clfr_tbl)
	    PIL_FreeMemory ((void*) g_dpe_clfr_tbl);
	if (g_dpe_marker_tbl)
	    PIL_FreeMemory ((void*) g_dpe_marker_tbl);
	if (g_dpe_meter_tbl)
	    PIL_FreeMemory ((void*) g_dpe_meter_tbl);
	if (g_dpe_dropper_tbl)
	    PIL_FreeMemory ((void*) g_dpe_dropper_tbl);
	if (g_dpe_queue_tbl)
	    PIL_FreeMemory ((void*) g_dpe_queue_tbl);
	if (g_dpe_scheduler_tbl)
	    PIL_FreeMemory ((void*) g_dpe_scheduler_tbl);
	errtype = -1;
    }

    /* set up default path for Best Effort traffic */

    /* allocate Best Effort Flow Id and Queue ID */
    errtype = allocate_flow_id_qos (&g_best_effort_flow_tuple.flow_id);
    for (i = 0; i < g_max_num_ports; i++) {
         errtype = allocate_queue_id_qos (i, 
				&g_best_effort_flow_tuple.queue_id);
	QOS_DBG_PRINT (LOG_INFO, "queue id = %d \n", 
			g_best_effort_flow_tuple.queue_id);
	}
    g_best_effort_flow_tuple.next_block_id = 0;

    dpe_setup_besteffort_path ();

    PRETURN (errtype);
}

/*****************************************************************************
*
*   Function:   fpm_dpe_mgr_shutdown()
*
*   Parameters: None 
*
*   Description: This function shuts-down the DPE Manager. It deallocates 
*                memory for all the tables. It assumes that a higher level
*                application removes all QoS configurations before this
*                function is called, and so does not perform and QoS
*                shutdown intelligently.
*
*   Return: 0 on success, -1 on error
*****************************************************************************/

int fpm_dpe_mgr_shutdown ()
{
    int i;
    NPF_DATA data;
    dpe_EntryId_t* prev_tbl_entry;

    PDBG_TRACE ();

    PIL_DestroyCriticalSection (&g_dpe_global_lock);

    for (i = 0; i < g_max_flow_id; i++) {
	if (g_dpe_clfr_tbl[i] != NULL) {
	    while (npf_list_popFront (&(g_dpe_clfr_tbl[i]->prev_tbl_entry_list),
				      (void**) data) != -1) {
		prev_tbl_entry = (dpe_EntryId_t*) data;
		PIL_FreeMemory ((void*) prev_tbl_entry);
	    }/* end loop on all entries in the Prev Table Entry List */
	    npf_list_destroy (&g_dpe_clfr_tbl[i]->prev_tbl_entry_list);
	    PIL_FreeMemory ((void*) g_dpe_clfr_tbl[i]);
	}
	if (g_dpe_marker_tbl[i] != NULL) {
	    while (npf_list_popFront 
		   (&(g_dpe_marker_tbl[i]->prev_tbl_entry_list),
		    (void**) data) != -1) {
		prev_tbl_entry = (dpe_EntryId_t*) data;
		PIL_FreeMemory ((void*) prev_tbl_entry);
	    }/* end loop on all entries in the Prev Table Entry List */
	    npf_list_destroy (&g_dpe_marker_tbl[i]->prev_tbl_entry_list);
	    PIL_FreeMemory ((void*) g_dpe_marker_tbl[i]);
	}
	if (g_dpe_meter_tbl[i] != NULL) {
	    while (npf_list_popFront (&(g_dpe_meter_tbl[i]->prev_tbl_entry_list),
				      (void**) data) != -1) {
		prev_tbl_entry = (dpe_EntryId_t*) data;
		PIL_FreeMemory ((void*) prev_tbl_entry);
	    }/* end loop on all entries in the Prev Table Entry List */
	    npf_list_destroy (&g_dpe_meter_tbl[i]->prev_tbl_entry_list);
	    PIL_FreeMemory ((void*) g_dpe_meter_tbl[i]);
	}
	if (g_dpe_dropper_tbl[i] != NULL) {
	    while (npf_list_popFront 
		   (&(g_dpe_dropper_tbl[i]->prev_tbl_entry_list),
		    (void**) data) != -1) {
		prev_tbl_entry = (dpe_EntryId_t*) data;
		PIL_FreeMemory ((void*) prev_tbl_entry);
	    }/* end loop on all entries in the Prev Table Entry List */
	    npf_list_destroy (&g_dpe_dropper_tbl[i]->prev_tbl_entry_list);
	    PIL_FreeMemory ((void*) g_dpe_dropper_tbl[i]);
	}
	if (g_dpe_queue_tbl[i] != NULL) {
	    while (npf_list_popFront (&(g_dpe_queue_tbl[i]->prev_tbl_entry_list),
				      (void**) data) != -1) {
		prev_tbl_entry = (dpe_EntryId_t*) data;
		PIL_FreeMemory ((void*) prev_tbl_entry);
	    }/* end loop on all entries in the Prev Table Entry List */
	    npf_list_destroy (&g_dpe_queue_tbl[i]->prev_tbl_entry_list);
	    PIL_FreeMemory ((void*) g_dpe_queue_tbl[i]);
	}
    } /* end for all tables of size MAX_FLOW_ID */
    for (i = 0; i < g_max_queue_id; i++) {
	if (g_dpe_scheduler_tbl[i] != NULL) {
	    while (npf_list_popFront 
		   (&(g_dpe_scheduler_tbl[i]->prev_tbl_entry_list),
		    (void**) data) != -1) {
		prev_tbl_entry = (dpe_EntryId_t*) data;
		PIL_FreeMemory ((void*) prev_tbl_entry);
	    }/* end loop on all entries in the Prev Table Entry List */
	    npf_list_destroy (&g_dpe_scheduler_tbl[i]->prev_tbl_entry_list);
	    PIL_FreeMemory ((void*) g_dpe_scheduler_tbl[i]);
	}
    } /* end for table of size MAX_QUEUE_ID */

    PIL_FreeMemory ((void*) g_dpe_clfr_tbl);
    PIL_FreeMemory ((void*) g_dpe_marker_tbl);
    PIL_FreeMemory ((void*) g_dpe_meter_tbl);
    PIL_FreeMemory ((void*) g_dpe_dropper_tbl);
    PIL_FreeMemory ((void*) g_dpe_queue_tbl);
    PIL_FreeMemory ((void*) g_dpe_scheduler_tbl);

    PRETURN (FPPI_SUCCESS);
}

/*****************************************************************************
*
*   Function:   dpe_setup_besteffort_path()
*
*   Parameters: None 
*
*   Description: This function is for setting up a Best Effort path. 
*                This configures the SRTC Meter, WRED Buffer Manager and  
*                DRR scheduler with default paramters. It also allocates and
*                sets up a flow id/queue id for the BE traffic path. 
*
*   Return: 0 on success, -1 on failure
*****************************************************************************/
int dpe_setup_besteffort_path ()
{
    int err = 0, i;
    dpe_flow_tuple_t downstream_flowtuple, failnext_flowtuple;
    npf_ds_Meter_t meter;

    QOS_DBG_TRACE ();
    /* Set up defaults for Scheduler & Queue for every port */
    for (i = 0; i < g_max_num_ports; i++) {
        QOS_DBG_PRINT (LOG_INFO, "i= %d flow id = %d\n, q id = %d", i,
		        g_best_effort_flow_tuple.flow_id,
		        g_best_effort_flow_tuple.queue_id);

	qos_add_default_scheduler_entry (i, g_best_effort_flow_tuple);
	qos_add_default_queue_entry (i, g_best_effort_flow_tuple.queue_id);
    }
    /* set up WRED Buf Mgr for the best effort Flow id */
    qos_add_default_red_instance (g_best_effort_flow_tuple.flow_id);
    /* set up the meter for the best effort path */
    qos_create_meter (g_best_effort_flow_tuple);
    downstream_flowtuple.flow_id = 0;
    downstream_flowtuple.queue_id = 0;
    downstream_flowtuple.next_block_id = 0;
    failnext_flowtuple.flow_id = 0;
    failnext_flowtuple.queue_id = 0;
    failnext_flowtuple.next_block_id = 0;
    qos_setup_meter (g_best_effort_flow_tuple,
		     downstream_flowtuple, failnext_flowtuple);
    meter.failnexttableid = 0;
    meter.failnextinstid = 0;
    meter.Specific =  PIL_AllocateMemory (sizeof (npf_ds_TBParam_t),
					  PIL_MEM_ZERO_INIT);
   ((npf_ds_TBParam_t*)meter.Specific)->meter_type = npf_ds_TBParam;
   ((npf_ds_TBParam_t*)meter.Specific)->Type = npf_ds_SrTCMBlind;
   ((npf_ds_TBParam_t*)meter.Specific)->Rate = 100000000;
   ((npf_ds_TBParam_t*)meter.Specific)->BurstSize = 100000000;
   ((npf_ds_TBParam_t*)meter.Specific)->Interval = 100000000;
    qos_set_meter_params (g_best_effort_flow_tuple, 
			  &meter);

    PRETURN (err);
}

int dpe_get_besteffort_flow_tuple (uint32_t* flowid, uint32_t* queueid)
{
    *flowid = g_best_effort_flow_tuple.flow_id;
    *queueid = g_best_effort_flow_tuple.queue_id;

    return 0;
}

/*************************************************************************
 *                                                                       *
 *                         DPE MANAGEMENT ROUTINES                       *
 *                                                                       *
 *************************************************************************/
/**************************************************************************
*
*   Function: dpe_add_entry
*
*   Parameters: IN: Interface,
*               IN: table_id
*               IN: instance_id
*               IN: dpe_entry - void*
*               IN: Next table id
*               IN: Next instance_id
* 
*   Description: add a new entry in one of the DPE tables. This function
*                has the entire logic of flow id/queue id generation and
*                propagation. It also ensures that valid parameters are 
*                passed to the data plane, in the correct order as much as
*                possible.
*
*   PRETURN : 0 on success, -1 for error 
*************************************************************************/

int dpe_add_entry ( uint32_t             intf, 
		    npf_ds_TableId_t     table_id,
		    npf_ds_InstanceId_t  inst_id,
		    void*                dpe_entry,
		    npf_ds_TableId_t     next_tbl_id,
		    npf_ds_InstanceId_t  next_inst_id )
{
    uint32_t            curr_queue_id, curr_flow_id, err;
    dpe_EntryId_t*      prev_tbl_entry;
    dpe_flow_tuple_t    flow_tuple;
    dpe_scheduler_t*    sched_entry;
    dpe_meter_t*        meter_entry;
    dpe_queue_t*        queue_entry;
    dpe_clfr_t*         clfr_entry;
    uint32_t            status;

    QOS_DBG_TRACE ();
	
    curr_flow_id = g_max_flow_id + 1;
    curr_queue_id= g_max_queue_id + 1;
    status = 0;

    DPE_LOCK(); /* Critical Section Begin */
    /* Assumption - the last Data Path Element can only be the SCHEDULER */ 
    if ( table_id == DPE_SCHEDULER_TABLE_ID ) {
	curr_flow_id = g_max_flow_id + 1;
        curr_queue_id = g_max_queue_id + 1;

	flow_tuple.flow_id = curr_flow_id;
	flow_tuple.queue_id = curr_queue_id;
	flow_tuple.next_block_id = 0;
	status = DPE_STATUS_VALID; 
	goto CREATE;
    }
    /* if the current table is the QUEUE, allocate a Queue ID */
    else if ( table_id == DPE_QUEUE_TABLE_ID ) {
	/* allocate queue id 
	*/
	err = allocate_queue_id_qos (intf, &curr_queue_id);
	QOS_DBG_PRINT (LOG_INFO, "Generated new Queue ID: %d !\n",
		       curr_queue_id);
        flow_tuple.queue_id = curr_queue_id;

	/* THIS WILL HAPPEN AT THE QUEUE */
	/* generate a new flow id */
	err = allocate_flow_id_qos (&flow_tuple.flow_id);
	QOS_DBG_PRINT (LOG_INFO, 
		       "New Flow Id = %d \t Queue Id = %d!\n",
		       flow_tuple.flow_id, flow_tuple.queue_id);
	dpe_add_inst_to_prev_entry_list (table_id, inst_id, next_tbl_id,
					 next_inst_id);
	//dpe_propagate_flow_tuple (table_id, inst_id);
    } /* end if (current table == QUEUE) */
    else {
	/* now set up the back pointers - this table_id and inst_id must be
	   in the prev_tbl_list of the TABLE_ID and INSTANCE_ID of the current
	   DPE entry. For ex.
	   if this is a Queue entry, next block is Scheduler. So,
	   Scheduler_table[next_inst_id].prev_tbl = {curr_tbl_id, curr_inst_id}
	*/
	dpe_add_inst_to_prev_entry_list (table_id, inst_id, next_tbl_id,
					 next_inst_id);
	/* get downstream flow tuple */
	dpe_get_downstream_flow_tuple (next_tbl_id, next_inst_id, &flow_tuple);
	QOS_DBG_PRINT (LOG_INFO, " Flow Id = %d \n\t Queue Id = %d \n\t",
			flow_tuple.flow_id, flow_tuple.queue_id);
	
	/* propagate flow tuple to upstream blocks */
	if (flow_tuple.flow_id != g_max_flow_id + 1 &&
	    flow_tuple.queue_id != g_max_queue_id + 1)
	    dpe_propagate_flow_tuple (table_id, inst_id);
    }
    /**************************************************************
     * In order to be able to aggregate flows at the meter, we have
     * to a bit of a trick here. Essentially, the {fid, qid} we have
     * got now from the {next_tbl, next_inst} has to be preserved.
     * What we do for aggregating upstream blocks is this
     * - generate a new flow_id. 
     * - Use the same Queue id that we got from the next DPE 
     * - the new Flow ID will be given to any upstream blocks like Clfr
     * - Set up the meter so that it marks incoming packets in this way
     *    - it accepts flows with the upstream flow id
     *    - it changes the flow id to the downstream flow id when it
     *      leaves the meter
     * - we do this for both conforming and non conforming traffic
     *
     * THIS IS DONE IN THE SPECIAL CASE OF CREATING A NEW METER ENTRY
     **************************************************************/

    /* Get the STATUS of the DPE entry that corresponds to 
       [next_tbl, nxt_inst] */
    dpe_get_downstream_status (next_tbl_id, next_inst_id, &status);
    QOS_DBG_PRINT (LOG_INFO, "Status = %d \n", status);
    if (status == DPE_STATUS_PROPAGATED)
	status = DPE_STATUS_VALID;
 CREATE:

    /* create the new table entry, and add it to the correct table */
    err = dpe_create_table_entry (intf, table_id, inst_id,
				  (void*) dpe_entry, next_tbl_id, 
				  next_inst_id, flow_tuple, status);
    
    QOS_DBG_PRINT (LOG_INFO, "Current parameters:!\n\t"
		   " Interface   = %d \n\t"
		   "Table Id     = %d \n\t"
		   "Instance Id  = %d \n\t"
		   "Next Table Id= %d \n\t"
		   "Next Inst Id = %d \n\t"
		   "Flow Id      = %d \n\t"
		   "Queue Id     = %d \n\t"
		   "DPE Status   = %d \n\t", intf,
		   table_id, inst_id, next_tbl_id, next_inst_id,
		   flow_tuple.flow_id, flow_tuple.queue_id, status);

    /* if this DPE has now bcome VALID, propagate the status to
       any waiting DPEs upstream */
    if (status == DPE_STATUS_VALID)
	dpe_propagate_status (table_id, inst_id, status);
    /* IF new DPE was a Queue or meter, we need to propagate the new Flow Tuple to 
       upstream DPEs. In our implementation, Queue and Meter are the only DPEs
       that generate new flow id and/or queue ids */
    if (table_id == DPE_QUEUE_TABLE_ID || table_id == DPE_METER_TABLE_ID)
	dpe_propagate_flow_tuple (table_id, inst_id);
    dpe_propagate_params (table_id, inst_id);

    DPE_UNLOCK(); /* Critical Section end */

    PRETURN (0);
}

/**************************************************************************
*   Function: dpe_propagate_params
*
*   Parameters: IN: table_id
*               IN: instance_id
* 
*   Description: propagates the new DPE params to the underlying data plane.
*                This is a recursive function; it begins propagating params
*                from the current DPE, follows its backward pointers and 
*                recursively performs the same action.
*                Parameters are propagated only if the DPE is VALID and has
*                not yet been propagated to the data plane. Once the DPE's
*                parameters are propagated, its status is set to PROPAGATED.
*
*   PRETURN : 0 on success, -1 for error 
*************************************************************************/

int dpe_propagate_params (npf_ds_TableId_t     table_id,
			  npf_ds_InstanceId_t  inst_id)
{
    int                 err = 0;
    dpe_EntryId_t*      entry;
    dpe_flow_tuple_t    flow_tuple;
    dpe_scheduler_t*    sched_entry;
    dpe_meter_t*        meter_entry;
    dpe_queue_t*        queue_entry;
    dpe_clfr_t*         clfr_entry;
    dpe_marker_t*       marker_entry;
    dpe_dropper_t*      dropper_entry;
    DListItr*           itr;

    QOS_DBG_TRACE ();
    /*****************************************************************
     * PROPAGATE PARAMETERS TO UNDERLYING NPU
     * The exact point at which API calls are propagated to the 
     * underlying NPU could be different.
     * The following code is specific to Sauvie Island for IXP12xx NPU
     ****************************************************************/
    
    switch (table_id) {
    case DPE_SCHEDULER_TABLE_ID:
	    {
		sched_entry = (dpe_scheduler_t*)g_dpe_scheduler_tbl[inst_id];
		/***************************************************************
		 * For Sauvie Island, the DRR scheduler does not support 
		 * hierarchical scheduling. Hence the Scheduler parameters 
		 * passed in here cannot be used. We have to wait for the Queue 
		 * API call to pass in parameters to the Scheduler.
		 **************************************************************/
		if (sched_entry->status == DPE_STATUS_VALID)
		    sched_entry->status = DPE_STATUS_PROPAGATED;
		itr = npf_list_itrCreate 
		    (&(sched_entry->prev_tbl_entry_list));
		QOS_DBG_PRINT (LOG_INFO, "Propagating Scheduler params\n");
	    }
	break;
    case DPE_QUEUE_TABLE_ID:
	{
	    queue_entry = (dpe_queue_t*) g_dpe_queue_tbl[inst_id]; 
	    if (queue_entry->status == DPE_STATUS_VALID) {
		queue_entry->status = DPE_STATUS_PROPAGATED;
		/*********************************************************
		 * Now make a call to the DRR Scheduler block of Sauvie
		 * Island. 
		 * 
		 * SET UP DRR SCHEDULER 
		 ********************************************************/
		
		/* Need Rate params from Queue */
		sched_entry = 
		    g_dpe_scheduler_tbl[queue_entry->next_instance_id];
		QOS_DBG_PRINT (LOG_INFO, "Propagating Queue params\n");
		err = qos_add_scheduler_entry (sched_entry->intf, 
					       queue_entry->flow_tuple,
					       queue_entry->queue.min_rate,
					       queue_entry->queue.max_rate);
		if (err == 0) {
		/*********************************************************
		 * SET UP WRED BUFFER MANAGER PARAMETERS
		 * Default set up for Queue/Alg Dropper. If app makes a
		 * AlgDropper call, WRED params can be changed.
		 ********************************************************/
		err = qos_add_default_queue_entry 
		    (queue_entry->intf, 
		     queue_entry->flow_tuple.queue_id);
		err = qos_add_default_red_instance 
		    (queue_entry->flow_tuple.flow_id);
		}
	    }
	    itr = npf_list_itrCreate 
		(&(queue_entry->prev_tbl_entry_list));
	}
	break;
    case DPE_ALG_DROPPER_TABLE_ID:
	{
	    dropper_entry = g_dpe_dropper_tbl[inst_id];
	    if (dropper_entry->status == DPE_STATUS_VALID) {
		dropper_entry->status = DPE_STATUS_PROPAGATED;
	    /* add code here for overriding default RED params and EWMA
	       set up earlier for this {flow_id, queue_id} tuple */
		QOS_DBG_PRINT (LOG_INFO, "Propagating Dropper params\n");
		err = qos_set_dropper_params (dropper_entry->intf, 
					      dropper_entry->flow_tuple,
					      &dropper_entry->dropper);
	    }
	    itr = npf_list_itrCreate(&(dropper_entry->prev_tbl_entry_list));
	}
	break;
    case DPE_METER_TABLE_ID:
	{
	    meter_entry = (dpe_meter_t*) g_dpe_meter_tbl[inst_id];
	    if (meter_entry->status == DPE_STATUS_VALID) {
		meter_entry->status = DPE_STATUS_PROPAGATED;
		/* We have to set up the Meter for both Conforming and
		   non conforming traffic paths */

		QOS_DBG_PRINT (LOG_INFO, "Propagating Meter params\n");
		/* create a new meter - this prepares the meter for recving
		   packets with the upstream flow id */
		err = qos_create_meter (meter_entry->flow_tuple);
		/* now set up the meter for updating incoming flow id to
		   the downstream flow id, queue id */
		if (err == 0) {
		    err = qos_setup_meter (meter_entry->flow_tuple,
					   meter_entry->downstream_flow_tuple,
					   meter_entry->failnext_flow_tuple);
		    /* now set up the meter parameters */
		    err = qos_set_meter_params (meter_entry->flow_tuple, 
						&meter_entry->meter);
		}
	    }
	    itr = npf_list_itrCreate 
		(&(meter_entry->prev_tbl_entry_list));
	}
	break;
    case DPE_MARKER_TABLE_ID:
      {
	marker_entry = (dpe_marker_t*) g_dpe_marker_tbl[inst_id];
	if (marker_entry->status == DPE_STATUS_VALID) {
	    marker_entry->status = DPE_STATUS_PROPAGATED;
	    QOS_DBG_PRINT (LOG_INFO, "Propagating Marker params\n");
	    err = qos_add_marker (marker_entry->flow_tuple, 
				  &marker_entry->action);
	}
	itr = npf_list_itrCreate (&(marker_entry->prev_tbl_entry_list));
      }
        break;
    case DPE_CLASSIFIER_TABLE_ID:
	{
	    clfr_entry = (dpe_clfr_t*) g_dpe_clfr_tbl[inst_id];
	    if (clfr_entry->status == DPE_STATUS_VALID) {
	    QOS_DBG_PRINT (LOG_INFO, "Propagating Classifier params\n");
		if (g_current_router_config == DS_MPLS_LER ||
		    g_current_router_config == DS_MPLS_LSR) {
		    /* we have to call into MPLS to signal that Flow ID/Queue ID
		       have been generated for this Filter. MPLS Filters can be
		       either DSCP, Label or FEC based. So, we have to examine
		       what kind of classification filter we just got and then 
		       invoke MPLS callbacks with the new FID/QiD
		    */
		    clfr_entry = (dpe_clfr_t*) g_dpe_clfr_tbl[inst_id];
		    dpe_signal_mpls (clfr_entry);
		} /* end if (MPLS Classifier) */
		else if (g_current_router_config == DS_EDGE ||
			 g_current_router_config == DS_CORE) {
		    err = qos_add_classifier (clfr_entry->flow_tuple,
					      clfr_entry->filter,
					      &clfr_entry->filter_hdl);
		}
		clfr_entry->status = DPE_STATUS_PROPAGATED;
	    }
	    itr = npf_list_itrCreate (&(clfr_entry->prev_tbl_entry_list));
	}
	break;
    } /* end switch on current table id to decide what params to
	 propagate */
    
    /************************************************************************
     * Now we check if the DPE has any upstream blocks associated with it.
     * If they are, and their parameters are complete and not propagated yet,
     * do so now.
     ***********************************************************************/
    
    if (npf_list_itrFirst(itr) != -1) {
	do {
	    QOS_DBG_PRINT (LOG_INFO, "Found a back ptr!\n");
	    entry = (dpe_EntryId_t*) npf_list_itrGetData(itr);
	    /* recursively call this function for this table */
	    dpe_propagate_params (entry->table_id, entry->instance_id);
	}while (npf_list_itrNext(itr) != -1);
    }
    npf_list_itrDelete (itr);
    PRETURN (err);
}

/**************************************************************************
*   Function: dpe_propagate_status
*
*   Parameters: IN: table_id
*               IN: instance_id
*               IN: status
* 
*   Description: propagates the new DPE status to any waiting upstream DPEs
*                This function is also recursive. It starts with the given 
*                DPE, sets the status to VALID, traverses its backward pointers
*                and sets their status to VALID if they are currently INVALID.
*                This will subsequently result in propagating any of the 
*                new DPEs whose status changed from INVALID to VALID.  
*
*   PRETURN : 0 on success, -1 for error 
*************************************************************************/

int dpe_propagate_status (npf_ds_TableId_t     table_id,
			  npf_ds_InstanceId_t  inst_id,
			  uint32_t             status)
{
    dpe_EntryId_t*      entry;
    dpe_flow_tuple_t    flow_tuple;
    dpe_scheduler_t*    sched_entry;
    dpe_meter_t*        meter_entry;
    dpe_queue_t*        queue_entry;
    dpe_clfr_t*         clfr_entry;
    dpe_marker_t*       marker_entry;
    dpe_dropper_t*      dropper_entry;
    DListItr*           itr = NULL;

    QOS_DBG_TRACE ();
    switch (table_id) {
    case DPE_SCHEDULER_TABLE_ID:
	    {
		sched_entry = (dpe_scheduler_t*)g_dpe_scheduler_tbl[inst_id];
		if (sched_entry->status == DPE_STATUS_VALID) {
		    itr = npf_list_itrCreate 
			(&(sched_entry->prev_tbl_entry_list));
		    QOS_DBG_PRINT (LOG_INFO, "Propagating Scheduler status\n");
		}
	    }
	break;
    case DPE_QUEUE_TABLE_ID:
	{
	    queue_entry = (dpe_queue_t*) g_dpe_queue_tbl[inst_id]; 
	    if (queue_entry->status == DPE_STATUS_INVALID) {
		queue_entry->status = DPE_STATUS_VALID;
		itr = npf_list_itrCreate 
		    (&(queue_entry->prev_tbl_entry_list));
	    }
	}
	break;
    case DPE_ALG_DROPPER_TABLE_ID:
	{
	    dropper_entry = g_dpe_dropper_tbl[inst_id];
	    if (dropper_entry->status == DPE_STATUS_INVALID) {
		dropper_entry->status = DPE_STATUS_VALID;
		itr = npf_list_itrCreate(&(dropper_entry->prev_tbl_entry_list));
	    }
	}
	break;
    case DPE_METER_TABLE_ID:
	{
	    meter_entry = (dpe_meter_t*) g_dpe_meter_tbl[inst_id];
	    if (meter_entry->status == DPE_STATUS_INVALID) {
		meter_entry->status = DPE_STATUS_VALID;
		itr = npf_list_itrCreate 
		    (&(meter_entry->prev_tbl_entry_list));
	    }
	}
	break;
    case DPE_MARKER_TABLE_ID:
      {
	marker_entry = (dpe_marker_t*) g_dpe_marker_tbl[inst_id];
	if (marker_entry->status == DPE_STATUS_INVALID) {
	    marker_entry->status = DPE_STATUS_VALID;
	    QOS_DBG_PRINT (LOG_INFO, "Propagating Marker params\n");
	    itr = npf_list_itrCreate (&(marker_entry->prev_tbl_entry_list));
	}
      }
      break;
    case DPE_CLASSIFIER_TABLE_ID:
	{
	    clfr_entry = (dpe_clfr_t*) g_dpe_clfr_tbl[inst_id];
	    if (clfr_entry->status == DPE_STATUS_INVALID) {
		clfr_entry->status = DPE_STATUS_VALID;
		itr = npf_list_itrCreate (&(clfr_entry->prev_tbl_entry_list));
	    }
	}
	break;
    } /* end switch on current table id to decide what params to
	 propagate */
    
    /************************************************************************
     * Now we check if the DPE has any upstream blocks associated with it.
     * If they are, and their parameters are complete and not propagated yet,
     * do so now.
     ***********************************************************************/
    
    if (npf_list_itrFirst(itr) != -1) {
	do {
	    QOS_DBG_PRINT (LOG_INFO, "Found a back ptr!\n");
	    entry = (dpe_EntryId_t*) npf_list_itrGetData(itr);
	    /* recursively call this function for this table */
	    dpe_propagate_status (entry->table_id, entry->instance_id,
				  status);
	}while (npf_list_itrNext(itr) != -1);
    }    
    npf_list_itrDelete (itr);
    PRETURN (0);
}

/**************************************************************************
*
*   Function: dpe_delete_entry
*
*   Parameters: IN: Interface,
*               IN: table_id
*               IN: instance_id
*               IN: dpe_entry - void*
*               IN: Next table id
*               IN: Next instance_id
* 
*   Description: deletes an entry from one of the DPE tables. This function
*                simply removes the QoS parameters/configurations from the
*                data plane and de-allocates memory created for the DPE entry
*                during the previous Add Entry call.
*
*   PRETURN : 0 on success, -1 for error 
*************************************************************************/

int dpe_delete_entry ( uint32_t             intf,	        
		       npf_ds_TableId_t     table_id,
		       npf_ds_InstanceId_t  inst_id)
{
    dpe_scheduler_t*     sched_entry;
    dpe_queue_t*         queue_entry;
    dpe_dropper_t*       dropper_entry;
    dpe_meter_t*         meter_entry;
    dpe_marker_t*        marker_entry;
    dpe_clfr_t*          clfr_entry;
    dpe_EntryId_t*       prev_tbl_entry;
    NPF_DATA             data;

    QOS_DBG_TRACE ();

    switch (table_id) {
    case DPE_SCHEDULER_TABLE_ID:
	{
	    if (g_dpe_scheduler_tbl[inst_id] != NULL) {
		sched_entry = (dpe_scheduler_t*) g_dpe_scheduler_tbl[inst_id];
		/* Call into the platform specific implementation to
		   remove the Scheduler */
		qos_delete_scheduler (sched_entry->flow_tuple, 
				      sched_entry->intf);
		/* Release the Queue id */
		release_queue_id_qos (intf, dropper_entry->flow_tuple.flow_id);
		/* remove this {tbl,instance} from the prev_tbl_entry_list
		   of the (next_tbl, next_inst}
		*/
		dpe_delete_inst_from_prev_entry_list 
		    (sched_entry->sched_tbl_id, sched_entry->sched_instance_id,
		     sched_entry->next_tbl_id, sched_entry->next_instance_id);
		
		while (npf_list_popFront 
		       (&(g_dpe_scheduler_tbl[inst_id]->prev_tbl_entry_list),
			(void**) &data) != -1) {
		    prev_tbl_entry = (dpe_EntryId_t*) data;
		    PIL_FreeMemory ((void*) prev_tbl_entry);
		}
		npf_list_destroy (&sched_entry->prev_tbl_entry_list);
		PIL_FreeMemory ((void*) sched_entry);
		g_dpe_scheduler_tbl[inst_id] = (p_dpe_scheduler_t) NULL;
	    }
	}
	break;
    case DPE_QUEUE_TABLE_ID:
	{
	    if (g_dpe_queue_tbl[inst_id] != NULL) {
		queue_entry = (dpe_queue_t*) g_dpe_queue_tbl[inst_id];
		/* Call into the platform specific implementation to
		   remove the Queue */
		qos_delete_queue (queue_entry->flow_tuple, queue_entry->intf);
		qos_delete_scheduler (queue_entry->flow_tuple, 
				      queue_entry->intf);
		
		/* remove this {tbl,instance} from the prev_tbl_entry_list
		   of the (next_tbl, next_inst}
		*/
		dpe_delete_inst_from_prev_entry_list 
		    (queue_entry->queue_tbl_id, queue_entry->queue_instance_id,
		     queue_entry->next_tbl_id, queue_entry->next_instance_id);
		
		while (npf_list_popFront 
		       (&(g_dpe_queue_tbl[inst_id]->prev_tbl_entry_list),
			(void**) &data) != -1) {
		    prev_tbl_entry = (dpe_EntryId_t*) data;
		    PIL_FreeMemory ((void*) prev_tbl_entry);
		}
		npf_list_destroy (&queue_entry->prev_tbl_entry_list);
		PIL_FreeMemory ((void*) queue_entry);
		g_dpe_queue_tbl[inst_id] = (p_dpe_queue_t) NULL;
	    }
	}
	break;
    case DPE_ALG_DROPPER_TABLE_ID:
	{
	    if (g_dpe_dropper_tbl[inst_id] != NULL) {
		dropper_entry = (dpe_dropper_t*) g_dpe_dropper_tbl[inst_id];
		/* Call into the platform specific implementation to
		   remove the dropper */
		qos_delete_dropper (dropper_entry->flow_tuple);
		
		/* Release the flow id */
		release_flow_id_qos (dropper_entry->flow_tuple.flow_id);
		/* remove this {tbl,instance} from the prev_tbl_entry_list
		   of the (next_tbl, next_inst}
		*/
		dpe_delete_inst_from_prev_entry_list 
		    (dropper_entry->dropper_tbl_id, 
		     dropper_entry->dropper_instance_id,
		     dropper_entry->next_tbl_id, 
		     dropper_entry->next_instance_id);
		while (npf_list_popFront 
		       (&(g_dpe_dropper_tbl[inst_id]->prev_tbl_entry_list),
			(void**) &data) != -1) {
		    prev_tbl_entry = (dpe_EntryId_t*) data;
		    PIL_FreeMemory ((void*) prev_tbl_entry);
		}
		npf_list_destroy (&dropper_entry->prev_tbl_entry_list);
		PIL_FreeMemory ((void*) dropper_entry);
		g_dpe_dropper_tbl[inst_id] = (p_dpe_dropper_t) NULL;
	    }
	}
	break;
    case DPE_METER_TABLE_ID:
	{
	    if (g_dpe_meter_tbl[inst_id] != NULL) {
		meter_entry = (dpe_meter_t*) g_dpe_meter_tbl[inst_id];
		/* Call into the platform specific implementation to
		   remove the meter */
		qos_delete_meter (meter_entry->flow_tuple,
				  meter_entry->failnext_flow_tuple);
		
		/* remove this {tbl,instance} from the prev_tbl_entry_list
		   of the (next_tbl, next_inst}
		*/
		dpe_delete_inst_from_prev_entry_list 
		    (meter_entry->meter_tbl_id, meter_entry->meter_instance_id,
		     meter_entry->next_tbl_id, meter_entry->next_instance_id);
		dpe_delete_inst_from_prev_entry_list 
		    (meter_entry->meter_tbl_id, meter_entry->meter_instance_id,
		     meter_entry->meter.failnexttableid, 
		     meter_entry->meter.failnextinstid);
		
		while (npf_list_popFront 
		       (&(g_dpe_meter_tbl[inst_id]->prev_tbl_entry_list),
			(void**) &data) != -1) {
		    prev_tbl_entry = (dpe_EntryId_t*) data;
		    PIL_FreeMemory ((void*) prev_tbl_entry);
		}
		npf_list_destroy (&meter_entry->prev_tbl_entry_list);
		PIL_FreeMemory ((void*) meter_entry);
		g_dpe_meter_tbl[inst_id] = (p_dpe_meter_t) NULL;
	    }
	}
	break;
    case DPE_MARKER_TABLE_ID:
	{
	    if (g_dpe_marker_tbl[inst_id] != NULL) {
		marker_entry = (dpe_marker_t*) g_dpe_marker_tbl[inst_id];
		
		/* Call into MPLS? */
		qos_delete_marker (marker_entry->flow_tuple);
		/* remove this {tbl,instance} from the prev_tbl_entry_list
		   of the (next_tbl, next_inst}
		*/
		dpe_delete_inst_from_prev_entry_list 
		    (marker_entry->marker_tbl_id, 
		     marker_entry->marker_instance_id,
		     marker_entry->next_tbl_id, marker_entry->next_instance_id);
		
		while (npf_list_popFront 
		       (&(g_dpe_marker_tbl[inst_id]->prev_tbl_entry_list),
			(void**) &data) != -1) {
		    prev_tbl_entry = (dpe_EntryId_t*) data;
		    PIL_FreeMemory ((void*) prev_tbl_entry);
		}
		npf_list_destroy (&marker_entry->prev_tbl_entry_list);
		PIL_FreeMemory ((void*) marker_entry);
		g_dpe_marker_tbl[inst_id] = (p_dpe_marker_t) NULL;
	    }
	}
	break;
    case DPE_CLASSIFIER_TABLE_ID:
	{
	    if (g_dpe_clfr_tbl[inst_id] != NULL) {
		clfr_entry = (dpe_clfr_t*) g_dpe_clfr_tbl[inst_id];
		/* Call into MPLS? */
		
		qos_delete_classifier (clfr_entry->filter_hdl);
		/* remove this {tbl,instance} from the prev_tbl_entry_list
		   of the (next_tbl, next_inst}
		*/
		dpe_delete_inst_from_prev_entry_list 
		    (clfr_entry->clfr_tbl_id, clfr_entry->clfr_instance_id,
		     clfr_entry->next_tbl_id, clfr_entry->next_instance_id);
		
		while (npf_list_popFront 
		       (&(g_dpe_clfr_tbl[inst_id]->prev_tbl_entry_list),
			(void**) &data) != -1) {
		    prev_tbl_entry = (dpe_EntryId_t*) data;
		    PIL_FreeMemory ((void*) prev_tbl_entry);
		}
		npf_list_destroy (&clfr_entry->prev_tbl_entry_list);
		PIL_FreeMemory ((void*) clfr_entry);
		g_dpe_clfr_tbl[inst_id] = (p_dpe_clfr_t) NULL;
	    }
	}
	break;
    } /* end switch on table_id */
    PRETURN (0);
}

/**************************************************************************
*
*   Function: dpe_mod_entry
*
*   Parameters: IN: Interface,
*               IN: table_id
*               IN: instance_id
*               IN: dpe_entry - void*
*               IN: Next table id
*               IN: Next instance_id
* 
*   Description: modifies an entry in one of the DPE tables. Modifying an entry
*                is simple. The current DPE entry's backward pointers (links to
*                upstream DPEs) are stored, and the entry is deleted, so that 
*                all QoS configuration parameters are removed from the data plane
*                Subsequently, the DPE entry is re-added. And the backward links
*                are restored. 
*                
*                A series of modify calls could potentially change the entire 
*                DPE path since potentially next table and next instance ids can
*                be changed. Hence, flow id/queue id could completely change.
*
*   PRETURN : 0 on success, -1 for error 
*************************************************************************/

int dpe_mod_entry ( uint32_t             intf,
		     npf_ds_TableId_t     table_id,
		     npf_ds_InstanceId_t  inst_id,
		     void*                dpe_entry,
		     npf_ds_TableId_t     next_tbl_id,
		     npf_ds_InstanceId_t  next_inst_id )
{
    dpe_scheduler_t*     sched_entry;
    dpe_queue_t*         queue_entry;
    dpe_dropper_t*       dropper_entry;
    dpe_meter_t*         meter_entry;
    dpe_marker_t*        marker_entry;
    dpe_clfr_t*          clfr_entry;
    dpe_EntryId_t*       entry;
    DList                back_links;
    DListItr*            itr;
    uint32_t             prev_intf;
    npf_ds_TableId_t     prev_next_tbl_id;
    npf_ds_InstanceId_t  prev_next_inst_id;
    int                  err = 0;
    
    QOS_DBG_TRACE ();

    /****************************************************************
     * - store the back links of the DPE that is to be modified
     * - delete the DPE entry
     * - add a new one with the new parameters
     * - add the back links again to the DPE
     ***************************************************************/

    npf_list_init (&back_links, PIL_FreeMemory);
    switch (table_id) {
    case DPE_SCHEDULER_TABLE_ID:
	{
	    sched_entry = (dpe_scheduler_t*) g_dpe_scheduler_tbl[inst_id];
	    prev_intf = sched_entry->intf;
	    prev_next_tbl_id = sched_entry->next_tbl_id;
	    prev_next_inst_id = sched_entry->next_instance_id;
	    itr = npf_list_itrCreate(&(sched_entry->prev_tbl_entry_list));
	}
	break;
    case DPE_QUEUE_TABLE_ID:
	{
	    queue_entry = (dpe_queue_t*) g_dpe_queue_tbl[inst_id];
	    prev_intf = queue_entry->intf;
	    prev_next_tbl_id = queue_entry->next_tbl_id;
	    prev_next_inst_id = queue_entry->next_instance_id;
	    itr = npf_list_itrCreate(&(queue_entry->prev_tbl_entry_list));
	}
	break;
    case DPE_ALG_DROPPER_TABLE_ID:
	{
	    dropper_entry = (dpe_dropper_t*) g_dpe_dropper_tbl[inst_id];
	    prev_intf = dropper_entry->intf;
	    prev_next_tbl_id = dropper_entry->next_tbl_id;
	    prev_next_inst_id = dropper_entry->next_instance_id;
	    itr = npf_list_itrCreate(&(dropper_entry->prev_tbl_entry_list));
	}
	break;
    case DPE_METER_TABLE_ID:
	{
	    meter_entry = (dpe_meter_t*) g_dpe_meter_tbl[inst_id];
	    prev_intf = intf;
	    prev_next_tbl_id = meter_entry->next_tbl_id;
	    prev_next_inst_id = meter_entry->next_instance_id;
	    itr = npf_list_itrCreate(&(meter_entry->prev_tbl_entry_list));
	}
	break;
    case DPE_MARKER_TABLE_ID:
	{
	    marker_entry = (dpe_marker_t*) g_dpe_marker_tbl[inst_id];
	    prev_intf = intf;
	    prev_next_tbl_id = marker_entry->next_tbl_id;
	    prev_next_inst_id = marker_entry->next_instance_id;
	    itr = npf_list_itrCreate(&(marker_entry->prev_tbl_entry_list));
	}
	break;
    case DPE_CLASSIFIER_TABLE_ID:
	{
	    clfr_entry = (dpe_clfr_t*) g_dpe_clfr_tbl[inst_id];
	    prev_intf = intf;
	    prev_next_tbl_id = clfr_entry->next_tbl_id;
	    prev_next_inst_id = clfr_entry->next_instance_id;
	    itr = npf_list_itrCreate(&(clfr_entry->prev_tbl_entry_list));
	}
	break;
    } /* end switch on table_id */
    if (npf_list_itrFirst(itr) != -1) {
	do {
	    entry = (dpe_EntryId_t*) npf_list_itrGetData(itr);
	    npf_list_pushFront (&back_links, entry);
	}while (npf_list_itrNext(itr) != -1);
    }
    npf_list_itrDelete (itr);

    /* now delete the DPE */
    dpe_delete_entry (prev_intf, table_id, inst_id);
    /* now add the DPE with the new parameters */
    dpe_add_entry (intf, table_id, inst_id, dpe_entry, next_tbl_id,
		   next_inst_id);
    /* now put back the back pointers to the upstream DPE */
    itr = npf_list_itrCreate (&back_links);
    if (npf_list_itrFirst(itr) != -1) {
	do {
	    entry = (dpe_EntryId_t*) npf_list_itrGetData(itr);
	    //npf_list_pushFront (p_list, entry);
	    dpe_add_inst_to_prev_entry_list (entry->table_id, 
					     entry->instance_id, table_id,
					     inst_id);
	}while (npf_list_itrNext(itr) != -1);
    }
    npf_list_itrDelete (itr);
    npf_list_destroy (&back_links);
    PRETURN (err);
}

/*************************************************************************
 *                                                                       *
 *                         DPE GET STATISTICS ROUTINES                   *
 *                                                                       *
 *************************************************************************/
/**************************************************************************
*
*   Function: dpe_get_clfr_stats
*
*   Parameters: IN: table_id
*               IN: instance_id
*               OUT:npf_ds_ClfrStats_t
* 
*   Description: Gets statistics counters from the specified clfr entry.
*
*   PRETURN : 0 on success, -1 on error 
*************************************************************************/

int dpe_get_clfr_stats (npf_ds_TableId_t     tbl_id,
			npf_ds_InstanceId_t  inst_id,
			npf_ds_ClfrStats_t*  clfr_stats)
{
    int                err = 0;
    dpe_clfr_t*        dpe_clfr;
    unsigned long long defaulted_pkt_count=0;
    unsigned long long pkt_count=0;
    unsigned long long dropped_pkt_count=0;
    unsigned long long defaulted_byte_count=0;
    unsigned long long byte_count=0;
    unsigned long long dropped_byte_count=0;

    QOS_DBG_TRACE();
    
    if (tbl_id != DPE_CLASSIFIER_TABLE_ID) {
	QOS_DBG_PRINT (LOG_ERR, "Invalid Table Id! \n");
	err = -1;
	PRETURN (err);
    }
    if (g_dpe_clfr_tbl[inst_id] == NULL){
	QOS_DBG_PRINT (LOG_ERR, "Invalid Instance Id! \n");
	err = -1;
	PRETURN (err);
    }
    dpe_clfr = (dpe_clfr_t*) g_dpe_clfr_tbl[inst_id];

    err = qos_get_clfr_stats (dpe_clfr->filter_hdl, &defaulted_pkt_count,
			      &pkt_count, &dropped_pkt_count, 
			      &defaulted_byte_count,
			      &byte_count, &dropped_byte_count);

    QOS_DBG_PRINT (LOG_INFO, "Clfr stats: \n\t"
		   "Defaulted pkt count = %d \n\t"
		   "packet count = %d \n\t"
		   "Dropped Pkt Count = %d \n\t"
		   "defaulted byte count = %d \n\t"
		   "byte count = %d \n\t"
		   "dropped byte count = %d \n\t", defaulted_pkt_count,
		   pkt_count, dropped_pkt_count, defaulted_byte_count,
		   byte_count, dropped_byte_count);

    clfr_stats->default_packets = defaulted_pkt_count;
    clfr_stats->default_bytes = defaulted_byte_count;
    clfr_stats->total_packets = pkt_count;
    clfr_stats->total_bytes = byte_count;

    PRETURN (err);
}

/**************************************************************************
*
*   Function: dpe_get_marker_stats
*
*   Parameters: IN: table_id
*               IN: instance_id
*               OUT:npf_ds_MarkerStats_t
* 
*   Description: Gets statistics counters from the specified marker
*
*   PRETURN : 0 on success, -1 on error 
*************************************************************************/

int dpe_get_marker_stats (npf_ds_TableId_t      tbl_id,
			  npf_ds_InstanceId_t   inst_id,
			  npf_ds_MarkerStats_t* marker_stats)
{
    int                err = 0;
    dpe_marker_t*      dpe_marker;
    unsigned long long pkts_marked=0;
    unsigned long long pkts_dropped=0, bytes_marked=0, bytes_dropped=0;

    QOS_DBG_TRACE();
    
    if (tbl_id != DPE_MARKER_TABLE_ID) {
	QOS_DBG_PRINT (LOG_ERR, "Invalid Table Id! \n");
	err = -1;
	PRETURN (err);
    }
    if (g_dpe_marker_tbl[inst_id] == NULL){
	QOS_DBG_PRINT (LOG_ERR, "Invalid Instance Id! \n");
	err = -1;
	PRETURN (err);
    }
    dpe_marker = (dpe_marker_t*) g_dpe_marker_tbl[inst_id];

    err = qos_get_marker_stats (&pkts_marked, &pkts_dropped, &bytes_marked,
				&bytes_dropped);
    
    QOS_DBG_PRINT (LOG_INFO, "Marker stats: \n\t"
		   "Num marked  = %ld \n\t"
		   "Num dropped = %ld \n\t", pkts_marked, pkts_dropped,
		   bytes_marked, bytes_dropped);
    
    marker_stats->packets_marked = pkts_marked;
    marker_stats->packets_dropped = pkts_dropped;
    marker_stats->bytes_marked = bytes_marked;
    marker_stats->bytes_dropped = bytes_dropped;

    PRETURN (err);
}


/**************************************************************************
*
*   Function: dpe_get_meter_stats
*
*   Parameters: IN: table_id
*               IN: instance_id
*               OUT:npf_ds_MeterStats_t
* 
*   Description: Gets statistics counters from the specified meter
*
*   PRETURN :  0 on success, -1 on error 
*************************************************************************/

int dpe_get_meter_stats (npf_ds_TableId_t     tbl_id,
			 npf_ds_InstanceId_t  inst_id,
			 npf_ds_MeterStats_t* meter_stats)
{
    int                err = 0;
    dpe_meter_t*       dpe_meter;
    unsigned long long conf_pkt_count=0, conf_byte_count=0;
    unsigned long long nonconf_pkt_count=0, nonconf_byte_count=0;

    QOS_DBG_TRACE();
    if (tbl_id != DPE_METER_TABLE_ID) {
	QOS_DBG_PRINT (LOG_ERR, "Invalid Table Id! \n");
	err = -1;
	PRETURN (err);
    }
    if (g_dpe_meter_tbl[inst_id] == NULL){
	QOS_DBG_PRINT (LOG_ERR, "Invalid Instance Id! \n");
	err = -1;
	PRETURN (err);
    }
    dpe_meter = (dpe_meter_t*) g_dpe_meter_tbl[inst_id];
    //dpe_dbg_print (dpe_Meter, (void*) dpe_meter);
    
    err = qos_get_meter_stats (dpe_meter->flow_tuple.flow_id,
			       &conf_pkt_count, &conf_byte_count,
			       &nonconf_pkt_count, &nonconf_byte_count);

    meter_stats->conf_packet_count = (npf_ds_counter_t) conf_pkt_count;
    meter_stats->conf_byte_count = (npf_ds_counter_t) conf_byte_count;
    meter_stats->nonconf_packet_count = (npf_ds_counter_t) nonconf_pkt_count;
    meter_stats->nonconf_byte_count = (npf_ds_counter_t) nonconf_byte_count;

    QOS_DBG_PRINT (LOG_INFO, "Meter stats: \n Conf Pkt Count = %d \t"
		   "Conf Byte count = %d\n"
		   "Non Conf pkt count = %d \t Non conf byte count = %d\n",
		   conf_pkt_count, conf_byte_count, nonconf_pkt_count,
		   nonconf_byte_count);

    PRETURN (err);
}

/**************************************************************************
*
*   Function: dpe_get_dropper_stats
*
*   Parameters: IN: table_id
*               IN: instance_id
*               OUT:npf_ds_DropperStats_t
* 
*   Description: Gets statistics counters from the specified dropper
*
*   PRETURN :  0 on success, -1 on error
*************************************************************************/

int dpe_get_dropper_stats (npf_ds_TableId_t     tbl_id,
			   npf_ds_InstanceId_t  inst_id,
			   npf_ds_DropperStats_t* flow_stats)
{
    int                err = 0;
    dpe_dropper_t*     dpe_dropper;
    unsigned long long pkts_dropped=0, pkts_consumed=0;

    QOS_DBG_TRACE();
    if (tbl_id != DPE_ALG_DROPPER_TABLE_ID) {
	QOS_DBG_PRINT (LOG_ERR, "Invalid Table Id! \n");
	err = -1;
	PRETURN (err);
    }
    if (g_dpe_dropper_tbl[inst_id] == NULL){
	QOS_DBG_PRINT (LOG_ERR, "Invalid Instance Id! \n");
	err = -1;
	PRETURN (err);
    }
    dpe_dropper = (dpe_dropper_t*) g_dpe_dropper_tbl[inst_id];
    //dpe_dbg_print (dpe_AlgDropper, (void*) dpe_dropper);
    
    err = qos_get_dropper_stats (dpe_dropper->flow_tuple.flow_id,
				 &pkts_dropped, &pkts_consumed);

    flow_stats->packets_dropped = (npf_ds_counter_t) pkts_dropped;
    flow_stats->packets_consumed = (npf_ds_counter_t) pkts_consumed;
    QOS_DBG_PRINT (LOG_INFO, "Dropper stats: Pkts Dropped = %d \t"
		   "Pkts consumed = %d\n", pkts_dropped, pkts_consumed);

    PRETURN (err);
}

/**************************************************************************
*
*   Function: dpe_get_queue_stats
*
*   Parameters: IN: table_id
*               IN: instance_id
*               OUT:npf_ds_QueueStats_t
* 
*   Description: Gets statistics counters from the specified queue
*
*   PRETURN : 0 on success, -1 on error
*************************************************************************/

int dpe_get_queue_stats (npf_ds_TableId_t     tbl_id,
			 npf_ds_InstanceId_t  inst_id,
			 npf_ds_QueueStats_t* q_stats)
{
    int                err = 0;
    dpe_queue_t*       dpe_q;
    unsigned long long pkts_dropped=0, pkts_consumed=0;

    QOS_DBG_TRACE();
    if (tbl_id != DPE_QUEUE_TABLE_ID) {
	QOS_DBG_PRINT (LOG_ERR, "Invalid Table Id! \n");
	err = -1;
	PRETURN (err);
    }
    if (g_dpe_queue_tbl[inst_id] == NULL){
	QOS_DBG_PRINT (LOG_ERR, "Invalid Instance Id! \n");
	err = -1;
	PRETURN (err);
    }
    dpe_q = (dpe_queue_t*) g_dpe_queue_tbl[inst_id];
    //dpe_dbg_print (dpe_Queue, (void*) dpe_q);
    
    err = qos_get_queue_stats (dpe_q->intf, /* Queue Group Id */
			       dpe_q->flow_tuple.queue_id,
			       &pkts_dropped, &pkts_consumed);

    q_stats->packet_count = (npf_ds_counter_t) pkts_dropped;
    q_stats->byte_count = (npf_ds_counter_t) pkts_consumed;
    QOS_DBG_PRINT (LOG_INFO, "Queue stats: Pkts Dropped = %d \t"
		   "Pkts consumed = %d\n", pkts_dropped, pkts_consumed);

    PRETURN (err);
}

/**************************************************************************
*
*   Function: dpe_get_scheduler_stats
*
*   Parameters: IN: table_id
*               IN: instance_id
*               OUT:npf_ds_SchedulerStats_t
* 
*   Description: Gets statistics counters from the specified scheduler
*
*   PRETURN :  0 on success, -1 on error 
*************************************************************************/

int dpe_get_scheduler_stats (npf_ds_TableId_t         tbl_id,
			     npf_ds_InstanceId_t      inst_id,
			     npf_ds_SchedulerStats_t* sched_stats)
{
    int                err = 0;
    dpe_scheduler_t*   dpe_sched;
    unsigned long long pkt_cnt=0, byte_cnt=0;

    QOS_DBG_TRACE();
    if (tbl_id != DPE_SCHEDULER_TABLE_ID) {
	QOS_DBG_PRINT (LOG_ERR, "Invalid Table Id! \n");
	err = -1;
	PRETURN (err);
    }
    if (g_dpe_scheduler_tbl[inst_id] == NULL){
	QOS_DBG_PRINT (LOG_ERR, "Invalid Instance Id! \n");
	err = -1;
	PRETURN (err);
    }
    dpe_sched = (dpe_scheduler_t*) g_dpe_scheduler_tbl[inst_id];
    //dpe_dbg_print (dpe_Scheduler, (void*) dpe_sched);
    
    err = qos_get_scheduler_stats (dpe_sched->intf, /* Queue Group Id */
				   dpe_sched->flow_tuple.queue_id,
				   &pkt_cnt, &byte_cnt);

    sched_stats->packet_count = (npf_ds_counter_t) pkt_cnt;
    sched_stats->byte_count = (npf_ds_counter_t) byte_cnt;
    QOS_DBG_PRINT (LOG_INFO, "Scheduler stats: Pkt Cnt = %d \t"
		   "Byte Cnt = %d\n", pkt_cnt, byte_cnt);

    PRETURN (err);
}


/*************************************************************************
 *                                                                       *
 *              DPE TABLE MANAGEMENT ROUTINES                            *
 *                                                                       *
 *************************************************************************/
/**************************************************************************
*
*   Function: dpe_create_table_entry
*
*   Parameters: IN: Interface,
*               IN: table_id
*               IN: instance_id
*               IN: dpe_entry - void*
*               IN: Next table id
*               IN: Next instance_id
*               IN: Flow Tuple (flow_id, queue_id, next_blk_id)
* 
*   Description: creates a new entry in one of the DPE tables. This function
*                allocates memory for the new DPE entry & copies the passed-in
*                parameters.
*                As a special case of the METER DPE, this function allocates
*                a new flow id for upstream blocks (since flows can aggregate
*                at the meter. It also uses the fail next table and fail next
*                instance to set the meter's actions.
*
*   PRETURN : 0 on success, -1 for error 
*************************************************************************/

int dpe_create_table_entry (uint32_t             intf,
			    npf_ds_TableId_t     table_id,
			    npf_ds_InstanceId_t  inst_id,
			    void*                dpe_entry,
			    npf_ds_TableId_t     next_tbl_id,
			    npf_ds_InstanceId_t  next_inst_id,
			    dpe_flow_tuple_t     flow_tuple,
			    uint32_t             status)
{
    npf_ds_Scheduler_t*  ds_scheduler;
    npf_ds_Q_t*          ds_queue;
    npf_ds_AlgDrop_t*    ds_dropper;
    npf_ds_Meter_t*      ds_meter;
    npf_ds_Action_t*     ds_marker;
    npf_ds_BaseFilter_t* ds_clfr;
    dpe_scheduler_t*     sched_entry;
    dpe_queue_t*         queue_entry;
    dpe_dropper_t*       dropper_entry;
    dpe_meter_t*         meter_entry;
    dpe_marker_t*        marker_entry;
    dpe_clfr_t*          clfr_entry;
    dpe_EntryId_t*       prev_tbl_entry;
    uint32_t             tbl_id, instance_id, failnext_tbl_id, failnext_inst_id;
    dpe_flow_tuple_t     failnext_flow_tuple;
    uint32_t             upstream_flow_id;
    int                  err;

    QOS_DBG_TRACE ();

    tbl_id = (uint32_t) table_id;
    instance_id = (uint32_t) inst_id;

    QOS_DBG_PRINT (LOG_INFO, "Table Id = %d \t Instance Id = %d \n\t",
		   tbl_id, instance_id);
    switch (table_id) {
    case DPE_SCHEDULER_TABLE_ID:
	{
	    ds_scheduler = (npf_ds_Scheduler_t*) dpe_entry;
	    
	    /* allocate memory for the Scheduler table entry */
	    if (g_dpe_scheduler_tbl[instance_id] == NULL) {
		QOS_DBG_PRINT (LOG_INFO, "Scheduler entry is NULL!\n");
		g_dpe_scheduler_tbl[instance_id] = (dpe_scheduler_t*) 
		    PIL_AllocateMemory (sizeof (dpe_scheduler_t),
					PIL_MEM_ZERO_INIT);
		npf_list_init 
		    (&(g_dpe_scheduler_tbl[instance_id]->prev_tbl_entry_list), 
		     PIL_FreeMemory);
		g_dpe_scheduler_tbl[instance_id]->prev_tbl_count = 0;
	    }
	    /* copy all scheduler parameters into the new entry created */
	    sched_entry = (dpe_scheduler_t*) (g_dpe_scheduler_tbl[instance_id]);
	    sched_entry->status = status;
	    sched_entry->intf = intf;
	    sched_entry->sched_tbl_id = table_id;
	    sched_entry->sched_instance_id = inst_id;
	    /* Copy individual parameters from npf_ds_Scheduler_t */
	    dpe_copy_sched_params (&sched_entry->scheduler, ds_scheduler);

	    sched_entry->next_tbl_id = next_tbl_id;
	    sched_entry->next_instance_id = next_inst_id;
	    sched_entry->flow_tuple.flow_id = flow_tuple.flow_id;
	    sched_entry->flow_tuple.queue_id = flow_tuple.queue_id;
	    sched_entry->flow_tuple.next_block_id = DPE_NULL_TABLE_ID;
	}
	break;
    case DPE_QUEUE_TABLE_ID:
	{
	    /* now store all the scheduler parameters into the Queue table 
	     */
	    ds_queue = (npf_ds_Q_t*) dpe_entry;
	    
	    /* allocate memory for the Queue table entry */
	    if (g_dpe_queue_tbl[instance_id] == NULL) {
		QOS_DBG_PRINT (LOG_INFO, "Queue entry is NULL!\n");
		g_dpe_queue_tbl[instance_id] = (dpe_queue_t*) 
		    PIL_AllocateMemory (sizeof (dpe_queue_t),
					PIL_MEM_ZERO_INIT);
		npf_list_init 
		    (&(g_dpe_queue_tbl[instance_id]->prev_tbl_entry_list), 
		     PIL_FreeMemory);
		g_dpe_queue_tbl[instance_id]->prev_tbl_count = 0;
	    }
	    /* copy all queue parameters into the new entry created */
	    queue_entry = (dpe_queue_t*) (g_dpe_queue_tbl[instance_id]);
	    queue_entry->status = status;
	    queue_entry->intf = intf;
	    queue_entry->queue_tbl_id = table_id;
	    queue_entry->queue_instance_id = inst_id;

	    dpe_copy_queue_params (&queue_entry->queue, ds_queue);
 
	    queue_entry->next_tbl_id = next_tbl_id;
	    queue_entry->next_instance_id = next_inst_id;
	    queue_entry->flow_tuple.flow_id = flow_tuple.flow_id;
	    queue_entry->flow_tuple.queue_id = flow_tuple.queue_id;
	    queue_entry->flow_tuple.next_block_id = next_tbl_id;
	}
	break;
    case DPE_ALG_DROPPER_TABLE_ID:
	{
	    /* now store all the dropper parameters into the dropper table 
	     */
	    ds_dropper = (npf_ds_AlgDrop_t*) dpe_entry;
	    
	    /* allocate memory for the dropper table entry */
	    if (g_dpe_dropper_tbl[instance_id] == NULL) {
		QOS_DBG_PRINT (LOG_INFO, "dropper entry is NULL!\n");
		g_dpe_dropper_tbl[instance_id] = (dpe_dropper_t*) 
		    PIL_AllocateMemory (sizeof (dpe_dropper_t),
					PIL_MEM_ZERO_INIT);
		npf_list_init 
		    (&(g_dpe_dropper_tbl[instance_id]->prev_tbl_entry_list), 
		     PIL_FreeMemory);
		g_dpe_dropper_tbl[instance_id]->prev_tbl_count = 0;
	    }
	    /* copy all dropper parameters into the new entry created */
	    dropper_entry = (dpe_dropper_t*) (g_dpe_dropper_tbl[instance_id]);
	    dropper_entry->status = status;
	    dropper_entry->intf  = intf;
	    dropper_entry->dropper_tbl_id = table_id;
	    dropper_entry->dropper_instance_id = inst_id;
	    
	    dpe_copy_dropper_params (&dropper_entry->dropper, ds_dropper);

	    dropper_entry->next_tbl_id = next_tbl_id;
	    dropper_entry->next_instance_id = next_inst_id;
	    dropper_entry->flow_tuple.flow_id = flow_tuple.flow_id;
	    dropper_entry->flow_tuple.queue_id = flow_tuple.queue_id;
	    dropper_entry->flow_tuple.next_block_id = next_tbl_id;
	}
	break;
    case DPE_MARKER_TABLE_ID:
	{
	    /* now store all the marker parameters into the Marker table 
	     */
	    ds_marker = (npf_ds_Action_t*) dpe_entry;
	    
	    /* allocate memory for the Marker table entry */
	    if (g_dpe_marker_tbl[instance_id] == NULL) {
		QOS_DBG_PRINT (LOG_ERR, "Marker entry is NULL!\n");
		g_dpe_marker_tbl[instance_id] = (dpe_marker_t*) 
		    PIL_AllocateMemory (sizeof (dpe_marker_t),
					PIL_MEM_ZERO_INIT);
		npf_list_init 
		    (&(g_dpe_marker_tbl[instance_id]->prev_tbl_entry_list), 
		     PIL_FreeMemory);
		g_dpe_marker_tbl[instance_id]->prev_tbl_count = 0;
	    }
	    /* copy all marker parameters into the new entry created */
	    marker_entry = (dpe_marker_t*) (g_dpe_marker_tbl[instance_id]);
	    marker_entry->status = status;
	    marker_entry->marker_tbl_id = table_id;
	    marker_entry->marker_instance_id = inst_id;
	    
    	    QOS_DBG_PRINT (LOG_INFO, "Marker->Specific = 0x%x \n\t"
		   "*marker->Specific = 0x%x \n\t"
		   " Dscp = 0x%x \n\t",
		   ds_marker->Specific, *ds_marker->Specific,
		   ((npf_ds_DscpMarkAct_t*)ds_marker->Specific)->Dscp);

	    dpe_copy_marker_params (&marker_entry->action, ds_marker);

	    marker_entry->next_tbl_id = next_tbl_id;
	    marker_entry->next_instance_id = next_inst_id;
	    marker_entry->flow_tuple.flow_id = flow_tuple.flow_id;
	    marker_entry->flow_tuple.queue_id = flow_tuple.queue_id;
	    marker_entry->flow_tuple.next_block_id = next_tbl_id;
	}
	break;
    case DPE_METER_TABLE_ID:
	{
	    /* now store all the meter parameters into the Meter table 
	     */
	    ds_meter = (npf_ds_Meter_t*) dpe_entry;
	    
	    /* allocate memory for the meter table entry */
	    if (g_dpe_meter_tbl[instance_id] == NULL) {
		QOS_DBG_PRINT (LOG_INFO, "Meter entry is NULL!\n");
		g_dpe_meter_tbl[instance_id] = (dpe_meter_t*) 
		    PIL_AllocateMemory (sizeof (dpe_meter_t),
					PIL_MEM_ZERO_INIT);
		npf_list_init 
		    (&(g_dpe_meter_tbl[instance_id]->prev_tbl_entry_list), 
		     PIL_FreeMemory);
		g_dpe_meter_tbl[instance_id]->prev_tbl_count = 0;
	    }
	    /* copy all meter parameters into the new entry created */
	    meter_entry = (dpe_meter_t*) (g_dpe_meter_tbl[instance_id]);
	    meter_entry->status = status;
	    meter_entry->meter_tbl_id = table_id;
	    meter_entry->meter_instance_id = inst_id;

	    dpe_copy_meter_params (&meter_entry->meter, ds_meter);

	    meter_entry->next_tbl_id = next_tbl_id;
	    meter_entry->next_instance_id = next_inst_id;

	    /**************************************************************
	     * In order to be able to aggregate flows at the meter, we have
	     * to a bit of a trick here. Essentially, the {fid, qid} we have
	     * got now from the {next_tbl, next_inst} has to be preserved.
	     * What we do for aggregating upstream blocks is this
	     * - generate a new flow_id. 
	     * - Use the same Queue id that we got from the next DPE 
	     * - the new Flow ID will be given to any upstream blocks like 
	     * Clfr
	     * - Set up the meter so that it marks incoming packets in this 
	     * way
	     *    - it accepts flows with the upstream flow id
	     *    - it changes the flow id to the downstream flow id when it
	     *      leaves the meter
	     * - we do this for both conforming and non conforming traffic
	     **************************************************************/

	    /* store the downstream flow_id, queue_id for conforming traffic */
	    meter_entry->downstream_flow_tuple.flow_id = flow_tuple.flow_id;
	    meter_entry->downstream_flow_tuple.queue_id = flow_tuple.queue_id;
	    meter_entry->downstream_flow_tuple.next_block_id = next_tbl_id;

	    /* generate a new flow_id for upstream traffic */
	    err = allocate_flow_id_qos (&upstream_flow_id);
	    QOS_DBG_PRINT (LOG_INFO, "METER: New Flow id for upstream"
			   "blocks = %d \n", upstream_flow_id);
	    /* set up the upstream flow tuple */
	    meter_entry->flow_tuple.flow_id = upstream_flow_id; /* new flow id */
	    /* queue id and next tbl id remains the same */
	    meter_entry->flow_tuple.queue_id = flow_tuple.queue_id;
	    meter_entry->flow_tuple.next_block_id = next_tbl_id;

	    /* Now look at the FailNext Table Id for the meter */
	    QOS_DBG_PRINT (LOG_INFO, "Fail Next Tbl for Meter = %d \t"
			   "Fail Next Instance Id = %d \n",
			   (meter_entry->meter).failnexttableid,
			   (meter_entry->meter).failnextinstid);
	    failnext_tbl_id = (uint32_t) (meter_entry->meter).failnexttableid;
	    failnext_inst_id = (uint32_t) (meter_entry->meter).failnextinstid;
	    
	    /* - Get the flow id and queue id from the next tbl and next inst
	       - add a previous table entry to the next tbl and next inst
	    
	       Add back pointer for fail next DPE also */
	    dpe_add_inst_to_prev_entry_list (table_id, inst_id,
					     failnext_tbl_id, failnext_inst_id);
	    dpe_get_downstream_flow_tuple (failnext_tbl_id,
					   failnext_inst_id,
					   &meter_entry->failnext_flow_tuple);
	} /* end (METER) */
	break;
    case DPE_CLASSIFIER_TABLE_ID:
	{
	    /* now store all the classifier parameters into the Clfr table 
	     */
	    ds_clfr = (npf_ds_BaseFilter_t*) dpe_entry;
	    
	    /* allocate memory for the Clfr table entry */
	    if (g_dpe_clfr_tbl[instance_id] == NULL) {
		QOS_DBG_PRINT (LOG_INFO, "Clfr entry is NULL!\n");
		g_dpe_clfr_tbl[instance_id] = (dpe_clfr_t*) 
		    PIL_AllocateMemory (sizeof (dpe_clfr_t),
					PIL_MEM_ZERO_INIT);
		npf_list_init 
		    (&(g_dpe_clfr_tbl[instance_id]->prev_tbl_entry_list), 
		     PIL_FreeMemory);
		g_dpe_clfr_tbl[instance_id]->prev_tbl_count = 0;
	    }
	    /* copy all clfr parameters into the new entry created */
	    clfr_entry = (dpe_clfr_t*) (g_dpe_clfr_tbl[instance_id]);
	    clfr_entry->status = status;
	    clfr_entry->clfr_tbl_id = table_id;
	    clfr_entry->clfr_instance_id = inst_id;
	    /* need to allocate memory for the  filter */
	 
            if (ds_clfr->type == npf_ds_IPv4Filter) {
		clfr_entry->filter = (npf_ds_BaseFilter_t*) 
		  PIL_AllocateMemory (sizeof (npf_ds_IPv4Filter_t), 
				      PIL_MEM_ZERO_INIT);
		if (!clfr_entry->filter) {
			QOS_DBG_PRINT (LOG_ERR, "Could not alloc clfr!\n");
			exit (1);
	        }
           }
	    else if (ds_clfr->type == npf_ds_IPv6Filter) {
		clfr_entry->filter = (npf_ds_BaseFilter_t*) 
		  PIL_AllocateMemory (sizeof (npf_ds_IPv6Filter_t), 
				      PIL_MEM_ZERO_INIT);
		if (!clfr_entry->filter) {
			QOS_DBG_PRINT (LOG_ERR, "Could not alloc clfr!\n");
			exit (1);
	        }
            }
	    else if (ds_clfr->type == npf_ds_MPLSFilter) {
		clfr_entry->filter = (npf_ds_BaseFilter_t*) 
		  PIL_AllocateMemory (sizeof (npf_ds_MPLSFilter_t), 
				      PIL_MEM_ZERO_INIT);
		if (!clfr_entry->filter) {
			QOS_DBG_PRINT (LOG_ERR, "Could not alloc clfr!\n");
			exit (1);
	        }
           }
	    dpe_copy_clfr_params (clfr_entry->filter, ds_clfr);

	    clfr_entry->next_tbl_id = next_tbl_id;
	    clfr_entry->next_instance_id = next_inst_id;
	    clfr_entry->flow_tuple.flow_id = flow_tuple.flow_id;
	    clfr_entry->flow_tuple.queue_id = flow_tuple.queue_id;
	    clfr_entry->flow_tuple.next_block_id = next_tbl_id;
	}
	break;
    } /* end switch (table_id) to add an entry */

    PRETURN (0);
}

/**************************************************************************
*
*   Function: dpe_get_downstream_flow_tuple
*
*   Parameters: IN: Next table id
*               IN: Next instance_id
*               OUT: Flow tuple from {next tbl, next inst}
*
*   Description: copies the {flowid, queueid} from downstream DPE
*
*   PRETURN : 0 on success, -1 for error 
*************************************************************************/
int dpe_get_downstream_flow_tuple (npf_ds_TableId_t     next_tbl_id,
				   npf_ds_InstanceId_t  next_inst_id,
				   dpe_flow_tuple_t*     flow_tuple)
{
    int err = 0;

    QOS_DBG_TRACE();

    switch (next_tbl_id) {
    case DPE_SCHEDULER_TABLE_ID:
	{
	    /* check if instance id corresponds to a valid entry */
	    if (g_dpe_scheduler_tbl[next_inst_id] == NULL)
		{
		    flow_tuple->flow_id = g_max_flow_id + 1;
		    flow_tuple->queue_id = g_max_queue_id + 1;
		    flow_tuple->next_block_id = DPE_NULL_TABLE_ID;
		    QOS_DBG_PRINT (LOG_INFO,"Next Instance does not have a valid"
				   "entry\n");
		}
	    else {
		/* copy Flow ID, Queue ID from the next DPE block */
		flow_tuple->flow_id = 
		    g_dpe_scheduler_tbl[next_inst_id]->flow_tuple.flow_id;
		flow_tuple->queue_id = 
		    g_dpe_scheduler_tbl[next_inst_id]->flow_tuple.queue_id;
		flow_tuple->next_block_id = next_tbl_id;
	    }
	}
	break;
    case DPE_QUEUE_TABLE_ID:
	{
	    /* check if instance id corresponds to a valid entry */
	    if (g_dpe_queue_tbl[next_inst_id] == NULL)
		{
		    flow_tuple->flow_id = g_max_flow_id + 1;
		    flow_tuple->queue_id = g_max_queue_id + 1;
		    flow_tuple->next_block_id = DPE_NULL_TABLE_ID;
		    QOS_DBG_PRINT (LOG_INFO,"Next Instance does not have a valid"
				   "entry\n");
		}
	    else {
		/* copy Flow ID, Queue ID from the next DPE block */
		flow_tuple->flow_id = 
		    g_dpe_queue_tbl[next_inst_id]->flow_tuple.flow_id;
		flow_tuple->queue_id = 
		    g_dpe_queue_tbl[next_inst_id]->flow_tuple.queue_id;
		flow_tuple->next_block_id = next_tbl_id;
	    }
	}
	break;
    case DPE_ALG_DROPPER_TABLE_ID:
	{
	    /* check if instance id corresponds to a valid entry */
	    if (g_dpe_dropper_tbl[next_inst_id] == NULL)
		{
		    flow_tuple->flow_id = g_max_flow_id + 1;
		    flow_tuple->queue_id = g_max_queue_id + 1;
		    flow_tuple->next_block_id = DPE_NULL_TABLE_ID;
		    QOS_DBG_PRINT (LOG_INFO,"Next Instance does not have a valid"
				   "entry\n");
		}
	    else {
		/* copy Flow ID, Queue ID from the next DPE block */
		flow_tuple->flow_id = 
		    g_dpe_dropper_tbl[next_inst_id]->flow_tuple.flow_id;
		flow_tuple->queue_id = 
		    g_dpe_dropper_tbl[next_inst_id]->flow_tuple.queue_id;
		flow_tuple->next_block_id = next_tbl_id;    
	    }
	}
	break;
    case DPE_MARKER_TABLE_ID:
	{
	    /* check if instance id corresponds to a valid entry */
	    if (g_dpe_marker_tbl[next_inst_id] == NULL)
		{
		    flow_tuple->flow_id = g_max_flow_id + 1;
		    flow_tuple->queue_id = g_max_queue_id + 1;
		    flow_tuple->next_block_id = DPE_NULL_TABLE_ID;
		    QOS_DBG_PRINT (LOG_INFO,"Next Instance does not have a valid"
				   "entry\n");
		}
	    else {
		/* copy Flow ID, Queue ID from the next DPE block */
		flow_tuple->flow_id = 
		    g_dpe_marker_tbl[next_inst_id]->flow_tuple.flow_id;
		flow_tuple->queue_id = 
		    g_dpe_marker_tbl[next_inst_id]->flow_tuple.queue_id;
		flow_tuple->next_block_id = next_tbl_id;
	    }
	}
	break;
    case DPE_METER_TABLE_ID:
	{
	    /* check if instance id corresponds to a valid entry */
	    if (g_dpe_meter_tbl[next_inst_id] == NULL)
		{
		    flow_tuple->flow_id = g_max_flow_id + 1;
		    flow_tuple->queue_id = g_max_queue_id + 1;
		    flow_tuple->next_block_id = 0;
		    QOS_DBG_PRINT (LOG_INFO,"Next Instance does not have a valid"
				   "entry\n");
		}
	    else {
		/* copy Flow ID, Queue ID from the next DPE block */
		flow_tuple->flow_id = 
		    g_dpe_meter_tbl[next_inst_id]->flow_tuple.flow_id;
		flow_tuple->queue_id = 
		    g_dpe_meter_tbl[next_inst_id]->flow_tuple.queue_id;
		flow_tuple->next_block_id = next_tbl_id;
	    }
	}
	break;
    case DPE_CLASSIFIER_TABLE_ID:
	{
	    /* check if instance id corresponds to a valid entry */
	    if (g_dpe_clfr_tbl[next_inst_id] == NULL)
		{
		    flow_tuple->flow_id = g_max_flow_id + 1;
		    flow_tuple->queue_id = g_max_queue_id + 1;
		    flow_tuple->next_block_id = 0;
		    QOS_DBG_PRINT (LOG_INFO,"Next Instance does not have a valid"
				   "entry\n");
		}
	    else {
		/* copy Flow ID, Queue ID from the next DPE block */
		flow_tuple->flow_id = 
		    g_dpe_clfr_tbl[next_inst_id]->flow_tuple.flow_id;
		flow_tuple->queue_id = 
		    g_dpe_clfr_tbl[next_inst_id]->flow_tuple.queue_id;
		flow_tuple->next_block_id = next_tbl_id;
	    }
	}
	break;
    }/* end (switch on next_tbl_id) */    
    PRETURN (err);
}

/**************************************************************************
*
*   Function: dpe_propagate_flow_tuple
*
*   Parameters: IN: table_id
*               IN: instance_id
* 
*   Description: propagates the Flow Tuple to all upstream blocks. This is a
*                recursive function that updates the flow tuple of all 
*                upstream DPEs. 
*
*   PRETURN : 0 on success, -1 for error 
*************************************************************************/

int dpe_propagate_flow_tuple (npf_ds_TableId_t     table_id,
			      npf_ds_InstanceId_t  inst_id)
{
    int               err = 0;
    dpe_EntryId_t*    entry;
    DListItr*         itr;
    dpe_flow_tuple_t* p_flow_tuple;

    switch (table_id) {
    case DPE_SCHEDULER_TABLE_ID:
	{
	    if(g_dpe_scheduler_tbl[inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_ERR, "Entry is NULL\n");
		PRETURN (-1);
	    }
	    else {
		QOS_DBG_PRINT (LOG_INFO, "Propagating Flow Tuple from Sched\n");
		p_flow_tuple = &(g_dpe_scheduler_tbl[inst_id]->flow_tuple);
		itr = npf_list_itrCreate 
		    (&(g_dpe_scheduler_tbl[inst_id]->prev_tbl_entry_list));
		if (npf_list_itrFirst(itr) != -1) {
		    do {
			entry = (dpe_EntryId_t*) npf_list_itrGetData(itr);
			/* update the corr. table & instance's Flow Tuple */
			dpe_update_flow_tuple (entry->table_id, 
					       entry->instance_id,
					       table_id, 
					       inst_id,
					       p_flow_tuple);
		    }while (npf_list_itrNext(itr) != -1);
		}/* end if iterator was successfully created */
		/* now we need to recursively propagate the Flow tuple to all
		   upstream DPEs
		*/
		itr = npf_list_itrCreate 
		    (&(g_dpe_scheduler_tbl[inst_id]->prev_tbl_entry_list));
	    } /* end if (entry was non NULL) */
	}
	break;
    case DPE_QUEUE_TABLE_ID:
	{
	    if(g_dpe_queue_tbl[inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_ERR, "Entry is NULL\n");
		PRETURN (-1);
	    }
	    else {
		QOS_DBG_PRINT (LOG_INFO, "Propagating Flow Tuple from Queue\n");
		p_flow_tuple = &(g_dpe_queue_tbl[inst_id]->flow_tuple);
		itr = npf_list_itrCreate 
		    (&(g_dpe_queue_tbl[inst_id]->prev_tbl_entry_list));
		if (npf_list_itrFirst(itr) != -1) {
		    do {
			entry = (dpe_EntryId_t*) npf_list_itrGetData(itr);
			/* update the corr. table & instance's Flow Tuple */
			dpe_update_flow_tuple (entry->table_id, 
					       entry->instance_id, 
					       table_id, 
					       inst_id,
					       p_flow_tuple);
		    }while (npf_list_itrNext(itr) != -1);
		}/* end if iterator was successfully created */
		/* now we need to recursively propagate the Flow tuple to all
		   upstream DPEs
		*/
		itr = npf_list_itrCreate 
		    (&(g_dpe_queue_tbl[inst_id]->prev_tbl_entry_list));
	    } /* end if (entry was non NULL) */
	}
	break;
    case DPE_ALG_DROPPER_TABLE_ID:
	{
	    if(g_dpe_dropper_tbl[inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_ERR, "Entry is NULL\n");
		PRETURN (-1);
	    }
	    else {
		QOS_DBG_PRINT (LOG_INFO, "Propagating Flow Tuple from Dropper\n");
		p_flow_tuple = &(g_dpe_dropper_tbl[inst_id]->flow_tuple);
		itr = npf_list_itrCreate 
		    (&(g_dpe_dropper_tbl[inst_id]->prev_tbl_entry_list));
		if (npf_list_itrFirst(itr) != -1) {
		    do {
			entry = (dpe_EntryId_t*) npf_list_itrGetData(itr);
			/* update the corr. table & instance's Flow Tuple */
			dpe_update_flow_tuple (entry->table_id, 
					       entry->instance_id, 
					       table_id, 
					       inst_id,
					       p_flow_tuple);
		    }while (npf_list_itrNext(itr) != -1);
		}/* end if iterator was successfully created */
		/* now we need to recursively propagate the Flow tuple to all
		   upstream DPEs
		*/
		itr = npf_list_itrCreate 
		    (&(g_dpe_dropper_tbl[inst_id]->prev_tbl_entry_list));
	    } /* end if (entry was non NULL) */
	}
	break;
    case DPE_METER_TABLE_ID:
	{
	    if(g_dpe_meter_tbl[inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_ERR, "Entry is NULL\n");
		PRETURN (-1);
	    }
	    else {
		QOS_DBG_PRINT (LOG_INFO, "Propagating Flow Tuple from Meter\n");
		p_flow_tuple = &(g_dpe_meter_tbl[inst_id]->flow_tuple);
		itr = npf_list_itrCreate 
		    (&(g_dpe_meter_tbl[inst_id]->prev_tbl_entry_list));
		if (npf_list_itrFirst(itr) != -1) {
		    do {
			entry = (dpe_EntryId_t*) npf_list_itrGetData(itr);
			/* update the corr. table & instance's Flow Tuple */
			dpe_update_flow_tuple (entry->table_id, 
					       entry->instance_id,
					       table_id, 
					       inst_id,
					       p_flow_tuple);
		    }while (npf_list_itrNext(itr) != -1);
		}/* end if iterator was successfully created */
		/* now we need to recursively propagate the Flow tuple to all
		   upstream DPEs
		*/
		itr = npf_list_itrCreate 
		    (&(g_dpe_meter_tbl[inst_id]->prev_tbl_entry_list));
	    } /* end if (entry was non NULL) */
	}
	break;
    case DPE_MARKER_TABLE_ID:
	{
	    if(g_dpe_marker_tbl[inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_ERR, "Entry is NULL\n");
		PRETURN (-1);
	    }
	    else {
		QOS_DBG_PRINT (LOG_INFO, "Propagating Flow Tuple from Marker\n");
		p_flow_tuple = &(g_dpe_marker_tbl[inst_id]->flow_tuple);
		itr = npf_list_itrCreate 
		    (&(g_dpe_marker_tbl[inst_id]->prev_tbl_entry_list));
		if (npf_list_itrFirst(itr) != -1) {
		    do {
			entry = (dpe_EntryId_t*) npf_list_itrGetData(itr);
			/* update the corr. table & instance's Flow Tuple */
			dpe_update_flow_tuple (entry->table_id, 
					       entry->instance_id,
					       table_id, 
					       inst_id,
					       p_flow_tuple);
		    }while (npf_list_itrNext(itr) != -1);
		}/* end if iterator was successfully created */
		/* now we need to recursively propagate the Flow tuple to all
		   upstream DPEs
		*/
		itr = npf_list_itrCreate 
		    (&(g_dpe_marker_tbl[inst_id]->prev_tbl_entry_list));
	    } /* end if (entry was non NULL) */
	}
	break;
    case DPE_CLASSIFIER_TABLE_ID:
	{
	    if(g_dpe_clfr_tbl[inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_ERR, "Next Entry is NULL\n");
		PRETURN (-1);
	    }
	    else {
		QOS_DBG_PRINT (LOG_INFO, "Propagating Flow Tuple from Clfr\n");
		p_flow_tuple = &(g_dpe_clfr_tbl[inst_id]->flow_tuple);
		itr = npf_list_itrCreate 
		    (&(g_dpe_clfr_tbl[inst_id]->prev_tbl_entry_list));
		if (npf_list_itrFirst(itr) != -1) {
		    do {
			entry = (dpe_EntryId_t*) npf_list_itrGetData(itr);
			/* update the corr. table & instance's Flow Tuple */
			dpe_update_flow_tuple (entry->table_id, 
					       entry->instance_id,
					       table_id, 
					       inst_id,
					       p_flow_tuple);
		    }while (npf_list_itrNext(itr) != -1);
		}/* end if iterator was successfully created */
		/* now we need to recursively propagate the Flow tuple to all
		   upstream DPEs
		*/
		itr = npf_list_itrCreate 
		    (&(g_dpe_clfr_tbl[inst_id]->prev_tbl_entry_list));
	    } /* end if (entry was non NULL) */
	}
	break;
    } /* end switch (next tbl id) */
    if (npf_list_itrFirst(itr) != -1) {
	do {
	    QOS_DBG_PRINT (LOG_INFO, "Found a back ptr!\n");
	    entry = (dpe_EntryId_t*) npf_list_itrGetData(itr);
	    /* recursively call this function for this table */
	    dpe_propagate_flow_tuple (entry->table_id, entry->instance_id);
	}while (npf_list_itrNext(itr) != -1);
    }    
    npf_list_itrDelete (itr);
    PRETURN (err);
}

/**************************************************************************
*
*   Function: dpe_update_flow_tuple
*
*   Parameters: IN: table_id
*               IN: instance_id
* 
*   Description: updates the Flow Tuple of this DPE entry
*
*   PRETURN : 0 on success, -1 for error 
*************************************************************************/

int dpe_update_flow_tuple (npf_ds_TableId_t     table_id,
			   npf_ds_InstanceId_t  inst_id,
			   npf_ds_TableId_t     next_tbl_id,
			   npf_ds_InstanceId_t  next_inst_id,
			   dpe_flow_tuple_t*    flow_tuple)
{
    int               err = 0;
    DListItr*         itr;
    dpe_flow_tuple_t* p_flow_tuple;

    switch (table_id) {
    case DPE_SCHEDULER_TABLE_ID:
	{
	    if(g_dpe_scheduler_tbl[inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_ERR, "Entry is NULL\n");
		PRETURN (-1);
	    }
	    else {
		QOS_DBG_PRINT (LOG_INFO, "Updating Flow Tuple for Sched\n");
		g_dpe_scheduler_tbl[inst_id]->flow_tuple.flow_id = 
		    flow_tuple->flow_id;
		g_dpe_scheduler_tbl[inst_id]->flow_tuple.queue_id = 
		    flow_tuple->queue_id;
		g_dpe_scheduler_tbl[inst_id]->flow_tuple.next_block_id = 
		    flow_tuple->next_block_id;
	    } /* end if (entry was non NULL) */
	}
	break;
    case DPE_QUEUE_TABLE_ID:
	{
	    if(g_dpe_queue_tbl[inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_ERR, "Entry is NULL\n");
		PRETURN (-1);
	    }
	    else {
		QOS_DBG_PRINT (LOG_INFO, "Updating Flow Tuple for Queue\n");
		g_dpe_queue_tbl[inst_id]->flow_tuple.flow_id = 
		    flow_tuple->flow_id;
		g_dpe_queue_tbl[inst_id]->flow_tuple.queue_id = 
		    flow_tuple->queue_id;
		g_dpe_queue_tbl[inst_id]->flow_tuple.next_block_id = 
		    flow_tuple->next_block_id;
	    } /* end if (entry was non NULL) */
	}
	break;
    case DPE_ALG_DROPPER_TABLE_ID:
	{
	    if(g_dpe_dropper_tbl[inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_ERR, "Entry is NULL\n");
		PRETURN (-1);
	    }
	    else {
		QOS_DBG_PRINT (LOG_INFO, "Updating Flow Tuple for Dropper\n");
		g_dpe_dropper_tbl[inst_id]->flow_tuple.flow_id = 
		    flow_tuple->flow_id;
		g_dpe_dropper_tbl[inst_id]->flow_tuple.queue_id = 
		    flow_tuple->queue_id;
		g_dpe_dropper_tbl[inst_id]->flow_tuple.next_block_id = 
		    flow_tuple->next_block_id;
	    } /* end if (entry was non NULL) */
	}
	break;
    case DPE_METER_TABLE_ID:
	{
	    if(g_dpe_meter_tbl[inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_ERR, "Entry is NULL\n");
		PRETURN (-1);
	    }
	    else {
		QOS_DBG_PRINT (LOG_INFO, "Updating Flow Tuple for Meter\n");
		if (g_dpe_meter_tbl[inst_id]->next_instance_id == next_inst_id) 
		    {
			QOS_DBG_PRINT (LOG_INFO, "Updating flow tuple for"
				       "Conforming path\n");
			g_dpe_meter_tbl[inst_id]->downstream_flow_tuple.flow_id = 
			    flow_tuple->flow_id;
			g_dpe_meter_tbl[inst_id]->downstream_flow_tuple.queue_id = 
			    flow_tuple->queue_id;
			g_dpe_meter_tbl[inst_id]->downstream_flow_tuple.next_block_id = 
			    flow_tuple->next_block_id;
			/* the other value we need to update is the queue id for
			   upstream blocks. Flow id for upstream blocks is generated
			   by the meter, so we should not change it.
			   upstream queue id = downstream queue id from conforming
			   path.
			*/
			g_dpe_meter_tbl[inst_id]->flow_tuple.queue_id = 
			    flow_tuple->queue_id;
		    }
		else if (g_dpe_meter_tbl[inst_id]->meter.failnextinstid == 
			 next_inst_id) {
		    QOS_DBG_PRINT (LOG_INFO, "Updating flow tuple for non"
				   "Conforming path\n");
		    g_dpe_meter_tbl[inst_id]->failnext_flow_tuple.flow_id = 
			flow_tuple->flow_id;
		    g_dpe_meter_tbl[inst_id]->failnext_flow_tuple.queue_id = 
			flow_tuple->queue_id;
		    g_dpe_meter_tbl[inst_id]->failnext_flow_tuple.next_block_id = 
			flow_tuple->next_block_id;
		}
		
	    } /* end if (entry was non NULL) */
	}
	break;
    case DPE_MARKER_TABLE_ID:
	{
	    if(g_dpe_marker_tbl[inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_ERR, "Entry is NULL\n");
		PRETURN (-1);
	    }
	    else {
		QOS_DBG_PRINT (LOG_INFO, "Updating Flow Tuple for Marker\n");
		g_dpe_marker_tbl[inst_id]->flow_tuple.flow_id = 
		    flow_tuple->flow_id;
		g_dpe_marker_tbl[inst_id]->flow_tuple.queue_id = 
		    flow_tuple->queue_id;
		g_dpe_marker_tbl[inst_id]->flow_tuple.next_block_id = 
		    flow_tuple->next_block_id;
	    } /* end if (entry was non NULL) */
	}
	break;
    case DPE_CLASSIFIER_TABLE_ID:
	{
	    if(g_dpe_clfr_tbl[inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_ERR, "Next Entry is NULL\n");
		PRETURN (-1);
	    }
	    else {
		QOS_DBG_PRINT (LOG_INFO, "Updating Flow Tuple for Clfr\n");
		g_dpe_clfr_tbl[inst_id]->flow_tuple.flow_id = 
		    flow_tuple->flow_id;
		g_dpe_clfr_tbl[inst_id]->flow_tuple.queue_id = 
		    flow_tuple->queue_id;
		g_dpe_clfr_tbl[inst_id]->flow_tuple.next_block_id = 
		    flow_tuple->next_block_id;
	    } /* end if (entry was non NULL) */
	}
	break;
    } /* end switch (next tbl id) */

    dpe_dbg_print_tbl_entry (table_id, inst_id);

    PRETURN (err);
}

/**************************************************************************
*
*   Function: dpe_get_downstream_status
*
*   Parameters: IN: table_id
*               IN: instance_id
*               IN: Next table id
*               IN: Next instance_id
* 
*   Description: creates a back link in (nxt_tbl, nxt_inst} to (tbl, inst}
*
*   PRETURN : 0 on success, -1 for error 
*************************************************************************/
int dpe_get_downstream_status (npf_ds_TableId_t     table_id,
			       npf_ds_InstanceId_t  inst_id,
			       uint32_t*            status)
{
    QOS_DBG_TRACE ();

    switch (table_id) {
    case DPE_SCHEDULER_TABLE_ID:
	{
	    if(g_dpe_scheduler_tbl[inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_INFO, "Entry is NULL\n");
		*status = DPE_STATUS_INVALID;
	    }
	    else {
		*status = g_dpe_scheduler_tbl[inst_id]->status;
		QOS_DBG_PRINT (LOG_INFO, "Status for Sched = %d \n", *status);
	    } /* end if (entry was non NULL) */
	}
	break;
    case DPE_QUEUE_TABLE_ID:
	{
	    if(g_dpe_queue_tbl[inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_INFO, "Entry is NULL\n");
		*status = DPE_STATUS_INVALID;
	    }
	    else {
		*status = g_dpe_queue_tbl[inst_id]->status;
		QOS_DBG_PRINT (LOG_INFO, "Status for Queue = %d \n", *status);
	    } /* end if (entry was non NULL) */
	}
	break;
    case DPE_ALG_DROPPER_TABLE_ID:
	{
	    if(g_dpe_dropper_tbl[inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_INFO, "Entry is NULL\n");
		*status = DPE_STATUS_INVALID;
	    }
	    else {
		*status = g_dpe_dropper_tbl[inst_id]->status;
		QOS_DBG_PRINT (LOG_INFO, "Status for Dropper = %d \n", *status);
	    } /* end if (entry was non NULL) */
	}
	break;
    case DPE_METER_TABLE_ID:
	{
	    if(g_dpe_meter_tbl[inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_INFO, "Entry is NULL\n");
		*status = DPE_STATUS_INVALID;
	    }
	    else {
		*status = g_dpe_meter_tbl[inst_id]->status;
		QOS_DBG_PRINT (LOG_INFO, "Status for Meter = %d \n", *status);
	    } /* end if (entry was non NULL) */
	}
	break;
    case DPE_MARKER_TABLE_ID:
	{
	    if(g_dpe_marker_tbl[inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_INFO, "Entry is NULL\n");
		*status = DPE_STATUS_INVALID;
	    }
	    else {
		*status = g_dpe_marker_tbl[inst_id]->status;
		QOS_DBG_PRINT (LOG_INFO, "Status for Marker = %d \n", *status);
	    } /* end if (entry was non NULL) */
	}
	break;
    case DPE_CLASSIFIER_TABLE_ID:
	{
	    if(g_dpe_clfr_tbl[inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_INFO, "Next Entry is NULL\n");
		*status = DPE_STATUS_INVALID;
	    }
	    else {
		*status = g_dpe_clfr_tbl[inst_id]->status;
		QOS_DBG_PRINT (LOG_INFO, "Status for Clfr = %d \n", *status);
	    }
	}
	break;
    } /* end switch (next tbl id) */
    PRETURN (0);
}

/**************************************************************************
*
*   Function: dpe_add_inst_to_prev_entry_list
*
*   Parameters: IN: table_id
*               IN: instance_id
*               IN: Next table id
*               IN: Next instance_id
* 
*   Description: creates a back link in {nxt_tbl, nxt_inst} to {tbl, inst}.
*                If the entry next_inst does not exist in next_tbl, a dummy
*                entry is created, and a backward link is installed from
*                there to the current DPE.
*
*   PRETURN : 0 on success, -1 for error 
*************************************************************************/
int dpe_add_inst_to_prev_entry_list ( npf_ds_TableId_t     table_id,
				      npf_ds_InstanceId_t  inst_id,
				      npf_ds_TableId_t     next_tbl_id,
				      npf_ds_InstanceId_t  next_inst_id)
{
    int                  err = 0;
    dpe_EntryId_t*       prev_tbl_entry;
    dpe_scheduler_t*     sched_entry;
    dpe_queue_t*         queue_entry;
    dpe_dropper_t*       dropper_entry;
    dpe_meter_t*         meter_entry;
    dpe_marker_t*        marker_entry;
    dpe_clfr_t*          clfr_entry;
    
    prev_tbl_entry = (dpe_EntryId_t*) 
	PIL_AllocateMemory (sizeof (dpe_EntryId_t), PIL_MEM_ZERO_INIT);
    if (!prev_tbl_entry) {
	QOS_DBG_PRINT (LOG_ERR, "Out of memory!\n");
	exit (1);
    }
    prev_tbl_entry->table_id = table_id;
    prev_tbl_entry->instance_id = inst_id;
    /* add prev_dpe_entry to the next table's entry pointed to by 
       next_tbl_id and next_inst_id
       - check if there is any entry corr. to the instance id
         passed in. If not, the add is out of order
    */
    /* If Entry is empty, create the next table's entry and
       fill up only the prev_tbl_entry_list of the entry
    */
    switch (next_tbl_id) {
    case DPE_SCHEDULER_TABLE_ID:
	{
	    if(g_dpe_scheduler_tbl[next_inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_INFO, "Add Queue is before Sched!\n");
		g_dpe_scheduler_tbl[next_inst_id] = (dpe_scheduler_t*) 
		    PIL_AllocateMemory (sizeof (dpe_scheduler_t),
					PIL_MEM_ZERO_INIT);
		sched_entry = (dpe_scheduler_t*) (g_dpe_scheduler_tbl[next_inst_id]);
		sched_entry->status = DPE_STATUS_INVALID;
		sched_entry->intf = 0;
		sched_entry->sched_tbl_id = 0;
		sched_entry->sched_instance_id = 0;
		sched_entry->next_tbl_id = 0;
		sched_entry->next_instance_id = 0;
		sched_entry->flow_tuple.flow_id = g_max_flow_id + 1;
		sched_entry->flow_tuple.queue_id = g_max_queue_id + 1;
		sched_entry->flow_tuple.next_block_id = DPE_NULL_TABLE_ID;
		npf_list_init (&(sched_entry->prev_tbl_entry_list), 
			       PIL_FreeMemory);
	    }
	    QOS_DBG_PRINT (LOG_INFO, "Adding back pointer from Scheduler\n");
	    npf_list_pushFront 
		(&(g_dpe_scheduler_tbl[next_inst_id]->prev_tbl_entry_list),
		 prev_tbl_entry);
	    g_dpe_scheduler_tbl[next_inst_id]->prev_tbl_count++;
	}
	break;
    case DPE_QUEUE_TABLE_ID:
	{
	    if(g_dpe_queue_tbl[next_inst_id] == NULL) {
		g_dpe_queue_tbl[next_inst_id] = (dpe_queue_t*) 
		    PIL_AllocateMemory (sizeof (dpe_queue_t),
					PIL_MEM_ZERO_INIT);
		queue_entry = (dpe_queue_t*) (g_dpe_queue_tbl[next_inst_id]);
		queue_entry->status = DPE_STATUS_INVALID;
		queue_entry->intf = 0;
		queue_entry->queue_tbl_id = 0;
		queue_entry->queue_instance_id = 0;
		queue_entry->next_tbl_id = 0;
		queue_entry->next_instance_id = 0;
		queue_entry->flow_tuple.flow_id = g_max_flow_id + 1;
		queue_entry->flow_tuple.queue_id = g_max_queue_id + 1;
		queue_entry->flow_tuple.next_block_id = DPE_NULL_TABLE_ID;
		npf_list_init (&(queue_entry->prev_tbl_entry_list), 
			       PIL_FreeMemory);
	    }
	    QOS_DBG_PRINT (LOG_INFO, "Adding back pointer from Queue\n");
	    npf_list_pushFront 
		(&(g_dpe_queue_tbl[next_inst_id]->prev_tbl_entry_list),
		 prev_tbl_entry);
	    g_dpe_queue_tbl[next_inst_id]->prev_tbl_count++;
	}
	break;
    case DPE_ALG_DROPPER_TABLE_ID:
	{
	    if(g_dpe_dropper_tbl[next_inst_id] == NULL) {
		g_dpe_dropper_tbl[next_inst_id] = (dpe_dropper_t*) 
		    PIL_AllocateMemory (sizeof (dpe_dropper_t),
					PIL_MEM_ZERO_INIT);
		dropper_entry = (dpe_dropper_t*) (g_dpe_dropper_tbl[next_inst_id]);
		dropper_entry->status = DPE_STATUS_INVALID;
		dropper_entry->intf = 0;
		dropper_entry->dropper_tbl_id = 0;
		dropper_entry->dropper_instance_id = 0;
		dropper_entry->next_tbl_id = 0;
		dropper_entry->next_instance_id = 0;
		dropper_entry->flow_tuple.flow_id = g_max_flow_id + 1;
		dropper_entry->flow_tuple.queue_id = g_max_queue_id + 1;
		dropper_entry->flow_tuple.next_block_id = DPE_NULL_TABLE_ID;
		npf_list_init (&(dropper_entry->prev_tbl_entry_list), 
			       PIL_FreeMemory);
	    }
	    QOS_DBG_PRINT (LOG_INFO, "Adding back pointer from Dropper\n");
	    npf_list_pushFront 
		(&(g_dpe_dropper_tbl[next_inst_id]->prev_tbl_entry_list),
		 prev_tbl_entry);
	    g_dpe_dropper_tbl[next_inst_id]->prev_tbl_count++;
	}
	break;
    case DPE_METER_TABLE_ID:
	{
	    if(g_dpe_meter_tbl[next_inst_id] == NULL) {
		g_dpe_meter_tbl[next_inst_id] = (dpe_meter_t*) 
		    PIL_AllocateMemory (sizeof (dpe_meter_t),
					PIL_MEM_ZERO_INIT);
		meter_entry = (dpe_meter_t*) (g_dpe_meter_tbl[next_inst_id]);
		meter_entry->status = DPE_STATUS_INVALID;
		meter_entry->meter_tbl_id = 0;
		meter_entry->meter_instance_id = 0;
		meter_entry->next_tbl_id = 0;
		meter_entry->next_instance_id = 0;
		meter_entry->flow_tuple.flow_id = g_max_flow_id + 1;
		meter_entry->flow_tuple.queue_id = g_max_queue_id + 1;
		meter_entry->flow_tuple.next_block_id = DPE_NULL_TABLE_ID;
		npf_list_init (&(meter_entry->prev_tbl_entry_list), 
			       PIL_FreeMemory);
	    }
	    QOS_DBG_PRINT (LOG_INFO, "Adding back pointer from Meter\n");
	    npf_list_pushFront 
		(&(g_dpe_meter_tbl[next_inst_id]->prev_tbl_entry_list),
		 prev_tbl_entry);
	    g_dpe_meter_tbl[next_inst_id]->prev_tbl_count++;
	}
	break;
    case DPE_MARKER_TABLE_ID:
	{
	    if(g_dpe_marker_tbl[next_inst_id] == NULL) {
		g_dpe_marker_tbl[next_inst_id] = (dpe_marker_t*) 
		    PIL_AllocateMemory (sizeof (dpe_marker_t),
					PIL_MEM_ZERO_INIT);
		marker_entry = (dpe_marker_t*) (g_dpe_marker_tbl[next_inst_id]);
		marker_entry->status = DPE_STATUS_INVALID;
		marker_entry->marker_tbl_id = 0;
		marker_entry->marker_instance_id = 0;
		marker_entry->next_tbl_id = 0;
		marker_entry->next_instance_id = 0;
		marker_entry->flow_tuple.flow_id = g_max_flow_id + 1;
		marker_entry->flow_tuple.queue_id = g_max_queue_id + 1;
		marker_entry->flow_tuple.next_block_id = DPE_NULL_TABLE_ID;
		npf_list_init (&(marker_entry->prev_tbl_entry_list), 
			       PIL_FreeMemory);
	    }
	    QOS_DBG_PRINT (LOG_INFO, "Adding back pointer from Marker\n");
	    npf_list_pushFront 
		(&(g_dpe_marker_tbl[next_inst_id]->prev_tbl_entry_list),
		 prev_tbl_entry);
	    g_dpe_marker_tbl[next_inst_id]->prev_tbl_count++;
	}
	break;
    case DPE_CLASSIFIER_TABLE_ID:
	{
	    if(g_dpe_clfr_tbl[next_inst_id] == NULL) {
		g_dpe_clfr_tbl[next_inst_id] = (dpe_clfr_t*) 
		    PIL_AllocateMemory (sizeof (dpe_clfr_t),
					PIL_MEM_ZERO_INIT);
		clfr_entry = (dpe_clfr_t*) (g_dpe_clfr_tbl[next_inst_id]);
		clfr_entry->status = DPE_STATUS_INVALID;
		clfr_entry->clfr_tbl_id = 0;
		clfr_entry->clfr_instance_id = 0;
		clfr_entry->next_tbl_id = 0;
		clfr_entry->next_instance_id = 0;
		clfr_entry->flow_tuple.flow_id = g_max_flow_id + 1;
		clfr_entry->flow_tuple.queue_id = g_max_queue_id + 1;
		clfr_entry->flow_tuple.next_block_id = DPE_NULL_TABLE_ID;
		npf_list_init (&(clfr_entry->prev_tbl_entry_list), 
			       PIL_FreeMemory);
	    }
	    QOS_DBG_PRINT (LOG_INFO, "Adding back pointer from Queue\n");
	    npf_list_pushFront 
		(&(g_dpe_clfr_tbl[next_inst_id]->prev_tbl_entry_list),
		 prev_tbl_entry);
	    g_dpe_clfr_tbl[next_inst_id]->prev_tbl_count++;
	}
	break;
    } /* end switch (next tbl id) */
    
    PRETURN (err);
}

/**************************************************************************
*
*   Function: dpe_delete_inst_from_prev_entry_list
*
*   Parameters: IN: table_id
*               IN: instance_id
*               IN: Next table id
*               IN: Next instance_id
* 
*   Description: deletes a back link in {nxt_tbl, nxt_inst} to {tbl, inst}
*
*   PRETURN : 0 on success, -1 for error 
*************************************************************************/
int dpe_delete_inst_from_prev_entry_list ( npf_ds_TableId_t     table_id,
					   npf_ds_InstanceId_t  inst_id,
					   npf_ds_TableId_t     next_tbl_id,
					   npf_ds_InstanceId_t  next_inst_id)
{
    int            err = 0;
    dpe_EntryId_t* entry;
    DListItr*      itr;

    switch (next_tbl_id) {
    case DPE_SCHEDULER_TABLE_ID:
	{
	    if(g_dpe_scheduler_tbl[next_inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_ERR, "Next Entry is NULL\n");
		PRETURN (0);
	    }
	    else {
		QOS_DBG_PRINT (LOG_INFO, "Removing back pointer from"
			       "Scheduler\n");
		itr = npf_list_itrCreate 
		    (&(g_dpe_scheduler_tbl[next_inst_id]->prev_tbl_entry_list));
		if (npf_list_itrFirst(itr) != -1) {
		    do {
			entry = (dpe_EntryId_t*) npf_list_itrGetData(itr);
			if (entry->table_id == table_id &&
			    entry->instance_id == inst_id) {
			    npf_list_itrRemoveFree (itr);
			    g_dpe_scheduler_tbl[next_inst_id]->prev_tbl_count--;
			} /* if Entry was found !*/
		    }while (npf_list_itrNext(itr) != -1);
		}/* end if iterator was successfully created */
	    } /* end if (entry was non NULL) */
	    npf_list_itrDelete (itr);
	}
	break;
    case DPE_QUEUE_TABLE_ID:
	{
	    if(g_dpe_queue_tbl[next_inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_ERR, "Next Entry is NULL\n");
		PRETURN (0);
	    }
	    else {
		QOS_DBG_PRINT (LOG_INFO, "Removing back pointer from Queue\n");
		itr = npf_list_itrCreate 
		    (&(g_dpe_queue_tbl[next_inst_id]->prev_tbl_entry_list));
		if (npf_list_itrFirst(itr) != -1) {
		    do {
			entry = (dpe_EntryId_t*) npf_list_itrGetData(itr);
			if (entry->table_id == table_id &&
			    entry->instance_id == inst_id) {
			    npf_list_itrRemoveFree (itr);
			    g_dpe_queue_tbl[next_inst_id]->prev_tbl_count--;
			} /* if Entry was found !*/
		    }while (npf_list_itrNext(itr) != -1);
		}/* end if iterator was successfully created */
	    } /* end if (entry was non NULL) */
	    npf_list_itrDelete (itr);
	}
	break;
    case DPE_ALG_DROPPER_TABLE_ID:
	{
	    if(g_dpe_dropper_tbl[next_inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_ERR, "Next Entry is NULL\n");
		PRETURN (0);
	    }
	    else {
		QOS_DBG_PRINT (LOG_INFO, "Removing back pointer from Dropper\n");
		itr = npf_list_itrCreate 
		    (&(g_dpe_dropper_tbl[next_inst_id]->prev_tbl_entry_list));
		if (npf_list_itrFirst(itr) != -1) {
		    do {
			entry = (dpe_EntryId_t*) npf_list_itrGetData(itr);
			if (entry->table_id == table_id &&
			    entry->instance_id == inst_id) {
			    npf_list_itrRemoveFree (itr);
			    g_dpe_dropper_tbl[next_inst_id]->prev_tbl_count--;
			} /* if Entry was found !*/
		    }while (npf_list_itrNext(itr) != -1);
		}/* end if iterator was successfully created */
	    } /* end if (entry was non NULL) */
	    npf_list_itrDelete (itr);
	}
	break;
    case DPE_METER_TABLE_ID:
	{
	    if(g_dpe_meter_tbl[next_inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_ERR, "Next Entry is NULL\n");
		PRETURN (0);
	    }
	    else {
		QOS_DBG_PRINT (LOG_INFO, "Removing back pointer from Meter\n");
		itr = npf_list_itrCreate 
		    (&(g_dpe_meter_tbl[next_inst_id]->prev_tbl_entry_list));
		if (npf_list_itrFirst(itr) != -1) {
		    do {
			entry = (dpe_EntryId_t*) npf_list_itrGetData(itr);
			if (entry->table_id == table_id &&
			    entry->instance_id == inst_id) {
			    npf_list_itrRemoveFree (itr);
			    g_dpe_meter_tbl[next_inst_id]->prev_tbl_count--;
			} /* if Entry was found !*/
		    }while (npf_list_itrNext(itr) != -1);
		}/* end if iterator was successfully created */
	    } /* end if (entry was non NULL) */
	    npf_list_itrDelete (itr);
	}
	break;
    case DPE_MARKER_TABLE_ID:
	{
	    if(g_dpe_marker_tbl[next_inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_ERR, "Next Entry is NULL\n");
		PRETURN (0);
	    }
	    else {
		QOS_DBG_PRINT (LOG_INFO, "Removing back pointer from Marker\n");
		itr = npf_list_itrCreate 
		    (&(g_dpe_marker_tbl[next_inst_id]->prev_tbl_entry_list));
		if (npf_list_itrFirst(itr) != -1) {
		    do {
			entry = (dpe_EntryId_t*) npf_list_itrGetData(itr);
			if (entry->table_id == table_id &&
			    entry->instance_id == inst_id) {
			    npf_list_itrRemoveFree (itr);
			    g_dpe_marker_tbl[next_inst_id]->prev_tbl_count--;
			} /* if Entry was found !*/
		    }while (npf_list_itrNext(itr) != -1);
		}/* end if iterator was successfully created */
	    } /* end if (entry was non NULL) */
	    npf_list_itrDelete (itr);
	}
	break;
    case DPE_CLASSIFIER_TABLE_ID:
	{
	    if(g_dpe_clfr_tbl[next_inst_id] == NULL) {
		QOS_DBG_PRINT (LOG_ERR, "Next Entry is NULL\n");
		PRETURN (0);
	    }
	    else {
		QOS_DBG_PRINT (LOG_INFO, "Removing back pointer from Clfr\n");
		itr = npf_list_itrCreate 
		    (&(g_dpe_clfr_tbl[next_inst_id]->prev_tbl_entry_list));
		if (npf_list_itrFirst(itr) != -1) {
		    do {
			entry = (dpe_EntryId_t*) npf_list_itrGetData(itr);
			if (entry->table_id == table_id &&
			    entry->instance_id == inst_id) {
			    npf_list_itrRemoveFree (itr);
			    g_dpe_clfr_tbl[next_inst_id]->prev_tbl_count--;
			} /* if Entry was found !*/
		    }while (npf_list_itrNext(itr) != -1);
		}/* end if iterator was successfully created */
	    } /* end if (entry was non NULL) */
	    npf_list_itrDelete (itr);
	}
	break;
    } /* end switch (next tbl id) */
    
    PRETURN (err);
}

/*************************************************************************
 *                                                                       *
 *              DPE SEARCH MANAGEMENT ROUTINES                           *
 *                                                                       *
 *************************************************************************/
/**************************************************************************
*
*   Function: QueryFlowTupleByDSCP
*
*   Parameters: IN: DSCP
*               OUT: Flow Tuple (flow_id, queue_id)
* 
*   Description: looks up the Classifier Table based on DSCP
*                and returns a Flow tuple
*
*   PRETURN : 0 on success, -1 for error 
*************************************************************************/

int QueryFlowTupleByDSCP (uint32_t  dscp,
			  uint32_t* flowid,
			  uint32_t* queueid)
{
    dpe_clfr_t*          clfr_entry;
    npf_ds_BaseFilter_t* filter;
    npf_ds_IPv4Filter_t* ipv4_filter;
    npf_ds_IPv6Filter_t* ipv6_filter;
    int                  err = 0, i;

    QOS_DBG_TRACE();

    for (i = 0; i < g_max_flow_id; i++) {
	if (g_dpe_clfr_tbl[i] != NULL) {
	    clfr_entry = (dpe_clfr_t*) g_dpe_clfr_tbl[i];
	    filter = (npf_ds_BaseFilter_t*) (clfr_entry->filter);

	    if (filter->type == npf_ds_IPv4Filter) {
		ipv4_filter = (npf_ds_IPv4Filter_t*) filter;
		if (ipv4_filter->Dscp == dscp) {
		    /* FOUND DSCP! */
		    *flowid  = clfr_entry->flow_tuple.flow_id;
		    *queueid = clfr_entry->flow_tuple.queue_id;
		    PRETURN (0);
		} /* end if (FOUND DSCP) */
	    }/* end if (IPv4 filter) */
	    else if (filter->type == npf_ds_IPv6Filter) {
		ipv6_filter = (npf_ds_IPv6Filter_t*) filter;
		if (ipv6_filter->Dscp == dscp) {
		    /* FOUND DSCP! */
		    *flowid  = clfr_entry->flow_tuple.flow_id;
		    *queueid = clfr_entry->flow_tuple.queue_id;
		    PRETURN (0);
		} /* end if (FOUND DSCP) */
	    } /* end if (IPv6 filter) */
	} /* end if table entry is non NULL */
    } /* end for loop */
    
    /* if we havent returned by now, we havent found it */
    err = -1;
    PRETURN (err);
}

/**************************************************************************
*
*   Function: QueryFlowTupleByLabel
*
*   Parameters: IN: npf_mplsLbl
*               OUT: Flow Tuple (flow_id, queue_id)
* 
*   Description: looks up the Classifier Table based on MPLS Label
*                and returns a Flow tuple
*
*   PRETURN : 0 on success, -1 for error 
*************************************************************************/

int QueryFlowTupleByLabel (npf_mplsLbl mpls_label,
			   uint32_t*   flowid,
			   uint32_t*   queueid)
{
    dpe_clfr_t*          clfr_entry;
    npf_ds_BaseFilter_t* filter;
    npf_ds_MPLSFilter_t* mpls_filter;
    uint32_t             shim_lbl, vpi, vci, clp, dlci;
    int                  dlcilen;
    int                  err = 0, i;

    QOS_DBG_TRACE();

    for (i = 0; i < g_max_flow_id; i++) {
	if (g_dpe_clfr_tbl[i] != NULL) {
	    clfr_entry = (dpe_clfr_t*) g_dpe_clfr_tbl[i];
	    filter = (npf_ds_BaseFilter_t*) (clfr_entry->filter);
	    
	    if (filter->type == npf_ds_MPLSFilter) {
		mpls_filter = (npf_ds_MPLSFilter_t*) filter;
		if (mpls_filter->type == npf_ds_Label) {
		    if (mpls_filter->filter.label.mplsLbl_type == 
			mpls_label.mplsLbl_type) {
			/* Label Types have matched. */ 
			switch (mpls_label.mplsLbl_type) {
			case SHIMLBL:
			    {
				shim_lbl = 
	    mpls_filter->filter.label.npf_mplsLblValue.shimLbl.label_value;
	if (mpls_label.npf_mplsLblValue.shimLbl.label_value == shim_lbl) {
				    /* FOUND LABEL */
				    *flowid  = clfr_entry->flow_tuple.flow_id;
				    *queueid = clfr_entry->flow_tuple.queue_id;
				    PRETURN (0);    
				}
			    }
			    break;
			case ATMLBL:
			    {
		vpi = mpls_filter->filter.label.npf_mplsLblValue.atmLbl.VPI;
		vci = mpls_filter->filter.label.npf_mplsLblValue.atmLbl.VCI;
		clp = mpls_filter->filter.label.npf_mplsLblValue.atmLbl.CLP;
			if (mpls_label.npf_mplsLblValue.atmLbl.VPI == vpi &&
			    mpls_label.npf_mplsLblValue.atmLbl.VCI == vci &&
			    mpls_label.npf_mplsLblValue.atmLbl.CLP == clp) {
			    /* FOUND LABEL */
			    *flowid  = clfr_entry->flow_tuple.flow_id;
			    *queueid = clfr_entry->flow_tuple.queue_id;
			    PRETURN (0); 
			}
			    }
			    break;
			case FRLBL:
			    {
	dlcilen = mpls_filter->filter.label.npf_mplsLblValue.frLbl.dlciLen;
		dlci = mpls_filter->filter.label.npf_mplsLblValue.frLbl.dlci;
		if (mpls_label.npf_mplsLblValue.frLbl.dlciLen == dlcilen &&
		    mpls_label.npf_mplsLblValue.frLbl.dlci == dlci) {
		    /* FOUND LABEL */
		    *flowid  = clfr_entry->flow_tuple.flow_id;
		    *queueid = clfr_entry->flow_tuple.queue_id;
		    PRETURN (0); 
		}
			    }
			    break;
			} /* end switch on LABEL TYPE */
		    } /* end if (LABEL TYPES HAVE MATCHED) */
		}/* end if (MPLS Label) */
	    } /* end if (MPLS filter) */
	} /* end if table entry is non NULL */
    } /* end for loop */
    /* if we havent returned by now, we havent found it */
    err = -1;
    PRETURN (err);
}

/**************************************************************************
*
*   Function: QueryFlowTupleByFEC
*
*   Parameters: IN: npf_mplsFec
*               OUT: Flow Tuple (flow_id, queue_id)
* 
*   Description: looks up the Classifier Table based on MPLS FEC
*                and returns a Flow tuple
*
*   PRETURN : 0 on success, -1 for error 
*************************************************************************/

int QueryFlowTupleByFEC (npf_mplsFec mpls_fec,
			 uint32_t*   flowid,
			 uint32_t*   queueid)
{
    dpe_clfr_t*          clfr_entry;
    npf_ds_BaseFilter_t* filter;
    npf_ds_MPLSFilter_t* mpls_filter;
    int                  prefix_len;
    uint32_t             addr_family, as_id;
    IPA                  host_addr, prefix;
    int                  err = 0, i;

    QOS_DBG_TRACE();

    for (i = 0; i < g_max_flow_id; i++) {
	if (g_dpe_clfr_tbl[i] != NULL) {
	    clfr_entry = (dpe_clfr_t*) g_dpe_clfr_tbl[i];
	    filter = (npf_ds_BaseFilter_t*) (clfr_entry->filter);
	    
	    if (filter->type == npf_ds_MPLSFilter) {
		mpls_filter = (npf_ds_MPLSFilter_t*) filter;
		if (mpls_filter->type == npf_ds_FEC) {
		    if (mpls_filter->filter.fec.type == mpls_fec.type) {
			switch (mpls_fec.type) {
			case npf_HostFEC:
			    {
				addr_family = 
		mpls_filter->filter.fec.npf_fecIdentifier.hostaddress.addrFamily;
				host_addr =
		mpls_filter->filter.fec.npf_fecIdentifier.hostaddress.hostAddr;
	if (addr_family == mpls_fec.npf_fecIdentifier.hostaddress.addrFamily &&
    host_addr == (uint32_t)mpls_fec.npf_fecIdentifier.hostaddress.hostAddr)
	    {
		/* FOUND MPLS Label! */
		*flowid  = clfr_entry->flow_tuple.flow_id;
		*queueid = clfr_entry->flow_tuple.queue_id;
		PRETURN (0);
	    }
			    }
			    break;
			case npf_PrefixFEC:
			    {
      prefix_len = mpls_filter->filter.fec.npf_fecIdentifier.prefix.prefixLength;
	prefix = mpls_filter->filter.fec.npf_fecIdentifier.prefix.prefix;
	if (prefix_len == mpls_fec.npf_fecIdentifier.prefix.prefixLength &&
	    prefix == mpls_fec.npf_fecIdentifier.prefix.prefix) {
	    /* FOUND MPLS Label! */
	    *flowid  = clfr_entry->flow_tuple.flow_id;
	    *queueid = clfr_entry->flow_tuple.queue_id;
	    PRETURN (0);
	}
			    }
			    break;
			case npf_ASFEC:
			    {
			as_id = mpls_filter->filter.fec.npf_fecIdentifier.asId;
			if (as_id == mpls_fec.npf_fecIdentifier.asId) {
			    /* FOUND MPLS Label! */
			    *flowid  = clfr_entry->flow_tuple.flow_id;
			    *queueid = clfr_entry->flow_tuple.queue_id;
			    PRETURN (0)
				}
			    }
			    break;
			}/* end switch on FEC TYPE */
		    } /* end (IF FEC TYPES HAVE MATCHED) */
		}/* end if (MPLS Label) */
	    } /* end if (MPLS filter) */
	} /* end if table entry is non NULL */
    } /* end for loop */
    /* if we havent returned by now, we havent found it */
    err = -1;
    PRETURN (err);
}

/**************************************************************************
*
*   Function: dpe_signal_mpls
*
*   Parameters: IN: dpe_clfr_t* - Classifier entry that got a new FId/Qid
*               
*   Description: looks into the Classifier entry, determines the filter type
*                and invokes the callback registered by MPLS
*
*   PRETURN : 0 on success, -1 for error 
*************************************************************************/
int dpe_signal_mpls (dpe_clfr_t* clfr_entry)
{
    int      err = 0;
    uint32_t filter_type;
    void*    key;

    QOS_DBG_TRACE ();

    dpe_get_clfrFilter_params (clfr_entry, &filter_type, &key);
    
    InvokeFlowEventCallback (filter_type, key, clfr_entry->flow_tuple.flow_id,
			     clfr_entry->flow_tuple.queue_id);
   
    PRETURN (err);
}


/**************************************************************************
*
*   Function: dpe_get_clfrFilter_params
*
*   Parameters: IN: dpe_clfr_t* - Classifier entry
*               OUT: TYPE - Filter type - DSCP,FEC, LABEL
*               OUT: Key - will contain the actual filter value
*                    Memory is allocated by this function. Caller MUST free the 
*                    memory after use using PIL_FreeMemory()
*              
*   Description: looks into the Classifier entry, determines the filter type
*                and returns the actual filter value
*
*   PRETURN : 0 on success, -1 for error 
*************************************************************************/
int dpe_get_clfrFilter_params (dpe_clfr_t* clfr_entry,
			       uint32_t*   type,  /* OUT: Type: DSCP, FEC or 
						     Label */ 
			       void**      key) /* OUT: Dscp, npf_mplsFec, 
						   npf_mplsLbl - Value */
{
    int err = 0;
    npf_ds_BaseFilter_t* filter;
    npf_ds_IPv4Filter_t* ipv4_filter;
    npf_ds_IPv6Filter_t* ipv6_filter;
    npf_ds_MPLSFilter_t* mpls_filter;
    uint32_t*            dscp;
    npf_mplsLbl*         mpls_lbl;
    npf_mplsFec*         mpls_fec;

    QOS_DBG_TRACE ();
    filter = (npf_ds_BaseFilter_t*) (clfr_entry->filter);
    
    switch (filter->type) {
    case npf_ds_IPv4Filter:
	{
	    QOS_DBG_PRINT (LOG_INFO, "Filter type = DSCP\n");
	    ipv4_filter = (npf_ds_IPv4Filter_t*) filter;
	    *type = DSCP_KEY;
	    dscp = (uint32_t*) PIL_AllocateMemory (sizeof (uint32_t),
						   PIL_MEM_ZERO_INIT);
	    *dscp = ipv4_filter->Dscp; 
	    *key = dscp;
	}
	break;
    case npf_ds_IPv6Filter:
	{
	    QOS_DBG_PRINT (LOG_INFO, "Filter type = DSCP\n");
	    ipv6_filter = (npf_ds_IPv6Filter_t*) filter;
	    *type = DSCP_KEY;
	    dscp = (uint32_t*) PIL_AllocateMemory (sizeof (uint32_t),
						   PIL_MEM_ZERO_INIT);
	    *dscp = ipv4_filter->Dscp;
	    *key = dscp; 
	}
	break;
    case npf_ds_MPLSFilter:
	{
	    QOS_DBG_PRINT (LOG_INFO, "Filter type = MPLS\n");
	    mpls_filter = (npf_ds_MPLSFilter_t*) filter;
	    QOS_DBG_PRINT (LOG_INFO, "MPLS Filter type - %d \n",
			   mpls_filter->type);
	    switch (mpls_filter->type) {
	    case npf_ds_Label:
		{
		    QOS_DBG_PRINT (LOG_INFO, "Filter type = MPLS Label \n");
		    mpls_lbl = (npf_mplsLbl*) 
			PIL_AllocateMemory (sizeof (npf_mplsLbl), 
					    PIL_MEM_ZERO_INIT);
		    *key = mpls_lbl;
		    switch (mpls_filter->filter.label.mplsLbl_type) {
		    case SHIMLBL:
			{
			    QOS_DBG_PRINT (LOG_INFO, "Filter type = MPLS Shim"
					   "Lbl\n");
			    *type = LABEL_KEY;
			    mpls_lbl->npf_mplsLblValue.shimLbl.label_value = 
		  mpls_filter->filter.label.npf_mplsLblValue.shimLbl.label_value;
			}
			break;
		    case ATMLBL:
			{
			    QOS_DBG_PRINT (LOG_INFO, "Filter type = MPLS ATM"
					   "Lbl\n");
			    *type = LABEL_KEY;
			    mpls_lbl->npf_mplsLblValue.atmLbl.VPI = 
			mpls_filter->filter.label.npf_mplsLblValue.atmLbl.VPI;
			    mpls_lbl->npf_mplsLblValue.atmLbl.VCI = 
			mpls_filter->filter.label.npf_mplsLblValue.atmLbl.VCI;
			    mpls_lbl->npf_mplsLblValue.atmLbl.CLP = 
			mpls_filter->filter.label.npf_mplsLblValue.atmLbl.CLP;
			}
			break;
		    case FRLBL:
			{
			    QOS_DBG_PRINT (LOG_INFO, "Filter type = MPLS FR"
					   "Lbl\n");
			    *type = LABEL_KEY;
			    mpls_lbl->npf_mplsLblValue.frLbl.dlciLen = 
			mpls_filter->filter.label.npf_mplsLblValue.frLbl.dlciLen;
			    mpls_lbl->npf_mplsLblValue.frLbl.dlci = 
			mpls_filter->filter.label.npf_mplsLblValue.frLbl.dlci;
			}
			break;
		    } /* end of diff. MPLS Label types */
		} /* end if (MPLS Label) */
		break;
	    case npf_ds_FEC:
		{
		    QOS_DBG_PRINT (LOG_INFO, "Filter type = MPLS FEC \n");
		    mpls_fec = (npf_mplsFec*) 
			PIL_AllocateMemory (sizeof (npf_mplsFec), 
					    PIL_MEM_ZERO_INIT);
		    *key = mpls_fec;
		    switch (mpls_filter->filter.fec.type) {
		    case npf_HostFEC:
			{
			    QOS_DBG_PRINT (LOG_INFO, "Filter type = MPLS Host"
					   "FEC\n");
			    *type = FEC_KEY;
			    mpls_fec->npf_fecIdentifier.hostaddress.addrFamily =
		mpls_filter->filter.fec.npf_fecIdentifier.hostaddress.addrFamily;
			    mpls_fec->npf_fecIdentifier.hostaddress.hostAddr =
		mpls_filter->filter.fec.npf_fecIdentifier.hostaddress.hostAddr;
			}
			break;
		    case npf_PrefixFEC:
			{
			    QOS_DBG_PRINT (LOG_INFO, "Filter type = MPLS Prefix"
					   "FEC\n");
			    *type = FEC_KEY;
			    mpls_fec->npf_fecIdentifier.prefix.prefixLength = 
		mpls_filter->filter.fec.npf_fecIdentifier.prefix.prefixLength;
			    mpls_fec->npf_fecIdentifier.prefix.prefix = 
		mpls_filter->filter.fec.npf_fecIdentifier.prefix.prefix;
			}
			break;
		    case npf_ASFEC:
			{
			    QOS_DBG_PRINT (LOG_INFO, "Filter type = MPLS AS"
					   "FEC\n");
			    *type = FEC_KEY;
			    mpls_fec->npf_fecIdentifier.asId = 
				mpls_filter->filter.fec.npf_fecIdentifier.asId;
			}
			break;
		    } /* end switch on FEC TYPE) */
		} /* end if (MPLS FEC) */
	 	break;
   	default:
		QOS_DBG_PRINT (LOG_INFO, "Unknown FIlter type - %d !\n",
				mpls_filter->type);
		break;
	    } /* end switch (MPLS LABEL or FEC) */
	}
	break;
    } /* end switch (FILTER TYPE) */
    PRETURN (err);
}

/*************************************************************************
 *                                                                       *
 *              COPY ROUTINES FOR COPYING NPF STRUCTURES                 *
 *                                                                       *
 *************************************************************************/
/**************************************************************************
*
*   Function: dpe_copy_sched_params
*
*   Parameters: IN: pointer to destination
*               IN: pointer to copy struct from
*
*   Description: Copies the contents of the source structure into the dest
*
*   PRETURN : None 
*************************************************************************/
void dpe_copy_sched_params (npf_ds_Scheduler_t* sched_entry, /* destination */
			    npf_ds_Scheduler_t* ds_scheduler) /* source */
{
    int count, i;

    QOS_DBG_TRACE();
    QOS_DBG_PRINT (LOG_INFO, "Copying Scheduler parameters!\n");

    /* Copy individual parameters from npf_ds_Scheduler_t */
    sched_entry->Method = npf_ds_SchedulerDRR;
    /* use previously alloc-ed memory for min Rate since it is a simple struct */
    if (!ds_scheduler->min_rate) {
	if (sched_entry->min_rate)
	    PIL_FreeMemory ((void*) sched_entry->min_rate);
	sched_entry->min_rate = NULL;
    }
    else {
	sched_entry->min_rate = (npf_ds_MinRate_t*) 
	    PIL_AllocateMemory (sizeof (npf_ds_MinRate_t), 
				PIL_MEM_ZERO_INIT);
	if (!sched_entry->min_rate) {
	    QOS_DBG_PRINT (LOG_ERR, "Out of memory!\n");
	    exit (1);
	}
	memcpy (sched_entry->min_rate, 
		ds_scheduler->min_rate, sizeof (npf_ds_MinRate_t));
    }
    if (!ds_scheduler->max_rate) {
	if (sched_entry->max_rate)
	    PIL_FreeMemory ((void*) sched_entry->max_rate);
	sched_entry->max_rate = NULL;
	PRETURN (/*void*/);
    }
    
    if (!sched_entry->max_rate)
	sched_entry->max_rate = (npf_ds_MaxRate_t*) 
	    PIL_AllocateMemory (sizeof (npf_ds_MaxRate_t), 
				PIL_MEM_ZERO_INIT);
    if (!sched_entry->max_rate) {
	QOS_DBG_PRINT (LOG_ERR, "Out of memory!\n");
	exit (1);
    }
    sched_entry->max_rate->Type = 
	ds_scheduler->max_rate->Type;
    if (ds_scheduler->max_rate->Type == Type_SLMaxRate) {
	memcpy (&sched_entry->max_rate->maxrate.slmaxrate,
		&ds_scheduler->max_rate->maxrate.slmaxrate,
		sizeof (npf_ds_SLMaxRate_t));
    }
    else if (ds_scheduler->max_rate->Type == Type_MLMaxRate) {
	/* if count has changed, we need to re alloc memory, so free 
	   previously alloc-ed memory */
	if (sched_entry->max_rate->maxrate.mlmaxrate.maxrate_list)
	    PIL_FreeMemory 
		((void*) 
		 sched_entry->max_rate->maxrate.mlmaxrate.maxrate_list);

	count = ds_scheduler->max_rate->maxrate.mlmaxrate.maxrate_count; 
	sched_entry->max_rate->maxrate.mlmaxrate.maxrate_count=
	    count;
	sched_entry->max_rate->maxrate.mlmaxrate.maxrate_list=
	    (npf_ds_SLMaxRate_t*) 
	    PIL_AllocateMemory (count * sizeof (npf_ds_SLMaxRate_t),
				PIL_MEM_ZERO_INIT);
	for(i = 0; i < count; i++) {
	    memcpy (&sched_entry->max_rate->maxrate.mlmaxrate.maxrate_list[i],
		    &ds_scheduler->max_rate->maxrate.mlmaxrate.maxrate_list[i],
		    sizeof (npf_ds_SLMaxRate_t));
	}/* end for loop */
    }/* end if(maxrate) */

    QOS_DBG_PRINT (LOG_INFO, "Done Copying Scheduler parameters!\n");
    PRETURN (/*void*/);
}

/**************************************************************************
*
*   Function: dpe_copy_queue_params
*
*   Parameters: IN: pointer to destination
*               IN: pointer to copy struct from
*
*   Description: Copies the contents of the source structure into the dest
*
*   PRETURN : None 
*************************************************************************/
void dpe_copy_queue_params (npf_ds_Q_t* queue_entry, /* destination */
			    npf_ds_Q_t* ds_queue) /* source */
{
    int count, i;

    QOS_DBG_TRACE();
    QOS_DBG_PRINT (LOG_INFO, "Copying Queue parameters!\n");

    //if (!queue_entry->min_rate)
	queue_entry->min_rate = (npf_ds_MinRate_t*) 
	    PIL_AllocateMemory (sizeof (npf_ds_MinRate_t), PIL_MEM_ZERO_INIT);
    if (!queue_entry->min_rate) {
	QOS_DBG_PRINT (LOG_ERR, "Out of memory!\n");
	exit (1);
    }
    memcpy (queue_entry->min_rate, ds_queue->min_rate, 
	    sizeof (npf_ds_MinRate_t));
    
    //if (!queue_entry->max_rate)
	queue_entry->max_rate = (npf_ds_MaxRate_t*) 
	    PIL_AllocateMemory (sizeof (npf_ds_MaxRate_t), PIL_MEM_ZERO_INIT);
    if (!queue_entry->max_rate) {
	QOS_DBG_PRINT (LOG_ERR, "Out of memory!\n");
	exit (1);
    }
    queue_entry->max_rate->Type = 
	ds_queue->max_rate->Type;
    if (ds_queue->max_rate->Type == Type_SLMaxRate) {
	memcpy (&queue_entry->max_rate->maxrate.slmaxrate,
		&ds_queue->max_rate->maxrate.slmaxrate,
		sizeof (npf_ds_SLMaxRate_t));
    }
    else if (ds_queue->max_rate->Type == Type_MLMaxRate) {
	/* free up prev. memory since count might have changed */
	if (queue_entry->max_rate->maxrate.mlmaxrate.maxrate_list)
	    PIL_FreeMemory 
		((void*) queue_entry->max_rate->maxrate.mlmaxrate.maxrate_list);

	count = ds_queue->max_rate->maxrate.mlmaxrate.maxrate_count; 
	queue_entry->max_rate->maxrate.mlmaxrate.maxrate_count=
	    count;
	queue_entry->max_rate->maxrate.mlmaxrate.maxrate_list=
	    (npf_ds_SLMaxRate_t*) 
	    PIL_AllocateMemory (count * sizeof (npf_ds_SLMaxRate_t),
				PIL_MEM_ZERO_INIT);
	for(i = 0; i < count; i++) {
	    memcpy (&queue_entry->max_rate->maxrate.mlmaxrate.maxrate_list[i],
		    &ds_queue->max_rate->maxrate.mlmaxrate.maxrate_list[i],
		    sizeof (npf_ds_SLMaxRate_t));
	}/* end for loop */
    }/* end if (maxRate) */
    PRETURN (/*void*/);
}


/**************************************************************************
*
*   Function: dpe_copy_dropper_params
*
*   Parameters: IN: pointer to destination
*               IN: pointer to copy struct from
*
*   Description: Copies the contents of the source structure into the dest
*
*   PRETURN : None 
*************************************************************************/
void dpe_copy_dropper_params (npf_ds_AlgDrop_t* dropper_entry, /* destination */
			      npf_ds_AlgDrop_t* ds_dropper) /* source */
{
    QOS_DBG_TRACE();
    QOS_DBG_PRINT (LOG_INFO, "Copying Dropper parameters!\n");

    /* copy the npf_ds_AlgDrop_t contents passed in */
    dropper_entry->Type = ds_dropper->Type;
    /* Single or Multiple Queue Dropper */
    if (ds_dropper->Type == Type_sQDrop) {
	memcpy (&dropper_entry->dropper.sqalgdrop,
		&ds_dropper->dropper.sqalgdrop,
		sizeof (npf_ds_SQAlgDrop_t));
	if (ds_dropper->dropper.sqalgdrop.Type == Type_randomDrop) {
	    //if (!dropper_entry->dropper.sqalgdrop.Specific)
	    dropper_entry->dropper.sqalgdrop.Specific = 
		PIL_AllocateMemory (sizeof (npf_ds_RandomDrop_t),
				    PIL_MEM_ZERO_INIT);
	    
	    memcpy (dropper_entry->dropper.sqalgdrop.Specific,
		    ds_dropper->dropper.sqalgdrop.Specific,
		    sizeof (npf_ds_RandomDrop_t));
	    
	} /* end if (Random Dropper) */
	else if (ds_dropper->dropper.sqalgdrop.Type == Type_alwaysDrop) {
	    /* Always Drop */
	    dropper_entry->dropper.sqalgdrop.Specific = NULL;
	} /* end if always Drop */
    } /* end if (sQDropper) */
    else if (ds_dropper->Type == Type_mQDrop) {
	/* not supported currently */
    }
    PRETURN (/*void*/);
}

/**************************************************************************
*
*   Function: dpe_copy_meter_params
*
*   Parameters: IN: pointer to destination
*               IN: pointer to copy struct from
*
*   Description: Copies the contents of the source structure into the dest
*
*   PRETURN : None 
*************************************************************************/
void dpe_copy_meter_params (npf_ds_Meter_t* meter_entry, /* destination */
			    npf_ds_Meter_t* ds_meter) /* source */
{
    QOS_DBG_TRACE();
    QOS_DBG_PRINT (LOG_INFO, "Copying Meter parameters!\n");

    /* copy contents of npf_ds_Meter_t */
    memcpy (meter_entry, ds_meter, sizeof (npf_ds_Meter_t));
    QOS_DBG_PRINT (LOG_INFO, "fail next tbl = %d\t fail next inst = %d\n\t"
		   "fail next tbl from ds =%d \t inst from ds = %d\n",
		   meter_entry->failnexttableid,  
		   meter_entry->failnextinstid,  
		   ds_meter->failnexttableid,  
		   ds_meter->failnextinstid);
  
    if (*ds_meter->Specific == npf_ds_TBParam) {
	//if (!meter_entry->Specific)
	meter_entry->Specific =  PIL_AllocateMemory 
	    (sizeof (npf_ds_TBParam_t), PIL_MEM_ZERO_INIT);
	memcpy (meter_entry->Specific, ds_meter->Specific, 
		sizeof (npf_ds_TBParam_t));
    } /* end if (Token Bucket Meter) */
    PRETURN (/*void*/);
}

/**************************************************************************
*
*   Function: dpe_copy_marker_params
*
*   Parameters: IN: pointer to destination
*               IN: pointer to copy struct from
*
*   Description: Copies the contents of the source structure into the dest
*
*   PRETURN : None 
*************************************************************************/
void dpe_copy_marker_params (npf_ds_Action_t* marker_entry, /* destination */
			     npf_ds_Action_t* ds_marker) /* source */
{
    QOS_DBG_TRACE();
    QOS_DBG_PRINT (LOG_INFO, "Copying Marker parameters!\n");
    /* copy parameters from npf_ds_Action_t */
    /* if DSCP Marker */
    QOS_DBG_PRINT (LOG_INFO, "Marker->Specific = 0x%x \n\t"
		   "*marker->Specific = 0x%x \n\t"
		   " Dscp = 0x%x \n\t",
		   ds_marker->Specific, *ds_marker->Specific,
		   ((npf_ds_DscpMarkAct_t*)ds_marker->Specific)->Dscp);

    if (*ds_marker->Specific == npf_ds_DscpMarkAct) {
	//if (!marker_entry->Specific)
	marker_entry->Specific =  PIL_AllocateMemory 
	    (sizeof (npf_ds_DscpMarkAct_t), PIL_MEM_ZERO_INIT);
	memcpy (marker_entry->Specific, ds_marker->Specific,
		sizeof (npf_ds_DscpMarkAct_t));
	QOS_DBG_PRINT (LOG_INFO, "DSCP mark = 0x%x\n\t",
		       ((npf_ds_DscpMarkAct_t*)marker_entry->Specific)->Dscp);

    } /* end if DSCP Marker */
    else //if (ds_marker->Specific == npf_ds_MPLSMarkAct) 
	{
	    //if (!marker_entry->Specific)
	    marker_entry->Specific = PIL_AllocateMemory 
		(sizeof (npf_ds_MPLSMarkAct_t), PIL_MEM_ZERO_INIT);
	    memcpy (marker_entry->Specific, ds_marker->Specific,
		    sizeof (npf_ds_MPLSMarkAct_t));
	}/* end if MPLS Marker */
    PRETURN (/*void*/);
}

/**************************************************************************
*
*   Function: dpe_copy_clfr_params
*
*   Parameters: IN: pointer to destination
*               IN: pointer to copy struct from
*
*   Description: Copies the contents of the source structure into the dest
*
*   PRETURN : None 
*************************************************************************/
void dpe_copy_clfr_params (npf_ds_BaseFilter_t* clfr_entry, /* destination */
			   npf_ds_BaseFilter_t* ds_clfr) /* source */
{
    QOS_DBG_TRACE();
    QOS_DBG_PRINT (LOG_INFO, "Copying classifier parameters!\n");
    /* copy the filter from basefilter_t */
    if (ds_clfr->type == npf_ds_IPv4Filter) {
	memcpy (clfr_entry, ds_clfr, sizeof (npf_ds_IPv4Filter_t));
    } /* end if (IPv4 filter) */
    else if (ds_clfr->type == npf_ds_IPv6Filter) {
	memcpy (clfr_entry, ds_clfr, sizeof (npf_ds_IPv6Filter_t));
    } /* end if (IPv6 filter) */
    else if (ds_clfr->type == npf_ds_MPLSFilter) {
	memcpy (clfr_entry, ds_clfr, sizeof (npf_ds_MPLSFilter_t));
    } /* end if (MPLS filter) */
    PRETURN (/*void*/);
}

/*************************************************************************
 *                                                                       *
 *                         DPE HELPER ROUTINES                           *
 *                                                                       *
 *************************************************************************/

/**************************************************************************
*
*   Function: dpe_dbg_print()
*
*   Parameters: dpe_TableType_t
*               dpe_entry
*
*   Description: Dumps debug info about the entry
*
*   PRETURN : None 
*************************************************************************/

void dpe_dbg_print (dpe_TableType_t table_type,
		    void*         dpe_entry )
{
    npf_ds_Scheduler_t*  ds_scheduler;
    npf_ds_Q_t*          ds_queue;
    npf_ds_AlgDrop_t*    ds_dropper;
    npf_ds_Meter_t*      ds_meter;
    npf_ds_Action_t*     ds_marker;
    npf_ds_BaseFilter_t* ds_clfr;
    dpe_scheduler_t*     sched_entry;
    dpe_queue_t*         queue_entry;
    dpe_dropper_t*       dropper_entry;
    dpe_meter_t*         meter_entry;
    dpe_marker_t*        marker_entry;
    dpe_clfr_t*          clfr_entry;
    DListItr*            itr;
    dpe_EntryId_t*       prev_entry;

    QOS_DBG_TRACE ();

    switch (table_type) {
    case dpe_Scheduler:
	{
	    QOS_DBG_PRINT (LOG_INFO, 
			   "SCHEDULER ENTRY\t"
			   "---------------\n\t");
	    sched_entry = (dpe_scheduler_t*) dpe_entry;
	    ds_scheduler = (npf_ds_Scheduler_t*) &(sched_entry->scheduler);
	    QOS_DBG_PRINT (LOG_INFO, 
			   "Table Id     = %d \n\t"
			   "Instance Id  = %d \n\t"
			   "Next Tbl Id  = %d \n\t"
			   "Next Inst Id = %d \n\t"
			   "Flow Id      = %d \n\t"
			   "Queue Id     = %d \n\t"
			   "Next Blk Id  = %d \n\t",
			   sched_entry->sched_tbl_id, 
			   sched_entry->sched_instance_id,
			   sched_entry->next_tbl_id,
			   sched_entry->next_instance_id,
			   sched_entry->flow_tuple.flow_id,
			   sched_entry->flow_tuple.queue_id,
			   sched_entry->flow_tuple.next_block_id);
	    QOS_DBG_PRINT (LOG_INFO, "Previous table-entry list!\n\t");
	    if ((itr = npf_list_itrCreate (&sched_entry->prev_tbl_entry_list)) 
		== NULL)
		QOS_DBG_PRINT (LOG_ERR, "Error traversing prev_entry list!\n");
	    if (npf_list_itrFirst (itr) != -1) {
		do {
		    prev_entry = (dpe_EntryId_t*) npf_list_itrGetData (itr);
		    QOS_DBG_PRINT (LOG_INFO, 
				   "Table Id    = %d \n\tInstance Id = %d \n\t",
				   prev_entry->table_id, 
				   prev_entry->instance_id);
		} while (npf_list_itrNext (itr) != -1);
	    } /* end start iterating on the list */
	    npf_list_itrDelete (itr);

	    QOS_DBG_PRINT (LOG_INFO, "Scheduler Parameters\n\t");
	    //dpe_dbg_print_scheduler_entry (ds_scheduler);
	}
	break;
    case dpe_Queue:
	{
	    QOS_DBG_PRINT (LOG_INFO, 
			   "QUEUE ENTRY\t"
			   "-----------\n\t");
	    queue_entry = (dpe_queue_t*) dpe_entry;
	    ds_queue = (npf_ds_Q_t*) &(queue_entry->queue);
	    QOS_DBG_PRINT (LOG_INFO, 
			   "Table Id     = %d \n\t"
			   "Instance Id  = %d \n\t"
			   "Next Tbl Id  = %d \n\t"
			   "Next Inst Id = %d \n\t"
			   "Flow Id      = %d \n\t"
			   "Queue Id     = %d \n\t"
			   "Next Blk Id  = %d \n\t",
			   queue_entry->queue_tbl_id, 
			   queue_entry->queue_instance_id,
			   queue_entry->next_tbl_id,
			   queue_entry->next_instance_id,
			   queue_entry->flow_tuple.flow_id,
			   queue_entry->flow_tuple.queue_id,
			   queue_entry->flow_tuple.next_block_id);
	    QOS_DBG_PRINT (LOG_INFO, "Previous table-entry list!\n\t");
	    if ((itr = npf_list_itrCreate (&queue_entry->prev_tbl_entry_list)) 
		== NULL)
		QOS_DBG_PRINT (LOG_ERR, "Error traversing prev_entry list!\n");
	    if (npf_list_itrFirst (itr) != -1) {
		do {
		    prev_entry = (dpe_EntryId_t*) npf_list_itrGetData (itr);
		    QOS_DBG_PRINT (LOG_INFO, 
				   "Table Id    = %d \n\tInstance Id = %d \n\t",
				   prev_entry->table_id, 
				   prev_entry->instance_id);
		} while (npf_list_itrNext (itr) != -1);
	    } /* end start iterating on the list */
	    npf_list_itrDelete (itr);
	    QOS_DBG_PRINT (LOG_INFO, "Queue Parameters\n\t");
	    //dpe_dbg_print_queue_entry (ds_queue);
	}
	break;
    case dpe_AlgDropper:
	{
	    QOS_DBG_PRINT (LOG_INFO, 
			   "DROPPER ENTRY\t"
			   "-------------\n\t");
	    dropper_entry = (dpe_dropper_t*) dpe_entry;
	    ds_dropper = (npf_ds_AlgDrop_t*) &(dropper_entry->dropper);
	    QOS_DBG_PRINT (LOG_INFO, 
			   "Table Id     = %d \n\t"
			   "Instance Id  = %d \n\t"
			   "Next Tbl Id  = %d \n\t"
			   "Next Inst Id = %d \n\t"
			   "Flow Id      = %d \n\t"
			   "Queue Id     = %d \n\t"
			   "Next Blk Id  = %d \n\t",
			   dropper_entry->dropper_tbl_id, 
			   dropper_entry->dropper_instance_id,
			   dropper_entry->next_tbl_id,
			   dropper_entry->next_instance_id,
			   dropper_entry->flow_tuple.flow_id,
			   dropper_entry->flow_tuple.queue_id,
			   dropper_entry->flow_tuple.next_block_id);
	    QOS_DBG_PRINT (LOG_INFO, "Previous table-entry list!\n\t");
	    if ((itr = npf_list_itrCreate (&dropper_entry->prev_tbl_entry_list)) 
		== NULL)
		QOS_DBG_PRINT (LOG_ERR, "Error traversing prev_entry list!\n");
	    if (npf_list_itrFirst (itr) != -1) {
		do {
		    prev_entry = (dpe_EntryId_t*) npf_list_itrGetData (itr);
		    QOS_DBG_PRINT (LOG_INFO, 
				   "Table Id    = %d \n\tInstance Id = %d \n\t",
				   prev_entry->table_id, 
				   prev_entry->instance_id);
		} while (npf_list_itrNext (itr) != -1);
	    } /* end start iterating on the list */
	    npf_list_itrDelete (itr);
	    QOS_DBG_PRINT (LOG_INFO, "Algo Dropper Parameters\n\t");
	    //dpe_dbg_print_dropper_entry (ds_dropper);
	}
	break;
    case dpe_Marker:
	{
	    QOS_DBG_PRINT (LOG_INFO, 
			   "MARKER ENTRY\t"
			   "------------\n\t");
	    marker_entry = (dpe_marker_t*) dpe_entry;
	    ds_marker = (npf_ds_Action_t*) &(marker_entry->action);
	    QOS_DBG_PRINT (LOG_INFO, 
			   "Table Id     = %d \n\t"
			   "Instance Id  = %d \n\t"
			   "Next Tbl Id  = %d \n\t"
			   "Next Inst Id = %d \n\t"
			   "Flow Id      = %d \n\t"
			   "Queue Id     = %d \n\t"
			   "Next Blk Id  = %d \n\t",
			   marker_entry->marker_tbl_id, 
			   marker_entry->marker_instance_id,
			   marker_entry->next_tbl_id,
			   marker_entry->next_instance_id,
			   marker_entry->flow_tuple.flow_id,
			   marker_entry->flow_tuple.queue_id,
			   marker_entry->flow_tuple.next_block_id);
	    QOS_DBG_PRINT (LOG_INFO, "Previous table-entry list!\n\t");
	    if ((itr = npf_list_itrCreate (&marker_entry->prev_tbl_entry_list)) 
		== NULL)
		QOS_DBG_PRINT (LOG_ERR, "Error traversing prev_entry list!\n");
	    if (npf_list_itrFirst (itr) != -1) {
		do {
		    prev_entry = (dpe_EntryId_t*) npf_list_itrGetData (itr);
		    QOS_DBG_PRINT (LOG_INFO, 
				   "Table Id    = %d \n\tInstance Id = %d \n\t",
				   prev_entry->table_id, 
				   prev_entry->instance_id);
		} while (npf_list_itrNext (itr) != -1);
	    } /* end start iterating on the list */
	    npf_list_itrDelete (itr);
	    QOS_DBG_PRINT (LOG_INFO, "Marker Parameters\n\t");
	    //dpe_dbg_print_marker_entry (ds_marker);
	}
	break;
    case dpe_Meter:
	{
	    QOS_DBG_PRINT (LOG_INFO, 
			   "METER ENTRY\t"
			   "-----------\n\t");
	    meter_entry = (dpe_meter_t*) dpe_entry;
	    ds_meter = (npf_ds_Meter_t*) &(meter_entry->meter);
	    QOS_DBG_PRINT (LOG_INFO, 
			   "Table Id     = %d \n\t"
			   "Instance Id  = %d \n\t"
			   "Next Tbl Id  = %d \n\t"
			   "Next Inst Id = %d \n\t"
			   "Upstream Flow Id      = %d \n\t"
			   "Upstream Queue Id     = %d \n\t"
			   "Downstream Flow id  = %d \n\t"
			   "Downstream queue id = %d \n\t"
			   "downstream nxt blk  = %d \n\t"
			   "Failnext flow id = %d \n\t"
			   "Failnext queue id= %d \n\t"
			   "failnext Next Blk Id  = %d \n\t",
			   meter_entry->meter_tbl_id, 
			   meter_entry->meter_instance_id,
			   meter_entry->next_tbl_id,
			   meter_entry->next_instance_id,
			   meter_entry->flow_tuple.flow_id,
			   meter_entry->flow_tuple.queue_id,
			   meter_entry->downstream_flow_tuple.flow_id,
			   meter_entry->downstream_flow_tuple.queue_id,
			   meter_entry->downstream_flow_tuple.next_block_id,
			   meter_entry->failnext_flow_tuple.flow_id,
			   meter_entry->failnext_flow_tuple.queue_id,
			   meter_entry->failnext_flow_tuple.next_block_id);
	    QOS_DBG_PRINT (LOG_INFO, "Previous table-entry list!\n\t");
	    if ((itr = npf_list_itrCreate (&meter_entry->prev_tbl_entry_list)) 
		== NULL)
		QOS_DBG_PRINT (LOG_ERR, "Error traversing prev_entry list!\n");
	    if (npf_list_itrFirst (itr) != -1) {
		do {
		    prev_entry = (dpe_EntryId_t*) npf_list_itrGetData (itr);
		    QOS_DBG_PRINT (LOG_INFO, 
				   "Table Id    = %d \n\tInstance Id = %d \n\t",
				   prev_entry->table_id, 
				   prev_entry->instance_id);
		} while (npf_list_itrNext (itr) != -1);
	    } /* end start iterating on the list */
	    npf_list_itrDelete (itr);
	    QOS_DBG_PRINT (LOG_INFO, "Meter Parameters\n\t");
	    //dpe_dbg_print_meter_entry (ds_meter);
	}
	break;
    case dpe_Classifier:
	{
	    QOS_DBG_PRINT (LOG_INFO, 
			   "CLASSIFIER ENTRY\t"
			   "----------------\n\t");
	    clfr_entry = (dpe_clfr_t*) dpe_entry;
	    ds_clfr = (npf_ds_BaseFilter_t*) (clfr_entry->filter);
	    QOS_DBG_PRINT (LOG_INFO, 
			   "Table Id     = %d \n\t"
			   "Instance Id  = %d \n\t"
			   "Next Tbl Id  = %d \n\t"
			   "Next Inst Id = %d \n\t"
			   "Flow Id      = %d \n\t"
			   "Queue Id     = %d \n\t"
			   "Next Blk Id  = %d \n\t",
			   clfr_entry->clfr_tbl_id, 
			   clfr_entry->clfr_instance_id,
			   clfr_entry->next_tbl_id,
			   clfr_entry->next_instance_id,
			   clfr_entry->flow_tuple.flow_id,
			   clfr_entry->flow_tuple.queue_id,
			   clfr_entry->flow_tuple.next_block_id);
	    QOS_DBG_PRINT (LOG_INFO, "Previous table-entry list!\n\t");
	    if ((itr = npf_list_itrCreate (&clfr_entry->prev_tbl_entry_list)) 
		== NULL)
		QOS_DBG_PRINT (LOG_INFO, "No back pointers!\n");
	    if (npf_list_itrFirst (itr) != -1) {
		do {
		    prev_entry = (dpe_EntryId_t*) npf_list_itrGetData (itr);
		    QOS_DBG_PRINT (LOG_INFO, 
				   "Table Id    = %d \n\tInstance Id = %d \n\t",
				   prev_entry->table_id, 
				   prev_entry->instance_id);
		} while (npf_list_itrNext (itr) != -1);
	    } /* end start iterating on the list */
	    npf_list_itrDelete (itr);
	    QOS_DBG_PRINT (LOG_INFO, "Classifier Parameters\n\t");
	    //dpe_dbg_print_clfr_entry (ds_clfr);
	}
	break;
    } /* end switch (table_id) to print an entry */

    PRETURN (/*void*/);
}

/**************************************************************************
*
*   Function: dpe_dbg_print_tbl_entry()
*
*   Parameters: dpe_TableType_t
*               instance id
*
*   Description: Dumps debug info about the entry indexed by the instance
*
*   PRETURN : None 
*************************************************************************/

void dpe_dbg_print_tbl_entry (npf_ds_TableId_t table_id, 
			      npf_ds_InstanceId_t inst_id)
{
    uint32_t instance_id;

    QOS_DBG_TRACE ();
    
    instance_id = (uint32_t) inst_id;
    switch (table_id) {
    case DPE_SCHEDULER_TABLE_ID:
	dpe_dbg_print (dpe_Scheduler, 
		       (dpe_scheduler_t*) g_dpe_scheduler_tbl[instance_id]);
	break;
    case DPE_QUEUE_TABLE_ID:
	dpe_dbg_print (dpe_Queue, 
		       (dpe_queue_t*) g_dpe_queue_tbl[instance_id]);
	break;
    case DPE_ALG_DROPPER_TABLE_ID:
	dpe_dbg_print (dpe_AlgDropper, 
		       (dpe_dropper_t*) g_dpe_dropper_tbl[instance_id]);
	break;
    case DPE_METER_TABLE_ID:
	dpe_dbg_print (dpe_Meter, 
		       (dpe_meter_t*) g_dpe_meter_tbl[instance_id]);
	break;
    case DPE_MARKER_TABLE_ID:
	dpe_dbg_print (dpe_Marker, 
		       (dpe_marker_t*) g_dpe_marker_tbl[instance_id]);
	break;
    case DPE_CLASSIFIER_TABLE_ID:
	dpe_dbg_print (dpe_Classifier, 
		       (dpe_clfr_t*) g_dpe_clfr_tbl[instance_id]);
	break;
    } /* end switch on table id for printing */
    PRETURN (/*void*/);
}
