/*****************************************************************************
*                  INTEL CORPORATION PROPRIETARY INFORMATION
*  This software is supplied under the terms of a license agreement or
*  nondisclosure agreement with Intel Corporation and may not be copied
*  or disclosed except in accordance with the terms of that agreement.
*  Copyright (c) 2002 Intel Corporation. All Rights Reserved.
*****************************************************************************/

/*****************************************************************************
* $Source: /cvs/constellations/ixa_3.0/src/cp_pdk/ForwardingPlane/FPModule/QoSManager/src/qos_event_callbacks.c,v $                                                                     * $Revision: 1.5 $
* $Date: 2003/11/17 16:57:26 $
*
* Description:
*     Implementation of the QoS flow id event code. This includes the
*     callback mechanism as well as some query calls to query for flow ids.
*
*****************************************************************************/


#include "qos_event_callbacks.h"
#include "list.h"
#include "log.h"
#define DBG_TRACE() npf_logger_Write(LOG_FP_QOS, LOG_TRACE, "%s\n", "entering");
#ifdef RETURN
#undef RETURN
#endif
#define RETURN( val ) \
    { \
        npf_logger_Write(LOG_FP_QOS, LOG_TRACE, "%s\n", "leaving"); \
        return val; \
    }
#define DBG_PRINT( lvl, fmt, args... ) \
        npf_logger_Write(LOG_FP_QOS, lvl, fmt, ## args)


/*****************************************************************************
*    Name: g_cbhandle
*
*    Description:
*        The counter used to generate callback handles when callbacks are
*        registered.
*
*/
static uint32_t g_cbhandle = 0;

/*****************************************************************************
*    Name: flow_event_callbacks
*
*    Description:
*        This list maintains the state for the registered callbacks.
*
*/
static DList /*<FlowEventCallbackInfo_t>*/ flow_event_callbacks;

/*****************************************************************************
*    Name: GetCBHandle
*
*    Description:
*        Allocates a new callback handle. Used when a callback is registered.
*
*    Parameters:
*        None
*
*    Returns:
*        The new callback handle.
*
*/
static uint32_t GetCBHandle() {
    DBG_TRACE();
    RETURN( ++g_cbhandle );
}

/*****************************************************************************
*    Name: InitCBInfo
*
*    Description:
*        Utility function used to initialize a callback info structure.
*
*    Parameters:
*        OUT: cbinfo   - A pointer to the callback info structure to initialize.
*        IN:  callback - The callback function that is being registered.
*        IN:  context  - The context to initialize the callback info with.
*
*    Preconditions:
*        cbinfo != NULL
*
*    Returns:
*        0 if successful, -1 otherwise.
*
*/
static int InitCBInfo( 
        FlowEventCallbackInfo_t *cbinfo, 
        uint32_t context,
        QOS_FLOW_EVENT_CB callback ) {

    DBG_TRACE();
    if ( !cbinfo ) {
        DBG_PRINT( LOG_ERR, "Got a NULL cbinfo pointer\n" );
        RETURN( -1 );
    }
    cbinfo->callback = callback;
    cbinfo->context = context;
    cbinfo->cbhandle = GetCBHandle();
    RETURN( 0 );
}

/*****************************************************************************
*    Name: RemoveCallback
*
*    Description:
*        Utility function that is used to remove a callback from the internal
*        list of registered callbacks. This function does not deallocate any
*        memory. It simply finds the callback based on the callback handle,
*        removes the callback information from the list if it is found and
*        returns a pointer to the callback info. It is up to the calling
*        function to free the callback info as needed.
*
*    Parameters:
*        OUT: cbinfo      - A pointer to a callback info struct pointer. This
*                           is only valid if the callback that is being
*                           removed is found. The calling function would use
*                           this pointer to deallocate space reserved for the
*                           callback info.
*        IN:  handle      - The callback handle to search for.
*        IN:  cbinfo_list - The callback info list to search thru.
*
*    Preconditions:
*        cbinfo != NULL
*
*    Postconditions:
*        If the handle is found the callback info will be removed from the
*        list.
*
*    Returns:
*        0 for success, -1 otherwise.
*
*/
static uint32_t RemoveCallback(
        FlowEventCallbackInfo_t **cbinfo,
        uint32_t handle,
        DList *cbinfo_list ) {

    DListItr *itr;
    uint32_t retval = -1;

    DBG_TRACE();
    if ( !cbinfo ) {
        DBG_PRINT( LOG_ERR, "Got a NULL cbinfo pointer\n" );
        RETURN( -1 );
    }
    if ( ( itr = npf_list_itrCreate( cbinfo_list ) ) == NULL ) {
        DBG_PRINT( LOG_ERR, "Unable to create list iterator\n" );
        RETURN( -1 );
    }
    npf_list_itrFirst( itr );
    do {
        (*cbinfo) = npf_list_itrGetData( itr );
        if ( (*cbinfo)->cbhandle == handle ) {
            npf_list_itrRemove( itr, (void **)cbinfo );
            retval = 0;
            DBG_PRINT( LOG_INFO, "Found the cbinfo to remove\n" );
        }
    } while ( npf_list_itrNext( itr ) != -1 && retval == -1 );
    npf_list_itrDelete( itr );

    RETURN( retval );
}

/*****************************************************************************
*    Name: RegisterPendingFlowCB
*
*    Description:
*        Allows a client to register a callback for flow id events.
*
*    Parameters:
*        IN:  context  - This value will be associated with the callback
*                        function and will be passed back to the client when
*                        the callback is invoked.
*        IN:  callback - The callback function to register.
*        OUT: cbhandle - An id generated by the QoS Manager that will be used
*                        by QoS Manager to identify the registered callback.
*                        This will be used by the client when the callback is
*                        deregistered.
*
*    Preconditions:
*        cbhandle != NULL
*
*    Returns:
*        0 if successful, -1 otherwise.
*
*/
int RegisterPendingFlowCB( uint32_t context,
        QOS_FLOW_EVENT_CB callback, uint32_t *cbhandle ) {

    FlowEventCallbackInfo_t *cbinfo = (FlowEventCallbackInfo_t *)PIL_AllocateMemory( sizeof( FlowEventCallbackInfo_t ), PIL_MEM_ZERO_INIT );
    
    DBG_TRACE();
    if ( !cbinfo ) {
        DBG_PRINT( LOG_CRIT, "Unable to allocate cbinfo pointer\n" );
        RETURN( -1 );
    }
    if ( !cbhandle ) {
        DBG_PRINT( LOG_ERR, "Got a NULL cbhandle pointer\n" );
        RETURN( -1 );
    }
    if ( InitCBInfo( cbinfo, context, callback ) != 0 ) {
        PIL_FreeMemory( cbinfo );
        DBG_PRINT( LOG_ERR, "Unable to initialize cbinfo struct\n" );
        RETURN( -1 );
    }
    npf_list_pushBack( &flow_event_callbacks, (void *)cbinfo );
    *cbhandle = cbinfo->cbhandle;
    DBG_PRINT( LOG_INFO, "Successfully registered callback\n" );
    RETURN( 0 );
}

/*****************************************************************************
*    Name: DeregisterPendingFlowCB
*
*    Description:
*        Allows a client to deregister its flow id event callback.
*
*    Parameters:
*        IN:  cbhandle - The callback handle that was returned to the client
*                        when the callback was registered.
*
*    Returns:
*        0 if succesful, -1 otherwise.
*
*/
int DeregisterPendingFlowCB( uint32_t cbhandle ) {

    FlowEventCallbackInfo_t *cbinfo;
    int retval;

    DBG_TRACE();
    retval = RemoveCallback( &cbinfo, cbhandle, &flow_event_callbacks );

    if ( retval == 0 ) {
        PIL_FreeMemory( cbinfo );
        DBG_PRINT( LOG_INFO, "Successfully removed callback\n" );
        RETURN( 0 );
    }
    RETURN( -1 );
}

int InitFlowEventCallback() {
    npf_list_init( &flow_event_callbacks, PIL_FreeMemory );
    return 0;
}

/*****************************************************************************
*    Name: InvokeFlowEventCallback
*
*    Description:
*        This utility function is called when the event callbacks need to
*        be called. The parameters passed in will be passed to the clients
*        via the callback.
*
*    Parameters:
*        IN:  type   - The type of classification entry that was added.
*        IN:  key    - The classification entry.
*        IN:  flowid - The flow id associated with the classification entry.
*
*    Preconditions:
*        key != NULL
*
*    Returns:
*        0 if successful, -1 otherwise.
*
*/
int InvokeFlowEventCallback( uint32_t type, void *key, uint32_t flowid,
        uint32_t queueid ) {

    DListItr *itr;
    FlowEventCallbackInfo_t *cbinfo = NULL;

    DBG_TRACE();
    if ( ( itr = npf_list_itrCreate( &flow_event_callbacks ) ) == NULL ) {
        DBG_PRINT( LOG_ERR, "Unable to create list iterator\n" );
        RETURN( -1 );
    }
    npf_list_itrFirst( itr );
    do {
        cbinfo = npf_list_itrGetData( itr );
        if ( cbinfo ) {
            DBG_PRINT( LOG_INFO, "Invoking client callback\n" );
            cbinfo->callback( cbinfo->context, type, key, flowid, queueid );
            DBG_PRINT( LOG_INFO, "Returned from client callback\n" );
        }
        else {
            DBG_PRINT( LOG_CRIT, "Found a NULL pointer in callback list\n" );
        }
    } while ( npf_list_itrNext( itr ) != -1 );
    npf_list_itrDelete( itr );
    RETURN( 0 );
}
