/*****************************************************************************
*                  INTEL CORPORATION PROPRIETARY INFORMATION
*  This software is supplied under the terms of a license agreement or
*  nondisclosure agreement with Intel Corporation and may not be copied
*  or disclosed except in accordance with the terms of that agreement.
*  Copyright (c) 2001 Intel Corporation. All Rights Reserved.
*****************************************************************************/
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <assert.h>
#include <errno.h>

#include "FPModule.h"
#include "ixasdk20_macros.h"
#include "fp_common.h"
#include "npf_ds_types.h"
#include "dpe_mgr.h"
#include "log.h"
#include "qos_mgr.h"

/* include file for Resource Manager */
#include <ix/microace/rm.h>
/* include file for all QoS Ace stubs, etc. */
#include <idl/qos_types_stub_c.h>
#include <idl/classifier_stub_c.h>
#include <idl/dscp_classifier_stub_c.h>
#include <idl/ipv4_range_classifier_stub_c.h>
#include <idl/marker_stub_c.h>
#include <idl/dscp_marker_stub_c.h>
#include <idl/meter_stub_c.h>
#include <idl/srtc_meter_stub_c.h>
#include <idl/bufmgr_stub_c.h>
#include <idl/wred_bufmgr_stub_c.h>
#include <idl/scheduler_stub_c.h>
#include <idl/drr_scheduler_stub_c.h>

extern router_config    g_current_router_config;
static ix_error         err =0;
static ix_base_t        g_clfr_ace, g_marker_ace, g_meter_ace;
static ix_base_t        g_bufmgr_ace, g_sched_ace;
static ix_ossl_thread_t g_clfr_ace_cap_thread, g_marker_ace_cap_thread;
static ix_ossl_thread_t g_meter_ace_cap_thread, g_bufmgr_ace_cap_thread;
static ix_ossl_thread_t g_sched_ace_cap_thread;

/* these variables define how data path elements can be linked together
   on the specific platform/NPU */
static uint32_t         g_clfr_count, g_marker_count, g_meter_count; 
static uint32_t         g_dropper_count, g_queue_count, g_scheduler_count;
static uint32_t         g_shaper_count;

static uint32_t         g_max_clfr_next_tbl_count, g_max_meter_next_tbl_count;
static uint32_t         g_max_marker_next_tbl_count;
static uint32_t         g_max_dropper_next_tbl_count;
static uint32_t         g_max_queue_next_tbl_count;
static uint32_t         g_max_scheduler_next_tbl_count;
static uint32_t         g_max_shaper_next_tbl_count;

/* some constants */
#define WRED_BUFMGR_EWMA          2   /* Default value for EWMA */
#define WRED_BUFMGR_MAX_Q_DEPTH   128 /* Default Max Q Depth (rmconfig.h)*/
#define WRED_BUFMGR_MIN_Q_DEPTH   127 /* Default Min Q Depth (rmconfig.h)*/
#define WRED_BUFMGR_MAX_DROP_PROB 1   /* Default Max Drop probability */
#define DRR_SCHEDULER_CREDIT_DEF  65534 /* Default Credit limit for DRR sched */
#define DRR_SCHEDULER_QUANTUM_DEF 64  /* Default Quantum for scheduler */
#define WRED_NEXT_BLOCK_ID       8    /* Next Block Id of Meter - Queue */
#define METER_NEXT_BLOCK_ID      6
#define MARKER_NEXT_BLOCK_ID     7

static float g_drr_quantum_factor;

static ix_error idl_thread_run(void* cookie, void** dummy)
{
    (void)dummy;
    return ix_cap_loop((ix_cap*)cookie);
}

/**************************************************************************
*
*   Function: generate_quantum()
*
*   Parameters: IN: npf_ds_MinRate_t
*               IN: npf_ds_MaxRate_t
*               OUT: Quantum for DRR scheduler
*
*   Description: This function takes in the Queue parameters (currently only
*                Min Rate is supported) and generates a quantum number for the
*                DRR scheduler.
*
*   PRETURN : 0
*************************************************************************/
static int generate_quantum (npf_ds_MinRate_t* min_rate,
			     npf_ds_MaxRate_t* max_rate,
			     quantum_t* quantum)
{
    uint32_t priority, abs, rel;

    priority = min_rate->Priority;
    abs      = min_rate->Abs;
    rel      = min_rate->Rel;

    *quantum  = rel * g_drr_quantum_factor;
    QOS_DBG_PRINT (LOG_INFO, "Scheduler Quantum = %ld\n",
		   (unsigned long) *quantum);

    return (0);
}

/**************************************************************************
*
*   Function: qos_init_fp()
*
*   Parameters: None
*
*   Description: Performs platform specific initializations. For IXA SDK 2.0,
*                this involves creating ACE proxies for all the correct ACEs
*                being used from the Sauvie Island QoS components.
*
*                Based on the current router configuration, this function 
*                initializes only specific ACEs, and the related stub functions.
*
*   PRETURN : 0 for success, -1 for failure
*************************************************************************/

int qos_init_fp ()
{
    QOS_DBG_TRACE ();
    QOS_DBG_PRINT (LOG_INFO, "Initializing stubs...\n");
    QOS_DBG_PRINT (LOG_INFO, "Current router configuration = %d\n", 
		   g_current_router_config);

    InitFlowEventCallback();

    g_drr_quantum_factor = ((float) DRR_SCHEDULER_QUANTUM_DEF)/50;
    QOS_DBG_PRINT (LOG_INFO, "Quantum Factor ALPHA = %f \n",
		   g_drr_quantum_factor);
    

    switch (g_current_router_config) {
    case IPv4_ROUTER:
	break;
    case MPLS:
	break;
    case DS_MPLS_LER:
    case DS_MPLS_LSR:
	{
	    /* Classifier and Marker are MPLS (Strawberry Hill) components
	       MPLS Manager is responsible for intitializing them
	       No QoS Aces to be initialized, other than common ACEs
	       -- Meter, Buffer Manager (Algo Dropper, Queue)
	       -- Scheduler  (DRR Scheduler, Shaper)
	    */
	    QOS_DBG_PRINT (LOG_INFO, "Configuring for DS_MPLS_ILER !\n");
	    memset (&g_meter_ace, 0, sizeof(ix_base_t));
	    g_meter_ace.m_pCap = (ix_cap*) 
		PIL_AllocateMemory (sizeof(ix_cap), PIL_MEM_ZERO_INIT);
	    if (!g_meter_ace.m_pCap) {
		QOS_DBG_PRINT (LOG_CRIT, "Unable to allocate Meter ACE Cap!\n");
		exit (1);
	    }
	    ix_cap_init (g_meter_ace.m_pCap, "srtc_meter");
	    ix_ossl_create_thread (&g_meter_ace_cap_thread, idl_thread_run, 
				   (void *)g_meter_ace.m_pCap);
	    //ECK (stub_meter_init(&g_meter_ace, g_meter_ace.m_pCap ));
	    ECK (stub_srtc_meter_init(&g_meter_ace, g_meter_ace.m_pCap ));
	    ix_oms_connect (&g_meter_ace, "srtc_meter");
	    ix_oms_wait_connected (&g_meter_ace);
	    QOS_DBG_PRINT (LOG_INFO, "Created meter proxy!\n");
	}
	break;
	/***************************************************************
	 * Pure QoS cases are not supported currently
	 ***************************************************************/
    case DS_EDGE:
	{
	    QOS_DBG_PRINT (LOG_INFO, "Configuring for DS_EDGE!\n");
	    memset (&g_clfr_ace, 0, sizeof(ix_base_t));
	    g_clfr_ace.m_pCap = (ix_cap *) 
		PIL_AllocateMemory (sizeof(ix_cap), PIL_MEM_ZERO_INIT);
	    if (!g_clfr_ace.m_pCap) {
		QOS_DBG_PRINT (LOG_CRIT, "Unable to alloc  Clfr ACE CAP\n");
		exit (1);
	    }
	    ix_cap_init (g_clfr_ace.m_pCap, "ipv4_range_classifier");
	    
	    memset (&g_marker_ace, 0, sizeof(ix_base_t));
	    g_marker_ace.m_pCap = (ix_cap *) 
		PIL_AllocateMemory (sizeof(ix_cap), PIL_MEM_ZERO_INIT);
	    if (!g_marker_ace.m_pCap) {
		QOS_DBG_PRINT (LOG_CRIT, "Unable to alloc DSCP Marker ACE"
			       "CAP\n");
		exit (1);
	    }
	    ix_cap_init (g_marker_ace.m_pCap, "dscp_marker");
	    
	    memset (&g_meter_ace, 0, sizeof(ix_base_t));
	    g_meter_ace.m_pCap = (ix_cap*) 
		PIL_AllocateMemory (sizeof(ix_cap), PIL_MEM_ZERO_INIT);
	    if (!g_meter_ace.m_pCap) {
		QOS_DBG_PRINT (LOG_CRIT, "Unable to alloc Meter ACE Cap!\n");
		exit (1);
	    }
	    ix_cap_init (g_meter_ace.m_pCap, "srtc_meter");
	    
	    /* Create a thread for the ACEs */
	    
	    ix_ossl_create_thread (&g_clfr_ace_cap_thread, idl_thread_run, 
				   (void *)g_clfr_ace.m_pCap);
	    ix_ossl_create_thread (&g_meter_ace_cap_thread, idl_thread_run, 
				   (void *)g_meter_ace.m_pCap);
	    ix_ossl_create_thread (&g_marker_ace_cap_thread, idl_thread_run, 
				   (void *)g_marker_ace.m_pCap);
	    
	    /* initialize the stubs of the ACEs */
	    //ECK (stub_classifier_init(&g_clfr_ace, g_clfr_ace.m_pCap));
	    ECK (stub_ipv4_range_classifier_init (&g_clfr_ace, 
						  g_clfr_ace.m_pCap));
	    //ECK (stub_meter_init(&g_meter_ace, g_meter_ace.m_pCap));
	    ECK (stub_srtc_meter_init(&g_meter_ace, g_meter_ace.m_pCap));
	    //ECK (stub_marker_init (&g_marker_ace, g_marker_ace.m_pCap));
	    ECK (stub_dscp_marker_init (&g_marker_ace, 
					g_marker_ace.m_pCap ));
	    
	    QOS_DBG_PRINT (LOG_INFO, "Connecting to OMS...\n");
	    /* connect to the actual Aces running on the micro-engine. 
	       The names of these Aces have to be the ones registered with 
	       the OMS Naming Service*/
	    ix_oms_connect (&g_clfr_ace, "ipv4_range_classifier");
	    ix_oms_connect (&g_marker_ace, "dscp_marker");
	    ix_oms_connect (&g_meter_ace, "srtc_meter");

	    /* Wait in a connected state for things to happen */ 
	    ix_oms_wait_connected (&g_clfr_ace);
	    ix_oms_wait_connected (&g_marker_ace);
	    ix_oms_wait_connected (&g_meter_ace);
	}
	break;
    case DS_CORE:
	{
	    /*
	      - DSCP Classifier
	      - No Marker
	      - Meter, WRED Buf Mgr, DRR Scheduler
	    */
	    QOS_DBG_PRINT (LOG_INFO, "Configuring for DS_CORE !\n");
	    memset (&g_clfr_ace, 0, sizeof(ix_base_t));
	    g_clfr_ace.m_pCap = (ix_cap *) 
		PIL_AllocateMemory (sizeof(ix_cap), PIL_MEM_ZERO_INIT);
	    if (!g_clfr_ace.m_pCap) {
		QOS_DBG_PRINT (LOG_CRIT, "Unable to allocate  Classifier ACE"
			       "CAP\n");
		exit (1);
	    }
	    ix_cap_init (g_clfr_ace.m_pCap, "dscp_classifier");

	    /* Create a thread for the ACEs */
	    ix_ossl_create_thread (&g_clfr_ace_cap_thread, idl_thread_run, 
				   (void *)g_clfr_ace.m_pCap);

	    /* initialize the stubs of the ACEs */
	    //ECK (stub_classifier_init (&g_clfr_ace, g_clfr_ace.m_pCap));
	    ECK (stub_dscp_classifier_init (&g_clfr_ace, 
					    g_clfr_ace.m_pCap ));
	    QOS_DBG_PRINT (LOG_INFO, "Connecting to OMS...\n");
	    /* connect to the actual Aces running on the micro-engine. 
	       The names of these Aces have to be the ones registered with 
	       the OMS Naming Service*/
	    ix_oms_connect (&g_clfr_ace, "dscp_classifier");
	    /* Wait in a connected state for things to happen */ 
	    ix_oms_wait_connected (&g_clfr_ace);
	}
	break;
    } /* end switch (router_config) */

    /* Common ACEs that are required for all cases are
       -- WRED Buffer Manager
       -- DRR  Scheduler
    */
    
    memset (&g_bufmgr_ace, 0, sizeof(ix_base_t));
    g_bufmgr_ace.m_pCap = (ix_cap*) PIL_AllocateMemory (sizeof(ix_cap),
							PIL_MEM_ZERO_INIT);
    if (!g_bufmgr_ace.m_pCap) {
	QOS_DBG_PRINT (LOG_CRIT, "Unable to allocate Buf Mgr ACE Cap!\n");
	exit (1);
    }
    ix_cap_init (g_bufmgr_ace.m_pCap, "wred_bufmgr");
    
    memset (&g_sched_ace, 0, sizeof(ix_base_t));
    g_sched_ace.m_pCap = (ix_cap*) PIL_AllocateMemory (sizeof(ix_cap),
						       PIL_MEM_ZERO_INIT);
    if (!g_sched_ace.m_pCap) {
	QOS_DBG_PRINT (LOG_CRIT, "Unable to allocate Scheduler ACE Cap!\n");
	exit (1);
    }
    ix_cap_init (g_sched_ace.m_pCap, "drr_scheduler");
    /* create OSSL threads for the CAP (Communication Access
       Point) */
    ix_ossl_create_thread (&g_bufmgr_ace_cap_thread, idl_thread_run, 
			   (void *)g_bufmgr_ace.m_pCap);
    ix_ossl_create_thread (&g_sched_ace_cap_thread, idl_thread_run, 
			   (void *)g_sched_ace.m_pCap);
    /* initialize the stubs of the ACEs */

    //ECK (stub_bufmgr_init (&g_bufmgr_ace, g_bufmgr_ace.m_pCap));
    ECK (stub_wred_bufmgr_init(&g_bufmgr_ace, g_bufmgr_ace.m_pCap ));
    //ECK (stub_scheduler_init (&g_sched_ace, g_sched_ace.m_pCap));
    ECK (stub_drr_scheduler_init(&g_sched_ace, g_sched_ace.m_pCap ));
    
    ix_oms_connect (&g_bufmgr_ace, "wred_bufmgr");
    ix_oms_connect (&g_sched_ace, "drr_scheduler");
    /* Wait in a connected state for things to happen */ 
    ix_oms_wait_connected (&g_bufmgr_ace);
    ix_oms_wait_connected (&g_sched_ace);
    
    switch (g_current_router_config) {
    case IPv4_ROUTER:
	    break;
    case MPLS:
    case DS_MPLS_LER:
    case DS_MPLS_LSR:
	{
	    /* # of instances data path elements  */
	    g_clfr_count = g_meter_count = g_shaper_count = 1;
	    g_dropper_count = g_queue_count = g_scheduler_count = 1;
	    g_marker_count = 0;
 
	    g_max_clfr_next_tbl_count = 1; /* Classifier --> Meter */
	    g_max_marker_next_tbl_count = 0; /* NO Marker */
	    g_max_meter_next_tbl_count = 2; /* Meter --> Dropper, Q */
	    g_max_dropper_next_tbl_count = 1; /* Dropper --> Queue */
	    g_max_queue_next_tbl_count = 1; /* Queue --> Scheduler */
	    g_max_shaper_next_tbl_count = 0; /* Shaper --> Scheduler */
	    g_max_scheduler_next_tbl_count = 0; /* Scheduler --> 0 */
	}
	break;
    case DS_EDGE:
	{
	    /* # of instances data path elements  */
	    g_clfr_count = g_marker_count = g_meter_count = g_shaper_count = 1;
	    g_dropper_count = g_queue_count = g_scheduler_count = 1;
	    
	    g_max_clfr_next_tbl_count = 4; /* Clfr --> Meter, Marker, Dropper, 
					      Queue */
	    g_max_meter_next_tbl_count = 3; /* Meter --> Marker, Dropper, 
					       Queue */
	    g_max_marker_next_tbl_count = 3; /* Marker --> Meter, Dropper, 
						Queue */
	    g_max_dropper_next_tbl_count = 1; /* Dropper --> Queue */
	    g_max_queue_next_tbl_count = 1; /* Queue --> Shaper */
	    g_max_shaper_next_tbl_count = 1; /* Shaper --> Scheduler */
	    g_max_scheduler_next_tbl_count = 0; /* Scheduler --> 0 */
	}
	break;
    case DS_CORE:
	{
	    /* # of instances data path elements  */
	    g_clfr_count = g_meter_count = g_shaper_count = 1;
	    g_dropper_count = g_queue_count = g_scheduler_count = 1;
	    g_marker_count = 0;

	    g_max_clfr_next_tbl_count = 3; /* Classifier --> Meter, Dropper,
					    Queue */
	    g_max_marker_next_tbl_count = 0; /* NO Marker */
	    g_max_meter_next_tbl_count = 2; /* Meter --> Dropper, Queue */
	    g_max_dropper_next_tbl_count = 1; /* Dropper --> Queue */
	    g_max_queue_next_tbl_count = 1; /* Queue --> Shaper */
	    g_max_shaper_next_tbl_count = 1; /* Shaper --> Scheduler */
	    g_max_scheduler_next_tbl_count = 0; /* Scheduler --> 0 */
	}
	break;
    }

    PRETURN (0);
}

/**************************************************************************
*
*   Function: qos_shutdown_fp()
*
*   Parameters: None
*
*   Description: Performs platform specific cleanup/shutdown
*
*   PRETURN : 0 on success, -1 for failure 
*************************************************************************/

int qos_shutdown_fp ()
{
    /*disconnecting from OMS */
    QOS_DBG_TRACE ();
    switch (g_current_router_config) {
    case IPv4_ROUTER:
	break;
    case MPLS:
	break;
    case DS_MPLS_LER:
    case DS_MPLS_LSR:
	{
	    QOS_DBG_PRINT (LOG_INFO, "Shutting down DS_MPLS_ILER!\n");
	    /* Nothing to do for Classifier, Marker */
	    
	    ECK (ix_oms_disconnect (&g_meter_ace));
	    //ECK (stub_meter_fini (&g_meter_ace));
	    ECK (stub_srtc_meter_fini (&g_meter_ace));
	    
	    PIL_FreeMemory ((void*)g_meter_ace.m_pCap);
	}
	break;
    case DS_EDGE:
	{
	    QOS_DBG_PRINT (LOG_INFO, "Shutting down DS_EDGE !\n");

	    ECK (ix_oms_disconnect(&g_clfr_ace));
	    ECK (ix_oms_disconnect(&g_meter_ace));

	    /*destroying stubs */
	    //ECK (stub_classifier_fini (&g_clfr_ace));
	    ECK (stub_ipv4_range_classifier_fini(&g_clfr_ace));

	    //ECK (stub_meter_fini (&g_meter_ace));
	    ECK (stub_srtc_meter_fini (&g_meter_ace));

	    PIL_FreeMemory ((void*)g_clfr_ace.m_pCap);
	    PIL_FreeMemory ((void*)g_meter_ace.m_pCap);
	}
	break;
    case DS_CORE:
	{
	    QOS_DBG_PRINT (LOG_INFO, "Shutting down DS_CORE !\n");
	    ECK (ix_oms_disconnect(&g_clfr_ace));

	    /*destroying stubs */
	    //ECK (stub_classifier_fini (&g_clfr_ace));
	    ECK (stub_dscp_classifier_fini(&g_clfr_ace));

	    PIL_FreeMemory ((void*)g_clfr_ace.m_pCap);
	}
	break;
    } /* end switch (router_config) */

    ECK (ix_oms_disconnect(&g_bufmgr_ace));
    ECK (ix_oms_disconnect(&g_sched_ace));

    /*destroying stubs */
    //ECK (stub_bufmgr_fini (&g_bufmgr_ace));
    ECK (stub_wred_bufmgr_fini(&g_bufmgr_ace));
    //ECK (stub_scheduler_fini (&g_sched_ace));
    ECK (stub_drr_scheduler_fini(&g_sched_ace));

    PIL_FreeMemory ((void*)g_bufmgr_ace.m_pCap);
    PIL_FreeMemory ((void*)g_sched_ace.m_pCap);

    /* Close communication with Resource Manager */
    QOS_DBG_PRINT (LOG_INFO, "Closing communication with Resource Manager!\n");
    err = RmTerm ();
    if (err != 0) {
	QOS_DBG_PRINT (LOG_INFO, "Error disconnecting from Resource Manager!\n");
    }
    PRETURN (err);
}

/**************************************************************************
*
*   Function: qos_get_caps
*
*   Parameters: npf_ds_IfCaps_t* Interface capabilities
*
*   Description: Returns the QoS Capabilities supported on this data plane
*
*   PRETURN : 0
*************************************************************************/

int qos_get_caps (npf_ds_IfCaps_t* qos_caps)
{
    QOS_DBG_TRACE ();

    /*	Common components - SRTC Meter, WRED Buf Mgr, DRR Scheduler */

    /* Basic Classification capabilities */
    qos_caps->ClfrCaps_Count = g_clfr_count;
    qos_caps->ClfrCaps_list[0].Base.type = npf_ds_ClfrCaps;
    qos_caps->ClfrCaps_list[0].Base.tableid = DPE_CLASSIFIER_TABLE_ID;
    qos_caps->ClfrCaps_list[0].Base.Direction = npf_ds_IngressOnly;
    qos_caps->ClfrCaps_list[0].Spec = 0;
    qos_caps->ClfrCaps_list[0].Base.stats_supported = 0;
    qos_caps->ClfrCaps_list[0].Base.next_tables_count = 
	g_max_clfr_next_tbl_count;
    switch (g_current_router_config) {
    case IPv4_ROUTER:
	break;
    case MPLS:
	break;
    case DS_MPLS_LER:
	{
	    /* Classifier = CIDR-based FEC classifier
	       Marker     = MPLS marker
	    */
	    QOS_DBG_PRINT (LOG_INFO, "Collecting Caps for DS_MPLS_ILER!\n");
	    qos_caps->ClfrCaps_list[0].Spec |= 0x031F; 
	    /* Bit 10 - MPLS FEC
	       Bit 9  - MPLS Label
	       Bits 0-5 - IPv4 and DSCP filter. */
	    /****************************************************
	     * this is a hack, actually. Even though FP can report IPv4 and
	     * DSCP filters separately, code in the DiffServ module inthe
	     * control plane checks only for IPv4. So, if FP reports DSCP,
	     * this cannot be validated
	     ******************************************************/
	    qos_caps->ClfrCaps_list[0].Base.stats_supported = 0x0001; 
	    /* Pkt Count */
	    /* Classifier can be followed by Marker ONLY in MPLS+QoS */
	    qos_caps->ClfrCaps_list[0].Base.next_tables_list[0] = 
		DPE_METER_TABLE_ID;

	    /* MPLS Marker Capabilities */
	    qos_caps->MarkerCaps_Count = 0;

	       /* fill Meter Capabilities */
	    qos_caps->MeterCaps_Count = g_meter_count;
	    qos_caps->MeterCaps_list[0].Spec = 0;
	    /* Bit 3 - srTCMBlind */
	    qos_caps->MeterCaps_list[0].Spec |= 0x0001; /* Bit 1 - Simple */
	    qos_caps->MeterCaps_list[0].Base.type = npf_ds_MeterCaps;
	    qos_caps->MeterCaps_list[0].Base.tableid = DPE_METER_TABLE_ID;
	    qos_caps->MeterCaps_list[0].Base.Direction = npf_ds_IngressOnly;
	    qos_caps->MeterCaps_list[0].Base.stats_supported = 0;
	    qos_caps->MeterCaps_list[0].Base.stats_supported = 0x0003;
	    /* Byte count&
	       pkt count */
	    qos_caps->MeterCaps_list[0].Base.next_tables_count = 
		g_max_meter_next_tbl_count;
	    qos_caps->MeterCaps_list[0].Base.next_tables_list[0] = 
		DPE_ALG_DROPPER_TABLE_ID;
	    qos_caps->MeterCaps_list[0].Base.next_tables_list[1] = 
		DPE_QUEUE_TABLE_ID;
	    qos_caps->MeterCaps_list[0].failnext_tables_count =
		g_max_meter_next_tbl_count;
	    qos_caps->MeterCaps_list[0].failnext_tables_list[0] =
		DPE_ALG_DROPPER_TABLE_ID;
	    qos_caps->MeterCaps_list[0].failnext_tables_list[1] =
		DPE_QUEUE_TABLE_ID;
	}
	break;
    case DS_MPLS_LSR:
	{
	    /* Classifier = Label-based classifier
	       Marker     = MPLS marker
	    */
	    QOS_DBG_PRINT (LOG_INFO, "Collecting Caps for DS_MPLS_LSR!\n");
	    qos_caps->ClfrCaps_list[0].Spec = 0x011F; 
	    /* bit 9 - MPLS Label
	       bit 0-5 - IPv4 + DSCP classifier */
	    /****************************************************
	     * this is a hack, actually. Even though FP can report IPv4 and
	     * DSCP filters separately, code in the DiffServ module inthe
	     * control plane checks only for IPv4. So, if FP reports DSCP,
	     * this cannot be validated
	     ******************************************************/
	    qos_caps->ClfrCaps_list[0].Base.stats_supported = 0x0001; 
	    /* Pkt Count */
	    /* Classifier can be followed by Marker ONLY in MPLS+QoS */
	    qos_caps->ClfrCaps_list[0].Base.next_tables_list[0] = 
		DPE_METER_TABLE_ID;

	    /* MPLS Marker Capabilities */
	    qos_caps->MarkerCaps_Count = 0;

	    /* fill Meter Capabilities */
	    qos_caps->MeterCaps_Count = g_meter_count;
	    qos_caps->MeterCaps_list[0].Spec = 0;
	    /* Bit 3 - srTCMBlind */
	    qos_caps->MeterCaps_list[0].Spec |= 0x0001; /* Bit 1 - Simple */
	    qos_caps->MeterCaps_list[0].Base.type = npf_ds_MeterCaps;
	    qos_caps->MeterCaps_list[0].Base.tableid = DPE_METER_TABLE_ID;
	    qos_caps->MeterCaps_list[0].Base.Direction = npf_ds_IngressOnly;
	    qos_caps->MeterCaps_list[0].Base.stats_supported = 0;
	    qos_caps->MeterCaps_list[0].Base.stats_supported = 0x0003; 
	    /* Byte count & pkt count */	    
	    qos_caps->MeterCaps_list[0].Base.next_tables_count = 
		g_max_meter_next_tbl_count;
	    qos_caps->MeterCaps_list[0].Base.next_tables_list[0] = 
		DPE_ALG_DROPPER_TABLE_ID;
	    qos_caps->MeterCaps_list[0].Base.next_tables_list[1] = 
		DPE_QUEUE_TABLE_ID;
	    qos_caps->MeterCaps_list[0].failnext_tables_count =
		g_max_meter_next_tbl_count;
	    qos_caps->MeterCaps_list[0].failnext_tables_list[0] =
		DPE_ALG_DROPPER_TABLE_ID;
	    qos_caps->MeterCaps_list[0].failnext_tables_list[1] =
		DPE_QUEUE_TABLE_ID;
	}
	break;
    case DS_EDGE:
	{
	    /* IPv4 Range Matching Classifier
	       DSCP Marker
	    */
	    QOS_DBG_PRINT (LOG_INFO, "Collecting Capabilities for DS_EDGE !\n");
	    qos_caps->ClfrCaps_list[0].Spec |= 0x001F; 
	    /* IPv4 Range Matching Clfr */
	    qos_caps->ClfrCaps_list[0].Base.stats_supported = 0x0001; 
	    /* Pkt Count */
	    /* Classifier can be followed by 
	       -- Marker
	       -- Meter
	       -- Buf Mgr (Algo Dropper)
	    */
	    qos_caps->ClfrCaps_list[0].Base.next_tables_list[0] = 
		DPE_MARKER_TABLE_ID;
	    qos_caps->ClfrCaps_list[0].Base.next_tables_list[1] = 
		DPE_METER_TABLE_ID;
	    qos_caps->ClfrCaps_list[0].Base.next_tables_list[2] = 
		DPE_ALG_DROPPER_TABLE_ID;
	    qos_caps->ClfrCaps_list[0].Base.next_tables_list[3] = 
		DPE_QUEUE_TABLE_ID;

	    /* DSCP Marker Capabilities */
	    qos_caps->MarkerCaps_Count = g_marker_count;
	    qos_caps->MarkerCaps_list[0].Base.type = npf_ds_MarkerCaps;
	    qos_caps->MarkerCaps_list[0].Base.tableid = DPE_MARKER_TABLE_ID;
	    qos_caps->MarkerCaps_list[0].Base.Direction = npf_ds_IngressOnly;
	    qos_caps->MarkerCaps_list[0].Spec = 0; 
	    qos_caps->MarkerCaps_list[0].Spec |= 0x0001; /* DSCP Marker */
	    qos_caps->MarkerCaps_list[0].Base.stats_supported = 0; 
	    qos_caps->MarkerCaps_list[0].Base.stats_supported = 0x0001; 
	    /*Pkt Count*/
	    qos_caps->MarkerCaps_list[0].Base.next_tables_count = 
		g_max_marker_next_tbl_count;
	    /* Marker --> Meter
	       Marker --> Dropper */
	    qos_caps->MarkerCaps_list[0].Base.next_tables_list[0] = 
		DPE_METER_TABLE_ID;
	    qos_caps->MarkerCaps_list[0].Base.next_tables_list[1] = 
		DPE_ALG_DROPPER_TABLE_ID;
	    qos_caps->MarkerCaps_list[0].Base.next_tables_list[2] = 
		DPE_QUEUE_TABLE_ID;

	    /* fill Meter Capabilities */
	    qos_caps->MeterCaps_Count = g_meter_count;
	    qos_caps->MeterCaps_list[0].Spec = 0;
	    /* Bit 3 - srTCMBlind */
	    qos_caps->MeterCaps_list[0].Spec |= 0x0001; /* Bit 1 - Simple */
	    qos_caps->MeterCaps_list[0].Base.type = npf_ds_MeterCaps;
	    qos_caps->MeterCaps_list[0].Base.tableid = DPE_METER_TABLE_ID;
	    qos_caps->MeterCaps_list[0].Base.Direction = npf_ds_IngressOnly;
	    qos_caps->MeterCaps_list[0].Base.stats_supported = 0;
	    qos_caps->MeterCaps_list[0].Base.stats_supported = 0x0003; 
	    /* Byte count & pkt count */
	    qos_caps->MeterCaps_list[0].Base.next_tables_count = 
		g_max_meter_next_tbl_count;
	    
	    qos_caps->MeterCaps_list[0].Base.next_tables_list[0] = 
		DPE_MARKER_TABLE_ID;
	    qos_caps->MeterCaps_list[0].Base.next_tables_list[1] = 
		DPE_ALG_DROPPER_TABLE_ID;
	    qos_caps->MeterCaps_list[0].Base.next_tables_list[2] = 
		DPE_QUEUE_TABLE_ID;
	    
	    qos_caps->MeterCaps_list[0].failnext_tables_count =
		g_max_meter_next_tbl_count;
	    qos_caps->MeterCaps_list[0].failnext_tables_list[0] = 
		DPE_MARKER_TABLE_ID;
	    qos_caps->MeterCaps_list[0].failnext_tables_list[1] = 
		DPE_ALG_DROPPER_TABLE_ID;
	    qos_caps->MeterCaps_list[0].failnext_tables_list[2] = 
		DPE_QUEUE_TABLE_ID;
	}
	break;
    case DS_CORE:
	{
	    /* DSCP Classifier
	       No marker
	    */
	    QOS_DBG_PRINT (LOG_INFO, "Collecting Caps for DS_MPLS_LSR!\n");
	    qos_caps->ClfrCaps_list[0].Spec = 0x001F; /* DSCP */
	    qos_caps->ClfrCaps_list[0].Base.stats_supported = 0x0001; 
	    /* Pkt Count */
	    /* Classifier can be followed by 
	       -- Meter
	       -- Buf Mgr (Algo Dropper)
	    */
	    qos_caps->ClfrCaps_list[0].Base.next_tables_list[0] = 
		DPE_METER_TABLE_ID;
	    qos_caps->ClfrCaps_list[0].Base.next_tables_list[1] = 
		DPE_ALG_DROPPER_TABLE_ID;
	    qos_caps->ClfrCaps_list[0].Base.next_tables_list[2] = 
		DPE_QUEUE_TABLE_ID;

	    /* fill Meter Capabilities */
	    qos_caps->MeterCaps_Count = g_meter_count;
	    qos_caps->MeterCaps_list[0].Spec = 0;
	    /* Bit 3 - srTCMBlind */
	    qos_caps->MeterCaps_list[0].Spec |= 0x0001; /* Bit 1 - Simple */
	    qos_caps->MeterCaps_list[0].Base.type = npf_ds_MeterCaps;
	    qos_caps->MeterCaps_list[0].Base.tableid = DPE_METER_TABLE_ID;
	    qos_caps->MeterCaps_list[0].Base.Direction = npf_ds_IngressOnly;
	    qos_caps->MeterCaps_list[0].Base.stats_supported = 0;
	    qos_caps->MeterCaps_list[0].Base.stats_supported = 0x0003; 
	    /* Byte count & pkt count */
	    qos_caps->MeterCaps_list[0].Base.next_tables_count = 
		g_max_meter_next_tbl_count;
	    qos_caps->MeterCaps_list[0].Base.next_tables_list[0] = 
		DPE_ALG_DROPPER_TABLE_ID;
	    qos_caps->MeterCaps_list[0].Base.next_tables_list[1] = 
		DPE_QUEUE_TABLE_ID;

	    qos_caps->MeterCaps_list[0].failnext_tables_count =
		g_max_meter_next_tbl_count;
	    qos_caps->MeterCaps_list[0].failnext_tables_list[0] =
		DPE_ALG_DROPPER_TABLE_ID;
	    qos_caps->MeterCaps_list[0].failnext_tables_list[1] =
		DPE_QUEUE_TABLE_ID;
	}
	break;
    } /*end switch (router_config) */
 
     /* fill in Algo Dropper capabilities */
    qos_caps->DropperCaps_Count = g_dropper_count;
    qos_caps->DropperCaps_list[0].Type = 0;
    qos_caps->DropperCaps_list[0].Type |= 0x0018; /* Bit 4 - mQDrop */
    qos_caps->DropperCaps_list[0].Base.type = npf_ds_AlgDropCaps;
    qos_caps->DropperCaps_list[0].Base.tableid = DPE_ALG_DROPPER_TABLE_ID;
    qos_caps->DropperCaps_list[0].Base.Direction = npf_ds_IngressOnly;
    qos_caps->DropperCaps_list[0].Base.stats_supported = 0;
    qos_caps->DropperCaps_list[0].Base.stats_supported |= 0x0001; /* Pkt count
								   */ 
    qos_caps->DropperCaps_list[0].Base.next_tables_count = 
	g_max_dropper_next_tbl_count;
    qos_caps->DropperCaps_list[0].Base.next_tables_list[0] = 
	DPE_QUEUE_TABLE_ID;

    /* fill in Queue (Buffer Manager) capabilities */
    qos_caps->QueueCaps_Count = g_queue_count;
    qos_caps->QueueCaps_list[0].Base.type = npf_ds_QCaps;
    qos_caps->QueueCaps_list[0].Base.tableid = DPE_QUEUE_TABLE_ID;
    qos_caps->QueueCaps_list[0].Base.Direction = npf_ds_EgressOnly;
    qos_caps->DropperCaps_list[0].Base.stats_supported = 0; 
    qos_caps->DropperCaps_list[0].Base.stats_supported |= 0x0001; /* Pkt count
								   */
    qos_caps->QueueCaps_list[0].Base.next_tables_count = 
	g_max_queue_next_tbl_count;
    qos_caps->QueueCaps_list[0].Base.next_tables_list[0] = 
	DPE_SCHEDULER_TABLE_ID;
    
    /* fill in Shaper capabilities */
    qos_caps->ShaperCaps_Count = g_shaper_count;
    qos_caps->ShaperCaps_list[0].Base.type = npf_ds_ShaperCaps;
    qos_caps->ShaperCaps_list[0].Base.tableid = DPE_SHAPER_TABLE_ID;
    qos_caps->ShaperCaps_list[0].Base.Direction = npf_ds_EgressOnly;
    qos_caps->ShaperCaps_list[0].MaxLevels = 1; /* One level supported */
    qos_caps->ShaperCaps_list[0].Base.next_tables_count = 0;
    qos_caps->ShaperCaps_list[0].Base.next_tables_list[0] = -1;
    /* not followed by anything */

    /* fill in Scheduler capabilities */
    qos_caps->SchedulerCaps_Count = g_scheduler_count;
    qos_caps->SchedulerCaps_list[0].Base.type = npf_ds_SchedulerCaps;
    qos_caps->SchedulerCaps_list[0].Base.tableid = DPE_SCHEDULER_TABLE_ID;
    qos_caps->SchedulerCaps_list[0].Base.Direction = npf_ds_EgressOnly;
    qos_caps->SchedulerCaps_list[0].Base.stats_supported = 0;
    qos_caps->SchedulerCaps_list[0].Base.stats_supported |= 0x0003; 
    /* Pkt & byte count */
    qos_caps->SchedulerCaps_list[0].ServiceDisc = npf_ds_SchedulerDRR;
    qos_caps->SchedulerCaps_list[0].MinMaxRate = 0x0003; 
    /* Minrate only - bit 1*/
    qos_caps->SchedulerCaps_list[0].Base.next_tables_count = 0; 

    PRETURN (0);
}

/*************************************************************************
 *                                                                       *
 *                         ADD ENTRY ROUTINES                            *
 *                                                                       *
 *************************************************************************/
/**************************************************************************
*
*   Function: qos_add_classifier()
*
*   Parameters: IN: Flow Tuple
*               IN: npf_ds_BaseFilter_t 
*               OUT: filter handle
*
*   Description: Adds a DSCP or IPv4 Range Matching filter and returns
*                a filter id that should be used for deleting
*
*   PRETURN : 0 on success, -1 for failure
*************************************************************************/
int qos_add_classifier (dpe_flow_tuple_t flow_tpl, npf_ds_BaseFilter_t* clfr,
			filter_handle_t* filter_handle)
{
    flow_id_t            f_id;
    queue_id_t           q_id;
    block_id_t           next_blk_id;
    flow_tuple_t         flow_tuple;
    char                 dscp;
    range_filter_t range_filter;
    priority_t           filter_priority;
    npf_ds_IPv4Filter_t* ipv4_filter;
    unsigned long        result;
    int                  err = 0;

    QOS_DBG_TRACE ();
    QOS_DBG_PRINT (LOG_INFO, "Classifier params\n\t"
		   "Flow Id              = %d \n\t"
		   "Queue Id             = %d \n\t"
		   "Next Blk Id          = %d \n\t",
		   flow_tpl.flow_id, flow_tpl.queue_id,
		   flow_tpl.next_block_id);

    f_id = (unsigned long) flow_tpl.flow_id;
    q_id = (unsigned long) flow_tpl.queue_id;
    next_blk_id = (unsigned long) flow_tpl.next_block_id;

    if (clfr->type == npf_ds_IPv4Filter)
	ipv4_filter = (npf_ds_IPv4Filter_t*) clfr;
    else {
	QOS_DBG_PRINT (LOG_ERR, "Unknown filter type!\n");
	err = -1;
	PRETURN (err);
    }
    /* create flow tuple as reqd. by the clfr ACEs */
    flow_tuple.flow = f_id;
    flow_tuple.queue = q_id;
    if (flow_tpl.next_block_id == DPE_METER_TABLE_ID)
	flow_tuple.next_block = METER_NEXT_BLOCK_ID;
    else if (flow_tpl.next_block_id == DPE_MARKER_TABLE_ID)
	flow_tuple.next_block = MARKER_NEXT_BLOCK_ID;
    else if (flow_tpl.next_block_id == DPE_ALG_DROPPER_TABLE_ID ||
	     flow_tpl.next_block_id == DPE_QUEUE_TABLE_ID)
	flow_tuple.next_block = WRED_NEXT_BLOCK_ID;
    else {
	QOS_DBG_PRINT (LOG_ERR, "Invalid Next block Id !\n");
	err = -1;
	PRETURN (err);
    }
    switch (g_current_router_config) {
    case DS_EDGE:
	{
	    /* Edge routers have a IPv4 Range Matching classifier */

	    /* for the range filter, specify exact match by setting
	       range.start = range.end = IP addr
	    */
	    range_filter.source_addresses.start = ipv4_filter->SrcAddr;
	    range_filter.source_addresses.end = ipv4_filter->SrcAddr;
	    range_filter.dest_addresses.start = ipv4_filter->DstAddr;
	    range_filter.dest_addresses.end = ipv4_filter->DstAddr;
	    range_filter.source_ports.start = ipv4_filter->SrcL4PortMin;
	    range_filter.source_ports.end = ipv4_filter->SrcL4PortMax;
	    range_filter.dest_ports.start = ipv4_filter->DstL4PortMin;
	    range_filter.dest_ports.end = ipv4_filter->DstL4PortMax;
	    range_filter.protocol = ipv4_filter->Protocol;
	    filter_priority = (unsigned long) clfr->Precedence;

	    QOS_DBG_PRINT (LOG_INFO, "IPv4 Clfr params!\n\t"
			   "Src Addr     = 0x%x \n\t"
			   "Dst Addr     = 0x%x \n\t"
			   "Src Port Min = 0x%x \n\t"
			   "Src Port Max = 0x%x \n\t"
			   "Dst Port Min = 0x%x \n\t"
			   "Dst Port Max = 0x%x \n\t"
			   "Priority     = 0x%x \n\t"
			   "Protocol     = 0x%x \n\t",
			   ipv4_filter->SrcAddr, ipv4_filter->DstAddr,
			   ipv4_filter->SrcL4PortMin,
			   ipv4_filter->SrcL4PortMax,
			   ipv4_filter->DstL4PortMin,
			   ipv4_filter->DstL4PortMax, filter_priority,
			   ipv4_filter->Protocol);
			   
	    ECK (stub_ipv4_range_classifier_add_filter (&g_clfr_ace,
							(const range_filter_t*) 
							&range_filter,
							filter_priority,
							(const flow_tuple_t*) 
							&flow_tuple,
							filter_handle,
							&result));
	    if (result == QOS_RTN_SUCCESS)
		QOS_DBG_PRINT (LOG_INFO, "IPv4 Clfr:add_filter()"
			       "successful!\n");
	    else if (result == QOS_RTN_ERR_FIDRANGE)
		QOS_DBG_PRINT (LOG_INFO, "IPv4 Clfr:add_filter() Error:"
			       "Flow ID out of range \n");
	    else if (result == QOS_RTN_ERR_QIDRANGE)
		QOS_DBG_PRINT (LOG_INFO, "IPv4 Clfr:add_filter() Error:"
			       "Queue ID not in range \n");
	    else if (result == CLS_RTN_ERR_INVALID_ARG)
		QOS_DBG_PRINT (LOG_INFO, "IPv4 Clfr:add_filter(): Invalid"
			       "argument \n");
	    else if (result == CLS_RTN_ERR_PRIORITY_IN_USE)
		QOS_DBG_PRINT (LOG_INFO, "IPv4 Clfr:add_filter(): Priority"
			       "in use \n");
	    else if (result == QOS_RTN_ERR_OOM)
		QOS_DBG_PRINT (LOG_INFO, "IPv4 Clfr:add_filter(): Out of"
			       "memory to add filter \n");
	    
	    if (result != QOS_RTN_SUCCESS) {
		err = -1;
		PRETURN (err);
	    }
	    QOS_DBG_PRINT (LOG_INFO, "IPv4 Clfr!\n\t"
			   " Filter handle = 0x%x \n\t", *filter_handle);
	}
	break;
    case DS_CORE:
	{
	    /* Core routers have a DSCP/TOS classifier */
	    dscp = ipv4_filter->Dscp;
	    QOS_DBG_PRINT (LOG_INFO, "IPv4 Clfr params!\n\t"
			   " DSCP = 0x%x \n\t", dscp);
	    ECK (stub_dscp_classifier_add_filter (&g_clfr_ace,
						  dscp,
						  (const flow_tuple_t*) 
						  &flow_tuple,
						  filter_handle,
						  &result));
	    if (result == QOS_RTN_SUCCESS)
		QOS_DBG_PRINT (LOG_INFO, "DSCP Clfr:add_filter()"
			       "successful!\n");
	    else if (result == QOS_RTN_ERR_FIDRANGE)
		QOS_DBG_PRINT (LOG_INFO, "DSCP Clfr:add_filter() Error:"
			       "Flow ID out of range \n");
	    else if (result == QOS_RTN_ERR_QIDRANGE)
		QOS_DBG_PRINT (LOG_INFO, "DCSP Clfr:add_filter() Error:"
			       "Queue ID not in range \n");
	    else if (result == CLS_RTN_SUCCESS_REPLACED)
		QOS_DBG_PRINT (LOG_INFO, "DSCP Clfr:add_filter() Previous"
			       "DSCP/TOS replaced successfully!\n");
	    if (result != QOS_RTN_SUCCESS || 
		result != CLS_RTN_SUCCESS_REPLACED) {
		err = -1;
		PRETURN (err);
	    }
	    QOS_DBG_PRINT (LOG_INFO, "DSCP Clfr!\n\t"
			   " Filter handle = 0x%x \n\t", *filter_handle);
	}
	break;
    } /* end switch (router config) */

    PRETURN (err);
}

/**************************************************************************
*
*   Function: qos_add_marker()
*
*   Parameters: IN: Flow Tuple
*               IN: DSCP mark to apply
*
*   Description: sets up a mapping between a DSCP and Flow ID in the DSCP
*                Marker
*
*   PRETURN : 0 on success, -1 for failure 
*************************************************************************/
int qos_add_marker (dpe_flow_tuple_t flow_tuple, npf_ds_Action_t* action)
{
    flow_id_t             f_id;
    npf_ds_DscpMarkAct_t* dscp_mark;
    unsigned long         result;
    int                   err = 0;
    
    QOS_DBG_TRACE ();
    QOS_DBG_PRINT (LOG_INFO, "Marker params\n\t"
		   "Flow Id              = %d \n\t"
		   "Queue Id             = %d \n\t"
		   "Next Blk Id          = %d \n\t",
		   flow_tuple.flow_id, flow_tuple.queue_id,
		   flow_tuple.next_block_id);

    f_id = (unsigned long) flow_tuple.flow_id;
    if (*action->Specific == npf_ds_DscpMarkAct) {
      dscp_mark = (npf_ds_DscpMarkAct_t*) action->Specific;
      QOS_DBG_PRINT (LOG_INFO, "Marker type = DSCP\n");
    }
    else 
      QOS_DBG_PRINT (LOG_INFO, "Unknown Marker type!\n");

    QOS_DBG_PRINT (LOG_INFO, "DSCP mark = 0x%x \n\t", dscp_mark->Dscp);
    
    ECK (stub_dscp_marker_set_mapping (&g_marker_ace,
				       f_id,
				       dscp_mark->Dscp, &result));
    if (result == QOS_RTN_SUCCESS)
      QOS_DBG_PRINT (LOG_INFO, "DSCP Marker:add_marker() successful!\n");
    else if (result == QOS_RTN_ERR_FIDRANGE)
      QOS_DBG_PRINT (LOG_INFO, "DSCP Marker:add_marker() Error:"
		     "Flow ID out of range \n");
    if (result != QOS_RTN_SUCCESS) {
      err = -1;
    }    
    
    PRETURN (err);
}
/**************************************************************************
*
*   Function: qos_create_meter()
*
*   Parameters: Flow Tuple
*
*   Description: Installs a new meter 
*
*   PRETURN : 0 on success, -1 for failure 
*************************************************************************/

int qos_create_meter (dpe_flow_tuple_t flow_tuple)
{
    flow_id_t            f_id;
    queue_id_t           q_id;
    block_id_t           next_blk_id;
    unsigned long        result;
    int                  err = 0;

    QOS_DBG_TRACE ();
    QOS_DBG_PRINT (LOG_INFO, "SRTC Meter Parameters!\n\t"
		   "Flow Id              = %d \n\t"
		   "Queue Id             = %d \n\t"
		   "Next Blk Id          = %d \n",
		   flow_tuple.flow_id, flow_tuple.queue_id,
		   flow_tuple.next_block_id);

    f_id = (unsigned long) flow_tuple.flow_id;
    q_id = (unsigned long) flow_tuple.queue_id;
    next_blk_id = (unsigned long) flow_tuple.next_block_id;

    /* Create a meter using the Flow ID */
    ECK (stub_meter_create_meter (&g_meter_ace, f_id, &result));

    if (result == QOS_RTN_SUCCESS)
	QOS_DBG_PRINT (LOG_INFO, "Meter:create_meter() successful!\n");
    else if (result == QOS_RTN_ERR_FIDRANGE)
	QOS_DBG_PRINT (LOG_INFO, "Meter:create_meter() Error : Flow ID out"
		       " of range \n");
    else if (result == QOS_RTN_ERR_FIDINUSE)
	QOS_DBG_PRINT (LOG_INFO, "Meter:create_meter() Error : Flow ID in use"
		       "\n");

    if (result != QOS_RTN_SUCCESS) {
	err = -1;
	PRETURN (err);
    }
    PRETURN (err);
}

/**************************************************************************
*
*   Function: qos_setup_meter()
*
*   Parameters: IN: Flow Tuple
*               IN: Downstream flow tuple
*               IN: Fail next flow tuple
*
*   Description: Sets up a meter for updating incoming flow ids to map to
*                the corresponding outgoing downstream flow ids, for both
*                conforming and non-conforming traffic 
*
*   PRETURN : 0 on success, -1 for failure 
*************************************************************************/

int qos_setup_meter (dpe_flow_tuple_t flow_tuple, 
		     dpe_flow_tuple_t downstream_flow_tuple,
		     dpe_flow_tuple_t failnext_flow_tuple)
{
    flow_id_t            f_id, downstream_fid, failnext_f_id;
    queue_id_t           q_id, failnext_q_id;
    block_id_t           next_blk_id, failnext_blk_id;
    meter_action_t       meter_action;
    unsigned long        result;
    int                  err = 0;
    
    QOS_DBG_TRACE ();
    QOS_DBG_PRINT (LOG_INFO, "SRTC Meter Parameters!\n\t"
		   "Flow Id              = %d \n\t"
		   "Queue Id             = %d \n\t"
		   "Next Blk Id          = %d \n"
		   "Downstream flow id   = %d \n\t"
		   "Downstream queue id  = %d \n\t"
		   "Downstream NBlk id   = %d \n\t"
		   "Fail next Flow Id    = %d \n\t"
		   "fail next Queue Id   = %d \n\t"
		   "fail next Next Blk Id= %d \n",
		   flow_tuple.flow_id, 
		   flow_tuple.queue_id,
		   flow_tuple.next_block_id,
		   downstream_flow_tuple.flow_id, 
		   downstream_flow_tuple.queue_id,
		   downstream_flow_tuple.next_block_id, 
		   failnext_flow_tuple.flow_id, 
		   failnext_flow_tuple.queue_id, 
		   failnext_flow_tuple.next_block_id);

    /* Incoming flow id, queue id, next block id */
    f_id = (unsigned long) flow_tuple.flow_id;
    q_id = (unsigned long) flow_tuple.queue_id;
    downstream_fid = (unsigned long) downstream_flow_tuple.flow_id;

    //next_blk_id = (unsigned long) flow_tuple.next_block_id;
    /* For Strawberry Hill and Sauvie Island, Next block of Meter
       will always be the Buffer Mgr (Queue+dropper). This value
       has to be hardcoded since there is no way to query this.
    */
    next_blk_id = WRED_NEXT_BLOCK_ID;

    /* Set up the path for conforming traffic */
    meter_action.next_block = next_blk_id;  /* TO DO: NEED TO MAP THIS ID */
    meter_action.next_fid  = downstream_fid; /* map incoming flow id */
    meter_action.next_qid  = q_id; /* queue id will not change */
    meter_action.apply_qid = 0;  /* Dont change Q ID */
    meter_action.apply_mark = 0; /* dont change DSCP */
    meter_action.dscp_mark = '0'; /* some char. ignored anyway */

    /* Make the set_action call */
    /* CONFORMANCE LEVEL = GREEN */
    ECK (stub_meter_set_action (&g_meter_ace, f_id, srtcm_conf_level_t_GREEN,
				&meter_action, &result));
	 
    if (result == QOS_RTN_SUCCESS)
	QOS_DBG_PRINT (LOG_INFO, "Meter:set_action() successful!\n");
    else if (result == QOS_RTN_ERR_FIDRANGE)
	QOS_DBG_PRINT (LOG_INFO, "Meter:set_action() Error : Flow ID out"
		       " of range \n");
    else if (result == QOS_RTN_ERR_QIDRANGE)
	QOS_DBG_PRINT (LOG_INFO, "Meter:set_action() Error : Queue ID out"
		       " of range \n");
    else if (result == QOS_RTN_ERR_FIDINUSE)
	QOS_DBG_PRINT (LOG_INFO, "Meter:set_action() Error : Flow ID in use"
		       "\n");
    else if (result == MTR_RTN_ERR_CONFLVLRANGE)
	QOS_DBG_PRINT (LOG_INFO, "Meter:set_action() Error : Conf. level out"
		       " of range \n");

    if (result != QOS_RTN_SUCCESS) {
	err = -1;
	//PRETURN (err);
    }
    /* Set up the path for non conforming traffic */

    failnext_f_id = (unsigned long) failnext_flow_tuple.flow_id;
    failnext_q_id = (unsigned long) failnext_flow_tuple.queue_id;
    //failnext_blk_id = (unsigned long) failnext_flow_tuple.next_block_id;
    failnext_blk_id = WRED_NEXT_BLOCK_ID;

    meter_action.next_block = failnext_blk_id;  /* TO DO: NEED TO MAP THIS ID */
    meter_action.next_fid  = failnext_f_id;
    meter_action.next_qid  = failnext_q_id;
    meter_action.apply_qid = 1;  /* change Q ID to fail next Q id */
    meter_action.apply_mark = 0; /* dont change DSCP */
    meter_action.dscp_mark = '0'; /* some char. ignored anyway */

    /* Make the set_action call */
    /* CONFORMANCE LEVEL = GREEN */
    ECK (stub_meter_set_action (&g_meter_ace, f_id, 
				srtcm_conf_level_t_RED,
				&meter_action, &result));
	 
    if (result == QOS_RTN_SUCCESS)
	QOS_DBG_PRINT (LOG_INFO, "Meter:set_action() successful!\n");
    else if (result == QOS_RTN_ERR_FIDRANGE)
	QOS_DBG_PRINT (LOG_INFO, "Meter:set_action() Error : Flow ID out"
		       " of range \n");
    else if (result == QOS_RTN_ERR_QIDRANGE)
	QOS_DBG_PRINT (LOG_INFO, "Meter:set_action() Error : Queue ID out"
		       " of range \n");
    else if (result == QOS_RTN_ERR_FIDINUSE)
	QOS_DBG_PRINT (LOG_INFO, "Meter:set_action() Error : Flow ID in use"
		       "\n");
    else if (result == MTR_RTN_ERR_CONFLVLRANGE)
	QOS_DBG_PRINT (LOG_INFO, "Meter:set_action() Error : Conf. level out"
		       " of range \n");

    if (result != QOS_RTN_SUCCESS) {
	err = -1;
	PRETURN (err);
    }
    PRETURN (err);
}

/**************************************************************************
*
*   Function: qos_setup_meter()
*
*   Parameters: IN: Flow Tuple
*               IN: Downstream flow tuple
*               IN: Fail next flow tuple
*
*   Description: Sets up a meter for updating incoming flow ids to map to
*                the corresponding outgoing downstream flow ids, for both
*                conforming and non-conforming traffic 
*
*   PRETURN : 0 on success, -1 for failure 
*************************************************************************/
int qos_set_meter_params (dpe_flow_tuple_t flow_tuple, npf_ds_Meter_t* meter)
{
    flow_id_t            f_id;
    queue_id_t           q_id;
    block_id_t           next_blk_id;
    srtcm_rate_params_t  srtc_params;
    npf_ds_TBParam_t*    tb_params;
    unsigned long        result;
    int                  err = 0;

    QOS_DBG_TRACE ();
    QOS_DBG_PRINT (LOG_INFO, "SRTC Meter Parameters!\n\t"
		   "Flow Id              = %d \n\t"
		   "Queue Id             = %d \n\t"
		   "Next Blk Id          = %d \n",
		   flow_tuple.flow_id, flow_tuple.queue_id,
		   flow_tuple.next_block_id);

    f_id = (unsigned long) flow_tuple.flow_id;
    q_id = (unsigned long) flow_tuple.queue_id;
    next_blk_id = (unsigned long) flow_tuple.next_block_id;

    /* now set the SRTC parameters */
    if (*meter->Specific == npf_ds_TBParam)
	tb_params = (npf_ds_TBParam_t*) (meter->Specific);
    else {
	QOS_DBG_PRINT (LOG_ERR, "Unsupported Meter type - %d\n",
		       meter->Specific);
	PRETURN (-1);
    }
    QOS_DBG_PRINT (LOG_INFO, "SRTC parameters\n\t Rate: %d \n\t"
		   "Burst = %d \n\t Interval = %d \n\t",
		   tb_params->Rate, tb_params->BurstSize, 
		   tb_params->Interval); 

    /* convert from kbits/s to bytes/s */
    srtc_params.cir = (unsigned int) (tb_params->Rate*125); 
    /* committed information rate (bps) */
    srtc_params.cbs = (unsigned int) tb_params->BurstSize; /* NEED TO CONVERT 
							     FROM PACKETS TO
							     BYTES */
    srtc_params.ebs = 0;
    /* committed information rate (bps) */
    ECK (stub_srtc_meter_set_rate (&g_meter_ace, f_id, &srtc_params, &result));
    if (result == QOS_RTN_SUCCESS)
	QOS_DBG_PRINT (LOG_INFO, "SRTC Meter:set_rate() successful!\n");
    else if (result == QOS_RTN_ERR_FIDRANGE)
	QOS_DBG_PRINT (LOG_INFO, "SRTC Meter:set_rate() Error : Flow ID out"
		       " of range \n");
    else if (result == QOS_RTN_ERR_FIDINUSE)
	QOS_DBG_PRINT (LOG_INFO, "SRTC Meter:set_rate() Error : Flow ID not in"
		       "use \n");

    if (result != QOS_RTN_SUCCESS) {
	err = -1;
	PRETURN (err);
    }

    PRETURN (err);
}

/**************************************************************************
 *
 *   Function: qos_set_dropper_params()
 *
 *   Parameters: Interface (port/queue_group id)
 *               Flow id
 *
 *   Description: Sets up the EWMA constant, etc. for the corr. Queue Id.
 *
 *   PRETURN : 0 on success, -1 for failure 
 *************************************************************************/

int qos_set_dropper_params (uint32_t intf, dpe_flow_tuple_t flow_tuple,
			    npf_ds_AlgDrop_t* dropper)
{
    flow_id_t            f_id;
    queue_id_t           queue_id;
    queue_grp_t          queue_grp_id;
    unsigned long        ewma;
    unsigned short       max_q_depth; /* Max Queue Depth */
    unsigned short       min_q_depth; /* Min Queue Depth */
    float                max_drop_prob; /* Maximum Drop probablity [0,1] */
    wred_params_t        wred_params;
    npf_ds_RandomDrop_t* random_dropper;
    unsigned long        result;
    int                  err = 0;

    QOS_DBG_TRACE ();
    QOS_DBG_PRINT (LOG_INFO, "Queue Parameters!\n\t"
		   "Interface Id = %d \n\t"
		   " Queue Id    = %d \n\t", intf, flow_tuple.queue_id);

    f_id = (unsigned long ) flow_tuple.flow_id;
    queue_id = (unsigned long ) flow_tuple.queue_id;
    queue_grp_id = (unsigned long) intf;

    /* Now set the Queue parameters */
    if (dropper->dropper.sqalgdrop.Type == Type_randomDrop) {
	random_dropper = (npf_ds_RandomDrop_t*) 
	    dropper->dropper.sqalgdrop.Specific;
    	ewma = (unsigned long) random_dropper->Weight;
    }

    if (dropper->dropper.sqalgdrop.Type == Type_alwaysDrop) { 
	/* NULL if absolute dropper */
	max_q_depth = 0;
	min_q_depth = 0;
	max_drop_prob = (float) WRED_BUFMGR_MAX_DROP_PROB;
	
	wred_params.max_t = max_q_depth;
	wred_params.min_t = min_q_depth;
	wred_params.max_p = max_drop_prob;
	
	ECK (stub_wred_bufmgr_add_red_instance (&g_bufmgr_ace, f_id,
						&wred_params, &result));
    }
    else if (random_dropper != NULL) { 
	/* currently, only probability field of npf_ds_RandomDrop_t is 
	   of use */
	max_q_depth = (unsigned short) WRED_BUFMGR_MAX_Q_DEPTH;
	min_q_depth = (unsigned short) WRED_BUFMGR_MIN_Q_DEPTH;
	max_drop_prob = (float) random_dropper->ProbMax;
	
	wred_params.max_t = max_q_depth;
	wred_params.min_t = min_q_depth;
	wred_params.max_p = max_drop_prob;
	
	ECK (stub_wred_bufmgr_add_red_instance (&g_bufmgr_ace, f_id,
						&wred_params, &result));
    }
    
    QOS_DBG_PRINT (LOG_INFO, "Setting EWMA for Q id = %d\n\t"
		   "Queue grp id = %d \n\t"
		   "Ewma = %d \n\t", queue_id, queue_grp_id, ewma);
    
    ECK (stub_wred_bufmgr_set_ewma_const (&g_bufmgr_ace, queue_grp_id,
					  queue_id, ewma, &result));

    if (result == QOS_RTN_SUCCESS)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:set_ewma() successful!\n");
    else if (result == QOS_RTN_ERR_QIDRANGE)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:set_ewma() Error : Queue ID out"
		       " of range \n");
    else if (result == QOS_RTN_ERR_QGRPRANGE)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:set_ewma() Error : Group ID out"
		       " of range \n");
    else if (result == QOS_RTN_ERR_QIDINUSE)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr: set_ewma() Error: Queue Id"
		       " currently in use\n");
    else if (result == QOS_RTN_ERR_QIDUNUSE)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr: set_ewma() Error: Queue Id not"
		       " currently in use\n");
    else if (result == WRED_RTN_ERR_EWMARANGE)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:set_ewma() Error : EWMA is not "
		       "in valid range \n");
    else 
	QOS_DBG_PRINT (LOG_ERR, "WRED Buf Mgr:set_ewma() Unknown error %ld\n",
			result);
    if (result != QOS_RTN_SUCCESS)
	err = -1;
    
    PRETURN (err);
}

/**************************************************************************
 *
 *   Function: qos_add_default_queue_entry()
 *
 *   Parameters: Interface (port/queue_group id)
 *               Flow id
 *
 *   Description: Sets up the EWMA constant for the corr. Queue Id.
 *
 *   PRETURN : 0 on success, -1 for failure 
 *************************************************************************/

int qos_add_default_queue_entry (uint32_t intf, uint32_t q_id)
{
    queue_id_t     queue_id;
    queue_grp_t    queue_grp_id;
    unsigned long  ewma;
    unsigned long  result;
    int            err = 0;

    QOS_DBG_TRACE ();
    QOS_DBG_PRINT (LOG_INFO, "Queue Parameters!\n\t"
		   "Interface Id = %d \n\t"
		   " Queue Id    = %d \n\t", intf, q_id);

    queue_id = (unsigned long ) q_id;
    queue_grp_id = (unsigned long) intf;

    /* Add a Queue - Prepare the Buffer Manager ACE for the new
       new Queue ID
    */ 
    ECK (stub_bufmgr_add_queue (&g_bufmgr_ace, queue_grp_id,
				queue_id, &result));

    if (result == QOS_RTN_SUCCESS)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:add_queue() successful!\n");
    else if (result == QOS_RTN_ERR_QIDRANGE)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:add_queue() Error : Queue ID out"
		       " of range \n");
    else if (result == QOS_RTN_ERR_QIDUNUSE)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:add_queue() Error : Queue ID not"
		       " in use \n");

    /* Now set the Queue parameters */
    ewma = (unsigned long) WRED_BUFMGR_EWMA;

    ECK (stub_wred_bufmgr_set_ewma_const (&g_bufmgr_ace, queue_grp_id,
					  queue_id, ewma, &result));

    if (result == QOS_RTN_SUCCESS)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:set_ewma() successful!\n");
    else if (result == QOS_RTN_ERR_QIDRANGE)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:set_ewma() Error : Queue ID out"
		       " of range \n");
    else if (result == QOS_RTN_ERR_QGRPRANGE)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:set_ewma() Error : Group ID out"
		       " of range \n");
    else if (result == QOS_RTN_ERR_QIDINUSE)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr: set_ewma() Error: Queue Id"
		       " currently in use\n");
    else if (result == QOS_RTN_ERR_QIDUNUSE)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr: set_ewma() Error: Queue Id not"
		       " currently in use\n");
    else if (result == WRED_RTN_ERR_EWMARANGE)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:set_ewma() Error : EWMA is not "
		       "in valid range \n");
    else 
	QOS_DBG_PRINT (LOG_ERR, "WRED Buf Mgr:set_ewma() Unknown error %ld\n",
			result);
    if (result != QOS_RTN_SUCCESS)
	err = -1;
    
    PRETURN (err);
}

/**************************************************************************
*
*   Function: qos_add_default_red_instance()
*
*   Parameters: Flow id
*
*   Description: Installs a default RED parameter for the corr. flow id
*
*   PRETURN : 0 on success, -1 for failure 
*************************************************************************/

int qos_add_default_red_instance (uint32_t flow_id)
{
    flow_id_t      f_id;
    unsigned short max_q_depth; /* Max Queue Depth - obtained from rmconfig.h */
    unsigned short min_q_depth; /* Min Queue Depth - obtained from rmconfig.h */
    float          max_drop_prob; /* Maximum Drop probablity [0,1] */
    wred_params_t  wred_params;
    unsigned long  result;
    int            err = 0;

    QOS_DBG_TRACE ();
    QOS_DBG_PRINT (LOG_INFO, " WRED Parameters!\n\t"
		   " Flow Id    = %d \n\t",flow_id);

    f_id = (unsigned long ) flow_id;
    max_q_depth = (unsigned short) WRED_BUFMGR_MAX_Q_DEPTH;
    min_q_depth = (unsigned short) WRED_BUFMGR_MIN_Q_DEPTH;
    max_drop_prob = (float) WRED_BUFMGR_MAX_DROP_PROB;

    wred_params.max_t = max_q_depth;
    wred_params.min_t = min_q_depth;
    wred_params.max_p = max_drop_prob;

    ECK (stub_wred_bufmgr_add_red_instance (&g_bufmgr_ace, f_id,
					    &wred_params, &result));

    if (result == QOS_RTN_SUCCESS)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:add_red() successful!\n");
    else if (result == QOS_RTN_ERR_FIDRANGE)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:set_red() Error : Flow ID out"
		       " of range \n");
    else if (result == WRED_RTN_ERR_MAXTRANGE)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:set_red() Error : Max Threshold"
		       " ID out of range \n");
    else if (result == WRED_RTN_ERR_MINTRANGE)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr: set_red() Error: Min Threshold"
		       "out of range\n");
    else if (result == WRED_RTN_ERR_MAXPRANGE)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:set_red() Error : Max"
		       "Probability out of range \n");
    else 
	QOS_DBG_PRINT (LOG_ERR, "WRED Buf Mgr:set_red() Unknown error %d\n",
			result);
    if (result != QOS_RTN_SUCCESS)
	err = -1;
    
    PRETURN (err);
}

/**************************************************************************
*
*   Function: qos_add_scheduler_entry()
*
*   Parameters: Interface (port/queue_group id)
*               Flow tuple
*               npf_ds_MinRate_t*, npf_ds_MaxRate_t*
*
*   Description: Sets up a scheduler for the specified queue
*
*   PRETURN : 0 on success, -1 for failure 
*************************************************************************/

int qos_add_scheduler_entry (uint32_t intf, dpe_flow_tuple_t flow_tuple,
			     npf_ds_MinRate_t* min_rate, 
			     npf_ds_MaxRate_t* max_rate)
{
    quantum_t      quantum;
    credit_t       credit_limit;
    flow_id_t      flow_id;
    queue_id_t     queue_id;
    queue_grp_t    queue_grp_id;
    unsigned long  result;
    int            err = 0;

    QOS_DBG_TRACE ();
    QOS_DBG_PRINT (LOG_INFO, "Scheduler Parameters!\n\t"
		   "Interface Id = %d \n\t"
		   " Flow Id     = %d \n\t"
		   " Queue Id    = %d \n\t", intf, flow_tuple.flow_id,
		   flow_tuple.queue_id);

    queue_grp_id = (unsigned long) intf;
    flow_id  = (unsigned long) flow_tuple.flow_id;
    queue_id = (unsigned long) flow_tuple.queue_id;

    /* by default, quantum will be DRR_SCHEDULER_QUANTUM_DEF (100 bytes) */
    generate_quantum (min_rate, max_rate, &quantum);
    credit_limit = DRR_SCHEDULER_CREDIT_DEF; /* 1518 bytes */
    QOS_DBG_PRINT (LOG_INFO, " Parameters before set_drr() \n Quantum = %lu\n"
		   " Credit limit = %lu \n\t",
		   quantum, credit_limit);

    ECK (stub_drr_scheduler_set_drr (&g_sched_ace, queue_grp_id, queue_id,
				     quantum, credit_limit, &result));
    if (result == QOS_RTN_SUCCESS)
	QOS_DBG_PRINT (LOG_INFO, "Scheduler:set_drr() successful!\n");
    else if (result == QOS_RTN_ERR_QIDRANGE)
	QOS_DBG_PRINT (LOG_INFO, "Scheduler:set_drr() Error : Queue ID out"
		       " of range \n");
    else if (result == QOS_RTN_ERR_QGRPRANGE)
	QOS_DBG_PRINT (LOG_INFO, "Scheduler:set_drr() Error : Group ID out"
		       "of range \n");
    else if (result == DRR_RTN_ERR_QUANTUMRANGE)
	QOS_DBG_PRINT (LOG_INFO, "Scheduler:set_drr() Error : Quantum out"
		       "of range \n");
    else if (result == DRR_RTN_ERR_CREDITRANGE)
	QOS_DBG_PRINT (LOG_INFO, "Scheduler:set_drr() Error : Credit out"
		       "of range \n");

    if (result != QOS_RTN_SUCCESS)
	err = -1;

    PRETURN (err);
}

/**************************************************************************
*
*   Function: qos_add_default_scheduler_entry()
*
*   Parameters: Interface (port/queue_group id)
*               Flow tuple
*
*   Description: Sets up a default scheduler for the specified queue
*
*   PRETURN : 0 on success, -1 for failure 
*************************************************************************/

int qos_add_default_scheduler_entry (uint32_t intf, dpe_flow_tuple_t flow_tuple)
{
    quantum_t      quantum;
    credit_t       credit_limit;
    flow_id_t      flow_id;
    queue_id_t     queue_id;
    queue_grp_t    queue_grp_id;
    unsigned long  result;
    int            err = 0;

    QOS_DBG_TRACE ();
    QOS_DBG_PRINT (LOG_INFO, "Scheduler Parameters!\n\t"
		   "Interface Id = %d \n\t"
		   " Flow Id     = %d \n\t"
		   " Queue Id    = %d \n\t", intf, flow_tuple.flow_id,
		   flow_tuple.queue_id);

    queue_grp_id = (unsigned long) intf;
    flow_id  = (unsigned long) flow_tuple.flow_id;
    queue_id = (unsigned long) flow_tuple.queue_id;

    /* by default, quantum will be DRR_SCHEDULER_QUANTUM_DEF (100 bytes) */
    quantum = DRR_SCHEDULER_QUANTUM_DEF;
    credit_limit = DRR_SCHEDULER_CREDIT_DEF; /* 1518 bytes */
    QOS_DBG_PRINT (LOG_INFO, "Scheduler Params!\n\t"
		   "Quantum = %ld \n\t Credit Limit = %ld\n\t",
		    quantum, credit_limit);

    ECK (stub_drr_scheduler_set_drr (&g_sched_ace, queue_grp_id, queue_id,
				     quantum, credit_limit, &result));
    if (result == QOS_RTN_SUCCESS)
	QOS_DBG_PRINT (LOG_INFO, "Scheduler:set_drr() successful!\n");
    else if (result == QOS_RTN_ERR_QIDRANGE)
	QOS_DBG_PRINT (LOG_INFO, "Scheduler:set_drr() Error : Queue ID out"
		       " of range \n");
    else if (result == QOS_RTN_ERR_QGRPRANGE)
	QOS_DBG_PRINT (LOG_INFO, "Scheduler:set_drr() Error : Group ID out"
		       "of range \n");
    else if (result == DRR_RTN_ERR_QUANTUMRANGE)
	QOS_DBG_PRINT (LOG_INFO, "Scheduler:set_drr() Error : Quantum out"
		       "of range \n");
    else if (result == DRR_RTN_ERR_CREDITRANGE)
	QOS_DBG_PRINT (LOG_INFO, "Scheduler:set_drr() Error : Credit out"
		       "of range \n");

    if (result != QOS_RTN_SUCCESS)
	err = -1;

    PRETURN (err);
}

/*************************************************************************
 *                                                                       *
 *                         MODIFY ENTRY ROUTINES                         *
 *                                                                       *
 *************************************************************************/

/**************************************************************************
*
*   Function: qos_modify_scheduler_entry()
*
*   Parameters: Interface (port/queue_group id)
*               Flow tuple
*               npf_ds_Scheduler
*
*   Description: Modifies a scheduler for the specified queue
*
*   PRETURN : None 
*************************************************************************/

int qos_modify_scheduler_entry (uint32_t intf, dpe_flow_tuple_t flow_tuple,
				npf_ds_MinRate_t* min_rate,
				npf_ds_MaxRate_t* max_rate)
{
    quantum_t      quantum;
    credit_t       credit_limit;
    flow_id_t      flow_id;
    queue_id_t     queue_id;
    queue_grp_t    queue_grp_id;
    unsigned long  result;
    int            err = 0;

    QOS_DBG_TRACE ();
    QOS_DBG_PRINT (LOG_INFO, "Scheduler Parameters!\n\t"
		   "Interface Id = %d \n\t"
		   " Flow Id     = %d \n\t"
		   " Queue Id    = %d \n\t", intf, flow_tuple.flow_id,
		   flow_tuple.queue_id);

    queue_grp_id = (unsigned long) intf;
    flow_id  = (unsigned long) flow_tuple.flow_id;
    queue_id = (unsigned long) flow_tuple.queue_id;

    /* by default, quantum will be DRR_SCHEDULER_QUANTUM_DEF (100 bytes) */
    generate_quantum (min_rate, max_rate, &quantum);
    credit_limit = DRR_SCHEDULER_CREDIT_DEF; /* 1518 bytes */
    
    ECK (stub_drr_scheduler_set_drr (&g_sched_ace, queue_grp_id, queue_id,
				     quantum, credit_limit, &result));
    if (result == QOS_RTN_SUCCESS)
	QOS_DBG_PRINT (LOG_INFO, "Scheduler:set_drr() successful!\n");
    else if (result == QOS_RTN_ERR_QIDRANGE)
	QOS_DBG_PRINT (LOG_INFO, "Scheduler:set_drr() Error : Queue ID out"
		       " of range \n");
    else if (result == QOS_RTN_ERR_QGRPRANGE)
	QOS_DBG_PRINT (LOG_INFO, "Scheduler:set_drr() Error : Group ID out"
		       "of range \n");
    else if (result == DRR_RTN_ERR_QUANTUMRANGE)
	QOS_DBG_PRINT (LOG_INFO, "Scheduler:set_drr() Error : Quantum out"
		       "of range \n");
    else if (result == DRR_RTN_ERR_CREDITRANGE)
	QOS_DBG_PRINT (LOG_INFO, "Scheduler:set_drr() Error : Credit out"
		       "of range \n");

    if (result != QOS_RTN_SUCCESS)
	err = -1;

    PRETURN (err);
}

/**************************************************************************
*
*   Function: qos_modify_meter()
*
*   Parameters: Flow Tuple
*               npf_ds_TBParam_t - token bucket params
*
*   Description: Modifies a meter 
*
*   PRETURN : None 
*************************************************************************/

int qos_modify_meter (dpe_flow_tuple_t flow_tuple, npf_ds_Meter_t* meter)
{
    flow_id_t            f_id;
    srtcm_rate_params_t  srtc_params;
    npf_ds_TBParam_t*    tb_params;
    unsigned long        result;
    int                  err = 0;

    QOS_DBG_TRACE ();
    QOS_DBG_PRINT (LOG_INFO, "SRTC Meter Parameters!\n\t"
		   "Flow Id              = %d \n\t",
		   flow_tuple.flow_id);

    f_id = (unsigned long) flow_tuple.flow_id;


    /* now set the SRTC parameters */
    if (*meter->Specific == npf_ds_TBParam)
	tb_params = (npf_ds_TBParam_t*) (meter->Specific);
    else {
	QOS_DBG_PRINT (LOG_ERR, "Unsupported Meter type - %d\n",
		       meter->Specific);
	PRETURN (-1);
    }
    QOS_DBG_PRINT (LOG_INFO, "SRTC parameters\n\t Rate: %d \n\t"
		   "Burst = %d \n\t Interval = %d \n\t",
		   tb_params->Rate, tb_params->BurstSize, 
		   tb_params->Interval); 

    /* convert KBits/s to Bytes/s */
    srtc_params.cir = (unsigned int) ((tb_params->Rate*1024)/8); 
    /* committed information rate (bps) */
    srtc_params.cbs = (unsigned int) tb_params->BurstSize; /* NEED TO CONVERT 
							     FROM PACKETS TO
							     BYTES */
    srtc_params.ebs = 0;
    /* committed information rate (bps) */
    ECK (stub_srtc_meter_set_rate (&g_meter_ace, f_id, &srtc_params, &result));
    if (result == QOS_RTN_SUCCESS)
	QOS_DBG_PRINT (LOG_INFO, "SRTC Meter:set_rate() successful!\n");
    else if (result == QOS_RTN_ERR_FIDRANGE)
	QOS_DBG_PRINT (LOG_INFO, "SRTC Meter:set_rate() Error : Flow ID out"
		       " of range \n");
    else if (result == QOS_RTN_ERR_FIDINUSE)
	QOS_DBG_PRINT (LOG_INFO, "SRTC Meter:set_rate() Error : Flow ID not in"
		       "use \n");

    if (result != QOS_RTN_SUCCESS) {
	err = -1;
	PRETURN (err);
    }

    PRETURN (err);
}

/*************************************************************************
 *                                                                       *
 *                         DELETE ENTRY ROUTINES                         *
 *                                                                       *
 *************************************************************************/
/**************************************************************************
*
*   Function: qos_delete_classifier()
*
*   Parameters: IN: Filter ID
*
*   Description: Deletes a classifier installed earlier
*
*   PRETURN : 0 on success, -1 for failure 
*************************************************************************/
int qos_delete_classifier (filter_handle_t filter_handle)
{
    unsigned long        result;
    int                  err = 0;

    QOS_DBG_TRACE ();

    QOS_DBG_PRINT (LOG_INFO, "Delete Classifier params \n\t"
		   "Filter id = 0x%x \n", filter_handle);

    ECK (stub_classifier_delete_filter (&g_clfr_ace, filter_handle,
					&result));
    if (result == QOS_RTN_SUCCESS)
      QOS_DBG_PRINT (LOG_INFO, "Clfr:del_filter() successful!\n");
    else if (result == CLS_RTN_ERR_INVALID_FILTERID)
      QOS_DBG_PRINT (LOG_INFO, "Clfr:del_filter(): Invalid filter id \n");
    else if (result == QOS_RTN_ERR_OOM)
      QOS_DBG_PRINT (LOG_INFO, "Clfr:del_filter(): Out of memory!\n");
    
    if (result != QOS_RTN_SUCCESS)
      err = -1;
    PRETURN (err);
}


/**************************************************************************
*
*   Function: qos_delete_marker()
*
*   Parameters: IN: Flow ID
*
*   Description: Deletes a marker mapping installed earlier
*
*   PRETURN : 0 on success, -1 for failure 
*************************************************************************/
int qos_delete_marker (dpe_flow_tuple_t flow_tuple)
{
    flow_id_t     f_id;
    unsigned long result;
    int           err = 0;

    QOS_DBG_TRACE ();

    QOS_DBG_PRINT (LOG_INFO, "Delete marker params \n\t"
		   "Flow id = %d \n", flow_tuple.flow_id);

    f_id = (unsigned long) flow_tuple.flow_id;

    ECK (stub_marker_del_mapping (&g_marker_ace, f_id, &result));
    if (result == QOS_RTN_SUCCESS)
      QOS_DBG_PRINT (LOG_INFO, "Marker:del_mapping() successful!\n");
    else if (result == QOS_RTN_ERR_FIDRANGE)
      QOS_DBG_PRINT (LOG_INFO, "Marker:del_mapping(): Flow id out of"
		     "range!\n");
    else if (result == QOS_RTN_ERR_FIDUNUSE)
      QOS_DBG_PRINT (LOG_INFO, "Marker:del_mapping(): Flow id unused!\n");
    
    if (result != QOS_RTN_SUCCESS)
      err = -1;
    PRETURN (err);
}

/**************************************************************************
*
*   Function: qos_delete_meter()
*
*   Parameters: Flow Tuple
*               npf_ds_TBParam_t - token bucket params
*
*   Description: Deletes a meter 
*
*   PRETURN : 0 on success, -1 for failure
*************************************************************************/

int qos_delete_meter (dpe_flow_tuple_t flow_tuple, 
		      dpe_flow_tuple_t failnext_flow_tuple)
{
    flow_id_t     f_id, failnext_f_id;
    unsigned long result;
    int           err = 0;

    QOS_DBG_TRACE ();
    QOS_DBG_PRINT (LOG_INFO, "Delete Meter Parameters!\n\t"
		   " Flow Id     = %d \n\t"
		   " Queue Id    = %d \n\t", flow_tuple.flow_id,
		   flow_tuple.queue_id);

    f_id = (unsigned long) flow_tuple.flow_id;
    failnext_f_id = (unsigned long) failnext_flow_tuple.flow_id;

    /* Delete the meter for Flow id = conformant traffic */
    ECK (stub_meter_del_meter (&g_meter_ace, f_id, &result));
	 
    if (result == QOS_RTN_SUCCESS)
	QOS_DBG_PRINT (LOG_INFO, "Meter:del_meter() successful!\n");
    else if (result == QOS_RTN_ERR_FIDRANGE)
	QOS_DBG_PRINT (LOG_INFO, "Meter:del_meter() Error : Flow ID out"
		       " of range \n");
    else if (result == QOS_RTN_ERR_FIDUNUSE)
	QOS_DBG_PRINT (LOG_INFO, "Meter:del_meter() Error : Flow ID not in use"
		       "\n");

    if (result != QOS_RTN_SUCCESS)
	err = -1;

    PRETURN (err);
}

/**************************************************************************
*
*   Function: qos_delete_dropper()
*
*   Parameters: Flow Tuple
*
*   Description: Deletes a dropper 
*
*   PRETURN : 0 on success, -1 for failure
*************************************************************************/

int qos_delete_dropper (dpe_flow_tuple_t flow_tuple) 
{
    flow_id_t     f_id;
    unsigned long result;
    int           err = 0;

    QOS_DBG_TRACE ();
    QOS_DBG_PRINT (LOG_INFO, "Delete Dropper Parameters!\n\t"
		   " Flow Id     = %d \n\t"
		   " Queue Id    = %d \n\t", flow_tuple.flow_id,
		   flow_tuple.queue_id);

    f_id = (unsigned long) flow_tuple.flow_id;

    /* Delete the dropper for the Flow id */
    ECK (stub_wred_bufmgr_del_red_instance (&g_bufmgr_ace, f_id, &result));
	 
    if (result == QOS_RTN_SUCCESS)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr: del_red() successful!\n");
    else if (result == QOS_RTN_ERR_FIDRANGE)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr: del_red() Error : Flow ID out"
		       " of range \n");
    else if (result == QOS_RTN_ERR_FIDUNUSE)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr: del_red() Error : Flow ID not"
		       "in use \n");

    if (result != QOS_RTN_SUCCESS)
	err = -1;

    PRETURN (err);
}

/**************************************************************************
*
*   Function: qos_delete_queue()
*
*   Parameters: IN: Flow Tuple
*               IN: Queue Group Id (Interface)
*   Description: Deletes a queue
*
*   PRETURN : 0 on success, -1 for failure
*************************************************************************/

int qos_delete_queue (dpe_flow_tuple_t flow_tuple, uint32_t intf) 
{
    queue_id_t    q_id;
    queue_grp_t   q_grp_id;
    unsigned long result;
    int           err = 0;

    QOS_DBG_TRACE ();

    q_id = (unsigned long) flow_tuple.queue_id;
    q_grp_id = (unsigned long) intf;

    QOS_DBG_PRINT (LOG_INFO, "Delete Queue Parameters!\n\t"
		   " Flow Id     = %d \n\t"
		   " Queue Id    = %d \n\t"
		   " Queue Grp Id= %d \n\t", flow_tuple.flow_id,
		   flow_tuple.queue_id, q_grp_id);

    /* Delete the Queue for the Queue id for the interface */
    ECK (stub_bufmgr_del_queue (&g_bufmgr_ace, q_grp_id, q_id, &result));
	 
    if (result == QOS_RTN_SUCCESS)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr: del_q() successful!\n");
    else if (result == QOS_RTN_ERR_QIDRANGE)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr: del_q() Error : Q ID out"
		       " of range \n");
/*      else if (result == QOS_RTN_ERR_QIDUNUSED) */
/*  	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr: del_q() Error : Q ID not" */
/*  		       "in use \n"); */

    if (result != QOS_RTN_SUCCESS)
	err = -1;

    PRETURN (err);
}

/**************************************************************************
*
*   Function: qos_delete_scheduler()
*
*   Parameters: IN: Flow Tuple
*               IN: Queue Group Id (Interface)
*   Description: Deletes a Queue in Scheduler
*
*   PRETURN : 0 on success, -1 for failure
*************************************************************************/

int qos_delete_scheduler (dpe_flow_tuple_t flow_tuple, uint32_t intf) 
{
    queue_id_t    q_id;
    queue_grp_t   q_grp_id;
    unsigned long result;
    int           err = 0;

    QOS_DBG_TRACE ();

    q_id = (unsigned long) flow_tuple.queue_id;
    q_grp_id = (unsigned long) intf;

    QOS_DBG_PRINT (LOG_INFO, "Delete Scheduler Parameters!\n\t"
		   " Flow Id     = %d \n\t"
		   " Queue Id    = %d \n\t"
		   " Queue Grp Id= %d \n\t", flow_tuple.flow_id,
		   flow_tuple.queue_id, q_grp_id);

    /* Delete the Queue for the Queue id for the interface */
    ECK (stub_scheduler_delete_queue (&g_sched_ace, q_grp_id, q_id, &result));
	 
    if (result == QOS_RTN_SUCCESS)
	QOS_DBG_PRINT (LOG_INFO, "Scheduler: del_q() successful!\n");
    else if (result == QOS_RTN_ERR_QIDRANGE)
	QOS_DBG_PRINT (LOG_INFO, "Scheduler: del_q() Error : Q ID out"
		       " of range \n");
    /*  else if (result == QOS_RTN_ERR_QIDUNUSED) */
/*  	QOS_DBG_PRINT (LOG_INFO, "Scheduler: del_q() Error : Q ID not" */
/*  		       "in use \n"); */

    if (result != QOS_RTN_SUCCESS)
	err = -1;

    PRETURN (err);
}

/*************************************************************************
 *                                                                       *
 *                         GET STATISTICS ROUTINES                       *
 *                                                                       *
 *************************************************************************/
/**************************************************************************
*
*   Function: qos_get_clfr_stats
*
*   Parameters: IN: Flow Id
*               OUT:Packet count, Byte count
* 
*   Description: Gets statistics counters from the specified classifier
*
*   PRETURN : 0 on success, -1 for failure 
*************************************************************************/

int qos_get_clfr_stats (filter_handle_t filter_hdl,
			unsigned long long* defaulted_pkt_count,
			unsigned long long* pkt_count,
			unsigned long long* dropped_pkt_count,
			unsigned long long* defaulted_byte_count,
			unsigned long long* byte_count,
			unsigned long long* dropped_byte_count)
{
    filter_id_t    filter;
    counter_t      num_default, packet_count, num_dropped;
    long           result;
    int            err = 0;
    
    QOS_DBG_TRACE ();
    QOS_DBG_PRINT (LOG_INFO, "Get Clfr Statistics Parameters!\n\t"
		   "Filter handle = %d \n", filter_hdl);

    filter = (unsigned long) filter_hdl;

    ECK (stub_classifier_get_dropped_count (&g_clfr_ace, &num_dropped,
					    &result));
    ECK (stub_classifier_get_defaulted_count (&g_clfr_ace, &num_default, 
					      &result));
    ECK (stub_classifier_get_packet_count (&g_clfr_ace, filter, &packet_count,
					   &result));
    
    if (result == QOS_RTN_SUCCESS)
	QOS_DBG_PRINT (LOG_INFO, "Clfr:get_stats() successful!\n");
    else if (result == CLS_RTN_ERR_INVALID_FILTERID)
	QOS_DBG_PRINT (LOG_INFO, "Clfr:get_stats() Error : Filter ID out"
		       " invalid \n");
    else if (result == QOS_RTN_ERR_NOSTATSAVAIL)
	QOS_DBG_PRINT (LOG_INFO, "Meter:get_stats() Error : No Stats"
		       "available \n");
/*      else if (result == QOS_RTN_ERR_SUCCESS_32BIT_STATS) */
/*  	QOS_DBG_PRINT (LOG_INFO, "Clfr:get_stats() : Only 32-bit" */
/*  		       "Stats available \n"); */

    if (result != QOS_RTN_SUCCESS) {
	err = -1;
	PRETURN (err);
    }
    
    *defaulted_pkt_count = num_default;
    *pkt_count = packet_count;
    *dropped_pkt_count = num_dropped;
    *defaulted_byte_count = 0;
    *byte_count = 0;
    *dropped_byte_count = 0;

    PRETURN (err);
}

/**************************************************************************
*
*   Function: qos_get_marker_stats
*
*   Parameters: OUT: number marked
*               OUT: number dropped
* 
*   Description: Gets statistics counters from the specified marker
*
*   PRETURN : 0 on success, -1 for failure 
*************************************************************************/

int qos_get_marker_stats (unsigned long long* pkts_marked,
			  unsigned long long* pkts_dropped,
			  unsigned long long* bytes_marked,
			  unsigned long long* bytes_dropped)
{
    counter_t      num_marked = 0, num_dropped = 0;
    long           result;
    int            err = 0;
    
    QOS_DBG_TRACE ();

    ECK (stub_marker_get_stats (&g_marker_ace, &num_marked,
				&num_dropped, &result));

    if (result == QOS_RTN_SUCCESS)
	QOS_DBG_PRINT (LOG_INFO, "Clfr:get_stats() successful!\n");
/*      else if (result == QOS_RTN_ERR_SUCCESS_32BIT_STATS) */
/*  	QOS_DBG_PRINT (LOG_INFO, "Clfr:get_stats() : Only 32-bit" */
/*  		       "Stats available \n"); */

    if (result != QOS_RTN_SUCCESS) {
	err = -1;
	*pkts_marked = 0;
	*pkts_dropped = 0;
    }
    
    *pkts_marked = num_marked;
    *pkts_dropped = num_dropped;
    *bytes_marked = 0;
    *bytes_dropped = 0;

    PRETURN (err);
}


/**************************************************************************
*
*   Function: qos_get_meter_stats
*
*   Parameters: IN: Flow Id
*               OUT:Packet count, Byte count
* 
*   Description: Gets statistics counters from the specified meter
*
*   PRETURN : 0 on success, -1 for failure 
*************************************************************************/

int qos_get_meter_stats (uint32_t flow_id,
			 unsigned long long* conf_pkt_count,
			 unsigned long long* conf_byte_count,
			 unsigned long long* nonconf_pkt_count,
			 unsigned long long* nonconf_byte_count)
{
    flow_id_t      f_id;
    meter_stats_t  conf_stats, nonconf_stats;
    long           result;
    int            err = 0;
    
    QOS_DBG_TRACE ();
    QOS_DBG_PRINT (LOG_INFO, "Get Meter Statistics Parameters!\n\t"
		   "Flow Id    = %d \n", flow_id);

    f_id = (unsigned long) flow_id;

    /* Get stats for conforming (GREEN) */
    ECK (stub_meter_get_stats (&g_meter_ace, f_id, srtcm_conf_level_t_GREEN,
			       &conf_stats, &result));
    /* Get stats for Non conforming (GREEN) */
    ECK (stub_meter_get_stats (&g_meter_ace, f_id, srtcm_conf_level_t_RED,
			       &nonconf_stats, &result));
    
    if (result == QOS_RTN_SUCCESS)
	QOS_DBG_PRINT (LOG_INFO, "Meter:get_stats() successful!\n");
    else if (result == QOS_RTN_ERR_FIDRANGE)
	QOS_DBG_PRINT (LOG_INFO, "Meter:get_stats() Error : Flow ID out"
		       " of range \n");
    else if (result == QOS_RTN_ERR_NOSTATSAVAIL)
	QOS_DBG_PRINT (LOG_INFO, "Meter:get_stats() Error : No Stats"
		       "available \n");
    else if (result == MTR_RTN_ERR_CONFLVLRANGE)
  	QOS_DBG_PRINT (LOG_INFO, "Meter:get_stats() Error : Conf. Level" 
  		       "not valid \n");
/*      else if (result == QOS_RTN_ERR_FIDUNUSED) */
/*    	QOS_DBG_PRINT (LOG_INFO, "Meter:get_stats() Error : Flow Id"  */
/*    		       "not in use \n"); */
/*      else if (result == QOS_RTN_ERR_SUCCESS_32BIT_STATS) */
/*  	QOS_DBG_PRINT (LOG_INFO, "Meter:get_stats() : Only 32-bit" */
/*  		       "Stats available \n"); */

    if (result != QOS_RTN_SUCCESS) {
	err = -1;
	PRETURN (err);
    }

    *conf_pkt_count = conf_stats.pkt_count;
    *conf_byte_count = conf_stats.byte_count;
    *nonconf_pkt_count = nonconf_stats.pkt_count;
    *nonconf_byte_count = nonconf_stats.byte_count;

    PRETURN (err);
}

/**************************************************************************
*
*   Function: qos_get_dropper_stats
*
*   Parameters: IN: Flow Id
*               OUT:Packets dropped, Packets consumed
* 
*   Description: Gets statistics counters from the specified dropper
*
*   PRETURN : 0 on success, -1 for failure 
*************************************************************************/

int qos_get_dropper_stats (uint32_t flow_id,
			   unsigned long long* pkts_dropped,
			   unsigned long long* pkts_consumed)
{
    flow_id_t      f_id;
    bufmgr_stats_t flow_stats;
    long           result;
    int            err = 0;
    
    QOS_DBG_TRACE ();
    QOS_DBG_PRINT (LOG_INFO, "Get Dropper Statistics Parameters!\n\t"
		   "Flow Id    = %d \n", flow_id);

    f_id = (unsigned long) flow_id;
 
    ECK (stub_wred_bufmgr_get_flow_stats (&g_bufmgr_ace, f_id,
				     &flow_stats, &result));
    
    if (result == QOS_RTN_SUCCESS)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:get_stats() successful!\n");
    else if (result == QOS_RTN_ERR_FIDRANGE)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:get_stats() Error : Flow ID out"
		       " of range \n");
    else if (result == QOS_RTN_ERR_NOSTATSAVAIL)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:get_stats() Error : No Stats"
		       "available \n");
/*      else if (result == QOS_RTN_ERR_FIDUNUSED) */
/*    	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:get_stats() Error : Flow Id"  */
/*    		       "not in use \n"); */
/*      else if (result == QOS_RTN_ERR_SUCCESS_32BIT_STATS) */
/*  	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:get_stats() : Only 32-bit" */
/*  		       "Stats available \n"); */

    if (result != QOS_RTN_SUCCESS) {
	err = -1;
	PRETURN (err);
    }

    *pkts_dropped = flow_stats.pkt_dropped;
    *pkts_consumed = flow_stats.pkt_consumed;

    PRETURN (err);
}

/**************************************************************************
*
*   Function: qos_get_queue_stats
*
*   Parameters: IN: Queue Group Id
*               IN: Queue Id
*               OUT:Packets dropped, Packets consumed
* 
*   Description: Gets statistics counters from the specified queue
*
*   PRETURN : 0 on success, -1 for failure 
*************************************************************************/

int qos_get_queue_stats (uint32_t queue_grp_id, uint32_t queue_id,
			 unsigned long long* pkts_dropped,
			 unsigned long long* pkts_consumed)
{
    queue_id_t     q_id;
    queue_grp_t    q_grp_id;
    bufmgr_stats_t q_stats;
    long           result;
    int            err = 0;
    
    QOS_DBG_TRACE ();
    QOS_DBG_PRINT (LOG_INFO, "Get Queue Statistics Parameters!\n\t"
		   "Queue Id    = %d \n\t"
		   "Queue Grp Id= %d \n",queue_id, queue_grp_id);

    q_id = (unsigned long) queue_id;
    q_grp_id = (unsigned long) queue_grp_id;
 
    ECK (stub_bufmgr_get_queue_stats (&g_bufmgr_ace, q_grp_id, q_id,
				      &q_stats, &result));
    
    if (result == QOS_RTN_SUCCESS)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:get_stats() successful!\n");
    else if (result == QOS_RTN_ERR_QIDRANGE)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:get_stats() Error : Q ID out"
		       " of range \n");
    else if (result == QOS_RTN_ERR_QGRPRANGE)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:get_stats() Error : Q Grp"
		       "out of range \n");
    else if (result == QOS_RTN_ERR_NOSTATSAVAIL)
	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:get_stats() Error : No Stats"
		       "available \n");
/*      else if (result == QOS_RTN_ERR_QIDUNUSED) */
/*  	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:get_stats() Error : Q Id" */
/*  		       "not in use \n"); */
/*      else if (result == QOS_RTN_ERR_SUCCESS_32BIT_STATS) */
/*  	QOS_DBG_PRINT (LOG_INFO, "WRED Buf Mgr:get_stats() : Only 32-bit" */
/*  		       "Stats available \n"); */

    if (result != QOS_RTN_SUCCESS) {
	err = -1;
	PRETURN (err);
    }

    *pkts_dropped = q_stats.pkt_dropped;
    *pkts_consumed = q_stats.pkt_consumed;

    PRETURN (err);
}

/**************************************************************************
*
*   Function: qos_get_scheduler_stats
*
*   Parameters: IN: Queue Group Id
*               IN: Queue Id
*               OUT:Packet Count, Byte Count
* 
*   Description: Gets statistics counters from the specified scheduler
*
*   PRETURN : 0 on success, -1 for failure 
*************************************************************************/

int qos_get_scheduler_stats (uint32_t queue_grp_id, uint32_t queue_id,
			     unsigned long long* packet_count,
			     unsigned long long* byte_count)
{
    queue_id_t  q_id;
    queue_grp_t q_grp_id;
    counter_t   pkt_cnt, byte_cnt;
    long        result;
    int         err = 0;
    
    QOS_DBG_TRACE ();
    QOS_DBG_PRINT (LOG_INFO, "Get Scheduler Statistics Parameters!\n\t"
		   "Queue Id    = %d \n\t"
		   "Queue Grp Id= %d \n",queue_id, queue_grp_id);

    q_id = (unsigned long) queue_id;
    q_grp_id = (unsigned long) queue_grp_id;
 
    ECK (stub_scheduler_get_stats (&g_sched_ace, q_grp_id, q_id,
				   &pkt_cnt, &byte_cnt, &result));

    if (result == QOS_RTN_SUCCESS)
	QOS_DBG_PRINT (LOG_INFO, "DRR Scheduler:get_stats() successful!\n");
    else if (result == QOS_RTN_ERR_QIDRANGE)
	QOS_DBG_PRINT (LOG_INFO, "DRR Scheduler:get_stats() Error : Q ID out"
		       " of range \n");
    else if (result == QOS_RTN_ERR_QGRPRANGE)
	QOS_DBG_PRINT (LOG_INFO, "DRR Scheduler:get_stats() Error : Q Grp"
		       "out of range \n");
    else if (result == QOS_RTN_ERR_NOSTATSAVAIL)
	QOS_DBG_PRINT (LOG_INFO, "DRR Scheduler:get_stats() Error : No Stats"
		       "available \n");

    if (result != QOS_RTN_SUCCESS) {
	err = -1;
	PRETURN (err);
    }
    *packet_count = pkt_cnt;
    *byte_count   = byte_cnt;

    PRETURN (err);
}
