/*****************************************************************************
*                  INTEL CORPORATION PROPRIETARY INFORMATION
*  This software is supplied under the terms of a license agreement or
*  nondisclosure agreement with Intel Corporation and may not be copied
*  or disclosed except in accordance with the terms of that agreement.
*  Copyright (c) 2001 Intel Corporation. All Rights Reserved.
*****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <errno.h>

#include "FPModule.h"
#include "backendapi_types.h"
#include "qos_mgr_impl.h"

/* Callback Handles - Callbacks registered with Back end API */
static FPPI_CBHANDLE g_dpe_cb_hdl;

static FPPI_CONTEXT g_dpe_cntxt;

static void fpm_qos_dpe_cb (FPPI_CBCORRELATOR  cbcorrelator,
				FPPI_CONTEXT       context,
				void*              response_data);

static DList	   ReqList;
PilCriticalSection LOCK1;
#define REQ_LOCK()\
         PIL_EnterCriticalSection( &LOCK1 )
#define REQ_UNLOCK() \
        PIL_LeaveCriticalSection( &LOCK1 )


/*****************************************************************************
*   Function:   CompareRequest()					     *
*									     *
*   Description: Compare function for RequestList			     *
*									     *
*****************************************************************************/
static int CompareRequest(const void *data, const void *key)
{
    return(data == key);
}
/*****************************************************************************
*									     *
*   Function:   fpm_qos_mgr_init()					     *
*									     *
*   Description: This function initializes the QoS Module. It registers for  *
*                callbacks with the Back-end API, and initializes the DPE    *
*                Manager and FP Engine (platform specific QoS component)     *
*									     *
*****************************************************************************/
int fpm_qos_mgr_init()
{
    int errtype;

    PIL_InitializeCriticalSection(&LOCK1);

    /* Assigning Contexts for registering for callbacks */

    g_dpe_cntxt = (void *) FPPI_DS_DpeOperation;

    /* Register DPE API callbacks */
    errtype = BENDAPI_register_callback (&fpm_qos_dpe_cb, 
					 &g_dpe_cb_hdl,
					 FPPI_DS_DpeOperation, 
					 g_dpe_cntxt);
    if (FPPI_SUCCESS != errtype) {
	return (FPPI_FAILURE);
    }

    npf_list_init(&ReqList, PIL_FreeMemory);
    ReqList.compare = &CompareRequest;
    if(fpm_qos_dpe_init() != FPPI_SUCCESS)
    {
	return (FPPI_FAILURE);
    }

    return (0);
}

/******************************************************************************
*   Function:   fpm_qos_mgr_shutdown()					      *
*									      *
*   Description: This function shuts-down the QoS Module. It deregisters for  *
*                callbacks with the Back-end API, de-initializes the FP Engine*
*									      *
******************************************************************************/
int fpm_qos_mgr_shutdown ()
{
    PIL_DestroyCriticalSection(&LOCK1);
    npf_list_destroy(&ReqList);
    if(fpm_qos_dpe_shutdown() != FPPI_SUCCESS)
	   return (FPPI_FAILURE);
    return (FPPI_SUCCESS);
}

/******************************************************************************
*   Function:   fpm_qos_dpe_cb()					      *
*									      *
*   Description: This is callback function for Back-end API registration      *
*                For each user request, this is called.			      *
*		 Each time, new request object(fpm_qos_request_t) created     *
*		 which holds user passed values. Te object is deleted when    *
*		 response is generated					      *
******************************************************************************/
void fpm_qos_dpe_cb(FPPI_CBCORRELATOR  cbcorrelator,
			 FPPI_CONTEXT  context,
			 void*         response_data)
{
    fpm_qos_request_t	*req;
    REQ_LOCK();
    if(response_data == NULL)
    {
	REQ_UNLOCK();
	return;
    }
    req = (fpm_qos_request_t *)PIL_AllocateMemory(sizeof(fpm_qos_request_t), PIL_MEM_ZERO_INIT);
    if(req == NULL)
    {
	REQ_UNLOCK();
	return;
    }
    memcpy(&req->dpe, response_data, sizeof(FPPI_DS_Dpe_t));
    req->corr = cbcorrelator;
    req->context = context;
    npf_list_pushBack(&ReqList, req);
    fpm_qos_dpe_operation(req);
    REQ_UNLOCK();
}

/******************************************************************************
*   Function:   send_ix_cc_error()					      *
*									      *
*   Description: Helper function to send response			      *
******************************************************************************/
void send_ix_cc_error(NPF_DS_ErrorType_t error)
{
    FPPI_DS_Dpe_t  dpe;
    dpe.error = error;
    REQ_LOCK();
    BENDAPI_Report_Status (FPPI_SUCCESS, FPPI_DS_DpeOperation, 0, (void *)&dpe);
    REQ_UNLOCK();
    return;
}

/******************************************************************************
*   Function:   send_ix_cc_error()					      *
*									      *
*   Description: Helper function to send response			      *
******************************************************************************/
void send_ix_cc_async_response(fpm_qos_request_t *resp)
{
    FPPI_DS_Dpe_t  dpe;
    REQ_LOCK();
    memcpy(&dpe, &resp->dpe, sizeof(FPPI_DS_Dpe_t));
    if(npf_list_searchByKey(&ReqList, resp) == NULL) 
    {
	#ifdef FPM_QOSMGR_DEBUG
	printf("No Request Found in the List, ????????????\n");
	#endif
	REQ_UNLOCK();
	return;
    }
    BENDAPI_Report_Status (FPPI_SUCCESS, FPPI_DS_DpeOperation, resp->corr, &dpe);
    npf_list_removeByKeyFree(&ReqList, resp);
    /* PIL_FreeMemory(resp);*/
    REQ_UNLOCK();
    return;
}
