/*****************************************************************************
*                  INTEL CORPORATION PROPRIETARY INFORMATION
*  This software is supplied under the terms of a license agreement or
*  nondisclosure agreement with Intel Corporation and may not be copied
*  or disclosed except in accordance with the terms of that agreement.
*  Copyright (c) 2002 Intel Corporation. All Rights Reserved.
*****************************************************************************/
#include <linux/module.h>
#include "ix_ossl.h"
#include "shim_ipv4_mgr.h"
#include "shim_mgr.h"
#include "shim_list.h"

#include "cci/cc.h"
#include "ix_cc.h"
#include "cc/ix_cc_rtmv4.h"
#include "cc/ix_cc_ipv4.h"
#include "cc/ix_cc_fpm_core.h"
#include "cc/ipv4/internal/ix_cc_ipv4_init.h"
#include "cc/ix_cc_stkdrv.h"

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

MODULE_LICENSE("Proprietary"); 

#define sh_list_head(list) ((list)->head)
#define sh_list_tail(list) ((list)->tail)
#define sh_list_size(list) ((list)->size)

#define DEV_NUMBER_SHIM 230

DList responseQueue;
ix_uint32 g_numPorts;
extern ix_cc_ipv4_context* g_pIpv4Context;

int sh_read(int, void *, int);
int sh_write(int, const void *, size_t);
ix_error sh_ipv4_cb(ix_error, void *);
ix_error sh_ipv4_lookup_route_cb(ix_error, void *,
                                  ix_cc_rtmv4_next_hop_info *);
ix_uint32 sh_onCB_cc_link_status(ix_uint32,
                   ix_uint32, ix_cc_fpm_link_status);
ix_uint32 sh_onCB_cc_ipv4_exception(ix_uint32,
                   ix_uint32, ix_uint32);
ix_error sh_l2_op_cb(ix_error, void *);
ix_error sh_cm_get_port_num_cb(ix_error arg_Result, void* arg_pContext,
                               ix_uint32* arg_pNumPorts);
ix_error sh_cm_get_port_fecaps_cb(ix_error arg_Result, void* arg_pContext,
                                  ix_cc_properties* arg_pProperty);
extern void ix_ph_receive_packet_fpm(ix_uint32 bladeid, ix_uint32 portid,
                      ix_uint32 len, char* buffer);
static ix_error sh_list_init(DList *list, void (*destroy)(void *data)); 
static void sh_list_destroy(DList *list); 
static ix_error sh_list_pushBack(DList *list, const void *data); 
static ix_error sh_list_insNext(DList *list, DListElmt *element, const void *data); 
static ix_error sh_list_popFront(DList *list, void **data); 
static ix_error sh_list_remove(DList *list, DListElmt *element, void **data); 
static void sh_free_responseQueueData(SH_GeneralResponseMessage* respMsg);

static struct file_operations shim_config_fops = {
   read:  sh_read,
   write: sh_write,
}; 

/* Allocation of the Major number */
ix_int32 g_sh_MajorNumber = DEV_NUMBER_SHIM;
ix_ossl_sem_t g_sh_queue_sid = 0;
/**
 * NAME: init_module
 *
 * DESCRIPTION:This function gets called when the module is inserted into
 * the kernel. This function registers the shim device driver with the
 * kernel. 
 * 
 * @Param:  - IN - none
 *
 * @Param:  - OUT - none
 * 
 * @Return:  0 on success
 *          -1 on failure
 *
 * Error Codes:
 *
 *        None. 
 */

int init_module(void)
{
#ifdef IX_DEBUG
   ix_ossl_message_log("Loading shim driver\n");
#endif
   if (register_chrdev(g_sh_MajorNumber, "Shim", &shim_config_fops) == 0)
   {
#ifdef IX_DEBUG
      ix_ossl_message_log("Shim driver is registered with major number: %ld\n", g_sh_MajorNumber);
      ix_ossl_message_log("USAGE:\n");
      ix_ossl_message_log("mknod /dev/Shim c %ld <minor>\n", g_sh_MajorNumber);
      ix_ossl_message_log("with different minor numbers\n\n");
#endif
   } 
   else
   {
#ifdef IX_DEBUG
      ix_ossl_message_log("Shim driver cannot be registered !!!\n");
#endif
      return -1;
   } 

   sh_list_init(&responseQueue, ix_ossl_free); 
   IX_ERROR_CR(ix_ossl_sem_init(IX_OSSL_SEM_UNAVAILABLE, &g_sh_queue_sid));
   return 0;
} /* init_module */             

/**
 * NAME: init_module
 *
 * DESCRIPTION:This function gets called when the module is removed from
 * the kernel. This function unregisters the shim device driver from the
 * kernel. 
 * 
 * @Param:  - IN - none
 *
 * @Param:  - OUT - none
 * 
 * @Return:  0 on success
 *          -1 on failure
 *
 * Error Codes:
 *
 *        None. 
 */

void cleanup_module(void)
{
#ifdef IX_DEBUG
   ix_ossl_message_log("Unloading Shim driver \n");
#endif
   
   if(unregister_chrdev(g_sh_MajorNumber, "Shim") == 0)
   {
#ifdef IX_DEBUG
      ix_ossl_message_log("Shim driver cleaned successfully ...\n");
#endif
   }
   else
   {
#ifdef IX_DEBUG
      ix_ossl_message_log("Shim driver could not be cleaned up !!!\n");
#endif
   } 

   sh_list_destroy(&responseQueue);
   IX_ERROR_CR(ix_ossl_sem_fini(g_sh_queue_sid)); 
} /* cleanup_module */      

/**
 * NAME: sh_write
 *
 * DESCRIPTION:This is the write() function implementation for the shim 
 * device driver. The write function calls the core component function.
 * 
 * @Param:  - IN -  filedes: File descriptor for the character device opened
 *          - IN -  buf:     Buffer containing the data
 *          - IN -  nbyte:   Number of bytes in the buffer  
 *
 * @Param:  - OUT - none
 * 
 * @Return:  Number of bytes written on success
 *           -1 on failure
 *
 * Error Codes:
 *
 *        None. 
 */

int sh_write(
   int              filedes,
   const void*      buf,
   size_t           nbyte)
{
   SH_Context_t*   shContext; /* Pointer to shim context */
   ix_error        errorCode; /* Error code */
  
#ifdef IX_DEBUG
   ix_ossl_message_log("Entering sh_write() with buffer size: %d\n", nbyte);
#endif 
   shContext = (SH_Context_t*)buf;
   switch (shContext->compName) 
   {
      case SH_IPV4_CC:
      switch(shContext->u.ipv4Action)
      {
         case SH_IPV4_ASYNC_ADD_ROUTE:
#ifdef IX_DEBUG
              ix_ossl_message_log("In sh_write(): ix_cc_ipv4_async_add_route will be called\n");
#endif 

              errorCode = ix_cc_ipv4_async_add_route(
                             *((ix_uint32*)(shContext->params[0])), 
                             *((ix_uint32*)(shContext->params[1])),
                             *((ix_cc_rtmv4_nhid*)(shContext->params[2])),
                             sh_ipv4_cb,
                             (void*)buf);
              shContext->requestStatus = errorCode;
              if (errorCode != IX_SUCCESS)
                 return -1;
              else
                 return nbyte;     
              break;

        case SH_IPV4_ASYNC_DEL_ROUTE: 
#ifdef IX_DEBUG
             ix_ossl_message_log("In sh_write(): ix_cc_ipv4_async_delete_route will be called\n");
#endif 

             errorCode = ix_cc_ipv4_async_delete_route(
                            *((ix_uint32*)(shContext->params[0])),
                            *((ix_uint32*)(shContext->params[1])),
                            sh_ipv4_cb,
                             (void*)buf);
             shContext->requestStatus = errorCode;
             if (errorCode != IX_SUCCESS)
                return -1;
             else
                return nbyte;     
             break;

        case SH_IPV4_ASYNC_UPDATE_ROUTE:
#ifdef IX_DEBUG
             ix_ossl_message_log("In sh_write(): ix_cc_ipv4_async_update_route will be called\n");
#endif 
             errorCode = ix_cc_ipv4_async_update_route(
                            *((ix_uint32*)(shContext->params[0])),
                            *((ix_uint32*)(shContext->params[1])),
                            *((ix_cc_rtmv4_nhid*)(shContext->params[2])),
                            sh_ipv4_cb,
                            (void*)buf);
             shContext->requestStatus = errorCode;
             if (errorCode != IX_SUCCESS)
                return -1;
             else
                return nbyte;     
             break;

       case SH_IPV4_ASYNC_LOOKUP_ROUTE:
#ifdef IX_DEBUG
            ix_ossl_message_log("In sh_write(): ix_cc_ipv4_async_lookup_route will be called\n");
#endif 
            errorCode = ix_cc_ipv4_async_lookup_route(
                            *((ix_uint32*)(shContext->params[0])),
                            sh_ipv4_lookup_route_cb,
                            (void*)buf);
             shContext->requestStatus = errorCode;
             if (errorCode != IX_SUCCESS)
                return -1;
             else
                return nbyte;     
             break;
                           
       case SH_IPV4_ASYNC_FLUSH_PREFIX:
#ifdef IX_DEBUG
            ix_ossl_message_log("In sh_write(): ix_cc_ipv4_async_purge_routes will be called\n");
#endif 
            errorCode = ix_cc_ipv4_async_purge_routes();
            shContext->requestStatus = errorCode;
            if (errorCode != IX_SUCCESS)
               return -1;
            else
               return nbyte;     
            break;

       case SH_IPV4_ASYNC_ADD_NHOP:
#ifdef IX_DEBUG
            ix_ossl_message_log("In sh_write(): ix_cc_ipv4_async_add_next_hop will be called\n");
#endif 
            errorCode = ix_cc_ipv4_async_add_next_hop(
                           *((ix_cc_rtmv4_nhid*)(shContext->params[0])),
                           ((ix_cc_rtmv4_next_hop_info*)(shContext->params[1])),
                           sh_ipv4_cb, 
                           (void*)buf); 
            shContext->requestStatus = errorCode;
            if (errorCode != IX_SUCCESS)
               return -1;
            else
               return nbyte;     
            break;

       case SH_IPV4_ASYNC_DEL_NHOP:
#ifdef IX_DEBUG
            ix_ossl_message_log("In sh_write(): ix_cc_ipv4_async_delete_next_hop will be called\n");
#endif 
            errorCode = ix_cc_ipv4_async_delete_next_hop(
                           *((ix_cc_rtmv4_nhid*)(shContext->params[0])),
                           sh_ipv4_cb,
                           (void*)buf);
            shContext->requestStatus = errorCode;
            if (errorCode != IX_SUCCESS)
               return -1;
            else
               return nbyte;     
            break;

      case SH_IPV4_ASYNC_UPDATE_NHOP:
#ifdef IX_DEBUG
           ix_ossl_message_log("In sh_write(): ix_cc_ipv4_async_update_next_hop will be called\n");
#endif 
           errorCode = ix_cc_ipv4_async_update_next_hop(
                          *((ix_cc_rtmv4_nhid*)(shContext->params[0])),
                          ((ix_cc_rtmv4_next_hop_info*)(shContext->params[1])),
                          sh_ipv4_cb,
                          (void*)buf);
            shContext->requestStatus = errorCode;
            if (errorCode != IX_SUCCESS)
               return -1;
            else
               return nbyte;     
            break;

      case SH_IPV4_ASYNC_GET_NHOP:
#ifdef IX_DEBUG
           ix_ossl_message_log("In sh_write(): ix_cc_ipv4_async_get_next_hop will be called\n");
#endif 
/*           errorCode = ix_cc_ipv4_async_get_next_hop(
                          *((ix_cc_rtmv4_nhid*)(shContext->params[0])),
                           Call back function here ,
                          (void*)buf);
            shContext->requestStatus = errorCode;
            if (errorCode != IX_SUCCESS)
               return -1;
            else
               return nbyte;     */
            break;

       case SH_IPV4_ASYNC_FLUSH_NHOP:
#ifdef IX_DEBUG
            ix_ossl_message_log("In sh_write(): ix_cc_ipv4_async_purge_rtm will be called\n");
#endif 
            errorCode = ix_cc_ipv4_async_purge_rtm();
            shContext->requestStatus = errorCode;
            if (errorCode != IX_SUCCESS)
               return -1;
            else
               return nbyte;     
            break;

       case SH_IPV4_GET_NHOP:
#ifdef IX_DEBUG
            ix_ossl_message_log("In sh_write(): ix_cc_rtmv4_get_next_hop will be called\n");
#endif 
            errorCode = ix_cc_rtmv4_get_next_hop(
                           g_pIpv4Context->hRtm,
                           *((ix_cc_rtmv4_nhid*)(shContext->params[1])),
                            ((ix_cc_rtmv4_next_hop_info*)(shContext->params[2])));
            shContext->requestStatus = errorCode;
            if (errorCode != IX_SUCCESS)
               return -1;
            else
               return nbyte;     
            break;
 
       case SH_IPV4_GET_ROUTE:
#ifdef IX_DEBUG
            ix_ossl_message_log("In sh_write(): ix_cc_rtmv4_get_route will be called\n");
#endif 
            errorCode = ix_cc_rtmv4_get_route(
                           g_pIpv4Context->hRtm,
                           *((ix_uint32*)(shContext->params[1])),
                           *((ix_uint32*)(shContext->params[2])), 
                           ((ix_cc_rtmv4_nhid*)(shContext->params[3])));
            shContext->requestStatus = errorCode;
            if (errorCode != IX_SUCCESS)
               return -1;
            else
               return nbyte;     
            break;

        default:
#ifdef IX_DEBUG
           ix_ossl_message_log("In sh_write(): Undefined action for IPv4 Core Component \n");
#endif
      }
      break;
       
      case SH_FPM_CC:
      switch(shContext->u.fpmAction)
      {
         case SH_FPM_REGISTER_LINKST:
#ifdef IX_DEBUG
            ix_ossl_message_log("In sh_write(): ix_cc_fpm_register_link_status_cb will be called\n");
#endif 
            errorCode = ix_cc_fpm_register_link_status_cb((void*)&sh_onCB_cc_link_status);
            shContext->requestStatus = errorCode;
            if (errorCode != IX_SUCCESS)
               return -1;
            else
               return nbyte;     
            break;

         case SH_FPM_REGISTER_IPV4EXCP:
#ifdef IX_DEBUG
            ix_ossl_message_log("In sh_write(): ix_cc_fpm_register_ipv4_exception_cb will be called\n");
#endif 
            errorCode = ix_cc_fpm_register_ipv4_exception_cb((void*)&sh_onCB_cc_ipv4_exception);
            shContext->requestStatus = errorCode;
            if (errorCode != IX_SUCCESS)
               return -1;
            else
               return nbyte;     
            break;
          
         case SH_FPM_REGISTER_PKTHDLR:
#ifdef IX_DEBUG
            ix_ossl_message_log("In sh_write(): ix_cc_fpm_register_pkt_hdlr_cb will be called\n");
#endif 
            errorCode = ix_cc_fpm_register_pkt_hdlr_cb((ix_cc_fpm_pkt_hdlr_cb)&ix_ph_receive_packet_fpm);
            shContext->requestStatus = errorCode;
            if (errorCode != IX_SUCCESS)
               return -1;
            else
               return nbyte;     
            break;

         case SH_FPM_DEREGISTER_LINKST:
#ifdef IX_DEBUG
            ix_ossl_message_log("In sh_write(): ix_cc_fpm_deregister_link_status_cb will be called\n");
#endif 
            errorCode = ix_cc_fpm_deregister_link_status_cb();
            shContext->requestStatus = errorCode;
            if (errorCode != IX_SUCCESS)
               return -1;
            else
               return nbyte;     
            break;
            
         case SH_FPM_DEREGISTER_IPV4EXCP:
#ifdef IX_DEBUG
            ix_ossl_message_log("In sh_write(): ix_cc_fpm_deregister_ipv4_exception_cb will be called\n");
#endif 
            errorCode = ix_cc_fpm_deregister_ipv4_exception_cb();
            shContext->requestStatus = errorCode;
            if (errorCode != IX_SUCCESS)
               return -1;
            else
               return nbyte;     
            break;

         case SH_FPM_DEREGISTER_PKTHDLR:
#ifdef IX_DEBUG
            ix_ossl_message_log("In sh_write(): ix_cc_fpm_deregister_pkt_hdlr_cb will be called\n");
#endif 
            errorCode = ix_cc_fpm_deregister_pkt_hdlr_cb();
            shContext->requestStatus = errorCode;
            if (errorCode != IX_SUCCESS)
               return -1;
            else
               return nbyte;     
            break;
           
         case SH_FPM_ADD_L2ENTRY:
#ifdef IX_DEBUG
              ix_ossl_message_log("In sh_write(): ix_cc_fpm_sync_add_l2_entry will be called\n");
#endif 
              errorCode = ix_cc_fpm_sync_add_l2_entry (
                             *((ix_uint32*)(shContext->params[0])),
                             *((ix_uint32*)(shContext->params[1])),
                             ((ix_cc_l2tm_entry*)(shContext->params[2])));

              shContext->requestStatus = errorCode;
              if (errorCode != IX_SUCCESS)
                 return -1;
              else
                 return nbyte;     
              break;

         case SH_FPM_DEL_L2ENTRY:
#ifdef IX_DEBUG
              ix_ossl_message_log("In sh_write(): ix_cc_fpm_sync_del_l2_entry will be called\n");
#endif 
              errorCode = ix_cc_fpm_sync_del_l2_entry (
                             *((ix_uint32*)(shContext->params[0])));
              shContext->requestStatus = errorCode;
              if (errorCode != IX_SUCCESS)
                 return -1;
              else
                 return nbyte;     
              break;

         case SH_FPM_ASYNC_ADD_L2ENTRY:
#ifdef IX_DEBUG
              ix_ossl_message_log("In sh_write(): ix_cc_fpm_async_add_l2_entry will be called\n");
#endif 
              errorCode = ix_cc_fpm_async_add_l2_entry (
                             *((ix_uint32*)(shContext->params[0])),
                             *((ix_uint32*)(shContext->params[1])),
                              ((ix_cc_l2tm_entry*)(shContext->params[2])),
                              sh_l2_op_cb,
                              (void*)buf); 
              shContext->requestStatus = errorCode;
              if (errorCode != IX_SUCCESS)
                 return -1;
              else
                 return nbyte;     
              break;

         case SH_FPM_ASYNC_DEL_L2ENTRY:
#ifdef IX_DEBUG
              ix_ossl_message_log("In sh_write(): ix_cc_fpm_async_del_l2_entry will be called\n");
#endif 
              errorCode = ix_cc_fpm_async_del_l2_entry (
                             *((ix_uint32*)(shContext->params[0])),
                              sh_l2_op_cb, 
                              (void*)buf); 
              shContext->requestStatus = errorCode;
              if (errorCode != IX_SUCCESS)
                 return -1;
              else
                 return nbyte;     
              break;
                   
         case SH_FPM_ASYNC_ADD_ARPENTRY:
#ifdef IX_DEBUG
              ix_ossl_message_log("In sh_write(): ix_cc_fpm_async_add_arp_entry will be called\n");
#endif 
              errorCode = ix_cc_fpm_async_add_arp_entry (
                             ((ix_cc_eth_tx_next_hop_info*)(shContext->params[0])),
                              sh_ipv4_cb, 
                              (void*)buf); 
              shContext->requestStatus = errorCode;
              if (errorCode != IX_SUCCESS)
                 return -1;
              else
                 return nbyte;     
              break;

         case SH_FPM_ASYNC_DEL_ARPENTRY:
#ifdef IX_DEBUG
              ix_ossl_message_log("In sh_write(): ix_cc_fpm_async_del_arp_entry will be called\n");
#endif 
              errorCode = ix_cc_fpm_async_del_arp_entry (
                             *((ix_uint32*)(shContext->params[0])),
                              sh_ipv4_cb, 
                              (void*)buf); 
              shContext->requestStatus = errorCode;
              if (errorCode != IX_SUCCESS)
                 return -1;
              else
                 return nbyte;     
              break;

         case SH_FPM_FLUSH_ARP:
#ifdef IX_DEBUG
              ix_ossl_message_log("In sh_write(): ix_cc_fpm_sync_purge_arp_table will be called\n");
#endif 
              errorCode = ix_cc_fpm_sync_purge_arp_table();
              shContext->requestStatus = errorCode;
              if (errorCode != IX_SUCCESS)
                 return -1;
              else
                 return nbyte;     
              break;

         case SH_FPM_CONFIG_CPADDR:
#ifdef IX_DEBUG
              ix_ossl_message_log("In sh_write(): ix_ph_set_cp_address will be called\n");
#endif 
              ix_ph_set_cp_address(*((ix_uint32*)(shContext->params[0])));
              return nbyte;
         default:
#ifdef IX_DEBUG
           ix_ossl_message_log("In sh_write(): Undefined action for FPM Core Component \n");
#endif
      }
      break;

      case SH_STKDRV_CC:
      switch(shContext->u.stkdrvAction)
      {
         case SH_STKDRV_ASYNC_GET_NUMPORTS:
#ifdef IX_DEBUG
              ix_ossl_message_log("In sh_write(): ix_cc_stkdrv_async_get_num_ports will be called\n");
#endif 
              errorCode = ix_cc_stkdrv_async_get_num_ports (
                           sh_cm_get_port_num_cb,
                           (void*)buf);
              shContext->requestStatus = errorCode;
              if (errorCode != IX_SUCCESS)
                 return -1;
              else
                 return nbyte;     
              break;

         case SH_STKDRV_ASYNC_GET_PROPERTY:
#ifdef IX_DEBUG
              ix_ossl_message_log("In sh_write(): ix_cc_stkdrv_async_get_property will be called\n");
#endif 
              errorCode = ix_cc_stkdrv_async_get_property (
                             sh_cm_get_port_fecaps_cb, 
                             (void*)buf,
                            *((ix_uint32*)(shContext->params[0])),
                            *((ix_uint32*)(shContext->params[1])));
              shContext->requestStatus = errorCode;
              if (errorCode != IX_SUCCESS)
                 return -1;
              else
                 return nbyte;     
              break;
 
         case SH_STKDRV_ASYNC_SET_PROPERTY:
#ifdef IX_DEBUG
              ix_ossl_message_log("In sh_write(): ix_cc_async_set_property will be called\n");
#endif 
              errorCode = ix_cc_async_set_property (
                            *((ix_uint32*)(shContext->params[0])),
                             ((ix_cc_properties*)(shContext->params[1]))); 
              shContext->requestStatus = errorCode;
              if (errorCode != IX_SUCCESS)
                 return -1;
              else
                 return nbyte;     
              break;
 
         default:
#ifdef IX_DEBUG
           ix_ossl_message_log("In sh_write(): Undefined action for Stack Driver Core Component\n");
#endif
      }

      default:
#ifdef IX_DEBUG
           ix_ossl_message_log("In sh_write(): No support for this core component\n");
#endif
   }
   return -1;   
}

/**
 * NAME: sh_read
 *
 * DESCRIPTION:This is the read() function implementation for the shim 
 * device driver. The read function removes a response from the kernel
 * queue and copies the same to the user space. The function blocks
 * indefinitely if there is no response message in the queue.
 * 
 * @Param:  - IN -  filedes: File descriptor for the character device opened
 *          - IN -  nbyte:   Buffer size
 *
 * @Param:  - OUT -  buf:     Buffer into which the data is to be copied
 * 
 * @Return:  Number of bytes written on success
 *           -1 on failure
 *
 * Error Codes:
 *
 *        None. 
 */

int sh_read(int filedes, void* buf, int nbyte)
{
   SH_GeneralResponseMessage* respMsg; /* Pointer to the common response
                                          structure */
   ix_uint32                  count;   /* Local counter */
   char*                      pos;     /* Pointer that keeps track of the 
                                          bytes already copied into 
                                          the buffer */

   pos = (char*)buf;

   IX_ERROR_CR(ix_ossl_sem_init(IX_OSSL_SEM_UNAVAILABLE, &g_sh_queue_sid));

   if(sh_list_head(&responseQueue) == NULL)
      IX_ERROR_CR(ix_ossl_sem_take(g_sh_queue_sid, IX_OSSL_WAIT_FOREVER)); 
   /* if(sh_list_head(&responseQueue) == NULL)
      return 0; */

   if(sh_list_head(&responseQueue) != NULL)
   {
      sh_list_popFront(&responseQueue, (void**)&respMsg);
      ix_ossl_memcpy(pos, respMsg, sizeof(SH_ResponseMessageHeader));
      pos += sizeof(SH_ResponseMessageHeader); 
      for (count = 0; count < MAX_NUM_RESP; count++)
      {
         if(respMsg->outData[count].length != 0)
         {
            ix_ossl_memcpy(pos, &respMsg->outData[count].length, sizeof(ix_uint32));      
            pos += sizeof(ix_uint32);
            ix_ossl_memcpy(pos, respMsg->outData[count].responseData, respMsg->outData[count].length);
            pos += respMsg->outData[count].length;
         }
      } 
   }
   sh_free_responseQueueData(respMsg);

#ifdef IX_DEBUG
   ix_ossl_message_log("Exiting sh_read()\n");
#endif 

   return (pos - (char*)buf);
}

/**
 * NAME: sh_ipv4_cb
 *
 * DESCRIPTION:This is the proxy callback function for ipv4. It builds
 * a message and puts it in the queue.
 * 
 * @Param:  - IN -  arg_Result: Error code
 *          - IN -  arg_pContext:  pointer to shim context
 *
 * @Param:  - OUT -  None.
 * 
 * @Return:  IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL - invalid context
 * IX_CC_ERROR_OOM - memory allocation failure
 */

ix_error sh_ipv4_cb(ix_error arg_Result, void* arg_pContext)
{
   SH_Context_t*              shContext;  /* Pointer to the shim context */
   SH_GeneralResponseMessage* respMsg;    /* Pointer to the common response
                                             structure */
   ix_error                   err;        /* Error code */ 

#ifdef IX_DEBUG
   ix_ossl_message_log("Entering sh_ipv4_cb()\n");
#endif 

   if (!arg_pContext)
   {
#ifdef IX_DEBUG
      ix_ossl_message_log("Invalid context\n");
#endif
      err = IX_ERROR_NEW(IX_CC_ERROR_NULL, IX_ERROR_LEVEL_WARNING);
      return err;
   }
   shContext = (SH_Context_t*)arg_pContext;
   shContext->responseStatus = arg_Result;
   respMsg = (SH_GeneralResponseMessage*)ix_ossl_malloc(sizeof(SH_GeneralResponseMessage));
   if (respMsg == NULL)
   {
#ifdef IX_DEBUG
      ix_ossl_message_log("Failed to allocate memory\n");
#endif
      err = IX_ERROR_NEW(IX_CC_ERROR_OOM, IX_ERROR_LEVEL_WARNING);
      return err;      
   }
   ix_ossl_memset(respMsg, 0, sizeof(SH_GeneralResponseMessage));
   respMsg->respHeader.compName = shContext->compName;
   respMsg->respHeader.shContext = arg_pContext;
   respMsg->respHeader.u.ipv4Action = shContext->u.ipv4Action;
   sh_list_pushBack(&responseQueue, (const void*)respMsg);

   IX_ERROR_CR(ix_ossl_sem_give(g_sh_queue_sid)); 
#ifdef IX_DEBUG
   ix_ossl_message_log("Exiting sh_ipv4_cb()\n");
#endif 

   return IX_SUCCESS;
}
 
/**
 * NAME: sh_ipv4_lookup_route_cb
 *
 * DESCRIPTION:This is the proxy callback function for ipv4 route lookup. 
 * It builds a message and puts it in the queue.
 * 
 * @Param:  - IN -  arg_Result: Error code
 *          - IN -  arg_pContext:  pointer to shim context
 *          - IN -  arg_pNHInfo: Next hop information
 *
 * @Param:  - OUT -  None.
 * 
 * @Return:  IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL - invalid context
 * IX_CC_ERROR_OOM - memory allocation failure
 */

ix_error sh_ipv4_lookup_route_cb (ix_error arg_Result, void* arg_pContext,
                                  ix_cc_rtmv4_next_hop_info* arg_pNHInfo)
{
   SH_Context_t*              shContext;  /* Pointer to the shim context */
   SH_GeneralResponseMessage* respMsg;    /* Pointer to the common response
                                             structure */
   ix_error                   err;        /* Error code */ 

#ifdef IX_DEBUG
   ix_ossl_message_log("Entering sh_ipv4_lookup_route_cb()\n");
#endif 

   if (!arg_pContext)
   {
#ifdef IX_DEBUG
      ix_ossl_message_log("Invalid context\n");
#endif
      err = IX_ERROR_NEW(IX_CC_ERROR_NULL, IX_ERROR_LEVEL_WARNING);
      return err;
   }
   shContext = (SH_Context_t*)arg_pContext;
   shContext->responseStatus = arg_Result;
   respMsg = (SH_GeneralResponseMessage*)ix_ossl_malloc(sizeof(SH_GeneralResponseMessage));
   if (respMsg == NULL)
   {
#ifdef IX_DEBUG
      ix_ossl_message_log("Failed to allocate memory\n");
#endif
      err = IX_ERROR_NEW(IX_CC_ERROR_OOM, IX_ERROR_LEVEL_WARNING);
      return err;      
   }
   ix_ossl_memset(respMsg, 0, sizeof(SH_GeneralResponseMessage));
   respMsg->respHeader.compName = shContext->compName;
   respMsg->respHeader.shContext = arg_pContext;
   respMsg->respHeader.u.ipv4Action = shContext->u.ipv4Action;
   respMsg->outData[0].length = sizeof(ix_cc_rtmv4_next_hop_info);
   respMsg->outData[0].responseData = (ix_cc_rtmv4_next_hop_info*)ix_ossl_malloc(sizeof(ix_cc_rtmv4_next_hop_info));
   ix_ossl_memcpy(respMsg->outData[0].responseData, arg_pNHInfo, 
                 sizeof(ix_cc_rtmv4_next_hop_info));

   sh_list_pushBack(&responseQueue, (const void*)respMsg);

   IX_ERROR_CR(ix_ossl_sem_give(g_sh_queue_sid)); 
#ifdef IX_DEBUG
   ix_ossl_message_log("Exiting sh_ipv4_lookup_route_cb()\n");
#endif 

   return IX_SUCCESS;

}

/**
 * NAME: sh_onCB_cc_link_status
 *
 * DESCRIPTION:This is the proxy callback function for ipv4 route lookup. 
 * It builds a message and puts it in the queue.
 * 
 * @Param:  - IN -  bladeId: Blade identifier
 *          - IN -  portId:  Port identifier that has originated the event
 *          - IN -  linkState: New link state
 *
 * @Param:  - OUT -  None.
 * 
 * @Return:  IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_OOM - memory allocation failure
 */

ix_uint32 sh_onCB_cc_link_status (ix_uint32 bladeId,
                   ix_uint32 portId, ix_cc_fpm_link_status linkState)
{
   SH_GeneralResponseMessage* respMsg;    /* Pointer to the common response
                                             structure */
   ix_error                   err;        /* Error code */

#ifdef IX_DEBUG
   ix_ossl_message_log("Entering sh_onCB_cc_link_status()\n");
#endif 

   respMsg = (SH_GeneralResponseMessage*)ix_ossl_malloc(sizeof(SH_GeneralResponseMessage));
   if (respMsg == NULL)
   {
#ifdef IX_DEBUG
      ix_ossl_message_log("Failed to allocate memory\n");
#endif
      err = IX_ERROR_NEW(IX_CC_ERROR_OOM, IX_ERROR_LEVEL_WARNING);
      return err;      
   }
   ix_ossl_memset(respMsg, 0, sizeof(SH_GeneralResponseMessage));
   respMsg->respHeader.compName = SH_FPM_CC;
   respMsg->respHeader.shContext = NULL;
   respMsg->respHeader.u.fpmAction = SH_FPM_EVENT_LINKST;
   respMsg->outData[0].length = sizeof(ix_uint32);
   respMsg->outData[0].responseData = (ix_uint32*)ix_ossl_malloc(sizeof(ix_uint32));
   ix_ossl_memcpy(respMsg->outData[0].responseData, &bladeId, 
                 sizeof(ix_uint32));
   respMsg->outData[1].length = sizeof(ix_uint32);
   respMsg->outData[1].responseData = (ix_uint32*)ix_ossl_malloc(sizeof(ix_uint32));
   ix_ossl_memcpy(respMsg->outData[1].responseData, &portId, 
                 sizeof(ix_uint32));
   respMsg->outData[2].length = sizeof(ix_cc_fpm_link_status);
   respMsg->outData[2].responseData = (ix_cc_fpm_link_status*)ix_ossl_malloc(sizeof(ix_cc_fpm_link_status));
   ix_ossl_memcpy(respMsg->outData[2].responseData, &linkState, 
                 sizeof(ix_cc_fpm_link_status));
   sh_list_pushBack(&responseQueue, (const void*)respMsg);

   IX_ERROR_CR(ix_ossl_sem_give(g_sh_queue_sid)); 
#ifdef IX_DEBUG
   ix_ossl_message_log("Exiting sh_onCB_cc_link_status()\n");
#endif 

   return IX_SUCCESS;
}

/**
 * NAME: sh_onCB_cc_ipv4_exception
 *
 * DESCRIPTION:This is the proxy callback function for ipv4 direct host
 * attach event. It builds a message and puts it in the queue.
 * 
 * @Param:  - IN -  bladeId: Blade identifier
 *          - IN -  portId:  Port identifier that has originated the event
 *          - IN -  dest: Destination address that caused this event
 *
 * @Param:  - OUT -  None.
 * 
 * @Return:  IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_OOM - memory allocation failure
 */

ix_uint32 sh_onCB_cc_ipv4_exception (ix_uint32 bladeId,
                   ix_uint32 portId, ix_uint32 dest)
{
   SH_GeneralResponseMessage* respMsg;    /* Pointer to the common response
                                             structure */
   ix_error                   err;        /* Error code */

#ifdef IX_DEBUG
   ix_ossl_message_log("Entering sh_onCB_cc_ipv4_exception()\n");
#endif 

   respMsg = (SH_GeneralResponseMessage*)ix_ossl_malloc(sizeof(SH_GeneralResponseMessage));
   if (respMsg == NULL)
   {
#ifdef IX_DEBUG
      ix_ossl_message_log("Failed to allocate memory\n");
#endif
      err = IX_ERROR_NEW(IX_CC_ERROR_OOM, IX_ERROR_LEVEL_WARNING);
      return err;      
   }
   ix_ossl_memset(respMsg, 0, sizeof(SH_GeneralResponseMessage));
   respMsg->respHeader.compName = SH_FPM_CC;
   respMsg->respHeader.shContext = NULL;
   respMsg->respHeader.u.fpmAction = 
   respMsg->outData[0].length = sizeof(ix_uint32);
   respMsg->outData[0].responseData = (ix_uint32*)ix_ossl_malloc(sizeof(ix_uint32));
   ix_ossl_memcpy(respMsg->outData[0].responseData, &bladeId, 
                 sizeof(ix_uint32));
   respMsg->outData[1].length = sizeof(ix_uint32);
   respMsg->outData[1].responseData = (ix_uint32*)ix_ossl_malloc(sizeof(ix_uint32));
   ix_ossl_memcpy(respMsg->outData[1].responseData, &portId, 
                 sizeof(ix_uint32));
   respMsg->outData[2].length = sizeof(ix_uint32);
   respMsg->outData[2].responseData = (ix_uint32*)ix_ossl_malloc(sizeof(ix_uint32));
   ix_ossl_memcpy(respMsg->outData[2].responseData, &dest, 
                 sizeof(ix_uint32));
   sh_list_pushBack(&responseQueue, (const void*)respMsg);

   IX_ERROR_CR(ix_ossl_sem_give(g_sh_queue_sid)); 
#ifdef IX_DEBUG
   ix_ossl_message_log("Exiting sh_onCB_cc_ipv4_exception()\n");
#endif 

   return IX_SUCCESS;
}

/**
 * NAME: sh_cm_get_port_fecaps_cb
 *
 * DESCRIPTION:This is the proxy callback function for CM. It builds
 * a message and puts it in the queue.
 * 
 * @Param:  - IN -  arg_Result: Error code
 *          - IN -  arg_pContext:  pointer to shim context
 *
 * @Param:  - OUT - arg_pProperty: Pointer to port properties
 * 
 * @Return:  IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL - invalid context
 * IX_CC_ERROR_OOM - memory allocation failure
 */

ix_error sh_cm_get_port_fecaps_cb(ix_error arg_Result, void* arg_pContext,
                                  ix_cc_properties* arg_pProperty)
{
   SH_Context_t*              shContext;/* Pointer to shim context */
   SH_GeneralResponseMessage* respMsg;  /* Pointer to common 
                                           response structure */
   ix_error                   err;      /* Error code */ 

#ifdef IX_DEBUG
   ix_ossl_message_log("Entering sh_cm_get_port_fecaps_cb()\n");
#endif 

   if (!arg_pContext)
   {
#ifdef IX_DEBUG
      ix_ossl_message_log("Invalid context\n");
#endif
      err = IX_ERROR_NEW(IX_CC_ERROR_NULL, IX_ERROR_LEVEL_WARNING);
      return err;
   }
   
   shContext = (SH_Context_t*)arg_pContext;
   shContext->responseStatus = arg_Result;
   respMsg = (SH_GeneralResponseMessage*)ix_ossl_malloc(sizeof(SH_GeneralResponseMessage));
   if (respMsg == NULL)
   {
#ifdef IX_DEBUG
      ix_ossl_message_log("Failed to allocate memory\n");
#endif
      err = IX_ERROR_NEW(IX_CC_ERROR_OOM, IX_ERROR_LEVEL_WARNING);
      return err;      
   }
   ix_ossl_memset(respMsg, 0, sizeof(SH_GeneralResponseMessage));
   respMsg->respHeader.compName = shContext->compName;
   respMsg->respHeader.shContext = arg_pContext;
   respMsg->respHeader.u.stkdrvAction = shContext->u.stkdrvAction;
   respMsg->outData[0].length = sizeof(ix_cc_properties) * g_numPorts;
   respMsg->outData[0].responseData = (ix_cc_properties*)ix_ossl_malloc(sizeof(ix_cc_properties) * g_numPorts);
   ix_ossl_memcpy(respMsg->outData[0].responseData, arg_pProperty, 
                 sizeof(ix_cc_properties) * g_numPorts);

   sh_list_pushBack(&responseQueue, (const void*)respMsg);

   IX_ERROR_CR(ix_ossl_sem_give(g_sh_queue_sid)); 

#ifdef IX_DEBUG
   ix_ossl_message_log("Exiting sh_cm_get_port_fecaps_cb()\n");
#endif 

   return IX_SUCCESS;
}

/**
 * NAME: sh_cm_get_port_num_cb
 *
 * DESCRIPTION:This is the proxy callback function for CM. It builds
 * a message and puts it in the queue.
 * 
 * @Param:  - IN -  arg_Result: Error code
 *          - IN -  arg_pContext:  pointer to shim context
 *
 * @Param:  - OUT - arg_pNumPorts: Pointer to number of ports
 * 
 * @Return:  IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL - invalid context
 * IX_CC_ERROR_OOM - memory allocation failure
 */

ix_error sh_cm_get_port_num_cb(ix_error arg_Result, void* arg_pContext,
                               ix_uint32* arg_pNumPorts)
{
   SH_Context_t*              shContext;/* Pointer to shim context */
   SH_GeneralResponseMessage* respMsg;  /* Pointer to common 
                                           response structure */
   ix_error                   err;      /* Error code */ 

#ifdef IX_DEBUG
   ix_ossl_message_log("Entering sh_cm_get_port_num_cb()\n");
#endif 

   if (!arg_pContext)
   {
#ifdef IX_DEBUG
      ix_ossl_message_log("Invalid context\n");
#endif
      err = IX_ERROR_NEW(IX_CC_ERROR_NULL, IX_ERROR_LEVEL_WARNING);
      return err;
   }

   g_numPorts = *arg_pNumPorts; 
   shContext = (SH_Context_t*)arg_pContext;
   shContext->responseStatus = arg_Result;
   respMsg = (SH_GeneralResponseMessage*)ix_ossl_malloc(sizeof(SH_GeneralResponseMessage));
   if (respMsg == NULL)
   {
#ifdef IX_DEBUG
      ix_ossl_message_log("Failed to allocate memory\n");

#endif
      err = IX_ERROR_NEW(IX_CC_ERROR_OOM, IX_ERROR_LEVEL_WARNING);
      return err;      
   }
   ix_ossl_memset(respMsg, 0, sizeof(SH_GeneralResponseMessage));
   respMsg->respHeader.compName = shContext->compName;
   respMsg->respHeader.shContext = arg_pContext;
   respMsg->respHeader.u.stkdrvAction = shContext->u.stkdrvAction;
   respMsg->outData[0].length = sizeof(ix_uint32);
   respMsg->outData[0].responseData = (ix_uint32*)ix_ossl_malloc(sizeof(ix_uint32));
   ix_ossl_memcpy(respMsg->outData[0].responseData, arg_pNumPorts, 
                 sizeof(ix_uint32));

   sh_list_pushBack(&responseQueue, (const void*)respMsg);

   IX_ERROR_CR(ix_ossl_sem_give(g_sh_queue_sid)); 
#ifdef IX_DEBUG
   ix_ossl_message_log("Exiting sh_cm_get_port_num_cb()\n");
#endif 

   return IX_SUCCESS;

}
 
/**
 * NAME: sh_l2_op_cb
 *
 * DESCRIPTION:This is the proxy callback function for L2. It builds
 * a message and puts it in the queue.
 * 
 * @Param:  - IN -  arg_Result: Error code
 *          - IN -  arg_pContext:  pointer to shim context
 *
 * @Param:  - OUT - None.
 * 
 * @Return:  IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL - invalid context
 * IX_CC_ERROR_OOM - memory allocation failure
 */

ix_error sh_l2_op_cb(ix_error arg_Result, void* arg_pContext)
{
   SH_Context_t*              shContext;/* Pointer to shim context */
   SH_GeneralResponseMessage* respMsg;  /* Pointer to common 
                                           response structure */
   ix_error                   err;      /* Error code */ 

#ifdef IX_DEBUG
   ix_ossl_message_log("Entering sh_l2_op_cb()\n");
#endif 

   if (!arg_pContext)
   {
#ifdef IX_DEBUG
      ix_ossl_message_log("Invalid context\n");
#endif
      err = IX_ERROR_NEW(IX_CC_ERROR_NULL, IX_ERROR_LEVEL_WARNING);
      return err;
   }

   /* shContext = (SH_Context_t*)arg_pContext;
   shContext->responseStatus = arg_Result; */
   respMsg = (SH_GeneralResponseMessage*)ix_ossl_malloc(sizeof(SH_GeneralResponseMessage));
   if (respMsg == NULL)
   {
#ifdef IX_DEBUG
      ix_ossl_message_log("Failed to allocate memory\n");
#endif
      err = IX_ERROR_NEW(IX_CC_ERROR_OOM, IX_ERROR_LEVEL_WARNING);
      return err;      
   }
   ix_ossl_memset(respMsg, 0, sizeof(SH_GeneralResponseMessage));

   /* Accessing the user allocated memory causes an exception in the 
      Linux kernel for this particular callback - cause unknown */ 
   /* respMsg->respHeader.compName = shContext->compName;
   respMsg->respHeader.shContext = arg_pContext;
   respMsg->respHeader.u.fpmAction = shContext->u.fpmAction; */
   respMsg->respHeader.responseStatus = arg_Result;
   respMsg->respHeader.compName = SH_FPM_CC;
   respMsg->respHeader.shContext = arg_pContext;
   respMsg->respHeader.u.fpmAction = SH_FPM_ASYNC_ADD_L2ENTRY;


   sh_list_pushBack(&responseQueue, (const void*)respMsg);

   IX_ERROR_CR(ix_ossl_sem_give(g_sh_queue_sid)); 
#ifdef IX_DEBUG
   ix_ossl_message_log("Exiting sh_l2_op_cb()\n");
#endif 

   return IX_SUCCESS;
}

/**
 * NAME: sh_list_init
 *
 * DESCRIPTION:This is the initialisation function for the list.
 * 
 * @Param:  - IN -  list: Pointer to DList structure
 *          - IN -  destroy: Pointer to the destroy function for list node
 *
 * @Param:  - OUT - None.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure. 
 *
 * Error Codes:
 * IX_CC_ERROR_OOR - Out of range value error 
 */

static ix_error sh_list_init(DList *list, void (*destroy)(void *data)) 
{
    ix_error  err = IX_SUCCESS;    /* Error code */

#ifdef IX_DEBUG
   ix_ossl_message_log("Entering sh_list_init()\n");
#endif 

    list->size = 0;
    list->destroy = destroy;
    list->compare = NULL;
    list->head = NULL;
    list->tail = NULL;
    err = ix_ossl_mutex_init(IX_OSSL_MUTEX_UNLOCK, &list->lock);
    if (err != IX_SUCCESS)
    {
        err = IX_ERROR_WARNING(IX_CC_ERROR_OOR,
                 ("Failed to create mutex"));
    }
#ifdef IX_DEBUG
   ix_ossl_message_log("Exiting sh_list_init()\n");
#endif 

    return err;
}

/**
 * NAME: sh_list_destroy
 *
 * DESCRIPTION:This is the destroy function for the list.
 * 
 * @Param:  - IN -  list: Pointer to DList structure
 *
 * @Param:  - OUT - None.
 * 
 * @Return: None.
 *
 * Error Codes:
 *      None.
 */

static void sh_list_destroy(DList *list) 
{
    void *data;

#ifdef IX_DEBUG
   ix_ossl_message_log("Entering sh_list_destroy()\n");
#endif 

    if (list==NULL) return;

    while (sh_list_size(list) > 0) 
    {

       if (sh_list_remove(list, sh_list_tail(list), (void **)&data) == 0 && 
           list->destroy != NULL) {

            /* Call a user-defined function to free 
               dynamically allocated data.  */
            list->destroy(data);
       }
    }

    ix_ossl_mutex_fini(list->lock);
    /* No operations are allowed now, 
       but clear the structure as a precaution. */
    ix_ossl_memset(list, 0, sizeof(DList));
#ifdef IX_DEBUG
   ix_ossl_message_log("Exiting sh_list_destroy()\n");
#endif 

    return;
}

/**
 * NAME: sh_list_pushBack
 *
 * DESCRIPTION:This is the pushBack function for the list.
 * 
 * @Param:  - IN -  list: Pointer to DList structure
 *          - IN -  data: Pointer to the data to be inserted
 * @Param:  - OUT - None.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.  
 *
 * Error Codes:
 * IX_CC_ERROR_NULL - invalid list/data
 * IX_CC_ERROR_OOM - memory allocation failure
 */

static ix_error sh_list_pushBack(DList *list, const void *data) 
{
    ix_error err;  /* Error code */

#ifdef IX_DEBUG
   ix_ossl_message_log("Entering sh_list_pushBack()\n");
#endif 

    if ((list == NULL) || (data == NULL))
    {
#ifdef IX_DEBUG
      ix_ossl_message_log("Invalid list or data\n");
#endif
      err = IX_ERROR_NEW(IX_CC_ERROR_NULL, IX_ERROR_LEVEL_WARNING);
      return err;
    }

    ix_ossl_mutex_lock(list->lock, IX_OSSL_MUTEX_LOCK);
        err = sh_list_insNext(list, sh_list_tail(list), data);
    ix_ossl_mutex_unlock(list->lock);
#ifdef IX_DEBUG
   ix_ossl_message_log("Exiting sh_list_pushBack()\n");
#endif 

    return err;
}

/**
 * NAME: sh_list_insNext
 *
 * DESCRIPTION:This is the insertion function for the list.
 * 
 * @Param:  - IN -  list: Pointer to DList structure
 *          - IN -  element: Pointer to the last element of the list
 *          - IN -  data: Pointer to the data to be inserted
 * @Param:  - OUT - None.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.  
 *
 * Error Codes:
 * IX_CC_ERROR_NULL - invalid list/data
 * IX_CC_ERROR_OOM - memory allocation failure
 */

static ix_error sh_list_insNext(DList *list, DListElmt *element, const void *data) 
{
    DListElmt *new_element;  /* Temporary pointer to list element */
    ix_error  err;           /* Error code */

#ifdef IX_DEBUG
   ix_ossl_message_log("Entering sh_list_insNext()\n");
#endif 

    if ((list == NULL) || (data == NULL))
    {
#ifdef IX_DEBUG
      ix_ossl_message_log("Invalid context\n");
#endif
      err = IX_ERROR_NEW(IX_CC_ERROR_NULL, IX_ERROR_LEVEL_WARNING);
      return err;
    }

    /* Allocate storage for the element. Do it now to save lock time */

    if ((new_element = (DListElmt *)ix_ossl_malloc
                   (sizeof(DListElmt))) == NULL)
    {
#ifdef IX_DEBUG
      ix_ossl_message_log("Failed to allocate memory\n");
#endif
      err = IX_ERROR_NEW(IX_CC_ERROR_OOM, IX_ERROR_LEVEL_WARNING);
      return err;      
    } 

    ix_ossl_mutex_lock(list->lock, IX_OSSL_MUTEX_LOCK);

    /* Do not allow a NULL element unless the list is empty. */

    if (element == NULL && sh_list_size(list) != 0) 
    {
        ix_ossl_mutex_unlock(list->lock);
        ix_ossl_free(new_element);
#ifdef IX_DEBUG
        ix_ossl_message_log("Invalid list/element\n");
#endif
        err = IX_ERROR_NEW(IX_CC_ERROR_NULL, IX_ERROR_LEVEL_WARNING);
        return err;
    }

    /* Insert the new element into the list. */
    new_element->data = (void *)data;

    if (sh_list_size(list) == 0) {

        /* Handle insertion when the list is empty. */
        list->head = new_element;
        list->head->prev = NULL;
        list->head->next = NULL;
        list->tail = new_element;

    } else {

        /* Handle insertion when the list is not empty. */
        new_element->next = element->next;
        new_element->prev = element;

        if (element->next == NULL)
            list->tail = new_element;
        else
            element->next->prev = new_element;

        element->next = new_element;
    }

    /* Adjust the size of the list to account for the inserted element. */

    list->size++;
    ix_ossl_mutex_unlock(list->lock);
#ifdef IX_DEBUG
   ix_ossl_message_log("Exiting sh_list_insNext()\n");
#endif 

    return IX_SUCCESS;
}

/**
 * NAME: sh_list_popFront
 *
 * DESCRIPTION: This is the popFront function of the list.
 * 
 * @Param:  - IN -  list: Pointer to DList structure
 * @Param:  - OUT - data: Data removed
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.  
 *
 * Error Codes:
 * IX_CC_ERROR_NULL - invalid list/data
 * IX_CC_ERROR_OOM - memory allocation failure
 */

static ix_error sh_list_popFront(DList *list, void **data) 
{
    ix_error  err;  /* Error code */

#ifdef IX_DEBUG
   ix_ossl_message_log("Entering sh_list_popFront()\n");
#endif 

    if (list == NULL) 
    {
#ifdef IX_DEBUG
      ix_ossl_message_log("Invalid list\n");
#endif
      err = IX_ERROR_NEW(IX_CC_ERROR_NULL, IX_ERROR_LEVEL_WARNING);
      return err;
    }

    // Same thread can enter PIL critical section mutiple times w/o blocking
    ix_ossl_mutex_lock(list->lock, IX_OSSL_MUTEX_LOCK);
        err = sh_list_remove(list, sh_list_head(list), data);
    ix_ossl_mutex_unlock(list->lock);
#ifdef IX_DEBUG
   ix_ossl_message_log("Exiting sh_list_popFront()\n");
#endif 

    return err;
}

/**
 * NAME: sh_list_remove
 *
 * DESCRIPTION: Removes the DListElemt but does not destroy the user data
 *    Returns the pointer the the user data which got removed
 * 
 * @Param:  - IN -  list: Pointer to DList structure
 *          - IN -  element: Pointer to the element to be removed
 * @Param:  - OUT - data: Data removed
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.  
 *
 * Error Codes:
 * IX_CC_ERROR_NULL - invalid list/data
 * IX_CC_ERROR_OOM - memory allocation failure
 */

static ix_error sh_list_remove(DList *list, DListElmt *element, void **data) 
{
    ix_error   err;  /* Error code */

#ifdef IX_DEBUG
   ix_ossl_message_log("Entering sh_list_remove()\n");
#endif 

    if (list == NULL) 
    {
#ifdef IX_DEBUG
      ix_ossl_message_log("Invalid list\n");
#endif
      err = IX_ERROR_NEW(IX_CC_ERROR_NULL, IX_ERROR_LEVEL_WARNING);
      return err;
    }

    ix_ossl_mutex_lock(list->lock, IX_OSSL_MUTEX_LOCK);

    /* Do not allow a NULL element or removal from an empty list. */
    if (element == NULL || sh_list_size(list) == 0) 
    {
        ix_ossl_mutex_unlock(list->lock);
#ifdef IX_DEBUG
        ix_ossl_message_log("Invalid element\n");
#endif
        err = IX_ERROR_NEW(IX_CC_ERROR_NULL, IX_ERROR_LEVEL_WARNING);
        return err;
    }

    /* Remove the element from the list. */
    *data = element->data;

    if (element == list->head) {

        /* Handle removal from the head of the list. */
        list->head = element->next;

        if (list->head == NULL)
            list->tail = NULL;
        else
            element->next->prev = NULL;

    } else {

        /* Handle removal from other than the head of the list. */
        element->prev->next = element->next;

        if (element->next == NULL)
            list->tail = element->prev;
        else
            element->next->prev = element->prev;
    }

    /* Free the storage allocated by the abstract data type. */
    ix_ossl_free(element);

    /* Adjust the size of the list to account for the removed element. */
    list->size--;
    ix_ossl_mutex_unlock(list->lock);
#ifdef IX_DEBUG
   ix_ossl_message_log("Exiting sh_list_remove()\n");
#endif 

    return 0;

}

/**
 * NAME: sh_free_responseQueueData
 *
 * DESCRIPTION: Function that frees the response queue node.
 * 
 * @Param:  - IN -  respMsg: Pointer to the common response structure.
 * @Param:  - OUT - None.
 * 
 * @Return: None.
 *
 * Error Codes:
 *    None.
 */

static void sh_free_responseQueueData(SH_GeneralResponseMessage* respMsg)
{
   ix_uint32   count;   /* Local counter */

#ifdef IX_DEBUG
   ix_ossl_message_log("Entering sh_free_responseQueueData()\n");
#endif 

   for (count = 0; count < MAX_NUM_RESP; count++)
   {
      if(respMsg->outData[count].length != 0)
         if(respMsg->outData[count].responseData != NULL)
            ix_ossl_free(respMsg->outData[count].responseData);
   } 
   ix_ossl_free(respMsg);
#ifdef IX_DEBUG
   ix_ossl_message_log("Exiting sh_free_responseQueueData()\n");
#endif 

}
#endif
