/*****************************************************************************
*                  INTEL CORPORATION PROPRIETARY INFORMATION
*  This software is supplied under the terms of a license agreement or
*  nondisclosure agreement with Intel Corporation and may not be copied
*  or disclosed except in accordance with the terms of that agreement.
*  Copyright (c) 2002 Intel Corporation. All Rights Reserved.
*****************************************************************************/

/*****************************************************************************
* $Source: /cvs/constellations/ixa_3.0/src/cp_pdk/ForwardingPlane/FPModule/inc/qos_flow_events.h,v $
* $Revision: 1.6 $
* $Date: 2003/11/17 16:57:28 $
*
* Description:
*     This file defines the interface to the QoS Manager's Flow ID event 
*     notification mechanism. Other FPModule managers can register for 
*     events if they need some data from QoS Manager to update their own 
*     tables (for instance MPLS Manager needing flow IDs). In addition there
*     are also functions to query flow IDs.
*
*****************************************************************************/

#ifndef _QOS_FLOW_EVENTS_H
#define _QOS_FLOW_EVENTS_H

#include <stdint.h>
#include "npf_mpls_types.h"
#include "npf_common.h"


#define FEC_KEY 0
#define LABEL_KEY 1
#define DSCP_KEY 2

/*****************************************************************************
*    Name: QOS_FLOW_EVENT_CB
*
*    Description:
*        This is the signature for the flow id event callback. Any component
*        that wishes to register for flow id events must implement and
*        register a function matching this signature.
*
*    Parameters:
*        IN:  context    - The client context that was provided when the 
*                          callback was registered. The context is not
*                          modified in any way, it is simply passed back to
*                          the client.
*        IN:  entry_type - The type of QoS entry that was added with the flow.
*                          This will be either a FEC, DSCP or label depending
*                          on what type of QoS entry was added.
*        IN:  entry_key  - The QoS info that was added, either a FEC, DSCP or
*                          label.
*        IN:  flowid     - The flow id that has been assigned to the
*                          classification key in entry_key.
*        IN:  queueid    - The queue id that has been assigned to the
*                          classification key in entry_key.
*
*    Postconditions:
*        entry_key must not be deleted or modified
*
*    Returns:
*        Nothing
*
*/
typedef void (*QOS_FLOW_EVENT_CB)( uint32_t context, uint32_t entry_type,
        void *entry_key, uint32_t flowid, uint32_t queueid );

/*****************************************************************************
*    Name: RegisterPendingFlowCB
*
*    Description:
*        Allows a client to register a callback for flow id events.
*
*    Parameters:
*        IN:  context  - This value will be associated with the callback
*                        function and will be passed back to the client when
*                        the callback is invoked.
*        IN:  callback - The callback function to register.
*        OUT: cbhandle - An id generated by the QoS Manager that will be used
*                        by QoS Manager to identify the registered callback.
*                        This will be used by the client when the callback is
*                        deregistered.
*
*    Preconditions:
*        cbhandle != NULL
*
*    Returns:
*        0 if successful, -1 otherwise.
*
*/
extern int RegisterPendingFlowCB( uint32_t context, 
        QOS_FLOW_EVENT_CB callback, uint32_t *cbhandle );

/*****************************************************************************
*    Name: DeregisterPendingFlowCB
*
*    Description:
*        Allows a client to deregister its flow id event callback.
*
*    Parameters:
*        IN:  cbhandle - The callback handle that was returned to the client
*                        when the callback was registered.
*
*    Returns:
*        0 if succesful, -1 otherwise.
*
*/
extern int DeregisterPendingFlowCB( uint32_t cbhandle );

/*****************************************************************************
*    Name: QueryFlowByFEC
*
*    Description:
*        Allows an FPModule component to query the DPE Manager for a flowid
*        based on a FEC. If a flow id has already been associated with this
*        FEC it will be returned.
*
*    Parameters:
*        IN:  fec  - The FEC to look for.
*        OUT: flowid - The associated flow id if one is available. Only valid
*                      if function returns 0.
*
*    Preconditions:
*        fec != NULL
*        flowid != NULL
*
*    Returns:
*        0 if a valid flow id is associated with the FEC, -1 otherwise.
*
*/
extern int QueryFlowTupleByFEC( npf_mplsFec fec, uint32_t *flowid, 
        uint32_t *queueid );

/*****************************************************************************
*    Name: QueryFlowTupleByDSCP
*
*    Description:
*        Allows an FPModule component to query the DPE Manager for a flowid
*        based on a DSCP. If a flow id has already been associated with this
*        DSCP it will be returned.
*
*    Parameters:
*        IN:  dscp  - The DSCP to look for.
*        OUT: flowid - The associated flow id if one is available. Only valid
*                      if function returns 0.
*
*    Preconditions:
*        flowid != NULL
*
*    Returns:
*        0 if a valid flow id is associated with the DSCP, -1 otherwise.
*
*/
extern int QueryFlowTupleByDSCP( uint32_t DSCP, uint32_t *flowid, 
        uint32_t *queueid );

/*****************************************************************************
*    Name: QueryFlowTupleByLabel
*
*    Description:
*        Allows an FPModule component to query the DPE Manager for a flowid
*        based on a label. If a flow id has already been associated with this
*        label it will be returned.
*
*    Parameters:
*        IN:  label  - The label to look for.
*        OUT: flowid - The associated flow id if one is available. Only valid
*                      if function returns 0.
*
*    Preconditions:
*        flowid != NULL
*
*    Returns:
*        0 if a valid flow id is associated with the label, -1 otherwise.
*
*/
extern int QueryFlowTupleByLabel( npf_mplsLbl label, uint32_t *flowid,
        uint32_t *queueid );

#endif
