/*****************************************************************************
*                  INTEL CORPORATION PROPRIETARY INFORMATION
*  This software is supplied under the terms of a license agreement or
*  nondisclosure agreement with Intel Corporation and may not be copied
*  or disclosed except in accordance with the terms of that agreement.
*  Copyright (c) 2001 Intel Corporation. All Rights Reserved.
*****************************************************************************/

#ifndef _DPE_MANAGER_H
#define _DPE_MANAGER_H

#include <stdio.h>
#include <assert.h>
#include <signal.h>
#include "backendapi.h"
#include "pdk_macros.h"
#include "list.h"
#include "npf_ds_types.h"

#define DPE_BLOCK_ID             10
#define DPE_CLASSIFIER_TABLE_ID  DPE_BLOCK_ID + 1
#define DPE_MARKER_TABLE_ID      DPE_BLOCK_ID + 2
#define DPE_METER_TABLE_ID       DPE_BLOCK_ID + 3
#define DPE_ALG_DROPPER_TABLE_ID DPE_BLOCK_ID + 4
#define DPE_QUEUE_TABLE_ID       DPE_BLOCK_ID + 5
#define DPE_SHAPER_TABLE_ID      DPE_BLOCK_ID + 6
#define DPE_SCHEDULER_TABLE_ID   DPE_BLOCK_ID + 7
#define DPE_NULL_TABLE_ID        DPE_BLOCK_ID + 8

/* Current state of DPE table entry. VALID indicates that
   all parameters are ready for populating to the underlying
   NPU/data plane
*/
#define DPE_STATUS_INVALID       0
#define DPE_STATUS_VALID         1
#define DPE_STATUS_PROPAGATED    2  /* indicates that the parameters of this
				       DPE have been propagated to the 
				       data plane */
typedef struct _flow_tuple {
    uint32_t flow_id;
    uint32_t queue_id;
    uint32_t next_block_id;
} dpe_flow_tuple_t;

typedef struct _dpe_entry_id {
    NPF_DS_TableId_t table_id;
    NPF_DS_InstanceId_t instance_id;
} dpe_EntryId_t;

typedef unsigned long filter_handle_t;

typedef enum {
    dpe_Classifier,
    dpe_Marker,
    dpe_Meter,
    dpe_AlgDropper,
    dpe_Queue,
    dpe_Scheduler
} dpe_TableType_t;

typedef struct _clfr_elmnt {
    uint32_t             status; /* VALID, INVALID, PROPAGATED */
    NPF_DS_TableId_t     clfr_tbl_id;
    NPF_DS_InstanceId_t  clfr_instance_id;
    NPF_DS_BaseFilter_t* filter;
    NPF_DS_TableId_t     next_tbl_id;
    NPF_DS_InstanceId_t  next_instance_id;
    dpe_flow_tuple_t     flow_tuple;
    uint32_t             prev_tbl_count;
    DList                prev_tbl_entry_list; /* this contains TableId & InstID
						 of the previous QoS Block
					      */
    filter_handle_t      filter_hdl;         /* handle returned on adding a 
					      	new filter */ 
} dpe_clfr_t, *p_dpe_clfr_t;

typedef struct _marker_elmnt {
    uint32_t             status; /* VALID, INVALID */
    NPF_DS_TableId_t     marker_tbl_id;
    NPF_DS_InstanceId_t  marker_instance_id;
    NPF_DS_Action_t      action;
    NPF_DS_TableId_t     next_tbl_id;
    NPF_DS_InstanceId_t  next_instance_id;
    dpe_flow_tuple_t     flow_tuple;
    uint32_t             prev_tbl_count;
    DList                prev_tbl_entry_list; /* this contains TableId & InstID
						 of the previous QoS Block
					      */
} dpe_marker_t, *p_dpe_marker_t;

typedef struct _meter_elmnt {
    uint32_t             status; /* VALID, INVALID */
    NPF_DS_TableId_t     meter_tbl_id;
    NPF_DS_InstanceId_t  meter_instance_id;
    NPF_DS_Meter_t       meter;
    NPF_DS_TableId_t     next_tbl_id;
    NPF_DS_InstanceId_t  next_instance_id;
    dpe_flow_tuple_t     flow_tuple;    /* Upstream flow_tuple for Clfr, etc. */
    dpe_flow_tuple_t     downstream_flow_tuple; /* downstream flow tuple for
						   conforming traffic */
    dpe_flow_tuple_t     failnext_flow_tuple;
    uint32_t             prev_tbl_count;
    DList                prev_tbl_entry_list; /* this contains TableId & InstID
						 of the previous QoS Block
					      */
} dpe_meter_t, *p_dpe_meter_t;

typedef struct _AlgDropper_elmnt {
    uint32_t             status; /* VALID, INVALID */
    uint32_t             intf;
    NPF_DS_TableId_t     dropper_tbl_id;
    NPF_DS_InstanceId_t  dropper_instance_id;
    NPF_DS_AlgDrop_t     dropper;
    NPF_DS_TableId_t     next_tbl_id;
    NPF_DS_InstanceId_t  next_instance_id;
    dpe_flow_tuple_t     flow_tuple;
    uint32_t             prev_tbl_count;
    DList                prev_tbl_entry_list; /* this contains TableId & InstID
						 of the previous QoS Block
					      */
} dpe_dropper_t, *p_dpe_dropper_t;

typedef struct _queue_elmnt {
    uint32_t             status; /* VALID, INVALID */
    uint32_t             intf;
    NPF_DS_TableId_t     queue_tbl_id;
    NPF_DS_InstanceId_t  queue_instance_id;
    NPF_DS_Q_t           queue;
    NPF_DS_TableId_t     next_tbl_id;
    NPF_DS_InstanceId_t  next_instance_id;
    dpe_flow_tuple_t     flow_tuple;
    uint32_t             prev_tbl_count;
    DList                prev_tbl_entry_list; /* this contains TableId & InstID
						 of the previous QoS Block
					      */
} dpe_queue_t, *p_dpe_queue_t;

typedef struct _scheduler_elmnt {
    uint32_t             status; /* VALID, INVALID */
    uint32_t             intf;
    NPF_DS_TableId_t     sched_tbl_id;
    NPF_DS_InstanceId_t  sched_instance_id;
    NPF_DS_Scheduler_t   scheduler;
    NPF_DS_TableId_t     next_tbl_id;
    NPF_DS_InstanceId_t  next_instance_id;
    dpe_flow_tuple_t     flow_tuple;
    uint32_t             prev_tbl_count;
    DList                prev_tbl_entry_list; /* this contains TableId & InstID
						 of the previous QoS Block
					      */
} dpe_scheduler_t, *p_dpe_scheduler_t;


/* Initialization and shutdown routines, called by FPManager */
int fpm_dpe_mgr_init();
int fpm_dpe_mgr_shutdown();

int dpe_setup_besteffort_path ();
int dpe_get_besteffort_flow_tuple (uint32_t* flowid, uint32_t* queue_id);

int dpe_add_entry ( uint32_t            intf,
                    NPF_DS_TableId_t    table_id,
                    NPF_DS_InstanceId_t instance_id,
                    void*               entry,
                    NPF_DS_TableId_t    next_table_id,
                    NPF_DS_InstanceId_t next_instance_id );

int dpe_mod_entry ( uint32_t             intf,
		     NPF_DS_TableId_t     table_id,
		     NPF_DS_InstanceId_t  inst_id,
		     void*                dpe_entry,
		     NPF_DS_TableId_t     next_tbl_id,
		     NPF_DS_InstanceId_t  next_inst_id );

int dpe_propagate_params (uint32_t             intf, 
			  NPF_DS_TableId_t     table_id);

int dpe_propagate_status (NPF_DS_TableId_t     table_id,
			  NPF_DS_InstanceId_t  inst_id,
			  uint32_t             status);

int dpe_create_table_entry ( uint32_t             intf, 
			     NPF_DS_TableId_t     table_id,
			     NPF_DS_InstanceId_t  inst_id,
			     void*                dpe_entry,
			     NPF_DS_TableId_t     next_tbl_id,
			     NPF_DS_InstanceId_t  next_inst_id,
			     dpe_flow_tuple_t     flow_tuple,
			     uint32_t             status);

int dpe_delete_entry ( uint32_t             intf,	        
		       NPF_DS_TableId_t     table_id,
		       NPF_DS_InstanceId_t  inst_id);

int dpe_signal_mpls (dpe_clfr_t* clfr_entry);

int dpe_get_clfrFilter_params (dpe_clfr_t* clfr_entry,
			       uint32_t*   type,  /* OUT: Type: DSCP, FEC or 
						     Label */ 
			       void**      key); /* OUT: Dscp, npf_mplsFec, 
						   npf_mplsLbl - Value */

int dpe_get_downstream_flow_tuple (NPF_DS_TableId_t     next_tbl_id,
				   NPF_DS_InstanceId_t  next_inst_id,
				   dpe_flow_tuple_t*     flow_tuple);

int dpe_add_inst_to_prev_entry_list ( NPF_DS_TableId_t     table_id,
				      NPF_DS_InstanceId_t  inst_id,
				      NPF_DS_TableId_t     next_tbl_id,
				      NPF_DS_InstanceId_t  next_inst_id);

int dpe_delete_inst_from_prev_entry_list ( NPF_DS_TableId_t     table_id,
					   NPF_DS_InstanceId_t  inst_id,
					   NPF_DS_TableId_t     next_tbl_id,
					   NPF_DS_InstanceId_t  next_inst_id);

int dpe_update_flow_tuple (NPF_DS_TableId_t     table_id,
			   NPF_DS_InstanceId_t  inst_id,
			   NPF_DS_TableId_t     next_tbl_id,
			   NPF_DS_InstanceId_t  next_inst_id,
			   dpe_flow_tuple_t*    flow_tuple);

int dpe_propagate_flow_tuple (NPF_DS_TableId_t     table_id,
			      NPF_DS_InstanceId_t  inst_id);

int dpe_get_downstream_status (NPF_DS_TableId_t     table_id,
			       NPF_DS_InstanceId_t  inst_id,
			       uint32_t*            status);

void dpe_copy_sched_params (NPF_DS_Scheduler_t* sched_entry, /* destination */
			    NPF_DS_Scheduler_t* src_sched); /* source */

void dpe_copy_queue_params (NPF_DS_Q_t* queue_entry, /* destination */
			    NPF_DS_Q_t* ds_queue); /* source */

void dpe_copy_dropper_params (NPF_DS_AlgDrop_t* dropper_entry, /* destination */
			      NPF_DS_AlgDrop_t* ds_dropper); /* source */

void dpe_copy_marker_params (NPF_DS_Action_t* marker_entry, /* destination */
			     NPF_DS_Action_t* ds_marker); /* source */

void dpe_copy_meter_params (NPF_DS_Meter_t* meter_entry, /* destination */
			    NPF_DS_Meter_t* ds_meter); /* source */

void dpe_copy_clfr_params (NPF_DS_BaseFilter_t* clfr_entry, /* destination */
			   NPF_DS_BaseFilter_t* ds_clfr); /* source */
/*************************************************************************
 *                         GET STATISTICS ROUTINES                       *
 *************************************************************************/

int dpe_get_clfr_stats (NPF_DS_TableId_t     tbl_id,
			NPF_DS_InstanceId_t  inst_id,
			NPF_DS_ClfrStats_t* clfr_stats);

int dpe_get_marker_stats (NPF_DS_TableId_t     tbl_id,
			  NPF_DS_InstanceId_t  inst_id,
			  NPF_DS_MarkerStats_t* marker_stats);

int dpe_get_meter_stats (NPF_DS_TableId_t     tbl_id,
			 NPF_DS_InstanceId_t  inst_id,
			 NPF_DS_MeterStats_t* meter_stats);

int dpe_get_dropper_stats (NPF_DS_TableId_t     tbl_id,
			   NPF_DS_InstanceId_t  inst_id,
			   NPF_DS_DropperStats_t* q_stats);

int dpe_get_queue_stats (NPF_DS_TableId_t     tbl_id,
			 NPF_DS_InstanceId_t  inst_id,
			 NPF_DS_QueueStats_t* q_stats);

int dpe_get_scheduler_stats (NPF_DS_TableId_t         tbl_id,
			     NPF_DS_InstanceId_t      inst_id,
			     NPF_DS_SchedulerStats_t* sched_stats);

void dpe_dbg_print (dpe_TableType_t table_type,
		    void*         dpe_entry );

void dpe_dbg_print_tbl_entry (NPF_DS_TableId_t table_id, 
			      NPF_DS_InstanceId_t inst_id);

/* Function for getting QoS capabilitites, called by B&D */
//int fpm_dpe_getCaps (FPPI_FE_Caps* dpe_caps);

#endif /*_DPE_MANAGER_H_ */
