/*****************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 1998-2003 By Intel Corporation.  All rights reserved.  
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *****************************************************************************/

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include "ipaddr.h"

/* Most of the ip manipulation routines are in standard C,
   check ipaddr.h
 */

/*
   Display an ip address in readable format.
   input:  struct in_addr or npf_ipaddr
   output: buff = "xx.xx.xx.xx"
 */
int npf_ipaddr_addr2char(struct in_addr in, char *buff)
{
    sprintf(buff, "%s", inet_ntoa(in));
    return (0); 
}

/* Validate ipaddress - detect invalid or multicast ipaddress
 * input : long host order
 * output: -1 if first & last octet is 0 or first octet >= 224 (multicast)
 *          0 if success */
int npf_ipaddr_validate(uint32_t ipaddr, uint32_t submask) {

    /* failure if first and the last octet is 0 */
    if ((ipaddr>>24) == 0 || (ipaddr&0xFF) == 0)
        return -1;
    /* failure if first octet >= 224 (0xE0) */
    if ((ipaddr>>24) >= 0xE0)
        return -1;
#if 0
{
    uint32_t addr;
    uint32_t mask;

    /* copy from kernel's bad_mask */
    addr = htonl(ipaddr);
    mask = htonl(submask);

    if (addr & (mask = ~mask))
        return -1;
    if (submask & (submask+1))
        return -1;
}
#endif
    return (0);
}

int npf_ipaddr_prefixlen(uint32_t ipaddr) {

    int i, len = 0;
    for (i=0; i<32; i++) {
        if ( ((ipaddr << i)&0x80000000) == 0x80000000 ) 
            len ++;
    }

    return len;

}

/***************************************************************************
 *                              IPv6 Functions
 ***************************************************************************/

int npf_ipv6addr_validate(IPv6A addr, IPv6A submask) {
    return 0;
}

int npf_ipv6addr_compare(IPv6A addr1, IPv6A addr2) {
    if ((addr1[0] == addr2[0]) && (addr1[1] == addr2[1]) &&
        (addr1[2] == addr2[2]) && (addr1[3] == addr2[3]) )
        return 0;
    return -1;
}

int npf_ipv6addr_copy(IPv6A dest, IPv6A src) {
    memcpy(dest, src, sizeof(IPv6A));
    return 0;
}

int npf_ipv6addr_isZero(IPv6A addr) {
    if ( (!addr[0]) && (!addr[1]) && (!addr[2]) && (!addr[3]) )
        return 0;
    return -1;
}

/****************************************************************************
 * Function: npf_ipv6addr_swapByteOrder()
     *
 * ---------------------------                                              *
 * Parameters: v6Addr   - ptr to IPv6 address                               *
 *                                                                          *
 * Description: Swaps the byte order of a 16 byte IPv6 address. Note that   *
 *              the prototype for this method is not similar to the         *
 *              corresponding 32 bit system byte swapping methods           *
 *              (htonl ...). The input parameter is a pointer to a struct   *
 *              and the return type is void.                                *
 *                                                                          *
 * Return: None                                                             *
 *                                                                          *
 ****************************************************************************/

void npf_ipv6addr_swapByteOrder(char *addr) {

    unsigned long i, tmp, one, two;
    unsigned long limit = 4;

    for (i = 0; i < limit; i++)
    {
        tmp = i * 4;
        one = addr[tmp];
        two = addr[tmp + 1];
        addr[tmp]     = addr[tmp + 3];
        addr[tmp + 3] = one;
        addr[tmp + 1] = addr[tmp + 2];
        addr[tmp + 2] = two;
    }
}

/* Calculate IPv6 prefix length. Assume the passing address fileds 
 * (unsigned 32-bit int) are in network byte order.
 */
int PrefixLen(IPv6A v6mask)
{
    int len = 0;
    int i, j;

    for (i = 0; i < 4; i++) {
	for ( j = 0; j < 32; j ++ ) {
	    if ( ( (ntohl(v6mask[i]) << j)&0x80000000 ) == 0x80000000 ) 
		len ++;
	    else
		return len;
	}
    }
    return len;	
}

/* Calculate IPv6 prefix length. */
int Equalv6Addrs(IPv6A addr1, IPv6A addr2)
{
    int i;

    for (i = 0; i < 4; i++) {
	if ( addr1[i] != addr2[i] )
	    return FALSE;
    }
    return TRUE;	
}

#ifdef TEST
int main() {

    IPADDR ipa;
    char sip[16];
    int i = EL2HLT;

    inet_aton("224.2.3.1", &ipa);
    printf("network order ipa.s_addr = %.2X\n", ipa.s_addr);
    npf_ipaddr_addr2char(ipa, sip);
    printf ("npf_ipaddr_addr2char -> sip = %s\n", sip);
    if (npf_ipaddr_validate(ntohl(ipa.s_addr)) == -1)
        printf("Validation failure\n");

    return (0);

}
#endif	/* TEST */
