/*****************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 1998-2003 By Intel Corporation.  All rights reserved.  
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *****************************************************************************/

/*****************************************************************************
*                                                                            *
*  -------------------------------- list.c --------------------------------  *
*                                                                            *
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "list.h"
#include "log.h"

/*****************************************************************************
*                                                                            *
*  --------------------------- Public Interface ---------------------------  *
*                                                                            *
*****************************************************************************/

/* ----------------------------- npf_list_init ---------------------------- */ 

void npf_list_init(DList *list, void (*destroy)(void *data)) {

    //Initialize the list.

    list->size = 0;
    list->destroy = destroy;
    list->compare = NULL;
    list->head = NULL;
    list->tail = NULL;
    PIL_InitializeCriticalSection(&list->lock);

    return;

}

/* --------------------------- npf_list_destroy --------------------------- */
/*
 *    This delete all the list element and DESTROY/FREE the user data
 *    using the destroy function defined by user which is passed in at
 *    the list initialization time
 */

void npf_list_destroy(DList *list) {

    void *data;

    if (list==NULL) return;

    //Remove each element.

    while (npf_list_size(list) > 0) {

       if (npf_list_remove(list, npf_list_tail(list), (void **)&data) == 0 && 
           list->destroy != NULL) {

            // Call a user-defined function to free dynamically allocated data. 
            list->destroy(data);
       }
    }

    PIL_DestroyCriticalSection(&list->lock);

    // No operations are allowed now, but clear the structure as a precaution.
    memset(list, 0, sizeof(DList));

    return;

}

/* -------------------------- npf_list_pushFront -------------------------- */

int npf_list_pushFront(DList *list, const void *data) {

    int ret;
    assert(list); //assert(data);
    if (data == NULL) return -1;

    // Same thread can enter PIL critical section mutiple times w/o blocking
    PIL_EnterCriticalSection(&list->lock);
        ret = npf_list_insPrev(list, npf_list_head(list), data);
    PIL_LeaveCriticalSection(&list->lock);

    return ret;

}

/* -------------------------- npf_list_pushBack --------------------------- */

int npf_list_pushBack(DList *list, const void *data) {

    int ret;
    assert(list); assert(data);

    // Same thread can enter PIL critical section mutiple times w/o blocking
    PIL_EnterCriticalSection(&list->lock);
        ret = npf_list_insNext(list, npf_list_tail(list), data);
    PIL_LeaveCriticalSection(&list->lock);

    return ret;

}

/* -------------------------- npf_list_popFront --------------------------- */

int npf_list_popFront(DList *list, void **data) {

    int ret;
    assert(list);

    // Same thread can enter PIL critical section mutiple times w/o blocking
    PIL_EnterCriticalSection(&list->lock);
        ret = npf_list_remove(list, npf_list_head(list), data);
    PIL_LeaveCriticalSection(&list->lock);

    return ret;

}

/* -------------------------- npf_list_popBack ---------------------------- */

int npf_list_popBack(DList *list, void **data) {

    int ret;
    assert(list);

    // Same thread can enter PIL critical section mutiple times w/o blocking
    PIL_EnterCriticalSection(&list->lock);
        ret = npf_list_remove(list, npf_list_tail(list), data);
    PIL_LeaveCriticalSection(&list->lock);

    return ret;

}

/* ------------------------ npf_list_popFrontFree ------------------------- */

int npf_list_popFrontFree(DList *list) {

    int ret;
    assert(list);

    // Same thread can enter PIL critical section mutiple times w/o blocking
    PIL_EnterCriticalSection(&list->lock);
        ret = npf_list_removeFree(list, npf_list_head(list));
    PIL_LeaveCriticalSection(&list->lock);

    return ret;

}

/* ------------------------ npf_list_popBackFree -------------------------- */

int npf_list_popBackFree(DList *list) {

    int ret;
    assert(list);

    // Same thread can enter PIL critical section mutiple times w/o blocking
    PIL_EnterCriticalSection(&list->lock);
        ret = npf_list_removeFree(list, npf_list_tail(list));
    PIL_LeaveCriticalSection(&list->lock);

    return ret;

}

/* --------------------------- npf_list_isExist --------------------------- */

int npf_list_isExist(DList *list, const void *data) {
    printf("npf_list_isExist not done yet\n");
    return 0;
}

/* ----------------------- npf_list_getFirstData -------------------------- */

void *npf_list_getFirstData(DList *list) {
    
    void *data;
    // lock to avoid critical section between get head and get data
    PIL_EnterCriticalSection(&list->lock);
        if (npf_list_head(list) != NULL)
            data = npf_list_data(npf_list_head(list));
        else
            data = NULL;
    PIL_LeaveCriticalSection(&list->lock);
    return data;
    
}

/* ------------------------ npf_list_getLastData -------------------------- */

void *npf_list_getLastData(DList *list) {

    void *data;
    // lock to avoid critical section between get tail and get data
    PIL_EnterCriticalSection(&list->lock);
        if (npf_list_tail(list) != NULL)
            data = npf_list_data(npf_list_tail(list));
        else
            data = NULL;
    PIL_LeaveCriticalSection(&list->lock);
    return data;

}

/* ------------------------------------------------------------------------
 
                              ITERATOR OPERATIONS

    User needs define a iterator pointer and use ItrCreate() to allocate
    an iterator memory.  After using iterator, user responses for calling
    ItrDelete to free the iterator memory.
    See the example in the header file
    For thread safety, iterator operation should be protected in the user 
    defined LOCK
 
   ------------------------------------------------------------------------ */

/* --------------------------- npf_list_itrCreate ------------------------- */
/*
 *    Return NULL if list is empty
 */
DListItr *npf_list_itrCreate(DList *list) {
    DListItr *new_itr;
    assert(list);
    if (npf_list_size(list) == 0)
        return NULL;
    else {
        // Allocate storage for iterator
	if ((new_itr = (DListItr *)PIL_AllocateMemory
	                   (sizeof(DListItr), PIL_MEM_ZERO_INIT)) == NULL)
            return NULL;
        new_itr->list = list;
        new_itr->current = (DListElmt *)PTR_UNDEFINED;
    }
    return new_itr;
}


/* --------------------------- npf_list_itrDelete ------------------------- */
int npf_list_itrDelete(DListItr *itr) {
    // assert(itr); itrCreate return NULL and it may go itrDelete directly
    if (itr == NULL) return 0;
    memset(itr, 0, sizeof(DListItr));
    PIL_FreeMemory(itr);
    return 0;
}
    


/* --------------------------- npf_list_itrFirst -------------------------- */
/*
 *    Return -1 if list is empty, and undefined the iterator
 *              if itr is not defined
 *    Return 0 if success
 */
int npf_list_itrFirst(DListItr *itr) {
    //assert(itr);
    if (itr == NULL) return -1;
    if (npf_list_size(itr->list) == 0) {
        itr->current = (DListElmt *)PTR_UNDEFINED;
        return -1;
    } else {
        itr->current = npf_list_head(itr->list);
        return 0;
    }
}

/* --------------------------- npf_list_itrLast --------------------------- */
/*
 *    Return -1 if list is emplty and undefined the iterator.
 *    Return 0 if success
 */
int npf_list_itrLast(DListItr *itr) {
    assert(itr);
    if (npf_list_size(itr->list) == 0) {
        itr->current = (DListElmt *)PTR_UNDEFINED;
        return -1;
    } else {
        itr->current = npf_list_tail(itr->list);
        return 0;
    }
}

/* --------------------------- npf_list_itrNext --------------------------- */
/*
 *    Return -1 if itr->current is undefined; if next is NULL
 */
int npf_list_itrNext(DListItr *itr) {
    void *undef = (void *)PTR_UNDEFINED;
    assert(itr); assert(itr->current);
    if (itr->current == (DListElmt *)undef)
        return -1;
    if (npf_list_isTail(itr->current)) {
        itr->current = (DListElmt *)PTR_UNDEFINED;
        return -1;
    } else {
        itr->current = npf_list_next(itr->current);
        return 0;
    }
}

/* --------------------------- npf_list_itrPrev --------------------------- */
/*
 *    Return -1 if itr->current is undefined; if prev is NULL
 */
int npf_list_itrPrev(DListItr *itr) {
    void *undef = (void *)PTR_UNDEFINED;
    assert(itr); assert(itr->current);
    if (itr->current == (DListElmt *)undef)
        return -1;
    if (npf_list_isHead(itr->current)) {
        itr->current = (DListElmt *)PTR_UNDEFINED;
        return -1;
    } else {
        itr->current = npf_list_prev(itr->current);
        return 0;
    }
}

/* ------------------------- npf_list_itrInsNext -------------------------- */
/*
 *    Return -1 if input itr is undefined, or failed and set curret undefined
 *    Return 0 if success and set the itr to the one after the new inserted
 */

int npf_list_itrInsNext(DListItr *itr, const void *data) {
    void *undef = (void *)PTR_UNDEFINED;
    assert(itr); assert(itr->list); assert(itr->current); assert(data);
    if (itr->current == (DListElmt *)undef)
        return -1;
    if (npf_list_insNext(itr->list, itr->current, data) == -1) {
        itr->current = (DListElmt *)PTR_UNDEFINED;
        return -1;
    }
    // set the itr to the one after to the new inserted
    if (npf_list_isTail(itr->current->next))
        itr->current = (DListElmt *)undef;
    else
        itr->current = itr->current->next->next;

    return 0;
}

/* ------------------------- npf_list_itrInsPrev -------------------------- */
/*
 *    Return -1 if input itr is undefined, or fail and set current undefined
 *    Return 0 if success and remain the itr to the input one (one after new
 *    inserted)
 */

int npf_list_itrInsPrev(DListItr *itr, const void *data) {
    void *undef = (void *)PTR_UNDEFINED;
    assert(itr); assert(itr->list); assert(itr->current); assert(data);
    if (itr->current == (DListElmt *)undef)
        return -1;
    if (npf_list_insPrev(itr->list, itr->current, data) == -1) {
        itr->current = (DListElmt *)PTR_UNDEFINED;
        return -1;
    }
    return 0;
}

/* -------------------------- npf_list_itrRemove -------------------------- */
/*
 *    Return -1 if input itr is undefined, or fail and set current undefined
 *    Return 0 if success and set the itr to the one next to removed one,
 *    set undefined if removed one is the last one
 */

int npf_list_itrRemove(DListItr *itr, void **data) {
    DListElmt *nextToRemoved;
    void *undef = (void *)PTR_UNDEFINED;
    assert(itr); assert(itr->list); assert(itr->current);
    if (itr->current == (DListElmt *)undef)
        return -1;
    nextToRemoved = itr->current->next;
    if (npf_list_remove(itr->list, itr->current, data) == -1) {
        itr->current = (DListElmt *)PTR_UNDEFINED;
        return -1;
    }
    if (nextToRemoved == NULL)
        itr->current = (DListElmt *)undef;
    else
        itr->current = nextToRemoved;
    return 0;
}

/* ------------------------ npf_list_itrRemoveFree ------------------------ */
/*
 *    Same as npf_list_itrRemove but it DESTROY/FREE the user data
 *    by using user defined free function
 */

int npf_list_itrRemoveFree(DListItr *itr) {
    void *data;
    void *undef = (void *)PTR_UNDEFINED;
    assert(itr); assert(itr->list); assert(itr->current);
    if (itr->current == (DListElmt *)undef)
        return -1;
    if (npf_list_itrRemove(itr, (void **)&data) == 0) {
        if (itr->list->destroy != NULL) {
            // Call a user-defined function to free dynamically allocated data. 
            itr->list->destroy(data);
        }
        return 0;
    } else {
        itr->current = (DListElmt *)PTR_UNDEFINED;
        return -1;
    }
}

/* -------------------------- npf_list_itrGetData ------------------------- */
/*
 *    Return the data pointer which points to the user data
 *    Return NULL if itr->current is undefined
 */
void *npf_list_itrGetData(DListItr *itr) {
    void *undef = (void *)PTR_UNDEFINED;
    assert(itr); assert(itr->current);
    if (itr->current == (DListElmt *)undef)
        return NULL;
    else
        return itr->current->data;
}

/* ------------------------------------------------------------------------
 
                                MAP OPERATIONS

    User should provide their own list->compare function. Return 1 if the
    input void *key matches the void *data->key. The compare algorithm is
    up to user's implementation
    list->compare(const void *data, const void *key)
 
   ------------------------------------------------------------------------ */

/* ------------------------ npf_list_removeByKey -------------------------- */
/*
 *    Search the list element by Key, remove the list element and output
 *    the data. 0 if success, -1 otherwise
 */
int npf_list_removeByKey(DList *list, const void *key, void **data) {

    void     *tmpdata;
    DListItr *itr;
    int      ret;

    tmpdata = NULL;
    ret = -1;

    if (list->compare == NULL) return -1;

    PIL_EnterCriticalSection(&list->lock);
        if ((itr = npf_list_itrCreate(list)) == NULL) {
            PIL_LeaveCriticalSection(&list->lock);
            return -1;
        }
        npf_list_itrFirst(itr);
        do {
            tmpdata = npf_list_itrGetData(itr);
            if (list->compare(tmpdata, key) == 1) {
                ret = npf_list_itrRemove(itr, data);
                break;
            }
        } while (npf_list_itrNext(itr) != -1);
        npf_list_itrDelete(itr);
    PIL_LeaveCriticalSection(&list->lock);
    return ret;
}

/* ---------------------- npf_list_removeByKeyFree ------------------------ */
/*
 *    Same as npf_list_removeKey but it free the data automatically by
 *    calling user provided destroy function
 */
int npf_list_removeByKeyFree(DList *list, const void *key) {

    void     *data;
    DListItr *itr;
    int      ret;

    data = NULL;
    ret = -1;

    if (list->compare == NULL) return -1;

    PIL_EnterCriticalSection(&list->lock);
        if ((itr = npf_list_itrCreate(list)) == NULL) {
            PIL_LeaveCriticalSection(&list->lock);
            return -1;
        }
        npf_list_itrFirst(itr);
        do {
            data = npf_list_itrGetData(itr);
            if (list->compare(data, key) == 1) {
                ret = npf_list_itrRemoveFree(itr);
                break;
            }
        } while (npf_list_itrNext(itr) != -1);
        npf_list_itrDelete(itr);
    PIL_LeaveCriticalSection(&list->lock);
    return ret;
}

/* ------------------------ npf_list_searchByKey -------------------------- */
/*
 *    User should define list->compare(const void *data, const void *key)
 *    key is one field inside the data, return 1 if the key inside data
 *    matches the pass in key
 */

void *npf_list_searchByKey(DList *list, const void *key) {
    
    void     *data, *retval;
    DListItr *itr;

    retval = NULL;

    if (list->compare == NULL) return NULL;

    PIL_EnterCriticalSection(&list->lock);
        if ((itr = npf_list_itrCreate(list)) == NULL) {
            PIL_LeaveCriticalSection(&list->lock);
            return NULL;
        }
        npf_list_itrFirst(itr);
        do {
            data = npf_list_itrGetData(itr);
            if (list->compare(data, key) == 1) {
                retval = data;
                break;
            }
        } while (npf_list_itrNext(itr) != -1);
        npf_list_itrDelete(itr);
    PIL_LeaveCriticalSection(&list->lock);
    return retval;
    
}


/*****************************************************************************
*                                                                            *
*  --------------------------- Private Interface --------------------------  *
*                                                                            *
*****************************************************************************/

/* ---------------------------- npf_list_insNext -------------------------- */

int npf_list_insNext(DList *list, DListElmt *element, const void *data) {

    DListElmt *new_element;

    // element maybe NULL when list is empty
    assert(list); assert(data);

    // Allocate storage for the element. Do it now to save lock time

    if ((new_element = (DListElmt *)PIL_AllocateMemory
                   (sizeof(DListElmt), PIL_MEM_ZERO_INIT)) == NULL)
        return -1;

    PIL_EnterCriticalSection(&list->lock);

    // Do not allow a NULL element unless the list is empty.

    if (element == NULL && npf_list_size(list) != 0) {
        PIL_LeaveCriticalSection(&list->lock);
        PIL_FreeMemory(new_element);
        return -1;
    }

    // Insert the new element into the list.
    new_element->data = (void *)data;

    if (npf_list_size(list) == 0) {

        // Handle insertion when the list is empty.
        list->head = new_element;
        list->head->prev = NULL;
        list->head->next = NULL;
        list->tail = new_element;

    } else {

        // Handle insertion when the list is not empty.
        new_element->next = element->next;
        new_element->prev = element;

        if (element->next == NULL)
            list->tail = new_element;
        else
            element->next->prev = new_element;

        element->next = new_element;

    }

    // Adjust the size of the list to account for the inserted element.

    list->size++;
    PIL_LeaveCriticalSection(&list->lock);

    return 0;

}

/* ---------------------------- npf_list_insPrev -------------------------- */

int npf_list_insPrev(DList *list, DListElmt *element, const void *data) {

    DListElmt          *new_element;

    // element maybe NULL when list is empty
    assert(list); assert(data);

    // Allocate storage for the element. Do it now to save lock time
    if ((new_element = (DListElmt *)PIL_AllocateMemory
                   (sizeof(DListElmt), PIL_MEM_ZERO_INIT)) == NULL)
        return -1;

    PIL_EnterCriticalSection(&list->lock);

    // Do not allow a NULL element unless the list is empty.
    if (element == NULL && npf_list_size(list) != 0) {
        PIL_LeaveCriticalSection(&list->lock);
        PIL_FreeMemory(new_element);
        return -1;
    }

    // Insert the new element into the list.
    new_element->data = (void *)data;

    if (npf_list_size(list) == 0) {

        // Handle insertion when the list is empty.
        list->head = new_element;
        list->head->prev = NULL;
        list->head->next = NULL;
        list->tail = new_element;

     } else {

        // Handle insertion when the list is not empty.
        new_element->next = element; 
        new_element->prev = element->prev;

        if (element->prev == NULL)
            list->head = new_element;
        else
            element->prev->next = new_element;

        element->prev = new_element;

    }

    // Adjust the size of the list to account for the new element.

    list->size++;
    PIL_LeaveCriticalSection(&list->lock);

    return 0;

}

/* ---------------------------- npf_list_remove --------------------------- */
/*
 *    Remove the DListElemt but does not destroy the user data
 *    Return the pointer the the user data which got removed
 */

int npf_list_remove(DList *list, DListElmt *element, void **data) {

    assert(list);

    PIL_EnterCriticalSection(&list->lock);

    // Do not allow a NULL element or removal from an empty list.
    if (element == NULL || npf_list_size(list) == 0) {
        PIL_LeaveCriticalSection(&list->lock);
        return -1;
    }

    // Remove the element from the list.
    *data = element->data;

    if (element == list->head) {

        // Handle removal from the head of the list.
        list->head = element->next;

        if (list->head == NULL)
            list->tail = NULL;
        else
            element->next->prev = NULL;

    } else {

        // Handle removal from other than the head of the list.
        element->prev->next = element->next;

        if (element->next == NULL)
            list->tail = element->prev;
        else
            element->next->prev = element->prev;

    }

    // Free the storage allocated by the abstract data type.
    PIL_FreeMemory(element);

    // Adjust the size of the list to account for the removed element.
    list->size--;
    PIL_LeaveCriticalSection(&list->lock);

    return 0;

}

/* -------------------------- npf_list_removeFree ------------------------- */
/*
 *    Remove the DListElemt and DESTROY/FREE the user data using user
 *    defined destroy function passed at the list initialization time
 */

int npf_list_removeFree(DList *list, DListElmt *element) {

    void *data;

    assert(list);

    if (npf_list_remove(list, element, (void **)&data) == 0) {
        if (list->destroy != NULL) {
            // Call a user-defined function to free dynamically allocated data. 
            list->destroy(data);
        }
    } else 
        return -1;

    return 0;

}

#ifdef TESTLIST

/*****************************************************************************
*                                                                            *
*  Test on multi-thread to prove this is the multi-thread safety list        *
*                                                                            *
*****************************************************************************/

#define SIZEIT 100	// Q: Failed if size > 1019, why?
#include <errno.h>
#include <sys/poll.h>

pthread_t threads[SIZEIT];
DList list;

int foo(int in) {

    int *data, i;
    DListItr *itr;

    for (i = 0; i < 10; i++) {
        if ((data = (int*)malloc(sizeof(int))) == NULL)
            return -ENOMEM;
        *data = i;

        if (npf_list_pushFront(&list, data) != 0)
            return -1;
        npf_logger_write(LOG_ALL, LOG_DEBUG, "Thread %d - %dth push %d Size: %d", 
                                  in, i, *data, npf_list_size(&list));
        poll(0,0,10);

        itr = npf_list_itrCreate(&list);
        // LOCK
        npf_list_itrFirst(itr);
        npf_logger_write(LOG_ALL, LOG_DEBUG, "Thread %d - %dth check %d",
                                  in, i, *((int *)npf_list_itrGetData(itr)));
        // UNLOCK
        npf_list_itrDelete(itr);

        if (npf_list_popBack(&list, (void **)&data) != 0)
            return -1;
        npf_logger_write(LOG_ALL, LOG_DEBUG, "Thread %d - %dth pop  %d Size: %d", 
                                  in, i, *data, npf_list_size(&list));
        list.destroy(data);
        poll(0,0,10);

    }
    return 0;
}

int main() {

    int i;
    void *status;

    npf_logger_open("/dev/null");
    npf_logger_setdisplay(1);
    npf_logger_write(LOG_ALL, LOG_DEBUG, "Test List on Multi-thread program!");

    // free() is the function to call at the destroy time
    npf_list_init(&list, free);

    for ( i = 0; i < SIZEIT; i++) {
        printf("Creating Thread %d\n", i);
        pthread_create(&threads[i], NULL, (void *(*)(void *))foo, (void*)i);
    }

    for ( i = 0; i < SIZEIT; i++)
        pthread_join(threads[i], &status);

    // Make sure Delete and Free the list before exit
    npf_list_destroy(&list);

    return 0;
}
#endif /* TESTLIST */
