/*****************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 1998-2003 By Intel Corporation.  All rights reserved.  
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *****************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>
#include <time.h>
#include <stdarg.h>
#include <string.h>

#include "pil.h"
#include "list.h"
#include "log.h"

// This array is used to translate the enum value to a string for output in
// the log message prefix.
// NOTE THIS MUST BE UPDATED IF THE ENUM IS CHANGED
// update this if npf_logger_level in log.h modified
static char *npf_logger_level_names[LOG_LEVEL_MAX+1] = { "UNUSED", "TRACE", "INFO", "UNUSED", "ERR",
 "UNUSED", "UNUSED", "UNUSED", "CRIT" };

// This array is used to translate the enum value to a string for output in
// the log message prefix.
// NOTE THIS MUST BE UPDATED IF THE ENUM IS CHANGED
// the log message prefix.
// NOTE THIS MUST BE UPDATED IF THE ENUM IS CHANGED
//update this if npf_logger_component in log.h modified
static char *npf_logger_comp_names[LOG_MAX_COMPS] = { "PDKMGR", "NS", "LBM", "TPM", "CM", "IF", "BD" , "IPV4", "IPV6", "IPV6TM", "DIFFSERV", "INTSERV", "MPLS", "TOPO", "CB", "VIC", "PH", "SH", "APP", "TPI", "DRS", "GTP" , "FP_GTP", "FP_QOS", "FP_MPLS", "FP_CM","FP_ATM", "FP_IPV4", "FP_IPV6", "FP_IPV6TM", "FP_BOOT", "FP_BD","IXA_STUBS", "CNF_TST", "SQE_STUBS", "ATM" };

typedef struct {
    uint32_t level;
} log_info;

typedef struct _logger {

    PilCriticalSection lock;
    char *filename;
    char *scratch;
    FILE *file;
    npf_logger_verbosity verbosity;
    log_info component_info[LOG_MAX_COMPS];

} npf_logger;

/*
 * list of loggers
 */
//static npf_logger Logger;

PilCriticalSection LoggerLock;
DList LoggerList;
int LoggerRefCount = 0;

/*
   Backwards compatibility hack
 */

void * CPLoggerHandle = NULL;
void * FPLoggerHandle = NULL;

int npf_logger_compare(const void *a, const void *b)
{
    return (a==b);
}

/*
   initialize PIL lock & logger list
 */
NPF_RET npf_logger_Init()
{
    if (!LoggerRefCount)
    {
        PIL_Init();
        PIL_InitializeCriticalSection(&LoggerLock);
        npf_list_init(&LoggerList,NULL);

        LoggerList.compare = npf_logger_compare;
    }

    LoggerRefCount++;

    return NPF_SUCCESS;
}

/*
   create and initialize logger instance
 */
NPF_RET npf_logger_StartEx(void **handle, char *filename) 
{
    npf_logger *Logger;
    FILE *fp = NULL;

    if (!handle)
    {
        return NPF_INVALID_PARAMETERS;
    }
    *handle = NULL;
    PIL_EnterCriticalSection(&LoggerLock);

    Logger = (npf_logger*)PIL_AllocateMemory(sizeof(npf_logger),0);

    if (!Logger)
    {
        return NPF_OUT_OF_MEMORY;
    }

    PIL_InitializeCriticalSection(&Logger->lock);
    PIL_EnterCriticalSection(&Logger->lock);
    if ( strcmp( filename, "stdout" ) == 0 ) {
        Logger->file = stdout;
    }
    else if ( strcmp( filename, "stderr" ) == 0 ) {
        Logger->file = stderr;
    }
    else if ( (fp = fopen(filename, "a")) != NULL ) {
        Logger->filename = filename;
        Logger->file = fp;
    } else {
        /* someone is writing? not possible cuz we're in CS */
    }
    Logger->verbosity = LOG_VERBOSITY;

    // Initialize all component log levels

    Logger->component_info[LOG_PDKMGR].level = LOG_PDKMGR_LVL;
    Logger->component_info[LOG_NS].level = LOG_NS_LVL;
    Logger->component_info[LOG_LBM].level = LOG_LBM_LVL;
    Logger->component_info[LOG_TPM].level = LOG_TPM_LVL;
    Logger->component_info[LOG_CM].level = LOG_CM_LVL;
    Logger->component_info[LOG_IF].level = LOG_IF_LVL;
    Logger->component_info[LOG_BD].level = LOG_BD_LVL;
    Logger->component_info[LOG_IPV4].level = LOG_IPV4_LVL;
    Logger->component_info[LOG_DS].level = LOG_DS_LVL;
    Logger->component_info[LOG_INTSERV].level = LOG_INTSERV_LVL;
    Logger->component_info[LOG_MPLS].level = LOG_MPLS_LVL;
    Logger->component_info[LOG_TOPO].level = LOG_TOPO_LVL;
    Logger->component_info[LOG_CB].level = LOG_CB_LVL;
    Logger->component_info[LOG_VIC].level = LOG_VIC_LVL;
    Logger->component_info[LOG_PH].level = LOG_PH_LVL;
    Logger->component_info[LOG_APP].level = LOG_APP_LVL;
    Logger->component_info[LOG_TPI].level = LOG_TPI_LVL;
    Logger->component_info[LOG_DRS].level = LOG_DRS_LVL;
    Logger->component_info[LOG_GTP].level = LOG_GTP_LVL;
    Logger->component_info[LOG_FP_QOS].level = LOG_FP_QOS_LVL;
    Logger->component_info[LOG_FP_MPLS].level = LOG_FP_MPLS_LVL;
    Logger->component_info[LOG_CNF_TST].level = LOG_CNF_TST_LVL;
    Logger->component_info[LOG_IXA_STUBS].level = LOG_STUBS_LVL;

    Logger->scratch = (char*)PIL_AllocateMemory(sizeof(char)*1024,0);

    PIL_LeaveCriticalSection(&Logger->lock);

    *handle = (void*)Logger;
    npf_list_pushFront(&LoggerList,handle);

    PIL_LeaveCriticalSection(&LoggerLock);

    return NPF_SUCCESS;
}

NPF_RET npf_logger_StopEx(void *handle)
{
    npf_logger *tmpLogger;
    npf_logger *Logger = (npf_logger*)handle;

    if (!handle)
    {
        return NPF_INVALID_PARAMETERS;
    }

    PIL_EnterCriticalSection(&Logger->lock);

    if (Logger->file) {
        fflush(Logger->file); 
        fclose(Logger->file);
        Logger->file = 0;
    } 

    PIL_FreeMemory(Logger->scratch);

    PIL_LeaveCriticalSection(&Logger->lock);
    PIL_DestroyCriticalSection(&Logger->lock);

    npf_list_removeByKey(&LoggerList,handle,(void **)&tmpLogger);

    PIL_FreeMemory(handle);

    return NPF_SUCCESS;
}

NPF_RET npf_logger_SetLevelEx( void *handle, npf_logger_component comp, uint32_t level)
{
    npf_logger *Logger = (npf_logger*)handle;
    if (!handle)
    {
        return NPF_INVALID_PARAMETERS;
    }

    if ( comp < 0 || comp >= LOG_MAX_COMPS ) {
        return NPF_INVALID_PARAMETERS;
    }
    if ( level >> LOG_NUM_LEVELS ) {
        return NPF_INVALID_PARAMETERS;
    }
    PIL_EnterCriticalSection(&Logger->lock);
    Logger->component_info[comp].level = level;
    PIL_LeaveCriticalSection(&Logger->lock);
    return NPF_SUCCESS;
}

NPF_RET npf_logger_SetVerbosityEx( void *handle, npf_logger_verbosity verb )
{
    npf_logger *Logger = (npf_logger*)handle;
    if (!handle)
    {
        return NPF_INVALID_PARAMETERS;
    }

    if ( verb < 0 || verb > LOG_VERB_MAX ) {
        return NPF_INVALID_PARAMETERS;
    }
    PIL_EnterCriticalSection(&Logger->lock);
    Logger->verbosity = verb;
    PIL_LeaveCriticalSection(&Logger->lock);
    return NPF_SUCCESS;
}

/* This could be replaced by C standard predefined macros
   __DATE__ & __TIME__, except the 0.001 second
 */
char *get_now_time( void ) {
    time_t secs;
    unsigned long long millis;
    struct tm *tmp;
    static char time_string[32];

    millis = PIL_GetCurrentTime();
    secs = (time_t)millis / 1000;
    millis = millis % 1000;

    tmp = localtime(&secs);

    strftime(time_string, sizeof(time_string), "%m/%d %H:%M:%S", tmp);
    sprintf(&time_string[strlen(time_string)], ".%03d", (int)millis);
    return(time_string);
} /* end get_now_time */

/*
   Never use it, use npf_logger_write(component, level, format, ...) instead 
   Write the message to log file and standard output if display is set, but
   time stamp doesn not show on standard output
*/
void npf_logger_logEx(void *handle, char *file,
    const char *func, int line, npf_logger_component comp, 
    npf_logger_level level, char *format, ...)
{

    npf_logger *Logger = (npf_logger*)handle;

    va_list args;
    char *prefix_ptr;
    int prefix_len = 0;

    if (!handle)
    {
        return;
    }

    if (!Logger || !Logger->file) {        // check if log file exists
        return;
    }

    PIL_EnterCriticalSection(&Logger->lock);
    if ( Logger->component_info[comp].level & level ) {
        va_start(args, format);
	Logger->scratch[0] = 0;
        /*bzero( prefix, 512 );*/
        prefix_ptr = Logger->scratch;
        switch ( Logger->verbosity ) {
            case LOG_VERB_TIME :
                sprintf( prefix_ptr, "%s:", get_now_time() );
                prefix_len = strlen( prefix_ptr );
                prefix_ptr += prefix_len;
            case LOG_VERB_PID :
                /*sprintf( prefix_ptr, "%d:", getpid() ); */
                sprintf( prefix_ptr, "%d:", (int)PIL_GetCurrentThread() );
                prefix_len = strlen( prefix_ptr );
                prefix_ptr += prefix_len;
            case LOG_VERB_FILE :
                sprintf( prefix_ptr, "%s:", file );
                prefix_len = strlen( prefix_ptr );
                prefix_ptr += prefix_len;
            case LOG_VERB_LINE :
                sprintf( prefix_ptr, "%d:", line );
                prefix_len = strlen( prefix_ptr );
                prefix_ptr += prefix_len;
            case LOG_VERB_FUNC :
                sprintf( prefix_ptr, "%s:", func );
                prefix_len = strlen( prefix_ptr );
                prefix_ptr += prefix_len;
            case LOG_VERB_LEVEL :
                sprintf( prefix_ptr, "%s:", npf_logger_level_names[level] );
                prefix_len = strlen( prefix_ptr );
                prefix_ptr += prefix_len;
            case LOG_VERB_COMP :
                sprintf( prefix_ptr, "%s: ", npf_logger_comp_names[comp] );
                prefix_len = strlen( prefix_ptr );
                prefix_ptr += prefix_len;
            default :
                // Do Nothing
        }

        vsprintf(prefix_ptr, format, args);
        va_end(args);
        fprintf(Logger->file, "%s\n", Logger->scratch);
        fflush(Logger->file);
    }
    PIL_LeaveCriticalSection(&Logger->lock);
}

#ifdef TESTLOG

int main() {
    int tmpint = 100;
    char *tmpchar="FP";
    npf_logger_open("pdk.log");
    npf_logger_setlevel(LOG_INFO);
    npf_logger_setmaxline(1500);
    npf_logger_setdisplay(1);
    npf_logger_setcomponent(LOG_NS, LOG_CM);
    npf_logger_write(LOG_CM, LOG_ERR, "Connection Refuse!");
    npf_logger_write(LOG_NS, LOG_INFO, "Callback %dth from %s!", tmpint, tmpchar);
    npf_logger_write(LOG_IPv4, LOG_INFO, "This will not show up!");
    npf_logger_close();

    return 0;
}

#endif /* TESTLOG */
