
/*****************************************************************************
 *                  INTEL CORPORATION PROPRIETARY INFORMATION
 *  This software is supplied under the terms of a license agreement or
 *  nondisclosure agreement with Intel Corporation and may not be copied
 *  or disclosed except in accordance with the terms of that agreement.
 *  Copyright (c) 2001 Intel Corporation. All Rights Reserved.
 ******************************************************************************/

/*****************************************************************************
 * $Source: /cvs/constellations/ixa_3.0/src/cp_pdk/common/DataTypes/src/pdk_hash.c,v $
 * $Revision: 1.5 $
 * $Date: 2003/11/17 16:57:51 $
 * $Author: mjuneja $
 * 
 *  Description: Implementation of Hash Data structure
 * ***************************************************************************/

#include "pdk_hash.h"

/**************************************************************************
 * Function: pdkHashInit
 *
 * Description: Initializes the hash control block
 *
 * Params:
 * IN OUT hashCb: Pointer to the control block
 * IN len: Length of the hash array. Or number of unique hash indexes
 * IN compFn: Function pointer to compare the keys
 * IN indFn: Function pointer to find the index
 * IN freeFn: Function pointer to free memory blocks
 *
 * Returns:
 *    PDK_HASH_FAILED : On failure
 *    PDK_HASH_OK     : On success
 **************************************************************************/

PdkHashRet pdkHashInit ( PdkHashCb           *hashCb,
                         uint32_t            len,
                         PdkHashCompFn       compFn,
                         PdkHashIndexFn      indFn,
                         PdkHashFreeFn       freeFn )
{
    uint32_t i;

    if (len <= 0 || compFn == NULL || indFn == NULL)
       return PDK_HASH_FAILED;

    hashCb->len = len;
    hashCb->compFn = compFn;
    hashCb->indFn = indFn;

    hashCb->array = (DList *)PIL_AllocateMemory(sizeof(DList) * len,
                                                   PIL_MEM_ZERO_INIT);
    if (hashCb->array == (DList *)NULL)
	return PDK_HASH_FAILED;

    for (i = 0; i < len; i++)
    {
	npf_list_init(&hashCb->array[i], freeFn);
    }

    return PDK_HASH_OK;
} /* end of pdkHashInit() */

/************************************************************************
 * Function: pdkHashInsert
 *
 * Description: Inserts a data block pointer as per the key given
 *
 * Params:
 *
 * IN hashCb: Pointer to control block
 * IN key: Pointer to hash key
 * IN data: data pointer to be inserted
 *
 * Returns:
 *    PDK_HASH_FAILED : On failure
 *    PDK_HASH_OK     : On success
 ************************************************************************/

PdkHashRet pdkHashInsert ( PdkHashCb *hashCb, void *key, void *data)
{
    uint32_t  index;
    DList     *dList;
    DListItr  *dListItr;

    index = hashCb->indFn(key);
    if ((index == -1) || (index >= hashCb->len))
	return PDK_HASH_FAILED;

    dList = &hashCb->array[index];

    if (npf_list_size(dList) == 0)
    {
	if (npf_list_pushFront(dList, data) == -1)
	{
	    return PDK_HASH_FAILED;
	}
	return PDK_HASH_OK;
    }

    dListItr = npf_list_itrCreate(dList);
    if (dListItr == NULL)
    {
	return PDK_HASH_FAILED;
    }

    /* This could be assertion as itrCreate didn't fail */
    if (npf_list_itrFirst(dListItr) == -1)
    {
	npf_list_itrDelete(dListItr);
	return PDK_HASH_FAILED;
    }

    do {
	if (hashCb->compFn(dListItr->current->data, key) == PDK_HASH_EQUAL)
	{
	    /* Duplicate key */
	    npf_list_itrDelete(dListItr);
	    return PDK_HASH_DUPLICATE;
	}

	if (npf_list_itrNext(dListItr) == -1)
	{
	    /* We have reached the end of list. Since no entry found, add it */
	    npf_list_itrDelete(dListItr);
	    if (npf_list_pushBack(dList, data) == -1)
	    {
		return PDK_HASH_FAILED;
	    }
	    break;
	}
    }
    while (1);

    return PDK_HASH_OK;
} /* end of pdkHashInsert */


/***************************************************************************
 * Function: pdkHashDelete
 *
 * Description: Remove the data block pointer from hash list.
 *              This function doesn't free the data. The data pointer is
 *              returned to caller.
 *
 * Params:
 *
 * IN hashCb: Pointer to control block
 * IN key: Pointer to hash key
 * IN OUT data: Pointer to pointer to data which is indicated by key
 *
 * Returns:
 *    PDK_HASH_FAILED : On failure
 *    PDK_HASH_OK     : On success
 **************************************************************************/

PdkHashRet pdkHashDelete(PdkHashCb *hashCb, void *key, void **data)
{
    uint32_t  index;
    DList     *dList;
    DListItr  *dListItr;

    index = hashCb->indFn(key);

    if ((index == -1) || (index >= hashCb->len))
    {
	return PDK_HASH_FAILED;
    }

    dList = &hashCb->array[index];

    if (npf_list_size(dList) == 0)
    {
	return PDK_HASH_FAILED;
    }

    dListItr = npf_list_itrCreate(dList);
    if (dListItr == NULL)
    {
	return PDK_HASH_FAILED;
    }

    /* This could be assertion as itrCreate didn't fail */
    if (npf_list_itrFirst(dListItr) == -1)
    {
	npf_list_itrDelete(dListItr);
	return PDK_HASH_FAILED;
    }

    do {
	if (hashCb->compFn(dListItr->current->data, key) == PDK_HASH_EQUAL)
	{
	    if (npf_list_itrRemove(dListItr, data) == -1)
	    {
		return PDK_HASH_FAILED;
	    }
	    npf_list_itrDelete(dListItr);
	    break;
	}

	if (npf_list_itrNext(dListItr) == -1)
	{
	    /* We have reached the end of list. */
	    npf_list_itrDelete(dListItr);
	    return PDK_HASH_FAILED;
	}
    }
    while (1);

    return PDK_HASH_OK;
} /* end of pdkHashDelete */


/***************************************************************************
 * Function: pdkHashDeleteFree
 *
 * Description: Remove the data block pointer from hash list.
 *              This function also frees the data. 
 *
 * Params:
 *
 * IN hashCb: Pointer to control block
 * IN key: Pointer to hash key
 *
 * Returns:
 *    PDK_HASH_FAILED : On failure
 *    PDK_HASH_OK     : On success
 **************************************************************************/

PdkHashRet pdkHashDeleteFree(PdkHashCb *hashCb, void *key)
{
    uint32_t  index;
    DList     *dList;
    DListItr  *dListItr;

    index = hashCb->indFn(key);

    if ((index == -1) || (index >= hashCb->len))
    {
	return PDK_HASH_FAILED;
    }

    dList = &hashCb->array[index];

    if (npf_list_size(dList) == 0)
    {
	return PDK_HASH_FAILED;
    }

    dListItr = npf_list_itrCreate(dList);
    if (dListItr == NULL)
    {
	return PDK_HASH_FAILED;
    }

    /* This could be assertion as itrCreate didn't fail */
    if (npf_list_itrFirst(dListItr) == -1)
    {
	npf_list_itrDelete(dListItr);
	return PDK_HASH_FAILED;
    }

    do {
	if (hashCb->compFn(dListItr->current->data, key) == PDK_HASH_EQUAL)
	{
	    if (npf_list_itrRemoveFree(dListItr) == -1)
	    {
		return PDK_HASH_FAILED;
	    }
	    npf_list_itrDelete(dListItr);
	    break;
	}

	if (npf_list_itrNext(dListItr) == -1)
	{
	    /* We have reached the end of list. */
	    npf_list_itrDelete(dListItr);
	    return PDK_HASH_FAILED;
	}
    }
    while (1);

    return PDK_HASH_OK;
} /* end of pdkHashDeleteFree */


/****************************************************************************
 * Function: pdkHashFind
 *
 * Description: Find the data block pointer from hash indicated by the key
 *
 * Params:
 *
 * IN hashCb: Pointer to control block
 * IN key: Pointer to hash key
 * IN OUT data: Pointer to pointer to data which is indicated by key
 *
 * Returns:
 *    PDK_HASH_FAILED : On failure
 *    PDK_HASH_OK     : On success
 ***************************************************************************/

PdkHashRet pdkHashFind(PdkHashCb *hashCb, void *key, void **data)
{
    uint32_t  index;
    DList     *dList;
    DListItr  *dListItr;

    index = hashCb->indFn(key);
    if ((index == -1) || (index >= hashCb->len))
    {
	return PDK_HASH_FAILED;
    }

    dList = &hashCb->array[index];

    if (npf_list_size(dList) == 0)
    {
	return PDK_HASH_FAILED;
    }

    dListItr = npf_list_itrCreate(dList);
    if (dListItr == NULL)
    {
	return PDK_HASH_FAILED;
    }

    /* This could be assertion as itrCreate didn't fail */
    if (npf_list_itrFirst(dListItr) == -1)
    {
	npf_list_itrDelete(dListItr);
	return PDK_HASH_FAILED;
    }

    do {
	if (hashCb->compFn(dListItr->current->data, key) == PDK_HASH_EQUAL)
	{
            // Copy found Data Pointer in user supplied address 
	    *data = dListItr->current->data;
	    npf_list_itrDelete(dListItr);
	    break;
	}

	if (npf_list_itrNext(dListItr) == -1)
	{
	    /* We have reached the end of list. */
	    npf_list_itrDelete(dListItr);
	    return PDK_HASH_FAILED;
	}
    }
    while (1);

    return PDK_HASH_OK;
} /* end of pdkHashFind */

/**************************************************************************
 * Function: pdkHashDestroy
 *
 * Description:  This deletes all the Hash List elements and fress the user
 *               data using the free function defined by user which is passed
 *               in at Hash Init.
 *
 * Params:
 * IN hashCb: Pointer to the control block
 *
 * Returns:
 *    PDK_HASH_FAILED : On failure
 *    PDK_HASH_OK     : On success
 **************************************************************************/

PdkHashRet pdkHashDestroy ( PdkHashCb       *hashCb )
{
    uint32_t  i;

    /* Perform Shutdown of  DLists in Hash Table */
    for (i = 0; i < hashCb->len; i++)
    {
	npf_list_destroy(&hashCb->array[i]);
    }
    
    /* Free Buffer reserved for Array of DLists */
    PIL_FreeMemory(hashCb->array);

    return PDK_HASH_OK;
} /* end of pdkHashDestroy() */
