/*****************************************************************************
*                  INTEL CORPORATION PROPRIETARY INFORMATION
*  This software is supplied under the terms of a license agreement or
*  nondisclosure agreement with Intel Corporation and may not be copied
*  or disclosed except in accordance with the terms of that agreement.
*      Copyright (c) 2001 Intel Corporation. All Rights Reserved.
*****************************************************************************/
 
 
/*****************************************************************************
* $Source: /CODEBASE/common/Datatypes/src/rwlock.c
* $Revision: 1.6 $
* $Date: 2003/11/17 16:57:51 $
*
* Description:
*    The implementation of read-write lock, this implementation fixed 
*	 a livelock bug in pthread rwlock 
*
* Assumptions:
*     1. Critical section is available from PIL
*     2. Since critical sections in PIL can be entered for multiple times,
*	     this rwlock can also be nested, as long as it is released in order.
*     
*****************************************************************************/
#include "rwlock.h"


/*****************************************************************************
*
*   Function: init()
*
*   Parameters:
*        lock - the rwlock to be initialized
*
*   Description: Initialize a rwlock
*
*   Return: TRUE is successful, FALSE if not 
*
*****************************************************************************/ 
Bool init(RWLOCK * lock)
{
	lock->state = INITIALIZED;
	lock->nr = 0;
	lock->nw = 0;
	PIL_InitializeCriticalSection(&(lock->r)); // always success
	PIL_InitializeCriticalSection(&(lock->w));

	return TRUE;
} // init 


/*****************************************************************************
*
*   Function: acquire()
*
*   Parameters:
*        lock - the rwlock to be acquired
*		 op - RW_READ or RW_WRITE operation
*
*   Description: Acquire a rwlock
*
*   Return: TRUE is successful, FALSE if not
*
*****************************************************************************/ 
Bool acquire(RWLOCK * lock, RW_OPERATION op)
{
	if(lock->state == UNINITIALIZED)
		return FALSE;

	if(op == RW_WRITE)
	{
		++lock->nw;
		PIL_EnterCriticalSection(&lock->w); // always success
		lock->mode = RW_WRITE;
	}
	else
	{
		PIL_EnterCriticalSection(&lock->r);

		if(++lock->nr == 1 || lock->nw > 0) 
		{
			PIL_EnterCriticalSection(&lock->w); // always success
		}
		lock->mode = RW_READ;
		PIL_LeaveCriticalSection(&lock->r);
	}
	
	return TRUE;
} // acquire


/*****************************************************************************
*
*   Function: release()
*
*   Parameters:
*        lock - the rwlock to be released
*
*   Description: Release a rwlock
*
*   Return: TRUE is successful, FALSE if not
*
*****************************************************************************/ 
Bool release(RWLOCK * lock)
{
	if(lock->state == UNINITIALIZED)
		return FALSE;

	if(lock->mode == RW_WRITE)
	{
		PIL_LeaveCriticalSection(&lock->w);
		--lock->nw;
	}
	else
	{
		PIL_EnterCriticalSection(&lock->r);
		if(--lock->nr == 0 || lock->nw > 0)
		{
			PIL_LeaveCriticalSection(&lock->w);
		}
		PIL_LeaveCriticalSection(&lock->r);
	} 

	return TRUE;
} // release;   


/*****************************************************************************
*
*   Function: destroy()
*
*   Parameters:
*        lock - the rwlock to be destroyed
*
*   Description: Destroy a rwlock
*
*   Return: TRUE is successful, FALSE if not
*
*****************************************************************************/  
Bool destroy(RWLOCK * lock)
{
	if(lock->nr || lock->nw)
		return FALSE;

	PIL_DestroyCriticalSection(&(lock->r));
	PIL_DestroyCriticalSection(&(lock->w));
	lock->state = UNINITIALIZED;

	return TRUE;
	
} // destroy 
