/******************************************************************************
//           INTEL CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or
//  nondisclosure agreement with Intel Corporation and may not be copied
//  or disclosed except in accordance with the terms of that agreement.
//      Copyright (c) 1999-2000 Intel Corporation.  All Rights Reserved.
//
// $Workfile:   Sync.c  $
// $Revision: 1.6 $
// $Date: 2003/11/17 16:57:53 $
//
// Description: Implementation of the synchronization functions
//
******************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <pil.h>

#include "csportability.h"
#include "cserror.h"

/******************************************************************************
// Method:  CsSyncCreate
//
// Parameters:
//     OUT handle - The unique identifier for the new synchronization object.
//
// Description:
//     Used to create a synchronization object. The synchronization module
//     does not maintaining the object in a list, the calling function must
//     keep track of the id for future usage.
//
// Returns:
//     CS_RET_SUCCESS if successful.
//     CS_RET_OUTOFMEMORY if out of memory.
//     CS_RET_UNKNOWN for other errors.
//
// Comments:
//     Requires: Nothing
//     Promises: The newly created synchronization object will have the
//               identifier that is returned.
******************************************************************************/
ULONG CsSyncCreate( ULONG *handle ) 
{

    PilSemaphore sem;

    if (!(sem = PIL_CreateSemaphore(1,1)))
    {    
        return CS_RET_UNKNOWN;
    }

    *handle = (ULONG)sem;
    return CS_RET_SUCCESS;
}


/******************************************************************************
// Method:  CsSyncDestroy
//
// Parameters:
//     handle - the id of the synchronization object to destroy.
//
// Description:
//     Used to destroy a previously created synchronization object. Any
//     synchronization objects passed to this method will no longer be
//     usable.
//
// Returns:
//     CS_RET_SUCCESSFUL if successful, CS_RET_INVALIDHANDLE if the handle
//     passed in is not valid.
//
// Comments:
//     Requires: The handle must be that of an active synchronization object.
//     Promises: The synchronization object will be destroyed and all of its
//               resources freed.
******************************************************************************/
ULONG CsSyncDestroy( ULONG handle ) 
{

    if ( !handle ) 
    {
        return CS_RET_INVALIDHANDLE;
    }

    if((PIL_DestroySemaphore( (PilSemaphore)handle )) == PIL_INVALID_HANDLE)
    {
        return CS_RET_INVALIDHANDLE;
    }

    return CS_RET_SUCCESS;
}


/******************************************************************************
// Method:  CsSyncLock
//
// Parameters:
//     handle - the id of the synchronization object to try to lock.
//
// Description:
//     Used to set the synchronization object to the locked state. If the
//     synchronization object is currently locked the calling function will
//     block.
//
// Returns:
//     CS_RET_SUCCESS if successful, CS_RET_INVALIDHANDLE if the handle
//     provided is invalid.
//
// Comments:
//     Requires: The handle must be valid.
//     Promises: The synchronization object will be locked.
******************************************************************************/
ULONG CsSyncLock( ULONG handle ) 
{
    if ( !handle ) 
    {
        return CS_RET_INVALIDHANDLE;
    }

    if( (PIL_AcquireSemaphore( (PilSemaphore)handle, PIL_INFINITE_TIMEOUT )) != PIL_OK)
    {  
        return CS_RET_INVALIDHANDLE;
    }

    return CS_RET_SUCCESS;

}


/******************************************************************************
// Method:  CsSyncUnlock
//
// Parameters:
//     handle - the id of the synchronization object to unlock.
//
// Description:
//     Used to set the synchronization object to an unlocked state. This will
//     make the object available to be locked again.
//
// Returns:
//     CS_RET_SUCCESS if successful, CS_RET_INVALIDHANDLE if an invalid
//     handle is passed.
//
// Comments:
//     Requires: The handle must be valid.
//     Promises: The synchronization object will be unlocked.
//
******************************************************************************/
ULONG CsSyncUnlock( ULONG handle ) 
{
    if ( !handle ) 
    {
        return CS_RET_INVALIDHANDLE;
    }

    if( (PIL_ReleaseSemaphore( (PilSemaphore)handle )) == PIL_INVALID_HANDLE)
    {
      return CS_RET_INVALIDHANDLE;
    }

    return CS_RET_SUCCESS;
}


