/******************************************************************************
//           INTEL CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or
//  nondisclosure agreement with Intel Corporation and may not be copied
//  or disclosed except in accordance with the terms of that agreement.
//      Copyright (c) 1999-2000 Intel Corporation.  All Rights Reserved.
//
// $Workfile:   pgeneral.c  $
// $Revision: 1.6 $
// $Date: 2003/11/17 16:57:53 $
//
// Description: Implementation of general functions
//
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include "timer.h"
#include "pgeneral.h"
#include "cserror.h"
#include "csportability.h"
#include "ptcpip.h"

#include "slist.h"

#define FILENAMELEN 256
#define FILENAME "portablog"

/* global variables */

BOOL g_portInit=FALSE;

/* end globals */

/******************************************************************************
//
// METHOD:       ULONG CsPortInit(  void (*pCbCsNetReceive),
//                                  void (*pCbCsNetDisconnect),
//                                  void (*pCbCsTimerExpire),
//                                  void (*pCbCsNetAccept)) 
//
// ARGUMENTS:    Callback function pointers for Receive,Disconnect,Timeout
//               and accept are the arguments
//
// DESCRIPTION:  Initialize the portability required variables and starts
//               Client/Server Threads and Initializes Windows Socket Library.
//
// RETURNS:      CS_RET_SUCCESS
//               CS_RET_INVALIDARGS
//               CS_RET_UNKNOWN
//               CS_RET_ALREADYINITED
//               CS_RET_NETWORKFAILED
//
******************************************************************************/

ULONG CsPortInit( 
		 void (*pCbCsNetReceive)( ULONG handle, 
                                   const char *message, 
                                   int length ),
		 void (*pCbCsNetDisconnect)( ULONG handle ),
		 void (*pCbCsTimerExpire)( ULONG handle ),
		 void (*pCbCsNetAccept)( ULONG handle, char *addr) ) 
{

    if ( g_portInit == TRUE ) 
    {
	return CS_RET_ALREADYINITED;
    }

    if ( pCbCsNetReceive == NULL || 
       pCbCsNetDisconnect == NULL || 
       pCbCsTimerExpire == NULL ) 
    {
	return CS_RET_INVALIDARGS;
    }

    CallbackNetReceive = pCbCsNetReceive;
    CallbackNetDisconnect = pCbCsNetDisconnect;
    CallbackNetAccept = pCbCsNetAccept;

    if ( !TimerInit( pCbCsTimerExpire ) ) 
    {
	return CS_RET_UNKNOWN;
    }

    if(! NetworkInit())
    {
	CSTRACE((CSTRACE_ERR, "[PORTAB] Network Init failed.\n"));
	return CS_RET_NETWORKFAILED;
    }

    // ignore SIGPIPE, CsNetSend will return CS_RET_NETWORKFAILED
    // instead of crashing.
    signal(SIGPIPE, SIG_IGN);

    g_portInit = TRUE;
    return CS_RET_SUCCESS;
}/*CsPortInit*/

/******************************************************************************
//
// METHOD:      void CsPortDeinit( void )
//
// ARGUMENTS:   None.
//
// DESCRIPTION: Clean up internal values, object state...etc., and shutdown
//              the server or client.
//
// RETURNS:     Nothing.
//              
//
******************************************************************************/

void CsPortDeinit( void ) 
{

    if ( g_portInit == FALSE ) 
	return;

    TimerDeinit();
    NetworkDeinit();
    g_portInit = FALSE;

}


/******************************************************************************
//
// METHOD:      ULONG CsLogStatus( USHORT severity,  USHORT event, char *Msg)
//
// ARGUMENTS:   Severity : Error severity to log
//              event    : error event value
//              Msg      : Error message string pointer.
//
// DESCRIPTION: Logs the passed information to a Log file.
//		NOTE: This function is not used in the current 
//		      implementation. It has been provided as an
//		      example of how the Log Status could be
//		      implemented.              
//
// RETURNS:     CS_RET_SUCCESS
//              
//
******************************************************************************/

ULONG CsLogStatus( USHORT severity,  USHORT event, char *Msg ) 
{
  
    static char *ErrStr[] = 
    {
	"Success" ,
	"Out of Memory" ,
	"Init Error" ,
	"Already Initialized" ,
	"Invalid Handle" ,
	"Invalid Argument" ,
	"Invalid Server" ,
	"Error Connecting Server" ,
	"Already Connected" ,
	"Error Sending Data" ,
	"Error Receiving Data" ,
	"Server Not Found" ,
	"Invalid Event" ,
	"Server Already Listening" ,
	"Unknown Error",
	"No Impact",
	"Minor",
	"Major",
	"Critical",
	"Show Stopper"
    };
  
    time_t currentTime;
    FILE *LogFileHandle;
    char dbuffer [9];
  
    time( &currentTime );
    strcpy( dbuffer, ctime( &currentTime ) );
  
    LogFileHandle = fopen(FILENAME,"a");
    fwrite(dbuffer,sizeof(char),8,LogFileHandle);
    fwrite(" ",sizeof(char),1,LogFileHandle);
  
    if ( (severity == 0) && (event == 0) )
    {
	fwrite(Msg,sizeof(char),10,LogFileHandle);
    }
    else
    {
	fwrite(ErrStr[severity],sizeof(char),6,LogFileHandle);
	fwrite(" ",sizeof(char),1,LogFileHandle);
	fwrite(ErrStr[event],sizeof(char),10,LogFileHandle);
    }
  
    fwrite("\x0D\x0A",sizeof(char),2,LogFileHandle);
    fclose(LogFileHandle);
  
  return CS_RET_SUCCESS;
}


