/******************************************************************************
//             INTEL CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or 
//  nondisclosure agreement with Intel Corporation and may not be copied 
//  or disclosed except in accordance with the terms of that agreement. 
//      Copyright (c) 2000-2001 Intel Corporation. All Rights Reserved.
//
//  $Workfile:   ptcpip.c  $
//  $Revision: 1.6 $
//  $Date: 2003/11/17 16:57:55 $
//
//  Description: Implementation of networking functions
//
******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <pil.h>
#include <stddef.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#ifdef _VXWORKS_
#include <ioLib.h>
#include <pipeDrv.h>
#include <sockLib.h>
#else
#include <fcntl.h>
#endif

#include "cserror.h"
#include "csdatatypes.h"
#include "csportability.h"
#include "pgeneral.h"
#include "ptcpip.h"
#include "slist.h"

#ifdef SOLARIS
 #include <strings.h>
#endif

/* global variables*/
static BOOL    g_ActingAsClient=FALSE;
static BOOL    g_ActingAsServer=FALSE;
static int     g_ListenSocket;

static LPLIST  g_PortSockList;
static int     g_socketPipe[2];  /*pipe used to signal a thread to close*/
static PilThread g_ListenThreadHandle;

/*Sync objects*/
static ULONG   g_DataLock;
static ULONG   g_SocketLock;

/*Function prototypes*/
static ULONG AppendSocket(SOCKET sHandle, PilThread tHandle);
static ULONG DeleteSocket(SOCKET handle );
static PsockStruct FindSocket(SOCKET handle );
static ULONG DeleteAndCloseSocketList(void);
static uint32_t ListenThread(void *listenSock);
static uint32_t ProcessSocketEvent(void *);
static ULONG CloseThisSocket(SOCKET handle);

/******************************************************************************
// Method:  ULONG AppendSocket(SOCKET sHandle, PilThread tHandle)
//
// Parameters:
//     sHandle - a client socket to add to list.
//     tHandle - the thread handle of the thread which is listening on the 
//                 socket.
//
// Description:
//     Used to append a new client socket accepted by server into the list
//     while maintaining the order of the list. The list must be kept in
//     ascending order by socket value.
//
// Returns:
//     CS_RET_SUCCESS if successful.
//     CS_RET_OUTOFMEMORY if memory allocation fails.
// 
******************************************************************************/
ULONG AppendSocket(SOCKET sHandle, PilThread tHandle)
{
    sockStruct *newSocket;

    newSocket = (sockStruct *)malloc(sizeof(sockStruct));
    if( newSocket == NULL )
    {
        CSTRACE((CSTRACE_INFO, "[PORTAB] unable to allocate memory "
                                 "for socket 0x%lx.\n", sHandle));
        return CS_RET_OUTOFMEMORY;
    }

    newSocket->sockHandle = sHandle;
    newSocket->threadHandle = tHandle;

    CsSyncLock(g_SocketLock);

    if( ListAddItem(g_PortSockList, newSocket) == FALSE )
    {
        CSTRACE((CSTRACE_INFO, "[PORTAB] unable to add socket 0x%lx "
                   "to socketlist.\n", sHandle));
        free(newSocket);

        CsSyncUnlock(g_SocketLock);

        return CS_RET_OUTOFMEMORY;
    }

    CsSyncUnlock(g_SocketLock);

    CSTRACE( (CSTRACE_INFO,"[PORTAB]Appended socket %ld.\n", sHandle));
    return CS_RET_SUCCESS;

}/*AppendSocket*/

/******************************************************************************
// Method:  ULONG DeleteSocket(SOCKET handle)
//
// Parameters:
//     handle - A client socket to delete from the list.
//
// Description:
//     This method will search for the given handle in the list of client
//           sockets and will remove it from the list and also free its memory.
//
// Returns:
//     CS_RET_SUCCESS if successful.
//     CS_RET_INVALIDHANDLE if handle is not found.
//     CS_RET_OUTOFMEMORY if memory allocation fails.
//
******************************************************************************/
ULONG DeleteSocket(SOCKET handle ) 
{
    LPLISTITR itr;
    sockStruct *curStruct;

    CsSyncLock(g_SocketLock);

    itr = ListItrCreate(g_PortSockList);

    if( itr == NULL )
    {
        CSTRACE((CSTRACE_ERR, "[PORTAB] unable to create itr for "
                   "g_PortSockList!\n"));

        CsSyncUnlock(g_SocketLock);

        return CS_RET_OUTOFMEMORY;
    }

    while( (curStruct = (sockStruct *)ListItrGetNext(itr)) != NULL )
    {
        if( curStruct->sockHandle == handle )
        {
            /* found!*/


            if( ListItrRemoveItem(itr) == FALSE )
            {
                CSTRACE((CSTRACE_ERR, "[PORTAB] unable to remove socket "
                            "0x%lx from socket list!\n"));

                ListItrDelete(itr);
                CsSyncUnlock(g_SocketLock);

                return CS_RET_OUTOFMEMORY;
            }
            free(curStruct);

            ListItrDelete(itr);
            CsSyncUnlock(g_SocketLock);
              
            return CS_RET_SUCCESS;
        }/*if(curStruct-> ...*/

    }/*while(...*/

    ListItrDelete(itr);
    CsSyncUnlock(g_SocketLock);

    return CS_RET_INVALIDHANDLE;
}/*DeleteSocket*/


/******************************************************************************
// Method:  PsockStruct FindSocket(SOCKET handle)
//
// Parameters:
//     handle - A client socket to look for
//
// Description:
//     This method will search for the given handle in the list and return 
//	a PsockStruct allocated and filled. The return pointer has to be deleted.
//
// Returns:
//     CS_RET_SUCCESS       if found
//     CS_RET_INVALIDHANDLE if not found
//     CS_RET_OUTOFMEMORY if memory allocation fails.
//
******************************************************************************/
PsockStruct FindSocket(SOCKET handle ) 
{
    LPLISTITR itr;
    sockStruct *curStruct;
    sockStruct *retStruct;

    CsSyncLock(g_SocketLock);

    itr = ListItrCreate(g_PortSockList);

    if( itr == NULL )
    {
        CSTRACE((CSTRACE_ERR, "[PORTAB] unable to create itr for "
               "g_PortSockList!\n"));

        CsSyncUnlock(g_SocketLock);
        return NULL;
    }

    while( (curStruct = (sockStruct *)ListItrGetNext(itr)) != NULL )
    {
        if( curStruct->sockHandle == handle )
        {
            /* found!*/
	    retStruct = (sockStruct *)calloc(sizeof(sockStruct),1);
	    retStruct->sockHandle = curStruct->sockHandle;
	    retStruct->threadHandle = curStruct->threadHandle;
            ListItrDelete(itr);
            CsSyncUnlock(g_SocketLock);

            return retStruct;
        }      
    }

    ListItrDelete(itr);
    CsSyncUnlock(g_SocketLock);

    return NULL;
}/*FindSocket*/


/******************************************************************************
// Method:  ULONG DeleteAndCloseSocketList()
//
// Parameters:
//     None
//
// Description:
//     This method deletes all the elements from the socket list of clients
//     connected to this server and also frees associated memory.
//
// Returns:
//     CS_RET_SUCCESS          if successful.
//     CS_RET_OUTOFMEMORY if memory allocation fails.
//
// 
******************************************************************************/
ULONG DeleteAndCloseSocketList(void)
{

    sockStruct *curStruct=NULL;
    sockStruct tempsock;

    while (1)
    {
    	CsSyncLock(g_SocketLock);
    	curStruct = (sockStruct *)ListGetFirst(g_PortSockList);
	if (NULL != curStruct)
	{
		tempsock.sockHandle = curStruct->sockHandle;
		tempsock.threadHandle = curStruct->threadHandle;
        	CloseThisSocket(tempsock.sockHandle);
    		CsSyncUnlock(g_SocketLock);
        	PIL_WaitForThread(tempsock.threadHandle, PIL_INFINITE_TIMEOUT);
	}
	else
	{
    		CsSyncUnlock(g_SocketLock);
    		return CS_RET_SUCCESS;
	}
    }

#if 0
    while( (curStruct = (sockStruct *)ListItrGetNext(itr)) != NULL )
    {
	sockThread = curStruct->threadHandle;
        CSTRACE((CSTRACE_INFO, "[PORTAB] closing sock %d thread 0x%x!\n", curStruct->sockHandle, curStruct->threadHandle));
        CloseThisSocket(curStruct->sockHandle);
        PIL_WaitForThread(sockThread, PIL_INFINITE_TIMEOUT);

        CSTRACE((CSTRACE_INFO, "[PORTAB] Thread closed successfully!\n"));

        if( ListItrRemoveItem(itr) == FALSE )
        {
            CSTRACE((CSTRACE_ERR, "[PORTAB] unable to remove socket "
                   "0x%lx from socket list!\n"));

            ListItrDelete(itr);
            CsSyncUnlock(g_SocketLock);

            return CS_RET_OUTOFMEMORY;
        }

        free(curStruct);

    }

    ListItrDelete(itr);
    CsSyncUnlock(g_SocketLock);

    return CS_RET_SUCCESS;
#endif

}/*DeleteAndCloseSocketList*/

/******************************************************************************
// 
// METHOD: NetworkInit()
//
// ARGUMENTS: none
//
// DESCRIPTION: Initialize network module
//
// RETURNS: TRUE if successful, FALSE otherwise.
******************************************************************************/

BOOL NetworkInit()
{


    g_PortSockList = ListCreate();
    if( g_PortSockList == NULL )
    {
        return FALSE;
    }    

#ifdef _VXWORKS_
    pipeDevCreate("/pipe/interconnect/listen",1,strlen(TERMINATE_STR)+1);
    g_socketPipe[0] = g_socketPipe[1] = open("/pipe/interconnect/listen",O_RDWR,0777);
#else
    if(pipe(g_socketPipe))
    {
        CSTRACE((CSTRACE_ERR, "[PORTAB] Unable to create pipe.\n"));
        ListDelete(g_PortSockList);
        return FALSE;    
    }
#endif

    //All synchronization objects need to be created globally
    if(CsSyncCreate(&g_DataLock) != CS_RET_SUCCESS)
    {
        close(g_socketPipe[0]);
#ifndef _VXWORKS_
        close(g_socketPipe[1]);
#endif
        ListDelete(g_PortSockList);
        return FALSE;        
    }
    
    if(CsSyncCreate(&g_SocketLock)!= CS_RET_SUCCESS)
    {
        CsSyncDestroy(g_DataLock);
        close(g_socketPipe[0]);
#ifndef _VXWORKS_
        close(g_socketPipe[1]);
#endif
        ListDelete(g_PortSockList);
        return FALSE;        
    }

    return TRUE;
}/*NetworkInit*/


/******************************************************************************
// 
// METHOD: NetworkDeinit()
//
// ARGUMENTS: none
//
// DESCRIPTION: deinitialize network module
//
// RETURNS: TRUE if successful, FALSE otherwise.
******************************************************************************/

BOOL NetworkDeinit()
{

    BOOL errFlag = TRUE;

    if(g_ActingAsServer == TRUE)
    {
        /* listen alive, stop listen first*/
        if(!(errFlag = CsNetStopListen()))
        {
            CSTRACE((CSTRACE_ERR, "[PORTAB] NetStopListen failed.\n"));
        }
    }

    /*Here this is useful only for the client side. The server side calls
      this function in CsNetStopListen*/
    DeleteAndCloseSocketList();

    //All synchronization objects need to be destroyed globally
    CsSyncDestroy(g_DataLock);
    CsSyncDestroy(g_SocketLock);
    
    close(g_socketPipe[0]);
#ifndef _VXWORKS_
    close(g_socketPipe[1]);
#endif

    ListDelete(g_PortSockList);

    /*Set all flags to FALSE*/
    g_ActingAsServer = FALSE;
    g_ActingAsClient = FALSE;
    return errFlag;
}/*NetworkDeinit*/


/******************************************************************************
//
// METHOD:      void *ListenThread(socket listenSock)
//
// ARGUMENTS:   listenSock     listening socket
//
// DESCRIPTION: This thread handles the listening and accepting of socket.
//
//
// RETURNS:     NULL
//
******************************************************************************/
uint32_t ListenThread(void *listenSock)
{
    ULONG     handle;
    ULONG     *handlePtr;
    int       connCliFD = 0;

    int       highestFD;
    fd_set    rfds;
    int       retval;
    char      recvBuffer[MAX_RECV_LEN];

    PilThread ThreadHandle;
    int connectedCliAddrLen;
    struct sockaddr_in  connectedCliAddr;

    if(!listenSock)
        return 0;

    CSTRACE((CSTRACE_INFO, "[PORTAB] Listen thread started, "
           "listen sock is 0x%x. listen thread 0x%08x g 0x%0x\n", *((int *)listenSock) ));

    /*Enter loop for accepting new connections*/
    while(1)
    {

	connectedCliAddrLen = sizeof(struct sockaddr_in);
        connCliFD = accept(*((int *)listenSock), 
                           (struct sockaddr *) &connectedCliAddr,
                            &connectedCliAddrLen);

        if(connCliFD == -1 )
        {
            switch(errno)
            {
            case EWOULDBLOCK:
                    CSTRACE((CSTRACE_INFO, "Received EWOULDBLOCK."
                                            " Everything OK\n"));
                    break;
            case EBADF:
                    CSTRACE((CSTRACE_ERR, "EBADF\n"));
                    free(listenSock);
                    PIL_ExitThread(0);
                    break;
            case ENOTSOCK:
                    CSTRACE((CSTRACE_ERR, "ENOTSOCK\n"));
                    free(listenSock);
                    PIL_ExitThread(0);
                    break;
            case EOPNOTSUPP:
                    CSTRACE((CSTRACE_ERR, "EOPNOTSUPP\n"));
                    free(listenSock);
                    PIL_ExitThread(0);
                    break;
            case EFAULT:
                    CSTRACE((CSTRACE_ERR, "EFAULT\n"));
                    free(listenSock);
                    PIL_ExitThread(0);
                    break;
            case EPERM:
                    CSTRACE((CSTRACE_ERR, "EPERM\n"));
                    free(listenSock);
                    PIL_ExitThread(0);
                    break;
            default:
                    CSTRACE((CSTRACE_ERR, "Unknown\n"));
                    free(listenSock);
                    PIL_ExitThread(0);
                    break;
            }/*switch*/
        }
        else
        {
            /*New Socket connected*/

           handle = SOCKETToULONG(connCliFD);    
           handlePtr = (ULONG *) calloc(1, sizeof(ULONG));    
           if(!handlePtr)
           {
                CSTRACE((CSTRACE_ERR, "[PORTAB] Calloc failed.\n"));
                free(listenSock);
                PIL_ExitThread(0);        
           }
           *handlePtr = handle;

           /*Create new thread for the newly conected socket*/
/*
           if(pthread_create(&ThreadHandle, NULL, ProcessSocketEvent, 
                                handlePtr) != 0)
*/
           if ( 0 == (ThreadHandle = PIL_CreateThread(
                                    ProcessSocketEvent,
                                    DEFAULT_STACK_SIZE,
                                    NULL,
                                    DEFAULT_THREAD_PRIORITY,
                                    handlePtr,
                                    0)))
            {
                /*thread creation failed*/
                free(handlePtr);
                CSTRACE( (CSTRACE_ERR,"[PORTAB]Client:Can't create "
                                    "client thread.\n"));

                if( CloseThisSocket(handle) == CS_RET_SUCCESS )
                {
                    CSTRACE( (CSTRACE_INFO, "[PORTAB]Successfully"
                        " disconnected the socket %X.\n", handle));
                }
            }
            else
            {
                CSTRACE((CSTRACE_INFO, "[PORTAB]: Successfully created"
                    " client thread, thread id is 0x%08X\n", ThreadHandle));

                /*Add the client socket to socket list*/
                AppendSocket(handle, ThreadHandle);
            }/*if(PIL_CreateThread...*/


        }/*if(connCliFD == -1)*/

        /*Do a select now. Wait for either a new connection, or the 
          terminate command on the pipe*/

        FD_ZERO(&rfds);
        FD_SET(*((int *)listenSock) , &rfds);
        FD_SET(g_socketPipe[0], &rfds);

        if( *((int *)listenSock) > g_socketPipe[0])
                highestFD = *((int *)listenSock) + 1;
        else
                highestFD = g_socketPipe[0] + 1;

        retval = select(highestFD, &rfds, NULL, NULL, NULL);

        if(retval == -1)
        {
            CSTRACE((CSTRACE_INFO, "[PORTAB] SELECT returned -1."
                " Listen thread exiting...\n"));

            free(listenSock);
            PIL_ExitThread(0);
        }

        /*check if data was written to the pipe */
        if(FD_ISSET(g_socketPipe[0], &rfds))
        {
            retval = read(g_socketPipe[0], recvBuffer, MAX_RECV_LEN);
                
            /*if there is an error, or if the pipe contains the 
              terminate string, exit */
            if( (retval <= 0) ||
                !strncmp(recvBuffer,TERMINATE_STR , retval) )
            {
                CSTRACE((CSTRACE_ERR, "[PORTAB]Terminate command "
                                      "received\n"));
                free(listenSock);
                PIL_ExitThread(0);
            }

            /*go to the start of the loop*/
            continue;        
        }/*if(FD_ISSET.... */

    CSTRACE((CSTRACE_INFO, "[PORTAB] Returned from SELECT\n"));


    /*Select was for the socket. So go back to the top*/

    }/*while(1)*/

}/*ListenThread*/

/******************************************************************************
//
// METHOD:      void *ProcessSocketEvent(void *handlePtr)
//
// ARGUMENTS:   handlePtr : Socket handle pointer
//
// DESCRIPTION: This thread handles the read on behalf of client.
//
//
// RETURNS:     NULL
//              
//
******************************************************************************/
uint32_t ProcessSocketEvent(void* handlePtr)
{
    int sockFD = 0;
    char recvBuffer[MAX_RECV_LEN];
    int nread = 0;

    struct sockaddr name;
    int namelen;
    char *netaddr;

    CSTRACE ((CSTRACE_INFO, "[PORTAB]Entered Client Thread.\n"));

    if(!handlePtr)
    {
        CSTRACE((CSTRACE_ERR, "[PORTAB] handlePtr is NULL\n"));
        return 0;
    }

    if(g_ActingAsServer)
    {
        namelen = sizeof(name);

        if( getpeername( *((int *)handlePtr), &name, &namelen))
        {
            CSTRACE((CSTRACE_ERR, "PORTAB: Couldn't get socket info\n"));
            free(handlePtr);
            return 0;
        }

        /*inform the upper layer of the new socket*/
        netaddr = inet_ntoa(  ((struct sockaddr_in *)name.sa_data)->sin_addr);
        CallbackNetAccept( *((ULONG *)handlePtr), netaddr);
    }/*if(g_ActingAsServer)*/

    /*Start normal activities*/

    sockFD = ULONGToSOCKET( *((ULONG*) handlePtr) );

    /*read data */
    while(1)
    {
        nread = read(sockFD, recvBuffer, MAX_RECV_LEN);
        if(nread < 0)
        {
          /* error in read */

            CSTRACE((CSTRACE_ERR, "[PORTAB] Read returned %d!\n", nread));

            switch(errno)
            {
            case EINTR:
            case EAGAIN:
                nread=0;
                break;
            case EIO:
            case EISDIR:
            case EBADF:
            case EINVAL:
            case EFAULT:
            default:
perror("Listen Thread read: ");
                CSTRACE((CSTRACE_INFO,"[PORTAB] read error! socket already"
                                 "deleted?!\n"));

                CallbackNetDisconnect(sockFD);

                CSTRACE((CSTRACE_INFO, "[PORTAB] thread 0x%08X exited...\n",
                                        PIL_GetCurrentThreadId()));

                CSTRACE((CSTRACE_INFO, "[PORTAB] closing sockFD %d\n",sockFD));
                close(sockFD);

                /*remove from socket list*/
                DeleteSocket(sockFD);

                free(handlePtr);
                PIL_ExitThread(0);                    
            }/*switch*/

        } /* if( nread < 0 ) */
        else if(nread == 0)
        {                                        

            CSTRACE((CSTRACE_ERR, "[PORTAB]EOF detected!!!\n"));
            
            CallbackNetDisconnect(sockFD);

            CSTRACE((CSTRACE_INFO, "[PORTAB] thread 0x%08X exited...\n",
               PIL_GetCurrentThreadId()));

                CSTRACE((CSTRACE_INFO, "[PORTAB] closing sockFD %d\n",sockFD));
            close(sockFD);

            /*remove from socket list*/
            DeleteSocket(sockFD);
            
            free(handlePtr);
            PIL_ExitThread(0);                        
        }
        else
        {
                /* we have new data! */
            CSTRACE((CSTRACE_INFO, "[PORTAB]Number of bytes "
               "received %d,Socket %X\n", nread, sockFD));

            CallbackNetReceive(sockFD, recvBuffer, nread);

        }/*else*/

    } /* while(1) */                        

}/*ProcessSocketEvent*/

/******************************************************************************
// TCP/IP Functions
******************************************************************************/

/******************************************************************************
//
// METHOD:      ULONG CsNetSend( ULONG handle,  LPBYTE message, int length )
//
// ARGUMENTS:   handle  : Socket handle
//              message : Pointer to the data to be sent  
//              length  : Size of the data
//  
// DESCRIPTION: Sends data to the specified socket.
//              
//
// RETURNS:     Returns CS_RET_SUCCESS, if successful otherwise returns
//              appropriate error values.
//              
//
******************************************************************************/
ULONG CsNetSend( ULONG handle,  LPBYTE message, int length )
{
    char *CurDataPtr = (char *)message;
    int nwritten = 0;
#ifdef TIMESTAMP
	{
        struct timeval timestamp1;
        gettimeofday (&timestamp1, 0);
        printf("======> timestamp1: tv_sec = %d, tv_usec = %d\n",
            timestamp1.tv_sec, timestamp1.tv_usec);
    }  
#endif
    /*Keep writing as long as there is some data left to be sent*/
    while(length > 0) 
    {
        nwritten = write(ULONGToSOCKET(handle), CurDataPtr, length);
        if( nwritten == -1 )
        {
            /* error occured during write*/
            CSTRACE((CSTRACE_ERR, "[PORTAB] write failed! errno = %d.\n",
                                   errno));
            return CS_RET_SOCKETCLOSED;
        }

        length -= nwritten;
        CurDataPtr += nwritten;
        CSTRACE((CSTRACE_INFO, "[PORTAB]%d bytes written.\n", nwritten));
    }/*while(length > 0)*/

    return CS_RET_SUCCESS;
}/*CsNetSend*/

/******************************************************************************
//
// METHOD:      ULONG CsNetConnect(char *addr,USHORT port,ULONG *handle,
//                                CS_IPADDR *ServerIp);
//
// ARGUMENTS:   addr    - IP address or Domain name of the Server
//              port    - port of the Server
//              handle  - This is a out variable, returned by the Server   
//                         if it connects successfully.
//              ServerIp- Backup Server IP address
//  
// DESCRIPTION: Connects to the server listening at the port. (param 2)
//              
//
// RETURNS:     CS_RET_SUCCESS
//              CS_RET_INVALIDHOSTNAME
//              CS_RET_SOCKETCREATERR
//              CS_RET_CONNECTFAILED
//              CS_RET_NOTINITIALIZED
//
******************************************************************************/

ULONG CsNetConnect(char *addr, USHORT port, ULONG *handle, CS_IPADDR *ServerIp)
{
    #ifndef _VXWORKS_
    struct hostent *pHost;
    #endif
    struct sockaddr_in sin;
    PilThread ThreadHandle;
    SOCKET clientSocket = 0;
    struct in_addr ipAddr;
    ULONG  *handlePtr;

    CSTRACE( (CSTRACE_INFO, "[PORTAB]ULONG CsNetConnect(.....\n") );

    if( g_portInit == FALSE )
    {
        CSTRACE((CSTRACE_ERR, "[PORTAB] Need to call CsPortInit() first "
                                  "before calling this function.\n"));
        return CS_RET_NOTINITIALIZED;
    }

#ifdef _VXWORKS_
    ipAddr.s_addr = htonl(ServerIp->addr);

#else

#ifdef SOLARIS
    /*doesn't support inet_aton yet, so use inet_addr*/
    if((ipAddr.s_addr = CSINET_ADDR(addr)))
#else
    if( inet_aton(addr, &ipAddr) == 0 )
#endif
   {
        pHost = gethostbyname (addr);
        if(NULL == pHost)
        {
              /*Unable to resolve host name*/
              CSTRACE( (CSTRACE_ERR, 
                        "[PORTAB]Error resolving Host Name %s.\n", addr) );
              return CS_RET_INVALIDHOSTNAME;
        }

        memcpy((char *)&(ipAddr.s_addr), pHost->h_addr, 
                    sizeof(ipAddr.s_addr));
    }/*if(inet_aton...*/

    ServerIp->addr = ipAddr.s_addr;
#endif

    /* create the socket*/
    if((clientSocket = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
        /*Unable to create socket*/
        *handle = 0;
        CSTRACE( (CSTRACE_ERR, "[PORTAB]Error creating a socket for client\n"));
        return CS_RET_SOCKETCREATERR;
    }

    CSTRACE( (CSTRACE_INFO, 
            "[PORTAB]Successfully created a socket for client. "
            "Socket %X.\n", clientSocket) );

    bzero((char *) &sin, sizeof(sin));
    sin.sin_family        = AF_INET;
    sin.sin_addr.s_addr = ipAddr.s_addr; /*already in Network Byte Order*/
    sin.sin_port        = CSHTONS(port);

    /* connect to remote endpoint*/
    CSTRACE( (CSTRACE_INFO, "[PORTAB]Trying to connect at IP %s, "
            "IP in Hex %X, Port %X.\n", addr, ipAddr.s_addr, port) );

    if(connect(clientSocket, (struct sockaddr *)&sin, sizeof (sin)) < 0)
    {
        CSTRACE( (CSTRACE_ERR, "[PORTAB]Error connecting at IP %s, %X, "
                "Port %X.\n", addr, ipAddr.s_addr, port) );

        /* connect failed */
        close (clientSocket);
        *handle = 0;
        return CS_RET_CONNECTFAILED;
    }

    CSTRACE( (CSTRACE_INFO, "[PORTAB]Client connected to server.\n"));

    *handle = SOCKETToULONG(clientSocket);

    /* handlePtr will be deleted when the thread quits*/
    handlePtr = calloc(1, sizeof(ULONG));
    if(handlePtr == NULL)
    {
        CSTRACE((CSTRACE_ERR, "[PORTAB]: Memory allocation failed\n"));
        close(clientSocket);
        return CS_RET_OUTOFMEMORY;
    }

    *handlePtr = *handle;

/*
    if(pthread_create(&ThreadHandle, NULL, ProcessSocketEvent, 
                        handlePtr) == 0)
*/
    if ( 0 != (ThreadHandle = PIL_CreateThread(
                              ProcessSocketEvent,
                              DEFAULT_STACK_SIZE,
                              NULL,
                              DEFAULT_THREAD_PRIORITY,
                              handlePtr,
                              0)))
    {
        CSTRACE((CSTRACE_INFO, 
               "[PORTAB]Successfully created client thread, "
               "thread id is 0x%08X\n", ThreadHandle));

        /* add the socket to socket list*/
        AppendSocket(clientSocket, ThreadHandle);

        CsSyncLock(g_DataLock);
        g_ActingAsClient = TRUE;
        CsSyncUnlock(g_DataLock);

        return CS_RET_SUCCESS;                
    }

    /* thread creation failed*/
    free(handlePtr);

    CSTRACE( (CSTRACE_ERR, 
            "[PORTAB]Client:Can't create client thread.\n"));

    if( CloseThisSocket(clientSocket) == CS_RET_SUCCESS )
    {

        CSTRACE( (CSTRACE_INFO, 
                "[PORTAB]Successfully disconnected the socket %X.\n", 
                clientSocket) );        
    }
    /* out of resource*/
    return CS_RET_OUTOFMEMORY;

}/*CsNetConnect*/


/******************************************************************************
//
// METHOD:      ULONG CsNetListen( USHORT port )
//
// ARGUMENTS:   port : To Listen for connection.
//
// DESCRIPTION: Starts a server and waits for the client connection.
//              
// RETURNS:     Returns CS_RET_SUCCESS, if successful otherwise returns
//              appropriate error values.
//             
******************************************************************************/
ULONG CsNetListen( USHORT port )
{

    struct sockaddr_in serv_addr;
    int sockFD;
    int reuse = 1;
    ULONG *handlePtr  = NULL;
    int retval = 0;

    CSTRACE( (CSTRACE_INFO, "[PORTAB]ULONG CsNetListen(....).\n") );

    if( g_portInit == FALSE )
    {
        CSTRACE((CSTRACE_ERR, "[PORTAB] Need to call CsPortInit() first "
               "before calling this function.\n"));
        return CS_RET_NOTINITIALIZED;
    }

    CsSyncLock(g_DataLock);
    
    if(g_ActingAsClient == TRUE) 
    {
        CsSyncUnlock(g_DataLock);
        CSTRACE( (CSTRACE_ERR, 
                "[PORTAB]CsNetListen Failed because this code is acting "
                "as Client\n") );
        return CS_RET_ACTINGASCLIENT;
    }

    if(g_ActingAsServer == TRUE)
    {
        CsSyncUnlock(g_DataLock);
        CSTRACE( (CSTRACE_ERR, "[PORTAB]Socket Listening already\n") );        
        return CS_RET_ALREADYINLISTEN;
    }

    CsSyncUnlock(g_DataLock);

    /*Create a socket*/
    if((sockFD = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
        CSTRACE( (CSTRACE_ERR, "[PORTAB]socket() function failed\n"));
        return CS_RET_INVALIDHANDLE;                            
    }

    /*Make this a Non Blocking socket*/ 
#ifdef _VXWORKS_
/* test, this seems to be causing some problems
    retval = ioctl(sockFD, FIONBIO, 1);
 */
#else
    retval = fcntl(sockFD, F_SETFL, O_NONBLOCK);
#endif
    if(retval == -1)
    {
        close(sockFD);
        return CS_RET_SOCKETCREATERR;
    }

    CSTRACE( (CSTRACE_INFO, "[PORTAB]Socket value is %d.\n", sockFD));

    if( setsockopt(sockFD, SOL_SOCKET, SO_REUSEADDR, (void *)&reuse, sizeof(reuse)) <0 ) 
    {
        close(sockFD);
        return CS_RET_SOCKETCREATERR;
    }

    /*initialize server addr structure bytes to nulls*/
    bzero((char*) &serv_addr, sizeof(serv_addr));

    /*bind server address*/
    serv_addr.sin_family      = AF_INET;
    serv_addr.sin_addr.s_addr = INADDR_ANY;
    serv_addr.sin_port        = CSHTONS(port);

    if( bind(sockFD, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) <0 )
    {
        CSTRACE( (CSTRACE_ERR, "Server:Can't bind to local address.\n"));
        return CS_RET_LISTENFAILED;
    }

    CSTRACE((CSTRACE_INFO, "[PORTAB]After bind.\n"));
    /*queue upto 5 connection requests while
    current one gets accepted*/
    if(listen(sockFD, MAX_LISTEN_SOCKS) <0 ) 
    {
        CSTRACE((CSTRACE_ERR, "[PORTAB] Listen failed\n"));
        return CS_RET_LISTENFAILED;
    }

    CSTRACE((CSTRACE_INFO, "[PORTAB]listen successful.\n"));

    CsSyncLock(g_DataLock);
    g_ActingAsServer = TRUE;
    g_ListenSocket = sockFD;
    CsSyncUnlock(g_DataLock);

    /* start the listen thread
    handlePtr will be free when listen thread exists*/
    handlePtr = (ULONG *) calloc(1, sizeof(ULONG));

    if( handlePtr == NULL )
    {
        return CS_RET_OUTOFMEMORY;
    }

    *handlePtr = sockFD;

/*
    if(  pthread_create(&g_ListenThreadHandle, NULL, ListenThread, handlePtr) 
        != 0 )
*/
    if ( 0 == (g_ListenThreadHandle = PIL_CreateThread(
                                      ListenThread,
                                      DEFAULT_STACK_SIZE,
                                      NULL,
                                      DEFAULT_THREAD_PRIORITY,
                                      handlePtr,
                                      0)))
    {
        CSTRACE((CSTRACE_INFO, 
               "[PORTAB] could not create listening thread!\n"));

        CloseThisSocket(sockFD);
        free(handlePtr);

        return CS_RET_OUTOFMEMORY;
    }
    CSTRACE((CSTRACE_INFO, "[PORTAB] CsNetListen successful!\n"));

    return CS_RET_SUCCESS;

}/*CsNetListen*/


/******************************************************************************
//
// METHOD:      ULONG CsNetStopListen()
//
// ARGUMENTS:
//
// DESCRIPTION: Stops the Server.
//              
//
// RETURNS:     Returns CS_RET_SUCCESS, if succeeds otherwise returns
//              appropriate error values.
//
******************************************************************************/
ULONG CsNetStopListen(void)
{

    BOOL ActingAsServer = FALSE;
    SOCKET ListenSocket;

    int retval;

    CSTRACE( (CSTRACE_INFO, "[PORTAB]ULONG CsNetStopListen(...\n") );

    if( g_portInit == FALSE )
    {
        CSTRACE((CSTRACE_ERR, "[PORTAB] Need to call CsPortInit() first "
               "before calling this function.\n"));
        return CS_RET_NOTINITIALIZED;
    }

    CsSyncLock(g_DataLock);
    ActingAsServer = g_ActingAsServer;
    g_ActingAsServer = FALSE;
    ListenSocket = g_ListenSocket;
    CsSyncUnlock(g_DataLock);

    if(ActingAsServer == FALSE)
    {
        CSTRACE( (CSTRACE_ERR, "Server Not Listening.\n") );
        return CS_RET_NOSERVER;
    }

    /*delete & close list of all connected clients' socket descriptors*/
    DeleteAndCloseSocketList();
    CSTRACE((CSTRACE_INFO, "[PORTAB] Deleted and closed socket list\n"));

    /*To stop server thread*/
    CsSyncLock(g_DataLock);
    g_ListenSocket = 0;
    CsSyncUnlock(g_DataLock);

    retval = write(g_socketPipe[1],TERMINATE_STR , sizeof(TERMINATE_STR ));

    CSTRACE((CSTRACE_INFO,"WRITE RETURNED %d BYTES\n ", retval));
    if(retval == sizeof(TERMINATE_STR )) 
    {
        CSTRACE( (CSTRACE_INFO, 
                "[PORTAB]Server: Waiting for listening thread\n"));

        PIL_WaitForThread(g_ListenThreadHandle, PIL_INFINITE_TIMEOUT);

        CSTRACE( (CSTRACE_INFO, "[PORTAB]Server: Successfully "
                "Disconnected.\n") );        

        close(ListenSocket);
        return CS_RET_SUCCESS;
    }

    CSTRACE((CSTRACE_ERR, "[PORTAB] Couldn't close Listen Socket\n"));

    close(ListenSocket);
    return CS_RET_UNKNOWN;

}/*CsNetStopListen*/


/******************************************************************************
//
// METHOD:      ULONG CsNetDisconnect ( ULONG handle )
//
// ARGUMENTS:   handle : Socket handle
//
// DESCRIPTION: Disconnect a client from the server.
//              
//
// RETURNS:     Returns CS_RET_SUCCESS, if succeeds otherwise returns
//              appropriate error values.
//              
//
******************************************************************************/
ULONG CsNetDisconnect ( ULONG handle )
{
    SOCKET sockFD = 0;
    PsockStruct curStruct;

    CSTRACE( (CSTRACE_INFO, "[PORTAB]ULONG CsNetDisconnect(...).\n") );

    if( g_portInit == FALSE )
    {
        CSTRACE((CSTRACE_ERR, "[PORTAB] Need to call CsPortInit() first "
               "before calling this function.\n"));
        return CS_RET_NOTINITIALIZED;
    }

    sockFD = ULONGToSOCKET(handle);

    if( (curStruct =  FindSocket(sockFD)) != NULL) 
    {
        /* found the socket  */
        CloseThisSocket(curStruct->sockHandle);
        PIL_WaitForThread(curStruct->threadHandle, PIL_INFINITE_TIMEOUT);

        CSTRACE((CSTRACE_INFO, "[PORTAB] Thread closed successfully!\n"));
	free (curStruct);
        /* delete socket*/
       // DeleteSocket(handle);

    }
    else
    {
        CSTRACE((CSTRACE_ERR, "[PORTAB] Socket not found: %d\n", handle));
        return CS_RET_INVALIDHANDLE;
    }
                                    
    CSTRACE((CSTRACE_INFO, "[PORTAB] Socket Disconnected\n"));
    return CS_RET_SUCCESS;
}/*CsNetDisconnect*/


/******************************************************************************
//
// METHOD:      ULONG CSHTONL( ULONG value )
//
// ARGUMENTS:   value: host byte order
//
// DESCRIPTION:     function converts a u_long from host 
//                  to TCP/IP network byte order.
//              
//
// RETURNS:     Returns the TCP/IP network byte order
//              
//
******************************************************************************/

ULONG CSHTONL( ULONG value )

{
    return htonl(value);
}

/******************************************************************************
//
// METHOD:      ULONG CSNTOHL( ULONG value )
//
// ARGUMENTS:   value : Network order byte.
//
// DESCRIPTION:   function converts a u_long from TCP/IP 
//                network order to host byte order.
//              
//
// RETURNS:     Returns host byte order.
//              
//
******************************************************************************/

ULONG CSNTOHL( ULONG value )

{

    return ntohl(value);

}


/******************************************************************************
//
// METHOD:      USHORT CSHTONS( USHORT value )
//
// ARGUMENTS:   value : host byte order
//
// DESCRIPTION:   function converts a u_short from host 
//                to TCP/IP network byte order.
//              
//
// RETURNS:     Returns network byte order.
//              
//
******************************************************************************/

USHORT CSHTONS( USHORT value )
{
    return htons(value);
}


/******************************************************************************
//
// METHOD:      USHORT CSNTOHS( USHORT value )
//
// ARGUMENTS:   value : network byte order
//
// DESCRIPTION:  function converts a u_short from TCP/IP 
//               network byte order to host byte order.
//              
//
// RETURNS:     Returns host byte order.
//              
//
******************************************************************************/
USHORT CSNTOHS( USHORT value )
{
    return ntohs(value);
}


/******************************************************************************
//
// METHOD:      ULONG CloseThisSocket(SOCKET handle)
//
// ARGUMENTS:   Handle : Socket handle to close the socket
//
// DESCRIPTION: This is internal function, used by TCPIP module
//              top layers should call this function.
//
// RETURNS:     Returns CS_RET_SUCCESS, if succeeds else returns
//              the corresponding error value.
//
******************************************************************************/
ULONG CloseThisSocket(SOCKET handle)
{
    if(shutdown(handle, 2) != 0)
    {
        CSTRACE((CSTRACE_ERR,"Error closing socket %lu.\n",handle));
        switch(errno)
        {
        case EBADF:
            CSTRACE((CSTRACE_ERR, "[PORTAB] Invalid FD\n"));
            break;
        case ENOTSOCK:
            CSTRACE((CSTRACE_ERR, "[PORTAB] FD isn't a socket\n"));
            break;
        case ENOTCONN:
            CSTRACE((CSTRACE_ERR, "[PORTAB] Socket not connected\n"));
            break;
        default:
            CSTRACE((CSTRACE_ERR, "[PORTAB] Unknown error during shutdown\n"));
        }
        return CS_RET_UNKNOWN;
    }
    else
    {
      CSTRACE( (CSTRACE_INFO, 
                "[PORTAB]Successfully closed: socket %lu.\n", 
                handle) );
      
      return CS_RET_SUCCESS;
    }
}/*CloseThisSocket*/

/******************************************************************************
//
// METHOD:      char *CSINET_NTOA(const CS_IPADDR *in)
//
// ARGUMENTS:   in - structure that represents an Internet host address.
//
// DESCRIPTION: function converts an (Ipv4) Internet network address 
//              into a string in Internet standard dotted format.
//              
//
// RETURNS:     returns a char pointer to a buffer containing the text 
//              address in standard ".'' notation.
//              
//
******************************************************************************/
char *CSINET_NTOA(const CS_IPADDR *in)
{
    struct in_addr i;

    i.s_addr = in->addr;

    return inet_ntoa(i);
}

/******************************************************************************
//
// METHOD:      ULONG CSINET_ADDR(char *cp)
//
// ARGUMENTS:   cp - Dotted IP address string.
//
// DESCRIPTION: function converts a string containing an (Ipv4) 
//              Internet Protocol dotted address into a proper address 
//              for the IN_ADDR structure.
//
//              
//
// RETURNS:     returns an unsigned long value containing a suitable 
//              binary representation of the Internet address given.
//              
//
******************************************************************************/
ULONG CSINET_ADDR(char *cp)

{
    return inet_addr(cp);
}

/******************************************************************************
//
// METHOD:      ULONG64 SwapUlong64( ULONG64 InValue )
//
// ARGUMENTS:   InValue - The value to be swapped
//
// DESCRIPTION: function converts a ULONG64 from TCP/IP 
//              network byte order to host byte order and vice versa.
//              
//
// RETURNS:     Returns swapped value.
//              
//
******************************************************************************/
#ifndef CS_BIG_ENDIAN
ULONG64 SwapUlong64(ULONG64 InValue)
{
    ULONG64 OutValue;
    ULONG * IPtr = (ULONG *)&InValue;
    ULONG * OPtr = (ULONG *)&OutValue;

    /*It doesn't matter if this is htonl or ntohl. We just need to swap*/
    OPtr[0] = CSHTONL(IPtr[1]);
    OPtr[1] = CSHTONL(IPtr[0]);
    return OutValue;
}
#endif /*!CS_BIG_ENDIAN*/

/******************************************************************************
//
// METHOD:      ULONG64 CSNTOHL64( ULONG64 value )
//
// ARGUMENTS:   value - network byte order
//
// DESCRIPTION: function converts a ULONG64 from TCP/IP 
//              network byte order to host byte order.
//              
//
// RETURNS:     Returns host byte order.
//              
//
******************************************************************************/

ULONG64 CSNTOHL64( ULONG64 value )
{
#ifdef CS_BIG_ENDIAN
    /*Network & host byte order are the same. Simply return value*/
    return value;
#else
    return SwapUlong64(value);
#endif
}

/******************************************************************************
//
// METHOD:      ULONG64 CSHTONL64( ULONG64 value )
//
// ARGUMENTS:   value - host byte order
//
// DESCRIPTION: function converts a ULONG64 from host byte order to TCP/IP 
//              network byte order.
//              
//
// RETURNS:     Returns network byte order.
//              
//
******************************************************************************/

ULONG64 CSHTONL64( ULONG64 value )
{
#ifdef CS_BIG_ENDIAN
    /*Network & host byte order are the same. Simply return value*/
    return value;
#else
        return SwapUlong64(value);
#endif
}
