/******************************************************************************
//           INTEL CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or
//  nondisclosure agreement with Intel Corporation and may not be copied
//  or disclosed except in accordance with the terms of that agreement.
//      Copyright (c) 1999-2000 Intel Corporation.  All Rights Reserved.
//
// $Workfile:   timer.c  $
// $Revision: 1.6 $
// $Date: 2003/11/17 16:57:56 $
//
// Description:
//     This file contains the implementation of the timer primitives for
//     the COPS portability module. Both the external interface functions
//     and the internal methods are implemented in this file.
//
******************************************************************************/

#include <time.h>
#include <unistd.h>
#include <assert.h>
#include <stdlib.h>
#include "csportability.h"
#include "timer.h"
#include "cserror.h"
#include <errno.h>
#include "pil.h"
#include <signal.h>
#include <stdio.h>
#ifdef _VXWORKS_
#include <sys/times.h>
#include "ioLib.h"
#include "pipeDrv.h"
#else
#include <sys/time.h>
#endif

#define TERMINATE_THREAD 0
#define WAKE_UP_THREAD 1

void (*TimeoutCallback)(ULONG handle);  // Pointer to timer callback
static struct TimerInfo *g_timerList;  // Pointer to timer linked list
static BOOL g_runTimer;                // flag to start/stop timer processing
static BOOL g_initialized;             // flag for initialization state
static PilThread g_timerThread;        // thread for timer processing
static ULONG g_sync;                   // handle to sync object. Any method that
                                // modifies the timer list must be sync'ed
static int g_timerPipe[2];      // timer pipe used to signal a new timer
                                // inserted


/******************************************************************************
//
//  Function prototypes
//
******************************************************************************/
static ULONG AddTimer( ULONG handle, ULONG seconds, TIMER_CALLBACK cbFunc );
static ULONG UpdateTimer( ULONG handle, ULONG seconds, TIMER_CALLBACK cbFunc );
static ULONG RemoveTimer( ULONG handle, BOOL external );
static struct TimerInfo *FindTimer( ULONG handle, BOOL external );
static int InsertTimer( struct TimerInfo *newTimer );
static int DeleteTimer( ULONG handle, BOOL external );
static void FreeTimers( void );
static void LogError( char *info );
static int CheckTimerList(void);
static uint32_t TimerThread(void *dataPtr);

#ifdef _DEBUG
static void PrintTimers( void );
#endif


/******************************************************************************
// Method:  TimerInit
//
// Parameters:
//     CbCsTimeout - Callback to invoke when a timer goes off.
//
// Description:
//     This will initialize the timer module. Allocating space and
//     creating any threads as needed. This routine should only be called
//     once. If it is called a second time it will simply return
//     immediately.
//
// Returns:
//     1 if successful, 0 if unsuccessful.
//
// Comments:
//     Requires: The callback must not be NULL.
//     Promises: The timer module will be completely initialized.
//
******************************************************************************/
int TimerInit( void (*CbCsTimeout)(ULONG handle) ) 
{

    if ( g_initialized ) 
    {
        LogError( "Can't initialized module when already initialized" );
        return 0;
    }

    if ( !CbCsTimeout ) 
    {
        LogError( "Can't initialize with NULL callback" );
        return 0;
    }

    /* Set up callback, list and signal handler. Then create the condition
    variable for timer notification, get a synchronization object and 
    start the timer thread(the thread that will be used to 
    invoke callbacks).*/

    TimeoutCallback = CbCsTimeout;
    g_timerList = NULL;

    if ( CsSyncCreate( &g_sync ) != CS_RET_SUCCESS ) 
    {
        LogError( "Unable to create synchronization object" );
        return 0;
    }

#ifdef _VXWORKS_
    pipeDevCreate("/pipe/Interconnect/timers",10,sizeof(int));
    g_timerPipe[0] = g_timerPipe[1] = open("/pipe/Interconnect/timers",
	O_RDWR, 0666);

    if (g_timerPipe[0] == ERROR)
#else

    if ( pipe( g_timerPipe ) )
#endif
    {
        CSTRACE((CSTRACE_ERR, "[PORTAB]Unable to open timer pipe" ));
        return 0;
    }


    g_runTimer = TRUE;
    if ( (g_timerThread = PIL_CreateThread(
                          TimerThread,
                          DEFAULT_STACK_SIZE,
                          NULL,
                          DEFAULT_THREAD_PRIORITY,
                          NULL,
                          0)) == 0 ) 
    {
        CsSyncLock(g_sync);
        close(g_timerPipe[0]);
        close(g_timerPipe[1]);
        CsSyncUnlock(g_sync);

        CsSyncDestroy(g_sync);
        return 0;
    }

    g_initialized = TRUE;
    return 1;

}/*TimerInit*/


/******************************************************************************
// Method:  TimerDeinit
//
// Parameters:
//     None
//
// Description:
//     This routine is used to deinitialize the timer module. It will join
//     any threads and deallocate any memory as needed. It will also
//     stop all current timers without signalling them first. In other
//     words, if there are active timers when this is called, they will
//     never go off. This routine may only be called if the timer module
//     has already been initialized. If the timer module hasn't been
//     initialized, this routine will do nothing.
//
// Returns:
//     1 if successful, 0 otherwise.
//
// Comments:
//     Requires: The module must have been previously initialized.
//     Promises: The module will be completely cleaned up, threads 
//               ended, memory deallocated, timers turned off.
//
******************************************************************************/
int TimerDeinit( void ) 
{

    int message = TERMINATE_THREAD;

    if ( !g_initialized ) 
    {
        LogError( "Can't deinitialize uninitialized module" );
        return 0;
    }

    CSTRACE((CSTRACE_INFO, "[PORTAB] TimerDeinit: Closing timer thread...\n"));

    CsSyncLock( g_sync );

    write(g_timerPipe[1], (char*)&message, sizeof(int));

    FreeTimers();
    g_runTimer = FALSE;

    if ( PIL_WaitForThread( g_timerThread, PIL_INFINITE_TIMEOUT ) ) 
    {
        LogError( "Unable to join timer callback thread" );
    }

    CSTRACE((CSTRACE_INFO, "[PORTAB] Timer thread closed successfully.\n"));

    if ( close( g_timerPipe[0] ) ) 
    {
        CSTRACE((CSTRACE_ERR, 
                 "[PORTAB]Unable to close read end of timer pipe" ));
    }

    if ( close( g_timerPipe[1] ) ) 
    {
        CSTRACE((CSTRACE_ERR, 
                 "[PORTAB]Unable to close write end of timer pipe" ));
    }


    CSTRACE((CSTRACE_INFO, "[PORTAB] Pipe closed successfully.\n"));

    g_initialized = FALSE;
    CsSyncUnlock( g_sync );

    CSTRACE((CSTRACE_INFO, "[PORTAB] Trying to delete lock\n"));
    if ( CsSyncDestroy( g_sync ) != CS_RET_SUCCESS ) 
    {
        LogError( "Unable to destroy synchronization object" );
    }
    CSTRACE((CSTRACE_INFO, "[PORTAB] Timer Deinit Done\n"));
    return 1;
}


/******************************************************************************
// Method:  AddTimer
//
// Parameters:
//     handle - The handle of the new timer to add.
//     seconds - The number of seconds until the alarm should go off.
//     cbFunc - callback function to be called when timer expires
//
// Description:
//     Add a new timer to the list of pending timers. This method will 
//     be called if the handle provided is not already in the list. The
//     new timer will be added to the list in ascending order by alarm 
//     time.
//
// Returns:
//     CS_RET_SUCCESS if the call completes successfully.
//     CS_RET_OUTOFMEMORY if memory cannot be allocated for the new 
//     timer.
//
// Comments:
//     Requires: The handle must not be in the list already.
//               The number of seconds must be greater than 0.
//     Promises: The new timer will be added to the list in order.
//
******************************************************************************/
ULONG AddTimer( ULONG handle, ULONG seconds, TIMER_CALLBACK cbFunc ) 
{

    struct TimerInfo *newTimer;
    int message = WAKE_UP_THREAD;

    /* Before doing anything to the timer list first turn off the running
     timer. Then allocate a new timer and insert it into the list.*/

    newTimer = malloc( sizeof( struct TimerInfo ) );
    if ( !newTimer ) 
    {
        LogError( "Unable to allocate memory for new timer in AddTimer" );
        return CS_RET_OUTOFMEMORY;
    }
    newTimer->handle = handle;
    newTimer->time = time( NULL ) + seconds;
    newTimer->cbFunc = cbFunc;
    InsertTimer( newTimer );
    
    write(g_timerPipe[1], (char*)&message, sizeof(int));

    return CS_RET_SUCCESS;
}


/******************************************************************************
// Method:  UpdateTimer
//
// Parameters:
//     handle - The handle of the timer to update.
//     seconds - The number of seconds until the alarm should go off.
//     cbFunc - callback function that is called when the timer expires.
//
// Description:
//     Update a timer in the list of pending timers. This method will only
//     be called if the handle provided is already in the list. The
//     existing timer will have its alarm time changed to match the new
//     time and its entry will be moved in the list such that the list
//     maintains its ascending order by alarm time.
//
// Returns:
//     CS_RET_SUCCESS if the call completes successfully.
//     CS_RET_OUTOFMEMORY if memory cannot be allocated for the new timer.
//
// Comments:
//     Requires: The handle must already be in the list.
//               The number of seconds must be greater than 0.
//     Promises: The existing timer will be updated.
//               The timer list will maintain its order.
//
******************************************************************************/

ULONG UpdateTimer( ULONG handle, ULONG seconds, TIMER_CALLBACK cbFunc ) 
{

    struct TimerInfo *newTimer;
    int message = WAKE_UP_THREAD;
    
    /* Before doing anything to the timer list first turn off the running
     timer. Then allocate a new timer to replace the original, delete the
     original and insert the replacement.*/

    newTimer = malloc( sizeof( struct TimerInfo ) );
    if ( !newTimer ) 
    {
        LogError( "Unable to allocate memory for new timer in UpdateTimer" );
        return CS_RET_OUTOFMEMORY;
    }

    if (!cbFunc)
        DeleteTimer( handle, FALSE );
    else
        DeleteTimer( handle, TRUE );

    newTimer->handle = handle;
    newTimer->time = time( NULL ) + seconds;
    newTimer->cbFunc = cbFunc;

    InsertTimer( newTimer );

    write(g_timerPipe[1], (char*)&message, sizeof(int));

    return CS_RET_SUCCESS;
}


/******************************************************************************
// Method:  RemoveTimer
//
// Parameters:
//     handle - The handle of the timer to remove.
//     external - The flag to indicate whether to look for a handle
//     created by CsTimerSetEx or passed in by CsTimerSet.
//
// Description:
//     Given a handle to a timer, search thru the list and find it and
//     delete it. If the timer to be deleted is the one that is currently
//     being waited on, the timer will not go off.
//
// Returns:
//     CS_RET_SUCCESS if the call completes successfully.
//     CS_RET_INVALIDHANDLE if the timer handle to remove could not be found.
//
// Comments:
//     Requires: The handle must be in the list.
//     Promises: If the handle was in the list it and it was created by the
//               appropriate TimerSet function, it will be removed.
//               The timer corresponding to that handle will not go off.
//               Any memory allocated for the timer will be freed.
//               The list will maintain its order.
//
******************************************************************************/

ULONG RemoveTimer( ULONG handle, BOOL external ) 
{

    /* Before doing anything to the timer list first turn off the running
     timer. Then attempt to delete the timer. Finally, if there are still
     timers in the list, restart the interval timer. */
    
    if ( !DeleteTimer( handle, external ) ) 
    {
        return CS_RET_INVALIDHANDLE;
    }
 
   return CS_RET_SUCCESS;
}


/******************************************************************************
// Method:  FindTimer
//
// Parameters:
//     handle - The handle to find.
//     external - flag to indicate whether the timer was created by 
//                CsTimerSet or CsTimerSetEx
// Description:
//     Used to search thru the list to find a timer that matches the handle
//     provided. If a match is found(there can be only one), return a
//     pointer to that timer.
//
// Returns:
//     A pointer to the timer that matches the handle if successful, a NULL
//     pointer otherwise.
//
// Comments:
//     Requires: Nothing
//     Promises: If the timer is found a pointer to it will be returned.
//               The list will not be modified.
//
******************************************************************************/

struct TimerInfo *FindTimer( ULONG handle, BOOL external ) 
{

    struct TimerInfo *tempPtr = g_timerList;
    while ( tempPtr ) 
    {
        if (( tempPtr->handle == handle ) &&
            ( (external && (tempPtr->cbFunc)) ||
              (!external && !(tempPtr->cbFunc)) ))
        {
            return tempPtr;
        }
        else 
        {
            tempPtr = tempPtr->next;
        }
    }
    return NULL;
}

/******************************************************************************
// Method:  PrintTimers
//
// Parameters:
//     None
//
// Description:
//     Convenience routine provided to traverse the list and print out the
//     handle of each timer in the list. Useful for debugging.
//
// Returns:
//     None
//
// Comments:
//     Requires: Nothing
//     Promises: The list will not be modified.
//
******************************************************************************/
#ifdef _DEBUG

void PrintTimers( void ) 
{

    struct TimerInfo *tempPtr = g_timerList;

    while ( tempPtr ) 
    {
        printf( "%ld ", tempPtr->handle );
        tempPtr = tempPtr->next;
    }
    printf( "\n" );
}
#endif /*_DEBUG*/


/******************************************************************************
// Method:  FreeTimers
//
// Parameters:
//     None
//
// Description:
//     This will traverse the list, freeing each timer on the list. None
//     of the timers will be allowed to fire while being deleted.
//
// Returns:
//     None
//
// Comments:
//     Requires: The interval timer must be turned off prior to calling
//               this routine.
//     Promises: The list will be emptied.
//               All memory will be deallocated.
//
******************************************************************************/

void FreeTimers( void ) 
{

    struct TimerInfo *tempPtr = g_timerList;
    struct TimerInfo *target = NULL;

    while ( tempPtr ) 
    {
        target = tempPtr;
        tempPtr = target->next;
        free( target );
    }
    g_timerList = NULL;
}


/******************************************************************************
// Method:  InsertTimer
//
// Parameters:
//     timer - a pointer to the new timer to add.
//
// Description:
//     Used to insert a new timer into the list while maintaining the order
//     of the list. The list must be kept in ascending order by alarm time.
//
// Returns:
//     1 if successful, 0 otherwise.
//
// Comments:
//     Requires: The parameter must not be NULL.
//     Promises: The new timer will be added to the list.
//               The list will maintain its order.
//
******************************************************************************/

int InsertTimer( struct TimerInfo *timer ) 
{

    struct TimerInfo *tempPtr = g_timerList;
    struct TimerInfo *backPtr = NULL;
    BOOL done = FALSE;

    if ( !timer ) 
    {
        return 0;
    }

    timer->next = NULL;

    while ( tempPtr && !done ) 
    {
        assert( timer->handle != tempPtr->handle );
        if ( tempPtr->time < timer->time ) 
        {
            backPtr = tempPtr;
            tempPtr = tempPtr->next;
        }
        else 
        {
            timer->next = tempPtr;
            done = TRUE;
        }
    }
    
    if ( tempPtr == g_timerList ) 
    {
        g_timerList = timer;
    }
    
    if ( backPtr ) 
    {
        backPtr->next = timer;
     }
    return 1;
}


/******************************************************************************
// Method:  DeleteTimer
//
// Parameters:
//     handle - The handle of the timer to delete.
//     external - flag to indicate whether the timer was created by the COPS 
//                layer or some other layer.
// Description:
//     This method will search for the given handle in the list of pending
//     timers and will remove it from the list and also free its memory.
//
// Returns:
//     1 if successful, 0 otherwise.
//
// Comments:
//     Requires: Nothing
//     Promises: If the handle is found that timer will be removed from the
//               list.
//               Any memory allocated for the timer will be deallocated.
//               The list will maintain its order.
//
******************************************************************************/

int DeleteTimer( ULONG handle, BOOL external ) 
{
    struct TimerInfo *tempPtr = g_timerList;
    struct TimerInfo *backPtr = NULL;
    struct TimerInfo *target = NULL;

    while ( tempPtr && !target ) 
    {
        /*handles should match AND (if external, cbFunc should be valid OR
          if internal, cbFunc should be NULL). In English: Check if the handle
          was generated by CsTimerSetEx or CsTimerSet*/
        if (( tempPtr->handle == handle ) &&
            ( (external && (tempPtr->cbFunc)) ||
              (!external && !(tempPtr->cbFunc)) ))
        {
            target = tempPtr;
        }
        else 
        {
            backPtr = tempPtr;
            tempPtr = tempPtr->next;
        }
    }

    if ( target ) 
    {
        if ( target == g_timerList ) 
        {
            g_timerList = target->next;
        }
        if ( backPtr ) 
        {
            backPtr->next = target->next;
        }
        free( target );
        return 1;
    }
    return 0;
}

/******************************************************************************
// Method:  LogError
//
// Parameters:
//     info - the error string to print
//
// Description:
//     Convenience routine to print out errors to a log file.
//
// Returns:
//     None
//
// Comments:
//     Requires: Nothing
//     Promises: The info provided will be printed with the log messages.
//
******************************************************************************/

void LogError( char *info ) 
{

    CSTRACE((CSTRACE_ERR, "[Portab] %s\n", info));
}


/******************************************************************************
// Method: TimerThread 
//
// Parameters:
//     dataPtr    This parameter is not used. It has been given just for
//                  conformance.
//
// Description:
//     This is where the timer callback thread runs. It waits for timers
//     to go off and will then invoke the timer callback with the timer
//     handle as a parameter.
//
// Returns:
//     TRUE if processing should continue, FALSE otherwise.
//
// Comments:
//     Requires: Nothing
//     Promises: Will invoke callbacks on any signalled timers.
//               Will stop processing signals when required.
//
******************************************************************************/

uint32_t TimerThread(void *dataPtr ) 
{

    int alarmTime;
    fd_set rfds;
    int highestFD;
    int retval;
    int data;
    int iTime;
    struct timeval sleeptime;

    while (1)
    {

        alarmTime = CheckTimerList();

        /* calculate the amount of time to sleep */

        CSTRACE((CSTRACE_INFO, 
                "[PORTAB] CheckTimerList() returned %d.\n", alarmTime));

        if ( alarmTime == 0 )
        {
            /* no pending timer, sleep indefinitely */
            FD_ZERO(&rfds);
            FD_SET(g_timerPipe[0], &rfds);
            highestFD = g_timerPipe[0] + 1;

            CSTRACE((CSTRACE_INFO, "[PORTAB] alarmTime is 0, sleep "
                     "until a timer is set.\n"));

            retval = select( highestFD, &rfds, NULL, NULL, NULL);

            CSTRACE((CSTRACE_INFO, "[PORTAB] wake up from indefinite "
                     "sleep!\n"));

            if ( retval )
            {
                /* there is data to be read */
                read(g_timerPipe[0], (char*)&data, sizeof(data));
                if ( data == TERMINATE_THREAD )
                {
                    CSTRACE((CSTRACE_INFO, 
                             "[PORTAB] read TERMINATE_THREAD "
                             "from the pipe, exiting..\n"));
                    PIL_ExitThread(0);
                }
            }
            else
            {
                /* error! pipe closed?  exit thread */
                CSTRACE((CSTRACE_INFO, "[PORTAB] select return %d, "
                         "timer thread existing...\n", retval));
                PIL_ExitThread(0);                
            }/*if(retval)*/

        }
        else
        {
            /* there are timers, sleep until the next timer expires */

            iTime = alarmTime - time(NULL);

            if ( iTime > 0 )
            {
                FD_ZERO(&rfds);
                FD_SET(g_timerPipe[0], &rfds);
                highestFD = g_timerPipe[0] + 1;

                sleeptime.tv_sec = iTime;
                sleeptime.tv_usec = 0;

                CSTRACE((CSTRACE_INFO, "[PORTAB] TimerThread sleep for %d"
                     " sec.\n", iTime));

                retval = select( highestFD, &rfds, NULL, NULL, &sleeptime);

                CSTRACE((CSTRACE_INFO, "[PORTAB] TimerThread wake up!\n"));

                if ( retval )
                {
                    /* there is data to be read */
                    read(g_timerPipe[0], (char*)&data, sizeof(data));
                    if ( data == TERMINATE_THREAD )
                    {
                        CSTRACE((CSTRACE_INFO, 
                                 "[PORTAB] read TERMINATE_THREAD "
                                 "from the pipe, exiting..\n"));
                        PIL_ExitThread(0);
                    }
                }
                else
                {
                    CSTRACE((CSTRACE_ERR,"PORTAB: Timer Expired\n"));
                }/*if(retval)*/

            }/*if(iTime > 0)*/

        }/*if(alarmTimer == 0)*/

    }/*while(1)*/

}


/******************************************************************************
// Method:  CheckTimerList
//
// Parameters:
//     None
//
// Description:
//     This is where the timer callback is invoked. Before the callback
//     is invoked its associated timer is removed from the timer list.
//
// Returns:
//     None
//
// Comments:
//     Requires: Nothing
//     Promises: Will invoke callbacks on any expired timers.
//
******************************************************************************/

int CheckTimerList(void) 
{

    ULONG activeHandle;
    time_t currentTime = time( NULL );
    BOOL over = FALSE;
    struct TimerInfo *tempPtr = g_timerList;
    TIMER_CALLBACK ExtTimeoutCallback = NULL;
    BOOL external = FALSE;
    
    /* Only need to traverse the list until the alarm time of the current 
     list element is past the current time.
     For each timer that has expired remove the timer from the list and
     signal the callback thread with the handle of the expired timer.*/
    
    while ( tempPtr && !over ) 
    {
        if ( tempPtr->time <= currentTime ) 
        {
            activeHandle = tempPtr->handle;
            CsSyncLock( g_sync );

            if (!(tempPtr->cbFunc))
            {
                DeleteTimer( activeHandle, FALSE );
                external = FALSE;
            }
            else
            {
                DeleteTimer( activeHandle, TRUE );
                external = TRUE;
                ExtTimeoutCallback = tempPtr->cbFunc;
            }

            tempPtr = tempPtr->next;
            CsSyncUnlock( g_sync );

            if (external)
                ExtTimeoutCallback( activeHandle );
            else
                TimeoutCallback( activeHandle );
        }
        else 
        {
            over = TRUE;
        }
    }

    if ( g_timerList ) 
    {
        return g_timerList->time;
    }
    
    return 0;

}


/******************************************************************************
// Method:  CsTimerSet
//
// Parameters:
//     handle - handle to the timer object
//     seconds - number of seconds until the timer expires
//
// Description:
//     A timer object is created by this function and after the expiration
//     period the callback function CbCsTimerCallback is called. This function
//     can also be used to set a timer for a different time(in order to do
//     this, the function must be called before the timer goes off). The
//     handle is provided by the caller and must be unique for each different
//     timer.
//
// Returns:
//     CS_RET_SUCCESS if successful, CS_RET_INVALIDHANDLE otherwise.
//
// Comments:
//     Requires: Nothing
//     Promises: If the handle is associated with an existing timer that
//               timer's time will be updated.
//               If the handle does not match an existing timer a new
//               timer will be created.
//
******************************************************************************/

ULONG CsTimerSet( ULONG handle, ULONG seconds ) 
{

    int res;

    /*This is the API call used by COPS to set a timer. First determine if
     the timer is being created or reset and then perform whichever
     action is needed.*/
    


    CsSyncLock( g_sync );
    if ( FindTimer( handle, FALSE ) ) 
    {
        res = UpdateTimer( handle, seconds, NULL );
    }
    else 
    {
        res = AddTimer( handle, seconds, NULL );
    }
    
    CsSyncUnlock( g_sync );
    return res;
}


/******************************************************************************
// Method:  CsTimerSetEx
//
// Parameters:
//     handle - handle returned for the timer object
//     seconds - number of seconds until the timer expires
//     cbFunc - callback function that should be called when the timer expires
//
// Description:
//     A timer object is created by this function and after the expiration
//     period the callback function CbCsTimerCallback is called. This function
//     can also be used to set a timer for a different time(in order to do
//     this, the function must be called before the timer goes off). A unique
//     handle is generated if the timer doesnt exist and passed back.
//
// Returns:
//     CS_RET_SUCCESS if successful, CS_RET_INVALIDHANDLE otherwise.
//
// Comments:
//     Requires: handle should be a valid ULONG *
//     Promises: If the handle is associated with an existing timer that
//               timer's time will be updated.
//               If the handle does not match an existing timer a new
//               timer will be created.
//
******************************************************************************/

ULONG CsTimerSetEx( ULONG *handle, ULONG seconds, TIMER_CALLBACK cbFunc ) 
{
    int res;
    UCHAR *temp = NULL;

    /*This is the API call used by layers other than COPS to set a 
     timer. First determine if the timer is being created or reset and 
     then perform whichever action is needed.*/

    if (!handle)
        return CS_RET_INVALIDARGS;

    CsSyncLock( g_sync );
    if ( FindTimer( *handle, TRUE ) ) 
    {
        res = UpdateTimer( *handle, seconds, cbFunc );
    }
    else 
    {
        /* malloc to generate a unique handle  
         * will be freed when timer is deleted */
        temp = malloc(sizeof(UCHAR));
        *handle = (ULONG)temp;        
        res = AddTimer( *handle, seconds, cbFunc );
    }
    
    CsSyncUnlock( g_sync );
    return res;
}

/******************************************************************************
// Method:  CsTimerDelete
//
// Parameters:
//     handle - handle to the timer object to delete.
//
// Description:
//     The timer matching the handle passed in will be deleted.
//
// Returns:
//     CS_RET_SUCCESS if successful, CS_RET_INVALIDHANDLE otherwise.
//
// Comments:
//     Requires: Nothing
//     Promises: If the handle is found that timer will be deleted.
//               Any memory allocated for the timer will be deallocated.
//
******************************************************************************/

ULONG CsTimerDelete( ULONG handle ) 
{

    int res;

    /*
     This is the API call used by COPS to set a timer. Attempt to remove
     the timer specified.*/
    
    CsSyncLock( g_sync );
    res = RemoveTimer( handle, FALSE );
    CsSyncUnlock( g_sync );
    return res;
}


/******************************************************************************
// Method:  CsTimerDeleteEx
//
// Parameters:
//     handle - handle to the timer object to delete.
//
// Description:
//     The timer matching the handle passed in will be deleted.
//
// Returns:
//     CS_RET_SUCCESS if successful, CS_RET_INVALIDHANDLE otherwise.
//
// Comments:
//     Requires: Nothing
//     Promises: Will Look for timer created with CsTimerSetEx only. 
//               If the handle is found that timer will be deleted.
//               Any memory allocated for the timer will be deallocated.
//
******************************************************************************/

ULONG CsTimerDeleteEx( ULONG handle ) 
{
    int res;    
 
    CsSyncLock( g_sync );
    res = RemoveTimer( handle, TRUE );
    CsSyncUnlock( g_sync );
    return res;
}
