/*****************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 1998-2003 By Intel Corporation.  All rights reserved.  
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *****************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

#include "list.h"
#include "stack.h"
#include "types.h"

#define MAX_ARGS 128

extern FILE* yyin;

int g_verbose = 0;

#ifdef TRACE_ON
#define TRACE(s,args...) fprintf(stderr,s,##args);
#else
#define TRACE(s,args...)
#endif

#define check_opts(index, total_num) \
    ( ( ((index)+1) < (total_num) ) &&   (strcmp(argv[i+1],"--infile"))&&  \
                               (strcmp(argv[i+1],"--prefix"))&&  \
                               (strcmp(argv[i+1],"--suffix"))&&  \
                               (strcmp(argv[i+1],"--suffix"))&&  \
                               (strcmp(argv[i+1],"--cppoptions"))&&  \
                               (strcmp(argv[i+1],"--verbose"))&&  \
                               (strcmp(argv[i+1],"--help")) )?1:0



extern void writeheader(FILE* f);

FILE* openfile(char* infile, char* prefix, char* root, char* suffix, char* ext)
{
    FILE* f;
    char tmp[128];

    if (!root) return stdout;

    if (!prefix && !suffix)
    {
        f = stdout;
    }
    else
    {
        snprintf(tmp,127,"%s%s%s.%s",prefix?prefix:"",root,suffix?suffix:"",ext);
        f = fopen(tmp,"w");
    }
    writeheader(f);

    if ('h' == *ext)
    {
        if (infile)
            snprintf(tmp,127,"#include \"%s\"\n\n",infile);
        else
            snprintf(tmp,127,"#include \"%s.h\"\n\n",root);
    }
    else
        snprintf(tmp,127,"#include \"%s%s%s.%s\"\n\n",prefix?prefix:"",root,suffix?suffix:"","h");

    fprintf(f,tmp);

    return f;
}

int usage()
{
    printf("usage:\n"
           "\t--infile <filename>   input file. reads from stdin if not set\n"
           "\t--cppoptions \"<options>\"   options to pass to cpp\n"
           "\t--prefix <pre>        prepended to output files\n"
           "\t--suffix <suf>        appended to output files\n"
           "\t\tif suffix and prefix are not set, output is sent to stdout\n"
           "\t--verbose             print some extra messages\n"
           "\t--help                print this\n");
    return 0;
}

void closefile(FILE* f)
{
    if (f!=stdin && f!=stdout) fclose(f);
}

int parseopts(char* cmdopts[], int i, char* o)
{
    char* options;
    char* arg;
    options = strdup(o);
    TRACE("i:%d : options:%s\n",i,options);
    arg = strtok(options," \t");
    while (arg)
    {
        TRACE("arg:%s\n",arg);
        cmdopts[i++] = arg;
        arg = strtok(NULL," \t");
    }
    TRACE("i:%d\n",i);
    return i;
}

FILE* preprocess(char* filename, char* options)
{
    int p[2];

    if (-1 == pipe(p)) return 0;

    if (!fork()) /* child */
    {
        int numopts,i;
        char* cmdopts[MAX_ARGS];
        char* arg;

        close(p[0]);
        close(STDOUT_FILENO);
        dup(p[1]);

        cmdopts[0] = "cpp";

        numopts = parseopts(
                    cmdopts,
                    1,
                    /*"-C -P -I- -nostdinc -nostdinc++");*/
                    "-C -I- -nostdinc -nostdinc++");

        if (options)
            numopts = parseopts(cmdopts,numopts,options);

        cmdopts[numopts++] = filename;
        cmdopts[numopts] = 0;

        i = 0;
        while (cmdopts[i])
        {
            TRACE("%d>%s\n",i,cmdopts[i]);
            i++;
        }
        execvp("cpp", cmdopts);
    }
    else /* parent */
    {
        close(p[1]);
        return fdopen(p[0],"r");
    }
}

int main (int argc, char* argv[])
{
    int   i;
    char* inputfilename = NULL;
    char* outputext = NULL;
    char* prefix = NULL;
    char* suffix = NULL;
    char* cppoptions = NULL;
    char  outputfilename[128];

    FILE* inputfile = NULL;
    FILE* outputfile = NULL;

    /* If no command line options passed print help */
    if (argc == 1)
    {
         printf("\n Invoke TPCodeGen with proper command options !!! \n");
         printf(" ------------------------------------------------ \n");
         return usage();
    }

    for (i=1; i<argc; i++)
    {
       if ( strstr(argv[i],"--infile") )
       {
         if ( check_opts(i, argc) )
         {
           inputfilename = strdup(argv[i+1]);
           printf("inputfilename = %s\n",inputfilename);
         }
       }
       else if ( strstr(argv[i],"--prefix") )
       {
         if ( check_opts(i, argc) )
         {
            prefix = strdup(argv[i+1]);
            printf("prefix = %s\n",prefix);
         }
       }
       else if ( strstr(argv[i],"--suffix") )
       {
         if ( check_opts(i, argc) )
         {
            suffix = strdup(argv[i+1]);
            printf("suffix = %s\n",suffix);
         }
       }
       else if ( strstr(argv[i],"--cppoptions") )
       {
         if ( check_opts(i, argc) )
         {
            cppoptions = strdup(argv[i+1]);
            printf("cppoptions = %s\n",cppoptions);
         }
       }
       else if ( strstr(argv[i],"--help") )
       {
         printf("argv[%d] = %s\n",i,strstr(argv[i],"--help"));
         return usage();
       }
       else if ( strstr(argv[i],"--verbose") )
       {
          g_verbose = 1;
          printf("SET VERBOSE FLAG :g_verbose = %d \n", g_verbose);
       }
       else 
       {
          ; /* Continue */
       }
    }

    yyin = preprocess(inputfilename,cppoptions);

    inittypes();
    initfunctions();
    if (yyparse()) return 0;

    if (g_verbose)
    {
        printalltypes();
        printallfunctions();
    }

    outputfile = openfile(inputfilename,prefix,"encap",suffix,"c");
    writeallencaps(outputfile,0);
    closefile(outputfile);

    outputfile = openfile(inputfilename,prefix,"encap",suffix,"h");
    writeallencaps(outputfile,1);
    closefile(outputfile);

    outputfile = openfile(inputfilename,prefix,"decap",suffix,"c");
    writealldecaps(outputfile,0);
    closefile(outputfile);

    outputfile = openfile(inputfilename,prefix,"decap",suffix,"h");
    writealldecaps(outputfile,1);
    closefile(outputfile);
}
