/*****************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 1998-2003 By Intel Corporation.  All rights reserved.  
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *****************************************************************************/

/*****************************************************************************
*                                                                            *
*  -------------------------------- list.h --------------------------------  *
*                                                                            *
*****************************************************************************/

#ifndef _LIST_H_
#define _LIST_H_

#ifdef __cplusplus
extern "C" {
#endif

#include "pil.h"

// Define a structure for doubly-linked list elements, not shown by outside.
typedef struct DListElmt_ {
    void               *data;
    struct DListElmt_  *prev;
    struct DListElmt_  *next;
} DListElmt;

// Define a structure for doubly-linked lists. 
typedef struct DList_ {
    int                size;
    PilCriticalSection lock;

    int                (*match)(const void *key1, const void *key2);
    int                (*compare)(const void *data, const void *key);
    void               (*destroy)(void *data);

    DListElmt          *head;
    DListElmt          *tail;
} DList;

// Define a structure for doubly-linked list iterator
typedef struct DListItr_ {
    DList	*list;
    DListElmt	*current;
} DListItr;

/*****************************************************************************
*                                                                            *
*  --------------------------- Public Interface ---------------------------  *
*                                                                            *
*****************************************************************************/

/*
 *    User allocates the memory of list and the data, may let LIST destroy
 *    the memory of the data for user by using removeFree/popFree
 *
 *    Example: Create a list and push one data then pop out
 *             DList pipeline; struct segment *seg; void *data;
 *             npf_list_init(&pipeline, free);
 *             seg = (struct segment *)malloc(sizeof(struct segment));
 *             npf_list_pushBack(&pipeline, seg);
 *             ...
 *             npf_list_popFront(&pipeline, (void **)&data);
 *             if (list->destroy != NULL) list->destroy(data); 
 */

#define npf_list_size(list) ((list)->size)
extern void npf_list_init(DList *list, void (*destroy)(void *data));
extern void npf_list_destroy(DList *list);	// Remove and free data

extern int npf_list_pushFront(DList *list, const void *data);
extern int npf_list_pushBack(DList *list, const void *data);
extern int npf_list_popFront(DList *list, void **data);
extern int npf_list_popBack(DList *list, void **data);
extern int npf_list_popFrontFree(DList *list);
extern int npf_list_popBackFree(DList *list);
extern int npf_list_isExist(DList *list, const void *data);
extern void *npf_list_getFirstData(DList *list);
extern void *npf_list_getLastData(DList *list);

/*
 *    User needs call ItrCreate before use and ItrDelete after done
 *    For thread safety, Iterator operations should be protected in the user 
 *    defined lock
 *
 *    Example: Insert a new DATA into the third position of the list
 *             Here does not do error check, i.e. assume list size >= 2
 *             Functions return -1 if error, like no next element.
 *
 *             DListItr *itr;
 *             itr = npf_list_itrCreate(&list);
 *             LOCK();
 *             npf_list_itrFirst(itr);
 *             npf_list_itrNext(itr);
 *             npf_list_itrInsNext(itr, DATA);
 *             UNLOCK();
 *             npf_list_itrDelete(itr);
 *
 *    Example: Traverse the List
 *
 *             DListItr *itr;
 *             if ((itr = npf_list_itrCreate(&list)) != NULL) {
 *                 npf_list_itrFirst(itr);
 *                 do { ... } while (npf_list_itrNext(itr) != -1)
 *                 npf_list_itrDelete(itr);
 *             }
 *
 *    npf_list_itrCreate        - Create a undefined itr, ret NULL if size 0
 *    npf_list_itrDelete        - Delete a iterator
 *    npf_list_itrFirst		- Get first iterator
 *    npf_list_itrLast		- Get last iterator
 *    npf_list_itrNext		- Move to the next iterator
 *    npf_list_itrPrev		- Move to the previous iterator
 *    npf_list_itrInsNext	- Insert one after and itr is next to new one
 *    npf_list_itrInsPrev	- Insert one before and itr is next to new
 *    npf_list_itrRemove	- Remove and iterator move to one after removed
 *    npf_list_itrRemoveFree	- Remove and Free the data, itr move to next
 *    npf_list_itrGetData	- Return the void *ptr to the user data
 */
#define npf_list_itrListSize(itr) (npf_list_size((itr)->list))
extern DListItr *npf_list_itrCreate(DList *list);
extern int npf_list_itrDelete(DListItr *itr);
extern int npf_list_itrFirst(DListItr *itr);
extern int npf_list_itrLast(DListItr *itr);
extern int npf_list_itrNext(DListItr *itr);
extern int npf_list_itrPrev(DListItr *itr);
extern int npf_list_itrInsNext(DListItr *itr, const void *data);
extern int npf_list_itrInsPrev(DListItr *itr, const void *data);
extern int npf_list_itrRemove(DListItr *itr, void **data);
extern int npf_list_itrRemoveFree(DListItr *itr);
extern void *npf_list_itrGetData(DListItr *itr);

/*
 *    Map functions - Map is the list contains the key.
 *    User should define their own list->compare functions to return 1
 *    if the input void*key matches the data->key
 */
extern int npf_list_removeByKey(DList *list, const void *key, void **data);
extern int npf_list_removeByKeyFree(DList *list, const void *key);
extern void *npf_list_searchByKey(DList *list, const void *key);


/*****************************************************************************
*                                                                            *
*  --------------------------- Private Interface --------------------------  *
*                                                                            *
*****************************************************************************/ 

/*
 *    DListElmt is not shown by outside
 */

int npf_list_insNext(DList *list, DListElmt *element, const void *data);
int npf_list_insPrev(DList *list, DListElmt *element, const void *data);
int npf_list_remove(DList *list, DListElmt *element, void **data);
int npf_list_removeFree(DList *list, DListElmt *element);

#define npf_list_head(list) ((list)->head)
#define npf_list_tail(list) ((list)->tail)
#define npf_list_isHead(element) ((element)->prev == NULL ? 1 : 0)
#define npf_list_isTail(element) ((element)->next == NULL ? 1 : 0)
#define npf_list_data(element) ((element)->data)
#define npf_list_next(element) ((element)->next)
#define npf_list_prev(element) ((element)->prev)

#ifdef __cplusplus
}
#endif

#endif /* _LIST_H_ */

/* ------------------------------- Todo --------------------------------  

  Race Condition still exists at itrInsNext, itrInsPrev, itrRemove, 
  itrRemoveFree.
  The pass-in arguments have a DListItr which is obtaines before the Lock.

  Should let user define their own match function and pass-in the function
  pointer also the user_context. Have a traverse func inside to run
  the match func to remove the matched item.
  Do not delete by order number, it can cause race condition.

  Insert just append to the end/start of the list, or insert before/after
  a item which is selected by user defined match func.

  Q: does this list allow the duplicates item, if yes then user should
     take the response that this won't happen or they don't care ?

  Note :

  User RESPONSE for the additional lock if required.

  For example, user make a npf_list call to retrieve an information, like
  the size of the list, or the head/tail of the list, or the pre/next node
  of a giving node; then based on the return value to make another npf_list
  call like insert or remove.  In this case user should response for lock
  the whole transactions to prevent race condition.

  User do not need to lock in the case that only one npf_list function is
  called and there is no subsequence npf_list call(s) based on the result
  of the first npf_list call.

  npf_list read call is protected by rwlock internelly so user do not need 
  to lock it, however user can still take advantage of rwlock outside if 
  needed.

*/
