/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: BASEOBJ.H                                                 //
// Module:   Platform Independent Layer Base Object declarations       //
// Author:   Dan Baumberger                                            //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 2000 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////

#ifndef __BASEOBJ_H_INCLUDED__
#define __BASEOBJ_H_INCLUDED__

#include "llist.h"

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
//                                                                   //
//                     C O M M O N  T H R E A D S                    //
//                                                                   //
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

typedef struct PilBaseThread_s {
    LIST_NODE Link;         // THIS MUST BE FIRST!  This is used to link
                            // the thread structures together.  It if is
                            // not first, the linked list macros will not
                            // work!
    char      tag[ 4 ];     // The tag value for the handle.  It should
                            // contain 'THRD'.
    uint32_t  Value;        // The thread local value associated with this
                            // instance of the handle.
} PilBaseThread_t;

extern LIST_NODE          __g_pilThreadList;
extern PilCriticalSection __g_pilThreadCS;

///////////////////////////////////////////////////////////////////////
//
// pilInitBaseThread
//
// Initializes the base thread routines.  This must be called before
// any of the base thread functions are used, preferably from PIL_Init.
//
//

void
pilInitBaseThread(
    );

///////////////////////////////////////////////////////////////////////
//
// pilDeInitBaseThread
//
// Deinitializes the base thread routines.  This needs to be called
// when the base thread functions are no longer needed, preferably
// from PIL_DeInit.
//
//

void
pilDeInitBaseThread(
    );

///////////////////////////////////////////////////////////////////////
//
// pilLockThreadList
//
// Grabs the lock associated with the list of allocated threads.  The
// lock needs to be obtained before using any of the thread list
// functions EXCEPT pilAddToThreadList() and pilRemoveFromThreadList().
//
//

#define pilLockThreadList()   PIL_EnterCriticalSection( &__g_pilThreadCS )

///////////////////////////////////////////////////////////////////////
//
// pilUnLockThreadList
//
// Releases the lock acquired with pilLockThreadList.
//
//

#define pilUnlockThreadList() PIL_LeaveCriticalSection( &__g_pilThreadCS )

///////////////////////////////////////////////////////////////////////
//
// pilAddToThreadList
//
// Adds the given PilBaseThread_t-derived structure to the list of
// allocated threads.  Note the thread lock does NOT need to be acquired
// for this macro.
//
//

#define pilAddToThreadList( pilBaseThread )                             \
    pilLockThreadList();                                                \
    INSERT_TAIL_LIST( &__g_pilThreadList, (LIST_NODE*) pilBaseThread ); \
    pilUnlockThreadList()

///////////////////////////////////////////////////////////////////////
//
// pilRemoveFromThreadList
//
// Removes the given PilBaseThread_t-derived structure from the list of
// allocated threads.  Note the thread lock does NOT need to be acquired
// for this macro.
//
//

#define pilRemoveFromThreadList( pilBaseThread )     \
    pilLockThreadList();                             \
    REMOVE_ENTRY_LIST( (LIST_NODE*) pilBaseThread ); \
    pilUnlockThreadList()

///////////////////////////////////////////////////////////////////////
//
// pilGetFirstInThreadList
//
// Returns the first thread handle from the list of allocated thread
// handles.  Note that the thread lock should have already been acquired
// via pilLockThreadList().
//
//

#define pilGetFirstInThreadList() GET_FIRST_ENTRY_FROM_LIST( &__g_pilThreadList )

///////////////////////////////////////////////////////////////////////
//
// pilGetNextInThreadList
//
// Given a PilBaseThread_t-derived structure, this macro will return the
// next structure in the list of allocated threads.  Note that this macro
// needs to be protected by first calling pilLockThreadList().
//
//

#define pilGetNextInThreadList( pilBaseThread ) \
    GET_NEXT_ENTRY_FROM_LIST( &__g_pilThreadList, (LIST_NODE*) pilBaseThread )

///////////////////////////////////////////////////////////////////////
//
// pilInitBaseThreadHandle
//
// Correctly initializes the members of the base thread structure.
//
//

void
pilInitBaseThreadHandle(
    PilBaseThread_t* pThread
    );

///////////////////////////////////////////////////////////////////////
//
// pilIsValidThreadHandle
//
// Returns 1 if the thread handle is a valid thread handle, otherwise
// it returns 0.
//
//

uint32_t
pilIsValidThreadHandle(
    PilThread Thread
    );

#endif // __BASEOBJ_H_INCLUDED__
