/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: THREAD.C                                                  //
// Module:   Platform Independence Layer Thread Management Routines    //
// Author:   Dan Baumberger                                            //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 2000 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////

#include "pil.h"
#include "basepobj.h"

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
//                                                                   //
//                          G L O B A L S                            //
//                                                                   //
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

// __g_pilThreadList is the head of the list of currently created
// thread structures.
LIST_NODE          __g_pilThreadList;

// __g_pilThreadCS protects the thread list above.  This critical section
// must be entered before accessing the list.
PilCriticalSection __g_pilThreadCS;

///////////////////////////////////////////////////////////////////////
//
// pilInitBaseThread
//
// Initializes the list head and prepares the critical section object
// that protects the list.
//
//

void
pilInitBaseThread(
    )
{
    INITIALIZE_LIST_HEAD(&__g_pilThreadList);
    PIL_InitializeCriticalSection( &__g_pilThreadCS );
}

///////////////////////////////////////////////////////////////////////
//
// pilDeInitBaseThread
//
// Cleans up the critical section protecting the thread list.  Note
// that the list does not need to be cleaned up. Any thread that have
// not be deallocated will be considered leaks.
//
//

void
pilDeInitBaseThread(
    )
{
    PIL_DestroyCriticalSection( &__g_pilThreadCS );
}

///////////////////////////////////////////////////////////////////////
//
// pilInitBaseThreadHandle
//
// Correctly initializes the tag and Value fields of the base thread
// structure.
//
//

void
pilInitBaseThreadHandle(
    PilBaseThread_t* pThread // The base thread to initialize
    )
{
    // Initialize the tag value so that IsValidThreadHandle will
    // view this as a valid handle.
    pThread->tag[ 0 ] = 'T';
    pThread->tag[ 1 ] = 'H';
    pThread->tag[ 2 ] = 'R';
    pThread->tag[ 3 ] = 'D';

    // By default, the Value field needs to be set to 0.
    pThread->Value = 0;
}

///////////////////////////////////////////////////////////////////////
//
// pilIsValidThreadHandle
//
// Checks the tag value to ensure that it is the same as was set during
// initialization of the thread structure.
//
//

uint32_t
pilIsValidThreadHandle(
    PilThread Thread  // The handle of the thread to verify
    )
{
    PilBaseThread_t* pThread = (PilBaseThread_t*) Thread;

    // Obviously, NULL pointers are not valid thread handles.
    if( pThread == NULL ) {
        return( 0 );
    }

    // Check the tag to ensure it is the same as initialization.
    if( ( pThread->tag[ 0 ] = 'T' ) &&
        ( pThread->tag[ 1 ] = 'H' ) &&
        ( pThread->tag[ 2 ] = 'R' ) &&
        ( pThread->tag[ 3 ] = 'D' ) ) {
        return( 1 );
    } else {
        return( 0 );
    }
}

///////////////////////////////////////////////////////////////////////
//
// PIL_GetThreadValue
//
//
// 

PilResult
PIL_GetThreadValue(
    PilThread Thread,	// @parm The handle of the thread to obtain the
                        //   value of.
    uint32_t* pValue    // @parm Pointer to a location to store the 
                        //   the value.
    )
{
    PilBaseThread_t* pThread = (PilBaseThread_t*) Thread;

    if( !pilIsValidThreadHandle( Thread) ) {
        return( PIL_INVALID_HANDLE );
    }

    if( pValue == NULL ) {
        return( PIL_INVALID_PARAM );
    }

    *pValue = pThread->Value;
    return( PIL_OK );
}

///////////////////////////////////////////////////////////////////////
//
// PIL_SetThreadValue
//
//
//

PilResult
PIL_SetThreadValue(
    PilThread Thread, // @parm The handle of the thread to set the value
                      //   of.
    uint32_t  Value   // @parm The 32-bit value to store with this thread.
    )
{
    PilBaseThread_t* pThread = (PilBaseThread_t*) Thread;

    if( !pilIsValidThreadHandle( Thread ) ) {
        return( PIL_INVALID_HANDLE );
    }

    pThread->Value = Value;
    return( PIL_OK );
}

