/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILHEAP.C                                                 //
// Module:   PIL Local Heap Management Routines.                       //
// Author:   JP Sugarbroad                                             //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1996 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////

#include "pil.h"

// size MUST be first
typedef struct block_s {
    long size;            // Block size, ORed with BLOCK_FREE if free
    struct block_s *prev;

    // Doubly-linked free list
    void *__free_list[0];
} block_t;

#define NEXT_FREE(b) ((block_t *)(b)->__free_list[0])
#define NEXT_FREE_PTR(b) ((block_t **)&((b)->__free_list[0]))
#define PREV_FREE(b) ((block_t **)(b)->__free_list[1])

// Free bit
#define BLOCK_FREE (1UL<<31)

// Nice utility macros
#define IS_FREE(b) ((b)->size < 0)
#define IS_USED(b) ((b)->size >= 0)
#define DATA_SIZE(b) ((b)->size & ~BLOCK_FREE)
#define NEXT_BLOCK(b) ((block_t *)((char *)((b)+1)+DATA_SIZE(b)))

// Minimum block size (The size of the two free pointers)
#define MIN_DATA_SIZE 8

// Macro to check maximum size
#define SIZE_TOO_LARGE(s) (s >= BLOCK_FREE)

// uint32_t alignment macro
#define ALIGN4(x) (((x)+3)&~3)

// This is a PilHeap
typedef struct {
#ifdef DEBUG
    uint32_t magic;
    uint32_t size;
#endif
    PilCriticalSection cs;
    block_t *first_free;
} heap_t;

#ifdef DEBUG
#   define PIL_HEAP_MAGIC (0x6EA0109F)

#   define check_heap(h,r) do { \
        if (!(h) || (h)->magic != PIL_HEAP_MAGIC) { \
            ASSERTMSG(FALSE,"Attempt to use bad heap"); \
            return r; \
        } \
        ASSERT((h)->size >= 0 && (h)->size < BLOCK_FREE); \
    } while (0)

#   define check_block(h,b,r) do { \
        if (   (void *)(b) < (void *)(h) \
            || (void *)(b) >= (h)->size+(void *)(h)) \
        { \
            ASSERTMSG(FALSE,"Attempt to pass bad heap block"); \
            return r; \
        } \
        ASSERT(IS_USED(b)); \
    } while (0)
#else
#   define check_heap(h,r)
#   define check_block(h,b,r)
#endif

///////////////////////////////////////////////////////////////////////
//
// lock_heap (internal procedure)
//
// Locks the heap to prevent concurrent access
//

static void lock_heap(heap_t *heap)
{
    PIL_EnterCriticalSection(&heap->cs);
}

///////////////////////////////////////////////////////////////////////
//
// unlock_heap (internal procedure)
//
// Unlocks the heap to allow others to access it
//

static void unlock_heap(heap_t *heap)
{
    PIL_LeaveCriticalSection(&heap->cs);
}

///////////////////////////////////////////////////////////////////////
//
// add_free_list (internal procedure)
//
// Put a block on the free list
//

static void add_free_list(heap_t *heap, block_t *block)
{
    NEXT_FREE(block) = heap->first_free;
    if (NEXT_FREE(block)) {
        PREV_FREE(NEXT_FREE(block)) = NEXT_FREE_PTR(block);
    }
    heap->first_free = block;
    PREV_FREE(heap->first_free) = &heap->first_free;
}

///////////////////////////////////////////////////////////////////////
//
// move_free_list (internal procedure)
//
// Replace a block with another one on the free list
//

static void move_free_list(heap_t *heap, block_t *to, block_t *from)
{
    PREV_FREE(to) = PREV_FREE(from);
    if (PREV_FREE(to)) {
        *PREV_FREE(to) = to;
    }
    NEXT_FREE(to) = NEXT_FREE(from);
    if (NEXT_FREE(to)) {
        PREV_FREE(NEXT_FREE(to)) = NEXT_FREE_PTR(to);
    }
}

///////////////////////////////////////////////////////////////////////
//
// del_free_list (internal procedure)
//
// Remove a block from the free list
//

static void del_free_list(heap_t *heap, block_t *block)
{
    if (PREV_FREE(block)) {
        *PREV_FREE(block) = NEXT_FREE(block);
    }
    if (NEXT_FREE(block)) {
        PREV_FREE(NEXT_FREE(block)) = PREV_FREE(block);
    }
}

///////////////////////////////////////////////////////////////////////
//
// fix_pointers (internal procedure)
//
// Fix the prev pointer for a new block
//

static void fix_pointers(block_t *block)
{
    block_t *next = NEXT_BLOCK(block);

    if (next->size) next->prev = block;
}

///////////////////////////////////////////////////////////////////////
//
// split_block (internal procedure)
//
// Split a block into two blocks. The original block is first, and is
// marked used. The other block is free.
//

static BOOL split_block(heap_t *heap, block_t *block, long size)
{
    block_t *new_block;
    long new_block_size;
    BOOL free;

    new_block_size = DATA_SIZE(block) - size - sizeof(block_t);
    if (new_block_size >= MIN_DATA_SIZE) {
        free = IS_FREE(block);
        block->size = size;

        new_block = NEXT_BLOCK(block);
        new_block->size = new_block_size;
        new_block->prev = block;
        fix_pointers(new_block);
        if (free) {
            new_block->size |= BLOCK_FREE;
            move_free_list(heap,new_block,block);
        } else {
            unlock_heap(heap);
            PIL_HeapFree((PilHeap) heap,new_block);
            lock_heap(heap);
        }
        return FALSE;
    } else {
        return TRUE;
    }
}

///////////////////////////////////////////////////////////////////////
//
// PIL_CreateHeap
//
// Create a new managed heap.
//

PilHeap PIL_CreateHeap(uint32_t dwMaxHeapSize)
{
    heap_t *heap;
    block_t *block;
    uint32_t allocsize;

    ASSERT(   dwMaxHeapSize >= MIN_DATA_SIZE
           && !SIZE_TOO_LARGE(dwMaxHeapSize));

    dwMaxHeapSize = ALIGN4(dwMaxHeapSize);

    // Allocate the heap
    allocsize = sizeof(heap_t) +  // Heap header
                sizeof(block_t) + // Block header
                dwMaxHeapSize +   // Requested size
                4;                // End marker
    heap = (heap_t *) PIL_AllocateMemory(allocsize,0);
    if (heap == NULL) return (PilHeap) NULL;

    // Location of the first block
    block = (block_t *) (heap+1);

#ifdef DEBUG
    heap->magic = PIL_HEAP_MAGIC; // Magic number
    heap->size = allocsize - 4;   // Size of heap
#endif
    PIL_InitializeCriticalSection(&heap->cs);
    heap->first_free = NULL;      // Start of free list

    // Create the initial free block
    block->size = dwMaxHeapSize | BLOCK_FREE;
    block->prev = NULL;
    add_free_list(heap,block);

    // Create the end marker
    block = NEXT_BLOCK(block);
    block->size = 0;

    return (PilHeap) heap;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_DestroyHeap
//
// Delete a managed heap created using CreateHeap.
//

PilResult PIL_DestroyHeap(PilHeap Heap)
{
    heap_t *heap = (heap_t *) Heap;

    check_heap(heap,PIL_INVALID_HANDLE);

    PIL_FreeMemory(heap);
    return PIL_OK;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_HeapAlloc
//
// Allocate a block of the given size of the managed heap.
//

void * PIL_HeapAlloc(PilHeap Heap, uint32_t dwSize)
{
    heap_t *heap = (heap_t *) Heap;
    block_t *block = heap->first_free;

    check_heap(heap,NULL);

    if (dwSize <= 0) return NULL; // Bad block size
    if (SIZE_TOO_LARGE(dwSize)) return NULL; // Block too big
    if (!block) return NULL; // No free blocks

    if (dwSize < MIN_DATA_SIZE) dwSize = MIN_DATA_SIZE;

    dwSize = ALIGN4(dwSize);

    lock_heap(heap);

    // Search for a large enough free block
    while (block && DATA_SIZE(block) < dwSize) {
        block = NEXT_FREE(block);
    }
    if (!block) {
        unlock_heap(heap);
        return NULL; // Didn't find one
    }

    if (split_block(heap,block,dwSize)) {
        block->size &= ~BLOCK_FREE;
        del_free_list(heap,block);
    }

    unlock_heap(heap);

    return block+1;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_HeapFree
//
// Deallocate a memory block from a managed heap.
//

PilResult PIL_HeapFree(PilHeap Heap, void * pBuf)
{
    heap_t *heap = (heap_t *) Heap;
    block_t *block = ((block_t *) pBuf)-1;
    block_t *other;
    BOOL in_free_list = FALSE;

    check_heap(heap,PIL_INVALID_HANDLE);
    check_block(heap,block,PIL_INVALID_PARAM);

    lock_heap(heap);

    // Coalesce right
    other = NEXT_BLOCK(block);
    if (IS_FREE(other)) {
        block->size = DATA_SIZE(block) + sizeof(block_t) + DATA_SIZE(other);
        move_free_list(heap,block,other);
        in_free_list = TRUE;
    }

    // Coalesce left
    other = block->prev;
    if (other && IS_FREE(other)) {
        other->size = DATA_SIZE(other) + sizeof(block_t) + DATA_SIZE(block);
        if (in_free_list) {
            del_free_list(heap,block);
        } else {
            in_free_list = TRUE;
        }
        block = other;
    }

    if (in_free_list) {
        // Fix the prev pointer
        fix_pointers(block);
    } else {
        // We didn't have any blocks to coalesce with.
        add_free_list(heap,block);
    }

    // Mark the block free
    block->size |= BLOCK_FREE;

    unlock_heap(heap);
    return PIL_OK;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_HeapShrink
//
// Reduce the size of an allocated block.
//

PilResult PIL_HeapShrink(PilHeap Heap, void * pBuf, uint32_t dwNewSize)
{
    heap_t *heap = (heap_t *) Heap;
    block_t *block = ((block_t *) pBuf)-1;
    BOOL r;

    check_heap(heap,PIL_INVALID_HANDLE);
    check_block(heap,block,PIL_INVALID_PARAM);

    if (DATA_SIZE(block) < dwNewSize) return PIL_INVALID_PARAM;
    if (dwNewSize <= 0) return PIL_INVALID_PARAM;

    if (dwNewSize < MIN_DATA_SIZE) dwNewSize = MIN_DATA_SIZE;
    dwNewSize = ALIGN4(dwNewSize);

    lock_heap(heap);

    r = split_block(heap,block,dwNewSize);

    unlock_heap(heap);
    return r ? PIL_ERROR : PIL_OK;
}
