/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILMBOX.C                                                 //
// Module:   Platform Isolation Layer Mailbox Interface                //
// Author:   JP Sugarbroad                                             //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////

#include "pil.h"

extern void* memcpy( void*, const void*, int );

#ifdef DEBUG
    // Magic number for debug checking
#   define PIL_MBOX_MAGIC (0xA2CB0098)

    // Macro to check for invalid mailboxes
#   define pil_checkmbox(mbox,r) do { \
        if (!(mbox) || (mbox)->magic != PIL_MBOX_MAGIC) { \
            ASSERTMSG(FALSE,"Attempt to use bad mailbox"); \
            return r; \
        } \
    } while (0);
#else // !DEBUG
#   define pil_checkmbox(mbox,r)
#endif

// The structure behind PilMailbox
typedef struct mbox_s {
#ifdef DEBUG
    uint32_t magic;
#endif
    PilCriticalSection lock; // Mutex lock for access to other entries
    PilSemaphore msgs;       // Semaphore of messages available
    PilSemaphore free;       // Semaphore of free slots
    uint32_t mlen;              // Length of a message
    uint32_t qmax, qlen, qoff;  // Maximum length, current length, offset of queue
    void * data;              // Message data
} mbox_t;

// Implementation note:
//
// mbox_t.data is used as a circular queue, with qoff pointing to the current
// head of the queue. All offsets/lengths are multiplied by mlen before being
// used to offset data.

///////////////////////////////////////////////////////////////////////
//
// PIL_CreateMailbox
//
// Creates a new Mailbox
//

PilMailbox PIL_CreateMailbox(uint32_t dwMessageLength, uint32_t dwQueueLength)
{
    mbox_t *mbox;

    ASSERT(dwMessageLength > 0);
    ASSERT(dwQueueLength > 0);

    // Allocate and initialize the mailbox

    mbox = (mbox_t *) PIL_AllocateMemory(sizeof(mbox_t),0);
    if (mbox == NULL) return (PilMailbox) NULL;

#ifdef DEBUG
    mbox->magic = PIL_MBOX_MAGIC;
#endif

    PIL_InitializeCriticalSection(&mbox->lock);

    mbox->mlen = dwMessageLength;
    mbox->qmax = dwQueueLength;
    mbox->qlen = 0;
    mbox->qoff = 0;

    mbox->data = PIL_AllocateMemory(dwMessageLength * dwQueueLength, 0);
    if (mbox->data == NULL) goto out_nodata;

    mbox->msgs = PIL_CreateSemaphore(0,dwQueueLength);
    if (mbox->msgs == NULL) goto out_nomsgs;

    mbox->free = PIL_CreateSemaphore(dwQueueLength,dwQueueLength);
    if (mbox->free == NULL) goto out_nofree;

    return( ( PilMailbox ) mbox );

  out_nofree:
    PIL_DestroySemaphore(mbox->msgs);
  out_nomsgs:
    PIL_FreeMemory(mbox->data);
  out_nodata:
    PIL_DestroyCriticalSection(&mbox->lock);
    PIL_FreeMemory(mbox);
    return (PilMailbox) NULL;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_DestroyMailbox
//
// Destroys a mailbox object
//

PilResult PIL_DestroyMailbox(PilMailbox Mailbox)
{
    mbox_t *mbox = (mbox_t *) Mailbox;

    pil_checkmbox(mbox,PIL_INVALID_HANDLE); // Check for invalid mailbox
#ifdef DEBUG
    mbox->magic = 0; // Prevent use of destroyed mailboxes
#endif
    PIL_EnterCriticalSection(&mbox->lock);
    PIL_DestroyCriticalSection(&mbox->lock);
    PIL_DestroySemaphore(mbox->free);
    PIL_DestroySemaphore(mbox->msgs);
    PIL_FreeMemory(mbox->data);
    PIL_FreeMemory(mbox);
    return PIL_OK;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_ReadMailbox
//
// Get a message from the mailbox, waiting up to dwTimeout for one to
// arrive
//

PilResult PIL_ReadMailbox(PilMailbox Mailbox, void * pMessage, uint32_t dwTimeout)
{
    mbox_t *mbox = (mbox_t *) Mailbox;

    pil_checkmbox(mbox,PIL_INVALID_HANDLE); // Check for invalid mailbox
#ifdef DEBUG
    if (!pMessage) {
        ASSERTMSG(FALSE,"Invalid pMessage in PIL_ReadMailbox");
        return PIL_INVALID_PARAM;
    }
#endif

    // Try to grab a message
    if (PIL_AcquireSemaphore(mbox->msgs, dwTimeout) == PIL_TIMEOUT) {
        return PIL_TIMEOUT;
    }
    ASSERT(mbox->qlen); // Make sure there is space

    // Acquire exclusive access
    PIL_EnterCriticalSection(&mbox->lock);

    // Copy out a message
    memcpy(pMessage, ((char *)mbox->data) + mbox->qoff*mbox->mlen, mbox->mlen);

    // Decrement number of messages
    if (--mbox->qlen) {
        // More messages remaining. Point qoff to next message.
        if (mbox->qmax == ++mbox->qoff) mbox->qoff = 0;
    } else {
        // No more messages. Reset qoff to 0.
        mbox->qoff = 0;
    }
    
    // Release exclusive access
    PIL_LeaveCriticalSection(&mbox->lock);

    // Notify writers of a new free slot
    PIL_ReleaseSemaphore(mbox->free);
    return PIL_OK;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_WriteMailbox
//
// Add a message to the mailbox, waiting up to dwTimeout for a space
// to become available.
//

PilResult PIL_WriteMailbox(PilMailbox Mailbox, void * pMessage, uint32_t dwTimeout)
{
    mbox_t *mbox = (mbox_t *) Mailbox;
    uint32_t pos;

    pil_checkmbox(mbox,PIL_INVALID_HANDLE); // Check for invalid mailbox
#ifdef DEBUG
    if (!pMessage) {
        ASSERTMSG(FALSE,"Invalid pMessage in PIL_WriteMailbox");
        return PIL_INVALID_PARAM;
    }
#endif

    // Acquire a free slot
    if (PIL_AcquireSemaphore(mbox->free, dwTimeout) == PIL_TIMEOUT) {
        return PIL_TIMEOUT;
    }
    ASSERT(mbox->qlen < mbox->qmax); // Make sure there is space

    // Get exclusive access
    PIL_EnterCriticalSection(&mbox->lock);

    // Find where to put the message
    pos = mbox->qoff + mbox->qlen;
    if (pos > mbox->qmax) pos -= mbox->qmax;

    // Copy in the message
    memcpy(((char *)mbox->data) + pos*mbox->mlen, pMessage, mbox->mlen);

    // Increment number of messages
    mbox->qlen++;

    // Release exclusive access
    PIL_LeaveCriticalSection(&mbox->lock);

    // Notify readers of a new message
    PIL_ReleaseSemaphore(mbox->msgs);
    return PIL_OK;
}
