/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILTMR.C                                                  //
// Module:   Platform Isolation Layer Timer Routines                   //
// Author:   JP Sugarbroad                                             //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////

#include "pil.h"

#ifdef DEBUG
    // Magic number for debug checking
#   define PIL_TIMER_MAGIC (0x31415926)

    // Macro to check for invalid timers
#   define pil_checktmr(t,r) do { \
        if (!(t) || (t)->magic != PIL_TIMER_MAGIC) { \
            ASSERTMSG(FALSE,"Attempt to use bad timer"); \
            return r; \
        } \
        ASSERT((t)->function != NULL); \
        ASSERT((t)->interval >= 0); \
    } while (0);
#else // !DEBUG
#   define pil_checktmr(t,r)
#endif

// The structure behind PilTimer
typedef struct timer_s {
#ifdef DEBUG
    uint32_t magic;
#endif
    // Linked list and expiry time
    struct timer_s **prev, *next;
    uint64_t expire;
    
    // Passed from user
    PilTimerRoutine function;
    uint32_t interval;
    void *data; // a.k.a. Context
    PilTimerType type;
} timer_t;

static timer_t *timer_list;

static PilCriticalSection timer_lock;
static PilSemaphore timer_change;
static PilEvent timer_enable;
static int timer_disabled, timer_dead;
static PilEvent timer_thread_done;
static PilThread timer_thread;

///////////////////////////////////////////////////////////////////////
//
// timer_proc (internal procedure)
//
// Thread procedure which runs timers
//

static uint32_t timer_proc(void *arg)
{
    while (!timer_dead) {
        int64_t time_to_next;

        if (timer_disabled) PIL_WaitForEvent(timer_enable,PIL_INFINITE_TIMEOUT);

        PIL_EnterCriticalSection(&timer_lock);
        // Yes, it could be a do while loop
      next_timer:
        if (!timer_list) {
            time_to_next = (uint64_t) PIL_INFINITE_TIMEOUT;
            goto no_timer;
        }
        time_to_next = timer_list->expire - PIL_GetCurrentTime();
        if (time_to_next <= 0) {
            timer_t *t = timer_list;
            if ((timer_list = t->next)) t->next->prev = &timer_list;
            t->prev = NULL;
            t->next = NULL;

            PIL_LeaveCriticalSection(&timer_lock);
            if (t->type == PIL_TIMER_REPEATING) PIL_StartTimer((PilTimer) t);
            (*t->function)((PilTimer) t, t->data);
            PIL_EnterCriticalSection(&timer_lock);

            goto next_timer;
        }
      no_timer:
        PIL_LeaveCriticalSection(&timer_lock);

        // Wait for timer list change or next timer firing time
        ASSERT(time_to_next < 0x100000000LL);
        PIL_AcquireSemaphore(timer_change,(uint32_t) time_to_next);
    }
    PIL_SetEvent(timer_thread_done);
    return 0;
}

///////////////////////////////////////////////////////////////////////
//
// timer_init
//
// Called by PIL_Init to initialize timers.
//
// RETURNS
//
// PIL_NOMEM            Insufficient memory to initialize
// PIL_OK               Initialization complete
//

PilResult timer_init(void)
{
    timer_list = NULL;
    timer_dead = 0;

    timer_change = PIL_CreateBinarySemaphore(FALSE);
    if (timer_change == NULL) return PIL_NOMEM;

    timer_enable = PIL_CreateEvent(FALSE,FALSE);
    if (timer_enable == NULL) {
        PIL_DestroySemaphore(timer_change);
        return PIL_NOMEM;
    }

    timer_thread_done = PIL_CreateEvent(FALSE,FALSE);
    if (timer_thread_done == NULL) {
        PIL_DestroyEvent(timer_enable);
        PIL_DestroySemaphore(timer_change);
        return PIL_NOMEM;
    }

    PIL_InitializeCriticalSection(&timer_lock);

    timer_thread = PIL_CreateThread(timer_proc,0,"Timer thread",8,NULL,0);
    if (timer_thread == NULL) {
        PIL_DestroyCriticalSection(&timer_lock);
        PIL_DestroyEvent(timer_thread_done);
        PIL_DestroyEvent(timer_enable);
        PIL_DestroySemaphore(timer_change);
        return PIL_NOMEM;
    }

    return PIL_OK;
}

///////////////////////////////////////////////////////////////////////
//
// timer_deinit
//
// Called by PIL_DeInit to deinitialize timers.
//

void timer_deinit(void)
{
    timer_dead = 1;
    PIL_ReleaseSemaphore(timer_change);
    if (PIL_WaitForEvent(timer_thread_done,1000) != PIL_OK) {
        PIL_TerminateThread(timer_thread);
    }
    PIL_DestroyThread(timer_thread,NULL);
    PIL_DestroyCriticalSection(&timer_lock);
    PIL_DestroyEvent(timer_thread_done);
    PIL_DestroyEvent(timer_enable);
    PIL_DestroySemaphore(timer_change);
}

///////////////////////////////////////////////////////////////////////
//
// PIL_CreateTimer
//
// Creates a new timer that fires once or repeats dwInterval milliseconds.
// The timer is not scheduled to fire until PIL_StartTimer is called.
//

PilTimer PIL_CreateTimer(
    PilTimerRoutine TimerRoutine,
    uint32_t        dwInterval,
    void *          Context,
    PilTimerType    Type
) {
    timer_t *t;

#ifdef DEBUG
    if (dwInterval < 0 && dwInterval != PIL_INFINITE_TIMEOUT) {
        ASSERTMSG(FALSE,"Bad dwInterval passed to PIL_CreateTimer");
        return (PilTimer) NULL;
    }
#endif

    // Allocate and initialize the timer

    t = (timer_t *) PIL_AllocateMemory(sizeof(timer_t),PIL_MEM_ZERO_INIT);
    if (t == NULL) return (PilTimer) NULL;

#ifdef DEBUG
    t->magic = PIL_TIMER_MAGIC;
#endif
    t->function = TimerRoutine;
    t->interval = dwInterval;
    t->data = Context;
    t->type = Type;

    return (PilTimer) t;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_DestroyTimer
//
// Destroys a timer object created using PIL_CreateTimer.
//

PilResult PIL_DestroyTimer(PilTimer Timer)
{
    timer_t *t = (timer_t *) Timer;

    pil_checktmr(t,PIL_INVALID_HANDLE); // Check for invalid timer
#ifdef DEBUG
    t->magic = 0; // Prevent use of a destroyed timer
#endif

    PIL_FreeMemory(t);
    return PIL_OK;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_ResetTimer
//
// Resets the timer parameters allowing specification of a new interval,
// timer routine, or timer type.
//

PilResult PIL_ResetTimer(
    PilTimer        Timer,
    PilTimerRoutine TimerRoutine,
    uint32_t        dwInterval,
    void *          Context,
    PilTimerType    Type
) {
    timer_t *t = (timer_t *) Timer;

    pil_checktmr(t,PIL_INVALID_HANDLE); // Check for invalid timer

#ifdef DEBUG
    if (!TimerRoutine) {
        ASSERTMSG(FALSE,"Bad TimerRoutine passed to PIL_ResetTimer");
        return PIL_INVALID_PARAM;
    }
    if (dwInterval < 0 && dwInterval != PIL_INFINITE_TIMEOUT) {
        ASSERTMSG(FALSE,"Bad dwInterval passed to PIL_ResetTimer");
        return PIL_INVALID_PARAM;
    }
#endif

    // Illegal to reset a running timer
    if (t->prev) return PIL_ERROR;

    // Just change the data structure
    t->function = TimerRoutine;
    t->interval = dwInterval;
    t->data = Context;
    t->type = Type;
    return PIL_OK;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_StartTimer
//
// Schedules a timer to fire at the interval it was created with.
//

PilResult PIL_StartTimer(PilTimer Timer)
{
    timer_t *t = (timer_t *) Timer, **cur;

    pil_checktmr(t,PIL_INVALID_HANDLE); // Check for invalid timer

    if (t->prev) return PIL_ERROR;

    if (t->interval == PIL_INFINITE_TIMEOUT) {
        t->prev = (void *) 1L;
        return PIL_OK;
    }

    t->expire = PIL_GetCurrentTime() + t->interval;

    // Lock the timer list
    PIL_EnterCriticalSection(&timer_lock);

    // Add it to the (sorted) list
    cur = &timer_list;
    while (*cur && (*cur)->expire < t->expire) {
        cur = &(*cur)->next;
    }
    t->prev = cur;
    if ((t->next = *cur)) {
        t->next->prev = &t->next;
    }
    *cur = t;

    // If we modified the timer list head, notify thread.
    if (timer_list == t) PIL_ReleaseSemaphore(timer_change);

    // Unlock the timer list
    PIL_LeaveCriticalSection(&timer_lock);

    return PIL_OK;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_StopTimer
//
// Removes the timer from the queue.
//

PilResult PIL_StopTimer(PilTimer Timer)
{
    timer_t *t = (timer_t *) Timer;

    pil_checktmr(t,PIL_INVALID_HANDLE); // Check for invalid timer

    if (!t->prev) return PIL_ERROR;

    if (t->interval == PIL_INFINITE_TIMEOUT) {
        t->prev = NULL;
        return PIL_OK;
    }

    // Lock the timer list
    PIL_EnterCriticalSection(&timer_lock);

    // Take it out of the list
    if (((*t->prev) = t->next)) t->next->prev = t->prev;

    // If we are stopping the first timer, notify thread.
    if (t->prev == &timer_list) PIL_ReleaseSemaphore(timer_change);

    // Unlock the timer list
    PIL_LeaveCriticalSection(&timer_lock);

    t->prev = NULL;
    t->next = NULL;

    return PIL_OK;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_DisableTimers
//
// Disables all timer callbacks until PIL_EnableTimers() is called.
//

PilResult PIL_DisableTimers()
{
    if (timer_disabled) return PIL_ERROR;

    timer_disabled = 1;
    PIL_ReleaseSemaphore(timer_change);

    return PIL_OK;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_EnableTimers
//
// Re-enables timers disabled by PIL_DisableTimers().
//

PilResult PIL_EnableTimers()
{
    if (!timer_disabled) return PIL_ERROR;

    timer_disabled = 0;
    PIL_SetEvent(timer_enable);

    return PIL_OK;
}
