/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: TIME.C                                                    //
// Module:   Platform Isolation Layer Time Helper Routines             //
// Author:   JP Sugarbroad                                             //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////

//
// Usage notes:
//
// Use this file if your system does not provide 64-bit time. To use it,
// implement a routine time32() which returns the uint32_t time value (in
// milliseconds). This file provides PIL_GetCurrentTime, avoiding the wrap
// problem resulting from a simple cast from uint32_t to uint64_t.
//
// Don't forget to put time_init and time_deinit into PIL_Init and PIL_DeInit
// respectively.
//

#include "pil.h"

// External routine to provide 32-bit time

extern uint32_t time32(void);

static uint32_t last_time,time_high;
static PilCriticalSection time_lock;
static PilTimer time_timer;

/////////////////////////////////////////////////////////////////////////
//
// time_check (internal procedure)
//
// Checks for 32-bit time wrap
//

static void time_check(PilTimer t, void *arg)
{
    uint32_t new_time;

    PIL_EnterCriticalSection(&time_lock);
    new_time = time32();
    if (new_time < last_time) time_high++;
    last_time = new_time;
    PIL_LeaveCriticalSection(&time_lock);
}

/////////////////////////////////////////////////////////////////////////
//
// time_init
//
// Initializes the time helper routines
//

PilResult time_init(void)
{
    time_timer = PIL_CreateTimer(time_check,0x7FFFFFFF,NULL,PIL_TIMER_REPEATING);
    if (time_timer == NULL) return PIL_NOMEM;
    PIL_StartTimer(time_timer);
    PIL_InitializeCriticalSection(&time_lock);
    last_time = time32();
    time_high = 0;
    return PIL_OK;
}

/////////////////////////////////////////////////////////////////////////
//
// time_deinit
//
// De-initializes the time helper routines
//

void time_deinit(void)
{
    PIL_StopTimer(time_timer);
    PIL_EnterCriticalSection(&time_lock);
    PIL_DestroyCriticalSection(&time_lock);
    PIL_DestroyTimer(time_timer);
}

///////////////////////////////////////////////////////////////////////
//
// PIL_GetCurrentTime
//
// Returns a 64-bit time value in milliseconds.  Note that this may not
// be the actual clock time in milliseconds but instead may be the
// number of milliseconds since system start.  It is guaranteed that
// the difference between any two calls of this function is the elapsed
// real time in milliseconds.
//

uint64_t PIL_GetCurrentTime()
{
    PIL_StopTimer(time_timer);
    time_check(NULL,NULL);
    PIL_StartTimer(time_timer);
    return (((uint64_t) time_high) << 32) | last_time;
}
