/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PIL.H                                                     //
// Module:   Platform Isolation Layer                                  //
// Author:   Daniel Baumberger                                         //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
// PURPOSE:                                                            //
//                                                                     //
//   Top level declarations of the Platform Independence Layer.        //
//                                                                     //
/////////////////////////////////////////////////////////////////////////

#ifndef __PIL_H_INCLUDED__
#define __PIL_H_INCLUDED__

//
// Enable AUTODUCK processing
// @doc
//
// @topic PIL API |
// PIL stands for the Platform Independence Layer (or Library).  It
// provides an operating system independent API for many common
// operations.  The API is documented in the follow sections.
//
// PIL also provides some common debugging routines.  It provides
// appropriately defined <f TRACE> and <f ASSERT> macros (i.e. they are
// automatically disable when compiled in release mode).  With a 
// specially build version of PIL, it keeps track of all the resources
// allocated through it and can report on which ones were not freed.
//
// Finally, PIL can provide other runtime support needs that an application
// may have.  This may include implementing parts of the C or C++ run-time
// for environments that do not have native support.  Other common utility
// code, such as Heap management, can be including in PIL allowing all
// clients to benefit from the shared code.
//
// The PIL API abstracts the common operating system services that 
// applications and drivers use.  By porting the PIL functionality to the
// new platform and rebuilding depending applications or drivers, much,
// if not all, of the code does not need to be ported since it is not
// dependent on any single operating system but only on PIL.
//
// The PIL services are broken into two categories: required and optional
// services.  Required services are available in all implementations.
// Optional services may not be available on every implementation due to
// difficulty in emulation or the services just do not make sense for that
// implementation.  The required services fall into these categories:
//
// <tab>- Asychronous Procedure Calls (Apc)<nl>
// <tab>- Critical sections<nl>
// <tab>- Events<nl>
// <tab>- Heap management<nl>
// <tab>- Initialization and de-initialization<nl>
// <tab>- Mailbox (messaging) management<nl>
// <tab>- Memory management<nl>
// <tab>- Semaphores<nl>
// <tab>- Threading<nl>
// <tab>- Timers<nl>
//
// Optional services include:
//
// <tab>- Dynamic Libraries<nl>
// <tab>- Input/Output port access functions<nl>
// <tab>- Interlocked services<nl>
// <tab>- Registry Access<nl>
//
// PIL also provides some utility services:
//
// <tab>- Miscellaneous utility services<nl>
// <tab>- <ms ASSERT> and <ms DEBUG> macros<nl>
// <tab>- Allocated resource tracking<nl>
// 
// @topic PIL Naming Conventions |
//
// All required PIL functions start with the prefix <bold PIL_>.  This clearly
// distinguishes PIL calls from any platform or application specific
// functions.  Any function that starts with <bold PIL_> is guaranteed to be
// available on any platform that PIL has an implementation for.  Optional
// PIL services have the prefix <bold PILOS_> to distinguish that they may not
// be available in all implementations.  When porting code with calls to
// <bold PILOS_> functions, the PIL implementation notes for the target
// implementation need to be consulted to see if those optional services
// are available.
//
// Platform specific functions may be added to specific implementations of
// PIL if they follow a certain naming scheme.  They cannot start with 
// <bold PIL_> or <bold PILOS_> but rather they must start with the PIL
// implementation specific name such as <bold PIL32_> for Win32 or 
// <bold PILVX_> for VxWorks.  If the service is needed in more than one
// implementation of PIL, it should be made a <bold PILOS_> optional service.
//
// All PIL data structures start with the <bold Pil> prefix and follow the
// Pascal naming convention such as <ms PilSemaphore> or 
// <ms PilCriticalSection>.  PIL constants start <bold PIL_> such <ms PIL_OK>
// or <ms PIL_ERROR>.
//
// @topic Standard C Run-time |
//
// In any environment that provides a C compiler, it is a good assumption
// that at least part of the standard C run-time library is available.  PIL
// does not explicitly export those run-time functions, rather is can be
// assumed that they are available elsewhere.  However, under certain
// environments, not all C run-time functions will be available.  In those
// cases, PIL may implement parts, although certainly not all, of the C
// run-time functions with their standard prototypes.
//
// It is not specified that any implementation of PIL implement the C run-time
// functions.  However, as noted above, any specific implementation may be
// supporting them.  Consult the implementation specific notes for the
// implementation for details on which run-time library support, if any,
// is supported.
//
// @topic Berkeley Sockets |
//
// For platforms where networking support is available, the prefferred
// API for network programming is the standard Berkeley Sockets API as
// defined in the POSIX.1g specification and discussed in "Unix Network
// Programming" written by Richard Stevens.  PIL does not attempt to
// place an abstraction to this API or provide any extensions to it.
// Aynchronous extentions, such as WINSOCK supports, are not very portable.
// It is best to stick with standard Berkeley synchronous sockets API.
//
// @topic Return Codes |
// The API documentation in the following sections documents many of the
// return codes that are returned as result codes from PIL functions.
// However, implementations may return codes that are not explicitly
// documented.  This arises from the way that certain implementations need
// to implement those services.
//
// Certain return codes, if not otherwise, documented, may be returned
// from almost any function.  These codes include:
//
// @flag PIL_ERROR | Unless otherwise documented for a particular API call,
//   this denotes an implementation specific general error.  The debug
//   implementation of PIL should provide more information on why this
//   error occurred.
// @flag PIL_INVALID_PARAM | Although documented in most cases, any function
//   can return this error code if it determines that any of the given 
//   parameters is not valid.
// @flag PIL_NOMEM | Many implementations of functions may require memory
//   to be allocated that other implementations may not need.  In this case,
//   PIL_NOMEM is the proper return code.
//
//

//#define __int8_t_defined
#ifdef _VXWORKS_
#include <unistd.h>
#define inet_aton(cp, inp) (inet_aton(cp, inp)==0)?1:0
#else
#include <stdint.h>
#endif

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#ifndef BOOL
#define BOOL unsigned int
#endif

#ifndef NULL
#define NULL 0
#endif

typedef uint32_t PilResult;

//
// PIL Return Codes
//

#define PIL_OK              0x00
#define PIL_ERROR           0x01
#define PIL_NOMEM           0x02
#define PIL_INVALID_HANDLE  0x03
#define PIL_INVALID_PARAM   0x04
#define PIL_NOT_FOUND       0x05
#define PIL_INITIALIZED     0x06
#define PIL_NOT_INITIALIZED 0x07
#define PIL_TIMEOUT         0x10

#define PIL_NO_MORE_ITEMS       0x20
#define PIL_INSUFFICIENT_BUFFER 0x21

#define PIL_NOT_IMPLEMENTED 0xFF 

#define PIL_INFINITE_TIMEOUT ( ( (uint32_t) -1 ) / 2)

#ifdef __cplusplus
    extern "C" {
#endif // __cplusplus

#include "pilinit.h"
#include "pilapc.h"
#include "pilcs.h"
#include "pilevnt.h"
#include "pilintrl.h"
#include "pilio.h"
#include "pilmem.h"
#include "pilsem.h"
#include "pilthd.h"
#include "piltmr.h"
#include "pilheap.h"
#include "pilmbox.h"
#include "pildebug.h"
#include "pillib.h"
#include "pilmisc.h"
        
#ifdef __cplusplus
}
#endif // __cplusplus

#endif // __PIL_H_INCLUDED__
