/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PIL_LNX.H                                                 //
// Module:   Platform Isolation Layer Internal Routines                //
// Author:   Venky Gokulrangan                                         //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
// PURPOSE: Data Structure Declaration and function declaration for    //
//          internal implementation of PIL in Linux User Mode          //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//
// HISTORY:
//
/////////////////////////////////////////////////////////////////////////

#ifndef __PIL_LNX_H_INCLUDED__
#define __PIL_LNX_H_INCLUDED__

#ifdef __cplusplus
    extern "C" {
#endif

#include "pil.h"
#include "llist.h"
#include "cqueue.h"
#include <pthread.h>

extern int errno ;


//////////////////////////////////////////////////////////////////////////////
//
//  Timer  Structures
//
// The following structures are used to implement the timer routines
// Each of the thread has a Timer Header structure and pointers to the
// timer objects created in the thread. A thread can have upto MAX_TIMERS
// threads
//
//////////////////////////////////////////////////////////////////////////////

#define MAX_TIMERS        0x10
#define PIL_TIMER_RUNNING 0x01


//////////////////////////////////////////////////////////////////////////////
//
// Information maintained per timer created
//
// Notes:
//
// 1. If a semaphore with specified timeout is used by a thread,
// the Handle var is set to 1 by the semaphore. Upon expiration of the 
// timer, Expiration Handler resets this Handle to indicate that the timer
// expired. And the callback is called.. Usually, the callback checks the
// flag and understands the timeout and notifies its clients accordingly
//
// 2. Timer count is decremented using a heart-beat style timer. If no
// more timer is present, then the heart beat is stopped
//
//////////////////////////////////////////////////////////////////////////////

typedef struct _PilTimer_t
{
#ifdef RESTRACK
   LIST_NODE Link ;
#endif
   PilTimer          Handle ;  // Flag to be used by semaphore implementaion
   DWORD             dwTime ;  // in millisec -- firing interval
   DWORD             dwCurr ;  // curr expiration in count
   PilTimerRoutine   ProcedureAddress ; // Function to trigger
   PVOID             Data;     // Data to be passed to the routine
   int               Recur ;   // recurring timer
   int               Flags ;   // Flags for indicating RUNNING or STOPPED

} PilTimer_t ;
  

//
// Timer management information maintained per thread
//

typedef struct _PilTimerHeader_t
{
#ifdef RESTACK
   LIST_NODE Link ;
#endif
   WORD           NumTimers ;             // Num of Timers Active
   WORD           Mask;                   // Indicates Availability
   PilTimer_t*    TObject [MAX_TIMERS] ;  // 16 timers/thread
   PilSemaphore   AccessSem ;             // Access to the objects above

} PilTimerHeader_t ;


//////////////////////////////////////////////////////////////////////////////
//
//  Thread Management information
//  StartAddress, ThreadParam, dwOptions are client specified information
//  Tid is maintained to perform operations using POSIX's thread library
//
//////////////////////////////////////////////////////////////////////////////

typedef struct _PilThread_t
{
#ifdef RESTRACK
   LIST_NODE Link ;
#endif

  PilThreadRoutine StartAddress ; // Address of the 
  PVOID            ThreadParam ;  // Passed by the thread creator
  DWORD            dwOptions ;    // Thread Options
  pthread_t        Tid;           // Id of the thread

} PilThread_t ;


//////////////////////////////////////////////////////////////////////////////
//
//  Semaphore Management Structures
//
//  Semid is maintained to operate with the native LINUX calls.
//  dwMax is used to maintain the maximum count, so that in case a release
//  semaphore indulges in exceeding the limit, we can flag it.
//
//////////////////////////////////////////////////////////////////////////////

typedef struct _PilSemaphore_t
{
#ifdef RESTRACK
   LIST_NODE Link ;
#endif

   int   Semid ;
   DWORD dwMax ;

} PilSemaphore_t ;


// Event Structures
//
// Events are simulated on top of semaphores
// We use a flag and a count of waiting threads to trigger the event
// for all waiters.
//////////////////////////////////////////////////////////////////////////////


typedef struct _PilEvent_t
{

#ifdef RESTRACK
   LIST_NODE Link ;
#endif

   BOOL      bManualReset ;  // Manual or Auto Reset
   BOOL      bSignalled ;    // Initial State

   int       NumWaiters ;    // Number of anxious waiters
   BOOL      Flag ;          // Status of the event

   // events are built on top of semaphores

   PilSemaphore AccessSem ;  // Sync access to the above vars
   PilSemaphore EventSem ;   // Event for the waiters to wait on

} PilEvent_t ;



//////////////////////////////////////////////////////////////////////////////
//
//  Mailbox Structures
//
//////////////////////////////////////////////////////////////////////////////

//
// This contains the functions for mailbox..
// The functions are intended to be used on a per process basis
// using the static library mechanism
//

typedef struct _MboxDesc_t
{
   CQHandle       Handle ;     // Handle to the Queue
   DWORD          dwQLen;      // Maximum number of messages in the queue
   DWORD          dwMailSize ; // Size of individual message

   PilSemaphore   hReadSem ;   // Reader Semaphore
   PilSemaphore   hWriteSem ;  // Writer Semaphore

   PilSemaphore   pLock;       // Semaphore protects the circ queue

} MboxDesc;



#ifdef __cplusplus
   } ;
#endif

#endif
