/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILCS.H                                                   //
// Module:   Platform Isolation Layer Critical Sections                //
// Author:   Daniel Baumberger                                         //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Declares the Critical Section API of PIL.                           //
//                                                                     //
/////////////////////////////////////////////////////////////////////////

#ifndef __PILCS_H_INCLUDED__
#define __PILCS_H_INCLUDED__

typedef char PilCriticalSection[ 32 ];

//
// @doc
// @topic Critical Sections (Required) |
//
// Critical sections are a method of synchronizing access to shared resources.
// They differ from semaphores in that they are essentially binary: either a
// thread has it or it does not and will block trying to get it.  There is
// no timeout value associated with waiting for the critical section.
//
// Another major difference is the user of the critical section owns the
// memory for the critical section.  The owning process must allocate the
// memory by declaring a variable of type <f PilCriticalSection> for each
// critical section and initialize them by using 
// <f PIL_InitializeCriticalSection>.  Once the critical section is no longer
// needed, it must be discarded using <f PIL_DestroyCriticalSection>.
//
// A thread attempts to obtain a critical section using 
// <f PIL_EnterCriticalSection>.  When a thread is done with the critical
// section, it releases the critical section by calling 
// <f PIL_LeaveCriticalSection>.  <f PIL_TryEnterCriticalSection> can be
// used to not block for a critical section but test if it can be acquired.
//
// Once a thread owns a critical section, subsequent calls to 
// <f PIL_EnterCriticalSection> will not block.  This prevents deadlocks
// with a thread waiting for itself.  A thread must call 
// <f PIL_LeaveCriticalSection> an equal number of times as
// <f PIL_EnterCriticalSection> in order to release the critical section.
//

///////////////////////////////////////////////////////////////////////
//
// PIL_DestroyCriticalSection
//
// @func <f PIL_DestroyCriticalSection> destroys any memory assoicated
//   with a critical section that has been initialized via
//   <f PIL_InitializeCriticalSection>.  The behavior is undefined to
//   destroy a critical section that has not been initialized via
//   <f PIL_InitializeCriticalSection> or has threads waiting on it.
//
// @rdesc This function does not return a value.
//
//

void
PIL_DestroyCriticalSection(
    PilCriticalSection* pCriticalSection // @parm The critical section object
                                         //   to destroy.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_EnterCriticalSection
//
// @func <f PIL_EnterCriticalSection> enters a critical section gaining
//   access to the shared resource(s).  If another thread has the critical
//   section, this call will block until the critical section may be acquired.
//   There is no timeout.  The thread will wait forever until it can gain
//   access.
//
//   Once a critical section has successfully been obtained, the owning
//   thread may make calls to <f PIL_EnterCriticalSection> without blocking.
//   Note that if a thread makes multiple calls to <f PIL_EnterCriticalSection>,
//   it must call <f PIL_LeaveCriticalSection> an equal number of times or
//   the critical section will not be released.
//
//   The behavior is undefined to enter a critical section that has not 
//   been initialized using <f PIL_InitializeCriticalSection>.
//
// @rdesc This function does not return a value.
//
//

void
PIL_EnterCriticalSection(
    PilCriticalSection* pCriticalSection // @parm Pointer to the critical
                                         //   section to enter.
    );


///////////////////////////////////////////////////////////////////////
//
// PIL_InitializeCriticalSection
//
// @func <f PIL_InitializeCriticalSection> initializes the memory 
//   associated with a critical section object preparing it for use.
//   All critical sections must first be initialized with this call.
//
// @rdesc This function does not return a value.
//
//

void
_PIL_InitializeCriticalSection(
    PilCriticalSection* pCriticalSection // @parm Pointer to a chunk of memory
                                         //   to initialize as a critical
                                         //   section.
    , const char* name
    );

#define PIL_InitializeCriticalSection(a) \
    _PIL_InitializeCriticalSection(a, __FUNCTION__ );
	
///////////////////////////////////////////////////////////////////////
//
// PIL_LeaveCriticalSection
//
// @func <f PIL_LeaveCriticalSection> leaves a critical section, releasing
//   the next thread to enter the critical section if one is waiting.
//
//   The behavior is undefined if a critical section is left without first
//   being initialized using <f PIL_InitializeCriticalSection>.
//
// @rdesc This function does not return a value.
//
//

void
PIL_LeaveCriticalSection(
    PilCriticalSection* pCriticalSection // @parm Pointer to the critical
                                         //   to leave.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_TryEnterCriticalSection
//
// @func <f PIL_TryEnterCriticalSection> attempts to enter a critical
//   section.  The difference between this function and 
//   <f PIL_EnterCriticalSection> is that this function will return <ms FALSE>
//   if it could not enter the critical section because another thread has
//   the critical section.  It will not block.  It will return <ms TRUE> if
//   the critical section was successfully entered.  If the function returns
//   <ms TRUE>, the critical section was acquired successfully and must be
//   released using <f PIL_LeaveCriticalSection> when the thread is done
//   with it.
//
//   It is undefined to enter a critical section that has not been initialized
//   using <f PIL_EnterCriticalSection>.
//
// @rdesc Returns one of the following:
//
// @flag TRUE | The critical section was successfully entered.
// @flag FALSE | Another thread currently owns the critical section.
//
//

BOOL
PIL_TryEnterCriticalSection(
    PilCriticalSection* pCriticalSection // @parm Pointer to the critical
                                         //   to try to enter.
    );

#endif // __PILCS_H_INCLUDED__
