/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILDEBUG.H                                                //
// Module:   Platform Isolation Layer Debug Routines/Macros            //
// Author:   Venky Gokulrangan                                         //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
// PURPOSE: Debug Functions and Macro Declarations for PIL             //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//
// HISTORY:
//
/////////////////////////////////////////////////////////////////////////

#ifndef __PILDEBUG_H_INCLUDED
#define __PILDEBUG_H_INCLUDED

// 
// @doc
//
// @topic DEBUG services |
//
// The DEBUG services of PIL aid in the debugging of application written
// to the PIL API.  Most of these services automatically turn off and not
// generate code when compiled in release mode.  
//


#ifdef __cplusplus
   extern "C" {
#endif // __cplusplus

#define MAX_BUFFER_LEN 1024

// Maximum length of the string that can be printed

uint32_t
PIL_Trace(
     char*  Format,
     ... ) ;

#ifdef DEBUG

#ifndef ASSERT

//
// @func void | ASSERT |The <f ASSERT> macro checks the given condition and 
//   prints a message on the appropriate output device if that condition is
//   not true.  The format of the message is as follows:
//
// <ms *** ASSERTION FAILED:> <it filename> <ms line> <it number> <ms ***>
//
// Where <it filename> and <it number> will be filled in with the 
// appropriate information.
//
// Note when the <ms DEBUG> symbol is not defined, this macro will not
// generate any code.
//
// @parm BOOL | bCondition | The condition expression that must be <ms TRUE>
//   or the assertion fails.
//
  
#define ASSERT(x) if (!(x)) { PIL_Trace ( "\n *** ASSERTION FAILED: %s line %d: " #x " ***\n",  __FILE__, __LINE__ ); }
#endif //  ASSERT

#else

#ifndef ASSERT
#define ASSERT(x)
#endif // ASSERT

#endif //  DEBUG 

#ifdef DEBUG

#ifndef ASSERTMSG

//
// @func void | ASSERTMSG | The <f ASSERTMSG> macro works just like the <f ASSERT>
//   macro except that it allows a custom message to be printed when the
//   assertion fails:
//
// <ms *** ASSERTION FAILED:> <it filename> <ms line> <it number> <ms ***>
// <ms ***> <it custom message>
//
// Where <it filename>, <it number>, and <it custom message> will be filled
// in with the appropriate information.
//
// Note when the <ms DEBUG> symbol is not defined, this macro will not
// generate any code.
//
// @parm BOOL | bCondition | The condition expression that must be <ms TRUE>
//   or the assertion fails.
// @parm char* | Msg | The custom message to display if the assertion fails.
//


#define ASSERTMSG(x,msg) if (!(x)) { PIL_Trace ( "\n ** ASSERTION FAILED: %s line %d: " #x "\n", __FILE__, __LINE__ ) ; PIL_Trace ( " *** %s\n", msg ) ; }
#endif // ASSERTMSG

#else

#ifndef ASSERTMSG
#define ASSERTMSG(x,msg)
#endif // ASSERTMSG

#endif // DEBUG

#ifdef DEBUG

//
// @func int32_t | TRACE | The <f TRACE> macro works just like the <ms printf()>
//   C run-time library function except that when the <ms DEBUG> symbol is
//   not defined, it does not generate any code.  <f TRACE> accepts all
//   formatting commands that <ms printf()> does.  Consult an appropriate
//   C run-time library reference for acceptable parameters.
//
// @parm char* | fmt | The <ms printf>-style format string following by
//                     any additional parameters required by that format
//                     string.
//
// @rdesc The total number of characters output to the debug terminal.
//
// @func int32_t | TRACEx | The <f TRACEx> macros operate in the same manner
//   as the <f TRACE> macro except the number of parameters is fixed.  
//   <ms TRACE1> through <ms TRACE6> are defined.
//
//   These macros will not generate any code unless the <ms DEBUG> symbol is
//   defined.
//
// @parm char* | fmt | The <ms printf>-stype format string followed by
//                     1 to 6 additional parameters.
//
// @rdesc The total number of characters output to the debug terminal.
//

#define TRACE                          PIL_Trace
#define TRACE1(fmt,p1)                 PIL_Trace(fmt,p1)
#define TRACE2(fmt,p1,p2)              PIL_Trace(fmt,p1,p2)
#define TRACE3(fmt,p1,p2,p3)           PIL_Trace(fmt,p1,p2,p3)
#define TRACE4(fmt,p1,p2,p3,p4)        PIL_Trace(fmt,p1,p2,p3,p4)
#define TRACE5(fmt,p1,p2,p3,p4,p5)     PIL_Trace(fmt,p1,p2,p3,p4,p5)
#define TRACE6(fmt,p1,p2,p3,p4,p5,p6)  PIL_Trace(fmt,p1,p2,p3,p4,p5,p6)

#else

#ifdef _MSC_VER
#define TRACE                          
#else
#define TRACE(args...)
#endif

#define TRACE1(fmt,p1)                 
#define TRACE2(fmt,p1,p2)              
#define TRACE3(fmt,p1,p2,p3)           
#define TRACE4(fmt,p1,p2,p3,p4)        
#define TRACE5(fmt,p1,p2,p3,p4,p5)     
#define TRACE6(fmt,p1,p2,p3,p4,p5,p6)  

#endif

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
//                                                                   //
//                 R E S O U R C E  T R A C K I N G                  //
//                                                                   //
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

//
// @topic Resource Tracking |
//
// In special builds, PIL includes functionality to track all allocated
// resources.  This can be very useful in debugging to find memory leaks.
// It can also be used to determine at run-time how much of each resource
// is really being allocated.
//
// The functions described in this section enable applications to access
// the resource information that PIL is tracking.  This enables an 
// application to be built that will dynamically monitor resource 
// allocation.  Also included is a simple function to simply dump all
// resources onto an appropriate debugging device.
//
// Resource tracking is always enabled in the <ms DEBUG> builds of PIL.  This
// aids in detecting bad handles.  A retail build of PIL can also have
// resource tracking by enabling the <ms RESTRACK> flag when compiling.
//

typedef struct {
    uint32_t cApc;
    uint32_t cCriticalSection;
    uint32_t cEvent;
    uint32_t cHeap;
    uint32_t cLib;
    uint32_t cMailbox;
    uint32_t cMemory;
    uint32_t cSemaphore;
    uint32_t cSocket;
    uint32_t cThread;
    uint32_t cTimer;
} PilResCounts;

enum PilResourceTypes {
    RT_APC,
    RT_CRITICAL_SECTION,
    RT_EVENT,
    RT_HEAP,
    RT_MAILBOX,
    RT_MEMORY,
    RT_SEMAPHORE,
    RT_THREAD,
    RT_TIMER,
    RT_LIB,
    RT_SOCKET
};

///////////////////////////////////////////////////////////////////////
//
// PIL_DumpResources
//
// @func Dumps the currently allocated resources on a suitable debugging
//   terminal.
//
// @rdesc This function does not return a value.
//
//

void
PIL_DumpResources(
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_GetResourceCount
//
// @func Retrieves the current number of each tracked resource.
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK            | The operation completed successfully.
// @flag PIL_INVALID_PARAM | <p pCounts> is not a valid pointer.
//
//

PilResult
PIL_GetResourceCount(
    PilResCounts* pCounts
    );


#ifdef __cplusplus
   };
#endif // __cplusplus

#endif // __PILDEBUG_H_INCLUDED__
