/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILEVNT.H                                                 //
// Module:   Platform Isolation Layer Events                           //
// Author:   Daniel Baumberger                                         //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
// PURPOSE:                                                            //
//                                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//
// HISTORY:
//
// 
/////////////////////////////////////////////////////////////////////////

#ifndef __PILEVNT_H_INCLUDED__
#define __PILEVNT_H_INCLUDED__

//
// @doc
// @topic Events (Required) |
//
// Events are a synchronization method for determining if something has
// happened.  They maintain only a "signaled" or "non-signaled" state.
// Threads can wait on an event to be awoken when something happens.
// They are very similar to binary semaphores with the additional semantic
// that the state, signaled or non-signaled, can be changed without
// blocking.  The only blocking call is <f PIL_WaitForEvent> which
// waits until an event becomes signaled.
//
// Events come in two types: manual or automatic reset.  Manual reset
// events require an explicit call to set the event to non-signaled once
// it has become signaled.  Automatic reset events automatically reset back
// to non-signaled once a thread has been awoken.  In other words, manual
// reset events let all threads waiting on the event through while the 
// automatic reset events allow only a single thread.
//
// Events are created via <f PIL_CreateEvent>.  At creation time, the 
// signal state can be set and what type of event it is (manual or automatic
// reset).  Events are destroyed using <f PIL_DestroyEvent> when they are
// no longer necessary.  <f PIL_ResetEvent> sets a manual reset event back
// to the non-signaled state.  <f PIL_SetEvent> sets either type of event
// to the signaled state.  Finally, <f PIL_WaitForEvent> waits until an 
// event is signaled, with an optional timeout.
//

typedef void* PilEvent;

///////////////////////////////////////////////////////////////////////
//
// PIL_CreateEvent
//
// @func <f PIL_CreateEvent> creates a new manual or automatic reset
//   event.  <p bManualReset> determines if the event is manual or
//   automatic reset.  <ms TRUE> (non-zero) creates a manual reset event,
//   <ms FALSE> (zero) creates an automatic reset event. 
//   
//   <p bSignaled> determines the initial state of the event.  <ms TRUE>
//   sets the initial state to signaled.  <ms FALSE> will make the event
//   non-signaled.  For most purposes, this parameter will be <ms FALSE>.
//
// @rdesc Returns one of the following:
//
// @flag PilEvent | The handle of the newly created event.
// @flag NULL     | The event could no be created.
//
//

PilEvent
PIL_CreateEvent(
    BOOL bManualReset,  // @parm The type of event to create.  <ms TRUE> if
                        //   the event is a manual reset or
                        //   <ms FALSE> if it is automatic.
    BOOL bSignaled      // @parm <ms TRUE> if the initial state of the event
                        //   is signaled or <ms FALSE> if it is non-signaled.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_DestroyEvent
//
// @func <f PIL_DestroyEvent> destroys an event created with 
//   <f PIL_CreateEvent>.  The behavior is undefined if an event is
//   destroyed with threads waiting on it.
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK             | The event was successfully destroyed.
// @flag PIL_INVALID_HANDLE | The handle of the event is not a valid event
//                            handle.
//
//

PilResult
PIL_DestroyEvent(
    PilEvent Event  // @parm The handle of the event to destroy.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_ResetEvent
//
// @func <f PIL_ResetEvent> resets the state of a manual reset event
//   back to non-signaled.  It does nothing for an automatic reset event.
//
// @rdesc Return one of the following:
//
// @flag PIL_OK             | The event has been successfully reset to
//                            non-signaled.
// @flag PIL_INVALID_HANDLE | The handle of the event is not a valid event
//                            handle.
//
//

PilResult
PIL_ResetEvent(
    PilEvent Event  // @parm The handle of the manual reset event to reset
                    //   to non-signaled.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_SetEvent
//
// @func <f PIL_SetEvent> sets the state of an event to the signaled
//   state.  For manual reset events, the event stays in the signaled
//   state and all threads waiting (or come to wait when in the signaled
//   state) will be let through until the event is explicitly reset.  For
//   automatic reset events, a single thread waiting on the event is woken
//   (or the first thread to wait after being signaled) and then the state
//   is automatically set back to non-signaled.
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK             | The event was successfully signaled.
// @flag PIL_INVALID_HANDLE | The handle of the event is not a valid event
//                             handle.
//

PilResult
PIL_SetEvent(
    PilEvent Event  // @parm The handle of the event to set to the signaled
                    //   state.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_WaitForEvent
//
// @func <f PIL_WaitForEvent> waits for an event to become signaled.
//   The function returns when the event is set to the signaled state
//   or when the timeout period expires.  The state of the event after
//   a successful wait is dependent on the type of event.  For manual
//   reset events, the <f PIL_WaitForEvent> does not affect its state.
//   For automatic reset events, the state is automatically reset to
//   non-signaled after the first successful wait.
//
//   The timeout specifies how long the thread should be suspended waiting
//   for the event to be signaled in milliseconds.  The value of
//   <ms PIL_INFINITE_TIMEOUT> will wait forever.  A value of <ms 0> allows
//   the state of the event to be checked without blocking.
//
//   Note that on all platforms, the granularity of the wait may not be one
//   millisecond but may be more like 10 or 20 milliseconds.  In that case,
//   the wait will be rounded up to the nearest integral unit.
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK             | The wait was successful and the event is
//                            signaled.
// @flag PIL_INVALID_HANDLE | The handle of the event is not a valid event
//                            handle.
// @flag PIL_TIMEOUT        | The timeout period expired before the event
//                            became signaled.
//
//

PilResult
PIL_WaitForEvent(
    PilEvent Event,     // @parm The handle of the event to wait for
    uint32_t dwTimeout  // @parm The time, in milliseconds, to wait for the
                        //   event or <ms PIL_INFINITE_TIMEOUT> to wait 
                        //   forever.
    );

#endif // __PILEVNT_H_INCLUDED__

