/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILHEAP.H                                                 //
// Module:   PIL  Heap Management Internal Declarations                //
// Author:   Daniel Baumberger                                         //
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      // 
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
//                                                                     //
//                                                                     //
// PURPOSE: PIL Interface to heap management for client code           //
//                                                                     //
/////////////////////////////////////////////////////////////////////////

#ifndef _HEAP_H_INCLUDED_
#define _HEAP_H_INCLUDED_

//
// @doc
// @topic Heap management (Required) |
//
// PIL's heap management allows the client to create heaps and manage them
// separately from the underlying operating system.  An application can
// also create multiple heaps in order to better manage buffering.  For
// example, if an application manipulates buffers of a couple different
// sizes, a heap could be created for each so that each heap manages buffers
// of only one size.  PIL does not have a pre-set limit on the number of
// heaps.
//
// A new heap is created using <f PIL_CreateHeap>.  At this time the 
// total size of the heap is given.  Blocks of memory can be allocated from
// a specific heap using <f PIL_HeapAlloc>.  Blocks are returned to the
// heap using <f PIL_HeapFree>.  When a heap is no longer needed, it can
// be freed using <f PIL_DestroyHeap>.  Finally, blocks from a heap can
// be shrunk using <f PIL_HeapShrink>.  The heap manager does not allow
// blocks to be increased in size.
//
// The heap manager uses 8 bytes of overhead per allocation.  The initial
// size of the heap is guaranteed to be the size given during 
// <f PIL_CreateHeap>.  Each sub-allocation requires 8 bytes out of the
// heap.  So if a heap is created with a size of 10,240 bytes, you cannot
// get 10 allocations of 1,024 bytes out of the heap, but rather 9 allocations
// of 1.024 bytes with 952 bytes left in the heap.  Memory is allocated in a
// first-fit algorithm.  Free blocks of memory are automatically combined
// during <f PIL_HeapFree>.
//

#ifdef __cplusplus
   extern "C" {
#endif

typedef struct tHeapHeader
{
  struct    tHeapHeader* pPrev;    // Pointer to the previous block
  uint32_t  dwSize:31;             // The size of the block
  BOOL      bInUse:1;              // TRUE if allocated, or FALSE

} HeapHeader, *PHeapHeader;

typedef struct tHeapDesc
{

#ifdef RESTRACK
   LIST_NODE Link ;
#endif

  PHeapHeader  phhHead;  // The header record of the mem pool
  PHeapHeader  phhTail;  // The tail of the memory pool
  uint32_t     dwSize;   // The size of the entire pool
  PilSemaphore LockSem ; // Sync. access to heap

} HeapDesc;

typedef void*    PilHeap ;
typedef void**   PPVOID ;

///////////////////////////////////////////////////////////////////////
//
// PIL_CreateHeap
//
// @func <f PIL_CreateHeap> creates a new heap with the given number
//   of bytes.  When the heap is first created, it is guaranteed to contain
//   <p MaxHeapSize> bytes, if allocated in one chunk.  Each sub-allocation
//   requires 8 bytes of overhead, which is also allocated from the heap.
//
//   The maximum size of any one heap is 2^31 or 2,147,483,648 bytes.
//
// @rdesc Returns one of the following:
// 
// @flag PilHeap | The handle of the newly created heap.
// @flag NULL    | The heap could not be allocated from the underlying 
//                 operating system.
//
//

PilHeap
PIL_CreateHeap (
    uint32_t MaxHeapSize  // @parm The maximum size of the heap in bytes.
    );  

//////////////////////////////////////////////////////////////////////////////
//
// PIL_DestroyHeap
//
// @func <f PIL_DestroyHeap> de-allocates a heap created using
//   <f PIL_CreateHeap>.  All blocks allocated from that heap become
//   invalid after this call.
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK             | The heap was successfully destroyed.
// @flag PIL_INVALID_HANDLE | The handle of the heap is not a valid heap
//                            handle.
//

PilResult
PIL_DestroyHeap (
    PilHeap hHeap  // @parm The handle of the heap to destroy.
    );  


//////////////////////////////////////////////////////////////////////////////
//
// PIL_HeapAlloc
//
// @func <f PIL_HeapAlloc> allocates <p Size> bytes from the given heap.
//   The maximum size of any single allocation is the size of the heap when
//   created.
//
// @rdesc One of the following:
//
// @flag void* | Pointer to the newly allocated memory block.
// @flag NULL  | The heap does not contain enough free memory to allocate
//               the block.
//
//

void*
PIL_HeapAlloc (
    PilHeap  hHeap,  // @parm The handle of the heap to allocate memory from.
    uint32_t Size    // @parm The size, in bytes, of the block to allocate.
    ) ;   


//////////////////////////////////////////////////////////////////////////////
//
// PIL_HeapFree
//
// @func <f PIL_HeapFree> returns a block allocated using <f PIL_HeapAlloc>
//   back into the heap.  When a block is freed, the heap manager performs a
//   quick heap compaction to maximize the free space in the heap.  The
//   behavior is undefined to free a block to a heap that was not allocated from
//   the heap.
//

PilResult
PIL_HeapFree (
    PilHeap hHeap,  // @parm The handle of the heap to free the block back
                    //   into.
    void*   pBuf    // @parm The pointer of the buffer to free back into the
                    //   heap.
     );

//////////////////////////////////////////////////////////////////////////////
//
// PIL_HeapShrink
//
// @func <f PIL_HeapShrink> allows a block allocated from a heap to be made
//   smaller, freeing the reset of the block back into the heap.  Not all 
//   blocks can be shrunk due to overhead in the heap.  A block cannot be
//   made larger using this function.  The behavior is undefined if a block
//   is shunk in a heap that it was not allocated from.
//
// @rdesc Returns one of the following:
// 
// @flag PIL_OK             | The operation completed successfully.
// @flag PIL_INVALID_HANDLE | The handle of the heap is not a valid heap
//                            handle.
// @flag PIL_INVALID_PARAM  | The new size is not less than the old size.
// @flag PIL_ERROR          | The block cannot be resized due to lack of
//                            space in the heap.
//
//

PilResult
PIL_HeapShrink (
    PilHeap   hHeap,   // @parm The handle of the heap the block belongs to.
    void*     pBuf,    // @parm The pointer to the buffer to shrink.
    uint32_t  NewSize  // @parm The new, smaller size of the block.
    );

//////////////////////////////////////////////////////////////////////////////
//
// Function:   PIL_ScanHeap
// Parameters: 1. Handle to the Heao to be scanned
//
// Purpose:    This goes through the list from block to next and prints
//             the information contained in every block
//
// Returns:    NONE
//
// NOTES:
//       This function is just a Diagnostic Function for DEBUG purposes
//

#ifdef DEBUG
void
PIL_ScanHeap (
    PilHeap hHeap  // Debug Routine
    );
#endif

#ifdef __cplusplus
   }
#endif

#endif // _HEAP_H_INCLUDED_
