/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILIO.H                                                   //
// Module:   Platform Isolation Layer IO and PCI Routines              //
// Author:   JP Sugarbroad                                             //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////

#ifndef __PILPCI_H_INCLUDED__
#define __PILPCI_H_INCLUDED__

//
// @doc
//
// @topic Input/Output register access (Optional) |
//
// The Input/Ouput register access service is mainly for drivers.  It
// allows a driver to access the hardware ports in a fairly portable
// manner.  It is not entirely portable since not all architectures
// will use port IO.  However, large sections of driver code can be 
// ported between environments easily when using these functions rather
// than the native OS.
//
// Support is broken into two pieces.  The first lets a driver find
// its associated hardware on the PCI bus by searching by either class
// or by vendor/device ID.  Once the device is found, PIL provides functions
// to access the PCI configuration space for that device.
//
// The other type of access is raw port IO.  This, of course, requires the
// port number to be know.  It does not, however, require searching for
// a device on the PCI bus.
//

// PilPciDevice is an opaque handle to a PCI device in the system
typedef void *PilPciDevice;

/////////////////////////////////////////////////////////////////////////
//
// PILOS_FindPciDeviceSignature
//
// @func <f PILOS_FindPciDeviceSignature> locates a specific PCI device
//   by its signature (vendor id/product id).
//
// @rdesc Returns one of the following:
//
// @flag PilPciDevice | The handle of the PCI device.
// @flag NULL         | The specific PCI device could not be found.
//
//

PilPciDevice
PILOS_FindPciDeviceSignature( 
    uint16_t     Vendor, // @parm The vendor number to match from the device
                         //   configuration space.
    uint16_t     Id,     // @parm The device ID number to match from the device 
                         //   configuration space.
    PilPciDevice Last    // @parm The value returned from a previous call to 
                         //   this function to continue searching from.
    );

///////////////////////////////////////////////////////////////////////
//
// PILOS_FindPciDeviceClass
//
// @func <f PILOS_FindPciDeviceClass> locates a PCI device based on its
//   class.
//
// @rdesc Returns one of the following:
//
// @flag PilPciDevice | The handle of the PCI device.
// @flag NULL         | The specific PCI device could not be found.
//
//

PilPciDevice
PILOS_FindPciDeviceClass( 
    uint16_t     Class, // @parm The class code to match from the device 
                        //   configuration space.
    PilPciDevice Last   // @parm The value returned from a previous call to 
                        //   this function to continue searching from.
    );


/////////////////////////////////////////////////////////////////////////
//
// PILOS_ReadPortByte
//
// @func <f PILOS_ReadPortByte> reads a single byte from <p wPort>.
//
// @rdesc The byte value read from the port.
//
//

uint8_t
PILOS_ReadPortByte( 
    uint16_t wPort // @parm The port number to read a byte from.
    );

/////////////////////////////////////////////////////////////////////////
//
// PILOS_ReadPortWord
//
// @func <f PILOS_ReadPortWord> reads a single word from <p wPort>.
//
// @rdesc The word value read from the port.
//
//

uint16_t
PILOS_ReadPortWord( 
    uint16_t wPort // @parm The port number to read a word from.
    );

/////////////////////////////////////////////////////////////////////////
//
// PILOS_ReadPortDword
//
// @func <f PILOS_ReadPortDword> reads a single dword from <p wPort>.
//
// @rdesc The dword value read from the port.
//
//

uint32_t
PILOS_ReadPortDword( 
    uint16_t wPort // @parm The port number to read a dword from.
    );

/////////////////////////////////////////////////////////////////////////
//
// PILOS_WritePortByte
//
// @func <f PILOS_WritePortByte> writes a single byte to <p wPort>.
//
// @rdesc This function does not return a value.
//
//

void
PILOS_WritePortByte( 
    uint16_t wPort, // @parm The port to write <p Value> to.
    uint8_t Value   // @parm The value to write to <p wPort>.
    );

/////////////////////////////////////////////////////////////////////////
//
// PILOS_WritePortWord
//
// @func <f PILOS_WritePortWord> writes a single word to <p wPort>.
//
// @rdesc This function does not return a value.
//
//

void
PILOS_WritePortWord( 
    uint16_t wPort, // @parm The port to write <p Value> to.
    uint16_t Value  // @parm The value to write to <p wPort>.
    );

/////////////////////////////////////////////////////////////////////////
//
// PILOS_WritePortDword
//
// @func <f PILOS_WritePortDword> writes a single dword to <p wPort>.
//
// @rdesc This function does not return a value.
//
//

void
PILOS_WritePortDword( 
    uint16_t wPort, // @parm The port to write <p Value> to.
    uint32_t Value  // @parm The value to write to <p wPort>.
    );

/////////////////////////////////////////////////////////////////////////
//
// PILOS_ReadPciConfigByte
//
// @func <f PILOS_ReadPciConfigByte> reads a single byte value from the
//   device at the given offset in the PCI configuration space.
//
// @rdesc The value from the offset in the PCI configuration space.
//
//

uint8_t
PILOS_ReadPciConfigByte( 
    PilPciDevice Device, // @parm The specific device to read from.
    uint8_t      Offset  // @parm The byte offset into the device's PCI 
                         //   configuration space to read from.
    );

/////////////////////////////////////////////////////////////////////////
//
// PILOS_ReadPciConfigWord
//
// @func <f PILOS_ReadPciConfigWord> reads a single word value from the
//   device at the given offset in the PCI configuration space.
//
// @rdesc The value from the offset in the PCI configuration space.
//
//

uint16_t
PILOS_ReadPciConfigWord( 
    PilPciDevice Device, // @parm The specific device to read from.
    uint8_t      Offset  // @parm The byte offset into the device's PCI 
                         //   configuration space to read from.
    );

/////////////////////////////////////////////////////////////////////////
//
// PILOS_ReadPciConfigDword
//
// @func <f PILOS_ReadPciConfigDword> reads a single dword value from the
//   device at the given offset in the PCI configuration space.
//
// @rdesc The value from the offset in the PCI configuration space.
//
//

uint32_t
PILOS_ReadPciConfigDword( 
    PilPciDevice Device, // @parm The specific device to read from.
    uint8_t      Offset  // @parm The byte offset into the device's PCI 
                         //   configuration space to read from.
    );

/////////////////////////////////////////////////////////////////////////
//
// PILOS_WritePciConfigByte
//
// @func <f PILOS_WritePciConfigByte> writes a single byte value into the
//   device's PCI configuration space at the given byte offset.
//
// @rdesc This function does not return a value.
//
//

void
PILOS_WritePciConfigByte( 
    PilPciDevice Device, // @parm The device to write PCI configuration data 
                         //   from.
    uint8_t      Offset, // @parm The byte offset into the PCI configuration
                         //   space to read from.
    uint8_t      Value   // @parm The value to write into the PCI configuration
                         //   space.
    );

/////////////////////////////////////////////////////////////////////////
//
// PILOS_WritePciConfigWord
//
// @func <f PILOS_WritePciConfigWord> writes a single word value into the
//   device's PCI configuration space at the given byte offset.
//
// @rdesc This function does not return a value.
//
//

void
PILOS_WritePciConfigWord( 
    PilPciDevice Device, // @parm The device to write PCI configuration data 
                         //   from.
    uint8_t      Offset, // @parm The byte offset into the PCI configuration
                         //   space to read from.
    uint16_t     Value   // @parm The value to write into the PCI configuration
                         //   space.
    );

/////////////////////////////////////////////////////////////////////////
//
// PILOS_WritePciConfigDword
//
// @func <f PILOS_WritePciConfigDword> writes a single dword value into the
//   device's PCI configuration space at the given byte offset.
//
// @rdesc This function does not return a value.
//
//

void
PILOS_WritePciConfigDword( 
    PilPciDevice Device, // @parm The device to write PCI configuration data 
                         //   from.
    uint8_t      Offset, // @parm The byte offset into the PCI configuration
                         //   space to read from.
    uint32_t     Value   // @parm The value to write into the PCI configuration
                         //   space.
    );


#endif
