/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILMBOX.H                                                 //
// Module:   Platform Isolation Layer Mailbox Interface                //
// Author:   Venky Gokulrangan                                         //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
// PURPOSE: PIL Interface to the mailbox module                        //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//
// HISTORY:
//
/////////////////////////////////////////////////////////////////////////

#ifndef __MBOX_H_INCLUDED__
#define __MBOX_H_INCLUDED__

//
// @doc
//
// @topic Mailbox management (Required) |
//
// The mailbox manager in PIL provides basic, platform independent messaging
// services.  Mailboxes allow multiple threads to pass information back and
// forth between them.  It also provides a way to synchronize on those
// messages.  In other words, a thread can block waiting for messages in its
// mailbox.  When a message arrives, it wakes up, reads the message, and does
// whatever action is required.
//
// The format of the messages is not defined.  When the mailbox is created,
// the application tells PIL how big each message is and how many to store.
// PIL will allocate the necessary memory to maintain the queue of messages.
// PIL retains a copy of the data and not a pointer.  Therefore, once a
// message has been written into the mailbox, the original does not have to
// be preserved.  Pointers can easily be passed in this paradigm by specifying
// a size of pointer for the message size.
//
// A new mailbox is created using <f PIL_CreateMailbox>.  Here the size of
// each message and the number of messages to store is communicated.  There is
// no inherent limit on the number of mailboxes, the size of messages, or the
// number of messages other than the memory required to store these 
// structures.  When a mailbox is no longer needed, it can be destroyed using
// <f PIL_DestroyMailbox>.
//
// <f PIL_ReadMailbox> reads a message from the mailbox.  If there are no
// messages in the mailbox, it can optionally block waiting for messages.
// <f PIL_WriteMailbox> copies a message into the mailbox, waking any 
// threads that may be waiting on the mailbox.  <f PIL_WriteMailbox> can
// also optionally block if the mailbox is full.
//
// Note that if multiple threads are waiting to read a mailbox, messages
// will be given to threads in a first in first out order.  That is, the
// first thread waiting will get the first message, the next thread the
// second, and so on.  The same holds true for writing to a mailbox.
//

#ifdef __cplusplus
extern "C" {
#endif

typedef void* PilMailbox ; // essentially a pointer to MBoxDesc

///////////////////////////////////////////////////////////////////////
//
// PIL_CreateMailbox
//
// @func <f PIL_CreateMailbox> creates a new message queue. <p QueueLength>
//   messages can be stored in the mailbox, each occupying <p MessageLength>
//   bytes.  <f PIL_CreateMailbox> pre-allocates all the memory necessary
//   to store all <p QueueLength> messages so that at run-time the mailbox
//   either has available slots or it does not.  It does not go out to the
//   OS and try to get more.
//
// @rdesc Returns one of the following:
//
// @flag PilMailbox | The handle of the newly created mailbox.
// @flag NULL       | The mailbox could not be created.  The most common
//                    reason is there isn't enough memory to allocate
//                    the mailbox data structures.
//
//

PilMailbox
PIL_CreateMailbox ( 
    uint32_t  QueueLen,       // @parm The size of each message in bytes stored
                             //   in each mailbox slot.
    uint32_t  MessageLength // @parm The maximum number of messages that can
                             //   be stored in the mailbox.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_DestroyMailbox
//
// @func <f PIL_DestroyMailbox> destroys a mailbox created with
//   <f PIL_CreateMailbox>, releasing all memory associated with the
//   mailbox.  The behavior is undefined if a mailbox is destroyed that
//   has readers or writers blocked on it.  All messages in the queue are
//   lost when the mailbox is destroyed.  Since the contents of the messages
//   are not fixed, if the message pointers to allocated resources, those
//   resources will not be freed.
//
// @rdesc Returns on of the following:
//
// @flag PIL_OK             | The operation completed successfully.
// @flag PIL_INVALID_HANDLE | The handle of the mailbox is not a valid mailbox
//                            handle.
//
//

PilResult 
PIL_DestroyMailbox (
    PilMailbox hMailbox  // @parm The handle of the mailbox to destroy.
    );


///////////////////////////////////////////////////////////////////////
//
// PIL_ReadMailbox
//
// @func <f PIL_ReadMailbox> reads the next message from the mailbox,
//   copying the message into the buffer provided by the caller in
//   <p pMessage>.  If a message is not available, the call will block
//   for the period specified by <p Timeout>.
//
//   This function may be used in non-blocking contexts, such as time
//   and Apc callbacks, if the timeout is zero.
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK             | A message was successfully read and copied into
//                            the given buffer.
// @flag PIL_INVALID_HANDLE | The handle of the mailbox is not a valid
//                            mailbox handle.
// @flag PIL_INVALID_PARAM  | <p pMessage> is not a valid pointer.
// @flag PIL_TIMEOUT        | The timeout period expired before a message
//                            was available.
//
//

PilResult
PIL_ReadMailbox (
   PilMailbox hMailbox,  // @parm The handle of the mailbox from which to read
                         //   a message.
   void*      pmbmMsg,   // @parm A pointer to a buffer where the message
                         //   contents will be copied.
   uint32_t   Timeout    // @parm The time to wait for a message in
                         //   milliseconds to wait for the message or
                         //   <ms PIL_INFINITE_TIMEOUT> to wait forever.
   );


///////////////////////////////////////////////////////////////////////
//
// PIL_WriteMailbox
//
// @func <f PIL_WriteMailbox> copies the message pointed to by <p pMessage>
//   into the next available slot in the mailbox queue.  If no slot is
//   available to write the message, <f PIL_WriteMailbox> will block for
//   <p Timeout> milliseconds.
//
//   The function may be used in non-blocking contexts, such as timer and
//   Apc callbacks, if the timeout is zero.
//
// @rdesc Returns on of the following:
//
// @flag PIL_OK             | The message was successfully written into
//                            the mailbox.
// @flag PIL_INVALID_HANDLE | The handle of the mailbox is not a valid
//                            mailbox handle.
// @flag PIL_INVALID_PARAM  | <p pMessage> is not a valid pointer.
// @flag PIL_TIMEOUT        | The timeout period expired before a slot
//                            was available.
//

PilResult
PIL_WriteMailbox (
    PilMailbox   hMailbox, // @parm The handle of the mailbox in which to
                           //   write a message.
    void*        pmbmMsg,  // @parm A pointer to a buffer where the message
                           //   contents will be copied from.
    uint32_t     Timeout   // @parm The time to wait for a free slot in
                           //   milliseconds or <ms PIL_INFINITE_TIMEOUT> to
                           //   wait forever.
    );


#ifdef __cplusplus
}
#endif

#endif // __PILMBOX_H_INCLUDED__
