/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILMEM.H                                                  //
// Module:   Platform Isolation Layer Memory Routines                  //
// Author:   Daniel Baumberger                                         //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
// PURPOSE:                                                            //
//                                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//
// HISTORY:
//
/////////////////////////////////////////////////////////////////////////

#ifndef __PILMEM_H_INCLUDED__
#define __PILMEM_H_INCLUDED__

//
// @doc
//
// @topic Memory (Required) |
//
// The memory routines allow a client to acquire or return
// memory to the underlying operating system.  Memory can be allocated
// by specifying a number of bytes using <f PIL_AllocateMemory> or a
// specified number of pages using <f PIL_AllocatePages>.  Both routines
// allow a flag to specify if the memory should be initialized to zero and
// optionally with pages, to lock the pages in memory.  The inverse operation
// is <f PIL_FreeMemory> and <f PIL_FreePages>.  These two functions cannot
// be intermixed.  I.e. if the memory is allocated using 
// <f PIL_AllocateMemory>, it needs to be freed with <f PIL_FreeMemory>.
// 
// PIL also provides a mechanism to get the page size from the underlying
// environment in a portable manner via the <f PIL_GetPageSize> function.
//

#define PIL_MEM_ZERO_INIT    0x00000001
#define PIL_MEM_PAGE_LOCK    0x00000002

////////////////////////////////////////////////////////////////////////
//
// PIL_AllocateMemory
//
// @func <f PIL_AllocateMemory> allocates a new block of memory from the
//   underlying operating system.  The memory block can be optionally
//   initialized to zero by specifying the <ms PIL_MEM_ZERO_INIT> flag.
//   There is no special alignment on the block of memory.
//
// @rdesc Returns one of the following:
//
// @flag void* | Pointer to the newly allocated block of memory.
// @flag NULL  | The memory could not be allocated.
//
//

void*
PIL_AllocateMemory(
    uint32_t Size,   // @parm The size of the block in bytes.
    uint32_t Flags   // @parm Special options for the allocation.  Currently
                     //   only <ms PIL_MEM_ZERO_INIT> is supported.
    );

////////////////////////////////////////////////////////////////////////
//
// PIL_AllocatePages
//
// @func <f PIL_AllocatePages> allocates one or more contiguous pages.
//   <p pPhysAddr>, when specified, will contain the physical address
//   of the pages on platforms where it is supported.  Otherwise, it
//   will contain the same value as the linear address.
//
//   All platforms may not support the <ms PIL_MEM_PAGE_LOCK> flag.
//   The allocation will fail if this flag is specified and page locked
//   memory is not supported.
//
// @rdesc Returns one of the following:
//
// @flag void* | The linear address of the newly allocated block.
// @flag NULL  | The pages could not be allocated or could not be locked.
//
//

void*
PIL_AllocatePages(
    uint32_t  NumPages,  // @parm The number of pages to allocate.  This
                         //   must be at least 1.
    uint32_t  Flags,     // @parm Optional flags for allocation.
                         //   <ms PIL_MEM_ZERO_INIT> will initialize all the
                         //   pages to 0.  <ms PIL_MEM_PAGE_LOCK> will lock
                         //   the pages in memory so they are not swapped
                         //   out to disk.
    uint32_t* pPhysAddr  // @parm Optional pointer to store the physical
                         //   address of the newly allocated pages.
    );

////////////////////////////////////////////////////////////////////////
//
// PIL_FreeMemory
//
// @func <f PIL_FreeMemory> frees memory allocated using 
//   <f PIL_AllocateMemory>.  Freeing memory via <f PIL_FreeMemory>
//   that was not allocated using <f PIL_AllocateMemory> is undefined
//   and may cause the system to crash.
//
// @rdesc This function does not return a value.
//
//

void
PIL_FreeMemory(
    void* pMem  // @parm Pointer to the linear address of the memory to free.
    );

////////////////////////////////////////////////////////////////////////
//
// PIL_FreePages
//
// @func <f PIL_FreePages> frees pages allocated via <f PIL_AllocatePages>.
//   The number of pages specified to <f PIL_FreePages> must match that
//   specified to <f PIL_AllocatePages> or problems may result.  Freeing
//   memory via <f PIL_FreePages> that was not allocated using
//   <f PIL_AllocatePages> is undefined and may cause the system to crash.
//
// @rdesc This function does not return a value.
//
//

void
PIL_FreePages(
    void*    pMem,     // @parm The linear address of the pages.
    uint32_t NumPages  // @parm The number of pages to free.
    );

////////////////////////////////////////////////////////////////////////
//
// PIL_GetPageSize
//
// @func <f PIL_GetPageSize> returns the size of a page in bytes on
//   the native operating environment.  This page size value can then
//   be used to calculate the number of pages necessary with
//   <f PIL_AllocatePages>.
//
// @rdesc The size of the page in bytes.
//
//

uint32_t
PIL_GetPageSize(
    );

#endif // __PILMEM_H_INCLUDED__



