/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILSEM.H                                                  //
// Module:   Platform Isolation Layer Semaphores                       //
// Author:   Daniel Baumberger                                         //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
// PURPOSE:                                                            //
//                                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//
// HISTORY:
//
// 
/////////////////////////////////////////////////////////////////////////

#ifndef __PILSEM_H_INCLUDED__
#define __PILSEM_H_INCLUDED__

//
// @doc
//
// @topic Semaphores (Required) |
//
// Semaphores are the basic sychronization mechanism in PIL.  All PIL
// semaphores are counting semaphores, but a macro is provided to make
// a "binary" semaphore.  Semaphores are first-come-first-serve.  I.e.
// the priority of the thread waiting on the semaphore does not matter.
//
// When created, a semaphore is given both an initial count and a maximum
// count.  The maximum count is the total number of resources the semaphore
// is protecting.  For simple synchronization, the count will be 1.  The
// initial count is the number of resources free upon creation.  Most of the
// time, this will be 0, so that all threads initially block on the semaphore,
// or the same as the maximum count.  Note that specifying a value in
// between the initial and maximum count may not work on all platforms.
//
// Semaphores are created with <f PIL_CreateSemaphore>.  When successful, this
// call will return a handle to the newly created semaphore.  
// <f PIL_AcquireSemaphore> will wait on the semaphore, or in other words, 
// reduce the count of the semaphore by 1.  If the count is 0, the thread
// will block.  The timeout parameter to <f PIL_AcquireSemaphore> determines
// how long to wait in this case.
//
// <f PIL_ReleaseSemaphore> increments the count of the semaphore by 1, 
// possibly signaling a thread waiting on the semaphore.  
// <f PIL_ReleaseSemaphore> cannot block but a context switch may occur if
// the signaled thread is higher priority.
//
// All semaphores need to be destroyed using <f PIL_DestroySemaphore> when
// they are no longer needed.  Note that destroying a semaphore that
// another thread is using is undefined and will probably cause the machine
// to crash.
//

typedef void* PilSemaphore;

///////////////////////////////////////////////////////////////////////
//
// PIL_AcquireSemaphore
//
// @func <f PIL_AcquireSemaphore> attempts to grab the semaphore, possibly
//   blocking for a period of time.  It effectively decrements the semaphore's
//   count by one.   
//
//   A timeout value of zero specifies that the acquire should not block.
//   If the semaphore is free, <ms PIL_OK> is returned and the semaphore 
//   is acquired.  If the semaphore is not free, <f PIL_AcquireSemaphore>
//   will return <ms PIL_TIMEOUT> to signal that the semaphore is not free
//   and the zero-length timeout expired.
//
//   Note that on all platforms, the timeout granularity may not be one 
//   millisecond but may be something larger such as 10 or 20 milliseconds.
//   Values will be rounded up to the nearest integral value.
//
// @rdesc Returns on of the following:
//
// @flag PIL_OK      | The semaphore was acquired successfully.
// @flag PIL_TIMEOUT | The timeout period expired before the semaphore could
//                     be acquired.
//
//

PilResult
PIL_AcquireSemaphore(
    PilSemaphore Sem,       // @parm The handle of the semaphore to acquire.
    uint32_t     Timeout    // @parm The amount of time, in milliseconds, 
                            //   to wait for the semaphore.  A value of 
                            //   <ms PIL_INFINITE_TIMEOUT> can be specified
                            //   to wait forever.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_CreateBinarySemaphore
//
// @func <f PIL_CreateBinarySemaphore> is equivalent to 
//   <f PIL_CreateSemaphore>( bInitalCount, 1 ).  Please refer to 
//   <f PIL_CreateSemaphore> for more details.
//
// @parm BOOL | bInitialCount | <ms TRUE> (non-zero) if the semaphore is
//                              initially free, <ms FALSE> (zero) if it is
//                              not.
//
// @rdesc Returns one of the following:
//
// @flag PilSemaphore | The handle of the newly created semaphore object.
// @flag NULL         | The semaphore could not be created.
//
//

#define PIL_CreateBinarySemaphore( bInitialCount ) PIL_CreateSemaphore( bInitialCount, 1 )


///////////////////////////////////////////////////////////////////////
//
// PIL_CreateSemaphore
//
// @func <f PIL_CreateSemaphore> creates a new semaphore object and returns
//   the handle of the new object.  If the object cannot be created, most
//   likely because of memory constraints, the function returns <ms NULL>.  
//
//   The <p InitialCount> parameter determines the starting value of the
//   counting semaphore.  For most purposes, this will either be 0 (i.e.
//   there are no free resources) or the same as the <p MaximumCount>
//   parameter (all resources free).  The <p InitialCount> must be less than
//   or equal to the MaximumCount.
//
//   The <p MaximumCount> parameter determines the maximum count of the
//   semaphore.  It must be at least 1 and greater than or equal to
//   <p InitialCount>.  
//
// @rdesc Returns one of the following:
//
// @flag PilSemaphore | The handle of the newly created semaphore object.
// @flag NULL         | The semaphore could not be created.
//
//

PilSemaphore
PIL_CreateSemaphore(
    uint32_t InitialCount,  // @parm The count the semaphore should have
                            //   when it is created.
    uint32_t MaximumCount   // @parm The maximum count the semaphore can
                            //   have.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_DestroySemaphore
//
// @func <f PIL_DestroySemaphore> destroys a semaphore created with 
//   <f PIL_CreateSemaphore>.  All semaphores must be destroyed using
//   this function for the memory to be reclaimed.  Semaphores should
//   not be destroyed while they are still being used.  
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK             | The semaphore has been successfully destroyed.
// @flag PIL_INVALID_HANDLE | The handle of the semaphore is not a valid
//                            semaphore handle.
//
//

PilResult
PIL_DestroySemaphore(
    PilSemaphore Sem  // @parm The handle of the semaphore to destroy.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_GetSemaphoreCount
//
// @func <f PIL_GetSemaphoreCount> returns the current count of the 
//   semaphore between the initial and maximum values that it was created
//   with.
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK             | The operation completed successfully.
// @flag PIL_INVALID_HANDLE | The semaphore handle is not a valid handle
//                            for a semaphore.
// @flag PIL_INVALID_PARAM  | The pointer to the value is not a valid pointer.
//
//

PilResult
PIL_GetSemaphoreCount(
    PilSemaphore Sem,    // @parm The handle of the semaphore for which to
                         //   receive the count.
    uint32_t*    pValue  // @parm Pointer to a variable to place the semaphore
                         //   value in.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_ReleaseSemaphore
//
// @func <f PIL_ReleaseSemaphore> releases a semaphore, essentially 
//   incrementing the count of the semaphore.  A semaphore does not
//   have to be first acquired before it can be released.  It is not an
//   error to release a semaphore that is already at the maximum count; 
//   it will essentially be ignored.  
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK             | The semaphore was released successfully
//                            (incremented).
// @flag PIL_INVALID_HANDLE | The handle of the semaphore is not a valid
//                            semaphore handle.
//
//

PilResult
PIL_ReleaseSemaphore(
    PilSemaphore Sem  // @parm The handle of the semaphore to release.
    );

#endif // __PILSEM_H_INCLUDED__
