/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILTHD.H                                                  //
// Module:   Platform Isolation Layer Thread Management Routines       //
// Author:   Daniel Baumberger                                         //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
// PURPOSE:                                                            //
//                                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//
// HISTORY:
//
/////////////////////////////////////////////////////////////////////////

#ifndef __PILTHD_H_INCLUDED__
#define __PILTHD_H_INCLUDED__

//
// @doc
//
// @topic Threading (Required) |
//
// A thread, also known as a task, is the basic unit of execution.  It
// contains a copy of the CPU execution state for itself but it shares
// an address with other threads created under a process (or with all
// other threads if it runs in the kernel).  
//
// All threads are created using the <f PIL_CreateThread> API.  Certain
// attributes of the thread can be specified when it is created such as
// its priority, stack size, a name, and which function represents the
// thread's starting address.  Once created, the thread starts executing
// independently unless the <ms PIL_THREAD_SUSPENDED> flag is used when
// the thread is created, which delays the thread from executing until a
// <f PIL_ResumeThread> call is made on the thread.
//
// Certain thread functions operate on a thread by another thread and some
// are for use by the thread itself.  Examples of the former are
// <f PIL_DestroyThread>, <f PIL_GetThreadPriority>, <f PIL_ResumeThread>, 
// <f PIL_SuspendThread>, and <f PIL_TerminateThread>.  All of these
// functions require the handle of the thread.  A thread may obtain its
// own handle by calling <f PIL_GetCurrentThread>.  Functions used by the
// thread include <f PIL_DelayThread> and <f PIL_ExitThread>.
//
// All threads must call <f PIL_ExitThread> before their thread function
// returns.  Failing to do so is platform dependent and may cause
// unpredictable results.   After a thread has successfully exited, the
// creator of the thread (or somebody else) needs to call <f PIL_DestroyThread>
// to free up the data structures associated with the thread.  
// <f PIL_DestroyThread> returns the exit code for the thread.
//
// Another thread can forcibly cause another thread to terminate by calling
// <f PIL_TerminateThread>.  Note that on all platforms, the thread data
// structures may not be cleaned up complete when calling this function.  
// <f PIL_DestroyThread> is still required to release the memory associated
// with the thread.
//
// PIL allows a single, 32-bit value to be associated with each thread
// handle.  The value may be set using the <f PIL_SetThreadValue> function
// and retrieved via the <f PIL_GetThreadValue> function.
// 
// In some cases, the priority level mapping that PIL does may not be
// sufficient.  PIL provides a mechanism to retrieve and set the 
// native priority level of a thread without any interpretation by PIL.
// These native priority values are operating system specific.  
// <f PIL_GetNativeThreadPriority> retrieves the native priority from the
// underlying operating system.  Inversly, <f PIL_SetNativeThreadPriority>
// allows the thread priority to be set without any priority mapping by
// PIL.
//
// <f PIL_WaitForThread> allows thread to wait until another thread has
// exited.  The waiting thread has the option to just check if another
// thread has exited, wait for a short while for it to exit, or wait
// forever until the thread exits.
//

#define PIL_THREAD_SUSPENDED 0x0001

#define PIL_MIN_PRIORITY 1
#define PIL_MAX_PRIORITY 8

typedef void* PilThread;

typedef uint32_t (*PilThreadRoutine)( void* Param );

///////////////////////////////////////////////////////////////////////
//
// PIL_CreateThread
//
// @func <f PIL_CreateThread> creates a thread an optionally starts it
//   executing.  The <p StartingAddress> parameter is the address (i.e.
//   a function) where the thread will begin executing.  The specified
//   function must have the following prototype:
//
//   <ms uint32_t PilThreadRoutine ( void* Param );>
//
//   Note: thread functions are assumed to be using the standard C
//   calling convention (cdecl).
//
//   The <p Param> specified in the <f PIL_CreateThread> call will be
//   passed to the thread function.  The contents of the parameter are
//   not specified and are completely up to the caller.
//
//   <p dwStackSize> is the size of the stack the thread should have, in
//   bytes.  Note that not all platforms allow threads to have a specified
//   stack size.  However, for portability, a stack size should be specified
//   anyway.  To conserve memory, the stack should not be very large.
//
//   The <p Name> is just a string describing the thread.  It is entirely
//   optional but can help in resource tracking if one is specified.
//
//   The priority of the thread can range from 1 (lowest) to 8 (highest).  
//
//   <f PIL_CreateThread> currently has only one option for <p dwOptions>: 
//   <ms PIL_THREAD_SUSPENDED>.  When this option is specified, the thread
//   will not begin executing immediately upon creation but will have to
//   started using the <f PIL_ResumeThread> API.
//
// @rdesc Returns one of the following:
//
// @flag PilThread | The handle of the of the newly created thread.
// @flag NULL      | The thread could not be created.
//
//

PilThread
_PIL_CreateThread(
    PilThreadRoutine StartAddress, // @parm The starting address the thread
                                   //   will begin execution.
    uint32_t         StackSize,    // @parm The size of the thread's stack
                                   //   in bytes.
    char*            Name,         // @parm The optional name of the thread.
    uint32_t         Priority,     // @parm The priority the thread will
                                   //   execute at.
    void*            Param,        // @parm An optional parameter to pass to
                                   //   the thread function.
    uint32_t         Options       // @parm Thread creation options.
    );

#define PIL_CreateThread(a,b,c,d,e,f) \
	_PIL_CreateThread(a,b, \
		c?c:__FUNCTION__, \
		d,e,f)

///////////////////////////////////////////////////////////////////////
//
// PIL_DelayThread
//
// @func <f PIL_DelayThread> delays the execution of the currently executing
//   thread for the specified number of milliseconds.  The thread is not
//   alertable during this wait so it cannot be signaled.  Note that on
//   all platforms, the resolution of the wait may not be 1 millisecond
//   but could be 10 or 20 milliseconds.  In those cases, the delay will
//   be rounded up to the nearest supported interval.
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK | The delay was completed successfully.
//
//

PilResult
PIL_DelayThread(
    uint32_t     Delay  // @parm The number of milliseconds to delay the
                        //   currently executing thead.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_DestroyThread
//
// @func <f PIL_DestroyThread> deallocates all memory associated with
//   a thread.  Note that this assumes the thread is already terminated.
//   After this call, the thread handle will be invalid and cannot be
//   used by anyone, including the thread itself.  All threads must be
//   destroyed using the function after they terminate to free associated
//   memory.
//
//   <p pStatus> is an optional parameter that allows the caller to retrieve
//   the exit status of the thread when it terminated.  The exit status is
//   the value passed to <f PIL_ExitThread> when the thread voluntarily exited.
//   This parameter may be <ms NULL> if the status is not required.
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK             | The thread was destroyed successfully.
// @flag PIL_INVALID_HANDLE | The handle of the thread is not a valid thread
//                            handle.
//
//

PilResult
PIL_DestroyThread(
    PilThread Thread,	// @parm The handle of the thread to destroy.
    uint32_t* pStatus   // @parm An optional pointer to a variable to
                        //   retrieve the exit status of the thread.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_ExitThread
//
// @func <f PIL_ExitThread> stops execution of the current thread normally.
//   Note that it does not destroy the thread, but simply stops execution.
//   <f PIL_DestroyThread> must be called to clean up after the thread.
//   <bold This function never returns>.  
//
//   The thread may specify a reason code as to why it exits.  If there is
//   no reason, it should be set to 0.
//
// @rdesc This function does not return.
//
//

void
PIL_ExitThread(
    uint8_t Status  // @parm The reason code the thread has for exiting.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_GetCurrentThread
//
// @func <f PIL_GetCurrentThread> returns the handle of the currently
//   executing thread for use in other thread functions.  This function
//   cannot fail and will always return a valid handle.  <it Note that this
//   may not be lightweight and may take longer as the number of threads
//   increase>.
//
// @rdesc The handle of the currently executing thread
//
//

PilThread
PIL_GetCurrentThread(
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_GetCurrentThreadId
//
// Returns the threadid of the currently executing thread.
//

unsigned long PIL_GetCurrentThreadId();

///////////////////////////////////////////////////////////////////////
//
// PIL_GetNativeThreadPriority
//
// @func <f PIL_GetNativeThreadPriority> retrieves the current
//   priority of the given thread as maintained by the underlying
//   operating system.  The values returned by this function are
//   operating system specific and are useful only when priorities
//   outside what PIL offers are required.
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK             | The priority was retrieved successfully.
// @flag PIL_INVALID_HANDLE | The handle of the thread is not a valid thread
//                            handle.
// @flag PIL_INVALID_PARAM  | The pointer to the priority value is not a
//                            valid pointer.
//
//

PilResult
PIL_GetNativeThreadPriority(
    PilThread Thread,	// @parm The handle of the thread to obtain the
                        //   priority of.
    uint32_t* pPri		// @parm A pointer to a variable to place the
                        //   priority value in.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_GetThreadPriority
//
// @func <f PIL_GetThreadPriority> retrieves the current priority of the
//   given thread. 
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK             | The priority was retrieved successfully.
// @flag PIL_INVALID_HANDLE | The handle of the thread is not a valid thread
//                            handle.
// @flag PIL_INVALID_PARAM  | The pointer to the priority value is not a
//                            valid pointer.
//
//

PilResult
PIL_GetThreadPriority(
    PilThread Thread,	// @parm The handle of the thread to obtain the
                        //   priority of.
    uint32_t* pPri		// @parm A pointer to a variable to place the
                        //   priority value in.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_GetThreadValue
//
// @func <f PIL_GetThreadValue> retrieves the 32-bit value associated
//   with the given thread.  If no value has been previously set, this
//   function will return 0.
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK             | The value was retrieved successfully.
// @flag PIL_INVALID_HANDLE | The handle of the thread is not a valid thread
//                            handle.
// @flag PIL_INVALID_PARAM  | The pointer to the value is not a
//                            valid pointer.
//
// 

PilResult
PIL_GetThreadValue(
    PilThread Thread,	// @parm The handle of the thread to obtain the
                        //   value of.
    uint32_t* pValue    // @parm Pointer to a location to store the 
                        //   the value.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_ResumeThread
//
// @func <f PIL_ResumeThread> resumes the execution of a thread suspended
//   either a creation time via the <ms PIL_THREAD_SUSPENDED> flag or via
//   <f PIL_SuspendThread>.  Execution continues at the same priority level
//   the thread was at when it was suspended.
//
//   Resuming a thread that is not suspended has no effect.
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK             | The thread was resumed successfully.
// @flag PIL_INVALID_HANDLE | The handle of the thread is not a valid thread
//                            handle.
// @flag PIL_ERROR          | The thread could not be resumed.  Most likely
//                            this is a result of the thread being in a state
//                            that is cannot be resumed due to some operating
//                            system dependent reason.
//
//

PilResult
PIL_ResumeThread(
    PilThread Thread    // @parm The handle of the thread to resume
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_SetNativeThreadPriority
//
// @func <f PIL_SetNativeThreadPriority> sets the priority of a thread
//   by using values native to the underlying operating system.  Note
//   that these values used by this function are operating system
//   specific and are useful only for setting priorities outside of
//   what PIL offers.
//
// @rdesc Returns on of the following:
//
// @flag PIL_OK             | The thread's priority was adjusted successfully.
// @flag PIL_INVALID_HANDLE | The handle of the thread is not a valid thread
//                            handle.
// @flag PIL_INVALID_PARAM  | The new priority value is not a valid priority.
//
//

PilResult
PIL_SetNativeThreadPriority(
    PilThread Thread,   // @parm The handle of the thread to adjust the
                        //   priority of.
    uint32_t  Priority  // @parm The new priority of the thread.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_SetThreadPriority
//
// @func <f PIL_SetThreadPriority> sets the priority of a thread.  Threads
//   may adjust their own priority be specifying <f PIL_GetCurrentThread>
//   as the first parameter.  Although this call will not block, if the
//   priority of the thread is adjusted above the priority of the currently
//   executing thread, a task switch may occur.
//
// @rdesc Returns on of the following:
//
// @flag PIL_OK             | The thread's priority was adjusted successfully.
// @flag PIL_INVALID_HANDLE | The handle of the thread is not a valid thread
//                            handle.
// @flag PIL_INVALID_PARAM  | The new priority value is not a valid priority.
//
//

PilResult
PIL_SetThreadPriority(
    PilThread Thread,   // @parm The handle of the thread to adjust the
                        //   priority of.
    uint32_t  Priority  // @parm The new priority of the thread.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_SetThreadValue
//
// @func <f PIL_SetThreadValue> allows a single 32-bit value to be
//   associated with a particular thread handle.  Only one 32-bit
//   value is provided for each thread handle and thus is shared
//   by all.  No interpretation of this value is done by PIL.
//
// @rdesc Returns on of the following:
//
// @flag PIL_OK             | The thread's priority was adjusted successfully.
// @flag PIL_INVALID_HANDLE | The handle of the thread is not a valid thread
//                            handle.
//
//

PilResult
PIL_SetThreadValue(
    PilThread Thread, // @parm The handle of the thread to set the value
                      //   of.
    uint32_t  Value   // @parm The 32-bit value to store with this thread.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_SuspendThread
//
// @func <f PIL_SuspendThread> causes the given thread to stop executing.
//   This is different than adjusting the thread priority to minimum.
//   Once this call returns, the thread has stopped executing and will
//   not continue until a <f PIL_ResumeThread> call is made.
//
//   Suspending a thread that is currently blocked with a timeout is
//   undefined.  
//
//   Suspending a thread that is already suspended has no effect.
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK             | The thread was suspended successfully.
// @flag PIL_INVALID_HANDLE | The handle of the thread is not a valid thread
//                            handle.
// @flag PIL_ERROR          | The thread could not be suspended.  Most likely
//                            this is a result of the thread being in a state
//                            that is cannot be suspended due to some
//                            operating system dependent reason.
//
//

PilResult
PIL_SuspendThread(
    PilThread Thread    // @parm The handle of the thread to suspend
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_TerminateThread
//
// @func <f PIL_TerminateThread> forcibly terminates a thread, not giving
//   it a chance to do any cleanup.  This should be used only in abnormal
//   circumstances.   Note that the thread must still be destroyed using
//   <f PIL_DestroyThread>.
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK             | The thread was terminated successfully.
// @flag PIL_INVALID_HANDLE | The handle of the thread is not a valid thread
//                            handle.
//
//

PilResult
PIL_TerminateThread(
    PilThread Thread  // @parm The handle of the thread to terminate.
    );

////////////////////////////////////////////////////////////////////////
//
// PIL_WaitForThread
//
// @func <f PIL_WaitForThread> will block until the given thread exits.
//   The timeout value specifies how long to wait for the thread to
//   exit.  A value of <ms PIL_INFINITE_TIMEOUT> will wait forever.
//   A timeout value of <ms 0> will check if the thread has exited.
//
//   Note that on all platforms, the granularity of the wait may not be one
//   millisecond but may be more like 10 or 20 milliseconds.  In that case,
//   the wait will be rounded up to the nearest integral unit.
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK             | The wait was successful and the thread has
//                            exited.
// @flag PIL_INVALID_HANDLE | The handle of the event is not a valid event
//                            handle.
// @flag PIL_TIMEOUT        | The timeout period expired before the event
//                            became signaled.
//
// 

PilResult
PIL_WaitForThread(
    PilThread Thread,   // @parm The handle of the thread to wait for.
    uint32_t dwTimeout  // @parm The time, in milliseconds, to wait for the
                        //   thread or <ms PIL_INFINITE_TIMEOUT> to wait 
                        //   forever.
    );

#endif // __PILTHD_H_INCLUDED__
