/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILTMR.H                                                  //
// Module:   Platform Isolation Layer Timer Routines                   //
// Author:   Daniel Baumberger                                         //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
// PURPOSE:                                                            //
//                                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//
// HISTORY:
//
// 
/////////////////////////////////////////////////////////////////////////

#ifndef __PILTMR_H_INCLUDED__
#define __PILTMR_H_INCLUDED__

//
// @doc
//
// @topic Timers (Required) |
//
// Timers provide a mechanism to generate a callback after a certain amount
// of time has passed.  In addition, they can be set to provide a callback
// on a periodic basis.
//
// Timers are created using <f PIL_CreateTimer>, which specifies the type
// (single or repeating) and also specifies the initial interval of the timer.
// A timer is scheduled to fire using <f PIL_StartTimer>.  A timer can be
// removed from the list of active timers using <f PIL_StopTimer>.
// <f PIL_ResetTimer> allows the timer to be reconfigured with a different
// callback function or interval.  Finally, all timers need to be freed
// using <f PIL_DestroyTimer>.
//
// Due to platform differences in the execution of the timer callbacks, it
// cannot be assumed that the timer runs in any specific context.  Therefore,
// many operations are disallowed during the execution of the timer routine.
// They include:
//
// <tab>- Any PIL function the Creates or Destroys an object<nl>
// <tab>- Memory allocation functions<nl>
// <tab>- Blocking synchronization functions<nl>
// <tab>- Read/Write mailbox functions with a non-zero timeout<nl>
//

typedef void* PilTimer;

typedef void (*PilTimerRoutine)( PilTimer Timer, void* Context );

typedef enum ePIL_TIMER_TYPE {
    PIL_TIMER_SINGLE,
    PIL_TIMER_REPEATING
} PilTimerType;  

///////////////////////////////////////////////////////////////////////
//
// PIL_CreateTimer
//
// @func <f PIL_CreateTimer> creates a new timer object.  It does not
//   automatically schedule the timer to fire.  <f PIL_StartTimer> will
//   start a newly created timer.
//
//   The timer routine must have the following prototype:
//
//   <ms void TimerRoutine( PilTimer Timer, void* Context );>
//
//   The handle of the timer that just expired is passed to the timer
//   routine along with the user defined context pass from 
//   <f PIL_CreateTimer> or <f PIL_ResetTimer>.
//
//   The interval specifies the amount of time before the timer expires.
//   For repeating timers, it will be the period in which it fires.  Note
//   that not all platforms will have a resolution of one millisecond, but
//   more like 10 or 20 milliseconds.  On those platforms, the interval
//   will be rounded up to the next integral unit.
//
// @rdesc Returns of the following:
//
// @flag PilTimer | The handle of the newly created timer.
// @flag NULL     | The timer could not be created.
//
//

PilTimer
PIL_CreateTimer(
    PilTimerRoutine  TimerRoutine,  // @parm The function to call when the
                                    //   timer expires.
    uint32_t         Interval,      // @parm The amount of time to wait until
                                    //   the timer fires in milliseconds.
    void*            Context,       // @parm The context info to pass to the
                                    //   timer routine.
    PilTimerType     Type           // @parm <ms PIL_TIMER_SINGLE> or
                                    //   <ms PIL_TIMER_REPEATING>.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_DestroyTimer
//
// @func <f PIL_DestroyTimer> destroys a timer object created via 
//   <f PIL_CreateTimer>.  The timer must be first stopped by waiting
//   for it to expire or calling <f PIL_StopTimer>.  The behavior of
//   destroying a timer that is still scheduled is undefined.
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK             | The timer was destroyed successfully.
// @flag PIL_INVALID_HANDLE | The handle of the timer is not a valid timer
//                            handle.                            
//
//

PilResult
PIL_DestroyTimer(
    PilTimer  Timer     // @parm The handle of the timer to destroy.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_DisableTimers
//
// @func <f PIL_DisableTimers> globally disables timer callbacks until 
//   <f PIL_EnableTimers> is called.  After this call, no timer callback
//   will be called.   Care should be taken to avoid long periods of time
//   when timers are disabled so that the timeout values are not skewed.
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK    | The timers have been disabled successfully.
// @flag PIL_ERROR | The timers are already disabled.
//
//

PilResult
PIL_DisableTimers(
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_EnableTimers
//
// @func <f PIL_EnableTimers> re-enables timer callbacks that have been
//   disabled via <f PIL_DisableTimers>.  Note that timers that would
//   have expired during the period when timers were disabled will most
//   likely fire immediately.
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK    | The timers have been enabled successfully.
// @flag PIL_ERROR | The timers are already enabled.
//
//

PilResult
PIL_EnableTimers(
    );


///////////////////////////////////////////////////////////////////////
//
// PIL_ResetTimer
//
// @func <f PIL_ResetTimer> allows an already created timer to be
//   re-configured.  The prototype is very similar to <f PIL_CreateTimer>.
//   Refer to <f PIL_CreateTimer> for descriptions of the parameters.
//
// @rdesc Returns one of the following:
// 
// @flag PIL_OK             | The timer was successfully reset.
// @flag PIL_INVALID_HANDLE | The handle of the timer is not a valid timer
//                            handle.
// @flag PIL_INVALID_PARAM  | The timer routine is not valid.
//
//

PilResult
PIL_ResetTimer(
    PilTimer        Timer,         // @parm The handle of the timer to reset.
    PilTimerRoutine TimerRoutine,  // @parm The new routine to call when the
                                   //   timer fires.
    uint32_t        Interval,      // @parm The new interval to wait before
                                   //   calling the timer routine.
    void*           Context,       // @parm The new context to pass to the
                                   //   timer routine.
    PilTimerType    Type           // @parm <ms PIL_TIMER_SINGLE> or
                                   //   <ms PIL_TIMER_REPEATING>.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_StartTimer
//
// @func <f PIL_StartTimer> schedules a timer according to the parameters
//   specified when it was created or when reset with <f PIL_ResetTimer>.
//   A repeating timer cannot be scheduled more than once unless first
//   stopped.  A single shot timer cannot be scheduled more than once unless
//   stopped or it has fired.
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK             | The timer was scheduled successfully.
// @flag PIL_INVALID_HANDLE | The handle of the timer is not a valid timer
//                            handle.
// @flag PIL_ERROR          | The timer is already scheduled.

PilResult
PIL_StartTimer(
    PilTimer Timer  // @parm The handle of the timer to schedule.
    );

///////////////////////////////////////////////////////////////////////
//
// PIL_StopTimer
//
// @func <f PIL_StopTimer> stops a timer that has been scheduled.  
//   Repeating timers must be stopped with this function before they can
//   be destroyed.
//
// @rdesc Returns one of the following:
//
// @flag PIL_OK             | The timer was stopped successfully.
// @flag PIL_INVALID_HANDLE | The handle of the timer is not a valid timer
//                            handle.
// @flag PIL_ERROR          | The timer is not scheduled.

PilResult
PIL_StopTimer(
    PilTimer Timer  // @parm The handle of the timer to stop.
    );

#endif // __PILTMR_INCLUDED__
