/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILCS.C                                                   //
// Module:   Platform Isolation Layer Critical Sections                //
// Author:   JP Sugarbroad                                             //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////

#include "pil_int.h"
#include <string.h>
#include <stdio.h>

// We do not use PilSemaphore because here the user allocates the memory
// for us.

#ifdef DEBUG
    // Magic number for debug checking
#   define PIL_CS_MAGIC (0xC0FFEE)

    // Macro to check for invalid critical sections
#   define pil_checkcs(x,r) do { \
        if (!(x) || (x)->magic != PIL_CS_MAGIC) { \
            ASSERTMSG(FALSE,"Attempt to use bad critical section."); \
            return r; \
        } \
    } while (0);
#else // !DEBUG
#   define pil_checkcs(x,r)
#endif

#ifdef DEBUG
static pillist cslist = NULL;
pillist _PIL_GetCriticalSectionList() { return cslist; }
#endif

///////////////////////////////////////////////////////////////////////
//
// PIL_InitializeCriticalSection
//
// Initializes a user-provided block of memory as a critical section
// object.
//

void _PIL_InitializeCriticalSection(PilCriticalSection *pCriticalSection, const char *name)
{
pthread_mutexattr_t attr;
    pilcs_t *arg = (pilcs_t *) pCriticalSection;

    ASSERT(sizeof(pilcs_t) <= sizeof(PilCriticalSection));
#ifdef DEBUG
    arg->magic = PIL_CS_MAGIC;
    {
        pthread_mutexattr_t attr;
        
        pthread_mutexattr_init(&attr);
        pthread_mutexattr_settype(&attr,PTHREAD_MUTEX_RECURSIVE_NP);

        pthread_mutex_init(&arg->mutex,&attr);

        pthread_mutexattr_destroy(&attr);
    }
#else
        pthread_mutexattr_init(&attr);
        pthread_mutexattr_settype(&attr,PTHREAD_MUTEX_RECURSIVE_NP);

        pthread_mutex_init(&arg->mutex,&attr);
        pthread_mutexattr_destroy(&attr);
#endif

#ifdef DEBUG
    arg->info = (csinfo_t*)PIL_AllocateMemory(sizeof(csinfo_t),0);
    arg->info->owner = NULL;
    arg->info->blocking = NULL;

    if (!name)
        name = "unnamed";

    arg->info->name = (char*)PIL_AllocateMemory(strlen(name)+1,0);
    strcpy(arg->info->name,name);

    _PIL_AddToList(&cslist,arg->info);
#endif
}

///////////////////////////////////////////////////////////////////////
//
// PIL_DestroyCriticalSection
//
// Destroys a critical section object.
//

void PIL_DestroyCriticalSection(PilCriticalSection *pCriticalSection)
{
#ifdef DEBUG
    csinfo_t *info;
#endif
    pilcs_t *arg = (pilcs_t *) pCriticalSection;

    arg = arg;
    pil_checkcs(arg,);
#ifdef DEBUG
    arg->magic = 0; // Prevent use of destroyed critical section
#endif

#ifdef DEBUG
    info = arg->info;
    _PIL_RemoveFromList(&cslist,info);

    PIL_FreeMemory(info);
#endif
    
    pthread_mutex_destroy(&arg->mutex);
}

///////////////////////////////////////////////////////////////////////
//
// PIL_EnterCriticalSection
//
// Enters a critical section, blocking until no other threads are in
// the critical section.
//

void PIL_EnterCriticalSection(PilCriticalSection *pCriticalSection)
{
    pilcs_t *arg = (pilcs_t *) pCriticalSection;
#ifdef DEBUG
    int r;
#endif

    pil_checkcs(arg,);

#ifdef DEBUG
    t = (thread_t*)PIL_GetCurrentThread();
    _PIL_AddToList(&(arg->info->blocking),t);
#endif

#ifdef DEBUG
    if ((r = pthread_mutex_lock(&arg->mutex))) {
        PIL_Trace("Error in PIL_EnterCriticalSection: %s\n",strerror(r));
    }
#else
    pthread_mutex_lock(&arg->mutex);
#endif
    
#ifdef DEBUG
    if (t)
    {
        _PIL_RemoveFromList(&(arg->info->blocking),t);
        arg->info->owner = t;
    }
#endif
}

///////////////////////////////////////////////////////////////////////
//
// PIL_TryEnterCriticalSection
//
// Attempts to enter a critical section, returning a FALSE error code
// if it cannot immediately enter.
//
// Note: TRUE != !FALSE
//

BOOL PIL_TryEnterCriticalSection(PilCriticalSection *pCriticalSection)
{
    pilcs_t *arg = (pilcs_t *) pCriticalSection;
#ifdef DEBUG
    int r;
#endif

    pil_checkcs(arg,TRUE);
#ifdef DEBUG
    if ((r = pthread_mutex_trylock(&arg->mutex))) {
        if (r == EBUSY) return FALSE;
        PIL_Trace("Error in pthread_mutex_trylock: %s\n",strerror(r));
        return FALSE;
    }
    return TRUE;
#else
    return pthread_mutex_trylock(&arg->mutex) ? FALSE : TRUE;
#endif
}

///////////////////////////////////////////////////////////////////////
//
// PIL_LeaveCriticalSection
//
// Leaves a critical section, enabling another thread to enter.
//

void PIL_LeaveCriticalSection(PilCriticalSection *pCriticalSection)
{
    pilcs_t *arg = (pilcs_t *) pCriticalSection;
#ifdef DEBUG
    int r;
#endif

#ifdef DEBUG
    arg->info->owner = NULL;
#endif

    pil_checkcs(arg,);
#ifdef DEBUG
    if ((r = pthread_mutex_unlock(&arg->mutex))) {
        PIL_Trace("Error in pthread_mutex_unlock: %s\n",strerror(r));
    }
#else
    pthread_mutex_unlock(&arg->mutex);
#endif
}
