/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILEVNT.C                                                 //
// Module:   Platform Isolation Layer Events                           //
// Author:   JP Sugarbroad                                             //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////

#include "pil_int.h"

#ifdef DEBUG
    // Magic number for debug checking
#   define PIL_EVENT_MAGIC (0x3D4ACB50)

    // Macro to check for invalid events
#   define pil_checkevnt(ev,r) do { \
        if (!(ev) || (ev)->magic != PIL_EVENT_MAGIC) { \
            ASSERTMSG(FALSE,"Attempt to use bad event"); \
            return r; \
        } \
    } while (0);
#else // !DEBUG
#   define pil_checkevnt(ev,r)
#endif

// Automatic-reset events are the same as binary semaphores...

// The structure behind PilEvent
typedef struct event_s {
#ifdef DEBUG
    uint32_t magic; // Magic number
#endif
    BOOL manual; // TRUE if this is a manual-reset event
    union {
        struct {
            BOOL signaled;
            pthread_mutex_t mx;
            pthread_cond_t cv;
        } m; // manual
        PilSemaphore a; // automatic
    } u;
} event_t;

///////////////////////////////////////////////////////////////////////
//
// PIL_CreateEvent
//
// Creates a new manual or automatic reset event object.
//

PilEvent PIL_CreateEvent(BOOL bManualReset, BOOL bSignaled)
{
    event_t *ev;

    // Allocate and initialize the event

    ev = (event_t *) PIL_AllocateMemory(sizeof(event_t),0);
    if (ev == NULL) return (PilEvent) NULL;

#ifdef DEBUG
    ev->magic = PIL_EVENT_MAGIC;
#endif
    ev->manual = bManualReset;
    if (bManualReset) {
        ev->u.m.signaled = bSignaled ? TRUE : FALSE;
        pthread_mutex_init(&ev->u.m.mx,NULL);
        pthread_cond_init(&ev->u.m.cv,NULL);
    } else {
        // Treat it like a binary semaphore
        ev->u.a = PIL_CreateBinarySemaphore(bSignaled);
        if (ev->u.a == NULL) {
            PIL_FreeMemory(ev);
            return (PilEvent) NULL;
        }
    }
    return (PilEvent) ev;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_DestroyEvent
//
// Destroys an event object created using PIL_CreateEvent.
//

PilResult PIL_DestroyEvent(PilEvent Event)
{
    event_t *ev = (event_t *) Event;

    pil_checkevnt(ev,PIL_INVALID_HANDLE); // Check for invalid event
#ifdef DEBUG
    ev->magic = 0; // Prevent use of destroyed event
#endif
    if (ev->manual) {
        pthread_cond_destroy(&ev->u.m.cv);
        pthread_mutex_destroy(&ev->u.m.mx);
    } else {
        PIL_DestroySemaphore(ev->u.a);
    }
    PIL_FreeMemory(ev);
    return PIL_OK;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_SetEvent
//
// Sets an event into the signaled state.  For auto-reset events,
// the state of the event will automatically be reset to non-signaled
// after a thread waiting on the event is wakened.
//

PilResult PIL_SetEvent(PilEvent Event)
{
    event_t *ev = (event_t *) Event;

    pil_checkevnt(ev,PIL_INVALID_HANDLE); // Check for invalid event

    if (ev->manual) {
        ev->u.m.signaled = TRUE;
        pthread_cond_broadcast(&ev->u.m.cv); // wake up waiters
        return PIL_OK;
    } else {
        // There is code in pilsem.c so that over-releasing a binary
        // semaphore does not produce a warning message.
        return PIL_ReleaseSemaphore(ev->u.a);
    }
}

///////////////////////////////////////////////////////////////////////
//
// PIL_ResetEvent
//
// Resets a manual reset event to the non-signaled state.  It does not
// have any effect on an auto-reset event.
//

PilResult PIL_ResetEvent(PilEvent Event)
{
    event_t *ev = (event_t *) Event;

    pil_checkevnt(ev,PIL_INVALID_HANDLE); // Check for invalid event

    if (ev->manual) {
        ev->u.m.signaled = FALSE;
    }
    // As per spec, does nothing to automatic-reset events
    return PIL_OK;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_WaitForEvent
//
// Waits for an event for the given number of milliseconds.  It will
// return when the event is set to the signaled state or a timeout
// occurs.
//

PilResult PIL_WaitForEvent(PilEvent Event, uint32_t dwTimeout)
{
    event_t *ev = (event_t *) Event;

    pil_checkevnt(ev,PIL_INVALID_HANDLE); // Check for invalid event

    if (ev->manual) {
        if (dwTimeout == PIL_INFINITE_TIMEOUT) {
            pthread_mutex_lock(&ev->u.m.mx);
            while (!ev->u.m.signaled) {
                pthread_cond_wait(&ev->u.m.cv,&ev->u.m.mx);
            }
            pthread_mutex_unlock(&ev->u.m.mx);
            return PIL_OK;
        } else {
            struct timeval tv;
            struct timespec ts;
            uint32_t t = dwTimeout / 1000;
            int r = 0;

            gettimeofday(&tv,NULL);
            ts.tv_sec = tv.tv_sec + t;
            ts.tv_nsec = (tv.tv_usec+(dwTimeout-t*1000)*1000)*1000;
            pthread_mutex_lock(&ev->u.m.mx);
            while (!ev->u.m.signaled && r != ETIMEDOUT) {
                r = pthread_cond_timedwait(&ev->u.m.cv,&ev->u.m.mx,&ts);
            }
            pthread_mutex_unlock(&ev->u.m.mx);
            switch (r) {
              case 0: return PIL_OK;
              case ETIMEDOUT: return PIL_TIMEOUT;
              default: return PIL_ERROR;
            }
        }
    } else {
        // Treat it like a binary semaphore
        return PIL_AcquireSemaphore(ev->u.a, dwTimeout);
    }
}
