/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILSEM.C                                                  //
// Module:   Platform Isolation Layer Semaphores                       //
// Author:   JP Sugarbroad                                             //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////

#include "pil_int.h"
#include <string.h>
#include <stdio.h>

#ifdef DEBUG
    // Magic number for debug checking
#   define PIL_SEM_MAGIC (0x1A2B3C4D)

#   define pil_checksem(sem,r) do { \
        if (!(sem) || (sem)->magic != PIL_SEM_MAGIC) { \
            ASSERTMSG(FALSE,"Attempt to use bad semaphore."); \
            return r; \
        } \
    } while (0);
#else // !DEBUG
#   define pil_checksem(sem,r)
#endif

// The structure behind PilSemaphore
typedef struct sem_s
{
#ifdef DEBUG
    uint32_t magic;
#endif
    uint32_t cur,max;
    pthread_mutex_t mx;
    pthread_cond_t cv;
} sem_t;

///////////////////////////////////////////////////////////////////////
//
// PIL_CreateSemaphore
//
// Creates a new semaphore object, returning a handle to the new object.
//

PilSemaphore PIL_CreateSemaphore(uint32_t dwInitialCount, uint32_t dwMaximumCount)
{
    sem_t *result;
    
    // Sanity check
    ASSERT(dwInitialCount >= 0 && dwInitialCount <= dwMaximumCount);
    if (dwInitialCount < 0 || dwInitialCount > dwMaximumCount) return (PilSemaphore) NULL;

    // Allocate and initialize the semaphore

    result = (sem_t *) PIL_AllocateMemory(sizeof(sem_t),0);
    if (result == NULL) return (PilSemaphore) NULL;

#ifdef DEBUG
    result->magic = PIL_SEM_MAGIC;
#endif
    result->cur = dwInitialCount;
    result->max = dwMaximumCount;

    pthread_mutex_init(&result->mx,NULL);
    pthread_cond_init(&result->cv,NULL);

    return (PilSemaphore) result;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_DestroySemaphore
//
// Destroys a semaphore created with PIL_CreateSemaphore or 
// PIL_CreateBinarySemaphore.  
//

PilResult PIL_DestroySemaphore(PilSemaphore Sem)
{
    sem_t *arg = (sem_t *) Sem;

    pil_checksem(arg,PIL_INVALID_HANDLE); // Check for invalid semaphore
#ifdef DEBUG
    arg->magic = 0; // Prevent use of destroyed semaphore
#endif

    pthread_cond_destroy(&arg->cv);
    pthread_mutex_destroy(&arg->mx);

    PIL_FreeMemory(arg);
    return PIL_OK;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_GetSemaphoreCount
//
// Returns the current count of the semaphore between 0 and the maximum
// count the semaphore was created with.
//

PilResult PIL_GetSemaphoreCount(PilSemaphore Sem, uint32_t *pValue)
{
    sem_t *arg = (sem_t *) Sem;

    pil_checksem(arg,PIL_INVALID_HANDLE); // Check for invalid semaphore
#ifdef DEBUG
    if (!pValue) {
        PIL_Trace("Bad pValue passed to PIL_GetSemaphoreCount.\n");
        return PIL_INVALID_PARAM;
    }
#endif

    pthread_mutex_lock(&arg->mx);
    *pValue = arg->cur;
    pthread_mutex_unlock(&arg->mx);

    return PIL_OK;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_AcquireSemaphore
//
// Acquires the given semaphore, decrementing the count, blocking if the 
// semaphore is not free.
//

PilResult PIL_AcquireSemaphore(PilSemaphore Sem, uint32_t Timeout)
{
    sem_t *arg = (sem_t *) Sem;

    pil_checksem(arg,PIL_INVALID_HANDLE); // Check for invalid semaphore

    if (Timeout == 0) {
        PilResult result = PIL_OK;

        pthread_mutex_lock(&arg->mx);
        if (arg->cur) {
            arg->cur--;
        } else {
            result = PIL_TIMEOUT;
        }
        pthread_mutex_unlock(&arg->mx);
        return result;
    }

    if (Timeout == PIL_INFINITE_TIMEOUT) {
        pthread_mutex_lock(&arg->mx);
        while (!arg->cur) {
            pthread_cond_wait(&arg->cv,&arg->mx);
        }
        arg->cur--;
        pthread_mutex_unlock(&arg->mx);
        return PIL_OK;
    } else {
        struct timeval tv;
        struct timespec ts;
        uint32_t t = Timeout / 1000;
        int r = 0;

        gettimeofday(&tv,NULL);
        ts.tv_sec = tv.tv_sec + t;
        ts.tv_nsec = (tv.tv_usec+(Timeout-t*1000)*1000)*1000;
        pthread_mutex_lock(&arg->mx);
        while (!arg->cur && r != ETIMEDOUT) {
            r = pthread_cond_timedwait(&arg->cv,&arg->mx,&ts);
        }
        if (r == 0) arg->cur--;
        pthread_mutex_unlock(&arg->mx);
        switch (r) {
          case 0:
            return PIL_OK;
          case ETIMEDOUT:
            return PIL_TIMEOUT;
          default:
            return PIL_ERROR;
        }
    }
}

///////////////////////////////////////////////////////////////////////
//
// PIL_ReleaseSemaphore
//
// Releases a semaphore, incrementing the count by one, waking a thread
// waiting on the semaphore.
//

PilResult PIL_ReleaseSemaphore(PilSemaphore Sem)
{
    sem_t *arg = (sem_t *) Sem;

    pil_checksem(arg,PIL_INVALID_HANDLE); // Check for invalid semaphore

    pthread_mutex_lock(&arg->mx);
    if (arg->cur < arg->max) {
        arg->cur++;
        pthread_cond_signal(&arg->cv);
    }
    pthread_mutex_unlock(&arg->mx);

    return PIL_OK;
}
