/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILTHD.C                                                  //
// Module:   Platform Isolation Layer Thread Management Routines       //
// Author:   JP Sugarbroad                                             //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////

#include <string.h>
#include "pil_int.h"

#ifdef DEBUG
    // Magic number for debug checking
#   define PIL_THREAD_MAGIC (0xB4DC3002)

    // Macro to check for invalid critical sections
#   define pil_checkthread(x,r) do { \
        if (!(x) || (x)->magic != PIL_THREAD_MAGIC) { \
            ASSERTMSG(FALSE,"Attempt to use bad thread id."); \
            return r; \
        } \
    } while (0);
#else // !DEBUG
#   define pil_checkthread(x,r)
#endif

// Structure used to communicate between PIL_CreateThread and thread_glue
typedef struct {
    PilThreadRoutine routine;
    void * data;
    uint32_t options;
    pthread_mutex_t mx;
    pthread_cond_t cv;
} create_t;

static pillist thread_list = NULL;

#ifdef DEBUG
pillist _PIL_GetThreadList()
{
    return thread_list;
}
#endif

/////////////////////////////////////////////////////////////////////////
//
// thread_glue (internal procedure)
//
// Glue code for PIL_CreateThread
//

void *thread_glue(void *arg)
{
    create_t *c = (create_t *) arg;
    PilThreadRoutine routine = c->routine;
    void * data = c->data;
    uint32_t options = c->options;

    pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS,NULL);
    pthread_mutex_lock(&c->mx);
    pthread_cond_signal(&c->cv);
    pthread_mutex_unlock(&c->mx);
    if (options & PIL_THREAD_SUSPENDED) pthread_kill(pthread_self(),SIGSTOP);

#ifdef DEBUG
    t = PIL_GetCurrentThread();
    if (t)
      t->status = THREAD_ACTIVE;
#endif
    (*routine)(data);
    PIL_ExitThread(0);
    return(NULL);
}

///////////////////////////////////////////////////////////////////////
//
// PIL_CreateThread
//
// Creates a new thread executing at the given address.  
//
// NOTE: Currently does not support priority
//

PilThread _PIL_CreateThread(
    PilThreadRoutine StartAddress,
    uint32_t dwStackSize,
    char *Name,
    uint32_t Priority,
    void * Param,
    uint32_t dwOptions
) 
{
    thread_t *t;
    create_t c = {
        StartAddress,
        Param,
        dwOptions,
        PTHREAD_MUTEX_INITIALIZER,
        PTHREAD_COND_INITIALIZER
    };
    int r;

    // Allocate and initialize the thread

    t = (thread_t *) PIL_AllocateMemory(sizeof(thread_t),0);
    if (t == NULL) return (PilThread) NULL;

#ifdef DEBUG
    t->magic = PIL_THREAD_MAGIC;
#endif
    if (Name != NULL) {
        t->name = (char *) PIL_AllocateMemory(strlen(Name)+1,0);
        if (t->name != NULL) {
            strcpy(t->name,Name);
        }
    } else {
        t->name = NULL;
    }
#ifdef DEBUG
    t->status = THREAD_SUSPENDED;
#endif

    pthread_mutex_lock(&c.mx);

    if ((r = pthread_create(&t->thread,NULL,thread_glue,&c))) {
        if (t->name) PIL_FreeMemory(t->name);
        PIL_FreeMemory(t);
        return (PilThread) NULL;
    }

    pthread_cond_wait(&c.cv,&c.mx);
    pthread_mutex_unlock(&c.mx);

    if (dwOptions & PIL_THREAD_SUSPENDED) {
        // Allow time for the thread to suspend itself
        nanosleep(&(const struct timespec){0,100000000},NULL);
    }

    _PIL_AddToList(&thread_list,t);

    return (PilThread) t;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_DelayThread
//
// Delays execution of the currently running thread by the given number
// of milliseconds.
//

PilResult PIL_DelayThread(uint32_t dwDelay)
{
    struct timespec ts;

    ts.tv_sec = dwDelay / 1000;
    ts.tv_nsec = (dwDelay % 1000) * 1000000;

    while (nanosleep(&ts,&ts) == -1) {
        if (errno != EINTR) return PIL_ERROR;
    }
    return PIL_OK;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_DestroyThread
//
// Destroys the thread by releasing the stack and any other memory
// relating to the thread.  The thread must have exited before this
// function is called!
//
// NOTE: This will block if the thread isn't dead!
//

PilResult PIL_DestroyThread(PilThread Thread, uint32_t *pdwStatus)
{
    thread_t *t = (thread_t *) Thread;

    pil_checkthread(t,PIL_INVALID_HANDLE);

    switch (pthread_join(t->thread,(void **) pdwStatus)) {
      case ESRCH:
        return PIL_INVALID_HANDLE;
      case 0:
        break;
      default:
        return PIL_ERROR;
    }

    _PIL_RemoveFromList(&thread_list,t);

    return PIL_OK;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_ExitThread
//
// Exits the currently executing thread function.  The thread data
// structures are not destroyed upon calling this function but must
// still be destroyed with PIL_DestroyThread.  Note that this function
// does not return.
//

void PIL_ExitThread(uint8_t dwStatus)
{
#ifdef DEBUG
    thread_t *t;
    t = (thread_t*)PIL_GetCurrentThread();
    if (t)
      t->status = THREAD_EXITED;
#endif
    pthread_exit((void *)(uint32_t)dwStatus);
}

int ThreadCompare(void* a, void* b)
{
    thread_t *t = (thread_t*)a;
    return pthread_equal(t->thread,(pthread_t)b);
}

///////////////////////////////////////////////////////////////////////
//
// PIL_GetCurrentThread
//
// Returns the handle of the currently executing thread.
//

PilThread PIL_GetCurrentThread()
{
    pthread_t search = pthread_self();
    thread_t *t;

    t = (thread_t*)_PIL_FindInList(&thread_list,(void *)search,ThreadCompare);

    return (PilThread) t;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_GetCurrentThreadId
//
// Returns the threadid of the currently executing thread.
//

unsigned long PIL_GetCurrentThreadId()
{
    return pthread_self();
}

///////////////////////////////////////////////////////////////////////
//
// PIL_GetThreadPriority
//
// Returns the priority of the given thread.
//

PilResult PIL_GetThreadPriority(PilThread Thread, uint32_t *pdwPriority)
{
    return PIL_NOT_IMPLEMENTED;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_ResumeThread
//
// Resumes a thread that has either been suspended via PIL_SuspendThread
// or created using the PIL_THREAD_SUSPENDED creation flag.
//

PilResult PIL_ResumeThread(PilThread Thread)
{
    thread_t *t = (thread_t *) Thread;

    pil_checkthread(t,PIL_INVALID_HANDLE);

    switch (pthread_kill(t->thread,SIGCONT)) {
      case 0:
#ifdef DEBUG
        t->status = THREAD_ACTIVE;
#endif
        return PIL_OK;
      case ESRCH:
        return PIL_INVALID_HANDLE;
      default:
        return PIL_ERROR;
   }
}

///////////////////////////////////////////////////////////////////////
//
// PIL_SetThreadPriority
//
// Changes the priority of the given thread.
//

PilResult PIL_SetThreadPriority(PilThread Thread, uint32_t dwPriority)
{
    return PIL_NOT_IMPLEMENTED;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_SuspendThread
//
// Suspends the given thread stopping it from executing.  The thread
// can be resumed with PIL_ResumeThread.
//

PilResult PIL_SuspendThread(PilThread Thread)
{
    thread_t *t = (thread_t *) Thread;

    pil_checkthread(t,PIL_INVALID_HANDLE);

    switch (pthread_kill(t->thread,SIGSTOP)) {
      case 0:
#ifdef DEBUG
        t->status = THREAD_SUSPENDED;
#endif
        return PIL_OK;
      case ESRCH:
        return PIL_INVALID_HANDLE;
      default:
        return PIL_ERROR;
   }
}

///////////////////////////////////////////////////////////////////////
//
// PIL_TerminateThread
//
// Unconditionally terminates the thread.  This essentially causes the
// thread to exit but the thread still needs to be destroyed via
// PIL_DestroyThread.
//

PilResult PIL_TerminateThread(PilThread Thread)
{
    thread_t *t = (thread_t *) Thread;

    pil_checkthread(t,PIL_INVALID_HANDLE);

    switch (pthread_cancel(t->thread)) {
      case 0:
#ifdef DEBUG
        t->status = THREAD_INVALID;
#endif
        break;
      case ESRCH:
        return PIL_INVALID_HANDLE;
      default:
        return PIL_ERROR;
    }

    return PIL_ResumeThread(Thread);
}


///////////////////////////////////////////////////////////////////////
//
// PIL_WaitForThread
//
// this pil api is being defined temporarily by renin. it was not originally 
// defined, and from the usage in the windows version, it appears that this 
// call is used to wait for the thread to complete before killing it. So
// i am using the pthread_join call in linux which appears to do the job.
// But someone should look into this and make sure its right.
//

PilResult PIL_WaitForThread(PilThread Thread, uint32_t dwTimeout)
{
    int		*thread_return;
    thread_t *t = (thread_t *) Thread;

    pil_checkthread(t,PIL_INVALID_HANDLE);

    switch (pthread_join(t->thread, (void **) &thread_return)) {
      case 0:
        return PIL_OK;
      case ESRCH:
        return PIL_INVALID_HANDLE;
      default:
        return PIL_ERROR;
    }
}
