/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: CQUEUE.C                                                  //
// Module:   Platform Independent Layer Mailbox Services               //
// Author:   Venky Gokulrangan                                         //
//           Inspired by Dan Baumberger's work on mailbox in C++       //
//           Kinda replicated in C to maintain the same interface      //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
// PURPOSE: Implement Fixed Sized Circular Queue for Linux             //
//                                                                     //
// REQUIRES: Some functions from cqueue.c are called                   //
//                                                                     //
/////////////////////////////////////////////////////////////////////////

//#include "cqueue.h"
#include "pil_vx.h"


//////////////////////////////////////////////////////////////////////////////
//
// Function:   IsFull
// Parameters: Pointer to the Circular Queue Header
// Purpose:    Checks to see if we have reached the limits of the queue
//
// NOTES:
// 1. The corresponding Handle MUST HAVE BEEN LOCKED before calling this function
// This is just to promise the light-weghtedness of the function
// Otherwise the TRUTH of the return value is NOT guaranteed
//
// 2. NULL Handle should not be passed in to this function
// Since returning TRUE OR FALSE for a NULL Handle will not make sense
//
// 3. This function is internal to this file and not exposed outside
//

static
BOOL
IsFull (
     PilCQ* pCQ)
{
   return ( pCQ->dwCount == pCQ->dwQLen ) ;
}


//////////////////////////////////////////////////////////////////////////////
//
// Function:   IsEmpty
// Purpose:    Checks to see if the Circular Queue is empty
// Returns:    TRUE if the Queue is empty; FALSE otherwise
//
// NOTES:
// 1. The corresponding Handle MUST HAVE BEEN LOCKED before calling this function
// This is just to promise the light-weghtedness of the function
// Otherwise the TRUTH of the return value is NOT guaranteed
//
// 2. NULL Handle should not be passed in to this function
// Since returning TRUE OR FALSE for a NULL Handle will not make sense
//
// 3. This function is internal to this file and not exposed outside
//

static
BOOL
IsEmpty (
     PilCQ* pCQ)
{
   return ( pCQ->dwCount == 0 ) ;
}

//////////////////////////////////////////////////////////////////////////////
//
// Function:   CreateCQ
// Parameters: Number of Units in the Queue
// Purpose:    Creates a circular queue of the size as specified.
//             The number of nodes in this queue is fixed and is determined
//             by the size parameter "dwQLen"
//             1. First it allocates the header
//             2. Then it creates the semaphore that serializes the access to
//                the semaphore
//             3. Then it creates the queue itself..  Each of the
//
// Returns:    Handle of the created circular queue
//             NULL, if unable to create the circular queue
//
// Notes:      In any case, if an error occurs, all the resources allocated
//             uptil the previous point are removed
//

CQHandle
CreateCQ (
       DWORD dwQLen,
       DWORD dwMboxSize )
{
   BOOL    Error = FALSE ;
   PilCQ*  Handle ;

   Handle = (PilCQ *)PIL_AllocateMemory ( sizeof (PilCQ), PIL_MEM_ZERO_INIT ) ;
   if ( Handle )
   {
      Handle->dwQLen     = dwQLen ;
      Handle->dwMailSize = dwMboxSize ;
      Handle->dwHead = Handle->dwTail = Handle->dwCount = 0;

      Handle->AccessSem = PIL_CreateSemaphore (dwQLen, dwQLen) ;
      if ( !Handle->AccessSem )
         Error = TRUE ;
      else
      {
         // Allocate the actual circular buffer
         Handle->pQ = PIL_AllocateMemory ( dwMboxSize * dwQLen ,
                                           PIL_MEM_ZERO_INIT ) ;
         if ( !Handle->pQ )
            Error = TRUE ;
      }

      if ( Error )
      {
         if ( Handle->AccessSem )
            PIL_DestroySemaphore ( Handle->AccessSem ) ;

         if ( Handle->pQ )
            PIL_FreeMemory ( Handle->pQ ) ;

         PIL_FreeMemory ( Handle ) ;
         Handle = NULL ;
      }
   }

   return (CQHandle)Handle ;
}


//////////////////////////////////////////////////////////////////////////////
//
// Function:   Enqueue
// Parameters: 1. Handle to the Circular Queue
//             2. The address to copy the mail message from
// Purpose:    1. Checks if an empty slot is available in the circular queue
//             2. If so, it copies the message from the user address space
//                on to this queue element.
//             3. Usual error and sanity checks are performed
// Returns:    QUEUE_ERROR if unable to perform the operation
//             QUEUE_FULL  if no space is available
//             QUEUE_OK    if succesful
//

QueueReturn
Enqueue ( 
   CQHandle hCQ,
   PVOID Object ) // Object itself is an address
{
   QueueReturn Ret ;
   PilCQ* Handle = (PilCQ *)hCQ ;

   if ( !Handle )
      return QUEUE_ERROR ;

   if ( PIL_AcquireSemaphore ( Handle->AccessSem, PIL_INFINITE_TIMEOUT ) != PIL_OK )
      return QUEUE_ERROR ;

   if ( IsFull (Handle) )
      Ret = QUEUE_FULL ;
   else
   {
      memcpy ( (PVOID)((PBYTE)Handle->pQ + Handle->dwTail * Handle->dwMailSize), 
                      Object, Handle->dwMailSize ) ;
      Handle->dwTail = (Handle->dwTail+1) % Handle->dwQLen ;
      Handle->dwCount++ ;
      Ret = QUEUE_OK ;
   }

   PIL_ReleaseSemaphore ( Handle->AccessSem ) ;
   return Ret ;
}

//////////////////////////////////////////////////////////////////////////////
//
// Function:   QueueReturn
// Parameters: 1. Handle to the Circular Queue
//             2. The address to copy the mail message to
// Purpose:    1. Checks if atleast one element in the queue is present
//             2. If so, it copies the element to the user address space
//             3. Usual error and sanity checks are performed
// Returns:    QUEUE_ERROR if unable to perform the operation
//             QUEUE_FULL  if no element is available to copy from to user
//             QUEUE_OK    if succesful
//
QueueReturn
DeQueue (
         CQHandle hCQ,
         PVOID    pObject)
{
   QueueReturn Ret = QUEUE_OK ;
   PilCQ* Handle = (PilCQ *)hCQ ;

   if ( !Handle )
      return QUEUE_ERROR ;

   if ( PIL_AcquireSemaphore ( Handle->AccessSem, PIL_INFINITE_TIMEOUT ) != PIL_OK )
      return QUEUE_ERROR ;

   if ( IsEmpty (Handle) )
      Ret = QUEUE_EMPTY ;
   else
   {
      memcpy ( pObject, (PVOID)((PBYTE)Handle->pQ + Handle->dwHead * Handle->dwMailSize), 
                        Handle->dwMailSize ) ;
      Handle->dwHead = (Handle->dwHead + 1) % Handle->dwQLen ;
      Handle->dwCount-- ;
   }

   PIL_ReleaseSemaphore ( Handle->AccessSem ) ;
   return Ret ;
}
