/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILEVNT.C                                                 //
// Module:   Platform Isolation Layer Memory Routines                  //
// Author:   Venky Gokulrangan                                         //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
// PURPOSE: Implement functions for PIL Event Routines for VxWorks     //
//                                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//
// HISTORY:
//
/////////////////////////////////////////////////////////////////////////

#include "pil_vx.h"


///////////////////////////////////////////////////////////////////////
// Notes:    Events are simulated on top of semaphores
//           But multiple threads can wait on the same semaphore
//           Thus for every thread, waiting on the same semaphore
//           the NumWaiters will be incremented by 1. Upon the release
//           of the semaphore, signal (V) operation will be performed
//           internally by this routine "NumWaiters" times.. At that
//           instant any process, waiting on the semaphore will be
//           "signlled"
//           Since multiple threads can access the same event
//           the event itself is protected by an access semaphore
///////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////
//
// Function: PIL_CreateEvent
//
// Purpose:  Creates an event object
//           In VxWorks, there is no concept of an event
// Notes:    Refer Notes above
//
// RETURNS
//
// A handle to the newly created event object or NULL if the object 
// could not be created.
//

PilEvent
PIL_CreateEvent(
   BOOL bManualReset,   // TRUE for manual reset, FALSE for auto-reset
   BOOL bSignaled      // TRUE for signaled, FALSE for non-signaled
   )
{
   // BOOL is not type safe.. Need to check for values for TRUE/FALSE ?
   PilEvent_t* NewInfo = NULL;

   if ( (NewInfo = (PilEvent_t *)PIL_AllocateMemory ( 
                           sizeof (PilEvent_t), PIL_MEM_ZERO_INIT ) ) )
   {
      NewInfo->bManualReset = bManualReset ;
      NewInfo->bSignalled   = bSignaled ;
      NewInfo->Flag         = bSignaled ;

      NewInfo->EventSem     = PIL_CreateSemaphore (bSignaled, TRUE) ;
      NewInfo->AccessSem    = PIL_CreateBinarySemaphore (TRUE, TRUE) ;

      /* GWG - add check here to make sure objects were created. If not
         make sure NULL is returned. */
      if (NewInfo->EventSem == NULL || NewInfo->AccessSem == NULL)
      {
          if (NewInfo->EventSem != NULL)
          {
              PIL_DestroySemaphore(NewInfo->EventSem);
          }

          if (NewInfo->AccessSem != NULL)
          {
              PIL_DestroySemaphore(NewInfo->AccessSem);
          }

          PIL_FreeMemory(NewInfo);

          /* Set return value to NULL */
          NewInfo = NULL;
      }
       

   }
   else
   {
      TRACE("%s: Unable to allocate memory.\n", __FUNCTION__ ) ;
   }

   return (PilEvent) NewInfo ;
}


///////////////////////////////////////////////////////////////////////
//
// Function: PIL_DestroyEvent
//
// Purpose:  Destroys an event object created using PIL_CreateEvent.
//           The memory will be freed and the semaphores associated
//           with this Event object will be destroyed as well.
//
// Returns:
// PIL_OK          Event successfully destroyed.
// PIL_INVALID_HANDLE  The event handle is not valid.
// PIL_ERROR           If semaphore could not be destroyed
//

PilResult
PIL_DestroyEvent(
   PilEvent Handle// The handle of the event to destroy
   )
{
   PilResult Result = PIL_OK  ;
   PilEvent_t* p = (PilEvent_t *)Handle ;
   // Notify all the waiters that it is no longer available

   if ( !p )
      return PIL_INVALID_HANDLE ;
   // CHECK FOR VALIDITY OF THE HANDLE!!!

   if ( PIL_DestroySemaphore ( p->EventSem ) != PIL_OK )
      Result = PIL_ERROR ;

   if ( PIL_DestroySemaphore (p->AccessSem) != PIL_OK )
      Result = PIL_ERROR ;

   PIL_FreeMemory ( p ) ;

#ifdef DEBUG
#endif

   return Result ;
}


PilResult
PIL_ResetEvent(
   PilEvent Handle // The handle of the event to reset
   )
{
   PilEvent_t *p = (PilEvent_t *)Handle ;
   PilResult  Result = PIL_OK ;

   if ( !p )
      return PIL_INVALID_HANDLE ;

  // Locate the Event

   PIL_AcquireSemaphore (p->AccessSem, PIL_INFINITE_TIMEOUT) ;

   if ( p->Flag )
      p->Flag = FALSE ;

   PIL_ReleaseSemaphore (p->AccessSem) ;

   return Result;
}

///////////////////////////////////////////////////////////////////////
//
// Function: PIL_SetEvent
//
// Purpose: Sets an event into the signaled state. The Internal flag
// used to maintain the state of the event is set to TRUE. When the evnt
// is set and there are waiters waiting on this event, all of the 
// waiters are woken up, (by fakely signalling the semaphore. The result
// from PIL_ReleaseSemaphore () is ignored.. PIL_Semaphores check 
// against the maximum value to make sure that they do not increase
// the count of the semaphore.
//
// RETURNS
//
// PIL_OK      The event was succefully signaled
// PIL_INVALID_HANDLE   The event handle is not valid.
// PIL_ERROR            The semaphore operation failed
//

PilResult
PIL_SetEvent (
        PilEvent Handle)    // The handle of the event to set
{
   int i;
   PilResult  Result = PIL_OK ;
   PilEvent_t* p = (PilEvent_t *)Handle ;

   PIL_AcquireSemaphore ( p->AccessSem, PIL_INFINITE_TIMEOUT) ;

   // Increment the value of the semaphore by as many waiters
   for ( i=0; i<p->NumWaiters; i++ )
      PIL_ReleaseSemaphore (p->EventSem) ;

   // Successfully notified all the waiters
   // Reset the number of waiters

   p->NumWaiters = 0 ; // Reset number of waiters
   p->Flag   = TRUE ;  // Force  state to be sure

   if ( p->bManualReset == FALSE ) // Auto Reset
      p->Flag = FALSE ; // RESET

   PIL_ReleaseSemaphore ( p->AccessSem ) ;

   return Result ;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_WaitForEvent
//
// Waits for the the state of the event to be signalled
// 1. Initially it locks the access semaphore
// 1. Checks to see if the Flag is TRUE. ( Already set)
// 2.  If it is already set, then the event is true, so it returns
//     after releasing the access semaphore
// 3.  If it is not already set, then this has to wait on the event
//     sem by incrementing the number of waiters on this event.
// 4.  Since the Num of waiters is an event variable, the access sem
//     is released right after the value is incremented.
// Return Values:
//
// PIL_OK      The event was succefully signaled
// PIL_INVALID_HANDLE   The event handle is not valid.
// PIL_TIMEOUT          The timeout expired before the event was signalled
//
//

PilResult
PIL_WaitForEvent (
         PilEvent  Handle,
         DWORD     dwTimeOut)
{
   PilEvent_t *p = (PilEvent_t *)Handle ;

   PIL_AcquireSemaphore ( p->AccessSem, PIL_INFINITE_TIMEOUT ) ;

   if ( p->Flag != TRUE ) // We need to wait
   {
      p->NumWaiters++ ;

      PIL_ReleaseSemaphore ( p->AccessSem ) ;
      if ( PIL_AcquireSemaphore ( p->EventSem, dwTimeOut ) == PIL_TIMEOUT)
         return PIL_TIMEOUT ;
   }
   else
      PIL_ReleaseSemaphore ( p->AccessSem ) ;

   return PIL_OK ;
}
