/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILHEAP.C                                                 //
// Module:   PIL Local Heap Management Routines.                       //
// Author:   Venky Gokulrangan                                         //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
// PURPOSE:                                                            //
//   Implements a thread  safe heap manager.                           //
//                                                                     //
// NOTES:                                                              //
//   This code was derived from Dan Baumberger's Heap Manager which    //
//   was derived from Dynamic Heap Memory Manager written by Ramanan G //
//   during Dec 30 1993.                                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//
// HISTORY:
/////////////////////////////////////////////////////////////////////////

#include "pil_vx.h"


///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
//                                                                   //
//                  P U B L I C  F U N C T I O N S                   //
//                                                                   //
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////
//
// PIL_CreateHeap
//
// DESCRIPTION
//   Creates a new managed heap.
//
// RETURNS
//   The handle of the new heap or NULL.
//
//

PilHeap
PIL_CreateHeap (
        DWORD dwMaxHeapSize )   // The maximum amount of mem in the heap
{
   HeapDesc* pHeap;
   ASSERTMSG(dwMaxHeapSize > sizeof ( HeapHeader ),
           "HeapCreate: pool size is too small to allocate anything!" ) ;

   if( !(pHeap = PIL_AllocateMemory(sizeof(HeapDesc), PIL_MEM_ZERO_INIT)) )
   {
      ASSERTMSG(FALSE,"HeapCreate: cannot allocate descriptor!" );
      return( (PilHeap)NULL );
   }

   pHeap->dwSize = dwMaxHeapSize + sizeof( HeapHeader );
   if(!(pHeap->phhHead = PIL_AllocateMemory( pHeap->dwSize,PIL_MEM_ZERO_INIT )))
   {
      ASSERTMSG(FALSE,"HeapCreate: cannot allocate memory pool!" );
      PIL_FreeMemory ( pHeap ) ;
      return( (PilHeap)NULL );
   }

   if ( !(pHeap->LockSem = PIL_CreateSemaphore (1, 1)) )
   {
      ASSERTMSG(FALSE,"HeapCreate: Semaphore Creation failed!\n" ) ;
      PIL_FreeMemory ( pHeap->phhHead );
      PIL_FreeMemory ( pHeap ) ;
      return (PilHeap)NULL ;
   }

   pHeap->phhTail         = ( PHeapHeader )
                             ( (PBYTE) pHeap->phhHead +
                              dwMaxHeapSize +
                              sizeof( HeapHeader ) );

   pHeap->phhHead->pPrev  = NULL;
   pHeap->phhHead->dwSize = dwMaxHeapSize;
   pHeap->phhHead->bInUse = FALSE;

   return( (PilHeap)pHeap );
}

///////////////////////////////////////////////////////////////////////
//
// PIL_DestroyHeap
//
// DESCRIPTION
//   Deletes a managed heap created using HeapCreate.
//
//

PilResult
PIL_DestroyHeap (
   PilHeap hHeap  )    // The heap to delete
{
    HeapDesc* pHeap = (HeapDesc*) hHeap;

    if( !pHeap )
    {
        ASSERTMSG(pHeap, "HeapDelete: NULL HeapHandle!" );
        return PIL_INVALID_HANDLE ;
    }

    if( pHeap->phhHead )
    {
       PIL_FreeMemory (pHeap->phhHead);
    }

    if ( pHeap->LockSem )
    {
       PIL_DestroySemaphore (pHeap->LockSem) ;
    }

    PIL_FreeMemory (pHeap) ;
    return PIL_OK ;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_AllocHeap
//
// DESCRIPTION
//   Allocates a block of the given size of the managed heap.
//
//

PVOID
PIL_AllocHeap (
     PilHeap hHeap,    // The heap to allocate from
     DWORD   dwSize  ) // The size of block to allocate
{
    HeapDesc*   pHeap = (HeapDesc*) hHeap;
    DWORD       dwAllocSize;
    PHeapHeader phhTemp, phhAllocMem, phhNext ;

    if( !pHeap )
    {
        ASSERTMSG(FALSE,"PIL_AllocHeap: NULL HeapHandle!" );
        return( NULL );
    }

    // add the size of our header into the allocation size.
    // DWord align the allocation

    dwAllocSize = dwSize + sizeof( HeapHeader );
    dwAllocSize = ( ((dwAllocSize - 1) >> 2 ) + 1 ) << 2 ;

    TRACE("Allocating heap of size: %d", dwAllocSize) ;

    // GVR: Notes: For the user level, instead of having to disable the interrupts
    // (which the user level thread/process cannot do anyway) we will acquire a private
    // semaphore and release when we are done.

    PIL_AcquireSemaphore ( pHeap->LockSem, PIL_INFINITE_TIMEOUT ) ;
    phhTemp = pHeap->phhHead;

    do
    {
        if( ( phhTemp->dwSize >= dwSize ) && !phhTemp->bInUse )
        {
            // Found a free block of the right size.
            phhAllocMem           = phhTemp;
            phhTemp               = (PHeapHeader)
                                    ((PBYTE) phhTemp + dwAllocSize);

            if( phhTemp < pHeap->phhTail )
            {
                phhTemp->dwSize        = phhAllocMem->dwSize - dwAllocSize;
                phhAllocMem->dwSize    = dwAllocSize - sizeof( HeapHeader );
                phhAllocMem->bInUse   = TRUE;
                phhTemp->bInUse       = FALSE;
                phhTemp->pPrev        = phhAllocMem;
                phhNext               = (PHeapHeader)
                                        ((PBYTE) phhTemp + sizeof(HeapHeader) + 
                                          phhTemp->dwSize) ;
                if ( phhNext < pHeap->phhTail )
                   phhNext->pPrev = phhTemp ;
            }
            else
            {
                phhAllocMem->dwSize    = dwSize;
                phhAllocMem->bInUse    = TRUE;
            }
            PIL_ReleaseSemaphore ( pHeap->LockSem ) ;
            return( (PVOID) ((PBYTE) phhAllocMem + sizeof( HeapHeader ) ) );
        }
        else
        {
           // This block is too small or somebody already beat us
           // to it.  Let's find the next block.
           phhTemp                 =  (PHeapHeader)
                            ((PBYTE) phhTemp +  phhTemp->dwSize + sizeof(HeapHeader) ) ;
        }
    } while( phhTemp < pHeap->phhTail ) ;

    // We've walked the entire list and haven't found a suitable block.
    // Return an error.

    PIL_ReleaseSemaphore ( pHeap->LockSem ) ;

    return( NULL );
}

///////////////////////////////////////////////////////////////////////
//
// PIL_FreeHeap
//
// DESCRIPTION
//   Deallocates a memory block from a managed heap.
//

PilResult
PIL_FreeHeap (
    PilHeap hHeap,    // The heap to deallocate from
    PVOID   pBuf )    // The pointer to free
{
    HeapDesc* pHeap = (HeapDesc*) hHeap;
    PHeapHeader phhHeap, phhTemp, phhNext;

    phhHeap = pHeap->phhHead;
    phhTemp = ( PHeapHeader ) ((PBYTE) pBuf - sizeof( HeapHeader ));

    if( ( phhTemp < pHeap->phhHead ) ||
        ( phhTemp > pHeap->phhTail ) )
    {
        ASSERTMSG(FALSE,"FreeHeap: pointer was not allocated from this heap!" );
        PIL_ReleaseSemaphore ( pHeap->LockSem ) ;
        return PIL_ERROR ;
    }

    if ( !phhTemp->bInUse )
    {
       ASSERTMSG(FALSE, "Memory already freed OR not in use\n") ;
       PIL_ReleaseSemaphore ( pHeap->LockSem ) ;
       return PIL_ERROR ;
    }

    PIL_AcquireSemaphore ( pHeap->LockSem, PIL_INFINITE_TIMEOUT ) ; // Infinite timeout
    phhNext = (PHeapHeader)
              ( (PBYTE)phhTemp + phhTemp->dwSize + sizeof (HeapHeader) ) ;

    // CASE 1: Freed entry is the head of the list
    if( phhTemp->pPrev == NULL )
    {
        // CASE 1a: Next entry is free
        if( !phhNext->bInUse )
        {
            phhTemp->bInUse              = FALSE;
            // GVR -- Size of the header is added only once since the 
       // first header is mandatory
            phhTemp->dwSize              += sizeof( HeapHeader ) +
                                           phhNext->dwSize;

            phhTemp = (PHeapHeader)( (PBYTE)phhTemp + phhTemp->dwSize ) ;

            // Check the next heaeder
            if ( phhNext < pHeap->phhTail )
               phhNext->pPrev = phhTemp ;

            PIL_ReleaseSemaphore ( pHeap->LockSem ) ;
            return PIL_OK ;
        }
        // CASE 1b: Next entry is not free
        else
        {
            phhTemp->bInUse = FALSE;
            PIL_ReleaseSemaphore ( pHeap->LockSem ) ;
            return PIL_OK ;
        }
    }

    // CASE 2: Freed entry in in the middle of the list
    phhNext = (PHeapHeader)
              ( (PBYTE)phhTemp + phhTemp->dwSize + sizeof(HeapHeader) ) ;

    if( ( phhTemp->pPrev != NULL ) &&
        ( phhNext <  pHeap->phhTail ) )
    {
        // CASE 2a: Prev and Next are both free
        if ( ( !phhTemp->pPrev->bInUse ) &&
             ( !phhNext->bInUse ) )
        {
            phhTemp->pPrev->bInUse       = FALSE;
            phhTemp->pPrev->dwSize       += 2 * sizeof( HeapHeader ) +
                                           phhTemp->dwSize +
                                           phhNext->dwSize;

            phhNext = (PHeapHeader)
                      ( (PBYTE)phhNext + phhNext->dwSize + sizeof(HeapHeader)) ;

            if ( phhNext < pHeap->phhTail )
               phhNext->pPrev = phhTemp->pPrev ;

            PIL_ReleaseSemaphore (pHeap->LockSem) ;
            return PIL_OK ;
        }

        // CASE 2b: Prev is free
        if( ( !phhTemp->pPrev->bInUse ) &&
            ( phhNext->bInUse ) )
        {
            phhTemp->pPrev->bInUse = FALSE;
            phhTemp->pPrev->dwSize += phhTemp->dwSize + sizeof( HeapHeader );
            phhNext->pPrev = phhTemp->pPrev ;

            PIL_ReleaseSemaphore ( pHeap->LockSem ) ;
            return PIL_OK ;
        }

        // CASE 2c: Next is free
        if( ( phhTemp->pPrev->bInUse ) &&
            ( !phhNext->bInUse ) )
        {
            phhTemp->bInUse              = FALSE;
            phhTemp->dwSize              += phhNext->dwSize +
                                           sizeof( HeapHeader );

            phhNext = (PHeapHeader)
                      ( (PBYTE) phhNext + phhNext->dwSize + sizeof(HeapHeader) ) ;

            // Set the block after the next's PREV to this'
            if ( phhNext <  pHeap->phhTail) // Bounds Check within buffer
                phhNext->pPrev = phhTemp ;

            PIL_ReleaseSemaphore ( pHeap->LockSem ) ;
            return PIL_OK;
        }

        // CASE 2d: Neither Prev or Next is free
        if( ( phhTemp->pPrev->bInUse ) &&
            ( phhNext->bInUse ) )
        {
            phhTemp->bInUse = FALSE;
            PIL_ReleaseSemaphore ( pHeap->LockSem ) ;
            return PIL_OK;
        }
    }

    // CASE 3: Freed entry is the end of the list
    if ( phhNext < pHeap->phhTail )
    {
        // CASE 3a: Prev is free
        if( !phhTemp->pPrev->bInUse )
        {
            phhTemp->pPrev->bInUse = FALSE;
            phhTemp->pPrev->dwSize += phhTemp->dwSize + sizeof( HeapHeader );

            PIL_ReleaseSemaphore ( pHeap->LockSem ) ;
            return PIL_OK;
        }
        // CASE 3b: Prev is NOT free
        else
        {
            phhTemp->bInUse = FALSE;
            PIL_ReleaseSemaphore ( pHeap->LockSem ) ;
            return PIL_OK;
        }
    }

    PIL_ReleaseSemaphore ( pHeap->LockSem ) ;

    ASSERTMSG(FALSE,"FreeHeap: heap is in bad shape!" );
    return PIL_ERROR ;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_ShrinkHeap
//
// DESCRIPTION
//   Reduces the size of an allocated block by constructing a new
//   header at the end of the new requested size, updating the links,
//   and calling HeapFree on the newly constructed block.  HeapFree
//   will take care of the case where the next block (after the new
//   block) is free.
//
//

PilResult
PIL_ShrinkHeap (
    PilHeap    hHeap,        // The heap to shrink a buffer from
    PVOID      pBuf,         // The pointer to the buffer to shrink
    DWORD      dwNewSize )   // The new (smaller) size
{
    HeapDesc*   pHeap = (HeapDesc*) hHeap;
    PHeapHeader phhTemp, phhNewBlock, phhNext ;
    
    // Using the client buffer pointer, find the heap header for this block

    phhTemp = (PHeapHeader) ((PBYTE) pBuf - sizeof( HeapHeader ) );

    //GVR Dont allow if we have just enough room for the header
    if( phhTemp->dwSize > ( dwNewSize + sizeof( HeapHeader ) ) )
    {
        ASSERTMSG(FALSE,"PIL_ShrinkHeap: insufficient memory to shrink block!" );
        return PIL_NOMEM ;
    }

    // Check to make sure the new block is SMALLER than the current one.

    if( phhTemp->dwSize <= dwNewSize )
    {
        ASSERTMSG(FALSE,"PIL_ShrinkHeap:Block size is the same or bigger than original!" );
        return PIL_INVALID_PARAM ;
    }

    PIL_AcquireSemaphore ( pHeap->LockSem, PIL_INFINITE_TIMEOUT ) ;

    phhNext             = (PHeapHeader)
                          ( (PBYTE) pBuf + phhTemp->dwSize + sizeof(HeapHeader) ) ;

    // Calculate the address of the new heap header.
    phhNewBlock         = (PHeapHeader) ( (PBYTE) pBuf + dwNewSize );

    // Insert the block into the linked list of buffers.
    phhNewBlock->pPrev  = phhTemp;

    // Calculate the new size and mark the block as InUse.  By saying the
    // the block is InUse, HeapFree can then be used the free it, taking
    // care of the details of merging the free blocks.

    phhNewBlock->dwSize  = phhTemp->dwSize - dwNewSize - sizeof( HeapHeader );
    phhNewBlock->bInUse = TRUE;
    
    // If the block being split it not the last block, we need to set the
    // pPrev pointer of that next block back to the newly created block.

    if ( phhNext < pHeap->phhTail )
       phhNext->pPrev = phhNewBlock ;

    // Update the original block.
    phhTemp->dwSize      = dwNewSize;
    PIL_ReleaseSemaphore ( pHeap->LockSem ) ;

    // Free the "new" block.  This will handle the merging of free memory
    // blocks.

    PIL_FreeHeap( hHeap, (PVOID)((PBYTE) phhNewBlock + sizeof( HeapHeader ) ) );
    return PIL_OK ;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_ScanHeap
//
// DESCRIPTION
//   Displays the current heap on the "stdio" device.
//
//

#ifdef  DEBUG

VOID
PIL_ScanHeap (
   PilHeap hHeap )
{
   HeapDesc*   pHeap = (HeapDesc*) hHeap;
   HeapHeader* phh;
   DWORD       dw;
    
   if( !pHeap )
   {
      ASSERTMSG(FALSE,"PIL_HeapScan: NULL heap handle!" );
      return;
   }

   PIL_AcquireSemaphore ( pHeap->LockSem, PIL_INFINITE_TIMEOUT ) ;
   phh = pHeap->phhHead;

   TRACE1 ("\nPIL_HeapScan( %x )\n",hHeap);

   while( phh < pHeap->phhTail )
   {
      if( phh->bInUse )
      {
         TRACE1 ("Used block of %d bytes.\n",phh->dwSize);
      }
      else
      {
         TRACE1 ("Free block of %d bytes.\n",phh->dwSize);
      }

      TRACE1 ("\tHeader:         %x\n",phh);
      TRACE1 ("\tUser Pointer:   %x\n",(PBYTE) phh + sizeof(HeapHeader));
      TRACE1 ("\tPrevious block: %x\n",phh->pPrev);
      phh = (PHeapHeader) ( (PBYTE)phh + phh->dwSize + sizeof(HeapHeader)) ;
   }

   PIL_ReleaseSemaphore ( pHeap->LockSem ) ;
}

#endif //  DEBUG
