/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILMEM.C                                                  //
// Module:   Platform Isolation Layer Memory Routines                  //
// Author:   Venky Gokulrangan                                         //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
// PURPOSE: Implement functions for PIL Memory Routines for VxWorks    //
//                                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//
// HISTORY:
//
/////////////////////////////////////////////////////////////////////////

#include "pil_vx.h"



////////////////////////////////////////////////////////////////////////
//
// PIL_AllocateMemory
//
// Allocates a block of memory of at least dwSize bytes in size.
//

PVOID
PIL_AllocateMemory(
    DWORD dwSize,  // Size, in bytes, to allocate
    DWORD dwFlags  // Special flags for allocation
    )
{
   PVOID pMem ;  // pointer to the allocated memory block

   // Sanity check..

   if ( (dwSize <= 0) ||
        ((dwFlags & PIL_MEM_PAGE_LOCK) == PIL_MEM_PAGE_LOCK))
      return NULL ;

   // DWORD Align it
   // Subtracting 1 and dividing by 4 and adding 1 will give in DWORD chunks
   // Then multiply by 4 will give the result
   // Size-- ; Size >>= 2;  Size++; Size << 2

   dwSize  = ( ((dwSize - 1) >> 2 ) + 1 ) << 2 ;
   pMem = (PVOID)malloc (dwSize) ;

   if ( (pMem != NULL) &&
        ((dwFlags & PIL_MEM_ZERO_INIT) == PIL_MEM_ZERO_INIT)
      )
   {
      memset ( pMem, 0, dwSize ) ;
   }

   return pMem ;
}

////////////////////////////////////////////////////////////////////////
//
// PIL_AllocatePages
//
// Allocates dwNumPages of contiguous pages of memory which must be 
// at least 1
//

PVOID
PIL_AllocatePages(
    DWORD  dwNumPages,  // The number of pages to allocate
    DWORD  dwFlags,     // Special flags for allocation
    PDWORD pdwPhysAddr  // Physical address of the pages
    )
{
   // In VxWorks, everything is in memory

   DWORD dwNumBytes = dwNumPages * PIL_BYTES_PER_PAGE ;


   // So we neeed to just return what PIL_Allocate_Memory does for us

   // Sanity Checking .. Checking for < 0 values will be meaningless
   // since DWORD is unsigned

   if ( dwNumPages <= 0 )
      return NULL ;

   return PIL_AllocateMemory (dwNumBytes, dwFlags) ;
}

////////////////////////////////////////////////////////////////////////
//
// PIL_FreeMemory
//
// Frees memory allocated using PIL_AllocateMemory.
//
//

VOID
PIL_FreeMemory(
    PVOID pMem  // Pointer to the memory to free
    )
{
   if ( pMem != NULL )  // we can do assertions
   {
      free (pMem) ;
   }
}

////////////////////////////////////////////////////////////////////////
//
// PIL_FreePages
//
// Frees dwNumPages allocated through PIL_AllocatePages.  dwNumPages
// must be the same as specified for PIL_AllocatePages.
//
//

VOID
PIL_FreePages(
    PVOID pMem,       // The linear address of the first page
    DWORD dwNumPages  // The number of pages to free
    )
{
   // Since we are guaranteed, that pMem points to the location which was
   // assigned by PIL_AllocatePages, and dwNumPages will be the same as the
   // one by PIL_AllocatePages, we assume that the memory location points to
   // a validly-assigned memory location by PIL routines

   if ( pMem != NULL )
   {
#ifdef DEBUG2
   printf ( "%s: Freeing Pages: %ld\n", __FUNCTION__, dwNumPages ) ;
#endif
      PIL_FreeMemory (pMem) ;
   }
}
