/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILSEM.C                                                  //
// Module:   Platform Isolation Layer Semaphores                       //
// Author:   Venky Gokulrangan                                         //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
// PURPOSE: VxWorks Semaphore Implementation                           //
//                                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//
// HISTORY:
//
// 
/////////////////////////////////////////////////////////////////////////

#include "semLib.h"
#include "pil_vx.h"

static
PilSemaphore_t*
SemCreate (
   DWORD dwInit,
   DWORD dwMax,
   BOOL  bSem )
{
   PilSemaphore_t* Handle;
   SEM_ID          SemId;

   if ( (Handle = (PilSemaphore_t *) PIL_AllocateMemory(
                                      sizeof(PilSemaphore_t ), 
                                      PIL_MEM_ZERO_INIT)) != NULL)
   {
      // Assume if the dwMaxCount equals 1, then it is a binary semaphore
      // pil.h #define's the PIL_BinarySemphoreCreate and routes the call to
      //  PIL_CreateSemaphore and so we can positively assume that.

      if ( bSem == TRUE )
      {
         Handle->dwCurrCount = (dwInit == 0 ? 0 : 1 );
         SemId = semBCreate ( SEM_Q_PRIORITY, 
                              Handle->dwCurrCount ? SEM_FULL:SEM_EMPTY);
      }
      else
         SemId = semCCreate ( SEM_Q_PRIORITY,
                              dwInit ) ;
      if ( SemId == NULL )
      {
         TRACE("COULD NOT CREATE SEM.. INSPECT") ;
         PIL_FreeMemory (Handle);
         Handle=NULL ;
      }
      else
      {
         Handle->Semid = SemId ;
         Handle->dwMax = dwMax ; 
         if (bSem == FALSE) 
            Handle->dwCurrCount = dwInit ; // initialize with count
      }
   }

   return Handle;
}


///////////////////////////////////////////////////////////////////////
//
// PIL_CreateSemaphore
//
//

PilSemaphore
PIL_CreateBinarySemaphore(
    DWORD dwInitialCount,   // Initial count of the semaphore
    DWORD dwMaximumCount   // Maximum count of the semaphore
    )
{
  // Result is just used as an internal variable. If at some point we    
  // are extending this routing to indicate the error type then use this
  // Result variable
   PilResult  Result = PIL_OK ;   
   PilSemaphore_t* p = NULL ;

   // could have used ( dwInitialCount/dwInitialCount & ~0x01)
   // just making it more readable :)

   if ( (dwInitialCount > dwMaximumCount) ||
        ( ! ( (dwInitialCount == 0) || (dwInitialCount == 1) ) ) ||
        ( ! ( (dwMaximumCount == 0) || (dwMaximumCount == 1) ) ) )
   {
      Result = PIL_INVALID_PARAM ;
   }
   else
   {
      p = SemCreate ( dwInitialCount, dwMaximumCount, TRUE) ;      
   }
   return (PilSemaphore)p ;
}


///////////////////////////////////////////////////////////////////////
//
// PIL_CreateSemaphore
//
//

PilSemaphore
PIL_CreateSemaphore(
    DWORD dwInitialCount,   // Initial count of the semaphore
    DWORD dwMaximumCount   // Maximum count of the semaphore
    )
{
  // Result is just used as an internal variable. If at some point we    
  // are extending this routing to indicate the error type then use this
  // Result variable
   PilResult  Result = PIL_OK ;   
   PilSemaphore_t* p = NULL ;

   if ( dwInitialCount > dwMaximumCount ) 
   {
      Result = PIL_INVALID_PARAM ;
   }
   else
   {
      p = SemCreate ( dwInitialCount, dwMaximumCount, FALSE ) ;      
   }
   return (PilSemaphore)p ;
}


///////////////////////////////////////////////////////////////////////
//
// PIL_AcquireSemaphore
//
//

PilResult
PIL_AcquireSemaphore(
    PilSemaphore  hSem,      // The handle of the semaphore to acquire
    DWORD         Timeout
    )
{ 
   int             ctTimeout =0;  // in ClockTicks ;

   // semTake requres that the Timeout be in clockticks rather than
   // in millsec.. Therefore, we convert the millisec into ticks by the
   // following manipulation

   if ( NULL == hSem )
      return PIL_INVALID_HANDLE ;
   else if ((WORD)Timeout <= 0)
      return PIL_INVALID_PARAM ;
   else
   {
      if ( Timeout == PIL_INFINITE_TIMEOUT )
         ctTimeout = WAIT_FOREVER ;
      else if (Timeout > 0)
      {
         // To get 1 ClockTick, we must spend atleast 16.66 msec..
         // Try for alternate implementations!!

         ctTimeout = getTicks (Timeout); // 50/3 is 16.666.. msec
      }

      if ( semTake ( ( (PilSemaphore_t *)hSem)->Semid, ctTimeout ) == ERROR )
      {
         if ( Timeout != PIL_INFINITE_TIMEOUT )
            return PIL_TIMEOUT ;
         else
            return PIL_ERROR ;
      }
      ( (PilSemaphore_t *)hSem)->dwCurrCount-- ;
      TRACE("Acquired Sem: VAL: %d", ( (PilSemaphore_t *)hSem)->dwCurrCount );
   }
   return PIL_OK ;
}


///////////////////////////////////////////////////////////////////////
//
// PIL_DestroySemaphore
//
//

PilResult
PIL_DestroySemaphore(
    PilSemaphore hSem   // The handle of the semaphore to destroy
    )
{
   if ( hSem == NULL )
      return PIL_INVALID_HANDLE ;

   return semDelete ( ((PilSemaphore_t *)hSem)->Semid ) == OK ? PIL_OK : PIL_ERROR  ;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_GetSemaphoreCount
//
//

PilResult
PIL_GetSemaphoreCount(
    PilSemaphore Sem,// The handle of the semaphore to get the count of
    PDWORD       pValue
    )
{
  // VxWorks does not support it.. So please note that PIL does not
  // support it either..
   ASSERT(Sem) ;
   if (!Sem)
      return PIL_INVALID_HANDLE ;

   *pValue = ( ( (PilSemaphore_t *)Sem)->dwCurrCount ) ;
   return PIL_OK ;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_ReleaseSemaphore
//
//

PilResult
PIL_ReleaseSemaphore(
    PilSemaphore hSem   // The handle of the semaphore to release
    )
{
   PilResult Ret = PIL_OK ;
   PilSemaphore_t* pSem = (PilSemaphore_t *)hSem ;

   ASSERT(hSem);
   if ( !hSem )
      Ret = PIL_INVALID_HANDLE ;
   else
   {
      if ( ( (pSem->dwCurrCount+1) > pSem->dwMax ) ||
           semGive (pSem->Semid) != OK )
      {
         TRACE("%s Overflow Error", __FUNCTION__);
         Ret = PIL_ERROR ;
      }
      else
      {
         ((PilSemaphore_t *)hSem)->dwCurrCount++ ;
      }
   }
   return Ret ;
}
