/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILTHD.C                                                  //
// Module:   Platform Isolation Layer Thread Management Routines       //
// Author:   Venky Gokulrangan                                         //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
// PURPOSE:  PIL Implementation for VxWorks Threads                    //
//                                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//
// HISTORY:
//
/////////////////////////////////////////////////////////////////////////

#include "pil_vx.h"
#include "pil_int.h"
#include "taskLib.h"
#include "taskVarLib.h"

PVOID pTaskData ;  // For thread (task) specific data

#ifdef DEBUG
pillist _PIL_GetThreadList() { return thread_list; }
#endif

/*
 * Implementation Notes:
 * Task Id of the task from VxWorks is going to be returned to the Create
 * Thread Call. The problem is for a call like GetCurrentThread (), we need
 * to make the TID back to the Opaque Handle, which means we need to add
 * the pointer to the PilThread as an added variable to the Task Control
 * Block. Adding this thread specific variable to the TCB adds overhead
 * to the task switching ( as this is considered as task context info) and
 * a few microseconds have to be wasted (WindRiver Programmer's Guide Page 51)
 * As such we will just use directly the Task Id, unless we land in problems
 *
 * For the PilThread_t structure, we will keep it since we need to pass it
 * to the Start_A_Thread function.
 */

int
Create_A_Thread ( PVOID  vp )
{
   uint32_t ret;
   PilThread_t* thread = (PilThread_t *) vp;

   // Just update our thread id
   thread->Tid = taskIdSelf () ;

   // set the thread specific information
   if (taskVarAdd (0, (int *)&pTaskData ) != OK )
      TRACE("TASKVARADD FALURE -- INSPECT")
        ; // this is needed in this line

   pTaskData = vp ; // set the malloced address so we can free upon exit

   if ( (thread->dwOptions & PIL_THREAD_SUSPENDED) == PIL_THREAD_SUSPENDED)
      taskSuspend ( taskIdSelf () ) ;

   // Some one has resumed us.. wake up ..

   // Now this task runs in the context
   ret = thread->StartAddress ( thread->ThreadParam ) ;
   PIL_ExitThread( ret );
   return ret;
}


///////////////////////////////////////////////////////////////////////
//
// PIL_CreateThread
//
//

PilThread
_PIL_CreateThread(
   PilThreadRoutine StartAddress,   // The function for the thread to exec
   DWORD            dwStackSize,   // The initial stack size or 0
   CHAR*            Name,         // The optional name for the thread
   DWORD            Priority,      // The initial priority of the thread
   PVOID            Param,         // An optional parameter for the thread
   DWORD            dwOptions      // Thread creation options
   )
{
   PilThread_t*  Thread ;
   int  Tid ;

/*
   Since our task  starts at 100, and we want all
   tasks to have the same priority so that vxworks will round robin them,
   default all priorities to 100 (for now)
 */
   if (dwStackSize < 30000)
   {
     dwStackSize = 30000;
   }
   Priority = 100;
   Thread = PIL_AllocateMemory ( sizeof (PilThread_t), PIL_MEM_ZERO_INIT ) ;
   if ( Thread )
   {
      Thread->StartAddress = StartAddress ;
      Thread->ThreadParam  = Param ;
      Thread->dwOptions    = dwOptions ;
      Thread->ExitEvent    = PIL_CreateEvent(TRUE,FALSE);

      Tid = taskSpawn ( Name, Priority, 
//#ifdef DEBUG
   0,
//#else
//   (VX_UNBREAKABLE & ~VX_DEALLOC_STACK),
//#endif
      dwStackSize, Create_A_Thread, (int)Thread, 0,0,0,0,0,0,0,0,0 );

#ifdef DEBUG
      _PIL_AddToList(&thread_list,Thread);
#endif
   }

   return (PilThread) Thread;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_DelayThread
//
//

PilResult
PIL_DelayThread(
   DWORD     dwDelay   // Delay in milliseconds to delay
   )
{
   DWORD ctTimeout ; // in clockTicks as close as possible

   // as before get in ticks
   ctTimeout = getTicks (dwDelay) ;

   // Delay execution
   return ( taskDelay ( ctTimeout ) == OK ? PIL_OK : PIL_ERROR ) ;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_DestroyThread
//
//

PilResult
PIL_DestroyThread(
   PilThread Thread,   // The handle of the thread to destroy
        PDWORD    pdwStatus
   )
{
   // Now we need to destroy the actual task specific data
   int Var;   // Thread specific data
   int BakVar ; // Just make sure we have a back up of the Var
   PilThread_t*  Task ; // Out own task specific data

   ASSERT(Thread) ;
   if ( !Thread )
      return PIL_INVALID_HANDLE ;

   Task = (PilThread_t *)Thread ;
   if ( (taskVarGet ( Task->Tid, &Var ) == ERROR) || !Var )
   {
      TRACE("taskVarGet Failure -- INSPECT");
      return PIL_ERROR; // Unable to get past this situation
   }

   BakVar = Var ;  // backup the variable
   taskVarDelete ( Task->Tid, &Var ) ; // Free up the taskVar

   PIL_DestroyEvent(Task->ExitEvent);
   free ((PVOID)BakVar) ;   // Free up the allocated memory for this task

   // Ignore the return status from taskDelete because sometimes the
   // taskTerminate might have called and terminated it..

   taskDelete (Task->Tid) ;

#ifdef DEBUG
   _PIL_RemoveFromList(&thread_list, Thread);
#endif

   return PIL_OK ;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_ExitThread
//
//

VOID
PIL_ExitThread(
   uint8_t dwResult
   )
{
// Since PIL_ExitThread runs in the context of the current thread,
// there is no necessity for us to obtain it using taskVarGet
// because, when the taskSwitch happens, automatically, pTaskData
// points to the current thread obeject

   PilThread_t*  Task ; // Out own task specific data

   Task = (PilThread_t *) pTaskData ; // Get the Pil_Thread strucuture
   Task->Status = dwResult ;
   PIL_SetEvent(Task->ExitEvent);

   exit (dwResult) ; // Exit code is dwResult.. Will never return
}

///////////////////////////////////////////////////////////////////////
//
// PIL_GetCurrentThread
//
//
// This function will have to map the TID to the corresponding
// Thread handle.. And the problem is VxWorks does not know it. So
// we need to store task specific data using the taskVatrAdd function
// during task initialization and retrieve the corresponding info
// back from VxWorks.
//

PilThread
PIL_GetCurrentThread(
   )
{
// Since PIL_GetCurrentThread runs in the context of the current
// thread, there is no necessity for us to obtain it using taskVarGet
// because, when the taskSwitch happens, automatically, pTaskData
// points to the current thread obeject
   return pTaskData ; 
}

// This exists for unix and need for vxworks as well
unsigned long PIL_GetCurrentThreadId()
{
    return (unsigned long)PIL_GetCurrentThread();
}

///////////////////////////////////////////////////////////////////////
//
// PIL_GetThreadPriority
//
//

PilResult
PIL_GetThreadPriority(
   PilThread Thread,   // The handle of the thread to retrieve priority
   PDWORD    pPriority
   )
{
   int Priority ; // VxWorks needs a pointer to int
   PilThread_t* Task = (PilThread_t *)Thread ;

   ASSERT(Thread) ;
   if ( !Thread || (taskPriorityGet (Task->Tid, &Priority) != OK) )
      return PIL_INVALID_HANDLE ;

   *pPriority = Priority ;

   return PIL_OK ;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_SetThreadPriority
//
//

PilResult
PIL_SetThreadPriority(
   PilThread Thread,      // The handle of the thread to change
   DWORD     dwPriority   // The new priority level
   )
{
   PilThread_t* Task = (PilThread_t *)Thread ;

   ASSERT(Thread) ;
   return(  Thread ? 
       (taskPrioritySet(Task->Tid, dwPriority) == OK ? PIL_OK : PIL_ERROR) :
            PIL_INVALID_HANDLE) ;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_SuspendThread
//
//

PilResult
PIL_SuspendThread(
   PilThread Thread   // The handle of the thread to suspend
   )
{
   PilThread_t* Task = (PilThread_t *)Thread ;

   ASSERT(Thread) ;
   return(  Thread ?
       (taskSuspend(Task->Tid) == OK ? PIL_OK : PIL_INVALID_HANDLE) :
            PIL_INVALID_HANDLE) ;
}


///////////////////////////////////////////////////////////////////////
//
// PIL_ResumeThread
//
//

PilResult
PIL_ResumeThread(
   PilThread Thread   // The handle of the thread to resume
   )
{
   PilThread_t* Task = (PilThread_t *)Thread ;

   ASSERT(Thread) ;
   return(  Thread ?
       (taskResume(Task->Tid) == OK ? PIL_OK : PIL_INVALID_HANDLE) :
            PIL_INVALID_HANDLE) ;  
}



///////////////////////////////////////////////////////////////////////
//
// PIL_TerminateThread
//
//

PilResult
PIL_TerminateThread(
   PilThread Thread
   )
{
  // explicit Termincation of thread should be avoided letting the code
  // of the thread to pass back exit codes.. In this case we set a 
  // special value  all 1's as the status..

  // Also the thread data strucuture maintained by us will have to be
  // freed upon exit, which has to be performed by PIL_DestroyThread

  // set the status that there was no code associated with the thread
  // status

   PilThread_t* Task = (PilThread_t *)Thread ;

   ASSERT(Thread) ;
   if (!Thread)
      return PIL_INVALID_HANDLE ;

   Task->Status = 0xFFFFFFFF; // Just indicate the status

   // setting the status will help tasks such as
   return taskDelete (Task->Tid) == OK ? PIL_OK : PIL_ERROR ;
}

////////////////////////////////////////////////////////////////////////
//
// PIL_WaitForThread
//
//

PilResult
PIL_WaitForThread(
    PilThread Thread,
    uint32_t dwTimeout
    )
{
    PilThread_t* Task = (PilThread_t *)Thread;

    ASSERT(Thread);

    if (!Thread)
        return PIL_INVALID_HANDLE;

    return PIL_WaitForEvent(Task->ExitEvent,dwTimeout);
}
