/////////////////////////////////////////////////////////////////////////
//                                                                     //
// Filename: PILTMR.C                                                  //
// Module:   Platform Isolation Layer Timer Routines                   //
// Author:   Daniel Baumberger                                         //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
//      Copyright (C) 1999 Intel Corporation  ALL RIGHTS RESERVED      //
//                                                                     //
//             INTEL CORPORATION PROPRIETARY INFORMATION               //
//                                                                     //
//     This software is supplied under the terms of a license agree-   //
//     ment or non-disclosure agreement with Intel Corporation         //
//     and may not be copied nor disclosed except in accordance        //
//     with the terms of that agreement.                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//                                                                     //
// PURPOSE: Implement PIL Timer Routines for VxWorks                   //
//                                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
//
// HISTORY:
//
// 
/////////////////////////////////////////////////////////////////////////

#include "pil_vx.h"
#include "syslib.h"
#include "time.h"

/* 
 * Implementation Notes:
 * If there are t ticks per 1000 msec(1 sec),
 * then ( x * t) / 1000 will be there for x millisec (x is the interval)
 * sysClktRateGet () will give t and x will be passed in as the parameter
 * We will add 1 to make sure that , there is atleast one 1 tick
 * PLEASE REMEMBER THAT THE CLOCK RATE MUST BE A MULTIPLE OF 1 msec to make
 * sure that there is no "residual effect" of the timer mechanism..
 */


int
getTicks (
       DWORD dwInterval )
{
   // INterval must be in msec.. We assume a 10 msec resolution and
   // have coded it here
   int ticks;

   ticks = ( sysClkRateGet() * dwInterval) / 1000 ;
   if (ticks <= 0 )  // guard.. does not cost for '<'
      ticks = 1 ;

   return ticks;
}

/*
 * The common call back routine is supposed to handle all the timer events
 * The onus of responsibility of making sure that the corresponding procedure
 * runs within the period of the timer handler lies with the programmer that
 * writes the TimerRoutine
 */

VOID
CallBack (
       PVOID p )
{
   PilTimer_t* Tmr = (PilTimer_t *)p ; // The function should not overrun 

   if (!Tmr)
      return ;

   if ( Tmr->Recur ) // Repeating timers.. Start watchdog again
   {
      if (wdStart ( Tmr->wdId, getTicks(Tmr->dwInterval), CallBack, (int)Tmr) != OK )
        // printf or VxWorks should automatically log it.
             ;
   }

   // Even if restart failed, make sure that atleast the current expiration 
   // runs atleast..

   if ( Tmr->ProcedureAddress )
      Tmr->ProcedureAddress (Tmr, Tmr->Data) ;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_CreateTimer
//
//

PilTimer
PIL_CreateTimer(
   PilTimerRoutine  TimerRoutine,   // The routine to call for the timer
   DWORD            dwInterval,   // Interval in ms before timer fires
   PVOID            Context,         // Context to pass to timer handler
   PilTimerType     Type             // SINGLE or REPEATING timer
   )
{
   PilTimer_t*  Handle;

   // Preconditions 

   if ( !TimerRoutine || (dwInterval <= 0) ||
        ( (Type != PIL_TIMER_SINGLE) && (Type != PIL_TIMER_REPEATING) )
      )
      return NULL ;

   // Allocate memory for the handle
   Handle = (PilTimer_t *) PIL_AllocateMemory( sizeof (PilTimer_t), PIL_MEM_ZERO_INIT ) ;
   if (  Handle )
   {
      Handle->Recur            = (Type == PIL_TIMER_REPEATING) ;
      Handle->dwInterval       =  dwInterval ;
      Handle->ProcedureAddress =  TimerRoutine ;
      Handle->Data             =  Context ;
   }

   return Handle ;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_DestroyTimer
//
//

PilResult 
PIL_DestroyTimer(
   PilTimer  Timer      // The handle of the timer to destroy
   )
{
   WDOG_ID wdId ;

   if ( !Timer )
      return PIL_INVALID_HANDLE ;

   wdId = ((PilTimer_t *)Timer)->wdId ;
   PIL_FreeMemory (Timer) ;  // Dont check back the code

   // If after deleting the wdTimer, we did a Free memory, returning
   // error becomes a minor tech issue.. So we just focus on the 
   // wdDelete function

   // Timer has been freed already intentionally
   return ( wdDelete (wdId) == OK ? PIL_OK : PIL_ERROR ) ;   
}

///////////////////////////////////////////////////////////////////////
//
// PIL_ResetTimer
//
//

PilResult
PIL_ResetTimer(
   PilTimer        Handle,          // The handle of the timer to reset
   PilTimerRoutine TimerRoutine,   // The routine to call on timer expr.
   DWORD           dwInterval,       // The interval before expiration
   PVOID           Context,        // Pointer to the parameter
   PilTimerType    Type             // SINGLE or REPEATING timer
   )
{
   PilTimer_t* Timer = (PilTimer_t *) Handle ;
  // cancel the existing timer and restart a new one
   if ( !Timer )
      return PIL_INVALID_HANDLE ;

   if ( !dwInterval || !TimerRoutine )
     return PIL_INVALID_PARAM ;

   // The new wdStart resets the previous timer for this one

   if ( wdStart ( Timer->wdId, getTicks(dwInterval), CallBack, (int)Timer) != OK )
      return PIL_ERROR ;
   else
      return PIL_OK ;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_StartTimer
//
//

PilResult 
PIL_StartTimer(
   PilTimer  Handle   // The handle of the timer to destroy
   )
{
   PilTimer_t* Timer  = (PilTimer_t *)Handle ;

   if ( !Timer )
      return PIL_INVALID_HANDLE ;

   // Schedule the timer
   return wdStart ( Timer->wdId,
                    getTicks(Timer->dwInterval),
                    CallBack,
                    (int)Timer
                   ) != OK ? PIL_OK : PIL_ERROR ;
}

///////////////////////////////////////////////////////////////////////
//
// PIL_StopTimer
//
//

PilResult 
PIL_StopTimer(
   PilTimer  Timer      // The handle of the timer to destroy
   )
{
   if (!Timer)
      return PIL_INVALID_HANDLE ;

   return ( wdCancel ( ((PilTimer_t *)Timer)->wdId ) == OK ? PIL_OK : PIL_ERROR );
}


///////////////////////////////////////////////////////////////////////
//
// PIL_GetCurrentTime
//
// Returns a 64-bit time value in milliseconds.  Note that this may not
// be the actual clock time in milliseconds but instead may be the
// number of milliseconds since system start.  It is guaranteed that
// the difference between any two calls of this function is the elapsed
// real time in milliseconds.
//

uint64_t PIL_GetCurrentTime()
{
    time_t	now;

    time(&now);
    return (( (unsigned long) now)*1000 );

}

