/**
 *============================================================================
 * = COPYRIGHT 
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or 
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      FP Module Core Component - CCI - init function
 *
 * = FILENAME
 *      ix_cc_fpm.c
 *
 * = DESCRIPTION
 *      The file contains the main functions of the FPModule ingress core 
 *      component. It contains the init() and fini() functions, message and
 *      packet handler routines, and also the routines that handle packets in
 *      the inbound and outbound directions.
 *
 * = AUTHOR
 *       Rajeev Muralidhar
 *       Rajeev.D.Muralidahr@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      06/24/2002
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */

/*#define IX_FILE_IDENT "$Id: ix_cc_fpm.c,v 1.6.2.1 2004/03/24 11:39:52 smds Exp $";*/

#include "cc/ix_cc_fpm_core.h"
#include "ix_cc.h"

#if !defined(IX_INCLUDE_REGISTRY) /* configuration values header file */
#include "ix_config.h"
#endif

#define REPOSITORY_PATH_LOCAL_BLADE_ID "System_Properties/BLADE_ID"

/* Pointer to FPM context. This will be allocated during init() of the CC */
ix_cc_fpm_context* g_pFPMContext;

/* Instead of using hashdefine for IX Buffer size, ix_rm_buffer_free_list_get_info function is called and the IX buffer size is copied from the returned structure. */
/* #define MAX_IX_BUFFER_SIZE 1024 */

/* local functions */
static void ix_cc_fpm_dump_context (ix_cc_fpm_context* context);
static ix_error 
ix_cc_fpm_translate_ix_buffer (
			       ix_hw_buffer_meta* arg_pMeta,
			       ix_buffer_handle* arg_pBufferHandle,
			       ix_buffer_handle* arg_pNextBufferHandle,
			       char** arg_pBuffer,
			       ix_uint32* arg_pIsEOP
			       );

/**
 * NAME: ix_cc_fpm_init
 *
 * DESCRIPTION: This primitive shall be called and returned successfully before 
 *              requesting any service from FPModule Core component. This 
 *              primitive should be called only once, although no explicit logic
 *              is in place for this. The behavior is undefined if this is 
 *              called more than once.
 *              This function performs all initializations required by a core 
 *              component (registering msg/pkt handlers).
 *              Also, this does not have any micro block component, so 
 *              symbol patching is not done. 
 * 
 * @Param:  - IN arg_hCcHandle - handle to FP Module core component, created by 
 *            core component infrastructure; this shall be used later to get 
 *            other services from core component infrastructure.
 *
 * @Param:  - OUT arg_ppContext - location where the pointer to the control 
 *            block allocated by FP Module core component will be stored. 
 *            The control block is internal to the FPM core component and it 
 *            contains information about internal data structures, allocated 
 *            memory and other relevant information. Later this pointer to be 
 *            passed into ix_cc_fpm_fini function for freeing memory and other 
 *            house keeping operations when FPM component is being destroyed.  
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_OOM   - memory allocation failure      
 * IX_CC_FPM_ERROR_CCI - failure from CCI 
 * 
 */

ix_error ix_cc_fpm_init(
			ix_cc_handle arg_hCcHandle,
			void** arg_ppContext
                        )
{
    ix_error err = IX_SUCCESS;
    ix_cc_init_context* init_context = (ix_cc_init_context*) *arg_ppContext;
    ix_cc_properties_node *pPropNode;

    FPM_CC_DBG_TRACE();
    if (init_context == NULL) 
      return IX_ERROR_WARNING (IX_CC_ERROR_NULL, ("FP Module CC init called"
						    " with NULL context"));    
    
    /* allocate memory for FP module control structure */
    g_pFPMContext = ix_ossl_malloc (sizeof (ix_cc_fpm_context));
    if (g_pFPMContext == NULL)
	return IX_ERROR_WARNING (IX_CC_ERROR_OOM_SYSTEM, 
				 ("Could not allocate"
				  "memory for FPM ctx"));
    /* initialize the structure */
    ix_ossl_memset(g_pFPMContext,0,sizeof(ix_cc_fpm_context));

    g_pFPMContext->fpm_hFPMHandle = arg_hCcHandle;
    g_pFPMContext->fpm_hFreeList  = init_context->hFreeList;
    g_pFPMContext->fpm_linkState_cb = NULL;
    g_pFPMContext->fpm_pktHdlr_cb = NULL;
    g_pFPMContext->fpm_ipv4Exception_cb = NULL;    

    /**
     * Retrieve local blade property information from init context
     * (This will work for both level-1 and level-0 modes)
     */
    pPropNode = init_context->pPropertyList;
    /* Get blade Id from first node */
    if(pPropNode != NULL)
    {
        g_pFPMContext->fpm_localBladeId  = pPropNode->pPropertyInfo->blade_id;
    }/* end if */

    
/*
  In a system where CCI is not used, we have to register the message/packet 
  handlers with the Resource Manager. If CCI is available, we will register
  these handlers with CCI (Core Component Infrastructure)
 */
#if defined(IX_EXCLUDE_CCI)
    FPM_CC_DBG_PRINT ("IN %s %d \n", __FUNCTION__, __LINE__);
    IX_ERROR_CRT (ix_rm_message_handler_register (IX_CC_FP_MODULE_MSG_INPUT,
						  ix_cc_fpm_msg_handler,
						  g_pFPMContext),


		  IX_CC_FPM_ERROR_ADD_HANDLER,
		  IX_ERROR_LEVEL_WARNING);
    IX_ERROR_CRT (ix_rm_packet_handler_register (IX_CC_FP_MODULE_PKT_INPUT,
						 ix_cc_fpm_pkt_handler,
						 g_pFPMContext),
		  IX_CC_FPM_ERROR_ADD_HANDLER,
		  IX_ERROR_LEVEL_WARNING);
#else
    FPM_CC_DBG_PRINT ("IN %s %d \n", __FUNCTION__, __LINE__);
    IX_ERROR_CGT (ix_cci_cc_add_message_handler (g_pFPMContext->fpm_hFPMHandle,
						 IX_CC_FP_MODULE_MSG_INPUT,
						 ix_cc_fpm_msg_handler,
						 IX_INPUT_TYPE_SINGLE_SRC),
		  err,
		  IX_CC_FPM_ERROR_ADD_HANDLER,
		  IX_ERROR_LEVEL_WARNING, label_1);
    
    /* register the packet handler to receive packets destined for CP 
       from stack driver */
    IX_ERROR_CGT (ix_cci_cc_add_packet_handler (g_pFPMContext->fpm_hFPMHandle,
						IX_CC_FP_MODULE_PKT_INPUT,
						(ix_pkt_handler) 
						ix_cc_fpm_pkt_handler,
						IX_INPUT_TYPE_SINGLE_SRC),
		  err,
		  IX_CC_FPM_ERROR_ADD_HANDLER,
		  IX_ERROR_LEVEL_LOCAL, label_1);
#endif
    /* store the FPM context in the context passed in */
    *arg_ppContext = (void*) g_pFPMContext;
    ix_cc_fpm_dump_context (g_pFPMContext);    

    FPM_CC_DBG_PRINT ("Leaving %s %d \n", __FUNCTION__, __LINE__);
    return IX_SUCCESS; label_1:
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_ossl_message_log("FPM CORE: %ld\n", (ix_uint32)err);
#else
    ix_error_dump (stdout, err);
#endif
    return err;
} /* end ix_cc_fpm_init */


/**
 * NAME: ix_cc_fpm_fini
 *
 * DESCRIPTION: This primitive shall be called to terminate services from 
 *              FP Module core component. 
 * 
 * @Param:  - IN arg_hCcHandle - handle to FP Module core component
 *
 * @Param:  - IN arg_pContext - pointer to the control block memory allocated 
 * earlier in ix_cc_fpm_init function. 
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_FPM_ERROR_INVALID_INPUT_PARAM - invalid input parameter(s)
 * IX_CC_ERROR_OOM - failure of memory deletion
 * IX_CC_FPM_ERROR_CCI - failure from CCI
 * 
 */

ix_error ix_cc_fpm_fini (
			 ix_cc_handle arg_hCcHandle,
			 void* arg_pContext
			 )
{
    if (arg_pContext == NULL)
	return IX_ERROR_WARNING (IX_CC_ERROR_NULL, ("FP Module CC fini called"
						    " with NULL context"));
    g_pFPMContext = (ix_cc_fpm_context*) arg_pContext;
    ix_cc_fpm_dump_context (g_pFPMContext);    
    
    g_pFPMContext->fpm_linkState_cb = NULL;
    g_pFPMContext->fpm_pktHdlr_cb = NULL;
    g_pFPMContext->fpm_ipv4Exception_cb = NULL;

#if defined(IX_EXCLUDE_CCI)
    IX_ERROR_CRT(ix_rm_message_handler_unregister(IX_CC_FP_MODULE_MSG_INPUT),
		 IX_CC_FPM_ERROR_CCI,
		 IX_ERROR_LEVEL_WARNING);
#else
    IX_ERROR_CRT(ix_cci_cc_remove_message_handler(IX_CC_FP_MODULE_MSG_INPUT),
		 IX_CC_FPM_ERROR_CCI,
		 IX_ERROR_LEVEL_WARNING);
#endif    
    ix_ossl_free ((ix_cc_fpm_context*) arg_pContext);    
    return IX_SUCCESS;
} /* end ix_cc_fpm_fini */


/*****************************************************************************
 *      MESSAGE HANDLER, GENERAL CALLBACK FUNCTION
 ***************************************************************************/
/*
 * NAME: ix_cc_fpm_msg_handler
 *
 * DESCRIPTION: This primitive is the message handler function for FP Module
 *              core component. FP Module ingress core component receives 
 *              messages from its egress counterpart through this message 
 *              handler function. 
 * 
 * @Param:  - IN arg_hDataToken - buffer handle embedding information for the 
 *            message passed in arg_UserData.
 * @Param:  - IN ix_uint32 arg_UserData - defines the type of message; this is
 *            defined by each core component.
 * @Param:  - IN void* arg_pContext - pointer to FP Module core component, 
 *            defined & initialized by FPM ingress CC during init().
 *
 * @Param:  - OUT - None
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 * IX_CC_ERROR_NULL - null argument
 * IX_CC_FPM_ERROR_INVALID_CONTEXT - context is unusable
 * IX_CC_ERROR_UNDEFINED_MSG - un supported message
 * IX_CC_FPM_ERROR_MSG_LIBRARY - error from message support library
 * 
 */
ix_error ix_cc_fpm_msg_handler (
				ix_buffer_handle arg_hDataToken,
				ix_uint32 arg_UserData,
				void* arg_pContext
				)
{
    ix_error                          err = IX_SUCCESS;
    void                              *pContext = NULL;
    void                              *pMsg = NULL;
    ix_cc_properties_msg              *pPropMsg;
    ix_cc_properties                  *pProperty;
    ix_uint32                         port_id = 0;
    ix_cc_link_status                 link_status;    

    FPM_CC_DBG_TRACE();
    /* Validate arg_pContext */
    if (arg_pContext == NULL) {
	return IX_ERROR_WARNING (IX_CC_ERROR_NULL,
				 ("arg_pContext is NULL"));
    }/* end if(arg_pContext == NULL) */
    
    if ((arg_UserData < IX_CC_COMMON_MSG_ID_FIRST) ||
	(arg_UserData >= IX_CC_FPM_MSG_LAST)) {
	
        return IX_ERROR_WARNING (IX_CC_ERROR_UNDEFINED_MSG,
				 ("Invalid Message Type"));
    }/* end if */    
    IX_MSUP_EXTRACT_MSG (arg_hDataToken,&pMsg,&pContext);

#ifdef IX_CC_ETH    
    if (arg_UserData != IX_CC_FPM_MSG_PURGE_ARP_TABLE) {
	if (pMsg == NULL) {
	    return IX_ERROR_WARNING (IX_CC_ERROR_NULL,
				     ("Invalid Buffer"));
	}/* end if(pContext == NULL) */
    }
#endif
    
    FPM_CC_DBG_PRINT ("%s %d\n", __FUNCTION__, __LINE__);
    /* Check for message type */
    switch (arg_UserData) {
	/* Handle Property update message.
	   Link status event is delivered as a property update 
	*/
    case IX_CC_COMMON_MSG_ID_PROP_UPDATE:
	{
	  ix_cc_fpm_link_status linkstatus;
	  pPropMsg = (ix_cc_properties_msg*) pMsg;
	  pProperty = &pPropMsg->propData;
	  
	  FPM_CC_DBG_PRINT ("%s %d\n", __FUNCTION__, __LINE__);

	  
	  /* Get port Id from the message */
	  port_id = pProperty->port_id;
	  link_status = pProperty->link_state;
	  
	  if (link_status == IX_CC_LINK_STATUS_DOWN) {
	    FPM_CC_DBG_PRINT ("Recvd link down for Port %ld \n", (ix_uint32)port_id);
	    linkstatus = IX_CC_FPM_LINK_DOWN;
	  }
	  else if (link_status == IX_CC_LINK_STATUS_UP) {
	    FPM_CC_DBG_PRINT ("Recvd link up for Port %ld \n", (ix_uint32)port_id);
	    linkstatus = IX_CC_FPM_LINK_UP;
	  }
	    /* invoke callback into FP Module delivering the link status
	       event 
	    */
	    if (g_pFPMContext->fpm_linkState_cb != NULL) {
	      FPM_CC_DBG_PRINT ("Invoking link status callback!\n");
	      g_pFPMContext->fpm_linkState_cb (g_pFPMContext->fpm_localBladeId,
					       port_id,
					       linkstatus);
	    }
	    break;
	}
    case IX_NEW_DIRECT_HOST_MSG_ID:
      {
	ix_cc_fpm_direct_host_exception_data* ipv4_event_data = 
	  (ix_cc_fpm_direct_host_exception_data*) pMsg;
	
	FPM_CC_DBG_PRINT ("IPv4 exception (Direct Host) received"
			  "Port Id = %ld \t Dest IP = 0x%x \n",
			  (ix_uint32)ipv4_event_data->port_id, ipv4_event_data->dest_ip);
	/* invoke callback into FP Module delivering the IPv4 exception
	   for Directly connected host
	*/
	if (g_pFPMContext->fpm_ipv4Exception_cb != NULL) {
	  FPM_CC_DBG_PRINT ("Invoking Ipv4 exception callback!\n");
	  g_pFPMContext->fpm_ipv4Exception_cb (g_pFPMContext->fpm_localBladeId,
					       ipv4_event_data->port_id,
					       ipv4_event_data->dest_ip);
	}
      }
      break;
    default:
	err = IX_ERROR_WARNING (IX_CC_ERROR_UNDEFINED_MSG,
				("Msg handler invoked with invalid msg ID &d", 
				 arg_UserData));
	break;
    }/* end switch(arg_UserData) */
    
    IX_ERROR_CRT (ix_rm_buffer_free_chain (arg_hDataToken),
		  IX_CC_FPM_ERROR_BUFFER_FREE, 0); 
    return err;
} 


/* end ix_cc_fpm_msg_handler */
/**
 * NAME: ix_cc_fpm_pkt_handler
 *
 * DESCRIPTION: This primitive is the packet handler function for FP Module
 *              core component. FP Module ingress core component receives 
 *              packets destined for remote control plane (protocol packets)
 *              from the stack driver; these packets will then be tunneled
 *              to the control plane.
 * 
 * @Param:  - IN ix_buffer_handle arg_hBuffer - buffer handle containing
 *                                                 the packet
 * @Param:  - IN ix_uint32 arg_ExceptionCode - Exception code for the packet.
 * @Param:  - IN void* arg_ComponentContext - pointer to FPModule core component
 *               context. This context was defined by the core component and 
 *               passed to CCI through ix_cc_fpm_init function.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 */
ix_error ix_cc_fpm_pkt_handler (
				ix_buffer_handle arg_hBuffer,
				ix_uint32 arg_ExceptionCode,
				void* arg_pContext
				)
{
    ix_error err = IX_SUCCESS;
    ix_buffer_handle curr_buf_hdl = arg_hBuffer;
    ix_buffer_handle next_buf_hdl = IX_NULL_BUFFER_HANDLE;
    ix_uint32 port_id, pkt_len, buffer_len;
    ix_hw_buffer_meta* pMeta;
    ix_uint32 isEOP = 0;
    char** buffer = NULL, *packet = NULL;

/* 
    FPM_CC_DBG_TRACE ();
*/

    if (arg_pContext == NULL)
      
      return IX_ERROR_WARNING (IX_CC_ERROR_NULL, 
			       ("Packet Handler invoked with NULL context"));
    if (arg_hBuffer == IX_NULL_BUFFER_HANDLE)
	return IX_ERROR_WARNING (IX_CC_ERROR_NULL,
				 ("Null buffer handle in Pkt Handler"));
    
    /*
      Extract the input port ID & packet size from the meta-data of 
      arg_hBuffer
      Extract buffer length also (length of this buffer)
    */
    
    IX_ERROR_CGT (ix_rm_buffer_get_meta(arg_hBuffer, (void**)&pMeta), err,
		  IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING, label);
    port_id = (ix_uint32) IX_NTOH16 (IX_RM_MEM_UINT16_READ 
				     (&pMeta->m_InputPort));
    pkt_len = (ix_uint32) IX_NTOH16 (IX_RM_MEM_UINT16_READ 
				     (&pMeta->m_PacketSize));
    buffer_len = (ix_uint32) IX_NTOH16 (IX_RM_MEM_UINT16_READ 
					(&pMeta->m_BufferSize));

    /* allocate memory for the whole packet */
    buffer = (char**) ix_ossl_malloc (sizeof (char*));
    if (buffer == NULL) {
      err= IX_ERROR_WARNING (IX_CC_ERROR_OOM_SYSTEM,
			       ("Unable to allocate memory for packet !\n"));
      goto label;
    }
    *buffer = (char*) ix_ossl_malloc (pkt_len);
    if (*buffer == NULL) {
      err= IX_ERROR_WARNING (IX_CC_ERROR_OOM_SYSTEM,
			       ("Unable to allocate memory for packet !\n"));
      goto label;
    }
    packet = *buffer;

/* 
    FPM_CC_DBG_PRINT ("********Recvd packet from stack driver!***********\n");
    FPM_CC_DBG_PRINT ("Port id = %d \t Packet size = %d Buffer size = %d \n", 
		      port_id, pkt_len, buffer_len);

*/
    /* Translate the ix_buffer into a VxWorks buffer. */
    IX_ERROR_CG(ix_cc_fpm_translate_ix_buffer (pMeta,
					       &curr_buf_hdl, 
					       &next_buf_hdl,
					       buffer,
					       &isEOP),
		err, label);
    
    /* Handle buffer chains. */
    while(!isEOP)
      {
	FPM_CC_DBG_PRINT ("Getting next buffer in chain\n");
	*buffer += buffer_len;
	curr_buf_hdl = next_buf_hdl;
	next_buf_hdl = IX_NULL_BUFFER_HANDLE;
	IX_ERROR_CGT(ix_rm_buffer_get_meta(curr_buf_hdl, 
					   (void**)&pMeta),
		     err, IX_CC_ERROR_INTERNAL, 
		     IX_ERROR_LEVEL_WARNING, label);
	buffer_len = (ix_uint32) IX_NTOH16 (IX_RM_MEM_UINT16_READ 
					    (&pMeta->m_BufferSize));
	/* Translate the ix_buffer into a char buffer. */
	IX_ERROR_CG(ix_cc_fpm_translate_ix_buffer (pMeta, &curr_buf_hdl, &
						   next_buf_hdl, buffer,
						   &isEOP),
		    err, label);
      }
    
/* 
    FPM_CC_DBG_PRINT ("Finished with buffer chains"
		      "Packet size = %d Buffer size = %d \n", 
		      pkt_len, buffer_len);
*/

    /* Call the registered callback into FPModule */
    if (g_pFPMContext->fpm_pktHdlr_cb != NULL) {
      g_pFPMContext->fpm_pktHdlr_cb (g_pFPMContext->fpm_localBladeId,
				     port_id,
				     pkt_len,
				     packet);
    }
/*  Free the Allocated Memory. ph_write_packet should have 
    copied the packet in his buffer 
*/

 label:
    IX_ERROR_CRT (ix_rm_buffer_free_chain (arg_hBuffer),
		  IX_CC_FPM_ERROR_BUFFER_FREE, 0); 

    arg_hBuffer = IX_NULL_BUFFER_HANDLE;

    if (buffer) {
      if (*buffer)
        ix_ossl_free (*buffer);
      ix_ossl_free (buffer);
    }
    return err;
}


/**
 * NAME: _ix_cc_fpm_translate_ix_buffer
 *
 * DESCRIPTION: Translates an ix_buffer into a char buffer
 * 
 * @Param:  - IN 
 *
 * @Return: IX_SUCCESS
 *          IX_CC_ERROR_INTERNAL
 *          IX_ERROR_OOM_SYSTEM
 */
ix_error ix_cc_fpm_translate_ix_buffer(
				       ix_hw_buffer_meta* arg_pMeta,
				       ix_buffer_handle* arg_pBufferHandle,
				       ix_buffer_handle* arg_pNextBufferHandle,
				       char** arg_pBuffer,
				       ix_uint32* arg_pIsEOP
				       )
{
  ix_error err = IX_SUCCESS;
  void* pData;
  
  ix_uint16 bufferLength = IX_NTOH16 (IX_RM_MEM_UINT16_READ(&arg_pMeta->m_BufferSize));


  IX_ERROR_CRT (ix_cc_hw_get_packet_data (*arg_pBufferHandle, &pData),
		IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
  
  /* copy the buffer */
  ix_ossl_memcpy (*arg_pBuffer, pData, bufferLength);

  IX_ERROR_CRT (ix_rm_buffer_get_next (*arg_pBufferHandle, 
				       arg_pNextBufferHandle),
		IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
  
  IX_ERROR_CRT (ix_rm_buffer_is_eop(*arg_pBufferHandle, arg_pIsEOP),
		IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
  
  /* We dont need to free the ix_buffer here, since we will call 
     buffer_free_chain from the packet handler function
     
     err = ix_rm_buffer_free (*arg_pBufferHandle);
  */  

  *arg_pBufferHandle = IX_NULL_BUFFER_HANDLE;
  if(err != IX_SUCCESS) {
    err = IX_ERROR_CHAIN(err, IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING,
			 ("RM could not free the buffer with handle 0x%x\n",
			  *arg_pBufferHandle));
    return err;
  }
  return IX_SUCCESS;
}


/**
 * NAME: ix_cc_fpm_sync_send_packet
 *
 * DESCRIPTION: This function is invoked by FP Module (or the tunnel module)
 *              to send packets from the remote CP out of the system.
 *              These packets will be converted to ix_buffer and then
 *              handed off to the stack driver.
 * 
 * @Param:  - IN ix_uint32 arg_PortId - Port on which packet is to be sent
 * @Param:  - IN char* arg_pBuffer - Char buffer containing the packet
 * @Param:  - IN ix_uint32 arg_PacketLen - Packet length
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 * Error Codes:
 * 
 */
ix_error ix_cc_fpm_sync_send_packet (
				     ix_uint32 arg_PortID,
				     char*     arg_pBuffer,
				     ix_uint32 arg_PacketLen
				     )
{
  ix_error err = IX_SUCCESS;
  ix_buffer_handle* buffer_hdls;
  ix_uint32 num_ix_bufs = 1; /* the number of ix_buffers needed for this packet */
  ix_uint32 i, k, buf_chaining = 0, num_bytes_in_buffer = 0;
  ix_uint32 num_bytes_to_copy = 0;
  ix_uint32 num_bytes_copied = 0, user_data = 0;
  ix_uint32 ix_buf_size;
  ix_buffer_free_list_info flInfo;
  ix_hw_buffer_meta* buffer_meta;
  void* buffer_data;
  char* bufptr; /* pointer to packet, will be used for copying into 
		   multiple ix_buffers */
  ix_uint32 cell_count = 0;

/*
  FPM_CC_DBG_TRACE ();

  FPM_CC_DBG_PRINT ("Received packet of len %d for port %d \n", arg_PacketLen,
		    arg_PortID);
*/

  /* Find ix_buf_size */
  err = ix_rm_buffer_free_list_get_info(g_pFPMContext->fpm_hFreeList, &flInfo);
  /* check error value of this function */
  if (err != IX_SUCCESS) {
    return IX_ERROR_WARNING (IX_CC_ERROR_INTERNAL,
                            ("Unable to get ix_buffer size"));
  }
  ix_buf_size = flInfo.m_DataElementSize;

  num_ix_bufs = ((arg_PacketLen + 127) / ix_buf_size) + 1;
  if (num_ix_bufs > 1)
    buf_chaining = 1;
 

/*
  FPM_CC_DBG_PRINT ("Number of ix_buffers needed = %d \n", num_ix_bufs);
*/
  
  bufptr = arg_pBuffer;

  /* allocate the ix_buffer_handles */
  buffer_hdls = (ix_buffer_handle*) ix_ossl_malloc (num_ix_bufs * sizeof 
						    (ix_buffer_handle));
  if (buffer_hdls == NULL)
    return IX_ERROR_WARNING (IX_CC_ERROR_OOM_SYSTEM, 
			     ("Could not allocate ix_buffer handles for packet"));

/*
  FPM_CC_DBG_PRINT ("Allocated %d buffer handles!\n", num_ix_bufs);
*/

  for (i = 0; i < num_ix_bufs; i++) {
    /* allocate the ix_buffers using the buffer free list pointer got during
       initialization */
    buffer_hdls[i] = IX_NULL_BUFFER_HANDLE;
    err = ix_rm_buffer_alloc (g_pFPMContext->fpm_hFreeList,
			      &(buffer_hdls[i]));
    if (err != IX_SUCCESS) {
      FPM_CC_DBG_PRINT ("Unable to allocate ix_buffers!\n");
      for (k = 0; k < i; k++)
      {
         if (buffer_hdls[k])
            ix_rm_buffer_free (buffer_hdls[k]);
      }
      ix_ossl_free (buffer_hdls);
      return IX_ERROR_WARNING (IX_CC_ERROR_OOM_SYSTEM, 
			      ("Could not allocate ix_buffer for the packet"));
    }
  }
/*

  FPM_CC_DBG_PRINT ("Allocated %d buffers !\n", num_ix_bufs);
*/
  
  /* create the meta data information for the first buffer */

  err = ix_rm_buffer_get_meta (buffer_hdls[0], (void**)&buffer_meta);
  if (err != IX_SUCCESS)
    return IX_ERROR_WARNING (IX_CC_ERROR_INTERNAL, 
			     ("Unable to get packet metadata"));

  buffer_meta->m_OutputPort = arg_PortID;
  buffer_meta->m_PacketSize = IX_HTON16 (arg_PacketLen);
  /* set the offset to 128 for the first buffer in the chain ALWAYS */
  buffer_meta->m_Offset = 128;
  

  /* get a handle to the buffer's packet part */
  err = ix_cc_hw_get_packet_data (buffer_hdls[0], &buffer_data);
  if (err != IX_SUCCESS)
    return IX_ERROR_WARNING (IX_CC_ERROR_INTERNAL, 
			     ("Unable to get packet data"));

  /* copy the packet data into the ix_buffer - for this first buffer, we can
     copy only upto (ix_buf_size - 128) bytes of the original packet since 
     the first buffer's offset is set to 128
   */  
  if (arg_PacketLen < ix_buf_size - 128) {
    num_bytes_in_buffer = arg_PacketLen;
    num_bytes_copied = arg_PacketLen;
    num_bytes_to_copy = 0;
  }
  else {
    num_bytes_in_buffer = ix_buf_size - 128;
    num_bytes_copied = num_bytes_in_buffer;
    num_bytes_to_copy = arg_PacketLen - num_bytes_copied;
  }

/*
  FPM_CC_DBG_PRINT ("Bytes to copy = %d!\n", num_bytes_in_buffer);  
*/
  /* fill up the BufferSize metadata now */
  buffer_meta->m_BufferSize = num_bytes_in_buffer; 
  ix_ossl_memcpy (buffer_data, arg_pBuffer,
		  num_bytes_in_buffer);

  /* TO DO : For a single buffer, what about next buffer pointer ? */

  /* advance the bufptr by the amount of bytes just copied */

  bufptr += num_bytes_in_buffer;
/*
  FPM_CC_DBG_PRINT ("Copied %d bytes into first ix_buffer!\n",
		    num_bytes_in_buffer);
*/

 /* Calculate cell count for this buffer */
  cell_count = ix_cc_calculate_cell_count(num_bytes_in_buffer, 1);

 /* Write the cell count info to buffer handle */
 ix_rm_buffer_cell_count_set(&buffer_hdls[0], cell_count); 

  if (buf_chaining == 1) {
    for (i = 1; i < num_ix_bufs; i++) {

      FPM_CC_DBG_PRINT ("Copying buffer number : %ld\n", (ix_uint32)i);

      err = ix_rm_buffer_get_meta (buffer_hdls[i], (void**)&buffer_meta);
      if (err != IX_SUCCESS) {
          for (k = 0; k < num_ix_bufs; k++)
              if (buffer_hdls[k])
                  ix_rm_buffer_free (buffer_hdls[k]);
	  return IX_ERROR_WARNING (IX_CC_ERROR_INTERNAL, 
				 ("Unable to get packet metadata"));
      }
      buffer_meta->m_OutputPort = arg_PortID;
      buffer_meta->m_PacketSize = IX_HTON16 (arg_PacketLen);
      /* now set the offset to 0 */
      buffer_meta->m_Offset = 0;
      /* get a handle to the buffer's packet part */
      err = ix_cc_hw_get_packet_data (buffer_hdls[i], &buffer_data);
      if (err != IX_SUCCESS) {
          for (k = 0; k < num_ix_bufs; k++)
              if (buffer_hdls[k])
                  ix_rm_buffer_free (buffer_hdls[k]);
	  return IX_ERROR_WARNING (IX_CC_ERROR_INTERNAL, 
				 ("Unable to get packet metadata"));
      }
      
      if (num_bytes_to_copy < ix_buf_size) {
	num_bytes_in_buffer = num_bytes_to_copy;
	num_bytes_to_copy = 0;
      }
      else {
	num_bytes_in_buffer = ix_buf_size;
	num_bytes_copied += num_bytes_in_buffer;
	num_bytes_to_copy = arg_PacketLen - num_bytes_copied;
      }
      FPM_CC_DBG_PRINT ("Num bytes in buffer = %ld \n Num bytes copied = %ld \n"
			"Num bytes to copy = %ld \n", (ix_uint32)num_bytes_in_buffer,
			(ix_uint32)num_bytes_copied, (ix_uint32)num_bytes_to_copy);

      buffer_meta->m_BufferSize = num_bytes_in_buffer;
      ix_ossl_memcpy (buffer_data, bufptr, num_bytes_in_buffer);
      bufptr += num_bytes_in_buffer;
     /* Calculate cell count for this buffer */
     cell_count = ix_cc_calculate_cell_count(num_bytes_in_buffer, 0);

     /* Write the cell count info to buffer handle */
     ix_rm_buffer_cell_count_set(&buffer_hdls[i], cell_count); 

      /* TO DO: if this is the last buffer in the chain, set EOP bit */
      
	
      /* now chain the ix_buffers */
      err = ix_rm_buffer_link (&buffer_hdls[i-1], &buffer_hdls[i]);
      if (err != IX_SUCCESS) {
          for (k = 0; k < num_ix_bufs; k++)
              if (buffer_hdls[k])
                  ix_rm_buffer_free (buffer_hdls[k]);
          return IX_ERROR_WARNING (IX_CC_ERROR_INTERNAL, 
			       ("Unable to get packet metadata"));
      }
    }
  }

  /* TO DO: now the ix_buffer_handle can be handed off to stack driver */
  
  err = ix_rm_message_send (IX_CC_STKDRV_FPM_INPUT,
			    buffer_hdls[0],
			    user_data);

/* First of all we should not have allocated the memory 
   for buffer handle. Even If we did we have to clean it by our self. 
   no other function we free this. */

  ix_ossl_free (buffer_hdls);    


  if (err != IX_SUCCESS) {
    return IX_ERROR_WARNING (IX_CC_FPM_ERROR_PACKET_SEND,
			     ("Unable to send packet"));
  }
  return IX_SUCCESS;
}

/**
 * NAME: ix_cc_fpm_register_link_status_cb
 *
 * DESCRIPTION: This function is called by the FP Module's platform
 *              specific component to receive link status upates.
 *  
 * @Param:  - IN  arg_LinkState_cb: Link state callback function
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 *          Return code is not interpreted by the core component.
 * Error Codes:
 * 
 * IX_CC_FPM_ERROR_CALLBACK - user callback returns error message
 * 
 */
ix_error ix_cc_fpm_register_link_status_cb (ix_cc_fpm_link_status_cb 
					    arg_LinkState_cb)
{
    if (arg_LinkState_cb != NULL) {
	FPM_CC_DBG_PRINT ("Registering callback for Link state events\n");
	/* store callback in a global function pointer */
	g_pFPMContext->fpm_linkState_cb = arg_LinkState_cb;
	return IX_SUCCESS;
    }
    
    return IX_CC_FPM_ERROR_INVALID_INPUT_PARAM;
}

/**
 * NAME: ix_cc_fpm_deregister_link_status_cb
 *
 * DESCRIPTION: This function is called by the FP Module's platform
 *              specific component to deregister its callback
 *  
 * @Param:  - IN  - None
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 *          Return code is not interpreted by the core component.
 * Error Codes:
 * 
 * IX_CC_FPM_ERROR_CALLBACK - user callback returns error message
 * 
 */
ix_error ix_cc_fpm_deregister_link_status_cb ()
{
    g_pFPMContext->fpm_linkState_cb = NULL;
    return IX_SUCCESS;
}

/**
 * NAME: ix_cc_fpm_register_packet_cb
 *
 * DESCRIPTION: This function is called by the FP Module's platform
 *              specific component to receive packets destined to the CP
 *  
 * @Param:  - IN  arg_Packet_cb: Packet receive callback function
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 *          Return code is not interpreted by the core component.
 * Error Codes:
 * 
 * IX_CC_FPM_ERROR_CALLBACK - user callback returns error message
 * 
 */
ix_error ix_cc_fpm_register_pkt_hdlr_cb (ix_cc_fpm_pkt_hdlr_cb arg_Packet_cb)
{
    if (arg_Packet_cb != NULL) {
	FPM_CC_DBG_PRINT ("Registering callback for packets\n");
	/* store callback in a global function pointer */
	g_pFPMContext->fpm_pktHdlr_cb = arg_Packet_cb;
	return IX_SUCCESS;
    }
    return IX_CC_FPM_ERROR_INVALID_INPUT_PARAM;
}

/**
 * NAME: ix_cc_fpm_deregister_pkt_hdlr_cb
 *
 * DESCRIPTION: This function is called by the FP Module's platform
 *              specific component to deregister its callback
 *  
 * @Param:  - IN  - None
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 *          Return code is not interpreted by the core component.
 

* Error Codes:
 * 
 * IX_CC_FPM_ERROR_CALLBACK - user callback returns error message
 * 
 */

ix_error ix_cc_fpm_deregister_pkt_hdlr_cb ()
{
    g_pFPMContext->fpm_pktHdlr_cb = NULL;
    return IX_SUCCESS;
}

/**
 * NAME: ix_cc_fpm_register_ipv4_exception_cb 
 *
 * DESCRIPTION: This function is called by the FP Module's platform
 *              specific component to receive exceptions for directly
 *              connected hosts (IPv4 exception for an unrouteable
 *              packet to a directly connected host).
 *  
 * @Param:  - IN  arg_ipv4Exception_cb: Callback function that is 
 *            to be invoked for the event
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 *          Return code is not interpreted by the core component.
 * Error Codes:
 * 
 * IX_CC_FPM_ERROR_CALLBACK - user callback returns error message
 * 
 */
ix_error ix_cc_fpm_register_ipv4_exception_cb (ix_cc_fpm_ipv4_exception_cb 
					       arg_ipv4Exception_cb)
{
    if (arg_ipv4Exception_cb != NULL) {
      FPM_CC_DBG_PRINT ("Registering callback for IPv4 Exceptions"
			"(Direct Host routes)\n");
      /* store callback in a global function pointer */
      g_pFPMContext->fpm_ipv4Exception_cb = arg_ipv4Exception_cb;
      return IX_SUCCESS;
    }
    return IX_CC_FPM_ERROR_INVALID_INPUT_PARAM;
}

/**
 * NAME: ix_cc_fpm_deregister_ipv4_exception_cb
 *
 * DESCRIPTION: This function is called by the FP Module's platform
 *              specific component to deregister its callback
 *  
 * @Param:  - IN  - None
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 *          Return code is not interpreted by the core component.
 * Error Codes:
 * 
 * IX_CC_FPM_ERROR_CALLBACK - user callback returns error message
 * 
 */

ix_error ix_cc_fpm_deregister_ipv4_exception_cb ()
{
    g_pFPMContext->fpm_ipv4Exception_cb = NULL;
    return IX_SUCCESS;
}

/**
 * NAME: ix_cc_fpm_dump_context
 *
 * DESCRIPTION: This function dumps the content of the FP Module cc context
 *  
 * @Param:  - IN  - ix_cc_fpm_context* context 
 *
 * @Param:  - OUT - none
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 *          Return code is not interpreted by the core component.
 * Error Codes:
 * 
 * IX_CC_FPM_ERROR_CALLBACK - user callback returns error message
 * 
 */
void ix_cc_fpm_dump_context (ix_cc_fpm_context* context)
{
    FPM_CC_DBG_PRINT ("In %s \n Printing FPM Context \n", __FUNCTION__);
    FPM_CC_DBG_PRINT ("Blade ID = %ld\n", (ix_uint32)context->fpm_localBladeId);
    return;
}
