/*****************************************************************************
 *                            Intel Proprietary
 *
 * Copyright (c) 1998-2003 By Intel Corporation.  All rights reserved.  
 * No part of this program or publication may be reproduced, transmitted,
 * transcribed, stored in a retrieval system, or translated into any language
 * or computer language in any form or by any means, electronic, mechanical,
 * magnetic, optical, chemical, manual, or otherwise, without the prior
 * written permission of:
 *                        Intel Corporation
 *                        2200 Mission College Blvd.
 *                        Santa Clara, CA  95052-8119
 *****************************************************************************/

/*
* This header file defines typedefs, constants, and functions
* that apply to the NPF Interface Management API
*/

#ifndef __NPF_IF_H_
#define __NPF_IF_H_

#ifdef __cplusplus
extern "C" {
#endif

typedef NPF_uint32_t NPF_IfHandle_t;

/*
* Interface Types
*/
typedef enum {
    NPF_IF_TYPE_UNK=1, /* Interface type unknown */
    NPF_IF_TYPE_LAN=2, /* Generic LAN interface */
    NPF_IF_TYPE_ATM=3, /* ATM UNI interface */
    NPF_IF_TYPE_POS=4, /* Packet over SONET interface */
    NPF_IF_TYPE_IPV4=5, /* IPv4 logical interface */
    NPF_IF_TYPE_IPV6=6, /* IPv6 logical interface */
    NPF_IF_TYPE_IPV6INV4=7, /* IPv6 Tunnel interface */
    NPF_IF_TYPE_MPLS=8,     /* MPLS interface */
    NPF_IF_TYPE_HIGHEST=9   /* Highest defined value */
} NPF_IfType_t;


/*
* Operational Status code, returned in the asynchronous
* response from NPF_IfOperStatusGet().
*/

typedef enum {
    NPF_IF_OPER_STATUS_UP = 1, /* Operationally UP */
    NPF_IF_OPER_STATUS_DOWN = 2, /* Operationally DOWN */
    NPF_IF_OPER_STATUS_UNKNOWN /* Status unknown */
} NPF_IfOperStatus_t;



/*
* Administrative Status code, used in the NPF_IfGeneric_t
* structure passed to NPF_IfAttrSet() and NPF_IfCreateAndSet().
*/

typedef enum {
    NPF_IF_ADMIN_STATUS_UP = 1, /* Administratively UP */
    NPF_IF_ADMIN_STATUS_DOWN = 2 /* Administratively DOWN */
} NPF_IfAdminStatus_t;

/*
* Structure to relate two interfaces, passed to NPF_IfBind().
*/

typedef struct {
    NPF_IfHandle_t parent; /* Parent interface handle */
    NPF_IfHandle_t child; /* Child interface handle */
} NPF_IfBinding_t;


/*
* Statistics, from MIB-II Interface Group (RFC 2863), returned
* by asynchronous callback from NPF_IfGenericStatsGet().
*/

typedef struct {
    NPF_uint64_t bytesRx; /* Receive Bytes */
    NPF_uint64_t ucPackRx; /* Receive Unicast Packets */
    NPF_uint64_t mcPackRx; /* Receive Multicast Packets */
    NPF_uint64_t bcPackRx; /* Receive Broadcast packets */
    NPF_uint64_t dropRx; /* Receive packets dropped */
    NPF_uint64_t errorRx; /* Receive errors */
    NPF_uint64_t protoRx; /* Receive unknown protocol */
    NPF_uint64_t bytesTx; /* Transmit bytes */
    NPF_uint64_t ucPackTx; /* Transmit Unicast Packets */
    NPF_uint64_t mcPackTx; /* Transmit Multicast Packets */
    NPF_uint64_t bcPackTx; /* Transmit Broadcast Packets */
    NPF_uint64_t dropTx; /* Transmit dropped packets */
    NPF_uint64_t errorTx; /* Transmit errors */
} NPF_IfStatistics_t;



/*
 * IPv4 Interface attributes
 */
typedef struct {
    NPF_IPv4NetAddr_t addr; /* Primary IPv4 Address and plen */
    NPF_uint16_t mtu; /* IPv4 Max Transmission Unit */
    NPF_IPv4UnicastPrefixTableHandle_t FIB_Handle;/* Forwarding Info Base */
} NPF_IfIPv4_t;

/*
 * MPLS Interface attributes
 */
typedef struct {
    NPF_uint32_t flags;     /* Port Flags */
    NPF_uint32_t mtuSize;   /* Labelled Pkt Size */
}NPF_IfMPLS_t;

/*
* LAN Speed codes for setting Ethernet speed.
*/

typedef enum {
    NPF_IF_LAN_SPEED_10M = 1, /* 10 megabits/second */
    NPF_IF_LAN_SPEED_100M = 2, /* 100 megabits/second */
    NPF_IF_LAN_SPEED_1G = 3, /* 1 gigabit/second */
    NPF_IF_LAN_SPEED_10G = 4, /* 10 gigabit/second */
    NPF_IF_LAN_SPEED_AUTO = 5 /* Set autosense */
} NPF_IfLAN_Speed_t;


/*
* Generic LAN Interface attributes, used in the union within
* NPF_IfGeneric_t.
*/

typedef struct {
    NPF_uint32_t port; /* Port number */
    NPF_boolean_t promisc; /* Promiscuous Mode */
    NPF_IfLAN_Speed_t speed; /* Speed control */
} NPF_IfLAN_t;

/*
* ATM VPI/VCI types, used in various structures to
* identify ATM Vccs.
*/
typedef NPF_uint16_t NPF_VccVPI_t; /* VPI is 8 or 12 bits */
typedef NPF_uint16_t NPF_VccVCI_t; /* VCI is 16 bits */


typedef struct { /* ATM Vcc Address (VPI/VCI) structure */
    NPF_VccVPI_t VPI; /* VPI number */
    NPF_VccVCI_t VCI; /* VCI number */
} NPF_VccAddr_t;


/*
* ATM UNI Vcc AAL type code, used in the ATM Vcc
* attribute structure, NPF_IfVcc_t.
*/

typedef enum {
    NPF_IF_VCC_AAL1 = 1,       /* AAL 1 */
    NPF_IF_VCC_AAL2 = 2,       /* AAL 2 */
    NPF_IF_VCC_AAL3_4 = 3,     /* AAL 3&4 */
    NPF_IF_VCC_AAL5_LLC = 4,   /* AAL 5 with LLC/SNAP */
    NPF_IF_VCC_AAL5_VCMUX = 5, /* AAL 5, VC-multiplexed */
    NPF_IF_VCC_AALRAW = 6,     /* AALRAW */
    NPF_IF_VCC_AAL0 = 7
} NPF_IfAAL_t;


/*
 *	ATM per VCC statistics
 */
typedef struct
{
		NPF_VccAddr_t	vccAddr;			/* VPI/VCI value to which stats apply */
		NPF_uint64_t	cellsClp1Rx;	/* Ingress Total Cells for CLP0 + CLP1 */
		NPF_uint64_t	cellsClp0Rx;	/* Ingress Total Cells for CLP0 */
		NPF_uint64_t	cellsTaggedRx;	/* Ingress Cells changed from CLP0 to CLP1 */
		NPF_uint64_t	cellsDiscardedRx;	/* Ingress total cells discarded */

		NPF_uint64_t	cellsClp1Tx;	/* Egress Total Cells for CLP0 + CLP1 */
		NPF_uint64_t	cellsClp0Tx;	/* Egress Total Cells for CLP 0 */
		
		NPF_uint64_t	framesRx;		/* Ingress AAL frames */
		NPF_uint64_t	framesDiscardedRx;	/* Ingress AAL Frames Discarded */
		NPF_uint64_t	framesTx;		/* Egress AAL Frames */
		
		NPF_uint64_t	bytesRx;			/* Ingress Bytes */
		NPF_uint64_t	bytesTx;			/* Egress Bytes */
}NPF_ATM_VccStats_t;


/*
 * VCC Operational Status 
 */
typedef struct
{
	NPF_VccAddr_t					vccAddr;					/* VCC Address */
	NPF_IfOperStatus_t			operStatus;				/* Operational Status */
}NPF_VccOperStatus_t;

/*
 *   AAL2 Connection ID.
 */
typedef struct
{
		NPF_VccAddr_t  connection;     /* VPI/VCI */
		NPF_uint16_t   aal2Cid;        /* AAL2 Channel Id */
}NPF_ATM_AAL2_ConnId_t;

/*
 *	ATM Cross-Connect
 */
typedef struct
{
		NPF_IfHandle_t		sourceIf;			/* Source Interface */
		NPF_VccAddr_t		sourceVcc;			/* Source VPI/VCI */
		NPF_IfHandle_t		destinationIf;		/* Destination interface */
		NPF_VccAddr_t		destinationVcc;	/* Destination VPI/VCI */
}NPF_ATM_ConnectionXc_t;

/*
 * ATM Address Type 
 */
typedef enum
{
		NPF_IF_ATM_E164_ADDR		=  0,  /* E164 ATM Address */
		NPF_IF_ATM_NSAP_ADDR		=  1   /* NSAP ATM Address */
}NPF_ATM_AddressType_t;

typedef NPF_uint32_t NPF_E164_Address[32];
typedef NPF_uint32_t NPF_NSAP_Address[32];


/* 
 *	ATM Address: E.164 Address and NSAP Address
 */
typedef struct
{
	NPF_ATM_AddressType_t		addrCh;					/* Address Choice */
	union
	{
			NPF_E164_Address		e164Address;			/* E.164 ATM Address */
			NPF_NSAP_Address		nsapAddress;			/* NSAP ATM address */
	}u;
}NPF_ATM_Address_t;

/*
 * Interface Type - Public/Private
 */
typedef enum
{
		NPF_PUBLIC = 0,
		NPF_PRIVATE = 1
}NPF_ATM_InfType_t;

/*
 *	Interface - UNI/NNI
 */
typedef enum
{
		NPF_NNI = 0,		/* Interface type is NNI */
		NPF_UNI = 1			/* Interface type is UNI */
}NPF_ATM_Inf_t;

/* 
 *	UNI Device Type - user side/network side
 */
typedef enum
{
		NPF_USER = 0,		/* User side */
		NPF_NETWORK  = 1	/* Network Side */
}NPF_ATM_DevType_t;

/*
 *	ATM Interface Attributes, used in the union within NPF_IfGeneric_t
 */
typedef struct
{
	NPF_uint32_t					port;						/* Port number */
	NPF_ATM_Address_t				atmAddress;				/* ATM Address */
	NPF_uint16_t					maxVpiBits;				/* MAX VPI bits */
	NPF_uint16_t					maxVciBits;				/* MAX VCI bits */
	NPF_ATM_Inf_t					atmInf;					/* ATM Interface (UNI/NNI) */
	NPF_ATM_InfType_t				infType;					/* Interface Type (Public/Private) */
	NPF_ATM_DevType_t				devType;					/* UNI Device Type (End User/ N/w node) */
	NPF_uchar8_t					rxCpId[16];				/* Ingress CPID */
	NPF_uchar8_t					txCpId[16];				/* Egress CPID */
}NPF_IfATM_t;


/*
* Packet over SONET (POS) Interface Attributes, used in the
* union within NPF_IfGeneric_t.
*/

typedef struct {
    NPF_uint32_t port; /* Port number */
} NPF_IfPOS_t;


/*
 * IPv6 Address Flags
 */

#define NPF_IF_IPV6_ADDR_FLAGS_NONE  0x00
#define NPF_IF_IPV6_ADDR_ANYCAST     0x01
#define NPF_IF_IPV6_ADDR_PREFERRED   0x02
#define NPF_IF_IPV6_ADDR_DEPRICATED  0x04

/*
 * IPv6 unicast zone structure
 */
#define NPF_IF_IPV6UC_SCOPE_LINK_LOCAL 0
#define NPF_IF_IPV6UC_SCOPE_SITE_LOCAL 1
#define NPF_IF_IPV6UC_SCOPE_GLOBAL     2

/*
 * IPv6 Interface attributes
 */

typedef struct {
    NPF_uint32_t                n_addr;     /* Number of IPv6 Addresses */
    NPF_IPv6NetAddr_t           *addr;      /* IPv6 Address */
    NPF_IPv6UCZoneTable_t       zoneTable;  /* Zone index for each scope */
    NPF_uint16_t                mtu;        /* IPv6 Max Transmission Unit */
    NPF_IPv6PrefixTableHandle_t FIB_Handle; /* Forwarding Info Base */
} NPF_IfIPv6_t;

/*
 * IPv6inv4 Tunnel Interface attributes
 */
typedef struct {
    NPF_IPv6inv4TunnelType_t tunnelType;  /* Type of v6inv4 Tunnel */
    NPF_IPv4Address_t        IPv4DstAddr; /* Destination IPv4 address */
    NPF_IPv4Address_t        IPv4SrcAddr; /* Source IPv4 address */
    NPF_uint8_t              IPv4TTL;   /* Time to live in IPv4 header */
    NPF_uint8_t              IPv4DSCP;  /* DiffServ Code Point in IPv4 header */
} NPF_IfIPv6inv4_t;

/*
 * ICMP v6 Rate Limiting types
 */
typedef enum {
    NPF_IF_ICMPV6_RATE_TIMER = 1,     /* Timer based rate control */
    NPF_IF_ICMPV6_RATE_BANDWIDTH = 2  /* Bandwidth based rate control */
} NPF_ICMPv6RateLimitType_t;

/*
* The Interface structure, passed to NPF_IfAttrSet()
* and NPF_IfCreateAndSet().
*/

typedef struct {
    NPF_IfType_t type; /* Logical interface type */
    NPF_uint64_t speed; /* Speed in Kbits/second */
    NPF_IfAdminStatus_t adminStatus; /* Administrative up/down */
    union  /* @!union-switch:type! Type specific attributes (by if_type code) */
    {
        NPF_IfIPv4_t IPv4_Attr; /* @!union-case:NPF_IF_TYPE_IPV4! IPv4 Interface attributes */
        NPF_IfLAN_t LAN_Attr; /* @!union-case:NPF_IF_TYPE_LAN! LAN interface attributes */
        NPF_IfATM_t ATM_Attr; /* @!union-case:NPF_IF_TYPE_ATM! ATM UNI interface attributes */
        NPF_IfPOS_t POS_Attr; /* @!union-case:NPF_IF_TYPE_POS! POS interface attributes */
	NPF_IfIPv6_t IPv6_Attr; /* @!union-case:NPF_IF_TYPE_IPV6! IPv6 interface attributes */
	NPF_IfIPv6inv4_t IPv6inv4_Attr; /* @!union-case:NPF_IF_TYPE_IPV6INV4! IPv6 tunnel attributes */
	NPF_IfMPLS_t MPLS_Attr; /* @!union-case:NPF_IF_TYPE_MPLS! MPLS Iface attributes */
    } u;
} NPF_IfGeneric_t;

/*
* Completion Callback Types, to be found in the callback
* data structure, NPF_IfCallbackData_t.
*/

typedef enum NPF_IfCallbackType {
    NPF_IF_CREATE = 1,
    NPF_IF_DELETE = 2,
    NPF_IF_BIND = 3,
    NPF_IF_STATS_GET = 4,
    NPF_IF_VCC_STATS_GET = 5,
    NPF_IF_ATTR_SET = 6,
    NPF_IF_CREATE_AND_SET = 7,
    NPF_IF_ENABLE = 8,
    NPF_IF_DISABLE = 9,
    NPF_IF_OPER_STATUS_GET = 10,
    NPF_IF_LAN_SRC_ADDR_GET = 11,
    NPF_IF_LAN_SRC_ADDR_SET = 12,
    NPF_IF_LAN_ADDR_LIST_SET = 13,
    NPF_IF_LAN_ADDR_LIST_ADD = 14,
    NPF_IF_LAN_PROMISC_SET = 15,
    NPF_IF_LAN_PROMISC_CLEAR = 16,
    NPF_IF_LAN_FULL_DUPLEX_SET = 17,
    NPF_IF_LAN_FULL_DUPLEX_CLEAR = 18,
    NPF_IF_LAN_SPEED_SET = 19,
    NPF_IF_LAN_FLOW_CONTROL_TX_ENABLE = 20,
    NPF_IF_LAN_FLOW_CONTROL_TX_DISABLE = 21,
    NPF_IF_LAN_FLOW_CONTROL_RX_ENABLE = 22,
    NPF_IF_LAN_FLOW_CONTROL_RX_DISABLE = 23,
    NPF_IF_IPV4ADDR_SET = 24,
    NPF_IF_IPV4MTU_SET = 25,
    NPF_IF_IPV4FIB_SET = 26,
    NPF_IF_ATM_VCC_SET = 27,
    NPF_IF_ATM_VCC_BIND = 28,
    NPF_IF_ATM_VCC_DELETE = 29,
    NPF_IF_IPV6_ADDR_ADD = 30,
    NPF_IF_IPV6_ADDR_DELETE = 31,
    NPF_IF_IPV6_UCZONE_SET = 32,
    NPF_IF_IPV6_MTU_SET = 33,
    NPF_IF_IPV6_FIB_SET = 34,
    NPF_IF_IPV6INV4_DSTADDR_SET = 35,
    NPF_IF_IPV6INV4_SRCADDR_SET = 36,
    NPF_IF_IPV6INV4_TTL_SET = 37,
    NPF_IF_IPV6INV4_DSCP_SET = 38,
    NPF_IF_ICMPV6_RATE_SET = 39,
    NPF_IF_MPLSMTU_SET = 40,
	 NPF_IF_ATM_VCC_ENABLE					= 41,
	 NPF_IF_ATM_VCC_DISABLE					= 42,
	 NPF_IF_ATM_VCC_OPER_STATUS_GET		= 43,
	 NPF_IF_ATM_VCC_STATS_GET				= 44,
	 NPF_IF_ATM_AAL2_CHANNEL_SET			= 45,
	 NPF_IF_ATM_AAL2_CHANNEL_DEL			= 46,
	 NPF_IF_ATM_CONNECTION_XC_SET			= 47,
	 NPF_IF_ATM_CONNECTION_XC_DEL			= 48,
	 NPF_IF_ATM_OAM_CP_SET					= 49,
	 NPF_IF_ATM_OAM_CC_SET					= 50,	
	 NPF_IF_ATM_OAM_LB_SET					= 51,
	 NPF_IF_ATM_OAM_MONITOR					= 52,
	 NPF_IF_ATM_OAM_ALARM_SET				= 53,
	 NPF_IF_ATM_OAM_ALARM_CLEAR			= 54,
	 NPF_IF_ATM_OAM_PM_SET					= 55
} NPF_IfCallbackType_t;


/*
* Asynchronous error codes (returned in function callbacks)
*/


/* Callback/event reg. error */
#define NPF_IF_E_ALREADY_REGISTERED ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR)

/* Callback/event handle invalid */
#define NPF_IF_E_BAD_CALLBACK_HANDLE ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+1)

/* Callback function is NULL */
#define NPF_IF_E_BAD_CALLBACK_FUNCTION ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+2)

/* Invalid parameter */
#define NPF_IF_E_INVALID_PARAM ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+3)

/* Invalid child i/f handle */
#define NPF_IF_E_INVALID_CHILD_HANDLE ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+4)

/* Invalid parent i/f handle */
#define NPF_IF_E_INVALID_PARENT_HANDLE ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+5)

/* Invalid interface handle */
#define NPF_IF_E_INVALID_HANDLE ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+6)

/* Invalid VPI/VCI address */
#define NPF_IF_E_NO_SUCH_VCC ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+7)

/* Invalid IP address */
#define NPF_IF_E_INVALID_IPADDR ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+8)

/* Invalid IP net prefix length */
#define NPF_IF_E_INVALID_NETPLEN ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+9)

/* Invalid IP MTU specification */
#define NPF_IF_E_INVALID_MTU ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+10)

/* Invalid interface attribute */
#define NPF_IF_E_INVALID_ATTRIBUTE ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+11)

/* Error  interface not created */
#define NPF_IF_E_NOT_CREATED ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+12)

/* Invalid MAC address */
#define NPF_IF_E_INVALID_MAC_ADDR ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+13)

/* Invalid FIB handle */
#define NPF_IF_E_INVALID_FIB_HANDLE ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+14)

/* Invalid ATM UNI i/f handle */
#define NPF_IF_E_INVALID_ATM_HANDLE ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+15)

/* Invalid layer 3 i/f handle */
#define NPF_IF_E_INVALID_L3_HANDLE ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+16)

/* Array length <= 0 or too big */
#define NPF_IF_E_BAD_ARRAY_LENGTH ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+17)

/* Interface has no source addr. */
#define NPF_IF_E_NO_SRC_ADDRESS ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+18)

/* Invalid generic interface speed */
#define NPF_IF_E_INVALID_SPEED ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+19)

/* Invalid VPI/VCI */
#define NPF_IF_E_INVALID_VCC_ADDRESS ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+20)

/* Invalid Interface Type */
#define NPF_IF_E_INVALID_IF_TYPE ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+21)

/* Invalid Administrative Status code */
#define NPF_IF_E_INVALID_ADMIN_STATUS ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+22)

/* Invalid Port number */
#define NPF_IF_E_INVALID_PORT_NUMBER ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+23)

/* Invalid Promiscuous Mode code */
#define NPF_IF_E_INVALID_PROMISC_MODE ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+24)

/* Invalid LAN speed code */
#define NPF_IF_E_INVALID_LAN_SPEED ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+25)

/* Invalid ATM AAL code */
#define NPF_IF_E_INVALID_ATM_AAL ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+26)

/* Invalid ATM QoS specification */
#define NPF_IF_E_INVALID_ATM_QOS ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+27)

/* Invalid ICMP rate limiting type */
#define NPF_IF_E_INVALID_ICMP_RATELIMITTYPE ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+28)

/* Invalid ICMP rate limiting value */
#define NPF_IF_E_INVALID_ICMP_RATELIMITVALUE ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+29)

/* Invalid IPv6 scope value */
#define NPF_IF_E_INVALID_SCOPEVALUE ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+30)

/* Invalid TTL value */
#define NPF_IF_E_INVALID_TTL ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+31)

/* Invalid DSCP value */
#define NPF_IF_E_INVALID_DSCP ((NPF_IfErrorType_t) NPF_INTERFACES_BASE_ERR+32)

typedef NPF_uint32_t NPF_IfErrorType_t;


/*
* An asynchronous response contains an interface handle,
* a error or success code, and in some cases a function-
* specific structure embedded in a union. One or more of
* these is passed to the callback function as an array
* within the NPF_IfCallbackData_t structure (below).
*/
typedef struct {        /* Asynchronous Response Structure */
	NPF_IfHandle_t ifHandle;     /* Interface handle for this response */
	NPF_IfErrorType_t error;     /* Error code for this response */
	union {             /* Function-specific structures: */
		NPF_uint32_t unused;     /* Default */
		NPF_uint32_t arrayIndex;    /* NPF_IfCreateAndSet index */
		NPF_IfStatistics_t *ifStats;   /* NPF_IfGenericStatsGet() */
		NPF_IfOperStatus_t operStat;   /* NPF_IfOperStatusGet() */
		NPF_ATM_VccStats_t *vccStats;  /* NPF_IfVccStatsGet() */
		NPF_IfHandle_t child;       /* NPF_IfBind(), handle=parent*/
		NPF_MAC_Address_t MACaddr;  /* NPF_IfLAN_SrcAddrGet() */
		NPF_VccAddr_t VccAddr;      /* NPF_IfATM_VccSet(), */
		NPF_VccOperStatus_t vccOperStat; /* NPF_IfATM_VccOperStatusGet() */
		NPF_ATM_AAL2_ConnId_t  aal2ConnId; /* NPF_IfATM_AAL2_ChannelSet() */
		NPF_ATM_ConnectionXc_t connXcInfo; /* NPF_IfATM_ConnectionXcSet/Del () */
    } u;
} NPF_IfAsyncResponse_t;


/*
* The callback function receives the following structure containing
* one or more asynchronous responses from a single function call.
* There are several possibilities:
* 1. The called function does a single request
* - n_resp = 1, and the resp array has just one element.
* - allOK = TRUE if the request completed without error
* and the only return value is the response code.
* - if allOK = FALSE, the "resp" structure has the error code.
* 2. the called function supports an array of requests
* a. All completed successfully, at the same time, and the
* only returned value is the response code:
* - allOK = TRUE, n_resp = 0.
* b. Some completed, but not all
* - allOK = FALSE, n_resp = the number completed
* - the "resp" array will contain one element for
* each completed request, with the error code
* in the NPF_IfAsyncResponse_t structure, along
* with any other information needed to identify
* which request element the response belongs to.
* - Callback function invocations are repeated in
* this fashion until all requests are complete.
* Responses are not repeated for request elements
* already indicated as complete in earlier callback
* function invocations.
*/

typedef struct {
    NPF_IfCallbackType_t type; 		/* Which function was called? */
    NPF_boolean_t allOK; 		/* TRUE if all completed OK */
    NPF_uint32_t n_resp; 		/* Number of responses in array */
    NPF_IfAsyncResponse_t *resp; 	/* Pointer to response structures*/
} NPF_IfCallbackData_t;


/*
* Event types
*/

typedef enum {
    NPF_IF_OPR_UP = 1, /* Interface went oper UP */
    NPF_IF_OPR_DOWN = 2, /* Interface went oper DOWN */
    NPF_IF_COUNTER_DISCONTINUITY = 3, /* Counter discontinuity occurred*/
	 NPF_IF_ATM_EVENT_OAM_AIS_ALARM = 4,
	 NPF_IF_ATM_EVENT_OAM_RDI_ALARM = 5,
	 NPF_IF_ATM_EVENT_OAM_LB = 6,
	 NPF_IF_ATM_EVENT_OAM_FPM_STATS = 7,
	 NPF_IF_ATM_EVENT_OAM_BR_STATS = 8,
	 NPF_IF_ATM_EVENT_ATM_VCC_UP_EVENT = 9,
	 NPF_IF_ATM_EVENT_ATM_VCC_DOWN_EVENT = 10
} NPF_IfEvent_t;


/*
* Event notification structure and array
*/
typedef struct NPF_IfEventData {
    NPF_IfEvent_t eventType; /* Event type */
    NPF_IfHandle_t handle; /* Interface handle */
} NPF_IfEventData_t;


typedef struct {
    NPF_uint16_t n_data; /* Number of events in array */
    NPF_IfEventData_t *eventData; /* Array of event notifications */
} NPF_IfEventArray_t;


typedef NPF_uint32_t NPF_IfEventHandlerHandle_t;

/*
* Completion Callback Function
*/
typedef void (*NPF_IfCallbackFunc_t) (
		NPF_IN NPF_userContext_t userContext,
		NPF_IN NPF_correlator_t correlator,
		NPF_IN NPF_IfCallbackData_t *ifCallbackData);


/*
* Completion Callback Registration Function
*/
NPF_error_t NPF_IfRegister(
		NPF_IN NPF_userContext_t userContext,
		NPF_IN NPF_IfCallbackFunc_t ifCallbackFunc,
		NPF_OUT NPF_callbackHandle_t *CallbackHandle);

/*
* Completion Callback Deregistration Function
*/

NPF_error_t NPF_IfDeregister(
		NPF_IN NPF_callbackHandle_t CallbackHandle);
/*
* Event Handler Function
*/
typedef void (*NPF_IfEventHandlerFunc_t) (
		NPF_IN NPF_userContext_t userContext,
		NPF_IN NPF_IfEventArray_t ifEventArray);

/*
* Event Handler Registration Function
*/
NPF_error_t NPF_IfEventRegister(
		NPF_IN NPF_userContext_t userContext,
		NPF_IN NPF_IfEventHandlerFunc_t ifEventHandlerFunc,
		NPF_OUT NPF_IfEventHandlerHandle_t *ifEventHandlerHandle);


/*
* Event Handler Deregistration Function
*/
NPF_error_t NPF_IfEventDeregister(
		NPF_IN NPF_IfEventHandlerHandle_t ifEventHandlerHandle);

/*
* Function to Create one or more Interfaces
*/
NPF_error_t NPF_IfCreate(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_if,
		NPF_IN NPF_IfType_t if_Type);

/*
* Function to Delete one or more Interfaces
*/
NPF_error_t NPF_IfDelete(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray);

/*
* Function to Bind one or more pairs of Interfaces
*/
NPF_error_t NPF_IfBind(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t if_nbinds,
		NPF_IN NPF_IfBinding_t *if_bindArray);

/*
* Function to Read Statistics from one or more Interfaces
*/
NPF_error_t NPF_IfGenericStatsGet(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray);

/*
* Function to Read Statistics from one or
* more Vccs on a single ATM UNI interface
*/
NPF_error_t NPF_IfVccStatsGet(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_IfHandle_t if_Handle,
		NPF_IN NPF_uint32_t n_Vccs,
		NPF_IN NPF_VccAddr_t *if_VccAddrArray);

/*
* Function to Set all Interface Attributes for one
* or more interfaces
*/
NPF_error_t NPF_IfAttrSet(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray,
		NPF_IN NPF_IfGeneric_t *if_StructArray);

/*
* Function to create one or more interfaces and set all of
* their attributes
*/
NPF_error_t NPF_IfCreateAndSet(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_if,
		NPF_IN NPF_IfGeneric_t *if_StructArray);

/*
* Function to Enable one or more Interfaces
*/
NPF_error_t NPF_IfEnable(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray);

/*
* Function to Disable one or more Interfaces
*/
NPF_error_t NPF_IfDisable(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray);

/*
* Function to Return the Operational Status of one
* or more Interfaces
*/
NPF_error_t NPF_IfOperStatusGet(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray);

/*
* Function to Return the Source MAC Address of one
* or more LAN Interfaces
*/
NPF_error_t NPF_IfLAN_SrcAddrGet(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray);

/*
* Function to Set the Source MAC Address of one
* or more LAN Interfaces
*/
NPF_error_t NPF_IfLAN_SrcAddrSet(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray,
		NPF_IN NPF_MAC_Address_t *if_LAN_SrcAddrArray);

/*
* Function to Set a List of Receive MAC Addresses for
* one or more LAN Interfaces
*/
NPF_error_t NPF_IfLAN_MAC_RcvAddrListSet(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray,
		NPF_IN NPF_uint32_t if_n_MAC_Addrs,
		NPF_IN NPF_MAC_Address_t *if_LAN_AddrArray);

/*
* Function to Add to the List of Receive MAC Addresses
* for one or more LAN Interfaces
*/
NPF_error_t NPF_IfLAN_MAC_RcvAddrListAdd(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray,
		NPF_IN NPF_uint32_t if_n_MAC_Addrs,
		NPF_IN NPF_MAC_Address_t *if_LAN_AddrArray);

/*
* Function to Set one or more LAN Interfaces
* in Promiscuous Mode
*/
NPF_error_t NPF_IfLAN_PromiscSet(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray);

/*
* Function to Turn off Promiscuous Mode in one or more
* LAN Interfaces
*/
NPF_error_t NPF_IfLAN_PromiscClear(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray);

/*
* Function to set full duplex mode on one or more
* 10/100 and gigabit Ethernet LAN Interfaces
*/
NPF_error_t NPF_IfLAN_FullDuplexSet(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray);

/*
* Function to disable full duplex mode on one or more
* 10/100 and gigabit Ethernet LAN Interfaces
*/
NPF_error_t NPF_IfLAN_FullDuplexClear(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray);

/*
* Function to set interface speed or autosense
* on one or more 10/100 Ethernet LAN Interfaces
*/
NPF_error_t NPF_IfLAN_SpeedSet(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray,
		NPF_IN NPF_IfLAN_Speed_t *speedArray);

/*
* Function to enable transmitting Flow Control messages on
* one or more 10/100 and gigabit Ethernet LAN Interfaces
*/
NPF_error_t NPF_IfLAN_FlowControlTxEnable(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray);

/*
* Function to disable transmitting Flow Control messages
* on one or more 10/100 and gigabit Ethernet LAN Interfaces
*/
NPF_error_t NPF_IfLAN_FlowControlTxDisable(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray);

/*
* Function to enable responding to Flow Control messages
* on one or more 10/100 and gigabit Ethernet LAN Interfaces
*/
NPF_error_t NPF_IfLAN_FlowControlRxEnable(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray);

/*
* Function to disable responding to Flow Control messages
* on one or more 10/100 and gigabit Ethernet LAN Interfaces
*/
NPF_error_t NPF_IfLAN_FlowControlRxDisable(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray);

/*
* Function to set the primary IP address and prefix length
* on one or more IPv4 Interfaces
*/
NPF_error_t NPF_IfIPv4AddrSet(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray,
		NPF_IN NPF_IPv4NetAddr_t *if_IPv4AddrArray);

/*
* Function to set the MTU on one or more IPv4 interfaces
*/
NPF_error_t NPF_IfIPv4MTUSet(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray,
		NPF_IN NPF_uint16_t *if_IPv4MTUArray);

/*
* Function to Associate a FIB to one or more IPv4 Interfaces
*/
NPF_error_t NPF_IfIPv4FIBSet(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray,
		NPF_IN NPF_IPv4UnicastPrefixTableHandle_t if_FIB_Handle);


/*
 * Assign IP address and network prefix to interface
 */
NPF_error_t NPF_IfIPv6AddrAdd(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t     if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t         n_handles,
		NPF_IN NPF_IfHandle_t       *if_HandleArray,
		NPF_IN NPF_IPv6NetAddr_t    *if_IPv6AddrArray);

/*
 * Delete IP address and network prefix from interface
 */
NPF_error_t NPF_IfIPv6AddrDelete(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t     if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t         n_handles,
		NPF_IN NPF_IfHandle_t       *if_HandleArray,
		NPF_IN NPF_IPv6NetAddr_t    *if_IPv6AddrArray);

/*
 * Set an IPv6 Interface's Scope Specific Zones address
 */
NPF_error_t NPF_IfIPv6UCZoneSet(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t     if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t         n_handles,
		NPF_IN NPF_IfHandle_t       *if_HandleArray,
		NPF_IN NPF_uint8_t          *if_scopeArray,
		NPF_IN NPF_uint32_t         *if_zoneArray);

/*
 * Set an IPv6 Interface's MTU
 */
NPF_error_t NPF_IfIPv6MTUSet(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t     if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t         n_handles,
		NPF_IN NPF_IfHandle_t       *if_HandleArray,
		NPF_IN NPF_uint16_t         *if_IPv6MTU_Array);


/*
 * Associate a FIB to an IPv6 Interface
 */
NPF_error_t NPF_IfIPv6FIBSet(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t     if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t         n_handles,
		NPF_IN NPF_IfHandle_t       *if_HandleArray,
		NPF_IN NPF_IPv6PrefixTableHandle_t if_FIB_Handle);

/*
 * Set IPv6 in IPv4 Tunnel Destination Address
 */
NPF_error_t NPF_IfIPv6inv4DstAddrSet(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t     if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t         n_handles,
		NPF_IN NPF_IfHandle_t       *if_HandleArray,
		NPF_IN NPF_IPv4Address_t    *if_IPv4AddressArray);

/*
 * Set IPv6 in IPv4 Tunnel Source Address
 */
NPF_error_t NPF_IfIPv6inv4SrcAddrSet(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t     if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t         n_handles,
		NPF_IN NPF_IfHandle_t       *if_HandleArray,
		NPF_IN NPF_IPv4Address_t    *if_IPv4AddressArray);

/*
 * Set IPv6 in IPv4 Tunnel Time To Live
 */
NPF_error_t NPF_IfIPv6inv4TTLSet(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t     if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t         n_handles,
		NPF_IN NPF_IfHandle_t       *if_HandleArray,
		NPF_IN NPF_uint8_t          *if_IPv4TTLArray);

/*
 * Set IPv6 in IPv4 Tunnel DSCP
 */
NPF_error_t NPF_IfIPv6inv4DSCPSet(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t     if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t         n_handles,
		NPF_IN NPF_IfHandle_t       *if_HandleArray,
		NPF_IN NPF_uint8_t          *if_IPv4DSCPArray);

/*
 * Set ICMPIv6 Rate Limiting Parameters
 */
NPF_error_t NPF_IfIICMPv6RateLimitSet(
		NPF_IN NPF_callbackHandle_t      if_cbHandle,
		NPF_IN NPF_correlator_t          if_cbCorrelator,
		NPF_IN NPF_errorReporting_t      if_errorReporting,
		NPF_IN NPF_ICMPv6RateLimitType_t if_rateLimitType,
		NPF_IN NPF_uint32_t              *if_rateLimitValue);

/*
* Function to set the MTU on one or more MPLS interfaces
*/
NPF_error_t NPF_IfMPLSMTUSet(
		NPF_IN NPF_callbackHandle_t if_cbHandle,
		NPF_IN NPF_correlator_t if_cbCorrelator,
		NPF_IN NPF_errorReporting_t if_errorReporting,
		NPF_IN NPF_uint32_t n_handles,
		NPF_IN NPF_IfHandle_t *if_HandleArray,
		NPF_IN NPF_uint16_t *if_MPLSMTUArray);


#ifdef __cplusplus
}
#endif

#endif /* __NPF_IF_ */
