/*****************************************************************************
*                  INTEL CORPORATION PROPRIETARY INFORMATION
*  This software is supplied under the terms of a license agreement or
*  nondisclosure agreement with Intel Corporation and may not be copied
*  or disclosed except in accordance with the terms of that agreement.
*  Copyright (c) 2001 Intel Corporation. All Rights Reserved.
*****************************************************************************/


/*****************************************************************************
 * Description: 
 * This header file defines typedefs, constants, for the NP Forum 
 * IPv4 Unicast Forwarding API
******************************************************************************/

#ifndef __NPF_IPV4_TYPES_H
#define __NPF_IPV4_TYPES_H


#ifdef __cplusplus
extern "C" {
#endif

#include "npf_common.h"

/*-----------------------------------------------------------------
* 
* Common Data Types
*
*-----------------------------------------------------------------*/
/* 
* Table support enumeration.
*/

typedef enum {
	NPF_IPV4UC_UNIFIED_ONLY = 1,
	NPF_IPV4UC_BOTH_SUPPORTED = 2
} NPF_IPv4UC_SupportedMode_t;
/*
* Table preference enumeration. No preference value may only be returned
* by implementations that returned "both supported" to the support query. 
*/

typedef enum {
	NPF_IPV4UC_DISCRETE_PREFERRED = 1,
	NPF_IPV4UC_UNIFIED_PREFERRED = 2,
	NPF_IPV4UC_NO_PREFERENCE = 3
} NPF_IPv4UC_PreferredMode_t;

/*
* IPv4 unicast prefix type:
*/

typedef NPF_IPv4NetAddr_t NPF_IPv4UC_Prefix_t;


/*
* Next hop type definition
*/

typedef enum {
	NPF_IPV4UC_NH_BASIC = 1,
	NPF_IPV4UC_NH_DIRECT_ATTACH = 2,
	NPF_IPV4UC_NH_SEND_TO_CP = 3,
	NPF_IPV4UC_NH_DISCARD = 4,
	NPF_IPV4UC_NH_REMOTE = 5
} NPF_IPv4UC_NextHopType_t;


/*
* Media Address structure:
*/

typedef struct {
	NPF_MediaType_t type;
	union {
		NPF_MAC_Address_t MAC_Address;
		NPF_VccAddr_t ATM_Vc;
	}u;
} NPF_MediaAddress_t;

/*
* IPv4 unicast next hop structure: weight, egressInterface and
* nextHopIP fields are valid only for IPV4UC_BASIC,
* IPV4UC_DIRECT_ATTACH, and IPV4UC_REMOTE types.
*/

typedef struct {
	NPF_IPv4UC_NextHopType_t type;
	NPF_uint16_t weight;
	NPF_IfHandle_t egressInterface;
	NPF_IPv4Address_t nextHopIP;
	NPF_MediaAddress_t mediaAddress;
} NPF_IPv4UC_NextHop_t;

/*
* IPv4 unicast Next Hop Entry: nextHopArray points to an array
* (one or more) of NPF_IPv4UC_NextHop_t structures.
* nextHopCount indicates how many next hops are in the array.
* An array is passed because a single prefix may use multiple
* next hops.
*/

typedef struct {
	NPF_uint32_t nextHopCount;
	NPF_IPv4UC_NextHop_t *nextHopArray;
} NPF_IPv4UC_NextHopArray_t;


/*
* IPv4 unicast Address Resolution entry:
*/

typedef struct {
	NPF_IPv4Address_t IP_Address;
	NPF_IfHandle_t interfaceHandle;
	NPF_MediaAddress_t mediaAddress;
} NPF_IPv4UC_AddResEntry_t;

/*
* This structure contains the key of an Address Resolution
* table entry, consisting of IP address and interface handle.
*/

typedef struct {
	NPF_IPv4Address_t IP_Address;
	NPF_IfHandle_t interfaceHandle;
} NPF_IPv4UC_AddResKey_t;

/*
* Meaningful structure name used in adress resolution
* asynchronous callback data.
*/

typedef NPF_IPv4UC_AddResEntry_t NPF_IPv4UC_AddResQueryResp_t;

/*
* Common table handle types
*/ 

typedef NPF_uint32_t NPF_IPv4UC_AddResTableHandle_t;

/*
* Asynchronous error codes (returned in function callbacks) 
*/

typedef NPF_uint32_t NPF_IPv4UC_ReturnCode_t;


#define IPV4_ERR(n) ((NPF_IPv4UC_ReturnCode_t) NPF_IPV4_BASE_ERR + (n)) 
#define NPF_IPV4UC_TABLE_FULL IPV4_ERR(0)
#define NPF_IPV4UC_TABLE_ENTRY_DOES_NOT_EXIST IPV4_ERR(1)
#define NPF_IPV4UC_FUNCTION_NOT_SUPPORTED IPV4_ERR(2)
#define NPF_IPV4UC_INVALID_HANDLE IPV4_ERR(3) 
#define NPF_IPV4UC_INSUFFICIENT_STORAGE IPV4_ERR(4)


/*-----------------------------------------------------------------
*
* Discrete Mode Data Types 
*
*-----------------------------------------------------------------*/



/*
* This structure contains the query results for a single 
* prefix table entry.
*/

typedef struct {
	NPF_IPv4UC_Prefix_t prefix;
	NPF_uint32_t nextHopIdentifier; 
} NPF_IPv4UC_PrefixQueryResp_t;

/*
* This structure contains the query results for a single
* next hop table entry. 
*/

typedef struct {
	NPF_uint32_t nextHopIdentifier;
	NPF_IPv4UC_NextHopArray_t nextHopArray;
} NPF_IPv4UC_NextHopQueryResp_t; 

/*
* Discrete mode handle types
*/

typedef NPF_uint32_t NPF_IPv4UC_PrefixTableHandle_t; 

typedef NPF_uint32_t NPF_IPv4UC_NextHopTableHandle_t;

typedef NPF_uint32_t NPF_IPv4UC_FwdTableHandle_t;


/*
* Asynchronous response structure for NPF_IPv4UC_PrefixTableHandleCreate()
*/ 

typedef struct {
	NPF_IPv4UC_FwdTableHandle_t extHandle;
	NPF_IPv4UC_PrefixTableHandle_t intHandle;
} NPF_IPv4UC_PfxCreateResp_t;



/*-----------------------------------------------------------------
*
* Unified Mode Data Types
*
*-----------------------------------------------------------------*/ 

/*
* This structure contains the query results for a single FIB table
* entry.
*/ 

typedef struct {
	NPF_IPv4UC_Prefix_t prefix;
	NPF_IPv4UC_NextHopArray_t nextHopArray;
} NPF_IPv4UC_FibQueryResp_t;

/*
* Unified table handle types
*/

typedef NPF_uint32_t NPF_IPv4UC_FibTableHandle_t;


/*
* Asynchronous response structure for NPF_IPv4UC_FIBTableHandleCreate()
*/

typedef struct {
	NPF_IPv4UC_FwdTableHandle_t extHandle; 
	NPF_IPv4UC_FibTableHandle_t intHandle;
} NPF_IPv4UC_FibCreateResp_t;



/*-----------------------------------------------------------------
* 
* Completion Callback Data Types
*
*-----------------------------------------------------------------*/


/* 
* Common callback definition:
*/

typedef enum NPF_IPv4UC_CallbackType {
	NPF_IPV4UC_PREFIX_TABLE_HANDLE_CREATE = 1,
	NPF_IPV4UC_PREFIX_TABLE_HANDLE_DELETE = 2, 
	NPF_IPV4UC_PREFIX_ENTRY_ADD = 3,
	NPF_IPV4UC_PREFIX_ENTRY_DELETE = 4,
	NPF_IPV4UC_PREFIX_TABLE_FLUSH = 5,
	NPF_IPV4UC_PREFIX_TABLE_ATTRIBUTE_QUERY = 6,
	NPF_IPV4UC_PREFIX_ENTRY_QUERY = 7, 
	NPF_IPV4UC_PREFIX_NEXT_HOP_TABLE_BIND = 8,
	NPF_IPV4UC_NEXT_HOP_TABLE_HANDLE_CREATE = 9,
	NPF_IPV4UC_NEXT_HOP_TABLE_HANDLE_DELETE = 10,
	NPF_IPV4UC_NEXT_HOP_ENTRY_ADD = 11,
	NPF_IPV4UC_NEXT_HOP_ENTRY_DELETE = 12, 
	NPF_IPV4UC_NEXT_HOP_TABLE_FLUSH = 13,
	NPF_IPV4UC_NEXT_HOP_TABLE_ATTRIBUTE_QUERY = 14,
	NPF_IPV4UC_NEXT_HOP_ENTRY_QUERY = 15,
	NPF_IPV4UC_FIB_TABLE_HANDLE_CREATE = 16,
	NPF_IPV4UC_FIB_TABLE_HANDLE_DELETE = 17, 
	NPF_IPV4UC_FIB_ENTRY_ADD = 18,
	NPF_IPV4UC_FIB_ENTRY_DELETE = 19,
	NPF_IPV4UC_FIB_TABLE_FLUSH = 20,
	NPF_IPV4UC_FIB_TABLE_ATTRIBUTE_QUERY = 21,
	NPF_IPV4UC_FIB_ENTRY_QUERY = 22, 
	NPF_IPV4UC_ADDRESS_RES_TABLE_HANDLE_CREATE = 23,
	NPF_IPV4UC_ADDRESS_RES_TABLE_HANDLE_DELETE = 24,
	NPF_IPV4UC_ADDRESS_RES_ENTRY_ADD = 25,
	NPF_IPV4UC_ADDRESS_RES_ENTRY_DELETE = 26,
	NPF_IPV4UC_ADDRESS_RES_TABLE_FLUSH = 27, 
	NPF_IPV4UC_ADDRESS_RES_TABLE_ATTRIBUTE_QUERY = 28,
	NPF_IPV4UC_ADDRESS_RES_ENTRY_QUERY = 29
} NPF_IPv4UC_CallbackType_t;


/* 
* An asynchronous response contains a return code indicating
* an error or success of a particular request operation.
* The structure may also contain other optional information
* that was requested by the operation orthe information may
* assist in correlating the response to the corresponding request 
* operation when multiple operations are requested by the application.
*/

typedef struct {
	NPF_IPv4UC_ReturnCode_t returnCode;
	union { 
		NPF_IPv4UC_PfxCreateResp_t prefixTableHandles;
		NPF_IPv4UC_Prefix_t prefix;
		NPF_IPv4UC_PrefixQueryResp_t prefixQueryResult;
		NPF_IPv4UC_NextHopTableHandle_t nextHopTableHandle;
		NPF_uint32_t nextHopIdentifier; 
		NPF_IPv4UC_NextHopQueryResp_t nextHopQueryResult;
		NPF_IPv4UC_FibCreateResp_t fibTableHandles;
		NPF_IPv4UC_Prefix_t fibPrefix;
		NPF_IPv4UC_FibQueryResp_t fibQueryResult;
		NPF_IPv4UC_AddResTableHandle_t addResTableHandle; 
		NPF_IPv4UC_AddResKey_t addResKey;
		NPF_IPv4UC_AddResQueryResp_t addResQueryResult;
		NPF_uint32_t tableSpaceRemaining;
		NPF_uint32_t unused;
	} u; 
} NPF_IPv4UC_AsyncResponse_t;


/*
* This structure is passed to the application as a paramter on a registered
* completion callback. The type field inidcates which function invocation 
* led to this response. The other three fields contain values depending
* upon the invoking function, whether or not a single operation was
* requested and whether the operations were successful or not.
*
* There are several possibilities: 
*
* The application invokes a function requesting a single operation:
* - If allOK = TRUE, then numResp = 0 and the "resp" pointer is NULL.
* This indicates the operation completed successfully and there is
* no other additional response data to return. 
* - If allOK = FALSE, then numResp = 1 and the "resp" pointer points to
* a response structure. If the returnCode field indicates NPF_NO_ERROR,
* the operation completed successfully and there is additional response
* data in the structure. Otherwise, the operation failed and the reason
* is indicated by the returnCode. 
* The application invokes a function requesting multiple operations:
* - If all operations completed successfully at the same time and there
* is no additional response data to provide, then allOK = TRUE,
* numResp = 0 and the "resp" pointer is NULL.
* - If all operations completed successfully at the same time, but there 
* is additional response data to provide, then allOK = FALSE, numResp
* indicates the total number of requested operations and the "resp"
* pointer points to an array of response structures. The returnCode
* field will indicate NPF_NO_ERROR.
* - If some operations completed, but not all, then: 
* > allOK = FALSE, numResp = the number of request operations
* completed.
* > The "resp" pointer will point to an array of response structures,
* each one containing one element for each completed request. For
* operations that completed successfully, the returnCode field will 
* indicate NPF_NO_ERROR and additional response data may be present,
* depending on the type of function invocation. For operations that
* failed, the reason is indicated by the returnCode field.
*/

typedef struct {
	NPF_IPv4UC_CallbackType_t type;
	NPF_boolean_t allOK;
	NPF_uint32_t numResp;
	NPF_IPv4UC_AsyncResponse_t *resp; 
} NPF_IPv4UC_CallbackData_t;




/*-----------------------------------------------------------------
*
* Event Notification Data Types 
*
*-----------------------------------------------------------------*/


/*
* Event Notification Types 
*/

typedef enum NPF_IPv4UC_Event {
	NPF_IPV4UC_PREFIX_TBL_MISS = 1,
	NPF_IPV4UC_NEXT_HOP_TBL_MISS = 2,
	NPF_IPV4UC_ADD_RES_TBL_MISS = 3, 
	NPF_IPV4UC_FIB_PREFIX_MISS = 4,
	NPF_IPV4UC_FWDTBL_REFRESH = 5
} NPF_IPv4UC_Event_t;


/* This event is triggered when the forwarding plane is unable to find a */ 
/* next hop identifier for a specific prefix. This event is optional. */

typedef struct {
	NPF_IPv4UC_PrefixTableHandle_t pfxTableHandle;
	NPF_IPv4Address_t destIP_Address;
} NPF_IPv4UC_PrefixTblMiss_t; 

/* This event is triggered when the forwarding plane is unable to find a */
/* next hop table entry for a specific next hop identifier. This event */
/* is optional. */

typedef struct { 
	NPF_IPv4UC_NextHopTableHandle_t nextHopTableHandle;
	NPF_uint32_t nextHopIdentifier;
} NPF_IPv4UC_NextHopTblMiss_t;

/* This event is triggered when the forwarding plane is unable to find a */ 
/* FIB table entry for a specific IP address. This event is optional */

typedef struct {
	NPF_IPv4UC_FibTableHandle_t fibTableHandle;
	NPF_IPv4Address_t destIP_Address;
} NPF_IPv4UC_FIB_PrefixMiss_t; 

/* This event is triggered when the forwarding plane is unable to find an*/
/* address resolution entry for a specific next hop. This event is */
/* optional. */

typedef struct { 
	NPF_IPv4UC_AddResTableHandle_t addResTableHandle;
	NPF_IfHandle_t interfaceHandle;
	NPF_IPv4Address_t IP_Address;
} NPF_IPv4UC_AddResTblMiss_t;

/*
* This structure defines the enumerations for the table type used in
* the NPF_IPv4UC_FwdTbl_Refresh_t structure above.
*/ 

typedef enum NPF_IPv4UC_TableType {
	NPF_IPV4UC_FIB_TABLE = 1,
	NPF_IPV4UC_PREFIX_TABLE = 2
}NPF_IPv4UC_TableType_t;


/* This event is triggered when the application or the IPv4 API */
/* implementation needs to be notified that a FIB needs t be refreshed */
/* on the forwarding plane. This event is optional. */

typedef struct {
	NPF_IPv4UC_TableType_t tableHandleType; 
	union {
		NPF_IPv4UC_FibTableHandle_t fibTableHandle;
		NPF_IPv4UC_PrefixTableHandle_t prefixTableHandle;
	} u;
} NPF_IPv4UC_FwdTbl_Refresh_t; 

/*
* Event Notification Structures
*/

typedef struct {
	NPF_IPv4UC_Event_t type; 
	union {
		NPF_IPv4UC_PrefixTblMiss_t prefixTblMiss;
		NPF_IPv4UC_NextHopTblMiss_t nextHopTblMiss;
		NPF_IPv4UC_AddResTblMiss_t addResTblMiss;
		NPF_IPv4UC_FIB_PrefixMiss_t fibPrefixMiss; 
		NPF_IPv4UC_FwdTbl_Refresh_t fwdTableRefreshRequest;
	} u;
} NPF_IPv4UC_EventData_t;


/* 
* This structure is provided when the event notification handler
* is invoked. It specifies one or more IPv4 unicast forwarding events.
*/

typedef struct { 
	NPF_uint32_t numEvents;
	NPF_IPv4UC_EventData_t *eventArray;
} NPF_IPv4UC_EventArray_t;


#ifdef __cplusplus
}
#endif
#endif /* __NPF_IPV4_TYPES_H */ 


