/*****************************************************************************
*                  INTEL CORPORATION PROPRIETARY INFORMATION
*  This software is supplied under the terms of a license agreement or
*  nondisclosure agreement with Intel Corporation and may not be copied
*  or disclosed except in accordance with the terms of that agreement.
*      Copyright (c) 2001 Intel Corporation. All Rights Reserved.
*****************************************************************************/
 
 
/*****************************************************************************
* $Source: /CODEBASE/npf_api/npf_ns_api.h
* $Revision: 1.6 $
* $Date: 2003/11/17 16:58:00 $
*
* Description:
*    The namespace API as part of NPF APIs
*
*****************************************************************************/
#ifndef _NPF_NS_API_H_
#define _NPF_NS_API_H_

#ifdef __cplusplus
extern "C" {
#endif 

#include "npf_common.h"
#include "npf_ns_types.h"

/*****************************************************************************
*   Function: npf_ns_create()
*
*   Parameters:
*       pathStr - null-terminated path string of the node to be created
*       nodeType - type of the node to be created: type, instance, or alias
*       dataType - data type to be refed by this node
*       hInstance - namepsace handle to the instance node for alias
*       hNode - the output namespace handle to the node created
*
*   Description: create a namespace node in open mode. The caller must 
*       close the handle after using it.
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_NS_NODE_EXISTS if a node already exists at the path
*           NPF_OUT_OF_MEMORY if the system is out of memory
*           NPF_COMPONENT_UNINITIATED if namespace is not initialized
*           NPF_SUCCESS if successful
*
*****************************************************************************/ 
NPF_RET npf_ns_create(char *         pathStr,
                      NPF_NSNODETYPE nodeType,
                      NPF_DATATYPE   dataType,
                      NPF_NSHDL      hInstance,
                      NPF_NSHDL *    hNode);

/*****************************************************************************
*
*   Function: npf_ns_alias()
*
*   Parameters:
*       canonName - null-terminated canonical path string
*       aliasName - null-terminated alias path string
*
*   Description: create an alias of a namespace node in close mode
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_NS_NODE_EXISTS if the alias exists
*           NPF_OUT_OF_MEMORY the system is out of memory
*           NPF_COMPONENT_UNINITIATED if namespace is not initialized
*           NPF_SUCCESS if successful
*
*****************************************************************************/ 
NPF_RET npf_ns_alias(char * canonName,
                     char * aliasName);

/*****************************************************************************
*
*   Function: npf_ns_delete()
*
*   Parameters:
*       name - null-terminated path string of a namespace node to be removed
*
*   Description: remove a namespace node at the given path. This node can
*       be removed if there is no reference to it and its descedants.
*       If the node is removed, all of its descedants are also removed.
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_NS_NODE_IN_USE if the node is in use
*           NPF_COMPONENT_UNINITIATED if namespace is not initialized
*           NPF_SUCCESS if successful
*
*****************************************************************************/ 
NPF_RET npf_ns_delete(char * name);


/*****************************************************************************
*
*   Function: npf_ns_rename()
*
*   Parameters:
*       oldName - null-terminated old path string of a namespace node
*       newName - null-terminated new path string of the namespace node
*
*   Description: rename a namespace node, which causes moving a namespace
*        node from one location on the hierachy to another location
*
*   Return: NPF_NS_NODE_EXISTS if a node already exists
*           NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_OUT_OF_MEMORY if the system is out of memory
*           NPF_COMPONENT_UNINITIATED if namespace is not initialized
*           NPF_SUCCESS if successful
*
*****************************************************************************/ 
NPF_RET npf_ns_rename(char * oldName,
                      char * newName);

/*****************************************************************************
*
*   Function: npf_ns_open()
*
*   Parameters:
*       name - null-terminated path string of a namespace node
*       hNode - output namespace handle to the node
*
*   Description: open a namespace  node, which causing the ref count to
*       the node increamented by 1.  The caller must close it after using it.
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIATED if namespace is not initialized
*           NPF_SUCCESS if successful
*
*****************************************************************************/ 
NPF_RET npf_ns_open(char *      name,
                    NPF_NSHDL * hNode);


/*****************************************************************************
*
*   Function: npf_ns_getPath()
*
*   Parameters:
*       hNode - an open namespace handle to a node
*       path - null-terminated path string, allocated by caller
*
*   Description: Get the path string of a namespace node from an open handle.
*       The caller must allocate the memory of size NPF_NS_MAXPATH for
*       path string before calling this function. 
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIATED if namespace is not initialized
*           NPF_OUT_OF_MEMORY if the system run out of memory
*           NPF_SUCCESS if successful
*
*****************************************************************************/
NPF_RET npf_ns_getPath(NPF_NSHDL hNode,
                       char * path);

/*****************************************************************************
*
*   Function: npf_ns_getDataType()
*
*   Parameters:
*       hNode - an open namespace handle
*       dataType - output data type of the node
*
*   Description: get the data type referenced by the node from an open
*       handle.   
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIATED if namespace is not initialized
*           NPF_SUCCESS if successful
*
*****************************************************************************/
NPF_RET npf_ns_getDataType(NPF_NSHDL      hNode,
                           NPF_DATATYPE * dataType);


/*****************************************************************************
*
*   Function: npf_ns_getDataHandle()
*
*   Parameters:
*       hNode - an open namespace handle to an instance node
*       hData - output data handle
*
*   Description: get the data handle referenced by the node
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIATED if namespace is not initialized
*           NPF_SUCCESS if successful
*
*****************************************************************************/
NPF_RET npf_ns_getDataHandle(NPF_NSHDL     hNode,
                             NPF_HANDLE  * hData);


/*****************************************************************************
*
*   Function: npf_ns_getInstParent()
*
*   Parameters:
*       hNode - an open namespace handle to an instance node
*       hInstParent - output handle to the instant parent node
*
*   Description: For an instant node, get an open handle to the instant
*       parent node, which is its grand parent in the namespace hierarchy.
*       The caller must close this handle after using it.  
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIATED if namespace is not initialized
*           NPF_SUCCESS if successful
*
*****************************************************************************/
NPF_RET npf_ns_getInstParent(NPF_NSHDL   hNode,
                             NPF_NSHDL * hInstParent);


/*****************************************************************************
*
*   Function: npf_ns_getParentNode()
*
*   Parameters:
*       hNode - an open namespace handle to an instance node
*       hParent - output open handle to the parent node
*
*   Description: Get an open handle to the parent node in the namespace
*       hierarchy. The caller must close this handle after using it.   
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIATED if namespace is not initialized
*           NPF_SUCCESS if successful
*
*****************************************************************************/
NPF_RET npf_ns_getParentNode(NPF_NSHDL   hNode,
                             NPF_NSHDL * hParent);
                                            

/* the sub API for node iterator */
/*****************************************************************************
*
*   Function: npf_ns_first()
*
*   Parameters:
*		nodeItr - a node iterator
*
*   Description: Get an open namespec handle to the first child in the
*       directory. The caller must close this handle after using it.
*
*   Return: The handle to the first child, or NULL if none.
*
*****************************************************************************/
NPF_NSHDL npf_ns_first(NODE_ITERATOR nodeItr);


/*****************************************************************************
*
*   Function: npf_ns_next()
*
*   Parameters:
*		nodeItr - a node iterator
*
*   Description: Get an open namespace handle to the next child in the
*       directory. The caller must close this handle after using it.
*
*   Return: The handle to the next child, or NULL if none.
*
*****************************************************************************/
NPF_NSHDL npf_ns_next(NODE_ITERATOR nodeItr);


/*****************************************************************************
*
*   Function: npf_ns_prev()
*
*   Parameters:
*		nodeItr - a node iterator
*
*   Description: Get an open namespace handle to the previous child in
*       the directory. The caller must close this handle after using it.
*
*   Return: The handle to the previous child, or NULL if none.
*
*****************************************************************************/
NPF_NSHDL npf_ns_prev(NODE_ITERATOR nodeItr);


/*****************************************************************************
*
*   Function: npf_ns_count()
*
*   Parameters:
*		nodeItr - a node iterator
*
*   Description: Get the count of children in the directory.
*
*   Return: The count if succesful, -1 otherwise.
*
*****************************************************************************/ 
int       npf_ns_count(NODE_ITERATOR nodeItr);


/*****************************************************************************
*
*   Function: npf_ns_done()
*
*   Parameters:
*		nodeItr - a node iterator
*
*   Description: Done with this directory handle
*
*   Return: none
*
*****************************************************************************/
void      npf_ns_done(NODE_ITERATOR nodeItr); 


/*****************************************************************************
*
*   Function: npf_ns_readDir()
*
*   Parameters:
*       name - the null-terminated path string of a node
*       dirItr - node iterator for the directory
*
*   Description: read the directory part of the node
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIATED if namespace is not initialized
*           NPF_OUT_OF_MEMORY if the system is out of memory
*           NPF_SUCCESS if successful
*
*****************************************************************************/ 
NPF_RET npf_ns_readDir(char *      name,
                       NODE_ITERATOR * dirItr);

/*****************************************************************************
*
*   Function: npf_ns_addAssociate()
*
*   Parameters:
*       hNode - an open namespace handle to a node
*       associate - a namespace node to be associated
*
*   Description: add an associate to the node
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIATED if namespace is not initialized
*           NPF_SUCCESS if successful
*
*****************************************************************************/
NPF_RET npf_ns_addAssociate(NPF_NSHDL hNode, 
                            NPF_NSHDL associate);

/*****************************************************************************
*
*   Function: npf_ns_delAssociate()
*
*   Parameters:
*       hNode - an open namespace handle to a node
*       associate - handle to the accociated node
*
*   Description: delete an associate from the node
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIATED if namespace is not initialized
*           NPF_SUCCESS if successful
*
*****************************************************************************/
NPF_RET npf_ns_delAssociate(NPF_NSHDL hNode, 
                            NPF_NSHDL associate);


/*****************************************************************************
*
*   Function: npf_ns_enumAssociate()
*
*   Parameters:
*       hNode - an open namespace handle to a node
*       associateItr - the output iterator of associates
*
*   Description: enumerate the associates of the node
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIATED if namespace is not initialized
*           NPF_OUT_OF_MEMORY if the system is out of memory
*           NPF_SUCCESS if successful
*
*****************************************************************************/
NPF_RET npf_ns_enumAssociate(NPF_NSHDL hNode, 
                             NODE_ITERATOR * associateItr); 


/*****************************************************************************
*
*   Function: npf_ns_close()
*
*   Parameters:
*       hNode - a namespace node handle
*
*   Description: Close this handle to the namespace node. This causes
*       the internal ref count decreamented by 1.
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIATED if namespace is not initialized
*           NPF_SUCCESS if successful
*
*****************************************************************************/ 
NPF_RET npf_ns_close(NPF_NSHDL hNode);


/*****************************************************************************
*
*   Function: npf_ns_getCanonName()
*
*   Parameters:
*       nameStr - null-terminated path string to a node
*       canonName - null-terminated canonical path string, allocated by
*                   caller
*
*   Description: Get the canonical path string of a namespace node, typically,
*        an alias node. The caller must allocate the memory of size
*        NPF_NS_MAXPATH for canonName string before calling this function.
*
*   Return: NPF_INVALID_PARAMETERS if an input parameter is invalid
*           NPF_COMPONENT_UNINITIATED if namespace is not initialized
*           NPF_SUCCESS if successful
*
*****************************************************************************/ 
NPF_RET npf_ns_getCanonName(char * nameStr,
                            char * canonName);

#ifdef __cplusplus
}
#endif 

#endif /* _NPF_NS_API_H_ */   
