/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation.All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 5
 *
 * = FILENAME
 *      cc.h
 *
 * = DESCRIPTION
 *      Public API functions for core components and message/packet handlers
 *
 * = AUTHOR
 *       Roland Arajs
 *       roland.arajs@intel.com
 *
 * = CHANGE HISTORY
 *      4/8/2002 1:50:41 PM - creation time
 *
 * ============================================================================
 * $Id: cc.h,v 1.9 2003/08/20 21:53:00 ageorges Exp $
 */

#if !defined(__CC_H__)
#define __CC_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */

#include "ix_ossl.h"
#include "ix_rm.h"
#include "cci/cci_types.h"


/**
 * TYPENAME: ix_cc_get_instance
 * 
 * DESCRIPTION: Function signature for user-defined function that will determine
 *      the CC instance based on passed to ix_cci_cc_create() function. There will be
 *      a global instance of this function as a support for CC multiple instance facility.
 *      The applications will have to register/unregister their own function prior to the
 *      invokation of ix_cci_cc_create(), as the code will invoke the global instance of
 *      the function which will be responsible on retrieving the CC instance based on the
 *      passed context to the ix_cci_cc_create() call.
 *
 * @Param:  - IN arg_pContext: context pointer that is passed to the ix_cci_cc_create()
 *      function.
 */
typedef ix_uint32 (*ix_cc_get_instance)( void* arg_pContext );


/**
 * Global function pointer for CC instance retrieval function
 */
extern ix_cc_get_instance g_CCGetInstanceFPtr;


/**
 * NAME: _ix_rm_cci_cc_default_get_instance
 *
 * DESCRIPTION: This function will retrieve the default instance of a CC. For now
 *  the function will return 0 at all times.
 * 
 * @Param:  - IN arg_pContext - context passed to the CC creation function
 *
 * @Return: This function returns 0 at all times.
 */
ix_uint32 _ix_rm_cci_cc_default_get_instance(
                                              void* arg_pContext
                                            );

/**
 * MACRO NAME: IX_CCI_CC_GET_INSTANCE_REGISTER
 *
 * DESCRIPTION: Ths macro will register the active CC get instance function. If another function
 *  is already registered (different from the CCI default one) the macro will return an error (0).
 *  If NULL is passed to the macro then the macro will return 0. In the case of success the macro will 
 *  return a non 0 value.
 *
 * @Param:  - IN arg_pCCGetInstanceFPtr - a function pointer of ix_cc_get_instance type to be registered
 *  with the system.
 *
 * @Return: Non 0 if successful or 0 otherwise
 */
#define IX_CCI_CC_GET_INSTANCE_REGISTER( \
                                         arg_pCCGetInstanceFPtr \
                                       ) \
                    ((g_CCGetInstanceFPtr != _ix_rm_cci_cc_default_get_instance) ? 0 : \
                    ((arg_pCCGetInstanceFPtr == 0) ? 0 : \
                    ((g_CCGetInstanceFPtr = arg_pCCGetInstanceFPtr), 1)))


/**
 * MACRO NAME: IX_CCI_CC_GET_INSTANCE_UNREGISTER
 *
 * DESCRIPTION: Ths macro will unregister the active CC get instance function by restoring the default
 *  CCI function
 *
 */
#define IX_CCI_CC_GET_INSTANCE_UNREGISTER( \
                                         ) \
                    (g_CCGetInstanceFPtr = _ix_rm_cci_cc_default_get_instance)
                                       
/**
 * TYPENAME: ix_cc_init
 * 
 * DESCRIPTION: Function signature for user-defined core component 
 *              initialization function. This function is passed into and run
 *              by the ix_cci_cc_create() function.
 *
 * @Param:  - INOUT arg_ppContext: In: Pointer to context provided by 
 *                  ix_cci_cc_create().
 *                  Out: Pointer to output context that will be passed
 *                  to the user-defined ix_cc_fini function, which will be called
 *                  by ix_cci_cc_destroy(). The user is responsible for 
 *                  dynamically allocating memory for this context within this
 *                  function. This context can be used for storing pointers to 
 *                  resources that will be shared between the core component and 
 *                  any corresponding microblock(s).
 */
typedef ix_error (*ix_cc_init)( ix_cc_handle hCC, void** ppContext );


/**
 * TYPENAME: ix_cc_fini
 * 
 * DESCRIPTION: Function signature for user-defined core component termination
 *              function. This function is passed into the ix_cc_create() 
 *              function and run during ix_cc_destroy().
 *
 * @Param:  - OUT arg_pContext: Pointer to context allocated and initialized
 *              by user-defined initialization function of type ix_cc_init.
 */
typedef ix_error (*ix_cc_fini)( ix_cc_handle hCC, void* arg_pContext );

/**
 * TYPENAME: ix_event_func
 * 
 * DESCRIPTION: Function signature for user-defined event function that is
 *              registered in the infrastructure using function
 *              ix_cci_cc_add_event_handler().
 *
 * @Param:  - IN arg_pContext: Pointer to context of CC or engine calling 
 *                  ix_cci_cc_add_event_handler().
 */
typedef ix_error (* ix_event_func)( void * arg_pContext );


/**
 * TYPENAME: ix_msg_handler
 * 
 * DESCRIPTION: Function signature for user-defined message handler.
 *
 * @Param:  - IN arg_hDataToken. Message buffer passed to the handler if the
 *                  handler is the target of an ix_cci_send_message() 
 *                  function.
 * @Param:  - IN arg_UserData. Data passed to the handler by 
 *                  ix_cci_send_message().
 *
 * @Param:  - IN arg_pComponentContext. User-defined context set by function
 *                  ix_cci_cc_add_message_handler().
 *
 */
typedef ix_error (* ix_msg_handler) ( ix_buffer_handle arg_hDataToken,
                                      ix_uint32 arg_UserData,                    
                                      void* arg_pComponentContext );

/**
 * TYPENAME: ix_pkt_handler
 * 
 * DESCRIPTION: Function signature for user-defined message handler.
 *
 * @Param:  - IN arg_hDataToken. Packet buffer passed to the handler if the
 *                  handler is the target of an ix_cci_send_packet() 
 *                  function.
 * @Param:  - IN arg_ExceptionCode. Exception data passed to the handler, along
 *                  with the buffer handle, by ix_cci_send_packet().
 *
 * @Param:  - IN arg_pComponentContext. User-defined context set by function
 *                  ix_cci_cc_add_packet_handler().
 *
 */
typedef ix_error (* ix_pkt_handler) ( ix_buffer_handle arg_hDataToken,
                                      ix_uint32 arg_Exception,                    
                                      void* arg_pComponentContext );

/**
 * TYPENAME: ix_event_type
 * 
 * DESCRIPTION: Specifies the type of event handler registered using 
 *              ix_cci_cc_add_event_handler().
 *
 */
typedef enum ix_e_event_type
{
    IX_EVENT_TYPE_FIRST = 0,
    IX_EVENT_TYPE_PERIODIC = IX_EVENT_TYPE_FIRST,
    IX_EVENT_TYPE_ONESHOT,
    IX_EVENT_TYPE_LAST
} ix_event_type;


/**
 * TYPENAME: ix_input_type
 * 
 * DESCRIPTION: Specifies whether an input ID will receive input from a single
 *              source or from multiple sources. Inputs from multiple sources
 *              requires the infrastructure to lock the message/packet input 
 *              queue while the incoming buffer handle and user/exception data
 *              are being written to the queue.
 *
 */
typedef enum ix_e_input_type
{
    IX_INPUT_TYPE_FIRST = 0,
    IX_INPUT_TYPE_SINGLE_SRC = IX_INPUT_TYPE_FIRST,
    IX_INPUT_TYPE_MULTI_SRC,
    IX_INPUT_TYPE_LAST
} ix_input_type;





/**
 * NAME: ix_cci_cc_create
 *
 * DESCRIPTION: Creates a core component. This function should be called from 
 *              the user-defined init function in the core component's execution 
 *              engine.
 * 
 * @Param:  - IN arg_EngHandle: Stored in the core component so that functions 
 *                  ix_cci_cc_add_packet_handler() and 
 *                  ix_cci_cc_add_message_handler() can be automatically added 
 *                  to the default policy of the execution engine.
 * @Param:  - IN arg_InitFunc: Pointer to initialization function that will be 
 *                  run. This function will create resources shared between the 
 *                  core and any microblocks with which the component will 
 *                  interact. It may also patch symbols to those microblocks. 
 * @Param:  - IN arg_pContextIn: Pointer to context that will be passed into
 *                  arg_InitFunc.
 * @Param:  - IN arg_FiniFunc: Pointer to component termination function. 
 *
 * @Return: IX_SUCCESS if successful or an ix_error token encapsulating one of 
 *          the following error codes:
 *              IX_CCI_ERR_OS_MEM_ALLOC not enough memory 
 *              Any errors returned by ix_cc_init callback function.
 *              IX_CCI_ASSERTION invalid function parameter (debug compilations 
 *                  only)
 */
IX_EXPORT_FUNCTION 
ix_error ix_cci_cc_create(
                           ix_exe_handle arg_EngHandle,
                           ix_cc_init arg_InitFunc,
                           ix_cc_fini arg_FiniFunc,
                           void* arg_pContextIn,
                           ix_cc_handle* arg_phComponent
                         );



/**
 * NAME: ix_cci_cc_destroy
 *
 * DESCRIPTION: Destroys the component. The function will perform internal 
 *              cleanup and invoke the user-defined fini() function associated 
 *              with the component. This function should be called from the 
 *              user-defined fini function for the core component's execution 
 *              engine.
 * 
 * @Param:  - IN arg_hComponent: Handle to core component originally initialized 
 *                  using ix_cci_cc_create().
 *
 * @Return: IX_SUCCESS if successful or an ix_error token encapsulating one of 
 *          the following error codes:
 *              Any errors returned by the ix_cc_fini callback function.
 *	            IX_CCI_ASSERTION invalid function parameter (debug compilations 
 *                  only)
 */
IX_EXPORT_FUNCTION 
ix_error ix_cci_cc_destroy( ix_cc_handle arg_hComponent );



/**
 * NAME: ix_cci_cc_add_event_handler
 *
 * DESCRIPTION: Adds an event handler to a core component. This function should 
 *              be called within the user's core-component code.
 *
 * NOTE:        The execution engine only checks for events between the handling 
 *              of messages and packets. Therefore, if a handler takes a long 
 *              time to process a packet/message, the event will be triggered 
 *              considerably later (or less often) than requested in the arg_ms 
 *              parameter. The event function is called in the thread context of 
 *              the execution engine. Event functions take strict priority over 
 *              message and packet handlers. 
 *
 * @Param:  - IN arg_hComponent. Handle to core component or execution engine.
 * @Param:  - IN arg_ms: The minimum number of milliseconds from now when event 
 *                  will be triggered.
 * @Param:  - IN arg_EventFunc: Pointer to user-defined event handler.
 * @Param:  - IN arg_EventType: Set to IX_EVENT_TYPE_PERIODIC for periodic 
 *                  events or IX_EVENT_TYPE_ONESHOT otherwise.
 * @Param:  - IN arg_Priority: If the event scheduler sees two or more events 
 *                  that are scheduled to occur at the same time, this value 
 *                  prioritizes between them.
 * @Param:  - OUT arg_phEvent: Pointer to event handle, used for removing event 
 *                  handlers.
 *
 * @Return: IX_SUCCESS if successful or an ix_error token encapsulating one of the
 *          following error codes:
 *	            IX_CCI_ERR_OS_MEM_ALLOC not enough memory for internal event 
 *                  structure.
 *              IX_CCI_ERR_CC_ADD_HANDLER_INVALID_HANDLE invalid handle passed 
 *                  in first argument.
 *              IX_CCI_ERR_EVENT_HANDLER_GET_TIME error getting system time.
 *              IX_CCI_ASSERTION invalid function parameter (debug compilations 
 *                  only)
 */
IX_EXPORT_FUNCTION 
ix_error ix_cci_cc_add_event_handler(
                                      ix_cc_handle   arg_hComponent,
                                      ix_uint32 arg_ms,
                                      ix_event_func arg_EventFunc,
                                      ix_event_type arg_EventType,
                                      ix_uint32 arg_Priority,
                                      ix_event_handle* arg_phEvent
                                    );


/**
 * NAME: ix_cci_cc_add_event_handler_ex
 *
 * DESCRIPTION: Simlar to ix_cci_cc_add_event_handler(), but this function allows
 *              the user to specify the event's own context instead of the CC/engine 
 *              context.
 *
 * @Param:  - IN arg_hComponent. Handle to core component or exeuction engine.
 * @Param:  - IN arg_ms: The minimum number of milliseconds from now when event 
 *                  will be triggered.
 * @Param:  - IN arg_EventFunc: Pointer to user-defined event handler.
 * @Param:  - IN arg_EventType: Set to IX_EVENT_TYPE_PERIODIC for periodic 
 *                  events or IX_EVENT_TYPE_ONESHOT otherwise.
 * @Param:  - IN arg_Priority: If the event scheduler sees two or more events 
 *                  that are scheduled to occur at the same time, this value 
 *                  prioritizes between them.
 * @Param:  - IN arg_EventContext: Context to be passed to event when the event
 *                  occurs.
 * @Param:  - OUT arg_phEvent: Pointer to event handle, used for removing event 
 *                  handlers.
 *
 * @Return: IX_SUCCESS if successful or an ix_error token encapsulating one of the
 *          following error codes:
 *	            IX_CCI_ERR_OS_MEM_ALLOC not enough memory for internal event 
 *                  structure.
 *              IX_CCI_ERR_CC_ADD_HANDLER_INVALID_HANDLE invalid handle passed 
 *                  in first argument.
 *              IX_CCI_ERR_EVENT_HANDLER_GET_TIME error getting system time.
 *              IX_CCI_ASSERTION invalid function parameter (debug compilations 
 *                  only)
 */
IX_EXPORT_FUNCTION 
ix_error ix_cci_cc_add_event_handler_ex(
                                         ix_cc_handle   arg_hComponent,
                                         ix_uint32 arg_ms,
                                         ix_event_func arg_EventFunc,
                                         ix_event_type arg_EventType,
                                         ix_uint32 arg_Priority,
                                         void* arg_EventContext,
                                         ix_event_handle* arg_phEvent
                                       );



/**
 * NAME: ix_cci_change_event
 *
 * DESCRIPTION: Changes the period of a periodic event
 *
 * @Param:  - IN arg_hEvent. Handle to event.
 * @Param:  - IN arg_ms: The new event period.
 *
 * @Return: IX_SUCCESS if successful or an ix_error token encapsulating one of the
 *          following error codes:
 *              IX_CCI_ERROR_ASSERTION invalid handle passed 
 *                  in first argument or event not periodic (debug compilations only).
 */

IX_EXPORT_FUNCTION
ix_error ix_cci_change_event(
                              ix_event_handle arg_hEvent,
                              ix_uint32 arg_ms
                            );


/**
 * NAME: ix_cci_cc_remove_event_handler
 *
 * DESCRIPTION: Removes an event handler from a core component. This function 
 *              should be called within the user's core-component code and is 
 *              only required for removing periodic events and unexpired 
 *              one-shot events.
 *
 * NOTE:        It is not necessary to remove a one-shot event handler after it 
 *              has expired. On the other hand, although this function checks 
 *              that the input parameters are valid, attempting to remove an 
 *              expired event will not cause the function to return an error.
 * 
 * @Param:  - IN arg_hEvent: Handle of event to be removed.
 *
 * @Return: IX_SUCCESS if successful or an ix_error token encapsulating one of 
 *          the following error codes:
 *	            IX_CCI_ASSERTION invalid function parameter (debug compilations 
 *                  only).
 *              IX_CCI_ERR_EVENT_NOT_FOUND no trace  of event found, even on 
 *                  expired-event list (debug compilations only) 
 */
IX_EXPORT_FUNCTION 
ix_error ix_cci_cc_remove_event_handler(
                                         ix_cc_handle arg_hComponent,                 
                                         ix_event_handle arg_hEvent
                                       );


/**
 * NAME: ix_cci_cc_add_message_handler
 *
 * DESCRIPTION: Add a message handler to a core component and associate it with 
 *              an input ID. This function should be called in the init function 
 *              of the core component's execution engine. Attempting to 
 *              associate a handler to an ID that already is associated with a 
 *              handler will cause the function to return an error.
 * 
 * @Param:  - IN arg_hComponent, used to gain access to the execution engine, 
 *                  and thereby to the execution engine's policy, so that the 
 *                  message handler can be added to a default policy. 
 * @Param:  - IN arg_InputID: Input ID to be associated with the handler.
 * @Param:  - IN arg_Handler: Pointer to user-defined message handler
 * @Param:  - IN arg_SourceType. Enumerated type to indicate whether the input 
 *                  will be receiving messages from one source or multiple 
 *                  sources. Valid values are: IX_INPUT_TYPE_SINGLE_SRC and 
 *                  IX_INPUT_TYPE_MULTI_SRC. This tells the framework to lock 
 *                  when writing data into the queue.
 *
 * @Return: IX_SUCCESS if successful or an ix_error token encapsulating one of 
 *          the following error codes:
 *              IX_CCI_ERR_OS_MEM_ALLOC could not allocate enough memory for 
 *                  internal handler structure.
 *              IX_CCI_TOKEN_PROCESSOR_CONTAINER_FULL attempted to add handler 
 *                  to a full policy that cannot grow.
 *              IX_CCI_ERR_TKP_CONTAINER_MEM_ALLOC memory allocation error when 
 *                  attempting to add handler to a full policy.
 *              IX_CCI_ERR_MSG_SET_MODE error reported by Resource Manager when 
 *                  setting handler's mode of operation.
 *              IX_CCI_ASSERTION invalid function parameter (debug compilations 
 *                  only).
 */
IX_EXPORT_FUNCTION 
ix_error ix_cci_cc_add_message_handler(
                                        ix_cc_handle arg_hComponent,
                                        ix_uint32 arg_InputID,
                                        ix_msg_handler arg_Handler,
                                        ix_input_type arg_SourceType
                                      );


/**
 * NAME: ix_cci_cc_remove_message_handler
 *
 * DESCRIPTION: Removes a message handler previously added with 
 *              ix_cci_cc_add_message_handler(). This function should be called 
 *              from the user-defined fini() function in the core component's 
 *              execution engine. It must be called before calling 
 *              ix_cci_cc_destroy() on the associated core component.
 * 
 * @Param:  - IN arg_InputID: Input ID associated with the message handler to be 
 *                  removed. After calling this function, any messages arriving 
 *                  on this input ID will be dropped.
 * @Return: IX_SUCCESS if successful or an ix_error token encapsulating one of 
 *          the following error codes:
 *              IX_CCI_ERR_DISABLE_MSG_HANDLER error disabling message handler 
 *                  in Resource Manager.
 *              IX_CCI_ASSERTION invalid function parameter (debug compilations 
 *                  only).
 */
IX_EXPORT_FUNCTION 
ix_error ix_cci_cc_remove_message_handler( ix_uint32 arg_InputID );




/**
 * NAME: ix_cci_cc_add_packet_handler
 *
 * DESCRIPTION: Add a packet handler to a core component and associate it with 
 *              an input ID. This function should be called in the init function 
 *              of the core component's execution engine. Attempting to 
 *              associate a handler to an ID that already is associated with a 
 *              handler will cause the function to return an error.  
 * 
 * @Param:  - IN arg_hComponent: used to gain access to the execution engine, 
 *                  and thereby to the execution engine's policy, so that the 
 *                  packet handler can be added to a default policy. 
 * @Param:  - IN arg_Handler: Pointer to user-defined packet handler function.
 * @Param:  - IN arg_InputID: Input ID to be associated with the handler. 
 * @Param:  - IN arg_SourceType. Enumerated type to indicate whether the input 
 *                  will be receiving packets from one source or multiple 
 *                  sources. Valid values are: IX_INPUT_TYPE_SINGLE_SRC and 
 *                  IX_INPUT_TYPE_MULTI_SRC. This tells the framework to lock 
 *                  when writing data into the queue.
 *
 * @Return: IX_SUCCESS if successful or an ix_error token encapsulating one of 
 *          the following error codes:
 *              IX_CCI_ERR_OS_MEM_ALLOC could not allocate enough memory for 
 *                  internal handler structure.
 *              IX_CCI_TOKEN_PROCESSOR_CONTAINER_FULL attempted to add handler 
 *                  to a full policy that cannot grow.
 *              IX_CCI_ERR_TKP_CONTAINER_MEM_ALLOC memory allocation error when 
 *                  attempting to add handler to a full policy.
 *              IX_CCI_ERR_MSG_SET_MODE error reported by Resource Manager when 
 *                  setting handler's mode of operation.
 *              IX_CCI_ASSERTION invalid function parameter (debug compilations 
 *                  only).
 */
IX_EXPORT_FUNCTION 
ix_error ix_cci_cc_add_packet_handler(
                                        ix_cc_handle arg_hComponent,
                                        ix_uint32 arg_InputID,
                                        ix_pkt_handler arg_Handler,
                                        ix_input_type arg_SourceType
                                     );

               		
/**
 * NAME: ix_cci_cc_remove_packet_handler
 *
 * DESCRIPTION: Removes a packet handler previously added with 
 *              ix_cci_cc_add_packet_handler(). This function should be called 
 *              from the user-defined fini() function in the core component's 
 *              execution engine. It must be called before calling 
 *              ix_cci_cc_destroy() on the associated core component.
 * 
 * @Param:  - IN arg_InputID: Input ID associated with the packet handler to be 
 *                  removed. After calling this function, any packets arriving 
 *                  on this input ID will be dropped.
 *
 * @Return: IX_SUCCESS if successful or an ix_error token encapsulating one of 
 *          the following error codes:
 *              IX_CCI_ERR_DISABLE_MSG_HANDLER error disabling message handler 
 *                  in Resource Manager.
 *              IX_CCI_ASSERTION invalid function parameter (debug compilations 
 *                  only).
 */
IX_EXPORT_FUNCTION 
ix_error ix_cci_cc_remove_packet_handler( ix_uint32 arg_InputID );


/**
 * NAME: ix_cci_get_message
 *
 * DESCRIPTION: Gets a message from the message queue identified by arg_Id.  
 *              This function is for use by user-space applications. If there 
 *              are no messages on the specified queue, the function will block.
 * 
 * @Param:  - IN arg_QueueId: Provides a queue ID. This ID must be in the same 
 *                  range as core-component message-input Ids, but must not 
 *                  correspond to any input ID associated with a core-component 
 *                  message handler. 
 * @Param:  - INOUT arg_pMsgData: Pointer to a buffer, initialized by the 
 *                  caller, where the function will return the message. If the 
 *                  size of the message buffer is not sufficient to hold the 
 *                  message, the function will return an error. 
 * @Param:  - INOUT arg_pMsgSize: As an input, specifies the size of the message 
 *                  buffer. As an output, specifies the size of the message 
 *                  stored in the message buffer. 
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_cci_get_message(
                             ix_uint32 arg_QueueId, 
                             ix_uint32* arg_pMsgData,
                             ix_uint32* arg_pMsgSize
                           );



/**
 * NAME: ix_cci_send_message
 *
 * DESCRIPTION: Sends a message from a core component's packet/message handler to 
 *              another core component. Microblocks to not support messages 
 *              being sent to them; messages sent to microblocks will be 
 *              dropped. The user may call this function at any time after the 
 *              core framework has been initialized, but it is typically called 
 *              from within a core component's packet or message handler, to 
 *              send a message to another core component (or even itself). The 
 *              infrastructure will also call this function, or an internal 
 *              equivalent, when a microblock sends a message to a core 
 *              component. If this function is called to send a message to an ID 
 *              that does not have a handler attached, a default message handler 
 *              will be called, which simply drops the message and frees the 
 *              message's shared-memory resources.  
 * 
 * @Param:  - IN arg_OutputId: Output ID where message is to be sent.
 * @Param:  - IN arg_Handle: Handle to message.
 * @Param:  - IN arg_UserData: User-defined variable that can be used by the 
 *                  receiving message handler for any purpose, such as 
 *                  identifying the message type.
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_cci_send_message(
                              ix_uint32 arg_OutputId,
                              ix_buffer_handle arg_Handle,
                              ix_uint32 arg_UserData
                            );


/**
 * NAME: ix_cci_send_packet
 *
 * DESCRIPTION: Sends a packet from a core component's packet/message handler  
 *              to another core component. The user may call this function at 
 *              any time after the core framework has been initialized, but it 
 *              is typically called from within a core component's packet or 
 *              message handler, to send a packet to another core component 
 *              (or even itself). The infrastructure will also call an internal 
 *              equivalent to this function when a microblock sends a packet
 *              to a core component. If this function is called to send a 
 *              packet to an ID that does not have a handler attached, a default 
 *              packet handler will be called, which simply drops the message 
 *              and frees the packet's shared-memory resources.  
 *  
 * NOTE:        The equivalent function invoked when a microblock sends a packet
 *              to a core component has a third parameter for sending an 
 *              exception code. When a packet handler receives a packet in 
 *              response to ix_cci_send_packet(), the exception code is always
 *              set to zero.
 * 
 * @Param:  - IN arg_OutputId: Output ID where packet is to be sent.
 * @Param:  - IN arg_Handle: Handle to packet.
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_cci_send_packet(
                             ix_uint32 arg_OutputId,
                             ix_buffer_handle arg_Handle
                           );




#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__CC_H__) */
