/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation.All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 5
 *
 * = LIBRARY
 *      Core Framework Library
 *
 * = MODULE
 *      Library Interface
 *
 * = FILENAME
 *      engine.h
 *
 * = DESCRIPTION
 *      
 *
 * = AUTHOR
 *       Roland Arajs
 *       roland.arajs@intel.com
 *
 * = CHANGE HISTORY
 *      4/8/2002 8:24:39 AM - creation time
 *
 * ============================================================================
 * $Id: engine.h,v 1.6 2003/03/04 19:30:25 ageorges Exp $
 */

#if !defined(__ENGINE_H__)
#define __ENGINE_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */

#include "ix_symbols.h"
#include "cci/cci_types.h"


/**
 * TYPENAME: ix_exe_init
 * 
 * DESCRIPTION: Function signature required for user-defined initialization 
 *              function. The idea of the function is to store pointers to
 *              core components, policies, etc in the context which will
 *              be passed to a user-defined ix_exe_fini function when the
 *              execution engine closes down. The memory for the context 
 *              pointed to by ppContext must be dynamically allocated by the 
 *              initialization function. The initialization function must 
 *              return the context pointer via ppContext so that it may be 
 *              stored by the execution engine. 
 */
typedef ix_error (*ix_exe_init)( ix_exe_handle arg_ExeHandle, 
                                 void** arg_ppContext );


/**
 * TYPENAME: ix_exe_fini
 * 
 * DESCRIPTION: Function signature required for user-defined termination 
 *              function. The function pointer is stored by API function
 *              ix_exe_run() and is executed by the ix_exe_shutdown() 
 *              function. 
 * @Param:  - IN arg_ExeHandle: Handle created by ix_exe_run() and passed
 *                  into this function by ix_exe_shutdown(). This handle
 *                  is used by all API functions related to the execution
 *                  engine.
 * @Param:  - IN arg_pContext: Context originally allocated and initialized
 *                  by the user-defined execution engine initialization 
 *                  function (see ix_exe_init above). This context should 
 *                  be used to store pointers to resources initialized in
 *                  the execution engine (such as core components).
 */
typedef ix_error (*ix_exe_fini)( ix_exe_handle ExeHandle, void* pContext );


/**
 * NAME: ix_cci_exe_run
 *
 * DESCRIPTION: Creates and runs an execution engine (thread of control). 
 *              This function should be called from the user's application 
 *              after ix_cci_cci_init() has been called to initialize the core 
 *              framework. Note that this function calls the user-defined init()
 *              function associated with this engine, and stores the output 
 *              context internally so that it can be passed to the user-defined 
 *              fini() function for this engine when ix_cci_exe_shutdown() is 
 *              called.
 * 
 * @Param:  - IN arg_FileName: If the operating environment is process-based 
 *                  and is required to spawn a new process for each execution 
 *                  engine, this parameter specifies the filename where the 
 *                  execution engine's process entry point is defined. 
 *                  Otherwise, the parameter should be set to NULL.
 * @Param:  - IN arg_InitFunc: Specifies a user-defined initialization function 
 *                  that will be called by ix_cci_exe_run(). This function is 
 *                  responsible for creating the execution-control tree and 
 *                  setting up the message and packet handlers. If the operating 
 *                  environment is process-based and is required to spawn a new 
 *                  process for each execution engine, this parameter should be 
 *                  set to NULL, and the initialization should be performed by 
 *                  the process' entry function.
 * @Param:  - IN arg_FiniFunc: Specifies a user-defined cleanup function that 
 *                  will be called in ix_cci_exe_run() as a result of 
 *                  ix_cci_exe_shutdown() being called in a control thread. 
 *                  After the cleanup function has returned the ix_cci_exe_run() 
 *                  function will terminate. If the operating environment is 
 *                  process-based and is required to spawn a new process for 
 *                  each execution engine, this parameter should be set to NULL, 
 *                  and the engine's fini pointer should be initialized by the 
 *                  process' entry function.
 * @Param:  - IN arg_Name: Name to be associated with the current engine.
 * @Param:  - OUT arg_pHandle: Pointer to new execution engine's handle. 
 *
 * @Return: IX_SUCCESS if successful or an ix_error token encapsulating one of 
 *          the following error codes:
 *              IX_CCI_ERR_ENG_INIT_SEMA_PROP_READ could not read registry 
 *                  properties for the global synchronization semaphore.
 *              IX_CCI_ERR_ENG_PROP_CLOSE error closing property for global 
 *                  semaphore.
 *              IX_CCI_ERR_ENG_INIT_ENGINE_PROP_READ could not read properties 
 *                  for shared-memory integers used by all execution engines.
 *              IX_CCI_ERR_ENG_PROP_CLOSE error closing property for shared-
 *                  memory integers.
 *              IX_CCI_ERR_ENG_THREAD_SPAWN error spawning engine thread.
 *              IX_CCI_ERR_OS_MEM_ALLOC memory allocation error.
 *              IX_CCI_ERR_TKP_CONTAINER_MEM_ALLOC memory allocation error.
 *              IX_CCI_ERR_ENG_THREAD_ID could not read current thread ID.
 *              IX_CCI_ERR_SHARED_MEM_ALLOC shared-memory allocation error.
 *              IX_CCI_ERR_ENG_INIT_SEMA_INIT could not initialize this engine's 
 *                  semaphore.
 *              Errors returned from ix_cci_policy_create(), 
 *                  ix_cci_policy_add_branch(), and ix_cci_exe_add_policy() 
 *                  while trying to create the default policy tree.
 *              Errors returned from the ix_exe_init-style callback function.
 *              IX_CCI_NO_SELECTED_CHILD no event/message/packet handler was 
 *                  added to the engine (neither directly, nor via one of its 
 *                  core components).
 *              IX_CCI_ASSERTION invalid function parameter (debug compilations 
 *                  only).
 */
IX_EXPORT_FUNCTION 
ix_error ix_cci_exe_run(
                         const char* arg_FileName,
                         ix_exe_init arg_InitFunc,
                         ix_exe_fini arg_FiniFunc,
                         const char* arg_Name,
                         ix_exe_handle* arg_pHandle
                       );


/**
 * NAME: ix_cci_exe_shutdown
 *
 * DESCRIPTION: Sets a flag in the execution engine, telling it to shut down. 
 *              Each execution engine will contain a semaphore that the engine 
 *              will unlock when it has finished processing and has executed the 
 *              execution engine's fini() function defined by the user. This 
 *              shutdown function will block until the execution engine 
 *              indicates shutdown is complete. This function should be called 
 *              in the user's application.
 * 
 * @Param:  - IN arg_Handle: Handle to execution engine originally initialized 
 *                  by ix_cci_exe_create().
 *
 * @Return: IX_SUCCESS if successful or an ix_error token encapsulating one of 
 *          the following error codes:
 *              Fatal errors returned by event/message/packet handlers	
 *              Errors returned by ixe_exe_fini-style callback function
 *              IX_CCI_ASSERTION invalid function parameter (debug compilations 
 *                  only).
 */
IX_EXPORT_FUNCTION 
ix_error ix_cci_exe_shutdown( ix_exe_handle arg_Handle );



/**
 * NAME: ix_cci_exe_add_policy
 *
 * DESCRIPTION: Adds a branch node (scheduling policy) to an execution engine. 
 *              This function should be called from the user-defined init() 
 *              function for the execution engine in order to replace the 
 *              exeucution engine's default policy tree with a different 
 *              scheduling policy or policy tree.
 *
 * NOTE:        This function, in conjunction with ix_cci_policy_add_branch() 
 *              and ix_cci_policy_add_leaf(), should only be called to replace 
 *              the default policy tree with a different policy or policy tree. 
 *              All message and packet handlers must be added to the execution 
 *              engine's core components before calling this function. Failure 
 *              to do so will result in an error being returned from the 
 *              add-handler function. Care should be taken that all IDs are 
 *              added to the new policy tree using ix_cci_policy_add_leaf(); 
 *              otherwise the handlers will not run when packets or messages 
 *              arrive on those IDs. 
 * 
 * @Param:  - IN arg_hParent: Execution engine where policy will be added.
 * @Param:  - IN arg_hChild: Policy to be added.
 * @Param:  - IN arg_NumCredits: The number of times a handler will be called 
 *                  before another handler is selected by the policy tree 
 *                  (assuming the handler's queue has enough entries and that 
 *                  no new messages/packets are received with higher priority).
 *
 * @Return: 0 if successful or an ix_error token encapsulating the following 
 *          error code:
 *              IX_CCI_ASSERTION invalid function parameter (debug compilations 
 *                  only).
 */
IX_EXPORT_FUNCTION 
ix_error ix_cci_exe_add_policy(
                                ix_exe_handle arg_hParent,
                                ix_policy_handle arg_hChild,
                                ix_uint32 arg_NumCredits
                              );


/**
 * NAME: ix_cci_exe_get_info
 *
 * DESCRIPTION: Gets the execution engine handle for the execution engine
 *              the caller is running in.  If this function is called from a 
 *              non-engine thread, this function will return information about 
 *              the default engine. The default engine is set by function 
 *              ix_cci_exe_set_default(). If this function has not been called, 
 *              the default engine will be engine 0.
 * 
 * @Param:  - OUT   arg_pExeHandle: Handle of engine running this thread.
 * @Param:  - OUT   arg_pEngineNumber: A unique engine number. The first 
 *                      execution engine created is engine number 0, the next is
 *                      engine number 1, etc.
 * @Param:  - OUT   arg_ppContext: Pointer to engine context initialized by
 *                      the user-defined ix_exe_init-style function passed into
 *                      ix_cci_exe_run() when the engine was created.
 *
 * @Return: IX_SUCCESS if successful or an ix_error token encapsulating the 
 *          following error code: 
 *              IX_CCI_ERR_ENG_THREAD_ID error getting current thread ID or 
 *                  default engine no longer exists.
 */
IX_EXPORT_FUNCTION 
ix_error ix_cci_exe_get_info( ix_exe_handle* arg_pExeHandle,
                              ix_uint32* arg_pEngineNumber,
                              void** arg_ppContext );



/**
 * NAME: ix_cci_exe_set_default
 *
 * DESCRIPTION: Sets an engine as the default engine whose information is 
 *              returned in ix_cci_exe_get_info() when the latter function is
 *              called from a non-engine thread. If this function is never
 *              called, the default engine is the first engine created after
 *              ix_cci_init().
 * 
 * @Param:  - IN    arg_hEngine: Handle of engine to be set as the default.
 *
 * @Return: IX_SUCCESS if successful or an ix_error token encapsulating the 
 *          following error code:
 *              IX_CCI_ASSERTION invalid function parameter (debug compilations 
 *                  only).
 */
IX_EXPORT_FUNCTION 
ix_error ix_cci_exe_set_default(
                                 ix_exe_handle arg_hEngine
                               );


#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__ENGINE_H__) */
