/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation.All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 5
 *
 * = FILENAME
 *      internal_cci_handles.h
 *
 * = DESCRIPTION
 *  Defines all the handles used in the core-component infrastructure. Each
 *  different type of handle will have a type code in the most-significant 
 *  bits. This will allow for error checking to make sure that the wrong
 *  handle isn't being used with an API.
 *
 *  Handles are the most convenient way to handle shared-memory objects.
 *  In the core component infrastructure, the following items need to be in
 *  shared memory:
 *      Packet and message queues
 *      Policies 
 *      Engines
 *  In addition, the APIs specify that core components have handles
 *
 *  We want to map different types of handles to different bits. This 
 *  allows type checking and inclusion of information about important
 *  parents in the handle. For example, a core component will know 
 *  about its parent policy and the execution engine it is running in
 *  without having to work up the policy tree.
 *
 *  In this implementation, the array sizes for each CCI type will be as 
 *  follows:
 *      Packet queues   192 (8 bits)
 *      Message queues  192 (8 bits)
 *      Core components  64 (6 bits) or possibly up to 256 (8 bits)
 *      Policies        512 (9 bits)
 *      Engines          64 (6 bits)
 *      Total bits:         37 bits
 *
 *  This is too much to be expressed in a single 32-bit handle. However,
 *  if we add a number of bits to express the handle type, types at the
 *  bottom of the execution control tree can use the same bit space. 
 *  Also, since send-message and send-packet functions do not need to
 *  access a core component, index information for packet queues, 
 *  message queues, and core components can use the same handle bits.
 *
 *  Since there are 5 handle types, we need 3 bits for handle type information.
 *  info. We will distribute unused reserved bits in the handle in such a way as
 *  to maximize the extensibility. Bearing this in mind, the handle bit 
 *  assignments, in this implementation, will be as follows:
 *
 *  31   29        26        21        18        10        7            0
 *  ----------------------------------------------------------------------
 *  | Type | Rsvd | Engine Idx | Rsvd | Policy Idx | Rsvd | Queue/CC Idx |       
 *  ----------------------------------------------------------------------
 *
 * Event handles can also share the same bit space as queues and CCs or, since
 * events are not associated with policies, the bit mask could take up as many
 * as 21 bits. This implementation will use the same byte mask as for queues
 * and CCs.
 *
 * = AUTHOR
 *       Roland Arajs
 *       roland.arajs@intel.com
 *
 * = CHANGE HISTORY
 *      3/14/2002 11:58:34 AM - creation time
 *
 * ============================================================================
 * $Id: internal_cci_handles.h,v 1.6 2002/12/06 23:51:19 rlarajs Exp $
 */

#if !defined(__INTERNAL_CCI_HANDLES_H__)
#define __INTERNAL_CCI_HANDLES_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */

#include "cci/cci_defs.h"
#include "cci/internal/internal_cci_error.h"

#define IX_CCI_H_CC_START_BIT       0
#define IX_CCI_H_CC_OVER_BIT        8 
#define IX_CCI_H_TKQ_START_BIT      0
#define IX_CCI_H_TKQ_OVER_BIT       8
#define IX_CCI_H_EVENT_START_BIT    0
#define IX_CCI_H_EVENT_OVER_BIT     21
#define IX_CCI_H_POLICY_START_BIT   10
#define IX_CCI_H_POLICY_OVER_BIT    19
#define IX_CCI_H_ENGINE_START_BIT   21
#define IX_CCI_H_ENGINE_OVER_BIT    27

/**
 * DESCRIPTION: Mask to bits in handle that index into a component
 *              array
 */
#define IX_CCI_HTYPE_MASK       0xE0000000UL

#define IX_CCI_ENGINE_MASK_BELOW    ((1UL << IX_CCI_H_ENGINE_START_BIT) - 1)
#define IX_CCI_ENGINE_MASK_ABOVE    ~((1UL << IX_CCI_H_ENGINE_OVER_BIT) - 1)
#define IX_CCI_ENGINE_IDX_MASK      ~(IX_CCI_ENGINE_MASK_ABOVE \
                                    | IX_CCI_ENGINE_MASK_BELOW) 

#define IX_CCI_POLICY_MASK_BELOW    ((1UL << IX_CCI_H_POLICY_START_BIT) - 1)
#define IX_CCI_POLICY_MASK_ABOVE    ~((1UL << IX_CCI_H_POLICY_OVER_BIT) - 1)
#define IX_CCI_POLICY_IDX_MASK      ~(IX_CCI_POLICY_MASK_ABOVE \
                                    | IX_CCI_POLICY_MASK_BELOW) 

#define IX_CCI_TKQ_IDX_MASK         ((1UL << IX_CCI_H_TKQ_OVER_BIT) - 1)


#define IX_CCI_CC_IDX_MASK          ((1UL << IX_CCI_H_CC_OVER_BIT) - 1)

#define IX_CCI_EVENT_IDX_MASK       ((1UL << IX_CCI_H_EVENT_OVER_BIT) - 1)

#define IX_CCI_INVALID_INDEX        0xFFFFFFFFUL

/**
 * TYPENAME: ix_cci_handle_types
 * 
 * DESCRIPTION: Besides indexes for engines, policies, and queues/CCs, the 
 *              ix_tk_handle type is given in the most significant 3 bits of the
 *              handle.
 */
typedef enum ix_e_cci_handle_types
{
    IX_CCI_HANDLE_TYPES_INVALID = 0x00000000,
    IX_CCI_HANDLE_TYPES_EVENT  =  0x40000000,
    IX_CCI_HANDLE_TYPES_MSGTKQ =  0x60000000,
    IX_CCI_HANDLE_TYPES_PKTTKQ =  0x80000000,
    IX_CCI_HANDLE_TYPES_POLICY =  0xA0000000,
    IX_CCI_HANDLE_TYPES_CC =      0xC0000000,
    IX_CCI_HANDLE_TYPES_EXE =     0xE0000000,
} ix_cci_handle_types;


#define IX_CCI_H_TYPE_MASK      0xE0000000UL


/**
 * MACRO NAME: IX_CCI_GET_HANDLE_TYPE
 *
 * DESCRIPTION: 
 *
 * @Param:  - IN arg_Handle
 *
 * @Return ix_cci_handle_types enumerated handle type
 */
#define IX_CCI_GET_HANDLE_TYPE( arg_Handle ) \
    ((arg_Handle) & IX_CCI_H_TYPE_MASK)



              

/**
 * MACRO NAME: IX_CCI_EVENT_IDX_TO_HANDLE
 *
 * DESCRIPTION: Converts an event shared-memory index to a handle, given an
 *              existing CC handle.
 * @Param:  - IN Event index
 * @Param:  - IN CC handle (could also be an engine handle). The engine IDX
 *               is retained in the event handle.
 *  Returns handle if index is valid, or IX_CCI_HANDLE_TYPES_INVALID 
 *  otherwise (if programming logic check is not optimized out)
 */
#if defined(_IX_CCI_DEBUG_) && !defined(_CCT_EXCLUDE_)
#define IX_CCI_EVENT_IDX_TO_HANDLE( index, cc_handle )                      \
    ( index < IX_CCI_MAX_EVENTS                                             \
    ? index | IX_CCI_HANDLE_TYPES_EVENT                                     \
            | (cc_handle & IX_CCI_ENGINE_IDX_MASK)                          \
    : IX_CCI_HANDLE_TYPES_INVALID)
#else
#define IX_CCI_EVENT_IDX_TO_HANDLE( index, cc_handle)                       \
     (index | IX_CCI_HANDLE_TYPES_EVENT                                     \
            | (cc_handle & IX_CCI_ENGINE_IDX_MASK)) 
#endif
                                  
                                                             

/**
 * MACRO NAME: IX_MSG_IDX_TO_HANDLE
 *
 * DESCRIPTION: Converts a message-queue shared-memory index to a handle.
 * @Param:  - IN index: Message index
 * @Param:  - IN policy_handle: The engine and policy indexes are retained in 
 *               the message handle.
 * Returns handle if index is valid, or IX_CCI_HANDLE_TYPES_INVALID 
 * otherwise (if programming logic check is not optimized out)
 */
#if defined(_IX_CCI_DEBUG_) && !defined(_CCT_EXCLUDE_)
#define IX_MSG_IDX_TO_HANDLE( index, policy_handle )                \
    ( index < IX_CCI_MAX_INPUT_IDS                                  \
    ? index | IX_CCI_HANDLE_TYPES_MSGTKQ                            \
            | (policy_handle & ~IX_CCI_H_TYPE_MASK)                 \
    : IX_CCI_HANDLE_TYPES_INVALID)   
#else
#define IX_MSG_IDX_TO_HANDLE( index, policy_handle )                \
    ( index | IX_CCI_HANDLE_TYPES_MSGTKQ                            \
            | (policy_handle & ~IX_CCI_H_TYPE_MASK)                         
#endif                                

/**
 * MACRO NAME: IX_PKT_IDX_TO_HANDLE
 *
 * DESCRIPTION: Converts a message-queue shared-memory index to a handle.
 * @Param:  - IN index: Packet index
 * @Param:  - IN policy_handle: The engine and policy indexes are retained in 
 *               the packet handle.
 * Returns handle if index is valid, or IX_CCI_HANDLE_TYPES_INVALID 
 * otherwise (if programming logic check is not optimized out)
 */
#if defined(_IX_CCI_DEBUG_) && !defined(_CCT_EXCLUDE_)
#define IX_PKT_IDX_TO_HANDLE( index, policy_handle )                \
    ( index < IX_CCI_MAX_INPUT_IDS                                  \
    ? index | IX_CCI_HANDLE_TYPES_PKTTKQ                            \
            | (policy_handle  & ~IX_CCI_H_TYPE_MASK)                \
    : IX_CCI_HANDLE_TYPES_INVALID)   
#else
#define IX_PKT_IDX_TO_HANDLE( index, policy_handle )                \
    ( index | IX_CCI_HANDLE_TYPES_PKTTKQ )                          \
            | (policy_handle & ~IX_CCI_H_TYPE_MASK)
#endif                                

/**
 * MACRO NAME: IX_CC_IDX_TO_HANDLE
 *
 * DESCRIPTION: Converts a message-queue shared-memory index to a handle.
 *
 * @Param:  - IN index: CC index into engines CC list
 * @Param:  - IN engine_handle: The engine index is retained in the CC handle.
 *
 * Returns handle if index is valid, or IX_CCI_HANDLE_TYPES_INVALID 
 * otherwise (if programming logic check is not optimized out)
 */
#if defined(_IX_CCI_DEBUG_) && !defined(_CCT_EXCLUDE_)
#define IX_CC_IDX_TO_HANDLE( index, engine_handle )                 \
    ( index < IX_CCI_MAX_CCS                                        \
    ? (index << IX_CCI_H_CC_START_BIT) | IX_CCI_HANDLE_TYPES_CC     \
        | (engine_handle  & ~IX_CCI_H_TYPE_MASK)                    \
    : IX_CCI_HANDLE_TYPES_INVALID)   
#else
#define IX_CC_IDX_TO_HANDLE( index, engine_handle )                 \
    (index << IX_CCI_H_CC_START_BIT) | IX_CCI_HANDLE_TYPES_CC       \
    | (engine_handle & ~IX_CCI_H_TYPE_MASK)
#endif                                

/**
 * MACRO NAME: IX_POLICY_IDX_TO_HANDLE
 *
 * DESCRIPTION: Converts a message-queue shared-memory index to a handle.
 *
 * @Param:  - IN index: Policy index in shared memory
 * @Param:  - IN engine_handle: The engine index is retained in the policy
 *               handle.
 *
 * Returns handle if index is valid, or IX_CCI_HANDLE_TYPES_INVALID 
 * otherwise (if programming logic check is not optimized out)
 */
#if defined(_IX_CCI_DEBUG_) && !defined(_CCT_EXCLUDE_)
#define IX_POLICY_IDX_TO_HANDLE( index, engine_handle )                     \
    ( index < IX_CCI_MAX_POLICIES                                           \
    ? (index << IX_CCI_H_POLICY_START_BIT) | IX_CCI_HANDLE_TYPES_POLICY     \
        | (engine_handle & ~IX_CCI_H_TYPE_MASK)                             \
    : IX_CCI_HANDLE_TYPES_INVALID)   
#else
#define IX_POLICY_IDX_TO_HANDLE( index, engine_handle )                                   \
    (index << IX_CCI_H_POLICY_START_BIT) | IX_CCI_HANDLE_TYPES_POLICY       \
    | (engine_handle & ~IX_CCI_H_TYPE_MASK)
#endif                                

/**
 * MACRO NAME: IX_ENGINE_IDX_TO_HANDLE
 *
 * DESCRIPTION: Converts a message-queue shared-memory index to a handle.
 *
 * @Param:  - IN index: Engine index in shared memory
 *
 *              Returns handle if index is valid, or IX_CCI_HANDLE_TYPES_INVALID 
 *              otherwise (if programming logic check is not optimized out)
 */
#if defined(_IX_CCI_DEBUG_) && !defined(_CCT_EXCLUDE_)
#define IX_ENGINE_IDX_TO_HANDLE( index )                                    \
    ( index < IX_CCI_MAX_ENGINES                                            \
    ? (index << IX_CCI_H_ENGINE_START_BIT) | IX_CCI_HANDLE_TYPES_EXE        \
    : IX_CCI_HANDLE_TYPES_INVALID)   
#else
#define IX_ENGINE_IDX_TO_HANDLE( index )                                    \
    (index << IX_CCI_H_ENGINE_START_BIT) | IX_CCI_HANDLE_TYPES_EXE                          
#endif                                

/**
 * MACRO NAME: IX_HAND_TO_MSG_IDX
 *
 * DESCRIPTION: Converts message-queue handle to index into message queue
 *              structure array in shared memory.
 *
 *              Returns index if handle is valid, or IX_CCI_INVALID_INDEX 
 *              otherwise (if programming logic check is not optimized out)
 */
#if defined(_IX_CCI_DEBUG_) && !defined(_CCT_EXCLUDE_)
#define IX_HAND_TO_MSG_IDX( tkq_handle )                                \
    ( IX_CCI_HANDLE_TYPES_MSGTKQ == (tkq_handle & IX_CCI_H_TYPE_MASK)       \
    ?  ( tkq_handle & IX_CCI_TKQ_IDX_MASK ) : IX_CCI_INVALID_INDEX )
#else
#define IX_HAND_TO_MSG_IDX( tkq_handle )                                \
    ( tkq_handle & IX_CCI_TKQ_IDX_MASK )
#endif

/**
 * MACRO NAME: IX_HAND_TO_PKT_IDX
 *
 * DESCRIPTION: Converts packet-queue handle to index into packet queue
 *              structure array in shared memory.
 *
 *              Returns index if handle is valid, or IX_CCI_INVALID_INDEX 
 *              otherwise (if programming logic check is not optimized out)
 *              
 */
#if defined(_IX_CCI_DEBUG_) && !defined(_CCT_EXCLUDE_)
#define IX_HAND_TO_PKT_IDX( tkq_handle )                                \
    ( IX_CCI_HANDLE_TYPES_PKTTKQ == (IX_CCI_H_TYPE_MASK & tkq_handle)       \
    ?  ( tkq_handle & IX_CCI_TKQ_IDX_MASK ) : IX_CCI_INVALID_INDEX )
#else
#define IX_HAND_TO_PKT_IDX( tkq_handle )                                \
    ( tkq_handle & IX_CCI_TKQ_IDX_MASK )
#endif


/**
 * MACRO NAME: IX_HAND_TO_CC_IDX
 *
 * DESCRIPTION: Converts packet-queue handle to index into core component
 *              structure array in execution engine local memory.
 *
 *              Returns index if handle is valid, or IX_CCI_INVALID_INDEX 
 *              otherwise (if programming logic check is not optimized out)
 *              
 */
#if defined(_IX_CCI_DEBUG_) && !defined(_CCT_EXCLUDE_)
#define IX_HAND_TO_CC_IDX( tkq_handle )                                 \
    ( IX_CCI_HANDLE_TYPES_CC == (IX_CCI_H_TYPE_MASK & tkq_handle)           \
    ?  ( tkq_handle & IX_CCI_CC_IDX_MASK ) : IX_CCI_INVALID_INDEX )
#else
#define IX_HAND_TO_CC_IDX( tkq_handle )                                 \
    ( tkq_handle & IX_CCI_CC_IDX_MASK )
#endif

/**
 * MACRO NAME: IX_HAND_TO_POLICY_IDX
 *
 * DESCRIPTION: Converts packet-queue handle to index into policy
 *              structure array in shared memory.
 *
 *              Returns index if handle is valid, or IX_CCI_INVALID_INDEX 
 *              otherwise (if programming logic check is not optimized out)
 *              
 */
#if defined(_IX_CCI_DEBUG_) && !defined(_CCT_EXCLUDE_)
#define IX_HAND_TO_POLICY_IDX( tkq_handle )                             \
    ( IX_CCI_HANDLE_TYPES_POLICY == (IX_CCI_H_TYPE_MASK & tkq_handle)       \
    ?  ( (tkq_handle & IX_CCI_POLICY_IDX_MASK)                              \
           >>  IX_CCI_H_POLICY_START_BIT ) : IX_CCI_INVALID_INDEX )
#else
#define IX_HAND_TO_POLICY_IDX( tkq_handle )                             \
    ( (tkq_handle & IX_CCI_POLICY_IDX_MASK) >> IX_CCI_H_POLICY_START_BIT )
#endif

/**
 * MACRO NAME: IX_HAND_TO_ENGINE_IDX
 *
 * DESCRIPTION: Converts packet-queue handle to index into engine index in
 *              shared memory
 *
 *              Returns index if handle is valid, or IX_CCI_INVALID_INDEX 
 *              otherwise (if programming logic check is not optimized out)
 *              
 */
#if defined(_IX_CCI_DEBUG_) && !defined(_CCT_EXCLUDE_)
#define IX_HAND_TO_ENGINE_IDX( tkq_handle )                                 \
    ( IX_CCI_HANDLE_TYPES_EXE == (IX_CCI_H_TYPE_MASK & tkq_handle)          \
      || IX_CCI_HANDLE_TYPES_CC == (IX_CCI_H_TYPE_MASK & tkq_handle)        \
      || IX_CCI_HANDLE_TYPES_EVENT == (IX_CCI_H_TYPE_MASK & tkq_handle)     \
    ?  ( (tkq_handle & IX_CCI_ENGINE_IDX_MASK)                              \
           >>  IX_CCI_H_ENGINE_START_BIT ) : IX_CCI_INVALID_INDEX )
#else
#define IX_HAND_TO_ENGINE_IDX( tkq_handle )                             \
    ( (tkq_handle & IX_CCI_ENGINE_IDX_MASK) >> IX_CCI_H_ENGINE_START_BIT )
#endif




#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__INTERNAL_CCI_HANDLES_H__) */
