/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation.All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 5
 *
 * = FILENAME
 *      internal_exe_ctrl_if.h
 *
 * = DESCRIPTION
 *      Defines the common interface used for communicating message and payload
 *      data in the framework superstructure. This interface is shared by 
 *      end-point extensions, scheduling policies, and library-based execution
 *      engines.
 *
 * = AUTHOR
 *       Roland Arajs
 *       roland.arajs@intel.com
 *
 * = CHANGE HISTORY
 *      1/28/2002 9:55:59 AM - creation time
 *
 * ============================================================================
 * $Id: internal_exe_ctrl_if.h,v 1.7 2003/08/18 21:48:07 ageorges Exp $
 */



#if !defined(__INTERNAL_EXE_CTRL_IF_H__)
#define __INTERNAL_EXE_CTRL_IF_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */

#include "ix_ossl.h"
#include "ix_rm.h"
#include "rm/internal/internal_tk_input_notify.h"
#include "cci/cci_defs.h"
#include "cci/cci_types.h"



/**
 * Following definitions are used for providing process function status:
 * either data still pending, or high-priority override
 */
#define IX_CCI_DATA_STILL_PENDING       1UL
#define IX_CCI_HIGH_PRIORITY_OVERRIDE   2UL
            

/**
 * Forward declarations
 */
typedef struct ix_s_cci_exe_ctrl_if ix_cci_exe_ctrl_if;


typedef struct ix_s_tkp_entry ix_tkp_entry;

typedef struct ix_s_engine_proc_node ix_engine_proc_node;
/**
 * TYPENAME: ix_cci_process_tokens_ptr
 * 
 * DESCRIPTION: Called by parent after the engine synchronization mechanism has
 *              been unlocked in response to a token (message or pkt) arriving
 *              on an ECI leaf.
 */

typedef 
ix_error (*ix_cci_process_tokens_ptr) 
                    (
                      ix_cci_exe_ctrl_if* arg_pTokenProcessor,
                      ix_uint32* arg_pNumInputsPending,
                      ix_uint32* arg_pProcessCount,
                      ix_engine_proc_node* arg_pEngProcNode
                    );

/**
 * TYPENAME: ix_cci_eci_fini_ptr
 * 
 * DESCRIPTION: Called by parent after the engine synchronization mechanism has
 *              been unlocked in response to a token (message or pkt) arriving
 *              on an EPE.
 */

typedef 
ix_error (*ix_cci_eci_fini_ptr) 
                    ( ix_cci_exe_ctrl_if* arg_pTokenProcessor );



/**
 * TYPENAME: ix_cci_exe_ctrl_if (typedef appears in 
 *           internal_token_proc_container.h)
 * 
 * DESCRIPTION: Structure common to framework superstructure components 
 *              (end-point extensions, scheduling policies, and execution
 *              engines).
 *
 */
struct ix_s_cci_exe_ctrl_if
{
    /* token-processor interface. Parent calls process-tokens function. This
    *  block selects a child (if there is more than one) and calls the 
    *  process-tokens function in that child.
    */
    ix_cci_process_tokens_ptr       pProcessFunc; /* Process fn called by parent */
    ix_cci_eci_fini_ptr             pFiniFunc;

    /* Input notify structure (ix_tk_input_notify) is in shared memory. In order 
     * to access this memory from more than one address space, this memory is 
     * initialized before this structure. 
     */
    ix_tk_input_notify*             pTkInNotify;
};



/**
 * TYPENAME: ix_eci_leaf
 * 
 * DESCRIPTION: For execution control tree leaves
 *
 */
typedef struct ix_s_eci_leaf
{
    ix_cci_exe_ctrl_if  tEci;
    ix_communication_id vCommId;

    void*               pUserContext; 
    ix_handle           hUserCtxOwner;
} ix_eci_leaf;



/**
 * NAME: ix_cci_eci_init
 *
 * DESCRIPTION:   Initializes a common token interface structure
 * 
 * @Param:  - IN arg_pTokenProcIf. Pointer to target structure.
 * @Param:  - IN arg_ProcessTokensFuncPtr. Process function to be associated
 *                  with the target structure.
 * @Param:  - IN arg_FiniFuncPtr. Termination function to be associated with
 *                  the target structure.
 * @Param:  - IN arg_TkHandle. Specifies the location of the corresponding
 *                  child-notify structure in shared memory.
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 *
 * NOTE: This function is for calling by the ix_cci_eci_node_init and
 *       ix_cci_eci_leaf_init functions only.
 */
ix_error ix_cci_eci_init(
                          ix_cci_exe_ctrl_if*       arg_pTokenProcIf,
                          ix_cci_process_tokens_ptr arg_ProcessTokensFuncPtr,
                          ix_cci_eci_fini_ptr       arg_FiniFuncPtr,
                          ix_tk_handle              arg_TkHandle
                        );



/**
 * NAME: ix_cci_eci_fini
 *
 * DESCRIPTION:   
 * 
 * @Param:  - IN arg_pTokenProcessor. Pointer to ix_eci_node. The argument is of
 *                  type ix_cci_exe_ctrl_if* so that its function pointer can be
 *                  called within the execution control tree without the parent
 *                  having to know whether it is calling a fini function for
 *                  and eci node or leaf.
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
ix_error ix_cci_eci_fini(
                               ix_cci_exe_ctrl_if* arg_pTokenProcessor
                             );


/**
 * NAME: ix_cci_set_notify_id
 *
 * DESCRIPTION:   Sets the notification ID in the token-notify interface
 *                corresponding to this execution-control construct.
 * 
 * @Param:  - IN    arg_pEci. Pointer to execution control structure
 * @Param:  - IN    arg_vId.  New ID for corresponding token interface
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
ix_error ix_cci_set_notify_id(
                               ix_cci_exe_ctrl_if*   arg_pEci,
                               ix_uint32             arg_vId
                             );




/**
 * NAME: ix_cci_eci_leaf_init
 *
 * DESCRIPTION:   Structure used by token queues 
 * 
 * @Param:  - IN arg_pEciLeaf. Pointer to target structure.
 * @Param:  - IN arg_ProcessTokensFuncPtr. Process function to be associated
 *               w  ith the target structure.
 * @Param:  - IN arg_FiniFuncPtr. Termination function for target structure.
 * @Param:  - IN arg_TkHandle. Specifies the location of the corresponding
 *                  child-notify structure in shared memory.
 * @Param:  - IN arg_pUserContext. Context passed into 
 *                  ix_cci_cc_add_message_handler() or 
 *                  ix_cci_cc_add_message_handler(), which may be a CC context
 *                  or engine context.
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
ix_error ix_cci_eci_leaf_init(
                               ix_eci_leaf*               arg_pEciLeaf,
                               ix_cci_process_tokens_ptr  arg_ProcessTokensFuncPtr,
                               ix_cci_eci_fini_ptr        arg_FiniFuncPtr,
                               ix_tk_handle               arg_TkHandle,
                               void*                      arg_pUserContext
                             );



#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__INTERNAL_EXE_CTRL_IF_H__) */
