/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is suplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or diclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation.All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 4
 *
 * = FILENAME
 *      policy.h
 *
 * = DESCRIPTION
 *      Defines public APIs for policies in the Core Component Infrastructure.
 *
 * = AUTHOR
 *       Roland Arajs
 *       roland.arajs@intel.com
 *
 * = CHANGE HISTORY
 *      4/8/2002 11:03:47 AM - creation time
 *
 * ============================================================================
 * $Id: policy.h,v 1.3 2002/09/27 17:44:11 rlarajs Exp $
 */

#if !defined(__POLICY_H__)
#define __POLICY_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */


#include "cci/cci_types.h"


/**
 * TYPENAME: ix_policy_type
 * 
 * DESCRIPTION: 
 *
 */
typedef enum ix_e_policy_type
{
    IX_POLICY_TYPE_FIRST = 0,
    IX_POLICY_TYPE_RR = IX_POLICY_TYPE_FIRST,   /* Round-robin */
    IX_POLICY_TYPE_WRR,                         /* Weighted round-robin */
    IX_POLICY_TYPE_SP,                          /* Strict priority */
    IX_POLICY_TYPE_LAST
} ix_policy_type;



/**
 * NAME: ix_cci_policy_create
 *
 * DESCRIPTION: Create a scheduling policy. This function should be called in 
 *              the user-defined init() function of an execution engine.
 * 
 * @Param:  - IN argPolicyType: Enumerated type indicating the policy type. 
 * @Param:  - OUT arg_pHandle: Pointer to handle of new policy.
 *
 * @Return: IX_SUCCESS if successful or an ix_error token encapsulating one of 
 *          the following error codes:
 *              IX_CCI_ERR_POLICY_CREATE memory allocation error
 *              IX_CCI_ASSERTION invalid function parameter (debug compilations 
 *                  only).
 */
IX_EXPORT_FUNCTION 
ix_error ix_cci_policy_create(
                               ix_policy_type argPolicyType,
                               ix_policy_handle* arg_pHandle
                             );


/**
 * NAME: ix_cci_policy_destroy
 *
 * DESCRIPTION: Destroy a scheduling policy. This function should be called 
 *              from the user-defined fini() function of an execution engine.
 * 
 * NOTE:        This function does not destroy any subtree policies. The user 
 *              is responsible for destroying all policies that he/she has 
 *              created. The core component infrastructure will ensure that 
 *              notifications are dropped gracefully from a tree that is in the 
 *              process of being dismantled.
 *
 * @Param:  - IN arg_Handle: Handle to policy (becomes invalid after this 
 *                  function returns). 
 *
 * @Return: IX_SUCCESS if successful or an ix_error token encapsulating of the 
 *          following error code:
 *              IX_CCI_ASSERTION invalid function parameter (debug compilations 
 *                  only).
 */
IX_EXPORT_FUNCTION 
ix_error ix_cci_policy_destroy( ix_policy_handle arg_Handle );



/**
 * NAME: ix_cci_policy_add_branch
 *
 * DESCRIPTION: Adds a branch node (scheduling policy) to an another scheduling 
 * policy. Up to 32 nodes (branches and leaves) can be added to an individual 
 * policy. This function should be called from the user-defined init() function 
 * of an execution engine on an existing policy, created using 
 * ix_cci_policy_create().
 * 
 * @Param:  - IN arg_hPolicy: Scheduling policy where branch will be added.
 * @Param:  - IN arg_hChild: Branch (policy) to be added.
 * @Param:  - IN arg_WeightOrPriority: Weight (if parent is WRR policy), 
 *                  priority (if parent is strict priority policy). This 
 *                  parameter is ignored if the parent is a round-robin 
 *                  policy.
 *
 * @Return: IX_SUCCESS if successful or an ix_error token encapsulating one of 
 *          the following error codes:
 *              IX_CCI_TOKEN_PROCESSOR_CONTAINER_FULL parent policy is full and 
 *                  was initialized to be fixed in size
 *              IX_CCI_ERR_TKP_CONTAINER_MEM_ALLOC memory allocation error
 *              IX_CCI_ASSERTION invalid function parameter (debug compilations 
 *                  only).
 */
IX_EXPORT_FUNCTION 
ix_error ix_cci_policy_add_branch(
                                   ix_policy_handle arg_hPolicy,
                                   ix_policy_handle arg_hChild,
                                   ix_uint32 arg_WeightOrPriority
                                 );



/**
 * NAME: ix_cci_policy_add_leaf
 *
 * DESCRIPTION: Adds a leaf node (packet/message input node) to a scheduling 
 *              policy. Up to 32 nodes (branches and leaves) can be added to an 
 *              individual policy.This function should be called from the 
 *              user-defined init() function of an execution engine on an 
 *              existing policy, created using ix_cci_policy_create().
 * 
 * @Param:  - IN arg_hPolicy: Scheduling policy where leaf will be added.
 * @Param:  - IN arg_InputId: Packet/message-input ID.
 * @Param:  - IN arg_IsPacketId: Used as a Boolean (1 for packet-input ID, 
 *                  0 for message-input ID).
 * @Param   - IN arg_WeightOrPriority: Weight (if parent is WRR policy), or 
 *                  priority (if parent is strict priority policy). This 
 *                  parameter is ignored if the parent is a round-robin policy.
 *
 * @Return: IX_SUCCESS if successful or an ix_error token encapsulating one of 
 *          the following error codes:
 *              IX_CCI_TOKEN_PROCESSOR_CONTAINER_FULL parent policy is full and 
 *                  was initialized to be fixed in size
 *              IX_CCI_ERR_TKP_CONTAINER_MEM_ALLOC memory allocation error
 *              IX_CCI_ASSERTION invalid function parameter (debug compilations 
 *                  only).
 */
IX_EXPORT_FUNCTION 
ix_error ix_cci_policy_add_leaf(
                                 ix_policy_handle arg_hPolicy,
                                 ix_uint32 arg_InputId,
                                 ix_uint32 arg_IsPacketId,
                                 ix_uint32 arg_WeightOrPriority
                               );



#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__POLICY_H__) */
