/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation.All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 5
 *
 * = FILENAME
 *      cc.c
 *
 * = DESCRIPTION
 *      Core component functions.
 *
 * = AUTHOR
 *       Roland Arajs
 *       roland.arajs@intel.com
 *
 *
 * = CHANGE HISTORY
 *      4/28/2002 8:25:54 PM - creation time
 *
 * ============================================================================
 * $Id: cc.c,v 1.11 2003/08/25 18:49:41 ageorges Exp $
 */
#include "ix_ossl.h"
#include "ix_cc.h"

#if !(_IX_OS_TYPE == _IX_OS_LINUX_KERNEL_)
#include <stddef.h>
#endif /*(_IX_OS_TYPE == _IX_OS_LINUX_KERNEL_) */

#include "cci/internal/internal_cc.h"
#include "cci/internal/internal_engine.h"
#include "cci/internal/internal_cci_handles.h"



/**
 * Global function pointer for CC instance retrieval function
 */
ix_cc_get_instance g_CCGetInstanceFPtr = _ix_rm_cci_cc_default_get_instance;



/*******************************************************************************
 * Public API
 */

ix_error ix_cci_cc_create(
                           ix_exe_handle arg_EngHandle,
                           ix_cc_init arg_InitFunc,
                           ix_cc_fini arg_FiniFunc,
                           void* arg_pContextIn,
                           ix_cc_handle* arg_phComponent
                         )
{
    /* 
     * Get execution engine pointer via the token-notify interface.
     * Check that the engine handle is valid and that the init and fini 
     * function pointers are not NULL
     */
    ix_error err;
    ix_engine_proc_node* pEngine;
    ix_uint32 idxEng = IX_HAND_TO_ENGINE_IDX( arg_EngHandle );
    ix_cc_node* pCC;
    ix_cc_node** ppCCEntry;
    ix_tk_input_notify* pEngNotifyIf;
    ix_cc_handle hCC;
    ix_uint32 idxCC = 0;
    IX_CCI_PARAM_ASSERT( 
        IX_CCI_GET_HANDLE_TYPE(arg_EngHandle) == IX_CCI_HANDLE_TYPES_EXE );
    IX_CCI_PARAM_ASSERT( idxEng != IX_CCI_INVALID_INDEX )
    IX_CCI_PARAM_ASSERT( arg_InitFunc != NULL  )
    IX_CCI_PARAM_ASSERT( arg_FiniFunc != NULL )
    IX_CCI_PARAM_ASSERT( arg_phComponent != NULL )

    pEngNotifyIf = g_pEngine + idxEng;
    pEngine = (ix_engine_proc_node*)pEngNotifyIf->pProcIfContext;
    
    /* Create a CC in the execution engine's CC container */
    pCC = IX_OSSL_MALLOC( sizeof(ix_cc_node) );
    IX_CCI_ERR_RET( pCC==NULL, IX_CCI_ERR_OS_MEM_ALLOC )
    ix_ossl_memset( pCC, 0, sizeof(ix_cc_node) );

    /* Put the CC node in the engine's CC list, keeping track
     * of the position of the CC in the list 
     */
    ppCCEntry = &pEngine->pCCList;
    while( *ppCCEntry )
    {
        idxCC++;
        ppCCEntry = &(*ppCCEntry)->pNext;
    } 
    *ppCCEntry = pCC;

    /* Set the values of the node elements */
    pCC->pFiniFunc = arg_FiniFunc;
    pCC->pNext = NULL;
    pCC->vIsUserInitComplete = 0;

    /* Convert the CC index into a handle and update O/P arg */
    *arg_phComponent = hCC = IX_INSTANCE_TO_HANDLE( IX_CC_IDX_TO_HANDLE( idxCC, arg_EngHandle ), 
        (g_CCGetInstanceFPtr(arg_pContextIn)));

    IX_CCI_ASSERT( hCC != (ix_uint32)IX_CCI_HANDLE_TYPES_INVALID );

    /** 
     * Call the user's CC-initialization function, passing the context provided
     * by this function's arguments, and storing the returned context
     * in the CC node.
     */
    pCC->pUserContext = arg_pContextIn;
    err = arg_InitFunc( hCC, &pCC->pUserContext );
    pCC->vIsUserInitComplete = 1;

    return err;
}

ix_error ix_cci_cc_destroy( ix_cc_handle arg_hComponent )
{
    /* Convert CC handle into CC index and engine index */ 
    ix_uint32 idxCC;
    ix_uint32 idxEngine;
    ix_cc_node* pCCEntry;
    ix_tk_input_notify* pEngNotifyIf;
    ix_engine_proc_node* pEngine;
    ix_error err = IX_SUCCESS;
    ix_uint32 i = 0;

    idxCC = IX_HAND_TO_CC_IDX( arg_hComponent );
    IX_CCI_PARAM_ASSERT( idxCC != IX_CCI_INVALID_INDEX )

    idxEngine = IX_HAND_TO_ENGINE_IDX( arg_hComponent );
    IX_CCI_ASSERT( idxEngine != IX_CCI_INVALID_INDEX )
    
    /* Get pointer to engine's execution control construct via Notify I/F */
    pEngNotifyIf = g_pEngine + idxEngine;
    pEngine = (ix_engine_proc_node*)pEngNotifyIf->pProcIfContext;

    /* Find the CC in the engine's CC list*/
    pCCEntry = pEngine->pCCList;
    while( i++ < idxCC )
    {
        IX_CCI_ASSERT( pCCEntry->pNext != NULL )
        pCCEntry = pCCEntry->pNext;
    }
    /* 
     * Call the user-defined fini function with user-defined context associated
     * with that CC
     */
    err = pCCEntry->pFiniFunc( arg_hComponent, pCCEntry->pUserContext );
    /* 
     * Clear the fields associated with the CC, but leave it up to the engine
     * shutdown sequence to clear the CC list. (Has to stay in engine's CC list
     * so that the handles of other CCs remain valid.)
     */
    pCCEntry->pFiniFunc = NULL;
    pCCEntry->pUserContext = NULL;
    return err;
}


/**
 * Internal API
 */

/**
 * NAME: _ix_rm_cci_cc_default_get_instance
 *
 * DESCRIPTION: This function will retrieve the default instance of a CC. For now
 *  the function will return 0 at all times.
 * 
 * @Param:  - IN arg_pContext - context passed to the CC creation function
 *
 * @Return: This function returns 0 at all times.
 */
ix_uint32 _ix_rm_cci_cc_default_get_instance(
                                              void* arg_pContext
                                            )
{
    (void)arg_pContext;

    return 0;
}
