/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation.All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 5
 *
 * = FILENAME
 *      cci_main.c
 *
 * = DESCRIPTION
 *      Contains main CCI initialization and closedown functions.
 *
 * = AUTHOR
 *       Roland Arajs
 *       roland.arajs@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CHANGE HISTORY
 *      4/2/2002 8:46:57 AM - creation time
 *
 * ============================================================================
 * $Id: cci_main.c,v 1.16 2003/08/20 21:56:02 ageorges Exp $
 */
#include "ix_ossl.h"

#if !( _IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_ )
#include <stddef.h>
#endif /* !( _IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_ ) */

#include "ix_rm.h"

#include "cci/cci_defs.h"
#include "cci/cci_main.h"
#include "cci/internal/internal_cci_error.h"
#include "cci/internal/internal_engine.h"


#define IX_CCI_NUM_FERROR_WORDS     2




/**
 * TYPENAME: ix_cci_property_handle
 * 
 * DESCRIPTION: 
 *
 */
typedef enum ix_e_cci_config_property_handle
{
    IX_CCI_PROPERTY_HANDLE_FIRST = 0,
    IX_CCI_PROPERTY_HANDLE_MSGTKQ = IX_CCI_PROPERTY_HANDLE_FIRST,
    IX_CCI_PROPERTY_HANDLE_SHUTDOWNFLAGS,
    IX_CCI_PROPERTY_HANDLE_SEMA,
    IX_CCI_PROPERTY_HANDLE_LAST
} ix_cci_config_property_handle;




static ix_configuration_property_handle    m_aPropertyHandle[IX_CCI_PROPERTY_HANDLE_LAST];




/**
 * Internal function declarations
 */

/**
 * NAME: shared_mem_init
 *
 * DESCRIPTION:   
 * 
 * @Param:  - OUT   arg_ppTkNotifyBlock: Where function will return the block of
 *                      token-notify interfaces
 * @Param:  - OUT   arg_ppUintBlock: Where function will return the shared-memory
 *                      integer block
 * @Param:  - OUT   arg_ppSema: Where function will return the pointer to a global
 *                      semaphore
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
 
static ix_error shared_mem_init(
                                 ix_tk_input_notify** arg_ppTkNotifyBlock,
                                 ix_uint32** arg_ppUintBlock,
                                 ix_ossl_sem_t** arg_ppSema
                               );



/**
 * NAME: create_registry_properties
 *
 * DESCRIPTION:   Creates registry properties and stores their handles in the
 *                module array m_aPropertyHandle
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
static ix_error create_registry_properties( void );


/**
 * NAME: delete_registry_properties
 *
 * DESCRIPTION:   Deletes registry properties created by 
 *                create_registry_properties().
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
static ix_error delete_registry_properties( void );


/**
 * NAME: register_shared_mem_blocks
 *
 * DESCRIPTION:   
 * 
 * @Param:  - IN    arg_PropTkNotify
 * @Param:  - IN    arg_PropUint
 * @Param:  - IN    arg_PropSema
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
static ix_error register_shared_mem_blocks(
                                            ix_uint32 arg_PropTkNotify,
                                            ix_uint32 arg_PropUint,
                                            ix_uint32 arg_PropSema
                                          );




ix_error ix_cci_init( void )
{
    ix_error err;
    ix_tk_input_notify*     pMsgTkq = NULL; 
    ix_uint32*              pShutdownFlags = NULL;
    ix_ossl_sem_t*          pSema = NULL;

#if !defined(_CCT_EXCLUDE_)     
    ix_error cleanup_err;
#endif

	/* Initialize message logging if in CCI debug mode */
    err = IX_CCI_INIT_ERROR_LOGGING();
    IX_CCI_ERR_RET( err != IX_SUCCESS, IX_CCI_ERR_INIT_ERR_LOG )


    /* Initialize the shared-memory blocks */
    err = shared_mem_init( &pMsgTkq, &pShutdownFlags, &pSema );
    IX_CCI_IXERR_RET( err )


    /* Initialize the semaphore */
    err = IX_OSSL_SEM_INIT( IX_OSSL_SEM_AVAILABLE, pSema );
    IX_CCI_ERR_GOTO( err != IX_SUCCESS, 
                      IX_CCI_ERR_GLOB_SEMA_INIT, ret_err2, err )


    /* Create registry properties that will be used by the CCI */
    err = create_registry_properties();
    IX_CCI_IXERR_GOTO( ret_err3, err )


    /* Register the the shared memory pointers in the registry */
    err = register_shared_mem_blocks( (ix_uint32) pMsgTkq,
                                     (ix_uint32) pShutdownFlags,
                                     (ix_uint32) pSema );
    IX_CCI_IXERR_GOTO( ret_err4, err )

    IX_CCI_CC_GET_INSTANCE_UNREGISTER();
    
    return IX_SUCCESS; /* Normal return */

    /* Error return path */
#if !defined(_CCT_EXCLUDE_)
ret_err4:
    cleanup_err = delete_registry_properties();
    IX_CCI_ERR_DUMP( cleanup_err );
ret_err3:
    cleanup_err = IX_OSSL_SEM_FINI( *pSema );
    IX_CCI_ERR_DUMP( cleanup_err );
ret_err2:
    cleanup_err = IX_RM_MEM_FREE( pMsgTkq );
    IX_CCI_ERR_DUMP( cleanup_err );
    cleanup_err = IX_RM_MEM_FREE( pShutdownFlags );
    IX_CCI_ERR_DUMP( cleanup_err );
    cleanup_err = IX_RM_MEM_FREE( pSema );
    IX_CCI_ERR_DUMP( cleanup_err );

    return err;
#endif

}


ix_error ix_cci_fini( void )
{
    ix_uint8*   pMsgTkq = NULL; 
    ix_uint32*  pShutdownFlags = NULL;
    ix_uint32*  pSema = NULL;
    ix_error    err1 = 0;

    /* Lookup registry  to get addresses of the two
     * areas of shared memory initialized by ix_cci_init()
     * Free shared memory 
     */
    ix_error err = ix_rm_cp_property_get_value_uint32(
        m_aPropertyHandle[IX_CCI_PROPERTY_HANDLE_MSGTKQ],
        (ix_uint32*)&pMsgTkq );

    if( err == IX_SUCCESS )
    {
        err = IX_RM_MEM_FREE( pMsgTkq );
    }
    err1 = ix_rm_cp_property_get_value_uint32(
                m_aPropertyHandle[IX_CCI_PROPERTY_HANDLE_SHUTDOWNFLAGS],
                (ix_uint32*)&pShutdownFlags );
    if( err1 == IX_SUCCESS )
    {
        err1 = IX_RM_MEM_FREE( pShutdownFlags );
    }
    IX_CCI_ERR_FWD_OR_DUMP( err1, err )
    err1 = ix_rm_cp_property_get_value_uint32(
                m_aPropertyHandle[IX_CCI_PROPERTY_HANDLE_SEMA],
                (ix_uint32*)&pSema );
    if( err1 == IX_SUCCESS )
    {
        err1 = IX_OSSL_SEM_FINI( (ix_ossl_sem_t)*pSema );
        IX_CCI_ERR_FWD_OR_DUMP( err1, err )

        err1 = IX_RM_MEM_FREE( pSema );
        IX_CCI_ERR_FWD_OR_DUMP( err1, err )
    }
    IX_CCI_ERR_FWD_OR_DUMP( err1, err )

    /* Delete CCI properties in registry */
    err1 = delete_registry_properties();
    IX_CCI_ERR_FWD_OR_DUMP( err1, err )
    return err;
}


static ix_error shared_mem_init(
                                 ix_tk_input_notify** arg_ppTkNotifyBlock,
                                 ix_uint32** arg_ppUintBlock,
                                 ix_ossl_sem_t** arg_ppSema
                               )
{
    ix_error err;
#if !defined(_CCT_EXCLUDE_)
    ix_error cleanup_err;
#endif

	/* Map global token queue pointers to RM shared memory, placing all
     * common memory types in one contiguous block
     */
    ix_uint32 sizeTkNotifyBlock = (2 * IX_CCI_MAX_INPUT_IDS 
                                   + IX_CCI_MAX_POLICIES
                                   + IX_CCI_MAX_ENGINES)
                                   * sizeof(ix_tk_input_notify);

    ix_uint32 sizeUintBlock = (IX_CCI_NUM_SHUTDOWN_WORDS 
                               + IX_CCI_NUM_SHMEM_COUNTS
                               + IX_CCI_NUM_FERROR_WORDS
                               + IX_CCI_NUM_DEFAULT_ENGINES) 
                               * sizeof(ix_uint32);

    err = IX_RM_MEM_ALLOC( IX_CCI_SHARED_MEMORY_TYPE, IX_CCI_MEMORY_CHANNEL,
                           sizeTkNotifyBlock,
                           (void**)arg_ppTkNotifyBlock );
    IX_CCI_ERR_RET( err != IX_SUCCESS, IX_CCI_ERR_SHARED_MEM_ALLOC )

    ix_ossl_memset( *arg_ppTkNotifyBlock, 0, sizeTkNotifyBlock );

    /* Allocate enough space for a shutdown byte for each execution engine
     * plus 4 words for storing engine count, policy count, message-tkin
     * count, and packet tkin count.
     */
    err = IX_RM_MEM_ALLOC( IX_CCI_SHARED_MEMORY_TYPE, IX_CCI_MEMORY_CHANNEL, 
                            sizeUintBlock,
                            (void**)arg_ppUintBlock );
    IX_CCI_ERR_GOTO( err != IX_SUCCESS,
                     IX_CCI_ERR_SHARED_MEM_ALLOC, ret_err1, err )
    ix_ossl_memset( *arg_ppUintBlock, 0, sizeUintBlock );



    /* Allocate space for a semaphore to control access to shared memory */
    err = IX_RM_MEM_ALLOC( IX_CCI_SHARED_MEMORY_TYPE, IX_CCI_MEMORY_CHANNEL, 
                           sizeof(ix_ossl_sem_t),
                           (void**)arg_ppSema );
    IX_CCI_ERR_GOTO( err != IX_SUCCESS,
                     IX_CCI_ERR_SHARED_MEM_ALLOC, ret_err2, err )

    return IX_SUCCESS;    /* Normal return */

    /* Error return path */
#if !defined(_CCT_EXCLUDE_)
ret_err2:
    cleanup_err = IX_RM_MEM_FREE( *arg_ppSema );
    IX_CCI_ERR_DUMP( cleanup_err );

ret_err1:
    cleanup_err = IX_RM_MEM_FREE( *arg_ppTkNotifyBlock );
    IX_CCI_ERR_DUMP( cleanup_err );

    return err;
#endif
}


static ix_error create_registry_properties( void )
{
    /* Create all the properties used by the CCI */
    ix_error err;
#if !defined(_CCT_EXCLUDE_)
    ix_error cleanup_err;
#endif
    err = ix_rm_cp_property_create( IX_CP_CORE_PROPERTY_ROOT, 
                  g_szCciSharedPtrMsgTkq,
                  0,
                  &m_aPropertyHandle[IX_CCI_PROPERTY_HANDLE_MSGTKQ] );
    IX_CCI_ERR_RET( err != IX_SUCCESS, IX_CCI_ERR_PROPERTY_CREATE )

                              
    err = ix_rm_cp_property_create( IX_CP_CORE_PROPERTY_ROOT, 
                  g_szCciSharedPtrShutdownFlags,
                  0,
                  &m_aPropertyHandle[IX_CCI_PROPERTY_HANDLE_SHUTDOWNFLAGS] );
    IX_CCI_ERR_GOTO( err != IX_SUCCESS,
                     IX_CCI_ERR_PROPERTY_CREATE, ret_err1, err )


    err = ix_rm_cp_property_create( IX_CP_CORE_PROPERTY_ROOT, 
                  g_szCciSharedPtrShmemSema,
                  0,
                  &m_aPropertyHandle[IX_CCI_PROPERTY_HANDLE_SEMA] );
    IX_CCI_ERR_GOTO( err != IX_SUCCESS,
                     IX_CCI_ERR_PROPERTY_CREATE, ret_err2, err )

    return IX_SUCCESS;    /* Normal return */

    /* Error return path */
#if !defined(_CCT_EXCLUDE_)
ret_err2:
    cleanup_err = ix_rm_cp_property_delete( 
            m_aPropertyHandle[IX_CCI_PROPERTY_HANDLE_SHUTDOWNFLAGS] );
    IX_CCI_ERR_DUMP( cleanup_err );

ret_err1:
    cleanup_err = ix_rm_cp_property_delete( 
            m_aPropertyHandle[IX_CCI_PROPERTY_HANDLE_MSGTKQ] );
    IX_CCI_ERR_DUMP( cleanup_err );

    return err;
#endif 
}

ix_error delete_registry_properties( void )
{
    /* Delete all the registry properties used by the CCI */
    ix_error err = ix_rm_cp_property_delete( 
        m_aPropertyHandle[IX_CCI_PROPERTY_HANDLE_SEMA] );

    ix_error err1 = ix_rm_cp_property_delete( 
            m_aPropertyHandle[IX_CCI_PROPERTY_HANDLE_SHUTDOWNFLAGS] );
    IX_CCI_ERR_FWD_OR_DUMP( err1, err )

    err1 = ix_rm_cp_property_delete( 
            m_aPropertyHandle[IX_CCI_PROPERTY_HANDLE_MSGTKQ] );
    IX_CCI_ERR_FWD_OR_DUMP( err1, err )

    return err;
}

static ix_error register_shared_mem_blocks(
                                            ix_uint32 arg_PropTkNotify,
                                            ix_uint32 arg_PropUint,
                                            ix_uint32 arg_PropSema
                                          )
{
    /** 
     * Register the three shared-memory blocks used by the CCI in 
     * the registry
     */
    ix_error err = ix_rm_cp_property_set_value_uint32( 
                m_aPropertyHandle[IX_CCI_PROPERTY_HANDLE_MSGTKQ],
                arg_PropTkNotify );
    IX_CCI_ERR_RET( err != IX_SUCCESS, IX_CCI_ERR_PROPERTY_SET )
                              

    err = ix_rm_cp_property_set_value_uint32(
                m_aPropertyHandle[IX_CCI_PROPERTY_HANDLE_SHUTDOWNFLAGS],
                arg_PropUint );
    IX_CCI_ERR_RET( err != IX_SUCCESS, IX_CCI_ERR_PROPERTY_SET )


    err = ix_rm_cp_property_set_value_uint32(
                m_aPropertyHandle[IX_CCI_PROPERTY_HANDLE_SEMA],
                arg_PropSema );
    IX_CCI_ERR_RET( err != IX_SUCCESS, IX_CCI_ERR_PROPERTY_SET )

    return IX_SUCCESS; /* Normal return */
}


