/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation.All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 5
 *
 * = FILENAME
 *      eci_leaf.c
 *
 * = DESCRIPTION
 *    Functions supporting constructs at the leaves of the execution control 
 *    tree.
 *
 * = AUTHOR
 *       Roland Arajs
 *       roland.arajs@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CHANGE HISTORY
 *      4/28/2002 2:27:08 PM - creation time
 *
 * ============================================================================
 * $Id: eci_leaf.c,v 1.21 2003/09/19 21:10:28 ageorges Exp $
 */

#include "ix_ossl.h"
#include "cci/cc.h"
#include "cci/policy.h"
#include "cci/internal/internal_engine.h"
#include "cci/internal/internal_eci_node.h"
#include "cci/internal/internal_cci_error.h"
#include "cci/internal/internal_cci_handles.h"
#include "rm/internal/internal_tk_input_notify.h"
#include "cci/internal/internal_tk_notify_if.h"
#include "cci/internal/internal_policy.h"


/**
 * Typedefs of private execution-control structures. There are two of them: one
 * for messages and one for packets. The only difference is in the type of 
 * handler. Since the typedef of a packet handler has the same signature as a
 * message handler, the handler could have been part of an ix_eci_leaf.
 * However, since the handlers each have their own typedefs and the function
 * signatures might diverge in a later release of this software, each handler
 * gets its own ix_eci_leaf-derived structure.
 */

/**
 * TYPENAME: ix_msg_eci_leaf
 * 
 * DESCRIPTION: Execution-control interface for message handlers
 *
 */
typedef struct ix_s_msg_eci_leaf
{
    ix_eci_leaf     tEciLeaf;
    ix_msg_handler  pMsgHandler;
} ix_msg_eci_leaf;

/**
 * TYPENAME: ix_pkt_eci_leaf
 * 
 * DESCRIPTION: Execution-control interface for packet handlers
 *
 */
typedef struct ix_s_pkt_eci_leaf
{
    ix_eci_leaf     tEciLeaf;
    ix_pkt_handler  pPktHandler;
} ix_pkt_eci_leaf;




/*******************************************************************************
 * Declarations of functions with local file scope only
 */

/**
 * NAME: get_cci_context
 *
 * DESCRIPTION: Gets user context from execution engine or CC, depending on
 *              the arg_hComponent handle type. Also returns engine's execution-
 *              control interface.
 * 
 * @Param:  - IN    arg_hComponent. Handle to execution engine or core component
 * @Param:  - OUT   arg_ppContext. Where function will return the context to be
 *                      used with this component and its handlers.
 * @Param:  - OUT   arg_ppEngineProcIf. Where function will return pointer to
 *                      engine's execution-control interface.
 * @Param:  - OUT   arg_pIsContextValid. Indicates whether the context returned
 *                      in arg2 is valid.
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
static ix_error get_cci_context(
                                 ix_handle arg_hComponent,
                                 void** arg_ppContext,
                                 ix_engine_proc_node** arg_ppEngineProcIf,
                                 ix_uint32* arg_pIsContextValid
                               );


/**
 * NAME: process_msg_tokens
 *
 * DESCRIPTION:   Execution-control process function for message handlers
 * 
 * @Param:  - IN    arg_pTokenProcessor: Pointer to ix_msg_eci_leaf structure
 * @Param:  - OUT   arg_pNumInputsPending: Where function returns whether there
 *                      are more inputs to be processed.
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
static ix_error process_msg_tokens(
                                    ix_cci_exe_ctrl_if* arg_pTokenProcessor,
                                    ix_uint32* arg_pPendingStatus,
                                    ix_uint32* arg_pProcCount,
                                    ix_engine_proc_node* arg_pEngProcNode
                                  );


/**
 * NAME: process_pkt_tokens
 *
 * DESCRIPTION:   Execution-control process function for packet handlers
 * 
 * @Param:  - IN    arg_pTokenProcessor: Pointer to ix_msg_eci_leaf structure
 * @Param:  - OUT   arg_pNumInputsPending: Where function returns whether there
 *                      are more inputs to be processed.
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
static ix_error process_pkt_tokens(
                                    ix_cci_exe_ctrl_if* arg_pTokenProcessor,
                                    ix_uint32* arg_pPendingStatus,
                                    ix_uint32* arg_pProcCount,
                                    ix_engine_proc_node* arg_pEngProcNode
                                  );


/**
 * NAME: eci_leaf_fini
 *
 * DESCRIPTION:   
 * 
 * @Param:  - IN    arg_pTokenProcessor: Pointer to eci message or packet leaf
 *                      to be terminated.
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
static ix_error eci_leaf_fini 
                    ( ix_cci_exe_ctrl_if* arg_pTokenProcessor );



/*******************************************************************************
 * Public APIs: Packet and message handlers
 */
ix_error ix_cci_cc_add_message_handler(
                                        ix_cc_handle arg_hComponent,
                                        ix_uint32 arg_InputId,
                                        ix_msg_handler arg_Handler,
                                        ix_input_type arg_SourceType
                                      )
{
    ix_error err;
    ix_msg_eci_leaf* pEciLeaf;
    ix_handle hMsgHandler;
    ix_tkp_entry* pEngChildEntry;
    void* pContext = NULL;
    ix_engine_proc_node* pEngProcIf = NULL;
    ix_uint32 vIsUserContextValid = 0;

    IX_CCI_PARAM_ASSERT( arg_InputId > IX_COMM_LAST_UBLOCK_ID 
                         && arg_InputId < 256 )
    IX_CCI_PARAM_ASSERT( arg_Handler != NULL )
    IX_CCI_PARAM_ASSERT( arg_SourceType < IX_INPUT_TYPE_LAST )
    
    /* Get context from execution engine or CC associated with handle arg */
    err = get_cci_context( arg_hComponent, &pContext, 
                           &pEngProcIf, &vIsUserContextValid );
    IX_CCI_DBG_IXERR_RET( err )
    /*
     * Create an ECI (execution-control interface) leaf and set up its
     * communication ID, handler, and engine/CC context
     */
    pEciLeaf = (ix_msg_eci_leaf*)IX_OSSL_MALLOC( sizeof(ix_msg_eci_leaf) );

    IX_CCI_ERR_RET( pEciLeaf==NULL, IX_CCI_ERR_OS_MEM_ALLOC );
    ix_ossl_memset( pEciLeaf, 0, sizeof(ix_msg_eci_leaf) );


    /* Calculate handle for token-notify interface */
    hMsgHandler = IX_CCI_HANDLE_TYPES_MSGTKQ 
                  | (arg_InputId - IX_COMM_LAST_UBLOCK_ID - 1);

    err = ix_cci_eci_leaf_init( (ix_eci_leaf*)pEciLeaf, process_msg_tokens,
                                eci_leaf_fini,
                                hMsgHandler,
                                pContext
                              );
    IX_CCI_DBG_IXERR_GOTO( err_ret1, err )

    pEciLeaf->pMsgHandler = arg_Handler;
    /*
     * If context is not set, save handle to engine/CC so that it can be
     * retrieved the first time the message handler is invoked in response
     * to a message on its comm ID.
     */
    if( !vIsUserContextValid )
        pEciLeaf->tEciLeaf.hUserCtxOwner = arg_hComponent;
    else
        pEciLeaf->tEciLeaf.hUserCtxOwner = 0;

    /* Initialize the token-notify interface */
    err = ix_cci_tk_notify_register_eci( hMsgHandler, pEciLeaf );
    IX_CCI_DBG_IXERR_GOTO( err_ret2, err )
    /*
     * Copy the engine's new-token indicator reference. The leaf proc fn will
     * refer to this to decide whether to check if a token has been received
     * in a part of the policy tree with a higher strict priority.
     */
    pEciLeaf->tEciLeaf.tEci.pTkInNotify->pIsNewToken = pEngProcIf->pIsNewToken;
    /*
     * Check engine to see if this is the first handler being added to the 
     * default policy (if no selected child, we still have a default policy
     * and this is the first message/packet handler. If so, add this handler
     * directly as child 0 of the engine, and add the handler to the
     * default policy.
     */
    pEngChildEntry = pEngProcIf->tNode.tChildren.pEntries;
    if( 
        (pEngProcIf->tNode.vSelectedChild | IX_CCI_NO_EVENT_PRESENT)
        == IX_CCI_NO_SELECTED_CHILD 
      )
    {
        ix_eci_node* pEngExeRoot;
        ix_eci_node* pMsgPolicyEci;

        /* Directly add handler as child 0  */
        pEngProcIf->tNode.vSelectedChild = 0;
        pEngProcIf->tNode.tChildren.pEntries[0].pEci
            = (ix_cci_exe_ctrl_if*)pEciLeaf;

        /* Get pointer to default message policy */
        pEngExeRoot = (ix_eci_node*)pEngChildEntry[1].pEci;
        pMsgPolicyEci = (ix_eci_node*)pEngExeRoot->tChildren.pEntries[0].pEci;
        
        /* Add handler to message policy */
        err = ix_cci_policy_add_leaf(
                                      pMsgPolicyEci->hSelf,
                                      arg_InputId,
                                      0, /* Message, not packet */
                                      IX_CCI_MESSAGE_RUN_GRANULARITY  /* Weight irrelevant for RR policy */
                                    );
        IX_CCI_IXERR_GOTO( err_ret2, err )

        /* Change handler's token-notify parent to be engine instead of policy */
        ix_cci_connect_parent(  pEciLeaf->tEciLeaf.tEci.pTkInNotify,
                                pEngProcIf->tNode.tEci.pTkInNotify
                              );
    }
    
    /*
     * Otherwise, compare engine's child 0 with child 2. If they are not the 
     * same, the default policy must be used.
     */
    else if( pEngChildEntry[0].pEci != pEngChildEntry[2].pEci )
    {
        ix_eci_node* pMsgPolicyEci;
        ix_eci_node* pEngExeRoot;
        ix_eci_node* pDefaultPolicyEci;
        pEngExeRoot = (ix_eci_node*)pEngChildEntry[0].pEci;
        /* Get pointer to default message policy */
        pDefaultPolicyEci = (ix_eci_node*)pEngChildEntry[1].pEci;
        pMsgPolicyEci = (ix_eci_node*)pDefaultPolicyEci->tChildren.pEntries[0].pEci;


        /**
         * Compare engine's child 0 with child 1. If they are the same, the
         * default policy is active. If not, the first handler is directly
         * selected by the execution engine. Change child 0 to be the same
         * as child 1 to activate the default policy tree.
         */
        if( pEngExeRoot != pDefaultPolicyEci )
        {
            /* 
             * Find out whether child 0 is a message or a packet and revert
             * the child's token-notify parent to the token-notify interface
             * of the message or packet policy
             */
            ix_cci_exe_ctrl_if* pMsgEci = 
                pMsgPolicyEci->tChildren.pEntries[0].pEci;
            if( (ix_cci_exe_ctrl_if*)pEngExeRoot == pMsgEci )
            {
                ix_cci_connect_parent( pMsgEci->pTkInNotify, 
                                       pMsgPolicyEci->tEci.pTkInNotify );
            }
            else
            {
                ix_eci_node* pPktPolicyEci = 
                    (ix_eci_node*)pDefaultPolicyEci->tChildren.pEntries[1].pEci;
                ix_cci_connect_parent( 
                    pPktPolicyEci->tChildren.pEntries[0].pEci->pTkInNotify,
                    pPktPolicyEci->tEci.pTkInNotify );
            }
            
            /**
             * Instead of direct connection between a leaf and the engine,
             * make the default tree the engine root
             */
            pEngChildEntry[0].pEci = (ix_cci_exe_ctrl_if*)pDefaultPolicyEci;
        }

        /* Add handler to default policy */
        err = ix_cci_policy_add_leaf(
                                      pMsgPolicyEci->hSelf,
                                      arg_InputId,
                                      0, /* Message, not packet */
                                      IX_CCI_MESSAGE_RUN_GRANULARITY  /* Weight irrelevant for RR policy */
                                    );
        IX_CCI_IXERR_GOTO( err_ret2, err )
    }   
    /*
     * Register ECI leaf message ring with RM
     */
    err = ix_rm_message_set_receive_mode( 
                                          arg_InputId, 
                                          IX_COMM_ID_MODE_GET_SELECT 
                                        );
    IX_CCI_ERR_GOTO( err != IX_SUCCESS, IX_CCI_ERR_MSG_SET_MODE, err_ret2, err )
    
    err = ix_rm_message_set_consumer_mode( arg_InputId, 0 );
    IX_CCI_ERR_GOTO( err != IX_SUCCESS, IX_CCI_ERR_MSG_SET_MODE, err_ret2, err )

    err = ix_rm_message_set_producer_mode( arg_InputId, arg_SourceType );
    IX_CCI_ERR_GOTO( err != IX_SUCCESS, IX_CCI_ERR_MSG_SET_MODE, err_ret2, err )

    err = ix_rm_message_enable_cc_mode( arg_InputId, 
                                        pEciLeaf->tEciLeaf.tEci.pTkInNotify 
                                      );
    IX_CCI_ERR_GOTO( err != IX_SUCCESS, IX_CCI_ERR_MSG_SET_MODE, err_ret2, err )
    return IX_SUCCESS;

    /* Code only gets here if an error has occurred */
#if !defined(_CCT_EXCLUDE_)
err_ret2:
    {
        ix_error cleanup_err = eci_leaf_fini( (ix_cci_exe_ctrl_if*)pEciLeaf );
        IX_CCI_ERR_DUMP( cleanup_err );
        return err;
    }
#if !defined(_CCI_ERROR_PATH_TEST_) && defined(_IX_CCI_DEBUG_)
err_ret1:
    IX_OSSL_FREE( pEciLeaf );
    return err;
#endif
#endif  /* _CCT_EXCLUDE_ */
}


ix_error ix_cci_cc_remove_message_handler( ix_uint32 arg_InputID )
{
    ix_error err;
    ix_tk_input_notify* pNotifyIf = NULL;
    ix_cci_exe_ctrl_if* pEciIf;

    /* Unregister from RM */
    IX_CCI_PARAM_ASSERT( arg_InputID > IX_COMM_LAST_UBLOCK_ID 
                         && arg_InputID < 256 )
    err = ix_rm_message_disable_cc_mode( arg_InputID, (void**)&pNotifyIf );
    IX_CCI_ERR_RET( err != IX_SUCCESS, IX_CCI_ERR_DISABLE_MSG_HANDLER )    

    /* Close down the execution-control interface and free memory resources */
    pEciIf = (ix_cci_exe_ctrl_if*)pNotifyIf->pProcIfContext;
    err = eci_leaf_fini( pEciIf );
    IX_CCI_ERR_DUMP( err ); 
    return err;
}


ix_error ix_cci_cc_add_packet_handler(
                                        ix_cc_handle arg_hComponent,
                                        ix_uint32 arg_InputId,
                                        ix_pkt_handler arg_Handler,
                                        ix_input_type arg_SourceType
                                     )
{
    ix_error err;
    ix_pkt_eci_leaf* pEciLeaf;
    ix_handle hPktHandler;
    ix_tkp_entry* pEngChildEntry;
    void* pContext = NULL;
    ix_engine_proc_node* pEngProcIf = NULL;
    ix_uint32 vIsValidContext = 0;

    IX_CCI_PARAM_ASSERT( arg_InputId > IX_COMM_LAST_UBLOCK_ID 
                         && arg_InputId < 256 )
    IX_CCI_PARAM_ASSERT( arg_Handler != NULL )
    IX_CCI_PARAM_ASSERT( arg_SourceType < IX_INPUT_TYPE_LAST )
    /* Get context from execution engine or CC associated with handle arg */
    err = get_cci_context( arg_hComponent, &pContext, 
                           &pEngProcIf, &vIsValidContext );
    IX_CCI_IXERR_RET( err )
    /*
     * Create an ECI (execution-control interface) leaf and set up its
     * communication ID, handler, and engine/CC context
     */
    pEciLeaf = (ix_pkt_eci_leaf*)IX_OSSL_MALLOC( sizeof(ix_pkt_eci_leaf) );
    IX_CCI_ERR_RET( pEciLeaf==NULL, IX_CCI_ERR_OS_MEM_ALLOC );
    ix_ossl_memset( pEciLeaf, 0, sizeof(ix_msg_eci_leaf) );


    /* Calculate handle for token-notify interface */
    hPktHandler = IX_CCI_HANDLE_TYPES_PKTTKQ 
                  | (arg_InputId - IX_COMM_LAST_UBLOCK_ID - 1);

    err = ix_cci_eci_leaf_init( (ix_eci_leaf*)pEciLeaf, process_pkt_tokens,
                                eci_leaf_fini,
                                hPktHandler,
                                pContext
                              );
    IX_CCI_DBG_IXERR_GOTO( err_ret1, err )

    pEciLeaf->pPktHandler = arg_Handler;
    /*
     * If context is not set, save handle to engine/CC so that it can be
     * retrieved the first time the message handler is invoked in response
     * to a message on its comm ID.
     */
    if( !vIsValidContext )
        pEciLeaf->tEciLeaf.hUserCtxOwner = arg_hComponent;
    else
        pEciLeaf->tEciLeaf.hUserCtxOwner = 0;

    /* Initialize the token-notify interface */
    err = ix_cci_tk_notify_register_eci( hPktHandler, pEciLeaf );
    IX_CCI_DBG_IXERR_GOTO( err_ret2, err )
    /*
     * Copy the engine's new-token indicator reference. The leaf proc fn will
     * refer to this to decide whether to check if a token has been received
     * in a part of the policy tree with a higher strict priority.
     */
    pEciLeaf->tEciLeaf.tEci.pTkInNotify->pIsNewToken = pEngProcIf->pIsNewToken;
    /*
     * Check engine to see if this is the first handler being added to the 
     * default policy (if no selected child, we still have a default policy
     * and this is the first message/packet handler. If so, add this handler
     * directly as child 0 of the engine, and add the handler to the
     * default policy.
     */
    pEngChildEntry = pEngProcIf->tNode.tChildren.pEntries;
    if( 
        (pEngProcIf->tNode.vSelectedChild | IX_CCI_NO_EVENT_PRESENT)
        == IX_CCI_NO_SELECTED_CHILD 
      )
    {
        ix_eci_node* pEngExeRoot;
        ix_eci_node* pPktPolicyEci;

        /* Directly add handler as child 0  */
        pEngProcIf->tNode.vSelectedChild = 0;
        pEngProcIf->tNode.tChildren.pEntries[0].pEci
            = (ix_cci_exe_ctrl_if*)pEciLeaf;

        /* Get pointer to default message policy */
        pEngExeRoot = (ix_eci_node*)pEngChildEntry[1].pEci;
        pPktPolicyEci = (ix_eci_node*)pEngExeRoot->tChildren.pEntries[1].pEci;
        
        /* Add handler to message policy */
        err = ix_cci_policy_add_leaf(
                                      pPktPolicyEci->hSelf,
                                      arg_InputId,
                                      1, /* Packet, not message */
                                      IX_CCI_PACKET_RUN_GRANULARITY  /* Weight irrelevant for RR policy */
                                    );
        IX_CCI_IXERR_GOTO( err_ret2, err )

        /* Change handler's token-notify parent to be engine instead of policy */
        ix_cci_connect_parent(  pEciLeaf->tEciLeaf.tEci.pTkInNotify,
                                pEngProcIf->tNode.tEci.pTkInNotify
                              );

    }
    
    /*
     * Otherwise, compare engine's child 0 with child 2. If they are not the 
     * same, the default policy must be used.
     */
    else if( pEngChildEntry[0].pEci != pEngChildEntry[2].pEci )
    {
        ix_eci_node* pPktPolicyEci;
        ix_eci_node* pEngExeRoot;
        ix_eci_node* pDefaultPolicyEci;
        pEngExeRoot = (ix_eci_node*)pEngChildEntry[0].pEci;
        /* Get pointer to default packet policy */
        pDefaultPolicyEci = (ix_eci_node*)pEngChildEntry[1].pEci;
        pPktPolicyEci = (ix_eci_node*)pDefaultPolicyEci->tChildren.pEntries[1].pEci;


        /**
         * Compare engine's child 0 with child 1. If they are the same, the
         * default policy is active. If not, the first handler is directly
         * selected by the execution engine. Change child 0 to be the same
         * as child 1 to activate the default policy tree.
         */
        if( pEngExeRoot != pDefaultPolicyEci )
        {
            /* 
             * Find out whether child 0 is a message or a policy and revert
             * the child's token-notify parent to the token-notify interface
             * of the message or packet policy
             */
            ix_cci_exe_ctrl_if* pPktEci =
                pPktPolicyEci->tChildren.pEntries[0].pEci;
            if( (ix_cci_exe_ctrl_if*)pEngExeRoot == pPktEci )
            {
                ix_cci_connect_parent( 
                                       pPktEci->pTkInNotify,
                                       pPktPolicyEci->tEci.pTkInNotify
                                     );
            }
            else
            {
                ix_eci_node* pMsgPolicyEci = 
                    (ix_eci_node*)pDefaultPolicyEci->tChildren.pEntries[0].pEci;
                ix_cci_connect_parent( 
                    pMsgPolicyEci->tChildren.pEntries[0].pEci->pTkInNotify,
                    pMsgPolicyEci->tEci.pTkInNotify );
            }
            
            /**
             * Instead of direct connection between a leaf and the engine,
             * make the default tree the engine root
             */
            pEngChildEntry[0].pEci = (ix_cci_exe_ctrl_if*)pDefaultPolicyEci;
        }

        /* Add handler to default policy */
        err = ix_cci_policy_add_leaf(
                                      pPktPolicyEci->hSelf,
                                      arg_InputId,
                                      1, /* Packet, not message */
                                      IX_CCI_PACKET_RUN_GRANULARITY  /* Weight irrelevant for RR policy */
                                    );
        IX_CCI_IXERR_GOTO( err_ret2, err )
    }   
    /*
     * Register ECI leaf packet ring with RM
     */
    err = ix_rm_packet_set_receive_mode( 
                                         arg_InputId, 
                                         IX_COMM_ID_MODE_GET_SELECT 
                                       );
    IX_CCI_ERR_GOTO( err != IX_SUCCESS, IX_CCI_ERR_PKT_SET_MODE, err_ret2, err )
    
    err = ix_rm_packet_set_consumer_mode( arg_InputId, 0 );
    IX_CCI_ERR_GOTO( err != IX_SUCCESS, IX_CCI_ERR_PKT_SET_MODE, err_ret2, err )

    err = ix_rm_packet_set_producer_mode( arg_InputId, arg_SourceType );
    IX_CCI_ERR_GOTO( err != IX_SUCCESS, IX_CCI_ERR_PKT_SET_MODE, err_ret2, err )

    err = ix_rm_packet_enable_cc_mode( arg_InputId, 
                                       pEciLeaf->tEciLeaf.tEci.pTkInNotify 
                                     );
    IX_CCI_ERR_GOTO( err != IX_SUCCESS, IX_CCI_ERR_PKT_SET_MODE, err_ret2, err )
    return IX_SUCCESS;

    /* Code only gets here if an error has occurred */
#if !defined(_CCT_EXCLUDE_)
err_ret2:
    {
        ix_error cleanup_err = eci_leaf_fini( (ix_cci_exe_ctrl_if*)pEciLeaf );
        IX_CCI_ERR_DUMP( cleanup_err );
        return err;
    }
#if !defined(_CCI_ERROR_PATH_TEST_) && defined(_IX_CCI_DEBUG_)
err_ret1:
    IX_OSSL_FREE( pEciLeaf );
    return err;
#endif
#endif  /* _CCT_EXCLUDE_ */
}



ix_error ix_cci_cc_remove_packet_handler( ix_uint32 arg_InputID )
{
    ix_error err;
    ix_cci_exe_ctrl_if* pEciIf;
    ix_tk_input_notify* pNotifyIf = NULL;

    /* Unregister from RM */
    IX_CCI_PARAM_ASSERT( arg_InputID > IX_COMM_LAST_UBLOCK_ID 
                         && arg_InputID < 256 )
    err = ix_rm_packet_disable_cc_mode( arg_InputID, (void**)&pNotifyIf );
    IX_CCI_ERR_RET( err != IX_SUCCESS, IX_CCI_ERR_DISABLE_PKT_HANDLER )    

    /* Close down the execution-control interface and free memory resources */
    pEciIf = (ix_cci_exe_ctrl_if*)pNotifyIf->pProcIfContext;
    err = eci_leaf_fini( pEciIf );
    IX_CCI_ERR_DUMP( err ); 
    return err;
}



ix_error ix_cci_send_message(
                              ix_uint32 arg_OutputId,
                              ix_buffer_handle arg_Handle,
                              ix_uint32 arg_UserData
                            )
{
    /** 
     * At present this function just encapsulates the send-message 
     * function in the RM 
     */
    return ix_rm_message_send( arg_OutputId, arg_Handle, arg_UserData );
}


ix_error ix_cci_send_packet(
                             ix_uint32 arg_OutputId,
                             ix_buffer_handle arg_Handle
                           )
{
    /** 
     * At present this function just encapsulates the send-packet 
     * function in the RM 
     */
    return ix_rm_packet_send( arg_OutputId, arg_Handle, 0 );
}


/*******************************************************************************
 * Functions called from other modules in the CCI
 */

ix_error 
ix_cci_eci_leaf_init(
                      ix_eci_leaf*               arg_pEciLeaf,
                      ix_cci_process_tokens_ptr  arg_ProcessTokensFuncPtr,
                      ix_cci_eci_fini_ptr        arg_FiniFuncPtr,
                      ix_tk_handle               arg_TkHandle,
                      void*                      arg_pUserContext
                    )
{
    /* Initialize base common token interface construct */
    ix_error err = ix_cci_eci_init( (ix_cci_exe_ctrl_if*)arg_pEciLeaf,
                     arg_ProcessTokensFuncPtr, arg_FiniFuncPtr, arg_TkHandle );
    IX_CCI_DBG_IXERR_RET( err )

    /* Copy the user context for passing to the handler */
    arg_pEciLeaf->pUserContext = arg_pUserContext;

    /* Convert handle to communication ID */
    arg_pEciLeaf->vCommId = (arg_TkHandle & IX_CCI_TKQ_IDX_MASK)
                            + IX_COMM_LAST_UBLOCK_ID + 1;
    return IX_SUCCESS;
}






/*******************************************************************************
 * Implementation of functions with local file scope only
 */
static ix_error get_cci_context(
                                 ix_handle arg_hComponent,
                                 void** arg_ppContext,
                                 ix_engine_proc_node** arg_ppEngineProcIf,
                                 ix_uint32* arg_pIsContextValid
                               )
{
    /**
     * Convert the handle to an engine index, and use this to get access to the 
     * execution-control interface via the token-notify interface. 
     */
    ix_engine_proc_node* pEngEci;
    int idxEng = IX_HAND_TO_ENGINE_IDX( arg_hComponent );
    IX_CCI_PARAM_ASSERT( idxEng != IX_CCI_INVALID_INDEX )

     *arg_ppEngineProcIf 
        = pEngEci = (ix_engine_proc_node*)g_pEngine[idxEng].pProcIfContext;
    
    /*
     * If handle is for an execution engine, get context directly from the
     * engine. If context is NULL check engine's user-init-completion status to
     * ascertain whether the context should be deemed to be valid.
     */
    if( IX_CCI_GET_HANDLE_TYPE(arg_hComponent) == (ix_uint32)IX_CCI_HANDLE_TYPES_EXE )
    {
        *arg_ppContext = pEngEci->pUserContext;
        *arg_pIsContextValid = pEngEci->vIsUserInitComplete;
    }
    
    /*
     * Otherwise get CC index from handle and use that to index into the 
     * engine's CC list. Get context from indexed CC.
     */
    else
    {
        ix_cc_node* pCCListEntry;
        int idxCC = IX_HAND_TO_CC_IDX( arg_hComponent );
        IX_CCI_ASSERT( idxEng != IX_CCI_INVALID_INDEX )
        pCCListEntry = pEngEci->pCCList;
        while( idxCC-- > 0 )
        {
            pCCListEntry = pCCListEntry->pNext;
        }
        *arg_ppContext = pCCListEntry->pUserContext;
        /*
         * If context is NULL, check CC's user-init_completion status to 
         * ascertain whether the context should be deemed to be valid.
         */
        *arg_pIsContextValid = pCCListEntry->vIsUserInitComplete;

    }
    return IX_SUCCESS;
}


static ix_error process_msg_tokens 
                    (
                      ix_cci_exe_ctrl_if* arg_pTokenProcessor,
                      ix_uint32* arg_pPendingStatus,
                      ix_uint32* arg_pProcCount,
                      ix_engine_proc_node* arg_pEngProcNode
                    )
{
    ix_error err;
    ix_buffer_handle hMsgBuf = 0;
    ix_uint32 vUserData = 0;
    ix_msg_eci_leaf* pMsgLeaf = (ix_msg_eci_leaf*)arg_pTokenProcessor;
    ix_uint32 vCommId = pMsgLeaf->tEciLeaf.vCommId;
    ix_uint32 vNumPending = 0;

    /* See if there are any messages pending */
    err = ix_rm_message_peek( vCommId, &vNumPending );
    IX_CCI_ERR_RET( err != IX_SUCCESS, IX_CCI_ERR_MSG_PEEK )
    if( vNumPending == 0 )   /* Return if nothing to process */
    {
        return IX_SUCCESS;
    }

    err = ix_cci_handle_events( arg_pEngProcNode );

#if defined(_IX_CCI_DEBUG_)
    IX_CCI_IXERR_LOG_RET_IF_FATAL( err, IX_CCI_ERR_EVENT_HANDLER );
#endif /* defined(_IX_CCI_DEBUG_) */

    /* 
     * Check new-token indicator, and if set, check for high-priority 
     * override 
     */
    if( *(arg_pEngProcNode->pIsNewToken) )
    {
        ix_uint32 vOverride =
            ix_cci_is_high_priority_override( 
                                      pMsgLeaf->tEciLeaf.tEci.pTkInNotify,
                                      arg_pEngProcNode 
                                    );
        if( vOverride )
        {
            *arg_pPendingStatus = IX_CCI_DATA_STILL_PENDING 
                                  | IX_CCI_HIGH_PRIORITY_OVERRIDE;
            return IX_SUCCESS;
        }
    }
    

    /* If leaf's user context is invalid, get the context from the engine or CC */
    if( pMsgLeaf->tEciLeaf.hUserCtxOwner )
    {
        void* pContext;
        ix_uint32 vIsContextValid = 0;
        ix_engine_proc_node* pEngProcNode;
        err = get_cci_context( 
                               pMsgLeaf->tEciLeaf.hUserCtxOwner,
                               &pContext,
                               &pEngProcNode,
                               &vIsContextValid
                             );
        IX_CCI_IXERR_RET( err )
        IX_CCI_ERR_RET( !vIsContextValid, IX_CCI_ERR_CC_INIT_INCOMPLETE )
        pMsgLeaf->tEciLeaf.pUserContext = pContext;
        pMsgLeaf->tEciLeaf.hUserCtxOwner = 0;
    }

    do
    {

        /* Get the message */
        err = ix_rm_message_get( vCommId, &hMsgBuf, &vUserData );
        /* IX_CCI_ERR_RET( err != IX_SUCCESS, IX_CCI_ERR_MESSAGE_GET ) */
        
        /* if there is no more data just exit the loop */
        if(err != IX_SUCCESS)
        {
            vNumPending = 0;
            break;
        }

    
        /* Call the user-registered message handler */
        err = pMsgLeaf->pMsgHandler( hMsgBuf, vUserData, 
                                     pMsgLeaf->tEciLeaf.pUserContext );

        /*
         * Following CCI error macro call may be temporary, depending on whether
         * CCI will handle chained errors or not.
         */
/*        IX_CCI_IXERR_LOG_RET_IF_FATAL( err, IX_CCI_ERR_MSG_HANDLER ) */

    }
    while((--vNumPending > 0) && (--(*arg_pProcCount) > 0));

    /* Check queue status again if there was only one token on the ring */
    if(vNumPending == 0 )
    {
        err = ix_rm_message_peek( vCommId, &vNumPending );
        IX_CCI_ERR_RET( err != IX_SUCCESS, IX_CCI_ERR_PKT_PEEK )
        if( vNumPending != 0 )
            *arg_pPendingStatus = IX_CCI_DATA_STILL_PENDING;
    }
    else
        *arg_pPendingStatus = IX_CCI_DATA_STILL_PENDING;


    return IX_SUCCESS;
}

static ix_error process_pkt_tokens(
                                    ix_cci_exe_ctrl_if* arg_pTokenProcessor,
                                    ix_uint32* arg_pPendingStatus,
                                    ix_uint32* arg_pProcCount,
                                    ix_engine_proc_node* arg_pEngProcNode
                                  )
{
    ix_error err;
    ix_buffer_handle hPktBuf = 0;
    ix_uint32 vUserData = 0;
    ix_uint32 vNumPending = 0;
    ix_pkt_eci_leaf* pPktLeaf = (ix_pkt_eci_leaf*)arg_pTokenProcessor;
    ix_uint32 vCommId = pPktLeaf->tEciLeaf.vCommId;
    
    /* See if there are any messages pending */
    err = ix_rm_packet_peek( vCommId, &vNumPending );
    IX_CCI_ERR_RET( err != IX_SUCCESS, IX_CCI_ERR_PKT_PEEK )
    if( vNumPending == 0 )   /* Return if nothing to process */
    {
        return IX_SUCCESS;
    }

    err = ix_cci_handle_events( arg_pEngProcNode );

#if defined(_IX_CCI_DEBUG_)
    IX_CCI_IXERR_LOG_RET_IF_FATAL( err, IX_CCI_ERR_EVENT_HANDLER );
#endif /* defined(_IX_CCI_DEBUG_) */

    /* 
     * Check new-token indicator, and if set, check for high-priority 
     * override 
     */
    if( *(arg_pEngProcNode->pIsNewToken) )
    {
        ix_uint32 vOverride =
            ix_cci_is_high_priority_override( 
                                      pPktLeaf->tEciLeaf.tEci.pTkInNotify,
                                      arg_pEngProcNode 
                                    );
        /**
         * If override, exit to allow the high-priority queue to be serviced.
         * Since there are still tokens to be processed here, set the pending
         * flag so that the execution-ctrl code will return to process this
         * queue later.
         */
        if( vOverride )
        {
            *arg_pPendingStatus = IX_CCI_DATA_STILL_PENDING 
                                  | IX_CCI_HIGH_PRIORITY_OVERRIDE;
            return IX_SUCCESS;
        }
    }

    /* If leaf's user context is invalid, get the context from the engine or CC */
    if( pPktLeaf->tEciLeaf.hUserCtxOwner )
    {
        void* pContext;
        ix_engine_proc_node* pEngProcNode;
        ix_uint32 vIsContextValid = 0;
        err = get_cci_context( 
                               pPktLeaf->tEciLeaf.hUserCtxOwner,
                               &pContext,
                               &pEngProcNode,
                               &vIsContextValid
                             );
        IX_CCI_IXERR_RET( err )
        IX_CCI_ERR_RET( !vIsContextValid, IX_CCI_ERR_CC_INIT_INCOMPLETE )
        pPktLeaf->tEciLeaf.pUserContext = pContext;
        pPktLeaf->tEciLeaf.hUserCtxOwner = 0;
    }

    do
    {
        err = ix_rm_packet_get( vCommId, &hPktBuf, &vUserData );
        /* IX_CCI_ERR_RET( err != IX_SUCCESS, IX_CCI_ERR_PACKET_GET ) */

        /* if there is no more data just exit the loop */
        if(err != IX_SUCCESS)
        {
            vNumPending = 0;
            break;
        }

        /* Call the user-registered packet handler */

        err = pPktLeaf->pPktHandler( hPktBuf, vUserData, 
                                     pPktLeaf->tEciLeaf.pUserContext );

/*        IX_CCI_IXERR_LOG_RET_IF_FATAL( err, IX_CCI_ERR_PKT_HANDLER ); */
    }
    while((--vNumPending > 0) && (--(*arg_pProcCount) > 0));

    if(vNumPending == 0 )
    {
        err = ix_rm_packet_peek( vCommId, &vNumPending );
        IX_CCI_ERR_RET( err != IX_SUCCESS, IX_CCI_ERR_PKT_PEEK )
        if( vNumPending != 0 )
            *arg_pPendingStatus = IX_CCI_DATA_STILL_PENDING;
    }
    else
        *arg_pPendingStatus = IX_CCI_DATA_STILL_PENDING;


    return IX_SUCCESS;
}

static ix_error eci_leaf_fini 
                    ( ix_cci_exe_ctrl_if* arg_pTokenProcessor )
{
    /**
     * Set pointer to this leaf in parent policy to NULL
     */
    ix_tk_input_notify* pParent = arg_pTokenProcessor->pTkInNotify->pParent;
    if( pParent == NULL )   /* Not added to a policy or engine */
    {
        IX_OSSL_FREE( arg_pTokenProcessor );
    }
    else if( pParent->pParent == NULL )
    {
        /**
         * This handler is the only handler being run under the execution
         * engine and is directly connected to the engine instead of the
         * default policy.
         * Call engine function to remove the leaf 
         */
        ix_cci_exe_remove_leaf( pParent );
    }
    else   /* parent is policy (normal case) */
    {
        ix_cci_policy_clear_entry( pParent, arg_pTokenProcessor );
        IX_OSSL_FREE( arg_pTokenProcessor );
    }
    return IX_SUCCESS;
}
