/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation.All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 5
 *
 * = FILENAME
 *      engine.c
 *
 * = DESCRIPTION
 *      Source code for execution engine.  
 *
 * = AUTHOR
 *       Roland Arajs
 *       roland.arajs@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CHANGE HISTORY
 *      4/9/2002 9:34:47 AM - creation time
 *
 * ============================================================================
 * $Id: engine.c,v 1.15 2003/08/18 21:48:17 ageorges Exp $
 */

#include "ix_ossl.h"
#include "ix_rm.h"

#include "cci/cci_defs.h"
#include "cci/internal/internal_engine.h"
#include "cci/internal/internal_cci_handles.h"


ix_cci_eng_thread_args g_aThreadArgs[IX_CCI_MAX_ENGINES];
static ix_uint8* pShutdownFlags;
static ix_ossl_sem_t* m_pSema;
static ix_uint32    m_vEngineCount;


/**
 * Public function definitions
 */
ix_error ix_cci_exe_run(
                         const char* arg_FileName,
                         ix_exe_init arg_InitFunc,
                         ix_exe_fini arg_FiniFunc,
                         const char* arg_Name,
                         ix_exe_handle* arg_pHandle
                       )
{
    ix_error err = 0;
    ix_ossl_thread_t tid;
    ix_cci_eng_thread_args* pThreadArgs;
    volatile ix_uint8* pShutdownFlag;
    ix_configuration_property_handle hShutdownFlagsProperty = 0;

    IX_CCI_PARAM_ASSERT( arg_InitFunc != NULL );
    IX_CCI_PARAM_ASSERT( arg_FiniFunc != NULL );
    IX_CCI_PARAM_ASSERT( arg_pHandle != NULL );

    /* If entering this function for the first time, get the pointer to the
     * global semaphore in shared memory
     */
    if( m_vEngineCount == 0 ) 
    {
        ix_configuration_property_handle hSemaProperty;
            /* Get the shared memory addresses initialized in ix_cci_init()  */
        err = ix_rm_cp_property_open( IX_CP_CORE_PROPERTY_ROOT,
                                      g_szCciSharedPtrShmemSema,
                                      &hSemaProperty );
        IX_CCI_ERR_RET( err, IX_CCI_ERR_ENG_INIT_SEMA_PROP_READ )

        err = ix_rm_cp_property_get_value_uint32( hSemaProperty,
                                                 (ix_uint32*) &m_pSema );
        
        ix_rm_cp_property_close( hSemaProperty );
        IX_CCI_ERR_RET( err != IX_SUCCESS, IX_CCI_ERR_ENG_PROP_CLOSE )

        err = ix_rm_cp_property_open( IX_CP_CORE_PROPERTY_ROOT,
                                      g_szCciSharedPtrShutdownFlags,
                                      &hShutdownFlagsProperty );
        IX_CCI_ERR_RET( err != IX_SUCCESS,
                         IX_CCI_ERR_ENG_INIT_ENGINE_PROP_READ )

        ix_rm_cp_property_get_value_uint32( hShutdownFlagsProperty,
                                            (ix_uint32*)&pShutdownFlags );

        err = ix_rm_cp_property_close( hShutdownFlagsProperty );
        IX_CCI_ERR_RET( err != IX_SUCCESS, IX_CCI_ERR_ENG_PROP_CLOSE )
    }
    pShutdownFlag = pShutdownFlags + m_vEngineCount;
    ix_ossl_sem_take( *m_pSema, IX_OSSL_WAIT_FOREVER );
    /* Semaphore is released in engine thread after global engine structures
     * and variables are updated.
     */
    *arg_pHandle = IX_ENGINE_IDX_TO_HANDLE( m_vEngineCount );
    IX_CCI_ASSERT( *arg_pHandle != IX_CCI_HANDLE_TYPES_INVALID );
    pThreadArgs = g_aThreadArgs + m_vEngineCount++;
    pThreadArgs->pInitFunc = arg_InitFunc;
    pThreadArgs->pFiniFunc = arg_FiniFunc;

    if( arg_FileName )
    {
        /* Only use code here in OS's that have kernels with partitioned address
         * space. Spawn process/task/thread using O/S-defined mechanism. This is
         * a placeholder only. We support VxWorks and Linux, which both use a
         * monolithic address space in the kernel.
         */
        return err;
    }
    
    err = ix_ossl_thread_create( ix_cci_run_engine, pThreadArgs, arg_Name, &tid );
    IX_CCI_ERR_RET( err != IX_SUCCESS, IX_CCI_ERR_ENG_THREAD_SPAWN );

    /* Wait until the the spawned thread has actually started */
    while( *pShutdownFlag != IX_ENG_THREAD_AWAKE )
    {
        /*
         * If shutdown flag is IX_ENG_SHUTDOWN_COMPLETE, an error has occurred
         * during engine resource allocation; the global semaphore will be set
         * until we have had a chance to read the engine error status.
         */
        if( *pShutdownFlag == IX_ENG_SHUTDOWN_COMPLETE )
        {
            ix_error* pShutdownStatus  = ((ix_error*)pShutdownFlags) 
                                         + IX_CCI_NUM_SHUTDOWN_WORDS - 1;
            ix_ossl_sem_give( *m_pSema );
            m_vEngineCount--;
            return *pShutdownStatus;
        }
        ix_ossl_sleep( 1 );
    }

    return IX_SUCCESS;
}


ix_error ix_cci_exe_shutdown( ix_exe_handle arg_Handle )
{
    /* Convert handle to index in shared-memory engine array */ 
    ix_uint32 engIdx = IX_HAND_TO_ENGINE_IDX( arg_Handle );
    ix_engine_proc_node* pEngine;
    ix_ossl_sem_t* pEngSema;
    volatile ix_uint8* pShutdownFlag;
    volatile ix_uint32* pShutdownStatus = ((ix_uint32*)pShutdownFlags) 
                                    + IX_CCI_NUM_SHUTDOWN_WORDS - 1;
    IX_CCI_PARAM_ASSERT( 
        IX_CCI_GET_HANDLE_TYPE(arg_Handle) == IX_CCI_HANDLE_TYPES_EXE );
    IX_CCI_PARAM_ASSERT( engIdx != IX_CCI_INVALID_INDEX );
    /* 
     * Calculate which shared-memory shutdown byte correponds to this
     * execution engine, and set to start shutdown.
     */
    pShutdownFlag = pShutdownFlags + engIdx;

    /* Check engine has not already shut down due to an error */
    if( *pShutdownFlag != IX_ENG_SHUTDOWN_COMPLETE )
    {
        /* Tell engine to shut down  */
        *pShutdownFlag = IX_ENG_SHUTDOWN_START;

        /* Get pointer to engine's execution control i/f from toke-notify i/f */
        pEngine = (ix_engine_proc_node*)g_pEngine[engIdx].pProcIfContext;
        pEngSema = &pEngine->tNode.tEci.pTkInNotify->vSema;
        ix_ossl_sem_give(*pEngSema);

        /*
         * Wait for engine to set its shutdown flag to indicate shutdown completion.
         * Note that the although the following code is not optimal as far as 
         * execution efficiency is concerned, it is preferable to allocating a
         * semaphore in RM shared-memory space for this purpose alone.
         */
        while( *pShutdownFlag != IX_ENG_SHUTDOWN_COMPLETE )
        {
            ix_ossl_sleep_tick( 1 );
        }
    }
    /* 
     * Global semaphore is taken by engine while storing its exit status 
     * Give it back here.    
     */
    ix_ossl_sem_give( *m_pSema );
    if( m_vEngineCount )
        m_vEngineCount--;
    return *pShutdownStatus;
}


