/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation.All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 5
 *
 * = FILENAME
 *      engine_impl.cpp
 *
 * = DESCRIPTION
 *      This module contains the engine code that runs in the engine thread
 *
 * = AUTHOR
 *       Roland Arajs
 *       roland.arajs@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CHANGE HISTORY
 *      4/26/2002 10:50:31 AM - creation time
 *
 * ============================================================================
 * $Id: engine_impl.c,v 1.29 2003/12/11 22:25:28 ageorges Exp $
 */

#include "ix_ossl.h"

#if !(_IX_OS_TYPE == _IX_OS_LINUX_KERNEL_)
#include <stddef.h>
#endif /*(_IX_OS_TYPE == _IX_OS_LINUX_KERNEL_) */

#include "ix_rm.h"

#include "cci/cci_defs.h"
#include "cci/engine.h"
#include "cci/policy.h"
#include "cci/internal/internal_engine.h"
#include "cci/internal/internal_eci_node.h"
#include "cci/internal/internal_tk_notify_if.h"
#include "cci/internal/internal_cci_error.h"
#include "cci/internal/internal_cci_handles.h"


/**
 * TYPENAME: ix_exe_loop_ctrl
 * 
 * DESCRIPTION: Controls whether the execution-control loop is active or inactive
 *              and waiting for shutdown, or has received an unhandled fatal 
 *              error (causing the code to exit the loop immediately)
 *
 */
typedef enum ix_e_exe_loop_ctrl
{
    IX_EXE_LOOP_CTRL_ACTIVE,
    IX_EXE_LOOP_CTRL_WAIT_FOR_SHUTDOWN,
    IX_EXE_LOOP_CTRL_ABORT
} ix_exe_loop_ctrl;


/**
 * Shared memory pointers to core component infrastructure objects
 */
static ix_uint32   m_IsSharedSpaceInit;
ix_tk_input_notify*     g_pMsgTkq; 
ix_tk_input_notify*     g_pPktTkq;
ix_tk_input_notify*     g_pPolicy;
ix_tk_input_notify*     g_pEngine;
ix_uint8*               g_pShutdownFlags;
ix_uint32*              g_pEngShutdownStatus;
ix_uint32*              g_pEngineCount;
ix_uint32*              g_pPolicyCount;
ix_uint32*              g_pDefaultEngineIdx;

ix_ossl_sem_t* g_pShmemSema;

ix_ferror_func* g_pFatalErrorHandler;
void** g_ppFatalErrorContext;


/*******************************************************************************
 * Internal (private) function declarations
 */

/**
 * NAME: init_shared_memory
 *
 * DESCRIPTION:   Initializes global pointers to shared memory allocated by the
 *                ix_cci_init() function.
 * 
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
static void ix_cci_init_shared_memory( void );



/**
 * NAME: add_default_policy
 *
 * DESCRIPTION:   
 * 
 * @Param:  - IN    arg_hEngine: Handle to engine that needs a default policy 
 *                      attached
 * @Param:  - OUT   arg_phRootPolicy: Where function returns handle to default 
 *                      policy
 * @Param:  - OUT   arg_phMsgPolicy: Where function returns handle to message 
 *                      sub-policy
 * @Param:  - OUT   arg_phPktPolicy: Where function returns handle to packet 
 *                      sub-policy
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
static ix_error add_default_policy( 
                                    ix_exe_handle arg_hEngine,
                                    ix_policy_handle* arg_phRootPolicy,
                                    ix_policy_handle* arg_phMsgPolicy,
                                    ix_policy_handle* arg_phPktPolicy
                                  );


/**
 * NAME: create_engine_proc_node
 *
 * DESCRIPTION:   Creates a structure for handling the execution-control interface
 *                and registers the new interface in the corresponding token-
 *                control interface.
 * 
 * @Param:  - OUT   arg_ppEngProcNode: Where function will return a pointer to 
 *                      the engine's execution-control interface
 * @Param:  - OUT   arg_phEngine: Where the function will return the engine's 
 *                      handle
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
static ix_error create_engine_proc_node(
                                         ix_engine_proc_node** arg_ppEngProcNode,
                                         ix_exe_handle* arg_phEngine
                                       );



/**
 * NAME: engine_proc_node_init
 *
 * DESCRIPTION:   Initialize engine construct 
 * 
 * @Param:  - IN    arg_TkHandle. Specifies the location of the corresponding
 *                      child-notify structure in shared memory.
 * @Param:  - INOUT arg_pNode. Pointer to target engine-proc node structure.
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
static ix_error engine_proc_node_init(
                           ix_tk_handle               arg_TkHandle,
                           ix_engine_proc_node*       arg_pNode
                         );


/**
 * NAME: run_loop
 *
 * DESCRIPTION:   Implements execution engine's run loop algorithm.
 * 
 * @Param:  - IN    arg_pEngProcIf. Pointer to engine's execution-control 
 *                      construct.
 * @Param:  - IN   arg_pShutdownWord. Pointer to word within shutdown flags
 *                      associated with execution engine. 
 * @Param:  - IN   arg_ShutdownBitMask. Bit mask on shutdown word corresponding 
 *                      to this engine
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
static ix_error run_loop( 
                          ix_engine_proc_node* arg_pEngProcIf, 
                          ix_uint8* arg_pShutdownWord
                        );


/**
 * NAME: eng_h_to_ptr
 *
 * DESCRIPTION:   
 * 
 * @Param:  - IN    arg_hEngine: Handle to engine
 *
 * @Return: pointer to engine. This function assumes arg_hEngine is a valid
 *          handle to an execution engine.
 */
static ix_engine_proc_node* eng_h_to_ptr( ix_exe_handle arg_hEngine );



/**
 * NAME: process_fatal_errors
 *
 * DESCRIPTION: Calls the fatal-error handler registered using 
 *              ix_cci_register_fatal_error_handler().
 * 
 * @Param:  - IN    arg_Error: Fatal error to be sent to registered handler
 * @Param:  - IN    arg_pEngine: Engine where error was detected
 *
 * @Return: 1 if fatal-error handler has been called; 0 otherwise
 */
static int process_fatal_errors(
                                 ix_error arg_Error,
                                 ix_engine_proc_node* arg_pEngine
                               );



/**
 * NAME: terminate_engine
 *
 * DESCRIPTION:   Returns engine resources to the O/S
 * 
 * @Param: IN       arg_hEngine. Engine handle
 * @Parma: INOUT    arg_pEngine. Pointer to engine's execution-control structure
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
static void terminate_engine(
                             ix_exe_handle arg_hEngine, 
                             ix_engine_proc_node* arg_pEngine
                            );




/*******************************************************************************
 * Definitions of functions internal to the CCI
 */
ix_error ix_cci_run_engine( void* arg_pEngThreadArgs )
{
    ix_error err;
    ix_error cleanup_err;
    ix_uint32 thisEngIdx = 0;
    ix_uint8* pShutdownFlag = NULL;
    ix_engine_proc_node* pEngineProcIf = NULL;
    ix_exe_handle hEngine;
    ix_policy_handle hRootPolicy;
    ix_policy_handle hMsgPolicy;
    ix_policy_handle hPktPolicy;
    int vIsUserInitComplete = 0;

    ix_cci_eng_thread_args* pArgs = (ix_cci_eng_thread_args*)arg_pEngThreadArgs;
    err = create_engine_proc_node( &pEngineProcIf, &hEngine );

    /* Above stores the default policy as child 0. Change this to child 1 
     * This allows the first eci_leaf to be connected directly to the engine
     * bypassing the policy tree until another leaf is added to the engine.
     * Set flag to indicate that no policy or eci_leaf is currently selected.
     * If the user adds a communication ID to a CC in the user-defined 
     * initialization function (invoked below), this flag will be reset
     * to 0 to indicate that there are handlers to process.
     *
     * Since engine resource allocation is complete, except for what may be
     * allocated by the user-defined initialization, allow the main thread to
     * continue.
     */
    thisEngIdx = IX_HAND_TO_ENGINE_IDX( hEngine );

    pShutdownFlag = g_pShutdownFlags + thisEngIdx;

    
    /**
     * Output from create_engine_proc_node() isn't tested until here because
     * the shutdown flag has to be set to release the main thread.
     */
    IX_CCI_IXERR_GOTO( ret_err0, err )

    /*
     * The semaphore was taken by the main thread running function 
     * ix_cci_exe_run() in engine.c. Although policy count will be updated
     * by add_default_policy() the semaphore will be taken and given again
     * within the public API functions called by add_default_policy().
     */
    ix_ossl_sem_give( *g_pShmemSema );


    /**
     * Set up the default policy tree in the engine
     */
    err = add_default_policy( hEngine, &hRootPolicy, &hMsgPolicy, &hPktPolicy );
    IX_CCI_IXERR_GOTO( ret_err1, err )


    pEngineProcIf->tNode.tChildren.pEntries[1].pEci =
        pEngineProcIf->tNode.tChildren.pEntries[0].pEci;
    /* Set custom policy pointer to NULL  */
    pEngineProcIf->tNode.tChildren.pEntries[2].pEci = NULL;

    /* Indicate no policy or eci_leaf is currently selected */
    pEngineProcIf->tNode.vSelectedChild = IX_CCI_NO_SELECTED_CHILD;
    pEngineProcIf->tNode.tChildren.pEntries[0].vWeight = 0;
    pEngineProcIf->tNode.tChildren.pEntries[0].pEci = NULL;


    /* Get user context by running user-provided init function */
    err = pArgs->pInitFunc( hEngine, &pEngineProcIf->pUserContext );
    IX_CCI_IXERR_GOTO( ret_err2, err )
    /* 
     * Indicate user init function has completed, so that handlers know that
     * if the context is NULL, it is because the user has set it that way.
     */
    pEngineProcIf->vIsUserInitComplete = vIsUserInitComplete = 1;

    /*  Check that the user has added at least one handler */
    IX_CCI_ERR_GOTO( 
        pEngineProcIf->tNode.vSelectedChild == IX_CCI_NO_SELECTED_CHILD,
        IX_CCI_ERR_ENG_NO_HANDLERS, ret_err2, err )

    /* Execute engine process function until shutdown flag is detected */
    err = run_loop( pEngineProcIf, pShutdownFlag );

    /* Call user-defined execution engine fini function */
    cleanup_err = pArgs->pFiniFunc( hEngine, pEngineProcIf->pUserContext );
    IX_CCI_ERR_FWD_OR_DUMP( cleanup_err, err )

    /* Clean up execution control tree (events and CCs) */
    terminate_engine( hEngine, pEngineProcIf );


    /* Clean up private engine context and exit */
#if !defined(_CCT_EXCLUDE_)
ret_err2:
#endif
    cleanup_err = ix_cci_policy_destroy( hPktPolicy );
    IX_CCI_ERR_FWD_OR_DUMP( cleanup_err, err )

    cleanup_err = ix_cci_policy_destroy( hMsgPolicy );
    IX_CCI_ERR_FWD_OR_DUMP( cleanup_err, err )

    cleanup_err = ix_cci_policy_destroy( hRootPolicy );
    IX_CCI_ERR_FWD_OR_DUMP( cleanup_err, err )

#if !defined(_CCT_EXCLUDE_)
ret_err1:
#endif
    cleanup_err = ix_cci_eci_node_fini( (ix_cci_exe_ctrl_if*)pEngineProcIf );
    IX_CCI_ERR_FWD_OR_DUMP( cleanup_err, err )

    
    /* Set global thread-exit status */
    ix_ossl_sem_take( *g_pShmemSema, IX_OSSL_WAIT_FOREVER );

    /* Give sema back in ix_cci_exe_shutdown() */
#if !defined(_CCT_EXCLUDE_)
ret_err0:
#endif
    /* 
     * Clear shutdown flag to indicate that engine cleanup is complete 
     */
    if( pShutdownFlag )
    {
        int vIsUserInitShutdown = *pShutdownFlag==IX_ENG_SHUTDOWN_START ? 1: 0;
        if( err != IX_SUCCESS && *g_pEngShutdownStatus == IX_SUCCESS )
            *g_pEngShutdownStatus = err;
        *pShutdownFlag = IX_ENG_SHUTDOWN_COMPLETE;
        if( !vIsUserInitShutdown )
            ix_ossl_sem_give( *g_pShmemSema );
    }
    if( !vIsUserInitComplete )
    {
        /* Assume if an engine cannot start system must close down */
        m_IsSharedSpaceInit = 0;
    }
#if defined(_TEST_RT_)
    /* Rational Test RT gets confused if subthreads terminate!  */
    for( ; ; )
        ix_ossl_sleep( 0 );
#endif
    return err; 
}


/*******************************************************************************
 * Public API Functions
 */

ix_error ix_cci_exe_add_policy(
                                ix_exe_handle arg_hParent,
                                ix_policy_handle arg_hChild,
                                ix_uint32 arg_NumCredits
                              )
{
    ix_uint32 idxEngine;
    ix_uint32 idxPolicy;
    ix_tk_input_notify* pEngNotifyIf;
    ix_tk_input_notify* pPolNotifyIf;
    ix_engine_proc_node* pEngCommonTkIf;
    ix_cci_exe_ctrl_if* pPolCommonTkIf;

    /* Get the shared-memory pointers corresponding to the notify interfaces
     * of the engine and policy 
     */
    IX_CCI_PARAM_ASSERT( 
        IX_CCI_GET_HANDLE_TYPE(arg_hParent) == IX_CCI_HANDLE_TYPES_EXE );
    idxEngine = IX_HAND_TO_ENGINE_IDX( arg_hParent );
    IX_CCI_PARAM_ASSERT( idxEngine != IX_CCI_INVALID_INDEX );
    idxPolicy = IX_HAND_TO_POLICY_IDX( arg_hChild );
    IX_CCI_PARAM_ASSERT( idxPolicy != IX_CCI_INVALID_INDEX );
    pEngNotifyIf = g_pEngine + idxEngine;
    pPolNotifyIf = g_pPolicy + idxPolicy;
    
    /* Add engine's token-notify pointer to policy's token-input interface  */
    pPolNotifyIf->pParent = pEngNotifyIf;

    /* 
     * Add policy as child 0 and child 2. Then when a message or packet handler
     * is added to the engine (via a CC), we know that the current active
     * child is neither a communication ID nor the default policy, and 
     * therefore cannot be added as a leaf to the default policy
     */
    pEngCommonTkIf = (ix_engine_proc_node*)pEngNotifyIf->pProcIfContext;
    pPolCommonTkIf = (ix_cci_exe_ctrl_if*)pPolNotifyIf->pProcIfContext;
    pEngCommonTkIf->tNode.tChildren.pEntries[0].pEci = pPolCommonTkIf;
    pEngCommonTkIf->tNode.tChildren.pEntries[2].pEci = pPolCommonTkIf;

    /* Make policy the selected child */
    pEngCommonTkIf->tNode.vSelectedChild = 0;

    /* Update the granularity of the exeuction engine */
    pEngCommonTkIf->vGranularity = arg_NumCredits;
    return IX_SUCCESS;
}

/* Undocumented API function */
ix_error ix_cci_exe_get_info( ix_exe_handle* arg_pExeHandle,
                              ix_uint32* arg_pEngineNumber,
                              void** arg_pEngineContext )
{
    int i;
    int vEngineCount;
    ix_ossl_thread_t tid;
    ix_error err = ix_ossl_thread_get_id( &tid );
    ix_engine_proc_node* pEngEci = NULL;
    ix_exe_handle hEngine = IX_CCI_HANDLE_TYPES_INVALID;
    IX_CCI_ERR_RET( err, IX_CCI_ERR_ENG_THREAD_ID )
    /*
     * In case each thread has its own memory space, check that global
     * shared-memory pointers are set up and, if not, initialize them.
     */
    if( !m_IsSharedSpaceInit )
    {
        ix_cci_init_shared_memory();
    }

    /* Search through all the execution engines for corresponding thread ID */
    vEngineCount = *g_pEngineCount;
    for( i=0; i<vEngineCount; i++ )
    {
        hEngine = IX_ENGINE_IDX_TO_HANDLE(i);
        pEngEci = eng_h_to_ptr( hEngine );
        if( pEngEci == NULL )
            continue;
        if( pEngEci->vThreadId == tid )
            break;
    }
    if( i==vEngineCount )
    {
        i = *g_pDefaultEngineIdx;
        hEngine = IX_ENGINE_IDX_TO_HANDLE(i);
        pEngEci = eng_h_to_ptr( hEngine );
        IX_CCI_ERR_RET( pEngEci == NULL, IX_CCI_ERR_ENG_THREAD_ID );
    }
    if( arg_pExeHandle ) 
        *arg_pExeHandle = hEngine;
    if( arg_pEngineNumber )
        *arg_pEngineNumber = i;
    if( arg_pEngineContext )
        *arg_pEngineContext = pEngEci->pUserContext;
    return IX_SUCCESS;
}


ix_error ix_cci_exe_set_default( ix_exe_handle arg_hEngine )
{
    if( !m_IsSharedSpaceInit )
    {
        ix_cci_init_shared_memory();
    }
    *g_pDefaultEngineIdx = IX_HAND_TO_ENGINE_IDX( arg_hEngine );
    IX_CCI_PARAM_ASSERT( *g_pDefaultEngineIdx != IX_CCI_INVALID_INDEX );
    return IX_SUCCESS;
}


/*******************************************************************************
 * Definitions of internal functions with local file scope only
 */
static ix_error add_default_policy(
                                    ix_exe_handle arg_hEngine,
                                    ix_policy_handle* arg_phRootPolicy,
                                    ix_policy_handle* arg_phMsgPolicy,
                                    ix_policy_handle* arg_phPktPolicy
                                  )
{

    /*  Create a strict-priority policy*/
    ix_error  err = ix_cci_policy_create( IX_POLICY_TYPE_SP, arg_phRootPolicy );
    IX_CCI_IXERR_RET( err )

    err = ix_cci_policy_create( IX_POLICY_TYPE_WRR, arg_phMsgPolicy );
    IX_CCI_IXERR_GOTO( ret_err1, err )

    err = ix_cci_policy_create( IX_POLICY_TYPE_WRR, arg_phPktPolicy );
    IX_CCI_IXERR_GOTO( ret_err2, err )

    /* Attach the message policy and packet policy to the root policy */
    err = ix_cci_policy_add_branch( *arg_phRootPolicy, *arg_phMsgPolicy, IX_CCI_MESSAGE_POLICY_RUN_GRANULARITY );
    IX_CCI_IXERR_GOTO( ret_err3, err )

    err = ix_cci_policy_add_branch( *arg_phRootPolicy, *arg_phPktPolicy, IX_CCI_PACKET_POLICY_RUN_GRANULARITY );
    IX_CCI_IXERR_GOTO( ret_err3, err )

    /* Attach root policy to execution engine */
    err = ix_cci_exe_add_policy( arg_hEngine, *arg_phRootPolicy, 
                                 IX_CCI_EXE_RUN_GRANULARITY );
    IX_CCI_IXERR_GOTO( ret_err3, err )

    return IX_SUCCESS; /* Normal return */

    /* Error return path */
#if !defined(_CCT_EXCLUDE_)
    {
        ix_error cleanup_err;
ret_err3:
        cleanup_err = ix_cci_policy_destroy( *arg_phPktPolicy );
        IX_CCI_ERR_FWD_OR_DUMP( cleanup_err, err )
ret_err2:
        cleanup_err = ix_cci_policy_destroy( *arg_phMsgPolicy );
        IX_CCI_ERR_FWD_OR_DUMP( cleanup_err, err )

ret_err1:
        cleanup_err = ix_cci_policy_destroy( *arg_phRootPolicy );
        IX_CCI_ERR_FWD_OR_DUMP( cleanup_err, err )

        return err;
    }
#endif
}


static void terminate_engine(
                              ix_exe_handle arg_hEngine, 
                              ix_engine_proc_node* arg_pEngine
                            )
{
    ix_cc_node*    pCCNode;
    ix_cc_node*    pDelCCNode;
    ix_event_node* pEventNode;
    ix_cc_handle   hCC;
    ix_event_node* pDelEventNode = NULL;
    ix_uint32      idx = 0;

    /* Delete the CC list, if it exists */
    pCCNode = arg_pEngine->pCCList;
    while( pCCNode )
    {
        pDelCCNode = pCCNode;
        if( pCCNode->pFiniFunc )
        {
            hCC = IX_CC_IDX_TO_HANDLE( idx, arg_hEngine );
            pCCNode->pFiniFunc( hCC, pCCNode->pUserContext );
        }
        pCCNode = pCCNode->pNext;
        IX_OSSL_FREE( pDelCCNode );
        idx++;
    }
    arg_pEngine->pCCList = NULL;

    /* Delete any outstanding events */
    pEventNode = arg_pEngine->pEventList;
    while( pEventNode )
    {
        pDelEventNode = pEventNode;
        pEventNode = pEventNode->pNext;
        IX_OSSL_FREE( pDelEventNode );
    }
    arg_pEngine->pEventList = NULL;

    /* Delete memory resources associated with expired events */
    pEventNode = arg_pEngine->pExpiredEventList;
    while( pEventNode )
    {
        pDelEventNode = pEventNode;
        pEventNode = pEventNode->pNext;
        IX_OSSL_FREE( pDelEventNode );
    }
    arg_pEngine->pExpiredEventList = NULL;


    /* Free the RM shared-memory integer used to store new-token indicator */
    IX_RM_MEM_FREE( arg_pEngine->pIsNewToken );

    /* 
     * Assume that if we're closing down an engine, we're closing down the whole
     * CCI. For VxWorks, we can't assume the static will be reset to 0 when the
     * CCI starts up again because the code may not be re-downloaded.
     */
    m_IsSharedSpaceInit = 0;
}


static ix_error create_engine_proc_node(
                                         ix_engine_proc_node** arg_ppEngProcNode,
                                         ix_exe_handle* arg_phEngine
                                       )
{
    ix_error err = 0;
    ix_uint32 thisEngIdx;
    ix_tk_input_notify* pEngTkNotifyIf;
    ix_engine_proc_node* pEngineProcIf = NULL;
    ix_tk_handle hEngine;

    /* If not already initialized, init global pointers to shared memory */
    if( !m_IsSharedSpaceInit )
    {
        ix_cci_init_shared_memory();
    }
    
    /**
     * Set the "shutdown" flag to signal the control thread that this thread
     * is now awake.
     * Create the a new engine structure in shared memory 
     * Note that the global semaphore in shared memory will already be locked
     * by the ix_exe_run function 
     */
    
    /* Initialize engine' token_input_notify interface and update 
     * the engine count 
     */
    thisEngIdx = (*g_pEngineCount)++;

    pEngTkNotifyIf = &g_pEngine[thisEngIdx];
    *arg_phEngine = hEngine = IX_ENGINE_IDX_TO_HANDLE( thisEngIdx );
    IX_CCI_ASSERT( *arg_phEngine != IX_CCI_HANDLE_TYPES_INVALID );


    /*
     * Initialize the engine's processing  node.
     */
    pEngineProcIf = (ix_engine_proc_node*)
                                IX_OSSL_MALLOC( sizeof(ix_engine_proc_node) );

    IX_CCI_ERR_RET( !pEngineProcIf, IX_CCI_ERR_OS_MEM_ALLOC )
    ix_ossl_memset( pEngineProcIf, 0, sizeof(ix_engine_proc_node) );


    err = engine_proc_node_init( hEngine, pEngineProcIf );
    IX_CCI_IXERR_GOTO( ret_err, err )

    *arg_ppEngProcNode = pEngineProcIf;

    /* Register the node with the token-notify interface */
    err = ix_cci_tk_notify_register_eci( hEngine, pEngineProcIf );
    IX_CCI_IXERR_GOTO( ret_err, err )

    /* Set the new-token indicator in the token-notify interface */
    pEngTkNotifyIf->pIsNewToken = pEngineProcIf->pIsNewToken;
    return IX_SUCCESS;    /* Normal return */

    /* Error return path */
#if !defined(_CCT_EXCLUDE_)
ret_err:
        IX_OSSL_FREE( pEngineProcIf );
        ix_ossl_sem_give( *g_pShmemSema );
        return err;
#endif
}



void ix_cci_exe_remove_leaf( ix_tk_input_notify* arg_pEngine )
{
    int idxPolicy;
    int i;
    ix_engine_proc_node* pEngEci = (ix_engine_proc_node*)arg_pEngine->pProcIfContext;
    ix_eci_node* pRootPolicy;

    /* Temporarily store pointer to leaf */
    ix_tkp_entry* pChildEntry = pEngEci->tNode.tChildren.pEntries;
    ix_cci_exe_ctrl_if* pLeafEci = pChildEntry[0].pEci;

    /* Change active child to be default policy (child 0<-child 1) */
    pChildEntry[0] = pChildEntry[1];

    /* Remove child from default policy tree */

    pRootPolicy = (ix_eci_node*)pEngEci->tNode.tChildren.pEntries[0].pEci;
    
    
    /**
     * Root policy has two child policies: one for messages and one for packets.
     * Since we don't know where the leaf is, search for it in both policies.
     */
    for( idxPolicy=0; idxPolicy<2; idxPolicy++ )
    {
        /* Find the node pointer in the msg/pkt policy's child container */
        ix_tkp_entry* pTkpEntry;
        int numEntries;
        ix_eci_node* pChildPolicy 
            = (ix_eci_node*)pRootPolicy->tChildren.pEntries[idxPolicy].pEci;
        /* 
         * Shutting down of policies and handles can be done by both the user
         * and the CCI system. Check the node hasn't already closed down
         */
        if( pChildPolicy == NULL )
            continue;

        pTkpEntry = pChildPolicy->tChildren.pEntries;
        numEntries = pChildPolicy->tChildren.vNumEntries;

        /* Search for leaf in selected policy */
        for( i=0; i<numEntries; i++ )
        {
            /* If found clear the associated policy fields */
            if( pTkpEntry[i].pEci == pLeafEci )
            {
                /* Free the leaf's memory resources */
                IX_OSSL_FREE( pLeafEci );

                /* Set the entry to NULL */
                pTkpEntry[i].pEci = NULL;
                pTkpEntry[i].vCredits = 0;
                pTkpEntry[i].vWeight = 0;
                break;
            }
        }
        if( i < numEntries )
            break;
    }
}

static ix_engine_proc_node* eng_h_to_ptr( ix_exe_handle arg_hEngine )
{
    /* Since this is an internal function we know handle is for engine */
    int idxEng = IX_HAND_TO_ENGINE_IDX( arg_hEngine );
    ix_tk_input_notify* pEngNotifyIf = g_pEngine + idxEng;
    return (ix_engine_proc_node*)pEngNotifyIf->pProcIfContext;
}


static void ix_cci_init_shared_memory( void )
{
    ix_configuration_property_handle hSemaProperty;
    ix_configuration_property_handle hNotifyIfProperty;
    ix_configuration_property_handle hShutdownFlagsProperty;

    g_pMsgTkq = NULL; 
    g_pPktTkq = NULL;
    g_pPolicy = NULL;
    g_pEngine = NULL;
    g_pShutdownFlags = NULL;
    g_pEngShutdownStatus = NULL;
    g_pEngineCount = NULL;
    g_pPolicyCount = NULL;
    g_pDefaultEngineIdx = NULL;

    
    /**
     * If the code has got as far as here, the following property functions
     * are bound to work. If they do not, we can't do anything about it 
     * anyway. If access can't be gained to the shutdown flags, the system
     * is going to hang.
     */
    ix_rm_cp_property_open( IX_CP_CORE_PROPERTY_ROOT,
                            g_szCciSharedPtrShutdownFlags,
                            &hShutdownFlagsProperty );

    ix_rm_cp_property_get_value_uint32( hShutdownFlagsProperty,
                                        (ix_uint32*)&g_pShutdownFlags );
    g_pEngShutdownStatus = ((ix_uint32*)g_pShutdownFlags) 
                                + IX_CCI_NUM_SHUTDOWN_WORDS - 1;
    g_pEngineCount = g_pEngShutdownStatus + 1;
    g_pPolicyCount = g_pEngineCount + 1;
    g_pDefaultEngineIdx = g_pPolicyCount + 1;
    g_pFatalErrorHandler = (ix_ferror_func*) (g_pDefaultEngineIdx + 1 ); 
    g_ppFatalErrorContext = (void**)(g_pFatalErrorHandler + 1);
    /* Get the shared memory addresses initialized in ix_cci_init()  */
    ix_rm_cp_property_open( IX_CP_CORE_PROPERTY_ROOT,
                                  g_szCciSharedPtrShmemSema,
                                  &hSemaProperty );

    ix_rm_cp_property_get_value_uint32( hSemaProperty,
                                        (ix_uint32*) &g_pShmemSema );

    ix_rm_cp_property_open( IX_CP_CORE_PROPERTY_ROOT,
                            g_szCciSharedPtrMsgTkq,
                            &hNotifyIfProperty );

    ix_rm_cp_property_get_value_uint32( hNotifyIfProperty,
                                           (ix_uint32*)&g_pMsgTkq );

    g_pPktTkq = g_pMsgTkq + IX_CCI_MAX_INPUT_IDS;
    g_pPolicy = g_pPktTkq + IX_CCI_MAX_INPUT_IDS;
    g_pEngine = g_pPolicy + IX_CCI_MAX_POLICIES;


    m_IsSharedSpaceInit = 1;

    
    ix_rm_cp_property_close( hNotifyIfProperty );

    ix_rm_cp_property_close( hSemaProperty );

}


static ix_error engine_proc_node_init(
                           ix_tk_handle               arg_TkHandle,
                           ix_engine_proc_node*       arg_pNode
                         )
{
    
    /**
     * The processing and fini function ptrs are set to NULL because the funcs
     * are called directly in ix_cci_run_engine(). The execution engine only has
     * one active child.  But the engine needs to know whether a token queue is 
     * connected or the default policy, or the user's custom policy. Child 0 is 
     * the selected policy or token queue, child 1 is the default policy, child 
     * 2 is the user's custom policy (if added).
     */
    ix_error err = ix_cci_eci_node_init( (ix_eci_node*)arg_pNode,
                                         NULL, NULL, arg_TkHandle, 3, 0 );
    IX_CCI_IXERR_RET( err )

    /* Initialize CC list as empty */
    arg_pNode->pCCList = NULL;

    /* Indicate user-defined engine init function has not been run yet */
    arg_pNode->vIsUserInitComplete = 0;
    
    /* Initialize the event list, due list, and expired list as empty */
    arg_pNode->pEventList = NULL;
    arg_pNode->vSizeEventList = 0;
    arg_pNode->pDueEventList = NULL;
    arg_pNode->pExpiredEventList = NULL;
    arg_pNode->pActiveEvent = NULL;

    /* To support ix_cci_exe_get_handle() function, store the thread ID */
    err = ix_ossl_thread_get_id( &arg_pNode->vThreadId );
    IX_CCI_ERR_RET( err != IX_SUCCESS, IX_CCI_ERR_ENG_THREAD_ID );

    /* Initialize the new-token indicator in RM shared-memory space */
    err = IX_RM_MEM_ALLOC( IX_MEMORY_TYPE_DRAM, IX_CCI_MEMORY_CHANNEL, 
                           sizeof(ix_uint32), (void**)&arg_pNode->pIsNewToken );
    IX_CCI_ERR_RET( err != IX_SUCCESS, IX_CCI_ERR_SHARED_MEM_ALLOC );
    return IX_SUCCESS;
}

static ix_error run_loop( 
                          ix_engine_proc_node* arg_pEngProcIf, 
                          ix_uint8* arg_pShutdownFlag
                        )
{
    /* Calculate which shutdown bit of which global shutdown byte to 
     * monitor
     */
    ix_exe_loop_ctrl vLoopCtrl = IX_EXE_LOOP_CTRL_ACTIVE;
    ix_error err = IX_SUCCESS;
    ix_tk_input_notify* pNotifyIf = arg_pEngProcIf->tNode.tEci.pTkInNotify;
    ix_ossl_sem_t* pEngSema = &pNotifyIf->vSema;
    ix_uint32 vSemaTimeout = pNotifyIf->vSemaTimeout;
    volatile ix_uint8* pShutdownFlag = arg_pShutdownFlag;
    *pShutdownFlag = IX_ENG_THREAD_AWAKE;


    /* If there are only event handlers, check for events only */
    if( arg_pEngProcIf->tNode.vSelectedChild == IX_CCI_EVENT_ONLY_PRESENT )
    {
        while( *pShutdownFlag != IX_ENG_SHUTDOWN_START 
               && vLoopCtrl != IX_EXE_LOOP_CTRL_ABORT )
        {
            ix_uint32 eventCount = arg_pEngProcIf->vGranularity;

            /* Wait for event-check interval */
            ix_ossl_sem_take( *pEngSema, vSemaTimeout );

            while( eventCount-- 
                   && vLoopCtrl != IX_EXE_LOOP_CTRL_WAIT_FOR_SHUTDOWN )
            {
                /* Check events */
                err = ix_cci_handle_events( arg_pEngProcIf );
                /*
                 * Any errors that get returned are fatal. If a fatal-error
                 * handler is present call it and wait for shutdown
                 */
                if( err )
                {
                    /*
                     * If a fatal-error handler is present, call it and wait for 
                     * shutdown. Otherwise exit the exe ctrl loop immediately
                     * and let the CCI close down the engine the best it can.
                     */
                    process_fatal_errors( err, arg_pEngProcIf );
                    vLoopCtrl = IX_EXE_LOOP_CTRL_WAIT_FOR_SHUTDOWN;
                    break;
                }

            } /* end while pending  */
            /* Cede control to the next task (execution engine) */
/*            ix_ossl_sleep( 0 );*/
            /* Check the active event list to set the engine-sempahore timeout */
            if( arg_pEngProcIf->pEventList )
            {
                ix_event_node* pNode = arg_pEngProcIf->pEventList->pNext;
                vSemaTimeout = arg_pEngProcIf->pEventList->vPeriodMilliSec;
                while( pNode )
                {
                    if(vSemaTimeout > pNode->vPeriodMilliSec )
                    {
                        vSemaTimeout = pNode->vPeriodMilliSec; 
                    }
                    pNode = pNode->pNext;
                }

            }
            else
                vSemaTimeout = IX_CCI_EVENT_CHECK_INTERVAL;
        }
    }
    /*
     * Otherwise check for events, messages, and packets. Give priority to 
     * events, then messages, then packets.
     */
    else 
    {
        /* Child 0 is the engine's only selectable child */
        ix_cci_exe_ctrl_if* pRootPolicy =
            arg_pEngProcIf->tNode.tChildren.pEntries[0].pEci;
        ix_cci_process_tokens_ptr 
            pRootPolicyProcFunc = pRootPolicy->pProcessFunc;
        while( *pShutdownFlag != IX_ENG_SHUTDOWN_START 
                && vLoopCtrl != IX_EXE_LOOP_CTRL_ABORT )
        {
            ix_uint32 isStillPending = 1;
            ix_uint32 procCount = arg_pEngProcIf->vGranularity;

            /* Wait for event-check interval */
            ix_ossl_sem_take( *pEngSema, vSemaTimeout );

            while( isStillPending  
                   && vLoopCtrl != IX_EXE_LOOP_CTRL_WAIT_FOR_SHUTDOWN )
            {
                /* Check events */
                err = ix_cci_handle_events( arg_pEngProcIf );
            
                /*
                 * Any errors that get returned are fatal. If a fatal-error
                 * handler is present call it and wait for shutdown
                 */
                if( err )
                {
                    /*
                     * If a fatal-error handler is present, call it and wait for 
                     * shutdown. Otherwise exit the exe ctrl loop immediately
                     * and let the CCI close down the engine the best it can.
                     */
                    *g_pEngShutdownStatus = err;
                    process_fatal_errors( err, arg_pEngProcIf );
                    vLoopCtrl = IX_EXE_LOOP_CTRL_WAIT_FOR_SHUTDOWN;
                    break;
                }
                /* Above function only returns fatal errors */
                IX_CCI_IXERR_RET( err )

                /* Check policy tree (call the root policy's process function) */
                isStillPending = 0;
                err = pRootPolicyProcFunc( pRootPolicy, &isStillPending, 
                                           &procCount, arg_pEngProcIf );
                if( err )
                {
                    /*
                     * Treat message and packet handlers in the same way as event
                     * handlers.
                     */
                    process_fatal_errors( err, arg_pEngProcIf );
                    vLoopCtrl = IX_EXE_LOOP_CTRL_WAIT_FOR_SHUTDOWN;
                    break;
                }
                /*
                 * If code has reached here, engine has used up its processing 
                 * budget or there were no more tokens to process. Cede control
                 * to the next task (execution engine).
                 */
/*                ix_ossl_sleep( 0 );*/
                procCount = arg_pEngProcIf->vGranularity;
            } /* end while pending  */
            /* Check the active event list to set the engine-sempahore timeout */
            if( arg_pEngProcIf->pEventList )
                vSemaTimeout = arg_pEngProcIf->pEventList->vPeriodMilliSec;
            else
                vSemaTimeout = IX_CCI_EVENT_CHECK_INTERVAL;

        }   /* shutdown flag for this execution engine detected */
    }   /* end else */

    return err;
}

static int process_fatal_errors(
                                 ix_error arg_Error,
                                 ix_engine_proc_node* arg_pEngine
                               )
{
    /* Return 0 if no fatal-error handler has been registered */
    if( *g_pFatalErrorHandler == NULL )
        return 0;
    (*g_pFatalErrorHandler)( 
                             arg_Error, 
                             arg_pEngine->tNode.hSelf, 
                             *g_ppFatalErrorContext 
                           );
    return 1;
}
