/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation.All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 5
 *
 * = FILENAME
 *      event_node.c
 *
 * = DESCRIPTION
 *      Code for storing and handling events
 *
 * = AUTHOR
 *       Roland Arajs
 *       roland.arajs@intel.com
 *
 * = CHANGE HISTORY
 *      5/1/2002 8:05:28 AM - creation time
 *
 * ============================================================================
 * $Id: event_node.c,v 1.21 2003/12/11 22:25:28 ageorges Exp $
 */

#include "ix_ossl.h"

#include "cci/internal/internal_event.h"
#include "cci/internal/internal_cci_handles.h"
#include "cci/internal/internal_cci_error.h"
#include "cci/internal/internal_engine.h"

/*******************************************************************************
 * Internal Function Declarations
 */

/**
 * NAME: create_event
 *
 * DESCRIPTION:   
 * 
 * @Param:  - IN    arg_pEngine: Where the event will eventually be activated.
 *                      Used to calculate/re-use event handle
 * @Param:  - IN    arg_hCC: Handle to core component. Used to calculate event 
 *                      handle.
 * @Param:  - OUT   arg_phEvent: Event handle
 * @Param:  - OUT   arg_pEventNode: Where function will return the event pointer
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
static ix_error create_event(
                              ix_engine_proc_node* arg_pEngine,
                              ix_cc_handle      arg_hCC,
                              ix_event_handle*  arg_phEvent,
                              ix_event_node**   arg_ppEventNode
                            );



/**
 * NAME: activate_event
 *
 * DESCRIPTION:   
 * 
 * @Param:  - IN arg_pEvent: Pointer to event that is to be activated
 * @Param:  - IN arg_When: The time from now in ms when the event will be 
 *                  triggered
 * @Param:  - IN arg_pEngine: The engine where the event will be triggered
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
static ix_error activate_event(
                                ix_event_node* arg_pEvent,
                                ix_uint32 arg_When,
                                ix_engine_proc_node* arg_pEngine
                              );




/**
 * NAME: process_due_events
 *
 * DESCRIPTION:   
 * 
 * @Param:  - INOUT arg_pEngProcIf. Engine where events will be processed 
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
static ix_error process_due_events(
                                    ix_engine_proc_node* arg_pEngProcIf
                                  );



/**
 * NAME: reschedule_event
 *
 * DESCRIPTION:   Reschedules an event, either because a due event is periodic
 *                or because there is no context ready yet.
 * 
 * @Param:  - IN    arg_pDueList. List of events that are due to be processed.
 *                      This function will reschedule the item at the front
 *                      of the due list, pop the event off this list, and put 
 *                      it back on the active list
 * @Param:  - INOUT arg_ppActiveList. Where event will be rescheduled 
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
static void reschedule_event(
                              ix_event_node* arg_pDueList,
                              ix_event_node** arg_ppActiveList
                            );



/**
 * NAME: get_user_context
 *
 * DESCRIPTION:   Gets the user context associated with an event.
 * 
 * @Param:  - INOUT  arg_pEventNode. Pointer to event node. If context is not
 *                      in the node, the function gets the context from the 
 *                      event owner (CC or execution engine), and stores the
 *                      context in the node.
 *
 * @Return: user context from event or from event owner (CC or execution engine)
 */
static void* get_user_context( ix_event_node* arg_pEventNode, 
                               ix_uint32* arg_pIsContextValid );






/*******************************************************************************
 * Public APIs
 */

ix_error ix_cci_cc_add_event_handler(
                                      ix_cc_handle   arg_hComponent,
                                      ix_uint32 arg_ms,
                                      ix_event_func arg_EventFunc,
                                      ix_event_type arg_EventType,
                                      ix_uint32 arg_Priority,
                                      ix_event_handle* arg_phEvent
                                    )
{
    /**
     * Although this API associates an event with a core component, the core 
     * component infrastructure actually associates the event with an 
     * execution engine. The core component's user context will be passed to
     * the event handler.
     */
    /* Convert handle to engine and core component indexes */
    ix_uint32 idxCC;
    ix_uint32 idxEngine;
    ix_cc_node* pCCEntry;
    ix_tk_input_notify* pEngNotifyIf;
    ix_engine_proc_node* pEngine;
    ix_event_node* pEventNode;
    ix_uint32 i = 0;
    void* pUserContext;
    ix_cci_handle_types vHandleType;
    ix_error err;

    idxEngine = IX_HAND_TO_ENGINE_IDX( arg_hComponent );
    IX_CCI_PARAM_ASSERT( idxEngine != IX_CCI_INVALID_INDEX )
    IX_CCI_PARAM_ASSERT( arg_EventFunc != NULL )
    IX_CCI_PARAM_ASSERT( arg_EventType < IX_EVENT_TYPE_LAST )
    IX_CCI_PARAM_ASSERT( arg_phEvent != NULL )
    IX_CCI_PARAM_ASSERT( !(arg_EventType == IX_EVENT_TYPE_PERIODIC 
                           && arg_ms == 0) )
    
    /* Get pointer to engine's execution-control I/F via token-notify I/F */
    pEngNotifyIf = g_pEngine + idxEngine;
    pEngine = (ix_engine_proc_node*)pEngNotifyIf->pProcIfContext;

    /* Create a new event or re-use one that is on the engine's expired list */
    err = create_event( pEngine, arg_hComponent, arg_phEvent, &pEventNode );
    IX_CCI_IXERR_RET( err )

    
    /**
     * arg-hComponent could actually be a handle to an execution engine. Attempt
     * to get the context from the engine or component. If no context is defined
     * yet, store the CC/EXE handle so that the context can be retrieved later.
     */
    vHandleType = IX_CCI_GET_HANDLE_TYPE( arg_hComponent );
    switch( vHandleType )
    {
    case IX_CCI_HANDLE_TYPES_EXE:
        pUserContext = pEngine->pUserContext;
        /*
         *  If user context is not valid, because user is initializing event 
         *  before updating engine's context pointer in the user-defined 
         *  engine-initialization function, we must store the handle of the 
         *  engine so that the context can be searched for later when the event 
         *  is actually received.
         */
        if( pEngine->vIsUserInitComplete )
        {
            pEventNode->pEventContext = pUserContext;
            pEventNode->hUserCtxOwner = IX_CCI_HANDLE_TYPES_INVALID;
        }
        else
        {
            pEventNode->pEventContext = NULL;
            pEventNode->hUserCtxOwner = arg_hComponent;
        }
        break;

    case IX_CCI_HANDLE_TYPES_CC:
        {
            idxCC = IX_HAND_TO_CC_IDX( arg_hComponent );
            IX_CCI_ASSERT( idxCC != IX_CCI_INVALID_INDEX )

            /* Get execution ctrl I/F of CC to get user context */
            pCCEntry = pEngine->pCCList;
            while( i++ < idxCC )
            {
                pCCEntry = pCCEntry->pNext;
            }
             /* Set event context to CC user context */
            pUserContext = pCCEntry->pUserContext;
            /*
             *  If user context is not valid, because user is initializing event 
             *  before updating CC's context pointer in the user-defined 
             *  CC-initialization function, we must store the handle of the CC 
             *  so that the context can be searched for later when the event is 
             *  actually received.
             */
            if( pCCEntry->vIsUserInitComplete )
            {
                pEventNode->pEventContext = pUserContext;
                pEventNode->hUserCtxOwner = IX_CCI_HANDLE_TYPES_INVALID;
            }
            else
            {
                pEventNode->pEventContext = NULL;
                pEventNode->hUserCtxOwner = arg_hComponent;
            }
        }
        break;
    default:
        return IX_CCI_ERR_NEW( IX_CCI_ERR_CC_ADD_HANDLER_INVALID_HANDLE );
    }


    /* Fill out the other event fields from the function arguments */
    pEventNode->pEventFunc = arg_EventFunc;
    pEventNode->vPriority = arg_Priority;
    pEventNode->vEventType = arg_EventType;
    /* 
     * In execution engine, clear the MS bit of selected-child member to
     * indicate a handler is present.
     */
    pEngine->tNode.vSelectedChild &= IX_CCI_EVENT_ONLY_PRESENT;
    /* 
     * Convert the user-provided time into ix_ossl_time_t format and add it
     * to the current system time. Store the result in the event node's 
     * expiry-time field.
     */
    err = activate_event( pEventNode, arg_ms, pEngine );
    IX_CCI_IXERR_RET( err )
    return IX_SUCCESS;
}


ix_error ix_cci_cc_add_event_handler_ex(
                                         ix_cc_handle   arg_hComponent,
                                         ix_uint32 arg_ms,
                                         ix_event_func arg_EventFunc,
                                         ix_event_type arg_EventType,
                                         ix_uint32 arg_Priority,
                                         void* arg_EventContext,
                                         ix_event_handle* arg_phEvent
                                       )
{
    /**
     * Although this API associates an event with a core component, the core 
     * component infrastructure actually associates the event with an 
     * execution engine.
     */
    /* Convert handle to engine and core component indexes */
    ix_uint32 idxEngine;
    ix_tk_input_notify* pEngNotifyIf;
    ix_engine_proc_node* pEngine;
    ix_event_node* pEventNode;
    ix_error err;

    idxEngine = IX_HAND_TO_ENGINE_IDX( arg_hComponent );
    IX_CCI_PARAM_ASSERT( idxEngine != IX_CCI_INVALID_INDEX )
    IX_CCI_PARAM_ASSERT( arg_EventFunc != NULL )
    IX_CCI_PARAM_ASSERT( arg_EventType < IX_EVENT_TYPE_LAST )
    IX_CCI_PARAM_ASSERT( arg_phEvent != NULL )
    IX_CCI_PARAM_ASSERT( !(arg_EventType == IX_EVENT_TYPE_PERIODIC 
                           && arg_ms == 0) )
    
    /* Get pointer to engine's execution-control I/F via token-notify I/F */
    pEngNotifyIf = g_pEngine + idxEngine;
    pEngine = (ix_engine_proc_node*)pEngNotifyIf->pProcIfContext;

    /* Create a new event or re-use one that is on the engine's expired list */
    err = create_event( pEngine, arg_hComponent, arg_phEvent, &pEventNode );
    IX_CCI_IXERR_RET( err )


    /* Set the user context from function argument rather than CC or engine */
    pEventNode->hUserCtxOwner = IX_CCI_HANDLE_TYPES_INVALID;
    pEventNode->pEventContext = arg_EventContext;


    /* Fill out the other event fields from the function arguments */
    pEventNode->pEventFunc = arg_EventFunc;
    pEventNode->vPriority = arg_Priority;
    pEventNode->vEventType = arg_EventType;
    /* 
     * In execution engine, clear the MS bit of selected-child member to
     * indicate a handler is present.
     */
    pEngine->tNode.vSelectedChild &= IX_CCI_EVENT_ONLY_PRESENT;
    /* 
     * Convert the user-provided time into ix_ossl_time_t format and add it
     * to the current system time. Store the result in the event node's 
     * expiry-time field.
     */
    err = activate_event( pEventNode, arg_ms, pEngine );
    IX_CCI_IXERR_RET( err )
    return IX_SUCCESS;
}

ix_error ix_cci_cc_remove_event_handler(
                                         ix_cc_handle arg_hComponent,                 
                                         ix_event_handle arg_hEvent
                                       )
{
    /* Get execution engine index from CC handle */
    ix_engine_proc_node* pEngine;
    ix_event_node* pEvent;
    ix_event_node* pPrevEvent = NULL;
    ix_uint32 idxEngine = IX_HAND_TO_ENGINE_IDX( arg_hComponent ); 
    IX_CCI_PARAM_ASSERT( idxEngine != IX_CCI_INVALID_INDEX )
    IX_CCI_PARAM_ASSERT( 
        IX_CCI_GET_HANDLE_TYPE(arg_hEvent) == IX_CCI_HANDLE_TYPES_EVENT)

    /* Get pointer to engine's execution control i/f from toke-notify i/f */
    pEngine = (ix_engine_proc_node*)g_pEngine[idxEngine].pProcIfContext;

    
    /**
     * First check if we want to delete the active event
     */
    pEvent = pEngine->pActiveEvent;
    if( pEvent )
    {
        /* Check if this is the event we are looking for */
        if( pEvent->hEvent == arg_hEvent )
        {
            /* Set the active event to NULL to signal that we have deleted it */
            pEngine->pActiveEvent = NULL;

            /* Add this event to the expired list */
            pEvent->pNext = pEngine->pExpiredEventList;
            pEngine->pExpiredEventList = pEvent;
            return IX_SUCCESS;
        }
    }


    /*
     * Search through the list of active events for event with handle 
     * corresponding to function argument
     */
    pEvent = pEngine->pEventList;
    if( pEvent )
    {
        /* If event is at beginning of list */
        if( pEvent->hEvent == arg_hEvent )
        {
            /* Set beginning of list to next event */
            pEngine->pEventList = pEvent->pNext;

            /* Add this event to the expired list */
            pEvent->pNext = pEngine->pExpiredEventList;
            pEngine->pExpiredEventList = pEvent;
            return IX_SUCCESS;
        }
        /* Otherwise, search for the event in the active list and unlink */
        for( ; ; )
        {
            pPrevEvent = pEvent;
            pEvent = pEvent->pNext;
            if( !pEvent )
                break;
            if( pEvent->hEvent == arg_hEvent )
            {
                /* Move event from active event list to expired event list */
                pPrevEvent->pNext = pEvent->pNext;
                pEvent->pNext = pEngine->pExpiredEventList;
                pEngine->pExpiredEventList = pEvent;
                return IX_SUCCESS;
            }
        }
    }
    /*
     * If event was not in active list, search due list
     */
    pEvent = pEngine->pDueEventList;
    
    if( pEvent )
    {
        /* If event is at beginning of list */
        if( pEvent->hEvent == arg_hEvent )
        {
            /* Set beginning of list to next event */
            pEngine->pDueEventList = pEvent->pNext;

            /* Add this event to the expired list */
            pEvent->pNext = pEngine->pExpiredEventList;
            pEngine->pExpiredEventList = pEvent;
            return IX_SUCCESS;
        }
        /* Otherwise, search for the event in the due list and unlink */
        for( ; ; )
        {
            pPrevEvent = pEvent;
            pEvent = pEvent->pNext;
            if( !pEvent )
                break;
            if( pEvent->hEvent == arg_hEvent )
            {
                /* Move event from active event list to expired event list */
                pPrevEvent->pNext = pEvent->pNext;
                pEvent->pNext = pEngine->pExpiredEventList;
                pEngine->pExpiredEventList = pEvent;
                return IX_SUCCESS;
            }
        }
    }
#if defined(_IX_CCI_DEBUG_)
    /* 
     * Event not in active list. If in debug mode, check that event is on 
     * expired list 
     */
    pEvent = pEngine->pExpiredEventList;
    while( pEvent )
    {
        if( pEvent->hEvent == arg_hEvent )
            break;
        pEvent = pEvent->pNext;
    }
    /* If not found there, return an error */
    IX_CCI_ERR_RET( !pEvent, IX_CCI_ERR_EVENT_NOT_FOUND )
#endif
    return IX_SUCCESS;
}



/*******************************************************************************
 * Functions called by other CCI modules
 */

ix_error ix_cci_handle_events( ix_engine_proc_node* arg_pEngProcNode )
{
    ix_ossl_time_t sysTime = {0, 0};
    ix_event_node* pActiveList;
    /* Get the current system time */
    /* If there are any expired events in the active event list*/
    while( arg_pEngProcNode->pEventList )
    {
        /**
         * Don't try optimizing this loop by taking pActiveList outside. If one
         * of the processed event handlers deletes an active event, the pointer
         * will go out-of-date. For the same reason, the due list is kept in
         * the engine rather than as an automatic variable in this function
         */
        ix_event_node* pDueList = NULL;
        ix_error err = ix_ossl_time_get( &sysTime );
        IX_CCI_ERR_RET( err, IX_CCI_ERR_EVENT_HANDLER_GET_TIME )
        pActiveList = arg_pEngProcNode->pEventList;
        IX_CCI_ASSERT( arg_pEngProcNode->pDueEventList == NULL );

        /* Create a temporary event list and add the expired events in 
         * priority order
         */
        while( pActiveList && !IX_OSSL_TIME_LT(sysTime, pActiveList->vExpiryTime) )
        {
            ix_event_node* pWkng = arg_pEngProcNode->pDueEventList;

            if( !pWkng || pActiveList->vPriority > pWkng->vPriority )
            {
                /**
                 * If due list is empty or active-list item priority is greater
                 * than any item currently on the due list, then add item to 
                 * start of due list
                 */
                arg_pEngProcNode->pDueEventList = pDueList = pActiveList;
                arg_pEngProcNode->pEventList = pActiveList = pActiveList->pNext;
                pDueList->pNext = pWkng;
            }
            else
            {
                /**
                 * Otherwise, add item in due list to maintain priority order
                 */
                ix_event_node* pNext = pWkng->pNext;
                while( pNext && pActiveList->vPriority < pNext->vPriority )
                {
                    pWkng = pNext;
                    pNext = pWkng->pNext;
                }

                pWkng->pNext = pActiveList;
                pActiveList = arg_pEngProcNode->pEventList = pActiveList->pNext;
                pWkng->pNext->pNext = pNext;

            }
        }
        /* If the temporary list is empty exit from event handler */
        if( !pDueList )
            break;
        /* Execute events in priority order */
        err = process_due_events( arg_pEngProcNode );
        IX_CCI_IXERR_RET( err )
    }
    return IX_SUCCESS;
}

/*
NOTE: Maybe uncomment this function when eci leafs check whether events are
      pending.
int ix_cci_is_event_pending( const ix_event_node* const* arg_ppEventList )
{
    ix_ossl_time_t sysTime;
*/
    /* Get the system time and compare it with first entry in event list */
/*
    ix_ossl_time_get( &sysTime );
*/
    /* Return true if event has expired; false otherwise */
/*
    return IX_OSSL_TIME_GT( (**arg_ppEventList).vExpiryTime, sysTime ) ? 0 : 1;
}
*/

/*******************************************************************************
 * Implementation of internal module functions
 */
static ix_error process_due_events(
                                    ix_engine_proc_node* arg_pEngProcIf
                                  )
{
    ix_error err;
    ix_uint32 vIsContextValid;
    ix_event_node* pCurrentEvent = arg_pEngProcIf->pDueEventList;
    void* pUserContext;
    while( pCurrentEvent )
    {
        /* Get the user context */
        pUserContext = get_user_context( pCurrentEvent, &vIsContextValid );
        /* remove the current event from the due list */
        arg_pEngProcIf->pDueEventList = pCurrentEvent->pNext;

        /* User context must be valid by now */
        IX_CCI_ASSERT( vIsContextValid );

        /* store the active event */
        arg_pEngProcIf->pActiveEvent = pCurrentEvent;

        /* invoke the event handler */
        err = pCurrentEvent->pEventFunc( pUserContext );

        /* retrieve the active event */
        pCurrentEvent = arg_pEngProcIf->pActiveEvent;
        arg_pEngProcIf->pActiveEvent = NULL;

#if defined(_IX_CCI_DEBUG_)
        /* 
         * If there is an error, check the severity. If fatal return. 
         * Otherwise, log the error. The normal processing code
         * will then drop one-shot events and reschedule periodic ones.
         */
        IX_CCI_IXERR_LOG_RET_IF_FATAL( err, IX_CCI_ERR_EVENT_HANDLER );
#endif /* defined(_IX_CCI_DEBUG_) */

        /* In the case the event has been removed then do nothing */
        if(pCurrentEvent != NULL)
        {
            /* Put one-shot events on the front of the expired list */
            if( pCurrentEvent->vEventType == IX_EVENT_TYPE_ONESHOT )
            {
                pCurrentEvent->pNext = arg_pEngProcIf->pExpiredEventList;
                arg_pEngProcIf->pExpiredEventList = pCurrentEvent;
            }
            else 
            {
                /**
                 * Reschedule periodic events by putting them back on the active
                 * event list
                 */
                reschedule_event( pCurrentEvent, &arg_pEngProcIf->pEventList );
            }
        }
        
        pCurrentEvent = arg_pEngProcIf->pDueEventList;
    }

    return IX_SUCCESS;
}

static void* get_user_context( ix_event_node* arg_pEventNode,
                               ix_uint32* arg_pIsContextValid )
{
    ix_handle hEventOwner = arg_pEventNode->hUserCtxOwner;
    int idxEng;
    int idxCC;
    void* pContext;
    ix_uint32 vOwnerType;
    ix_engine_proc_node* pEngProcIf;
    ix_cc_node* pCCProcIf;
    vOwnerType = IX_CCI_GET_HANDLE_TYPE( hEventOwner );
    switch( vOwnerType )
    {
    case IX_CCI_HANDLE_TYPES_INVALID:
        pContext = arg_pEventNode->pEventContext;
        *arg_pIsContextValid = 1;
        break;
    case IX_CCI_HANDLE_TYPES_CC:
        /* Get the engine's execution control i/f via its notify i/f */
        idxEng = IX_HAND_TO_ENGINE_IDX( hEventOwner ); 
        pEngProcIf = (ix_engine_proc_node*)g_pEngine[idxEng].pProcIfContext;
        /* Get the CC's execution control i/f from the engine's ctrl i/f */
        idxCC = IX_HAND_TO_CC_IDX( hEventOwner );
        pCCProcIf = pEngProcIf->pCCList;
        while( idxCC-- )
        {
            pCCProcIf = pCCProcIf->pNext;
        }
        /* Get the user context from the CC */
        pContext = pCCProcIf->pUserContext;
        /* If the context is set up  */
        if( pCCProcIf->vIsUserInitComplete )
        {
            /* Change owner handle to reflect this */
            arg_pEventNode->hUserCtxOwner = IX_CCI_HANDLE_TYPES_INVALID;

            /* Set event-node context to CC context */
            arg_pEventNode->pEventContext = pContext;
            *arg_pIsContextValid = 1;
        }
        else
        {
            *arg_pIsContextValid = 0;
        }
        break;
    default:    /* Engine handle */
        /* Get the engine's execution control i/f via its notify i/f */
        idxEng = IX_HAND_TO_ENGINE_IDX( hEventOwner ); 
        pEngProcIf = (ix_engine_proc_node*)g_pEngine[idxEng].pProcIfContext;
        pContext = pEngProcIf->pUserContext;
        /* If the context is set up change owner handle to reflect this */
        if( pEngProcIf->vIsUserInitComplete )
        {
            arg_pEventNode->hUserCtxOwner = IX_CCI_HANDLE_TYPES_INVALID;
            arg_pEventNode->pEventContext = pContext;
            *arg_pIsContextValid = 1;
        }
        else
        {
            *arg_pIsContextValid = 0;
        }
    }
    return pContext;
}


static void reschedule_event(
                              ix_event_node* arg_pDueList,
                              ix_event_node** arg_ppActiveList
                            )
{
    ix_ossl_time_t schedTime;
    ix_event_node* pNextActive;
    ix_event_node* pActiveList = *arg_ppActiveList;
    ix_ossl_time_get( &schedTime );
    IX_OSSL_TIME_ADD( schedTime, arg_pDueList->vPeriod );
    IX_OSSL_TIME_NORMALIZE( schedTime );
    IX_OSSL_TIME_SET( arg_pDueList->vExpiryTime, schedTime );

    /* If active list is empty or rescheduled event expires before
     * any event on the active list, store the event on the front of
     * the list.
     */
    if( !pActiveList 
        || IX_OSSL_TIME_LT( arg_pDueList->vExpiryTime,
                            pActiveList->vExpiryTime ) )
    {
        *arg_ppActiveList = arg_pDueList;
        arg_pDueList->pNext = pActiveList;
    }
    else
    {
        /** 
         * Insert the scheduled event in the active list according 
         * to expiry time
         */
        pNextActive = pActiveList->pNext;
        while( pNextActive
               && IX_OSSL_TIME_LT(pNextActive->vExpiryTime, schedTime) )
        {
            pActiveList = pNextActive;
            pNextActive = pActiveList->pNext;
        }
        pActiveList->pNext = arg_pDueList;
        arg_pDueList->pNext = pNextActive;
    }
}



static ix_error create_event(
                              ix_engine_proc_node* arg_pEngine,
                              ix_cc_handle arg_hCC,
                              ix_event_handle* arg_phEvent,
                              ix_event_node** arg_ppEventNode
                            )
{
    /* If there are events on the expired list, use one of those. Otherwise
     * create a new event.
     */
    ix_event_node* pEventNode;

    if( arg_pEngine->pExpiredEventList ) /* re-use an expired event */
    {
        pEventNode = arg_pEngine->pExpiredEventList;
        arg_pEngine->pExpiredEventList = pEventNode->pNext;
        
        *arg_phEvent = pEventNode->hEvent;
    }
    else /* create a new event */
    {
        /* Allocate memory for an event node */
        pEventNode = IX_OSSL_MALLOC( sizeof(ix_event_node) );
        IX_CCI_ERR_RET( !pEventNode, IX_CCI_ERR_OS_MEM_ALLOC )
        ix_ossl_memset( pEventNode, 0, sizeof(ix_event_node) );
        /* 
         * Convert event index to handle and store in function output argument. 
         * Since the engine's event list is sorted in expiration time order rather
         * than by event handle, also store the event handle in the event itself in
         * case the user wishes to cancel the event.
         */
        *arg_phEvent = pEventNode->hEvent =
            IX_CCI_EVENT_IDX_TO_HANDLE( arg_pEngine->vSizeEventList, arg_hCC );
        arg_pEngine->vSizeEventList++;
        IX_CCI_ASSERT( *arg_phEvent != IX_CCI_HANDLE_TYPES_INVALID )
    }
    *arg_ppEventNode = pEventNode;
    return IX_SUCCESS;
}


static ix_error activate_event(
                                ix_event_node* arg_pEvent,
                                ix_uint32 arg_When,
                                ix_engine_proc_node* arg_pEngine
                              )
{
    /* Convert the user-provided time into ix_ossl_time_t format and add it
     * to the current system time. Store the result in the event node
     */
    ix_error err;
    ix_event_node* pEventEntry;
    ix_event_node* pPrevEventEntry = NULL;
    ix_ossl_time_t sysTime;

    /* Convert time in ms to time in seconds and nanoseconds */
    arg_pEvent->vExpiryTime.sec = arg_When / 1000;
    arg_pEvent->vExpiryTime.nsec = (arg_When % 1000) * 1000000;
    IX_OSSL_TIME_NORMALIZE( arg_pEvent->vExpiryTime );
    arg_pEvent->vPeriod = arg_pEvent->vExpiryTime;
    arg_pEvent->vPeriodMilliSec = arg_When;
    /**
     * If there is an error getting the system time, put the event on
     * the expired list and return an error
     */
    err = ix_ossl_time_get( &sysTime );
    IX_CCI_ERR_GOTO( err != IX_SUCCESS, IX_CCI_ERR_EVENT_HANDLER_GET_TIME,
                     err_ret, err );
    
    IX_OSSL_TIME_ADD( arg_pEvent->vExpiryTime, sysTime );
    IX_OSSL_TIME_NORMALIZE( arg_pEvent->vExpiryTime );

    
    /* Add the event node in the appropriate time-sorted location in the 
     * engine's event list.
     */
    pEventEntry = arg_pEngine->pEventList;
    while( pEventEntry 
           && IX_OSSL_TIME_LT( pEventEntry->vExpiryTime, 
                               arg_pEvent->vExpiryTime ) )
    {
        pPrevEventEntry = pEventEntry;
        pEventEntry = pEventEntry->pNext;
    }
    if( !pPrevEventEntry )
    {
        arg_pEvent->pNext = pEventEntry;
        arg_pEngine->pEventList = arg_pEvent;
        /* First event: change engine semaphore's timeout to event timeoout */
        arg_pEngine->tNode.tEci.pTkInNotify->vSemaTimeout = arg_When;
    }
    else
    {
        pPrevEventEntry->pNext = arg_pEvent;
        arg_pEvent->pNext = pEventEntry;
    }
    return IX_SUCCESS;

err_ret:
    arg_pEvent->pNext = arg_pEngine->pExpiredEventList;
    arg_pEngine->pExpiredEventList = arg_pEvent;
    return err;
}


ix_error ix_cci_change_event(
                              ix_event_handle arg_hEvent,
                              ix_uint32 arg_ms
                            )
{
    ix_uint32 idxEng;
    ix_tk_input_notify* pEngTkNotifyIf;
    ix_engine_proc_node* pEngineProcIf;
    ix_event_node* pEvent;
    ix_ossl_time_t vPeriod;

    IX_CCI_PARAM_ASSERT( IX_CCI_GET_HANDLE_TYPE(arg_hEvent) == IX_CCI_HANDLE_TYPES_EVENT );

    /* Get the engine index from the event handle */
    idxEng = IX_HAND_TO_ENGINE_IDX( arg_hEvent );

    /* Get the Engine ECI via the engine's shared-memory token pointer */
    pEngTkNotifyIf = &g_pEngine[idxEng];
    pEngineProcIf = (ix_engine_proc_node*) pEngTkNotifyIf->pProcIfContext;

    /* Search through event lists for event core */
    pEvent = pEngineProcIf->pEventList;
    while( pEvent )
    {
        if( pEvent->hEvent == arg_hEvent )
        {
            /* Assert an error if event type is not periodic */
            IX_CCI_PARAM_ASSERT( pEvent->vEventType == IX_EVENT_TYPE_PERIODIC );

            /* Set new period and exit */
            pEvent->vPeriodMilliSec = arg_ms;

            vPeriod.sec = arg_ms / 1000;
            vPeriod.nsec = (arg_ms % 1000) * 1000000;
            IX_OSSL_TIME_NORMALIZE( vPeriod );
            pEvent->vPeriod = vPeriod;
            return IX_SUCCESS;
        }
        pEvent = pEvent->pNext;
    }
    pEvent = pEngineProcIf->pDueEventList;
    while( pEvent )
    {
        if( pEvent->hEvent == arg_hEvent )
        {
            /* Assert an error if event type is not periodic */
            IX_CCI_PARAM_ASSERT( pEvent->vEventType == IX_EVENT_TYPE_PERIODIC );

            /* Set new period and exit */
            pEvent->vPeriodMilliSec = arg_ms;

            vPeriod.sec = arg_ms / 1000;
            vPeriod.nsec = (arg_ms % 1000) * 1000000;
            IX_OSSL_TIME_NORMALIZE( vPeriod );
            pEvent->vPeriod = vPeriod;
            return IX_SUCCESS;
        }
        pEvent = pEvent->pNext;
    }
    /* If event isn't on active list or due list return an error */
    return IX_CCI_ERR_NEW( IX_CCI_ERR_EVENT_NOT_FOUND );
}
