/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation.All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 5
 *
 * = FILENAME
 *      policy.c
 *
 * = DESCRIPTION
 *      Functions to support policies
 *
 * = AUTHOR
 *       Roland Arajs
 *       roland.arajs@intel.com
 *
 * = CHANGE HISTORY
 *      5/1/2002 4:01:25 PM - creation time
 *
 * ============================================================================
 * $Id: policy.c,v 1.21 2003/01/25 00:17:43 ageorges Exp $
 */

#include "ix_ossl.h"
#include "ix_types.h"
#include "ix_rm.h"
#include "cci/internal/internal_cci_handles.h"
#include "cci/internal/internal_engine.h"
#include "cci/internal/internal_policy.h"
#include "cci/internal/internal_cci_error.h"
#include "cci/internal/internal_eci_node.h"
#include "cci/internal/internal_tk_notify_if.h"


/*******************************************************************************
 * Declarations of internal functions with file scope only
 */




/**
 * NAME: ix_cci_process_wrr_policy
 *       ix_cci_process_rr_policy
 *       ix_cci_process_sp_policy
 *
 * DESCRIPTION:   Processing functions in the execution control tree. 
 *                Three types of policy process functions are available, for
 *                round-robin, weighted round-robin, and strict-priority
 *                algorithms. 
 * 
 * @Param:  - IN  arg_pPolicy           Pointer to policy's process-interface 
 *                                      construct
 * @Param:  - OUT arg_pProcStatus       Pointer where the function returns the
 *                                      whether there are items still left in the 
 *                                      queue to process or high-priority 
 *                                      override condition
 * @Param:  - INOUT arg_pProcCount      Pointer to engine process count
 * @Param:  - INOUT arg_pEngProcNode    Pointer to engine proc node, which is 
 *                                      passed down to ECI leafs so that they
 *                                      may process events without having to
 *                                      return to the main engine run_loop()
 *                                      function.
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
ix_error ix_cci_process_wrr_policy(
                                    ix_cci_exe_ctrl_if* arg_pPolicy,
                                    ix_uint32* arg_pProcStatus,
                                    ix_uint32* arg_pProcCount,
                                    ix_engine_proc_node* arg_pEngProcNode
                                  );
ix_error ix_cci_process_rr_policy(
                                   ix_cci_exe_ctrl_if* arg_pPolicy,
                                   ix_uint32* arg_pNumInputsPending,
                                   ix_uint32* arg_pProcCount,
                                   ix_engine_proc_node* arg_pEngProcNode
                                 );

ix_error ix_cci_process_sp_policy(
                                   ix_cci_exe_ctrl_if* arg_pPolicy,
                                   ix_uint32* arg_pNumInputsPending,
                                   ix_uint32* arg_pProcCount,
                                   ix_engine_proc_node* arg_pEngProcNode
                                 );


/*******************************************************************************
 * Internal macros
 */

void ix_cci_get_first_pending(                                           
                               volatile ix_uint32* arg_FlagArray,                            
                               ix_uint32 arg_NumFlags,                             
                               ix_uint32* arg_pChildIdx,                             
                               ix_uint32 arg_IsClearFlag
                             )                                           
{                                                                           
    volatile ix_uint32* pFlagArray = arg_FlagArray;                                  
    ix_uint32 numQFlags = (arg_NumFlags + 3) >> 2;                          
    ix_uint32 i;                                                            
    ix_uint32 j = 0;                                                        
    volatile ix_uint8* pFlag;                                                        
    for( i=0; i<numQFlags; i++ )                                            
    {                                                                       
        if( *pFlagArray > 0 )                                               
        {                                                                   
            pFlag = (ix_uint8*) pFlagArray;                                 
            for( ; j<4; j++ )                                               
            {                                                               
                if( *pFlag )                                                
                {  
                    if( arg_IsClearFlag )
                        *pFlag = 0;
                    break;                                                  
                }                                                           
                pFlag++;                                                    
            }                                                               
            break;                                                          
        }                                                                   
        pFlagArray++;                                                       
    }                                                                       
    *arg_pChildIdx = i == numQFlags ?  IX_CCI_NO_SELECTED_CHILD : ((i*4)+j);  
} 

                                 
void ix_cci_get_next_pending(                                    
                              volatile ix_uint32* arg_FlagArray,                     
                              ix_uint32 arg_StartFlag,                     
                              ix_uint32 arg_NumFlags,                      
                              ix_uint32* arg_pChildIdx
                            )                                    
{
    ix_uint32 numQFlags = (arg_NumFlags + 3) >> 2;                  
    ix_uint32 qFlagIdx = (arg_StartFlag + 3) >> 2;                  
    ix_uint32 flagIdx = qFlagIdx << 2;                              
    volatile ix_uint8* pFlags = (ix_uint8*)arg_FlagArray + arg_StartFlag;    
    ix_uint32 byteOff = flagIdx - arg_StartFlag;
    ix_uint32 i = byteOff;                                          
    ix_uint32 j;                                                    
    *arg_pChildIdx = IX_CCI_NO_SELECTED_CHILD;
    while( i > 0 )                                                  
    {                                                               
        if(*pFlags > 0)                                             
        {                                                           
            *arg_pChildIdx = flagIdx - i;
            break;
        }
        if( *arg_pChildIdx != IX_CCI_NO_SELECTED_CHILD )              
            break;                                                  
        pFlags++;
        i--;                                                        
    }                                                               
    if( *arg_pChildIdx != IX_CCI_NO_SELECTED_CHILD )                  
        return;                                                      
    for( i=qFlagIdx; i<numQFlags; i++ )                             
    {                                                               
        if( *(ix_uint32*)pFlags > 0 )                               
        {                                                           
            for( j=0; j<4; j++ )                                    
            {                                                       
                if( *pFlags )                                       
                {                                                   
                    *arg_pChildIdx = 4*i + j;       
                    return;                                          
                }                                                   
                pFlags++;                                           
            }                                                       
        }                                                           
        else
            pFlags += sizeof(ix_uint32);
    }             
    for( pFlags=(ix_uint8*)arg_FlagArray, i=0; i<qFlagIdx; i++ )    
    {                                                               
        if( *(ix_uint32*)pFlags > 0 )                               
        {                                                           
            for( j=0; j<4; j++ )                                    
            {                                                       
                if( *pFlags )                                       
                {                                                   
                    *arg_pChildIdx = 4*i + j;                                 
                    return;                                          
                }                                                   
                pFlags++;                                           
            }                                                       
        }                                                           
        else
            pFlags += sizeof(ix_uint32);
    }                                                               
}
                 
ix_error ix_cci_process_selected_child(                                          
                                        ix_eci_node* arg_pPolicy,                             
                                        ix_uint32* arg_pIsStillPending,
                                        ix_uint32* arg_pProcCount,
                                        ix_engine_proc_node* arg_pEngProcNode
                                      )                                          
{                                                                               
    /* Call the process function in the selected child */  
    ix_error err;
    ix_uint32 idxChild = arg_pPolicy->vSelectedChild;
    ix_cci_exe_ctrl_if* pChild =                                                
        arg_pPolicy->tChildren.pEntries[idxChild].pEci;                         

    err = pChild->pProcessFunc( pChild, arg_pIsStillPending, 
                          arg_pProcCount, arg_pEngProcNode );                        
    IX_CCI_IXERR_RET( err );
                                                                                
    /* If child process shows data is still pending, re-enable child's          
     * pending flag.                                                            
     */                                                                         
    if( *arg_pIsStillPending & IX_CCI_DATA_STILL_PENDING )                                                
    {                                                                           
        arg_pPolicy->pPending->flag[idxChild] = 1;                              
    } 
    return IX_SUCCESS;
} 






/*******************************************************************************
 * Public API function definitions
 */
ix_error ix_cci_policy_create(
                               ix_policy_type argPolicyType,
                               ix_policy_handle* arg_pHandle
                             )
{
    ix_error err;
    ix_uint32 idxPolicy;
    ix_policy_handle hPolicy;
    ix_uint32 initContainerSize = 0;
    ix_uint32 containerGrowIncr = 0;
    ix_cci_process_tokens_ptr procFunc = NULL;
    ix_eci_node* pEciNode;
    IX_CCI_PARAM_ASSERT( argPolicyType < IX_POLICY_TYPE_LAST )
    IX_CCI_PARAM_ASSERT( arg_pHandle != NULL )

    /* Get the global semaphore */
    ix_ossl_sem_take( *g_pShmemSema, IX_OSSL_WAIT_FOREVER );

    idxPolicy = (*g_pPolicyCount)++;
    *arg_pHandle = hPolicy = IX_POLICY_IDX_TO_HANDLE( idxPolicy, 0 );
    IX_CCI_ASSERT( *arg_pHandle != IX_CCI_HANDLE_TYPES_INVALID );

    pEciNode = IX_OSSL_MALLOC( sizeof(ix_eci_node) );

    IX_CCI_ERR_GOTO( !pEciNode, IX_CCI_ERR_POLICY_CREATE, err_ret1, err )
    ix_ossl_memset( pEciNode, 0, sizeof(ix_eci_node) );

    switch( argPolicyType )
    {
    case IX_POLICY_TYPE_RR:
        initContainerSize = IX_CCI_INIT_RR_POLICY_SIZE;
        containerGrowIncr = IX_CCI_RR_POLICY_GROW_INCR;
        procFunc = ix_cci_process_rr_policy;
        break;
    case IX_POLICY_TYPE_WRR:
        initContainerSize = IX_CCI_INIT_WRR_POLICY_SIZE;
        containerGrowIncr = IX_CCI_WRR_POLICY_GROW_INCR;
        procFunc = ix_cci_process_wrr_policy;
        break;
    case IX_POLICY_TYPE_SP:
        initContainerSize = IX_CCI_INIT_SP_POLICY_SIZE;
        containerGrowIncr = IX_CCI_SP_POLICY_GROW_INCR;
        procFunc = ix_cci_process_sp_policy;
        break;

    default:
        IX_CCI_ERR_NOCHECK_GOTO( IX_CCI_ERR_INVALID_POLICY_TYPE, err_ret2, err )
    }
    
    /* Call node-initialization function */
    err = ix_cci_eci_node_init( pEciNode, procFunc, ix_cci_eci_node_fini,
                                hPolicy, initContainerSize, containerGrowIncr );
    IX_CCI_IXERR_GOTO( err_ret2, err )

    /* Initialize the token-notify interface */
    ix_cci_tk_notify_register_eci( hPolicy, pEciNode );


    ix_ossl_sem_give( *g_pShmemSema );
    return err; /* Normal error return */

err_ret2:
    IX_OSSL_FREE( pEciNode );
#if !defined(_CCT_EXCLUDE_)
err_ret1:
    ix_ossl_sem_give( *g_pShmemSema );
#endif
    return err;
}



/* Policy processing functions */
ix_error ix_cci_process_wrr_policy(
                                    ix_cci_exe_ctrl_if* arg_pPolicy,
                                    ix_uint32* arg_pIsInputPending,
                                    ix_uint32* arg_pProcCount,
                                    ix_engine_proc_node* arg_pEngProcNode
                                  )
{
    ix_error err;
    ix_uint32 vSelectedChild;
    ix_eci_node* pPolicy = (ix_eci_node*)arg_pPolicy;
    volatile ix_uint32* pQFlagPending;
    volatile ix_uint8* pFlagPending;
    ix_uint32 numFlags;

    IX_CCI_ASSERT( arg_pPolicy != NULL )
    IX_CCI_ASSERT( arg_pIsInputPending != NULL )
    IX_CCI_ASSERT( arg_pProcCount != NULL )
    vSelectedChild = pPolicy->vSelectedChild;
    pQFlagPending = pPolicy->pPending->qflag;
    pFlagPending = pPolicy->pPending->flag;
    numFlags = pPolicy->tChildren.vNumEntries;

    /* Examine pending flag for selected child */
    /* If not set */
    if( !pFlagPending[vSelectedChild] )
    {
        /* Find next pending, clearing its pending flag if found */
        if(++vSelectedChild == numFlags)
            vSelectedChild = 0;

        ix_cci_get_next_pending( pQFlagPending, 
                                 vSelectedChild, /* this is the next search position */
                                 numFlags,
                                 &vSelectedChild );
        /* Exit if no pending child is found */
        if( vSelectedChild == IX_CCI_NO_SELECTED_CHILD )
        {
            return IX_SUCCESS;
        }
        else
        {
            pPolicy->vSelectedChild = vSelectedChild;
        }
    }
    /* While selected child has pending data */
    do
    {
        ix_tkp_entry* pEntry = 
            pPolicy->tChildren.pEntries + vSelectedChild;
        /*
         * Call process function for selected child, and if child indicates
         * there is more to process, set pending flag for that child. Function
         * returns with pending/high-priority-override indicator in 2nd arg.
         */
        pFlagPending[vSelectedChild] = 0;
        *arg_pIsInputPending = 0;
        err = ix_cci_process_selected_child( pPolicy, arg_pIsInputPending, 
                                             arg_pProcCount, arg_pEngProcNode );
        IX_CCI_IXERR_RET( err );

        /* 
         * Decrement number of credits for selected child 
         * If number of credits has dropped to zero or selected child has no
         * more data
         */
        if( 0 == --(pEntry->vCredits) 
            || !(*arg_pIsInputPending & IX_CCI_DATA_STILL_PENDING) )
        {
            /* Set credits to weight */
            pEntry->vCredits = pEntry->vWeight;

            /* Find next pending */
            if(++vSelectedChild == numFlags)
                vSelectedChild = 0;

            ix_cci_get_next_pending( pQFlagPending, 
                                     vSelectedChild, /* this is the next search position */
                                     numFlags,
                                     &vSelectedChild );
            /* If new pending child */
            if( vSelectedChild != IX_CCI_NO_SELECTED_CHILD )
            {
                /* Set selected child to new pending child */
                pPolicy->vSelectedChild = vSelectedChild;

                /* Set out arg to indicate policy still has data to process */
                *arg_pIsInputPending |= IX_CCI_DATA_STILL_PENDING;
            }
            else
            {
                /* Start next search from next child */
                pPolicy->vSelectedChild 
                    = vSelectedChild == numFlags-1 ? 0 : vSelectedChild + 1;
            }
        }
/*        ix_ossl_sleep( 0 );*/
        /* 
         * Exit if nothing more to process or high-priority override is set or 
         * max process count is reached 
         */
    } while( *arg_pIsInputPending == IX_CCI_DATA_STILL_PENDING 
             && *arg_pProcCount > 0 );
    return IX_SUCCESS;
}

ix_error ix_cci_process_rr_policy(
                                   ix_cci_exe_ctrl_if* arg_pPolicy,
                                   ix_uint32* arg_pIsInputPending,
                                   ix_uint32* arg_pProcCount,
                                   ix_engine_proc_node* arg_pEngProcNode
                                 )
{
    ix_error err;
    ix_uint32 vSelectedChild;
    ix_uint32 numFlags;
    volatile ix_uint32* pQFlagPending;
    volatile ix_uint8* pFlagPending;
    ix_uint32 nextSearchPos;
    ix_eci_node* pPolicy = (ix_eci_node*)arg_pPolicy;

    IX_CCI_ASSERT( arg_pPolicy != NULL )
    IX_CCI_ASSERT( arg_pIsInputPending != NULL )
    pQFlagPending = pPolicy->pPending->qflag;
    pFlagPending = pPolicy->pPending->flag;

    numFlags = pPolicy->tChildren.vNumEntries;
    vSelectedChild = pPolicy->vSelectedChild;

    /* Examine pending flag for selected child */
    /* If not set */
    if( !pFlagPending[vSelectedChild] )
    {
        /* Find next pending, clearing its pending flag if found */
        ix_uint32 nextSearchPos 
            = vSelectedChild == numFlags-1 ? 0 : vSelectedChild + 1;

        ix_cci_get_next_pending( pQFlagPending, 
                                 nextSearchPos,
                                 numFlags,
                                 &vSelectedChild );
        /* Exit if no pending child is found */
        if( vSelectedChild == IX_CCI_NO_SELECTED_CHILD )
        {
            return IX_SUCCESS;
        }
        else
            pPolicy->vSelectedChild = vSelectedChild;
    }
    /* While selected child has pending data */
    do
    {
        /* Clear pending flag of selected child */
        pFlagPending[vSelectedChild] = 0;
        /*
         * Call process function for selected child. Function returns with pending/
         * high-priority-override indicator in *arg_pIsInputPending.
         */
        *arg_pIsInputPending = 0;
        err = ix_cci_process_selected_child( pPolicy, arg_pIsInputPending, 
                                             arg_pProcCount, arg_pEngProcNode );
        IX_CCI_IXERR_RET( err );

        /* Find next child with data pending (maintaining its pending flag) */     
        nextSearchPos = vSelectedChild == numFlags - 1 ? 0: vSelectedChild + 1;
        ix_cci_get_next_pending( pQFlagPending,                                 
                                 nextSearchPos, numFlags, &vSelectedChild );

        /* If pending child found */                                            
        if( vSelectedChild != IX_CCI_NO_SELECTED_CHILD )                          
        {
            /* Set policy's selected flag to new child index */                     
            pPolicy->vSelectedChild = vSelectedChild;                            
            /* Set output argument */
            *arg_pIsInputPending |= IX_CCI_DATA_STILL_PENDING;                                             
        }
        else
        {
            pPolicy->vSelectedChild 
                = vSelectedChild == numFlags-1 ? 0 : vSelectedChild + 1;
        }
/*        ix_ossl_sleep( 0 );*/
        /* 
         * Exit if nothing more to process or high-priority override is set or 
         * max process count is reached 
         */
    } while( *arg_pIsInputPending == IX_CCI_DATA_STILL_PENDING 
             && *arg_pProcCount > 0 );
    return IX_SUCCESS;
}

ix_error ix_cci_process_sp_policy(
                                   ix_cci_exe_ctrl_if* arg_pPolicy,
                                   ix_uint32* arg_pIsInputPending,
                                   ix_uint32* arg_pProcCount,
                                   ix_engine_proc_node* arg_pEngProcNode
                                 )
{
    /* This function assumes *arg_pIsInputPending is set to 0 on entry */
    ix_error err;
    ix_uint32 numFlags;
    volatile ix_uint32* pQFlagPending;
    ix_eci_node* pPolicy = (ix_eci_node*)arg_pPolicy;
    ix_uint32 vSelectedChild = IX_CCI_NO_SELECTED_CHILD;
    IX_CCI_ASSERT( arg_pPolicy != NULL )
    IX_CCI_ASSERT( arg_pIsInputPending != NULL )

    pQFlagPending = pPolicy->pPending->qflag;
    numFlags = pPolicy->tChildren.vNumEntries;

    do
    {
        /* Check the pending flags. Get and clear the first that is set.*/
        *arg_pIsInputPending = 0;
        ix_cci_get_first_pending( pQFlagPending, numFlags, &vSelectedChild, 1 );
        if( vSelectedChild != IX_CCI_NO_SELECTED_CHILD )
        {
            /* Run the selected child's process function. If process function 
             * indicates more data needs to processed from that queue, set its 
             * pending flag again
             */
            pPolicy->vSelectedChild = vSelectedChild;
            err = ix_cci_process_selected_child( pPolicy, arg_pIsInputPending,
                                                 arg_pProcCount, arg_pEngProcNode );
            IX_CCI_IXERR_RET( err );
            if( !(*arg_pIsInputPending & IX_CCI_DATA_STILL_PENDING) )
            {
                /* Find out whether any other child has data pending */     
                ix_cci_get_first_pending( pQFlagPending, 
                                          numFlags, &vSelectedChild, 0 );
                if( vSelectedChild != IX_CCI_NO_SELECTED_CHILD)
                {
                    *arg_pIsInputPending |= IX_CCI_DATA_STILL_PENDING;                                             
                }
            }
        }
        if( *arg_pIsInputPending & IX_CCI_HIGH_PRIORITY_OVERRIDE
            && arg_pEngProcNode->pPriorityChangeUpNode == pPolicy )
        {
            /* 
             * Clear the override bit to avoid being forced to return further up
             * the policy tree than necessary.
             */
            *arg_pIsInputPending &= ~IX_CCI_HIGH_PRIORITY_OVERRIDE;
        }
/*      ix_ossl_sleep( 0 );*/
    } while( *arg_pIsInputPending == IX_CCI_DATA_STILL_PENDING 
             && *arg_pProcCount > 0);
    
    return IX_SUCCESS;
}


ix_error ix_cci_policy_destroy( ix_policy_handle arg_Handle )
{
    ix_error err;
    /* Convert handle to token-notify interface pointer */
    ix_uint32 idxPolicy = IX_HAND_TO_POLICY_IDX( arg_Handle );
    ix_cci_exe_ctrl_if* pPolicy;

    IX_CCI_PARAM_ASSERT( idxPolicy != IX_CCI_INVALID_INDEX )

    /* Get pointer to policy structure from token-notify I/F */
    pPolicy = (ix_cci_exe_ctrl_if*)g_pPolicy[idxPolicy].pProcIfContext;

    /* Call node fini function */
    err = ix_cci_eci_node_fini( pPolicy );
    IX_CCI_IXERR_RET( err )
    return IX_SUCCESS;
}


ix_error ix_cci_policy_add_branch(
                                   ix_policy_handle arg_hPolicy,
                                   ix_policy_handle arg_hChild,
                                   ix_uint32 arg_WeightOrPriority
                                 )
{
    /* Convert handles to pointers */
    ix_error err;
    ix_uint32 idxChild;
    ix_cci_exe_ctrl_if* pChild;
    ix_eci_node* pPolicy;
    ix_token_processor_container* pContainer;

    ix_uint32 idxPolicy = IX_HAND_TO_POLICY_IDX( arg_hPolicy );


    /* Check both handles are valid policy handles */
    IX_CCI_PARAM_ASSERT( idxPolicy != IX_CCI_INVALID_INDEX )

    idxChild = IX_HAND_TO_POLICY_IDX( arg_hChild );
    IX_CCI_PARAM_ASSERT( idxChild != IX_CCI_INVALID_INDEX );

    /* Add child policy to parent policy's container */
    pPolicy = (ix_eci_node*)g_pPolicy[idxPolicy].pProcIfContext;
    pChild = (ix_cci_exe_ctrl_if*)g_pPolicy[idxChild].pProcIfContext;
    pContainer = &pPolicy->tChildren;
    err = ix_cci_tkpc_add_child( pContainer,pChild, arg_WeightOrPriority );
    IX_CCI_IXERR_RET( err )

    /* If parent's selected child indicates no selected child, set it to 0 */
    if( pPolicy->vSelectedChild == IX_CCI_NO_SELECTED_CHILD )
        pPolicy->vSelectedChild = 0;

    /* Store parent's notify i/f pointer in the child's notify i/f */
    g_pPolicy[idxChild].pParent = g_pPolicy + idxPolicy;

    return IX_SUCCESS;
}


ix_error ix_cci_policy_add_leaf(
                                 ix_policy_handle arg_hPolicy,
                                 ix_uint32 arg_InputId,
                                 ix_uint32 arg_IsPacketId,
                                 ix_uint32 arg_WeightOrPriority
                               )
{
    /* Convert handles to pointers */
    ix_cci_exe_ctrl_if* pChild;
    ix_eci_node* pPolicyEci;
    ix_tk_input_notify* pPolicyNotifyIf;
    ix_token_processor_container* pContainer;

    ix_uint32 idxPolicy = IX_HAND_TO_POLICY_IDX( arg_hPolicy );


    /* Check first function arg is a valid policy handle */
    IX_CCI_PARAM_ASSERT( idxPolicy != IX_CCI_INVALID_INDEX )
    IX_CCI_PARAM_ASSERT( arg_InputId > IX_COMM_LAST_UBLOCK_ID 
                         && arg_InputId < 256 )

    /**
     * Check index to packet or message queue is reasonable and convert index
     * to execution-control structure
     */
    if( arg_IsPacketId ) /* Packet token queue */
    {
        IX_CCI_ASSERT( arg_InputId < IX_CCI_MAX_INPUT_IDS )
        pChild = (ix_cci_exe_ctrl_if*)
            g_pPktTkq[arg_InputId - IX_COMM_LAST_UBLOCK_ID -1].pProcIfContext;
    }
    else    /* Message token queue */
    {
        IX_CCI_ASSERT( arg_InputId < IX_CCI_MAX_INPUT_IDS )
        pChild = (ix_cci_exe_ctrl_if*)
            g_pMsgTkq[arg_InputId - IX_COMM_LAST_UBLOCK_ID -1].pProcIfContext;
    }

    /* 
     * Connect token-notify interface of child to token-notify interface of 
     * policy 
     */
    pPolicyNotifyIf = g_pPolicy + idxPolicy;
    pPolicyEci = (ix_eci_node*)g_pPolicy[idxPolicy].pProcIfContext;
    pContainer = &pPolicyEci->tChildren;
    ix_cci_connect_parent( pChild->pTkInNotify, pPolicyNotifyIf );

    /* Add pkt/msg token queue to parent policy's container */
    return ix_cci_tkpc_add_child( pContainer, pChild, arg_WeightOrPriority );
}


void ix_cci_policy_clear_entry(
                                ix_tk_input_notify* arg_pPolicyNotifyIf,
                                ix_cci_exe_ctrl_if* arg_pEntry
                              )
{
    ix_tkp_entry* pTkpEntry;
    int numEntries;
    int i;
    /*
     * Get pointer to policy from handle and from there get pointer to 
     * entries in child container and number of entries in child container.
     */
    ix_eci_node* pPolicy = (ix_eci_node*)arg_pPolicyNotifyIf->pProcIfContext;
    pTkpEntry = pPolicy->tChildren.pEntries;
    numEntries = pPolicy->tChildren.vNumEntries;

    /* Search for leaf in child container */
    for( i=0; i<numEntries; i++ )
    {
        /* If found clear the associated policy fields */
        if( pTkpEntry[i].pEci == arg_pEntry )
        {
            pTkpEntry[i].pEci = NULL;
            pTkpEntry[i].vCredits = 0;
            pTkpEntry[i].vWeight = 0;
            break;
        }
    }
}

int ix_cci_is_high_priority_override( 
                                      ix_tk_input_notify* arg_pNotifyIf,
                                      ix_engine_proc_node* arg_pEngProcIf
                                    )
{
    int vRetVal = 0;
    ix_tk_input_notify* pNotifyIf = arg_pNotifyIf;
    /* Clear new-token indicator */
    *(pNotifyIf->pIsNewToken) = 0;
    pNotifyIf = pNotifyIf->pParent;

    /* Search up tree for SP policy nodes */
    while( pNotifyIf )
    {
        ix_eci_node* pNode = pNotifyIf->pProcIfContext;
        if( pNode->tEci.pProcessFunc == ix_cci_process_sp_policy )
        {
            ix_uint32 childIdx = IX_CCI_NO_SELECTED_CHILD;
            ix_uint32 numFlags = pNode->tChildren.vNumEntries;
            volatile ix_uint32* pQFlagPending = pNode->pPending->qflag;
            /* Get first pending flag without changing it.*/
            ix_cci_get_first_pending( pQFlagPending, numFlags, &childIdx, 0 );
            if( childIdx < pNode->vSelectedChild )
            {
                /*
                 * SP policy nearest policy root with priority override gets
                 * stored in the engine proc node structure
                 */
                arg_pEngProcIf->pPriorityChangeUpNode = pNode;
                vRetVal = IX_CCI_HIGH_PRIORITY_OVERRIDE;
            }
        }
        pNotifyIf = pNotifyIf->pParent;
    }
    return vRetVal;
}
