/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation.All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 5
 *
 * = FILENAME
 *      tk_notify.c
 *
 * = DESCRIPTION
 *      All engines, policies, and token queues consist of a token_notify
 *      interface and an execution-control interface .
 *      The code in this module contains functions that support the 
 *      ix_tk_input_notify construct declared in internal_exe_ctrl_if.h. 
 *      Constructs of this type will exist in RM-allocated shared memory, and 
 *      are accessed in the thread of a send-message or send-packet function, to
 *      set policy pending flags and wake up the engine by releasing the engine 
 *      semaphore.
 *
 * = AUTHOR
 *       Roland Arajs
 *       roland.arajs@intel.com
 *
 * = CHANGE HISTORY
 *      4/28/2002 10:46:21 AM - creation time
 *
 * ============================================================================
 * $Id: tk_notify.c,v 1.7 2003/08/18 21:48:17 ageorges Exp $
 */

#include "ix_ossl.h"

#if !( _IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_ )
#include <stddef.h>
#endif /* !( _IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_ ) */

#include "ix_rm.h"

#include "cci/internal/internal_exe_ctrl_if.h"
#include "cci/internal/internal_tk_notify_if.h"
#include "cci/internal/internal_cci_handles.h"
#include "cci/internal/internal_engine.h"

#if defined(_UT_COVERAGE_)
#include "cci/cci_engine_stubs.h"
#endif



/**
 * Definitions of private functions called from elsewhere in CCI
 */
ix_error ix_cci_tk_notify_register_eci(
                                        ix_tk_handle    arg_hTkNotify,
                                        void*           arg_pEci
                                      )
{
    ix_error err;
    ix_tk_input_notify* pTkNotify = NULL;
    ix_uint32   i;

#if defined(_UT_COVERAGE_)
    incr_tk_notify_reg_eci_counter();
#endif
    
    /* Convert token-notify handle to shared-memory pointer */
    err = ix_cci_tk_notify_h_to_ptr( arg_hTkNotify, &pTkNotify );
    IX_CCI_DBG_IXERR_RET( err )

    /**
     * The API functions get indirect access to the execution control constructs
     * of execution engines and policies through the token-notify interface.
     * This is because the APIs are defined using handles, and the handles are
     * indexes to these constructs, not the execution-control constructs.
     */
    pTkNotify->pProcIfContext = arg_pEci;

    /* Clear pending flags */
    for( i=0; i<IX_CCI_POLICY_MAX_QFLAGS; i++)
    pTkNotify->tPending.qflag[i] = 0;

    /* Initialize tk-notify connect fields to disconnected status */
    pTkNotify->pParent = NULL;
    pTkNotify->vId = IX_CCI_NO_SELECTED_CHILD;

    /* 
     * If this is notify interface for an engine, initialize the semaphore in
     * this interface and set corresponding timeout value 
     */
    if( (arg_hTkNotify & IX_CCI_H_TYPE_MASK) == IX_CCI_HANDLE_TYPES_EXE )
    {
        err = IX_OSSL_SEM_INIT( IX_OSSL_SEM_UNAVAILABLE, &pTkNotify->vSema );
        pTkNotify->vSemaTimeout = IX_CCI_EVENT_CHECK_INTERVAL;
        IX_CCI_ERR_RET( err != IX_SUCCESS, 
                        IX_CCI_ERR_ENG_INIT_SEMA_INIT )
    }
    return err;
}

ix_error ix_cci_connect_parent(
                                ix_tk_input_notify*   arg_pTkNotify,
                                ix_tk_input_notify*   arg_pParent
                              )
{
    /*
    ix_error err;
    ix_tk_input_notify* pTkNotify = NULL;
    ix_tk_input_notify* pParent = NULL;
    */
    /* Convert token-notify handles to shared-memory pointer */
    /*
    err = ix_cci_tk_notify_h_to_ptr( arg_hParent, &pParent );
    IX_CCI_IXERR_RET( err )
    err = ix_cci_tk_notify_h_to_ptr( arg_hTkNotify, &pTkNotify );
    IX_CCI_IXERR_RET( err )
    */

    /* Store parent pointer */
    arg_pTkNotify->pParent = arg_pParent;    

    return IX_SUCCESS;
}

ix_error ix_cci_tk_notify_fini( ix_tk_input_notify* arg_pTkNotify )
{
    ix_uint32 i;
#if defined(_UT_COVERAGE_)
    incr_tk_notify_fini_counter();
#endif
    /* It might be nice to disable inputs to the communication ID */

        
   /* If pending flag is set, return an error */
    for( i=0; i<IX_CCI_POLICY_MAX_QFLAGS; i++)
    IX_CCI_ERR_RET( arg_pTkNotify->tPending.qflag[i] > 0,
                    IX_CCI_ERR_TK_NOTIFY_FINI_STILL_PENDING )

    /* Just reset all the fields we can */
    arg_pTkNotify->pParent = NULL;
    arg_pTkNotify->pProcIfContext = NULL;
    arg_pTkNotify->vId = IX_CCI_NO_SELECTED_CHILD;
    if( arg_pTkNotify->vSema )
        IX_OSSL_SEM_FINI( arg_pTkNotify->vSema );
    return IX_SUCCESS;
}


ix_error ix_cci_tk_notify_h_to_ptr(
                                    ix_tk_handle    arg_hTkNotify,
                                    ix_tk_input_notify** arg_ppTkNotify
                                  )
{
     ix_tk_input_notify* pTkNotify = NULL;
    
    /* Convert token-notify handles to shared-memory pointer */
    ix_cci_handle_types handleType = IX_CCI_GET_HANDLE_TYPE( arg_hTkNotify );
    switch( handleType )
    {
    ix_uint32 idx;
    case IX_CCI_HANDLE_TYPES_PKTTKQ:
        idx = IX_HAND_TO_PKT_IDX( arg_hTkNotify );
        pTkNotify = g_pPktTkq + idx;
        break;
    case IX_CCI_HANDLE_TYPES_MSGTKQ:
        idx = IX_HAND_TO_MSG_IDX( arg_hTkNotify );
        pTkNotify = g_pMsgTkq + idx;
        break;
    case IX_CCI_HANDLE_TYPES_POLICY:
        idx = IX_HAND_TO_POLICY_IDX( arg_hTkNotify );
        pTkNotify = g_pPolicy + idx;
        break;
    case IX_CCI_HANDLE_TYPES_EXE:
        idx = IX_HAND_TO_ENGINE_IDX( arg_hTkNotify );
        pTkNotify = g_pEngine + idx;
        break;
    default:
        /* Any other handle type is invalid in this function */
        return IX_CCI_ERR_NEW( IX_CCI_ERR_TK_NOTIFY_INVALID_HANDLE );
    }
    *arg_ppTkNotify = pTkNotify;
    return IX_SUCCESS;
}

ix_error ix_cci_set_tk_notify_id(
                                  ix_tk_input_notify* arg_pTkNotify,
                                  ix_uint32 arg_Id
                                )
{
    IX_CCI_ASSERT( arg_pTkNotify != NULL )
    arg_pTkNotify->vId = arg_Id;
    return IX_SUCCESS;
}
