/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation.All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 5
 *
 * = FILENAME
 *      token_proc_container.c
 *
 * = DESCRIPTION
 *      Supports execution control interfaces that are branch nodes and 
 *      therefore have child nodes. It defines a container for these nodes.
 *
 * = AUTHOR
 *       Roland Arajs
 *       roland.arajs@intel.com
 *
 * = CHANGE HISTORY
 *      2/12/2002 10:39:27 AM - creation time
 *
 * ============================================================================
 * $Id: token_proc_container.c,v 1.8 2003/01/18 00:02:45 ageorges Exp $
 */
#include "ix_ossl.h"

#if !( _IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_ )
#include <stddef.h>
#endif /* !( _IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_ ) */

#include "cci/internal/internal_token_proc_container.h"
#include "cci/internal/internal_cci_error.h"


#if defined(_UT_COVERAGE_)
#include "cci/cci_engine_stubs.h"
#endif


/*******************************************************************************
 * Declarations of functions with file scope only
 */

/**
 * NAME: grow_container_add_child
 *
 * DESCRIPTION: Called if attempting to add a child to a full container. The
 *              container grows by the amount specified by the container's
 *              vGrowIncr member. If the vGrowIncr value is zero, the function
 *              returns an error.
 * 
 * @Param:  - IN arg_pContainer - pointer to container where child is to be 
 *                  added
 * @Param:  - IN arg_pChild - pointer to child.
 * @Param:  - IN arg_Weight - weight or priority to be assigned to child
 * @Param:  - IN arg_InsertPos - insertion position of child into container
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
static ix_error grow_container_add_child(
                          ix_token_processor_container* arg_pContainer,
                          ix_cci_exe_ctrl_if*           arg_pChild,
                          ix_uint32                     arg_Weight,
                          ix_uint32                     arg_InsertPos
                        );



/*******************************************************************************
 * Implementation of functions with CCI scope (usage in internal header file)
 */

/*
ix_error ix_cci_tkpc_remove_child(
                                   ix_token_processor_container*  arg_pContainer,
                                   ix_uint32                      arg_ChildId
                                 )
{
    ix_cci_exe_ctrl_if* pEci;
#if defined(_UT_COVERAGE_)
    incr_tkpc_remove_child_counter();
#endif

    IX_CCI_ASSERT( arg_pContainer != NULL )
    IX_CCI_ASSERT( arg_ChildId < arg_pContainer->vNumEntries )

    pEci = arg_pContainer->pEntries[ arg_ChildId ].pEci;
    pEci->pFiniFunc( pEci );
*/
    /* Don't delete the memory resources for this interface because it is
     * part of a container where the other members have handles whose values are
     * based on their container indexes.
     */
    /* Set the fini function to NULL to make sure it cannot be called again */
/*
    pEci->pFiniFunc = NULL;
    return IX_SUCCESS;
}
*/

/**
 * NAME: ix_cci_tkproc_container_init
 *
 * DESCRIPTION:   Container of ix_tkp_entry elements containing execution
 *                control pointers and weights/priorities.
 * 
 * @Param:  - IN arg_pContainer - a pointer to an uninitialized container
 *
 * @Return: 0 if successful or a valid ix_error token on failure.
 */
ix_error 
ix_cci_tkproc_container_init( ix_token_processor_container* arg_pContainer,
                              ix_uint32 arg_InitSize,
                              ix_uint32 arg_GrowIncr
                            )
{
#if defined(_UT_COVERAGE_)
    incr_tkpc_init_counter();
#endif
    IX_CCI_ASSERT( arg_pContainer != NULL )
    IX_CCI_ASSERT( arg_InitSize || arg_GrowIncr )

    if( arg_InitSize )
        arg_pContainer->pEntries = 
            IX_OSSL_MALLOC( sizeof(ix_tkp_entry) * arg_InitSize );
    IX_CCI_ERR_RET( !arg_pContainer->pEntries, 
                    IX_CCI_ERR_TKP_CONTAINER_MEM_ALLOC )
    ix_ossl_memset( arg_pContainer->pEntries, 0, 
                    sizeof(ix_tkp_entry) * arg_InitSize );
    arg_pContainer->vGrowIncr = arg_GrowIncr;
    arg_pContainer->vCapacity = arg_InitSize;
    arg_pContainer->vNumEntries = 0;

    return IX_SUCCESS;
}

ix_error ix_cci_tkproc_container_fini(
                                      ix_token_processor_container* arg_pContainer
                                    )
{
#if defined(_UT_COVERAGE_)
    incr_tkpc_fini_counter();
#endif
    IX_CCI_ASSERT( arg_pContainer != NULL )
    if( arg_pContainer->pEntries != NULL )
    {
        /* Call the fini functions in each child */
        ix_uint32 i;
        ix_cci_exe_ctrl_if* pCtrlIf;
        for( i=0; i<arg_pContainer->vNumEntries; i++ )
        {
            pCtrlIf = arg_pContainer->pEntries[i].pEci;
            if( pCtrlIf && pCtrlIf->pFiniFunc )
                pCtrlIf->pFiniFunc( pCtrlIf );
        }

        /* Delete the memory resources for this container */
        IX_OSSL_FREE( arg_pContainer->pEntries );
        arg_pContainer->pEntries = NULL;
    }
    
    return IX_SUCCESS;
}


ix_error ix_cci_tkpc_add_child(
                          ix_token_processor_container* arg_pContainer,
                          ix_cci_exe_ctrl_if*           arg_pChild,
                          ix_uint32                     arg_Weight
                        )
{
    ix_error err = 0;
    ix_uint32 insPos;
    ix_uint32 i;
    ix_uint32 size;
    ix_tkp_entry* pList;

    IX_CCI_ASSERT( arg_pContainer != NULL )
    IX_CCI_ASSERT( arg_pChild != NULL )
#if defined(_UT_COVERAGE_)
    incr_tkpc_add_child_counter();
#endif
    size = arg_pContainer->vNumEntries;
    pList = arg_pContainer->pEntries;

    if( arg_Weight == 0 )
        arg_Weight = 1;     /* Set to minimum weight */

    /* This container is sorted by weight (high to low). Find location
     * in container where this entry will belong
     */
    for( insPos=0; insPos<size; insPos++ )
    {
        if( arg_Weight > pList[insPos].vWeight )
            break;
    }

    /* If the container is full */
    if( size == arg_pContainer->vCapacity )
    {
        err = grow_container_add_child( arg_pContainer,
                                        arg_pChild,
                                        arg_Weight,
                                        insPos
                                      );
        IX_CCI_IXERR_RET( err )
    }
    else
    {
        /* Spare capacity in container: */
        /* Move the elements from the insert position up by one, updating the 
         * token-notify ID
         */
        for( i=size; i>insPos; i-- )
        {
            pList[i] = pList[i-1];
            ix_cci_set_notify_id( pList[i].pEci, i );
        }
        /* Insert the new entry values from the function arguments and update 
         * the token-notify ID for the new entry
         */
        pList[insPos].pEci = arg_pChild;
        pList[insPos].vWeight = arg_Weight;
        pList[insPos].vCredits = arg_Weight;
        ix_cci_set_notify_id( arg_pChild, insPos );
    }

    /* Increment the container's entry count */
    arg_pContainer->vNumEntries++;
     
    return IX_SUCCESS;
}



/**
 * Internal function implementation
 */
static ix_error grow_container_add_child(
                          ix_token_processor_container* arg_pContainer,
                          ix_cci_exe_ctrl_if*           arg_pChild,
                          ix_uint32                     arg_Weight,
                          ix_uint32                     arg_InsertPos
                        )
{
    ix_uint32 i;
    ix_uint32 size = arg_pContainer->vNumEntries;
    ix_tkp_entry* pList = arg_pContainer->pEntries;

    /* Return an error if the container cannot grow */
    ix_uint32 incr = arg_pContainer->vGrowIncr;
    ix_tkp_entry* pNewList;
    IX_CCI_ERR_RET( incr==0, IX_CCI_TOKEN_PROCESSOR_CONTAINER_FULL )

    /* Create a new list with incremented size*/
    pNewList = IX_OSSL_MALLOC( sizeof(ix_tkp_entry) * (size + incr));
    IX_CCI_ERR_RET( !pNewList, IX_CCI_ERR_TKP_CONTAINER_MEM_ALLOC )

    ix_ossl_memset( pNewList, 0, sizeof(ix_tkp_entry) * (size + incr) );

    arg_pContainer->vCapacity = size + incr;
    /* Copy the old list up to the insert pos to the new list */
    for( i=0; i<arg_InsertPos; i++ )
    {
        pNewList[i] = pList[i];
    }
    /* Insert the new entry values from the function args */
    pNewList[arg_InsertPos].pEci = arg_pChild;
    pNewList[arg_InsertPos].vWeight = arg_Weight;
    pNewList[arg_InsertPos].vCredits = arg_Weight;

    /* Send this new entry index to its token-notify interface */
    ix_cci_set_notify_id( arg_pChild, arg_InsertPos );
    
    /**
     * Insert the rest of the old list to the new list and update
     * the token-notify IDs with the new list position
     */
    for( i=arg_InsertPos; i<size; i++ )
    {
        pNewList[i+1] = pList[i];
        ix_cci_set_notify_id( pList[i].pEci, i+1 );
    }
    
    /* Delete the old list */
    IX_OSSL_FREE( pList );

    /* Update the container with the new list */
    arg_pContainer->pEntries = pNewList;
    return IX_SUCCESS;
}
