/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      atomic_operations.h
 *
 * = DESCRIPTION
 *      This file will declare all function prototypes for atomic operations support
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      5/24/2002 10:46:15 AM - creation time 
 * ============================================================================
 * $Id: atomic_operations.h,v 1.7 2003/08/08 20:24:58 ageorges Exp $
 */

#if !defined(__ATOMIC_OPERATIONS_H__)
#define __ATOMIC_OPERATIONS_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */


/**
 * NAME: ix_rm_atomic_sram_swap
 *
 * DESCRIPTION: This function will perform a sram memory atomic swap.The
 *  value referred by arg_pValue1 will be swapped with the value referred
 *  by arg_pValue2. On return *arg_pValue1 will be changed to previous *arg_pValue2
 *  and vice-versa. The operation is atomic only relative to the second argument 
 *  memory location and that only if that points to a SRAM memory location. So,
 *  in order for the operation to be atomic, arg_pValue2 must point to a SRAM location.
 * 
 * @Param:  - INOUT arg_pValue1 - the address of the first operand of the swap operation
 * @Param:  - INOUT arg_pValue2 - the address of the second operand of the swap operation.
 *      This is the operand relative to which the operation is atomic.
 */
IX_EXPORT_FUNCTION 
void ix_rm_atomic_sram_swap(
                             ix_uint32* arg_pValue1, 
                             ix_uint32* arg_pValue2 
                           );


/**
 * NAME: ix_rm_atomic_sram_add
 *
 * DESCRIPTION: This function will increment the value addressed by arg_pSramValue by arg_IncrementValue.
 *  The arg_pSramValue address should be a SRAM address, as it will not be checked.
 *  The value at arg_pSramValue saturates at 0. So if you add a negative number that is larger
 *  in absolute value than the value at arg_pSramValue then the result at the arg_pSramValue location
 *  will be 0
 * 
 * @Param:  - IN arg_IncrementValue - the value to be added to the memory location addressed by
 *  arg_pSramValue;
 * @Param:  - INOUT arg_pSramValue - the value at this location will be incremented by arg_IncrementValue
 *
 */
IX_EXPORT_FUNCTION 
void ix_rm_atomic_sram_add( 
                            ix_int32 arg_IncrementValue, 
                            ix_uint32* arg_pSramValue
                          );


/**
 * NAME: ix_rm_atomic_sram_test_and_add
 *
 * DESCRIPTION: This function will increment the value addressed by arg_pSramValue by arg_IncrementValue
 *  and will return the value at arg_pSramValue location before increment operation.
 *  The arg_pSramValue address should be a SRAM address, as it will not be checked.
 *  The value at arg_pSramValue saturates at 0. So if you add a negative number that is larger
 *  in absolute value than the value at arg_pSramValue then the result at the arg_pSramValue location
 *  will be 0
 * 
 * @Param:  - IN arg_IncrementValue - the value to be added to the memory location addressed by
 *  arg_pSramValue;
 * @Param:  - INOUT arg_pSramValue - the value at this location will be incremented by arg_IncrementValue
 *
 * @Return: Returns the value before the atomic add has been performed.
 */
IX_EXPORT_FUNCTION 
ix_uint32 ix_rm_atomic_sram_test_and_add( 
                                          ix_int32 arg_IncrementValue, 
                                          ix_uint32* arg_pSramValue
                                        );


/**
 * NAME: ix_rm_atomic_sram_subtract
 *
 * DESCRIPTION: This function will decrement the value addressed by arg_pSramValue by arg_DecrementValue.
 *  The arg_pSramValue address should be a SRAM address, as it will not be checked.
 *  The value at arg_pSramValue saturates at 0. So if you subtract a positive number that is larger
 *  in absolute value than the value at arg_pSramValue then the result at the arg_pSramValue location
 *  will be 0
 * 
 * @Param:  - IN arg_DecrementValue - the value to be subtracted from the memory location addressed by
 *  arg_pSramValue;
 * @Param:  - INOUT arg_pSramValue - the value at this location will be decremented by arg_DecrementValue
 *
 */
IX_EXPORT_FUNCTION 
void ix_rm_atomic_sram_subtract( 
                                 ix_int32 arg_DecrementValue, 
                                 ix_uint32* arg_pSramValue
                               );


/**
 * NAME: ix_rm_atomic_sram_test_and_subtract
 *
 * DESCRIPTION: This function will decrement the value addressed by arg_pSramValue by arg_DecrementValue
 *  and will return the value at arg_pSramValue location before decrement operation.
 *  The arg_pSramValue address should be a SRAM address, as it will not be checked.
 *  The value at arg_pSramValue saturates at 0. So if you subtract a positive number that is larger
 *  in absolute value than the value at arg_pSramValue then the result at the arg_pSramValue location
 *  will be 0
 * 
 * @Param:  - IN arg_DecrementValue - the value to be subtracted from the memory location addressed by
 *  arg_pSramValue;
 * @Param:  - INOUT arg_pSramValue - the value at this location will be decremented by arg_DecrementValue
 *
 * @Return: Returns the value before the atomic subtract has been performed.
 */
IX_EXPORT_FUNCTION 
ix_uint32 ix_rm_atomic_sram_test_and_subtract( 
                                               ix_int32 arg_DecrementValue, 
                                               ix_uint32* arg_pSramValue
                                             );


/**
 * NAME: ix_rm_atomic_sram_bit_set
 *
 * DESCRIPTION: This function will atomically set the bits of the value addressed 
 *  by arg_pSramValue corresponding to the bits set (1) in the arg_SetValue argument.
 *  The arg_pSramValue address must be a SRAM address, as it will not be checked.
 * 
 * @Param:  - IN arg_SetValue - the value specifying which bits of the memory location 
 *  addressed by arg_pSramValue should be set
 * @Param:  - INOUT arg_pSramValue - the value at this location will have the bits
 *  specified by arg_SetValue argument set upon return
 *
 */
IX_EXPORT_FUNCTION 
void ix_rm_atomic_sram_bit_set( 
                                ix_uint32 arg_SetValue, 
                                ix_uint32* arg_pSramValue
                              );


/**
 * NAME: ix_rm_atomic_sram_bit_test_and_set
 *
 * DESCRIPTION: This function will atomically set the bits of the value addressed 
 *  by arg_pSramValue corresponding to the bits set (1) in the arg_SetValue argument.
 *  The arg_pSramValue address must be a SRAM address, as it will not be checked.
 *  Upon return, the function will return the previous value at the arg_pSramValue
 *  location
 * 
 * @Param:  - IN arg_SetValue - the value specifying which bits of the memory location 
 *  addressed by arg_pSramValue should be set
 * @Param:  - INOUT arg_pSramValue - the value at this location will have the bits
 *  specified by arg_SetValue argument set upon return
 *
 * @Return: Returns the value at the arg_pSramValue location before the atomic set 
 *  has been performed.
 */
IX_EXPORT_FUNCTION 
ix_uint32 ix_rm_atomic_sram_bit_test_and_set( 
                                              ix_uint32 arg_SetValue, 
                                              ix_uint32* arg_pSramValue
                                            );


/**
 * NAME: ix_rm_atomic_sram_bit_clear
 *
 * DESCRIPTION: This function will atomically clear the bits of the value addressed 
 *  by arg_pSramValue corresponding to the bits set (1) in the arg_ClearValue argument.
 *  The arg_pSramValue address must be a SRAM address, as it will not be checked.
 * 
 * @Param:  - IN arg_ClearValue - the value specifying which bits of the memory location 
 *  addressed by arg_pSramValue should be cleared
 * @Param:  - INOUT arg_pSramValue - the value at this location will have the bits
 *  specified by arg_ClearValue argument cleared upon return
 *
 */
IX_EXPORT_FUNCTION 
void ix_rm_atomic_sram_bit_clear( 
                                  ix_uint32 arg_ClearValue, 
                                  ix_uint32* arg_pSramValue
                                );


/**
 * NAME: ix_rm_atomic_sram_bit_test_and_clear
 *
 * DESCRIPTION: This function will atomically clear the bits of the value addressed 
 *  by arg_pSramValue corresponding to the bits set (1) in the arg_ClearValue argument.
 *  The arg_pSramValue address must be a SRAM address, as it will not be checked.
 *  Upon return, the function will return the previous value at the arg_pSramValue
 *  location
 * 
 * @Param:  - IN arg_ClearValue - the value specifying which bits of the memory location 
 *  addressed by arg_pSramValue should be cleared
 * @Param:  - INOUT arg_pSramValue - the value at this location will have the bits
 *  specified by arg_ClearValue argument cleared upon return
 *
 * @Return: Returns the value at the arg_pSramValue location before the atomic clear 
 *  has been performed.
 */
IX_EXPORT_FUNCTION 
ix_uint32 ix_rm_atomic_sram_bit_test_and_clear( 
                                                ix_uint32 arg_ClearValue, 
                                                ix_uint32* arg_pSramValue
                                              );


/**
 * NAME: ix_rm_atomic_scratch_swap
 *
 * DESCRIPTION: This function will perform a scratch memory atomic swap.The
 *  value referred by arg_pValue1 will be swapped with the value referred
 *  by arg_pValue2. On return *arg_pValue1 will be changed to previous *arg_pValue2
 *  and vice-versa. The operation is atomic only relative to the second argument 
 *  memory location and that only if that points to a SCRATCH memory location. So,
 *  in order for the operation to be atomic, arg_pValue2 must point to a SCRATCH location.
 * 
 * @Param:  - INOUT arg_pValue1 - the address of the first operand of the swap operation
 * @Param:  - INOUT arg_pValue2 - the address of the second operand of the swap operation
 *      This is the operand relative to which the operation is atomic.
 */
IX_EXPORT_FUNCTION 
void ix_rm_atomic_scratch_swap(
                                ix_uint32* arg_pValue1, 
                                ix_uint32* arg_pValue2 
                              );




/**
 * NAME: ix_rm_atomic_scratch_add
 *
 * DESCRIPTION: This function will increment the value addressed by arg_pScratchValue by arg_IncrementValue.
 *  The arg_pScratchValue address should be a SCRATCH address, as it will not be checked.
 * 
 * @Param:  - IN arg_IncrementValue - the value to be added to the memory location addressed by
 *  arg_pScratchValue;
 * @Param:  - INOUT arg_pScratchValue - the value at this location will be incremented by arg_IncrementValue
 *
 */
IX_EXPORT_FUNCTION 
void ix_rm_atomic_scratch_add( 
                               ix_uint32 arg_IncrementValue, 
                               ix_uint32* arg_pScratchValue
                             );


/**
 * NAME: ix_rm_atomic_scratch_test_and_add
 *
 * DESCRIPTION: This function will increment the value addressed by arg_pScratchValue by arg_IncrementValue
 *  and will return the value at arg_pScratchValue location before increment operation.
 *  The arg_pScratchValue address should be a SCRATCH address, as it will not be checked.
 * 
 * @Param:  - IN arg_IncrementValue - the value to be added to the memory location addressed by
 *  arg_pScratchValue;
 * @Param:  - INOUT arg_pScratchValue - the value at this location will be incremented by arg_IncrementValue
 *
 * @Return: Returns the value before the atomic add has been performed.
 */
IX_EXPORT_FUNCTION 
ix_uint32 ix_rm_atomic_scratch_test_and_add( 
                                             ix_uint32 arg_IncrementValue, 
                                             ix_uint32* arg_pScratchValue
                                           );


/**
 * NAME: ix_rm_atomic_scratch_subtract
 *
 * DESCRIPTION: This function will decrement the value addressed by arg_pScratchValue by arg_DecrementValue.
 *  The arg_pScratchValue address should be a SCRATCH address, as it will not be checked.
 *  The value at arg_pScratchValue saturates at 0. So if you subtract a number that is larger
 *  than the value at arg_pScratchValue then the result at the arg_pScratchValue location
 *  will be 0
 * 
 * @Param:  - IN arg_DecrementValue - the value to be subtracted from the memory location addressed by
 *  arg_pScratchValue;
 * @Param:  - INOUT arg_pScratchValue - the value at this location will be decremented by arg_DecrementValue
 *
 */
IX_EXPORT_FUNCTION 
void ix_rm_atomic_scratch_subtract( 
                                    ix_uint32 arg_DecrementValue, 
                                    ix_uint32* arg_pScratchValue
                                  );


/**
 * NAME: ix_rm_atomic_scratch_test_and_subtract
 *
 * DESCRIPTION: This function will decrement the value addressed by arg_pScratchValue by arg_DecrementValue
 *  and will return the value at arg_pScratchValue location before decrement operation.
 *  The arg_pScratchValue address should be a SCRATCH address, as it will not be checked.
 *  The value at arg_pScratchValue saturates at 0. So if you subtract a number that is larger
 *  than the value at arg_pScratchValue then the result at the arg_pScratchValue location
 *  will be 0
 * 
 * @Param:  - IN arg_DecrementValue - the value to be subtracted from the memory location addressed by
 *  arg_pScratchValue;
 * @Param:  - INOUT arg_pScratchValue - the value at this location will be decremented by arg_DecrementValue
 *
 * @Return: Returns the value before the atomic subtract has been performed.
 */
IX_EXPORT_FUNCTION 
ix_uint32 ix_rm_atomic_scratch_test_and_subtract( 
                                                  ix_uint32 arg_DecrementValue, 
                                                  ix_uint32* arg_pScratchValue
                                                );


/**
 * NAME: ix_rm_atomic_scratch_bit_set
 *
 * DESCRIPTION: This function will atomically set the bits of the value addressed 
 *  by arg_pScratchValue corresponding to the bits set (1) in the arg_SetValue argument.
 *  The arg_pScratchValue address must be a SCRATCH address, as it will not be checked.
 * 
 * @Param:  - IN arg_SetValue - the value specifying which bits of the memory location 
 *  addressed by arg_pScratchValue should be set
 * @Param:  - INOUT arg_pScratchValue - the value at this location will have the bits
 *  specified by arg_SetValue argument set upon return
 *
 */
IX_EXPORT_FUNCTION 
void ix_rm_atomic_scratch_bit_set( 
                                   ix_uint32 arg_SetValue, 
                                   ix_uint32* arg_pScratchValue
                                 );


/**
 * NAME: ix_rm_atomic_scratch_bit_test_and_set
 *
 * DESCRIPTION: This function will atomically set the bits of the value addressed 
 *  by arg_pScratchValue corresponding to the bits set (1) in the arg_SetValue argument.
 *  The arg_pScratchValue address must be a SCRATCH address, as it will not be checked.
 *  Upon return, the function will return the previous value at the arg_pScratchValue
 *  location
 * 
 * @Param:  - IN arg_SetValue - the value specifying which bits of the memory location 
 *  addressed by arg_pScratchValue should be set
 * @Param:  - INOUT arg_pScratchValue - the value at this location will have the bits
 *  specified by arg_SetValue argument set upon return
 *
 * @Return: Returns the value at the arg_pScratchValue location before the atomic set 
 *  has been performed.
 */
IX_EXPORT_FUNCTION 
ix_uint32 ix_rm_atomic_scratch_bit_test_and_set( 
                                                 ix_uint32 arg_SetValue, 
                                                 ix_uint32* arg_pScratchValue
                                               );


/**
 * NAME: ix_rm_atomic_scratch_bit_clear
 *
 * DESCRIPTION: This function will atomically clear the bits of the value addressed 
 *  by arg_pScratchValue corresponding to the bits set (1) in the arg_ClearValue argument.
 *  The arg_pScratchValue address must be a SCRATCH address, as it will not be checked.
 * 
 * @Param:  - IN arg_ClearValue - the value specifying which bits of the memory location 
 *  addressed by arg_pScratchValue should be cleared
 * @Param:  - INOUT arg_pScratchValue - the value at this location will have the bits
 *  specified by arg_ClearValue argument cleared upon return
 *
 */
IX_EXPORT_FUNCTION 
void ix_rm_atomic_scratch_bit_clear( 
                                     ix_uint32 arg_ClearValue, 
                                     ix_uint32* arg_pScratchValue
                                   );


/**
 * NAME: ix_rm_atomic_scratch_bit_test_and_clear
 *
 * DESCRIPTION: This function will atomically clear the bits of the value addressed 
 *  by arg_pScratchValue corresponding to the bits set (1) in the arg_ClearValue argument.
 *  The arg_pScratchValue address must be a SCRATCH address, as it will not be checked.
 *  Upon return, the function will return the previous value at the arg_pScratchValue
 *  location
 * 
 * @Param:  - IN arg_ClearValue - the value specifying which bits of the memory location 
 *  addressed by arg_pScratchValue should be cleared
 * @Param:  - INOUT arg_pScratchValue - the value at this location will have the bits
 *  specified by arg_ClearValue argument cleared upon return
 *
 * @Return: Returns the value at the arg_pScratchValue location before the atomic clear 
 *  has been performed.
 */
IX_EXPORT_FUNCTION 
ix_uint32 ix_rm_atomic_scratch_bit_test_and_clear( 
                                                   ix_uint32 arg_ClearValue, 
                                                   ix_uint32* arg_pScratchValue
                                                 );


#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__ATOMIC_OPERATIONS_H__) */
