/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      buffer.h
 *
 * = DESCRIPTION
 *      This file will define all exported data types used in the Buffer API module. Also
 *      will declare all function prototypes of the exported interface.
 *      The Buffer API has two parts: 
 *          - an API that is used to create buffer free lists of varying element sizes. The buffer
 *      API supports both hardware and software buffers defined that way based on if the managing
 *      of the buffers has or not hardware support. The software and hardware buffers are both composed
 *      of meta and payload data but they are different in the way the information in buffer handles is
 *      packed, and in how the meta data is laid out.
 *      This API is quite generic and independent of packet descriptor layout defined by the microblock
 *      infrastructure or by the sowftware.
 *          - an API that is used to access fields in the packet descriptor (meta data).
 *      This API is specific to layout and fields of packet descriptor.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      4/23/2002 10:10:19 AM - creation time 
 * ============================================================================
 * $Id: buffer.h,v 1.29 2003/09/19 21:10:54 ageorges Exp $
 */

#if !defined(__BUFFER_H__)
#define __BUFFER_H__

#if defined(_IX_RM_IMPL_HARDWARE_) 

    #if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)


        #if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

            #include <vxworks/ixp2400/ixdp2400.h>

        #else

            #include <vxworks/ixp2800/ixdp2000.h>

        #endif

    #endif /* end (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) */

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */



/**
 * Generic Buffer API
 */

/**
 * TYPENAME: ix_buffer_handle
 * 
 * DESCRIPTION: Generic type for buffer handle which could be a hardware or a software 
 *          buffer handle.
 *
 */
typedef ix_handle ix_buffer_handle;


/**
 * The following is the hardware buffer handle mapping.
 * This is according to Microace Building Blocks for Sausalito Design Document.
 *      LW	Bits 	Size	Field	    Description
 *      -----------------------------------------------------------------
 *      0	31:31	1	    EOP flag	This buffer includes EOP
 *          30:30	1	    SOP flag	This buffer includes SOP
 *	        29:24	6	    Cell Count	Cell Count for buffer
 *	        23:0	24	    Address 	Address in SRAM of buffer descriptor expressed
 *                                      as a LongWord index offset into SRAM channel 0
 *
 */

/**
 * The following is the current software buffer handle mapping.
 *      LW	Bits 	Size	Field	    Description
 *      -----------------------------------------------------------------
 *      0	31:31	1	    EOP flag	This buffer includes EOP
 *          30:30	1	    SOP flag	This buffer includes SOP
 *	        29:22	8   	FL Index	Free list index that owns this buffer
 *	        21:1	21	    Buf Index 	Buffer index inside specified FL
 *          0:0     1       Buf Type    This bit is 1 at all times to differentiate it
 *                                      from hardware buffers whose handle will have last
 *                                      bit set to 0 at all times (8 byte meta alignment)
 *
 */


/**
 * DESCRIPTION: This symbol defines a null buffer handle
 *
 */
#define IX_NULL_BUFFER_HANDLE   ((ix_buffer_handle)0)



/**
 * TYPENAME: ix_buffer_type
 * 
 * DESCRIPTION: This enumerated type differentiate between software and hardware
 *          buffers. The hardware ones have hardware support for managing them
 *          and can be used by both uCode and core API. The software buffers
 *          will be managed totaly in software and must be used just by the 
 *          core apps.
 */
typedef enum ix_e_buffer_type
{
    IX_BUFFER_TYPE_FIRST = 0,
    IX_BUFFER_TYPE_HARDWARE = IX_BUFFER_TYPE_FIRST,
    IX_BUFFER_TYPE_SOFTWARE,
    IX_BUFFER_TYPE_LAST
} ix_buffer_type;



/**
 * There will be a maximum of IX_BUF_MAX_HW_FL_NUMBER hardware FLs corresponding to the maximum number
 * of free lists required by the uCode. We start with a miximum IX_BUF_MAX_FL_NUMBER buffer free lists 
 * in the system out which IX_BUF_MAX_HW_FL_NUMBER are hardware supported in the case we support just
 * one SRAM channel and IX_BUF_MAX_HW_FL_NUMBER * IX_SRAM_CHANNELS_NUMBER in the case we support multiple
 * channels for HW free lists.
 */

/**
 * DESCRIPTION: This symbol defines the maximum number of bits required to 
 *          represent a hardware FL ID. In the case of SRAM multichannel HW
 *          free list support, that corresponds to only one channel.
 *
 */
#define IX_BUF_HW_FL_ID_BITS         4U

 
/**
 * DESCRIPTION: This symbol defines the maximum number of bits required to 
 *          represent a FL ID.
 *
 */
#define IX_BUF_FL_ID_BITS            8U


#if (29U <= IX_BUF_FL_ID_BITS)

#error "The buffer index and FL index don't fit into software handle"

#else

    /**
     * DESCRIPTION: This symbol defines the maximum number of bits required to 
     *          represent a FL ID.
     *
     */
    #define IX_BUF_HANDLE_SW_INDEX_BITS (29U - IX_BUF_FL_ID_BITS)


#endif /* (29 >= IX_BUF_FL_ID_BITS) */



/**
 * DESCRIPTION: This symbol defines the maximum number of hardware FL on this system.
 *  In the case of SRAM multichannel HW free list support, the number will represent
 *  the number of HW free lists per channel
 *
 */
#define IX_BUF_MAX_HW_FL_NUMBER      (1U << IX_BUF_HW_FL_ID_BITS)


/**
 * DESCRIPTION: This symbol defines the maximum number of FL on this system.
 *
 */
#define IX_BUF_MAX_FL_NUMBER         (1U << IX_BUF_FL_ID_BITS)


/**
 * DESCRIPTION: This symbol defines the maximum number of software FL on this system.
 *
 */

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

#define IX_BUF_MAX_SW_FL_NUMBER      (IX_BUF_MAX_FL_NUMBER - IX_BUF_MAX_HW_FL_NUMBER * IX_SRAM_CHANNELS_NUMBER)

#else

#define IX_BUF_MAX_SW_FL_NUMBER      (IX_BUF_MAX_FL_NUMBER - IX_BUF_MAX_HW_FL_NUMBER)

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */



/**
 * DESCRIPTION: This symbol defines the first software FL ID on this system.
 *
 */
#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

#define IX_BUF_FIRST_SW_FL_ID        (IX_BUF_MAX_HW_FL_NUMBER * IX_SRAM_CHANNELS_NUMBER)

#else

#define IX_BUF_FIRST_SW_FL_ID        (IX_BUF_MAX_HW_FL_NUMBER)

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */




/**
 * TYPENAME: ix_buffer_free_list_handle
 * 
 * DESCRIPTION: Generic type for a buffer free list handle.
 *
 */
typedef ix_handle ix_buffer_free_list_handle;


/**
 * TYPENAME: ix_sw_buffer_meta
 * 
 * DESCRIPTION: This structure defines the internal meta data common to all buffer objects.
 *          This hidden structure provides the means to link, test and navigate lists of buffers.
 *
 */
typedef struct ix_s_sw_buffer_meta
{
    ix_buffer_handle    m_hNext; /* This is the link to the next buffer */
    ix_buffer_handle    m_hSelf; /* This is a reference to buffer handle */
    ix_uint32*          m_pData; /* This is a pointer to the data area of the buffer */
    ix_uint32           m_DataSize; /* This represents the active data size in the buffer expressed in bytes */
} ix_sw_buffer_meta;

/**
 * DESCRIPTION: This symbol will declare all the fields and the layout of the common
 *          part of the buffer meta data. Applications that need extra fields can do
 *          so in the following manner without the expense of an extra indirection.
 *          typedef struct ix_s_atm_buffer_meta
 *          {
 *              IX_DECLARE_HW_BUFFER_META_DATA
 *              ix_uint32       m_CellHeader;
 *          } ix_atm_buffer_meta;
 *
 */

#if defined(_IX_RM_SPLIT_META_DATA_)

#define IX_DECLARE_HW_BUFFER_META_DATA \
        /* long word 0 */ \
        ix_uint32       m_Reserved0;  /* reserved */ \
        /* long word 1 */ \
        ix_uint16       m_BufferSize; /* size of the current buffer in bytes */ \
        ix_uint16       m_Offset; /* the offset of start data in the buffer in bytes */ \
        /* long word 2 */ \
        ix_uint16       m_PacketSize; /* size of the entire packet in bytes */ \
         \
         \
        /** \
         * The following represents packed buffer information. The provided macros \
         * should be used to extract the particular fields. The layout is as follows: \
         * +--------------+-------------+---------------+ \
         * | Free List ID |    RxStat   |  Header type  | \
         * | 15:12 4 bit  | 11:8 4 bit  |    7:0 8 bit  | \
         * +--------------+-------------+---------------+ \
         *  15:12   4 bits      free_list_id    Free list ID for the buffer
         *  11:8    4 bits      rx_stat         Receive status flag
         *   7:0    8 bits      header_type     Type of header at "offset" bytes into the packet
         * We don't use bit fields to avoid mapping complications \
         */ \
         \
        ix_uint16       m_BufferInfo; /* packed buffer information */ \
        /* long word 3 */ \
        ix_uint16       m_InputPort; /* input port on ingress processor */ \
        ix_uint16       m_OutputPort; /* output port on egress processor */ \
        /* long word 4 */ \
        ix_uint16       m_NextHopID; /* next hop IP ID */ \
        ix_uint8        m_FabricPort; /* Output port for fabric indicating destination blade */ \
        \
        ix_uint8        m_Reserved1     : 4; /* reserved */ \
        ix_uint8        m_NextHopIDType : 4; \
        \
        /* long word 5 */ \
        ix_uint32       m_ColorID       : 8; \
        ix_uint32       m_FlowID        : 24; /* flow identifier (QoS flow id or MPLS label/flow id) */ \
        \
        /* long word 6 */ \
        ix_uint16       m_ClassID; /* class id */ \
        ix_uint16       m_Reserved2; /* reserved */ \
        /* long word 7 */ \
        ix_uint32       m_PacketNext; /* pointer to next packet (unused in cell mode) */ \

#else

#define IX_DECLARE_HW_BUFFER_META_DATA \
        /* long word 0 */ \
        ix_uint32       m_HwNext; /* reserved hardware link to the next data */ \
        /* long word 1 */ \
        ix_uint16       m_BufferSize; /* size of the current buffer in bytes */ \
        ix_uint16       m_Offset; /* the offset of start data in the buffer in bytes */ \
        /* long word 2 */ \
        ix_uint16       m_PacketSize; /* size of the entire packet in bytes */ \
         \
         \
        /** \
         * The following represents packed buffer information. The provided macros \
         * should be used to extract the particular fields. The layout is as follows: \
         * +--------------+-------------+---------------+ \
         * | Free List ID |    RxStat   |  Header type  | \
         * | 15:12 4 bit  | 11:8 4 bit  |    7:0 8 bit  | \
         * +--------------+-------------+---------------+ \
         *  15:12   4 bits      free_list_id    Free list ID for the buffer
         *  11:8    4 bits      rx_stat         Receive status flag
         *   7:0    8 bits      header_type     Type of header at "offset" bytes into the packet
         * We don't use bit fields to avoid mapping complications \
         */ \
         \
        ix_uint16       m_BufferInfo; /* packed buffer information */ \
        /* long word 3 */ \
        ix_uint16       m_InputPort; /* input port on ingress processor */ \
        ix_uint16       m_OutputPort; /* output port on egress processor */ \
        /* long word 4 */ \
        ix_uint16       m_NextHopID; /* next hop IP ID */ \
        ix_uint8        m_FabricPort; /* Output port for fabric indicating destination blade */ \
        \
        ix_uint8        m_Reserved1     : 4; /* reserved */ \
        ix_uint8        m_NextHopIDType : 4; \
        \
        /* long word 5 */ \
        ix_uint32       m_ColorID       : 8; \
        ix_uint32       m_FlowID        : 24; /* flow identifier (QoS flow id or MPLS label/flow id) */ \
        \
        /* long word 6 */ \
        ix_uint16       m_ClassID; /* class id */ \
        ix_uint16       m_Reserved2; /* reserved */ \
        /* long word 7 */ \
        ix_uint32       m_PacketNext; /* pointer to next packet (unused in cell mode) */ \


#endif /* defined(_IX_RM_SPLIT_META_DATA_) */


/**
 * TYPENAME: ix_hw_buffer_meta
 * 
 * DESCRIPTION: This structure defines the information and the layout of the common
 *          buffer meta data. If applications will need to add extra fields then they
 *          can do so.
 *
 */
typedef struct ix_s_hw_buffer_meta
{
    IX_DECLARE_HW_BUFFER_META_DATA
} ix_hw_buffer_meta;



#if defined(_IX_RM_SPLIT_META_DATA_)


/**
 * TYPENAME: ix_hw_internal_buffer_meta
 * 
 * DESCRIPTION: This is the internal HW meta data in case of split meta data. This
 *  portion should be accessible just by RM, as the other apps should not care about
 *  m_HwNext field.
 *
 */
typedef struct ix_s_hw_internal_buffer_meta
{
    /* long word 0 */ 
    ix_uint32       m_HwNext; /* reserved hardware link to the next data */ 

    /* long word 1 */ 
    ix_hw_buffer_meta*  m_pMeta; /* address of the proper meta data */ 
} ix_hw_internal_buffer_meta;


#endif /* defined(_IX_RM_SPLIT_META_DATA_) */



/**
 * MACRO NAME: IX_RM_HW_META_GET_HEADER_TYPE
 *
 * DESCRIPTION: This macro retrieves the header type information from the packed m_BufferInfo 
 *          field of the ix_hw_buffer_meta structure.
 *
 * @Param:  - IN arg_BufferInfo - ix_uint16 value of the m_BufferInfo field of the 
 *          ix_hw_buffer_meta structure.
 *
 * @Return: ix_uint32 value representing the header type in the buffer referred by
 *          ix_hw_buffer_meta structure.
 */
#define IX_RM_HW_META_GET_HEADER_TYPE( \
                                       arg_BufferInfo \
                                     ) \
                                     ((ix_uint32)IX_GET_BIT_FIELD16(arg_BufferInfo, 0U, 7U))
                             

/**
 * MACRO NAME: IX_RM_HW_META_SET_HEADER_TYPE
 *
 * DESCRIPTION: This macro sets the header type information into the packed m_BufferInfo 
 *          field of the ix_hw_buffer_meta structure.
 *
 * @Param:  - INOUT arg_BufferInfo - ix_uint16 value of the m_BufferInfo field of the 
 *          ix_hw_buffer_meta structure. On return this field will have the header type
 *          part adjusted to the new value.
 * @Param:  - IN arg_HeaderType - ix_uint16 or ix_uint32 value representing the new header type.
 *
 * @Return: ix_uint16 value representing the updated m_BufferInfo field of the ix_hw_buffer_meta 
 *          structure.
 */
#define IX_RM_HW_META_SET_HEADER_TYPE( \
                                       arg_BufferInfo, \
                                       arg_HeaderType \
                                     ) \
                                     IX_SET_BIT_FIELD16(arg_BufferInfo, arg_HeaderType, 0U, 7U)
                             


/**
 * MACRO NAME: IX_RM_HW_META_GET_FREE_LIST_ID
 *
 * DESCRIPTION: This macro retrieves the free list id from the packed m_BufferInfo 
 *          field of the ix_hw_buffer_meta structure.
 *
 * @Param:  - IN arg_BufferInfo - ix_uint16 value of the m_BufferInfo field of the 
 *          ix_hw_buffer_meta structure.
 *
 * @Return: ix_uint32 value representing the free list id of the buffer referred by
 *          ix_hw_buffer_meta structure.
 */
#define IX_RM_HW_META_GET_FREE_LIST_ID( \
                                        arg_BufferInfo \
                                      ) \
                                      ((ix_uint32)IX_GET_BIT_FIELD16(arg_BufferInfo, 12U, 11U + IX_BUF_HW_FL_ID_BITS))
                             

/**
 * MACRO NAME: IX_RM_HW_META_SET_FREE_LIST_ID
 *
 * DESCRIPTION: This macro sets the free list id into the packed m_BufferInfo 
 *          field of the ix_hw_buffer_meta structure.
 *
 * @Param:  - INOUT arg_BufferInfo - ix_uint16 value of the m_BufferInfo field of the 
 *          ix_hw_buffer_meta structure. On return this field will have the header type
 *          part adjusted to the new value.
 * @Param:  - IN arg_FreeListID - ix_uint32 value representing the free list ID who owns the buffer.
 *
 * @Return: ix_uint16 value representing the updated m_BufferInfo field of the ix_hw_buffer_meta 
 *          structure.
 */
#define IX_RM_HW_META_SET_FREE_LIST_ID( \
                                        arg_BufferInfo, \
                                        arg_FreeListID \
                                      ) \
                                      IX_SET_BIT_FIELD16(arg_BufferInfo, arg_FreeListID, 12U, 11U + IX_BUF_HW_FL_ID_BITS)



/**
 * MACRO NAME: IX_RM_HW_META_GET_RX_STAT
 *
 * DESCRIPTION: This macro retrieves the receive status information from the packed m_BufferInfo 
 *          field of the ix_hw_buffer_meta structure.
 *
 * @Param:  - IN arg_BufferInfo - ix_uint16 value of the m_BufferInfo field of the 
 *          ix_hw_buffer_meta structure.
 *
 * @Return: ix_uint32 value representing the receive status info of the buffer referred by
 *          ix_hw_buffer_meta structure.
 */
#define IX_RM_HW_META_GET_RX_STAT( \
                                   arg_BufferInfo \
                                 ) \
                                 ((ix_uint32)IX_GET_BIT_FIELD16(arg_BufferInfo, 8U, 11U))
                             

/**
 * MACRO NAME: IX_RM_HW_META_SET_RX_STAT
 *
 * DESCRIPTION: This macro sets the receive status information into the packed m_BufferInfo 
 *          field of the ix_hw_buffer_meta structure.
 *
 * @Param:  - INOUT arg_BufferInfo - ix_uint16 value of the m_BufferInfo field of the 
 *          ix_hw_buffer_meta structure. On return this field will have the receive
 *          status part adjusted to the new value.
 * @Param:  - IN arg_RxStat - ix_uint32 value representing the new receive status.
 *
 * @Return: ix_uint16 value representing the updated m_BufferInfo field of the ix_hw_buffer_meta 
 *          structure.
 */
#define IX_RM_HW_META_SET_RX_STAT( \
                                   arg_BufferInfo, \
                                   arg_RxStat \
                                 ) \
                                 IX_SET_BIT_FIELD16(arg_BufferInfo, arg_RxStat, 8U, 11U)



/**
 * NAME: ix_rm_hw_buffer_free_list_create
 *
 * DESCRIPTION: This function will create a hardware buffer free list with arg_ElementsNumber entries. 
 *          Memory will be allocated in SRAM and DRAM. A handle to the new created free list
 *          is returned. The size of SRAM required for the meta should be at least the size imposed
 *          by the hardware which at this moment is 4 bytes. The sizes will be rounded to 
 *          the next power of 2 always, so it is better that preferred sizes are passed in.
 *          Maximum IX_BUF_MAX_HW_FL_NUMBER hardware free lists can be created.
 * 
 * @Param:  - IN arg_ElementsNumber - the number of entries to be allocated for the free list.
 * @Param:  - IN arg_SRAMSize - the size of SRAM to be allocated for each entry.
 * @Param:  - IN arg_DRAMSize - the size of DRAM to be allocated for each entry.
 * @Param:  - IN arg_SRAMChannel - the SRAM channel where the memory should be allocated from.  
 * @Param:  - IN arg_DRAMChannel - the DRAM channel where the memory should be allocated from.  
 * @Param:  - OUT arg_pFreeListHandle - location where the newly created free list handle should
 *          be stored.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_hw_buffer_free_list_create( 
                                           ix_uint32 arg_ElementsNumber, 
                                           ix_uint32 arg_SRAMSize, 
                                           ix_uint32 arg_DRAMSize, 
                                           ix_uint32 arg_SRAMChannel, 
                                           ix_uint32 arg_DRAMChannel, 
                                           ix_buffer_free_list_handle* arg_pFreeListHandle 
                                         );


/**
 * NAME: ix_rm_sw_buffer_free_list_create
 *
 * DESCRIPTION: This function will create a software buffer free list that will allocate and
 *          manage arg_NumElements elements. The caller has to specify the sizes of the meta
 *          and data parts of the buffer. The sizes will be rounded to the next power of
 *          2 always, so it is better that preferred sizes are passed in. The meta and data blocks
 *          can be allocated in a flexible way in different types of memory, but SRAM and especially
 *          SCRATCH memory should be used only if badly needed. The sizes of the meta and data can 
 *          not be 0 at the same time. Maximum IX_BUF_MAX_SW_FL_NUMBER software free lists can be 
 *          created.
 * 
 * @Param:  - IN arg_MetaElementSize - the size of a meta element. The size will be rounded
 *          to the next power of 2
 * @Param:  - IN arg_MetaMemoryType - the type of the meta memory. For now we will allow SRAM,
 *          DRAM and SCRATCH. SRAM and especially SCRATCH should be used with great care.
 * @Param:  - IN arg_DataElementSize - the size of a data element. The size will be rounded
 *          to the next power of 2
 * @Param:  - IN arg_DataMemoryType - the type of the data memory. For now we will allow SRAM,
 *          DRAM and SCRATCH. SRAM and especially SCRATCH should be used with great care. 
 * @Param:  - IN arg_ElementsNumber - the number of memory-buffer elements that this manager will 
 *          allocate and manage. 
 * @Param:  - OUT arg_pPoolManagerHandle - address where the handle of the new created pool
 *          manager will be stored upon return.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_sw_buffer_free_list_create(
                                           ix_uint32 arg_MetaElementSize, 
                                           ix_memory_type arg_MetaMemoryType,
                                           ix_uint32 arg_DataElementSize, 
                                           ix_memory_type arg_DataMemoryType, 
                                           ix_uint32 arg_ElementsNumber,
                                           ix_buffer_free_list_handle* arg_pFreeListHandle 
                                         );



/**
 * NAME: ix_rm_buffer_free_list_delete
 *
 * DESCRIPTION: This function will delete the specified buffer free list. The associated memory
 *          will be released.
 * 
 * @Param:  - IN arg_hFreeList - represents the handle of the free list that should 
 *          be deleted.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_buffer_free_list_delete( 
                                        ix_buffer_free_list_handle arg_hFreeList 
                                      );




/**
 * TYPENAME: ix_buffer_free_list_info
 * 
 * DESCRIPTION: This structure contains all the information associated with a buffer
 *          free list. Information about start addresses of the controlled meta and
 *          data memory and buffer element sizes for both meta and data will be included.
 *
 */
typedef struct ix_s_buffer_free_list_info
{
    /* specifies if the free list is a software(1) or a hardware(0) one */
    ix_uint32       m_FreeListType;
    
#if defined(_IX_RM_SPLIT_META_DATA_)

    ix_uint32*      m_pInternalMetaBaseAddress; /* specifies the start of the internal controlled meta memory */
    ix_memory_type  m_InternalMetaMemoryType; /* internal meta block memory type */
    ix_uint32       m_InternalMetaMemoryChannel; /* internal meta block memory channel */
    ix_uint32       m_InternalMetaElementSize; /* specifies the buffer internal meta element size */

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */


    ix_uint32*      m_pMetaBaseAddress; /* specifies the start of the controlled meta memory */
    ix_memory_type  m_MetaMemoryType; /* meta block memory type */
    ix_uint32       m_MetaMemoryChannel; /* meta block memory channel */
    ix_uint32       m_MetaElementSize; /* specifies the buffer meta element size */

    ix_uint32*      m_pDataBaseAddress; /* specifies the start of the controlled data memory */
    ix_memory_type  m_DataMemoryType; /* data block memory type */
    ix_uint32       m_DataMemoryChannel; /* meta block memory channel */
    ix_uint32       m_DataElementSize; /* specifies the buffer data element size  */

    ix_uint32       m_NumElements; /* specifies the number of buffer elements in the list */

    /**
     * Specifies the SRAM controller QArray index for the corresponding hardware free list
     * and the free list index for the software free list
     */
    ix_uint32       m_FreeListInfo;
    
    
    /**
     * Specifies the free list Id for hardware free lists
     */
    ix_uint32       m_FreeListInfo1;

} ix_buffer_free_list_info;




/**
 * NAME: ix_rm_buffer_free_list_get_info
 *
 * DESCRIPTION: This function will retrieve the buffer free list info. Once a buffer
 *          free list is created the caller has no control over where the memory required
 *          will be allocated. Later the caller might need the start addresses in 
 *          order to pass them to the microcode.
 * 
 * @Param:  - IN arg_hFreeList - specifies the free list handle for which we
 *          we want to retrieve the information.
 * @Param:  - OUT arg_pFreeListInfo - a pointer to a structure where the function
 *          will store the requested info on return.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_buffer_free_list_get_info( 
                                          ix_buffer_free_list_handle arg_hFreeList, 
                                          ix_buffer_free_list_info* arg_pFreeListInfo 
                                        );


/**
 * NAME: ix_rm_buffer_alloc
 *
 * DESCRIPTION: This function will allocate a buffer from the specified buffer free list.  
 * 
 * @Param:  - IN arg_hFreeList - handle of the free list from which the buffer will 
 *          be allocated. 
 * @Param:  - OUT arg_pBufferHandle - location where the new allocated buffer handle will 
 *          be stored. 
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_buffer_alloc( 
                             ix_buffer_free_list_handle arg_hFreeList, 
                             ix_buffer_handle* arg_pBufferHandle 
                           );


/**
 * NAME: ix_rm_buffer_free
 *
 * DESCRIPTION: This function will free the specified buffer returning it to the right 
 *  buffer free list. Be aware that if the buffer is the start of a buffer chain,
 *  then the rest of the chain will be not freed and the chain will be broken.
 * 
 * @Param:  - IN arg_hBuffer - handle of the buffer to be freed.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_buffer_free( 
                            ix_buffer_handle arg_hBuffer 
                          );


/**
 * NAME: ix_rm_buffer_free_chain
 *
 * DESCRIPTION: This function will free the specified buffer returning it to the right 
 *  buffer free list. If the buffer is the head of a buffer chain then the entire
 *  chain will be freed. If this buffer is in the middle of a chain then the end of the
 *  chain will be freed starting from this buffer. The chain has to be unlinked at this 
 *  buffer boundary in order to avoid chain inconsistency.
 * 
 * @Param:  - IN arg_hBuffer - handle of the buffer head of chain to be freed.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_buffer_free_chain( 
                                  ix_buffer_handle arg_hBuffer 
                                );


/**
 * NAME: ix_rm_buffer_get_meta
 *
 * DESCRIPTION: Returns the base of meta data portion for the specified buffer.  
 * 
 * @Param:  - IN arg_hBuffer - handle of the buffer for which the meta data is requested.
 * @Param:  - OUT arg_pMetaData - location where the meta data address should be written. 
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_buffer_get_meta( 
                                ix_buffer_handle arg_hBuffer, 
                                void** arg_pMetaData 
                              );


/**
 * NAME: ix_rm_buffer_get_data
 *
 * DESCRIPTION: Returns the base of data portion of the specified buffer.  
 * 
 * @Param:  - IN arg_hBuffer - handle of the buffer for which the data is requested.
 * @Param:  - OUT arg_pData - location where the data address should be written.  
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure. 
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_buffer_get_data( 
                                ix_buffer_handle arg_hBuffer, 
                                void** arg_pData 
                              );






/**
 * DESCRIPTION: This value is the start address of the SRAM channel 0. It is important to be
 *          used to retrieve hardware buffers info in a fast fashion.
 *
 */
#if defined(_IX_RM_IMPL_HARDWARE_)

#include <halMmap.h>

#define IX_SRAM_CH0_START_ADDRESS ((ix_uint32)Hal_sram_ch0_virtAddr)
#define IX_SRAM_CH1_START_ADDRESS ((ix_uint32)Hal_sram_ch1_virtAddr)

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

#define IX_SRAM_CH2_START_ADDRESS ((ix_uint32)Hal_sram_ch2_virtAddr)
#define IX_SRAM_CH3_START_ADDRESS ((ix_uint32)Hal_sram_ch3_virtAddr)

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */

#else

#define IX_SRAM_CH0_START_ADDRESS ix_rm_get_sram_channel_start_address(0)
#define IX_SRAM_CH1_START_ADDRESS ix_rm_get_sram_channel_start_address(1)

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

#define IX_SRAM_CH2_START_ADDRESS ix_rm_get_sram_channel_start_address(2)
#define IX_SRAM_CH3_START_ADDRESS ix_rm_get_sram_channel_start_address(3)

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */


/**
 * NAME: ix_rm_buffer_is_eop
 *
 * DESCRIPTION: This function will return non 0 if this buffer is the last in a list of concatenated memory buffers.
 *          This is the equivalent of the End Of Packet property of a buffer.
 * 
 * @Param:  - IN arg_hBuffer - handle to a buffer whose EOP property we are interested in
 * @Param:  - OUT arg_pIsEOP - an address where the EOP property of the buffer will be stored. 
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure. On return *arg_pIsEOP will be 0
 *          if the buffer is not the last in a chain or non 0 otherwise.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_buffer_is_eop( 
                              ix_buffer_handle arg_hBuffer,
                              ix_uint32* arg_pIsEOP
                            );


/**
 * NAME: ix_rm_buffer_is_sop
 *
 * DESCRIPTION: This function will return non 0 if this buffer is the first in a list of concatenated memory buffers.
 *          This is the equivalent of the Start Of Packet property of a buffer.
 * 
 * @Param:  - IN arg_hBuffer - handle to a buffer whose SOP property we are interested in.
 * @Param:  - OUT arg_pIsSOP - an address where the SOP property of the buffer will be stored. 
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure. On return *arg_pIsSOP will be 0
 *          if the buffer is not the first in a chain or non 0 otherwise.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_buffer_is_sop( 
                              ix_buffer_handle arg_hBuffer,
                              ix_uint32* arg_pIsSOP
                            );



/**
 * NAME: ix_rm_buffer_cell_count_set
 *
 * DESCRIPTION: This function will set the cell count field into the
 *  passed hardware buffer handle
 * 
 * @Param:  - INOUT arg_pBufferHandle - address of the buffer handle to be modified
 * @Param:  - IN arg_CellCount - the new cell count, which can have values 0 to 63
 *  for now. Larger values will be truncated to 6bit values.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_buffer_cell_count_set(
                                      ix_buffer_handle* arg_pBufferHandle,
                                      ix_uint32 arg_CellCount
                                    );


/**
 * NAME: ix_rm_buffer_cell_count_get
 *
 * DESCRIPTION: This function will retrieve the cell count field from the
 *  passed hardware buffer handle
 * 
 * @Param:  - IN arg_hBuffer - buffer handle of interest
 * @Param:  - OUT arg_pCellCount - the location where the cell count will be returned
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_buffer_cell_count_get(
                                      ix_buffer_handle arg_hBuffer,
                                      ix_uint32* arg_pCellCount
                                    );


/**
 * NAME: ix_rm_buffer_get_type
 *
 * DESCRIPTION: This function will retrieve the type of the buffer whose handle has been passed
 *          as the first parameter.
 * 
 * @Param:  - IN arg_hBuffer - handle of the buffer whose type we want to retrieve
 * @Param:  - OUT arg_pBufferType - address where the buffer type will be stored on return. 
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure. On return the *arg_pBufferType
 *          variable will have the type of the buffer.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_buffer_get_type(
                                ix_buffer_handle arg_hBuffer,
                                ix_buffer_type* arg_pBufferType
                              );





/**
 * NAME: ix_rm_buffer_get_next
 *
 * DESCRIPTION: This function will return the next link for this buffer. If there are no buffers linked
 *          to this one the arg_pNextBufferHandle location will be set to IX_NULL_BUFFER_HANDLE.
 * 
 * @Param:  - IN arg_hBuffer - handle of the buffer for which we need the next link.
 * @Param:  - OUT arg_pNextBufferHandle - the location where to store the next link for this buffer. 
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_buffer_get_next(
                                ix_buffer_handle arg_hBuffer, 
                                ix_buffer_handle* arg_pNextBufferHandle
                              );




/**
 * NAME: ix_rm_buffer_link
 *
 * DESCRIPTION: This function will link the buffer referred by arg_pNextBufferHandle to the buffer 
 *          referred by arg_pBufferHandle. The buffers will be in a linked list construct. If the 
 *          current buffer (arg_pBufferHandle) has already a link then the function will return an
 *          error. On return the handles of the two buffers will be modified to specify the new
 *          EOP/SOP state.
 * 
 * @Param:  - IN arg_pBufferHandle - the reference to the handle of the buffer where to perform the link
 * @Param:  - IN arg_pNextBufferHandle - the reference to the handle of the buffer to be linked.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure. If the 
 *          current buffer (arg_pBufferHandle) has already a link then the function will return an
 *          error.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_buffer_link(
                            ix_buffer_handle* arg_pBufferHandle, 
                            ix_buffer_handle* arg_pNextBufferHandle
                          );



/**
 * NAME: ix_rm_buffer_unlink
 *
 * DESCRIPTION: This function will unlink the buffer referred by arg_pBufferHandle. If the buffer has
 *          a link that will be set to IX_NULL_BUFFER_HANDLE and the unlinked buffer handle will be
 *          returned at the arg_pUnlinkedBufferHandle location. On return the handles of the two buffers will be modified to specify the new
 *          EOP/SOP state. In order this function to work *arg_pBufferHandle should have the SOP flag set
 *          which mean that we can unlink just at a beginning of a chain.
 * 
 * @Param:  - IN arg_pBufferHandle - the reference to the handle of the buffer that will go from a linked to an unliked state.
 * @Param:  - OUT arg_pUnlinkedBufferHandle - location where the handle of the unlinked buffer will be stored.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_buffer_unlink(
                              ix_buffer_handle* arg_pBufferHandle,
                              ix_buffer_handle* arg_pUnlinkedBufferHandle
                            );


#if defined( _IX_RM_COMPONENT_MULTI_THREADED_ )


/**
 * NAME: ix_rm_sw_free_list_init_sync
 *
 * DESCRIPTION: This function will initialize the synchronization for the specified
 *          buffer free list. A detailed explanation on how sychronization
 *          mechanism works is found in the L0 HLD/LLD and also in sync_mechanism.h
 *          file.
 * 
 * @Param:  - IN arg_hSwFreeList - this is the handle of the free list for which we
 *              want to set up the synchronization mechanism
 * @Param:  - IN arg_LockFunction - this is the lock function of the sync mechanism.
 * @Param:  - IN arg_UnlockFunction - this is the unlock function of the sync mechanism.
 * @Param:  - IN arg_pContext - context that will be passed to each invocation of lock
 *          and unlock function. It might point to a synchronization primitive to be
 *          used by the specified lock and unlock functions.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_sw_free_list_init_sync(
                                       ix_buffer_free_list_handle    arg_hSwFreeList, 
                                       ix_sa_lock_ptr                arg_LockFunction,
                                       ix_sa_unlock_ptr              arg_UnlockFunction,
                                       void*                         arg_pContext
                                     );


#endif /* defined( _IX_RM_COMPONENT_MULTI_THREADED_ ) */


/**
 * NAME: ix_rm_buffer_make_safe_single
 *
 * DESCRIPTION: This function will set the next link buffer for a single buffer (SOP|EOP flags set)
 *  to IX_NULL_BUFFER_HANDLE. If the buffer is not a single buffer(i.e. belongs to a chain), then an
 *  error will be returned. This function should be used in conjunction with ix_rm_buffer_make_safe_chained()
 *  to fix stale buffer handles.
 * 
 * @Param:  - INOUT  arg_hBuffer - buffer handle to the buffer to be changed. On return the meta
 *  data will be changed in such a way that the next link will be set to IX_NULL_BUFFER_HANDLE
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_buffer_make_safe_single(
                                        ix_buffer_handle arg_hBuffer
                                      );


/**
 * NAME: ix_rm_buffer_make_safe_chained
 *
 * DESCRIPTION: This function will reset the EOP flag for the passed buffer handle if the next
 *  buffer link is not IX_NULL_BUFFER_HANDLE. In the case the buffer handle does not have the SOP
 *  flag set, an error will be returned. This function should be used in conjunction with 
 *  ix_rm_buffer_make_safe_single() to fix stale buffer handles.
 * 
 * @Param:  - INOUT  arg_pBufferHandle - address to the buffer handle to be updated.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_buffer_make_safe_chained(
                                         ix_buffer_handle* arg_pBufferHandle
                                       );




#if defined (_IX_RM_BUFFER_DEBUG_)


/**
 * NAME: ix_rm_free_list_print_available_buffers
 *
 * DESCRIPTION: This function will print information about the available buffers in a
 *  certain free list
 * 
 * @Param:  - IN arg_FreeListType - describes the type of the free list: 0 for HW
 *  and 1 for SW
 * @Param:  - IN arg_FreeListNumber - describes the free list number we are interested
 *  in
 */
IX_EXPORT_FUNCTION 
void ix_rm_free_list_print_available_buffers(
                                              ix_uint32 arg_FreeListType, 
                                              ix_uint32 arg_FreeListNumber
                                            );


/**
 * NAME: ix_rm_free_list_print_buffers_info
 *
 * DESCRIPTION: This function will print information about the buffers belonging to a certain
 *  HW free list
 * 
 * @Param:  - IN arg_FreeListNumber - describes the free list number we are interested
 *  in
 */
IX_EXPORT_FUNCTION 
void ix_rm_free_list_print_buffers_info(
                                         ix_uint32 arg_FreeListNumber
                                       );



/**
 * NAME: ix_rm_free_list_print_info
 *
 * DESCRIPTION: This function will print information about a certain free list
 * 
 * @Param:  - IN arg_FreeListType - describes the type of the free list: 0 for HW
 *  and 1 for SW
 * @Param:  - IN arg_FreeListNumber - describes the free list number we are interested
 *  in
 */
IX_EXPORT_FUNCTION 
void ix_rm_free_list_print_info(
                                 ix_uint32 arg_FreeListType, 
                                 ix_uint32 arg_FreeListNumber
                               );


/**
 * NAME: ix_rm_buffer_print_meta
 *
 * DESCRIPTION: This function will print the meta data information for the passed buffer
 * 
 * @Param:  - IN arg_hBuffer - buffer whose meta data we would like to print
 *
 */
IX_EXPORT_FUNCTION 
void ix_rm_buffer_print_meta(
                              ix_buffer_handle arg_hBuffer
                            );


/**
 * NAME: ix_rm_buffer_print_data
 *
 * DESCRIPTION: This function will print the data for the passed buffer
 * 
 * @Param:  - IN arg_hBuffer - buffer whose meta data we would like to print
 *
 */
IX_EXPORT_FUNCTION 
void ix_rm_buffer_print_data(
                              ix_buffer_handle arg_hBuffer
                            );



/**
 * NAME: ix_rm_buffer_print_debug_info
 *
 * DESCRIPTION: This function will print the debug info for the passed buffer
 * 
 * @Param:  - IN arg_hBuffer - buffer whose debug info we would like to print
 *
 */
IX_EXPORT_FUNCTION 
void ix_rm_buffer_print_debug_info(
                                    ix_buffer_handle arg_hBuffer
                                  );

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */


#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__BUFFER_H__) */
