/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 4
 *
 * = LIBRARY
 *      Resource Manager Library
 *
 * = MODULE
 *      Communication API
 *
 * = FILENAME
 *      communication.h
 *
 * = DESCRIPTION
 *      This file will define all exported data types used in the Communication API module. Also
 *      will declare all function prototypes of the exported interface. 
 *      The Resource Manager Communication API implements the mechanism to transport exception
 *      packets and control messages between the microblocks and their Core counterpart. All
 *      packet buffers exchanged between microblocks and Core counterparts have to go through
 *      Resource Manager. A callback interface and a polling (blocking and nonblocking) 
 *      interface are provided. Only one can be used at a time.
 *
 * = AUTHOR
 *       Adrian Georgescu
 *       adrian.georgescu@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      10/17/2001 10:11:08 AM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 * $Id: communication.h,v 1.35 2003/09/05 19:52:13 ageorges Exp $
 */

#if !defined(__COMMUNICATION_H__)
#define __COMMUNICATION_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */


#if !defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_)
/**
 * This symbol is used to create the internal comm Id queues into OS memory
 */
#if (_IX_OS_TYPE_ != _IX_OS_LINUX_KERNEL_)

/**
 * With the Linux kernel we can not allocate more than 128K at a time
 * and the code will fail for the memory allocation as the space for the
 * internal queues should be contiguous by design.
 */
#define _IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_ 1

#else

#define _IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_ 1

#endif /* end (_IX_OS_TYPE_ != _IX_OS_LINUX_KERNEL_) */

#endif /* end !defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_) */

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) || (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

#if (_IX_BOARD_TYPE_ == _IX_IXDP2400_)

    #if !defined(_IX_COMPILE_PCI_COMM_IN_)

        /*
         * This symbol enables PCI communication between ingress and egress processors
         * on IXDP2400 platform.
         */

        #define _IX_COMPILE_PCI_COMM_IN_ 1

    #endif /* end !defined(_IX_COMPILE_PCI_COMM_IN_) */

#elif (_IX_BOARD_TYPE_ == _IX_IXDP2401_)

    #if defined(_IX_COMPILE_PCI_COMM_IN_)

        /* In this case just disable the PCI communication */

        #undef _IX_COMPILE_PCI_COMM_IN_ 

    #endif /* end !defined(_IX_COMPILE_PCI_COMM_IN_) */

#endif /* (_IX_BOARD_TYPE_ == _IX_IXDP2400_) */

#endif /* end (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) || (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

#elif (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) || (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

#if (_IX_BOARD_TYPE_ == _IX_IXDP2800_)

    #if !defined(_IX_COMPILE_PCI_COMM_IN_)

        /*
         * This symbol enables PCI communication between ingress and egress processors
         * on IXDP2800 platform.
         */

        #define _IX_COMPILE_PCI_COMM_IN_ 1

    #endif /* end !defined(_IX_COMPILE_PCI_COMM_IN_) */

#elif (_IX_BOARD_TYPE_ == _IX_IXDP2801_)

    #if defined(_IX_COMPILE_PCI_COMM_IN_)

        /* In this case just disable the PCI communication */
        #undef _IX_COMPILE_PCI_COMM_IN_ 

    #endif /* end !defined(_IX_COMPILE_PCI_COMM_IN_) */

#endif /* (_IX_BOARD_TYPE_ == _IX_IXDP2800_) */

#endif /* end (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) || (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */


#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */


/* #define _IX_TEST_THREAD_SWITCHES_ */

#if defined(_IX_TEST_THREAD_SWITCHES_)

extern ix_uint32 g_PacketInterruptCount;

#endif /* end defined (_IX_TEST_THREAD_SWITCHES_) */

#if !defined(_IX_COMM_DISPATCH_THREADS_TUNABLE_)

/* This is for tunable dispatch threads that yield */
#define _IX_COMM_DISPATCH_THREADS_TUNABLE_

#endif /* end !defined(_IX_COMM_DISPATCH_THREADS_TUNABLE_) */


/**
 * TYPENAME: ix_comm_data_handler
 * 
 * DESCRIPTION: This is the definition of the generic data handler function 
 *	pointer. The data handler function takes as parameters a buffer handle 
 *	which in general refers to a packet or message handle, a 32 bit 
 *	unsigned integer where extra user data is passed to the handler, and a 
 *	context pointer that was passed to Resource Manager at the time the 
 *	data handler has been registered with a particular communication ID and 
 *	that will be passed as an argument to the data handler upon each 
 *	invocation. When data is sent to that communication ID through 
 *	send_packet or send_message functions, on the receiving side, the 
 *	registered handler will be invoked. The send_packet and send_message 
 *	have an arg_UserData parameter that will passed as arg_UserData 
 *	argument to the registered handler for the packet or message 
 *	communication ID. This type can be used as a parameter 
 *	for all registration functions but in the actual implementation of the 
 *	handler the first parameter can be specialized to alias ix_handle types 
 *	in order to avoid any confusion for the programmer.
 */
typedef ix_error (*ix_comm_data_handler)( 
                                          ix_buffer_handle arg_hBuffer, 
                                          ix_uint32 arg_UserData,
                                          void* arg_pContext 
                                        );




/**
 * DESCRIPTION: Least significant bit of the local identifier into a communication
 *  ID descriptor.
 *
 */
#define IX_COMM_ID_LOCAL_ID_LSB     0U


/**
 * DESCRIPTION: Most significant bit of the local identifier into a communication
 *  ID descriptor. 
 *
 */
#define IX_COMM_ID_LOCAL_ID_MSB     7U



/**
 * A subsystem ID of 0 means this system. A subsystem ID of 1 means the peer system.
 */


/**
 * DESCRIPTION: Least significant bit of the subsystem ID field into a communication
 *  ID descriptor.
 *
 */
#define IX_COMM_ID_SUBSYSTEM_ID_LSB   8U



/**
 * DESCRIPTION: Most significant bit of the subsystem ID field into a communication
 *  ID descriptor. 
 *
 */
#define IX_COMM_ID_SUBSYSTEM_ID_MSB   8U



/**
 * DESCRIPTION: Least significant bit of the blade ID field into a communication
 *  ID descriptor.
 *
 */
#define IX_COMM_ID_BLADE_ID_LSB     9U



/**
 * DESCRIPTION: Most significant bit of the blade ID field into a communication
 *  ID descriptor.
 *
 */
#define IX_COMM_ID_BLADE_ID_MSB     24U



/**
 * TYPENAME: ix_communication_id
 * 
 * DESCRIPTION: This type describes a communication ID. The communication ID 
 *	has packed inside several pieces of information such as local 
 *	communication ID, system ID and the blade ID. Macros are be provided 
 *	for constructing communication IDs and for accessing different bit 
 *	fields. The local communication IDs are used in the following manner: 
 *	the IDs 0 to IX_COMM_LAST_UBLOCK_ID are reserved for communication with 
 *	the microblocks. A send on a such ID will be sent to the corresponding 
 *	microblock. The IDs IX_COMM_LAST_UBLOCK_ID + 1 to 
 *	IX_COMM_LOCAL_ID_NUMBER are used for communication with the core 
 *	components. The uCode will send the exception packets and messages to 
 *	one of these IDs on which core components will listen.
 *      The current mapping of a communication ID is as follows:
 *      -----------------------------
 *          Field:          Local ID
 *          Bits:           IX_COMM_ID_LOCAL_ID_LSB:IX_COMM_ID_LOCAL_ID_MSB
 *          Size:           IX_COMM_ID_LOCAL_ID_MSB + 1 - IX_COMM_ID_LOCAL_ID_LSB
 *          Description:    This field describes the local communication 
 *	                      identifier
 *          -----------------------------
 *          Field:          Subsystem ID
 *          Bits:           IX_COMM_ID_SUBSYSTEM_ID_LSB:IX_COMM_ID_SUBSYSTEM_ID_MSB
 *          Size:           IX_COMM_ID_SUBSYSTEM_ID_MSB + 1 - IX_COMM_ID_SUBSYSTEM_ID_LSB
 *          Description:    This field describes the subsystem type and is 
 *	                      used mainly for remote communication.
 *          -----------------------------
 *          Field:          Blade ID
 *          Bits:           IX_COMM_ID_BLADE_ID_LSB:IX_COMM_ID_BLADE_ID_MSB
 *          Size:           IX_COMM_ID_BLADE_ID_MSB + 1 - IX_COMM_ID_BLADE_ID_LSB
 *          Description:    This field describes the blade identifier and is
 *                          mainly used for remote communication.
 *
 */
typedef ix_uint32 ix_communication_id;




/**
 * DESCRIPTION: This symbol describes the first local ID coresponding to a 
 *  microblock.
 *
 */
#define IX_COMM_FIRST_UBLOCK_ID    0UL



/**
 * DESCRIPTION: This symbol describes the last local ID coresponding to a 
 *  microblock.
 *
 */
#define IX_COMM_LAST_UBLOCK_ID     63UL


/**
 * DESCRIPTION: This symbol describes the first local ID coresponding to a 
 *  core component.
 *
 */
#define IX_COMM_FIRST_CORE_ID     (IX_COMM_LAST_UBLOCK_ID + 1UL)


/**
 * DESCRIPTION: This symbol describes the number of communication IDs allocated to 
 *  microblocks on a particular subsystem.
 *
 */
#define IX_COMM_UBLOCK_ID_NUMBER   (IX_COMM_LAST_UBLOCK_ID + 1UL - IX_COMM_FIRST_UBLOCK_ID)



/**
 * DESCRIPTION: This symbol describes the  total number of communication IDs on a 
 *  particular subsystem.
 *
 */
#define IX_COMM_LOCAL_ID_NUMBER    (1U << (IX_COMM_ID_LOCAL_ID_MSB + 1U - IX_COMM_ID_LOCAL_ID_LSB))


/**
 * DESCRIPTION: This symbol describes the number of communication IDs allocated to 
 *  core components on a particular subsystem.
 *
 */
#define IX_COMM_CORE_ID_NUMBER   (IX_COMM_LOCAL_ID_NUMBER - IX_COMM_UBLOCK_ID_NUMBER)



/**
 * DESCRIPTION: This value will specify that the communication ID refers to this particular system. A system
 *          is usually composed from an ingress subsystem and an egress subsystem but could be comprised just
 *          of one NP which will perform both ingress and egress duties. Each system will have allocated a
 *          certain ID from 1,65535. This way would be fast to perform communication on the same subsystem
 *          by specifying a IX_COMM_LOCAL_BLADE_ID blade ID.
 *
 */
#define IX_COMM_LOCAL_BLADE_ID   0UL



/**
 * DESCRIPTION: This mask is used to determine if the comm ID refers to the local 
 *  subsystem or not.
 *
 */
#define IX_COMM_REMOTE_SUBSYSTEM_MASK (IX_BIT_FIELD_MASK32(IX_COMM_ID_SUBSYSTEM_ID_LSB, \
                                        IX_COMM_ID_SUBSYSTEM_ID_MSB) | IX_BIT_FIELD_MASK32(IX_COMM_ID_BLADE_ID_LSB, \
                                        IX_COMM_ID_BLADE_ID_MSB))


/**
 * Macros for ix_communication_id type handling
 */

/**
 * MACRO NAME: IX_RM_COMM_ID_GET_LOCAL_ID
 *
 * DESCRIPTION: This macro will retrieve the local ID corresponding to the passed communication ID.
 *
 * @Param:  - IN arg_CommID - a communication ID of type ix_communication_id
 *
 * @Return: An ix_uint32 value will be returned specifying the local ID corresponding to the passed 
 *          communication ID.
 */
#define IX_RM_COMM_ID_GET_LOCAL_ID( \
                                    arg_CommID \
                                  ) \
                                  IX_GET_BIT_FIELD32(arg_CommID, \
                                    IX_COMM_ID_LOCAL_ID_LSB, \
                                    IX_COMM_ID_LOCAL_ID_MSB)


/**
 * MACRO NAME: IX_RM_COMM_ID_GET_SUBSYSTEM_ID
 *
 * DESCRIPTION: This macro will retrieve the subsystem type corresponding to the passed communication ID.
 *
 * @Param:  - IN arg_CommID - a communication ID of type ix_communication_id
 *
 * @Return: An ix_uint32 value will be returned specifying the subsystem ID corresponding to the passed 
 *          communication ID.
 */
#define IX_RM_COMM_ID_GET_SYSTEM_ID( \
                                     arg_CommID \
                                   ) \
                                   (IX_GET_BIT_FIELD32(arg_CommID, \
                                      IX_COMM_ID_SUBSYSTEM_ID_LSB, \
                                      IX_COMM_ID_SUBSYSTEM_ID_MSB))



/**
 * MACRO NAME: IX_RM_COMM_ID_GET_BLADE_ID
 *
 * DESCRIPTION: This macro will retrieve the blade ID corresponding to the passed communication ID.
 *
 * @Param:  - IN arg_CommID - a communication ID of type ix_communication_id
 *
 * @Return: An ix_uint32 value will be returned specifying the blade ID corresponding to the passed 
 *          communication ID.
 */
#define IX_RM_COMM_ID_GET_BLADE_ID( \
                                    arg_CommID \
                                  ) \
                                  IX_GET_BIT_FIELD32(arg_CommID, \
                                    IX_COMM_ID_BLADE_ID_LSB, \
                                    IX_COMM_ID_BLADE_ID_MSB)



/**
 * MACRO NAME: IX_RM_COMM_MAKE_ID
 *
 * DESCRIPTION: This macro will create and return a new communication ID based on the local ID and subsystem
 *          type for the specified blade ID.
 *          
 * @Param:  - IN arg_LocalID - an ix_uint32 value representing the local ID on the specified blade and 
 *          subsystem type
 * @Param:  - IN arg_SubsystemID - an ix_uint32 value representing the subsytem ID of the destination. A 
 *          subsystem ID of 0 means the current subsystem and a value of 1 means the peer subsystem.
 * @Param:  - IN arg_BladeID - the ID of the destination system.
 *
 * @Return: Returns an ix_communication_id value representing the identifier where we want to communicate.
 *          If the identifier corresponds to a remote system or subsystem we can not listen on that ID from
 *          this subsystem, instead we can just send data to that destination.
 */
#define IX_RM_COMM_MAKE_ID( \
                            arg_LocalID, \
                            arg_SubsystemID, \
                            arg_BladeID \
                          ) \
                          ((ix_communication_id) /* typecast */ \
                          IX_MAKE_BIT_FIELD32(arg_LocalID, \
                            IX_COMM_ID_LOCAL_ID_LSB, \
                            IX_COMM_ID_LOCAL_ID_MSB) | \
                          IX_MAKE_BIT_FIELD32(arg_SubsystemID, \
                            IX_COMM_ID_SUBSYSTEM_ID_LSB, \
                            IX_COMM_ID_SUBSYSTEM_ID_MSB) | \
                          IX_MAKE_BIT_FIELD32(arg_BladeID, \
                            IX_COMM_ID_BLADE_ID_LSB, \
                            IX_COMM_ID_BLADE_ID_MSB))


/**
 * MACRO NAME: IX_RM_COMM_MAKE_LOCAL_ID
 *
 * DESCRIPTION: This macro will create and return a new communication ID based on the local ID. This 
 *          communication ID will belong to the local subsystem.
 *          
 * @Param:  - IN arg_LocalID - an ix_uint32 value representing the local ID ofthe destination
 *
 * @Return: Returns an ix_communication_id value representing the identifier where we want to communicate.
 *          The newly created communication ID is a local one and as a consequence we can send data to
 *          this destination and at the same time we could listen on this communication ID.
 */
#define IX_RM_COMM_MAKE_LOCAL_ID( \
                                  arg_LocalID \
                                 ) \
                                 IX_RM_COMM_MAKE_ID(arg_LocalID, 0, 0)
                                  

/**
 * DESCRIPTION: This value represents the start value to be used for notification messages
 */
#define IX_NOTIFICATION_MESSAGE     0xffff0000UL



/**
 * DESCRIPTION: This symbol represents the user data value to be passed along
 *  with the configuration property handle that specifies that the CP data has
 *  changed.
 */
#define IX_CP_NOTIFICATION_DATA_CHANGED      IX_NOTIFICATION_MESSAGE + 2UL


/**
 * TYPENAME: ix_comm_id_mode
 * 
 * DESCRIPTION: This enumerated type describes the receive modes a communication ID could be
 *          in. In the get/select mode the sent data will be stored in an internal queue
 *          and will be later on retrieved by the receiving side. In the callback mode
 *          the data is passed directly to a callback registered by the receiving side.
 *          At the initialization time all the communication IDs are in callback mode.
 *
 */
typedef enum ix_e_comm_id_mode
{
    IX_COMM_ID_MODE_FIRST = 0,
    IX_COMM_ID_MODE_CALLBACK = IX_COMM_ID_MODE_FIRST,
    IX_COMM_ID_MODE_GET_SELECT,
    IX_COMM_ID_MODE_LAST
} ix_comm_id_mode;


#if defined (_IX_COMM_DISPATCH_THREADS_TUNABLE_)

/**
 * The following symbols define the internal dispatch threads behaviour
 */

/**
 * DESCRIPTION: Defines the packet dispatch thread priority
 *
 */
#define IX_RM_COMM_PACKET_DISPATCH_THREAD_PRIORITY IX_OSSL_THREAD_PRI_MEDIUM


/**
 * DESCRIPTION: This constant defines after how many data tokens the packet dispatch
 *  thread will yield processing. It should be higher than 0.
 *
 */
#define IX_RM_COMM_PACKET_DISPATCH_THREAD_YIELD_COUNT   128


/**
 * DESCRIPTION: This constant defines how many milliseconds the packet dispatch thread
 *  will sleep after processing IX_RM_COMM_PACKET_DISPATCH_THREAD_YIELD_COUNT data tokens.
 *
 */
#define IX_RM_COMM_PACKET_DISPATCH_THREAD_YIELD_DELAY   0


/**
 * DESCRIPTION: Defines the message dispatch thread priority
 *
 */
#define IX_RM_COMM_MESSAGE_DISPATCH_THREAD_PRIORITY IX_OSSL_THREAD_PRI_MEDIUM


/**
 * DESCRIPTION: This constant defines after how many data tokens the message dispatch
 *  thread will yield processing. It should be higher than 0.
 *
 */
#define IX_RM_COMM_MESSAGE_DISPATCH_THREAD_YIELD_COUNT   128


/**
 * DESCRIPTION: This constant defines how many milliseconds the message dispatch thread
 *  will sleep after processing IX_RM_COMM_MESSAGE_DISPATCH_THREAD_YIELD_COUNT data tokens.
 *
 */
#define IX_RM_COMM_MESSAGE_DISPATCH_THREAD_YIELD_DELAY   0


#endif /* end defined (_IX_COMM_DISPATCH_THREADS_TUNABLE_) */

 
/**
 * NAME: ix_rm_packet_set_receive_mode
 *
 * DESCRIPTION: This function will set the packet receive mode for the communication Id.
 *  That could be IX_COMM_ID_MODE_CALLBACK or IX_COMM_ID_MODE_GET_SELECT. The change from
 *  one mode to the other should be done explicitely. If this communication ID is in 
 *  get/select mode, all buffers in the internal queue will be dropped and all waiting tasks
 *  will be awaken. If the communication ID is  in callback mode, the registered callback 
 *  will be notified that it has been disconnected from the communication Id, and the the 
 *  mode will become get/select.
 *
 * @Param:  - IN arg_CommunicationId - the communication ID for which we are 
 *	setting the packet receive mode.
 * @Param:  - IN arg_ReceiveMode - packet receive mode. The two possible values are 
 *  IX_COMM_ID_MODE_CALLBACK for callback mode and IX_COMM_ID_MODE_GET_SELECT for 
 *  get/select mode
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_packet_set_receive_mode(
                                        ix_communication_id arg_CommunicationId, 
                                        ix_comm_id_mode arg_ReceiveMode
                                      );

/**
 * NAME: ix_rm_message_set_receive_mode
 *
 * DESCRIPTION: This function will set the message receive mode for the communication Id.
 *  That could be IX_COMM_ID_MODE_CALLBACK or IX_COMM_ID_MODE_GET_SELECT. The change from
 *  one mode to the other should be done explicitely. If this communication ID is in 
 *  get/select mode, all buffers in the internal queue will be dropped and all waiting tasks
 *  will be awaken. If the communication ID is  in callback mode, the registered callback 
 *  will be notified that it has been disconnected from the communication Id, and the the 
 *  mode will become get/select.
 *
 * @Param:  - IN arg_CommunicationId - the communication ID for which we are 
 *	setting the message receive mode.
 * @Param:  - IN arg_ReceiveMode - message receive mode. The two possible values are 
 *  IX_COMM_ID_MODE_CALLBACK for callback mode and IX_COMM_ID_MODE_GET_SELECT for 
 *  get/select mode
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_message_set_receive_mode(
                                         ix_communication_id arg_CommunicationId, 
                                         ix_comm_id_mode arg_ReceiveMode
                                       );

/**
 * The following functions apply just to the get/select receive mode
 */

/**
 * NAME: ix_rm_packet_set_consumer_mode
 *
 * DESCRIPTION: This function will set the packet consumer mode for the communication Id
 *  If arg_ConsumerMode is 0 then the communication Id will be set to single consumer 
 *  mode otherwise the multiple consumer mode will be set.
 *  If the comm Id is in IX_COMM_ID_MODE_CALLBACK receive mode then the function will
 *  return an error.
 *
 * @Param:  - IN arg_CommunicationId - the communication ID for which we are 
 *	setting the packet consumer mode.
 * @Param:  - IN arg_ConsumerMode - consumer mode for Comm Id. 0 for single consumer and
 *  different from 0 for multiple consumer.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_packet_set_consumer_mode(
                                         ix_communication_id arg_CommunicationId, 
                                         ix_uint32 arg_ConsumerMode
                                       );


/**
 * NAME: ix_rm_packet_set_producer_mode
 *
 * DESCRIPTION: This function will set the packet producer mode for the communication Id
 *  If arg_ConsumerMode is 0 then the communication Id will be set to single producer 
 *  mode otherwise the multiple producer mode will be set.
 *  If the comm Id is in IX_COMM_ID_MODE_CALLBACK receive mode then the function will
 *  return an error.
 *
 * @Param:  - IN arg_CommunicationId - the communication ID for which we are 
 *	setting the packet producer mode.
 * @Param:  - IN arg_ProducerMode - producer mode for Comm Id. 0 for single producer and
 *  different from 0 for multiple producer.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_packet_set_producer_mode(
                                         ix_communication_id arg_CommunicationId, 
                                         ix_uint32 arg_ProducerMode
                                       );


/**
 * NAME: ix_rm_message_set_consumer_mode
 *
 * DESCRIPTION: This function will set the message consumer mode for the communication Id
 *  If arg_ConsumerMode is 0 then the communication Id will be set to single consumer 
 *  mode otherwise the multiple consumer mode will be set.
 *  If the comm Id is in IX_COMM_ID_MODE_CALLBACK receive mode then the function will
 *  return an error.
 *
 * @Param:  - IN arg_CommunicationId - the communication ID for which we are 
 *	setting the message consumer mode.
 * @Param:  - IN arg_ConsumerMode - consumer mode for Comm Id. 0 for single consumer and
 *  different from 0 for multiple consumer.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_message_set_consumer_mode(
                                          ix_communication_id arg_CommunicationId, 
                                          ix_uint32 arg_ConsumerMode
                                       );

/**
 * NAME: ix_rm_message_set_producer_mode
 *
 * DESCRIPTION: This function will set the message producer mode for the communication Id
 *  If arg_ConsumerMode is 0 then the communication Id will be set to single producer 
 *  mode otherwise the multiple producer mode will be set.
 *  If the comm Id is in IX_COMM_ID_MODE_CALLBACK receive mode then the function will
 *  return an error.
 *
 * @Param:  - IN arg_CommunicationId - the communication ID for which we are 
 *	setting the message producer mode.
 * @Param:  - IN arg_ProducerMode - producer mode for Comm Id. 0 for single producer and
 *  different from 0 for multiple producer.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_message_set_producer_mode(
                                          ix_communication_id arg_CommunicationId, 
                                          ix_uint32 arg_ProducerMode
                                        );


/**
 * NAME: ix_rm_packet_handler_register
 *
 * DESCRIPTION: This function will register a packet handler for the 
 *	specified communication ID. If the 
 *  _IX_COMM_ID_NOTIFICATION_DISCONNECT_ENABLE_ was defined at the 
 *  time of RM compilation, if another packet handler is in place, that 
 *	packet handler will be invoked with a IX_NULL_BUFFER_HANDLE as the 
 *	first parameter, and IX_COMM_ID_NOTIFICATION_DISCONNECT value as the 
 *	user data parameter. From this point on, the new registered handler 
 *	will be used for processing packets sent to this destination. 
 * 
 * @Param:  - IN arg_CommunicationId - the communication ID for which we are 
 *	registering the packet handler.
 * @Param:  - IN arg_PacketHandler - the packet handler that we want to 
 *	register.
 * @Param:  - IN arg_pContext - context data that will be passed to the 
 *	packet handler on each invocation. The programmer has to make sure that 
 *	the context data is valid as long the packet handler is registered.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_packet_handler_register( 
                                        ix_communication_id arg_CommunicationId, 
                                        ix_comm_data_handler arg_PacketHandler, 
                                        void* arg_pContext 
                                      );



/**
 * NAME: ix_rm_packet_handler_unregister
 *
 * DESCRIPTION: This function will unregister the current packet handler for 
 *	the specified communication ID. From this point on, the packets sent to 
 *	this communication ID will undergo the default processing which at this 
 *	time is to drop the packet (the composing buffers will be returned to 
 *	the free list that owns them).
 * 
 * @Param:  - IN arg_CommunicationId - the communication ID for which we 
 *	unregister the packet handler.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_packet_handler_unregister( 
                                          ix_communication_id arg_CommunicationId 
                                        );




/**
 * NAME: ix_rm_message_handler_register
 *
 * DESCRIPTION: This function will register a message handler for the 
 *	specified communication ID. If another message handler is in place, 
 *	that message handler will be invoked with a IX_NULL_BUFFER_HANDLE as 
 *	the first parameter, and IX_COMM_ID_NOTIFICATION_DISCONNECT value as 
 *	the user data parameter. From this point on, the new registered handler 
 *	will be used for processing messages sent to this destination.
 * 
 * @Param:  - IN arg_CommunicationId - the communication ID for which we are 
 *	registering the message handler.
 * @Param:  - IN arg_MessageHandler - the message handler that we want to 
 *	register.
 * @Param:  - IN arg_pContext - context data that will be passed to the 
 *	message handler on each invocation. The programmer has to make sure 
 *	that the context data is valid as long the message handler is 
 *	registered.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_message_handler_register( 
                                         ix_communication_id arg_CommunicationId, 
                                         ix_comm_data_handler arg_MessageHandler, 
                                         void* arg_pContext 
                                       );



/**
 * NAME: ix_rm_message_handler_unregister
 *
 * DESCRIPTION: This function will unregister the current message handler for 
 *	the specified communication ID. From this point on, the messages sent 
 *	to this communication ID will undergo the default processing which at 
 *	this time is to drop the message (the composing buffers will be 
 *	returned to the free list that owns them).
 *
 * @Param:  - IN arg_CommunicationId - the communication ID for which we 
 *	unregister the message handler.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_message_handler_unregister( 
                                           ix_communication_id arg_CommunicationId 
                                         );




/**
 * NAME: ix_rm_packet_send
 *
 * DESCRIPTION: This function will send a packet to the specified 
 *	communication ID. If the communication ID is in get/select mode and the 
 *	internal packet queue is full then an error will be returned specifying 
 *	that the internal queue is full.
 * 
 * @Param:  - IN arg_CommunicationId - the destination communication ID. 
 * @Param:  - IN arg_PacketBuffer - the first buffer handle from the packet 
 *	that will be sent to the specified communication ID.
 * @Param:  - IN arg_UserData - custom user data that will be passed to the 
 *	packet handler registered with the specified communication ID. This 
 *	argument can be used to pass extra information to the destination.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_packet_send( 
                            ix_communication_id arg_CommunicationId, 
                            ix_buffer_handle arg_PacketBuffer,
                            ix_uint32 arg_UserData
                          );



/**
 * NAME: ix_rm_packet_send_wait
 *
 * DESCRIPTION: This function will send a packet to the specified 
 *	communication ID. If the communication ID is in get/select mode and the 
 *	internal packet queue is full then the send function will wait for the 
 *	specified timeout for the internal queue to become available before 
 *	returning an error.
 * 
 * @Param:  - IN arg_CommunicationId - the destination communication ID. 
 * @Param:  - IN arg_PacketBuffer - the first buffer handle from the packet 
 *	that will be sent to the specified communication ID.
 * @Param:  - IN arg_UserData - custom user data that will be passed to the 
 *	packet handler registered with the specified communication ID. This 
 *	argument can be used to pass extra information to the destination.
 * @Param:  - IN arg_Timeout - timeout specifying for how long we will wait 
 *	for the case when the communication ID is in get/select mode for the 
 *	internal data queue to become available. It is expressed in 
 *	milliseconds.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_packet_send_wait( 
                                 ix_communication_id arg_CommunicationId, 
                                 ix_buffer_handle arg_PacketBuffer,
                                 ix_uint32 arg_UserData,
                                 ix_uint32 arg_Timeout
                               );



/**
 * NAME: ix_rm_message_send
 *
 * DESCRIPTION: This function will send a message to the specified 
 *	communication ID. If the communication ID is in get/select mode and the 
 *	internal message queue is full then an error will be returned 
 *	specifying that the internal queue is full.
 * 
 * @Param:  - IN arg_CommunicationId - the destination communication ID. 
 * @Param:  - IN arg_MessageBuffer - the first buffer handle from the message 
 *	that will be sent to the specified communication ID.
 * @Param:  - IN arg_UserData - custom user data that will be passed to the 
 *	message handler registered with the specified communication ID. This 
 *	argument can be used to pass extra information to the destination.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_message_send( 
                             ix_communication_id arg_CommunicationId, 
                             ix_buffer_handle arg_MessageBuffer,
                             ix_uint32 arg_UserData
                           );



/**
 * NAME: ix_rm_message_send_wait
 *
 * DESCRIPTION: This function will send a message to the specified 
 *	communication ID.If the communication ID is in get/select mode and the 
 *	internal message queue is full then the send function will wait for the 
 *	specified timeout for the internal queue to become available before 
 *	returning an error.
 * 
 * @Param:  - IN arg_CommunicationId - the destination communication ID. 
 * @Param:  - IN arg_MessageBuffer - the first buffer handle from the message 
 *	that will be sent to the specified communication ID.
 * @Param:  - IN arg_UserData - custom user data that will be passed to the 
 *	message handler registered with the specified communication ID. This 
 *	argument can be used to pass extra information to the destination.
 * @Param:  - IN arg_Timeout - timeout specifying for how long we will wait 
 *	for the case when the communication ID is in get/select mode for the 
 *	internal data queue to become available. It is expressed in 
 *	milliseconds.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_message_send_wait( 
                                  ix_communication_id arg_CommunicationId, 
                                  ix_buffer_handle arg_MessageBuffer,
                                  ix_uint32 arg_UserData,
                                  ix_uint32 arg_Timeout
                                );


/**
 * NAME: ix_rm_packet_peek
 *
 * DESCRIPTION: This function will retrieve the number of packets stored in
 *  the internal queue for the specified comm Id. The number might not be the 
 *  same after the function returns. If the comm Id is in IX_COMM_ID_MODE_CALLBACK
 *  receive mode, then an error will be returned.
 * 
 * @Param:  - IN arg_CommunicationId - the local communication ID from which 
 *	we want to get information. 
 * @Param:  - OUT arg_pAvailablePackets - address where the number of available
 *  packets will be stored.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for 
 *	failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_packet_peek(
                            ix_communication_id arg_CommunicationId, 
                            ix_uint32* arg_pAvailablePackets
                          );


/**
 * NAME: ix_rm_packet_get
 *
 * DESCRIPTION: This function will retrieve the next packet and associated 
 *	user data from a get/select communication ID. If the internal packet queue is empty 
 *  then an error will be returned immediately. If the comm Id is in IX_COMM_ID_MODE_CALLBACK
 *  receive mode then an error will be returned.
 * 
 * @Param:  - IN arg_CommunicationId - the local communication ID from which 
 *	we want to get data. 
 * @Param:  - IN arg_pPacketBuffer - address where the first buffer handle 
 *	from the packet from the head of the internal queue will be extracted.
 * @Param:  - IN arg_pUserData - address where the associated user data will 
 *	be stored.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for 
 *	failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_packet_get(
                           ix_communication_id arg_CommunicationId, 
                           ix_buffer_handle* arg_pPacketBuffer,
                           ix_uint32* arg_UserData
                         );


/**
 * NAME: ix_rm_packet_get_wait
 *
 * DESCRIPTION: This function will retrieve the next packet and associated 
 *	user data from a get/select communication ID. If the comm Id is in IX_COMM_ID_MODE_CALLBACK
 *  receive mode then an error will be returned. If the internal packet queue is empty then the 
 *	function will wait for the specified timeout for the data to become 
 *	available before returning an error.
 * 
 * @Param:  - IN arg_CommunicationId - the local communication ID from which 
 *	we want to get data. 
 * @Param:  - IN arg_pPacketBuffer - address where the first buffer handle 
 *	from the packet from the head of the internal queue will be extracted.
 * @Param:  - IN arg_pUserData - address where the associated user data will 
 *	be stored.
 * @Param:  - IN arg_Timeout - timeout specifying for how long we will wait 
 *	for the case when the communication ID is in get/select mode for data to 
 *	become available. It is expressed in milliseconds.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for 
 *	failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_packet_get_wait(
                                ix_communication_id arg_CommunicationId, 
                                ix_buffer_handle* arg_pPacketBuffer,
                                ix_uint32* arg_UserData,
                                ix_uint32 arg_Timeout
                              );

/**
 * NAME: ix_rm_message_peek
 *
 * DESCRIPTION: This function will retrieve the number of messages stored in
 *  the internal queue for the specified comm Id. The number might not be the 
 *  same after the function returns. If the comm Id is in IX_COMM_ID_MODE_CALLBACK
 *  receive mode, then an error will be returned.
 * 
 * @Param:  - IN arg_CommunicationId - the local communication ID from which 
 *	we want to get information. 
 * @Param:  - OUT arg_pAvailableMessages - address where the number of available
 *  messages will be stored.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for 
 *	failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_message_peek(
                            ix_communication_id arg_CommunicationId, 
                            ix_uint32* arg_pAvailableMessages
                          );

/**
 * NAME: ix_rm_message_get
 *
 * DESCRIPTION: This function will retrieve the next message and associated 
 *	user data from a get/select communication ID. If the internal packet queue is empty 
 *  then an error will be returned immediately. If the comm Id is in IX_COMM_ID_MODE_CALLBACK
 *  receive mode then an error will be returned.
 * 
 * @Param:  - IN arg_CommunicationId - the local communication ID from which 
 *	we want to get data. 
 * @Param:  - IN arg_pMessageBuffer - address where the first buffer handle 
 *	from the message from the head of the internal queue will be extracted.
 * @Param:  - IN arg_pUserData - address where the associated user data will 
 *	be stored.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for 
 *	failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_message_get(
                            ix_communication_id arg_CommunicationId, 
                            ix_buffer_handle* arg_pMessageBuffer,
                            ix_uint32* arg_UserData
                          );


/**
 * NAME: ix_rm_message_get_wait
 *
 * DESCRIPTION: This function will retrieve the next message and associated 
 *	user data from a get/select communication ID. If the comm Id is in IX_COMM_ID_MODE_CALLBACK
 *  receive mode then an error will be returned.If the internal message queue is empty then the 
 *	function will wait for the specified timeout for the data to become 
 *	available before returning an error.
 * 
 * @Param:  - IN arg_CommunicationId - the local communication ID from which 
 *	we want to get data. 
 * @Param:  - IN arg_pMessageBuffer - address where the first buffer handle 
 *	from the message from the head of the internal queue will be extracted.
 * @Param:  - IN arg_pUserData - address where the associated user data will 
 *	be stored.
 * @Param:  - IN arg_Timeout - timeout specifying for how long we will wait 
 *	for the case when the communication ID is in get/select mode for data to 
 *	become available. It is expressed in milliseconds.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for 
 *	failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_message_get_wait(
                                 ix_communication_id arg_CommunicationId, 
                                 ix_buffer_handle* arg_pMessageBuffer,
                                 ix_uint32* arg_UserData,
                                 ix_uint32 arg_Timeout
                               );



/**
 * DESCRIPTION: Flag that specifies that we are interested in packet 
 *	processing for a specific communication Id. If this flag is set for a 
 *	comm Id into the action set on the select input then the function will 
 *	return if packets have been received on the comm Id. On return the flag 
 *	will be set only if packets have been received on the comm Id.
 *
 */
#define IX_COMM_SELECT_PACKET_MASK         0x1


/**
 * DESCRIPTION: Flag that specifies that we are interested in message 
 *	processing for a specific communication Id. If this flag is set for a 
 *	comm Id into the action set on the select input then the function will 
 *	return if messages have been received on the comm Id. On return the 
 *	flag will be set only if messages have been received on the comm Id.
 *
 */
#define IX_COMM_SELECT_MESSAGE_MASK        0x2


/**
 * DESCRIPTION: On return from the select call this flag will specify that an 
 *	error occurred for the specified comm Id.
 *
 */
#define IX_COMM_SELECT_ERROR_MASK          0x4

/**
 * TYPENAME: ix_select_action_set
 * 
 * DESCRIPTION: This type represents an array of core local communication ID 
 *	masks that will be passed to the select function. The index 0 in the 
 *	array coresponds to the first core communication ID (select does not 
 *	work on the reserved uBlock comm IDs). If the select caller wants to be 
 *	notified on the arrival of packets or messages on a communication ID 
 *	then the corresponding mask should be set with the corresponding flags.
 *
 */
typedef ix_bit_mask32 ix_comm_select_action_set[IX_COMM_CORE_ID_NUMBER];



/**
 * NAME: ix_rm_comm_select
 *
 * DESCRIPTION: This function will wait on a local set of core component 
 *	communication IDs to have messages and/or data available. The 
 *	arg_SelectActionSet parameter is an array of 32 bit flags specifying 
 *	what kind of data we are waiting for on each of the specified 
 *	communication IDs. The first element of the array coresponds to the 
 *	first core communication ID. You can not wait for uBlock comm IDs for 
 *	data. The function will wait for data for a maximum time specified by 
 *	arg_Timeout parameter. On return the whole select action set will be 
 *	flushed and each mask will be updated to specify if data is available 
 *	or an error has happened on that communication ID.
 *  If one of the requested comm IDs is not in the IX_COMM_ID_MODE_GET_SELECT
 *  receive mode for the corresponding active flags then an error will be
 *  returned.
 * 
 * @Param:  - INOUT arg_SelectActionSet - array of 32bit flags representing 
 *	what kind of data we are waiting on packets/messages for each CC end 
 *	point. On return the flags will be cleared and will specify only the 
 *	data type that has been received on each local CC communication ID.
 * @Param:  - IN arg_Timeout - timeout specifying for how long we want to 
 *	wait for data to become available. It is expressed in milliseconds.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for 
 *	failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_comm_select( 
                            ix_comm_select_action_set arg_SelectActionSet, 
                            ix_uint32 arg_Timeout 
                          );




/**
 * NAME: ix_rm_ublock_packet_comm_init
 *
 * DESCRIPTION: This function will initialize the packet communication from 
 *	the core to the microblock corresponding to the passed communication 
 *	Id. The passed hardware ring will be associated with the communication 
 *	Id, and from this point on all packets sent on this Id will be queued 
 *	into the specified ring.
 * 
 * @Param:  - IN arg_CommunicationId - the microblock communication Id of 
 *	interest
 * @Param:  - IN arg_hHwRing - the hardware ring on which packet 
 *	communication from the core to microblock will be performed.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for 
 *	failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_ublock_packet_comm_init(
                                        ix_communication_id arg_CommunicationId, 
                                        ix_hw_ring_handle arg_hHwRing
                                      );



/**
 * NAME: ix_rm_ublock_message_comm_init
 *
 * DESCRIPTION: This function will initialize the message communication from 
 *	the core to the microblock corresponding to the passed communication 
 *	Id. The passed hardware ring will be associated with the communication 
 *	Id, and from this point on all messages sent on this Id will be queued 
 *	into the specified ring.
 * 
 * @Param:  - IN arg_CommunicationId - the microblock communication Id of 
 *	interest
 * @Param:  - IN arg_hHwRing - the hardware ring on which message 
 *	communication from the core to microblock will be performed.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for 
 *	failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_ublock_message_comm_init(
                                         ix_communication_id arg_CommunicationId, 
                                         ix_hw_ring_handle arg_hHwRing
                                       );

/**
 * The following functions are for internal CCI support
 */

/* CCI support enabled */
#if !defined( _IX_RM_CCI_SUPPORT_EXCLUDE_ )

/**
 * NAME: ix_rm_packet_enable_cc_mode
 *
 * DESCRIPTION: This function will enable packet CC mode for the comm Id. This mode
 *  will work just in the single consumer mode. If the comm Id is in multiple consumer
 *  mode then an error will be returned. In this mode an internal specific CC processing
 *  function will be called when the comm Id is in single consumer get/select mode and the internal
 *  queue goes from empty to non empty state.
 * @Param:  - IN arg_CommunicationId - the communication ID for which we are 
 *	enabling the packet CC mode.
 * @Param:  - IN arg_pCCModeContext - context to be passed to the internal 
 *  specific CC processing function the comm Id is in single consumer get/select 
 *  mode and the internal queue goes from empty to non empty state
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_packet_enable_cc_mode(
                                      ix_communication_id arg_CommunicationId, 
                                      void* arg_pCCModeContext
                                    );



/**
 * NAME: ix_rm_packet_disable_cc_mode
 *
 * DESCRIPTION: This function will disable packet CC mode for the comm Id and 
 *              returns the context originally registered by the CCI so that
 *              the appropriate termination operations can be performed on it.
 * 
 * @Param:  - IN    arg_CommunicationId - the communication ID for which we are 
 *	                    disabling the packet CC mode. 
 * @Param:  - OUT   arg_ppContext - context registered by CCI
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_packet_disable_cc_mode(
                                       ix_communication_id arg_CommunicationId,
                                       void ** arg_ppContext
                                     );



/**
 * NAME: ix_rm_message_enable_cc_mode
 *
 * DESCRIPTION: This function will enable message CC mode for the comm Id. This mode
 *  will work just in the single consumer mode. If the comm Id is in multiple consumer
 *  mode then an error will be returned. In this mode an internal specific CC processing
 *  function will be called when the comm Id is in single consumer get/select mode and the internal
 *  queue goes from empty to non empty state.
 * @Param:  - IN arg_CommunicationId - the communication ID for which we are 
 *	enabling the message CC mode.
 * @Param:  - IN arg_pCCModeContext - context to be passed to the internal 
 *  specific CC processing function the comm Id is in single consumer get/select 
 *  mode and the internal queue goes from empty to non empty state
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_message_enable_cc_mode(
                                       ix_communication_id arg_CommunicationId, 
                                       void* arg_pCCModeContext
                                     );



/**
 * NAME: ix_rm_message_disable_cc_mode
 *
 * DESCRIPTION: This function will disable message CC mode for the comm Id and 
 *              returns the context originally registered by the CCI so that
 *              the appropriate termination operations can be performed on it.
 * 
 * @Param:  - IN    arg_CommunicationId - the communication ID for which we are 
 *	                    disabling the message CC mode. 
 * @Param:  - OUT   arg_ppContext - context registered by CCI
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_message_disable_cc_mode(
                                        ix_communication_id arg_CommunicationId,
                                        void** arg_ppContext
                                      );



#endif /* !defined( _IX_RM_CCI_SUPPORT_EXCLUDE_ ) */


#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__COMMUNICATION_H__) */
