/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 4
 *
 * = LIBRARY
 *      Resource Manager Library
 *
 * = MODULE
 *      Library Interface
 *
 * = FILENAME
 *      configuration_property.h
 *
 * = DESCRIPTION
 *      This file will define all types and function prototypes for the Configuration 
 *      Property API. We provide this API to allow easy and consistent application 
 *      configuration. A notification mechanism based on the Observer/Observable design 
 *      pattern will provide property configuration across the whole system.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      12/18/2001 7:50:14 AM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 * $Id: configuration_property.h,v 1.7 2003/07/14 20:20:13 ageorges Exp $
 */

#if !defined(__CONFIGURATION_PROPERTY_H__)
#define __CONFIGURATION_PROPERTY_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */



/**
 * This API can be seen as System Repository API. Properties will represent name-
 * value entities that can be set and accessed throughout the system in a consitent
 * manner. Properties will be defined by property handles that will link together the
 * name and value pair. We opted for a hierarchical implementation of the registry similar
 * to the file system implementation. We will set a limit on the number of properties
 * that can be created in the system. Each property will have a set of attributes associated
 * with it. There could be properties that can act just as nodes in the hierarchy and that
 * will have no value associated to them. Once a property has been created user can 
 * register with the property to receive notifications if the property changes.
 */



/**
 * TYPENAME: ix_configuration_property_handle
 * 
 * DESCRIPTION: This type represents a generic type for configuration property handle
 *
 */
typedef ix_handle ix_configuration_property_handle;

/**
 * External variable for the core properties tree.
 */
extern ix_configuration_property_handle g_hCoreRegistryRoot;


/**
 * DESCRIPTION: Defines a NULL configuration property handle
 *
 */
#define IX_NULL_CONFIGURATION_PROPERTY_HANDLE  ((ix_configuration_property_handle)0)



/**
 * NAME: ix_rm_cp_get_core_root
 *
 * DESCRIPTION: This function will retrieve the core property tree root.
 * 
 * @Return: Returns the core property tree root.
 */
IX_EXPORT_FUNCTION 
ix_configuration_property_handle ix_rm_cp_get_core_root(void);




/**
 * DESCRIPTION: Defines the root of the core properties tree
 *
 */
#define IX_CP_CORE_PROPERTY_ROOT ix_rm_cp_get_core_root()



/**
 * DESCRIPTION: This symbol defines the maximum length for a property name
 *
 */
#define IX_CP_MAX_PROPERTY_NAME_LENGTH  76UL 



/**
 * DESCRIPTION: This symbol describes the character that will be used as path delimiter
 *  into the CP name
 *
 */
#define IX_CP_PATH_NAME_DELIMITER   '/'




/**
 * TYPENAME: ix_cp_options_flags
 * 
 * DESCRIPTION: 
 *
 */
typedef enum ix_e_cp_options_flags
{
    IX_CP_OPTIONS_FLAGS_FIRST = 0,
    IX_CP_OPTIONS_FLAGS_CONSTANT = 0x1,
    IX_CP_OPTIONS_FLAGS_DATA = 0x2,
    IX_CP_OPTIONS_FLAGS_DATA_UINT32 = 0x4,
    IX_CP_OPTIONS_FLAGS_LAST
} ix_cp_options_flags;




/**
 * TYPENAME: ix_s_cp_property_info
 * 
 * DESCRIPTION: This structure represents the structure of the property attributes information
 *          that will be returned to the user upon request.
 *          
 */
typedef struct ix_s_cp_property_info
{
    ix_uint32                           m_Options;
    ix_uint32                           m_DataSize;
    ix_configuration_property_handle    m_hParentProperty;
    ix_configuration_property_handle    m_hPropertyNext;
    ix_configuration_property_handle    m_hPropertyPrev;
    ix_configuration_property_handle    m_hChildrenPropertiesList;
    char                                m_aName[IX_CP_MAX_PROPERTY_NAME_LENGTH];
} ix_cp_property_info;

/**
 * NAME: ix_rm_cp_property_create
 *
 * DESCRIPTION: This function will create a new property entry at the specified level.
 *          At the parent level the property name should be unique.
 * 
 * @Param:  - IN arg_hParentProperty - Represents the property where the new property will
 *          be created. This value can be IX_CP_CORE_PROPERTY_ROOT in which case the new
 *          property will be created at the top level of the registry.
 * @Param:  - IN arg_pPropertyName - represents the name of the new property to be
 *          created at this level. The name should be unique at the parent property
 *          level or an error will be returned.
 * @Param:  - IN arg_PropertyOptions - represents the options to be applied to the
 *          new created property.
 * @Param:  - OUT arg_pProperty - represents the location where the handle of the new
 *          created property will be stored. In the case of failure this value
 *          will be set to IX_CP_NULL_PROPERTY upon return.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure. If the parent property
 *          is invalid then an error will be returned. If a property with the same name exists
 *          at the specified location, an error will be returned.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_cp_property_create(
                                   ix_configuration_property_handle arg_hParentProperty, 
                                   const char* arg_pPropertyName,
                                   ix_uint32 arg_PropertyOptions,
                                   ix_configuration_property_handle* arg_pProperty 
                                 );



/**
 * NAME: ix_rm_cp_property_delete
 *
 * DESCRIPTION: This function will delete the specified property from the registry.
 * 
 * @Param:  - IN arg_hProperty - this is the property to deleted from the registry.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_cp_property_delete(
                                   ix_configuration_property_handle arg_hProperty
                                 );




/**
 * NAME: ix_rm_cp_property_open
 *
 * DESCRIPTION: This function will retrieve the handle of the property named arg_pPropertyName
 *          at the level of arg_hParentProperty. Once successful the returned handle is a valid
 *          property handle that can be used for further work.
 * 
 * @Param:  - IN arg_hParentProperty - represents the level where we are looking for
 *          the arg_pPropertyName named property.
 * @Param:  - IN arg_pPropertyName - the name of the property we are looking for. The
 *          name can be a full path name in which case the search will go in the tree
 *          structure at the arg_hParentProperty level.
 * @Param:  - OUT arg_pProperty - represents the location where the required property
 *          handle will be stored. If the property does not exist at the specified level
 *          this value will be set to IX_CP_NULL_PROPERTY upon return.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_cp_property_open(
                                 ix_configuration_property_handle arg_hParentProperty, 
                                 const char* arg_pPropertyName,
                                 ix_configuration_property_handle* arg_pProperty 
                               );




/**
 * NAME: ix_rm_cp_property_close
 *
 * DESCRIPTION: This function will invalidate the property handle. From this point on
 *          the use of this handle in any operation will return an error.
 * 
 * @Param:  - IN arg_hProperty - the handle of the property that is to be closed.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_cp_property_close(
                                  ix_configuration_property_handle arg_hProperty
                                );



/**
 * NAME: ix_rm_cp_property_attach
 *
 * DESCRIPTION: This function will register notification communication Id 
 *	with the property specified by arg_hProperty. When this property 
 *	changes, a message will be sent to the registered communication Id with 
 *	the handle of the property that has changed. It depends on the 
 *	communication Id setup when the message will be processed.
 * 
 * @Param:  - IN arg_hProperty - the handle of the property whose changes we 
 *	are interested in.
 * @Param:  - IN arg_hNotificationCommId - the communication Id where 
 *	notification events should be sent to.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for 
 *	failure. If the communication Id is already registered to receive 
 *	notification messages then an error will be returned.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_cp_property_attach(
                                   ix_configuration_property_handle arg_hProperty,
                                   ix_communication_id arg_hNotificationCommId 
                                 );




/**
 * NAME: ix_rm_cp_property_detach
 *
 * DESCRIPTION: This function will unregister the notification communication 
 *	Id with the given property referred by arg_hProperty.
 * 
 * @Param:  - IN arg_hProperty - the handle of the property of interest.
 * @Param:  - IN arg_hNotificationCommId - the communication Id that we want 
 *	to unregister from the property notification list.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for 
 *	failure. If the given communication Id is not registered for 
 *	notification with the specified property, then an error will be 
 *	returned.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_cp_property_detach(
                                   ix_configuration_property_handle arg_hProperty,
                                   ix_communication_id arg_hNotificationCommId 
                                 );




/**
 * NAME: ix_rm_cp_property_set_value
 *
 * DESCRIPTION: This function will associate a value with the property if none is
 *          already associated, and will set the property's value to the specified one.
 * 
 * @Param:  - IN arg_hProperty - the handle of the property whose value we want to set
 * @Param:  - IN arg_DataSize - the size in bytes of the data value.
 * @Param:  - IN arg_pData - address where the data of the value to be set resides.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_cp_property_set_value(
                                      ix_configuration_property_handle arg_hProperty,
                                      ix_uint32 arg_DataSize,
                                      const void* arg_pData
                                    );




/**
 * NAME: ix_rm_cp_property_get_value
 *
 * DESCRIPTION: This function will retrieve the value of the specified property if a
 *          a value is associated with the property.
 * 
 * @Param:  - IN arg_hProperty - the handle of the property whose value we want to retrieve.
 * @Param:  - INOUT arg_pDataSize - on input specifies the size of the buffer addressed by
 *          arg_pData. On output specifies the number of bytes copied to the buffer.
 * @Param:  - OUT arg_pData - location where the property's value should be copied. 
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_cp_property_get_value(
                                      ix_configuration_property_handle arg_hProperty,
                                      ix_uint32* arg_pDataSize,
                                      void* arg_pData
                                    );


/**
 * NAME: ix_rm_cp_property_set_value_uint32
 *
 * DESCRIPTION: This function will associate a value with the property if none is
 *          already associated, and will set the property's value as a uint32 value
 *          to the specified value.
 * 
 * @Param:  - IN arg_hProperty - the handle of the property whose value we want to set
 * @Param:  - IN arg_Value - the new value for the property.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_cp_property_set_value_uint32(
                                             ix_configuration_property_handle arg_hProperty,
                                             ix_uint32 arg_Value
                                           );




/**
 * NAME: ix_rm_cp_property_get_value_uint32
 *
 * DESCRIPTION: This function will retrieve the value of the specified property in the case
 *          property value type is uint32. If the value has a different type then an error will be
 *          returned.
 * 
 * @Param:  - IN arg_hProperty - the handle of the property whose value we want to retrieve.
 * @Param:  - OUT arg_pValue - the location where uint32 value of the property should be stored. 
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_cp_property_get_value_uint32(
                                             ix_configuration_property_handle arg_hProperty,
                                             ix_uint32* arg_pValue
                                           );





/**
 * NAME: ix_rm_cp_property_delete_value
 *
 * DESCRIPTION: This function will delete the value of the property. From this point
 *          on, this property will have no value associated with it.
 * 
 * @Param:  - IN arg_hProperty - the handle of the property whose valeu we want to delete.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_cp_property_delete_value(
                                         ix_configuration_property_handle arg_hProperty
                                       );



/**
 * NAME: ix_rm_cp_property_get_info
 *
 * DESCRIPTION: This function will return the information associated with the property.
 * 
 * @Param:  - IN arg_hProperty - the handle of the property whose attributes we want to
 *          retrieve.
 * @Param:  - OUT arg_pInfo - address where property's information will be stored.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_cp_property_get_info(
                                     ix_configuration_property_handle arg_hProperty, 
                                     ix_cp_property_info* arg_pInfo 
                                   );





/**
 * NAME: ix_rm_cp_property_get_subproperty
 *
 * DESCRIPTION: This function will return the handle of the property with the arg_SubpropertyIndex
 *          index. The subproperties indexes start from 0.
 * 
 * @Param:  - IN arg_hProperty - the handle of the property whose subproperties we
 *          are interested in.
 * @Param:  - IN arg_SubpropertyIndex - the index of the subproperty we want to retrieve.
 * @Param:  - OUT arg_pSubproperty - the address where the subproperty handle will be stored. 
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_cp_property_get_subproperty(
                                            ix_configuration_property_handle arg_hProperty, 
                                            ix_uint32 arg_SubpropertyIndex, 
                                            ix_configuration_property_handle* arg_pSubproperty 
                                          );




#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__CONFIGURATION_PROPERTY_H__) */
