/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      counter_64bit.h
 *
 * = DESCRIPTION
 *      This file will declare all types and function prototypes for the 64 bit counter API.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      5/2/2002 1:44:20 PM - creation time 
 * ============================================================================
 * $Id: counter_64bit.h,v 1.6 2002/09/18 17:27:15 lbmiller Exp $
 */

#if !defined(__COUNTER_64BIT_H__)
#define __COUNTER_64BIT_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */



/**
 * TYPENAME: ix_counter_64bit_handle
 * 
 * DESCRIPTION: Generic type for a 64 bit counter handle
 *
 */
typedef ix_handle ix_counter_64bit_handle;


/**
 * DESCRIPTION: This symbol defines a null 64 bit counter handle
 *
 */
#define IX_NULL_COUNTER_64BIT_HANDLE   ((ix_counter_64bit_handle)0)


/**
 * NAME: ix_rm_counter_64bit_new
 *
 * DESCRIPTION: This function will allocate an array of new 64 bit counters and upon return
 *  will fill the array of counter handles with the handles of the newly created counters.
 *  The caller will specify the number of counters to be created through arg_CounterNumber
 *  parameter. The second parameter is the virtual address of a memory block to be used for
 *  the internal 32 bit counters coresponding to each counter. The block of memory should be
 *  an array of 32 bit integers with at least arg_CounterNumber elements. The first element
 *  of this array will be used as the internal 32 bit counter of the first 64 bit counter 
 *  created and so on. The caller can this way patch into the microcode just the offset of
 *  the start of this block. The last parameter is a pointer to the first element of an array
 *  of 64 bit counter handles that will be filled upon return. The caller has to allocate this
 *  handle array before calling this function.
 *  The third parameter will specify the internal 32 bit timer overflow time for all these timers.
 *  The internal 32 bit timer is guaranteed to be read and reset more often than the specified
 *  overflow time. This way the internal timer will be guarateed not to overflow.
 *
 * @Param:  - IN arg_CounterNumber - the number of counters to be created by this call.
 * @Param:  - IN arg_InternalCounterMemoryBlock - address of the memory block to be used for the
 *  internal part of the timers. This should be common memory, allocated by a call to ix_rm_mem_alloc()
 *  function. 
 * @Param:  - IN arg_InternalOverflowTime - the overflow time for the internal 32bit ME counter
 * @Param:  - OUT arg_pCounterHandle - address of the first element of a counter handle array
 *  where the handles of the newly created counters will be stored upon return.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_counter_64bit_new(
                                  ix_uint32 arg_CounterNumber, 
                                  ix_uint32* arg_InternalCounterMemoryBlock,
                                  ix_uint32 arg_InternalOverflowTime,
                                  ix_counter_64bit_handle* arg_pCounterHandle
                                );




/**
 * NAME: ix_rm_counter_64bit_delete
 *
 * DESCRIPTION: This function will delete the passed counter. From this point on the
 *          handle is invalid and should not be used. All resources related to the
 *          counter will be released.
 * 
 * @Param:  - IN arg_hCounter - handle of the counter to be deleted
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_counter_64bit_delete(
                                     ix_counter_64bit_handle arg_hCounter 
                                   );


/**
 * NAME: ix_rm_counter_64bit_get_internal_overflow_time
 *
 * DESCRIPTION: This function will get the overflow time for the internal 32bit and store at the
 *  passed location.
 * 
 * @Param:  - IN arg_hCounter - the handle of the counter of interest.
 * @Param:  - IN arg_pInternalOverflowTime - address where the current internal counter overflow
 *  time will be stored upon return.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_counter_64bit_get_internal_overflow_time(
                                                         ix_counter_64bit_handle arg_hCounter, 
                                                         ix_uint32* arg_pInternalOverflowTime
                                                       );


/**
 * NAME: ix_rm_counter_64bit_set_internal_overflow_time
 *
 * DESCRIPTION: This function will set the overflow time for the internal 32bit
 *          ME counter to the new value.
 * 
 * @Param:  - IN arg_hCounter - the handle of the counter of interest.
 * @Param:  - IN arg_InternalOverflowTime - the new overflow time for the internal 32bit ME counter
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_counter_64bit_set_internal_overflow_time(
                                                         ix_counter_64bit_handle arg_hCounter, 
                                                         ix_uint32 arg_InternalOverflowTime
                                                       );



/**
 * NAME: ix_rm_counter_64bit_get_value
 *
 * DESCRIPTION: This function will return the 64 bit core value of the counter. Internally
 *          the function will read the ME counter and reset it to 0 and increment the core 
 *          counter with the read value.
 * 
 * @Param:  - IN arg_hCounter - the handle of the counter of interest.
 * @Param:  - OUT arg_pCounterValue - address where the value of the counter should be
 *          stored on return
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_counter_64bit_get_value(
                                        ix_counter_64bit_handle arg_hCounter, 
                                        ix_uint64* arg_pCounterValue 
                                      );




/**
 * NAME: ix_rm_counter_64bit_set_value
 *
 * DESCRIPTION: This function will set the value of the core counter to the specified
 *          value.
 * 
 * @Param:  - IN arg_hCounter - the handle of the counter of interest.
 * @Param:  - IN arg_CounterValue - the new value for the core counter
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_counter_64bit_set_value(
                                        ix_counter_64bit_handle arg_hCounter, 
                                        ix_uint64 arg_CounterValue 
                                      );





#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__COUNTER_64BIT_H__) */
