/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      fast_memory_operations.h
 *
 * = DESCRIPTION
 *      This file defines all types and function prototypes related to the fast memory
 *      operations. These operations have been added to copy data from OS memory to the
 *      RM managed memory and vice versa. We provide a fast copy from two locations in
 *      the RM managed memory areas. These fuctions will take advantage of CPU instructions
 *      that allow for fast memory transfer.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      9/10/2002 12:12:46 PM - creation time 
 * ============================================================================
 * $Id: fast_memory_operations.h,v 1.2 2002/09/18 17:27:15 lbmiller Exp $
 */

#if !defined(__FAST_MEMORY_OPERATIONS_H__)
#define __FAST_MEMORY_OPERATIONS_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */


/**
 * NAME: ix_rm_managed_to_os_memory_copy
 *
 * DESCRIPTION: This function will copy a managed memory area into OS memory. This function
 *  will take advantage of the CPU fast instructions only if data is 4 byte aligned.
 * 
 * @Param:  - INOUT arg_pDest - this is the OS destination memory location where the data will be copied to
 * @Param:  - IN arg_pSrc - this is the managed memory location where the data will be copied from
 * @Param:  - IN arg_Count - this is the amount of data to be copied.
 *
 * @Return: The address of the destination buffer
 */
IX_EXPORT_FUNCTION 
void* ix_rm_managed_to_os_memory_copy(
                                       void* arg_pDest, 
                                       const void* arg_pSrc,
                                       ix_uint32 arg_Count 
                                     );


/**
 * NAME: ix_rm_os_to_managed_memory_copy
 *
 * DESCRIPTION: This function will copy a OS memory area into managed memory. This function
 *  will take advantage of the CPU fast instructions only if data is 4 byte aligned.
 * 
 * @Param:  - INOUT arg_pDest - this is the managed destination memory location where the data will be copied to
 * @Param:  - IN arg_pSrc - this is the OS memory location where the data will be copied from
 * @Param:  - IN arg_Count - this is the amount of data to be copied.
 *
 * @Return: The address of the destination buffer
 */
IX_EXPORT_FUNCTION 
void* ix_rm_os_to_managed_memory_copy(
                                       void* arg_pDest, 
                                       const void* arg_pSrc,
                                       ix_uint32 arg_Count 
                                     );


/**
 * NAME: ix_rm_managed_to_managed_memory_copy
 *
 * DESCRIPTION: This function will copy a managed memory area into managed memory. This function
 *  will take advantage of the CPU fast instructions only if data is 4 byte aligned.
 * 
 * @Param:  - INOUT arg_pDest - this is the managed destination memory location where the data will be copied to
 * @Param:  - IN arg_pSrc - this is the managed memory location where the data will be copied from
 * @Param:  - IN arg_Count - this is the amount of data to be copied.
 *
 * @Return: The address of the destination buffer
 */
IX_EXPORT_FUNCTION 
void* ix_rm_managed_to_managed_memory_copy(
                                            void* arg_pDest, 
                                            const void* arg_pSrc,
                                            ix_uint32 arg_Count 
                                          );



#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__FAST_MEMORY_OPERATIONS_H__) */
