/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      hardware.h
 *
 * = DESCRIPTION
 *      This file will define all exported data types used in the Hardware Resource Manager module. Also
 *      will declare all function prototypes of the exported interface.
 *      This API manage MEv2 hardware features including SRAM queues, rings, scratch rings etc.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      4/23/2002 12:37:47 PM - creation time 
 * ============================================================================
 * $Id: hardware.h,v 1.10 2003/09/19 21:10:54 ageorges Exp $
 */

#if !defined(__HARDWARE_H__)
#define __HARDWARE_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */



/**
 * The SRAM controllers for the IXP2000 series of processors support a data structure
 * called Q-Array, which handles basic queue management commands through hardware. These
 * hardware supported queues enable faster turn around times for packets in the fast path.
 * Each element in the Q-Array is a queue descriptor and can be used to point to a queue, 
 * ring or journal. The Q-Array supports up to 64 queue descriptors on chip.
 */

/**
 * TYPENAME: ix_hw_queue_handle
 * 
 * DESCRIPTION: This is a generic type for hardware queue descriptor.
 *          The following is the mapping of the hardware queue descriptor:
 *          +----------+----------------+-------------+
 *          | Channel  |    Reserved    | Queue Index |
 *          | (30..31) |                |    (0..5)   |
 *          +----------+----------------+-------------+
 *
 */
typedef ix_handle ix_hw_queue_handle;


/**
 * Queue handle fields accessors and modifiers.
 * These macros should be used for portability in the case the encoding will change.
 */


/**
 * MACRO NAME: IX_RM_HW_QUEUE_GET_INDEX
 *
 * DESCRIPTION: This macro extracts the queue index from a queue handle. 
 *
 * @Param:  - IN arg_hHwQueue - represents a queue handle of type ix_hw_queue_handle
 *
 * @Return: a ix_uint32 value representing the queue index of the handle.
 */
#define IX_RM_HW_QUEUE_GET_INDEX( \
                                  arg_hHwQueue \
                                ) \
                                IX_GET_BIT_FIELD32(arg_hHwQueue, 0U, 5U)


/**
 * MACRO NAME: IX_RM_HW_QUEUE_SET_INDEX
 *
 * DESCRIPTION: This macro will set the queue index in a queue handle.
 *
 * @Param:  - IN arg_hHwQueue - represents a queue handle of type ix_hw_queue_handle.
 * @Param:  - IN arg_QueueIndex - represents the queue index of type ix_uint32.
 *
 * @Return: The new value of the queue handle is returned.
 */
#define IX_RM_HW_QUEUE_SET_INDEX( \
                                  arg_hHwQueue, \
                                  arg_QueueIndex \
                                ) \
                                IX_SET_BIT_FIELD32(arg_hHwQueue, arg_QueueIndex, 0U, 5U)
 
 
/**
 * MACRO NAME: IX_RM_HW_QUEUE_GET_CHANNEL
 *
 * DESCRIPTION: This macro extracts the queue channel from a queue handle.
 *
 * @Param:  - IN arg_hHwQueue - represents a queue handle of type ix_hw_queue_handle
 *
 * @Return: a ix_uint32 value representing the queue channel of the handle.
 */
#define IX_RM_HW_QUEUE_GET_CHANNEL( \
                                    arg_hHwQueue \
                                  ) \
                                  IX_GET_BIT_FIELD32(arg_hHwQueue, 30U, 31U)


/**
 * MACRO NAME: IX_RM_HW_QUEUE_SET_CHANNEL
 *
 * DESCRIPTION: This macro will set the queue channel in a queue handle.
 *
 * @Param:  - IN arg_hHwQueue - represents a queue handle of type ix_hw_queue_handle.
 * @Param:  - IN arg_QueueChannel - represents the queue channel of type ix_uint32.
 *
 * @Return: The new value of the queue handle is returned.
 */
#define IX_RM_HW_QUEUE_SET_CHANNEL( \
                                    arg_hHwQueue, \
                                    arg_QueueChannel \
                                  ) \
                                  IX_SET_BIT_FIELD32(arg_hHwQueue, arg_QueueChannel, 30U, 31U)


/**
 * TYPENAME: ix_hw_ring_handle
 * 
 * DESCRIPTION: This is a generic type for hardware ring descriptor.
 *          The following is the mapping of the ring handle encoding:
 *          +------------+------------+-----------+-----------+
 *          |  Channel   |    Size    | Mem type  | Ring index|
 *          |   8 bit    |    8 bit   |    8 bit  |   8 bit   |
 *          +------------+------------+-----------+-----------+
 *
 */
typedef ix_handle ix_hw_ring_handle;



/**
 * Ring handle fields accessors and modifiers.
 * These macros should be used for portability in the case the encoding will change.
 */


/**
 * MACRO NAME: IX_RM_HW_RING_GET_CHANNEL
 *
 * DESCRIPTION: This macro extracts the channel from a ring handle
 *
 * @Param:  - IN arg_hHwRing - is the ring handle whose channel we need. Should be of type
 *          ix_hw_ring_handle.
 *
 * @Return: a ix_uint32 value representing the channel
 */
#define IX_RM_HW_RING_GET_CHANNEL( \
                                   arg_hHwRing \
                                 ) \
                                 IX_GET_BIT_FIELD32(arg_hHwRing, 24U, 31U)


/**
 * MACRO NAME: IX_RM_HW_RING_SET_CHANNEL
 *
 * DESCRIPTION: This macro will set the ring channel in the ring handle
 *
 * @Param:  - IN arg_hHwRing - the ring handle whose channel we want to set. 
 *          Should be of type ix_hw_ring_handle.
 * @Param:  - IN arg_HwRingChannel - the channel value for the ring. Should be of type ix_uint32.
 *
 * @Return: The new value of the ring handle is returned.
 */
#define IX_RM_HW_RING_SET_CHANNEL( \
                                   arg_hHwRing, \
                                   arg_HwRingChannel \
                                 ) \
                                 IX_SET_BIT_FIELD32(arg_hHwRing, arg_HwRingChannel, 24U, 31U)


/**
 * MACRO NAME: IX_RM_HW_RING_GET_SIZE
 *
 * DESCRIPTION: This macro retrieves the size field from a ring handle.
 *
 * @Param:  - IN arg_hHwRing - is the ring handle whose size we need. Should be of type
 *          ix_hw_ring_handle.
 *
 * @Return: a ix_uint32 value representing the size.
 */
#define IX_RM_HW_RING_GET_SIZE( \
                                arg_hHwRing \
                              ) \
                              IX_GET_BIT_FIELD32(arg_hHwRing, 16U, 23U)


/**
 * MACRO NAME: IX_RM_HW_RING_SET_SIZE
 *
 * DESCRIPTION: This macro sets the size field in a ring handle.
 *
 * @Param:  - IN arg_hHwRing - the ring handle whose size we want to set. 
 *          Should be of type ix_hw_ring_handle.
 * @Param:  - IN arg_HwRingSize - the size value for the ring. Should be of type ix_uint32.
 *
 * @Return: The new value of the ring handle is returned.
 */
#define IX_RM_HW_RING_SET_SIZE( \
                                arg_hHwRing, \
                                arg_HwRingSize \
                              ) \
                              IX_SET_BIT_FIELD32(arg_hHwRing, arg_HwRingSize, 16U, 23U)


/**
 * MACRO NAME: IX_RM_HW_RING_GET_MEMORY_TYPE
 *
 * DESCRIPTION: This macro retrieves the memory type from a ring handle.
 *
 * @Param:  - IN arg_hHwRing - is the ring handle whose memory type we need. Should be 
 *          of type ix_hw_ring_handle.
 *
 * @Return: a ix_memory_type value representing the memory type.
 */
#define IX_RM_HW_RING_GET_MEMORY_TYPE( \
                                       arg_hHwRing \
                                     ) \
                                     ((ix_memory_type)IX_GET_BIT_FIELD32(arg_hHwRing, 8U, 15U))
                        

/**
 * MACRO NAME: IX_RM_HW_RING_SET_MEMORY_TYPE
 *
 * DESCRIPTION: This macro sets the memory type in a ring handle.
 *
 * @Param:  - IN arg_hHwRing - the ring handle whose memory type we want to set. 
 *          Should be of type ix_hw_ring_handle.
 * @Param:  - IN arg_HwRingMemoryType - the memory type value for the ring. Should be of 
 *          type ix_memory_type.
 *
 * @Return: The new value of the ring handle is returned.
 */
#define IX_RM_HW_RING_SET_MEMORY_TYPE( \
                                       arg_hHwRing, \
                                       arg_HwRingMemoryType \
                                     ) \
                                     IX_SET_BIT_FIELD32(arg_hHwRing, arg_HwRingMemoryType, 8U, 15U)


/**
 * MACRO NAME: IX_RM_HW_RING_GET_INDEX
 *
 * DESCRIPTION: This macro will retrieve the ring index from a ring handle.
 *
 * @Param:  - IN arg_hHwRing - is the ring handle whose ring index we need. Should be 
 *          of type ix_hw_ring_handle.
 *
 * @Return: a ix_uint32 value representing the ring index.
 */
#define IX_RM_HW_RING_GET_INDEX( \
                                 arg_hHwRing \
                               ) \
                               IX_GET_BIT_FIELD32(arg_hHwRing, 0U, 7U)


/**
 * MACRO NAME: IX_RM_HW_RING_SET_INDEX
 *
 * DESCRIPTION: This macro will set the ring index in a ring handle.
 *
 * @Param:  - IN arg_hHwRing - the ring handle whose ring index we want to set. 
 *          Should be of type ix_hw_ring_handle.
 * @Param:  - IN arg_HwRingIndex - the ring index value for the ring. Should be of type ix_uint32.
 *
 * @Return: The new value of the ring handle is returned.
 */
#define IX_RM_HW_RING_SET_INDEX( \
                                 arg_hHwRing, \
                                 arg_HwRingIndex \
                               ) \
                               IX_SET_BIT_FIELD32(arg_hHwRing, arg_HwRingIndex, 0U, 7U)




/**
 * TYPENAME: ix_sram_ring_size
 * 
 * DESCRIPTION: This enumerated type describes the allowed sizes for an SRAM ring
 *
 */
typedef enum ix_e_sram_ring_size
{
    IX_SRAM_RING_SIZE_FIRST = 0,
    IX_SRAM_RING_SIZE_512 = IX_SRAM_RING_SIZE_FIRST,
    IX_SRAM_RING_SIZE_1K,
    IX_SRAM_RING_SIZE_2K,
    IX_SRAM_RING_SIZE_4K,
    IX_SRAM_RING_SIZE_8K,
    IX_SRAM_RING_SIZE_16K,
    IX_SRAM_RING_SIZE_32K,
    IX_SRAM_RING_SIZE_64K,
    IX_SRAM_RING_SIZE_LAST
} ix_sram_ring_size;




/**
 * TYPENAME: ix_scratch_ring_size
 * 
 * DESCRIPTION: This enumerated type describes the allowed sizes for an SCRATCH ring
 *
 */
typedef enum ix_e_scratch_ring_size
{
    IX_SCRATCH_RING_SIZE_FIRST = 0,
    IX_SCRATCH_RING_SIZE_128 = IX_SCRATCH_RING_SIZE_FIRST,
    IX_SCRATCH_RING_SIZE_256,
    IX_SCRATCH_RING_SIZE_512,
    IX_SCRATCH_RING_SIZE_1K,
    IX_SCRATCH_RING_SIZE_LAST
} ix_scratch_ring_size;




/**
 * Hardware SRAM queues functions
 */

/**
 * NAME: ix_rm_hw_queue_create
 *
 * DESCRIPTION: This function will reserve arg_QueuesNumber entries in the Q-Array of the SRAM
 *          controller on the specified channel. Note that no queue descriptors are actually
 *          allocated by this call. The queues should be contiguos otherwise the call will fail.
 *          *arv_pQueueHandleBase will have the handle of the first queue upon return. 
 *          A macro will be provided to get Nth handle relative to the base handle when required.
 * 
 * @Param:  - IN arg_Channel - the channel of SRAM controller in question.
 * @Param:  - IN arg_QueuesNumber - number of queues to be allocated.
 * @Param:  - IN arg_Flags - flags passed to the request.
 * @Param:  - OUT arv_pQueueHandleBase - address where the first queue handle will be stored. 
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure. If the Resource Manager can not find
 *          arg_QueuesNumber contiguos queues then the call will fail.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_hw_queue_create( 
                                ix_uint32 arg_Channel, 
                                ix_uint32 arg_QueuesNumber, 
                                ix_uint32 arg_Flags, 
                                ix_hw_queue_handle* arv_pQueueHandleBase 
                              );


/**
 * DESCRIPTION: This symbol defines a mask for a hardware queue handle
 *
 */
#define IX_HW_QUEUE_HANDLE_SIGNATURE 0xabfee28a

/**
 * MACRO NAME: IX_RM_HW_QUEUE_CREATE_HANDLE
 *
 * DESCRIPTION: Creates a queue handle given the channel and queue index
 *
 * @Param:  - IN arv_QueueChannel - ix_uint32 value specifying the channel.
 * @Param:  - IN arv_QueueIndex - ix_uint32 value specifying the queue index.
 *
 * @Return: A new created queue handle is returned.
 */
#define IX_RM_HW_QUEUE_CREATE_HANDLE( \
                                      arv_QueueChannel, \
                                      arv_QueueIndex \
                                    ) \
                                    ((ix_hw_queue_handle) \
                                    (IX_MAKE_BIT_FIELD32(arv_QueueChannel,30U,31U) | \
                                    (IX_BIT_FIELD_MASK32(6U,29U) & IX_HW_QUEUE_HANDLE_SIGNATURE) | \
                                    IX_MAKE_BIT_FIELD32(arv_QueueIndex,0U,5U)))



/**
 * MACRO NAME: IX_RM_HW_QUEUE_HANDLE_IS_VALID
 *
 * DESCRIPTION: This macro will check if the passed hardware queue handle is valid
 *
 * @Param:  - IN arg_hHwQueue - a ix_hw_queue_handle that represents the queue handle of interest.
 *
 * @Return: Returns non zero if the handle is valid or zero otherwise.
 */
#define IX_RM_HW_QUEUE_HANDLE_IS_VALID( \
                                        arg_hHwQueue \
                                      ) \
                             ((((IX_HW_QUEUE_HANDLE_SIGNATURE ^ (arg_hHwQueue)) & \
                                IX_BIT_FIELD_MASK32(6U, 29U)) == 0) ? 1 : 0)


/**
 * MACRO NAME: IX_RM_HW_QUEUE_GET_NEXT_N_HANDLE
 *
 * DESCRIPTION: This macro will return the Nth next queue handle from the passed queue handle
 *          base. This macro is provided for portability. It makes the assumption that the 
 *          queues in cause are contiguos.
 *
 * @Param:  - IN arv_hHwQueueBase - represents the base queue handle of type ix_hw_queue_handle
 * @Param:  - IN arg_Index - a ix_uint32 value representing the index of the required handle
 *          relative to arv_QueueHandle
 *
 * @Return: A new created queue handle is returned.
 */
#define IX_RM_HW_QUEUE_GET_NEXT_N_HANDLE( \
                                          arv_hHwQueueBase, \
                                          arg_Index \
                                        ) \
                                        IX_RM_HW_QUEUE_CREATE_HANDLE( \
                                           IX_RM_HW_QUEUE_GET_CHANNEL(arv_hHwQueueBase), \
                                           IX_RM_HW_QUEUE_GET_INDEX(arv_hHwQueueBase)+arg_Index \
                                        ) 

/**
 * NAME: ix_rm_hw_queue_delete
 *
 * DESCRIPTION: This function deletes the specified queue. The Q Array entry will be freed.  
 * 
 * @Param:  - IN arg_hHwQueue - represents the queue handle that should be deleted.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_hw_queue_delete( 
                                ix_hw_queue_handle arg_hHwQueue 
                              );




/**
 * NAME: ix_rm_hw_queue_array_get_base_address
 *
 * DESCRIPTION: This function will return the virtual base address of the core allocated
 *          queue descriptor array for each SRAM memory channel. Using this address the
 *          user can get the physical offset for the queue descriptor array by a call to
 *          ix_rm_get_phys_offset() call. This value has to be patched into the microcode
 *          that will perform the other half of the QArray initialization on each SRAM
 *          channel.
 * 
 * @Param:  - IN arg_Channel - represents the SRAM channel of interest.
 * @Param:  - IN arg_pQArrayBaseAddress - represents the address where the queue descriptor
 *          array base address will be stored upon return.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_hw_queue_array_get_base_address(
                                                ix_uint32 arg_Channel, 
                                                ix_uint32** arg_pQArrayBaseAddress
                                              );



/**
 * NAME: ix_rm_hw_enqueue
 *
 * DESCRIPTION: Enqueue an SRAM entry to the tail of specified queue. The first 4 bytes
 *          of each entry is reserved for hardware use. The caller needs to set it in the
 *          specified format as in [IXP2000-CAS]. Essentially the LSB 24 bits are used as
 *          a next address and the 8 MSB bits are used as a cell count, EOP and SOP flags. By
 *          default these will be set to 0xff.
 * 
 * @Param:  - IN arg_hHwQueue - queue handle where to enqueue the entry.
 * @Param:  - IN arg_hHwBuffer - the entry to be enqueued of type ix_buffer_handle. For
 *          queues only hardware buffer handles can be interpreted by the hardware and enqueued
 *          and dequeued.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_hw_enqueue( 
                           ix_hw_queue_handle arg_hHwQueue, 
                           ix_buffer_handle arg_hHwBuffer 
                         );



/**
 * NAME: ix_rm_hw_dequeue
 *
 * DESCRIPTION: This function will dequeue an SRAM entry from the specified queue.  
 * 
 * @Param:  - IN arg_hHwQueue - queue handle from where to dequeue the entry. 
 * @Param:  - OUT arg_pHwBufferHandle - location where to return the entry should 
 *          be stored. For queues the dequeued value is a 32 bit integer with a
 *          predefined packing which is modeled by the ix_buffer_handle type.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_hw_dequeue( 
                           ix_hw_queue_handle arg_hHwQueue, 
                           ix_buffer_handle* arg_pHwBufferHandle 
                         );


/**
 * The scratch pad memory in IXP2000 support rings of various sizes. During scratch ring
 * creation, the Resource Manager initializes the ring registers with appropriate base
 * address and size fields. The scratch memory needed for the ring is also allocated. Once
 * the ring is created, the applications can call Resource Manager functions to put/get data
 * in these rings. These rings are accessible from microengines also. A total of 16 rings
 * are supported by the hardware and the ring numbers are from 0 to 15.
 * SRAM rings are supported by the Q-Array. The number of rings supported is restricted only
 * by the entries free in the Q-Array.
 */

/**
 * SRAM and SCRATCH ring functions
 */


/**
 * NAME: ix_rm_hw_sram_ring_create
 *
 * DESCRIPTION: This function will create a ring in SRAM on the specified channel.
 *          The ring size is limited to several values  as specified by ix_sram_ring_size 
 *          enumerated type. The Resource Manager allocates necessary memory to hold the 
 *          ring data in SRAM. The handle returned can be used on the core to consume and 
 *          produce data to this ring.
 * 
 * @Param:  - IN arg_Channel - specifies the channel for the memory.
 * @Param:  - IN arg_RingSize - represents the ring element size as specified by the ix_sram_ring_size
 *          enumerated type.
 * @Param:  - OUT arg_pHwRingHandle - represents the address where created ring handle should be stored 
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_hw_sram_ring_create( 
                                    ix_uint32 arg_Channel,
                                    ix_sram_ring_size arg_RingSize,
                                    ix_hw_ring_handle* arg_pHwRingHandle 
                                  );


/**
 * NAME: ix_rm_hw_scratch_ring_create
 *
 * DESCRIPTION: This function will create a ring in SRAM on the specified channel.
 *          The ring size is limited to several values  as specified by ix_scratch_ring_size 
 *          enumerated type. The Resource Manager allocates necessary memory to hold the 
 *          ring data in SCRATCH. The handle returned can be used on the core to consume and 
 *          produce data to this ring.
 * 
 * @Param:  - IN arg_Channel - specifies the channel for the memory.
 * @Param:  - IN arg_RingSize - represents the ring element size as specified by the ix_scratch_ring_size
 *          enumerated type.
 * @Param:  - IN arg_RingInternalId - specifies the internal scratch ring Id to be associated with the ring.
 *          If the specified Id is use then the call will fail. This approach has been taken due to limitation
 *          of the ME instruction set which requires that the ring Id should be a compile time constant.
 * @Param:  - OUT arg_pHwRingHandle - represents the address where created ring handle should be stored 
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_hw_scratch_ring_create( 
                                       ix_uint32 arg_Channel,
                                       ix_scratch_ring_size arg_RingSize,
                                       ix_uint32 arg_RingInternalId,
                                       ix_hw_ring_handle* arg_pHwRingHandle 
                                     );


/**
 * NAME: ix_rm_hw_ring_delete
 *
 * DESCRIPTION: This function deletes the specified ring. The ring handle is freed and the
 *          Resource Manager reclaims allocated memory.
 * 
 * @Param:  - IN arg_hHwRing - represents the ring to be deleted.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_hw_ring_delete( 
                               ix_hw_ring_handle arg_hHwRing 
                             );



/**
 * NAME: ix_rm_hw_ring_put
 *
 * DESCRIPTION: This function will produce arg_DataSize 32 bit words from 
 *	        arg_pData address on the ring specified by arg_hHwRing. If the ring 
 *	        will not accomodate the entire data, then the function will return an 
 *	        error.
 * 
 * @Param:  - IN arg_hHwRing - represents the handle of the ring where data
 *	        will be stored.
 * @Param:  - INOUT arg_pDataSize - represents the number of 32 bit words to be 
 *	        written on the ring on the input and the actual number of 32 bit words
 *          written on the output.
 * @Param:  - IN arg_pData - represents the location of the data that is 
 *	        intended to be stored.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_hw_ring_put( 
                            ix_hw_ring_handle arg_hHwRing, 
                            ix_uint32* arg_pDataSize,
                            const ix_uint32* arg_pData 
                          );



/**
 * NAME: ix_rm_hw_ring_get
 *
 * DESCRIPTION: This function will consume from the ring arg_DataSize 32 bit words and store
 *          them at the location specified by arg_pData. 
 * 
 * @Param:  - IN arg_hHwRing - represents the ring where the data is consumed from.
 * @Param:  - INOUT arg_pDataSize - represents the number of 32 bit words to be 
 *	        read from the ring on the input and the actual number of 32 bit words
 *          read on the output.
 * @Param:  - OUT arg_pData - the location where the consumed data should be stored. 
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure. 
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_hw_ring_get( 
                            ix_hw_ring_handle arg_hHwRing, 
                            ix_uint32* arg_pDataSize, 
                            ix_uint32* arg_pData 
                          );



#if defined (_IX_RM_DEBUG_) && defined (_IX_RM_IMPL_HARDWARE_)



/**
 * NAME: ix_rm_scratch_ring_print_info
 *
 * DESCRIPTION: This function will print information about the requested scratch ring 
 *  number.
 * 
 * @Param:  - IN arg_RingNumber - a number representing the scratch ring whose info
 *  we are interested in.
 *
 */
IX_EXPORT_FUNCTION 
void ix_rm_scratch_ring_print_info(
                                    ix_uint32 arg_RingNumber
                                  );


/**
 * NAME: ix_rm_scratch_ring_print_data
 *
 * DESCRIPTION: This function will print the data existing in the requested scratch ring 
 *  number.
 * 
 * @Param:  - IN arg_RingNumber - a number representing the scratch ring whose data
 *  we are interested in.
 *
 */
IX_EXPORT_FUNCTION 
void ix_rm_scratch_ring_print_data(
                                    ix_uint32 arg_RingNumber
                                  );



/**
 * NAME: ix_rm_sram_ring_print_info
 *
 * DESCRIPTION: This function will print information about the requested sram ring 
 *  number.
 * 
 * @Param:  - IN arg_ChannelNumber - the channel number of the SRAM ring of interest.
 * @Param:  - IN arg_RingNumber - a number representing the sram ring whose info
 *  we are interested in.
 *
 */
IX_EXPORT_FUNCTION 
void ix_rm_sram_ring_print_info(
                                 ix_uint32 arg_ChannelNumber,
                                 ix_uint32 arg_RingNumber
                               );


/**
 * NAME: ix_rm_sram_ring_print_data
 *
 * DESCRIPTION: This function will print the data existing in the requested sram ring 
 *  number.
 * 
 * @Param:  - IN arg_ChannelNumber - the channel number of the SRAM ring of interest.
 * @Param:  - IN arg_RingNumber - a number representing the sram ring whose data
 *  we are interested in.
 *
 */
IX_EXPORT_FUNCTION 
void ix_rm_sram_ring_print_data(
                                 ix_uint32 arg_ChannelNumber,
                                 ix_uint32 arg_RingNumber
                               );


#endif /* defined (_IX_RM_DEBUG_) && defined (_IX_RM_IMPL_HARDWARE_) */


#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__HARDWARE_H__) */
