/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      internal_buffer.h
 *
 * = DESCRIPTION
 *      This file defines all data types and function prototypes required internaly by the 
 *      Buffer API module.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      4/23/2002 12:54:57 PM - creation time 
 * ============================================================================
 * $Id: internal_buffer.h,v 1.16 2003/09/19 21:11:03 ageorges Exp $
 */

#if !defined(__INTERNAL_BUFFER_H__)
#define __INTERNAL_BUFFER_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */


/**
 * The following are support macros required for buffer processing
 */

/**
 * MACRO NAME: IX_RM_BUFFER_GET_TYPE
 *
 * DESCRIPTION: This macro will retrieve the type of the buffer
 *          represented by the passed handle.
 *
 * @Param:  - IN arg_hBuffer - handle to a buffer.
 *
 * @Return: an ix_buffer_type enumerated type representing the buffer type.
 */
#define IX_RM_BUFFER_GET_TYPE( \
                                arg_hBuffer \
                             ) \
                             ((ix_buffer_type)IX_GET_BIT_FIELD32(arg_hBuffer, 0U, 0U))



/**
 * MACRO NAME: IX_RM_BUF_SW_GET_FL_ID
 *
 * DESCRIPTION: This macro will retrieve the FL ID from the software buffer handle passed as a
 *          parameter.
 *
 * @Param:  - IN arg_hSwBuffer - handle to a software buffer
 *
 * @Return: an ix_uint32 value representing the FL ID this buffer belongs to
 */
#define IX_RM_BUF_SW_GET_FL_ID( \
                               arg_hSwBuffer \
                             ) \
                             IX_GET_BIT_FIELD32(arg_hSwBuffer, IX_BUF_HANDLE_SW_INDEX_BITS + 1U, \
                             IX_BUF_HANDLE_SW_INDEX_BITS + IX_BUF_FL_ID_BITS)


/**
 * MACRO NAME: IX_RM_BUF_SW_GET_INDEX
 *
 * DESCRIPTION: This macro will retrieve the buffer index from the software buffer handle passed as a
 *          parameter.
 *
 * @Param:  - IN arg_hSwBuffer - handle to a software buffer.
 *
 * @Return: an ix_uint32 value representing the buffer index for the passed buffer handle.
 */
#define IX_RM_BUF_SW_GET_INDEX( \
                                arg_hSwBuffer \
                              ) \
                              IX_GET_BIT_FIELD32(arg_hSwBuffer, 1U, IX_BUF_HANDLE_SW_INDEX_BITS)


#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

/**
 * MACRO NAME: IX_RM_BUF_GET_FL_SRAM_CHANNEL
 *
 * DESCRIPTION: This macro will retrieve the buffer FL meta SRAM channel for the hardware buffer 
 *          handle passed as a parameter.
 *
 * @Param:  - IN arg_hHwBuffer - handle to a hardware buffer.
 *
 * @Return: an ix_uint32 value representing the SRAM FL meta memory channel for the passed buffer handle.
 */
#define IX_RM_BUF_GET_FL_SRAM_CHANNEL( \
                                       arg_hHwBuffer \
                                     ) \
                                     (IX_GET_BIT_FIELD32(arg_hHwBuffer, 1U, 2U))

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */

                                        
/**
 * MACRO NAME: IX_RM_BUF_HW_GET_META_OFFSET
 *
 * DESCRIPTION: This macro will retrieve the buffer meta offset for the hardware buffer handle passed as a
 *          parameter.
 *
 * @Param:  - IN arg_hHwBuffer - handle to a hardware buffer.
 *
 * @Return: an ix_uint32 value representing the buffer meta offset for the passed buffer handle.
 */
#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)


/**
 * In this case the last bit is always 0, and the bits 1 and 2 will represent
 * the SRAM channel number.
 */

#define IX_RM_BUF_HW_GET_META_OFFSET( \
                                      arg_hHwBuffer \
                                    ) \
                                    (IX_GET_BIT_FIELD32(arg_hHwBuffer, 3U, 23U) << 5U)

#else

#define IX_RM_BUF_HW_GET_META_OFFSET( \
                                      arg_hHwBuffer \
                                    ) \
                                    (IX_GET_BIT_FIELD32(arg_hHwBuffer, 0U, 23U) << 2U)

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */

/**
 * DESCRIPTION: Describes the SRAM channel where the meta data of the hardware buffers
 *      will be allocated.
 *
 */
#define IX_HW_FL_META_CHANNEL 0

/**
 * DESCRIPTION: Defines the hardware free list SRAM meta base address. That would be the start of
 *      different SRAM channels based on the SRAM channel where the buffers' meta will be created
 *
 */
#if (IX_HW_FL_META_CHANNEL == 0)

#define IX_HW_FL_META_BASE_ADDRESS IX_SRAM_CH0_START_ADDRESS

#elif (IX_HW_FL_META_CHANNEL == 1)
                                  
#define IX_HW_FL_META_BASE_ADDRESS IX_SRAM_CH1_START_ADDRESS

    #if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

        #if (IX_HW_FL_META_CHANNEL == 2)

        #define IX_HW_FL_META_BASE_ADDRESS IX_SRAM_CH2_START_ADDRESS

        #elif (IX_HW_FL_META_CHANNEL == 3)
                                  
        #define IX_HW_FL_META_BASE_ADDRESS IX_SRAM_CH3_START_ADDRESS

        #endif /* end (IX_HW_FL_META_CHANNEL == 2) */

    #endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */

#endif /* end (IX_HW_FL_META_CHANNEL == 0) */

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

extern ix_uint8* _g_aSRAMChannelsStartAddress[IX_SRAM_CHANNELS_NUMBER];

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */


/**
 * MACRO NAME: IX_RM_BUF_HW_GET_META_ADDRESS
 *
 * DESCRIPTION: This macro will retrieve the buffer meta address for the hardware buffer handle passed as a
 *          parameter.
 *
 * @Param:  - IN arg_hHwBuffer - handle to a hardware buffer.
 *
 * @Return: the meta address for the hardware buffer coresponding to the passed buffer handle.
 */
#if defined(_IX_RM_SPLIT_META_DATA_)

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

#define IX_RM_BUF_HW_GET_META_ADDRESS( \
                                       arg_hHwBuffer \
                                     ) \
                                     ((ix_hw_buffer_meta*)(((ix_hw_internal_buffer_meta*)(IX_RM_BUF_HW_GET_META_OFFSET(arg_hHwBuffer) + \
                                        _g_aSRAMChannelsStartAddress[(arg_hHwBuffer & 0x00000006) >> 1]))->m_pMeta))

#else

#define IX_RM_BUF_HW_GET_META_ADDRESS( \
                                       arg_hHwBuffer \
                                     ) \
                                     ((ix_hw_buffer_meta*)(((ix_hw_internal_buffer_meta*)(IX_RM_BUF_HW_GET_META_OFFSET(arg_hHwBuffer) + \
                                        (ix_uint8*)IX_HW_FL_META_BASE_ADDRESS))->m_pMeta))

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */

                                    
#else

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

#define IX_RM_BUF_HW_GET_META_ADDRESS( \
                                       arg_hHwBuffer \
                                     ) \
                                     ((ix_hw_buffer_meta*)(IX_RM_BUF_HW_GET_META_OFFSET(arg_hHwBuffer) + \
                                        _g_aSRAMChannelsStartAddress[(arg_hHwBuffer & 0x00000006) >> 1]))

#else

#define IX_RM_BUF_HW_GET_META_ADDRESS( \
                                       arg_hHwBuffer \
                                     ) \
                                     ((ix_hw_buffer_meta*)(IX_RM_BUF_HW_GET_META_OFFSET(arg_hHwBuffer) + (ix_uint8*)IX_HW_FL_META_BASE_ADDRESS))

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */


#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

#if defined(_IX_RM_SPLIT_META_DATA_)

/**
 * MACRO NAME: IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS
 *
 * DESCRIPTION: This macro will retrieve the buffer internal meta address for the hardware 
 *  buffer handle passed as a parameter.
 *
 * @Param:  - IN arg_hHwBuffer - handle to a hardware buffer.
 *
 * @Return: The internal meta address for the hardware buffer coresponding to the passed buffer handle.
 */

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

#define IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS( \
                                                arg_hHwBuffer \
                                              ) \
                                              ((ix_hw_internal_buffer_meta*)(IX_RM_BUF_HW_GET_META_OFFSET(arg_hHwBuffer) + \
                                               _g_aSRAMChannelsStartAddress[(arg_hHwBuffer & 0x00000006) >> 1]))

#else

#define IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS( \
                                                arg_hHwBuffer \
                                              ) \
                                              ((ix_hw_internal_buffer_meta*)(IX_RM_BUF_HW_GET_META_OFFSET(arg_hHwBuffer) + (ix_uint8*)IX_HW_FL_META_BASE_ADDRESS))

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

 

/**
 * MACRO NAME: IX_RM_BUF_HW_GET_FL_ID
 *
 * DESCRIPTION: This macro will retrieve the FL ID of the FL that owns the buffer passed as a
 *          parameter. We are going to use the IX_RM_MEM_UINT16_READ macro to read the m_BufferInfo
 *          member of the hardware buffer meta for Foreign Model portability.
 *
 * @Param:  - IN arg_hHwBuffer - handle to a hardware buffer.
 *
 * @Return: an ix_uint32 value representing the FL ID of the FL that owns this buffer.
 */

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

#define IX_RM_BUF_HW_GET_FL_ID( \
                               arg_hHwBuffer \
                             ) \
                             ((IX_RM_HW_META_GET_FREE_LIST_ID(IX_RM_MEM_UINT16_READ(&( \
                             IX_RM_BUF_HW_GET_META_ADDRESS(arg_hHwBuffer)->m_BufferInfo)))) \
                             + IX_RM_BUF_GET_FL_SRAM_CHANNEL(arg_hHwBuffer) * IX_BUF_MAX_HW_FL_NUMBER)

#else

#define IX_RM_BUF_HW_GET_FL_ID( \
                               arg_hHwBuffer \
                             ) \
                             (IX_RM_HW_META_GET_FREE_LIST_ID(IX_RM_MEM_UINT16_READ(&( \
                             IX_RM_BUF_HW_GET_META_ADDRESS(arg_hHwBuffer)->m_BufferInfo))))

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */

                                      


/**
 * MACRO NAME: IX_RM_BUF_GET_FL_ADDRESS
 *
 * DESCRIPTION: This macro will return the address of the FL that owns the buffer
 *          passed as a parameter
 *
 * @Param:  - IN arg_hBuffer - handle to a hardware buffer.
 *
 * @Return: The address of the FL that owns the buffer passed as a parameter
 */
#define IX_RM_BUF_GET_FL_ADDRESS( \
                                  arg_hBuffer \
                                ) \
                                ((_ix_buffer_free_list*)g_FLManager.m_aFreeListAddress[(IX_RM_BUFFER_GET_TYPE(arg_hBuffer) ? \
                                IX_RM_BUF_SW_GET_FL_ID(arg_hBuffer) : \
                                IX_RM_BUF_HW_GET_FL_ID(arg_hBuffer))])



/**
 * MACRO NAME: IX_RM_BUF_GET_EOP
 *
 * DESCRIPTION: This macro will return a non zero value if the buffer is the last in a list of
 *      concatenated buffers (end of a packet). 
 *
 * @Param:  - IN arg_hHwBuffer - handle to a buffer.
 *
 * @Return: Returns an ix_uint32 value that will be non 0 if the buffer is the last in a list of
 *      concatenated buffers (end of a packet).
 */
#define IX_RM_BUF_GET_EOP( \
                           arg_hBuffer \
                         ) \
                         IX_GET_BIT_FIELD32(arg_hBuffer, 31U, 31U)

/**
 * MACRO NAME: IX_RM_BUF_SET_EOP
 *
 * DESCRIPTION: This macro will set the EOP flag to a buffer handle. 
 *
 * @Param:  - IN arg_hBuffer - handle to a buffer.
 * @Param:  - IN arg_EOPFlag - the EOP flag. The value range is 0..1.
 *
 * @Return: Returns the updated buffer handle
 */
#define IX_RM_BUF_SET_EOP( \
                           arg_hBuffer, \
                           arg_EOPFlag \
                         ) \
                         IX_SET_BIT_FIELD32(arg_hBuffer, arg_EOPFlag, 31U, 31U)


/**
 * MACRO NAME: IX_RM_BUF_GET_SOP
 *
 * DESCRIPTION: This macro will return a non zero value if the buffer is the first in a list of
 *      concatenated buffers (end of a packet). 
 *
 * @Param:  - IN arg_hBuffer - handle to a buffer.
 *
 * @Return: Returns an ix_uint32 value that will be non 0 if the buffer is the first in a list of
 *      concatenated buffers (end of a packet).
 */
#define IX_RM_BUF_GET_SOP( \
                           arg_hBuffer \
                         ) \
                         IX_GET_BIT_FIELD32(arg_hBuffer, 30U, 30U)



/**
 * MACRO NAME: IX_RM_BUF_SET_SOP
 *
 * DESCRIPTION: This macro will set the SOP flag to a buffer handle. 
 *
 * @Param:  - IN arg_hBuffer - handle to a buffer.
 * @Param:  - IN arg_SOPFlag - the SOP flag. The value range is 0..1.
 *
 * @Return: Returns the updated buffer handle
 */
#define IX_RM_BUF_SET_SOP( \
                           arg_hBuffer, \
                           arg_SOPFlag \
                         ) \
                         IX_SET_BIT_FIELD32(arg_hBuffer, arg_SOPFlag, 30U, 30U)




/**
 * DESCRIPTION: This mask determines if a buffer is part of a chain or not
 *
 */
#define IX_RM_SINGLE_BUFFER_MASK    0xc0000000

                                                                                     

/**
 * MACRO NAME: IX_RM_BUF_SW_MAKE_HANDLE
 *
 * DESCRIPTION: This macro will create a new software buffer handle based on the free list's
 *          index and the the buffer index inside that free list.
 *
 * @Param:  - IN arg_FLID - an ix_uint32 value representing the free list's ID that owns
 *          this buffer.
 * @Param:  - IN arg_BufferIndex - an ix_uint32 value representing the buffer index.
 *
 * @Return: An ix_buffer_handle type value representing the newly created handle.
 */
#define IX_RM_BUF_SW_MAKE_HANDLE( \
                                  arg_FLID, \
                                  arg_BufferIndex \
                                ) \
                                ((ix_buffer_handle) /* typecast */ \
                                IX_MAKE_BIT_FIELD32(1U, 31U, 31U) | /* EOP flag */ \
                                IX_MAKE_BIT_FIELD32(1U, 30U, 30U) | /* SOP flag */ \
                                /* free list id bit field */ \
                                IX_MAKE_BIT_FIELD32(arg_FLID, IX_BUF_HANDLE_SW_INDEX_BITS + 1U, \
                                IX_BUF_HANDLE_SW_INDEX_BITS + IX_BUF_FL_ID_BITS) | \
                                IX_MAKE_BIT_FIELD32(arg_BufferIndex, 1U, IX_BUF_HANDLE_SW_INDEX_BITS) | \
                                IX_MAKE_BIT_FIELD32(1U, 0U, 0U)  /* buffer type flag */ \
                                )

/**
 * MACRO NAME: IX_RM_BUF_HW_MAKE_HANDLE
 *
 * DESCRIPTION: This macro will create a new hardware buffer handle based on the meta address
 *          expressed as a longword offset into the SRAM memory channel 0.
 *
 * @Param:  - IN arg_LongWordOffset - an ix_uint32 value representing the longword offset into the 
 *          SRAM memory channel 0 of the buffer's meta data.
 *
 * @Return: An ix_buffer_handle type value representing the newly created handle.
 */
#define IX_RM_BUF_HW_MAKE_HANDLE( \
                                  arg_LongWordOffset \
                                ) \
                                ((ix_buffer_handle) /* typecast */ \
                                IX_MAKE_BIT_FIELD32(0x000000c0U, 24U, 31U) | /* set EOP|SOP and buffer count to 0xc0 */ \
                                IX_MAKE_BIT_FIELD32(arg_LongWordOffset, 0U, 23U)  /* meta longword offset */ \
                                )



/**
 * DESCRIPTION: This value will used to build a free list handle
 *
 */
#define IX_BUF_FL_HANDLE_SIGNATURE  0xfa319cd1



/**
 * MACRO NAME: IX_RM_BUF_FL_MAKE_HANDLE
 *
 * DESCRIPTION: This macro will create a FL handle based on the passed FL ID. 
 *
 * @Param:  - IN arg_FLID - this is a ix_uint32 value representing the FL ID 
 *          to build a FL handle upon.
 *
 * @Return: returns a ix_buffer_free_list_handle coresponding to the passed ID.
 */
#define IX_RM_BUF_FL_MAKE_HANDLE( \
                                 arg_FLID \
                               ) \
                               ((ix_buffer_free_list_handle) /* typecast */ \
                               (IX_BIT_FIELD_MASK32(IX_BUF_FL_ID_BITS, 31U) & IX_BUF_FL_HANDLE_SIGNATURE)| \
                               IX_MAKE_BIT_FIELD32(arg_FLID, 0U, IX_BUF_FL_ID_BITS - 1U))                   
                               



/**
 * MACRO NAME: IX_RM_BUF_FL_GET_FL_ID
 *
 * DESCRIPTION: This macro will extract the FL ID out of a FL handle
 *
 * @Param:  - IN arg_hFLHandle - a ix_buffer_free_list_handle that represents the
 *          FL handle of interest.
 *
 * @Return: returns an ix_uint32 value representing the FL ID.
 */
#define IX_RM_BUF_FL_GET_FL_ID( \
                               arg_hFLHandle \
                             ) \
                             IX_GET_BIT_FIELD32(arg_hFLHandle, 0U, IX_BUF_FL_ID_BITS - 1U)



/**
 * MACRO NAME: IX_RM_BUF_FL_IS_VALID
 *
 * DESCRIPTION: This macro will check if the passed free list handle is valid
 *
 * @Param:  - IN arg_hFLHandle - a ix_buffer_free_list_handle that represents the
 *          FL handle of interest.
 *
 * @Return: Returns non zero if the handle is valid or zero otherwise.
 */
#define IX_RM_BUF_FL_IS_VALID( \
                               arg_hFLHandle \
                             ) \
                             ((((IX_BUF_FL_HANDLE_SIGNATURE ^ (arg_hFLHandle)) & \
                                IX_BIT_FIELD_MASK32(IX_BUF_FL_ID_BITS, 31U)) == 0) ? 1 : 0)


/**
 * This is a forward declaration
 */
typedef struct _ix_s_buffer_free_list _ix_buffer_free_list;



/**
 * TYPENAME: _ix_buffer_free_list_fini_ptr
 * 
 * DESCRIPTION: This type defines a buffer free list finalizer function.
 *          The function will take as the first parameter a pointer to the free list
 *          to work on. This function will perform the custom finalization step for the
 *          buffer free list. This will be done before the common finalization has
 *          been performed.
 *
 */
typedef 
ix_error (* _ix_buffer_free_list_fini_ptr)( 
                                            _ix_buffer_free_list* arg_pFreeList 
                                          );



/**
 * TYPENAME: _ix_mc_cell_allocator_ptr
 * 
 * DESCRIPTION: This type defines a buffer free list buffer allocator function.
 *          The function will take as the first parameter a pointer to the free list
 *          to work on. The second parameter will be the location where the allocated 
 *          buffer handle should be stored upon return.
 *
 */
typedef 
ix_error (* _ix_buffer_allocator_ptr)( 
                                       _ix_buffer_free_list* arg_pFreeList, 
                                       ix_buffer_handle* arg_pBufferHandle
                                     );



/**
 * TYPENAME: _ix_buffer_deallocator_ptr
 * 
 * DESCRIPTION: This type defines a buffer free list buffer deallocator function.
 *          The function will take as the first parameter a pointer to the free list
 *          to work on. The second parameter will be the handle of the buffer to be released.
 *
 */
typedef 
ix_error (* _ix_buffer_deallocator_ptr)( 
                                         _ix_buffer_free_list* arg_pFreeList, 
                                         ix_buffer_handle arg_hBuffer
                                       );


/**
 * TYPENAME: _ix_buffer_free_list_vmt
 * 
 * DESCRIPTION: This structure is the equivalent of a virtual method table in C++.
 *          Each buffer free list will have an associated VMT that will control behaviour
 *          of that buffer free list. This will allow for a great flexibility and 
 *          different free lists can be implemented in the same way. We will be able to implement
 *          the common behaviour in the same place and we abstract the different behaviour and
 *          let users to customize that different behaviour. It's the idea stated by the 
 *          template method design pattern.  The finalizer will perform
 *          the custom finalization step before the common finalization is to be performed.
 *          We believed that is good to abstract the buffer allocator and deallocator behaviour
 *          in order to allow different implementations.
 *
 */
typedef struct _ix_buffer_free_list_vmt
{
    _ix_buffer_free_list_fini_ptr   m_FiniFuncPtr; /* free list finalizer */
    _ix_buffer_allocator_ptr        m_AllocatorFuncPtr; /* buffer allocator */
    _ix_buffer_deallocator_ptr      m_DeallocatorFuncPtr; /* buffer deallocator */
} _ix_buffer_free_list_vmt;



/**
 * TYPENAME: _ix_sw_free_list_data
 * 
 * DESCRIPTION: This structure defines the data specific just to software free lists.
 *
 */
typedef struct _ix_s_sw_free_list_data
{
    ix_uint32                       m_NumFreeElements; /* number of elements in the free list */

    /* this flag signals if the PM has been requested to be deleted */
    ix_uint32                       m_MarkedForDeletion; 

#if defined( _IX_RM_COMPONENT_MULTI_THREADED_ )

    ix_synchronization_adapter      m_SyncAdapter; /* synchronization adapter */

#endif /* defined( _IX_RM_COMPONENT_MULTI_THREADED_ ) */

} _ix_sw_free_list_data;



/**
 * TYPENAME: _ix_hw_free_list_data
 * 
 * DESCRIPTION: This structure defines the data specific just to hardware free lists.
 *
 */
typedef struct _ix_s_hw_free_list_data
{
    ix_uint32       m_SRAMQueueId;

#if defined(_IX_RM_SPLIT_META_DATA_)

    ix_uint32*                      m_pInternalMetaBaseAddress; /* internal meta block base address */
    ix_memory_type                  m_InternalMetaMemoryType; /* internal meta block memory type */
    ix_uint32                       m_InternalMetaMemoryChannel; /* internal meta block memory channel */
    ix_uint32                       m_InternalMetaElementShift; /* internal meta element shift value */

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter      m_SyncAdapter; /* synchronization adapter */

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

} _ix_hw_free_list_data;


/**
 * TYPENAME: _ix_free_list_data
 * 
 * DESCRIPTION: This union describes type specific (sw or hw) free list data
 *
 */
typedef union _ix_u_free_list_data
{
    _ix_hw_free_list_data   m_HwData; /* hardware specific data */
    _ix_sw_free_list_data   m_SwData; /* software specific data */
} _ix_free_list_data;

#if defined (_IX_RM_BUFFER_DEBUG_)

/**
 * TYPENAME: ix_buffer_last_core_allocation_status
 * 
 * DESCRIPTION: This enumeration represents the last allocation status of a buffer on the core
 *  side.
 *
 */
typedef enum ix_e_buffer_last_core_allocation_status
{
    IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_FIRST = 0,
    IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_FREE = IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_FIRST,
    IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_ALLOCATED_BY_CORE,
    IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_RECEIVED_FROM_ME_AS_PACKET,
    IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_RECEIVED_FROM_ME_AS_MESSAGE,
    IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_LAST
} ix_buffer_last_core_allocation_status;


/**
 * TYPENAME: ix_buffer_debug_info
 * 
 * DESCRIPTION: This structure contains debug information for the buffers
 *
 */
typedef struct ix_s_buffer_debug_info
{
    ix_buffer_last_core_allocation_status m_LastCoreAllocationStatus;
    ix_uint32 m_SendAsPacketCount;
    ix_uint32 m_SendAsMessageCount;
    ix_buffer_handle m_hBuffer; /* Placeholder for the buffer handle */
} ix_buffer_debug_info;


#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

/**
 * TYPENAME: _ix_s_buffer_free_list
 * 
 * DESCRIPTION: This type defines a buffer free list structure.
 *
 */
struct _ix_s_buffer_free_list
{
    ix_uint32*                      m_pMetaBaseAddress; /* meta block base address */
    ix_memory_type                  m_MetaMemoryType; /* meta block memory type */
    ix_uint32                       m_MetaMemoryChannel; /* meta block memory channel */
    ix_uint32                       m_MetaElementShift; /* meta element shift value */

    ix_uint32*                      m_pDataBaseAddress; /* data block base address */
    ix_memory_type                  m_DataMemoryType; /* data block memory type */
    ix_uint32                       m_DataMemoryChannel; /* data block memory channel */
    ix_uint32                       m_DataElementShift; /* data element shift value */

    ix_uint32                       m_NumElements; /* number of elements */

    _ix_buffer_free_list_vmt        m_VMT; /* free list function table */

    ix_buffer_handle                m_FreeListHead; /* head of the free list */

    _ix_free_list_data              m_FLData; /* type specific data */

#if defined (_IX_RM_BUFFER_DEBUG_)

    ix_buffer_debug_info*           m_pDebugInfo; /* debug information for the buffers */

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

};



/**
 * TYPENAME: _ix_buffer_free_list_manager
 * 
 * DESCRIPTION: This structure represents the free list manager. It manages both
 *          the hardware and software end free lists.
 *
 */
typedef struct _ix_s_buffer_free_list_manager
{
    _ix_buffer_free_list*           m_aFreeListAddress[IX_BUF_MAX_FL_NUMBER];

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

    _ix_buffer_free_list            m_aHwFreeList[IX_SRAM_CHANNELS_NUMBER][IX_BUF_MAX_HW_FL_NUMBER];

#else

    _ix_buffer_free_list            m_aHwFreeList[IX_BUF_MAX_HW_FL_NUMBER];

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */

    _ix_buffer_free_list            m_aSwFreeList[IX_BUF_MAX_SW_FL_NUMBER];

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter      m_SyncAdapter; /* synchronization adapter */

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

} _ix_buffer_free_list_manager;



#if defined( _IX_RM_MULTI_THREADED_ )

/**
 * NAME: _ix_rm_fl_manager_init_sync
 *
 * DESCRIPTION: This function will initialize the synchronization for the free list
 *          manager.
 *          A detailed explanation on how sychronization mechanism works is found in 
 *          the L0 HLD/LLD and also in sync_mechanism.h file.
 * 
 * @Param:  - IN arg_LockFunction - this is the lock function of the sync mechanism.
 * @Param:  - IN arg_UnlockFunction - this is the unlock function of the sync mechanism.
 * @Param:  - IN arg_pContext - context that will be passed to each invocation of lock
 *          and unlock function. It might point to a synchronization primitive to be
 *          used by the specified lock and unlock functions.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_fl_manager_init_sync(
                                      ix_sa_lock_ptr            arg_LockFunction,
                                      ix_sa_unlock_ptr          arg_UnlockFunction,
                                      void*                     arg_pContext
                                    );


#endif /* defined( _IX_RM_MULTI_THREADED_ ) */




/**
 * The following are the default hardware free list VMT functions.
 */

/**
 * NAME: _ix_rm_hw_free_list_fini
 *
 * DESCRIPTION: This function will perform the proper cleanup for a hardware free list.
 * 
 * @Param:  - IN arg_pHwFreeList - address of the hardware free list structure this function
 *          will work on.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_hw_free_list_fini( 
                                   _ix_buffer_free_list* arg_pHwFreeList 
                                 );



/**
 * NAME: _ix_rm_hw_buffer_allocator
 *
 * DESCRIPTION: This function will get a new hardware buffer from the hardware free list.
 * 
 * @Param:  - IN arg_pHwFreeList - address of the hardware free list structure this function
 *          will work on.
 * @Param:  - OUT arg_pHwBufferHandle - address where the handle of the newly allocated 
 *          hardware buffer will be stored.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_hw_buffer_allocator( 
                                     _ix_buffer_free_list* arg_pHwFreeList, 
                                     ix_buffer_handle* arg_pHwBufferHandle
                                   );



/**
 * NAME: _ix_rm_hw_buffer_deallocator
 *
 * DESCRIPTION: This function will return the specified hardware buffer to the free list.
 * 
 * @Param:  - IN arg_pHwFreeList - address of the hardware free list structure this function
 *          will work on.
 * @Param:  - IN arg_hHwBuffer - the handle of the hardware buffer to be deleted.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_hw_buffer_deallocator( 
                                       _ix_buffer_free_list* arg_pHwFreeList, 
                                       ix_buffer_handle arg_hHwBuffer
                                     );



/**
 * The following are the default software free list VMT functions.
 */

/**
 * NAME: _ix_rm_sw_free_list_fini
 *
 * DESCRIPTION: This function will perform the proper cleanup for a software free list.
 * 
 * @Param:  - IN arg_pSwFreeList - address of the software free list structure this function
 *          will work on.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_sw_free_list_fini( 
                                   _ix_buffer_free_list* arg_pSwFreeList 
                                 );



/**
 * NAME: _ix_rm_sw_buffer_allocator
 *
 * DESCRIPTION: This function will get a new software buffer from the software free list.
 * 
 * @Param:  - IN arg_pSwFreeList - address of the software free list structure this function
 *          will work on.
 * @Param:  - OUT arg_pSwBufferHandle - address where the handle of the newly allocated 
 *          software buffer will be stored.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_sw_buffer_allocator( 
                                     _ix_buffer_free_list* arg_pSwFreeList, 
                                     ix_buffer_handle* arg_pSwBufferHandle
                                   );



/**
 * NAME: _ix_rm_sw_buffer_deallocator
 *
 * DESCRIPTION: This function will return the specified software buffer to the free list.
 * 
 * @Param:  - IN arg_pSwFreeList - address of the software free list structure this function
 *          will work on.
 * @Param:  - IN arg_hSwBuffer - the handle of the software buffer to be deleted.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_sw_buffer_deallocator( 
                                     _ix_buffer_free_list* arg_pSwFreeList, 
                                     ix_buffer_handle arg_hSwBuffer
                                   );



/**
 * NAME: _ix_rm_buffer_init
 *
 * DESCRIPTION: This function should be called at the time the resource manager initializer 
 *          function is called. The function will initialize all the Buffer API.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_buffer_init(void);


/**
 * NAME: _ix_rm_buffer_fini
 *
 * DESCRIPTION: This function should be called at the time the resource manager finalizer function
 *          is called. The function will provide cleanup for Buffer API.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_buffer_fini(void);


#if defined (_IX_RM_BUFFER_DEBUG_)



/**
 * NAME: _ix_rm_buffer_debug_info_get
 *
 * DESCRIPTION: This function will retrieve the debug info coresponding to a buffer
 * 
 * @Param:  - IN arg_hBuffer - the handle to the buffer whose debug info we are trying to
 *  retrieve
 * @Param:  - OUT arg_pDebugInfo - an address where the pointer to the corresponding debug
 *  info will be stored on return
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_buffer_debug_info_get(
                                       ix_buffer_handle arg_hBuffer,
                                       ix_buffer_debug_info** arg_pDebugInfo 
                                     );


/**
 * NAME: _ix_rm_free_list_get_available_buffers
 *
 * DESCRIPTION: This function will 
 * 
 * @Param:  - IN arg_hFreeList - the free list whose number of available buffers
 *  we are interested in.
 * @Param:  - OUT arg_pNoAvailableBuffers - address where on return the number of 
 *  available buffers will be stored upon return
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_free_list_get_available_buffers(
                                                 ix_buffer_free_list_handle arg_hFreeList, 
                                                 ix_uint32* arg_pNoAvailableBuffers 
                                               );


#endif /* defined (_IX_RM_BUFFER_DEBUG_) */



/**
 * This global variable represents an array of available free lists in this system.
 */
extern _ix_buffer_free_list_manager g_FLManager;



#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

/**
 * SRAM QArray entry ID array coresponding the first buffer free list
 * These values will need to be patched into the uCode.
 */
extern ix_uint32 g_StartBufferFreeListQArrayID[IX_SRAM_CHANNELS_NUMBER];

#else
 
/**
 * SRAM QArray entry ID on the active channel coresponding the first buffer free list
 * This value will need to be patched into the uCode.
 */
extern ix_uint32 g_StartBufferFreeListQArrayID;

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */


#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__INTERNAL_BUFFER_H__) */
