/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 4
 *
 * = LIBRARY
 *      Resource Manager Library
 *
 * = MODULE
 *      Communication API
 *
 * = FILENAME
 *      internal_communication.h
 *
 * = DESCRIPTION
 *      This file defines all data types and function prototypes required internaly by the 
 *      Communication API module.
 *
 * = AUTHOR
 *       Adrian Georgescu
 *       adrian.georgescu@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      10/26/2001 11:27:44 AM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 * $Id: internal_communication.h,v 1.40 2003/09/19 23:11:52 ageorges Exp $
 */

#if !defined(__INTERNAL_COMMUNICATION_H__)
#define __INTERNAL_COMMUNICATION_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */


#if defined(_IX_COMM_ID_NOTIFICATION_DISCONNECT_ENABLE_)

/**
 * DESCRIPTION: This symbol represents the user data value to be
 *  passed for disconnect notification along with IX_NULL_BUFFER_HANDLE to the
 *  registered data handler.
 */
#define IX_COMM_ID_NOTIFICATION_DISCONNECT      IX_NOTIFICATION_MESSAGE  + 1UL

#endif /* end defined(_IX_COMM_ID_NOTIFICATION_DISCONNECT_ENABLE_) */


/**
 * NAME: _ix_rm_comm_data_drop
 *
 * DESCRIPTION: This function will be the default processing function for each
 *          communication ID. This will return the buffers to the owning free list.
 * 
 * @Param:  - IN arg_hBuffer - the handle of the buffer to be processed
 * @Param:  - IN arg_UserData - the user data associated with the buffer to process
 * @Param:  - IN arg_pContext - context that has been passed to the comm ID at the 
 *          default handler registration time.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_comm_data_drop( 
                                ix_buffer_handle arg_hBuffer, 
                                ix_uint32 arg_UserData,
                                void* arg_pContext 
                              );

/**
 * NAME: _ix_rm_comm_ublock_packet_handler
 *
 * DESCRIPTION: This function will be the the internal processing function for the uBlock
 *  packet end point. This function will store the buffer handle and user data onto the
 *  associated end point ring which will passed through arg_pContext argument.
 * 
 * @Param:  - IN arg_hBuffer - the handle of the buffer to be processed
 * @Param:  - IN arg_UserData - the user data associated with the buffer to process
 * @Param:  - IN arg_pContext - context that has been passed to the comm ID at the 
 *          default handler registration time.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_comm_ublock_packet_handler( 
                                ix_buffer_handle arg_hBuffer, 
                                ix_uint32 arg_UserData,
                                void* arg_pContext 
                              );


/**
 * NAME: _ix_rm_comm_ublock_message_handler
 *
 * DESCRIPTION: This function will be the the internal processing function for the uBlock
 *  message end point. This function will store the buffer handle and user data onto the
 *  associated end point ring which will passed through arg_pContext argument.
 * 
 * @Param:  - IN arg_hBuffer - the handle of the buffer to be processed
 * @Param:  - IN arg_UserData - the user data associated with the buffer to process
 * @Param:  - IN arg_pContext - context that has been passed to the comm ID at the 
 *          default handler registration time.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_comm_ublock_message_handler( 
                                ix_buffer_handle arg_hBuffer, 
                                ix_uint32 arg_UserData,
                                void* arg_pContext 
                              );

/**
 * TYPENAME: ix_comm_handler_type
 * 
 * DESCRIPTION: This internal type  defines all possible handler types 
 *
 */
typedef enum _ix_e_comm_handler_type
{
    IX_COMM_HANDLER_TYPE_FIRST = 0,
    IX_COMM_HANDLER_TYPE_PACKET = IX_COMM_HANDLER_TYPE_FIRST,
    IX_COMM_HANDLER_TYPE_MESSAGE,
    IX_COMM_HANDLER_TYPE_LAST
} _ix_comm_handler_type;



/**
 * TYPENAME: ix_send_call_type
 * 
 * DESCRIPTION: This enumeration describes the two possible send call types: blocking
 *  and nonblocking. For the blocking calls then a valid timeout should be passed to
 *  the behaviour function for the endpoint. This parameter should be passed immediately
 *  after standard parameters for the behaviour function. The next parameter should be
 *  the timeout.
 *
 */
typedef enum ix_e_send_call_type
{
    IX_SEND_CALL_TYPE_FIRST = 0,
    IX_SEND_CALL_TYPE_NONBLOCK = IX_SEND_CALL_TYPE_FIRST,
    IX_SEND_CALL_TYPE_BLOCK,
    IX_SEND_CALL_TYPE_LAST
} ix_send_call_type;



/**
 * TYPENAME: _ix_comm_ep_behaviour_handler
 * 
 * DESCRIPTION: This type has been introduced to allow better flexibility on the
 *  send side of communication.
 */
typedef ix_error (*_ix_comm_ep_behaviour_handler)( 
                                                   ix_buffer_handle arg_hBuffer, 
                                                   ix_uint32 arg_UserData,
                                                   void* arg_pContext,
                                                   ...
                                                 );


/**
 * NAME: _ix_rm_handler_register
 *
 * DESCRIPTION: This function will register a data handler for the 
 *	specified communication ID. If the 
 *  _IX_COMM_ID_NOTIFICATION_DISCONNECT_ENABLE_ was defined at the 
 *  time of RM compilation, if another data handler is in place, that 
 *	data handler will be invoked with a IX_NULL_BUFFER_HANDLE as the 
 *	first parameter, and IX_COMM_ID_NOTIFICATION_DISCONNECT value as the 
 *	user data parameter. From this point on, the new registered handler 
 *	will be used for processing data sent to this destination. If this 
 *	communication ID is in get/select mode, the handler will be invoked with 
 *	all unprocessed data from the internal queue, and all waiting tasks 
 *	will be awaken and signaled that they have been disconnected from the 
 *	communication Id. This function can be used for both packet and message
 *  handlers through arg_HandlerType parameter 
 * 
 * @Param:  - IN arg_CommunicationId - the communication ID for which we are 
 *	registering the data handler.
 * @Param:  - IN arg_HandlerType - the type of data handler we want to register.
 *  The posibilities are packet and message.
 * @Param:  - IN arg_DataHandler - the data handler that we want to 
 *	register.
 * @Param:  - IN arg_pContext - context data that will be passed to the 
 *	data handler on each invocation. The programmer has to make sure that 
 *	the context data is valid as long the packet handler is registered.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error _ix_rm_handler_register( 
                                  ix_communication_id arg_CommunicationId, 
                                  _ix_comm_handler_type arg_HandlerType,
                                  _ix_comm_ep_behaviour_handler arg_DataHandler, 
                                  void* arg_pContext 
                                );


/**
 * DESCRIPTION: This symbol defines the number of 32 bit dwords the ME
 *          will write on the hardware ring for each exception packet passed
 *          to the core.
 */
#define IX_UBLOCK_PACKET_DATA_SIZE    2UL


/**
 * DESCRIPTION: This symbol defines the number of 32 bit dwords the ME
 *          will write on the hardware ring for each message passed
 *          to the core.
 */
#define IX_UBLOCK_MESSAGE_DATA_SIZE   2UL


/**
 * DESCRIPTION: This symbol defines the size of uBlock/core packet data ring size
 *
 */
#define IX_UBLOCK_PACKET_DATA_RING_SIZE IX_SCRATCH_RING_SIZE_128


/**
 * DESCRIPTION: This symbol defines the size of uBlock/core message data ring size
 *
 */
#define IX_UBLOCK_MESSAGE_DATA_RING_SIZE IX_SCRATCH_RING_SIZE_128



/**
 * DESCRIPTION: This is the predefined scratch ring ID for the uBlock/core packet 
 *  communication ring.
 *
 */
#define IX_UBLOCK_PACKET_RING_ID    0UL


/**
 * DESCRIPTION: This is the predefined scratch ring ID for the uBlock/core message 
 *  communication ring.
 *
 */
#define IX_UBLOCK_MESSAGE_RING_ID    1UL


/**
 * DESCRIPTION: This symbol defines the number of 32 bit dwords a CC
 *          will write on the comm Id ring for each packet passed.
 */
#define IX_CORE_PACKET_DATA_SIZE_EXP   1
#define IX_CORE_PACKET_DATA_SIZE    (1UL << IX_CORE_PACKET_DATA_SIZE_EXP)


/**
 * DESCRIPTION: This symbol defines the number of 32 bit dwords a CC
 *          will write on the comm Id ring for each message passed.
 */
#define IX_CORE_MESSAGE_DATA_SIZE_EXP   1
#define IX_CORE_MESSAGE_DATA_SIZE   (1UL << IX_CORE_MESSAGE_DATA_SIZE_EXP)


/**
 * DESCRIPTION: This symbol defines the size of core/core packet data ring size
 *  This value MUST be a power of 2 at all times
 */
#define IX_CORE_PACKET_DATA_RING_SIZE 256


/**
 * DESCRIPTION: This symbol defines the size of core/core message data ring size
 *  This value MUST be a power of 2 at all times
 */
#define IX_CORE_MESSAGE_DATA_RING_SIZE 128



/**
 * DESCRIPTION: This symbol defines the number of 32 bit dwords that will
 *  written on the PCI end-point ring for each packet passed.
 *
 */
#define IX_CORE_PCI_PACKET_DATA_SIZE    3



/**
 * DESCRIPTION: This symbol defines the number of 32 bit dwords that will
 *  written on the PCI end-point ring for each message passed.
 *
 */
#define IX_CORE_PCI_MESSAGE_DATA_SIZE    3


/**
 * DESCRIPTION: This symbol defines the size of PCI packet data ring size
 *  This value MUST be a power of 2 at all times
 */
#define IX_CORE_PCI_PACKET_DATA_RING_SIZE 128


/**
 * DESCRIPTION: This symbol defines the size of PCI message data ring size
 *  This value MUST be a power of 2 at all times
 */
#define IX_CORE_PCI_MESSAGE_DATA_RING_SIZE 128


/**
 * TYPENAME: ix_wait_notification_node
 * 
 * DESCRIPTION: This structure defines a flow control notification node. It will be used
 *          to store notification tokens for the case when a data queue goes out of the
 *          empty or full states.
 *
 */
typedef struct ix_s_wait_notification_node ix_wait_notification_node;

struct ix_s_wait_notification_node
{
    ix_ossl_sem_t               m_Semaphore; /* semaphore registered to be signaled */
    ix_wait_notification_node*  m_pNext; /* next notification node in the list */
};



/**
 * TYPENAME: ix_end_point_data_queue
 * 
 * DESCRIPTION: This structure defines an end point data queue which is implemented as 
 *          a circular ring. The queue holds no data and has to be initialized externally
 *          in order to function. As it is, the queue can be used only in a single threaded
 *          environment, but external synchronization and flow control can make it work in
 *          a multithreaded multi producer/multi consumer environment.
 *          The design of this ring queue was done with the processing speed in mind.
 *
 */

typedef struct ix_s_end_point_data_queue
{
    volatile ix_uint32  m_Put;
    volatile ix_uint32  m_Get;

    ix_uint32           m_Size; /* size will be at all times a power of 2 */
    ix_uint32           m_Capacity; /* capacity will be size - 1 */
    ix_uint32           m_ElementSize; /* element size expressed in bytes */

    ix_uint32*          m_pStartAddress; /* start of the memory block controlled by the ring */

} ix_end_point_data_queue;



/**
 * MACRO NAME: IX_EP_DATA_QUEUE_IS_FULL
 *
 * DESCRIPTION: This macro will determine if the internal end point data queue is full or
 *  not. The data queue is declared full if there is just one space left in the queue.
 *
 * @Param:  - IN arg_pEPDataQueue - the address of the end point data queue whose status
 *  we are interested in
 *
 * @Return: Returns true if the queue is full and false if it is not
 */
#define IX_EP_DATA_QUEUE_IS_FULL( \
                                  arg_pEPDataQueue \
                                ) \
                                (((arg_pEPDataQueue)->m_Put - (arg_pEPDataQueue)->m_Get) == (arg_pEPDataQueue)->m_Capacity)

/**
 * MACRO NAME: IX_EP_DATA_QUEUE_IS_EMPTY
 *
 * DESCRIPTION: This macro will determine if the internal end point data queue is empty or
 *  not. 
 *
 * @Param:  - IN arg_pEPDataQueue - the address of the end point data queue whose status
 *  we are interested in
 *
 * @Return: Returns true if the queue is empty and false if it is not
 */
#define IX_EP_DATA_QUEUE_IS_EMPTY( \
                                   arg_pEPDataQueue \
                                 ) \
                                 ((arg_pEPDataQueue)->m_Get == (arg_pEPDataQueue)->m_Put)

/**
 * MACRO NAME: IX_EP_DATA_QUEUE_ELEMENTS_AVAILABLE
 *
 * DESCRIPTION: This macro will determine how many elements are available in the internal
 *  data queue.
 *
 * @Param:  - IN arg_pEPDataQueue - the address of the end point data queue whose status
 *  we are interested in
 *
 * @Return: Returns the number of elements available in the data queue
 */
#define IX_EP_DATA_QUEUE_ELEMENTS_AVAILABLE( \
                                             arg_pEPDataQueue \
                                           ) \
                                           ((arg_pEPDataQueue)->m_Put - (arg_pEPDataQueue)->m_Get)


                                
/**
 * DESCRIPTION: Mask that describes the bit specifying the comm Id receive mode mask.
 *  If the value is 0 then we are in callback receive mode. If the value is 1 then we 
 *  are in the buffered get/select mode.
 *
 */
#define IX_COMM_ID_RECEIVE_MODE_MASK    0x1


/**
 * DESCRIPTION: Mask that describes the bit specifying the comm Id producer mode mask.
 *  If the value is 0 then we are in the single producer mode. If it is 1 then we are
 *  in the multiple producer mode.
 */
#define IX_COMM_ID_PRODUCER_MODE_MASK   0x2


/**
 * DESCRIPTION: Mask that describes the bit specifying the comm Id consumer mode mask.
 *  If the value is 0 then we are in the single consumer mode. If it is 1 then we are
 *  in the multiple consumer mode.
 */
#define IX_COMM_ID_CONSUMER_MODE_MASK   0x4


/**
 * DESCRIPTION: Mask that describes the bit specifying the comm Id CCI enabled mode mask.
 *  If the value is 0 then the CCI notification mode is not enabled. If it is 1 then the
 *  the CCI notification mode is enabled.
 */
#define IX_COMM_ID_CC_ENABLED_MODE_MASK 0x8

/**
 * TYPENAME: ix_end_point_cc_data
 * 
 * DESCRIPTION: This structure represents the core component end point data. The two notification
 *          queues are used for flow control. The synchronization semaphore is used for exclusive
 *          access to the data in MT environment.
 *
 */
typedef struct ix_s_end_point_cc_data
{
    ix_end_point_data_queue     m_EPDataQueue; /* EP data queue */
    ix_uint32                   m_Flags; /* EP configuration flags */

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_wait_notification_node*  m_pSendNotificationQueue; 
        /* full data queue flow control notification queue */
    ix_wait_notification_node*  m_pReceiveNotificationQueue; 
        /* empty data queue flow control notification queue */

    ix_ossl_sem_t               m_SyncSemaphore; /* access synchronization semaphore */

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* CCI support */
#if !defined( _IX_RM_CCI_SUPPORT_EXCLUDE_ )

    void*                       m_CCNotificationContext; /* this is the context for CC notification */

#endif /* !defined( _IX_RM_CCI_SUPPORT_EXCLUDE_ ) */

} ix_end_point_cc_data;



/**
 * TYPENAME: ix_end_point_ub_data
 * 
 * DESCRIPTION: This structure represents the microblock end point data. This will be different form the
 *          core component end point data due to the different requirements and implementation of the 
 *          communication to the microblocks.
 *
 */
typedef struct ix_s_end_point_ub_data
{
    ix_hw_ring_handle   m_hHwRing;
    ix_uint32           m_Flags;
} ix_end_point_ub_data;



/**
 * TYPENAME: ix_end_point_data
 * 
 * DESCRIPTION: This union defines the end point data.
 *
 */
typedef union ix_u_end_point_data
{
    ix_end_point_cc_data m_CoreCompData;
    ix_end_point_ub_data m_UBlockData;
} ix_end_point_data;



/**
 * TYPENAME: ix_end_point_behaviour
 * 
 * DESCRIPTION: This structure defines the behaviour associated with an end point. When
 *          a sender will pass data to a communication ID, then this behaviour will decide
 *          how data will be processed further.
 *
 */
typedef struct ix_s_end_point_behaviour
{
    _ix_comm_ep_behaviour_handler   m_BehaviourHandler;
    void*                           m_pInternalBehaviourContext;
    void*                           m_pExternalBehaviourContext;
} ix_end_point_behaviour;



/**
 * TYPENAME: ix_end_point
 * 
 * DESCRIPTION: This structure defines the end-point abstraction. An end point has a behivour and
 *          data associated with it. Based on this information the data is routed accordingly by
 *          the end point. This abstraction will be able to handle many possible data passing
 *          configurations.
 */
typedef struct ix_s_end_point
{
    ix_end_point_behaviour      m_EPBehaviour;
    ix_end_point_data           m_EPData;
} ix_end_point;



/**
 * Processing functions for communication Ids
 */

/**
 * NAME: _ix_rm_comm_data_drop
 *
 * DESCRIPTION: This function will be the default processing function for each
 *          communication ID. This will return the buffers to the owning free list.
 * 
 * @Param:  - IN arg_hBuffer - the handle of the buffer to be processed
 * @Param:  - IN arg_UserData - the user data associated with the buffer to process
 * @Param:  - IN arg_pContext - context that has been passed to the comm ID at the 
 *          default handler registration time.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_comm_data_drop( 
                                ix_buffer_handle arg_hBuffer, 
                                ix_uint32 arg_UserData,
                                void* arg_pContext 
                              );

/**
 * NAME: _ix_rm_comm_data_store_mp
 *
 * DESCRIPTION: This function will be the default get/select processing function for each
 *  communication ID. The end point will store the data received and will notify
 *  all pending tasks that new data is available..
 * 
 * @Param:  - IN arg_hBuffer - the handle of the buffer to be processed
 * @Param:  - IN arg_UserData - the user data associated with the buffer to process
 * @Param:  - IN arg_pContext - context that has been passed to the comm ID at the 
 *  default handler registration time.
 * @Param:  - IN arg_SendCallType - specifies if the send call was blocking or unblocking.
 * @Param:  - IN arg_Timeout - specifies the timeout for the operation for the blocking
 *  call.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_comm_data_store_mp( 
                                    ix_buffer_handle arg_hBuffer, 
                                    ix_uint32 arg_UserData,
                                    void* arg_pContext,
                                    ix_send_call_type arg_SendCallType,
                                    ix_uint32 arg_Timeout
                                  );


/**
 * NAME: _ix_rm_comm_data_store_sp
 *
 * DESCRIPTION: This function will be the default get/select processing function for each
 *  communication ID in single producer mode. The end point will store the data received and will notify
 *  all pending tasks that new data is available.
 * 
 * @Param:  - IN arg_hBuffer - the handle of the buffer to be processed
 * @Param:  - IN arg_UserData - the user data associated with the buffer to process
 * @Param:  - IN arg_pContext - context that has been passed to the comm ID at the 
 *  default handler registration time.
 * @Param:  - IN arg_SendCallType - specifies if the send call was blocking or unblocking.
 * @Param:  - IN arg_Timeout - specifies the timeout for the operation for the blocking
 *  call.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_comm_data_store_sp( 
                                    ix_buffer_handle arg_hBuffer, 
                                    ix_uint32 arg_UserData,
                                    void* arg_pContext,
                                    ix_send_call_type arg_SendCallType,
                                    ix_uint32 arg_Timeout
                                  );



/**
 * Internal macros used for the internal communication function implementation
 */


/**
 * No check is done to guarantee that both the buffer handle and the user data fit
 * into the queue. We make the asumption that the controlled array is a multiple of
 * the element size (2 ix_uint32 values in our case)
 */

/* CCI support enabled */
#if !defined( _IX_RM_CCI_SUPPORT_EXCLUDE_ )

#define _IX_EP_DATA_QUEUE_PUT_COMM_ELEMENT_MP( \
                                               arg_pEPDataQueue, \
                                               arg_hBuffer, \
                                               arg_UserData, \
                                               arg_pEPCCData \
                                             ) \
    do \
    { \
        register ix_uint32 put = arg_pEPDataQueue->m_Put; \
        register ix_uint32* putAddress = (ix_uint32*)((ix_uint8*)arg_pEPDataQueue->m_pStartAddress + (put & arg_pEPDataQueue->m_Capacity) * arg_pEPDataQueue->m_ElementSize); \
        *(putAddress++) = arg_hBuffer; \
        *putAddress = arg_UserData; \
        if(put == arg_pEPDataQueue->m_Get) \
        { \
            arg_pEPDataQueue->m_Put = ++put; \
            if(arg_pEPCCData->m_pReceiveNotificationQueue) \
            { \
                _IX_RM_EP_DATA_WAIT_QUEUE_NOTIFY(arg_pEPCCData->m_pReceiveNotificationQueue); \
            } \
            if((arg_pEPCCData->m_Flags & IX_COMM_ID_CC_ENABLED_MODE_MASK) != 0) \
                ix_cci_notify(arg_pEPCCData->m_CCNotificationContext); \
        } \
        else \
            arg_pEPDataQueue->m_Put = ++put; \
    } \
    while(0)

#else

#define _IX_EP_DATA_QUEUE_PUT_COMM_ELEMENT_MP( \
                                               arg_pEPDataQueue, \
                                               arg_hBuffer, \
                                               arg_UserData, \
                                               arg_pEPCCData \
                                             ) \
    do \
    { \
        register ix_uint32 put = arg_pEPDataQueue->m_Put; \
        register ix_uint32* putAddress = (ix_uint32*)((ix_uint8*)arg_pEPDataQueue->m_pStartAddress + (put & arg_pEPDataQueue->m_Capacity) * arg_pEPDataQueue->m_ElementSize); \
        *(putAddress++) = arg_hBuffer; \
        *putAddress = arg_UserData; \
        if((arg_pEPCCData->m_pReceiveNotificationQueue) && (put == arg_pEPDataQueue->m_Get)) \
        { \
            arg_pEPDataQueue->m_Put = ++put; \
            _IX_RM_EP_DATA_WAIT_QUEUE_NOTIFY(arg_pEPCCData->m_pReceiveNotificationQueue); \
        } \
        else \
            arg_pEPDataQueue->m_Put = ++put; \
    } \
    while(0)

#endif /* !defined( _IX_RM_CCI_SUPPORT_EXCLUDE_ ) */


/**
 * This macro will produce in the single producer case, when less locking
 * is required.
 */

/* CCI support enabled */
#if !defined( _IX_RM_CCI_SUPPORT_EXCLUDE_ )

#define _IX_EP_DATA_QUEUE_PUT_COMM_ELEMENT_SP( \
                                               arg_pEPDataQueue, \
                                               arg_hBuffer, \
                                               arg_UserData, \
                                               arg_pEPCCData \
                                             ) \
    do \
    { \
        register ix_uint32 put = arg_pEPDataQueue->m_Put; \
        register ix_uint32* putAddress = (ix_uint32*)((ix_uint8*)arg_pEPDataQueue->m_pStartAddress + (put & arg_pEPDataQueue->m_Capacity) * arg_pEPDataQueue->m_ElementSize); \
        *(putAddress++) = arg_hBuffer; \
        *putAddress = arg_UserData; \
        if(put == arg_pEPDataQueue->m_Get) \
        { \
            arg_pEPDataQueue->m_Put = ++put; \
            if(arg_pEPCCData->m_pReceiveNotificationQueue) \
            { \
                _IX_RM_EP_DATA_ACTION_LOCK(arg_pEPCCData); \
                _IX_RM_EP_DATA_WAIT_QUEUE_NOTIFY(arg_pEPCCData->m_pReceiveNotificationQueue); \
                _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData); \
            } \
            if((arg_pEPCCData->m_Flags & IX_COMM_ID_CC_ENABLED_MODE_MASK) != 0) \
                ix_cci_notify(arg_pEPCCData->m_CCNotificationContext); \
        } \
        else \
            arg_pEPDataQueue->m_Put = ++put; \
    } \
    while(0)

#else

#define _IX_EP_DATA_QUEUE_PUT_COMM_ELEMENT_SP( \
                                               arg_pEPDataQueue, \
                                               arg_hBuffer, \
                                               arg_UserData, \
                                               arg_pEPCCData \
                                             ) \
    do \
    { \
        register ix_uint32 put = arg_pEPDataQueue->m_Put; \
        register ix_uint32* putAddress = (ix_uint32*)((ix_uint8*)arg_pEPDataQueue->m_pStartAddress + (put & arg_pEPDataQueue->m_Capacity) * arg_pEPDataQueue->m_ElementSize); \
        *(putAddress++) = arg_hBuffer; \
        *putAddress = arg_UserData; \
        if((arg_pEPCCData->m_pReceiveNotificationQueue) && (put == arg_pEPDataQueue->m_Get)) \
        { \
            arg_pEPDataQueue->m_Put = ++put; \
            _IX_RM_EP_DATA_ACTION_LOCK(arg_pEPCCData); \
            _IX_RM_EP_DATA_WAIT_QUEUE_NOTIFY(arg_pEPCCData->m_pReceiveNotificationQueue); \
            _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData); \
        } \
        else \
            arg_pEPDataQueue->m_Put = ++put; \
    } \
    while(0)

#endif /* !defined( _IX_RM_CCI_SUPPORT_EXCLUDE_ ) */



/**
 * No check is done to guarantee that both the buffer handle and the user data are
 * into the queue. We make the asumption that the controlled array is a multiple of
 * the element size (2 ix_uint32 values in our case) and the put operation is atomic.
 */
#define _IX_EP_DATA_QUEUE_GET_COMM_ELEMENT_MC( \
                                               arg_pEPDataQueue, \
                                               arg_pBufferHandle, \
                                               arg_pUserData, \
                                               arg_pEPCCData \
                                             ) \
    do \
    { \
        register ix_uint32 get = arg_pEPDataQueue->m_Get; \
        register ix_uint32* getAddress = (ix_uint32*)((ix_uint8*)arg_pEPDataQueue->m_pStartAddress + (get & arg_pEPDataQueue->m_Capacity) * arg_pEPDataQueue->m_ElementSize); \
        *arg_pBufferHandle = *(getAddress++); \
        *arg_pUserData = *getAddress; \
        if((arg_pEPCCData->m_pSendNotificationQueue) && \
            (arg_pEPDataQueue->m_Put - get == arg_pEPDataQueue->m_Capacity)) \
        { \
            arg_pEPDataQueue->m_Get = ++get; \
            _IX_RM_EP_DATA_WAIT_QUEUE_NOTIFY(arg_pEPCCData->m_pSendNotificationQueue); \
        } \
        else \
            arg_pEPDataQueue->m_Get = ++get;\
    } \
    while(0)



/**
 * This macro will consume in the single consumer mode. We are truing to do less locking
 * in this mode.
 */
#define _IX_EP_DATA_QUEUE_GET_COMM_ELEMENT_SC( \
                                               arg_pEPDataQueue, \
                                               arg_pBufferHandle, \
                                               arg_pUserData, \
                                               arg_pEPCCData \
                                             ) \
    do \
    { \
        register ix_uint32 get = arg_pEPDataQueue->m_Get; \
        register ix_uint32* getAddress = (ix_uint32*)((ix_uint8*)arg_pEPDataQueue->m_pStartAddress + (get & arg_pEPDataQueue->m_Capacity) * arg_pEPDataQueue->m_ElementSize); \
        *arg_pBufferHandle = *(getAddress++); \
        *arg_pUserData = *getAddress; \
        if((arg_pEPCCData->m_pSendNotificationQueue) && \
            (arg_pEPDataQueue->m_Put - get == arg_pEPDataQueue->m_Capacity)) \
        { \
            arg_pEPDataQueue->m_Get = ++get; \
            _IX_RM_EP_DATA_ACTION_LOCK(arg_pEPCCData); \
            _IX_RM_EP_DATA_WAIT_QUEUE_NOTIFY(arg_pEPCCData->m_pSendNotificationQueue); \
            _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData); \
        } \
        else \
            arg_pEPDataQueue->m_Get = ++get;\
    } \
    while(0)

/* notification queue related macros */
#define _IX_RM_EP_DATA_WAIT_QUEUE_NOTIFY( \
                                          arg_pWaitNotificationQueue \
                                        ) \
    do \
    { \
        ix_wait_notification_node* pWNNode = arg_pWaitNotificationQueue; \
        while(pWNNode) \
        { \
            /* signal the entity waiting on the semaphore */ \
            ix_ossl_sem_give(pWNNode->m_Semaphore); \
             \
            /* go to the next node in the list */ \
            pWNNode = pWNNode->m_pNext; \
             \
            /* remove the current notification node */ \
            ix_ossl_free(arg_pWaitNotificationQueue); \
             \
            /* set the head of the list to the next node */ \
            arg_pWaitNotificationQueue = pWNNode; \
        } \
    } \
    while(0)


#define _IX_RM_EP_DATA_WAIT_QUEUE_ADD( \
                                       arg_pWaitNotificationQueue, \
                                       arg_Semaphore \
                                      ) \
    do \
    { \
        ix_wait_notification_node* pWNNode = (ix_wait_notification_node*)ix_ossl_malloc(sizeof(ix_wait_notification_node)); \
        if(pWNNode == 0) \
        { \
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_NOTIFICATION_ELEM_ALLOCATION_FAILED, s_aFunctionIdentifier, \
                "the notification element allocation failed!", 0); \
        } \
        else \
        { \
            pWNNode->m_Semaphore = arg_Semaphore; \
            pWNNode->m_pNext = arg_pWaitNotificationQueue; \
            arg_pWaitNotificationQueue = pWNNode; \
        } \
    } \
    while(0)


#define _IX_RM_EP_DATA_WAIT_QUEUE_REMOVE( \
                                          arg_pWaitNotificationQueue, \
                                          arg_Semaphore \
                                        ) \
    do \
    { \
        ix_wait_notification_node *pWNPrevNode = 0, *pWNNode = arg_pWaitNotificationQueue; \
        while(pWNNode) \
        { \
            if(pWNNode->m_Semaphore == arg_Semaphore) \
            { \
                if(pWNPrevNode == 0) \
                    arg_pWaitNotificationQueue = pWNNode->m_pNext; \
                else \
                    pWNPrevNode->m_pNext = pWNNode->m_pNext; \
                \
                /* free the node */ \
                ix_ossl_free(pWNNode); \
                 \
                break; \
            } \
            pWNPrevNode = pWNNode; \
            pWNNode = pWNNode->m_pNext; \
        } \
    } \
    while(0)

/* data synchronization macros */
#define _IX_RM_EP_DATA_ACTION_LOCK( \
                                    arg_pEPCCData \
                                  ) \
    do \
    { \
        ix_ossl_sem_take(arg_pEPCCData->m_SyncSemaphore,IX_OSSL_WAIT_FOREVER); \
    } \
    while(0)
                                        


#define _IX_RM_EP_DATA_ACTION_UNLOCK( \
                                      arg_pEPCCData \
                                    ) \
    do \
    { \
        ix_ossl_sem_give(arg_pEPCCData->m_SyncSemaphore);\
    } \
    while(0)





/**
 * PCI queueing/dequing macros
 */

/**
 * No check is done to guarantee that both the buffer handle the user data and destination
 * comm Id fit into the queue. We make the asumption that the controlled array is a multiple
 * of the element size (3 ix_uint32 values in our case)
 */

#define _IX_EP_DATA_QUEUE_PCI_PUT_COMM_ELEMENT_MP( \
                                               arg_pEPDataQueue, \
                                               arg_hBuffer, \
                                               arg_UserData, \
                                               arg_Destination, \
                                               arg_pEPCCData \
                                             ) \
    do \
    { \
        register ix_uint32 put = arg_pEPDataQueue->m_Put; \
        register ix_uint32* putAddress = (ix_uint32*)((ix_uint8*)arg_pEPDataQueue->m_pStartAddress + (put & arg_pEPDataQueue->m_Capacity) * arg_pEPDataQueue->m_ElementSize); \
        *(putAddress++) = arg_hBuffer; \
        *(putAddress++) = arg_UserData; \
        *putAddress = arg_Destination; \
        if((arg_pEPCCData->m_pReceiveNotificationQueue) && (put == arg_pEPDataQueue->m_Get)) \
        { \
            arg_pEPDataQueue->m_Put = ++put; \
            _IX_RM_EP_DATA_WAIT_QUEUE_NOTIFY(arg_pEPCCData->m_pReceiveNotificationQueue); \
        } \
        else \
            arg_pEPDataQueue->m_Put = ++put; \
    } \
    while(0)


/**
 * No check is done to guarantee that all the buffer handle ,the user data and the 
 * destination communication Id are into the queue. We make the asumption that the 
 * controlled array is a multiple of the element size (3 ix_uint32 values in our case) 
 * and the put operation is atomic.
 */

#define _IX_EP_DATA_QUEUE_PCI_GET_COMM_ELEMENT_SC( \
                                               arg_pEPDataQueue, \
                                               arg_pBufferHandle, \
                                               arg_pUserData, \
                                               arg_pDestination, \
                                               arg_pEPCCData \
                                             ) \
    do \
    { \
        register ix_uint32 get = arg_pEPDataQueue->m_Get; \
        register ix_uint32* getAddress = (ix_uint32*)((ix_uint8*)arg_pEPDataQueue->m_pStartAddress + (get & arg_pEPDataQueue->m_Capacity) * arg_pEPDataQueue->m_ElementSize); \
        *arg_pBufferHandle = *(getAddress++); \
        *arg_pUserData = *(getAddress++); \
        *arg_pDestination = *getAddress; \
        if((arg_pEPCCData->m_pSendNotificationQueue) && \
            (arg_pEPDataQueue->m_Put - get == arg_pEPDataQueue->m_Capacity)) \
        { \
            arg_pEPDataQueue->m_Get = ++get; \
            _IX_RM_EP_DATA_ACTION_LOCK(arg_pEPCCData); \
            _IX_RM_EP_DATA_WAIT_QUEUE_NOTIFY(arg_pEPCCData->m_pSendNotificationQueue); \
            _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData); \
        } \
        else \
            arg_pEPDataQueue->m_Get = ++get; \
    } \
    while(0)




/**
 * The following structures are designed for implementation of a semaphore pool
 */


/**
 * TYPENAME: ix_sem_free_node
 * 
 * DESCRIPTION: This structure defines a free semaphore pool node
 *
 */
typedef struct ix_s_sem_free_node  ix_sem_free_node;

struct ix_s_sem_free_node
{
    ix_ossl_sem_t       m_Semaphore; /* the free semaphore */
    ix_sem_free_node*   m_pNext; /* next node in the free list */
};


/**
 * TYPENAME: ix_sem_pool
 * 
 * DESCRIPTION: This structure defines a semaphore pool. This structure holds no data,
 *          and has to be initialized externally by the provided function.
 *
 */
typedef struct ix_s_sem_pool
{
    ix_ossl_sem_t*      m_pSemaphores; /* pointer to an array of semaphore handles */
    ix_uint32           m_Count; /* number of semaphores in the array */
    ix_sem_free_node*   m_pSemFreeList; /* semaphore free list */
} ix_sem_pool;



/**
 * DESCRIPTION: This symbol defines the size of the internal communication
 *      semaphore pool.
 *
 */
#define IX_COMM_SEM_POOL_SIZE   512UL


/**
 * TYPENAME: ix_end_point_manager
 * 
 * DESCRIPTION: This structure represents the end-point manager. It manages both
 *          the message and packets end points.
 *
 */
typedef struct ix_s_end_point_manager
{
    ix_end_point    m_aPacketEndPoint[IX_COMM_LOCAL_ID_NUMBER];

    ix_end_point    m_aMessageEndPoint[IX_COMM_LOCAL_ID_NUMBER];

    /* this is the semaphore pool used by the communication API */
    ix_sem_pool     m_CommSemaphorePool;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter      m_SyncAdapter; /* synchronization adapter */

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

} ix_end_point_manager;


/**
 * NAME: _ix_rm_sem_pool_init
 *
 * DESCRIPTION: This function will initialize a semaphore pool. The arg_count specifies
 *          how many semaphores we would like to have into the pool. The structure referred
 *          by arg_pSemPool will be initialized on successful return. The count should be 
 *          greater than 1 because the first semaphore in the created array will be used
 *          for operation synchronization.
 * 
 * @Param:  - INOUT arg_pSemPool - pointer to a semaphore pool that will be initialized.
 * @Param:  - IN arg_Count - the number of semaphores requested to be in the pool.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_sem_pool_init(
                               ix_sem_pool* arg_pSemPool,
                               ix_uint32 arg_Count
                             );


/**
 * NAME: _ix_rm_sem_pool_fini
 *
 * DESCRIPTION: This function will destroy the passed semaphore pool releasing all 
 *          semaphores and associated data. Upon return the passed pool structure will
 *          be totally reset and can be reused.
 * 
 * @Param:  - INOUT arg_pSemPool - the semaphore pool that we would like to cleanup. 
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_sem_pool_fini(
                               ix_sem_pool* arg_pSemPool
                             );


/**
 * NAME: _ix_rm_sem_pool_get_sem
 *
 * DESCRIPTION: This function will get a semaphore from the requested pool. This
 *          semaphore has to be returned to the pool in order to avoid resource 
 *          exhaustion.
 * 
 * @Param:  - IN arg_pSemPool - address of the pool of interest.
 * @Param:  - OUT arg_pSem - address where the available semaphore will be returned.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_sem_pool_get_sem( 
                                  ix_sem_pool* arg_pSemPool,
                                  ix_ossl_sem_t* arg_pSem 
                                );

/**
 * NAME: _ix_rm_sem_pool_release_sem
 *
 * DESCRIPTION: This function will return a semaphore to a given pool. No check is made
 *          to see if the semaphore really belongs to the pool. Also there will be no 
 *          check to see if the semaphore has been already released.
 * 
 * @Param:  - IN arg_pSemPool - semaphore pool of interest
 * @Param:  - IN arg_Sem - semaphore to be released
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_sem_pool_release_sem( 
                                      ix_sem_pool* arg_pSemPool,
                                      ix_ossl_sem_t arg_Sem 
                                    );


/**
 * Multi producer queueing functions
 */


/**
 * NAME: _ix_rm_cc_queue_element_nonblock_mp
 *
 * DESCRIPTION: This function will queue a data element into specified core component end point
 *          data queue. The function is non blocking. If the queue is full then an error will be
 *          returned right away. In case of error it is the duty of the caller to decide what to 
 *          do and eventually drop the buffer.
 * 
 * @Param:  - IN arg_pEPCCData -  core component end point data address this fuction will work on
 * @Param:  - IN arg_hBuffer - buffer handle to be stored onto the end point data queue
 * @Param:  - IN arg_UserData - user data to be stored onto the end point data queue
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_cc_queue_element_nonblock_mp( 
                                              ix_end_point_cc_data* arg_pEPCCData,
                                              ix_buffer_handle arg_hBuffer,
                                              ix_uint32 arg_UserData
                                            );



/**
 * NAME: _ix_rm_cc_queue_element_block_mp
 *
 * DESCRIPTION: This function will queue a data element into specified core component end point
 *          data queue. The function is blocking. If the queue is full then the call will wait for
 *          the specified timeout for the queue to become available. If the timeout is infinite then
 *          the function will return only when was able to queue the element. In the case of error
 *          it is the duty of the caller to decide what to do and eventually drop the buffer.
 * 
 * @Param:  - IN arg_pEPCCData -  core component end point data address this fuction will work on
 * @Param:  - IN arg_hBuffer - buffer handle to be stored onto the end point data queue
 * @Param:  - IN arg_UserData - user data to be stored onto the end point data queue
 * @Param:  - IN arg_Timeout - timeout specifying for how long to wait for the queue to become available.
 *          It is expressed in milliseconds.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_cc_queue_element_block_mp( 
                                           ix_end_point_cc_data* arg_pEPCCData,
                                           ix_buffer_handle arg_hBuffer,
                                           ix_uint32 arg_UserData,
                                           ix_uint32 arg_Timeout
                                         );

/**
 * Multi consumer dequeueing functions
 */


/**
 * NAME: _ix_rm_cc_dequeue_element_nonblock_mc
 *
 * DESCRIPTION: This function will dequeue a data element from the specified core component end point
 *          data queue. The function is non blocking. If the queue is empty then an error will be
 *          returned right away. In case of error it is the duty of the caller to decide what to 
 *          do.
 * 
 * @Param:  - IN arg_pEPCCData -  core component end point data address this fuction will work on
 * @Param:  - OUT arg_pBufferHandle - address where the dequeued buffer handle will be stored on return.
 * @Param:  - OUT arg_UserData - address where the dequeued user data will be stored on return.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_cc_dequeue_element_nonblock_mc( 
                                                ix_end_point_cc_data* arg_pEPCCData,
                                                ix_buffer_handle* arg_pBufferHandle,
                                                ix_uint32* arg_UserData
                                              );


/**
 * NAME: _ix_rm_cc_dequeue_element_block_mc
 *
 * DESCRIPTION: This function will dequeue a data element from the specified core component end point
 *          data queue. The function is blocking. If the queue is empty then the call will wait for
 *          the specified timeout for data to become available. If the timeout is infinite then
 *          the function will return only when was able to dequeue an element. In the case of error
 *          it is the duty of the caller to decide what to do.
 * 
 * @Param:  - IN arg_pEPCCData -  core component end point data address this fuction will work on
 * @Param:  - OUT arg_pBufferHandle - address where the dequeued buffer handle will be stored on return.
 * @Param:  - OUT arg_UserData - address where the dequeued user data will be stored on return.
 * @Param:  - IN arg_Timeout - timeout specifying for how long to wait for data to become available.
 *          It is expressed in milliseconds.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */ix_error _ix_rm_cc_dequeue_element_block_mc( 
                                                ix_end_point_cc_data* arg_pEPCCData,
                                                ix_buffer_handle* arg_pBufferHandle,
                                                ix_uint32* arg_pUserData,
                                                ix_uint32 arg_Timeout
                                              );



/**
 * Single producer queueing functions
 */


/**
 * NAME: _ix_rm_cc_queue_element_nonblock_sp
 *
 * DESCRIPTION: This function will queue a data element into specified core component end point
 *          data queue. The function is non blocking. If the queue is full then an error will be
 *          returned right away. In case of error it is the duty of the caller to decide what to 
 *          do and eventually drop the buffer.
 * 
 * @Param:  - IN arg_pEPCCData -  core component end point data address this fuction will work on
 * @Param:  - IN arg_hBuffer - buffer handle to be stored onto the end point data queue
 * @Param:  - IN arg_UserData - user data to be stored onto the end point data queue
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_cc_queue_element_nonblock_sp( 
                                              ix_end_point_cc_data* arg_pEPCCData,
                                              ix_buffer_handle arg_hBuffer,
                                              ix_uint32 arg_UserData
                                            );



/**
 * NAME: _ix_rm_cc_queue_element_block_sp
 *
 * DESCRIPTION: This function will queue a data element into specified core component end point
 *          data queue. The function is blocking. If the queue is full then the call will wait for
 *          the specified timeout for the queue to become available. If the timeout is infinite then
 *          the function will return only when was able to queue the element. In the case of error
 *          it is the duty of the caller to decide what to do and eventually drop the buffer.
 * 
 * @Param:  - IN arg_pEPCCData -  core component end point data address this fuction will work on
 * @Param:  - IN arg_hBuffer - buffer handle to be stored onto the end point data queue
 * @Param:  - IN arg_UserData - user data to be stored onto the end point data queue
 * @Param:  - IN arg_Timeout - timeout specifying for how long to wait for the queue to become available.
 *          It is expressed in milliseconds.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_cc_queue_element_block_sp( 
                                           ix_end_point_cc_data* arg_pEPCCData,
                                           ix_buffer_handle arg_hBuffer,
                                           ix_uint32 arg_UserData,
                                           ix_uint32 arg_Timeout
                                         );

/**
 * Single consumer dequeueing functions
 */


/**
 * NAME: _ix_rm_cc_dequeue_element_nonblock_sc
 *
 * DESCRIPTION: This function will dequeue a data element from the specified core component end point
 *          data queue. The function is non blocking. If the queue is empty then an error will be
 *          returned right away. In case of error it is the duty of the caller to decide what to 
 *          do.
 * 
 * @Param:  - IN arg_pEPCCData -  core component end point data address this fuction will work on
 * @Param:  - OUT arg_pBufferHandle - address where the dequeued buffer handle will be stored on return.
 * @Param:  - OUT arg_UserData - address where the dequeued user data will be stored on return.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_cc_dequeue_element_nonblock_sc( 
                                                ix_end_point_cc_data* arg_pEPCCData,
                                                ix_buffer_handle* arg_pBufferHandle,
                                                ix_uint32* arg_UserData
                                              );


/**
 * NAME: _ix_rm_cc_dequeue_element_block_sc
 *
 * DESCRIPTION: This function will dequeue a data element from the specified core component end point
 *          data queue. The function is blocking. If the queue is empty then the call will wait for
 *          the specified timeout for data to become available. If the timeout is infinite then
 *          the function will return only when was able to dequeue an element. In the case of error
 *          it is the duty of the caller to decide what to do.
 * 
 * @Param:  - IN arg_pEPCCData -  core component end point data address this fuction will work on
 * @Param:  - OUT arg_pBufferHandle - address where the dequeued buffer handle will be stored on return.
 * @Param:  - OUT arg_UserData - address where the dequeued user data will be stored on return.
 * @Param:  - IN arg_Timeout - timeout specifying for how long to wait for data to become available.
 *          It is expressed in milliseconds.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */ix_error _ix_rm_cc_dequeue_element_block_sc( 
                                                ix_end_point_cc_data* arg_pEPCCData,
                                                ix_buffer_handle* arg_pBufferHandle,
                                                ix_uint32* arg_pUserData,
                                                ix_uint32 arg_Timeout
                                              );


/**
 * NAME: _ix_rm_cc_pci_queue_element_nonblock_mp
 *
 * DESCRIPTION: This function will queue a pci data element into specified end point
 *          data queue. The function is non blocking. If the queue is full then an error will be
 *          returned right away. In case of error it is the duty of the caller to decide what to 
 *          do and eventually drop the buffer. The element in this case will include the destination
 *          comm Id.
 * 
 * @Param:  - IN arg_pEPCCData -  core component end point data address this fuction will work on
 * @Param:  - IN arg_hBuffer - buffer handle to be stored onto the end point data queue
 * @Param:  - IN arg_UserData - user data to be stored onto the end point data queue
 * @Param:  - IN arg_Destination - destination comm Id to be stored onto the end point data queue
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_cc_pci_queue_element_nonblock_mp( 
                                           ix_end_point_cc_data* arg_pEPCCData,
                                           ix_buffer_handle arg_hBuffer,
                                           ix_uint32 arg_UserData,
                                           ix_communication_id arg_Destination
                                         );


/**
 * NAME: _ix_rm_cc_pci_queue_element_block_mp
 *
 * DESCRIPTION: This function will queue a pci data element into the specified end point
 *          data queue. The function is blocking. If the queue is full then the call will wait for
 *          the specified timeout for the queue to become available. If the timeout is infinite then
 *          the function will return only when was able to queue the element. In the case of error
 *          it is the duty of the caller to decide what to do and eventually drop the buffer.
 *          The element in this case will include the destination comm Id.
 * 
 * @Param:  - IN arg_pEPCCData -  core component end point data address this fuction will work on
 * @Param:  - IN arg_hBuffer - buffer handle to be stored onto the end point data queue
 * @Param:  - IN arg_UserData - user data to be stored onto the end point data queue
 * @Param:  - IN arg_Destination - destination comm Id to be stored onto the end point data queue
 * @Param:  - IN arg_Timeout - timeout specifying for how long to wait for the queue to become available.
 *          It is expressed in milliseconds.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_cc_pci_queue_element_block_mp( 
                                        ix_end_point_cc_data* arg_pEPCCData,
                                        ix_buffer_handle arg_hBuffer,
                                        ix_uint32 arg_UserData,
                                        ix_communication_id arg_Destination,
                                        ix_uint32 arg_Timeout
                                      );


/**
 * NAME: _ix_rm_comm_pci_data_store_mp
 *
 * DESCRIPTION: This function will be the default processing function for each PCI
 *  end point. The end point will store the data received and will notify
 *  all pending tasks that new data is available..
 * 
 * @Param:  - IN arg_hBuffer - the handle of the buffer to be processed
 * @Param:  - IN arg_UserData - the user data associated with the buffer to process
 * @Param:  - IN arg_pContext - context that has been passed to the comm ID at the 
 *  default handler registration time.
 * @Param:  - IN arg_Destination - destination comm Id to be stored onto the end point data queue
 * @Param:  - IN arg_SendCallType - specifies if the send call was blocking or unblocking.
 * @Param:  - IN arg_Timeout - specifies the timeout for the operation for the blocking
 *  call.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_comm_pci_data_store_mp( 
                                    ix_buffer_handle arg_hBuffer, 
                                    ix_uint32 arg_UserData,
                                    void* arg_pContext,
                                    ix_communication_id arg_Destination,
                                    ix_send_call_type arg_SendCallType,
                                    ix_uint32 arg_Timeout
                                  );

/**
 * NAME: _ix_rm_process_queue_entries
 *
 * DESCRIPTION: This function will pass all entries from the queue to the specified
 *          handler for processing.
 * 
 * @Param:  - INOUT arg_pEPCCData - cc end point internal queue to be processed
 * @Param:  - IN arg_DataHandler - data handler to be used for processing all queue elements.
 * @Param:  - IN arg_pContext - context to be passed to the data handler.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_process_queue_entries( 
                                       ix_end_point_cc_data* arg_pEPCCData,
                                       ix_comm_data_handler arg_DataHandler,
                                       void* arg_pContext
                                     );



/**
 * NAME: _ix_rm_cc_select
 *
 * DESCRIPTION: This function will wait on a local set of core component communication IDs 
 *          to have messages and/or data available. The arg_SelectActionSet parameter is an
 *          array of 32 bit flags specifying what kind of data we are waiting for on each of
 *          the specified CC identifiers. The first element of the array coresponds to the
 *          first CC end point. You can not wait for uBlock points for data. The function will
 *          wait for data for a maximum time specified by arg_Timeout parameter.
 * 
 * @Param:  - INOUT arg_SelectActionSet - array of 32bit flags representing what kind of data we
 *          are waiting on packets/messages for each CC end point. On return the flags will be 
 *          cleared and will specify only the data type that has been received on each local
 *          CC communication ID.
 * @Param:  - IN arg_Timeout - timeout specifying for how long we want to wait for data
 *          to become available. It is expressed in milliseconds.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_cc_select( 
                           ix_comm_select_action_set arg_SelectActionSet, 
                           ix_uint32 arg_Timeout 
                         );

/**
 * NAME: _ix_rm_communication_init
 *
 * DESCRIPTION: This function should be called at the time the resource manager initializer 
 *          function is called. The function will initialize all the Communication API.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_communication_init(void);


/**
 * NAME: _ix_rm_communication_fini
 *
 * DESCRIPTION: This function should be called at the time the resource manager finalizer function
 *          is called. The function will provide cleanup for Communication API.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_communication_fini(void);


/* CCI support enabled */
#if !defined( _IX_RM_CCI_SUPPORT_EXCLUDE_ )

/**
 * NAME: ix_cci_notify
 *
 * DESCRIPTION: This function will be called when the commId becomes not empty to
 *  notify the registered CC object interested. This CC mode enabled will work just
 *  in single consumer mode.
 * 
 * @Param:  - IN pContext - context registered by a CC with the 
 *  comm Id.
 *
 */
void ix_cci_notify( 
                    void* pContext
                  );


#endif /* !defined( _IX_RM_CCI_SUPPORT_EXCLUDE_ ) */


/**
 * MACRO NAME: IX_RM_UBLOCK_TO_CORE_GET_DESTINATION
 *
 * DESCRIPTION: This macro will retrieve the communication destination from a data
 *          element's second longword that is sent from UCode to core
 *
 * @Param:  - IN - this is 32bit integer that is sent as the second longword into
 *          a data element from the uCode to core. The first longword will represent
 *          a hardware buffer handle.
 *
 * @Return: Returns an ix_uint32 value representing the local communication ID of the
 *          destination.
 */
#define IX_RM_UBLOCK_TO_CORE_GET_DESTINATION( \
                                              arg_Data \
                                            ) \
                                            IX_GET_BIT_FIELD32(arg_Data, 0U, IX_COMM_ID_LOCAL_ID_MSB)



/**
 * MACRO NAME: IX_RM_UBLOCK_TO_CORE_GET_USER_DATA
 *
 * DESCRIPTION: This macro will retrieve the communication user data from a data
 *          element's second longword that is sent from UCode to core
 *
 * @Param:  - IN - this is 32bit integer that is sent as the second longword into
 *          a data element from the uCode to core. The first longword will represent
 *          a hardware buffer handle.
 *
 * @Return: Returns an ix_uint32 value representing the user data for the data element
 *          sent from the uCode to core.
 */
#define IX_RM_UBLOCK_TO_CORE_GET_USER_DATA( \
                                            arg_Data \
                                          ) \
                                          IX_GET_BIT_FIELD32(arg_Data, 10U, 30U)
                                    


#if defined (_IX_RM_TEST_AI_INTERRUPTS_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

#define IX_COMM_SW_INTERRUPT    0x00

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

#define IX_COMM_SW_INTERRUPT    IRQ_INT_SWI

#endif /* (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) */

#endif /* defined (_IX_RM_TEST_AI_INTERRUPTS_) */


#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

/**
 * DESCRIPTION: This symbol describes the uCode to core packet interrupt
 *
 */


#define IX_COMM_PACKET_INTERRUPT_0    0x10
#define IX_COMM_PACKET_INTERRUPT_2    0x12

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

#define IX_COMM_PACKET_INTERRUPT_1    0x11
#define IX_COMM_PACKET_INTERRUPT_3    0x13

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */ 

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)


#define IX_COMM_PACKET_INTERRUPT_0    IRQ_IXP2000_THDA0 /* IRQ_INT_THDA0 */
#define IX_COMM_PACKET_INTERRUPT_2    IRQ_IXP2000_THDA2 /* IRQ_INT_THDA2 */

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

#define IX_COMM_PACKET_INTERRUPT_1    IRQ_IXP2000_THDA1 /* IRQ_INT_THDA1 */
#define IX_COMM_PACKET_INTERRUPT_3    IRQ_IXP2000_THDA3 /* IRQ_INT_THDA3 */

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */ 

#endif /* (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) */



#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

/**
 * DESCRIPTION: This symbol describes the uCode to core message interrupt
 *
 */

#define IX_COMM_MESSAGE_INTERRUPT_0    0x18
#define IX_COMM_MESSAGE_INTERRUPT_2    0x1a

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

#define IX_COMM_MESSAGE_INTERRUPT_1    0x19
#define IX_COMM_MESSAGE_INTERRUPT_3    0x1b

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)


#define IX_COMM_MESSAGE_INTERRUPT_0    IRQ_IXP2000_THDB0 /* IRQ_INT_THDB0 */
#define IX_COMM_MESSAGE_INTERRUPT_2    IRQ_IXP2000_THDB2 /* IRQ_INT_THDB2 */

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

#define IX_COMM_MESSAGE_INTERRUPT_1    IRQ_IXP2000_THDB1 /* IRQ_INT_THDB1 */
#define IX_COMM_MESSAGE_INTERRUPT_3    IRQ_IXP2000_THDB3 /* IRQ_INT_THDB3 */

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */

#endif /* (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) */


#include <halMmap.h>
#define IX_XSCALE_CSR_BASE_ADDR ((ix_uint32)Hal_xscale_local_intctl_virtAddr)


/**
 * MACRO NAME: IX_RM_COMM_PACKET_INTERRUPT_DISABLE
 *
 * DESCRIPTION: This macro will disable the specified uCode to core 
 *          communication interrupt.
 *
 * @Param:  - IN - arg_Interrupt an ix_uint32 value specifying the interrupt to 
 *          be disabled
 *
 */
#if defined(_IX_RM_IMPL_HARDWARE_) && !defined (_IX_RM_TEST_AI_INTERRUPTS_)

    #if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

    #define IX_RM_COMM_PACKET_INTERRUPT_DISABLE( \
                                                 arg_Dummy \
                                               ) \
        do \
        { \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x1e0)) = 0xffffffff; \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x1e8)) = 0xffffffff; \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x18)) = ((1 << IX_COMM_PACKET_INTERRUPT_0) | \
            (1 << IX_COMM_PACKET_INTERRUPT_2)); \
        } \
        while(0)

    #else

    #define IX_RM_COMM_PACKET_INTERRUPT_DISABLE( \
                                                 arg_Dummy \
                                               ) \
        do \
        { \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x1e0)) = 0xffffffff; \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x1e4)) = 0xffffffff; \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x1e8)) = 0xffffffff; \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x1ec)) = 0xffffffff; \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x18)) = ((1 << IX_COMM_PACKET_INTERRUPT_0) | \
            (1 << IX_COMM_PACKET_INTERRUPT_1) | (1 << IX_COMM_PACKET_INTERRUPT_2) | \
            (1 << IX_COMM_PACKET_INTERRUPT_3)); \
        } \
        while(0)

    #endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

#else

#define IX_RM_COMM_PACKET_INTERRUPT_DISABLE( \
                                             arg_Dummy \
                                           ) \
    do \
    { \
    } \
    while(0)

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */



/**
 * MACRO NAME: IX_RM_COMM_PACKET_INTERRUPT_ENABLE
 *
 * DESCRIPTION: This macro will enable the specified uCode to core 
 *          communication interrupt.
 *
 * @Param:  - IN - arg_Interrupt an ix_uint32 value specifying the interrupt to 
 *          be enabled
 *
 */

#if defined(_IX_RM_IMPL_HARDWARE_) && !defined (_IX_RM_TEST_AI_INTERRUPTS_)

    #if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

    #define IX_RM_COMM_PACKET_INTERRUPT_ENABLE( \
                                                arg_Dummy \
                                              ) \
        do \
        { \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x10)) = ((1 << IX_COMM_PACKET_INTERRUPT_0) | \
            (1 << IX_COMM_PACKET_INTERRUPT_2)); \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x160)) = 0xffffffff; \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x168)) = 0xffffffff; \
        } \
        while(0)
    
    #else

    #define IX_RM_COMM_PACKET_INTERRUPT_ENABLE( \
                                                 arg_Dummy \
                                              ) \
        do \
        { \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x10)) = ((1 << IX_COMM_PACKET_INTERRUPT_0) | \
            (1 << IX_COMM_PACKET_INTERRUPT_1) | (1 << IX_COMM_PACKET_INTERRUPT_2) | \
            (1 << IX_COMM_PACKET_INTERRUPT_3)); \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x160)) = 0xffffffff; \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x164)) = 0xffffffff; \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x168)) = 0xffffffff; \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x16c)) = 0xffffffff; \
        } \
        while(0)

    #endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

#else

#define IX_RM_COMM_PACKET_INTERRUPT_ENABLE( \
                                            arg_Dummy \
                                          ) \
    do \
    { \
    } \
    while(0)
                                    
#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */


/**
 * MACRO NAME: IX_RM_COMM_MESSAGE_INTERRUPT_DISABLE
 *
 * DESCRIPTION: This macro will disable the specified uCode to core 
 *          communication interrupt.
 *
 * @Param:  - IN - arg_Interrupt an ix_uint32 value specifying the interrupt to 
 *          be disabled
 *
 */
#if defined(_IX_RM_IMPL_HARDWARE_) && !defined (_IX_RM_TEST_AI_INTERRUPTS_)

    #if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

    #define IX_RM_COMM_MESSAGE_INTERRUPT_DISABLE( \
                                                  arg_Dummy \
                                                ) \
        do \
        { \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x200)) = 0xffffffff; \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x208)) = 0xffffffff; \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x18)) = ((1 << IX_COMM_MESSAGE_INTERRUPT_0) | \
            (1 << IX_COMM_MESSAGE_INTERRUPT_2)); \
        } \
        while(0)

    #else

    #define IX_RM_COMM_MESSAGE_INTERRUPT_DISABLE( \
                                                  arg_Dummy \
                                                ) \
        do \
        { \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x200)) = 0xffffffff; \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x204)) = 0xffffffff; \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x208)) = 0xffffffff; \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x20c)) = 0xffffffff; \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x18)) = ((1 << IX_COMM_MESSAGE_INTERRUPT_0) | \
            (1 << IX_COMM_MESSAGE_INTERRUPT_1) | (1 << IX_COMM_MESSAGE_INTERRUPT_2) | \
            (1 << IX_COMM_MESSAGE_INTERRUPT_3)); \
        } \
        while(0)

    #endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

#else

#define IX_RM_COMM_MESSAGE_INTERRUPT_DISABLE( \
                                              arg_Dummy \
                                            ) \
    do \
    { \
    } \
    while(0)

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */



/**
 * MACRO NAME: IX_RM_COMM_MESSAGE_INTERRUPT_ENABLE
 *
 * DESCRIPTION: This macro will enable the specified uCode to core 
 *          communication interrupt.
 *
 * @Param:  - IN - arg_Interrupt an ix_uint32 value specifying the interrupt to 
 *          be enabled
 *
 */

#if defined(_IX_RM_IMPL_HARDWARE_) && !defined (_IX_RM_TEST_AI_INTERRUPTS_)

    #if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

    #define IX_RM_COMM_MESSAGE_INTERRUPT_ENABLE( \
                                                  arg_Dummy \
                                               ) \
        do \
        { \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x10)) = ((1 << IX_COMM_MESSAGE_INTERRUPT_0) | \
            (1 << IX_COMM_MESSAGE_INTERRUPT_2)); \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x180)) = 0xffffffff; \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x188)) = 0xffffffff; \
        } \
        while(0)
    
    #else

    #define IX_RM_COMM_MESSAGE_INTERRUPT_ENABLE( \
                                                  arg_Dummy \
                                               ) \
        do \
        { \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x10)) = ((1 << IX_COMM_MESSAGE_INTERRUPT_0) | \
            (1 << IX_COMM_MESSAGE_INTERRUPT_1) | (1 << IX_COMM_MESSAGE_INTERRUPT_2) | \
            (1 << IX_COMM_MESSAGE_INTERRUPT_3)); \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x180)) = 0xffffffff; \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x184)) = 0xffffffff; \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x188)) = 0xffffffff; \
            *((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x18c)) = 0xffffffff; \
        } \
        while(0)

    #endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

#else

#define IX_RM_COMM_MESSAGE_INTERRUPT_ENABLE( \
                                             arg_Dummy \
                                           ) \
    do \
    { \
    } \
    while(0)
                                    
#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

/**
 * TYPENAME: ix_dispatch_engine_interface_func_ptr
 * 
 * DESCRIPTION: This is the definition of a function type for the dispatch engine
 *          interface. It is assumed that all functions of the interface are of this
 *          type. This approach will allow us to abstract the internal communication 
 *          API implementation. We will be able to allow the users to set their own 
 *          ISR and implement their own engine. It will be extremely useful for the
 *          case of a single threaded OS.
 */
typedef ix_error (*ix_dispatch_engine_interface_func_ptr)( 
                                                           void* arg_pContext,
                                                           ...
                                                         );



/**
 * TYPENAME: _ix_dispatch_engine_interface
 * 
 * DESCRIPTION: This type defines in an abstract way the communication dispatch engine
 *          interface. This abstraction provides support for the ME to core communication.
 *
 */
typedef struct _ix_s_dispatch_engine_interface
{
    /* this function will initialize the engine */
    ix_dispatch_engine_interface_func_ptr m_InitEngineFuncPtr;

    /* this function will finalize the engine */
    ix_dispatch_engine_interface_func_ptr m_FiniEngineFuncPtr;

    /* this function should be called in order to start the engine */
    ix_dispatch_engine_interface_func_ptr m_StartEngineFuncPtr;

    /* this function should be called in order to stop the engine */
    ix_dispatch_engine_interface_func_ptr m_StopEngineFuncPtr;

    /**
     * this function is the main processing function for the engine. In the case
     * of an engine with a separate thread then this function might be imagined as
     * the thread function.
     */
    ix_dispatch_engine_interface_func_ptr m_RunEngineFuncPtr;

    /* this function should be called in order to notify the engine that new data arrived */
    ix_dispatch_engine_interface_func_ptr m_NotifyEngineFuncPtr;

} _ix_dispatch_engine_interface;



/**
 * TYPENAME: _ix_default_dispatch_engine_context
 * 
 * DESCRIPTION: This type describes the default context for the dispatch engine interface
 *
 */
typedef struct _ix_s_default_dispatch_engine_context
{
    ix_uint32           m_StopProcessing; /* flag that signals that the engine should stop processing */
    ix_hw_ring_handle   m_hDataRing; /* this is the hardware ring where data is sent from the ME */
    ix_ossl_sem_t       m_SyncSemaphore; /* this is synchronization object used by the engine */
    ix_ossl_thread_t    m_DispatchThreadId;

    _ix_dispatch_engine_interface*  m_pEngineInterface; /* a reference to the coresponding engine interface */
} _ix_default_dispatch_engine_context;


/**
 * This is the internal communication semaphore pool
 */
extern ix_sem_pool* g_pCommSemaphorePool;


/**
 * Global end point manager
 */
extern ix_end_point_manager g_EndPointManager;

/**
 * This variable represents the default data handler for the end points.
 */
extern ix_comm_data_handler g_DefaultDataHandler;


#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__INTERNAL_COMMUNICATION_H__) */
