/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 4
 *
 * = LIBRARY
 *      Resource Manager Library
 *
 * = MODULE
 *      Library Interface
 *
 * = FILENAME
 *      internal_configuration_property.h
 *
 * = DESCRIPTION
 *      This file describes all internal types and function prototypes required for the
 *      internal implementation of the Configuration Property API.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      12/18/2001 8:23:24 AM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 * $Id: internal_configuration_property.h,v 1.8 2003/05/19 19:49:22 ageorges Exp $
 */

#if !defined(__INTERNAL_CONFIGURATION_PROPERTY_H__)
#define __INTERNAL_CONFIGURATION_PROPERTY_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */


/**
 * Configuration Property API implementation will make use of the sw buffer API. We
 * believe that the software buffer abstraction fits very well the requirements we have
 * for configuration properties.
 */



/**
 * DESCRIPTION: This symbol defines the size of the software buffer free list
 *  to be used by the configuration property manager.
 */
#define IX_CP_FREE_LIST_SIZE    2048UL


/**
 * DESCRIPTION: This symbol defines the size of the data associated with a configuration
 *  property software buffer
 *
 */
#define IX_CP_DATA_ELEMENT_SIZE     64UL

/**
 * DESCRIPTION: Defines the root of the uCode properties tree
 *
 */
#define IX_CP_UCODE_PROPERTY_ROOT g_hUCodeRegistryRoot


/**
 * TYPENAME: ix_cp_property_notification_token
 * 
 * DESCRIPTION: This structure defines a notification element that will be stored in
 *  a list of registered notification tokens for a particular property. At the 
 *  time a property changes, a message will be sent to the registered communication
 *  ID with the handle of the property that has changed. The processing of
 *  message can be done either in a synchronous or asynchromous way depending on
 *  communication ID listener setup.
 *
 */
typedef struct ix_s_cp_property_notification_token ix_cp_property_notification_token;

struct ix_s_cp_property_notification_token
{
    ix_communication_id                     m_CommunicationId; /* notification communication id */
    ix_cp_property_notification_token*      m_pNext; /* next notification token in the list */
};



/**
 * TYPENAME: ix_cp_property_attributes
 * 
 * DESCRIPTION: This structure represents the structure of the internal property attributes.
 *          
 *
 */
typedef struct ix_s_cp_property_attributes
{
    ix_sw_buffer_meta                   m_SwBufferMeta; /* required sw buffer meta */
    ix_uint32                           m_Options;
    ix_uint32                           m_ReferenceCount;
    ix_uint32                           m_DataSize;
    ix_configuration_property_handle    m_hParentProperty;
    ix_configuration_property_handle    m_hPropertyNext;
    ix_configuration_property_handle    m_hPropertyPrev;
    ix_configuration_property_handle    m_hChildrenPropertiesListHead;
    ix_configuration_property_handle    m_hChildrenPropertiesListTail;
    char                                m_aName[IX_CP_MAX_PROPERTY_NAME_LENGTH];
    ix_cp_property_notification_token*  m_pNotificationTokenList;
} ix_cp_property_attributes;



/**
 * TYPENAME: ix_cp_tree_predicate
 * 
 * DESCRIPTION: This type describes a predicate used to traverse the configuration
 *  property tree. The configuration property tree will be traversed till this 
 *  predicate will return a non NULL configuration property handle. That will be
 *  the handle for which the internal condition of the predicate has been satisfied.
 *
 */
typedef ix_configuration_property_handle (*ix_cp_tree_predicate)(ix_configuration_property_handle arg_hCP, void* arg_pContext);


/**
 * Predicate that will find if the configuration property has the passed name
 */
ix_configuration_property_handle _ix_cp_tree_predicate_has_name(
                                  ix_configuration_property_handle arg_hCP, 
                                  const char* arg_pCPName
                                  );

/**
 * Predicate that will delete the configuration property without any effects
 */
ix_configuration_property_handle _ix_cp_tree_predicate_delete(
                                  ix_configuration_property_handle arg_hCP, 
                                  void* arg_pContext
                                  );

#if 0 /* compile it out for metrics purposes */

/**
 * NAME: _ix_rm_cp_tree_pre_order_traverse
 *
 * DESCRIPTION: This function will 
 * 
 * @Param:  - IN arg_hCPRoot -  the CP that will be the root of this operation
 * @Param:  - IN arg_Predicate - the predicate to be used in this operation
 * @Param:  - IN arg_pPredicateContext - the context to be passed to the predicate
 *
 * @Return: IX_NULL_CONFIGURATION_PROPERTY_HANDLE for failure or a valid configuration
 *  property handle for success.
 */
ix_configuration_property_handle _ix_rm_cp_tree_pre_order_traverse( 
                                            ix_configuration_property_handle arg_hCPRoot, 
                                            ix_cp_tree_predicate arg_Predicate, 
                                            void* arg_pPredicateContext
                                          );

/**
 * NAME: _ix_rm_cp_tree_in_order_traverse
 *
 * DESCRIPTION: This function will 
 * 
 * @Param:  - IN arg_hCPRoot -  the CP that will be the root of this operation
 * @Param:  - IN arg_Predicate - the predicate to be used in this operation
 * @Param:  - IN arg_pPredicateContext - the context to be passed to the predicate
 *
 * @Return: IX_NULL_CONFIGURATION_PROPERTY_HANDLE for failure or a valid configuration
 *  property handle for success.
 */
ix_configuration_property_handle _ix_rm_cp_tree_in_order_traverse( 
                                           ix_configuration_property_handle arg_hCPRoot, 
                                           ix_cp_tree_predicate arg_Predicate, 
                                           void* arg_pPredicateContext
                                         );

#endif /* if 0 */


/**
 * NAME: _ix_rm_cp_tree_post_order_traverse
 *
 * DESCRIPTION: This function will 
 * 
 * @Param:  - IN arg_hCPRoot -  the CP that will be the root of this operation
 * @Param:  - IN arg_Predicate - the predicate to be used in this operation
 * @Param:  - IN arg_pPredicateContext - the context to be passed to the predicate
 *
 * @Return: IX_NULL_CONFIGURATION_PROPERTY_HANDLE for failure or a valid configuration
 *  property handle for success.
 */
ix_configuration_property_handle _ix_rm_cp_tree_post_order_traverse( 
                                             ix_configuration_property_handle arg_hCPRoot, 
                                             ix_cp_tree_predicate arg_Predicate, 
                                             void* arg_pPredicateContext
                                           );
/**
 * NAME: _ix_rm_cp_tree_child_forward_traverse
 *
 * DESCRIPTION: This function will 
 * 
 * @Param:  - IN arg_hCPRoot -  the CP that will be the root of this operation
 * @Param:  - IN arg_Predicate - the predicate to be used in this operation
 * @Param:  - IN arg_pPredicateContext - the context to be passed to the predicate
 *
 * @Return: IX_NULL_CONFIGURATION_PROPERTY_HANDLE for failure or a valid configuration
 *  property handle for success.
 */
ix_configuration_property_handle _ix_rm_cp_tree_children_forward_traverse( 
                                            ix_configuration_property_handle arg_hCPRoot, 
                                            ix_cp_tree_predicate arg_Predicate, 
                                            void* arg_pPredicateContext
                                          );


#if 0 /* compile it out for metrics purposes */

/**
 * NAME: _ix_rm_cp_tree_child_reverse_traverse
 *
 * DESCRIPTION: This function will 
 * 
 * @Param:  - IN arg_hCPRoot -  the CP that will be the root of this operation
 * @Param:  - IN arg_Predicate - the predicate to be used in this operation
 * @Param:  - IN arg_pPredicateContext - the context to be passed to the predicate
 *
 * @Return: IX_NULL_CONFIGURATION_PROPERTY_HANDLE for failure or a valid configuration
 *  property handle for success.
 */
ix_configuration_property_handle _ix_rm_cp_tree_children_reverse_traverse( 
                                            ix_configuration_property_handle arg_hCPRoot, 
                                            ix_cp_tree_predicate arg_Predicate, 
                                            void* arg_pPredicateContext
                                          );

#endif /* if 0 */

/**
 * TYPENAME: ix_cp_manager
 * 
 * DESCRIPTION: This structure represents the configuration property manager. 
 *  It manages both core and microcode property trees. 
 *
 */
typedef struct ix_s_cp_manager
{
    ix_buffer_free_list_handle          m_hCPFreeList; /* cp free list */

    ix_configuration_property_handle    m_hCoreRegistryRoot; /* core property tree root */

    /* notification support */
    ix_uint32                           m_NotificationInProgress;
    ix_configuration_property_handle    m_ActiveNotificationHandle;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter          m_SyncAdapter; /* synchronization adapter */

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

} ix_cp_manager;


/**
 * NAME: _ix_rm_configuration_property_init
 *
 * DESCRIPTION: This function should be called at the time the resource manager
 *  initializer function is called. The function will initialize all the Configuration
 *  Property API.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_configuration_property_init(void);


/**
 * NAME: _ix_rm_configuration_property_fini
 *
 * DESCRIPTION: This function should be called at the time the resource manager 
 *  finalizer function is called. The function will provide cleanup for Configuration
 *  Property API.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_configuration_property_fini(void);

/**
 * Global variable representing the configuration property manager.
 */
extern ix_cp_manager g_CPManager;


#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__INTERNAL_CONFIGURATION_PROPERTY_H__) */
