/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      internal_counter_64bit.h
 *
 * = DESCRIPTION
 *      This file will declare all types and function prototypes needed internally
 *      by the Counter 64 bit API.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      5/22/2002 4:14:54 PM - creation time 
 * ============================================================================
 * $Id: internal_counter_64bit.h,v 1.7 2003/09/19 21:15:28 ageorges Exp $
 */

#if !defined(__INTERNAL_COUNTER_64BIT_H__)
#define __INTERNAL_COUNTER_64BIT_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */

/**
 * TYPENAME: _ix_counter_64bit
 * 
 * DESCRIPTION: This structure describes the internal structure of a 64 bit counter
 *
 */
typedef struct _ix_s_counter_64bit _ix_counter_64bit;


/**
 * TYPENAME: ix_update_64bit_counter_func_ptr
 * 
 * DESCRIPTION: This function type is generic update function for each 64 bit counter
 *
 */
typedef void (*ix_update_64bit_counter_func_ptr)(_ix_counter_64bit* arg_p64bitCounter, void* arg_pContext);


/**
 * TYPENAME: _ix_64bit_counter_update_context
 * 
 * DESCRIPTION: This structure will be passed as a context to the counter updating 
 *  functions. The m_ActiveTimeout member is the active TO which represents the time elapsed
 *  since last update operation. The m_NewTimeout member represents the computed TO for
 *  the next update operation.
 *
 */
typedef struct _ix_s_64bit_counter_update_context
{
    ix_uint32 m_ActiveTimeout;
    ix_uint32 m_NewTimeout;
} _ix_64bit_counter_update_context;



/**
 * TYPENAME: _ix_counter_64bit
 * 
 * DESCRIPTION: This structure describes the internal structure of a 64 bit counter
 *
 */
struct _ix_s_counter_64bit
{
    ix_uint64           m_CoreValue; /* the core side of the counter */
    ix_uint32*          m_pMEValue;  /* the ME side of the counter */
    ix_uint32           m_Timeout; /* 32 bit counter overflow timeout expressed in milliseconds */
    ix_uint32           m_UpdateElapsedTime; /* time elapsed since last update */
    ix_uint32           m_Properties;

    ix_update_64bit_counter_func_ptr    m_UpdateFuncPtr; /* timer update function */

    _ix_counter_64bit*      m_pNext;
    ix_counter_64bit_handle m_hCounter; /* handle of the counter */
};



/**
 * DESCRIPTION: This mask will be used to differentiate between regular numbers and 
 *  counter handles
 *
 */
#define IX_64BIT_COUNTER_HANDLE_SIGNATURE    0x5f8a88dd


/**
 * DESCRIPTION: Defines the LSB of the counter index in a given block
 *
 */
#define IX_64BIT_COUNTER_HANDLE_COUNTER_INDEX_LSB   0UL 


/**
 * DESCRIPTION: Defines the MSB of the counter index in a given block
 *
 */
#define IX_64BIT_COUNTER_HANDLE_COUNTER_INDEX_MSB   9UL 


/**
 * DESCRIPTION: Defines the LSB of the block index of a counter
 *
 */
#define IX_64BIT_COUNTER_HANDLE_BLOCK_INDEX_LSB     10UL 


/**
 * DESCRIPTION: Defines the MSB of the block index of a counter
 *
 */
#define IX_64BIT_COUNTER_HANDLE_BLOCK_INDEX_MSB     19UL 



/**
 * DESCRIPTION: Defines how many counter blocks will be allocated into the system.
 *  One block will contain exactly IX_64BIT_COUNTER_PER_BLOCK counters.
 *
 */
#define IX_64BIT_COUNTER_BLOCKS        (1UL << (IX_64BIT_COUNTER_HANDLE_BLOCK_INDEX_MSB - \
    IX_64BIT_COUNTER_HANDLE_BLOCK_INDEX_LSB + 1UL))


/**
 * DESCRIPTION: Defines how many counter blocks will be initialized at the system startup.
 *  The other ones will be initialized only when extra timers will be required, just to
 *  conserve resources.
 *
 */
#define IX_64BIT_COUNTER_STARTUP_INITIALIZED_BLOCKS        4UL 


/**
 * DESCRIPTION: This symbol defines how many counters will be allocated in each counter
 *  block.
 *
 */
#define IX_64BIT_COUNTER_PER_BLOCK     (1UL << (IX_64BIT_COUNTER_HANDLE_COUNTER_INDEX_MSB - \
    IX_64BIT_COUNTER_HANDLE_COUNTER_INDEX_LSB + 1UL))



/**
 * MACRO NAME: IX_64BIT_COUNTER_HANDLE_GET_COUNTER_INDEX
 *
 * DESCRIPTION: Retrieves the counter index value in a block for the passed handle
 *
 * @Param:  - IN arg_h64bitCounter - handle of the 64 bit counter of interest
 *
 * @Return: an ix_uint32 value representing the counter index value in the block
 */
#define IX_64BIT_COUNTER_HANDLE_GET_COUNTER_INDEX( \
                                                   arg_h64bitCounter \
                                                 ) \
                                                 IX_GET_BIT_FIELD32(arg_h64bitCounter, \
                                                    IX_64BIT_COUNTER_HANDLE_COUNTER_INDEX_LSB, \
                                                    IX_64BIT_COUNTER_HANDLE_COUNTER_INDEX_MSB)

/**
 * MACRO NAME: IX_64BIT_COUNTER_HANDLE_GET_COUNTER_BLOCK_INDEX
 *
 * DESCRIPTION: Retrieves the counter block index value for the passed handle
 *
 * @Param:  - IN arg_h64bitCounter - handle of the 64 bit counter of interest
 *
 * @Return: an ix_uint32 value representing the counter block index value
 */
#define IX_64BIT_COUNTER_HANDLE_GET_COUNTER_BLOCK_INDEX( \
                                                   arg_h64bitCounter \
                                                 ) \
                                                 IX_GET_BIT_FIELD32(arg_h64bitCounter, \
                                                    IX_64BIT_COUNTER_HANDLE_BLOCK_INDEX_LSB, \
                                                    IX_64BIT_COUNTER_HANDLE_BLOCK_INDEX_MSB)
                                                 

/**
 * MACRO NAME: IX_64BIT_COUNTER_MAKE_HANDLE
 *
 * DESCRIPTION: This macro will create a 64 bit counter handle based on counter index
 *  and block index. 
 *
 * @Param:  - IN arg_CounterIndex - this is a ix_uint32 value representing counter index in a block
 * @Param:  - IN arg_BlockIndex - this is a ix_uint32 value representing block index of the counter
 *
 * @Return: returns a ix_counter_64bit_handle coresponding to the passed data.
 */
#define IX_64BIT_COUNTER_MAKE_HANDLE( \
                                 arg_CounterIndex, \
                                 arg_BlockIndex \
                               ) \
                               ((ix_counter_64bit_handle) /* typecast */ \
                               (IX_BIT_FIELD_MASK32(IX_64BIT_COUNTER_HANDLE_BLOCK_INDEX_MSB + 1U, 31U) & \
                               IX_64BIT_COUNTER_HANDLE_SIGNATURE) | \
                               IX_MAKE_BIT_FIELD32(arg_CounterIndex, IX_64BIT_COUNTER_HANDLE_COUNTER_INDEX_LSB, \
                               IX_64BIT_COUNTER_HANDLE_COUNTER_INDEX_MSB) | \
                               IX_MAKE_BIT_FIELD32(arg_BlockIndex, IX_64BIT_COUNTER_HANDLE_BLOCK_INDEX_LSB, \
                               IX_64BIT_COUNTER_HANDLE_BLOCK_INDEX_MSB))                   
                               

/**
 * MACRO NAME: IX_64BIT_COUNTER_HANDLE_IS_VALID
 *
 * DESCRIPTION: This macro will check if the passed 64 bit counter handle is valid
 *
 * @Param:  - IN arg_h64bitCounter - handle of the 64 bit counter of interest
 *
 * @Return: Returns non zero if the handle is valid or zero otherwise.
 */
#define IX_64BIT_COUNTER_HANDLE_IS_VALID( \
                                          arg_h64bitCounter \
                                        ) \
                                        ((((IX_64BIT_COUNTER_HANDLE_SIGNATURE ^ (arg_h64bitCounter)) & \
                                           IX_BIT_FIELD_MASK32(IX_64BIT_COUNTER_HANDLE_BLOCK_INDEX_MSB + 1U, 31U)) == 0) ? 1 : 0)

/**
 * DESCRIPTION: This symbol describes the default timeout for a 32 bit part of the
 *  counter expressed in milliseconds.
 *
 */
#define IX_64BIT_COUNTER_DEFAULT_TIMEOUT    50UL 



/**
 * DESCRIPTION: This symbol mask defines if a counter is active or it is in the free
 *  list (inactive).
 *
 */
#define IX_64BIT_COUNTER_PROPERTY_ACTIVE    0x01UL

/**
 * TYPENAME: _ix_counter_block
 * 
 * DESCRIPTION: This type defines a counter block type
 *
 */
typedef _ix_counter_64bit _ix_counter_block[IX_64BIT_COUNTER_PER_BLOCK];



/**
 * TYPENAME: _ix_64bit_counter_manager
 * 
 * DESCRIPTION: This structure describes a 64 bit counter manager
 *
 */
typedef struct _ix_s_64bit_counter_manager
{
    _ix_counter_64bit*  m_CounterBlocks[IX_64BIT_COUNTER_BLOCKS];
    ix_uint32           m_InitializedCounterBlocks;
    _ix_counter_64bit*  m_pCounterFreeList;
    ix_uint32           m_FreeCounters;
    _ix_counter_64bit*  m_pActiveCounterList;

    ix_uint32           m_ActiveTO; /* active timeout for updating timers */

    /* these following members are used for the updating mechanism */
    ix_ossl_sem_t       m_UpdateSemaphore; 
    ix_ossl_thread_t    m_UpdateThreadId;
    ix_uint32           m_UpdateProcessingDone;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter      m_SyncAdapter; /* synchronization adapter */

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

} _ix_64bit_counter_manager;


/**
 * NAME: _ix_rm_counter_64bit_sram_update
 *
 * DESCRIPTION: This function will perform the sram 64 bit update counter operation
 * 
 * @Param:  - INOUT arg_p64bitCounter - counter to be updated
 * @Param:  - INOUT arg_pContext - external data to be passed to the update operation
 *
 */
void _ix_rm_counter_64bit_sram_update(
                                       _ix_counter_64bit* arg_p64bitCounter, 
                                       void* arg_pContext
                                     );

/**
 * NAME: _ix_rm_counter_64bit_scratch_update
 *
 * DESCRIPTION: This function will perform the scratch 64 bit update counter operation
 * 
 * @Param:  - INOUT arg_p64bitCounter - counter to be updated
 * @Param:  - INOUT arg_pContext - external data to be passed to the update operation
 *
 */
void _ix_rm_counter_64bit_scratch_update(
                                          _ix_counter_64bit* arg_p64bitCounter, 
                                          void* arg_pContext
                                        );


/**
 * NAME: _ix_rm_counter_64bit_initialize_new_block
 *
 * DESCRIPTION: This function will initialize the next unused counter block for
 *  future use.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_counter_64bit_initialize_new_block(void);


/**
 * NAME: _ix_rm_counter_64bit_add_to_active_list
 *
 * DESCRIPTION: This function will add the passed counter to the active counter list
 * 
 * @Param:  - IN arg_p64bitCounter address of the counter to be added to active counter list
 *
 */
void _ix_rm_counter_64bit_add_to_active_list(
                                              _ix_counter_64bit* arg_p64bitCounter
                                            );

/**
 * NAME: _ix_rm_counter_64bit_remove_from_active_list
 *
 * DESCRIPTION: This function will remove the passed counter from the active counter list
 * 
 * @Param:  - IN arg_p64bitCounter address of the counter to be added to active counter list
 *
 */
void _ix_rm_counter_64bit_remove_from_active_list(
                                                   _ix_counter_64bit* arg_p64bitCounter
                                                 );

/**
 * NAME: _ix_rm_counter_64bit_update
 *
 * DESCRIPTION: This function will force an update of all counters
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_counter_64bit_update(void);


/**
 * NAME: _ix_rm_counter_64bit_init
 *
 * DESCRIPTION: This function should be called at the time the resource manager initializer 
 *          function is called. The function will initialize all the Counter 64 bit API.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_counter_64bit_init(void);


/**
 * NAME: _ix_rm_counter_64bit_fini
 *
 * DESCRIPTION: This function should be called at the time the resource manager finalizer function
 *          is called. The function will provide cleanup for Counter 64 bit API.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_counter_64bit_fini(void);



/**
 * Global 64 bit counter manager
 */
extern _ix_64bit_counter_manager g_64bitCounterManager;

#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__INTERNAL_COUNTER_64BIT_H__) */
