/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 4
 *
 * = LIBRARY
 *      Resource Manager Library
 *
 * = MODULE
 *      Hardware Resource Manager API
 *
 * = FILENAME
 *      internal_hardware.h
 *
 * = DESCRIPTION
 *      This file defines all data types and function prototypes required internaly by the 
 *      Hardware Resource Manager API module.
 *
 * = AUTHOR
 *       Adrian Georgescu
 *       adrian.georgescu@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      10/26/2001 11:29:08 AM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 * $Id: internal_hardware.h,v 1.5 2003/09/19 21:11:03 ageorges Exp $
 */

#if !defined(__INTERNAL_HARDWARE_H__)
#define __INTERNAL_HARDWARE_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */



/**
 * TYPENAME: ix_sram_queue_descriptor
 * 
 * DESCRIPTION: This structure represents the SRAM mapping of a queue descriptor
 *          structure. Some of the hardware specified fields are packed in a longword
 *          and we have chosen to provide access and modifier macros instead of using
 *          bit fields in order to avoid endianess problems.
 */
typedef struct ix_s_sram_queue_descriptor
{
    ix_uint32   m_HeadInfo; /* this field packs EOP, cell count and the head pointer*/
    ix_uint32   m_TailInfo; /* this field packs the tail pointer */
    ix_uint32   m_CountInfo; /* this field packs the the queue count */
    ix_uint32   m_Reserved;
} ix_sram_queue_descriptor;



/**
 * MACRO NAME: IX_RM_GET_QUEUE_DESC_EOP
 *
 * DESCRIPTION: This macro retrieves the EOP field from the m_HeadInfo
 *          member of a ix_sram_queue_descriptor structure.
 *
 * @Param:  - IN arg_HeadInfo - an ix_uint32 value representing the m_HeadInfo member of a 
 *          ix_sram_queue_descriptor structure.
 *
 * @Return: Returns an ix_uint32 value that is non zero if the top of the queue buffer
 *          is marked as EOP. In this case on the next dequeue operation the Q_count field
 *          will be decremented.
 */
#define IX_RM_GET_QUEUE_DESC_EOP( \
                                  arg_HeadInfo \
                                ) \
                                IX_GET_BIT_FIELD32(arg_HeadInfo, 31U, 31U)


/**
 * MACRO NAME: IX_RM_SET_QUEUE_DESC_EOP
 *
 * DESCRIPTION: This macro sets the EOP field from the m_HeadInfo
 *          member of a ix_sram_queue_descriptor structure.
 *
 * @Param:  - IN arg_HeadInfo - an ix_uint32 value representing the m_HeadInfo member of a 
 *          ix_sram_queue_descriptor structure.
 * @Param:  - IN arg_EOPFlag - an ix_uint32 value representing the new value of the EOP field. The allowed
 *          values are 0 and 1.
 *
 * @Return: Returns the new value of arg_HeadInfo.
 */
#define IX_RM_SET_QUEUE_DESC_EOP( \
                                  arg_HeadInfo, \
                                  arg_EOPFlag \
                                ) \
                                IX_SET_BIT_FIELD32(arg_HeadInfo, arg_EOPFlag, 31U, 31U)


/**
 * MACRO NAME: IX_RM_GET_QUEUE_DESC_SOP
 *
 * DESCRIPTION: This macro retrieves the SOP field from the m_HeadInfo
 *          member of a ix_sram_queue_descriptor structure.
 *
 * @Param:  - IN arg_HeadInfo - an ix_uint32 value representing the m_HeadInfo member of a 
 *          ix_sram_queue_descriptor structure.
 *
 * @Return: Returns an ix_uint32 value that is non zero if the top of the queue buffer
 *          is marked as SOP. 
 */
#define IX_RM_GET_QUEUE_DESC_SOP( \
                                  arg_HeadInfo \
                                ) \
                                IX_GET_BIT_FIELD32(arg_HeadInfo, 30U, 30U)


/**
 * MACRO NAME: IX_RM_SET_QUEUE_DESC_SOP
 *
 * DESCRIPTION: This macro sets the SOP field from the m_HeadInfo
 *          member of a ix_sram_queue_descriptor structure.
 *
 * @Param:  - IN arg_HeadInfo - an ix_uint32 value representing the m_HeadInfo member of a 
 *          ix_sram_queue_descriptor structure.
 * @Param:  - IN arg_SOPFlag - an ix_uint32 value representing the new value of the SOP field. The allowed
 *          values are 0 and 1.
 *
 * @Return: Returns the new value of arg_HeadInfo.
 */
#define IX_RM_SET_QUEUE_DESC_SOP( \
                                  arg_HeadInfo, \
                                  arg_SOPFlag \
                                ) \
                                IX_SET_BIT_FIELD32(arg_HeadInfo, arg_SOPFlag, 30U, 30U)
                                
                                
/**
 * MACRO NAME: IX_RM_GET_QUEUE_DESC_CELL_COUNT
 *
 * DESCRIPTION: This macro retrieves the cell count field from the m_HeadInfo
 *          member of a ix_sram_queue_descriptor structure.
 *
 * @Param:  - IN arg_HeadInfo - an ix_uint32 value representing the m_HeadInfo member of a 
 *          ix_sram_queue_descriptor structure.
 *
 * @Return: Returns an ix_uint32 that is the value of the cell count field from the m_HeadInfo.
 */
#define IX_RM_GET_QUEUE_DESC_CELL_COUNT( \
                                         arg_HeadInfo \
                                       ) \
                                       IX_GET_BIT_FIELD32(arg_HeadInfo, 24U, 29U)


/**
 * MACRO NAME: IX_RM_SET_QUEUE_DESC_CELL_COUNT
 *
 * DESCRIPTION: This macro sets the cell count field from the m_HeadInfo
 *          member of a ix_sram_queue_descriptor structure.
 *
 * @Param:  - IN arg_HeadInfo - an ix_uint32 value representing the m_HeadInfo member of a 
 *          ix_sram_queue_descriptor structure.
 * @Param:  - IN arg_CellCount - an ix_uint32 value representing the new value  of the cell 
 *          count field.
 *
 * @Return: Returns the new value of arg_HeadInfo.
 */
#define IX_RM_SET_QUEUE_DESC_CELL_COUNT( \
                                         arg_HeadInfo, \
                                         arg_CellCount \
                                       ) \
                                       IX_SET_BIT_FIELD32(arg_HeadInfo, arg_CellCount, 24U, 29U)


/**
 * MACRO NAME: IX_RM_GET_QUEUE_DESC_HEAD
 *
 * DESCRIPTION: This macro retrieves the head field from the m_HeadInfo
 *          member of a ix_sram_queue_descriptor structure.
 *
 * @Param:  - IN arg_HeadInfo - an ix_uint32 value representing the m_HeadInfo member of a 
 *          ix_sram_queue_descriptor structure.
 *
 * @Return: Returns an ix_uint32 that is the value of the head field from the m_HeadInfo.
 */
#define IX_RM_GET_QUEUE_DESC_HEAD( \
                                   arg_HeadInfo \
                                 ) \
                                 IX_GET_BIT_FIELD32(arg_HeadInfo, 0U, 23U)


/**
 * MACRO NAME: IX_RM_SET_QUEUE_DESC_HEAD
 *
 * DESCRIPTION: This macro sets the head field from the m_HeadInfo
 *          member of a ix_sram_queue_descriptor structure.
 *
 * @Param:  - IN arg_HeadInfo - an ix_uint32 value representing the m_HeadInfo member of a 
 *          ix_sram_queue_descriptor structure.
 * @Param:  - IN arg_Head - an ix_uint32 value representing the new value of the head field.
 *
 * @Return: Returns the new value of arg_HeadInfo.
 */
#define IX_RM_SET_QUEUE_DESC_HEAD( \
                                   arg_HeadInfo, \
                                   arg_Head \
                                 ) \
                                 IX_SET_BIT_FIELD32(arg_HeadInfo, arg_Head, 0U, 23U)


/**
 * MACRO NAME: IX_RM_GET_QUEUE_DESC_TAIL
 *
 * DESCRIPTION: This macro retrieves the tail field from the m_TailInfo
 *          member of a ix_sram_queue_descriptor structure.
 *
 * @Param:  - IN arg_TailInfo - an ix_uint32 value representing the m_TailInfo member of a 
 *          ix_sram_queue_descriptor structure.
 *
 * @Return: Returns an ix_uint32 that is the value of the tail field from the m_TailInfo.
 */
#define IX_RM_GET_QUEUE_DESC_TAIL( \
                                   arg_TailInfo \
                                 ) \
                                 IX_GET_BIT_FIELD32(arg_TailInfo, 0U, 23U)


/**
 * MACRO NAME: IX_RM_SET_QUEUE_DESC_TAIL
 *
 * DESCRIPTION: This macro sets the tail field from the m_HeadInfo
 *          member of a ix_sram_queue_descriptor structure.
 *
 * @Param:  - IN arg_TailInfo - an ix_uint32 value representing the m_TailInfo member of a 
 *          ix_sram_queue_descriptor structure.
 * @Param:  - IN arg_Tail - an ix_uint32 value representing the new value of the tail field.
 *
 * @Return: Returns the new value of arg_TailInfo.
 */
#define IX_RM_SET_QUEUE_DESC_TAIL( \
                                   arg_TailInfo, \
                                   arg_Tail \
                                 ) \
                                 IX_SET_BIT_FIELD32(arg_TailInfo, arg_Tail, 0U, 23U)


/**
 * MACRO NAME: IX_RM_GET_QUEUE_DESC_Q_COUNT
 *
 * DESCRIPTION: This macro retrieves the Q_count field from the m_CountInfo
 *          member of a ix_sram_queue_descriptor structure.
 *
 * @Param:  - IN arg_CountInfo - an ix_uint32 value representing the m_CountInfo member of a 
 *          ix_sram_queue_descriptor structure.
 *
 * @Return: Returns an ix_uint32 that is the value of the Q_count field from the m_CountInfo.
 */
#define IX_RM_GET_QUEUE_DESC_Q_COUNT( \
                                      arg_CountInfo \
                                    ) \
                                    IX_GET_BIT_FIELD32(arg_CountInfo, 0U, 23U)


/**
 * MACRO NAME: IX_RM_SET_QUEUE_DESC_Q_COUNT
 *
 * DESCRIPTION: This macro sets the Q_count field from the m_CountInfo
 *          member of a ix_sram_queue_descriptor structure.
 *
 * @Param:  - IN arg_CountInfo - an ix_uint32 value representing the m_CountInfo member of a 
 *          ix_sram_queue_descriptor structure.
 * @Param:  - IN arg_QCount - an ix_uint32 value representing the new value of the Q_count field.
 *
 * @Return: Returns the new value of arg_CountInfo.
 */
#define IX_RM_SET_QUEUE_DESC_Q_COUNT( \
                                      arg_CountInfo, \
                                      arg_QCount \
                                    ) \
                                    IX_SET_BIT_FIELD32(arg_CountInfo, arg_QCount, 0U, 23U)


/**
 * MACRO NAME: IX_RM_GET_QUEUE_DESC_SW_PRIVATE
 *
 * DESCRIPTION: This macro retrieves the SW_Private field from the m_CountInfo
 *          member of a ix_sram_queue_descriptor structure.
 *
 * @Param:  - IN arg_CountInfo - an ix_uint32 value representing the m_CountInfo member of a 
 *          ix_sram_queue_descriptor structure.
 *
 * @Return: Returns an ix_uint32 that is the value of the SW_Private field from the m_CountInfo.
 */
#define IX_RM_GET_QUEUE_DESC_SW_PRIVATE( \
                                         arg_CountInfo \
                                       ) \
                                       IX_GET_BIT_FIELD32(arg_CountInfo, 24U, 31U)


/**
 * MACRO NAME: IX_RM_SET_QUEUE_DESC_SW_PRIVATE
 *
 * DESCRIPTION: This macro sets the SW_Private field from the m_CountInfo
 *          member of a ix_sram_queue_descriptor structure.
 *
 * @Param:  - IN arg_CountInfo - an ix_uint32 value representing the m_CountInfo member of a 
 *          ix_sram_queue_descriptor structure.
 * @Param:  - IN arg_SwPrivate - an ix_uint32 value representing the new value of the SW_Private field.
 *
 * @Return: Returns the new value of arg_CountInfo.
 */
#define IX_RM_SET_QUEUE_DESC_SW_PRIVATE( \
                                         arg_CountInfo, \
                                         arg_SwPrivate \
                                       ) \
                                       IX_SET_BIT_FIELD32(arg_CountInfo, arg_SwPrivate, 24U, 31U)


/**
 * TYPENAME: ix_sram_ring_descriptor
 * 
 * DESCRIPTION: This structure represents the SRAM mapping of a ring descriptor
 *          structure. Some of the hardware specified fields are packed in a longword
 *          and we have chosen to provide access and modifier macros, instead of using
 *          bit fields in order to avoid endianess problems.
 */
typedef struct ix_s_sram_ring_descriptor
{
    ix_uint32   m_HeadInfo; /* this field packs ring size and the get pointer*/
    ix_uint32   m_TailInfo; /* this field packs the put pointer */
    ix_uint32   m_CountInfo; /* this field packs the the ring count */
    ix_uint32   m_Reserved;
} ix_sram_ring_descriptor;




/**
 * MACRO NAME: IX_RM_GET_RING_DESC_RING_SIZE
 *
 * DESCRIPTION: This macro retrieves the ring size field from the m_HeadInfo
 *          member of a ix_sram_ring_descriptor structure.
 *
 * @Param:  - IN arg_HeadInfo - an ix_uint32 value representing the m_HeadInfo member of a 
 *          ix_sram_ring_descriptor structure.
 *
 * @Return: Returns an ix_uint32 that is the value of the ring size field from the m_HeadInfo.
 */
#define IX_RM_GET_RING_DESC_RING_SIZE( \
                                        arg_HeadInfo \
                                      ) \
                                      IX_GET_BIT_FIELD32(arg_HeadInfo, 29U, 31U)


/**
 * MACRO NAME: IX_RM_SET_RING_DESC_RING_SIZE
 *
 * DESCRIPTION: This macro sets the ring size field from the m_HeadInfo
 *          member of a ix_sram_ring_descriptor structure.
 *
 * @Param:  - IN arg_HeadInfo - an ix_uint32 value representing the m_HeadInfo member of a 
 *          ix_sram_ring_descriptor structure.
 * @Param:  - IN arg_RingSize - an ix_uint32 value representing the new value of the ring 
 *          size field.
 *
 * @Return: Returns the new value of arg_HeadInfo.
 */
#define IX_RM_SET_RING_DESC_RING_SIZE( \
                                       arg_HeadInfo, \
                                       arg_RingSize \
                                     ) \
                                     IX_SET_BIT_FIELD32(arg_HeadInfo, arg_RingSize, 29U, 31U)


/**
 * MACRO NAME: IX_RM_GET_RING_DESC_HEAD
 *
 * DESCRIPTION: This macro retrieves the head field from the m_HeadInfo
 *          member of a ix_sram_ring_descriptor structure.
 *
 * @Param:  - IN arg_HeadInfo - an ix_uint32 value representing the m_HeadInfo member of a 
 *          ix_sram_ring_descriptor structure.
 *
 * @Return: Returns an ix_uint32 that is the value of the head field from the m_HeadInfo.
 */
#define IX_RM_GET_RING_DESC_HEAD( \
                                   arg_HeadInfo \
                                 ) \
                                 IX_GET_BIT_FIELD32(arg_HeadInfo, 0U, 23U)


/**
 * MACRO NAME: IX_RM_SET_RING_DESC_HEAD
 *
 * DESCRIPTION: This macro sets the head field from the m_HeadInfo
 *          member of a ix_sram_ring_descriptor structure.
 *
 * @Param:  - IN arg_HeadInfo - an ix_uint32 value representing the m_HeadInfo member of a 
 *          ix_sram_ring_descriptor structure.
 * @Param:  - IN arg_Head - an ix_uint32 value representing the new value of the head field.
 *
 * @Return: Returns the new value of arg_HeadInfo.
 */
#define IX_RM_SET_RING_DESC_HEAD( \
                                   arg_HeadInfo, \
                                   arg_Head \
                                 ) \
                                 IX_SET_BIT_FIELD32(arg_HeadInfo, arg_Head, 0U, 23U)


/**
 * MACRO NAME: IX_RM_GET_RING_DESC_TAIL
 *
 * DESCRIPTION: This macro retrieves the tail field from the m_TailInfo
 *          member of a ix_sram_ring_descriptor structure.
 *
 * @Param:  - IN arg_TailInfo - an ix_uint32 value representing the m_TailInfo member of a 
 *          ix_sram_ring_descriptor structure.
 *
 * @Return: Returns an ix_uint32 that is the value of the tail field from the m_TailInfo.
 */
#define IX_RM_GET_RING_DESC_TAIL( \
                                   arg_TailInfo \
                                 ) \
                                 IX_GET_BIT_FIELD32(arg_TailInfo, 0U, 23U)


/**
 * MACRO NAME: IX_RM_SET_RING_DESC_TAIL
 *
 * DESCRIPTION: This macro sets the tail field from the m_HeadInfo
 *          member of a ix_sram_ring_descriptor structure.
 *
 * @Param:  - IN arg_TailInfo - an ix_uint32 value representing the m_TailInfo member of a 
 *          ix_sram_ring_descriptor structure.
 * @Param:  - IN arg_Tail - an ix_uint32 value representing the new value of the tail field.
 *
 * @Return: Returns the new value of arg_TailInfo.
 */
#define IX_RM_SET_RING_DESC_TAIL( \
                                   arg_TailInfo, \
                                   arg_Tail \
                                 ) \
                                 IX_SET_BIT_FIELD32(arg_TailInfo, arg_Tail, 0U, 23U)


/**
 * MACRO NAME: IX_RM_GET_RING_DESC_RING_COUNT
 *
 * DESCRIPTION: This macro retrieves the ring count field from the m_CountInfo
 *          member of a ix_sram_ring_descriptor structure.
 *
 * @Param:  - IN arg_CountInfo - an ix_uint32 value representing the m_CountInfo member of a 
 *          ix_sram_ring_descriptor structure.
 *
 * @Return: Returns an ix_uint32 that is the value of the ring count field from the m_CountInfo.
 */
#define IX_RM_GET_RING_DESC_RING_COUNT( \
                                        arg_CountInfo \
                                      ) \
                                      IX_GET_BIT_FIELD32(arg_CountInfo, 0U, 23U)


/**
 * MACRO NAME: IX_RM_SET_RING_DESC_RING_COUNT
 *
 * DESCRIPTION: This macro sets the ring count field from the m_CountInfo
 *          member of a ix_sram_ring_descriptor structure.
 *
 * @Param:  - IN arg_CountInfo - an ix_uint32 value representing the m_CountInfo member of a 
 *          ix_sram_ring_descriptor structure.
 * @Param:  - IN arg_RingCount - an ix_uint32 value representing the new value of the ring count field.
 *
 * @Return: Returns the new value of arg_CountInfo.
 */
#define IX_RM_SET_RING_DESC_RING_COUNT( \
                                        arg_CountInfo, \
                                        arg_RingCount \
                                      ) \
                                      IX_SET_BIT_FIELD32(arg_CountInfo, arg_RingCount, 0U, 23U)



/**
 * TYPENAME: ix_sram_qarray_descriptor
 * 
 * DESCRIPTION: This union describes a QArray descriptor that can be either
 *          a queue descriptor or a ring descriptor.
 */
typedef union ix_u_sram_qarray_descriptor
{
    ix_sram_queue_descriptor    m_QueueDescriptor;
    ix_sram_ring_descriptor     m_RingDescriptor;
} ix_sram_qarray_descriptor;



/**
 * TYPENAME: ix_ring_op_impl_data
 * 
 * DESCRIPTION: This structure will include extra information for ring atomicity
 *  in enqueue and dequeue operations, and supporting data for the operations themselves,
 *  plus some extra ring information
 *
 */
typedef struct ix_s_ring_op_impl_data
{
    ix_uint32       m_RingSize;
    ix_uint32*      m_pRingBaseAddress;

#if defined(_IX_RM_MULTI_THREADED_)

    ix_ossl_sem_t   m_Semaphore; /* used for operation atomicity */

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

} ix_ring_op_impl_data;



/**
 * TYPENAME: ix_sram_qarray_channel_manager
 * 
 * DESCRIPTION: This structure defines a core side QArray manager for each SRAM channel.
 *          The m_aQArrayDescriptors array holds the descriptors information and the
 *          m_aQArrayUsage array describes the descriptor usage (e.g. if it is in use,
 *          if it is a queue or ring descriptor, etc.).
 *
 */
typedef struct ix_s_sram_qarray_channel_manager
{
    ix_sram_qarray_descriptor m_aQArrayDescriptors[IX_MAX_SRAM_QARRAY_ENTRIES];
    ix_uint32 m_aQArrayUsage[IX_MAX_SRAM_QARRAY_ENTRIES];
} ix_sram_qarray_channel_manager;


/**
 * TYPENAME: ix_sram_channel_ring_op_impl_data
 * 
 * DESCRIPTION: 
 *
 */
typedef ix_ring_op_impl_data ix_sram_channel_ring_op_impl_data[IX_MAX_SRAM_QARRAY_ENTRIES];



/**
 * TYPENAME: ix_sram_qarray_manager
 * 
 * DESCRIPTION: This is the system wide qarray manager structure
 *
 */
typedef struct ix_s_sram_qarray_manager
{
    /**
     * This is an array to to pointers to QArray channel managers.
     * The QArray channel managers have to be allocated in the common memory because part
     * of the initialization will be done by the uCode, so it need sto have access
     * to that memory.
     */
    ix_sram_qarray_channel_manager* m_aQArrayManagersPtr[IX_SRAM_CHANNELS_NUMBER];

    /* This array defines data required for implementation of ring put/get operations */
    ix_sram_channel_ring_op_impl_data m_aRingOpImplData[IX_SRAM_CHANNELS_NUMBER];

#if defined(_IX_RM_MULTI_THREADED_)

    ix_synchronization_adapter      m_SyncAdapter; /* synchronization adapter */

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

} ix_sram_qarray_manager;




/**
 * TYPENAME: ix_sram_qarray_entry_type
 * 
 * DESCRIPTION: This enumeration describes the possible qarray entry types.
 *
 */
typedef enum ix_e_sram_qarray_entry_type
{
    IX_SRAM_QARRAY_ENTRY_TYPE_FIRST = 0,
    IX_SRAM_QARRAY_ENTRY_TYPE_QUEUE = IX_SRAM_QARRAY_ENTRY_TYPE_FIRST,
    IX_SRAM_QARRAY_ENTRY_TYPE_RING,
    IX_SRAM_QARRAY_ENTRY_TYPE_LAST
} ix_sram_qarray_entry_type;


/**
 * The queue array descriptor usage double word will be mapped as follows:
 *      LW	Bits 	Size	Field	    Description
 *      -----------------------------------------------------------------
 *      0	 0:0	1	    In use flag	This entry is already reserved if the flag is 1
 *           1:1	1	    queue/ring  This field will be 0 for queue descriptor 
 *                          flag        and 1 for ring descriptor.
 *	         2:31	30	    Reserved	Reserved for future use.
 */


/**
 * MACRO NAME: IX_RM_GET_QARRAY_ENTRY_IN_USE
 *
 * DESCRIPTION: This macro will determine if the specified qarray entry is in use based on the
 *          corresponding QArray usage field.
 *
 * @Param:  - IN arg_QArrayUsage - this is a ix_uint32 QArray usage field coresponding to the
 *          QArray entry.
 *
 * @Return: Returns a ix_uint32 value that will be 1 if the entry is in use or 0 if not.
 */
#define IX_RM_GET_QARRAY_ENTRY_IN_USE( \
                                       arg_QArrayUsage \
                                     ) \
                                     IX_GET_BIT_FIELD32(arg_QArrayUsage, 0U, 0U)


/**
 * MACRO NAME: IX_RM_SET_QARRAY_ENTRY_IN_USE
 *
 * DESCRIPTION: This macro will set the in use flag into a QArray usage field.
 *
 * @Param:  - IN arg_QArrayUsage - this is a ix_uint32 QArray usage field coresponding to the
 *          QArray entry.
 * @Param:  - IN arg_InUseFlag - this ix_uint32 value will specify the new value of the in use 
 *          bit field flag.
 *
 * @Return: Returns the new modified value of the arg_QArrayUsage field.
 */
#define IX_RM_SET_QARRAY_ENTRY_IN_USE( \
                                       arg_QArrayUsage, \
                                       arg_InUseFlag \
                                     ) \
                                     IX_SET_BIT_FIELD32(arg_QArrayUsage, arg_InUseFlag, 0U, 0U)



/**
 * MACRO NAME: IX_RM_GET_QARRAY_ENTRY_TYPE
 *
 * DESCRIPTION: This macro will determine the type of the qarray entry based on the
 *          corresponding QArray usage field.
 *
 * @Param:  - IN arg_QArrayUsage - this is a ix_uint32 QArray usage field coresponding to the
 *          QArray entry.
 *
 * @Return: Returns a ix_sram_qarray_entry_type value that will specify the entry type.
 */
#define IX_RM_GET_QARRAY_ENTRY_TYPE( \
                                     arg_QArrayUsage \
                                   ) \
                                   ((ix_sram_qarray_entry_type) /* typecast */ \
                                   IX_GET_BIT_FIELD32(arg_QArrayUsage, 1U, 1U))


/**
 * MACRO NAME: IX_RM_SET_QARRAY_ENTRY_TYPE
 *
 * DESCRIPTION: This macro will set the entry type field into a QArray usage field.
 *
 * @Param:  - IN arg_QArrayUsage - this is a ix_uint32 QArray usage field coresponding to the
 *          QArray entry.
 * @Param:  - IN arg_EntryType - this ix_sram_qarray_entry_type value will specify the new entry type.
 *
 * @Return: Returns the new modified value of the arg_QArrayUsage field.
 */
#define IX_RM_SET_QARRAY_ENTRY_TYPE( \
                                     arg_QArrayUsage, \
                                     arg_EntryType \
                                   ) \
                                   IX_SET_BIT_FIELD32(arg_QArrayUsage, arg_EntryType, 1U, 1U)



/**
 * TYPENAME: ix_scratch_ring_manager
 * 
 * DESCRIPTION: This is the system wide scratch manager structure
 *
 */
typedef struct ix_s_scratch_ring_manager
{
    ix_uint32 m_aRingUsage[IX_MAX_SCRATCH_RINGS];

    /* This array defines data required for implementation of ring put/get operations */
    ix_ring_op_impl_data m_aRingOpImplData[IX_MAX_SCRATCH_RINGS];

#if defined(_IX_RM_MULTI_THREADED_)

    ix_synchronization_adapter      m_SyncAdapter; /* synchronization adapter */

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

} ix_scratch_ring_manager;



/**
 * NAME: _ix_rm_hardware_init
 *
 * DESCRIPTION: This function should be called at the time the resource manager initializer 
 *          function is called. The function will initialize all the Hardware API.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_hardware_init(void);


/**
 * NAME: _ix_rm_hardware_fini
 *
 * DESCRIPTION: This function should be called at the time the resource manager finalizer function
 *          is called. The function will provide cleanup for Hardware API.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_hardware_fini(void);


/**
 * This is a global QArray manager
 */
extern ix_sram_qarray_manager g_QArrayManager;

/**
 * This is a global SCRATCH ring manager
 */
extern ix_scratch_ring_manager g_SCRATCHRingManager;

#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__INTERNAL_HARDWARE_H__) */
