/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      internal_me_xscale_locks.h
 *
 * = DESCRIPTION
 *      This file defines all data types and function prototypes required internaly by the 
 *      Me/XScale locks API module.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      8/8/2003 1:34:09 PM - creation time 
 * ============================================================================
 * $Id: internal_me_xscale_locks.h,v 1.3 2003/10/10 21:09:07 ageorges Exp $
 */

#if !defined(__INTERNAL_ME_XSCALE_LOCKS_H__)
#define __INTERNAL_ME_XSCALE_LOCKS_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */

/**
 * DESCRIPTION: This value will used to build a Me/XScale lock handle
 *
 */
#define IX_ME_XSCALE_LOCK_HANDLE_SIGNATURE  0xc72ffea3


/**
 * DESCRIPTION: Defines the LSB of the lock index in a Me\XScale lock handle
 *
 */
#define IX_ME_XSCALE_LOCK_HANDLE_INDEX_LSB   0UL 


/**
 * DESCRIPTION: Defines the MSB of the lock index in a Me\XScale lock handle
 *
 */
#define IX_ME_XSCALE_LOCK_HANDLE_INDEX_MSB   5UL 



/**
 * DESCRIPTION: This symbol defines the maximun number of Me\XScale locks that 
 *  can be created in the system
 *
 */
#define IX_ME_XSCALE_LOCK_MAX_COUNT     (1UL << (IX_ME_XSCALE_LOCK_HANDLE_INDEX_MSB - \
    IX_ME_XSCALE_LOCK_HANDLE_INDEX_LSB + 1UL))



/**
 * MACRO NAME: IX_ME_XSCALE_LOCK_HANDLE_GET_LOCK_INDEX
 *
 * DESCRIPTION: Retrieves the Me\XScale lock internal index for the passed handle
 *
 * @Param:  - IN arg_hMeXScaleLock - handle of the Me\XScale lock of interest
 *
 * @Return: an ix_uint32 value representing the internal index value of the lock
 */
#define IX_ME_XSCALE_LOCK_HANDLE_GET_LOCK_INDEX( \
                                                 arg_hMeXScaleLock \
                                               ) \
                                               IX_GET_BIT_FIELD32(arg_hMeXScaleLock, \
                                                  IX_ME_XSCALE_LOCK_HANDLE_INDEX_LSB, \
                                                  IX_ME_XSCALE_LOCK_HANDLE_INDEX_MSB)


/**
 * MACRO NAME: IX_ME_XSCALE_LOCK_MAKE_HANDLE
 *
 * DESCRIPTION: This macro will create a Me\XScale lock handle based on an internal index
 *
 * @Param:  - IN arg_InternalIndex - this is a ix_uint32 value representing the internal index of the lock
 *
 * @Return: Returns a ix_me_xscale_lock_handle coresponding to the passed data.
 */
#define IX_ME_XSCALE_LOCK_MAKE_HANDLE( \
                                       arg_InternalIndex \
                                     ) \
                                   ((ix_me_xscale_lock_handle) /* typecast */ \
                                   (IX_BIT_FIELD_MASK32(IX_ME_XSCALE_LOCK_HANDLE_INDEX_MSB + 1U, 31U) & \
                                   IX_ME_XSCALE_LOCK_HANDLE_SIGNATURE) | \
                                   IX_MAKE_BIT_FIELD32(arg_InternalIndex, IX_ME_XSCALE_LOCK_HANDLE_INDEX_LSB, \
                                   IX_ME_XSCALE_LOCK_HANDLE_INDEX_MSB))                   
                               

/**
 * MACRO NAME: IX_ME_XSCALE_LOCK_HANDLE_IS_VALID
 *
 * DESCRIPTION: This macro will check if the passed Me\XScale lock handle is valid
 *
 * @Param:  - IN arg_hMeXScaleLock - handle of the Me\XScale lock of interest
 *
 * @Return: Returns non zero if the handle is valid or zero otherwise.
 */
#define IX_ME_XSCALE_LOCK_HANDLE_IS_VALID( \
                                           arg_hMeXScaleLock \
                                         ) \
                                         ((((IX_ME_XSCALE_LOCK_HANDLE_SIGNATURE ^ (arg_hMeXScaleLock)) & \
                                            IX_BIT_FIELD_MASK32(IX_ME_XSCALE_LOCK_HANDLE_INDEX_MSB + 1U, 31U)) == 0) ? 1 : 0)



/**
 * TYPENAME: _ix_me_xscale_lock
 * 
 * DESCRIPTION: This stucture describes internally an MR/XScale lock
 *
 */
typedef struct _ix_s_me_xscale_lock _ix_me_xscale_lock;

struct _ix_s_me_xscale_lock
{
    ix_uint32* m_pScratchMemory; /* Memory location where the bit lock is created */
    ix_uint32 m_BitPosition; /* bit position of the lock in the associated memory location */
    ix_me_xscale_lock_owner m_Owner; /* owner of the lock */

    _ix_me_xscale_lock* m_pNext; /* points to the next lock in the free list */
    ix_me_xscale_lock_handle m_hMeXScaleLock; /* handle of the lock */
};



/**
 * DESCRIPTION: This symbol will represent an active lcok (that has been allocated)
 *
 */
#define IX_ME_XSCALE_LOCK_ACTIVE ((_ix_me_xscale_lock*)0xffffffff)
                                 

/**
 * TYPENAME: _ix_me_xscale_lock_manager
 * 
 * DESCRIPTION: This structure describes a ME\XScale locks manager
 *
 */
typedef struct _ix_s_me_xscale_lock_manager
{
    ix_uint32* m_pLocksScratchArea;
    _ix_me_xscale_lock m_MeXScaleLocks[IX_ME_XSCALE_LOCK_MAX_COUNT];
    _ix_me_xscale_lock* m_pLocksFreeList;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter      m_SyncAdapter; /* synchronization adapter */

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

} _ix_me_xscale_lock_manager;


/**
 * NAME: _ix_rm_me_xscale_lock_init
 *
 * DESCRIPTION: This function should be called at the time the resource manager initializer 
 *          function is called. The function will initialize all the ME\XScale lock API.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_me_xscale_lock_init(void);


/**
 * NAME: _ix_rm_me_xscale_lock_fini
 *
 * DESCRIPTION: This function should be called at the time the resource manager finalizer function
 *          is called. The function will provide cleanup for ME\XScale lock API.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_me_xscale_lock_fini(void);


/**
 * Global ME\XScale lock manager
 */
extern _ix_me_xscale_lock_manager g_MeXScaleLockManager;



#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__INTERNAL_ME_XSCALE_LOCKS_H__) */
