/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 4
 *
 * = LIBRARY
 *      Resource Manager Library
 *
 * = MODULE
 *      Memory Manager API
 *
 * = FILENAME
 *      internal_memory.h
 *
 * = DESCRIPTION
 *      This file defines all data types and function prototypes required internaly by the 
 *      Memory Manager library module.
 *
 * = AUTHOR
 *       Adrian Georgescu
 *       adrian.georgescu@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      10/12/2001 11:41:58 AM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 * $Id: internal_memory.h,v 1.6 2003/07/15 20:55:03 ageorges Exp $
 */

#if !defined(__INTERNAL_MEMORY_H__)
#define __INTERNAL_MEMORY_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */

/**
 * Forward declaration
 */
typedef struct _ix_s_mm_memory_manager _ix_mm_memory_manager;



/**
 * TYPENAME: _ix_mm_allocator_ptr
 *
 * DESCRIPTION: This type defines a memory allocator function pointer. It gets as a parameter
 *          an address of the memory manager data should work on. The second parameter specifies
 *          the size of the memory block requested. Also as the third parameter gets
 *          an address where the address of the newly allocated block should be stored. The 
 *          memory alignment will be done according to the default alignment for the memory manager.
 *          The function will return an error code for failure or zero if successful.
 *
 */
typedef 
ix_error (* _ix_mm_allocator_ptr)( 
                                   _ix_mm_memory_manager* arg_pMemoryManager, 
                                   ix_uint32 arg_Size,
                                   void** arg_pMemory 
                                 );


/**
 * TYPENAME: _ix_mm_aligned_allocator_ptr
 * 
 * DESCRIPTION: This type defines an aligned memory allocator function pointer. It gets as a parameter
 *          an address of the memory manager data should work on. The second parameter specifies
 *          the size of the memory block requested. The third parameter specifies the alignment
 *          type as explained at ix_memory_alignment_type type description. The fourth parameter 
 *          specifies the required alignment for the memory block specified as a shift value. The
 *          alignment shift should be at least the default alignment for the manager or the latter
 *          will be used instead.
 *          Also as the fifth parameter gets an address where the address of the newly allocated block should be stored.
 *          The function will return an error code for failure or zero if successful.
 *
 */
typedef 
ix_error (* _ix_mm_aligned_allocator_ptr)( 
                                           _ix_mm_memory_manager* arg_pMemoryManager, 
                                           ix_uint32 arg_Size,
                                           ix_memory_alignment_type arg_AlignmentType,
                                           ix_uint32 arg_AlignmentShift,
                                           void** arg_pMemory 
                                         );



/**
 * TYPENAME: _ix_mm_deallocator_ptr
 * 
 * DESCRIPTION: This type defines a memory deallocator function pointer. The memory deallocator
 *          function will take as a parameter an address of the memory manager data should work on.
 *          As a second paameter takes the memory address of the block that should be freed.
 *
 */
typedef 
ix_error (* _ix_mm_deallocator_ptr)( 
                                     _ix_mm_memory_manager* arg_pMemoryManager, 
                                     void* arg_pMemory 
                                   );


#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * TYPENAME: _ix_mm_reserve_ptr
 * 
 * DESCRIPTION: This type defines a memory reservation function pointer. It gets as a parameter
 *          an address of the memory manager data should work on. The second parameter specifies
 *          the size of the memory block requested. The third parameter specifies the byte
 *          offset in the controlled memory block that we want to reserve. On return this value will
 *          modified to reflect the alignment requirements. The function will return an error 
 *          code for failure or zero if successful.
 *
 */
typedef 
ix_error (* _ix_mm_reserve_ptr)( 
                                 _ix_mm_memory_manager* arg_pMemoryManager, 
                                 ix_uint32 arg_Size,
                                 ix_uint32* arg_pByteOffset 
                               );

#else /* Westport hardware */

/**
 * TYPENAME: _ix_mm_reserve_ptr
 * 
 * DESCRIPTION: This type defines a memory reservation function pointer. It gets as a parameter
 *          an address of the memory manager data should work on. The second parameter specifies
 *          the size of the memory block requested. The third parameter specifies the
 *          ofsset type which applies just to aliased memory.The fourth parameter specifies the byte
 *          offset in the controlled memory block that we want to reserve. On return this value will
 *          modified to reflect the alignment requirements. The function will return an error 
 *          code for failure or zero if successful.
 *
 */
typedef 
ix_error (* _ix_mm_reserve_ptr)( 
                                 _ix_mm_memory_manager* arg_pMemoryManager, 
                                 ix_uint32 arg_Size,
                                 ix_memory_offset_type arg_OffsetType, 
                                 ix_uint32* arg_pByteOffset 
                               );

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */


#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * TYPENAME: _ix_mm_aligned_reserve_ptr
 * 
 * DESCRIPTION: This type defines an aligned memory reservation function pointer. It gets as a parameter
 *          an address of the memory manager data should work on. The second parameter specifies
 *          the size of the memory block requested. The third parameter specifies the alignment
 *          type as explained at ix_memory_alignment_type type description. The fourth parameter 
 *          specifies the required alignment for the memory block specified as a shift value. 
 *          The alignment shift should be at least the default alignment for the manager or 
 *          the latter will be used instead.
 *          The fifth parameter specifies the byte offset in the controlled memory block that we 
 *          want to reserve. On return this value will modified to reflect the alignment requirements. 
 *          The function will return an error code for failure or zero if successful.
 *
 */
typedef 
ix_error (* _ix_mm_aligned_reserve_ptr)( 
                                         _ix_mm_memory_manager* arg_pMemoryManager, 
                                         ix_uint32 arg_Size,
                                         ix_memory_alignment_type arg_AlignmentType,
                                         ix_uint32 arg_AlignmentShift,
                                         ix_uint32* arg_pByteOffset 
                                       );

#else /* Westport hardware */

/**
 * TYPENAME: _ix_mm_aligned_reserve_ptr
 * 
 * DESCRIPTION: This type defines an aligned memory reservation function pointer. It gets as a parameter
 *          an address of the memory manager data should work on. The second parameter specifies
 *          the size of the memory block requested. The third parameter specifies the alignment
 *          type as explained at ix_memory_alignment_type type description. The fourth parameter 
 *          specifies the required alignment for the memory block specified as a shift value. 
 *          The alignment shift should be at least the default alignment for the manager or 
 *          the latter will be used instead. The fifth parameter specifies the
 *          ofsset type which applies just to aliased memory.
 *          The sixth parameter specifies the byte offset in the controlled memory block that we 
 *          want to reserve. On return this value will modified to reflect the alignment requirements. 
 *          The function will return an error code for failure or zero if successful.
 *
 */
typedef 
ix_error (* _ix_mm_aligned_reserve_ptr)( 
                                         _ix_mm_memory_manager* arg_pMemoryManager, 
                                         ix_uint32 arg_Size,
                                         ix_memory_alignment_type arg_AlignmentType,
                                         ix_uint32 arg_AlignmentShift,
                                         ix_memory_offset_type arg_OffsetType, 
                                         ix_uint32* arg_pByteOffset 
                                       );

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */


/**
 * TYPENAME: _ix_mm_init_ptr
 * 
 * DESCRIPTION: This type defines a memory manager initializer function pointer. The 
 *          memory manager initializer gets as a parameter an address of the memory 
 *          manager data should work on. The initializer will perform all the necessary
 *          initialization of the memory manager structure. Initialization should be done
 *          before the memory manager becomes valid for use.
 *
 */
typedef 
ix_error (* _ix_mm_init_ptr)( 
                              _ix_mm_memory_manager* arg_pMemoryManager 
                            );


/**
 * TYPENAME: _ix_mm_fini_ptr
 * 
 * DESCRIPTION: This type defines a memory manager finalizer function pointer. The 
 *          memory manager finalizer gets as a parameter an address of the memory 
 *          manager data should work on. The finalizer will perform all necessary
 *          cleanup for the specified memory manager.
 *
 */
typedef 
ix_error (* _ix_mm_fini_ptr)( 
                              _ix_mm_memory_manager* arg_pMemoryManager 
                            );


/**
 * TYPENAME: _ix_mm_read_ptr
 * 
 * DESCRIPTION: This type defines a memory manager read function for 32 bit values.
 *          The memory manager read function takes as a parameter an address of the memory 
 *          manager data should work on. The other two parameters represent the address
 *          where the data should be read from and the address where the read data should
 *          be stored.
 */
typedef 
ix_error (* _ix_mm_read_ptr)( 
                              const _ix_mm_memory_manager* arg_pMemoryManager, 
                              const ix_uint32* arg_pFromAddress, 
                              ix_uint32* arg_pToAddress 
                            );


/**
 * TYPENAME: _ix_mm_write_ptr
 * 
 * DESCRIPTION: This type defines a memory manager write function for 32 bit values.
 *          The memory manager write function takes as a parameter an address of the memory 
 *          manager data should work on. The other two parameters represent the value to
 *          be written and the address where the read data should be written.
 *
 */
typedef 
ix_error (* _ix_mm_write_ptr)( 
                               const _ix_mm_memory_manager* arg_pMemoryManager, 
                               ix_uint32 arg_Value, 
                               ix_uint32* arg_pToAddress 
                             );


/**
 * TYPENAME: _ix_mm_vmt
 * 
 * DESCRIPTION: This structure is the equivalent of a virtual method table in C++.
 *          Each memory manager will have an associated VMT that will control behaviour
 *          of that memory manager. This will allow for a great flexibility and the
 *          memory managers for different types of memory can be implemented in the
 *          same way. The init, fini, alloc and free operations are rare, so the 2 extra 
 *          indirections are not of any importance. The read/write operations are here
 *          just for the simulation case that is anyway a slow operation. For the hardware
 *          case the programmer will use the fast macros.
 *
 */
typedef struct _ix_s_mm_vmt
{
    _ix_mm_init_ptr                 m_InitFuncPtr; /* initializer */
    _ix_mm_fini_ptr                 m_FiniFuncPtr; /* finalizer */
    _ix_mm_allocator_ptr            m_AllocFuncPtr; /* allocator */
    _ix_mm_aligned_allocator_ptr    m_AlignedAllocFuncPtr; /* aligned allocator */
    _ix_mm_deallocator_ptr          m_FreeFuncPtr; /* deallocator */
    _ix_mm_reserve_ptr              m_ReserveFuncPtr; /* reservation function */
    _ix_mm_aligned_reserve_ptr      m_AlignedReserveFuncPtr; /* aligned reservation function */
    _ix_mm_read_ptr                 m_ReadFuncPtr; /* read */
    _ix_mm_write_ptr                m_WriteFuncPtr; /* write */
} _ix_mm_vmt;


/**
 * TYPENAME: ix_mm_memory_block
 * 
 * DESCRIPTION: This structure defines a memory block cell. This structure is used
 *          to keep the free memory blocks in a linked list. These structure will be
 *          allocated on the OS heap and the unused ones will be kept in a recycled list
 *          in order to avoid frequent malloc/free operations that are expensive.
 *
 */
typedef struct _ix_s_mm_memory_block_cell
{
    ix_uint32                   		m_Size;
    ix_uint32*                  		m_pData;
    struct _ix_s_mm_memory_block_cell*  m_pNextCell;
    struct _ix_s_mm_memory_block_cell*  m_pPrevCell;
} _ix_mm_memory_block_cell;


#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * TYPENAME: _ix_s_mm_memory_manager
 * 
 * DESCRIPTION: This structure defines a memory manager object. This object has all the required
 *          information to be able to perform all memory required operations independently.
 *
 */
struct _ix_s_mm_memory_manager
{
    ix_memory_type              m_MemoryType; /* specifies the memory type */
    ix_uint32                   m_MemoryChannel; /* specifies the memory channel */
    ix_uint32                   m_MemoryBlockSize; /* specifies the controlled memory block size in bytes */
    ix_uint32                   m_DefaultAlignmentShift; /* shift value to obtain the required alignment */
    ix_uint32*                  m_pStartAddress; /* specifies the controlled memory block start address */
    ix_uint32*                  m_pEndAddress; /* specifies the controlled memory block's end address */
    ix_uint32                   m_ChannelPhysicalOffset; 
        /* specifies the physical offset of the controlled memory from the channel's start address */
    void*                       m_pChannelPhysicalAddress;
        /* this is the start physical address of the current channel */
    _ix_mm_memory_block_cell*   m_pRover; /* this is a starting pointer in the search of a free memory block */
    _ix_mm_memory_block_cell*   m_pBlockList; /* free memory block list */
    _ix_mm_memory_block_cell*   m_pRecycledBlockCellList; /* memory block cell recycled list */
    ix_uint32                   m_RecycledBlockCellCount; /* memory block cell recycled count */
    const _ix_mm_vmt*           m_pVMT; /* VMT. Pointer to a structure of function pointers which define memory manager's operations */

#if defined(_IX_RM_MULTI_THREADED_)

    ix_synchronization_adapter      m_SyncAdapter; /* synchronization adapter */

#endif /* end defined(_IX_RM_MULTI_THREADED_) */
    
};

#else /* Westport hardware */

/**
 * TYPENAME: _ix_s_mm_memory_manager
 * 
 * DESCRIPTION: This structure defines a memory manager object. This object has all the required
 *          information to be able to perform all memory required operations independently.
 *
 */
struct _ix_s_mm_memory_manager
{
    ix_memory_type              m_MemoryType; /* specifies the memory type */
    ix_memory_bank_type         m_MemoryBankType;   /* memory bank this area belongs to */     

    ix_uint32                   m_MemoryChannel; /* specifies the memory channel */
    ix_uint32                   m_MemoryBlockSize; /* specifies the controlled memory block size in bytes */
    ix_uint32                   m_DefaultAlignmentShift; /* shift value to obtain the required alignment */
    ix_uint32*                  m_pStartAddress; /* specifies the controlled memory block start address */
    ix_uint32*                  m_pEndAddress; /* specifies the controlled memory block's end address */
    ix_uint32                   m_ChannelPhysicalOffset; 
        /* specifies the physical offset of the controlled memory from the channel's start address */
    void*                       m_pChannelPhysicalAddress;
        /* this is the start physical address of the current channel */
    ix_uint32                   m_NonAliasChannelPhysicalOffset; 
        /* physical offset of the start of the non aliased memory block inside the memory channel */
    void*                       m_pNonAliasChannelPhysicalAddress;
        /* this is the start non alias physical address of the current channel in its memory bank */

    _ix_mm_memory_block_cell*   m_pRover; /* this is a starting pointer in the search of a free memory block */
    _ix_mm_memory_block_cell*   m_pBlockList; /* free memory block list */
    _ix_mm_memory_block_cell*   m_pRecycledBlockCellList; /* memory block cell recycled list */
    ix_uint32                   m_RecycledBlockCellCount; /* memory block cell recycled count */
    const _ix_mm_vmt*           m_pVMT; /* VMT. Pointer to a structure of function pointers which define memory manager's operations */

#if defined(_IX_RM_MULTI_THREADED_)

    ix_synchronization_adapter      m_SyncAdapter; /* synchronization adapter */

#endif /* end defined(_IX_RM_MULTI_THREADED_) */
    
};

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */


/**
 * TYPENAME: _ix_mm_memory_manager_group
 * 
 * DESCRIPTION: This structure defines a logical group of memory managers. All the memory
 *          managers for a particular type of memory will reside in a memory manager group.
 *
 */
typedef struct _ix_s_mm_memory_manager_group
{
    ix_uint32                   m_MemberCount; /* the numbers of members in the group */
    _ix_mm_memory_manager*      m_pMembers; /* array of members */
} _ix_mm_memory_manager_group;


/**
 * NAME: _ix_rm_memory_init
 *
 * DESCRIPTION: This function should be called at the time the resource manager initializer 
 *          function is called. The function will initialize all memory managers based on 
 *          the hardware configuration.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_memory_init(void);


/**
 * NAME: _ix_rm_memory_fini
 *
 * DESCRIPTION: This function should be called at the time the resource manager finalizer function
 *          is called. The function will finalize all memory managers.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_memory_fini(void);

#if 0

/**
 * NAME: _ix_rm_mm_print_memory_manager
 *
 * DESCRIPTION: This is a support function that will print on standard output the memory manager
 *          information passed as a parameter.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager structure that will be
 *          printed.
 */
void _ix_rm_mm_print_memory_manager( 
                                    _ix_mm_memory_manager* arg_pMemoryManager 
                                   );

#endif /* 0 */


/**
 * NAME: _ix_rm_mm_allocator
 *
 * DESCRIPTION: This is an allocator function for all types of memory.
 *          For special needs different allocators will be defined, but this is the generic one.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager data this allocator works on.
 * @Param:  - IN arg_Size - the size of the memory block requested.
 * @Param:  - OUT arg_pMemory - a pointer to where to store the new allocated block address .
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure. In the case the call fails the *arg_pMemory
 *          is set to NULL. Caution about the fact that memory allocation and reservation will 
 *          consider the alignment and will adjust the size and start address in accordance with 
 *          the default alignment.
 */
ix_error _ix_rm_mm_allocator( 
                              _ix_mm_memory_manager* arg_pMemoryManager, 
                              ix_uint32 arg_Size, 
                              void** arg_pMemory 
                            );


/**
 * NAME: _ix_rm_mm_aligned_allocator
 *
 * DESCRIPTION: This is an aligned allocator function for all types of memory.
 *          For special needs different aligned allocators will be defined, but this 
 *          is the generic one.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager data this allocator works on.
 * @Param:  - IN arg_Size - the size of the memory block requested.
 * @Param:  - IN arg_AlignmentType - specifies the type of alignment required as explained at 
 *          ix_memory_alignment_type type description
 * @Param:  - IN arg_AlignmentShift - this value specifies the required memory block alignment as the
 *          power of 2 of the alignment. If the specified alignment is less than the
 *          default alignment for the memory pool manager, then the default alignment will be used.
 * @Param:  - OUT arg_pMemory - a pointer to where to store the new allocated block address .
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure. In the case the call fails the *arg_pMemory
 *          is set to NULL. The returned memory will insure the right alignment for the given memory type
 *          and the requested alignment type.
 */
ix_error _ix_rm_mm_aligned_allocator( 
                                      _ix_mm_memory_manager* arg_pMemoryManager, 
                                      ix_uint32 arg_Size,
                                      ix_memory_alignment_type arg_AlignmentType,
                                      ix_uint32 arg_AlignmentShift,
                                      void** arg_pMemory 
                                    );



/**
 * NAME: _ix_rm_mm_deallocator
 *
 * DESCRIPTION: This function will deallocate previously allocated memory. It is generic deallocator
 *          that can be used for all memory types regardless alignment. If the that address is
 *          passed is not found in the used list then nothing will happen and an error will be
 *          returned.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager data this allocator works on.
 * @Param:  - IN arg_pMemory - the memory block that will be released.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_deallocator( 
                                _ix_mm_memory_manager* arg_pMemoryManager, 
                                void* arg_pMemory 
                              );


#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * NAME: _ix_rm_mm_reserve
 *
 * DESCRIPTION: This is a memory reservation function for all types of memory.
 *          For special needs different reservation functions will be defined, but this is the generic one.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager data this allocator works on.
 * @Param:  - IN arg_Size - the size of the memory block requested.
 * @Param:  - INOUT arg_pByteOffset - an address to byte offset in the memory block where we want to 
 *          reserve memory. On output this offset will be modified to accomodate the required alignment.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure. If any of the requested area is 
 *          in use then an error will be returned. Caution about the fact that memory 
 *          allocation and reservation will consider the alignment and will adjust the
 *          size and start address in accordance with the default alignment.
 *          
 */
ix_error _ix_rm_mm_reserve( 
                            _ix_mm_memory_manager* arg_pMemoryManager, 
                            ix_uint32 arg_Size,
                            ix_uint32* arg_pByteOffset 
                          );

#else /* Westport hardware */

/**
 * NAME: _ix_rm_mm_reserve
 *
 * DESCRIPTION: This is a memory reservation function for all types of memory.
 *          For special needs different reservation functions will be defined, but this is the generic one.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager data this allocator works on.
 * @Param:  - IN arg_Size - the size of the memory block requested.
 * @Param:  - IN arg_OffsetType - represents the type of the specified offset 
 * @Param:  - INOUT arg_pByteOffset - an address to byte offset in the memory block where we want to 
 *          reserve memory. On output this offset will be modified to accomodate the required alignment.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure. If any of the requested area is 
 *          in use then an error will be returned. Caution about the fact that memory 
 *          allocation and reservation will consider the alignment and will adjust the
 *          size and start address in accordance with the default alignment.
 *          
 */
ix_error _ix_rm_mm_reserve( 
                            _ix_mm_memory_manager* arg_pMemoryManager, 
                            ix_uint32 arg_Size,
                            ix_memory_offset_type arg_OffsetType, 
                            ix_uint32* arg_pByteOffset 
                          );

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */


#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * NAME: _ix_rm_mm_aligned_reserve
 *
 * DESCRIPTION: This is an aligned memory reservation function for all types of memory.
 *          For special needs different aligned reservation functions will be defined, but 
 *          this is the generic one.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager data this allocator works on.
 * @Param:  - IN arg_Size - the size of the memory block requested.
 * @Param:  - IN arg_AlignmentType - specifies the type of alignment required as explained at 
 *          ix_memory_alignment_type type description
 * @Param:  - IN arg_AlignmentShift - this value specifies the required memory block alignment as the
 *          power of 2 of the alignment. If the specified alignment is less than the
 *          default alignment for the memory pool manager then the default alignment will be used.
 * @Param:  - INOUT arg_pByteOffset - an address to byte offset in the memory block where we want to 
 *          reserve memory. On output this offset will be modified to accomodate the required alignment.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure. If any of the requested area is 
 *          in use then an error will be returned. Caution about the fact that memory 
 *          allocation and reservation will consider the allignment and will adjust the
 *          size and start address in accordance with the alignment.
 *          
 */
ix_error _ix_rm_mm_aligned_reserve( 
                                    _ix_mm_memory_manager* arg_pMemoryManager, 
                                    ix_uint32 arg_Size,
                                    ix_memory_alignment_type arg_AlignmentType,
                                    ix_uint32 arg_AlignmentShift,
                                    ix_uint32* arg_pByteOffset 
                                  );

#else /* Westport hardware */

/**
 * NAME: _ix_rm_mm_aligned_reserve
 *
 * DESCRIPTION: This is an aligned memory reservation function for all types of memory.
 *          For special needs different aligned reservation functions will be defined, but 
 *          this is the generic one.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager data this allocator works on.
 * @Param:  - IN arg_Size - the size of the memory block requested.
 * @Param:  - IN arg_AlignmentType - specifies the type of alignment required as explained at 
 *          ix_memory_alignment_type type description
 * @Param:  - IN arg_AlignmentShift - this value specifies the required memory block alignment as the
 *          power of 2 of the alignment. If the specified alignment is less than the
 *          default alignment for the memory pool manager then the default alignment will be used.
 * @Param:  - IN arg_OffsetType - represents the type of the specified offset 
 * @Param:  - INOUT arg_pByteOffset - an address to byte offset in the memory block where we want to 
 *          reserve memory. On output this offset will be modified to accomodate the required alignment.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure. If any of the requested area is 
 *          in use then an error will be returned. Caution about the fact that memory 
 *          allocation and reservation will consider the allignment and will adjust the
 *          size and start address in accordance with the alignment.
 *          
 */
ix_error _ix_rm_mm_aligned_reserve( 
                                    _ix_mm_memory_manager* arg_pMemoryManager, 
                                    ix_uint32 arg_Size,
                                    ix_memory_alignment_type arg_AlignmentType,
                                    ix_uint32 arg_AlignmentShift,
                                    ix_memory_offset_type arg_OffsetType, 
                                    ix_uint32* arg_pByteOffset 
                                  );

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */


/**
 * NAME: _ix_rm_mm_initializer
 *
 * DESCRIPTION: This function provides all the required initialization for the memory manager
 *          that is passed as an argument.There are some assumptions that this function 
 *          makes:  the type of memory should be set before calling the function;
 *                  the channel, or microengine number should be set before calling the function;
 *                  the start address should be set prior calling the function;
 *                  the size of controlled memory should be set prior calling the function;
 *                  the default alignment should be set prior calling the function;
 *                  the memory channel physical start address should be set prior calling the function;
 *                  the physiscal offset into the channel coresponding to the virtual start address
 *                  should be set prior calling the function;
 * 
 * @Param:  - IN arg_pMemoryManager - the address of the memory manager that will be initialized.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure. Upon return the referred memory manager
 *          is fully initialized in the case the call is successful.
 */
ix_error _ix_rm_mm_initializer( 
                                _ix_mm_memory_manager* arg_pMemoryManager 
                              );



/**
 * NAME: _ix_rm_mm_finalizer
 *
 * DESCRIPTION: This function is the finalizer for the memory manager passed as an argument.
 *          After the finalizer returns the memory manager will not have any memory allocated
 *          and should be ready for a new initialization for reuse. After the finalization the
 *          memory manager can not handle requests.
 * 
 * @Param:  - IN arg_pMemoryManager - the address of the memory manager that will be finalized.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_finalizer( 
                              _ix_mm_memory_manager* arg_pMemoryManager 
                            );


/**
 * NAME: _ix_rm_mm_split_cell
 *
 * DESCRIPTION: This function will split the passed cell into two cells and return the newly created
 *          cell. The passed cell will be adjusted to firstHalfSize size and the new cell will get the
 *          remaining size. The newly created cell will be inserted in the list after the current cell.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager data this function works on.
 * @Param:  - IN arg_pCell - represents a pointer to the cell to be split
 * @Param:  - IN arg_Size - represents the new size of the passed cell. The newly created cell will get
 *          the remaining memory.
 *
 * @Return: returns a new _ix_mm_memory_block_cell to be used by the memory manager or NULL for failure. 
 */
_ix_mm_memory_block_cell* _ix_rm_mm_split_cell(
                                                _ix_mm_memory_manager* arg_pMemoryManager, 
                                                _ix_mm_memory_block_cell* arg_pCell, 
                                                ix_uint32 arg_Size
                                              );

/**
 * NAME: _ix_rm_mm_new_cell
 *
 * DESCRIPTION: This function will return a new _ix_mm_memory_block_cell to be used by the memory
 *          manager.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager data this function works on.
 *
 * @Return: returns a new _ix_mm_memory_block_cell to be used by the memory manager or NULL for failure. 
 */
_ix_mm_memory_block_cell* _ix_rm_mm_new_cell( 
                                              _ix_mm_memory_manager* arg_pMemoryManager 
                                            );



/**
 * NAME: _ix_rm_mm_release_cell
 *
 * DESCRIPTION: This function will release the specified block memory cell. This function and the above
 *          allows to recycle the cells in order to avoid recurring heap allocations and deallocations.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager data this function works on.
 * @Param:  - IN arg_pCell - block cell that should be released;
 *
 */
void _ix_rm_mm_release_cell( 
                             _ix_mm_memory_manager* arg_pMemoryManager, 
                             _ix_mm_memory_block_cell* arg_pCell 
                           );


#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * NAME: _ix_rm_mm_read_dram
 *
 * DESCRIPTION: This function reads a DRAM 32 bit memory location from the arg_pFromAddress and stores
 *          the value to the arg_pToAddress. The memory pointers should guarantee 4 byte alignment. The
 *          alignment will be checked just in DEBUG mode.
 * 
 * @Param:  - IN arg_pMemoryManager - represents the address of the memory manager data should work on
 * @Param:  - IN arg_pFromAddress - represents the address we read from.
 * @Param:  - OUT arg_pToAddress - represents the address where to store the read value.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_read_dram( 
                              const _ix_mm_memory_manager* arg_pMemoryManager, 
                              const ix_uint32* arg_pFromAddress, 
                              ix_uint32* arg_pToAddress 
                            );



/**
 * NAME: _ix_rm_mm_read_sram
 *
 * DESCRIPTION: This function reads a SRAM 32 bit memory location from the arg_pFromAddress and stores
 *          the value to the arg_pToAddress. The memory pointers should guarantee 4 byte alignment. The
 *          alignment will be checked just in DEBUG mode.
 * 
 * @Param:  - IN arg_pMemoryManager - represents the address of the memory manager data should work on
 * @Param:  - IN arg_pFromAddress - represents the address we read from.
 * @Param:  - OUT arg_pToAddress - represents the address where to store the read value.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_read_sram( 
                              const _ix_mm_memory_manager* arg_pMemoryManager, 
                              const ix_uint32* arg_pFromAddress, 
                              ix_uint32* arg_pToAddress 
                            );


/**
 * NAME: _ix_rm_mm_read_scratch
 *
 * DESCRIPTION: This function reads a SCRATCH 32 bit memory location from the arg_pFromAddress and stores
 *          the value to the arg_pToAddress. The memory pointers should guarantee 4 byte alignment. The
 *          alignment will be checked just in DEBUG mode.
 * 
 * @Param:  - IN arg_pMemoryManager - represents the address of the memory manager data should work on.
 * @Param:  - IN arg_pFromAddress - represents the address we read from.
 * @Param:  - OUT arg_pToAddress - represents the address where to store the read value.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_read_scratch( 
                                 const _ix_mm_memory_manager* arg_pMemoryManager, 
                                 const ix_uint32* arg_pFromAddress, 
                                 ix_uint32* arg_pToAddress 
                               );



/**
 * NAME: _ix_rm_mm_write_dram
 *
 * DESCRIPTION: This function writes a 32 bit value to the specified DRAM address location. The memory pointer
 *          should guarantee 4 byte alignment. The alignment will be checked just in DEBUG mode.
 * 
 * @Param:  - IN arg_pMemoryManager - represents the address of the memory manager data should work on
 * @Param:  - IN arg_Value - a 32 bit value that will be written at the specified location.
 * @Param:  - IN arg_pToAddress - address where the data will be written.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_write_dram( 
                               const _ix_mm_memory_manager* arg_pMemoryManager, 
                               ix_uint32 arg_Value, 
                               ix_uint32* arg_pToAddress 
                             );


/**
 * NAME: _ix_rm_mm_write_sram
 *
 * DESCRIPTION: This function writes a 32 bit value to the specified SRAM address location. The memory pointer
 *          should guarantee 4 byte alignment. The alignment will be checked just in DEBUG mode.
 * 
 * @Param:  - IN arg_pMemoryManager - represents the address of the memory manager data should work on
 * @Param:  - IN arg_Value - a 32 bit value that will be written at the specified location.
 * @Param:  - IN arg_pToAddress - address where the data will be written.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_write_sram( 
                               const _ix_mm_memory_manager* arg_pMemoryManager, 
                               ix_uint32 arg_Value, 
                               ix_uint32* arg_pToAddress 
                             );


/**
 * NAME: _ix_rm_mm_write_scratch
 *
 * DESCRIPTION: This function writes a 32 bit value to the specified SCRACH address location. The memory pointer
 *          should guarantee 4 byte alignment. The alignment will be checked just in DEBUG mode.
 * 
 * @Param:  - IN arg_pMemoryManager - represents the address of the memory manager data should work on
 * @Param:  - IN arg_Value - a 32 bit value that will be written at the specified location.
 * @Param:  - IN arg_pToAddress - address where the data will be written.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_write_scratch( 
                                  const _ix_mm_memory_manager* arg_pMemoryManager, 
                                  ix_uint32 arg_Value, 
                                  ix_uint32* arg_pToAddress 
                                );


#else /* Westport hardware */

/**
 * NAME: _ix_rm_mm_read_south
 *
 * DESCRIPTION: This function reads a South bank 32 bit memory location from the arg_pFromAddress and stores
 *          the value to the arg_pToAddress. The memory pointers should guarantee 4 byte alignment. The
 *          alignment will be checked just in DEBUG mode.
 * 
 * @Param:  - IN arg_pMemoryManager - represents the address of the memory manager data should work on
 * @Param:  - IN arg_pFromAddress - represents the address we read from.
 * @Param:  - OUT arg_pToAddress - represents the address where to store the read value.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_read_south( 
                              const _ix_mm_memory_manager* arg_pMemoryManager, 
                              const ix_uint32* arg_pFromAddress, 
                              ix_uint32* arg_pToAddress 
                            );



/**
 * NAME: _ix_rm_mm_read_north
 *
 * DESCRIPTION: This function reads a North bank 32 bit memory location from the arg_pFromAddress and stores
 *          the value to the arg_pToAddress. The memory pointers should guarantee 4 byte alignment. The
 *          alignment will be checked just in DEBUG mode.
 * 
 * @Param:  - IN arg_pMemoryManager - represents the address of the memory manager data should work on
 * @Param:  - IN arg_pFromAddress - represents the address we read from.
 * @Param:  - OUT arg_pToAddress - represents the address where to store the read value.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_read_north( 
                              const _ix_mm_memory_manager* arg_pMemoryManager, 
                              const ix_uint32* arg_pFromAddress, 
                              ix_uint32* arg_pToAddress 
                            );


/**
 * NAME: _ix_rm_mm_write_south
 *
 * DESCRIPTION: This function writes a 32 bit value to the specified South bank address location. The memory pointer
 *          should guarantee 4 byte alignment. The alignment will be checked just in DEBUG mode.
 * 
 * @Param:  - IN arg_pMemoryManager - represents the address of the memory manager data should work on
 * @Param:  - IN arg_Value - a 32 bit value that will be written at the specified location.
 * @Param:  - IN arg_pToAddress - address where the data will be written.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_write_south( 
                               const _ix_mm_memory_manager* arg_pMemoryManager, 
                               ix_uint32 arg_Value, 
                               ix_uint32* arg_pToAddress 
                             );


/**
 * NAME: _ix_rm_mm_write_north
 *
 * DESCRIPTION: This function writes a 32 bit value to the specified North bank address location. The memory pointer
 *          should guarantee 4 byte alignment. The alignment will be checked just in DEBUG mode.
 * 
 * @Param:  - IN arg_pMemoryManager - represents the address of the memory manager data should work on
 * @Param:  - IN arg_Value - a 32 bit value that will be written at the specified location.
 * @Param:  - IN arg_pToAddress - address where the data will be written.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_write_north( 
                               const _ix_mm_memory_manager* arg_pMemoryManager, 
                               ix_uint32 arg_Value, 
                               ix_uint32* arg_pToAddress 
                             );


#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */


/**
 * The following section defines support for implementing algorithm functions that will
 * work on _ix_mm_memory_block_cell linked list. The idea is similar to that found in STL
 * with the exception that we do not define generic iterators.
 *
 * The _ix_rm_memory_block_cell_action function will get a start _ix_mm_memory_block_cell pointer
 * and an end _ix_mm_memory_block_cell plus a function pointer of the type _ix_mm_memory_block_cell_predicate
 * and will invoke the predicate on all the cells in the range [start, end). The caller must
 * be sure that "start" and "end" values are valid and starting from "start" you get to the "end".
 * If the predicate returns 0 then the processing stops and the _ix_mm_memory_block_cell pointer
 * for which the predicate fails will be returned.
 */


/**
 * This type defines a predicate function for a _ix_mm_memory_block_cell
 */
typedef 
ix_uint32 (* _ix_mm_memory_block_cell_predicate)( 
                                                  _ix_mm_memory_block_cell* arg_pCell, 
                                                  void* arg_pContext 
                                                );


/**
 * NAME: _ix_rm_memory_block_cell_action
 *
 * DESCRIPTION: This function iterates through alist of memory block cells and will apply
 *          the arg_Predicate to each entry in the logcal interval [arg_pStartCell, arg_pEndCell).
 *          The user have to make sure there is a finite path from arg_pStartCell to arg_pEndCell.
 *          The function will return 0 if all the invokations of the arg_Predicate predicate
 *          return non 0. If one invocation returns 0 then the function will return the address
 *          of the cell for which the predicate failed.
 * 
 * @Param:  - IN arg_pStartCell - the start cell address in the linked list;
 * @Param:  - IN arg_pEndCell - the end cell address in the linked list. The predicate will
 *          not be applied for this cell;
 * @Param:  - IN arg_Predicate - the active predicate that will be applied to each cell;
 * @Param:  - IN arg_pContext - a pointer to some context data that will be passed to arg_Predicate at
 *          each invocation;
 *
 * @Return: The function will return 0 if all the invocations of the arg_Predicate predicate
 *          return non 0. If one invokation returns 0 then the function will return the address
 *          of the cell for which the predicate failed.
 */
_ix_mm_memory_block_cell* _ix_rm_memory_block_cell_action( 
                                                           _ix_mm_memory_block_cell* arg_pStartCell, 
                                                           _ix_mm_memory_block_cell* arg_pEndCell, 
                                                           _ix_mm_memory_block_cell_predicate arg_Predicate, 
                                                           void* arg_pContext 
                                                         );


/**
 * NAME: _ix_rm_memory_block_cell_print
 *
 * DESCRIPTION: This function will print a memory block cell from a memory manager block list.  
 * 
 * @Param:  - IN arg_pCell - memory cell to print
 * @Param:  - IN arg_pContext - context argument that in this case should be a pointer to an
 *          ix_uint32 counter that will be incremented at each call;
 *
 * @Return: Returns 1 all the time.
 */
ix_uint32 _ix_rm_memory_block_cell_print( 
                                          _ix_mm_memory_block_cell* arg_pCell, 
                                          void* arg_pContext 
                                        );

#if 0

/**
 * NAME: _ix_rm_memory_recycled_cell_print
 *
 * DESCRIPTION: This function will print a memory block cell from a memory manager recycled
 *          cell list.  
 * 
 * @Param:  - IN arg_pCell - memory cell to print
 * @Param:  - IN arg_pContext - context argument that in this case should be a pointer to an
 *          ix_uint32 counter that will be incremented at each call;
 *
 * @Return: Returns 1 all the time.
 */
ix_uint32 _ix_rm_memory_recycled_cell_print( 
                                             _ix_mm_memory_block_cell* arg_pCell, 
                                             void* arg_pContext 
                                           );

#endif /* 0 */


/**
 * NAME: _ix_rm_memory_get_free_memory
 *
 * DESCRIPTION: This function is used ti retrieve the free memory for a memory manager.
 * 
 * @Param:  - IN arg_pCell - current memory cell;
 * @Param:  - IN arg_pContext - context argument that in this case should be a pointer to an
 *          ix_uint32 value that will be increased by he size of current cell;
 *
 * @Return: Returns 1 all the time.
 */
ix_uint32 _ix_rm_memory_get_free_memory( 
                                         _ix_mm_memory_block_cell* arg_pCell, 
                                         void* arg_pContext 
                                       );


/**
 * This global variable represents the memory manger groups active in this hardware configuration
 */
extern const _ix_mm_memory_manager_group g_aMemoryManagersGroups[IX_MEMORY_TYPE_LAST];

#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__INTERNAL_MEMORY_H__) */
