/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      internal_microengine.h
 *
 * = DESCRIPTION
 *      This file defines all data types and function prototypes required internaly by the 
 *      Microengine API module.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      6/4/2002 12:10:02 PM - creation time 
 * ============================================================================
 * $Id: internal_microengine.h,v 1.11 2003/09/29 20:18:09 ageorges Exp $
 */

#if !defined(__INTERNAL_MICROENGINE_H__)
#define __INTERNAL_MICROENGINE_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */


/**
 * TYPENAME: _ix_imported_symbol_node
 * 
 * DESCRIPTION: This structure represents a name/value pair of a symbol to be patched
 *  into the uCode that will be kept into an internal linked list.
 */
typedef struct _ix_s_imported_symbol_node _ix_imported_symbol_node;

struct _ix_s_imported_symbol_node
{
    ix_uint32                   m_Value;
    char*                       m_Name;
    _ix_imported_symbol_node*   m_pNext;
};




/**
 * TYPENAME: _ix_microengine_data_manager
 * 
 * DESCRIPTION: This structure manages the entire Microengine API
 *
 */
typedef struct _ix_s_microengine_data_manager
{
    /**
     * The handle to the active UOF image
     */
    void* m_UOFActiveHandle;

    /**
     * Specifies if the UOF has been loaded from a file or if it was hardcoded
     */
    ix_uint32 m_IsUOFLoadedFromFile;

    /**
     * Array of imported symbols lists
     */
    _ix_imported_symbol_node* m_aImportedSymbolsList[IX_MAX_MICROENGINE_NUMBER];


#if defined(_IX_RM_MULTI_THREADED_)

    ix_synchronization_adapter      m_SyncAdapter; /* synchronization adapter */

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

#if defined(_IX_RM_WORKBENCH_SUPPORT_)

    
    /**
     * Specifies if the UCode has been downloaded from Workbench
     */
    ix_uint32 m_IsUOFLoadedFromWorkbench;

    
    /**
     * Address of the buffer where the Workbench image have been copied
     */
    unsigned char* m_pUCodeBuffer;
    
    /**
     * Size of the Workbench image 
     */
    unsigned long m_UCodeSize;

#endif /* defined(_IX_RM_WORKBENCH_SUPPORT_) */


} _ix_microengine_data_manager;



/**
 * NAME: _ix_rm_microengine_init
 *
 * DESCRIPTION: This function should be called at the time the resource manager initializer 
 *  function is called. The function will initialize all the Microengine API.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_microengine_init(void);


/**
 * NAME: _ix_rm_microengine_fini
 *
 * DESCRIPTION: This function should be called at the time the resource manager finalizer function
 *  is called. The function will provide cleanup for Microengine API.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_microengine_fini(void);



/**
 * NAME: _ix_rm_microengine_patch_symbols
 *
 * DESCRIPTION: This function will patch imported symbols into the microcode without any data 
 *  locking. It just adds the symbols to the internal symbol lists.
 * 
 * @Param:  - IN arg_UengNumber - represents the microengine number for which this call is 
 *          intended.
 * @Param:  - IN arg_SymbolsNumber - represents the number of symbols in the arg_Symbols 
 *          array that will be patched.
 * @Param:  - IN arg_aSymbols - represents an array of symbols that will be patched into 
 *          the microcode.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error _ix_rm_microengine_patch_symbols( 
                                           ix_uint32 arg_MENumber, 
                                           ix_uint32 arg_SymbolsNumber, 
                                           const ix_imported_symbol arg_aSymbols[] 
                                         );


/**
 * NAME: _ix_rm_patch_qarray_symbols
 *
 * DESCRIPTION: This function will the required QArray symbols for all SRAM channels
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_patch_qarray_symbols(void);


/**
 * NAME: _ix_rm_microengine_register_internal_symbols
 *
 * DESCRIPTION: This function will register some internal symbols with the specified microengine.
 *  Many required symbols do not require user's explicit patching. They will be set and registered
 *  for patching by RM library. These are internal values of RM that are required for uCode
 *  proper setup.
 * 
 * @Param:  - IN arg_MENumber - the microengine's number whose symbols should be registered
 *  with the passed microengine
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_microengine_register_internal_symbols( 
                                                       ix_uint32 arg_MENumber
                                                     );



/**
 * NAME: _ix_rm_microengine_patch_all_symbols
 *
 * DESCRIPTION: This function will patch all symbols registered for patching for the specified
 *  microengine
 * 
 * @Param:  - IN arg_MENumber - the microengine's number whose symbols should be patched into
 *  the microcode.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_microengine_patch_all_symbols( 
                                               ix_uint32 arg_MENumber
                                             );



/**
 * Global Microengine API manager
 */
extern _ix_microengine_data_manager g_MicroengineAPIManager;



/**
 * Global UCLO error strings
 */
extern const char* g_aUCLOErrorStrings[]; 
extern const ix_uint32 g_UCLOErrorStringCount;


/**
 * Global HALME error strings
 */
extern const char* g_aHALMEErrorStrings[]; 
extern const ix_uint32 g_HALMEErrorStringCount;


#if defined(_IX_RM_WORKBENCH_SUPPORT_)


/**
 * NAME: _ix_rm_download_microcode
 *
 * DESCRIPTION: This function will be registered with the ME debug library and will perform
 *      the copying the the UOF file into an internal RM buffer. That will give the oportunity
 *      to RM to complete the the symbol patching before the code is downloaded into the
 *      uStore.
 * 
 * @Param:  - IN arg_pUCodeBuffer - address of the Workbench downloaded uCode image.
 * @Param:  - IN arg_pUCodeSize - size of the Workbench downloaded uCode image
 *      upon return
 *
 * @Return: 0 if successful or a nonzero for failure.
 */
unsigned int _ix_rm_download_microcode( 
                               const unsigned char* arg_pUCodeBuffer, 
                               unsigned long arg_UCodeSize
                             );



#endif /* defined(_IX_RM_WORKBENCH_SUPPORT_) */


#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__INTERNAL_MICROENGINE_H__) */
