/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      internal_pci_communication.h
 *
 * = DESCRIPTION
 *      This file describes all the internal types and function prototypes
 *      used in the PCI communication infrastructure.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 *      Aaron Luk
 *      aaron.luk@intel.com
 *
 * = CHANGE HISTORY
 *      7/9/2002 10:31:00 AM - creation time 
 * ============================================================================
 * $Id: internal_pci_communication.h,v 1.20 2003/03/04 19:32:11 ageorges Exp $
 */

#if !defined(__INTERNAL_PCI_COMMUNICATION_H__)
#define __INTERNAL_PCI_COMMUNICATION_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */


/**
 * DESCRIPTION: Size of the meta in each buffer in the HW buffer
 *              free list allocated for the PCI Rx thread.
 *
 */
#define IX_PCI_MAX_HW_META_SIZE (sizeof(ix_hw_buffer_meta))

/**
 * DESCRIPTION: Size of the meta in each buffer in the SW buffer
 *              free list allocated for the PCI Rx thread.
 *
 */
#define IX_PCI_MAX_SW_META_SIZE (sizeof(ix_sw_buffer_meta))


/**
 * DESCRIPTION: Size of the data in each buffer in the HW buffer
 *              free list allocated for the PCI Rx thread.
 *
 */
#define IX_PCI_MAX_HW_DATA_SIZE 2048

/**
 * DESCRIPTION: Size of the data in each buffer in the SW buffer
 *              free list allocated for the PCI Rx thread.
 *
 */
#define IX_PCI_MAX_SW_DATA_SIZE 2048

/**
 * DESCRIPTION: Number of buffers in the HW buffer free list
 *              allocated for the PCI Rx thread.
 *
 */
#define IX_PCI_HW_FL_SIZE 1024

/**
 * DESCRIPTION: Number of buffers in the SW buffer free list
 *              allocated for the PCI Rx thread.
 *
 */
#define IX_PCI_SW_FL_SIZE 1024

/**
 * DESCRIPTION: Tx timeout in milliseconds.
 *
 */
#if defined(_IX_RM_DEBUG_)
    #define IX_PCI_TX_TIMEOUT IX_OSSL_WAIT_FOREVER
#else
    #define IX_PCI_TX_TIMEOUT 2000
#endif /* defined(_IX_RM_DEBUG_) */

/**
 * DESCRIPTION: Value read from/written to mailbox register 1,
 *              indicates that Rx has finished processing a
 *              a buffer.
 *
 */
#define IX_PCI_SIGNAL_RX_FROM_TX 0x01

/**
 * DESCRIPTION: Value read from/written to mailbox register 1,
 *              indicates that Tx has a buffer ready to be
 *              transferred.
 *
 */
#define IX_PCI_SIGNAL_TX_FROM_RX 0x02

/**
 * DESCRIPTION: Value read from/written to mailbox register 1,
 *              indicates that everything went ok and processing
 *              can continue.
 *
 */
#define IX_PCI_RX_SIGNAL_OK 0x0
#define IX_PCI_TX_SIGNAL_OK 0x0

/**
 * DESCRIPTION: Value read from/written to mailbox register 1,
 *              indicates that an error has occurred on the other side.
 *
 */
#define IX_PCI_RX_SIGNAL_ERROR 0x4
#define IX_PCI_TX_SIGNAL_ERROR 0x8

/**
 * TYPENAME: ix_pci_buffer_info
 * 
 * DESCRIPTION: This type stores the information relating to the buffer
 *              to be transferred over PCI.
 *
 */
typedef struct ix_s_pci_buffer_info
{
    /* Indicates whether the buffer is HW or SW. */
    ix_buffer_type m_BufferType;

    /* Indicates whether this buffer is the last in a chain. */
    ix_uint32 m_EOPFlag;

    /* Identifies the memory type of the meta. */
    ix_memory_type m_MetaMemoryType;
    /* Size of the meta in bytes. */
    ix_uint32 m_MetaSize;
    /* Offset to the meta. */
    ix_uint32 m_MetaOffset;

    /* Identifies the memory type of the data. */
    ix_memory_type m_DataMemoryType;
    /* Size of the data in bytes. */
    ix_uint32 m_DataSize;
    /* Offset to the data. */
    ix_uint32 m_DataOffset;
} ix_pci_buffer_info;



/**
 * TYPENAME: ix_pci_buffer_data_type
 * 
 * DESCRIPTION: Used to indicate whether a buffer is a packet
 *              or a message.
 *
 */
typedef enum ix_e_pci_buffer_data_type
{
    IX_PCI_BUFFER_DATA_TYPE_FIRST = 0,
    IX_PCI_BUFFER_DATA_TYPE_PACKET = IX_PCI_BUFFER_DATA_TYPE_FIRST,
    IX_PCI_BUFFER_DATA_TYPE_MESSAGE,
    IX_PCI_BUFFER_DATA_TYPE_LAST
} ix_pci_buffer_data_type;

/**
 * TYPENAME: ix_pci_control_info
 * 
 * DESCRIPTION: Control information for a buffer to be transferred
 *              over PCI.
 *
 */
typedef struct ix_s_pci_control_info
{
    /* Indicates whether the buffer is a packet or a message. */
    ix_pci_buffer_data_type m_PCIBufferDataType;
    /**
     * Buffer information, identifies memory type and size for meta-data
     * and data, whether this is a HW or SW buffer, and whether this
     * buffer is the last buffer in a chain.
     */
    ix_pci_buffer_info m_BufferInfo;
    /* User data for this buffer chain. */
    ix_uint32 m_UserData;
    /* Destination communication ID for this buffer. */
    ix_communication_id m_DestinationCommId;

    /* Padding to make this structure a multiple of 32 bytes. */
    /*ix_uint32 res0;
    ix_uint32 res1;
    ix_uint32 res2;
    ix_uint32 res3;
    ix_uint32 res4;*/
} ix_pci_control_info;



/**
 * TYPENAME: ix_pci_engine_interface_func_ptr
 * 
 * DESCRIPTION: This is the definition of a function type for the dispatch engine
 *          interface. It is assumed that all functions of the interface are of this
 *          type. This approach will allow us to abstract the internal communication 
 *          API implementation. We will be able to allow the users to set their own 
 *          ISR and implement their own engine.
 *
 */
typedef ix_dispatch_engine_interface_func_ptr ix_pci_engine_interface_func_ptr;



/**
 * TYPENAME: _ix_rm_pci_engine_interface
 * 
 * DESCRIPTION: This type defines, in an abstract way, the PCI Tx and Rx thread
 *          interface. This abstraction provides support for the PCI Tx and Rx threads.
 *
 */
typedef _ix_dispatch_engine_interface _ix_rm_pci_engine_interface;



/**
 * TYPENAME: _ix_rm_pci_buffer_retrieve_func_ptr
 * 
 * DESCRIPTION: This is a blocking function that is called by the Tx thread
 *              in order to retrieve the data queued for the peer subsystem.
 *              We will provide a default implementation, but that can be changed
 *              as required.
 *
 */
typedef ix_error (*_ix_rm_pci_buffer_retrieve_func_ptr)(ix_pci_buffer_data_type* arg_pBufferType,
                                               ix_buffer_handle* arg_pBufferHandle,
                                               ix_uint32* arg_pUserData,
                                               ix_communication_id* arg_pCommId);




/**
 * TYPENAME: ix_pci_controller
 * 
 * DESCRIPTION: Control structure for the PCI communication threads.
 *
 */
typedef struct ix_s_pci_controller
{
    /**
     * Signal from the other side, contained in the mailbox_1 register,
     * will be read into here.
     */
    ix_uint32 m_SignalValue;
    /* Mailbox remote registers semaphore */
    ix_ossl_sem_t m_MBRegsSemaphore;

    /* Base address of remote PCI CSRs */
    ix_uint32 m_RemoteCSRBase;
    /* Base address of remote SRAM */
    ix_uint32 m_RemoteSRAMBase;
    /* Base address of remote DRAM */
    ix_uint32 m_RemoteDRAMBase;

    /* Tx data members. */
    /* Tx thread. */
    ix_ossl_thread_t m_TxThread;
    /* Flag that indicates termination of Tx thread. */
    ix_uint32 m_TxTerminateFlag;
    /**
     * Semaphore for Tx thread, notifies Tx that Rx has
     * finished processing a buffer.
     */
    ix_ossl_sem_t m_TxSemaphore;
    /* Control information for a buffer to transmit. */
    ix_pci_control_info* m_pTxPCIControlInfo;

    /* Physical offset of control information in the DRAM memory. */
    ix_uint32 m_TxPCIControlInfoOffset;

    /* Contains Tx thread functions. */
    _ix_rm_pci_engine_interface*  m_pTxInterface;
    /**
     * Pointer to function that retrieves a buffer to transmit
     * from the communication API.
     */
    _ix_rm_pci_buffer_retrieve_func_ptr m_TxBufferRetrieve;

    /* Rx data members. */
    /* Tx thread. */
    ix_ossl_thread_t m_RxThread;
    /* Flag that indicates termination of Rx thread. */
    ix_uint32 m_RxTerminateFlag;
    /**
     * Semaphore for Rx thread, notifies Rx that Tx has
     * a buffer ready for transmit.
     */
    ix_ossl_sem_t m_RxSemaphore;

    /* freelist of HW ix_buffers used by the Rx side */
    ix_buffer_free_list_handle m_hPCIHwFreeList;

    /* user free list related data */
    ix_buffer_free_list_handle m_hUserHwFreeList;

    /* freelist of SW ix_buffers used by the Rx side */
    ix_buffer_free_list_handle m_hPCISwFreeList;
    /* Contains Rx thread functions. */
    _ix_rm_pci_engine_interface*  m_pRxInterface;
} ix_pci_controller;



/**
 * NAME: _ix_rm_pci_communication_init
 *
 * DESCRIPTION: This function will be called by _ix_rm_comunication_init()
 *          at the time the resource manager initializer function is
 *          called. The function will initialize all the PCI communication.
 * @Param:  - IN arg_pContext - context that has been passed to RM at the initialization
 *          of the service.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_pci_communication_init(void* arg_pContext);


/**
 * NAME: _ix_rm_pci_communication_fini
 *
 * DESCRIPTION: This function will be called by _ix_rm_communication_fini()
 *          at the time the resource manager finalizer function is called.
 *          The function will provide cleanup for PCI communication.
 * @Param:  - IN arg_pContext - context that has been passed to RM at the initialization
 *          of the service.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_pci_communication_fini(void* arg_pContext);


/**
 * Global communication controller
 */
extern ix_pci_controller g_PCICommunicationController;


#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__INTERNAL_PCI_COMMUNICATION_H__) */
