/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 4
 *
 * = LIBRARY
 *      Resource Manager Library
 *
 * = MODULE
 *      Library Interface
 *
 * = FILENAME
 *      internal_sync.h
 *
 * = DESCRIPTION
 *      
 *
 * = AUTHOR
 *       Roland Arajs
 *       roland.arajs@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      11/27/2001 10:36:13 AM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 * $Id: internal_sync.h,v 1.3 2002/09/18 17:27:15 lbmiller Exp $
 */

#if !defined(__INTERNAL_SYNC_H__)
#define __INTERNAL_SYNC_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */


/**
 * We provide a generic synchronization adapter, where the users may provide their
 * own synchronization objects. The adapter provides placeholders for two functions: 
 * lock and unlock that have to be provided by the users. These functions must provide
 * the logic of the synchronization mechanism. It is up to the user what mechanism to 
 * use (e.g. semaphore, critical section, mutex, etc.) 
 * The adapter can be associated with some entity that needs to be synchronized and 
 * provides the flexibility that synchronization may be defined externally.
 */


/**
 * TYPENAME: ix_synchronization_adapter
 * 
 * DESCRIPTION: Synchronization adapter that consists of two sync-function placeholders.
 *
 */
typedef struct ix_s_synchronization_adapter
{
    ix_uint32           m_SyncActive;       
                                            /**
                                             * Flag that both of the following two
                                             * function pointers are non NULL. We introduce
                                             * this extra member to improve the runtime
                                             * performance.
                                             */

    ix_sa_lock_ptr      m_LockFuncPtr;      /* Lock function placeholder */
    ix_sa_unlock_ptr    m_UnlockFuncPtr;    /* Unlock function placeholder */
    void*               m_pContext;         /* Will be passed back to each 
                                             invocation of lock and unlock function */
} ix_synchronization_adapter;



/**
 * NAME: _ix_rm_sa_default_lock
 *
 * DESCRIPTION: This function is provided for internal lock support. The function
 *          makes the assumption that the context is a pointer to an OSSL semaphore.
 *          The call will aquire the semaphore, so any subsequent calls to _ix_rm_sa_default_lock
 *          with the same context will block till the semaphore is released.
 * 
 * @Param:  - INOUT arg_pContext - it will be interpreted as a pointer to an OSSL
 *          semaphore of the type ix_ossl_sem_t.
 *
 */
void _ix_rm_sa_default_lock( 
                             void* arg_pContext 
                           );


/**
 * NAME: _ix_rm_sa_default_unlock
 *
 * DESCRIPTION: This function is provided for internal unlock support. The function
 *          makes the assumption that the context is a pointer to an OSSL semaphore.
 *          The call will release the semaphore, so any tasks waiting on that object
 *          will be released and one willa quire the semaphore and continue running.
 * 
 * @Param:  - INOUT it will be interpreted as a pointer to an OSSL
 *          semaphore of the type ix_ossl_sem_t.
 *
 */
void _ix_rm_sa_default_unlock( 
                               void* arg_pContext 
                             );


/**
 * NAME: _ix_rm_sa_sync_adapter_init
 *
 * DESCRIPTION:   Initializes the synchronization adapter object.
 * 
 * @Param:  - IN arg_pAdapter - pointer to adapter to be initialized 
 * @Param:  - IN arg_LockFunction - the lock function
 * @Param:  - IN arg_UnlockFunction - the unlock function
 * @Param:  - IN arg_pContext - context to be passed to the lock and unlock functions
 */
void _ix_rm_sa_sync_adapter_init( 
                                 ix_synchronization_adapter*  arg_pAdapter,
                                 ix_sa_lock_ptr               arg_LockFunction,
                                 ix_sa_unlock_ptr             arg_UnlockFunction,
                                 void*                        arg_pContext
                               );



/**
 * MACRO NAME: IX_RM_SYNC_ADAPTER_LOCK
 *
 * DESCRIPTION: If _IX_RM_MULTI_THREADED_ symbol is defined, performs a data lock
 *              based on the user-defined mechanism. Otherwise, resolves to a NO-OP.
 *
 * @Param:  - IN arg_pAdapter       Requires pointer to an ix_synchronization_adapter structure
 */

#if defined(_IX_RM_MULTI_THREADED_)

#define IX_RM_SYNC_ADAPTER_LOCK( \
                                 arg_pAdapter \
                               ) \
                               ( arg_pAdapter->m_SyncActive \
                               ? arg_pAdapter->m_LockFuncPtr(arg_pAdapter->m_pContext) \
                               : ((void)0) )

#else

#define IX_RM_SYNC_ADAPTER_LOCK( \
                                 arg_pAdapter \
                               ) \
                               ((void)0)

#endif /* defined (_IX_RM_MULTI_THREADED_) */

/**
 * MACRO NAME: IX_RM_SYNC_ADAPTER_UNLOCK
 *
 * DESCRIPTION: If _IX_RM_MULTI_THREADED_ symbol is defined, performs a data unlock
 *              based on the user-defined mechanism. Otherwise, resolves to a NO-OP.
 *
 * @Param:  - IN arg_pAdapter       Requires pointer to an ix_synchronization_adapter structure
 */

#if defined(_IX_RM_MULTI_THREADED_)

#define IX_RM_SYNC_ADAPTER_UNLOCK( \
                                   arg_pAdapter \
                                 ) \
                                 ( arg_pAdapter->m_SyncActive \
                                 ? arg_pAdapter->m_UnlockFuncPtr(arg_pAdapter->m_pContext) \
                                 : ((void)0) )

#else

#define IX_RM_SYNC_ADAPTER_UNLOCK( \
                                   arg_pAdapter \
                                 ) \
                                 ((void)0)

#endif /* defined (_IX_RM_MULTI_THREADED_) */




#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__INTERNAL_SYNC_H__) */
