/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 4
 *
 * = LIBRARY
 *      Resource Manager Library
 *
 * = MODULE
 *      System API
 *
 * = FILENAME
 *      internal_system.h
 *
 * = DESCRIPTION
 *      This file defines all data types and function prototypes required internaly by the 
 *      System API module.
 *
 * = AUTHOR
 *       Adrian Georgescu
 *       adrian.georgescu@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      10/26/2001 11:31:21 AM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 * $Id: internal_system.h,v 1.9 2003/07/15 20:55:03 ageorges Exp $
 */

#if !defined(__INTERNAL_SYSTEM_H__)
#define __INTERNAL_SYSTEM_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */



/**
 * The next section provides error handling macros 
 */


/**
 * MACRO NAME: IX_RM_ERROR_NEW
 *
 * DESCRIPTION: This macro returns a new ix_error.
 *
 * @Param:  - IN arg_ErrorCode - of integral type represents the error code for this error.
 * @Param:  - IN arg_pFunctionName - of type const char* represents the function name where the error happens.
 * @Param:  - IN arg_pErrorReason - of type const char* represents some text that is passed to the macro.
 * @Param:  - IN arg_ErrorChain - of type ix_error represents the error this error should be chained to. 
 *
 * @Return: This macro returns a new ix_error.
 */
#if defined( _IX_RM_DEBUG_ )

#define IX_RM_ERROR_NEW( \
                         arg_ErrorCode, \
                         arg_pFunctionName, \
                         arg_pErrorReason, \
                         arg_ErrorChain \
                       ) \
                        (ix_ossl_message_log("Error in file: %s, line: %d\n\tFunction: %s\n\tReason: " \
                            "%s\n\tError Chain: 0x%08lx\n\tError : 0x%08lx\n", \
                            __FILE__, \
                            __LINE__, \
                            arg_pFunctionName, \
                            arg_pErrorReason, \
                            (ix_uint32)arg_ErrorChain, \
                            (ix_uint32)IX_ERROR_NEW(arg_ErrorCode, IX_ERROR_LEVEL_LOCAL) \
                        ), \
                        IX_ERROR_NEW(arg_ErrorCode, IX_ERROR_LEVEL_LOCAL))

#else

#define IX_RM_ERROR_NEW( \
                         arg_ErrorCode, \
                         arg_pFunctionName, \
                         arg_pErrorReason, \
                         arg_ErrorChain \
                       ) \
                        IX_ERROR_NEW((ix_uint32)arg_ErrorCode, IX_ERROR_LEVEL_LOCAL)

#endif /* defined _IX_RM_DEBUG_ */

/**
 * MACRO NAME: IX_RM_ERROR_LOG
 *
 * DESCRIPTION: This macro logs the error on the default output stream.
 *
 * @Param:  - IN arg_ErrorCode - of integral type represents the error code for this error.
 * @Param:  - IN arg_pFunctionName - of type const char* represents the function name where the error happens.
 * @Param:  - IN arg_pErrorReason - of type const char* represents some text that is passed to the macro.
 * @Param:  - IN arg_ErrorChain - of type ix_error represents the error this error should be chained to. 
 *
 * @Return: This macro returns a new ix_error.
 */

#define IX_RM_ERROR_LOG( \
                         arg_ErrorCode, \
                         arg_pFunctionName, \
                         arg_pErrorReason, \
                         arg_ErrorChain \
                       ) \
                        do \
                        { \
                            ix_ossl_message_log("Error in file: %s, line: %d\n\tFunction: %s\n\tReason: " \
                                "%s\n\tError Chain: 0x%08lx\n\tError : 0x%08lx\n", \
                                __FILE__, \
                                __LINE__, \
                                arg_pFunctionName, \
                                arg_pErrorReason, \
                                (ix_uint32)arg_ErrorChain, \
                                (ix_uint32)IX_ERROR_NEW(arg_ErrorCode, IX_ERROR_LEVEL_LOCAL)\
                            ); \
                        } \
                        while(0)


/**
 * MACRO NAME: IX_RM_ASSERT
 *
 * DESCRIPTION: This macro will return an error from the current function with specified comment
 *          if the condition is FALSE. If the condition arg_Condition is TRUE then nothing happens.
 *          In RELEASE mode this macro will be a no op.
 *
 * @Param:  - IN arg_Condition - the condition that should be checked. Should be of integral type.
 *          Nonzero is considered TRUE and zero FALSE.
 * @Param:  - IN arg_pFunctionName - of type const char* represents the function name where the error happens.
 * @Param:  - IN arg_pErrorReason - of type const char* represents a comment string that should be passed to
 *          to the returned error token.
 *
 */
#if defined(_IX_RM_DEBUG_)

#define IX_RM_ASSERT( \
                      arg_Condition, \
                      arg_pFunctionName, \
                      arg_pErrorReason \
                    ) \
                    if(!(arg_Condition)) \
                    { \
                        return IX_RM_ERROR_NEW( \
                                                IX_RM_ERROR_ASSERTION, \
                                                arg_pFunctionName, \
                                                arg_pErrorReason, \
                                                0 \
                                              ); \
                    }

#else

#define IX_RM_ASSERT( \
                      arg_Condition, \
                      arg_pFunctionName, \
                      arg_pErrorReason \
                    ) \
                    do \
                    { \
                    } \
                    while(0)

#endif /* defined _IX_RM_DEBUG_ */


/**
 * MACRO NAME: IX_RM_ERROR_CHAIN
 *
 * DESCRIPTION: This macro will chain Resource Manager's errors. 
 *
 * @Param:  - IN arg_ErrorChain - of ix_error type represents the error token.
 * @Param:  - IN arg_pFunctionName - of type const char* represents the function name where the error happens.
 * @Param:  - IN arg_pErrorReason - of type const char* represents some text that is passed to the macro.
 *
 * @Return: A new ix_error token is returned.
 */
#if defined(_IX_RM_DEBUG_)

#define IX_RM_ERROR_CHAIN( \
                           arg_ErrorChain, \
                           arg_pFunctionName, \
                           arg_pErrorReason \
                         ) \
                        (ix_ossl_message_log("Error in file: %s, line: %d\n\tFunction: %s\n\tReason: %s\n", \
                            __FILE__, \
                            __LINE__, \
                            arg_pFunctionName, \
                            arg_pErrorReason \
                        ), arg_ErrorChain)

#else

#define IX_RM_ERROR_CHAIN( \
                           arg_ErrorChain, \
                           arg_pFunctionName, \
                           arg_pErrorReason \
                         ) \
                         arg_ErrorChain

#endif /* defined _IX_RM_DEBUG_ */


/**
 * TYPENAME: _ix_init_object
 * 
 * DESCRIPTION: This structure defines an initailization object. This structure will
 *          keep track of all details related to the initailization process of 
 *          Resource Manager library.
 *
 */
typedef struct _ix_s_init_object
{
    ix_uint32       m_InitCount; /* an initailization count */

#if defined(_IX_RM_MULTI_THREADED_)

    ix_synchronization_adapter      m_SyncAdapter; /* synchronization adapter */

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

} _ix_init_object;


#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * TYPENAME: _ix_sys_config
 * 
 * DESCRIPTION: This structure is an internal version of the ix_sys_config structure
 *          that will provide the proper syncronization when needed.
 *
 */
typedef struct _ix_s_sys_config
{
    ix_sys_config       m_SysConfig; /* a copy of the original ix_sys_config */

#if defined(_IX_RM_MULTI_THREADED_)

    ix_synchronization_adapter      m_SyncAdapter; /* synchronization adapter */
    
#endif /* end defined(_IX_RM_MULTI_THREADED_) */

} _ix_sys_config;



/**
 * NAME: _ix_rm_system_init
 *
 * DESCRIPTION: This function should be called at the time the resource manager initializer 
 *          function is called. The function will initialize all the System API.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_system_init(void);


/**
 * NAME: _ix_rm_system_fini
 *
 * DESCRIPTION: This function should be called at the time the resource manager finalizer function
 *          is called. The function will provide cleanup for System API.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_system_fini(void);



/**
 * This object will hold system software configuration information.
 * It will be used by ix_rm_sys_config_get() and ix_rm_sys_config_set() 
 * interface calls.
 */
extern _ix_sys_config g_SystemConfiguration;

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */


/**
 * This object will keep track of all details related to the initailization process of 
 * Resource Manager library.
 */
extern _ix_init_object g_RMLibInitialization; 


#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__INTERNAL_SYSTEM_H__) */
