/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      me_notification.h
 *
 * = DESCRIPTION
 *      This file describes all data types and function prototypes exported by
 *      the ME notification API
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      8/8/2003 2:20:36 PM - creation time 
 * ============================================================================
 * $Id: me_notification.h,v 1.3 2003/10/10 21:09:00 ageorges Exp $
 */

#if !defined(__ME_NOTIFICATION_H__)
#define __ME_NOTIFICATION_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */



/**
 * TYPENAME: ix_me_transfer_register_type
 * 
 * DESCRIPTION: This enumerated type describes all the types of ME transfer
 *  registers. The XScale can read the SRAM write transfer registers of an ME
 *  and can write the SRAM or DRAM read transfer registers of an ME
 *
 */
typedef enum ix_e_me_transfer_register_type
{
    IX_ME_TRANSFER_REGISTER_TYPE_FIRST = 0,
    IX_ME_TRANSFER_REGISTER_TYPE_SRAM = IX_ME_TRANSFER_REGISTER_TYPE_FIRST,
    IX_ME_TRANSFER_REGISTER_TYPE_DRAM,
    IX_ME_TRANSFER_REGISTER_TYPE_LAST
} ix_me_transfer_register_type;


/**
 * NAME: ix_rm_me_transfer_register_read
 *
 * DESCRIPTION: This function will read the value of an ME transfer register for the
 *  specified ME, context and register type. Only the IX_ME_TRANSFER_REGISTER_TYPE_SRAM
 *  registers can be read. During the call, the code will check if the ME is in the 4 
 *  contexts mode or 8 contexts mode. In the first case the valid contexts numbers are
 *  0,2,4,6, whereas for the second case they are in the range 0-7.
 * 
 * @Param:  - IN arg_MENumber - the microengine whose transfer register we want to read
 * @Param:  - IN arg_ContextNumber - the ME context whose transfer register we want 
 *  to read
 * @Param:  - IN arg_RegisterType - the type of the ME transfer register we want to read.
 *  For reading just the IX_ME_TRANSFER_REGISTER_TYPE_SRAM is accepted.
 * @Param:  - IN arg_RegisterIndex - the index of the ME transfer register to be read.
 *  The value of the index is in the range 0-31 in the case the ME is in 4 contexts mode
 *  , and in the range 0-15 for the 8 contexts mode.
 * @Param:  - OUT arg_pRegisterValue - address where the value of the ME transfer register
 *  will be stored upon return.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_me_transfer_register_read(
                                          ix_uint32 arg_MENumber,
                                          ix_uint32 arg_ContextNumber, 
                                          ix_me_transfer_register_type arg_RegisterType,
                                          ix_uint32 arg_RegisterIndex,
                                          ix_uint32* arg_pRegisterValue
                                        );



/**
 * NAME: ix_rm_me_transfer_register_write
 *
 * DESCRIPTION: This function will write the value of an ME transfer register for the
 *  specified ME, context and register type. During the call, the code will check if the ME 
 *  is in the 4 contexts mode or 8 contexts mode. In the first case the valid contexts 
 *  numbers are 0,2,4,6, whereas for the second case they are in the range 0-7.
 * 
 * @Param:  - IN arg_MENumber - the microengine whose transfer register we want to write
 * @Param:  - IN arg_ContextNumber - the ME context whose transfer register we want 
 *  to write.
 * @Param:  - IN arg_RegisterType - the type of the ME transfer register we want to write.
 * @Param:  - IN arg_RegisterIndex - the index of the ME transfer register to be read.
 *  The value of the index is in the range 0-31 in the case the ME is in 4 contexts mode
 *  , and in the range 0-15 for the 8 contexts mode.
 * @Param:  - IN arg_RegisterValue - the value to be written in the ME transfer register.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_me_transfer_register_write(
                                           ix_uint32 arg_MENumber,
                                           ix_uint32 arg_ContextNumber, 
                                           ix_me_transfer_register_type arg_RegisterType,
                                           ix_uint32 arg_RegisterIndex,
                                           ix_uint32 arg_RegisterValue
                                         );



/**
 * NAME: ix_rm_me_signal
 *
 * DESCRIPTION: This function will send a specified signal to the specified ME and 
 *  context. The signalling will be achieved by writing the SAME_ME_REGISTER for the
 *  desired ME. There is a 3 cycles latency from the write CSR command untill the CSR
 *  actually written, and 8 cycles until the ME will signaled after the CSR write.
 *  That should be considered when issuing many signals to the same ME.
 *  During the call, the code will check if the ME is in the 4  contexts mode or 8 
 *  contexts mode. In the first case the valid context numbers are 0,2,4,6 whereas
 *  for the second case they are in the range 0-7.
 * 
 * @Param:  - IN arg_MENumber - the microengine whose which we want to signal.
 * @Param:  - IN arg_ContextNumber - the ME context for which the signal is intended.
 * @Param:  - IN arg_SignalNumber - the signal number to be sent to the specified ME
 *  context. Valid values are in the range 1-15
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_me_signal(
                          ix_uint32 arg_MENumber,
                          ix_uint32 arg_ContextNumber, 
                          ix_uint32 arg_SignalNumber
                        );





#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__ME_NOTIFICATION_H__) */
