/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      me_xscale_locks.h
 *
 * = DESCRIPTION
 *      This file describes all data types and function prototypes exported by
 *      the ME/XScale locks API
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      8/8/2003 1:33:17 PM - creation time 
 * ============================================================================
 * $Id: me_xscale_locks.h,v 1.3 2003/10/10 00:30:20 ageorges Exp $
 */

#if !defined(__ME_XSCALE_LOCKS_H__)
#define __ME_XSCALE_LOCKS_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */


/**
 * The ME/XScale locks will be used for synchronization between ME and XScale data access
 * and they are based on the SCRATCH memory atomic operations. On the XScale side the API will
 * suffice to do lock/unlock operations. For a particular lock object, the XScale application
 * will have to retrieve the SCRATCH byte offset of the underlying SCRATCH memory location
 * used for atomic set and test and clear and the bit position of the lock into that memory 
 * location, and patch these values into the uCode (or pass it through any other mean). On the
 * ME side, in order for MEs to acquire the lock they will have to perform a bit test and clear
 * for the passed SCRATCH offset on the bit position passed from the XScale. The returned value
 * will be masked for the XScale passed bit position of the lock and if the resulting
 * value from the operation is non 0 then there was a successful locking, otherwise
 * it means that the lock is held by somebody else. In order to release a held lock the MEs
 * will have to perform a bit set operation for the SCRATCH offset and bit position of the lock. 
 * On the XScale side the API will guaratee if the ME acquired the lock, then the lock will not
 * be released by the xScale, but on the ME side is the responsibility
 * of the programmers to insure this behaviour (the same ME/context that acquired the lock will
 * be the one that will release it).
 */

/**
 * TYPENAME: ix_me_xscale_lock_handle
 * 
 * DESCRIPTION: Generic type for me/xscale lock handle
 *
 */
typedef ix_handle ix_me_xscale_lock_handle;


/**
 * NAME: ix_rm_me_xscale_lock_new
 *
 * DESCRIPTION: This function will create a new me/xscale lock object.
 * 
 * @Param:  - OUT  arg_pMeXScaleLockHandle - location where a handle to the newly created
 *  me/xscale lock will be stored upon return.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_me_xscale_lock_new(
                                   ix_me_xscale_lock_handle* arg_pMeXScaleLockHandle
                                 );


/**
 * NAME: ix_rm_me_xscale_lock_delete
 *
 * DESCRIPTION: This function will delete the specified me/xscale lock object and
 *  will release all resources associated with it.
 * 
 * @Param:  - IN arg_hMeXScaleLock - the handle of the me/xscale lock to be deleted
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_me_xscale_lock_delete(
                                      ix_me_xscale_lock_handle arg_hMeXScaleLock
                                    );



/**
 * NAME: ix_rm_me_xscale_lock_acquire
 *
 * DESCRIPTION: This function will acquire the specified me/xscale lock. If the lock
 *  is already taken then an error will be returned. 
 * 
 * @Param:  - IN arg_hMeXScaleLock - the handle of the lock we are trying to acquire
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_me_xscale_lock_acquire( 
                                       ix_me_xscale_lock_handle arg_hMeXScaleLock
                                     );




/**
 * NAME: ix_rm_me_xscale_lock_release
 *
 * DESCRIPTION: This function will release the specified ME/XScale lock. If the lock
 *  has been acquired by the MEs, an error will be returned. 
 * 
 * @Param:  - IN arg_hMeXScaleLock - the handle of the lock we are releasing
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_me_xscale_lock_release(
                                       ix_me_xscale_lock_handle arg_hMeXScaleLock
                                     );




/**
 * TYPENAME: ix_me_xscale_lock_status
 * 
 * DESCRIPTION: This enumeration describes the possible status of a me/xscale lock
 *
 */
typedef enum ix_e_me_xscale_lock_status
{
    IX_ME_XSCALE_LOCK_STATUS_FIRST = 0,
    IX_ME_XSCALE_LOCK_STATUS_AVAILABLE = IX_ME_XSCALE_LOCK_STATUS_FIRST,
    IX_ME_XSCALE_LOCK_STATUS_UNAVAILABLE,
    IX_ME_XSCALE_LOCK_STATUS_LAST
} ix_me_xscale_lock_status;



/**
 * TYPENAME: ix_me_xscale_lock_owner
 * 
 * DESCRIPTION: This type describes all the possible owners of an XScale\ME lock
 *
 */
typedef enum ix_e_me_xscale_lock_owner
{
    IX_ME_XSCALE_LOCK_OWNER_FIRST = 0,
    IX_ME_XSCALE_LOCK_OWNER_NONE = IX_ME_XSCALE_LOCK_OWNER_FIRST,
    IX_ME_XSCALE_LOCK_OWNER_XSCALE,
    IX_ME_XSCALE_LOCK_OWNER_ME,
    IX_ME_XSCALE_LOCK_OWNER_LAST
} ix_me_xscale_lock_owner;


/**
 * TYPENAME: ix_me_xscale_lock_info
 * 
 * DESCRIPTION: This structure provides information about a given me/xscale lock
 *
 */
typedef struct ix_s_me_xscale_lock_info
{
    ix_uint32 m_ScratchOffset;  /* the byte offset of the underlying SCRATCH memory location of the lock */
    ix_uint32 m_BitPosition; /* bit position of the lock into the underlying SCRATCH memory location */
    ix_me_xscale_lock_status m_Status;  /* status of the lock */
    ix_me_xscale_lock_owner m_Owner;  /* owner of the lock */
} ix_me_xscale_lock_info;


/**
 * NAME: ix_rm_me_xscale_lock_get_info
 *
 * DESCRIPTION: This function will return useful information about the passed lock
 * 
 * @Param:  - IN arg_hMeXScaleLock - the handle of the lock we are interested in
 * @Param:  - OUT arg_pMeXScaleLockInfo - address of structure were the lock information
 *  will be stored upon return.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_me_xscale_lock_get_info(
                                        ix_me_xscale_lock_handle arg_hMeXScaleLock,
                                        ix_me_xscale_lock_info* arg_pMeXScaleLockInfo
                                      );



#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__ME_XSCALE_LOCKS_H__) */
