/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      microengine.h
 *
 * = DESCRIPTION
 *      This file will define all types exported by the Mcroengine API. At the same time will
 *      declare all function prototypes of the exported interface.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      6/4/2002 12:09:06 PM - creation time 
 * ============================================================================
 * $Id: microengine.h,v 1.14 2003/09/19 23:13:12 ageorges Exp $
 */

#if !defined(__MICROENGINE_H__)
#define __MICROENGINE_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */

/**
 * TYPENAME: ix_imported_symbol
 * 
 * DESCRIPTION: This structure represents a name/value pair of the symbols to be patched into 
 *          microcode.
 */
typedef struct ix_s_imported_symbol 
{
    ix_uint32           m_Value;
    char*               m_Name;
} ix_imported_symbol;



/**
 * NAME: ix_rm_ueng_set_ucode
 *
 * DESCRIPTION: This function sets the microcode image for microengines. The arg_pImageName 
 *  represents the image name containing the microcode in UOF format. The images are not 
 *  loaded into the microengines' microstore because symbols still 
 *  need to be patched. The images are loaded at the time of the ix_rm_ueng_load()
 *  call.
 * 
 * @Param:  - IN arg_pImageName - represents the image name containing the microcode in 
 *          UOF format.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_ueng_set_ucode( 
                               const char* arg_pImageName 
                             );


/**
 * NAME: ix_rm_ueng_map_ucode
 *
 * DESCRIPTION: This function sets the microcode image for microengines from a passed buffer.
 *  The images are not loaded into the microengines' microstore because symbols still 
 *  need to be patched. The images are loaded at the time of the ix_rm_ueng_load()
 *  call. The arg_pUOFImage should point to a RW memory location rather than a RO memory
 *  location.
 * 
 * @Param:  - IN arg_pUOFImage - represents the image buffer from where the microcode should be mapped
 * @Param:  - IN arg_Size - represents the size of the image buffer
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_ueng_map_ucode( 
                               void* arg_pUOFImage, 
                               ix_int32 arg_Size 
                             );

#if defined(_IX_RM_WORKBENCH_SUPPORT_)

/**
 * NAME: ix_rm_ueng_set_ucode_from_wb
 *
 * DESCRIPTION: This function will set the uCode downloaded from the WorkBench
 * 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_ueng_set_ucode_from_wb(void);

#endif /* defined(_IX_RM_WORKBENCH_SUPPORT_) */


/**
 * NAME: ix_rm_ueng_reset_all
 *
 * DESCRIPTION: This function stops and resets all active microengines.
 *  It removes the active UOF image from memory and brings the ME system
 *  into initial state.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_ueng_reset_all(void);


/**
 * NAME: ix_rm_ueng_patch_symbols
 *
 * DESCRIPTION: This function will patch imported symbols into the microcode. This call should
 *  be done after ix_rm_ueng_set_ucode() call and before ix_rm_ueng_load(). The RM library will
 *  store internally all patched symbols and will actually patch them at the time of
 *  ix_rm_ueng_load() call.
 * 
 * @Param:  - IN arg_UengNumber - represents the microengine number for which this call is 
 *          intended.
 * @Param:  - IN arg_SymbolsNumber - represents the number of symbols in the arg_Symbols 
 *          array that will be patched.
 * @Param:  - IN arg_aSymbols - represents an array of symbols that will be patched into 
 *          the microcode.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_ueng_patch_symbols( 
                                   ix_uint32 arg_MENumber, 
                                   ix_uint32 arg_SymbolsNumber, 
                                   const ix_imported_symbol arg_aSymbols[] 
                                 );


/**
 * NAME: ix_rm_ueng_load
 *
 * DESCRIPTION: This function will load the microcode into the microengine store. The 
 *          ix_rm_ueng_set_ucode() function should be called prior to this one. Also is
 *          expected that all imported symbols to be patched by this time.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_ueng_load(void);




/**
 * NAME: ix_rm_ueng_start
 *
 * DESCRIPTION: This function will start the passed microengine and will activate the
 *  specified contexts.
 * 
 * @Param:  - IN arg_MENumber - the microengine's number to be started.
 * @Param:  - IN arg_EnableContextMask - mask specifying what microengine contexts should
 *          be enabled at the start time.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_ueng_start(
                           ix_uint32 arg_MENumber, 
                           ix_bit_mask32 arg_EnableContextMask 
                         );



/**
 * TYPENAME: ix_me_start_node
 * 
 * DESCRIPTION: Structure describing an ME to be started
 *
 */
typedef struct ix_s_me_start_node
{
    ix_uint32       m_MENumber; /* ME number to be started */
    ix_bit_mask32   m_EnableContextMask; /* context mask for the contexts to be started for the specified ME */
} ix_me_start_node;


/**
 * NAME: ix_rm_ueng_start_group
 *
 * DESCRIPTION: This function will start the specified ME group and will activate the
 *  specified contexts. The MEs will be started in a very tight loop in this case.
 * 
 * @Param:  - IN arg_pMEGroup - the address of an array of ix_me_start_node structures
 *  specifying what MEs to be started.
 * @Param:  - IN ix_uint32 - the count of elements in the above passed ME group array.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_ueng_start_group(
                                 ix_me_start_node* arg_pMEGroup, 
                                 ix_uint32 arg_MEGroupCount 
                               );

/**
 * NAME: ix_rm_ueng_start_all
 *
 * DESCRIPTION: This function will start all the active MEs and will activate all the
 *  contexts. The MEs will be started in a very tight loop in this case. It depends on the
 *  loaded uCode which MEs are active.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_ueng_start_all(void);



/**
 * NAME: ix_rm_ueng_stop
 *
 * DESCRIPTION: This function will stop the specified microengine with all 
 *  active contexts.
 * 
 * @Param:  - IN arg_MENumber - the microengine's number to be stopped.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_ueng_stop(
                          ix_uint32 arg_MENumber
                        );



/**
 * NAME: ix_rm_ueng_reset
 *
 * DESCRIPTION: This function will reset the specified microengine. If the arg_ClearRegisters
 *  is non zero then all the microengine registers will be cleared and the microengine
 *  will be in the same state as at startup.
 * 
 * @Param:  - IN arg_MENumber - the microengine's number to be reset. 
 * @Param:  - IN arg_ClearRegisters - flag specifying if all registers should be cleared
 *          on reset.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_ueng_reset(
                           ix_uint32 arg_MENumber, 
                           ix_uint32 arg_ClearRegisters
                         );


/**
 * NAME: ix_rm_ueng_enable
 *
 * DESCRIPTION: This function will enable one or more of the microengine's contexts.
 *  The microengine should be started in order for this call to succeed.
 * 
 * @Param:  - IN arg_MENumber - the microengine's number whose contexts should be enabled. 
 * @Param:  - IN arg_EnableContextMask - mask specifying what microengine contexts should
 *          be enabled.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_ueng_enable( 
                            ix_uint32 arg_MENumber, 
                            ix_bit_mask32 arg_EnableContextMask 
                          );


/**
 * NAME: ix_rm_ueng_disable
 *
 *
 * DESCRIPTION: This function will disable one or more of the microengine's contexts.
 *  The microengine should be started in order for this call to succeed.
 * 
 * @Param:  - IN arg_MENumber - the microengine's number whose contexts should be disabled. 
 * @Param:  - IN arg_DisableContextMask - mask specifying what microengine contexts should
 *          be disabled.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_ueng_disable( 
                             ix_uint32 arg_MENumber, 
                             ix_bit_mask32 arg_DisableContextMask 
                           );


/**
 * This is a global flag that will signal if the uCode has been loaded or not
 * into the microstore.
 */
extern ix_uint32 g_MicrocodeLoaded;


#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__MICROENGINE_H__) */
