/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      remote_communication.h
 *
 * = DESCRIPTION
 *      This file defines all types and function prototypes exported by the remote
 *      communication API.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      11/27/2002 2:57:51 PM - creation time 
 * ============================================================================
 * $Id: remote_communication.h,v 1.2 2003/03/04 19:32:11 ageorges Exp $
 */

#if !defined(__REMOTE_COMMUNICATION_H__)
#define __REMOTE_COMMUNICATION_H__


#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */

/**
 * TYPENAME: ix_remote_comm_data_handler
 * 
 * DESCRIPTION: This is the definition of the generic remote data handler function 
 *	pointer. The data handler function takes as parameters a buffer handle 
 *	which in general refers to a packet or message handle, a 32 bit 
 *	unsigned integer where extra user data is passed to the handler, a remote
 *  communication Id and a context pointer that was passed to Resource Manager 
 *  at the time of remote communication set-up. When data is sent to that
 *  remote communication ID through send_packet or send_message functions, 
 *  the registered remote handler will be invoked. The send_packet and send_message 
 *	have an arg_UserData parameter that will passed as arg_UserData 
 *	argument.
 */
typedef ix_error (*ix_remote_comm_data_handler)( 
                                                 ix_buffer_handle arg_hBuffer, 
                                                 ix_uint32 arg_UserData,
                                                 void* arg_pContext, 
                                                 ix_communication_id arg_CommId
                                               );


/**
 * TYPENAME: ix_remote_comm_service_initializer
 * 
 * DESCRIPTION: This is the definition of the remote communication service initializer.
 *  The function takes as an argument a context. The context passed to the function is
 *  the one passed to Resource Manager at the time of the remote communication set-up.
 *
 */
typedef ix_error (*ix_remote_comm_service_initializer)(
                                                        void* arg_pContext
                                                      );


/**
 * TYPENAME: ix_remote_comm_service_finalizer
 * 
 * DESCRIPTION: This is the definition of the remote communication service finalizer.
 *  The function takes as an argument a context. The context passed to the function is
 *  the one passed to Resource Manager at the time of the remote communication set-up.
 *
 */
typedef ix_error (*ix_remote_comm_service_finalizer)(
                                                      void* arg_pContext
                                                    );



/**
 * TYPENAME: ix_remote_comm_service
 * 
 * DESCRIPTION: This structure defines a remote communication service.
 *  RM will support just one remote communication service at a time.
 *  However, the programmer can install one handler that can demultiplex
 *  the processing for different destinations.
 *  The contexts should be valid for as long this service is registered 
 *  with RM.
 */
typedef struct ix_s_remote_comm_service
{
    /* Will be called at the initialization of the service */
    ix_remote_comm_service_initializer  m_ServiceInitializer;

    /* Will be called at the finalization of the service */
    ix_remote_comm_service_finalizer    m_ServiceFinalizer;

    /* Will be passed as context to the service initializer and finalizer */
    void*                               m_pServiceContext;

    /* Will be invoked for each packet sent to a remote target */
    ix_remote_comm_data_handler         m_PacketHandler;
    /* Will be passed as context to the packet handler at each invocation */
    void*                               m_pPacketHandlerContext;

    /* Will be invoked for each message sent to a remote target */
    ix_remote_comm_data_handler         m_MessageHandler;
    /* Will be passed as context to the message handler at each invocation */
    void*                               m_pMessageHandlerContext;

} ix_remote_comm_service;



/**
 * NAME: ix_rm_remote_comm_service_register
 *
 * DESCRIPTION: This function will register a remote communication service if no other is
 *  already installed. The one in place should be first unregistered, and after that try
 *  to register a new one. If the registration is possible then the initializer of the new
 *  service is invoked, and if that call is successful, then the registration will be completed.
 * 
 * @Param:  - IN arg_pRemoteCommService - the new remote comm service to be installed.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_remote_comm_service_register(
                                             const ix_remote_comm_service* arg_pRemoteCommService
                                           );




/**
 * NAME: ix_rm_remote_comm_service_unregister
 *
 * DESCRIPTION: This function will unregister the currently registered remote communication
 *  service. At the time of this call the finalizer of the existing remote comm service will
 *  be invoked.
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_remote_comm_service_unregister(void);



/**
 * NAME: ix_rm_init_pci_remote_communication
 *
 * DESCRIPTION: This function will install the PCI remote communication service 
 *          between ingress and egress
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_init_pci_remote_communication(void);



/**
 * NAME: ix_rm_register_pci_communication_hw_free_list
 *
 * DESCRIPTION: This function will register a hardware free list to be used by the receive
 *      side of the PCI remote communication driver instead of the default hardware free
 *      list. The PCI driver automatically creates an internal free list that is used for 
 *      copying the buffers from the remote subsystem. This free list can be replaced by
 *      a free list of choice.
 * 
 * @Param:  - IN arg_hFreeList - the hardware free list to be used by the PCI remote 
 *      communication driver.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_register_pci_communication_hw_free_list(
                                                        ix_buffer_free_list_handle arg_hFreeList
                                                      );




/**
 * NAME: ix_rm_unregister_pci_communication_hw_free_list
 *
 * DESCRIPTION: This function will unregister the previously registered user hardware
 *      free list for the PCI remote communication driver. From this point on the default
 *      (PCI driver internal one) hardware free list will be used to allocate the required
 *      buffers.
 * 
 * @Param:  - OUT  arg_pFreeListHandle - if successful the previously registered user
 *      hardware free list handle will be returned to the caller at the passed address.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_unregister_pci_communication_hw_free_list(
                                                          ix_buffer_free_list_handle* arg_pFreeListHandle
                                                        );



#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__REMOTE_COMMUNICATION_H__) */
