/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      atomic_operations.c
 *
 * = DESCRIPTION
 *      This file will provide implementation for all atomic operations functions.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      5/24/2002 10:52:51 AM - creation time 
 * ============================================================================
 */

static const char s_FileIdentifier[] = "$Id: atomic_operations.c,v 1.15 2003/08/08 20:25:06 ageorges Exp $";

#include <halMmap.h>

#include <ix_ossl.h>
#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"

#include "rm/symbols.h"
#include "rm/internal/internal_support.h"

#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"

#include "rm/memory.h"
#include "rm/internal/internal_memory.h"

#include "rm/atomic_operations.h"


/**
 * SRAM atomic operations
 */

/**
 * NAME: ix_rm_atomic_sram_swap
 *
 * DESCRIPTION: This function will perform an sram memory atomic swap.The
 *  value referred by arg_pValue1 will be swapped with the value referred
 *  by arg_pValue2. On return *arg_pValue1 will be changed to previous *arg_pValue2
 *  and vice-versa. The operation is atomic only relative to the second argument 
 *  memory location and that only if that points to a SRAM memory location. So,
 *  in order for the operation to be atomic, arg_pValue2 must point to a SRAM location.
 * 
 * @Param:  - INOUT arg_pValue1 - the address of the first operand of the swap operation
 * @Param:  - INOUT arg_pValue2 - the address of the second operand of the swap operation.
 *      This is the operand relative to which the operation is atomic.
 */
void ix_rm_atomic_sram_swap(
                             ix_uint32* arg_pValue1, 
                             ix_uint32* arg_pValue2 
                           )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_atomic_sram_swap";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

#if defined(__GNUC__) 

    {
        /* ARM uses the SWP instruction to
         * atomically read and write a single
         * integer cell in memory.
         */
        __asm__("swp	%0, %1, [%2]"
	    : "=&r"(*arg_pValue1)	/* %0: data coming back from [%2] */
	    : "r"(*arg_pValue1)	/* %1: data to be written into [%2] */
	    , "r"(arg_pValue2)	/* %2: memory address to load from */
	    );
    }

#else

#error "Implement the atomic SRAM swap operation"

#endif /* defined(__GNUC__)  */

#else

    {
        ix_uint32 temp = *arg_pValue2;
        *arg_pValue2 = *arg_pValue1;
        *arg_pValue1 = temp;
    }

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

}


/**
 * NAME: ix_rm_atomic_sram_add
 *
 * DESCRIPTION: This function will increment the value addressed by arg_pSramValue by arg_IncrementValue.
 *  The arg_pSramValue address should be an SRAM address, as it will not be checked.
 *  The value at arg_pSramValue saturates at 0. So if you add a negative number that is larger
 *  in absolute value than the value at arg_pSramValue then the result at the arg_pSramValue location
 *  will be 0
 * 
 * @Param:  - IN arg_IncrementValue - the value to be added to the memory location addressed by
 *  arg_pSramValue;
 * @Param:  - INOUT arg_pSramValue - the value at this location will be incremented by arg_IncrementValue
 *
 */
void ix_rm_atomic_sram_add( 
                            ix_int32 arg_IncrementValue, 
                            ix_uint32* arg_pSramValue
                          )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_atomic_sram_add";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) || (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    {
        ix_uint32 channel;
        ix_uint32 offset;
        ix_rm_get_phys_offset((const void*)arg_pSramValue, 0, &channel, &offset, 0);
        switch(channel)
        {
            case 0:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch0_bit_add_virtAddr + offset);
                break;
            case 1:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch1_bit_add_virtAddr + offset);
                break;

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

            case 2:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch2_bit_add_virtAddr + offset);
                break;
            case 3:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch3_bit_add_virtAddr + offset);
                break;

#endif /* end #if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */

           default:
               break;

        }
        *arg_pSramValue = (ix_uint32)arg_IncrementValue;
    }

#else

#error "Implement SRAM atomic add"

#endif

#else

    if((arg_IncrementValue < 0) && (*arg_pSramValue < (ix_uint32)(-arg_IncrementValue)))
        *arg_pSramValue = 0;
    else
        *arg_pSramValue += arg_IncrementValue;

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

}


/**
 * NAME: ix_rm_atomic_sram_test_and_add
 *
 * DESCRIPTION: This function will increment the value addressed by arg_pSramValue by arg_IncrementValue
 *  and will return the value at arg_pSramValue location before increment operation.
 *  The arg_pSramValue address should be an SRAM address, as it will not be checked.
 *  The value at arg_pSramValue saturates at 0. So if you add a negative number that is larger
 *  in absolute value than the value at arg_pSramValue then the result at the arg_pSramValue location
 *  will be 0
 *
 * @Param:  - IN arg_IncrementValue - the value to be added to the memory location addressed by
 *  arg_pSramValue;
 * @Param:  - INOUT arg_pSramValue - the value at this location will be incremented by arg_IncrementValue
 *
 * @Return: Returns the value before the atomic add has been performed.
 */
ix_uint32 ix_rm_atomic_sram_test_and_add(
                                          ix_int32 arg_IncrementValue,
                                          ix_uint32* arg_pSramValue
                                        )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_atomic_sram_test_and_add";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) || (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    {
        ix_uint32 channel;
        ix_uint32 offset;
        ix_rm_get_phys_offset((const void*)arg_pSramValue, 0, &channel, &offset, 0);
        switch(channel)
        {
            case 0:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch0_bit_add_virtAddr + offset);
                break;
            case 1:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch1_bit_add_virtAddr + offset);
                break;

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

            case 2:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch2_bit_add_virtAddr + offset);
                break;
            case 3:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch3_bit_add_virtAddr + offset);
                break;

#endif /* end #if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */

           default:
               break;

        }

        ix_rm_atomic_sram_swap((ix_uint32*)&arg_IncrementValue, arg_pSramValue);

    }

#else

#error "Implement SRAM atomic test and add"

#endif

#else

    {
        ix_int32 temp = arg_IncrementValue;
        arg_IncrementValue = *arg_pSramValue;
        if((temp < 0) && (*arg_pSramValue < (ix_uint32)(-temp)))
            *arg_pSramValue = 0;
        else
            *arg_pSramValue += temp;
    }

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

    return (ix_uint32)arg_IncrementValue;
}


/**
 * NAME: ix_rm_atomic_sram_subtract
 *
 * DESCRIPTION: This function will decrement the value addressed by arg_pSramValue by arg_DecrementValue.
 *  The arg_pSramValue address should be an SRAM address, as it will not be checked.
 *  The value at arg_pSramValue saturates at 0. So if you subtract a positive number that is larger
 *  in absolute value than the value at arg_pSramValue then the result at the arg_pSramValue location
 *  will be 0
 *
 * @Param:  - IN arg_DecrementValue - the value to be subtracted from the memory location addressed by
 *  arg_pSramValue;
 * @Param:  - INOUT arg_pSramValue - the value at this location will be decremented by arg_DecrementValue
 *
 */
void ix_rm_atomic_sram_subtract(
                                  ix_int32 arg_DecrementValue,
                                  ix_uint32* arg_pSramValue
                                )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_atomic_sram_subtract";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) || (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    {
        ix_uint32 channel;
        ix_uint32 offset;
        ix_rm_get_phys_offset((const void*)arg_pSramValue, 0, &channel, &offset, 0);
        switch(channel)
        {
            case 0:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch0_bit_add_virtAddr + offset);
                break;
            case 1:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch1_bit_add_virtAddr + offset);
                break;

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

            case 2:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch2_bit_add_virtAddr + offset);
                break;
            case 3:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch3_bit_add_virtAddr + offset);
                break;

#endif /* end #if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */

           default:
               break;

        }

        *arg_pSramValue = (ix_uint32)(-arg_DecrementValue);

    }

#else

#error "Implement SRAM atomic subtract"

#endif

#else

    if((arg_DecrementValue > 0) && (*arg_pSramValue < (ix_uint32)arg_DecrementValue))
        *arg_pSramValue = 0;
    else
        *arg_pSramValue -= arg_DecrementValue;

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

}


/**
 * NAME: ix_rm_atomic_sram_test_and_subtract
 *
 * DESCRIPTION: This function will decrement the value addressed by arg_pSramValue by arg_DecrementValue
 *  and will return the value at arg_pSramValue location before decrement operation.
 *  The arg_pSramValue address should be an SRAM address, as it will not be checked.
 *  The value at arg_pSramValue saturates at 0. So if you subtract a positive number that is larger
 *  in absolute value than the value at arg_pSramValue then the result at the arg_pSramValue location
 *  will be 0
 *
 * @Param:  - IN arg_DecrementValue - the value to be subtracted from the memory location addressed by
 *  arg_pSramValue;
 * @Param:  - INOUT arg_pSramValue - the value at this location will be decremented by arg_DecrementValue
 *
 * @Return: Returns the value before the atomic subtract has been performed.
 */
ix_uint32 ix_rm_atomic_sram_test_and_subtract(
                                               ix_int32 arg_DecrementValue,
                                               ix_uint32* arg_pSramValue
                                             )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_atomic_sram_test_and_subtract";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

    arg_DecrementValue = (-arg_DecrementValue);

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) || (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    {
        ix_uint32 channel;
        ix_uint32 offset;
        ix_rm_get_phys_offset((const void*)arg_pSramValue, 0, &channel, &offset, 0);
        switch(channel)
        {
            case 0:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch0_bit_add_virtAddr + offset);
                break;
            case 1:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch1_bit_add_virtAddr + offset);
                break;

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

            case 2:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch2_bit_add_virtAddr + offset);
                break;
            case 3:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch3_bit_add_virtAddr + offset);
                break;

#endif /* end #if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */

           default:
               break;

        }

        ix_rm_atomic_sram_swap((ix_uint32*)&arg_DecrementValue, arg_pSramValue);

    }

#else

#error "Implement SRAM atomic test and subtract"

#endif

#else

    {
        ix_int32 temp = arg_DecrementValue;
        arg_DecrementValue = *arg_pSramValue;
        if((temp > 0) && (*arg_pSramValue < (ix_uint32)temp))
            *arg_pSramValue = 0;
        else
            *arg_pSramValue -= temp;
    }

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

    return (ix_uint32)arg_DecrementValue;
}



/**
 * NAME: ix_rm_atomic_sram_bit_set
 *
 * DESCRIPTION: This function will atomically set the bits of the value addressed 
 *  by arg_pSramValue corresponding to the bits set (1) in the arg_SetValue argument.
 *  The arg_pSramValue address must be a SRAM address, as it will not be checked.
 * 
 * @Param:  - IN arg_SetValue - the value specifying which bits of the memory location 
 *  addressed by arg_pSramValue should be set
 * @Param:  - INOUT arg_pSramValue - the value at this location will have the bits
 *  specified by arg_SetValue argument set upon return
 *
 */
void ix_rm_atomic_sram_bit_set( 
                                ix_uint32 arg_SetValue, 
                                ix_uint32* arg_pSramValue
                              )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_atomic_sram_bit_set";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) || (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    {
        ix_uint32 channel;
        ix_uint32 offset;
        ix_rm_get_phys_offset((const void*)arg_pSramValue, 0, &channel, &offset, 0);
        switch(channel)
        {
            case 0:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch0_bit_set_virtAddr + offset);
                break;
            case 1:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch1_bit_set_virtAddr + offset);
                break;

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

            case 2:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch2_bit_set_virtAddr + offset);
                break;
            case 3:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch3_bit_set_virtAddr + offset);
                break;

#endif /* end #if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */

           default:
               break;

        }
        *arg_pSramValue = arg_SetValue;
    }

#else

#error "Implement SRAM atomic bit set"

#endif

#else

    *arg_pSramValue |= arg_SetValue;

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

}


/**
 * NAME: ix_rm_atomic_sram_bit_test_and_set
 *
 * DESCRIPTION: This function will atomically set the bits of the value addressed 
 *  by arg_pSramValue corresponding to the bits set (1) in the arg_SetValue argument.
 *  The arg_pSramValue address must be a SRAM address, as it will not be checked.
 *  Upon return, the function will return the previous value at the arg_pSramValue
 *  location
 * 
 * @Param:  - IN arg_SetValue - the value specifying which bits of the memory location 
 *  addressed by arg_pSramValue should be set
 * @Param:  - INOUT arg_pSramValue - the value at this location will have the bits
 *  specified by arg_SetValue argument set upon return
 *
 * @Return: Returns the value at the arg_pSramValue location before the atomic set 
 *  has been performed.
 */
ix_uint32 ix_rm_atomic_sram_bit_test_and_set( 
                                              ix_uint32 arg_SetValue, 
                                              ix_uint32* arg_pSramValue
                                            )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_atomic_sram_bit_test_and_set";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) || (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    {
        ix_uint32 channel;
        ix_uint32 offset;
        ix_rm_get_phys_offset((const void*)arg_pSramValue, 0, &channel, &offset, 0);
        switch(channel)
        {
            case 0:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch0_bit_set_virtAddr + offset);
                break;
            case 1:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch1_bit_set_virtAddr + offset);
                break;

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

            case 2:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch2_bit_set_virtAddr + offset);
                break;
            case 3:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch3_bit_set_virtAddr + offset);
                break;

#endif /* end #if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */

           default:
               break;

        }

        ix_rm_atomic_sram_swap(&arg_SetValue, arg_pSramValue);

    }

#else

#error "Implement SRAM atomic bit test and set"

#endif

#else

    {
        ix_uint32 temp = arg_SetValue;
        arg_SetValue = *arg_pSramValue;
        *arg_pSramValue |= temp;
    }

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

    return arg_SetValue;
}


/**
 * NAME: ix_rm_atomic_sram_bit_clear
 *
 * DESCRIPTION: This function will atomically clear the bits of the value addressed 
 *  by arg_pSramValue corresponding to the bits set (1) in the arg_ClearValue argument.
 *  The arg_pSramValue address must be a SRAM address, as it will not be checked.
 * 
 * @Param:  - IN arg_ClearValue - the value specifying which bits of the memory location 
 *  addressed by arg_pSramValue should be cleared
 * @Param:  - INOUT arg_pSramValue - the value at this location will have the bits
 *  specified by arg_ClearValue argument cleared upon return
 *
 */
void ix_rm_atomic_sram_bit_clear( 
                                  ix_uint32 arg_ClearValue, 
                                  ix_uint32* arg_pSramValue
                                )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_atomic_sram_bit_clear";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) || (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    {
        ix_uint32 channel;
        ix_uint32 offset;
        ix_rm_get_phys_offset((const void*)arg_pSramValue, 0, &channel, &offset, 0);
        switch(channel)
        {
            case 0:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch0_bit_clr_virtAddr + offset);
                break;
            case 1:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch1_bit_clr_virtAddr + offset);
                break;

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

            case 2:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch2_bit_clr_virtAddr + offset);
                break;
            case 3:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch3_bit_clr_virtAddr + offset);
                break;

#endif /* end #if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */

           default:
               break;

        }
        *arg_pSramValue = arg_ClearValue;
    }

#else

#error "Implement SRAM atomic bit clear"

#endif

#else

    *arg_pSramValue &= ~arg_ClearValue;

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

}


/**
 * NAME: ix_rm_atomic_sram_bit_test_and_clear
 *
 * DESCRIPTION: This function will atomically clear the bits of the value addressed 
 *  by arg_pSramValue corresponding to the bits set (1) in the arg_ClearValue argument.
 *  The arg_pSramValue address must be a SRAM address, as it will not be checked.
 *  Upon return, the function will return the previous value at the arg_pSramValue
 *  location
 * 
 * @Param:  - IN arg_ClearValue - the value specifying which bits of the memory location 
 *  addressed by arg_pSramValue should be cleared
 * @Param:  - INOUT arg_pSramValue - the value at this location will have the bits
 *  specified by arg_ClearValue argument cleared upon return
 *
 * @Return: Returns the value at the arg_pSramValue location before the atomic clear 
 *  has been performed.
 */
ix_uint32 ix_rm_atomic_sram_bit_test_and_clear( 
                                                ix_uint32 arg_ClearValue, 
                                                ix_uint32* arg_pSramValue
                                              )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_atomic_sram_bit_test_and_clear";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) || (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    {
        ix_uint32 channel;
        ix_uint32 offset;
        ix_rm_get_phys_offset((const void*)arg_pSramValue, 0, &channel, &offset, 0);
        switch(channel)
        {
            case 0:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch0_bit_clr_virtAddr + offset);
                break;
            case 1:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch1_bit_clr_virtAddr + offset);
                break;

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

            case 2:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch2_bit_clr_virtAddr + offset);
                break;
            case 3:
                arg_pSramValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_sram_ch3_bit_clr_virtAddr + offset);
                break;

#endif /* end #if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */

           default:
               break;

        }

        ix_rm_atomic_sram_swap(&arg_ClearValue, arg_pSramValue);

    }

#else

#error "Implement SRAM atomic bit test and clear"

#endif

#else

    {
        ix_uint32 temp = arg_ClearValue;
        arg_ClearValue = *arg_pSramValue;
        *arg_pSramValue &= ~temp;
    }

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

    return arg_ClearValue;
}



/**
 * SCRATCH atomic operations
 */

/**
 * NAME: ix_rm_atomic_scratch_swap
 *
 * DESCRIPTION: This function will perform an scratch memory atomic swap.The
 *  value referred by arg_pValue1 will be swapped with the value referred
 *  by arg_pValue2. On return *arg_pValue1 will be changed to previous *arg_pValue2
 *  and vice-versa. The operation is atomic only relative to the second argument 
 *  memory location and that only if that points to a SCRATCH memory location. So,
 *  in order for the operation to be atomic, arg_pValue2 must point to a SCRATCH location.
 * 
 * @Param:  - INOUT arg_pValue1 - the address of the first operand of the swap operation
 * @Param:  - INOUT arg_pValue2 - the address of the second operand of the swap operation
 *      This is the operand relative to which the operation is atomic.
 */
void ix_rm_atomic_scratch_swap(
                                ix_uint32* arg_pValue1, 
                                ix_uint32* arg_pValue2 
                              )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_atomic_scratch_swap";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

#if defined(__GNUC__)

    {
        /* ARM uses the SWP instruction to
         * atomically read and write a single
         * integer cell in memory.
         */
        __asm__("swp	%0, %1, [%2]"
	    : "=&r"(*arg_pValue1)	/* %0: data coming back from [%2] */
	    : "r"(*arg_pValue1)	/* %1: data to be written into [%2] */
	    , "r"(arg_pValue2)	/* %2: memory address to load from */
	    );
    }

#else

#error "Implement the atomic SCRATCH swap operation"

#endif /* defined(__GNUC__)  */

#else

    {
        ix_uint32 temp = *arg_pValue2;
        *arg_pValue2 = *arg_pValue1;
        *arg_pValue1 = temp;
    }

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

}



/**
 * NAME: ix_rm_atomic_scratch_add
 *
 * DESCRIPTION: This function will increment the value addressed by arg_pScratchValue by arg_IncrementValue.
 *  The arg_pScratchValue address should be a SCRATCH address, as it will not be checked.
 * 
 * @Param:  - IN arg_IncrementValue - the value to be added to the memory location addressed by
 *  arg_pScratchValue;
 * @Param:  - INOUT arg_pScratchValue - the value at this location will be incremented by arg_IncrementValue
 *
 */
void ix_rm_atomic_scratch_add( 
                               ix_uint32 arg_IncrementValue, 
                               ix_uint32* arg_pScratchValue
                             )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_atomic_scratch_add";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) || (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    {
        ix_uint32 offset;
        ix_rm_get_phys_offset((const void*)arg_pScratchValue, 0, 0, &offset, 0);

        arg_pScratchValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_scratch_add_virtAddr + offset);

        *arg_pScratchValue = arg_IncrementValue;
    }

#else

#error "Implement SCRATCH atomic add"

#endif

#else

    *arg_pScratchValue += arg_IncrementValue;

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

}



/**
 * NAME: ix_rm_atomic_scratch_test_and_add
 *
 * DESCRIPTION: This function will increment the value addressed by arg_pScratchValue by arg_IncrementValue
 *  and will return the value at arg_pScratchValue location before increment operation.
 *  The arg_pScratchValue address should be a SCRATCH address, as it will not be checked.
 * 
 * @Param:  - IN arg_IncrementValue - the value to be added to the memory location addressed by
 *  arg_pScratchValue;
 * @Param:  - INOUT arg_pScratchValue - the value at this location will be incremented by arg_IncrementValue
 *
 * @Return: Returns the value before the atomic add has been performed.
 */
ix_uint32 ix_rm_atomic_scratch_test_and_add( 
                                             ix_uint32 arg_IncrementValue, 
                                             ix_uint32* arg_pScratchValue
                                           )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_atomic_scratch_test_and_add";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) || (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    {
        ix_uint32 offset;
        ix_rm_get_phys_offset((const void*)arg_pScratchValue, 0, 0, &offset, 0);

        arg_pScratchValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_scratch_add_virtAddr + offset);

        ix_rm_atomic_scratch_swap(&arg_IncrementValue, arg_pScratchValue);
    }

#else

#error "Implement SCRATCH atomic test and add"

#endif

#else

    {
        ix_uint32 temp = arg_IncrementValue;
        arg_IncrementValue = *arg_pScratchValue;
        *arg_pScratchValue += temp;
    }

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

    return arg_IncrementValue;
}



/**
 * NAME: ix_rm_atomic_scratch_subtract
 *
 * DESCRIPTION: This function will decrement the value addressed by arg_pScratchValue by arg_DecrementValue.
 *  The arg_pScratchValue address should be a SCRATCH address, as it will not be checked.
 *  The value at arg_pScratchValue saturates at 0. So if you subtract a number that is larger
 *  than the value at arg_pScratchValue then the result at the arg_pScratchValue location
 *  will be 0
 * 
 * @Param:  - IN arg_DecrementValue - the value to be subtracted from the memory location addressed by
 *  arg_pScratchValue;
 * @Param:  - INOUT arg_pScratchValue - the value at this location will be decremented by arg_DecrementValue
 *
 */
void ix_rm_atomic_scratch_subtract( 
                                    ix_uint32 arg_DecrementValue, 
                                    ix_uint32* arg_pScratchValue
                                  )

{
    static const char s_aFunctionIdentifier[] = "ix_rm_atomic_scratch_subtract";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) || (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    {
        ix_uint32 offset;
        ix_rm_get_phys_offset((const void*)arg_pScratchValue, 0, 0, &offset, 0);

        arg_pScratchValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_scratch_sub_virtAddr + offset);

        *arg_pScratchValue = arg_DecrementValue;
    }

#else

#error "Implement SCRATCH atomic subtract"

#endif

#else

    if((*arg_pScratchValue < arg_DecrementValue))
        *arg_pScratchValue = 0;
    else
        *arg_pScratchValue -= arg_DecrementValue;

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

}



/**
 * NAME: ix_rm_atomic_scratch_test_and_subtract
 *
 * DESCRIPTION: This function will decrement the value addressed by arg_pScratchValue by arg_DecrementValue
 *  and will return the value at arg_pScratchValue location before decrement operation.
 *  The arg_pScratchValue address should be a SCRATCH address, as it will not be checked.
 *  The value at arg_pScratchValue saturates at 0. So if you subtract a number that is larger
 *  than the value at arg_pScratchValue then the result at the arg_pScratchValue location
 *  will be 0
 * 
 * @Param:  - IN arg_DecrementValue - the value to be subtracted from the memory location addressed by
 *  arg_pScratchValue;
 * @Param:  - INOUT arg_pScratchValue - the value at this location will be decremented by arg_DecrementValue
 *
 * @Return: Returns the value before the atomic subtract has been performed.
 */
ix_uint32 ix_rm_atomic_scratch_test_and_subtract( 
                                                  ix_uint32 arg_DecrementValue, 
                                                  ix_uint32* arg_pScratchValue
                                                )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_atomic_scratch_test_and_subtract";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) || (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    {
        ix_uint32 offset;
        ix_rm_get_phys_offset((const void*)arg_pScratchValue, 0, 0, &offset, 0);

        arg_pScratchValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_scratch_sub_virtAddr + offset);

        ix_rm_atomic_scratch_swap(&arg_DecrementValue, arg_pScratchValue);
    }

#else

#error "Implement SCRATCH atomic test and subtract"

#endif

#else

    {
        ix_uint32 temp = arg_DecrementValue;
        arg_DecrementValue = *arg_pScratchValue;
        if((*arg_pScratchValue < temp))
            *arg_pScratchValue = 0;
        else
            *arg_pScratchValue -= temp;
    }

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

    return arg_DecrementValue;

}



/**
 * NAME: ix_rm_atomic_scratch_bit_set
 *
 * DESCRIPTION: This function will atomically set the bits of the value addressed 
 *  by arg_pScratchValue corresponding to the bits set (1) in the arg_SetValue argument.
 *  The arg_pScratchValue address must be a SCRATCH address, as it will not be checked.
 * 
 * @Param:  - IN arg_SetValue - the value specifying which bits of the memory location 
 *  addressed by arg_pScratchValue should be set
 * @Param:  - INOUT arg_pScratchValue - the value at this location will have the bits
 *  specified by arg_SetValue argument set upon return
 *
 */
void ix_rm_atomic_scratch_bit_set( 
                                   ix_uint32 arg_SetValue, 
                                   ix_uint32* arg_pScratchValue
                                 )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_atomic_scratch_bit_set";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) || (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    {
        ix_uint32 offset;
        ix_rm_get_phys_offset((const void*)arg_pScratchValue, 0, 0, &offset, 0);

        arg_pScratchValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_scratch_bit_set_virtAddr + offset);

        *arg_pScratchValue = arg_SetValue;
    }

#else

#error "Implement SCRATCH atomic bit set"

#endif

#else

    *arg_pScratchValue |= arg_SetValue;

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

}



/**
 * NAME: ix_rm_atomic_scratch_bit_test_and_set
 *
 * DESCRIPTION: This function will atomically set the bits of the value addressed 
 *  by arg_pScratchValue corresponding to the bits set (1) in the arg_SetValue argument.
 *  The arg_pScratchValue address must be a SCRATCH address, as it will not be checked.
 *  Upon return, the function will return the previous value at the arg_pScratchValue
 *  location
 * 
 * @Param:  - IN arg_SetValue - the value specifying which bits of the memory location 
 *  addressed by arg_pScratchValue should be set
 * @Param:  - INOUT arg_pScratchValue - the value at this location will have the bits
 *  specified by arg_SetValue argument set upon return
 *
 * @Return: Returns the value at the arg_pScratchValue location before the atomic set 
 *  has been performed.
 */
ix_uint32 ix_rm_atomic_scratch_bit_test_and_set( 
                                                 ix_uint32 arg_SetValue, 
                                                 ix_uint32* arg_pScratchValue
                                               )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_atomic_scratch_bit_test_and_set";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) || (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    {
        ix_uint32 offset;
        ix_rm_get_phys_offset((const void*)arg_pScratchValue, 0, 0, &offset, 0);

        arg_pScratchValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_scratch_bit_set_virtAddr + offset);

        ix_rm_atomic_scratch_swap(&arg_SetValue, arg_pScratchValue);
    }

#else

#error "Implement SCRATCH atomic bit test and set"

#endif

#else

    {
        ix_uint32 temp = arg_SetValue;
        arg_SetValue = *arg_pScratchValue;
        *arg_pScratchValue |= temp;
    }

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

    return arg_SetValue;

}



/**
 * NAME: ix_rm_atomic_scratch_bit_clear
 *
 * DESCRIPTION: This function will atomically clear the bits of the value addressed 
 *  by arg_pScratchValue corresponding to the bits set (1) in the arg_ClearValue argument.
 *  The arg_pScratchValue address must be a SCRATCH address, as it will not be checked.
 * 
 * @Param:  - IN arg_ClearValue - the value specifying which bits of the memory location 
 *  addressed by arg_pScratchValue should be cleared
 * @Param:  - INOUT arg_pScratchValue - the value at this location will have the bits
 *  specified by arg_ClearValue argument cleared upon return
 *
 */
void ix_rm_atomic_scratch_bit_clear( 
                                     ix_uint32 arg_ClearValue, 
                                     ix_uint32* arg_pScratchValue
                                   )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_atomic_scratch_bit_clear";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) || (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    {
        ix_uint32 offset;
        ix_rm_get_phys_offset((const void*)arg_pScratchValue, 0, 0, &offset, 0);

        arg_pScratchValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_scratch_bit_clear_virtAddr + offset);

        *arg_pScratchValue = arg_ClearValue;
    }

#else

#error "Implement SCRATCH atomic bit clear"

#endif

#else

    *arg_pScratchValue &= ~arg_ClearValue;

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

}



/**
 * NAME: ix_rm_atomic_scratch_bit_test_and_clear
 *
 * DESCRIPTION: This function will atomically clear the bits of the value addressed 
 *  by arg_pScratchValue corresponding to the bits set (1) in the arg_ClearValue argument.
 *  The arg_pScratchValue address must be a SCRATCH address, as it will not be checked.
 *  Upon return, the function will return the previous value at the arg_pScratchValue
 *  location
 * 
 * @Param:  - IN arg_ClearValue - the value specifying which bits of the memory location 
 *  addressed by arg_pScratchValue should be cleared
 * @Param:  - INOUT arg_pScratchValue - the value at this location will have the bits
 *  specified by arg_ClearValue argument cleared upon return
 *
 * @Return: Returns the value at the arg_pScratchValue location before the atomic clear 
 *  has been performed.
 */
ix_uint32 ix_rm_atomic_scratch_bit_test_and_clear( 
                                                   ix_uint32 arg_ClearValue, 
                                                   ix_uint32* arg_pScratchValue
                                                 )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_atomic_scratch_bit_test_and_clear";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) || (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    {
        ix_uint32 offset;
        ix_rm_get_phys_offset((const void*)arg_pScratchValue, 0, 0, &offset, 0);

        arg_pScratchValue = (ix_uint32*)((ix_uint8*)(ix_uint32)Hal_scratch_bit_clear_virtAddr + offset);

        ix_rm_atomic_scratch_swap(&arg_ClearValue, arg_pScratchValue);
    }

#else

#error "Implement SCRATCH atomic bit test and clear"

#endif

#else

    {
        ix_uint32 temp = arg_ClearValue;
        arg_ClearValue = *arg_pScratchValue;
        *arg_pScratchValue &= ~temp;
    }

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

    return arg_ClearValue;
}
