/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      buffer.c
 *
 * = DESCRIPTION
 *      This file will define all functions the exported Buffer API interface.
 *      The Buffer API has two parts: 
 *          - an API that is used to create buffer free lists of varying sizes. This API is 
 *      quite generic and independent of packet descriptor layout defined by the microblock
 *      infrastructure.
 *          - an API that is used to access fields in the packet descriptor (meta data).
 *      This API is specific to layout and fields of packet descriptor.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      4/23/2002 10:17:05 AM - creation time 
 * ============================================================================
 */

static const char s_aFileIdentifier[] = "$Id: buffer.c,v 1.48 2003/12/11 23:08:20 ageorges Exp $";


#include "ix_ossl.h"

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"

#include "rm/symbols.h"
#include "rm/internal/internal_support.h"
#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"
#include "rm/system.h"
#include "rm/internal/internal_system.h"
#include "rm/memory.h"

#include "rm/buffer.h"
#include "rm/internal/internal_buffer.h"
#include "rm/hardware.h"
#include "rm/internal/internal_hardware.h"

#if defined(_IX_RM_IMPL_HARDWARE_) 

    #if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

        #include <vmLib.h>

    #endif /* end (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) */

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */



/**
 * NAME: ix_rm_hw_buffer_free_list_create
 *
 * DESCRIPTION: This function will create a hardware buffer free list with arg_ElementsNumber entries. 
 *          Memory will be allocated in SRAM and DRAM. A handle to the new created free list
 *          is returned. The size of SRAM required for the meta should be at least the size imposed
 *          by the hardware which at this moment is 4 bytes. The sizes will be rounded to 
 *          the next power of 2 always, so it is better that preferred sizes are passed in.
 *          Maximum IX_BUF_MAX_HW_FL_NUMBER hardware free lists can be created.
 * 
 * @Param:  - IN arg_ElementsNumber - the number of entries to be allocated for the free list.
 * @Param:  - IN arg_SRAMSize - the size of SRAM to be allocated for each entry.
 * @Param:  - IN arg_DRAMSize - the size of DRAM to be allocated for each entry.
 * @Param:  - IN arg_SRAMChannel - the SRAM channel where the memory should be allocated from.  
 * @Param:  - IN arg_DRAMChannel - the DRAM channel where the memory should be allocated from.  
 * @Param:  - OUT arg_pFreeListHandle - location where the newly created free list handle should
 *          be stored.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_hw_buffer_free_list_create( 
                                           ix_uint32 arg_ElementsNumber, 
                                           ix_uint32 arg_SRAMSize, 
                                           ix_uint32 arg_DRAMSize, 
                                           ix_uint32 arg_SRAMChannel, 
                                           ix_uint32 arg_DRAMChannel, 
                                           ix_buffer_free_list_handle* arg_pFreeListHandle 
                                         )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_hw_buffer_free_list_create";

    ix_error err = IX_SUCCESS;

    ix_uint32 shiftSize;
    ix_uint32 size;
    ix_uint32 i;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_FLManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */


    _ix_buffer_free_list* pFreeList = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_ElementsNumber == 0)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_BAD_SIZE, s_aFunctionIdentifier, 
                        "the free list number of elements should be non zero!", err);
    }

    *arg_pFreeListHandle = 0;
    
    if(arg_SRAMSize < sizeof(ix_hw_buffer_meta))
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_BAD_META_SIZE, s_aFunctionIdentifier, 
                        "the meta size is not large enough!", err);
    }

    /* make sure that the requested sizes fit into a 32 bit number */
    size = (arg_SRAMSize > arg_DRAMSize) ? arg_SRAMSize : arg_DRAMSize;
    if((0xffffffff / size) <= arg_ElementsNumber)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_SIZE_TOO_LARGE, s_aFunctionIdentifier, 
                        "the requested meta or data size is too large!", err);
    }


#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the free list manager data */
    if(pSyncAdapter->m_SyncActive)
    {
        pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);
    }

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

    {
        ix_uint32 j;

        /* reserve a free list entry into the free list manager */
        for( i = arg_SRAMChannel * IX_BUF_MAX_HW_FL_NUMBER, j = 0; j < IX_BUF_MAX_HW_FL_NUMBER; i++, j++)
        {
            if(g_FLManager.m_aFreeListAddress[i] == 0)
            {
                pFreeList = g_FLManager.m_aFreeListAddress[i] = &g_FLManager.m_aHwFreeList[arg_SRAMChannel][j];
                break;
            }
        }
    }

#else

    /* reserve a free list entry into the free list manager */
    for( i = 0; i < IX_BUF_MAX_HW_FL_NUMBER; i++)
    {
        if(g_FLManager.m_aFreeListAddress[i] == 0)
        {
            pFreeList = g_FLManager.m_aFreeListAddress[i] = &g_FLManager.m_aHwFreeList[i];
            break;
        }
    }

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */


#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the free list data */
    if(pSyncAdapter->m_SyncActive)
    {
        pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);
    }

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* there is no pool manager slot available */
    if(pFreeList == 0)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_ALL_IN_USE, s_aFunctionIdentifier, 
                        "all available hardware free lists in the system are in use!", err);
    }

    *arg_pFreeListHandle = IX_RM_BUF_FL_MAKE_HANDLE(i);

#if defined(_IX_RM_SPLIT_META_DATA_)

    /* get the meta element shift size */
    for(shiftSize = 0; sizeof(ix_hw_internal_buffer_meta) > (1U << shiftSize); shiftSize++)
    {}

    pFreeList->m_FLData.m_HwData.m_InternalMetaElementShift = shiftSize;
    pFreeList->m_FLData.m_HwData.m_InternalMetaMemoryType = IX_MEMORY_TYPE_SRAM;

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

    pFreeList->m_FLData.m_HwData.m_InternalMetaMemoryChannel = arg_SRAMChannel;

#else

#if (IX_HW_FL_META_CHANNEL == 0)

    pFreeList->m_FLData.m_HwData.m_InternalMetaMemoryChannel = 0;

#elif (IX_HW_FL_META_CHANNEL == 1)
                                  
    pFreeList->m_FLData.m_HwData.m_InternalMetaMemoryChannel = 1;

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

#if (IX_HW_FL_META_CHANNEL == 2)

    pFreeList->m_FLData.m_HwData.m_InternalMetaMemoryChannel = 2;

#elif (IX_HW_FL_META_CHANNEL == 3)
                          
    pFreeList->m_FLData.m_HwData.m_InternalMetaMemoryChannel = 3;

#endif /* end (IX_HW_FL_META_CHANNEL == 2) */

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */

#endif /* end (IX_HW_FL_META_CHANNEL == 0) */

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */


#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

    /* get the meta element shift size */
    for(shiftSize = 0; arg_SRAMSize > (1U << shiftSize); shiftSize++)
    {}

    pFreeList->m_MetaElementShift = shiftSize;
    pFreeList->m_MetaMemoryType = IX_MEMORY_TYPE_SRAM;
    pFreeList->m_MetaMemoryChannel = arg_SRAMChannel;

    /* get the data element shift size */
    for(shiftSize = 0; arg_DRAMSize > (1U << shiftSize); shiftSize++)
    {}

    pFreeList->m_DataElementShift = shiftSize;
    pFreeList->m_DataMemoryType = IX_MEMORY_TYPE_DRAM;
    pFreeList->m_DataMemoryChannel = arg_DRAMChannel;

    pFreeList->m_NumElements = arg_ElementsNumber;

    /* setup the VMT functions */
    pFreeList->m_VMT.m_FiniFuncPtr = _ix_rm_hw_free_list_fini;
    pFreeList->m_VMT.m_AllocatorFuncPtr = _ix_rm_hw_buffer_allocator;
    pFreeList->m_VMT.m_DeallocatorFuncPtr = _ix_rm_hw_buffer_deallocator;

#if defined(_IX_RM_SPLIT_META_DATA_)

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

    /* make sure that the meta data will be 32 byte aligned for each element */
    if(pFreeList->m_FLData.m_HwData.m_InternalMetaElementShift < 5)
        pFreeList->m_FLData.m_HwData.m_InternalMetaElementShift = 5;

#else

    /* make sure that the meta data will be 8 byte aligned for each element */
    if(pFreeList->m_FLData.m_HwData.m_InternalMetaElementShift < 3)
        pFreeList->m_FLData.m_HwData.m_InternalMetaElementShift = 3;

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */


    size = (1 << pFreeList->m_FLData.m_HwData.m_InternalMetaElementShift) * pFreeList->m_NumElements;

    /* allocate meta memory */
    err = ix_rm_mem_alloc_aligned(pFreeList->m_FLData.m_HwData.m_InternalMetaMemoryType, 
        pFreeList->m_FLData.m_HwData.m_InternalMetaMemoryChannel, 
        size, IX_MEMORY_ALIGNMENT_TYPE_PHYSICAL_OFFSET, 5, 
        ((void**)&pFreeList->m_FLData.m_HwData.m_pInternalMetaBaseAddress));

    /* check for allocation errors */
    if(err)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_INTERNAL_META_ALLOCATION_FAILED, s_aFunctionIdentifier, 
                        "internal meta block allocation failed!", err);
        goto label_4;
    }

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */


#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

    /* make sure that the meta data will be 32 byte aligned for each element */
    if(pFreeList->m_MetaElementShift < 5)
        pFreeList->m_MetaElementShift = 5;

#else

    /* make sure that the meta data will be 8 byte aligned for each element */
    if(pFreeList->m_MetaElementShift < 3)
        pFreeList->m_MetaElementShift = 3;

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */


    size = (1 << pFreeList->m_MetaElementShift) * pFreeList->m_NumElements;

    /* allocate meta memory */
    err = ix_rm_mem_alloc_aligned(pFreeList->m_MetaMemoryType, pFreeList->m_MetaMemoryChannel, 
        size, IX_MEMORY_ALIGNMENT_TYPE_PHYSICAL_OFFSET, 5, ((void**)&pFreeList->m_pMetaBaseAddress));

    /* check for allocation errors */
    if(err)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_META_ALLOCATION_FAILED, s_aFunctionIdentifier, 
                        "meta block allocation failed!", err);
        goto label_3;
    }

    /* make sure that the payload data will be 8 byte aligned for each element */
    if(pFreeList->m_DataElementShift < 3)
        pFreeList->m_DataElementShift = 3;

    size = (1 << pFreeList->m_DataElementShift) * pFreeList->m_NumElements;

    /* allocate the data memory */
    err = ix_rm_mem_alloc_aligned(pFreeList->m_DataMemoryType, pFreeList->m_DataMemoryChannel, 
        size, IX_MEMORY_ALIGNMENT_TYPE_PHYSICAL_OFFSET, 11, ((void**)&pFreeList->m_pDataBaseAddress));
    if(err)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_DATA_ALLOCATION_FAILED, s_aFunctionIdentifier, 
                        "data block allocation failed!", err);

        goto label_2;
    }

#if defined (_IX_RM_BUFFER_DEBUG_)

    err = ix_rm_mem_alloc(IX_MEMORY_TYPE_DRAM, 0, sizeof(ix_buffer_debug_info) * pFreeList->m_NumElements, ((void**)&pFreeList->m_pDebugInfo));
    if(err)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_DEBUG_INFO_ALLOCATION_FAILED, s_aFunctionIdentifier, 
                        "debug info block allocation failed!", err);

        goto label_1;
    }

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */


    /* build the memory buffer free list and initialize each buffer's meta */
    {
        /* declare some local variables */
        ix_hw_buffer_meta* pMeta = (ix_hw_buffer_meta*)pFreeList->m_pMetaBaseAddress;

#if defined(_IX_RM_SPLIT_META_DATA_)

        ix_hw_internal_buffer_meta* pInternalMeta = (ix_hw_internal_buffer_meta*)pFreeList->m_FLData.m_HwData.m_pInternalMetaBaseAddress;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

#if defined (_IX_RM_BUFFER_DEBUG_)

        ix_buffer_debug_info* pDebugInfo = pFreeList->m_pDebugInfo;

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */


        ix_uint32 metaLongWordOffset;

#if defined(_IX_RM_SPLIT_META_DATA_)

        ix_uint32 metaLongWordSize = 1 << (pFreeList->m_FLData.m_HwData.m_InternalMetaElementShift - 2);

#else

        ix_uint32 metaLongWordSize = 1 << (pFreeList->m_MetaElementShift - 2);

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

        ix_uint16 bufferInfo = 0;
        ix_buffer_handle currentHandle;

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

        ix_uint16 freeListID = pFreeList - (_ix_buffer_free_list*)g_FLManager.m_aHwFreeList;

#else

        ix_uint16 freeListID = pFreeList - g_FLManager.m_aHwFreeList;

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */

        ix_sram_queue_descriptor* pQueueDescriptor;

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

        pFreeList->m_FLData.m_HwData.m_SRAMQueueId = g_StartBufferFreeListQArrayID[arg_SRAMChannel] + (freeListID % IX_BUF_MAX_HW_FL_NUMBER);

        /* Now get the free list ID on the specified channel */
        freeListID %= IX_BUF_MAX_HW_FL_NUMBER;

#else

        pFreeList->m_FLData.m_HwData.m_SRAMQueueId = g_StartBufferFreeListQArrayID + freeListID;

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */


        size = (1 << pFreeList->m_MetaElementShift);

        /* get the start meta longword offset */
        {
            ix_memory_type memoryType;
            ix_uint32 channel;
            void* physicalAddress;

#if defined(_IX_RM_SPLIT_META_DATA_)

            /* get the start meta byte offset */
            ix_rm_get_phys_offset((void*)pFreeList->m_FLData.m_HwData.m_pInternalMetaBaseAddress, 
                          &memoryType, &channel, &metaLongWordOffset, &physicalAddress);

#else

            /* get the start meta byte offset */
            ix_rm_get_phys_offset((void*)pFreeList->m_pMetaBaseAddress, 
                          &memoryType, &channel, &metaLongWordOffset, &physicalAddress);

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */


            /* adjust the byte offset to longword offset */
            metaLongWordOffset >>= 2;
        }

        /* build first buffer handle */
        currentHandle = IX_RM_BUF_HW_MAKE_HANDLE(metaLongWordOffset);

        /* initialize the free list head */
        pFreeList->m_FreeListHead = currentHandle;

        /* initialize the value coresponding to m_BufferInfo field */
        IX_RM_HW_META_SET_FREE_LIST_ID(bufferInfo, freeListID);

        /* build the linked list the way the SRAM controller expects it */
        for(i = 1, metaLongWordOffset += metaLongWordSize; 
            i < pFreeList->m_NumElements; 
            i++, metaLongWordOffset += metaLongWordSize, 
            pMeta = (ix_hw_buffer_meta*)((ix_uint8*)pMeta + size))
        {

#if defined (_IX_RM_BUFFER_DEBUG_)

            /* setup the debug info for the buffer */
            pDebugInfo->m_LastCoreAllocationStatus = IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_FREE;
            pDebugInfo->m_SendAsPacketCount = 0;
            pDebugInfo->m_SendAsMessageCount = 0;
            pDebugInfo->m_hBuffer = currentHandle;

            pDebugInfo++;

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

            /* get the handle of the next cell */
            currentHandle = IX_RM_BUF_HW_MAKE_HANDLE(metaLongWordOffset);

#if defined(_IX_RM_SPLIT_META_DATA_)

            /* initialize this buffer's meta in the common ME/core memory */
            (void)IX_RM_MEM_UINT32_WRITE(&pInternalMeta->m_HwNext, currentHandle);

            /* initialize the address of the real meta */
            (void)IX_RM_MEM_UINT32_WRITE(&pInternalMeta->m_pMeta, pMeta);

#else

            /* initialize this buffer's meta in the common ME/core memory */
            (void)IX_RM_MEM_UINT32_WRITE(&pMeta->m_HwNext, currentHandle);

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */


            (void)IX_RM_MEM_UINT16_WRITE(&pMeta->m_BufferInfo, bufferInfo);

#if defined(_IX_RM_SPLIT_META_DATA_)

            pInternalMeta = (ix_hw_internal_buffer_meta*)((ix_uint32*)pInternalMeta + metaLongWordSize);

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

        }

#if defined(_IX_RM_SPLIT_META_DATA_)

        /* initialize the last buffers's meta in the common ME/core memory */
        IX_RM_MEM_UINT32_WRITE(&pInternalMeta->m_HwNext, 0);

        /* initialize the address of the real meta */
        (void)IX_RM_MEM_UINT32_WRITE(&pInternalMeta->m_pMeta, pMeta);

#else

        /* initialize the last buffers's meta in the common ME/core memory */
        IX_RM_MEM_UINT32_WRITE(&pMeta->m_HwNext, 0);

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

        (void)IX_RM_MEM_UINT16_WRITE(&pMeta->m_BufferInfo, bufferInfo);


#if defined (_IX_RM_BUFFER_DEBUG_)

        /* setup the debug info for the buffer */
        pDebugInfo->m_LastCoreAllocationStatus = IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_FREE;
        pDebugInfo->m_SendAsPacketCount = 0;
        pDebugInfo->m_SendAsMessageCount = 0;
        pDebugInfo->m_hBuffer = currentHandle;

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */


        /**
         * Setup the underlying SRAM queue.
         */

        /* initialize the QArray descriptor entry */

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

        {
            ix_uint32 channel = pFreeList->m_MetaMemoryChannel;

            pQueueDescriptor = &g_QArrayManager.m_aQArrayManagersPtr[channel]->
                m_aQArrayDescriptors[freeListID + g_StartBufferFreeListQArrayID[channel]].m_QueueDescriptor;

        }

#else

        pQueueDescriptor = &g_QArrayManager.m_aQArrayManagersPtr[pFreeList->m_MetaMemoryChannel]->
            m_aQArrayDescriptors[freeListID + g_StartBufferFreeListQArrayID].m_QueueDescriptor;

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */


        (void)IX_RM_MEM_UINT32_WRITE(&pQueueDescriptor->m_HeadInfo, pFreeList->m_FreeListHead);

        /* get the handle of the last cell in the list */
        metaLongWordOffset -= metaLongWordSize;
        currentHandle = IX_RM_BUF_HW_MAKE_HANDLE(metaLongWordOffset);
        (void)IX_RM_MEM_UINT32_WRITE(&pQueueDescriptor->m_TailInfo, currentHandle);

        (void)IX_RM_MEM_UINT32_WRITE(&pQueueDescriptor->m_CountInfo, pFreeList->m_NumElements);

        IX_RM_MEM_UINT32_WRITE(&pQueueDescriptor->m_Reserved, 0);
    }

    return err;

#if defined (_IX_RM_BUFFER_DEBUG_)

label_1:


    /* release the allocated data memory */
    ix_rm_mem_free((void*)pFreeList->m_pDataBaseAddress);

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

label_2:

    /* release the allocated meta memory */
    ix_rm_mem_free((void*)pFreeList->m_pMetaBaseAddress);

label_3:

#if defined(_IX_RM_SPLIT_META_DATA_)

    /* release the allocated internal meta memory */
    ix_rm_mem_free((void*)pFreeList->m_FLData.m_HwData.m_pInternalMetaBaseAddress);

label_4:

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

    /* release the free list entry due to unsuccessful initailization */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the free list manager data */
    if(pSyncAdapter->m_SyncActive)
    {
        pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);
    }

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */


    /* cleanup the free list entry */
    ix_ossl_memset(pFreeList,0,sizeof(_ix_buffer_free_list));

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

    g_FLManager.m_aFreeListAddress[pFreeList - (_ix_buffer_free_list*)g_FLManager.m_aHwFreeList] = 0;

#else

    g_FLManager.m_aFreeListAddress[pFreeList - g_FLManager.m_aHwFreeList] = 0;

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */


#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the free list data */
    if(pSyncAdapter->m_SyncActive)
    {
        pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);
    }

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: ix_rm_sw_buffer_free_list_create
 *
 * DESCRIPTION: This function will create a software buffer free list that will allocate and
 *          manage arg_NumElements elements. The caller has to specify the sizes of the meta
 *          and data parts of the buffer. The sizes will be rounded to the next power of
 *          2 always, so it is better that preferred sizes are passed in. The meta and data blocks
 *          can be allocated in a flexible way in different types of memory, but SRAM and especially
 *          SCRATCH memory should be used only if badly needed. The sizes of the meta and data can 
 *          not be 0 at the same time. Maximum IX_BUF_MAX_SW_FL_NUMBER software free lists can be 
 *          created.
 * 
 * @Param:  - IN arg_MetaElementSize - the size of a meta element. The size will be rounded
 *          to the next power of 2
 * @Param:  - IN arg_MetaMemoryType - the type of the meta memory. For now we will allow SRAM,
 *          DRAM and SCRATCH. SRAM and especially SCRATCH should be used with great care.
 * @Param:  - IN arg_DataElementSize - the size of a data element. The size will be rounded
 *          to the next power of 2
 * @Param:  - IN arg_DataMemoryType - the type of the data memory. For now we will allow SRAM,
 *          DRAM and SCRATCH. SRAM and especially SCRATCH should be used with great care. 
 * @Param:  - IN arg_NumElements - the number of memory-buffer elements that this manager will 
 *          allocate and manage. 
 * @Param:  - OUT arg_pPoolManagerHandle - address where the handle of the new created pool
 *          manager will be stored upon return.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_sw_buffer_free_list_create(
                                           ix_uint32 arg_MetaElementSize, 
                                           ix_memory_type arg_MetaMemoryType,
                                           ix_uint32 arg_DataElementSize, 
                                           ix_memory_type arg_DataMemoryType, 
                                           ix_uint32 arg_ElementsNumber,
                                           ix_buffer_free_list_handle* arg_pFreeListHandle 
                                         )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_sw_buffer_free_list_create";

    ix_error err = 0;

    ix_uint32 shiftSize;
    ix_memory_type memoryType;
    ix_uint32 channel;
    ix_uint32 size;
    ix_uint32 maxChannelNo;
    ix_uint32 i;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_FLManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */


    _ix_buffer_free_list* pFreeList = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier); 

    *arg_pFreeListHandle = 0;

    if(arg_ElementsNumber == 0)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_BAD_SIZE, s_aFunctionIdentifier, 
                        "the free list number of elements should be non zero!", err);
    }

    if(arg_MetaElementSize < sizeof(ix_sw_buffer_meta))
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_BAD_META_SIZE, s_aFunctionIdentifier, 
                        "the meta size is not large enough!", err);
    }

    /* make sure that the requested sizes fit into a 32 bit number */
    size = (arg_MetaElementSize > arg_DataElementSize) ? arg_MetaElementSize : arg_DataElementSize;
    if((0xffffffff / size) <= arg_ElementsNumber)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_SIZE_TOO_LARGE, s_aFunctionIdentifier, 
                        "the requested meta or data size is too large!", err);
    }

    /* reserve a pool manager entry */
#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the PM manager data */
    if(pSyncAdapter->m_SyncActive)
    {
        pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);
    }

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* reserve a pool manager entry into PM manager */
    for( i = IX_BUF_FIRST_SW_FL_ID; i < IX_BUF_MAX_FL_NUMBER; i++)
    {
        if(g_FLManager.m_aFreeListAddress[i] == 0)
        {
            pFreeList = g_FLManager.m_aFreeListAddress[i] = &g_FLManager.m_aSwFreeList[i - IX_BUF_FIRST_SW_FL_ID];
            break;
        }
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the PM manager data */
    if(pSyncAdapter->m_SyncActive)
    {
        pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);
    }

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* there is no pool manager slot available */
    if(pFreeList == 0)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_ALL_IN_USE, s_aFunctionIdentifier, 
                        "all available hardware free lists in the system are in use!", err);
    }

    *arg_pFreeListHandle = IX_RM_BUF_FL_MAKE_HANDLE(i);

    /* get the meta element shift size */
    for(shiftSize = 0, size = 1; arg_MetaElementSize > (1U << shiftSize); shiftSize++)
    {}

    pFreeList->m_MetaElementShift = shiftSize;
    pFreeList->m_MetaMemoryType = arg_MetaMemoryType;

    /* get the data element shift size */
    for(shiftSize = 0; arg_DataElementSize > (1U << shiftSize); shiftSize++)
    {}

    pFreeList->m_DataElementShift = shiftSize;
    pFreeList->m_DataMemoryType = arg_DataMemoryType;

    pFreeList->m_NumElements = arg_ElementsNumber;

    /* setup the VMT functions */
    pFreeList->m_VMT.m_FiniFuncPtr = _ix_rm_sw_free_list_fini;
    pFreeList->m_VMT.m_AllocatorFuncPtr = _ix_rm_sw_buffer_allocator;
    pFreeList->m_VMT.m_DeallocatorFuncPtr = _ix_rm_sw_buffer_deallocator;

    /* allocate the meta memory */
    memoryType = pFreeList->m_MetaMemoryType;
    switch(memoryType)
    {
        case IX_MEMORY_TYPE_DRAM:
            maxChannelNo = IX_DRAM_CHANNELS_NUMBER;
            break;
        case IX_MEMORY_TYPE_SRAM:
            maxChannelNo = IX_SRAM_CHANNELS_NUMBER;
            break;
        case IX_MEMORY_TYPE_SCRATCH:
            maxChannelNo = IX_SCRATCH_CHANNELS_NUMBER;
            break;
        default:
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_BAD_MEMORY_TYPE, s_aFunctionIdentifier, 
                        "bad memory type!", err);
            goto label_3;
    }
    /* make sure that the meta data will be 8 byte aligned for each element */
    if(pFreeList->m_MetaElementShift < 3)
        pFreeList->m_MetaElementShift = 3;

    size = (1 << pFreeList->m_MetaElementShift) * pFreeList->m_NumElements;

    /* check memory availability on each channel */
    for(channel = 0; channel < maxChannelNo; channel++)
    {
        err = ix_rm_mem_alloc_aligned(memoryType, channel, size, 
            IX_MEMORY_ALIGNMENT_TYPE_VIRTUAL, 3, ((void**)&pFreeList->m_pMetaBaseAddress));
        if(err == 0)
            break;
    }

    /* check for allocation errors */
    if(err)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_META_ALLOCATION_FAILED, s_aFunctionIdentifier, 
                        "meta block allocation failed!", err);
        goto label_3;
    }

    /* set the used meta data memory channel */
    pFreeList->m_MetaMemoryChannel = channel;

    /* set the meta block to 0 */
    for(i = 0; i < (size / sizeof(ix_uint32)); i++)
    {
        IX_RM_MEM_UINT32_WRITE((ix_uint32*)pFreeList->m_pMetaBaseAddress + i, 0);
    }

    /* allocate the data memory */
    memoryType = pFreeList->m_DataMemoryType;
    switch(memoryType)
    {
        case IX_MEMORY_TYPE_DRAM:
            maxChannelNo = IX_DRAM_CHANNELS_NUMBER;
            break;
        case IX_MEMORY_TYPE_SRAM:
            maxChannelNo = IX_SRAM_CHANNELS_NUMBER;
            break;
        case IX_MEMORY_TYPE_SCRATCH:
            maxChannelNo = IX_SCRATCH_CHANNELS_NUMBER;
            break;
        default:
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_BAD_MEMORY_TYPE, s_aFunctionIdentifier, 
                        "bad memory type!", err);
            goto label_2;
    }
    /* make sure that the payload data will be 8 byte aligned for each element */
    if(pFreeList->m_DataElementShift < 3)
        pFreeList->m_DataElementShift = 3;

    size = (1 << pFreeList->m_DataElementShift) * pFreeList->m_NumElements;

    /* check memory availability on each channel */
    for(channel = 0; channel < maxChannelNo; channel++)
    {
        err = ix_rm_mem_alloc_aligned(memoryType, channel, size, 
            IX_MEMORY_ALIGNMENT_TYPE_VIRTUAL, 3, ((void**)&pFreeList->m_pDataBaseAddress));
        if(err == 0)
            break;
    }

    /* check for allocation errors */
    if(err)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_DATA_ALLOCATION_FAILED, s_aFunctionIdentifier, 
                        "data block allocation failed!", err);

        goto label_2;
    }

    /* set the used data memory channel */
    pFreeList->m_DataMemoryChannel = channel;

    /* set the data block to 0 */
    for(i = 0; i < (size / sizeof(ix_uint32)); i++)
    {
        IX_RM_MEM_UINT32_WRITE((ix_uint32*)pFreeList->m_pDataBaseAddress + i, 0);
    }

#if defined (_IX_RM_BUFFER_DEBUG_)

    err = ix_rm_mem_alloc(IX_MEMORY_TYPE_DRAM, 0, sizeof(ix_buffer_debug_info) * pFreeList->m_NumElements, ((void**)&pFreeList->m_pDebugInfo));
    if(err)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_DEBUG_INFO_ALLOCATION_FAILED, s_aFunctionIdentifier, 
                        "debug info block allocation failed!", err);

        goto label_1;
    }

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

    /* build the memory cell free list and initialize each buffer's meta */
    {
        /* declare some local variables */
#if defined (_IX_RM_BUFFER_DEBUG_)

        ix_buffer_debug_info* pDebugInfo = pFreeList->m_pDebugInfo;

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

        ix_sw_buffer_meta* pMeta = (ix_sw_buffer_meta*)pFreeList->m_pMetaBaseAddress;
        ix_uint32* pData = pFreeList->m_pDataBaseAddress;
        ix_uint32 metaSize = 1 << pFreeList->m_MetaElementShift;
        ix_uint32 dataSize = 1 << pFreeList->m_DataElementShift;
        ix_uint32 flID = pFreeList - g_FLManager.m_aSwFreeList + IX_BUF_FIRST_SW_FL_ID;
        ix_buffer_handle currentHandle = IX_RM_BUF_SW_MAKE_HANDLE(flID,0);
        ix_buffer_handle nextHandle;

        /* initialize the free list head */
        pFreeList->m_FreeListHead = currentHandle;

        for(i = 0; i < pFreeList->m_NumElements - 1; i++)
        {
#if defined (_IX_RM_BUFFER_DEBUG_)

            /* setup the debug info for the buffer */
            pDebugInfo->m_LastCoreAllocationStatus = IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_FREE;
            pDebugInfo->m_SendAsPacketCount = 0;
            pDebugInfo->m_SendAsMessageCount = 0;
            pDebugInfo->m_hBuffer = currentHandle;

            pDebugInfo++;

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

            /* get the handle of the next cell */
            nextHandle = IX_RM_BUF_SW_MAKE_HANDLE(flID,i+1);

            /* initialize this cell's meta */
            pMeta->m_hNext = nextHandle;
            pMeta->m_hSelf = IX_NULL_BUFFER_HANDLE;
            pMeta->m_pData = pData;

            /* set the data size to the buffer data size for now */
            pMeta->m_DataSize = dataSize;

            /* update the data and meta pointers */
            pMeta = (ix_sw_buffer_meta*)((ix_uint8*)pMeta + metaSize);
            pData = (ix_uint32*)((ix_uint8*)pData + dataSize);

            /* set the current handle to the next handle */
            currentHandle = nextHandle;
        }

        /* initialize the last cell's meta */
        pMeta->m_hNext = IX_NULL_BUFFER_HANDLE;
        pMeta->m_hSelf = currentHandle;
        pMeta->m_pData = pData;

#if defined (_IX_RM_BUFFER_DEBUG_)

        /* setup the debug info for the buffer */
        pDebugInfo->m_LastCoreAllocationStatus = IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_FREE;
        pDebugInfo->m_SendAsPacketCount = 0;
        pDebugInfo->m_SendAsMessageCount = 0;
        pDebugInfo->m_hBuffer = currentHandle;

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

        /* set the data size to the buffer data size for now */
        pMeta->m_DataSize = dataSize;

        /* all elements are in the free list */
        pFreeList->m_FLData.m_SwData.m_NumFreeElements = pFreeList->m_NumElements;
    }

    return err;

#if defined (_IX_RM_BUFFER_DEBUG_)

label_1:

    /* release the allocated meta memory */
    ix_rm_mem_free((void*)pFreeList->m_pDataBaseAddress);

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

label_2:

    /* release the allocated meta memory */
    ix_rm_mem_free((void*)pFreeList->m_pMetaBaseAddress);

label_3:

    /* release the free list entry due to unsuccessful initailization */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the free list manager data */
    if(pSyncAdapter->m_SyncActive)
    {
        pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);
    }

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* cleanup the free list entry */
    ix_ossl_memset(pFreeList,0,sizeof(_ix_buffer_free_list));
    g_FLManager.m_aFreeListAddress[(pFreeList - g_FLManager.m_aSwFreeList) + IX_BUF_FIRST_SW_FL_ID] = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the free list data */
    if(pSyncAdapter->m_SyncActive)
    {
        pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);
    }

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}



/**
 * NAME: ix_rm_buffer_free_list_delete
 *
 * DESCRIPTION: This function will delete the specified buffer free list. The associated memory
 *          will be released.
 * 
 * @Param:  - IN arg_hFreeList - represents the handle of the free list that should 
 *          be deleted.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_buffer_free_list_delete( 
                                        ix_buffer_free_list_handle arg_hFreeList 
                                      )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_buffer_free_list_delete";

    _ix_buffer_free_list* pFreeList = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* retrieve the free list stucture address coresponding to the passed handle */

#if defined( _IX_RM_DEBUG_ )

    /* do extra checking in debug mode */
    IX_RM_ASSERT(IX_RM_BUF_FL_IS_VALID(arg_hFreeList) != 0, s_aFunctionIdentifier, 
        "the passed free list handle is invalid!");
    {
        ix_uint32 flId = IX_RM_BUF_FL_GET_FL_ID(arg_hFreeList);

        pFreeList = g_FLManager.m_aFreeListAddress[flId];

        if(pFreeList == 0)
        {
            return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_INVALID, s_aFunctionIdentifier, 
                        "the associated free list for this buffer is invalid!", 0);
        }

    }

#else

    pFreeList = g_FLManager.m_aFreeListAddress[IX_RM_BUF_FL_GET_FL_ID(arg_hFreeList)];

#endif /* end defined( _IX_RM_DEBUG_ ) */

    return pFreeList->m_VMT.m_FiniFuncPtr(pFreeList);
}


/**
 * NAME: ix_rm_buffer_free_list_get_info
 *
 * DESCRIPTION: This function will retrieve the buffer free list info. Once a buffer
 *          free list is created the caller has no control over where the memory required
 *          will be allocated. Later the caller might need the start addresses in 
 *          order to pass them to the microcode.
 * 
 * @Param:  - IN arg_hFreeList - specifies the free list handle for which we
 *          we want to retrieve the information.
 * @Param:  - OUT arg_pFreeListInfo - a pointer to a structure where the function
 *          will store the requested info on return.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_buffer_free_list_get_info( 
                                          ix_buffer_free_list_handle arg_hFreeList, 
                                          ix_buffer_free_list_info* arg_pFreeListInfo 
                                        )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_buffer_free_list_get_info";

    _ix_buffer_free_list* pFreeList = 0;
    ix_uint32 flId = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_DEBUG_ )

    /* do extra checking in debug mode */
    IX_RM_ASSERT(IX_RM_BUF_FL_IS_VALID(arg_hFreeList) != 0, s_aFunctionIdentifier, 
        "the passed free list handle is invalid!");

    flId = IX_RM_BUF_FL_GET_FL_ID(arg_hFreeList);

    pFreeList = g_FLManager.m_aFreeListAddress[flId];

    if(pFreeList == 0)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_INVALID, s_aFunctionIdentifier, 
                    "the associated free list for this buffer is invalid!", 0);
    }

#else

    flId = IX_RM_BUF_FL_GET_FL_ID(arg_hFreeList);
    pFreeList = g_FLManager.m_aFreeListAddress[flId];

#endif /* end defined( _IX_RM_DEBUG_ ) */

    arg_pFreeListInfo->m_DataElementSize = 1 << pFreeList->m_DataElementShift;
    arg_pFreeListInfo->m_DataMemoryChannel = pFreeList->m_DataMemoryChannel;
    arg_pFreeListInfo->m_DataMemoryType = pFreeList->m_DataMemoryType;
    arg_pFreeListInfo->m_FreeListType = (flId < IX_BUF_MAX_HW_FL_NUMBER) ? 0 : 1;
    arg_pFreeListInfo->m_MetaElementSize = 1 << pFreeList->m_MetaElementShift;
    arg_pFreeListInfo->m_MetaMemoryChannel = pFreeList->m_MetaMemoryChannel;
    arg_pFreeListInfo->m_MetaMemoryType = pFreeList->m_MetaMemoryType;
    arg_pFreeListInfo->m_NumElements = pFreeList->m_NumElements;
    arg_pFreeListInfo->m_pDataBaseAddress = pFreeList->m_pDataBaseAddress;
    arg_pFreeListInfo->m_pMetaBaseAddress = pFreeList->m_pMetaBaseAddress;
    if(arg_pFreeListInfo->m_FreeListType == 0)
    {
        /* hardware free list */

        /* get the SRAM controller QArray index */
        arg_pFreeListInfo->m_FreeListInfo = pFreeList->m_FLData.m_HwData.m_SRAMQueueId;

        /* get the free list id */

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

        arg_pFreeListInfo->m_FreeListInfo1 = arg_pFreeListInfo->m_FreeListInfo - g_StartBufferFreeListQArrayID[pFreeList->m_MetaMemoryChannel];

#else

        arg_pFreeListInfo->m_FreeListInfo1 = arg_pFreeListInfo->m_FreeListInfo - g_StartBufferFreeListQArrayID;

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */


#if defined(_IX_RM_SPLIT_META_DATA_)

        arg_pFreeListInfo->m_InternalMetaElementSize = 1 << pFreeList->m_FLData.m_HwData.m_InternalMetaElementShift;
        arg_pFreeListInfo->m_InternalMetaMemoryChannel = pFreeList->m_FLData.m_HwData.m_InternalMetaMemoryChannel;
        arg_pFreeListInfo->m_InternalMetaMemoryType = pFreeList->m_FLData.m_HwData.m_InternalMetaMemoryType;
        arg_pFreeListInfo->m_pInternalMetaBaseAddress = pFreeList->m_FLData.m_HwData.m_pInternalMetaBaseAddress;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

    }
    else
    {
        /* software free list */

        /* get the free list index */
        arg_pFreeListInfo->m_FreeListInfo = pFreeList - g_FLManager.m_aSwFreeList;

        /* set this one to 0 */
        arg_pFreeListInfo->m_FreeListInfo1 = 0;

#if defined(_IX_RM_SPLIT_META_DATA_)

        arg_pFreeListInfo->m_InternalMetaElementSize = 0;
        arg_pFreeListInfo->m_InternalMetaMemoryChannel = 0;
        arg_pFreeListInfo->m_InternalMetaMemoryType = 0;
        arg_pFreeListInfo->m_pInternalMetaBaseAddress = 0;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

    }

    return IX_SUCCESS;
}


/**
 * NAME: ix_rm_buffer_alloc
 *
 * DESCRIPTION: This function will allocate a buffer from the specified buffer free list.  
 * 
 * @Param:  - IN arg_hFreeList - handle of the free list from which the buffer will 
 *          be allocated. 
 * @Param:  - OUT arg_pBufferHandle - location where the new allocated buffer handle will 
 *          be stored. 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_buffer_alloc( 
                             ix_buffer_free_list_handle arg_hFreeList, 
                             ix_buffer_handle* arg_pBufferHandle 
                           )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_buffer_alloc";

    _ix_buffer_free_list* pFreeList = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_DEBUG_ )

    /* do extra checking in debug mode */
    IX_RM_ASSERT(IX_RM_BUF_FL_IS_VALID(arg_hFreeList) != 0, s_aFunctionIdentifier, 
        "the passed free list handle is invalid!");
    {
        ix_uint32 flId = IX_RM_BUF_FL_GET_FL_ID(arg_hFreeList);

        pFreeList = g_FLManager.m_aFreeListAddress[flId];

        if(pFreeList == 0)
        {
            return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_INVALID, s_aFunctionIdentifier, 
                        "the associated free list for this buffer is invalid!", 0);
        }

    }

#else

    pFreeList = g_FLManager.m_aFreeListAddress[IX_RM_BUF_FL_GET_FL_ID(arg_hFreeList)];

#endif /* end defined( _IX_RM_DEBUG_ ) */

    return pFreeList->m_VMT.m_AllocatorFuncPtr(pFreeList, arg_pBufferHandle);
}


/**
 * NAME: ix_rm_buffer_free
 *
 * DESCRIPTION: This function will free the specified buffer returning it to the right 
 *  buffer free list. Be aware that if the buffer is the start of a buffer chain,
 *  then the rest of the chain will be not freed and the chain will be broken.
 * 
 * @Param:  - IN arg_hBuffer - handle of the buffer to be freed.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_buffer_free( 
                            ix_buffer_handle arg_hBuffer 
                          )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_buffer_free";

    _ix_buffer_free_list* pFreeList = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(IX_RM_BUFFER_GET_TYPE(arg_hBuffer) == IX_BUFFER_TYPE_HARDWARE)
    {
        /* hardware buffer */

        if(arg_hBuffer == IX_NULL_BUFFER_HANDLE)
            return IX_SUCCESS;

#if defined( _IX_RM_DEBUG_ )

        {
            ix_uint32 flId = IX_RM_BUF_HW_GET_FL_ID(arg_hBuffer);

#if defined(_IX_RM_SPLIT_META_DATA_)

            void* pMeta = IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(arg_hBuffer);

#else

            void* pMeta = IX_RM_BUF_HW_GET_META_ADDRESS(arg_hBuffer);

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

            /* do extra checking in debug mode */

            pFreeList = g_FLManager.m_aFreeListAddress[flId];

            if(pFreeList == 0)
            {
                return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_INVALID, s_aFunctionIdentifier, 
                            "the associated free list for this buffer is invalid!", 0);
            }

            {

#if defined(_IX_RM_SPLIT_META_DATA_)

                _ix_hw_free_list_data* pInternal = &pFreeList->m_FLData.m_HwData;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */


#if defined(_IX_RM_SPLIT_META_DATA_)

                if((pMeta < (void*)pInternal->m_pInternalMetaBaseAddress) || (pMeta >= 
                    (void*)((ix_uint8*)pInternal->m_pInternalMetaBaseAddress + 
                    (pFreeList->m_NumElements << pInternal->m_InternalMetaElementShift))))

#else

                if((pMeta < (void*)pFreeList->m_pMetaBaseAddress) || (pMeta >= 
                    (void*)((ix_uint8*)pFreeList->m_pMetaBaseAddress + 
                    (pFreeList->m_NumElements << pFreeList->m_MetaElementShift))))

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

                {
                    return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_HANDLE_INVALID, s_aFunctionIdentifier, 
                            "the passed buffer is not in the range of the owning free list!", 0);

                }
            }

        }

#else

        pFreeList = g_FLManager.m_aFreeListAddress[IX_RM_BUF_HW_GET_FL_ID(arg_hBuffer)];

#endif /* end defined( _IX_RM_DEBUG_ ) */

    }
    else
    {
        /* software buffer */

#if defined( _IX_RM_DEBUG_ )

        /* do extra checking in debug mode */
        ix_uint32 flId = IX_RM_BUF_SW_GET_FL_ID(arg_hBuffer);

        /* check if the ID coresponds to a software free list */
        if(flId < IX_BUF_FIRST_SW_FL_ID)
        {
            return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_ID_INVALID, s_aFunctionIdentifier, 
                        "the free list ID coresponds to a hardware free list instead of a software one!", 0);
        }

        pFreeList = g_FLManager.m_aFreeListAddress[flId];

#else

        pFreeList = g_FLManager.m_aFreeListAddress[IX_RM_BUF_SW_GET_FL_ID(arg_hBuffer)];

#endif /* end defined( _IX_RM_DEBUG_ ) */

    }

    return pFreeList->m_VMT.m_DeallocatorFuncPtr(pFreeList, arg_hBuffer);
}


/**
 * NAME: ix_rm_buffer_free_chain
 *
 * DESCRIPTION: This function will free the specified buffer returning it to the right 
 *  buffer free list. If the buffer is the head of a buffer chain then the entire
 *  chain will be freed. If this buffer is in the middle of a chain then the end of the
 *  chain will be freed starting from this buffer. The chain has to be unlinked at this 
 *  buffer boundary in order to avoid chain inconsistency.
 * 
 * @Param:  - IN arg_hBuffer - handle of the buffer head of chain to be freed.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_buffer_free_chain( 
                                  ix_buffer_handle arg_hBuffer 
                                )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_buffer_free_chain";

    ix_error err = IX_SUCCESS;

    ix_buffer_handle hNextBuffer = IX_NULL_BUFFER_HANDLE;

#if defined( _IX_RM_DEBUG_ )

    ix_uint32 flId = 0;

#endif /* end defined( _IX_RM_DEBUG_ ) */

    void* pMeta = 0;
    _ix_buffer_free_list* pFreeList = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    while(arg_hBuffer != IX_NULL_BUFFER_HANDLE)
    {
        if(IX_RM_BUFFER_GET_TYPE(arg_hBuffer) == IX_BUFFER_TYPE_HARDWARE)
        {
            /* hardware buffer */

#if defined( _IX_RM_DEBUG_ )

            {
                pMeta = IX_RM_BUF_HW_GET_META_ADDRESS(arg_hBuffer);

                flId = IX_RM_BUF_HW_GET_FL_ID(arg_hBuffer);

                /* do extra checking in debug mode */

                pFreeList = g_FLManager.m_aFreeListAddress[flId];

                if(pFreeList == 0)
                {
                    return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_INVALID, s_aFunctionIdentifier, 
                                "the associated free list for this buffer is invalid!", 0);
                }

                if((pMeta < (void*)pFreeList->m_pMetaBaseAddress) || (pMeta > 
                    (void*)((ix_uint8*)pFreeList->m_pMetaBaseAddress + 
                    (pFreeList->m_NumElements << pFreeList->m_MetaElementShift))))
                {
                    return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_HANDLE_INVALID, s_aFunctionIdentifier, 
                            "the passed buffer is not in the range of the owning free list!", 0);

                }

            }

#else

            pFreeList = g_FLManager.m_aFreeListAddress[IX_RM_BUF_HW_GET_FL_ID(arg_hBuffer)];

#endif /* end defined( _IX_RM_DEBUG_ ) */

            if(IX_RM_BUF_GET_EOP(arg_hBuffer) != 0)
                hNextBuffer = IX_NULL_BUFFER_HANDLE;
            else
            {

#if defined(_IX_RM_SPLIT_META_DATA_)

                hNextBuffer = IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(arg_hBuffer)->m_HwNext;

#else

                hNextBuffer = IX_RM_BUF_HW_GET_META_ADDRESS(arg_hBuffer)->m_HwNext;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */
            }

        }
        else
        {
            /* software buffer */

#if defined( _IX_RM_DEBUG_ )

            /* do extra checking in debug mode */
            flId = IX_RM_BUF_SW_GET_FL_ID(arg_hBuffer);

            /* check if the ID coresponds to a software free list */
            if(flId < IX_BUF_FIRST_SW_FL_ID)
            {
                return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_ID_INVALID, s_aFunctionIdentifier, 
                            "the free list ID coresponds to a hardware free list instead of a software one!", 0);
            }

            pFreeList = g_FLManager.m_aFreeListAddress[flId];

#else

            pFreeList = g_FLManager.m_aFreeListAddress[IX_RM_BUF_SW_GET_FL_ID(arg_hBuffer)];

#endif /* end defined( _IX_RM_DEBUG_ ) */

            if(IX_RM_BUF_GET_EOP(arg_hBuffer) != 0)
                hNextBuffer = IX_NULL_BUFFER_HANDLE;
            else
            {
                pMeta = (void*)((ix_uint8*)pFreeList->m_pMetaBaseAddress + 
                    (IX_RM_BUF_SW_GET_INDEX(arg_hBuffer) << pFreeList->m_MetaElementShift));
                hNextBuffer = ((ix_sw_buffer_meta*)pMeta)->m_hNext;
            }

        }

        /* free the buffer */
        err = pFreeList->m_VMT.m_DeallocatorFuncPtr(pFreeList, arg_hBuffer);

        arg_hBuffer = hNextBuffer;

    }


    
    /**
     * This implementation might cause stack overflow for large chains
     */

    /*

    if(IX_RM_BUF_GET_EOP(arg_hBuffer) != 0)
    {
        err = ix_rm_buffer_free(arg_hBuffer);
    }
    else
    {
        err = ix_rm_buffer_get_next(arg_hBuffer, &hNextBuffer);
        if(err != IX_SUCCESS)
            return err;

        err = ix_rm_buffer_free(arg_hBuffer);
        if(err != IX_SUCCESS)
            return err;

        if(hNextBuffer != IX_NULL_BUFFER_HANDLE)
            err = ix_rm_buffer_free_chain(hNextBuffer);
    }

    */

    return err;
}



/**
 * NAME: ix_rm_buffer_get_meta
 *
 * DESCRIPTION: Returns the base of meta data portion for the specified buffer.  
 * 
 * @Param:  - IN arg_hBuffer - handle of the buffer for which the meta data is requested.
 * @Param:  - OUT arg_pMetaData - location where the meta data address should be written. 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_buffer_get_meta( 
                                ix_buffer_handle arg_hBuffer, 
                                void** arg_pMetaData 
                              )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_buffer_get_meta";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(IX_RM_BUFFER_GET_TYPE(arg_hBuffer) == IX_BUFFER_TYPE_HARDWARE)
    {
        if(arg_hBuffer == IX_NULL_BUFFER_HANDLE)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_HANDLE_INVALID, s_aFunctionIdentifier, 
                        "the passed buffer handle is invalid!", 0);

#if defined( _IX_RM_DEBUG_ )

        /* do extra checking in debug mode */
        {
            ix_uint32 flID = IX_RM_BUF_HW_GET_FL_ID(arg_hBuffer);
            _ix_buffer_free_list* pFreeList = g_FLManager.m_aFreeListAddress[flID];

            if(pFreeList == 0)
            {
                return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_INVALID, s_aFunctionIdentifier, 
                            "the associated free list for this buffer is invalid!", 0);
            }
        }

#endif /* end defined( _IX_RM_DEBUG_ ) */

        /* hardware buffer */
        *arg_pMetaData = IX_RM_BUF_HW_GET_META_ADDRESS(arg_hBuffer);
    }
    else
    {
        /* software buffer */

#if defined( _IX_RM_DEBUG_ )

        /* do extra checking in debug mode */
        _ix_buffer_free_list* pFreeList = 0;
        ix_uint32 flId = IX_RM_BUF_SW_GET_FL_ID(arg_hBuffer);

        /* check if the ID coresponds to a software free list */
        if(flId < IX_BUF_FIRST_SW_FL_ID)
        {
            return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_ID_INVALID, s_aFunctionIdentifier, 
                        "the free list ID coresponds to a hardware free list instead of a software one!", 0);
        }

        pFreeList = g_FLManager.m_aFreeListAddress[flId];

#else

        _ix_buffer_free_list* pFreeList = 
            g_FLManager.m_aFreeListAddress[IX_RM_BUF_SW_GET_FL_ID(arg_hBuffer)];

#endif /* end defined( _IX_RM_DEBUG_ ) */


        *arg_pMetaData = (void*)((ix_uint8*)pFreeList->m_pMetaBaseAddress + 
            (IX_RM_BUF_SW_GET_INDEX(arg_hBuffer) << pFreeList->m_MetaElementShift));

    }

    return IX_SUCCESS;
}


/**
 * NAME: ix_rm_buffer_get_data
 *
 * DESCRIPTION: Returns the base of data portion of the specified buffer.  
 * 
 * @Param:  - IN arg_hBuffer - handle of the buffer for which the data is requested.
 * @Param:  - OUT arg_pData - location where the data address should be written.  
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure. 
 */
ix_error ix_rm_buffer_get_data( 
                                ix_buffer_handle arg_hBuffer, 
                                void** arg_pData 
                              )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_buffer_get_data";

#if defined( _IX_RM_DEBUG_ )

    ix_uint32 flId = 0;

#endif /* end defined( _IX_RM_DEBUG_ ) */

    register _ix_buffer_free_list* pFreeList = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(IX_RM_BUFFER_GET_TYPE(arg_hBuffer) == IX_BUFFER_TYPE_HARDWARE)
    {
        if(arg_hBuffer == IX_NULL_BUFFER_HANDLE)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_HANDLE_INVALID, s_aFunctionIdentifier, 
                        "the the passed buffer handle is invalid!", 0);

#if defined( _IX_RM_DEBUG_ )

        /* do extra checking in debug mode */
        {
            pFreeList = g_FLManager.m_aFreeListAddress[IX_RM_BUF_HW_GET_FL_ID(arg_hBuffer)];

            if(pFreeList == 0)
            {
                return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_INVALID, s_aFunctionIdentifier, 
                            "the associated free list for this buffer is invalid!", 0);
            }
        }

#endif /* end defined( _IX_RM_DEBUG_ ) */

        /* hardware buffer */
        pFreeList = g_FLManager.m_aFreeListAddress[IX_RM_BUF_HW_GET_FL_ID(arg_hBuffer)];

#if defined(_IX_RM_SPLIT_META_DATA_)

        *arg_pData = (void*)((ix_uint8*)pFreeList->m_pDataBaseAddress + 
            + ((((ix_uint8*)IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(arg_hBuffer) - 
            (ix_uint8*)pFreeList->m_FLData.m_HwData.m_pInternalMetaBaseAddress ) >> 
            pFreeList->m_FLData.m_HwData.m_InternalMetaElementShift)
            << pFreeList->m_DataElementShift));


#else

        *arg_pData = (void*)((ix_uint8*)pFreeList->m_pDataBaseAddress + 
            + ((((ix_uint8*)IX_RM_BUF_HW_GET_META_ADDRESS(arg_hBuffer) - 
            (ix_uint8*)pFreeList->m_pMetaBaseAddress) >> pFreeList->m_MetaElementShift)
            << pFreeList->m_DataElementShift));

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

    }
    else
    {

#if defined( _IX_RM_DEBUG_ )

        /* software buffer */
        flId = IX_RM_BUF_SW_GET_FL_ID(arg_hBuffer);

        /* check if the ID coresponds to a software free list */
        if(flId < IX_BUF_FIRST_SW_FL_ID)
        {
            return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_ID_INVALID, s_aFunctionIdentifier, 
                        "the free list ID coresponds to a hardware free list instead of a software one!", 0);

        }

#else

#endif /* end defined( _IX_RM_DEBUG_ ) */

        pFreeList = g_FLManager.m_aFreeListAddress[IX_RM_BUF_SW_GET_FL_ID(arg_hBuffer)];

        *arg_pData = (void*)((ix_uint8*)pFreeList->m_pDataBaseAddress + 
            (IX_RM_BUF_SW_GET_INDEX(arg_hBuffer) << pFreeList->m_DataElementShift));

    }

    return IX_SUCCESS;
}


/**
 * NAME: ix_rm_buffer_is_eop
 *
 * DESCRIPTION: This function will return non 0 if this buffer is the last in a list of concatenated memory buffers.
 *          This is the equivalent of the End Of Packet property of a buffer.
 * 
 * @Param:  - IN arg_hBuffer - handle to a buffer whose EOP property we are interested in
 * @Param:  - OUT arg_pIsEOP - an address where the EOP property of the buffer will be stored. 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure. On return *arg_pIsEOP will be 0
 *          if the buffer is not the last in a chain or non 0 otherwise.
 */
ix_error ix_rm_buffer_is_eop( 
                              ix_buffer_handle arg_hBuffer,
                              ix_uint32* arg_pIsEOP
                            )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_buffer_is_eop";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_hBuffer == IX_NULL_BUFFER_HANDLE)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_HANDLE_INVALID, s_aFunctionIdentifier, 
                   "the the passed buffer handle is invalid!", 0);

    *arg_pIsEOP = IX_RM_BUF_GET_EOP(arg_hBuffer);

    return IX_SUCCESS;
}


/**
 * NAME: ix_rm_buffer_is_sop
 *
 * DESCRIPTION: This function will return non 0 if this buffer is the first in a list of concatenated memory buffers.
 *          This is the equivalent of the Start Of Packet property of a buffer.
 * 
 * @Param:  - IN arg_hBuffer - handle to a buffer whose SOP property we are interested in.
 * @Param:  - OUT arg_pIsSOP - an address where the SOP property of the buffer will be stored. 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure. On return *arg_pIsSOP will be 0
 *          if the buffer is not the first in a chain or non 0 otherwise.
 */
ix_error ix_rm_buffer_is_sop( 
                              ix_buffer_handle arg_hBuffer,
                              ix_uint32* arg_pIsSOP
                            )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_buffer_is_sop";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_hBuffer == IX_NULL_BUFFER_HANDLE)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_HANDLE_INVALID, s_aFunctionIdentifier, 
                   "the the passed buffer handle is invalid!", 0);

    *arg_pIsSOP = IX_RM_BUF_GET_SOP(arg_hBuffer);

    return IX_SUCCESS;
}


/**
 * NAME: ix_rm_buffer_cell_count_set
 *
 * DESCRIPTION: This function will set the cell count field into the
 *  passed hardware buffer handle
 * 
 * @Param:  - INOUT arg_pBufferHandle - address of the buffer handle to be modified
 * @Param:  - IN arg_CellCount - the new cell count, which can have values 0 to 63
 *  for now. Larger values will be truncated to 6bit values.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_buffer_cell_count_set(
                                      ix_buffer_handle* arg_pBufferHandle,
                                      ix_uint32 arg_CellCount
                                    )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_buffer_cell_count_set";

    ix_error err = IX_SUCCESS;
    register ix_buffer_handle hBuffer = *arg_pBufferHandle;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(IX_RM_BUFFER_GET_TYPE(hBuffer) == IX_BUFFER_TYPE_HARDWARE)
    {
        if(hBuffer == IX_NULL_BUFFER_HANDLE)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_HANDLE_INVALID, s_aFunctionIdentifier, 
                        "the the passed buffer handle is invalid!", 0);
        IX_SET_BIT_FIELD32(hBuffer, arg_CellCount, 24U, 29U);
        *arg_pBufferHandle = hBuffer;
    }
    else
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_HANDLE_INVALID, s_aFunctionIdentifier, 
                        "the the passed buffer handle coresponds to a software buffer!", 0);

    return err;
}


/**
 * NAME: ix_rm_buffer_cell_count_get
 *
 * DESCRIPTION: This function will retrieve the cell count field from the
 *  passed hardware buffer handle
 * 
 * @Param:  - IN arg_hBuffer - buffer handle of interest
 * @Param:  - OUT arg_pCellCount - the location where the cell count will be returned
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_buffer_cell_count_get(
                                      ix_buffer_handle arg_hBuffer,
                                      ix_uint32* arg_pCellCount
                                    )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_buffer_cell_count_get";

    ix_error err = IX_SUCCESS;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(IX_RM_BUFFER_GET_TYPE(arg_hBuffer) == IX_BUFFER_TYPE_HARDWARE)
    {
        if(arg_hBuffer == IX_NULL_BUFFER_HANDLE)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_HANDLE_INVALID, s_aFunctionIdentifier, 
                        "the the passed buffer handle is invalid!", 0);
        *arg_pCellCount = IX_GET_BIT_FIELD32(arg_hBuffer, 24U, 29U);
    }
    else
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_HANDLE_INVALID, s_aFunctionIdentifier, 
                        "the the passed buffer handle coresponds to a software buffer!", 0);

    return err;

    return err;
}



/**
 * NAME: ix_rm_buffer_get_type
 *
 * DESCRIPTION: This function will retrieve the type of the buffer whose handle has been passed
 *          as the first parameter.
 * 
 * @Param:  - IN arg_hBuffer - handle of the buffer whose type we want to retrieve
 * @Param:  - OUT arg_pBufferType - address where the buffer type will be stored on return. 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure. On return the *arg_pBufferType
 *          variable will have the type of the buffer.
 */
ix_error ix_rm_buffer_get_type(
                                ix_buffer_handle arg_hBuffer,
                                ix_buffer_type* arg_pBufferType
                              )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_buffer_get_type";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_hBuffer == IX_NULL_BUFFER_HANDLE)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_HANDLE_INVALID, s_aFunctionIdentifier, 
                   "the the passed buffer handle is invalid!", 0);

    *arg_pBufferType = IX_RM_BUFFER_GET_TYPE(arg_hBuffer);

    return IX_SUCCESS;
}




/**
 * NAME: ix_rm_buffer_get_next
 *
 * DESCRIPTION: This function will return the next link for this buffer. If there are no buffers linked
 *          to this one the arg_pNextBufferHandle location will be set to IX_NULL_BUFFER_HANDLE.
 * 
 * @Param:  - IN arg_hBuffer - handle of the buffer for which we need the next link.
 * @Param:  - OUT arg_pNextBufferHandle - the location where to store the next link for this buffer. 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_buffer_get_next(
                                ix_buffer_handle arg_hBuffer, 
                                ix_buffer_handle* arg_pNextBufferHandle
                              )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_buffer_get_next";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((arg_hBuffer & IX_RM_SINGLE_BUFFER_MASK) == IX_RM_SINGLE_BUFFER_MASK)
    {
        *arg_pNextBufferHandle = IX_NULL_BUFFER_HANDLE;
        return IX_SUCCESS;
    }

    if(IX_RM_BUFFER_GET_TYPE(arg_hBuffer) == IX_BUFFER_TYPE_HARDWARE)
    {

        if(arg_hBuffer == IX_NULL_BUFFER_HANDLE)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_HANDLE_INVALID, s_aFunctionIdentifier, 
                   "the the passed buffer handle is invalid!", 0);
#if defined( _IX_RM_DEBUG_ )

        /* do extra checking in debug mode */
        {
            _ix_buffer_free_list* pFreeList = g_FLManager.m_aFreeListAddress[IX_RM_BUF_HW_GET_FL_ID(arg_hBuffer)];

            if(pFreeList == 0)
            {
                return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_INVALID, s_aFunctionIdentifier, 
                            "the associated free list for this buffer is invalid!", 0);
            }
        }

#endif /* end defined( _IX_RM_DEBUG_ ) */

        /* hardware buffer */

#if defined(_IX_RM_SPLIT_META_DATA_)

        *arg_pNextBufferHandle = IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(arg_hBuffer)->m_HwNext;

#else

        *arg_pNextBufferHandle = IX_RM_BUF_HW_GET_META_ADDRESS(arg_hBuffer)->m_HwNext;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

    }
    else
    {
        /* software buffer */
        _ix_buffer_free_list* pFreeList = 0;
        ix_uint32 flId = IX_RM_BUF_SW_GET_FL_ID(arg_hBuffer);
        ix_sw_buffer_meta* pMetaData;

#if defined( _IX_RM_DEBUG_ )

        /* check if the ID coresponds to a software free list */
        if(flId < IX_BUF_FIRST_SW_FL_ID)
        {
            return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_ID_INVALID, s_aFunctionIdentifier, 
                        "the free list ID coresponds to a hardware free list instead of a software one!", 0);
        }

#endif /* end defined( _IX_RM_DEBUG_ ) */

        pFreeList = g_FLManager.m_aFreeListAddress[flId];

        pMetaData = (void*)((ix_uint8*)pFreeList->m_pMetaBaseAddress + 
            (IX_RM_BUF_SW_GET_INDEX(arg_hBuffer) << pFreeList->m_MetaElementShift));

        *arg_pNextBufferHandle = pMetaData->m_hNext;
    }

    return IX_SUCCESS;
}




/**
 * NAME: ix_rm_buffer_link
 *
 * DESCRIPTION: This function will link the buffer referred by arg_pNextBufferHandle to the buffer 
 *          referred by arg_pBufferHandle. The buffers will be in a linked list construct. If the 
 *          current buffer (arg_pBufferHandle) has already a link then the function will return an
 *          error. On return the handles of the two buffers will be modified to specify the new
 *          EOP/SOP state.
 * 
 * @Param:  - IN arg_pBufferHandle - the reference to the handle of the buffer where to perform the link
 * @Param:  - IN arg_pNextBufferHandle - the reference to the handle of the buffer to be linked.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure. If the 
 *          current buffer (arg_pBufferHandle) has already a link then the function will return an
 *          error.
 */
ix_error ix_rm_buffer_link(
                            ix_buffer_handle* arg_pBufferHandle, 
                            ix_buffer_handle* arg_pNextBufferHandle
                          )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_buffer_link";
    register ix_buffer_handle hBuffer = *arg_pBufferHandle;
    register ix_buffer_handle hNextBuffer = *arg_pNextBufferHandle;
    register ix_buffer_handle hTempBuffer;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((hBuffer == IX_NULL_BUFFER_HANDLE))
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_HANDLE_INVALID, s_aFunctionIdentifier, 
                   "the passed buffer handles is invalid!", 0);

    if(IX_RM_BUF_GET_EOP(hBuffer) == 0)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_ALREADY_LINKED, s_aFunctionIdentifier, 
                    "the destination buffer has already a link!", 0);
    }

    if(hNextBuffer == IX_NULL_BUFFER_HANDLE)
    {
        if(IX_RM_BUFFER_GET_TYPE(hBuffer) == IX_BUFFER_TYPE_HARDWARE)
        {
#if defined( _IX_RM_DEBUG_ )

            /* do extra checking in debug mode */
            {
                _ix_buffer_free_list* pFreeList = g_FLManager.m_aFreeListAddress[IX_RM_BUF_HW_GET_FL_ID(hBuffer)];

                if(pFreeList == 0)
                {
                    return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_INVALID, s_aFunctionIdentifier, 
                                "the associated free list for this buffer is invalid!", 0);
                }
            }

#endif /* end defined( _IX_RM_DEBUG_ ) */

#if defined(_IX_RM_SPLIT_META_DATA_)

            IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(hBuffer)->m_HwNext = IX_NULL_BUFFER_HANDLE;

#else

            IX_RM_BUF_HW_GET_META_ADDRESS(hBuffer)->m_HwNext = IX_NULL_BUFFER_HANDLE;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

            return IX_SUCCESS;
        }
        else
        {
            /* software buffer */
            _ix_buffer_free_list* pFreeList = 0;
            ix_uint32 flId = IX_RM_BUF_SW_GET_FL_ID(hBuffer);
            register ix_sw_buffer_meta* pMeta = 0;

#if defined( _IX_RM_DEBUG_ )

            /* check if the ID coresponds to a software free list */
            if(flId < IX_BUF_FIRST_SW_FL_ID)
            {
                return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_ID_INVALID, s_aFunctionIdentifier, 
                            "the free list ID coresponds to a hardware free list instead of a software one!", 0);
            }

#endif /* end defined( _IX_RM_DEBUG_ ) */

            pFreeList = g_FLManager.m_aFreeListAddress[flId];

            pMeta = (void*)((ix_uint8*)pFreeList->m_pMetaBaseAddress + 
                (IX_RM_BUF_SW_GET_INDEX(hBuffer) << pFreeList->m_MetaElementShift));

            pMeta->m_hNext = IX_NULL_BUFFER_HANDLE;
        }
    }

    /* clear the SOP flag for hNextBuffer */
    IX_RM_BUF_SET_SOP(hNextBuffer, 0);

    /* clear the EOP flag for hBuffer */
    IX_RM_BUF_SET_EOP(hBuffer, 0);

    if(IX_RM_BUFFER_GET_TYPE(hBuffer) == IX_BUFFER_TYPE_HARDWARE)
    {

        /* hardware buffer */

#if defined(_IX_RM_SPLIT_META_DATA_)

        register ix_hw_internal_buffer_meta* pMeta = IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(hBuffer);

#else

        register ix_hw_buffer_meta* pMeta = IX_RM_BUF_HW_GET_META_ADDRESS(hBuffer);

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */


#if defined( _IX_RM_DEBUG_ )

        /* do extra checking in debug mode */
        {
            _ix_buffer_free_list* pFreeList = g_FLManager.m_aFreeListAddress[IX_RM_BUF_HW_GET_FL_ID(hBuffer)];

            if(pFreeList == 0)
            {
                return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_INVALID, s_aFunctionIdentifier, 
                            "the associated free list for this buffer is invalid!", 0);
            }
        }

#endif /* end defined( _IX_RM_DEBUG_ ) */

        /* check if the next buffer is single buffer */
        if(IX_RM_BUF_GET_EOP(hNextBuffer) != 0)
        {
            /* make the next of the hNextBuffer point to hBuffer */
            if(IX_RM_BUFFER_GET_TYPE(hNextBuffer) == IX_BUFFER_TYPE_HARDWARE)
            {

#if defined(_IX_RM_SPLIT_META_DATA_)

                IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(hNextBuffer)->m_HwNext = hBuffer;

#else

                IX_RM_BUF_HW_GET_META_ADDRESS(hNextBuffer)->m_HwNext = hBuffer;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

            }
            else
            {
                /* software buffer */
                _ix_buffer_free_list* pFreeList = 0;
                ix_uint32 flId = IX_RM_BUF_SW_GET_FL_ID(hNextBuffer);
                register ix_sw_buffer_meta* pMeta1 = 0;

#if defined( _IX_RM_DEBUG_ )

                /* check if the ID coresponds to a software free list */
                if(flId < IX_BUF_FIRST_SW_FL_ID)
                {
                    return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_ID_INVALID, s_aFunctionIdentifier, 
                                "the free list ID coresponds to a hardware free list instead of a software one!", 0);
                }

#endif /* end defined( _IX_RM_DEBUG_ ) */

                pFreeList = g_FLManager.m_aFreeListAddress[flId];

                pMeta1 = (void*)((ix_uint8*)pFreeList->m_pMetaBaseAddress + 
                    (IX_RM_BUF_SW_GET_INDEX(hNextBuffer) << pFreeList->m_MetaElementShift));

                pMeta1->m_hNext = hBuffer;
                pMeta1->m_hSelf = hNextBuffer;
            }
        }

        /* check if single buffer */
        if(IX_RM_BUF_GET_SOP(hBuffer) == 0)
        {
            /* in this case we get the next buffer which is the previous buffer in the chain */
            hTempBuffer = pMeta->m_HwNext;

            /* update the previous buffer of hBuffer */
            if(IX_RM_BUFFER_GET_TYPE(hNextBuffer) == IX_BUFFER_TYPE_HARDWARE)
            {

#if defined(_IX_RM_SPLIT_META_DATA_)

                IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(hTempBuffer)->m_HwNext = hBuffer;

#else

                IX_RM_BUF_HW_GET_META_ADDRESS(hTempBuffer)->m_HwNext = hBuffer;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

            }
            else
            {
                /* software buffer */
                _ix_buffer_free_list* pFreeList = 0;
                ix_uint32 flId = IX_RM_BUF_SW_GET_FL_ID(hTempBuffer);
                register ix_sw_buffer_meta* pMeta1 = 0;

#if defined( _IX_RM_DEBUG_ )

                /* check if the ID coresponds to a software free list */
                if(flId < IX_BUF_FIRST_SW_FL_ID)
                {
                    return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_ID_INVALID, s_aFunctionIdentifier, 
                                "the free list ID coresponds to a hardware free list instead of a software one!", 0);
                }

#endif /* end defined( _IX_RM_DEBUG_ ) */

                pFreeList = g_FLManager.m_aFreeListAddress[flId];

                pMeta1 = (void*)((ix_uint8*)pFreeList->m_pMetaBaseAddress + 
                    (IX_RM_BUF_SW_GET_INDEX(hTempBuffer) << pFreeList->m_MetaElementShift));

                pMeta1->m_hNext = hBuffer;
            }
        }

        /* link the buffers */
        pMeta->m_HwNext = hNextBuffer;

    }
    else
    {
        /* software buffer */
        _ix_buffer_free_list* pFreeList = 0;
        ix_uint32 flId = IX_RM_BUF_SW_GET_FL_ID(hBuffer);
        register ix_sw_buffer_meta* pMeta = 0;

#if defined( _IX_RM_DEBUG_ )

        /* check if the ID coresponds to a software free list */
        if(flId < IX_BUF_FIRST_SW_FL_ID)
        {
            return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_ID_INVALID, s_aFunctionIdentifier, 
                        "the free list ID coresponds to a hardware free list instead of a software one!", 0);
        }

#endif /* end defined( _IX_RM_DEBUG_ ) */

        pFreeList = g_FLManager.m_aFreeListAddress[flId];

        pMeta = (void*)((ix_uint8*)pFreeList->m_pMetaBaseAddress + 
            (IX_RM_BUF_SW_GET_INDEX(hBuffer) << pFreeList->m_MetaElementShift));

        /* check if the next buffer is single buffer */
        if(IX_RM_BUF_GET_EOP(hNextBuffer) != 0)
        {
            /* make the next of the hNextBuffer point to hBuffer */
            if(IX_RM_BUFFER_GET_TYPE(hNextBuffer) == IX_BUFFER_TYPE_HARDWARE)
            {

#if defined(_IX_RM_SPLIT_META_DATA_)

                IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(hNextBuffer)->m_HwNext = hBuffer;

#else

                IX_RM_BUF_HW_GET_META_ADDRESS(hNextBuffer)->m_HwNext = hBuffer;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

            }
            else
            {
                /* software buffer */
                register ix_sw_buffer_meta* pMeta1 = 0;
                pFreeList = 0;
                flId = IX_RM_BUF_SW_GET_FL_ID(hNextBuffer);

#if defined( _IX_RM_DEBUG_ )

                /* check if the ID coresponds to a software free list */
                if(flId < IX_BUF_FIRST_SW_FL_ID)
                {
                    return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_ID_INVALID, s_aFunctionIdentifier, 
                                "the free list ID coresponds to a hardware free list instead of a software one!", 0);
                }

#endif /* end defined( _IX_RM_DEBUG_ ) */

                pFreeList = g_FLManager.m_aFreeListAddress[flId];

                pMeta1 = (void*)((ix_uint8*)pFreeList->m_pMetaBaseAddress + 
                    (IX_RM_BUF_SW_GET_INDEX(hNextBuffer) << pFreeList->m_MetaElementShift));

                pMeta1->m_hNext = hBuffer;
                pMeta1->m_hSelf = hNextBuffer;
            }
        }

        /* check if single buffer */
        if(IX_RM_BUF_GET_SOP(hBuffer) == 0)
        {
            /* in this case we get the next buffer which is the head of the chain */
            hTempBuffer = pMeta->m_hNext;

            /* update the previous buffer of hBuffer */
            if(IX_RM_BUFFER_GET_TYPE(hNextBuffer) == IX_BUFFER_TYPE_HARDWARE)
            {

#if defined(_IX_RM_SPLIT_META_DATA_)

                IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(hTempBuffer)->m_HwNext = hBuffer;

#else

                IX_RM_BUF_HW_GET_META_ADDRESS(hTempBuffer)->m_HwNext = hBuffer;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

            }
            else
            {
                    /* software buffer */
                register ix_sw_buffer_meta* pMeta1 = 0;
                pFreeList = 0;
                flId = IX_RM_BUF_SW_GET_FL_ID(hTempBuffer);

#if defined( _IX_RM_DEBUG_ )

                /* check if the ID coresponds to a software free list */
                if(flId < IX_BUF_FIRST_SW_FL_ID)
                {
                    return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_ID_INVALID, s_aFunctionIdentifier, 
                                "the free list ID coresponds to a hardware free list instead of a software one!", 0);
                }

#endif /* end defined( _IX_RM_DEBUG_ ) */

                pFreeList = g_FLManager.m_aFreeListAddress[flId];

                pMeta1 = (void*)((ix_uint8*)pFreeList->m_pMetaBaseAddress + 
                    (IX_RM_BUF_SW_GET_INDEX(hTempBuffer) << pFreeList->m_MetaElementShift));

                pMeta1->m_hNext = hBuffer;
            }
        }

        /* link the buffers */
        pMeta->m_hNext = hNextBuffer;

    }

    /* update the handles */
    *arg_pBufferHandle = hBuffer;
    *arg_pNextBufferHandle = hNextBuffer;

    return IX_SUCCESS;
}



/**
 * NAME: ix_rm_buffer_unlink
 *
 * DESCRIPTION: This function will unlink the buffer referred by arg_pBufferHandle. If the buffer has
 *          a link that will be set to IX_NULL_BUFFER_HANDLE and the unlinked buffer handle will be
 *          returned at the arg_pUnlinkedBufferHandle location. On return the handles of the two buffers will be modified to specify the new
 *          EOP/SOP state. In order this function to work *arg_pBufferHandle should have the SOP flag set
 *          which mean that we can unlink just at a beginning of a chain.
 * 
 * @Param:  - IN arg_pBufferHandle - the reference to the handle of the buffer that will go from a linked to an unliked state.
 * @Param:  - OUT arg_pUnlinkedBufferHandle - location where the handle of the unlinked buffer will be stored.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_buffer_unlink(
                              ix_buffer_handle* arg_pBufferHandle,
                              ix_buffer_handle* arg_pUnlinkedBufferHandle
                            )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_buffer_unlink";

    register ix_buffer_handle hBuffer = *arg_pBufferHandle;
    register ix_buffer_handle hUnlinkedBuffer = IX_NULL_BUFFER_HANDLE;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(hBuffer == IX_NULL_BUFFER_HANDLE)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_HANDLE_INVALID, s_aFunctionIdentifier, 
                   "the passed buffer handle is invalid!", 0);

    if(IX_RM_BUF_GET_SOP(hBuffer) == 0)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_HANDLE_INVALID, s_aFunctionIdentifier, 
                   "the passed buffer is not an SOP buffer!", 0);

    if(IX_RM_BUF_GET_EOP(hBuffer) != 0)
    {
        *arg_pUnlinkedBufferHandle = IX_NULL_BUFFER_HANDLE;
        return IX_SUCCESS;
    }

    /* set the EOP flag for hBuffer */
    IX_RM_BUF_SET_EOP(hBuffer, 1);

    if(IX_RM_BUFFER_GET_TYPE(hBuffer) == IX_BUFFER_TYPE_HARDWARE)
    {
        /* hardware buffer */

#if defined(_IX_RM_SPLIT_META_DATA_)

        ix_hw_internal_buffer_meta* pMeta = IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(hBuffer);

#else

        ix_hw_buffer_meta* pMeta = IX_RM_BUF_HW_GET_META_ADDRESS(hBuffer);

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */


#if defined( _IX_RM_DEBUG_ )

        /* do extra checking in debug mode */
        {
            _ix_buffer_free_list* pFreeList = g_FLManager.m_aFreeListAddress[IX_RM_BUF_HW_GET_FL_ID(hBuffer)];

            if(pFreeList == 0)
            {
                return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_INVALID, s_aFunctionIdentifier, 
                            "the associated free list for this buffer is invalid!", 0);
            }
        }

#endif /* end defined( _IX_RM_DEBUG_ ) */
        
        hUnlinkedBuffer = pMeta->m_HwNext;

    }
    else
    {
        /* software buffer */
        _ix_buffer_free_list* pFreeList = 0;
        ix_uint32 flId = IX_RM_BUF_SW_GET_FL_ID(hBuffer);
        ix_sw_buffer_meta* pMeta = 0;

#if defined( _IX_RM_DEBUG_ )

        /* check if the ID coresponds to a software free list */
        if(flId < IX_BUF_FIRST_SW_FL_ID)
        {
            return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_ID_INVALID, s_aFunctionIdentifier, 
                        "the free list ID coresponds to a hardware free list instead of a software one!", 0);
        }

#endif /* end defined( _IX_RM_DEBUG_ ) */

        pFreeList = g_FLManager.m_aFreeListAddress[flId];

        pMeta = (void*)((ix_uint8*)pFreeList->m_pMetaBaseAddress + 
            (IX_RM_BUF_SW_GET_INDEX(hBuffer) << pFreeList->m_MetaElementShift));

        hUnlinkedBuffer = pMeta->m_hNext;
        pMeta->m_hSelf = hBuffer;
    }

    /* set the SOP flag for the returned link */
    IX_RM_BUF_SET_SOP(hUnlinkedBuffer, 1);

    *arg_pBufferHandle = hBuffer;
    *arg_pUnlinkedBufferHandle = hUnlinkedBuffer;

    return IX_SUCCESS;
}



#if defined( _IX_RM_COMPONENT_MULTI_THREADED_ )


/**
 * NAME: ix_rm_sw_free_list_init_sync
 *
 * DESCRIPTION: This function will initialize the synchronization for the specified
 *          buffer free list. A detailed explanation on how sychronization
 *          mechanism works is found in the L0 HLD/LLD and also in sync_mechanism.h
 *          file.
 * 
 * @Param:  - IN arg_hSwFreeList - this is the handle of the free list for which we
 *              want to set up the synchronization mechanism
 * @Param:  - IN arg_LockFunction - this is the lock function of the sync mechanism.
 * @Param:  - IN arg_UnlockFunction - this is the unlock function of the sync mechanism.
 * @Param:  - IN arg_pContext - context that will be passed to each invocation of lock
 *          and unlock function. It might point to a synchronization primitive to be
 *          used by the specified lock and unlock functions.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_sw_free_list_init_sync(
                                       ix_buffer_free_list_handle    arg_hSwFreeList, 
                                       ix_sa_lock_ptr                arg_LockFunction,
                                       ix_sa_unlock_ptr              arg_UnlockFunction,
                                       void*                         arg_pContext
                                     )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_sw_free_list_init_sync";

    ix_error err = 0;
    _ix_buffer_free_list* pFreeList = 0;
    ix_synchronization_adapter* pSyncAdapter;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* retrieve the free list stucture address coresponding to the passed handle */

#if defined( _IX_RM_DEBUG_ )

    /* do extra checking in debug mode */
    IX_RM_ASSERT(IX_RM_BUF_FL_IS_VALID(arg_hSwFreeList) != 0, s_aFunctionIdentifier, 
        "the passed free list handle is invalid!");
    {
        ix_uint32 flId = IX_RM_BUF_FL_GET_FL_ID(arg_hSwFreeList);

        pFreeList = g_FLManager.m_aFreeListAddress[flId];

        if(pFreeList == 0)
        {
            return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_INVALID, s_aFunctionIdentifier, 
                        "the associated free list for this buffer is invalid!", 0);
        }

    }

#else

    pFreeList = g_FLManager.m_aFreeListAddress[IX_RM_BUF_FL_GET_FL_ID(arg_hSwFreeList)];

#endif /* end defined( _IX_RM_DEBUG_ ) */

    pSyncAdapter = &pFreeList->m_FLData.m_SwData.m_SyncAdapter;

    if((arg_LockFunction != 0) && (arg_UnlockFunction != 0))
    {
        pSyncAdapter->m_pContext = arg_pContext;
        pSyncAdapter->m_LockFuncPtr = arg_LockFunction;
        pSyncAdapter->m_UnlockFuncPtr = arg_UnlockFunction;

        /* set the synchronization mechanism in active state */
        pSyncAdapter->m_SyncActive = 1;
    }
    else
    {
        pSyncAdapter->m_SyncActive = 0;
    }

    return err;
}
#endif /* defined( _IX_RM_COMPONENT_MULTI_THREADED_ ) */


/**
 * NAME: ix_rm_buffer_make_safe_single
 *
 * DESCRIPTION: This function will set the next link buffer for a single buffer (SOP|EOP flags set)
 *  to IX_NULL_BUFFER_HANDLE. If the buffer is not a single buffer(i.e. belongs to a chain), then an
 *  error will be returned. This function should be used in conjunction with ix_rm_buffer_make_safe_chained()
 *  to fix stale buffer handles.
 * 
 * @Param:  - INOUT  arg_hBuffer - buffer handle to the buffer to be changed. On return the meta
 *  data will be changed in such a way that the next link will be set to IX_NULL_BUFFER_HANDLE
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_buffer_make_safe_single(
                                        ix_buffer_handle arg_hBuffer
                                      )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_buffer_make_safe_single";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_hBuffer == IX_NULL_BUFFER_HANDLE)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_HANDLE_INVALID, s_aFunctionIdentifier, 
               "the the passed buffer handle is invalid!", 0);

    if((arg_hBuffer & IX_RM_SINGLE_BUFFER_MASK) != IX_RM_SINGLE_BUFFER_MASK)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_HANDLE_INVALID, s_aFunctionIdentifier, 
               "the the passed buffer handle does not belong to a single buffer!", 0);
    }

    if(IX_RM_BUFFER_GET_TYPE(arg_hBuffer) == IX_BUFFER_TYPE_HARDWARE)
    {
#if defined( _IX_RM_DEBUG_ )

        /* do extra checking in debug mode */
        {
            _ix_buffer_free_list* pFreeList = g_FLManager.m_aFreeListAddress[IX_RM_BUF_HW_GET_FL_ID(arg_hBuffer)];

            if(pFreeList == 0)
            {
                return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_INVALID, s_aFunctionIdentifier, 
                            "the associated free list for this buffer is invalid!", 0);
            }
        }

#endif /* end defined( _IX_RM_DEBUG_ ) */

        /* hardware buffer */

#if defined(_IX_RM_SPLIT_META_DATA_)

        IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(arg_hBuffer)->m_HwNext = IX_NULL_BUFFER_HANDLE;

#else

        IX_RM_BUF_HW_GET_META_ADDRESS(arg_hBuffer)->m_HwNext = IX_NULL_BUFFER_HANDLE;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

    }
    else
    {
        /* software buffer */
        _ix_buffer_free_list* pFreeList = 0;
        ix_uint32 flId = IX_RM_BUF_SW_GET_FL_ID(arg_hBuffer);
        ix_sw_buffer_meta* pMetaData;

#if defined( _IX_RM_DEBUG_ )

        /* check if the ID coresponds to a software free list */
        if(flId < IX_BUF_FIRST_SW_FL_ID)
        {
            return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_ID_INVALID, s_aFunctionIdentifier, 
                        "the free list ID coresponds to a hardware free list instead of a software one!", 0);
        }

#endif /* end defined( _IX_RM_DEBUG_ ) */

        pFreeList = g_FLManager.m_aFreeListAddress[flId];

        pMetaData = (void*)((ix_uint8*)pFreeList->m_pMetaBaseAddress + 
            (IX_RM_BUF_SW_GET_INDEX(arg_hBuffer) << pFreeList->m_MetaElementShift));

        pMetaData->m_hNext = IX_NULL_BUFFER_HANDLE;
    }

    return IX_SUCCESS;
}


/**
 * NAME: ix_rm_buffer_make_safe_chained
 *
 * DESCRIPTION: This function will reset the EOP flag for the passed buffer handle if the next
 *  buffer link is not IX_NULL_BUFFER_HANDLE. In the case the buffer handle does not have the SOP
 *  flag set, an error will be returned. This function should be used in conjunction with 
 *  ix_rm_buffer_make_safe_single() to fix stale buffer handles.
 * 
 * @Param:  - INOUT  arg_pBufferHandle - address to the buffer handle to be updated.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_buffer_make_safe_chained(
                                         ix_buffer_handle* arg_pBufferHandle
                                       )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_buffer_make_safe_chained";

    register ix_buffer_handle hBuffer = *arg_pBufferHandle;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(hBuffer == IX_NULL_BUFFER_HANDLE)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_HANDLE_INVALID, s_aFunctionIdentifier, 
               "the the passed buffer handle is invalid!", 0);

    if(IX_RM_BUF_GET_SOP(hBuffer) == 0)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_HANDLE_INVALID, s_aFunctionIdentifier, 
               "the the passed buffer handle is not an SOP handle!", 0);
    }

    if(IX_RM_BUF_GET_EOP(hBuffer) == 0)
        /* nothing to be done in this case */
        return IX_SUCCESS; 

    if(IX_RM_BUFFER_GET_TYPE(hBuffer) == IX_BUFFER_TYPE_HARDWARE)
    {
        /* hardware buffer */

#if defined( _IX_RM_DEBUG_ )

        /* do extra checking in debug mode */
        {
            _ix_buffer_free_list* pFreeList = g_FLManager.m_aFreeListAddress[IX_RM_BUF_HW_GET_FL_ID(hBuffer)];

            if(pFreeList == 0)
            {
                return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_INVALID, s_aFunctionIdentifier, 
                            "the associated free list for this buffer is invalid!", 0);
            }
        }

#endif /* end defined( _IX_RM_DEBUG_ ) */

        /* reset the EOP flag */
#if defined(_IX_RM_SPLIT_META_DATA_)

        if(IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(hBuffer)->m_HwNext != IX_NULL_BUFFER_HANDLE)

#else

        if(IX_RM_BUF_HW_GET_META_ADDRESS(hBuffer)->m_HwNext != IX_NULL_BUFFER_HANDLE)

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

        {
            IX_RM_BUF_SET_EOP(hBuffer, 0);
            *arg_pBufferHandle = hBuffer;
        }
    }
    else
    {
        /* software buffer */
        _ix_buffer_free_list* pFreeList = 0;
        ix_uint32 flId = IX_RM_BUF_SW_GET_FL_ID(hBuffer);
        ix_sw_buffer_meta* pMetaData;

#if defined( _IX_RM_DEBUG_ )

        /* check if the ID coresponds to a software free list */
        if(flId < IX_BUF_FIRST_SW_FL_ID)
        {
            return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_ID_INVALID, s_aFunctionIdentifier, 
                        "the free list ID coresponds to a hardware free list instead of a software one!", 0);
        }

#endif /* end defined( _IX_RM_DEBUG_ ) */

        pFreeList = g_FLManager.m_aFreeListAddress[flId];

        pMetaData = (void*)((ix_uint8*)pFreeList->m_pMetaBaseAddress + 
            (IX_RM_BUF_SW_GET_INDEX(hBuffer) << pFreeList->m_MetaElementShift));

        /* reset the EOP flag */
        if(pMetaData->m_hNext != IX_NULL_BUFFER_HANDLE)
        {
            IX_RM_BUF_SET_EOP(hBuffer, 0);
            pMetaData->m_hSelf = hBuffer;
            *arg_pBufferHandle = hBuffer;
        }
    }

    return IX_SUCCESS;
}

#if defined (_IX_RM_BUFFER_DEBUG_)


/**
 * NAME: ix_rm_free_list_print_available_buffers
 *
 * DESCRIPTION: This function will print information about the available buffers in a
 *  certain free list
 * 
 * @Param:  - IN arg_FreeListType - describes the type of the free list: 0 for HW
 *  and 1 for SW
 * @Param:  - IN arg_FreeListNumber - describes the free list number we are interested
 *  in
 */
void ix_rm_free_list_print_available_buffers(
                                              ix_uint32 arg_FreeListType, 
                                              ix_uint32 arg_FreeListNumber
                                            )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_free_list_print_available_buffers";

    _ix_buffer_free_list* pFreeList = 0;
    ix_uint32 availableBuffers;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_FreeListType > 1)
    {
        ix_ossl_message_log("The free list type should be 0 for HW free lists and 1 for SW free lists\n");
        return;
    }

    if(arg_FreeListType == 0)
    {
        if(arg_FreeListNumber >= IX_BUF_MAX_HW_FL_NUMBER)
        {
            ix_ossl_message_log("Bad free list number\n");
            return;
        }

        /* retrieve the free list of interest */
        if(g_FLManager.m_aFreeListAddress[arg_FreeListNumber] != 0)
        {
            pFreeList = g_FLManager.m_aFreeListAddress[arg_FreeListNumber];
        }
    }
    else
    {
        if(arg_FreeListNumber >= (IX_BUF_MAX_FL_NUMBER - IX_BUF_FIRST_SW_FL_ID))
        {
            ix_ossl_message_log("Bad free list number\n");
            return;
        }

        /* retrieve the free list of interest */
        if(g_FLManager.m_aFreeListAddress[arg_FreeListNumber + IX_BUF_FIRST_SW_FL_ID] != 0)
        {
            pFreeList = g_FLManager.m_aFreeListAddress[arg_FreeListNumber + IX_BUF_FIRST_SW_FL_ID];
        }
    }

    if(pFreeList == 0)
    {
        ix_ossl_message_log("The requested free list was not found\n");
        return;
    }

    ix_ossl_message_log("Available buffers for the %s free list %lu\n", 
        (arg_FreeListType == 0) ? "hardware" : "software", arg_FreeListNumber);

    if(arg_FreeListType == 0)
        _ix_rm_free_list_get_available_buffers(IX_RM_BUF_FL_MAKE_HANDLE(arg_FreeListNumber), &availableBuffers);
    else
        _ix_rm_free_list_get_available_buffers(IX_RM_BUF_FL_MAKE_HANDLE(arg_FreeListNumber + IX_BUF_FIRST_SW_FL_ID), &availableBuffers);

    
    ix_ossl_message_log("\tThere are %lu available buffers out %lu existing in the list!\n", availableBuffers, pFreeList->m_NumElements);

    ix_ossl_message_log("\n");

}



/**
 * NAME: ix_rm_free_list_print_buffers_info
 *
 * DESCRIPTION: This function will print information about the buffers belonging to a certain
 *  HW free list
 * 
 * @Param:  - IN arg_FreeListNumber - describes the free list number we are interested
 *  in
 */
void ix_rm_free_list_print_buffers_info(
                                         ix_uint32 arg_FreeListNumber
                                       )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_free_list_print_buffers_info";

    _ix_buffer_free_list* pFreeList = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_FreeListNumber >= IX_BUF_MAX_HW_FL_NUMBER)
    {
        ix_ossl_message_log("Bad free list number\n");
        return;
    }

    /* retrieve the free list of interest */
    if(g_FLManager.m_aFreeListAddress[arg_FreeListNumber] != 0)
    {
        pFreeList = g_FLManager.m_aFreeListAddress[arg_FreeListNumber];
    }

    {
        ix_buffer_debug_info* pDebugInfo;
        ix_uint32 i;
        ix_uint32 allocStatus[IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_LAST];

        ix_ossl_memset(allocStatus, 0, sizeof(allocStatus));
        for(i = 0, pDebugInfo = pFreeList->m_pDebugInfo; i < pFreeList->m_NumElements; i++, pDebugInfo++)
        {
            allocStatus[pDebugInfo->m_LastCoreAllocationStatus]++;
        }

        ix_ossl_message_log("In HW free list %lu, there are %lu buffers\n"
            "they have been last allocated on the core side as follows:\n", 
            arg_FreeListNumber, pFreeList->m_NumElements);
        ix_ossl_message_log("\tLast freed by the core: %lu\n",allocStatus[IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_FREE]);
        ix_ossl_message_log("\tLast allocated by the core: %lu\n",allocStatus[IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_ALLOCATED_BY_CORE]);
        ix_ossl_message_log("\tLast sent to the core as packet buffers: %lu\n",allocStatus[IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_RECEIVED_FROM_ME_AS_PACKET]);
        ix_ossl_message_log("\tLast sent to the core as message buffers: %lu\n",allocStatus[IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_RECEIVED_FROM_ME_AS_MESSAGE]);
    }

    return;
}


/**
 * NAME: ix_rm_free_list_print_info
 *
 * DESCRIPTION: This function will print information about a certain free list
 * 
 * @Param:  - IN arg_FreeListType - describes the type of the free list: 0 for HW
 *  and 1 for SW
 * @Param:  - IN arg_FreeListNumber - describes the free list number we are interested
 *  in
 */
void ix_rm_free_list_print_info(
                                 ix_uint32 arg_FreeListType, 
                                 ix_uint32 arg_FreeListNumber
                               )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_free_list_print_info";

    _ix_buffer_free_list* pFreeList = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_FreeListType > 1)
    {
        ix_ossl_message_log("The free list type should be 0 for HW free lists and 1 for SW free lists\n");
        return;
    }

    if(arg_FreeListType == 0)
    {
        if(arg_FreeListNumber >= IX_BUF_MAX_HW_FL_NUMBER)
        {
            ix_ossl_message_log("Bad free list number\n");
            return;
        }

        /* retrieve the free list of interest */
        if(g_FLManager.m_aFreeListAddress[arg_FreeListNumber] != 0)
        {
            pFreeList = g_FLManager.m_aFreeListAddress[arg_FreeListNumber];
        }
    }
    else
    {
        if(arg_FreeListNumber >= (IX_BUF_MAX_FL_NUMBER - IX_BUF_FIRST_SW_FL_ID))
        {
            ix_ossl_message_log("Bad free list number\n");
            return;
        }

        /* retrieve the free list of interest */
        if(g_FLManager.m_aFreeListAddress[arg_FreeListNumber + IX_BUF_FIRST_SW_FL_ID] != 0)
        {
            pFreeList = g_FLManager.m_aFreeListAddress[arg_FreeListNumber + IX_BUF_FIRST_SW_FL_ID];
        }
    }

    if(pFreeList == 0)
    {
        ix_ossl_message_log("The requested free list was not found\n");
        return;
    }

    ix_ossl_message_log("Information for the %s free list %lu\n", 
        (arg_FreeListType == 0) ? "hardware" : "software", arg_FreeListNumber);

    ix_ossl_message_log("\tMetaBaseAddress:       0x%p\n", pFreeList->m_pMetaBaseAddress);
    ix_ossl_message_log("\tMetaMemoryType:        ");

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

    switch(pFreeList->m_MetaMemoryType)
    {
        case IX_MEMORY_TYPE_DRAM:
            ix_ossl_message_log("DRAM");
            break;
        case IX_MEMORY_TYPE_SRAM:
            ix_ossl_message_log("SRAM");
            break;
        case IX_MEMORY_TYPE_SCRATCH:
            ix_ossl_message_log("SCRATCH");
            break;
        default:
            ix_ossl_message_log("LOCAL");
            break;
    }

#else

    switch(pFreeList->m_MetaMemoryType)
    {
        case IX_MEMORY_TYPE_DRAM_SOUTH:
            ix_ossl_message_log("South bank DRAM");
            break;
        case IX_MEMORY_TYPE_SRAM_MSG_SOUTH:
            ix_ossl_message_log("South bank MSG-SRAM");
            break;
        case IX_MEMORY_TYPE_DRAM_NORTH:
            ix_ossl_message_log("North bank aliased DRAM");
            break;
        case IX_MEMORY_TYPE_SRAM_NORTH:
            ix_ossl_message_log("North bank aliased SRAM");
            break;
        case IX_MEMORY_TYPE_SCRATCH:
            ix_ossl_message_log("North bank aliased SCRATCH");
            break;
        default:
            ix_ossl_message_log("Unknown");
            break;
    }

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

    ix_ossl_message_log("\n");
    ix_ossl_message_log("\tMetaMemoryChannel:     %lu\n", pFreeList->m_MetaMemoryChannel);
    ix_ossl_message_log("\tMetaMemorySize:        %d\n", 1 << pFreeList->m_MetaElementShift);

#if defined(_IX_RM_SPLIT_META_DATA_)

    if(arg_FreeListType == 0)
    {
        ix_ossl_message_log("\tInternalMetaBaseAddress:   0x%p\n", pFreeList->m_FLData.m_HwData.m_pInternalMetaBaseAddress);
        ix_ossl_message_log("\tInternalMetaMemoryType:    ");

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

        switch(pFreeList->m_FLData.m_HwData.m_InternalMetaMemoryType)
        {
            case IX_MEMORY_TYPE_DRAM:
                ix_ossl_message_log("DRAM");
                break;
            case IX_MEMORY_TYPE_SRAM:
                ix_ossl_message_log("SRAM");
                break;
            case IX_MEMORY_TYPE_SCRATCH:
                ix_ossl_message_log("SCRATCH");
                break;
            default:
                ix_ossl_message_log("LOCAL");
                break;
        }

#else

        switch(pFreeList->m_FLData.m_HwData.m_InternalMetaMemoryType)
        {
            case IX_MEMORY_TYPE_DRAM_SOUTH:
                ix_ossl_message_log("South bank DRAM");
                break;
            case IX_MEMORY_TYPE_SRAM_MSG_SOUTH:
                ix_ossl_message_log("South bank MSG-SRAM");
                break;
            case IX_MEMORY_TYPE_DRAM_NORTH:
                ix_ossl_message_log("North bank aliased DRAM");
                break;
            case IX_MEMORY_TYPE_SRAM_NORTH:
                ix_ossl_message_log("North bank aliased SRAM");
                break;
            case IX_MEMORY_TYPE_SCRATCH:
                ix_ossl_message_log("North bank aliased SCRATCH");
                break;
            default:
                ix_ossl_message_log("Unknown");
                break;
        }

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

        ix_ossl_message_log("\n");
        ix_ossl_message_log("\tInternalMetaMemoryChannel: %lu\n", pFreeList->m_FLData.m_HwData.m_InternalMetaMemoryChannel);
        ix_ossl_message_log("\tInternalMetaMemorySize:    %lu\n", 1 << pFreeList->m_FLData.m_HwData.m_InternalMetaElementShift);

    }

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

    ix_ossl_message_log("\tDataBaseAddress:       0x%p\n", pFreeList->m_pDataBaseAddress);
    ix_ossl_message_log("\tDataMemoryType:        ");

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

    switch(pFreeList->m_DataMemoryType)
    {
        case IX_MEMORY_TYPE_DRAM:
            ix_ossl_message_log("DRAM");
            break;
        case IX_MEMORY_TYPE_SRAM:
            ix_ossl_message_log("SRAM");
            break;
        case IX_MEMORY_TYPE_SCRATCH:
            ix_ossl_message_log("SCRATCH");
            break;
        default:
            ix_ossl_message_log("LOCAL");
            break;
    }

#else

    switch(pFreeList->m_DataMemoryType)
    {
        case IX_MEMORY_TYPE_DRAM_SOUTH:
            ix_ossl_message_log("South bank DRAM");
            break;
        case IX_MEMORY_TYPE_SRAM_MSG_SOUTH:
            ix_ossl_message_log("South bank MSG-SRAM");
            break;
        case IX_MEMORY_TYPE_DRAM_NORTH:
            ix_ossl_message_log("North bank aliased DRAM");
            break;
        case IX_MEMORY_TYPE_SRAM_NORTH:
            ix_ossl_message_log("North bank aliased SRAM");
            break;
        case IX_MEMORY_TYPE_SCRATCH:
            ix_ossl_message_log("North bank aliased SCRATCH");
            break;
        default:
            ix_ossl_message_log("Unknown");
            break;
    }

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

    ix_ossl_message_log("\n");
    ix_ossl_message_log("\tDataMemoryChannel:     %lu\n", pFreeList->m_DataMemoryChannel);
    ix_ossl_message_log("\tDataMemorySize:        %d\n", 1 << pFreeList->m_DataElementShift);

    ix_ossl_message_log("\tNumElements:           %lu\n", pFreeList->m_NumElements);

    if(arg_FreeListType == 0)
    {
        ix_ossl_message_log("\tSRAMQueueId:           %lu\n", pFreeList->m_FLData.m_HwData.m_SRAMQueueId);
    }

    ix_ossl_message_log("\n");
}



/**
 * NAME: ix_rm_buffer_print_meta
 *
 * DESCRIPTION: This function will print the meta data information for the passed buffer
 * 
 * @Param:  - IN arg_hBuffer - buffer whose meta data we would like to print
 *
 */
void ix_rm_buffer_print_meta(
                              ix_buffer_handle arg_hBuffer
                            )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_buffer_print_meta";

    ix_error err = IX_SUCCESS;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(IX_RM_BUFFER_GET_TYPE(arg_hBuffer) == IX_BUFFER_TYPE_HARDWARE)
    {
        /* hardware buffer */
        ix_hw_buffer_meta* pMeta;

        if(arg_hBuffer == IX_NULL_BUFFER_HANDLE)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_HANDLE_INVALID, s_aFunctionIdentifier, 
                        "the passed buffer handle is invalid!", 0);
            return;
        }

        {
            _ix_buffer_free_list* pFreeList = g_FLManager.m_aFreeListAddress[IX_RM_BUF_HW_GET_FL_ID(arg_hBuffer)];

            if(pFreeList == 0)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_INVALID, s_aFunctionIdentifier, 
                            "the associated free list for this buffer is invalid!", 0);
                return;
            }
        }

        pMeta = IX_RM_BUF_HW_GET_META_ADDRESS(arg_hBuffer);

        ix_ossl_message_log("Meta data for hardware buffer: 0x%08lx\n", arg_hBuffer);

#if defined(_IX_RM_SPLIT_META_DATA_)

        {
            ix_hw_internal_buffer_meta* pInternalMeta;

            pInternalMeta = IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(arg_hBuffer);

            ix_ossl_message_log("\tHwNext:        0x%08lx\n", pInternalMeta->m_HwNext);
            ix_ossl_message_log("\tReserved0:     0x%08lx\n", pMeta->m_Reserved0);

        }

#else

        ix_ossl_message_log("\tHwNext:        0x%08lx\n", pMeta->m_HwNext);

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

        ix_ossl_message_log("\tBufferSize:    %hu\n", pMeta->m_BufferSize);

        ix_ossl_message_log("\tOffset:        %hu\n", pMeta->m_Offset);

        ix_ossl_message_log("\tPacketSize:    %hu\n", pMeta->m_PacketSize);

        ix_ossl_message_log("\tFreeListID:    %lu\n", IX_RM_HW_META_GET_FREE_LIST_ID(pMeta->m_BufferInfo));
        ix_ossl_message_log("\tRxStat:        %lu\n", IX_RM_HW_META_GET_RX_STAT(pMeta->m_BufferInfo));
        ix_ossl_message_log("\tHeaderType:    %lu\n", IX_RM_HW_META_GET_HEADER_TYPE(pMeta->m_BufferInfo));

        ix_ossl_message_log("\tInputPort:     %hu\n", pMeta->m_InputPort);
        ix_ossl_message_log("\tOutputPort:    %hu\n", pMeta->m_OutputPort);

        ix_ossl_message_log("\tNextHopID:     %hu\n", pMeta->m_NextHopID);
        ix_ossl_message_log("\tFabricPort:    %hu\n", (ix_uint16)pMeta->m_FabricPort);
        ix_ossl_message_log("\tReserved1      0x%02hx\n", (ix_uint16)pMeta->m_Reserved1);
        ix_ossl_message_log("\tNextHopIDType: %hu\n", (ix_uint16)pMeta->m_NextHopIDType);

        {
            ix_uint32 temp;
            temp = pMeta->m_ColorID;
            ix_ossl_message_log("\tColorID:       %lu\n", temp);
            temp = pMeta->m_FlowID;
            ix_ossl_message_log("\tFlowID:        %lu\n", temp);
        }

        ix_ossl_message_log("\tClassID:       %hu\n", pMeta->m_ClassID);
        ix_ossl_message_log("\tReserved2:     0x%04hx\n", pMeta->m_Reserved2);

        ix_ossl_message_log("\tPacketNext:    0x%08lx\n", pMeta->m_PacketNext);

    }
    else
    {
        /* software buffer */
        ix_sw_buffer_meta* pMeta;
        _ix_buffer_free_list* pFreeList = 0;
        ix_uint32 flId = IX_RM_BUF_SW_GET_FL_ID(arg_hBuffer);

        /* check if the ID coresponds to a software free list */
        if(flId < IX_BUF_FIRST_SW_FL_ID)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_ID_INVALID, s_aFunctionIdentifier, 
                        "the free list ID coresponds to a hardware free list instead of a software one!", 0);
            return;
        }

        pFreeList = g_FLManager.m_aFreeListAddress[flId];

        pMeta = (void*)((ix_uint8*)pFreeList->m_pMetaBaseAddress + 
            (IX_RM_BUF_SW_GET_INDEX(arg_hBuffer) << pFreeList->m_MetaElementShift));

        ix_ossl_message_log("Meta data for software buffer: 0x%08lx\n", arg_hBuffer);
        ix_ossl_message_log("\tNext:          0x%08lx\n", pMeta->m_hNext);
        ix_ossl_message_log("\tSelf:          0x%08lx\n", pMeta->m_hSelf);
        ix_ossl_message_log("\tpData:         0x%p\n", pMeta->m_pData);
        ix_ossl_message_log("\tDataSize:      0x%lu\n\n", pMeta->m_DataSize);

    }
}


/**
 * NAME: ix_rm_buffer_print_data
 *
 * DESCRIPTION: This function will print the data for the passed buffer
 * 
 * @Param:  - IN arg_hBuffer - buffer whose meta data we would like to print
 *
 */
void ix_rm_buffer_print_data(
                              ix_buffer_handle arg_hBuffer
                            )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_buffer_print_data";

    ix_error err = IX_SUCCESS;
    ix_uint32 dataSize = 0;
    ix_uint32* pData = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(IX_RM_BUFFER_GET_TYPE(arg_hBuffer) == IX_BUFFER_TYPE_HARDWARE)
    {
        /* hardware buffer */

        if(arg_hBuffer == IX_NULL_BUFFER_HANDLE)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_HANDLE_INVALID, s_aFunctionIdentifier, 
                        "the passed buffer handle is invalid!", 0);
            return;
        }

        {
            _ix_buffer_free_list* pFreeList = g_FLManager.m_aFreeListAddress[IX_RM_BUF_HW_GET_FL_ID(arg_hBuffer)];

            if(pFreeList == 0)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_INVALID, s_aFunctionIdentifier, 
                            "the associated free list for this buffer is invalid!", 0);
                return;
            }

            dataSize = 1 << pFreeList->m_DataElementShift;
        }


        ix_ossl_message_log("Data for hardware buffer: 0x%08lx\n", arg_hBuffer);

    }
    else
    {
        /* software buffer */
        _ix_buffer_free_list* pFreeList = 0;
        ix_uint32 flId = IX_RM_BUF_SW_GET_FL_ID(arg_hBuffer);

        /* check if the ID coresponds to a software free list */
        if(flId < IX_BUF_FIRST_SW_FL_ID)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_ID_INVALID, s_aFunctionIdentifier, 
                        "the free list ID coresponds to a hardware free list instead of a software one!", 0);
            return;
        }

        pFreeList = g_FLManager.m_aFreeListAddress[flId];

        dataSize = 1 << pFreeList->m_DataElementShift;

        ix_ossl_message_log("Data for software buffer: 0x%08lx\n", arg_hBuffer);

    }

    err = ix_rm_buffer_get_data(arg_hBuffer, (void**)&pData);
    if(err != IX_SUCCESS)
        return;

    ix_ossl_message_log("\tStart address: 0x%p\n", pData);
    ix_ossl_message_log("\tSize:          %lu\n\n", dataSize);
    {
        ix_uint32 i,j;

        for(i = 0; i < dataSize; )
        {
            for(j = 0; (j < 4) && (i < dataSize) ; i++, j++, pData++)
            {
                ix_ossl_message_log("0x%08lx ", *pData);
            }
            ix_ossl_message_log("\n");
        }

    }
}


/**
 * NAME: ix_rm_buffer_print_debug_info
 *
 * DESCRIPTION: This function will print the debug info for the passed buffer
 * 
 * @Param:  - IN arg_hBuffer - buffer whose debug info we would like to print
 *
 */
void ix_rm_buffer_print_debug_info(
                                    ix_buffer_handle arg_hBuffer
                                  )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_buffer_print_debug_info";

    ix_error err = IX_SUCCESS;
    ix_buffer_debug_info* pDebugInfo;
    char* allocStatus;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    err =_ix_rm_buffer_debug_info_get(arg_hBuffer, &pDebugInfo);
    if(err != IX_SUCCESS)
    {
        ix_ossl_message_log("Could not retrieve the debug info for the buffer 0x%08lx!\n", arg_hBuffer);
        return;
    }
    
    ix_ossl_message_log("Debug info for buffer 0x%08lx!\n", arg_hBuffer);
    switch(pDebugInfo->m_LastCoreAllocationStatus)
    {
        case IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_FREE: 
            allocStatus = "buffer is in the free list";
            break;
        case IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_ALLOCATED_BY_CORE:
            allocStatus = "buffer has been allocated by core";
            break;
        case IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_RECEIVED_FROM_ME_AS_PACKET:
            allocStatus = "buffer has been sent to the core as a packet buffer";
            break;
        case IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_RECEIVED_FROM_ME_AS_MESSAGE:
            allocStatus = "buffer has been sent to the core as a packet buffer";
            break;
        default:
            allocStatus = "unknown";
            break;
    }
    ix_ossl_message_log("\tLast core allocation status: %s\n", allocStatus);
    ix_ossl_message_log("\tNumber of times sent as a packet buffer on the core: %lu\n", pDebugInfo->m_SendAsPacketCount);
    ix_ossl_message_log("\tNumber of times sent as a message buffer on the core: %lu\n", pDebugInfo->m_SendAsMessageCount);

    return;
}



#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

