/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      communication.c
 *
 * = DESCRIPTION
 *      This file will define functions of the exported Communication API interface. 
 *      The Resource Manager Communication API implements the mechanism to transport exception
 *      packets and control messages between the microblocks and their Core counterpart. All
 *      packet buffers exchanged between microblocks and Core counterparts have to go through
 *      Resource Manager. A callback interface and a polling (blocking and nonblocking) 
 *      interface are provided. Only one can be used at a time.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      4/26/2002 4:24:28 PM - creation time 
 * ============================================================================
 */

static const char s_aFileIdentifier[] = "$Id: communication.c,v 1.32 2003/10/06 20:35:29 ageorges Exp $";

#include "ix_ossl.h"
#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"

#include "rm/symbols.h"
#include "rm/internal/internal_support.h"

#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"

#include "rm/system.h"
#include "rm/internal/internal_system.h"

#include "rm/memory.h"

#include "rm/buffer.h"
#include "rm/internal/internal_buffer.h"

#include "rm/hardware.h"

#include "rm/communication.h"
#include "rm/internal/internal_communication.h"

#include "rm/remote_communication.h"
#include "rm/internal/internal_remote_communication.h"




/**
 * NAME: ix_rm_packet_set_receive_mode
 *
 * DESCRIPTION: This function will set the packet receive mode for the communication Id.
 *  That could be IX_COMM_ID_MODE_CALLBACK or IX_COMM_ID_MODE_GET_SELECT. The change from
 *  one mode to the other should be done explicitely. If this communication ID is in 
 *  get/select mode, all buffers in the internal queue will be dropped and all waiting tasks
 *  will be awaken. If the communication ID is  in callback mode, the registered callback 
 *  will be notified that it has been disconnected from the communication Id, and the the 
 *  mode will become get/select.
 *
 * @Param:  - IN arg_CommunicationId - the communication ID for which we are 
 *	setting the packet receive mode.
 * @Param:  - IN arg_ReceiveMode - packet receive mode. The two possible values are 
 *  IX_COMM_ID_MODE_CALLBACK for callback mode and IX_COMM_ID_MODE_GET_SELECT for 
 *  get/select mode
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_packet_set_receive_mode(
                                        ix_communication_id arg_CommunicationId, 
                                        ix_comm_id_mode arg_ReceiveMode
                                      )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_packet_set_receive_mode";

    ix_error err = IX_SUCCESS;

    ix_uint32 localId  = IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId);
    ix_end_point* pEndPoint = 0;
    ix_end_point_cc_data* pCCData = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_EndPointManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        if(localId > IX_COMM_LAST_UBLOCK_ID)
        {
            pEndPoint = &g_EndPointManager.m_aPacketEndPoint[localId];
            pCCData = &pEndPoint->m_EPData.m_CoreCompData;

#if defined( _IX_RM_MULTI_THREADED_ )

            /* lock the end point data */
            pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */
    
            /* change only if required */
            if((ix_comm_id_mode)(pCCData->m_Flags & IX_COMM_ID_RECEIVE_MODE_MASK) != arg_ReceiveMode)
            {
                ix_end_point_behaviour* pBehaviour = &pEndPoint->m_EPBehaviour;

                if(arg_ReceiveMode == IX_COMM_ID_MODE_GET_SELECT)
                {
                    /* we are in callback mode and want to get into get/select mode */

                    pBehaviour->m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)g_DefaultDataHandler;
                    pBehaviour->m_pExternalBehaviourContext = pCCData;
                    pBehaviour->m_pInternalBehaviourContext = 0;
                    pBehaviour->m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)_ix_rm_comm_data_store_mp;

                    IX_SET_BIT_FIELD32(pCCData->m_Flags, IX_COMM_ID_MODE_CALLBACK, 0, 0);     

#if defined(_IX_COMM_ID_NOTIFICATION_DISCONNECT_ENABLE_)

                    /* notify the previously registered callback about the disconnect */
                    if(pBehaviour->m_BehaviourHandler !=  (_ix_comm_ep_behaviour_handler)g_DefaultDataHandler)
                    {
                        err = pBehaviour->m_BehaviourHandler(IX_NULL_BUFFER_HANDLE, IX_COMM_ID_NOTIFICATION_DISCONNECT, pBehaviour->m_pExternalBehaviourContext);
                    }

#endif /* end defined(_IX_COMM_ID_NOTIFICATION_DISCONNECT_ENABLE_) */

                    /* by default we put the comm Id in the mp/mc mode */
                    IX_SET_BIT_FIELD32(pCCData->m_Flags, 1, 1, 1);/* set producer mode */
                    IX_SET_BIT_FIELD32(pCCData->m_Flags, 1, 2, 2);/* set consumer mode */
                    pCCData->m_Flags &= ~IX_COMM_ID_CC_ENABLED_MODE_MASK;
                    pCCData->m_CCNotificationContext = 0;
                }
                else
                {
                    /* we are in get/select mode and want to get into callback mode */
                    
                    pBehaviour->m_BehaviourHandler = (_ix_comm_ep_behaviour_handler) g_DefaultDataHandler;
                    pBehaviour->m_pExternalBehaviourContext = 0;
                    pBehaviour->m_pInternalBehaviourContext = 0;

                    if(!IX_EP_DATA_QUEUE_IS_EMPTY(&(pCCData->m_EPDataQueue)))
                    {
                        /* pass all queued elements to the newly installed handler */
                        err = _ix_rm_process_queue_entries(pCCData, g_DefaultDataHandler, 0);
                    }

                    IX_SET_BIT_FIELD32(pCCData->m_Flags, IX_COMM_ID_MODE_CALLBACK, 0, 0);

                    /* lock the access to the queue */
                    _IX_RM_EP_DATA_ACTION_LOCK(pCCData);

                    /* notify the receiveing queue to wake up all pending tasks */
                    _IX_RM_EP_DATA_WAIT_QUEUE_NOTIFY(pCCData->m_pReceiveNotificationQueue);                

                    /* unlock the access to the queue */
                    _IX_RM_EP_DATA_ACTION_UNLOCK(pCCData); 
                }
                /* set the receive mode flag */
                IX_SET_BIT_FIELD32(pCCData->m_Flags, arg_ReceiveMode, 0, 0);
            }

#if defined( _IX_RM_MULTI_THREADED_ )

            /* unlock the end point data */
            pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

        }
        else
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_UBLOCK, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change coresponds to an ublock!", err);
        }
    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_REMOTE, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change is remote!", err);
    }

    return err;
}


/**
 * NAME: ix_rm_message_set_receive_mode
 *
 * DESCRIPTION: This function will set the message receive mode for the communication Id.
 *  That could be IX_COMM_ID_MODE_CALLBACK or IX_COMM_ID_MODE_GET_SELECT. The change from
 *  one mode to the other should be done explicitely. If this communication ID is in 
 *  get/select mode, all buffers in the internal queue will be dropped and all waiting tasks
 *  will be awaken. If the communication ID is  in callback mode, the registered callback 
 *  will be notified that it has been disconnected from the communication Id, and the the 
 *  mode will become get/select.
 *
 * @Param:  - IN arg_CommunicationId - the communication ID for which we are 
 *	setting the message receive mode.
 * @Param:  - IN arg_ReceiveMode - message receive mode. The two possible values are 
 *  IX_COMM_ID_MODE_CALLBACK for callback mode and IX_COMM_ID_MODE_GET_SELECT for 
 *  get/select mode
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_message_set_receive_mode(
                                         ix_communication_id arg_CommunicationId, 
                                         ix_comm_id_mode arg_ReceiveMode
                                       )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_message_set_receive_mode";

    ix_error err = IX_SUCCESS;

    ix_uint32 localId  = IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId);
    ix_end_point* pEndPoint = 0;
    ix_end_point_cc_data* pCCData = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_EndPointManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        if(localId > IX_COMM_LAST_UBLOCK_ID)
        {
            pEndPoint = &g_EndPointManager.m_aMessageEndPoint[localId];
            pCCData = &pEndPoint->m_EPData.m_CoreCompData;

#if defined( _IX_RM_MULTI_THREADED_ )

            /* lock the end point data */
            pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */
    
            /* change only if required */
            if((ix_comm_id_mode)(pCCData->m_Flags & IX_COMM_ID_RECEIVE_MODE_MASK) != arg_ReceiveMode)
            {
                ix_end_point_behaviour* pBehaviour = &pEndPoint->m_EPBehaviour;

                if(arg_ReceiveMode == IX_COMM_ID_MODE_GET_SELECT)
                {
                    /* we are in callback mode and want to get into get/select mode */

                    pBehaviour->m_BehaviourHandler =  (_ix_comm_ep_behaviour_handler)g_DefaultDataHandler;
                    pBehaviour->m_pExternalBehaviourContext = pCCData;
                    pBehaviour->m_pInternalBehaviourContext = 0;
                    pBehaviour->m_BehaviourHandler =  (_ix_comm_ep_behaviour_handler)_ix_rm_comm_data_store_mp;

                    IX_SET_BIT_FIELD32(pCCData->m_Flags, IX_COMM_ID_MODE_CALLBACK, 0, 0);     
                   
#if defined(_IX_COMM_ID_NOTIFICATION_DISCONNECT_ENABLE_)

                    /* notify the previously registered callback about the disconnect */
                    if(pBehaviour->m_BehaviourHandler !=  (_ix_comm_ep_behaviour_handler)g_DefaultDataHandler)
                    {
                        err = pBehaviour->m_BehaviourHandler(IX_NULL_BUFFER_HANDLE, IX_COMM_ID_NOTIFICATION_DISCONNECT, pBehaviour->m_pExternalBehaviourContext);
                    }

#endif /* end defined(_IX_COMM_ID_NOTIFICATION_DISCONNECT_ENABLE_) */

                    /* by default we put the comm Id in the mp/mc mode */
                    IX_SET_BIT_FIELD32(pCCData->m_Flags, 1, 1, 1);/* set producer mode */
                    IX_SET_BIT_FIELD32(pCCData->m_Flags, 1, 2, 2);/* set consumer mode */
                    pCCData->m_Flags &= ~IX_COMM_ID_CC_ENABLED_MODE_MASK;
                    pCCData->m_CCNotificationContext = 0;
                }
                else
                {
                    /* we are in get/select mode and want to get into callback mode */
                    
                    pBehaviour->m_BehaviourHandler =  (_ix_comm_ep_behaviour_handler)g_DefaultDataHandler;
                    pBehaviour->m_pExternalBehaviourContext = 0;
                    pBehaviour->m_pInternalBehaviourContext = 0;


/**
 * 
 */
                    if(!IX_EP_DATA_QUEUE_IS_EMPTY(&pCCData->m_EPDataQueue))
                    {
                        /* pass all queued elements to the newly installed handler */
                        err = _ix_rm_process_queue_entries(pCCData, g_DefaultDataHandler, 0);
                    }
                    
                    IX_SET_BIT_FIELD32(pCCData->m_Flags, IX_COMM_ID_MODE_CALLBACK, 0, 0);

                    /* lock the access to the queue */
                    _IX_RM_EP_DATA_ACTION_LOCK(pCCData);

                    /* notify the receiveing queue to wake up all pending tasks */
                    _IX_RM_EP_DATA_WAIT_QUEUE_NOTIFY(pCCData->m_pReceiveNotificationQueue);                

                    /* unlock the access to the queue */
                    _IX_RM_EP_DATA_ACTION_UNLOCK(pCCData); 
                }
                /* set the receive mode flag */
                IX_SET_BIT_FIELD32(pCCData->m_Flags, arg_ReceiveMode, 0, 0);
            }

#if defined( _IX_RM_MULTI_THREADED_ )

            /* unlock the end point data */
            pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

        }
        else
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_UBLOCK, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change coresponds to an ublock!", err);
        }
    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_REMOTE, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change is remote!", err);
    }

    return err;
}


/**
 * The following functions apply just to the get/select receive mode
 */

/**
 * NAME: ix_rm_packet_set_consumer_mode
 *
 * DESCRIPTION: This function will set the packet consumer mode for the communication Id
 *  If arg_ConsumerMode is 0 then the communication Id will be set to single consumer 
 *  mode otherwise the multiple consumer mode will be set.
 *  If the comm Id is in IX_COMM_ID_MODE_CALLBACK receive mode then the function will
 *  return an error.
 *
 * @Param:  - IN arg_CommunicationId - the communication ID for which we are 
 *	setting the packet consumer mode.
 * @Param:  - IN arg_ConsumerMode - consumer mode for Comm Id. 0 for single consumer and
 *  different from 0 for multiple consumer.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_packet_set_consumer_mode(
                                         ix_communication_id arg_CommunicationId, 
                                         ix_uint32 arg_ConsumerMode
                                       )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_packet_set_consumer_mode";

    ix_error err = IX_SUCCESS;

    
    ix_uint32 localId  = IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId);
    ix_end_point* pEndPoint = 0;
    ix_end_point_cc_data* pCCData = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_EndPointManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        if(localId > IX_COMM_LAST_UBLOCK_ID)
        {
            pEndPoint = &g_EndPointManager.m_aPacketEndPoint[localId];
            pCCData = &pEndPoint->m_EPData.m_CoreCompData;

#if defined( _IX_RM_MULTI_THREADED_ )

            /* lock the end point data */
            pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

            if((pCCData->m_Flags & IX_COMM_ID_RECEIVE_MODE_MASK) == IX_COMM_ID_MODE_GET_SELECT)
            {
                if(arg_ConsumerMode != 0)
                {
                    /* multi consumer */
                    if((pCCData->m_Flags & IX_COMM_ID_CC_ENABLED_MODE_MASK) != 0)
                        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_CC_MODE_ENABLED, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change coresponds is has the CC mode enabled."
                            "Disable the CC mode and try again!", err);
                }

                /* set the consumer mode flag */
                IX_SET_BIT_FIELD32(pCCData->m_Flags, arg_ConsumerMode?1:0, 2, 2);
            }
            else
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_RECV_CALLBACK, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change is in callback mode!", err);
            }

#if defined( _IX_RM_MULTI_THREADED_ )

            /* unlock the end point data */
            pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

        }
        else
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_UBLOCK, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change coresponds to an ublock!", err);
        }
    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_REMOTE, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change is remote!", err);
    }

    return err;
}



/**
 * NAME: ix_rm_packet_set_producer_mode
 *
 * DESCRIPTION: This function will set the packet producer mode for the communication Id
 *  If arg_ConsumerMode is 0 then the communication Id will be set to single producer 
 *  mode otherwise the multiple producer mode will be set.
 *  If the comm Id is in IX_COMM_ID_MODE_CALLBACK receive mode then the function will
 *  return an error.
 *
 * @Param:  - IN arg_CommunicationId - the communication ID for which we are 
 *	setting the packet producer mode.
 * @Param:  - IN arg_ProducerMode - producer mode for Comm Id. 0 for single producer and
 *  different from 0 for multiple producer.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_packet_set_producer_mode(
                                         ix_communication_id arg_CommunicationId, 
                                         ix_uint32 arg_ProducerMode
                                       )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_packet_set_producer_mode";

    ix_error err = IX_SUCCESS;

    
    ix_uint32 localId  = IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId);
    ix_end_point* pEndPoint = 0;
    ix_end_point_cc_data* pCCData = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_EndPointManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        if(localId > IX_COMM_LAST_UBLOCK_ID)
        {
            pEndPoint = &g_EndPointManager.m_aPacketEndPoint[localId];
            pCCData = &pEndPoint->m_EPData.m_CoreCompData;

#if defined( _IX_RM_MULTI_THREADED_ )

            /* lock the end point data */
            pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

            if((pCCData->m_Flags & IX_COMM_ID_RECEIVE_MODE_MASK) == IX_COMM_ID_MODE_GET_SELECT)
            {
                /* set the producer mode flag */
                if(arg_ProducerMode != 0)
                {
                    /* multi producer */

                    /* set the multi producer behaviour function */
                    pEndPoint->m_EPBehaviour.m_BehaviourHandler =  (_ix_comm_ep_behaviour_handler)_ix_rm_comm_data_store_mp;
                }
                else
                {
                    /* single producer */

                    /* set the single producer behaviour function */
                    pEndPoint->m_EPBehaviour.m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)_ix_rm_comm_data_store_sp;
                }
                IX_SET_BIT_FIELD32(pCCData->m_Flags, arg_ProducerMode?1:0, 1, 1);
            }
            else
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_RECV_CALLBACK, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change is in callback mode!", err);
            }

#if defined( _IX_RM_MULTI_THREADED_ )

            /* unlock the end point data */
            pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

        }
        else
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_UBLOCK, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change coresponds to an ublock!", err);
        }
    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_REMOTE, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change is remote!", err);
    }

    return err;
}



/**
 * NAME: ix_rm_message_set_consumer_mode
 *
 * DESCRIPTION: This function will set the message consumer mode for the communication Id
 *  If arg_ConsumerMode is 0 then the communication Id will be set to single consumer 
 *  mode otherwise the multiple consumer mode will be set.
 *  If the comm Id is in IX_COMM_ID_MODE_CALLBACK receive mode then the function will
 *  return an error.
 *
 * @Param:  - IN arg_CommunicationId - the communication ID for which we are 
 *	setting the message consumer mode.
 * @Param:  - IN arg_ConsumerMode - consumer mode for Comm Id. 0 for single consumer and
 *  different from 0 for multiple consumer.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_message_set_consumer_mode(
                                          ix_communication_id arg_CommunicationId, 
                                          ix_uint32 arg_ConsumerMode
                                       )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_message_set_consumer_mode";

    ix_error err = IX_SUCCESS;

    
    ix_uint32 localId  = IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId);
    ix_end_point* pEndPoint = 0;
    ix_end_point_cc_data* pCCData = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_EndPointManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        if(localId > IX_COMM_LAST_UBLOCK_ID)
        {
            pEndPoint = &g_EndPointManager.m_aMessageEndPoint[localId];
            pCCData = &pEndPoint->m_EPData.m_CoreCompData;

#if defined( _IX_RM_MULTI_THREADED_ )

            /* lock the end point data */
            pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

            if((pCCData->m_Flags & IX_COMM_ID_RECEIVE_MODE_MASK) == IX_COMM_ID_MODE_GET_SELECT)
            {
                if(arg_ConsumerMode != 0)
                {
                    /* multi consumer */
                    if((pCCData->m_Flags & IX_COMM_ID_CC_ENABLED_MODE_MASK) != 0)
                        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_CC_MODE_ENABLED, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change coresponds is has the CC mode enabled."
                            "Disable the CC mode and try again!", err);
                }

                /* set the consumer mode flag */
                IX_SET_BIT_FIELD32(pCCData->m_Flags, arg_ConsumerMode?1:0, 2, 2);
            }
            else
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_RECV_CALLBACK, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change is in callback mode!", err);
            }

#if defined( _IX_RM_MULTI_THREADED_ )

            /* unlock the end point data */
            pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

        }
        else
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_UBLOCK, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change coresponds to an ublock!", err);
        }
    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_REMOTE, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change is remote!", err);
    }

    return err;
}


/**
 * NAME: ix_rm_message_set_producer_mode
 *
 * DESCRIPTION: This function will set the message producer mode for the communication Id
 *  If arg_ConsumerMode is 0 then the communication Id will be set to single producer 
 *  mode otherwise the multiple producer mode will be set.
 *  If the comm Id is in IX_COMM_ID_MODE_CALLBACK receive mode then the function will
 *  return an error.
 *
 * @Param:  - IN arg_CommunicationId - the communication ID for which we are 
 *	setting the message producer mode.
 * @Param:  - IN arg_ProducerMode - producer mode for Comm Id. 0 for single producer and
 *  different from 0 for multiple producer.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_message_set_producer_mode(
                                          ix_communication_id arg_CommunicationId, 
                                          ix_uint32 arg_ProducerMode
                                        )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_message_set_producer_mode";

    ix_error err = IX_SUCCESS;

    ix_uint32 localId  = IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId);
    ix_end_point* pEndPoint = 0;
    ix_end_point_cc_data* pCCData = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_EndPointManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        if(localId > IX_COMM_LAST_UBLOCK_ID)
        {
            pEndPoint = &g_EndPointManager.m_aMessageEndPoint[localId];
            pCCData = &pEndPoint->m_EPData.m_CoreCompData;

#if defined( _IX_RM_MULTI_THREADED_ )

            /* lock the end point data */
            pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

            if((pCCData->m_Flags & IX_COMM_ID_RECEIVE_MODE_MASK) == IX_COMM_ID_MODE_GET_SELECT)
            {
                /* set the producer mode flag */
                if(arg_ProducerMode != 0)
                {
                    /* multi producer */

                    /* set the multi producer behaviour function */
                    pEndPoint->m_EPBehaviour.m_BehaviourHandler =  (_ix_comm_ep_behaviour_handler)_ix_rm_comm_data_store_mp;
                }
                else
                {
                    /* single producer */

                    /* set the multi producer behaviour function */
                    pEndPoint->m_EPBehaviour.m_BehaviourHandler =  (_ix_comm_ep_behaviour_handler)_ix_rm_comm_data_store_sp;
                }
                IX_SET_BIT_FIELD32(pCCData->m_Flags, arg_ProducerMode?1:0, 1, 1);
            }
            else
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_RECV_CALLBACK, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change is in callback mode!", err);
            }

#if defined( _IX_RM_MULTI_THREADED_ )

            /* unlock the end point data */
            pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

        }
        else
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_UBLOCK, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change coresponds to an ublock!", err);
        }
    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_REMOTE, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change is remote!", err);
    }

    return err;
}


/**
 * NAME: ix_rm_packet_handler_register
 *
 * DESCRIPTION: This function will register a packet handler for the 
 *	specified communication ID. If the 
 *  _IX_COMM_ID_NOTIFICATION_DISCONNECT_ENABLE_ was defined at the 
 *  time of RM compilation, if another packet handler is in place, that 
 *	packet handler will be invoked with a IX_NULL_BUFFER_HANDLE as the 
 *	first parameter, and IX_COMM_ID_NOTIFICATION_DISCONNECT value as the 
 *	user data parameter. From this point on, the new registered handler 
 *	will be used for processing packets sent to this destination. 
 * 
 * @Param:  - IN arg_CommunicationId - the communication ID for which we are 
 *	registering the packet handler.
 * @Param:  - IN arg_PacketHandler - the packet handler that we want to 
 *	register.
 * @Param:  - IN arg_pContext - context data that will be passed to the 
 *	packet handler on each invocation. The programmer has to make sure that 
 *	the context data is valid as long the packet handler is registered.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_packet_handler_register( 
                                        ix_communication_id arg_CommunicationId, 
                                        ix_comm_data_handler arg_PacketHandler, 
                                        void* arg_pContext 
                                      )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_packet_handler_register";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    return _ix_rm_handler_register(arg_CommunicationId, IX_COMM_HANDLER_TYPE_PACKET, 
        (_ix_comm_ep_behaviour_handler)arg_PacketHandler, arg_pContext);
}



/**
 * NAME: ix_rm_packet_handler_unregister
 *
 * DESCRIPTION: This function will unregister the current packet handler for 
 *	the specified communication ID. From this point on, the packets sent to 
 *	this communication ID will undergo the default processing which at this 
 *	time is to drop the packet (the composing buffers will be returned to 
 *	the free list that owns them).
 * 
 * @Param:  - IN arg_CommunicationId - the communication ID for which we 
 *	unregister the packet handler.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_packet_handler_unregister( 
                                          ix_communication_id arg_CommunicationId 
                                        )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_packet_handler_unregister";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    return _ix_rm_handler_register(arg_CommunicationId, IX_COMM_HANDLER_TYPE_PACKET, 
        (_ix_comm_ep_behaviour_handler)g_DefaultDataHandler, 0);
}



/**
 * NAME: ix_rm_message_handler_register
 *
 * DESCRIPTION: This function will register a message handler for the 
 *	specified communication ID. If another message handler is in place, 
 *	that message handler will be invoked with a IX_NULL_BUFFER_HANDLE as 
 *	the first parameter, and IX_COMM_ID_NOTIFICATION_DISCONNECT value as 
 *	the user data parameter. From this point on, the new registered handler 
 *	will be used for processing messages sent to this destination.
 * 
 * @Param:  - IN arg_CommunicationId - the communication ID for which we are 
 *	registering the message handler.
 * @Param:  - IN arg_MessageHandler - the message handler that we want to 
 *	register.
 * @Param:  - IN arg_pContext - context data that will be passed to the 
 *	message handler on each invocation. The programmer has to make sure 
 *	that the context data is valid as long the message handler is 
 *	registered.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_message_handler_register( 
                                         ix_communication_id arg_CommunicationId, 
                                         ix_comm_data_handler arg_MessageHandler, 
                                         void* arg_pContext 
                                       )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_message_handler_register";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    return _ix_rm_handler_register(arg_CommunicationId, IX_COMM_HANDLER_TYPE_MESSAGE, 
        (_ix_comm_ep_behaviour_handler)arg_MessageHandler, arg_pContext);
}



/**
 * NAME: ix_rm_message_handler_unregister
 *
 * DESCRIPTION: This function will unregister the current message handler for 
 *	the specified communication ID. From this point on, the messages sent 
 *	to this communication ID will undergo the default processing which at 
 *	this time is to drop the message (the composing buffers will be 
 *	returned to the free list that owns them).
 *
 * @Param:  - IN arg_CommunicationId - the communication ID for which we 
 *	unregister the message handler.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_message_handler_unregister( 
                                           ix_communication_id arg_CommunicationId 
                                         )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_message_handler_unregister";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    return _ix_rm_handler_register(arg_CommunicationId, IX_COMM_HANDLER_TYPE_MESSAGE, 
        (_ix_comm_ep_behaviour_handler)g_DefaultDataHandler, 0);
}



/**
 * NAME: ix_rm_packet_send
 *
 * DESCRIPTION: This function will send a packet to the specified 
 *	communication ID. If the communication ID is in get/select mode and the 
 *	internal packet queue is full then an error will be returned specifying 
 *	that the internal queue is full.
 * 
 * @Param:  - IN arg_CommunicationId - the destination communication ID. 
 * @Param:  - IN arg_PacketBuffer - the first buffer handle from the packet 
 *	that will be sent to the specified communication ID.
 * @Param:  - IN arg_UserData - custom user data that will be passed to the 
 *	packet handler registered with the specified communication ID. This 
 *	argument can be used to pass extra information to the destination.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_packet_send( 
                            ix_communication_id arg_CommunicationId, 
                            ix_buffer_handle arg_PacketBuffer,
                            ix_uint32 arg_UserData
                          )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_packet_send";

    ix_end_point* pEndPoint = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined (_IX_RM_BUFFER_DEBUG_)

    /* Update the debug info */
    if(arg_PacketBuffer != IX_NULL_BUFFER_HANDLE)
    {
        ix_buffer_debug_info* pDebugInfo;
        if(_ix_rm_buffer_debug_info_get(arg_PacketBuffer, &pDebugInfo) == IX_SUCCESS)
        {
            pDebugInfo->m_SendAsPacketCount++;
        }

        if(IX_RM_BUF_GET_EOP(arg_PacketBuffer) == 0)
        {
            ix_buffer_handle hTempBuffer = arg_PacketBuffer;
            ix_buffer_handle hNextBuffer = IX_NULL_BUFFER_HANDLE;
            while(1)
            {

                hNextBuffer = IX_NULL_BUFFER_HANDLE;
                ix_rm_buffer_get_next(hTempBuffer, &hNextBuffer);
                if(hNextBuffer == IX_NULL_BUFFER_HANDLE)
                    break;

                if(_ix_rm_buffer_debug_info_get(hNextBuffer, &pDebugInfo) == IX_SUCCESS)
                {
                    pDebugInfo->m_SendAsMessageCount++;
                }


                if(IX_RM_BUF_GET_EOP(hNextBuffer) != 0)
                    break;

                hTempBuffer = hNextBuffer;
            }
        }
    }

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        pEndPoint = &g_EndPointManager.m_aPacketEndPoint[IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId)];
        return pEndPoint->m_EPBehaviour.m_BehaviourHandler(arg_PacketBuffer, arg_UserData,
           pEndPoint->m_EPBehaviour.m_pExternalBehaviourContext, IX_SEND_CALL_TYPE_NONBLOCK, 0);
    }
    else
    {
        pEndPoint = &g_RemoteCommManager.m_RemoteCommPacketEndPoint;
        return pEndPoint->m_EPBehaviour.m_BehaviourHandler(arg_PacketBuffer, arg_UserData,
           pEndPoint->m_EPBehaviour.m_pExternalBehaviourContext, arg_CommunicationId, IX_SEND_CALL_TYPE_NONBLOCK, 0);
    }
}



/**
 * NAME: ix_rm_packet_send_wait
 *
 * DESCRIPTION: This function will send a packet to the specified 
 *	communication ID. If the communication ID is in get/select mode and the 
 *	internal packet queue is full then the send function will wait for the 
 *	specified timeout for the internal queue to become available before 
 *	returning an error.
 * 
 * @Param:  - IN arg_CommunicationId - the destination communication ID. 
 * @Param:  - IN arg_PacketBuffer - the first buffer handle from the packet 
 *	that will be sent to the specified communication ID.
 * @Param:  - IN arg_UserData - custom user data that will be passed to the 
 *	packet handler registered with the specified communication ID. This 
 *	argument can be used to pass extra information to the destination.
 * @Param:  - IN arg_Timeout - timeout specifying for how long we will wait 
 *	for the case when the communication ID is in get/select mode for the 
 *	internal data queue to become available. It is expressed in 
 *	milliseconds.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_packet_send_wait( 
                                 ix_communication_id arg_CommunicationId, 
                                 ix_buffer_handle arg_PacketBuffer,
                                 ix_uint32 arg_UserData,
                                 ix_uint32 arg_Timeout
                               )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_packet_send_wait";

    ix_end_point* pEndPoint = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined (_IX_RM_BUFFER_DEBUG_)

    /* Update the debug info */
    if(arg_PacketBuffer != IX_NULL_BUFFER_HANDLE)
    {
        ix_buffer_debug_info* pDebugInfo;
        if(_ix_rm_buffer_debug_info_get(arg_PacketBuffer, &pDebugInfo) == IX_SUCCESS)
        {
            pDebugInfo->m_SendAsPacketCount++;
        }

        if(IX_RM_BUF_GET_EOP(arg_PacketBuffer) == 0)
        {
            ix_buffer_handle hTempBuffer = arg_PacketBuffer;
            ix_buffer_handle hNextBuffer = IX_NULL_BUFFER_HANDLE;
            while(1)
            {

                hNextBuffer = IX_NULL_BUFFER_HANDLE;
                ix_rm_buffer_get_next(hTempBuffer, &hNextBuffer);
                if(hNextBuffer == IX_NULL_BUFFER_HANDLE)
                    break;

                if(_ix_rm_buffer_debug_info_get(hNextBuffer, &pDebugInfo) == IX_SUCCESS)
                {
                    pDebugInfo->m_SendAsMessageCount++;
                }


                if(IX_RM_BUF_GET_EOP(hNextBuffer) != 0)
                    break;

                hTempBuffer = hNextBuffer;
            }
        }
    }

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        pEndPoint = &g_EndPointManager.m_aPacketEndPoint[IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId)];
        return pEndPoint->m_EPBehaviour.m_BehaviourHandler(arg_PacketBuffer, arg_UserData,
            pEndPoint->m_EPBehaviour.m_pExternalBehaviourContext, IX_SEND_CALL_TYPE_BLOCK, arg_Timeout);
    }
    else
    {
        pEndPoint = &g_RemoteCommManager.m_RemoteCommPacketEndPoint;
        return pEndPoint->m_EPBehaviour.m_BehaviourHandler(arg_PacketBuffer, arg_UserData,
            pEndPoint->m_EPBehaviour.m_pExternalBehaviourContext, arg_CommunicationId, IX_SEND_CALL_TYPE_BLOCK, arg_Timeout);
    }
}



/**
 * NAME: ix_rm_message_send
 *
 * DESCRIPTION: This function will send a message to the specified 
 *	communication ID. If the communication ID is in get/select mode and the 
 *	internal message queue is full then an error will be returned 
 *	specifying that the internal queue is full.
 * 
 * @Param:  - IN arg_CommunicationId - the destination communication ID. 
 * @Param:  - IN arg_MessageBuffer - the first buffer handle from the message 
 *	that will be sent to the specified communication ID.
 * @Param:  - IN arg_UserData - custom user data that will be passed to the 
 *	message handler registered with the specified communication ID. This 
 *	argument can be used to pass extra information to the destination.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_message_send( 
                             ix_communication_id arg_CommunicationId, 
                             ix_buffer_handle arg_MessageBuffer,
                             ix_uint32 arg_UserData
                           )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_message_send";

    ix_end_point* pEndPoint = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined (_IX_RM_BUFFER_DEBUG_)

    /* Update the debug info */
    if(arg_MessageBuffer != IX_NULL_BUFFER_HANDLE)
    {
        ix_buffer_debug_info* pDebugInfo;
        if(_ix_rm_buffer_debug_info_get(arg_MessageBuffer, &pDebugInfo) == IX_SUCCESS)
        {
            pDebugInfo->m_SendAsMessageCount++;
        }

        if(IX_RM_BUF_GET_EOP(arg_MessageBuffer) == 0)
        {
            ix_buffer_handle hTempBuffer = arg_MessageBuffer;
            ix_buffer_handle hNextBuffer = IX_NULL_BUFFER_HANDLE;
            while(1)
            {

                hNextBuffer = IX_NULL_BUFFER_HANDLE;
                ix_rm_buffer_get_next(hTempBuffer, &hNextBuffer);
                if(hNextBuffer == IX_NULL_BUFFER_HANDLE)
                    break;

                if(_ix_rm_buffer_debug_info_get(hNextBuffer, &pDebugInfo) == IX_SUCCESS)
                {
                    pDebugInfo->m_SendAsMessageCount++;
                }


                if(IX_RM_BUF_GET_EOP(hNextBuffer) != 0)
                    break;

                hTempBuffer = hNextBuffer;
            }
        }
    }

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        pEndPoint = &g_EndPointManager.m_aMessageEndPoint[IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId)];
        return pEndPoint->m_EPBehaviour.m_BehaviourHandler(arg_MessageBuffer, arg_UserData,
           pEndPoint->m_EPBehaviour.m_pExternalBehaviourContext, IX_SEND_CALL_TYPE_NONBLOCK, 0);
    }
    else
    {
        pEndPoint = &g_RemoteCommManager.m_RemoteCommMessageEndPoint;
        return pEndPoint->m_EPBehaviour.m_BehaviourHandler(arg_MessageBuffer, arg_UserData,
           pEndPoint->m_EPBehaviour.m_pExternalBehaviourContext, arg_CommunicationId, IX_SEND_CALL_TYPE_NONBLOCK, 0);
    }
}



/**
 * NAME: ix_rm_message_send_wait
 *
 * DESCRIPTION: This function will send a message to the specified 
 *	communication ID.If the communication ID is in get/select mode and the 
 *	internal message queue is full then the send function will wait for the 
 *	specified timeout for the internal queue to become available before 
 *	returning an error.
 * 
 * @Param:  - IN arg_CommunicationId - the destination communication ID. 
 * @Param:  - IN arg_MessageBuffer - the first buffer handle from the message 
 *	that will be sent to the specified communication ID.
 * @Param:  - IN arg_UserData - custom user data that will be passed to the 
 *	message handler registered with the specified communication ID. This 
 *	argument can be used to pass extra information to the destination.
 * @Param:  - IN arg_Timeout - timeout specifying for how long we will wait 
 *	for the case when the communication ID is in get/select mode for the 
 *	internal data queue to become available. It is expressed in 
 *	milliseconds.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_message_send_wait( 
                                  ix_communication_id arg_CommunicationId, 
                                  ix_buffer_handle arg_MessageBuffer,
                                  ix_uint32 arg_UserData,
                                  ix_uint32 arg_Timeout
                                )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_message_send_wait";

    ix_end_point* pEndPoint = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined (_IX_RM_BUFFER_DEBUG_)

    /* Update the debug info */
    if(arg_MessageBuffer != IX_NULL_BUFFER_HANDLE)
    {
        ix_buffer_debug_info* pDebugInfo;
        if(_ix_rm_buffer_debug_info_get(arg_MessageBuffer, &pDebugInfo) == IX_SUCCESS)
        {
            pDebugInfo->m_SendAsMessageCount++;
        }

        if(IX_RM_BUF_GET_EOP(arg_MessageBuffer) == 0)
        {
            ix_buffer_handle hTempBuffer = arg_MessageBuffer;
            ix_buffer_handle hNextBuffer = IX_NULL_BUFFER_HANDLE;
            while(1)
            {

                hNextBuffer = IX_NULL_BUFFER_HANDLE;
                ix_rm_buffer_get_next(hTempBuffer, &hNextBuffer);
                if(hNextBuffer == IX_NULL_BUFFER_HANDLE)
                    break;

                if(_ix_rm_buffer_debug_info_get(hNextBuffer, &pDebugInfo) == IX_SUCCESS)
                {
                    pDebugInfo->m_SendAsMessageCount++;
                }


                if(IX_RM_BUF_GET_EOP(hNextBuffer) != 0)
                    break;

                hTempBuffer = hNextBuffer;
            }
        }
    }

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        pEndPoint = &g_EndPointManager.m_aMessageEndPoint[IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId)];
        return pEndPoint->m_EPBehaviour.m_BehaviourHandler(arg_MessageBuffer, arg_UserData,
            pEndPoint->m_EPBehaviour.m_pExternalBehaviourContext, IX_SEND_CALL_TYPE_BLOCK, arg_Timeout);
    }
    else
    {
        pEndPoint = &g_RemoteCommManager.m_RemoteCommMessageEndPoint;
        return pEndPoint->m_EPBehaviour.m_BehaviourHandler(arg_MessageBuffer, arg_UserData,
            pEndPoint->m_EPBehaviour.m_pExternalBehaviourContext, arg_CommunicationId, IX_SEND_CALL_TYPE_BLOCK, arg_Timeout);
    }
}


/**
 * NAME: ix_rm_packet_peek
 *
 * DESCRIPTION: This function will retrieve the number of packets stored in
 *  the internal queue for the specified comm Id. The number might not be the 
 *  same after the function returns. If the comm Id is in IX_COMM_ID_MODE_CALLBACK
 *  receive mode, then an error will be returned.
 * 
 * @Param:  - IN arg_CommunicationId - the local communication ID from which 
 *	we want to get information. 
 * @Param:  - OUT arg_pAvailablePackets - address where the number of available
 *  packets will be stored.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for 
 *	failure.
 */
ix_error ix_rm_packet_peek(
                            ix_communication_id arg_CommunicationId, 
                            ix_uint32* arg_pAvailablePackets
                          )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_packet_peek";

    ix_error err = IX_SUCCESS;

    ix_uint32 localId  = IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId);
    ix_end_point* pEndPoint = 0;
    ix_end_point_data_queue* pCCDataQueue = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        if(localId > IX_COMM_LAST_UBLOCK_ID)
        {
            pEndPoint = &g_EndPointManager.m_aPacketEndPoint[localId];
            pCCDataQueue = &pEndPoint->m_EPData.m_CoreCompData.m_EPDataQueue;

            if((pEndPoint->m_EPData.m_CoreCompData.m_Flags & IX_COMM_ID_RECEIVE_MODE_MASK) == IX_COMM_ID_MODE_GET_SELECT)
            {
                *arg_pAvailablePackets = IX_EP_DATA_QUEUE_ELEMENTS_AVAILABLE(pCCDataQueue);
            }
            else
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_RECV_CALLBACK, s_aFunctionIdentifier, 
                            "the comm Id we are probing is in callback mode!", err);
            }

        }
        else
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_UBLOCK, s_aFunctionIdentifier, 
                            "the comm Id we are probing coresponds to an ublock!", err);
        }
    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_REMOTE, s_aFunctionIdentifier, 
                            "the comm Id we are probing is remote!", err);
    }

    return err;
}


/**
 * NAME: ix_rm_packet_get
 *
 * DESCRIPTION: This function will retrieve the next packet and associated 
 *	user data from a get/select communication ID. If the internal packet queue is empty 
 *  then an error will be returned immediately. If the comm Id is in IX_COMM_ID_MODE_CALLBACK
 *  receive mode then an error will be returned.
 * 
 * @Param:  - IN arg_CommunicationId - the local communication ID from which 
 *	we want to get data. 
 * @Param:  - IN arg_pPacketBuffer - address where the first buffer handle 
 *	from the packet from the head of the internal queue will be extracted.
 * @Param:  - IN arg_pUserData - address where the associated user data will 
 *	be stored.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for 
 *	failure.
 */
ix_error ix_rm_packet_get(
                           ix_communication_id arg_CommunicationId, 
                           ix_buffer_handle* arg_pPacketBuffer,
                           ix_uint32* arg_UserData
                         )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_packet_get";

    ix_error err = IX_SUCCESS;

    ix_uint32 localId  = IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId);
    ix_end_point* pEndPoint = 0;
    ix_end_point_cc_data* pCCData = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        if(localId > IX_COMM_LAST_UBLOCK_ID)
        {
            pEndPoint = &g_EndPointManager.m_aPacketEndPoint[localId];

            pCCData = &pEndPoint->m_EPData.m_CoreCompData;

            if((pCCData->m_Flags & IX_COMM_ID_RECEIVE_MODE_MASK) == IX_COMM_ID_MODE_GET_SELECT)
            {
                if((pCCData->m_Flags & IX_COMM_ID_CONSUMER_MODE_MASK) != 0)
                    /* multiple consumer mode */
                    err = _ix_rm_cc_dequeue_element_nonblock_mc(pCCData, arg_pPacketBuffer, arg_UserData);
                else
                    /* single consumer mode */
                    err = _ix_rm_cc_dequeue_element_nonblock_sc(pCCData, arg_pPacketBuffer, arg_UserData);
            }
            else
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_RECV_CALLBACK, s_aFunctionIdentifier, 
                            "the comm Id we are trying to get data from is in callback mode!", err);
            }
        }
        else
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_UBLOCK, s_aFunctionIdentifier, 
                            "the comm Id whose behaviour we are trying to get data from coresponds to an ublock!", err);
        }
    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_REMOTE, s_aFunctionIdentifier, 
                            "the comm Id whose behaviour we are trying to get data from is remote!", err);
    }

    return err;
}


/**
 * NAME: ix_rm_packet_get_wait
 *
 * DESCRIPTION: This function will retrieve the next packet and associated 
 *	user data from a get/select communication ID. If the comm Id is in IX_COMM_ID_MODE_CALLBACK
 *  receive mode then an error will be returned. If the internal packet queue is empty then the 
 *	function will wait for the specified timeout for the data to become 
 *	available before returning an error.
 * 
 * @Param:  - IN arg_CommunicationId - the local communication ID from which 
 *	we want to get data. 
 * @Param:  - IN arg_pPacketBuffer - address where the first buffer handle 
 *	from the packet from the head of the internal queue will be extracted.
 * @Param:  - IN arg_pUserData - address where the associated user data will 
 *	be stored.
 * @Param:  - IN arg_Timeout - timeout specifying for how long we will wait 
 *	for the case when the communication ID is in get/select mode for data to 
 *	become available. It is expressed in milliseconds.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for 
 *	failure.
 */
ix_error ix_rm_packet_get_wait(
                                ix_communication_id arg_CommunicationId, 
                                ix_buffer_handle* arg_pPacketBuffer,
                                ix_uint32* arg_UserData,
                                ix_uint32 arg_Timeout
                              )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_packet_get_wait";

    ix_error err = IX_SUCCESS;

    ix_uint32 localId  = IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId);
    ix_end_point* pEndPoint = 0;
    ix_end_point_cc_data* pCCData = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        if(localId > IX_COMM_LAST_UBLOCK_ID)
        {
            pEndPoint = &g_EndPointManager.m_aPacketEndPoint[localId];

            pCCData = &pEndPoint->m_EPData.m_CoreCompData;

            if((pCCData->m_Flags & IX_COMM_ID_RECEIVE_MODE_MASK) == IX_COMM_ID_MODE_GET_SELECT)
            {
                if((pCCData->m_Flags & IX_COMM_ID_CONSUMER_MODE_MASK) != 0)
                    /* multiple consumer mode */
                    err = _ix_rm_cc_dequeue_element_block_mc(pCCData, 
                        arg_pPacketBuffer, arg_UserData, arg_Timeout);
                else
                    /* single consumer mode */
                    err = _ix_rm_cc_dequeue_element_block_sc(pCCData, 
                        arg_pPacketBuffer, arg_UserData, arg_Timeout);            
            }
            else
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_RECV_CALLBACK, s_aFunctionIdentifier, 
                            "the comm Id we are trying to get data from is in callback mode!", err);
            }
        }
        else
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_UBLOCK, s_aFunctionIdentifier, 
                            "the comm Id whose behaviour we are trying to get data from coresponds to an ublock!", err);
        }
    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_REMOTE, s_aFunctionIdentifier, 
                            "the comm Id whose behaviour we are trying to get data from is remote!", err);
    }

    return err;
}



/**
 * NAME: ix_rm_message_peek
 *
 * DESCRIPTION: This function will retrieve the number of messages stored in
 *  the internal queue for the specified comm Id. The number might not be the 
 *  same after the function returns. If the comm Id is in IX_COMM_ID_MODE_CALLBACK
 *  receive mode, then an error will be returned.
 * 
 * @Param:  - IN arg_CommunicationId - the local communication ID from which 
 *	we want to get information. 
 * @Param:  - OUT arg_pAvailableMessages - address where the number of available
 *  messages will be stored.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for 
 *	failure.
 */
ix_error ix_rm_message_peek(
                            ix_communication_id arg_CommunicationId, 
                            ix_uint32* arg_pAvailableMessages
                          )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_message_peek";

    ix_error err = IX_SUCCESS;

    ix_uint32 localId  = IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId);
    ix_end_point* pEndPoint = 0;
    ix_end_point_data_queue* pCCDataQueue = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        if(localId > IX_COMM_LAST_UBLOCK_ID)
        {
            pEndPoint = &g_EndPointManager.m_aMessageEndPoint[localId];
            pCCDataQueue = &pEndPoint->m_EPData.m_CoreCompData.m_EPDataQueue;

            if((pEndPoint->m_EPData.m_CoreCompData.m_Flags & IX_COMM_ID_RECEIVE_MODE_MASK) == IX_COMM_ID_MODE_GET_SELECT)
            {
                *arg_pAvailableMessages = IX_EP_DATA_QUEUE_ELEMENTS_AVAILABLE(pCCDataQueue);
            }
            else
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_RECV_CALLBACK, s_aFunctionIdentifier, 
                            "the comm Id we are probing is in callback mode!", err);
            }
        }
        else
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_UBLOCK, s_aFunctionIdentifier, 
                            "the comm Id we are probing coresponds to an ublock!", err);
        }
    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_REMOTE, s_aFunctionIdentifier, 
                            "the comm Id we are probing is remote!", err);
    }

    return err;
}


/**
 * NAME: ix_rm_message_get
 *
 * DESCRIPTION: This function will retrieve the next message and associated 
 *	user data from a get/select communication ID. If the internal packet queue is empty 
 *  then an error will be returned immediately. If the comm Id is in IX_COMM_ID_MODE_CALLBACK
 *  receive mode then an error will be returned.
 * 
 * @Param:  - IN arg_CommunicationId - the local communication ID from which 
 *	we want to get data. 
 * @Param:  - IN arg_pMessageBuffer - address where the first buffer handle 
 *	from the message from the head of the internal queue will be extracted.
 * @Param:  - IN arg_pUserData - address where the associated user data will 
 *	be stored.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for 
 *	failure.
 */
ix_error ix_rm_message_get(
                            ix_communication_id arg_CommunicationId, 
                            ix_buffer_handle* arg_pMessageBuffer,
                            ix_uint32* arg_UserData
                          )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_essage_get";

    ix_error err = IX_SUCCESS;

    ix_uint32 localId  = IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId);
    ix_end_point* pEndPoint = 0;
    ix_end_point_cc_data* pCCData = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        if(localId > IX_COMM_LAST_UBLOCK_ID)
        {
            pEndPoint = &g_EndPointManager.m_aMessageEndPoint[localId];

            pCCData = &pEndPoint->m_EPData.m_CoreCompData;

            if((pCCData->m_Flags & IX_COMM_ID_RECEIVE_MODE_MASK) == IX_COMM_ID_MODE_GET_SELECT)
            {
                if((pCCData->m_Flags & IX_COMM_ID_CONSUMER_MODE_MASK) != 0)
                    /* multiple consumer mode */
                    err = _ix_rm_cc_dequeue_element_nonblock_mc(pCCData, arg_pMessageBuffer, arg_UserData);
                else
                    /* single consumer mode */
                    err = _ix_rm_cc_dequeue_element_nonblock_sc(pCCData, arg_pMessageBuffer, arg_UserData);
            }
            else
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_RECV_CALLBACK, s_aFunctionIdentifier, 
                            "the comm Id we are trying to get data from is in callback mode!", err);
            }
        }
        else
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_UBLOCK, s_aFunctionIdentifier, 
                            "the comm Id whose behaviour we are trying to get data from coresponds to an ublock!", err);
        }
    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_REMOTE, s_aFunctionIdentifier, 
                            "the comm Id whose behaviour we are trying to get data from is remote!", err);
    }
    return err;
}


/**
 * NAME: ix_rm_message_get_wait
 *
 * DESCRIPTION: This function will retrieve the next message and associated 
 *	user data from a get/select communication ID. If the comm Id is in IX_COMM_ID_MODE_CALLBACK
 *  receive mode then an error will be returned.If the internal message queue is empty then the 
 *	function will wait for the specified timeout for the data to become 
 *	available before returning an error.
 * 
 * @Param:  - IN arg_CommunicationId - the local communication ID from which 
 *	we want to get data. 
 * @Param:  - IN arg_pMessageBuffer - address where the first buffer handle 
 *	from the message from the head of the internal queue will be extracted.
 * @Param:  - IN arg_pUserData - address where the associated user data will 
 *	be stored.
 * @Param:  - IN arg_Timeout - timeout specifying for how long we will wait 
 *	for the case when the communication ID is in get/select mode for data to 
 *	become available. It is expressed in milliseconds.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for 
 *	failure.
 */
ix_error ix_rm_message_get_wait(
                                 ix_communication_id arg_CommunicationId, 
                                 ix_buffer_handle* arg_pMessageBuffer,
                                 ix_uint32* arg_UserData,
                                 ix_uint32 arg_Timeout
                               )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_message_get_wait";

    ix_error err = IX_SUCCESS;

    ix_uint32 localId  = IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId);
    ix_end_point* pEndPoint = 0;
    ix_end_point_cc_data* pCCData = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        if(localId > IX_COMM_LAST_UBLOCK_ID)
        {
            pEndPoint = &g_EndPointManager.m_aMessageEndPoint[localId];

            pCCData = &pEndPoint->m_EPData.m_CoreCompData;

            if((pCCData->m_Flags & IX_COMM_ID_RECEIVE_MODE_MASK) == IX_COMM_ID_MODE_GET_SELECT)
            {
                if((pCCData->m_Flags & IX_COMM_ID_CONSUMER_MODE_MASK) != 0)
                    /* multiple consumer mode */
                    err = _ix_rm_cc_dequeue_element_block_mc(pCCData, 
                        arg_pMessageBuffer, arg_UserData, arg_Timeout);
                else
                    /* single consumer mode */
                    err = _ix_rm_cc_dequeue_element_block_sc(pCCData, 
                        arg_pMessageBuffer, arg_UserData, arg_Timeout);            
            }
            else
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_RECV_CALLBACK, s_aFunctionIdentifier, 
                            "the comm Id we are trying to get data from is in callback mode!", err);
            }
        }
        else
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_UBLOCK, s_aFunctionIdentifier, 
                            "the comm Id whose behaviour we are trying to get data from coresponds to an ublock!", err);
        }
    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_REMOTE, s_aFunctionIdentifier, 
                            "the comm Id whose behaviour we are trying to get data from is remote!", err);
    }
    return err;
}


/**
 * NAME: ix_rm_comm_select
 *
 * DESCRIPTION: This function will wait on a local set of core component 
 *	communication IDs to have messages and/or data available. The 
 *	arg_SelectActionSet parameter is an array of 32 bit flags specifying 
 *	what kind of data we are waiting for on each of the specified 
 *	communication IDs. The first element of the array coresponds to the 
 *	first core communication ID. You can not wait for uBlock comm IDs for 
 *	data. The function will wait for data for a maximum time specified by 
 *	arg_Timeout parameter. On return the whole select action set will be 
 *	flushed and each mask will be updated to specify if data is available 
 *	or an error has happened on that communication ID.
 *  If one of the requested comm IDs is not in the IX_COMM_ID_MODE_GET_SELECT
 *  receive mode for the corresponding active flags then an error will be
 *  returned.
 * 
 * @Param:  - INOUT arg_SelectActionSet - array of 32bit flags representing 
 *	what kind of data we are waiting on packets/messages for each CC end 
 *	point. On return the flags will be cleared and will specify only the 
 *	data type that has been received on each local CC communication ID.
 * @Param:  - IN arg_Timeout - timeout specifying for how long we want to 
 *	wait for data to become available. It is expressed in milliseconds.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for 
 *	failure.
 */
ix_error ix_rm_comm_select( 
                            ix_comm_select_action_set arg_SelectActionSet, 
                            ix_uint32 arg_Timeout 
                          )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_comm_select";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    return _ix_rm_cc_select(arg_SelectActionSet, arg_Timeout);
}




/**
 * NAME: ix_rm_ublock_packet_comm_init
 *
 * DESCRIPTION: This function will initialize the packet communication from 
 *	the core to the microblock corresponding to the passed communication 
 *	Id. The passed hardware ring will be associated with the communication 
 *	Id, and from this point on all packets sent on this Id will be queued 
 *	into the specified ring.
 * 
 * @Param:  - IN arg_CommunicationId - the microblock communication Id of 
 *	interest
 * @Param:  - IN arg_hHwRing - the hardware ring on which packet 
 *	communication from the core to microblock will be performed.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for 
 *	failure.
 */
ix_error ix_rm_ublock_packet_comm_init(
                                        ix_communication_id arg_CommunicationId, 
                                        ix_hw_ring_handle arg_hHwRing
                                      )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_ublock_packet_comm_init";

    ix_error err = IX_SUCCESS;
    ix_uint32 localId  = IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId);

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        if(localId <= IX_COMM_LAST_UBLOCK_ID)
        {
            g_EndPointManager.m_aPacketEndPoint[localId].m_EPData.m_UBlockData.m_hHwRing = arg_hHwRing;

            err = _ix_rm_handler_register(arg_CommunicationId, IX_COMM_HANDLER_TYPE_PACKET, 
                (_ix_comm_ep_behaviour_handler)_ix_rm_comm_ublock_packet_handler, (void*)arg_hHwRing);
        }
        else
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_CC, s_aFunctionIdentifier, 
                            "the comm Id whose ring we are trying to set coresponds to core component!", err);
        }
    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_REMOTE, s_aFunctionIdentifier, 
                            "the comm Id whose behaviour we are trying to get data from is remote!", err);
    }

    return err;
}



/**
 * NAME: ix_rm_ublock_message_comm_init
 *
 * DESCRIPTION: This function will initialize the message communication from 
 *	the core to the microblock corresponding to the passed communication 
 *	Id. The passed hardware ring will be associated with the communication 
 *	Id, and from this point on all messages sent on this Id will be queued 
 *	into the specified ring.
 * 
 * @Param:  - IN arg_CommunicationId - the microblock communication Id of 
 *	interest
 * @Param:  - IN arg_hHwRing - the hardware ring on which message 
 *	communication from the core to microblock will be performed.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for 
 *	failure.
 */
ix_error ix_rm_ublock_message_comm_init(
                                         ix_communication_id arg_CommunicationId, 
                                         ix_hw_ring_handle arg_hHwRing
                                       )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_ublock_message_comm_init";

    ix_error err = IX_SUCCESS;
    ix_uint32 localId  = IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId);

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        if(localId <= IX_COMM_LAST_UBLOCK_ID)
        {
            g_EndPointManager.m_aPacketEndPoint[localId].m_EPData.m_UBlockData.m_hHwRing = arg_hHwRing;

            err = _ix_rm_handler_register(arg_CommunicationId, IX_COMM_HANDLER_TYPE_MESSAGE, 
                (_ix_comm_ep_behaviour_handler)_ix_rm_comm_ublock_message_handler, (void*)arg_hHwRing);
        }
        else
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_CC, s_aFunctionIdentifier, 
                            "the comm Id whose ring we are trying to set coresponds to core component!", err);
        }
    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_REMOTE, s_aFunctionIdentifier, 
                            "the comm Id whose behaviour we are trying to get data from is remote!", err);
    }

    return err;
}


/**
 * The following functions are for internal CCI support
 */

/* CCI support enabled */
#if !defined( _IX_RM_CCI_SUPPORT_EXCLUDE_ )

/**
 * NAME: ix_rm_packet_enable_cc_mode
 *
 * DESCRIPTION: This function will enable packet CC mode for the comm Id. This mode
 *  will work just in the single consumer mode. If the comm Id is in multiple consumer
 *  mode then an error will be returned. In this mode an internal specific CC processing
 *  function will be called when the comm Id is in single consumer get/select mode and the internal
 *  queue goes from empty to non empty state.
 * @Param:  - IN arg_CommunicationId - the communication ID for which we are 
 *	enabling the packet CC mode.
 * @Param:  - IN arg_pCCModeContext - context to be passed to the internal 
 *  specific CC processing function the comm Id is in single consumer get/select 
 *  mode and the internal queue goes from empty to non empty state
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_packet_enable_cc_mode(
                                      ix_communication_id arg_CommunicationId, 
                                      void* arg_pCCModeContext
                                    )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_packet_enable_cc_mode";

    ix_error err = IX_SUCCESS;

    
    ix_uint32 localId  = IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId);
    ix_end_point* pEndPoint = 0;
    ix_end_point_cc_data* pCCData = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_EndPointManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        if(localId > IX_COMM_LAST_UBLOCK_ID)
        {
            pEndPoint = &g_EndPointManager.m_aPacketEndPoint[localId];
            pCCData = &pEndPoint->m_EPData.m_CoreCompData;

#if defined( _IX_RM_MULTI_THREADED_ )

            /* lock the end point data */
            pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

            if((pCCData->m_Flags & IX_COMM_ID_RECEIVE_MODE_MASK) == IX_COMM_ID_MODE_GET_SELECT)
            {
                if((pCCData->m_Flags & IX_COMM_ID_CONSUMER_MODE_MASK) == 0)
                {
                    if((pCCData->m_Flags & IX_COMM_ID_CC_ENABLED_MODE_MASK) == 0)
                    {
                        pCCData->m_CCNotificationContext = arg_pCCModeContext;
                        pCCData->m_Flags |= IX_COMM_ID_CC_ENABLED_MODE_MASK;
                    }
                    else
                    {
                        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_CC_MODE_ENABLED, s_aFunctionIdentifier, 
                            "the cc mode is already enabled; first needs to be disabled!", err);
                    }
                }
                else
                {
                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_MULTIPLE_CONSUMER, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change is in multiple consumer mode!", err);
                }
            }
            else
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_RECV_CALLBACK, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change is in callback mode!", err);
            }

#if defined( _IX_RM_MULTI_THREADED_ )

            /* unlock the end point data */
            pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

        }
        else
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_UBLOCK, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change coresponds to an ublock!", err);
        }
    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_REMOTE, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change is remote!", err);
    }

    return err;
}


/**
 * NAME: ix_rm_packet_disable_cc_mode
 *
 * DESCRIPTION: This function will disable packet CC mode for the comm Id and 
 *              returns the context originally registered by the CCI so that
 *              the appropriate termination operations can be performed on it.
 * 
 * @Param:  - IN    arg_CommunicationId - the communication ID for which we are 
 *	                    disabling the packet CC mode. 
 * @Param:  - OUT   arg_ppContext - context registered by CCI
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_packet_disable_cc_mode(
                                       ix_communication_id arg_CommunicationId,
                                       void** arg_ppContext
                                     )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_packet_disable_cc_mode";

    ix_error err = IX_SUCCESS;

    
    ix_uint32 localId  = IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId);
    ix_end_point* pEndPoint = 0;
    ix_end_point_cc_data* pCCData = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_EndPointManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        if(localId > IX_COMM_LAST_UBLOCK_ID)
        {
            pEndPoint = &g_EndPointManager.m_aPacketEndPoint[localId];
            pCCData = &pEndPoint->m_EPData.m_CoreCompData;

#if defined( _IX_RM_MULTI_THREADED_ )

            /* lock the end point data */
            pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

            if((pCCData->m_Flags & IX_COMM_ID_RECEIVE_MODE_MASK) == IX_COMM_ID_MODE_GET_SELECT)
            {
                if((pCCData->m_Flags & IX_COMM_ID_CONSUMER_MODE_MASK) == 0)
                {
                    if((pCCData->m_Flags & IX_COMM_ID_CC_ENABLED_MODE_MASK) != 0)
                    {
                        pCCData->m_Flags &= ~IX_COMM_ID_CC_ENABLED_MODE_MASK;
                        *arg_ppContext = pCCData->m_CCNotificationContext;
                        pCCData->m_CCNotificationContext = 0;
                    }
                    else
                    {
                        *arg_ppContext = pCCData->m_CCNotificationContext;
                        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_CC_MODE_DISABLED, s_aFunctionIdentifier, 
                            "the cc mode is already disabled; first needs to be enabled!", err);
                    }
                }
                else
                {
                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_MULTIPLE_CONSUMER, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change is in multiple consumer mode!", err);
                }

            }
            else
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_RECV_CALLBACK, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change is in callback mode!", err);
            }

#if defined( _IX_RM_MULTI_THREADED_ )

            /* unlock the end point data */
            pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

        }
        else
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_UBLOCK, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change coresponds to an ublock!", err);
        }
    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_REMOTE, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change is remote!", err);
    }

    return err;
}



/**
 * NAME: ix_rm_message_enable_cc_mode
 *
 * DESCRIPTION: This function will enable message CC mode for the comm Id. This mode
 *  will work just in the single consumer mode. If the comm Id is in multiple consumer
 *  mode then an error will be returned. In this mode an internal specific CC processing
 *  function will be called when the comm Id is in single consumer get/select mode and the internal
 *  queue goes from empty to non empty state.
 * @Param:  - IN arg_CommunicationId - the communication ID for which we are 
 *	enabling the message CC mode.
 * @Param:  - IN arg_pCCModeContext - context to be passed to the internal 
 *  specific CC processing function the comm Id is in single consumer get/select 
 *  mode and the internal queue goes from empty to non empty state
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_message_enable_cc_mode(
                                       ix_communication_id arg_CommunicationId, 
                                       void* arg_pCCModeContext
                                     )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_message_enable_cc_mode";

    ix_error err = IX_SUCCESS;

    
    ix_uint32 localId  = IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId);
    ix_end_point* pEndPoint = 0;
    ix_end_point_cc_data* pCCData = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_EndPointManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        if(localId > IX_COMM_LAST_UBLOCK_ID)
        {
            pEndPoint = &g_EndPointManager.m_aMessageEndPoint[localId];
            pCCData = &pEndPoint->m_EPData.m_CoreCompData;

#if defined( _IX_RM_MULTI_THREADED_ )

            /* lock the end point data */
            pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

            if((pCCData->m_Flags & IX_COMM_ID_RECEIVE_MODE_MASK) == IX_COMM_ID_MODE_GET_SELECT)
            {
                if((pCCData->m_Flags & IX_COMM_ID_CONSUMER_MODE_MASK) == 0)
                {
                    if((pCCData->m_Flags & IX_COMM_ID_CC_ENABLED_MODE_MASK) == 0)
                    {
                        pCCData->m_CCNotificationContext = arg_pCCModeContext;
                        pCCData->m_Flags |= IX_COMM_ID_CC_ENABLED_MODE_MASK;
                    }
                    else
                    {
                        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_CC_MODE_ENABLED, s_aFunctionIdentifier, 
                            "the cc mode is already enabled; first needs to be disabled!", err);
                    }
                }
                else
                {
                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_MULTIPLE_CONSUMER, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change is in multiple consumer mode!", err);
                }
            }
            else
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_RECV_CALLBACK, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change is in callback mode!", err);
            }

#if defined( _IX_RM_MULTI_THREADED_ )

            /* unlock the end point data */
            pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

        }
        else
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_UBLOCK, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change coresponds to an ublock!", err);
        }
    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_REMOTE, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change is remote!", err);
    }

    return err;
}



/**
 * NAME: ix_rm_message_disable_cc_mode
 *
 * DESCRIPTION: This function will disable message CC mode for the comm Id and
 *              return the CC's context so that it can close its execution-
 *              control interface
 * 
 * @Param:  - IN    arg_CommunicationId - the communication ID for which we are 
 *	                    disabling the message CC mode. 
 * @Param:  - OUT   arg_ppContext - CCI context
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_message_disable_cc_mode(
                                        ix_communication_id arg_CommunicationId,
                                        void** arg_ppContext
                                      )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_message_disable_cc_mode";

    ix_error err = IX_SUCCESS;

    ix_uint32 localId  = IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId);
    ix_end_point* pEndPoint = 0;
    ix_end_point_cc_data* pCCData = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_EndPointManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        if(localId > IX_COMM_LAST_UBLOCK_ID)
        {
            pEndPoint = &g_EndPointManager.m_aMessageEndPoint[localId];
            pCCData = &pEndPoint->m_EPData.m_CoreCompData;

#if defined( _IX_RM_MULTI_THREADED_ )

            /* lock the end point data */
            pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

            if((pCCData->m_Flags & IX_COMM_ID_RECEIVE_MODE_MASK) == IX_COMM_ID_MODE_GET_SELECT)
            {
                if((pCCData->m_Flags & IX_COMM_ID_CONSUMER_MODE_MASK) == 0)
                {
                    if((pCCData->m_Flags & IX_COMM_ID_CC_ENABLED_MODE_MASK) != 0)
                    {
                        pCCData->m_Flags &= ~IX_COMM_ID_CC_ENABLED_MODE_MASK;
                        *arg_ppContext = pCCData->m_CCNotificationContext;
                        pCCData->m_CCNotificationContext = 0;
                    }
                    else
                    {
                        *arg_ppContext = pCCData->m_CCNotificationContext;
                        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_CC_MODE_DISABLED, s_aFunctionIdentifier, 
                            "the cc mode is already disabled; first needs to be enabled!", err);
                    }
                }
                else
                {
                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_MULTIPLE_CONSUMER, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change is in multiple consumer mode!", err);
                }

            }
            else
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_RECV_CALLBACK, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change is in callback mode!", err);
            }

#if defined( _IX_RM_MULTI_THREADED_ )

            /* unlock the end point data */
            pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

        }
        else
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_UBLOCK, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change coresponds to an ublock!", err);
        }
    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_REMOTE, s_aFunctionIdentifier, 
                            "the comm Id we are trying to change is remote!", err);
    }

    return err;
}

#endif /* !defined( _IX_RM_CCI_SUPPORT_EXCLUDE_ ) */

