/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      configuration_property.c
 *
 * = DESCRIPTION
 *      
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      4/25/2002 1:38:12 PM - creation time 
 * ============================================================================
 */

static const char s_FileIdentifier[] = "$Id: configuration_property.c,v 1.13 2003/09/17 17:05:52 ageorges Exp $";

#include "ix_ossl.h"

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"

#include "rm/symbols.h"
#include "rm/internal/internal_support.h"

#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"

#include "rm/system.h"
#include "rm/internal/internal_system.h"

#include "rm/memory.h"
#include "rm/buffer.h"
#include "rm/internal/internal_buffer.h"

#include "rm/hardware.h"
#include "rm/communication.h"

#include "rm/configuration_property.h"
#include "rm/internal/internal_configuration_property.h"



/**
 * NAME: ix_rm_cp_get_core_root
 *
 * DESCRIPTION: This function will retrieve the core property tree root.
 * 
 * @Return: Returns the core property tree root.
 */
ix_configuration_property_handle ix_rm_cp_get_core_root(void)
{
    return g_CPManager.m_hCoreRegistryRoot;
}


/**
 * NAME: ix_rm_cp_property_create
 *
 * DESCRIPTION: This function will create a new property entry at the specified level.
 *          At the parent level the property name should be unique.
 * 
 * @Param:  - IN arg_hParentProperty - Represents the property where the new property will
 *          be created. This value can be IX_CP_CORE_PROPERTY_ROOT in which case the new
 *          property will be created at the top level of the registry.
 * @Param:  - IN arg_pPropertyName - represents the name of the new property to be
 *          created at this level. The name should be unique at the parent property
 *          level or an error will be returned.
 * @Param:  - IN arg_PropertyOptions - represents the options to be applied to the
 *          new created property.
 * @Param:  - OUT arg_pProperty - represents the location where the handle of the new
 *          created property will be stored. In the case of failure this value
 *          will be set to IX_CP_NULL_PROPERTY upon return.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure. If the parent property
 *          is invalid then an error will be returned. If a property with the same name exists
 *          at the specified location, an error will be returned.
 */
ix_error ix_rm_cp_property_create(
                                   ix_configuration_property_handle arg_hParentProperty, 
                                   const char* arg_pPropertyName,
                                   ix_uint32 arg_PropertyOptions,
                                   ix_configuration_property_handle* arg_pProperty 
                                 )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_cp_property_create";

    ix_error err = IX_SUCCESS;

    ix_configuration_property_handle hNewCP;
    ix_cp_property_attributes* pCPAttributes;
    ix_cp_property_attributes* pCPParentAttributes;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_CPManager.m_SyncAdapter;        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check if the passed handle is not NULL */
    if(arg_hParentProperty == IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_CP_HANDLE_NULL, s_aFunctionIdentifier, 
                        "the passed configuration property handle NULL!", err);
    }

    /* check if the name is correct */
    if(arg_pPropertyName == 0)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_CP_NAME_NULL, s_aFunctionIdentifier, 
                        "the configuration property name is NULL!", err);
    }

    if(strlen(arg_pPropertyName) >= IX_CP_MAX_PROPERTY_NAME_LENGTH)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_CP_NAME_TOO_LONG, s_aFunctionIdentifier, 
                        "the configuration property name is too long!", err);
    }

    if(strchr(arg_pPropertyName,IX_CP_PATH_NAME_DELIMITER) != 0)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_CP_BAD_NAME, s_aFunctionIdentifier, 
                        "the configuration property name contains the path delimiter!", err);
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* look if that name already exists at this level */
    if(_ix_rm_cp_tree_children_forward_traverse(arg_hParentProperty, 
        (ix_cp_tree_predicate)_ix_cp_tree_predicate_has_name, (void*)arg_pPropertyName) != IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
    {
        err =  IX_RM_ERROR_NEW(IX_RM_ERROR_CP_ALREADY_EXISTS, s_aFunctionIdentifier, 
                        "a property with that name already exists at this node!", err);
        goto label_1;
    }

    /* create a new software buffer for the property */
    err =  ix_rm_buffer_alloc(g_CPManager.m_hCPFreeList, &hNewCP);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_CREATION_FAILED, s_aFunctionIdentifier, 
                        "failed to allocate a new configuration property!", err);
        goto label_1;
    }

    err = ix_rm_buffer_get_meta(hNewCP, (void**)&pCPAttributes);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                        "failed to access the new configuration property attributes!", err);
        goto label_2;
    }

    err = ix_rm_buffer_get_meta(arg_hParentProperty, (void**)&pCPParentAttributes);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                        "failed to access the parent configuration property attributes!", err);
        goto label_2;
    }

    pCPAttributes->m_DataSize = 0;
    pCPAttributes->m_hChildrenPropertiesListHead = 
        pCPAttributes->m_hChildrenPropertiesListTail = 
        IX_NULL_CONFIGURATION_PROPERTY_HANDLE;

    pCPAttributes->m_hPropertyNext = IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
    pCPAttributes->m_hPropertyPrev = pCPParentAttributes->m_hChildrenPropertiesListTail;
    pCPAttributes->m_hParentProperty = arg_hParentProperty;
    pCPAttributes->m_ReferenceCount = 0;
    pCPAttributes->m_pNotificationTokenList = 0;

    pCPAttributes->m_Options = arg_PropertyOptions;

    strcpy(pCPAttributes->m_aName, arg_pPropertyName);

    if(pCPParentAttributes->m_hChildrenPropertiesListTail != IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
    {
        err = ix_rm_buffer_get_meta(pCPParentAttributes->m_hChildrenPropertiesListTail, (void**)&pCPAttributes);
        if(err != IX_SUCCESS)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                            "failed to access the configuration property attributes!", err);
            goto label_2;
        }

        pCPAttributes->m_hPropertyNext = hNewCP;

        pCPParentAttributes->m_hChildrenPropertiesListTail = hNewCP;
    }
    else
    {
        pCPParentAttributes->m_hChildrenPropertiesListHead = 
            pCPParentAttributes->m_hChildrenPropertiesListTail = hNewCP;
    }

    *arg_pProperty = hNewCP;

    goto label_1;

label_2:

    /* free the sw buffer allocated for CP */
    ix_rm_buffer_free(hNewCP);

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}



/**
 * NAME: ix_rm_cp_property_delete
 *
 * DESCRIPTION: This function will delete the specified property from the registry.
 * 
 * @Param:  - IN arg_hProperty - this is the property to deleted from the registry.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_cp_property_delete(
                                   ix_configuration_property_handle arg_hProperty
                                 )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_cp_property_delete";

    ix_error err = IX_SUCCESS;

    ix_configuration_property_handle hCPParent;
    ix_cp_property_attributes* pCPAttributes;
    ix_cp_property_attributes* pCPParentAttributes;
    ix_cp_property_attributes* pCPOtherAttributes;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_CPManager.m_SyncAdapter;        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check if the passed handle is not NULL */
    if(arg_hProperty == IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_CP_HANDLE_NULL, s_aFunctionIdentifier, 
                        "the passed configuration property handle NULL!", err);
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* check if the call is done during notification */
    if(g_CPManager.m_NotificationInProgress != 0)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_NOTIFICATION_ACTIVE, s_aFunctionIdentifier, 
                        "the delete operation is forbidden during notification!", err);
        goto label_1;
    }

    err = ix_rm_buffer_get_meta(arg_hProperty, (void**)&pCPAttributes);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                        "failed to access the configuration property attributes!", err);
        goto label_1;
    }

    /* make sure the we are not trying to delete a constant property */
    if(pCPAttributes->m_Options & IX_CP_OPTIONS_FLAGS_CONSTANT)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_CONSTANT_DELETE, s_aFunctionIdentifier, 
                        "you are trying to delete one of the constant configuration properties!", err);
        goto label_1;
    }

    hCPParent = pCPAttributes->m_hParentProperty;

    err = ix_rm_buffer_get_meta(hCPParent, (void**)&pCPParentAttributes);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                        "failed to access the parent configuration property attributes!", err);
        goto label_1;
    }

    if(pCPAttributes->m_hPropertyPrev == IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
    {
        /* it is at the beginning of the parent's list */
        if(pCPAttributes->m_hPropertyNext == IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
        {
            /* it is the only property into the parent's list */
            pCPParentAttributes->m_hChildrenPropertiesListHead = 
                pCPParentAttributes->m_hChildrenPropertiesListTail = 
                IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
        }
        else
        {
            err = ix_rm_buffer_get_meta(pCPAttributes->m_hPropertyNext, (void**)&pCPOtherAttributes);
            if(err != IX_SUCCESS)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                                "failed to access the configuration property attributes!", err);
                goto label_1;
            }
            pCPOtherAttributes->m_hPropertyPrev = IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
            pCPParentAttributes->m_hChildrenPropertiesListHead = pCPAttributes->m_hPropertyNext;
        }
    }
    else 
    {
        if(pCPAttributes->m_hPropertyNext == IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
        {
            /* it is the last in the parent's list */
            err = ix_rm_buffer_get_meta(pCPAttributes->m_hPropertyPrev, (void**)&pCPOtherAttributes);
            if(err != IX_SUCCESS)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                                "failed to access the configuration property attributes!", err);
                goto label_1;
            }
            pCPOtherAttributes->m_hPropertyNext = IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
            pCPParentAttributes->m_hChildrenPropertiesListTail = pCPAttributes->m_hPropertyPrev;
        }
        else
        {
            /* is in between */
            ix_cp_property_attributes* pCPOtherAttributes1;

            /* it is the last in the parent's list */
            err = ix_rm_buffer_get_meta(pCPAttributes->m_hPropertyPrev, (void**)&pCPOtherAttributes);
            if(err != IX_SUCCESS)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                                "failed to access the configuration property attributes!", err);
                goto label_1;
            }

            err = ix_rm_buffer_get_meta(pCPAttributes->m_hPropertyNext, (void**)&pCPOtherAttributes1);
            if(err != IX_SUCCESS)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                                "failed to access the configuration property attributes!", err);
                goto label_1;
            }
            
            pCPOtherAttributes1->m_hPropertyPrev = pCPAttributes->m_hPropertyPrev;
            pCPOtherAttributes->m_hPropertyNext = pCPAttributes->m_hPropertyNext;
        }
    }

    /* delete the entire subtree starting from this root */
    _ix_rm_cp_tree_post_order_traverse(arg_hProperty, _ix_cp_tree_predicate_delete, 0);

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}




/**
 * NAME: ix_rm_cp_property_open
 *
 * DESCRIPTION: This function will retrieve the handle of the property named arg_pPropertyName
 *          at the level of arg_hParentProperty. Once successful the returned handle is a valid
 *          property handle that can be used for further work.
 * 
 * @Param:  - IN arg_hParentProperty - represents the level where we are looking for
 *          the arg_pPropertyName named property.
 * @Param:  - IN arg_pPropertyName - the name of the property we are looking for. The
 *          name can be a full path name in which case the search will go in the tree
 *          structure at the arg_hParentProperty level.
 * @Param:  - OUT arg_pProperty - represents the location where the required property
 *          handle will be stored. If the property does not exist at the specified level
 *          this value will be set to IX_CP_NULL_PROPERTY upon return.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_cp_property_open(
                                 ix_configuration_property_handle arg_hParentProperty, 
                                 const char* arg_pPropertyName,
                                 ix_configuration_property_handle* arg_pProperty 
                               )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_cp_property_open";

    ix_error err = IX_SUCCESS;

    char nodeLevelName[IX_CP_MAX_PROPERTY_NAME_LENGTH];
    ix_configuration_property_handle hCP;
    char* pDelim;
    ix_uint32 length;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_CPManager.m_SyncAdapter;        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    *arg_pProperty = IX_NULL_CONFIGURATION_PROPERTY_HANDLE;

    /* check if the passed handle is not NULL */
    if(arg_hParentProperty == IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_CP_HANDLE_NULL, s_aFunctionIdentifier, 
                        "the passed configuration property handle NULL!", err);
    }

    if(arg_pPropertyName == 0)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_CP_NAME_NULL, s_aFunctionIdentifier, 
                        "the configuration property name is NULL!", err);
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* look into the current subtree */
    hCP = arg_hParentProperty;

    while((hCP != IX_NULL_CONFIGURATION_PROPERTY_HANDLE) && (arg_pPropertyName != 0) && (*arg_pPropertyName != 0))
    {
        pDelim = strchr(arg_pPropertyName, IX_CP_PATH_NAME_DELIMITER);
        if(pDelim == 0)
            length = strlen(arg_pPropertyName);
        else
            length = pDelim - arg_pPropertyName;
        if(length >= IX_CP_MAX_PROPERTY_NAME_LENGTH)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_NAME_TOO_LONG, s_aFunctionIdentifier, 
                            "the configuration property name is too long!", err);
            goto label_1;
        }
        strncpy(nodeLevelName, arg_pPropertyName, length);
        nodeLevelName[length] = '\0';
        if(pDelim == 0)
            arg_pPropertyName = 0;
        else
            arg_pPropertyName += length + 1;
        hCP = _ix_rm_cp_tree_children_forward_traverse(hCP, (ix_cp_tree_predicate)_ix_cp_tree_predicate_has_name, nodeLevelName);
    }

    if(hCP != IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
    {
        ix_cp_property_attributes* pCPAttributes;
        err = ix_rm_buffer_get_meta(hCP, (void**)&pCPAttributes);
        if(err != IX_SUCCESS)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                            "failed to access the configuration property attributes!", err);
            goto label_1;
        }

        /* increment the reference count */
        pCPAttributes->m_ReferenceCount++;

    }

    *arg_pProperty = hCP;
    if(hCP == IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_PROPERTY_NOT_FOUND, s_aFunctionIdentifier, 
                            "the configuration property with the passed name has not been found at the parent level!", err);
        goto label_1;
    }

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: ix_rm_cp_property_close
 *
 * DESCRIPTION: This function will invalidate the property handle. From this point on
 *          the use of this handle in any operation will return an error.
 * 
 * @Param:  - IN arg_hProperty - the handle of the property that is to be closed.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_cp_property_close(
                                  ix_configuration_property_handle arg_hProperty
                                )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_cp_property_close";

    ix_error err = IX_SUCCESS;

    ix_cp_property_attributes* pCPAttributes;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_CPManager.m_SyncAdapter;        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check if the passed handle is not NULL */
    if(arg_hProperty == IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_CP_HANDLE_NULL, s_aFunctionIdentifier, 
                        "the passed configuration property handle NULL!", err);
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    err = ix_rm_buffer_get_meta(arg_hProperty, (void**)&pCPAttributes);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                        "failed to access the configuration property attributes!", err);
        goto label_1;
    }

    /* if the reference count goes to 0 then delete the property */
    if(pCPAttributes->m_ReferenceCount == 0)
    {
        ix_configuration_property_handle hCPParent;
        ix_cp_property_attributes* pCPParentAttributes;
        ix_cp_property_attributes* pCPOtherAttributes;

        /* check if the call is done during notification */
        if(g_CPManager.m_NotificationInProgress != 0)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_NOTIFICATION_ACTIVE, s_aFunctionIdentifier, 
                            "the delete operation is forbidden during notification!", err);
            goto label_1;
        }

        /* make sure the we are not trying to delete a constant property */
        if(pCPAttributes->m_Options & IX_CP_OPTIONS_FLAGS_CONSTANT)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_CONSTANT_DELETE, s_aFunctionIdentifier, 
                            "you are trying to delete one of the constant configuration properties!", err);
            goto label_1;
        }

        hCPParent = pCPAttributes->m_hParentProperty;

        err = ix_rm_buffer_get_meta(hCPParent, (void**)&pCPParentAttributes);
        if(err != IX_SUCCESS)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                            "failed to access the parent configuration property attributes!", err);
            goto label_1;
        }

        if(pCPAttributes->m_hPropertyPrev == IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
        {
            /* it is at the beginning of the parent's list */
            if(pCPAttributes->m_hPropertyNext == IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
            {
                /* it is the only property into the parent's list */
                pCPParentAttributes->m_hChildrenPropertiesListHead = 
                    pCPParentAttributes->m_hChildrenPropertiesListTail = 
                    IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
            }
            else
            {
                err = ix_rm_buffer_get_meta(pCPAttributes->m_hPropertyNext, (void**)&pCPOtherAttributes);
                if(err != IX_SUCCESS)
                {
                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                                    "failed to access the configuration property attributes!", err);
                    goto label_1;
                }
                pCPOtherAttributes->m_hPropertyPrev = IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
                pCPParentAttributes->m_hChildrenPropertiesListHead = pCPAttributes->m_hPropertyNext;
            }
        }
        else 
        {
            if(pCPAttributes->m_hPropertyNext == IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
            {
                /* it is the last in the parent's list */
                err = ix_rm_buffer_get_meta(pCPAttributes->m_hPropertyPrev, (void**)&pCPOtherAttributes);
                if(err != IX_SUCCESS)
                {
                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                                    "failed to access the configuration property attributes!", err);
                    goto label_1;
                }
                pCPOtherAttributes->m_hPropertyNext = IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
                pCPParentAttributes->m_hChildrenPropertiesListTail = pCPAttributes->m_hPropertyPrev;
            }
            else
            {
                /* is in between */
                ix_cp_property_attributes* pCPOtherAttributes1;

                /* it is the last in the parent's list */
                err = ix_rm_buffer_get_meta(pCPAttributes->m_hPropertyPrev, (void**)&pCPOtherAttributes);
                if(err != IX_SUCCESS)
                {
                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                                    "failed to access the configuration property attributes!", err);
                    goto label_1;
                }

                err = ix_rm_buffer_get_meta(pCPAttributes->m_hPropertyNext, (void**)&pCPOtherAttributes1);
                if(err != IX_SUCCESS)
                {
                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                                    "failed to access the configuration property attributes!", err);
                    goto label_1;
                }
            
                pCPOtherAttributes1->m_hPropertyPrev = pCPAttributes->m_hPropertyPrev;
                pCPOtherAttributes->m_hPropertyNext = pCPAttributes->m_hPropertyNext;
            }
        }

        /* delete the entire subtree starting from this root */
        _ix_rm_cp_tree_post_order_traverse(arg_hProperty, _ix_cp_tree_predicate_delete, 0);
    }
    else
        pCPAttributes->m_ReferenceCount--;

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}



/**
 * NAME: ix_rm_cp_property_attach
 *
 * DESCRIPTION: This function will register notification communication Id 
 *	with the property specified by arg_hProperty. When this property 
 *	changes, a message will be sent to the registered communication Id with 
 *	the handle of the property that has changed. It depends on the 
 *	communication Id setup when the message will be processed.
 * 
 * @Param:  - IN arg_hProperty - the handle of the property whose changes we 
 *	are interested in.
 * @Param:  - IN arg_hNotificationCommId - the communication Id where 
 *	notification events should be sent to.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for 
 *	failure. If the communication Id is already registered to receive 
 *	notification messages then an error will be returned.
 */
ix_error ix_rm_cp_property_attach(
                                   ix_configuration_property_handle arg_hProperty,
                                   ix_communication_id arg_hNotificationCommId 
                                 )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_cp_property_attach";

    ix_error err = IX_SUCCESS;

    ix_cp_property_attributes* pCPAttributes;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_CPManager.m_SyncAdapter;        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check if the passed handle is not NULL */
    if(arg_hProperty == IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_CP_HANDLE_NULL, s_aFunctionIdentifier, 
                        "the passed configuration property handle NULL!", err);
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* check if the call is done during notification */
    if((g_CPManager.m_NotificationInProgress != 0) && (g_CPManager.m_ActiveNotificationHandle == arg_hProperty))
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_NOTIFICATION_ACTIVE, s_aFunctionIdentifier, 
                        "the attach operation is forbidden during notification for the same property!", err);
        goto label_1;
    }

    err = ix_rm_buffer_get_meta(arg_hProperty, (void**)&pCPAttributes);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                        "failed to access the configuration property attributes!", err);
        goto label_1;
    }

    /* add the notification comm Id to the list */
    {
        ix_cp_property_notification_token* pNewToken;
        ix_cp_property_notification_token* pNotificationToken = pCPAttributes->m_pNotificationTokenList;
        if(pNotificationToken == 0)
        {
            pNewToken = ix_ossl_malloc(sizeof(ix_cp_property_notification_token));
            if(pNewToken == 0)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ALLOCATE_NOTIFICATION_TOKEN, s_aFunctionIdentifier, 
                    "failed to allocate a notification token!", err);
                goto label_1;
            }
            pNewToken->m_pNext = 0;
            pNewToken->m_CommunicationId = arg_hNotificationCommId;
            pCPAttributes->m_pNotificationTokenList = pNewToken;
        }
        else
        {
            while(pNotificationToken->m_pNext != 0)
            {
                if(pNotificationToken->m_CommunicationId == arg_hNotificationCommId)
                    break;
                pNotificationToken = pNotificationToken->m_pNext;
            }
            if(pNotificationToken->m_pNext == 0)
            {
                pNewToken = ix_ossl_malloc(sizeof(ix_cp_property_notification_token));
                if(pNewToken == 0)
                {
                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ALLOCATE_NOTIFICATION_TOKEN, s_aFunctionIdentifier, 
                        "failed to allocate a notification token!", err);
                    goto label_1;
                }

                pNewToken->m_pNext = 0;
                pNewToken->m_CommunicationId = arg_hNotificationCommId;
                pNotificationToken->m_pNext = pNewToken;
            }
        }
    }

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}




/**
 * NAME: ix_rm_cp_property_detach
 *
 * DESCRIPTION: This function will unregister the notification communication 
 *	Id with the given property referred by arg_hProperty.
 * 
 * @Param:  - IN arg_hProperty - the handle of the property of interest.
 * @Param:  - IN arg_hNotificationCommId - the communication Id that we want 
 *	to unregister from the property notification list.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for 
 *	failure. If the given communication Id is not registered for 
 *	notification with the specified property, then an error will be 
 *	returned.
 */
ix_error ix_rm_cp_property_detach(
                                   ix_configuration_property_handle arg_hProperty,
                                   ix_communication_id arg_hNotificationCommId 
                                 )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_cp_property_detach";

    ix_error err = IX_SUCCESS;

    ix_cp_property_attributes* pCPAttributes;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_CPManager.m_SyncAdapter;        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check if the passed handle is not NULL */
    if(arg_hProperty == IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_CP_HANDLE_NULL, s_aFunctionIdentifier, 
                        "the passed configuration property handle NULL!", err);
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* check if the call is done during notification */
    if((g_CPManager.m_NotificationInProgress != 0) && (g_CPManager.m_ActiveNotificationHandle == arg_hProperty))
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_NOTIFICATION_ACTIVE, s_aFunctionIdentifier, 
                        "the dettach operation is forbidden during notification for the same property!", err);
        goto label_1;
    }

    err = ix_rm_buffer_get_meta(arg_hProperty, (void**)&pCPAttributes);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                        "failed to access the configuration property attributes!", err);
        goto label_1;
    }

    /* remove the notification comm Id from the list */
    {
        ix_cp_property_notification_token* pLastToken = 0;
        ix_cp_property_notification_token* pCurrentToken = pCPAttributes->m_pNotificationTokenList;
        while(pCurrentToken != 0)
        {
            if(pCurrentToken->m_CommunicationId == arg_hNotificationCommId)
            {
                if(pLastToken == 0)
                    pCPAttributes->m_pNotificationTokenList = pCurrentToken->m_pNext;
                else
                    pLastToken->m_pNext = pCurrentToken->m_pNext;

                ix_ossl_free(pCurrentToken);

                break;
            }
            pLastToken = pCurrentToken;
            pCurrentToken = pCurrentToken->m_pNext;
        }
    }

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}




/**
 * NAME: ix_rm_cp_property_set_value
 *
 * DESCRIPTION: This function will associate a value with the property if none is
 *          already associated, and will set the property's value to the specified one.
 * 
 * @Param:  - IN arg_hProperty - the handle of the property whose value we want to set
 * @Param:  - IN arg_DataSize - the size in bytes of the data value.
 * @Param:  - IN arg_pData - address where the data of the value to be set resides.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_cp_property_set_value(
                                      ix_configuration_property_handle arg_hProperty,
                                      ix_uint32 arg_DataSize,
                                      const void* arg_pData
                                    )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_cp_property_set_value";

    ix_error err = IX_SUCCESS;

    ix_cp_property_attributes* pCPAttributes;
    ix_configuration_property_handle hCP;
    ix_configuration_property_handle hCPPrevious;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_CPManager.m_SyncAdapter;        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check if the passed handle is not NULL */
    if(arg_hProperty == IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_CP_HANDLE_NULL, s_aFunctionIdentifier, 
                        "the passed configuration property handle NULL!", err);
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* check if the call is done during notification */
    if(g_CPManager.m_NotificationInProgress != 0)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_NOTIFICATION_ACTIVE, s_aFunctionIdentifier, 
                        "the set operation is forbidden during notification for the same property!", err);
#if defined( _IX_RM_MULTI_THREADED_ )

        pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

        return err;
    }

    err = ix_rm_buffer_get_meta(arg_hProperty, (void**)&pCPAttributes);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                        "failed to access the configuration property attributes!", err);
        goto label_1;
    }

    /* make sure the we are not trying to change a constant property */
    if((pCPAttributes->m_Options & IX_CP_OPTIONS_FLAGS_CONSTANT) && 
        (pCPAttributes->m_Options & IX_CP_OPTIONS_FLAGS_DATA))
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_CONSTANT_SET_VALUE, s_aFunctionIdentifier, 
                        "you are trying to modify one of the constant configuration properties!", err);
        goto label_1;
    }

    /* make sure we free the buffer chain for the property */
    if((pCPAttributes->m_Options & IX_CP_OPTIONS_FLAGS_DATA) && 
        (!((pCPAttributes->m_Options & IX_CP_OPTIONS_FLAGS_DATA_UINT32) != 0)))
    {
        if(pCPAttributes->m_DataSize > IX_CP_DATA_ELEMENT_SIZE)
        {
            /* clear the EOP bit */
            IX_RM_BUF_SET_EOP(arg_hProperty, 0);

            /* unlink and free the chained buffers */
            ix_rm_buffer_unlink(&arg_hProperty, &hCP);

            ix_rm_buffer_free_chain(hCP);
        }
    }

    /* make sure the data fits into the property */
    if(arg_DataSize <= IX_CP_DATA_ELEMENT_SIZE)
    {
        /* copy the passed data */
        ix_ossl_memcpy(pCPAttributes->m_SwBufferMeta.m_pData, arg_pData, arg_DataSize);
    }
    else
    {
        ix_uint32 size = arg_DataSize;
        ix_cp_property_attributes* pCPOtherAttributes = pCPAttributes;

        hCPPrevious = arg_hProperty;
        
        while(size > IX_CP_DATA_ELEMENT_SIZE)
        {
             /* copy the passed data */
            ix_ossl_memcpy(pCPOtherAttributes->m_SwBufferMeta.m_pData, arg_pData, IX_CP_DATA_ELEMENT_SIZE);

            /* adjust the data pointer */
            arg_pData = (ix_uint8*)arg_pData + IX_CP_DATA_ELEMENT_SIZE;

            /* adjust the size */
            size -= IX_CP_DATA_ELEMENT_SIZE;

            /* allocate a new buffer */
            err =  ix_rm_buffer_alloc(g_CPManager.m_hCPFreeList, &hCP);
            if(err != IX_SUCCESS)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_CREATION_FAILED, s_aFunctionIdentifier, 
                                "failed to allocate a new configuration property!", err);
                goto label_2;
            }

            err = ix_rm_buffer_get_meta(hCP, (void**)&pCPOtherAttributes);
            if(err != IX_SUCCESS)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                                "failed to access the new configuration property attributes!", err);
                goto label_2;
            }

            /* link the buffers */
            err = ix_rm_buffer_link(&hCPPrevious, &hCP);
            if(err != IX_SUCCESS)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_LINK, s_aFunctionIdentifier, 
                                "failed to link two configuration properties!", err);
                goto label_2;
            }

            /* set the previous CP */
            hCPPrevious = hCP;
        }

        /* copy the passed data */
        ix_ossl_memcpy(pCPOtherAttributes->m_SwBufferMeta.m_pData, arg_pData, size);

    }
    /* adjust the data size */
    pCPAttributes->m_DataSize = arg_DataSize;

    /* adjust the flags */
    pCPAttributes->m_Options |= IX_CP_OPTIONS_FLAGS_DATA;
    pCPAttributes->m_Options &= ~IX_CP_OPTIONS_FLAGS_DATA_UINT32;

    /* set the notification fields */
    if(pCPAttributes->m_pNotificationTokenList)
    {
        g_CPManager.m_NotificationInProgress = 1;
        g_CPManager.m_ActiveNotificationHandle = arg_hProperty;
    }

    goto label_1;

label_2:

    /* unlink and free the chained buffers */
    ix_rm_buffer_unlink(&arg_hProperty, &hCP);

    ix_rm_buffer_free_chain(hCP);

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* notify all interested parts about the change */
    if(g_CPManager.m_NotificationInProgress != 0)
    {
        ix_cp_property_notification_token* pNotificationToken = pCPAttributes->m_pNotificationTokenList;
        while(pNotificationToken != 0)
        {
            ix_rm_message_send(pNotificationToken->m_CommunicationId, 0, arg_hProperty);
            pNotificationToken = pNotificationToken->m_pNext;
        }
        g_CPManager.m_NotificationInProgress = 0;
        g_CPManager.m_ActiveNotificationHandle = IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
    }

    return err;
}




/**
 * NAME: ix_rm_cp_property_get_value
 *
 * DESCRIPTION: This function will retrieve the value of the specified property if a
 *          a value is associated with the property.
 * 
 * @Param:  - IN arg_hProperty - the handle of the property whose value we want to retrieve.
 * @Param:  - INOUT arg_pDataSize - on input specifies the size of the buffer addressed by
 *          arg_pData. On output specifies the number of bytes copied to the buffer.
 * @Param:  - OUT arg_pData - location where the property's value should be copied. 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
IX_EXPORT_FUNCTION 
ix_error ix_rm_cp_property_get_value(
                                      ix_configuration_property_handle arg_hProperty,
                                      ix_uint32* arg_pDataSize,
                                      void* arg_pData
                                    )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_cp_property_get_value";

    ix_error err = IX_SUCCESS;

    ix_cp_property_attributes* pCPAttributes;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_CPManager.m_SyncAdapter;        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check if the passed handle is not NULL */
    if(arg_hProperty == IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_CP_HANDLE_NULL, s_aFunctionIdentifier, 
                        "the passed configuration property handle NULL!", err);
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    err = ix_rm_buffer_get_meta(arg_hProperty, (void**)&pCPAttributes);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                        "failed to access the configuration property attributes!", err);
        goto label_1;
    }

    if((pCPAttributes->m_Options & IX_CP_OPTIONS_FLAGS_DATA) == 0)
    {
        /* no data associated with this CP */
        *arg_pDataSize = 0;
        
        goto label_1;
    }

    if(pCPAttributes->m_Options & IX_CP_OPTIONS_FLAGS_DATA_UINT32)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_DATA_UINT32, s_aFunctionIdentifier, 
                        "the data ofthis property is of type ix_uint32 instead of opaque!", err);
        goto label_1;
    }

    if(pCPAttributes->m_DataSize > *arg_pDataSize)
    {
        /* set the required size */
        *arg_pDataSize = pCPAttributes->m_DataSize;

        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_BUFFER_TOO_SMALL, s_aFunctionIdentifier, 
                        "the copy buffer for the configuration property value is too small!", err);
        goto label_1;
    }

    *arg_pDataSize = 0;

    /* transfer the data */
    {
        ix_uint8* pData;
        ix_uint8* pDataTransfer = arg_pData;
        ix_uint32 size = pCPAttributes->m_DataSize;

        if(size > IX_CP_DATA_ELEMENT_SIZE)
        {
            /* clear the EOP bit */
            IX_RM_BUF_SET_EOP(arg_hProperty, 0);
        }

        while(size > IX_CP_DATA_ELEMENT_SIZE)
        {
            pData = (ix_uint8*)pCPAttributes->m_SwBufferMeta.m_pData;

            ix_ossl_memcpy(pDataTransfer, pData, IX_CP_DATA_ELEMENT_SIZE);

            err = ix_rm_buffer_get_next(arg_hProperty, &arg_hProperty);
            if(err != IX_SUCCESS)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_BUFFER_GET_NEXT_FAILED, s_aFunctionIdentifier, 
                        "failed to get next buffer for the configuration property!", err);
                goto label_1;
            }

            err = ix_rm_buffer_get_meta(arg_hProperty, (void**)&pCPAttributes);
            if(err != IX_SUCCESS)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                                "failed to access the configuration property attributes!", err);
                goto label_1;
            }

            pDataTransfer += IX_CP_DATA_ELEMENT_SIZE;

            size -= IX_CP_DATA_ELEMENT_SIZE;
        }

        pData = (ix_uint8*)pCPAttributes->m_SwBufferMeta.m_pData;

        ix_ossl_memcpy(pDataTransfer, pData, size);
    }

    *arg_pDataSize = pCPAttributes->m_DataSize;

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: ix_rm_cp_property_set_value_uint32
 *
 * DESCRIPTION: This function will associate a value with the property if none is
 *          already associated, and will set the property's value as a uint32 value
 *          to the specified value.
 * 
 * @Param:  - IN arg_hProperty - the handle of the property whose value we want to set
 * @Param:  - IN arg_Value - the new value for the property.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_cp_property_set_value_uint32(
                                             ix_configuration_property_handle arg_hProperty,
                                             ix_uint32 arg_Value
                                           )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_cp_property_set_value_uint32";

    ix_error err = IX_SUCCESS;

    ix_cp_property_attributes* pCPAttributes;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_CPManager.m_SyncAdapter;        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check if the passed handle is not NULL */
    if(arg_hProperty == IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_CP_HANDLE_NULL, s_aFunctionIdentifier, 
                        "the passed configuration property handle NULL!", err);
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* check if the call is done during notification */
    if(g_CPManager.m_NotificationInProgress != 0)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_NOTIFICATION_ACTIVE, s_aFunctionIdentifier, 
                        "the set operation is forbidden during notification for the same property!", err);

#if defined( _IX_RM_MULTI_THREADED_ )

        pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

        return err;
    }

    err = ix_rm_buffer_get_meta(arg_hProperty, (void**)&pCPAttributes);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                        "failed to access the configuration property attributes!", err);
        goto label_1;
    }

    /* make sure the we are not trying to change a constant property */
    if((pCPAttributes->m_Options & IX_CP_OPTIONS_FLAGS_CONSTANT) && 
        (pCPAttributes->m_Options & IX_CP_OPTIONS_FLAGS_DATA))
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_CONSTANT_SET_VALUE, s_aFunctionIdentifier, 
                        "you are trying to modify one of the constant configuration properties!", err);
        goto label_1;
    }

    /* make sure we free the buffer chain for the property */
    if((pCPAttributes->m_Options & IX_CP_OPTIONS_FLAGS_DATA) && 
        (!((pCPAttributes->m_Options & IX_CP_OPTIONS_FLAGS_DATA_UINT32) != 0)))
    {
        ix_configuration_property_handle hCP;

        if(pCPAttributes->m_DataSize > IX_CP_DATA_ELEMENT_SIZE)
        {
            /* clear the EOP bit */
            IX_RM_BUF_SET_EOP(arg_hProperty, 0);

            /* unlink and free the chained buffers */
            ix_rm_buffer_unlink(&arg_hProperty, &hCP);

            ix_rm_buffer_free_chain(hCP);
        }
    }

    *(ix_uint32*)pCPAttributes->m_SwBufferMeta.m_pData = arg_Value;

    /* adjust the data size */
    pCPAttributes->m_DataSize = sizeof(ix_uint32);

    /* adjust the flags */
    pCPAttributes->m_Options |= IX_CP_OPTIONS_FLAGS_DATA;
    pCPAttributes->m_Options |= IX_CP_OPTIONS_FLAGS_DATA_UINT32;

    /* set the notification fields */
    if(pCPAttributes->m_pNotificationTokenList)
    {
        g_CPManager.m_NotificationInProgress = 1;
        g_CPManager.m_ActiveNotificationHandle = arg_hProperty;
    }

    goto label_1;

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* notify all interested parts about the change */
    if(g_CPManager.m_NotificationInProgress != 0)
    {
        ix_cp_property_notification_token* pNotificationToken = pCPAttributes->m_pNotificationTokenList;
        while(pNotificationToken != 0)
        {
            ix_rm_message_send(pNotificationToken->m_CommunicationId, 0, arg_hProperty);
            pNotificationToken = pNotificationToken->m_pNext;
        }
        g_CPManager.m_NotificationInProgress = 0;
        g_CPManager.m_ActiveNotificationHandle = IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
    }

    return err;
}




/**
 * NAME: ix_rm_cp_property_get_value_uint32
 *
 * DESCRIPTION: This function will retrieve the value of the specified property in the case
 *          property value type is uint32. If the value has a different type then an error will be
 *          returned.
 * 
 * @Param:  - IN arg_hProperty - the handle of the property whose value we want to retrieve.
 * @Param:  - OUT arg_pValue - the location where uint32 value of the property should be stored. 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_cp_property_get_value_uint32(
                                             ix_configuration_property_handle arg_hProperty,
                                             ix_uint32* arg_pValue
                                           )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_cp_property_get_value_uint32";

    ix_error err = IX_SUCCESS;

    ix_cp_property_attributes* pCPAttributes;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_CPManager.m_SyncAdapter;        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check if the passed handle is not NULL */
    if(arg_hProperty == IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_CP_HANDLE_NULL, s_aFunctionIdentifier, 
                        "the passed configuration property handle NULL!", err);
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    err = ix_rm_buffer_get_meta(arg_hProperty, (void**)&pCPAttributes);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                        "failed to access the configuration property attributes!", err);
        goto label_1;
    }

    if((pCPAttributes->m_Options & IX_CP_OPTIONS_FLAGS_DATA) == 0)
    {
        *arg_pValue = 0xffffffff;

        goto label_1;
    }

    if((pCPAttributes->m_Options & IX_CP_OPTIONS_FLAGS_DATA_UINT32) == 0)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_DATA_OPAQUE, s_aFunctionIdentifier, 
                        "the data of this property is of opaque type instead of ix_uint32!", err);
        goto label_1;
    }

    *arg_pValue = *(ix_uint32*)pCPAttributes->m_SwBufferMeta.m_pData;

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}





/**
 * NAME: ix_rm_cp_property_delete_value
 *
 * DESCRIPTION: This function will delete the value of the property. From this point
 *          on, this property will have no value associated with it.
 * 
 * @Param:  - IN arg_hProperty - the handle of the property whose valeu we want to delete.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_cp_property_delete_value(
                                         ix_configuration_property_handle arg_hProperty
                                       )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_cp_property_delete_value";

    ix_error err = IX_SUCCESS;

    ix_cp_property_attributes* pCPAttributes;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_CPManager.m_SyncAdapter;        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check if the passed handle is not NULL */
    if(arg_hProperty == IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_CP_HANDLE_NULL, s_aFunctionIdentifier, 
                        "the passed configuration property handle NULL!", err);
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    err = ix_rm_buffer_get_meta(arg_hProperty, (void**)&pCPAttributes);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                        "failed to access the configuration property attributes!", err);
        goto label_1;
    }

    if(pCPAttributes->m_Options & IX_CP_OPTIONS_FLAGS_DATA)
    {
        ix_uint32 size = pCPAttributes->m_DataSize;
        pCPAttributes->m_Options &=  ~IX_CP_OPTIONS_FLAGS_DATA;
        pCPAttributes->m_DataSize = 0;

        if(pCPAttributes->m_Options & IX_CP_OPTIONS_FLAGS_DATA_UINT32)
        {
            pCPAttributes->m_Options &= ~IX_CP_OPTIONS_FLAGS_DATA_UINT32;
        }
        else
        {
            /* unlink and free the chained buffers */
            ix_configuration_property_handle hCP = IX_NULL_CONFIGURATION_PROPERTY_HANDLE;

            if(size > IX_CP_DATA_ELEMENT_SIZE)
            {
                /* clear the EOP bit */
                IX_RM_BUF_SET_EOP(arg_hProperty, 0);

                /* unlink and free the chained buffers */
                ix_rm_buffer_unlink(&arg_hProperty, &hCP);

                ix_rm_buffer_free_chain(hCP);
            }
        }
    }

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}



/**
 * NAME: ix_rm_cp_property_get_info
 *
 * DESCRIPTION: This function will return the information associated with the property.
 * 
 * @Param:  - IN arg_hProperty - the handle of the property whose attributes we want to
 *          retrieve.
 * @Param:  - OUT arg_pInfo - address where property's information will be stored.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_cp_property_get_info(
                                     ix_configuration_property_handle arg_hProperty, 
                                     ix_cp_property_info* arg_pInfo 
                                   )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_cp_property_get_info";

    ix_error err = IX_SUCCESS;

    ix_cp_property_attributes* pCPAttributes;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_CPManager.m_SyncAdapter;        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check if the passed handle is not NULL */
    if(arg_hProperty == IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_CP_HANDLE_NULL, s_aFunctionIdentifier, 
                        "the passed configuration property handle NULL!", err);
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    err = ix_rm_buffer_get_meta(arg_hProperty, (void**)&pCPAttributes);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                        "failed to access the configuration property attributes!", err);
        goto label_1;
    }

    arg_pInfo->m_DataSize = pCPAttributes->m_DataSize;
    arg_pInfo->m_hChildrenPropertiesList = pCPAttributes->m_hChildrenPropertiesListHead;
    arg_pInfo->m_hParentProperty = pCPAttributes->m_hParentProperty;
    arg_pInfo->m_hPropertyNext = pCPAttributes->m_hPropertyNext;
    arg_pInfo->m_hPropertyPrev = pCPAttributes->m_hPropertyPrev;
    arg_pInfo->m_Options = pCPAttributes->m_Options;
    ix_ossl_memcpy(arg_pInfo->m_aName, pCPAttributes->m_aName, IX_CP_MAX_PROPERTY_NAME_LENGTH);

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}




/**
 * NAME: ix_rm_cp_property_get_subproperty
 *
 * DESCRIPTION: This function will return the handle of the property with the arg_SubpropertyIndex
 *          index. The subproperties indexes start from 0.
 * 
 * @Param:  - IN arg_hProperty - the handle of the property whose subproperties we
 *          are interested in.
 * @Param:  - IN arg_SubpropertyIndex - the index of the subproperty we want to retrieve.
 * @Param:  - OUT arg_pSubproperty - the address where the subproperty handle will be stored. 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_cp_property_get_subproperty(
                                            ix_configuration_property_handle arg_hProperty, 
                                            ix_uint32 arg_SubpropertyIndex, 
                                            ix_configuration_property_handle* arg_pSubproperty 
                                          )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_cp_property_get_subproperty";

    ix_error err = IX_SUCCESS;

    ix_cp_property_attributes* pCPAttributes;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_CPManager.m_SyncAdapter;        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check if the passed handle is not NULL */
    if(arg_hProperty == IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_CP_HANDLE_NULL, s_aFunctionIdentifier, 
                        "the passed configuration property handle NULL!", err);
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    err = ix_rm_buffer_get_meta(arg_hProperty, (void**)&pCPAttributes);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                        "failed to access the configuration property attributes!", err);
        goto label_1;
    }

    *arg_pSubproperty = IX_NULL_CONFIGURATION_PROPERTY_HANDLE;

    /* find the requested index */
    {
        ix_configuration_property_handle hCP = pCPAttributes->m_hChildrenPropertiesListHead;
        ix_uint32 local_index = 0;

        while(hCP != IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
        {
            if(local_index == arg_SubpropertyIndex)
            {
                *arg_pSubproperty = hCP;
                break;
            }
            local_index++;

            err = ix_rm_buffer_get_meta(hCP, (void**)&pCPAttributes);
            if(err != IX_SUCCESS)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                                "failed to access the configuration property attributes!", err);
                goto label_1;
            }

            hCP = pCPAttributes->m_hPropertyNext;
        }
    }

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


