/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      counter_64bit.c
 *
 * = DESCRIPTION
 *      This file will define all functions for the 64 bit counter API.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      5/22/2002 4:09:55 PM - creation time 
 * ============================================================================
 */

static const char s_FileIdentifier[] = "$Id: counter_64bit.c,v 1.11 2003/08/08 18:50:24 ageorges Exp $";

#include "ix_ossl.h"

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"

#include "rm/symbols.h"
#include "rm/internal/internal_support.h"

#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"

#include "rm/system.h"
#include "rm/internal/internal_system.h"

#include "rm/memory.h"

#include "rm/counter_64bit.h"
#include "rm/internal/internal_counter_64bit.h"

/**
 * NAME: ix_rm_counter_64bit_new
 *
 * DESCRIPTION: This function will allocate an array of new 64 bit counters and upon return
 *  will fill the array of counter handles with the handles of the newly created counters.
 *  The caller will specify the number of counters to be created through arg_CounterNumber
 *  parameter. The second parameter is the virtual address of a memory block to be used for
 *  the internal 32 bit counters coresponding to each counter. The block of memory should be
 *  an array of 32 bit integers with at least arg_CounterNumber elements. The first element
 *  of this array will be used as the internal 32 bit counter of the first 64 bit counter 
 *  created and so on. The caller can this way patch into the microcode just the offset of
 *  the start of this block. The last parameter is a pointer to the first element of an array
 *  of 64 bit counter handles that will be filled upon return. The caller has to allocate this
 *  handle array before calling this function.
 *  The third parameter will specify the internal 32 bit timer overflow time for all these timers.
 *  The internal 32 bit timer is guaranteed to be read and reset more often than the specified
 *  overflow time. This way the internal timer will be guarateed not to overflow.
 *
 * @Param:  - IN arg_CounterNumber - the number of counters to be created by this call.
 * @Param:  - IN arg_InternalCounterMemoryBlock - address of the memory block to be used for the
 *  internal part of the timers. This should be common memory, allocated by a call to ix_rm_mem_alloc()
 *  function. 
 * @Param:  - IN arg_InternalOverflowTime - the overflow time for the internal 32bit ME counter
 * @Param:  - OUT arg_pCounterHandle - address of the first element of a counter handle array
 *  where the handles of the newly created counters will be stored upon return.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_counter_64bit_new(
                                  ix_uint32 arg_CounterNumber, 
                                  ix_uint32* arg_InternalCounterMemoryBlock,
                                  ix_uint32 arg_InternalOverflowTime,
                                  ix_counter_64bit_handle* arg_pCounterHandle
                                )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_counter_64bit_new";

    ix_error err = IX_SUCCESS;
    ix_memory_type memoryType;
    ix_update_64bit_counter_func_ptr updateFunction;
    ix_uint32 i;
    ix_uint32 forceUpdate = 0;

    _ix_counter_64bit* p64bitCounter;
   

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_64bitCounterManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check if the overflow time is non zero */
    if(arg_InternalOverflowTime == 0)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_BAD_OVERFLOW_TIME, s_aFunctionIdentifier, 
                        "the passed overflow time is unacceptable!", err);
    }

    /* check to see if we have a valid number request */
    if(arg_CounterNumber == 0)
        return IX_SUCCESS;

    /* check to see if the passed memory block is OK */
    {
        ix_uint32 channel;
        ix_uint32 offset;
        void* physicalAddress;

        err = ix_rm_get_phys_offset(arg_InternalCounterMemoryBlock, &memoryType,
            &channel, &offset, &physicalAddress);
        if((err != IX_SUCCESS) || 
            ((memoryType != IX_MEMORY_TYPE_SRAM) && (memoryType != IX_MEMORY_TYPE_SCRATCH)))
        {
            return IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_BAD_INTERNAL_MEMORY, s_aFunctionIdentifier, 
                        "the passed internal memory block is not a SRAM or SCRATCH valid block!", err);
        }

        if(memoryType == IX_MEMORY_TYPE_SRAM)
            updateFunction = _ix_rm_counter_64bit_sram_update;
        else
            updateFunction = _ix_rm_counter_64bit_scratch_update;
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    if(arg_CounterNumber <= (g_64bitCounterManager.m_FreeCounters + 
        (IX_64BIT_COUNTER_BLOCKS - g_64bitCounterManager.m_InitializedCounterBlocks) *
        IX_64BIT_COUNTER_PER_BLOCK))
    {
        /* make sure we can accomodate the requested number of counters */
        while(arg_CounterNumber > g_64bitCounterManager.m_FreeCounters)
        {
            /* try to initialize a new 64 bit counter block */
            err = _ix_rm_counter_64bit_initialize_new_block();
            if(err != IX_SUCCESS)
            {
                break;
            }
        }

        if(arg_CounterNumber > g_64bitCounterManager.m_FreeCounters)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_NOT_ENOUGH_COUNTERS, s_aFunctionIdentifier, 
                            "there are not enough counters to satisfy the allocation request!", err);
            goto label_1;
        }
    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_NOT_ENOUGH_COUNTERS, s_aFunctionIdentifier, 
                            "there are not enough counters to satisfy the allocation request!", err);
        goto label_1;
    }

    for(i = 0; i < arg_CounterNumber; i++, arg_pCounterHandle++, arg_InternalCounterMemoryBlock++)
    {
        p64bitCounter = g_64bitCounterManager.m_pCounterFreeList;
        g_64bitCounterManager.m_pCounterFreeList = p64bitCounter->m_pNext;

        g_64bitCounterManager.m_FreeCounters--;

        p64bitCounter->m_CoreValue = 0;
        p64bitCounter->m_Properties = IX_64BIT_COUNTER_PROPERTY_ACTIVE;
        p64bitCounter->m_Timeout = arg_InternalOverflowTime;
        p64bitCounter->m_UpdateElapsedTime = 0;
        p64bitCounter->m_UpdateFuncPtr = updateFunction;
        p64bitCounter->m_pMEValue = arg_InternalCounterMemoryBlock;
        p64bitCounter->m_pNext = 0;

        IX_RM_MEM_UINT32_WRITE(p64bitCounter->m_pMEValue, 0);
        
        _ix_rm_counter_64bit_add_to_active_list(p64bitCounter);

        *arg_pCounterHandle = p64bitCounter->m_hCounter;
    }

    if(arg_InternalOverflowTime < g_64bitCounterManager.m_ActiveTO)
    {
        g_64bitCounterManager.m_ActiveTO = arg_InternalOverflowTime;
        forceUpdate = 1;
    }

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    if(forceUpdate != 0)
        return _ix_rm_counter_64bit_update();
    else
        return err;
}




/**
 * NAME: ix_rm_counter_64bit_delete
 *
 * DESCRIPTION: This function will delete the passed counter. From this point on the
 *  handle is invalid and should not be used. All resources related to the
 *  counter will be released.
 * 
 * @Param:  - IN arg_hCounter - handle of the counter to be deleted
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_counter_64bit_delete(
                                     ix_counter_64bit_handle arg_hCounter 
                                   )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_counter_64bit_delete";

    ix_error err = IX_SUCCESS;
    _ix_counter_64bit* pCounter = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_64bitCounterManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    {
        ix_uint32 blockId = IX_64BIT_COUNTER_HANDLE_GET_COUNTER_BLOCK_INDEX(arg_hCounter);
        
#if defined( _IX_RM_DEBUG_ )

        /* do extra checking in debug mode */
        if(IX_64BIT_COUNTER_HANDLE_IS_VALID(arg_hCounter) == 0)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_HANDLE_INVALID, s_aFunctionIdentifier, 
                        "the passed 64 bit counter handle is invalid!", err);
            goto label_1;
        }
#else


#endif /* end defined( _IX_RM_DEBUG_ ) */

        pCounter = g_64bitCounterManager.m_CounterBlocks[blockId];
        if(pCounter == 0)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_HANDLE_INVALID, s_aFunctionIdentifier, 
                        "the passed 64 bit counter handle is invalid!", err);
            goto label_1;
        }

        /* get the counter of interest */
        pCounter += IX_64BIT_COUNTER_HANDLE_GET_COUNTER_INDEX(arg_hCounter);

        /* check if the counter is active */
        if(pCounter->m_Properties & IX_64BIT_COUNTER_PROPERTY_ACTIVE)
        {
            _ix_rm_counter_64bit_remove_from_active_list(pCounter);

            /* add the counter to the free list */
            pCounter->m_Properties = 0;
            pCounter->m_UpdateElapsedTime = 0;
            pCounter->m_CoreValue = 0;
            pCounter->m_pMEValue = 0;
            pCounter->m_UpdateFuncPtr = 0;
            pCounter->m_Timeout = 0;

            pCounter->m_pNext = g_64bitCounterManager.m_pCounterFreeList;
            g_64bitCounterManager.m_pCounterFreeList = pCounter;

            g_64bitCounterManager.m_FreeCounters++;

            /* adjust the active timeout */
            if(g_64bitCounterManager.m_pActiveCounterList == 0)
            {
                g_64bitCounterManager.m_ActiveTO = IX_OSSL_WAIT_FOREVER;
            }
            else
            {
                if(g_64bitCounterManager.m_pActiveCounterList->m_Timeout > g_64bitCounterManager.m_ActiveTO)
                    g_64bitCounterManager.m_ActiveTO = g_64bitCounterManager.m_pActiveCounterList->m_Timeout;
            }
        }
    }

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: ix_rm_counter_64bit_get_internal_overflow_time
 *
 * DESCRIPTION: This function will get the overflow time for the internal 32bit and store at the
 *  passed location.
 * 
 * @Param:  - IN arg_hCounter - the handle of the counter of interest.
 * @Param:  - IN arg_pInternalOverflowTime - address where the current internal counter overflow
 *  time will be stored upon return.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_counter_64bit_get_internal_overflow_time(
                                                         ix_counter_64bit_handle arg_hCounter, 
                                                         ix_uint32* arg_pInternalOverflowTime
                                                       )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_counter_64bit_get_internal_overflow_time";

    ix_error err = IX_SUCCESS;
    _ix_counter_64bit* pCounter = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_64bitCounterManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    {
        ix_uint32 blockId = IX_64BIT_COUNTER_HANDLE_GET_COUNTER_BLOCK_INDEX(arg_hCounter);
        
#if defined( _IX_RM_DEBUG_ )

        /* do extra checking in debug mode */
        if(IX_64BIT_COUNTER_HANDLE_IS_VALID(arg_hCounter) == 0)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_HANDLE_INVALID, s_aFunctionIdentifier, 
                        "the passed 64 bit counter handle is invalid!", err);
            goto label_1;
        }
#else


#endif /* end defined( _IX_RM_DEBUG_ ) */

        pCounter = g_64bitCounterManager.m_CounterBlocks[blockId];
        if(pCounter == 0)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_HANDLE_INVALID, s_aFunctionIdentifier, 
                        "the passed 64 bit counter handle is invalid!", err);
            goto label_1;
        }

        /* get the counter of interest */
        pCounter += IX_64BIT_COUNTER_HANDLE_GET_COUNTER_INDEX(arg_hCounter);

        /* check if the counter is active */
        if((pCounter->m_Properties & IX_64BIT_COUNTER_PROPERTY_ACTIVE) == 0)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_INACTIVE, s_aFunctionIdentifier, 
                        "the passed 64 bit counter is inactive!", err);
            goto label_1;
        }

        *arg_pInternalOverflowTime = pCounter->m_Timeout;

    }

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: ix_rm_counter_64bit_set_internal_overflow_time
 *
 * DESCRIPTION: This function will set the overflow time for the internal 32bit
 *  ME counter to the new value.
 * 
 * @Param:  - IN arg_hCounter - the handle of the counter of interest.
 * @Param:  - IN arg_InternalOverflowTime - the new overflow time for the internal 32bit ME counter
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_counter_64bit_set_internal_overflow_time(
                                                         ix_counter_64bit_handle arg_hCounter, 
                                                         ix_uint32 arg_InternalOverflowTime
                                                       )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_counter_64bit_set_internal_overflow_time";

    ix_error err = IX_SUCCESS;
    _ix_counter_64bit* pCounter = 0;
    ix_uint32 forceUpdate = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_64bitCounterManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check if the overflow time is non zero */
    if(arg_InternalOverflowTime == 0)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_BAD_OVERFLOW_TIME, s_aFunctionIdentifier, 
                        "the passed overflow time is unacceptable!", err);
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    {
        ix_uint32 blockId = IX_64BIT_COUNTER_HANDLE_GET_COUNTER_BLOCK_INDEX(arg_hCounter);
        
#if defined( _IX_RM_DEBUG_ )

        /* do extra checking in debug mode */
        if(IX_64BIT_COUNTER_HANDLE_IS_VALID(arg_hCounter) == 0)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_HANDLE_INVALID, s_aFunctionIdentifier, 
                        "the passed 64 bit counter handle is invalid!", err);
            goto label_1;
        }
#else


#endif /* end defined( _IX_RM_DEBUG_ ) */

        pCounter = g_64bitCounterManager.m_CounterBlocks[blockId];
        if(pCounter == 0)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_HANDLE_INVALID, s_aFunctionIdentifier, 
                        "the passed 64 bit counter handle is invalid!", err);
            goto label_1;
        }

        /* get the counter of interest */
        pCounter += IX_64BIT_COUNTER_HANDLE_GET_COUNTER_INDEX(arg_hCounter);

        /* check if the counter is active */
        if((pCounter->m_Properties & IX_64BIT_COUNTER_PROPERTY_ACTIVE) == 0)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_INACTIVE, s_aFunctionIdentifier, 
                        "the passed 64 bit counter is inactive!", err);
            goto label_1;
        }

        if(pCounter->m_Timeout != arg_InternalOverflowTime)
        {
            pCounter->m_Timeout = arg_InternalOverflowTime;

            /* now we have to extract the counter from the list and insert it into the ordered list */
            _ix_rm_counter_64bit_remove_from_active_list(pCounter);
            _ix_rm_counter_64bit_add_to_active_list(pCounter);
            if(arg_InternalOverflowTime < g_64bitCounterManager.m_ActiveTO)
            {
                g_64bitCounterManager.m_ActiveTO = arg_InternalOverflowTime;
                forceUpdate = 1;
            }
        }

    }

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    if(forceUpdate != 0)
        return _ix_rm_counter_64bit_update();
    else
        return err;
}



/**
 * NAME: ix_rm_counter_64bit_get_value
 *
 * DESCRIPTION: This function will return the 64 bit core value of the counter. Internally
 *  the function will read the ME counter and reset it to 0 and increment the core 
 *  counter with the read value.
 * 
 * @Param:  - IN arg_hCounter - the handle of the counter of interest.
 * @Param:  - OUT arg_pCounterValue - address where the value of the counter should be
 *          stored on return
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_counter_64bit_get_value(
                                        ix_counter_64bit_handle arg_hCounter, 
                                        ix_uint64* arg_pCounterValue 
                                      )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_counter_64bit_get_value";

    ix_error err = IX_SUCCESS;
    _ix_counter_64bit* pCounter = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_64bitCounterManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    {
        ix_uint32 blockId = IX_64BIT_COUNTER_HANDLE_GET_COUNTER_BLOCK_INDEX(arg_hCounter);
        
#if defined( _IX_RM_DEBUG_ )

        /* do extra checking in debug mode */
        if(IX_64BIT_COUNTER_HANDLE_IS_VALID(arg_hCounter) == 0)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_HANDLE_INVALID, s_aFunctionIdentifier, 
                        "the passed 64 bit counter handle is invalid!", err);
            goto label_1;
        }
#else


#endif /* end defined( _IX_RM_DEBUG_ ) */

        pCounter = g_64bitCounterManager.m_CounterBlocks[blockId];
        if(pCounter == 0)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_HANDLE_INVALID, s_aFunctionIdentifier, 
                        "the passed 64 bit counter handle is invalid!", err);
            goto label_1;
        }

        /* get the counter of interest */
        pCounter += IX_64BIT_COUNTER_HANDLE_GET_COUNTER_INDEX(arg_hCounter);

        /* check if the counter is active */
        if((pCounter->m_Properties & IX_64BIT_COUNTER_PROPERTY_ACTIVE) == 0)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_INACTIVE, s_aFunctionIdentifier, 
                        "the passed 64 bit counter is inactive!", err);
            goto label_1;
        }

        /* force an update */
        pCounter->m_UpdateFuncPtr(pCounter, 0);

        *arg_pCounterValue = pCounter->m_CoreValue;
    }

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}




/**
 * NAME: ix_rm_counter_64bit_set_value
 *
 * DESCRIPTION: This function will set the value of the core counter to the specified
 *  value.
 * 
 * @Param:  - IN arg_hCounter - the handle of the counter of interest.
 * @Param:  - IN arg_CounterValue - the new value for the core counter
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_counter_64bit_set_value(
                                        ix_counter_64bit_handle arg_hCounter, 
                                        ix_uint64 arg_CounterValue 
                                      )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_counter_64bit_set_value";

    ix_error err = IX_SUCCESS;
    _ix_counter_64bit* pCounter = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_64bitCounterManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    {
        ix_uint32 blockId = IX_64BIT_COUNTER_HANDLE_GET_COUNTER_BLOCK_INDEX(arg_hCounter);
        
#if defined( _IX_RM_DEBUG_ )

        /* do extra checking in debug mode */
        if(IX_64BIT_COUNTER_HANDLE_IS_VALID(arg_hCounter) == 0)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_HANDLE_INVALID, s_aFunctionIdentifier, 
                        "the passed 64 bit counter handle is invalid!", err);
            goto label_1;
        }
#else


#endif /* end defined( _IX_RM_DEBUG_ ) */

        pCounter = g_64bitCounterManager.m_CounterBlocks[blockId];
        if(pCounter == 0)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_HANDLE_INVALID, s_aFunctionIdentifier, 
                        "the passed 64 bit counter handle is invalid!", err);
            goto label_1;
        }

        /* get the counter of interest */
        pCounter += IX_64BIT_COUNTER_HANDLE_GET_COUNTER_INDEX(arg_hCounter);

        /* check if the counter is active */
        if((pCounter->m_Properties & IX_64BIT_COUNTER_PROPERTY_ACTIVE) == 0)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_INACTIVE, s_aFunctionIdentifier, 
                        "the passed 64 bit counter is inactive!", err);
            goto label_1;
        }

        pCounter->m_CoreValue = arg_CounterValue;
    }

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


