/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      fast_memory_operations.c
 *
 * = DESCRIPTION
 *      This file provides implementation of all functions related to the fast memory
 *      operations. These operations have been added to copy data from OS memory to the
 *      RM managed memory and vice versa. We provide a fast copy from two locations in
 *      the RM managed memory areas. These fuctions will take advantage of CPU instructions
 *      that allow for fast memory transfer.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      9/10/2002 1:49:42 PM - creation time 
 * ============================================================================
 */

static const char s_FileIdentifier[] = "$Id: fast_memory_operations.c,v 1.5 2002/11/07 23:35:09 ageorges Exp $";

#include "ix_ossl.h"

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"

#include "rm/internal/internal_support.h"

#include "rm/fast_memory_operations.h"


/**
 * For now these functions have the same implementation but they will be different
 * at lease for the RM Foreign Model implementation where the managed memory belongs
 * to the transactor.
 */


/**
 * NAME: ix_rm_managed_to_os_memory_copy
 *
 * DESCRIPTION: This function will copy a managed memory area into OS memory. This function
 *  will take advantage of the CPU fast instructions only if data is 4 byte aligned.
 * 
 * @Param:  - INOUT arg_pDest - this is the OS destination memory location where the data will be copied to
 * @Param:  - IN arg_pSrc - this is the managed memory location where the data will be copied from
 * @Param:  - IN arg_Count - this is the amount of data to be copied.
 *
 * @Return: The address of the destination buffer
 */
void* ix_rm_managed_to_os_memory_copy(
                                       void* arg_pDest, 
                                       const void* arg_pSrc,
                                       ix_uint32 arg_Count 
                                     )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_managed_to_os_memory_copy";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* first check if source and destination have 4 byte aligned addresses */
    if(((ix_uint32)arg_pDest & 0x03) && ((ix_uint32)arg_pSrc & 0x03))
    {
        return ix_ossl_memcpy(arg_pDest, arg_pSrc, arg_Count);
    }
    else
    {

#if defined( _IX_RM_IMPL_HARDWARE_ )

#if defined(__GNUC__) 

        __asm__(
            "mov r8, %2 \n" /* put in r8 the arg_Count */
            "mov r9, r8, lsr #5 \n" /* get the number of 32 byte chunks */
            "cmp r9, #0 \n" /* see if it is 0 */
            "beq m_to_o_copy_4 \n" /* if zero go to the 4 byte chunks */
            "add r9, %0, r9, lsl #5 \n" /* calculate the end address */
            "\n"
            "m_to_o_loop_32:\n"
            "\n"
            "ldmia %0!, {r0-r7} \n" /* load 32 bytes from the source address */
            "stmia %1!, {r0-r7} \n" /* store 32 bytes to the destination address */
            "cmp %0, r9 \n" /* see if we are done */
            "blo m_to_o_loop_32 \n"
            "\n"
            "m_to_o_copy_4:\n" /* 4 byte chunks */
            "\n"
            "and r8, r8, #31 \n"
            "cmp r8, #0 \n" /* see if the remaining number of bytes to copy is 0 */
            "beq m_to_o_copy_end \n" /* we are done */
            "mov r9, r8, lsr #2 \n" /* get the number of 4 byte chunks */
            "cmp r9, #0 \n" /* see if it is 0 */
            "beq m_to_o_copy_1 \n" /* if zero go to the 1 byte chunks */
            "add r9, %0, r9, lsl #2 \n" /* calculate the end address */
            "\n"
            "m_to_o_loop_4:\n"
            "\n"
            "ldmia %0!, {r0} \n" /* load 4 bytes from the source address */
            "stmia %1!, {r0} \n" /* store 4 bytes to the destination address */
            "cmp %0, r9 \n" /* see if we are done */
            "blo m_to_o_loop_4 \n"
            "\n"
            "m_to_o_copy_1:\n"
            "\n"
            "and r8, r8, #3 \n"
            "cmp r8, #0 \n" /* see if the remaining number of bytes to copy is 0 */
            "beq m_to_o_copy_end \n" /* we are done */
            "add r9, %0, r8" /* calcuate the end address */
            "\n"
            "m_to_o_loop_1: \n"
            "\n"
            "ldrb r0, [%0], #1 \n" /* load 1 byte from the source address */
            "strb r0, [%1], #1 \n" /* store 1 byte to the destination address */
            "cmp %0, r9 \n" /* see if we are done */
            "blo m_to_o_loop_1 \n"
            "\n"
            "m_to_o_copy_end: \n"
            "\n"
            : /* no outputs */
            : "r"(arg_pSrc),"r"(arg_pDest),"r"(arg_Count)
            :"r0","r1","r2","r3","r4","r5",
            "r6","r7","r8","r9"
            );

        return arg_pDest;

#else

#error "Implement fast memory operations!"

#endif /* defined(__GNUC__) */

#else

        return ix_ossl_memcpy(arg_pDest, arg_pSrc, arg_Count);

#endif /* defined( _IX_RM_IMPL_HARDWARE_ ) */

    }
}


/**
 * NAME: ix_rm_os_to_managed_memory_copy
 *
 * DESCRIPTION: This function will copy a OS memory area into managed memory. This function
 *  will take advantage of the CPU fast instructions only if data is 4 byte aligned.
 * 
 * @Param:  - INOUT arg_pDest - this is the managed destination memory location where the data will be copied to
 * @Param:  - IN arg_pSrc - this is the OS memory location where the data will be copied from
 * @Param:  - IN arg_Count - this is the amount of data to be copied.
 *
 * @Return: The address of the destination buffer
 */
void* ix_rm_os_to_managed_memory_copy(
                                       void* arg_pDest, 
                                       const void* arg_pSrc,
                                       ix_uint32 arg_Count 
                                     )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_os_to_managed_memory_copy";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* first check if source and destination have 4 byte aligned addresses */
    if(((ix_uint32)arg_pDest & 0x03) && ((ix_uint32)arg_pSrc & 0x03))
    {
        return ix_ossl_memcpy(arg_pDest, arg_pSrc, arg_Count);
    }
    else
    {

#if defined( _IX_RM_IMPL_HARDWARE_ )

#if defined(__GNUC__) 

        __asm__(
            "mov r8, %2 \n" /* put in r8 the arg_Count */
            "mov r9, r8, lsr #5 \n" /* get the number of 32 byte chunks */
            "cmp r9, #0 \n" /* see if it is 0 */
            "beq o_to_m_copy_4 \n" /* if zero go to the 4 byte chunks */
            "add r9, %0, r9, lsl #5 \n" /* calculate the end address */
            "\n"
            "o_to_m_loop_32:\n"
            "\n"
            "ldmia %0!, {r0-r7} \n" /* load 32 bytes from the source address */
            "stmia %1!, {r0-r7} \n" /* store 32 bytes to the destination address */
            "cmp %0, r9 \n" /* see if we are done */
            "blo o_to_m_loop_32 \n"
            "\n"
            "o_to_m_copy_4:\n" /* 4 byte chunks */
            "\n"
            "and r8, r8, #31 \n"
            "cmp r8, #0 \n" /* see if the remaining number of bytes to copy is 0 */
            "beq o_to_m_copy_end \n" /* we are done */
            "mov r9, r8, lsr #2 \n" /* get the number of 4 byte chunks */
            "cmp r9, #0 \n" /* see if it is 0 */
            "beq o_to_m_copy_1 \n" /* if zero go to the 1 byte chunks */
            "add r9, %0, r9, lsl #2 \n" /* calculate the end address */
            "\n"
            "o_to_m_loop_4:\n"
            "\n"
            "ldmia %0!, {r0} \n" /* load 4 bytes from the source address */
            "stmia %1!, {r0} \n" /* store 4 bytes to the destination address */
            "cmp %0, r9 \n" /* see if we are done */
            "blo o_to_m_loop_4 \n"
            "\n"
            "o_to_m_copy_1:\n"
            "\n"
            "and r8, r8, #3 \n"
            "cmp r8, #0 \n" /* see if the remaining number of bytes to copy is 0 */
            "beq o_to_m_copy_end \n" /* we are done */
            "add r9, %0, r8" /* calcuate the end address */
            "\n"
            "o_to_m_loop_1: \n"
            "\n"
            "ldrb r0, [%0], #1 \n" /* load 1 byte from the source address */
            "strb r0, [%1], #1 \n" /* store 1 byte to the destination address */
            "cmp %0, r9 \n" /* see if we are done */
            "blo o_to_m_loop_1 \n"
            "\n"
            "o_to_m_copy_end: \n"
            "\n"
            : /* no outputs */
            : "r"(arg_pSrc),"r"(arg_pDest),"r"(arg_Count)
            :"r0","r1","r2","r3","r4","r5",
            "r6","r7","r8","r9"
            );

        return arg_pDest;

#else

#error "Implement fast memory operations!"

#endif /* defined(__GNUC__) */

#else

        return ix_ossl_memcpy(arg_pDest, arg_pSrc, arg_Count);

#endif /* defined( _IX_RM_IMPL_HARDWARE_ ) */

    }
}


/**
 * NAME: ix_rm_managed_to_managed_memory_copy
 *
 * DESCRIPTION: This function will copy a managed memory area into managed memory. This function
 *  will take advantage of the CPU fast instructions only if data is 4 byte aligned.
 * 
 * @Param:  - INOUT arg_pDest - this is the managed destination memory location where the data will be copied to
 * @Param:  - IN arg_pSrc - this is the managed memory location where the data will be copied from
 * @Param:  - IN arg_Count - this is the amount of data to be copied.
 *
 * @Return: The address of the destination buffer
 */
void* ix_rm_managed_to_managed_memory_copy(
                                            void* arg_pDest, 
                                            const void* arg_pSrc,
                                            ix_uint32 arg_Count 
                                          )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_managed_to_managed_memory_copy";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* first check if source and destination have 4 byte aligned addresses */
    if(((ix_uint32)arg_pDest & 0x03) && ((ix_uint32)arg_pSrc & 0x03))
    {
        return ix_ossl_memcpy(arg_pDest, arg_pSrc, arg_Count);
    }
    else
    {

#if defined( _IX_RM_IMPL_HARDWARE_ )

#if defined(__GNUC__) 

        __asm__(
            "mov r8, %2 \n" /* put in r8 the arg_Count */
            "mov r9, r8, lsr #5 \n" /* get the number of 32 byte chunks */
            "cmp r9, #0 \n" /* see if it is 0 */
            "beq m_to_m_copy_4 \n" /* if zero go to the 4 byte chunks */
            "add r9, %0, r9, lsl #5 \n" /* calculate the end address */
            "\n"
            "m_to_m_loop_32:\n"
            "\n"
            "ldmia %0!, {r0-r7} \n" /* load 32 bytes from the source address */
            "stmia %1!, {r0-r7} \n" /* store 32 bytes to the destination address */
            "cmp %0, r9 \n" /* see if we are done */
            "blo m_to_m_loop_32 \n"
            "\n"
            "m_to_m_copy_4:\n" /* 4 byte chunks */
            "\n"
            "and r8, r8, #31 \n"
            "cmp r8, #0 \n" /* see if the remaining number of bytes to copy is 0 */
            "beq m_to_m_copy_end \n" /* we are done */
            "mov r9, r8, lsr #2 \n" /* get the number of 4 byte chunks */
            "cmp r9, #0 \n" /* see if it is 0 */
            "beq m_to_m_copy_1 \n" /* if zero go to the 1 byte chunks */
            "add r9, %0, r9, lsl #2 \n" /* calculate the end address */
            "\n"
            "m_to_m_loop_4:\n"
            "\n"
            "ldmia %0!, {r0} \n" /* load 4 bytes from the source address */
            "stmia %1!, {r0} \n" /* store 4 bytes to the destination address */
            "cmp %0, r9 \n" /* see if we are done */
            "blo m_to_m_loop_4 \n"
            "\n"
            "m_to_m_copy_1:\n"
            "\n"
            "and r8, r8, #3 \n"
            "cmp r8, #0 \n" /* see if the remaining number of bytes to copy is 0 */
            "beq m_to_m_copy_end \n" /* we are done */
            "add r9, %0, r8" /* calcuate the end address */
            "\n"
            "m_to_m_loop_1: \n"
            "\n"
            "ldrb r0, [%0], #1 \n" /* load 1 byte from the source address */
            "strb r0, [%1], #1 \n" /* store 1 byte to the destination address */
            "cmp %0, r9 \n" /* see if we are done */
            "blo m_to_m_loop_1 \n"
            "\n"
            "m_to_m_copy_end: \n"
            "\n"
            : /* no outputs */
            : "r"(arg_pSrc),"r"(arg_pDest),"r"(arg_Count)
            :"r0","r1","r2","r3","r4","r5",
            "r6","r7","r8","r9"
            );

        return arg_pDest;

#else

#error "Implement fast memory operations!"

#endif /* defined(__GNUC__) */

#else

        return ix_ossl_memcpy(arg_pDest, arg_pSrc, arg_Count);

#endif /* defined( _IX_RM_IMPL_HARDWARE_ ) */

    }
}

