/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      hardware.c
 *
 * = DESCRIPTION
 *      This file will define all exported functions by the Hardware Resource Manager API.
 *      This API manage MEv2 hardware features including SRAM queues, rings, scratch rings etc.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      4/23/2002 12:36:18 PM - creation time 
 * ============================================================================
 */

static const char s_aFileIdentifier[] = "$Id: hardware.c,v 1.56 2003/12/12 00:21:48 ageorges Exp $";

#include "ix_ossl.h"

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"

#include "rm/symbols.h"
#include "rm/internal/internal_support.h"

#include <hal_sram.h>
#include <hal_scratch.h>

#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"

#include "rm/system.h"
#include "rm/internal/internal_system.h"

#include "rm/memory.h"

#include "rm/buffer.h"
#include "rm/internal/internal_buffer.h"

#include "rm/hardware.h"

#include "rm/atomic_operations.h"

#include "rm/microengine.h"

#include "rm/internal/internal_hardware.h"



/**
 * Hardware SRAM queues functions
 */

/**
 * NAME: ix_rm_hw_queue_create
 *
 * DESCRIPTION: This function will reserve arg_QueuesNumber entries in the Q-Array of the SRAM
 *          controller on the specified channel. Note that no queue descriptors are actually
 *          allocated by this call. The queues should be contiguos otherwise the call will fail.
 *          *arv_pQueueHandleBase will have the handle of the first queue upon return. 
 *          A macro will be provided to get Nth handle relative to the base handle when required.
 * 
 * @Param:  - IN arg_Channel - the channel of SRAM controller in question.
 * @Param:  - IN arg_QueuesNumber - number of queues to be allocated.
 * @Param:  - IN arg_Flags - flags passed to the request.
 * @Param:  - OUT arv_pQueueHandleBase - address where the first queue handle will be stored. 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure. If the Resource Manager can not find
 *          arg_QueuesNumber contiguos queues then the call will fail.
 */
ix_error ix_rm_hw_queue_create( 
                             ix_uint32 arg_Channel, 
                             ix_uint32 arg_QueuesNumber, 
                             ix_uint32 arg_Flags, 
                             ix_hw_queue_handle* arv_pQueueHandleBase 
                           )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_hw_queue_create";

    ix_error err = 0;
    ix_uint32 i;
    ix_uint32 j;
    ix_uint32 k;
    ix_sram_qarray_channel_manager* pQArrayChannelManager = 0;
    ix_uint32 entryUsage;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_QArrayManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(g_MicrocodeLoaded != 0)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_HW_QUEUE_CREATE_FAILED, s_aFunctionIdentifier, 
                        "you can not create a hardware queue after the uCode has been loaded!", err);
    }

    IX_RM_ASSERT(arg_Channel < IX_SRAM_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
    pQArrayChannelManager = g_QArrayManager.m_aQArrayManagersPtr[arg_Channel];
    if (pQArrayChannelManager == NULL)
    {
        return IX_ERROR_NEW(IX_RM_ERROR_HW_QARRAY_MGR_ALLOC_FAILED, IX_ERROR_LEVEL_LOCAL);
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the QArray manager data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* find arg_QueuesNumber of consecutive QArray entries that are not in use */
    for(i = 0; i < IX_MAX_SRAM_QARRAY_ENTRIES; i++)
    {
        entryUsage = IX_RM_MEM_UINT32_READ(&pQArrayChannelManager->m_aQArrayUsage[i]);

        /* if in use go to the next entry */
        if(IX_RM_GET_QARRAY_ENTRY_IN_USE(entryUsage))
        {
            continue;
        }

        /* check to see if there are arg_QueuesNumber consecutive free entries */
        for(j = 1, k = i; (j < arg_QueuesNumber) && (k < IX_MAX_SRAM_QARRAY_ENTRIES); j++, k++)
        {
            entryUsage = IX_RM_MEM_UINT32_READ(&pQArrayChannelManager->m_aQArrayUsage[k]);

            /* if in use exit the loop */
            if(IX_RM_GET_QARRAY_ENTRY_IN_USE(entryUsage))
            {
                break;
            }
        }

        
        /**
         * If we found arg_QueuesNumber consecutive free entries then mark them to be
         * in use.
         */
        if (j == arg_QueuesNumber)
        {
            ix_sram_queue_descriptor* pQueueDescriptor;
            for(k = i; k < i + arg_QueuesNumber; k++)
            {
                entryUsage = 0;
                IX_RM_MEM_UINT32_WRITE(&pQArrayChannelManager->m_aQArrayUsage[k],IX_RM_SET_QARRAY_ENTRY_IN_USE(entryUsage,1));

                /* initialize the QArray descriptor entry */
                pQueueDescriptor = &pQArrayChannelManager->m_aQArrayDescriptors[k].m_QueueDescriptor;
                IX_RM_MEM_UINT32_WRITE(&pQueueDescriptor->m_HeadInfo, 0);
                IX_RM_MEM_UINT32_WRITE(&pQueueDescriptor->m_TailInfo, 0);
                IX_RM_MEM_UINT32_WRITE(&pQueueDescriptor->m_CountInfo, 0);
                IX_RM_MEM_UINT32_WRITE(&pQueueDescriptor->m_Reserved, 0);

            }
            break;
        }
        else
        {
            /* set to check the next entry */
            i = k;
        }

    }

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the QArray manager data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    if(i < IX_MAX_SRAM_QARRAY_ENTRIES)
    {
        /* we found the required number of free entries */
        ix_hw_queue_handle queueHandle = IX_RM_HW_QUEUE_CREATE_HANDLE(0,0);

        IX_RM_HW_QUEUE_SET_INDEX(queueHandle, i);
        IX_RM_HW_QUEUE_SET_CHANNEL(queueHandle, arg_Channel);

        *arv_pQueueHandleBase = queueHandle;
    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_HW_QUEUE_CREATE_FAILED, s_aFunctionIdentifier, 
                        "the reservation of QArray entries failed!", err);
    }

    return err;
}



/**
 * NAME: ix_rm_hw_queue_delete
 *
 * DESCRIPTION: This function deletes the specified queue. The Q Array entry will be freed.  
 * 
 * @Param:  - IN arg_hHwQueue - represents the queue handle that should be deleted.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_hw_queue_delete( 
                                ix_hw_queue_handle arg_hHwQueue 
                              )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_hw_queue_delete";

    /* Naming convention for automatic variables */
    ix_error err = 0;
    ix_uint32 memoryChannel;
    ix_uint32 queueIndex;
    ix_sram_qarray_channel_manager* pQArrayChannelManager = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_QArrayManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    IX_RM_ASSERT(IX_RM_HW_QUEUE_HANDLE_IS_VALID(arg_hHwQueue), s_aFunctionIdentifier, "bad queue handle!");

    memoryChannel = IX_RM_HW_QUEUE_GET_CHANNEL(arg_hHwQueue);
    IX_RM_ASSERT(memoryChannel < IX_SRAM_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");

    queueIndex = IX_RM_HW_QUEUE_GET_INDEX(arg_hHwQueue);
    IX_RM_ASSERT(queueIndex < IX_MAX_SRAM_QARRAY_ENTRIES, s_aFunctionIdentifier, "bad queue index!");

    pQArrayChannelManager = g_QArrayManager.m_aQArrayManagersPtr[memoryChannel];
    if (pQArrayChannelManager == NULL)
    {
        return IX_ERROR_NEW(IX_RM_ERROR_HW_QARRAY_MGR_ALLOC_FAILED, IX_ERROR_LEVEL_LOCAL);
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the QArray manager data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    {
        ix_uint32 entryUsage;

        /* This is the common memory, so we have to do a read, modify, write set of operations */
        entryUsage = IX_RM_MEM_UINT32_READ(&pQArrayChannelManager->m_aQArrayUsage[queueIndex]);

        IX_RM_SET_QARRAY_ENTRY_IN_USE(entryUsage, 0U);

        (void)IX_RM_MEM_UINT32_WRITE(&pQArrayChannelManager->m_aQArrayUsage[queueIndex], entryUsage);
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the QArray manager data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */


    return err;
}



/**
 * NAME: ix_rm_hw_enqueue
 *
 * DESCRIPTION: Enqueue an SRAM entry to the tail of specified queue. The first 4 bytes
 *          of each entry is reserved for hardware use. The caller needs to set it in the
 *          specified format as in [IXP2000-CAS]. Essentially the LSB 24 bits are used as
 *          a next address and the 8 MSB bits are used as a cell count, EOP and SOP flags. By
 *          default these will be set to 0xff.
 * 
 * @Param:  - IN arg_hHwQueue - queue handle where to enqueue the entry.
 * @Param:  - IN arg_hBuffer - the entry to be enqueued of type ix_buffer_handle. For
 *          queues only these handles can be interpreted by the hardware and enqueued
 *          and dequeued.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_hw_enqueue( 
                           ix_hw_queue_handle arg_hHwQueue, 
                           ix_buffer_handle arg_hBuffer 
                         )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_hw_enqueue";

    /* Naming convention for automatic variables */
    ix_error err = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check in debug mode the validity of the queue handle */
    IX_RM_ASSERT(IX_RM_HW_QUEUE_HANDLE_IS_VALID(arg_hHwQueue), s_aFunctionIdentifier, "bad queue handle!");

    IX_RM_ASSERT(IX_RM_HW_QUEUE_GET_CHANNEL(arg_hHwQueue) < IX_SRAM_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
    IX_RM_ASSERT(IX_RM_HW_QUEUE_GET_INDEX(arg_hHwQueue) < IX_MAX_SRAM_QARRAY_ENTRIES, s_aFunctionIdentifier, "bad queue index!");

    arg_hBuffer = (0xc000000 | IX_RM_BUF_HW_GET_META_OFFSET(arg_hBuffer));
    
#if defined(_IX_RM_IMPL_HARDWARE_)

    SRAM_QUEUE_ENQUEUE(IX_RM_HW_QUEUE_GET_CHANNEL(arg_hHwQueue),
        IX_RM_HW_QUEUE_GET_INDEX(arg_hHwQueue),
        arg_hBuffer);

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

    return err;
}



/**
 * NAME: ix_rm_hw_dequeue
 *
 * DESCRIPTION: This function will dequeue an SRAM entry from the specified queue.  
 * 
 * @Param:  - IN arg_hHwQueue - queue handle from where to dequeue the entry. 
 * @Param:  - OUT arg_pBufferHandle - location where to return the entry should 
 *          be stored. For queues the dequeued value is a 32 bit integer with a
 *          predefined packing which is modeled by the ix_buffer_handle type.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_hw_dequeue( 
                           ix_hw_queue_handle arg_hHwQueue, 
                           ix_buffer_handle* arg_pBufferHandle 
                         )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_hw_dequeue";

    /* Naming convention for automatic variables */
    ix_error err = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check in debug mode the validity of the queue handle */
    IX_RM_ASSERT(IX_RM_HW_QUEUE_HANDLE_IS_VALID(arg_hHwQueue), s_aFunctionIdentifier, "bad queue handle!");

    IX_RM_ASSERT(IX_RM_HW_QUEUE_GET_CHANNEL(arg_hHwQueue) < IX_SRAM_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
    IX_RM_ASSERT(IX_RM_HW_QUEUE_GET_INDEX(arg_hHwQueue) < IX_MAX_SRAM_QARRAY_ENTRIES, s_aFunctionIdentifier, "bad queue index!");

#if defined(_IX_RM_IMPL_HARDWARE_)

    *arg_pBufferHandle = (ix_buffer_handle)SRAM_QUEUE_DEQUEUE(
        IX_RM_HW_QUEUE_GET_CHANNEL(arg_hHwQueue),
        IX_RM_HW_QUEUE_GET_INDEX(arg_hHwQueue));

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

    return err;
}



/**
 * The scratch pad memory in IXP2000 support rings of various sizes. During scratch ring
 * creation, the Resource Manager initializes the ring registers with appropriate base
 * address and size fields. The scratch memory needed for the ring is also allocated. Once
 * the ring is created, the applications can call Resource Manager functions to put/get data
 * in these rings. These rings are accessible from microengines also. A total of 16 rings
 * are supported by the hardware and the ring numbers are from 0 to 15.
 * SRAM rings are supported by the Q-Array. The number of rings supported is restricted only
 * by the entries free in the Q-Array.
 */

/**
 * SRAM and SCRATCH ring functions
 */


/**
 * Static support functions for ring creation. They just specialize creation for
 * SRAM and SCRATCH rings. For details read the explanation from ix_rm_hw_ring_create().
 */

/**
 * NAME: ix_rm_hw_sram_ring_create
 *
 * DESCRIPTION: This function will create a ring in SRAM on the specified channel.
 *          The ring size is limited to several values  as specified by ix_sram_ring_size 
 *          enumerated type. The Resource Manager allocates necessary memory to hold the 
 *          ring data in SRAM. The handle returned can be used on the core to consume and 
 *          produce data to this ring.
 * 
 * @Param:  - IN arg_Channel - specifies the channel for the memory.
 * @Param:  - IN arg_RingSize - represents the ring element size as specified by the ix_sram_ring_size
 *          enumerated type.
 * @Param:  - OUT arg_pHwRingHandle - represents the address where created ring handle should be stored 
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_hw_sram_ring_create( 
                                    ix_uint32 arg_Channel,
                                    ix_sram_ring_size arg_RingSize,
                                    ix_hw_ring_handle* arg_pHwRingHandle 
                                  )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_hw_sram_ring_create";

    ix_error err = 0;
    ix_uint32 size;
    ix_uint32 alignment;
    ix_sram_qarray_channel_manager* pQArrayChannelManager = 0;
    ix_uint32 entryUsage;
    ix_uint32 i;
    ix_uint8* pMemory = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_QArrayManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(g_MicrocodeLoaded != 0)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_HW_RING_CREATE_FAILED, s_aFunctionIdentifier, 
                        "you can not create a sram hardware ring after the uCode has been loaded!", err);
    }

    IX_RM_ASSERT(arg_Channel < IX_SRAM_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
    pQArrayChannelManager = g_QArrayManager.m_aQArrayManagersPtr[arg_Channel];
    if (pQArrayChannelManager == NULL)
    {
        return IX_ERROR_NEW(IX_RM_ERROR_HW_QARRAY_MGR_ALLOC_FAILED, IX_ERROR_LEVEL_LOCAL);
    }

    /* get the right alignment */
    alignment = 11 + (ix_uint32)arg_RingSize;

    /* update the size */
    size = 1 << alignment;

    /* allocate the necessary memory from the coresponding SRAM channel */
    err = ix_rm_mem_alloc_aligned(IX_MEMORY_TYPE_SRAM, arg_Channel, size, 
        IX_MEMORY_ALIGNMENT_TYPE_PHYSICAL_OFFSET, alignment, (void**)&pMemory);
    if(err)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_HW_RING_CREATE_FAILED, s_aFunctionIdentifier, 
                        "ring memory allocation failed!", err);
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the QArray manager data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* find a QArray entry that is not in use */
    for(i = 0; i < IX_MAX_SRAM_QARRAY_ENTRIES; i++)
    {
        entryUsage = IX_RM_MEM_UINT32_READ(&pQArrayChannelManager->m_aQArrayUsage[i]);

        /* if in use go to the next entry */
        if(IX_RM_GET_QARRAY_ENTRY_IN_USE(entryUsage))
        {
            continue;
        }
        
        /* we found a free entry, mark it as a ring in use entry */
        entryUsage = 0;
        IX_RM_SET_QARRAY_ENTRY_IN_USE(entryUsage,1);
        IX_RM_SET_QARRAY_ENTRY_TYPE(entryUsage,IX_SRAM_QARRAY_ENTRY_TYPE_RING);
        (void)IX_RM_MEM_UINT32_WRITE(&pQArrayChannelManager->m_aQArrayUsage[i],entryUsage);
        break;
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the QArray manager data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* if we found an entry then setup the ring */
    if(i < IX_MAX_SRAM_QARRAY_ENTRIES)
    {
        ix_uint32 offset;
        ix_uint32 channel;
        ix_memory_type memType;
        void* physicalAddress;

        /* calculate the physical offset in bytes */
        err = ix_rm_get_phys_offset(pMemory, &memType, &channel, &offset, &physicalAddress);
        if(err)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_HW_RING_CREATE_FAILED, s_aFunctionIdentifier, 
                        "failed to get physical memory offset for the ring memory!", err);
            goto label_1;
        }

#if defined(_IX_RM_MULTI_THREADED_)

        /* initialize the operation implementation data */
        err = ix_ossl_sem_init(IX_OSSL_SEM_AVAILABLE, &g_QArrayManager.m_aRingOpImplData[arg_Channel][i].m_Semaphore);
        if(err != IX_SUCCESS)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_HW_RING_CREATE_FAILED, s_aFunctionIdentifier, 
                        "the creation of the operations semaphore failed!", err);
            goto label_1;
        }

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

        /* set the size of the ring in long words */
        g_QArrayManager.m_aRingOpImplData[arg_Channel][i].m_RingSize = (size >> 2);
        g_QArrayManager.m_aRingOpImplData[arg_Channel][i].m_pRingBaseAddress = (ix_uint32*)pMemory;

        /* calculate the offset in long words */
        offset >>= 2;

        {
            ix_sram_ring_descriptor* pRingDescriptor = 
                &pQArrayChannelManager->m_aQArrayDescriptors[i].m_RingDescriptor;

            /* initialize the QArray descriptor entry */
            (void)IX_RM_MEM_UINT32_WRITE(&pRingDescriptor->m_HeadInfo, 
                IX_MAKE_BIT_FIELD32(alignment - 11, 29U, 31U) | 
                IX_MAKE_BIT_FIELD32(offset, 0U, 23U));
            (void)IX_RM_MEM_UINT32_WRITE(&pRingDescriptor->m_TailInfo, 
                IX_MAKE_BIT_FIELD32(offset, 0U, 23U));
            IX_RM_MEM_UINT32_WRITE(&pRingDescriptor->m_CountInfo, 0);
            IX_RM_MEM_UINT32_WRITE(&pRingDescriptor->m_Reserved, 0);

        }

        /* set the ring handle */
        {
            ix_hw_ring_handle ringHandle = IX_NULL_HANDLE;

            IX_RM_HW_RING_SET_SIZE(ringHandle, alignment - 11);
            IX_RM_HW_RING_SET_CHANNEL(ringHandle, arg_Channel);
            IX_RM_HW_RING_SET_MEMORY_TYPE(ringHandle, IX_MEMORY_TYPE_SRAM);
            IX_RM_HW_RING_SET_INDEX(ringHandle, i);

            *arg_pHwRingHandle = ringHandle;
        }

    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_HW_RING_CREATE_FAILED, s_aFunctionIdentifier, 
                        "the reservation of QArray ring entry failed!", err);
        goto label_1;
    }

    return err;

label_1:

    /* free the allocated SRAM memory */
    if(pMemory != 0)
        ix_rm_mem_free(pMemory);

    /* make the entry available */
    entryUsage = 0;
    IX_RM_SET_QARRAY_ENTRY_IN_USE(entryUsage,0);
    (void)IX_RM_MEM_UINT32_WRITE(&pQArrayChannelManager->m_aQArrayUsage[i],entryUsage);

    return err;
}


/**
 * NAME: ix_rm_hw_scratch_ring_create
 *
 * DESCRIPTION: This function will create a ring in SRAM on the specified channel.
 *          The ring size is limited to several values  as specified by ix_scratch_ring_size 
 *          enumerated type. The Resource Manager allocates necessary memory to hold the 
 *          ring data in SCRATCH. The handle returned can be used on the core to consume and 
 *          produce data to this ring.
 * 
 * @Param:  - IN arg_Channel - specifies the channel for the memory.
 * @Param:  - IN arg_RingSize - represents the ring element size as specified by the ix_scratch_ring_size
 *          enumerated type.
 * @Param:  - IN arg_RingInternalId - specifies the internal scratch ring Id to be associated with the ring.
 *          If the specified Id is use then the call will fail. This approach has been taken due to limitation
 *          of the ME instruction set which requires that the ring Id should be a compile time constant.
 * @Param:  - OUT arg_pHwRingHandle - represents the address where created ring handle should be stored 
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_hw_scratch_ring_create( 
                                       ix_uint32 arg_Channel,
                                       ix_scratch_ring_size arg_RingSize,
                                       ix_uint32 arg_RingInternalId,
                                       ix_hw_ring_handle* arg_pHwRingHandle 
                                     )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_hw_scratch_ring_create";

    ix_error err = 0;
    ix_uint32 size;
    ix_uint32 alignment;
    ix_uint8* pMemory = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_SCRATCHRingManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    IX_RM_ASSERT(arg_RingInternalId < IX_MAX_SCRATCH_RINGS, s_aFunctionIdentifier, "bad scratch ring ID!");

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the SCRATCH ring manager data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    if(g_SCRATCHRingManager.m_aRingUsage[arg_RingInternalId] == 1)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_HW_RING_CREATE_FAILED, s_aFunctionIdentifier, 
                        "the requested SCRATCH ring is in use!", err);
    }
    else
    {
        /* set the ring as being in use */
        g_SCRATCHRingManager.m_aRingUsage[arg_RingInternalId] = 1;
    }        

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the SCRATCH ring manager data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    if(err != IX_SUCCESS)
        return err;

    /* get the right alignment */
    alignment = 9 + (ix_uint32)arg_RingSize;

    /* update the size */
    size = 1 << alignment;

    /* allocate the necessary memory from the coresponding SRAM channel */
    err = ix_rm_mem_alloc_aligned(IX_MEMORY_TYPE_SCRATCH, arg_Channel, size, 
        IX_MEMORY_ALIGNMENT_TYPE_PHYSICAL_OFFSET, alignment, (void**)&pMemory);
    if(err)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_HW_RING_CREATE_FAILED, s_aFunctionIdentifier, 
                        "ring memory allocation failed!", err);
        goto label_1;
    }

#if defined(_IX_RM_MULTI_THREADED_)

    /* initialize the operation implementation data */
    err = ix_ossl_sem_init(IX_OSSL_SEM_AVAILABLE, &g_SCRATCHRingManager.m_aRingOpImplData[arg_RingInternalId].m_Semaphore);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_HW_RING_CREATE_FAILED, s_aFunctionIdentifier, 
                    "the creation of the operations semaphore failed!", err);
        goto label_1;
    }

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    /* set the size of the ring in long words */
    g_SCRATCHRingManager.m_aRingOpImplData[arg_RingInternalId].m_RingSize = (size >> 2);

    {
        ix_uint32 value;
        ix_hw_ring_handle ringHandle = IX_NULL_HANDLE;
        ix_uint32 offset;
        ix_uint32 channel;
        ix_memory_type memType;
        void* physicalAddress;

        /* calculate the physical offset in bytes */
        err = ix_rm_get_phys_offset(pMemory, &memType, &channel, &offset, &physicalAddress);
        if(err)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_HW_RING_CREATE_FAILED, s_aFunctionIdentifier, 
                        "failed to get physical memory offset for the ring memory!", err);
            goto label_1;
        }

        IX_RM_HW_RING_SET_SIZE(ringHandle, alignment - 9);
        IX_RM_HW_RING_SET_CHANNEL(ringHandle, arg_Channel);
        IX_RM_HW_RING_SET_MEMORY_TYPE(ringHandle, IX_MEMORY_TYPE_SCRATCH);
        IX_RM_HW_RING_SET_INDEX(ringHandle, arg_RingInternalId);

        *arg_pHwRingHandle = ringHandle;

        /* initialize the SCRATCH ring CSRs */
        value = IX_MAKE_BIT_FIELD32(alignment - 9, 30U, 31U) |
            IX_MAKE_BIT_FIELD32(offset, 0U, 13U);

        
#if defined(_IX_RM_IMPL_HARDWARE_)

        SET_SCRATCH_CSR(RING_X_BASE, arg_RingInternalId, value);
        SET_SCRATCH_CSR(RING_X_HEAD, arg_RingInternalId, 0);
        SET_SCRATCH_CSR(RING_X_TAIL, arg_RingInternalId, 0);

        /* Added as a fix for a IXP2400 B0 board issue */
        {
            ix_uint32 temp;
            temp = GET_SCRATCH_CSR(RING_X_TAIL, arg_RingInternalId);
        }

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

    }

    return err;

label_1:

    /* free the allocated SRAM memory */
    if(pMemory != 0)
        ix_rm_mem_free(pMemory);

    /* set the ring as being in available */
    g_SCRATCHRingManager.m_aRingUsage[arg_RingInternalId] = 0;

    return err;
}



/**
 * Static support functions for ring deletion. They just specialize deletion for
 * SRAM and SCRATCH rings. For details read the explanation from ix_rm_hw_ring_delete().
 */
static 
ix_error _ix_rm_hw_sram_ring_delete( 
                                     ix_hw_ring_handle arg_HwRingHandle
                                   )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_hw_sram_ring_delete";

    ix_error err = 0;
    ix_uint32 memoryChannel;
    ix_uint32 ringIndex;

    ix_sram_qarray_channel_manager* pQArrayChannelManager = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_QArrayManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    memoryChannel = IX_RM_HW_RING_GET_CHANNEL(arg_HwRingHandle);
    IX_RM_ASSERT(memoryChannel < IX_SRAM_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");

    ringIndex = IX_RM_HW_RING_GET_INDEX(arg_HwRingHandle);
    IX_RM_ASSERT(ringIndex < IX_MAX_SRAM_QARRAY_ENTRIES, s_aFunctionIdentifier, "bad ring index!");

    pQArrayChannelManager = g_QArrayManager.m_aQArrayManagersPtr[memoryChannel];
    if (pQArrayChannelManager == NULL)
    {
        return IX_ERROR_NEW(IX_RM_ERROR_HW_QARRAY_MGR_ALLOC_FAILED, IX_ERROR_LEVEL_LOCAL);
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the QArray manager data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* set the size to 0 first in order to avoid data corruption */

    /* set the size of the ring in long words */
    g_QArrayManager.m_aRingOpImplData[memoryChannel][ringIndex].m_RingSize = 0;

#if defined(_IX_RM_MULTI_THREADED_)

    /* initialize the operation implementation data */
    if(g_QArrayManager.m_aRingOpImplData[memoryChannel][ringIndex].m_Semaphore != 0)
    {
        ix_ossl_sem_fini(g_QArrayManager.m_aRingOpImplData[memoryChannel][ringIndex].m_Semaphore);
        g_QArrayManager.m_aRingOpImplData[memoryChannel][ringIndex].m_Semaphore = 0;
    }

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

#if defined(_IX_RM_IMPL_HARDWARE_)

    {
        ix_sram_ring_descriptor* pRingDescriptor = 
            &pQArrayChannelManager->m_aQArrayDescriptors[ringIndex].m_RingDescriptor;

        /* reset the ring QArray entry */
        (void)IX_RM_MEM_UINT32_WRITE(&pRingDescriptor->m_HeadInfo,0);
        (void)IX_RM_MEM_UINT32_WRITE(&pRingDescriptor->m_TailInfo,0);
        (void)IX_RM_MEM_UINT32_WRITE(&pRingDescriptor->m_CountInfo, 0);
        (void)IX_RM_MEM_UINT32_WRITE(&pRingDescriptor->m_Reserved, 0);
    }

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    {
        ix_uint32 entryUsage;

        /* This is the common memory, so we have to do a read, modify, write set of operations */
        entryUsage = IX_RM_MEM_UINT32_READ(&pQArrayChannelManager->m_aQArrayUsage[ringIndex]);

        IX_RM_SET_QARRAY_ENTRY_IN_USE(entryUsage, 0U);

        (void)IX_RM_MEM_UINT32_WRITE(&pQArrayChannelManager->m_aQArrayUsage[ringIndex], entryUsage);
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the QArray manager data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    if(g_QArrayManager.m_aRingOpImplData[memoryChannel][ringIndex].m_pRingBaseAddress != 0)
    {
        /* release the ring memory */
        err = ix_rm_mem_free(g_QArrayManager.m_aRingOpImplData[memoryChannel][ringIndex].m_pRingBaseAddress);
        if(err)
        {
            return IX_RM_ERROR_NEW(IX_RM_ERROR_HW_RING_DELETE_FAILED, s_aFunctionIdentifier, 
                            "failed to release sram ring's memory!", 0);
        }
        g_QArrayManager.m_aRingOpImplData[memoryChannel][ringIndex].m_pRingBaseAddress = 0;
    }

    return err;
}


static 
ix_error _ix_rm_hw_scratch_ring_delete( 
                                        ix_hw_ring_handle arg_HwRingHandle
                                      )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_hw_scratch_ring_delete";

    ix_error err = 0;
    ix_uint32 memoryChannel;
    ix_uint32 ringIndex;

#if defined(_IX_RM_IMPL_HARDWARE_)

    ix_uint8* pMemory = 0;

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_SCRATCHRingManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    memoryChannel = IX_RM_HW_RING_GET_CHANNEL(arg_HwRingHandle);
    IX_RM_ASSERT(memoryChannel < IX_SCRATCH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");

    ringIndex = IX_RM_HW_RING_GET_INDEX(arg_HwRingHandle);
    IX_RM_ASSERT(ringIndex < IX_MAX_SCRATCH_RINGS, s_aFunctionIdentifier, "bad ring index!");

#if defined(_IX_RM_IMPL_HARDWARE_)

    err = ix_rm_get_virtual_address(IX_MEMORY_TYPE_SCRATCH, memoryChannel,
        IX_GET_BIT_FIELD32(GET_SCRATCH_CSR(RING_X_BASE, ringIndex), 0U, 13U), &pMemory);
    if(err)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_HW_RING_DELETE_FAILED, s_aFunctionIdentifier, 
                        "could not retrieve the virtual address of the scratch ring base!", 0);
    }

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */


#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the SCRATCH ring manager data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* set the size to 0 first in order to avoid data corruption */

    /* set the size of the ring in long words */
    g_SCRATCHRingManager.m_aRingOpImplData[ringIndex].m_RingSize = 0;

#if defined(_IX_RM_MULTI_THREADED_)

    /* finalize the operation implementation data */
    if(g_SCRATCHRingManager.m_aRingOpImplData[ringIndex].m_Semaphore != 0)
    {
        ix_ossl_sem_fini(g_SCRATCHRingManager.m_aRingOpImplData[ringIndex].m_Semaphore);
        g_SCRATCHRingManager.m_aRingOpImplData[ringIndex].m_Semaphore = 0;
    }

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    /* reset the ring and set it as inactive */
    SET_SCRATCH_CSR(RING_X_BASE, ringIndex, 0);
    SET_SCRATCH_CSR(RING_X_HEAD, ringIndex, 0);
    SET_SCRATCH_CSR(RING_X_TAIL, ringIndex, 0);

    /* Added as a fix for a IXP2400 B0 board issue */
    {
        ix_uint32 temp;
        temp = GET_SCRATCH_CSR(RING_X_TAIL, ringIndex);
    }

    g_SCRATCHRingManager.m_aRingUsage[ringIndex] = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the SCRATCH ring manager data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined(_IX_RM_IMPL_HARDWARE_)

    /* release the ring memory */
    err = ix_rm_mem_free(pMemory);
    if(err)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_HW_RING_DELETE_FAILED, s_aFunctionIdentifier, 
                        "failed to release scratch ring's memory!", 0);
    }

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    return err;
}

/**
 * NAME: ix_rm_hw_ring_delete
 *
 * DESCRIPTION: This function deletes the specified ring. The ring handle is freed and the
 *          Resource Manager reclaims allocated memory.
 * 
 * @Param:  - IN arg_HwRingHandle - represents the ring to be deleted.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_hw_ring_delete( 
                               ix_hw_ring_handle arg_HwRingHandle 
                             )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_hw_ring_delete";

    ix_uint32 memoryType;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    memoryType = IX_RM_HW_RING_GET_MEMORY_TYPE(arg_HwRingHandle);

    switch(memoryType)
    {
        case IX_MEMORY_TYPE_SRAM:
            return _ix_rm_hw_sram_ring_delete(arg_HwRingHandle);
        case IX_MEMORY_TYPE_SCRATCH:
            return _ix_rm_hw_scratch_ring_delete(arg_HwRingHandle);
        default:
            return IX_RM_ERROR_NEW(IX_RM_ERROR_BAD_MEMORY_TYPE, s_aFunctionIdentifier, 
                        "the memory type does not support rings!", 0);

    }
}



/**
 * NAME: ix_rm_hw_ring_put
 *
 * DESCRIPTION: This function will produce arg_DataSize 32 bit words from 
 *	        arg_pData address on the ring specified by arg_hHwRing. If the ring 
 *	        will not accomodate the entire data, then the function will return an 
 *	        error.
 * 
 * @Param:  - IN arg_hHwRing - represents the handle of the ring where data
 *	        will be stored.
 * @Param:  - INOUT arg_pDataSize - represents the number of 32 bit words to be 
 *	        written on the ring on the input and the actual number of 32 bit words
 *          written on the output.
 * @Param:  - IN arg_pData - represents the location of the data that is 
 *	        intended to be stored.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_hw_ring_put( 
                            ix_hw_ring_handle arg_hHwRing, 
                            ix_uint32* arg_pDataSize,
                            const ix_uint32* arg_pData 
                          )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_hw_ring_put";

    ix_error err = IX_SUCCESS;
    ix_uint32 size = *arg_pDataSize;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(size == 0)
        return IX_SUCCESS;


#if defined(_IX_RM_IMPL_HARDWARE_)

    switch(IX_RM_HW_RING_GET_MEMORY_TYPE(arg_hHwRing))
    {
        case IX_MEMORY_TYPE_SRAM:
            {
                register ix_uint32 channel = IX_RM_HW_RING_GET_CHANNEL(arg_hHwRing);
                register ix_uint32 ringIndex = IX_RM_HW_RING_GET_INDEX(arg_hHwRing);

                if(IX_RM_GET_QARRAY_ENTRY_IN_USE(g_QArrayManager.m_aQArrayManagersPtr[channel]->m_aQArrayUsage[ringIndex])==0)
                    return IX_RM_ERROR_NEW(IX_RM_ERROR_HW_SRAM_RING_PUT_FAILED, s_aFunctionIdentifier,
                                    "the passed SRAM ring is uninitialized!", err);
                else
                {
                    ix_ring_op_impl_data* pRingData = &g_QArrayManager.m_aRingOpImplData[channel][ringIndex];
                    ix_uint32 result;
                    ix_uint32* pCurrentSize = &g_QArrayManager.m_aQArrayManagersPtr[channel]->m_aQArrayDescriptors[ringIndex].m_RingDescriptor.m_Reserved;

#if defined(_IX_RM_MULTI_THREADED_)

                    ix_ossl_sem_t sem = pRingData->m_Semaphore;

                    ix_ossl_sem_take(sem, IX_OSSL_WAIT_FOREVER);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

                    if(size > 1)
                    {
                
                        if((size + IX_RM_MEM_UINT32_READ(pCurrentSize)) <= 
                            pRingData->m_RingSize)
                        {
                            for(;size != 0; size--,arg_pData++)
                            {
                                result = *arg_pData;
                                SRAM_RING_PUT(channel, ringIndex, result);
                                if(result == 0)
                                {
                                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_HW_SRAM_RING_PUT_FAILED, s_aFunctionIdentifier, 
                                        "the SRAM put operation failed!", err);
                                    break;
                                }
                                ix_rm_atomic_sram_add(1,pCurrentSize);
                            }
                        }
                        else
                        {
                            err = IX_RM_ERROR_NEW(IX_RM_ERROR_HW_SRAM_RING_PUT_FAILED, s_aFunctionIdentifier, 
                                "the ammount of data to be put is too large!", err);
                        }
                    }
                    else
                    {
                        if((size + IX_RM_MEM_UINT32_READ(pCurrentSize)) <= 
                            pRingData->m_RingSize)
                        {
                            result = *arg_pData;
                            SRAM_RING_PUT(channel, ringIndex, result);
                            if(result == 0)
                            {
                                err = IX_RM_ERROR_NEW(IX_RM_ERROR_HW_SRAM_RING_PUT_FAILED, s_aFunctionIdentifier, 
                                    "the SRAM put operation failed!", err);
                                break;
                            }
                            ix_rm_atomic_sram_add(1,pCurrentSize);
                            size--;
                        }
                        else
                        {
                            err = IX_RM_ERROR_NEW(IX_RM_ERROR_HW_SRAM_RING_PUT_FAILED, s_aFunctionIdentifier, 
                                "the ammount of data to be put is too large!", err);
                        }
                    }


#if defined(_IX_RM_MULTI_THREADED_)

                    ix_ossl_sem_give(sem);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

                }
            }
            break;
        case IX_MEMORY_TYPE_SCRATCH:
            {
                register ix_uint32 ringIndex = IX_RM_HW_RING_GET_INDEX(arg_hHwRing);

                if(g_SCRATCHRingManager.m_aRingUsage[ringIndex] == 0)
                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_HW_SCRATCH_RING_PUT_FAILED, s_aFunctionIdentifier, 
                            "the passed SCRATCH ring is uninitialized!", err);
                else
                {
                    
#if defined(_IX_RM_MULTI_THREADED_)

                    ix_ossl_sem_t sem = g_SCRATCHRingManager.m_aRingOpImplData[ringIndex].m_Semaphore;

                    ix_ossl_sem_take(sem, IX_OSSL_WAIT_FOREVER);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

                    if(size > 1)
                    {
                        
                        /**
                         * Ring full is declared at three quarters of the size, so check if the
                         * number of LW is less than 1/4 of the size
                         */

                        if(size <= g_SCRATCHRingManager.m_aRingOpImplData[ringIndex].m_RingSize >> 2)
                        {
                            
                            /**
                             * The check for ring full should be done just for the first LW.
                             * Logically we can go ahead and store the whole sequence without
                             * causing any problem. Even more, if we check for each LW then we
                             * can leave the ring in an inconsistent state if ring full condition
                             * is declared before we finish writing the whole sequence.
                             */

                            if(SCRATCH_RING_GET_IS_FULL_STATUS(ringIndex) != 0)
                            {
                                err = IX_RM_ERROR_NEW(IX_RM_ERROR_HW_SCRATCH_RING_PUT_FAILED, s_aFunctionIdentifier,
                                        "the SCRATCH ring is full!", err);
                            }
                            else
                            {
                                for( ;size != 0; size--, arg_pData++)
                                {
                                    SCRATCH_RING_PUT(ringIndex, *arg_pData);
                                }
                            }
                        }
                        else
                        {
                            err = IX_RM_ERROR_NEW(IX_RM_ERROR_HW_SCRATCH_RING_PUT_FAILED, s_aFunctionIdentifier,
                                "the ammount of data to be put is too large!", err);
                        }

                    }
                    else
                    {
                        if(SCRATCH_RING_GET_IS_FULL_STATUS(ringIndex) == 0)
                        {
                            SCRATCH_RING_PUT(ringIndex, *arg_pData);
                            size--;
                        }
                        else
                        {
                            err = IX_RM_ERROR_NEW(IX_RM_ERROR_HW_SCRATCH_RING_PUT_FAILED, s_aFunctionIdentifier, 
                                "the SCRATCH ring is full!", err);
                        }
                    }


#if defined(_IX_RM_MULTI_THREADED_)

                    ix_ossl_sem_give(sem);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

                }
            }
            break;
        default:
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_BAD_MEMORY_TYPE, s_aFunctionIdentifier, 
                    "the memory type does not support rings!", 0);

    }

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    *arg_pDataSize -= size;

    return err;
}


/**
 * NAME: ix_rm_hw_ring_get
 *
 * DESCRIPTION: This function will consume from the ring arg_DataSize 32 bit words and store
 *          them at the location specified by arg_pData. 
 * 
 * @Param:  - IN arg_hHwRing - represents the ring where the data is consumed from.
 * @Param:  - INOUT arg_pDataSize - represents the number of 32 bit words to be 
 *	        read from the ring on the input and the actual number of 32 bit words
 *          read on the output.
 * @Param:  - OUT arg_pData - the location where the consumed data should be stored. 
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure. 
 */
ix_error ix_rm_hw_ring_get( 
                            ix_hw_ring_handle arg_hHwRing, 
                            ix_uint32* arg_pDataSize, 
                            ix_uint32* arg_pData 
                          )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_hw_ring_get";

    /* Naming convention for automatic variables */
    ix_error err = 0;
    ix_uint32 size = *arg_pDataSize;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

    switch(IX_RM_HW_RING_GET_MEMORY_TYPE(arg_hHwRing))
    {
        case IX_MEMORY_TYPE_SRAM:
            {
                register ix_uint32 channel = IX_RM_HW_RING_GET_CHANNEL(arg_hHwRing);
                register ix_uint32 ringIndex = IX_RM_HW_RING_GET_INDEX(arg_hHwRing);

                if(IX_RM_GET_QARRAY_ENTRY_IN_USE(g_QArrayManager.m_aQArrayManagersPtr[channel]->m_aQArrayUsage[ringIndex]) == 0)
                    return IX_RM_ERROR_NEW(IX_RM_ERROR_HW_SRAM_RING_GET_FAILED, s_aFunctionIdentifier, 
                                    "the passed SRAM ring is uninitialized!", err);
                else
                {
                    ix_ring_op_impl_data* pRingData = &g_QArrayManager.m_aRingOpImplData[channel][ringIndex];
                    ix_uint32* pCurrentSize = &g_QArrayManager.m_aQArrayManagersPtr[channel]->m_aQArrayDescriptors[ringIndex].m_RingDescriptor.m_Reserved;


#if defined(_IX_RM_MULTI_THREADED_)

                    ix_ossl_sem_t sem = pRingData->m_Semaphore;

                    ix_ossl_sem_take(sem, IX_OSSL_WAIT_FOREVER);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

                    if(size > 1)
                    {

                        if(size <= IX_RM_MEM_UINT32_READ(pCurrentSize))
                        {
                            for(;size != 0; size--,arg_pData++)
                            {
                                *arg_pData = SRAM_RING_GET(IX_RM_HW_RING_GET_CHANNEL(arg_hHwRing), \
                                    IX_RM_HW_RING_GET_INDEX(arg_hHwRing));
                            }
                            ix_rm_atomic_sram_subtract(*arg_pDataSize,pCurrentSize);
                        }
                        else
                        {
                            err = IX_RM_ERROR_NEW(IX_RM_ERROR_HW_SRAM_RING_GET_FAILED, s_aFunctionIdentifier, 
                                    "not enough data into the ring!", IX_RM_MEM_UINT32_READ(pCurrentSize));
                        }
                    }
                    else
                    {
                        if(1 <= IX_RM_MEM_UINT32_READ(pCurrentSize))
                        {
                            *arg_pData = SRAM_RING_GET(IX_RM_HW_RING_GET_CHANNEL(arg_hHwRing), \
                                    IX_RM_HW_RING_GET_INDEX(arg_hHwRing));
                            ix_rm_atomic_sram_subtract(1,pCurrentSize);
                            size--;
                        }
                        else
                        {
                            err = IX_RM_ERROR_NEW(IX_RM_ERROR_HW_SRAM_RING_GET_FAILED, s_aFunctionIdentifier, 
                                    "not enough data into the ring!", IX_RM_MEM_UINT32_READ(pCurrentSize));
                        }
                    }

#if defined(_IX_RM_MULTI_THREADED_)

                    ix_ossl_sem_give(sem);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

                }
            }
            break;
        case IX_MEMORY_TYPE_SCRATCH:
            {
                register ix_uint32 ringIndex = IX_RM_HW_RING_GET_INDEX(arg_hHwRing);

                if(g_SCRATCHRingManager.m_aRingUsage[ringIndex] == 0)
                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_HW_SCRATCH_RING_GET_FAILED, s_aFunctionIdentifier, 
                            "the passed SCRATCH ring is uninitialized!", err);
                else
                {

#if defined(_IX_RM_MULTI_THREADED_)

                    ix_ossl_sem_t sem = g_SCRATCHRingManager.m_aRingOpImplData[ringIndex].m_Semaphore;

                    ix_ossl_sem_take(sem, IX_OSSL_WAIT_FOREVER);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

                    if(size > 1)
                    {
                        for(;size != 0; size--,arg_pData++)
                        {
                            /* there is no way to know if there is any data on the ring */
                            *arg_pData = SCRATCH_RING_GET(IX_RM_HW_RING_GET_INDEX(arg_hHwRing));
                        }

                    }
                    else
                    {
                        /* there is no way to know if there is any data on the ring */
                        *arg_pData = SCRATCH_RING_GET(IX_RM_HW_RING_GET_INDEX(arg_hHwRing));
                        size--;
                    }


#if defined(_IX_RM_MULTI_THREADED_)

                    ix_ossl_sem_give(sem);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

                }
            }
            break;
        default:
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_BAD_MEMORY_TYPE, s_aFunctionIdentifier, 
                        "the memory type does not support rings!", 0);

    }

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    *arg_pDataSize -= size;

    return err;
}

/**
 * NAME: ix_rm_hw_queue_array_get_base_address
 *
 * DESCRIPTION: This function will return the virtual base address of the core allocated
 *          queue descriptor array for each SRAM memory channel. Using this address the
 *          user can get the physical offset for the queue descriptor array by a call to
 *          ix_rm_get_phys_offset() call. This value has to be patched into the microcode
 *          that will perform the other half of the QArray initialization on each SRAM
 *          channel.
 * 
 * @Param:  - IN arg_Channel - represents the SRAM channel of interest.
 * @Param:  - IN arg_pQArrayBaseAddress - represents the address where the queue descriptor
 *          array base address will be stored upon return.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_hw_queue_array_get_base_address(
                                                ix_uint32 arg_Channel, 
                                                ix_uint32** arg_pQArrayBaseAddress
                                              )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_hw_queue_array_get_base_address";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    IX_RM_ASSERT(arg_Channel < IX_SRAM_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
    if (g_QArrayManager.m_aQArrayManagersPtr[arg_Channel] == NULL)
    {
        return IX_ERROR_NEW(IX_RM_ERROR_NULL_MEMORY_ADDRESS, IX_ERROR_LEVEL_LOCAL);
    }
    *arg_pQArrayBaseAddress = (ix_uint32*)g_QArrayManager.m_aQArrayManagersPtr[arg_Channel]->m_aQArrayDescriptors;

    return IX_SUCCESS;
}


#if defined (_IX_RM_DEBUG_) && defined (_IX_RM_IMPL_HARDWARE_)



/**
 * NAME: ix_rm_scratch_ring_print_info
 *
 * DESCRIPTION: This function will print information about the requested scratch ring 
 *  number.
 * 
 * @Param:  - IN arg_RingNumber - a number representing the scratch ring whose info
 *  we are interested in.
 *
 */
void ix_rm_scratch_ring_print_info(
                                    ix_uint32 arg_RingNumber
                                  )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_scratch_ring_print_info";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_RingNumber >= IX_MAX_SCRATCH_RINGS)
    {
        ix_ossl_message_log("Bad SCRATCH ring number. The numbers should be in the range [0, %d]!\n", IX_MAX_SCRATCH_RINGS - 1);
        return;
    }

    if(g_SCRATCHRingManager.m_aRingUsage[arg_RingNumber] == 0)
        ix_ossl_message_log("The SCRATCH ring %lu is not in use by the core!\n", arg_RingNumber);


    {
        ix_uint32 base;
        ix_uint32 head;
        ix_uint32 tail;
        ix_uint32 size;

        base = GET_SCRATCH_CSR(RING_X_BASE, arg_RingNumber);
        head = GET_SCRATCH_CSR(RING_X_HEAD, arg_RingNumber);
        tail = GET_SCRATCH_CSR(RING_X_TAIL, arg_RingNumber);

        ix_ossl_message_log("Information for SCRATCH ring %lu: \n", arg_RingNumber);

        size = IX_GET_BIT_FIELD32(base, 30U, 31U);
        size = 128 << size;
        ix_ossl_message_log("\tSize: %lu 32bit words\n", size);

        ix_ossl_message_log("\tRing base SCRATCH offset: 0x%08lx\n",IX_GET_BIT_FIELD32(base, 0U, 13U));
        ix_ossl_message_log("\tRing base virtual address: 0x%p\n",(ix_uint8*)((ix_uint32)Hal_scratch_rd_wr_swap_virtAddr) + IX_GET_BIT_FIELD32(base, 0U, 13U));

        ix_ossl_message_log("\tRing head index: %lu\n", head & (size - 1));
        ix_ossl_message_log("\tRing tail index: %lu\n", tail & (size - 1));
    }

    return;
}


/**
 * NAME: ix_rm_scratch_ring_print_data
 *
 * DESCRIPTION: This function will print the data existing in the requested scratch ring 
 *  number.
 * 
 * @Param:  - IN arg_RingNumber - a number representing the scratch ring whose data
 *  we are interested in.
 *
 */
void ix_rm_scratch_ring_print_data(
                                    ix_uint32 arg_RingNumber
                                  )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_scratch_ring_print_data";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_RingNumber >= IX_MAX_SCRATCH_RINGS)
    {
        ix_ossl_message_log("Bad SCRATCH ring number. The numbers should be in the range [0, %d]!\n", IX_MAX_SCRATCH_RINGS - 1);
        return;
    }

    if(g_SCRATCHRingManager.m_aRingUsage[arg_RingNumber] == 0)
        ix_ossl_message_log("The SCRATCH ring %lu is not in use by the core!\n", arg_RingNumber);


    {
        ix_uint32 base;
        ix_uint32 size;
        ix_uint32 i;
        ix_uint32* pBase = 0;

        base = GET_SCRATCH_CSR(RING_X_BASE, arg_RingNumber);

        ix_ossl_message_log("Data for SCRATCH ring %lu: \n", arg_RingNumber);

        size = IX_GET_BIT_FIELD32(base, 30U, 31U);
        size = 128 << size;

        pBase = (ix_uint32*)((ix_uint8*)((ix_uint32)Hal_scratch_rd_wr_swap_virtAddr) + IX_GET_BIT_FIELD32(base, 0U, 13U));

        for(i = 0; i < size; i += 4)
        {
            ix_ossl_message_log(" 0x%08lx 0x%08lx ", pBase[i], pBase[i+1]);
            ix_ossl_message_log(" 0x%08lx 0x%08lx \n", pBase[i+2], pBase[i+3]);
        };
    }

    return;
}



/**
 * NAME: ix_rm_sram_ring_print_info
 *
 * DESCRIPTION: This function will print information about the requested sram ring 
 *  number.
 * 
 * @Param:  - IN arg_ChannelNumber - the channel number of the SRAM ring of interest.
 * @Param:  - IN arg_RingNumber - a number representing the sram ring whose info
 *  we are interested in.
 *
 */
void ix_rm_sram_ring_print_info(
                                 ix_uint32 arg_ChannelNumber,
                                 ix_uint32 arg_RingNumber
                               )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_sram_ring_print_info";

    ix_sram_qarray_channel_manager* pQArrayChannelManager = 0;
    ix_uint32 entryUsage;

    (void)s_aFunctionIdentifier;

    if(arg_ChannelNumber >= IX_SRAM_CHANNELS_NUMBER)
    {
        ix_ossl_message_log("The requested SRAM channel number is invalid!\n");
        return;
    }

    pQArrayChannelManager = g_QArrayManager.m_aQArrayManagersPtr[arg_ChannelNumber];
    if (pQArrayChannelManager == NULL)
    {
        ix_ossl_message_log("No QArray manager for SRAM channel %lu\n",
                            arg_ChannelNumber);
        return;
    }

    entryUsage = IX_RM_MEM_UINT32_READ(&pQArrayChannelManager->m_aQArrayUsage[arg_RingNumber]);

    if(IX_RM_GET_QARRAY_ENTRY_IN_USE(entryUsage) == 0)
    {
        ix_ossl_message_log("The coresponding QArray descriptor for SRAM ring %lu is not initialized!\n", arg_RingNumber);
        return;
    }
        
    if(IX_RM_GET_QARRAY_ENTRY_TYPE(entryUsage) != IX_SRAM_QARRAY_ENTRY_TYPE_RING)
    {
        ix_ossl_message_log("The coresponding QArray descriptor for SRAM ring %lu is initialized as a queue!\n", arg_RingNumber);
        return;
    }

    {
        ix_uint32 size;
        ix_uint32 head;
        ix_uint32 tail;
        ix_uint32 count;
        ix_uint8* pBase;
        ix_uint32 offset;
        ix_sram_ring_descriptor* pRingDescriptor = 
            &pQArrayChannelManager->m_aQArrayDescriptors[arg_RingNumber].m_RingDescriptor;

        /* initialize the QArray descriptor entry */
        head = IX_RM_MEM_UINT32_READ(&pRingDescriptor->m_HeadInfo);
        size = 512 << IX_GET_BIT_FIELD32(head, 29U, 31U);
        /* get the head offset in bytes */
        head = IX_GET_BIT_FIELD32(head, 0U, 23U) << 2;

        tail = IX_RM_MEM_UINT32_READ(&pRingDescriptor->m_TailInfo);
        /* get the tail offset in bytes */
        tail = IX_GET_BIT_FIELD32(tail, 0U, 23U) << 2;

        count = IX_RM_MEM_UINT32_READ(&pRingDescriptor->m_CountInfo);
        count = IX_GET_BIT_FIELD32(count, 0U, 23U);


        ix_ossl_message_log("Information for SRAM ring %lu on channel %lu: \n", arg_RingNumber, arg_ChannelNumber);

        ix_ossl_message_log("\tRing size in 32bit words: %lu\n", size);

        pBase = (ix_uint8*)g_QArrayManager.m_aRingOpImplData[arg_ChannelNumber][arg_RingNumber].m_pRingBaseAddress;

        if(ix_rm_get_phys_offset(pBase, 0, 0, &offset, 0) != IX_SUCCESS)
        {
            ix_ossl_message_log("Can not retrieve base physical offset for the SRAM ring!\n");
            return;
        }

        ix_ossl_message_log("\tRing base offset:          0x%08lx\n", offset);
        ix_ossl_message_log("\tRing base virtual address: 0x%p\n", pBase);

        if(ix_rm_get_virtual_address(IX_MEMORY_TYPE_SRAM,
            arg_ChannelNumber, head, &pBase) != IX_SUCCESS)
        {
            ix_ossl_message_log("Can not retrieve the virtual address of the head pointer for the SRAM ring!\n");
            return;
        }

        ix_ossl_message_log("\tRing head offset:          0x%08lx\n", head);
        ix_ossl_message_log("\tRing head virtual address: 0x%p\n", pBase);

        if(ix_rm_get_virtual_address(IX_MEMORY_TYPE_SRAM, 
            arg_ChannelNumber, tail, &pBase) != IX_SUCCESS)
        {
            ix_ossl_message_log("Can not retrieve the virtual address of the tail pointer for the SRAM ring!\n");
            return;
        }
        ix_ossl_message_log("\tRing tail offset:          0x%08lx\n", tail);
        ix_ossl_message_log("\tRing tail virtual address: 0x%p\n", pBase);
    }

    return;
}


/**
 * NAME: ix_rm_sram_ring_print_data
 *
 * DESCRIPTION: This function will print the data existing in the requested sram ring 
 *  number.
 * 
 * @Param:  - IN arg_ChannelNumber - the channel number of the SRAM ring of interest.
 * @Param:  - IN arg_RingNumber - a number representing the sram ring whose data
 *  we are interested in.
 *
 */
void ix_rm_sram_ring_print_data(
                                 ix_uint32 arg_ChannelNumber,
                                 ix_uint32 arg_RingNumber
                               )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_sram_ring_print_data";

    ix_sram_qarray_channel_manager* pQArrayChannelManager = 0;
    ix_uint32 entryUsage;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_ChannelNumber >= IX_SRAM_CHANNELS_NUMBER)
    {
        ix_ossl_message_log("The requested SRAM channel number is invalid!\n");
        return;
    }

    pQArrayChannelManager = g_QArrayManager.m_aQArrayManagersPtr[arg_ChannelNumber];
    if (pQArrayChannelManager == NULL)
    {
        ix_ossl_message_log("No QArray manager for SRAM channel %lu\n",
                            arg_ChannelNumber);
        return;
    }

    entryUsage = IX_RM_MEM_UINT32_READ(&pQArrayChannelManager->m_aQArrayUsage[arg_RingNumber]);

    if(IX_RM_GET_QARRAY_ENTRY_IN_USE(entryUsage) == 0)
    {
        ix_ossl_message_log("The coresponding QArray descriptor for SRAM ring %lu is not initialized!\n", arg_RingNumber);
        return;
    }
        
    if(IX_RM_GET_QARRAY_ENTRY_TYPE(entryUsage) != IX_SRAM_QARRAY_ENTRY_TYPE_RING)
    {
        ix_ossl_message_log("The coresponding QArray descriptor for SRAM ring %lu is initialized as a queue!\n", arg_RingNumber);
        return;
    }

    {
        ix_uint32 size;
        ix_uint32 i;
        ix_uint32* pBase;
        ix_sram_ring_descriptor* pRingDescriptor = 
            &pQArrayChannelManager->m_aQArrayDescriptors[arg_RingNumber].m_RingDescriptor;

        /* initialize the QArray descriptor entry */
        size = IX_RM_MEM_UINT32_READ(&pRingDescriptor->m_HeadInfo);
        size = 512 << IX_GET_BIT_FIELD32(size, 29U, 31U);

        ix_ossl_message_log("Data for SRAM ring %lu on channel %lu: \n", arg_RingNumber, arg_ChannelNumber);

        pBase = (ix_uint32*)g_QArrayManager.m_aRingOpImplData[arg_ChannelNumber][arg_RingNumber].m_pRingBaseAddress;

        for(i = 0; i < size; i += 4)
        {
            ix_ossl_message_log(" 0x%08lx 0x%08lx ", pBase[i], pBase[i+1]);
            ix_ossl_message_log(" 0x%08lx 0x%08lx \n", pBase[i+2] , pBase[i+3]);
        };
    }

    return;
}


#endif /* defined (_IX_RM_DEBUG_) && defined (_IX_RM_IMPL_HARDWARE_) */
