/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      hash.c
 *
 * = DESCRIPTION
 *      This file defines all functions and variables required by the Hash API
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      11/21/2002 10:50:28 AM - creation time 
 * ============================================================================
 */

static const char s_FileIdentifier[] = "$Id: hash.c,v 1.6 2002/12/05 20:22:50 ageorges Exp $";

#include "ix_ossl.h"

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"

#include "rm/symbols.h"
#include "rm/internal/internal_support.h"

#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"

/* #define _IX_HASH_APPLY_SWAP_*/

#include "rm/hash.h"
#include "rm/internal/internal_hash.h"



/**
 * NAME: ix_rm_hash_48_hash
 *
 * DESCRIPTION: This function will perform a 48 bit hash operation
 * 
 * @Param:  - INOUT - arg_pHash48 on input holds the data to be hashed and on
 *      output contains the hashed value.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_hash_48_hash(
                             ix_hash_48* arg_pHash48
                           )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_hash_48_hash";

    ix_error err = IX_SUCCESS;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_HashAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined(_IX_RM_IMPL_HARDWARE_HASH_) 

    {
        register ix_uint32 value;

        /* write data into the hash registers in order lw3,lw2,lw1,lw0 */
        value = arg_pHash48->m_LW1;
        *((volatile ix_uint32*)(IX_HASH_OPERAND_BASE_ADDRESS + IX_HASH_48_OPERAND_LW1_OFFSET)) =
            IX_HASH_32BIT_SWAP(value);

        value = arg_pHash48->m_LW0;
        *((volatile ix_uint32*)(IX_HASH_OPERAND_BASE_ADDRESS + IX_HASH_48_OPERAND_LW0_OFFSET)) =
            IX_HASH_32BIT_SWAP(value);

        IX_RM_HASH_WAIT_FOR_COMPLETION();

        /* read the result of the hash operation */
        value = *((ix_uint32*)(IX_HASH_OPERAND_BASE_ADDRESS + IX_HASH_48_OPERAND_LW0_OFFSET));
        arg_pHash48->m_LW0 = IX_HASH_32BIT_SWAP(value);

        value = *((ix_uint32*)(IX_HASH_OPERAND_BASE_ADDRESS + IX_HASH_48_OPERAND_LW1_OFFSET));
        arg_pHash48->m_LW1 = IX_HASH_32BIT_SWAP(value);
    }

#else

    err = _ix_rm_hash_48_software_hash(arg_pHash48);

#endif /* end defined(_IX_RM_IMPL_HARDWARE_HASH_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: ix_rm_hash_48_multiplier_set
 *
 * DESCRIPTION: This function will set the multiplier for the 48 bit hash
 * 
 * @Param:  - IN arg_pHashMultiplier48 - the new value of the 48 bit hash multiplier.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_hash_48_multiplier_set(
                                       const ix_hash_multiplier_48* arg_pHashMultiplier48
                                     )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_hash_48_multiplier_set";

    ix_error err = IX_SUCCESS;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_HashAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined(_IX_RM_IMPL_HARDWARE_HASH_)
    
    {
        register ix_uint32 value;

        value = arg_pHashMultiplier48->m_LW0;
        *((volatile ix_uint32*)(IX_HASH_MULTIPLIER_BASE_ADDRESS + IX_HASH_48_MULTIPLIER_LW0_OFFSET)) =
            IX_HASH_32BIT_SWAP(value);

        value = arg_pHashMultiplier48->m_LW1;
        *((volatile ix_uint32*)(IX_HASH_MULTIPLIER_BASE_ADDRESS + IX_HASH_48_MULTIPLIER_LW1_OFFSET)) =
            IX_HASH_32BIT_SWAP(value);
    }
    
#else

    g_HashAPIManager.m_HashMultiplier48.m_LW0 = arg_pHashMultiplier48->m_LW0;
    g_HashAPIManager.m_HashMultiplier48.m_LW1 = arg_pHashMultiplier48->m_LW1;

#endif /* end defined(_IX_RM_IMPL_HARDWARE_HASH_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: ix_rm_hash_48_multiplier_get
 *
 * DESCRIPTION: This function will retrieves the multiplier for the 48 bit hash
 * 
 * @Param:  - OUT arg_pHashMultiplier48 - the locationwhere the 48 bit hash multiplier
 *      will be stored on successful return
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_hash_48_multiplier_get(
                                       ix_hash_multiplier_48* arg_pHashMultiplier48
                                     )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_hash_48_multiplier_get";

    ix_error err = IX_SUCCESS;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_HashAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined(_IX_RM_IMPL_HARDWARE_HASH_) 

    {
        register ix_uint32 value;

        value = *((ix_uint32*)(IX_HASH_MULTIPLIER_BASE_ADDRESS + IX_HASH_48_MULTIPLIER_LW0_OFFSET));
        arg_pHashMultiplier48->m_LW0 = IX_HASH_32BIT_SWAP(value);
            
        value = *((ix_uint32*)(IX_HASH_MULTIPLIER_BASE_ADDRESS + IX_HASH_48_MULTIPLIER_LW1_OFFSET));
        arg_pHashMultiplier48->m_LW1 = IX_HASH_32BIT_SWAP(value);
    }
    
#else

    arg_pHashMultiplier48->m_LW0 = g_HashAPIManager.m_HashMultiplier48.m_LW0;
    arg_pHashMultiplier48->m_LW1 = g_HashAPIManager.m_HashMultiplier48.m_LW1;

#endif /* end defined(_IX_RM_IMPL_HARDWARE_HASH_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: ix_rm_hash_64_hash
 *
 * DESCRIPTION: This function will perform a 64 bit hash operation
 * 
 * @Param:  - INOUT - arg_pHash64 on input holds the data to be hashed and on
 *      output contains the hashed value.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_hash_64_hash(
                             ix_hash_64* arg_pHash64
                           )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_hash_64_hash";

    ix_error err = IX_SUCCESS;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_HashAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined(_IX_RM_IMPL_HARDWARE_HASH_) 

    {
        register ix_uint32 value;

        /* write data into the hash registers in order lw3,lw2,lw1,lw0 */
        value = arg_pHash64->m_LW1;
        *((volatile ix_uint32*)(IX_HASH_OPERAND_BASE_ADDRESS + IX_HASH_64_OPERAND_LW1_OFFSET)) =
            IX_HASH_32BIT_SWAP(value);

        value = arg_pHash64->m_LW0;
        *((volatile ix_uint32*)(IX_HASH_OPERAND_BASE_ADDRESS + IX_HASH_64_OPERAND_LW0_OFFSET)) =
            IX_HASH_32BIT_SWAP(value);

        IX_RM_HASH_WAIT_FOR_COMPLETION();

        /* read the result of the hash operation */
        value = *((ix_uint32*)(IX_HASH_OPERAND_BASE_ADDRESS + IX_HASH_64_OPERAND_LW0_OFFSET));
        arg_pHash64->m_LW0 = IX_HASH_32BIT_SWAP(value);
        
        value = *((ix_uint32*)(IX_HASH_OPERAND_BASE_ADDRESS + IX_HASH_64_OPERAND_LW1_OFFSET));
        arg_pHash64->m_LW1 = IX_HASH_32BIT_SWAP(value);
    }

#else

    err = _ix_rm_hash_64_software_hash(arg_pHash64);

#endif /* end defined(_IX_RM_IMPL_HARDWARE_HASH_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: ix_rm_hash_64_multiplier_set
 *
 * DESCRIPTION: This function will set the multiplier for the 64 bit hash
 * 
 * @Param:  - IN arg_pHashMultiplier64 - the new value of the 64 bit hash multiplier.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_hash_64_multiplier_set(
                                       const ix_hash_multiplier_64* arg_pHashMultiplier64
                                     )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_hash_64_multiplier_set";

    ix_error err = IX_SUCCESS;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_HashAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined(_IX_RM_IMPL_HARDWARE_HASH_) 

    {
        register ix_uint32 value;

        value = arg_pHashMultiplier64->m_LW0;
        *((volatile ix_uint32*)(IX_HASH_MULTIPLIER_BASE_ADDRESS + IX_HASH_64_MULTIPLIER_LW0_OFFSET)) =
            IX_HASH_32BIT_SWAP(value);

        value = arg_pHashMultiplier64->m_LW1;
        *((volatile ix_uint32*)(IX_HASH_MULTIPLIER_BASE_ADDRESS + IX_HASH_64_MULTIPLIER_LW1_OFFSET)) =
            IX_HASH_32BIT_SWAP(value);
    }
    
#else

    g_HashAPIManager.m_HashMultiplier64.m_LW0 = arg_pHashMultiplier64->m_LW0;
    g_HashAPIManager.m_HashMultiplier64.m_LW1 = arg_pHashMultiplier64->m_LW1;

#endif /* end defined(_IX_RM_IMPL_HARDWARE_HASH_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: ix_rm_hash_64_multiplier_get
 *
 * DESCRIPTION: This function will retrieves the multiplier for the 64 bit hash
 * 
 * @Param:  - OUT arg_pHashMultiplier64 - the locationwhere the 64 bit hash multiplier
 *      will be stored on successful return
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_hash_64_multiplier_get(
                                       ix_hash_multiplier_64* arg_pHashMultiplier64
                                     )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_hash_64_multiplier_get";

    ix_error err = IX_SUCCESS;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_HashAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined(_IX_RM_IMPL_HARDWARE_HASH_) 

    {
        register ix_uint32 value;

        value = *((ix_uint32*)(IX_HASH_MULTIPLIER_BASE_ADDRESS + IX_HASH_64_MULTIPLIER_LW0_OFFSET));
        arg_pHashMultiplier64->m_LW0 = IX_HASH_32BIT_SWAP(value);
        
        value = *((ix_uint32*)(IX_HASH_MULTIPLIER_BASE_ADDRESS + IX_HASH_64_MULTIPLIER_LW1_OFFSET));
        arg_pHashMultiplier64->m_LW1 = IX_HASH_32BIT_SWAP(value);

    }

#else

    arg_pHashMultiplier64->m_LW0 = g_HashAPIManager.m_HashMultiplier64.m_LW0;
    arg_pHashMultiplier64->m_LW1 = g_HashAPIManager.m_HashMultiplier64.m_LW1;

#endif /* end defined(_IX_RM_IMPL_HARDWARE_HASH_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: ix_rm_hash_128_hash
 *
 * DESCRIPTION: This function will perform a 128 bit hash operation
 * 
 * @Param:  - INOUT - arg_pHash128 on input holds the data to be hashed and on
 *      output contains the hashed value.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_hash_128_hash(
                              ix_hash_128* arg_pHash128
                            )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_hash_128_hash";

    ix_error err = IX_SUCCESS;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_HashAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined(_IX_RM_IMPL_HARDWARE_HASH_) 

    {
        register ix_uint32 value;

        /* write data into the hash registers in order lw3,lw2,lw1,lw0 */
        value = arg_pHash128->m_LW3;
        *((volatile ix_uint32*)(IX_HASH_OPERAND_BASE_ADDRESS + IX_HASH_128_OPERAND_LW3_OFFSET)) =
            IX_HASH_32BIT_SWAP(value);

        value = arg_pHash128->m_LW2;
        *((volatile ix_uint32*)(IX_HASH_OPERAND_BASE_ADDRESS + IX_HASH_128_OPERAND_LW2_OFFSET)) =
            IX_HASH_32BIT_SWAP(value);

        value = arg_pHash128->m_LW1;
        *((volatile ix_uint32*)(IX_HASH_OPERAND_BASE_ADDRESS + IX_HASH_128_OPERAND_LW1_OFFSET)) =
            IX_HASH_32BIT_SWAP(value);

        value = arg_pHash128->m_LW0;
        *((volatile ix_uint32*)(IX_HASH_OPERAND_BASE_ADDRESS + IX_HASH_128_OPERAND_LW0_OFFSET)) =
            IX_HASH_32BIT_SWAP(value);

        IX_RM_HASH_WAIT_FOR_COMPLETION();

        /* read the result of the hash operation */
        value = *((ix_uint32*)(IX_HASH_OPERAND_BASE_ADDRESS + IX_HASH_128_OPERAND_LW0_OFFSET));
        arg_pHash128->m_LW0 = IX_HASH_32BIT_SWAP(value);

        value = *((ix_uint32*)(IX_HASH_OPERAND_BASE_ADDRESS + IX_HASH_128_OPERAND_LW1_OFFSET));
        arg_pHash128->m_LW1 = IX_HASH_32BIT_SWAP(value);

        value = *((ix_uint32*)(IX_HASH_OPERAND_BASE_ADDRESS + IX_HASH_128_OPERAND_LW2_OFFSET));
        arg_pHash128->m_LW2 = IX_HASH_32BIT_SWAP(value);

        value = *((ix_uint32*)(IX_HASH_OPERAND_BASE_ADDRESS + IX_HASH_128_OPERAND_LW3_OFFSET));
        arg_pHash128->m_LW3 = IX_HASH_32BIT_SWAP(value);
    }

#else

    err = _ix_rm_hash_128_software_hash(arg_pHash128);

#endif /* end defined(_IX_RM_IMPL_HARDWARE_HASH_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: ix_rm_hash_128_multiplier_set
 *
 * DESCRIPTION: This function will set the multiplier for the 128 bit hash
 * 
 * @Param:  - IN arg_pHashMultiplier128 - the new value of the 128 bit hash multiplier.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_hash_128_multiplier_set(
                                        const ix_hash_multiplier_128* arg_pHashMultiplier128
                                      )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_hash_128_multiplier_set";

    ix_error err = IX_SUCCESS;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_HashAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined(_IX_RM_IMPL_HARDWARE_HASH_) 

    {
        register ix_uint32 value;

        value = arg_pHashMultiplier128->m_LW0;
        *((volatile ix_uint32*)(IX_HASH_MULTIPLIER_BASE_ADDRESS + IX_HASH_128_MULTIPLIER_LW0_OFFSET)) =
            IX_HASH_32BIT_SWAP(value);

        value = arg_pHashMultiplier128->m_LW1;
        *((volatile ix_uint32*)(IX_HASH_MULTIPLIER_BASE_ADDRESS + IX_HASH_128_MULTIPLIER_LW1_OFFSET)) =
            IX_HASH_32BIT_SWAP(value);

        value = arg_pHashMultiplier128->m_LW2;
        *((volatile ix_uint32*)(IX_HASH_MULTIPLIER_BASE_ADDRESS + IX_HASH_128_MULTIPLIER_LW2_OFFSET)) =
            IX_HASH_32BIT_SWAP(value);

        value = arg_pHashMultiplier128->m_LW3;
        *((volatile ix_uint32*)(IX_HASH_MULTIPLIER_BASE_ADDRESS + IX_HASH_128_MULTIPLIER_LW3_OFFSET)) =
            IX_HASH_32BIT_SWAP(value);
    }
    
#else

    g_HashAPIManager.m_HashMultiplier128.m_LW0 = arg_pHashMultiplier128->m_LW0;
    g_HashAPIManager.m_HashMultiplier128.m_LW1 = arg_pHashMultiplier128->m_LW1;
    g_HashAPIManager.m_HashMultiplier128.m_LW2 = arg_pHashMultiplier128->m_LW2;
    g_HashAPIManager.m_HashMultiplier128.m_LW3 = arg_pHashMultiplier128->m_LW3;

#endif /* end defined(_IX_RM_IMPL_HARDWARE_HASH_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: ix_rm_hash_128_multiplier_get
 *
 * DESCRIPTION: This function will retrieves the multiplier for the 128 bit hash
 * 
 * @Param:  - OUT arg_pHashMultiplier128 - the locationwhere the 128 bit hash multiplier
 *      will be stored on successful return
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_hash_128_multiplier_get(
                                        ix_hash_multiplier_128* arg_pHashMultiplier128
                                      )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_hash_128_multiplier_get";

    ix_error err = IX_SUCCESS;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_HashAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined(_IX_RM_IMPL_HARDWARE_HASH_) 

    {
        register ix_uint32 value;

        value = *((ix_uint32*)(IX_HASH_MULTIPLIER_BASE_ADDRESS + IX_HASH_128_MULTIPLIER_LW0_OFFSET));
        arg_pHashMultiplier128->m_LW0 = IX_HASH_32BIT_SWAP(value);

        value = *((ix_uint32*)(IX_HASH_MULTIPLIER_BASE_ADDRESS + IX_HASH_128_MULTIPLIER_LW1_OFFSET));
        arg_pHashMultiplier128->m_LW1 = IX_HASH_32BIT_SWAP(value);

        value = *((ix_uint32*)(IX_HASH_MULTIPLIER_BASE_ADDRESS + IX_HASH_128_MULTIPLIER_LW2_OFFSET));
        arg_pHashMultiplier128->m_LW2 = IX_HASH_32BIT_SWAP(value);

        value = *((ix_uint32*)(IX_HASH_MULTIPLIER_BASE_ADDRESS + IX_HASH_128_MULTIPLIER_LW3_OFFSET));
        arg_pHashMultiplier128->m_LW3 = IX_HASH_32BIT_SWAP(value);
    }

#else

    arg_pHashMultiplier128->m_LW0 = g_HashAPIManager.m_HashMultiplier128.m_LW0;
    arg_pHashMultiplier128->m_LW1 = g_HashAPIManager.m_HashMultiplier128.m_LW1;
    arg_pHashMultiplier128->m_LW2 = g_HashAPIManager.m_HashMultiplier128.m_LW2;
    arg_pHashMultiplier128->m_LW3 = g_HashAPIManager.m_HashMultiplier128.m_LW3;

#endif /* end defined(_IX_RM_IMPL_HARDWARE_HASH_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}

