/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      internal_buffer.c
 *
 * = DESCRIPTION
 *      This file will define all the internal supporting functions for the 
 *      Buffer API.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      4/23/2002 10:13:46 AM - creation time 
 * ============================================================================
 */

static const char s_aFileIdentifier[] = "$Id: internal_buffer.c,v 1.41 2003/10/13 18:45:06 ageorges Exp $";

#include "ix_ossl.h"

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"

#include "rm/symbols.h"
#include "rm/internal/internal_support.h"

#include <hal_sram.h>

#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"


#include "rm/system.h"
#include "rm/internal/internal_system.h"

#include "rm/memory.h"

#include "rm/buffer.h"
#include "rm/internal/internal_buffer.h"

#include "rm/hardware.h"
#include "rm/internal/internal_hardware.h"

/* Just to have the Hw buffer working */
/* #define _IX_RM_LIB_DEVELOPMENT_ 1 */

/**
 * This global variable represents an array of available free lists in this system.
 */
_ix_buffer_free_list_manager g_FLManager;


#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

/**
 * SRAM QArray entry ID array coresponding the first buffer free list
 * These values will need to be patched into the uCode.
 */
ix_uint32 g_StartBufferFreeListQArrayID[IX_SRAM_CHANNELS_NUMBER];

ix_uint8* _g_aSRAMChannelsStartAddress[IX_SRAM_CHANNELS_NUMBER];

#else
 
/**
 * SRAM QArray entry ID on the active channel coresponding the first buffer free list
 * This value will need to be patched into the uCode.
 */
ix_uint32 g_StartBufferFreeListQArrayID;

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */



#if defined( _IX_RM_MULTI_THREADED_ )

/**
 * Synchronization semaphore for free list manager
 */
static ix_ossl_sem_t s_FreeListManagerSemaphore = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */



/**
 * The following are the default hardware free list VMT functions.
 */

/**
 * NAME: _ix_rm_hw_free_list_fini
 *
 * DESCRIPTION: This function will perform the proper cleanup for a hardware free list.
 * 
 * @Param:  - IN arg_pHwFreeList - address of the hardware free list structure this function
 *          will work on.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_hw_free_list_fini( 
                                   _ix_buffer_free_list* arg_pHwFreeList 
                                 )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_hw_free_list_fini";

    ix_error err = IX_SUCCESS;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* destroy the memory buffer free list and free the associated memory */
    {
        /* declare some local variables */

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

        ix_uint32 freeListID = arg_pHwFreeList - (_ix_buffer_free_list*)g_FLManager.m_aHwFreeList;

#else

        ix_uint32 freeListID = arg_pHwFreeList - g_FLManager.m_aHwFreeList;

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */

        ix_sram_queue_descriptor* pQueueDescriptor;

        /**
         * Reset the underlying SRAM queue.
         */

        /* finalize the QArray descriptor entry */

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

        pQueueDescriptor = &g_QArrayManager.m_aQArrayManagersPtr[arg_pHwFreeList->m_MetaMemoryChannel]->
            m_aQArrayDescriptors[freeListID + g_StartBufferFreeListQArrayID[arg_pHwFreeList->m_MetaMemoryChannel]].m_QueueDescriptor;

#else

        pQueueDescriptor = &g_QArrayManager.m_aQArrayManagersPtr[arg_pHwFreeList->m_MetaMemoryChannel]->
            m_aQArrayDescriptors[freeListID + g_StartBufferFreeListQArrayID].m_QueueDescriptor;

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */

        IX_RM_MEM_UINT32_WRITE(&pQueueDescriptor->m_HeadInfo, 0);
        IX_RM_MEM_UINT32_WRITE(&pQueueDescriptor->m_TailInfo, 0);
        IX_RM_MEM_UINT32_WRITE(&pQueueDescriptor->m_CountInfo, 0);
        IX_RM_MEM_UINT32_WRITE(&pQueueDescriptor->m_Reserved, 0);

        /* free all memory */

#if defined (_IX_RM_BUFFER_DEBUG_)

        /* release the allocated debug info memory */
        ix_rm_mem_free(arg_pHwFreeList->m_pDebugInfo);

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

        /* release the allocated data memory */
        ix_rm_mem_free(arg_pHwFreeList->m_pDataBaseAddress);

        /* release the allocated meta memory */
        ix_rm_mem_free(arg_pHwFreeList->m_pMetaBaseAddress);

#if defined(_IX_RM_SPLIT_META_DATA_)

        /* release the allocated debug info memory */
        ix_rm_mem_free(arg_pHwFreeList->m_FLData.m_HwData.m_pInternalMetaBaseAddress);

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

        /* cleanup the free list entry */
        ix_ossl_memset(arg_pHwFreeList,0,sizeof(_ix_buffer_free_list));

        g_FLManager.m_aFreeListAddress[freeListID] = 0;
    }

    return err;
}



/**
 * NAME: _ix_rm_hw_buffer_allocator
 *
 * DESCRIPTION: This function will get a new hardware buffer from the hardware free list.
 * 
 * @Param:  - IN arg_pHwFreeList - address of the hardware free list structure this function
 *          will work on.
 * @Param:  - OUT arg_pHwBufferHandle - address where the handle of the newly allocated 
 *          hardware buffer will be stored.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_hw_buffer_allocator( 
                                     _ix_buffer_free_list* arg_pHwFreeList, 
                                     ix_buffer_handle* arg_pHwBufferHandle
                                   )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_hw_buffer_allocator";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    *arg_pHwBufferHandle = IX_NULL_BUFFER_HANDLE;

#if defined( _IX_RM_LIB_DEVELOPMENT_ )

    {
        if((*arg_pHwBufferHandle = arg_pHwFreeList->m_FreeListHead) != IX_NULL_BUFFER_HANDLE)
        {

#if defined(_IX_RM_SPLIT_META_DATA_)

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

            /* add the SRAM channel info to the buffer handle */
            *arg_pHwBufferHandle |= arg_pHwFreeList->m_FLData.m_HwData.m_InternalMetaMemoryChannel << 1;

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */

            arg_pHwFreeList->m_FreeListHead = IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(*arg_pHwBufferHandle)->m_HwNext;

#else

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

            /* add the SRAM channel info to the buffer handle */
            *arg_pHwBufferHandle |= arg_pHwFreeList->m_MetaMemoryChannel << 1;

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */

            arg_pHwFreeList->m_FreeListHead = IX_RM_BUF_HW_GET_META_ADDRESS(*arg_pHwBufferHandle)->m_HwNext;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

#if defined (_IX_RM_BUFFER_DEBUG_)

            {
                ix_buffer_debug_info* pDebugInfo;
                if(_ix_rm_buffer_debug_info_get(*arg_pHwBufferHandle, &pDebugInfo) == IX_SUCCESS)
                {
                    pDebugInfo->m_hBuffer = *arg_pHwBufferHandle;
                    pDebugInfo->m_LastCoreAllocationStatus = IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_ALLOCATED_BY_CORE;
                    pDebugInfo->m_SendAsMessageCount = 0;
                    pDebugInfo->m_SendAsPacketCount = 0;
                }
            }


#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

            return IX_SUCCESS;
        }
        else
            return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_NO_FREE_BUFFERS, s_aFunctionIdentifier, 
                    "free list out of free buffers!", 0);    
    }

#else

    {

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)
    
#if defined(_IX_RM_SPLIT_META_DATA_)

        ix_uint32 channel = arg_pHwFreeList->m_FLData.m_HwData.m_InternalMetaMemoryChannel;

#else

        ix_uint32 channel = arg_pHwFreeList->m_MetaMemoryChannel;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

        if((*arg_pHwBufferHandle = SRAM_QUEUE_DEQUEUE(channel 
            ,g_StartBufferFreeListQArrayID[channel] + 
            (arg_pHwFreeList - g_FLManager.m_aHwFreeList[channel])))
            == IX_NULL_BUFFER_HANDLE)

#else

        if((*arg_pHwBufferHandle = SRAM_QUEUE_DEQUEUE(arg_pHwFreeList->m_MetaMemoryChannel 
            ,g_StartBufferFreeListQArrayID + (arg_pHwFreeList - g_FLManager.m_aHwFreeList)))
            == IX_NULL_BUFFER_HANDLE)

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */

            return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_NO_FREE_BUFFERS, s_aFunctionIdentifier, 
                        "free list out of free buffers!", 0);
        else
        {

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

            /* set the channel number in the handle */
            *arg_pHwBufferHandle |= channel << 1;
            
#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */

#if defined (_IX_RM_BUFFER_DEBUG_)

            {
                ix_buffer_debug_info* pDebugInfo;
                if(_ix_rm_buffer_debug_info_get(*arg_pHwBufferHandle, &pDebugInfo) == IX_SUCCESS)
                {
                    pDebugInfo->m_hBuffer = *arg_pHwBufferHandle;
                    pDebugInfo->m_LastCoreAllocationStatus = IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_ALLOCATED_BY_CORE;
                    pDebugInfo->m_SendAsMessageCount = 0;
                    pDebugInfo->m_SendAsPacketCount = 0;
                }
            }


#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

            return IX_SUCCESS;
        }

    }

#endif /* defined( _IX_RM_LIB_DEVELOPMENT_ ) */

}



/**
 * NAME: _ix_rm_hw_buffer_deallocator
 *
 * DESCRIPTION: This function will return the specified hardware buffer to the free list.
 * 
 * @Param:  - IN arg_pHwFreeList - address of the hardware free list structure this function
 *          will work on.
 * @Param:  - IN arg_hHwBuffer - the handle of the hardware buffer to be deleted.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_hw_buffer_deallocator( 
                                     _ix_buffer_free_list* arg_pHwFreeList, 
                                     ix_buffer_handle arg_hHwBuffer
                                   )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_hw_buffer_deallocator";

    ix_error err = IX_SUCCESS;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_LIB_DEVELOPMENT_ )

    /* return the buffer to the free list */

#if defined(_IX_RM_SPLIT_META_DATA_)

    IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(arg_hHwBuffer)->m_HwNext = arg_pHwFreeList->m_FreeListHead;

#else

    IX_RM_BUF_HW_GET_META_ADDRESS(arg_hHwBuffer)->m_HwNext = arg_pHwFreeList->m_FreeListHead;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

#if defined (_IX_RM_BUFFER_DEBUG_)

        {
            ix_buffer_debug_info* pDebugInfo;
            if(_ix_rm_buffer_debug_info_get(arg_hHwBuffer, &pDebugInfo) == IX_SUCCESS)
            {
                pDebugInfo->m_hBuffer |= IX_RM_SINGLE_BUFFER_MASK;
                if(pDebugInfo->m_LastCoreAllocationStatus == IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_FREE)
                {
                    ix_ossl_message_log("The buffer with the handle 0x%08lx has been freed before by the core!\n", pDebugInfo->m_hBuffer);
                }
                else
                {
                    pDebugInfo->m_LastCoreAllocationStatus = IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_FREE;
                    pDebugInfo->m_SendAsMessageCount = 0;
                    pDebugInfo->m_SendAsPacketCount = 0;
                }
            }
        }


#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

    /* remove the channel info from the handle */
    arg_hHwBuffer &= 0xfffffff8;
    
    arg_pHwFreeList->m_FreeListHead = arg_hHwBuffer | IX_RM_SINGLE_BUFFER_MASK;

#else

    {

#if defined (_IX_RM_BUFFER_DEBUG_)

        {
            ix_buffer_debug_info* pDebugInfo;
            if(_ix_rm_buffer_debug_info_get(arg_hHwBuffer, &pDebugInfo) == IX_SUCCESS)
            {
                pDebugInfo->m_hBuffer |= IX_RM_SINGLE_BUFFER_MASK;
                if(pDebugInfo->m_LastCoreAllocationStatus == IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_FREE)
                {
                    ix_ossl_message_log("The buffer with the handle 0x%08lx has been freed before by the core!\n", arg_hHwBuffer);
                }
                else
                {
                    pDebugInfo->m_LastCoreAllocationStatus = IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_FREE;
                    pDebugInfo->m_SendAsMessageCount = 0;
                    pDebugInfo->m_SendAsPacketCount = 0;
                }
            }
        }


#endif /* defined (_IX_RM_BUFFER_DEBUG_) */


        arg_hHwBuffer = IX_RM_SINGLE_BUFFER_MASK | IX_RM_BUF_HW_GET_META_OFFSET(arg_hHwBuffer);

        {

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

            ix_uint32 channel = arg_pHwFreeList->m_MetaMemoryChannel;

            /* remove the channel info from the handle */
            arg_hHwBuffer &= 0xfffffff8;

            SRAM_QUEUE_ENQUEUE(channel ,
                 g_StartBufferFreeListQArrayID[channel] + 
                 (arg_pHwFreeList - g_FLManager.m_aHwFreeList[channel]), arg_hHwBuffer);

#else

            SRAM_QUEUE_ENQUEUE(arg_pHwFreeList->m_MetaMemoryChannel , 
                 g_StartBufferFreeListQArrayID + (arg_pHwFreeList - g_FLManager.m_aHwFreeList), arg_hHwBuffer);

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */

        }

    }

#endif /* defined( _IX_RM_LIB_DEVELOPMENT_ ) */

    return err;
}



/**
 * The following are the default software free list VMT functions.
 */

/**
 * NAME: _ix_rm_sw_free_list_fini
 *
 * DESCRIPTION: This function will perform the proper cleanup for a software free list.
 * 
 * @Param:  - IN arg_pSwFreeList - address of the software free list structure this function
 *          will work on.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_sw_free_list_fini( 
                                   _ix_buffer_free_list* arg_pSwFreeList 
                                 )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_sw_free_list_fini";

    ix_error err = IX_SUCCESS;

#if defined( _IX_RM_COMPONENT_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &arg_pSwFreeList->m_FLData.m_SwData.m_SyncAdapter;

#endif /* defined( _IX_RM_COMPONENT_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_COMPONENT_MULTI_THREADED_ )

    /* lock the free list data */
    if(pSyncAdapter->m_SyncActive)
    {
        pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);
    }

#endif /* defined( _IX_RM_COMPONENT_MULTI_THREADED_ ) */

    /* check if there are any buffers in use */
    if(arg_pSwFreeList->m_NumElements > arg_pSwFreeList->m_FLData.m_SwData.m_NumFreeElements)
    {
        /* if there are still cells in use then just mark the free list for deletion */
        arg_pSwFreeList->m_FLData.m_SwData.m_MarkedForDeletion = 1;
    }
    else
    {
        /* free all memory */

#if defined (_IX_RM_BUFFER_DEBUG_)

        /* release the allocated debug info memory */
        ix_rm_mem_free(arg_pSwFreeList->m_pDebugInfo);

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

        /* release the allocated meta memory */
        ix_rm_mem_free(arg_pSwFreeList->m_pMetaBaseAddress);

        /* release the allocated data memory */
        ix_rm_mem_free(arg_pSwFreeList->m_pDataBaseAddress);

        arg_pSwFreeList->m_FLData.m_SwData.m_MarkedForDeletion = 0;
        arg_pSwFreeList->m_FLData.m_SwData.m_NumFreeElements = 0;
        arg_pSwFreeList->m_FreeListHead = IX_NULL_BUFFER_HANDLE;

        /* cleanup the free list manager entry */
        memset(arg_pSwFreeList,0,sizeof(_ix_buffer_free_list));
        g_FLManager.m_aFreeListAddress[(arg_pSwFreeList - g_FLManager.m_aSwFreeList) + IX_BUF_FIRST_SW_FL_ID] = 0;

    }

#if defined( _IX_RM_COMPONENT_MULTI_THREADED_ )

    /* unlock the free list data */
    if(pSyncAdapter->m_SyncActive)
    {
        pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);
    }

#endif /* defined( _IX_RM_COMPONENT_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: _ix_rm_sw_buffer_allocator
 *
 * DESCRIPTION: This function will get a new software buffer from the software free list.
 * 
 * @Param:  - IN arg_pSwFreeList - address of the software free list structure this function
 *          will work on.
 * @Param:  - OUT arg_pSwBufferHandle - address where the handle of the newly allocated 
 *          software buffer will be stored.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_sw_buffer_allocator( 
                                     _ix_buffer_free_list* arg_pSwFreeList, 
                                     ix_buffer_handle* arg_pSwBufferHandle
                                   )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_sw_buffer_allocator";

    ix_error err = IX_SUCCESS;

    ix_sw_buffer_meta* pBufferMeta = 0;

#if defined( _IX_RM_COMPONENT_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &arg_pSwFreeList->m_FLData.m_SwData.m_SyncAdapter;

#endif /* defined( _IX_RM_COMPONENT_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_COMPONENT_MULTI_THREADED_ )

    /* lock the free list data */
    if(pSyncAdapter->m_SyncActive)
    {
        pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);
    }

#endif /* defined( _IX_RM_COMPONENT_MULTI_THREADED_ ) */

    /* check if there are any buffers free */
    if(arg_pSwFreeList->m_FLData.m_SwData.m_NumFreeElements > 0)
    {
        register ix_buffer_handle bufferHandle = arg_pSwFreeList->m_FreeListHead;

#if defined (_IX_RM_BUFFER_DEBUG_)

        {
            ix_buffer_debug_info* pDebugInfo;
            if(_ix_rm_buffer_debug_info_get(bufferHandle, &pDebugInfo) == IX_SUCCESS)
            {
                pDebugInfo->m_hBuffer = bufferHandle;
                pDebugInfo->m_LastCoreAllocationStatus = IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_ALLOCATED_BY_CORE;
                pDebugInfo->m_SendAsMessageCount = 0;
                pDebugInfo->m_SendAsPacketCount = 0;
            }
        }

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */


        /* get the buffer meta */
        pBufferMeta = (ix_sw_buffer_meta*)((ix_uint8*)arg_pSwFreeList->m_pMetaBaseAddress + 
            (IX_RM_BUF_SW_GET_INDEX(bufferHandle) << arg_pSwFreeList->m_MetaElementShift));

        /* show that the buffer is in use */
        pBufferMeta->m_hSelf = bufferHandle;

        /* set the free list head to the next buffer */
        arg_pSwFreeList->m_FreeListHead = pBufferMeta->m_hNext;

        /* decrement the number of free buffers */
        arg_pSwFreeList->m_FLData.m_SwData.m_NumFreeElements--;

        /* return the head of the list */
        *arg_pSwBufferHandle = bufferHandle;
    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_NO_FREE_BUFFERS, s_aFunctionIdentifier, 
                        "free list out of free buffers!", err);
    }

#if defined( _IX_RM_COMPONENT_MULTI_THREADED_ )

    /* unlock the free list data */
    if(pSyncAdapter->m_SyncActive)
    {
        pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);
    }

#endif /* defined( _IX_RM_COMPONENT_MULTI_THREADED_ ) */
    return err;
}



/**
 * NAME: _ix_rm_sw_buffer_deallocator
 *
 * DESCRIPTION: This function will return the specified software buffer to the free list.
 * 
 * @Param:  - IN arg_pSwFreeList - address of the software free list structure this function
 *          will work on.
 * @Param:  - IN arg_hSwBuffer - the handle of the software buffer to be deleted.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_sw_buffer_deallocator( 
                                     _ix_buffer_free_list* arg_pSwFreeList, 
                                     ix_buffer_handle arg_hSwBuffer
                                   )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_sw_buffer_deallocator";

    ix_error err = IX_SUCCESS;

    ix_sw_buffer_meta* pBufferMeta = 0;

#if defined( _IX_RM_COMPONENT_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &arg_pSwFreeList->m_FLData.m_SwData.m_SyncAdapter;

#endif /* defined( _IX_RM_COMPONENT_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_COMPONENT_MULTI_THREADED_ )

    /* lock the free list data */
    if(pSyncAdapter->m_SyncActive)
    {
        pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);
    }

#endif /* defined( _IX_RM_COMPONENT_MULTI_THREADED_ ) */

#if defined (_IX_RM_BUFFER_DEBUG_)

    {
        ix_buffer_debug_info* pDebugInfo;
        if(_ix_rm_buffer_debug_info_get(arg_hSwBuffer, &pDebugInfo) == IX_SUCCESS)
        {
            pDebugInfo->m_hBuffer |= IX_RM_SINGLE_BUFFER_MASK;
            if(pDebugInfo->m_LastCoreAllocationStatus == IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_FREE)
            {
                ix_ossl_message_log("The buffer with the handle 0x%08lx has been freed before by the core!\n", pDebugInfo->m_hBuffer);
            }
            else
            {
                pDebugInfo->m_LastCoreAllocationStatus = IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_FREE;
                pDebugInfo->m_SendAsMessageCount = 0;
                pDebugInfo->m_SendAsPacketCount = 0;
            }
        }
    }

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */


    /* get the buffer's meta */
    pBufferMeta = (ix_sw_buffer_meta*)((ix_uint8*)arg_pSwFreeList->m_pMetaBaseAddress + 
            (IX_RM_BUF_SW_GET_INDEX(arg_hSwBuffer) << arg_pSwFreeList->m_MetaElementShift));

    if(pBufferMeta->m_hSelf != IX_NULL_BUFFER_HANDLE)
    {
        /* mark the buffer as free */
        pBufferMeta->m_hSelf = IX_NULL_BUFFER_HANDLE;

        /* put the released buffer at the beginning of the free list */
        pBufferMeta->m_hNext = arg_pSwFreeList->m_FreeListHead;

        /* set the flags properly so that for the next allocation we don't have problems */
        arg_hSwBuffer |= IX_RM_SINGLE_BUFFER_MASK;

        arg_pSwFreeList->m_FreeListHead = arg_hSwBuffer;

        /* increment the number of free cells */
        arg_pSwFreeList->m_FLData.m_SwData.m_NumFreeElements++;

        /* check if the free list is marked for deletion */
        if(arg_pSwFreeList->m_FLData.m_SwData.m_MarkedForDeletion)
        {
            if(arg_pSwFreeList->m_FLData.m_SwData.m_NumFreeElements == arg_pSwFreeList->m_NumElements)
            {
                /* free all memory */

                /* release the allocated meta memory */
                err = ix_rm_mem_free(arg_pSwFreeList->m_pMetaBaseAddress);

                /* release the allocated data memory */
                err = ix_rm_mem_free(arg_pSwFreeList->m_pDataBaseAddress);

                arg_pSwFreeList->m_FLData.m_SwData.m_MarkedForDeletion = 0;
                arg_pSwFreeList->m_FLData.m_SwData.m_NumFreeElements = 0;
                arg_pSwFreeList->m_FreeListHead = IX_NULL_BUFFER_HANDLE;

                /* cleanup the free list manager entry */
                memset(arg_pSwFreeList,0,sizeof(_ix_buffer_free_list));
                g_FLManager.m_aFreeListAddress[(arg_pSwFreeList - g_FLManager.m_aSwFreeList) + IX_BUF_FIRST_SW_FL_ID] = 0;
            }
        }
    }


#if defined( _IX_RM_COMPONENT_MULTI_THREADED_ )

    /* unlock the free list data */
    if(pSyncAdapter->m_SyncActive)
    {
        pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);
    }

#endif /* defined( _IX_RM_COMPONENT_MULTI_THREADED_ ) */

    return err;
}



#if defined( _IX_RM_MULTI_THREADED_ )

/**
 * NAME: _ix_rm_fl_manager_init_sync
 *
 * DESCRIPTION: This function will initialize the synchronization for the free list
 *  manager. A detailed explanation on how sychronization mechanism works is found in 
 *  the L0 HLD/LLD and also in sync_mechanism.h file.
 * 
 * @Param:  - IN arg_LockFunction - this is the lock function of the sync mechanism.
 * @Param:  - IN arg_UnlockFunction - this is the unlock function of the sync mechanism.
 * @Param:  - IN arg_pContext - context that will be passed to each invocation of lock
 *          and unlock function. It might point to a synchronization primitive to be
 *          used by the specified lock and unlock functions.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_fl_manager_init_sync(
                                      ix_sa_lock_ptr            arg_LockFunction,
                                      ix_sa_unlock_ptr          arg_UnlockFunction,
                                      void*                     arg_pContext
                                    )    
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_fl_manager_init_sync";

    ix_error err = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((arg_LockFunction != 0) && (arg_UnlockFunction != 0))
    {
        g_FLManager.m_SyncAdapter.m_pContext = arg_pContext;
        g_FLManager.m_SyncAdapter.m_LockFuncPtr = arg_LockFunction;
        g_FLManager.m_SyncAdapter.m_UnlockFuncPtr = arg_UnlockFunction;

        /* set the synchronization mechanism in active state */
        g_FLManager.m_SyncAdapter.m_SyncActive = 1;
    }
    else
    {
        g_FLManager.m_SyncAdapter.m_SyncActive = 0;
    }

    return err;
}

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */


/**
 * NAME: _ix_rm_buffer_init
 *
 * DESCRIPTION: This function should be called at the time the resource manager initializer 
 *          function is called. The function will initialize all the Buffer API.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_buffer_init(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_buffer_init";

    ix_error err = 0;
    ix_hw_queue_handle hQueue;

	IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    err = ix_ossl_sem_init(IX_OSSL_SEM_AVAILABLE, &s_FreeListManagerSemaphore);
    if(err)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_API_INIT_FAILED, s_aFunctionIdentifier, 
                        "initialization of free list manager's semaphore failed!", err);
    }

    err = _ix_rm_fl_manager_init_sync(_ix_rm_sa_default_lock, _ix_rm_sa_default_unlock, 
        (void*)&s_FreeListManagerSemaphore); 
    if(err)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_API_INIT_FAILED, s_aFunctionIdentifier, 
                        "synchronization mechanism initialization for free list manager failed!", err);
    }

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

    /* Initialize the SRAM memory start address array */

    _g_aSRAMChannelsStartAddress[0] = (ix_uint8*)IX_SRAM_CH0_START_ADDRESS;
    _g_aSRAMChannelsStartAddress[1] = (ix_uint8*)IX_SRAM_CH1_START_ADDRESS;

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

    _g_aSRAMChannelsStartAddress[2] = (ix_uint8*)IX_SRAM_CH2_START_ADDRESS;
    _g_aSRAMChannelsStartAddress[3] = (ix_uint8*)IX_SRAM_CH3_START_ADDRESS;

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */


#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

    {
        ix_uint32 i;

        /* allocate QArray entries dedicated to buffer free lists for each channel */

        for(i = 0; i < IX_SRAM_CHANNELS_NUMBER; i++)
        {
            err = ix_rm_hw_queue_create(i, IX_BUF_MAX_HW_FL_NUMBER, 0, &hQueue);
            if(err != IX_SUCCESS)
            {
                if(IX_ERROR_GET_CODE(err) != IX_RM_ERROR_HW_QARRAY_MGR_ALLOC_FAILED)
                    return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_QUEUES_ALLOC_FAILED, s_aFunctionIdentifier, 
                                "allocation of buffer free list QArray entries failed!", err);
                else
                {
                    err = IX_SUCCESS; /* reset the error code in this case */
                    continue;
                }

            }  
            /* Set the first QArray descriptor ID coresponding to buffer free lists */
            g_StartBufferFreeListQArrayID[i] = IX_RM_HW_QUEUE_GET_INDEX(hQueue);

        }
    }

#else

    /* allocate QArray entries dedicated to buffer free lists */
    err = ix_rm_hw_queue_create(IX_HW_FL_META_CHANNEL, IX_BUF_MAX_HW_FL_NUMBER, 0, &hQueue);
    if(err != IX_SUCCESS)
    {
        if(IX_ERROR_GET_CODE(err) != IX_RM_ERROR_HW_QARRAY_MGR_ALLOC_FAILED)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_QUEUES_ALLOC_FAILED, s_aFunctionIdentifier, 
                        "allocation of buffer free list QArray entries failed!", err);
        else
            err = IX_SUCCESS; /* reset the error code in this case */
    } 

    /* Set the first QArray descriptor ID coresponding to buffer free lists */
    g_StartBufferFreeListQArrayID = IX_RM_HW_QUEUE_GET_INDEX(hQueue);

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */


    return err;
}


/**
 * NAME: _ix_rm_buffer_fini
 *
 * DESCRIPTION: This function should be called at the time the resource manager finalizer function
 *          is called. The function will provide cleanup for Buffer API.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_buffer_fini(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_buffer_fini";

    ix_error err = 0;
    ix_hw_queue_handle hQueue;
    ix_uint32 i;
    _ix_buffer_free_list* pFreeList;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_FLManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

	IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */


    /* delete all in use free lists */
    for(i = 0; i < IX_BUF_MAX_HW_FL_NUMBER; i++)
    {
        pFreeList = g_FLManager.m_aFreeListAddress[i];
        if(pFreeList != 0)
        {
            /* list is still in use so delete it */
            _ix_rm_hw_free_list_fini(pFreeList);
        }
    }

    for(i = IX_BUF_MAX_HW_FL_NUMBER; i < IX_BUF_MAX_FL_NUMBER; i++)
    {
        pFreeList = g_FLManager.m_aFreeListAddress[i];
        if(pFreeList != 0)
        {
            /* list is still in use so delete it */
            pFreeList->m_FLData.m_SwData.m_NumFreeElements = pFreeList->m_NumElements;
            _ix_rm_sw_free_list_fini(pFreeList);
        }
    }

#if defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_)

    {
        ix_uint32 j;

        for(j = 0; j < IX_SRAM_CHANNELS_NUMBER; j++)
        {
            /* release all QArray entries dedicated to buffer free lists */
            for(i = 0; i < IX_BUF_MAX_HW_FL_NUMBER; i++)
            {
                hQueue = IX_RM_HW_QUEUE_CREATE_HANDLE(j, i + g_StartBufferFreeListQArrayID[j]);
                err = ix_rm_hw_queue_delete(hQueue);
            }
        }
    }

#else

    /* release all QArray entries dedicated to buffer free lists */
    for(i = 0; i < IX_BUF_MAX_HW_FL_NUMBER; i++)
    {
        hQueue = IX_RM_HW_QUEUE_CREATE_HANDLE(IX_HW_FL_META_CHANNEL, i + g_StartBufferFreeListQArrayID);
        err = ix_rm_hw_queue_delete(hQueue);
    }

#endif /* end defined(_IX_RM_MULTIPLE_CHANNEL_HW_FREE_LIST_) */


    /* finalize the synchronization mechanism */
#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

    _ix_rm_fl_manager_init_sync(0, 0, (void*)0); 

    err = ix_ossl_sem_fini(s_FreeListManagerSemaphore);
    if(err != 0)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_API_FINI_FAILED, s_aFunctionIdentifier, 
                        "finalization of free list manager's semaphore failed!", err);
    }

    s_FreeListManagerSemaphore = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


#if defined (_IX_RM_BUFFER_DEBUG_)


/**
 * NAME: _ix_rm_buffer_debug_info_get
 *
 * DESCRIPTION: This function will retrieve the debug info coresponding to a buffer
 * 
 * @Param:  - IN arg_hBuffer - the handle to the buffer whose debug info we are trying to
 *  retrieve
 * @Param:  - OUT arg_pDebugInfo - an address where the pointer to the corresponding debug
 *  info will be stored on return
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_buffer_debug_info_get(
                                       ix_buffer_handle arg_hBuffer,
                                       ix_buffer_debug_info** arg_pDebugInfo 
                                     )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_buffer_debug_info_get";

#if defined( _IX_RM_DEBUG_ )

    ix_uint32 flId = 0;

#endif /* end defined( _IX_RM_DEBUG_ ) */

    register _ix_buffer_free_list* pFreeList = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(IX_RM_BUFFER_GET_TYPE(arg_hBuffer) == IX_BUFFER_TYPE_HARDWARE)
    {
        if(arg_hBuffer == IX_NULL_BUFFER_HANDLE)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_HANDLE_INVALID, s_aFunctionIdentifier, 
                        "the the passed buffer handle is invalid!", 0);

#if defined( _IX_RM_DEBUG_ )

        /* do extra checking in debug mode */
        {
            pFreeList = g_FLManager.m_aFreeListAddress[IX_RM_BUF_HW_GET_FL_ID(arg_hBuffer)];

            if(pFreeList == 0)
            {
                return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_INVALID, s_aFunctionIdentifier, 
                            "the associated free list for this buffer is invalid!", 0);
            }
        }

#endif /* end defined( _IX_RM_DEBUG_ ) */

        /* hardware buffer */
        pFreeList = g_FLManager.m_aFreeListAddress[IX_RM_BUF_HW_GET_FL_ID(arg_hBuffer)];

#if defined(_IX_RM_SPLIT_META_DATA_)

        *arg_pDebugInfo = pFreeList->m_pDebugInfo + 
            + ((((ix_uint8*)IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(arg_hBuffer) - 
            (ix_uint8*)pFreeList->m_FLData.m_HwData.m_pInternalMetaBaseAddress ) >> 
            pFreeList->m_FLData.m_HwData.m_InternalMetaElementShift));


#else

        *arg_pDebugInfo = pFreeList->m_pDebugInfo + 
            + ((((ix_uint8*)IX_RM_BUF_HW_GET_META_ADDRESS(arg_hBuffer) - 
            (ix_uint8*)pFreeList->m_pMetaBaseAddress) >> pFreeList->m_MetaElementShift));

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

    }
    else
    {

#if defined( _IX_RM_DEBUG_ )

        /* software buffer */
        flId = IX_RM_BUF_SW_GET_FL_ID(arg_hBuffer);

        /* check if the ID coresponds to a software free list */
        if(flId < IX_BUF_FIRST_SW_FL_ID)
        {
            return IX_RM_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_ID_INVALID, s_aFunctionIdentifier, 
                        "the free list ID coresponds to a hardware free list instead of a software one!", 0);

        }

#else

#endif /* end defined( _IX_RM_DEBUG_ ) */

        pFreeList = g_FLManager.m_aFreeListAddress[IX_RM_BUF_SW_GET_FL_ID(arg_hBuffer)];

        *arg_pDebugInfo = pFreeList->m_pDebugInfo + (IX_RM_BUF_SW_GET_INDEX(arg_hBuffer));

    }

    return IX_SUCCESS;
}


/**
 * NAME: _ix_rm_free_list_get_available_buffers
 *
 * DESCRIPTION: This function will 
 * 
 * @Param:  - IN arg_hFreeList - the free list whose number of available buffers
 *  we are interested in.
 * @Param:  - OUT arg_pNoAvailableBuffers - address where on return the number of 
 *  available buffers will be stored upon return
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_free_list_get_available_buffers(
                                                 ix_buffer_free_list_handle arg_hFreeList, 
                                                 ix_uint32* arg_pNoAvailableBuffers 
                                               )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_free_list_get_available_buffers";

    ix_error err = IX_SUCCESS;
    ix_error internalErr = IX_SUCCESS;

    ix_buffer_handle hBufferChain = IX_NULL_BUFFER_HANDLE;
    ix_buffer_handle hBuffer = IX_NULL_BUFFER_HANDLE;
    ix_uint32 availableBuffers = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    for(;;)
    {
        /* allocate a new buffer */
        internalErr = ix_rm_buffer_alloc(arg_hFreeList, &hBuffer);
        if(internalErr != IX_SUCCESS)
            break;

        /* and chain it */
        if(hBufferChain != IX_NULL_BUFFER_HANDLE)
        {
            internalErr = ix_rm_buffer_link(&hBuffer, &hBufferChain);
            if(internalErr != IX_SUCCESS)
            {
                /* free the newly allocated buffer */
                ix_rm_buffer_free_chain(hBuffer);
                break;
            }
        }
        /* make hBuffer the head of the chain */
        hBufferChain = hBuffer;
        availableBuffers++;
    }

    if(internalErr != IX_ERROR_NEW(IX_RM_ERROR_BUF_FREE_LIST_ALL_IN_USE, IX_ERROR_LEVEL_LOCAL))
        err = internalErr;

    /* Free the newly created buffer chain */
    if(hBufferChain != IX_NULL_BUFFER_HANDLE)
        ix_rm_buffer_free_chain(hBufferChain);

    *arg_pNoAvailableBuffers = availableBuffers;

    return err;
}


#endif /* defined (_IX_RM_BUFFER_DEBUG_) */


/* Just to have the Hw buffer working */
/* #undef _IX_RM_LIB_DEVELOPMENT_ */

