/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 4
 *
 * = LIBRARY
 *      Resource Manager Library
 *
 * = MODULE
 *      Communication API
 *
 * = FILENAME
 *      internal_communication.c
 *
 * = DESCRIPTION
 *      This file will define all the internal supporting functions for the 
 *      Communication API.
 *
 * = AUTHOR
 *       Adrian Georgescu
 *       adrian.georgescu@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      10/26/2001 11:34:33 AM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */

static const char s_aFileIdentifier[] = "$Id: internal_communication.c,v 1.92 2003/10/02 21:49:54 ageorges Exp $";

#include "ix_ossl.h"

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"

#include "rm/symbols.h"

#include <hal_scratch.h>



#if defined(_IX_RM_IMPL_HARDWARE_)

    /* #define _IX_RM_TEST_AI_INTERRUPTS_ */


    #if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

        #include <iv.h>
        #include <intLib.h>

    #elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        #include <asm/irq.h>
        #include <asm/arch/irqs.h>

    #endif /* end (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) */

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */


#include "rm/internal/internal_support.h"
#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"
#include "rm/system.h"
#include "rm/internal/internal_system.h"

#include "rm/memory.h"

#include "rm/buffer.h"
#include "rm/internal/internal_buffer.h"
#include "rm/hardware.h"

#include "rm/communication.h"
#include "rm/internal/internal_communication.h"

#include "rm/internal/internal_pci_communication.h"

#include "rm/remote_communication.h"
#include "rm/internal/internal_remote_communication.h"


#if defined(_IX_TEST_THREAD_SWITCHES_)

ix_uint32 g_PacketInterruptCount = 0;

#endif /* end defined (_IX_TEST_THREAD_SWITCHES_) */


/**
 * Test the me/core communication by allocating a hw free list the uCode 
 * can allocate buffers from. 
 */
/* #define _IX_RM_TEST_ME_CORE_COMM_   1 */
/* #define _IX_RM_TEST_SRAM_RINGS_ 1 */

/**
 * Default handlers declarations
 */

static ix_error _ix_rm_comm_default_ublock_packet_handler
                              (
                                ix_buffer_handle arg_hBuffer,
                                ix_uint32 arg_UserData,
                                void* arg_pContext
                              );

static ix_error _ix_rm_comm_default_ublock_message_handler
                              (
                                ix_buffer_handle arg_hBuffer,
                                ix_uint32 arg_UserData,
                                void* arg_pContext
                              );


static ix_error _ix_rm_end_point_manager_init
                                      (
                                        ix_end_point_manager* arg_pEPManager
                                      );
static ix_error _ix_rm_end_point_manager_fini
                                      (
                                        ix_end_point_manager* arg_pEPManager
                                      );

/**
 * Dispatch engines interface functions
 */
static ix_error _ix_rm_packet_dispatch_engine_init(void* pContext);
static ix_error _ix_rm_message_dispatch_engine_init(void* pContext);
static ix_error _ix_rm_dispatch_engine_fini(void* pContext);
static ix_error _ix_rm_packet_dispatch_engine_start(void* pContext);
static ix_error _ix_rm_message_dispatch_engine_start(void* pContext);
static ix_error _ix_rm_packet_dispatch_engine_stop(void* pContext);
static ix_error _ix_rm_message_dispatch_engine_stop(void* pContext);
static ix_error _ix_rm_packet_dispatch_engine_run(void* pContext);
static ix_error _ix_rm_message_dispatch_engine_run(void* pContext);
static ix_error _ix_rm_dispatch_engine_notify(void* pContext);


/**
 * This is the packet engine interface
 */
static _ix_dispatch_engine_interface s_PacketEngineInterface = 
{
    (ix_dispatch_engine_interface_func_ptr)_ix_rm_packet_dispatch_engine_init,
    (ix_dispatch_engine_interface_func_ptr)_ix_rm_dispatch_engine_fini,
    (ix_dispatch_engine_interface_func_ptr)_ix_rm_packet_dispatch_engine_start,
    (ix_dispatch_engine_interface_func_ptr)_ix_rm_packet_dispatch_engine_stop,
    (ix_dispatch_engine_interface_func_ptr)_ix_rm_packet_dispatch_engine_run,
    (ix_dispatch_engine_interface_func_ptr)_ix_rm_dispatch_engine_notify
};


/**
 * This is the message engine interface
 */
static _ix_dispatch_engine_interface s_MessageEngineInterface = 
{
    (ix_dispatch_engine_interface_func_ptr)_ix_rm_message_dispatch_engine_init,
    (ix_dispatch_engine_interface_func_ptr)_ix_rm_dispatch_engine_fini,
    (ix_dispatch_engine_interface_func_ptr)_ix_rm_message_dispatch_engine_start,
    (ix_dispatch_engine_interface_func_ptr)_ix_rm_message_dispatch_engine_stop,
    (ix_dispatch_engine_interface_func_ptr)_ix_rm_message_dispatch_engine_run,
    (ix_dispatch_engine_interface_func_ptr)_ix_rm_dispatch_engine_notify
};

/**
 * This is the ME to core packet dispatch engine context
 */
static _ix_default_dispatch_engine_context s_PacketEngineContext = 
{
    0,
    0,
    0,
    0,
    &s_PacketEngineInterface
};


/**
 * This is the ME to core message dispatch engine context
 */
static _ix_default_dispatch_engine_context s_MessageEngineContext = 
{
    0,
    0,
    0,
    0,
    &s_MessageEngineInterface
};



/**
 * This is the internal communication semaphore pool
 */
ix_sem_pool* g_pCommSemaphorePool = 0;



/**
 * Global end point manager
 */
ix_end_point_manager g_EndPointManager;


#if defined( _IX_RM_MULTI_THREADED_ )

/**
 * Synchronization semaphore for end point manager
 */
static ix_ossl_sem_t s_EPManagerSemaphore = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */


#if defined(_IX_RM_IMPL_HARDWARE_)

/* ISRs for uBlock/core packet/message communication */
#if defined (_IX_RM_TEST_AI_INTERRUPTS_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

static void _ix_ublock_to_core_isr(int arg_value);

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

static void _ix_ublock_to_core_isr(int irq, void* context, struct pt_regs* regs);

#endif

#else

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

static void _ix_ublock_to_core_packet_isr(int arg_value);
static void _ix_ublock_to_core_message_isr(int arg_value);

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

static void _ix_ublock_to_core_packet_isr(int irq, void* context, struct pt_regs* regs);
static void _ix_ublock_to_core_message_isr(int irq, void* context, struct pt_regs* regs);

#endif

#endif /* end  defined (_IX_RM_TEST_AI_INTERRUPTS_) */

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */




/**
 * This variable represents the default data handler for the end points.
 * For now we set the default behaviour to "drop data"
 */
ix_comm_data_handler g_DefaultDataHandler = _ix_rm_comm_data_drop;


#if defined (_IX_RM_TEST_AI_INTERRUPTS_)

void trigger_software_interrupt(void)
{
    *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x54))) |= 1;
}

ix_uint32 software_interrupt_test_thread_run_terminate = 0;
ix_uint32 software_interrupt_test_thread_run_done = 0;
ix_ossl_thread_t software_interrupt_test_thread_id = 0;

#define IX_SOFTWARE_INTERRUPT_TEST_THREAD_TIMEOUT   1000

ix_error _ix_rm_software_interrupt_test_thread_run(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_software_interrupt_test_thread_run";

    ix_error err = 0;

    ix_buffer_free_list_handle hHwFreeList;
    ix_buffer_handle hBuffer;
    ix_uint32 data = 100;

    err = ix_rm_hw_buffer_free_list_create(128, sizeof(ix_hw_buffer_meta), 64, 0, 0, &hHwFreeList);
    if(err != IX_SUCCESS)
        return err;

    while(software_interrupt_test_thread_run_terminate == 0)
    {
        err = ix_rm_buffer_alloc(hHwFreeList, &hBuffer);
        if(err != IX_SUCCESS)
            goto label_1;

#if defined(_IX_RM_IMPL_HARDWARE_)

        SCRATCH_RING_PUT(IX_UBLOCK_MESSAGE_RING_ID, hBuffer);
        SCRATCH_RING_PUT(IX_UBLOCK_MESSAGE_RING_ID, data);

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

        trigger_software_interrupt();
        ix_ossl_sleep(IX_SOFTWARE_INTERRUPT_TEST_THREAD_TIMEOUT);
    }

label_1:

    ix_rm_buffer_free_list_delete(hHwFreeList);

    software_interrupt_test_thread_run_terminate = 0;
    software_interrupt_test_thread_run_done = 1;

    return err;
}

#endif /* end  defined (_IX_RM_TEST_AI_INTERRUPTS_) */




#if defined (_IX_RM_TEST_ME_CORE_COMM_)

ix_uint32 g_TotalBuffersDropped = 0;

#endif /* end  defined (_IX_RM_TEST_ME_CORE_COMM_) */



/**
 * NAME: _ix_rm_comm_data_drop
 *
 * DESCRIPTION: This function will be the default processing function for each
 *          communication ID. This will return the buffers to the owning free list.
 * 
 * @Param:  - IN arg_hBuffer - the handle of the buffer to be processed
 * @Param:  - IN arg_UserData - the user data associated with the buffer to process
 * @Param:  - IN arg_pContext - context that has been passed to the comm ID at the 
 *          default handler registration time.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_comm_data_drop( 
                                ix_buffer_handle arg_hBuffer, 
                                ix_uint32 arg_UserData,
                                void* arg_pContext 
                              )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_comm_data_drop";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_hBuffer != IX_NULL_BUFFER_HANDLE)
    {

#if defined (_IX_RM_TEST_ME_CORE_COMM_)

        g_TotalBuffersDropped++;

#endif /* end  defined (_IX_RM_TEST_ME_CORE_COMM_) */

        return ix_rm_buffer_free_chain(arg_hBuffer);
    }
    else
        return IX_SUCCESS;
}


/**
 * NAME: _ix_rm_comm_default_ublock_packet_handler
 *
 * DESCRIPTION: This function will be the default processing function for the uBlock
 *  packet end point.
 * 
 * @Param:  - IN arg_hBuffer - the handle of the buffer to be processed
 * @Param:  - IN arg_UserData - the user data associated with the buffer to process
 * @Param:  - IN arg_pContext - context that has been passed to the comm ID at the 
 *          default handler registration time.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_comm_default_ublock_packet_handler(
                                ix_buffer_handle arg_hBuffer,
                                ix_uint32 arg_UserData,
                                void* arg_pContext
                              )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_comm_default_ublock_packet_handler";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_hBuffer != IX_NULL_BUFFER_HANDLE)
        return ix_rm_buffer_free_chain(arg_hBuffer);
    else
        return IX_SUCCESS;
}


/**
 * NAME: _ix_rm_comm_ublock_packet_handler
 *
 * DESCRIPTION: This function will be the the internal processing function for the uBlock
 *  packet end point. This function will store the buffer handle and user data onto the
 *  associated end point ring which will passed through arg_pContext argument.
 * 
 * @Param:  - IN arg_hBuffer - the handle of the buffer to be processed
 * @Param:  - IN arg_UserData - the user data associated with the buffer to process
 * @Param:  - IN arg_pContext - context that has been passed to the comm ID at the 
 *          default handler registration time.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_comm_ublock_packet_handler( 
                                ix_buffer_handle arg_hBuffer, 
                                ix_uint32 arg_UserData,
                                void* arg_pContext 
                              )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_comm_ublock_packet_handler";

    ix_hw_ring_handle hRing = (ix_hw_ring_handle)arg_pContext;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_hBuffer != IX_NULL_BUFFER_HANDLE)
    {
        if(IX_RM_BUFFER_GET_TYPE(arg_hBuffer) == IX_BUFFER_TYPE_HARDWARE)
        {
            /* hardware buffer */
            ix_uint32 size = 2;
            ix_uint32 data[2];

            data[0] = (ix_uint32)arg_hBuffer;
            data[1] = arg_UserData;

            return ix_rm_hw_ring_put(hRing, &size, data);
        }
        else
        {
            /* for software buffer we just free the buffer chain */
            return ix_rm_buffer_free_chain(arg_hBuffer);
        }
    }
    else
        return IX_SUCCESS;
}


/**
 * NAME: _ix_rm_comm_default_ublock_message_handler
 *
 * DESCRIPTION: This function will be the default processing function for the uBlock
 *  message end point.
 * 
 * @Param:  - IN arg_hBuffer - the handle of the buffer to be processed
 * @Param:  - IN arg_UserData - the user data associated with the buffer to process
 * @Param:  - IN arg_pContext - context that has been passed to the comm ID at the 
 *          default handler registration time.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_comm_default_ublock_message_handler( 
                                ix_buffer_handle arg_hBuffer, 
                                ix_uint32 arg_UserData,
                                void* arg_pContext 
                              )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_comm_default_ublock_message_handler";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_hBuffer != IX_NULL_BUFFER_HANDLE)
        return ix_rm_buffer_free_chain(arg_hBuffer);
    else
        return IX_SUCCESS;
}

/**
 * NAME: _ix_rm_comm_ublock_message_handler
 *
 * DESCRIPTION: This function will be the the internal processing function for the uBlock
 *  message end point. This function will store the buffer handle and user data onto the
 *  associated end point ring which will passed through arg_pContext argument.
 * 
 * @Param:  - IN arg_hBuffer - the handle of the buffer to be processed
 * @Param:  - IN arg_UserData - the user data associated with the buffer to process
 * @Param:  - IN arg_pContext - context that has been passed to the comm ID at the 
 *          default handler registration time.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_comm_ublock_message_handler( 
                                ix_buffer_handle arg_hBuffer, 
                                ix_uint32 arg_UserData,
                                void* arg_pContext 
                              )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_comm_ublock_message_handler";

    ix_hw_ring_handle hRing = (ix_hw_ring_handle)arg_pContext;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_hBuffer != IX_NULL_BUFFER_HANDLE)
    {
        if(IX_RM_BUFFER_GET_TYPE(arg_hBuffer) == IX_BUFFER_TYPE_HARDWARE)
        {
            /* hardware buffer */
            ix_uint32 size = 2;
            ix_uint32 data[2];

            data[0] = (ix_uint32)arg_hBuffer;
            data[1] = arg_UserData;

            return ix_rm_hw_ring_put(hRing, &size, data);
        }
        else
        {
            /* for software buffer we just free the buffer chain */
            return ix_rm_buffer_free_chain(arg_hBuffer);
        }
    }
    else
        return IX_SUCCESS;
}

/**
 * NAME: _ix_rm_comm_data_store_mp
 *
 * DESCRIPTION: This function will be the default get/select processing function for each
 *  communication ID. The end point will store the data received and will notify
 *  all pending tasks that new data is available..
 * 
 * @Param:  - IN arg_hBuffer - the handle of the buffer to be processed
 * @Param:  - IN arg_UserData - the user data associated with the buffer to process
 * @Param:  - IN arg_pContext - context that has been passed to the comm ID at the 
 *  default handler registration time.
 * @Param:  - IN arg_SendCallType - specifies if the send call was blocking or unblocking.
 * @Param:  - IN arg_Timeout - specifies the timeout for the operation for the blocking
 *  call.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_comm_data_store_mp( 
                                    ix_buffer_handle arg_hBuffer, 
                                    ix_uint32 arg_UserData,
                                    void* arg_pContext,
                                    ix_send_call_type arg_SendCallType,
                                    ix_uint32 arg_Timeout
                                  )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_comm_data_store_mp";

    ix_end_point_cc_data* pCCData = (ix_end_point_cc_data*)arg_pContext;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_SendCallType == IX_SEND_CALL_TYPE_NONBLOCK)
    {
        return _ix_rm_cc_queue_element_nonblock_mp(pCCData, arg_hBuffer, arg_UserData);
    }
    else
    {
        return _ix_rm_cc_queue_element_block_mp(pCCData, arg_hBuffer, arg_UserData, arg_Timeout);
    }
}


/**
 * NAME: _ix_rm_comm_data_store_sp
 *
 * DESCRIPTION: This function will be the default get/select processing function for each
 *  communication ID in single producer mode. The end point will store the data received and will notify
 *  all pending tasks that new data is available.
 * 
 * @Param:  - IN arg_hBuffer - the handle of the buffer to be processed
 * @Param:  - IN arg_UserData - the user data associated with the buffer to process
 * @Param:  - IN arg_pContext - context that has been passed to the comm ID at the 
 *  default handler registration time.
 * @Param:  - IN arg_SendCallType - specifies if the send call was blocking or unblocking.
 * @Param:  - IN arg_Timeout - specifies the timeout for the operation for the blocking
 *  call.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_comm_data_store_sp( 
                                    ix_buffer_handle arg_hBuffer, 
                                    ix_uint32 arg_UserData,
                                    void* arg_pContext,
                                    ix_send_call_type arg_SendCallType,
                                    ix_uint32 arg_Timeout
                                  )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_comm_data_store_sp";

    ix_end_point_cc_data* pCCData = (ix_end_point_cc_data*)arg_pContext;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_SendCallType == IX_SEND_CALL_TYPE_NONBLOCK)
    {
        return _ix_rm_cc_queue_element_nonblock_sp(pCCData, arg_hBuffer, arg_UserData);
    }
    else
    {
        return _ix_rm_cc_queue_element_block_sp(pCCData, arg_hBuffer, arg_UserData, arg_Timeout);
    }
}


/**
 * NAME: _ix_rm_comm_pci_data_store_mp
 *
 * DESCRIPTION: This function will be the default processing function for each PCI
 *  end point. The end point will store the data received and will notify
 *  all pending tasks that new data is available..
 * 
 * @Param:  - IN arg_hBuffer - the handle of the buffer to be processed
 * @Param:  - IN arg_UserData - the user data associated with the buffer to process
 * @Param:  - IN arg_pContext - context that has been passed to the comm ID at the 
 *  default handler registration time.
 * @Param:  - IN arg_Destination - destination comm Id to be stored onto the end point data queue
 * @Param:  - IN arg_SendCallType - specifies if the send call was blocking or unblocking.
 * @Param:  - IN arg_Timeout - specifies the timeout for the operation for the blocking
 *  call.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_comm_pci_data_store_mp( 
                                    ix_buffer_handle arg_hBuffer, 
                                    ix_uint32 arg_UserData,
                                    void* arg_pContext,
                                    ix_communication_id arg_Destination,
                                    ix_send_call_type arg_SendCallType,
                                    ix_uint32 arg_Timeout
                                  )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_comm_pci_data_store_mp";

    ix_end_point_cc_data* pCCData = (ix_end_point_cc_data*)arg_pContext;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_SendCallType == IX_SEND_CALL_TYPE_NONBLOCK)
    {
        return _ix_rm_cc_pci_queue_element_nonblock_mp(pCCData, arg_hBuffer, arg_UserData, arg_Destination);
    }
    else
    {
        return _ix_rm_cc_pci_queue_element_block_mp(pCCData, arg_hBuffer, arg_UserData, arg_Destination, arg_Timeout);
    }
}



/**
 * NAME: _ix_rm_handler_register
 *
 * DESCRIPTION: This function will register a data handler for the 
 *	specified communication ID. If the 
 *  _IX_COMM_ID_NOTIFICATION_DISCONNECT_ENABLE_ was defined at the 
 *  time of RM compilation, if another data handler is in place, that 
 *	data handler will be invoked with a IX_NULL_BUFFER_HANDLE as the 
 *	first parameter, and IX_COMM_ID_NOTIFICATION_DISCONNECT value as the 
 *	user data parameter. From this point on, the new registered handler 
 *	will be used for processing data sent to this destination. If this 
 *	communication ID is in get/select mode, the handler will be invoked with 
 *	all unprocessed data from the internal queue, and all waiting tasks 
 *	will be awaken and signaled that they have been disconnected from the 
 *	communication Id. This function can be used for both packet and message
 *  handlers through arg_HandlerType parameter 
 * 
 * @Param:  - IN arg_CommunicationId - the communication ID for which we are 
 *	registering the data handler.
 * @Param:  - IN arg_HandlerType - the type of data handler we want to register.
 *  The posibilities are packet and message.
 * @Param:  - IN arg_DataHandler - the data handler that we want to 
 *	register.
 * @Param:  - IN arg_pContext - context data that will be passed to the 
 *	data handler on each invocation. The programmer has to make sure that 
 *	the context data is valid as long the packet handler is registered.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error _ix_rm_handler_register( 
                                  ix_communication_id arg_CommunicationId, 
                                  _ix_comm_handler_type arg_HandlerType,
                                  _ix_comm_ep_behaviour_handler arg_DataHandler, 
                                  void* arg_pContext 
                                )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_handler_register";

    ix_error err = IX_SUCCESS;

    ix_uint32 localId  = IX_RM_COMM_ID_GET_LOCAL_ID(arg_CommunicationId);
    ix_end_point* pEndPoint = 0;
    ix_end_point_behaviour* pBehaviour = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((arg_CommunicationId & IX_COMM_REMOTE_SUBSYSTEM_MASK) == 0)
    {
        switch(arg_HandlerType)
        {
            case IX_COMM_HANDLER_TYPE_PACKET:
                pEndPoint = &g_EndPointManager.m_aPacketEndPoint[localId];
                break;
            case IX_COMM_HANDLER_TYPE_MESSAGE:
                pEndPoint = &g_EndPointManager.m_aMessageEndPoint[localId];
                break;
            default:
                IX_RM_ASSERT(arg_HandlerType < IX_COMM_HANDLER_TYPE_LAST , s_aFunctionIdentifier, "bad handler type!");
        }
        pBehaviour = &pEndPoint->m_EPBehaviour;

        if(localId > IX_COMM_LAST_UBLOCK_ID)
        {
            ix_end_point_cc_data* pCCData = &pEndPoint->m_EPData.m_CoreCompData;

            {

#if defined( _IX_RM_MULTI_THREADED_ )

                ix_synchronization_adapter* pSyncAdapter = &g_EndPointManager.m_SyncAdapter;

                /* lock the end point data */
                pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

                if((pCCData->m_Flags & IX_COMM_ID_RECEIVE_MODE_MASK) == IX_COMM_ID_MODE_CALLBACK)
                {
                    /* CB mode always */
                    _ix_comm_ep_behaviour_handler dataHandler = pBehaviour->m_BehaviourHandler;

                    if((arg_DataHandler != dataHandler) || (arg_pContext != pBehaviour->m_pExternalBehaviourContext))
                    {

                        pBehaviour->m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)g_DefaultDataHandler;
                        pBehaviour->m_pExternalBehaviourContext = arg_pContext;
                        pBehaviour->m_pInternalBehaviourContext = 0;
                        pBehaviour->m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)arg_DataHandler;

                        IX_SET_BIT_FIELD32(pCCData->m_Flags, IX_COMM_ID_MODE_CALLBACK, 0, 0);     

#if defined(_IX_COMM_ID_NOTIFICATION_DISCONNECT_ENABLE_)
                        
                        if(dataHandler != (_ix_comm_ep_behaviour_handler)g_DefaultDataHandler)
                        {
                            err = dataHandler(IX_NULL_BUFFER_HANDLE, IX_COMM_ID_NOTIFICATION_DISCONNECT, pBehaviour->m_pExternalBehaviourContext);
                        }

#endif /* end defined(_IX_COMM_ID_NOTIFICATION_DISCONNECT_ENABLE_) */

                    }  
                }
                else
                {
                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_RECV_GET_SELECT, s_aFunctionIdentifier, 
                                "the comm Id we are trying to change coresponds is in get/select mode!", err);
                }

#if defined( _IX_RM_MULTI_THREADED_ )

                /* unlock the end point data */
                pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

            }
        }
        else
        {
            {

#if defined( _IX_RM_MULTI_THREADED_ )

                ix_synchronization_adapter* pSyncAdapter = &g_EndPointManager.m_SyncAdapter;

                /* lock the end point data */
                pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

                {
                    /* CB mode always */
                    _ix_comm_ep_behaviour_handler dataHandler = pBehaviour->m_BehaviourHandler;

                    if(arg_DataHandler != dataHandler)
                    {

                        pBehaviour->m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)g_DefaultDataHandler;
                        pBehaviour->m_pExternalBehaviourContext = arg_pContext;
                        pBehaviour->m_pInternalBehaviourContext = 0;
                        pBehaviour->m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)arg_DataHandler;

                    }

                    if(((arg_HandlerType == IX_COMM_HANDLER_TYPE_PACKET) && (arg_DataHandler != 
                        (_ix_comm_ep_behaviour_handler)_ix_rm_comm_ublock_packet_handler)) ||
                        ((arg_HandlerType == IX_COMM_HANDLER_TYPE_MESSAGE) && (arg_DataHandler != 
                        (_ix_comm_ep_behaviour_handler)_ix_rm_comm_ublock_message_handler)))
                    {
                        pEndPoint->m_EPData.m_UBlockData.m_hHwRing = 0;
                        pEndPoint->m_EPData.m_UBlockData.m_Flags = 0;
                    }
                   
                }

#if defined( _IX_RM_MULTI_THREADED_ )

                /* unlock the end point data */
                pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

            }
        }
    }
    else
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_REMOTE, s_aFunctionIdentifier, 
                            "the comm Id whose behaviour we are trying to change is remote!", err);
    }

    return err;
}



/**
 * NAME: _ix_rm_end_point_manager_init
 *
 * DESCRIPTION: This function will initialize the passed end point manager.
 * 
 * @Param:  - INOUT arg_pEPManager - end point manager to initialize 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_end_point_manager_init(
                                        ix_end_point_manager* arg_pEPManager
                                      )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_end_point_manager_init";

    ix_error err = 0;
    ix_uint32 i;
    ix_end_point* pEndPoint;
    ix_end_point_cc_data* pCCEPData;
    ix_end_point_data_queue* pCCEPDataQueue;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* initialize the internal communication semaphore pool */
    err = _ix_rm_sem_pool_init(&arg_pEPManager->m_CommSemaphorePool, IX_COMM_SEM_POOL_SIZE);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_SEM_POOL_INIT_FAILED, s_aFunctionIdentifier, 
                        "the initialization internal communication semaphore pool!", err);
    }

    g_pCommSemaphorePool = &arg_pEPManager->m_CommSemaphorePool;

    /* initialize the uBlock end points */
    for(i = 0; i < IX_COMM_UBLOCK_ID_NUMBER; i++)
    {
        /* initialize the packet end point for the comm Id */
        pEndPoint = &arg_pEPManager->m_aPacketEndPoint[i];

        pEndPoint->m_EPBehaviour.m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)_ix_rm_comm_default_ublock_packet_handler;
        pEndPoint->m_EPBehaviour.m_pInternalBehaviourContext = pEndPoint;
        pEndPoint->m_EPBehaviour.m_pExternalBehaviourContext = 0;
        pEndPoint->m_EPData.m_UBlockData.m_hHwRing = 0;
        pEndPoint->m_EPData.m_UBlockData.m_Flags = 0;

        /* initialize the message end point for the comm Id */
        pEndPoint = &arg_pEPManager->m_aMessageEndPoint[i];

        pEndPoint->m_EPBehaviour.m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)_ix_rm_comm_default_ublock_message_handler;
        pEndPoint->m_EPBehaviour.m_pInternalBehaviourContext = pEndPoint;
        pEndPoint->m_EPBehaviour.m_pExternalBehaviourContext = 0;
        pEndPoint->m_EPData.m_UBlockData.m_hHwRing = 0;
        pEndPoint->m_EPData.m_UBlockData.m_Flags = 0;
    }

    /* initialize the end points coresponding to the core components */


    for(i = IX_COMM_UBLOCK_ID_NUMBER; i < IX_COMM_LOCAL_ID_NUMBER; i++)
    {
        /* reset the internal data queue for the packet end point for the comm Id */
        pEndPoint = &arg_pEPManager->m_aPacketEndPoint[i];
        pCCEPData = &pEndPoint->m_EPData.m_CoreCompData;
        pCCEPDataQueue = &pCCEPData->m_EPDataQueue;

        pCCEPDataQueue->m_pStartAddress = 0;
        pCCEPDataQueue->m_Size = 0;
        pCCEPDataQueue->m_Capacity = 0;
        pCCEPDataQueue->m_ElementSize = 0;
        pCCEPDataQueue->m_Put = pCCEPDataQueue->m_Get = 0;

        /* reset the internal data queue for the message end point for the comm Id */
        pEndPoint = &arg_pEPManager->m_aMessageEndPoint[i];
        pCCEPData = &pEndPoint->m_EPData.m_CoreCompData;
        pCCEPDataQueue = &pCCEPData->m_EPDataQueue;

        pCCEPDataQueue->m_pStartAddress = 0;
        pCCEPDataQueue->m_Size = 0;
        pCCEPDataQueue->m_Capacity = 0;
        pCCEPDataQueue->m_ElementSize = 0;
        pCCEPDataQueue->m_Put = pCCEPDataQueue->m_Get = 0;
    }

    for(i = IX_COMM_UBLOCK_ID_NUMBER; i < IX_COMM_LOCAL_ID_NUMBER; i++)
    {
        /* initialize the packet end point for the comm Id */
        pEndPoint = &arg_pEPManager->m_aPacketEndPoint[i];

        pEndPoint->m_EPBehaviour.m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)g_DefaultDataHandler;
        pEndPoint->m_EPBehaviour.m_pInternalBehaviourContext = pEndPoint;
        pEndPoint->m_EPBehaviour.m_pExternalBehaviourContext = 0;

        pCCEPData = &pEndPoint->m_EPData.m_CoreCompData;

        /* initialize the data queue */
        pCCEPDataQueue = &pCCEPData->m_EPDataQueue;

#if defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_)

        pCCEPDataQueue->m_pStartAddress = (ix_uint32*)ix_ossl_malloc(IX_CORE_PACKET_DATA_RING_SIZE * IX_CORE_PACKET_DATA_SIZE * sizeof(ix_uint32));
        if(pCCEPDataQueue->m_pStartAddress == 0)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_DATA_ALLOCATION_FAILED, s_aFunctionIdentifier, 
                            "the allocation of end point packet queues data area failed!", err);
            goto label_1;
        }

#else

        err = ix_rm_mem_alloc(IX_MEMORY_TYPE_DRAM, 
                          0, 
                          IX_CORE_PACKET_DATA_RING_SIZE * IX_CORE_PACKET_DATA_SIZE * sizeof(ix_uint32) , 
                          (void**)&pCCEPDataQueue->m_pStartAddress
                        );
        if(err != IX_SUCCESS)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_DATA_ALLOCATION_FAILED, s_aFunctionIdentifier, 
                            "the allocation of end point packet queues data area failed!", err);
            goto label_1;
        }

#endif /* defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_) */

        pCCEPDataQueue->m_Size = IX_CORE_PACKET_DATA_RING_SIZE;
        pCCEPDataQueue->m_Capacity = pCCEPDataQueue->m_Size - 1;
        pCCEPDataQueue->m_ElementSize = IX_CORE_PACKET_DATA_SIZE * sizeof(ix_uint32);
        pCCEPDataQueue->m_Put = pCCEPDataQueue->m_Get = 0;

        pCCEPData->m_Flags = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

        pCCEPData->m_pSendNotificationQueue = 0; 
        pCCEPData->m_pReceiveNotificationQueue = 0; 

        err = ix_ossl_sem_init(IX_OSSL_SEM_AVAILABLE, &pCCEPData->m_SyncSemaphore);
        if(err != IX_SUCCESS)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_SEMA_CREATION_FAILED, s_aFunctionIdentifier, 
                            "the creation of the EP sync semaphore failed!", err);
            goto label_3;
        }

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

        /* initialize the message end point for the comm Id */
        pEndPoint = &arg_pEPManager->m_aMessageEndPoint[i];

        pEndPoint->m_EPBehaviour.m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)g_DefaultDataHandler;
        pEndPoint->m_EPBehaviour.m_pInternalBehaviourContext = pEndPoint;
        pEndPoint->m_EPBehaviour.m_pExternalBehaviourContext = 0;

        pCCEPData = &pEndPoint->m_EPData.m_CoreCompData;

        /* initialize the data queue */
        pCCEPDataQueue = &pCCEPData->m_EPDataQueue;

#if defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_)

        pCCEPDataQueue->m_pStartAddress = (ix_uint32*)ix_ossl_malloc(IX_CORE_MESSAGE_DATA_RING_SIZE * IX_CORE_MESSAGE_DATA_SIZE * sizeof(ix_uint32));
        if(pCCEPDataQueue->m_pStartAddress == 0)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_DATA_ALLOCATION_FAILED, s_aFunctionIdentifier, 
                            "the allocation of end point message queues data area failed!", err);
            goto label_2;
        }

#else

        err = ix_rm_mem_alloc(IX_MEMORY_TYPE_DRAM, 
                          0, 
                          IX_CORE_MESSAGE_DATA_RING_SIZE * IX_CORE_MESSAGE_DATA_SIZE * sizeof(ix_uint32) , 
                          (void**)&pCCEPDataQueue->m_pStartAddress
                        );
        if(err != IX_SUCCESS)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_DATA_ALLOCATION_FAILED, s_aFunctionIdentifier, 
                            "the allocation of end point message queues data area failed!", err);
            goto label_2;
        }

#endif /* defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_) */

        pCCEPDataQueue->m_Size = IX_CORE_MESSAGE_DATA_RING_SIZE;
        pCCEPDataQueue->m_Capacity = pCCEPDataQueue->m_Size - 1;
        pCCEPDataQueue->m_ElementSize = IX_CORE_MESSAGE_DATA_SIZE * sizeof(ix_uint32);
        pCCEPDataQueue->m_Put = pCCEPDataQueue->m_Get = 0;

        pCCEPData->m_Flags = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

        pCCEPData->m_pSendNotificationQueue = 0; 
        pCCEPData->m_pReceiveNotificationQueue = 0; 

        err = ix_ossl_sem_init(IX_OSSL_SEM_AVAILABLE, &pCCEPData->m_SyncSemaphore);
        if(err != IX_SUCCESS)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_SEMA_CREATION_FAILED, s_aFunctionIdentifier, 
                            "the creation of the EP sync semaphore failed!", err);
            goto label_3;
        }

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    }

    err = _ix_rm_remote_comm_init();
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_REMOTE_COMM_API_INIT_FAILED, s_aFunctionIdentifier, 
                        "the initialization of the remote communication failed!", err);
        goto label_3;
    }

    return err;

label_3:

    for(i = IX_COMM_UBLOCK_ID_NUMBER; i < IX_COMM_LOCAL_ID_NUMBER; i++)
    {
        pEndPoint = &arg_pEPManager->m_aPacketEndPoint[i];
        pCCEPData = &pEndPoint->m_EPData.m_CoreCompData;

        if(pCCEPData->m_SyncSemaphore != 0)
        {
            ix_ossl_sem_fini(pCCEPData->m_SyncSemaphore);
            pCCEPData->m_SyncSemaphore = 0;
        }

        pEndPoint = &arg_pEPManager->m_aMessageEndPoint[i];
        pCCEPData = &pEndPoint->m_EPData.m_CoreCompData;

        if(pCCEPData->m_SyncSemaphore != 0)
        {
            ix_ossl_sem_fini(pCCEPData->m_SyncSemaphore);
            pCCEPData->m_SyncSemaphore = 0;
        }
    }


label_2:


    for(i = IX_COMM_UBLOCK_ID_NUMBER; i < IX_COMM_LOCAL_ID_NUMBER; i++)
    {
        /* cleanup the internal data queue for the message end point for the comm Id */
        pEndPoint = &arg_pEPManager->m_aMessageEndPoint[i];
        pCCEPDataQueue = &pCCEPData->m_EPDataQueue;

        if(pCCEPDataQueue->m_pStartAddress != 0)
        {

#if defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_)

            ix_ossl_free(pCCEPDataQueue->m_pStartAddress);

#else

            ix_rm_mem_free(pCCEPDataQueue->m_pStartAddress);

#endif /* defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_) */

            pCCEPDataQueue->m_pStartAddress = 0;

        }

        pCCEPDataQueue->m_Size = 0;
        pCCEPDataQueue->m_Capacity = 0;
        pCCEPDataQueue->m_ElementSize = 0;
        pCCEPDataQueue->m_Put = pCCEPDataQueue->m_Get = 0;

    }


label_1:

    for(i = IX_COMM_UBLOCK_ID_NUMBER; i < IX_COMM_LOCAL_ID_NUMBER; i++)
    {
        /* cleanup the internal data queue for the packet end point for the comm Id */
        pEndPoint = &arg_pEPManager->m_aPacketEndPoint[i];
        pCCEPDataQueue = &pCCEPData->m_EPDataQueue;

        if(pCCEPDataQueue->m_pStartAddress != 0)
        {

#if defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_)

            ix_ossl_free(pCCEPDataQueue->m_pStartAddress);

#else

            ix_rm_mem_free(pCCEPDataQueue->m_pStartAddress);

#endif /* defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_) */

            pCCEPDataQueue->m_pStartAddress = 0;

        }

        pCCEPDataQueue->m_Size = 0;
        pCCEPDataQueue->m_Capacity = 0;
        pCCEPDataQueue->m_ElementSize = 0;
        pCCEPDataQueue->m_Put = pCCEPDataQueue->m_Get = 0;
    }

    return err;
}

/* Queue type parameter should be 0 for packets and 1 for messages */
void _ix_rm_comm_display_internal_queue(ix_uint32 arg_CommIdIndex, ix_uint32 arg_QueueType)
{
    ix_end_point* pEndPoint = 0;
    ix_end_point_data_queue* pEPDataQueue = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_EndPointManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    if(arg_QueueType == 0)
        pEndPoint = &g_EndPointManager.m_aPacketEndPoint[arg_CommIdIndex];
    else
        pEndPoint = &g_EndPointManager.m_aMessageEndPoint[arg_CommIdIndex];

    pEPDataQueue = &pEndPoint->m_EPData.m_CoreCompData.m_EPDataQueue;

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the end point data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    {
        ix_uint32 size = pEPDataQueue->m_Size * pEPDataQueue->m_ElementSize / sizeof(ix_uint32);
        ix_uint32 i = 0;

        ix_ossl_message_log("Information for the internal %s queue for comm ID %lu\n", 
            (arg_QueueType == 0) ? "packet" : "message", arg_CommIdIndex);

        ix_ossl_message_log("Queue size in elements: %lu\n", pEPDataQueue->m_Size);
        ix_ossl_message_log("Queue capacity in elements: %lu\n", pEPDataQueue->m_Capacity);
        ix_ossl_message_log("Element size: %lu\n", pEPDataQueue->m_ElementSize);
        ix_ossl_message_log("Put index: 0x%08lx\n", pEPDataQueue->m_Put);
        ix_ossl_message_log("Get index: 0x%08lx\n", pEPDataQueue->m_Get);
        ix_ossl_message_log("Queue data start address: 0x%p\n", pEPDataQueue->m_pStartAddress);

        for(i = 0; i < size; i+=4)
        {
            ix_ossl_message_log("0x%08lx 0x%08lx 0x%08lx 0x%08lx \n",
                pEPDataQueue->m_pStartAddress[i],
                pEPDataQueue->m_pStartAddress[i+1],
                pEPDataQueue->m_pStartAddress[i+2],
                pEPDataQueue->m_pStartAddress[i+3]);
        }
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the end point data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

}


/**
 * NAME: _ix_rm_end_point_manager_fini
 *
 * DESCRIPTION: This function will finalize the passed end point manager.
 * 
 * @Param:  - INOUT arg_pEPManager - the end point manager to claenup 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_end_point_manager_fini( 
                                        ix_end_point_manager* arg_pEPManager
                                      )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_end_point_manager_fini";

    ix_error err = 0;
    ix_uint32 i;
    ix_end_point* pEndPoint;
    ix_end_point_cc_data* pCCEPData;
    ix_end_point_data_queue* pCCEPDataQueue;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    _ix_rm_remote_comm_fini();

    /* finalize the uBlock end points */
    for(i = 0; i < IX_COMM_UBLOCK_ID_NUMBER; i++)
    {
        /* finalize the packet end point for the comm Id */
        pEndPoint = &arg_pEPManager->m_aPacketEndPoint[i];

        pEndPoint->m_EPBehaviour.m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)g_DefaultDataHandler;
        pEndPoint->m_EPBehaviour.m_pInternalBehaviourContext = pEndPoint;
        pEndPoint->m_EPBehaviour.m_pExternalBehaviourContext = 0;
        pEndPoint->m_EPData.m_UBlockData.m_hHwRing = 0;
        pEndPoint->m_EPData.m_UBlockData.m_Flags = 0;

        /* finalize the message end point for the comm Id */
        pEndPoint = &arg_pEPManager->m_aMessageEndPoint[i];

        pEndPoint->m_EPBehaviour.m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)g_DefaultDataHandler;
        pEndPoint->m_EPBehaviour.m_pInternalBehaviourContext = pEndPoint;
        pEndPoint->m_EPBehaviour.m_pExternalBehaviourContext = 0;
        pEndPoint->m_EPData.m_UBlockData.m_hHwRing = 0;
        pEndPoint->m_EPData.m_UBlockData.m_Flags = 0;
    }

    for(i = IX_COMM_UBLOCK_ID_NUMBER; i < IX_COMM_LOCAL_ID_NUMBER; i++)
    {
        /* finalize the packet end point for the comm Id */
        pEndPoint = &arg_pEPManager->m_aPacketEndPoint[i];

        pEndPoint->m_EPBehaviour.m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)g_DefaultDataHandler;
        pEndPoint->m_EPBehaviour.m_pInternalBehaviourContext = pEndPoint;
        pEndPoint->m_EPBehaviour.m_pExternalBehaviourContext = 0;

        pCCEPData = &pEndPoint->m_EPData.m_CoreCompData;

        pCCEPDataQueue = &pCCEPData->m_EPDataQueue;

        /* finalize the data queue */
        pCCEPDataQueue = &pCCEPData->m_EPDataQueue;
        pCCEPDataQueue->m_Put = pCCEPDataQueue->m_Get = 0; 
        
        pCCEPDataQueue->m_Capacity = 0;
        pCCEPDataQueue->m_Size = 0;
        pCCEPDataQueue->m_ElementSize = 0;

        if(pCCEPDataQueue->m_pStartAddress != 0)
        {

#if defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_)

            ix_ossl_free(pCCEPDataQueue->m_pStartAddress);

#else

            ix_rm_mem_free(pCCEPDataQueue->m_pStartAddress);

#endif /* defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_) */

            pCCEPDataQueue->m_pStartAddress = 0;

        }
        
        pCCEPData->m_Flags = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

        if(pCCEPData->m_pSendNotificationQueue != 0)
        {
            _IX_RM_EP_DATA_ACTION_LOCK(pCCEPData);
            _IX_RM_EP_DATA_WAIT_QUEUE_NOTIFY(pCCEPData->m_pSendNotificationQueue);
            _IX_RM_EP_DATA_ACTION_UNLOCK(pCCEPData);
        }
        if(pCCEPData->m_pReceiveNotificationQueue != 0)
        {
            _IX_RM_EP_DATA_ACTION_LOCK(pCCEPData);
            _IX_RM_EP_DATA_WAIT_QUEUE_NOTIFY(pCCEPData->m_pReceiveNotificationQueue);
            _IX_RM_EP_DATA_ACTION_UNLOCK(pCCEPData);
        }

        /* delete the EP sync semaphore */
        if(pCCEPData->m_SyncSemaphore != 0)
        {
            ix_ossl_sem_fini(pCCEPData->m_SyncSemaphore);
            pCCEPData->m_SyncSemaphore = 0;
        }

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

        /* finalize the message end point for the comm Id */
        pEndPoint = &arg_pEPManager->m_aMessageEndPoint[i];

        pEndPoint->m_EPBehaviour.m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)g_DefaultDataHandler;
        pEndPoint->m_EPBehaviour.m_pInternalBehaviourContext = pEndPoint;
        pEndPoint->m_EPBehaviour.m_pExternalBehaviourContext = 0;

        pCCEPData = &pEndPoint->m_EPData.m_CoreCompData;

        /* finalize the data queue */
        pCCEPDataQueue = &pCCEPData->m_EPDataQueue;
        pCCEPDataQueue->m_Put = pCCEPDataQueue->m_Get; 

        pCCEPDataQueue->m_Capacity = 0;
        pCCEPDataQueue->m_Size = 0;
        pCCEPDataQueue->m_ElementSize = 0;

        if(pCCEPDataQueue->m_pStartAddress != 0)
        {

#if defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_)

            ix_ossl_free(pCCEPDataQueue->m_pStartAddress);

#else

            ix_rm_mem_free(pCCEPDataQueue->m_pStartAddress);

#endif /* defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_) */

            pCCEPDataQueue->m_pStartAddress = 0;

        }
        
        pCCEPData->m_Flags = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

        if(pCCEPData->m_pSendNotificationQueue != 0)
        {
            _IX_RM_EP_DATA_ACTION_LOCK(pCCEPData);
            _IX_RM_EP_DATA_WAIT_QUEUE_NOTIFY(pCCEPData->m_pSendNotificationQueue);
            _IX_RM_EP_DATA_ACTION_UNLOCK(pCCEPData);
        }
        if(pCCEPData->m_pReceiveNotificationQueue != 0)
        {
            _IX_RM_EP_DATA_ACTION_LOCK(pCCEPData);
            _IX_RM_EP_DATA_WAIT_QUEUE_NOTIFY(pCCEPData->m_pReceiveNotificationQueue);
            _IX_RM_EP_DATA_ACTION_UNLOCK(pCCEPData);
        }

        /* delete the EP sync semaphore */
        if(pCCEPData->m_SyncSemaphore != 0)
        {
            ix_ossl_sem_fini(pCCEPData->m_SyncSemaphore);
            pCCEPData->m_SyncSemaphore = 0;
        }

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    }


    if(g_pCommSemaphorePool)
    {
        g_pCommSemaphorePool = 0;
        _ix_rm_sem_pool_fini(&arg_pEPManager->m_CommSemaphorePool);
    };

    return err;
}

#if defined (_IX_RM_TEST_ME_CORE_COMM_)

ix_buffer_free_list_handle s_hTestHwFreeList = 0;

#endif /* end  defined (_IX_RM_TEST_ME_CORE_COMM_) */


#if defined (_IX_RM_TEST_SRAM_RINGS_)

ix_hw_ring_handle s_hTestSRAMRing = 0;

#endif /* end  defined (_IX_RM_TEST_SRAM_RINGS_) */


/**
 * NAME: _ix_rm_communication_init
 *
 * DESCRIPTION: This function should be called at the time the resource manager initializer 
 *          function is called. The function will initialize all the Communication API.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_communication_init(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_communication_init";

    ix_error err = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* reset the default data handler */
    g_DefaultDataHandler = _ix_rm_comm_data_drop;

#if defined( _IX_RM_MULTI_THREADED_ )

    err = ix_ossl_sem_init(IX_OSSL_SEM_AVAILABLE, &s_EPManagerSemaphore);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_API_INIT_FAILED, s_aFunctionIdentifier, 
                        "initialization of end point manager's semaphore failed!", err);
        goto label_1;
    }

    _ix_rm_sa_sync_adapter_init(&g_EndPointManager.m_SyncAdapter, 
        _ix_rm_sa_default_lock, _ix_rm_sa_default_unlock, (void*)&s_EPManagerSemaphore); 

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */


#if defined(_IX_RM_IMPL_HARDWARE_) 

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

    /* enable all interrupt levels */
    intLevelSet(0);

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

#endif /* end (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) */

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    /* initialize the end point manager */
    err = _ix_rm_end_point_manager_init(&g_EndPointManager);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_MGR_INIT_FAILED, s_aFunctionIdentifier, 
                        "the initialization of the end point manager failed!", err);
        goto label_2;
    }

    err = s_PacketEngineContext.m_pEngineInterface->m_InitEngineFuncPtr(&s_PacketEngineContext);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_PKT_DISP_ENG_INIT_FAILED, s_aFunctionIdentifier, 
                        "the initialization of the packet dispatch engine failed!", err);
        goto label_3;
    }

    err = s_PacketEngineContext.m_pEngineInterface->m_StartEngineFuncPtr(&s_PacketEngineContext);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_PKT_DISP_ENG_START_FAILED, s_aFunctionIdentifier, 
                        "startingthe packet dispatch engine failed!", err);
        goto label_4;
    }

    err = s_MessageEngineContext.m_pEngineInterface->m_InitEngineFuncPtr(&s_MessageEngineContext);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_MSG_DISP_ENG_INIT_FAILED, s_aFunctionIdentifier, 
                        "the initialization of the message dispatch engine failed!", err);
        goto label_5;
    }

    err = s_MessageEngineContext.m_pEngineInterface->m_StartEngineFuncPtr(&s_MessageEngineContext);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_MSG_DISP_ENG_START_FAILED, s_aFunctionIdentifier, 
                        "starting the message dispatch engine failed!", err);
        goto label_6;

    }

#if defined(_IX_COMPILE_PCI_COMM_IN_)

#if defined(_IX_RM_IMPL_HARDWARE_)


    err = ix_rm_init_pci_remote_communication();
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_INIT_FAILED, s_aFunctionIdentifier, 
                        "initialization of PCI communication failed!", err);
        goto label_7;
    }

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

#endif /* end defined(_IX_COMPILE_PCI_COMM_IN_) */


#if defined (_IX_RM_TEST_AI_INTERRUPTS_)

    /* create the AI software interrupt test thread */

    /* We trigger the software interrupt in the context of the message engine for now */
    ix_ossl_thread_create((ix_ossl_thread_entry_point_t)_ix_rm_software_interrupt_test_thread_run,
        &s_MessageEngineContext, "RmSwInt", &software_interrupt_test_thread_id);

    ix_ossl_thread_set_priority(software_interrupt_test_thread_id, IX_OSSL_THREAD_PRI_HIGH);

#endif /* end  defined (_IX_RM_TEST_AI_INTERRUPTS_) */

#if defined (_IX_RM_TEST_ME_CORE_COMM_)

    /* create a hardware free list in order for uCode to allocate from it */
    err = ix_rm_hw_buffer_free_list_create(1024, sizeof(ix_hw_buffer_meta), 32, 0, 0, 
        &s_hTestHwFreeList);
    if(err != IX_SUCCESS)
    {
        goto label_8;
    }

#endif /* end  defined (_IX_RM_TEST_ME_CORE_COMM_) */

#if defined (_IX_RM_TEST_SRAM_RINGS_)

    /* create a hardware free list in order for uCode to allocate from it */
    err = ix_rm_hw_sram_ring_create(0, IX_SRAM_RING_SIZE_1K, &s_hTestSRAMRing);
    if(err != IX_SUCCESS)
    {
        goto label_9;
    }

#endif /* end  defined (_IX_RM_TEST_SRAM_RINGS_) */

    return err;

#if defined (_IX_RM_TEST_ME_CORE_COMM_)

label_9:

    /* free the test hw free list */
    if(s_hTestHwFreeList != 0)
    {
        ix_rm_buffer_free_list_delete(s_hTestHwFreeList);
        s_hTestHwFreeList = 0;
    }

    ix_ossl_message_log("Total number of dropped buffers is: %u\n", g_TotalBuffersDropped);

    /* reset the dropped buffers count */
    g_TotalBuffersDropped = 0;

#endif /* end  defined (_IX_RM_TEST_ME_CORE_COMM_) */


#if defined (_IX_RM_TEST_AI_INTERRUPTS_)

    /* terminate the AI software interrupt test thread */

    if(software_interrupt_test_thread_run_done == 0)
    {
        software_interrupt_test_thread_run_terminate = 1;
    
        /* wait for thread to terminate */
        while(software_interrupt_test_thread_run_terminate != 0)
        {
            ix_ossl_sleep(1);
        };
    }
    software_interrupt_test_thread_id = 0;

#endif /* end  defined (_IX_RM_TEST_AI_INTERRUPTS_) */

#if defined (_IX_RM_TEST_ME_CORE_COMM_)

label_8:

#endif /* end  defined (_IX_RM_TEST_ME_CORE_COMM_) */


#if defined(_IX_COMPILE_PCI_COMM_IN_)

#if defined(_IX_RM_IMPL_HARDWARE_)

label_7:

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

#endif /* end defined(_IX_COMPILE_PCI_COMM_IN_) */

    s_MessageEngineContext.m_pEngineInterface->m_StopEngineFuncPtr(&s_MessageEngineContext);

label_6:

    s_MessageEngineContext.m_pEngineInterface->m_FiniEngineFuncPtr(&s_MessageEngineContext);

label_5:

    s_PacketEngineContext.m_pEngineInterface->m_StopEngineFuncPtr(&s_PacketEngineContext);

label_4:

    s_PacketEngineContext.m_pEngineInterface->m_FiniEngineFuncPtr(&s_PacketEngineContext);

label_3:

    /* finalize the end point manager */
    _ix_rm_end_point_manager_fini(&g_EndPointManager);

    /* finalize the synchronization mechanism */
#if defined( _IX_RM_MULTI_THREADED_ )

label_2:

    _ix_rm_sa_sync_adapter_init(&g_EndPointManager.m_SyncAdapter, 0, 0, (void*)0); 

    ix_ossl_sem_fini(s_EPManagerSemaphore);

    s_EPManagerSemaphore = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

label_1:

    return err;
}


/**
 * NAME: _ix_rm_communication_fini
 *
 * DESCRIPTION: This function should be called at the time the resource manager finalizer function
 *          is called. The function will provide cleanup for Communication API.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_communication_fini(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_communication_fini";

    ix_error err = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined (_IX_RM_TEST_ME_CORE_COMM_)

    /* free the test hw free list */
    if(s_hTestHwFreeList != 0)
    {
        ix_rm_buffer_free_list_delete(s_hTestHwFreeList);
        s_hTestHwFreeList = 0;
    }

    ix_ossl_message_log("Total number of dropped buffers is: %u\n", g_TotalBuffersDropped);

    /* reset the dropped buffers count */
    g_TotalBuffersDropped = 0;

#endif /* end  defined (_IX_RM_TEST_ME_CORE_COMM_) */

#if defined (_IX_RM_TEST_SRAM_RINGS_)

    /* free the test SRAM ring */
    if(s_hTestSRAMRing != 0)
    {
        ix_rm_hw_ring_delete(s_hTestSRAMRing);
        s_hTestSRAMRing = 0;
    }

#endif /* end  defined (_IX_RM_TEST_SRAM_RINGS_) */


#if defined (_IX_RM_TEST_AI_INTERRUPTS_)

    /* terminate the AI software interrupt test thread */

    if(software_interrupt_test_thread_run_done == 0)
    {
        software_interrupt_test_thread_run_terminate = 1;
    
        /* wait for thread to terminate */
        while(software_interrupt_test_thread_run_terminate != 0)
        {
            ix_ossl_sleep(1);
        };
    }
    software_interrupt_test_thread_id = 0;

#endif /* end  defined (_IX_RM_TEST_AI_INTERRUPTS_) */

    err = s_MessageEngineContext.m_pEngineInterface->m_StopEngineFuncPtr(&s_MessageEngineContext);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_MSG_DISP_ENG_STOP_FAILED, s_aFunctionIdentifier, 
                        "stoping the message dispatch engine failed!", err);
    }

    err = s_MessageEngineContext.m_pEngineInterface->m_FiniEngineFuncPtr(&s_MessageEngineContext);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_MSG_DISP_ENG_FINI_FAILED, s_aFunctionIdentifier, 
                        "the finalization of the message dispatch engine failed!", err);
    }

    err = s_PacketEngineContext.m_pEngineInterface->m_StopEngineFuncPtr(&s_PacketEngineContext);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_PKT_DISP_ENG_STOP_FAILED, s_aFunctionIdentifier, 
                        "stoping the packet dispatch engine failed!", err);
    }

    err = s_PacketEngineContext.m_pEngineInterface->m_FiniEngineFuncPtr(&s_PacketEngineContext);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_PKT_DISP_ENG_FINI_FAILED, s_aFunctionIdentifier, 
                        "the finalization of the packet dispatch engine failed!", err);
    }

    /* finalize the end point manager */
    err = _ix_rm_end_point_manager_fini(&g_EndPointManager);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_MGR_FINI_FAILED, s_aFunctionIdentifier, 
                        "the finalization of the end point manager failed!", err);
    }

    /* finalize the synchronization mechanism */
#if defined( _IX_RM_MULTI_THREADED_ )

    _ix_rm_sa_sync_adapter_init(&g_EndPointManager.m_SyncAdapter, 0, 0, (void*)0); 

    err = ix_ossl_sem_fini(s_EPManagerSemaphore);
    if(err != 0)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_API_FINI_FAILED, s_aFunctionIdentifier, 
                        "finalization of end point manager's semaphore failed!", err);
    }

    s_EPManagerSemaphore = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * The following section provides default implementation for the dispatch communication 
 * interface for message and packet passing.
 */


/**
 * NAME: _ix_rm_packet_dispatch_engine_init
 *
 * DESCRIPTION: This function is an implementation of the dispatch engine interface
 *          init function. It will provide all the initialization code for the packet
 *          dispatch engine.
 * 
 * @Param:  - INOUT pContext - dispatch engine context this function will work on. 
 *          In this implementation the context is a pointer to a _ix_default_dispatch_engine_context
 *          structure.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_packet_dispatch_engine_init(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_packet_dispatch_engine_init";

    ix_error err = 0;

    _ix_default_dispatch_engine_context* pDefaultDispatchEngineContext = 
        (_ix_default_dispatch_engine_context*)pContext;

    ix_scratch_ring_size aSize; 

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    pDefaultDispatchEngineContext->m_StopProcessing = 0;

    /* create a ME to core communication hardware ring */
    aSize = IX_UBLOCK_PACKET_DATA_RING_SIZE;
    err = ix_rm_hw_scratch_ring_create(0, aSize, IX_UBLOCK_PACKET_RING_ID, 
        &pDefaultDispatchEngineContext->m_hDataRing); 
    if(err !=IX_SUCCESS)
       return IX_RM_ERROR_NEW(IX_RM_ERROR_HW_RING_CREATE_FAILED, s_aFunctionIdentifier, 
            "failed to create the internal communication ring!", 0);

    /* create a synchronization object */
    err =  _ix_rm_sem_pool_get_sem(g_pCommSemaphorePool, &pDefaultDispatchEngineContext->m_SyncSemaphore);
    if(err !=IX_SUCCESS)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_SEM_GET_FAILED, s_aFunctionIdentifier, 
            "failed to get a synchronization semaphore!", 0);

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

    /* Initialize and enable the interrupt */
    {
        STATUS status;

#if defined (_IX_RM_TEST_AI_INTERRUPTS_)


        status = intConnect(INUM_TO_IVEC(IX_COMM_SW_INTERRUPT), _ix_ublock_to_core_isr, (int)pDefaultDispatchEngineContext);
        if(status != 0)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_ISR_SETUP_FAILED, s_aFunctionIdentifier, 
                "failed to setup communication ISR!", (ix_error)IX_COMM_SW_INTERRUPT);

#else

        /* clear the interrupt source */
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x60))) = 0xffffffff;
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x68))) = 0xffffffff;

        status = intConnect(INUM_TO_IVEC(IX_COMM_PACKET_INTERRUPT_0), _ix_ublock_to_core_packet_isr, (int)pDefaultDispatchEngineContext);
        if(status != 0)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_ISR_SETUP_FAILED, s_aFunctionIdentifier,
                "failed to setup communication ISR!", (ix_error)IX_COMM_PACKET_INTERRUPT_0);

        status = intConnect(INUM_TO_IVEC(IX_COMM_PACKET_INTERRUPT_2), _ix_ublock_to_core_packet_isr, (int)pDefaultDispatchEngineContext);
        if(status != 0)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_ISR_SETUP_FAILED, s_aFunctionIdentifier,
                "failed to setup communication ISR!", (ix_error)IX_COMM_PACKET_INTERRUPT_2);


#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)


#else

        /* clear the interrupt source */
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x64))) = 0xffffffff;
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x6C))) = 0xffffffff;

        status = intConnect(INUM_TO_IVEC(IX_COMM_PACKET_INTERRUPT_1), _ix_ublock_to_core_packet_isr, (int)pDefaultDispatchEngineContext);
        if(status != 0)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_ISR_SETUP_FAILED, s_aFunctionIdentifier,
                "failed to setup communication ISR!", (ix_error)IX_COMM_PACKET_INTERRUPT_1);

        status = intConnect(INUM_TO_IVEC(IX_COMM_PACKET_INTERRUPT_3), _ix_ublock_to_core_packet_isr, (int)pDefaultDispatchEngineContext);
        if(status != 0)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_ISR_SETUP_FAILED, s_aFunctionIdentifier,
                "failed to setup communication ISR!", (ix_error)IX_COMM_PACKET_INTERRUPT_3);

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

#endif /* end defined (_IX_RM_TEST_AI_INTERRUPTS_) */

    }

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    {
        int status;

#if defined (_IX_RM_TEST_AI_INTERRUPTS_)

        status = request_irq(IX_COMM_SW_INTERRUPT, _ix_ublock_to_core_isr, SA_SHIRQ, "Resource Manager", (void*)pDefaultDispatchEngineContext);
        if(status != 0)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_ISR_SETUP_FAILED, s_aFunctionIdentifier, 
                "failed to setup communication ISR!", (ix_error)(-status));

        enable_irq(IX_COMM_SW_INTERRUPT);

#else

        /* clear the interrupt source */

        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x60))) = 0xffffffff;
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x68))) = 0xffffffff;

        status = request_irq(IX_COMM_PACKET_INTERRUPT_0, _ix_ublock_to_core_packet_isr, SA_INTERRUPT, "Resource Manager", (void*)pDefaultDispatchEngineContext);
        if(status != 0)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_ISR_SETUP_FAILED, s_aFunctionIdentifier, 
                "failed to setup communication ISR!", (ix_error)IX_COMM_PACKET_INTERRUPT_0);

        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x160))) = 0xffffffff;
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x10))) = 1<<16;

        enable_irq(IX_COMM_PACKET_INTERRUPT_0);

        status = request_irq(IX_COMM_PACKET_INTERRUPT_2, _ix_ublock_to_core_packet_isr, SA_INTERRUPT, "Resource Manager", (void*)pDefaultDispatchEngineContext);
        if(status != 0)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_ISR_SETUP_FAILED, s_aFunctionIdentifier, 
                "failed to setup communication ISR!", (ix_error)IX_COMM_PACKET_INTERRUPT_2);

        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x168))) = 0xffffffff;
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x10))) = 1 << 18;
        enable_irq(IX_COMM_PACKET_INTERRUPT_2);

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)


#else

        /* clear the interrupt source */
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x64))) = 0xffffffff;
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x6C))) = 0xffffffff;

        status = request_irq(IX_COMM_PACKET_INTERRUPT_1, _ix_ublock_to_core_packet_isr, SA_INTERRUPT, "Resource Manager", (void*)pDefaultDispatchEngineContext);
        if(status != 0)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_ISR_SETUP_FAILED, s_aFunctionIdentifier, 
                "failed to setup communication ISR!", (ix_error)IX_COMM_PACKET_INTERRUPT_1);

        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x164))) = 0xffffffff;
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x10))) = 1<<17;
        enable_irq(IX_COMM_PACKET_INTERRUPT_1);

        status = request_irq(IX_COMM_PACKET_INTERRUPT_3, _ix_ublock_to_core_packet_isr, SA_INTERRUPT, "Resource Manager", (void*)pDefaultDispatchEngineContext);
        if(status != 0)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_ISR_SETUP_FAILED, s_aFunctionIdentifier, 
                "failed to setup communication ISR!", (ix_error)IX_COMM_PACKET_INTERRUPT_3);

        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x16c))) = 0xffffffff;
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x10))) = 1<<19;
        enable_irq(IX_COMM_PACKET_INTERRUPT_3);


#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

#endif /* end defined (_IX_RM_TEST_AI_INTERRUPTS_) */
    }

#endif /* end (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) */

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    /* enable the interrupt */
    IX_RM_COMM_PACKET_INTERRUPT_ENABLE(0);

    return err;
}


/**
 * NAME: _ix_rm_message_dispatch_engine_init
 *
 * DESCRIPTION: This function is an implementation of the dispatch engine interface
 *          init function. It will provide all the initialization code for the message
 *          dispatch engine.
 * 
 * @Param:  - INOUT pContext - dispatch engine context this function will work on. 
 *          In this implementation the context is a pointer to a _ix_default_dispatch_engine_context
 *          structure.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_message_dispatch_engine_init(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_message_dispatch_engine_init";

    ix_error err = 0;

    _ix_default_dispatch_engine_context* pDefaultDispatchEngineContext = 
        (_ix_default_dispatch_engine_context*)pContext;

    ix_scratch_ring_size aSize; 

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    pDefaultDispatchEngineContext->m_StopProcessing = 0;

    /* create a ME to core communication hardware ring */
    aSize = IX_UBLOCK_MESSAGE_DATA_RING_SIZE;
    err = ix_rm_hw_scratch_ring_create(0, aSize, IX_UBLOCK_MESSAGE_RING_ID,  
        &pDefaultDispatchEngineContext->m_hDataRing); 
    if(err !=IX_SUCCESS)
       return IX_RM_ERROR_NEW(IX_RM_ERROR_HW_RING_CREATE_FAILED, s_aFunctionIdentifier, 
            "failed to create the internal communication ring!", 0);

    /* create a synchronization object */
    err = _ix_rm_sem_pool_get_sem(g_pCommSemaphorePool, &pDefaultDispatchEngineContext->m_SyncSemaphore);
    if(err !=IX_SUCCESS)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_SEM_GET_FAILED, s_aFunctionIdentifier, 
            "failed to get a synchronization semaphore!", 0);

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

    /* Initialize and enable the interrupt */
    {
        STATUS status;

#if defined (_IX_RM_TEST_AI_INTERRUPTS_)

        status = intConnect(INUM_TO_IVEC(IX_COMM_SW_INTERRUPT), _ix_ublock_to_core_isr, (int)pDefaultDispatchEngineContext);
        if(status != 0)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_ISR_SETUP_FAILED, s_aFunctionIdentifier, 
                "failed to setup communication ISR!", (ix_error)IX_COMM_SW_INTERRUPT);


#else

        /* clear the interrupt source */
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x80))) = 0xffffffff;
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x88))) = 0xffffffff;

        status = intConnect(INUM_TO_IVEC(IX_COMM_MESSAGE_INTERRUPT_0), _ix_ublock_to_core_message_isr, (int)pDefaultDispatchEngineContext);
        if(status != 0)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_ISR_SETUP_FAILED, s_aFunctionIdentifier, 
                "failed to setup communication ISR!", (ix_error)IX_COMM_MESSAGE_INTERRUPT_0);

        status = intConnect(INUM_TO_IVEC(IX_COMM_MESSAGE_INTERRUPT_2), _ix_ublock_to_core_message_isr, (int)pDefaultDispatchEngineContext);
        if(status != 0)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_ISR_SETUP_FAILED, s_aFunctionIdentifier, 
                "failed to setup communication ISR!", (ix_error)IX_COMM_MESSAGE_INTERRUPT_2);


#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)


#else

        /* clear the interrupt source */
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x84))) = 0xffffffff;
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x8C))) = 0xffffffff;

        status = intConnect(INUM_TO_IVEC(IX_COMM_MESSAGE_INTERRUPT_1), _ix_ublock_to_core_message_isr, (int)pDefaultDispatchEngineContext);
        if(status != 0)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_ISR_SETUP_FAILED, s_aFunctionIdentifier, 
                "failed to setup communication ISR!", (ix_error)IX_COMM_MESSAGE_INTERRUPT_1);

        status = intConnect(INUM_TO_IVEC(IX_COMM_MESSAGE_INTERRUPT_3), _ix_ublock_to_core_message_isr, (int)pDefaultDispatchEngineContext);
        if(status != 0)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_ISR_SETUP_FAILED, s_aFunctionIdentifier, 
                "failed to setup communication ISR!", (ix_error)IX_COMM_MESSAGE_INTERRUPT_3);


#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

#endif /* end defined (_IX_RM_TEST_AI_INTERRUPTS_) */

    }

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    {
        int status;

#if defined (_IX_RM_TEST_AI_INTERRUPTS_)

        status = request_irq(IX_COMM_SW_INTERRUPT, _ix_ublock_to_core_isr, SA_SHIRQ, "Resource Manager", (void*)pDefaultDispatchEngineContext);
        if(status != 0)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_ISR_SETUP_FAILED, s_aFunctionIdentifier, 
                "failed to setup communication ISR!", (ix_error)IX_COMM_SW_INTERRUPT);

        enable_irq(IX_COMM_SW_INTERRUPT);
       
#else

        /* clear the interrupt source */
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x80))) = 0xffffffff;
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x88))) = 0xffffffff;

        status = request_irq(IX_COMM_MESSAGE_INTERRUPT_0, _ix_ublock_to_core_message_isr, SA_INTERRUPT, "Resource Manager", (void*)pDefaultDispatchEngineContext);
        if(status != 0)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_ISR_SETUP_FAILED, s_aFunctionIdentifier,
                "failed to setup communication ISR!", (ix_error)IX_COMM_MESSAGE_INTERRUPT_0);

        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x180))) = 0xffffffff;
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x10))) = 1<<24;
        enable_irq(IX_COMM_MESSAGE_INTERRUPT_0);

        status = request_irq(IX_COMM_MESSAGE_INTERRUPT_2, _ix_ublock_to_core_message_isr, SA_INTERRUPT, "Resource Manager", (void*)pDefaultDispatchEngineContext);
        if(status != 0)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_ISR_SETUP_FAILED, s_aFunctionIdentifier, 
                "failed to setup communication ISR!", (ix_error)IX_COMM_MESSAGE_INTERRUPT_2);

        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x188))) = 0xffffffff;
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x10))) = 1<<26;
        enable_irq(IX_COMM_MESSAGE_INTERRUPT_2);


#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)


#else

        /* clear the interrupt source */
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x84))) = 0xffffffff;
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x8C))) = 0xffffffff;

        status = request_irq(IX_COMM_MESSAGE_INTERRUPT_1, _ix_ublock_to_core_message_isr, SA_INTERRUPT, "Resource Manager", (void*)pDefaultDispatchEngineContext);
        if(status != 0)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_ISR_SETUP_FAILED, s_aFunctionIdentifier, 
                "failed to setup communication ISR!", (ix_error)IX_COMM_MESSAGE_INTERRUPT_1);

        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x184))) = 0xffffffff;
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x10))) = 1<<25;
        enable_irq(IX_COMM_MESSAGE_INTERRUPT_1);

        status = request_irq(IX_COMM_MESSAGE_INTERRUPT_3, _ix_ublock_to_core_message_isr, SA_INTERRUPT, "Resource Manager", (void*)pDefaultDispatchEngineContext);
        if(status != 0)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_ISR_SETUP_FAILED, s_aFunctionIdentifier, 
                "failed to setup communication ISR!", (ix_error)IX_COMM_MESSAGE_INTERRUPT_3);

        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x18c))) = 0xffffffff;
        *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x10))) = 1<<27;
        enable_irq(IX_COMM_MESSAGE_INTERRUPT_3);


#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

#endif /* end defined (_IX_RM_TEST_AI_INTERRUPTS_) */
    }

#endif /* end (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) */

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    /* enable the interrupt */
    IX_RM_COMM_MESSAGE_INTERRUPT_ENABLE(0);

    return err;
}


/**
 * NAME: _ix_rm_dispatch_engine_fini
 *
 * DESCRIPTION: This function is an implementation of the dispatch engine interface
 *          fini function. It will provide all the finalization code for the
 *          dispatch engine.
 * 
 * @Param:  - INOUT pContext - dispatch engine context this function will work on. 
 *          In this implementation the context is a pointer to a _ix_default_dispatch_engine_context
 *          structure.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_dispatch_engine_fini(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_dispatch_engine_fini";

    ix_error err = 0;

    _ix_default_dispatch_engine_context* pDefaultDispatchEngineContext = 
        (_ix_default_dispatch_engine_context*)pContext;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* terminate the dispatch thread */

    /* delete the synchronization object */
    err = _ix_rm_sem_pool_release_sem(g_pCommSemaphorePool, pDefaultDispatchEngineContext->m_SyncSemaphore);

    /* delete the ME to core communication hardware ring */
    err = ix_rm_hw_ring_delete(pDefaultDispatchEngineContext->m_hDataRing);

    return err;
}


/**
 * NAME: _ix_rm_packet_dispatch_engine_start
 *
 * DESCRIPTION: This function is an implementation of the packet dispatch engine interface
 *          start function. This function will be called in order to start the dispatch
 *          engine processing.
 * 
 * @Param:  - INOUT pContext - dispatch engine context this function will work on. 
 *          In this implementation the context is a pointer to a _ix_default_dispatch_engine_context
 *          structure.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_packet_dispatch_engine_start(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_packet_dispatch_engine_start";

    ix_error err = 0;
    ix_int32 threadPriority = IX_OSSL_THREAD_PRI_HIGH - 2;

    _ix_default_dispatch_engine_context* pDefaultDispatchEngineContext = 
        (_ix_default_dispatch_engine_context*)pContext;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* start the engine */

    /* create the dispatch thread */
    err = ix_ossl_thread_create((ix_ossl_thread_entry_point_t)pDefaultDispatchEngineContext->m_pEngineInterface->m_RunEngineFuncPtr,
        pContext, "RmPktEng", &pDefaultDispatchEngineContext->m_DispatchThreadId);
    if(err !=IX_SUCCESS)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_DISPATCH_TH_INIT_FAILED, s_aFunctionIdentifier, 
            "failed to create the dispatch thread!", 0);

    /* set the dispatch thread priority to according to the compile time constants */

#if defined (_IX_COMM_DISPATCH_THREADS_TUNABLE_)

    threadPriority = IX_RM_COMM_PACKET_DISPATCH_THREAD_PRIORITY;

#else

    threadPriority = IX_OSSL_THREAD_PRI_HIGH - 2 /* IX_OSSL_THREAD_PRI_MEDIUM */;

#endif /* end defined (_IX_COMM_DISPATCH_THREADS_TUNABLE_) */

    err = ix_ossl_thread_set_priority(pDefaultDispatchEngineContext->m_DispatchThreadId, threadPriority);
    if(err !=IX_SUCCESS)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_DISPATCH_TH_INIT_FAILED, s_aFunctionIdentifier,
            "failed to set the dispatch thread priority!", 0);

    return err;
}


/**
 * NAME: _ix_rm_message_dispatch_engine_start
 *
 * DESCRIPTION: This function is an implementation of the message dispatch engine interface
 *          start function. This function will be called in order to start the dispatch
 *          engine processing.
 * 
 * @Param:  - INOUT pContext - dispatch engine context this function will work on. 
 *          In this implementation the context is a pointer to a _ix_default_dispatch_engine_context
 *          structure.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_message_dispatch_engine_start(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_message_dispatch_engine_start";

    ix_error err = 0;
    ix_int32 threadPriority = IX_OSSL_THREAD_PRI_HIGH - 2;

    _ix_default_dispatch_engine_context* pDefaultDispatchEngineContext = 
        (_ix_default_dispatch_engine_context*)pContext;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* start the engine */

    /* create the dispatch thread */
    err = ix_ossl_thread_create((ix_ossl_thread_entry_point_t)pDefaultDispatchEngineContext->m_pEngineInterface->m_RunEngineFuncPtr,
        pContext, "RmMsgEng", &pDefaultDispatchEngineContext->m_DispatchThreadId);
    if(err !=IX_SUCCESS)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_DISPATCH_TH_INIT_FAILED, s_aFunctionIdentifier, 
            "failed to create the dispatch thread!", 0);

#if defined (_IX_COMM_DISPATCH_THREADS_TUNABLE_)

    threadPriority = IX_RM_COMM_MESSAGE_DISPATCH_THREAD_PRIORITY;

#else

    threadPriority = IX_OSSL_THREAD_PRI_HIGH - 1;

#endif /* end defined (_IX_COMM_DISPATCH_THREADS_TUNABLE_) */

    err = ix_ossl_thread_set_priority(pDefaultDispatchEngineContext->m_DispatchThreadId, threadPriority);
    if(err !=IX_SUCCESS)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_DISPATCH_TH_INIT_FAILED, s_aFunctionIdentifier, 
            "failed to set the dispatch thread priority!", 0);

    return err;
}



/**
 * NAME: _ix_rm_packet_dispatch_engine_stop
 *
 * DESCRIPTION: This function is an implementation of the packet dispatch engine interface
 *          stop function. This function will be called in order to stop the dispatch 
 *          engine processing.
 * 
 * @Param:  - INOUT pContext - dispatch engine context this function will work on. 
 *          In this implementation the context is a pointer to a _ix_default_dispatch_engine_context
 *          structure.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_packet_dispatch_engine_stop(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_packet_dispatch_engine_stop";

    ix_error err = 0;

    _ix_default_dispatch_engine_context* pDefaultDispatchEngineContext = 
        (_ix_default_dispatch_engine_context*)pContext;

    ix_uint32 ringId = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) && defined(_IX_RM_IMPL_HARDWARE_)

#if defined (_IX_RM_TEST_AI_INTERRUPTS_)

    /* Unregister the software interrupt handler */
    disable_irq(IX_COMM_SW_INTERRUPT);
    free_irq(IX_COMM_SW_INTERRUPT, pContext);

#else

    /* Unregister the interrupt handlers */

    disable_irq(IX_COMM_PACKET_INTERRUPT_0);
    free_irq(IX_COMM_PACKET_INTERRUPT_0, pContext);
    disable_irq(IX_COMM_PACKET_INTERRUPT_2);
    free_irq(IX_COMM_PACKET_INTERRUPT_2, pContext);

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)


#else

    disable_irq(IX_COMM_PACKET_INTERRUPT_1);
    free_irq(IX_COMM_PACKET_INTERRUPT_1, pContext);
    disable_irq(IX_COMM_PACKET_INTERRUPT_3);
    free_irq(IX_COMM_PACKET_INTERRUPT_3, pContext);

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

#endif /* end defined (_IX_RM_TEST_AI_INTERRUPTS_) */

#endif /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) && defined(_IX_RM_IMPL_HARDWARE_) */

    pDefaultDispatchEngineContext->m_StopProcessing = 1;

    /* put some null data to exit the loop */
    ringId = IX_RM_HW_RING_GET_INDEX(pDefaultDispatchEngineContext->m_hDataRing);

#if defined(_IX_RM_IMPL_HARDWARE_)

    SCRATCH_RING_PUT(ringId, 0);

#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

    /* wake up the thread if waiting for data */
    err = pDefaultDispatchEngineContext->m_pEngineInterface->m_NotifyEngineFuncPtr(pContext);

    /* wait for the thread to terminate */
    while(pDefaultDispatchEngineContext->m_StopProcessing != 0)
    {
        ix_ossl_sleep(1);
    };
    return err;
}

/**
 * NAME: _ix_rm_message_dispatch_engine_stop
 *
 * DESCRIPTION: This function is an implementation of the message dispatch engine interface
 *          stop function. This function will be called in order to stop the dispatch 
 *          engine processing.
 * 
 * @Param:  - INOUT pContext - dispatch engine context this function will work on. 
 *          In this implementation the context is a pointer to a _ix_default_dispatch_engine_context
 *          structure.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_message_dispatch_engine_stop(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_message_dispatch_engine_stop";

    ix_error err = 0;

    _ix_default_dispatch_engine_context* pDefaultDispatchEngineContext = 
        (_ix_default_dispatch_engine_context*)pContext;

    ix_uint32 ringId = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) && defined(_IX_RM_IMPL_HARDWARE_)

#if defined (_IX_RM_TEST_AI_INTERRUPTS_)

    /* Unregister the software interrupt handler */
    disable_irq(IX_COMM_SW_INTERRUPT);
    free_irq(IX_COMM_SW_INTERRUPT, pContext);

#else

    /* Unregister the interrupt handlers */
    disable_irq(IX_COMM_MESSAGE_INTERRUPT_0);
    free_irq(IX_COMM_MESSAGE_INTERRUPT_0, pContext);
    disable_irq(IX_COMM_MESSAGE_INTERRUPT_2);
    free_irq(IX_COMM_MESSAGE_INTERRUPT_2, pContext);

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)


#else

    disable_irq(IX_COMM_MESSAGE_INTERRUPT_1);
    free_irq(IX_COMM_MESSAGE_INTERRUPT_1, pContext);
    disable_irq(IX_COMM_MESSAGE_INTERRUPT_3);
    free_irq(IX_COMM_MESSAGE_INTERRUPT_3, pContext);

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

#endif /* end defined (_IX_RM_TEST_AI_INTERRUPTS_) */

#endif /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) && defined(_IX_RM_IMPL_HARDWARE_) */

    pDefaultDispatchEngineContext->m_StopProcessing = 1;

    /* put some null data to exit the loop */
    ringId = IX_RM_HW_RING_GET_INDEX(pDefaultDispatchEngineContext->m_hDataRing);

#if defined(_IX_RM_IMPL_HARDWARE_)

    SCRATCH_RING_PUT(ringId, 0);
    
#endif /* defined(_IX_RM_IMPL_HARDWARE_) */

    /* wake up the thread if waiting for data */
    err = pDefaultDispatchEngineContext->m_pEngineInterface->m_NotifyEngineFuncPtr(pContext);

    /* wait for the thread to terminate */
    while(pDefaultDispatchEngineContext->m_StopProcessing != 0)
    {
        ix_ossl_sleep(1);
    };
    return err;
}




#if defined (_IX_COMM_DISPATCH_THREADS_TUNABLE_)

/**
 * NAME: _ix_rm_packet_dispatch_engine_run
 *
 * DESCRIPTION: This function is an implementation of the dispatch engine interface
 *          run function. This is the processing function for the packet dispatch engine.
 * 
 * @Param:  - INOUT pContext - dispatch engine context this function will work on. 
 *          In this implementation the context is a pointer to a _ix_default_dispatch_engine_context
 *          structure.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_packet_dispatch_engine_run(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_packet_dispatch_engine_run";

    ix_error err = 0;

    _ix_default_dispatch_engine_context* pDefaultDispatchEngineContext = 
        (_ix_default_dispatch_engine_context*)pContext;

    register ix_buffer_handle hBuffer;
    register ix_uint32 userData;
    register ix_end_point_behaviour* pEndPointBehavior;
    register ix_end_point* pEndPointStart = g_EndPointManager.m_aPacketEndPoint;
    register ix_uint32 currentBufferCounter;

#if defined (_IX_RM_TEST_ME_CORE_COMM_)

    register packetCounter = 0;
    ix_ossl_time_t startTime, lastTime;

#define TEST_COMM_PACKET_COUNT  1000000

#endif /* end  defined (_IX_RM_TEST_ME_CORE_COMM_) */


    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    while(pDefaultDispatchEngineContext->m_StopProcessing == 0)
    {
        /* reset the counter */
        currentBufferCounter = 0;

        /* wait for data */
        ix_ossl_sem_take(pDefaultDispatchEngineContext->m_SyncSemaphore,IX_OSSL_WAIT_FOREVER);

        /* retrieve data until the ring is empty */
        while(1)
        {
            hBuffer = SCRATCH_RING_GET(IX_UBLOCK_PACKET_RING_ID);
            /* if we got a spurious wakeup then just exit the loop */
            if(hBuffer == 0)
                break;

label_1:

#if defined (_IX_RM_BUFFER_DEBUG_)

            {
                ix_buffer_debug_info* pDebugInfo;
                if(_ix_rm_buffer_debug_info_get(hBuffer, &pDebugInfo) == IX_SUCCESS)
                {
                    pDebugInfo->m_hBuffer = hBuffer;
                    pDebugInfo->m_LastCoreAllocationStatus = IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_RECEIVED_FROM_ME_AS_PACKET;
                    pDebugInfo->m_SendAsMessageCount = 0;
                    pDebugInfo->m_SendAsPacketCount = 0;
                }
            }

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

            /* fix the chain in the case we don't have a single buffer */
            if(IX_RM_BUF_GET_EOP(hBuffer) == 0)
            {
                ix_buffer_handle hTempBuffer = hBuffer;
                ix_buffer_handle hNextBuffer = IX_NULL_BUFFER_HANDLE;
                while(1)
                {

#if defined(_IX_RM_SPLIT_META_DATA_)

                    hNextBuffer = IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(hTempBuffer)->m_HwNext;

#else

                    hNextBuffer = IX_RM_BUF_HW_GET_META_ADDRESS(hTempBuffer)->m_HwNext;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

#if defined (_IX_RM_BUFFER_DEBUG_)

                    {
                        ix_buffer_debug_info* pDebugInfo;
                        if(_ix_rm_buffer_debug_info_get(hNextBuffer, &pDebugInfo) == IX_SUCCESS)
                        {
                            pDebugInfo->m_hBuffer = hNextBuffer;
                            pDebugInfo->m_LastCoreAllocationStatus = IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_RECEIVED_FROM_ME_AS_PACKET;
                            pDebugInfo->m_SendAsMessageCount = 0;
                            pDebugInfo->m_SendAsPacketCount = 0;
                        }
                    }

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */


                    if(IX_RM_BUF_GET_EOP(hNextBuffer) != 0)
                    {

#if defined(_IX_RM_SPLIT_META_DATA_)

                        IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(hNextBuffer)->m_HwNext = hTempBuffer;

#else

                        IX_RM_BUF_HW_GET_META_ADDRESS(hNextBuffer)->m_HwNext = hTempBuffer;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

                        break;
                    }
                    hTempBuffer = hNextBuffer;
                }
            }

            userData = SCRATCH_RING_GET(IX_UBLOCK_PACKET_RING_ID);
            /* check if were able to retrieve two long words, and if not wait for the second */
            while(userData == 0)
            {
                userData = SCRATCH_RING_GET(IX_UBLOCK_PACKET_RING_ID);
                ix_ossl_sleep(0);
            }

            /* Now send the packet to the coresponding local EP for processing */
            pEndPointBehavior = &(pEndPointStart + IX_RM_UBLOCK_TO_CORE_GET_DESTINATION(userData))->m_EPBehaviour;
            err = pEndPointBehavior->m_BehaviourHandler(hBuffer, 
                IX_RM_UBLOCK_TO_CORE_GET_USER_DATA(userData),
                pEndPointBehavior->m_pExternalBehaviourContext, IX_SEND_CALL_TYPE_NONBLOCK, 0);
            if(err != IX_SUCCESS)
            {
                /* What to do here? Drop the packet? Yes for now! */
                ix_rm_buffer_free_chain((ix_buffer_handle)hBuffer);
            }

            if(++currentBufferCounter >= IX_RM_COMM_PACKET_DISPATCH_THREAD_YIELD_COUNT)
            {
                currentBufferCounter = 0;
                ix_ossl_sleep(IX_RM_COMM_PACKET_DISPATCH_THREAD_YIELD_DELAY);
            }

#if defined (_IX_RM_TEST_ME_CORE_COMM_)

            if(packetCounter++ == 0)
            {
                ix_ossl_time_get(&startTime);
            }
            else
            {
                if(packetCounter == TEST_COMM_PACKET_COUNT)
                {
                    ix_ossl_time_get(&lastTime);

                    ix_ossl_message_log("Processed %u packets\n", packetCounter);

                    ix_ossl_message_log("Start time: sec: %u nsec: %u\n", startTime.sec, startTime.nsec);
                    ix_ossl_message_log("Stop time: sec: %u nsec: %u\n", lastTime.sec, lastTime.nsec);

                    /* calculate the time elapsed from start */
                    IX_OSSL_TIME_SUB(lastTime, startTime);
                    ix_ossl_message_log("Elapsed time: %u msec\n", (lastTime.sec * 1000 + lastTime.nsec / 1000000));
                    ix_ossl_message_log("Stop time: sec: %u nsec: %u\n", lastTime.sec, lastTime.nsec);
                    break;
                }
            }

#endif /* end  defined (_IX_RM_TEST_ME_CORE_COMM_) */


        }

#if defined (_IX_RM_TEST_ME_CORE_COMM_)

        if(packetCounter >= TEST_COMM_PACKET_COUNT)
        {
            while(pDefaultDispatchEngineContext->m_StopProcessing == 0)
            {
                ix_ossl_sleep(10);
            }
            break;
        }

#endif /* end  defined (_IX_RM_TEST_ME_CORE_COMM_) */

        /* enable the packet interrupt */
        IX_RM_COMM_PACKET_INTERRUPT_ENABLE(0);

        /**
         * Check to see if data arrived between the time the ring was last checked and the time the
         * packet interrupt has been enabled.
         */
        hBuffer = SCRATCH_RING_GET(IX_UBLOCK_PACKET_RING_ID);
        if(hBuffer != 0)
        {
            IX_RM_COMM_PACKET_INTERRUPT_DISABLE(0);
            goto label_1;
        }
    }

    pDefaultDispatchEngineContext->m_StopProcessing = 0;

    return err;
}

#else

/**
 * NAME: _ix_rm_packet_dispatch_engine_run
 *
 * DESCRIPTION: This function is an implementation of the dispatch engine interface
 *          run function. This is the processing function for the packet dispatch engine.
 * 
 * @Param:  - INOUT pContext - dispatch engine context this function will work on. 
 *          In this implementation the context is a pointer to a _ix_default_dispatch_engine_context
 *          structure.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_packet_dispatch_engine_run(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_packet_dispatch_engine_run";

    ix_error err = 0;

    _ix_default_dispatch_engine_context* pDefaultDispatchEngineContext = 
        (_ix_default_dispatch_engine_context*)pContext;

    register ix_buffer_handle hBuffer;
    register ix_uint32 userData;
    register ix_end_point_behaviour* pEndPointBehavior;
    register ix_end_point* pEndPointStart = g_EndPointManager.m_aPacketEndPoint;

#if defined (_IX_RM_TEST_ME_CORE_COMM_)

    register packetCounter = 0;
    ix_ossl_time_t startTime, lastTime;

#define TEST_COMM_PACKET_COUNT  10000000

#endif /* end  defined (_IX_RM_TEST_ME_CORE_COMM_) */


    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    while(pDefaultDispatchEngineContext->m_StopProcessing == 0)
    {
        /* wait for data */
        ix_ossl_sem_take(pDefaultDispatchEngineContext->m_SyncSemaphore,IX_OSSL_WAIT_FOREVER);

        /* retrieve data until the ring is empty */
        while(1)
        {
            hBuffer = SCRATCH_RING_GET(IX_UBLOCK_PACKET_RING_ID);
            /* if we got a spurious wakeup then just exit the loop */
            if(hBuffer == 0)
                break;

label_1:

#if defined (_IX_RM_BUFFER_DEBUG_)

            {
                ix_buffer_debug_info* pDebugInfo;
                if(_ix_rm_buffer_debug_info_get(hBuffer, &pDebugInfo) == IX_SUCCESS)
                {
                    pDebugInfo->m_hBuffer = hBuffer;
                    pDebugInfo->m_LastCoreAllocationStatus = IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_RECEIVED_FROM_ME_AS_PACKET;
                    pDebugInfo->m_SendAsMessageCount = 0;
                    pDebugInfo->m_SendAsPacketCount = 0;
                }
            }

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

            /* fix the chain in the case we don't have a single buffer */
            if(IX_RM_BUF_GET_EOP(hBuffer) == 0)
            {
                ix_buffer_handle hTempBuffer = hBuffer;
                ix_buffer_handle hNextBuffer = IX_NULL_BUFFER_HANDLE;
                while(1)
                {

#if defined(_IX_RM_SPLIT_META_DATA_)

                    hNextBuffer = IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(hTempBuffer)->m_HwNext;

#else

                    hNextBuffer = IX_RM_BUF_HW_GET_META_ADDRESS(hTempBuffer)->m_HwNext;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

#if defined (_IX_RM_BUFFER_DEBUG_)

                    {
                        ix_buffer_debug_info* pDebugInfo;
                        if(_ix_rm_buffer_debug_info_get(hNextBuffer, &pDebugInfo) == IX_SUCCESS)
                        {
                            pDebugInfo->m_hBuffer = hNextBuffer;
                            pDebugInfo->m_LastCoreAllocationStatus = IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_RECEIVED_FROM_ME_AS_PACKET;
                            pDebugInfo->m_SendAsMessageCount = 0;
                            pDebugInfo->m_SendAsPacketCount = 0;
                        }
                    }

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */


                    if(IX_RM_BUF_GET_EOP(hNextBuffer) != 0)
                    {

#if defined(_IX_RM_SPLIT_META_DATA_)

                        IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(hNextBuffer)->m_HwNext = hTempBuffer;

#else

                        IX_RM_BUF_HW_GET_META_ADDRESS(hNextBuffer)->m_HwNext = hTempBuffer;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

                        break;
                    }
                    hTempBuffer = hNextBuffer;
                }
            }

            userData = SCRATCH_RING_GET(IX_UBLOCK_PACKET_RING_ID);
            /* check if were able to retrieve two long words, and if not wait for the second */
            while(userData == 0)
            {
                userData = SCRATCH_RING_GET(IX_UBLOCK_PACKET_RING_ID);
                ix_ossl_sleep(0);
            }

            /* Now send the packet to the coresponding local EP for processing */
            pEndPointBehavior = &(pEndPointStart + IX_RM_UBLOCK_TO_CORE_GET_DESTINATION(userData))->m_EPBehaviour;
            err = pEndPointBehavior->m_BehaviourHandler(hBuffer, 
                IX_RM_UBLOCK_TO_CORE_GET_USER_DATA(userData),
                pEndPointBehavior->m_pExternalBehaviourContext, IX_SEND_CALL_TYPE_NONBLOCK, 0);
            if(err != IX_SUCCESS)
            {
                /* What to do here? Drop the packet? Yes for now! */
                ix_rm_buffer_free_chain((ix_buffer_handle)hBuffer);
            }

#if defined (_IX_RM_TEST_ME_CORE_COMM_)

            if(packetCounter++ == 0)
            {
                ix_ossl_time_get(&startTime);
            }
            else
            {
                if(packetCounter == TEST_COMM_PACKET_COUNT)
                {
                    ix_ossl_time_get(&lastTime);

                    ix_ossl_message_log("Processed %u packets\n", packetCounter);

                    ix_ossl_message_log("Start time: sec: %u nsec: %u\n", startTime.sec, startTime.nsec);
                    ix_ossl_message_log("Stop time: sec: %u nsec: %u\n", lastTime.sec, lastTime.nsec);

                    /* calculate the time elapsed from start */
                    IX_OSSL_TIME_SUB(lastTime, startTime);
                    ix_ossl_message_log("Elapsed time: %u msec\n", (lastTime.sec * 1000 + lastTime.nsec / 1000000));
                    ix_ossl_message_log("Stop time: sec: %u nsec: %u\n", lastTime.sec, lastTime.nsec);
                    break;
                }
            }

#endif /* end  defined (_IX_RM_TEST_ME_CORE_COMM_) */

        }

#if defined (_IX_RM_TEST_ME_CORE_COMM_)

        if(packetCounter >= TEST_COMM_PACKET_COUNT)
        {
            while(pDefaultDispatchEngineContext->m_StopProcessing == 0)
            {
                ix_ossl_sleep(10);
            }
            break;
        }

#endif /* end  defined (_IX_RM_TEST_ME_CORE_COMM_) */

        /* enable the packet interrupt */
        IX_RM_COMM_PACKET_INTERRUPT_ENABLE(0);

        /**
         * Check to see if data arrived between the time the ring was last checked and the time the
         * packet interrupt has been enabled.
         */
        hBuffer = SCRATCH_RING_GET(IX_UBLOCK_PACKET_RING_ID);
        if(hBuffer != 0)
        {
            IX_RM_COMM_PACKET_INTERRUPT_DISABLE(0);
            goto label_1;
        }
    }

    pDefaultDispatchEngineContext->m_StopProcessing = 0;

    return err;
}

#endif /* end defined (_IX_COMM_DISPATCH_THREADS_TUNABLE_) */


#if defined (_IX_COMM_DISPATCH_THREADS_TUNABLE_)

/**
 * NAME: _ix_rm_message_dispatch_engine_run
 *
 * DESCRIPTION: This function is an implementation of the dispatch engine interface
 *          run function. This is the processing function for the message dispatch engine.
 * 
 * @Param:  - INOUT pContext - dispatch engine context this function will work on. 
 *          In this implementation the context is a pointer to a _ix_default_dispatch_engine_context
 *          structure.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_message_dispatch_engine_run(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_message_dispatch_engine_run";

    ix_error err = 0;

    _ix_default_dispatch_engine_context* pDefaultDispatchEngineContext = 
        (_ix_default_dispatch_engine_context*)pContext;

    register ix_buffer_handle hBuffer;
    register ix_uint32 userData;
    register ix_end_point_behaviour* pEndPointBehavior;
    register ix_end_point* pEndPointStart = g_EndPointManager.m_aMessageEndPoint;
    register ix_uint32 currentBufferCounter;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    while(pDefaultDispatchEngineContext->m_StopProcessing == 0)
    {
        /* reset the buffer current counter */
        currentBufferCounter = 0;

        /* wait for data */
        ix_ossl_sem_take(pDefaultDispatchEngineContext->m_SyncSemaphore,IX_OSSL_WAIT_FOREVER);

        /* retrieve data until the ring is empty */
        while(1)
        {
            hBuffer = SCRATCH_RING_GET(IX_UBLOCK_MESSAGE_RING_ID);
            /* if we got a spurious wakeup then just exit the loop */
            if(hBuffer == 0)
                break;

label_1:

#if defined (_IX_RM_BUFFER_DEBUG_)

            {
                ix_buffer_debug_info* pDebugInfo;
                if(_ix_rm_buffer_debug_info_get(hBuffer, &pDebugInfo) == IX_SUCCESS)
                {
                    pDebugInfo->m_hBuffer = hBuffer;
                    pDebugInfo->m_LastCoreAllocationStatus = IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_RECEIVED_FROM_ME_AS_MESSAGE;
                    pDebugInfo->m_SendAsMessageCount = 0;
                    pDebugInfo->m_SendAsPacketCount = 0;
                }
            }

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

            /* fix the chain in the case we don't have a single buffer */
            if(IX_RM_BUF_GET_EOP(hBuffer) == 0)
            {
                ix_buffer_handle hTempBuffer = hBuffer;
                ix_buffer_handle hNextBuffer = IX_NULL_BUFFER_HANDLE;
                while(1)
                {

#if defined(_IX_RM_SPLIT_META_DATA_)

                    hNextBuffer = IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(hTempBuffer)->m_HwNext;

#else

                    hNextBuffer = IX_RM_BUF_HW_GET_META_ADDRESS(hTempBuffer)->m_HwNext;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

#if defined (_IX_RM_BUFFER_DEBUG_)

                    {
                        ix_buffer_debug_info* pDebugInfo;
                        if(_ix_rm_buffer_debug_info_get(hNextBuffer, &pDebugInfo) == IX_SUCCESS)
                        {
                            pDebugInfo->m_hBuffer = hNextBuffer;
                            pDebugInfo->m_LastCoreAllocationStatus = IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_RECEIVED_FROM_ME_AS_MESSAGE;
                            pDebugInfo->m_SendAsMessageCount = 0;
                            pDebugInfo->m_SendAsPacketCount = 0;
                        }
                    }

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */


                    if(IX_RM_BUF_GET_EOP(hNextBuffer) != 0)
                    {

#if defined(_IX_RM_SPLIT_META_DATA_)

                        IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(hNextBuffer)->m_HwNext = hTempBuffer;

#else

                        IX_RM_BUF_HW_GET_META_ADDRESS(hNextBuffer)->m_HwNext = hTempBuffer;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

                        break;
                    }
                    hTempBuffer = hNextBuffer;
                }
            }

            userData = SCRATCH_RING_GET(IX_UBLOCK_MESSAGE_RING_ID);
            /* check if were able to retrieve two long words, and if not wait for the second */
            while(userData == 0)
            {
                userData = SCRATCH_RING_GET(IX_UBLOCK_MESSAGE_RING_ID);
                ix_ossl_sleep(0);
            }

            /* Now send the packet to the coresponding local EP for processing */
            pEndPointBehavior = &(pEndPointStart + IX_RM_UBLOCK_TO_CORE_GET_DESTINATION(userData))->m_EPBehaviour;
            err = pEndPointBehavior->m_BehaviourHandler(hBuffer, 
                IX_RM_UBLOCK_TO_CORE_GET_USER_DATA(userData),
                pEndPointBehavior->m_pExternalBehaviourContext, IX_SEND_CALL_TYPE_NONBLOCK, 0);
            if(err != IX_SUCCESS)
            {
                /* What to do here? Drop the packet? Yes for now! */
                ix_rm_buffer_free_chain((ix_buffer_handle)hBuffer);

            }

            if(++currentBufferCounter >= IX_RM_COMM_MESSAGE_DISPATCH_THREAD_YIELD_COUNT)
            {
                currentBufferCounter = 0;
                ix_ossl_sleep(IX_RM_COMM_MESSAGE_DISPATCH_THREAD_YIELD_DELAY);
            }

        }

        /* enable the message interrupt */
        IX_RM_COMM_MESSAGE_INTERRUPT_ENABLE(0);

        /**
         * Check to see if data arrived between the time the ring was last checked and the time the
         * packet interrupt has been enabled.
         */
        hBuffer = SCRATCH_RING_GET(IX_UBLOCK_MESSAGE_RING_ID);
        if(hBuffer != 0)
        {
            IX_RM_COMM_MESSAGE_INTERRUPT_DISABLE(0);
            goto label_1;
        }
    }

    pDefaultDispatchEngineContext->m_StopProcessing = 0;

    return err;
}

#else

/**
 * NAME: _ix_rm_message_dispatch_engine_run
 *
 * DESCRIPTION: This function is an implementation of the dispatch engine interface
 *          run function. This is the processing function for the message dispatch engine.
 * 
 * @Param:  - INOUT pContext - dispatch engine context this function will work on. 
 *          In this implementation the context is a pointer to a _ix_default_dispatch_engine_context
 *          structure.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_message_dispatch_engine_run(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_message_dispatch_engine_run";

    ix_error err = 0;

    _ix_default_dispatch_engine_context* pDefaultDispatchEngineContext = 
        (_ix_default_dispatch_engine_context*)pContext;

    register ix_buffer_handle hBuffer;
    register ix_uint32 userData;
    register ix_end_point_behaviour* pEndPointBehavior;
    register ix_end_point* pEndPointStart = g_EndPointManager.m_aMessageEndPoint;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    while(pDefaultDispatchEngineContext->m_StopProcessing == 0)
    {
        /* wait for data */
        ix_ossl_sem_take(pDefaultDispatchEngineContext->m_SyncSemaphore,IX_OSSL_WAIT_FOREVER);

        /* retrieve data until the ring is empty */
        while(1)
        {
            hBuffer = SCRATCH_RING_GET(IX_UBLOCK_MESSAGE_RING_ID);
            /* if we got a spurious wakeup then just exit the loop */
            if(hBuffer == 0)
                break;

label_1:

#if defined (_IX_RM_BUFFER_DEBUG_)

            {
                ix_buffer_debug_info* pDebugInfo;
                if(_ix_rm_buffer_debug_info_get(hBuffer, &pDebugInfo) == IX_SUCCESS)
                {
                    pDebugInfo->m_hBuffer = hBuffer;
                    pDebugInfo->m_LastCoreAllocationStatus = IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_RECEIVED_FROM_ME_AS_MESSAGE;
                    pDebugInfo->m_SendAsMessageCount = 0;
                    pDebugInfo->m_SendAsPacketCount = 0;
                }
            }

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */

            /* fix the chain in the case we don't have a single buffer */
            if(IX_RM_BUF_GET_EOP(hBuffer) == 0)
            {
                ix_buffer_handle hTempBuffer = hBuffer;
                ix_buffer_handle hNextBuffer = IX_NULL_BUFFER_HANDLE;
                while(1)
                {

#if defined(_IX_RM_SPLIT_META_DATA_)

                    hNextBuffer = IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(hTempBuffer)->m_HwNext;

#else

                    hNextBuffer = IX_RM_BUF_HW_GET_META_ADDRESS(hTempBuffer)->m_HwNext;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

#if defined (_IX_RM_BUFFER_DEBUG_)

                    {
                        ix_buffer_debug_info* pDebugInfo;
                        if(_ix_rm_buffer_debug_info_get(hNextBuffer, &pDebugInfo) == IX_SUCCESS)
                        {
                            pDebugInfo->m_hBuffer = hNextBuffer;
                            pDebugInfo->m_LastCoreAllocationStatus = IX_BUFFER_LAST_CORE_ALLOCATION_STATUS_RECEIVED_FROM_ME_AS_MESSAGE;
                            pDebugInfo->m_SendAsMessageCount = 0;
                            pDebugInfo->m_SendAsPacketCount = 0;
                        }
                    }

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */


                    if(IX_RM_BUF_GET_EOP(hNextBuffer) != 0)
                    {

#if defined(_IX_RM_SPLIT_META_DATA_)

                        IX_RM_BUF_HW_GET_INTERNAL_META_ADDRESS(hNextBuffer)->m_HwNext = hTempBuffer;

#else

                        IX_RM_BUF_HW_GET_META_ADDRESS(hNextBuffer)->m_HwNext = hTempBuffer;

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

                        break;
                    }
                    hTempBuffer = hNextBuffer;
                }
            }

            userData = SCRATCH_RING_GET(IX_UBLOCK_MESSAGE_RING_ID);
            /* check if were able to retrieve two long words, and if not wait for the second */
            while(userData == 0)
            {
                userData = SCRATCH_RING_GET(IX_UBLOCK_MESSAGE_RING_ID);
                ix_ossl_sleep(0);
            }

            /* Now send the packet to the coresponding local EP for processing */
            pEndPointBehavior = &(pEndPointStart + IX_RM_UBLOCK_TO_CORE_GET_DESTINATION(userData))->m_EPBehaviour;
            err = pEndPointBehavior->m_BehaviourHandler(hBuffer, 
                IX_RM_UBLOCK_TO_CORE_GET_USER_DATA(userData),
                pEndPointBehavior->m_pExternalBehaviourContext, IX_SEND_CALL_TYPE_NONBLOCK, 0);
            if(err != IX_SUCCESS)
            {
                /* What to do here? Drop the packet? Yes for now! */
                ix_rm_buffer_free_chain((ix_buffer_handle)hBuffer);

            }

        }

        /* enable the message interrupt */
        IX_RM_COMM_MESSAGE_INTERRUPT_ENABLE(0);
        
        /**
         * Check to see if data arrived between the time the ring was last checked and the time the
         * packet interrupt has been enabled.
         */

        hBuffer = SCRATCH_RING_GET(IX_UBLOCK_MESSAGE_RING_ID);

        if(hBuffer != 0)
        {
            IX_RM_COMM_MESSAGE_INTERRUPT_DISABLE(0);
            goto label_1;
        }

    }

    pDefaultDispatchEngineContext->m_StopProcessing = 0;

    return err;
}

#endif /* end defined (_IX_COMM_DISPATCH_THREADS_TUNABLE_) */


/**
 * NAME: _ix_rm_dispatch_engine_notify
 *
 * DESCRIPTION: This function is an implementation of the dispatch engine interface
 *          notify function. This function will be invoked by the ISR when a data token
 *          is sent from the ME to the core.
 * 
 * @Param:  - INOUT pContext - dispatch engine context this function will work on. 
 *          In this implementation the context is a pointer to a _ix_default_dispatch_engine_context
 *          structure.
 *
 * @Return: Returns 0 for success and non zero for failure.
 */
ix_error _ix_rm_dispatch_engine_notify(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_dispatch_engine_notify";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* signal the sync object contained in the context */
    return ix_ossl_sem_give(((_ix_default_dispatch_engine_context*)pContext)->m_SyncSemaphore);
}




/**
 * NAME: _ix_rm_sem_pool_init
 *
 * DESCRIPTION: This function will initialize a semaphore pool. The arg_count specifies
 *          how many semaphores we would like to have into the pool. The structure referred
 *          by arg_pSemPool will be initialized on successful return. The count should be 
 *          greater than 1 because the first semaphore in the created array will be used
 *          for operation synchronization.
 * 
 * @Param:  - INOUT arg_pSemPool - pointer to a semaphore pool that will be initialized.
 * @Param:  - IN arg_Count - the number of semaphores requested to be in the pool.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_sem_pool_init(
                               ix_sem_pool* arg_pSemPool,
                               ix_uint32 arg_Count
                             )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_sem_pool_init";

    ix_error err = 0;
    ix_uint32 i;
    ix_sem_free_node* pSemNode;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    arg_pSemPool->m_pSemaphores = (ix_ossl_sem_t*)ix_ossl_malloc(arg_Count * sizeof(ix_ossl_sem_t));
    if(arg_pSemPool->m_pSemaphores == 0)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_SEM_POOL_INIT_FAILED, s_aFunctionIdentifier, 
                "the semaphore pool allocation failed!", 0);
    }

    arg_pSemPool->m_Count = arg_Count;

    /* reset the pool data */
    for(i = 0; i < arg_Count; i++)
    {
        arg_pSemPool->m_pSemaphores[i] = 0;
    }
    arg_pSemPool->m_pSemFreeList = 0;

    /* initialize the pool */

    /* create the semaphores */
    for(i = 0; i < arg_Count; i++)
    {
        /* create a semaphore */
        err =  ix_ossl_sem_init(IX_OSSL_SEM_UNAVAILABLE,&arg_pSemPool->m_pSemaphores[i]);
        if(err != IX_SUCCESS)
            goto label_1;
    }

    /* initialize the semaphore controlling the pool */
    ix_ossl_sem_give(arg_pSemPool->m_pSemaphores[0]);

    /* add the semaphores to the free list except the first one the will be used for synchronization */
    for(i = 1; i < arg_pSemPool->m_Count; i++)
    {
        pSemNode = (ix_sem_free_node*)ix_ossl_malloc(sizeof(ix_sem_free_node));
        if(pSemNode == 0)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_SEM_NODE_ALLOCATION_FAILED, s_aFunctionIdentifier, 
                "the semaphore node allocation failed!", 0);
            goto label_1;
        }

        /* setup the newly allocated node and add it to the free list */
        pSemNode->m_Semaphore = arg_pSemPool->m_pSemaphores[i];
        pSemNode->m_pNext = arg_pSemPool->m_pSemFreeList;
        arg_pSemPool->m_pSemFreeList = pSemNode;
    }

    return err;

label_1:

    /* destroy the free list */
    {
        ix_sem_free_node* pSemNextNode;

        pSemNode = arg_pSemPool->m_pSemFreeList;
        while(pSemNode)
        {
            pSemNextNode = pSemNode->m_pNext;
            ix_ossl_free(pSemNode);
            pSemNode = pSemNextNode;
        }
        arg_pSemPool->m_pSemFreeList = 0;
    }

    /* destroy all created semaphores */
    for(i = 0; i < arg_pSemPool->m_Count; i++)
    {
        if(arg_pSemPool->m_pSemaphores[i] != 0)
        {
            ix_ossl_sem_fini(arg_pSemPool->m_pSemaphores[i]);
            arg_pSemPool->m_pSemaphores[i] = 0;
        }
    }

    /* free the semaphore array */
    ix_ossl_free(arg_pSemPool->m_pSemaphores);
    arg_pSemPool->m_pSemaphores = 0;

    return err;
}


/**
 * NAME: _ix_rm_sem_pool_fini
 *
 * DESCRIPTION: This function will destroy the passed semaphore pool releasing all 
 *          semaphores and associated data. Upon return the passed pool structure will
 *          be totally reset and can be reused.
 * 
 * @Param:  - INOUT arg_pSemPool - the semaphore pool that we would like to cleanup. 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_sem_pool_fini(
                               ix_sem_pool* arg_pSemPool
                             )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_sem_pool_fini";

    ix_error err = 0;
    ix_sem_free_node *pSemNode, *pSemNextNode;
    ix_uint32 i;
    ix_ossl_sem_t syncSem = arg_pSemPool->m_pSemaphores[0];

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* lock access to the pool */
    ix_ossl_sem_take(syncSem,IX_OSSL_WAIT_FOREVER);

    /* destroy the free list */
    pSemNode = arg_pSemPool->m_pSemFreeList;
    while(pSemNode)
    {
        pSemNextNode = pSemNode->m_pNext;
        ix_ossl_free(pSemNode);
        pSemNode = pSemNextNode;
    }
    arg_pSemPool->m_pSemFreeList = 0;

    /* destroy all created semaphores */
    for(i = 1; i < arg_pSemPool->m_Count; i++)
    {
        if(arg_pSemPool->m_pSemaphores[i] != 0)
        {
            ix_ossl_sem_fini(arg_pSemPool->m_pSemaphores[i]);
            arg_pSemPool->m_pSemaphores[i] = 0;
        }
    }

    /* free the semaphore array */
    ix_ossl_free(arg_pSemPool->m_pSemaphores);
    arg_pSemPool->m_pSemaphores = 0;
    arg_pSemPool->m_Count = 0;

    /* unlock the access to the pool */
    ix_ossl_sem_give(syncSem);

    /* destroy the synchronization semaphore for the pool */
    ix_ossl_sem_fini(syncSem);

    return err;
}



/**
 * NAME: _ix_rm_sem_pool_get_sem
 *
 * DESCRIPTION: This function will get a semaphore from the requested pool. This
 *          semaphore has to be returned to the pool in order to avoid resource 
 *          exhaustion.
 * 
 * @Param:  - IN arg_pSemPool - address of the pool of interest.
 * @Param:  - OUT arg_pSem - address where the available semaphore will be returned.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_sem_pool_get_sem( 
                                  ix_sem_pool* arg_pSemPool,
                                  ix_ossl_sem_t* arg_pSem 
                                )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_sem_pool_get_sem";

    ix_error err = 0;
    ix_sem_free_node *pSemNode;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* lock access to the pool */
    ix_ossl_sem_take(arg_pSemPool->m_pSemaphores[0],IX_OSSL_WAIT_FOREVER);

    if(arg_pSemPool->m_pSemFreeList == 0)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_SEM_GET_FAILED, s_aFunctionIdentifier,
                "no more semaphores available in the pool!", 0);
        goto label_1;
    }

    pSemNode = arg_pSemPool->m_pSemFreeList;
    arg_pSemPool->m_pSemFreeList = pSemNode->m_pNext;
    *arg_pSem = pSemNode->m_Semaphore;
    ix_ossl_free(pSemNode);

    /* unlock the access to the pool */
    ix_ossl_sem_give(arg_pSemPool->m_pSemaphores[0]);

label_1:
    return err;
}



/**
 * NAME: _ix_rm_sem_pool_release_sem
 *
 * DESCRIPTION: This function will return a semaphore to a given pool. No check is made
 *          to see if the semaphore really belongs to the pool. Also there will be no 
 *          check to see if the semaphore has been already released.
 * 
 * @Param:  - IN arg_pSemPool - semaphore pool of interest
 * @Param:  - IN arg_Sem - semaphore to be released
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_sem_pool_release_sem( 
                                      ix_sem_pool* arg_pSemPool,
                                      ix_ossl_sem_t arg_Sem 
                                    )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_sem_pool_release_sem";

    ix_error err = 0;
    ix_sem_free_node *pSemNode;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* lock access to the pool */
    ix_ossl_sem_take(arg_pSemPool->m_pSemaphores[0],IX_OSSL_WAIT_FOREVER);

    pSemNode = (ix_sem_free_node*)ix_ossl_malloc(sizeof(ix_sem_free_node));
    if(pSemNode == 0)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_SEM_NODE_ALLOCATION_FAILED, s_aFunctionIdentifier, 
            "the semaphore node allocation failed!", 0);
        goto label_1;
    }

    /* reset the semaphore if necessary */
    ix_ossl_sem_take(arg_Sem, IX_OSSL_WAIT_NONE);

    /* setup and add the node to the free list */
    pSemNode->m_pNext = arg_pSemPool->m_pSemFreeList;
    pSemNode->m_Semaphore = arg_Sem;
    arg_pSemPool->m_pSemFreeList = pSemNode;

    /* unlock the access to the pool */
    ix_ossl_sem_give(arg_pSemPool->m_pSemaphores[0]);

label_1:
    return err;
}

/**
 * Multi producer queueing functions
 */

/**
 * NAME: _ix_rm_cc_queue_element_nonblock_mp
 *
 * DESCRIPTION: This function will queue a data element into specified core component end point
 *          data queue. The function is non blocking. If the queue is full then an error will be
 *          returned right away. In case of error it is the duty of the caller to decide what to 
 *          do and eventually drop the buffer.
 * 
 * @Param:  - IN arg_pEPCCData -  core component end point data address this fuction will work on
 * @Param:  - IN arg_hBuffer - buffer handle to be stored onto the end point data queue
 * @Param:  - IN arg_UserData - user data to be stored onto the end point data queue
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_cc_queue_element_nonblock_mp( 
                                           ix_end_point_cc_data* arg_pEPCCData,
                                           ix_buffer_handle arg_hBuffer,
                                           ix_uint32 arg_UserData
                                         )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_cc_queue_element_nonblock_mp";

    ix_error err = 0;
    ix_end_point_data_queue *pDQueue = &arg_pEPCCData->m_EPDataQueue;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check if the queue is full */
    if(IX_EP_DATA_QUEUE_IS_FULL(pDQueue))
    {
        return IX_ERROR_NEW(IX_RM_ERROR_COMM_QUEUE_FULL, IX_ERROR_LEVEL_WARNING);
    }

    /* lock the access to the queue */
    _IX_RM_EP_DATA_ACTION_LOCK(arg_pEPCCData);

    /* check again if the queue is full now as we have access locked */
    if(IX_EP_DATA_QUEUE_IS_FULL(pDQueue))
    {
        _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData);
        return IX_ERROR_NEW(IX_RM_ERROR_COMM_QUEUE_FULL, IX_ERROR_LEVEL_WARNING);
    }

    /* put the data on the queue and notify the receive queue about this */
    _IX_EP_DATA_QUEUE_PUT_COMM_ELEMENT_MP(pDQueue, arg_hBuffer, arg_UserData, arg_pEPCCData);

    /* unlock the access to the queue */
    _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData);

    return err;
}


/**
 * NAME: _ix_rm_cc_queue_element_block_mp
 *
 * DESCRIPTION: This function will queue a data element into specified core component end point
 *          data queue. The function is blocking. If the queue is full then the call will wait for
 *          the specified timeout for the queue to become available. If the timeout is infinite then
 *          the function will return only when was able to queue the element. In the case of error
 *          it is the duty of the caller to decide what to do and eventually drop the buffer.
 * 
 * @Param:  - IN arg_pEPCCData -  core component end point data address this fuction will work on
 * @Param:  - IN arg_hBuffer - buffer handle to be stored onto the end point data queue
 * @Param:  - IN arg_UserData - user data to be stored onto the end point data queue
 * @Param:  - IN arg_Timeout - timeout specifying for how long to wait for the queue to become available.
 *          It is expressed in milliseconds.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_cc_queue_element_block_mp( 
                                        ix_end_point_cc_data* arg_pEPCCData,
                                        ix_buffer_handle arg_hBuffer,
                                        ix_uint32 arg_UserData,
                                        ix_uint32 arg_Timeout
                                      )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_cc_queue_element_block_mp";

    ix_error err = 0;

    ix_uint32 timeout;
    ix_ossl_time_t startTime, lastTime;

    ix_ossl_sem_t fullSem = 0;

    ix_end_point_data_queue *pDQueue = &(arg_pEPCCData->m_EPDataQueue);

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);


    /* check if the queue is full */
    if(IX_EP_DATA_QUEUE_IS_FULL(pDQueue))
    {
        /* get the start time */
        ix_ossl_time_get(&startTime);

label_1:
        if(fullSem == 0)
        {
            /* get a new semaphore */
            err =  _ix_rm_sem_pool_get_sem(g_pCommSemaphorePool, &fullSem);
            if(err !=IX_SUCCESS)
                return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_SEM_GET_FAILED, s_aFunctionIdentifier, 
                    "failed to get a blocking full semaphore!", 0);
        }

         /* lock the access to the notification list */
        _IX_RM_EP_DATA_ACTION_LOCK(arg_pEPCCData);

        /* add only if the queue is full */
        if(IX_EP_DATA_QUEUE_IS_FULL(pDQueue))
        {
            /* add the empty semaphore to the send notification list */
            _IX_RM_EP_DATA_WAIT_QUEUE_ADD(arg_pEPCCData->m_pSendNotificationQueue, fullSem);
        }

        /* unlock the access to the notification list */
        _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData);

        if(err != IX_SUCCESS)
            goto label_2;

        /* check again the queue if it is full. Somebody might have consumed from it meanwhile */
        if(IX_EP_DATA_QUEUE_IS_FULL(pDQueue))
        {
            /* get the last time */
            ix_ossl_time_get(&lastTime);

            /* calculate the time elapsed from start */
            IX_OSSL_TIME_SUB(lastTime, startTime);

            /* calculate the elapsed time in millisecons */
            timeout = lastTime.sec * 1000 + lastTime.nsec / 1000000;

            /* check if we passed the timeout */
            if(timeout > arg_Timeout)
            {
                err = IX_ERROR_NEW(IX_OSSL_ERROR_TIMEOUT, IX_ERROR_LEVEL_LOCAL);
                goto label_2;
            }

            /* adjust the timeout */
            timeout = arg_Timeout - timeout;

            /* wait on the full semaphore for the queue to become available */
            err = ix_ossl_sem_take(fullSem,timeout);

            /* check for errors or timeout */
            if(err != IX_SUCCESS)
            {
                /* lock the access to the notification list */
                _IX_RM_EP_DATA_ACTION_LOCK(arg_pEPCCData);

                /* make sure we remove the full semaphore from the send notification list */
                _IX_RM_EP_DATA_WAIT_QUEUE_REMOVE(arg_pEPCCData->m_pSendNotificationQueue, fullSem);

                /* unlock the access to the notification list */
                _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData);

                goto label_2;
            };

            /* check again if the queue is full. Another thread might have put data on the queue meanwhile */
            if(IX_EP_DATA_QUEUE_IS_FULL(pDQueue))
            {
                /* wait again for the queue to become available */
                goto label_1;
            }
        }
    }

    /* lock the access to the queue */
    _IX_RM_EP_DATA_ACTION_LOCK(arg_pEPCCData);

    /* check again if the queue is full now as we have the queue access locked */
    if(IX_EP_DATA_QUEUE_IS_FULL(pDQueue))
    {
        _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData);
        /* wait again for the queue to become available */
        goto label_1;
    }

    /* put the data on the queue and notify the receive queue about that */
    _IX_EP_DATA_QUEUE_PUT_COMM_ELEMENT_MP(pDQueue, arg_hBuffer, arg_UserData, arg_pEPCCData);

    /* unlock the access to the queue */
    _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData);

label_2:

    if(fullSem != 0)
        _ix_rm_sem_pool_release_sem(g_pCommSemaphorePool, fullSem);

    return err;
}


/**
 * Single producer queueing functions
 */

/**
 * NAME: _ix_rm_cc_queue_element_nonblock_sp
 *
 * DESCRIPTION: This function will queue a data element into specified core component end point
 *          data queue. The function is non blocking. If the queue is full then an error will be
 *          returned right away. In case of error it is the duty of the caller to decide what to 
 *          do and eventually drop the buffer.
 * 
 * @Param:  - IN arg_pEPCCData -  core component end point data address this fuction will work on
 * @Param:  - IN arg_hBuffer - buffer handle to be stored onto the end point data queue
 * @Param:  - IN arg_UserData - user data to be stored onto the end point data queue
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_cc_queue_element_nonblock_sp( 
                                           ix_end_point_cc_data* arg_pEPCCData,
                                           ix_buffer_handle arg_hBuffer,
                                           ix_uint32 arg_UserData
                                         )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_cc_queue_element_nonblock_sp";

    ix_error err = 0;
    ix_end_point_data_queue *pDQueue = &arg_pEPCCData->m_EPDataQueue;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check if the queue is full */
    if(IX_EP_DATA_QUEUE_IS_FULL(pDQueue))
    {
        return IX_ERROR_NEW(IX_RM_ERROR_COMM_QUEUE_FULL, IX_ERROR_LEVEL_WARNING);
    }

    /* put the data on the queue and notify the receive queue about this */
    _IX_EP_DATA_QUEUE_PUT_COMM_ELEMENT_SP(pDQueue, arg_hBuffer, arg_UserData, arg_pEPCCData);


    return err;
}


/**
 * NAME: _ix_rm_cc_queue_element_block_sp
 *
 * DESCRIPTION: This function will queue a data element into specified core component end point
 *          data queue. The function is blocking. If the queue is full then the call will wait for
 *          the specified timeout for the queue to become available. If the timeout is infinite then
 *          the function will return only when was able to queue the element. In the case of error
 *          it is the duty of the caller to decide what to do and eventually drop the buffer.
 * 
 * @Param:  - IN arg_pEPCCData -  core component end point data address this fuction will work on
 * @Param:  - IN arg_hBuffer - buffer handle to be stored onto the end point data queue
 * @Param:  - IN arg_UserData - user data to be stored onto the end point data queue
 * @Param:  - IN arg_Timeout - timeout specifying for how long to wait for the queue to become available.
 *          It is expressed in milliseconds.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_cc_queue_element_block_sp( 
                                        ix_end_point_cc_data* arg_pEPCCData,
                                        ix_buffer_handle arg_hBuffer,
                                        ix_uint32 arg_UserData,
                                        ix_uint32 arg_Timeout
                                      )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_cc_queue_element_block_sp";

    ix_error err = 0;

    ix_ossl_sem_t fullSem = 0;

    ix_end_point_data_queue *pDQueue = &(arg_pEPCCData->m_EPDataQueue);

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check if the queue is full */
    if(IX_EP_DATA_QUEUE_IS_FULL(pDQueue))
    {
        /* get a new semaphore */
        err =  _ix_rm_sem_pool_get_sem(g_pCommSemaphorePool, &fullSem);
        if(err !=IX_SUCCESS)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_SEM_GET_FAILED, s_aFunctionIdentifier, 
                "failed to get a blocking full semaphore!", 0);

         /* lock the access to the notification list */
        _IX_RM_EP_DATA_ACTION_LOCK(arg_pEPCCData);

        /* add only if the queue is full */
        if(IX_EP_DATA_QUEUE_IS_FULL(pDQueue))
        {
            /* add the empty semaphore to the send notification list */
            _IX_RM_EP_DATA_WAIT_QUEUE_ADD(arg_pEPCCData->m_pSendNotificationQueue, fullSem);
        }

        /* unlock the access to the notification list */
        _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData);

        if(err != IX_SUCCESS)
            goto label_1;

        /* wait on the full semaphore for the queue to become available */
        err = ix_ossl_sem_take(fullSem, arg_Timeout);

        /* check for errors or timeout */
        if(err != IX_SUCCESS)
        {
            /* lock the access to the notification list */
            _IX_RM_EP_DATA_ACTION_LOCK(arg_pEPCCData);

            /* make sure we remove the full semaphore from the send notification list */
            _IX_RM_EP_DATA_WAIT_QUEUE_REMOVE(arg_pEPCCData->m_pSendNotificationQueue, fullSem);

            /* unlock the access to the notification list */
            _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData);

            goto label_1;
        }
    }

    /* put the data on the queue and notify the receive queue about that */
    _IX_EP_DATA_QUEUE_PUT_COMM_ELEMENT_SP(pDQueue, arg_hBuffer, arg_UserData, arg_pEPCCData);

label_1:

    if(fullSem != 0)
        _ix_rm_sem_pool_release_sem(g_pCommSemaphorePool, fullSem);

    return err;
}

/**
 * Multi consumer dequeueing functions
 */

/**
 * NAME: _ix_rm_cc_dequeue_element_nonblock_mc
 *
 * DESCRIPTION: This function will dequeue a data element from the specified core component end point
 *          data queue. The function is non blocking. If the queue is empty then an error will be
 *          returned right away. In case of error it is the duty of the caller to decide what to 
 *          do.
 * 
 * @Param:  - IN arg_pEPCCData -  core component end point data address this fuction will work on
 * @Param:  - OUT arg_pBufferHandle - address where the dequeued buffer handle will be stored on return.
 * @Param:  - OUT arg_UserData - address where the dequeued user data will be stored on return.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_cc_dequeue_element_nonblock_mc( 
                                             ix_end_point_cc_data* arg_pEPCCData,
                                             ix_buffer_handle* arg_pBufferHandle,
                                             ix_uint32* arg_pUserData
                                           )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_cc_dequeue_element_nonblock_mc";

    ix_error err = 0;
    ix_end_point_data_queue *pDQueue = &arg_pEPCCData->m_EPDataQueue;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check if the queue is empty */
    if(IX_EP_DATA_QUEUE_IS_EMPTY(pDQueue))
    {
        return IX_ERROR_NEW(IX_RM_ERROR_COMM_QUEUE_EMPTY, IX_ERROR_LEVEL_WARNING);
    }

    /* lock the access to the queue */
    _IX_RM_EP_DATA_ACTION_LOCK(arg_pEPCCData);

    /* check again if the queue is empty now as we have access locked */
    if(IX_EP_DATA_QUEUE_IS_EMPTY(pDQueue))
    {
        _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData);
        return IX_ERROR_NEW(IX_RM_ERROR_COMM_QUEUE_EMPTY, IX_ERROR_LEVEL_WARNING);
    }

    /* get the data from the queue and notify the send queue that space is available */
    _IX_EP_DATA_QUEUE_GET_COMM_ELEMENT_MC(pDQueue, arg_pBufferHandle, arg_pUserData, arg_pEPCCData);

    /* unlock the access to the queue */
    _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData);

    return err;
}


/**
 * NAME: _ix_rm_cc_dequeue_element_block_mc
 *
 * DESCRIPTION: This function will dequeue a data element from the specified core component end point
 *          data queue. The function is blocking. If the queue is empty then the call will wait for
 *          the specified timeout for data to become available. If the timeout is infinite then
 *          the function will return only when was able to dequeue an element. In the case of error
 *          it is the duty of the caller to decide what to do.
 * 
 * @Param:  - IN arg_pEPCCData -  core component end point data address this fuction will work on
 * @Param:  - OUT arg_pBufferHandle - address where the dequeued buffer handle will be stored on return.
 * @Param:  - OUT arg_UserData - address where the dequeued user data will be stored on return.
 * @Param:  - IN arg_Timeout - timeout specifying for how long to wait for data to become available.
 *          It is expressed in milliseconds.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_cc_dequeue_element_block_mc( 
                                          ix_end_point_cc_data* arg_pEPCCData,
                                          ix_buffer_handle* arg_pBufferHandle,
                                          ix_uint32* arg_pUserData,
                                          ix_uint32 arg_Timeout
                                        )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_cc_dequeue_element_block_mc";

    ix_error err = 0;

    ix_uint32 timeout;
    ix_ossl_time_t startTime, lastTime;

    ix_ossl_sem_t emptySem = 0;

    ix_end_point_data_queue *pDQueue = &arg_pEPCCData->m_EPDataQueue;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check if the queue is empty */
    if(IX_EP_DATA_QUEUE_IS_EMPTY(pDQueue))
    {
        /* get the start time */
        ix_ossl_time_get(&startTime);

label_1:
        
        if(emptySem == 0)
        {
            /* get a new semaphore */
            err =  _ix_rm_sem_pool_get_sem(g_pCommSemaphorePool, &emptySem);
            if(err !=IX_SUCCESS)
                return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_SEM_GET_FAILED, s_aFunctionIdentifier, 
                    "failed to get a blocking empty semaphore!", 0);
        }

        /* lock the access to the notification list */
        _IX_RM_EP_DATA_ACTION_LOCK(arg_pEPCCData);

        /* add only if the queue is empty */
        if(IX_EP_DATA_QUEUE_IS_EMPTY(pDQueue))
        {
            /* add the empty semaphore to the receive notification list */
            _IX_RM_EP_DATA_WAIT_QUEUE_ADD(arg_pEPCCData->m_pReceiveNotificationQueue, emptySem);
        }

        /* unlock the access to the notification list */
        _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData);

        if(err != IX_SUCCESS)
            goto label_2;

        /* check again the queue if it is empty. Somebody might have produced to it meanwhile */
        if(IX_EP_DATA_QUEUE_IS_EMPTY(pDQueue))
        {
            /* get the last time */
            ix_ossl_time_get(&lastTime);

            /* calculate the time elapsed from start */
            IX_OSSL_TIME_SUB(lastTime, startTime);

            /* calculate the elapsed time in millisecons */
            timeout = lastTime.sec * 1000 + lastTime.nsec / 1000000;

            /* check if we passed the timeout */
            if(timeout > arg_Timeout)
            {
                err = IX_ERROR_NEW(IX_OSSL_ERROR_TIMEOUT, IX_ERROR_LEVEL_LOCAL);
                goto label_2;
            }

            /* adjust the timeout */
            timeout = arg_Timeout - timeout;

            /* wait on the empty semaphore for the queue to become available */
            err = ix_ossl_sem_take(emptySem,timeout);

            /* check for errors or timeout */
            if(err != IX_SUCCESS)
            {
                /* lock the access to the notification list */
                _IX_RM_EP_DATA_ACTION_LOCK(arg_pEPCCData);

                /* make sure we remove the empty semaphore from the receive notification list */
                _IX_RM_EP_DATA_WAIT_QUEUE_REMOVE(arg_pEPCCData->m_pReceiveNotificationQueue, emptySem);

                /* unlock the access to the notification list */
                _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData);

                goto label_2;
            }

            /* check again if the queue is empty. Another thread might have consumed data from the queue meanwhile */
            if(IX_EP_DATA_QUEUE_IS_EMPTY(pDQueue))
            {
                /* wait again for the queue to become available */
                goto label_1;
            }
        }
    }

    /* lock the access to the queue */
    _IX_RM_EP_DATA_ACTION_LOCK(arg_pEPCCData);

    /* check again if the queue is empty now as we have the queue access locked */
    if(IX_EP_DATA_QUEUE_IS_EMPTY(pDQueue))
    {
        _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData);
        /* wait again for the queue to become available */
        goto label_1;
    }

    /* get the data from the queue and notify the send queue that space is available */
    _IX_EP_DATA_QUEUE_GET_COMM_ELEMENT_MC(pDQueue, arg_pBufferHandle, arg_pUserData, arg_pEPCCData);

    /* unlock the access to the queue */
    _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData);

label_2:

    if(emptySem != 0)
        _ix_rm_sem_pool_release_sem(g_pCommSemaphorePool, emptySem);

    return err;
}


/**
 * Single consumer dequeueing functions
 */

/**
 * NAME: _ix_rm_cc_dequeue_element_nonblock_sc
 *
 * DESCRIPTION: This function will dequeue a data element from the specified core component end point
 *          data queue. The function is non blocking. If the queue is empty then an error will be
 *          returned right away. In case of error it is the duty of the caller to decide what to 
 *          do.
 * 
 * @Param:  - IN arg_pEPCCData -  core component end point data address this fuction will work on
 * @Param:  - OUT arg_pBufferHandle - address where the dequeued buffer handle will be stored on return.
 * @Param:  - OUT arg_UserData - address where the dequeued user data will be stored on return.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_cc_dequeue_element_nonblock_sc( 
                                             ix_end_point_cc_data* arg_pEPCCData,
                                             ix_buffer_handle* arg_pBufferHandle,
                                             ix_uint32* arg_pUserData
                                           )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_cc_dequeue_element_nonblock_sc";

    ix_error err = 0;
    ix_end_point_data_queue *pDQueue = &arg_pEPCCData->m_EPDataQueue;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check if the queue is empty */
    if(IX_EP_DATA_QUEUE_IS_EMPTY(pDQueue))
    {
        return IX_ERROR_NEW(IX_RM_ERROR_COMM_QUEUE_EMPTY, IX_ERROR_LEVEL_WARNING);
    }

    /* get the data from the queue and notify the send queue that space is available */
    _IX_EP_DATA_QUEUE_GET_COMM_ELEMENT_SC(pDQueue, arg_pBufferHandle, arg_pUserData, arg_pEPCCData);

    return err;
}


/**
 * NAME: _ix_rm_cc_dequeue_element_block_sc
 *
 * DESCRIPTION: This function will dequeue a data element from the specified core component end point
 *          data queue. The function is blocking. If the queue is empty then the call will wait for
 *          the specified timeout for data to become available. If the timeout is infinite then
 *          the function will return only when was able to dequeue an element. In the case of error
 *          it is the duty of the caller to decide what to do.
 * 
 * @Param:  - IN arg_pEPCCData -  core component end point data address this fuction will work on
 * @Param:  - OUT arg_pBufferHandle - address where the dequeued buffer handle will be stored on return.
 * @Param:  - OUT arg_UserData - address where the dequeued user data will be stored on return.
 * @Param:  - IN arg_Timeout - timeout specifying for how long to wait for data to become available.
 *          It is expressed in milliseconds.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_cc_dequeue_element_block_sc( 
                                          ix_end_point_cc_data* arg_pEPCCData,
                                          ix_buffer_handle* arg_pBufferHandle,
                                          ix_uint32* arg_pUserData,
                                          ix_uint32 arg_Timeout
                                        )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_cc_dequeue_element_block_sc";

    ix_error err = 0;

    ix_ossl_sem_t emptySem = 0;

    ix_end_point_data_queue *pDQueue = &arg_pEPCCData->m_EPDataQueue;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check if the queue is empty */
    if(IX_EP_DATA_QUEUE_IS_EMPTY(pDQueue))
    {
        /* get the start time */

        /* get a new semaphore */
        err =  _ix_rm_sem_pool_get_sem(g_pCommSemaphorePool, &emptySem);
        if(err !=IX_SUCCESS)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_SEM_GET_FAILED, s_aFunctionIdentifier, 
                "failed to get a blocking empty semaphore!", 0);

        /* lock the access to the notification list */
        _IX_RM_EP_DATA_ACTION_LOCK(arg_pEPCCData);

        /* add only if the queue is empty */
        if(IX_EP_DATA_QUEUE_IS_EMPTY(pDQueue))
        {
            /* add the empty semaphore to the receive notification list */
            _IX_RM_EP_DATA_WAIT_QUEUE_ADD(arg_pEPCCData->m_pReceiveNotificationQueue, emptySem);
        }

        /* unlock the access to the notification list */
        _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData);

        if(err != IX_SUCCESS)
            goto label_1;

        /* wait on the empty semaphore for the queue to become available */
        err = ix_ossl_sem_take(emptySem, arg_Timeout);

        /* check for errors or timeout */
        if(err != IX_SUCCESS)
        {
            /* lock the access to the notification list */
            _IX_RM_EP_DATA_ACTION_LOCK(arg_pEPCCData);

            /* make sure we remove the empty semaphore from the receive notification list */
            _IX_RM_EP_DATA_WAIT_QUEUE_REMOVE(arg_pEPCCData->m_pReceiveNotificationQueue, emptySem);

            /* unlock the access to the notification list */
            _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData);

            goto label_1;
        }
    }

    /* get the data from the queue and notify the send queue that space is available */
    _IX_EP_DATA_QUEUE_GET_COMM_ELEMENT_SC(pDQueue, arg_pBufferHandle, arg_pUserData, arg_pEPCCData);

label_1:

    if(emptySem != 0)
        _ix_rm_sem_pool_release_sem(g_pCommSemaphorePool, emptySem);

    return err;
}

/**
 * NAME: _ix_rm_process_queue_entries
 *
 * DESCRIPTION: This function will pass all entries from the queue to the specified
 *          handler for processing.
 * 
 * @Param:  - INOUT arg_pEPCCData - cc end point internal queue to be processed
 * @Param:  - IN arg_DataHandler - data handler to be used for processing all queue elements.
 * @Param:  - IN arg_pContext - context to be passed to the data handler.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_process_queue_entries( 
                                       ix_end_point_cc_data* arg_pEPCCData,
                                       ix_comm_data_handler arg_DataHandler,
                                       void* arg_pContext
                                     )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_process_queue_entries";

    ix_error err = IX_SUCCESS;
    ix_buffer_handle hBuffer;
    ix_uint32 userData;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);
    while(_ix_rm_cc_dequeue_element_nonblock_mc(arg_pEPCCData, &hBuffer, &userData) ==
        IX_SUCCESS)
    {
        err = arg_DataHandler(hBuffer, userData, arg_pContext);
    }

    return err;
}


/**
 * NAME: _ix_rm_cc_select
 *
 * DESCRIPTION: This function will wait on a local set of core component communication IDs 
 *          to have messages and/or data available. The arg_SelectActionSet parameter is an
 *          array of 32 bit flags specifying what kind of data we are waiting for on each of
 *          the specified CC identifiers. The first element of the array coresponds to the
 *          first CC end point. You can not wait for uBlock points for data. The function will
 *          wait for data for a maximum time specified by arg_Timeout parameter.
 * 
 * @Param:  - INOUT arg_SelectActionSet - array of 32bit flags representing what kind of data we
 *          are waiting on packets/messages for each CC end point. On return the flags will be 
 *          cleared and will specify only the data type that has been received on each local
 *          CC communication ID.
 * @Param:  - IN arg_Timeout - timeout specifying for how long we want to wait for data
 *          to become available. It is expressed in milliseconds.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_cc_select( 
                           ix_comm_select_action_set arg_SelectActionSet, 
                           ix_uint32 arg_Timeout 
                         )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_cc_select";

    ix_error err = 0;
    ix_uint32 isEmptySet = 1;

    struct
    {
        ix_end_point_cc_data* m_pEPCCData;
        ix_uint32 m_InternalId;
    } requestedSet[2 * IX_COMM_CORE_ID_NUMBER];
    ix_uint32 i,j;
    ix_uint32 action;
    ix_end_point_cc_data* pData;

    ix_end_point_cc_data* pEPCCData;
    ix_uint32 cbMode = 0;

    ix_uint32 timeout;
    ix_ossl_time_t startTime, lastTime;

    ix_ossl_sem_t emptySem = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* build the requested set and clear the select action set */
    for(i = 0, j = 0; i < IX_COMM_CORE_ID_NUMBER; i++)
    {
        action = arg_SelectActionSet[i];
        if(action & IX_COMM_SELECT_PACKET_MASK)
        {
            pEPCCData = 
                &g_EndPointManager.m_aPacketEndPoint[IX_COMM_FIRST_CORE_ID + i].m_EPData.m_CoreCompData;
            if((pEPCCData->m_Flags & IX_COMM_ID_RECEIVE_MODE_MASK) == IX_COMM_ID_MODE_CALLBACK)
            {
                cbMode = 1;
                break;
            }
            else
                requestedSet[j].m_pEPCCData = pEPCCData;

            requestedSet[j].m_InternalId = i;
            j++;
        }

        if(action & IX_COMM_SELECT_MESSAGE_MASK)
        {
            pEPCCData = 
                &g_EndPointManager.m_aMessageEndPoint[IX_COMM_FIRST_CORE_ID + i].m_EPData.m_CoreCompData;
            if((pEPCCData->m_Flags & IX_COMM_ID_RECEIVE_MODE_MASK) == IX_COMM_ID_MODE_CALLBACK)
            {
                cbMode = 1;
                break;
            }
            else
                requestedSet[j].m_pEPCCData = pEPCCData;

            requestedSet[j].m_InternalId = i | 0x80000000;
            j++;
        }

        arg_SelectActionSet[i] = 0;
    }

    if(cbMode != 0)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_RECV_CALLBACK, s_aFunctionIdentifier, 
                            "one of the comm Ids we are trying to get data from is in callback mode!", err);

    requestedSet[j].m_pEPCCData = 0;

    for(j = 0, pData = requestedSet[0].m_pEPCCData, isEmptySet = 1; pData != 0; pData = requestedSet[++j].m_pEPCCData)
    {
        isEmptySet &= IX_EP_DATA_QUEUE_IS_EMPTY(&pData->m_EPDataQueue);
    }

    if(isEmptySet)
    {
        /* get a new semaphore */
        err =  _ix_rm_sem_pool_get_sem(g_pCommSemaphorePool, &emptySem);
        if(err !=IX_SUCCESS)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_SEM_GET_FAILED, s_aFunctionIdentifier, 
                "failed to get a blocking empty semaphore!", 0);

        /* get the start time */
        ix_ossl_time_get(&startTime);

label_1:

        /* add the empty semaphore to the receive notification list of each cc data */
        for(j = 0, pData = requestedSet[0].m_pEPCCData, isEmptySet = 1; pData != 0; pData = requestedSet[++j].m_pEPCCData)
        {
            /* lock the access to the notification list */
            _IX_RM_EP_DATA_ACTION_LOCK(pData);

            /* add the empty semaphore to the receive notification list */
            _IX_RM_EP_DATA_WAIT_QUEUE_ADD(pData->m_pReceiveNotificationQueue, emptySem);

            /* unlock the access to the notification list */
            _IX_RM_EP_DATA_ACTION_UNLOCK(pData);

            isEmptySet &= IX_EP_DATA_QUEUE_IS_EMPTY(&pData->m_EPDataQueue);
        }

        if(isEmptySet)
        {
            /* get the last time */
            ix_ossl_time_get(&lastTime);

            /* calculate the time elapsed from start */
            IX_OSSL_TIME_SUB(lastTime, startTime);

            /* calculate the elapsed time in millisecons */
            timeout = lastTime.sec * 1000 + lastTime.nsec / 1000000;

            /* check if we passed the timeout */
            if(timeout > arg_Timeout)
            {
                err = IX_ERROR_NEW(IX_OSSL_ERROR_TIMEOUT, IX_ERROR_LEVEL_LOCAL);
                goto label_2;
            }

            /* adjust the timeout */
            timeout = arg_Timeout - timeout;

            /* wait on the empty semaphore for the queue to become available */
            err = ix_ossl_sem_take(emptySem,timeout);

            /* check for errors or timeout */
            if(err != IX_SUCCESS)
            {
                /* remove the empty semaphore from the receive notification list of each cc data */
                for(j = 0, pData = requestedSet[0].m_pEPCCData; pData != 0; pData = requestedSet[++j].m_pEPCCData)
                {
                    /* lock the access to the notification list */
                    _IX_RM_EP_DATA_ACTION_LOCK(pData);

                    /* add the empty semaphore to the receive notification list */
                    _IX_RM_EP_DATA_WAIT_QUEUE_REMOVE(pData->m_pReceiveNotificationQueue, emptySem);

                    /* unlock the access to the notification list */
                    _IX_RM_EP_DATA_ACTION_UNLOCK(pData);
                }

                goto label_2;
            };

            /* remove the empty semaphore from the receive notification list of each cc data */
            for(j = 0, pData = requestedSet[0].m_pEPCCData, isEmptySet = 1; pData != 0; pData = requestedSet[++j].m_pEPCCData)
            {
                /* lock the access to the notification list */
                _IX_RM_EP_DATA_ACTION_LOCK(pData);

                /* add the empty semaphore to the receive notification list */
                _IX_RM_EP_DATA_WAIT_QUEUE_REMOVE(pData->m_pReceiveNotificationQueue, emptySem);

                /* unlock the access to the notification list */
                _IX_RM_EP_DATA_ACTION_UNLOCK(pData);

                isEmptySet &= IX_EP_DATA_QUEUE_IS_EMPTY(&pData->m_EPDataQueue);
            }

            if(isEmptySet)
                goto label_1;
        }
    }

    for(j = 0, pData = requestedSet[0].m_pEPCCData, isEmptySet = 1; pData != 0; pData = requestedSet[++j].m_pEPCCData)
    {
        i = IX_EP_DATA_QUEUE_IS_EMPTY(&pData->m_EPDataQueue);
        isEmptySet &= IX_EP_DATA_QUEUE_IS_EMPTY(&pData->m_EPDataQueue);

        /* setup the return status */
        if(!i)
        {
            i = requestedSet[j].m_InternalId;
            if(i & 0x80000000)
            {
                i &= ~0x80000000;
                arg_SelectActionSet[i] |= IX_COMM_SELECT_MESSAGE_MASK;
            }
            else
            {
                arg_SelectActionSet[i] |= IX_COMM_SELECT_PACKET_MASK;
            }
        }
    }

    if(isEmptySet)
        goto label_1;

label_2:

    if(emptySem != 0)
        _ix_rm_sem_pool_release_sem(g_pCommSemaphorePool, emptySem);

    return err;
}

#if defined(_IX_RM_IMPL_HARDWARE_)

/* ISRs for uBlock/core packet/message communication */
#if defined (_IX_RM_TEST_AI_INTERRUPTS_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

/* ISR for uBlock/core packet/message communication */
void _ix_ublock_to_core_isr(int arg_value)
{
    *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x54))) = 0;

    ix_ossl_message_log("Interrupt\n");

    ((_ix_default_dispatch_engine_context*)arg_value)->m_pEngineInterface->m_NotifyEngineFuncPtr((void*)arg_value);
}

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

/* ISR for uBlock/core packet/message communication */
void _ix_ublock_to_core_isr(int irq, void* context, struct pt_regs* regs)
{
    *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x54))) = 0;

    ix_ossl_message_log("Interrupt\n");

    ((_ix_default_dispatch_engine_context*)context)->m_pEngineInterface->m_NotifyEngineFuncPtr((void*)context);
}

#endif /* end (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) */

#else

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

void _ix_ublock_to_core_packet_isr(int arg_value)
{
    /* clear the interrupt source */
    *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x60))) = 0xffffffff;
    *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x68))) = 0xffffffff;

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)


#else

    *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x64))) = 0xffffffff;
    *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x6C))) = 0xffffffff;

#endif

    /* disable the packet interrupt just not to flood the core with interrupts */
    IX_RM_COMM_PACKET_INTERRUPT_DISABLE(0);

#if defined(_IX_TEST_THREAD_SWITCHES_)

    /* Increment the interrupt count */
    g_PacketInterruptCount++;

#endif /* end defined (_IX_TEST_THREAD_SWITCHES_) */

    /* ix_ossl_message_log("Interrupt thread 1\n"); */

    ((_ix_default_dispatch_engine_context*)arg_value)->m_pEngineInterface->m_NotifyEngineFuncPtr((void*)arg_value);
}


void _ix_ublock_to_core_message_isr(int arg_value)
{
    /* clear the interrupt source */
    *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x80))) = 0xffffffff;
    *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x88))) = 0xffffffff;

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)


#else

    *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x84))) = 0xffffffff;
    *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x8C))) = 0xffffffff;

#endif

    /* disable the message interrupt just not to flood the core with interrupts */
    IX_RM_COMM_MESSAGE_INTERRUPT_DISABLE(0);

    ((_ix_default_dispatch_engine_context*)arg_value)->m_pEngineInterface->m_NotifyEngineFuncPtr((void*)arg_value);
}

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

void _ix_ublock_to_core_packet_isr(int irq, void* context, struct pt_regs* regs)
{
    /* clear the interrupt source */

    *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x60))) = 0xffffffff;
    *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x68))) = 0xffffffff;


#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)


#else

    *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x64))) = 0xffffffff;
    *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x6C))) = 0xffffffff;

#endif

    /* disable the packet interrupt just not to flood the core with interrupts */
    IX_RM_COMM_PACKET_INTERRUPT_DISABLE(0);

    /* ix_ossl_message_log("Interrupt thread 1\n"); */

    ((_ix_default_dispatch_engine_context*)context)->m_pEngineInterface->m_NotifyEngineFuncPtr(context);

}


void _ix_ublock_to_core_message_isr(int irq, void* context, struct pt_regs* regs)
{
    /* clear the interrupt source */
    *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x80))) = 0xffffffff;
    *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x88))) = 0xffffffff;

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)


#else

    *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x84))) = 0xffffffff;
    *(((volatile ix_uint32*)(IX_XSCALE_CSR_BASE_ADDR + 0x8C))) = 0xffffffff;

#endif

    /* disable the message interrupt just not to flood the core with interrupts */
    IX_RM_COMM_MESSAGE_INTERRUPT_DISABLE(0);

    ((_ix_default_dispatch_engine_context*)context)->m_pEngineInterface->m_NotifyEngineFuncPtr((void*)context);

}

#endif

#endif /* end  defined (_IX_RM_TEST_AI_INTERRUPTS_) */


#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */





/**
 * PCI communication support. 
 * We need all of these functions because there is a need to put 3 longwords on the
 * internal queue at a time, different from the normal processing which requires just
 * 2. We need to pass the destination commId as well.
 */


/**
 * NAME: _ix_rm_cc_pci_queue_element_nonblock_mp
 *
 * DESCRIPTION: This function will queue a pci data element into specified end point
 *          data queue. The function is non blocking. If the queue is full then an error will be
 *          returned right away. In case of error it is the duty of the caller to decide what to 
 *          do and eventually drop the buffer. The element in this case will include the destination
 *          comm Id.
 * 
 * @Param:  - IN arg_pEPCCData -  core component end point data address this fuction will work on
 * @Param:  - IN arg_hBuffer - buffer handle to be stored onto the end point data queue
 * @Param:  - IN arg_UserData - user data to be stored onto the end point data queue
 * @Param:  - IN arg_Destination - destination comm Id to be stored onto the end point data queue
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_cc_pci_queue_element_nonblock_mp( 
                                           ix_end_point_cc_data* arg_pEPCCData,
                                           ix_buffer_handle arg_hBuffer,
                                           ix_uint32 arg_UserData,
                                           ix_communication_id arg_Destination
                                         )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_cc_pci_queue_element_nonblock_mp";

    ix_error err = 0;
    ix_end_point_data_queue *pDQueue = &arg_pEPCCData->m_EPDataQueue;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check if the queue is full */
    if(IX_EP_DATA_QUEUE_IS_FULL(pDQueue))
    {
        return IX_ERROR_NEW(IX_RM_ERROR_COMM_QUEUE_FULL, IX_ERROR_LEVEL_WARNING);
    }

    /* lock the access to the queue */
    _IX_RM_EP_DATA_ACTION_LOCK(arg_pEPCCData);

    /* check again if the queue is full now as we have access locked */
    if(IX_EP_DATA_QUEUE_IS_FULL(pDQueue))
    {
        _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData);
        return IX_ERROR_NEW(IX_RM_ERROR_COMM_QUEUE_FULL, IX_ERROR_LEVEL_WARNING);
    }

    /* put the data on the queue and notify the receive queue about this */
    _IX_EP_DATA_QUEUE_PCI_PUT_COMM_ELEMENT_MP(pDQueue, arg_hBuffer, arg_UserData, arg_Destination, arg_pEPCCData);

    /* unlock the access to the queue */
    _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData);

    return err;
}


/**
 * NAME: _ix_rm_cc_pci_queue_element_block_mp
 *
 * DESCRIPTION: This function will queue a pci data element into the specified end point
 *          data queue. The function is blocking. If the queue is full then the call will wait for
 *          the specified timeout for the queue to become available. If the timeout is infinite then
 *          the function will return only when was able to queue the element. In the case of error
 *          it is the duty of the caller to decide what to do and eventually drop the buffer.
 *          The element in this case will include the destination comm Id.
 * 
 * @Param:  - IN arg_pEPCCData -  core component end point data address this fuction will work on
 * @Param:  - IN arg_hBuffer - buffer handle to be stored onto the end point data queue
 * @Param:  - IN arg_UserData - user data to be stored onto the end point data queue
 * @Param:  - IN arg_Destination - destination comm Id to be stored onto the end point data queue
 * @Param:  - IN arg_Timeout - timeout specifying for how long to wait for the queue to become available.
 *          It is expressed in milliseconds.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_cc_pci_queue_element_block_mp( 
                                        ix_end_point_cc_data* arg_pEPCCData,
                                        ix_buffer_handle arg_hBuffer,
                                        ix_uint32 arg_UserData,
                                        ix_communication_id arg_Destination,
                                        ix_uint32 arg_Timeout
                                      )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_cc_pci_queue_element_block_mp";

    ix_error err = 0;

    ix_uint32 timeout;
    ix_ossl_time_t startTime, lastTime;

    ix_ossl_sem_t fullSem = 0;

    ix_end_point_data_queue *pDQueue = &(arg_pEPCCData->m_EPDataQueue);

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);


    /* check if the queue is full */
    if(IX_EP_DATA_QUEUE_IS_FULL(pDQueue))
    {
        /* get the start time */
        ix_ossl_time_get(&startTime);

label_1:
        if(fullSem == 0)
        {
            /* get a new semaphore */
            err =  _ix_rm_sem_pool_get_sem(g_pCommSemaphorePool, &fullSem);
            if(err !=IX_SUCCESS)
                return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_SEM_GET_FAILED, s_aFunctionIdentifier, 
                    "failed to get a blocking full semaphore!", 0);
        }

         /* lock the access to the notification list */
        _IX_RM_EP_DATA_ACTION_LOCK(arg_pEPCCData);

        /* add only if the queue is full */
        if(IX_EP_DATA_QUEUE_IS_FULL(pDQueue))
        {
            /* add the empty semaphore to the send notification list */
            _IX_RM_EP_DATA_WAIT_QUEUE_ADD(arg_pEPCCData->m_pSendNotificationQueue, fullSem);
        }

        /* unlock the access to the notification list */
        _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData);

        if(err != IX_SUCCESS)
            goto label_2;

        /* check again the queue if it is full. Somebody might have consumed from it meanwhile */
        if(IX_EP_DATA_QUEUE_IS_FULL(pDQueue))
        {
            /* get the last time */
            ix_ossl_time_get(&lastTime);

            /* calculate the time elapsed from start */
            IX_OSSL_TIME_SUB(lastTime, startTime);

            /* calculate the elapsed time in millisecons */
            timeout = lastTime.sec * 1000 + lastTime.nsec / 1000000;

            /* check if we passed the timeout */
            if(timeout > arg_Timeout)
            {
                err = IX_ERROR_NEW(IX_OSSL_ERROR_TIMEOUT, IX_ERROR_LEVEL_LOCAL);
                goto label_2;
            }

            /* adjust the timeout */
            timeout = arg_Timeout - timeout;

            /* wait on the full semaphore for the queue to become available */
            err = ix_ossl_sem_take(fullSem,timeout);

            /* check for errors or timeout */
            if(err != IX_SUCCESS)
            {
                /* lock the access to the notification list */
                _IX_RM_EP_DATA_ACTION_LOCK(arg_pEPCCData);

                /* make sure we remove the full semaphore from the send notification list */
                _IX_RM_EP_DATA_WAIT_QUEUE_REMOVE(arg_pEPCCData->m_pSendNotificationQueue, fullSem);

                /* unlock the access to the notification list */
                _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData);

                goto label_2;
            };

            /* check again if the queue is full. Another thread might have put data on the queue meanwhile */
            if(IX_EP_DATA_QUEUE_IS_FULL(pDQueue))
            {
                /* wait again for the queue to become available */
                goto label_1;
            }
        }
    }

    /* lock the access to the queue */
    _IX_RM_EP_DATA_ACTION_LOCK(arg_pEPCCData);

    /* check again if the queue is full now as we have the queue access locked */
    if(IX_EP_DATA_QUEUE_IS_FULL(pDQueue))
    {
        _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData);
        /* wait again for the queue to become available */
        goto label_1;
    }

    /* put the data on the queue and notify the receive queue about that */
    _IX_EP_DATA_QUEUE_PCI_PUT_COMM_ELEMENT_MP(pDQueue, arg_hBuffer, arg_UserData, arg_Destination, arg_pEPCCData);

    /* unlock the access to the queue */
    _IX_RM_EP_DATA_ACTION_UNLOCK(arg_pEPCCData);

label_2:

    if(fullSem != 0)
        _ix_rm_sem_pool_release_sem(g_pCommSemaphorePool, fullSem);

    return err;
}





