/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      internal_configuration_property.c
 *
 * = DESCRIPTION
 *      This file provides implementation for all support functions on the configuration
 *      property.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      5/14/2002 10:17:30 AM - creation time 
 * ============================================================================
 */

static const char s_FileIdentifier[] = "$Id: internal_configuration_property.c,v 1.10 2003/09/17 17:06:02 ageorges Exp $";

#include "ix_ossl.h"

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"

#include "rm/symbols.h"
#include "rm/internal/internal_support.h"

#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"

#include "rm/system.h"
#include "rm/internal/internal_system.h"

#include "rm/memory.h"
#include "rm/buffer.h"
#include "rm/internal/internal_buffer.h"
#include "rm/hardware.h"
#include "rm/communication.h"

#include "rm/configuration_property.h"
#include "rm/internal/internal_configuration_property.h"


/**
 * The configuration property API will be implemented on top of Buffer API by
 * using a software free list as a pool of properties.
 */

/**
 * Global variable representing the configuration property manager.
 */
ix_cp_manager g_CPManager;


#if defined( _IX_RM_MULTI_THREADED_ )

/**
 * Synchronization semaphore for CP manager
 */
static ix_ossl_sem_t s_CPManagerSemaphore;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */



/**
 * Predicate that will find if the configuration property has the passed name
 */
ix_configuration_property_handle _ix_cp_tree_predicate_has_name(
                                  ix_configuration_property_handle arg_hCP, 
                                  const char* arg_pCPName
                                  )
{
    static const char s_aFunctionIdentifier[] = "_ix_cp_tree_predicate_has_name";

    ix_error err = IX_SUCCESS;

    ix_cp_property_attributes* pCPAttributes;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    err = ix_rm_buffer_get_meta(arg_hCP, (void**)&pCPAttributes);
    if(err != IX_SUCCESS)
        return IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
    
    return (strcmp(arg_pCPName,pCPAttributes->m_aName) == 0) ? arg_hCP :
        IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
}


/**
 * Predicate that will delete the configuration property without any effects
 */
ix_configuration_property_handle _ix_cp_tree_predicate_delete(
                                  ix_configuration_property_handle arg_hCP, 
                                  void* arg_pContext
                                  )
{
    static const char s_aFunctionIdentifier[] = "_ix_cp_tree_predicate_has_name";

    ix_error err = IX_SUCCESS;

    ix_cp_property_attributes* pCPAttributes;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    err = ix_rm_buffer_get_meta(arg_hCP, (void**)&pCPAttributes);
    if(err != IX_SUCCESS)
        return IX_NULL_CONFIGURATION_PROPERTY_HANDLE;

    /* remove the notification comm Id to the list */
    {
        ix_cp_property_notification_token* pLastToken;
        ix_cp_property_notification_token* pCurrentToken = pCPAttributes->m_pNotificationTokenList;
        while(pCurrentToken != 0)
        {
            pLastToken = pCurrentToken;
            pCurrentToken = pCurrentToken->m_pNext;
            ix_ossl_free(pLastToken);
        }
        pCPAttributes->m_pNotificationTokenList = 0;
    }

    if(pCPAttributes->m_DataSize <= IX_CP_DATA_ELEMENT_SIZE)
    {
        ix_rm_buffer_free(arg_hCP);
    }
    else
    {
        /* clear the EOP bit */
        IX_RM_BUF_SET_EOP(arg_hCP, 0);

        ix_rm_buffer_free_chain(arg_hCP);
    }

    return IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
}

#if 0 /* compile it out for metrics purposes */

/**
 * NAME: _ix_rm_cp_tree_pre_order_traverse
 *
 * DESCRIPTION: This function will 
 * 
 * @Param:  - IN arg_hCPRoot -  the CP that will be the root of this operation
 * @Param:  - IN arg_Predicate - the predicate to be used in this operation
 * @Param:  - IN arg_pPredicateContext - the context to be passed to the predicate
 *
 * @Return: IX_NULL_CONFIGURATION_PROPERTY_HANDLE for failure or a valid configuration
 *  property handle for success.
 */
ix_configuration_property_handle _ix_rm_cp_tree_pre_order_traverse( 
                                            ix_configuration_property_handle arg_hCPRoot, 
                                            ix_cp_tree_predicate arg_Predicate, 
                                            void* arg_pPredicateContext
                                          )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_cp_tree_pre_order_traverse";

    ix_error err = IX_SUCCESS;

    ix_configuration_property_handle hCPReturn;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if((hCPReturn = arg_Predicate(arg_hCPRoot, arg_pPredicateContext)) == 
        IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
    {
        ix_cp_property_attributes* pCPAttributes;
        ix_configuration_property_handle hCP; 
        err = ix_rm_buffer_get_meta(arg_hCPRoot, (void**)&pCPAttributes);
        if(err != IX_SUCCESS)
            return IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
        hCP = pCPAttributes->m_hChildrenPropertiesListHead;
        while(hCP != IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
        {
            if((hCPReturn = _ix_rm_cp_tree_pre_order_traverse(hCP, 
                arg_Predicate, arg_pPredicateContext)) != 
                IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
                return hCPReturn;
            err = ix_rm_buffer_get_meta(hCP, (void**)&pCPAttributes);
            if(err != IX_SUCCESS)
                return IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
            hCP = pCPAttributes->m_hPropertyNext;
        }
    }

    return hCPReturn;
}

/**
 * NAME: _ix_rm_cp_tree_in_order_traverse
 *
 * DESCRIPTION: This function will 
 * 
 * @Param:  - IN arg_hCPRoot -  the CP that will be the root of this operation
 * @Param:  - IN arg_Predicate - the predicate to be used in this operation
 * @Param:  - IN arg_pPredicateContext - the context to be passed to the predicate
 *
 * @Return: IX_NULL_CONFIGURATION_PROPERTY_HANDLE for failure or a valid configuration
 *  property handle for success.
 */
ix_configuration_property_handle _ix_rm_cp_tree_in_order_traverse( 
                                           ix_configuration_property_handle arg_hCPRoot, 
                                           ix_cp_tree_predicate arg_Predicate, 
                                           void* arg_pPredicateContext
                                         )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_cp_tree_in_order_traverse";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    return IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
}

#endif /* if 0 */

/**
 * NAME: _ix_rm_cp_tree_post_order_traverse
 *
 * DESCRIPTION: This function will 
 * 
 * @Param:  - IN arg_hCPRoot -  the CP that will be the root of this operation
 * @Param:  - IN arg_Predicate - the predicate to be used in this operation
 * @Param:  - IN arg_pPredicateContext - the context to be passed to the predicate
 *
 * @Return: IX_NULL_CONFIGURATION_PROPERTY_HANDLE for failure or a valid configuration
 *  property handle for success.
 */
ix_configuration_property_handle _ix_rm_cp_tree_post_order_traverse( 
                                             ix_configuration_property_handle arg_hCPRoot, 
                                             ix_cp_tree_predicate arg_Predicate, 
                                             void* arg_pPredicateContext
                                           )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_cp_tree_post_order_traverse";

    ix_error err = IX_SUCCESS;

    ix_configuration_property_handle hCPReturn;
    ix_cp_property_attributes* pCPAttributes;
    ix_configuration_property_handle hCP; 

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    err = ix_rm_buffer_get_meta(arg_hCPRoot, (void**)&pCPAttributes);
    if(err != IX_SUCCESS)
        return IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
    hCP = pCPAttributes->m_hChildrenPropertiesListTail;
    while(hCP != IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
    {
        if((hCPReturn = _ix_rm_cp_tree_post_order_traverse(hCP, 
            arg_Predicate, arg_pPredicateContext)) != 
            IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
            return hCPReturn;
        err = ix_rm_buffer_get_meta(hCP, (void**)&pCPAttributes);
        if(err != IX_SUCCESS)
            return IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
        hCP = pCPAttributes->m_hPropertyPrev;
    }

    return arg_Predicate(arg_hCPRoot, arg_pPredicateContext);
}


/**
 * NAME: _ix_rm_cp_tree_children_forward_traverse
 *
 * DESCRIPTION: This function will 
 * 
 * @Param:  - IN arg_hCPRoot -  the CP that will be the root of this operation
 * @Param:  - IN arg_Predicate - the predicate to be used in this operation
 * @Param:  - IN arg_pPredicateContext - the context to be passed to the predicate
 *
 * @Return: IX_NULL_CONFIGURATION_PROPERTY_HANDLE for failure or a valid configuration
 *  property handle for success.
 */
ix_configuration_property_handle _ix_rm_cp_tree_children_forward_traverse( 
                                            ix_configuration_property_handle arg_hCPRoot, 
                                            ix_cp_tree_predicate arg_Predicate, 
                                            void* arg_pPredicateContext
                                          )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_cp_tree_children_forward_traverse";

    ix_error err = IX_SUCCESS;

    ix_configuration_property_handle hCPReturn;
    ix_cp_property_attributes* pCPAttributes;
    ix_configuration_property_handle hCP; 

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    err = ix_rm_buffer_get_meta(arg_hCPRoot, (void**)&pCPAttributes);
    if(err != IX_SUCCESS)
        return IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
    hCP = pCPAttributes->m_hChildrenPropertiesListHead;
    while(hCP != IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
    {
        if((hCPReturn = arg_Predicate(hCP, arg_pPredicateContext)) != 
            IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
            return hCPReturn;
        err = ix_rm_buffer_get_meta(hCP, (void**)&pCPAttributes);
        if(err != IX_SUCCESS)
            return IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
        hCP = pCPAttributes->m_hPropertyNext;
    }

    return IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
}

#if 0 /* compile it out for metrics purposes */

/**
 * NAME: _ix_rm_cp_tree_children_reverse_traverse
 *
 * DESCRIPTION: This function will 
 * 
 * @Param:  - IN arg_hCPRoot -  the CP that will be the root of this operation
 * @Param:  - IN arg_Predicate - the predicate to be used in this operation
 * @Param:  - IN arg_pPredicateContext - the context to be passed to the predicate
 *
 * @Return: IX_NULL_CONFIGURATION_PROPERTY_HANDLE for failure or a valid configuration
 *  property handle for success.
 */
ix_configuration_property_handle _ix_rm_cp_tree_children_reverse_traverse( 
                                            ix_configuration_property_handle arg_hCPRoot, 
                                            ix_cp_tree_predicate arg_Predicate, 
                                            void* arg_pPredicateContext
                                          )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_cp_tree_children_reverse_traverse";

    ix_error err = IX_SUCCESS;
    ix_configuration_property_handle hCPReturn;
    ix_cp_property_attributes* pCPAttributes;
    ix_configuration_property_handle hCP; 

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    err = ix_rm_buffer_get_meta(arg_hCPRoot, (void**)&pCPAttributes);
    if(err != IX_SUCCESS)
        return IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
    hCP = pCPAttributes->m_hChildrenPropertiesListTail;
    while(hCP != IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
    {
        if((hCPReturn = arg_Predicate(hCP, arg_pPredicateContext)) != 
            IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
            return hCPReturn;
        err = ix_rm_buffer_get_meta(hCP, (void**)&pCPAttributes);
        if(err != IX_SUCCESS)
            return IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
        hCP = pCPAttributes->m_hPropertyPrev;
    }

    return IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
}

#endif /* if 0 */

/**
 * NAME: _ix_rm_configuration_property_init
 *
 * DESCRIPTION: This function should be called at the time the resource manager
 *  initializer function is called. The function will initialize all the Configuration
 *  Property API.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_configuration_property_init(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_configuration_property_init";

    ix_error err = 0;

	IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    /* initialize the configuration property manager synchronization */
    err = ix_ossl_sem_init(IX_OSSL_SEM_AVAILABLE, &s_CPManagerSemaphore);
    if(err)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_CP_API_INIT_FAILED, s_aFunctionIdentifier, 
                        "initialization of CP manager's semaphore failed!", err);
    }

    _ix_rm_sa_sync_adapter_init(&g_CPManager.m_SyncAdapter, _ix_rm_sa_default_lock, 
        _ix_rm_sa_default_unlock, &s_CPManagerSemaphore); 
        

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* create a a software buffer free list for use by the CP manager */
    err = ix_rm_sw_buffer_free_list_create(sizeof(ix_cp_property_attributes),
        IX_MEMORY_TYPE_DRAM, IX_CP_DATA_ELEMENT_SIZE, IX_MEMORY_TYPE_DRAM,
        IX_CP_FREE_LIST_SIZE, &g_CPManager.m_hCPFreeList);
    if(err != IX_SUCCESS)
    {
        err =  IX_RM_ERROR_NEW(IX_RM_ERROR_CP_API_INIT_FAILED, s_aFunctionIdentifier, 
                   "the buffer free list initialization for CP manager failed!", err);
        goto label_1;

    }

    /* create the two trees root entries */
    {
        ix_cp_property_attributes* pCPAttributes;

        /* create the core CP root */
        /* create a new software buffer for the property */
        err =  ix_rm_buffer_alloc(g_CPManager.m_hCPFreeList, &g_CPManager.m_hCoreRegistryRoot);
        if(err != IX_SUCCESS)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_CREATION_FAILED, s_aFunctionIdentifier, 
                            "failed to allocate a the core root configuration property!", err);
            goto label_2;
        }

        err = ix_rm_buffer_get_meta(g_CPManager.m_hCoreRegistryRoot, (void**)&pCPAttributes);
        if(err != IX_SUCCESS)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES, s_aFunctionIdentifier, 
                            "failed to access the core root configuration property attributes!", err);
            goto label_2;
        }


        pCPAttributes->m_DataSize = 0;
        pCPAttributes->m_hChildrenPropertiesListHead = 
        pCPAttributes->m_hChildrenPropertiesListTail = 
            IX_NULL_CONFIGURATION_PROPERTY_HANDLE;

        pCPAttributes->m_hPropertyNext = IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
        pCPAttributes->m_hPropertyPrev = IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
        pCPAttributes->m_hParentProperty = IX_NULL_CONFIGURATION_PROPERTY_HANDLE;
        pCPAttributes->m_ReferenceCount = 0;
        pCPAttributes->m_pNotificationTokenList = 0;

        pCPAttributes->m_Options = IX_CP_OPTIONS_FLAGS_CONSTANT;

        strcpy(pCPAttributes->m_aName, "core_root");
    }

    return err;

label_2:

    if(g_CPManager.m_hCoreRegistryRoot != IX_NULL_CONFIGURATION_PROPERTY_HANDLE)
        ix_rm_buffer_free(g_CPManager.m_hCoreRegistryRoot);

    ix_rm_buffer_free_list_delete(g_CPManager.m_hCPFreeList);

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    /* release the allocated semaphore */
    ix_ossl_sem_fini(s_CPManagerSemaphore);
    s_CPManagerSemaphore = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: _ix_rm_configuration_property_fini
 *
 * DESCRIPTION: This function should be called at the time the resource manager 
 *  finalizer function is called. The function will provide cleanup for Configuration
 *  Property API.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_configuration_property_fini(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_configuration_property_fini";

    ix_error err = 0;

	IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* delete the entire trees */

    /* delete the free list */
    err = ix_rm_buffer_free_list_delete(g_CPManager.m_hCPFreeList);

#if defined( _IX_RM_MULTI_THREADED_ )

    _ix_rm_sa_sync_adapter_init(&g_CPManager.m_SyncAdapter, 0, 0, (void*)0);

    err = ix_ossl_sem_fini(s_CPManagerSemaphore);
    if(err != 0)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_CP_API_FINI_FAILED, s_aFunctionIdentifier, 
                        "finalization of CP manager's semaphore failed!", err);
    }

    s_CPManagerSemaphore = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}
