/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      internal_counter_64bit.c
 *
 * = DESCRIPTION
 *      This file will define all variables and functions needed internally
 *      by the Counter 64 bit API.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      5/22/2002 4:16:40 PM - creation time 
 * ============================================================================
 */

static const char s_FileIdentifier[] = "$Id: internal_counter_64bit.c,v 1.15 2003/09/19 21:15:56 ageorges Exp $";

#include "ix_ossl.h"

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"

#include "rm/symbols.h"
#include "rm/internal/internal_support.h"

#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"

#include "rm/system.h"
#include "rm/internal/internal_system.h"

#include "rm/memory.h"

#include "rm/counter_64bit.h"
#include "rm/internal/internal_counter_64bit.h"

#include "rm/atomic_operations.h"

/**
 * Global 64 bit counter manager
 */
_ix_64bit_counter_manager g_64bitCounterManager;


#if defined( _IX_RM_MULTI_THREADED_ )

/**
 * Synchronization semaphore for 64 bit counter manager
 */
static ix_ossl_sem_t s_64bitCounterManagerSemaphore;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

static ix_error _ix_rm_counter_64bit_thread_update_run
                                               (
                                                 _ix_64bit_counter_manager* arg_p64bitCounterManager,
                                                 ...
                                               );

/**
 * NAME: _ix_rm_counter_64bit_sram_update
 *
 * DESCRIPTION: This function will perform the sram 64 bit update counter operation
 * 
 * @Param:  - INOUT arg_p64bitCounter - counter to be updated
 * @Param:  - INOUT arg_pContext - external data to be passed to the update operation
 *
 */
void _ix_rm_counter_64bit_sram_update(
                                       _ix_counter_64bit* arg_p64bitCounter, 
                                       void* arg_pContext
                                     )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_counter_64bit_default_update";

    _ix_64bit_counter_update_context* pUpdateData = (_ix_64bit_counter_update_context*)arg_pContext;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_pContext == 0)
    {
        /* this is a forced update */
        goto label_1;
    }
    else
    {
        arg_p64bitCounter->m_UpdateElapsedTime += pUpdateData->m_ActiveTimeout;

        /* check to see if the counter will overflow till the next update */
        if((arg_p64bitCounter->m_UpdateElapsedTime + pUpdateData->m_ActiveTimeout)
            >= arg_p64bitCounter->m_Timeout)
        {
            /* update the counter now */
            goto label_1;
        }
    }
    return;

label_1:

    {
        ix_uint32 temp = 0;
        ix_rm_atomic_sram_swap(&temp, arg_p64bitCounter->m_pMEValue);

        arg_p64bitCounter->m_CoreValue += temp;
        arg_p64bitCounter->m_UpdateElapsedTime = 0;
    }

    return;
}


/**
 * NAME: _ix_rm_counter_64bit_scratch_update
 *
 * DESCRIPTION: This function will perform the scratch 64 bit update counter operation
 * 
 * @Param:  - INOUT arg_p64bitCounter - counter to be updated
 * @Param:  - INOUT arg_pContext - external data to be passed to the update operation
 *
 */
void _ix_rm_counter_64bit_scratch_update(
                                          _ix_counter_64bit* arg_p64bitCounter, 
                                          void* arg_pContext
                                        )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_counter_64bit_default_update";

    _ix_64bit_counter_update_context* pUpdateData = (_ix_64bit_counter_update_context*)arg_pContext;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_pContext == 0)
    {
        /* this is a forced update */
        goto label_1;
    }
    else
    {
        arg_p64bitCounter->m_UpdateElapsedTime += pUpdateData->m_ActiveTimeout;

        /* check to see if the counter will overflow till the next update */
        if((arg_p64bitCounter->m_UpdateElapsedTime + pUpdateData->m_ActiveTimeout)
            >= arg_p64bitCounter->m_Timeout)
        {
            /* update the counter now */
            goto label_1;
        }
    }
    return;

label_1:

    {
        ix_uint32 temp = 0;
        ix_rm_atomic_scratch_swap(&temp, arg_p64bitCounter->m_pMEValue);

        arg_p64bitCounter->m_CoreValue += temp;
        arg_p64bitCounter->m_UpdateElapsedTime = 0;
    }

    return;
}



/**
 * NAME: _ix_rm_counter_64bit_initialize_new_block
 *
 * DESCRIPTION: This function will initialize the next unused counter block for
 *  future use.
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_counter_64bit_initialize_new_block(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_counter_64bit_initialize_new_block";

    ix_error err = IX_SUCCESS;
    _ix_counter_64bit* pCounter;
    ix_uint32 i;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(g_64bitCounterManager.m_InitializedCounterBlocks >= IX_64BIT_COUNTER_BLOCKS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_ALL_BLOCKS_IN_USE, s_aFunctionIdentifier, 
                    "all counter blocks are in use!", err);
    }

    pCounter = ix_ossl_malloc(sizeof(_ix_counter_block));
    if(pCounter == 0)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_MEMORY_INIT_FAILED, s_aFunctionIdentifier, 
                    "initialization of 64 bit counter block memory failed!", err);
    }

    g_64bitCounterManager.m_CounterBlocks[g_64bitCounterManager.m_InitializedCounterBlocks] = pCounter;

    for(i = 0; i < IX_64BIT_COUNTER_PER_BLOCK; i++, pCounter++)
    {
        pCounter->m_CoreValue = 0;
        pCounter->m_Timeout = IX_64BIT_COUNTER_DEFAULT_TIMEOUT;
        pCounter->m_UpdateElapsedTime = 0;
        pCounter->m_UpdateFuncPtr = 0;
        pCounter->m_Properties = 0;

        pCounter->m_pMEValue = 0;
        pCounter->m_hCounter = IX_64BIT_COUNTER_MAKE_HANDLE(i, g_64bitCounterManager.m_InitializedCounterBlocks);

        /* add the new counter at the beginning of the free list */
        pCounter->m_pNext = g_64bitCounterManager.m_pCounterFreeList;
        g_64bitCounterManager.m_pCounterFreeList = pCounter;
    }

    g_64bitCounterManager.m_InitializedCounterBlocks++;

    g_64bitCounterManager.m_FreeCounters += IX_64BIT_COUNTER_PER_BLOCK;

    return err;
}

/**
 * NAME: _ix_rm_counter_64bit_thread_update_run
 *
 * DESCRIPTION: This function is the update thread function
 * 
 * @Param:  - IN arg_p64bitCounterManager - address of the manager data the function
 *  will work on. 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_counter_64bit_thread_update_run(
                                                 _ix_64bit_counter_manager* arg_p64bitCounterManager,
                                                 ...
                                               )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_counter_64bit_thread_update_run";

    ix_error err = IX_SUCCESS;
    _ix_counter_64bit* pCounter;
    _ix_64bit_counter_update_context updateContext;
    void* pContext;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &arg_p64bitCounterManager->m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */


    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    while(arg_p64bitCounterManager->m_UpdateProcessingDone == 0)
    {
        err = ix_ossl_sem_take(arg_p64bitCounterManager->m_UpdateSemaphore, arg_p64bitCounterManager->m_ActiveTO);
        if(err == IX_SUCCESS)
        {
            /* set the context to 0 to force a global update */
            pContext = 0;
        }
        else if(IX_ERROR_GET_CODE(err) == IX_OSSL_ERROR_TIMEOUT)
        {
            /* update the timeout */
            updateContext.m_ActiveTimeout = arg_p64bitCounterManager->m_ActiveTO;
            pContext = &updateContext;
        }
        else
            break;

#if defined( _IX_RM_MULTI_THREADED_ )

        pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

        /* update all counters */
        pCounter = arg_p64bitCounterManager->m_pActiveCounterList;

        while(pCounter != 0)
        {
            pCounter->m_UpdateFuncPtr(pCounter, pContext);

            pCounter = pCounter->m_pNext;
        }

#if defined( _IX_RM_MULTI_THREADED_ )

        pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    }

    arg_p64bitCounterManager->m_UpdateProcessingDone = 0;

    return err;
}


#if defined( _IX_RM_LIB_DEVELOPMENT_ )

ix_uint32 s_CounterTestUpdateDone = 0;
ix_uint32 s_CounterTestUpdateDelay = 1000;

/**
 * NAME: _ix_rm_counter_64bit_thread_test_update_run
 *
 * DESCRIPTION: This function is the test update thread function for testing purposes
 * 
 * @Param:  - IN arg_p64bitCounterManager - address of the manager data the function
 *  will work on. 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_counter_64bit_thread_test_update_run( 
                                                 _ix_64bit_counter_manager* arg_p64bitCounterManager,
                                                 ...
                                               )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_counter_64bit_thread_update_run";

    ix_error err = IX_SUCCESS;
    _ix_counter_64bit* pCounter;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &arg_p64bitCounterManager->m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */


    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    while(s_CounterTestUpdateDone == 0)
    {
        
        ix_ossl_sleep(s_CounterTestUpdateDelay);

        if(s_CounterTestUpdateDone != 0)
            break;

#if defined( _IX_RM_MULTI_THREADED_ )

        pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

        /* update all counters */
        pCounter = arg_p64bitCounterManager->m_pActiveCounterList;

        while(pCounter != 0)
        {
            (*pCounter->m_pMEValue)++;

            pCounter = pCounter->m_pNext;
        }

#if defined( _IX_RM_MULTI_THREADED_ )

        pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    }

    s_CounterTestUpdateDone = 0;

    return err;
}

#endif  /* defined( _IX_RM_LIB_DEVELOPMENT_ ) */


/**
 * NAME: _ix_rm_counter_64bit_update
 *
 * DESCRIPTION: This function will force an update of all counters
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_counter_64bit_update(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_counter_64bit_update";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    return ix_ossl_sem_give(g_64bitCounterManager.m_UpdateSemaphore);
}



/**
 * NAME: _ix_rm_counter_64bit_add_to_active_list
 *
 * DESCRIPTION: This function will add the passed counter to the active counter list
 * 
 * @Param:  - IN arg_p64bitCounter address of the counter to be added to active counter list
 *
 */
void _ix_rm_counter_64bit_add_to_active_list(
                                              _ix_counter_64bit* arg_p64bitCounter
                                            )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_counter_64bit_add_to_active_list";

    _ix_counter_64bit* p64bitCounterPrev = 0;
    _ix_counter_64bit* p64bitCounterCurrent = g_64bitCounterManager.m_pActiveCounterList;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* insert the counter into the active counter list */
    if(p64bitCounterCurrent == 0)
    {
        g_64bitCounterManager.m_pActiveCounterList = arg_p64bitCounter;
    }
    else
    {
        while(p64bitCounterCurrent != 0)
        {
            if(p64bitCounterCurrent->m_Timeout >= arg_p64bitCounter->m_Timeout)
                break;
            p64bitCounterPrev = p64bitCounterCurrent;
            p64bitCounterCurrent = p64bitCounterCurrent->m_pNext;
        }

        arg_p64bitCounter->m_pNext = p64bitCounterCurrent;
        if(p64bitCounterPrev == 0)
        {
            g_64bitCounterManager.m_pActiveCounterList = arg_p64bitCounter;
        }
        else
        {
            p64bitCounterPrev->m_pNext = arg_p64bitCounter;
        }
    }
}



/**
 * NAME: _ix_rm_counter_64bit_remove_from_active_list
 *
 * DESCRIPTION: This function will remove the passed counter from the active counter list
 * 
 * @Param:  - IN arg_p64bitCounter address of the counter to be added to active counter list
 *
 */
void _ix_rm_counter_64bit_remove_from_active_list(
                                                   _ix_counter_64bit* arg_p64bitCounter
                                                 )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_counter_64bit_add_to_active_list";

    _ix_counter_64bit* p64bitCounterPrev = 0;
    _ix_counter_64bit* p64bitCounterCurrent = g_64bitCounterManager.m_pActiveCounterList;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* remove the counter from the list of active counters */
    while(p64bitCounterCurrent != 0)
    {
        if(p64bitCounterCurrent == arg_p64bitCounter)
        {
            if(p64bitCounterPrev == 0)
            {
                g_64bitCounterManager.m_pActiveCounterList = arg_p64bitCounter->m_pNext;
            }
            else
            {
                p64bitCounterPrev->m_pNext = p64bitCounterCurrent->m_pNext;
            }
            break;
        }
        p64bitCounterPrev = p64bitCounterCurrent;
        p64bitCounterCurrent = p64bitCounterCurrent->m_pNext;
    }
}

/**
 * NAME: _ix_rm_counter_64bit_init
 *
 * DESCRIPTION: This function should be called at the time the resource manager initializer 
 *  function is called. The function will initialize all the Counter 64 bit API.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_counter_64bit_init(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_counter_64bit_init";

    ix_error err = IX_SUCCESS;
    ix_uint32 i;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    /* initialize the synchronization mechanism */
    err = ix_ossl_sem_init(IX_OSSL_SEM_AVAILABLE, &s_64bitCounterManagerSemaphore);
    if(err)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_INIT_FAILED, s_aFunctionIdentifier, 
                        "initialization of 64 bit counter manager's semaphore failed!", err);
    }

    _ix_rm_sa_sync_adapter_init(&g_64bitCounterManager.m_SyncAdapter, 
        _ix_rm_sa_default_lock, _ix_rm_sa_default_unlock, (void*)&s_64bitCounterManagerSemaphore);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    g_64bitCounterManager.m_ActiveTO = IX_OSSL_WAIT_FOREVER;
    ix_ossl_memset(g_64bitCounterManager.m_CounterBlocks, 0, sizeof(g_64bitCounterManager.m_CounterBlocks));
    g_64bitCounterManager.m_InitializedCounterBlocks = 0;
    g_64bitCounterManager.m_pActiveCounterList = 0;
    g_64bitCounterManager.m_pCounterFreeList = 0;
    g_64bitCounterManager.m_UpdateSemaphore = 0;
    g_64bitCounterManager.m_UpdateThreadId = 0;
    g_64bitCounterManager.m_FreeCounters = 0;

    /* initialize some counters */
    for(i = 0; i < IX_64BIT_COUNTER_STARTUP_INITIALIZED_BLOCKS; i++)
    {
        err = _ix_rm_counter_64bit_initialize_new_block();
        if(err != IX_SUCCESS)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_BLOCK_INIT_FAILED, s_aFunctionIdentifier, 
                        "initialization of a new 64 bit counter block failed!", err);
            goto label_1;
        }
    }

    /* setup the update mechanism */
    g_64bitCounterManager.m_UpdateProcessingDone = 0;

    /* create the update semaphore */
    err = ix_ossl_sem_init(IX_OSSL_SEM_UNAVAILABLE, &g_64bitCounterManager.m_UpdateSemaphore);
    if(err)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_UPDATE_SEM_INIT_FAILED, s_aFunctionIdentifier, 
                        "initialization of 64 bit counter update semaphore failed!", err);
        goto label_2;
    }
    /* create the update thread */
    err = ix_ossl_thread_create((ix_ossl_thread_entry_point_t)_ix_rm_counter_64bit_thread_update_run,
        &g_64bitCounterManager, "Rm64bitCnt", &g_64bitCounterManager.m_UpdateThreadId);
    if(err !=IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_UPDATE_TH_INIT_FAILED, s_aFunctionIdentifier, 
            "failed to create the update thread!", 0);
        goto label_3;
    }

    /* set the update thread priority to high */
    err = ix_ossl_thread_set_priority(g_64bitCounterManager.m_UpdateThreadId, IX_OSSL_THREAD_PRI_HIGH);
    if(err !=IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_UPDATE_TH_INIT_FAILED, s_aFunctionIdentifier, 
            "failed to set the update thread priority!", 0);
        goto label_4;
    }

#if defined( _IX_RM_LIB_DEVELOPMENT_ )

    {
        ix_ossl_thread_t tid;

        ix_ossl_thread_create((ix_ossl_thread_entry_point_t)_ix_rm_counter_64bit_thread_test_update_run,
            &g_64bitCounterManager, "Rm64bitTst", &tid);

        ix_ossl_thread_set_priority(tid, IX_OSSL_THREAD_PRI_HIGH);
    }

#endif  /* defined( _IX_RM_LIB_DEVELOPMENT_ ) */

    return err;

label_4:

    /* stop the update thread */
    g_64bitCounterManager.m_UpdateProcessingDone = 1;

    ix_ossl_sem_give(g_64bitCounterManager.m_UpdateSemaphore);

    /* wait some time for the update thread to terminate */
    ix_ossl_sleep(1);

    g_64bitCounterManager.m_UpdateThreadId = 0;

label_3:

    ix_ossl_sem_fini(g_64bitCounterManager.m_UpdateSemaphore);

    g_64bitCounterManager.m_UpdateSemaphore = 0;

label_2:

    /* free the memory associated with the counters */
    for(i = 0; i < g_64bitCounterManager.m_InitializedCounterBlocks; i++)
    {
        if(g_64bitCounterManager.m_CounterBlocks[i] != 0)
        {
            ix_ossl_free(g_64bitCounterManager.m_CounterBlocks[i]);
            g_64bitCounterManager.m_CounterBlocks[i] = 0;
        }
    }

    g_64bitCounterManager.m_InitializedCounterBlocks = 0;

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    /* finalize the synchronization mechanism */
    _ix_rm_sa_sync_adapter_init(&g_64bitCounterManager.m_SyncAdapter,0, 0, (void*)0); 

    ix_ossl_sem_fini(s_64bitCounterManagerSemaphore);

    s_64bitCounterManagerSemaphore = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: _ix_rm_counter_64bit_fini
 *
 * DESCRIPTION: This function should be called at the time the resource manager finalizer function
 *  is called. The function will provide cleanup for Counter 64 bit API.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_counter_64bit_fini(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_counter_64bit_fini";

    ix_error err = IX_SUCCESS;
    ix_uint32 i;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_64bitCounterManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_LIB_DEVELOPMENT_ )

    s_CounterTestUpdateDone = 1;
    while(s_CounterTestUpdateDone != 0)
    {
        ix_ossl_sleep(s_CounterTestUpdateDelay);
    } 

#endif  /* defined( _IX_RM_LIB_DEVELOPMENT_ ) */

    /* stop the update thread */
    g_64bitCounterManager.m_UpdateProcessingDone = 1;

    ix_ossl_sem_give(g_64bitCounterManager.m_UpdateSemaphore);

    /* wait for the update thread to terminate */
    while(g_64bitCounterManager.m_UpdateProcessingDone != 0)
    {
        ix_ossl_sleep(1);
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    g_64bitCounterManager.m_ActiveTO = IX_OSSL_WAIT_FOREVER;
    g_64bitCounterManager.m_pActiveCounterList = 0;
    g_64bitCounterManager.m_pCounterFreeList = 0;

    /* free the memory associated with the counters */
    for(i = 0; i < g_64bitCounterManager.m_InitializedCounterBlocks; i++)
    {
        if(g_64bitCounterManager.m_CounterBlocks[i] != 0)
        {
            ix_ossl_free(g_64bitCounterManager.m_CounterBlocks[i]);
        }
    }
    g_64bitCounterManager.m_InitializedCounterBlocks = 0;

    ix_ossl_memset(g_64bitCounterManager.m_CounterBlocks, 0, sizeof(g_64bitCounterManager.m_CounterBlocks));

    g_64bitCounterManager.m_UpdateThreadId = 0;

    ix_ossl_sem_fini(g_64bitCounterManager.m_UpdateSemaphore);

    g_64bitCounterManager.m_UpdateSemaphore = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* finalize the synchronization mechanism */
    _ix_rm_sa_sync_adapter_init(&g_64bitCounterManager.m_SyncAdapter,0, 0, (void*)0); 

    err = ix_ossl_sem_fini(s_64bitCounterManagerSemaphore);
    if(err != 0)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_64BIT_COUNTER_FINI_FAILED, s_aFunctionIdentifier, 
                        "finalization of 64 bit counter manager's semaphore failed!", err);
    }

    s_64bitCounterManagerSemaphore = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}
