/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      internal_hash.c
 *
 * = DESCRIPTION
 *      This file defines all the internal function prototypes and variables
 *      required by the hash API.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      11/21/2002 10:48:02 AM - creation time 
 * ============================================================================
 */

static const char s_FileIdentifier[] = "$Id: internal_hash.c,v 1.7 2002/12/21 01:54:59 ageorges Exp $";

#include "ix_ossl.h"

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"

#include "rm/symbols.h"
#include "rm/internal/internal_support.h"

#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"

#include "rm/system.h"
#include "rm/internal/internal_system.h"

#include "rm/hash.h"
#include "rm/internal/internal_hash.h"

#if defined( _IX_RM_MULTI_THREADED_ )

static ix_ossl_sem_t s_HashAPIManagerSemaphore = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

/**
 * Global Hash API manager
 */
_ix_hash_data_manager g_HashAPIManager;


/**
 * NAME: _ix_rm_hash_init
 *
 * DESCRIPTION: This function should be called at the time the resource manager initializer 
 *          function is called. The function will initialize all the Hash API.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_hash_init(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_hash_init";

    ix_error err = 0;

	IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);


#if defined( _IX_RM_MULTI_THREADED_ )

    /* initialize the synchronization semaphore */
    err = ix_ossl_sem_init(IX_OSSL_SEM_AVAILABLE, &s_HashAPIManagerSemaphore);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_HASH_API_INIT_FAILED, s_aFunctionIdentifier, 
                        "initialization of hash API manager's semaphore failed!", err);
    }

    /* initialize synchronization */
    _ix_rm_sa_sync_adapter_init(&g_HashAPIManager.m_SyncAdapter, 
        _ix_rm_sa_default_lock, _ix_rm_sa_default_unlock, (void*)&s_HashAPIManagerSemaphore);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* Reset the hash multipliers */
    {
        ix_hash_multiplier_48 hashMultiplier48 = {1,0};
        ix_hash_multiplier_64 hashMultiplier64 = {1,0};
        ix_hash_multiplier_128 hashMultiplier128 = {1,0,0,0};

        ix_rm_hash_48_multiplier_set(&hashMultiplier48);
        ix_rm_hash_64_multiplier_set(&hashMultiplier64);
        ix_rm_hash_128_multiplier_set(&hashMultiplier128);
    }

#if defined(_IX_RM_IMPL_HARDWARE_)

#else

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    return err;
}


/**
 * NAME: _ix_rm_hash_fini
 *
 * DESCRIPTION: This function should be called at the time the resource manager finalizer function
 *          is called. The function will provide cleanup for Hash API.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_hash_fini(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_hash_fini";

    ix_error err = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

#else

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* finalize synchronization */
    _ix_rm_sa_sync_adapter_init(&g_HashAPIManager.m_SyncAdapter, 0, 0, 0);

    /* finalize the memory managers' synchronization semaphore */
    err = ix_ossl_sem_fini(s_HashAPIManagerSemaphore);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_HASH_API_FINI_FAILED, s_aFunctionIdentifier, 
                        "finalization of hash API manager's semaphore failed!", err);
    }

    s_HashAPIManagerSemaphore = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}

#if !defined(_IX_RM_IMPL_HARDWARE_HASH_) 


/**
 * This function will perform multiplication of 2 16-bit integers with no carry and return 
 * the result in a 32-bit integer. So 3 * 3 will result in 5 rather than 9
 */

unsigned long _ix_rm_hash_integral_multiply_no_carry(unsigned short arg_Operand1, unsigned short arg_Operand2)
{
    int i;
    register unsigned long temp1;
    register unsigned long temp2 = 0;
    
    for(i = 0; i < 16; i++)
    {
        if(arg_Operand2 & (1 << i))
        {
            temp1 = ((unsigned long)arg_Operand1) << i;
            temp2 ^= temp1;
        }
    }
    return temp2;
}


/**
 * This function will provide fast polynomial multiplication modulo 2 where the polynomial
 * coeficients will be bits in an array of 16-bit integers. The polynomial multiplication
 * is done by splitting them in 16  bit integers.
 */

void _ix_rm_hash_polynomial_multiply_no_carry(const unsigned short* arg_pData1, int arg_Size1, const unsigned short* arg_pData2, int arg_Size2, unsigned short* arg_pResult)
{
    int i,j,k,l;
    int count = arg_Size1 + arg_Size2 - 1;
    unsigned long temp, temp1 = 0, temp2 = 0;

    for(i = 0, j = 1; i < count; i++, j++)
    {
        k = j - arg_Size2;
        if(k < 0)
            k = 0;
        l = i - k;
        for(;(k <= i) && (k < arg_Size1); k++, l--)
        {
            temp = _ix_rm_hash_integral_multiply_no_carry(arg_pData1[k] ,arg_pData2[l]);
            temp1 ^= temp & 0x0000ffff;
            temp2 ^= temp >> 16;
        }
        temp2 ^= temp1 >> 16;
        arg_pResult[i] = (unsigned short)(temp1 & 0x0000ffff);
        temp1 = temp2;
        temp2 = 0;
    }
    arg_pResult[count] = (unsigned short)temp1;
}


/**
 * This function will provide modulo 2 polynomial divison. The polynomial
 * coeficients will be bits in an array of 16-bit integers.
 */

void _ix_rm_hash_polynomial_divide_remainder_no_carry(unsigned short* arg_pData1, int arg_Size1, unsigned short* arg_pData2, int arg_Size2, unsigned short* arg_pResult)
{
    int i,j,k,l,m,p;
    unsigned short temp;

    for(j = arg_Size2 - 1; (j >= 0) && (arg_pData2[j] == 0); j--);
    for(l = 1; l < 16; l++)
        if((arg_pData2[j] >> l) == 0)
            break;

    for(i = arg_Size1 - 1; (i >= 0) && (arg_pData1[i] == 0); i--);
    for(k = 1; k < 16; k++)
        if((arg_pData1[i] >> k) == 0)
            break;

    m = (j << 4) + l;
    p = (i << 4) + k;

    while(p >= m)
    {
        p -= m;
        k = p & 0x000f;
        p >>= 4;

        temp = 0;
        for(l = 0; l <= j; l++, p++)
        {
            arg_pData1[p] ^= temp;
            arg_pData1[p] ^= arg_pData2[l] << k;
            temp = arg_pData2[l] >> (16 - k);
        }

        for(; (i >= 0) && (arg_pData1[i] == 0); i--);
        for(k = 1; k < 16; k++)
            if((arg_pData1[i] >> k) == 0)
                break;
        p = (i << 4) + k;
    }

    for(i = 0; i < arg_Size2; i++)
        arg_pResult[i] = arg_pData1[i];
}

/**
 * NAME: _ix_rm_hash_48_software_hash
 *
 * DESCRIPTION: This function will perform a software 48 bit hash operation
 * 
 * @Param:  - INOUT - arg_pHash48 on input holds the data to be hashed and on
 *      output contains the hashed value.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_hash_48_software_hash( 
                                       ix_hash_48* arg_pHash48
                                     )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_hash_48_software_hash";

    ix_error err = IX_SUCCESS;

    /* the generator polynomial is 1 + x^10 + x^25 + x^36 + x^48 */
    static unsigned short s_48HashGenerator[4] ={1025, 512, 16, 1};

    register unsigned long temp;

    static unsigned short data1[3];
    static unsigned short data2[3];
    static unsigned short data3[6];

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);
    
    temp = arg_pHash48->m_LW0;
    data1[0] = (unsigned short)(temp & 0x0000ffff);
    data1[1] = (unsigned short)(temp >> 16);

    temp = arg_pHash48->m_LW1;
    data1[2] = (unsigned short)(temp & 0x0000ffff);

    temp = g_HashAPIManager.m_HashMultiplier48.m_LW0;
    data2[0] = (unsigned short)(temp & 0x0000ffff);
    data2[1] = (unsigned short)(temp >> 16);

    temp = g_HashAPIManager.m_HashMultiplier48.m_LW1;
    data2[2] = (unsigned short)(temp & 0x0000ffff);

    _ix_rm_hash_polynomial_multiply_no_carry(data1, 3,data2, 3, data3);

    _ix_rm_hash_polynomial_divide_remainder_no_carry(data3, 6, s_48HashGenerator, 4, data1);

    arg_pHash48->m_LW0 = ((unsigned long)data1[1] << 16) + data1[0];
    arg_pHash48->m_LW1 = (unsigned long)data1[2];

    return err;
}


/**
 * NAME: _ix_rm_hash_64_software_hash
 *
 * DESCRIPTION: This function will perform a software 64 bit hash operation
 * 
 * @Param:  - INOUT - arg_pHash64 on input holds the data to be hashed and on
 *      output contains the hashed value.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_hash_64_software_hash( 
                                       ix_hash_64* arg_pHash64
                                     )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_hash_64_software_hash";

    ix_error err = IX_SUCCESS;

    /* the generator polynomial is 1 + x^17 + x^35 + x^54 + x^64 */
    static unsigned short s_64HashGenerator[5] ={1, 2, 8, 64, 1};

    register unsigned long temp;

    static unsigned short data1[4];
    static unsigned short data2[4];
    static unsigned short data3[8];

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);
    
    temp = arg_pHash64->m_LW0;
    data1[0] = (unsigned short)(temp & 0x0000ffff);
    data1[1] = (unsigned short)(temp >> 16);

    temp = arg_pHash64->m_LW1;
    data1[2] = (unsigned short)(temp & 0x0000ffff);
    data1[3] = (unsigned short)(temp >> 16);

    temp = g_HashAPIManager.m_HashMultiplier64.m_LW0;
    data2[0] = (unsigned short)(temp & 0x0000ffff);
    data2[1] = (unsigned short)(temp >> 16);

    temp = g_HashAPIManager.m_HashMultiplier64.m_LW1;
    data2[2] = (unsigned short)(temp & 0x0000ffff);
    data2[3] = (unsigned short)(temp >> 16);

    _ix_rm_hash_polynomial_multiply_no_carry(data1, 4, data2, 4, data3);

    _ix_rm_hash_polynomial_divide_remainder_no_carry(data3, 8, s_64HashGenerator, 5, (unsigned short*)data1);

    arg_pHash64->m_LW0 = ((unsigned long)data1[1] << 16) + data1[0];
    arg_pHash64->m_LW1 = ((unsigned long)data1[3] << 16) + data1[2];

    return err;
}


/**
 * NAME: _ix_rm_hash_128_software_hash
 *
 * DESCRIPTION: This function will perform a software 128 bit hash operation
 * 
 * @Param:  - INOUT - arg_pHash128 on input holds the data to be hashed and on
 *      output contains the hashed value.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_hash_128_software_hash( 
                                        ix_hash_128* arg_pHash128
                                      )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_hash_128_software_hash";

    ix_error err = IX_SUCCESS;

    /* the generator polynomial is 1 + x^33 + x^69 + x^98 + x^128 */
    static unsigned short s_128HashGenerator[9] ={1, 0, 2, 0, 32, 0, 4, 0, 1};

    register unsigned long temp;

    static unsigned short data1[8];
    static unsigned short data2[8];
    static unsigned short data3[16];

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);
    
    temp = arg_pHash128->m_LW0;
    data1[0] = (unsigned short)(temp & 0x0000ffff);
    data1[1] = (unsigned short)(temp >> 16);

    temp = arg_pHash128->m_LW1;
    data1[2] = (unsigned short)(temp & 0x0000ffff);
    data1[3] = (unsigned short)(temp >> 16);

    temp = arg_pHash128->m_LW2;
    data1[4] = (unsigned short)(temp & 0x0000ffff);
    data1[5] = (unsigned short)(temp >> 16);

    temp = arg_pHash128->m_LW3;
    data1[6] = (unsigned short)(temp & 0x0000ffff);
    data1[7] = (unsigned short)(temp >> 16);

    temp = g_HashAPIManager.m_HashMultiplier128.m_LW0;
    data2[0] = (unsigned short)(temp & 0x0000ffff);
    data2[1] = (unsigned short)(temp >> 16);

    temp = g_HashAPIManager.m_HashMultiplier128.m_LW1;
    data2[2] = (unsigned short)(temp & 0x0000ffff);
    data2[3] = (unsigned short)(temp >> 16);

    temp = g_HashAPIManager.m_HashMultiplier128.m_LW2;
    data2[4] = (unsigned short)(temp & 0x0000ffff);
    data2[5] = (unsigned short)(temp >> 16);

    temp = g_HashAPIManager.m_HashMultiplier128.m_LW3;
    data2[6] = (unsigned short)(temp & 0x0000ffff);
    data2[7] = (unsigned short)(temp >> 16);

    _ix_rm_hash_polynomial_multiply_no_carry( data1, 8, data2, 8, data3);

    _ix_rm_hash_polynomial_divide_remainder_no_carry(data3, 16, s_128HashGenerator, 9, (unsigned short*)data1);

    arg_pHash128->m_LW0 = ((unsigned long)data1[1] << 16) + data1[0];
    arg_pHash128->m_LW1 = ((unsigned long)data1[3] << 16) + data1[2];
    arg_pHash128->m_LW2 = ((unsigned long)data1[5] << 16) + data1[4];
    arg_pHash128->m_LW3 = ((unsigned long)data1[7] << 16) + data1[6];

    return err;
}


#endif /* end !defined(_IX_RM_IMPL_HARDWARE_HASH_) */
