/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      internal_me_xscale_locks.c
 *
 * = DESCRIPTION
 *      This file will define all variables and functions needed internally
 *      by the ME\XScale locks API.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      8/8/2003 1:32:18 PM - creation time 
 * ============================================================================
 */

static const char s_FileIdentifier[] = "$Id: internal_me_xscale_locks.c,v 1.3 2003/10/10 21:08:28 ageorges Exp $";

#include "ix_ossl.h"

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"

#include "rm/symbols.h"
#include "rm/internal/internal_support.h"

#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"

#include "rm/system.h"
#include "rm/internal/internal_system.h"

#include "rm/memory.h"

#include "rm/me_xscale_locks.h"
#include "rm/internal/internal_me_xscale_locks.h"

#include "rm/atomic_operations.h"

/**
 * Global ME\XScale lock manager
 */
_ix_me_xscale_lock_manager g_MeXScaleLockManager;

#if defined( _IX_RM_MULTI_THREADED_ )

/**
 * Synchronization semaphore for ME\XScale lock manager
 */
static ix_ossl_sem_t s_MeXScaleLockManagerSemaphore;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */


/**
 * NAME: _ix_rm_me_xscale_lock_init
 *
 * DESCRIPTION: This function should be called at the time the resource manager initializer 
 *          function is called. The function will initialize all the ME\XScale lock API.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_me_xscale_lock_init(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_me_xscale_lock_init";

    ix_error err = IX_SUCCESS;
    ix_uint32 i;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    /* initialize the synchronization mechanism */
    err = ix_ossl_sem_init(IX_OSSL_SEM_AVAILABLE, &s_MeXScaleLockManagerSemaphore);
    if(err)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_ME_XSCALE_LOCK_INIT_FAILED, s_aFunctionIdentifier, 
                        "initialization of ME\\XScale lock manager's semaphore failed!", err);
    }

    _ix_rm_sa_sync_adapter_init(&g_MeXScaleLockManager.m_SyncAdapter, 
        _ix_rm_sa_default_lock, _ix_rm_sa_default_unlock, (void*)&s_MeXScaleLockManagerSemaphore);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* Initialize the manager */
    g_MeXScaleLockManager.m_pLocksFreeList = 0;
    g_MeXScaleLockManager.m_pLocksScratchArea = 0;

    /* Find how much SCRATCH we need to allocate expressed in LW */
    i = (IX_ME_XSCALE_LOCK_MAX_COUNT + 31) / 32;
    
    /* Allocate the required SCRATCH memory */
    err = ix_rm_mem_alloc(IX_MEMORY_TYPE_SCRATCH, 0, i*4, (void**)&g_MeXScaleLockManager.m_pLocksScratchArea);
    if(err !=IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_XSCALE_LOCK_MEM_ALLOC_FAILED, s_aFunctionIdentifier, 
            "failed to allocate the scratch locks memory!", 0);
        goto label_1;
    }

    /* Initialize all the locks */
    for(i = 0; i < IX_ME_XSCALE_LOCK_MAX_COUNT; i++)
    {
        _ix_me_xscale_lock* pLock = &g_MeXScaleLockManager.m_MeXScaleLocks[i];
        pLock->m_Owner = IX_ME_XSCALE_LOCK_OWNER_NONE;
        pLock->m_pScratchMemory = g_MeXScaleLockManager.m_pLocksScratchArea + (i / 32);
        pLock->m_BitPosition = i % 32;
        pLock->m_hMeXScaleLock = IX_ME_XSCALE_LOCK_MAKE_HANDLE(i);
        pLock->m_pNext = g_MeXScaleLockManager.m_pLocksFreeList;
        g_MeXScaleLockManager.m_pLocksFreeList = pLock;
    }

    /* make all the locks unavailable */
    for(i = 0; i < ((IX_ME_XSCALE_LOCK_MAX_COUNT + 31) / 32); i++)
    {
        g_MeXScaleLockManager.m_pLocksScratchArea[i] = 0;
    }

    return err;

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    /* finalize the synchronization mechanism */
    _ix_rm_sa_sync_adapter_init(&g_MeXScaleLockManager.m_SyncAdapter,0, 0, (void*)0); 

    ix_ossl_sem_fini(s_MeXScaleLockManagerSemaphore);

    s_MeXScaleLockManagerSemaphore = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: _ix_rm_me_xscale_lock_fini
 *
 * DESCRIPTION: This function should be called at the time the resource manager finalizer function
 *          is called. The function will provide cleanup for ME\XScale lock API.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_me_xscale_lock_fini(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_me_xscale_lock_fini";

    ix_error err = IX_SUCCESS;
    ix_uint32 i;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_MeXScaleLockManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);


#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* finalize all the locks */
    for(i = 0; i < IX_ME_XSCALE_LOCK_MAX_COUNT; i++)
    {
        _ix_me_xscale_lock* pLock = &g_MeXScaleLockManager.m_MeXScaleLocks[i];

        pLock->m_Owner = IX_ME_XSCALE_LOCK_OWNER_NONE;
        pLock->m_pScratchMemory = 0;
        pLock->m_BitPosition = 0;
        pLock->m_hMeXScaleLock = 0;
        pLock->m_pNext = 0;
    }

    /* make all the locks unavailable */
    for(i = 0; i < ((IX_ME_XSCALE_LOCK_MAX_COUNT + 31) / 32); i++)
    {
        g_MeXScaleLockManager.m_pLocksScratchArea[i] = 0;
    }

    /* finalize the manager */
    if(g_MeXScaleLockManager.m_pLocksScratchArea != 0)
    {
        ix_rm_mem_free(g_MeXScaleLockManager.m_pLocksScratchArea);
        g_MeXScaleLockManager.m_pLocksScratchArea = 0;
    }

    g_MeXScaleLockManager.m_pLocksFreeList = 0;


#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* finalize the synchronization mechanism */
    _ix_rm_sa_sync_adapter_init(&g_MeXScaleLockManager.m_SyncAdapter,0, 0, (void*)0); 

    err = ix_ossl_sem_fini(s_MeXScaleLockManagerSemaphore);
    if(err != 0)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_ME_XSCALE_LOCK_FINI_FAILED, s_aFunctionIdentifier, 
                        "finalization of ME\\XScale lock manager's semaphore failed!", err);
    }

    s_MeXScaleLockManagerSemaphore = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}

