/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      internal_memory.c
 *
 * = DESCRIPTION
 *      This file will define all the internal supporting functions for the Memory Manager API.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      4/30/2002 10:41:08 AM - creation time 
 * ============================================================================
 */

static const char s_aFileIdentifier[] = "$Id: internal_memory.c,v 1.59 2003/12/13 01:12:57 mrcastel Exp $";


#include "ix_ossl.h"

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"

#include "rm/symbols.h"
#include "rm/internal/internal_support.h"

#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"
#include "rm/system.h"
#include "rm/internal/internal_system.h"

#include "rm/memory.h"
#include "rm/internal/internal_memory.h"

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

    #include <vmLib.h>

    #if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

        #include <vxworks/ixp2400/ixdp2400.h>

    #else

        #include <vxworks/ixp2800/ixdp2000.h>

    #endif

#endif /* end (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) */

#include <halMmap.h>
#include <halMev2Api.h>
#include <uclo.h>

#else

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

#include "rm/memmap.h"

#endif /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */



#if defined(_IX_RM_IMPL_HARDWARE_)

#else

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * Define memory buffers for DRAM memory channels
 */
#define IX_TESTING_DRAM_CHANNEL_MEM_SIZE        16*1024*1024
static unsigned char s_aDRAMMemoryBuffers[IX_DRAM_CHANNELS_NUMBER][IX_TESTING_DRAM_CHANNEL_MEM_SIZE];

/**
 * Define memory buffers for SRAM memory channels
 */
#define IX_TESTING_SRAM_CHANNEL_MEM_SIZE        8*1024*1024
static unsigned char s_aSRAMMemoryBuffers[IX_SRAM_CHANNELS_NUMBER][IX_TESTING_SRAM_CHANNEL_MEM_SIZE];

/**
 * Define memory buffers for SCRATCH memory channels
 */
#define IX_TESTING_SCRATCH_CHANNEL_MEM_SIZE     4096*sizeof(ix_uint32)
static unsigned char s_aSCRATCHMemoryBuffers[IX_SCRATCH_CHANNELS_NUMBER][IX_TESTING_SCRATCH_CHANNEL_MEM_SIZE];

/**
 * Define memory buffers for LOCAL memory for each microengine
 */
#define IX_TESTING_LOCAL_MEM_SIZE       1024
static unsigned char s_aLOCALMemoryBuffers[IX_MAX_MICROENGINE_NUMBER][IX_TESTING_LOCAL_MEM_SIZE];

#else /* Westport hardware */

/**
 * Define memory buffers for South DRAM memory channels
 */
#define IX_TESTING_DRAM_SOUTH_CHANNEL_MEM_SIZE        8*1024*1024
static unsigned char s_aSouthDRAMMemoryBuffers[IX_DRAM_SOUTH_CHANNELS_NUMBER][IX_TESTING_DRAM_SOUTH_CHANNEL_MEM_SIZE];

/**
 * Define memory buffers for South MSG-SRAM memory channels
 */
#define IX_TESTING_SRAM_MSG_SOUTH_CHANNEL_MEM_SIZE        4*1024*1024
static unsigned char s_aSouthSRAM_MSGMemoryBuffers[IX_SRAM_MSG_SOUTH_CHANNELS_NUMBER][IX_TESTING_SRAM_MSG_SOUTH_CHANNEL_MEM_SIZE];

/**
 * Define memory buffers for North DRAM memory channels
 */
#define IX_TESTING_DRAM_NORTH_CHANNEL_MEM_SIZE        2*1024*1024
static unsigned char s_aNorthDRAMMemoryBuffers[IX_DRAM_NORTH_CHANNELS_NUMBER][IX_TESTING_DRAM_NORTH_CHANNEL_MEM_SIZE];

/**
 * Define memory buffers for North SRAM memory channels
 */
#define IX_TESTING_SRAM_NORTH_CHANNEL_MEM_SIZE        1*1024*1024
static unsigned char s_aNorthSRAMMemoryBuffers[IX_SRAM_NORTH_CHANNELS_NUMBER][IX_TESTING_SRAM_NORTH_CHANNEL_MEM_SIZE];

/**
 * Define memory buffers for North SCRATCH memory channels
 */
#define IX_TESTING_SCRATCH_CHANNEL_MEM_SIZE     1*1024*1024
static unsigned char s_aSCRATCHMemoryBuffers[IX_SCRATCH_CHANNELS_NUMBER][IX_TESTING_SCRATCH_CHANNEL_MEM_SIZE];

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */


#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * This is the VMT for all DRAM memory managers.
 */
static const _ix_mm_vmt s_DRAMMemoryManagerVMT = 
    {
        _ix_rm_mm_initializer,
        _ix_rm_mm_finalizer,
        _ix_rm_mm_allocator,
        _ix_rm_mm_aligned_allocator,
        _ix_rm_mm_deallocator,
        _ix_rm_mm_reserve,
        _ix_rm_mm_aligned_reserve,
        _ix_rm_mm_read_dram,
        _ix_rm_mm_write_dram
    };


/**
 * This is the VMT for all SRAM memory managers.
 */
static const _ix_mm_vmt s_SRAMMemoryManagerVMT = 
    {
        _ix_rm_mm_initializer,
        _ix_rm_mm_finalizer,
        _ix_rm_mm_allocator,
        _ix_rm_mm_aligned_allocator,
        _ix_rm_mm_deallocator,
        _ix_rm_mm_reserve,
        _ix_rm_mm_aligned_reserve,
        _ix_rm_mm_read_sram,
        _ix_rm_mm_write_sram
    };


/**
 * This is the VMT for all SCRATCH memory managers.
 */
static const _ix_mm_vmt s_SCRATCHMemoryManagerVMT =
    {
        _ix_rm_mm_initializer,
        _ix_rm_mm_finalizer,
        _ix_rm_mm_allocator,
        _ix_rm_mm_aligned_allocator,
        _ix_rm_mm_deallocator,
        _ix_rm_mm_reserve,
        _ix_rm_mm_aligned_reserve,
        _ix_rm_mm_read_scratch,
        _ix_rm_mm_write_scratch
    };


/**
 * This is the VMT for all LOCAL memory managers.
 */
static const _ix_mm_vmt s_LOCALMemoryManagerVMT = 
    {
        _ix_rm_mm_initializer,
        _ix_rm_mm_finalizer,
        _ix_rm_mm_allocator,
        _ix_rm_mm_aligned_allocator,
        _ix_rm_mm_deallocator,
        _ix_rm_mm_reserve,
        _ix_rm_mm_aligned_reserve,
        0,
        0
    };

/**
 * This is the array of DRAM memory managers
 */
static _ix_mm_memory_manager s_aDRAMMemoryManagers[IX_DRAM_CHANNELS_NUMBER];

/**
 * This is the array of SRAM memory managers
 */
static _ix_mm_memory_manager s_aSRAMMemoryManagers[IX_SRAM_CHANNELS_NUMBER];

/**
 * This is the array of SCRATCH memory managers
 */
static _ix_mm_memory_manager s_aSCRATCHMemoryManagers[IX_SCRATCH_CHANNELS_NUMBER];

/**
 * This is the array of LOCAL memory managers
 */
static _ix_mm_memory_manager s_aLOCALMemoryManagers[IX_MAX_MICROENGINE_NUMBER];


/**
 * Memory managers groups
 */
const _ix_mm_memory_manager_group g_aMemoryManagersGroups[IX_MEMORY_TYPE_LAST] =
    {
        {IX_DRAM_CHANNELS_NUMBER, s_aDRAMMemoryManagers},
        {IX_SRAM_CHANNELS_NUMBER, s_aSRAMMemoryManagers},
        {IX_SCRATCH_CHANNELS_NUMBER, s_aSCRATCHMemoryManagers},
        {IX_MAX_MICROENGINE_NUMBER, s_aLOCALMemoryManagers}
    };

#else /* Westport hardware */

/**
 * This is the VMT for all South DRAM memory managers.
 */
static const _ix_mm_vmt s_SouthDRAMMemoryManagerVMT = 
    {
        _ix_rm_mm_initializer,
        _ix_rm_mm_finalizer,
        _ix_rm_mm_allocator,
        _ix_rm_mm_aligned_allocator,
        _ix_rm_mm_deallocator,
        _ix_rm_mm_reserve,
        _ix_rm_mm_aligned_reserve,
        _ix_rm_mm_read_south,
        _ix_rm_mm_write_south
    };


/**
 * This is the VMT for all South MSG-SRAM memory managers.
 */
static const _ix_mm_vmt s_SouthSRAM_MSGMemoryManagerVMT = 
    {
        _ix_rm_mm_initializer,
        _ix_rm_mm_finalizer,
        _ix_rm_mm_allocator,
        _ix_rm_mm_aligned_allocator,
        _ix_rm_mm_deallocator,
        _ix_rm_mm_reserve,
        _ix_rm_mm_aligned_reserve,
        _ix_rm_mm_read_south,
        _ix_rm_mm_write_south
    };

/**
 * This is the VMT for all North DRAM memory managers.
 */
static const _ix_mm_vmt s_NorthDRAMMemoryManagerVMT = 
    {
        _ix_rm_mm_initializer,
        _ix_rm_mm_finalizer,
        _ix_rm_mm_allocator,
        _ix_rm_mm_aligned_allocator,
        _ix_rm_mm_deallocator,
        _ix_rm_mm_reserve,
        _ix_rm_mm_aligned_reserve,
        _ix_rm_mm_read_north,
        _ix_rm_mm_write_north
    };


/**
 * This is the VMT for all North SRAM memory managers.
 */
static const _ix_mm_vmt s_NorthSRAMMemoryManagerVMT = 
    {
        _ix_rm_mm_initializer,
        _ix_rm_mm_finalizer,
        _ix_rm_mm_allocator,
        _ix_rm_mm_aligned_allocator,
        _ix_rm_mm_deallocator,
        _ix_rm_mm_reserve,
        _ix_rm_mm_aligned_reserve,
        _ix_rm_mm_read_north,
        _ix_rm_mm_write_north
    };


/**
 * This is the VMT for all North SCRATCH memory managers.
 */
static const _ix_mm_vmt s_SCRATCHMemoryManagerVMT =
    {
        _ix_rm_mm_initializer,
        _ix_rm_mm_finalizer,
        _ix_rm_mm_allocator,
        _ix_rm_mm_aligned_allocator,
        _ix_rm_mm_deallocator,
        _ix_rm_mm_reserve,
        _ix_rm_mm_aligned_reserve,
        _ix_rm_mm_read_north,
        _ix_rm_mm_write_north
    };


/**
 * This is the array of South DRAM memory managers
 */
static _ix_mm_memory_manager s_aSouthDRAMMemoryManagers[IX_DRAM_SOUTH_CHANNELS_NUMBER];

/**
 * This is the array of South MSG-SRAM memory managers
 */
static _ix_mm_memory_manager s_aSouthSRAM_MSGMemoryManagers[IX_SRAM_MSG_SOUTH_CHANNELS_NUMBER];

/**
 * This is the array of North DRAM memory managers
 */
static _ix_mm_memory_manager s_aNorthDRAMMemoryManagers[IX_DRAM_NORTH_CHANNELS_NUMBER];

/**
 * This is the array of North SRAM memory managers
 */
static _ix_mm_memory_manager s_aNorthSRAMMemoryManagers[IX_SRAM_NORTH_CHANNELS_NUMBER];

/**
 * This is the array of North SCRATCH memory managers
 */
static _ix_mm_memory_manager s_aSCRATCHMemoryManagers[IX_SCRATCH_CHANNELS_NUMBER];


/**
 * Memory managers groups
 */
const _ix_mm_memory_manager_group g_aMemoryManagersGroups[IX_MEMORY_TYPE_LAST] =
    {
        {IX_DRAM_SOUTH_CHANNELS_NUMBER, s_aSouthDRAMMemoryManagers},
        {IX_SRAM_MSG_SOUTH_CHANNELS_NUMBER, s_aSouthSRAM_MSGMemoryManagers},
        {IX_DRAM_NORTH_CHANNELS_NUMBER, s_aNorthDRAMMemoryManagers},
        {IX_SRAM_NORTH_CHANNELS_NUMBER, s_aNorthSRAMMemoryManagers},
        {IX_SCRATCH_CHANNELS_NUMBER, s_aSCRATCHMemoryManagers},
    };

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */


#if defined( _IX_RM_TEST_MACROS_ )

/**
 * This static function is used just for the purpose of testing macros
 * defined in the in this API module.
 */
static void _ix_rm_test_macros(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_test_macros";

    ix_uint8    value8Bit1 = 32;
    ix_uint8    value8Bit2;

    ix_uint16   value16Bit1 = 320;
    ix_uint16   value16Bit2;

    ix_uint32   value32Bit1 = 3200;
    ix_uint32   value32Bit2;

    ix_uint64   value64Bit1 = 320000;
    ix_uint64   value64Bit2;

    (void)s_aFunctionIdentifier;

    /* test read/write macros */
    value8Bit2 = IX_RM_MEM_UINT8_READ(&value8Bit1);
    value8Bit2++;
    value8Bit2 = IX_RM_MEM_UINT8_WRITE(&value8Bit1, value8Bit2);

    value16Bit2 = IX_RM_MEM_UINT16_READ(&value16Bit1);
    value16Bit2++;
    value16Bit2 = IX_RM_MEM_UINT16_WRITE(&value16Bit1, value16Bit2);

    value32Bit2 = IX_RM_MEM_UINT32_READ(&value32Bit1);
    value32Bit2++;
    value32Bit2 = IX_RM_MEM_UINT32_WRITE(&value32Bit1, value32Bit2);

    value64Bit2 = IX_RM_MEM_UINT64_READ(&value64Bit1);
    value64Bit2++;
    value64Bit2 = IX_RM_MEM_UINT64_WRITE(&value64Bit1, value64Bit2);   
}

#endif /* end defined( _IX_RM_TEST_MACROS_ ) */


#if defined( _IX_RM_MULTI_THREADED_ )

/**
 * Synchronization semaphore for memory API
 */
static ix_ossl_sem_t s_MemoryManagerSemaphore = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */


#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * NAME: _ix_rm_memory_init
 *
 * DESCRIPTION: This function should be called at the time the resource manager initializer 
 *          function is called. The function will initialize all memory managers based on 
 *          the hardware configuration.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_memory_init(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_memory_init";

    ix_error err = IX_SUCCESS;

    register _ix_mm_memory_manager* pMemoryManager;
    register ix_uint32   channelIndex;
    register ix_uint32   memType;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

    
    /**
     * Call this function first, which in turn calls halMe_Init() and does
     * the memory mapping initialization.
     */
    UcLo_InitLib();

    {
        ix_uint32 i;
        unsigned int sys;       /* Size of memory mapped for the OS */
        unsigned int sdram;
        unsigned int sram0;
        unsigned int sram1;
        unsigned int sram2;
        unsigned int sram3;
        Hal_SysMemInfo_T sysMemInfo;

        halMe_SysMemVirtSize(&sys, &sdram, &sram0, &sram1, &sram2, &sram3);
        halMe_GetSysMemInfo(&sysMemInfo);

        /*
	    ix_ossl_message_log("Memory sizes: \n\tDRAM: 0x%x\n\tSRAM_0: 0x%x\n\tSRAM_1: 0x%x\n\tSRAM_2: 0x%x\n\tSRAM_3: 0x%x\n",
		sdram, sram0, sram1, sram2, sram3);
        */

        /* Initialize the managed DRAM memory */
        s_aDRAMMemoryManagers[0].m_MemoryType = IX_MEMORY_TYPE_DRAM;
        s_aDRAMMemoryManagers[0].m_MemoryChannel = 0;
        s_aDRAMMemoryManagers[0].m_DefaultAlignmentShift = 3; /* 8 byte alignment */
        s_aDRAMMemoryManagers[0].m_pStartAddress = (ix_uint32*)((ix_uint32)Hal_dram_ch0_virtAddr + sysMemInfo.dramDesc[0].meDramOffset );
        s_aDRAMMemoryManagers[0].m_MemoryBlockSize = sdram;
        s_aDRAMMemoryManagers[0].m_pChannelPhysicalAddress = (ix_uint32*)(ix_uint32)halMe_GetPhyXaddr(Hal_dram_ch0_virtAddr  + sysMemInfo.dramDesc[0].meDramOffset, 0);
        s_aDRAMMemoryManagers[0].m_ChannelPhysicalOffset = (ix_uint32)s_aDRAMMemoryManagers[0].m_pChannelPhysicalAddress - IX_DRAM_BASE_ADDR_PHYSICAL_CHANNEL_0;
        s_aDRAMMemoryManagers[0].m_pVMT = &s_DRAMMemoryManagerVMT;

        /* Initialize the SRAM memory channel 0 */
        s_aSRAMMemoryManagers[0].m_MemoryType = IX_MEMORY_TYPE_SRAM;
        s_aSRAMMemoryManagers[0].m_MemoryChannel = 0;
        s_aSRAMMemoryManagers[0].m_DefaultAlignmentShift = 2; /* 4 byte alignment */
        s_aSRAMMemoryManagers[0].m_pStartAddress = (ix_uint32*)((ix_uint32)Hal_sram_ch0_virtAddr + sysMemInfo.sramChan[0].sramOffset);
        s_aSRAMMemoryManagers[0].m_MemoryBlockSize = sram0;
        s_aSRAMMemoryManagers[0].m_pChannelPhysicalAddress = (ix_uint32*)(ix_uint32)halMe_GetPhyXaddr(Hal_sram_ch0_virtAddr + sysMemInfo.sramChan[0].sramOffset, 0);
        s_aSRAMMemoryManagers[0].m_ChannelPhysicalOffset = (ix_uint32)s_aSRAMMemoryManagers[0].m_pChannelPhysicalAddress - IX_SRAM_BASE_ADDR_PHYSICAL_CHANNEL_0;
        s_aSRAMMemoryManagers[0].m_pVMT = &s_SRAMMemoryManagerVMT;

        /* Initialize the SRAM memory channel 1 */
        s_aSRAMMemoryManagers[1].m_MemoryType = IX_MEMORY_TYPE_SRAM;
        s_aSRAMMemoryManagers[1].m_MemoryChannel = 1;
        s_aSRAMMemoryManagers[1].m_DefaultAlignmentShift = 2; /* 4 byte alignment */
        s_aSRAMMemoryManagers[1].m_pStartAddress = (ix_uint32*)((ix_uint32)Hal_sram_ch1_virtAddr + sysMemInfo.sramChan[1].sramOffset);
        s_aSRAMMemoryManagers[1].m_MemoryBlockSize = sram1;
        s_aSRAMMemoryManagers[1].m_pChannelPhysicalAddress = (ix_uint32*)(ix_uint32)halMe_GetPhyXaddr(Hal_sram_ch1_virtAddr + sysMemInfo.sramChan[1].sramOffset, 0);
        s_aSRAMMemoryManagers[1].m_ChannelPhysicalOffset = (ix_uint32)s_aSRAMMemoryManagers[1].m_pChannelPhysicalAddress - IX_SRAM_BASE_ADDR_PHYSICAL_CHANNEL_1;
        s_aSRAMMemoryManagers[1].m_pVMT = &s_SRAMMemoryManagerVMT;


#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

        /* Initialize the SRAM memory channel 2 */
        s_aSRAMMemoryManagers[2].m_MemoryType = IX_MEMORY_TYPE_SRAM;
        s_aSRAMMemoryManagers[2].m_MemoryChannel = 2;
        s_aSRAMMemoryManagers[2].m_DefaultAlignmentShift = 2; /* 4 byte alignment */
        s_aSRAMMemoryManagers[2].m_pStartAddress = (ix_uint32*)((ix_uint32)Hal_sram_ch2_virtAddr + sysMemInfo.sramChan[2].sramOffset);
        s_aSRAMMemoryManagers[2].m_MemoryBlockSize = sram2;
        s_aSRAMMemoryManagers[2].m_pChannelPhysicalAddress = (ix_uint32*)(ix_uint32)halMe_GetPhyXaddr(Hal_sram_ch2_virtAddr + sysMemInfo.sramChan[2].sramOffset, 0);
        s_aSRAMMemoryManagers[2].m_ChannelPhysicalOffset = (ix_uint32)s_aSRAMMemoryManagers[2].m_pChannelPhysicalAddress - IX_SRAM_BASE_ADDR_PHYSICAL_CHANNEL_2;
        s_aSRAMMemoryManagers[2].m_pVMT = &s_SRAMMemoryManagerVMT;

        /* Initialize the SRAM memory channel 3 */
        s_aSRAMMemoryManagers[3].m_MemoryType = IX_MEMORY_TYPE_SRAM;
        s_aSRAMMemoryManagers[3].m_MemoryChannel = 3;
        s_aSRAMMemoryManagers[3].m_DefaultAlignmentShift = 2; /* 4 byte alignment */
        s_aSRAMMemoryManagers[3].m_pStartAddress = (ix_uint32*)((ix_uint32)Hal_sram_ch3_virtAddr + sysMemInfo.sramChan[3].sramOffset);
        s_aSRAMMemoryManagers[3].m_MemoryBlockSize = sram3;
        s_aSRAMMemoryManagers[3].m_pChannelPhysicalAddress = (ix_uint32*)(ix_uint32)halMe_GetPhyXaddr(Hal_sram_ch3_virtAddr + sysMemInfo.sramChan[3].sramOffset, 0);
        s_aSRAMMemoryManagers[3].m_ChannelPhysicalOffset = (ix_uint32)s_aSRAMMemoryManagers[3].m_pChannelPhysicalAddress - IX_SRAM_BASE_ADDR_PHYSICAL_CHANNEL_3;
        s_aSRAMMemoryManagers[3].m_pVMT = &s_SRAMMemoryManagerVMT;

#endif /* end #if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */

        /* Initialize the SCRATCH memory */
        s_aSCRATCHMemoryManagers[0].m_MemoryType = IX_MEMORY_TYPE_SCRATCH;
        s_aSCRATCHMemoryManagers[0].m_MemoryChannel = 0;
        s_aSCRATCHMemoryManagers[0].m_DefaultAlignmentShift = 2; /* 4 byte alignment */
        s_aSCRATCHMemoryManagers[0].m_pStartAddress = (ix_uint32*)(ix_uint32)Hal_scratch_rd_wr_swap_virtAddr;
        s_aSCRATCHMemoryManagers[0].m_MemoryBlockSize = IX_SCRATCH_MEM_SIZE;
        s_aSCRATCHMemoryManagers[0].m_pChannelPhysicalAddress = (ix_uint32*)(ix_uint32)halMe_GetPhyXaddr(Hal_scratch_rd_wr_swap_virtAddr, 0);
        s_aSCRATCHMemoryManagers[0].m_ChannelPhysicalOffset = (ix_uint32)s_aSCRATCHMemoryManagers[0].m_pChannelPhysicalAddress - IX_SCRATCH_BASE_PHYSICAL_ADDR;
        s_aSCRATCHMemoryManagers[0].m_pVMT = &s_SCRATCHMemoryManagerVMT;

        /* LOCAL memory initialization */
        for(i = 0; i < IX_MAX_MICROENGINE_NUMBER; i++)
        {
            s_aLOCALMemoryManagers[i].m_MemoryType = IX_MEMORY_TYPE_LOCAL;
            s_aLOCALMemoryManagers[i].m_MemoryChannel = i;
            s_aLOCALMemoryManagers[i].m_DefaultAlignmentShift = 2; /* 4 byte alignment */
            s_aLOCALMemoryManagers[i].m_pStartAddress = 0;
            s_aLOCALMemoryManagers[i].m_MemoryBlockSize = IX_LOCAL_MEM_SIZE;
            s_aLOCALMemoryManagers[i].m_pChannelPhysicalAddress = 0;
            s_aLOCALMemoryManagers[i].m_ChannelPhysicalOffset = 0;
            s_aLOCALMemoryManagers[i].m_pVMT = &s_LOCALMemoryManagerVMT;
        }

    }

#else

    {
        size_t i;

        /* DRAM memory initialization */
        for(i = 0; i < IX_DRAM_CHANNELS_NUMBER; i++)
        {
            ix_ossl_memset(&s_aDRAMMemoryManagers[i], 0, sizeof(_ix_mm_memory_manager));

            s_aDRAMMemoryManagers[i].m_MemoryType = IX_MEMORY_TYPE_DRAM;
            s_aDRAMMemoryManagers[i].m_MemoryChannel = i;
            s_aDRAMMemoryManagers[i].m_DefaultAlignmentShift = 3; /* 8 byte alignment */
            if((((ix_uint32)&s_aDRAMMemoryBuffers[i]) & 0x07) == 0)
            {
                s_aDRAMMemoryManagers[i].m_pStartAddress = (ix_uint32*)&s_aDRAMMemoryBuffers[i];
                s_aDRAMMemoryManagers[i].m_MemoryBlockSize = IX_TESTING_DRAM_CHANNEL_MEM_SIZE;
            }
            else
            {
                s_aDRAMMemoryManagers[i].m_pStartAddress = ((ix_uint32*)&s_aDRAMMemoryBuffers[i]) + 1 ;
                s_aDRAMMemoryManagers[i].m_MemoryBlockSize = IX_TESTING_DRAM_CHANNEL_MEM_SIZE - 8;
            }
            s_aDRAMMemoryManagers[i].m_ChannelPhysicalOffset = 0;
            s_aDRAMMemoryManagers[i].m_pChannelPhysicalAddress = 0;
            s_aDRAMMemoryManagers[i].m_pVMT = &s_DRAMMemoryManagerVMT;
        }

        /* SRAM memory initialization */
        for(i = 0; i < IX_SRAM_CHANNELS_NUMBER; i++)
        {
            ix_ossl_memset(&s_aSRAMMemoryManagers[i], 0, sizeof(_ix_mm_memory_manager));

            s_aSRAMMemoryManagers[i].m_MemoryType = IX_MEMORY_TYPE_SRAM;
            s_aSRAMMemoryManagers[i].m_MemoryChannel = i;
            s_aSRAMMemoryManagers[i].m_DefaultAlignmentShift = 2; /* 4 byte alignment */
            if((((ix_uint32)&s_aSRAMMemoryBuffers[i]) & 0x07) == 0)
            {
                s_aSRAMMemoryManagers[i].m_pStartAddress = (ix_uint32*)&s_aSRAMMemoryBuffers[i];
                s_aSRAMMemoryManagers[i].m_MemoryBlockSize = IX_TESTING_SRAM_CHANNEL_MEM_SIZE;
            }
            else
            {
                s_aSRAMMemoryManagers[i].m_pStartAddress = ((ix_uint32*)&s_aSRAMMemoryBuffers[i]) + 1 ;
                s_aSRAMMemoryManagers[i].m_MemoryBlockSize = IX_TESTING_SRAM_CHANNEL_MEM_SIZE - 8;
            }
            s_aSRAMMemoryManagers[i].m_ChannelPhysicalOffset = 0;
            s_aSRAMMemoryManagers[i].m_pChannelPhysicalAddress = 0;
            s_aSRAMMemoryManagers[i].m_pVMT = &s_SRAMMemoryManagerVMT;
        }

        /* SCRATCH memory initialization */
        for(i = 0; i < IX_SCRATCH_CHANNELS_NUMBER; i++)
        {
            ix_ossl_memset(&s_aSCRATCHMemoryManagers[i], 0, sizeof(_ix_mm_memory_manager));

            s_aSCRATCHMemoryManagers[i].m_MemoryType = IX_MEMORY_TYPE_SCRATCH;
            s_aSCRATCHMemoryManagers[i].m_MemoryChannel = i;
            s_aSCRATCHMemoryManagers[i].m_DefaultAlignmentShift = 2; /* 4 byte alignment */
            if((((ix_uint32)&s_aSCRATCHMemoryBuffers[i]) & 0x07) == 0)
            {
                s_aSCRATCHMemoryManagers[i].m_pStartAddress = (ix_uint32*)&s_aSCRATCHMemoryBuffers[i];
                s_aSCRATCHMemoryManagers[i].m_MemoryBlockSize = IX_TESTING_SCRATCH_CHANNEL_MEM_SIZE;
            }
            else
            {
                s_aSCRATCHMemoryManagers[i].m_pStartAddress = ((ix_uint32*)&s_aSCRATCHMemoryBuffers[i]) + 1 ;
                s_aSCRATCHMemoryManagers[i].m_MemoryBlockSize = IX_TESTING_SCRATCH_CHANNEL_MEM_SIZE - 8;
            }
            s_aSCRATCHMemoryManagers[i].m_ChannelPhysicalOffset = 0;
            s_aSCRATCHMemoryManagers[i].m_pChannelPhysicalAddress = 0;
            s_aSCRATCHMemoryManagers[i].m_pVMT = &s_SCRATCHMemoryManagerVMT;
        }

        /* LOCAL memory initialization */
        for(i = 0; i < IX_MAX_MICROENGINE_NUMBER; i++)
        {
            ix_ossl_memset(&s_aLOCALMemoryManagers[i], 0, sizeof(_ix_mm_memory_manager));

            s_aLOCALMemoryManagers[i].m_MemoryType = IX_MEMORY_TYPE_LOCAL;
            s_aLOCALMemoryManagers[i].m_MemoryChannel = i;
            s_aLOCALMemoryManagers[i].m_DefaultAlignmentShift = 2; /* 4 byte alignment */
            s_aLOCALMemoryManagers[i].m_pStartAddress = (ix_uint32*)&s_aLOCALMemoryBuffers[i];
            s_aLOCALMemoryManagers[i].m_MemoryBlockSize = IX_TESTING_LOCAL_MEM_SIZE;
            s_aLOCALMemoryManagers[i].m_ChannelPhysicalOffset = 0;
            s_aLOCALMemoryManagers[i].m_pChannelPhysicalAddress = 0;
            s_aLOCALMemoryManagers[i].m_pVMT = &s_LOCALMemoryManagerVMT;
        }
    }

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* initialize the synchronization semaphore */
    /* for now all memory managers will use the same semaphore for synchronization */
    err = ix_ossl_sem_init(IX_OSSL_SEM_AVAILABLE, &s_MemoryManagerSemaphore);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_MEMORY_INIT_FAILED, s_aFunctionIdentifier, 
                        "initialization of memory managers' semaphore failed!", err);
    }

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* iterate through all memory managers and invoke the init function */

    for(memType = IX_MEMORY_TYPE_DRAM; memType <= IX_MEMORY_TYPE_LOCAL; memType++)
    {
        /* search each memory manager in the group */
        for(channelIndex = 0; channelIndex < g_aMemoryManagersGroups[memType].m_MemberCount; channelIndex++)
        {
            pMemoryManager = g_aMemoryManagersGroups[memType].m_pMembers + channelIndex;

#if defined( _IX_RM_MULTI_THREADED_ )

            /* initialize synchronization */
            /* for now all memory managers will use the same semaphore for synchronization */
            _ix_rm_sa_sync_adapter_init(&pMemoryManager->m_SyncAdapter, 
                _ix_rm_sa_default_lock, _ix_rm_sa_default_unlock, (void*)&s_MemoryManagerSemaphore);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

            /* invoke the init function */
            err = pMemoryManager->m_pVMT->m_InitFuncPtr(pMemoryManager);

            /* if there was an error then return */
            if(err != IX_SUCCESS) 
                return err;
        }
    }
    return err;
}

#else /* Westport hardware */


/**
 * TYPENAME: ix_memory_init_info
 * 
 * DESCRIPTION: Initialization info for each memory type
 *
 */
typedef struct ix_s_memory_init_info
{
    ix_uint32   m_Size;
    void*       m_pVirtualAddress;
    void*       m_pPhysicalAddress;
    ix_uint32   m_ChannelPhysicalOffset;
    void*       m_pNonAliasPhysicalAddress;
    ix_uint32   m_NonAliasChannelPhysicalOffset;
} ix_memory_init_info;



/**
 * TYPENAME: ix_memory_system_init_info
 * 
 * DESCRIPTION: Initialization info for the whole memory in the system
 *
 */
typedef struct ix_s_memory_system_init_info
{
    ix_memory_init_info     m_aSouthDRAMInfo[IX_DRAM_SOUTH_CHANNELS_NUMBER];
    ix_memory_init_info     m_aSouthSRAM_MSGInfo[IX_SRAM_MSG_SOUTH_CHANNELS_NUMBER];
    ix_memory_init_info     m_aNorthDRAMInfo[IX_DRAM_NORTH_CHANNELS_NUMBER];
    ix_memory_init_info     m_aNorthSRAMInfo[IX_SRAM_NORTH_CHANNELS_NUMBER];
    ix_memory_init_info     m_aSCRATCHInfo[IX_SCRATCH_CHANNELS_NUMBER];
} ix_memory_system_init_info;

#if defined(_IX_RM_IMPL_HARDWARE_)

/**
 * NAME: _ix_rm_memory_system_init_info_get
 *
 * DESCRIPTION: This function will 
 * 
 * @Param:  - IN arg_pMemorySystemInfo - address where the system memory info should be updated
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_memory_system_init_info_get( 
                                             ix_memory_system_init_info* arg_pMemorySystemInfo
                                           )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_memory_system_init_info_get";

    ix_error err = IX_SUCCESS;
    ix_memory_init_info* pMemoryInitInfo;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* Initialize the south DRAM info */
    pMemoryInitInfo = &arg_pMemorySystemInfo->m_aSouthDRAMInfo[0];
    pMemoryInitInfo->m_Size = IX_WESTPORT_DRAM_SOUTH_SIZE;
    pMemoryInitInfo->m_pVirtualAddress = g_westport_dram_south_base_addr;
    pMemoryInitInfo->m_pPhysicalAddress = (void*)IX_WESTPORT_PHYS_DRAM_SOUTH_BASE_ADDR;
    pMemoryInitInfo->m_ChannelPhysicalOffset = 0;
    pMemoryInitInfo->m_pNonAliasPhysicalAddress = (void*)IX_WESTPORT_PHYS_DRAM_SOUTH_BASE_ADDR;
    pMemoryInitInfo->m_NonAliasChannelPhysicalOffset = 0;

    /* Initialize the north DRAM info */
    pMemoryInitInfo = &arg_pMemorySystemInfo->m_aNorthDRAMInfo[0];
    pMemoryInitInfo->m_Size = IX_WESTPORT_DRAM_NORTH_SIZE;
    pMemoryInitInfo->m_pVirtualAddress = g_westport_dram_north_base_addr;
    pMemoryInitInfo->m_pPhysicalAddress = (void*)IX_WESTPORT_PHYS_DRAM_NORTH_BASE_ADDR;
    pMemoryInitInfo->m_ChannelPhysicalOffset = 0;
    pMemoryInitInfo->m_pNonAliasPhysicalAddress = (void*)IX_WESTPORT_PHYS_DRAM_NORTH_BASE_ADDR_NON_ALIAS;
    pMemoryInitInfo->m_NonAliasChannelPhysicalOffset = 0;

    /* Initialize the south MSG-SRAM info */
    pMemoryInitInfo = &arg_pMemorySystemInfo->m_aSouthSRAM_MSGInfo[0];
    pMemoryInitInfo->m_Size = IX_WESTPORT_MSG_SRAM_SOUTH_SIZE;
    pMemoryInitInfo->m_pVirtualAddress = g_westport_msg_sram_south_base_addr;
    pMemoryInitInfo->m_pPhysicalAddress = (void*)IX_WESTPORT_PHYS_MSG_SRAM_SOUTH_BASE_ADDR;
    pMemoryInitInfo->m_ChannelPhysicalOffset = 0;
    pMemoryInitInfo->m_pNonAliasPhysicalAddress = (void*)IX_WESTPORT_PHYS_MSG_SRAM_SOUTH_BASE_ADDR;
    pMemoryInitInfo->m_NonAliasChannelPhysicalOffset = 0;

    /* Initialize the north SRAM info */
    pMemoryInitInfo = &arg_pMemorySystemInfo->m_aNorthSRAMInfo[0];
    pMemoryInitInfo->m_Size = IX_WESTPORT_SRAM_NORTH_SIZE;
    pMemoryInitInfo->m_pVirtualAddress = g_westport_sram_north_base_addr;
    pMemoryInitInfo->m_pPhysicalAddress = (void*)IX_WESTPORT_PHYS_SRAM_NORTH_BASE_ADDR;
    pMemoryInitInfo->m_ChannelPhysicalOffset = 0;
    pMemoryInitInfo->m_pNonAliasPhysicalAddress = (void*)IX_WESTPORT_PHYS_SRAM_NORTH_BASE_ADDR_NON_ALIAS;
    pMemoryInitInfo->m_NonAliasChannelPhysicalOffset = 0;

    /* Initialize the north SCRATCH info */
    pMemoryInitInfo = &arg_pMemorySystemInfo->m_aSCRATCHInfo[0];
    pMemoryInitInfo->m_Size = IX_WESTPORT_SCRATCH_SIZE;
    pMemoryInitInfo->m_pVirtualAddress = g_westport_scratch_base_addr;
    pMemoryInitInfo->m_pPhysicalAddress = (void*)IX_WESTPORT_PHYS_SCRATCH_BASE_ADDR;
    pMemoryInitInfo->m_ChannelPhysicalOffset = 0;
    pMemoryInitInfo->m_pNonAliasPhysicalAddress = (void*)IX_WESTPORT_PHYS_SCRATCH_BASE_ADDR_NON_ALIAS;
    pMemoryInitInfo->m_NonAliasChannelPhysicalOffset = 0;

    return err;
}

#else

static ix_memory_system_init_info s_MemorySystemInitInfo =
{
    /* South DRAM */
    {
        /* Channel 0 */
        {
            IX_TESTING_DRAM_SOUTH_CHANNEL_MEM_SIZE,
            s_aSouthDRAMMemoryBuffers[0],
            (void*)0x10000000,
            0x10000000,
            (void*)0x10000000,
            0x10000000
        }
    },

    /* South MSG_SRAM */
    {
        /* Channel 0 */
        {
            IX_TESTING_SRAM_MSG_SOUTH_CHANNEL_MEM_SIZE,
            s_aSouthSRAM_MSGMemoryBuffers[0],
            (void*)0x48000000,
            0x00000000,
            (void*)0x48000000,
            0x00000000
        }
    },

    /* North DRAM */
    {
        /* Channel 0 */
        {
            IX_TESTING_DRAM_NORTH_CHANNEL_MEM_SIZE,
            s_aNorthDRAMMemoryBuffers[0],
            (void*)0xd0000000,
            0x00000000,
            (void*)0x20000000,
            0x20000000
        }
    },

    /* North SRAM */
    {
        /* Channel 0 */
        {
            IX_TESTING_SRAM_NORTH_CHANNEL_MEM_SIZE,
            s_aNorthSRAMMemoryBuffers[0],
            (void*)0x62000000,
            0x00000000,
            (void*)0x90000000,
            0x10000000
        }
    },

    /* North SCRATCH */
    {
        /* Channel 0 */
        {
            IX_TESTING_SCRATCH_CHANNEL_MEM_SIZE,
            s_aSCRATCHMemoryBuffers[0],
            (void*)0x86000000,
            0x00000000,
            (void*)0xc6000000,
            0x00000000
        }
    }
};

/**
 * NAME: _ix_rm_memory_system_init_info_get
 *
 * DESCRIPTION: This function will 
 * 
 * @Param:  - IN arg_pMemorySystemInfo - address where the system memory info should be updated
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_memory_system_init_info_get( 
                                             ix_memory_system_init_info* arg_pMemorySystemInfo
                                           )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_memory_system_init_info_get";

    ix_error err = IX_SUCCESS;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    ix_ossl_memcpy(arg_pMemorySystemInfo, &s_MemorySystemInitInfo, sizeof(s_MemorySystemInitInfo));

    return err;
}

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */



/**
 * NAME: _ix_rm_memory_init
 *
 * DESCRIPTION: This function should be called at the time the resource manager initializer 
 *          function is called. The function will initialize all memory managers based on 
 *          the hardware configuration.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_memory_init(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_memory_init";

    ix_error err = IX_SUCCESS;

    register _ix_mm_memory_manager* pMemoryManager;
    register ix_uint32   channelIndex;
    register ix_uint32   memType;

    ix_memory_system_init_info memorySystemInitInfo;
    ix_uint32 i;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    _ix_rm_memory_system_init_info_get(&memorySystemInitInfo);

    /* Initialize the managed South DRAM memory */
    for(i = 0; i < IX_DRAM_SOUTH_CHANNELS_NUMBER; i++)
    {
        s_aSouthDRAMMemoryManagers[i].m_MemoryType = IX_MEMORY_TYPE_DRAM_SOUTH;
        s_aSouthDRAMMemoryManagers[i].m_MemoryBankType = IX_MEMORY_BANK_TYPE_SOUTH;
        s_aSouthDRAMMemoryManagers[i].m_MemoryChannel = i;
        s_aSouthDRAMMemoryManagers[i].m_DefaultAlignmentShift = 3; /* 8 byte alignment */
        s_aSouthDRAMMemoryManagers[i].m_pStartAddress = (ix_uint32*)memorySystemInitInfo.m_aSouthDRAMInfo[i].m_pVirtualAddress;
        s_aSouthDRAMMemoryManagers[i].m_MemoryBlockSize = memorySystemInitInfo.m_aSouthDRAMInfo[i].m_Size;
        s_aSouthDRAMMemoryManagers[i].m_pChannelPhysicalAddress = (ix_uint32*)memorySystemInitInfo.m_aSouthDRAMInfo[i].m_pPhysicalAddress;
        s_aSouthDRAMMemoryManagers[i].m_ChannelPhysicalOffset = memorySystemInitInfo.m_aSouthDRAMInfo[i].m_ChannelPhysicalOffset;
        s_aSouthDRAMMemoryManagers[i].m_pNonAliasChannelPhysicalAddress = (ix_uint32*)memorySystemInitInfo.m_aSouthDRAMInfo[i].m_pNonAliasPhysicalAddress;
        s_aSouthDRAMMemoryManagers[i].m_NonAliasChannelPhysicalOffset = memorySystemInitInfo.m_aSouthDRAMInfo[i].m_NonAliasChannelPhysicalOffset;
        s_aSouthDRAMMemoryManagers[i].m_pVMT = &s_SouthDRAMMemoryManagerVMT;
    }

    /* Initialize the managed South MSG-SRAM memory */
    for(i = 0; i < IX_SRAM_MSG_SOUTH_CHANNELS_NUMBER; i++)
    {
        s_aSouthSRAM_MSGMemoryManagers[i].m_MemoryType = IX_MEMORY_TYPE_SRAM_MSG_SOUTH;
        s_aSouthSRAM_MSGMemoryManagers[i].m_MemoryBankType = IX_MEMORY_BANK_TYPE_SOUTH;
        s_aSouthSRAM_MSGMemoryManagers[i].m_MemoryChannel = i;
        s_aSouthSRAM_MSGMemoryManagers[i].m_DefaultAlignmentShift = 2; /* 4 byte alignment */
        s_aSouthSRAM_MSGMemoryManagers[i].m_pStartAddress = (ix_uint32*)memorySystemInitInfo.m_aSouthSRAM_MSGInfo[i].m_pVirtualAddress;
        s_aSouthSRAM_MSGMemoryManagers[i].m_MemoryBlockSize = memorySystemInitInfo.m_aSouthSRAM_MSGInfo[i].m_Size;
        s_aSouthSRAM_MSGMemoryManagers[i].m_pChannelPhysicalAddress = (ix_uint32*)memorySystemInitInfo.m_aSouthSRAM_MSGInfo[i].m_pPhysicalAddress;
        s_aSouthSRAM_MSGMemoryManagers[i].m_ChannelPhysicalOffset = memorySystemInitInfo.m_aSouthSRAM_MSGInfo[i].m_ChannelPhysicalOffset;
        s_aSouthSRAM_MSGMemoryManagers[i].m_pNonAliasChannelPhysicalAddress = (ix_uint32*)memorySystemInitInfo.m_aSouthSRAM_MSGInfo[i].m_pNonAliasPhysicalAddress;
        s_aSouthSRAM_MSGMemoryManagers[i].m_NonAliasChannelPhysicalOffset = memorySystemInitInfo.m_aSouthSRAM_MSGInfo[i].m_NonAliasChannelPhysicalOffset;
        s_aSouthSRAM_MSGMemoryManagers[i].m_pVMT = &s_SouthSRAM_MSGMemoryManagerVMT;
    }

    /* Initialize the managed North DRAM memory */
    for(i = 0; i < IX_DRAM_NORTH_CHANNELS_NUMBER; i++)
    {
        s_aNorthDRAMMemoryManagers[i].m_MemoryType = IX_MEMORY_TYPE_DRAM_NORTH;
        s_aNorthDRAMMemoryManagers[i].m_MemoryBankType = IX_MEMORY_BANK_TYPE_NORTH;
        s_aNorthDRAMMemoryManagers[i].m_MemoryChannel = i;
        s_aNorthDRAMMemoryManagers[i].m_DefaultAlignmentShift = 3; /* 8 byte alignment */
        s_aNorthDRAMMemoryManagers[i].m_pStartAddress = (ix_uint32*)memorySystemInitInfo.m_aNorthDRAMInfo[i].m_pVirtualAddress;
        s_aNorthDRAMMemoryManagers[i].m_MemoryBlockSize = memorySystemInitInfo.m_aNorthDRAMInfo[i].m_Size;
        s_aNorthDRAMMemoryManagers[i].m_pChannelPhysicalAddress = (ix_uint32*)memorySystemInitInfo.m_aNorthDRAMInfo[i].m_pPhysicalAddress;
        s_aNorthDRAMMemoryManagers[i].m_ChannelPhysicalOffset = memorySystemInitInfo.m_aNorthDRAMInfo[i].m_ChannelPhysicalOffset;
        s_aNorthDRAMMemoryManagers[i].m_pNonAliasChannelPhysicalAddress = (ix_uint32*)memorySystemInitInfo.m_aNorthDRAMInfo[i].m_pNonAliasPhysicalAddress;
        s_aNorthDRAMMemoryManagers[i].m_NonAliasChannelPhysicalOffset = memorySystemInitInfo.m_aNorthDRAMInfo[i].m_NonAliasChannelPhysicalOffset;
        s_aNorthDRAMMemoryManagers[i].m_pVMT = &s_NorthDRAMMemoryManagerVMT;
    }

    /* Initialize the managed North SRAM memory */
    for(i = 0; i < IX_SRAM_NORTH_CHANNELS_NUMBER; i++)
    {
        s_aNorthSRAMMemoryManagers[i].m_MemoryType = IX_MEMORY_TYPE_SRAM_NORTH;
        s_aNorthSRAMMemoryManagers[i].m_MemoryBankType = IX_MEMORY_BANK_TYPE_NORTH;
        s_aNorthSRAMMemoryManagers[i].m_MemoryChannel = i;
        s_aNorthSRAMMemoryManagers[i].m_DefaultAlignmentShift = 2; /* 4 byte alignment */
        s_aNorthSRAMMemoryManagers[i].m_pStartAddress = (ix_uint32*)memorySystemInitInfo.m_aNorthSRAMInfo[i].m_pVirtualAddress;
        s_aNorthSRAMMemoryManagers[i].m_MemoryBlockSize = memorySystemInitInfo.m_aNorthSRAMInfo[i].m_Size;
        s_aNorthSRAMMemoryManagers[i].m_pChannelPhysicalAddress = (ix_uint32*)memorySystemInitInfo.m_aNorthSRAMInfo[i].m_pPhysicalAddress;
        s_aNorthSRAMMemoryManagers[i].m_ChannelPhysicalOffset = memorySystemInitInfo.m_aNorthSRAMInfo[i].m_ChannelPhysicalOffset;
        s_aNorthSRAMMemoryManagers[i].m_pNonAliasChannelPhysicalAddress = (ix_uint32*)memorySystemInitInfo.m_aNorthSRAMInfo[i].m_pNonAliasPhysicalAddress;
        s_aNorthSRAMMemoryManagers[i].m_NonAliasChannelPhysicalOffset = memorySystemInitInfo.m_aNorthSRAMInfo[i].m_NonAliasChannelPhysicalOffset;
        s_aNorthSRAMMemoryManagers[i].m_pVMT = &s_NorthSRAMMemoryManagerVMT;
    }

    /* Initialize the managed North SCRATCH memory */
    for(i = 0; i < IX_SRAM_NORTH_CHANNELS_NUMBER; i++)
    {
        s_aSCRATCHMemoryManagers[i].m_MemoryType = IX_MEMORY_TYPE_SCRATCH;
        s_aSCRATCHMemoryManagers[i].m_MemoryBankType = IX_MEMORY_BANK_TYPE_NORTH;
        s_aSCRATCHMemoryManagers[i].m_MemoryChannel = i;
        s_aSCRATCHMemoryManagers[i].m_DefaultAlignmentShift = 2; /* 4 byte alignment */
        s_aSCRATCHMemoryManagers[i].m_pStartAddress = (ix_uint32*)memorySystemInitInfo.m_aSCRATCHInfo[i].m_pVirtualAddress;
        s_aSCRATCHMemoryManagers[i].m_MemoryBlockSize = memorySystemInitInfo.m_aSCRATCHInfo[i].m_Size;
        s_aSCRATCHMemoryManagers[i].m_pChannelPhysicalAddress = (ix_uint32*)memorySystemInitInfo.m_aSCRATCHInfo[i].m_pPhysicalAddress;
        s_aSCRATCHMemoryManagers[i].m_ChannelPhysicalOffset = memorySystemInitInfo.m_aSCRATCHInfo[i].m_ChannelPhysicalOffset;
        s_aSCRATCHMemoryManagers[i].m_pNonAliasChannelPhysicalAddress = (ix_uint32*)memorySystemInitInfo.m_aSCRATCHInfo[i].m_pNonAliasPhysicalAddress;
        s_aSCRATCHMemoryManagers[i].m_NonAliasChannelPhysicalOffset = memorySystemInitInfo.m_aSCRATCHInfo[i].m_NonAliasChannelPhysicalOffset;
        s_aSCRATCHMemoryManagers[i].m_pVMT = &s_SCRATCHMemoryManagerVMT;
    }


#if defined( _IX_RM_MULTI_THREADED_ )

    /* initialize the synchronization semaphore */
    /* for now all memory managers will use the same semaphore for synchronization */
    err = ix_ossl_sem_init(IX_OSSL_SEM_AVAILABLE, &s_MemoryManagerSemaphore);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_MEMORY_INIT_FAILED, s_aFunctionIdentifier, 
                        "initialization of memory managers' semaphore failed!", err);
    }

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* iterate through all memory managers and invoke the init function */

    for(memType = IX_MEMORY_TYPE_FIRST; memType < IX_MEMORY_TYPE_LAST; memType++)
    {
        /* search each memory manager in the group */
        for(channelIndex = 0; channelIndex < g_aMemoryManagersGroups[memType].m_MemberCount; channelIndex++)
        {
            pMemoryManager = g_aMemoryManagersGroups[memType].m_pMembers + channelIndex;

#if defined( _IX_RM_MULTI_THREADED_ )

            /* initialize synchronization */
            /* for now all memory managers will use the same semaphore for synchronization */
            _ix_rm_sa_sync_adapter_init(&pMemoryManager->m_SyncAdapter, 
                _ix_rm_sa_default_lock, _ix_rm_sa_default_unlock, (void*)&s_MemoryManagerSemaphore);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

            /* invoke the init function */
            err = pMemoryManager->m_pVMT->m_InitFuncPtr(pMemoryManager);

            /* if there was an error then return */
            if(err != IX_SUCCESS) 
                return err;
        }
    }
    return err;
}

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */



/**
 * NAME: _ix_rm_memory_fini
 *
 * DESCRIPTION: This function should be called at the time the resource manager finalizer function
 *          is called. The function will finalize all memory managers.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_memory_fini(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_memory_fini";

    ix_error err = IX_SUCCESS;

    register _ix_mm_memory_manager* pMemoryManager;
    register ix_uint32   channelIndex;
    register ix_uint32   memType;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* iterate through all memory managers and invoke the fini function */

    for(memType = IX_MEMORY_TYPE_FIRST; memType < IX_MEMORY_TYPE_LAST; memType++)
    {
        /* search each memory manager in the group */
        for(channelIndex = 0; channelIndex < g_aMemoryManagersGroups[memType].m_MemberCount; channelIndex++)
        {
            pMemoryManager = g_aMemoryManagersGroups[memType].m_pMembers + channelIndex;

            /* invoke the init function */
            err = pMemoryManager->m_pVMT->m_FiniFuncPtr(pMemoryManager);

            /* if there was an error then return */
            if(err != IX_SUCCESS) 
                return err;
#if defined( _IX_RM_MULTI_THREADED_ )

            /* finalize synchronization */
            _ix_rm_sa_sync_adapter_init(&pMemoryManager->m_SyncAdapter, 0, 0, 0);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

        }
    }
    
#if defined( _IX_RM_MULTI_THREADED_ )

    /* finalize the memory managers' synchronization semaphore */
    err = ix_ossl_sem_fini(s_MemoryManagerSemaphore);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_MEMORY_FINI_FAILED, s_aFunctionIdentifier, 
                        "finalization of memory managers' semaphore failed!", err);
    }

    s_MemoryManagerSemaphore = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


#if 0

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * NAME: _ix_rm_mm_print_memory_manager
 *
 * DESCRIPTION: This is a support function that will print on standard output the memory manager
 *          information passed as a parameter.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager structure that will be
 *          printed.
 */
void _ix_rm_mm_print_memory_manager( 
                                     _ix_mm_memory_manager* arg_pMemoryManager 
                                   )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mm_print_memory_manager";

    ix_ossl_message_log("%s: memory manager address: 0x%p\n",s_aFunctionIdentifier, arg_pMemoryManager);
    switch(arg_pMemoryManager->m_MemoryType)
    {
        case IX_MEMORY_TYPE_DRAM:
            ix_ossl_message_log("\tm_MemoryType: DRAM\n");
            break;
        case IX_MEMORY_TYPE_SRAM:
            ix_ossl_message_log("\tm_MemoryType: SRAM\n");
            break;
        case IX_MEMORY_TYPE_SCRATCH:
            ix_ossl_message_log("\tm_MemoryType: SCRATCH\n");
            break;
        case IX_MEMORY_TYPE_LOCAL:
            ix_ossl_message_log("\tm_MemoryType: LOCAL\n");
            break;
        default:
            ix_ossl_message_log("\tm_MemoryType: INVALID\n");
            break;
    }

    ix_ossl_message_log("\tm_MemoryChannel: %lu\n",arg_pMemoryManager->m_MemoryChannel);
    ix_ossl_message_log("\tm_MemoryBlockSize: %lu\n",arg_pMemoryManager->m_MemoryBlockSize);
    ix_ossl_message_log("\tm_pStartAddress: 0x%p",arg_pMemoryManager->m_pStartAddress);
    ix_ossl_message_log("\tm_pEndAddress: 0x%p\n",arg_pMemoryManager->m_pEndAddress);
    ix_ossl_message_log("\tm_ChannelPhysicalOffset: %lu\n",arg_pMemoryManager->m_ChannelPhysicalOffset);
    ix_ossl_message_log("\tm_pChannelPhysicalAddress: 0x%p\n", arg_pMemoryManager->m_pChannelPhysicalAddress);
    ix_ossl_message_log("\tm_pRover: 0x%p\n",arg_pMemoryManager->m_pRover);
    ix_ossl_message_log("\tm_pBlockList: 0x%p\n",arg_pMemoryManager->m_pBlockList);

    /* print the whole list now */
    {
        ix_uint32 count = 0;
        /* iterate through entire list and print all entries */
        _ix_rm_memory_block_cell_action(arg_pMemoryManager->m_pBlockList, 0, 
            _ix_rm_memory_block_cell_print, &count);
        /* print the total number of entries */
        ix_ossl_message_log("\t\t%lu entries\n",count);
    }

    ix_ossl_message_log("\tm_RecycledBlockCellCount: %lu\n",arg_pMemoryManager->m_RecycledBlockCellCount);
    ix_ossl_message_log("\tm_pRecycledBlockCellList: 0x%p\n",arg_pMemoryManager->m_pRecycledBlockCellList);
    /* print the whole list now */
    {
        ix_uint32 count = 0;
        /* iterate through entire list and print all entries */
        _ix_rm_memory_block_cell_action(arg_pMemoryManager->m_pBlockList, 0, 
            _ix_rm_memory_recycled_cell_print, &count);
        /* print the total number of entries */
        ix_ossl_message_log("\t\t%lu entries\n",count);
    }

    ix_ossl_message_log("\tm_pVMT: %p\n",arg_pMemoryManager->m_pVMT);
    /* print the whole VMT now */
    ix_ossl_message_log("\t\tm_AllocFuncPtr: 0x%p\n",arg_pMemoryManager->m_pVMT->m_AllocFuncPtr);
    ix_ossl_message_log("\t\tm_FiniFuncPtr: 0x%p\n",arg_pMemoryManager->m_pVMT->m_FiniFuncPtr);
    ix_ossl_message_log("\t\tm_FreeFuncPtr: 0x%p\n",arg_pMemoryManager->m_pVMT->m_FreeFuncPtr);
    ix_ossl_message_log("\t\tm_InitFuncPtr: 0x%p\n",arg_pMemoryManager->m_pVMT->m_InitFuncPtr);
    ix_ossl_message_log("\t\tm_ReadFuncPtr: 0x%p\n",arg_pMemoryManager->m_pVMT->m_ReadFuncPtr);
    ix_ossl_message_log("\t\tm_WriteFuncPtr: 0x%p\n",arg_pMemoryManager->m_pVMT->m_WriteFuncPtr);
}

#else /* Westport hardware */

/**
 * NAME: _ix_rm_mm_print_memory_manager
 *
 * DESCRIPTION: This is a support function that will print on standard output the memory manager
 *          information passed as a parameter.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager structure that will be
 *          printed.
 */
void _ix_rm_mm_print_memory_manager( 
                                     _ix_mm_memory_manager* arg_pMemoryManager 
                                   )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mm_print_memory_manager";

    ix_ossl_message_log("%s: memory manager address: %p\n",s_aFunctionIdentifier, arg_pMemoryManager);
    switch(arg_pMemoryManager->m_MemoryType)
    {
        case IX_MEMORY_TYPE_DRAM_SOUTH:
            ix_ossl_message_log("\tm_MemoryType: South bank DRAM");
            break;
        case IX_MEMORY_TYPE_SRAM_MSG_SOUTH:
            ix_ossl_message_log("\tm_MemoryType: South bank MSG-SRAM");
            break;
        case IX_MEMORY_TYPE_DRAM_NORTH:
            ix_ossl_message_log("\tm_MemoryType: North bank aliased DRAM");
            break;
        case IX_MEMORY_TYPE_SRAM_NORTH:
            ix_ossl_message_log("\tm_MemoryType: North bank aliased SRAM");
            break;
        case IX_MEMORY_TYPE_SCRATCH:
            ix_ossl_message_log("\tm_MemoryType: North bank aliased SCRATCH");
            break;
        default:
            ix_ossl_message_log("\tm_MemoryType: Unknown");
            break;
    }

    ix_ossl_message_log("\tm_MemoryBankType: %s\n", (arg_pMemoryManager->m_MemoryBankType == IX_MEMORY_BANK_TYPE_SOUTH) ? "SOUTH" : "NORTH");
    ix_ossl_message_log("\tm_MemoryChannel: %lu\n",arg_pMemoryManager->m_MemoryChannel);
    ix_ossl_message_log("\tm_MemoryBlockSize: %lu\n",arg_pMemoryManager->m_MemoryBlockSize);
    ix_ossl_message_log("\tm_pStartAddress: 0x%p",arg_pMemoryManager->m_pStartAddress);
    ix_ossl_message_log("\tm_pEndAddress: 0x%p\n",arg_pMemoryManager->m_pEndAddress);
    ix_ossl_message_log("\tm_ChannelPhysicalOffset: %lu\n",arg_pMemoryManager->m_ChannelPhysicalOffset);
    ix_ossl_message_log("\tm_pChannelPhysicalAddress: 0x%p\n", arg_pMemoryManager->m_pChannelPhysicalAddress);
    ix_ossl_message_log("\tm_pNonAliasChannelPhysicalAddress: 0x%p\n", arg_pMemoryManager->m_pChannelPhysicalAddress);
    ix_ossl_message_log("\tm_NonAliasChannelPhysicalOffset: %lu\n",arg_pMemoryManager->m_NonAliasChannelPhysicalOffset);
    ix_ossl_message_log("\tm_pRover: 0x%p\n",arg_pMemoryManager->m_pRover);
    ix_ossl_message_log("\tm_pBlockList: 0x%p\n",arg_pMemoryManager->m_pBlockList);

    /* print the whole list now */
    {
        ix_uint32 count = 0;
        /* iterate through entire list and print all entries */
        _ix_rm_memory_block_cell_action(arg_pMemoryManager->m_pBlockList, 0, 
            _ix_rm_memory_block_cell_print, &count);
        /* print the total number of entries */
        ix_ossl_message_log("\t\t%lu entries\n",count);
    }

    ix_ossl_message_log("\tm_RecycledBlockCellCount: %lu\n",arg_pMemoryManager->m_RecycledBlockCellCount);
    ix_ossl_message_log("\tm_pRecycledBlockCellList: 0x%p\n",arg_pMemoryManager->m_pRecycledBlockCellList);
    /* print the whole list now */
    {
        ix_uint32 count = 0;
        /* iterate through entire list and print all entries */
        _ix_rm_memory_block_cell_action(arg_pMemoryManager->m_pBlockList, 0, 
            _ix_rm_memory_recycled_cell_print, &count);
        /* print the total number of entries */
        ix_ossl_message_log("\t\t%lu entries\n",count);
    }

    ix_ossl_message_log("\tm_pVMT: 0x%p\n",arg_pMemoryManager->m_pVMT);
    /* print the whole VMT now */
    ix_ossl_message_log("\t\tm_AllocFuncPtr: 0x%p\n",arg_pMemoryManager->m_pVMT->m_AllocFuncPtr);
    ix_ossl_message_log("\t\tm_FiniFuncPtr: 0x%p\n",arg_pMemoryManager->m_pVMT->m_FiniFuncPtr);
    ix_ossl_message_log("\t\tm_FreeFuncPtr: 0x%p\n",arg_pMemoryManager->m_pVMT->m_FreeFuncPtr);
    ix_ossl_message_log("\t\tm_InitFuncPtr: 0x%p\n",arg_pMemoryManager->m_pVMT->m_InitFuncPtr);
    ix_ossl_message_log("\t\tm_ReadFuncPtr: 0x%p\n",arg_pMemoryManager->m_pVMT->m_ReadFuncPtr);
    ix_ossl_message_log("\t\tm_WriteFuncPtr: 0x%p\n",arg_pMemoryManager->m_pVMT->m_WriteFuncPtr);
}

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

#endif /* 0 */

/**
 * NAME: _ix_rm_mm_allocator
 *
 * DESCRIPTION: This is an allocator function for all types of memory.
 *          For special needs different allocators will be defined, but this is the generic one.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager data this allocator works on.
 * @Param:  - IN arg_Size - the size of the memory block requested.
 * @Param:  - OUT arg_pMemory - a pointer to where to store the new allocated block address .
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure. In the case the call fails the *arg_pMemory
 *          is set to NULL. Caution about the fact that memory allocation and reservation will 
 *          consider the alignment and will adjust the size and start address in accordance with 
 *          the default alignment.
 */
ix_error _ix_rm_mm_allocator( 
                              _ix_mm_memory_manager* arg_pMemoryManager, 
                              ix_uint32 arg_Size, 
                              void** arg_pMemory 
                            )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mm_allocator";

    ix_error err = IX_SUCCESS;
    register _ix_mm_memory_block_cell* pActiveCell = 0;
    register _ix_mm_memory_block_cell* pRover = 0;
    ix_uint32 alignmentMask;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &arg_pMemoryManager->m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */


    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* adjust the size according to the default alignment */
    alignmentMask = ((1 << arg_pMemoryManager->m_DefaultAlignmentShift) - 1);
    arg_Size += alignmentMask;
    arg_Size &= ~alignmentMask;

#if defined(_IX_RM_MULTI_THREADED_)

    /* lock the memory manager data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    /* start from the rover location in the list */
    pActiveCell = pRover = arg_pMemoryManager->m_pRover;

    /* Search at the end of block list first */
    while(pActiveCell != 0)
    {
        /* check if the size of the block is large enough */
        if(pActiveCell->m_Size < arg_Size)
            pActiveCell = pActiveCell->m_pNextCell;
        else
        {
            /* The size matches exactly.*/
            /* If there is no size adjustment due to alignment we allocate the whole cell. */
            if(pActiveCell->m_Size == arg_Size)
            {
                /* set the cell to be in use */
                pActiveCell->m_Size = 0;

                /* set the memory address to this cell's data pointer */
                *arg_pMemory = pActiveCell->m_pData;

                /* reset the rover */
                if(pRover == pActiveCell)
                    arg_pMemoryManager->m_pRover = arg_pMemoryManager->m_pBlockList; 

                /* done with the allocation process */
                goto allocation_done;
            }
            else
            {
                register _ix_mm_memory_block_cell* pNewCell = 0;

                /* split the current cell and make pActiveCell of size arg_Size */
                pNewCell = _ix_rm_mm_split_cell(arg_pMemoryManager, pActiveCell, arg_Size);

                /* we could not split the cell */
                if(pNewCell == 0)
                {
                    /* set the memory address to 0 */
                    *arg_pMemory = 0;

                    /* set the error code */
                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_CELL_SPLIT_FAILED, s_aFunctionIdentifier, 
                        "memory cell split failed!", err);

                    /* done with the allocation process */
                    goto allocation_done;
                }
                else
                {
                    /* mark the new cell as being in use */
                    pActiveCell->m_Size = 0;

                    /* set the new cell's data pointer and the returned address to this value */
                    *arg_pMemory = pActiveCell->m_pData;

                    /* reset the rover */
                    arg_pMemoryManager->m_pRover = pNewCell; 

                    /* done with the allocation process */
                    goto allocation_done;
                }
            }
        }
    }

    /* we did not find any block that will fit our size */

    /* start from the beginning in the list */
    pActiveCell = arg_pMemoryManager->m_pBlockList;

    while(pActiveCell != pRover)
    {
        /* check if the size of the block is large enough */
        if(pActiveCell->m_Size < arg_Size)
            pActiveCell = pActiveCell->m_pNextCell;
        else
        {
            /* The size matches exactly.*/
            /* If there is no size adjustment due to alignment we allocate the whole cell. */
            if(pActiveCell->m_Size == arg_Size)
            {
                /* set the cell to be in use */
                pActiveCell->m_Size = 0;

                /* set the memory address to this cell's data pointer */
                *arg_pMemory = pActiveCell->m_pData;

                /* done with the allocation process */
                goto allocation_done;
            }
            else
            {
                register _ix_mm_memory_block_cell* pNewCell = 0;

                /* split the current cell and make pActiveCell of size arg_Size */
                pNewCell = _ix_rm_mm_split_cell(arg_pMemoryManager, pActiveCell, arg_Size);

                /* we could not split the cell */
                if(pNewCell == 0)
                {
                    /* set the memory address to 0 */
                    *arg_pMemory = 0;

                    /* set the error code */
                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_CELL_SPLIT_FAILED, s_aFunctionIdentifier, 
                        "memory cell split failed!", err);

                    /* done with the allocation process */
                    goto allocation_done;
                }
                else
                {
                    /* mark the new cell as being in use */
                    pActiveCell->m_Size = 0;

                    /* set the new cell's data pointer and the returned address to this value */
                    *arg_pMemory = pActiveCell->m_pData;

                    /* reset the rover */
                    arg_pMemoryManager->m_pRover = pNewCell; 

                    /* done with the allocation process */
                    goto allocation_done;
                }
            }
        }
    }
#ifdef _IX_RM_DEBUG_
    /*
     * Don't print any message in case of empty optional memory channel
     */
    if (arg_pMemoryManager->m_MemoryBlockSize == 0)
    {
        err = IX_ERROR_NEW(IX_RM_ERROR_CANNOT_ALLOCATE_SIZE,
                           IX_ERROR_LEVEL_LOCAL);
    }
    else
#endif
    err = IX_RM_ERROR_NEW(IX_RM_ERROR_CANNOT_ALLOCATE_SIZE, s_aFunctionIdentifier, 
        "the requested size can not be allocated!", err);


allocation_done:

#if defined(_IX_RM_MULTI_THREADED_)

    /* unlock the memory manager data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    /* return the error code */
    return err;
}



/**
 * NAME: _ix_rm_mm_aligned_allocator
 *
 * DESCRIPTION: This is an aligned allocator function for all types of memory.
 *          For special needs different aligned allocators will be defined, but this 
 *          is the generic one.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager data this allocator works on.
 * @Param:  - IN arg_Size - the size of the memory block requested.
 * @Param:  - IN arg_AlignmentType - specifies the type of alignment required as explained at 
 *          ix_memory_alignment_type type description
 * @Param:  - IN arg_AlignmentShift - this value specifies the required memory block alignment as the
 *          power of 2 of the alignment. If the specified alignment is less than the
 *          default alignment for the memory pool manager, then the default alignment will be used.
 * @Param:  - OUT arg_pMemory - a pointer to where to store the new allocated block address .
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure. In the case the call fails the *arg_pMemory
 *          is set to NULL. The returned memory will insure the right alignment for the given memory type
 *          and the requested alignment type.
 */
ix_error _ix_rm_mm_aligned_allocator( 
                                      _ix_mm_memory_manager* arg_pMemoryManager, 
                                      ix_uint32 arg_Size,
                                      ix_memory_alignment_type arg_AlignmentType,
                                      ix_uint32 arg_AlignmentShift,
                                      void** arg_pMemory 
                                    )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mm_aligned_allocator";

    ix_error err = IX_SUCCESS;
    register _ix_mm_memory_block_cell* pActiveCell = 0;
    register _ix_mm_memory_block_cell* pRover = 0;
    ix_uint32 alignmentMask;
    ix_uint32 sizeAdjustment;
    ix_uint32 temp;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &arg_pMemoryManager->m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */


    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* adjust the size according to the default alignment */
    alignmentMask = ((1 << arg_pMemoryManager->m_DefaultAlignmentShift) - 1);
    arg_Size += alignmentMask;
    arg_Size &= ~alignmentMask;

    /* adjust the alignment mask according to arg_AlignmentShift parameter */
    if(arg_AlignmentShift > arg_pMemoryManager->m_DefaultAlignmentShift)
    {
        alignmentMask = ((1 << arg_AlignmentShift) - 1);
    }

#if defined(_IX_RM_MULTI_THREADED_)

    /* lock the memory manager data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    /* start from the rover location in the list */
    pActiveCell = pRover = arg_pMemoryManager->m_pRover;

    /* Search at the end of block list first */
    while(pActiveCell != 0)
    {
        /* calculate the size adjustment due to the required alignment */
        switch(arg_AlignmentType)
        {
            case IX_MEMORY_ALIGNMENT_TYPE_PHYSICAL:
                temp = ((ix_uint32)pActiveCell->m_pData - (ix_uint32)arg_pMemoryManager->m_pStartAddress) +
                    (ix_uint32)arg_pMemoryManager->m_pChannelPhysicalAddress;
                break;
            case IX_MEMORY_ALIGNMENT_TYPE_PHYSICAL_OFFSET:
                temp = ((ix_uint32)pActiveCell->m_pData - (ix_uint32)arg_pMemoryManager->m_pStartAddress) +
                    (ix_uint32)arg_pMemoryManager->m_ChannelPhysicalOffset;
                break;

#if (_IX_HARDWARE_TYPE_ == _IX_HW_WESTPORT_) 

            case IX_MEMORY_ALIGNMENT_TYPE_PHYSICAL_NON_ALIAS:
                temp = ((ix_uint32)pActiveCell->m_pData - (ix_uint32)arg_pMemoryManager->m_pStartAddress) +
                    (ix_uint32)arg_pMemoryManager->m_pNonAliasChannelPhysicalAddress;
                break;
            case IX_MEMORY_ALIGNMENT_TYPE_PHYSICAL_OFFSET_NON_ALIAS:
                temp = ((ix_uint32)pActiveCell->m_pData - (ix_uint32)arg_pMemoryManager->m_pStartAddress) +
                    (ix_uint32)arg_pMemoryManager->m_NonAliasChannelPhysicalOffset;
                break;

#endif /* (_IX_HARDWARE_TYPE_ == _IX_HW_WESTPORT_) */

            default:
                temp = (ix_uint32)pActiveCell->m_pData;
                break;
        }
        sizeAdjustment = temp & alignmentMask;
        if(sizeAdjustment != 0)
            sizeAdjustment = (1 << arg_AlignmentShift) - sizeAdjustment;

        /* check if the size of the block is large enough */
        if(pActiveCell->m_Size < arg_Size + sizeAdjustment)
            pActiveCell = pActiveCell->m_pNextCell;
        else
        {
            register _ix_mm_memory_block_cell* pNewCell = 0;

            /* if there is some size adjustment then split the cell */
            if(sizeAdjustment != 0)
            {
                /* split the cell and make the pActiveCell of the size sizeAdjustment */
                pNewCell = _ix_rm_mm_split_cell(arg_pMemoryManager, pActiveCell, sizeAdjustment);

                /* the cell split failed */
                if(pNewCell == 0)
                {
                    /* set the memory address to 0 */
                    *arg_pMemory = 0;

                    /* set the error code */
                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_CELL_SPLIT_FAILED, s_aFunctionIdentifier, 
                        "memory cell split failed!", err);

                    /* done with the allocation process */
                    goto allocation_done;
                }
                else
                {
                    /* set the newly created cell as the active cell */
                    pActiveCell = pNewCell;

                }
            }
            /* The size matches exactly.*/
            if(pActiveCell->m_Size == arg_Size)
            {
                /* allocate the whole cell */

                /* set the cell to be in use */
                pActiveCell->m_Size = 0;

                /* set the memory address to this cell's data pointer */
                *arg_pMemory = pActiveCell->m_pData;

                /* reset the rover */
                if(pRover == pActiveCell)
                    arg_pMemoryManager->m_pRover = arg_pMemoryManager->m_pBlockList; 

                /* done with the allocation process */
                goto allocation_done;
            }
            else
            {
                /* our size is larger so we split the cell */

                /* split the current cell and make pActiveCell of size arg_Size */
                pNewCell = _ix_rm_mm_split_cell(arg_pMemoryManager, pActiveCell, arg_Size);

                /* we could not split the cell */
                if(pNewCell == 0)
                {
                    /* set the memory address to 0 */
                    *arg_pMemory = 0;

                    /* set the error code */
                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_CELL_SPLIT_FAILED, s_aFunctionIdentifier, 
                        "memory cell split failed!", err);

                    /* done with the allocation process */
                    goto allocation_done;
                }
                else
                {
                    /* mark the new cell as being in use */
                    pActiveCell->m_Size = 0;

                    /* set the new cell's data pointer and the returned address to this value */
                    *arg_pMemory = pActiveCell->m_pData;

                    /* reset the rover */
                    arg_pMemoryManager->m_pRover = pNewCell; 

                    /* done with the allocation process */
                    goto allocation_done;
                }
            }
        }
    }

    /* we did not find any block that will fit our size */

    /* start from the beginning in the list */
    pActiveCell = arg_pMemoryManager->m_pBlockList;

    while(pActiveCell != pRover)
    {
        /* calculate the size adjustment due to the required alignment */
        switch(arg_AlignmentType)
        {
            case IX_MEMORY_ALIGNMENT_TYPE_PHYSICAL:
                temp = ((ix_uint32)pActiveCell->m_pData - (ix_uint32)arg_pMemoryManager->m_pStartAddress) +
                    (ix_uint32)arg_pMemoryManager->m_pChannelPhysicalAddress;
                break;
            case IX_MEMORY_ALIGNMENT_TYPE_PHYSICAL_OFFSET:
                temp = ((ix_uint32)pActiveCell->m_pData - (ix_uint32)arg_pMemoryManager->m_pStartAddress) +
                    (ix_uint32)arg_pMemoryManager->m_ChannelPhysicalOffset;
                break;

#if (_IX_HARDWARE_TYPE_ == _IX_HW_WESTPORT_) 

            case IX_MEMORY_ALIGNMENT_TYPE_PHYSICAL_NON_ALIAS:
                temp = ((ix_uint32)pActiveCell->m_pData - (ix_uint32)arg_pMemoryManager->m_pStartAddress) +
                    (ix_uint32)arg_pMemoryManager->m_pNonAliasChannelPhysicalAddress;
                break;
            case IX_MEMORY_ALIGNMENT_TYPE_PHYSICAL_OFFSET_NON_ALIAS:
                temp = ((ix_uint32)pActiveCell->m_pData - (ix_uint32)arg_pMemoryManager->m_pStartAddress) +
                    (ix_uint32)arg_pMemoryManager->m_NonAliasChannelPhysicalOffset;
                break;

#endif /* (_IX_HARDWARE_TYPE_ == _IX_HW_WESTPORT_) */

            default:
                temp = (ix_uint32)pActiveCell->m_pData;
                break;
        }
        sizeAdjustment = temp & alignmentMask;
        if(sizeAdjustment != 0)
            sizeAdjustment = (1 << arg_AlignmentShift) - sizeAdjustment;

        /* check if the size of the block is large enough */
        if(pActiveCell->m_Size < arg_Size + sizeAdjustment)
            pActiveCell = pActiveCell->m_pNextCell;
        else
        {
            register _ix_mm_memory_block_cell* pNewCell = 0;

            /* if there is some size adjustment then split the cell */
            if(sizeAdjustment != 0)
            {
                /* split the cell and make the pActiveCell of the size sizeAdjustment */
                pNewCell = _ix_rm_mm_split_cell(arg_pMemoryManager, pActiveCell, sizeAdjustment);

                /* the cell split failed */
                if(pNewCell == 0)
                {
                    /* set the memory address to 0 */
                    *arg_pMemory = 0;

                    /* set the error code */
                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_CELL_SPLIT_FAILED, s_aFunctionIdentifier, 
                        "memory cell split failed!", err);

                    /* done with the allocation process */
                    goto allocation_done;
                }
                else
                {
                    /* set the newly created cell as the active cell */
                    pActiveCell = pNewCell;

                }
            }
            /* The size matches exactly.*/
            if(pActiveCell->m_Size == arg_Size)
            {
                /* allocate the whole cell */

                /* set the cell to be in use */
                pActiveCell->m_Size = 0;

                /* set the memory address to this cell's data pointer */
                *arg_pMemory = pActiveCell->m_pData;

                /* done with the allocation process */
                goto allocation_done;
            }
            else
            {
                /* our size is larger so we split the cell */

                /* split the current cell and make pActiveCell of size arg_Size */
                pNewCell = _ix_rm_mm_split_cell(arg_pMemoryManager, pActiveCell, arg_Size);

                /* we could not split the cell */
                if(pNewCell == 0)
                {
                    /* set the memory address to 0 */
                    *arg_pMemory = 0;

                    /* set the error code */
                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_CELL_SPLIT_FAILED, s_aFunctionIdentifier, 
                        "memory cell split failed!", err);

                    /* done with the allocation process */
                    goto allocation_done;
                }
                else
                {
                    /* mark the new cell as being in use */
                    pActiveCell->m_Size = 0;

                    /* set the new cell's data pointer and the returned address to this value */
                    *arg_pMemory = pActiveCell->m_pData;

                    /* reset the rover */
                    arg_pMemoryManager->m_pRover = pNewCell; 

                    /* done with the allocation process */
                    goto allocation_done;
                }
            }
        }
    }
#ifdef _IX_RM_DEBUG_
    /*
     * Don't print any message in case of empty optional memory channel
     */
    if (arg_pMemoryManager->m_MemoryBlockSize == 0)
    {
        err = IX_ERROR_NEW(IX_RM_ERROR_CANNOT_ALLOCATE_SIZE,
                           IX_ERROR_LEVEL_LOCAL);
    }
    else
#endif
    err = IX_RM_ERROR_NEW(IX_RM_ERROR_CANNOT_ALLOCATE_SIZE, s_aFunctionIdentifier, 
        "the requested size can not be allocated!", err);


allocation_done:

#if defined(_IX_RM_MULTI_THREADED_)

    /* unlock the memory manager data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    /* return the error code */
    return err;
}

/**
 * NAME: _ix_rm_mm_deallocator
 *
 * DESCRIPTION: This function will deallocate previously allocated memory. It is generic 
 *          deallocator that can be used for all memory types regardless alignment. If 
 *          the that address is passed is not found in the used list then nothing will happen and an error will be
 *          returned.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager data this allocator works on.
 * @Param:  - IN arg_pMemory - the memory block that will be released.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_deallocator( 
                                _ix_mm_memory_manager* arg_pMemoryManager, 
                                void* arg_pMemory 
                              )

{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mm_deallocator";

    ix_error err = IX_SUCCESS;

    register _ix_mm_memory_block_cell* pActiveCell = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &arg_pMemoryManager->m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_MULTI_THREADED_)

    /* lock the memory manager data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    /* start at the beginning of the list */
    pActiveCell = arg_pMemoryManager->m_pBlockList;

    /* look for a matching cell */
    while(pActiveCell != 0)
    {
        /* check if passed over our address. The list is ordered based on data address. */
        if((void*)pActiveCell->m_pData < arg_pMemory)
        {
            /* go to the next cell */
            pActiveCell = pActiveCell->m_pNextCell;
        }
        else if(pActiveCell->m_pData == arg_pMemory)
        {
            /* we found the cell */

            /* check if the cell has not been already released */
            if(pActiveCell->m_Size != 0)
            {
                /* set the error code */
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_DATA_ALREADY_RELEASED, s_aFunctionIdentifier, 
                        "the memory you are trying to free has been freed before!", err);

                /* done with the deallocation process */
                break;
            }
            else
            {
                register _ix_mm_memory_block_cell* a_pCell;

                a_pCell = pActiveCell->m_pNextCell;

                if(a_pCell)
                {
                    /* set the size for the active cell */
                    pActiveCell->m_Size = ((ix_uint8*)a_pCell->m_pData) - ((ix_uint8*)pActiveCell->m_pData);

                    /* see if we can link with the next cell */
                    if(a_pCell->m_Size)
                    {
                        /* are we at the end of the list? */
                        if(a_pCell->m_pNextCell)
                        {
                            /* set active's cell next cell */
                            pActiveCell->m_pNextCell = a_pCell->m_pNextCell;
                            /* set next cell's previous cell */
                            pActiveCell->m_pNextCell->m_pPrevCell = pActiveCell;
                        }
                        else
                        {
                            /* we are the end, so set the next cell to 0 */
                            pActiveCell->m_pNextCell = 0;
                        }

                        /* set the size for the entire merged cell */
                        pActiveCell->m_Size += a_pCell->m_Size;

                        /* reset the rover */
                        if(a_pCell == arg_pMemoryManager->m_pRover)
                            arg_pMemoryManager->m_pRover = pActiveCell;

                        /* release the next cell */
                        _ix_rm_mm_release_cell(arg_pMemoryManager, a_pCell);
                    }
                }
                else
                {
                    /* set the size for the active cell */
                    pActiveCell->m_Size = ((ix_uint8*)arg_pMemoryManager->m_pEndAddress) - ((ix_uint8*)pActiveCell->m_pData);
                }

                /* see if we can merge with the previous cell */
                a_pCell = pActiveCell->m_pPrevCell;

                /* check if the previous cell is a free one */
                if(a_pCell && a_pCell->m_Size)
                {
                    /* we merge with the previous cell */
                    
                    if(pActiveCell->m_pNextCell)
                    {
                        /* set previous's cell next cell */
                        a_pCell->m_pNextCell = pActiveCell->m_pNextCell;
                        /* set next cell previous cell */
                        a_pCell->m_pNextCell->m_pPrevCell = a_pCell;
                    }
                    else
                    {
                        /* this is the last cell in the list */
                        a_pCell->m_pNextCell = 0;
                    }

                    /* set the size for the entire merged cell */
                    a_pCell->m_Size += pActiveCell->m_Size;

                    /* reset the rover */
                    if(pActiveCell == arg_pMemoryManager->m_pRover)
                        arg_pMemoryManager->m_pRover = a_pCell;

                    /* release the active cell */
                    _ix_rm_mm_release_cell(arg_pMemoryManager, pActiveCell);

                }
                
            }

            /* done with the deallocation process */
            break;
        }
        else
        {
            /**
             * This means that we passed our location and we did not find the 
             * memory location. Based on the fact that the list is sorted then
             * we cancel the the serch.
             */

            /* set the error code */
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_BAD_MEMORY_ADDRESS, s_aFunctionIdentifier, 
                        "the memory you are tying to free has an invalid pointer!", err);

            /* done with the deallocation process */
            break;

        }
    }

#if defined(_IX_RM_MULTI_THREADED_)

    /* unlock the memory manager data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    /* return the error code */
    return err;
}

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * NAME: _ix_rm_mm_reserve
 *
 * DESCRIPTION: This is a memory reservation function for all types of memory.
 *          For special needs different reservation functions will be defined, but this is the generic one.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager data this allocator works on.
 * @Param:  - IN arg_Size - the size of the memory block requested.
 * @Param:  - INOUT arg_pByteOffset - an address to byte offset in the memory block where we want to 
 *          reserve memory. On output this offset will be modified to accomodate the required alignment.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure. If any of the requested area is 
 *          in use then an error will be returned. Caution about the fact that memory 
 *          allocation and reservation will consider the alignment and will adjust the
 *          size and start address in accordance with the default alignment.
 *          
 */
ix_error _ix_rm_mm_reserve( 
                            _ix_mm_memory_manager* arg_pMemoryManager, 
                            ix_uint32 arg_Size,
                            ix_uint32* arg_pByteOffset 
                          )

#else /* Westport hardware */

/**
 * NAME: _ix_rm_mm_reserve
 *
 * DESCRIPTION: This is a memory reservation function for all types of memory.
 *          For special needs different reservation functions will be defined, but this is the generic one.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager data this allocator works on.
 * @Param:  - IN arg_Size - the size of the memory block requested.
 * @Param:  - IN arg_OffsetType - represents the type of the specified offset 
 * @Param:  - INOUT arg_pByteOffset - an address to byte offset in the memory block where we want to 
 *          reserve memory. On output this offset will be modified to accomodate the required alignment.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure. If any of the requested area is 
 *          in use then an error will be returned. Caution about the fact that memory 
 *          allocation and reservation will consider the alignment and will adjust the
 *          size and start address in accordance with the default alignment.
 *          
 */
ix_error _ix_rm_mm_reserve( 
                            _ix_mm_memory_manager* arg_pMemoryManager, 
                            ix_uint32 arg_Size,
                            ix_memory_offset_type arg_OffsetType, 
                            ix_uint32* arg_pByteOffset 
                          )

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mm_reserve";

    ix_error err = IX_SUCCESS;
    register _ix_mm_memory_block_cell* pActiveCell = 0;
    register ix_uint32* pData;
    ix_uint32* pDataTemp;
    ix_uint32 alignmentMask;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &arg_pMemoryManager->m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */


    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* calculate the required alignment */
    alignmentMask = ((1 << arg_pMemoryManager->m_DefaultAlignmentShift) - 1);

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

    /* set the data pointer of the requested memory */
    pData = (ix_uint32*)((ix_uint8*)arg_pMemoryManager->m_pStartAddress + *arg_pByteOffset -
        arg_pMemoryManager->m_ChannelPhysicalOffset);

#else /* Westport hardware */

    /* set the data pointer of the requested memory */
    if(arg_OffsetType == IX_MEMORY_OFFSET_TYPE_ALIASED)
    {
        pData = (ix_uint32*)((ix_uint8*)arg_pMemoryManager->m_pStartAddress + *arg_pByteOffset -
            arg_pMemoryManager->m_ChannelPhysicalOffset);
    }
    else
    {
        pData = (ix_uint32*)((ix_uint8*)arg_pMemoryManager->m_pStartAddress + *arg_pByteOffset -
            arg_pMemoryManager->m_NonAliasChannelPhysicalOffset);
    }

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

	
    pDataTemp = (ix_uint32*)((ix_uint8*)pData + arg_Size);

    if((pDataTemp >= arg_pMemoryManager->m_pEndAddress) || (pDataTemp < arg_pMemoryManager->m_pStartAddress))
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_CANNOT_RESERVE_MEMORY, s_aFunctionIdentifier,
                                "memory reservation failed!", err);
    }
    
    /* adjust the size according to the required alignment */
    arg_Size += ((ix_uint32)pData & alignmentMask);

    /* adjust the start address according to the required alignment */
    pData = (ix_uint32*)((ix_uint32)pData & (~alignmentMask));

    if((pData < arg_pMemoryManager->m_pStartAddress) || (pData >= arg_pMemoryManager->m_pEndAddress))
        return IX_RM_ERROR_NEW(IX_RM_ERROR_MEMORY_BAD_OFFSET, s_aFunctionIdentifier, 
                        "the passed offset is invalid!", err);

    /* the block has to be aligned at the end */
    arg_Size += alignmentMask;
    arg_Size &= ~alignmentMask;

    /* adjust the byte offset according to the alignment */
    *arg_pByteOffset = (ix_uint8*)pData - (ix_uint8*)arg_pMemoryManager->m_pStartAddress;

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

    *arg_pByteOffset += arg_pMemoryManager->m_ChannelPhysicalOffset;

#else /* Westport hardware */

    if(arg_OffsetType == IX_MEMORY_OFFSET_TYPE_ALIASED)
    {
        *arg_pByteOffset += arg_pMemoryManager->m_ChannelPhysicalOffset;
    }
    else
    {
        *arg_pByteOffset += arg_pMemoryManager->m_NonAliasChannelPhysicalOffset;
    }
    
#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */


#if defined(_IX_RM_MULTI_THREADED_)

    /* lock the memory manager data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    /* search for a cell that contains this data pointer */
    pActiveCell = arg_pMemoryManager->m_pBlockList;
    while(pActiveCell != 0)
    {
        if(pData >= pActiveCell->m_pData)
        {
            if((pActiveCell->m_Size == 0) || (ix_uint8*)pData >= (ix_uint8*)pActiveCell->m_pData + pActiveCell->m_Size)
            {
                /* go to the next cell */
                pActiveCell = pActiveCell->m_pNextCell;
                continue;
            }

            /* found a location where requested memory block might fit */

            /* check if the requested memory will fit into the block */
            if((((ix_uint8*)pData) + arg_Size) <= (((ix_uint8*)pActiveCell->m_pData) + pActiveCell->m_Size))
            {
                
                /**
                 * Here we have three possibilities:
                 *      the requested area is at the beginning of the current block;
                 *      the requested area is in the middle of the current block;
                 *      the requested area is at the end of the current block;
                 */

                /* check if we are the beginning of the current block */
                if(pData == pActiveCell->m_pData)
                {
                    /* check if the sizes are different */
                    if(arg_Size == pActiveCell->m_Size)
                    {
                        /* reserve the whole block */

                        /* set the cell as being in use */
                        pActiveCell->m_Size = 0;
                        break;
                    }
                    else
                    {
                        /* split the cell in 2 and mark the first one as being in use */
                        register _ix_mm_memory_block_cell* pNewCell = 0;

                        /* split the current cell and make pActiveCell of size arg_Size */
                        pNewCell = _ix_rm_mm_split_cell(arg_pMemoryManager, pActiveCell, arg_Size);

                        /* we could not split the cell */
                        if(pNewCell == 0)
                        {
                            /* set the error code */
                            err = IX_RM_ERROR_NEW(IX_RM_ERROR_CELL_SPLIT_FAILED, s_aFunctionIdentifier, 
                                "memory cell split failed!", err);

                            /* done with the reservation process */
                            break;
                        }
                        else
                        {
                            /* mark the new cell as being in use */
                            pActiveCell->m_Size = 0;

                            /* reset the rover */
                            if(pActiveCell == arg_pMemoryManager->m_pRover)
                                arg_pMemoryManager->m_pRover = pNewCell;

                            /* done with the reservation process */
                            break;
                        }
                    }
                }
                else
                {
                    /* we are in the middle or at the end of the cell */

                    /* split the cell in 2 */
                    register _ix_mm_memory_block_cell* pNewCell = 0;

                    /* check if we are at the end of the current block */
                    if((((ix_uint8*)pData) + arg_Size) != (((ix_uint8*)pActiveCell->m_pData) + pActiveCell->m_Size))
                    {
                        /* we are in the middle */

                        /* first split the cell in two */
                        
                        /**
                         * first split the cell in two such that the requested block will be at the end
                         * of the active cell
                         */

                        pNewCell = _ix_rm_mm_split_cell(arg_pMemoryManager, pActiveCell, arg_Size + 
                            (((ix_uint8*)pData) - ((ix_uint8*)pActiveCell->m_pData)));

                        /* we could not split the cell */
                        if(pNewCell == 0)
                        {
                            /* set the error code */
                            err = IX_RM_ERROR_NEW(IX_RM_ERROR_CELL_SPLIT_FAILED, s_aFunctionIdentifier, 
                                "memory cell split failed!", err);

                            /* done with the reservation process */
                            break;
                        }

                    }

                    /* here we are the end of the active cell */

                    /* there is no need to see if we have to reserve the whole block */

                    /* split the cell in 2 and mark the second one as being in use */

                    pNewCell = _ix_rm_mm_split_cell(arg_pMemoryManager, pActiveCell,
                            (((ix_uint8*)pData) - ((ix_uint8*)pActiveCell->m_pData)));

                    /* we could not split the cell */
                    if(pNewCell == 0)
                    {
                        /* set the error code */
                        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CELL_SPLIT_FAILED, s_aFunctionIdentifier, 
                            "memory cell split failed!", err);

                        /* done with the reservation process */
                        break;
                    }
                    else
                    {
                        /* mark the new cell as being in use */
                        pNewCell->m_Size = 0;

                        /* done with the reservation process */
                        break;
                    }          
                }

            }
            else
            {
                /* the memory is either in use or does not have the required size */
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_MEMORY_IN_USE, s_aFunctionIdentifier, 
                        "the memory you are tying to reserve is partially in use!", err);
                break;
            }

        }
        else
        {
            /* we already passed all cells that might contain this offset */
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_MEMORY_IN_USE, s_aFunctionIdentifier, 
                    "the memory you are tying to reserve is partially in use!", err);
            break;
        }
    }

#if defined(_IX_RM_MULTI_THREADED_)

    /* unlock the memory manager data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */
  
    return err;
}


#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * NAME: _ix_rm_mm_aligned_reserve
 *
 * DESCRIPTION: This is an aligned memory reservation function for all types of memory.
 *          For special needs different aligned reservation functions will be defined, but 
 *          this is the generic one.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager data this allocator works on.
 * @Param:  - IN arg_Size - the size of the memory block requested.
 * @Param:  - IN arg_AlignmentType - specifies the type of alignment required as explained at 
 *          ix_memory_alignment_type type description
 * @Param:  - IN arg_AlignmentShift - this value specifies the required memory block alignment as the
 *          power of 2 of the alignment. If the specified alignment is less than the
 *          default alignment for the memory pool manager then the default alignment will be used.
 * @Param:  - INOUT arg_pByteOffset - an address to byte offset in the memory block where we want to 
 *          reserve memory. On output this offset will be modified to accomodate the required alignment.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure. If any of the requested area is 
 *          in use then an error will be returned. Caution about the fact that memory 
 *          allocation and reservation will consider the allignment and will adjust the
 *          size and start address in accordance with the alignment.
 *          
 */
ix_error _ix_rm_mm_aligned_reserve( 
                                    _ix_mm_memory_manager* arg_pMemoryManager, 
                                    ix_uint32 arg_Size,
                                    ix_memory_alignment_type arg_AlignmentType,
                                    ix_uint32 arg_AlignmentShift,
                                    ix_uint32* arg_pByteOffset 
                                  )

#else /* Westport hardware */

/**
 * NAME: _ix_rm_mm_aligned_reserve
 *
 * DESCRIPTION: This is an aligned memory reservation function for all types of memory.
 *          For special needs different aligned reservation functions will be defined, but 
 *          this is the generic one.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager data this allocator works on.
 * @Param:  - IN arg_Size - the size of the memory block requested.
 * @Param:  - IN arg_AlignmentType - specifies the type of alignment required as explained at 
 *          ix_memory_alignment_type type description
 * @Param:  - IN arg_AlignmentShift - this value specifies the required memory block alignment as the
 *          power of 2 of the alignment. If the specified alignment is less than the
 *          default alignment for the memory pool manager then the default alignment will be used.
 * @Param:  - IN arg_OffsetType - represents the type of the specified offset 
 * @Param:  - INOUT arg_pByteOffset - an address to byte offset in the memory block where we want to 
 *          reserve memory. On output this offset will be modified to accomodate the required alignment.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure. If any of the requested area is 
 *          in use then an error will be returned. Caution about the fact that memory 
 *          allocation and reservation will consider the allignment and will adjust the
 *          size and start address in accordance with the alignment.
 *          
 */
ix_error _ix_rm_mm_aligned_reserve( 
                                    _ix_mm_memory_manager* arg_pMemoryManager, 
                                    ix_uint32 arg_Size,
                                    ix_memory_alignment_type arg_AlignmentType,
                                    ix_uint32 arg_AlignmentShift,
                                    ix_memory_offset_type arg_OffsetType, 
                                    ix_uint32* arg_pByteOffset 
                                  )

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mm_reserve";

    ix_error err = IX_SUCCESS;
    register _ix_mm_memory_block_cell* pActiveCell = 0;
    register ix_uint32* pData;
    ix_uint32* pDataTemp;
    ix_uint32 alignmentMask;
    ix_uint32 temp;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &arg_pMemoryManager->m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */


    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* adjust the alignment shift to be minimum the default alignment */
    if(arg_AlignmentShift < arg_pMemoryManager->m_DefaultAlignmentShift)
    {
        arg_AlignmentShift = arg_pMemoryManager->m_DefaultAlignmentShift;
    }

    /* calculate the required alignment */
    alignmentMask = ((1 << arg_AlignmentShift) - 1);

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

    /* set the data pointer of the requested memory */
    pData = (ix_uint32*)(((ix_uint8*)arg_pMemoryManager->m_pStartAddress) + (ix_uint32)*arg_pByteOffset -
        (ix_uint32)arg_pMemoryManager->m_ChannelPhysicalOffset);

#else /* Westport hardware */

        /* set the data pointer of the requested memory */
    if(arg_OffsetType == IX_MEMORY_OFFSET_TYPE_ALIASED)
    {
        pData = (ix_uint32*)((ix_uint8*)arg_pMemoryManager->m_pStartAddress + *arg_pByteOffset -
            arg_pMemoryManager->m_ChannelPhysicalOffset);
    }
    else
    {
        pData = (ix_uint32*)((ix_uint8*)arg_pMemoryManager->m_pStartAddress + *arg_pByteOffset -
            arg_pMemoryManager->m_NonAliasChannelPhysicalOffset);
    }

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

    pDataTemp = (ix_uint32*)((ix_uint8*)pData + arg_Size);
    if((pDataTemp >= arg_pMemoryManager->m_pEndAddress) || (pDataTemp < arg_pMemoryManager->m_pStartAddress))
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_CANNOT_RESERVE_MEMORY, s_aFunctionIdentifier, 
                                "memory reservation failed!", err);
    }

    /* calculate the size adjustment due to the required alignment */
    switch(arg_AlignmentType)
    {
        case IX_MEMORY_ALIGNMENT_TYPE_PHYSICAL:
            temp = ((ix_uint32)pData - (ix_uint32)arg_pMemoryManager->m_pStartAddress) + 
                (ix_uint32)arg_pMemoryManager->m_pChannelPhysicalAddress;
            break;
        case IX_MEMORY_ALIGNMENT_TYPE_PHYSICAL_OFFSET:
            temp = ((ix_uint32)pData - (ix_uint32)arg_pMemoryManager->m_pStartAddress) + 
                (ix_uint32)arg_pMemoryManager->m_ChannelPhysicalOffset;
            break;

#if (_IX_HARDWARE_TYPE_ == _IX_HW_WESTPORT_) 

        case IX_MEMORY_ALIGNMENT_TYPE_PHYSICAL_NON_ALIAS:
            temp = ((ix_uint32)pData - (ix_uint32)arg_pMemoryManager->m_pStartAddress) +
                (ix_uint32)arg_pMemoryManager->m_pNonAliasChannelPhysicalAddress;
            break;
        case IX_MEMORY_ALIGNMENT_TYPE_PHYSICAL_OFFSET_NON_ALIAS:
            temp = ((ix_uint32)pData - (ix_uint32)arg_pMemoryManager->m_pStartAddress) +
                (ix_uint32)arg_pMemoryManager->m_NonAliasChannelPhysicalOffset;
            break;

#endif /* (_IX_HARDWARE_TYPE_ == _IX_HW_WESTPORT_) */

        default:
            temp =(ix_uint32)pData;
            break;
    }

    /* adjust the size according to the required alignment */
    arg_Size += (temp & alignmentMask);

    /* adjust the start address according to the required alignment */
    pData = (ix_uint32*)((ix_uint32)pData - (temp & alignmentMask));

    if((pData < arg_pMemoryManager->m_pStartAddress) || (pData >= arg_pMemoryManager->m_pEndAddress))
        return IX_RM_ERROR_NEW(IX_RM_ERROR_MEMORY_BAD_OFFSET, s_aFunctionIdentifier, 
                        "the passed offset is invalid!", err);

    /* calculate the default alignment */
    alignmentMask = ((1 << arg_pMemoryManager->m_DefaultAlignmentShift) - 1);

    /* the block has to be aligned at the end */
    arg_Size += alignmentMask;
    arg_Size &= ~alignmentMask;

    /* adjust the byte offset according to the alignment */
    *arg_pByteOffset = (ix_uint8*)pData - (ix_uint8*)arg_pMemoryManager->m_pStartAddress;

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

    *arg_pByteOffset += arg_pMemoryManager->m_ChannelPhysicalOffset;

#else /* Westport hardware */

    if(arg_OffsetType == IX_MEMORY_OFFSET_TYPE_ALIASED)
    {
        *arg_pByteOffset += arg_pMemoryManager->m_ChannelPhysicalOffset;
    }
    else
    {
        *arg_pByteOffset += arg_pMemoryManager->m_NonAliasChannelPhysicalOffset;
    }
    
#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

#if defined(_IX_RM_MULTI_THREADED_)

    /* lock the memory manager data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    /* search for a cell that contains this data pointer */
    pActiveCell = arg_pMemoryManager->m_pBlockList;
    while(pActiveCell != 0)
    {
        if(pData >= pActiveCell->m_pData)
        {
            if((pActiveCell->m_Size == 0) || (ix_uint8*)pData >= (ix_uint8*)pActiveCell->m_pData + pActiveCell->m_Size)
            {
                /* go to the next cell */
                pActiveCell = pActiveCell->m_pNextCell;
                continue;
            }

            /* found a location where requested memory block might fit */

            /* check if the requested memory will fit into the block */
            if((((ix_uint8*)pData) + arg_Size) <= (((ix_uint8*)pActiveCell->m_pData) + pActiveCell->m_Size))
            {
                
                /**
                 * Here we have three possibilities:
                 *      the requested area is at the beginning of the current block;
                 *      the requested area is in the middle of the current block;
                 *      the requested area is at the end of the current block;
                 */

                /* check if we are the beginning of the current block */
                if(pData == pActiveCell->m_pData)
                {
                    /* check if the sizes are different */
                    if(arg_Size == pActiveCell->m_Size)
                    {
                        /* reserve the whole block */

                        /* set the cell as being in use */
                        pActiveCell->m_Size = 0;
                        break;
                    }
                    else
                    {
                        /* split the cell in 2 and mark the first one as being in use */
                        register _ix_mm_memory_block_cell* pNewCell = 0;

                        /* split the current cell and make pActiveCell of size arg_Size */
                        pNewCell = _ix_rm_mm_split_cell(arg_pMemoryManager, pActiveCell, arg_Size);

                        /* we could not split the cell */
                        if(pNewCell == 0)
                        {
                            /* set the error code */
                            err = IX_RM_ERROR_NEW(IX_RM_ERROR_CELL_SPLIT_FAILED, s_aFunctionIdentifier, 
                                "memory cell split failed!", err);

                            /* done with the reservation process */
                            break;
                        }
                        else
                        {
                            /* mark the new cell as being in use */
                            pActiveCell->m_Size = 0;

                            /* reset the rover */
                            if(pActiveCell == arg_pMemoryManager->m_pRover)
                                arg_pMemoryManager->m_pRover = pNewCell;

                            /* done with the reservation process */
                            break;
                        }
                    }
                }
                else
                {
                    /* we are in the middle or at the end of the cell */

                    /* split the cell in 2 */
                    register _ix_mm_memory_block_cell* pNewCell = 0;

                    /* check if we are at the end of the current block */
                    if((((ix_uint8*)pData) + arg_Size) != (((ix_uint8*)pActiveCell->m_pData) + pActiveCell->m_Size))
                    {
                        /* we are in the middle */

                        /* first split the cell in two */
                        
                        /**
                         * first split the cell in two such that the requested block will be at the end
                         * of the active cell
                         */

                        pNewCell = _ix_rm_mm_split_cell(arg_pMemoryManager, pActiveCell, arg_Size + 
                            (((ix_uint8*)pData) - ((ix_uint8*)pActiveCell->m_pData)));

                        /* we could not split the cell */
                        if(pNewCell == 0)
                        {
                            /* set the error code */
                            err = IX_RM_ERROR_NEW(IX_RM_ERROR_CELL_SPLIT_FAILED, s_aFunctionIdentifier, 
                                "memory cell split failed!", err);

                            /* done with the reservation process */
                            break;
                        }

                    }

                    /* here we are the end of the active cell */

                    /* there is no need to see if we have to reserve the whole block */

                    /* split the cell in 2 and mark the second one as being in use */

                    pNewCell = _ix_rm_mm_split_cell(arg_pMemoryManager, pActiveCell,
                            (((ix_uint8*)pData) - ((ix_uint8*)pActiveCell->m_pData)));

                    /* we could not split the cell */
                    if(pNewCell == 0)
                    {
                        /* set the error code */
                        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CELL_SPLIT_FAILED, s_aFunctionIdentifier, 
                            "memory cell split failed!", err);

                        /* done with the reservation process */
                        break;
                    }
                    else
                    {
                        /* mark the new cell as being in use */
                        pNewCell->m_Size = 0;

                        /* done with the reservation process */
                        break;
                    }          
                }

            }
            else
            {
                /* the memory is either in use or does not have the required size */
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_MEMORY_IN_USE, s_aFunctionIdentifier, 
                        "the memory you are tying to reserve is partially in use!", err);
                break;
            }

        }
        else
        {
            /* we already passed all cells that might contain this offset */
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_MEMORY_IN_USE, s_aFunctionIdentifier, 
                    "the memory you are tying to reserve is partially in use!", err);
            break;
        }
    }

#if defined(_IX_RM_MULTI_THREADED_)

    /* unlock the memory manager data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */
  
    return err;
}


/**
 * NAME: _ix_rm_mm_initializer
 *
 * DESCRIPTION: This function provides all the required initialization for the memory manager
 *          that is passed as an argument.There are some assumptions that this function 
 *          makes:  the type of memory should be set before calling the function;
 *                  the channel, or microengine number should be set before calling the function;
 *                  the start address should be set prior calling the function;
 *                  the size of controlled memory should be set prior calling the function;
 * 
 * @Param:  - IN arg_pMemoryManager - the address of the memory manager that will be initialized.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure. Upon return the referred memory manager
 *          is fully initialized in the case the call is successful.
 */
ix_error _ix_rm_mm_initializer( 
                                _ix_mm_memory_manager* arg_pMemoryManager 
                              )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mm_initializer";

    ix_error err = IX_SUCCESS;
    ix_uint32 alignmentMask;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &arg_pMemoryManager->m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */


    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_MULTI_THREADED_)

    /* lock the memory manager data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    alignmentMask = ((1 << arg_pMemoryManager->m_DefaultAlignmentShift) - 1);

    /* check proper alignment */
    IX_RM_ASSERT(!((ix_uint32)arg_pMemoryManager->m_pStartAddress & alignmentMask), s_aFunctionIdentifier, "bad start virtual address alignment!");

#if defined(IX_PLATFORM_2401) || defined(IX_PLATFORM_2801)
    /* Platform specific
     * On IXDP2x01 platforms some SRAM channels are optional and HAL
     * returns -1 as physical address for such a channel. However,
     * Memory Manager needs to perform its initialization even for
     * empty channel (otherwise it generates address faults later).
     */
#else

    IX_RM_ASSERT(!((ix_uint32)arg_pMemoryManager->m_pChannelPhysicalAddress & alignmentMask), s_aFunctionIdentifier, "bad start physical memory channel address alignment!");
    IX_RM_ASSERT(!((ix_uint32)arg_pMemoryManager->m_ChannelPhysicalOffset & alignmentMask), s_aFunctionIdentifier, "bad physical offset alignment!");

#endif /* defined(IX_PLATFORM_2401) || defined(IX_PLATFORM_2801) */
    
    /* set the end memory address */
    arg_pMemoryManager->m_pEndAddress = (ix_uint32*)((ix_uint8*)arg_pMemoryManager->m_pStartAddress + arg_pMemoryManager->m_MemoryBlockSize);


    /* initialize the recycled block cell list and counter */
    arg_pMemoryManager->m_RecycledBlockCellCount = 0;
    arg_pMemoryManager->m_pRecycledBlockCellList = 0;

    /* initialize the block list */

    /* allocate a new cell */
    arg_pMemoryManager->m_pBlockList = _ix_rm_mm_new_cell(arg_pMemoryManager);

    /* check if the allocation succeeded */
    if(arg_pMemoryManager->m_pBlockList == 0)
    {
        /* we failed to allocate a new cell */

        /* set the error code */
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_CANNOT_CREATE_CELL, s_aFunctionIdentifier, 
                        "memory cell allocation failed when trying to initialize the memory manager!", err);
    }
    else
    {
        /* initialize the block list head */

        /* set the data pointer */
        arg_pMemoryManager->m_pBlockList->m_pData = (ix_uint32*)arg_pMemoryManager->m_pStartAddress;

        /* set the size */
        arg_pMemoryManager->m_pBlockList->m_Size = arg_pMemoryManager->m_MemoryBlockSize;

        /* set the links */
        arg_pMemoryManager->m_pBlockList->m_pNextCell = 0;
        arg_pMemoryManager->m_pBlockList->m_pPrevCell = 0;

        /* set the rover to the head of the block list */
        arg_pMemoryManager->m_pRover = arg_pMemoryManager->m_pBlockList;
    }

#if defined(_IX_RM_MULTI_THREADED_)

    /* unlock the memory manager data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    return err;
}


/**
 * NAME: _ix_rm_mm_finalizer
 *
 * DESCRIPTION: This function is the finalizer for the memory manager passed as an argument.
 *          After the finalizer returns the memory manager will not have any memory allocated
 *          and should be ready for a new initialization for reuse. After the finalization the
 *          memory manager can not handle requests.
 * 
 * @Param:  - IN arg_pMemoryManager - the address of the memory manager that will be finalized.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_finalizer( 
                              _ix_mm_memory_manager* arg_pMemoryManager 
                            )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mm_finalizer";

    ix_error err = IX_SUCCESS;

    register _ix_mm_memory_block_cell* a_pCell;
    register _ix_mm_memory_block_cell* a_pNextCell;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &arg_pMemoryManager->m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_MULTI_THREADED_)

    /* lock the memory manager data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    /* release all memory by destroying the block list */
    a_pCell = arg_pMemoryManager->m_pBlockList;

    while(a_pCell)
    {
        a_pNextCell = a_pCell->m_pNextCell;
        ix_ossl_free(a_pCell);
        a_pCell = a_pNextCell;
    }

    /* set the block list to 0 */
    arg_pMemoryManager->m_pBlockList = 0;

    /* set the rover to 0 */
    arg_pMemoryManager->m_pRover = 0;

    /* finalize the recycled cell list */
    a_pCell = arg_pMemoryManager->m_pRecycledBlockCellList;
    while(a_pCell)
    {
        a_pNextCell = a_pCell->m_pNextCell;
        ix_ossl_free(a_pCell);
        a_pCell = a_pNextCell;
    }

    /* set the recycled cell list to 0 */
    arg_pMemoryManager->m_pRecycledBlockCellList = 0;

    /* reset the recycled cell count */
    arg_pMemoryManager->m_RecycledBlockCellCount = 0;

#if defined(_IX_RM_MULTI_THREADED_)

    /* unlock the memory manager data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    return err;
}

/**
 * NAME: _ix_rm_mm_split_cell
 *
 * DESCRIPTION: This function will split the passed cell into two cells and return the newly created
 *          cell. The passed cell will be adjusted to firstHalfSize size and the new cell will get the
 *          remaining size. The newly created cell will be inserted in the list after the current cell.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager data this function works on.
 * @Param:  - IN arg_pCell - represents a pointer to the cell to be split
 * @Param:  - IN arg_Size - represents the new size of the passed cell. The newly created cell will get
 *          the remaining memory.
 *
 * @Return: returns a new _ix_mm_memory_block_cell to be used by the memory manager or NULL for failure. 
 */
_ix_mm_memory_block_cell* _ix_rm_mm_split_cell(
                                                _ix_mm_memory_manager* arg_pMemoryManager, 
                                                _ix_mm_memory_block_cell* arg_pCell, 
                                                ix_uint32 arg_Size
                                              )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mm_split_cell";

    register _ix_mm_memory_block_cell* pNewCell = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* allocate a new cell */
    pNewCell = _ix_rm_mm_new_cell(arg_pMemoryManager);

    /* we got a new cell */
    if(pNewCell != 0)
    {
        /* set the newly created cell size */
        pNewCell->m_Size = arg_pCell->m_Size - arg_Size;

        /* adjust the current cell size */
        arg_pCell->m_Size = arg_Size;

        /* set the next cell of the new created cell to the active's cell next cell*/
        pNewCell->m_pNextCell = arg_pCell->m_pNextCell;

        /* set the new created cell prev cell to the active cell */
        pNewCell->m_pPrevCell = arg_pCell;

        /* set the active's cell next cell to the newly created cell */
        arg_pCell->m_pNextCell = pNewCell;

        if(pNewCell->m_pNextCell)
        {
            /* set the next active's cell previous cell to the newly created cell */
            pNewCell->m_pNextCell->m_pPrevCell = pNewCell;
        }

        /* set the new cell's data pointer to this value */
        pNewCell->m_pData = (ix_uint32*)((ix_uint8*)arg_pCell->m_pData + arg_pCell->m_Size);
    }
        
    return pNewCell;
}


/**
 * NAME: _ix_rm_mm_new_cell
 *
 * DESCRIPTION: This function will return a new _ix_mm_memory_block_cell to be used by the memory
 *          manager.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager data this function works on.
 *
 * @Return: returns a new _ix_mm_memory_block_cell to be used by the memory manager or NULL for failure. 
 */
_ix_mm_memory_block_cell* _ix_rm_mm_new_cell( 
                                              _ix_mm_memory_manager* arg_pMemoryManager 
                                            )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mm_new_cell";
    static const ix_uint32  s_CellAllocationCount = 10;

    /* Naming convention for automatic variables */
    register _ix_mm_memory_block_cell* a_pCell = 0;
    register _ix_mm_memory_block_cell* a_pCellList = arg_pMemoryManager->m_pRecycledBlockCellList;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(a_pCellList)
    {
        /* return the first entry in the linked list */
        a_pCell = a_pCellList;
        /* set the list head to the next cell in the list */
        arg_pMemoryManager->m_pRecycledBlockCellList = a_pCellList->m_pNextCell;
        /* decrement the counter */
        arg_pMemoryManager->m_RecycledBlockCellCount--;
    }
    else
    {
        ix_uint32 i;
        /* allocate s_CellAllocationCount entries */
        for(i = 0 ; i < s_CellAllocationCount; i++)
        {
            a_pCell = (_ix_mm_memory_block_cell*) ix_ossl_malloc(sizeof(_ix_mm_memory_block_cell));
            if(a_pCell)
            {
                /* add the cell at the beginning of the list */
                a_pCell->m_pNextCell = a_pCellList;
                /* adjust the cell list's head */
                a_pCellList = a_pCell;
                /* increment the cell's counter */
                arg_pMemoryManager->m_RecycledBlockCellCount++;
            }
        }

        if(a_pCellList)
        {
            /* return the first entry in the linked list */
            a_pCell = a_pCellList;
            /* set the cell list head */
            arg_pMemoryManager->m_pRecycledBlockCellList = a_pCellList->m_pNextCell;
            /* decrement the cell counter */
            arg_pMemoryManager->m_RecycledBlockCellCount--;
        }
    }

    return a_pCell;
}

/**
 * NAME: _ix_rm_mm_release_cell
 *
 * DESCRIPTION: This function will release the specified block memory cell. This function and the above
 *          allows to recycle the cells in order to avoid recurring heap allocations and deallocations.
 * 
 * @Param:  - IN arg_pMemoryManager - a pointer to the memory manager data this function works on.
 * @Param:  - IN arg_pCell - block cell that should be released;
 *
 */
void _ix_rm_mm_release_cell( 
                             _ix_mm_memory_manager* arg_pMemoryManager, 
                             _ix_mm_memory_block_cell* arg_pCell 
                           )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mm_release_cell";
    /* this value is used as a high water mark */
    static const ix_uint32  s_CellHighWaterMark = 100;
    /* this value is used as alow water mark */
    static const ix_uint32  s_CellLowWaterMark = 50;

    
    /**
     * If the number of unused cells goes over the high water mark then
     * we release all the cells till we reach the low water mark. The values
     * have been chosen arbitrarily for now. Based on the type of apps these
     * values could be chosen in an optimal way.
     */


    register _ix_mm_memory_block_cell* a_pCell = 0;
    register _ix_mm_memory_block_cell* a_pCellList = arg_pMemoryManager->m_pRecycledBlockCellList;
    register ix_uint32 a_CellCount = arg_pMemoryManager->m_RecycledBlockCellCount;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(a_CellCount >= s_CellHighWaterMark)
    {
        /* release all the cells down to low watermark */
        for(;a_CellCount > s_CellLowWaterMark;a_CellCount--)
        {
            a_pCell = a_pCellList;
            a_pCellList = a_pCellList->m_pNextCell;
            ix_ossl_free(a_pCell);
        }
    }

    /* add the current cell to the list */
    arg_pCell->m_pNextCell = a_pCellList;
    a_CellCount++;

    /* adjust the memory manager */
    arg_pMemoryManager->m_pRecycledBlockCellList = arg_pCell;
    arg_pMemoryManager->m_RecycledBlockCellCount = a_CellCount;
}


#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * NAME: _ix_rm_mm_read_dram
 *
 * DESCRIPTION: This function reads a DRAM 32 bit memory location from the arg_pFromAddress and stores
 *          the value to the arg_pToAddress. The memory pointers should guarantee 4 byte alignment. The
 *          alignment will be checked just in DEBUG mode.
 * 
 * @Param:  - IN arg_pMemoryManager - represents the address of the memory manager data should work on
 * @Param:  - IN arg_pFromAddress - represents the address we read from.
 * @Param:  - OUT arg_pToAddress - represents the address where to store the read value.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_read_dram( 
                              const _ix_mm_memory_manager* arg_pMemoryManager, 
                              const ix_uint32* arg_pFromAddress, 
                              ix_uint32* arg_pToAddress 
                            )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mm_read_dram";

    ix_error err = IX_SUCCESS;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* this is for debug purposes */
    *arg_pToAddress = *arg_pFromAddress;

    return err;
}



/**
 * NAME: _ix_rm_mm_read_sram
 *
 * DESCRIPTION: This function reads a SRAM 32 bit memory location from the arg_pFromAddress and stores
 *          the value to the arg_pToAddress. The memory pointers should guarantee 4 byte alignment. The
 *          alignment will be checked just in DEBUG mode.
 * 
 * @Param:  - IN arg_pMemoryManager - represents the address of the memory manager data should work on
 * @Param:  - IN arg_pFromAddress - represents the address we read from.
 * @Param:  - OUT arg_pToAddress - represents the address where to store the read value.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_read_sram( 
                              const _ix_mm_memory_manager* arg_pMemoryManager, 
                              const ix_uint32* arg_pFromAddress, 
                              ix_uint32* arg_pToAddress 
                            )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mm_read_sram";

    ix_error err = IX_SUCCESS;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* this is for debug purposes */
    *arg_pToAddress = *arg_pFromAddress;

    return err;
}



/**
 * NAME: _ix_rm_mm_read_scratch
 *
 * DESCRIPTION: This function reads a SCRATCH 32 bit memory location from the arg_pFromAddress and stores
 *          the value to the arg_pToAddress. The memory pointers should guarantee 4 byte alignment. The
 *          alignment will be checked just in DEBUG mode.
 * 
 * @Param:  - IN arg_pMemoryManager - represents the address of the memory manager data should work on.
 * @Param:  - IN arg_pFromAddress - represents the address we read from.
 * @Param:  - OUT arg_pToAddress - represents the address where to store the read value.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_read_scratch( 
                                 const _ix_mm_memory_manager* arg_pMemoryManager, 
                                 const ix_uint32* arg_pFromAddress, 
                                 ix_uint32* arg_pToAddress 
                               )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mm_read_scratch";

    ix_error err = IX_SUCCESS;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* this is for debug purposes */
    *arg_pToAddress = *arg_pFromAddress;

    return err;
}



/**
 * NAME: _ix_rm_mm_write_dram
 *
 * DESCRIPTION: This function writes a 32 bit value to the specified DRAM address location. The memory pointer
 *          should guarantee 4 byte alignment. The alignment will be checked just in DEBUG mode.
 * 
 * @Param:  - IN arg_pMemoryManager - represents the address of the memory manager data should work on
 * @Param:  - IN arg_Value - a 32 bit value that will be written at the specified location.
 * @Param:  - IN arg_pToAddress - address where the data will be written.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_write_dram( 
                               const _ix_mm_memory_manager* arg_pMemoryManager, 
                               ix_uint32 arg_Value, 
                               ix_uint32* arg_pToAddress 
                             )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mm_write_dram";

    ix_error err = IX_SUCCESS;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* this is for debug purposes */
    *arg_pToAddress = arg_Value;

    return err;
}



/**
 * NAME: _ix_rm_mm_write_sram
 *
 * DESCRIPTION: This function writes a 32 bit value to the specified SRAM address location. The memory pointer
 *          should guarantee 4 byte alignment. The alignment will be checked just in DEBUG mode.
 * 
 * @Param:  - IN arg_pMemoryManager - represents the address of the memory manager data should work on
 * @Param:  - IN arg_Value - a 32 bit value that will be written at the specified location.
 * @Param:  - IN arg_pToAddress - address where the data will be written.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_write_sram( 
                               const _ix_mm_memory_manager* arg_pMemoryManager, 
                               ix_uint32 arg_Value, 
                               ix_uint32* arg_pToAddress 
                             )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mm_write_sram";

    ix_error err = IX_SUCCESS;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* this is for debug purposes */
    *arg_pToAddress = arg_Value;

    return err;
}



/**
 * NAME: _ix_rm_mm_write_scratch
 *
 * DESCRIPTION: This function writes a 32 bit value to the specified SCRACH address location. The memory pointer
 *          should guarantee 4 byte alignment. The alignment will be checked just in DEBUG mode.
 * 
 * @Param:  - IN arg_pMemoryManager - represents the address of the memory manager data should work on
 * @Param:  - IN arg_Value - a 32 bit value that will be written at the specified location.
 * @Param:  - IN arg_pToAddress - address where the data will be written.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_write_scratch( 
                                  const _ix_mm_memory_manager* arg_pMemoryManager, 
                                  ix_uint32 arg_Value, 
                                  ix_uint32* arg_pToAddress 
                                )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mm_write_scratch";

    ix_error err = IX_SUCCESS;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* this is for debug purposes */
    *arg_pToAddress = arg_Value;

    return err;
}

#else /* Westport hardware */

/**
 * NAME: _ix_rm_mm_read_south
 *
 * DESCRIPTION: This function reads a South bank 32 bit memory location from the arg_pFromAddress and stores
 *          the value to the arg_pToAddress. The memory pointers should guarantee 4 byte alignment. The
 *          alignment will be checked just in DEBUG mode.
 * 
 * @Param:  - IN arg_pMemoryManager - represents the address of the memory manager data should work on
 * @Param:  - IN arg_pFromAddress - represents the address we read from.
 * @Param:  - OUT arg_pToAddress - represents the address where to store the read value.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_read_south( 
                              const _ix_mm_memory_manager* arg_pMemoryManager, 
                              const ix_uint32* arg_pFromAddress, 
                              ix_uint32* arg_pToAddress 
                            )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mm_read_south";

    ix_error err = IX_SUCCESS;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* this is for debug purposes */
    *arg_pToAddress = *arg_pFromAddress;

    return err;
}


/**
 * NAME: _ix_rm_mm_read_north
 *
 * DESCRIPTION: This function reads a North bank 32 bit memory location from the arg_pFromAddress and stores
 *          the value to the arg_pToAddress. The memory pointers should guarantee 4 byte alignment. The
 *          alignment will be checked just in DEBUG mode.
 * 
 * @Param:  - IN arg_pMemoryManager - represents the address of the memory manager data should work on
 * @Param:  - IN arg_pFromAddress - represents the address we read from.
 * @Param:  - OUT arg_pToAddress - represents the address where to store the read value.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_read_north( 
                              const _ix_mm_memory_manager* arg_pMemoryManager, 
                              const ix_uint32* arg_pFromAddress, 
                              ix_uint32* arg_pToAddress 
                            )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mm_read_north";

    ix_error err = IX_SUCCESS;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* this is for debug purposes */
    *arg_pToAddress = *arg_pFromAddress;

    return err;
}

/**
 * NAME: _ix_rm_mm_write_south
 *
 * DESCRIPTION: This function writes a 32 bit value to the specified South bank address location. The memory pointer
 *          should guarantee 4 byte alignment. The alignment will be checked just in DEBUG mode.
 * 
 * @Param:  - IN arg_pMemoryManager - represents the address of the memory manager data should work on
 * @Param:  - IN arg_Value - a 32 bit value that will be written at the specified location.
 * @Param:  - IN arg_pToAddress - address where the data will be written.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_write_south( 
                               const _ix_mm_memory_manager* arg_pMemoryManager, 
                               ix_uint32 arg_Value, 
                               ix_uint32* arg_pToAddress 
                             )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mm_write_south";

    ix_error err = IX_SUCCESS;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* this is for debug purposes */
    *arg_pToAddress = arg_Value;

    return err;
}


/**
 * NAME: _ix_rm_mm_write_north
 *
 * DESCRIPTION: This function writes a 32 bit value to the specified North bank address location. The memory pointer
 *          should guarantee 4 byte alignment. The alignment will be checked just in DEBUG mode.
 * 
 * @Param:  - IN arg_pMemoryManager - represents the address of the memory manager data should work on
 * @Param:  - IN arg_Value - a 32 bit value that will be written at the specified location.
 * @Param:  - IN arg_pToAddress - address where the data will be written.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mm_write_north( 
                               const _ix_mm_memory_manager* arg_pMemoryManager, 
                               ix_uint32 arg_Value, 
                               ix_uint32* arg_pToAddress 
                             )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mm_write_north";

    ix_error err = IX_SUCCESS;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* this is for debug purposes */
    *arg_pToAddress = arg_Value;

    return err;
}

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */


/**
 * NAME: _ix_rm_memory_block_cell_action
 *
 * DESCRIPTION: This function iterates through alist of memory block cells and will apply
 *          the arg_Predicate to each entry in the logcal interval [arg_pStartCell, arg_pEndCell).
 *          The user have to make sure there is a finite path from arg_pStartCell to arg_pEndCell.
 *          The function will return 0 if all the invokations of the arg_Predicate predicate
 *          return non 0. If one invocation returns 0 then the function will return the address
 *          of the cell for which the predicate failed.
 * 
 * @Param:  - IN arg_pStartCell - the start cell address in the linked list;
 * @Param:  - IN arg_pEndCell - the end cell address in the linked list. The predicate will
 *          not be applied for this cell;
 * @Param:  - IN arg_Predicate - the active predicate that will be applied to each cell;
 * @Param:  - IN arg_pContext - a pointer to some context data that will be passed to arg_Predicate at
 *          each invocation;
 *
 * @Return: The function will return 0 if all the invocations of the arg_Predicate predicate
 *          return non 0. If one invokation returns 0 then the function will return the address
 *          of the cell for which the predicate failed.
 */
_ix_mm_memory_block_cell* _ix_rm_memory_block_cell_action( 
                                                           _ix_mm_memory_block_cell* arg_pStartCell, 
                                                           _ix_mm_memory_block_cell* arg_pEndCell, 
                                                           _ix_mm_memory_block_cell_predicate arg_Predicate, 
                                                           void* arg_pContext 
                                                         )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_memory_block_cell_action";

    /* Naming convention for automatic variables */
    _ix_mm_memory_block_cell* arg_pCurrentCell;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* iterate through list entrues and apply the predicate to each entry */
    for(arg_pCurrentCell = arg_pStartCell; arg_pCurrentCell != arg_pEndCell; arg_pCurrentCell = arg_pCurrentCell->m_pNextCell) 
    {
        if(arg_Predicate(arg_pCurrentCell, arg_pContext) == 0)
            return arg_pCurrentCell;
    }

    return 0;
}


/**
 * NAME: _ix_rm_memory_block_cell_print
 *
 * DESCRIPTION: This function will print a memory block cell from a memory manager block list.  
 * 
 * @Param:  - IN arg_pCell - memory cell to print
 * @Param:  - IN arg_pContext - context argument that in this case should be a pointer to an
 *          ix_uint32 counter that will be incremented at each call;
 *
 * @Return: Returns 1 all the time.
 */
ix_uint32 _ix_rm_memory_block_cell_print( 
                                          _ix_mm_memory_block_cell* arg_pCell, 
                                          void* arg_pContext 
                                        )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_memory_block_cell_print";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);
    
    /* print each entry */
    ix_ossl_message_log("\tEntry: %p\t",arg_pCell);
    if(arg_pCell->m_Size)
    {
        ix_ossl_message_log(" - free\n");
        ix_ossl_message_log("\t\tm_pData: 0x%p\n",arg_pCell->m_pData);
        ix_ossl_message_log("\t\tm_Size: %lu\n",arg_pCell->m_Size);
    }
    else
    {
        _ix_mm_memory_block_cell* pNextCell = arg_pCell->m_pNextCell;
        ix_ossl_message_log(" - in use\n");
        ix_ossl_message_log("\t\tm_pData: 0x%p\n",arg_pCell->m_pData);
        if(pNextCell)
            ix_ossl_message_log("\t\tactual size: %u\n",(ix_uint8*)pNextCell->m_pData - (ix_uint8*)arg_pCell->m_pData);
        else
            /* we can not calculate the size in this context */
            ix_ossl_message_log("\t\tactual size: ...\n");
    }

    ix_ossl_message_log("\t\tm_pNextCell: 0x%p\n",arg_pCell->m_pNextCell);
    ix_ossl_message_log("\t\tm_pPrevCell: 0x%p\n",arg_pCell->m_pPrevCell);
    ix_ossl_message_log("\n");

    /* increment the counter */
    (*(ix_uint32*)arg_pContext)++;
    return 1;
}

#if 0

/**
 * NAME: _ix_rm_memory_recycled_cell_print
 *
 * DESCRIPTION: This function will print a memory block cell from a memory manager recycled
 *          cell list.  
 * 
 * @Param:  - IN arg_pCell - memory cell to print
 * @Param:  - IN arg_pContext - context argument that in this case should be a pointer to an
 *          ix_uint32 counter that will be incremented at each call;
 *
 * @Return: Returns 1 all the time.
 */
ix_uint32 _ix_rm_memory_recycled_cell_print( 
                                             _ix_mm_memory_block_cell* arg_pCell, 
                                             void* arg_pContext 
                                           )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_memory_recycled_cell_print";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* print each entry */
    ix_ossl_message_log("\t\tentry: 0x%p\n",arg_pCell);
    ix_ossl_message_log("\t\tm_pNextCell: 0x%p\n",arg_pCell->m_pNextCell);
    ix_ossl_message_log("\n");

    /* increment the counter */
    (*(ix_uint32*)arg_pContext)++;
    return 1;
}

#endif /* 0 */

/**
 * NAME: _ix_rm_memory_get_free_memory
 *
 * DESCRIPTION: This function is used ti retrieve the free memory for a memory manager.
 * 
 * @Param:  - IN arg_pCell - current memory cell;
 * @Param:  - IN arg_pContext - context argument that in this case should be a pointer to an
 *          ix_uint32 value that will be increased by he size of current cell;
 *
 * @Return: Returns 1 all the time.
 */
ix_uint32 _ix_rm_memory_get_free_memory( 
                                         _ix_mm_memory_block_cell* arg_pCell, 
                                         void* arg_pContext 
                                       )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_memory_get_free_memory";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    *(ix_uint32*)arg_pContext += arg_pCell->m_Size;
    return 1;
}



/**
 * Read/Write support functions
 */


#if !defined(_IX_RM_IMPL_HARDWARE_)

/**
 * NAME: _ix_rm_mem_uint8_read
 *
 * DESCRIPTION: This function is provided as a IX_RM_MEM_UINT8_READ macro support function.
 *          Returns directly the read value. 
 * 
 * 
 * @Param:  - IN arg_pFromAddr - represents the address where to read from.
 *
 * @Return: The unsigned 8 bit value at the specified memory address. If the operation fails then the return
 *          value will be (ix_uint8)(-1) but there is no guarantee that we get that value because
 *          of an error.
 */
ix_uint8 _ix_rm_mem_uint8_read( 
                                const ix_uint8* arg_pFromAddr 
                              )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mem_uint8_read";

    /* Naming convention for automatic variables */
    ix_uint32 a_ReadValue;
    register ix_uint32* a_pMemAddress;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* read a 32 bit value and return just required byte */

    /* get a 32 bit aligned address */
    a_pMemAddress = (ix_uint32*)(((ix_uint32)arg_pFromAddr) & 0xfffffffc);
    
    /* read a 32 bit value */
    a_ReadValue = _ix_rm_mem_uint32_read(a_pMemAddress);

    /* return the part we are interested in */
    return *(ix_uint8*)(((ix_uint8*)&a_ReadValue) + ((ix_uint32)arg_pFromAddr & 0x00000003));
}


#endif /* !defined(_IX_RM_IMPL_HARDWARE_) */



#if !defined(_IX_RM_IMPL_HARDWARE_)


/**
 * NAME: _ix_rm_mem_uint16_read
 *
 * DESCRIPTION: This function is provided as a IX_RM_MEM_UINT16_READ macro support function.
 *          The passed memory address where to read from should be 2 bytes aligned or the
 *          results might be unpredictable. In DEBUG mode the alignment will be checked but
 *          not in RELEASE.
 * 
 * 
 * @Param:  - IN arg_pFromAddr - represents the address where to read from. The address should be
 *          2 bytes aligned or unpredictable behaviour might result.
 *
 * @Return: The value at the specified memory address. If the operation fails then the return
 *          value will be (ix_uint16)(-1) but there is no guarantee that we get that value because
 *          of an error.
 */
ix_uint16 _ix_rm_mem_uint16_read( 
                                  const ix_uint16* arg_pFromAddr 
                                )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mem_uint16_read";

    /* Naming convention for automatic variables */
    ix_uint32 a_ReadValue;
    register ix_uint32* a_pMemAddress;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);
    
#if defined(_IX_RM_DEBUG_)

    /* check the alignment in DEBUG mode */
    if(((ix_uint32)arg_pFromAddr & 0x00000001) != 0)
        return (ix_uint16)(-1);

#endif /* defined _IX_RM_DEBUG_ */

    /* read a 32 bit value and return just required word */

    /* get a 32 bit aligned address */
    a_pMemAddress = (ix_uint32*)(((ix_uint32)arg_pFromAddr) & 0xfffffffc);
    
    /* read a 32 bit value */
    a_ReadValue = _ix_rm_mem_uint32_read(a_pMemAddress);

    /* return the part we are interested in */
    return *(ix_uint16*)(((ix_uint8*)&a_ReadValue) + ((ix_uint32)arg_pFromAddr & 0x00000002));
}


#endif /* !defined(_IX_RM_IMPL_HARDWARE_) */



#if !defined(_IX_RM_IMPL_HARDWARE_)

/**
 * NAME: _ix_rm_mem_uint32_read
 *
 * DESCRIPTION: This function is provided as a IX_RM_MEM_UINT32_READ macro support function.
 *          The passed memory address where to read from should be 4 bytes aligned or the
 *          results might be unpredictable. In DEBUG mode the alignment will be checked but
 *          not in RELEASE.
 * 
 * 
 * @Param:  - IN arg_pFromAddr - represents the address where to read from. The address should be
 *          4 bytes aligned or unpredictable behaviour might result.
 *
 * @Return: The value at the specified memory address. If the operation fails then the return
 *          value will be (ix_uint32)(-1) but there is no guarantee that we get that value because
 *          of an error.
 */
ix_uint32 _ix_rm_mem_uint32_read( 
                                  const ix_uint32* arg_pFromAddr 
                                )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mem_uint32_read";

    register _ix_mm_memory_manager* pMemoryManager;
    register ix_uint32   channelIndex;
    register ix_uint32   memType;

    ix_error a_InternalError;

    ix_uint32 a_ReadValue = (ix_uint32)(-1);

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_DEBUG_)

    /* check the alignment in DEBUG mode */
    if(((ix_uint32)arg_pFromAddr & 0x00000003) != 0)
        return (ix_uint32)(-1);

#endif /* defined _IX_RM_DEBUG_ */

    /* for NULL we just return immediately */
    if(arg_pFromAddr == 0)
        return a_ReadValue;

    /* search each memory manager group */
    for(memType = IX_MEMORY_TYPE_FIRST; memType <= IX_MEMORY_TYPE_SCRATCH; memType++)
    {
        /* search each memory manager in the group */
        for(channelIndex = 0; channelIndex < g_aMemoryManagersGroups[memType].m_MemberCount; channelIndex++)
        {
            pMemoryManager = g_aMemoryManagersGroups[memType].m_pMembers + channelIndex;
            /* check if our address is in the address range of this memory manager */
            if((arg_pFromAddr >= pMemoryManager->m_pStartAddress) && (arg_pFromAddr < pMemoryManager->m_pEndAddress))
            {
                /* call the read function for this memory manager */
                a_InternalError = pMemoryManager->m_pVMT->m_ReadFuncPtr(pMemoryManager, arg_pFromAddr, &a_ReadValue);
                /* check the outcome of the free operation */
                if(a_InternalError)
                    a_ReadValue = (ix_uint32)(-1);

                /* we are done */
                break;
            }
        }
    }
    return a_ReadValue;
}

#endif /* end !defined(_IX_RM_IMPL_HARDWARE_) */



#if !defined(_IX_RM_IMPL_HARDWARE_)


/**
 * NAME: _ix_rm_mem_uint64_read
 *
 * DESCRIPTION: This function is provided as a IX_RM_MEM_UINT64_READ macro support function.
 *          The passed memory address where to read from should be 8 bytes aligned or the
 *          results might be unpredictable. In DEBUG mode the alignment will be checked but
 *          not in RELEASE.
 * 
 * @Param:  - IN arg_pFromAddr - represents the address where to read from. The address should be
 *          8 bytes aligned or unpredictable behaviour might result.
 *
 * @Return: The value at the specified memory address. If the operation fails then the return
 *          value will be (ix_uint64)(-1) but there is no guarantee that we get that value because
 *          of an error.
 */
ix_uint64 _ix_rm_mem_uint64_read( 
                                  const ix_uint64* arg_pFromAddr 
                                )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mem_uint64_read";

    /* Naming convention for automatic variables */
    ix_uint64 a_ReadValue;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_DEBUG_)

    /* check the alignment in DEBUG mode */
    if(((ix_uint32)arg_pFromAddr & 0x00000007) != 0)
        return (ix_uint64)(-1);

#endif /* defined _IX_RM_DEBUG_ */

    /* read a 32 bit value */
    *((ix_uint32*)&a_ReadValue) = _ix_rm_mem_uint32_read((ix_uint32*)arg_pFromAddr);

    /* read the next 32 bit value */
    *(((ix_uint32*)&a_ReadValue) + 1) = _ix_rm_mem_uint32_read(((ix_uint32*)arg_pFromAddr + 1));

    return a_ReadValue;
}


#endif /* end !defined(_IX_RM_IMPL_HARDWARE_) */


#if !defined(_IX_RM_IMPL_HARDWARE_)

/**
 * NAME: _ix_rm_mem_uint8_write
 *
 * DESCRIPTION: This function is provided as a IX_RM_MEM_UINT8_WRITE macro support function.
 *          This function will write the 8 bit unsigned arg_Value into the location referred 
 *          by arg_pToAddr. No memory alignment is required.  
 * 
 * 
 * @Param:  - IN arg_pToAddr - represents the memory address where the value should be written.
 * @Param:  - IN arg_Value - represents the value that should be written at the location specified by
 *          arg_pToAddr.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mem_uint8_write( 
                                 ix_uint8* arg_pToAddr, 
                                 ix_uint8 arg_Value 
                               )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mem_uint8_write";

    ix_error err = IX_SUCCESS;
    ix_uint32 a_ReadValue;
    register ix_uint32* a_pMemAddress;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* read a 32 bit value alter it with our byte and write it back */

    /* get a 32 bit aligned address */
    a_pMemAddress = (ix_uint32*)(((ix_uint32)arg_pToAddr) & 0xfffffffc);
    
    /* read a 32 bit value */
    a_ReadValue = _ix_rm_mem_uint32_read(a_pMemAddress);

    /* modify the read value */
    *(ix_uint8*)(((ix_uint8*)&a_ReadValue) + ((ix_uint32)arg_pToAddr & 0x00000003)) = arg_Value;

    /* write back the value */
    err = _ix_rm_mem_uint32_write(a_pMemAddress, a_ReadValue);

    return err;
}


#endif /* end !defined(_IX_RM_IMPL_HARDWARE_) */


#if !defined(_IX_RM_IMPL_HARDWARE_)

/**
 * NAME: _ix_rm_mem_uint16_write
 *
 * DESCRIPTION: This function is provided as a IX_RM_MEM_UINT16_WRITE macro support function.
 *          This function will write the 16 bit unsigned arg_Value into the location referred 
 *          by arg_pToAddr. The passed memory address where to write to, should be 2 bytes aligned or the
 *          results might be unpredictable. In DEBUG mode the alignment will be checked but
 *          not in RELEASE.
 * 
 * @Param:  - IN arg_pToAddr - represents the memory address where the value should be written.
 *          The passed memory address where to write to, should be 2 bytes aligned or the
 *          results might be unpredictable.
 *
 * @Param:  - IN arg_Value - represents the value that should be written at the location specified by
 *          arg_pToAddr.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mem_uint16_write( 
                                  ix_uint16* arg_pToAddr, 
                                  ix_uint16 arg_Value 
                               )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mem_uint16_write";

    ix_error err = IX_SUCCESS;
    ix_uint32 a_ReadValue;
    register ix_uint32* a_pMemAddress;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check the alignment in DEBUG mode */
    IX_RM_ASSERT(((ix_uint32)arg_pToAddr & 0x00000001) == 0, s_aFunctionIdentifier, "bad memory address alignment!");

    /* read a 32 bit value alter it with our byte and write it back */

    /* get a 32 bit aligned address */
    a_pMemAddress = (ix_uint32*)(((ix_uint32)arg_pToAddr) & 0xfffffffc);
    
    /* read a 32 bit value */
    a_ReadValue = _ix_rm_mem_uint32_read(a_pMemAddress);

    /* modify the read value */
    *(ix_uint16*)(((ix_uint8*)&a_ReadValue) + ((ix_uint32)arg_pToAddr & 0x00000002)) = arg_Value;

    /* write back the value */
    err = _ix_rm_mem_uint32_write(a_pMemAddress, a_ReadValue);

    return err;
}


#endif /* end !defined(_IX_RM_IMPL_HARDWARE_) */


#if !defined(_IX_RM_IMPL_HARDWARE_)

/**
 * NAME: _ix_rm_mem_uint32_write
 *
 * DESCRIPTION: This function is provided as a IX_RM_MEM_UINT32_WRITE macro support function.
 *          This function will write the 32 bit unsigned arg_Value into the location referred 
 *          by arg_pToAddr. The passed memory address where to write to, should be 4 bytes aligned or the
 *          results might be unpredictable. In DEBUG mode the alignment will be checked but
 *          not in RELEASE.
 * 
 * @Param:  - IN arg_pToAddr - represents the memory address where the value should be written.
 *          The passed memory address where to write to, should be 4 bytes aligned or the
 *          results might be unpredictable.
 * @Param:  - IN arg_Value - represents the value that should be written at the location specified by
 *          arg_pToAddr.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mem_uint32_write( 
                                  ix_uint32* arg_pToAddr, 
                                  ix_uint32 arg_Value 
                               )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mem_uint32_write";

    register _ix_mm_memory_manager* pMemoryManager;
    register ix_uint32   channelIndex;
    register ix_uint32   memType;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check the alignment in DEBUG mode */
    IX_RM_ASSERT(((ix_uint32)arg_pToAddr & 0x00000003) == 0, s_aFunctionIdentifier, "bad memory address alignment!");

    /* for NULL we just return immediately */
    if(arg_pToAddr == 0)
        return IX_RM_ERROR_NULL_MEMORY_ADDRESS;

    /* search each memory manager group */
    for(memType = IX_MEMORY_TYPE_FIRST; memType <= IX_MEMORY_TYPE_SCRATCH; memType++)
    {
        /* search each memory manager in the group */
        for(channelIndex = 0; channelIndex < g_aMemoryManagersGroups[memType].m_MemberCount; channelIndex++)
        {
            pMemoryManager = g_aMemoryManagersGroups[memType].m_pMembers + channelIndex;
            /* check if our address is in the address range of this memory manager */
            if((arg_pToAddr >= pMemoryManager->m_pStartAddress) && (arg_pToAddr < pMemoryManager->m_pEndAddress))
            {
                /* call the read function for this memory manager */
                return pMemoryManager->m_pVMT->m_WriteFuncPtr(pMemoryManager, arg_Value, arg_pToAddr);
            }
        }
    }
    /* the memory address is not found in any memory manager's range */
    return IX_RM_ERROR_BAD_MEMORY_ADDRESS;
}


#endif /* end !defined(_IX_RM_IMPL_HARDWARE_) */

#if !defined(_IX_RM_IMPL_HARDWARE_)


/**
 * NAME: _ix_rm_mem_uint64_write
 *
 * DESCRIPTION: This function is provided as a IX_RM_MEM_UINT64_WRITE macro support function.
 *          This function will write the 64 bit unsigned arg_Value into the location referred 
 *          by arg_pToAddr. The passed memory address where to write to, should be 8 bytes aligned or the
 *          results might be unpredictable. In DEBUG mode the alignment will be checked but
 *          not in RELEASE.
 * 
 * @Param:  - IN arg_pToAddr - represents the memory address where the value should be written.
 *          The passed memory address where to write to, should be 8 bytes aligned or the
 *          results might be unpredictable.
 * @Param:  - IN arg_Value - represents the value that should be written at the location specified by
 *          arg_pToAddr.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_mem_uint64_write( 
                                 ix_uint64* arg_pToAddr, 
                                 ix_uint64 arg_Value 
                               )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_mem_uint64_write";

    /* Naming convention for automatic variables */
    ix_error err = IX_SUCCESS;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);
    
    /* check the alignment in DEBUG mode */
    IX_RM_ASSERT(((ix_uint32)arg_pToAddr & 0x00000007) == 0, s_aFunctionIdentifier, "bad memory address alignment!");

    /* write two 32 bit values */

    /* write the first 32 bit value */
    err = _ix_rm_mem_uint32_write((ix_uint32*)arg_pToAddr, *((ix_uint32*)&arg_Value));

    /* check if we succeeded */
    if(err != IX_SUCCESS)
        return err;
    else
    {
       return _ix_rm_mem_uint32_write(((ix_uint32*)arg_pToAddr + 1), *((ix_uint32*)&arg_Value + 1));
    }
}

#endif /* end !defined(_IX_RM_IMPL_HARDWARE_) */


