/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 4
 *
 * = LIBRARY
 *      Resource Manager Library
 *
 * = MODULE
 *      Microengine API
 *
 * = FILENAME
 *      internal_microengine.c
 *
 * = DESCRIPTION
 *      This file will define all the internal supporting functions for the 
 *      Microengine API.
 *
 * = AUTHOR
 *       Adrian Georgescu
 *       adrian.georgescu@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      10/26/2001 11:39:04 AM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */

static const char s_aFileIdentifier[] = "$Id: internal_microengine.c,v 1.44 2003/09/30 19:30:59 ageorges Exp $";

#define _IX_RM_PRINT_PATCHING_SYMBOLS_MESSAGES_

#define _IX_RM_WORKBENCH_SUPPORT_


#include "ix_ossl.h"

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"

#include "rm/symbols.h"

#if defined(_IX_RM_IMPL_HARDWARE_)

#include <uclo.h>
#include <halMev2Api.h>

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

#include "rm/internal/internal_support.h"

#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"

#include "rm/system.h"
#include "rm/internal/internal_system.h"

#include "rm/memory.h"

#include "rm/buffer.h"

#include "rm/hardware.h"

#include "rm/communication.h"

#include "rm/configuration_property.h"

#include "rm/microengine.h"
#include "rm/internal/internal_microengine.h"


#if defined( _IX_RM_MULTI_THREADED_ )

static ix_ossl_sem_t s_MicroengineAPIManagerSemaphore = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */


const char* g_aUCLOErrorStrings[] = 
{
    "UCLO_FAILURE",
    "UCLO_FILEFAIL",
    "UCLO_BADOBJ",
    "UCLO_MEMFAIL",
    "UCLO_BADARG",
    "UCLO_NOOBJ",
    "UCLO_IMGNOTFND",
    "UCLO_SYMNOTFND",
    "UCLO_NEIGHNOTFND",
    "UCLO_TABLEFULL",
    "UCLO_REDEFINE",
    "UCLO_UOFINCOMPAT",
    "UCLO_UOFVERINCOMPAT",
    "UCLO_CNVREADONLY",
    "UCLO_UNINITVAR",
    "UCLO_EXPRFAIL",
    "UCLO_EXCDDRAMSIZE",
    "UCLO_EXCDSRAM0SIZE",
    "UCLO_EXCDSRAM1SIZE",
    "UCLO_EXCDSRAM2SIZE",
    "UCLO_EXCDSRAM3SIZE",
    "UCLO_EXCDSCRTHSIZE",
    "UCLO_EXCDLMEMSIZE"
};

const ix_uint32 g_UCLOErrorStringCount = sizeof(g_aUCLOErrorStrings)/sizeof(char*);


const char* g_aHALMEErrorStrings[] = 
{
   "HALME_FAIL the operation failed!"
   "HALME_BADARG bad function argument!"
   "HALME_DUPBKP duplicate break point!"
   "HALME_NOUSTORE not ustore available!"
   "HALME_BADADDR bad address!"
   "HALME_BADLIB bad debug library -- wasn't initialized!"
   "HALME_INACTIVE micro-engine inactive!"
   "HALME_ACTIVE micro-engine active!"
   "HALME_RESET micro-engine is in reset!"
   "HALME_TIMEOUT the operation exceeded the time limit!"
   "HALME_ISSET condition/evaluation is set/true!"
   "HALME_NOTSET condition/evaluation is not set/false!"
};

const ix_uint32 g_HALMEErrorStringCount = sizeof(g_aHALMEErrorStrings)/sizeof(char*);

/**
 * This array defines the microengine number mapping
 */
#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

unsigned char g_aMicroengineNumberMapping[IX_MAX_MICROENGINE_NUMBER] = 
{
    /* 0 */ 0x00,
    /* 1 */ 0x01,
    /* 2 */ 0x02,
    /* 3 */ 0x03,

    /* 4 */ 0x10,
    /* 5 */ 0x11,
    /* 6 */ 0x12,
    /* 7 */ 0x13
};

#elif (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

unsigned char g_aMicroengineNumberMapping[IX_MAX_MICROENGINE_NUMBER] = 
{
    /* 0 */     0x00,
    /* 1 */     0x01,
    /* 2 */     0x02,
    /* 3 */     0x03,
    /* 4 */     0x04,
    /* 5 */     0x05,
    /* 6 */     0x06,
    /* 7 */     0x07,

    /* 8 */     0x10,
    /* 9 */     0x11,
    /* 10 */    0x12,
    /* 11 */    0x13,
    /* 12 */    0x14,
    /* 13 */    0x15,
    /* 14 */    0x16,
    /* 15 */    0x17
};    
#endif


/**
 * Global Microengine API manager
 */
_ix_microengine_data_manager g_MicroengineAPIManager;


/**
 * NAME: _ix_rm_microengine_init
 *
 * DESCRIPTION: This function should be called at the time the resource manager initializer 
 *          function is called. The function will initialize all the Microengine API.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_microengine_init(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_microengine_init";

    ix_error err = 0;
    ix_uint32 i;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    g_MicroengineAPIManager.m_UOFActiveHandle = 0;
    g_MicroengineAPIManager.m_IsUOFLoadedFromFile = 0;

    for(i = 0; i < IX_MAX_MICROENGINE_NUMBER; i++)
    {
        g_MicroengineAPIManager.m_aImportedSymbolsList[i] = 0;
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    /* initialize the synchronization semaphore */
    err = ix_ossl_sem_init(IX_OSSL_SEM_AVAILABLE, &s_MicroengineAPIManagerSemaphore);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_ME_API_INIT_FAILED, s_aFunctionIdentifier, 
                        "initialization of microengine API manager's semaphore failed!", err);
    }

    /* initialize synchronization */
    _ix_rm_sa_sync_adapter_init(&g_MicroengineAPIManager.m_SyncAdapter, 
        _ix_rm_sa_default_lock, _ix_rm_sa_default_unlock, (void*)&s_MicroengineAPIManagerSemaphore);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: _ix_rm_microengine_fini
 *
 * DESCRIPTION: This function should be called at the time the resource manager finalizer function
 *          is called. The function will provide cleanup for Microengine API.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_microengine_fini(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_microengine_fini";

    ix_error err = 0;
    ix_uint32 i;
    _ix_imported_symbol_node* pSymbolNode;
    _ix_imported_symbol_node* pNextSymbolNode;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

    {
        int status;

        /* stop and reset all microengines */
        for(i = 0; i < IX_MAX_MICROENGINE_NUMBER; i++)
        {
            status = halMe_IsMeEnabled(g_aMicroengineNumberMapping[i]);
            switch(status)
            {
                case HALME_ENABLED:
                    halMe_Stop(g_aMicroengineNumberMapping[i], 0xffffffff);
                    /* intentional fall through */
                case HALME_DISABLED:
                    halMe_Reset(1<<g_aMicroengineNumberMapping[i],1);
                    /* intentional fall through */
                case HALME_RESET:
                    halMe_ClrReset(1<<g_aMicroengineNumberMapping[i]);
                    break;
                default:
                    break;
            }
        }
    }

    if(g_MicroengineAPIManager.m_UOFActiveHandle != 0)
    {
        UcLo_DeleObj(g_MicroengineAPIManager.m_UOFActiveHandle);
        g_MicroengineAPIManager.m_UOFActiveHandle = 0;
    }

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    /* mark that the uCode is not loaded */
    g_MicrocodeLoaded = 0;

    g_MicroengineAPIManager.m_UOFActiveHandle = 0;
    g_MicroengineAPIManager.m_IsUOFLoadedFromFile = 0;

    for(i = 0; i < IX_MAX_MICROENGINE_NUMBER; i++)
    {
        /* free the uCode imported symbol lists if required */
        pSymbolNode = g_MicroengineAPIManager.m_aImportedSymbolsList[i];
        while(pSymbolNode != 0)
        {
            pNextSymbolNode = pSymbolNode->m_pNext;
            if(pSymbolNode->m_Name != 0)
                ix_ossl_free(pSymbolNode->m_Name);
            ix_ossl_free(pSymbolNode);
            pSymbolNode = pNextSymbolNode;
        }
        g_MicroengineAPIManager.m_aImportedSymbolsList[i] = 0;
    }

#if defined( _IX_RM_MULTI_THREADED_ )

    /* finalize synchronization */
    _ix_rm_sa_sync_adapter_init(&g_MicroengineAPIManager.m_SyncAdapter, 0, 0, 0);

    /* finalize the memory managers' synchronization semaphore */
    err = ix_ossl_sem_fini(s_MicroengineAPIManagerSemaphore);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_ME_API_FINI_FAILED, s_aFunctionIdentifier, 
                        "finalization of microengine API manager's semaphore failed!", err);
    }

    s_MicroengineAPIManagerSemaphore = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: _ix_rm_microengine_patch_symbols
 *
 * DESCRIPTION: This function will patch imported symbols into the microcode without any data 
 *  locking. It just adds the symbols to the internal symbol lists.
 * 
 * @Param:  - IN arg_MENumber - represents the microengine number for which this call is 
 *          intended.
 * @Param:  - IN arg_SymbolsNumber - represents the number of symbols in the arg_Symbols 
 *          array that will be patched.
 * @Param:  - IN arg_aSymbols - represents an array of symbols that will be patched into 
 *          the microcode.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error _ix_rm_microengine_patch_symbols( 
                                           ix_uint32 arg_MENumber, 
                                           ix_uint32 arg_SymbolsNumber, 
                                           const ix_imported_symbol arg_aSymbols[] 
                                         )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_microengine_patch_symbols";

    ix_error err = IX_SUCCESS;
    _ix_imported_symbol_node* pNewNode;
    ix_uint32 i;
    ix_uint32 length;
    char* pSymbolName;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* get the coresponding internal ME index */
    for(i = 0; i < IX_MAX_MICROENGINE_NUMBER; i++)
        if(arg_MENumber == g_aMicroengineNumberMapping[i])
            arg_MENumber = i;

    for (i = 0; i < arg_SymbolsNumber; i++)
    {
        pSymbolName = arg_aSymbols[i].m_Name;
        if(pSymbolName != 0)
        {
            pNewNode = ix_ossl_malloc(sizeof(_ix_imported_symbol_node));
            if(pNewNode == 0)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_PATCH_SYMBOLS_FAILED, s_aFunctionIdentifier, 
                        "allocation of symbol node failed!", err);
                break;
            }
            length = strlen(pSymbolName) + 1;
            pNewNode->m_Name = ix_ossl_malloc(length);
            if(pNewNode->m_Name == 0)
            {
                ix_ossl_free(pNewNode);
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_PATCH_SYMBOLS_FAILED, s_aFunctionIdentifier, 
                        "allocation of symbol name memory failed!", err);
                break;
            }
            ix_ossl_memcpy(pNewNode->m_Name, pSymbolName, length);
            pNewNode->m_Value = arg_aSymbols[i].m_Value;
            pNewNode->m_pNext = g_MicroengineAPIManager.m_aImportedSymbolsList[arg_MENumber];
            g_MicroengineAPIManager.m_aImportedSymbolsList[arg_MENumber] = pNewNode;
        }
    }

    return err;
}


/**
 * NAME: _ix_rm_patch_qarray_symbols
 *
 * DESCRIPTION: This function will the required QArray symbols for all SRAM channels
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_patch_qarray_symbols(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_patch_qarray_symbols";

    ix_error err = IX_SUCCESS;

    ix_imported_symbol aSymbol;

    ix_uint32* pAddress;

    ix_memory_type memType;
    ix_uint32 memChannel, i, j;
    ix_uint32 offset;
    void* pPhysicalAddress;

    ix_uint32 importSymbolFound;
    static char buf[50];

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    for (i = 0; i < IX_SRAM_CHANNELS_NUMBER; i++)
    {
        sprintf(buf, "QARRAY_INIT_SRAM_BASE_CH%d", (int)i);

        importSymbolFound = 0;

        offset = 0;
        if (ix_rm_hw_queue_array_get_base_address(i,&pAddress) == IX_SUCCESS)
        {
            err = ix_rm_get_phys_offset((void*)pAddress, &memType, &memChannel,
                                        &offset, &pPhysicalAddress);
            IX_RM_ASSERT(err == IX_SUCCESS, s_aFunctionIdentifier,
                         "cannot find phys offset");
            offset >>= 2;   /* word offset shall be patched */
        }

#if defined(IX_PLATFORM_2401) || defined(IX_PLATFORM_2801)

        /* Improvment: In Aleutian Islands we need to now in microcode
         * that SRAM channel is not present. Let's mark in with -1.
         * It does not affect Angel Island since it always has both SRAM
         * channels.
         */
        else
        {
            offset = (ix_uint32)-1;
        }

#endif /* IX_PLATFORM_2X01 */

        /* add information about the channel */
        /* We take that out as it causes some problems in the uCode */
        /* offset |= (i << 30); */

        aSymbol.m_Name = (char*)buf;
        aSymbol.m_Value = offset;

        for(j = 0; j < IX_MAX_MICROENGINE_NUMBER; j++)
        {

#if defined(_IX_RM_IMPL_HARDWARE_)

            if(UcLo_MeBindSymbol(g_MicroengineAPIManager.m_UOFActiveHandle, 
                g_aMicroengineNumberMapping[j], aSymbol.m_Name, aSymbol.m_Value) == UCLO_SUCCESS)
            {

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

                _ix_rm_microengine_patch_symbols(g_aMicroengineNumberMapping[j], 1, &aSymbol);
                importSymbolFound++;

#if defined(_IX_RM_IMPL_HARDWARE_)

            }

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

        }

        if(importSymbolFound == 0)
        {
            ix_ossl_message_log("Warning: None of the microengines imports the symbol %s! \n", aSymbol.m_Name);
            ix_ossl_message_log("\tSRAM controller QArray for channel %lu will not be initialized!\n", i);
        }
    }

    return err;
}



/**
 * NAME: _ix_rm_microengine_register_internal_symbols
 *
 * DESCRIPTION: This function will register some internal symbols with the specified microengine.
 *  Many required symbols do not require user's explicit patching. They will be set and registered
 *  for patching by RM library. These are internal values of RM that are required for uCode
 *  proper setup.
 * 
 * @Param:  - IN arg_MENumber - the microengine's number whose symbols should be registered
 *  with the passed microengine
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_microengine_register_internal_symbols( 
    ix_uint32 arg_MENumber)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_microengine_register_internal_symbols";

    ix_error err = IX_SUCCESS;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    return err;
} /* _ix_rm_microengine_register_internal_symbols() */


/**
 * NAME: _ix_rm_microengine_patch_all_symbols
 *
 * DESCRIPTION: This function will patch all symbols registered for patching for the specified
 *  microengine
 * 
 * @Param:  - IN arg_MENumber - the microengine's number whose symbols should be patched into
 *  the microcode.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_microengine_patch_all_symbols( 
                                               ix_uint32 arg_MENumber
                                             )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_microengine_patch_all_symbols";

    ix_error err = IX_SUCCESS;
    ix_uint32 i;
    _ix_imported_symbol_node* pSymbolNode;
    _ix_imported_symbol_node* pNextSymbolNode;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    for(i = 0; i < IX_MAX_MICROENGINE_NUMBER; i++)
    {

        /* free the uCode imported symbol lists as we patch them */
        pSymbolNode = g_MicroengineAPIManager.m_aImportedSymbolsList[i];
        while(pSymbolNode != 0)
        {

#if defined(_IX_RM_IMPL_HARDWARE_)

            {
                int status;

#if defined( _IX_RM_DEBUG_ ) && defined(_IX_RM_PRINT_PATCHING_SYMBOLS_MESSAGES_)

                ix_ossl_message_log("Patching for ME: 0x%02x symbol: %s to value: 0x%08lx\n", 
                    g_aMicroengineNumberMapping[i], pSymbolNode->m_Name, pSymbolNode->m_Value);

#endif /* defined ( _IX_RM_DEBUG_ ) && defined(_IX_RM_PRINT_PATCHING_SYMBOLS_MESSAGES_) */

                if((status = UcLo_MeBindSymbol(g_MicroengineAPIManager.m_UOFActiveHandle, 
                    g_aMicroengineNumberMapping[i], pSymbolNode->m_Name, 
                    (int)pSymbolNode->m_Value) != UCLO_SUCCESS))
                {

#if defined(_IX_RM_PRINT_PATCHING_SYMBOLS_MESSAGES_)

                    ix_error internalErr;
                    
                    internalErr = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_PATCH_SYMBOLS_FAILED, s_aFunctionIdentifier,
                            "failed to bind one symbol!", status);

                    if((unsigned int)(status - UCLO_FAILURE) < (unsigned int)g_UCLOErrorStringCount)
                        ix_ossl_message_log("Failed to patch for ME: 0x%02x symbol: %s\n with UCLO error code: %s\n", 
                            g_aMicroengineNumberMapping[i], pSymbolNode->m_Name, g_aUCLOErrorStrings[status-UCLO_FAILURE]);
                    else
                        ix_ossl_message_log("Failed to patch for ME: 0x%02x symbol: %s\n", 
                            g_aMicroengineNumberMapping[i], pSymbolNode->m_Name);
                }

#endif /* defined(_IX_RM_PRINT_PATCHING_SYMBOLS_MESSAGES_) */

                status = UCLO_SUCCESS;
            }

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

            pNextSymbolNode = pSymbolNode->m_pNext;

            
            /**
             * In debug mode do not delete the nodes just to have the messages
             * printed out correctly. In VxWorks where the printing is done in a 
             * different thread the symbol names might get out of scope.
             */

#if !defined ( _IX_RM_DEBUG_ )

            if(pSymbolNode->m_Name != 0)
                ix_ossl_free(pSymbolNode->m_Name);
            ix_ossl_free(pSymbolNode);

#endif /* end !defined ( _IX_RM_DEBUG_ ) */

            pSymbolNode = pNextSymbolNode;
            if(err != IX_SUCCESS)
            {
                g_MicroengineAPIManager.m_aImportedSymbolsList[i] = pSymbolNode;
                return err;
            }
        }
        g_MicroengineAPIManager.m_aImportedSymbolsList[i] = 0;
    }

    return err;
}

#if defined(_IX_RM_WORKBENCH_SUPPORT_)


/**
 * NAME: _ix_rm_download_microcode
 *
 * DESCRIPTION: This function will be registered with the ME debug library and will perform
 *      the copying the the UOF file into an internal RM buffer. That will give the oportunity
 *      to RM to complete the the symbol patching before the code is downloaded into the
 *      uStore.
 * 
 * @Param:  - IN arg_pUCodeBuffer - address of the Workbench downloaded uCode image.
 * @Param:  - IN arg_pUCodeSize - size of the Workbench downloaded uCode image
 *      upon return
 *
 * @Return: 0 if successful or a nonzero for failure.
 */
unsigned int _ix_rm_download_microcode( 
                               const unsigned char* arg_pUCodeBuffer, 
                               unsigned long arg_UCodeSize
                             )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_download_microcode";

    unsigned char* pBuffer;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_MicroengineAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    if((arg_pUCodeBuffer == 0) || (arg_UCodeSize == 0))
    {
        ix_ossl_message_log("Bad UOF buffer and/or size!\n");
        return 1;
    }

    pBuffer = (unsigned char*)ix_ossl_malloc(arg_UCodeSize);
    if(pBuffer == 0)
    {
        ix_ossl_message_log("Can not allocate buffer for the UOF image!\n");
        return 1;
    }

    ix_ossl_memcpy(pBuffer, arg_pUCodeBuffer, arg_UCodeSize);

    g_MicroengineAPIManager.m_UCodeSize = arg_UCodeSize;
    g_MicroengineAPIManager.m_pUCodeBuffer = pBuffer;

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return 0;
}


#endif /* defined(_IX_RM_WORKBENCH_SUPPORT_) */
