/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      internal_pci_communication.c
 *
 * = DESCRIPTION
 *      This file provides the implementation of the support functions
 *      of the PCI communication infrastructure.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 *      Aaron Luk
 *      aaron.luk@intel.com
 *
 * = CHANGE HISTORY
 *      7/9/2002 10:32:53 AM - creation time
 * ============================================================================
 */

static const char s_FileIdentifier[] = "$Id: internal_pci_communication.c,v 1.84 2003/09/05 19:52:33 ageorges Exp $";


/* Define IX_PCI_A0_SILICON to indicate that we are using A0 silicon. */
/*#define IX_PCI_A0_SILICON           1*/

/* Define IX_PCI_TEST_INTERRUPTS to test the interrupts only. */
/*#define IX_PCI_TEST_INTERRUPTS 1*/

#include "ix_ossl.h"

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

/**
 * INCLUDE_PCI and INCLUDE_EXTRA_PCI_MAPPING must be defined
 * to include the PCI support from the BSP.
 */
#define INCLUDE_PCI                 1
#define INCLUDE_EXTRA_PCI_MAPPING   1



#endif /* (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) */

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"


#include "rm/symbols.h"
#include "rm/internal/internal_support.h"

#include <hal_scratch.h>

#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"
#include "rm/system.h"
#include "rm/internal/internal_system.h"

#include "rm/memory.h"

#include "rm/buffer.h"
#include "rm/internal/internal_buffer.h"
#include "rm/hardware.h"
#include "rm/communication.h"
#include "rm/internal/internal_communication.h"

#include "rm/remote_communication.h"
#include "rm/internal/internal_remote_communication.h"

#include "rm/internal/internal_pci_communication.h"

#include "rm/fast_memory_operations.h"


#if defined(_IX_COMPILE_PCI_COMM_IN_)

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

/* Include VxWorks headers for interrupts. */
#include <intLib.h>
#include <iv.h>

/* BSP-specific headers. */
#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

    #include <vxworks/ixp2400/ixdp2400.h>
    #include <vxworks/ixp2400/ixdp2400Pci.h>
    #include <vxworks/ixp2400/ixdp2400misc.h>

#else

    #include <vxworks/ixp2800/ixdp2000.h>
    #include <vxworks/ixp2800/ixdp2000Pci.h>
    #include <vxworks/ixp2800/ixdp2000misc.h>

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */


/* We test to see if the BSP PCI support is enabled. */
#if !defined(INCLUDE_PCI)
#error "Must define INCLUDE_PCI"
#endif /* end !defined(INCLUDE_PCI) */

/* We test to see if the BSP PCI EXTRA MAPPING support is enabled. */
#if !defined(INCLUDE_EXTRA_PCI_MAPPING)
#error "Must define INCLUDE_EXTRA_PCI_MAPPING"
#endif /* end !defined(INCLUDE_EXTRA_PCI_MAPPING) */

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

#include <linux/pci.h>
#include <asm/irq.h>
#include <asm/arch/ixp2000.h>
#include <asm/arch/irqs.h>
#include <asm/arch/pci.h>

#else

#error "PCI support is not enabled on this OS!"

#endif /* _IX_OS_TYPE_ */

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */



/**
 * MACRO NAME: IX_PCI_32BIT_SWAP
 *
 * DESCRIPTION: This macro will perform byte swapping for a 32 bit word and return
 *  the new swapped value
 *
 * @Param:  - INOUT arg_Data - it is a ix_uint32 value that will be byte swapped
 *  inside the macro.
 * @Return: The swapped value will be returned.
 */
#if defined(_IX_RM_IMPL_HARDWARE_) /* && (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) */
    #define IX_PCI_32BIT_SWAP( \
                               arg_Data \
                             ) \
                            (((arg_Data) >> 24) | ((arg_Data) << 24) | \
                            (((arg_Data) & 0x0000ff00) << 8) | \
                            (((arg_Data) & 0x00ff0000) >> 8))
#else
    #define IX_PCI_32BIT_SWAP( \
                               arg_Data \
                             ) \
                            arg_Data
#endif


/* Define macros, registers and interrupt vector for VxSim. */
#if !defined(_IX_RM_IMPL_HARDWARE_)

    /* Register containing upper bits of PCI memory addresses from the CPU. */
    ix_uint32 g_ixPCIAddrExt = 0;
    #define IX_PCI_ADDR_EXT (ix_uint32)&g_ixPCIAddrExt

    /* Vendor and device IDs used to find PCI devices. */
    #define IX_VENDOR_ID 1
    #define IX_DEVICE_ID 1

    /* Bit mask to get upper 3 bits of a remote PCI address. */
    #define IX_PCI_MEM_BASE_MASK 0

    /* Base address of PCI memory as seen from the CPU. */
    #define IX_CPU_PCI_MEM_ADRS 0

    /* Base address of remote PCI CSRs. */
    #define IX_REMOTE_PCI_CSR_BASE 0

    /* BSP-specific register offsets. */
    /* Doorbell register, used to trigger and acknowledge doorbell interrupt. */
    ix_uint32 g_ixPCILocalDoorbell = 0;
    #define IX_PCI_LOCAL_DOORBELL (ix_uint32)&g_ixPCILocalDoorbell
    ix_uint32 g_ixPCILocalDoorbellSetup = 0;
    #define IX_PCI_LOCAL_DOORBELL_SETUP (ix_uint32)&g_ixPCILocalDoorbellSetup
    ix_uint32 g_ixPCIRemoteDoorbell = 0;
    #define IX_PCI_REMOTE_DOORBELL (ix_uint32)&g_ixPCIRemoteDoorbell
    ix_uint32 g_ixPCIRemoteDoorbellSetup = 0;
    #define IX_PCI_REMOTE_DOORBELL_SETUP (ix_uint32)&g_ixPCIRemoteDoorbellSetup
    /**
     * Mailbox register, used to indicate whether to notify Tx or Rx,
     * and whether an error occured.
     */
    ix_uint32 g_ixPCILocalMailbox1 = 0;
    #define IX_PCI_LOCAL_MAILBOX_1 (ix_uint32)&g_ixPCILocalMailbox1
    #define IX_PCI_REMOTE_MAILBOX_1 IX_PCI_LOCAL_MAILBOX_1
    /* Mailbox register, used to contain data address from other side. */
    ix_uint32 g_ixPCILocalMailbox2 = 0;
    #define IX_PCI_LOCAL_MAILBOX_2 (ix_uint32)&g_ixPCILocalMailbox2
    #define IX_PCI_REMOTE_MAILBOX_2 IX_PCI_LOCAL_MAILBOX_2

    /* PCI doorbell interrupt vector */
    #define IX_PCI_DOORBELL_INT 13

    /* Register read/write macros. */
    #define _IX_PCI_REG_WRITE(reg,data)  \
                (*((volatile ix_uint32 *)(reg)) = (data));

    #define _IX_PCI_REG_READ(reg,result) \
                ((result) = *(volatile ix_uint32 *)(reg))


#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

    int isNpuMaster(void)

#else

    int isNPUMaster(void)

#endif /* end (_IX_HARDWARE_TYPE_ != _IX_HW_2800_) */


    {
        return 0;
    }

#else

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

/* Define IXP2400-specific macros, registers and interrupt vector. */
#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

    /* Register read/write macros. */
    #define _IX_PCI_REG_WRITE IXP2400_REG_WRITE
    #define _IX_PCI_REG_READ IXP2400_REG_READ

    /* Register containing upper bits of PCI memory addresses from the CPU. */
    #define IX_PCI_ADDR_EXT IXP2400_PCI_ADDR_EXT

    /* Vendor and device IDs used to find PCI devices. */
    #define IX_VENDOR_ID IXP2400_VENDOR_ID
    #define IX_DEVICE_ID IXP2400_DEVICE_ID

    /* Bit mask to get upper 3 bits of a remote PCI address. */
    #define IX_PCI_MEM_BASE_MASK PCI_MEM_BASE

#if !defined(CPU_PCI_MEM_ADRS)

#error "CPU_PCI_MEM_ADRS is not defined"

#endif

    /* Base address of PCI memory as seen from the CPU. */
    #define IX_CPU_PCI_MEM_ADRS CPU_PCI_MEM_ADRS

    /* Base address of remote PCI CSRs. */
    #define IX_REMOTE_PCI_CSR_BASE SLAVE_PCI_CSR_BASE

    /* BSP-specific register offsets. */
    /* Doorbell register, used to trigger and acknowledge doorbell interrupt. */
    #define IX_PCI_LOCAL_DOORBELL IXP2400_DOORBELL
    #define IX_PCI_LOCAL_DOORBELL_SETUP IXP2400_DOORBELL_SETUP
    #define IX_PCI_REMOTE_DOORBELL (IX_REMOTE_PCI_CSR_BASE + 0x060)
    #define IX_PCI_REMOTE_DOORBELL_SETUP (IX_REMOTE_PCI_CSR_BASE + 0x064)
    /**
     * Mailbox register, used to indicate whether to notify Tx or Rx,
     * and whether an error occured.
     */
    #define IX_PCI_LOCAL_MAILBOX_1 IXP2400_MAILBOX1
    #define IX_PCI_REMOTE_MAILBOX_1 (IX_REMOTE_PCI_CSR_BASE + 0x54)
    /* Mailbox register, used to contain data address from other side. */
    #define IX_PCI_LOCAL_MAILBOX_2 IXP2400_MAILBOX2
    #define IX_PCI_REMOTE_MAILBOX_2 (IX_REMOTE_PCI_CSR_BASE + 0x58)

    /* PCI doorbell interrupt vector */
    #define IX_PCI_DOORBELL_INT 13

#elif(_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

    #define INT_VEC_PIL INT_LVL_PIL

    /* Register read/write macros. */
    #define _IX_PCI_REG_WRITE IXP2000_REG_PUT
    #define _IX_PCI_REG_READ IXP2000_REG_GET

    /* Register containing upper bits of PCI memory addresses from the CPU. */
    #define IX_PCI_ADDR_EXT IXP2000_PCI_ADDR_EXT

    /* Vendor and device IDs used to find PCI devices. */
    #define IX_VENDOR_ID IXP2000_VENDOR_ID
    #define IX_DEVICE_ID IXP2000_DEVICE_ID

    /* Bit mask to get upper 3 bits of a remote PCI address. */
    #define IX_PCI_MEM_BASE_MASK PCI_MEM_BASE

    /* Base address of PCI memory as seen from the CPU. */
    #define IX_CPU_PCI_MEM_ADRS CPU_PCI_MEM_ADRS

    /* Base address of remote PCI CSRs. */
    #define IX_REMOTE_PCI_CSR_BASE 0xFE000

    /* BSP-specific register offsets. */
    /* Doorbell register, used to trigger and acknowledge doorbell interrupt. */
    #define IX_PCI_LOCAL_DOORBELL IXP2000_XSCALE_DOORBELL
    #define IX_PCI_LOCAL_DOORBELL_SETUP IXP2000_XSCALE_DOORBELL_SETUP
    #define IX_PCI_REMOTE_DOORBELL (IX_REMOTE_PCI_CSR_BASE + 0x060)
    #define IX_PCI_REMOTE_DOORBELL_SETUP (IX_REMOTE_PCI_CSR_BASE + 0x064)
    /**
     * Mailbox register, used to indicate whether to notify Tx or Rx,
     * and whether an error occured.
     */
    #define IX_PCI_LOCAL_MAILBOX_1 IXP2000_MAILBOX1
    #define IX_PCI_REMOTE_MAILBOX_1 (IX_REMOTE_PCI_CSR_BASE + 0x54)
    /* Mailbox register, used to contain data address from other side. */
    #define IX_PCI_LOCAL_MAILBOX_2 IXP2000_MAILBOX2
    #define IX_PCI_REMOTE_MAILBOX_2 (IX_REMOTE_PCI_CSR_BASE + 0x58)

    /* PCI doorbell interrupt vector */
    #define IX_PCI_DOORBELL_INT 13
#else

#error "This hardware configuration is not supported"

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    /* Define IXP2400-specific macros, registers and interrupt vector. */
#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

    #define isNPUMaster npu_is_pcimaster

    /* Register read/write macros. */
    #define _IX_PCI_REG_WRITE(reg, data) \
        (*(volatile unsigned long*)(reg) = (data))
    #define _IX_PCI_REG_READ(reg, result) \
        ((result) = *(volatile unsigned long*)(reg))

    /* Register containing upper bits of PCI memory addresses from the CPU. */
    #define IX_PCI_ADDR_EXT IXP2000_PCI_ADDR_EXT

    /* Vendor and device IDs used to find PCI devices. */
    #define IX_VENDOR_ID 0x8086
    #define IX_DEVICE_ID 0x9001

    /* Bit mask to get upper 3 bits of a remote PCI address. */
    #define IX_PCI_MEM_BASE_MASK 0xe0000000

    /* Base address of PCI memory as seen from the CPU. */
    extern unsigned long pci_window_address;
    #define IX_CPU_PCI_MEM_ADRS pci_window_address

    /* Base address of remote PCI CSRs. */
    #define IX_REMOTE_PCI_CSR_BASE 0xfe000

    /* BSP-specific register offsets. */
    /* Doorbell register, used to trigger and acknowledge doorbell interrupt. */
    #define IX_PCI_LOCAL_DOORBELL IXP2000_XSCALE_DOORBELL
    #define IX_PCI_LOCAL_DOORBELL_SETUP IXP2000_XSCALE_DOORBELL_SETUP

    #define IX_PCI_REMOTE_DOORBELL (IX_REMOTE_PCI_CSR_BASE + 0x060)
    #define IX_PCI_REMOTE_DOORBELL_SETUP (IX_REMOTE_PCI_CSR_BASE + 0x064)
    /**
     * Mailbox register, used to indicate whether to notify Tx or Rx,
     * and whether an error occured.
     */
    #define IX_PCI_LOCAL_MAILBOX_1 IXP2000_PCI_MAILBOX1
    #define IX_PCI_REMOTE_MAILBOX_1 (IX_REMOTE_PCI_CSR_BASE + 0x54)
    /* Mailbox register, used to contain data address from other side. */
    #define IX_PCI_LOCAL_MAILBOX_2 IXP2000_PCI_MAILBOX2
    #define IX_PCI_REMOTE_MAILBOX_2 (IX_REMOTE_PCI_CSR_BASE + 0x58)

    /* PCI doorbell interrupt vector */
    #define IX_PCI_DOORBELL_INT IRQ_IXP2000_PCI_DOORBELL

#else

    #define isNpuMaster npu_is_pcimaster

    /* Register read/write macros. */
    #define _IX_PCI_REG_WRITE(reg, data) \
        (*(volatile unsigned long*)(reg) = (data))
    #define _IX_PCI_REG_READ(reg, result) \
        ((result) = *(volatile unsigned long*)(reg))

    /* Register containing upper bits of PCI memory addresses from the CPU. */
    #define IX_PCI_ADDR_EXT IXP2000_PCI_ADDR_EXT

    /* Vendor and device IDs used to find PCI devices. */
    #define IX_VENDOR_ID 0x8086
    #define IX_DEVICE_ID 0x9004

    /* Bit mask to get upper 3 bits of a remote PCI address. */
    #define IX_PCI_MEM_BASE_MASK 0xe0000000

    /* Base address of PCI memory as seen from the CPU. */
    extern unsigned long pci_window_address;
    #define IX_CPU_PCI_MEM_ADRS pci_window_address

    /* Base address of remote PCI CSRs. */
    #define IX_REMOTE_PCI_CSR_BASE 0xfe000

    /* BSP-specific register offsets. */
    /* Doorbell register, used to trigger and acknowledge doorbell interrupt. */
    #define IX_PCI_LOCAL_DOORBELL IXP2000_XSCALE_DOORBELL
    #define IX_PCI_LOCAL_DOORBELL_SETUP IXP2000_XSCALE_DOORBELL_SETUP

    #define IX_PCI_REMOTE_DOORBELL (IX_REMOTE_PCI_CSR_BASE + 0x060)
    #define IX_PCI_REMOTE_DOORBELL_SETUP (IX_REMOTE_PCI_CSR_BASE + 0x064)
    /**
     * Mailbox register, used to indicate whether to notify Tx or Rx,
     * and whether an error occured.
     */
    #define IX_PCI_LOCAL_MAILBOX_1 IXP2000_PCI_MAILBOX1
    #define IX_PCI_REMOTE_MAILBOX_1 (IX_REMOTE_PCI_CSR_BASE + 0x54)
    /* Mailbox register, used to contain data address from other side. */
    #define IX_PCI_LOCAL_MAILBOX_2 IXP2000_PCI_MAILBOX2
    #define IX_PCI_REMOTE_MAILBOX_2 (IX_REMOTE_PCI_CSR_BASE + 0x58)

    /* PCI doorbell interrupt vector */
    #define IX_PCI_DOORBELL_INT IRQ_IXP2000_PCI_DOORBELL

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

#else

#error "PCI support is not enabled on this OS!"

#endif /* _IX_OS_TYPE_ */

#endif /* _IX_RM_IMPL_HARDWARE_ */


/**
 * MACRO NAME: _IX_PCI_GET_CPU_ADDRESS
 *
 * DESCRIPTION: This macro takes the base address of the remote data
 *              and sets the IX_PCI_ADDR_EXT register based on
 *              the first 3 bits of the remote base address.  The
 *              IX_PCI_ADDR_EXT register is used for the upper
 *              bits of PCI memory addresses from the CPU.  The macro
 *              then uses the remote base address as an offset from
 *              the base address of the PCI memory as seen from the CPU.
 *              Use this macro before accessing PCI memory, and after
 *              finishing the PCI acceses, set the IX_PCI_ADDR_EXT
 *              to 0x0.
 *
 * @Param:  - IN arg_pRemoteAddress - base address of remote data.
 *
 * @Return:
 */
#define _IX_PCI_GET_CPU_ADDRESS( \
                                arg_remoteAddress \
                                ) \
    do \
    { \
        /** \
         * Set the upper address bits used for CSR bus direct access to the \
         * PCI bus to the upper 3 bits of the remote base address. \
         */ \
        _IX_PCI_REG_WRITE(IX_PCI_ADDR_EXT, ((ix_uint32)arg_remoteAddress & IX_PCI_MEM_BASE_MASK) >> 16); \
        /** \
         * Since the first 3 bits of the remote address will be taken \
         * from the IX_PCI_ADDR_EXT register that we set above, \
         * we clear the first 3 bits of the remote base address. \
         */ \
        ((ix_uint32)arg_remoteAddress) &= ~IX_PCI_MEM_BASE_MASK; \
        /** \
         * Add the remote base address to the base address of the \
         * PCI memory as seen from the CPU. \
         */ \
        ((ix_uint32)arg_remoteAddress) += IX_CPU_PCI_MEM_ADRS; \
    } \
    while(0)

/**
 * MACRO NAME: _IX_PCI_READ_TX_ERROR_SIGNAL
 *
 * DESCRIPTION: This macro reads the signal value stored in the PCI
 *              controller to check whether an error occurred on the
 *              other side.
 *
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the thread.
 * @Param:  - OUT ix_uint32* arg_pCmd - pointer to where command will
 *          be stored.
 *
 * @Return:
 */

#define _IX_PCI_READ_TX_ERROR_SIGNAL( \
                                arg_pContext, \
                                arg_pCmd \
                                ) \
    do \
    { \
    /** \
     * The least significant bit of the signal indicates whether \
     * there was an error or not. \
     */ \
        (ix_uint32)(*((ix_uint32*)arg_pCmd)) = ((ix_pci_controller*)arg_pContext)->m_SignalValue & IX_PCI_TX_SIGNAL_ERROR; \
    } \
    while(0)


/**
 * MACRO NAME: _IX_PCI_READ_RX_ERROR_SIGNAL
 *
 * DESCRIPTION: This macro reads the signal value stored in the PCI
 *              controller to check whether an error occurred on the
 *              other side.
 *
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the thread.
 * @Param:  - OUT ix_uint32* arg_pCmd - pointer to where command will
 *          be stored.
 *
 * @Return:
 */

#define _IX_PCI_READ_RX_ERROR_SIGNAL( \
                                arg_pContext, \
                                arg_pCmd \
                                ) \
    do \
    { \
    /** \
     * The least significant bit of the signal indicates whether \
     * there was an error or not. \
     */ \
        (ix_uint32)(*((ix_uint32*)arg_pCmd)) = ((ix_pci_controller*)arg_pContext)->m_SignalValue & IX_PCI_RX_SIGNAL_ERROR; \
    } \
    while(0)


/**
 * PCI engines interface functions
 */

/**
 * NAME: _ix_rm_pci_tx_engine_init
 *
 * DESCRIPTION: This function initializes the data structures and memory
 *              for the Tx thread of the PCI communication.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Tx thread.
 *
 * @Return: IX_SUCCESS
 *          IX_RM_ERROR_COMM_PCI_TX_SEMA_CREATION_FAILED
 *          IX_RM_ERROR_COMM_PCI_TX_CTRL_INFO_ALLOC_FAILED
 *
 */
static ix_error _ix_rm_pci_tx_engine_init(void* pContext);

/**
 * NAME: _ix_rm_pci_rx_engine_init
 *
 * DESCRIPTION: This function initializes the data structures and memory
 *              for the Rx thread of the PCI communication.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Rx thread.
 *
 * @Return: IX_SUCCESS
 *          IX_RM_ERROR_COMM_PCI_RX_SEMA_CREATION_FAILED
 *          IX_RM_ERROR_COMM_PCI_RX_HW_FL_CREATE_FAILED
 *          IX_RM_ERROR_COMM_PCI_RX_SW_FL_CREATE_FAILED
 *
 */
static ix_error _ix_rm_pci_rx_engine_init(void* pContext);

/**
 * NAME: _ix_rm_pci_tx_engine_fini
 *
 * DESCRIPTION: This function finalizes the data structures and
 *              memory for the Tx thread of the PCI communication.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Tx thread.
 *
 * @Return: IX_SUCCESS, or an appropriate error code if memory cannot be
 *          freed or the semaphore cannot be freed.
 *
 */
static ix_error _ix_rm_pci_tx_engine_fini(void* pContext);

/**
 * NAME: _ix_rm_pci_rx_engine_fini
 *
 * DESCRIPTION: This function finalizes the data structures and
 *              memory for the Rx thread of the PCI communication.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Rx thread.
 *
 * @Return: IX_SUCCESS, or an appropriate error code if free lists
 *          cannot be deleted or the semaphore cannot be freed.
 *
 */
static ix_error _ix_rm_pci_rx_engine_fini(void* pContext);

/**
 * NAME: _ix_rm_pci_tx_engine_start
 *
 * DESCRIPTION: This function creates and starts the Tx thread.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Tx thread.
 *
 * @Return: IX_SUCCESS
 *          IX_RM_ERROR_COMM_PCI_TX_TH_INIT_FAILED
 *
 */
static ix_error _ix_rm_pci_tx_engine_start(void* pContext);

/**
 * NAME: _ix_rm_pci_rx_engine_start
 *
 * DESCRIPTION: This function creates and starts the Rx thread.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Rx thread.
 *
 * @Return: IX_SUCCESS
 *          IX_RM_ERROR_COMM_PCI_RX_TH_INIT_FAILED
 *
 */
static ix_error _ix_rm_pci_rx_engine_start(void* pContext);

/**
 * NAME: _ix_rm_pci_tx_engine_stop
 *
 * DESCRIPTION: This function stops the Tx thread.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Tx thread.
 *
 * @Return: IX_SUCCESS
 *
 */
static ix_error _ix_rm_pci_tx_engine_stop(void* pContext);

/**
 * NAME: _ix_rm_pci_rx_engine_stop
 *
 * DESCRIPTION: This function stops the Rx thread.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Rx thread.
 *
 * @Return: IX_SUCCESS
 *
 */
static ix_error _ix_rm_pci_rx_engine_stop(void* pContext);

/**
 * NAME: _ix_rm_pci_tx_engine_run
 *
 * DESCRIPTION: This function contains the Tx thread functionality.
 *              It retrieves buffers to transmit, constructs
 *              control information for that buffer, and notifies
 *              the Rx side that a buffer is ready for transfer.
 *              For buffer chains, the function sends each buffer
 *              in the chain, one at a time.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Tx thread.
 *
 * @Return: IX_SUCCESS, or an appropriate error code for failure.
 *
 */
static ix_error _ix_rm_pci_tx_engine_run(void* pContext);

/**
 * NAME: _ix_rm_pci_rx_engine_run
 *
 * DESCRIPTION: This function contains the Rx thread functionality.
 *              It waits on a semaphore until Tx on the other side
 *              signals that it has a buffer ready for transfer.
 *              Once it wakes up, it reads the buffer control info
 *              over the PCI, and uses that info to transfer the
 *              buffer.  After receiving an entire buffer chain,
 *              it sends the buffer chain to the destination commID.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Rx thread.
 *
 * @Return: IX_SUCCESS, or an appropriate error code for failure.
 *
 */
static ix_error _ix_rm_pci_rx_engine_run(void* pContext);

/**
 * NAME: _ix_rm_pci_tx_engine_notify
 *
 * DESCRIPTION: Notifies the Tx thread that Rx on the other side
 *              has finished processing a buffer.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Tx thread.
 *
 * @Return: IX_SUCCESS, or an appropriate error code for failure.
 *
 */
static ix_error _ix_rm_pci_tx_engine_notify(void* pContext);

/**
 * NAME: _ix_rm_pci_rx_engine_notify
 *
 * DESCRIPTION: Notifies the Rx thread that Tx on the other side
 *              has a buffer ready for transfer.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Rx thread.
 *
 * @Return: IX_SUCCESS, or an appropriate error code for failure.
 *
 */
static ix_error _ix_rm_pci_rx_engine_notify(void* pContext);

/**
 * NAME: _ix_rm_pci_buffer_retrieve
 *
 * DESCRIPTION: Retrieves a buffer to be transferred over the PCI.  When
 *              a client wants to shut down the PCI communication, it
 *              should send a NULL packet, something that should make
 *              this function return an error so that the Tx thread will
 *              wake up and exit.
 * @Param:  - IN ix_pci_buffer_data_type* arg_pBufferType - indicates
 *          whether a packet or a message is being transferred.
 * @Param:  - IN ix_buffer_handle* arg_pBufferHandle - pointer to the handle
 *          of the buffer to be transferred.
 * @Param:  - IN ix_uint32* arg_pUserData - pointer to user data.
 * @Param:  - IN ix_communication_id* arg_pCommId - pointer to destination
 *          communication ID.
 *
 * @Return: IX_SUCCESS, or an appropriate error code for failure.
 *
 */
static ix_error _ix_rm_pci_buffer_retrieve(ix_pci_buffer_data_type* arg_pBufferType,
                                               ix_buffer_handle* arg_pBufferHandle,
                                               ix_uint32* arg_pUserData,
                                               ix_communication_id* arg_pCommId);

/**
 * NAME: _ix_rm_pci_signal_rx_subsystem
 *
 * DESCRIPTION: This function, called from the Tx thread, signals
 *              the Rx thread on the other side that a buffer is ready
 *              for transfer, or that an error occured on Tx.
 *
 * @Param:  - IN ix_uint32 arg_cmd - indicates to Rx whether an error
 *          has occurred on Tx or not.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Tx thread.
 *
 * @Return: IX_SUCCESS, or an appropriate error code for failure.
 *
 */
static ix_error _ix_rm_pci_signal_rx_subsystem(ix_uint32 arg_cmd, void* arg_pContext);

/**
 * NAME: _ix_rm_pci_get_tx_control_info
 *
 * DESCRIPTION: This function, called from the Rx thread, gets
 *              the control information for the buffer to be
 *              transferred over PCI.
 *
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Rx thread.
 * @Param:  - OUT ix_pci_control_info* arg_pControlInfo - pointer to
 *          where control info will be stored.
 *
 * @Return: IX_SUCCESS, or an appropriate error code for failure.
 *
 */
static ix_error _ix_rm_pci_get_tx_control_info(void* arg_pContext, ix_pci_control_info* arg_pControlInfo);

/**
 * NAME: _ix_rm_pci_signal_tx_subsystem
 *
 * DESCRIPTION: This function, called from the Rx thread, signals
 *              the Tx thread on the other side that Rx has finished
 *              processing a buffer, or that an error occured on Rx.
 *
 * @Param:  - IN ix_uint32 arg_cmd - indicates to Tx whether an error
 *          has occurred on Rx or not.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Rx thread.
 *
 * @Return: IX_SUCCESS, or an appropriate error code for failure.
 *
 */
static ix_error _ix_rm_pci_signal_tx_subsystem(ix_uint32 arg_cmd, void* arg_pContext);

/**
 * NAME: _ix_rm_pci_copy_meta_from_remote_subsystem
 *
 * DESCRIPTION: This function, called from the Rx thread, copies
 *              the buffer meta-data from the other side.
 *
 * @Param:  - IN ix_pci_buffer_info* arg_pBufferInfo - pointer to
 *          control information for the buffer to be transferred.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Rx thread.
 * @Param:  - OUT void* arg_pLocalDestination - address to where
 *          remote data will be written.
 * @Param:  - OUT ix_uint32* arg_pDataSize - address to where the size of the
 *          data portion of the remote buffer will be written.
 *
 * @Return: IX_SUCCESS, or an appropriate error code for failure.
 *
 */
static ix_error _ix_rm_pci_copy_meta_from_remote_subsystem(
                                               ix_pci_buffer_info* arg_pBufferInfo,
                                               void* arg_pContext,
                                               void* arg_pLocalDestination,
                                               ix_uint32* arg_pDataSize
                                               );
/**
 * NAME: _ix_rm_pci_copy_data_from_remote_subsystem
 *
 * DESCRIPTION: This function, called from the Rx thread, copies
 *              the buffer data from the other side.
 *
 * @Param:  - IN ix_pci_buffer_info* arg_pBufferInfo - pointer to
 *          control information for the buffer to be transferred.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Rx thread.
 * @Param:  - IN ix_uint32 arg_dataSize - size of the remote data.
 * @Param:  - OUT void* arg_pLocalDestination - address to where
 *          remote data will be written.
 *
 * @Return: IX_SUCCESS, or an appropriate error code for failure.
 *
 */
ix_error _ix_rm_pci_copy_data_from_remote_subsystem(
                                               ix_pci_buffer_info* arg_pBufferInfo,
                                               ix_uint32 arg_dataSize,
                                               void* arg_pContext,
                                               void* arg_pLocalDestination
                                               );

/**
 * NAME: ix_pci_doorbell_isr
 *
 * DESCRIPTION: Interrupt service routine for the PCI doorbell.  Reads
 *              the signal value from the local mailbox_1 register
 *              into the PCI controller data member for signal value.
 *              It then notifies the appropriate thread being signalled.
 *
 * @Param:  - IN int arg_Value - required input parameter for VxWorks
 *          ISR - ignored in this function.
 *
 * @Return: IX_SUCCESS, or an appropriate error code for failure.
 *
 */

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

static void ix_pci_doorbell_isr(int arg_Value);

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

static void ix_pci_doorbell_isr(int irq, void* context, struct pt_regs* regs);

#else

#error "PCI support is not enabled on this OS!"

#endif /* _IX_OS_TYPE_ */


/**
 * This is the PCI Tx engine interface
 */
static _ix_rm_pci_engine_interface s_PCITxEngineInterface =
{
    (ix_pci_engine_interface_func_ptr)_ix_rm_pci_tx_engine_init,
    (ix_pci_engine_interface_func_ptr)_ix_rm_pci_tx_engine_fini,
    (ix_pci_engine_interface_func_ptr)_ix_rm_pci_tx_engine_start,
    (ix_pci_engine_interface_func_ptr)_ix_rm_pci_tx_engine_stop,
    (ix_pci_engine_interface_func_ptr)_ix_rm_pci_tx_engine_run,
    (ix_pci_engine_interface_func_ptr)_ix_rm_pci_tx_engine_notify
};


/**
 * This is the PCI Rx engine interface
 */
static _ix_rm_pci_engine_interface s_PCIRxEngineInterface =
{
    (ix_pci_engine_interface_func_ptr)_ix_rm_pci_rx_engine_init,
    (ix_pci_engine_interface_func_ptr)_ix_rm_pci_rx_engine_fini,
    (ix_pci_engine_interface_func_ptr)_ix_rm_pci_rx_engine_start,
    (ix_pci_engine_interface_func_ptr)_ix_rm_pci_rx_engine_stop,
    (ix_pci_engine_interface_func_ptr)_ix_rm_pci_rx_engine_run,
    (ix_pci_engine_interface_func_ptr)_ix_rm_pci_rx_engine_notify
};

/**
 * Global communication controller
 */
ix_pci_controller g_PCICommunicationController;


/**
 * NAME: _ix_rm_pci_communication_init
 *
 * DESCRIPTION: This function will be called by _ix_rm_comunication_init()
 *          at the time the resource manager initializer function is
 *          called. The function will initialize all the PCI communication.
 * @Param:  - IN arg_pContext - context that has been passed to RM at the initialization
 *          of the service.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_pci_communication_init(void* arg_pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_communication_init";

    ix_error err = IX_SUCCESS;
    ix_subsystem_type subsystemType;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* Reset and initialize global controller data. */
    ix_ossl_memset(&g_PCICommunicationController, 0, sizeof(g_PCICommunicationController));

    g_PCICommunicationController.m_pTxInterface = &s_PCITxEngineInterface;
    g_PCICommunicationController.m_pRxInterface = &s_PCIRxEngineInterface;
    g_PCICommunicationController.m_TxBufferRetrieve = _ix_rm_pci_buffer_retrieve;

    /* Create the remote Mailbox  registers semaphore */
    err = ix_ossl_sem_init(IX_OSSL_SEM_AVAILABLE, &g_PCICommunicationController.m_MBRegsSemaphore);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_INIT_FAILED, s_aFunctionIdentifier,
                        "the creation of the remote Mailbox registers semaphore failed!", err);
    }

    /* Set the subsystem type - the master NPU is Egress. */
#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

        if(isNpuMaster())

#else

        if(isNPUMaster())

#endif /* end (_IX_HARDWARE_TYPE_ != _IX_HW_2800_) */

        subsystemType = IX_SUBSYSTEM_TYPE_EGRESS;
    else
        subsystemType = IX_SUBSYSTEM_TYPE_INGRESS;

    /* Set the base addresses for VxSim testing. */
#if !defined(_IX_RM_IMPL_HARDWARE_)
    {
        /* Set the remote CSR base address */
        g_PCICommunicationController.m_RemoteCSRBase = 0;

        /* Set the remote SRAM base address */
        g_PCICommunicationController.m_RemoteSRAMBase = 0;

        /* Set the remote DRAM base address */
        g_PCICommunicationController.m_RemoteDRAMBase = 0;
    }
    /**
     * Find the PCI devices and set the base addreses of the
     * remote CSRs, SRAM, and DRAM.
     */
#else

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

    {
        int busNo;
        int deviceNo;
        int funcNo;

        /* Find the remote PCI device. */
        if(pciFindDevice(IX_VENDOR_ID, IX_DEVICE_ID,
            (subsystemType == IX_SUBSYSTEM_TYPE_EGRESS) ? 0 : 1,
            &busNo, &deviceNo, &funcNo) != OK)
        {
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_GET_REMOTE_DEV_FAILED, s_aFunctionIdentifier,
                        "obtaining the remote PCI device failed!", err);
        }

        /* fill in the remote memory information */

        /* get the remote CSR base address */
        pciConfigInLong(busNo, deviceNo, funcNo, 0x10,
            &g_PCICommunicationController.m_RemoteCSRBase);

        /*
        ix_ossl_message_log("\tCSR Bar: 0x%08lx\n", g_PCICommunicationController.m_RemoteCSRBase);
        */

        /* do the fancy PCI masking */
        g_PCICommunicationController.m_RemoteCSRBase &= ~0x0000000F;

        {
            ix_uint32 remoteCSRBase;

            ix_uint32 temp = 0;

            ix_uint32 extRegVal;

            /* Get remote CSR base address from controller. */
            remoteCSRBase = g_PCICommunicationController.m_RemoteCSRBase;

            /* Disable all PCI interrupts. */
            intDisable(INT_VEC_PIL);

            _IX_PCI_REG_READ(IX_PCI_ADDR_EXT, extRegVal);

            /* Get the address for desired PCI memory offset as seen from the CPU. */
            _IX_PCI_GET_CPU_ADDRESS(remoteCSRBase);

            _IX_PCI_REG_WRITE(remoteCSRBase + IX_PCI_REMOTE_MAILBOX_1, IX_PCI_32BIT_SWAP(temp));

            #if defined(IX_PCI_A0_SILICON)
                /* A read cycle to work around the A0 bug that requires a read cycle between writes. */
                _IX_PCI_REG_READ(remoteCSRBase + IX_PCI_REMOTE_MAILBOX_1, temp);
            #endif /* end defined(IX_PCI_A0_SILICON) */

            /* Reset the upper word used in PCI addresses back to its original value. */
            _IX_PCI_REG_WRITE(IX_PCI_ADDR_EXT, extRegVal);

            /* Re-enable all PCI interrupts. */
            intEnable(INT_VEC_PIL);

        }

        /* get the remote SRAM base address */
        pciConfigInLong(busNo, deviceNo, funcNo, 0x14,
            &g_PCICommunicationController.m_RemoteSRAMBase);
        
        /*
        ix_ossl_message_log("\tSRAM Bar:0x%08lx\n", g_PCICommunicationController.m_RemoteSRAMBase);
        */

        /* do the fancy PCI masking */
        g_PCICommunicationController.m_RemoteSRAMBase &= ~0x0000000F;

        /* get the remote DRAM base address */
        pciConfigInLong(busNo, deviceNo, funcNo, 0x18,
            &g_PCICommunicationController.m_RemoteDRAMBase);

        /*
        ix_ossl_message_log("\tDRAM Bar:0x%08lx\n", g_PCICommunicationController.m_RemoteDRAMBase);
        */

        /* do the fancy PCI masking */
        g_PCICommunicationController.m_RemoteDRAMBase &= ~0x0000000F;

    }

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    {
        struct pci_dev* pDev;

        /* Find the remote PCI device. */
        pDev = pci_find_device(IX_VENDOR_ID, IX_DEVICE_ID, NULL);
        if(pDev == NULL)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_GET_REMOTE_DEV_FAILED, s_aFunctionIdentifier,
                        "obtaining the remote PCI device failed!", err);
/*
        if(subsystemType == IX_SUBSYSTEM_TYPE_INGRESS)
        {
            pDev = pci_find_device(IX_VENDOR_ID, IX_DEVICE_ID, pDev);
            if(pDev == NULL)
                return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_GET_REMOTE_DEV_FAILED, s_aFunctionIdentifier,
                        "obtaining the remote PCI device failed!", err);
        }
*/
        /* fill in the remote memory information */

        /* get the remote CSR base address */
        pci_read_config_dword(pDev, 0x10, (u32*)&g_PCICommunicationController.m_RemoteCSRBase);

        /* do the fancy PCI masking */
        g_PCICommunicationController.m_RemoteCSRBase &= ~0x0000000F;

        {
            ix_uint32 remoteCSRBase;

            ix_uint32 temp = 0;

            ix_uint32 extRegVal;

            unsigned long flags;

            /* Get remote CSR base address from controller. */
            remoteCSRBase = g_PCICommunicationController.m_RemoteCSRBase;

            /* Disable all PCI interrupts. */
            local_irq_save(flags);

            _IX_PCI_REG_READ(IX_PCI_ADDR_EXT, extRegVal);

            /* Get the address for desired PCI memory offset as seen from the CPU. */
            _IX_PCI_GET_CPU_ADDRESS(remoteCSRBase);

            _IX_PCI_REG_WRITE(remoteCSRBase + IX_PCI_REMOTE_MAILBOX_1, IX_PCI_32BIT_SWAP(temp));

            #if defined(IX_PCI_A0_SILICON)
                /* A read cycle to work around the A0 bug that requires a read cycle between writes. */
                _IX_PCI_REG_READ(remoteCSRBase + IX_PCI_REMOTE_MAILBOX_1, temp);
            #endif /* end defined(IX_PCI_A0_SILICON) */

            /* Reset the upper word used in PCI addresses back to its original value. */
            _IX_PCI_REG_WRITE(IX_PCI_ADDR_EXT, extRegVal);

            /* Re-enable all PCI interrupts. */
            local_irq_restore(flags);

        }


        /* get the remote SRAM base address */
        pci_read_config_dword(pDev, 0x14, (u32*)&g_PCICommunicationController.m_RemoteSRAMBase);

        /* do the fancy PCI masking */
        g_PCICommunicationController.m_RemoteSRAMBase &= ~0x0000000F;

        /* get the remote DRAM base address */
        pci_read_config_dword(pDev, 0x18, (u32*)&g_PCICommunicationController.m_RemoteDRAMBase);

        /* do the fancy PCI masking */
        g_PCICommunicationController.m_RemoteDRAMBase &= ~0x0000000F;

    }

#else

#error "PCI support is not enabled on this OS!"

#endif /* _IX_OS_TYPE_ */

#endif /* end !defined(_IX_RM_IMPL_HARDWARE_) */


    /* Call the initialization function for the Rx thread. */
    err = g_PCICommunicationController.m_pRxInterface->m_InitEngineFuncPtr(&g_PCICommunicationController);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_RX_INIT_FAILED, s_aFunctionIdentifier,
                        "the initialization of the PCI Rx engine failed!", err);
    }

    /* Call the function to start the Rx thread. */
    err = g_PCICommunicationController.m_pRxInterface->m_StartEngineFuncPtr(&g_PCICommunicationController);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_RX_START_FAILED, s_aFunctionIdentifier,
                        "starting the PCI Rx engine failed!", err);
    }

    /* Call the initialization function for the Tx thread. */
    err = g_PCICommunicationController.m_pTxInterface->m_InitEngineFuncPtr(&g_PCICommunicationController);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_TX_INIT_FAILED, s_aFunctionIdentifier,
                        "the initialization of the PCI Tx engine failed!", err);
    }

    /* Call the function to start the Tx thread. */
    err = g_PCICommunicationController.m_pTxInterface->m_StartEngineFuncPtr(&g_PCICommunicationController);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_TX_START_FAILED, s_aFunctionIdentifier,
                        "starting the PCI Tx engine failed!", err);
    }

    /* Connect and enable the interrupt, if we are running on hardware. */
#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

    {
        STATUS ret_status = OK;

        /* enable the doorbell interrupt */
        /* *((volatile ix_uint32 *)IX_PCI_LOCAL_DOORBELL_SETUP) = 0x0; */

        ret_status = intConnect(INUM_TO_IVEC(IX_PCI_DOORBELL_INT), ix_pci_doorbell_isr, 0);
        if(ret_status != OK)
        {
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_INT_CONNECT_FAILED, s_aFunctionIdentifier,
                            "connecting the PCI doorbell interrupt failed!", err);
        }
        ret_status = intEnable(IX_PCI_DOORBELL_INT);
        if(ret_status != OK)
        {
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_INT_ENABLE_FAILED, s_aFunctionIdentifier,
                            "enabling the PCI doorbell interrupt failed!", err);
        }
    }

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    {
        int status;

        status = request_irq(IX_PCI_DOORBELL_INT, ix_pci_doorbell_isr, SA_INTERRUPT, "Resource Manager", 0);
        if(status != 0)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_INT_CONNECT_FAILED, s_aFunctionIdentifier,
                            "connecting the PCI doorbell interrupt failed!", err);

        enable_irq(IX_PCI_DOORBELL_INT);
    }

#else

#error "PCI support is not enabled on this OS!"

#endif /* _IX_OS_TYPE_ */

#endif /* end  defined(_IX_RM_IMPL_HARDWARE_) */

    return err;
}


/**
 * NAME: _ix_rm_pci_communication_fini
 *
 * DESCRIPTION: This function will be called by _ix_rm_communication_fini()
 *          at the time the resource manager finalizer function is called.
 *          The function will provide cleanup for PCI communication.
 * @Param:  - IN arg_pContext - context that has been passed to RM at the initialization
 *          of the service.
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_pci_communication_fini(void* arg_pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_communication_fini";

    ix_error err = IX_SUCCESS;

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

    STATUS ret_status = OK;

#endif

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

        /* Disable the interrupt. */
        ret_status = intDisable(IX_PCI_DOORBELL_INT);
        if(ret_status != OK)
        {
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_INT_DISABLE_FAILED, s_aFunctionIdentifier,
                            "disabling the PCI doorbell interrupt failed!", err);
        }

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        /* Unregister the software interrupt handler */
        disable_irq(IX_PCI_DOORBELL_INT);
        free_irq(IX_PCI_DOORBELL_INT, 0);

#else

#error "PCI support is not enabled on this OS!"

#endif /* _IX_OS_TYPE_ */

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    /* Call the function to stop the Tx thread. */
    err = g_PCICommunicationController.m_pTxInterface->m_StopEngineFuncPtr(&g_PCICommunicationController);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_TX_STOP_FAILED, s_aFunctionIdentifier,
                        "stopping the PCI Tx engine failed!", err);
    }

    /* Call the function to clean up the Tx thread. */
    err = g_PCICommunicationController.m_pTxInterface->m_FiniEngineFuncPtr(&g_PCICommunicationController);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_TX_FINI_FAILED, s_aFunctionIdentifier,
                        "the finalization of the PCI Tx engine failed!", err);
    }

    /* Call the function to stop the Rx thread. */
    err = g_PCICommunicationController.m_pRxInterface->m_StopEngineFuncPtr(&g_PCICommunicationController);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_RX_STOP_FAILED, s_aFunctionIdentifier,
                        "stopping the PCI Rx engine failed!", err);
    }

    /* Call the function to clean up the Rx thread. */
    err = g_PCICommunicationController.m_pRxInterface->m_FiniEngineFuncPtr(&g_PCICommunicationController);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_RX_FINI_FAILED, s_aFunctionIdentifier,
                        "the finalization of the PCI Rx engine failed!", err);
    }

    return err;
}


/**
 * Static internal functions
 */

/**
 * NAME: _ix_rm_pci_tx_engine_init
 *
 * DESCRIPTION: This function initializes the data structures and memory
 *              for the Tx thread of the PCI communication.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Tx thread.
 *
 * @Return: IX_SUCCESS
 *          IX_RM_ERROR_COMM_PCI_TX_SEMA_CREATION_FAILED
 *          IX_RM_ERROR_COMM_PCI_TX_CTRL_INFO_ALLOC_FAILED
 *
 */
ix_error _ix_rm_pci_tx_engine_init(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_tx_engine_init";

    ix_error err = IX_SUCCESS;

    ix_pci_controller* pPCIController = 
        (ix_pci_controller*)pContext;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* Create the semaphore for the Tx thread. */
    err = ix_ossl_sem_init(IX_OSSL_SEM_UNAVAILABLE, &pPCIController->m_TxSemaphore);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_TX_SEMA_CREATION_FAILED, s_aFunctionIdentifier, 
                        "the creation of the PCI Tx semaphore failed!", err);
    }

    /* Allocate DRAM memory for the Tx thread control info. */
    err = ix_rm_mem_alloc(IX_MEMORY_TYPE_DRAM, 0, sizeof(ix_pci_control_info),(void**)&pPCIController->m_pTxPCIControlInfo); 
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_TX_CTRL_INFO_ALLOC_FAILED, s_aFunctionIdentifier, 
                        "the creation of the PCI Tx control info failed!", err);
        goto label_1;
    }

    /* Get physical offset of the Tx thread control info. */
    {
        ix_memory_type memType;
        ix_uint32 channel;
        void* pPhysicalAddress;
        ix_uint32 i;
        ix_memory_info memoryInfo;

        err = ix_rm_get_phys_offset(pPCIController->m_pTxPCIControlInfo, &memType, 
            &channel, &pPCIController->m_TxPCIControlInfoOffset, &pPhysicalAddress);
        if(err != IX_SUCCESS)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_TX_CTRL_INFO_GET_PHYS_OFFSET_FAILED, s_aFunctionIdentifier, 
                        "getting the physical offset of the PCI Tx control info failed!", err);
            goto label_2;
        }

        /**
         * We are getting the offset from the beginning DRAM 
         * memory, not from the beginning of the channel
         */

        /* We make the assumption that memory channels are contiguous */
        for(i = 0; i < channel; i++)
        {
            ix_rm_mem_info(memType, i, &memoryInfo);
            pPCIController->m_TxPCIControlInfoOffset += memoryInfo.m_TotalSize;
        }

        /* Set physical offset to be the virtual address, in software testing. */
        #if !defined(_IX_RM_IMPL_HARDWARE_)
            pPCIController->m_TxPCIControlInfoOffset = (ix_uint32)pPCIController->m_pTxPCIControlInfo;
        #endif /* end !defined(_IX_RM_IMPL_HARDWARE_) */

    }


    return err;

label_2:

    if(pPCIController->m_pTxPCIControlInfo != 0)
    {
        ix_rm_mem_free(pPCIController->m_pTxPCIControlInfo);
        pPCIController->m_pTxPCIControlInfo = 0;
    }

label_1:

    /* Free the Tx semaphore if it was created. */
    if(pPCIController->m_TxSemaphore != 0)
    {
        ix_ossl_sem_fini(pPCIController->m_TxSemaphore);
        pPCIController->m_TxSemaphore = 0;
    }

    return err;
}

/**
 * NAME: _ix_rm_pci_rx_engine_init
 *
 * DESCRIPTION: This function initializes the data structures and memory
 *              for the Rx thread of the PCI communication.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Rx thread.
 *
 * @Return: IX_SUCCESS
 *          IX_RM_ERROR_COMM_PCI_RX_SEMA_CREATION_FAILED
 *          IX_RM_ERROR_COMM_PCI_RX_HW_FL_CREATE_FAILED
 *          IX_RM_ERROR_COMM_PCI_RX_SW_FL_CREATE_FAILED
 *
 */
ix_error _ix_rm_pci_rx_engine_init(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_rx_engine_init";

    ix_error err = IX_SUCCESS;

    ix_pci_controller* pPCIController = 
        (ix_pci_controller*)pContext;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* Create the semaphore for the Rx thread. */
    err = ix_ossl_sem_init(IX_OSSL_SEM_UNAVAILABLE, &pPCIController->m_RxSemaphore);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_RX_SEMA_CREATION_FAILED, s_aFunctionIdentifier, 
                        "the creation of the PCI Rx semaphore failed!", err);
    }

    /* Create the HW free list for the Rx thread. */
    err = ix_rm_hw_buffer_free_list_create(IX_PCI_HW_FL_SIZE, IX_PCI_MAX_HW_META_SIZE, 
                                           IX_PCI_MAX_HW_DATA_SIZE, 0, 0, &pPCIController->m_hPCIHwFreeList);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_RX_HW_FL_CREATE_FAILED, s_aFunctionIdentifier, 
                        "the creation of the PCI Rx HW free list failed!", err);
        goto label_1;
    }

    /* Create the SW free list for the Rx thread. */
    err = ix_rm_sw_buffer_free_list_create(IX_PCI_MAX_SW_META_SIZE, IX_MEMORY_TYPE_DRAM, 
                                           IX_PCI_MAX_SW_DATA_SIZE, IX_MEMORY_TYPE_DRAM, 
                                           IX_PCI_SW_FL_SIZE,  &pPCIController->m_hPCISwFreeList);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_RX_SW_FL_CREATE_FAILED, s_aFunctionIdentifier, 
                        "the creation of the PCI Rx SW free list failed!", err);
        goto label_2;
    }

    
    /**
     * Initialize the user hardware free list data
     */
    g_PCICommunicationController.m_hUserHwFreeList = 0;

    return err;

label_2:
    /* Delete the HW free list. */
    ix_rm_buffer_free_list_delete(pPCIController->m_hPCIHwFreeList);

label_1:
    /* Free the Rx semaphore if it was created. */
    if(pPCIController->m_RxSemaphore != 0)
    {
        ix_ossl_sem_fini(pPCIController->m_RxSemaphore);
        pPCIController->m_RxSemaphore = 0;
    }

    return err;
}


/**
 * NAME: _ix_rm_pci_tx_engine_fini
 *
 * DESCRIPTION: This function finalizes the data structures and
 *              memory for the Tx thread of the PCI communication.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Tx thread.
 *
 * @Return: IX_SUCCESS, or an appropriate error code if free lists
 *          cannot be deleted or the semaphore cannot be freed.
 *
 */
ix_error _ix_rm_pci_tx_engine_fini(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_tx_engine_fini";

    ix_error err = IX_SUCCESS;

    ix_pci_controller* pPCIController = 
        (ix_pci_controller*)pContext;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* Free the memory allocated to the Tx control info. */
    err = ix_rm_mem_free(pPCIController->m_pTxPCIControlInfo); 
    if(err != IX_SUCCESS)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_TX_CTRL_INFO_FREE_FAILED, s_aFunctionIdentifier, 
                        "freeing the memory for the Tx control info failed!", err);

    /* Free the semaphore for the Tx thread. */
    err = ix_ossl_sem_fini(pPCIController->m_TxSemaphore);
    if(err != IX_SUCCESS)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_TX_SEMA_FINI_FAILED, s_aFunctionIdentifier, 
                        "Tx semaphore finalization failed!", err);

    return err;

}

/**
 * NAME: _ix_rm_pci_rx_engine_fini
 *
 * DESCRIPTION: This function finalizes the data structures and
 *              memory for the Rx thread of the PCI communication.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Rx thread.
 *
 * @Return: IX_SUCCESS, or an appropriate error code if free lists
 *          cannot be deleted or the semaphore cannot be freed.
 *
 */
ix_error _ix_rm_pci_rx_engine_fini(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_rx_engine_fini";

    ix_error err = IX_SUCCESS;

    ix_pci_controller* pPCIController = 
        (ix_pci_controller*)pContext;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /**
     * Finalize the user hardware free list data
     */
    g_PCICommunicationController.m_hUserHwFreeList = 0;

    /* Delete the SW free list for the Rx thread. */
    err = ix_rm_buffer_free_list_delete(pPCIController->m_hPCISwFreeList);
    if(err != IX_SUCCESS)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_RX_SW_FL_DELETE_FAILED, s_aFunctionIdentifier, 
                        "the deletion of the PCI Rx SW free list failed!", err);

    /* Delete the HW free list for the Rx thread. */
    err = ix_rm_buffer_free_list_delete(pPCIController->m_hPCIHwFreeList);
    if(err != IX_SUCCESS)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_RX_HW_FL_DELETE_FAILED, s_aFunctionIdentifier, 
                        "the deletion of the PCI Rx HW free list failed!", err);

    /* Free the semaphore for the Rx thread. */
    err = ix_ossl_sem_fini(pPCIController->m_RxSemaphore);
    if(err != IX_SUCCESS)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_RX_SEMA_FINI_FAILED, s_aFunctionIdentifier, 
                        "Rx semaphore finalization failed!", err);

    return err;
}


/**
 * NAME: _ix_rm_pci_rx_engine_start
 *
 * DESCRIPTION: This function creates and starts the Rx thread.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Rx thread.
 *
 * @Return: IX_SUCCESS
 *          IX_RM_ERROR_COMM_PCI_RX_TH_INIT_FAILED
 *
 */
ix_error _ix_rm_pci_rx_engine_start(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_rx_engine_start";

    ix_error err = IX_SUCCESS;

    ix_pci_controller* pPCIController = 
        (ix_pci_controller*)pContext;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* start the engine */

    /* create the PCI Rx thread */
    err = ix_ossl_thread_create((ix_ossl_thread_entry_point_t)pPCIController->m_pRxInterface->m_RunEngineFuncPtr,
        pContext, "RmPciRx", &pPCIController->m_RxThread);
    if(err !=IX_SUCCESS)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_RX_TH_INIT_FAILED, s_aFunctionIdentifier, 
            "failed to create the PCI Rx thread!", 0);

    /* set the PCI Rx thread priority to high */
    err = ix_ossl_thread_set_priority(pPCIController->m_RxThread, IX_OSSL_THREAD_PRI_HIGH-3);
    if(err != IX_SUCCESS)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_RX_TH_INIT_FAILED, s_aFunctionIdentifier, 
            "failed to set the PCI Rx thread priority!", 0);

    return err;
}


/**
 * NAME: _ix_rm_pci_tx_engine_start
 *
 * DESCRIPTION: This function creates and starts the Tx thread.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Tx thread.
 *
 * @Return: IX_SUCCESS
 *          IX_RM_ERROR_COMM_PCI_TX_TH_INIT_FAILED
 *
 */
ix_error _ix_rm_pci_tx_engine_start(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_tx_engine_start";

    ix_error err = IX_SUCCESS;

    ix_pci_controller* pPCIController = 
        (ix_pci_controller*)pContext;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* start the engine */

    /* create the PCI Tx thread */
    err = ix_ossl_thread_create((ix_ossl_thread_entry_point_t)pPCIController->m_pTxInterface->m_RunEngineFuncPtr,
        pContext, "RmPciTx", &pPCIController->m_TxThread);
    if(err !=IX_SUCCESS)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_TX_TH_INIT_FAILED, s_aFunctionIdentifier, 
            "failed to create the PCI Tx thread!", 0);

    /* set the PCI Tx thread priority to high */
    err = ix_ossl_thread_set_priority(pPCIController->m_TxThread, IX_OSSL_THREAD_PRI_HIGH-3);
    if(err !=IX_SUCCESS)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_TX_TH_INIT_FAILED, s_aFunctionIdentifier, 
            "failed to set the PCI Tx thread priority!", 0);

    return err;
}

/**
 * NAME: _ix_rm_pci_tx_engine_stop
 *
 * DESCRIPTION: This function stops the Tx thread.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Tx thread.
 *
 * @Return: IX_SUCCESS
 *
 */
ix_error _ix_rm_pci_tx_engine_stop(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_tx_engine_stop";

    ix_error err = IX_SUCCESS;

    ix_pci_controller* pPCIController = 
        (ix_pci_controller*)pContext;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    pPCIController->m_TxTerminateFlag = 1;

    
    /**
     * Set the signal value for this subsystem to an error, so that
     * Tx will not process any more data.
     */
    pPCIController->m_SignalValue = IX_PCI_RX_SIGNAL_ERROR;

    /* wake up the thread if waiting for return from Rx */
    err = pPCIController->m_pTxInterface->m_NotifyEngineFuncPtr(pContext);

    /**
     * Send an erroneous buffer to trigger an error on the PCI communication.
     * This way Tx will exit its loop gracefully if it is waiting for data.
     * 
     */

    err = ix_rm_packet_send_wait(IX_RM_COMM_MAKE_ID(100, 1, 0), 
                                 IX_NULL_BUFFER_HANDLE,
                                 0,
                                 IX_OSSL_WAIT_FOREVER); 

    /* wait for the thread to terminate */
    while(pPCIController->m_TxTerminateFlag != 0)
    {
        ix_ossl_sleep(1);
    };

    return err;
}


/**
 * NAME: _ix_rm_pci_rx_engine_stop
 *
 * DESCRIPTION: This function stops the Rx thread.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Rx thread.
 *
 * @Return: IX_SUCCESS
 *
 */
ix_error _ix_rm_pci_rx_engine_stop(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_rx_engine_stop";

    ix_error err = IX_SUCCESS;

    ix_pci_controller* pPCIController = 
        (ix_pci_controller*)pContext;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    pPCIController->m_RxTerminateFlag = 1;

    /**
     * Set the signal value for this subsystem to an error, so that
     * Rx will not process any more data.
     */
    pPCIController->m_SignalValue = IX_PCI_TX_SIGNAL_ERROR;

    /* wake up the thread if waiting for data */
    err = pPCIController->m_pRxInterface->m_NotifyEngineFuncPtr(pContext);

    /* wait for the thread to terminate */
    while(pPCIController->m_RxTerminateFlag != 0)
    {
        ix_ossl_sleep(1);
    };

    return err;
}

/**
 * NAME: _ix_rm_pci_tx_engine_run
 *
 * DESCRIPTION: This function contains the Tx thread functionality.
 *              It retrieves buffers to transmit, constructs
 *              control information for that buffer, and notifies
 *              the Rx side that a buffer is ready for transfer.
 *              For buffer chains, the function sends each buffer
 *              in the chain, one at a time.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Tx thread.
 *
 * @Return: IX_SUCCESS, or an appropriate error code for failure.
 *
 */
ix_error _ix_rm_pci_tx_engine_run(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_tx_engine_run";

    ix_error err = IX_SUCCESS;

    ix_pci_controller* pPCIController = 
        (ix_pci_controller*)pContext;
    ix_pci_control_info* pPCIControlInfo = pPCIController->m_pTxPCIControlInfo;

    ix_pci_buffer_data_type aPCIBufferDataType;
    ix_buffer_handle hBuffer = IX_NULL_BUFFER_HANDLE;
    ix_buffer_handle hChain = IX_NULL_BUFFER_HANDLE;
    ix_uint32 userData;
    ix_communication_id commId;

    ix_memory_type memType; 
    void* pPhysicalAddress;
    ix_uint32 channel;

    ix_memory_info memoryInfo;
    ix_uint32 i;

    void* pData;

    _ix_buffer_free_list* pFreeList;

    ix_uint32 command;
    ix_uint16 offset = 0;

#if defined(IX_PCI_TEST_INTERRUPTS)
    static ix_uint32 s_txIntCount = 0;
#endif /* defined(IX_PCI_TEST_INTERRUPTS) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(IX_PCI_TEST_INTERRUPTS)

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

    if(isNpuMaster())

#else

    if(isNPUMaster())

#endif /* end (_IX_HARDWARE_TYPE_ != _IX_HW_2800_) */

        _ix_rm_pci_signal_rx_subsystem(0, (void*)pPCIController);

#endif /* defined(IX_PCI_TEST_INTERRUPTS) */

    while(pPCIController->m_TxTerminateFlag == 0)
    {

#if defined(IX_PCI_TEST_INTERRUPTS)
       
        ix_ossl_sem_take(pPCIController->m_TxSemaphore, IX_OSSL_WAIT_FOREVER);

        if((++s_txIntCount % 100000) == 0)
            ix_ossl_message_log("int count: %lu\n", s_txIntCount);
        _ix_rm_pci_signal_rx_subsystem(0, (void*)pPCIController);        
        
        
        
        
#else

        /* Retrieve the buffer to be transferred over PCI. */
        err = pPCIController->m_TxBufferRetrieve(&aPCIBufferDataType, &hBuffer, &userData, &commId);
        if((err != IX_SUCCESS) || (hBuffer == IX_NULL_BUFFER_HANDLE))
            continue;

        /* Initialize the control info common to an entire chain. */
        pPCIControlInfo->m_PCIBufferDataType = aPCIBufferDataType;
        pPCIControlInfo->m_UserData = userData;
        pPCIControlInfo->m_DestinationCommId = commId;
        
        /**
         * In the case of a buffer chain, we have to iterate through the chain
         * and send each buffer at a time. We have to be careful to synchronize 
         * with the receiving side.
         */

        /* We unlink this buffer from the chain. */
        err = ix_rm_buffer_unlink(&hBuffer, &hChain);
        if(err != IX_SUCCESS)
            goto label_1;


        while(hChain != IX_NULL_BUFFER_HANDLE)
        {
            /* Now fill the control information structure. */
            pPCIControlInfo->m_BufferInfo.m_EOPFlag = 0;
            pPCIControlInfo->m_BufferInfo.m_BufferType = IX_RM_BUFFER_GET_TYPE(hBuffer);
            err = ix_rm_buffer_get_meta(hBuffer, &pData);
            if(err != IX_SUCCESS)
                goto label_1;

            /* retrieve the offset for hardware buffers */
            if(pPCIControlInfo->m_BufferInfo.m_BufferType == IX_BUFFER_TYPE_HARDWARE)
                offset = ((ix_hw_buffer_meta*)pData)->m_Offset;

            /* Get physical offset into the type of memory of the meta and write it into the control info. */
            err = ix_rm_get_phys_offset(pData, &memType,
                &channel, &pPCIControlInfo->m_BufferInfo.m_MetaOffset, &pPhysicalAddress);
            if(err != IX_SUCCESS)
                goto label_1;

            /**
             * We are getting the offset from the beginning of that 
             * type of memory, not from the beginning of the channel
             */

            /* We make the assumption that memory channels are contiguous */
            for(i = 0; i < channel; i++)
            {
                ix_rm_mem_info(memType, i, &memoryInfo);
                pPCIControlInfo->m_BufferInfo.m_MetaOffset += memoryInfo.m_TotalSize;
            }

            /* Set physical offset to be the same as virtual address, in software testing. */
            #if !defined(_IX_RM_IMPL_HARDWARE_)
                pPCIControlInfo->m_BufferInfo.m_MetaOffset = (ix_uint32)pData;
            #endif /* end !defined(_IX_RM_IMPL_HARDWARE_) */

            err = ix_rm_buffer_get_data(hBuffer, &pData);
            if(err != IX_SUCCESS)
                goto label_1;

            /* adjust the offset for hardware buffers */
            if(pPCIControlInfo->m_BufferInfo.m_BufferType == IX_BUFFER_TYPE_HARDWARE)
                pData = (ix_uint8*)pData + offset;

            /* Get physical offset into the type of memory of the data and write it into the control info. */
            err = ix_rm_get_phys_offset(pData, &memType,
                &channel, &pPCIControlInfo->m_BufferInfo.m_DataOffset, &pPhysicalAddress);
            if(err != IX_SUCCESS)
                goto label_1;

            /**
             * We are getting the offset from the beginning of that 
             * type of memory, not from the beginning of the channel
             */

            /* We make the assumption that memory channels are contiguous */
            for(i = 0; i < channel; i++)
            {
                ix_rm_mem_info(memType, i, &memoryInfo);
                pPCIControlInfo->m_BufferInfo.m_DataOffset += memoryInfo.m_TotalSize;
            }

            /* Set physical offset to be the same as virtual address, in software testing. */
            #if !defined(_IX_RM_IMPL_HARDWARE_)
                pPCIControlInfo->m_BufferInfo.m_DataOffset = (ix_uint32)pData;
            #endif /* end !defined(_IX_RM_IMPL_HARDWARE_) */

            /**
             * Get the meta size, meta memory type, data size,
             * and data memory type from the buffer's free list.
             */
            pFreeList = IX_RM_BUF_GET_FL_ADDRESS(hBuffer);
            pPCIControlInfo->m_BufferInfo.m_MetaSize = 1 << pFreeList->m_MetaElementShift;
            pPCIControlInfo->m_BufferInfo.m_MetaMemoryType = pFreeList->m_MetaMemoryType;
            pPCIControlInfo->m_BufferInfo.m_DataSize = 1 << pFreeList->m_DataElementShift;
            pPCIControlInfo->m_BufferInfo.m_DataMemoryType = pFreeList->m_DataMemoryType;

            /* Write to the mailbox register on the Rx side, and signal the Rx PCI interrupt */
            _ix_rm_pci_signal_rx_subsystem(IX_PCI_TX_SIGNAL_OK, (void*)pPCIController);

            /* Wait for completion of transfer. */
            err = ix_ossl_sem_take(pPCIController->m_TxSemaphore, IX_PCI_TX_TIMEOUT);
            /* Check for timeout. */
            if(err != IX_SUCCESS)
                goto label_1;

            /* Check for errors from the Rx side. */
            _IX_PCI_READ_RX_ERROR_SIGNAL(pPCIController, &command);
            if(command != IX_PCI_RX_SIGNAL_OK)
                goto label_2;
            
            /* Free the sent buffer. */
            ix_rm_buffer_free(hBuffer);
            hBuffer = hChain;
            
            /* We unlink this buffer from the chain. */
            err = ix_rm_buffer_unlink(&hBuffer, &hChain);
            if(err != IX_SUCCESS)
                goto label_1;
        }

        /* Send the last buffer in the chain. */
        /* Now fill the control information structure. */
        pPCIControlInfo->m_BufferInfo.m_EOPFlag = 1;
        pPCIControlInfo->m_BufferInfo.m_BufferType = IX_RM_BUFFER_GET_TYPE(hBuffer);
        err = ix_rm_buffer_get_meta(hBuffer, &pData);
        if(err != IX_SUCCESS)
            goto label_1;

        /* retrieve the offset for hardware buffers */
        if(pPCIControlInfo->m_BufferInfo.m_BufferType == IX_BUFFER_TYPE_HARDWARE)
            offset = ((ix_hw_buffer_meta*)pData)->m_Offset;

        /* Set physical offset to be the same as virtual address, in software testing. */
        #if !defined(_IX_RM_IMPL_HARDWARE_)
            pPCIControlInfo->m_BufferInfo.m_MetaOffset = (ix_uint32)pData;
        #else

            /* Get physical offset and channel of the meta and write it into the control info. */
            err = ix_rm_get_phys_offset(pData, &memType,
                &channel, &pPCIControlInfo->m_BufferInfo.m_MetaOffset, &pPhysicalAddress);
            if(err != IX_SUCCESS)
                goto label_1;

            /**
             * We are getting the offset from the beginning of that 
             * type of memory, not from the beginning of the channel
             */

            /* We make the assumption that memory channels are contiguous */
            for(i = 0; i < channel; i++)
            {
                ix_rm_mem_info(memType, i, &memoryInfo);
                pPCIControlInfo->m_BufferInfo.m_MetaOffset += memoryInfo.m_TotalSize;
            }

        #endif /* end !defined(_IX_RM_IMPL_HARDWARE_) */

        err = ix_rm_buffer_get_data(hBuffer, &pData);
        if(err != IX_SUCCESS)
            goto label_1;

        /* adjust the offset for hardware buffers */
        if(pPCIControlInfo->m_BufferInfo.m_BufferType == IX_BUFFER_TYPE_HARDWARE)
            pData = (ix_uint8*)pData + offset;

        /* Set physical offset to be the same as virtual address, in software testing. */
        #if !defined(_IX_RM_IMPL_HARDWARE_)
        
            pPCIControlInfo->m_BufferInfo.m_DataOffset = (ix_uint32)pData;

        #else
            /* Get physical offset and channel of the data and write it into the control info. */
            err = ix_rm_get_phys_offset(pData, &memType,
                &channel, &pPCIControlInfo->m_BufferInfo.m_DataOffset, &pPhysicalAddress);
            if(err != IX_SUCCESS)
                goto label_1;

            /**
             * We are getting the offset from the beginning of that 
             * type of memory, not from the beginning of the channel
             */

            /* We make the assumption that memory channels are contiguous */
            for(i = 0; i < channel; i++)
            {
                ix_rm_mem_info(memType, i, &memoryInfo);
                pPCIControlInfo->m_BufferInfo.m_DataOffset += memoryInfo.m_TotalSize;
            }

        #endif /* end !defined(_IX_RM_IMPL_HARDWARE_) */

        /**
         * Get the meta size, meta memory type, data size,
         * and data memory type from the buffer's free list.
         */
        pFreeList = IX_RM_BUF_GET_FL_ADDRESS(hBuffer);
        pPCIControlInfo->m_BufferInfo.m_MetaSize = 1 << pFreeList->m_MetaElementShift;
        pPCIControlInfo->m_BufferInfo.m_MetaMemoryType = pFreeList->m_MetaMemoryType;
        pPCIControlInfo->m_BufferInfo.m_DataSize = 1 << pFreeList->m_DataElementShift;
        pPCIControlInfo->m_BufferInfo.m_DataMemoryType = pFreeList->m_DataMemoryType;

        /* Write to the mailbox register on the Rx side, and signal the Rx PCI interrupt */
        _ix_rm_pci_signal_rx_subsystem(IX_PCI_TX_SIGNAL_OK, (void*)pPCIController);


        /* Wait for completion of transfer. */
        err = ix_ossl_sem_take(pPCIController->m_TxSemaphore, IX_PCI_TX_TIMEOUT);
        /* Check for timeout. */
        if(err != IX_SUCCESS)
            goto label_1;


        /* Check for errors from the Rx side. */
        _IX_PCI_READ_RX_ERROR_SIGNAL(pPCIController, &command);
        if(command != IX_PCI_RX_SIGNAL_OK)
            goto label_2;

        /* Free the sent buffer. */
        ix_rm_buffer_free(hBuffer);
        hBuffer = IX_NULL_BUFFER_HANDLE;
        hChain = IX_NULL_BUFFER_HANDLE;

        /* Everything went fine, continue the loop. */
        continue;

label_1:

        /* Signal Rx that an error occurred. */
        _ix_rm_pci_signal_rx_subsystem(IX_PCI_TX_SIGNAL_ERROR, (void*)pPCIController);

label_2:

        /* Free the remaining buffer chain. */
        if(hBuffer != IX_NULL_BUFFER_HANDLE)
            ix_rm_buffer_free_chain(hBuffer);

        if(hChain != IX_NULL_BUFFER_HANDLE)
            ix_rm_buffer_free_chain(hChain);


#endif /* defined(IX_PCI_TEST_INTERRUPTS) */
    }

    pPCIController->m_TxTerminateFlag = 0;

    return err;
}


/**
 * NAME: _ix_rm_pci_rx_engine_run
 *
 * DESCRIPTION: This function contains the Rx thread functionality.
 *              It waits on a semaphore until Tx on the other side
 *              signals that it has a buffer ready for transfer.
 *              Once it wakes up, it reads the buffer control info
 *              over the PCI, and uses that info to transfer the
 *              buffer.  After receiving an entire buffer chain,
 *              it sends the buffer chain to the destination commID.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Rx thread.
 *
 * @Return: IX_SUCCESS, or an appropriate error code for failure.
 *
 */
ix_error _ix_rm_pci_rx_engine_run(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_rx_engine_run";

    ix_error err = IX_SUCCESS;

    ix_pci_controller* pPCIController = 
        (ix_pci_controller*)pContext;

    ix_uint32 command;

    ix_pci_control_info controlInfo;
    ix_buffer_handle hChain = IX_NULL_BUFFER_HANDLE;
    ix_buffer_handle hBuffer = IX_NULL_BUFFER_HANDLE;
    ix_buffer_handle hCurrBuffer = IX_NULL_BUFFER_HANDLE;
    ix_communication_id commId = 0;
    ix_uint32 userData = 0;
    ix_pci_buffer_data_type pciBufferDataType = 0;

    void* pData;
    void* pMeta;

    ix_uint32 dataSize, actualDataSize;
    ix_uint32 remainingDataSize;
    ix_buffer_type bufferType = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    while(pPCIController->m_RxTerminateFlag == 0)
    {
        /* wait from a signal from the PCI ISR that data is available */
        ix_ossl_sem_take(pPCIController->m_RxSemaphore, IX_OSSL_WAIT_FOREVER);
#if defined(IX_PCI_TEST_INTERRUPTS)

        /*_ix_rm_pci_signal_tx_subsystem(0, pPCIController);*/
        ix_ossl_sem_give(pPCIController->m_TxSemaphore);

#else

        /* Check for errors from the Tx side. */
        _IX_PCI_READ_TX_ERROR_SIGNAL(pPCIController, &command);
        if(command != IX_PCI_TX_SIGNAL_OK)
            goto label_2;

        /**
         * Get the control info and build a local buffer chain and retrieve
         * the peer buffers.
         */
        _ix_rm_pci_get_tx_control_info((void*)pPCIController, &controlInfo);

        /* We are at the start of a new chain. */
        if(hChain == IX_NULL_BUFFER_HANDLE)
        {
            commId = controlInfo.m_DestinationCommId;
            userData = controlInfo.m_UserData;
            pciBufferDataType = controlInfo.m_PCIBufferDataType;
            
            /**
             * If the first buffer in the chain is of a different
             * type than the following buffers in the chain,
             * then declare an error.
             */
            bufferType = controlInfo.m_BufferInfo.m_BufferType;
        }
        else
        {
            if (bufferType != controlInfo.m_BufferInfo.m_BufferType)
                goto label_1;
        }

        /* Check the meta size, and allocate the buffer from the appropriate free list. */
        if(bufferType == IX_BUFFER_TYPE_HARDWARE)
        {
           if(controlInfo.m_BufferInfo.m_MetaSize > IX_PCI_MAX_HW_META_SIZE)
               goto label_1;
           dataSize = IX_PCI_MAX_HW_DATA_SIZE;
           if(pPCIController->m_hUserHwFreeList)
                err = ix_rm_buffer_alloc(pPCIController->m_hUserHwFreeList, &hBuffer);
           else
                err = ix_rm_buffer_alloc(pPCIController->m_hPCIHwFreeList, &hBuffer);
           if(err != IX_SUCCESS)
               goto label_1;
        }
        else
        {
           if(controlInfo.m_BufferInfo.m_MetaSize > IX_PCI_MAX_SW_META_SIZE)
               goto label_1;
           dataSize = IX_PCI_MAX_SW_DATA_SIZE;
           err = ix_rm_buffer_alloc(pPCIController->m_hPCISwFreeList, &hBuffer);
           if(err != IX_SUCCESS)
               goto label_1;
        }

        /* Get the meta address of the newly allocated buffer */
        err = ix_rm_buffer_get_meta(hBuffer, &pMeta);
        if(err != IX_SUCCESS)
            goto label_1;

        /* Copy the meta from the remote subsystem */
        err = _ix_rm_pci_copy_meta_from_remote_subsystem(&controlInfo.m_BufferInfo,
            (void*)pPCIController, pMeta, &remainingDataSize);
        if(err != IX_SUCCESS)
            goto label_1;

        /**
         * If the local data size is greater than or equal to the remote
         * data size, copy the remote data into the local data.
         * Note that in the case where the local data size is greater than
         * the remote data size, the local buffer will not be filled to its
         * full capacity.
         */

        /* Adjust the required size */
        if(bufferType == IX_BUFFER_TYPE_HARDWARE)
            actualDataSize = dataSize - ((ix_hw_buffer_meta*)pMeta)->m_Offset;
        else
            actualDataSize = dataSize;

        if(actualDataSize >= remainingDataSize)
        {
            /* Get the data address of the newly allocated buffer */
            err = ix_rm_buffer_get_data(hBuffer, &pData);
            if(err != IX_SUCCESS)
                goto label_1;

            /* Make sure we are copying at the right offset */
            if(bufferType == IX_BUFFER_TYPE_HARDWARE)
                pData += ((ix_hw_buffer_meta*)pMeta)->m_Offset;

            /* Copy the data from the remote subsystem */
            err = _ix_rm_pci_copy_data_from_remote_subsystem(
                &controlInfo.m_BufferInfo, remainingDataSize,
                (void*)pPCIController, pData);
            if(err != IX_SUCCESS)
                goto label_1;
        }
        /**
         * If the local data size is less than the remote
         * data size, we must create a chain of local buffers
         * to contain the remote data.
         */
        else
        {
            while(remainingDataSize > actualDataSize)
            {
                /* Set the buffer size in the meta-data. */
                if(bufferType == IX_BUFFER_TYPE_HARDWARE)
                    ((ix_hw_buffer_meta*)pMeta)->m_BufferSize = (ix_uint16)actualDataSize;
                else
                    ((ix_sw_buffer_meta*)pMeta)->m_DataSize = actualDataSize;

                /* Get the data address of the newly allocated buffer */
                err = ix_rm_buffer_get_data(hBuffer, &pData);
                if(err != IX_SUCCESS)
                    goto label_1;

                /* Make sure we are copying at the right offset */
                if(bufferType == IX_BUFFER_TYPE_HARDWARE)
                    pData += dataSize - actualDataSize;

                /* Copy the data from the remote subsystem */
                err = _ix_rm_pci_copy_data_from_remote_subsystem(
                       &controlInfo.m_BufferInfo, actualDataSize,
                       (void*)pPCIController, pData);
                if(err != IX_SUCCESS)
                    goto label_1;

                if(hChain == IX_NULL_BUFFER_HANDLE)
                {
                    /* Start the buffer chain. */
                    hCurrBuffer = hBuffer;
                    hChain = hCurrBuffer;
                }
                else
                {
                    /* Link the buffer with the chain */
                    err = ix_rm_buffer_link(&hCurrBuffer, &hBuffer);
                    if(err != IX_SUCCESS)
                        goto label_1;
                    hCurrBuffer = hBuffer;
                    /**
                     * Set the EOP flag that the first buffer in
                     * this chain is not the EOP.
                     */
                    IX_RM_BUF_SET_EOP(hChain, 0);
                }

                /* Allocate the next buffer from the appropriate free list. */
                if(bufferType == IX_BUFFER_TYPE_HARDWARE)
                {
                    if(pPCIController->m_hUserHwFreeList)
                        err = ix_rm_buffer_alloc(pPCIController->m_hUserHwFreeList, &hBuffer);
                   else
                        err = ix_rm_buffer_alloc(pPCIController->m_hPCIHwFreeList, &hBuffer);
                    if(err != IX_SUCCESS)
                        goto label_1;
                    
                    /* Get the meta address of the newly allocated buffer */
                    err = ix_rm_buffer_get_meta(hBuffer, &pMeta);
                    if(err != IX_SUCCESS)
                        goto label_1;

                    /* set the next buffer offset to 0 */
                    ((ix_hw_buffer_meta*)pMeta)->m_Offset = 0;
                }
                else
                {
                    err = ix_rm_buffer_alloc(pPCIController->m_hPCISwFreeList, &hBuffer);
                    if(err != IX_SUCCESS)
                        goto label_1;
                    /* Get the meta address of the newly allocated buffer */
                    err = ix_rm_buffer_get_meta(hBuffer, &pMeta);
                    if(err != IX_SUCCESS)
                        goto label_1;
                }

                /* Increment data pointer in control info. */
                (ix_uint8*)controlInfo.m_BufferInfo.m_DataOffset += actualDataSize;

                /* Update the size of the rest of the packet. */
                remainingDataSize -= actualDataSize;

                /* update the actual data size */
                actualDataSize = dataSize;
            } /* end while(remainingDataSize > dataSize) */

            /* Handle the last segment of this buffer. */

            /* Set the buffer size in the meta-data. */
            if(bufferType == IX_BUFFER_TYPE_HARDWARE)
                ((ix_hw_buffer_meta*)pMeta)->m_BufferSize = (ix_uint16)remainingDataSize;
            else
                ((ix_sw_buffer_meta*)pMeta)->m_DataSize = remainingDataSize;

            /* Get the data address of the newly allocated buffer */
            err = ix_rm_buffer_get_data(hBuffer, &pData);
            if(err != IX_SUCCESS)
                goto label_1;

            /* Make sure we are copying at the right offset */
            if(bufferType == IX_BUFFER_TYPE_HARDWARE)
                pData += dataSize - actualDataSize;

            /* Copy the data from the remote subsystem */
            err = _ix_rm_pci_copy_data_from_remote_subsystem(
                   &controlInfo.m_BufferInfo, remainingDataSize,
                   (void*)pPCIController, pData);
            if(err != IX_SUCCESS)
                goto label_1;

        }

        if(hChain == IX_NULL_BUFFER_HANDLE)
        {
            /* Start the buffer chain. */
            hCurrBuffer = hBuffer;
            hChain = hCurrBuffer;
        }
        else
        {
            /* Link the buffer with the chain */
            err = ix_rm_buffer_link(&hCurrBuffer, &hBuffer);
            if(err != IX_SUCCESS)
                goto label_1;
            hCurrBuffer = hBuffer;
            /**
             * Set the EOP flag that the first buffer in
             * this chain is not the EOP.
             */
            IX_RM_BUF_SET_EOP(hChain, 0);
        }

        /* Signal the Tx side that we are done with this buffer */
        _ix_rm_pci_signal_tx_subsystem(IX_PCI_RX_SIGNAL_OK, pPCIController);

        if(controlInfo.m_BufferInfo.m_EOPFlag != 0)
        {
            /* We've got all the chain */

            /* Send the chain to the destination */
            if(pciBufferDataType == IX_PCI_BUFFER_DATA_TYPE_PACKET)
                err = ix_rm_packet_send(commId, hChain, userData);
            else
                err = ix_rm_message_send(commId, hChain, userData);
            if(err != IX_SUCCESS)
            {
                /* Free the buffer chain. */
                ix_rm_buffer_free_chain(hChain);
            }
            hChain = IX_NULL_BUFFER_HANDLE;
            hBuffer = IX_NULL_BUFFER_HANDLE;
            hCurrBuffer = IX_NULL_BUFFER_HANDLE;
        }

        /* Wait for the next buffer. */
        continue;

label_1:
        /**
         * Notify Tx that an error has occurred on the Rx side.  Tx should
         * drop the chain that it is currently sending.
         */
        _ix_rm_pci_signal_tx_subsystem(IX_PCI_RX_SIGNAL_ERROR, pPCIController);


label_2:

        if(hChain != IX_NULL_BUFFER_HANDLE)
        {
            ix_rm_buffer_free_chain(hChain);
            hChain = IX_NULL_BUFFER_HANDLE;
        }

        if(hBuffer != IX_NULL_BUFFER_HANDLE)
        {
            ix_rm_buffer_free(hBuffer);
            hBuffer = IX_NULL_BUFFER_HANDLE;
        }

        hCurrBuffer = IX_NULL_BUFFER_HANDLE;

#endif /* defined(IX_PCI_TEST_INTERRUPTS) */

    }



    pPCIController->m_RxTerminateFlag = 0;

    /* ix_ossl_message_log("Rx terminated 0x%x\n", err); */

    return err;
}


/**
 * NAME: _ix_rm_pci_tx_engine_notify
 *
 * DESCRIPTION: Notifies the Tx thread that Rx on the other side
 *              has finished processing a buffer.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Tx thread.
 *
 * @Return: IX_SUCCESS, or an appropriate error code for failure.
 *
 */
ix_error _ix_rm_pci_tx_engine_notify(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_tx_engine_notify";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* signal the sync object contained in the context */
    return ix_ossl_sem_give(((ix_pci_controller*)pContext)->m_TxSemaphore);
}

/**
 * NAME: _ix_rm_pci_rx_engine_notify
 *
 * DESCRIPTION: Notifies the Rx thread that Tx on the other side
 *              has a buffer ready for transfer.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Rx thread.
 *
 * @Return: IX_SUCCESS, or an appropriate error code for failure.
 *
 */
ix_error _ix_rm_pci_rx_engine_notify(void* pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_rx_engine_notify";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* signal the sync object contained in the context */
    return ix_ossl_sem_give(((ix_pci_controller*)pContext)->m_RxSemaphore);
}

/**
 * NAME: _ix_rm_pci_buffer_retrieve
 *
 * DESCRIPTION: Retrieves a buffer to be transferred over the PCI.  When
 *              a client wants to shut down the PCI communication, it
 *              should send a NULL packet, something that should make
 *              this function return an error so that the Tx thread will
 *              wake up and exit.
 * @Param:  - IN ix_pci_buffer_data_type* arg_pBufferType - indicates
 *          whether a packet or a message is being transferred.
 * @Param:  - IN ix_buffer_handle* arg_pBufferHandle - pointer to the handle
 *          of the buffer to be transferred.
 * @Param:  - IN ix_uint32* arg_pUserData - pointer to user data.
 * @Param:  - IN ix_communication_id* arg_pCommId - pointer to destination
 *          communication ID.
 *
 * @Return: IX_SUCCESS, or an appropriate error code for failure.
 *
 */
ix_error _ix_rm_pci_buffer_retrieve(ix_pci_buffer_data_type* arg_pBufferType,
                                               ix_buffer_handle* arg_pBufferHandle,
                                               ix_uint32* arg_pUserData,
                                               ix_communication_id* arg_pCommId)
{

    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_buffer_retrieve";

    ix_error err = IX_SUCCESS;
    ix_uint32 isEmptySet = 1;

    ix_end_point_cc_data* pPacketData;
    ix_end_point_cc_data* pMessageData;
    ix_end_point_data_queue* pPacketQueue;
    ix_end_point_data_queue* pMessageQueue;

    ix_ossl_sem_t emptySem = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* get the PCI data queues */
    pPacketData = &g_RemoteCommManager.m_RemoteCommPacketEndPoint.m_EPData.m_CoreCompData;
    pPacketQueue = &pPacketData->m_EPDataQueue;
    pMessageData = &g_RemoteCommManager.m_RemoteCommMessageEndPoint.m_EPData.m_CoreCompData;
    pMessageQueue = &pMessageData->m_EPDataQueue;

    /* check if data is already available */
    isEmptySet &= IX_EP_DATA_QUEUE_IS_EMPTY(pPacketQueue);
    isEmptySet &= IX_EP_DATA_QUEUE_IS_EMPTY(pMessageQueue);

    if(isEmptySet)
    {
        /* get a new semaphore */
        err =  _ix_rm_sem_pool_get_sem(g_pCommSemaphorePool, &emptySem);
        if(err !=IX_SUCCESS)
            return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_SEM_GET_FAILED, s_aFunctionIdentifier, 
                "failed to get a blocking empty semaphore!", 0);


        /* add the empty semaphore to the receive notification list of PCI message data */

        /* lock the access to the notification list */
        _IX_RM_EP_DATA_ACTION_LOCK(pMessageData);

        /* add the empty semaphore to the receive notification list */
        _IX_RM_EP_DATA_WAIT_QUEUE_ADD(pMessageData->m_pReceiveNotificationQueue, emptySem);

        if(err != IX_SUCCESS)
        {
            /* unlock the access to the notification list */
            _IX_RM_EP_DATA_ACTION_UNLOCK(pMessageData);

            if(emptySem != 0)
                _ix_rm_sem_pool_release_sem(g_pCommSemaphorePool, emptySem);
            return err;
        }

        /* unlock the access to the notification list */
        _IX_RM_EP_DATA_ACTION_UNLOCK(pMessageData);

        isEmptySet &= IX_EP_DATA_QUEUE_IS_EMPTY(pMessageQueue);

        /* add the empty semaphore to the receive notification list of PCI packet data */

        /* lock the access to the notification list */
        _IX_RM_EP_DATA_ACTION_LOCK(pPacketData);

        /* add the empty semaphore to the receive notification list */
        _IX_RM_EP_DATA_WAIT_QUEUE_ADD(pPacketData->m_pReceiveNotificationQueue, emptySem);
        if(err != IX_SUCCESS)
        {
            /* unlock the access to the notification list */
            _IX_RM_EP_DATA_ACTION_UNLOCK(pPacketData);

            _IX_RM_EP_DATA_ACTION_LOCK(pMessageData);
            _IX_RM_EP_DATA_WAIT_QUEUE_REMOVE(pMessageData->m_pReceiveNotificationQueue, emptySem);
            _IX_RM_EP_DATA_ACTION_UNLOCK(pMessageData);

            if(emptySem != 0)
                _ix_rm_sem_pool_release_sem(g_pCommSemaphorePool, emptySem);
            return err;
        }

        /* unlock the access to the notification list */
        _IX_RM_EP_DATA_ACTION_UNLOCK(pPacketData);

        isEmptySet &= IX_EP_DATA_QUEUE_IS_EMPTY(pPacketQueue);


        if(isEmptySet)
        {

label_1:            
            /* wait on the empty semaphore for the queue to become available */
            ix_ossl_sem_take(emptySem, IX_OSSL_WAIT_FOREVER);

            isEmptySet &= IX_EP_DATA_QUEUE_IS_EMPTY(pMessageQueue);

            isEmptySet &= IX_EP_DATA_QUEUE_IS_EMPTY(pPacketQueue);

            if(isEmptySet)
                goto label_1;
        }

        /* remove the empty semaphore from the receive notification list of PCI message data */

        /* lock the access to the notification list */
        _IX_RM_EP_DATA_ACTION_LOCK(pMessageData);

        /* remove the empty semaphore to the receive notification list */
        _IX_RM_EP_DATA_WAIT_QUEUE_REMOVE(pMessageData->m_pReceiveNotificationQueue, emptySem);

        /* unlock the access to the notification list */
        _IX_RM_EP_DATA_ACTION_UNLOCK(pMessageData);

        /* remove the empty semaphore from the receive notification list of PCI packet data */

        /* lock the access to the notification list */
        _IX_RM_EP_DATA_ACTION_LOCK(pPacketData);

        /* remove the empty semaphore to the receive notification list */
        _IX_RM_EP_DATA_WAIT_QUEUE_REMOVE(pPacketData->m_pReceiveNotificationQueue, emptySem);

        /* unlock the access to the notification list */
        _IX_RM_EP_DATA_ACTION_UNLOCK(pPacketData);
    }


    /* We have got some data. Retrieve it and pass it to the caller. */
    {
        ix_communication_id commId;

        /* We try for messages first which are supposed to be higher priority */
        if(!IX_EP_DATA_QUEUE_IS_EMPTY(pMessageQueue))
        {
            *arg_pBufferType = IX_PCI_BUFFER_DATA_TYPE_MESSAGE;
            _IX_EP_DATA_QUEUE_PCI_GET_COMM_ELEMENT_SC(pMessageQueue, 
                                                   arg_pBufferHandle, arg_pUserData, 
                                                   &commId, pMessageData);
        }
        if(!IX_EP_DATA_QUEUE_IS_EMPTY(pPacketQueue))
        {
            *arg_pBufferType = IX_PCI_BUFFER_DATA_TYPE_PACKET;
            _IX_EP_DATA_QUEUE_PCI_GET_COMM_ELEMENT_SC(pPacketQueue, 
                                                   arg_pBufferHandle, arg_pUserData, 
                                                   &commId, pPacketData);
        }

        *arg_pCommId = IX_RM_COMM_ID_GET_LOCAL_ID(commId);
    }

    if(emptySem != 0)
        _ix_rm_sem_pool_release_sem(g_pCommSemaphorePool, emptySem);

    return err;
}


/**
 * NAME: _ix_rm_pci_signal_rx_subsystem
 *
 * DESCRIPTION: This function, called from the Tx thread, signals
 *              the Rx thread on the other side that a buffer is ready
 *              for transfer, or that an error occured on Tx.
 *
 * @Param:  - IN ix_uint32 arg_cmd - indicates to Rx whether an error
 *          has occurred on Tx or not.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Tx thread.
 *
 * @Return: IX_SUCCESS, or an appropriate error code for failure.
 *
 */

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

ix_error _ix_rm_pci_signal_rx_subsystem(ix_uint32 arg_cmd, void* arg_pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_signal_rx_subsystem";

    ix_error err = IX_SUCCESS;

    ix_pci_controller* pPCIController = (ix_pci_controller*)arg_pContext;

    ix_uint32 remoteCSRBase;

    ix_uint32 temp;

    ix_uint32 extRegVal;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);
    
    /* Get remote CSR base address from controller. */
    remoteCSRBase = pPCIController->m_RemoteCSRBase;

    ix_ossl_sem_take(pPCIController->m_MBRegsSemaphore, IX_OSSL_WAIT_FOREVER);

#if defined(_IX_RM_IMPL_HARDWARE_)

    /* Disable all PCI interrupts. */
    intDisable(INT_VEC_PIL);

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    _IX_PCI_REG_READ(IX_PCI_ADDR_EXT, extRegVal);

    /* Get the address for desired PCI memory offset as seen from the CPU. */
    _IX_PCI_GET_CPU_ADDRESS(remoteCSRBase);

    /**
     * Write to remote mailbox register 1 to indicate that Tx is signalling Rx,
     * and whether an error occured on Tx or not.
     */

    do
    {
        _IX_PCI_REG_READ(remoteCSRBase + IX_PCI_REMOTE_MAILBOX_1, temp);
    }
    while(temp != 0);
    temp = IX_PCI_SIGNAL_RX_FROM_TX | arg_cmd;
    _IX_PCI_REG_WRITE(remoteCSRBase + IX_PCI_REMOTE_MAILBOX_1, IX_PCI_32BIT_SWAP(temp));


    #if defined(IX_PCI_A0_SILICON)
        /* A read cycle to work around the A0 bug that requires a read cycle between writes. */
        _IX_PCI_REG_READ(remoteCSRBase + IX_PCI_REMOTE_MAILBOX_1, temp);
    #endif /* end defined(IX_PCI_A0_SILICON) */


#if !defined(IX_PCI_TEST_INTERRUPTS)
    /* Write the offset in DRAM of the control data to remote mailbox register 2. */
    temp = pPCIController->m_TxPCIControlInfoOffset;
    _IX_PCI_REG_WRITE(remoteCSRBase + IX_PCI_REMOTE_MAILBOX_2, IX_PCI_32BIT_SWAP(temp));

    #if defined(IX_PCI_A0_SILICON)
        /* A read cycle to work around the A0 bug that requires a read cycle between writes. */
        _IX_PCI_REG_READ(remoteCSRBase + IX_PCI_REMOTE_MAILBOX_2, temp);
    #endif /* end defined(IX_PCI_A0_SILICON) */

#endif /* !defined(IX_PCI_TEST_INTERRUPTS) */

    /* Set the doorbell interrupt on the other side. */
    temp = 0x00000001;
    _IX_PCI_REG_WRITE(remoteCSRBase + IX_PCI_REMOTE_DOORBELL, IX_PCI_32BIT_SWAP(temp));

    /* Reset the upper word used in PCI addresses back to its original value. */
    _IX_PCI_REG_WRITE(IX_PCI_ADDR_EXT, extRegVal);

#if defined(_IX_RM_IMPL_HARDWARE_)

    /* Re-enable all PCI interrupts. */
    intEnable(INT_VEC_PIL);

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    ix_ossl_sem_give(pPCIController->m_MBRegsSemaphore);

    return err;
}

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

ix_error _ix_rm_pci_signal_rx_subsystem(ix_uint32 arg_cmd, void* arg_pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_signal_rx_subsystem";

    ix_error err = IX_SUCCESS;

    ix_pci_controller* pPCIController = (ix_pci_controller*)arg_pContext;

    ix_uint32 remoteCSRBase;

    ix_uint32 temp;

    ix_uint32 extRegVal;

    unsigned long flags;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);
    
    /* Get remote CSR base address from controller. */
    remoteCSRBase = pPCIController->m_RemoteCSRBase;

    ix_ossl_sem_take(pPCIController->m_MBRegsSemaphore, IX_OSSL_WAIT_FOREVER);

    /* Disable all interrupts. */
    local_irq_save(flags);

    _IX_PCI_REG_READ(IX_PCI_ADDR_EXT, extRegVal);

    /* Get the address for desired PCI memory offset as seen from the CPU. */
    _IX_PCI_GET_CPU_ADDRESS(remoteCSRBase);

    /**
     * Write to remote mailbox register 1 to indicate that Tx is signalling Rx,
     * and whether an error occured on Tx or not.
     */

    do
    {
        _IX_PCI_REG_READ(remoteCSRBase + IX_PCI_REMOTE_MAILBOX_1, temp);
    }
    while(temp != 0);
    temp = IX_PCI_SIGNAL_RX_FROM_TX | arg_cmd;
    _IX_PCI_REG_WRITE(remoteCSRBase + IX_PCI_REMOTE_MAILBOX_1, IX_PCI_32BIT_SWAP(temp));


    #if defined(IX_PCI_A0_SILICON)
        /* A read cycle to work around the A0 bug that requires a read cycle between writes. */
        _IX_PCI_REG_READ(remoteCSRBase + IX_PCI_REMOTE_MAILBOX_1, temp);
    #endif /* end defined(IX_PCI_A0_SILICON) */


#if !defined(IX_PCI_TEST_INTERRUPTS)
    /* Write the offset in DRAM of the control data to remote mailbox register 2. */
    temp = pPCIController->m_TxPCIControlInfoOffset;
    _IX_PCI_REG_WRITE(remoteCSRBase + IX_PCI_REMOTE_MAILBOX_2, IX_PCI_32BIT_SWAP(temp));

    #if defined(IX_PCI_A0_SILICON)
        /* A read cycle to work around the A0 bug that requires a read cycle between writes. */
        _IX_PCI_REG_READ(remoteCSRBase + IX_PCI_REMOTE_MAILBOX_2, temp);
    #endif /* end defined(IX_PCI_A0_SILICON) */

#endif /* !defined(IX_PCI_TEST_INTERRUPTS) */

    /* Set the doorbell interrupt on the other side. */
    temp = 0x00000001;
    _IX_PCI_REG_WRITE(remoteCSRBase + IX_PCI_REMOTE_DOORBELL, IX_PCI_32BIT_SWAP(temp));

    /* Reset the upper word used in PCI addresses back to its original value. */
    _IX_PCI_REG_WRITE(IX_PCI_ADDR_EXT, extRegVal);

    /* Re-enable all interrupts. */
    local_irq_restore(flags);

    ix_ossl_sem_give(pPCIController->m_MBRegsSemaphore);

    return err;
}

#else

#error "PCI support is not enabled on this OS!"

#endif /* _IX_OS_TYPE_ */


/**
 * NAME: _ix_rm_pci_get_tx_control_info
 *
 * DESCRIPTION: This function, called from the Rx thread, gets
 *              the control information for the buffer to be
 *              transferred over PCI.
 *
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Rx thread.
 * @Param:  - OUT ix_pci_control_info* arg_pControlInfo - pointer to
 *          where control info will be stored.
 *
 * @Return: IX_SUCCESS, or an appropriate error code for failure.
 *
 */

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

ix_error _ix_rm_pci_get_tx_control_info(void* arg_pContext, ix_pci_control_info* arg_pControlInfo)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_get_tx_control_info";

    ix_uint32 remoteDRAMBase;

    static ix_uint32 remoteDataAddress = 0;

    ix_uint32 remoteVirtualDataAddress;

    ix_uint32 extRegVal;

#if defined(IX_PCI_A0_SILICON)
    ix_uint32* pSource;
    ix_uint32* pDestination;
    ix_uint32 i;
    ix_uint32 temp;
#endif /*#if defined(IX_PCI_A0_SILICON) */

    ix_pci_controller* pPCIController = (ix_pci_controller*)arg_pContext;


    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(remoteDataAddress == 0)
    {
        /* Get remote DRAM base address from controller. */
        remoteDRAMBase = pPCIController->m_RemoteDRAMBase;

        /* Read address of remote control info from mailbox register 2. */
        _IX_PCI_REG_READ(IX_PCI_LOCAL_MAILBOX_2, remoteDataAddress);

        /*remoteDataAddress = IX_PCI_32BIT_SWAP(remoteDataAddress);*/

        remoteDataAddress += remoteDRAMBase;
    }

    remoteVirtualDataAddress = remoteDataAddress;

#if defined(_IX_RM_IMPL_HARDWARE_)

    /* Disable all PCI interrupts. */
    intDisable(INT_VEC_PIL); 

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    /* Get the address for desired PCI memory offset as seen from the CPU. */
    _IX_PCI_REG_READ(IX_PCI_ADDR_EXT, extRegVal);

    _IX_PCI_GET_CPU_ADDRESS(remoteVirtualDataAddress);

    /**
     * Copy the remote bytes to the local destination.
     */
    #if defined(IX_PCI_A0_SILICON)
        for(i = 0, pSource = (ix_uint32*)remoteVirtualDataAddress,
            pDestination = (ix_uint32*)arg_pControlInfo
            ; i < ((sizeof(ix_pci_control_info) + 3) >> 2)
            ;i++, pSource++, pDestination++)
        {
            *pDestination = *pSource;
            temp = *pSource;
        }
    #else
        /* ix_ossl_memcpy((void*)arg_pControlInfo, (void*)remoteDataAddress,
            sizeof(ix_pci_control_info)); */
        /* *arg_pControlInfo = *((ix_pci_control_info*)remoteVirtualDataAddress); */
        ix_rm_managed_to_managed_memory_copy(arg_pControlInfo,
                                            (void*)remoteVirtualDataAddress,
                                            sizeof(ix_pci_control_info) 
                                          );
    #endif /*#if defined(IX_PCI_A0_SILICON) */
    

    /* Reset the upper word in PCI addresses back to its original value. */
    _IX_PCI_REG_WRITE(IX_PCI_ADDR_EXT, extRegVal);

#if defined(_IX_RM_IMPL_HARDWARE_)

    /* Re-enable all PCI interrupts. */
    intEnable(INT_VEC_PIL); 

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    return IX_SUCCESS;
}

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

ix_error _ix_rm_pci_get_tx_control_info(void* arg_pContext, ix_pci_control_info* arg_pControlInfo)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_get_tx_control_info";

    ix_uint32 remoteDRAMBase;

    static ix_uint32 remoteDataAddress = 0;

    ix_uint32 remoteVirtualDataAddress;

    ix_uint32 extRegVal;

    ix_pci_controller* pPCIController = (ix_pci_controller*)arg_pContext;

    unsigned long flags;


    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(remoteDataAddress == 0)
    {
        /* Get remote DRAM base address from controller. */
        remoteDRAMBase = pPCIController->m_RemoteDRAMBase;

        /* Read address of remote control info from mailbox register 2. */
        _IX_PCI_REG_READ(IX_PCI_LOCAL_MAILBOX_2, remoteDataAddress);

        remoteDataAddress += remoteDRAMBase;
    }

    remoteVirtualDataAddress = remoteDataAddress;

    /* Disable all interrupts. */
    local_irq_save(flags);

    /* Get the address for desired PCI memory offset as seen from the CPU. */
    _IX_PCI_REG_READ(IX_PCI_ADDR_EXT, extRegVal);

    _IX_PCI_GET_CPU_ADDRESS(remoteVirtualDataAddress);

    /**
     * Copy the remote bytes to the local destination.
     */
    /* *arg_pControlInfo = *((ix_pci_control_info*)remoteVirtualDataAddress); */
    ix_rm_managed_to_managed_memory_copy(arg_pControlInfo,
                                        (void*)remoteVirtualDataAddress,
                                        sizeof(ix_pci_control_info) 
                                      );
    

    /* Reset the upper word in PCI addresses back to its original value. */
    _IX_PCI_REG_WRITE(IX_PCI_ADDR_EXT, extRegVal);

    /* Re-enable all interrupts. */
    local_irq_restore(flags);

    /* ix_ossl_sleep(0); */

    return IX_SUCCESS;
}

#else

#error "PCI support is not enabled on this OS!"

#endif /* _IX_OS_TYPE_ */

/**
 * NAME: _ix_rm_pci_signal_tx_subsystem
 *
 * DESCRIPTION: This function, called from the Rx thread, signals
 *              the Tx thread on the other side that Rx has finished
 *              processing a buffer, or that an error occured on Rx.
 *
 * @Param:  - IN ix_uint32 arg_cmd - indicates to Tx whether an error
 *          has occurred on Rx or not.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Rx thread.
 *
 * @Return: IX_SUCCESS, or an appropriate error code for failure.
 *
 */

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

ix_error _ix_rm_pci_signal_tx_subsystem(ix_uint32 arg_cmd, void* arg_pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_signal_tx_subsystem";

    ix_uint32 remoteCSRBase;

    ix_uint32 temp;

    ix_uint32 extRegVal;

    ix_pci_controller* pPCIController = (ix_pci_controller*)arg_pContext;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* Get remote CSR base address from controller. */
    remoteCSRBase = pPCIController->m_RemoteCSRBase;

    ix_ossl_sem_take(pPCIController->m_MBRegsSemaphore, IX_OSSL_WAIT_FOREVER);

#if defined(_IX_RM_IMPL_HARDWARE_)

    /* Disable all PCI interrupts. */
    intDisable(INT_VEC_PIL); 

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    _IX_PCI_REG_READ(IX_PCI_ADDR_EXT, extRegVal);

    /* Get the address for desired PCI memory offset as seen from the CPU. */
    _IX_PCI_GET_CPU_ADDRESS(remoteCSRBase);

    /**
     * Write to remote mailbox register 1 to indicate that Rx is signalling Tx,
     * and whether Rx processed a buffer succesfully or not.
     */

    do
    {
        _IX_PCI_REG_READ(remoteCSRBase + IX_PCI_REMOTE_MAILBOX_1, temp);
    }
    while(temp != 0);
    temp = IX_PCI_SIGNAL_TX_FROM_RX | arg_cmd;
    _IX_PCI_REG_WRITE(remoteCSRBase + IX_PCI_REMOTE_MAILBOX_1, IX_PCI_32BIT_SWAP(temp));

    
    #if defined(IX_PCI_A0_SILICON)
        /* A read cycle to work around the A0 bug that requires a read cycle between writes. */
        _IX_PCI_REG_READ(remoteCSRBase + IX_PCI_REMOTE_MAILBOX_1, temp);
    #endif /* end defined(IX_PCI_A0_SILICON) */
    
    /* Set the doorbell interrupt on the other side. */
    temp = 0x00000001;
    _IX_PCI_REG_WRITE(remoteCSRBase + IX_PCI_REMOTE_DOORBELL, IX_PCI_32BIT_SWAP(temp));

    #if defined(IX_PCI_A0_SILICON)
        /* A read cycle to work around the A0 bug that requires a read cycle between writes. */
        _IX_PCI_REG_READ(remoteCSRBase + IX_PCI_REMOTE_MAILBOX_1, temp);
    #endif /* end defined(IX_PCI_A0_SILICON) */

    /* Reset the upper word used in PCI addresses back to its original value. */
    _IX_PCI_REG_WRITE(IX_PCI_ADDR_EXT, extRegVal);

#if defined(_IX_RM_IMPL_HARDWARE_)

    /* Re-enable all PCI interrupts. */
    intEnable(INT_VEC_PIL); 

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    ix_ossl_sem_give(pPCIController->m_MBRegsSemaphore);

    return IX_SUCCESS;
}

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

ix_error _ix_rm_pci_signal_tx_subsystem(ix_uint32 arg_cmd, void* arg_pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_signal_tx_subsystem";

    ix_uint32 remoteCSRBase;

    ix_uint32 temp;

    ix_uint32 extRegVal;

    ix_pci_controller* pPCIController = (ix_pci_controller*)arg_pContext;

    unsigned long flags;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* Get remote CSR base address from controller. */
    remoteCSRBase = pPCIController->m_RemoteCSRBase;

    ix_ossl_sem_take(pPCIController->m_MBRegsSemaphore, IX_OSSL_WAIT_FOREVER);

    /* Disable all interrupts. */
    local_irq_save(flags);

    _IX_PCI_REG_READ(IX_PCI_ADDR_EXT, extRegVal);

    /* Get the address for desired PCI memory offset as seen from the CPU. */
    _IX_PCI_GET_CPU_ADDRESS(remoteCSRBase);

    /**
     * Write to remote mailbox register 1 to indicate that Rx is signalling Tx,
     * and whether Rx processed a buffer succesfully or not.
     */


    do
    {
        _IX_PCI_REG_READ(remoteCSRBase + IX_PCI_REMOTE_MAILBOX_1, temp);
    }
    while(temp != 0);
    temp = IX_PCI_SIGNAL_TX_FROM_RX | arg_cmd;
    _IX_PCI_REG_WRITE(remoteCSRBase + IX_PCI_REMOTE_MAILBOX_1, IX_PCI_32BIT_SWAP(temp));

    /* Set the doorbell interrupt on the other side. */
    temp = 0x00000001;
    _IX_PCI_REG_WRITE(remoteCSRBase + IX_PCI_REMOTE_DOORBELL, IX_PCI_32BIT_SWAP(temp));

    /* Reset the upper word used in PCI addresses back to its original value. */
    _IX_PCI_REG_WRITE(IX_PCI_ADDR_EXT, extRegVal);

    /* Re-enable all interrupts. */
    local_irq_restore(flags);

    ix_ossl_sem_give(pPCIController->m_MBRegsSemaphore);

    return IX_SUCCESS;
}

#else

#error "PCI support is not enabled on this OS!"

#endif /* _IX_OS_TYPE_ */


/**
 * NAME: _ix_rm_pci_copy_meta_from_remote_subsystem
 *
 * DESCRIPTION: This function, called from the Rx thread, copies
 *              the buffer meta-data from the other side.
 *
 * @Param:  - IN ix_pci_buffer_info* arg_pBufferInfo - pointer to
 *          control information for the buffer to be transferred.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Rx thread.
 * @Param:  - OUT void* arg_pLocalDestination - address to where
 *          remote data will be written.
 * @Param:  - OUT ix_uint32* arg_pDataSize - address to where the size of the
 *          data portion of the remote buffer will be written.
 *
 * @Return: IX_SUCCESS, or an appropriate error code for failure.
 *
 */

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

ix_error _ix_rm_pci_copy_meta_from_remote_subsystem(
                                               ix_pci_buffer_info* arg_pBufferInfo,
                                               void* arg_pContext,
                                               void* arg_pLocalDestination,
                                               ix_uint32* arg_pDataSize
                                               )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_copy_meta_from_remote_subsystem";

    ix_error err = IX_SUCCESS;

    ix_uint32 remoteSRAMBase;
    ix_uint32 remoteDRAMBase;
    ix_uint32 remoteDataBase;

    ix_uint32 remoteDataAddress;
    ix_uint32 remoteDataSize;

    ix_uint32* pDestination = arg_pLocalDestination;

    ix_uint32 extRegVal;

    ix_uint32 origNext = 0;
    ix_uint16 orig_bufferInfo = 0;

#if defined(IX_PCI_A0_SILICON)
    ix_uint32* pSource;
    ix_uint32 i;
    ix_uint32 temp;
#endif /* end defined(IX_PCI_A0_SILICON) */

    ix_pci_controller* pPCIController = (ix_pci_controller*)arg_pContext;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* Get remote SRAM and DRAM base addresss from controller. */
    remoteSRAMBase = pPCIController->m_RemoteSRAMBase;
    remoteDRAMBase = pPCIController->m_RemoteDRAMBase;

    /* Get the base address, size, and offset of the remote meta-data. */
    if (arg_pBufferInfo->m_BufferType == IX_BUFFER_TYPE_HARDWARE)
    {
        ix_hw_buffer_meta* pHWMeta = (ix_hw_buffer_meta*)arg_pLocalDestination;
        remoteDataBase = remoteSRAMBase;

        /* For HW meta, save the old values. */

#if !defined(_IX_RM_SPLIT_META_DATA_)

        origNext = pHWMeta->m_HwNext;

#endif /* !defined(_IX_RM_SPLIT_META_DATA_) */

        orig_bufferInfo = pHWMeta->m_BufferInfo;
    }
    else
        remoteDataBase = remoteDRAMBase;

    #if defined(IX_PCI_A0_SILICON)
        /* size expressed in 32 bit words */
        remoteDataSize = (arg_pBufferInfo->m_MetaSize + 3) >> 2;
    #else
        remoteDataSize = arg_pBufferInfo->m_MetaSize;
    #endif /* defined(IX_PCI_A0_SILICON) */

    remoteDataAddress = arg_pBufferInfo->m_MetaOffset;

    remoteDataAddress += remoteDataBase;

#if defined(_IX_RM_IMPL_HARDWARE_)

    /* Disable all PCI interrupts. */
    intDisable(INT_VEC_PIL); 

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    /* Get the address for desired PCI memory offset as seen from the CPU. */
    _IX_PCI_REG_READ(IX_PCI_ADDR_EXT, extRegVal);
  
    _IX_PCI_GET_CPU_ADDRESS(remoteDataAddress);

    /**
     * Copy the remote bytes to the local destination.
     */
    if (arg_pBufferInfo->m_BufferType == IX_BUFFER_TYPE_SOFTWARE)
    {
        /**
         * For SW buffers, we only need to copy the m_DataSize
         * field of the meta-data.
         */
        ((ix_sw_buffer_meta*)pDestination)->m_DataSize = 
            ((ix_sw_buffer_meta*)remoteDataAddress)->m_DataSize;
    }
    else
    {
        #if defined(IX_PCI_A0_SILICON)
            for(i = 0, pSource = (ix_uint32*)remoteDataAddress
                ; i < remoteDataSize; i++, pSource++, pDestination++)
            {
                *pDestination = *pSource;
                /* A read cycle to work around the A0 bug. */
                temp = *pSource;
            }
        #else
            /* ix_ossl_memcpy((void*)pDestination, (void*)remoteDataAddress,
                remoteDataSize);*/
            ix_rm_managed_to_managed_memory_copy(pDestination,
                                                (void*)remoteDataAddress,
                                                remoteDataSize
                                              );
        #endif /*#if defined(IX_PCI_A0_SILICON) */
    }

    /* Reset the upper word in PCI addresses back to its original value. */
    _IX_PCI_REG_WRITE(IX_PCI_ADDR_EXT, extRegVal);

#if defined(_IX_RM_IMPL_HARDWARE_)

    /* Re-enable all PCI interrupts. */
    intEnable(INT_VEC_PIL); 

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    if (arg_pBufferInfo->m_BufferType == IX_BUFFER_TYPE_HARDWARE)
    {
        ix_hw_buffer_meta* pHWMeta = (ix_hw_buffer_meta*)arg_pLocalDestination;

        /* For HW meta, restore the old values. */

#if !defined(_IX_RM_SPLIT_META_DATA_)

        pHWMeta->m_HwNext = origNext;

#endif /* !defined(_IX_RM_SPLIT_META_DATA_) */

        IX_RM_HW_META_SET_FREE_LIST_ID(pHWMeta->m_BufferInfo, 
            IX_RM_HW_META_GET_FREE_LIST_ID(orig_bufferInfo));

        /* Return the data size for this buffer. */
        *arg_pDataSize = pHWMeta->m_BufferSize;
    }
    else
    {
        /* Return the data size for this buffer. */
        *arg_pDataSize = ((ix_sw_buffer_meta*)pDestination)->m_DataSize;
    }

    return err;
}

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

ix_error _ix_rm_pci_copy_meta_from_remote_subsystem(
                                               ix_pci_buffer_info* arg_pBufferInfo,
                                               void* arg_pContext,
                                               void* arg_pLocalDestination,
                                               ix_uint32* arg_pDataSize
                                               )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_copy_meta_from_remote_subsystem";

    ix_error err = IX_SUCCESS;

    ix_uint32 remoteSRAMBase;
    ix_uint32 remoteDRAMBase;
    ix_uint32 remoteDataBase;

    ix_uint32 remoteDataAddress;
    ix_uint32 remoteDataSize;

    ix_uint32* pDestination = arg_pLocalDestination;

    ix_uint32 extRegVal;

    ix_uint32 origNext = 0;
    ix_uint16 orig_bufferInfo = 0;

    ix_pci_controller* pPCIController = (ix_pci_controller*)arg_pContext;

    unsigned long flags;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* Get remote SRAM and DRAM base addresss from controller. */
    remoteSRAMBase = pPCIController->m_RemoteSRAMBase;
    remoteDRAMBase = pPCIController->m_RemoteDRAMBase;

    /* Get the base address, size, and offset of the remote meta-data. */
    if (arg_pBufferInfo->m_BufferType == IX_BUFFER_TYPE_HARDWARE)
    {
        ix_hw_buffer_meta* pHWMeta = (ix_hw_buffer_meta*)arg_pLocalDestination;
        remoteDataBase = remoteSRAMBase;

        /* For HW meta, save the old values. */

#if !defined(_IX_RM_SPLIT_META_DATA_)

        origNext = pHWMeta->m_HwNext;

#endif /* !defined(_IX_RM_SPLIT_META_DATA_) */

        orig_bufferInfo = pHWMeta->m_BufferInfo;
    }
    else
        remoteDataBase = remoteDRAMBase;

    remoteDataSize = arg_pBufferInfo->m_MetaSize;

    remoteDataAddress = arg_pBufferInfo->m_MetaOffset;

    remoteDataAddress += remoteDataBase;

    /* Disable all interrupts. */
    local_irq_save(flags);

    /* Get the address for desired PCI memory offset as seen from the CPU. */
    _IX_PCI_REG_READ(IX_PCI_ADDR_EXT, extRegVal);
  
    _IX_PCI_GET_CPU_ADDRESS(remoteDataAddress);

    /**
     * Copy the remote bytes to the local destination.
     */
    if (arg_pBufferInfo->m_BufferType == IX_BUFFER_TYPE_SOFTWARE)
    {
        /**
         * For SW buffers, we only need to copy the m_DataSize
         * field of the meta-data.
         */
        ((ix_sw_buffer_meta*)pDestination)->m_DataSize =
            ((ix_sw_buffer_meta*)remoteDataAddress)->m_DataSize;
    }
    else
    {
        /* ix_ossl_memcpy((void*)pDestination, (void*)remoteDataAddress,
            remoteDataSize); */
        ix_rm_managed_to_managed_memory_copy(pDestination,
                                            (void*)remoteDataAddress,
                                            remoteDataSize
                                          );
    }

    /* Reset the upper word in PCI addresses back to its original value. */
    _IX_PCI_REG_WRITE(IX_PCI_ADDR_EXT, extRegVal);

    /* Re-enable all interrupts. */
    local_irq_restore(flags);

    /* ix_ossl_sleep(0); */

    if (arg_pBufferInfo->m_BufferType == IX_BUFFER_TYPE_HARDWARE)
    {
        ix_hw_buffer_meta* pHWMeta = (ix_hw_buffer_meta*)arg_pLocalDestination;

        /* For HW meta, restore the old values. */

#if !defined(_IX_RM_SPLIT_META_DATA_)

        pHWMeta->m_HwNext = origNext;

#endif /* !defined(_IX_RM_SPLIT_META_DATA_) */

        IX_RM_HW_META_SET_FREE_LIST_ID(pHWMeta->m_BufferInfo, 
            IX_RM_HW_META_GET_FREE_LIST_ID(orig_bufferInfo));

        /* Return the data size for this buffer. */
        *arg_pDataSize = pHWMeta->m_BufferSize;
    }
    else
    {
        /* Return the data size for this buffer. */
        *arg_pDataSize = ((ix_sw_buffer_meta*)pDestination)->m_DataSize;
    }

    return err;
}

#else

#error "PCI support is not enabled on this OS!"

#endif /* _IX_OS_TYPE_ */


/**
 * NAME: _ix_rm_pci_copy_data_from_remote_subsystem
 *
 * DESCRIPTION: This function, called from the Rx thread, copies
 *              the buffer data from the other side.
 *
 * @Param:  - IN ix_pci_buffer_info* arg_pBufferInfo - pointer to
 *          control information for the buffer to be transferred.
 * @Param:  - IN void* arg_pContext - pointer to the context that provides
 *          control information to the Rx thread.
 * @Param:  - IN ix_uint32 arg_dataSize - size of the remote data.
 * @Param:  - OUT void* arg_pLocalDestination - address to where
 *          remote data will be written.
 *
 * @Return: IX_SUCCESS, or an appropriate error code for failure.
 *
 */

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

ix_error _ix_rm_pci_copy_data_from_remote_subsystem(
                                               ix_pci_buffer_info* arg_pBufferInfo,
                                               ix_uint32 arg_dataSize,
                                               void* arg_pContext,
                                               void* arg_pLocalDestination
                                               )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_copy_from_remote_subsystem";

    ix_error err = IX_SUCCESS;

    ix_uint32 remoteSRAMBase;
    ix_uint32 remoteDRAMBase;
    ix_uint32 remoteDataBase;

    ix_uint32 remoteDataAddress;
    ix_uint32 remoteDataSize;

    ix_uint32* pDestination = arg_pLocalDestination;

    ix_uint32 extRegVal;

#if defined(IX_PCI_A0_SILICON)
    ix_uint32* pSource;
    ix_uint32 i;
    ix_uint32 temp;
#endif /* end defined(IX_PCI_A0_SILICON) */

    ix_pci_controller* pPCIController = (ix_pci_controller*)arg_pContext;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* Get remote SRAM and DRAM base addresss from controller. */
    remoteSRAMBase = pPCIController->m_RemoteSRAMBase;
    remoteDRAMBase = pPCIController->m_RemoteDRAMBase;

    if (arg_pBufferInfo->m_DataMemoryType == IX_MEMORY_TYPE_SRAM)
        remoteDataBase = remoteSRAMBase;
    else if (arg_pBufferInfo->m_DataMemoryType == IX_MEMORY_TYPE_DRAM)
        remoteDataBase = remoteDRAMBase;
    else
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_UNSUPPORTED_MEMORY_TYPE, s_aFunctionIdentifier, 
                    "tried to transfer unsupported memory type!", err);

    #if defined(IX_PCI_A0_SILICON)
        /* size expressed in 32 bit words */
        remoteDataSize = (arg_dataSize + 3) >> 2;
    #else
        remoteDataSize = arg_dataSize;
    #endif /* defined(IX_PCI_A0_SILICON) */

    remoteDataAddress = arg_pBufferInfo->m_DataOffset;

    remoteDataAddress += remoteDataBase;

#if defined(_IX_RM_IMPL_HARDWARE_)

    /* Disable all PCI interrupts. */
    intDisable(INT_VEC_PIL); 

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    /* Get the address for desired PCI memory offset as seen from the CPU. */
    _IX_PCI_REG_READ(IX_PCI_ADDR_EXT, extRegVal);
  
    _IX_PCI_GET_CPU_ADDRESS(remoteDataAddress);

    /**
     * Copy the remote bytes to the local destination.
     */
    #if defined(IX_PCI_A0_SILICON)
        for(i = 0, pSource = (ix_uint32*)remoteDataAddress
            ; i < remoteDataSize; i++, pSource++, pDestination++)
        {
            *pDestination = *pSource;
            /* A read cycle to work around the A0 bug. */
            temp = *pSource;
        }
    #else
        /* ix_ossl_memcpy((void*)pDestination, (void*)remoteDataAddress, remoteDataSize); */
        ix_rm_managed_to_managed_memory_copy(pDestination,
                                            (void*)remoteDataAddress,
                                            remoteDataSize
                                          );
    #endif /*#if defined(IX_PCI_A0_SILICON) */

    /* Reset the upper word in PCI addresses back to its original value. */
    _IX_PCI_REG_WRITE(IX_PCI_ADDR_EXT, extRegVal);

#if defined(_IX_RM_IMPL_HARDWARE_)

    /* Re-enable all PCI interrupts. */
    intEnable(INT_VEC_PIL); 

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    return err;
}

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

#define PCI_MAX_TRANSFER_SIZE   64

ix_error _ix_rm_pci_copy_data_from_remote_subsystem(
                                               ix_pci_buffer_info* arg_pBufferInfo,
                                               ix_uint32 arg_dataSize,
                                               void* arg_pContext,
                                               void* arg_pLocalDestination
                                               )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_pci_copy_from_remote_subsystem";

    ix_error err = IX_SUCCESS;

    ix_uint32 remoteSRAMBase;
    ix_uint32 remoteDRAMBase;
    ix_uint32 remoteDataBase;

    ix_uint32 remoteDataAddress, remoteTempDataAddress;
    ix_uint32 remoteDataSize;

    ix_uint8* pDestination = arg_pLocalDestination;

    ix_uint32 extRegVal;

    ix_pci_controller* pPCIController = (ix_pci_controller*)arg_pContext;

    unsigned long flags;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* Get remote SRAM and DRAM base addresss from controller. */
    remoteSRAMBase = pPCIController->m_RemoteSRAMBase;
    remoteDRAMBase = pPCIController->m_RemoteDRAMBase;

    if (arg_pBufferInfo->m_DataMemoryType == IX_MEMORY_TYPE_SRAM)
        remoteDataBase = remoteSRAMBase;
    else if (arg_pBufferInfo->m_DataMemoryType == IX_MEMORY_TYPE_DRAM)
        remoteDataBase = remoteDRAMBase;
    else
        return IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_PCI_UNSUPPORTED_MEMORY_TYPE, s_aFunctionIdentifier, 
                    "tried to transfer unsupported memory type!", err);

    remoteDataSize = arg_dataSize;

    remoteTempDataAddress = arg_pBufferInfo->m_DataOffset;

    remoteTempDataAddress  += remoteDataBase;

    while(remoteDataSize > PCI_MAX_TRANSFER_SIZE)
    {
        remoteDataAddress = remoteTempDataAddress;

        /* Disable all interrupts. */
        local_irq_save(flags);

        /* Get the address for desired PCI memory offset as seen from the CPU. */
        _IX_PCI_REG_READ(IX_PCI_ADDR_EXT, extRegVal);
  
        _IX_PCI_GET_CPU_ADDRESS(remoteDataAddress);

        /**
         * Copy the remote bytes to the local destination.
         */
        /* ix_ossl_memcpy((void*)pDestination, (void*)remoteDataAddress, remoteDataSize); */
        ix_rm_managed_to_managed_memory_copy(pDestination,
                                            (void*)remoteDataAddress,
                                            PCI_MAX_TRANSFER_SIZE
                                          );

        remoteDataSize -= PCI_MAX_TRANSFER_SIZE;
        remoteTempDataAddress += PCI_MAX_TRANSFER_SIZE;
        pDestination += PCI_MAX_TRANSFER_SIZE;

        /* Reset the upper word in PCI addresses back to its original value. */
        _IX_PCI_REG_WRITE(IX_PCI_ADDR_EXT, extRegVal);

        /* Re-enable all interrupts. */
        local_irq_restore(flags);
    }

    remoteDataAddress = remoteTempDataAddress;

    /* Disable all interrupts. */
    local_irq_save(flags);

    /* Get the address for desired PCI memory offset as seen from the CPU. */
    _IX_PCI_REG_READ(IX_PCI_ADDR_EXT, extRegVal);
  

    _IX_PCI_GET_CPU_ADDRESS(remoteDataAddress);


    /**
     * Copy the remote bytes to the local destination.
     */
    /* ix_ossl_memcpy((void*)pDestination, (void*)remoteDataAddress, remoteDataSize); */
    ix_rm_managed_to_managed_memory_copy(pDestination,
                                        (void*)remoteDataAddress,
                                        remoteDataSize
                                      );

    /* Reset the upper word in PCI addresses back to its original value. */
    _IX_PCI_REG_WRITE(IX_PCI_ADDR_EXT, extRegVal);

    /* Re-enable all interrupts. */
    local_irq_restore(flags);

    /* ix_ossl_sleep(0); */

    return err;
}

#else

#error "PCI support is not enabled on this OS!"

#endif /* _IX_OS_TYPE_ */


#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

/**
 * NAME: ix_pci_doorbell_isr
 *
 * DESCRIPTION: Interrupt service routine for the PCI doorbell.  Reads
 *              the signal value from the local mailbox_1 register
 *              into the PCI controller data member for signal value.
 *              It then notifies the appropriate thread being signalled.
 *
 * @Param:  - IN int arg_Value - required input parameter for VxWorks
 *          ISR - ignored in this function.
 *
 */
void ix_pci_doorbell_isr(int arg_Value)
{
    static const char s_aFunctionIdentifier[] = "ix_pci_doorbell_isr";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* acknowledge the interrupt */
    _IX_PCI_REG_WRITE(IX_PCI_LOCAL_DOORBELL, 0xffffffff);

    /* ix_ossl_message_log("Doorbell interrupt raised!\n"); */

    /* Read mailbox value into the control structure member for the signal value. */
    _IX_PCI_REG_READ(IX_PCI_LOCAL_MAILBOX_1, g_PCICommunicationController.m_SignalValue);

    /* reset the mailbox register */
    _IX_PCI_REG_WRITE(IX_PCI_LOCAL_MAILBOX_1, 0x00);

    if((g_PCICommunicationController.m_SignalValue & IX_PCI_SIGNAL_TX_FROM_RX) != 0)
    {
        /* ix_ossl_message_log("Doorbell interrupt raised - data received\n!"); */
        /* Signal Tx that Rx on the other side has finished processing a buffer. */
        g_PCICommunicationController.m_pTxInterface->m_NotifyEngineFuncPtr(&g_PCICommunicationController);
    }
    if((g_PCICommunicationController.m_SignalValue & IX_PCI_SIGNAL_RX_FROM_TX) != 0)
    {
        /* ix_ossl_message_log("Doorbell interrupt raised - data acknowledged\n!"); */
        /* Signal Rx that Tx on the other side has a buffer ready. */
        g_PCICommunicationController.m_pRxInterface->m_NotifyEngineFuncPtr(&g_PCICommunicationController);
    }

    return;
}

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

/**
 * NAME: ix_pci_doorbell_isr
 *
 * DESCRIPTION: Interrupt service routine for the PCI doorbell.  Reads
 *              the signal value from the local mailbox_1 register
 *              into the PCI controller data member for signal value.
 *              It then notifies the appropriate thread being signalled.
 *
 * @Param:  - IN int irq - required input parameter for Linux
 *          ISR - ignored in this function.
 *
 */
void ix_pci_doorbell_isr(int irq, void* context, struct pt_regs* regs)
{
    static const char s_aFunctionIdentifier[] = "ix_pci_doorbell_isr";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* acknowledge the interrupt */
    _IX_PCI_REG_WRITE(IX_PCI_LOCAL_DOORBELL, 0xffffffff);

    /* ix_ossl_message_log("Doorbell interrupt raised!\n"); */

    /* Read mailbox value into the control structure member for the signal value. */
    _IX_PCI_REG_READ(IX_PCI_LOCAL_MAILBOX_1, g_PCICommunicationController.m_SignalValue);

    /* reset the mailbox register */
    _IX_PCI_REG_WRITE(IX_PCI_LOCAL_MAILBOX_1, 0x00);

    if((g_PCICommunicationController.m_SignalValue & IX_PCI_SIGNAL_TX_FROM_RX) != 0)
    {
        /* Signal Tx that Rx on the other side has finished processing a buffer. */
        g_PCICommunicationController.m_pTxInterface->m_NotifyEngineFuncPtr(&g_PCICommunicationController);
    }
    if((g_PCICommunicationController.m_SignalValue & IX_PCI_SIGNAL_RX_FROM_TX) != 0)
    {
        /* Signal Rx that Tx on the other side has a buffer ready. */
        g_PCICommunicationController.m_pRxInterface->m_NotifyEngineFuncPtr(&g_PCICommunicationController);
    }

    return;
}

#else

#error "PCI support is not enabled on this OS!"

#endif /* _IX_OS_TYPE_ */

#endif /* end defined(_IX_COMPILE_PCI_COMM_IN_) */
