/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      internal_remote_communication.c
 *
 * = DESCRIPTION
 *      This file defines all internal variables and functions required by the remote
 *      communication API.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      11/27/2002 3:39:16 PM - creation time 
 * ============================================================================
 */

static const char s_FileIdentifier[] = "$Id: internal_remote_communication.c,v 1.5 2003/09/05 19:52:34 ageorges Exp $";


#include "ix_ossl.h"

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"

#include "rm/symbols.h"
#include "rm/internal/internal_support.h"
#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"
#include "rm/system.h"
#include "rm/internal/internal_system.h"

#include "rm/memory.h"

#include "rm/buffer.h"
#include "rm/internal/internal_buffer.h"
#include "rm/hardware.h"
#include "rm/communication.h"
#include "rm/internal/internal_communication.h"

#include "rm/remote_communication.h"
#include "rm/internal/internal_remote_communication.h"


/**
 * Global remote communication manager
 */
ix_remote_comm_manager g_RemoteCommManager;

#if defined( _IX_RM_MULTI_THREADED_ )

static ix_ossl_sem_t s_RemoteCommManagerSemaphore = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */



/**
 * Default remote service initializer and finalizer declarations
 */

static ix_error _ix_rm_remote_comm_default_service_initializer(
                                                                void* arg_pContext
                                                              );

static ix_error _ix_rm_remote_comm_default_service_finalizer(
                                                              void* arg_pContext
                                                            );


/**
 * Default remote communication service
 */
ix_remote_comm_service g_DefaultRemoteService = 
{
    _ix_rm_remote_comm_default_service_initializer,
    _ix_rm_remote_comm_default_service_finalizer,
    0,
    _ix_rm_remote_comm_default_packet_handler,
    0,
    _ix_rm_remote_comm_default_message_handler,
    0
};

/**
 * NAME: _ix_rm_remote_comm_default_packet_handler
 *
 * DESCRIPTION: This function will be the default packet remote communication handler
 * 
 * @Param:  - IN arg_hBuffer - the handle of the buffer to be processed
 * @Param:  - IN arg_UserData - the user data associated with the buffer to process
 * @Param:  - IN arg_pContext - context that has been passed to RM at the initialization
 *  of the service
 * @Param:  - IN arg_CommId - the remote communication Id
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_remote_comm_default_packet_handler(
                                ix_buffer_handle arg_hBuffer,
                                ix_uint32 arg_UserData,
                                void* arg_pContext,
                                ix_communication_id arg_CommId
                              )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_remote_comm_default_packet_handler";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_hBuffer != IX_NULL_BUFFER_HANDLE)
        return ix_rm_buffer_free_chain(arg_hBuffer);
    else
        return IX_SUCCESS;
}


/**
 * NAME: _ix_rm_remote_comm_default_message_handler
 *
 * DESCRIPTION: This function will be the default message remote communication handler
 * 
 * @Param:  - IN arg_hBuffer - the handle of the buffer to be processed
 * @Param:  - IN arg_UserData - the user data associated with the buffer to process
 * @Param:  - IN arg_pContext - context that has been passed to RM at the initialization
 *  of the service
 * @Param:  - IN arg_CommId - the remote communication Id
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_remote_comm_default_message_handler(
                                ix_buffer_handle arg_hBuffer,
                                ix_uint32 arg_UserData,
                                void* arg_pContext,
                                ix_communication_id arg_CommId
                              )
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_remote_comm_default_message_handler";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(arg_hBuffer != IX_NULL_BUFFER_HANDLE)
        return ix_rm_buffer_free_chain(arg_hBuffer);
    else
        return IX_SUCCESS;
}


/**
 * NAME: _ix_rm_remote_comm_default_service_initializer
 *
 * DESCRIPTION: This function is the initializer for the default remote communication service
 * 
 * @Param:  - INOUT arg_pContext - context that was passed to RM at the registration of the service. 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error _ix_rm_remote_comm_default_service_initializer(void* arg_pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_remote_comm_default_service_initializer";

    ix_error err = IX_SUCCESS;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    return err;
}

/**
 * NAME: _ix_rm_remote_comm_default_service_finalizer
 *
 * DESCRIPTION: This function is the finalizer for the default remote communication service
 * 
 * @Param:  - INOUT arg_pContext - context that was passed to RM at the registration of the service. 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */

ix_error _ix_rm_remote_comm_default_service_finalizer(void* arg_pContext)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_remote_comm_default_service_finalizer";

    ix_error err = IX_SUCCESS;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    return err;
}


/**
 * NAME: _ix_rm_remote_comm_init
 *
 * DESCRIPTION: This function should be called at the time the resource manager initializer 
 *          function is called. The function will initialize all the Remote Communication API.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_remote_comm_init(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_remote_comm_init";

    ix_error err = IX_SUCCESS;
    ix_end_point* pEndPoint;
    ix_end_point_cc_data* pCCEPData;
    ix_end_point_data_queue* pCCEPDataQueue;
    void* pPacketQueueData = 0;
    void* pMessageQueueData = 0;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    /* initialize the synchronization semaphore */
    err = ix_ossl_sem_init(IX_OSSL_SEM_AVAILABLE, &s_RemoteCommManagerSemaphore);
    if(err != IX_SUCCESS)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_REMOTE_COMM_API_INIT_FAILED, s_aFunctionIdentifier, 
                        "initialization of remote communication manager's semaphore failed!", err);
    }

    /* initialize synchronization */
    _ix_rm_sa_sync_adapter_init(&g_RemoteCommManager.m_SyncAdapter, 
        _ix_rm_sa_default_lock, _ix_rm_sa_default_unlock, (void*)&s_RemoteCommManagerSemaphore);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /**
     * Initialize the remote communication end points
     */

#if defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_)

    /* Initialize the end point queues memory areas */
    pPacketQueueData = ix_ossl_malloc(IX_CORE_PCI_PACKET_DATA_RING_SIZE * IX_CORE_PCI_PACKET_DATA_SIZE * sizeof(ix_uint32));
    if(pPacketQueueData == 0)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_DATA_ALLOCATION_FAILED, s_aFunctionIdentifier, 
                        "the allocation of remote end point packet queue data area failed!", err);
        goto label_1;
    }

    pMessageQueueData = ix_ossl_malloc(IX_CORE_PCI_MESSAGE_DATA_RING_SIZE * IX_CORE_PCI_MESSAGE_DATA_SIZE * sizeof(ix_uint32));
    if(pMessageQueueData == 0)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_DATA_ALLOCATION_FAILED, s_aFunctionIdentifier, 
                        "the allocation of remote end point message queue data area failed!", err);
        goto label_1;
    }

#else

    /* Initialize the end point queues memory areas */
    err = ix_rm_mem_alloc(IX_MEMORY_TYPE_DRAM, 
                          0, 
                          IX_CORE_PCI_PACKET_DATA_RING_SIZE * IX_CORE_PCI_PACKET_DATA_SIZE * sizeof(ix_uint32) , 
                          &pPacketQueueData
                        );
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_DATA_ALLOCATION_FAILED, s_aFunctionIdentifier, 
                        "the allocation of remote end point packet queue data area failed!", err);
        goto label_1;
    }

    err = ix_rm_mem_alloc(IX_MEMORY_TYPE_DRAM, 
                          0, 
                          IX_CORE_PCI_MESSAGE_DATA_RING_SIZE * IX_CORE_PCI_MESSAGE_DATA_SIZE * sizeof(ix_uint32) , 
                          &pMessageQueueData
                        );
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_DATA_ALLOCATION_FAILED, s_aFunctionIdentifier, 
                        "the allocation of remote end point message queue data area failed!", err);
        goto label_1;
    }

#endif /* defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_) */

    /* initialize the packet PCI end point */
    pEndPoint = &g_RemoteCommManager.m_RemoteCommPacketEndPoint;

    pEndPoint->m_EPBehaviour.m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)_ix_rm_remote_comm_default_packet_handler;
    pEndPoint->m_EPBehaviour.m_pInternalBehaviourContext = 0;
    pEndPoint->m_EPBehaviour.m_pExternalBehaviourContext = 0;

    pCCEPData = &pEndPoint->m_EPData.m_CoreCompData;

    /* initialize the data queue */
    pCCEPDataQueue = &pCCEPData->m_EPDataQueue;
    pCCEPDataQueue->m_Put = pCCEPDataQueue->m_Get = 0; 
    pCCEPDataQueue->m_pStartAddress = pPacketQueueData;
    pCCEPDataQueue->m_Size = IX_CORE_PCI_PACKET_DATA_RING_SIZE;
    pCCEPDataQueue->m_Capacity = pCCEPDataQueue->m_Size - 1;
    pCCEPDataQueue->m_ElementSize = IX_CORE_PCI_PACKET_DATA_SIZE * sizeof(ix_uint32);

    pCCEPData->m_Flags = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    pCCEPData->m_pSendNotificationQueue = 0; 
    pCCEPData->m_pReceiveNotificationQueue = 0; 

    err = ix_ossl_sem_init(IX_OSSL_SEM_AVAILABLE, &pCCEPData->m_SyncSemaphore);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_SEMA_CREATION_FAILED, s_aFunctionIdentifier, 
                        "the creation of the remote communication packet EP sync semaphore failed!", err);
        goto label_1;
    }

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* initialize the message PCI end point */
    pEndPoint = &g_RemoteCommManager.m_RemoteCommMessageEndPoint;

    pEndPoint->m_EPBehaviour.m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)_ix_rm_remote_comm_default_message_handler;
    pEndPoint->m_EPBehaviour.m_pInternalBehaviourContext = 0;
    pEndPoint->m_EPBehaviour.m_pExternalBehaviourContext = 0;

    pCCEPData = &pEndPoint->m_EPData.m_CoreCompData;

    /* initialize the data queue */
    pCCEPDataQueue = &pCCEPData->m_EPDataQueue;
    pCCEPDataQueue->m_Put = pCCEPDataQueue->m_Get = 0; 
    pCCEPDataQueue->m_pStartAddress = pMessageQueueData;
    pCCEPDataQueue->m_Size = IX_CORE_PCI_MESSAGE_DATA_RING_SIZE;
    pCCEPDataQueue->m_Capacity = pCCEPDataQueue->m_Size - 1;
    pCCEPDataQueue->m_ElementSize = IX_CORE_PCI_MESSAGE_DATA_SIZE * sizeof(ix_uint32);

    pCCEPData->m_Flags = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    pCCEPData->m_pSendNotificationQueue = 0; 
    pCCEPData->m_pReceiveNotificationQueue = 0; 

    err = ix_ossl_sem_init(IX_OSSL_SEM_AVAILABLE, &pCCEPData->m_SyncSemaphore);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_COMM_EP_SEMA_CREATION_FAILED, s_aFunctionIdentifier, 
                        "the creation of the remote communication message EP sync semaphore failed!", err);
        goto label_1;
    }

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* Initialize the default comm service */
    g_RemoteCommManager.m_RemoteCommService.m_ServiceInitializer = g_DefaultRemoteService.m_ServiceInitializer;
    g_RemoteCommManager.m_RemoteCommService.m_ServiceFinalizer = g_DefaultRemoteService.m_ServiceFinalizer;
    g_RemoteCommManager.m_RemoteCommService.m_pServiceContext = g_DefaultRemoteService.m_pServiceContext;
    g_RemoteCommManager.m_RemoteCommService.m_PacketHandler = g_DefaultRemoteService.m_PacketHandler;
    g_RemoteCommManager.m_RemoteCommService.m_pPacketHandlerContext = g_DefaultRemoteService.m_pPacketHandlerContext;
    g_RemoteCommManager.m_RemoteCommService.m_MessageHandler = g_DefaultRemoteService.m_MessageHandler;
    g_RemoteCommManager.m_RemoteCommService.m_pMessageHandlerContext = g_DefaultRemoteService.m_pMessageHandlerContext;

    return err;

label_1:

    if(pPacketQueueData != 0)
    {

#if defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_)

        ix_ossl_free(pPacketQueueData);

#else

        ix_rm_mem_free(pPacketQueueData);

#endif /* defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_) */
        
    }

    if(pMessageQueueData != 0)
    {

#if defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_)

        ix_ossl_free(pMessageQueueData);

#else

        ix_rm_mem_free(pMessageQueueData);

#endif /* defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_) */

    }

#if defined( _IX_RM_MULTI_THREADED_ )

    pEndPoint = &g_RemoteCommManager.m_RemoteCommMessageEndPoint;
    pCCEPData = &pEndPoint->m_EPData.m_CoreCompData;

    if(pCCEPData->m_SyncSemaphore != 0)
    {
        ix_ossl_sem_fini(pCCEPData->m_SyncSemaphore);
        pCCEPData->m_SyncSemaphore = 0;
    }

    pEndPoint = &g_RemoteCommManager.m_RemoteCommPacketEndPoint;
    pCCEPData = &pEndPoint->m_EPData.m_CoreCompData;

    if(pCCEPData->m_SyncSemaphore != 0)
    {
        ix_ossl_sem_fini(pCCEPData->m_SyncSemaphore);
        pCCEPData->m_SyncSemaphore = 0;
    }

    if(s_RemoteCommManagerSemaphore != 0)
    {
        ix_ossl_sem_fini(s_RemoteCommManagerSemaphore);
        s_RemoteCommManagerSemaphore = 0;
    }

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: _ix_rm_remote_comm_fini
 *
 * DESCRIPTION: This function should be called at the time the resource manager finalizer function
 *          is called. The function will provide cleanup for Remote Communication API.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_remote_comm_fini(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_rm_remote_comm_fini";

    ix_error err = IX_SUCCESS;
    ix_end_point* pEndPoint;
    ix_end_point_cc_data* pCCEPData;
    ix_end_point_data_queue* pCCEPDataQueue;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_RemoteCommManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */


    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* finalize the existing remote communication service */
    g_RemoteCommManager.m_RemoteCommService.m_ServiceFinalizer(g_RemoteCommManager.m_RemoteCommService.m_pServiceContext);

    /* finalize the packet remote communication end point */
    pEndPoint = &g_RemoteCommManager.m_RemoteCommPacketEndPoint;

    pEndPoint->m_EPBehaviour.m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)_ix_rm_remote_comm_default_packet_handler;
    pEndPoint->m_EPBehaviour.m_pInternalBehaviourContext = 0;
    pEndPoint->m_EPBehaviour.m_pExternalBehaviourContext = 0;

    pCCEPData = &pEndPoint->m_EPData.m_CoreCompData;

    pCCEPDataQueue = &pCCEPData->m_EPDataQueue;

    /* finalize the data queue */
    pCCEPDataQueue->m_Capacity = 1;
    pCCEPDataQueue->m_Put = pCCEPDataQueue->m_Get = 0;

    /* free the data queue associated memory */
    if(pCCEPDataQueue->m_pStartAddress != 0)
    {

#if defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_)

        ix_ossl_free(pCCEPDataQueue->m_pStartAddress);

#else

        ix_rm_mem_free(pCCEPDataQueue->m_pStartAddress);

#endif /* defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_) */
        
    }

    pCCEPDataQueue->m_pStartAddress = 0;

    pCCEPDataQueue->m_Capacity = 0;
    pCCEPDataQueue->m_Size = 0;
    
    pCCEPData->m_Flags = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    if(pCCEPData->m_pSendNotificationQueue != 0)
    {
        _IX_RM_EP_DATA_ACTION_LOCK(pCCEPData);
        _IX_RM_EP_DATA_WAIT_QUEUE_NOTIFY(pCCEPData->m_pSendNotificationQueue);
        _IX_RM_EP_DATA_ACTION_UNLOCK(pCCEPData);
    }
    if(pCCEPData->m_pReceiveNotificationQueue != 0)
    {
        _IX_RM_EP_DATA_ACTION_LOCK(pCCEPData);
        _IX_RM_EP_DATA_WAIT_QUEUE_NOTIFY(pCCEPData->m_pReceiveNotificationQueue);
        _IX_RM_EP_DATA_ACTION_UNLOCK(pCCEPData);
    }

    /* delete the EP sync semaphore */
    if(pCCEPData->m_SyncSemaphore != 0)
    {
        ix_ossl_sem_fini(pCCEPData->m_SyncSemaphore);
        pCCEPData->m_SyncSemaphore = 0;
    }

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* finalize the message remote communication end point */
    pEndPoint = &g_RemoteCommManager.m_RemoteCommMessageEndPoint;

    pEndPoint->m_EPBehaviour.m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)_ix_rm_remote_comm_default_message_handler;
    pEndPoint->m_EPBehaviour.m_pInternalBehaviourContext = 0;
    pEndPoint->m_EPBehaviour.m_pExternalBehaviourContext = 0;

    pCCEPData = &pEndPoint->m_EPData.m_CoreCompData;
    pCCEPDataQueue = &pCCEPData->m_EPDataQueue;

    /* finalize the data queue */
    pCCEPDataQueue->m_Capacity = 1;

    pCCEPDataQueue->m_Put = pCCEPDataQueue->m_Get = 0;

    /* free the data queue associated memory */
    if(pCCEPDataQueue->m_pStartAddress != 0)
    {

#if defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_)

        ix_ossl_free(pCCEPDataQueue->m_pStartAddress);

#else

        ix_rm_mem_free(pCCEPDataQueue->m_pStartAddress);

#endif /* defined(_IX_RM_COMM_ID_QUEUES_IN_OS_MEMORY_) */
        
    }

    pCCEPDataQueue->m_pStartAddress = 0;

    pCCEPDataQueue->m_Capacity = 0;
    pCCEPDataQueue->m_Size = 0;
    
    pCCEPData->m_Flags = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    if(pCCEPData->m_pSendNotificationQueue != 0)
    {
        _IX_RM_EP_DATA_ACTION_LOCK(pCCEPData);
        _IX_RM_EP_DATA_WAIT_QUEUE_NOTIFY(pCCEPData->m_pSendNotificationQueue);
        _IX_RM_EP_DATA_ACTION_UNLOCK(pCCEPData);
    }
    if(pCCEPData->m_pReceiveNotificationQueue != 0)
    {
        _IX_RM_EP_DATA_ACTION_LOCK(pCCEPData);
        _IX_RM_EP_DATA_WAIT_QUEUE_NOTIFY(pCCEPData->m_pReceiveNotificationQueue);
        _IX_RM_EP_DATA_ACTION_UNLOCK(pCCEPData);
    }

    /* delete the EP sync semaphore */
    if(pCCEPData->m_SyncSemaphore != 0)
    {
        ix_ossl_sem_fini(pCCEPData->m_SyncSemaphore);
        pCCEPData->m_SyncSemaphore = 0;
    }

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* finalize synchronization */
    _ix_rm_sa_sync_adapter_init(&g_RemoteCommManager.m_SyncAdapter, 0, 0, 0);

    /* finalize the manager's synchronization semaphore */
    ix_ossl_sem_fini(s_RemoteCommManagerSemaphore);
    s_RemoteCommManagerSemaphore = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


