/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 4
 *
 * = LIBRARY
 *      Resource Manager Library
 *
 * = MODULE
 *      System API
 *
 * = FILENAME
 *      internal_system.c
 *
 * = DESCRIPTION
 *      This file will define all the internal supporting functions for the 
 *      System API.
 *
 * = AUTHOR
 *       Adrian Georgescu
 *       adrian.georgescu@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      10/26/2001 11:40:05 AM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */

static const char s_aFileIdentifier[] = "$Id: internal_system.c,v 1.22 2003/08/20 18:48:42 ageorges Exp $";

#include "ix_ossl.h"

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"

#include "rm/symbols.h"
#include "rm/internal/internal_support.h"

#include "rm/system.h"
#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"
#include "rm/internal/internal_system.h"

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

/* BSP-specific headers. */
#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

    #include <vxworks/ixp2400/ixdp2400misc.h>

#else

    #include <vxworks/ixp2800/ixdp2000misc.h>

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) && defined(_IX_RM_IMPL_HARDWARE_)

#include <asm/arch/hardware.h>

#else

#endif /* end (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) */

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */


#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * This object will hold system wide configuration information.
 * It will be used by ix_rm_sys_config_get() and ix_rm_sys_config_set() 
 * interface calls.
 */
_ix_sys_config g_SystemConfiguration;

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */


#if defined( _IX_RM_MULTI_THREADED_ )

/**
 * Synchronization semaphore for system configuration
 */
static ix_ossl_sem_t s_SystemConfigurationSemaphore = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

 

/**
 * This object will keep track of all details related to the initailization process of 
 * Resource Manager library.
 */
_ix_init_object g_RMLibInitialization;

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * NAME: _ix_rm_system_init
 *
 * DESCRIPTION: This function should be called at the time the resource manager initializer 
 *          function is called. The function will initialize all the System API.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_system_init(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_system_init";

    ix_error err = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

	IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* reset the system configuration information */
    ix_ossl_memset(&g_SystemConfiguration.m_SysConfig, 0, sizeof(g_SystemConfiguration.m_SysConfig));

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter = &g_SystemConfiguration.m_SyncAdapter;

    if(s_SystemConfigurationSemaphore == 0)
    {
        /* initialize the configuration property manager synchronization */
        err = ix_ossl_sem_init(IX_OSSL_SEM_AVAILABLE, &s_SystemConfigurationSemaphore);
        if(err)
        {
            return IX_RM_ERROR_NEW(IX_RM_ERROR_LIB_INIT_FAILED, s_aFunctionIdentifier, 
                            "initialization of system configurationsemaphore failed!", err);
        }

        _ix_rm_sa_sync_adapter_init(pSyncAdapter, _ix_rm_sa_default_lock, 
            _ix_rm_sa_default_unlock, (void*)&s_SystemConfigurationSemaphore);
    }

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */


    /* initialize the system configuration */
    {
        g_SystemConfiguration.m_SysConfig.m_Endianess = 0; /* big endian for now */

        g_SystemConfiguration.m_SysConfig.m_NumberOfDramChannels = IX_DRAM_CHANNELS_NUMBER;
        g_SystemConfiguration.m_SysConfig.m_NumberOfMicroengines = IX_MAX_MICROENGINE_NUMBER;
        g_SystemConfiguration.m_SysConfig.m_NumberOfSramChannels = IX_SRAM_CHANNELS_NUMBER;

        g_SystemConfiguration.m_SysConfig.m_NumberOfPorts = 0;
        ix_ossl_memset(g_SystemConfiguration.m_SysConfig.m_aPorts,0,
            sizeof(g_SystemConfiguration.m_SysConfig.m_aPorts));

        g_SystemConfiguration.m_SysConfig.m_SystemId = 0;

#if defined(_IX_RM_IMPL_HARDWARE_)

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

        if(isNpuMaster())

#else

        if(isNPUMaster())

#endif /* end (_IX_HARDWARE_TYPE_ != _IX_HW_2800_) */

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        if(npu_is_pcimaster())

#endif /* end (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) */

            g_SystemConfiguration.m_SysConfig.m_SubsystemType = IX_SUBSYSTEM_TYPE_EGRESS;
        else
            g_SystemConfiguration.m_SysConfig.m_SubsystemType = IX_SUBSYSTEM_TYPE_INGRESS;

#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)

        g_SystemConfiguration.m_SysConfig.m_ChipId = getProductID(0xffffffff);

#endif /* end (_IX_OS_TYPE_ == _IX_OS_VXWORKS_) */

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    }

    return err;
}


/**
 * NAME: _ix_rm_system_fini
 *
 * DESCRIPTION: This function should be called at the time the resource manager finalizer function
 *          is called. The function will provide cleanup for System API.
 * 
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 */
ix_error _ix_rm_system_fini(void)
{
    static const char s_aFunctionIdentifier[] = "_ix_rm_system_fini";

    ix_error err = 0;

	IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_MULTI_THREADED_)

    _ix_rm_sa_sync_adapter_init(&g_SystemConfiguration.m_SyncAdapter, 0, 0, 0);

    /* release the synchronization semaphore to the system */
    err = ix_ossl_sem_fini(s_SystemConfigurationSemaphore);
    if(err != 0)
    {
        return IX_RM_ERROR_NEW(IX_RM_ERROR_LIB_FINI_FAILED, s_aFunctionIdentifier, 
                        "finalization of system configuration synchronization semaphore failed!", err);
    }

    s_SystemConfigurationSemaphore = 0;

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    return err;
}

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */
