/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      me_notification.c
 *
 * = DESCRIPTION
 *      This file defines all functions and varibles required by
 *      the ME notification API
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      8/8/2003 2:21:06 PM - creation time 
 * ============================================================================
 */

static const char s_FileIdentifier[] = "$Id: me_notification.c,v 1.4 2003/10/10 21:08:50 ageorges Exp $";

#include "ix_ossl.h"

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"

#include "rm/symbols.h"
#include "rm/internal/internal_support.h"

#include <core_io.h>
#include <halMmap.h>

#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"

#include "rm/system.h"
#include "rm/internal/internal_system.h"

#include "rm/me_notification.h"


/**
 * ME CSRs offsets
 */

#define IX_RM_CTX_ENABLES_OFFSET          0x18
#define IX_RM_SAME_ME_SIGNAL_OFFSET       0x108


/**
 * NAME: ix_rm_me_transfer_register_read
 *
 * DESCRIPTION: This function will read the value of an ME transfer register for the
 *  specified ME, context and register type. Only the IX_ME_TRANSFER_REGISTER_TYPE_SRAM
 *  registers can be read. During the call, the code will check if the ME is in the 4 
 *  contexts mode or 8 contexts mode. In the first case the valid contexts numbers are
 *  0,2,4,6, whereas for the second case they are in the range 0-7.
 * 
 * @Param:  - IN arg_MENumber - the microengine whose transfer register we want to read
 * @Param:  - IN arg_ContextNumber - the ME context whose transfer register we want 
 *  to read
 * @Param:  - IN arg_RegisterType - the type of the ME transfer register we want to read.
 *  For reading just the IX_ME_TRANSFER_REGISTER_TYPE_SRAM is accepted.
 * @Param:  - IN arg_RegisterIndex - the index of the ME transfer register to be read.
 *  The value of the index is in the range 0-31 in the case the ME is in 4 contexts mode
 *  , and in the range 0-15 for the 8 contexts mode.
 * @Param:  - OUT arg_pRegisterValue - address where the value of the ME transfer register
 *  will be stored upon return.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_me_transfer_register_read(
                                          ix_uint32 arg_MENumber,
                                          ix_uint32 arg_ContextNumber, 
                                          ix_me_transfer_register_type arg_RegisterType,
                                          ix_uint32 arg_RegisterIndex,
                                          ix_uint32* arg_pRegisterValue
                                        )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_me_transfer_register_read";

    ix_error err = IX_SUCCESS;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

    IX_RM_ASSERT(((arg_MENumber <= 0x03) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x13))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x03 and 0x10 - 0x13!");

#elif (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

    IX_RM_ASSERT(((arg_MENumber <= 0x07) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x17))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x07 and 0x10 - 0x17!");

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

    if((READ_LWORD((ix_uint8*)(ix_uint32)Hal_cap_me_local_csr_virtAddr 
        + (arg_MENumber << 10) + IX_RM_CTX_ENABLES_OFFSET) & 0x80000000) == 0)
    {
        /* 8 contexts mode */

        IX_RM_ASSERT((arg_ContextNumber <= 7), 
            s_aFunctionIdentifier, "Bad context number. Valid numbers are: 0-7!");

        IX_RM_ASSERT((arg_RegisterIndex <= 15), 
            s_aFunctionIdentifier, "Bad transfer register index number. Valid numbers are: 0-15!");

    }
    else
    {
        /* 4 contexts mode */

        IX_RM_ASSERT(((arg_ContextNumber <= 7) && ((arg_ContextNumber % 2) == 0)), 
            s_aFunctionIdentifier, "Bad context number. Valid numbers are: 0,2,4,6!");

        IX_RM_ASSERT((arg_RegisterIndex <= 31), 
            s_aFunctionIdentifier, "Bad transfer register index number. Valid numbers are: 0-31!");

    }

    IX_RM_ASSERT((arg_RegisterType == IX_ME_TRANSFER_REGISTER_TYPE_SRAM), 
        s_aFunctionIdentifier, "Bad transfer register type. Valid type is: IX_ME_TRANSFER_REGISTER_TYPE_SRAM!");

    *arg_pRegisterValue = READ_LWORD(((ix_uint8*)(ix_uint32)Hal_cap_me_xfer_csr_virtAddr 
        + (arg_MENumber << 10) + 
        /* the following is not needed as we can read just SRAM registers */
        /* (((arg_RegisterType == IX_ME_TRANSFER_REGISTER_TYPE_SRAM) ? 0 : 1) << 9) + */
        (((arg_ContextNumber << 4 /* multiply by 16 */) + arg_RegisterIndex) << 2)));

    return err;
}



/**
 * NAME: ix_rm_me_transfer_register_write
 *
 * DESCRIPTION: This function will write the value of an ME transfer register for the
 *  specified ME, context and register type. During the call, the code will check if the ME 
 *  is in the 4 contexts mode or 8 contexts mode. In the first case the valid contexts 
 *  numbers are 0,2,4,6, whereas for the second case they are in the range 0-7.
 * 
 * @Param:  - IN arg_MENumber - the microengine whose transfer register we want to write
 * @Param:  - IN arg_ContextNumber - the ME context whose transfer register we want 
 *  to write.
 * @Param:  - IN arg_RegisterType - the type of the ME transfer register we want to write.
 * @Param:  - IN arg_RegisterIndex - the index of the ME transfer register to be read.
 *  The value of the index is in the range 0-31 in the case the ME is in 4 contexts mode
 *  , and in the range 0-15 for the 8 contexts mode.
 * @Param:  - IN arg_RegisterValue - the value to be written in the ME transfer register.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_me_transfer_register_write(
                                           ix_uint32 arg_MENumber,
                                           ix_uint32 arg_ContextNumber, 
                                           ix_me_transfer_register_type arg_RegisterType,
                                           ix_uint32 arg_RegisterIndex,
                                           ix_uint32 arg_RegisterValue
                                         )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_me_transfer_register_write";

    ix_error err = IX_SUCCESS;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

    IX_RM_ASSERT(((arg_MENumber <= 0x03) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x13))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x03 and 0x10 - 0x13!");

#elif (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

    IX_RM_ASSERT(((arg_MENumber <= 0x07) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x17))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x07 and 0x10 - 0x17!");

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

    if((READ_LWORD((ix_uint8*)(ix_uint32)Hal_cap_me_local_csr_virtAddr 
        + (arg_MENumber << 10) + IX_RM_CTX_ENABLES_OFFSET) & 0x80000000) == 0)
    {
        /* 8 contexts mode */

        IX_RM_ASSERT((arg_ContextNumber <= 7), 
            s_aFunctionIdentifier, "Bad context number. Valid numbers are: 0-7!");

        IX_RM_ASSERT((arg_RegisterIndex <= 15), 
            s_aFunctionIdentifier, "Bad transfer register index number. Valid numbers are: 0-15!");

    }
    else
    {
        /* 4 contexts mode */

        IX_RM_ASSERT(((arg_ContextNumber <= 7) && ((arg_ContextNumber % 2) == 0)), 
            s_aFunctionIdentifier, "Bad context number. Valid numbers are: 0,2,4,6!");

        IX_RM_ASSERT((arg_RegisterIndex <= 31), 
            s_aFunctionIdentifier, "Bad transfer register index number. Valid numbers are: 0-31!");

    }

    IX_RM_ASSERT(((arg_RegisterType >= IX_ME_TRANSFER_REGISTER_TYPE_FIRST) && 
        (arg_RegisterType < IX_ME_TRANSFER_REGISTER_TYPE_LAST)), 
        s_aFunctionIdentifier, "Bad transfer register type!");

    WRITE_LWORD(((ix_uint32)Hal_cap_me_xfer_csr_virtAddr 
        + (arg_MENumber << 10) + 
        (((arg_RegisterType == IX_ME_TRANSFER_REGISTER_TYPE_SRAM) ? 0 : 1) << 9) + 
        (((arg_ContextNumber << 4 /* multiply by 16 */) + arg_RegisterIndex) << 2)),
        arg_RegisterValue);

    return err;
}



/**
 * NAME: ix_rm_me_signal
 *
 * DESCRIPTION: This function will send a specified signal to the specified ME and 
 *  context. The signalling will be achieved by writing the SAME_ME_REGISTER for the
 *  desired ME. There is a 3 cycles latency from the write CSR command untill the CSR
 *  actually written, and 8 cycles until the ME will signaled after the CSR write.
 *  That should be considered when issuing many signals to the same ME.
 *  During the call, the code will check if the ME is in the 4  contexts mode or 8 
 *  contexts mode. In the first case the valid context numbers are 0,2,4,6 whereas
 *  for the second case they are in the range 0-7.
 * 
 * @Param:  - IN arg_MENumber - the microengine whose which we want to signal.
 * @Param:  - IN arg_ContextNumber - the ME context for which the signal is intended.
 * @Param:  - IN arg_SignalNumber - the signal number to be sent to the specified ME
 *  context. Valid values are in the range 1-15
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_me_signal(
                          ix_uint32 arg_MENumber,
                          ix_uint32 arg_ContextNumber, 
                          ix_uint32 arg_SignalNumber
                        )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_me_signal";

    ix_error err = IX_SUCCESS;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

    IX_RM_ASSERT(((arg_MENumber <= 0x03) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x13))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x03 and 0x10 - 0x13!");

#elif (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

    IX_RM_ASSERT(((arg_MENumber <= 0x07) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x17))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x07 and 0x10 - 0x17!");

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

    if((READ_LWORD((ix_uint8*)(ix_uint32)Hal_cap_me_local_csr_virtAddr 
        + (arg_MENumber << 10) + IX_RM_CTX_ENABLES_OFFSET) & 0x80000000) == 0)
    {
        /* 8 contexts mode */

        IX_RM_ASSERT((arg_ContextNumber <= 7), 
            s_aFunctionIdentifier, "Bad context number. Valid numbers are: 0-7!");

    }
    else
    {
        /* 4 contexts mode */

        IX_RM_ASSERT(((arg_ContextNumber <= 7) && ((arg_ContextNumber % 2) == 0)), 
            s_aFunctionIdentifier, "Bad context number. Valid numbers are: 0,2,4,6!");

    }

    
    IX_RM_ASSERT(((arg_SignalNumber <= 15) && (arg_SignalNumber > 0)), 
        s_aFunctionIdentifier, "Bad signal number. Valid numbers are: 1-15!");


    WRITE_LWORD(((ix_uint32)Hal_cap_me_local_csr_virtAddr 
        + (arg_MENumber << 10) + IX_RM_SAME_ME_SIGNAL_OFFSET), 
        ((arg_SignalNumber << 3) + arg_ContextNumber));

    return err;
}

#undef IX_RM_CTX_ENABLES_OFFSET
#undef IX_RM_SAME_ME_SIGNAL_OFFSET

