/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      me_xscale_locks.c
 *
 * = DESCRIPTION
 *      This file defines all functions and varibles required by
 *      the ME\XScale locks API
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      8/8/2003 1:32:39 PM - creation time 
 * ============================================================================
 */

static const char s_FileIdentifier[] = "$Id: me_xscale_locks.c,v 1.3 2003/10/10 21:08:50 ageorges Exp $";

#include "ix_ossl.h"

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"

#include "rm/symbols.h"
#include "rm/internal/internal_support.h"

#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"

#include "rm/system.h"
#include "rm/internal/internal_system.h"

#include "rm/memory.h"

#include "rm/me_xscale_locks.h"
#include "rm/internal/internal_me_xscale_locks.h"

#include "rm/atomic_operations.h"



/**
 * NAME: ix_rm_me_xscale_lock_new
 *
 * DESCRIPTION: This function will create a new me/xscale lock object.
 * 
 * @Param:  - OUT  arg_pMeXScaleLockHandle - location where a handle to the newly created
 *  me/xscale lock will be stored upon return.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_me_xscale_lock_new(
                                   ix_me_xscale_lock_handle* arg_pMeXScaleLockHandle
                                 )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_me_xscale_lock_new";

    ix_error err = IX_SUCCESS;
    _ix_me_xscale_lock* pLock;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_MeXScaleLockManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    if(g_MeXScaleLockManager.m_pLocksFreeList == 0)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_XSCALE_LOCK_OUT_OF_LOCKS, s_aFunctionIdentifier, 
                            "there are no more ME\\XSCale locks available!", err);
        goto label_1;
    }

    pLock = g_MeXScaleLockManager.m_pLocksFreeList;
    g_MeXScaleLockManager.m_pLocksFreeList = pLock->m_pNext;
    pLock->m_Owner = IX_ME_XSCALE_LOCK_OWNER_NONE;
    pLock->m_pNext = IX_ME_XSCALE_LOCK_ACTIVE;

    /* make the lock available */
    {
        ix_uint32 mask = 1 << pLock->m_BitPosition;
        ix_rm_atomic_scratch_bit_set(mask, pLock->m_pScratchMemory);
    }

    *arg_pMeXScaleLockHandle = pLock->m_hMeXScaleLock;

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: ix_rm_me_xscale_lock_delete
 *
 * DESCRIPTION: This function will delete the specified me/xscale lock object and
 *  will release all resources associated with it.
 * 
 * @Param:  - IN arg_hMeXScaleLock - the handle of the me/xscale lock to be deleted
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_me_xscale_lock_delete(
                                      ix_me_xscale_lock_handle arg_hMeXScaleLock
                                    )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_me_xscale_lock_delete";

    ix_error err = IX_SUCCESS;
    _ix_me_xscale_lock* pLock;
    ix_uint32 i;
    ix_uint32 mask;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_MeXScaleLockManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    IX_RM_ASSERT((IX_ME_XSCALE_LOCK_HANDLE_IS_VALID(arg_hMeXScaleLock) != 0), 
        s_aFunctionIdentifier, "Bad ME\\XScale lock handle!");

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    pLock = g_MeXScaleLockManager.m_MeXScaleLocks + 
        IX_ME_XSCALE_LOCK_HANDLE_GET_LOCK_INDEX(arg_hMeXScaleLock);

    if(pLock->m_pNext != IX_ME_XSCALE_LOCK_ACTIVE)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_XSCALE_LOCK_NOT_ACTIVE, s_aFunctionIdentifier, 
                            "the ME\\XScale lock you are trying to delete is not active!", err);
        goto label_1;
    }

    mask = 1 << pLock->m_BitPosition;
    i = ix_rm_atomic_scratch_bit_test_and_clear(mask, pLock->m_pScratchMemory);
    if((i & mask) == 0)
    {
        /* the lock was already acquired */
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_XSCALE_LOCK_IS_LOCKED, s_aFunctionIdentifier, 
                            "the ME\\XScale lock you are trying to delete is already locked!", err);
        goto label_1;
    }

    pLock->m_Owner = IX_ME_XSCALE_LOCK_OWNER_NONE;
    pLock->m_pNext = g_MeXScaleLockManager.m_pLocksFreeList;
    g_MeXScaleLockManager.m_pLocksFreeList = pLock;

    /* we leave the lock in unavailable state */

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}



/**
 * NAME: ix_rm_me_xscale_lock_acquire
 *
 * DESCRIPTION: This function will acquire the specified me/xscale lock. If the lock
 *  is already taken then an error will be returned. 
 * 
 * @Param:  - IN arg_hMeXScaleLock - the handle of the lock we are trying to acquire
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_me_xscale_lock_acquire( 
                                       ix_me_xscale_lock_handle arg_hMeXScaleLock
                                     )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_me_xscale_lock_acquire";

    ix_error err = IX_SUCCESS;
    _ix_me_xscale_lock* pLock;
    ix_uint32 i;
    ix_uint32 mask;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_MeXScaleLockManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    IX_RM_ASSERT((IX_ME_XSCALE_LOCK_HANDLE_IS_VALID(arg_hMeXScaleLock) != 0), 
        s_aFunctionIdentifier, "Bad ME\\XScale lock handle!");

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    pLock = g_MeXScaleLockManager.m_MeXScaleLocks + 
        IX_ME_XSCALE_LOCK_HANDLE_GET_LOCK_INDEX(arg_hMeXScaleLock);

    if(pLock->m_pNext != IX_ME_XSCALE_LOCK_ACTIVE)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_XSCALE_LOCK_NOT_ACTIVE, s_aFunctionIdentifier, 
                            "the ME\\XScale lock you are trying to acquire is not active!", err);
        goto label_1;
    }

    mask = 1 << pLock->m_BitPosition;
    i = ix_rm_atomic_scratch_bit_test_and_clear(mask, pLock->m_pScratchMemory);
    if((i & mask) == 0)
    {
        /* the lock was already acquired */
        if(pLock->m_Owner == IX_ME_XSCALE_LOCK_OWNER_NONE)
            pLock->m_Owner = IX_ME_XSCALE_LOCK_OWNER_ME;

        err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_XSCALE_LOCK_IS_LOCKED, s_aFunctionIdentifier, 
                            "the ME\\XScale lock you are trying to acquire is already locked!", err);
        goto label_1;
    }
    else
    {
        /* the lock is ours */
        pLock->m_Owner = IX_ME_XSCALE_LOCK_OWNER_XSCALE;
    }

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: ix_rm_me_xscale_lock_release
 *
 * DESCRIPTION: This function will release the specified ME/XScale lock. If the lock
 *  has been acquired by the MEs, an error will be returned. 
 * 
 * @Param:  - IN arg_hMeXScaleLock - the handle of the lock we are releasing
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_me_xscale_lock_release(
                                       ix_me_xscale_lock_handle arg_hMeXScaleLock
                                     )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_me_xscale_lock_release";

    ix_error err = IX_SUCCESS;
    _ix_me_xscale_lock* pLock;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_MeXScaleLockManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    IX_RM_ASSERT((IX_ME_XSCALE_LOCK_HANDLE_IS_VALID(arg_hMeXScaleLock) != 0), 
        s_aFunctionIdentifier, "Bad ME\\XScale lock handle!");

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    pLock = g_MeXScaleLockManager.m_MeXScaleLocks + 
        IX_ME_XSCALE_LOCK_HANDLE_GET_LOCK_INDEX(arg_hMeXScaleLock);

    if(pLock->m_pNext != IX_ME_XSCALE_LOCK_ACTIVE)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_XSCALE_LOCK_NOT_ACTIVE, s_aFunctionIdentifier, 
                            "the ME\\XScale lock you are trying to release is not active!", err);
        goto label_1;
    }

    switch(pLock->m_Owner)
    {
        case IX_ME_XSCALE_LOCK_OWNER_NONE:
            {
                ix_uint32 mask;
                mask = 1 << pLock->m_BitPosition;

                /* if it is locked and we are not the owner, then ME locked it */
                if((mask & (*pLock->m_pScratchMemory)) == 0)
                {
                    pLock->m_Owner = IX_ME_XSCALE_LOCK_OWNER_ME;
                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_XSCALE_LOCK_IS_LOCKED_BY_ME, s_aFunctionIdentifier, 
                            "the ME\\XScale lock you are trying to release is locked by the MEs!", err);
                }
            }
            goto label_1;
        case IX_ME_XSCALE_LOCK_OWNER_ME:
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_XSCALE_LOCK_IS_LOCKED_BY_ME, s_aFunctionIdentifier, 
                            "the ME\\XScale lock you are trying to release is locked by the MEs!", err);
            goto label_1;
        default:
            break;
    }

    ix_rm_atomic_scratch_bit_set((1 << pLock->m_BitPosition), pLock->m_pScratchMemory);

    pLock->m_Owner = IX_ME_XSCALE_LOCK_OWNER_NONE;

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: ix_rm_me_xscale_lock_get_info
 *
 * DESCRIPTION: This function will return useful information about the passed lock
 * 
 * @Param:  - IN arg_hMeXScaleLock - the handle of the lock we are interested in
 * @Param:  - OUT arg_pMeXScaleLockInfo - address of structure were the lock information
 *  will be stored upon return.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_me_xscale_lock_get_info(
                                        ix_me_xscale_lock_handle arg_hMeXScaleLock,
                                        ix_me_xscale_lock_info* arg_pMeXScaleLockInfo
                                      )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_me_xscale_lock_get_info";

    ix_error err = IX_SUCCESS;
    _ix_me_xscale_lock* pLock;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_MeXScaleLockManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    IX_RM_ASSERT((IX_ME_XSCALE_LOCK_HANDLE_IS_VALID(arg_hMeXScaleLock) != 0), 
        s_aFunctionIdentifier, "Bad ME\\XScale lock handle!");

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    pLock = g_MeXScaleLockManager.m_MeXScaleLocks + 
        IX_ME_XSCALE_LOCK_HANDLE_GET_LOCK_INDEX(arg_hMeXScaleLock);

    if(pLock->m_pNext != IX_ME_XSCALE_LOCK_ACTIVE)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_XSCALE_LOCK_NOT_ACTIVE, s_aFunctionIdentifier, 
                            "the ME\\XScale lock you are trying to get information about is not active!", err);
        goto label_1;
    }

    arg_pMeXScaleLockInfo->m_BitPosition = pLock->m_BitPosition;

    err = ix_rm_get_phys_offset(pLock->m_pScratchMemory, 0, 0, 
        &arg_pMeXScaleLockInfo->m_ScratchOffset, 0);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_XSCALE_LOCK_BAD_MEMORY, s_aFunctionIdentifier, 
                            "bad memory location for ME\\XScale lock!", err);
        goto label_1;
    }

    arg_pMeXScaleLockInfo->m_Status = (((*pLock->m_pScratchMemory) & (1 << pLock->m_BitPosition)) != 0) ?
        IX_ME_XSCALE_LOCK_STATUS_AVAILABLE : IX_ME_XSCALE_LOCK_STATUS_UNAVAILABLE;

    if((arg_pMeXScaleLockInfo->m_Status == IX_ME_XSCALE_LOCK_STATUS_UNAVAILABLE) &&
        (pLock->m_Owner != IX_ME_XSCALE_LOCK_OWNER_XSCALE))
        pLock->m_Owner = IX_ME_XSCALE_LOCK_OWNER_ME;

    arg_pMeXScaleLockInfo->m_Owner = pLock->m_Owner;

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


