/**i
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 4
 *
 * = LIBRARY
 *      Resource Manager Library
 *
 * = MODULE
 *      Memory Manager API
 *
 * = FILENAME
 *      memory.c
 *
 * = DESCRIPTION
 *      The file defines all Memory Manager interface functions.
 *
 * = AUTHOR
 *       Adrian Georgescu
 *       adrian.georgescu@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      10/17/2001 3:06:05 PM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */

static const char s_aFileIdentifier[] = "$Id: memory.c,v 1.28 2003/09/22 20:12:30 ageorges Exp $";

#include "ix_ossl.h"

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"


#include "rm/symbols.h"
#include "rm/internal/internal_support.h"

#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"

#include "rm/system.h"
#include "rm/internal/internal_system.h"

#include "rm/memory.h"
#include "rm/internal/internal_memory.h"

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * NAME: ix_rm_mem_alloc
 *
 * DESCRIPTION: This function allocates common ME/Core memory on behalf of the caller.
 *          On return, the location referred by arg_pMemoryAddr will be set to NULL if failure 
 *          or to the allocated memory block address if successful. The returned address will
 *          aligned according to the default alignment for the type of memory.
 *          If the size does not comply with memory alignment then the actual allocated size will
 *          be rounded up to insure that alignment will be satisfied for further allocations.
 * 
 * @Param:  - IN arg_MemType - represents the type of memory requested.
 * @Param:  - IN arg_MemChannel - represents the channel from which memory is requested.
 * @Param:  - IN arg_Size - represents the size in bytes of the block requested.
 * @Param:  - OUT arg_pMemoryAddr - represents the address where the allocated memory address
 *          should be stored.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure. The location referred by 
 *          arg_pMemoryAddr will be set to NULL if failure, or to the allocated memory block
 *          address if successful.
 */

#else /* Westport hardware */

/**
 * NAME: ix_rm_mem_alloc
 *
 * DESCRIPTION: This function allocates common ME/NPE/Core memory on behalf 
 *	of the caller. On return, the location referred by arg_pMemoryAddr will 
 *	be set to NULL if failure  or to the allocated memory block address if 
 *	successful. The returned address will be aligned according to the 
 *	default virtual alignment for the type of memory. If the size does not 
 *	comply with memory alignment then the actual allocated size will be 
 *	rounded up to insure that alignment will be satisfied for further 
 *	allocations.
 * 
 * @Param:  - IN arg_MemType - represents the type of memory requested.
 * @Param:  - IN arg_MemChannel - represents the channel from which memory is 
 *	requested.
 * @Param:  - IN arg_Size - represents the size in bytes of the block 
 *	requested.
 * @Param:  - OUT arg_pMemoryAddr - represents the address where the 
 *	allocated memory address should be stored.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure. 
 *	The location referred by arg_pMemoryAddr will be set to NULL if 
 *	failure, or to the allocated memory block address if successful.
 */

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

ix_error ix_rm_mem_alloc( 
                          ix_memory_type arg_MemType, 
                          ix_uint32 arg_MemChannel, 
                          ix_uint32 arg_Size, 
                          void** arg_pMemoryAddr 
                        )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_mem_alloc";

    /* Naming convention for automatic variables */
    ix_error err = 0;
    register _ix_mm_memory_manager* pMemoryManager;

    (void)s_aFunctionIdentifier;

    /* check the arg_MemType in DEBUG mode */
    IX_RM_ASSERT(arg_MemType <= IX_MEMORY_TYPE_SCRATCH, s_aFunctionIdentifier, "bad memory type!");

    /* check the arg_Size in DEBUG mode */
    IX_RM_ASSERT(arg_Size != 0, s_aFunctionIdentifier, "the request of 0 size memory is not acceptable!");

#if defined(_IX_RM_DEBUG_)

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

    switch(arg_MemType)
    {
        case IX_MEMORY_TYPE_DRAM:
            IX_RM_ASSERT(arg_MemChannel < IX_DRAM_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SRAM:
            IX_RM_ASSERT(arg_MemChannel < IX_SRAM_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        default:
            /* this is the SCRATCH memory */
            IX_RM_ASSERT(arg_MemChannel < IX_SCRATCH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
    }

#else /* Westport hardware */

    switch(arg_MemType)
    {
        case IX_MEMORY_TYPE_DRAM_SOUTH:
            IX_RM_ASSERT(arg_MemChannel < IX_DRAM_SOUTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SRAM_MSG_SOUTH:
            IX_RM_ASSERT(arg_MemChannel < IX_SRAM_MSG_SOUTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_DRAM_NORTH:
            IX_RM_ASSERT(arg_MemChannel < IX_DRAM_NORTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SRAM_NORTH:
            IX_RM_ASSERT(arg_MemChannel < IX_SRAM_NORTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SCRATCH:
            IX_RM_ASSERT(arg_MemChannel < IX_SCRATCH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        default:
            break;
    }

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

#endif /* defined(_IX_RM_DEBUG_) */

    pMemoryManager = g_aMemoryManagersGroups[arg_MemType].m_pMembers + arg_MemChannel;

    err = pMemoryManager->m_pVMT->m_AllocFuncPtr(pMemoryManager, arg_Size, (void**)arg_pMemoryAddr);

#ifdef _IX_RM_DEBUG_

    if(err != 0 && pMemoryManager->m_MemoryBlockSize != 0)
    {
        err = IX_RM_ERROR_CHAIN(err, s_aFunctionIdentifier, "memory allocation failed!");
    }

#endif /* _IX_RM_DEBUG_ */

    return err;
}

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * NAME: ix_rm_mem_alloc_aligned
 *
 * DESCRIPTION: This function allocates common ME/Core memory on behalf of the caller.
 *          On return, the location referred by arg_pMemoryAddr will be set to NULL if failure 
 *          or to the allocated memory block address if successful. The returned address will
 *          be alligned according to the arg_AlignmentSfift parameter and  arg_AlignmentType alignment
 *          type. If arg_AlignmentSfift parameter is less than default alignment then the return
 *          memory will be aligned according to the default alignment. 
 *          If the alignment type is not IX_MEMORY_ALIGNMENT_TYPE_VIRTUAL then the resulting virtual
 *          address might not be aligned according to the arg_AlignmentShift but the physiscal offset
 *          or physical address will be properly aligned.
 *          The actual allocated size will be rounded up to accomodate the required alignment.
 * 
 *
 * @Param:  - IN arg_MemType - represents the type of memory requested.
 * @Param:  - IN arg_MemChannel - represents the channel from which memory is requested.
 * @Param:  - IN arg_Size - represents the size in bytes of the block requested.
 * @Param:  - IN arg_AlignmentType - specifies the type of alignment required as explained at 
 *          ix_memory_alignment_type type description
 * @Param:  - IN arg_AlignmentShift - this value specifies the required memory block alignment as the
 *          power of 2 of the alignment. If the specified alignment is less than the
 *          default alignment for the corresonding memory then the default alignment will be used.
 * @Param:  - OUT arg_pMemoryAddr - represents the address where the allocated memory address
 *          should be stored.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure. The location referred by 
 *          arg_pMemoryAddr will be set to NULL if failure, or to the allocated memory block
 *          address if successful.
 */

#else /* Westport hardware */

/**
 * NAME: ix_rm_mem_alloc_aligned
 *
 * DESCRIPTION: This function allocates common ME/NPE/Core memory on behalf 
 *	of the caller. On return, the location referred by arg_pMemoryAddr will 
 *	be set to NULL for failure or to the allocated memory block address if 
 *	successful. The returned address will be aligned according to the 
 *	arg_AlignmentShift parameter and  arg_AlignmentType alignment type. If 
 *	arg_AlignmentShift parameter is less than default alignment then the 
 *	return memory will be aligned according to the default alignment. If 
 *	the alignment type is not IX_MEMORY_ALIGNMENT_TYPE_VIRTUAL then the 
 *	resulting virtual address might not be aligned according to the 
 *	arg_AlignmentShift but the physical offset or physical address will be 
 *	properly aligned. The actual allocated size will be rounded up to 
 *	accommodate the required alignment.
 * 
 * @Param:  - IN arg_MemType - represents the type of memory requested.
 * @Param:  - IN arg_MemChannel - represents the channel from which memory is 
 *	requested.
 * @Param:  - IN arg_Size - represents the size in bytes of the block 
 *	requested.
 * @Param:  - IN arg_AlignmentType - specifies the type of alignment required 
 *	as explained at ix_memory_alignment_type type description
 * @Param:  - IN arg_AlignmentShift - this value specifies the required 
 *	memory block alignment as the power of 2 of the alignment. If the 
 *	specified alignment is less than the default alignment for the 
 *	corresponding memory then the default alignment will be used.
 * @Param:  - OUT arg_pMemoryAddr - represents the address where the 
 *	allocated memory address should be stored.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure. 
 *	The location referred by arg_pMemoryAddr will be set to NULL if 
 *	failure, or to the allocated memory block address if successful.
 */

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

ix_error ix_rm_mem_alloc_aligned( 
                                  ix_memory_type arg_MemType, 
                                  ix_uint32 arg_MemChannel, 
                                  ix_uint32 arg_Size,
                                  ix_memory_alignment_type arg_AlignmentType,
                                  ix_uint32 arg_AlignmentShift,
                                  void** arg_pMemoryAddr 
                                )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_mem_alloc_aligned";

    ix_error err = 0;
    register _ix_mm_memory_manager* pMemoryManager;

    (void)s_aFunctionIdentifier;

    /* check the arg_MemType in DEBUG mode */
    IX_RM_ASSERT(arg_MemType <= IX_MEMORY_TYPE_SCRATCH, s_aFunctionIdentifier, "bad memory type!");
    /* check the arg_Size in DEBUG mode */
    IX_RM_ASSERT(arg_Size != 0, s_aFunctionIdentifier, "the request of 0 size memory is not acceptable!");

#if defined(_IX_RM_DEBUG_)

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

    switch(arg_MemType)
    {
        case IX_MEMORY_TYPE_DRAM:
            IX_RM_ASSERT(arg_MemChannel < IX_DRAM_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SRAM:
            IX_RM_ASSERT(arg_MemChannel < IX_SRAM_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        default:
            /* this is the SCRATCH memory */
            IX_RM_ASSERT(arg_MemChannel < IX_SCRATCH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
    }

#else /* Westport hardware */

    switch(arg_MemType)
    {
        case IX_MEMORY_TYPE_DRAM_SOUTH:
            IX_RM_ASSERT(arg_MemChannel < IX_DRAM_SOUTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SRAM_MSG_SOUTH:
            IX_RM_ASSERT(arg_MemChannel < IX_SRAM_MSG_SOUTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_DRAM_NORTH:
            IX_RM_ASSERT(arg_MemChannel < IX_DRAM_NORTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SRAM_NORTH:
            IX_RM_ASSERT(arg_MemChannel < IX_SRAM_NORTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SCRATCH:
            IX_RM_ASSERT(arg_MemChannel < IX_SCRATCH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        default:
            break;
    }

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

#endif /* defined(_IX_RM_DEBUG_) */

    pMemoryManager = g_aMemoryManagersGroups[arg_MemType].m_pMembers + arg_MemChannel;

    err = pMemoryManager->m_pVMT->m_AlignedAllocFuncPtr(pMemoryManager, arg_Size, arg_AlignmentType, arg_AlignmentShift, (void**)arg_pMemoryAddr);

#ifdef _IX_RM_DEBUG_

    if(err != 0 && pMemoryManager->m_MemoryBlockSize != 0)
    {
        err = IX_RM_ERROR_CHAIN(err, s_aFunctionIdentifier, "memory allocation failed!");
    }

#endif /* _IX_RM_DEBUG_ */

    return err;
}

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * NAME: ix_rm_mem_reserve
 *
 * DESCRIPTION: This function will reserve an area of memory at the specified memory channel byte offset
 *          and of specified size. The function is provided as support for the microC compiler and
 *          linker. This call should be made before any memory allocation has been made in order
 *          to reduce the chance of failure. If the memory channel byte offset is not aligned according
 *          to the default alignment for the memory type then the allocated size will be larger to
 *          accomodate the alignment. In general the byte offset and size should comply with
 *          the memory alignment, otherwise more memory will be reserved to insure the right alignment.
 * 
 * @Param:  - IN arg_MemType - represents the type of memory to be reserved.
 * @Param:  - IN arg_MemChannel - represents the channel from which memory is to be reserved.
 * @Param:  - IN arg_Size - represents the size in bytes of the memory block to be reserved.  
 * @Param:  - INOUT arg_pByteOffset - represents the address of byte offset of the memory block 
 *          requested for reservation. On return this value will be adjusted to accomodate the
 *          the memory type default alignment.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure. If any portion of the memory requested
 *          for reservation has been already allocated then the call will fail.
 */
ix_error ix_rm_mem_reserve( 
                            ix_memory_type arg_MemType, 
                            ix_uint32 arg_MemChannel, 
                            ix_uint32 arg_Size, 
                            ix_uint32* arg_pByteOffset 
                          )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_mem_reserve";

    ix_error err = 0;
    register _ix_mm_memory_manager* pMemoryManager;

    (void)s_aFunctionIdentifier;

    /* check the arg_MemType in DEBUG mode */
    IX_RM_ASSERT(arg_MemType <= IX_MEMORY_TYPE_SCRATCH, s_aFunctionIdentifier, "bad memory type!");
    /* check the arg_Size in DEBUG mode */
    IX_RM_ASSERT(arg_Size != 0, s_aFunctionIdentifier, "the request of 0 size memory is not acceptable!");

#if defined(_IX_RM_DEBUG_)

    switch(arg_MemType)
    {
        case IX_MEMORY_TYPE_DRAM:
            IX_RM_ASSERT(arg_MemChannel < IX_DRAM_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SRAM:
            IX_RM_ASSERT(arg_MemChannel < IX_SRAM_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        default:
            /* this is the SCRATCH memory */
            IX_RM_ASSERT(arg_MemChannel < IX_SCRATCH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
    }

#endif /* defined(_IX_RM_DEBUG_) */

    pMemoryManager = g_aMemoryManagersGroups[arg_MemType].m_pMembers + arg_MemChannel;

    err = pMemoryManager->m_pVMT->m_ReserveFuncPtr(pMemoryManager, arg_Size, arg_pByteOffset);

#ifdef _IX_RM_DEBUG_

    if(err != 0)
    {
        err = IX_RM_ERROR_CHAIN(err, s_aFunctionIdentifier, "memory reservation failed!");
    }

#endif /* _IX_RM_DEBUG_ */

    return err;
}

#else /* Westport hardware */

/**
 * NAME: ix_rm_mem_reserve
 *
 * DESCRIPTION: This function will reserve an area of memory at the specified 
 *	memory channel byte offset and type, and of specified size. This call 
 *	should be made before any memory allocation has been made in order to 
 *	reduce the chance of failure. If the memory channel byte offset is not 
 *	aligned according to the default alignment for the memory type, then 
 *	the allocated size will be larger to accommodate the alignment. In 
 *	general the byte offset and size should comply with the memory 
 *	alignment, otherwise, more memory will be reserved to insure the right 
 *	alignment. For the North bank memory, if the offset type is 
 *	IX_MEMORY_OFFSET_TYPE_ALIASED, then it is relative to the South alias 
 *	window, otherwise it is relative to the start of the memory channel in 
 *	the North bank.
 * 
 * @Param:  - IN arg_MemType - represents the type of memory to be reserved.
 * @Param:  - IN arg_MemChannel - represents the channel from which memory is 
 *	to be reserved.
 * @Param:  - IN arg_Size - represents the size in bytes of the memory block 
 *	to be reserved. 
 * @Param:  - IN arg_OffsetType - represents the type of the specified offset 
 * @Param:  - INOUT arg_pByteOffset - represents the address of byte offset 
 *	of the memory block requested for reservation. On return this value 
 *	will be adjusted to accommodate the memory type default alignment.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure. 
 *	If any portion of the memory requested for reservation has been already 
 *	allocated then the call will fail.
 */
ix_error ix_rm_mem_reserve( 
                            ix_memory_type arg_MemType, 
                            ix_uint32 arg_MemChannel, 
                            ix_uint32 arg_Size,
                            ix_memory_offset_type arg_OffsetType, 
                            ix_uint32* arg_pByteOffset 
                          )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_mem_reserve";

    ix_error err = 0;
    register _ix_mm_memory_manager* pMemoryManager;

    (void)s_aFunctionIdentifier;

    /* check the arg_MemType in DEBUG mode */
    IX_RM_ASSERT(arg_MemType <= IX_MEMORY_TYPE_SCRATCH, s_aFunctionIdentifier, "bad memory type!");
    /* check the arg_Size in DEBUG mode */
    IX_RM_ASSERT(arg_Size != 0, s_aFunctionIdentifier, "the request of 0 size memory is not acceptable!");

#if defined(_IX_RM_DEBUG_)

    switch(arg_MemType)
    {
        case IX_MEMORY_TYPE_DRAM_SOUTH:
            IX_RM_ASSERT(arg_MemChannel < IX_DRAM_SOUTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SRAM_MSG_SOUTH:
            IX_RM_ASSERT(arg_MemChannel < IX_SRAM_MSG_SOUTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_DRAM_NORTH:
            IX_RM_ASSERT(arg_MemChannel < IX_DRAM_NORTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SRAM_NORTH:
            IX_RM_ASSERT(arg_MemChannel < IX_SRAM_NORTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SCRATCH:
            IX_RM_ASSERT(arg_MemChannel < IX_SCRATCH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        default:
            break;
    }

#endif /* defined(_IX_RM_DEBUG_) */

    pMemoryManager = g_aMemoryManagersGroups[arg_MemType].m_pMembers + arg_MemChannel;

    err = pMemoryManager->m_pVMT->m_ReserveFuncPtr(pMemoryManager, arg_Size, arg_OffsetType, arg_pByteOffset);

#ifdef _IX_RM_DEBUG_

    if(err != 0)
    {
        err = IX_RM_ERROR_CHAIN(err, s_aFunctionIdentifier, "memory reservation failed!");
    }

#endif /* _IX_RM_DEBUG_ */

    return err;
}

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * NAME: ix_rm_mem_reserve_aligned
 *
 * DESCRIPTION: This function will reserve an area of memory at the specified memory channel byte 
 *          offset and of specified size. The function is provided as support for the microC compiler and
 *          linker. This call should be made before any memory allocation has been made in order
 *          to reduce the chance of failure. If the byte offset is not be aligned according to 
 *          arg_AlignmentShift or - if that is less than default alignment - to the default alignment
 *          and alignment type then internal alignment will be performed and more data might be 
 *          reserved. 
 *          If the alignment type is not IX_MEMORY_ALIGNMENT_TYPE_VIRTUAL then the resulting virtual
 *          address might not be aligned according to the arg_AlignmentShift but the physiscal offset
 *          or physical address will be properly aligned.
 *          In general the byte offset and size should comply with the memory alignment, 
 *          otherwise more memory will be reserved to insure the right alignment. On return *arg_pByteOffset might
 *          be modified to reflect alignment requirements.
 * 
 * @Param:  - IN arg_MemType - represents the type of memory to be reserved.
 * @Param:  - IN arg_MemChannel - represents the channel from which memory is to be reserved.
 * @Param:  - IN arg_Size - represents the size in bytes of the memory block to be reserved.  
 * @Param:  - IN arg_AlignmentType - specifies the type of alignment required as explained at 
 *          ix_memory_alignment_type type description
 * @Param:  - IN arg_AlignmentShift - this value specifies the required memory block alignment as the
 *          power of 2 of the alignment. If the specified alignment is less than the
 *          default alignment for the corresonding memory then the default alignment will be used.
 * @Param:  - INOUT arg_pByteOffset - represents the address of byte offset of the memory block 
 *          requested for reservation. On return this value will be adjusted to accomodate the
 *          the memory type default alignment.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure. If any portion of the memory requested
 *          for reservation has been already allocated then the call will fail.
 */
ix_error ix_rm_mem_reserve_aligned( 
                                    ix_memory_type arg_MemType, 
                                    ix_uint32 arg_MemChannel, 
                                    ix_uint32 arg_Size,
                                    ix_memory_alignment_type arg_AlignmentType,
                                    ix_uint32 arg_AlignmentShift,
                                    ix_uint32* arg_pByteOffset 
                                  )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_mem_reserve_aligned";

    ix_error err = 0;
    register _ix_mm_memory_manager* pMemoryManager;

    (void)s_aFunctionIdentifier;

    /* check the arg_MemType in DEBUG mode */
    IX_RM_ASSERT(arg_MemType <= IX_MEMORY_TYPE_SCRATCH, s_aFunctionIdentifier, "bad memory type!");
    /* check the arg_Size in DEBUG mode */
    IX_RM_ASSERT(arg_Size != 0, s_aFunctionIdentifier, "the request of 0 size memory is not acceptable!");

#if defined(_IX_RM_DEBUG_)

    switch(arg_MemType)
    {
        case IX_MEMORY_TYPE_DRAM:
            IX_RM_ASSERT(arg_MemChannel < IX_DRAM_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SRAM:
            IX_RM_ASSERT(arg_MemChannel < IX_SRAM_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        default:
            /* this is the SCRATCH memory */
            IX_RM_ASSERT(arg_MemChannel < IX_SCRATCH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
    }

#endif /* defined(_IX_RM_DEBUG_) */

    pMemoryManager = g_aMemoryManagersGroups[arg_MemType].m_pMembers + arg_MemChannel;

    err = pMemoryManager->m_pVMT->m_AlignedReserveFuncPtr(pMemoryManager, arg_Size, arg_AlignmentType, arg_AlignmentShift, arg_pByteOffset);

#ifdef _IX_RM_DEBUG_

    if(err != 0)
    {
        err = IX_RM_ERROR_CHAIN(err, s_aFunctionIdentifier, "memory reservation failed!");
    }

#endif /* _IX_RM_DEBUG_ */

    return err;
}

#else /* Westport hardware */

/**
 * NAME: ix_rm_mem_reserve_aligned
 *
 * DESCRIPTION: This function will reserve an area of memory at the specified 
 *	memory channel byte offset and type, and of specified size. This call 
 *	should be made before any memory allocation has been made in order to 
 *	reduce the chance of failure. If the byte offset is not aligned 
 *	according to arg_AlignmentShift or - if that is less than default 
 *	alignment - to the default alignment for alignment type, then internal 
 *	alignment will be performed and more data might be reserved. If the 
 *	alignment type is not IX_MEMORY_ALIGNMENT_TYPE_VIRTUAL then the 
 *	resulting virtual address might not be aligned according to the 
 *	arg_AlignmentShift but the physical offset or physical address will be 
 *	properly aligned. In general, the byte offset and size should comply 
 *	with the memory alignment, otherwise more memory will be reserved to 
 *	insure the right alignment. On return, *arg_pByteOffset might be 
 *	modified to reflect alignment requirements. For the North bank memory, 
 *	if the offset type is IX_MEMORY_OFFSET_TYPE_ALIASED, then it is 
 *	relative to the South alias window, otherwise it is relative to the 
 *	start of the memory channel in the North bank.
 * 
 * @Param:  - IN arg_MemType - represents the type of memory to be reserved.
 * @Param:  - IN arg_MemChannel - represents the channel from which memory is 
 *	to be reserved.
 * @Param:  - IN arg_Size - represents the size in bytes of the memory block 
 *	to be reserved.  
 * @Param:  - IN arg_AlignmentType - specifies the type of alignment required 
 *	as explained at ix_memory_alignment_type type description
 * @Param:  - IN arg_AlignmentShift - this value specifies the required 
 *	memory block alignment as the power of 2 of the alignment. If the 
 *	specified alignment is less than the default alignment for the 
 *	corresponding memory then the default alignment will be used.
 * @Param:  - IN arg_OffsetType - represents the type of the specified offset 
 * @Param:  - INOUT arg_pByteOffset - represents the address of byte offset 
 *	of the memory block requested for reservation. On return this value 
 *	will be adjusted to accommodate the memory type default alignment.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure. 
 *	If any portion of the memory requested for reservation has been already 
 *	allocated then the call will fail.
 */
ix_error ix_rm_mem_reserve_aligned( 
                                 ix_memory_type arg_MemType, 
                                 ix_uint32 arg_MemChannel, 
                                 ix_uint32 arg_Size,
                                 ix_memory_alignment_type arg_AlignmentType,
                                 ix_uint32 arg_AlignmentShift,
                                 ix_memory_offset_type arg_OffsetType, 
                                 ix_uint32* arg_pByteOffset 
                                  )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_mem_reserve_aligned";

    ix_error err = 0;
    register _ix_mm_memory_manager* pMemoryManager;

    (void)s_aFunctionIdentifier;

    /* check the arg_MemType in DEBUG mode */
    IX_RM_ASSERT(arg_MemType <= IX_MEMORY_TYPE_SCRATCH, s_aFunctionIdentifier, "bad memory type!");
    /* check the arg_Size in DEBUG mode */
    IX_RM_ASSERT(arg_Size != 0, s_aFunctionIdentifier, "the request of 0 size memory is not acceptable!");

#if defined(_IX_RM_DEBUG_)

    switch(arg_MemType)
    {
        case IX_MEMORY_TYPE_DRAM_SOUTH:
            IX_RM_ASSERT(arg_MemChannel < IX_DRAM_SOUTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SRAM_MSG_SOUTH:
            IX_RM_ASSERT(arg_MemChannel < IX_SRAM_MSG_SOUTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_DRAM_NORTH:
            IX_RM_ASSERT(arg_MemChannel < IX_DRAM_NORTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SRAM_NORTH:
            IX_RM_ASSERT(arg_MemChannel < IX_SRAM_NORTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SCRATCH:
            IX_RM_ASSERT(arg_MemChannel < IX_SCRATCH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        default:
            break;
    }

#endif /* defined(_IX_RM_DEBUG_) */

    pMemoryManager = g_aMemoryManagersGroups[arg_MemType].m_pMembers + arg_MemChannel;

    err = pMemoryManager->m_pVMT->m_AlignedReserveFuncPtr(pMemoryManager, arg_Size, arg_AlignmentType, 
        arg_AlignmentShift, arg_OffsetType, arg_pByteOffset);

#ifdef _IX_RM_DEBUG_

    if(err != 0)
    {
        err = IX_RM_ERROR_CHAIN(err, s_aFunctionIdentifier, "memory reservation failed!");
    }

#endif /* _IX_RM_DEBUG_ */

    return err;
}

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

/**
 * NAME: ix_rm_mem_free
 *
 * DESCRIPTION: This function will free memory previously allocated or reserved. Passing an 
 *          address that has not been allocated or that has already been freed will result in
 *          an error. Passing NULL will result in an no op. Accessing the memory location 
 *          after the block has been freed will result in unpredictable behaviour.
 * 
 * 
 * @Param:  - IN arg_pMemory - represents the address of the memory block to be freed.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure. Passing an address that has 
 *          not been allocated or that has already been freed will result in an error. 
 *          Passing NULL will result in an no op.
 */
ix_error ix_rm_mem_free( 
                         void* arg_pMemory 
                       )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_mem_free";

    /* Naming convention for automatic variables */
    ix_error err = 0;
    register _ix_mm_memory_manager* pMemoryManager;
    register ix_uint32   channelIndex;
    register ix_uint32   memType;

    (void)s_aFunctionIdentifier;

    /* for NULL we just return immediately */
    if(arg_pMemory == 0)
        return err;

    /* search each memory manager group */
    for(memType = IX_MEMORY_TYPE_FIRST; memType <= IX_MEMORY_TYPE_SCRATCH; memType++)
    {
        /* search each memory manager in the group */
        for(channelIndex = 0; channelIndex < g_aMemoryManagersGroups[memType].m_MemberCount; channelIndex++)
        {
            pMemoryManager = g_aMemoryManagersGroups[memType].m_pMembers + channelIndex;
            /* check if our address is in the address range of this memory manager */
            if(((ix_uint8*)arg_pMemory >= (ix_uint8*)pMemoryManager->m_pStartAddress) && 
                ((ix_uint8*)arg_pMemory < (ix_uint8*)pMemoryManager->m_pEndAddress))
            {
                /* call the free function for this memory manager */
                err = pMemoryManager->m_pVMT->m_FreeFuncPtr(pMemoryManager, arg_pMemory);

#ifdef _IX_RM_DEBUG_

                /* check the outcome of the free operation */
                if(err)
                {
                    err = IX_RM_ERROR_CHAIN(err, s_aFunctionIdentifier, 
                        "the memory address is in range but could not be freed!");
                }

#endif /* _IX_RM_DEBUG_ */
                
                return err;
            }
        }
    }

    /* the memory address is not found in any memory manager's range */
    return IX_RM_ERROR_NEW(IX_RM_ERROR_BAD_MEMORY_ADDRESS, s_aFunctionIdentifier, 
                        "the memory address could not be found!", err);
}



/**
 * NAME: ix_rm_mem_info
 *
 * DESCRIPTION: This function retrieves memory information for the specified memory type
 *          and specified channel. On return the structure pointed by arg_pMemoryInfo will be
 *          filled with the memory information.
 * 
 * @Param:  - IN arg_MemType - represents the type of memory for which we want to get the info.
 * @Param:  - IN arg_MemChannel - represents the channel for which we want to get the info.
 * @Param:  - OUT arg_pMemoryInfo - on return the structure pointed by arg_pMemoryInfo will be
 *          filled with the memory information 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_mem_info( 
                         ix_memory_type arg_MemType, 
                         ix_uint32 arg_MemChannel, 
                         ix_memory_info* arg_pMemoryInfo 
                       )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_mem_info";

    /* Naming convention for automatic variables */
    ix_error err = 0;
    register _ix_mm_memory_manager* pMemoryManager;
    
    (void)s_aFunctionIdentifier;

    /* check the arg_MemType in DEBUG mode */
    IX_RM_ASSERT(arg_MemType <= IX_MEMORY_TYPE_SCRATCH, s_aFunctionIdentifier, "bad memory type!");
    IX_RM_ASSERT(arg_pMemoryInfo, s_aFunctionIdentifier, "the ix_memory_info* parameter should be" 
        "non 0 and point to a valid structure!");

#if defined(_IX_RM_DEBUG_)

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

    switch(arg_MemType)
    {
        case IX_MEMORY_TYPE_DRAM:
            IX_RM_ASSERT(arg_MemChannel < IX_DRAM_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SRAM:
            IX_RM_ASSERT(arg_MemChannel < IX_SRAM_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        default:
            /* this is the SCRATCH memory */
            IX_RM_ASSERT(arg_MemChannel < IX_SCRATCH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
    }

#else /* Westport hardware */

    switch(arg_MemType)
    {
        case IX_MEMORY_TYPE_DRAM_SOUTH:
            IX_RM_ASSERT(arg_MemChannel < IX_DRAM_SOUTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SRAM_MSG_SOUTH:
            IX_RM_ASSERT(arg_MemChannel < IX_SRAM_MSG_SOUTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_DRAM_NORTH:
            IX_RM_ASSERT(arg_MemChannel < IX_DRAM_NORTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SRAM_NORTH:
            IX_RM_ASSERT(arg_MemChannel < IX_SRAM_NORTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SCRATCH:
            IX_RM_ASSERT(arg_MemChannel < IX_SCRATCH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        default:
            break;
    }

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

#endif /* defined(_IX_RM_DEBUG_) */

    pMemoryManager = g_aMemoryManagersGroups[arg_MemType].m_pMembers + arg_MemChannel;

    arg_pMemoryInfo->m_pStartAddress = pMemoryManager->m_pStartAddress;
    arg_pMemoryInfo->m_TotalSize = pMemoryManager->m_MemoryBlockSize;
    arg_pMemoryInfo->m_DefaultAlignmentShift = pMemoryManager->m_DefaultAlignmentShift;
    arg_pMemoryInfo->m_ChannelPhysicalOffset = pMemoryManager->m_ChannelPhysicalOffset;
    arg_pMemoryInfo->m_pChannelPhysicalAddress = pMemoryManager->m_pChannelPhysicalAddress;

#if (_IX_HARDWARE_TYPE_ == _IX_HW_WESTPORT_) 

    arg_pMemoryInfo->m_MemoryBankType = pMemoryManager->m_MemoryBankType;
    arg_pMemoryInfo->m_NonAliasChannelPhysicalOffset = pMemoryManager->m_NonAliasChannelPhysicalOffset;
    arg_pMemoryInfo->m_pNonAliasChannelPhysicalAddress = pMemoryManager->m_pNonAliasChannelPhysicalAddress;

#endif /* (_IX_HARDWARE_TYPE_ == _IX_HW_WESTPORT_) */

    arg_pMemoryInfo->m_FreeSize = 0;

#if defined(_IX_RM_MULTI_THREADED_)

    {

        ix_synchronization_adapter* pSyncAdapter = &pMemoryManager->m_SyncAdapter;

        /* lock the memory manager data */
        pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

        if(_ix_rm_memory_block_cell_action(pMemoryManager->m_pBlockList, 0, 
            _ix_rm_memory_get_free_memory, &arg_pMemoryInfo->m_FreeSize) != 0)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_UNKNOWN, s_aFunctionIdentifier, 
                            "could not calculate free memory!", err);
        }

#if defined(_IX_RM_MULTI_THREADED_)

        /* unlock the memory manager data */
        pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);
    }

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    return err;
}

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * NAME: ix_rm_mem_local_alloc
 *
 * DESCRIPTION: This function will allocate a block of local memory for the specified ME.  
 *          The location referred by arg_pByteOffset will be set to (ix_uint32)(-1) if 
 *          failure, or to the allocated memory block offset if successful. 
 *          The local memory will not be accessible to the Core applications.
 *          The apps will get the offset and patch it in the uCode for the specific ME. The 
 *          returned offset will be 4 bytes aligned. If the size does not comply with memory alignment
 *          then the actual allocated size will be rounded up to 4 byte boundary.
 * 
 * @Param:  - IN arg_MENumber - represents the ME number for which local memory should be 
 *          allocated.
 * @Param:  - IN arg_Size - represents the size in bytes of the block requested.
 * @Param:  - OUT arg_pByteOffset - represents the address where the allocated memory offset in bytes
 *          should be stored.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure. The location referred by 
 *          arg_pByteOffset will be set to (ix_uint32)(-1) if failure, or to the allocated memory block
 *          offset if successful.
 */
ix_error ix_rm_mem_local_alloc( 
                                ix_uint32 arg_MENumber, 
                                ix_uint32 arg_Size, 
                                ix_uint32* arg_pByteOffset 
                              )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_mem_local_alloc";

    /* Naming convention for automatic variables */
    ix_error err = 0;
    ix_uint32* a_pMemoryAddress;
    ix_uint32 i;
    register _ix_mm_memory_manager* pMemoryManager;

    (void)s_aFunctionIdentifier;

    /* check the arg_MENumber in DEBUG mode */
#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

    IX_RM_ASSERT(((arg_MENumber <= 0x03) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x13))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x03 and 0x10 - 0x13!");

#elif (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

    IX_RM_ASSERT(((arg_MENumber <= 0x07) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x17))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x07 and 0x10 - 0x17!");

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */
    
    /* check the arg_Size in DEBUG mode */
    IX_RM_ASSERT(arg_Size != 0, s_aFunctionIdentifier, "the request of 0 size memory is not acceptable!");

    /* get the coresponding internal ME index */
    for(i = 0; i < IX_MAX_MICROENGINE_NUMBER; i++)
        if(arg_MENumber == g_aMicroengineNumberMapping[i])
            arg_MENumber = i;

    pMemoryManager = g_aMemoryManagersGroups[IX_MEMORY_TYPE_LOCAL].m_pMembers + arg_MENumber;

    {
        err = pMemoryManager->m_pVMT->m_AllocFuncPtr(pMemoryManager, arg_Size, (void**)&a_pMemoryAddress);
        if(err)
        {
            err = IX_RM_ERROR_CHAIN(err, s_aFunctionIdentifier, "local memory allocation failed!");
            /* set the offset to (ix_uint32)(-1) to indicate error */
            *arg_pByteOffset = (ix_uint32)(-1);
        }
        else
        {
            /* calculate the offset */
            *arg_pByteOffset = (ix_uint32)((ix_uint8*)a_pMemoryAddress - (ix_uint8*)pMemoryManager->m_pStartAddress);
        }
    }

    return err;
}



/**
 * NAME: ix_rm_mem_local_reserve
 *
 * DESCRIPTION: This function will reserve an area of memory at the specified byte offset and
 *          of specified size. The function is provided as support for the microC compiler and
 *          linker. This call should be made before any memory allocation has been made in order
 *          to reduce the chance of failure. If the byte offset will not be 4 byte aligned, then 
 *          internal alignment will be performed and more data might be reserved. In general 
 *          the byte offset and size should comply with the memory alignment, otherwise more 
 *          memory will be reserved to insure the right alignment. 
 * 
 * 
 * @Param:  - IN arg_MENumber - represents the ME number for which local memory should be 
 *          reserved.
 * @Param:  - IN arg_Size - represents the size in bytes of the memory block to be reserved.  
 * @Param:  - INOUT arg_pByteOffset - represents the address of byte offset of the memory block 
 *          requested for reservation. On return this value will be adjusted to reflect the
 *          default alignment for the memory block.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure. If any portion of the memory requested
 *          for reservation has been already allocated then the call will fail.
 */
ix_error ix_rm_mem_local_reserve( 
                                  ix_uint32 arg_MENumber, 
                                  ix_uint32 arg_Size, 
                                  ix_uint32* arg_pByteOffset 
                                )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_mem_local_reserve";

    /* Naming convention for automatic variables */
    ix_error err = 0;
    ix_uint32 i;
    register _ix_mm_memory_manager* pMemoryManager;

    (void)s_aFunctionIdentifier;

    /* check the arg_MENumber in DEBUG mode */
#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

    IX_RM_ASSERT(((arg_MENumber <= 0x03) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x13))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x03 and 0x10 - 0x13!");

#elif (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

    IX_RM_ASSERT(((arg_MENumber <= 0x07) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x17))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x07 and 0x10 - 0x17!");

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */
    
    /* check the arg_Size in DEBUG mode */
    IX_RM_ASSERT(arg_Size != 0, s_aFunctionIdentifier, "the request of 0 size memory is not acceptable!");

    /* get the coresponding internal ME index */
    for(i = 0; i < IX_MAX_MICROENGINE_NUMBER; i++)
        if(arg_MENumber == g_aMicroengineNumberMapping[i])
            arg_MENumber = i;

    pMemoryManager = g_aMemoryManagersGroups[IX_MEMORY_TYPE_LOCAL].m_pMembers + arg_MENumber;

    err = pMemoryManager->m_pVMT->m_ReserveFuncPtr(pMemoryManager, arg_Size, arg_pByteOffset);

    if(err != 0)
    {
        err = IX_RM_ERROR_CHAIN(err, s_aFunctionIdentifier, "local memory reservation failed!");
    }

    return err;
}



/**
 * NAME: ix_rm_mem_local_free
 *
 * DESCRIPTION: This function will free memory previously allocated or reserved. Passing an 
 *          address that has not been allocated or that has already been freed will result in
 *          an error. Passing NULL will result in an no op. Accessing the memory location 
 *          after the block has been freed will result in unpredictable behaviour.   
 * 
 * 
 * @Param:  - IN arg_MENumber - represents the ME number for which local memory should be 
 *          freed.
 * @Param:  - IN arg_MemoryOffset - represents the memory offset in bytes of the memory 
 *          block to be freed.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure. Passing an address that has 
 *          not been allocated or that has already been freed will result in an error. 
 *          Passing NULL will result in an no op.
 */
ix_error ix_rm_mem_local_free( 
                               ix_uint32 arg_MENumber, 
                               ix_uint32 arg_MemoryOffset 
                             )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_mem_local_free";

    ix_error err = IX_SUCCESS;

    ix_uint32 i;

    register ix_uint32* a_pMemoryAddress;
    register _ix_mm_memory_manager* pMemoryManager;

    (void)s_aFunctionIdentifier;
    (void)err;

    /* check the arg_MENumber in DEBUG mode */
#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

    IX_RM_ASSERT(((arg_MENumber <= 0x03) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x13))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x03 and 0x10 - 0x13!");

#elif (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

    IX_RM_ASSERT(((arg_MENumber <= 0x07) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x17))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x07 and 0x10 - 0x17!");

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

    /* get the coresponding internal ME index */
    for(i = 0; i < IX_MAX_MICROENGINE_NUMBER; i++)
        if(arg_MENumber == g_aMicroengineNumberMapping[i])
            arg_MENumber = i;

    pMemoryManager = g_aMemoryManagersGroups[IX_MEMORY_TYPE_LOCAL].m_pMembers + arg_MENumber;

    a_pMemoryAddress = (ix_uint32*)((ix_uint8*)pMemoryManager->m_pStartAddress + arg_MemoryOffset - pMemoryManager->m_ChannelPhysicalOffset);
    if((a_pMemoryAddress < pMemoryManager->m_pEndAddress) && (a_pMemoryAddress >=pMemoryManager->m_pStartAddress))
    {
        return pMemoryManager->m_pVMT->m_FreeFuncPtr(pMemoryManager, a_pMemoryAddress);
    }
    else
    {
        /* the memory address is not found in any memory manager's range */
        return IX_RM_ERROR_NEW(IX_RM_ERROR_BAD_MEMORY_ADDRESS, s_aFunctionIdentifier, 
                        "the memory address could not be found!", err);
    }
}


/**
 * NAME: ix_rm_mem_local_info
 *
 * DESCRIPTION: This function retrieves memory information for the specified memory type
 *          and specified channel. On return the structure pointed by arg_pMemoryInfo will be
 *          filled with the memory information.
 * 
 * @Param:  - IN arg_MENumber - represents the microegine number for which we want to get the info.
 * @Param:  - OUT arg_pMemoryInfo - on return the structure pointed by arg_pMemoryInfo will be
 *          filled with the memory information 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_mem_local_info( 
                               ix_memory_type arg_MENumber, 
                               ix_memory_info* arg_pMemoryInfo 
                             )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_mem_local_info";

    /* Naming convention for automatic variables */
    ix_error err = 0;
    ix_uint32 i;

    register _ix_mm_memory_manager* pMemoryManager;

    (void)s_aFunctionIdentifier;
    
    /* check the arg_MENumber in DEBUG mode */
#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

    IX_RM_ASSERT(((arg_MENumber <= 0x03) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x13))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x03 and 0x10 - 0x13!");

#elif (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

    IX_RM_ASSERT(((arg_MENumber <= 0x07) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x17))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x07 and 0x10 - 0x17!");

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */
    
    IX_RM_ASSERT(arg_pMemoryInfo, s_aFunctionIdentifier, "the ix_memory_info* parameter should be" 
        "non 0 and point to a valid structure!");

    /* get the coresponding internal ME index */
    for(i = 0; i < IX_MAX_MICROENGINE_NUMBER; i++)
        if(arg_MENumber == g_aMicroengineNumberMapping[i])
            arg_MENumber = i;

    pMemoryManager = g_aMemoryManagersGroups[IX_MEMORY_TYPE_LOCAL].m_pMembers + arg_MENumber;

    arg_pMemoryInfo->m_pStartAddress = pMemoryManager->m_pStartAddress;
    arg_pMemoryInfo->m_TotalSize = pMemoryManager->m_MemoryBlockSize;
    arg_pMemoryInfo->m_DefaultAlignmentShift = pMemoryManager->m_DefaultAlignmentShift;
    arg_pMemoryInfo->m_ChannelPhysicalOffset = pMemoryManager->m_ChannelPhysicalOffset;
    arg_pMemoryInfo->m_pChannelPhysicalAddress = pMemoryManager->m_pChannelPhysicalAddress;
    arg_pMemoryInfo->m_FreeSize = 0;

#if defined(_IX_RM_MULTI_THREADED_)
    {
        ix_synchronization_adapter* pSyncAdapter = &pMemoryManager->m_SyncAdapter;

        /* lock the memory manager data */
        pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

        if(_ix_rm_memory_block_cell_action(pMemoryManager->m_pBlockList, 0, 
            _ix_rm_memory_get_free_memory, &arg_pMemoryInfo->m_FreeSize) != 0)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_UNKNOWN, s_aFunctionIdentifier, 
                            "could not calculate free memory!", err);
        }

#if defined(_IX_RM_MULTI_THREADED_)

        /* unlock the memory manager data */
        pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);
    }

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    return err;
}

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * NAME: ix_rm_get_phys_offset
 *
 * DESCRIPTION: Get the physical offset into the channel memory in bytes  and the physical
 *          address for the passed virtual memory address. The function also returns the 
 *          type of memory and the channel corresponding to the input address. If any of the
 *          OUT parameters are NULL then that piece of data will not be returned.
 * 
 * @Param:  - IN arg_pMemory - represents the adress of virtual memory for which we want to get
 *          the physical offset.
 * @Param:  - OUT arg_pMemType - represents the location where to store the corresponding memory
 *          type.
 * @Param:  - OUT arg_pMemChannel - represents the location where to store the corresponding
 *          memory channel.
 * @Param:  - OUT arg_pOffset - represents the address where to store the physical offset into the channel.
 * @Param:  - OUT arg_pPhysicalAddress - represents the address where to store the physical address.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_get_phys_offset( 
                                const void* arg_pMemory, 
                                ix_memory_type* arg_pMemType, 
                                ix_uint32* arg_pMemChannel, 
                                ix_uint32* arg_pOffset,
                                void** arg_pPhysicalAddress
                              )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_get_phys_offset";

    /* Naming convention for automatic variables */
    ix_error err = 0;

    register _ix_mm_memory_manager* pMemoryManager;
    register ix_uint32   channelIndex;
    register ix_uint32   memType;

    (void)s_aFunctionIdentifier;

    /* for NULL we just return immediately */
    if(arg_pMemory == 0)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_NULL_MEMORY_ADDRESS, s_aFunctionIdentifier, 
                        "the passed memory address is NULL!", err);

    /* search each memory manager group */
    for(memType = IX_MEMORY_TYPE_DRAM; memType <= IX_MEMORY_TYPE_SCRATCH; memType++)
    {
        /* search each memory manager in the group */
        for(channelIndex = 0; channelIndex < g_aMemoryManagersGroups[memType].m_MemberCount; channelIndex++)
        {
            pMemoryManager = g_aMemoryManagersGroups[memType].m_pMembers + channelIndex;

            /* check if our address is in the address range of this memory manager */
            if(((ix_uint8*)arg_pMemory >= (ix_uint8*)pMemoryManager->m_pStartAddress) && 
                ((ix_uint8*)arg_pMemory < (ix_uint8*)pMemoryManager->m_pEndAddress))
            {
                if(arg_pOffset != 0)
                {
                    *arg_pOffset = (ix_uint32)((ix_uint8*)arg_pMemory - (ix_uint8*)pMemoryManager->m_pStartAddress) 
                        + pMemoryManager->m_ChannelPhysicalOffset;
                }

                if(arg_pPhysicalAddress != 0)
                {
                    *arg_pPhysicalAddress = (ix_uint8*)pMemoryManager->m_pChannelPhysicalAddress + 
                        (ix_uint32)((ix_uint8*)arg_pMemory - (ix_uint8*)pMemoryManager->m_pStartAddress) 
                        + pMemoryManager->m_ChannelPhysicalOffset;
                }

                if(arg_pMemType != 0)
                    *arg_pMemType = memType;

                if(arg_pMemChannel != 0)
                    *arg_pMemChannel = channelIndex;

                return err;
            }
        }
    }
    /* the memory address is not found in any memory manager's range */
    return IX_RM_ERROR_NEW(IX_RM_ERROR_BAD_MEMORY_ADDRESS, s_aFunctionIdentifier, 
                        "the memory address could not be found!", err);
}

#else /* Westport hardware */

/**
 * NAME: ix_rm_get_phys_offset
 *
 * DESCRIPTION: Get the physical offset into the channel memory in bytes and
 *	the physical address for the passed virtual memory address. The 
 *	function also returns the type of memory and the channel corresponding 
 *	to the input address. If any of the OUT parameters are NULL then that 
 *	piece of data will not be returned. For the North bank memory, 
 *	if the passed offset type is IX_MEMORY_OFFSET_TYPE_ALIASED, then the 
 *	returned offset is relative to the South alias window and the physical 
 *	address is relative to the South bank, otherwise the offset is relative 
 *	to the start of the memory channel in the North bank and the physical 
 *	address is relative to the North bank.
 * 
 * @Param:  - IN arg_pMemory - represents the adress of virtual memory for 
 *	which we want to get the physical offset.
 * @Param:  - IN arg_OffsetType - represents the type of the offset and 
 *	physical address to be returned 
 * @Param:  - OUT arg_pMemType - represents the location where to store the
 *	corresponding memory type.
 * @Param:  - OUT arg_pMemChannel - represents the location where to store 
 *	the corresponding memory channel.
 * @Param:  - OUT arg_pOffset - represents the address where to store the
 *	physical offset into the channel.
 * @Param:  - OUT arg_pPhysicalAddress - represents the address where to 
 *	store the physical address.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_get_phys_offset( 
                                const void* arg_pMemory,
                                ix_memory_offset_type arg_OffsetType, 
                                ix_memory_type* arg_pMemType, 
                                ix_uint32* arg_pMemChannel, 
                                ix_uint32* arg_pOffset,
                                void** arg_pPhysicalAddress
                              )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_get_phys_offset";

    /* Naming convention for automatic variables */
    ix_error err = 0;

    register _ix_mm_memory_manager* pMemoryManager;
    register ix_uint32   channelIndex;
    register ix_uint32   memType;

    (void)s_aFunctionIdentifier;

    /* for NULL we just return immediately */
    if(arg_pMemory == 0)
        return IX_RM_ERROR_NEW(IX_RM_ERROR_NULL_MEMORY_ADDRESS, s_aFunctionIdentifier, 
                        "the passed memory address is NULL!", err);

    /* search each memory manager group */
    for(memType = IX_MEMORY_TYPE_FIRST; memType < IX_MEMORY_TYPE_LAST; memType++)
    {
        /* search each memory manager in the group */
        for(channelIndex = 0; channelIndex < g_aMemoryManagersGroups[memType].m_MemberCount; channelIndex++)
        {
            pMemoryManager = g_aMemoryManagersGroups[memType].m_pMembers + channelIndex;

            /* check if our address is in the address range of this memory manager */
            if(((ix_uint8*)arg_pMemory >= (ix_uint8*)pMemoryManager->m_pStartAddress) && 
                ((ix_uint8*)arg_pMemory < (ix_uint8*)pMemoryManager->m_pEndAddress))
            {
                if(arg_OffsetType == IX_MEMORY_OFFSET_TYPE_ALIASED)
                {
                    if(arg_pOffset != 0)
                    {
                        *arg_pOffset = (ix_uint32)((ix_uint8*)arg_pMemory - (ix_uint8*)pMemoryManager->m_pStartAddress) 
                            + pMemoryManager->m_ChannelPhysicalOffset;
                    }

                    if(arg_pPhysicalAddress != 0)
                    {
                        *arg_pPhysicalAddress = (ix_uint8*)pMemoryManager->m_pChannelPhysicalAddress + 
                            (ix_uint32)((ix_uint8*)arg_pMemory - (ix_uint8*)pMemoryManager->m_pStartAddress) 
                            + pMemoryManager->m_ChannelPhysicalOffset;
                    }
                }
                else
                {
                    if(arg_pOffset != 0)
                    {
                        *arg_pOffset = (ix_uint32)((ix_uint8*)arg_pMemory - (ix_uint8*)pMemoryManager->m_pStartAddress) 
                            + pMemoryManager->m_NonAliasChannelPhysicalOffset;
                    }

                    if(arg_pPhysicalAddress != 0)
                    {
                        *arg_pPhysicalAddress = (ix_uint8*)pMemoryManager->m_pChannelPhysicalAddress + 
                            (ix_uint32)((ix_uint8*)arg_pMemory - (ix_uint8*)pMemoryManager->m_pStartAddress) 
                            + pMemoryManager->m_NonAliasChannelPhysicalOffset;
                    }
                }

                if(arg_pMemType != 0)
                    *arg_pMemType = memType;

                if(arg_pMemChannel != 0)
                    *arg_pMemChannel = channelIndex;

                return err;
            }
        }
    }
    /* the memory address is not found in any memory manager's range */
    return IX_RM_ERROR_NEW(IX_RM_ERROR_BAD_MEMORY_ADDRESS, s_aFunctionIdentifier, 
                        "the memory address could not be found!", err);
}

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */


#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * NAME: ix_rm_get_virtual_address
 *
 * DESCRIPTION: The function returns the virtual pointer into memory given the physical offset of the
 *          memory buffer expressed in bytes.
 * 
 * @Param:  - IN arg_MemType - represents the type of memory in question.
 * @Param:  - IN arg_MemChannel - represents the channel for which we want the virtual address.
 * @Param:  - IN arg_Offset - represents the physical offset
 * @Param:  - OUT arg_pMemoryAddr - represents the address where the virtual address should be stored. 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_get_virtual_address( 
                                    ix_memory_type arg_MemType, 
                                    ix_uint32 arg_MemChannel, 
                                    ix_uint32 arg_Offset, 
                                    ix_uint8** arg_pMemoryAddr 
                                  )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_get_virtual_address";

    /* Naming convention for automatic variables */
    ix_error err = 0;
    register _ix_mm_memory_manager* pMemoryManager;

    (void)s_aFunctionIdentifier;

    /* check the arg_MemType in DEBUG mode */
    IX_RM_ASSERT(arg_MemType <= IX_MEMORY_TYPE_SCRATCH, s_aFunctionIdentifier, "bad memory type!");

#if defined(_IX_RM_DEBUG_)

    switch(arg_MemType)
    {
        case IX_MEMORY_TYPE_DRAM:
            IX_RM_ASSERT(arg_MemChannel < IX_DRAM_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SRAM:
            IX_RM_ASSERT(arg_MemChannel < IX_SRAM_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        default:
            /* this is the SCRATCH memory */
            IX_RM_ASSERT(arg_MemChannel < IX_SCRATCH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
    }

#endif /* defined(_IX_RM_DEBUG_) */

    pMemoryManager = g_aMemoryManagersGroups[arg_MemType].m_pMembers + arg_MemChannel;

    if((arg_Offset < pMemoryManager->m_ChannelPhysicalOffset) || 
        ((arg_Offset - pMemoryManager->m_ChannelPhysicalOffset) >= pMemoryManager->m_MemoryBlockSize))
        return IX_RM_ERROR_NEW(IX_RM_ERROR_MEMORY_BAD_OFFSET, s_aFunctionIdentifier, 
                        "the passed offset is invalid!", err);

    /* calculate the virtual memory pointer */
    *arg_pMemoryAddr = ((ix_uint8*)pMemoryManager->m_pStartAddress) + arg_Offset 
        - pMemoryManager->m_ChannelPhysicalOffset;

    return err;
}

#else /* Westport hardware */

/**
 * NAME: ix_rm_get_virtual_address
 *
 * DESCRIPTION: The function returns the virtual pointer into memory given 
 *	the physical offset of the memory buffer expressed in bytes. 
 *	For the North bank memory, if the passed offset type is 
 *	IX_MEMORY_OFFSET_TYPE_ALIASED, then the offset is considered to be 
 *	relative to the South alias window, otherwise the offset is considered 
 *	to be relative to the start of the memory channel in the North bank.
 * 
 * @Param:  - IN arg_MemType - represents the type of memory in question.
 * @Param:  - IN arg_MemChannel - represents the channel for which we want 
 *	the virtual address.
 * @Param:  - IN arg_Offset - represents the physical offset in bytes
 * @Param:  - IN arg_OffsetType - represents the type of the specified offset 
 * @Param:  - OUT arg_pMemoryAddr - represents the address where the virtual 
 *	address should be stored. 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_get_virtual_address( 
                                    ix_memory_type arg_MemType, 
                                    ix_uint32 arg_MemChannel, 
                                    ix_uint32 arg_Offset, 
                                    ix_memory_offset_type arg_OffsetType, 
                                    ix_uint8** arg_pMemoryAddr 
                                  )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_get_virtual_address";

    ix_error err = 0;
    register _ix_mm_memory_manager* pMemoryManager;

    (void)s_aFunctionIdentifier;

    /* check the arg_MemType in DEBUG mode */
    IX_RM_ASSERT(arg_MemType < IX_MEMORY_TYPE_LAST, s_aFunctionIdentifier, "bad memory type!");

#if defined(_IX_RM_DEBUG_)

    switch(arg_MemType)
    {
        case IX_MEMORY_TYPE_DRAM_SOUTH:
            IX_RM_ASSERT(arg_MemChannel < IX_DRAM_SOUTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SRAM_MSG_SOUTH:
            IX_RM_ASSERT(arg_MemChannel < IX_SRAM_MSG_SOUTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_DRAM_NORTH:
            IX_RM_ASSERT(arg_MemChannel < IX_DRAM_NORTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SRAM_NORTH:
            IX_RM_ASSERT(arg_MemChannel < IX_SRAM_NORTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        case IX_MEMORY_TYPE_SCRATCH:
            IX_RM_ASSERT(arg_MemChannel < IX_SCRATCH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            break;
        default:
            break;
    }

#endif /* defined(_IX_RM_DEBUG_) */

    pMemoryManager = g_aMemoryManagersGroups[arg_MemType].m_pMembers + arg_MemChannel;

    /* calculate the virtual memory pointer */
    if(arg_OffsetType == IX_MEMORY_OFFSET_TYPE_ALIASED)
    {
        if((arg_Offset < pMemoryManager->m_ChannelPhysicalOffset) || 
            ((arg_Offset - pMemoryManager->m_ChannelPhysicalOffset) >= pMemoryManager->m_MemoryBlockSize))
            return IX_RM_ERROR_NEW(IX_RM_ERROR_MEMORY_BAD_OFFSET, s_aFunctionIdentifier, 
                            "the passed offset is invalid!", err);

        *arg_pMemoryAddr = ((ix_uint8*)pMemoryManager->m_pStartAddress) + arg_Offset 
            - pMemoryManager->m_ChannelPhysicalOffset;
    }
    else
    {
        if((arg_Offset < pMemoryManager->m_NonAliasChannelPhysicalOffset) || 
            ((arg_Offset - pMemoryManager->m_NonAliasChannelPhysicalOffset) >= pMemoryManager->m_MemoryBlockSize))
            return IX_RM_ERROR_NEW(IX_RM_ERROR_MEMORY_BAD_OFFSET, s_aFunctionIdentifier, 
                            "the passed offset is invalid!", err); 
        *arg_pMemoryAddr = ((ix_uint8*)pMemoryManager->m_pStartAddress) + arg_Offset 
            - pMemoryManager->m_NonAliasChannelPhysicalOffset;
    }

    return err;
}

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */


#if defined(_IX_RM_DEBUG_)


/**
 * NAME: ix_rm_mem_status_print
 *
 * DESCRIPTION: This function will a snapshot of a common memory block based on the type and
 *          channel. All allocated and free memory cells will be listed.
 * 
 * @Param:  - IN arg_MemType - represents the type of memory in question.
 * @Param:  - IN arg_MemChannel - represents the channel for which we want to print the status information.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

ix_error ix_rm_mem_status_print(
                                 ix_memory_type arg_MemType, 
                                 ix_uint32 arg_MemChannel 
                               )

{
    static const char s_aFunctionIdentifier[] = "ix_rm_mem_status_print";

    ix_error err = 0;
    register _ix_mm_memory_manager* pMemoryManager;
    ix_uint32 counter = 0;

    /* check the arg_MemType in DEBUG mode */
    IX_RM_ASSERT(arg_MemType <= IX_MEMORY_TYPE_LOCAL, s_aFunctionIdentifier, "bad memory type!");

    ix_ossl_message_log("Memory block\n");
    ix_ossl_message_log("\tType: ");
    switch(arg_MemType)
    {
        case IX_MEMORY_TYPE_DRAM:
            IX_RM_ASSERT(arg_MemChannel < IX_DRAM_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            ix_ossl_message_log("DRAM");
            break;
        case IX_MEMORY_TYPE_SRAM:
            IX_RM_ASSERT(arg_MemChannel < IX_SRAM_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            ix_ossl_message_log("SRAM");
            break;
        case IX_MEMORY_TYPE_SCRATCH:
            IX_RM_ASSERT(arg_MemChannel < IX_SCRATCH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            ix_ossl_message_log("SCRATCH");
            break;
        default:
            /* local memory */
#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

            IX_RM_ASSERT(((arg_MemChannel <= 0x03) || ((arg_MemChannel >= 0x10) && (arg_MemChannel <= 0x13))), 
                s_aFunctionIdentifier, "bad microengine number: 0x00-0x03 and 0x10 - 0x13!");

#elif (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

            IX_RM_ASSERT(((arg_MemChannel <= 0x07) || ((arg_MemChannel >= 0x10) && (arg_MemChannel <= 0x17))), 
                s_aFunctionIdentifier, "bad microengine number: 0x00-0x07 and 0x10 - 0x17!");

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

            {
                ix_uint32 i;

                /* get the coresponding internal ME index */
                for(i = 0; i < IX_MAX_MICROENGINE_NUMBER; i++)
                    if(arg_MemChannel == g_aMicroengineNumberMapping[i])
                        arg_MemChannel = i;
            }
    
            ix_ossl_message_log("LOCAL");
            break;
    }
    ix_ossl_message_log("\n");

    pMemoryManager = g_aMemoryManagersGroups[arg_MemType].m_pMembers + arg_MemChannel;
    ix_ossl_message_log("\tChannel: %lu\n", pMemoryManager->m_MemoryChannel);
    ix_ossl_message_log("\tBlock size: 0x%08lx\n", pMemoryManager->m_MemoryBlockSize);
    ix_ossl_message_log("\tDefault alignment shift: %lu\n", pMemoryManager->m_DefaultAlignmentShift);
    ix_ossl_message_log("\tStart address: 0x%p\n", pMemoryManager->m_pStartAddress);
    ix_ossl_message_log("\tEnd address: 0x%p\n", pMemoryManager->m_pEndAddress);
    ix_ossl_message_log("\tPhysical offset: 0x%08lx\n", pMemoryManager->m_ChannelPhysicalOffset);
    ix_ossl_message_log("\tChannel physical address: 0x%p\n", pMemoryManager->m_pChannelPhysicalAddress);

    _ix_rm_memory_block_cell_action(pMemoryManager->m_pBlockList, 0, _ix_rm_memory_block_cell_print, &counter);

    ix_ossl_message_log("\nTotal cells in the memory block: %lu\n\n", counter);

    return err;
}

#else /* Westport hardware */

ix_error ix_rm_mem_status_print(
                                 ix_memory_type arg_MemType, 
                                 ix_uint32 arg_MemChannel 
                               )

{
    static const char s_aFunctionIdentifier[] = "ix_rm_mem_status_print";

    ix_error err = 0;
    register _ix_mm_memory_manager* pMemoryManager;
    ix_uint32 counter = 0;

    /* check the arg_MemType in DEBUG mode */
    IX_RM_ASSERT(arg_MemType < IX_MEMORY_TYPE_LAST, s_aFunctionIdentifier, "bad memory type!");

    ix_ossl_message_log("Memory block\n");
    ix_ossl_message_log("\tType: ");
    switch(arg_MemType)
    {
        case IX_MEMORY_TYPE_DRAM_SOUTH:
            IX_RM_ASSERT(arg_MemChannel <= IX_DRAM_SOUTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            ix_ossl_message_log("South bank DRAM");
            break;
        case IX_MEMORY_TYPE_SRAM_MSG_SOUTH:
            IX_RM_ASSERT(arg_MemChannel <= IX_SRAM_MSG_SOUTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            ix_ossl_message_log("South bank MSG-SRAM");
            break;
        case IX_MEMORY_TYPE_DRAM_NORTH:
            IX_RM_ASSERT(arg_MemChannel <= IX_DRAM_NORTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            ix_ossl_message_log("North bank aliased DRAM");
            break;
        case IX_MEMORY_TYPE_SRAM_NORTH:
            IX_RM_ASSERT(arg_MemChannel <= IX_SRAM_NORTH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            ix_ossl_message_log("North bank aliased SRAM");
            break;
        case IX_MEMORY_TYPE_SCRATCH:
            IX_RM_ASSERT(arg_MemChannel <= IX_SCRATCH_CHANNELS_NUMBER, s_aFunctionIdentifier, "bad channel number!");
            ix_ossl_message_log("North bank aliased SCRATCH");
            break;
        default:
            ix_ossl_message_log("Unknown");
            break;
    }
    ix_ossl_message_log("\n");

    pMemoryManager = g_aMemoryManagersGroups[arg_MemType].m_pMembers + arg_MemChannel;
    ix_ossl_message_log("\tBank: %s\n", (pMemoryManager->m_MemoryBankType == IX_MEMORY_BANK_TYPE_SOUTH) ? "SOUTH" : "NORTH");
    ix_ossl_message_log("\tChannel: %lu\n", pMemoryManager->m_MemoryChannel);
    ix_ossl_message_log("\tBlock size: 0x%08lx\n", pMemoryManager->m_MemoryBlockSize);
    ix_ossl_message_log("\tDefault alignment shift: %lu\n", pMemoryManager->m_DefaultAlignmentShift);
    ix_ossl_message_log("\tStart address: 0x%p\n", pMemoryManager->m_pStartAddress);
    ix_ossl_message_log("\tEnd address: 0x%p\n", pMemoryManager->m_pEndAddress);
    ix_ossl_message_log("\tAliased physical offset: 0x%08lx\n", pMemoryManager->m_ChannelPhysicalOffset);
    ix_ossl_message_log("\tAliased physical address: 0x%p\n", pMemoryManager->m_pChannelPhysicalAddress);
    ix_ossl_message_log("\tNon aliased physical offset: 0x%08lx\n", pMemoryManager->m_NonAliasChannelPhysicalOffset);
    ix_ossl_message_log("\tNon aliased physical address: 0x%p\n", pMemoryManager->m_pNonAliasChannelPhysicalAddress);

    _ix_rm_memory_block_cell_action(pMemoryManager->m_pBlockList, 0, _ix_rm_memory_block_cell_print, &counter);

    ix_ossl_message_log("\nTotal cells in the memory block: %lu\n\n", counter);

    return err;
}

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

#endif /* end defined(_IX_RM_DEBUG_) */


#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

#if !defined( _IX_RM_IMPL_HARDWARE_ )

/**
 * NAME: ix_rm_get_sram_channel_start_address
 *
 * DESCRIPTION: This function will return the base address of the specified SRAM channel
 *
 * @Param:  - IN arg_Channel - represents the SRAM memory channel in question.
 * 
 */
void* ix_rm_get_sram_channel_start_address(ix_uint32 arg_Channel)
{
    static const char s_aFunctionIdentifier[] = "ix_rm_get_sram_channel_start_address";

    (void)s_aFunctionIdentifier;

#if defined(_IX_RM_DEBUG_)

    if(arg_Channel >= IX_SRAM_CHANNELS_NUMBER)
        ix_ossl_message_log("Error in file: %s, line: %d\n\tFunction: %s\n\tReason: " \
                            "%s\n", \
                            __FILE__, \
                            __LINE__, \
                            s_aFunctionIdentifier, \
                            "bad channel number!");

#endif /* defined(_IX_RM_DEBUG_) */

    return g_aMemoryManagersGroups[IX_MEMORY_TYPE_SRAM].m_pMembers[arg_Channel].m_pStartAddress;
}

#endif /* !defined( _IX_RM_IMPL_HARDWARE_ ) */

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */
