/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 4
 *
 * = LIBRARY
 *      Resource Manager Library
 *
 * = MODULE
 *      Microengine API
 *
 * = FILENAME
 *      microengine.c
 *
 * = DESCRIPTION
 *      This file will define functions of the exported Microengine API interface.      
 *
 * = AUTHOR
 *       Adrian Georgescu
 *       adrian.georgescu@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      10/23/2001 3:48:58 PM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */

static const char s_aFileIdentifier[] = "$Id: microengine.c,v 1.38 2003/09/29 20:18:31 ageorges Exp $";

/*
#define _IX_RM_WORKBENCH_SUPPORT_
*/

#include "ix_ossl.h"

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"

#include "rm/symbols.h"

#if defined(_IX_RM_IMPL_HARDWARE_)

#include <uclo.h>
#include <halMev2Api.h>

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */


#include "rm/internal/internal_support.h"

#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"

#include "rm/system.h"
#include "rm/internal/internal_system.h"

#include "rm/memory.h"

#include "rm/buffer.h"

#include "rm/hardware.h"

#include "rm/communication.h"

#include "rm/configuration_property.h"

#include "rm/microengine.h"
#include "rm/internal/internal_microengine.h"



/**
 * This is a global flag that will signal if the uCode has been loaded or not
 * into the microstore.
 */
ix_uint32 g_MicrocodeLoaded = 0;

/**
 * NAME: ix_rm_ueng_set_ucode
 *
 * DESCRIPTION: This function sets the microcode image for microengines. The arg_pImageName 
 *  represents the image name containing the microcode in UOF format. The images are not 
 *  loaded into the microengines' microstore because symbols still 
 *  need to be patched. The images are loaded at the time of the ix_rm_ueng_load()
 *  call.
 * 
 * @Param:  - IN arg_pImageName - represents the image name containing the microcode in 
 *          UOF format.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_ueng_set_ucode( 
                               const char* arg_pFileName 
                             )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_ueng_set_ucode";

    ix_error err = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_MicroengineAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    IX_RM_ASSERT((arg_pFileName != 0), s_aFunctionIdentifier, "bad UOF file name!");

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    {
        ix_uint32 i;
        _ix_imported_symbol_node* pSymbolNode;
        _ix_imported_symbol_node* pNextSymbolNode;

        for(i = 0; i < IX_MAX_MICROENGINE_NUMBER; i++)
        {
            /* free the uCode imported symbol lists if required */
            pSymbolNode = g_MicroengineAPIManager.m_aImportedSymbolsList[i];
            while(pSymbolNode != 0)
            {
                pNextSymbolNode = pSymbolNode->m_pNext;
                if(pSymbolNode->m_Name != 0)
                    ix_ossl_free(pSymbolNode->m_Name);
                ix_ossl_free(pSymbolNode);
                pSymbolNode = pNextSymbolNode;
            }
            g_MicroengineAPIManager.m_aImportedSymbolsList[i] = 0;
        }
    }

#if defined(_IX_RM_IMPL_HARDWARE_)

    {
        int status;

        if(g_MicroengineAPIManager.m_UOFActiveHandle != 0)
        {
            UcLo_DeleObj(g_MicroengineAPIManager.m_UOFActiveHandle);
            g_MicroengineAPIManager.m_UOFActiveHandle = 0;

#if defined(_IX_RM_WORKBENCH_SUPPORT_)

            g_MicroengineAPIManager.m_IsUOFLoadedFromWorkbench = 0;
            g_MicroengineAPIManager.m_pUCodeBuffer = 0;
            g_MicroengineAPIManager.m_UCodeSize = 0;

#endif /* defined(_IX_RM_WORKBENCH_SUPPORT_) */

        }

#if defined(_IX_RM_WORKBENCH_SUPPORT_)

        else
        {
            /* If the uCode has not been loaded then free the buffer */
            if(g_MicroengineAPIManager.m_pUCodeBuffer != 0)
            {
                ix_ossl_free(g_MicroengineAPIManager.m_pUCodeBuffer);
                g_MicroengineAPIManager.m_pUCodeBuffer = 0;
                g_MicroengineAPIManager.m_UCodeSize = 0;
            }
        }

#endif /* defined(_IX_RM_WORKBENCH_SUPPORT_) */


        if((status = UcLo_LoadObjFile(&g_MicroengineAPIManager.m_UOFActiveHandle, (char*)arg_pFileName)) != UCLO_SUCCESS)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_LOAD_UCODE_FAILED, s_aFunctionIdentifier, 
                "loading of the UOF image from file failed!", status);

#if defined (_IX_RM_DEBUG_)

            if(status - UCLO_FAILURE < g_UCLOErrorStringCount)
                ix_ossl_message_log("UCLO error code is: %s\n", g_aUCLOErrorStrings[status-UCLO_FAILURE]);

#endif /* end defined (_IX_RM_DEBUG_) */

        }
        else
        {
            UcLo_VarMemSeg_T varsMemSegments;
            if((status = UcLo_GetVarMemSegs(g_MicroengineAPIManager.m_UOFActiveHandle, &varsMemSegments)) != UCLO_SUCCESS)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_LOAD_UCODE_FAILED, s_aFunctionIdentifier, 
                    "failed to retrieve the variables memory segments!", status);

#if defined (_IX_RM_DEBUG_)

                if(status - UCLO_FAILURE < g_UCLOErrorStringCount)
                    ix_ossl_message_log("UCLO error code is: %s\n", g_aUCLOErrorStrings[status-UCLO_FAILURE]);

#endif /* end defined (_IX_RM_DEBUG_) */

            }
            else
            {
                ix_uint32 segmentReserved;

                if(g_pReservedMemoryAreas != 0)
                {
                    ix_memory_reserved_area* pReservedArea;

                    for(;;)
                    {
                        segmentReserved = 1;

                        /* SRAM channel 0 */
                        if(varsMemSegments.sram0Size != 0)
                        {
                            for(segmentReserved = 0, pReservedArea = g_pReservedMemoryAreas; pReservedArea->m_Size != 0; pReservedArea++)
                            {
                                if((pReservedArea->m_Type == IX_MEMORY_TYPE_SRAM) && (pReservedArea->m_Channel == 0))
                                {
                                    if((varsMemSegments.sram0BaseAddr >= pReservedArea->m_StartOffset) && 
                                        (varsMemSegments.sram0BaseAddr + varsMemSegments.sram0Size <= 
                                        pReservedArea->m_StartOffset + pReservedArea->m_Size))
                                    {
                                        segmentReserved = 1;
                                        break;
                                    }
                                }
                            }
                            if(segmentReserved == 0)
                                break;
                        }

                        /* SRAM channel 1 */
                        if(varsMemSegments.sram1Size != 0)
                        {
                            for(segmentReserved = 0, pReservedArea = g_pReservedMemoryAreas; pReservedArea->m_Size != 0; pReservedArea++)
                            {
                                if((pReservedArea->m_Type == IX_MEMORY_TYPE_SRAM) && (pReservedArea->m_Channel == 1))
                                {
                                    if((varsMemSegments.sram1BaseAddr >= pReservedArea->m_StartOffset) && 
                                        (varsMemSegments.sram1BaseAddr + varsMemSegments.sram1Size <= 
                                        pReservedArea->m_StartOffset + pReservedArea->m_Size))
                                    {
                                        segmentReserved = 1;
                                        break;
                                    }
                                }
                            }
                            if(segmentReserved == 0)
                                break;
                        }

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

                        /* SRAM channel 2 */
                        if(varsMemSegments.sram2Size != 0)
                        {
                            for(segmentReserved = 0, pReservedArea = g_pReservedMemoryAreas; pReservedArea->m_Size != 0; pReservedArea++)
                            {
                                if((pReservedArea->m_Type == IX_MEMORY_TYPE_SRAM) && (pReservedArea->m_Channel == 2))
                                {
                                    if((varsMemSegments.sram2BaseAddr >= pReservedArea->m_StartOffset) && 
                                        (varsMemSegments.sram2BaseAddr + varsMemSegments.sram2Size <= 
                                        pReservedArea->m_StartOffset + pReservedArea->m_Size))
                                    {
                                        segmentReserved = 1;
                                        break;
                                    }
                                }
                            }
                            if(segmentReserved == 0)
                                break;
                        }

                        /* SRAM channel 3 */
                        if(varsMemSegments.sram3Size != 0)
                        {
                            for(segmentReserved = 0, pReservedArea = g_pReservedMemoryAreas; pReservedArea->m_Size != 0; pReservedArea++)
                            {
                                if((pReservedArea->m_Type == IX_MEMORY_TYPE_SRAM) && (pReservedArea->m_Channel == 3))
                                {
                                    if((varsMemSegments.sram3BaseAddr >= pReservedArea->m_StartOffset) && 
                                        (varsMemSegments.sram3BaseAddr + varsMemSegments.sram3Size <= 
                                        pReservedArea->m_StartOffset + pReservedArea->m_Size))
                                    {
                                        segmentReserved = 1;
                                        break;
                                    }
                                }
                            }
                            if(segmentReserved == 0)
                                break;
                        }

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */

                        /* DRAM channel 0 */
                        if(varsMemSegments.sdramSize != 0)
                        {
                            for(segmentReserved = 0, pReservedArea = g_pReservedMemoryAreas; pReservedArea->m_Size != 0; pReservedArea++)
                            {
                                if((pReservedArea->m_Type == IX_MEMORY_TYPE_DRAM) && (pReservedArea->m_Channel == 0))
                                {
                                    if((varsMemSegments.sdramBaseAddr >= pReservedArea->m_StartOffset) && 
                                        (varsMemSegments.sdramBaseAddr + varsMemSegments.sdramSize <= 
                                        pReservedArea->m_StartOffset + pReservedArea->m_Size))
                                    {
                                        segmentReserved = 1;
                                        break;
                                    }
                                }
                            }
                            if(segmentReserved == 0)
                                break;
                        }

                        /* SCRATCH channel 0 */
                        if(varsMemSegments.scratchSize != 0)
                        {
                            for(segmentReserved = 0, pReservedArea = g_pReservedMemoryAreas; pReservedArea->m_Size != 0; pReservedArea++)
                            {
                                if((pReservedArea->m_Type == IX_MEMORY_TYPE_SCRATCH) && (pReservedArea->m_Channel == 0))
                                {
                                    if((varsMemSegments.scratchBaseAddr >= pReservedArea->m_StartOffset) && 
                                        (varsMemSegments.scratchBaseAddr + varsMemSegments.scratchSize <= 
                                        pReservedArea->m_StartOffset + pReservedArea->m_Size))
                                    {
                                        segmentReserved = 1;
                                        break;
                                    }
                                }
                            }
                            if(segmentReserved == 0)
                                break;
                        }

                        break;
                    }                   
                }
                else
                {
                    if((varsMemSegments.sram0Size != 0) || (varsMemSegments.sram1Size != 0) || 

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

                        (varsMemSegments.sram2Size != 0) || (varsMemSegments.sram3Size != 0) || 

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */

                        (varsMemSegments.sdramSize != 0) || (varsMemSegments.scratchSize != 0))
                        segmentReserved = 0;
                    else
                        segmentReserved = 1;
                }

                
                if(segmentReserved == 0)
                {
                    ix_ossl_message_log("\nMicrocode variables use memory segments that are not prereserved!\n");
                    ix_ossl_message_log("You will have to specify the following memory regions to be prereserved\n"
                        "at the RM initialization time (i.e. the call to ix_rm_init())!\n\n");

                    ix_ossl_message_log("Microcode variables memory segments.\n");

                    if(varsMemSegments.sram0Size != 0)
                        ix_ossl_message_log("\t SRAM Channel 0: Base offset: 0x%x Size: %u\n", varsMemSegments.sram0BaseAddr, varsMemSegments.sram0Size);

                    if(varsMemSegments.sram1Size != 0)
                        ix_ossl_message_log("\t SRAM Channel 1: Base offset: 0x%x Size: %u\n", varsMemSegments.sram1BaseAddr, varsMemSegments.sram1Size);

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)
                    
                    if(varsMemSegments.sram2Size != 0)
                        ix_ossl_message_log("\t SRAM Channel 2: Base offset: 0x%x Size: %u\n", varsMemSegments.sram2BaseAddr, varsMemSegments.sram2Size);
                    
                    if(varsMemSegments.sram3Size != 0)
                        ix_ossl_message_log("\t SRAM Channel 3: Base offset: 0x%x Size: %u\n", varsMemSegments.sram3BaseAddr, varsMemSegments.sram3Size);

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */

                    if(varsMemSegments.sdramSize != 0)
                        ix_ossl_message_log("\t DRAM: Base offset: 0x%x Size: %u\n", varsMemSegments.sdramBaseAddr, varsMemSegments.sdramSize);
                    
                    if(varsMemSegments.scratchSize != 0)
                        ix_ossl_message_log("\t SCRATCH: Base offset: 0x%x Size: %u\n", varsMemSegments.scratchBaseAddr, varsMemSegments.scratchSize);

                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_LOAD_UCODE_FAILED, s_aFunctionIdentifier, 
                        "microcode needs memory areas that are not prereserved!", status);
                }
                else
                {
                    g_MicroengineAPIManager.m_IsUOFLoadedFromFile = 1;
                }
            }
        }
    }

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}

/**
 * NAME: ix_rm_ueng_map_ucode
 *
 * DESCRIPTION: This function sets the microcode image for microengines from a passed buffer.
 *  The images are not loaded into the microengines' microstore because symbols still 
 *  need to be patched. The images are loaded at the time of the ix_rm_ueng_load()
 *  call. The arg_pUOFImage should point to a RW memory location rather than a RO memory
 *  location.
 * 
 * @Param:  - IN arg_pUOFImage - represents the image buffer from where the microcode should be mapped
 * @Param:  - IN arg_Size - represents the size of the image buffer
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_ueng_map_ucode( 
                               void* arg_pUOFImage, 
                               ix_int32 arg_Size 
                             )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_ueng_map_ucode";

    ix_error err = IX_SUCCESS;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_MicroengineAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    IX_RM_ASSERT((arg_pUOFImage != 0), s_aFunctionIdentifier, "bad UOF buffer!");

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    {
        ix_uint32 i;
        _ix_imported_symbol_node* pSymbolNode;
        _ix_imported_symbol_node* pNextSymbolNode;

        for(i = 0; i < IX_MAX_MICROENGINE_NUMBER; i++)
        {
            /* free the uCode imported symbol lists if required */
            pSymbolNode = g_MicroengineAPIManager.m_aImportedSymbolsList[i];
            while(pSymbolNode != 0)
            {
                pNextSymbolNode = pSymbolNode->m_pNext;
                if(pSymbolNode->m_Name != 0)
                    ix_ossl_free(pSymbolNode->m_Name);
                ix_ossl_free(pSymbolNode);
                pSymbolNode = pNextSymbolNode;
            }
            g_MicroengineAPIManager.m_aImportedSymbolsList[i] = 0;
        }
    }

#if defined(_IX_RM_IMPL_HARDWARE_)

    {
        int status;

        if(g_MicroengineAPIManager.m_UOFActiveHandle != 0)
        {
            UcLo_DeleObj(g_MicroengineAPIManager.m_UOFActiveHandle);
            g_MicroengineAPIManager.m_UOFActiveHandle = 0;

#if defined(_IX_RM_WORKBENCH_SUPPORT_)

            g_MicroengineAPIManager.m_IsUOFLoadedFromWorkbench = 0;
            g_MicroengineAPIManager.m_pUCodeBuffer = 0;
            g_MicroengineAPIManager.m_UCodeSize = 0;

#endif /* defined(_IX_RM_WORKBENCH_SUPPORT_) */

        }

#if defined(_IX_RM_WORKBENCH_SUPPORT_)

        else
        {
            /* If the uCode has not been loaded then free the buffer */
            if(g_MicroengineAPIManager.m_pUCodeBuffer != 0)
            {
                ix_ossl_free(g_MicroengineAPIManager.m_pUCodeBuffer);
                g_MicroengineAPIManager.m_pUCodeBuffer = 0;
                g_MicroengineAPIManager.m_UCodeSize = 0;
            }
        }

#endif /* defined(_IX_RM_WORKBENCH_SUPPORT_) */


        if((status = UcLo_MapObjAddr(&g_MicroengineAPIManager.m_UOFActiveHandle,
            arg_pUOFImage, arg_Size, 0)) != UCLO_SUCCESS)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_MAP_UCODE_FAILED, s_aFunctionIdentifier, 
                "mapping of the UOF image to memory failed!", status);

#if defined (_IX_RM_DEBUG_)

                if(status - UCLO_FAILURE < g_UCLOErrorStringCount)
                    ix_ossl_message_log("UCLO error code is: %s\n", g_aUCLOErrorStrings[status-UCLO_FAILURE]);

#endif /* end defined (_IX_RM_DEBUG_) */

        }
        else
        {
            UcLo_VarMemSeg_T varsMemSegments;
            if((status = UcLo_GetVarMemSegs(g_MicroengineAPIManager.m_UOFActiveHandle, &varsMemSegments)) != UCLO_SUCCESS)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_MAP_UCODE_FAILED, s_aFunctionIdentifier, 
                    "failed to retrieve the variables memory segments!", status);

#if defined (_IX_RM_DEBUG_)

                if(status - UCLO_FAILURE < g_UCLOErrorStringCount)
                    ix_ossl_message_log("UCLO error code is: %s\n", g_aUCLOErrorStrings[status-UCLO_FAILURE]);

#endif /* end defined (_IX_RM_DEBUG_) */

            }
            else
            {
                ix_uint32 segmentReserved;

                if(g_pReservedMemoryAreas != 0)
                {
                    ix_memory_reserved_area* pReservedArea;

                    for(;;)
                    {
                        segmentReserved = 1;

                        /* SRAM channel 0 */
                        if(varsMemSegments.sram0Size != 0)
                        {
                            for(segmentReserved = 0, pReservedArea = g_pReservedMemoryAreas; pReservedArea->m_Size != 0; pReservedArea++)
                            {
                                if((pReservedArea->m_Type == IX_MEMORY_TYPE_SRAM) && (pReservedArea->m_Channel == 0))
                                {
                                    if((varsMemSegments.sram0BaseAddr >= pReservedArea->m_StartOffset) && 
                                        (varsMemSegments.sram0BaseAddr + varsMemSegments.sram0Size <= 
                                        pReservedArea->m_StartOffset + pReservedArea->m_Size))
                                    {
                                        segmentReserved = 1;
                                        break;
                                    }
                                }
                            }
                            if(segmentReserved == 0)
                                break;
                        }

                        /* SRAM channel 1 */
                        if(varsMemSegments.sram1Size != 0)
                        {
                            for(segmentReserved = 0, pReservedArea = g_pReservedMemoryAreas; pReservedArea->m_Size != 0; pReservedArea++)
                            {
                                if((pReservedArea->m_Type == IX_MEMORY_TYPE_SRAM) && (pReservedArea->m_Channel == 1))
                                {
                                    if((varsMemSegments.sram1BaseAddr >= pReservedArea->m_StartOffset) && 
                                        (varsMemSegments.sram1BaseAddr + varsMemSegments.sram1Size <= 
                                        pReservedArea->m_StartOffset + pReservedArea->m_Size))
                                    {
                                        segmentReserved = 1;
                                        break;
                                    }
                                }
                            }
                            if(segmentReserved == 0)
                                break;
                        }

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

                        /* SRAM channel 2 */
                        if(varsMemSegments.sram2Size != 0)
                        {
                            for(segmentReserved = 0, pReservedArea = g_pReservedMemoryAreas; pReservedArea->m_Size != 0; pReservedArea++)
                            {
                                if((pReservedArea->m_Type == IX_MEMORY_TYPE_SRAM) && (pReservedArea->m_Channel == 2))
                                {
                                    if((varsMemSegments.sram2BaseAddr >= pReservedArea->m_StartOffset) && 
                                        (varsMemSegments.sram2BaseAddr + varsMemSegments.sram2Size <= 
                                        pReservedArea->m_StartOffset + pReservedArea->m_Size))
                                    {
                                        segmentReserved = 1;
                                        break;
                                    }
                                }
                            }
                            if(segmentReserved == 0)
                                break;
                        }

                        /* SRAM channel 3 */
                        if(varsMemSegments.sram3Size != 0)
                        {
                            for(segmentReserved = 0, pReservedArea = g_pReservedMemoryAreas; pReservedArea->m_Size != 0; pReservedArea++)
                            {
                                if((pReservedArea->m_Type == IX_MEMORY_TYPE_SRAM) && (pReservedArea->m_Channel == 3))
                                {
                                    if((varsMemSegments.sram3BaseAddr >= pReservedArea->m_StartOffset) && 
                                        (varsMemSegments.sram3BaseAddr + varsMemSegments.sram3Size <= 
                                        pReservedArea->m_StartOffset + pReservedArea->m_Size))
                                    {
                                        segmentReserved = 1;
                                        break;
                                    }
                                }
                            }
                            if(segmentReserved == 0)
                                break;
                        }

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */

                        /* DRAM channel 0 */
                        if(varsMemSegments.sdramSize != 0)
                        {
                            for(segmentReserved = 0, pReservedArea = g_pReservedMemoryAreas; pReservedArea->m_Size != 0; pReservedArea++)
                            {
                                if((pReservedArea->m_Type == IX_MEMORY_TYPE_DRAM) && (pReservedArea->m_Channel == 0))
                                {
                                    if((varsMemSegments.sdramBaseAddr >= pReservedArea->m_StartOffset) && 
                                        (varsMemSegments.sdramBaseAddr + varsMemSegments.sdramSize <= 
                                        pReservedArea->m_StartOffset + pReservedArea->m_Size))
                                    {
                                        segmentReserved = 1;
                                        break;
                                    }
                                }
                            }
                            if(segmentReserved == 0)
                                break;
                        }

                        /* SCRATCH channel 0 */
                        if(varsMemSegments.scratchSize != 0)
                        {
                            for(segmentReserved = 0, pReservedArea = g_pReservedMemoryAreas; pReservedArea->m_Size != 0; pReservedArea++)
                            {
                                if((pReservedArea->m_Type == IX_MEMORY_TYPE_SCRATCH) && (pReservedArea->m_Channel == 0))
                                {
                                    if((varsMemSegments.scratchBaseAddr >= pReservedArea->m_StartOffset) && 
                                        (varsMemSegments.scratchBaseAddr + varsMemSegments.scratchSize <= 
                                        pReservedArea->m_StartOffset + pReservedArea->m_Size))
                                    {
                                        segmentReserved = 1;
                                        break;
                                    }
                                }
                            }
                            if(segmentReserved == 0)
                                break;
                        }

                        break;
                    }
                }
                else
                {
                    if((varsMemSegments.sram0Size != 0) || (varsMemSegments.sram1Size != 0) || 

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

                        (varsMemSegments.sram2Size != 0) || (varsMemSegments.sram3Size != 0) || 

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */

                        (varsMemSegments.sdramSize != 0) || (varsMemSegments.scratchSize != 0))
                        segmentReserved = 0;
                    else
                        segmentReserved = 1;
                }

                if(segmentReserved == 0)
                {
                    ix_ossl_message_log("\nMicrocode variables use memory segments that are not prereserved!\n");
                    ix_ossl_message_log("You will have to specify the following memory regions to be prereserved\n"
                        "at the RM initialization time (i.e. the call to ix_rm_init())!\n\n");

                    ix_ossl_message_log("Microcode variables memory segments.\n");

                    if(varsMemSegments.sram0Size != 0)
                        ix_ossl_message_log("\t SRAM Channel 0: Base offset: 0x%x Size: %u\n", varsMemSegments.sram0BaseAddr, varsMemSegments.sram0Size);

                    if(varsMemSegments.sram1Size != 0)
                        ix_ossl_message_log("\t SRAM Channel 1: Base offset: 0x%x Size: %u\n", varsMemSegments.sram1BaseAddr, varsMemSegments.sram1Size);

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)
                    
                    if(varsMemSegments.sram2Size != 0)
                        ix_ossl_message_log("\t SRAM Channel 2: Base offset: 0x%x Size: %u\n", varsMemSegments.sram2BaseAddr, varsMemSegments.sram2Size);
                    
                    if(varsMemSegments.sram3Size != 0)
                        ix_ossl_message_log("\t SRAM Channel 3: Base offset: 0x%x Size: %u\n", varsMemSegments.sram3BaseAddr, varsMemSegments.sram3Size);

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2800_) */

                    if(varsMemSegments.sdramSize != 0)
                        ix_ossl_message_log("\t DRAM: Base offset: 0x%x Size: %u\n", varsMemSegments.sdramBaseAddr, varsMemSegments.sdramSize);
                    
                    if(varsMemSegments.scratchSize != 0)
                        ix_ossl_message_log("\t SCRATCH: Base offset: 0x%x Size: %u\n", varsMemSegments.scratchBaseAddr, varsMemSegments.scratchSize);

                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_MAP_UCODE_FAILED, s_aFunctionIdentifier, 
                        "microcode needs memory areas that are not prereserved!", status);
                }
                else
                {
                    g_MicroengineAPIManager.m_IsUOFLoadedFromFile = 0;
                }
            }
        }
    }
#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}

/**
 * NAME: ix_rm_ueng_reset_all
 *
 * DESCRIPTION: This function stops and resets all active microengines.
 *  It removes the active UOF image from memory and brings the ME system
 *  into initial state.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_ueng_reset_all(void)
{
    static const char s_aFunctionIdentifier[] = "ix_rm_ueng_reset_all";

    ix_error err = IX_SUCCESS;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_MicroengineAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    {
        ix_uint32 i;
        _ix_imported_symbol_node* pSymbolNode;
        _ix_imported_symbol_node* pNextSymbolNode;

        for(i = 0; i < IX_MAX_MICROENGINE_NUMBER; i++)
        {
            /* free the uCode imported symbol lists if required */
            pSymbolNode = g_MicroengineAPIManager.m_aImportedSymbolsList[i];
            while(pSymbolNode != 0)
            {
                pNextSymbolNode = pSymbolNode->m_pNext;
                if(pSymbolNode->m_Name != 0)
                    ix_ossl_free(pSymbolNode->m_Name);
                ix_ossl_free(pSymbolNode);
                pSymbolNode = pNextSymbolNode;
            }
            g_MicroengineAPIManager.m_aImportedSymbolsList[i] = 0;
        }
    }

#if defined(_IX_RM_IMPL_HARDWARE_)

    {
        int status;
        ix_uint32 i;

        /* stop and reset all microengines */
        for(i = 0; i < IX_MAX_MICROENGINE_NUMBER; i++)
        {
            status = halMe_IsMeEnabled(g_aMicroengineNumberMapping[i]);
            switch(status)
            {
                case HALME_ENABLED:
                    halMe_Stop(g_aMicroengineNumberMapping[i], 0xffffffff);
                    /* intentional fall through */
                case HALME_DISABLED:
                    halMe_Reset(1<<g_aMicroengineNumberMapping[i],1);
                    /* intentional fall through */
                case HALME_RESET:
                    halMe_ClrReset(1<<g_aMicroengineNumberMapping[i]);
                    break;
                default:
                    break;
            }
        }
    }

    if(g_MicroengineAPIManager.m_UOFActiveHandle != 0)
    {
        UcLo_DeleObj(g_MicroengineAPIManager.m_UOFActiveHandle);
        g_MicroengineAPIManager.m_UOFActiveHandle = 0;

#if defined(_IX_RM_WORKBENCH_SUPPORT_)

        g_MicroengineAPIManager.m_IsUOFLoadedFromWorkbench = 0;
        g_MicroengineAPIManager.m_pUCodeBuffer = 0;
        g_MicroengineAPIManager.m_UCodeSize = 0;

#endif /* defined(_IX_RM_WORKBENCH_SUPPORT_) */

    }

#if defined(_IX_RM_WORKBENCH_SUPPORT_)

    else
    {
        /* If the uCode has not been loaded then free the buffer */
        if(g_MicroengineAPIManager.m_pUCodeBuffer != 0)
        {
            ix_ossl_free(g_MicroengineAPIManager.m_pUCodeBuffer);
            g_MicroengineAPIManager.m_pUCodeBuffer = 0;
            g_MicroengineAPIManager.m_UCodeSize = 0;
        }
    }

#endif /* defined(_IX_RM_WORKBENCH_SUPPORT_) */


#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    /* mark that the uCode is not loaded */
    g_MicrocodeLoaded = 0;

    g_MicroengineAPIManager.m_UOFActiveHandle = 0;
    g_MicroengineAPIManager.m_IsUOFLoadedFromFile = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}

/**
 * NAME: ix_rm_ueng_patch_symbols
 *
 * DESCRIPTION: This function will patch imported symbols into the microcode.  
 * 
 * @Param:  - IN arg_UengNumber - represents the microengine number for which this call is 
 *          intended.
 * @Param:  - IN arg_SymbolsNumber - represents the number of symbols in the arg_Symbols 
 *          array that will be patched.
 * @Param:  - IN arg_aSymbols - represents an array of symbols that will be patched into 
 *          the microcode.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_ueng_patch_symbols( 
                                   ix_uint32 arg_MENumber, 
                                   ix_uint32 arg_SymbolsNumber, 
                                   const ix_imported_symbol arg_aSymbols[] 
                                 )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_ueng_patch_symbols";

    ix_error err = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_MicroengineAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

	IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

    IX_RM_ASSERT(((arg_MENumber <= 0x03) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x13))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x03 and 0x10 - 0x13!");

#elif (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

    IX_RM_ASSERT(((arg_MENumber <= 0x07) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x17))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x07 and 0x10 - 0x17!");

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    err = _ix_rm_microengine_patch_symbols(arg_MENumber, arg_SymbolsNumber, arg_aSymbols);

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}



/**
 * NAME: ix_rm_ueng_load
 *
 * DESCRIPTION: This function will load the microcode into the microengine store. The 
 *          ix_rm_ueng_set_ucode() function should be called prior to this one. Also is 
 *          expected that all imported symbols to be patched by this time.
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_ueng_load(void)
{
    static const char s_aFunctionIdentifier[] = "ix_rm_ueng_load";

    ix_error err = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_MicroengineAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

	IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    {
        ix_uint32 i;

        
        /**
         * Patch all QArray symbols on all MEs that import the symbols
         */
        _ix_rm_patch_qarray_symbols();


        /**
         * Don't forget to do the patching of all registered symbols
         */
        for(i = 0; i < IX_MAX_MICROENGINE_NUMBER; i++)
        {
            err = _ix_rm_microengine_register_internal_symbols(g_aMicroengineNumberMapping[i]);
            if(err != IX_SUCCESS)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_REGISTER_INTERNAL_SYMBOLS_FAILED, s_aFunctionIdentifier, 
                    "registration of local symbols failed!", err);
                goto label_1;
            }

            err = _ix_rm_microengine_patch_all_symbols(g_aMicroengineNumberMapping[i]);
            if(err != IX_SUCCESS)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_PATCH_SYMBOLS_FAILED, s_aFunctionIdentifier, 
                    "the patching of uCode symbols failed!", err);
                goto label_1;
            }
        }

#if defined(_IX_RM_IMPL_HARDWARE_)

        {
            int status;

            if((status = UcLo_WriteUimageAll(g_MicroengineAPIManager.m_UOFActiveHandle)) != UCLO_SUCCESS)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_LOAD_TO_USTORE_FAILED, s_aFunctionIdentifier, 
                    "loading of the UOF image to the microstore failed!", status);

#if defined (_IX_RM_DEBUG_)

                if(status - UCLO_FAILURE < g_UCLOErrorStringCount)
                    ix_ossl_message_log("UCLO error code is: %s\n", g_aUCLOErrorStrings[status-UCLO_FAILURE]);

#endif /* end defined (_IX_RM_DEBUG_) */

                goto label_1;

            }
        }

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

    }

    /* set the global flag to 1 to signal that the uCode has been loaded into the uSore */
    g_MicrocodeLoaded = 1;

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;

}


/**
 * NAME: ix_rm_ueng_start
 *
 * DESCRIPTION: This function will start the passed microengine and will activate the
 *  specified contexts.
 * 
 * @Param:  - IN arg_MENumber - the microengine's number to be started.
 * @Param:  - IN arg_EnableContextMask - mask specifying what microengine contexts should
 *          be enabled at the start time.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_ueng_start(
                           ix_uint32 arg_MENumber, 
                           ix_bit_mask32 arg_EnableContextMask 
                         )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_ueng_start";

    ix_error err = IX_SUCCESS;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_MicroengineAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

    IX_RM_ASSERT(((arg_MENumber <= 0x03) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x13))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x03 and 0x10 - 0x13!");

#elif (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

    IX_RM_ASSERT(((arg_MENumber <= 0x07) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x17))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x07 and 0x10 - 0x17!");

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined(_IX_RM_IMPL_HARDWARE_)

    {
        int status;

        status = UcLo_VerifyUengine(g_MicroengineAPIManager.m_UOFActiveHandle, arg_MENumber);
        if(status == UCLO_SUCCESS)
        {

            if((status = halMe_Start(arg_MENumber, arg_EnableContextMask)) != HALME_SUCCESS)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_START_FAILED, s_aFunctionIdentifier, 
                    "starting of the specified ME failed!", status);

#if defined (_IX_RM_DEBUG_)

                if(status - HALME_FAIL < g_HALMEErrorStringCount)
                    ix_ossl_message_log("HALME error code is: %s\n", g_aHALMEErrorStrings[status-HALME_FAIL]);

#endif /* end defined (_IX_RM_DEBUG_) */ 
        
            }
        }
        else
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_START_FAILED, s_aFunctionIdentifier, 
                    "The microengine's image is not valid!", status);

#if defined (_IX_RM_DEBUG_)

            if(status - UCLO_FAILURE < g_UCLOErrorStringCount)
                ix_ossl_message_log("UCLO error code is: %s\n", g_aUCLOErrorStrings[status-UCLO_FAILURE]);

#endif /* end defined (_IX_RM_DEBUG_) */ 

        }
    }

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}

/**
 * NAME: ix_rm_ueng_start_group
 *
 * DESCRIPTION: This function will start the specified ME group and will activate the
 *  specified contexts. The MEs will be started in a very tight loop in this case.
 * 
 * @Param:  - IN arg_pMEGroup - the address of an array of ix_me_start_node structures
 *  specifying what MEs to be started.
 * @Param:  - IN ix_uint32 - the count of elements in the above passed ME group array.
 *
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_ueng_start_group(
                                 ix_me_start_node* arg_pMEGroup, 
                                 ix_uint32 arg_MEGroupCount 
                               )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_ueng_start_group";

    ix_error err = IX_SUCCESS;
    ix_uint32 i;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_MicroengineAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    for(i = 0; i < arg_MEGroupCount; i++)
    {

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

        IX_RM_ASSERT(((arg_pMEGroup[i].m_MENumber <= 0x03) || ((arg_pMEGroup[i].m_MENumber >= 0x10) && (arg_pMEGroup[i].m_MENumber <= 0x13))), 
            s_aFunctionIdentifier, "bad microengine number: 0x00-0x03 and 0x10 - 0x13!");

#elif (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

        IX_RM_ASSERT(((arg_pMEGroup[i].m_MENumber <= 0x07) || ((arg_pMEGroup[i].m_MENumber >= 0x10) && (arg_pMEGroup[i].m_MENumber <= 0x17))), 
            s_aFunctionIdentifier, "bad microengine number: 0x00-0x07 and 0x10 - 0x17!");

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

    }

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined(_IX_RM_IMPL_HARDWARE_)

    {
        int status;
        ix_me_start_node* pStartNode;

        for(i = 0, pStartNode = arg_pMEGroup; i < arg_MEGroupCount; i++, pStartNode++)
        {

            status = UcLo_VerifyUengine(g_MicroengineAPIManager.m_UOFActiveHandle, pStartNode->m_MENumber);
            if(status != UCLO_SUCCESS)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_START_FAILED, s_aFunctionIdentifier, 
                        "The microengine's image is not valid!", status);

    #if defined (_IX_RM_DEBUG_)

                if(status - UCLO_FAILURE < g_UCLOErrorStringCount)
                    ix_ossl_message_log("UCLO error code is: %s\n", g_aUCLOErrorStrings[status-UCLO_FAILURE]);

    #endif /* end defined (_IX_RM_DEBUG_) */ 

                goto label_1;

            }
        }

        for(i = 0, pStartNode = arg_pMEGroup; i < arg_MEGroupCount; i++, pStartNode++)
        {

            if((status = halMe_Start(pStartNode->m_MENumber, pStartNode->m_EnableContextMask)) != HALME_SUCCESS)
            {
                ix_ossl_message_log("ME: 0x%02lx, context mask: 0x%08lx\n", pStartNode->m_MENumber, pStartNode->m_EnableContextMask); 
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_START_FAILED, s_aFunctionIdentifier, 
                    "starting of the specified ME failed!", status);

#if defined (_IX_RM_DEBUG_)

                if(status - HALME_FAIL < g_HALMEErrorStringCount)
                    ix_ossl_message_log("HALME error code is: %s\n", g_aHALMEErrorStrings[status-HALME_FAIL]);

#endif /* end defined (_IX_RM_DEBUG_) */ 

                goto label_1;        
            }
        }
        
    }

label_1:

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}

/**
 * NAME: ix_rm_ueng_start_all
 *
 * DESCRIPTION: This function will start all the active MEs and will activate all the
 *  contexts. The MEs will be started in a very tight loop in this case. It depends on the
 *  loaded uCode which MEs are active.
 * 
 * @Return: IX_ERROR_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_ueng_start_all(void)
{
    static const char s_aFunctionIdentifier[] = "ix_rm_rm_ueng_start_all";

    ix_error err = IX_SUCCESS;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_MicroengineAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined(_IX_RM_IMPL_HARDWARE_)

    {
        int status;
        ix_uint32 i;
        ix_uint32 activeME[IX_MAX_MICROENGINE_NUMBER];
        ix_uint32 activeMECount = 0;

        for(i = 0; i < IX_MAX_MICROENGINE_NUMBER; i++)
        {

            status = UcLo_VerifyUengine(g_MicroengineAPIManager.m_UOFActiveHandle, g_aMicroengineNumberMapping[i]);

            /* If things are all right store the ME number in order to be started */
            if(status == UCLO_SUCCESS)
                activeME[activeMECount++] = g_aMicroengineNumberMapping[i];
        }

        for(i = 0; i < activeMECount; i++)
        {

            if((status = halMe_Start(activeME[i], 0xffffffff)) != HALME_SUCCESS)
            {
                ix_ossl_message_log("ME: 0x%02lx\n", activeME[i]); 
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_START_FAILED, s_aFunctionIdentifier, 
                    "starting of the specified ME failed!", status);

#if defined (_IX_RM_DEBUG_)

                if(status - HALME_FAIL < g_HALMEErrorStringCount)
                    ix_ossl_message_log("HALME error code is: %s\n", g_aHALMEErrorStrings[status-HALME_FAIL]);

#endif /* end defined (_IX_RM_DEBUG_) */ 

                goto label_1;        
            }
        }
        
    }

label_1:

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}

/**
 * NAME: ix_rm_ueng_stop
 *
 * DESCRIPTION: This function will stop the specified microengine with all 
 *  active contexts.
 * 
 * @Param:  - IN arg_MENumber - the microengine's number to be stopped.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_ueng_stop(
                          ix_uint32 arg_MENumber
                        )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_ueng_stop";

    ix_error err = IX_SUCCESS;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_MicroengineAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

    IX_RM_ASSERT(((arg_MENumber <= 0x03) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x13))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x03 and 0x10 - 0x13!");

#elif (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

    IX_RM_ASSERT(((arg_MENumber <= 0x07) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x17))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x07 and 0x10 - 0x17!");

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined(_IX_RM_IMPL_HARDWARE_)

    {
        int status;

        /* Stop all contexts */
        if((status = halMe_Stop(arg_MENumber, 0xffffffff)) != HALME_SUCCESS)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_STOP_FAILED, s_aFunctionIdentifier, 
                "failed to stop the specified ME!", status);

#if defined (_IX_RM_DEBUG_)

            if(status - HALME_FAIL < g_HALMEErrorStringCount)
                ix_ossl_message_log("HLME error code is: %s\n", g_aHALMEErrorStrings[status-HALME_FAIL]);

#endif /* end defined (_IX_RM_DEBUG_) */ 
        
        }
    }

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}



/**
 * NAME: ix_rm_ueng_reset
 *
 * DESCRIPTION: This function will reset the specified microengine. If the arg_ClearRegisters
 *  is non zero then all the microengine registers will be cleared and the microengine
 *  will be in the same state as at startup.
 * 
 * @Param:  - IN arg_MENumber - the microengine's number to be reset. 
 * @Param:  - IN arg_ClearRegisters - flag specifying if all registers should be cleared
 *          on reset.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_ueng_reset(
                           ix_uint32 arg_MENumber, 
                           ix_uint32 arg_ClearRegisters
                         )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_ueng_reset";

    ix_error err = IX_SUCCESS;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_MicroengineAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

    IX_RM_ASSERT(((arg_MENumber <= 0x03) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x13))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x03 and 0x10 - 0x13!");

#elif (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

    IX_RM_ASSERT(((arg_MENumber <= 0x07) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x17))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x07 and 0x10 - 0x17!");

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined(_IX_RM_IMPL_HARDWARE_)

    /* Reset the ME first */
    halMe_Reset((1U << arg_MENumber), arg_ClearRegisters);

    /* Take the ME out of reset state */
    halMe_ClrReset(1U << arg_MENumber);

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: ix_rm_ueng_enable
 *
 * DESCRIPTION: This function will enable one or more of the microengine's contexts.
 *  The microengine should be started in order for this call to succeed.
 * 
 * @Param:  - IN arg_MENumber - the microengine's number whose contexts should be enabled. 
 * @Param:  - IN arg_EnableContextMask - mask specifying what microengine contexts should
 *          be enabled.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_ueng_enable( 
                            ix_uint32 arg_MENumber, 
                            ix_bit_mask32 arg_EnableContextMask 
                          )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_ueng_enable";

    ix_error err = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_MicroengineAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

	IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

    IX_RM_ASSERT(((arg_MENumber <= 0x03) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x13))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x03 and 0x10 - 0x13!");

#elif (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

    IX_RM_ASSERT(((arg_MENumber <= 0x07) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x17))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x07 and 0x10 - 0x17!");

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined(_IX_RM_IMPL_HARDWARE_)

    {
        int status;

        status = UcLo_VerifyUengine(g_MicroengineAPIManager.m_UOFActiveHandle, arg_MENumber);
        if(status == UCLO_SUCCESS)
        {

            if((status = halMe_Start(arg_MENumber, arg_EnableContextMask)) != HALME_SUCCESS)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_ENABLE_CTX_FAILED, s_aFunctionIdentifier, 
                    "failed to enable the specified ME contexts!", status);

#if defined (_IX_RM_DEBUG_)

                if(status - HALME_FAIL < g_HALMEErrorStringCount)
                    ix_ossl_message_log("HLME error code is: %s\n", g_aHALMEErrorStrings[status-HALME_FAIL]);

#endif /* end defined (_IX_RM_DEBUG_) */ 
        
            }
        }
        else
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_ENABLE_CTX_FAILED, s_aFunctionIdentifier, 
                    "The microengine's image is not valid!", status);

#if defined (_IX_RM_DEBUG_)

            if(status - UCLO_FAILURE < g_UCLOErrorStringCount)
                ix_ossl_message_log("UCLO error code is: %s\n", g_aUCLOErrorStrings[status-UCLO_FAILURE]);

#endif /* end defined (_IX_RM_DEBUG_) */ 


        }

    }

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}




/**
 * NAME: ix_rm_ueng_disable
 *
 *
 * DESCRIPTION: This function will disable one or more of the microengine's contexts.
 *  The microengine should be started in order for this call to succeed.
 * 
 * @Param:  - IN arg_MENumber - the microengine's number whose contexts should be disabled. 
 * @Param:  - IN arg_DisableContextMask - mask specifying what microengine contexts should
 *          be disabled.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_ueng_disable( 
                             ix_uint32 arg_MENumber, 
                             ix_bit_mask32 arg_DisableContextMask 
                           )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_ueng_disable";

    ix_error err = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_MicroengineAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

	IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if (_IX_HARDWARE_TYPE_ == _IX_HW_2400_)

    IX_RM_ASSERT(((arg_MENumber <= 0x03) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x13))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x03 and 0x10 - 0x13!");

#elif (_IX_HARDWARE_TYPE_ == _IX_HW_2800_)

    IX_RM_ASSERT(((arg_MENumber <= 0x07) || ((arg_MENumber >= 0x10) && (arg_MENumber <= 0x17))), 
        s_aFunctionIdentifier, "bad microengine number: 0x00-0x07 and 0x10 - 0x17!");

#endif /* end (_IX_HARDWARE_TYPE_ == _IX_HW_2400_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined(_IX_RM_IMPL_HARDWARE_)

    {
        int status;

        if((status = halMe_Stop(arg_MENumber, arg_DisableContextMask)) != HALME_SUCCESS)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_DISABLE_CTX_FAILED, s_aFunctionIdentifier, 
                "failed to disable the specified ME contexts!", status);

#if defined (_IX_RM_DEBUG_)

            if(status - HALME_FAIL < g_HALMEErrorStringCount)
                ix_ossl_message_log("HLME error code is: %s\n", g_aHALMEErrorStrings[status-HALME_FAIL]);

#endif /* end defined (_IX_RM_DEBUG_) */ 
        
        }
    }

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}

#if defined(_IX_RM_WORKBENCH_SUPPORT_)


/**
 * NAME: ix_rm_ueng_set_ucode_from_wb
 *
 * DESCRIPTION: This function will get the uCode downloaded from the WorkBench
 * 
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_ueng_set_ucode_from_wb(void)
{
    static const char s_aFunctionIdentifier[] = "ix_rm_ueng_load_ucode_from_wb";

    ix_error err = IX_SUCCESS;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_MicroengineAPIManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);


#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    {
        ix_uint32 i;
        _ix_imported_symbol_node* pSymbolNode;
        _ix_imported_symbol_node* pNextSymbolNode;

        for(i = 0; i < IX_MAX_MICROENGINE_NUMBER; i++)
        {
            /* free the uCode imported symbol lists if required */
            pSymbolNode = g_MicroengineAPIManager.m_aImportedSymbolsList[i];
            while(pSymbolNode != 0)
            {
                pNextSymbolNode = pSymbolNode->m_pNext;
                if(pSymbolNode->m_Name != 0)
                    ix_ossl_free(pSymbolNode->m_Name);
                ix_ossl_free(pSymbolNode);
                pSymbolNode = pNextSymbolNode;
            }
            g_MicroengineAPIManager.m_aImportedSymbolsList[i] = 0;
        }
    }

#if defined(_IX_RM_IMPL_HARDWARE_)

    {
        int status;

        if(g_MicroengineAPIManager.m_UOFActiveHandle != 0)
        {
            UcLo_DeleObj(g_MicroengineAPIManager.m_UOFActiveHandle);
            g_MicroengineAPIManager.m_UOFActiveHandle = 0;
            g_MicroengineAPIManager.m_IsUOFLoadedFromWorkbench = 0;
            g_MicroengineAPIManager.m_pUCodeBuffer = 0;
            g_MicroengineAPIManager.m_UCodeSize = 0;
        }
        else
        {
            /* If the uCode has not been loaded then free the buffer */
            if(g_MicroengineAPIManager.m_pUCodeBuffer != 0)
            {
                ix_ossl_free(g_MicroengineAPIManager.m_pUCodeBuffer);
                g_MicroengineAPIManager.m_pUCodeBuffer = 0;
                g_MicroengineAPIManager.m_UCodeSize = 0;
            }
        }

        /* Register a function with the udbg library */

        /* Waiting for the uCode to be downloaded */
        {
            ix_uint32 count = 0;

            ix_ossl_message_log("Download the microcode from WorkBench!\n");
            while(count < 90)
            {
                if(g_MicroengineAPIManager.m_pUCodeBuffer != 0)
                    ix_ossl_message_log(".");
                else
                {
                    ix_ossl_sleep(200);
                    break;
                };
            }
            if(count >= 90)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_LOAD_UCODE_FROM_WB_FAILED, s_aFunctionIdentifier, 
                 "the uCode download from WB timed out!", 0);
                goto label_1;
            }

            ix_ossl_message_log("Microcode download from WorkBench completed!\n");
        }

        /* Unregister the processing function from the udbg library */

        if((status = UcLo_MapObjAddr(&g_MicroengineAPIManager.m_UOFActiveHandle,
            g_MicroengineAPIManager.m_pUCodeBuffer, g_MicroengineAPIManager.m_UCodeSize, 0)) != UCLO_SUCCESS)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_LOAD_UCODE_FROM_WB_FAILED, s_aFunctionIdentifier, 
                "mapping of the UOF image to memory failed!", status);

#if defined (_IX_RM_DEBUG_)

                if(status - UCLO_FAILURE < g_UCLOErrorStringCount)
                    ix_ossl_message_log("UCLO error code is: %s\n", g_aUCLOErrorStrings[status-UCLO_FAILURE]);

#endif /* end defined (_IX_RM_DEBUG_) */

        }
        else
        {
            UcLo_VarMemSeg_T varsMemSegments;
            if((status = UcLo_GetVarMemSegs(g_MicroengineAPIManager.m_UOFActiveHandle, &varsMemSegments)) != UCLO_SUCCESS)
            {
                err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_LOAD_UCODE_FROM_WB_FAILED, s_aFunctionIdentifier, 
                    "failed to retrieve the variables memory segments!", status);

#if defined (_IX_RM_DEBUG_)

                if(status - UCLO_FAILURE < g_UCLOErrorStringCount)
                    ix_ossl_message_log("UCLO error code is: %s\n", g_aUCLOErrorStrings[status-UCLO_FAILURE]);

#endif /* end defined (_IX_RM_DEBUG_) */

            }
            else
            {
                ix_uint32 segmentReserved;

                if(g_pReservedMemoryAreas != 0)
                {
                    ix_memory_reserved_area* pReservedArea;

                    for(;;)
                    {
                        segmentReserved = 1;

                        /* SRAM channel 0 */
                        if(varsMemSegments.sram0Size != 0)
                        {
                            for(segmentReserved = 0, pReservedArea = g_pReservedMemoryAreas; pReservedArea->m_Size != 0; pReservedArea++)
                            {
                                if((pReservedArea->m_Type == IX_MEMORY_TYPE_SRAM) && (pReservedArea->m_Channel == 0))
                                {
                                    if((varsMemSegments.sram0BaseAddr >= pReservedArea->m_StartOffset) && 
                                        (varsMemSegments.sram0BaseAddr + varsMemSegments.sram0Size <= 
                                        pReservedArea->m_StartOffset + pReservedArea->m_Size))
                                    {
                                        segmentReserved = 1;
                                        break;
                                    }
                                }
                            }
                            if(segmentReserved == 0)
                                break;
                        }

                        /* SRAM channel 1 */
                        if(varsMemSegments.sram1Size != 0)
                        {
                            for(segmentReserved = 0, pReservedArea = g_pReservedMemoryAreas; pReservedArea->m_Size != 0; pReservedArea++)
                            {
                                if((pReservedArea->m_Type == IX_MEMORY_TYPE_SRAM) && (pReservedArea->m_Channel == 1))
                                {
                                    if((varsMemSegments.sram1BaseAddr >= pReservedArea->m_StartOffset) && 
                                        (varsMemSegments.sram1BaseAddr + varsMemSegments.sram1Size <= 
                                        pReservedArea->m_StartOffset + pReservedArea->m_Size))
                                    {
                                        segmentReserved = 1;
                                        break;
                                    }
                                }
                            }
                            if(segmentReserved == 0)
                                break;
                        }

                        /* SRAM channel 2 */
                        if(varsMemSegments.sram2Size != 0)
                        {
                            for(segmentReserved = 0, pReservedArea = g_pReservedMemoryAreas; pReservedArea->m_Size != 0; pReservedArea++)
                            {
                                if((pReservedArea->m_Type == IX_MEMORY_TYPE_SRAM) && (pReservedArea->m_Channel == 2))
                                {
                                    if((varsMemSegments.sram2BaseAddr >= pReservedArea->m_StartOffset) && 
                                        (varsMemSegments.sram2BaseAddr + varsMemSegments.sram2Size <= 
                                        pReservedArea->m_StartOffset + pReservedArea->m_Size))
                                    {
                                        segmentReserved = 1;
                                        break;
                                    }
                                }
                            }
                            if(segmentReserved == 0)
                                break;
                        }

                        /* SRAM channel 3 */
                        if(varsMemSegments.sram3Size != 0)
                        {
                            for(segmentReserved = 0, pReservedArea = g_pReservedMemoryAreas; pReservedArea->m_Size != 0; pReservedArea++)
                            {
                                if((pReservedArea->m_Type == IX_MEMORY_TYPE_SRAM) && (pReservedArea->m_Channel == 3))
                                {
                                    if((varsMemSegments.sram3BaseAddr >= pReservedArea->m_StartOffset) && 
                                        (varsMemSegments.sram3BaseAddr + varsMemSegments.sram3Size <= 
                                        pReservedArea->m_StartOffset + pReservedArea->m_Size))
                                    {
                                        segmentReserved = 1;
                                        break;
                                    }
                                }
                            }
                            if(segmentReserved == 0)
                                break;
                        }

                        /* DRAM channel 0 */
                        if(varsMemSegments.sdramSize != 0)
                        {
                            for(segmentReserved = 0, pReservedArea = g_pReservedMemoryAreas; pReservedArea->m_Size != 0; pReservedArea++)
                            {
                                if((pReservedArea->m_Type == IX_MEMORY_TYPE_DRAM) && (pReservedArea->m_Channel == 0))
                                {
                                    if((varsMemSegments.sdramBaseAddr >= pReservedArea->m_StartOffset) && 
                                        (varsMemSegments.sdramBaseAddr + varsMemSegments.sdramSize <= 
                                        pReservedArea->m_StartOffset + pReservedArea->m_Size))
                                    {
                                        segmentReserved = 1;
                                        break;
                                    }
                                }
                            }
                            if(segmentReserved == 0)
                                break;
                        }

                        /* SCRATCH channel 0 */
                        if(varsMemSegments.scratchSize != 0)
                        {
                            for(segmentReserved = 0, pReservedArea = g_pReservedMemoryAreas; pReservedArea->m_Size != 0; pReservedArea++)
                            {
                                if((pReservedArea->m_Type == IX_MEMORY_TYPE_SCRATCH) && (pReservedArea->m_Channel == 0))
                                {
                                    if((varsMemSegments.scratchBaseAddr >= pReservedArea->m_StartOffset) && 
                                        (varsMemSegments.scratchBaseAddr + varsMemSegments.scratchSize <= 
                                        pReservedArea->m_StartOffset + pReservedArea->m_Size))
                                    {
                                        segmentReserved = 1;
                                        break;
                                    }
                                }
                            }
                            if(segmentReserved == 0)
                                break;
                        }

                        break;
                    }
                }
                else
                {
                    if((varsMemSegments.sram0Size != 0) || (varsMemSegments.sram1Size != 0) || 
                        (varsMemSegments.sram2Size != 0) || (varsMemSegments.sram3Size != 0) || 
                        (varsMemSegments.sdramSize != 0) || (varsMemSegments.scratchSize != 0))
                        segmentReserved = 0;
                    else
                        segmentReserved = 1;
                }

                if(segmentReserved == 0)
                {
                    ix_ossl_message_log("\nMicrocode variables use memory segments that are not prereserved!\n");
                    ix_ossl_message_log("You will have to specify the following memory regions to be prereserved\n"
                        "at the RM initialization time (i.e. the call to ix_rm_init())!\n\n");

                    ix_ossl_message_log("Microcode variables memory segments.\n");

                    if(varsMemSegments.sram0Size != 0)
                        ix_ossl_message_log("\t SRAM Channel 0: Base offset: 0x%x Size: %u\n", varsMemSegments.sram0BaseAddr, varsMemSegments.sram0Size);

                    if(varsMemSegments.sram1Size != 0)
                        ix_ossl_message_log("\t SRAM Channel 1: Base offset: 0x%x Size: %u\n", varsMemSegments.sram1BaseAddr, varsMemSegments.sram1Size);
                    
                    if(varsMemSegments.sram2Size != 0)
                        ix_ossl_message_log("\t SRAM Channel 2: Base offset: 0x%x Size: %u\n", varsMemSegments.sram2BaseAddr, varsMemSegments.sram2Size);
                    
                    if(varsMemSegments.sram3Size != 0)
                        ix_ossl_message_log("\t SRAM Channel 3: Base offset: 0x%x Size: %u\n", varsMemSegments.sram3BaseAddr, varsMemSegments.sram3Size);

                    if(varsMemSegments.sdramSize != 0)
                        ix_ossl_message_log("\t DRAM: Base offset: 0x%x Size: %u\n", varsMemSegments.sdramBaseAddr, varsMemSegments.sdramSize);
                    
                    if(varsMemSegments.scratchSize != 0)
                        ix_ossl_message_log("\t SCRATCH: Base offset: 0x%x Size: %u\n", varsMemSegments.scratchBaseAddr, varsMemSegments.scratchSize);

                    err = IX_RM_ERROR_NEW(IX_RM_ERROR_ME_LOAD_UCODE_FROM_WB_FAILED, s_aFunctionIdentifier, 
                        "microcode needs memory areas that are not prereserved!", status);
                }
                else
                {
                    g_MicroengineAPIManager.m_IsUOFLoadedFromFile = 0;
                    g_MicroengineAPIManager.m_IsUOFLoadedFromWorkbench = 1;
                }
            }
        }
    }
#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

#if defined(_IX_RM_IMPL_HARDWARE_)

label_1:

#endif /* end defined(_IX_RM_IMPL_HARDWARE_) */

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}

#endif /* defined(_IX_RM_WORKBENCH_SUPPORT_) */

