/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      remote_communication.c
 *
 * = DESCRIPTION
 *      This file defines all variables and functions ezported by the remote
 *      communication API.
 *
 * = AUTHOR
 *      Adrian Georgescu
 *      adrian.georgescu@intel.com
 *
 * = CHANGE HISTORY
 *      12/2/2002 11:40:57 AM - creation time 
 * ============================================================================
 */

static const char s_FileIdentifier[] = "$Id: remote_communication.c,v 1.2 2003/03/04 19:32:11 ageorges Exp $";


#include "ix_ossl.h"

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"

#include "rm/symbols.h"
#include "rm/internal/internal_support.h"
#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"
#include "rm/system.h"
#include "rm/internal/internal_system.h"

#include "rm/memory.h"

#include "rm/buffer.h"
#include "rm/internal/internal_buffer.h"
#include "rm/hardware.h"
#include "rm/communication.h"
#include "rm/internal/internal_communication.h"

#include "rm/internal/internal_pci_communication.h"

#include "rm/remote_communication.h"
#include "rm/internal/internal_remote_communication.h"



/**
 * NAME: ix_rm_remote_comm_service_register
 *
 * DESCRIPTION: This function will register a remote communication service if no other is
 *  already installed. The one in place should be first unregistered, and after that try
 *  to register a new one. If the registration is possible then the initializer of the new
 *  service is invoked, and if that call is successful, then the registration will be completed.
 * 
 * @Param:  - IN arg_pRemoteCommService - the new remote comm service to be installed.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_remote_comm_service_register(
                                             const ix_remote_comm_service* arg_pRemoteCommService
                                           )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_remote_comm_service_register";

    ix_error err = IX_SUCCESS;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_RemoteCommManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    IX_RM_ASSERT((arg_pRemoteCommService->m_ServiceInitializer != 0) &&
        (arg_pRemoteCommService->m_ServiceFinalizer != 0) &&
        (arg_pRemoteCommService->m_PacketHandler != 0) &&
        (arg_pRemoteCommService->m_MessageHandler != 0)
        , s_aFunctionIdentifier, "NULL functions passed for the remote service!");

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    if(g_RemoteCommManager.m_RemoteCommService.m_ServiceInitializer != g_DefaultRemoteService.m_ServiceInitializer)
    {
        err = IX_RM_ERROR_NEW(IX_RM_ERROR_REMOTE_COMM_SERVICE_ACTIVE, s_aFunctionIdentifier, 
                        "Another remote communication service is active!", err);
        goto label_1;
    }
    
    /* initialize the new remote communication service */
    err = arg_pRemoteCommService->m_ServiceInitializer(arg_pRemoteCommService->m_pServiceContext);
    if(err != IX_SUCCESS)
    {
        err = IX_RM_ERROR_CHAIN(err, s_aFunctionIdentifier, "Remote service initializer failed!");
        goto label_1;
    }

    /* Install the new remote comm service */
    g_RemoteCommManager.m_RemoteCommService.m_ServiceInitializer = arg_pRemoteCommService->m_ServiceInitializer;
    g_RemoteCommManager.m_RemoteCommService.m_ServiceFinalizer = arg_pRemoteCommService->m_ServiceFinalizer;
    g_RemoteCommManager.m_RemoteCommService.m_pServiceContext = arg_pRemoteCommService->m_pServiceContext;
    g_RemoteCommManager.m_RemoteCommService.m_PacketHandler = arg_pRemoteCommService->m_PacketHandler;
    g_RemoteCommManager.m_RemoteCommService.m_pPacketHandlerContext = arg_pRemoteCommService->m_pPacketHandlerContext;
    g_RemoteCommManager.m_RemoteCommService.m_MessageHandler = arg_pRemoteCommService->m_MessageHandler;
    g_RemoteCommManager.m_RemoteCommService.m_pMessageHandlerContext = arg_pRemoteCommService->m_pMessageHandlerContext;

    {
        ix_end_point* pEndPoint;

        /* initialize the packet end point */
        pEndPoint = &g_RemoteCommManager.m_RemoteCommPacketEndPoint;

        /* initialize the remote packet end point processing */
        pEndPoint->m_EPBehaviour.m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)_ix_rm_remote_comm_default_packet_handler;
        pEndPoint->m_EPBehaviour.m_pExternalBehaviourContext = arg_pRemoteCommService->m_pPacketHandlerContext;
        pEndPoint->m_EPBehaviour.m_pInternalBehaviourContext = pEndPoint;
        pEndPoint->m_EPBehaviour.m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)arg_pRemoteCommService->m_PacketHandler;

        /* initialize the message end point */
        pEndPoint = &g_RemoteCommManager.m_RemoteCommMessageEndPoint;

        /* initialize the remote message end point processing */
        pEndPoint->m_EPBehaviour.m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)_ix_rm_remote_comm_default_message_handler;
        pEndPoint->m_EPBehaviour.m_pExternalBehaviourContext = arg_pRemoteCommService->m_pMessageHandlerContext;
        pEndPoint->m_EPBehaviour.m_pInternalBehaviourContext = pEndPoint;
        pEndPoint->m_EPBehaviour.m_BehaviourHandler = (_ix_comm_ep_behaviour_handler)arg_pRemoteCommService->m_MessageHandler;
    }

label_1:

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}




/**
 * NAME: ix_rm_remote_comm_service_unregister
 *
 * DESCRIPTION: This function will unregister the currently registered remote communication
 *  service. At the time of this call the finalizer of the existing remote comm service will
 *  be invoked.
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_remote_comm_service_unregister(void)
{
    static const char s_aFunctionIdentifier[] = "ix_rm_remote_comm_service_unregister";

    ix_error err = IX_SUCCESS;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_RemoteCommManager.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined( _IX_RM_MULTI_THREADED_ )

    /* lock the manager's data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    /* finalize the existing remote communication service */
    err = g_RemoteCommManager.m_RemoteCommService.m_ServiceFinalizer(g_RemoteCommManager.m_RemoteCommService.m_pServiceContext);
    if(err != IX_SUCCESS)
        err = IX_RM_ERROR_CHAIN(err, s_aFunctionIdentifier, "Remote service finalizer failed!");

    /* Install the default remote comm service */
    g_RemoteCommManager.m_RemoteCommService.m_ServiceInitializer = g_DefaultRemoteService.m_ServiceInitializer;
    g_RemoteCommManager.m_RemoteCommService.m_ServiceFinalizer = g_DefaultRemoteService.m_ServiceFinalizer;
    g_RemoteCommManager.m_RemoteCommService.m_pServiceContext = g_DefaultRemoteService.m_pServiceContext;
    g_RemoteCommManager.m_RemoteCommService.m_PacketHandler = g_DefaultRemoteService.m_PacketHandler;
    g_RemoteCommManager.m_RemoteCommService.m_pPacketHandlerContext = g_DefaultRemoteService.m_pPacketHandlerContext;
    g_RemoteCommManager.m_RemoteCommService.m_MessageHandler = g_DefaultRemoteService.m_MessageHandler;
    g_RemoteCommManager.m_RemoteCommService.m_pMessageHandlerContext = g_DefaultRemoteService.m_pMessageHandlerContext;

#if defined( _IX_RM_MULTI_THREADED_ )

    /* unlock the manager's data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    return err;
}


/**
 * NAME: ix_rm_init_pci_remote_communication
 *
 * DESCRIPTION: This function will install the PCI remote communication service 
 *          between ingress and egress
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_init_pci_remote_communication(void)
{
    static const char s_aFunctionIdentifier[] = "ix_rm_init_pci_remote_communication";

    ix_error err = IX_SUCCESS;

#if defined(_IX_COMPILE_PCI_COMM_IN_) && defined(_IX_RM_IMPL_HARDWARE_)

    ix_remote_comm_service pciRemoteService;

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    pciRemoteService.m_ServiceInitializer = _ix_rm_pci_communication_init;
    pciRemoteService.m_ServiceFinalizer = _ix_rm_pci_communication_fini;
    pciRemoteService.m_pServiceContext = 0;

    pciRemoteService.m_PacketHandler = (ix_remote_comm_data_handler)_ix_rm_comm_pci_data_store_mp;
    pciRemoteService.m_pPacketHandlerContext = &g_RemoteCommManager.m_RemoteCommPacketEndPoint.m_EPData.m_CoreCompData;

    pciRemoteService.m_MessageHandler = (ix_remote_comm_data_handler)_ix_rm_comm_pci_data_store_mp;
    pciRemoteService.m_pMessageHandlerContext = &g_RemoteCommManager.m_RemoteCommMessageEndPoint.m_EPData.m_CoreCompData;

    err = ix_rm_remote_comm_service_register(&pciRemoteService);
    if(err != IX_SUCCESS)
        err = IX_RM_ERROR_CHAIN(err, s_aFunctionIdentifier, "Initialization of PCI remote service failed!");


#else

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#endif /* end defined(_IX_COMPILE_PCI_COMM_IN_) */

    return err;
}

/**
 * NAME: ix_rm_register_pci_communication_hw_free_list
 *
 * DESCRIPTION: This function will register a hardware free list to be used by the receive
 *      side of the PCI remote communication driver instead of the default hardware free
 *      list. The PCI driver automatically creates an internal free list that is used for 
 *      copying the buffers from the remote subsystem. This free list can be replaced by
 *      a free list of choice.
 * 
 * @Param:  - IN arg_hFreeList - the hardware free list to be used by the PCI remote 
 *      communication driver.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_register_pci_communication_hw_free_list(
                                                        ix_buffer_free_list_handle arg_hFreeList
                                                      )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_register_pci_communication_hw_free_list";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_COMPILE_PCI_COMM_IN_) && defined(_IX_RM_IMPL_HARDWARE_)

    g_PCICommunicationController.m_hUserHwFreeList = arg_hFreeList;

#else


#endif /* end defined(_IX_COMPILE_PCI_COMM_IN_) */

    return IX_SUCCESS;
}




/**
 * NAME: ix_rm_unregister_pci_communication_hw_free_list
 *
 * DESCRIPTION: This function will unregister the previously registered user hardware
 *      free list for the PCI remote communication driver. From this point on the default
 *      (PCI driver internal one) hardware free list will be used to allocate the required
 *      buffers.
 * 
 * @Param:  - OUT  arg_pFreeListHandle - if successful the previously registered user
 *      hardware free list handle will be returned to the caller at the passed address.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error token for failure.
 */
ix_error ix_rm_unregister_pci_communication_hw_free_list(
                                                          ix_buffer_free_list_handle* arg_pFreeListHandle
                                                        )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_unregister_pci_communication_hw_free_list";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_COMPILE_PCI_COMM_IN_) && defined(_IX_RM_IMPL_HARDWARE_)

    g_PCICommunicationController.m_hUserHwFreeList = 0;

#else


#endif /* end defined(_IX_COMPILE_PCI_COMM_IN_) */

    return IX_SUCCESS;
}
