/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 4
 *
 * = LIBRARY
 *      Resource Manager Library
 *
 * = MODULE
 *      System API
 *
 * = FILENAME
 *      system.c
 *
 * = DESCRIPTION
 *      This file will define all functions of the exported System API interface. 
 *
 * = AUTHOR
 *       Adrian Georgescu
 *       adrian.georgescu@intel.com
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      10/23/2001 3:33:51 PM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 */

static const char s_aFileIdentifier[] = "$Id: system.c,v 1.43 2003/10/10 21:09:31 ageorges Exp $";

#include "ix_ossl.h"

#include "ix_os_type.h"
#include "ix_symbols.h"
#include "ix_hardware_configuration.h"

#include "ix_types.h"
#include "ix_macros.h"
#include "ix_error.h"


#include "rm/symbols.h"
#include "rm/internal/internal_support.h"
#include "rm/internal/internal_version.h"

#include "rm/sync_mechanism.h"
#include "rm/internal/internal_sync.h"

#include "rm/system.h"
#include "rm/internal/internal_system.h"

#include "rm/memory.h"

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

#include "rm/buffer.h"

#include "rm/hardware.h"

#include "rm/microengine.h"

#include "rm/communication.h"

#include "rm/configuration_property.h"

#include "rm/counter_64bit.h"

#include "rm/hash.h"

#include "rm/internal/internal_buffer.h"

#include "rm/internal/internal_hardware.h"

#include "rm/internal/internal_communication.h"

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */


#include "rm/internal/internal_memory.h"


#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

#include "rm/internal/internal_configuration_property.h"

#include "rm/internal/internal_microengine.h"

#include "rm/internal/internal_counter_64bit.h"

#include "rm/internal/internal_hash.h"

#include "rm/me_xscale_locks.h"

#include "rm/internal/internal_me_xscale_locks.h"

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

/**
 * Global reference to the prereserved memory areas
 */
ix_memory_reserved_area* g_pReservedMemoryAreas = 0;


#if defined( _IX_RM_MULTI_THREADED_ )

/**
 * Synchronization semaphore for system API
 */
static ix_ossl_sem_t s_RMSystemSemaphore = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

/**
 * NAME: ix_rm_version_get_string
 *
 * DESCRIPTION: This function will return the product's version string
 * 
 * @Return: Returns a string representing the product version string.
 */
const char* ix_rm_version_get_string(void)
{
    static const char s_aFunctionIdentifier[] = "ix_rm_version_get_string";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    if(g_pVersionSchemeString)
        return g_pVersionSchemeString;
    else
        return "The library has to be first initialized!\n";
}

/**
 * NAME: ix_rm_error_get_string
 *
 * DESCRIPTION: This is a support function that retrieves a string represenation of internal
 *          error codes. Internal error codes are passed within ix_error return value for all
 *          interface functions. This function can be used to get a string representation of
 *          that error code.
 * 
 * @Param:  - IN arg_Error - an internal error code for which we want a string representation.
 *
 * @Return: The string representation corresponding to the error code in cause.
 */
const char* ix_rm_error_get_string( 
                                    ix_rm_error_code arg_Error 
                                  )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_error_get_string";

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    switch(arg_Error)
    {
        case IX_RM_ERROR_SUCCESS:
            return "Success";
        /* assertion error code */
        case IX_RM_ERROR_ASSERTION:
            return "Assertion failed";

        /* memory manager error codes */
        case IX_RM_ERROR_MEMORY_INIT_FAILED:
            return "Memory API initialization failed";
        case IX_RM_ERROR_MEMORY_FINI_FAILED:
            return "Memory API finalization failed";
        case IX_RM_ERROR_BAD_MEMORY_TYPE:
            return "Inappropriate memory type in this context";
        case IX_RM_ERROR_CANNOT_CREATE_CELL:
            return "Cannot allocate internal memory block information structure";
        case IX_RM_ERROR_DATA_ALREADY_RELEASED:
            return "Memory has been already released";
        case IX_RM_ERROR_BAD_MEMORY_ADDRESS:
            return "The passed memory address is not in the Memory API range";
        case IX_RM_ERROR_MEMORY_IN_USE:
            return "The requested memory is partially in use";
        case IX_RM_ERROR_NULL_MEMORY_ADDRESS:
            return "Null memory address";
        case IX_RM_ERROR_CANNOT_ALLOCATE_SIZE:
            return "The allocation of the requested memory size cannot be satisfied";
        case IX_RM_ERROR_CANNOT_RESERVE_MEMORY:
            return "The requested amount of memory acnnot be reserved";
        case IX_RM_ERROR_CELL_SPLIT_FAILED:
            return "Internal memory block plit failed";
        case IX_RM_ERROR_MEMORY_BAD_OFFSET:
            return "The passed offset is not in the controlled memory block";

        /* general library error codes */
        case IX_RM_ERROR_LIB_INIT_FAILED:
            return "Failed to initialize the RM library";
        case IX_RM_ERROR_LIB_FINI_FAILED:
            return "Failed to finalize the RM library";

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

        /* communication API error codes */
        case IX_RM_ERROR_PKT_DISP_ENG_INIT_FAILED:
            return "Initialization of packet dispatch engine failed";
        case IX_RM_ERROR_MSG_DISP_ENG_INIT_FAILED:
            return "Initialization of message dispatch engine failed";
        case IX_RM_ERROR_PKT_DISP_ENG_FINI_FAILED:
            return "Finalization of packet dispatch engine failed";
        case IX_RM_ERROR_MSG_DISP_ENG_FINI_FAILED:
            return "Finalization of message dispatch engine failed";
        case IX_RM_ERROR_PKT_DISP_ENG_START_FAILED:
            return "Failed to start the packet dispatch engine";
        case IX_RM_ERROR_MSG_DISP_ENG_START_FAILED:
            return "Failed to start the message dispatch engine";
        case IX_RM_ERROR_PKT_DISP_ENG_STOP_FAILED:
            return "Failed to stop the packet dispatch engine";
        case IX_RM_ERROR_MSG_DISP_ENG_STOP_FAILED:
            return "Failed to stop the message dispatch engine";

        case IX_RM_ERROR_COMM_API_INIT_FAILED:
            return "Communication API initialization failed";
        case IX_RM_ERROR_COMM_API_FINI_FAILED:
            return "Communication API finalization failed";
        case IX_RM_ERROR_COMM_QUEUE_FULL:
            return "Communication ID internal queue is full";
        case IX_RM_ERROR_COMM_QUEUE_EMPTY:
            return "Communication ID internal queue is empty";
        case IX_RM_ERROR_COMM_NOTIFICATION_ELEM_ALLOCATION_FAILED:
            return "Failed to allocate a communication ID notification element";
        case IX_RM_ERROR_COMM_SEM_POOL_INIT_FAILED:
            return "Communication API semaphore pool initialization failed";
        case IX_RM_ERROR_COMM_SEM_NODE_ALLOCATION_FAILED:
            return "Failed to allocate a semaphore pool node";
        case IX_RM_ERROR_COMM_SEM_GET_FAILED:
            return "Failed to get a new notifiaction semaphore";
        case IX_RM_ERROR_COMM_ISR_SETUP_FAILED:
            return "Failed to setup the communication ISRs";
        case IX_RM_ERROR_COMM_DISPATCH_TH_INIT_FAILED:
            return "Failed to initialize one the dispatch thread engine";
        case IX_RM_ERROR_COMM_EP_DATA_ALLOCATION_FAILED:
            return "Failed to allocate the memory for the internal communication ID queue";
        case IX_RM_ERROR_COMM_EP_SEMA_CREATION_FAILED:
            return "Failed to create the end point synchronization semaphore";
        case IX_RM_ERROR_COMM_EP_MGR_INIT_FAILED:
            return "Initialization of the end point manager failed";
        case IX_RM_ERROR_COMM_EP_MGR_FINI_FAILED:
            return "Finalization of the end point manager failed";
        case IX_RM_ERROR_COMM_EP_PCI_PROXIES_INIT_FAILED:
            return "Initialization of the PCI end point proxies failed";
        case IX_RM_ERROR_COMM_EP_PCI_PROXIES_FINI_FAILED:
            return "Finalization of the PCI end point proxies failed";
        case IX_RM_ERROR_COMM_EP_REMOTE:
            return "The communication ID is remote";
        case IX_RM_ERROR_COMM_EP_UBLOCK:
            return "The communication ID coresponds to a microblock";
        case IX_RM_ERROR_COMM_EP_CC:
            return "The communication ID coresponds to a core component";
        case IX_RM_ERROR_COMM_EP_RECV_CALLBACK:
            return "The communication ID is in callback receive mode";
        case IX_RM_ERROR_COMM_EP_RECV_GET_SELECT:
            return "he communication ID is in get/select receive mode";
        case IX_RM_ERROR_COMM_EP_MULTIPLE_CONSUMER:
            return "The communication ID is in multiple consumer mode";
        case IX_RM_ERROR_COMM_EP_CC_MODE_ENABLED:
            return "The communication ID has the core component mode enabled";
        case IX_RM_ERROR_COMM_EP_CC_MODE_DISABLED:
            return "The communication ID has the core component mode disabled";

        case IX_RM_ERROR_REMOTE_COMM_API_INIT_FAILED:
            return "Remote communication API initialization failed";
        case IX_RM_ERROR_REMOTE_COMM_API_FINI_FAILED:
            return "Remote communication API finalization failed";
        case IX_RM_ERROR_REMOTE_COMM_SERVICE_ACTIVE:
            return "Another remote communication service is active";

        case IX_RM_ERROR_COMM_PCI_INIT_FAILED:
            return "Initialization of the PCI communication failed";
        case IX_RM_ERROR_COMM_PCI_FINI_FAILED:
            return "Finalization of the PCI communication failed";
        case IX_RM_ERROR_COMM_PCI_GET_REMOTE_DEV_FAILED:
            return "Cannot retrieve the remote device";
	    case IX_RM_ERROR_COMM_PCI_TX_INIT_FAILED:
            return "Initialization of the PCI transmit engine failed";
        case IX_RM_ERROR_COMM_PCI_RX_INIT_FAILED:
            return "Initialization of the PCI receive engine failed";
        case IX_RM_ERROR_COMM_PCI_TX_START_FAILED:
            return "Failed to start the PCI transmit engine";    
        case IX_RM_ERROR_COMM_PCI_RX_START_FAILED:
            return "Failed to start the PCI receive engine";
        case IX_RM_ERROR_COMM_PCI_TX_STOP_FAILED:
            return "Failed to stop the PCI transmit engine";
        case IX_RM_ERROR_COMM_PCI_RX_STOP_FAILED:
            return "Failed to stop the PCI receive engine";
        case IX_RM_ERROR_COMM_PCI_RX_FINI_FAILED:
            return "Finalization of the PCI transmit engine failed";
        case IX_RM_ERROR_COMM_PCI_TX_FINI_FAILED:
            return "Finalization of the PCI receive engine failed";
	    case IX_RM_ERROR_COMM_PCI_INT_CONNECT_FAILED:
            return "Failed to register the PCI communication interrupt ISR";
	    case IX_RM_ERROR_COMM_PCI_INT_ENABLE_FAILED:
            return "Failed to enable the PCI communication interrupt";
	    case IX_RM_ERROR_COMM_PCI_INT_DISABLE_FAILED:
            return "Failed to disable the PCI communication interrupt";
        case IX_RM_ERROR_COMM_PCI_RX_TH_INIT_FAILED:
            return "Failed to initialize the PCI transmit thread";
        case IX_RM_ERROR_COMM_PCI_TX_TH_INIT_FAILED:
            return "Failed to initialize the PCI receive thread";
        case IX_RM_ERROR_COMM_PCI_RX_SEMA_CREATION_FAILED:
            return "Failed to create the PCI receive semaphore";
        case IX_RM_ERROR_COMM_PCI_TX_SEMA_CREATION_FAILED:
            return "Failed to create the PCI transmit semaphore";
        case IX_RM_ERROR_COMM_PCI_TX_CTRL_INFO_ALLOC_FAILED:
            return "Failed to allocate the PCI transmit control information";
        case IX_RM_ERROR_COMM_PCI_TX_CTRL_INFO_GET_PHYS_OFFSET_FAILED:
            return "Failed to retrieve the PCI transmit control information offset";
        case IX_RM_ERROR_COMM_PCI_RX_HW_FL_CREATE_FAILED:
            return "Failed to create PCI receive hardware free list";
        case IX_RM_ERROR_COMM_PCI_RX_SW_FL_CREATE_FAILED:
            return "Failed to create PCI receive software free list";
	    case IX_RM_ERROR_COMM_PCI_TX_CTRL_INFO_FREE_FAILED:
            return "Failed to release PCI transmit control information";
	    case IX_RM_ERROR_COMM_PCI_TX_SEMA_FINI_FAILED:
            return "Failed to finalize PCI transmit semaphore";
	    case IX_RM_ERROR_COMM_PCI_RX_SW_FL_DELETE_FAILED:
            return "Failed to release PCI receive software free list";
	    case IX_RM_ERROR_COMM_PCI_RX_HW_FL_DELETE_FAILED:
            return "Failed to release PCI receive hardware free list";
	    case IX_RM_ERROR_COMM_PCI_RX_SEMA_FINI_FAILED:
            return "Failed to finalize PCI receive semaphore";
	    case IX_RM_ERROR_COMM_PCI_UNSUPPORTED_MEMORY_TYPE:
            return "Unsupported memory type for the PCI communication received buffer";

        /* hardware API error codes */
        case IX_RM_ERROR_HW_API_INIT_FAILED:
            return "Initialization of hardware API failed";
        case IX_RM_ERROR_HW_API_FINI_FAILED:
            return "Finalization of hardware API failed";
        case IX_RM_ERROR_HW_QARRAY_MGR_ALLOC_FAILED:
            return "Allocation of the QArray manager failed and the manager is not intialized";
        case IX_RM_ERROR_HW_QARRAY_MGR_DEALLOC_FAILED:
            return "Deallocation of the QArray manager failed";
        case IX_RM_ERROR_HW_QUEUE_CREATE_FAILED:
            return "Failed to create hardware queue";
        case IX_RM_ERROR_HW_RING_CREATE_FAILED:
            return "Failed to create hardware ring";
        case IX_RM_ERROR_HW_RING_DELETE_FAILED:
            return "Failed to release hardware ring";
        case IX_RM_ERROR_HW_SRAM_RING_PUT_FAILED:
            return "SRAM ring put operation failed";
        case IX_RM_ERROR_HW_SCRATCH_RING_PUT_FAILED:
            return "SCRATCH ring put operation failed";
        case IX_RM_ERROR_HW_SRAM_RING_GET_FAILED:
            return "SRAM ring get operation failed";
        case IX_RM_ERROR_HW_SCRATCH_RING_GET_FAILED:
            return "SCRATCH ring get operation failed";

        /* buffer API error codes */
        case IX_RM_ERROR_BUF_API_INIT_FAILED:
            return "Initialization of buffer API failed";
        case IX_RM_ERROR_BUF_API_FINI_FAILED:
            return "Finalization of buffer API failed";
        case IX_RM_ERROR_BUF_QUEUES_ALLOC_FAILED:
            return "Alocation of hardware queues for buffer free lists failed";
        case IX_RM_ERROR_BUF_HANDLE_INVALID:
            return "Invalid buffer handle";
        case IX_RM_ERROR_BUF_FREE_LIST_INVALID:
            return "Invalid free list handle";
        case IX_RM_ERROR_BUF_FREE_LIST_ID_INVALID:
            return "Invalid free list ID";
        case IX_RM_ERROR_BUF_FREE_LIST_BAD_SIZE:
            return "Bad free list size";
        case IX_RM_ERROR_BUF_FREE_LIST_SIZE_TOO_LARGE:
            return "Free list size too large";
        case IX_RM_ERROR_BUF_FREE_LIST_BAD_META_SIZE:
            return "Bad meta size";
        case IX_RM_ERROR_BUF_FREE_LIST_ALL_IN_USE:
            return "All free lists are in use";
        case IX_RM_ERROR_BUF_FREE_LIST_META_ALLOCATION_FAILED:
            return "Free list meta memory block allocation failed";

#if defined(_IX_RM_SPLIT_META_DATA_)

        case IX_RM_ERROR_BUF_FREE_LIST_INTERNAL_META_ALLOCATION_FAILED:
            return "Free list internal meta memory block allocation failed";

#endif /* defined(_IX_RM_SPLIT_META_DATA_) */

#if defined (_IX_RM_BUFFER_DEBUG_)

        case IX_RM_ERROR_BUF_FREE_LIST_DEBUG_INFO_ALLOCATION_FAILED:
            return "Free list debug information memory block allocation failed";

#endif /* defined (_IX_RM_BUFFER_DEBUG_) */


        case IX_RM_ERROR_BUF_FREE_LIST_DATA_ALLOCATION_FAILED:
            return "Free list data memory block allocation failed";
        case IX_RM_ERROR_BUF_FREE_LIST_FINI_FAILED:
            return "Free list finalization failed";
        case IX_RM_ERROR_BUF_FREE_LIST_CREATION_FAILED:
            return "Free list creation failed";
        case IX_RM_ERROR_BUF_NO_FREE_BUFFERS:
            return "Free list out of buffers";
        case IX_RM_ERROR_BUF_ALREADY_LINKED:
            return "Buffer already linked";

        /* configuration property API error codes */
        case IX_RM_ERROR_CP_API_INIT_FAILED:
            return "Initialization of configuration property API failed";
        case IX_RM_ERROR_CP_API_FINI_FAILED:
            return "Finalization of configuration property API failed";
        case IX_RM_ERROR_CP_HANDLE_NULL:
            return "Null configuration property handle";
        case IX_RM_ERROR_CP_NAME_TOO_LONG:
            return "Configuration property name too long";
        case IX_RM_ERROR_CP_NAME_NULL:
            return "Null configuration property name";
        case IX_RM_ERROR_CP_BAD_NAME:
            return "Bad configuration property name";
        case IX_RM_ERROR_CP_ALREADY_EXISTS:
            return "A configuration property with this name already exists at this node level";
        case IX_RM_ERROR_CP_CREATION_FAILED:
            return "Configuration property creation failed";
        case IX_RM_ERROR_CP_FAILED_TO_ACCESS_ATTRIBUTES:
            return "Failed to access configuration property attributes";
        case IX_RM_ERROR_CP_CONSTANT_DELETE:
            return "Cannot delete a constant configuration property";
        case IX_RM_ERROR_CP_CONSTANT_SET_VALUE:
            return "Cannot set value for a constant configuration property";
        case IX_RM_ERROR_CP_FAILED_TO_LINK:
            return "Failed to link a configuration property";
        case IX_RM_ERROR_CP_DATA_UINT32:
            return "Cannot change directly the data type from 32 bit to opaque for the configuration property";
        case IX_RM_ERROR_CP_DATA_OPAQUE:
            return "Cannot change directly the data type from opaque to 32 bit for the configuration property";
        case IX_RM_ERROR_CP_BUFFER_GET_NEXT_FAILED:
            return "Retrieveing next configuration property failed";
        case IX_RM_ERROR_CP_BUFFER_TOO_SMALL:
            return "Configuration property internal buffer too small for the data";
        case IX_RM_ERROR_CP_FAILED_TO_ALLOCATE_NOTIFICATION_TOKEN:
            return "Failed to allocate a configuration property notification token";
        case IX_RM_ERROR_CP_NOTIFICATION_ACTIVE:
            return "Operation cannot be performed while the notification is in progress";
        case IX_RM_ERROR_CP_PROPERTY_NOT_FOUND:
            return "Configuration property not found";

        /* 64 bit counters API error codes */
        case IX_RM_ERROR_64BIT_COUNTER_INIT_FAILED:
            return "Initialization of 64 bit counter API failed";
        case IX_RM_ERROR_64BIT_COUNTER_FINI_FAILED:
            return "Finalization of 64 bit counter API failed";
        case IX_RM_ERROR_64BIT_COUNTER_MEMORY_INIT_FAILED:
            return "Initializaton of the 64 bit counter memory failed";
        case IX_RM_ERROR_64BIT_COUNTER_BLOCK_INIT_FAILED:
            return "Initializaton of the 64 bit counter block failed";
        case IX_RM_ERROR_64BIT_COUNTER_ALL_BLOCKS_IN_USE:
            return "All 64 bit counter blocks are in use";
        case IX_RM_ERROR_64BIT_COUNTER_INACTIVE:
            return "64 bit counter is inactive";
        case IX_RM_ERROR_64BIT_COUNTER_HANDLE_INVALID:
            return "Invalid 64 bit counter handle";
        case IX_RM_ERROR_64BIT_COUNTER_UPDATE_TH_INIT_FAILED:
            return "Initialization of 64 bit counter update thread failed";
        case IX_RM_ERROR_64BIT_COUNTER_UPDATE_SEM_INIT_FAILED:
            return "Initialization of 64 bit counter update semaphore failed";
        case IX_RM_ERROR_64BIT_COUNTER_BAD_OVERFLOW_TIME:
            return "Bad 64 bit counter overflow time";
        case IX_RM_ERROR_64BIT_COUNTER_BAD_INTERNAL_MEMORY:
            return "64 bit counter bad type of internal memory";
        case IX_RM_ERROR_64BIT_COUNTER_NOT_ENOUGH_COUNTERS:
            return "Not enough 64 bit counters";
        case IX_RM_ERROR_64BIT_COUNTER_SWAP_MEM_ALLOC_FAILED:
            return "64 bit counter atomic swap failed";

        /* microengine API error codes */
        case IX_RM_ERROR_ME_API_INIT_FAILED:
            return "Initialization of microengine API failed";
        case IX_RM_ERROR_ME_API_FINI_FAILED:
            return "Finalization of microengine API failed";
        case IX_RM_ERROR_ME_MAP_UCODE_FAILED:
            return "Mapping of the UOF code to memory failed";
        case IX_RM_ERROR_ME_LOAD_UCODE_FAILED:
            return "Loading of the UOF code to memory failed";
        case IX_RM_ERROR_ME_LOAD_TO_USTORE_FAILED:
            return "Loading of the UOF code to micro store failed";
        case IX_RM_ERROR_ME_LOAD_UCODE_FROM_WB_FAILED:
            return "Loading of the UOF code from WorkBench failed";
        case IX_RM_ERROR_ME_START_FAILED:
            return "Failed to start the microengine";
        case IX_RM_ERROR_ME_STOP_FAILED:
            return "Failed to stop the microengine";
        case IX_RM_ERROR_ME_RESET_FAILED:
            return "Failed to reset the microengine";
        case IX_RM_ERROR_ME_ENABLE_CTX_FAILED:
            return "Failed to enable microengine's contexts";
        case IX_RM_ERROR_ME_DISABLE_CTX_FAILED:
            return "Failed to disable microengine's contexts";
        case IX_RM_ERROR_ME_REGISTER_INTERNAL_SYMBOLS_FAILED:
            return "Failed to register internal symbols";
        case IX_RM_ERROR_ME_PATCH_SYMBOLS_FAILED:
            return "Failed to patch symbols to micro code";

        /* hash API error codes */
        case IX_RM_ERROR_HASH_API_INIT_FAILED:
            return "Initialization of hash API failed";
        case IX_RM_ERROR_HASH_API_FINI_FAILED:
            return "Finalization of hash API failed";

        /* ME\XScale locks codes */
        case IX_RM_ERROR_ME_XSCALE_LOCK_INIT_FAILED:
            return "Initialization of ME\\XScale locks API failed";
        case IX_RM_ERROR_ME_XSCALE_LOCK_FINI_FAILED:
            return "Finalization of ME\\XScale locks API failed";
        case IX_RM_ERROR_ME_XSCALE_LOCK_MEM_ALLOC_FAILED:
            return "Failed to allocate the SCRATCH memory associate with the ME\\XScale locks";
        case IX_RM_ERROR_ME_XSCALE_LOCK_OUT_OF_LOCKS:
            return "The system is out of ME\\XScale locks";
        case IX_RM_ERROR_ME_XSCALE_LOCK_NOT_ACTIVE:
            return "The ME\\XScale lock is not active";
        case IX_RM_ERROR_ME_XSCALE_LOCK_IS_LOCKED:
            return "The ME\\XScale lock is already locked.";
        case IX_RM_ERROR_ME_XSCALE_LOCK_BAD_MEMORY:
            return "Bad SCRATCH memory location for the ME\\XScale lock";
        case IX_RM_ERROR_ME_XSCALE_LOCK_IS_LOCKED_BY_ME:
            return "The ME\\XScale lock is already locked by MEs.";


#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

        /* unknown error code */
        case IX_RM_ERROR_UNKNOWN:
            return "Unknown";
        default:
            return "Undocumented";
    }
    return 0;
}



/**
 * NAME: ix_rm_init
 *
 * DESCRIPTION: This function initializes the Resource Manager. It should be called before any
 *  other call is made to the Resource Manager. The arg_pReservedAreas parameter is the
 *  beginning of an array with memory areas that need to be prereserved. The end of the array
 *  will be signaled by an entry of size 0. If this parameter is 0 then it signals that there is
 *  no memory area that needs to be prereserved.
 *
 * @Param:  - IN arg_pReservedAreas - points to the beginning of an array of memory areas to be 
 *  prereserved at the initialization time.
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_init(const ix_memory_reserved_area* arg_pReservedAreas)
{
    static const char s_aFunctionIdentifier[] = "ix_rm_init";

    /* Naming convention for automatic variables */
    ix_error err = 0;
    ix_uint32 doInit = 0;
    ix_error internalErr;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = 0;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

    IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);


    /* initialize OSSL logging */
    ix_ossl_message_log_init();

#if defined( _IX_RM_MULTI_THREADED_ )

    pSyncAdapter = &g_RMLibInitialization.m_SyncAdapter;

    if(s_RMSystemSemaphore == 0)
    {
        /* initialize the configuration property manager synchronization */
        err = ix_ossl_sem_init(IX_OSSL_SEM_AVAILABLE, &s_RMSystemSemaphore);
        if(err)
        {
            return IX_RM_ERROR_NEW(IX_RM_ERROR_LIB_INIT_FAILED, s_aFunctionIdentifier,
                            "initialization of system semaphore failed!", err);
        }

        _ix_rm_sa_sync_adapter_init(pSyncAdapter, _ix_rm_sa_default_lock,
            _ix_rm_sa_default_unlock, (void*)&s_RMSystemSemaphore);
    }

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

#if defined(_IX_RM_MULTI_THREADED_)

    /* lock the library initialization data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    /* check and increment the initialization count */
    if(0 == g_RMLibInitialization.m_InitCount++)
    {
        /* we want to do a real init */
        doInit = 1;
    }

#if defined(_IX_RM_MULTI_THREADED_)

    /* unlock the library initialization data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    if(doInit)
    {

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

        ix_uint32 systemInitDone = 0;

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

        ix_uint32 memoryInitDone = 0;

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

        ix_uint32 bufferInitDone = 0;
        ix_uint32 communicationInitDone = 0;
        ix_uint32 hardwareInitDone = 0;
        ix_uint32 configurationPropertyInitDone = 0;
        ix_uint32 microengineInitDone = 0;
        ix_uint32 counter64bitInitDone = 0;
        ix_uint32 hashInitDone = 0;
        ix_uint32 meXScaleLocksInitDone = 0;

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

        /* initialize the prereserved areas address */
        g_pReservedMemoryAreas = 0;

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

        /* initialize the version string */
        _ix_rm_build_version_scheme_string();

        for(;;)
        {

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

            internalErr = _ix_rm_system_init();
            if(internalErr)
                break;
            systemInitDone = 1;

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

            internalErr = _ix_rm_memory_init();
            if(internalErr)
                break;
            memoryInitDone = 1;

            /* reserve some SRAM memory not to interfere with UCLO */
            if(arg_pReservedAreas != 0)
            {
                const ix_memory_reserved_area* pReservedArea;
                ix_memory_reserved_area* pInternalReservedArea;
                ix_uint32 i;

                for(pReservedArea = arg_pReservedAreas, i = 1; pReservedArea->m_Size != 0; i++, pReservedArea++)
                {}

                g_pReservedMemoryAreas = (ix_memory_reserved_area*)ix_ossl_malloc(i * sizeof(ix_memory_reserved_area));
                if(g_pReservedMemoryAreas == 0)
                {
                    internalErr = IX_RM_ERROR_NEW(IX_RM_ERROR_LIB_INIT_FAILED, s_aFunctionIdentifier,
                            "allocation of the reserved memory areas array failed!", 0);
                }
                else
                {
                    ix_ossl_memset(g_pReservedMemoryAreas, 0, i * sizeof(ix_memory_reserved_area));

                    for(pReservedArea = arg_pReservedAreas, pInternalReservedArea = g_pReservedMemoryAreas; pReservedArea->m_Size != 0; pInternalReservedArea++, pReservedArea++)
                    {
                        pInternalReservedArea->m_Channel = pReservedArea->m_Channel;
                        pInternalReservedArea->m_Type = pReservedArea->m_Type;
                        pInternalReservedArea->m_StartOffset = pReservedArea->m_StartOffset;

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

                        internalErr = ix_rm_mem_reserve(pInternalReservedArea->m_Type, pInternalReservedArea->m_Channel,
                            pReservedArea->m_Size, &pInternalReservedArea->m_StartOffset);

#else /* Westport hardware */

                        pInternalReservedArea->m_OffsetType = pReservedArea->m_OffsetType;
                        internalErr = ix_rm_mem_reserve(pInternalReservedArea->m_Type, pInternalReservedArea->m_Channel,
                            pReservedArea->m_Size, pInternalReservedArea->m_OffsetType, &pInternalReservedArea->m_StartOffset);

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

                        if(internalErr != IX_SUCCESS)
                        {
                            internalErr = IX_RM_ERROR_NEW(IX_RM_ERROR_LIB_INIT_FAILED, s_aFunctionIdentifier,
                                "failed to prereserve memory! Check the passed memory areas!", internalErr);
                            break;
                        }

                        /* adjust the size */
                        pInternalReservedArea->m_Size = pReservedArea->m_Size + (pReservedArea->m_StartOffset - pInternalReservedArea->m_StartOffset);

                    }
                }
            }
            if(internalErr)
                break;

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

            internalErr = _ix_rm_hardware_init();
            if(internalErr)
                break;
            hardwareInitDone = 1;

            internalErr = _ix_rm_buffer_init();
            if(internalErr)
                break;
            bufferInitDone = 1;

            internalErr = _ix_rm_communication_init();
            if(internalErr)
                break;
            communicationInitDone = 1;

            internalErr = _ix_rm_configuration_property_init();
            if(internalErr)
                break;
            configurationPropertyInitDone = 1;

            internalErr = _ix_rm_microengine_init();
            if(internalErr)
                break;
            microengineInitDone = 1;

            internalErr = _ix_rm_counter_64bit_init();
            if(internalErr)
                break;
            counter64bitInitDone = 1;

            internalErr = _ix_rm_hash_init();
            if(internalErr)
                break;
            hashInitDone = 1;

            internalErr = _ix_rm_me_xscale_lock_init();
            if(internalErr)
                break;
            meXScaleLocksInitDone = 1;

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

            break;
        }

        /* if an error happened do the cleanup */
        if(internalErr != 0)
        {

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

            if(meXScaleLocksInitDone != 0)
                _ix_rm_me_xscale_lock_fini();

            if(hashInitDone != 0)
                _ix_rm_hash_fini();

            if(counter64bitInitDone != 0)
                _ix_rm_counter_64bit_fini();

            if(microengineInitDone != 0)
                _ix_rm_microengine_fini();

            if(configurationPropertyInitDone != 0)
                _ix_rm_configuration_property_fini();

            if(communicationInitDone != 0)
                _ix_rm_communication_fini();

            if(bufferInitDone != 0)
                _ix_rm_buffer_fini();

            if(hardwareInitDone != 0)
                _ix_rm_hardware_fini();

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

            if(g_pReservedMemoryAreas != 0)
            {
                ix_ossl_free(g_pReservedMemoryAreas);
                g_pReservedMemoryAreas = 0;
            }

            if(memoryInitDone != 0)
                _ix_rm_memory_fini();

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

            if(systemInitDone != 0)
                _ix_rm_system_fini();

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

            err = IX_RM_ERROR_CHAIN(internalErr, s_aFunctionIdentifier, 
                        "Resource Manager library initialization failed!");

#if defined(_IX_RM_MULTI_THREADED_)

            /* lock the library initialization data */
            pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

            /* make the initialization count 0 */
            g_RMLibInitialization.m_InitCount = 0;

#if defined(_IX_RM_MULTI_THREADED_)

            /* unlock the library initialization data */
            pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

            /* release the synchronization semaphore to the system */
            ix_ossl_sem_fini(s_RMSystemSemaphore);

            s_RMSystemSemaphore = 0;

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

        }
    }

    return err;
}



/**
 * NAME: ix_rm_term
 *
 * DESCRIPTION: This call should be called when the Resource Manager services are not needed any
 *          more and it will provide all necessary cleanup.
 * 
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_term(void)
{
    static const char s_aFunctionIdentifier[] = "ix_rm_term";

    /* Naming convention for automatic variables */
    ix_error err = 0;
    ix_uint32 doTerm = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_RMLibInitialization.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

	IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

#if defined(_IX_RM_MULTI_THREADED_)

    /* lock the library finalization data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    /* check if we are not already done */
    if(0 == g_RMLibInitialization.m_InitCount)
        return err;

#if defined(_IX_RM_MULTI_THREADED_)

    /* unlock the library finalization data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    /* decrement and check the initialization count */
    if(0 == --g_RMLibInitialization.m_InitCount)
    {
        /* we want to do a real init */
        doTerm = 1;
    }

    /* check to see if we need to do the cleanup */
    if(doTerm)
    {
        ix_uint32 errorCount = 0;
        ix_error internalError = IX_SUCCESS;

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

        internalError = _ix_rm_me_xscale_lock_fini();
        if(internalError != IX_SUCCESS)
            errorCount++;

        internalError = _ix_rm_hash_fini();
        if(internalError != IX_SUCCESS)
            errorCount++;

        internalError = _ix_rm_counter_64bit_fini();
        if(internalError != IX_SUCCESS)
            errorCount++;

        internalError = _ix_rm_microengine_fini();
        if(internalError != IX_SUCCESS)
            errorCount++;

        internalError = _ix_rm_configuration_property_fini();
        if(internalError != IX_SUCCESS)
            errorCount++;

        internalError = _ix_rm_communication_fini();
        if(internalError != IX_SUCCESS)
            errorCount++;

        internalError = _ix_rm_buffer_fini();
        if(internalError != IX_SUCCESS)
            errorCount++;

        internalError = _ix_rm_hardware_fini();
        if(internalError != IX_SUCCESS)
            errorCount++;

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

        if(g_pReservedMemoryAreas != 0)
        {
            ix_ossl_free(g_pReservedMemoryAreas);
            g_pReservedMemoryAreas = 0;
        }

        internalError = _ix_rm_memory_fini();
        if(internalError != IX_SUCCESS)
            errorCount++;

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

        internalError = _ix_rm_system_fini();
        if(internalError != IX_SUCCESS)
            errorCount++;

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */

#if defined(_IX_RM_MULTI_THREADED_)

        /* release the synchronization semaphore to the system */
        internalError = ix_ossl_sem_fini(s_RMSystemSemaphore);
        if(internalError != IX_SUCCESS)
            errorCount++;

        s_RMSystemSemaphore = 0;

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

        if(errorCount != 0)
        {
            err = IX_RM_ERROR_NEW(IX_RM_ERROR_LIB_FINI_FAILED, s_aFunctionIdentifier, 
                            "one of the Resource Manager finalization steps failed!\n"
                            "The system might be in an inconsistent state!", internalError);
        }


    }

    return err;
}

#if (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) 

/**
 * NAME: ix_rm_sys_config_get
 *
 * DESCRIPTION: This function will retrieve the system specific configuration.  
 * 
 * @Param:  - OUT arg_pSysConfig - this is the location where the system configuration will be written.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_sys_config_get( 
                               ix_sys_config* arg_pSysConfig 
                             )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_sys_config_get";

    /* Naming convention for automatic variables */
    ix_error err = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_SystemConfiguration.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

	IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check the arg_pSysConfig in DEBUG mode */
    IX_RM_ASSERT(arg_pSysConfig, s_aFunctionIdentifier, "the arg_pSysConfig* parameter should be" 
        "non 0 and point to a valid structure!");

#if defined(_IX_RM_MULTI_THREADED_)

    /* lock the system configuration data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    /* copy the system configuartion data */
    ix_ossl_memcpy(arg_pSysConfig, &g_SystemConfiguration.m_SysConfig, sizeof(ix_sys_config));

#if defined(_IX_RM_MULTI_THREADED_)

    /* unlock the system configuration data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    return err;
}



/**
 * NAME: ix_rm_sys_config_set
 *
 * DESCRIPTION: This function will set the system configuration. Only certain members of the
 *          system configuration can be set, the others will be ignored.
 * 
 * @Param:  - IN ix_sys_config - location of the system configuration to be set.
 *
 * @Return: IX_SUCCESS if successful or a valid ix_error for failure.
 */
ix_error ix_rm_sys_config_set( 
                              const ix_sys_config* arg_pSysConfig 
                            )
{
    static const char s_aFunctionIdentifier[] = "ix_rm_sys_config_set";

    /* Naming convention for automatic variables */
    ix_error err = 0;

#if defined( _IX_RM_MULTI_THREADED_ )

    ix_synchronization_adapter* pSyncAdapter = &g_SystemConfiguration.m_SyncAdapter;

#endif /* defined( _IX_RM_MULTI_THREADED_ ) */

	IX_RM_UNUSED_VARIABLE(s_aFunctionIdentifier);

    /* check the arg_pSysConfig in DEBUG mode */
    IX_RM_ASSERT(arg_pSysConfig, s_aFunctionIdentifier, "the ix_sys_config* parameter should be" 
        "non 0 and point to a valid structure!");

#if defined(_IX_RM_MULTI_THREADED_)

    /* lock the system configuration data */
    pSyncAdapter->m_LockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    /* this is the only field that can be set */
    g_SystemConfiguration.m_SysConfig.m_Endianess = arg_pSysConfig->m_Endianess;

#if defined(_IX_RM_MULTI_THREADED_)

    /* unlock the system configuration data */
    pSyncAdapter->m_UnlockFuncPtr(pSyncAdapter->m_pContext);

#endif /* end defined(_IX_RM_MULTI_THREADED_) */

    return err;
}

#endif /* (_IX_HARDWARE_TYPE_ != _IX_HW_WESTPORT_) */




