/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      ix_cc_macros.h
 *
 * = DESCRIPTION
 *      Various macros used in the core components.  This header should be
 *      relatively stand-alone, not relying on other header files in order
 *      to compile.
 *
 * ============================================================================
 * $Id: ix_cc_macros.h,v 1.14 2003/10/02 16:41:53 ktseng Exp $
 */

#if !defined(__IX_CC_MACROS_H__)
#define __IX_CC_MACROS_H__

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
#include <linux/delay.h>
#endif /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

/*
 * If the OPTIONAL flag isn't already turned on, turn it on for DEBUG builds.
 */
#if (!defined(OPTIONAL) && defined(IX_DEBUG))
#define OPTIONAL
#endif /* !OPTIONAL && IX_DEBUG */

/*
 * IX_CC_OPTIONAL provides a simple way to turn on or off a block of code
 * based on the OPTIONAL build flag.  Use it like:
 * IX_CC_OPTIONAL(t = 0;)
 * or
 * IX_CC_OPTIONAL(
 *      i = 0;
 *      s = "";
 * )
 *
 * The lines inside the macro will either be compiled in or not, based on
 * the OPTIONAL build flag.
 */
#ifdef OPTIONAL
#define IX_CC_OPTIONAL(__s) __s
#else
#define IX_CC_OPTIONAL(__s)
#endif

/**
 * In some cases, wrap the statement in a do while zero.  This ensures a
 * statement is not "misused" -- i.e. is only used as a statement, not as an
 * expression.  Remove for TestRT because having the do while affects the line
 * coverage and code complexity scores.  Remove for Release because the macro
 * is only useful at compile-time, to ensure some macros are properly used.
 *
 * This macro should be used to create other macros.
 */
#if !defined(IX_DEBUG) || defined(TESTRT) 
#define IX_CC_WRAP(__s) __s
#else
#define IX_CC_WRAP(__s) do { __s; } while (0)
#endif

/*
 * IX_MIN returns the minimum value of 'a' or 'b'.
 */
#define IX_MIN(a, b) ((a) < (b)) ? (a) : (b)

/*
 * IX_MAX returns the maximum value of 'a' or 'b'.
 */
#define IX_MAX(a, b) ((a) > (b)) ? (a) : (b)

/* Declare private functions as static if debugging is not enabled. */
#if defined (IX_DEBUG)
	#define PRIVATE
#else
	#define PRIVATE static
#endif /* end defined(IX_DEBUG) */

/*IX_INCLUDE_REGISTRY and IX_EXCLUDE_CCI have been removed from here as they need
to defined at the GLOBAL level as compile-time defines*/

/* Comment out this line if you do not want IPv6 support. */
/*#define IX_IPV6_SUPPORT 1*/

/* Comment out this line if you do not want Ethernet support. */
/*#define IX_ETHERNET_INCLUDE 1*/

/* Number of bytes to reserve at head of SOP HW buffers allocated on the core. */
#define IX_CC_SOP_HW_BUFFER_PREPEND     128

/**
*
* defines max size for ublock image file
*/
#define IX_UBLOCK_IMAGE_NAME_MAX_SIZE   128  

/**
*   hz - following macros are for combining SRAM channel number and offset into one
*   long word for patching it to the microblocks. It is prefered way for microblocks 
*   to receive base address of shared SRAM 
*/

/**
 * DESCRIPTION: Least significant bit of SRAM channel number field into a physical offset
 *  
 */
#define IX_CHANNEL_NUMBER_LSB     30U

 /**
 * DESCRIPTION: Max value for the channel number
 *  4 channels - 0,1,2, and 3
 */
#define IX_CHANNEL_NUM_MAX_VALUE   3
 /**
 * DESCRIPTION: 32-bit number with channel bits set to 1's 
 */
#define IX_CHANNEL_NUMBER_BITS    (IX_CHANNEL_NUM_MAX_VALUE << IX_CHANNEL_NUMBER_LSB)
 
 /**
 * MACRO NAME: IX_CC_SET_SRAM_CHANNEL_INTO_OFFSET
 *
 * DESCRIPTION: This macro sets channel number into physical offset adress into
 * two upper bits of the long word with the following mapping
 *   00 - channel 0
 *   01 - channel 1
 *   10 - channel 2
 *   11 - channel 3
 *  @Param:  - IN arg_ChannelNumber -  channel number
 *  @Param:  - INOUT arg_Offset   - physical offset
  * @Return: packed long word inside the offset. 
 */
#define IX_CC_SET_SRAM_CHANNEL_INTO_OFFSET(arg_ChannelNumber,arg_Offset) \
    IX_CC_WRAP(arg_Offset &= ~IX_CHANNEL_NUMBER_BITS; \
               arg_Offset |= (arg_ChannelNumber << IX_CHANNEL_NUMBER_LSB))

/**
 * MACRO NAME: IX_CC_GET_SRAM_CHANNEL_AND_OFFSET
 *
 * DESCRIPTION: This macro reads channel number and physical offset from the
 *              value prepared for patching to microcode by the
 *              IX_CC_SET_SRAM_CHANNEL_INTO_OFFSET macro.
 * NOTE: It is allowed to use the same variable for the input arg_Value and
 *       one of output arguments.
 *  @Param:  - IN  arg_Value   - value containing channel and offset
 *  @Param:  - OUT arg_Channel - store the channel number here
 *  @Param:  - OUT arg_Offset  - store the physical offset here
 * @Return: None
 */
#define IX_CC_GET_SRAM_CHANNEL_AND_OFFSET(arg_Value, arg_Channel, arg_Offset) \
        {                                                                     \
            ix_uint32 val = (ix_uint32)(arg_Value);                           \
            (arg_Channel) = val >> IX_CHANNEL_NUMBER_LSB;                     \
            (arg_Offset) = val & ~IX_CHANNEL_NUMBER_BITS;                     \
        }

/* end of channel macros */

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
#define MAX_NUMBER_OF_BLADES 32
#else /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */
#define MAX_NUMBER_OF_BLADES 64
#endif /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

#define MAX_NUMBER_OF_PORTS_PER_BLADE 16

/* Macros to read and write register memory */
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
unsigned int get32(unsigned int addr);
void put32(unsigned int addr, unsigned int val);
#define GET32(addr)          get32(addr)
#define PUT32(addr,val)      put32(addr, val)
#define taskDelay(s)         mdelay(s)
#else
#define GET32(addr)          (*(volatile unsigned int *)(addr))
#define PUT32(addr,val)      (*(volatile unsigned int *)(addr) = (val))
#endif /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

#if defined(__cplusplus)
} /* extern "C" */
#endif /* __cplusplus */

#endif /*__IX_CC_MACROS_H__*/
