/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation.All rights reserved.
 *
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 5
 *
 * = FILENAME
 *      ix_cc_properties.h
 *
 * = DESCRIPTION
 *      Header for interface property structure used by the core components
 *      in receiving property updates.
 *
 * = AUTHOR
 *       Aaron Luk
 *       aaron.luk@intel.com
 *
 * = CHANGE HISTORY
 *      7/25/2002 3:45:01 PM - creation time
 *
 * ============================================================================
 * $Id: ix_cc_properties.h,v 1.31 2003/11/18 19:15:13 rranjeet Exp $
 * $History: $
 */

#if !defined(__IX_CC_PROPERTIES_H__)
#define __IX_CC_PROPERTIES_H__

#include "ix_rm.h"

#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */


/* Maximum length of property name in the registry.h file
 * for ex: "IP_ADDR_0" etc.
 */
#define IX_CC_MAX_LEN_PROPERTY_NAME 20

/* Length of MAC address in bytes. */
#define IX_CC_MAC_ADDR_LEN 6

/* Property IDs used to set properties. */

/**
 * Used to add a logical interface for a given
 * port.  In this release, it will simply
 * modify the IPv4 or IPv6 address and subnet
 * mask for the existing logical interface for
 * the port, as we support only 1 logical
 * interface per port in this release.
 */
#define IX_CC_SET_PROPERTY_ADD_INTERFACE_IPV4   0x00000001
#define IX_CC_SET_PROPERTY_ADD_INTERFACE_IPV6   0x00000002


/**
 * Used to delete a logical interface for a
 * given port.  In this release, it will simply
 * set the IPv4 or IPv6 address and subnet mask
 * for the existing logical interface for the
 * port to 0.0.0.0, as we support only 1
 * logical interface per port in this release.
 */
#define IX_CC_SET_PROPERTY_DEL_INTERFACE_IPV4   0x00000004
#define IX_CC_SET_PROPERTY_DEL_INTERFACE_IPV6   0x00000008

/**
 * Used to add a gateway address for a given
 * port.
 */
#define IX_CC_SET_PROPERTY_ADD_GATEWAY_IPV4     0x00000010
#define IX_CC_SET_PROPERTY_ADD_GATEWAY_IPV6     0x00000020

/**
 * Used to delete a gateway address for a
 * given port - sets gateway to 0.0.0.0.
 */
#define IX_CC_SET_PROPERTY_DEL_GATEWAY_IPV4     0x00000040
#define IX_CC_SET_PROPERTY_DEL_GATEWAY_IPV6     0x00000080

/**
 * Used to add a broadcast address for a given
 * port.
 */
#define IX_CC_SET_PROPERTY_ADD_BROADCAST_IPV4   0x00000100
#define IX_CC_SET_PROPERTY_ADD_BROADCAST_IPV6   0x00000200

/**
 * Used to delete a broadcast address for a
 * given port - sets broadcast to 0.0.0.0.
 */
#define IX_CC_SET_PROPERTY_DEL_BROADCAST_IPV4   0x00000400
#define IX_CC_SET_PROPERTY_DEL_BROADCAST_IPV6   0x00000800

/* Used to modify the MTU of a port. */
#define IX_CC_SET_PROPERTY_MTU                  0x00001000

/* Used to set the physical port status (up/down, etc.). */
#define IX_CC_SET_PROPERTY_PHYSICAL_IF_STATUS   0x00002000

/**
 * Used to set the link status (up/down, etc.).  This should
 * only be used by a component/application that can detect
 * when the link goes up or down.  For example, the proper
 * use of this would be when a cable is unplugged, and the
 * media driver would then trigger an interrupt to an
 * Interface CC.  The Interface CC would then invoke the
 * async_set_property API into the Stack Driver, which
 * would notify the appropriate property clients that the
 * link has gone down.
 */
#define IX_CC_SET_PROPERTY_LINK_STATUS          0x00004000

/* Used to set the MAC address for a port. */
#define IX_CC_SET_PROPERTY_MAC_ADDR             0x00008000

/**
 * Used to set link speed for a port -
 * unsupported in this release.
 */
#define IX_CC_SET_PROPERTY_LINK_SPEED           0x00010000


/**
 * TYPENAME: ix_cc_physical_if_status
 * 
 * DESCRIPTION: Enumeration for physical interface state.
 *              This may be expanded as necessary.
 *
 */
typedef enum ix_e_cc_physical_if_status
{
    IX_CC_PHYSICAL_IF_STATUS_FIRST = 0,
    IX_CC_PHYSICAL_IF_STATUS_DOWN = IX_CC_PHYSICAL_IF_STATUS_FIRST,
    IX_CC_PHYSICAL_IF_STATUS_UP,
    IX_CC_PHYSICAL_IF_STATUS_LAST
} ix_cc_physical_if_status;

/**
* TYPENAME: ix_cc_link_status
* 
* DESCRIPTION: Enumeration for link state.
*              This may be expanded as necessary.
*
*/
typedef enum ix_e_cc_link_status
{
    IX_CC_LINK_STATUS_FIRST = 0,
    IX_CC_LINK_STATUS_DOWN = IX_CC_LINK_STATUS_FIRST,
    IX_CC_LINK_STATUS_UP,
    IX_CC_LINK_STATUS_LAST
} ix_cc_link_status; 

/**
 * TYPENAME: ix_cc_media_type
 * 
 * DESCRIPTION: Media type for a port.
 *
 */
typedef enum ix_e_media_type
{
    IX_MEDIA_TYPE_FIRST = 0,
    IX_MEDIA_TYPE_FAST_ETHERNET = IX_MEDIA_TYPE_FIRST,
    IX_MEDIA_TYPE_1GB_ETHERNET,
    IX_MEDIA_TYPE_ATM,
    IX_MEDIA_TYPE_POS,
    IX_MEDIA_TYPE_UNKNOWN,
    IX_MEDIA_TYPE_LAST
} ix_media_type;


/**
 * TYPENAME: ix_ipv4_properties
 * 
 * DESCRIPTION: Structure to hold IPv4 interface properties.
 *
 */
typedef struct ix_s_ipv4_properties
{
    ix_uint32 ipv4_address;
    ix_uint32 ipv4_subnet_mask;
    ix_uint32 ipv4_broadcast;
    ix_uint32 ipv4_gateway;
} ix_ipv4_properties;


#define PROP_MAX_IPV6_ADDRESSES     7  
#define PROP_MAX_IPV4_ADDRESSES     1
/* Maximum number of virtual interfaces per physical
 * interface is number of IPv4 addresses +
 * number of IPv6 addresses allowed.
 */
#define PROP_MAX_VIRTUAL_IF \
        (PROP_MAX_IPV6_ADDRESSES + PROP_MAX_IPV4_ADDRESSES)


/**
 * TYPENAME: ix_ipv6_addr_prefixlen
 *
 * DESCRIPTION: Structure to hold IPv6 address and its
 *              prefix length.
 *
 */
typedef struct ix_s_ipv6_addr_prefixlen
{
    ix_uint128 addr;
    ix_uint8   prefixlen;
} ix_ipv6_addr_prefixlen;

/**
 * TYPENAME: ix_ipv6_properties
 * 
 * DESCRIPTION: Structure to hold IPv6 interface properties.
 * Each virtual interface has one ipv6 address and one
 * gateway address.
 *
 */
typedef struct ix_s_ipv6_properties
{
    ix_ipv6_addr_prefixlen ipv6_address;
    ix_uint128             ipv6_gateway;
} ix_ipv6_properties;

/**
 * TYPENAME: ix_cc_ip_version
 * 
 * DESCRIPTION: Identification for IP version.
 *
 */
typedef enum ix_e_ip_version
{
    IX_CC_IP_VERSION_IPV4,
    IX_CC_IP_VERSION_IPV6
} ix_cc_ip_version;


/**
 * TYPENAME: ix_cc_ip_properties
 * 
 * DESCRIPTION: Container for IP properties of a virtual interface.
 *
 */
typedef struct ix_s_cc_ip_properties
{
    ix_cc_ip_version ip_version;
    union
    {
        ix_ipv4_properties ipv4_prop;
        ix_ipv6_properties ipv6_prop;
    } protocol;
} ix_cc_ip_properties;


/**
 * TYPENAME: ix_cc_properties
 * 
 * DESCRIPTION: Structure to hold interface properties, used to
 *              update properties or get properties.
 *
 */
typedef struct ix_cc_s_properties
{
    /* ID of the blade (FP) */
    ix_uint32 blade_id;

    /* ID of physical interface(port) */
    ix_uint32 port_id; 

    /* IP properties. */
    /* Gives the valid number of virtual interfaces in the
     * following array.
     */
    ix_uint32 no_virtual_if;
    ix_cc_ip_properties ip_prop[PROP_MAX_VIRTUAL_IF];

    ix_uint16 mtu;

    /* physical if state */
    ix_cc_physical_if_status physical_if_state;

    /* link state */
    ix_cc_link_status link_state;

    /* link speed in megabits-per-second. */
    ix_uint32 link_speed;

    /* MAC address in network byte order. */
    ix_uint8 mac_addr[IX_CC_MAC_ADDR_LEN];

    /* Media type (Fast Ethernet, 1GB Ethernet, ATM, POS) */
    ix_media_type media_type;
} ix_cc_properties;



/**
 * TYPENAME: ix_cc_properties_node
 * 
 * DESCRIPTION: Element of linked list of property structures.
 *
 */
typedef struct ix_s_cc_properties_node ix_cc_properties_node;
struct ix_s_cc_properties_node
{
    /* Pointer to interface properties data. */
    ix_cc_properties* pPropertyInfo;

    /* Pointer to next node in linked list. */
    ix_cc_properties_node* pNextPropNode;
};


/**
 * TYPENAME: ix_cc_properties_msg
 * 
 * DESCRIPTION: Structure to hold property update messages.
 *
 */
typedef struct ix_s_cc_properties_msg
{
    ix_uint32 propID; /* ID of property update. */
    ix_cc_properties propData; /* properties structure with properties data. */
} ix_cc_properties_msg;


/**
 * NAME: ix_cc_async_set_property
 *
 * DESCRIPTION: This function sends update messages to the registered clients
 * who are interested in the respective property changes.
 * Note that although this is an asynchronous function, no callback will be
 * provided to indicate success or error of the actual property updates.
 * 
 * @Param:  - IN ix_uint32 arg_propId - id of the property to set.
 * @Param:  - IN ix_cc_properties* arg_pProperty - pointer to
 *          ix_cc_properties data structure.
 * @Return: IX_SUCCESS
 *          IX_CC_ERROR_OOM_SYSTEM
 */
ix_error ix_cc_async_set_property(
                                  ix_uint32 arg_propId,
                                  ix_cc_properties* arg_pProperty
                                  );



#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__IX_CC_PROPERTIES_H__) */
