/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      ix_netmacros.h
 *
 * = DESCRIPTION
 *      Contains macros for converting numbers between network and host byte
 *      order.  Many OSs offer macros, but sometimes they are included from
 *      different header files.  This file hides those details from the user.
 * ============================================================================
 * $Id: ix_netmacros.h,v 1.13 2003/11/06 19:00:24 ktseng Exp $
 */

#if !defined(__IX_NETMACROS_H__)
#define __IX_NETMACROS_H__

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

#include "ix_os_type.h"
#include "ix_cc_macros.h"

#define IX_BIG_ENDIAN           1
#define IX_LITTLE_ENDIAN        2

#if ((_IX_OS_TYPE_ == _IX_OS_LINUX_USER_) || (_IX_OS_TYPE_ == _IX_OS_VXWORKS_))

#include <netinet/in.h>

#elif (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

#include <linux/in.h>

#endif /* ((_IX_OS_TYPE_ == _IX_OS_LINUX_USER_) || (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)) */

/*
 * For Linux...
 */
#if ((_IX_OS_TYPE_ == _IX_OS_LINUX_USER_) || \
     (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_))
    #if defined(__LITTLE_ENDIAN) && (__BYTE_ORDER == __LITTLE_ENDIAN)
        #define IX_BYTE_ORDER IX_LITTLE_ENDIAN
    #endif /* Linux Little Endian */

    #if defined(__BIG_ENDIAN) && (__BYTE_ORDER == __BIG_ENDIAN)
        #define IX_BYTE_ORDER IX_BIG_ENDIAN
    #endif /* Linux Big Endian */
#endif /* Linux Byte Order */

/*
 * For VxWorks...
 */
#if (_IX_OS_TYPE_ == _IX_OS_VXWORKS_)
    #if defined(_LITTLE_ENDIAN) && (_BYTE_ORDER == _LITTLE_ENDIAN)
        #define IX_BYTE_ORDER IX_LITTLE_ENDIAN
    #endif /* VxWorks Little Endian */

    #if defined(_BIG_ENDIAN) && (_BYTE_ORDER == _BIG_ENDIAN)
        #define IX_BYTE_ORDER IX_BIG_ENDIAN
    #endif /* VxWorks Big Endian */
#endif /* VxWorks Byte Order */
/*
 * Windows is always Little Endian, for us.
 */
#if (_IX_OS_TYPE_ == _IX_OS_WIN32_)
    #define IX_BYTE_ORDER IX_LITTLE_ENDIAN
#endif /* Win32 Byte Order */

#ifndef IX_BYTE_ORDER
    #error Could not determine the byte order
#endif /* IX_BYTE_ORDER */

	
/*
 * Byte swap macros
 */
#ifndef __bswap_16
/* Swap bytes in 16 bit value.  */
#define __bswap_constant_16(x) \
        ((((x) >> 8) & 0xff) | (((x) & 0xff) << 8))
#define __bswap_16(x) __bswap_constant_16 (x)
#endif

#ifndef __bswap_32
/* Swap bytes in 32 bit value.  */
#define __bswap_constant_32(x) \
        ((((x) & 0xff000000) >> 24) | (((x) & 0x00ff0000) >>  8) |      \
        (((x) & 0x0000ff00) <<  8) | (((x) & 0x000000ff) << 24))
#define __bswap_32(x) __bswap_constant_32 (x)
#endif

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

#if (__BYTE_ORDER == __BIG_ENDIAN)

#define IX_NTOH32(__u) (__u)
#define IX_HTON32(__u) (__u)
#define IX_NTOH16(__u) (__u)
#define IX_HTON16(__u) (__u)
#endif /* BIG_ENDIAN */

#if (__BYTE_ORDER == __LITTLE_ENDIAN)
#define IX_NTOH32(__u) __bswap_32(__u)
#define IX_HTON32(__u) __bswap_32(__u)
#define IX_NTOH16(__u) __bswap_16(__u)
#define IX_HTON16(__u) __bswap_16(__u)
#endif /* LITTLE_ENDIAN */

#endif /* end (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

#if ((_IX_OS_TYPE_ == _IX_OS_LINUX_USER_) || \
     (_IX_OS_TYPE_ == _IX_OS_VXWORKS_))
#include <netinet/in.h>

#define IX_NTOH32(__u) ((ix_uint32) ntohl(__u))
#define IX_HTON32(__u) ((ix_uint32) htonl(__u))
#define IX_NTOH16(__u) ((ix_uint16) ntohs(__u))
#define IX_HTON16(__u) ((ix_uint16) htons(__u))
#endif /* Linux user || VxWorks */

#if (_IX_OS_TYPE_ == _IX_OS_WIN32_)
#include "windows.h"
#define IX_NTOH32(__u) ((ix_uint32) ntohl(__u))
#define IX_HTON32(__u) ((ix_uint32) htonl(__u))
#define IX_NTOH16(__u) ((ix_uint16) ntohs(__u))
#define IX_HTON16(__u) ((ix_uint16) htons(__u))
#endif /* Windows */




#if IX_BYTE_ORDER == IX_LITTLE_ENDIAN
/*
 * It is expected that both arg_pSrc and arg_pDst are of type ix_uint128*. Note
 * that after using this macro, the caller should only access the destination's
 * aUint32 array (not aUint8, aUint16, etc.).
 */
#define IX_NTOH128(arg_pSrc, arg_pDst) IX_CC_WRAP(                      \
    (arg_pDst)->aUint32[0] = IX_NTOH32((arg_pSrc)->aUint32[0]);         \
    (arg_pDst)->aUint32[1] = IX_NTOH32((arg_pSrc)->aUint32[1]);         \
    (arg_pDst)->aUint32[2] = IX_NTOH32((arg_pSrc)->aUint32[2]);         \
    (arg_pDst)->aUint32[3] = IX_NTOH32((arg_pSrc)->aUint32[3]))
#else /* IX_BYTE_ORDER == IX_BIG_ENDIAN */
#define IX_NTOH128(arg_pSrc, arg_pDst) IX_CC_WRAP(                      \
    (arg_pDst)->aUint32[0] = (arg_pSrc)->aUint32[0];                    \
    (arg_pDst)->aUint32[1] = (arg_pSrc)->aUint32[1];                    \
    (arg_pDst)->aUint32[2] = (arg_pSrc)->aUint32[2];                    \
    (arg_pDst)->aUint32[3] = (arg_pSrc)->aUint32[3])
#endif /* IX_BYTE_ORDER */
#define IX_HTON128(arg_src, arg_dst)   IX_NTOH128(arg_src, arg_dst)

/*
 * Sometimes we need to pack a 32-bit network value into a 16-bit network
 * location, or expand a 16-bit network value into a 32-bit network value.
 * These macros are slightly more efficient than converting to host values
 * then back to network of the new size.
 */
#if IX_BYTE_ORDER == IX_BIG_ENDIAN
#define IX_N32TO16(__u) ((ix_uint16) (__u))
#define IX_N16TO32(__u) ((ix_uint32) (__u))
#define IX_N32TO8(__u)  ((ix_uint8)  (__u))
#define IX_N8TO32(__u)  ((ix_uint32) (__u))
#else /* IX_LITTLE_ENDIAN */
/*
 * When shifting to the right, cast to uint just in case it's a signed number.
 * Shifting a negative number results in the upper bits being set.
 */
#define IX_N32TO16(__u) (((ix_uint32)(__u)) >> 16)
#define IX_N16TO32(__u) ((__u) << 16)
#define IX_N32TO8(__u)  (((ix_uint32)(__u)) >> 24)
#define IX_N8TO32(__u)  ((__u) << 24)
#endif

#if defined(__cplusplus)
} /* extern "C" */
#endif /* __cplusplus */

#endif /*__IX_NETMACROS_H__*/
