///////////////////////////////////////////////////////////////////////////////
//
//
//                  I N T E L   P R O P R I E T A R Y
//
//     COPYRIGHT [c]  2002 BY  INTEL  CORPORATION.  ALL RIGHTS
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT
//     THE PRIOR WRITTEN PERMISSION OF :
//
//                        INTEL  CORPORATION
//
//                     2200 MISSION COLLEGE BLVD
//
//               SANTA  CLARA,  CALIFORNIA  95052-8119
//
///////////////////////////////////////////////////////////////////////////////
//
//      Filename: dl_buf_multi.uc
//
//      Purpose: Packet buffer utility macros for multiple buffer freelists.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef	__DL_BUF_MULTI_UC__
#define	__DL_BUF_MULTI_UC__

#include <buf.uc>
#include <dl_system.h>


///////////////////////////////////////////////////////////////////////////////
// dl_buf_init:
// 
// Description: 
//   Initializes the buffering mechanism. This macro should be called once 
//   before any of the other dl_buf macros are called.
//   Typically this macro will create a buffer freelist.
//
// Outputs:
//
// Inputs:
//   freelist_id : The id of the buffer freelist to create. 
//                 (freelist_id MUST be a constant).  
// Size:   				:
//
///////////////////////////////////////////////////////////////////////////////
#macro dl_buf_init(freelist_id)
.begin

	#if (!isnum(freelist_id))
	#error "freelist_id must be a constant."
	#endif

	// Compute number of packet buffers.

	#define_eval NO_OF_PKT_BUFFERS	BUF_SRAM_SIZE_/**/freelist_id/**/ / META_DATA_SIZE_/**/freelist_id/**/

	// Sanity check. Check if the specified SDRAM memory is of suffcient size
	// to hold all buffers.

	#define_eval SDRAM_BUFFERS		BUF_SDRAM_SIZE_/**/freelist_id/**/ / BUFFER_SIZE_/**/freelist_id/**/

	#if	(SDRAM_BUFFERS != NO_OF_PKT_BUFFERS)
	#error	"The amount of memory allocated in SRAM for packet buffers doesn't \
			 match with the SDRAM size. Check BUF_SRAM_SIZE_/**/freelist_id/**/ \
			 and BUF_SDRAM_SIZE_/**/freelist_id/**/. Also check \
			 META_DATA_SIZE_/**/freelist_id/**/ and BUFFER_SIZE_/**/freelist_id/**/."
	#endif

	#undef	SDRAM_BUFFERS

	// Specifying a base of address of 0 for SRAM (and DRAM) simplifies the
	// arithmetic required to translate the handle to offset. It saves one
	// instruction. (See dl_buf_get_desc). However, dl_buf_alloc will return 0
	// for a null buffer (i.e no buffer available) and it may not be possible
	// to differentiate between a null buffer and a buffer at address 0.
	// So we simply make that one buffer at address 0 unavailable for
	// allocation/freeing, yet taking advantage of using a base address of 0.
	// We end up wasting a little memory (1 buffer) but that's ok.

	// If USE_IMPORT_VAR is defined BUF_SRAM_BASE_/**/freelist_id/**/ and
	// BUF_SDRAM_BASE_/**/freelist_id/**/ are patched symbols.So we need to use
	// gpr's to modify these values.
	// If USE_IMPORT_VAR is not defined we can modify these values using
	// #define_eval which is more efficient. But since this is init code, for
	// simplicity we can use gpr's in this case too.

	.reg _buf_sram_base _buf_sdram_base

	move(_buf_sram_base, BUF_SRAM_BASE_/**/freelist_id/**/)
	move(_buf_sdram_base, BUF_SDRAM_BASE_/**/freelist_id/**/)

	.if	(_buf_sram_base==0x00000000 || _buf_sram_base==0x40000000 || \
		 _buf_sram_base==0x80000000 || _buf_sram_base==0xC0000000)

	.reg _meta_data_size _buffer_size

	immed32[_meta_data_size, META_DATA_SIZE_/**/freelist_id/**/]
	immed32[_buffer_size, BUFFER_SIZE_/**/freelist_id/**/]

	alu[_buf_sram_base, _buf_sram_base, +, _meta_data_size]
	alu[_buf_sdram_base, _buf_sdram_base, +, _buffer_size]

	#define_eval NO_OF_PKT_BUFFERS	NO_OF_PKT_BUFFERS - 1

	.endif

	buf_freelist_create(NO_OF_PKT_BUFFERS, BUF_FREE_LIST_/**/freelist_id/**/,
						_buf_sdram_base, BUFFER_SIZE_/**/freelist_id/**/,
						_buf_sram_base, META_DATA_SIZE_/**/freelist_id/**/)

	#undef NO_OF_PKT_BUFFERS

.end
#endm


///////////////////////////////////////////////////////////////////////////////
// dl_buf_alloc:
// 
// Description: 
//   Allocates an available packet buffer from the specified freelist. If no 
//   buffer is available it returns 0.
//
// Outputs:
//   buf_handle : Handle of the newly allocated buffer [bits 23:0].
//                buf_handle must be a SRAM transfer register.
//
// Inputs:
//   freelist_id      : The id of the buffer freelist from which to allocate
//                      a buffer. (freelist_id MUST be a constant).
//   req_sig          : Signal to use in the dequeue operation.
//   sig_action       : What to do with during the dequeue operation - 
//                        - wait for signal(s) as specified by sig_action
//                        - do not wait for signal, just return (SIG_NONE)
//   FREELIST_ID_TYPE : This field is added to differentiate this macro from 
//                      the macro in dl_buf.uc which takes in the q-array entry
//                      number as an argument. This macro takes in the 
//                      as an argument. FREELIST_ID_TYPE is not used in
//                      freelist_id this macro; its only purpose is to create 
//                      a unique signature for this macro.
//
// Size:
//   3 - 4 instructions.
///////////////////////////////////////////////////////////////////////////////
#macro dl_buf_alloc(buf_handle, freelist_id, req_sig, sig_action, FREELIST_ID_TYPE)
.begin

	#if (!isnum(freelist_id))
	#error "freelist_id must be a constant."
	#endif

	.reg _freelist_id, _channel_number

	; Get the channel number from which to fetch the buffer.

	immed32[_channel_number, (BUF_SRAM_BASE_/**/freelist_id/**/ & (0x3 << 30))]

    ; Get the freelist id from which to fetch the buffer.

	alu[_freelist_id, --, B, BUF_FREE_LIST_/**/freelist_id/**/, <<24]

	; Fetch the buffer handle from the free list.

	sram[dequeue, buf_handle, _freelist_id, _channel_number], sig_done[req_sig]

    ; If required swap out and wait for the dequeue to complete.

	#if (!streq('sig_action', 'SIG_NONE'))
	ctx_arb[sig_action]
	#endif

.end
#endm


///////////////////////////////////////////////////////////////////////////////
// dl_buf_free:
// 
// Description: 
//   Frees the buffer that was previously allocated using dl_buf_alloc().
//   Only one buffer can be freed at a time by this macro. Chained buffers
//   (for large packets) are not supported by this macro.
//   EOP, SOP and cell count in the buffer handle will all be reset to zero 
//   by this macro. So the user need not reset them before calling this macro.
//
// Outputs:
//
// Inputs:
//   buf_handle       : The buffer handle to free.
//   freelist_id      : The id of the buffer freelist to which the buffer
//                      will be freed. (freelist_id MUST be a constant).
//   FREELIST_ID_TYPE : This field is added to differentiate this macro from 
//                      the macro in dl_buf.uc which takes in the q-array entry
//                      number as an argument. This macro takes in the 
//                      freelist_id as an argument. FREELIST_ID_TYPE is not 
//                      used in this macro; its only purpose is to create a 
//                      unique signature for this macro.
//
// Size:
//   4 - 5 instructions.
///////////////////////////////////////////////////////////////////////////////
#macro dl_buf_free(buf_handle, freelist_id, FREELIST_ID_TYPE)
.begin

	#if (!isnum(freelist_id))
	#error "freelist_id must be a constant."
	#endif

	.reg _channel_number _sram_addr

	; Mask out the cell count and sop and eop bits from the buffer handle before enqueuing.

	alu[_sram_addr, buf_handle, AND~, 0xFF, <<24]

	#ifdef CHANNEL_NUMBER_IN_BUFFER_HANDLE

	; Mask out the channel number from the buffer handle before enqueuing.

	alu[_sram_addr, _sram_addr, AND~, 0x6]

	#endif

	; Get the channel number on which to enqueue the buffer.

	immed32[_channel_number, (BUF_SRAM_BASE_/**/freelist_id/**/ & (0x3 << 30))]

	; Set the freelist id on which to enqueue the buffer.

	alu[_sram_addr, _sram_addr, OR, BUF_FREE_LIST_/**/freelist_id/**/, <<24]

	; Set the channel number on which to enqueue the buffer and free the buffer.

	sram[enqueue, --, _sram_addr, _channel_number]
	
.end
#endm


///////////////////////////////////////////////////////////////////////////////
// dl_buf_get_desc:
// 
// Description: 
//   Given a buffer handle, this macro returns the sram address where the 
//   meta-data (called the buffer descriptor) for this buffer handle 
//   is present. Bits 23:0 of the handle contain the sram address.
//   The address returned is an absolute byte address and not an offset 
//   from a given base address.
//
// Outputs:
//   sram_offset : SRAM byte address [23:0] and SRAM channel number [31:30]
//                 where the meta data for this buffer is present. 
//
// Inputs:
//   buf_handle  : The buffer handle.
//   freelist_id : The id of the buffer freelist from which to allocate a
//                 buffer (freelist_id MUST be a constant).
//
// Size:
//   2 - 6 instructions.
///////////////////////////////////////////////////////////////////////////////
#macro dl_buf_get_desc(sram_offset, buf_handle, freelist_id)
.begin

	#if (!isnum(freelist_id))
	#error "freelist_id must be a constant."
	#endif

	; Clear out bits 31:24 of buf_handle.

	#if (isnum(buf_handle))
	
	.reg _temp_handle
	
	immed32[_temp_handle, buf_handle]
	alu[sram_offset, _temp_handle, AND~, 0xFF, <<24]
	
	#else

	alu[sram_offset, buf_handle, AND~, 0xFF, <<24]

	#endif	// (isnum(buf_handle))

	#ifdef CHANNEL_NUMBER_IN_BUFFER_HANDLE

	; Mask out the channel number.

	alu[sram_offset, sram_offset, AND~, 0x6]

	#endif

	; Convert long word address to byte address.
	; sram_offset * 4 will give sram byte offset within a channel. 

	alu[sram_offset, --, B, sram_offset, <<2]

	; Set the channel number in sram_offset.

	#ifdef	USE_IMPORT_VAR

	// BUF_SRAM_BASE_/**/freelist_id/**/ is a patched symbol.

	.reg _channel_number

	immed32[_channel_number, ((BUF_SRAM_BASE_/**/freelist_id/**/ >> 30) & 0x3)]

	alu[sram_offset, sram_offset, OR, _channel_number, <<30]

	#else	// USE_IMPORT_VAR

	#define_eval  CHANNEL_NUMBER  ((BUF_SRAM_BASE_/**/freelist_id/**/ >> 30) & 0x3)

	#if (CHANNEL_NUMBER != 0)

	alu[sram_offset, sram_offset, OR, CHANNEL_NUMBER, <<30]	

	#endif	// CHANNEL_NUMBER

	#undef	CHANNEL_NUMBER

	#endif // USE_IMPORT_VAR

.end
#endm


///////////////////////////////////////////////////////////////////////////////
// dl_buf_get_desc:
// 
// Description: 
//   Given a buffer handle, this macro returns the sdram address where the 
//   packet data for this buffer handle is present. Bits 23:0 of the handle
//   contain the sram address. The corresponding dram address is computed by
//   multiplying the buffer handle address by a fixed ratio and adding the 
//   dram base address to the result.
//
// Outputs:
//   sdram_offset : SDRAM byte address where the packet data for this buffer
//                  is present. 
//
// Inputs:
//   buf_handle  : The buffer handle.
//   freelist_id : The id of the buffer freelist this buffer belongs to.
//                 (freelist_id MUST be a constant).
//
// Size:
//   2 - 6 instructions.
///////////////////////////////////////////////////////////////////////////////
#macro dl_buf_get_data(sdram_offset, buf_handle, freelist_id)
.begin

	#if (!isnum(freelist_id))
	#error "freelist_id must be a constant."
	#endif

	.reg _dl_rel_base

	; Clear out bits 31:24 of buf_handle.

	#if (isnum(buf_handle))
	
	.reg _temp_handle
	
	immed32[_temp_handle, buf_handle]
	alu[sdram_offset, _temp_handle, AND~, 0xFF, <<24]
	
	#else

	alu[sdram_offset, buf_handle, AND~, 0xFF, <<24]

	#endif	// (isnum(buf_handle))

	#ifdef CHANNEL_NUMBER_IN_BUFFER_HANDLE

	; Mask out the channel number.

	alu[sdram_offset, sdram_offset, AND~, 0x6]

	#endif

	; Compute SDRAM offset.

	#define_eval 	DL_DS_RATIO			(BUFFER_SIZE_/**/freelist_id/**/ / META_DATA_SIZE_/**/freelist_id/**/)
	#define_eval	MULT_FACTOR			LOG2(DL_DS_RATIO)
	#define_eval	MULT_FACTOR_PLUS_2	(MULT_FACTOR + 2)

	alu[sdram_offset, --, B, sdram_offset, <<MULT_FACTOR_PLUS_2]


	; Add SDRAM base address (i.e DL_REL_BASE) to the computed SDRAM offset.

 	#ifdef USE_IMPORT_VAR

	// DL_REL_BASE is a patched symbol. So simply use it.

	immed32[_dl_rel_base, DL_REL_BASE]
	alu[sdram_offset, _dl_rel_base, +, sdram_offset]

	#else // USE_IMPORT_VAR

	// Compute DL_REL_BASE.

	#define_eval    TEMP_BUF_BASE       (BUF_SRAM_BASE_/**/freelist_id/**/ & 0x3FFFFFFF)
	#define_eval	DL_REL_BASE			(BUF_SDRAM_BASE_/**/freelist_id/**/ - (TEMP_BUF_BASE * DL_DS_RATIO))

	#if (DL_REL_BASE != 0)

	immed32[_dl_rel_base, DL_REL_BASE]
	alu[sdram_offset, _dl_rel_base, +, sdram_offset]

	#endif // (DL_REL_BASE != 0)

	#endif // USE_IMPORT_VAR

	#undef DL_REL_BASE
	#undef TEMP_BUF_BASE

	#undef MULT_FACTOR_PLUS_2
	#undef MULT_FACTOR
	#undef DL_DS_RATIO

.end
#endm


///////////////////////////////////////////////////////////////////////////////
// dl_buf_get_data_from_meta:
// 
// Description: 
//   This macro is almost identical to dl_buf_get_data(). Instead of taking
//   a buffer  handle as an input parameter, it takes the sram offset where
//   the buffer meta data is present (Calling dl_buf_get_desc() will give you
//   that sram offset). It saves a couple of instructions.
//   See dl_buf_get_data() for more details.
//
// Outputs:
//   sdram_offset : SDRAM byte address where the packet data for this buffer
//                  is present. 
//
// Inputs:
//   sram_offset : The SRAM offset where the meta data for this buffer is
//                 present.
//   freelist_id : The id of the buffer freelist this buffer belongs to.
//                 (freelist_id MUST be a constant).
//
// Size:
//   2 - 6 instructions.
///////////////////////////////////////////////////////////////////////////////
#macro dl_buf_get_data_from_meta(sdram_offset, sram_offset, freelist_id)
.begin

	#if (!isnum(freelist_id))
	#error "freelist_id must be a constant."
	#endif

	.reg _dl_rel_base

	; Clear out bits 31:24 of buf_handle.

	#if (isnum(sram_offset))
	
	.reg _temp_handle
	
	immed32[_temp_handle, sram_offset]
	alu[sdram_offset, _temp_handle, AND~, 0xFF, <<24]
	
	#else

	alu[sdram_offset, sram_offset, AND~, 0xFF, <<24]

	#endif	// (isnum(buf_handle))

	#ifdef CHANNEL_NUMBER_IN_BUFFER_HANDLE

	; Mask out the channel number.

	alu[sdram_offset, sdram_offset, AND~, 0x6]

	#endif

	; Compute SDRAM offset.

	#define_eval 	TMP_DL_DS_RATIO			(BUFFER_SIZE_/**/freelist_id/**/ / META_DATA_SIZE_/**/freelist_id/**/)
	#define_eval	TMP_MULT_FACTOR			LOG2(TMP_DL_DS_RATIO)

	alu[sdram_offset, --, B, sdram_offset, <<TMP_MULT_FACTOR]


	; Add SDRAM base address (i.e DL_REL_BASE) to the computed SDRAM offset

 	#ifdef USE_IMPORT_VAR

	// DL_REL_BASE is a patched symbol. So simply use it.

	immed32[_dl_rel_base, DL_REL_BASE]
	alu[sdram_offset, _dl_rel_base, +, sdram_offset]

	#else // USE_IMPORT_VAR

	// Compute DL_REL_BASE.

	#define_eval    TEMP_BUF_BASE       (BUF_SRAM_BASE_/**/freelist_id/**/ & 0x3FFFFFFF)
	#define_eval	TMP_DL_REL_BASE			(BUF_SDRAM_BASE_/**/freelist_id/**/ - (TEMP_BUF_BASE * TMP_DL_DS_RATIO))

	#if (TMP_DL_REL_BASE != 0)

	immed32[_dl_rel_base, TMP_DL_REL_BASE]
	alu[sdram_offset, _dl_rel_base, +, sdram_offset]

	#endif // (DL_REL_BASE != 0)

	#endif // USE_IMPORT_VAR

	#undef TMP_DL_REL_BASE
	#undef TEMP_BUF_BASE

	#undef TMP_MULT_FACTOR
	#undef TMP_DL_DS_RATIO

.end
#endm


///////////////////////////////////////////////////////////////////////////////
// dl_buf_drop:
// 
// Description: 
//   Drop one buffer. This is identical to freeing the buffer. For freeing a
//   chain of buffers use dl_buf_drop_chain(). If FREELIST_MANAGER is defined
//   the buffer handle is instead written to a scratch ring that is read by 
//   the freelist manager microblock. In this case, we do a ctx_swap[]. So 
//   care must be taken when calling this macro to ensure threads do not go 
//   out of order after executing the ctx_swap.
//
// Outputs:
//
// Inputs:
//   buf_handle  : The buffer handle to drop.
//   freelist_id : The id of the buffer freelist to drop to.
//                 (freelist_id MUST be a constant).
//                 Unused if FREELIST_MANAGER is defined.
//
// Size:
//
///////////////////////////////////////////////////////////////////////////////
#macro dl_buf_drop(buf_handle, freelist_id)
.begin

	#ifdef FREELIST_MANAGER

	#if (DROP_RING > 11)
	#error "For ring number greater than 11, we cannot use the SCR_Ring#_Full \
			input state. Check DROP_RING"
	#else
	#define_eval	DROP_RING_FULL	SCR_Ring/**/DROP_RING/**/_status
	#endif

	.reg $scr
	.reg ring
	.sig sig_scr

check_ring_full#:

 	br_inp_state[DROP_RING_FULL, check_ring_full#]

	; Assume that any inter thread signaling is done after this.

	alu[$scr, --, B, buf_handle]
	alu[ring, --, b, DROP_RING, <<2]
	scratch[put, $scr, 0, ring, 1], ctx_swap[sig_scr]

	#else	// FREELIST_MANAGER

 	dl_buf_free(buf_handle, freelist_id)

	#endif // FREELIST_MANAGER

.end
#endm


///////////////////////////////////////////////////////////////////////////////
// dl_buf_drop:
// 
// Description: 
//   Drop one buffer. This is almost identical to dl_buf_drop(). The only
//   difference is that if FREELIST_MANAGER is defined, instead of doing a 
//   ctx_swap[] when writing to the scratch ring, it simply adds the signal
//   to a signal mask. This makes using the macro safer to use since we do
//   not ctx_swap[].
//
// Outputs:
//
// Inputs:
//   buf_handle  : The buffer handle to drop.
//   freelist_id : The id of the buffer freelist to drop to.
//                 (freelist_id MUST be a constant).
//                 Unused if FREELIST_MANAGER is defined.
//   scr         : SRAM transfer registers used for scratch[put]
//                 Unused if FREELIST_MANAGER is not defined.
//   sig_mask    : Signal mask to add the scratch[put] signal.
//                 Unused if FREELIST_MANAGER is not defined.
//   sig_scr     : Signal used for scratch[put].
//                 Unused if FREELIST_MANAGER is not defined.
//
// Size:
//
///////////////////////////////////////////////////////////////////////////////
#macro dl_buf_drop(buf_handle, freelist_id, scr, sig_mask, sig_scr)
.begin

	#ifdef FREELIST_MANAGER

	#if (DROP_RING > 11)
	#error "For ring number greater than 11, we cannot use the SCR_Ring#_Full \
			input state. Check DROP_RING"
	#else
	#define_eval	DROP_RING_FULL	SCR_Ring/**/DROP_RING/**/_status
	#endif

	.reg ring

check_ring_full#:

 	br_inp_state[DROP_RING_FULL, check_ring_full#]

	alu[scr[0], --, B, buf_handle]
	alu[ring, --, B, DROP_RING, <<2]
	scratch[put, scr[0], 0, ring, 1], sig_done[sig_scr]
	alu[sig_mask, sig_mask, OR, 1, <<&sig_scr]

	#else	// FREELIST_MANAGER

 	dl_buf_free(buf_handle, freelist_id)

	#endif	// FREELIST_MANAGER

.end
#endm


#endif // __DL_BUF_MULTI_UC__