///////////////////////////////////////////////////////////////////////////////
//
//
//                  I N T E L   P R O P R I E T A R Y
//
//     COPYRIGHT [c]  2002 BY  INTEL  CORPORATION.  ALL RIGHTS
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT
//     THE PRIOR WRITTEN PERMISSION OF :
//
//                        INTEL  CORPORATION
//
//                     2200 MISSION COLLEGE BLVD
//
//               SANTA  CLARA,  CALIFORNIA  95052-8119
//
///////////////////////////////////////////////////////////////////////////////
//
//      Filename: dl_meta_multi.uc
//
//      Purpose: Packet meta-data utility macros for multiple buffer freelists.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef __DL_META_MULTI_UC__
#define __DL_META_MULTI_UC__

#include <stdmac.uc>
#include <xbuf.uc>
#include <buf.uc>


///////////////////////////////////////////////////////////////////////////////
// dl_meta_load_cache:
//
// Description: 
//   Fetch meta-data from SRAM for the specified buffer using the buffer handle.
//			
//   Outputs:
//     dl_meta      : Read transfer register prefix (using xbuf_alloc) to read
//                    the meta-data from SRAM.
//                    At least NUM_LW wxfer registers should have been allocated
//
//   Inputs:
//     buf_handle   : Buffer handle 
//     freelist_id  : Id of the buffer freelist that the handle belongs to.
//                    (freelist_id MUST be a constant).
//     signal_number: Signal to be used in I/O operation
//                    If 'DL_SPLIT_META_DATA' is defined and the meta data to
//                    flush spans two SRAM channels, signal 'signal_number1' will
//                    be used by this macro. It needs to be defined by the code
//                    that calls this macro.
//     START_LW	    : The starting LW from which NUM_LW long words will be fetched.
//     NUM_LW       : Number of long words to fetch. Must be a constant.
//
//   Size:   				
//
///////////////////////////////////////////////////////////////////////////////
#macro dl_meta_load_cache(buffer_handle, freelist_id, dl_meta, signal_number, START_LW, NUM_LW)
.begin

	// If 'DL_SPLIT_META_DATA' is defined and the meta data to
	// load spans two SRAM channels, signal 'signal_number1' will
	// be used by this macro. It needs to be defined by the code
	// that calls this macro.

	#if (!isnum(freelist_id))
	#error "freelist_id must be a constant."
	#endif

	#define_eval _DL_SPLIT_META_DATA		DL_SPLIT_META_DATA_/**/freelist_id/**/

	#if (_DL_SPLIT_META_DATA == TRUE)

	// 	Meta-data in IXP2800 reference designs is split into 2 sram channels.
	// 	While loading or flushing the meta-data we need to account for this.
	
	.reg sram_offset

	; Get the SRAM offset for the meta data. Channel number will be
	; as implied by [31:30] of BUF_SRAM_BASE.

	dl_buf_get_desc(sram_offset, buffer_handle, freelist_id)
	 
	; Read meta data from SRAM

	#define_eval START_INDEX 		(START_LW)
	#define_eval TEMP_SRAM_OFFSET 	(START_LW*4)

	#if (START_LW >= BUF_SRAM_SPLIT_LW_/**/freelist_id/**/)

	; All LW's are in the split SRAM channel

	.reg split_sram_offset

	; Clear the channel number
	alu[split_sram_offset, sram_offset, AND~, 3, <<30]

	; Set the channel number
	alu[split_sram_offset, split_sram_offset, OR, BUF_SRAM_SPLIT_CHANNEL_/**/freelist_id/**/, <<30]

	sram[read, dl_meta[START_INDEX], split_sram_offset, TEMP_SRAM_OFFSET, NUM_LW], \
		sig_done[signal_number]

	#elif (START_LW + NUM_LW - 1 < BUF_SRAM_SPLIT_LW_/**/freelist_id/**/)

	; All LW's are in the default SRAM channel

	.reg default_sram_offset

	; Clear the channel number
	alu[default_sram_offset, sram_offset, AND~, 3, <<30]

	; Set the channel number
	alu[default_sram_offset, default_sram_offset, OR, BUF_SRAM_DEFAULT_CHANNEL_/**/freelist_id/**/, <<30]

	sram[read, dl_meta[START_INDEX], default_sram_offset, TEMP_SRAM_OFFSET, NUM_LW], \
		sig_done[signal_number]

	#else	// (START_LW >= BUF_SRAM_SPLIT_LW_/**/freelist_id/**/)

	; LW's are in default and split SRAM channels

	#define_eval NUM_LW_1				(BUF_SRAM_SPLIT_LW_/**/freelist_id/**/ - START_LW)

	.reg default_sram_offset

	; Clear the channel number
	alu[default_sram_offset, sram_offset, AND~, 3, <<30]

	; Set the channel number
	alu[default_sram_offset, default_sram_offset, OR, BUF_SRAM_DEFAULT_CHANNEL_/**/freelist_id/**/, <<30]

	sram[read, dl_meta[START_INDEX], default_sram_offset, TEMP_SRAM_OFFSET, NUM_LW_1], \
		sig_done[signal_number]

	#define_eval NUM_LW_2				(NUM_LW - NUM_LW_1)
	#define_eval START_INDEX_2			(BUF_SRAM_SPLIT_LW_/**/freelist_id/**/)
	#define_eval TEMP_SRAM_OFFSET_2		(START_INDEX_2*4)

	.reg split_sram_offset

	; Clear the channel number
	alu[split_sram_offset, sram_offset, AND~, 3, <<30]

	; Set the channel number
	alu[split_sram_offset, split_sram_offset, OR, BUF_SRAM_SPLIT_CHANNEL_/**/freelist_id/**/, <<30]

	sram[read, dl_meta[START_INDEX_2], split_sram_offset, TEMP_SRAM_OFFSET_2, NUM_LW_2], \
		sig_done[signal_number/**/1]

	#undef NUM_LW_1
	#undef NUM_LW_2
	#undef START_INDEX_2
	#undef TEMP_SRAM_OFFSET_2

	#endif	// (START_LW >= BUF_SRAM_SPLIT_LW_/**/freelist_id/**/)

	#undef START_INDEX
	#undef TEMP_SRAM_OFFSET

	#else	// 	(_DL_SPLIT_META_DATA == TRUE)

	.reg	sram_offset

	; Get the SRAM offset for the meta data 
	dl_buf_get_desc(sram_offset, buffer_handle)
	 
	; Read meta data from SRAM
								
	#define_eval TEMP_SRAM_OFFSET 	(START_LW*4)
	#define_eval START_INDEX 		(START_LW)

	sram[read, dl_meta[START_INDEX], sram_offset, TEMP_SRAM_OFFSET, NUM_LW],\
								 sig_done[signal_number]						

	#undef	TEMP_SRAM_OFFSET
	#undef  START_INDEX

	#endif	// (_DL_SPLIT_META_DATA == TRUE)

	#undef	_DL_SPLIT_META_DATA
	
.end
#endm


///////////////////////////////////////////////////////////////////////////////
// dl_meta_flush_cache:
//
// Description: 
//   Flushes the cache of meta data from GPR to SRAM.
//			
//   Outputs:
//
//   Inputs:
//	   wxfer_prefix : Write transfer register prefix (using xbuf_alloc).
//                    At least NUM_LW wxfer registers should have been allocated
//     buf_handle   : Buffer handle 
//     freelist_id  : Id of the buffer freelist that the handle belongs to.
//                    (freelist_id MUST be a constant).
//     req_sig      : Signal to be used in I/O operation
//                    If 'DL_SPLIT_META_DATA' is defined and the meta data to
//                    flush spans two SRAM channels, signal 'req_sig1' will
//                    be used by this macro. It needs to be defined by the code
//                    that calls this macro.
//     sig_action   : What to do with I/O operation - 
//                      - wait for signal(s) as specified by sig_action
//                      - do not wait for signal, just return (SIG_NONE)
//     START_LW	    : The starting LW from which NUM_LW long words will be flushed.
//     NUM_LW       : Number of long words to flush. Must be a constant.
//
//   Size:   				
//
///////////////////////////////////////////////////////////////////////////////
#macro dl_meta_flush_cache(wxfer_prefix, buf_handle, freelist_id, req_sig, sig_action, START_LW, NUM_LW)
.begin

	// If 'DL_SPLIT_META_DATA' is defined and the meta data to
	// load spans two SRAM channels, signal 'req_sig1' will
	// be used by this macro. It needs to be defined by the code
	// that calls this macro.

	#if (!isnum(freelist_id))
	#error "freelist_id must be a constant."
	#endif

	#if (!isnum(NUM_LW))
	#error	"NUM_LW should be a constant"
	#endif

	#if (NUM_LW == 0)
	#error	"NUM_LW cannot be zero"
	#endif

	#if (NUM_LW > META_CACHE_SIZE)
	#error	"NUM_LW cannot be greater than META_CACHE_SIZE"
	#endif

	#if (START_LW == 0)
	alu[wxfer_prefix[0], --, b, dl_meta[0]]
	#endif

	#define_eval END_LW (START_LW + NUM_LW)

	#if ((START_LW <= 1) && (END_LW > 1)) 
	alu[wxfer_prefix[1], --, b, dl_meta[1]]
	#endif

	#if ((START_LW <= 2) && (END_LW > 2)) 
	alu[wxfer_prefix[2], --, b, dl_meta[2]]
	#endif

	#if ((START_LW <= 3) && (END_LW > 3))
	alu[wxfer_prefix[3], --, b, dl_meta[3]]
	#endif

	#if ((START_LW <= 4) && (END_LW > 4))
	alu[wxfer_prefix[4], --, b, dl_meta[4]]
	#endif

	#if ((START_LW <= 5) && (END_LW > 5))
	alu[wxfer_prefix[5], --, b, dl_meta[5]]
	#endif

	#if ((START_LW <= 6) && (END_LW > 6))
	alu[wxfer_prefix[6], --, b, dl_meta[6]]
	#endif

	#if ((START_LW <= 7) && (END_LW > 7))
	alu[wxfer_prefix[7], --, b, dl_meta[7]]
	#endif

	dl_meta_flush(wxfer_prefix, buf_handle, freelist_id, req_sig, sig_action, START_LW, NUM_LW)

.end
#endm

///////////////////////////////////////////////////////////////////////////////
// dl_meta_flush:
//
// Description: 
//   Flushes the cache of meta data from GPR to SRAM.
//			
//   Outputs:
//
//   Inputs:
//	   wxfer_prefix : Write transfer register prefix (using xbuf_alloc).
//                    At least NUM_LW wxfer registers should have been allocated
//     buf_handle   : Buffer handle 
//     freelist_id  : Id of the buffer freelist that the handle belongs to.
//                    (freelist_id MUST be a constant).
//     req_sig      : Signal to be used in I/O operation
//                    If 'DL_SPLIT_META_DATA' is defined and the meta data to
//                    flush spans two SRAM channels, signal 'req_sig1' will
//                    be used by this macro. It needs to be defined by the code
//                    that calls this macro.
//     sig_action   : What to do with I/O operation - 
//                      - wait for signal(s) as specified by sig_action
//                      - do not wait for signal, just return (SIG_NONE)
//     START_LW	    : The starting LW from which NUM_LW long words will be flushed.
//     NUM_LW       : Number of long words to flush. Must be a constant.
//
//   Size:   				
//
///////////////////////////////////////////////////////////////////////////////
#macro dl_meta_flush(wxfer_prefix, buf_handle, freelist_id, req_sig, sig_action, \
					START_LW, NUM_LW)
.begin

	// If 'DL_SPLIT_META_DATA' is defined and the meta data to
	// load spans two SRAM channels, signal 'req_sig1' will
	// be used by this macro. It needs to be defined by the code
	// that calls this macro.

	#if (!isnum(freelist_id))
	#error "freelist_id must be a constant."
	#endif

	#if (!isnum(NUM_LW))
	#error	"NUM_LW should be a constant"
	#endif

	#if (NUM_LW == 0)
	#error	"NUM_LW cannot be zero"
	#endif

	#if (NUM_LW > META_CACHE_SIZE)
	#error	"NUM_LW cannot be greater than META_CACHE_SIZE"
	#endif


	#define_eval _DL_SPLIT_META_DATA		DL_SPLIT_META_DATA_/**/freelist_id/**/

	#if (_DL_SPLIT_META_DATA == TRUE)

	// 	Meta-data in IXP2800 reference designs is split into 2 sram channels.
	// 	While loading or flushing the meta-data we need to account for this.

	.reg sram_offset

	; Get the SRAM offset for the meta data 

	dl_buf_get_desc(sram_offset, buf_handle, freelist_id)

	; Write the meta data to sram

	#define_eval START_INDEX 		(START_LW)
	#define_eval TEMP_SRAM_OFFSET 	(START_LW*4)

	#if (START_LW >= BUF_SRAM_SPLIT_LW_/**/freelist_id/**/)

	; All LW's are in the split SRAM channel
	
	.reg split_sram_offset

	; Clear the channel number
	alu[split_sram_offset, sram_offset, AND~, 3, <<30]

	; Set the channel number
	alu[split_sram_offset, split_sram_offset, OR, BUF_SRAM_SPLIT_CHANNEL_/**/freelist_id/**/, <<30]

	sram[write, wxfer_prefix[START_INDEX], split_sram_offset, TEMP_SRAM_OFFSET, NUM_LW], \
		sig_done[req_sig]

	#elif (START_LW + NUM_LW - 1 < BUF_SRAM_SPLIT_LW_/**/freelist_id/**/)

	; All LW's are in the default SRAM channel

	.reg default_sram_offset

	; Clear the channel number
	alu[default_sram_offset, sram_offset, AND~, 3, <<30]

	; Set the channel number
	alu[default_sram_offset, default_sram_offset, OR, BUF_SRAM_DEFAULT_CHANNEL_/**/freelist_id/**/, <<30]

	sram[write, wxfer_prefix[START_INDEX], default_sram_offset, TEMP_SRAM_OFFSET, NUM_LW], \
		sig_done[req_sig]

	#else

	; LW's are in default and split SRAM channels

	#define_eval NUM_LW_1			(BUF_SRAM_SPLIT_LW_/**/freelist_id/**/ - START_LW)

	.reg default_sram_offset

	; Clear the channel number
	alu[default_sram_offset, sram_offset, AND~, 3, <<30]

	; Set the channel number
	alu[default_sram_offset, default_sram_offset, OR, BUF_SRAM_DEFAULT_CHANNEL_/**/freelist_id/**/, <<30]

	sram[write, wxfer_prefix[START_INDEX], default_sram_offset, TEMP_SRAM_OFFSET, NUM_LW_1], \
		sig_done[req_sig]

	#define_eval NUM_LW_2			(NUM_LW - NUM_LW_1)
	#define_eval START_INDEX_2		(BUF_SRAM_SPLIT_LW_/**/freelist_id/**/)
	#define_eval TEMP_SRAM_OFFSET_2	(START_INDEX_2*4)

	.reg split_sram_offset

	; Clear the channel number
	alu[split_sram_offset, sram_offset, AND~, 3, <<30]

	; Set the channel number
	alu[split_sram_offset, split_sram_offset, OR, BUF_SRAM_SPLIT_CHANNEL_/**/freelist_id/**/, <<30]

	sram[write, wxfer_prefix[START_INDEX_2], split_sram_offset, TEMP_SRAM_OFFSET_2, NUM_LW_2], \
		sig_done[req_sig/**/1]

	#undef NUM_LW_1
	#undef NUM_LW_2
	#undef START_INDEX_2
	#undef TEMP_SRAM_OFFSET_2

	#endif

	#undef START_INDEX
	#undef TEMP_SRAM_OFFSET

	#else	//	(_DL_SPLIT_META_DATA == TRUE)

	.reg sram_offset

	; Get the SRAM offset for the meta data 

	dl_buf_get_desc(sram_offset, buf_handle, freelist_id)

	; Write the meta data to sram

	#define_eval START_INDEX (START_LW)
	#define_eval START_BYTE  (START_LW*4)

	sram[write, wxfer_prefix[START_INDEX], sram_offset, START_BYTE, NUM_LW],\
											 sig_done[req_sig]

	#undef	START_INDEX
	#undef	START_BYTE

	#endif // (_DL_SPLIT_META_DATA == TRUE)

	#undef _DL_SPLIT_META_DATA

	#if (!streq('sig_action', 'SIG_NONE'))
	ctx_arb[sig_action]
	#endif

.end
#endm


///////////////////////////////////////////////////////////////////////////////
// dl_meta_write:
//
// Description: 
//   Writes meta data from GPR to SRAM.
//			
//   Outputs:
//
//   Inputs:
//	   wxfer_prefix : Write transfer register prefix (using xbuf_alloc).
//                    At least NUM_LW wxfer registers should have been allocated
//     buf_handle   : Buffer handle 
//     freelist_id  : Id of the buffer freelist that the handle belongs to.
//                    (freelist_id MUST be a constant).
//     buffer_next -: These fields contain the meta-data that needs to be
//     packet_next    written to SRAM.
//     req_sig      : Signal to be used in I/O operation
//                    If 'DL_SPLIT_META_DATA' is defined and the meta data to
//                    flush spans two SRAM channels, signal 'req_sig1' will
//                    be used by this macro. It needs to be defined by the code
//                    that calls this macro.
//     sig_action   : What to do with I/O operation - 
//                      - wait for signal(s) as specified by sig_action
//                      - do not wait for signal, just return (SIG_NONE)
//     START_LW	    : The starting LW from which NUM_LW long words will be flushed.
//     NUM_LW       : Number of long words to flush. Must be a constant.
//
//   Size:   				
//
///////////////////////////////////////////////////////////////////////////////
#macro dl_meta_write(wxfer_prefix, buf_handle, freelist_id, buffer_next, offset, 
					  buffer_size, packet_size, free_list_id, rx_stat, header_type, \
					  input_port, output_port, fabric_port, next_hop_id, flow_color_id, \
					  class_id, packet_next, req_sig, sig_action, START_LW, NUM_LW)

.begin

	// If 'DL_SPLIT_META_DATA' is defined and the meta data to
	// write spans two SRAM channels, signal 'req_sig1' will
	// be used by this macro. It needs to be defined by the code
	// that calls this macro.

	#if (!isnum(freelist_id))
	#error "freelist_id must be a constant."
	#endif

	#if (!isnum(NUM_LW))
	#error	"argument NUM_LW to dl_flush_meta should be a constant"
	#endif

	#if (NUM_LW == 0)
	#error	"argument NUM_LW to dl_flush_meta cannot be zero"
	#endif

	#if (NUM_LW > META_CACHE_SIZE)
	#error	"argument NUM_LW) to dl_flush_meta cannot be greater than META_CACHE_SIZE"
	#endif

	#if (START_LW == 0)
	alu[wxfer_prefix[0], --, b, buffer_next]
	#endif

	#define_eval END_LW (START_LW + NUM_LW)

	#if ((START_LW <= 1) && (END_LW > 1)) 
	alu[wxfer_prefix[1], offset, OR, buffer_size, <<16]
	#endif

	#if ((START_LW <= 2) && (END_LW > 2)) 
	.reg meta2
	alu[meta2, header_type, OR, packet_size, <<16]
	alu[meta2,  meta2, OR, rx_stat, <<8]
	alu[wxfer_prefix[2], meta2, OR, free_list_id, <<12]
	 #endif

	#if ((START_LW <= 3) && (END_LW > 3))
	alu[wxfer_prefix[3], output_port, OR, input_port, <<16]
	#endif

	#if ((START_LW <= 4) && (END_LW > 4))
	alu[wxfer_prefix[4], fabric_port, OR, next_hop_id, <<16]
	#endif

	#if ((START_LW <= 5) && (END_LW > 5))
	alu[wxfer_prefix[5], --, b, flow_color_id]
	#endif

	#if ((START_LW <= 6) && (END_LW > 6))
	alu[wxfer_prefix[6], --, b, class_id, <<16]
	#endif

	#if ((START_LW <= 7) && (END_LW > 7))
	alu[wxfer_prefix[7], --, b, packet_next]
	#endif

	#undef END_LW

	#define_eval _DL_SPLIT_META_DATA		DL_SPLIT_META_DATA_/**/freelist_id/**/

	#if (_DL_SPLIT_META_DATA == TRUE)

	; Meta-data in IXP2800 reference designs is split into 2 sram channels.
	; While loading or flushing the meta-data we need to account for this.

	.reg sram_offset

	; Get the SRAM offset for the meta data 
	dl_buf_get_desc[sram_offset, buf_handle]

	; Write the meta data to sram

	#define_eval START_INDEX 		(START_LW)
	#define_eval TEMP_SRAM_OFFSET 	(START_LW*4)

	#if (START_LW >= BUF_SRAM_SPLIT_LW_/**/freelist_id/**/)

	; All LW's are in the split SRAM channel

	.reg split_sram_offset

	; Clear the channel number
	alu[split_sram_offset, sram_offset, AND~, 3, <<30]

	; Set the channel number
	alu[split_sram_offset, split_sram_offset, OR, BUF_SRAM_SPLIT_CHANNEL_/**/freelist_id/**/, <<30]

	sram[write, wxfer_prefix[START_INDEX], split_sram_offset, TEMP_SRAM_OFFSET, NUM_LW], \
		sig_done[req_sig]

	#elif (START_LW + NUM_LW - 1 < BUF_SRAM_SPLIT_LW_/**/freelist_id/**/)

	; All LW's are in the default SRAM channel

	.reg default_sram_offset

	; Clear the channel number
	alu[default_sram_offset, sram_offset, AND~, 3, <<30]

	; Set the channel number
	alu[default_sram_offset, default_sram_offset, OR, BUF_SRAM_DEFAULT_CHANNEL_/**/freelist_id/**/, <<30]

	sram[write, wxfer_prefix[START_INDEX], default_sram_offset, TEMP_SRAM_OFFSET, NUM_LW], \
		sig_done[req_sig]

	#else

	; LW's are in split and default SRAM channels

	#define_eval NUM_LW_1				(BUF_SRAM_SPLIT_LW_/**/freelist_id/**/ - START_LW)

	.reg default_sram_offset

	; Clear the channel number
	alu[default_sram_offset, sram_offset, AND~, 3, <<30]

	; Set the channel number
	alu[default_sram_offset, default_sram_offset, OR, BUF_SRAM_DEFAULT_CHANNEL_/**/freelist_id/**/, <<30]

	sram[write, wxfer_prefix[START_INDEX], default_sram_offset, TEMP_SRAM_OFFSET, NUM_LW_1], \
		sig_done[req_sig]

	#define_eval NUM_LW_2				(NUM_LW - NUM_LW_1)
	#define_eval START_INDEX_2			(BUF_SRAM_SPLIT_LW_/**/freelist_id/**/)
	#define_eval TEMP_SRAM_OFFSET_2		(START_INDEX_2*4)

	.reg split_sram_offset

	; Clear the channel number
	alu[split_sram_offset, sram_offset, AND~, 3, <<30]

	; Set the channel number
	alu[split_sram_offset, split_sram_offset, OR, BUF_SRAM_SPLIT_CHANNEL_/**/freelist_id/**/, <<30]

	sram[write, wxfer_prefix[START_INDEX_2], split_sram_offset, TEMP_SRAM_OFFSET_2, NUM_LW_2], \
		sig_done[req_sig/**/1]

	#undef NUM_LW_1
	#undef NUM_LW_2
	#undef START_INDEX_2
	#undef TEMP_SRAM_OFFSET_2

	#endif

	#undef START_INDEX
	#undef TEMP_SRAM_OFFSET

	#else	//	(_DL_SPLIT_META_DATA == TRUE)

	.reg sram_offset

	; Write the meta data to sram

	dl_buf_get_desc[sram_offset, buf_handle]

	#define_eval START_INDEX (START_LW)
	#define_eval START_BYTE  (START_LW*4)

	sram[write, wxfer_prefix[START_INDEX], sram_offset, START_BYTE, NUM_LW],\
											 sig_done[req_sig]

	#undef	START_INDEX
	#undef	START_BYTE

	#endif // (_DL_SPLIT_META_DATA == TRUE)

	#undef _DL_SPLIT_META_DATA

	#if (!streq('sig_action', 'SIG_NONE'))
	ctx_arb[sig_action]
	#endif

.end
#endm


#endif // __DL_META_MULTI_UC__
