/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2001 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2X00 Network Processor, Release 6
 *
 * = LIBRARY
 *      Ethernet ARP Library
 *
 *
 * = FILENAME
 *      ix_cc_arp.h
 *
 * = DESCRIPTION
 *      This file will define all exported data types used in the Ethernet ARP library. Also
 *      will declare all function prototypes of the exported interface. 
 *      The Ethernet ARP library implements the protocol to obtain Ethernet header, including Ethernet destination
 *      and source address and protocol fields, for the given IP packet. Also ARP library manages ARP cache table and 
 *      aging of ARP addresses.
 *
 *
 * = AKNOWLEDGEMENTS
 *      
 *
 * = CREATION TIME
 *      11/17/2002 10:11:08 AM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 * $Id: ix_cc_arp.h,v 1.21.2.1 2004/03/09 19:44:31 ktseng Exp $ 
 */

#if !defined(__IX_CC_ARP_H__)
#define __IX_CC_ARP_H__

#include "ix_types.h"
#include "ix_error.h"

#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */


/* from RFC 1122  - ARP related*/
/*

                                                  |       | | | |S| |
                                                  |       | | | |H| |F
                                                  |       | | | |O|M|o
                                                  |       | |S| |U|U|o
                                                  |       | |H| |L|S|t
                                                  |       |M|O| |D|T|n
                                                  |       |U|U|M| | |o
                                                  |       |S|L|A|N|N|t
                                                  |       |T|D|Y|O|O|t
FEATURE                                           |SECTION| | | |T|T|e
--------------------------------------------------|-------|-|-|-|-|-|--
                                                  |       | | | | | |
Trailer encapsulation                             |2.3.1  | | |x| | |
Send Trailers by default without negotiation      |2.3.1  | | | | |x|
ARP                                               |2.3.2  | | | | | |
  Flush out-of-date ARP cache entries             |2.3.2.1|x| | | | |
  Prevent ARP floods                              |2.3.2.1|x| | | | |
  Cache timeout configurable                      |2.3.2.1| |x| | | |
  Save at least one (latest) unresolved pkt       |2.3.2.2| |x| | | |
*/

/**
 * 
 * Typedefs used by ARP Library
 * 
 */
/**
 * TYPENAME: ix_ether_addr
 * 
 * DESCRIPTION: Forward type definition for ethernet address structure.
 *
 */
typedef struct ix_s_ether_addr ix_ether_addr; 

 /**
 * TYPENAME: ix_cc_arp_local_ip_lookup
 * 
 * DESCRIPTION: Type definition for lookup of local ip address.
 *
 */
typedef ix_error (*ix_cc_arp_local_ip_lookup)(void* arg_pContext,ix_uint32 arg_IPv4Addr,
                                              ix_ether_addr* arg_pL2SrcAddr);

/**
 * TYPENAME: ix_cc_arp_local_port_lookup
 * 
 * DESCRIPTION: Type definition for lookup for local port id
 *
 */
typedef ix_error (*ix_cc_arp_local_port_lookup)(void* arg_pContext,ix_uint32 arg_PortId,
                                                    ix_ether_addr* arg_pL2SrcAddr, 
                                                    ix_uint32* arg_pIpAddr);
/**
 * TYPENAME: ix_cc_arp_send_pkt
 * 
 * DESCRIPTION: Type definition for lookup into local interface table callback.
 *
 */
typedef void (*ix_cc_arp_send_pkt)(ix_buffer_handle arg_hBuffer);


/**
 *  Definition of constants used by ARP library 
 *  
 *      
 */     
/*
 *      IEEE 802.3 Ethernet magic constants.  The frame sizes omit the preamble
 *      and FCS/CRC (frame check sequence). 
 */

#define IX_ETH_ALEN        6               /* Bytes in one ethernet addr   */
#define IX_ETH_HLEN        14              /* Total bytes in header.       */
#define IX_ETH_ZLEN        60              /* Min. bytes in frame sans FCS */
#define IX_ETH_DATA_LEN    1500            /* Max. bytes in payload        */
#define IX_ETH_FRAME_LEN   1514            /* Max. bytes in frame sans FCS */

/*
 *      These are the defined Ethernet Protocol ID's.
 */

#define IX_ETH_P_LOOP      0x0060          /* Ethernet Loopback packet     */
#define IX_ETH_P_IP        0x0800          /* Internet Protocol packet     */
#define IX_ETH_P_X25       0x0805          /* CCITT X.25                   */
#define IX_ETH_P_ARP       0x0806          /* Address Resolution packet    */


/* ARP protocol HARDWARE identifiers. */
#define IX_ARPHRD_ETHER    1               /* Ethernet               */

/* ARP protocol header length */
#define IX_ARP_HLEN        8
/* ARP protocol IP length value */
#define IX_IPV4_ALEN       4

 /* ARP protocol opcodes. */
#define IX_ARPOP_REQUEST   1               /* ARP request                  */
#define IX_ARPOP_REPLY     2               /* ARP reply                    */
#define IX_ARPOP_RREQUEST  3               /* RARP request                 */
#define IX_ARPOP_RREPLY    4               /* RARP reply                   */
#define IX_ARPOP_InREQUEST 8               /* InARP request                */
#define IX_ARPOP_InREPLY   9               /* InARP reply                  */
#define IX_ARPOP_NAK       10              /* (ATM)ARP NAK                 */

/* Some ARP helper defines */
#define IX_LOOPBACK(x)     (((x) & IX_HTON32(0xff000000)) == IX_HTON32(0x7f000000))
#define IX_MULTICAST(x)    (((x) & IX_HTON32(0xf0000000)) == IX_HTON32(0xe0000000))
#define IX_BADCLASS(x)     (((x) & IX_HTON32(0xf0000000)) == IX_HTON32(0xf0000000))
#define IX_ZERONET(x)      (((x) & IX_HTON32(0xff000000)) == IX_HTON32(0x00000000))
#define IX_LOCAL_MCAST(x)  (((x) & IX_HTON32(0xFFFFFF00)) == IX_HTON32(0xE0000000))

/**    
 *  Non-protocol specific ARP Library  definitions 
 *  
 *     
 */ 
#define IX_CC_ARP_EVENT_PRIORITY        1
#define IX_CC_ARP_CACHE_TIMEOUT         300000  /* cache time out in milliseconds - 5 minutes*/
#define IX_CC_ARP_CACHE_ENTRY_TIMEOUT   1200 /* arp cache entry timeout in seconds - 20 minutes */
#define IX_CC_ARP_L2_TABLE_NAME         "DEFAULT"
#define IX_ARP_PKT_LEN                  48
#define IX_CC_ARP_ETH_MIN_PKT_LEN       60      /* excluding 4-byte FCS */
#define IX_CC_ARP_PKT_PADDING_LEN       (IX_CC_ARP_ETH_MIN_PKT_LEN - IX_ARP_PKT_LEN)
#define IX_BUFFER_LEN					2048 /* ixa sdk buffers */
#define IX_ARP_MAX_TX					5  /* maximum number of transmited requests per entry */
#define IX_ARP_XMIT_WAIT				20 /* time in seconds for flood avoidance algorithm */
/**    
 *  ARP Enumerations 
 *  
 *     
 */ 
/* enumerate packet type */
enum ix_arp_pkt_type
{
    IX_CC_ARP_REPLY, 
	IX_CC_ARP_REQUEST,
	IX_CC_IP		            /* reserved for buffered IP packets */
};
    
/**
 *    
 * ARP Data Structures 
 *  
 */

/**
 * Data Structure: ix_arp_init_context
 *                 Ethernet CC uses this structure to initialize ARP module
 *  
 *  
 */

typedef struct ix_s_arp_init_context
{
    void* pContext; /* caller context, opaque to ARP module. Most likely, Eth CC will identify local interface */
                    /* table with this pointer. Used in two following callbacks */
    ix_cc_arp_local_ip_lookup   pIpLkupCb; /* local ip lookup function pointer */
    ix_cc_arp_local_port_lookup pPortLkupCb; /* local port lookup function pointer */
    ix_cc_arp_send_pkt          pArpSendPkt; /* send ARP packet function pointer - the function implementation 
                                                should free the buffer in case of error*/
    ix_cc_init_context*         pCcInitContext; /* pointer to core components init context */ 
}ix_arp_init_context;

struct ix_s_ether_addr
{
    ix_uint8 etherAddr[IX_ETH_ALEN];   /* Ethernet address */
} __attribute__ ((packed));

typedef struct ix_s_ether_hdr
{
    ix_ether_addr dstAddr;      /* Ethernet destination address */
    ix_ether_addr srcAddr;      /* Ethernet source address */
    ix_uint16     protFld;      /* packet type ID field */  
} __attribute__ ((packed)) ix_ether_hdr;


/*
 *      This structure defines an ethernet arp header.
 */

typedef struct ix_s_arp_hdr
{
    ix_uint16 arHrd;         /* format of hardware address   */
    ix_uint16 arPro;         /* format of protocol address   */
    ix_uint8  arHln;         /* length of hardware address   */
    ix_uint8  arPln;         /* length of protocol address   */
    ix_uint16 arOp;          /* ARP opcode (command)         */
}ix_arp_hdr;


typedef struct ix_s_cc_arp_next_hop_info
{
    ix_uint32       nextHopIp;  /* next hop IP          */
    ix_uint16       outputPort; /* egress output port   */
    ix_uint16       l2Idx;       /* layer-2 index       */
    ix_ether_addr   l2Addr;     /* layer-2 address      */
} ix_cc_arp_next_hop_info;

typedef struct ix_s_cc_arp_cache_entry
{
    ix_uint32           destIp;	        /* Destination IP - corresponds with Next Hop Ip Address */
    ix_ether_addr       destEthAddr;	/* hardware destination address corresponding with Next Hop Ip Address */
    ix_ossl_time_t      expTime;        /* time value for aging and timing out the entry */
	ix_ossl_time_t		lastSend;       /* time stamp when arp request was sent for the entry - used to prevent arp flooding */
    ix_uint32           arpXmit;	    /* value to manage ARP packets transmission to prevent flooding */
    ix_uint16           arpFlgs;       /* flags indicating valid, inuse or permanent entry */
    ix_buffer_handle    pktHolder;	    /* holds IP packet while hardware address gets resolved from the entry */	
    ix_uint32           l2Idx;	        /*	index into shared L2 table */
} ix_cc_arp_cache_entry;
/* try to keep this entry 32 byte */
#define IX_ARP_CACHE_ENTRY_SIZE sizeof(ix_cc_arp_cache_entry);
/**
 *	ARP flags values
**/
#define IX_ARPF_INUSE	0x01 /* entry in use */
#define IX_ARPF_VALID	0x02 /* valid entry - valid dest hw address */
#define IX_ARPF_PERM	0x04 /* permanent entry - does not timeout */

/**
 *    
 *  ARP Functions 
 *  
 */

ix_error ix_cc_arp_init(ix_arp_init_context* arg_pInitContext); 
ix_error ix_cc_arp_fini(void);

/* arp cache operations */
ix_error ix_cc_arp_create_entry(ix_cc_arp_next_hop_info* arg_pArpInfo,
                                        ix_buffer_handle* arg_pRtnPkt); 
ix_error ix_cc_arp_add_entry(ix_cc_arp_next_hop_info* arg_pArpInfo); 
ix_error ix_cc_arp_update_entry(ix_cc_arp_next_hop_info* arg_pArpInfo); 
ix_error ix_cc_arp_del_entry(ix_uint32 arg_L2Index);

/* arp protocol operations */
ix_error ix_cc_arp_resolve_l2_addr(ix_buffer_handle arg_Pkt, ix_buffer_handle* arg_pRtnPkt); 
ix_error ix_cc_arp_process_arp_pkts(ix_buffer_handle arg_ArpPkt,ix_buffer_handle* arg_pRtnPkt,
                                    ix_uint32* arg_pPktType); 

ix_error ix_cc_arp_create_gratuitous_arp(ix_uint16 arg_OutputPort,
                                        ix_buffer_handle *arg_pRtnPkt); 

/* debugging and management support */
ix_error ix_cc_arp_purge(void); 
ix_error ix_cc_arp_dump(void);


/* event handling */
ix_error ix_arp_create_event_handler(void* arg_pContext);
ix_error ix_arp_cache_timeout(void* pContext);

/**
 *    
 *  ARP Error codes
 *  
 */
enum ix_cc_arp_error {
        IX_CC_ARP_ERROR_FIRST = IX_ERROR_MAKE_GROUP(CC_ARP),
        IX_CC_ARP_ERROR_NEW_ENTRY_CREATED, 
        IX_CC_ARP_ERROR_INVALID_L2_INDEX, 
        IX_CC_ARP_ERROR_INCORRECT_HEADER, 
        IX_CC_ARP_ERROR_INVALID_PARAMETER,
        IX_CC_ARP_INVALID_HANDLE,
        IX_CC_ARP_ERROR_NHIP_NOT_FOUND,
        IX_CC_ARP_ERROR_NO_CACHE_ENTRY,
        IX_CC_ARP_ERROR_SIP_CONFLICT,
        IX_CC_ARP_ERROR_NO_LOCAL_MAC_ADDRESS,
        IX_CC_ARP_ERROR_ENTRY_EXIST,
		IX_CC_ARP_FAIL_UPDATE_L2TM,
		IX_CC_ARP_ERROR_EVENT_HANDLER_GET_TIME,
        IX_CC_ARP_ERROR_CCI,
        IX_CC_ARP_ERROR_LAST
};
#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__IX_CC_ARP_H__) */



