/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      IPv4 Forwarder Core Component (Fragmentation Handling)
 *
 * = FILENAME
 *      fragment.h
 *
 * = DESCRIPTION
 *      This file contains definition of fragmentation routines
 *
 *
 *      
 *
 * = CREATION TIME
 *      07/12/2002 9:39:00 PM
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 * $Id: fragment.h,v 1.10 2003/04/14 21:55:22 rranjeet Exp $
 */

#if !defined(__FRAGMENT_H__)
#define __FRAGMENT_H__


#include "ip4hdr.h"
#include "checksum.h"

#if defined(__cplusplus)
extern "C"
{
#endif /* end defined(__cplusplus) */

/**
 * Preprocessor symbol and macro definitions.
 */


/* min fragment: min header size + 8 bytes of data
   to make sure fragment offset is multiple of 8 */
#define IX_CC_FRAGV4_IP_FRAG_MIN_SIZE        0x1c
#define IX_CC_FRAGV4_IP_FRAG_MAX_SIZE        0x10000

#define	IX_CC_FRAGV4_IP_FRAG_OFF_MASK	0x1fff
#define	IX_CC_FRAGV4_IP_FRAG_MF		0x2000
#define	IX_CC_FRAGV4_IP_FRAG_DF		0x4000


/**
 * Type definitions.
 */
/** Structure to represent IPv4 Fragment
*/
typedef struct ix_s_cc_fragv4_fragment ix_cc_fragv4_fragment;
struct ix_s_cc_fragv4_fragment {

    /* pointer to the previous fragment that belongs to the
      same datagram ( maybe NULL ) */
    ix_cc_fragv4_fragment *pPrev;

    /* pointer to the next fragment that belongs to the
           same datagram ( maybe NULL ) */
    ix_cc_fragv4_fragment *pNext;

    /* pointer to the buffer which contains the network packet for 
           this fragment */
    ix_buffer_handle hBuffer;
    ix_uint16 bufferChaining; /* This is to know if buffer_chaining was done**/
    /** If buffer chaining was done, then buffer represents first buffer 
    **in the chain for this fragment **/        

    /* pointer to IP4 header ( a location in the buffer ) */
    ix_cc_fragv4_ip4hdr *pHeader;        
   
    /*16-bit value that include 3-bit flags and 
    offset IN 8-BYTE uint of this fragment from
     octet 0 of the original data. This is from the fragment 
     header. It's kept here for quick and easy access */

    ix_uint16 flagsAndOff;
    /* length of payload in this fragment measured in octets. This
       is IP.LEN - IP.HLEN. 
       It's kept here for quick and easy access */
    ix_uint16 dataLen;

};




/**
 * Prototypes for interface functions.
 */
/** Retrieves data from the chained buffers

    @param      arg_hBuffer     buffer handle with the chain
    @param      pOrigData     location which will contain the data address
                              for the entire data contained in the chain
    
    @returns an error token, IX_SUCCESS for success.
*/

ix_error _ix_cc_fragv4_get_data_from_chain(
                          ix_buffer_handle arg_hBuffer, 
                          void **pOrigData);


/** Create a new fragment from a datagram ( or a bigger fragment ) that came as a chain of buffers

 @param      arg_FragSize        size of the fragment ( header + data ) in the
                                number of octets
 @param      arg_pData        pointer to data of Packet buffer in the fragment or datagram
                                from which a new fragment is being created.
                                The fragment contains this buffer will be
                                refered to as "immediate parent" of the new
                                fragment.We copy data from this buffer to the 
                                buffer of the newly created fragment

 @param      arg_pMeta        pointer to meta data of original buffer that came in

 @param      arg_FlagsAndOff      16-bit value that include: 
                                - 3-bit flags ( do-not-fragment flag and
						more-fragment flag ) 
                                - absolute offset from data byte 0 of 
				the original UNFRAGMENTED datagram
                                to the data being carried in this fragment
                                ( measured in 8 octets units ). 
                                This operand is the the FLAGS and
                                FRAGMENT OFFSET field ( byte 6,7 ) in the IP
				header of the fragment created by this function

 @param      arg_DataOff         relative offset ( measured in the number of
                                bytes ) from the first byte of the "immediate 
				parent" fragment, ie. the fragment that contains
                                the header and payload being copied to
                                the fragment being created. 
                                If the "immediate parent" of the new fragment 
				is UNFRAGMENTED, this 'data_off' is the same as
                                the offset part of 'flags_and_off' operand.
                                If the "immediate parent" of the new fragment 
				is FRAGMENTED, 'data_off' is the number of bytes 
			        from byte 0 of the "immediate parent" fragment 	

  @param      arg_Length          length of payload in this fragment ( measured
				in octets )
    
  @param      arg_IsFirst         whether this is the first fragment of a
                                  datagram
  @param      pOrigData         pointer to data in original buffer chain

  @param      arg_ppNewFrag         where to return the pointer to the new
                                    fragment

    @returns an error token, zero for success.
*/
IX_EXPORT_FUNCTION
ix_error ix_cc_fragv4_fragment_create_from_chain(ix_uint32 arg_FragSize,
                                   ix_uint8 *arg_pData,
                                   ix_hw_buffer_meta *arg_pMeta,
                                   ix_uint16 arg_FlagsAndOff, 
                                   ix_uint16 arg_DataOff,
                                   ix_uint16 arg_Length,
                                   ix_uint32 arg_IsFirst,
                                   ix_buffer_free_list_handle arg_hHwFreeList,
                                   void *pOrigData,
                                   ix_cc_fragv4_fragment **arg_ppNewFrag);




/** Create a new fragment from a datagram ( or a bigger fragment )

 @param      arg_FragSize        size of the fragment ( header + data ) in the
                                number of octets
 @param      arg_pData        pointer to data of Packet buffer in the fragment or datagram
                                from which a new fragment is being created.
                                The fragment contains this buffer will be
                                refered to as "immediate parent" of the new
                                fragment.We copy data from this buffer to the 
                                buffer of the newly created fragment

 @param      arg_pMeta        pointer to meta data of original buffer that came in
 @param      arg_FlagsAndOff      16-bit value that include: 
                                - 3-bit flags ( do-not-fragment flag and
						more-fragment flag ) 
                                - absolute offset from data byte 0 of 
				the original UNFRAGMENTED datagram
                                to the data being carried in this fragment
                                ( measured in 8 octets units ). 
                                This operand is the the FLAGS and
                                FRAGMENT OFFSET field ( byte 6,7 ) in the IP
				header of the fragment created by this function

 @param      arg_DataOff         relative offset ( measured in the number of
                                bytes ) from the first byte of the "immediate 
				parent" fragment, ie. the fragment that contains
                                the header and payload being copied to
                                the fragment being created. 
                                If the "immediate parent" of the new fragment 
				is UNFRAGMENTED, this 'data_off' is the same as
                                the offset part of 'flags_and_off' operand.
                                If the "immediate parent" of the new fragment 
				is FRAGMENTED, 'data_off' is the number of bytes 
			        from byte 0 of the "immediate parent" fragment 	

  @param      arg_Length          length of payload in this fragment ( measured
				in octets )
    
  @param      arg_IsFirst         whether this is the first fragment of a
                                  datagram

  @param      arg_ppNewFrag         where to return the pointer to the new
                                    fragment

    @returns an error token, zero for success.
*/
IX_EXPORT_FUNCTION
ix_error ix_cc_fragv4_fragment_create(ix_uint32 arg_FragSize,
                                   ix_uint8 *arg_pData,
                                   ix_hw_buffer_meta *arg_pMeta,
                                   ix_uint16 arg_FlagsAndOff, 
                                   ix_uint16 arg_DataOff,
                                   ix_uint16 arg_Length,
                                   ix_uint32 arg_IsFirst,
                                   ix_buffer_free_list_handle arg_hHwFreeList,
                                   ix_cc_fragv4_fragment **arg_ppNewFrag);

/** Create a new fragment to wrap around a given buffer

    @param      arg_hBuffer     buffer which contains fragment
    @param      arg_HdrOff      offset from the beginning of buffer to the 
				beginning of the IP4 header inside the buffer.
    @param      arg_ppNewFrag   where to return the pointer to the new fragment
    @returns an error token, zero for success.
*/
IX_EXPORT_FUNCTION
ix_error ix_cc_fragv4_fragment_make_from_buffer(ix_buffer_handle arg_hBuffer,
                                        ix_uint32 arg_HdrOff,
                                        ix_cc_fragv4_fragment **arg_ppNewFrag);

/** Print a fragment ( for debugging )
    
    @param      arg_pFragment	pointer to the fragment
    @param      arg_PrintHeader
			1: print all fields in the basic 20-bytes header
			0: do not print fields in the header
    @param      arg_PrintData
			1: print all data bytes in the fragment payload
			0: do not print data bytes
    @returns an error token, IX_SUCCESS for success.
*/
IX_EXPORT_FUNCTION
ix_error ix_cc_fragv4_fragment_print( ix_cc_fragv4_fragment *arg_pFragment,
                    ix_uint32 arg_PrintHeader, ix_uint32 arg_PrintData);



/** Fragment a single fragment into smaller fragments with size MTU.
    New fragments with smaller data length will be created to keep
    all data from the original fragment. The original fragment will be 
    unchanged after this function. 

    @param      arg_pFragment    pointer to the fragment of interest
    @param      arg_Mtu     minimum transfer unit size in octets
    @param      arg_ppFirst   pointer to the newly devided first fragment of
                the given fragment
    @returns an error token, zero for success.
*/
IX_EXPORT_FUNCTION
ix_error ix_cc_fragv4_fragment_divide( ix_cc_fragv4_fragment *arg_pFragment, 
                                     ix_uint32 arg_Mtu,
                                    ix_buffer_free_list_handle arg_hHwFreeList,
				    ix_cc_fragv4_fragment **arg_ppFirst );


/** Delete a fragment: release all resources allocated for the fragment

    Note that this function deletes ALL elements of the fragment INCLUDING
    the buffer which contains the packet data.
    This function can be considered as the counterpart of ix_fragment_create
    and ix_fragment_devide. These 2 functions allocate new buffers, so
    there should be a function to delete the allocated buffer in the
    fragment AND the fragment itself.
    If error happens while deleting the buffer, the function returns 
    IMMEDIATELY with the fragment structure still intact. This allows
    users to examine the fragment and to figure out possible reasons for the
    error.

    @param      arg_pFragment     pointer to the fragment
    @returns an error token, zero for success.
*/
IX_EXPORT_FUNCTION
ix_error ix_cc_fragv4_fragment_delete_frag_and_data_buff( 
                                      ix_cc_fragv4_fragment *arg_pFragment);

/** Delete a fragment:  delete the fragment but not the data buffer
 
    Note that this function just delete the fragment, NOT
    the buffer that contains packet data associated with the fragment.
    This function is the counterpart of ix_fragment_make_from_buffer.
    Since users own the buffer that was passed to ix_fragment_make_from_buffer,
    they should take care of deleting the buffer. 

    @param      arg_ptr     pointer to the fragment
    @returns an error token, zero for success.
*/
IX_EXPORT_FUNCTION
ix_error ix_cc_fragv4_fragment_delete_frag_only(
                 ix_cc_fragv4_fragment *arg_pFragment);


/** Print a fragment ( for debugging )
    
    @param      arg_pFragment	pointer to the fragment
    @param      arg_PrintHeader
			1: print all fields in the basic 20-bytes header
			0: do not print fields in the header
    @param      arg_PrintData
			1: print all data bytes in the fragment payload
			0: do not print data bytes
    @returns an error token, IX_SUCCESS for success.
*/
IX_EXPORT_FUNCTION
ix_error ix_cc_fragv4_fragment_print( ix_cc_fragv4_fragment *arg_pFragment,
                                    ix_uint32 arg_PrintHeader,
				ix_uint32 arg_PrintData);

/**
 * Exported variables.
 */

#if defined(__cplusplus)
}
#endif /* end defined(__cplusplus) */

#endif /* end !defined(__FRAGMENT_H__) */
