/*
 *			INTEL CORPORATION PROPRIETARY INFORMATION
 *	This software is supplied under the terms of a license agreement or
 *	nondisclosure agreement with Intel Corporation and may not be copied
 *	or disclosed except in accordance with the terms of that agreement.
 *		Copyright (C) 2002 Intel Corporation. All rights Reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2000 Network Processor
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      IPv4 Forwarder Core Component Fragmentation Library
 *
 * = FILENAME
 *      checksum.c
 *
 * = DESCRIPTION
 *   This file contains functions to manipulate checksum in IP header      
 *
 */
#define IX_ERROR_FILE_IDENT "$Id: checksum.c,v 1.10 2003/11/04 22:15:39 rranjeet Exp $"

#include "ix_ossl.h"
#include "cc/fragmentation/internal/internal-netrw.h"
#include "cc/fragmentation/checksum.h"


/** Calculate sum -- just return 32-bit accumulator of a number of
                     16-bit words in network order byte ( big endian )
    
    @param      arg_pWord	pointer to the first 16-bit word in network byte order 
    @param      len 	length of data to be summed up ( measured in octets )
    @param      arg_pSum 	where to return 32-bit sum
    @returns an error token,or IX_SUCCESS for success.
*/
ix_error ix_cc_fragv4_checksum_sum32( ix_uint16 *arg_pWord, ix_uint32 arg_Len,
                                   ix_uint32 *arg_pSum )
{
    
    ix_uint8 *pTemp;
    ix_uint32 total=0;
    ix_uint16 val;
    ix_int32 words;

    if (arg_pWord == NULL)
        return IX_ERROR_LOCAL(IX_CC_ERROR_NULL,
                            (" Pointer passed is NULL ")); 
    if (arg_pSum == NULL)
        return IX_ERROR_LOCAL(IX_CC_ERROR_NULL,
                            ("arg_pSum is NULL ")); 

    words = arg_Len >> 1;

    pTemp = (ix_uint8 *)arg_pWord;
    while ( words -- > 0 )
    {
#ifdef SIM_NT
        IX_ERROR_CRT(_ix_cc_fragv4_netrw_read_int16(pTemp, &val),
         IX_CC_ERROR_INTERNAL, 0);
#else
        val = IX_NTOH16(IX_RM_MEM_UINT16_READ((ix_uint16*)(ix_uint8*)pTemp));
#endif        
        total += (ix_uint32)val;
        pTemp +=2;

    }
   /* pick up an odd byte */
   if ( arg_Len & 1 )
   {
#ifdef SIM_NT
        IX_ERROR_CRT(_ix_cc_fragv4_netrw_read_int16(pTemp, &val),
         IX_CC_ERROR_INTERNAL, 0);
#else
        val = IX_NTOH16(IX_RM_MEM_UINT16_READ((ix_uint16*)(ix_uint8*)pTemp));
#endif        
        val = val & 0xff00;
        total += (ix_uint32)val;
   }
    /* This sum is returned in host byte order */
    *arg_pSum = total;

    return IX_SUCCESS;
}

/** Calculate sum -- return 16-bit 1's compliment sum of a number of
                             16-bit words. 
    
    @param      arg_pWord	pointer to the first 16-bit word in network byte order
    @param      arg_Len 	length of data to be summed up ( measured in octets )
    @param      arg_pSum 	where to return 16-bit 1's compliment sum
                                This value is returned in host byte order
    @returns an error token, zero for success.
*/
ix_error ix_cc_fragv4_checksum_compliment_sum16( ix_uint16 *arg_pWord,
                                     ix_uint32 arg_Len,
					  ix_uint16 *arg_pSum)
{

    
    ix_uint32 sum32=0;


    if (arg_pWord == NULL)
	    return IX_ERROR_LOCAL(IX_CC_ERROR_NULL,
                            ("Passed pointer is NULL\n")); 

    if (arg_pSum == NULL)
	    return IX_ERROR_LOCAL(IX_CC_ERROR_NULL,
                            ("arg_pSum is NULL ")); 

    IX_ERROR_CR(ix_cc_fragv4_checksum_sum32(arg_pWord, arg_Len, &sum32));
   
    sum32 = ( sum32 & 0xffffu) + (sum32 >> 16);
    sum32 = ( sum32 & 0xffffu) + (sum32 >> 16);
    
    /* This value is returned in host byte order */
    *arg_pSum = (ix_uint16)(~sum32);
    
    return IX_SUCCESS;

}
 

  

/** Calculate header checksum for the basic 20-octet ipv4 header 
    
    @param      arg_pWord	pointer to the ip4 header
    @param      arg_pChksum 	where to return the 16-bit 1's complement 
                                checksum. This checksum is returned in
                                network byte order
    @param	arg_Replace whether to replace the old checksum with the newly
                        calulated checksum.
			1:  replace old checksum
			0:  old checksum unchange
    @returns an error token, or IX_SUCCESS for success.
*/
ix_error ix_cc_fragv4_checksum_ip4hsum( ix_cc_fragv4_ip4hdr *arg_pWord,
                            ix_uint16 *arg_pChksum,
				 ix_uint32 arg_Replace)
{
    
    ix_uint16 sum;
    ix_uint16 *pTemp, oldSum;
    ix_uint16 len;

    if (arg_pWord == NULL)
        return IX_ERROR_LOCAL(IX_CC_ERROR_NULL,
                            ("Passed Pointer is NULL\n")); 
    if (arg_pChksum == NULL)
	    return IX_ERROR_LOCAL(IX_CC_ERROR_NULL,
                            ("Passed Pointer is NULL\n")); 
    oldSum = 0;
    sum  = 0;

    if ( !arg_Replace )
    {
        IX_ERROR_CR(ix_cc_fragv4_ip4hdr_checksum_read(arg_pWord, &oldSum));
    	
    }

    IX_ERROR_CR(ix_cc_fragv4_ip4hdr_checksum_write(arg_pWord, 0));

    
    pTemp = ((ix_uint16 *)arg_pWord);    
    IX_ERROR_CRT(ix_cc_fragv4_ip4hdr_read_hlen_in_octets(arg_pWord, &len),
                 IX_CC_FRAGV4_ERROR_IP4HDR_READ, 0);    


    /* Note that sum is returned in host byte order */
    IX_ERROR_CR(ix_cc_fragv4_checksum_compliment_sum16(pTemp, len, &sum));
   
    if ( !arg_Replace )
    {
        /* Note that oldSum was returned in network byte order */
    	IX_ERROR_CR(ix_cc_fragv4_ip4hdr_checksum_write(arg_pWord, oldSum));
    	
    }
    else 
    {
    	IX_ERROR_CR(ix_cc_fragv4_ip4hdr_checksum_write(arg_pWord, IX_HTON16(sum)));
    
    }
    *arg_pChksum = IX_HTON16(sum);

    return IX_SUCCESS;
}
