/*
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *     Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * $Id: ix_cc_l2tm.h,v 1.7 2003/01/20 23:54:50 lbmiller Exp $
 *
 * Note that the format and location of the comments in this file are
 * significant, as they are used to generate documentation about the module.
 * Take care when modifying this file.
 */

/**
 * @file ix_cc_l2tm.h
 *
 * Public header file for users of the Layer Two Table Manager.  Allows a
 * client to control the L2TM to create and destroy tables, add and
 * remove entries, and look up entries.
 */

#if !defined(__IX_CC_L2TM_H__)
#define __IX_CC_L2TM_H__

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

#include "ix_cc_error.h"
#include "ix_types.h"

/*
 * These are the typedefs for the various exported enumerations and data
 * structures.  They are listed outside of the group to clean up the generated
 * documentation.
 */
typedef enum ix_e_cc_l2tm_ipaddr_type ix_cc_l2tm_ipaddr_type;
typedef enum ix_e_cc_l2tm_l2addr_type ix_cc_l2tm_l2addr_type;
typedef enum ix_e_cc_l2tm_memory_type ix_cc_l2tm_memory_type;
typedef enum ix_e_cc_l2tm_state ix_cc_l2tm_state;
typedef struct ix_s_cc_l2tm_atm_header ix_cc_l2tm_atm_header;
typedef struct ix_s_cc_l2tm_ether_header ix_cc_l2tm_ether_header;
typedef union ix_u_cc_l2tm_ipaddr ix_cc_l2tm_ipaddr;
typedef struct ix_s_cc_l2tm_entry ix_cc_l2tm_entry;
typedef struct ix_s_cc_l2tm_stats ix_cc_l2tm_stats;
typedef struct ix_s_cc_l2tm_symbols ix_cc_l2tm_symbols;
typedef struct ix_s_cc_l2tm_config ix_cc_l2tm_config;

/**
 * @defgroup L2TM Exported API
 * The L2 Table Manager exports an interface for clients to control the L2
 * table.  The interface consists of a number of enumerations, data structures,
 * functions, etc. This section describes the exported API.
 *
 * @{
 */

/**
 * Indicates the type of the IP address to be stored.
 *
 * @typename ix_cc_l2tm_ipaddr_type
 */
enum ix_e_cc_l2tm_ipaddr_type
{
    /**
     * The IP Address is for IPv4, and therefore 32 bits.
     */
    IX_CC_L2TM_IPADDR_IPV4,

    /**
     * The IP Address is for IPv6, and therefore 128 bits.
     */
    IX_CC_L2TM_IPADDR_IPV6,

    /**
     * Last entry in the IP address enumeration.
     */
    IX_CC_L2TM_IPADDR_LAST
};

/**
 * Indicates the type of the header to be stored.
 *
 * @typename ix_cc_l2tm_l2addr_type
 */
enum ix_e_cc_l2tm_l2addr_type
{
    /**
     * The L2 Header is for Ethernet.
     */
    IX_CC_L2TM_L2ADDR_ETHER,

    /**
     * The L2 Header is for ATM.
     */
    IX_CC_L2TM_L2ADDR_ATM,

    /**
     * Last entry in the header enumeration.
     */
    IX_CC_L2TM_L2ADDR_LAST
};

/**
 * Error codes specific to the L2 Table Manager.
 */
enum ix_e_cc_l2tm_error
{
    /**
     * Failed to create the L2 Table.
     */
    IX_CC_L2TM_ERROR_CREATE_L2TABLE = IX_ERROR_MAKE_GROUP(CC_L2TM),

    /**
     * Failed to delete the L2 Table.
     */
    IX_CC_L2TM_ERROR_DELETE_TABLE,

    /**
     * The L2TM with the specified name could not be found.
     */
    IX_CC_L2TM_ERROR_NOT_FOUND,

    /**
     * A problem with the registry was encountered.
     */
    IX_CC_L2TM_ERROR_REGISTRY,

    /**
     * The entry structure contains invalid information.
     */
    IX_CC_L2TM_ERROR_INVALID_ENTRY,

    /**
     * Last entry in the L2TM error enumeration.
     */
    IX_CC_L2TM_ERROR_LAST
};

/**
 * Types of memory supported for creating a new L2 Table.
 *
 * @typename ix_cc_l2tm_memory_type
 */
enum ix_e_cc_l2tm_memory_type
{
    /**
     * Use SRAM for the L2 Table.
     */
    IX_CC_L2TM_MEMORY_TYPE_SRAM,

    /**
     * Use DRAM for the L2 Table.
     */
    IX_CC_L2TM_MEMORY_TYPE_DRAM,

    /**
     * Last entry in the ix_cc_l2tm_memory_type enumeration.
     */
    IX_CC_L2TM_MEMORY_TYPE_LAST
};

/**
 * Enumerations of the states each of the entries could occupy.
 *
 * @typename ix_cc_l2tm_state
 */
enum ix_e_cc_l2tm_state
{
    /**
     * The entry is unused.  This is indicated by the 'in use' flag being zero.
     */
    IX_CC_L2TM_STATE_IDLE,

    /**
     * The entry is in use, but not usable by microblocks.  It has L3
     * information, but no L2 information at this time.  The 'in use' flag is
     * set, and the 'valid' flag in shared memory is zero.
     */
    IX_CC_L2TM_STATE_INCOMPLETE,

    /**
     * The entry is usable by the microblocks.  It has valid L3 information
     * and L2 information.  The 'in use' and 'valid' flags are set.
     */
    IX_CC_L2TM_STATE_VALID,

    /**
     * Last entry in the state enumeration.
     */
    IX_CC_L2TM_STATE_LAST
};

/**
 * Opaque handle to the L2Table Manager.  The handle is generated for each
 * client by calling ix_cc_l2tm_init() or ix_cc_l2tm_create().  Each generated
 * handle should be released by calling ix_cc_l2tm_fini() or
 * ix_cc_l2tm_destroy().
 */
typedef struct _ix_s_cc_l2tm* ix_cc_l2tm;

/**
 * Describes an ATM entry in the L2 Table.  This structure is used when adding
 * an L2 entry or looking up an L2 Entry by Next Hop IP address or L2 Index.
 *
 * @typename ix_cc_l2tm_atm_header
 */
struct ix_s_cc_l2tm_atm_header
{
    /**
     * Virtual Path Identifier / Virtual Channel Identifier.  The low 28 bits
     * are kept for the microblocks.  The high 4 bits are reserved and must be
     * zero.  The L2TM does not validate the VPI/VCI fields, other than to
     * limit it to 28 bits.
     */
    ix_uint32 vpiVci;

    /**
     * Virtual Channel Queue.  The low 16 bits are stored for the microblocks.
     * The high 16 bits are reserved and must be zero.  The L2TM does not
     * validate the vcq, other than to limit it to 16 bits.
     */
    ix_uint32 vcq;

    /**
     * LLC / SNAP Header.  The L2TM does not validate the LLC and SNAP
     * headers.
     */
    ix_uint64 llcSnapHeader;
};

/**
 * Describes an Ethernet entry in the L2 Table.  This structure is used when
 * adding an L2 entry or looking up an L2 Entry by Next Hop IP address or L2
 * Index.
 *
 * @typename ix_cc_l2tm_ether_header
 */
struct ix_s_cc_l2tm_ether_header
{
    /**
     * Destination MAC address.  The L2TM does not validate the destination
     * MAC address.
     */
    ix_uint8 destMAC[6];

    /**
     * Source MAC address.  The L2TM does not validate the source MAC address.
     */
    ix_uint8 sourceMAC[6];

    /**
     * Protocol type.  The L2TM does not validate the protocol.
     */
    ix_uint16 protocol;
};

/**
 * Union containing an IP address.  When using this union, a corresponding
 * ix_cc_l2tm_ipaddr_type indicates which address is in use.
 *
 * @typename ix_cc_l2tm_ipaddr
 * @see ix_cc_l2tm_ipaddr_type
 */
union ix_u_cc_l2tm_ipaddr
{
    /**
     * 32-bit address for IPv4.
     */
    ix_uint32       ipv4Address;

    /**
     * 128-bit address for IPv6.
     */
    ix_uint128      ipv6Address;
};

/**
 * Describes an entry in the L2 Table.  This structure is used when adding
 * an L2 Entry, retrieving an L2 Entry, or looking up an L2 Entry by Next Hop
 * IP address or L2 Index.
 *
 * Though this data is identical in meaning to the L2 Table Entry used by the
 * Interface TX microblock, the structure used to store the data is less
 * restrictive.  The L2TM internally packs the L2 Entry data into the
 * format the microblocks expect.
 *
 * @typename ix_cc_l2tm_entry
 */
struct ix_s_cc_l2tm_entry
{
    /**
     * Indicates the type of header to be added, or the type of header stored
     * in the specified location.
     */
    ix_cc_l2tm_l2addr_type      l2AddrType;

    /**
     * Indicates the type of address to be added, or the type of address
     * stored in the specified location.
     */
    ix_cc_l2tm_ipaddr_type      ipAddrType;

    /**
     * Specifies the L2 header of the entry being added or looked up.
     */
    union
    {
        /**
         * Header for an Ethernet entry.
         */
        ix_cc_l2tm_ether_header ether;

        /**
         * Header for an ATM entry.
         */
        ix_cc_l2tm_atm_header   atm;
    } header;

    /**
     * Specifies the L3 address of the entry being added or looked up.
     */
    ix_cc_l2tm_ipaddr   l3Address;
};

/**
 * Statistics about the L2 Table Manager and its associated table.  The
 * statistics represent a snapshot of the state of the L2TM and table,
 * as well as some static information such as the number of entries the
 * table could contain.
 *
 * @typename ix_cc_l2tm_stats
 * @see ix_cc_l2tm_get_statistics
 */
struct ix_s_cc_l2tm_stats
{
    /**
     * Number of 'Valid' L2 entries currently contained within the L2 Table.
     */
    ix_uint32   numberOfValidEntries;

    /**
     * Number of 'Incomplete' L2 entries currently contained within the L2
     * Table.
     */
    ix_uint32   numberOfIncompleteEntries;

    /**
     * Maximum number of entries the L2 Table could contain.
     */
    ix_uint32   tableCapacity;

    /**
     * Amount of heap memory used for the L2 Table Manager.
     */
    ix_uint32   heapMemoryUsed;

    /**
     * Amount of shared memory used for the L2 Table.
     */
    ix_uint32   sharedMemoryUsed;
};

/**
 * Contains values useful to microblocks.  This structure is returned by
 * ix_cc_l2tm_get_symbols().
 *
 * @typename ix_cc_l2tm_symbols
 */
struct ix_s_cc_l2tm_symbols
{
    /**
     * Base address of the L2 Table.
     */
    ix_uint32   l2TableBase;
};

/**
 * Configuration structure for creating a new L2 Table.
 *
 * @typename ix_cc_l2tm_config
 */
struct ix_s_cc_l2tm_config
{
    /**
     * Requested number of L2 entries for the table.
     */
    ix_uint32                   numEntries;

    /**
     * Type of memory to use for L2 Table.
     */
    ix_cc_l2tm_memory_type      memoryType;

    /**
     * Channel of memory to use for the L2 Table.  The allowable range of
     * channels for a given memory type is determined by the Resource
     * Manager.
     */
    ix_uint32                   memoryChannel;
};

/**
 * Returns a handle to a named L2 Table Manager. If the named L2TM does not
 * yet exist, its parameters are retrieved from the registry and the L2TM
 * is created.  If the L2TM already exists, it is provided to the caller.
 *
 * A named L2 Table Manager should be used when a single table must be shared
 * among multiple clients. If a table is intended to be used for only a
 * single client, the client could choose to use ix_cc_l2tm_create() or
 * ix_cc_l2tm_init().
 *
 * The name provided must match a configuration in the registry.  All clients
 * sharing a table must invoke ix_cc_l2tm_create() with the same name.
 *
 * A handle obtained using ix_cc_l2tm_create() should be freed using
 * ix_cc_l2tm_destroy().
 *
 * @param       arg_pName (IN)
 *              Name of the L2 Table Manager to be created.
 *
 * @param       arg_phL2tm (OUT)
 *              Handle to the requested L2 Table Manager.
 *
 * @if LLD @pseudocode
 * -# Look up the handle in the Registry
 * -# If the handle doesn't already exist:
 *      -# Read the configuration from the Registry
 *      -# Create a new L2TM using the configuration
 *      -# Assign the name to the new L2TM.
 *      -# Set the handle in the registry
 * -# Increment the reference count for the handle
 * -# Return the handle to the caller
 * @endif
 *
 * @return IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_ERROR_RANGE    One of the provided values was out of range.
 * @retval IX_CC_ERROR_NULL     A required pointer was NULL.
 * @retval IX_CC_ERROR_ALIGN    A pointer was not properly aligned.
 * @retval IX_CC_ERROR_OOM_SYSTEM Memory could not be allocated to create the
 *      L2TM handle.
 * @retval IX_CC_L2TM_ERROR_CREATE_L2TABLE The L2TM failed to create the L2
 *      Table.
 * @retval IX_CC_L2TM_ERROR_NOT_FOUND The L2TM with the specified name could
 *      not be found.
 * @retval IX_CC_L2TM_ERROR_REGISTRY A problem with the registry was
 *      encountered.
 */
ix_error ix_cc_l2tm_create(
    const char *arg_pName,
    ix_cc_l2tm *arg_phL2tm);

/**
 * Destroys a named L2 Table Manager. The named L2 Table Manager is freed
 * when the last client calls this function. This function should only be
 * used with handles obtained from ix_cc_l2tm_create().
 *
 * Depending on the runtime environment, ix_cc_l2tm_destroy() sometimes invokes
 * ix_cc_l2tm_fini(); see the warnings about calling ix_cc_l2tm_fini().
 * 
 * @param       arg_hL2tm (IN)
 *              Handle to the named L2 Table Manager to be destroyed.
 *
 * @if LLD @pseudocode
 * -# Get the L2TM's name
 * -# Decrement the handle's refcount
 * -# If the refcount is zero
 *      -# Destroy the L2TM
 *      -# Clear the handle from the Registry
 * @endif
 *
 * @return IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_ERROR_NULL     A required pointer was NULL.
 * @retval IX_CC_ERROR_ALIGN    A pointer was not properly aligned.
 * @retval IX_CC_L2TM_ERROR_DELETE_TABLE The L2TM was unable to free the
 *      resources for the L2 Table.
 * @retval IX_CC_L2TM_ERROR_REGISTRY A problem with the registry was
 *      encountered.
 *
 * @see ix_cc_l2tm_fini
 */
ix_error ix_cc_l2tm_destroy(
    ix_cc_l2tm arg_hL2tm);

/**
 * Creates a new L2TM for the client. Note that this function always
 * initializes a new unnamed L2TM; if the L2TM is intended to be shared with
 * other clients, use ix_cc_l2tm_create() instead.
 *
 * A handle obtained using ix_cc_l2tm_init() should be freed using
 * ix_cc_l2tm_fini().
 *
 * @param       arg_pConfig (IN)
 *              Configuration to use when creating the new L2 Table.
 *
 * @param       arg_phL2tm (OUT)
 *              Handle to the L2 Table Manager.
 *
 * @if LLD @pseudocode
 * -# Allocate memory from the OSSL for the handle's data structure
 * -# Allocate memory from the OSSL for the L3 information
 * -# Allocate memory from the RM for the L2 information
 * -# Create the hash table for mapping from IPv4 address to L2 Index.
 * -# Create the hash table for mapping from IPv6 address to L2 Index.
 * -# Set all entries to the 'Idle' state
 * -# Return the newly created handle
 * @endif
 *
 * @return IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_ERROR_RANGE    One of the provided values was out of range.
 * @retval IX_CC_ERROR_NULL     A required pointer was NULL.
 * @retval IX_CC_ERROR_ALIGN    A pointer was not properly aligned.
 * @retval IX_CC_ERROR_OOM_SYSTEM Memory could not be allocated to create the
 *      L2TM handle.
 * @retval IX_CC_L2TM_ERROR_CREATE_L2TABLE The L2TM failed to create the L2
 *      Table.
 */
ix_error ix_cc_l2tm_init(
    const ix_cc_l2tm_config   *arg_pConfig,
    ix_cc_l2tm                *arg_phL2tm);

/**
 * Destroys an L2 Table Manager.  The unnamed L2 Table Manager is freed
 * immediately; because init and fini do not account for multiple clients,
 * clients should use ix_cc_l2tm_create() and ix_cc_l2tm_destroy() when
 * a single table is to be shared.
 *
 * Because microblocks use the tables managed by the L2TM, the tables should be
 * valid whenever the microblocks are executing.  Therefore, any client
 * responsible for microblocks should disable its microengines before
 * destroying the L2TM.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @if LLD @pseudocode
 * -# Destroy the hash table for IPv6 address mapping
 * -# Destroy the hash table for IPv4 address mapping
 * -# Free the memory from the RM for the L2 information
 * -# Free the memory from the OSSL for the L3 information
 * -# Free the memory from the OSSL for the handle's structure
 * @endif
 *
 * @return IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_ERROR_NULL     A required pointer was NULL.
 * @retval IX_CC_ERROR_ALIGN    A pointer was not properly aligned.
 * @retval IX_CC_L2TM_ERROR_DELETE_TABLE The L2TM was unable to free the
 *      resources for the L2 Table.
 */
ix_error ix_cc_l2tm_fini(
    ix_cc_l2tm arg_hL2tm);

/**
 * Adds an L2 Entry to the L2TM's table database. Typically, clients would do
 * this any time a new Next Hop is detected on a network. Next Hops are
 * uniquely identified by their layer 2 index.  The L2TM does not perform any
 * checks for duplicate entries elsewhere in the table, but does check for an
 * existing entry at the location in the table indicated by the L2Index.
 *
 * ix_cc_l2tm_add_entry() should be used when assigning both valid L2
 * information and an IP address. If only the IP address is known, use
 * ix_cc_l2tm_add_l3_info() instead.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @param       arg_L2Index (IN)
 *              Index into the L2 Table to be added.
 *
 * @param       arg_pL2Entry (IN)
 *              Pointer to the L2 Entry structure to be added.
 *
 * @if LLD @pseudocode
 * -# Verify the index is in the 'Idle' state
 * -# Check the input structure for validity
 * -# Add the IP address into the appropriate hash table
 * -# Add the L3 information into the L3 table
 * -# Set the IP address type for the entry
 * -# Set the header type for the entry
 * -# Set the L2 information in the shared structure
 * -# Set the state to 'Valid'
 * @endif
 *
 * @return IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_ERROR_NULL     A required pointer was NULL.
 * @retval IX_CC_ERROR_ALIGN    A pointer was not properly aligned.
 * @retval IX_CC_ERROR_RANGE    The L2 Index was out of range (beyond the
 *      capacity of the L2 Table).
 * @retval IX_CC_L2TM_ERROR_INVALID_ENTRY The entry structure contains invalid
 *      information.
 * @retval IX_CC_ERROR_DUPLICATE_ENTRY An entry with the same contents already
 *      exists at the specified location.
 * @retval IX_CC_ERROR_CONFLICTING_ENTRY An entry with different contents
 *      already exists at the specified location.
 */
ix_error ix_cc_l2tm_add_entry(
    ix_cc_l2tm arg_hL2tm,
    ix_uint16 arg_L2Index,
    ix_cc_l2tm_entry *arg_pL2Entry);

/**
 * Updates an L2 Entry to the L2TM's table database. Typically, clients would
 * do this when a next hop's information has changed.  The L2TM does not
 * perform any checks for duplicate entries elsewhere in the table, but does
 * verify the entry has valid data.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @param       arg_L2Index (IN)
 *              Index into the L2 Table to be updated.
 *
 * @param       arg_pL2Entry (IN)
 *              Pointer to the L2 Entry structure to be added.
 *
 * @if LLD @pseudocode
 * -# Verify the index is in the 'Valid' or 'Incomplete' state
 * -# Check the input structure for validity
 * -# Look up the L3 type for the existing entry
 * -# Retrieve the L3 information for the existing entry
 * -# Remove the IP address from the appropriate hash table
 * -# Set the L3 information into the L3 table
 * -# Set the IP address type for the entry
 * -# Add the IP address into the appropriate hash table
 * -# Set the L2 information in the shared structure
 * -# Set the state to 'Valid'
 * @endif
 *
 * @return IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_ERROR_NULL     A required pointer was NULL.
 * @retval IX_CC_ERROR_ALIGN    A pointer was not properly aligned.
 * @retval IX_CC_ERROR_RANGE    The L2 Index was out of range (beyond the
 *      capacity of the L2 Table).
 * @retval IX_CC_ERROR_ENTRY_NOT_FOUND The L2 Index does not contain valid
 *      data.
 * @retval IX_CC_L2TM_ERROR_INVALID_ENTRY The entry structure contains invalid
 *      information.
 */
ix_error ix_cc_l2tm_update_entry(
    ix_cc_l2tm arg_hL2tm,
    ix_uint16 arg_L2Index,
    ix_cc_l2tm_entry *arg_pL2Entry);

/**
 * Removes an L2 Entry from the L2TM. A client would typically delete an L2
 * Entry when a next hop router is removed from the network.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @param       arg_L2Index (IN)
 *              Index into the L2 Table to be deleted.
 *
 * @if LLD @pseudocode
 * -# Verify the index is in the 'Valid' or 'Incomplete' state
 * -# Look up the L3 type for the existing entry
 * -# Retrieve the L3 information for the existing entry
 * -# Remove the IP address from the appropriate hash table
 * -# Set the state to 'Idle'
 * @endif
 *
 * @return IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_ERROR_NULL     A required pointer was NULL.
 * @retval IX_CC_ERROR_ALIGN    A pointer was not properly aligned.
 * @retval IX_CC_ERROR_RANGE    The L2 Index was out of range (beyond the
 *      capacity of the L2 Table).
 * @retval IX_CC_ERROR_ENTRY_NOT_FOUND The entry at the specified location
 *      does not exist.
 */
ix_error ix_cc_l2tm_delete_entry(
    ix_cc_l2tm arg_hL2tm,
    ix_uint16 arg_L2Index);

/**
 * Gets an L2 Entry from the L2TM's table database.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @param       arg_L2Index (IN)
 *              Index into the L2 Table to be retrieved.
 *
 * @param       arg_pState (OUT)
 *              Current state of the entry. This parameter is optional; if the
 *              pointer is NULL, the state is not returned.
 *
 * @param       arg_pL2Entry (OUT)
 *              Pointer to the L2 Entry structure to be retrieved. This
 *              parameter is optional; if the pointer is NULL, the contents of
 *              the structure are not returned.
 *
 * @if LLD @pseudocode
 * -# Verify the index is valid.
 * -# Get the state of the entry.
 * -# If the caller wants the contents of the entry
 *      -# If the state is 'Incomplete' or 'Valid', store the L3 information.
 *      -# If the state is 'Valid', store the L2 Information.
 * @endif
 *
 * @return IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_ERROR_NULL     A required pointer was NULL.
 * @retval IX_CC_ERROR_ALIGN    A pointer was not properly aligned.
 * @retval IX_CC_ERROR_RANGE    The L2 Index was out of range (beyond the
 *      capacity of the L2 Table).
 */
ix_error ix_cc_l2tm_get_entry(
    ix_cc_l2tm arg_hL2tm,
    ix_uint16 arg_L2Index,
    ix_cc_l2tm_state *arg_pState,
    ix_cc_l2tm_entry *arg_pL2Entry);

/**
 * Adds L3 information into the L2 Table.  The entry is marked as in use, but
 * not valid, as the L2 information is not yet valid.  It is expected that the
 * client will call ix_cc_l2tm_assign_l2_info() to complete the entry.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @param       arg_L2Index (IN)
 *              Index into the L2 Table to be reserved.
 *
 * @param       arg_pIpAddress (IN)
 *              L3 Information to add to the entry.
 *
 * @param       arg_IpAddressType (IN)
 *              Type of the L3 address to be added.
 *
 * @if LLD @pseudocode
 * -# Verify the index is in the 'Idle' state
 * -# Add the IP address into the appropriate hash table
 * -# Set the L3 type for the entry
 * -# Set the L3 information for the entry
 * -# Set the state to 'Incomplete'
 * @endif
 *
 * @return IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_ERROR_NULL     A required pointer was NULL.
 * @retval IX_CC_ERROR_ALIGN    A pointer was not properly aligned.
 * @retval IX_CC_ERROR_RANGE    The L2 Index was out of range (beyond the
 *      capacity of the L2 Table).
 * @retval IX_CC_ERROR_ENTRY_NOT_FOUND The entry at the specified location
 *      does not exist.
 */
ix_error ix_cc_l2tm_add_l3_info(
    ix_cc_l2tm                  arg_hL2tm,
    ix_uint16                   arg_L2Index,
    const ix_cc_l2tm_ipaddr    *arg_pIpAddress,
    ix_cc_l2tm_ipaddr_type      arg_IpAddressType);

/**
 * Updates L3 information in the L2 Table.  It is expected that the client will
 * call ix_cc_l2tm_assign_l2_info() to complete the entry.
 *
 * A caller might use this entry point to modify the IP address of an existing
 * entry in the table.  Note that doing so causes the L2 information to be
 * invalidated; the caller must re-assign new L2 information.  (If the caller
 * wants to change the L2 information simultaneously with updating the IP
 * address, use ix_cc_l2tm_update_entry() instead).
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @param       arg_L2Index (IN)
 *              Index into the L2 Table to be reserved.
 *
 * @param       arg_pIpAddress (IN)
 *              L3 Information to add to the entry.
 *
 * @param       arg_IpAddressType (IN)
 *              Type of the L3 address to be added.
 *
 * @if LLD @pseudocode
 * -# Verify the index is in the 'Valid' or 'Incomplete' state
 * -# Look up the L3 type for the existing entry
 * -# Retrieve the L3 information for the existing entry
 * -# Remove the IP address from the appropriate hash table
 * -# Add the IP address into the appropriate hash table
 * -# Set the L3 type for the entry
 * -# Set the L3 information for the entry
 * -# Remove the L2 information
 * -# Set the state to 'Incomplete'
 * -# If the original state was 'Valid', adjust the counters.
 * @endif
 *
 * @return IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_ERROR_NULL     A required pointer was NULL.
 * @retval IX_CC_ERROR_ALIGN    A pointer was not properly aligned.
 * @retval IX_CC_ERROR_RANGE    The L2 Index was out of range (beyond the
 *      capacity of the L2 Table).
 * @retval IX_CC_ERROR_ENTRY_NOT_FOUND The entry at the specified location
 *      does not exist.
 */
ix_error ix_cc_l2tm_update_l3_info(
    ix_cc_l2tm                  arg_hL2tm,
    ix_uint16                   arg_L2Index,
    const ix_cc_l2tm_ipaddr    *arg_pIpAddress,
    ix_cc_l2tm_ipaddr_type      arg_IpAddressType);

/**
 * Updates an L2 Entry with new information.  Typically this would be used to
 * complete an entry after ARP has determined the L2 information.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @param       arg_pL2Entry (IN)
 *              L2 Table Entry structure - uses the IP address from the data
 *              structure as an index to populate the L2 Table.
 *
 * @param       arg_pL2Index (OUT)
 *              Index in the L2 Table where the structure was added. This
 *              argument is optional; if the pointer is NULL, the L2 Index is
 *              not returned to the caller.
 *
 * @if LLD @pseudocode
 * -# Verify the index is in the 'Incomplete' state
 * -# Check the input structure for validity
 * -# Get the l2Index for the specified IP address
 * -# Set the header type for the entry
 * -# Set the L2 information in the shared structure
 * -# Set the state to 'Valid'
 * @endif
 *
 * @return IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_ERROR_NULL     A required pointer was NULL.
 * @retval IX_CC_ERROR_ALIGN    A pointer was not properly aligned.
 * @retval IX_CC_ERROR_RANGE    The L2 Index was out of range (beyond the
 *      capacity of the L2 Table).
 * @retval IX_CC_L2TM_ERROR_INVALID_ENTRY The entry structure contains invalid
 *      information.
 */
ix_error ix_cc_l2tm_assign_l2_info(
    ix_cc_l2tm arg_hL2tm,
    ix_cc_l2tm_entry *arg_pL2Entry,
    ix_uint16 *arg_pL2Index);

/**
 * Clears the L2 Information from an entry from the L2TM.  A client would
 * typically clear the L2 Information when an ARP timeout has occurred and
 * no further information is available for the next hop.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @param       arg_L2Index (IN)
 *              Index into the L2 Table to be updated.
 *
 * @if LLD @pseudocode
 * -# Verify the index is in the 'Valid' state
 * -# Remove the L2 information
 * -# Set the state to 'Incomplete'
 * @endif
 *
 * @return IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_ERROR_NULL     A required pointer was NULL.
 * @retval IX_CC_ERROR_ALIGN    A pointer was not properly aligned.
 * @retval IX_CC_ERROR_RANGE    The L2 Index was out of range (beyond the
 *      capacity of the L2 Table).
 * @retval IX_CC_ERROR_ENTRY_NOT_FOUND The entry at the specified location
 *      does not exist.
 */
ix_error ix_cc_l2tm_clear_l2_info(
    ix_cc_l2tm arg_hL2tm,
    ix_uint16 arg_L2Index);

/**
 * Gets the index of the entry containing a given IP address. This function
 * is a general-purpose means of converting an IP address to an L2Index, given
 * the IP address exists in the L2 Table.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @param       arg_IpAddressType (IN)
 *              Indicates the type of the L3 Address.
 *
 * @param       arg_pIpAddress (IN)
 *              Layer-3 address to be looked up.
 *
 * @param       arg_pL2Index (OUT)
 *              Index in the L2 Table where the structure was added. If this
 *              argument is NULL, the L2 Index is not returned to the caller.
 *
 * @if LLD @pseudocode
 * -# Look up the index from the appropriate hash table
 * @endif
 *
 * @return IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_ERROR_NULL     A required pointer was NULL.
 * @retval IX_CC_ERROR_ALIGN    A pointer was not properly aligned.
 * @retval IX_CC_ERROR_RANGE    Invalid L3 address type.
 * @retval IX_CC_ERROR_ENTRY_NOT_FOUND  The L3 address was not found in the
 *              L2 Table.
 */
ix_error ix_cc_l2tm_get_l2_index(
    ix_cc_l2tm                  arg_hL2tm,
    ix_cc_l2tm_ipaddr_type      arg_IpAddressType,
    const ix_cc_l2tm_ipaddr    *arg_pIpAddress,
    ix_uint16                  *arg_pL2Index);

/**
 * Gets symbols useful to the microblocks. Typically the symbols are patched
 * to the microblocks.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @param       arg_pSymbols (OUT)
 *              Structure with symbols that need to be patched.
 *
 * @if LLD @pseudocode
 * -# Fill in the symbols using available information
 * @endif
 *
 * @return IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_ERROR_NULL     A required pointer was NULL.
 * @retval IX_CC_ERROR_ALIGN    A pointer was not properly aligned.
 */
ix_error ix_cc_l2tm_get_symbols(
    ix_cc_l2tm arg_hL2tm,
    ix_cc_l2tm_symbols *arg_pSymbols);

/**
 * Gets statistics about the L2 Table.  The statistics describe the current
 * state of the L2TM and its associated table.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @param       arg_pStatistics (OUT)
 *              Structure containing statistics about the L2 Table.
 *
 * @if LLD @pseudocode
 * -# Fill in the statistics using available information
 * @endif
 *
 * @return IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_ERROR_NULL     A required pointer was NULL.
 * @retval IX_CC_ERROR_ALIGN    A pointer was not properly aligned.
 */
ix_error ix_cc_l2tm_get_statistics(
    ix_cc_l2tm          arg_hL2tm,
    ix_cc_l2tm_stats   *arg_pStatistics);

/**
 * Remove all entries from the L2 Table.  This is primarily intended to be used
 * for development and debugging.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @if LLD @pseudocode
 * -# Clear the contents of the IPv4 hash table
 * -# Clear the contents of the IPv6 hash table
 * -# Mark all entries as 'Idle'
 * @endif
 *
 * @return IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_ERROR_NULL     A required pointer was NULL.
 * @retval IX_CC_ERROR_ALIGN    A pointer was not properly aligned.
 */
ix_error ix_cc_l2tm_purge(
    ix_cc_l2tm          arg_hL2tm);

/**
 * Prints all entries in the L2 table to the screen for debugging purposes.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @if LLD @pseudocode
 * -# For each entry in the table:
 *      -# Get the state of the entry.
 *      -# If the entry is INCOMPLETE or VALID, print the IP Type and IP
 *         Address.
 *      -# If the entry is VALID, print the L2 Type and L2 Address.
 * @endif
 *
 * @return IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_ERROR_NULL     A required pointer was NULL.
 * @retval IX_CC_ERROR_ALIGN    A pointer was not properly aligned.
 */
ix_error ix_cc_l2tm_dump(
    ix_cc_l2tm          arg_hL2tm);

/**
 * @} End of External API
 */

#if defined(__cplusplus)
} /* extern "C" */
#endif /* __cplusplus */

/*vim:tabstop=8:expandtab:shiftwidth=4:tw=78*/
#endif /*__IX_CC_L2TM_H__*/
