/*
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *     Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * $Id: internal_l2tm.h,v 1.9 2003/08/19 23:00:05 ktseng Exp $
 */

/**
 * @file internal_l2tm.h
 * Private header file for the L2TM.
 */

#if !defined(__INTERNAL_L2TM_H__)
#define __INTERNAL_L2TM_H__

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

#include "ix_rm.h"
#include "ix_netmacros.h"
#include "ix_cc_macros.h"
#include "cc/ix_cc_l2tm.h"
#include "shared.h"
#include "cc/internal/internal_sig.h"
#include "cc/ix_cc_hash_table.h"
#include "cc/ix_cc_hash_table128.h"

/*
 * Generate an error to note which sections have yet to be implemented.
 */
#define TODO IX_ERROR_PANIC(IX_CC_ERROR_INTERNAL, ("Not implemented"))

/**
 * @defgroup L2TMInternals L2 Table Manager Internals
 * Contains data types, function prototypes, macros, and structures not
 * exported to the clients of the L2TM.
 *
 * @{
 */

/**
 * Calculate the amount of memory required for the shared table. This is used
 * during initialization when allocating memory from the Resource Manager.
 *
 * @param       arg_NumEntries
 *              Number of entries to be managed by the L2TM.
 *
 * @returns Number of bytes of shared memory required.
 */
#define _IX_CC_L2TM_GET_TABLE_MEM_SIZE(arg_NumEntries)                     \
    ((arg_NumEntries) * sizeof(_ix_cc_l2tm_shared_entry))

/**
 * Calculate the number of buckets needed to store IPv4 addresses. This is used
 * during initialization when creating a 32-bit hash table.
 *
 * @param       arg_NumEntries
 *              Number of entries to be managed by the L2TM.
 *
 * @returns Integer number of buckets to use when creating the 32-bit hash
 *      table.
 */
#define _IX_CC_L2TM_NUM_IPV4_BUCKETS(arg_NumEntries)                    \
        IX_MAX((ix_uint32)((arg_NumEntries) / 4), 1)

/**
 * Calculate the number of buckets needed to store IPv6 addresses. This is used
 * during initialization when creating a 128-bit hash table.
 *
 * @param       arg_NumEntries
 *              Number of entries to be managed by the L2TM.
 *
 * @returns Integer number of buckets to use when creating the 128-bit hash
 *      table.
 */
#define _IX_CC_L2TM_NUM_IPV6_BUCKETS(arg_NumEntries)                    \
        IX_MAX(((ix_uint32)((arg_NumEntries) / 4)), 1)

/**
 * Verify an L2TM handle is valid.  Checks for such things as NULL, alignment,
 * and the proper signature.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 */
#define _IX_CC_L2TM_CHECK_HANDLE(arg_hL2tm)                             \
    IX_ERROR_DCHECK_ARG_NULL(1, arg_hL2tm);                             \
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hL2tm, struct _ix_s_cc_l2tm);      \
    _IX_CC_SIG_CHECK((arg_hL2tm)->sig, _ix_cc_l2tm_sig);

/**
 * The number of bits in each element of the array.  This is used with the
 * 'inUse' array to track which entries are in use.
 */
#define _IX_CC_L2TM_BITS_PER_ARRAY_ELEMENT (sizeof(ix_uint32) * 8)

/**
 * Returns the offset into the array of the indicated index. This is used with
 * the 'inUse' array to track which entries are in use.
 *
 * @param       arg_L2Index (IN)
 *              Index of the entry to query.
 *
 * @returns Index into the 'inUse' array containing the flag for the specified
 *      entry.
 */
#define _IX_CC_L2TM_GET_ELEMENT_INDEX(arg_L2Index)                      \
    IX_GET_BIT_FIELD16(arg_L2Index, 5, 15)

/**
 * Returns the bit position of the indicated index. This is used with
 * the 'inUse' array to track which entries are in use.
 *
 * @param       arg_L2Index (IN)
 *              Index of the entry to query.
 *
 * @returns Bit position of the 'inUse' flag for the given entry.
 */
#define _IX_CC_L2TM_GET_BIT_OF_ELEMENT(arg_L2Index)                     \
    IX_GET_BIT_FIELD16(arg_L2Index, 0, 4)

/**
 * Calculate the amount of heap memory required to manage the L2 Table. This
 * is used during initialization to allocate memory from the OSSL, as well
 * as when reporting statistics.
 *
 * @param       arg_NumEntries (IN)
 *              The capacity of the L2 Table.
 *
 * @returns Number of bytes required on the heap.
 */
#define _IX_CC_L2TM_CALC_HEAP_SIZE(arg_NumEntries)                      \
    (sizeof(struct _ix_s_cc_l2tm) + /* Main struct */                   \
    sizeof(ix_cc_l2tm_ipaddr) * (arg_NumEntries) + /* pL3Table */      \
    sizeof(ix_uint32) * (                                               \
        ((arg_NumEntries) + _IX_CC_L2TM_BITS_PER_ARRAY_ELEMENT - 1) /   \
        _IX_CC_L2TM_BITS_PER_ARRAY_ELEMENT) + /* pInUse */              \
    sizeof(ix_cc_l2tm_l2addr_type) * (arg_NumEntries) +/* pL3AddrTypes */\
    sizeof(ix_cc_l2tm_ipaddr_type) * (arg_NumEntries)) /* pIpAddrTypes */

/**
 * Copy a locally constructed header structure to shared memory. Before
 * writing, the entry is disabled by setting its 'valid' flag to zero. The 32
 * bits containing the 'valid' bit are written last to ensure the bit is only
 * set when the update is complete.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @param       arg_L2Index (IN)
 *              Index into the L2 Table to be queried.
 *
 * @param       arg_LocalHeader (IN)
 *              Header to be copied into shared memory.
 */
#define _IX_CC_L2TM_WRITE_HEADER(arg_hL2tm, arg_L2Index, arg_LocalHeader) \
    IX_CC_WRAP(                                                         \
        /* Set first byte to zero; clears 'valid'. */                   \
        (void) IX_RM_MEM_UINT8_WRITE(                                   \
            (ix_uint8*) &(arg_hL2tm)->pL2Table[arg_L2Index], 0);        \
        /* Write last 64 bits */                                        \
        (void) IX_RM_MEM_UINT64_WRITE(                                  \
            ((ix_uint64*) &(arg_hL2tm)->pL2Table[arg_L2Index]) +1,      \
            *((ix_uint64*) &(arg_LocalHeader)+1));                      \
        /* Write 2nd 32 bits */                                         \
        (void) IX_RM_MEM_UINT32_WRITE(                                  \
            ((ix_uint32*) &(arg_hL2tm)->pL2Table[arg_L2Index]) +1,      \
            *((ix_uint32*) &(arg_LocalHeader)+1));                      \
        /* Write 1st 32 bits, with proper 'valid' flag */               \
        (void) IX_RM_MEM_UINT32_WRITE(                                  \
            ((ix_uint32*) &(arg_hL2tm)->pL2Table[arg_L2Index]),         \
            *(ix_uint32*) &(arg_LocalHeader)))

/**
 * Copy a header from shared memory into a local structure.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @param       arg_L2Index (IN)
 *              Index into the L2 Table to be queried.
 *
 * @param       arg_LocalHeader (OUT)
 *              Header to be copied into shared memory.
 */
#define _IX_CC_L2TM_READ_HEADER(arg_hL2tm, arg_L2Index, arg_LocalHeader) \
    IX_CC_WRAP(                                                         \
        *(ix_uint64*) &(arg_LocalHeader) = IX_RM_MEM_UINT64_READ(       \
            (ix_uint64*) &(arg_hL2tm)->pL2Table[arg_L2Index]);          \
        *(((ix_uint64*) &(arg_LocalHeader))+1) = IX_RM_MEM_UINT64_READ(   \
            ((ix_uint64*) &(arg_hL2tm)->pL2Table[arg_L2Index]) +1))

/**
 * Pack a caller-usable Ethernet structure into a local Ethernet structure.
 *
 * @param       arg_EthEntry (IN)
 *              Structure to pack
 *
 * @param       arg_LocalEth (OUT)
 *              Local structure where the structure is to be packed
 */
#define _IX_CC_L2TM_PACK_ETH_HEADER(arg_EthEntry, arg_LocalEth)         \
    do {                                                                \
        int __i;                                                        \
        for (__i = 0; __i < 6; __i++)                                   \
        {                                                               \
            (arg_LocalEth).destMAC[__i] = (arg_EthEntry).destMAC[__i];  \
            (arg_LocalEth).srcMAC[__i] = (arg_EthEntry).sourceMAC[__i]; \
        }                                                               \
        (arg_LocalEth).protocol[0] = (arg_EthEntry).protocol >> 8;      \
        (arg_LocalEth).protocol[1] = (arg_EthEntry).protocol & 0xFF;    \
        (arg_LocalEth).res0 = 0;                                        \
        (arg_LocalEth).res1 = 0;                                        \
    } while(0)

/**
 * Unpack a local Ethernet structure into a caller-usable Ethernet structure.
 *
 * @param       arg_LocalEth (IN)
 *              Local copy of the entry to be unpacked
 *
 * @param       arg_EthEntry (OUT)
 *              Structure the entry is to be unpacked into
 */
#define _IX_CC_L2TM_UNPACK_ETH_HEADER(arg_LocalEth, arg_EthEntry)       \
    do {                                                                \
        int __i;                                                        \
        for (__i = 0; __i < 6; __i++)                                   \
        {                                                               \
            (arg_EthEntry).destMAC[__i] = (arg_LocalEth).destMAC[__i];  \
            (arg_EthEntry).sourceMAC[__i] = (arg_LocalEth).srcMAC[__i]; \
        }                                                               \
        (arg_EthEntry).protocol =                                       \
            ((arg_LocalEth).protocol[0] << 8) | (arg_LocalEth).protocol[1];\
    } while(0)

/**
 * Pack a caller-usable ATM structure into a local ATM structure
 *
 * @param       arg_AtmEntry (IN)
 *              Structure to pack
 *
 * @param       arg_LocalAtm (OUT)
 *              Local structure where the structure is to be packed
 */
#define _IX_CC_L2TM_PACK_ATM_HEADER(arg_AtmEntry, arg_LocalAtm)         \
    IX_CC_WRAP(                                                         \
        (arg_LocalAtm).vpiVci   = IX_HTON32((arg_AtmEntry).vpiVci);     \
        (arg_LocalAtm).vcq      = IX_HTON32((arg_AtmEntry).vcq);        \
        (arg_LocalAtm).llcSnap0 = IX_HTON32(                            \
            (arg_AtmEntry).llcSnapHeader >> 32);   \
        (arg_LocalAtm).llcSnap1 = IX_HTON32((arg_AtmEntry).llcSnapHeader))

/**
 * Unpack a local ATM structure into a caller-usable ATM structure.
 *
 * @param       arg_LocalAtm (IN)
 *              Local copy of the entry to be unpacked
 *
 * @param       arg_AtmEntry (OUT)
 *              Structure the entry is to be unpacked into
 */
#define _IX_CC_L2TM_UNPACK_ATM_HEADER(arg_LocalAtm, arg_AtmEntry)       \
    IX_CC_WRAP(                                                         \
        (arg_AtmEntry).vpiVci = IX_NTOH32((arg_LocalAtm).vpiVci) &      \
            0x0FFFFFFF;                                                 \
        (arg_AtmEntry).vcq = IX_NTOH32((arg_LocalAtm).vcq);             \
        (arg_AtmEntry).llcSnapHeader = IX_NTOH32((arg_LocalAtm).llcSnap0);\
        (arg_AtmEntry).llcSnapHeader <<= 32;                            \
        (arg_AtmEntry).llcSnapHeader |= IX_NTOH32((arg_LocalAtm).llcSnap1))

/**
 * Disable an entry by clearing its 'Valid' flag.  Note that the current
 * implementation may destroy some data in the process.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @param       arg_L2Index (IN)
 *              Index into the L2 Table to be disabled.
 *
 */
#define _IX_CC_L2TM_DISABLE_ENTRY(arg_hL2tm, arg_L2Index) IX_CC_WRAP(   \
    (void) IX_RM_MEM_UINT32_WRITE(                                      \
        &(arg_hL2tm)->pL2Table[arg_L2Index].flags, 0))

/**
 * Gets the flags field from shared memory.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @param       arg_L2Index (IN)
 *              Index into the L2 Table to be queried.
 *
 * @returns 32-bit flags field, with the four flags in the high four bits.
 */
#define _IX_CC_L2TM_GET_FLAGS(arg_hL2tm, arg_L2Index)                   \
    IX_NTOH32(                                                          \
        IX_RM_MEM_UINT32_READ(&(arg_hL2tm)->pL2Table[arg_L2Index].flags))

/**
 * Sets the in-use flag for the given entry to one.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @param       arg_L2Index (IN)
 *              Index into the L2 Table to be set.
 */
#define _IX_CC_L2TM_SET_IN_USE(arg_hL2tm, arg_L2Index) IX_CC_WRAP(      \
    (arg_hL2tm)->pInUse[_IX_CC_L2TM_GET_ELEMENT_INDEX(arg_L2Index)] |=  \
        (1 << _IX_CC_L2TM_GET_BIT_OF_ELEMENT(arg_L2Index)))

/**
 * Sets the in-use flag for the given entry to zero.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @param       arg_L2Index (IN)
 *              Index into the L2 Table to be set.
 */
#define _IX_CC_L2TM_SET_NOT_IN_USE(arg_hL2tm, arg_L2Index) IX_CC_WRAP(  \
    (arg_hL2tm)->pInUse[_IX_CC_L2TM_GET_ELEMENT_INDEX(arg_L2Index)] &=  \
        (0xFFFFFFFFUL ^ (1 << _IX_CC_L2TM_GET_BIT_OF_ELEMENT(arg_L2Index))))

/**
 * Gets the in-use flag for the indicated entry, indicating whether the value
 * is in use or not.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @param       arg_L2Index (IN)
 *              Index into the L2 Table to be queried.
 *
 * @returns 0 if the element is not in use, or 1 if it is.
 */
#define _IX_CC_L2TM_IS_IN_USE(arg_hL2tm, arg_L2Index)                   \
    (((arg_hL2tm)->pInUse[_IX_CC_L2TM_GET_ELEMENT_INDEX(arg_L2Index)] >> \
        _IX_CC_L2TM_GET_BIT_OF_ELEMENT(arg_L2Index)) & 1)

/**
 * Set an entry to Idle.  This involves clearing the 'Valid' flag in shared
 * memory, and clearing the 'in use' flag in local memory.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @param       arg_L2Index (IN)
 *              Index into the L2 Table to be set.
 */
#define _IX_CC_L2TM_SET_IDLE(arg_hL2tm, arg_L2Index) IX_CC_WRAP(        \
    _IX_CC_L2TM_DISABLE_ENTRY(arg_hL2tm, arg_L2Index);                  \
    _IX_CC_L2TM_SET_NOT_IN_USE(arg_hL2tm, arg_L2Index))

/**
 * The 'valid' bit is the first flag.  Because the flags field is 32bits (even
 * though only 4 bits are available to be used as flags), express the number
 * in 32 bits.
 * @see _ix_u_cc_l2tm_shared_entry
 */
#define _IX_CC_L2TM_FLAG_VALID 0x80000000UL

/**
 * Maximum size of a name that can be assigned to a Table Manager.
 */
#define _IX_CC_L2TM_MAX_NAME_SIZE 128

/**
 * Private data structure used in the L2TM. This structure is referenced from
 * the externally opaque handle (ix_cc_l2tm), and contains all the information
 * necessary to maintain the L2 and L3 information for a single L2 Table.
 */
struct _ix_s_cc_l2tm
{
    _IX_CC_SIG_LOCAL(sig);

    /**
     * Base address of the L2 Table in shared memory.  This is the virtual
     * address of shared memory, and is used directly to access memory used by
     * the microblocks.
     */
    _ix_cc_l2tm_shared_entry *pL2Table;

    /**
     * Base address of the L3 Table in local memory. This table, which is
     * parallel to the L2 Table, contains either IPv4 or IPv6 addresses.
     */
    ix_cc_l2tm_ipaddr *pL3Table;

    /**
     * Bit array indicating which entries in the L2 Table are in use. This
     * information, combined with the 'valid' bit in the shared structure,
     * determines the state of each entry: 'Idle', 'Incomplete', or 'Valid'.
     */
    ix_uint32 *pInUse;

    /**
     * Array indicating the type of L3 address stored for each element. This
     * information is determined by the caller during an add or update
     * operation, and is used to determine the meaning of the contents of the
     * associated entry in the pL3Table.
     */
    ix_cc_l2tm_ipaddr_type *pIpAddrTypes;

    /**
     * Array indicating the type of L2 information stored for each element.
     * This information is determined by the caller during an add or update
     * operation, and is used to determine the meaning of the contents of the
     * associated entry in the pL2Table.
     */
    ix_cc_l2tm_l2addr_type *pL2AddrTypes;

    /**
     * Hash table for mapping IPv4 addresses to L2 Indexes. The key for this
     * hash table is the IPv4 address, and the data is the L2Index.
     */
    ix_cc_hash_table hIpv4Hash;

    /**
     * Hash table for mapping IPv6 addresses to L2 Indexes. The key for this
     * hash table is the IPv6 address, and the data is the L2Index.
     */
    ix_cc_hash_table128 hIpv6Hash;

    /**
     * Number of entries the L2 Table Manager could contain. The client
     * specifies this value during ix_cc_l2tm_init(), or through the registry
     * in ix_cc_l2tm_create().
     */
    ix_uint16 capacity;

    /**
     * Number of Valid entries in the L2 Table. This number changes when
     * entries are added and deleted.
     */
    ix_uint16 numValid;

    /**
     * Number of Incomplete entries in the L2 Table. This number changes when
     * entries are added and deleted.
     */
    ix_uint16 numIncomplete;

    /**
     * Name of the L2 Table.  Matches the name given during
     * ix_cc_l2tm_create().
     */
    char name[_IX_CC_L2TM_MAX_NAME_SIZE];
};

/**
 * Get the state for the entry at the given index.  This function uses several
 * pieces of information when determining the state of a entry:
 * -# Is the index in range?
 * -# Is the 'in use' flag set in local memory?
 * -# Is the 'valid' flag set in shared memory?
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @param       arg_L2Index (IN)
 *              Index into the L2 Table to be queried.
 *
 * @param       arg_pState (OUT)
 *              The state of the entry.
 *
 * @if LLD @pseudocode
 * -# If the index is out of range, return an error
 * -# If the entry is not in use, the state is 'Idle'. Return.
 * -# Get the 'valid' flag from shared memory.
 * -# If the 'valid' flag is zero, the state is 'Incomplete'. Return.
 * -# The state is 'Valid'.
 * @endif
 *
 * @return IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_ERROR_RANGE    The L2 Index was out of range (beyond the
 *      capacity of the L2 Table).
 */
ix_error _ix_cc_l2tm_get_state(
    ix_cc_l2tm arg_hL2tm,
    ix_uint16 arg_L2Index,
    ix_cc_l2tm_state *arg_pState);

/**
 * Checks the input structure for correct values.  This function checks the
 * header and address types, verifies the contents of the structures appear to
 * match the types, looks for values that are out of range, etc.  It does not,
 * however, verify that MAC addresses are accurate, that IP addresses exist on
 * the network, etc.
 *
 * @param       arg_pL2Entry (IN)
 *              Pointer to the L2 Entry structure to be checked.
 *
 * @if LLD @pseudocode
 * -# If the address type is out of range, return an error.
 * -# If the header type is ATM:
 *      -# Verify bits 28-31 of vpiVci are zero
 *      -# Verify bits 16-31 of vcq are zero
 * -# Else, if the header type is not Ethernet, return an error.
 * -# The entry is acceptable, return success.
 * @endif
 *
 * @return IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_L2TM_ERROR_INVALID_ENTRY The entry structure contains invalid
 *      information.
 */
ix_error _ix_cc_l2tm_verify_entry(
    ix_cc_l2tm_entry *arg_pL2Entry);

/**
 * Adds an L3 address into the appropriate hash table, and if successful,
 * records the data in the address type and l3 address tables.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @param       arg_L2Index (IN)
 *              Index into the L2 Table to be modified.
 *
 * @param       arg_L3AddressType (IN)
 *              Type of the L3 address to be added.
 *
 * @param       arg_pL3Address (IN)
 *              L3 address to be added.
 *
 * @if LLD @pseudocode
 * -# Add the L3 information into the L3 table
 * -# Set the IP address type for the entry
 * @endif
 *
 * @return IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_L2TM_ERROR_INVALID_ENTRY The entry structure contains invalid
 *      information.
 */
ix_error _ix_cc_l2tm_add_l3_address(
    ix_cc_l2tm                  arg_hL2tm, 
    ix_uint16                   arg_L2Index, 
    ix_cc_l2tm_ipaddr_type      arg_L3AddressType,
    const ix_cc_l2tm_ipaddr    *arg_pL3Address);

/**
 * Prints the L2 Information of a single entry to the screen.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *              
 * @param       arg_L2Index (IN)
 *              Index of the entry to be printed.
 *
 * @return IX_SUCCESS
 */
ix_error _ix_cc_l2tm_print_l2(
    ix_cc_l2tm arg_hL2tm,
    ix_uint16  arg_L2Index);

/**
 * Prints the L3 Information of a single entry to the screen.
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *              
 * @param       arg_L2Index (IN)
 *              Index of the entry to be printed.
 *
 * @return IX_SUCCESS
 */
ix_error _ix_cc_l2tm_print_l3(
    ix_cc_l2tm arg_hL2tm,
    ix_uint16  arg_L2Index);

/**
 * @} End of Internals
 */

#if defined(__cplusplus)
} /* extern "C" */
#endif /* __cplusplus */

/*vim:tabstop=8:expandtab:shiftwidth=4:tw=78*/
#endif /*__INTERNAL_L2TM_H__*/
