/*
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *     Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * $Id: shared.h,v 1.3 2003/01/24 23:58:48 lbmiller Exp $
 *
 * Note that the format and location of the comments in this file are
 * significant, as they are used to generate documentation about the module.
 * Take care when modifying this file.
 */

/**
 * @file shared.h
 * Private file used only within the L2 Table Manager.
 */

#ifndef __IX_CC_L2TM_SHARED_H__
#define __IX_CC_L2TM_SHARED_H__

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

/*
 * These are the typedefs for the various enumerations and data structures.
 * They are listed outside of the group to clean up the generated
 * documentation.
 */
typedef struct _ix_s_cc_l2tm_shared_atm _ix_cc_l2tm_shared_atm;
typedef struct _ix_s_cc_l2tm_shared_eth _ix_cc_l2tm_shared_eth;
typedef union _ix_u_cc_l2tm_shared_entry _ix_cc_l2tm_shared_entry;

/**
 * @defgroup L2TMShared Shared Structures
 *
 * L2 Table data structures shared with the microblocks.  Though the
 * microblocks define the format, they are not exported as C-style structures,
 * so the L2TM defines them here.
 *
 * @{
 */

/**
 * ATM entry in the shared table.  This structure is used by the L2TM when
 * storing and retrieving an ATM header in shared memory, and therefore
 * conforms to the specification defined by the ATM encap microblock.
 *
 * @code
 * LW   Bits   Size     Field
 *  0   31:28    4      Flags
 *  0   27: 0   28      Out VPI/VCI
 *  1   15: 0   16      VCQ#
 *  2   31: 0   32      LLC/SNAP Header
 *  3   31: 0   32      LLC/SNAP Header
 * @endcode
 *
 * @typename _ix_cc_l2tm_shared_atm
 */
struct _ix_s_cc_l2tm_shared_atm
{
    ix_uint32 vpiVci;           /**< 28-bit VPI/VCI */
    ix_uint32 vcq;              /**< 16-bit VCQ */
    ix_uint32 llcSnap0;         /**< first 32 bits of the LLC/SNAP */
    ix_uint32 llcSnap1;         /**< last 32 bits of the LLC/SNAP */
};

/**
 * Ethernet entry in the shared table. This structure is used by the L2TM when
 * storing and retrieving an Ethernet header in shared memory, and therefore
 * conforms to the specification defined by the Eth encap microblock.
 *
 * @code
 * LW   Bits   Size     Field
 *  0   31:28    4      Flags
 *  0   27:24    4      Reserved (set to zero)
 *  0   23: 0   24      Destination MAC
 *  1   31: 8   24      Destination MAC (3 bytes)
 *  2    7: 0    8      Source MAC
 *  3   31: 0   32      Source MAC
 *  4   31:24    8      Source MAC
 *  4   23: 8   16      Protocol
 *  4    7: 0    8      Reserved
 * @endcode
 *
 * @typename _ix_cc_l2tm_shared_eth
 */
struct _ix_s_cc_l2tm_shared_eth
{
    /**
     * The first field overlaps with the flags.  The low 4 bits are unused,
     * and must be set to zero.
     */
    ix_uint8 res0;

    /**
     * Destination MAC address.  Because the field spans a 4-byte boundary, it
     * is expressed as an array.
     */
    ix_uint8 destMAC[6];

    /**
     * Destination MAC address.  Because the field spans a 4-byte boundary, it
     * is expressed as an array.
     */
    ix_uint8 srcMAC[6];

    /**
     * L3 Protocol encapsulated in the packet.  Because the field spans a
     * 2-byte boundary, it is expressed as an array.
     */
    ix_uint8 protocol[2];

    /**
     * Reserved.  Must be set to zero.
     */
    ix_uint8 res1;
};

/**
 * Single entry in the L2 Table.  The first four bits contains flags.  The
 * remainder of the data is defined by the Layer 2 protocol the entry
 * represents.
 *
 * @typename _ix_cc_l2tm_shared_entry
 */
union _ix_u_cc_l2tm_shared_entry
{
    /**
     * Flags indicating to the microblocks how they should treat the
     * L2 Information.
     *
     * The flags (in binary) are:
     * - 1000:    Valid
     * - 0100:    Send to Core
     * - 0010:    Reserved
     * - 0001:    Reserved
     *
     * The flags occupy the first four bits of each entry. To simplify the
     * implementation, the flags are expressed as a 32-bit number, but only
     * the four most sigificant bits are modified using this field.
     */
    ix_uint32               flags;

    /**
     * ATM entry in the shared table.  This entry overlaps with the flags
     * field, but is mutually exclusive with eth.
     */
    _ix_cc_l2tm_shared_atm  atm;

    /**
     * Ethernet entry in the shared table.  This entry overlaps with the flags
     * field, but is mutually exclusive with atm.
     */
    _ix_cc_l2tm_shared_eth  eth;
};

/**
 * @} End of Shared
 */

#if defined(__cplusplus)
} /* extern "C" */
#endif /* __cplusplus */

/*vim:tabstop=8:expandtab:shiftwidth=4:tw=78*/
#endif /* __IX_CC_L2TM_SHARED_H__ */
