/*
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *     Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * $Id: l2tm.c,v 1.14 2003/07/15 23:01:44 ktseng Exp $
 */

/**
 * @file l2tm.c
 *
 * Implementation of the L2 Table Manager.
 */

#include "ix_cc_error.h"
#include "ix_ossl.h"
#include "ix_rm.h"
#include "cc/l2tm/internal/internal_l2tm.h"

/**
 * Unique signature for the L2TM.
 * {9D7A258F-CB7A-42de-B242-C31C4BDCDBD6}
 */
_IX_CC_SIG_GLOBAL(_ix_cc_l2tm_sig, 0x9d7a258f, 0xcb7a, 0x42de,
        0xb2, 0x42, 0xc3, 0x1c, 0x4b, 0xdc, 0xdb, 0xd6);

ix_error ix_cc_l2tm_init(
    const ix_cc_l2tm_config *arg_pConfig,
    ix_cc_l2tm *arg_phL2tm)
{
    ix_cc_l2tm hL2tm = NULL;
    ix_memory_type rmMemoryType = 0;
    ix_uint32 l2TableMemSize = 0;
    ix_error err = IX_SUCCESS;
    ix_uint32 memSize = 0;
    
    unsigned int tmpBaseAddr, tmpL3TableAddr, tmpInUseAddr, tmpIpAddrTypesAddr, tmpL2AddrTypesAddr;


    IX_ERROR_DCHECK_ARG_NULL(1, arg_pConfig);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_pConfig, ix_cc_l2tm_config);
    IX_ERROR_DCHECK_ARG_NULL(2, arg_phL2tm);
    IX_ERROR_DCHECK_ARG_ALIGN(2, arg_phL2tm, ix_cc_l2tm);

#if defined(IX_DEBUG)
    /*
     * Verify the client isn't requesting too many entries
     */
    if (arg_pConfig->numEntries > 0xFFFFUL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
            ("numEntries is too large: %ld", arg_pConfig->numEntries));
    }
#endif /* IX_DEBUG */

    memSize = _IX_CC_L2TM_CALC_HEAP_SIZE(arg_pConfig->numEntries);

    /*
     * Allocate memory from the OSSL for the handle's data structure
     * Allocate memory from the OSSL for the L3 information
     */
    hL2tm = ix_ossl_malloc(memSize);
    if (hL2tm == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
            ("Could not allocate %d bytes for the ix_cc_l2tm", memSize));
    }
    ix_ossl_memset(hL2tm, 0, memSize);

    _IX_CC_SIG_COPY(hL2tm->sig, _ix_cc_l2tm_sig);

    /*
     * Decode the memory type
     */
    switch (arg_pConfig->memoryType)
    {
        case IX_CC_L2TM_MEMORY_TYPE_SRAM:
            rmMemoryType = IX_MEMORY_TYPE_SRAM;
            break;

        case IX_CC_L2TM_MEMORY_TYPE_DRAM:
            rmMemoryType = IX_MEMORY_TYPE_DRAM;
            break;

        default:
            err = IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
                ("The Memory Type was not recognized: %ld",
                 arg_pConfig->memoryType));
            goto clean_mem;
    }

    l2TableMemSize = _IX_CC_L2TM_GET_TABLE_MEM_SIZE(arg_pConfig->numEntries);

    /*
     * Allocate memory from the RM for the L2 information
     */
    err = ix_rm_mem_alloc_aligned(rmMemoryType, arg_pConfig->memoryChannel,
        l2TableMemSize,
        IX_MEMORY_ALIGNMENT_TYPE_VIRTUAL,
        3,              /* shift by 3 bytes, gives 64-bit alignment */
        (void**) &hL2tm->pL2Table);
    if (err != IX_SUCCESS)
    {
        switch ((int) IX_ERROR_GET_CODE(err))
        {
            case IX_RM_ERROR_ASSERTION:
                err = IX_ERROR_CHAIN(err, IX_CC_ERROR_RANGE, 0,
                    ("Invalid channel: %d", arg_pConfig->memoryChannel));
                break;

            case IX_RM_ERROR_CANNOT_ALLOCATE_SIZE:
                err = IX_ERROR_CHAIN(err, IX_CC_ERROR_OOM, 0,
                    ("Out of %s memory\n"
                     "\tCould not allocate %d bytes from the RM",
                     ((rmMemoryType == IX_MEMORY_TYPE_SRAM) ?
                          "SRAM" : "DRAM"),
                     l2TableMemSize));
                break;

            default:
                err = IX_ERROR_CHAIN(err, IX_CC_ERROR_INTERNAL, 0,
                    ("Could not allocate %d bytes from the RM",
                     l2TableMemSize));
                break;
        }
        goto clean_mem;
    }
    ix_ossl_memset(hL2tm->pL2Table, 0, l2TableMemSize);

    /*
     * Create the hash table for mapping from IPv4 address to L2 Index.
     */
    IX_ERROR_CG(ix_cc_hash_table_init(
                _IX_CC_L2TM_NUM_IPV4_BUCKETS(arg_pConfig->numEntries),
                arg_pConfig->numEntries,
                IX_CC_HASH_TABLE_ALGO_MAGIC,
                &hL2tm->hIpv4Hash),
            err, clean_shared);

    /*
     * Create a hash table for IPv6 address mapping
     */
    IX_ERROR_CG(ix_cc_hash_table128_init(
                _IX_CC_L2TM_NUM_IPV6_BUCKETS(arg_pConfig->numEntries),
                arg_pConfig->numEntries,
                IX_CC_HASH_TABLE_ALGO_MAGIC,
                &hL2tm->hIpv6Hash),
            err, clean_ipv4Hash);

    /*
     * Set all entries to the 'Idle' state
     * NOTE: This has been accomplished by the two 'memset' operations above.
     */

    /*
     * The L3 Table starts immediately after the L2 Table Manager's handle.
     */

    tmpBaseAddr = (unsigned int) hL2tm;
    tmpL3TableAddr = tmpBaseAddr + sizeof(struct _ix_s_cc_l2tm);
    tmpInUseAddr   = tmpL3TableAddr + (sizeof(ix_cc_l2tm_ipaddr) * (arg_pConfig->numEntries));
    tmpIpAddrTypesAddr = tmpInUseAddr + (sizeof(ix_uint32) * 
			(((arg_pConfig->numEntries) + _IX_CC_L2TM_BITS_PER_ARRAY_ELEMENT - 1) / _IX_CC_L2TM_BITS_PER_ARRAY_ELEMENT));

    tmpL2AddrTypesAddr = tmpIpAddrTypesAddr + (sizeof(ix_cc_l2tm_ipaddr_type) * (arg_pConfig->numEntries));

    hL2tm->pL3Table = (ix_cc_l2tm_ipaddr*) tmpL3TableAddr;
    hL2tm->pInUse = (ix_uint32*) tmpInUseAddr;
    hL2tm->pIpAddrTypes = (ix_cc_l2tm_ipaddr_type*) tmpIpAddrTypesAddr;
    hL2tm->pL2AddrTypes = (ix_cc_l2tm_l2addr_type*) tmpL2AddrTypesAddr;

    hL2tm->capacity = arg_pConfig->numEntries;

    /*
     * Return the newly created handle
     */
    *arg_phL2tm = hL2tm;
    return IX_SUCCESS;

clean_ipv4Hash:
    (void) ix_cc_hash_table_fini(hL2tm->hIpv4Hash);
clean_shared:
    (void) ix_rm_mem_free(hL2tm->pL2Table);
clean_mem:
    (void) ix_ossl_memset(hL2tm, 0, sizeof(struct _ix_s_cc_l2tm));
    (void) ix_ossl_free(hL2tm);
    return err;
}

ix_error ix_cc_l2tm_fini(
    ix_cc_l2tm arg_hL2tm)
{
    _IX_CC_L2TM_CHECK_HANDLE(arg_hL2tm);

    /*
     * Destroy the hash table for IPv6 address mapping
     */
    IX_ERROR_CR(ix_cc_hash_table128_fini(arg_hL2tm->hIpv6Hash));

    /*
     * Destroy the hash table for IPv4 address mapping
     */
    IX_ERROR_CR(ix_cc_hash_table_fini(arg_hL2tm->hIpv4Hash));

    /*
     * Free the memory from the RM for the L2 information
     */
    IX_ERROR_CR(ix_rm_mem_free(arg_hL2tm->pL2Table));

#if defined(IX_DEBUG)
    /*
     * In Debug mode, clear out the structure.  This resets the signature so
     * we can detect if the handle is used again by accident.
     */
    ix_ossl_memset(arg_hL2tm, 0, sizeof(struct _ix_s_cc_l2tm));
#endif

    /*
     * Free the memory from the OSSL for the L3 information
     * Free the memory from the OSSL for the handle's structure
     */
    ix_ossl_free(arg_hL2tm);

    return IX_SUCCESS;
}

ix_error ix_cc_l2tm_add_entry(
    ix_cc_l2tm arg_hL2tm,
    ix_uint16 arg_L2Index,
    ix_cc_l2tm_entry *arg_pL2Entry)
{
    _ix_cc_l2tm_shared_entry localHeader;

    _IX_CC_L2TM_CHECK_HANDLE(arg_hL2tm);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pL2Entry);
    IX_ERROR_DCHECK_ARG_ALIGN(3, arg_pL2Entry, ix_cc_l2tm_entry);

#if defined(IX_DEBUG)
    {
        ix_cc_l2tm_state state;
        /*
         * Verify the index is in the 'Idle' state
         */
        IX_ERROR_CR(_ix_cc_l2tm_get_state(arg_hL2tm, arg_L2Index, &state));
        if (state != IX_CC_L2TM_STATE_IDLE)
        {
            /* Report an error and return success */
            ix_ossl_message_log("An entry already exists at index 0x%04X\n", arg_L2Index);
            return(IX_SUCCESS);
        }

        /*
         * Check the input structure for validity
         */
        IX_ERROR_CR(_ix_cc_l2tm_verify_entry(arg_pL2Entry));
    }
#endif /* IX_DEBUG */

    /*
     * Add the IP address into the appropriate hash table
     * Add the L3 information into the L3 table
     * Set the IP address type for the entry
     */
    IX_ERROR_CR(_ix_cc_l2tm_add_l3_address(arg_hL2tm, arg_L2Index,
            arg_pL2Entry->ipAddrType, &arg_pL2Entry->l3Address));

    /*
     * Set the header type for the entry
     */
    arg_hL2tm->pL2AddrTypes[arg_L2Index] = arg_pL2Entry->l2AddrType;

    /*
     * Set the L2 information in the shared structure
     */
    switch (arg_pL2Entry->l2AddrType)
    {
        case IX_CC_L2TM_L2ADDR_ETHER:
            _IX_CC_L2TM_PACK_ETH_HEADER(arg_pL2Entry->header.ether,
                localHeader.eth);
            break;

        case IX_CC_L2TM_L2ADDR_ATM:
            _IX_CC_L2TM_PACK_ATM_HEADER(arg_pL2Entry->header.atm,
                localHeader.atm);
            break;

        /*
         * No default action here because the input structure was already
         * checked.
         */
        default:
            break;
    }

    /*
     * Set the state to 'Valid'
     */
    localHeader.flags |= IX_HTON32(_IX_CC_L2TM_FLAG_VALID);
    _IX_CC_L2TM_SET_IN_USE(arg_hL2tm, arg_L2Index);

    /*
     * The header was constructed in local memory; send it out to shared
     * memory.
     */
    _IX_CC_L2TM_WRITE_HEADER(arg_hL2tm, arg_L2Index, localHeader);

    arg_hL2tm->numValid++;

    return IX_SUCCESS;
}

ix_error ix_cc_l2tm_update_entry(
    ix_cc_l2tm arg_hL2tm,
    ix_uint16 arg_L2Index,
    ix_cc_l2tm_entry *arg_pL2Entry)
{
    ix_cc_l2tm_state state;

    void *hashData = (void*) (ix_uint32) arg_L2Index;
    _ix_cc_l2tm_shared_entry localHeader;
    ix_error err = IX_SUCCESS;

    _IX_CC_L2TM_CHECK_HANDLE(arg_hL2tm);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pL2Entry);
    IX_ERROR_DCHECK_ARG_ALIGN(3, arg_pL2Entry, ix_cc_l2tm_entry);

    /*
     * Verify the index is in the 'Valid' or 'Incomplete' state
     */
    IX_ERROR_CR(_ix_cc_l2tm_get_state(arg_hL2tm, arg_L2Index, &state));

#if defined(IX_DEBUG)
    if (state == IX_CC_L2TM_STATE_IDLE)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
            ("Entry not found at index 0x%04X", arg_L2Index));
    }

    /*
     * Check the input structure for validity
     */
    IX_ERROR_CR(_ix_cc_l2tm_verify_entry(arg_pL2Entry));
#endif /* IX_DEBUG */

    /*
     * Look up the L3 type for the existing entry
     * Retrieve the L3 information for the existing entry
     * Remove the IP address from the appropriate hash table
     */
    switch (arg_hL2tm->pIpAddrTypes[arg_L2Index])
    {
        case IX_CC_L2TM_IPADDR_IPV4:
            IX_ERROR_CR(ix_cc_hash_table_delete_entry(arg_hL2tm->hIpv4Hash,
                        arg_hL2tm->pL3Table[arg_L2Index].ipv4Address));
            break;

        case IX_CC_L2TM_IPADDR_IPV6:
            IX_ERROR_CR(ix_cc_hash_table128_delete_entry(arg_hL2tm->hIpv6Hash,
                        arg_hL2tm->pL3Table[arg_L2Index].ipv6Address));
            break;

        default:
            break;
    }

    /*
     * Add the IP address into the appropriate hash table
     * Add the L3 information into the L3 table
     * Set the IP address type for the entry
     */
    IX_ERROR_CG(_ix_cc_l2tm_add_l3_address(arg_hL2tm, arg_L2Index,
            arg_pL2Entry->ipAddrType, &arg_pL2Entry->l3Address), err, cleanup);

    /*
     * Set the L2 information in the shared structure
     */
    switch (arg_pL2Entry->l2AddrType)
    {
        case IX_CC_L2TM_L2ADDR_ETHER:
            _IX_CC_L2TM_PACK_ETH_HEADER(arg_pL2Entry->header.ether,
                localHeader.eth);
            break;

        case IX_CC_L2TM_L2ADDR_ATM:
            _IX_CC_L2TM_PACK_ATM_HEADER(arg_pL2Entry->header.atm,
                localHeader.atm);
            break;

        /*
         * No default action here because the input structure was already
         * checked.
         */
        default:
            break;
    }

    /*
     * Set the header type for the entry
     */
    arg_hL2tm->pL2AddrTypes[arg_L2Index] = arg_pL2Entry->l2AddrType;

    /*
     * Set the state to 'Valid'
     */
    localHeader.flags |= IX_HTON32(_IX_CC_L2TM_FLAG_VALID);
    _IX_CC_L2TM_SET_IN_USE(arg_hL2tm, arg_L2Index);

    /*
     * The header was constructed in local memory; send it out to shared
     * memory.
     */
    _IX_CC_L2TM_WRITE_HEADER(arg_hL2tm, arg_L2Index, localHeader);

    /*
     * If the original state was 'Incomplete', adjust the counters.
     */
    if (state == IX_CC_L2TM_STATE_INCOMPLETE)
    {
        arg_hL2tm->numIncomplete--;
        arg_hL2tm->numValid++;
    }

    return IX_SUCCESS;

cleanup:
    /*
     * If the new IP address didn't go into the hash table, add the original
     * back in and return an error.
     */
    switch (arg_hL2tm->pIpAddrTypes[arg_L2Index])
    {
        case IX_CC_L2TM_IPADDR_IPV4:
            (void) ix_cc_hash_table_add_entry(arg_hL2tm->hIpv4Hash,
                  arg_hL2tm->pL3Table[arg_L2Index].ipv4Address,
                  hashData);
            break;

        case IX_CC_L2TM_IPADDR_IPV6:
            (void) ix_cc_hash_table128_add_entry(arg_hL2tm->hIpv6Hash,
                  arg_hL2tm->pL3Table[arg_L2Index].ipv6Address,
                  hashData);
            break;

        default:
            break;
    }

    return err;
}

ix_error ix_cc_l2tm_delete_entry(
    ix_cc_l2tm arg_hL2tm,
    ix_uint16 arg_L2Index)
{
    ix_cc_l2tm_state state;

    _IX_CC_L2TM_CHECK_HANDLE(arg_hL2tm);

    /*
     * Verify the index is in the 'Valid' or 'Incomplete' state
     */
    IX_ERROR_CR(_ix_cc_l2tm_get_state(arg_hL2tm, arg_L2Index, &state));

#if defined(IX_DEBUG)
    if (state == IX_CC_L2TM_STATE_IDLE)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
            ("The entry at index 0x%04lX is Idle", arg_L2Index));
    }
#endif /* IX_DEBUG */

    /*
     * Look up the L3 type for the existing entry
     * Retrieve the L3 information for the existing entry
     * Remove the IP address from the appropriate hash table
     */
    switch (arg_hL2tm->pIpAddrTypes[arg_L2Index])
    {
        case IX_CC_L2TM_IPADDR_IPV4:
        {
            IX_ERROR_CR(
                ix_cc_hash_table_delete_entry(arg_hL2tm->hIpv4Hash,
                    arg_hL2tm->pL3Table[arg_L2Index].ipv4Address));
        }
        break;

        case IX_CC_L2TM_IPADDR_IPV6:
        {
            IX_ERROR_CR(
                ix_cc_hash_table128_delete_entry(arg_hL2tm->hIpv6Hash,
                    arg_hL2tm->pL3Table[arg_L2Index].ipv6Address));
        }
        break;

        default:
            return IX_ERROR_LOCAL(IX_CC_ERROR_INTERNAL,
                ("The address type for the entry at 0x%04X is invalid: %d",
                 arg_L2Index, arg_hL2tm->pIpAddrTypes[arg_L2Index]));
    }

    /*
     * Set the state to 'Idle'
     */
    _IX_CC_L2TM_SET_IDLE(arg_hL2tm, arg_L2Index);

    switch (state)
    {
        case IX_CC_L2TM_STATE_VALID:
            arg_hL2tm->numValid--;
            break;

        case IX_CC_L2TM_STATE_INCOMPLETE:
            arg_hL2tm->numIncomplete--;
            break;

        /*
         * State has already been checked, no action for default.
         */
        default:
            break;
    }

    return IX_SUCCESS;
}

ix_error ix_cc_l2tm_get_entry(
    ix_cc_l2tm arg_hL2tm,
    ix_uint16 arg_L2Index,
    ix_cc_l2tm_state *arg_pState,
    ix_cc_l2tm_entry *arg_pL2Entry)
{
    ix_cc_l2tm_state state;

    _IX_CC_L2TM_CHECK_HANDLE(arg_hL2tm);
    IX_ERROR_DCHECK_ARG_ALIGN(3, arg_pState, ix_cc_l2tm_state);
    IX_ERROR_DCHECK_ARG_ALIGN(4, arg_pL2Entry, ix_cc_l2tm_entry);

    /*
     * Verify the index is valid.
     * Get the state of the entry.
     */
    IX_ERROR_CR(_ix_cc_l2tm_get_state(arg_hL2tm, arg_L2Index, &state));

    /*
     * If the caller wants the contents of the entry
     */
    if (arg_pL2Entry != NULL)
    {
        ix_ossl_memset(arg_pL2Entry, 0, sizeof(ix_cc_l2tm_entry));

        /*
         * If the state is 'Incomplete' or 'Valid', store the L3 information.
         */
        if ((state == IX_CC_L2TM_STATE_INCOMPLETE) ||
            (state == IX_CC_L2TM_STATE_VALID))
        {
            arg_pL2Entry->ipAddrType = arg_hL2tm->pIpAddrTypes[arg_L2Index];
            arg_pL2Entry->l3Address = arg_hL2tm->pL3Table[arg_L2Index];
        }
        else
        {
            arg_pL2Entry->ipAddrType = IX_CC_L2TM_IPADDR_LAST;
        }

        /*
         * If the state is 'Valid', store the L2 Information.
         */
        if (state == IX_CC_L2TM_STATE_VALID)
        {
            _ix_cc_l2tm_shared_entry localHeader;
            _IX_CC_L2TM_READ_HEADER(arg_hL2tm, arg_L2Index, localHeader);
            arg_pL2Entry->l2AddrType = arg_hL2tm->pL2AddrTypes[arg_L2Index];

            if (arg_pL2Entry->l2AddrType == IX_CC_L2TM_L2ADDR_ATM)
            {
                _IX_CC_L2TM_UNPACK_ATM_HEADER(localHeader.atm,
                        arg_pL2Entry->header.atm);
            }
            else /* Ethernet */
            {
                _IX_CC_L2TM_UNPACK_ETH_HEADER(localHeader.eth,
                    arg_pL2Entry->header.ether);
            }
        } /* state == Valid */
        else
        {
            arg_pL2Entry->l2AddrType = IX_CC_L2TM_L2ADDR_LAST;
        }
    } /* arg_pL2Entry != NULL */

    if (arg_pState != NULL)
    {
        *arg_pState = state;
    }

    return IX_SUCCESS;
}

ix_error ix_cc_l2tm_add_l3_info(
    ix_cc_l2tm                  arg_hL2tm,
    ix_uint16                   arg_L2Index,
    const ix_cc_l2tm_ipaddr    *arg_pIpAddress,
    ix_cc_l2tm_ipaddr_type      arg_IpAddressType)
{
    _IX_CC_L2TM_CHECK_HANDLE(arg_hL2tm);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pIpAddress);
    IX_ERROR_DCHECK_ARG_ALIGN(3, arg_pIpAddress, ix_cc_l2tm_ipaddr);

#if defined(IX_DEBUG)
    {
        ix_cc_l2tm_state state;

        /*
         * Verify the index is in the 'Idle' state
         */
        IX_ERROR_CR(_ix_cc_l2tm_get_state(arg_hL2tm, arg_L2Index, &state));
        if (state != IX_CC_L2TM_STATE_IDLE)
        {
            /* Report an error and return success */
            ix_ossl_message_log("An entry already exists at index 0x%04X\n", arg_L2Index);
            return(IX_SUCCESS);
        }
    }
#endif /* IX_DEBUG */

    /*
     * Add the IP address into the appropriate hash table
     * Set the L3 type for the entry
     * Set the L3 information for the entry
     */
    IX_ERROR_CR(_ix_cc_l2tm_add_l3_address(arg_hL2tm, arg_L2Index,
            arg_IpAddressType, arg_pIpAddress));

    /*
     * Set the state to 'Incomplete'
     */
    _IX_CC_L2TM_SET_IN_USE(arg_hL2tm, arg_L2Index);

    arg_hL2tm->numIncomplete++;

    return IX_SUCCESS;
}

ix_error ix_cc_l2tm_update_l3_info(
    ix_cc_l2tm                  arg_hL2tm,
    ix_uint16                   arg_L2Index,
    const ix_cc_l2tm_ipaddr    *arg_pIpAddress,
    ix_cc_l2tm_ipaddr_type      arg_IpAddressType)
{
    ix_cc_l2tm_state state;

    IX_ERROR_CR(_ix_cc_l2tm_get_state(arg_hL2tm, arg_L2Index, &state));

#if defined(IX_DEBUG)
    /*
     * Verify the index is in the 'Valid' or 'Incomplete' state
     */
    if (state == IX_CC_L2TM_STATE_IDLE)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
            ("Entry at index 0x%04X is not allocated", arg_L2Index));
    }
#endif /* IX_DEBUG */

    /*
     * Look up the L3 type for the existing entry
     * Retrieve the L3 information for the existing entry
     * Remove the IP address from the appropriate hash table
     */
    switch (arg_hL2tm->pIpAddrTypes[arg_L2Index])
    {
        case IX_CC_L2TM_IPADDR_IPV4:
            IX_ERROR_CR(ix_cc_hash_table_delete_entry(arg_hL2tm->hIpv4Hash,
                        arg_hL2tm->pL3Table[arg_L2Index].ipv4Address));
            break;

        case IX_CC_L2TM_IPADDR_IPV6:
            IX_ERROR_CR(ix_cc_hash_table128_delete_entry(arg_hL2tm->hIpv6Hash,
                        arg_hL2tm->pL3Table[arg_L2Index].ipv6Address));
            break;

        default:
            break;
    }

    /*
     * Add the IP address into the appropriate hash table
     * Set the L3 type for the entry
     * Set the L3 information for the entry
     */
    IX_ERROR_CR(_ix_cc_l2tm_add_l3_address(arg_hL2tm, arg_L2Index,
            arg_IpAddressType, arg_pIpAddress));

    /*
     * Remove the L2 information
     * Set the state to 'Incomplete'
     */
    _IX_CC_L2TM_DISABLE_ENTRY(arg_hL2tm, arg_L2Index);

    /*
     * If the original state was 'Valid', adjust the counters.
     */
    if (state == IX_CC_L2TM_STATE_VALID)
    {
        arg_hL2tm->numIncomplete++;
        arg_hL2tm->numValid--;
    }

    return IX_SUCCESS;
}

ix_error ix_cc_l2tm_assign_l2_info(
    ix_cc_l2tm arg_hL2tm,
    ix_cc_l2tm_entry *arg_pL2Entry,
    ix_uint16 *arg_pL2Index)
{
    ix_uint16 l2Index = 0;
    _ix_cc_l2tm_shared_entry localHeader;

    _IX_CC_L2TM_CHECK_HANDLE(arg_hL2tm);
    IX_ERROR_DCHECK_ARG_NULL(2, arg_pL2Entry);
    IX_ERROR_DCHECK_ARG_ALIGN(2, arg_pL2Entry, ix_cc_l2tm_entry);
    IX_ERROR_DCHECK_ARG_ALIGN(3, arg_pL2Index, ix_uint16);

    /*
     * Get the l2Index for the specified IP address
     */
    IX_ERROR_CR(ix_cc_l2tm_get_l2_index(arg_hL2tm, arg_pL2Entry->ipAddrType,
            &arg_pL2Entry->l3Address, &l2Index));

#if defined(IX_DEBUG)
    {
        ix_cc_l2tm_state state;
        /*
         * Verify the index is in the 'Incomplete' state
         */
        IX_ERROR_CR(_ix_cc_l2tm_get_state(arg_hL2tm, l2Index, &state));
        if (state != IX_CC_L2TM_STATE_INCOMPLETE)
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                ("The entry at index 0x%04lX is not Incomplete", l2Index));
        }

        /*
         * Check the input structure for validity
         */
        IX_ERROR_CR(_ix_cc_l2tm_verify_entry(arg_pL2Entry));
    }
#endif /* IX_DEBUG */

    /*
     * Set the header type for the entry
     */
    arg_hL2tm->pL2AddrTypes[l2Index] = arg_pL2Entry->l2AddrType;

    /*
     * Set the L2 information in the shared structure
     */
    switch (arg_pL2Entry->l2AddrType)
    {
        case IX_CC_L2TM_L2ADDR_ETHER:
            _IX_CC_L2TM_PACK_ETH_HEADER(arg_pL2Entry->header.ether,
                localHeader.eth);
            break;

        case IX_CC_L2TM_L2ADDR_ATM:
            _IX_CC_L2TM_PACK_ATM_HEADER(arg_pL2Entry->header.atm,
                localHeader.atm);
            break;

        /*
         * No default action here because the input structure was already
         * checked.
         */
        default:
            break;
    }

    /*
     * Set the state to 'Valid'
     */
    localHeader.flags |= IX_HTON32(_IX_CC_L2TM_FLAG_VALID);

    /*
     * The header was constructed in local memory; send it out to shared
     * memory.
     */
    _IX_CC_L2TM_WRITE_HEADER(arg_hL2tm, l2Index, localHeader);

    arg_hL2tm->numValid++;
    arg_hL2tm->numIncomplete--;

    if (arg_pL2Index != NULL)
    {
        *arg_pL2Index = l2Index;
    }

    return IX_SUCCESS;
}

ix_error ix_cc_l2tm_clear_l2_info(
    ix_cc_l2tm arg_hL2tm,
    ix_uint16 arg_L2Index)
{
    _IX_CC_L2TM_CHECK_HANDLE(arg_hL2tm);

#if defined(IX_DEBUG)
    {
        ix_cc_l2tm_state state;

        /*
         * Verify the index is in the 'Valid' state
         */
        IX_ERROR_CR(_ix_cc_l2tm_get_state(arg_hL2tm, arg_L2Index, &state));
        if (state != IX_CC_L2TM_STATE_VALID)
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
                ("The entry at index 0x%04lX is not Valid", arg_L2Index));
        }
    }
#endif /* IX_DEBUG */

    /*
     * Remove the L2 information
     * Set the state to 'Incomplete'
     */
    _IX_CC_L2TM_DISABLE_ENTRY(arg_hL2tm, arg_L2Index);
    arg_hL2tm->numValid--;
    arg_hL2tm->numIncomplete++;

    return IX_SUCCESS;
}

ix_error ix_cc_l2tm_get_l2_index(
    ix_cc_l2tm                  arg_hL2tm,
    ix_cc_l2tm_ipaddr_type      arg_IpAddressType,
    const ix_cc_l2tm_ipaddr    *arg_pIpAddress,
    ix_uint16                  *arg_pL2Index)
{
    ix_uint32 l2Index;

    _IX_CC_L2TM_CHECK_HANDLE(arg_hL2tm);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pIpAddress);
    IX_ERROR_DCHECK_ARG_ALIGN(3, arg_pIpAddress, ix_cc_l2tm_ipaddr);
    IX_ERROR_DCHECK_ARG_NULL(4, arg_pL2Index);
    IX_ERROR_DCHECK_ARG_ALIGN(4, arg_pL2Index, ix_uint16);

    /*
     * Look up the index from the appropriate hash table
     */
    switch (arg_IpAddressType)
    {
        case IX_CC_L2TM_IPADDR_IPV4:
            IX_ERROR_CR(ix_cc_hash_table_lookup(arg_hL2tm->hIpv4Hash,
                arg_pIpAddress->ipv4Address, NULL, (void*) &l2Index));
            break;

        case IX_CC_L2TM_IPADDR_IPV6:
            IX_ERROR_CR(ix_cc_hash_table128_lookup(arg_hL2tm->hIpv6Hash,
                arg_pIpAddress->ipv6Address, NULL, (void*) &l2Index));
            break;

        default:
            return IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
                ("Invalid IP Address Type: %d", arg_IpAddressType));
    }

    *arg_pL2Index = l2Index;

    return IX_SUCCESS;
}

ix_error ix_cc_l2tm_get_symbols(
    ix_cc_l2tm arg_hL2tm,
    ix_cc_l2tm_symbols *arg_pSymbols)
{
    ix_error err = IX_SUCCESS;
    ix_memory_type type = 0;
    ix_uint32 channel = 0;
    ix_uint32 baseAddress;

    _IX_CC_L2TM_CHECK_HANDLE(arg_hL2tm);
    IX_ERROR_DCHECK_ARG_NULL(2, arg_pSymbols);
    IX_ERROR_DCHECK_ARG_ALIGN(2, arg_pSymbols, ix_cc_l2tm_symbols);

    /*
     * Fill in the symbols using available information
     */
    err = ix_rm_get_phys_offset(arg_hL2tm->pL2Table, &type, &channel,
        &baseAddress, NULL);
    if (err != IX_SUCCESS)
    {
        return IX_ERROR_CHAIN(err, IX_CC_ERROR_INTERNAL, 0,
            ("Could not get Offset for address %p", arg_hL2tm->pL3Table));
    }

    if (type == IX_MEMORY_TYPE_SRAM)
    {
        IX_CC_SET_SRAM_CHANNEL_INTO_OFFSET(channel, baseAddress);
    }

    arg_pSymbols->l2TableBase = baseAddress;

    return IX_SUCCESS;
}

ix_error ix_cc_l2tm_get_statistics(
    ix_cc_l2tm          arg_hL2tm,
    ix_cc_l2tm_stats   *arg_pStatistics)
{
    _IX_CC_L2TM_CHECK_HANDLE(arg_hL2tm);
    IX_ERROR_DCHECK_ARG_NULL(2, arg_pStatistics);
    IX_ERROR_DCHECK_ARG_ALIGN(2, arg_pStatistics, ix_cc_l2tm_stats);

    /*
     * Fill in the statistics using available information
     */
    arg_pStatistics->numberOfValidEntries = arg_hL2tm->numValid;
    arg_pStatistics->numberOfIncompleteEntries = arg_hL2tm->numIncomplete;
    arg_pStatistics->tableCapacity = arg_hL2tm->capacity;
    arg_pStatistics->heapMemoryUsed =
        _IX_CC_L2TM_CALC_HEAP_SIZE(arg_hL2tm->capacity);
    arg_pStatistics->sharedMemoryUsed =
        _IX_CC_L2TM_GET_TABLE_MEM_SIZE(arg_hL2tm->capacity);

    return IX_SUCCESS;
}

ix_error ix_cc_l2tm_purge(
    ix_cc_l2tm          arg_hL2tm)
{
    ix_error err = IX_SUCCESS;
    ix_uint16 i;
    ix_uint32 hashEntry;

    _IX_CC_L2TM_CHECK_HANDLE(arg_hL2tm);

    /*
     * Mark all entries as 'Idle'
     */
    for (i = 0; i < arg_hL2tm->capacity; i++)
    {
        _IX_CC_L2TM_SET_IDLE(arg_hL2tm, i);
    }

    arg_hL2tm->numValid = 0;
    arg_hL2tm->numIncomplete = 0;

    /*
     * Clear the contents of the IPv4 hash table
     */
    err = ix_cc_hash_table_get_first_entry(arg_hL2tm->hIpv4Hash, &hashEntry, NULL);
    while (err == IX_SUCCESS)
    {
        IX_ERROR_CG(ix_cc_hash_table_delete_entry(arg_hL2tm->hIpv4Hash,
                    hashEntry), err, report);
        err = ix_cc_hash_table_get_first_entry(arg_hL2tm->hIpv4Hash,
            &hashEntry, NULL);
    }
    if (IX_ERROR_GET_CODE(err) != IX_CC_ERROR_ENTRY_NOT_FOUND)
    {
        goto report;
    }

    /*
     * Clear the contents of the IPv6 hash table
     */
    IX_ERROR_CR(ix_cc_hash_table128_purge(arg_hL2tm->hIpv6Hash));

    return IX_SUCCESS;

report:
    (void) ix_cc_hash_table128_purge(arg_hL2tm->hIpv6Hash);
    return err;
}

ix_error _ix_cc_l2tm_print_l3(
    ix_cc_l2tm arg_hL2tm,
    ix_uint16  arg_L2Index)
{
    if (arg_hL2tm->pIpAddrTypes[arg_L2Index] == IX_CC_L2TM_IPADDR_IPV4)
    {
        /*
         * IPv4
         */
        ix_uint32 ipAddr = arg_hL2tm->pL3Table[arg_L2Index].ipv4Address;
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
	ix_ossl_message_log("0x%04X: IPv4 %ld.%ld.%ld.%ld\n", arg_L2Index,
	            (ipAddr >> 24) & 0xFF,
	            (ipAddr >> 16) & 0xFF,
	            (ipAddr >>  8) & 0xFF,
	            (ipAddr      ) & 0xFF);
#else

        printf("0x%04X: IPv4 %ld.%ld.%ld.%ld\n", arg_L2Index,
            (ipAddr >> 24) & 0xFF,
            (ipAddr >> 16) & 0xFF,
            (ipAddr >>  8) & 0xFF,
            (ipAddr      ) & 0xFF);
#endif

    }
    else
    {
        /*
         * IPv6
         */
        ix_uint128 *pIpAddr =
            &arg_hL2tm->pL3Table[arg_L2Index].ipv6Address;

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_ossl_message_log("0x%04X: IPv6 %X:%X:%X:%X:%X:%X:%X:%X\n",
            arg_L2Index,
            (ix_uint16) (pIpAddr->aUint32[0] >> 16),
            (ix_uint16) (pIpAddr->aUint32[0] & 0xFFFF),
            (ix_uint16) (pIpAddr->aUint32[1] >> 16),
            (ix_uint16) (pIpAddr->aUint32[1] & 0xFFFF),
            (ix_uint16) (pIpAddr->aUint32[2] >> 16),
            (ix_uint16) (pIpAddr->aUint32[2] & 0xFFFF),
            (ix_uint16) (pIpAddr->aUint32[3] >> 16),
            (ix_uint16) (pIpAddr->aUint32[3] & 0xFFFF));

#else
        printf("0x%04X: IPv6 %X:%X:%X:%X:%X:%X:%X:%X\n",
            arg_L2Index,
            (ix_uint16) (pIpAddr->aUint32[0] >> 16),
            (ix_uint16) (pIpAddr->aUint32[0] & 0xFFFF),
            (ix_uint16) (pIpAddr->aUint32[1] >> 16),
            (ix_uint16) (pIpAddr->aUint32[1] & 0xFFFF),
            (ix_uint16) (pIpAddr->aUint32[2] >> 16),
            (ix_uint16) (pIpAddr->aUint32[2] & 0xFFFF),
            (ix_uint16) (pIpAddr->aUint32[3] >> 16),
            (ix_uint16) (pIpAddr->aUint32[3] & 0xFFFF));
#endif
    }

    return IX_SUCCESS;
}

ix_error _ix_cc_l2tm_print_l2(
    ix_cc_l2tm arg_hL2tm,
    ix_uint16  arg_L2Index)
{
    _ix_cc_l2tm_shared_entry localHeader;

    _IX_CC_L2TM_READ_HEADER(arg_hL2tm, arg_L2Index, localHeader);

    if (arg_hL2tm->pL2AddrTypes[arg_L2Index] == IX_CC_L2TM_L2ADDR_ETHER)
    {
        /*
         * Ethernet
         */
        ix_cc_l2tm_ether_header ether;
        _IX_CC_L2TM_UNPACK_ETH_HEADER(localHeader.eth, ether);

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_ossl_message_log("    ETH: "
               "Dest = %02X:%02X:%02X:%02X:%02X:%02X, "
               "Source = %02X:%02X:%02X:%02X:%02X:%02X, "
               "Proto = 0x%04X\n",
               ether.destMAC[0], ether.destMAC[1],
              ether.destMAC[2], ether.destMAC[3],
               ether.destMAC[4], ether.destMAC[5],
               ether.sourceMAC[0], ether.sourceMAC[1],
               ether.sourceMAC[2], ether.sourceMAC[3],
               ether.sourceMAC[4], ether.sourceMAC[5],
               ether.protocol);

#else
        printf("    ETH: "
               "Dest = %02X:%02X:%02X:%02X:%02X:%02X, "
               "Source = %02X:%02X:%02X:%02X:%02X:%02X, "
               "Proto = 0x%04X\n",
               ether.destMAC[0], ether.destMAC[1],
               ether.destMAC[2], ether.destMAC[3],
               ether.destMAC[4], ether.destMAC[5],
               ether.sourceMAC[0], ether.sourceMAC[1],
               ether.sourceMAC[2], ether.sourceMAC[3],
               ether.sourceMAC[4], ether.sourceMAC[5],
               ether.protocol);
#endif
    }
    else
    {
        /*
         * ATM
         */
        ix_cc_l2tm_atm_header atm;
        _IX_CC_L2TM_UNPACK_ATM_HEADER(localHeader.atm, atm);
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_ossl_message_log("    ATM: vpiVci = 0x%08lX, vcq = 0x%04lX, "
			    "LLC/SNAP = 0x%08lX%08lX\n",
			     atm.vpiVci, atm.vcq,
			    (ix_uint32) (atm.llcSnapHeader >> 32),
			    (ix_uint32) atm.llcSnapHeader);
#else
        printf("    ATM: vpiVci = 0x%08lX, vcq = 0x%04lX, "
               "LLC/SNAP = 0x%08lX%08lX\n",
               atm.vpiVci, atm.vcq,
               (ix_uint32) (atm.llcSnapHeader >> 32),
               (ix_uint32) atm.llcSnapHeader);
#endif
    }

    return IX_SUCCESS;
}

ix_error ix_cc_l2tm_dump(
    ix_cc_l2tm          arg_hL2tm)
{
    ix_uint16 currIndex;

    _IX_CC_L2TM_CHECK_HANDLE(arg_hL2tm);

    /*
     * For each entry in the table:
     */
    for (currIndex = 0; currIndex < arg_hL2tm->capacity; currIndex++)
    {
        ix_cc_l2tm_state state;

        /*
         * Get the state of the entry.
         */
        IX_ERROR_CR(_ix_cc_l2tm_get_state(arg_hL2tm, currIndex, &state));

        /*
         * If the entry is INCOMPLETE or VALID, print the IP Type and IP
         * Address.
         */
        if (state != IX_CC_L2TM_STATE_IDLE)
        {
            IX_ERROR_CR(_ix_cc_l2tm_print_l3(arg_hL2tm, currIndex));

            /*
             * If the entry is VALID, print the L2 Type and L2 Address.
             */
            if (state == IX_CC_L2TM_STATE_VALID)
            {
                IX_ERROR_CR(_ix_cc_l2tm_print_l2(arg_hL2tm, currIndex));
            }
            else
            {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
                ix_ossl_message_log("    (Incomplete entry; no L2 Header Information)\n");
#else
                printf("    (Incomplete entry; no L2 Header Information)\n");
#endif
            }
        }
    }

    return IX_SUCCESS;
}

ix_error _ix_cc_l2tm_get_state(
    ix_cc_l2tm arg_hL2tm,
    ix_uint16 arg_L2Index,
    ix_cc_l2tm_state *arg_pState)
{
    ix_cc_l2tm_state state;

    /*
     * If the index is out of range, return an error
     */
    if (arg_L2Index > arg_hL2tm->capacity)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
            ("The index is beyond the capacity of the table: 0x%04X",
             arg_L2Index));
    }

    /*
     * If the entry is not in use, the state is 'Idle'. Return.
     */
    if (!_IX_CC_L2TM_IS_IN_USE(arg_hL2tm, arg_L2Index))
    {
        state = IX_CC_L2TM_STATE_IDLE;
    }
    else
    {
        /*
         * Get the 'valid' flag from shared memory.
         */
        ix_uint32 flags = _IX_CC_L2TM_GET_FLAGS(arg_hL2tm, arg_L2Index);

        /*
         * If the 'valid' flag is zero, the state is 'Incomplete'. Return.
         * The state is 'Valid'.
         */
        if (flags & _IX_CC_L2TM_FLAG_VALID)
        {
            state = IX_CC_L2TM_STATE_VALID;
        }
        else
        {
            state = IX_CC_L2TM_STATE_INCOMPLETE;
        }
    }

    *arg_pState = state;
    return IX_SUCCESS;
}

#if defined(IX_DEBUG)
ix_error _ix_cc_l2tm_verify_entry(
    ix_cc_l2tm_entry *arg_pL2Entry)
{
    /*
     * If the address type is out of range, return an error.
     */
    if (arg_pL2Entry->ipAddrType >= IX_CC_L2TM_IPADDR_LAST)
    {
        return IX_ERROR_WARNING(IX_CC_L2TM_ERROR_INVALID_ENTRY,
            ("Undefined address type: %d", arg_pL2Entry->ipAddrType));
    }

    switch (arg_pL2Entry->l2AddrType)
    {
        /*
         * If the header type is ATM:
         */
        case IX_CC_L2TM_L2ADDR_ATM:
            /*
             * Verify bits 28-31 of vpiVci are zero
             */
            if (arg_pL2Entry->header.atm.vpiVci & 0xF0000000)
            {
                return IX_ERROR_WARNING(IX_CC_L2TM_ERROR_INVALID_ENTRY,
                    ("Invalid vpiVci: 0x%08lX",
                     arg_pL2Entry->header.atm.vpiVci));
            }

            /*
             * Verify bits 16-31 of vcq are zero
             */
            if (arg_pL2Entry->header.atm.vcq & 0xFFFF0000)
            {
                return IX_ERROR_WARNING(IX_CC_L2TM_ERROR_INVALID_ENTRY,
                    ("Invalid vcq: 0x%08lX",
                     arg_pL2Entry->header.atm.vcq));
            }
            break;

        /*
         * Else, if the header type is not Ethernet, return an error.
         */
        case IX_CC_L2TM_L2ADDR_ETHER:
            break;

        default:
            return IX_ERROR_WARNING(IX_CC_L2TM_ERROR_INVALID_ENTRY,
                ("Undefined header type: %d", arg_pL2Entry->l2AddrType));
    }

    /*
     * The entry is acceptable, return success.
     */
    return IX_SUCCESS;
}
#endif /* IX_DEBUG */

ix_error _ix_cc_l2tm_add_l3_address(
    ix_cc_l2tm                  arg_hL2tm, 
    ix_uint16                   arg_L2Index, 
    ix_cc_l2tm_ipaddr_type      arg_IpAddressType,
    const ix_cc_l2tm_ipaddr    *arg_pIpAddress)
{
    void *hashData = (void*) (ix_uint32) arg_L2Index;
    switch (arg_IpAddressType)
    {
        case IX_CC_L2TM_IPADDR_IPV4:
            IX_ERROR_CR(ix_cc_hash_table_add_entry(arg_hL2tm->hIpv4Hash,
                        arg_pIpAddress->ipv4Address, hashData));
            break;

        case IX_CC_L2TM_IPADDR_IPV6:
            IX_ERROR_CR(ix_cc_hash_table128_add_entry(arg_hL2tm->hIpv6Hash,
                        arg_pIpAddress->ipv6Address, hashData));

        default:
            break;
    }

    /*
     * Add the L3 information into the L3 table
     */
    arg_hL2tm->pL3Table[arg_L2Index] = *arg_pIpAddress;

    /*
     * Set the IP address type for the entry
     */
    arg_hL2tm->pIpAddrTypes[arg_L2Index] = arg_IpAddressType;

    return IX_SUCCESS;
}
