/*
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *     Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * $Id: named_l2tm.c,v 1.5 2003/01/24 23:58:48 lbmiller Exp $
 */

/**
 * @file named_l2tm.c
 *
 * Implementation of the functions that allow a L2 Table Manager to be used
 * by multiple clients.
 */

#include "ix_cc_error.h"
#include "ix_ossl.h"
#include "ix_rm.h"
#include "cc/ix_cc_l2tm.h"
#include "cc/l2tm/internal/internal_l2tm.h"
#include "cc/l2tm/internal/internal_named.h"

/**
 * Opens a specified key in the registry. The caller provides the name of the
 * L2TM to open. This function opens a handle that can be used to read
 * configuration items or the handle and reference count.
 *
 * @param       arg_pName (IN)
 *              Name of the L2 Table Manager to open.
 *
 * @param       arg_phL2tmReg (OUT)
 *              Handle to the portion of the registry related to this L2TM.
 *
 * @pseudocode
 * -# Open the L2 Table Manager section of the registry.
 * -# Open the L2 Table Manager in question.
 * -# Close the L2 Table Manager section.
 * -# Return the open handle to the caller.
 *
 * @returns IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_L2TM_ERROR_NOT_FOUND The L2TM with the specified name could
 *      not be found.
 * @retval IX_CC_L2TM_ERROR_REGISTRY A problem with the registry was
 *      encountered.
 */
static ix_error _ix_cc_l2tm_reg_open(
    const char *arg_pName,
    ix_configuration_property_handle *arg_phL2tmReg)
{
    ix_configuration_property_handle hRoot;
    ix_configuration_property_handle hL2tmReg;
    ix_error err;

    /*
     * Open the L2 Table Manager section of the registry.
     */
    IX_ERROR_CRT(ix_rm_cp_property_open(ix_rm_cp_get_core_root(),
                _IX_CC_L2TM_REG_ROOT, &hRoot),
            IX_CC_L2TM_ERROR_REGISTRY, IX_ERROR_LEVEL_WARNING);

    /*
     * Open the L2 Table Manager in question.
     */
    IX_ERROR_CGT(ix_rm_cp_property_open(hRoot, arg_pName, &hL2tmReg),
            err, IX_CC_L2TM_ERROR_NOT_FOUND, IX_ERROR_LEVEL_WARNING,
            cleanRoot);

    /*
     * Close the L2 Table Manager section.
     */
    IX_ERROR_CRT(ix_rm_cp_property_close(hRoot),
            IX_CC_L2TM_ERROR_REGISTRY, IX_ERROR_LEVEL_WARNING);

    /*
     * Return the open handle to the caller.
     */
    *arg_phL2tmReg = hL2tmReg;
    return IX_SUCCESS;

cleanRoot:
    (void) ix_rm_cp_property_close(hRoot);
    return err;
}

/*
 * Gets the Named L2TM's handle from the registry. If the handle does not exist
 * in the registry, it is not considered an error; this function reports that
 * the handle is NULL.
 *
 * @param       arg_hL2tmReg (IN)
 *              Handle to the registry, as returned by _ix_cc_l2tm_reg_open().
 *
 * @param       arg_phL2tm (OUT)
 *              Handle as obtained from the registry. An output handle of NULL
 *              indicates that the L2TM by that name does not exist.
 *
 * @pseudocode
 * -# Attempt to open the 'handle' property in the registry.
 * -# If the open was successful
 *      -# Read the handle.
 *      -# Close the property.
 *
 * @returns IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_L2TM_ERROR_REGISTRY A problem with the registry was
 *      encountered.
 */
static ix_error _ix_cc_l2tm_reg_get_handle(
    ix_configuration_property_handle arg_hL2tmReg,
    ix_cc_l2tm *arg_phL2tm)
{
    ix_configuration_property_handle hL2tmHandle;
    ix_error err;

    *arg_phL2tm = NULL;

    /*
     * Attempt to open the 'handle' property in the registry.
     */
    err = ix_rm_cp_property_open(arg_hL2tmReg,
                _IX_CC_L2TM_REG_HANDLE, &hL2tmHandle);

    /*
     * If the open was successful
     */
    if (err == IX_SUCCESS)
    {
        /*
         * Read the handle.
         */
        IX_ERROR_CGT(ix_rm_cp_property_get_value_uint32(
                    hL2tmHandle, (ix_uint32*) arg_phL2tm),
                err, IX_CC_L2TM_ERROR_REGISTRY, IX_ERROR_LEVEL_WARNING,
                cleanReg);

        /*
         * Close the property.
         */
        IX_ERROR_CRT(ix_rm_cp_property_close(hL2tmHandle),
                IX_CC_L2TM_ERROR_REGISTRY, IX_ERROR_LEVEL_WARNING);
    }

    return IX_SUCCESS;

cleanReg:
    (void) ix_rm_cp_property_close(hL2tmHandle);
    return err;
}

/*
 * Increment the number of clients using this L2TM. If the registry does not
 * already have a 'refcount' value, create it.
 *
 * @param       arg_hL2tmReg (IN)
 *              Handle to the registry, as returned by _ix_cc_l2tm_reg_open().
 *
 * @pseudocode
 * -# Attempt to open the 'refcount' property.
 * -# If the open was successful, read the existing value.
 * -# If the open was not successful, create it.
 * -# Increment the reference count.
 * -# Store the reference count in the registry.
 * -# If the property wasn't just created, close it.
 *
 * @returns IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_L2TM_ERROR_REGISTRY A problem with the registry was
 *      encountered.
 */
static ix_error _ix_cc_l2tm_reg_inc(
    ix_configuration_property_handle arg_hL2tmReg)
{
    ix_error err;
    ix_configuration_property_handle hRefCount;
    ix_uint32 refCount = 0;
    int created = 0;

    /*
     * Attempt to open the 'refcount' property.
     */
    err = ix_rm_cp_property_open(arg_hL2tmReg,
                _IX_CC_L2TM_REG_REFCOUNT, &hRefCount);

    if (err == IX_SUCCESS)
    {
        /*
         * If the open was successful, read the existing value.
         */
        IX_ERROR_CGT(ix_rm_cp_property_get_value_uint32(
                    hRefCount, &refCount),
                err, IX_CC_L2TM_ERROR_REGISTRY, IX_ERROR_LEVEL_WARNING,
                cleanReg);
    }
    else
    {
        /*
         * If the open was not successful, create it.
         */
        IX_ERROR_CRT(ix_rm_cp_property_create(arg_hL2tmReg,
                    _IX_CC_L2TM_REG_REFCOUNT, 0, &hRefCount),
                IX_CC_L2TM_ERROR_REGISTRY, IX_ERROR_LEVEL_WARNING);
        created = 1;
    }

    /*
     * Increment the reference count.
     */
    refCount++;

    /*
     * Store the reference count in the registry.
     */
    IX_ERROR_CRT(ix_rm_cp_property_set_value_uint32(hRefCount, refCount),
            IX_CC_L2TM_ERROR_REGISTRY, IX_ERROR_LEVEL_WARNING);

    /*
     * If the property wasn't just created, close it.
     */
    if (created == 0)
    {
        IX_ERROR_CRT(ix_rm_cp_property_close(hRefCount),
                IX_CC_L2TM_ERROR_REGISTRY, IX_ERROR_LEVEL_WARNING);
    }

    return IX_SUCCESS;

cleanReg:
    (void) ix_rm_cp_property_close(hRefCount);
    return err;
}

/*
 * Decrement the number of clients using this L2TM.
 *
 * @param       arg_hL2tmReg (IN)
 *              Handle to the registry, as returned by _ix_cc_l2tm_reg_open().
 *
 * @param       arg_pRefCount (OUT)
 *              Resulting reference count, after the decrement
 *
 * @pseudocode
 * -# Open the 'refcount' property.
 * -# Read the existing value from the registry.
 * -# Decrement the reference count.
 * -# Store the reference count in the registry.
 * -# Close the property.
 *
 * @returns IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_L2TM_ERROR_REGISTRY A problem with the registry was
 *      encountered.
 */
static ix_error _ix_cc_l2tm_reg_dec(
    ix_configuration_property_handle arg_hL2tmReg,
    ix_uint32 *arg_pRefCount)
{
    ix_error err;
    ix_configuration_property_handle hRefCount;
    ix_uint32 refCount = 0;

    /*
     * Open the 'refcount' property.
     */
    IX_ERROR_CRT(ix_rm_cp_property_open(arg_hL2tmReg,
                _IX_CC_L2TM_REG_REFCOUNT, &hRefCount),
            IX_CC_L2TM_ERROR_REGISTRY, IX_ERROR_LEVEL_WARNING);

    /*
     * Read the existing value from the registry.
     */
    IX_ERROR_CGT(ix_rm_cp_property_get_value_uint32(
                hRefCount, &refCount),
            err, IX_CC_L2TM_ERROR_REGISTRY, IX_ERROR_LEVEL_WARNING,
            cleanReg);

    /*
     * Decrement the reference count.
     */
    refCount--;

    /*
     * Store the reference count in the registry.
     */
    IX_ERROR_CGT(ix_rm_cp_property_set_value_uint32(hRefCount, refCount),
            err, IX_CC_L2TM_ERROR_REGISTRY, IX_ERROR_LEVEL_WARNING, cleanReg);

    /*
     * Close the property.
     */
    IX_ERROR_CRT(ix_rm_cp_property_close(hRefCount),
            IX_CC_L2TM_ERROR_REGISTRY, IX_ERROR_LEVEL_WARNING);

    *arg_pRefCount = refCount;
    return IX_SUCCESS;

cleanReg:
    (void) ix_rm_cp_property_close(hRefCount);
    return err;
}

/*
 * Set the handle for this L2 Table Manager in the registry.
 *
 * @param       arg_hL2tmReg (IN)
 *              Handle to the registry, as returned by _ix_cc_l2tm_reg_open().
 *
 * @param       arg_hL2tm (IN)
 *              Handle to the L2TM, as returned by ix_cc_l2tm_init().
 *
 * @pseudocode
 * -# Attempt to open the 'handle' property.
 * -# If it couldn't be opened, create it.
 * -# Store the handle in the registry.
 * -# If the property wasn't just created, close it.
 *
 * @returns IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_L2TM_ERROR_REGISTRY A problem with the registry was
 *      encountered.
 */
static ix_error _ix_cc_l2tm_reg_set_handle(
    ix_configuration_property_handle arg_hL2tmReg,
    ix_cc_l2tm arg_hL2tm)
{
    ix_error err;
    ix_configuration_property_handle hHandle;
    int created = 0;

    /*
     * Attempt to open the 'handle' property.
     */
    err = ix_rm_cp_property_open(arg_hL2tmReg,
                _IX_CC_L2TM_REG_HANDLE, &hHandle);

    /*
     * If it couldn't be opened, create it.
     */
    if (err != IX_SUCCESS)
    {
        IX_ERROR_CRT(ix_rm_cp_property_create(arg_hL2tmReg,
                    _IX_CC_L2TM_REG_HANDLE, 0, &hHandle),
                IX_CC_L2TM_ERROR_REGISTRY, IX_ERROR_LEVEL_WARNING);
        created = 1;
    }

    /*
     * Store the handle in the registry.
     */
    IX_ERROR_CGT(ix_rm_cp_property_set_value_uint32(hHandle,
                (ix_uint32) arg_hL2tm),
            err, IX_CC_L2TM_ERROR_REGISTRY, IX_ERROR_LEVEL_WARNING, cleanReg);

    /*
     * If the property wasn't just created, close it.
     */
    if (created == 0)
    {
        IX_ERROR_CRT(ix_rm_cp_property_close(hHandle),
                IX_CC_L2TM_ERROR_REGISTRY, IX_ERROR_LEVEL_WARNING);
    }

    return IX_SUCCESS;

cleanReg:
    (void) ix_rm_cp_property_close(hHandle);
    return err;
}

/*
 * Gets a single 32-bit value from the registry.
 *
 * @param       arg_hL2tmReg (IN)
 *              Handle to the registry, as returned by _ix_cc_l2tm_reg_open().
 *
 * @param       arg_pValueName (IN)
 *              Name of the value to read.
 *
 * @param       arg_pValue (OUT)
 *              Value contained in the registry.
 *
 * @pseudocode
 * -# Open the property.
 * -# Read the value.
 * -# Close the property.
 *
 * @returns IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_L2TM_ERROR_REGISTRY A problem with the registry was
 *      encountered.
 */
static ix_error _ix_cc_l2tm_reg_read_value(
    ix_configuration_property_handle arg_hL2tmReg,
    const char *arg_pValueName,
    ix_uint32 *arg_pValue)
{
    ix_error err;
    ix_configuration_property_handle hValue;

    /*
     * Open the property.
     */
    IX_ERROR_CRT(ix_rm_cp_property_open(arg_hL2tmReg,
                arg_pValueName, &hValue),
            IX_CC_L2TM_ERROR_REGISTRY, IX_ERROR_LEVEL_WARNING);

    /*
     * Read the value.
     */
    IX_ERROR_CGT(ix_rm_cp_property_get_value_uint32(hValue, arg_pValue),
            err, IX_CC_L2TM_ERROR_REGISTRY, IX_ERROR_LEVEL_WARNING, cleanReg);

    /*
     * Close the property.
     */
    IX_ERROR_CRT(ix_rm_cp_property_close(hValue),
            IX_CC_L2TM_ERROR_REGISTRY, IX_ERROR_LEVEL_WARNING);

    return IX_SUCCESS;

cleanReg:
    (void) ix_rm_cp_property_close(hValue);
    return err;
}

/*
 * Get the configuration of an L2 Table Manager from the registry.
 *
 * @param       arg_hL2tmReg (IN)
 *              Handle to the registry, as returned by _ix_cc_l2tm_reg_open().
 *
 * @param       arg_pConfig (OUT)
 *              Configuration structure used to create a new L2TM.
 *
 * @pseudocode
 * -# Retrieve the size of the L2TM.
 * -# Retrieve the type of memory to use.
 * -# Retrieve the memory channel to use.
 *
 * @returns IX_SUCCESS if successful or a valid ix_error for failure.
 * @retval IX_CC_L2TM_ERROR_REGISTRY A problem with the registry was
 *      encountered.
 */
static ix_error _ix_cc_l2tm_reg_get_config(
    ix_configuration_property_handle arg_hL2tmReg,
    ix_cc_l2tm_config *arg_pConfig)
{
    ix_uint32 tempVal;

    /*
     * Retrieve the size of the L2TM.
     */
    IX_ERROR_CR(_ix_cc_l2tm_reg_read_value(arg_hL2tmReg,
                _IX_CC_L2TM_REG_NUM_ENTRIES, &tempVal));
    arg_pConfig->numEntries = tempVal;

    /*
     * Retrieve the type of memory to use.
     */
    IX_ERROR_CR(_ix_cc_l2tm_reg_read_value(arg_hL2tmReg,
                _IX_CC_L2TM_REG_MEMORY_TYPE, &tempVal));
    arg_pConfig->memoryType = tempVal;

    /*
     * Retrieve the memory channel to use.
     */
    IX_ERROR_CR(_ix_cc_l2tm_reg_read_value(arg_hL2tmReg,
                _IX_CC_L2TM_REG_MEMORY_CHANNEL, &tempVal));
    arg_pConfig->memoryChannel = tempVal;

    return IX_SUCCESS;
}

ix_error ix_cc_l2tm_create(
    const char *arg_pName,
    ix_cc_l2tm *arg_phL2tm)
{
    ix_configuration_property_handle hL2tmReg;
    ix_error err;
    ix_cc_l2tm hL2tm;
    int created = 0;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_pName);
    IX_ERROR_DCHECK_ARG_NULL(2, arg_phL2tm);
    IX_ERROR_DCHECK_ARG_ALIGN(2, arg_phL2tm, ix_cc_l2tm);

    /*
     * Look up the handle in the Registry
     */
    IX_ERROR_CR(_ix_cc_l2tm_reg_open(arg_pName, &hL2tmReg));
    IX_ERROR_CG(_ix_cc_l2tm_reg_get_handle(hL2tmReg, &hL2tm), err, cleanReg);

    /*
     * If the handle doesn't already exist
     */
    if (hL2tm == NULL)
    {
        ix_cc_l2tm_config config;

        /*
         * Read the configuration from the Registry
         */
        IX_ERROR_CG(_ix_cc_l2tm_reg_get_config(hL2tmReg, &config),
                err, cleanReg);

        /*
         * Create a new L2TM using the configuration
         */
        IX_ERROR_CG(ix_cc_l2tm_init(&config, &hL2tm), err, cleanReg);
        created = 1;

        /*
         * Assign the name to the new L2TM.
         */
        strcpy(hL2tm->name, arg_pName);

        /*
         * Set the handle in the registry
         */
        IX_ERROR_CG(_ix_cc_l2tm_reg_set_handle(hL2tmReg, hL2tm),
                err, cleanL2tm);
    }

    /*
     * Increment the reference count for the handle
     */
    IX_ERROR_CG(_ix_cc_l2tm_reg_inc(hL2tmReg), err, cleanL2tm);

    IX_ERROR_CRT(ix_rm_cp_property_close(hL2tmReg),
            IX_CC_L2TM_ERROR_REGISTRY, IX_ERROR_LEVEL_WARNING);

    /*
     * Return the handle to the caller
     */
    *arg_phL2tm = hL2tm;

    return IX_SUCCESS;

cleanL2tm:
    if (created)
    {
        (void) ix_cc_l2tm_fini(hL2tm);
    }
cleanReg:
    (void) ix_rm_cp_property_close(hL2tmReg);
    return err;
}

ix_error ix_cc_l2tm_destroy(
    ix_cc_l2tm arg_hL2tm)
{
    ix_error err;
    ix_uint32 refCount = 0;
    ix_configuration_property_handle hL2tmReg;

    IX_ERROR_CR(_ix_cc_l2tm_reg_open(arg_hL2tm->name, &hL2tmReg));

    /*
     * Decrement the handle's refcount
     */
    IX_ERROR_CG(_ix_cc_l2tm_reg_dec(hL2tmReg, &refCount), err, clean_reg);

    /*
     * If the refcount is zero
     */
    if (refCount == 0)
    {
        /*
         * Destroy the L2TM
         */
        IX_ERROR_CG(ix_cc_l2tm_fini(arg_hL2tm), err, clean_reg);

        /*
         * Clear the handle from the Registry
         */
        IX_ERROR_CG(_ix_cc_l2tm_reg_set_handle(hL2tmReg, NULL), err, clean_reg);
    }

    IX_ERROR_CR(ix_rm_cp_property_close(hL2tmReg));

    return IX_SUCCESS;

clean_reg:
    (void) ix_rm_cp_property_close(hL2tmReg);
    return err;
}

