/**
 * ===========================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      ix_lkup.h
 *
 * = DESCRIPTION
 *      Main header file for the the core side of the Lookup Library.  This
 *      file should be included by both the users of the library, and by
 *      the implementations.  The users should not need to include other
 *      implementation-specific header files, unless they want to use 
 *      vendor-specific features of the implementation.
 *
 * ===========================================================================
 * $Id: ix_lkup.h,v 1.15 2003/10/24 23:58:56 ktseng Exp $
 */

#if !defined(__IX_LKUP_H__)
#define __IX_LKUP_H__

#include "ix_types.h"
#include "ix_error.h"

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

/*
 * NAME: ix_lkup
 *
 * DESCRIPTION: The ix_lkup is a handle that is returned when the
 * application first initializes and gets a handle to the lookup library.
 * To the user of this library, the ix_lkup is opaque. To the implementer of
 * the library, the details of the data structure are in the implementation-
 * specific section.
 */
typedef struct ix_s_lkup* ix_lkup;

/*
 * NAME: ix_lkup_table
 *
 * DESCRIPTION: The ix_lkup_table is a handle that is returned when a new
 * table is created by calling IX_LKUP_CREATE_TABLE() on a valid ix_lkup. As
 * with the ix_lkup, the contents of this structure are opaque to the
 * application. To the implementer of the library, the details of the data
 * structure are in the implementation-specific section.
 */
typedef struct ix_s_lkup_table* ix_lkup_table;

/*
 * NAME: ix_lkup_cookie
 *
 * DESCRIPTION: The ix_lkup_cookie is an opaque cookie that is passed
 * between some of the API calls. Primarily calls that are used for
 * enumerating the contents of a table. The cookie is an opaque entity that is
 * passed by the caller without interpretation.
 */
typedef void* ix_lkup_cookie;

/*
 * Constants used for data structures.
 */
#define IX_LKUP_MEM_SRAM        1
#define IX_LKUP_MEM_DRAM        2
#define IX_LKUP_MEM_TCAM        3

/*
 * NAME: ix_lkup_table_type
 *
 * DESCRIPTION: Enumeration that defines the different table types and the
 * associated search methods for that table.
 */
typedef enum ix_e_lkup_table_type
{
    IX_LKUP_TABLE_LPM,          /* Longest Prefix Match table */
    IX_LKUP_TABLE_EXACT,        /* Exact match on the key data */
    IX_LKUP_TABLE_RANGE,        /* Allows range searches, etc. */

    IX_LKUP_TABLE_LAST          /* Last entry in the enumeration */
} ix_lkup_table_type;

/*
 * NAME: ix_lkup_table_conf
 *
 * DESCRIPTION: Structure that is used to pass the configuration parameters
 * when a new table is created.
 */
typedef struct ix_s_lkup_table_conf
{
    /*
     * tableType is the type of table to create.  This defines the search
     * method and the data structures used for the table.
     */
    ix_lkup_table_type tableType;

    /*
     * keySize is the size of the key and mask data in bits.
     * NOTE: An implementation may round this up to a larger size for
     * implementation efficiency, e.g. if the caller specifies the key size
     * is 29 bits, the implementation may round up to 32 bits for the storage
     * if it makes it more convenient.
     */
    int keySize;

    /*
     * maxWeight specifies the acceptable range of the weights. For example,
     * if maxWeight is n, then the valid range of weights is 1 to n.  A value
     * of zero for weight is invalid.
     */
    int maxWeight;

    /*
     * sizeHint is a hint provided to the library on the expected number of
     * elements in the table.  The library can use this to size the data
     * structures appropriately.  In the case of a hardware assisted lookup
     * this will be used as the number of elements to reserve.
     */
    int sizeHint;

    /*
     * dataSize is the size of associated data that should be stored with the
     * keys.  This is the number of bits of data.
     */
    int dataSize;

    /*
     * dataLocation tells the library where it should store the associated
     * data.  The value of IX_LKUP_MEM_SRAM should be used to indicate the 
     * data should be stored in SRAM.  The value IX_LKUP_MEM_DRAM should be
     * used to indicate the data is stored in DRAM.  The value
     * IX_LKUP_MEM_TCAM should be used to indicate the data is stored internal
     * to the hardware device.
     */
    int dataLocation;

    /*
     * dataChannel indicates which of the memory channels specified with the
     * dataLocation should be used.
     */
    int dataChannel;

#ifdef SPLIT_V6_TRIE_TABLE

    /*
     * dataChannel16BitTrie indicates which of the memory channels specified 
     * with the dataLocation should be used for the initial 16-bit trie table
     * if the trie table is to be split.
     */
    int dataChannel16BitTrie;

#endif /* SPLIT_V6_TRIE_TABLE */

} ix_lkup_table_conf;

/*
 * NAME: ix_lkup_tcam_params
 *
 * DESCRIPTION: Data structure that is passed when initializing the TCAM
 * version of the library by calling ix_lkup_tcam_init(). This contains state
 * that is specific to the TCAM implementation.
 */
typedef struct ix_s_lkup_tcam_params
{
    /*
     * qdrChannel indicates which of the QDR SRAM channels the TCAM is
     * attached to.
     */
    int   qdrChannel;

    /*
     * qdrDeviceSelect indicates which device select is used to select the
     * TCAM device on the QDR SRAM interface.  This value will be specific to
     * the implementation of the board.
     */
    int   qdrDeviceSelect;
} ix_lkup_tcam_params;

/*
 * NAME: ix_lkup_sw_init
 *
 * DESCRIPTION: Invoked directly by the client.  Initializes the software
 * lookup management library and returns a handle to use in subsequent
 * operations.
 *
 * @Param:      - IN    arg_LibVersion
 *                      The API version the caller is expecting to use. For
 *                      the current implementation this must be 1. In the
 *                      future, the version number will be used for backwards
 *                      compatibility.
 *
 * @Param:      - OUT   arg_phLkup
 *                      Pointer to a lookup handle. If the call is successful,
 *                      the lookup library allocates a lookup handle with the
 *                      initialized values.
 *
 * Error Codes:
 * IX_LKUP_ERROR_UNSUPPORTED
 *      The version of the library that was requested is not supported.
 *
 * IX_LKUP_ERROR_RESOURCES
 *      Not enough resources are available to complete the request.
 */
ix_error ix_lkup_sw_init(
        int                      arg_LibVersion,
        ix_lkup                 *arg_phLkup);

/*
 * NAME: ix_lkup_tcam_init
 *
 * DESCRIPTION: Invoked directly by the client.  Initializes the TCAM
 * lookup management library and returns a handle to use in subsequent
 * operations. When the call returns, the TCAM should be initialized and ready
 * to accept new table creation operations. This API may be called once for
 * each TCAM search machine being established in the system.
 *
 * @Param:      - IN    arg_LibVersion
 *                      The API version the caller is expecting to use. For
 *                      the current implementation this must be 1. In the
 *                      future, the version number will be used for backwards
 *                      compatibility.
 *
 * @Param:      - IN    arg_pTcamParams
 *                      Pointer to a caller specified parameter block common
 *                      to all TCAM implementations.
 *
 * @Param:      - IN    arg_pVendorParams
 *                      Pointer to a vendor specific parameter block that
 *                      supports vendor specific extensions.
 *
 * @Param:      - OUT   arg_phLkup
 *                      Pointer to a lookup handle. If the call is successful,
 *                      the lookup library allocates a lookup handle with the
 *                      initialized values.
 *
 * Error Codes:
 * IX_LKUP_ERROR_UNSUPPORTED
 *      The version of the library that was requested is not supported.
 *
 * IX_LKUP_ERROR_INVALID
 *      Some of the parameters passed are invalid.
 *
 * IX_LKUP_ERROR_RESOURCES
 *      Not enough resources are available to complete the request.
 */
ix_error ix_lkup_tcam_init(
        int                      arg_LibVersion,
        ix_lkup_tcam_params     *arg_pTcamParams,
        void                    *arg_pVendorParams,
        ix_lkup                 *arg_phLkup);

/*
 * NAME: ix_lkup_create_table
 *
 * DESCRIPTION: Invoked via IX_LKUP_CREATE_TABLE(). Creates a new instance
 * of a search table. If the call is successful then a handle to the new
 * table is returned that can be used for operations on the table.
 *
 * @Param:      - IN    arg_hLkup
 *                      The ix_lkup that was returned when the init operation
 *                      completed.
 *
 * @Param:      - IN    arg_pTableConf
 *                      The table configuration state. This defines the type
 *                      of table as well as the size, search method, etc.
 *
 * @Param:      - IN    arg_pVndrConf
 *                      Pointer to vendor specific table configuration
 *                      information. This is intended to allow vendors to
 *                      expose additional features of their devices. If the
 *                      vendor specific features are not being used, the
 *                      caller should pass NULL pointer for this argument.
 *                      All implementations should accept NULL as a valid
 *                      value for this argument.
 *
 * @Param:      - OUT   arg_phTable
 *                      Pointer to an ix_lkup_table that is filled if the
 *                      operation completes successfully. If the call is 
 *                      successful, the lookup library allocates a table
 *                      handle with the initialized values. The storage for
 *                      the handle is managed by the library.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The handle was invalid, or some of the table configuration information
 *      was invalid or not supported.
 *
 * IX_LKUP_ERROR_RESOURCES
 *      Not enough resources were available to create the requested table.
 */
#define IX_LKUP_CREATE_TABLE(arg_hLkup, arg_pTableConf,                 \
                             arg_VndrConf, arg_phTable)                 \
    (arg_hLkup)->ldataCreateTable(arg_hLkup, arg_pTableConf,            \
                              arg_VndrConf, arg_phTable)
typedef ix_error (*ix_lkup_create_table) (
        ix_lkup                  arg_hLkup,
        ix_lkup_table_conf      *arg_pTableConf,
        void                    *arg_VndrConf,
        ix_lkup_table           *arg_phTable
        );

/*
 * NAME: ix_lkup_destroy_table
 *
 * DESCRIPTION: Invoked via IX_LKUP_DESTROY_TABLE(). Destroys a table that was
 * created earlier with IX_LKUP_CREATE_TABLE(). The application should remove
 * all entries from the table before it is destroyed because the implementation
 * may not be able to free these entries.
 *
 * @Param:      - IN    arg_hLkup
 *                      Handle that was obained through a successful call to
 *                      ix_lkup_sw_init() or ix_lkup_tcam_init().
 *
 * @Param:      - IN    arg_hTable
 *                      Handle for the table to be destroyed.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      One of the handles was invalid.
 */
#define IX_LKUP_DESTROY_TABLE(arg_hLkup, arg_hTable)                    \
    (arg_hLkup)->ldataDestroyTable(arg_hLkup, arg_hTable)
typedef ix_error (*ix_lkup_destroy_table) (
        ix_lkup                  arg_hLkup,
        ix_lkup_table            arg_hTable
        );

/*
 * NAME: ix_lkup_fini
 *
 * DESCRIPTION: Invoked via IX_LKUP_FINI().  Destroys a previously obtained
 * ix_lkup handle. This function is typically called when the application is
 * shutting down and is cleaning up. After successful completion of this call,
 * the handle becomes invalid and subsequent calls to this handle cannot be
 * made. Further, the tables created using this handle should not be used
 * anymore and should be destroyed before this function is called.
 *
 * @Param:      - IN    arg_hLkup
 *                      The handle to destroy. This handle was obtained
 *                      through a successful call to ix_lkup_sw_init() or
 *                      ix_lkup_tcam_init().
 *
 * Error Codes:
 * IX_LKUP_ERROR_IN_USE
 *      The handle has active tables in use
 *
 * IX_LKUP_ERROR_INVALID
 *      The handle was invalid
 */
#define IX_LKUP_FINI(arg_hLkup)                                         \
    (arg_hLkup)->ldataFini(arg_hLkup)
typedef ix_error (*ix_lkup_fini) (
        ix_lkup                  arg_hLkup
        );

/*
 * NAME: ix_lkup_add_entry
 *
 * DESTRIPTION: Invoked via IX_LKUP_ADD_ENTRY().  Adds an entry to a table that
 * was previously created by calling IX_LKUP_CREATE_TABLE().  The lookup
 * management library implementation may return an error if it detects a
 * duplicate but the function is not required to make this validation.
 *
 * @Param:      - IN    arg_hTable
 *                      The handle to the table to update.
 *
 * @Param:      - IN    arg_pKey
 *                      A pointer to the key data.  The length of the key was
 *                      defined when the table was created.
 *
 * @Param:      - IN    arg_pMask
 *                      A pointer to the mask data for this entry. The length
 *                      of the mask was defined when the table was created.
 *
 * @Param:      - IN    arg_Weight
 *                      The weight for this entry. For EXACT match tables this
 *                      must be zero. For weighted tables, this should be in 
 *                      the range defined when creating the table. For LPM the
 *                      weight is used to disambiguate between entries. Lower
 *                      values have higher precedence than higher values of
 *                      weight.
 *
 * @Param:      - IN    arg_pData
 *                      Pointer to the data to be associated with the key.
 *                      This data is opaque to the lookup library and is only
 *                      interpreted by the higher level appliciations.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      Some of the parameters passed are not valid.
 *
 * IX_LKUP_ERROR_EXISTS
 *      An entry with the same key, mask, weight and data is already in the
 *      table.
 *
 * IX_LKUP_ERROR_CONFLICT
 *      An entry with the same key, mask and weight, but different data is
 *      already in the table.
 *
 * IX_LKUP_ERROR_RESOURCES
 *      Not enough resources are available to complete the operation. Either
 *      the maximum table size has been exceeded or other resources could not
 *      be allocated.
 */
#define IX_LKUP_ADD_ENTRY(arg_hTable, arg_pKey, arg_pMask, arg_Weight,  \
                          arg_pData)                                    \
    (arg_hTable)->ltabAddEntry(arg_hTable, arg_pKey, arg_pMask,         \
                               arg_Weight, arg_pData)
typedef ix_error (*ix_lkup_add_entry) (
        ix_lkup_table            arg_hTable,
        ix_uint8                *arg_pKey,
        ix_uint8                *arg_pMask,
        int                      arg_Weight,
        ix_uint8                *arg_pData
        );

/*
 * NAME: ix_lkup_remove_entry
 *
 * DESCRIPTION: Invoked via IX_LKUP_REMOVE_ENTRY(). Removes an entry that was
 * previously added to the table. The entry is identified by the values used
 * to add the entry.
 *
 * @Param:      - IN    arg_hTable
 *                      Handle to the table to update.
 *
 * @Param:      - IN    arg_pKey
 *                      Pointer to the key data of the item to be removed. The
 *                      length of the data was defined when the table was
 *                      created.
 *
 * @Param:      - IN    arg_pMask
 *                      Pointer to the mask data of the entry to be removed.
 *                      The length of the mask data was defined when the table
 *                      was created.
 *
 * @Param:      - IN    arg_Weight
 *                      The weight for the entry to be removed. For EXACT
 *                      match tables this must be zero. For weighted tables,
 *                      this should be in the range defined when creating the
 *                      table. For LPM the weight is used to disambiguate
 *                      between entries.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The table handle was not valid
 *
 * IX_LKUP_ERROR_NOTFOUND
 *      A matching entry was not found
 */
#define IX_LKUP_REMOVE_ENTRY(arg_hTable, arg_pKey, arg_pMask,           \
                             arg_Weight)                                \
    (arg_hTable)->ltabRemoveEntry(arg_hTable, arg_pKey, arg_pMask,      \
                                  arg_Weight)
typedef ix_error (*ix_lkup_remove_entry) (
        ix_lkup_table            arg_hTable,
        ix_uint8                *arg_pKey,
        ix_uint8                *arg_pMask,
        int                      arg_Weight
        );

/*
 * NAME: ix_lkup_update_entry
 *
 * DESCRIPTION: Invoked via IX_LKUP_UPDATE_ENTRY(). Updates the data
 * associated with an element already in the table.
 *
 * @Param:      - IN    arg_hTable
 *                      Handle to the table to update.
 *
 * @Param:      - IN    arg_pKey
 *                      Pointer to the key data of the item to be updated. The
 *                      length of the data was defined when the table was
 *                      created.
 *
 * @Param:      - IN    arg_pMask
 *                      Pointer to the mask data of the entry to be updated.
 *                      The length of the mask data was defined when the table
 *                      was created.
 *
 * @Param:      - IN    arg_Weight
 *                      The weight for the entry to be updated. For EXACT
 *                      match tables this must be zero. For weighted tables,
 *                      this should be in the range defined when creating the
 *                      table. For LPM the weight is used to disambiguate
 *                      between entries.
 *
 * @Param:      - IN    arg_pNewData
 *                      Pointer to the new data to associate with the specific
 *                      key.  The size of the data was specified when the
 *                      table was created.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The table handle was not valid
 *
 * IX_LKUP_ERROR_NOTFOUND
 *      A matching entry was not found so the update was not performed
 */
#define IX_LKUP_UPDATE_ENTRY(arg_hTable, arg_pKey, arg_pMask,           \
                             arg_Weight, arg_pNewData)                  \
    (arg_hTable)->ltabUpdateEntry(arg_hTable, arg_pKey, arg_pMask,      \
                                  arg_Weight, arg_pNewData)
typedef ix_error (*ix_lkup_update_entry) (
        ix_lkup_table            arg_hTable,
        ix_uint8                *arg_pKey,
        ix_uint8                *arg_pMask,
        int                      arg_Weight,
        ix_uint8                *arg_pNewData
        );

/*
 * NAME: ix_lkup_search_table
 *
 * DESCRIPTION: Invoked via IX_LKUP_SEARCH_TABLE(). Searches a specific table
 * and returns the associated data.  The search key is compared with the
 * entries and masks. If multiple entries match then the one with the lowest
 * weight will be returned.
 *
 * @Param:      - IN    arg_hTable
 *                      Handle to the table to search.
 *
 * @Param:      - IN    arg_pKey
 *                      Pointer to the key data used for the search. The
 *                      length of the data was defined when the table was
 *                      created.
 *
 * @Param:      - OUT   arg_pData
 *                      Pointer to a location where the associated data will
 *                      be stored if the search was successful.
 *
 * Error Codes:
 * IX_LKUP_ERROR_NOTFOUND
 *      There was no entry that matched this search request.
 *
 * IX_LKUP_ERROR_INVALID
 *      The handle was invalid
 */
#define IX_LKUP_SEARCH_TABLE(arg_hTable, arg_pKey, arg_pData)           \
    (arg_hTable)->ltabSearchTable(arg_hTable, arg_pKey, arg_pData)
typedef ix_error (*ix_lkup_search_table) (
        ix_lkup_table            arg_hTable,
        ix_uint8                *arg_pKey,
        ix_uint8                *arg_pData
        );

/*
 * NAME: ix_lkup_find_entry
 *
 * DESCRIPTION: Invoked via IX_LKUP_FIND_ENTRY(). Searches a table for an
 * entry that matches the exact key, mask and weigth combination. This is
 * used to locate a specific entry for search types where multiple entries
 * may match a specific search.
 *
 * @Param:      - IN    arg_hTable
 *                      Handle to the table to search.
 *
 * @Param:      - IN    arg_pKey
 *                      Pointer to the key data used for the search. The
 *                      length of the data was defined when the table was
 *                      created.
 *
 * @Param:      - IN    arg_pMask
 *                      Pointer to the mask data to use for the search.
 *                      The length of the mask data was defined when the table
 *                      was created.
 *
 * @Param:      - IN    arg_Weight
 *                      The weight associated with the entry to find.
 *
 * @Param:      - OUT   arg_pData
 *                      Pointer to a location where the associated data will
 *                      be stored if the search was successful. It assumes
 *                      that the caller allocated sufficient space for the
 *                      data associated with the call.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The handle was invalid
 *
 * IX_LKUP_ERROR_NOTFOUND
 *      There was no entry that matched this search request
 */
#define IX_LKUP_FIND_ENTRY(arg_hTable, arg_pKey, arg_pMask, arg_Weight, \
                           arg_pData)                                   \
    (arg_hTable)->ltabFindEntry(arg_hTable, arg_pKey, arg_pMask,        \
                                arg_Weight, arg_pData)
typedef ix_error (*ix_lkup_find_entry) (
        ix_lkup_table            arg_hTable,
        ix_uint8                *arg_pKey,
        ix_uint8                *arg_pMask,
        int                      arg_Weight,
        ix_uint8                *arg_pData
        );

/*
 * NAME: ix_lkup_read_first_entry
 *
 * DESCRIPTION: Invoked via IX_LKUP_READ_FIRST_ENTRY(). This is a debug
 * function that is used to read the items that are stored in the table. This
 * function will get the first item in the table. To continue iterating
 * through all the entries the programmer will call this function once and
 * then repeatedly call IX_LKUP_READ_NEXT_ENTRY until all the entries have
 * been enumerated.
 *
 * @Param:      - IN    arg_hTable
 *                      Handle to the table to search.
 *
 * @Param:      - OUT   arg_pKey
 *                      Pointer where the key data will be stored if the call
 *                      is successful. The caller should make sure there is
 *                      sufficient storage for the key size of the table.
 *
 * @Param:      - OUT   arg_pMask
 *                      Pointer where the mask data will be stored. The caller
 *                      should ensure that this structure is of sufficient
 *                      size to store the mask.
 *
 * @Param:      - OUT   arg_pWeight
 *                      Pointer to where the associated data will be stored if
 *                      the call is successful.
 *
 * @Param:      - OUT   arg_pData
 *                      Pointer to the location where the associated data will
 *                      be stored if the call is successful.
 *
 * @Param:      - OUT   arg_pCookie
 *                      Pointer to an ix_lkup_cookie where the cookie is
 *                      stored. The cookie will be passed to the subsequent
 *                      call to IX_LKUP_READ_NEXT_ENTRY.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      Some of the parameters passed are not valid
 *
 * IX_LKUP_ERROR_NOTFOUND
 *      This error is returned when there are no entries in the table. In this
 *      case none of the data structures to be filled in have valid data.
 */
#define IX_LKUP_READ_FIRST_ENTRY(arg_hTable, arg_pKey, arg_pMask,       \
                                 arg_Weight, arg_pData, arg_pCookie)    \
    (arg_hTable)->ltabReadFirstEntry(arg_hTable, arg_pKey, arg_pMask,   \
                                     arg_Weight, arg_pData, arg_pCookie)
typedef ix_error (*ix_lkup_read_first_entry) (
        ix_lkup_table            arg_hTable,
        ix_uint8                *arg_pKey,
        ix_uint8                *arg_pMask,
        int                     *arg_pWeight,
        ix_uint8                *arg_pData,
        ix_lkup_cookie          *arg_pCookie
        );

/*
 * NAME: ix_lkup_read_next_entry
 *
 * DESCRIPTION: Invoked via IX_LKUP_READ_NEXT_ENTRY(). This is a debug
 * function that is used to read the items that are stored in the table. This
 * function is repeatedly called after IX_LKUP_READ_FIRST_ENTRY to enumerate
 * the contents of the table. Each call will return a cookie that is passed
 * to the next call.
 *
 * @Param:      - IN    arg_hTable
 *                      Handle to the table to search.
 *
 * @Param:      - IN    arg_OldCookie
 *                      The cookie that was returned by the last successful
 *                      call to IX_LKUP_READ_FIRST_ENTRY or
 *                      IX_LKUP_READ_NEXT_ENTRY.
 *
 * @Param:      - OUT   arg_pKey
 *                      Pointer where the key data will be stored if the call
 *                      is successful. The caller should make sure there is
 *                      sufficient storage for the key size of the table.
 *
 * @Param:      - OUT   arg_pMask
 *                      Pointer where the mask data will be stored. The caller
 *                      should ensure that this structure is of sufficient
 *                      size to store the mask.
 *
 * @Param:      - OUT   arg_pWeight
 *                      Pointer where the weight data should be stored.
 *
 * @Param:      - OUT   arg_pData
 *                      Pointer to where the associated data will be stored if
 *                      the call is successful.
 *
 * @Param:      - OUT   arg_pNewCookie
 *                      Pointer to an ix_lkup_cookie where the cookie is
 *                      stored. The cookie will be passed in a subsequent call
 *                      to IX_LKUP_READ_NEXT_ENTRY.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      Some of the parameters passed are not valid
 *
 * IX_LKUP_ERROR_NOTFOUND
 *      This error is returned when there are no more entries in the table.
 */
#define IX_LKUP_READ_NEXT_ENTRY(arg_hTable, arg_pOldCookie, arg_pKey,   \
                                arg_pMask, arg_Weight, arg_pData,       \
                                arg_pNewCookie)                         \
    (arg_hTable)->ltabReadNextEntry(arg_hTable, arg_pOldCookie,         \
                                    arg_pKey, arg_pMask, arg_Weight,    \
                                    arg_pData, arg_pNewCookie)
typedef ix_error (*ix_lkup_read_next_entry) (
        ix_lkup_table            arg_hTable,
        ix_lkup_cookie           arg_OldCookie,
        ix_uint8                *arg_pKey,
        ix_uint8                *arg_pMask,
        int                     *arg_pWeight,
        ix_uint8                *arg_pData,
        ix_lkup_cookie          *arg_pNewCookie
        );

/*
 * NAME: ix_lkup_reset_table
 *
 * DESCRIPTION: Invoked via IX_LKUP_RESET_TABLE(). Clears all the items in the
 * table and resets it to its initial state. Note that the implementation will
 * try to recover all the state, but it may not be possible in all
 * implementations. The callers should try to remove all known entries from
 * the table before calling this function.
 *
 * @Param:      - IN    arg_hTable
 *                      The handle to the table to update.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The table handle was not valid
 */
#define IX_LKUP_RESET_TABLE(arg_hTable)                                 \
    (arg_hTable)->ltabResetTable(arg_hTable)
typedef ix_error (*ix_lkup_reset_table) (
        ix_lkup_table            arg_hTable
        );

/*
 * NAME: ix_lkup_set_property
 *
 * DESCRIPTION: Invoked via IX_LKUP_SET_PROPERTY(). Allows the caller to set
 * special attributes of the table. No attributes are currently defined but
 * this function may be used by vendors to export special features of their
 * devices such as timeout abilities, etc.
 *
 * @Param:      - IN    arg_hTable
 *                      The handle to the table to update.
 *
 * @Param:      - IN    arg_Property
 *                      The value of the property to set. These will be
 *                      defined by the specific implementation.
 *
 * @Param:      - IN    arg_pValue
 *                      Pointer to data associated with the property.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The table handle was invalid.
 *
 * IX_LKUP_ERROR_UNSUPPORTED
 *      The defined property was not supported.
 */
#define IX_LKUP_SET_PROPERTY(arg_hTable, arg_Property, arg_pValue)      \
    (arg_hTable)->ltabSetProperty(arg_hTable, arg_Property, arg_pValue)
typedef ix_error (*ix_lkup_set_property) (
        ix_lkup_table            arg_hTable,
        int                      arg_Property,
        ix_uint8                *arg_pValue
        );

/*
 * NAME: ix_lkup_get_property
 *
 * DESCRIPTION: Invoked via IX_LKUP_GET_PROPERTY(). Allows the caller to get
 * special attributes of the table. No attributes are currently defined but
 * this function may be used by vendors to export special features of their
 * devices such as timeout abilities, etc.
 *
 * @Param:      - IN    arg_hTable
 *                      The handle to the table to update.
 *
 * @Param:      - IN    arg_Property
 *                      The value of the property to get. These will be
 *                      defined by the specific implementation.
 *
 * @Param:      - OUT   arg_pValue
 *                      Pointer where the associated data will be returned
 *                      with a successful call.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The table handle was invalid.
 *
 * IX_LKUP_ERROR_UNSUPPORTED
 *      The defined property was not supported
 */
#define IX_LKUP_GET_PROPERTY(arg_hTable, arg_Property, arg_pValue)      \
    (arg_hTable)->ltabGetProperty(arg_hTable, arg_Property, arg_pValue)
typedef ix_error (*ix_lkup_get_property) (
        ix_lkup_table            arg_hTable,
        int                      arg_Property,
        ix_uint8                *arg_pValue
        );

/*
 * NAME: ix_lkup_get_table_info
 *
 * DESCRIPTION: Invoked via IX_LKUP_GET_TABLE_INFO(). Returns values that need
 * to be used by the microengines.
 *
 * @Param:      - IN    arg_hTable
 *                      The handle to the table.
 *
 * @Param:      - OUT   arg_pTableId
 *                      Pointer to where the table ID will be stored. This
 *                      identifier encodes the information needed by the
 *                      microengines to perform the lookup. The contents and
 *                      format of the id are implementation specific. It is
 *                      the responsibility of the application using this API
 *                      to get this value to the microengines that will be
 *                      calling the search APIs.
 *
 * @Param:      - OUT   arg_pTableData1
 *                      Pointer to where the data1 information will be stored.
 *                      The table data is used by the microengine to find the
 *                      associated data with an element. These values need to
 *                      be passed to some of the microengine calls. The
 *                      contents and format of the data are implementation
 *                      specific.
 *
 * @Param:      - OUT   arg_pTableData2
 *                      Pointer to where the data2 information will be stored.
 *                      The table data is used by the microengine to find the
 *                      associated data with an element. These values need to
 *                      be passed to some of the microengine calls. The
 *                      contents and format of the data are implementation
 *                      specific.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The table handle was not valid
 */
#define IX_LKUP_GET_TABLE_INFO(arg_hTable, arg_pTableId,                \
                               arg_pTableData1, arg_pTableData2)        \
    (arg_hTable)->ltabGetTableInfo(arg_hTable, arg_pTableId,            \
                                   arg_pTableData1, arg_pTableData2)
typedef ix_error (*ix_lkup_get_table_info) (
        ix_lkup_table            arg_hTable,
        ix_uint32               *arg_pTableId,
        ix_uint32               *arg_pTableData1,
        ix_uint32               *arg_pTableData2
        );

/*
 * Error codes for the Lookup Library
 */
enum
{
    IX_LKUP_ERROR_UNSUPPORTED = IX_ERROR_MAKE_GROUP(LKUP),
    IX_LKUP_ERROR_INVALID,
    IX_LKUP_ERROR_RESOURCES,
    IX_LKUP_ERROR_NOTFOUND,
    IX_LKUP_ERROR_EXISTS,
    IX_LKUP_ERROR_CONFLICT,
    IX_LKUP_ERROR_IN_USE,

    /*
     * IX_LKUP_ERROR_INTERNAL is not an official error code defined by the
     * Lookup Library.  In a normal, working system, no entry point should
     * return this error code.  However, as an alternative to using assert()
     * or accessing invalid memory, etc., this code may be useful during
     * development to indicate an internal problem has occurred.
     */
    IX_LKUP_ERROR_INTERNAL,

    IX_LKUP_ERROR_LAST
};

/*****************************************************************************
 * Implementation-specific information.
 * The rest of the file is required for the implementation, and may be
 * interesting for the curious reader.  However, the reader wanting only
 * to use the Lookup Library need not read the rest of this file.
 */

/*
 * NAME: ix_s_lkup
 *
 * DESCRIPTION: This is the data structure that all implementations need to
 * fill out and return when the library is initialized. Only the
 * implementation should care about this data structure. The data structure
 * contains all the function pointers for the operations that can be performed
 * on an ix_lkup.
 */
typedef struct ix_s_lkup
{
    ix_lkup_create_table        ldataCreateTable;
    ix_lkup_destroy_table       ldataDestroyTable;
    ix_lkup_fini                ldataFini;
} ix_s_lkup;

/*
 * NAME: ix_s_lkup_table
 *
 * DESCRIPTION: This is the data structure that all implementations need to
 * fill out and return when a table is created. Only the implementation should
 * care about this data structure. The data structure contains function
 * pointers for the operations that can be performed on an ix_lkup_table.
 */
typedef struct ix_s_lkup_table
{
    ix_lkup_add_entry           ltabAddEntry;
    ix_lkup_remove_entry        ltabRemoveEntry;
    ix_lkup_update_entry        ltabUpdateEntry;
    ix_lkup_search_table        ltabSearchTable;
    ix_lkup_find_entry          ltabFindEntry;
    ix_lkup_read_first_entry    ltabReadFirstEntry;
    ix_lkup_read_next_entry     ltabReadNextEntry;
    ix_lkup_reset_table         ltabResetTable;
    ix_lkup_set_property        ltabSetProperty;
    ix_lkup_get_property        ltabGetProperty;
    ix_lkup_get_table_info      ltabGetTableInfo;
} ix_s_lkup_table;

#if defined(__cplusplus)
} /* extern "C" */
#endif /* __cplusplus */

#endif /*__IX_LKUP_H__*/
