/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      internal_swlpmv4.h
 *
 * = DESCRIPTION
 *      Defines structures, macros, etc., for implementing the software LPM
 *      for IPv4 (a portion of the Lookup Library).  Users of the Lookup
 *      Library do not need to include this header file.
 *
 * ============================================================================
 * $Id: internal_swlpmv4.h,v 1.36 2003/11/13 01:50:14 rranjeet Exp $
 */

#if !defined(__IX_LKUP_SWLPMV4_INTERNAL_H__)
#define __IX_LKUP_SWLPMV4_INTERNAL_H__

#include "ix_cc_error.h"
#include "lkup/swlpmv4.h"
#include "cc/ix_cc_hash_table.h"
#include "cc/ix_cc_index_pool.h"
#include "ix_netmacros.h"
#include "ix_rm.h"

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

#define _IX_LKUP_SWLPMV4_KEY_SIZE       32
#define _IX_LKUP_SWLPMV4_MAX_DATA_SIZE  16

/*
 * LKUP_DEBUG can be enabled to print out an _ix_trie_chain or _ix_local_trie
 * data structure.
 */
/*
#define LKUP_DEBUG
*/

/*
 * NAME: _ix_trie_info
 *
 * DESCRIPTION: Structure describing the contents of a trie.  Contains prefix
 * arrays and a bitmask. Each element of the arrays contains the data for its
 * position; the data is used to back-fill elements when routes are deleted.
 */
typedef struct _ix_s_trie_info
{
    /*
     * prefix3: Array of the 8 possible 3-bit prefixes:
     *  000x, 001x, 010x, 011x, 100x, 101x, 110x, 111x
     */
    ix_uint16 prefix3[8];

    /*
     * prefix2: Array of the four possible 2-bit prefixes:
     *  00xx, 01xx, 10xx, 11xx
     */
    ix_uint16 prefix2[4];

    /*
     * prefix1: Array of the two possible 1-bit prefixes:
     *  0xxx, 1xxx
     */
    ix_uint16 prefix1[2];

    /*
     * prefix4DataExists: Bitmask indicating which of the 16 elements
     * contains a 4-bit prefix.  This is only a bitmask, and not an array
     * of 16-bit entries like the other prefix arrays, because nothing
     * can overlap a 4-bit prefix and hide its data.  Therefore we only need
     * to know that a 4-bit prefix exists, because its data can be obtained
     * elsewhere.
     */
    ix_uint16 prefix4DataExists;
} _ix_trie_info;

/*
 * NAME: _ix_hi256_trie_info
 *
 * DESCRIPTION: Structure describing the contents of the hi256 Table. Contains
 * prefix arrays and a mask array. Each element of the prefix arrays contains
 * the data for its position; the data is used to back-fill elements when
 * routes are deleted.
 */
typedef struct _ix_s_hi256_trie_info
{
    /*
     * See the description of the prefix arrays in _ix_trie_info.
     */
    ix_uint16 prefix7[128];
    ix_uint16 prefix6[64];
    ix_uint16 prefix5[32];
    ix_uint16 prefix4[16];
    ix_uint16 prefix3[8];
    ix_uint16 prefix2[4];
    ix_uint16 prefix1[2];

    /*
     * The mask array indicates the mask of each of the entries in the hi256
     * Table. For example, if mask[0] == 0xFF, we know that hi256TableBase[0]
     * contains an 8-bit route.  If mask[0x20] == 0xE0, we know that
     * hi256TableBase[0x20] has a 3-bit route.
     */
    ix_uint8 mask[256];
} _ix_hi256_trie_info;

typedef struct {
   ix_uint32 value;
} tableElement;

/*
 * NAME: _ix_lkup_swlpmv4
 *
 * DESCRIPTION: Structure containing all the information necessary for the
 * software LPM algorithm to store and retrieve routes.
 */
typedef struct _ix_s_lkup_swlpmv4
{
    /*
     * From time to time, the pointer to this structure needs to be treated
     * as an ix_lkup_sw_table or an ix_lkup_table.  The first element of the
     * _ix_lkup_sw_table is an ix_lkup_table, and the first element of the
     * _ix_lkup_swlpmv4 is an _ix_lkup_sw_table.  Therefore, the single
     * pointer can look like any of the others from time to time.
     */
    struct ix_s_lkup_sw_table hSwTable[1];

    /*
     * numTries: the number of tries this algorithm has available for use
     * (not how many are currently in use).
     */
    ix_uint32            numTries;

    /*
     * memType: SRAM, DRAM, etc.
     */
    int                  memType;

    /*
     * hi64kTableBase: Pointer to the base address of the hi64k table.  This
     * is a virtual pointer into shared memory (SRAM, DRAM, etc.).
     */
    tableElement        *hi64kTableBase;

    /*
     * hi256TableBase: Pointer to the base address of the hi256 table.  This
     * is a virtual pointer into shared memory (SRAM, DRAM, etc.).
     */
    tableElement        *hi256TableBase;

    /*
     * trieTableBase: Pointer to the base address of the trie table.  This
     * is a virtual pointer into shared memory (SRAM, DRAM, etc.).
     */
    tableElement        *trieTableBase;

    /*
     * hTrieIndexes: Index pool used for allocating tries.
     */
    ix_cc_index_pool    hTrieIndexes;

    /*
     * hHash: handle to the hash table used to map tries to their trie info.
     */
    ix_cc_hash_table     hHash;

    /*
     * iteratorKey, iteratorMaskLen: Used in read_next_entry; last reported
     * key and netmask length of that key.
     */
    ix_uint32            iteratorKey;
    ix_uint32            iteratorMaskLen;

    /*
     * iteratorExamineData: Used in read_next_entry; indicates whether to
     * skip digging deeper into tries and instead just report data from the
     * existing trie.
     */
    ix_uint32            iteratorExamineData;

    /*
     * numTrieInfoStructs: The number of trie info structures available to use
     * by the algorithm.
     */
    ix_uint32            numTrieInfoStructs;

    /*
     * hi256TrieInfo: Trie information for the hi256 table.
     */
    _ix_hi256_trie_info  hi256TrieInfo;

    /*
     * pTrieInfoBase: Array of trie info structures for the tries in
     * trieTableBase.
     */
    _ix_trie_info       *pTrieInfoBase;

    /*
     * hTrieInfoIndexes: Index pool used for allocating trie info structures.
     */
    ix_cc_index_pool    hTrieInfoIndexes;
} *_ix_lkup_swlpmv4;

ix_error _ix_lkup_swlpmv4_destroy_table(
        ix_lkup arg_hLkup,
        ix_lkup_table arg_hTable
        );

ix_error _ix_lkup_swlpmv4_add_entry(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int arg_Weight,
        ix_uint8 *arg_pData
        );

ix_error _ix_lkup_swlpmv4_remove_entry(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int arg_Weight
        );

ix_error _ix_lkup_swlpmv4_update_entry(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int arg_Weight,
        ix_uint8 *arg_pNewData
        );

ix_error _ix_lkup_swlpmv4_search_table(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pData
        );

ix_error _ix_lkup_swlpmv4_find_entry(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int arg_Weight,
        ix_uint8 *arg_pData
        );

ix_error _ix_lkup_swlpmv4_read_first_entry(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int *arg_pWeight,
        ix_uint8 *arg_pData,
        ix_lkup_cookie *arg_pCookie
        );

ix_error _ix_lkup_swlpmv4_read_next_entry(
        ix_lkup_table arg_hTable,
        void *arg_OldCookie,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int *arg_pWeight,
        ix_uint8 *arg_pData,
        ix_lkup_cookie *arg_pNewCookie
        );

ix_error _ix_lkup_swlpmv4_reset_table(
        ix_lkup_table arg_hTable
        );

ix_error _ix_lkup_swlpmv4_set_property(
        ix_lkup_table arg_hTable,
        int arg_Property,
        ix_uint8 *arg_pValue
        );

ix_error _ix_lkup_swlpmv4_get_property(
        ix_lkup_table arg_hTable,
        int arg_Property,
        ix_uint8 *arg_pValue
        );

ix_error _ix_lkup_swlpmv4_get_table_info(
        ix_lkup_table arg_hTable,
        ix_uint32 *arg_pTableId,
        ix_uint32 *arg_pTableData1,
        ix_uint32 *arg_pTableData2
        );

/*
 * NAME: _IX_GET_NIBBLE
 * 
 * DESCRIPTION: Gets the nibble from __v covering the bit at position __p.
 */
#define _IX_GET_NIBBLE(__v, __p)                                        \
    (ix_uint8) IX_GET_BIT_FIELD32(__v, (__p) & 0x1C, ((__p) & 0x1C) + 3)

/*
 * NAME: _IX_GET_BYTE
 *
 * DESCRIPTION: Gets the byte from __v covering the bit at position __p.
 */
#define _IX_GET_BYTE(__v, __p)                                          \
    (ix_uint8) IX_GET_BIT_FIELD32(__v, (__p) & 0x18, ((__p) & 0x1C) + 7)

/*
 * NAME: _IX_GET_WORD
 *
 * DESCRIPTION: Gets the word from __v covering the bit at position __p.
 */
#define _IX_GET_WORD(__v, __p)                                          \
    (ix_uint16) IX_GET_BIT_FIELD32(__v, (__p) & 0x10, ((__p) & 0x1C) + 15)

/*
 * NAME: _IX_USE_HI64K_TABLE
 *
 * DESCRIPTION: Given the netmask length, tell whether the route belongs in
 * the hi64k table or the hi256 table.  Return zero for hi256, non-zero for
 * hi64k.
 */
#define _IX_USE_HI64K_TABLE(__l) ((__l) > 16)

/*
 * NAME: _IX_GET_PEER
 *
 * DESCRIPTION: Returns a value a given 'distance' from the provided value,
 * based on __maskNibble. See the following table for examples.
 *
 *  IP  mask  -> peer
 * 0x0  0xF   -> 0x1
 * 0x1  0xF   -> 0x0
 * 0x8  0xE   -> 0xA
 * 0xA  0xE   -> 0x8
 * 0x0  0xC   -> 0x4
 * 0x4  0xC   -> 0x0
 * 0x8  0x8   -> 0x0
 */
#define _IX_GET_PEER(__ipNibble, __maskNibble)                          \
    (((__ipNibble) ^ (~(__maskNibble) + 1)) & 0xF)

/*
 * NAME: _IX_GET_LAST_ENTRY
 *
 * DESCRIPTION: Gets the last value in the range defined by the IP/mask pair.
 * For example, the range defined by 0x0/0xF is {0x0} (a single entry).
 * For 0x0/0xE, the range is {0x0, 0x1}.  For 0x8/0xC, the range is {0x8, 0x9,
 * 0xA, 0xB}.
 */
#define _IX_GET_LAST_ENTRY(__ipNibble, __maskNibble)                    \
    ((__ipNibble) | (~(__maskNibble) & 0xF))

/*
 * NAME: _IX_CHAIN_LEN
 *
 * DESCRIPTION: The maximum length of a trie chain, including a "null"
 * terminator.  The hi256 table can have two trailing tries, and the
 * hi64k table can have four.  The max is four, plus one null, makes five.
 */
#define _IX_CHAIN_LEN 5

typedef enum _ix_e_lkup_direction
{
    _IX_LKUP_DIRECTION_FORWARD,
    _IX_LKUP_DIRECTION_BACKWARD,

    _IX_LKUP_DIRECTION_LAST
} _ix_lkup_direction;

/*
 * A local trie is a copy of a trie (its index in the shared trie table and
 * the contents) used to describe a portion of a route.
 * A trie chain is a local copy of the list of tries that are used to store
 * a route and its associated data.
 */
typedef struct _ix_s_local_trie
{
    /*
     * values: Byte-for-byte copy of the contents of one of the tries in
     * shared memory.  Having the value array here makes it possible to study
     * the contents without a bunch of hits to uncached memory.
     */
    ix_uint32 values[16];

    /*
     * pTrieInfo: Pointer to the trie info structure, if it is known to exist.
     * It could be NULL under two conditions: if we don't know yet whether it
     * exists (in which case trieInfoLookedUp == 0), or if we know it doesn't
     * exist yet (in which case trieInfoLookedUp == 1).
     */
    _ix_trie_info *pTrieInfo;

    /*
     * 'modified' is a bitmask of which values have changed and need to be
     * updated in shared memory.
     */
    ix_uint16 modified;

    /*
     * index: The position of the referenced trie in trieTableBase.
     */
    ix_uint32 index;

    /*
     * The ipPortion and maskPortion are the portion of the IP address and
     * netMask that this trie represents in the parent trie's entries.
     * Note that for the 0th trie, its values of ipPortion and maskPortion
     * refer to either the hi256 table or hi64k table.  If they refer to the 
     * hi64k table, the values have 16 bits of information.  If the refer to
     * the hi256 table, they have 8 bits of information.  All other tries
     * (at indexes 1 - 3) have 4 bits of information.
     */
    ix_uint16 ipPortion;
    ix_uint16 maskPortion;

    /*
     * trieInfoLookedUp indicates whether the value of pTrieInfo is accurate.
     */
    ix_uint8 trieInfoLookedUp;
} _ix_local_trie;

/*
 * NAME: _ix_trie_chain
 *
 * DESCRIPTION: Contains all the necessary information about a single route
 * in shared memory.  We want to minimize the number of times we go to shared
 * memory, so we copy some interesting portions to local memory, examine them,
 * change them, and write them back out to shared memory -- essentially our
 * own private short-term cache.
 */
typedef struct _ix_s_trie_chain
{
    /*
     * Local copy of up to four tries.  If a trie has not yet been allocated
     * for a position in the chain, its index is zero.
     */
    _ix_local_trie localTries[_IX_CHAIN_LEN];

    /*
     * Information about the route being processed.
     */
    ix_uint32 ipAddr;
    ix_uint32 netMask;
    ix_uint32 data;

    /*
     * Additional helpful information.  All of this could be (and is) obtained
     * from examining the previous fields, but is calculated once for easier
     * access.
     */
    ix_uint8 netMaskLen;
} _ix_trie_chain;

ix_error _ix_initialize_chain(
    const ix_uint8 *pKey,
    const ix_uint8 *pMask,
    const ix_uint8 *pData,
    _ix_trie_chain *pChain);

/*
 * trieIndex is host-byte-order index into the trie table.
 * pValues is an array of 16 ix_uint32s where the results are placed.
 */
ix_error _ix_copy_trie_from_shared(
    const _ix_lkup_swlpmv4 hLpmv4,
    ix_uint32 trieIndex,
    ix_uint32 *pValues);

ix_error _ix_get_hi256_entry(
    const _ix_lkup_swlpmv4 hLpmv4,
    const _ix_trie_chain *pChain,
    ix_uint8* pData);

ix_error _ix_add_hi256_entry(
    _ix_lkup_swlpmv4 hLpmv4,
    int update,
    const _ix_trie_chain *pChain);

ix_error _ix_del_hi256_entry(
    _ix_lkup_swlpmv4 hLpmv4,
    const _ix_trie_chain *pChain);

ix_error _ix_get_hi256_8bit_entry(
    const _ix_lkup_swlpmv4 hLpmv4,
    const _ix_trie_chain *pChain,
    ix_uint8 *pData);

ix_error _ix_add_hi256_8bit_entry(
    _ix_lkup_swlpmv4 hLpmv4,
    int update,
    const _ix_trie_chain *pChain);

ix_error _ix_del_hi256_8bit_entry(
    _ix_lkup_swlpmv4 hLpmv4,
    const _ix_trie_chain *pChain);

void _ix_get_best_fit_hi256_data(
    const _ix_lkup_swlpmv4 hLpmv4,
    ix_uint8 entryIndex,
    ix_uint16 *pNewData,
    ix_uint8 *pMask);

ix_uint16* _ix_get_hi256_prefix_entry(
    const _ix_lkup_swlpmv4 hLpmv4,
    ix_uint8 netMaskLen,
    ix_uint8 ipByte);

ix_error _ix_copy_existing_from_base(
    _ix_lkup_swlpmv4 hLpmv4,
    _ix_trie_chain *pChain,
    ix_uint16 *pStartOfNew,
    int *pBitPos);

ix_error _ix_copy_existing_tries(
    _ix_lkup_swlpmv4 hLpmv4,
    ix_uint32 entry,
    ix_uint32 portionMask,
    int *pBitPos,
    _ix_trie_chain *pChain,
    ix_uint16 *pStartOfNew);

ix_error _ix_get_trie_info(
    _ix_lkup_swlpmv4 hLpmv4,
    _ix_local_trie *pLocal
    );

ix_error _ix_make_trie_info(
    _ix_lkup_swlpmv4 hLpmv4,
    _ix_local_trie *pLocal);

ix_error _ix_resolve_trie_chain(
    _ix_lkup_swlpmv4 hLpmv4,
    _ix_trie_chain *pChain);


ix_error _ix_connect_local_tries(
    _ix_local_trie *pLocal,
    ix_uint8 ipNibble,
    ix_uint32 nextTrieIndex);

ix_error _ix_assign_local_data(
    _ix_lkup_swlpmv4 hLpmv4,
    _ix_trie_chain *pChain,
    int allowExisting);

void _ix_write_chain_to_shared(
    _ix_lkup_swlpmv4 hLpmv4,
    const _ix_trie_chain *pChain,
    _ix_lkup_direction direction);

ix_error _ix_lookup_local(
    _ix_lkup_swlpmv4 hLpmv4,
    ix_uint8 ipNibble,
    ix_uint8 maskNibble,
    _ix_local_trie *pLocal,
    ix_uint16 *pData);

ix_error _ix_get_route_chain(
    _ix_lkup_swlpmv4 hLpmv4,
    _ix_trie_chain *pChain,
    ix_uint8 *pChainLen,
    ix_uint16 *pExistingData);

ix_error _ix_cleanup_chain(
    _ix_lkup_swlpmv4 hLpmv4,
    _ix_trie_chain *pChain,
    ix_uint16 *pFirstToFree);

void _ix_dump_chain(
    _ix_lkup_swlpmv4 hLpmv4,
    const _ix_trie_chain *pChain);

void _ix_dump_local(
    const _ix_local_trie *pLocal);

ix_error _ix_free_trie_info(
    _ix_lkup_swlpmv4 hLpmv4,
    _ix_local_trie *pLocal);

ix_uint8 _ix_get_chain_len(
    ix_uint8 netMaskLen);

ix_uint16 _ix_get_best_fit_data(
    _ix_local_trie *pLocal,
    ix_uint8 entryIndex);

ix_uint16 *_ix_get_trie_prefix_entry(
    _ix_local_trie *pLocal,
    ix_uint8 maskNibble,
    ix_uint8 ipNibble);

void _ix_write_trie_to_shared(
    _ix_lkup_swlpmv4 hLpmv4,
    const _ix_local_trie *pLocal);

ix_error _ix_append_trie_to_chain(
    _ix_lkup_swlpmv4 hLpmv4,
    _ix_trie_chain *pChain,
    ix_uint32 entry,
    int bitPos,
    ix_uint32 portionMask,
    ix_uint16 *pTrieChainIndex);


ix_error _ix_get_next_chained_trie(
    const _ix_local_trie *pLocal,
    int continueSearch,
    ix_uint8 *pIpNibble,
    ix_uint32 *pTrieIndex);

/*
 * NAME ix_lkup_swlpmv4_netmask_len
 *
 * DESCRIPTION: Given a network mask (e.g. 0xFFFF0000), return its length
 * (e.g. 16).
 */
ix_error _ix_lkup_swlpmv4_netmask_len(
    ix_uint32 netMask,
    ix_uint8 *pNetMaskLen);

/*
 * Searches the hi256 table's trie information looking for a non-zero value
 * in one of the prefix arrays.  Modify hLpmv4->iteratorKey and
 * hLpmv4->iteratorMaskLen to refer to the route that was found.  If no
 * route was found, change hLpmv4->iteratorKey to 0xFF000000 and
 * hLpmv4->iteratorMaskLen to 7 (i.e. the last possible route this routine
 * could find).
 */
ix_error _ix_search_hi256_table_info(
    _ix_lkup_swlpmv4 hLpmv4,
    int *pFound,
    ix_uint8 *pKey,
    ix_uint8 *pMask,
    ix_uint8 *pData);

/*
 * Creates a chain and searches it looking for the next unreported route.
 * Examines and updates hLpmv4->iteratorKey and hLpmv4->iteratorMaskLen.
 */

ix_error _ix_search_trie_chain(
    _ix_lkup_swlpmv4 hLpmv4,
    ix_uint32 trieIndex, 		
    int fromHi64kTable,
    int continueSearch,
    int *pFound,
    ix_uint8 *pKey,
    ix_uint8 *pMask,
    ix_uint8 *pData);

/**
 * NAME: _ix_search_trie_data
 *
 * DESCRIPTION:
 *      Searches a local trie looking for data.
 *
 * @Param:  IN hLpmv4
 * @Param:  IN pLocal
 * @Param:  IN continueSearch
 *              Indicates whether to start searching from the 0 slot or from
 *              the point after *pIpNibble.
 *
 * @Param:  INOUT pIpNibble
 *              If continueSearch is non-zero, the search continues at the
 *              value past *pIpNibble.  On output, if *pData is non-zero,
 *              *pIpNibble is the relevant nibble of the found route.
 *
 * @Param:  INOUT pMaskNibble
 *              If continueSearch is non-zero, *pMaskNibble is used as a
 *              starting point.  On output, if *pData is non-zero,
 *              *pMaskNibble is the relevant nibble of the found route.
 *              
 * @Param:  OUT pData
 *              If *pData is non-zero, a route was found: *pIpNibble and 
 *              *pMaskNibble define the relevant portion of the route.  If
 *              *pData is zero, no route was found.
 */
ix_error _ix_search_trie_data(
    _ix_lkup_swlpmv4 hLpmv4,
    _ix_local_trie *pLocal,
    int continueSearch,
    ix_uint8 *pIpNibble,
    ix_uint8 *pMaskNibble,
    ix_uint16 *pData);

/*
 * Helper function that updates the iterator values and sets the output
 * parameters according to the provided ipAddr, maskLen and data.
 */
void _ix_assign_search_outputs(
    _ix_lkup_swlpmv4 hLpmv4,
    ix_uint32 ipAddr,
    int maskLen,
    ix_uint16 data,
    ix_uint8 *pKey,
    ix_uint8 *pMask,
    ix_uint8 *pData);

/*
 * Takes in the current ipNibble and maskNibble, and returns the next ip/mask
 * nibbles in the sequence.  The sequence follows these rules:
 * 1) Start with the least-specific netMask (8)
 * 2) For each netMask, report all unique addresses
 * 3) Repeat step 2 for each valid netMask
 *
 * The sequence is as follows:
 * 0/1 8/1
 * 0/2 4/2 8/2 C/2
 * 0/3 2/3 4/3 6/3 8/3 A/3 C/3 E/3
 * 0/4 1/4 2/4 3/4 4/4 5/4 6/4 7/4 8/4 9/4 A/4 B/4 C/4 D/4 E/4 F/4
 *
 * Returns zero if the results are good, or 1 if there is an overflow.
 */
int _ix_get_next_prefix(
    ix_uint8 *pIpNibble,
    ix_uint8 *pMaskNibble);

/*
 * NAME: _IX_TRIE_ROUTE_RATIO
 *
 * DESCRIPTION: The number of tries required to make one route. This is
 * extremely subjective and is obtained by trial-and-error. One test showed
 * a ratio of 1:1.94 (each trie yields 1.94 routes). The macro is used for
 * calculating the number of tries required to support a given number of
 * routes.
 */
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
#define _IX_TRIE_ROUTE_RATIO 1 
#else
#define _IX_TRIE_ROUTE_RATIO 0.5f
#endif
/*
 * NAME: _IX_INFO_PER_TRIE
 *
 * DESCRIPTION: The number of trie_info structures required per trie. This
 * is extremely subjective and is obtained by trial-and-error. One test showed
 * that the number should be roughly .20 (20% of the tries need a trie_info
 * structure.
 */
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
#define _IX_INFO_PER_TRIE    5 
#else
#define _IX_INFO_PER_TRIE    0.2f
#endif

/*
 * NAME: _IX_LKUP_SWLPMV4_NUM_BUCKETS
 *
 * DESCRIPTION: Returns the number of buckets to store a given number of
 * trie indexes.
 *
 * @Param:      - IN    __n
 *                      Number of tries
 *
 * @Returns:
 * Number of buckets to use
 */
#define _IX_LKUP_SWLPMV4_NUM_BUCKETS(__n) ((__n) * 4)

/*
 * NOTE: The following set of macros deal with host byte order values
 * not network!
 */


/* 
 * This constant can be changed based on how many bits are desired for 
 * trie index. Trie index cannot be less than 16. The following set of 
 * macros deal with host byte order values not networkl
 */

#define _IX_TRIE_INDEX_BITS		16

/* 
 * The macros below should not change even if the number of bits for trie
 * changes
 */

#define _IX_TRIE_INDEX_MASK 		((ix_uint32) ((1 << _IX_TRIE_INDEX_BITS) - 1))
#define _IX_TRIE_DATA_MASK		((ix_uint32) (~(_IX_TRIE_INDEX_MASK)))

#define _IX_TRIE_DATA(__t) ((__t) >> _IX_TRIE_INDEX_BITS)
#define _IX_TRIE_INDEX(__t) ((__t) & _IX_TRIE_INDEX_MASK)
#define _IX_TRIE_PACK_LOCAL(__n, __d)                                   \
    ((__d) << _IX_TRIE_INDEX_BITS) | ((__n) & _IX_TRIE_INDEX_MASK)



/*
 * NAME: _IX_SET_HI64K_ENTRY
 *
 * DESCRIPTION: Sets a single 32-bit value in the hi64k table
 *
 * @Param:      - IN    __h
 *                      Handle of the SWLPMv4
 *
 * @Param:      - IN    __k
 *                      32-bit key (IP address)
 *
 * @Param:      - IN    __v
 *                      32-bit value to set the entry to
 *
 * @Returns:
 * none
 */
#define _IX_SET_HI64K_ENTRY(__h, __k, __v)                              \
    (void) IX_RM_MEM_UINT32_WRITE(                                      \
                (__h)->hi64kTableBase + ((ix_uint32)((__k)) >> 16),      \
                IX_HTON32(__v)                                          \
                )                                                       \

/*
 * NAME: _IX_GET_HI64K_ENTRY
 *
 * DESCRIPTION: Gets a single 32-bit value from the hi64k table
 *
 * @Param:      - IN    __h
 *                      Handle of the SWLPMv4
 *
 * @Param:      - IN    __k
 *                      32-bit key (IP address)
 *
 * @Returns:
 * 32-bit value from the hi64kTable
 */
#define _IX_GET_HI64K_ENTRY(__h, __k)                                   \
    IX_NTOH32(                                                          \
            IX_RM_MEM_UINT32_READ(                                      \
                (__h)->hi64kTableBase + ((ix_uint32)((__k)) >> 16)      \
                )                                                       \
            )

/*
 * NAME: _IX_SET_HI256_ENTRY
 *
 * DESCRIPTION: Sets a single 32-bit value in the hi256 table
 *
 * @Param:      - IN    __h
 *                      Handle of the SWLPMv4
 *
 * @Param:      - IN    __k
 *                      8-bit index of the location to write
 *
 * @Param:      - IN    __v
 *                      32-bit value to set the entry to
 *
 * @Returns:
 * none
 */
#define _IX_SET_HI256_ENTRY(__h, __k, __v)                              \
    (void) IX_RM_MEM_UINT32_WRITE(                                      \
            (__h)->hi256TableBase + (((ix_uint32) (__k)) >> 24),        \
            IX_HTON32(__v))

/*
 * NAME: _IX_GET_HI256_ENTRY
 *
 * DESCRIPTION: Gets a single 32-bit value from the hi256 table
 *
 * @Param:      - IN    __h
 *                      Handle of the SWLPMv4
 *
 * @Param:      - IN    __k
 *                      32-bit key (IP address)
 *
 * @Returns:
 * 32-bit value from the hi256Table
 */
#define _IX_GET_HI256_ENTRY(__h, __k)                                   \
    IX_NTOH32(                                                          \
            IX_RM_MEM_UINT32_READ(                                      \
                (__h)->hi256TableBase + (((ix_uint32) (__k)) >> 24)     \
                )                                                       \
            )

/*
 * NAME: _IX_READ_HI256_VALUE
 *
 * DESCRIPTION: Gets a single 32-bit value from the hi256 table
 *
 * @Param:      - IN    __h
 *                      Handle of the SWLPMv4
 *
 * @Param:      - IN    __k
 *                      8-bit key - 1st octet of the IPv4 address
 *
 * @Returns:
 * 32-bit value from the hi256Table
 */
#define _IX_READ_HI256_VALUE(__h, __k)                                   \
    IX_NTOH32(                                                          \
            IX_RM_MEM_UINT32_READ(                                      \
                (__h)->hi256TableBase + ((ix_uint32) (__k))             \
                )                                                       \
            )



/*
 * NAME: _IX_SET_HI256_ENTRY_DATA
 *
 * DESCRIPTION: Sets a single 16-bit data value in the hi256 table
 *
 * @Param:      - IN    __h
 *                      Handle of the SWLPMv4
 *
 * @Param:      - IN    __i
 *                      8-bit index of the location to write into
 *
 * @Param:      - IN    __d
 *                      16-bit Data value to write
 *
 * Returns:
 * None
 */

#define _IX_SET_HI256_ENTRY_DATA(__h, __i, __d)                     \
	 (void) IX_RM_MEM_UINT32_WRITE(                                      \
            (__h)->hi256TableBase + ((ix_uint32) (__i)),        \
   	   		IX_HTON32((((_IX_READ_HI256_VALUE(__h, __i)) & ~_IX_TRIE_DATA_MASK) | (__d << _IX_TRIE_INDEX_BITS)))) 
	 



/*
 * NAME: _IX_SET_HI256_ENTRY_INDEX
 *
 * DESCRIPTION: Sets a single 16-bit next trie index value in the hi256 table
 *
 * @Param:      - IN    __h
 *                      Handle of the SWLPMv4
 *
 * @Param:      - IN    __i
 *                      8-bit index of the location to write into
 *
 * @Param:      - IN    __n
 *                      16-bit Next trie index value to write
 *
 * Returns:
 * None
 */


#define _IX_SET_HI256_ENTRY_INDEX(__h, __i, __n)                        \
    (void) IX_RM_MEM_UINT32_WRITE(                                      \
                &(__h)->hi256TableBase[__i].value,                   \
    IX_HTON32((_IX_READ_HI256_VALUE(__h, __i) & ~_IX_TRIE_INDEX_MASK) | (__n & _IX_TRIE_INDEX_MASK)))



/*
 * NAME: _IX_GET_HI256_ENTRY_DATA
 *
 * DESCRIPTION: Gets a single 16-bit data value from the hi256 table
 *
 * @Param:      - IN    __h
 *                      Handle of the SWLPMv4
 *
 * @Param:      - IN    __i
 *                      8-bit index of the location to write into
 *
 * Returns:
 * 16-bit Data value from the location in shared memory
 */

#define _IX_GET_HI256_ENTRY_DATA(__h, __i)                              \
   (_IX_READ_HI256_VALUE(__h, __i) >> _IX_TRIE_INDEX_BITS) 

/*
 * NAME: _IX_GET_HI256_ENTRY_INDEX
 *
 * DESCRIPTION: Gets a single 16-bit next trie index value from the hi256
 * table
 *
 * @Param:      - IN    __h
 *                      Handle of the SWLPMv4
 *
 * @Param:      - IN    __i
 *                      8-bit index of the location to write into
 *
 * Returns:
 * 16-bit next trie index value from the location in shared memory
 */

#define _IX_GET_HI256_ENTRY_INDEX(__h, __i)                             \
    (_IX_READ_HI256_VALUE(__h, __i) & _IX_TRIE_INDEX_MASK)

/*
 * NAME: _IX_SET_TRIE_ENTRY
 *
 * DESCRIPTION: Sets a single 32-bit data / next trie index pair into a trie.
 *
 * @Param:      - IN    __h
 *                      Handle of the SWLPMv4
 *
 * @Param:      - IN    __t
 *                      Index of the trie in trieTableBase to set into
 *
 * @Param:      - IN    __o
 *                      Offset into the trie (number from 0 to 15)
 *
 * @Param:      - IN    __v
 *                      32-bit data / next trie index pair to write
 *
 * @Returns:
 * 32-bit data and next trie pair
 */
#define _IX_SET_TRIE_ENTRY(__h, __t, __o, __v)                          \
    (void) IX_RM_MEM_UINT32_WRITE(                                      \
            (__h)->trieTableBase + (__t) * 16 + (__o),                  \
            IX_HTON32(__v))

/*
 * NAME: _IX_GET_TRIE_ENTRY
 *
 * DESCRIPTION: Gets a single 32-bit data / next trie index pair from a trie.
 *
 * @Param:      - IN    __h
 *                      Handle of the SWLPMv4
 *
 * @Param:      - IN    __t
 *                      Index of the trie in trieTableBase to read from
 *
 * @Param:      - IN    __o
 *                      Offset into the trie (number from 0 to 15)
 *
 * @Returns:
 * 32-bit data and next trie pair
 */
#define _IX_GET_TRIE_ENTRY(__h, __t, __o)                               \
    IX_NTOH32(                                                          \
        IX_RM_MEM_UINT32_READ(                                          \
            (__h)->trieTableBase + (__t) * 16 + (__o)))

/*
 * NAME: _IX_ROUTE_EXISTS
 *
 * DESCRIPTION: Helper macro for generating an ix_error with the code
 * IX_LKUP_ERROR_EXISTS or IX_LKUP_ERROR_CONFLICT.
 *
 * @Param:      - IN    __e
 *                      Part of the error code's name: 'EXISTS' or 'CONFLICT'
 *
 * @Param:      - IN    __c
 *                      Pointer to an _ix_trie_chain structure with the
 *                      ipAddr and netMaskLen fields set
 *
 * @Returns:
 * Completely formed ix_error with interesting information about the route
 */
#define _IX_ROUTE_EXISTS(__e, __c)                                      \
    IX_ERROR_WARNING(IX_LKUP_ERROR_##__e,                               \
        ("The route for %d.%d.%d.%d/%d->0x%04lX already exists",        \
         ((__c)->ipAddr >> 24) & 0xFF,                                  \
         ((__c)->ipAddr >> 16) & 0xFF,                                  \
         ((__c)->ipAddr >>  8) & 0xFF,                                  \
         ((__c)->ipAddr      ) & 0xFF,                                  \
         (__c)->netMaskLen, (__c)->data))

/*
 * NAME: _IX_ROUTE_NOTFOUND
 *
 * DESCRIPTION: Helper macro for generating an ix_error with the code
 * IX_LKUP_ERROR_NOTFOUND
 *
 * @Param:      - IN    __c
 *                      Pointer to an _ix_trie_chain structure with the
 *                      ipAddr and netMaskLen fields set
 *
 * @Returns:
 * Completely formed ix_error with interesting information about the route
 */
#define _IX_ROUTE_NOTFOUND(__c)                                         \
    IX_ERROR_WARNING(IX_LKUP_ERROR_NOTFOUND,                            \
        ("The route for %d.%d.%d.%d/%d does not exist",                 \
         ((__c)->ipAddr >> 24) & 0xFF,                                  \
         ((__c)->ipAddr >> 16) & 0xFF,                                  \
         ((__c)->ipAddr >>  8) & 0xFF,                                  \
         ((__c)->ipAddr      ) & 0xFF,                                  \
         (__c)->netMaskLen))

#define TODO \
    IX_ERROR_PANIC(IX_LKUP_ERROR_INTERNAL, \
        ("Function is not currently implemented"))

#if defined(__cplusplus)
} /* extern "C" */
#endif /* __cplusplus */

#endif /*__IX_LKUP_SWLPMV4_INTERNAL_H__*/
