/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      internal_swlpmv6.h
 *
 * = DESCRIPTION
 *      Defines structures, macros, etc., for implementing the software LPM
 *      for IPv6 (a portion of the Lookup Library).  Users of the Lookup
 *      Library do not need to include this header file.
 *
 * ============================================================================
 */

#if !defined(__IX_LKUP_SWLPMV6_INTERNAL_H__)
#define __IX_LKUP_SWLPMV6_INTERNAL_H__


#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

#include "ix_cc_error.h"
#include "ix_ossl.h"
#include "lkup/swlpmv6.h"
#include "cc/ix_cc_hash_table.h"
#include "cc/ix_cc_index_pool.h"
#include "ix_netmacros.h"
#include "ix_rm.h"

#define _IX_LKUP_SWLPMV6_KEY_SIZE       128
#define _IX_LKUP_SWLPMV6_MAX_DATA_SIZE  16

/*
 * LKUP_DEBUG can be enabled to print out an _ix_trie_chain or _ix_local_trie
 * data structure.
 */
/*
#define LKUP_DEBUG
*/

#if !(defined FALSE)
#define FALSE 0
#endif /* end !(defined FALSE) */

#if !(defined TRUE)
#define TRUE 1
#endif /* end !(defined TRUE) */


/* Operations */
typedef enum _ix_e_swlpmv6_op {

    swlpmv6_add = 1,
    swlpmv6_delete,
    swlpmv6_update,
    swlpmv6_lookup,

    swlpmv6_maximum

} ix_swlpmv6_op;

typedef struct {
    ix_uint16 data;
    ix_uint16 nextTrie;
} tableElement;

/*
 * An IPv6 Address is 128-bits long and is contained in an array of
 * 4 ix_uint32's. For instance the IPv6 address 3ffe::1234 would be
 * expanded out, using the usual rules, and would look like the following
 * in array form. This is in Host (Big Endian) Order.
 *
 *     IpAddrArray[0] = 0x3ffe0000
 *     IpAddrArray[1] = 0x0
 *     IpAddrArray[2] = 0x0
 *     IpAddrArray[3] = 0x1234
 */

/* Maximum length (in bits) of an IPv6 address */
#define MAX_IPV6_ADDR_LENGTH   128

/* Nodes (in the 1-bit trie) can have "NULL_DATA" as the data element
 * stored. However, if the data part of a node is NULL_DATA, a route
 * cannot end in that node.
 */
#define NULL_DATA       0

/* Since we perform a 16-8-8... lookup in the route table, the maximum
 * number of 8-bit tries in a "chain" is (128 - 16)/8 = 14
 */
#define MAX_8BIT_TRIES_PER_ROUTE    14

/* In this implementation we use a 1-bit trie to keep track of routes 
 * in the route table in SRAM/SDRAM. The 1-bit trie is kept in sync with 
 * the route table when routes are added, deleted and updated.
 */

/*
 * The structure of each node in the 1-bit trie table.
 * Each node has a data element which can contain either "NULL_DATA"
 * or a "real" data element (>0). Each node also contains a left and right
 * child. In addition, if the node corresponds to a bit on a BOUNDARY the
 * node stores the block index of the trie-block. This is very useful
 * when a route is deleted. It helps the deletion routine quickly locate
 * the block in the route table and make a quick decision if a trie-block
 * can been freed.
 */

typedef struct _ix_s_1bit_trie_node
{
    ix_uint16 data;
    ix_uint16 block_id;    /* only if it is a boundary bit */
    struct _ix_s_1bit_trie_node *left;
    struct _ix_s_1bit_trie_node *right;
 
    /* This is used during dump route when we don't want to report  
     * the same route twice 
     */
    ix_uint8 visited;

} ix_1bit_trie_node;

/*
 * NAME: _ix_lkup_swlpmv6
 *
 * DESCRIPTION: Structure containing all the information necessary for the
 * software LPM algorithm to store and retrieve routes.
 */
typedef struct _ix_s_lkup_swlpmv6
{
    /*
     * From time to time, the pointer to this structure needs to be treated
     * as an ix_lkup_sw_table or an ix_lkup_table.  The first element of the
     * _ix_lkup_sw_table is an ix_lkup_table, and the first element of the
     * _ix_lkup_swlpmv6 is an _ix_lkup_sw_table.  Therefore, the single
     * pointer can look like any of the others from time to time.
     */
    struct ix_s_lkup_sw_table hSwTable[1];

    /*
     * numTries: the number of tries this algorithm has available for use
     * (not how many are currently in use).
     */
    ix_uint32            numTries;

    /*
     * memType: SRAM, DRAM, etc.
     */
    int                  memType;

    /*
     * hi64kTableBase: Pointer to the base address of the hi64k table.  This
     * is a virtual pointer into shared memory (SRAM, DRAM, etc.).
     */
    tableElement        *hi64kTableBase;

    /*
     * trieTableBase: Pointer to the base address of the trie table.  This
     * is a virtual pointer into shared memory (SRAM, DRAM, etc.).
     */
    tableElement        *trieTableBase;

    /*
     * hTrieIndexes: Index pool used for allocating tries.
     */
    ix_cc_index_pool    hTrieIndexes;

    /*
     * iteratorKey, iteratorMaskLen: Used in read_next_entry; last reported
     * key and netmask length of that key.
     */
    ix_uint32            iteratorKey;
    ix_uint32            iteratorMaskLen;

    /*
     * iteratorExamineData: Used in read_next_entry; indicates whether to
     * skip digging deeper into tries and instead just report data from the
     * existing trie.
     */
    ix_uint32            iteratorExamineData;

    /* 
     * gRoot: Root of the 1-bit trie table 
     */

    ix_1bit_trie_node    *gRoot;

} *_ix_lkup_swlpmv6;

ix_error _ix_lkup_swlpmv6_destroy_table(
        ix_lkup arg_hLkup,
        ix_lkup_table arg_hTable
        );

ix_error _ix_lkup_swlpmv6_add_entry(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int arg_Weight,
        ix_uint8 *arg_pData
        );

ix_error _ix_lkup_swlpmv6_remove_entry(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int arg_Weight
        );

ix_error _ix_lkup_swlpmv6_update_entry(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int arg_Weight,
        ix_uint8 *arg_pNewData
        );

ix_error _ix_lkup_swlpmv6_search_table(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pData
        );

ix_error _ix_lkup_swlpmv6_find_entry(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int arg_Weight,
        ix_uint8 *arg_pData
        );

ix_error _ix_lkup_swlpmv6_read_first_entry(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int *arg_pWeight,
        ix_uint8 *arg_pData,
        ix_lkup_cookie *arg_pCookie
        );

ix_error _ix_lkup_swlpmv6_read_next_entry(
        ix_lkup_table arg_hTable,
        void *arg_OldCookie,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int *arg_pWeight,
        ix_uint8 *arg_pData,
        ix_lkup_cookie *arg_pNewCookie
        );

ix_error _ix_lkup_swlpmv6_reset_table(
        ix_lkup_table arg_hTable
        );

ix_error _ix_lkup_swlpmv6_set_property(
        ix_lkup_table arg_hTable,
        int arg_Property,
        ix_uint8 *arg_pValue
        );

ix_error _ix_lkup_swlpmv6_get_property(
        ix_lkup_table arg_hTable,
        int arg_Property,
        ix_uint8 *arg_pValue
        );

ix_error _ix_lkup_swlpmv6_get_table_info(
        ix_lkup_table arg_hTable,
        ix_uint32 *arg_pTableId,
        ix_uint32 *arg_pTableData1,
        ix_uint32 *arg_pTableData2
        );

typedef enum _ix_e_lkup_direction
{
    _IX_LKUP_DIRECTION_FORWARD,
    _IX_LKUP_DIRECTION_BACKWARD,

    _IX_LKUP_DIRECTION_LAST
} _ix_lkup_direction;

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
typedef struct
{
        _ix_lkup_swlpmv6 hLpmv6;
        ix_1bit_trie_node *pNode;
        ix_uint32 *pIpv6Addr;
        ix_uint8 *pPrefixLength;
        ix_uint16 bitPos;
        ix_uint16 *pData;
        ix_error err; 
}ix_store_element;
/*
 * NAME: __swlpmv6_dfs_v
 *
 * DESCRIPTION: This routine performs a depth first search of the 1-bit
 * trie table,using global structure  and this function is called
 * by __swlpmv6_dfs
 *
 * @Param:      - IN   none 
 *
 * @Returns:
 */
ix_error __swlpmv6_dfs_v(void);
#endif


/*
 * NAME: _IX_TRIE_ROUTE_RATIO
 *
 * DESCRIPTION: The number of tries required to make one route. This is
 * extremely subjective and is obtained by trial-and-error. One test showed
 * a ratio of 1:1.94 (each trie yields 1.94 routes). The macro is used for
 * calculating the number of tries required to support a given number of
 * routes.
 */
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
/*
 * In linux, no floating point support, so IX_TRIE_ROUTE_RATIO is the
 * number of bits to shift, rather than a strict multiple, effectively
 * resulting in a numTries = numRoutes / 2^IX_TRIE_ROUTE_RATIO.
 */
#define _IX_TRIE_ROUTE_RATIO 1 
#else
#define _IX_TRIE_ROUTE_RATIO 0.5f
#endif

/*
 * NAME: _IX_INFO_PER_TRIE
 *
 * DESCRIPTION: The number of trie_info structures required per trie. This
 * is extremely subjective and is obtained by trial-and-error. One test showed
 * that the number should be roughly .20 (20% of the tries need a trie_info
 * structure.
 * THIS VALUE IS CURRENTLY UNUSED IN THE IPV6 IMPLEMENTATION.
 */
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
#define _IX_INFO_PER_TRIE    5 
/*
 * In linux, no floating point support, so IX_TRIE_ROUTE_RATIO is the
 * number of bits to shift, rather than a strict multiple, effectively
 * resulting in a numTries = numRoutes / 2^IX_INFO_PER_TRIE.
 */
#else
#define _IX_INFO_PER_TRIE    0.2f
#endif
/*
 * NAME: _IX_LKUP_SWLPMV6_NUM_BUCKETS
 *
 * DESCRIPTION: Returns the number of buckets to store a given number of
 * trie indexes.
 *
 * @Param:      - IN    __n
 *                      Number of tries
 *
 * @Returns:
 * Number of buckets to use
 */
#define _IX_LKUP_SWLPMV6_NUM_BUCKETS(__n) ((__n) * 4)

/*
 * Reading and writing Route Table Entries
 * Each entry is 32 bits long, the format of an entry is as follows
 * -----------------------------------------------
 * |  (16 bit) NHID       |   (16 bit) Next trie |
 * -----------------------------------------------
 */

#define SIZEOF_TRIE_ENTRY   sizeof(ix_uint32)
#define SIZEOF_16BIT_TRIE   ( SIZEOF_TRIE_ENTRY * (1 << 16) )
#define SIZEOF_8BIT_TRIE    ( SIZEOF_TRIE_ENTRY * (1 << 8) )

/*
 * NOTE: The following set of macros deal with host byte order values
 * not network!
 */

#define DATA_SHIFT          16

#define _IX_TRIE_INDEX_MASK 0x0000FFFFUL
#define _IX_TRIE_DATA_MASK  0xFFFF0000UL

#define _IX_GET_DATA(__e) ((__e) >> 16)
#define _IX_GET_NEXT_TRIE(__e) ((__e) & _IX_TRIE_INDEX_MASK)
#define _IX_TRIE_PACK_LOCAL(__n, __d)                      \
    ((__d) << 16) | ((__n) & _IX_TRIE_INDEX_MASK)

/*
 * NAME: _IX_SET_ENTRY_16
 *
 * DESCRIPTION: Sets a single 32-bit value in the hi64k table
 *
 * @Param:      - IN    __h
 *                      Handle of the SWLPMv6
 *
 * @Param:      - IN    __i
 *                      16-bit index (Top 16 bits of the IPv6 address)
 *
 * @Param:      - IN    __v
 *                      32-bit value to set the entry to
 *
 * @Returns:
 * none
 */
#define _IX_SET_ENTRY_16(__h, __i, __v)                                     \
    (void) IX_RM_MEM_UINT32_WRITE(                                          \
                (ix_uint32 *)((__h)->hi64kTableBase) + (__i), IX_HTON32(__v))

/*
 * NAME: _IX_GET_ENTRY_16
 *
 * DESCRIPTION: Gets a single 32-bit value from the hi64k table
 *
 * @Param:      - IN    __h
 *                      Handle of the SWLPMv6
 *
 * @Param:      - IN    __i
 *                      16-bit index (Top 16 bits of the IPv6 address)
 *
 * @Returns:
 * 32-bit value from the hi64kTable
 */
#define _IX_GET_ENTRY_16(__h, __i)                                \
    IX_NTOH32(                                                    \
            IX_RM_MEM_UINT32_READ(                                \
                (ix_uint32 *)((__h)->hi64kTableBase) + (__i)))

/*
 * NAME: _IX_SET_ENTRY_8
 *
 * DESCRIPTION: Sets a single 32-bit value in the hi256 table
 *
 * @Param:      - IN    __h
 *                      Handle of the SWLPMv6
 *
 * @Param:      - IN    __b
 *                      Block index of the 8-bit trie
 *
 * @Param:      - IN    __o
 *                      8-bit offset of the location to write
 *
 * @Param:      - IN    __v
 *                      32-bit value to set the entry to
 *
 * @Returns:
 * none
 */
#define _IX_SET_ENTRY_8(__h, __b, __o, __v)                            \
    (void) IX_RM_MEM_UINT32_WRITE(                                     \
            (ix_uint32 *)((__h)->trieTableBase) + (((__b) << 8) + (__o)), \
                                      IX_HTON32(__v))

/*
 * NAME: _IX_GET_ENTRY_8
 *
 * DESCRIPTION: Gets a single 32-bit value from the hi256 table
 *
 * @Param:      - IN    __h
 *                      Handle of the SWLPMv6
 *
 * @Param:      - IN    __b
 *                      Block index of the 8-bit trie
 *
 * @Param:      - IN    __o
 *                      8-bit offset of the location
 *
 * @Returns:
 * 32-bit value from the hi256Table
 */
#define _IX_GET_ENTRY_8(__h, __b, __o)                                    \
    IX_NTOH32(                                                            \
            IX_RM_MEM_UINT32_READ(                                        \
                (ix_uint32 *)((__h)->trieTableBase) + (((__b) << 8) + (__o))))

/*
 * NAME: _IX_ROUTE_EXISTS
 *
 * DESCRIPTION: Helper macro for generating an ix_error with the code
 * IX_LKUP_ERROR_EXISTS or IX_LKUP_ERROR_CONFLICT.
 *
 * @Param:      - IN    __e
 *                      Part of the error code's name: 'EXISTS' or 'CONFLICT'
 *
 * @Param:      - IN    __a
 *                      The IPv6 address.
 *            
 * @Param:      - IN    __d
 *                      The data for this route.
 *            
 * @Param:      - IN    __l
 *                      The mask length for this route.
 *
 * @Returns:
 * Completely formed ix_error with interesting information about the route
 */
#define _IX_IPV6_ROUTE_EXISTS(__e, __a, __d, __l)                       \
    IX_ERROR_WARNING(IX_LKUP_ERROR_##__e,                               \
        ("The route for 0x%x:0x%x:0x%x:0x%x/%d->0x%lx already exists",  \
         (__a)[0], (__a)[1], (__a)[2], (__a)[3],                        \
         __l, __d))


/*
 * NAME: _IX_ROUTE_NOTFOUND
 *
 * DESCRIPTION: Helper macro for generating an ix_error with the code
 * IX_LKUP_ERROR_NOTFOUND
 *
 * @Param:      - IN    __a
 *                      The IPv6 address.
 * 
 * @Param:      - IN    __l
 *                      The mask length for this route.
 *
 * @Returns:
 * Completely formed ix_error with interesting information about the route
 */
#define _IX_IPV6_ROUTE_NOTFOUND(__a, __l)                              \
    IX_ERROR_WARNING(IX_LKUP_ERROR_NOTFOUND,                           \
        ("The route for 0x%x:0x%x:0x%x:0x%x/%d does not exist",        \
        (__a[0]), (__a[1]), (__a[2]), (__a[3]),                        \
        __l))

#define TODO \
    IX_ERROR_PANIC(IX_LKUP_ERROR_INTERNAL, \
        ("Function is not currently implemented"))


/* 
 * Miscellaneous Macros to perform various tasks
 */

/*
 * NAME: LEGAL_BIT_POSITION
 *
 * DESCRIPTION: Is this bit position legal. The total number of bits
 * in an IPv6 address is 128, so, the bit position has to be in the 
 * range 0 - 127.
 *
 * @Param:      - IN    _bitPos
 *                      The bit position.
 *
 * @Returns:
 * TRUE: if the bit position is in the valid range
 * FALSE : Otherwise.
 */

static
__inline__ int
LEGAL_BIT_POSITION(int _bitPos)
{
    if (((_bitPos) < 0) || (_bitPos) > MAX_IPV6_ADDR_LENGTH - 1)
    {
        return FALSE;
    }

    return TRUE;
}

/*
 * NAME: GENERATE_MASK
 *
 * DESCRIPTION: Generate a mask for a given mask length and base.
 * The value of base is either 8 or 16. 
 *
 * @Param:      - IN    _maskLen
 *                      The length of the mask required.
 *
 * @Param:      - IN    _base
 *                      The base 8 or 16.
 *
 * @Param:      - OUT   _mask
 *                      The generated mask.
 *
 * @Returns:
 * none.
 */

#define GENERATE_MASK(_maskLen, _base, _mask) \
{ \
    _mask = ~((1 << ((_base) - (_maskLen))) - 1); \
}

/*
 * NAME: CLEAR_BIT_AT_POSITION
 *
 * DESCRIPTION: Clears (sets to zero) a bit in an IPv6 address.
 *
 * @Param:      - IN    ipAddr
 *                      The IPv6 address
 *
 * @Param:      - IN    bitPos
 *                      bit position of the bit that needs to be cleared
 *
 * @Returns:
 * none.
 */
static
__inline__ void
CLEAR_BIT_AT_POSITION(ix_uint32 *ipAddr, ix_uint32 bitPos)
{
    ix_uint32 index, shift, mask;

    index = bitPos / 32;
    shift = bitPos % 32;

    mask = (0xFFFFFFFF ^ (1 << (32 - shift - 1)));

    ipAddr[index] = ipAddr[index] & mask;
}

/*
 * NAME: SET_BIT_AT_POSITION
 *
 * DESCRIPTION: Sets (sets to one) a bit in an IPv6 address.
 *
 * @Param:      - IN    ipAddr
 *                      The IPv6 address
 *
 * @Param:      - IN    bitPos
 *                      bit position of the bit that needs to be cleared
 *
 * @Returns:
 * none.
 */
static
__inline__ void
SET_BIT_AT_POSITION(ix_uint32 *ipAddr, ix_uint32 bitPos)
{
    ix_uint32 index, shift, mask;

    index = bitPos / 32;
    shift = bitPos % 32;

    mask = (1 << (32 - shift - 1));

    ipAddr[index] = ipAddr[index] | mask;
}


/*
 * NAME: CLEAR_BIT_AT_POSITION_16
 *
 * DESCRIPTION: Clears (sets to zero) a bit in an 16-bit chunk of an IPv6
 * address.
 *
 * @Param:      - IN    _ipAddr
 *                      A 16-bit portion of an IPv6 address
 *
 * @Param:      - IN    _bitPos
 *                      bit position of the bit that needs to be cleared
 *
 * @Returns:
 * none.
 */
#define CLEAR_BIT_AT_POSITION_16(_ipAddr, _bitPos) \
{ \
    ix_uint16 _bitMask = (0xFFFF ^ (1 << (16 - (_bitPos) - 1))); \
    _ipAddr = _ipAddr & _bitMask; \
}

/*
 * NAME: CLEAR_BIT_AT_POSITION_8
 *
 * DESCRIPTION: Clears (sets to zero) a bit in an 8-bit chunk of an IPv6
 * address.
 *
 * @Param:      - IN    _ipAddr
 *                      A 8-bit portion of an IPv6 address
 *
 * @Param:      - IN    _bitPos
 *                      bit position of the bit that needs to be cleared
 *
 * @Returns:
 * none.
 */
#define CLEAR_BIT_AT_POSITION_8(_ipAddr, _bitPos) \
{ \
    ix_uint8 _bitMask = (0xFF ^ (1 << (8 - ((_bitPos) % 8) - 1))); \
    _ipAddr = _ipAddr & _bitMask; \
}

/*
 * NAME: SET_BIT_POSITION_16
 *
 * DESCRIPTION: Sets (sets to one) a bit in an 16-bit chunk of an IPv6
 * address.
 *
 * @Param:      - IN    _ipAddr
 *                      A 16-bit portion of an IPv6 address
 *
 * @Param:      - IN    _bitPos
 *                      bit position of the bit that needs to be cleared
 *
 * @Returns:
 * none.
 */
#define SET_BIT_AT_POSITION_16(_ipAddr, _bitPos) \
{ \
    ix_uint16 _bitMask = 1 << (16 - (_bitPos) - 1); \
    _ipAddr = _ipAddr | _bitMask; \
}

/*
 * NAME: CLEAR_BIT_AT_POSITION_8
 *
 * DESCRIPTION: Sets (sets to one) a bit in an 8-bit chunk of an IP
 * address.
 *
 * @Param:      - IN    _ipAddr
 *                      A 8-bit portion of an IPv6 address
 *
 * @Param:      - IN    _bitPos
 *                      bit position of the bit that needs to be cleared
 *
 * @Returns:
 * none.
 */
#define SET_BIT_AT_POSITION_8(_ipAddr, _bitPos) \
{ \
    ix_uint8 _bitMask = 1 << (8 - ((_bitPos) % 8) -1); \
    _ipAddr = _ipAddr | _bitMask; \
}


/*
 * NAME: BIT_AT_POSITION_32
 *
 * DESCRIPTION: Returns the bit at position specified.
 *
 * @Param:      - IN    _ipAddr
 *                      A 32-bit portion of the IPv6 address.
 *
 * @Param:      - IN    _bitPos
 *                      bit position of the bit that needs to be returned
 *
 * @Param:      - OUT    _bit
 *                      The value of the bit at the position
 *
 * @Returns:
 * none.
 * 
 * For example, ipAddr = 0011 0001 and bitPost = 3, the value returned 
 * is 1
 *                         ^
 *                         |
 */

#define BIT_AT_POSITION_32(_ipAddr, _bitPos, _bit) \
{ \
    ix_uint32 _mask = 1 << (32 - (_bitPos) - 1); \
    _bit = (_mask & (_ipAddr)) >> (32 - (_bitPos) - 1); \
}

/*
 * NAME: BIT_AT_POSITION_16
 *
 * DESCRIPTION: Returns the bit at position specified.
 *
 * @Param:      - IN    _ipAddr
 *                      A 16-bit portion of the IPv6 address.
 *
 * @Param:      - IN    _bitPos
 *                      bit position of the bit that needs to be returned
 *
 * @Param:      - OUT    _bit
 *                      The value of the bit at the position
 *
 * @Returns:
 * none.
 *
 * For example, ipAddr = 0011 0001 and bitPost = 3, the value returned 
 * is 1
 *                         ^
 *                         |
 */

#define BIT_AT_POSITION_16(_ipAddr, _bitPos, _bit) \
{ \
    ix_uint16 _mask = 1 << (16 - (_bitPos) - 1); \
    _bit = (_mask & (_ipAddr)) >> (16 - (_bitPos) - 1); \
}

/*
 * NAME: BIT_AT_POSITION_8
 *
 * DESCRIPTION: Returns the bit at position specified.
 *
 * @Param:      - IN    _ipAddr
 *                      An 8-bit portion of the IPv6 address.
 *
 * @Param:      - IN    _bitPos
 *                      bit position of the bit that needs to be returned
 *
 * @Param:      - OUT    _bit
 *                      The value of the bit at the position
 *
 * @Returns:
 * none.
 * 
 * For example, ipAddr = 0011 0001 and bitPost = 3, the value returned 
 * is 1
 *                         ^
 *                         |
 */

#define BIT_AT_POSITION_8(_ipAddr, _bitPos, _bit) \
{ \
    ix_uint8 _mask = 1 << (8 - (_bitPos) - 1); \
    _bit = (_mask & (_ipAddr)) >> (8 - (_bitPos) - 1); \
}

/*
 * Since we get the IPv6 address as an array of four unsigned 32 bit
 * quantities, these macros make it useful to extract portions of
 * the IP address. 
 *
 * INDEX_IN_ARRAY() returns the index (between 0 and 3) to read from
 * and SHIFT_BY() returns the amount to shift by 
 */

/*
 * NAME: INDEX_IN_ARRAY
 *
 * DESCRIPTION: Generate an array index to read a portion (8 or 16 bit)
 * in an IPv6 address. The portion is specified by a bit position that
 * resides in that portion.
 *
 * @Param:      - IN    _bitPos
 *                      bit position specifying which portion to return
 *
 * @Param:      - OUT   _index  
 *                      The value of the index
 *
 * @Returns:
 * none.
 */
#define INDEX_IN_ARRAY(_bitPos, _index) \
{ \
    _index = ((_bitPos) / 32); \
}

/*
 * NAME: SHIFT_BY
 *
 * DESCRIPTION: Generate an value which will be used to shift a 32 bit value
 * (From the IPv6 array) to read a portion of the IP address. This is
 * used in combination with INDEX_IN_ARRAY described above.
 *
 * @Param:      - IN    _bitPos
 *                      bit position specifying the portion of the IP address
 *                      we are trying to read.
 *
 * @Param:      - OUT   _value  
 *                      The value to be used to shift the 32 bit value by.
 *
 * @Returns:
 * none.
 * 
 * For instance, if _bitPos is 18, the array index (from above) will be 0.
 * Now, we want to extract the 3rd byte (from the left) at array index 0.
 * So the value to shift by would be 8 bits or 1 byte);
 */
#define SHIFT_BY(_bitPos, _value) \
{ \
    if((_bitPos) < 15) \
    { \
        _value = 16; \
    } \
    else \
    { \
        _value = ((3 - ( ( (_bitPos) % 32 )/ 8 ) ) * 8); \
    } \
}

/*
 * NAME: BOUNDARY_BIT_POSITION
 *
 * DESCRIPTION: Checks to see if the bit position is a "boundary bit". Since
 * our route table is organized as a trie with stides 16-8-8... a boundary
 * bit is the last bit in a stride. So 15, 23, 31 etc.
 * 
 * @Param:      - IN    _bitPos
 *                      The bit position to be checked.
 *
 * @Returns:
 * TRUE: if it is boundary bit.
 * FALSE: Otherwise.
 */

static
__inline__ ix_uint8 BOUNDARY_BIT_POSITION(int _bitPos)
{
    if ((_bitPos) < 15)
    {
        return FALSE;
    }

    if ((_bitPos) == 15)
    {
        return TRUE;
    }
    else
    {
       /* If bitPos is 23, 31, 39 ..., 127 */
       if ((((_bitPos) + 1) % 8) == 0)
       {
           return TRUE;
       }
    }

    return FALSE;
}

/*
 * NAME: COMPUTE_STRIDE
 *
 * DESCRIPTION: Compute the stride (the 8-bit trie) this bit lies in.
 * If the bit lies in the 16-bit trie, -1 is returned.
 * 
 * @Param:      - IN    _bitPosition
 *                      The bit position to be checked.
 *
 * @Returns:
 * -1: If bitPosition lies in the 16-bit trie.
 * 0 - 14 : Indicating the level of the 8-bit trie.
 */
static __inline__
ix_int8 COMPUTE_STRIDE(ix_uint8 bitPosition)
{
    if (bitPosition < 16)
    {
        return (-1);
    }

    return ((bitPosition - 16)/8);
}

/*
 * NAME: bit_at_position
 *
 * DESCRIPTION: Returns the bit stored at a certain bit position in an
 * IPv6 address.
 * 
 * @Param:      - IN    ipAddr
 *                      The bit position to be checked.
 * 
 * @Param:      - IN    bitPos
 *                      The bit position to be checked.
 *
 * @Returns:
 * The value at bitPos (0/1)
 */
ix_uint8 bit_at_position(
    ix_uint32 *ipAddr,
    ix_uint8 bitPos);

/*
 * NAME: alloc_trie_node
 *
 * DESCRIPTION: This function allocates a trie node. The trie node 
 * represents an element in the 1-bit trie.
 * 
 * @Param:      - OUT   **ppNode
 *                      Pointer to the pointer to the allocated node.
 *
 * @Param:      - IN    arg_data
 *                      Data stored in the node. 
 *
 * @Returns:
 */

ix_error alloc_trie_node(
    ix_1bit_trie_node **ppNode,
    ix_uint16 arg_data);

/*
 * NAME: free_trie_node
 *
 * DESCRIPTION: This routine frees a trie node allocated by calling
 * alloc_trie_node(). There are no checks to see if the memory was in 
 * fact allocated by alloc_trie_node. It is assumed that the caller
 * will return the right pointer.
 * 
 * @Param:      - IN    *pNode
 *                      Pointer to the node.
 *
 * @Returns:
 */

void free_trie_node(
    ix_1bit_trie_node *pNode);

/*
 * NAME: init_trie_root()
 *
 * DESCRIPTION: This routine initializes the root of the 1-bit trie
 * in memory. This node (the root) doesn't contain any data.
 * 
 * @Param:      - IN    hLpmv6
 *                      Handle of the SW LPM V6 structure.
 *
 * @Returns:
 */
ix_error init_trie_root(
    _ix_lkup_swlpmv6 pLpmv6);

/*
 * NAME: fini_trie_root()
 *
 * DESCRIPTION: This routine frees the root of the 1-bit trie in memory.
 * 
 * @Param:      - IN    hLpmv6
 *                      Handle of the SW LPM V6 structure.
 *
 * @Returns:
 */
void fini_trie_root(
    _ix_lkup_swlpmv6 pLpmv6);


/*
 * NAME: __swlpmv6_add_route
 *
 * DESCRIPTION: This routine adds a route to the routing table in shared
 * memory. Care is taken to avoid overwriting "more specific" routes, if
 * they exist.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 *
 * @Param:      - IN    ipAddr
 *                      The IPv6 address
 *
 * @Param:      - IN    maskLength
 *                      Prefix length of the route we are trying to add
 *
 * @Param:      - IN    data
 *                      Data for this route. Typically the next-hop ID
 *
 * @Returns:
 *
 */
ix_error __swlpmv6_add_route(
    _ix_lkup_swlpmv6 hLpmv6,
    ix_uint32 *ipAddr,
    ix_uint8 maskLength,
    ix_uint16 data);

/*
 * NAME: __swlpmv6_delete_route
 *
 * DESCRIPTION: This routine handles deletion routes given an IPv6 address
 * and a prefix length. 
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 *
 * @Param:      - IN    ipAddr
 *                      The IPv6 address
 *
 * @Param:      - IN    maskLength
 *                      The prefix length for this route
 *
 * @Returns:
 * none.
 */
ix_error __swlpmv6_delete_route(
    _ix_lkup_swlpmv6 hLpmv6,
    ix_uint32 *ipAddr,
    ix_uint8 maskLength);

/*
 * NAME: __swlpmv6_update_route
 *
 * DESCRIPTION: This routine updates a route in the routing table in shared
 * memory. Care is taken to avoid overwriting "more specific" routes, if
 * they exist.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 *
 * @Param:      - IN    ipAddr
 *                      The IPv6 address
 *
 * @Param:      - IN    maskLength
 *                      Prefix length of the route we are trying to add
 *
 * @Param:      - IN    data
 *                      Data for this route. Typically the next-hop ID
 *
 * @Returns:
 *
 */
ix_error __swlpmv6_update_route(
    _ix_lkup_swlpmv6 hLpmv6,
    ix_uint32 *ipAddr,
    ix_uint8 maskLength,
    ix_uint16 data);


/*
 * NAME: __swlpmv6_find_in_route_table
 *
 * DESCRIPTION: This routine tries to find the longest prefix match for
 * a given IPv6 address.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 *
 * @Param:      - IN    ipAddr
 *                      The IPv6 address for which a lookup is being
 *                      performed.
 * 
 * @Param:      - OUT   pdata
 *                      Data returned as a result of the search. If pData
 *                      contains zero, it means that no route was found.
 *
 * @Returns:
 */
void __swlpmv6_find_in_route_table(
    _ix_lkup_swlpmv6 hLpmv6,
    ix_uint32 *ipAddr,
    ix_uint16 *pData);

/*
 * NAME: __swlpmv6_find_in_trie_table
 *
 * DESCRIPTION: This routine finds corresponiding to a given IP address
 * and a prefix length. This is different from the routine above because
 * it is not a longest prefix match. This routine walks down the 1-bit trie
 * in memory and stops when it has traversed down to the right level.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 *
 * @Param:      - IN    ipAddr
 *                      The IPv6 address
 *
 * @Param:      - IN    maskLength
 *                      The mask length of the route we are interested in.
 *
 * @Param:      - OUT   pData
 *                      The result of the search opeartion is returned here
 *
 * @Returns:
 * none.
 */
ix_error
__swlpmv6_find_in_trie_table(
    _ix_lkup_swlpmv6 hLpmv6,
    ix_uint32 *ipAddr,
    ix_uint8 maskLength,
    ix_uint16 *pData);

/*
 * NAME: __swlpmv6_reset_table
 *
 * DESCRIPTION: This routine
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 *
 * @Returns:
 * none.
 */
ix_error __swlpmv6_reset_table(
    _ix_lkup_swlpmv6 hLpmv6);

/*
 * NAME: __swlpmv6_read_first_entry
 *
 * DESCRIPTION: This routine reads the first entry.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 * @Param:      - OUT   pIpv6Addr
 *                      The IPv6 address that is set as a result of the search.
 * @Param:      - OUT   pPrefixLength
 *                      The prefix Length that is set as a resulf of the search.
 * @Param:      - OUT   pWeight
 *                      Unused.
 * @Param:      - OUT   pData
 *                      The data resturned as a result of the search.
 *
 * @Returns:
 * IX_LKUP_ERROR_INTERNAL
 * IX_SUCCESS
 */
ix_error
__swlpmv6_read_first_entry(_ix_lkup_swlpmv6 hLpmv6,
                           ix_uint32 *pIpv6Addr,
                           ix_uint8 *prefixLength,
                           int *pWeight,
                           ix_uint16 *pData);

/*
 * NAME: __swlpmv6_read_next_entry
 *
 * DESCRIPTION: This routine reads the next entry from the 1-bit trie.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 * @Param:      - OUT   pIpv6Addr
 *                      The IPv6 address that is set as a result of the search.
 * @Param:      - OUT   pPrefixLength
 *                      The prefix Length that is set as a resulf of the search.
 * @Param:      - OUT   pWeight
 *                      Unused.
 * @Param:      - OUT   pData
 *                      The data resturned as a result of the search.
 *        
 * @Returns:
 * IX_LKUP_ERROR_INTERNAL
 * IX_SUCCESS
 */
ix_error
__swlpmv6_read_next_entry(_ix_lkup_swlpmv6 hLpmv6,
                           ix_uint32 *pIpv6Addr,
                           ix_uint8 *pPrefixLength,
                           int *pWeight,
                           ix_uint16 *pData);

/*
 * NAME: _swlpmv6_swap32
 *
 * DESCRIPTION: This routine access each element as a 32 bit quantity
 * and converts it to Host Byte Order.
 * 
 * @Param:      - IN, OUT   array32
 *                          128 bit IPv6 address in Network Byte Order.
 *
 * @Returns:
 */
static __inline__
void _swlpmv6_swap32(
    ix_uint32 *array32)
{
    ix_uint8 i;

    for (i = 0; i < 4; i++)
    {
        array32[i] = IX_NTOH32(array32[i]);
    }
}

/* 
 * NAME: _swlpmv6_print_ipv6_address
 *
 * DESCRIPTION: Print an IPv6 address.
 *
 * @Param:  - IN arg_Ipv6Addr - an array representing the IPv6 address
 *
 * @Returns:
 */
static __inline__
void _swlpmv6_print_ipv6_address(ix_uint128 *arg_Ipv6Addr)
{
    ix_uint8 i = 0;
    
    for (i = 0; i < 4; i++)
    {
        ix_ossl_message_log("IPv6_Add[%d] 0x%lx \n", 
	                               i, arg_Ipv6Addr->aUint32[i]);
    }

    return;
}


#if defined(__cplusplus)
} /* extern "C" */
#endif /* __cplusplus */

#endif /*__IX_LKUP_SWLPMV6_INTERNAL_H__*/
