/**
 * ===========================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      sw.c
 *
 * = DESCRIPTION
 *      Base implementation of the Software Lookup Library.  This file
 *      implements ix_lkup_sw_init() and a thin layer for
 *      IX_LKUP_CREATE_TABLE() and IX_LKUP_DESTROY_TABLE().  This thin layer
 *      calls specific implementations to create and destroy tables.  The
 *      initial implementation supports only LPM for IPv4.  This file will
 *      need to be updated when other software implementations are available.
 *
 *      Each software algorithm implementation must follow the guidelines
 *      described in sw.h.
 *
 * ===========================================================================
 */

#define IX_ERROR_FILE_IDENT                                             \
        "$Id: sw.c,v 1.13 2003/01/15 00:14:57 rmurty Exp $"

#include "ix_cc_error.h"
#include "lkup/swlpmv4.h"
#include "lkup/swlpmv6.h"
#include "ix_ossl.h"

/*
 * NAME: _ix_s_sw_lkup
 *
 * DESCRIPTION: Structure containing an ix_s_lkup structure and the additional
 * information the Software library must have for the handle.  A pointer to 
 * this structure can be returned to the caller as if it was a pointer to the
 * ix_s_lkup structure.
 */
typedef struct _ix_s_sw_lkup
{
    ix_s_lkup lkup;
    int numTables;
} *_ix_sw_lkup;

/*
 * NAME: _ix_lkup_sw_fini
 *
 * DESCRIPTION: Invoked via IX_LKUP_FINI().  Destroys a previously obtained
 * ix_lkup handle. This function is typically called when the application is
 * shutting down and is cleaning up. After successful completion of this call,
 * the handle becomes invalid and subsequent calls to this handle cannot be
 * made. Further, the tables created using this handle should not be used
 * anymore and should be destroyed before this function is called.
 *
 * @Param:      - IN    arg_hLkup
 *                      The handle to destroy. This handle was obtained
 *                      through a successful call to ix_lkup_sw_init.
 *
 * Error Codes:
 * IX_LKUP_ERROR_IN_USE
 *      The handle has active tables in use
 *
 * IX_LKUP_ERROR_INVALID
 *      The handle was invalid
 */
static ix_error _ix_lkup_sw_fini(
        ix_lkup arg_hLkup
        )
{
    _ix_sw_lkup hSwLkup = (_ix_sw_lkup) arg_hLkup;

    if (arg_hLkup == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_hLkup was NULL"));
    }

    if (hSwLkup->numTables != 0)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_IN_USE,
            ("%d tables still exist; destroy them first",
             hSwLkup->numTables));
    }

    ix_ossl_free(arg_hLkup);
    return IX_SUCCESS;
}

/*
 * NAME: _ix_lkup_sw_create_table
 *
 * DESCRIPTION: Invoked via IX_LKUP_CREATE_TABLE(). Creates a new instance
 * of a search table. If the call is successful then a handle to the new
 * table is returned that can be used for operations on the table.
 *
 * Evaluates some of the parameters and determines which table creation
 * function to invoke. The current implementation calls
 * _ix_lkup_swlpmv4_create_table() if arg_pTableConf->tableType is
 * IX_LKUP_TABLE_LPM and arg_pTableConf->keySize is 32.  
 * It calls _ix_lkup_swlpmv6_create_table() if the arg_pTableConf->tableType
 * is IX_LKUP_TABLE_LPM and arg_pTableConf->keySize is 128. If either is
 * not true, _ix_lkup_sw_create_table() returns an error.  As more algorithms
 * and key sizes are supported, this function must change to accomodate.
 *
 * @Param:      - IN    arg_hLkup
 *                      The ix_lkup that was returned when the init operation
 *                      completed.
 *
 * @Param:      - IN    arg_pTableConf
 *                      The table configuration state. This defines the type
 *                      of table as well as the size, search method, etc.
 *
 * @Param:      - IN    arg_pVndrConf
 *                      Pointer to vendor specific table configuration
 *                      information.
 *
 * @Param:      - OUT   arg_phTable
 *                      Pointer to an ix_lkup_table that is filled if the
 *                      operation completes successfully. If the call is 
 *                      successful, the lookup library allocates a table
 *                      handle with the initialized values. The storage for
 *                      the handle is managed by the library.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The handle was invalid, or some of the table configuration information
 *      was invalid or not supported.
 *
 * IX_LKUP_ERROR_RESOURCES
 *      Not enough resources were available to create the requested table.
 */
static ix_error _ix_lkup_sw_create_table(
        ix_lkup arg_hLkup,
        ix_lkup_table_conf *arg_pTableConf,
        void *arg_pVndrConf,
        ix_lkup_table *arg_phTable
        )
{
    _ix_sw_lkup hSwLkup = (_ix_sw_lkup) arg_hLkup;

    if (arg_hLkup == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_hLkup was NULL"));
    }

    if (arg_pTableConf == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_pTableConf was NULL"));
    }

    if (arg_phTable == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_phTable was NULL"));
    }

    /*
     * If the caller wants an LPM for a 32-bit key, send the request to the
     * Software LPM for IPv4.
     */
    if ((arg_pTableConf->tableType == IX_LKUP_TABLE_LPM) &&
        (arg_pTableConf->keySize == 32))
    {
        IX_ERROR_CR(_ix_lkup_swlpmv4_create_table(
                arg_hLkup,
                arg_pTableConf,
                arg_pVndrConf,
                arg_phTable
                ));
        hSwLkup->numTables++;
        return IX_SUCCESS;
    }

    /*
     * If the caller wants an LPM for a 128-bit key, send the request to the
     * Software LPM for IPv6.
     */
    if ((arg_pTableConf->tableType == IX_LKUP_TABLE_LPM) &&
        (arg_pTableConf->keySize == 128))
    {
        IX_ERROR_CR(_ix_lkup_swlpmv6_create_table(
                arg_hLkup,
                arg_pTableConf,
                arg_pVndrConf,
                arg_phTable
                ));
        hSwLkup->numTables++;
        return IX_SUCCESS;
    }

    /*
     * When other software implementations are available, invoke their
     * create_table() functions from here.
     */

    return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
        ("Could not create the requested table type"));
}

/*
 * NAME: _ix_lkup_sw_destroy_table
 *
 * DESCRIPTION: Invoked via IX_LKUP_DESTROY_TABLE(). Destroys a table that was
 * created earlier with IX_LKUP_CREATE_TABLE(). The application should remove
 * all entries from the table before it is destroyed because the implementation
 * may not be able to free these entries.
 *
 * The implementation forwards the request to the algorithm's destroy function.
 *
 * @Param:      - IN    arg_hLkup
 *                      Handle that was obained through a successful call to
 *                      ix_lkup_sw_init().
 *
 * @Param:      - IN    arg_hTable
 *                      Handle for the table to be destroyed.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      One of the handles was invalid.
 */
static ix_error _ix_lkup_sw_destroy_table(
        ix_lkup arg_hLkup,
        ix_lkup_table arg_hTable
        )
{
    _ix_sw_lkup hSwLkup = (_ix_sw_lkup) arg_hLkup;
    ix_lkup_sw_table hSwTable = (ix_lkup_sw_table) arg_hTable;

    if (arg_hLkup == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_hLkup was NULL"));
    }

    if (arg_hTable == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_hTable was NULL"));
    }

    IX_ERROR_CR(hSwTable->ltabDestroyTable(arg_hLkup, arg_hTable));

    hSwLkup->numTables--;

    return IX_SUCCESS;
}

/*
 * NAME: ix_lkup_sw_init
 *
 * DESCRIPTION: Invoked directly by the client.  Initializes the software
 * lookup management library and returns a handle to use in subsequent
 * operations.
 *
 * @Param:      - IN    arg_LibVersion
 *                      The API version the caller is expecting to use. For
 *                      the current implementation this must be 1. In the
 *                      future, the version number will be used for backwards
 *                      compatibility.
 *
 * @Param:      - OUT   arg_phLkup
 *                      Pointer to a lookup handle. If the call is successful,
 *                      the lookup library allocates a lookup handle with the
 *                      initialized values.
 *
 * Error Codes:
 * IX_LKUP_ERROR_UNSUPPORTED
 *      The version of the library that was requested is not supported.
 *
 * IX_LKUP_ERROR_RESOURCES
 *      Not enough resources are available to complete the request (out 
 *      of memory).
 */
ix_error ix_lkup_sw_init(
        int arg_LibVersion,
        ix_lkup *arg_phLkup)
{
    _ix_sw_lkup hSwLkup = NULL;
    ix_lkup hLkup = NULL;

    if ((arg_LibVersion <= 0) ||
        (arg_LibVersion > _IX_LKUP_SW_CURRENT_VERSION))
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_UNSUPPORTED,
            ("Invalid library version requested (%ld)", arg_LibVersion));
    }

    hSwLkup = (_ix_sw_lkup) ix_ossl_malloc(sizeof(struct _ix_s_sw_lkup));
    if (hSwLkup == NULL)
    {
        return IX_ERROR_LOCAL(IX_LKUP_ERROR_RESOURCES,
            ("Could not allocate %d bytes from the OSSL",
             sizeof(ix_s_lkup)));
    }

    hSwLkup->numTables = 0;

    hLkup = &hSwLkup->lkup;

    hLkup->ldataCreateTable = _ix_lkup_sw_create_table;
    hLkup->ldataDestroyTable = _ix_lkup_sw_destroy_table;
    hLkup->ldataFini = _ix_lkup_sw_fini;

    *arg_phLkup = hLkup;

    return IX_SUCCESS;
}

