/**
 * ===========================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      swlpmv4.c
 *
 * = DESCRIPTION
 *      Main file implementing the Software Longest Prefix Match algorithm 
 *      for IPv4.  This implementation fulfills the requirements of the
 *      Lookup Library for a single algorithm.  Clients of this algorithm
 *      should not call these functions directly, but should instead use
 *      the functions and macros provided in ix_lkup.h.
 *
 * ===========================================================================
 */

#define IX_ERROR_FILE_IDENT                                             \
        "$Id: swlpmv4.c,v 1.46 2003/11/13 01:49:26 rranjeet Exp $"

#include "ix_cc_error.h"
#include "ix_cc_macros.h"
#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_netmacros.h"
#include "lkup/internal/internal_swlpmv4.h"

/*
 * NAME: _ix_alloc_swlpmv4
 *
 * DESCRIPTION: Allocates memory from the heap to store data associated with
 * the _ix_lkup_swlpmv4 handle.  All memory is pre-initialized to zero, and
 * all fields except those dealing with shared memory are initialized to their
 * appropriate values.
 *
 * @Param:      - IN    arg_NumTries
 *                      Number of tries to allocate
 *
 * @Param:      - IN    arg_NumTrieInfoStructs
 *                      Number of trie info structures to allocate
 *
 * @Param:      - OUT   arg_phLpmv4
 *                      Handle of the newly allocated structure
 *
 * Error Codes:
 * IX_LKUP_ERROR_RESOURCES
 */
static ix_error _ix_alloc_swlpmv4(
     ix_uint32 arg_NumTries,
     ix_uint32 arg_NumTrieInfoStructs,
     _ix_lkup_swlpmv4 *arg_phLpmv4)
{
    _ix_lkup_swlpmv4 hLpmv4 = NULL;
    ix_uint32 heapSize = 0;
    ix_error err;

    /*
     * The size of memory allocated on the heap for the Software LPM is:
     *  * The main data structure used by the algorithm (which includes the
     *    Lookup Library's function pointer structure ix_lkup_table);
     *
     *  * A list of ix_uint32s for keeping track of which tries are allocated;
     *  
     *  * All the trie info structures;
     *
     *  * A list of ix_uint32s for keeping track of which trie info structures
     *    are allocated.
     */
    heapSize =
        sizeof(struct _ix_s_lkup_swlpmv4) +
        arg_NumTrieInfoStructs  * sizeof(_ix_trie_info);

    hLpmv4 = (_ix_lkup_swlpmv4) ix_ossl_malloc(heapSize);
    if (hLpmv4 == NULL)
    {
        return IX_ERROR_LOCAL(IX_LKUP_ERROR_RESOURCES,
            ("Could not allocate %d bytes from the OSSL", heapSize));
    }

    /*
     * Initialize the SW LPM's memory (including the In Use arrays) to zero.
     */
    ix_ossl_memset(hLpmv4, 0, heapSize);

    IX_ERROR_CGT(ix_cc_hash_table_init(
                _IX_LKUP_SWLPMV4_NUM_BUCKETS(arg_NumTrieInfoStructs),
                arg_NumTrieInfoStructs,
                IX_CC_HASH_TABLE_ALGO_MAGIC,
                &hLpmv4->hHash),
            err, IX_LKUP_ERROR_RESOURCES, 0, clean_hLpmv4);

    /*
     * Fill in the members of the algorithm's structure
     */
    hLpmv4->numTries = arg_NumTries;
    hLpmv4->iteratorKey = 0;
    hLpmv4->iteratorMaskLen = 0;
    hLpmv4->numTrieInfoStructs = arg_NumTrieInfoStructs;

    /*
     * pTrieInfoBase starts after the structure.
     */
    hLpmv4->pTrieInfoBase = (_ix_trie_info*) &hLpmv4[1];

    IX_ERROR_CGT(ix_cc_index_pool_init(arg_NumTries,
                1, /* zero is reserved and not used */
                &hLpmv4->hTrieIndexes),
            err, IX_LKUP_ERROR_RESOURCES, 0, clean_hash);

    IX_ERROR_CGT(ix_cc_index_pool_init(arg_NumTrieInfoStructs,
                0, /* no values need to be reserved */
                &hLpmv4->hTrieInfoIndexes),
            err, IX_LKUP_ERROR_RESOURCES, 0, clean_hTrieIndexes);

    *arg_phLpmv4 = hLpmv4;
    return IX_SUCCESS;

clean_hTrieIndexes:
    (void) ix_cc_index_pool_fini(hLpmv4->hTrieIndexes);
clean_hash:
    (void) ix_cc_hash_table_fini(hLpmv4->hHash);
clean_hLpmv4:
    ix_ossl_free(hLpmv4);
    return err;
}

/*
 * NAME: _ix_alloc_shared_memory
 *
 * DESCRIPTION: Allocates memory from the Resource Manager that will be
 * shared with the IPv4 microblock.
 *
 * @Param:      - IN    arg_MemType
 *                      Type of memory to allocate (SRAM, DRAM, etc.)
 *
 * @Param:      - IN    arg_DataChannel
 *                      Channel the data will be used from
 *
 * @Param:      - IN    arg_TableSize
 *                      Roughly indicates the number of routes that will
 *                      be stored.
 *
 * @Param:      - OUT   arg_ppBase
 *                      Resulting pointer to the base of the shared memory.
 *
 * Error Codes:
 * IX_LKUP_ERROR_RESOURCES
 * IX_LKUP_ERROR_INVALID
 */
static ix_error _ix_alloc_shared_memory(
        ix_memory_type arg_MemType,
        ix_uint32 arg_DataChannel,
        ix_uint32 arg_TableSize,
        ix_uint32 **arg_ppBase)
{
    ix_error err = IX_SUCCESS;
    ix_uint8 *pSharedMemory = NULL;
    ix_uint32 i;

    err = ix_rm_mem_alloc_aligned(
            arg_MemType,
            arg_DataChannel,
            arg_TableSize,
            IX_MEMORY_ALIGNMENT_TYPE_VIRTUAL,
            3,              /* shift by 3 bytes, gives 64-bit alignment */
            (void**) &pSharedMemory);
    if (err != IX_SUCCESS)
    {
        if (IX_ERROR_GET_CODE(err) == IX_RM_ERROR_ASSERTION)
        {
            err = IX_ERROR_CHAIN(err,
                IX_LKUP_ERROR_INVALID, IX_ERROR_LEVEL_WARNING,
                ("Invalid channel: %ld", arg_DataChannel));
        }
        else
        {
            err = IX_ERROR_CHAIN(err,
                IX_LKUP_ERROR_RESOURCES, IX_ERROR_LEVEL_WARNING,
                ("Could not allocate %d bytes from the Resource Manager",
                 arg_TableSize));
        }
        return err;
    }

    if (pSharedMemory == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_RESOURCES,
            ("Resource Manager returned pSharedMemory == NULL"));
    }

    /*
     * Initialize the shared memory to zero 64 bits at a time.
     */
    for (i = 0; i < arg_TableSize; i += sizeof(ix_uint64))
    {
        IX_RM_MEM_UINT64_WRITE(pSharedMemory + i, 0);
    }

    *arg_ppBase = (ix_uint32*) pSharedMemory;
    return IX_SUCCESS;
}

/*
 * NAME: _ix_lkup_swlpmv4_create_table
 *
 * DESCRIPTION: Invoked via IX_LKUP_CREATE_TABLE().  Creates a new table
 * using a software LPM algorithm.  This function is intended to be one of
 * several that might be called by the Lookup Library's more generic table
 * creation function.  The more generic function will call
 * _ix_lkup_swlpmv4_create_table() only when the table type is LPM and when
 * the key size indicates IPv4.
 *
 * @Param:      - IN    arg_hLkup
 *                      handle obtained by calling ix_lkup_sw_init().
 *
 * @Param:      - IN    arg_pTableConf
 *                      pointer to structure containing the table's desired
 *                      configuration.  This algorithm imposes these
 *                      restrictions:
 *                      * tableType == IX_LKUP_TABLE_LPM
 *                      * keySize == 32 bits
 *                      * dataSize <= 16 bits
 *                      * dataLocation == SRAM or DRAM
 *
 * @Param:      - IN    arg_pVndrConf
 *                      vendor-specific configuration; must be NULL
 *
 * @Param:      - OUT   arg_phTable
 *                      table handle to be used in future calls to the
 *                      algorithm
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The handle was invalid, or some of the table configuration information
 *      was invalid or not supported.
 *
 * IX_LKUP_ERROR_RESOURCES
 *      Not enough resources were available to create the requested table.
 */
ix_error _ix_lkup_swlpmv4_create_table(
        ix_lkup arg_hLkup,
        ix_lkup_table_conf *arg_pTableConf,
        void *arg_pVndrConf,
        ix_lkup_table *arg_phTable
        )
{
    _ix_lkup_swlpmv4 hLpmv4 = NULL;
    ix_lkup_sw_table hSwTable = NULL;
    ix_lkup_table hTable = NULL;
    ix_uint32 numTries = 0;
    ix_uint32 numTrieInfoStructs = 0;
    ix_memory_type rmMemType = 0;
    ix_error err = IX_SUCCESS;
    ix_uint32 *pSharedMemory = NULL;
    ix_uint32 tableSize = 0;

    const ix_uint32 hi64kTableSize = 0x10000 * sizeof(ix_uint32);
    const ix_uint32 hi256TableSize = 0x100   * sizeof(ix_uint32);

    /*
     * This algorithm doesn't need to use arg_hLkup. Typically only TCAM
     * versions will be interested in hLkup during create_table.
     */
    (void) arg_hLkup;

    if ((arg_pTableConf->tableType !=  IX_LKUP_TABLE_LPM)               ||
        (arg_pTableConf->keySize   != _IX_LKUP_SWLPMV4_KEY_SIZE)        ||
        (arg_pTableConf->dataSize  >  _IX_LKUP_SWLPMV4_MAX_DATA_SIZE)   ||
        (arg_pVndrConf != NULL))
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("Invalid table configuration"));
    }

    switch (arg_pTableConf->dataLocation)
    {
        case IX_LKUP_MEM_SRAM:
            rmMemType = IX_MEMORY_TYPE_SRAM;
            break;

        case IX_LKUP_MEM_DRAM:
            rmMemType = IX_MEMORY_TYPE_DRAM;
            break;

        default:
            return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
                ("Invalid dataLocation: %d", arg_pTableConf->dataLocation));
    }

    /*
     * Calculate the number of tries and trie info structures.  Round the
     * numbers up to an even 32.  Because _IX_INFO_PER_TRIE and
     * _IX_TRIE_ROUTE_RATIO may be small numbers, the calculation might give a
     * result of zero.  By adding 32 instead of the normal 31 used for
     * rounding, we guarantee the number of structs will be non-zero.
     */
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    numTries = arg_pTableConf->sizeHint >> _IX_TRIE_ROUTE_RATIO;
#else
    numTries = arg_pTableConf->sizeHint * _IX_TRIE_ROUTE_RATIO;
#endif
    numTries = (numTries + 32) & 0xFFFFFFE0UL;
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    numTrieInfoStructs = numTries / _IX_INFO_PER_TRIE;
#else
    numTrieInfoStructs = numTries * _IX_INFO_PER_TRIE;
#endif
    numTrieInfoStructs = (numTrieInfoStructs + 32) & 0xFFFFFFE0UL;

    /*
     * Allocate enough memory for the tries, the hi64k table,
     * and the hi256 table.
     */
    tableSize =
        hi64kTableSize + hi256TableSize + 
        numTries * sizeof(ix_uint32) * 16;

    IX_ERROR_CR(_ix_alloc_swlpmv4(numTries, numTrieInfoStructs, &hLpmv4));
    IX_ERROR_CG(_ix_alloc_shared_memory(
                    rmMemType,
                    arg_pTableConf->dataChannel,
                    tableSize,
                    &pSharedMemory),
        err, clean_hLpmv4);

    hSwTable = hLpmv4->hSwTable;
    hTable = hSwTable->hTable;

    hSwTable->ltabDestroyTable = _ix_lkup_swlpmv4_destroy_table;

    hTable->ltabAddEntry  = _ix_lkup_swlpmv4_add_entry;
    hTable->ltabRemoveEntry = _ix_lkup_swlpmv4_remove_entry;
    hTable->ltabUpdateEntry = _ix_lkup_swlpmv4_update_entry;
    hTable->ltabSearchTable = _ix_lkup_swlpmv4_search_table;
    hTable->ltabFindEntry = _ix_lkup_swlpmv4_find_entry;
    hTable->ltabReadFirstEntry = _ix_lkup_swlpmv4_read_first_entry;
    hTable->ltabReadNextEntry = _ix_lkup_swlpmv4_read_next_entry;
    hTable->ltabResetTable = _ix_lkup_swlpmv4_reset_table;
    hTable->ltabSetProperty = _ix_lkup_swlpmv4_set_property;
    hTable->ltabGetProperty = _ix_lkup_swlpmv4_get_property;
    hTable->ltabGetTableInfo = _ix_lkup_swlpmv4_get_table_info;

    hLpmv4->memType = arg_pTableConf->dataLocation;
    hLpmv4->hi64kTableBase = (tableElement*) pSharedMemory;
    hLpmv4->hi256TableBase = &hLpmv4->hi64kTableBase[0x10000];
    hLpmv4->trieTableBase = &hLpmv4->hi256TableBase[0x100];

    *arg_phTable = hTable;

    return IX_SUCCESS;

clean_hLpmv4:
    (void) ix_cc_hash_table_fini(hLpmv4->hHash);
    hLpmv4->hHash = NULL;
    (void) ix_cc_index_pool_fini(hLpmv4->hTrieIndexes);
    hLpmv4->hTrieIndexes = NULL;
    (void) ix_cc_index_pool_fini(hLpmv4->hTrieInfoIndexes);
    hLpmv4->hTrieInfoIndexes = NULL;
    ix_ossl_free(hLpmv4);
    return err;
}

/*
 * NAME: _ix_lkup_swlpmv4_destroy_table
 *
 * DESCRIPTION: Invoked via IX_LKUP_DESTROY_TABLE().  Destroys a table that
 * was created earlier with IX_LKUP_CREATE_TABLE().  The application should
 * remove all entries from the table before it is destroyed.
 *
 * @Param:      - IN    arg_hLkup
 *                      The handle that was obtained from ix_lkup_sw_init().
 *
 * @Param:      - IN    arg_hTable
 *                      The handle for the table to be destroyed.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      One of the handles was invalid
 */
ix_error _ix_lkup_swlpmv4_destroy_table(
        ix_lkup arg_hLkup,
        ix_lkup_table arg_hTable
        )
{
    _ix_lkup_swlpmv4 hLpmv4 = (_ix_lkup_swlpmv4) arg_hTable;

#if defined(IX_DEBUG)
    if (arg_hTable == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_hTable must not be NULL"));
    }
#endif /* IX_DEBUG */

    (void) arg_hLkup;

    (void) ix_rm_mem_free(hLpmv4->hi64kTableBase);
    (void) ix_cc_hash_table_fini(hLpmv4->hHash);
    (void) ix_cc_index_pool_fini(hLpmv4->hTrieIndexes);
    (void) ix_cc_index_pool_fini(hLpmv4->hTrieInfoIndexes);
    ix_ossl_free(hLpmv4);
    return IX_SUCCESS;
}

/*
 * NAME: _ix_lkup_swlpmv4_add_entry
 *
 * DESTRIPTION: Invoked via IX_LKUP_ADD_ENTRY().  Adds an entry to a table that
 * was previously created by calling IX_LKUP_CREATE_TABLE().  This function
 * will return an error if it detects a duplicate entry.
 *
 * @Param:      - IN    arg_hTable
 *                      The handle to the table to update.
 *
 * @Param:      - IN    arg_pKey
 *                      A pointer to the key data.  The length of the key was
 *                      defined when the table was created (i.e. 32 bits).
 *                      The key for this algorithm is the IP prefix.
 *
 * @Param:      - IN    arg_pMask
 *                      A pointer to the mask data for this entry. The length
 *                      of the mask was defined when the table was created.
 *
 * @Param:      - IN    arg_Weight
 *                      unused
 *
 * @Param:      - IN    arg_pData
 *                      Pointer to the data to be associated with the key.
 *                      This data is opaque to the lookup library and is only
 *                      interpreted by the higher level appliciations.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      Some of the parameters passed are not valid.
 *
 * IX_LKUP_ERROR_EXISTS
 *      An entry with the same key, mask and data is already in the table.
 *      (This implementation ignores the weight.)
 *
 * IX_LKUP_ERROR_CONFLICT
 *      An entry with the same key and mask but different data is already in
 *      the table. (This implementation ignores the weight.)
 *
 * IX_LKUP_ERROR_RESOURCES
 *      Not enough resources are available to complete the operation. Either
 *      the maximum table size has been exceeded or other resources could not
 *      be allocated.
 *
 * IX_LKUP_ERROR_INTERNAL
 *      An internal error occurred, probably because of a bug in the
 *      algorithm.
 */
ix_error _ix_lkup_swlpmv4_add_entry(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int arg_Weight,
        ix_uint8 *arg_pData
        )
{
    _ix_lkup_swlpmv4 hLpmv4 = (_ix_lkup_swlpmv4) arg_hTable;
    ix_error err;
    _ix_trie_chain trieChain;
    int trieChainIndex = 0;
    ix_uint16 startOfNew = 0;
    int bitPos = 0;

    (void) arg_Weight;

#if defined(IX_DEBUG)
    if (arg_hTable == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_hTable must not be NULL"));
    }

    if (arg_pData == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_pData must not be NULL"));
    }
#endif /* IX_DEBUG */

    /*
     * Pre-initialize some values of the trie chain: ipAddr, netMask, netMaskLen,
     * data.
     */
    IX_ERROR_CR(_ix_initialize_chain(arg_pKey, arg_pMask, arg_pData, &trieChain));

    /*
     * If the key is entered directly into the hi256 table, handle it
     * specially.  For all other cases, use the trie chain.
     */
    if (trieChain.netMaskLen <= 8)
    {
        /*
         * This is not an update operation
         */
        return _ix_add_hi256_entry(hLpmv4, 0, &trieChain);
    }

    /*
     * Copy tries from shared memory into the trie chain.
     */
    IX_ERROR_CR(
        _ix_copy_existing_from_base(hLpmv4, &trieChain, &startOfNew, &bitPos));
    trieChainIndex = startOfNew;

    /*
     * Now allocate enough additional tries to finish the chain. The
     * last trie may be partially masked, so round up to the next full trie.
     */
    while (bitPos+3 >= 32-trieChain.netMaskLen)
    {
        IX_ERROR_CG(
            ix_cc_index_pool_alloc(hLpmv4->hTrieIndexes,
                &trieChain.localTries[trieChainIndex].index),
            err, cleanTrieChain);

        /*
         * Because this link in the chain is new, we know it doesn't have a
         * trie info yet.
         */
        trieChain.localTries[trieChainIndex].trieInfoLookedUp = 1;

        /*
         * Since the memory was already set to zero, don't need to change
         * the values[] field.
         */

        bitPos -= 4;
        trieChainIndex++;
    }

    /*
     * Hook all the local tries together
     */
    IX_ERROR_CG(
        _ix_resolve_trie_chain(hLpmv4, &trieChain),
        err, cleanTrieChain);

    /*
     * Write the local tries out to shared memory in reverse order so the
     * microengines don't see the new route until it's completely written.
     */
    _ix_write_chain_to_shared(hLpmv4, &trieChain, _IX_LKUP_DIRECTION_BACKWARD);

    /*
     * If the entire chain is new, hook it into the appropriate hi table.
     */
    if (startOfNew == 0)
    {
        /*
         * The Hi64k table contains the entries with masks of 17 or more bits.
         */
        if (_IX_USE_HI64K_TABLE(trieChain.netMaskLen))
        {
            _IX_SET_HI64K_ENTRY(hLpmv4, trieChain.ipAddr,
                trieChain.localTries[0].index);
        }
        else
        {
            _IX_SET_HI256_ENTRY_INDEX(hLpmv4, (trieChain.ipAddr >> 24 ),
                trieChain.localTries[0].index);
        }
    }

    return IX_SUCCESS;

cleanTrieChain:
    for (; trieChainIndex > startOfNew; trieChainIndex--)
    {
        (void) ix_cc_index_pool_free(hLpmv4->hTrieIndexes,
            trieChain.localTries[trieChainIndex].index);
    }
    return err;
}

/*
 * NAME: _ix_lkup_swlpmv4_remove_entry
 *
 * DESCRIPTION: Invoked via IX_LKUP_REMOVE_ENTRY(). Removes an entry that was
 * previously added to the table. The entry is identified by the values used
 * to add the entry.
 *
 * @Param:      - IN    arg_hTable
 *                      Handle to the table to update.
 *
 * @Param:      - IN    arg_pKey
 *                      Pointer to the key data of the item to be removed. The
 *                      length of the data was defined when the table was
 *                      created.
 *
 * @Param:      - IN    arg_pMask
 *                      Pointer to the mask data of the entry to be removed.
 *                      The length of the mask data as defined when the table
 *                      was created.
 *
 * @Param:      - IN    arg_Weight
 *                      unused
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The table handle was not valid
 *
 * IX_LKUP_ERROR_NOTFOUND
 *      A matching entry was not found
 *
 * IX_LKUP_ERROR_INTERNAL
 *      An internal error occurred, probably because of a bug in the
 *      algorithm.
 */
ix_error _ix_lkup_swlpmv4_remove_entry(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int arg_Weight
        )
{
    _ix_lkup_swlpmv4 hLpmv4 = (_ix_lkup_swlpmv4) arg_hTable;
    _ix_trie_chain trieChain;
    _ix_local_trie *pLocal;
    ix_uint8 chainLen = 0;
    ix_uint16 firstToFree = _IX_CHAIN_LEN;

    ix_uint8 ipNibble;
    ix_uint8 maskNibble;
    ix_uint8 currEntry;
    ix_uint8 lastEntry;
    ix_uint32 nextTrieIndex; 	

    (void) arg_Weight;

#if defined(IX_DEBUG)
    if (arg_hTable == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_hTable must not be NULL"));
    }
#endif /* IX_DEBUG */

    IX_ERROR_CRT(_ix_initialize_chain(arg_pKey, arg_pMask, NULL, &trieChain),
        IX_LKUP_ERROR_NOTFOUND, 0);

    if (trieChain.netMaskLen <= 8)
    {
        /*
         * Falls in the middle of the hi256 table.  Because this is the only
         * case where the tries aren't used, handle it specially.
         */
        return _ix_del_hi256_entry(hLpmv4, &trieChain);
    }

    /*
     * Make a chain using the existing route, if it exists. This operation
     * copies the tries from shared memory into local memory.
     */
    IX_ERROR_CR(_ix_get_route_chain(hLpmv4, &trieChain, &chainLen, NULL));
    pLocal = &trieChain.localTries[chainLen-1];

    /*
     * See if the last trie in the chain has associated trie info.
     */
    IX_ERROR_CR(_ix_get_trie_info(hLpmv4, pLocal));

    /*
     * We know the route already exists, or _ix_get_route_chain() would have 
     * failed.  Set all affected entries to zero.
     */
    ipNibble = _IX_GET_NIBBLE(trieChain.ipAddr, 32-trieChain.netMaskLen);
    maskNibble = _IX_GET_NIBBLE(trieChain.netMask, 32-trieChain.netMaskLen);
    currEntry = ipNibble;

    /*
     * If the trie has trie info, clear out the appropriate prefix.
     */
    if (pLocal->pTrieInfo != NULL)
    {
        switch (maskNibble)
        {
            case 0x8:
                pLocal->pTrieInfo->prefix1[currEntry >> 3] = 0;
                break;

            case 0xC:
                pLocal->pTrieInfo->prefix2[currEntry >> 2] = 0;
                break;

            case 0xE:
                pLocal->pTrieInfo->prefix3[currEntry >> 1] = 0;
                break;

            /*
             * The default is impossible.
             */
            default:
            case 0xF:
                pLocal->pTrieInfo->prefix4DataExists &= ~(1<<currEntry);
                break;
        }
    }

    /*
     * For each of the entries affected by this route, find the next-best
     * data, combine it with the existing next trie, and assign the new
     * entry value.
     */
    lastEntry = _IX_GET_LAST_ENTRY(ipNibble, maskNibble);
    for (currEntry = ipNibble; currEntry <= lastEntry; currEntry++)
    {
        ix_uint16 data = _ix_get_best_fit_data(pLocal, currEntry);
        nextTrieIndex = _IX_TRIE_INDEX(pLocal->values[currEntry]);
        pLocal->values[currEntry] = _IX_TRIE_PACK_LOCAL(nextTrieIndex, data);
        pLocal->modified |= 1 << currEntry;
    }

    /*
     * As a result of removing the data, the last trie might be empty. Remove
     * it if so.  That could leave the next higher trie empty as well, etc.
     * Free all unused tries.  This operation only affects local memory, no
     * changes are made to shared memory yet.
     */
    IX_ERROR_CR(_ix_cleanup_chain(hLpmv4, &trieChain, &firstToFree));

    /*
     * If the entire chain is going away, unhook it from the appropriate
     * hi table first. Doing so helps protect the microengines.
     */
    if (firstToFree == 0)
    {
        if (_IX_USE_HI64K_TABLE(trieChain.netMaskLen))
        {
            _IX_SET_HI64K_ENTRY(hLpmv4, trieChain.ipAddr, 0);
        }
        else
        {
            _IX_SET_HI256_ENTRY(hLpmv4, trieChain.ipAddr, 0);
        }
    }

    /*
     * Write the changes out to shared memory. Use the forward direction to
     * protect the microengines.
     */
    _ix_write_chain_to_shared(hLpmv4, &trieChain, _IX_LKUP_DIRECTION_FORWARD);

    for (; firstToFree < _IX_CHAIN_LEN; firstToFree++)
    {
        pLocal = &trieChain.localTries[firstToFree];
        if (pLocal->index != 0)
        {
            /*
             * Don't have to check pLocal->trieInfoLookedUp here because
             * _ix_cleanup_chain() already got all of them.
             */
            if (pLocal->pTrieInfo != NULL)
            {
                IX_ERROR_CR(_ix_free_trie_info(hLpmv4, pLocal));
            }

            IX_ERROR_CR(ix_cc_index_pool_free(hLpmv4->hTrieIndexes,
                pLocal->index));
        }
    }

    return IX_SUCCESS;
}

/*
 * NAME: _ix_lkup_swlpmv4_update_entry
 *
 * DESCRIPTION: Invoked via IX_LKUP_UPDATE_ENTRY(). Updates the data
 * associated with an element already in the table.
 *
 * @Param:      - IN    arg_hTable
 *                      Handle to the table to update.
 *
 * @Param:      - IN    arg_pKey
 *                      Pointer to the key data of the item to be updated. The
 *                      length of the data was defined when the table was
 *                      created.
 *
 * @Param:      - IN    arg_pMask
 *                      Pointer to the mask data of the entry to be updated.
 *                      The length of the mask data as defined when the table
 *                      was created.
 *
 * @Param:      - IN    arg_Weight
 *                      unused
 *
 * @Param:      - IN    arg_pNewData
 *                      Pointer to the new data to associate with the specific
 *                      key.  The size of the data was specified when the
 *                      table was created.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The table handle was not valid
 *
 * IX_LKUP_ERROR_NOTFOUND
 *      A matching entry was not found so the update was not performed
 *
 * IX_LKUP_ERROR_INTERNAL
 *      An internal error occurred, probably because of a bug in the
 *      algorithm.
 */
ix_error _ix_lkup_swlpmv4_update_entry(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int arg_Weight,
        ix_uint8 *arg_pNewData
        )
{
    _ix_lkup_swlpmv4 hLpmv4 = (_ix_lkup_swlpmv4) arg_hTable;
    _ix_trie_chain trieChain;
    _ix_local_trie *pLocal;
    ix_uint8 chainLen = 0;

    (void) arg_Weight;

#if defined(IX_DEBUG)
    if (arg_hTable == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_hTable must not be NULL"));
    }
#endif /* IX_DEBUG */

    IX_ERROR_CRT(_ix_initialize_chain(arg_pKey, arg_pMask, arg_pNewData,
            &trieChain), IX_LKUP_ERROR_NOTFOUND, 0);

    /*
     * If the key is entered directly into the hi256 table, handle it
     * specially.  For all other cases, use the trie chain.
     */
    if (trieChain.netMaskLen <= 8)
    {
        /*
         * This is an update operation
         */
        return _ix_add_hi256_entry(hLpmv4, 1, &trieChain);
    }

    /*
     * Make a chain using the existing route, if it exists.
     */
    IX_ERROR_CR(_ix_get_route_chain(hLpmv4, &trieChain, &chainLen, NULL));
    pLocal = &trieChain.localTries[chainLen-1];

    /*
     * See if the last trie in the chain has associated trie info.
     */
    IX_ERROR_CR(_ix_get_trie_info(hLpmv4, pLocal));

    /*
     * Assign the local data.  Allow this entry to overwrite any existing.
     */
    IX_ERROR_CR(_ix_assign_local_data(hLpmv4, &trieChain, 1));
    _ix_write_trie_to_shared(hLpmv4, pLocal);

    return IX_SUCCESS;
}

/*
 * NAME: _ix_lkup_swlpmv4_search_table
 *
 * DESCRIPTION: Invoked via IX_LKUP_SEARCH_TABLE(). Searches a specific table
 * and returns the associated data.  The search key is compared with the
 * entries and masks using the Longest Prefix Match algorithm.
 *
 * @Param:      - IN    arg_hTable
 *                      Handle to the table to search.
 *
 * @Param:      - IN    arg_pKey
 *                      Pointer to the key data used for the search. The
 *                      length of the data was defined when the table was
 *                      created.
 *
 * @Param:      - OUT   arg_pData
 *                      Pointer to a location where the associated data will
 *                      be stored.  The search is always considered
 *                      successful; in effect, the algorithm assumes a
 *                      'default route' with the data 'zero'.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The handle was invalid
 *
 * NOTE: This implementation does not return error codes in search_table if
 * a route was not found
 */
ix_error _ix_lkup_swlpmv4_search_table(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pData
        )
{
    _ix_lkup_swlpmv4 hLpmv4 = (_ix_lkup_swlpmv4) arg_hTable;
    ix_uint32 ipAddr;
    ix_uint32 tableEntry;
    ix_uint8 bitPos;
	ix_uint32 trieIndex;	
    ix_uint16 trieData = 0;

#if defined(IX_DEBUG)
    if (arg_hTable == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_hTable must not be NULL"));
    }
#endif /* IX_DEBUG */

    /*
     * Search the hi64k table first, because it will give the most specific
     * results.  If nothing is found there, try the hi256 table.
     */

    ipAddr = IX_NTOH32(*(ix_uint32*) arg_pKey);

    tableEntry = _IX_GET_HI64K_ENTRY(hLpmv4, ipAddr);
    bitPos = 12;

    /*
     * This loop relies on the last entry in the table having its 'next
     * trie' value set to zero, or bitPos will roll below zero.
     */
    while ((trieIndex = _IX_TRIE_INDEX(tableEntry)) != 0)
    {
        tableEntry = _IX_GET_TRIE_ENTRY(
                hLpmv4,
                trieIndex,
                _IX_GET_NIBBLE(ipAddr, bitPos));
        bitPos -= 4;

        /*
         * If data exists at this location, it supercedes anything found so
         * far.
         */
        if (_IX_TRIE_DATA(tableEntry) != 0)
        {
            trieData = _IX_TRIE_DATA(tableEntry);
        }
    }

    /*
     * If the hi64k table didn't turn up anything, try again with the hi256.
     */
    if (trieData == 0)
    {
        tableEntry = _IX_GET_HI256_ENTRY(hLpmv4, ipAddr);
        bitPos = 20;

        /*
         * If data exists at this location, it supercedes anything found so
         * far.
         */
        if (_IX_TRIE_DATA(tableEntry) != 0)
        {
            trieData = _IX_TRIE_DATA(tableEntry);
        }

        while ((trieIndex = _IX_TRIE_INDEX(tableEntry)) != 0)
        {
            tableEntry = _IX_GET_TRIE_ENTRY(
                    hLpmv4,
                    trieIndex,
                    _IX_GET_NIBBLE(ipAddr, bitPos));
            bitPos -= 4;

            /*
             * If data exists at this location, it supercedes anything found so
             * far.
             */
            if (_IX_TRIE_DATA(tableEntry) != 0)
            {
                trieData = _IX_TRIE_DATA(tableEntry);
            }
        }
    }

    *(ix_uint16*)arg_pData = IX_HTON16(trieData);

    return IX_SUCCESS;
}

/*
 * NAME: _ix_lkup_swlpmv4_find_entry
 *
 * DESCRIPTION: Invoked via IX_LKUP_FIND_ENTRY(). Searches a table for an
 * entry that matches the exact key and mask combination. This is used to
 * locate a specific entry for search types where multiple entries may match
 * a specific search.
 *
 * @Param:      - IN    arg_hTable
 *                      Handle to the table to search.
 *
 * @Param:      - IN    arg_pKey
 *                      Pointer to the key data used for the search. The
 *                      length of the data was defined when the table was
 *                      created.
 *
 * @Param:      - IN    arg_pMask
 *                      Pointer to the mask data to use for the search.
 *                      The length of the mask data as defined when the table
 *                      was created.
 *
 * @Param:      - IN    arg_Weight
 *                      unused
 *
 * @Param:      - OUT   arg_pData
 *                      Pointer to a location where the associated data will
 *                      be stored if the search was successful. It assumes
 *                      that the caller allocated sufficient space for the
 *                      data associated with the call.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The handle was invalid
 *
 * IX_LKUP_ERROR_NOTFOUND
 *      There was no entry that matched this search request
 *
 * IX_LKUP_ERROR_INTERNAL
 *      An internal error occurred, probably because of a bug in the
 *      algorithm.
 */
ix_error _ix_lkup_swlpmv4_find_entry(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int arg_Weight,
        ix_uint8 *arg_pData
        )
{
    _ix_lkup_swlpmv4 hLpmv4 = (_ix_lkup_swlpmv4) arg_hTable;
    _ix_trie_chain trieChain;
    ix_uint16 existingData = 0;

    (void) arg_Weight;

#if defined(IX_DEBUG)
    if (arg_hTable == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_hTable must not be NULL"));
    }
#endif /* IX_DEBUG */

    IX_ERROR_CRT(_ix_initialize_chain(arg_pKey, arg_pMask, NULL, &trieChain),
        IX_LKUP_ERROR_NOTFOUND, 0);

    if (trieChain.netMaskLen <= 8)
    {
        /*
         * Falls in the middle of the hi256 table.  Because this is the only
         * case where the tries aren't used, handle it specially.
         */
        return _ix_get_hi256_entry(hLpmv4, &trieChain, arg_pData);
    }

    /*
     * Make a chain using the existing route, if it exists.  Note that this
     * could be a very expensive operation, and is a good candidate for
     * optimization if it is needed.
     */
    IX_ERROR_CR(_ix_get_route_chain(hLpmv4, &trieChain, NULL, &existingData));

    *(ix_uint16*)arg_pData = IX_NTOH16(existingData);

    return IX_SUCCESS;
}

/*
 * NAME: _ix_lkup_swlpmv4_read_first_entry
 *
 * DESCRIPTION: Invoked via IX_LKUP_READ_FIRST_ENTRY(). This is a debug
 * function that is used to read the items that are stored in the table. This
 * function will get the first item in the table. To continue iterating
 * through all the entries the programmer will call this function once and
 * then repeatedly call IX_LKUP_READ_NEXT_ENTRY until all the entries have
 * been enumerated.
 *
 * The enumeration algorithm for _ix_swlpmv4_read_first_entry() and
 * ix_swlpmv4_read_next_entry() is:
 * + Start with the hi256 table:
 *      + Search all the prefixN arrays and return their data (this covers
 *        prefixes up to a length of 7)
 *      + Search the hi256 table's mask array looking for an entry with a mask
 *        of 0xFF.  When one is found, read the corresponding entry from the
 *        shared memory and return it.  (This covers prefixes of a length of
 *        8.)
 *      + Read the entries in the hi256 table one-by-one.  If an entry links
 *      to a trie, read in the trie and follow its links until the data is
 *      found. (This covers prefixes up to a length of 16.)
 * + Read the entries in the hi64k table one-by-one.  If an entry links to
 *   a trie, read in the trie and follow its links until the data is
 *   found.  (This covers prefixes up to a length of 32.)
 *
 * Any time a route is found, update the hLpmv4->iteratorKey and
 * iteratorMaskLen values.  _ix_swlpmv4_read_next_entry() continues the search
 * with that key and mask length; _ix_swlpmv4_read_first_entry() starts over
 * every time.  The cookie is not used.
 *
 * @Param:      - IN    arg_hTable
 *                      Handle to the table to search.
 *
 * @Param:      - OUT   arg_pKey
 *                      Pointer where the key data will be stored if the call
 *                      is successful. This caller should make sure there is
 *                      sufficient storage for the key size of the table.
 *
 * @Param:      - OUT   arg_pMask
 *                      Pointer where the mask data will be stored. The caller
 *                      should ensure that this structure is of sufficient
 *                      size to store the mask.
 *
 * @Param:      - OUT   arg_pWeight
 *                      unused
 *
 * @Param:      - OUT   arg_pData
 *                      Pointer to the location where the associated data will
 *                      be stored if the call is successful.
 *
 * @Param:      - OUT   arg_pCookie
 *                      Pointer to an ix_lkup_cookie where the cookie is
 *                      stored. The cookie will be passed to a subsequent call
 *                      to IX_LKUP_READ_NEXT_ENTRY. (not used)
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      Some of the parameters passed are not valid
 *
 * IX_LKUP_ERROR_NOTFOUND
 *      This error is returned when there are no entries in the table. In this
 *      case none of the data structures to be filled in have valid data.
 *
 * IX_LKUP_ERROR_INTERNAL
 *      An internal error occurred, probably because of a bug in the
 *      algorithm.
 */
ix_error _ix_lkup_swlpmv4_read_first_entry(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int *arg_pWeight,
        ix_uint8 *arg_pData,
        ix_lkup_cookie *arg_pCookie
        )
{
    _ix_lkup_swlpmv4 hLpmv4 = (_ix_lkup_swlpmv4) arg_hTable;

    hLpmv4->iteratorKey = 0;
    hLpmv4->iteratorMaskLen = 0;
    return _ix_lkup_swlpmv4_read_next_entry(arg_hTable, *arg_pCookie,
        arg_pKey, arg_pMask, arg_pWeight, arg_pData, arg_pCookie);
}

/*
 * NAME: _ix_lkup_swlpmv4_read_next_entry
 *
 * DESCRIPTION: Invoked via IX_LKUP_READ_NEXT_ENTRY(). This is a debug
 * function that is used to read the items that are stored in the table. This
 * function is repeatedly called after IX_LKUP_READ_FIRST_ENTRY to enumerate
 * the contents of the table. Each call will return a cookie that is passed
 * to the next call.
 *
 * @Param:      - IN    arg_hTable
 *                      Handle to the table to search.
 *
 * @Param:      - IN    arg_OldCookie
 *                      The cookie that was returned by the last successful
 *                      call to IX_LKUP_READ_FIRST_ENTRY or
 *                      IX_LKUP_READ_NEXT_ENTRY.
 *
 * @Param:      - OUT   arg_pKey
 *                      Pointer where the key data will be stored if the call
 *                      is successful. This caller should make sure there is
 *                      sufficient storage for the key size of the table.
 *
 * @Param:      - OUT   arg_pMask
 *                      Pointer where the mask data will be stored. The caller
 *                      should ensure that this structure is of sufficient
 *                      size to store the mask.
 *
 * @Param:      - OUT   arg_pWeight
 *                      unused
 *
 * @Param:      - OUT   arg_pData
 *                      Pointer to the location where the associated data will
 *                      be stored if the call is successful.
 *
 * @Param:      - OUT   arg_pNewCookie
 *                      Pointer to an ix_lkup_cookie where the cookie is
 *                      stored. The cookie will be passed to a subsequent call
 *                      to IX_LKUP_READ_NEXT_ENTRY. (not used)
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      Some of the parameters passed are not valid
 *
 * IX_LKUP_ERROR_NOTFOUND
 *      This error is returned when there are no more entries in the table.
 *
 * IX_LKUP_ERROR_INTERNAL
 *      An internal error occurred, probably because of a bug in the
 *      algorithm.
 */
ix_error _ix_lkup_swlpmv4_read_next_entry(
        ix_lkup_table arg_hTable,
        ix_lkup_cookie arg_OldCookie,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int *arg_pWeight,
        ix_uint8 *arg_pData,
        ix_lkup_cookie *arg_pNewCookie
        )
{
    _ix_lkup_swlpmv4 hLpmv4 = (_ix_lkup_swlpmv4) arg_hTable;
    int found;

    /*
     * Until we find out otherwise, assume we're starting the search where we
     * left off before.  This value will get reset to zero each time we move
     * from one algorithm to another, and sometimes within an algorithm.
     */
    int continueSearch = 1;

    (void) arg_OldCookie;
    (void) arg_pWeight;

#if defined(IX_DEBUG)
    if (arg_hTable == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_hTable must not be NULL"));
    }

    if (arg_pNewCookie == NULL)
    {
        IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_pNewCookie must not be NULL"));
    }
#endif /* IX_DEBUG */
    *arg_pNewCookie = NULL;

    /*
     * If the starting point has a mask length less than 8, look at the hi256
     * table's prefix arrays.
     */
    if (hLpmv4->iteratorMaskLen < 8)
    {
        IX_ERROR_CR(_ix_search_hi256_table_info(hLpmv4, &found, arg_pKey,
            arg_pMask, arg_pData));
        if (found)
        {
            return IX_SUCCESS;
        }

        /*
         * We're moving on to a new search algorithm.  Tell the algorithm
         * to start fresh.
         */
        continueSearch = 0;
        hLpmv4->iteratorKey = 0x00000000;
        hLpmv4->iteratorMaskLen = 8;
    }

    /*
     * If the starting point has a mask length of 8, search through the hi256
     * table's mask array looking for one with an entry of 0xFF.  Get the
     * corresponding entry and return it.
     */
    if (hLpmv4->iteratorMaskLen == 8)
    {
        int maskIndex = hLpmv4->iteratorKey >> 24;

        if (continueSearch)
        {
            maskIndex++;
        }

        for (; maskIndex < 0x100; maskIndex++)
        {
            if (hLpmv4->hi256TrieInfo.mask[maskIndex] == 0xFF)
            {
                _ix_assign_search_outputs(
                    hLpmv4, (maskIndex << 24), 8,
                    _IX_GET_HI256_ENTRY_DATA(hLpmv4, maskIndex),
                    arg_pKey, arg_pMask, arg_pData);
                return IX_SUCCESS;
            }
        }

        /*
         * We're moving on to a new search algorithm.  Tell the algorithm
         * to start fresh.
         */
        continueSearch = 0;
        hLpmv4->iteratorKey = 0x00000000;
        hLpmv4->iteratorMaskLen = 9;
    } /* if (mask length == 8) */

    /*
     * If the starting point has a mask length between 9 and 16, search
     * through the hi256 table "by hand" looking for an entry with a non-zero
     * trie index.
     */
    if (hLpmv4->iteratorMaskLen <= 16)
    {
        int entryIndex = hLpmv4->iteratorKey >> 24;

        for (; entryIndex < 0x100; entryIndex++)
        {
            ix_uint16 entry = _IX_GET_HI256_ENTRY_INDEX(hLpmv4, entryIndex);
            if (entry != 0)
            {
                /*
                 * If we're continuing an existing search through the hi256
                 * table, the contents of iteratorKey is already accurate
                 * enough.  However if this is a new search, set the
                 * iteratorKey for _ix_search_trie_chain().
                 */
                if (!continueSearch)
                {
                    hLpmv4->iteratorKey = entryIndex << 24;
                }

                IX_ERROR_CR(_ix_search_trie_chain(hLpmv4, entry,
                        0, /* Came from the hi256 table */
                        continueSearch,
                        &found, arg_pKey, arg_pMask, arg_pData));
                if (found)
                {
                    return IX_SUCCESS;
                }

                /*
                 * No more to be found from this element of the hi256 table.
                 * Reset the mask length to the base length.
                 */
                hLpmv4->iteratorMaskLen = 9;
            } /* if (entry != 0) */

            /*
             * We're moving on to a new element, so tell the worker
             * function that it's dealing with new data.
             */
            continueSearch = 0;

            /*
             * When moving to a new location, start by digging down the tries.
             */
            hLpmv4->iteratorExamineData = 0;
            hLpmv4->iteratorKey = 0x00000000;

            /*
             * The mask length is set to the maximum so _ix_search_trie_chain
             * will copy as many tries as possible.
             */
            hLpmv4->iteratorMaskLen = 17;
        } /* for (entryIndex < 0x100) */
    } /* if (mask length <= 16) */

    /*
     * Search through the hi64k table "by hand" looking for an entry with a 
     * non-zero trie index.
     */
    if (hLpmv4->iteratorMaskLen <= 32)
    {
        ix_uint32 entryIndex = hLpmv4->iteratorKey >> 16;
        for (; entryIndex < 0x10000; entryIndex++)
        {
            ix_uint16 entry = _IX_TRIE_INDEX(
                _IX_GET_HI64K_ENTRY(hLpmv4, entryIndex << 16));
            if (entry != 0)
            {
                /*
                 * If we're continuing an existing search through the hi64k
                 * table, the contents of iteratorKey is already accurate
                 * enough.  However if this is a new search, set the
                 * iteratorKey for _ix_search_trie_chain().
                 */
                if (!continueSearch)
                {
                    hLpmv4->iteratorKey = entryIndex << 16;
                }

                IX_ERROR_CR(_ix_search_trie_chain(hLpmv4, entry,
                        1, /* Came from the hi64k table */
                        continueSearch,
                        &found, arg_pKey, arg_pMask, arg_pData));
                if (found)
                {
                    return IX_SUCCESS;
                }

                /*
                 * No more to be found from this element of the hi64k table.
                 * Reset the mask length to the base length.
                 */
                hLpmv4->iteratorMaskLen = 17;

                /*
                 * We're moving on to a new element, so tell the worker
                 * function that it's dealing with new data.
                 */
                continueSearch = 0;
            } /* if (entry != 0) */

            /*
             * When moving to a new location, start by digging down the tries.
             */
            hLpmv4->iteratorExamineData = 0;
        } /* for (entryIndex < 0x10000) */
    } /* if (mask len <= 32) */

    hLpmv4->iteratorMaskLen = 33;

    /*
     * Avoid the overhead of strings and buffers by using the simple ix_error
     * mechanism.
     */
    return IX_ERROR_NEW(IX_LKUP_ERROR_NOTFOUND, IX_ERROR_LEVEL_WARNING);
}

/*
 * NAME: _ix_lkup_swlpmv4_reset_table
 *
 * DESCRIPTION: Invoked via IX_LKUP_RESET_TABLE(). Clears all the items in the
 * table and resets it to its initial state. 
 *
 * @Param:      - IN    arg_hTable
 *                      The handle to the table to update.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The table handle was not valid
 *
 * IX_LKUP_ERROR_INTERNAL
 *      An internal error occurred, probably because of a bug in the
 *      algorithm.
 */
ix_error _ix_lkup_swlpmv4_reset_table(
        ix_lkup_table arg_hTable
        )
{
    ix_error err = IX_SUCCESS;
    ix_uint32 i;
    ix_uint32 tableSize = 0;
    const ix_uint32 hi64kTableSize = 0x10000 * sizeof(ix_uint32);
    const ix_uint32 hi256TableSize = 0x100   * sizeof(ix_uint32);
    _ix_lkup_swlpmv4 hLpmv4 = (_ix_lkup_swlpmv4) arg_hTable;
    char *pSharedMemory = NULL;
    ix_uint32 key;

#if defined(IX_DEBUG)
    if (arg_hTable == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_hTable must not be NULL"));
    }
#endif /* IX_DEBUG */

    tableSize =
        hi64kTableSize + hi256TableSize + 
        hLpmv4->numTries * sizeof(ix_uint32) * 16;

    pSharedMemory = (char*) hLpmv4->hi64kTableBase;

    /*
     * Initialize the shared memory to zero 64 bits at a time.
     */
    for (i = 0; i < tableSize; i += sizeof(ix_uint64))
    {
        IX_RM_MEM_UINT64_WRITE(pSharedMemory + i, 0);
    }

    /*
     * Clear out the Trie Info structures
     */
    ix_ossl_memset(hLpmv4->pTrieInfoBase, 0,
        sizeof(_ix_trie_info) * hLpmv4->numTrieInfoStructs);
    ix_ossl_memset(&(hLpmv4->hi256TrieInfo), 0, sizeof(_ix_hi256_trie_info));

    IX_ERROR_CRT(ix_cc_index_pool_reset(hLpmv4->hTrieIndexes),
        IX_LKUP_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
    IX_ERROR_CRT(ix_cc_index_pool_reset(hLpmv4->hTrieInfoIndexes),
        IX_LKUP_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);

    err = ix_cc_hash_table_get_first_entry(hLpmv4->hHash, &key, NULL);
    while (err == IX_SUCCESS)
    {
        IX_ERROR_CRT(ix_cc_hash_table_delete_entry(hLpmv4->hHash, key),
            IX_LKUP_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING);
        err = ix_cc_hash_table_get_first_entry(hLpmv4->hHash, &key, NULL);
    }
    if (IX_ERROR_GET_CODE(err) != IX_CC_ERROR_ENTRY_NOT_FOUND)
    {
        return IX_ERROR_CHAIN(err,
                IX_LKUP_ERROR_INTERNAL, IX_ERROR_LEVEL_WARNING,
                ("Error while removing entries from the hash table"));
    }

    return IX_SUCCESS;
}

/*
 * NAME: _ix_lkup_swlpmv4_set_property
 *
 * DESCRIPTION: Invoked via IX_LKUP_SET_PROPERTY().  No properties are
 * supported in this implementation.
 *
 * @Param:      - IN    arg_hTable
 *                      The handle to the table to update.
 *
 * @Param:      - IN    arg_Property
 *                      unused
 *
 * @Param:      - IN    arg_pValue
 *                      unused
 *
 * Error Codes:
 * IX_LKUP_ERROR_UNSUPPORTED
 *      The defined property was not supported
 */
ix_error _ix_lkup_swlpmv4_set_property(
        ix_lkup_table arg_hTable,
        int arg_Property,
        ix_uint8 *arg_pValue
        )
{
    (void) arg_hTable;
    (void) arg_Property;
    (void) arg_pValue;
    return IX_ERROR_WARNING(IX_LKUP_ERROR_UNSUPPORTED,
        ("This version of the Lookup Library does not support properties"));
}

/*
 * NAME: _ix_lkup_swlpmv4_get_property
 *
 * DESCRIPTION: Invoked via IX_LKUP_GET_PROPERTY(). No properties are
 * supported in this implementation.
 *
 * @Param:      - IN    arg_hTable
 *                      The handle to the table.
 *
 * @Param:      - IN    arg_Property
 *                      unused
 *
 * @Param:      - OUT   arg_pValue
 *                      unused
 *
 * Error Codes:
 * IX_LKUP_ERROR_UNSUPPORTED
 *      The defined property was not supported
 */
ix_error _ix_lkup_swlpmv4_get_property(
        ix_lkup_table arg_hTable,
        int arg_Property,
        ix_uint8 *arg_pValue
        )
{
    (void) arg_hTable;
    (void) arg_Property;
    (void) arg_pValue;
    return IX_ERROR_WARNING(IX_LKUP_ERROR_UNSUPPORTED,
        ("This version of the Lookup Library does not support properties"));
}

/*
 * NAME: _ix_lkup_swlpmv4_get_table_info
 *
 * DESCRIPTION: Invoked via IX_LKUP_GET_TABLE_INFO(). Returns values that need
 * to be used by the microengines:
 *
 * @Param:      - IN    arg_hTable
 *                      The handle to the table.
 *
 * @Param:      - OUT   arg_pTableId
 *                      Encodes the information needed by the microengines to
 *                      perform the lookup. 
 *
 * @Param:      - OUT   arg_pTableData1
 *                      unused
 *
 * @Param:      - OUT   arg_pTableData2
 *                      unused
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The table handle was not valid
 *
 * IX_LKUP_ERROR_INTERNAL
 *      An internal error occurred, probably because of a bug in the
 *      algorithm.
 */
ix_error _ix_lkup_swlpmv4_get_table_info(
        ix_lkup_table arg_hTable,
        ix_uint32 *arg_pTableId,
        ix_uint32 *arg_pTableData1,
        ix_uint32 *arg_pTableData2
        )
{
    ix_memory_type memType = 0;
    ix_uint32 memChannel = 0;
    ix_uint32 tableId = 0;
    _ix_lkup_swlpmv4 hLpmv4 = (_ix_lkup_swlpmv4) arg_hTable;

#if defined(IX_DEBUG)
    if (arg_hTable == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_hTable must not be NULL"));
    }
#endif /* IX_DEBUG */

    /*
     * This implementation uses the hi64k table's base address as the
     * table ID.
     */
    IX_ERROR_CRT(ix_rm_get_phys_offset(hLpmv4->hi64kTableBase,
                &memType, &memChannel, &tableId, NULL),
            IX_LKUP_ERROR_INTERNAL, 0);
    if (memType == IX_MEMORY_TYPE_SRAM)
    {
        IX_CC_SET_SRAM_CHANNEL_INTO_OFFSET(memChannel, tableId);
    }
    *arg_pTableId = tableId;

    /*
     * Data1 is unused
     */
    *arg_pTableData1 = 0;

    /*
     * Data2 is unused.
     */
    *arg_pTableData2 = 0;

    return IX_SUCCESS;
}

