/**
 * ===========================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      swlpmv4_chain.c
 *
 * = DESCRIPTION
 *      Helper functions for creating and using trie chains.  This is part of
 *      the Lookup API's Software LPM algorithm for IPv4.
 *
 * ===========================================================================
 */

#define IX_ERROR_FILE_IDENT                                             \
        "$Id: swlpmv4_chain.c,v 1.21 2003/11/06 03:10:41 rranjeet Exp $"

#include "ix_cc_error.h"
#include "ix_ossl.h"
#include "lkup/internal/internal_swlpmv4.h"



/*
 * NAME: _ix_copy_existing_from_base
 *
 * DESCRIPTION: Creates a chain (i.e. list of tries used to store the data for
 * a given IP address), starting at the appropriate hi table, using as many
 * existing, populated tries from shared memory as possible. Returns the index
 * of the first unmodified entry in the trie chain.
 *
 * @Param:      - IN    hLpmv4
 *                      Handle to the SW LPM v4.
 *
 * @Param:      - INOUT pChain
 *                      Pointer to pre-allocated memory. On input, ipAddr
 *                      and netMaskLen are expected to be valid. The unused
 *                      local tries must have their index value set to zero.
 *                      On output, some of the local tries are initialized
 *                      with data from shared memory.
 *
 * @Param:      - OUT   pStartOfNew
 *                      The first unused local trie in the chain.
 *
 * @Param:      - OUT   pBitPos
 *                      The lowest bit of the last nibble of the IP address
 *                      that was copied.
 *
 * Error Codes:
 * None
 */
ix_error _ix_copy_existing_from_base(
    _ix_lkup_swlpmv4 hLpmv4,
    _ix_trie_chain *pChain,
    ix_uint16 *pStartOfNew,
    int *pBitPos)
{
    ix_uint32 entry;
    ix_uint32 portionMask;

    if (_IX_USE_HI64K_TABLE(pChain->netMaskLen))
    {
        /*
         * The Hi64k table contains the entries with masks of 17 or more bits.
         */
        entry = _IX_TRIE_INDEX(_IX_GET_HI64K_ENTRY(hLpmv4, pChain->ipAddr));
        *pBitPos = 12;
        portionMask = 0xFFFF;
    }
    else
    {
        /*
         * The Hi256 table contains entries with 16 or fewer bits.
         */
        entry = _IX_TRIE_INDEX(_IX_GET_HI256_ENTRY(hLpmv4, pChain->ipAddr));
        *pBitPos = 20;
        portionMask = 0xFF;
    }

    return _ix_copy_existing_tries(hLpmv4, entry, portionMask, pBitPos,
        pChain, pStartOfNew);
}

/*
 * NAME: _ix_copy_existing_tries
 *
 * DESCRIPTION: Creates a chain (i.e. list of tries used to store the data for
 * a given IP address) using as many existing, populated tries from shared
 * memory as possible. Returns the index of the first unmodified entry in the
 * trie chain.
 *
 * @Param:      - IN    arg_hLpmv4
 *                      Handle to the SW LPM v4.
 *
 * @Param:      - IN    arg_Entry
 *                      Index of the first trie to copy from shared memory.
 *
 * @Param:      - IN    arg_PortionMask
 *                      An entity (either the hi256 table, the hi64k table, or
 *                      a trie) refers to this trie.  The maskPortion is the
 *                      part of the net mask that entity covers.  E.g.
 *                      the portionMask would be 0xFF if the trie is
 *                      referenced from the hi256 table, 0xFFFF if the trie
 *                      is referenced from the hi64k table, or 0xF if a trie.
 *
 * @Param:      - INOUT arg_pBitPos
 *                      On input, is the bit position into the IP address
 *                      of the first trie that needs to be copied.  On
 *                      output, is the bit position into the IP address of
 *                      the last trie that was copied.  E.g. On input if
 *                      the tries are referenced from the hi256 table,
 *                      *pBitPos would be 20; if they are coming from the
 *                      hi64k table, *pBitPos would be 12. If they are coming
 *                      from a trie, *pBitPos could be 0, 4, 8, 12, or 16.
 * 
 * @Param:      - INOUT arg_pChain
 *                      Pointer to pre-allocated memory. On input, ipAddr
 *                      and netMaskLen are expected to be valid. The unused
 *                      local tries must have their index value set to zero.
 *                      On output, some of the local tries are initialized
 *                      with data from shared memory.
 *
 * @Param:      - OUT   arg_pStartOfNew
 *                      The first unused local trie in the chain.
 *
 * Error Codes:
 * None
 */
ix_error _ix_copy_existing_tries(
    _ix_lkup_swlpmv4 arg_hLpmv4,
    ix_uint32 arg_Entry,
    ix_uint32 arg_PortionMask,
    int *arg_pBitPos,
    _ix_trie_chain *arg_pChain,
    ix_uint16 *arg_pStartOfNew)
{
    ix_uint16 trieChainIndex;
    int bitPos = *arg_pBitPos;
    ix_uint32 entry = arg_Entry;
    ix_uint32 portionMask = arg_PortionMask;

    /*
     * Copy as much of the existing chain as we can, and need to.  Stop
     * if we run out of the chain, or if we get as much as we need.
     */
    trieChainIndex = 0;
    while ((entry) && (bitPos+3 >= 32-arg_pChain->netMaskLen))
    {
        _ix_local_trie *pLocal = &arg_pChain->localTries[trieChainIndex];

#ifdef OPTIONAL
        /*
         * bitPos is a uint and therefore rolling below zero means it
         * becomes very large -- greater than 32.  This section of code can
         * be included for sanity checking, but should never be triggered.
         */
        if (bitPos > 32)
        {
            return IX_ERROR_LOCAL(IX_LKUP_ERROR_INTERNAL,
                    ("bitPos rolled below zero"));
        }
#endif

        /*
         * Copy a single trie from shared memory into the first available
         * local trie in the chain.
         */
        IX_ERROR_CR(_ix_append_trie_to_chain(arg_hLpmv4, arg_pChain,
                    entry, bitPos,
                    portionMask, &trieChainIndex));

        /*
         * Dereference the current trie by looking at the appropriate
         * offset, according to the nibble of ipAddr we're concerned with
         */
        entry = _IX_TRIE_INDEX(
                pLocal->values[_IX_GET_NIBBLE(arg_pChain->ipAddr, bitPos)]);

        bitPos -= 4;

        /*
         * _ix_append_trie_to_chain() reports which index was filled.  We want
         * to report the next available index, so increment by one.
         */
        trieChainIndex++;
        portionMask = 0xF;
    }

    *arg_pStartOfNew = trieChainIndex;
    *arg_pBitPos = bitPos;

    return IX_SUCCESS;
}

/*
 * NAME: _ix_resolve_trie_chain
 *
 * DESCRIPTION: Iterate through the chain and hook all the elements together
 * based on the pChain->ipAddr. Allocate trie info structures for tries when
 * necessary.
 *
 * @Param:      - IN    hLpmv4
 *                      Handle to the SW LPM v4.
 *
 * @Param:      - INOUT pChain
 *                      Pointer to the trie chain with local tries that
 *                      need to be connected.  The chain must be fully
 *                      populated with valid data (ipAddr, netMask,
 *                      netMaskLen, valid tries). It is acceptable if the
 *                      tries have not had the trie info looked up.
 *
 * Error Codes:
 * IX_LKUP_ERROR_CONFLICT
 * IX_LKUP_ERROR_EXISTS
 * IX_LKUP_ERROR_INTERNAL
 * IX_LKUP_ERROR_RESOURCES
 */
ix_error _ix_resolve_trie_chain(
    _ix_lkup_swlpmv4 hLpmv4,
    _ix_trie_chain *pChain)
{
    /*
     * If the net mask is more than 16 bits, use hi64k.  In this case, the
     * first trie looks at bits 12-15.  Otherwise, use the 256 table.  The
     * first trie looks at bits 20-23.
     */
    ix_uint16 trieChainIndex;
    ix_uint8 bitPos = _IX_USE_HI64K_TABLE(pChain->netMaskLen) ? 12 : 20;
    ix_uint8 ipNibble = _IX_GET_NIBBLE(pChain->ipAddr, bitPos);
    ix_uint8 maskNibble = _IX_GET_NIBBLE(pChain->netMask, bitPos);

    /*
     * For each local trie that isn't intended to be empty, connect the trie
     * to the next trie in the list.  Note that if the end of the list is
     * reached, the 'next' trie has an index of zero.
     */
    for (trieChainIndex = 0;
        (pChain->localTries[trieChainIndex+1].index != 0) && (maskNibble == 0xF);
        trieChainIndex++)
    {
        _ix_local_trie *pLocal = &pChain->localTries[trieChainIndex];

		
        /*
         * Connect this trie to the next one
         */
        IX_ERROR_CR(_ix_connect_local_tries(pLocal, ipNibble,
                        pChain->localTries[trieChainIndex+1].index));

        /*
         * For future reference, make note of which entry in the previous
         * trie refers to this trie.  Makes it easier to build up the entire
         * ipAddress and netMask later on.
         */
        pLocal->ipPortion = _IX_GET_NIBBLE(pChain->ipAddr, bitPos+4);
        pLocal->maskPortion = _IX_GET_NIBBLE(pChain->netMask, bitPos+4);

        /*
         * Move on to the next trie.
         */
        bitPos -= 4;
        ipNibble   = _IX_GET_NIBBLE(pChain->ipAddr, bitPos);
        maskNibble = _IX_GET_NIBBLE(pChain->netMask, bitPos);
    }

    /*
     * The preceding loop doesn't assign the ipPortion and maskPortion
     * correctly for the first trie.  Reset it to the proper value.
     */
    if (_IX_USE_HI64K_TABLE(pChain->netMaskLen))
    {
        pChain->localTries[0].ipPortion   = _IX_GET_WORD(pChain->ipAddr, 16);
        pChain->localTries[0].maskPortion = _IX_GET_WORD(pChain->netMask, 16);
    }
    else
    {
        pChain->localTries[0].ipPortion   = _IX_GET_BYTE(pChain->ipAddr, 20);
        pChain->localTries[0].maskPortion = _IX_GET_BYTE(pChain->netMask, 20);
    }

    /*
     * Write the data into the trie.  Don't allow this entry to overwrite
     * an existing value.  This only affects local memory, not shared memory.
     */
    IX_ERROR_CR(_ix_assign_local_data(hLpmv4, pChain, 0));

    return IX_SUCCESS;
}

/*
 * NAME: _ix_connect_local_tries
 *
 * DESCRIPTION: Connects one trie to another.  This only needs to be done to
 * one position in the current trie (otherwise we'd assign the data, not
 * another trie).  It's ok if the current trie already points to the next
 * one.  It's also ok if the current trie already points to data at the
 * specified location.
 *
 * @Param:      - INOUT pLocal
 *                      Pointer to the local trie that needs to be connected
 *                      to the next trie
 *
 * @Param:      - IN    ipNibble
 *                      4-bit portion of the IP address; used as an index into
 *                      the local trie's 16-element array.
 *
 * @Param:      - IN    nextTrieIndex
 *                      Shared-memory index of the next trie to be assigned
 *                      into the ipNibble'th element of the local trie's array.
 *
 * Error Codes:
 * IX_LKUP_ERROR_CONFLICT
 */
ix_error _ix_connect_local_tries(
    _ix_local_trie *pLocal,
    ix_uint8 ipNibble,
    ix_uint32 nextTrieIndex)

{
    ix_uint32 trieIndex = _IX_TRIE_INDEX(pLocal->values[ipNibble]);
 	

    if (trieIndex == nextTrieIndex)
    {
        /*
         * Nothing to do, return.
         */
        return IX_SUCCESS;
    }

    if (trieIndex == 0)
    {
        /*
         * It doesn't exist yet, so assign it.  Preserve any existing data.
         */
        ix_uint16 data = _IX_TRIE_DATA(pLocal->values[ipNibble]);
        pLocal->values[ipNibble] = _IX_TRIE_PACK_LOCAL(nextTrieIndex, data);
        pLocal->modified |= 1 << ipNibble;
        return IX_SUCCESS;
    }

    return IX_ERROR_WARNING(IX_LKUP_ERROR_CONFLICT,
        ("Trie #%d already points to #%d in position %d",
         pLocal->index, trieIndex, ipNibble));
}

/*
 * NAME: _ix_assign_local_data
 *
 * DESCRIPTION: Assigns the data indicated in pChain->data into the
 * appropriate local trie in the trie chain. Only one local trie is affected,
 * but between zero and eight locations within that trie could be changed.
 *
 * @Param:      - IN    hLpmv4
 *                      Handle to the SW LPM v4.
 *
 * @Param:      - INOUT pChain
 *                      The trie chain.  On input, ipAddr, netMask and
 *                      netMaskLen are expected to be valid.  Also, it is
 *                      expected that the chain contains sufficient allocated
 *                      tries to store the data in the appropriate location.
 *                      On output, only one trie's data is affected.
 *
 * @Param:      - IN    allowExisting
 *                      Indicates whether the new data may overwrite existing
 *                      data. In general, set to 0 for new routes, or 1 when
 *                      updating routes.
 *
 * Error Codes:
 * IX_LKUP_ERROR_CONFLICT
 * IX_LKUP_ERROR_EXISTS
 * IX_LKUP_ERROR_INTERNAL
 * IX_LKUP_ERROR_RESOURCES
 */
ix_error _ix_assign_local_data(
    _ix_lkup_swlpmv4 hLpmv4,
    _ix_trie_chain *pChain,
    int allowExisting)
{
    /*
     * 'existing' is a bitmask of the entries in the trie that already point
     * to other data.  This may be ok, or it may indicate a conflicting route.
     */
    ix_uint16 existing = 0;

    int currEntry = 0;
    ix_uint16 existingData;
    _ix_local_trie *pLocal = &pChain->localTries[_IX_CHAIN_LEN-1];

    /*
     * The ipNibble and maskNibble are the portions of the IP address and
     * network mask that affect where the data is placed.  Start from the
     * least significant nibble.
     */
    ix_uint8 ipNibble = _IX_GET_NIBBLE(pChain->ipAddr, 32-pChain->netMaskLen);
    ix_uint8 maskNibble = _IX_GET_NIBBLE(pChain->netMask, 32-pChain->netMaskLen);
    ix_uint8 lastEntry;

#ifdef OPTIONAL
    if (pChain->localTries[0].index == 0)
    {
        return IX_ERROR_LOCAL(IX_LKUP_ERROR_INTERNAL,
            ("The first trie in the chain has index zero."));
    }
#endif

    /*
     * Search backward from the end for the first real trie.
     */
    while (pLocal->index == 0)
    {
        pLocal--;
#ifdef OPTIONAL
        if (pLocal < pChain->localTries)
        {
            return IX_ERROR_LOCAL(IX_LKUP_ERROR_INTERNAL,
                ("Internal Error: pLocal went out of range"));
        }
#endif
    }

    if (!allowExisting)
    {
        /*
         * Verify this route doesn't already exist.
         */
        IX_ERROR_CR(
            _ix_lookup_local(hLpmv4, ipNibble, maskNibble, pLocal,
                &existingData));
        if (existingData != 0)
        {
            if (existingData == pChain->data)
            {
                return _IX_ROUTE_EXISTS(EXISTS, pChain);
            }

            return _IX_ROUTE_EXISTS(CONFLICT, pChain);
        }
    }

    /*
     * It is ok to store the new data.  See which, if any, of the values of the
     * local trie have data other than what we want.
     */
    lastEntry = _IX_GET_LAST_ENTRY(ipNibble, maskNibble);
    for (currEntry = ipNibble; currEntry <= lastEntry; currEntry++)
    {
        if (_IX_TRIE_DATA(pLocal->values[currEntry]) != 0)
        {
            /*
             * Mark that this one already has some other data.
             */
            existing |= 1 << currEntry;
        }
    }

    /*
     * See if we need to create a trie info structure.  Of course, no need
     * if one already exists.
     *
     * The conditions where we need to allocate a new trie info structure are
     * where:
     * 1) We are about to add a route that overlaps with an existing route in
     *    this trie;
     * 2) We are about to add a route that looks exactly like a neighbor route.
     *    We need a trie info in this case because two neighbor routes that
     *    have the same data look like a single less specific route.  The trie
     *    info disabiguates the routes.
     */
    if (pLocal->pTrieInfo == NULL)
    {
        /*
         * Keep a flag indicating whether we need a trie info structure.
         * We immediately know we need a trie info if there is any existing
         * data but we don't allow there to be any.
         */
        int needTrieInfo = existing && !allowExisting;

        /*
         * Only look at the neighbor position if we don't have an overlap
         * situation.
         */
        ix_uint16 peerData = 0;

        /*
         * If we haven't already decided to make a trie info structure,
         * check for a peer.
         */
        if (!needTrieInfo)
        {
            /*
             * See if the 'peer' of the current nibble exists.
             */
            IX_ERROR_CR(_ix_lookup_local(hLpmv4,
                        _IX_GET_PEER(ipNibble, maskNibble), maskNibble,
                        pLocal, &peerData));
            if (peerData == pChain->data)
            {
                needTrieInfo = 1;
            }
        }

        if (needTrieInfo)
        {
            /*
             * Conditions have been met, create the trie info.
             */
            IX_ERROR_CR(_ix_make_trie_info(hLpmv4, pLocal));
        }
    }

    /*
     * Finally, assign the values to the local trie.  If there is no trie info,
     * simply set all the values.  If there is trie info, use it to determine
     * which (if any) values need to be set.
     */
    if (pLocal->pTrieInfo == NULL)
    {
        ix_uint32 trieIndex;  
        lastEntry = _IX_GET_LAST_ENTRY(ipNibble, maskNibble);
        for (currEntry = ipNibble; currEntry <= lastEntry; currEntry++)
        {
            trieIndex = _IX_TRIE_INDEX(pLocal->values[currEntry]);
            pLocal->values[currEntry] =
                _IX_TRIE_PACK_LOCAL(trieIndex, pChain->data);
            pLocal->modified |= 1 << currEntry;
        }
    }
    else
    {
        _ix_trie_info *pTrieInfo = pLocal->pTrieInfo;
        ix_uint16 *pPrefixData =
            _ix_get_trie_prefix_entry(pLocal, maskNibble, ipNibble);

        /*
         * If pPrefixData is non-NULL, the data might apply to several entries.
         * Go through each of them and find the most appropriate data to store.
         */
        if (pPrefixData != NULL)
        {
            lastEntry = _IX_GET_LAST_ENTRY(ipNibble, maskNibble);
            *pPrefixData = pChain->data;
            for (currEntry = ipNibble; currEntry <= lastEntry; currEntry++)
            {
                ix_uint16 data = _ix_get_best_fit_data(pLocal, currEntry);
                ix_uint16 oldData = _IX_TRIE_DATA(pLocal->values[currEntry]);
                /* ix_uint16 trieIndex = _IX_TRIE_INDEX(pLocal->values[currEntry]); */
				ix_uint32 trieIndex = _IX_TRIE_INDEX(pLocal->values[currEntry]);
                /*
                 * To avoid too many hits to shared memory, only mark it as
                 * modified if it really did change.
                 */
                if (data != oldData)
                {
                    pLocal->values[currEntry] =
                        _IX_TRIE_PACK_LOCAL(trieIndex, data);
                    pLocal->modified |= 1 << currEntry;
                }
            }
        }
        else
        {
            /*
             * Since pPrefixData is NULL, this entry doesn't have a
             * corresponding prefix array entry -- it must be a 4-bit value.
             */
            /* ix_uint16 trieIndex = _IX_TRIE_INDEX(pLocal->values[ipNibble]); */
			ix_uint32 trieIndex = _IX_TRIE_INDEX(pLocal->values[ipNibble]);
            pLocal->values[ipNibble] =
                _IX_TRIE_PACK_LOCAL(trieIndex, pChain->data);
            pLocal->modified |= 1 << ipNibble;
            pTrieInfo->prefix4DataExists |= 1 << ipNibble;
        }
    }

    return IX_SUCCESS;
}

/*
 * NAME: _ix_write_trie_to_shared
 *
 * DESCRIPTION: Copies the contents of a single trie to shared memory.
 *
 * @Param:      - IN    hLpmv4
 *                      Handle to the SW LPM v4.
 *
 * @Param:      - IN    pLocal
 *                      Pointer to the local trie to be copied.
 */
void _ix_write_trie_to_shared(
    _ix_lkup_swlpmv4 hLpmv4,
    const _ix_local_trie *pLocal)
{
    if (pLocal->modified != 0)
    {
        int valueIndex;
        for (valueIndex = 0; valueIndex < 16; valueIndex++)
        {
            if (pLocal->modified & (1 << valueIndex))
            {
                _IX_SET_TRIE_ENTRY(hLpmv4, pLocal->index, valueIndex,
                    pLocal->values[valueIndex]);
            }
        }
    }
}

/*
 * NAME: _ix_write_chain_to_shared
 *
 * DESCRIPTION: Writes a chain to shared memory, in the direction indicated.
 * The chain should be written in the forward direction when removing a route,
 * and in the backward direction when adding a route.
 *
 * @Param:      - IN    hLpmv4
 *                      Handle to the SW LPM v4.
 *
 * @Param:      - IN    pChain
 *                      Pointer to the chain to write to shared memory.
 *
 * @Param:      - IN    direction
 *                      Indicates whether to write the chain first-to-last
 *                      or last-to-first.  Helps ensure the microengines
 *                      always have access to valid information.
 */
void _ix_write_chain_to_shared(
    _ix_lkup_swlpmv4 hLpmv4,
    const _ix_trie_chain *pChain,
    _ix_lkup_direction direction)
{
    int trieChainIndex;

    if (direction == _IX_LKUP_DIRECTION_FORWARD)
    {
        for (trieChainIndex = 0;
            pChain->localTries[trieChainIndex].index != 0;
            trieChainIndex++)
        {
            _ix_write_trie_to_shared(hLpmv4,
                &pChain->localTries[trieChainIndex]);
        }
    }
    else
    {
        /*
         * Write from last to first.  Find the end and work backward.
         */
        for (trieChainIndex = 0;
            pChain->localTries[trieChainIndex+1].index != 0;
            trieChainIndex++);

        for (; trieChainIndex >= 0; trieChainIndex--)
        {
            _ix_write_trie_to_shared(hLpmv4,
                &pChain->localTries[trieChainIndex]);
        }
    }
}

/*
 * NAME: _ix_lookup_local
 *
 * DESCRIPTION: Searches a local trie for the data indicated by the portion
 * of the IP address and network mask provided.  Retrieves the trie info
 * for the local trie if it exists.
 *
 * @Param:      - IN    hLpmv4
 *                      Handle to the SW LPM v4.
 *
 * @Param:      - IN    ipNibble
 *                      The portion of the IP address to search for in this
 *                      trie.
 *
 * @Param:      - IN    maskNibble
 *                      The portion of the net mask to search for in this
 *                      trie.
 *
 * @Param:      - INOUT pLocal
 *                      Pointer to the local data to be searched.  If the trie
 *                      info was not already found, it will be as part of the
 *                      search.
 *
 * @Param:      - OUT   pData
 *                      Pointer to the resulting data.  Zero if the data
 *                      doesn't exist, or non-zero if it does.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INTERNAL
 */
ix_error _ix_lookup_local(
    _ix_lkup_swlpmv4 hLpmv4,
    ix_uint8 ipNibble,
    ix_uint8 maskNibble,
    _ix_local_trie *pLocal,
    ix_uint16 *pData)
{
    _ix_trie_info *pTrieInfo;
    ix_uint16 existingData;

    IX_ERROR_CR(_ix_get_trie_info(hLpmv4, pLocal));

    pTrieInfo = pLocal->pTrieInfo;
    if (pTrieInfo != NULL)
    {
        ix_uint16 *pPrefixEntry =
            _ix_get_trie_prefix_entry(pLocal, maskNibble, ipNibble);

        /*
         * pPrefixEntry == NULL means there is no prefix entry.  Examine
         * the prefix4DataExists field instead.
         */
        if (pPrefixEntry == NULL)
        {
            existingData = 0;
            if ((pTrieInfo->prefix4DataExists & (1 << ipNibble)) != 0)
            {
                existingData = _IX_TRIE_DATA(pLocal->values[ipNibble]);
            }
        }
        else
        {
            existingData = *pPrefixEntry;
        }
    }
    else
    {
        /*
         * Because there isn't a trie info structure, examine the values
         * themselves to determine the existing data.  Across the range of
         * entries, if any are different from any others, this particular entry
         * doesn't exist -- otherwise there would have been a trie info
         * structure to disabiguate the entries.
         */

        ix_uint8 currEntry;
        ix_uint8 lastEntry;
        currEntry = ipNibble & maskNibble;

        /*
         * Seed the 'existingData' field with the first entry.
         */
        existingData = _IX_TRIE_DATA(pLocal->values[currEntry]);

        /*
         * Look at all other entries; set existingData to zero and exit if any
         * are different.
         */
        lastEntry = _IX_GET_LAST_ENTRY(ipNibble, maskNibble);
        for (currEntry = ipNibble; currEntry <= lastEntry; currEntry++)
        {
            if (_IX_TRIE_DATA(pLocal->values[currEntry]) != existingData)
            {
                *pData = 0;
                return IX_SUCCESS;
            }
        }

        /*
         * At this point we know that data is consistent across the range of
         * values the targeted route covers.  Now see if that is because this
         * particular route exists, or because a less specific route exists.
         *
         * This is accomplished by looking at the 'peers' of the specified
         * ipNibble/maskNibble.  E.g. if ipNibble == 0x0 and maskNibble == 
         * 0xF, the single peer is 0x1.  If ipNibble == 0xA, and
         * maskNibble == 0xE, the two peers would be 0x8 and 0x9.
         *
         * The search is very similar to the previous one, but with a
         * different starting point.  If all those values are the same as
         * existingData, then our route doesn't exist.
         */
        currEntry = _IX_GET_PEER(ipNibble, maskNibble);
        lastEntry = _IX_GET_LAST_ENTRY(currEntry, maskNibble);
        for (; currEntry <= lastEntry; currEntry++)
        {
            /*
             * If any are different, then our previous search told us we
             * found the correct route.  Return it.
             */
            if (_IX_TRIE_DATA(pLocal->values[currEntry]) != existingData)
            {
                *pData = existingData;
                return IX_SUCCESS;
            }
        }

        /*
         * All the entries were the same: our route doesn't exist.
         */
        existingData = 0;
    }

    *pData = existingData;
    return IX_SUCCESS;
}

/*
 * NAME: _ix_get_chain_len
 *
 * DESCRIPTION: Returns the length of a chain required to hold a key with the
 * specified mask length.
 *
 * @Param:      - IN    netMaskLen
 *                      Length of the net mask used to determine how many
 *                      tries are required.  Value from 1 to 32.
 *
 * @Returns:
 * The length of a chain required to hold a key with the specified mask
 * length.
 */
ix_uint8 _ix_get_chain_len(
    ix_uint8 netMaskLen)
{
    const ix_uint8 chainLenForMaskLen[8] =
    {
        0, /* netMaskLen = { 1 -  4} --> hi256 table and no tries    */
        0, /* netMaskLen = { 5 -  8} --> hi256 table and no tries    */
        1, /* netMaskLen = { 9 - 12} --> hi256 table and one trie    */
        2, /* netMaskLen = {13 - 16} --> hi256 table and two tries   */
        1, /* netMaskLen = {17 - 20} --> hi64k table and one trie    */
        2, /* netMaskLen = {21 - 24} --> hi64k table and two tries   */
        3, /* netMaskLen = {25 - 28} --> hi64k table and three tries */
        4  /* netMaskLen = {29 - 32} --> hi64k table and four tries  */
    };

    return chainLenForMaskLen[(netMaskLen-1)>>2];
}

/*
 * NAME: _ix_get_route_chain
 *
 * DESCRIPTION: Copies all the tries involved in storing a route from shared
 * memory into local memory.  Returns the number of tries copied and the data
 * the route refers to.
 *
 * @Param:      - IN    hLpmv4
 *                      Handle to the SW LPM v4.
 *
 * @Param:      - INOUT pChain
 *                      Pointer to the chain to popluate. On input, the
 *                      ipAddr and netMaskLen are expected to be valid, and
 *                      all local tries should be empty. On output, as many
 *                      tries are filled as possible (based on ipAddr,
 *                      netMaskLen, and the contents of shared memory).
 *
 * @Param:      - OUT   pChainLen
 *                      The number of valid local tries.
 *
 * @Param:      - OUT   pExistingData
 *                      The data the requested route contains.
 *
 * Error Codes:
 * IX_LKUP_ERROR_NOTFOUND
 * IX_LKUP_ERROR_INTERNAL
 */
ix_error _ix_get_route_chain(
    _ix_lkup_swlpmv4 hLpmv4,
    _ix_trie_chain *pChain,
    ix_uint8 *pChainLen,
    ix_uint16 *pExistingData)
{
    ix_uint16 startOfNew = _IX_CHAIN_LEN;
    ix_uint16 lastExisting;
    ix_uint16 existingData = 0;
    int bitPos = 0;
    ix_uint8 ipNibble;
    ix_uint8 maskNibble;

    /*
     * Each route is kept in a series of tries in shared memory.  We need to
     * examine the set of tries involved in storing this particular route.
     * Copy the contents of these tries from shared memory to local memory,
     * which is cached.  This makes examining the contents of the memory more
     * efficient.
     */
    IX_ERROR_CR(
        _ix_copy_existing_from_base(hLpmv4, pChain, &startOfNew, &bitPos));

    /*
     * If the route actually exists in memory, then the startOfNew and bitPos
     * values have predictable values.  If they don't match the predicted
     * values, the route must not exist and we can return an error immediately.
     *
     * The length of the network mask is used to predict the expected value
     * for 'startOfNew'.
     */
    if (startOfNew != _ix_get_chain_len(pChain->netMaskLen))
    {
        return _IX_ROUTE_NOTFOUND(pChain);
    }

#ifdef OPTIONAL
    if (startOfNew == 0)
    {
        return IX_ERROR_LOCAL(IX_LKUP_ERROR_INTERNAL,
            ("empty chain"));
    }
#endif

    /*
     * Determine whether the route exists by examining the last trie in
     * the chain.
     */
    lastExisting = startOfNew - 1;
    bitPos += 4;
    ipNibble = _IX_GET_NIBBLE(pChain->ipAddr, bitPos);
    maskNibble = _IX_GET_NIBBLE(pChain->netMask, bitPos);

    IX_ERROR_CR(
        _ix_lookup_local(hLpmv4, ipNibble, maskNibble,
            &pChain->localTries[lastExisting], &existingData));

    if (existingData == 0)
    {
        return _IX_ROUTE_NOTFOUND(pChain);
    }

    if (pExistingData != NULL)
    {
        *pExistingData = existingData;
    }

    if (pChainLen != NULL)
    {
        *pChainLen = startOfNew;
    }

    return IX_SUCCESS;
}

#if defined(LKUP_DEBUG)
/*
 * NAME: _ix_dump_local
 *
 * DESCRIPTION: Debug function to dump the contents of a local trie
 *
 * @Param:      - IN    pLocal
 *                      Pointer to the local trie to dump
 */
void _ix_dump_local(
    const _ix_local_trie *pLocal)
{
    int j,k;
    ix_ossl_message_log("%d: ", pLocal->index);
    if (pLocal->trieInfoLookedUp)
    {
        _ix_trie_info *pTrieInfo = pLocal->pTrieInfo;
        if (pTrieInfo == NULL)
        {
            ix_ossl_message_log("this trie has no trieInfo\n");
        }
        else
        {
            char prefix3Text[255];
            ix_ossl_message_log("trieInfo = \n");
            ix_ossl_message_log("   prefix1[] = {0x%04X, 0x%04X}\n",
                    pTrieInfo->prefix1[0],
                    pTrieInfo->prefix1[1]);
            ix_ossl_message_log("   prefix2[] = {0x%04X, 0x%04X, 0x%04X, 0x%04X}\n",
                    pTrieInfo->prefix2[0],
                    pTrieInfo->prefix2[1],
                    pTrieInfo->prefix2[2],
                    pTrieInfo->prefix2[3]);
            sprintf(prefix3Text,
                    "   prefix3[] = {0x%04X, 0x%04X, 0x%04X, 0x%04X, "
                                    "0x%04X, 0x%04X, 0x%04X, 0x%04X}\n",
                    pTrieInfo->prefix3[0],
                    pTrieInfo->prefix3[1],
                    pTrieInfo->prefix3[2],
                    pTrieInfo->prefix3[3],
                    pTrieInfo->prefix3[4],
                    pTrieInfo->prefix3[5],
                    pTrieInfo->prefix3[6],
                    pTrieInfo->prefix3[7]);
            ix_ossl_message_log(prefix3Text);
            ix_ossl_message_log("   prefix4DataExists = 0x%04X\n",
                    pTrieInfo->prefix4DataExists);
        }
    }
    else
    {
        ix_ossl_message_log("trieInfo has not been retrieved\n");
    }
    ix_ossl_message_log("  ipPortion = 0x%X, maskPortion = 0x%X, modified = 0x%04X\n  ",
        pLocal->ipPortion, pLocal->maskPortion, pLocal->modified);
    for (j = 0; j < 4; j++)
    {
        for (k = 0; k < 4; k++)
        {
            ix_ossl_message_log("0x%08lX ", pLocal->values[j*4 + k]);
        }
        ix_ossl_message_log("\n  ");
    }
    ix_ossl_message_log("\n");
}

/*
 * NAME: _ix_dump_chain
 *
 * DESCRIPTION: Debug function to dump the contents of a trie chain.
 *
 * @Param:      - IN    hLpmv4
 *                      Handle to the SW LPM v4.
 *
 * @Param:      - IN    pChain
 *                      Pointer to the chain to dump.
 */
void _ix_dump_chain(
    _ix_lkup_swlpmv4 hLpmv4,
    const _ix_trie_chain *pChain)
{
    int i;
    ix_ossl_message_log("ipAddr = 0x%08lX, netMask = 0x%08lX, data = 0x%04lX\n",
            pChain->ipAddr, pChain->netMask, pChain->data);

    if (_IX_USE_HI64K_TABLE(pChain->netMaskLen))
    {
        ix_ossl_message_log("hi64kTable[0x%04lX] = 0x%08lX\n",
            pChain->ipAddr >> 16, _IX_GET_HI64K_ENTRY(hLpmv4, pChain->ipAddr));
    }
    else
    {
        ix_ossl_message_log("hi256Table[0x%04lX] = 0x%08lX\n",
            pChain->ipAddr >> 24, _IX_GET_HI256_ENTRY(hLpmv4, pChain->ipAddr));
    }

    for (i = 0; pChain->localTries[i].index != 0; i++)
    {
        _ix_dump_local(&pChain->localTries[i]);
    }
}
#endif /* LKUP_DEBUG */

/*
 * NAME: _ix_cleanup_chain
 *
 * DESCRIPTION: Frees all tries in a chain that are no longer needed.  Starts
 * from the bottom of the chain.  If the trie is empty, removes it from the
 * previous trie in the chain.  Continues up the chain.
 *
 * @Param:      - IN    hLpmv4
 *                      Handle to the SW LPM v4.
 *
 * @Param:      - INOUT pChain
 *                      Pointer to pre-allocated memory. On input, all members
 *                      of the chain are expected to be valid (though not all
 *                      tries need to have their trie info looked up).  This
 *                      function will look up the trie info for tries when
 *                      necessary.
 *
 * @Param:      - OUT   pFirstToFree
 *                      The index into the trie chain of the first local trie
 *                      that needs to be freed.  This function does not free
 *                      the trie because that should happen only after the
 *                      trie's contents are written to shared memory.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INTERNAL
 */
ix_error _ix_cleanup_chain(
    _ix_lkup_swlpmv4 hLpmv4,
    _ix_trie_chain *pChain,
    ix_uint16 *pFirstToFree)
{
    int trieIndex;
    _ix_local_trie *pLocal;
    *pFirstToFree = _IX_CHAIN_LEN;

    /*
     * Examine them all from last to first.
     */
    for (trieIndex = _IX_CHAIN_LEN-1; trieIndex >= 0; trieIndex--)
    {
        int entry;
        pLocal = &pChain->localTries[trieIndex];
        if (pLocal->index == 0)
        {
            /*
             * No need to study it if it doesn't exist.  Look at the previous
             * one in the list.
             */
            continue;
        }

        IX_ERROR_CR(_ix_get_trie_info(hLpmv4, pLocal));

        /*
         * Look at all the values.  If any are non-zero, no more cleanup
         * can be done.  Return immediately.
         */
        for (entry = 0; entry < 16; entry++)
        {
            if (pLocal->values[entry] != 0)
            {
                return IX_SUCCESS;
            }
        }

        /*
         * The trie is empty, so there is no reason to keep it here. 
         * Remove it from the previous trie in the chain.  Don't change
         * anything for the first trie in the chain because, since it is
         * linked to from the high table, it must be handled specially.
         */
        if (trieIndex != 0)
        {
            _ix_local_trie *pParent = &pChain->localTries[trieIndex-1];
            ix_uint32 data =
                _IX_TRIE_DATA(pParent->values[pLocal->ipPortion]);
            ix_uint32 newValue = _IX_TRIE_PACK_LOCAL(0, data);
            pParent->values[pLocal->ipPortion] = newValue;
            pParent->modified |= 1 << pLocal->ipPortion;
        }

        /*
         * Note for the caller that this trie should be freed for future
         * use.
         */
        *pFirstToFree = trieIndex;
    }

    return IX_SUCCESS;
}

/*
 * NAME: _ix_initialize_chain
 *
 * DESCRIPTION: Initialize an _ix_trie_chain structure based on the requested
 * key, mask and data.  This function can be called either with a pointer to
 * data, or without (i.e. pData is NULL).  If pData is non-NULL, the contents
 * of the array are expected to be valid.  Because this implementation cannot
 * store the data value zero, _ix_initialize_chain() returns an error if
 * *pData == zero.
 *
 * @Param:      - IN    pKey
 *                      Pointer to the byte array containing the key (32 bits
 *                      for this implementation.)
 *
 * @Param:      - IN    pMask
 *                      Pointer to the byte array containing the mask (32 bits
 *                      for this implementation.)
 *
 * @Param:      - IN    pData
 *                      Pointer to the byte array containing the data (16
 *                      bits for this implementation). If NULL, pChain->data
 *                      is initialized to zero.
 *
 * @Param:      - OUT   pChain
 *                      Pointer to the structure to initialize. All contents
 *                      of the structure are overwritten.  On output, ipAddr
 *                      is initialized with the contents of pKey, netMask
 *                      and netMaskLen are initialized based on pMask, and
 *                      data is initialized with the contents of pData (or
 *                      zero if pData is NULL).  All local tries are
 *                      initialized to zero.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 */
ix_error _ix_initialize_chain(
    const ix_uint8 *pKey,
    const ix_uint8 *pMask,
    const ix_uint8 *pData,
    _ix_trie_chain *pChain)
{
    if (pKey == NULL)
    {
        return ix_error_new(0, IX_LKUP_ERROR_INVALID, IX_ERROR_LEVEL_WARNING,
            ("pKey must not be NULL"));
    }

    if (pMask == NULL)
    {
        return ix_error_new(0, IX_LKUP_ERROR_INVALID, IX_ERROR_LEVEL_WARNING,
            ("pMask must not be NULL"));
    }

    ix_ossl_memset(pChain, 0, sizeof(_ix_trie_chain));
    pChain->ipAddr = IX_NTOH32(*(const ix_uint32*) pKey);
    pChain->netMask = IX_NTOH32(*(const ix_uint32*) pMask);
    pChain->ipAddr &= pChain->netMask;
    IX_ERROR_CR(_ix_lkup_swlpmv4_netmask_len(pChain->netMask,
                &pChain->netMaskLen));

    if (pData != NULL)
    {
        pChain->data = IX_NTOH16(*(const ix_uint16*) pData);

        /*
         * Because the LPM algorithm is unable to store the data value zero,
         * look for it here.
         */
        if (pChain->data == 0)
        {
            return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
                ("This implementation of the Lookup Library is unable to"
                 " store zero data"));
        }
    }
#ifdef OPTIONAL
    /*
     * Though we could initialize pChain->data to zero if pData is NULL,
     * there is no reason to because the entire chain has already been
     * initialized to zero in _ix_initialize_chain().
     */
    else
    {
        pChain->data = 0;
    }
#endif /* OPTIONAL */

    return IX_SUCCESS;
}

/*
 * NAME: _ix_append_trie_to_chain
 *
 * DESCRIPTION: Given a trie index, creates a local trie at the end of an
 * existing trie chain.  Causes the 16 32-bit values for the trie to be
 * copied from shared memory.
 *
 * @Param:      - IN    hLpmv4
 *                      Handle to the SW LPM v4.
 *
 * @Param:      - INOUT pChain
 *                      Pointer to the chain to receive the appended trie.
 *
 * @Param:      - IN    entry
 *                      Index of the trie to copy
 *
 * @Param:      - IN    bitPos
 *                      Bit position into the IP address and net mask the
 *                      trie represents.  E.g. if the trie for the 5 in
 *                      0x12345600 was being copied, the bitPos would be
 *                      12.
 *
 * @Param:      - IN    portionMask
 *                      An entity (either the hi256 table, the hi64k table, or
 *                      a trie) refers to this trie.  The maskPortion is the
 *                      part of the net mask that entity covers.  E.g.
 *                      the portionMask would be 0xFF if the trie is
 *                      referenced from the hi256 table, 0xFFFF if the trie
 *                      is referenced from the hi64k table, or 0xF if a trie.
 *
 * @Param:      - OUT   pTrieChainIndex
 *                      Index of the local trie in the chain that was filled.
 *
 * Error Codes:
 * None
 */
ix_error _ix_append_trie_to_chain(
    _ix_lkup_swlpmv4 hLpmv4,
    _ix_trie_chain *pChain,
    ix_uint32 entry,
    int bitPos,
    ix_uint32 portionMask,
    ix_uint16 *pTrieChainIndex)
{
    _ix_local_trie *pLocal = NULL;
    ix_uint16 trieChainIndex;

    /*
     * Search for the first empty local trie.  This could be optimized by
     * storing the last used index in the _ix_trie_chain structure.
     */
    for (trieChainIndex = 0;
        (trieChainIndex < _IX_CHAIN_LEN) && 
        (pChain->localTries[trieChainIndex].index != 0);
        trieChainIndex++);

#ifdef OPTIONAL
    /*
     * This can be turned on for sanity checking, but should never be
     * triggered.
     */
    if (trieChainIndex >= _IX_CHAIN_LEN)
    {
        return IX_ERROR_LOCAL(IX_LKUP_ERROR_INTERNAL,
            ("Cannot append another trie, out of links in the chain"));
    }
#endif

    pLocal = &pChain->localTries[trieChainIndex];
    pLocal->index = entry;
    pLocal->pTrieInfo = NULL;
    pLocal->trieInfoLookedUp = 0;
    pLocal->ipPortion = (pChain->ipAddr >> (bitPos + 4)) & portionMask;
    pLocal->maskPortion = (pChain->netMask >> (bitPos + 4)) & portionMask;
    IX_ERROR_CR(_ix_copy_trie_from_shared(hLpmv4, entry, pLocal->values));

    if (pTrieChainIndex)
    {
        *pTrieChainIndex = trieChainIndex;
    }

    return IX_SUCCESS;
}

/*
 * NAME: _ix_get_trie_info
 *
 * DESCRIPTION: Retrieve the trie info associated with a trie, if it exists.
 *
 * @Param:      - IN    hLpmv4
 *                      Handle to the SW LPM v4.
 *
 * @Param:      - INOUT pLocal
 *                      Pointer to the local trie to get the trie info for. On
 *                      input, trieInfoLookedUp and index must be accurate. On
 *                      output, trieInfoLookedUp is 1 and pTrieInfo either
 *                      points to the trie info structure for the given trie if
 *                      it exists, or NULL if it doesn't.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INTERNAL
 */
ix_error _ix_get_trie_info(
    _ix_lkup_swlpmv4 hLpmv4,
    _ix_local_trie *pLocal
    )
{
    int exists;

    if (pLocal->trieInfoLookedUp != 0)
    {
        /*
         * Already have the trie info structure.
         */
        return IX_SUCCESS;
    }

    IX_ERROR_CRT(ix_cc_hash_table_lookup(hLpmv4->hHash, pLocal->index,
                &exists, (void**) &pLocal->pTrieInfo),
            IX_LKUP_ERROR_INTERNAL, 0);
    if (!exists)
    {
        pLocal->pTrieInfo = NULL;
    }

    /*
     * Note that we have searched for the trie info now (even if we didn't
     * find one.
     */
    pLocal->trieInfoLookedUp = 1;

    return IX_SUCCESS;
}

/*
 * NAME: _ix_free_trie_info
 *
 * DESCRIPTION: Frees the trie info structure associated with the given local
 * trie.
 *
 * @Param:      - IN    hLpmv4
 *                      Handle to the SW LPM v4.
 *
 * @Param:      - INOUT pLocal
 *                      Pointer to local trie that should have its info
 *                      structure freed.  On output, pTrieInfo is NULL.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INTERNAL
 */
ix_error _ix_free_trie_info(
    _ix_lkup_swlpmv4 hLpmv4,
    _ix_local_trie *pLocal)
{
    /*
     * In order to free the trie info, we need to know its index. Calculate it
     * by determining how far from the base address this one is.
     */
    unsigned int trieInfoIndex = pLocal->pTrieInfo - hLpmv4->pTrieInfoBase;

    IX_ERROR_CRT(ix_cc_hash_table_delete_entry(hLpmv4->hHash, pLocal->index),
        IX_LKUP_ERROR_INTERNAL, 0);
    pLocal->pTrieInfo = NULL;

    IX_ERROR_CR(ix_cc_index_pool_free(hLpmv4->hTrieInfoIndexes,
        trieInfoIndex));
    return 0;
}

/*
 * NAME: _ix_copy_trie_from_shared
 *
 * DESCRIPTION: Copies the 16 32-bit values (data and next trie) from
 * a trie in shared memory into a local buffer.
 *
 * @Param:      - IN    hLpmv4
 *                      Handle to the SW LPM v4.
 *
 * @Param:      - IN    trieIndex
 *                      Index of the shared trie to copy
 *
 * @Param:      - OUT   pValues
 *                      Array of 16 ix_uint32s where the results are placed
 *
 * Error Codes:
 * None
 */
ix_error _ix_copy_trie_from_shared(
    _ix_lkup_swlpmv4 hLpmv4,
    ix_uint32 trieIndex,
    ix_uint32 *pValues)
{
    ix_uint8 offset;
    for (offset = 0; offset < 16; offset++)
    {
        pValues[offset] = _IX_GET_TRIE_ENTRY(hLpmv4, trieIndex, offset);
    }
    return IX_SUCCESS;
}

/*
 * NAME: _ix_make_trie_info
 *
 * DESCRIPTION: Allocates a trie info structure and fills it according to the
 * contents of a local trie.
 *
 * @Param:      - IN    hLpmv4
 *                      Handle to the SW LPM v4.
 *
 * @Param:      - INOUT pLocal
 *                      Pointer to the local trie that requires a trie info
 *                      structure.  On input, the index and the contents of
 *                      the values[] array must be valid.  On output,
 *                      trieInfoLookedUp is set to 1 and pTrieInfo points at
 *                      the newly allocated and initialized structure.
 *
 * Error Codes:
 * IX_LKUP_ERROR_RESOURCES
 */
ix_error _ix_make_trie_info(
    _ix_lkup_swlpmv4 hLpmv4,
    _ix_local_trie *pLocal)
{
    ix_uint32 trieInfoIndex = 0;
    ix_error err;
    ix_uint8 prefixIndex;
    ix_uint8 ipNibble;
    ix_uint8 maskNibble;
    _ix_trie_info *pTrieInfo;

    /*
     * Get an unused trie info structure.
     */
    IX_ERROR_CR(ix_cc_index_pool_alloc(hLpmv4->hTrieInfoIndexes,
        &trieInfoIndex));

    pTrieInfo = &hLpmv4->pTrieInfoBase[trieInfoIndex];

    /*
     * For the purposes of the _ix_lookup_local() calls below, it needs to
     * think a trie info doesn't exist.  Assign it later.
     */
    pLocal->pTrieInfo = NULL;
    pLocal->trieInfoLookedUp = 1;

    /*
     * Add it to the hash table, so it can be searched later.
     */
    err = ix_cc_hash_table_add_entry(hLpmv4->hHash,
            pLocal->index, (void*) pTrieInfo);
    if (err != IX_SUCCESS)
    {
        if (IX_ERROR_GET_CODE(err) == IX_CC_ERROR_FULL)
        {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
            err = IX_ERROR_CHAIN(err, IX_LKUP_ERROR_RESOURCES, 0,
                ("Try increasing _IX_TRIE_ROUTE_RATIO"));
#else
            err = IX_ERROR_CHAIN(err, IX_LKUP_ERROR_RESOURCES, 0,
                ("Try increasing _IX_TRIE_ROUTE_RATIO(%2.1f)",
                 _IX_TRIE_ROUTE_RATIO));
#endif
        }
        goto cleanStruct;
    }

    /*
     * Decode the data from the local trie and fill the trie info structure.
     */
    for (prefixIndex = 0; prefixIndex < 2; prefixIndex++)
    {
        ipNibble = prefixIndex << 3;
        maskNibble = 0x8;
        /*
         * Use _ix_lookup_local to find out if any 1-bit prefixes exist.
         */
        IX_ERROR_CG(_ix_lookup_local(hLpmv4, ipNibble, maskNibble, pLocal,
                    &pTrieInfo->prefix1[prefixIndex]),
            err, cleanStruct);
    }

    for (prefixIndex = 0; prefixIndex < 4; prefixIndex++)
    {
        ipNibble = prefixIndex << 2;
        maskNibble = 0xC;
        /*
         * Use _ix_lookup_local to find out if any 2-bit prefixes exist.
         */
        IX_ERROR_CG(_ix_lookup_local(hLpmv4, ipNibble, maskNibble,
                    pLocal, &pTrieInfo->prefix2[prefixIndex]),
            err, cleanStruct);
    }

    for (prefixIndex = 0; prefixIndex < 8; prefixIndex++)
    {
        ipNibble = prefixIndex << 1;
        maskNibble = 0xE;
        /*
         * Use _ix_lookup_local to find out if any 3-bit prefixes exist.
         */
        IX_ERROR_CG(_ix_lookup_local(hLpmv4, ipNibble, maskNibble,
                    pLocal, &pTrieInfo->prefix3[prefixIndex]),
            err, cleanStruct);
    }

    /*
     * If any value is set, but there is no entry in the prefix arrays
     * covering that entry, it must be a 4-bit entry.  Set the flag for that
     * position.
     */
    for (prefixIndex = 0; prefixIndex < 16; prefixIndex++)
    {
        if ((pTrieInfo->prefix1[prefixIndex >> 3] == 0) &&
            (pTrieInfo->prefix2[prefixIndex >> 2] == 0) &&
            (pTrieInfo->prefix3[prefixIndex >> 1] == 0) &&
            (_IX_TRIE_DATA(pLocal->values[prefixIndex]) != 0))
        {
            pTrieInfo->prefix4DataExists |= (1 << prefixIndex);
        }
    }

    pLocal->pTrieInfo = pTrieInfo;

    return IX_SUCCESS;

cleanStruct:
    (void) ix_cc_index_pool_free(hLpmv4->hTrieInfoIndexes, trieInfoIndex);
    return err;
}

/*
 * NAME: _ix_get_trie_prefix_entry
 *
 * DESCRIPTION: Retrieves the pointer to the element in one of the prefix
 * arrays, as determined by the netMaskLen and ipByte.
 *
 * @Param:      - IN    hLpmv4
 *                      The handle to the table.
 *
 * @Param:      - IN    maskNibble
 *                      The portion of the net mask to search for in this
 *                      trie.
 *
 * @Param:      - IN    ipNibble
 *                      The portion of the IP address to search for in this
 *                      trie.
 *
 * @Returns:
 * Pointer into one of the trie info prefix arrays.
 */
ix_uint16 *_ix_get_trie_prefix_entry(
    _ix_local_trie *pLocal,
    ix_uint8 maskNibble,
    ix_uint8 ipNibble)
{
    switch (maskNibble)
    {
        case 0x8:
            return &pLocal->pTrieInfo->prefix1[ipNibble >> 3];

        case 0xC:
            return &pLocal->pTrieInfo->prefix2[ipNibble >> 2];

        case 0xE:
            return &pLocal->pTrieInfo->prefix3[ipNibble >> 1];

        default:
            return NULL;
    }
}

/*
 * NAME: _ix_get_best_fit_data
 *
 * DESCRIPTION: Looks at one element in each prefix array, from most specific
 * to least (i.e. prefix4DataExists, prefix3, prefix2, prefix1), until a
 * non-zero value is found.  Returns the value and mask corresponding to the
 * position the value was found in.
 *
 * This function should be called after an element is removed from a trie in
 * order to back-fill the data with the most appropriate remaining data.
 *
 * @Param:      - IN    hLpmv4
 *                      The handle to the table.
 *
 * @Param:      - IN    entryIndex
 *                      Value from zero to 15, indicates the position in the
 *                      trie.  This value is used to find an index in each of
 *                      the prefix arrays.
 *
 * @Param:      - OUT   pNewData
 *                      The data that was found, or zero if none.
 *
 * @Param:      - OUT   pMask
 *                      The mask indicating the position where the data was
 *                      found, or zero if none.
 */
ix_uint16 _ix_get_best_fit_data(
    _ix_local_trie *pLocal,
    ix_uint8 entryIndex)
{
    /*
     * If the trie has trie info, look up to see which data is the best fit.
     */
    if (pLocal->pTrieInfo != NULL)
    {
        _ix_trie_info *pTrieInfo = pLocal->pTrieInfo;
        ix_uint16 data;

        /*
         * If a 4-bit prefix exists, it is already stored in the trie's data.
         */
        if (pTrieInfo->prefix4DataExists & (1 << entryIndex))
        {
            return _IX_TRIE_DATA(pLocal->values[entryIndex]);
        }

        /*
         * If a prefix3 entry exists, use it.
         */
        data = pTrieInfo->prefix3[entryIndex >> 1];
        if (data != 0)
        {
            return data;
        }

        /*
         * If a prefix2 entry exists, use it.
         */
        data = pTrieInfo->prefix2[entryIndex >> 2];
        if (data != 0)
        {
            return data;
        }

        /*
         * None of the others exist, use the prefix1 entry (which might be
         * zero).
         */
        return pTrieInfo->prefix1[entryIndex >> 3];
    }
    else
    {
        /*
         * If we don't have trie info, we don't have an overlap.  Zero is
         * always the right answer in this situation.
         */
        return 0;
    }
}

/*
 * NAME ix_lkup_swlpmv4_netmask_len
 *
 * DESCRIPTION: Given a network mask (e.g. 0xFFFF0000), return its length
 * (e.g. 16).
 */
ix_error _ix_lkup_swlpmv4_netmask_len(ix_uint32 netMask, ix_uint8 *pNetMaskLen)
{
    int bit;
    ix_uint32 mask = netMask;

    for (bit = 0; bit < 32; bit++)
    {
        if (mask & 1)
        {
            break;
        }
        mask >>= 1;
    }

    /*
     * If bit == 32, return 0.  One way is to remove the upper bits.
     */
    bit &= 0x1F;

    *pNetMaskLen = 32 - bit;

#if defined(IX_DEBUG)
    if (netMask != (0xFFFFFFFFUL << bit))
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("The netMask 0x%08lX is invalid", netMask));
    }
#endif /* IX_DEBUG */

    return IX_SUCCESS;
}

