/**
 * ===========================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      swlpmv4_find.c
 *
 * = DESCRIPTION
 *      Helper functions for enumerating the LPM tables.  This is part of the
 *      Lookup API's Software LPM algorithm for IPv4.
 *
 * ===========================================================================
 */

#define IX_ERROR_FILE_IDENT                                             \
        "$Id: swlpmv4_find.c,v 1.13 2003/11/06 03:04:59 rranjeet Exp $"

#include "ix_cc_error.h"
#include "lkup/internal/internal_swlpmv4.h"
#include "ix_ossl.h"

/*
 * NAME: _ix_search_trie_chain
 *
 * DESCRIPTION: Creates a trie chain using the iteratorKey and
 * iteratorMaskLen as the starting point.  Searches the chain for the next
 * route, as follows:
 * 1) Attempts to find a deeper trie than the current one.
 * 2) Once the deepest trie is found (a "local maximum" depth) the bottom
 *    trie is searched for data.  Any data found is returned, and the location
 *    is noted for subsequent calls.
 * 3) When the next call comes in, the current trie is searched for more data.
 *    When no more data is available in the current trie, the search backs up
 *    to the parent trie, and the search continues with step 1, attempting to
 *    find a different deeper trie.
 *
 * @Param:      - IN    hLpmv4
 *                      Handle to the SW LPM v4.
 *
 * @Param:      - IN    trieIndex
 *                      Index of the trie to use as the base of the chain.
 *
 * @Param:      - IN    fromHi64kTable
 *                      Indicates whether the chain is from the hi64kTable (1)
 *                      or from the hi256Table (0).
 *
 * @Param:      - IN    continueSearch
 *                      Indicates whether to start a new search (0) or
 *                      continue a previous search (1).
 *
 * @Param:      - OUT   pFound
 *                      Indicates whether a route was found in the trie chain.
 *
 * @Param:      - OUT   pKey
 *                      Pointer to a byte array where the key is stored.
 *
 * @Param:      - OUT   pMask
 *                      Pointer to a byte array where the mask is stored.
 *
 * @Param:      - OUT   pData
 *                      Pointer to a byte array where the data is stored.
 */
ix_error _ix_search_trie_chain(
    _ix_lkup_swlpmv4 hLpmv4,
    ix_uint32 trieIndex, 		
    int fromHi64kTable,
    int continueSearch,
    int *pFound,
    ix_uint8 *pKey,
    ix_uint8 *pMask,
    ix_uint8 *pData)
{
    _ix_trie_chain trieChain;
    _ix_local_trie *pLocal;
    ix_uint8 ipNibble;
    ix_uint8 maskNibble;
    ix_uint16 startOfNew;
    int bitPos = fromHi64kTable ? 12 : 20;
    ix_uint32 coveredMask = fromHi64kTable ? 0xFFFF : 0xFF;
    ix_uint16 data;
    ix_uint32 nextTrieIndex; 
    int lastValidTrieIndex;

    /*
     * Don't use _ix_initialize_chain() because it does a lot of things we
     * don't need.
     */
    ix_ossl_memset(&trieChain, 0, sizeof(_ix_trie_chain));
    trieChain.ipAddr = hLpmv4->iteratorKey;
    trieChain.netMaskLen = hLpmv4->iteratorMaskLen;
    trieChain.netMask = 0xFFFFFFFFUL << (32 - trieChain.netMaskLen);

    /*
     * Get as much of the chain as we need in order to look at the last
     * trie examined.  We might not need the last trie, or we might need
     * to get more later, but this is a good place to start.
     */
    IX_ERROR_CR(
        _ix_copy_existing_tries(hLpmv4, trieIndex, coveredMask, &bitPos,
            &trieChain, &startOfNew));
#ifdef OPTIONAL
    if (startOfNew == 0)
    {
        return IX_ERROR_LOCAL(IX_LKUP_ERROR_INTERNAL,
            ("startOfNew should not have been zero"));
    }
#endif

    lastValidTrieIndex = startOfNew - 1;
    pLocal = &trieChain.localTries[lastValidTrieIndex];

    /*
     * _ix_copy_existing_tries() reported the bitPos of the first empty entry.
     * We want the bitPos of the last valid one.
     */
    bitPos += 4;

    while (1)
    {
        /*
         * Prepare the ipNibble and maskNibble if they're starting points.  If
         * they're not the starting points, _ix_search_trie() treats them as
         * only output parameters.
         */
        if (continueSearch)
        {
            ipNibble = _IX_GET_NIBBLE(trieChain.ipAddr, bitPos);
            maskNibble = _IX_GET_NIBBLE(trieChain.netMask, bitPos);
        }
        else
        {
            ipNibble = 0;
            maskNibble = 0x8;
        }

        /*
         * If we've already dug down through this trie, don't do it again.
         */
        if (!hLpmv4->iteratorExamineData)
        {
            /*
             * Look at the chained tries for the current trie.
             */
            IX_ERROR_CR(_ix_get_next_chained_trie(pLocal, continueSearch,
                    &ipNibble, &nextTrieIndex));

            /*
             * If there is another trie to consider, copy the trie's contents
             * into the chain, after the current local copy.
             */
            while (nextTrieIndex != 0)
            {
                ix_uint16 trieChainIndex;

                /*
                 * _ix_append_trie_to_chain() assigns pLocal->ipPortion from
                 * trieChain.ipAddr, so assign ipAddr using the latest
                 * information.
                 */
                trieChain.ipAddr &= 0xFFFFFFFFUL << (bitPos + 4);
                trieChain.ipAddr |= ipNibble << bitPos;
                trieChain.netMask = 0xFFFFFFFFUL << bitPos;

                /*
                 * Going deeper into the ip address means our bit position moves
                 * down by 4.
                 */
                bitPos -= 4;

                IX_ERROR_CR(_ix_append_trie_to_chain(hLpmv4,
                    &trieChain, nextTrieIndex, bitPos,
                    0xF,            /* maskPortion */
                    &trieChainIndex));

                /*
                 * Point our helper variable at the next deeper local trie slot.
                 */
                pLocal = &trieChain.localTries[trieChainIndex];

                /*
                 * This counts as a new search.
                 */
                continueSearch = 0;

                /*
                 * Look at the chained tries for the current trie.
                 */
                IX_ERROR_CR(_ix_get_next_chained_trie(pLocal, continueSearch,
                        &ipNibble, &nextTrieIndex));
            }

            /*
             * Once we've reached the end of a chain, prevent digging down
             * again.
             */
            hLpmv4->iteratorExamineData = 1;
            
            /*
             * We'll examine the data next.  This counts as a new search.
             */
            continueSearch = 0;
        } /* if (!examineData) */

        /*
         * Examine the data in this trie
         */
        IX_ERROR_CR(_ix_search_trie_data(
            hLpmv4, pLocal, continueSearch,
            &ipNibble, &maskNibble, &data));
        if (data != 0)
        {
            int maskLen = 32 - bitPos;

            trieChain.ipAddr &= 0xFFFFFFFFUL << (bitPos + 4);
            trieChain.ipAddr |= ipNibble << bitPos;

            switch (maskNibble)
            {
                case 0x8:
                    maskLen -= 3;
                    break;

                case 0xC:
                    maskLen -= 2;
                    break;

                case 0xE:
                    maskLen -= 1;
                    break;

                case 0xF:
                /*
                 * The default case is impossible.  Combine with 0xF, which does
                 * nothing.
                 */
                default:
                    break;
            }

            /*
             * Write the results into the byte arrays -- these go back to the
             * caller.
             */
            _ix_assign_search_outputs(hLpmv4,
                trieChain.ipAddr, maskLen, data, pKey, pMask, pData);
            *pFound = 1;
            return IX_SUCCESS;
        }

        /*
         * We've exhausted the current trie (both deeper tries and the data)
         * so back off one trie and look for more links.
         */
        if (pLocal == trieChain.localTries)
        {
            /*
             * Can't back up past the beginning.  We're done searching
             * this chain.
             */
            *pFound = 0;
            return IX_SUCCESS;
        }

        /*
         * To avoid any confusion, clear out the trie we just finished.
         * Setting the index is sufficient.
         */
        pLocal->index = 0;

        /*
         * Back up to the previous trie.
         */
        pLocal--;
        bitPos += 4;

        /*
         * We're continuing a previously suspended search.
         */
        continueSearch = 1;

        /*
         * Because we're backing up by one trie, we know there's possibly
         * more sub-tries in that one.  Continue searching for chained tries
         * before switching to data.
         */
        hLpmv4->iteratorExamineData = 0;

        /*
         * Chop off the trieChain's information to match the new position.
         */
        trieChain.ipAddr &= 0xFFFFFFFFUL << bitPos;
        trieChain.netMask = 0xFFFFFFFFUL << bitPos;
    }
    
    /*
     * Unreachable
     */
}

/*
 * NAME: _ix_assign_search_outputs
 *
 * DESCRIPTION: Copies an IP address (32-bit number), net mask length (1-32),
 * and associated data (16-bit number) into the byte arrays to send to the
 * caller.  Also notes the last found route in the iterator fields.
 *
 * @Param:      - INOUT hLpmv4
 *                      Handle to the SW LPM v4. On output, the iterator
 *                      fields are updated according to the ipAddr and
 *                      maskLen.
 *
 * @Param:      - IN    ipAddr
 *                      IP address to return to the caller.
 *
 * @Param:      - IN    maskLen
 *                      Net mask length of the route that was found.
 *
 * @Param:      - IN    data
 *                      Data associated with the route.
 *
 * @Param:      - OUT   pKey
 *                      Byte array where the IP address is stored.
 *
 * @Param:      - OUT   pMask
 *                      Byte array where the network mask is stored.
 *
 * @Param:      - OUT   pData
 *                      Byte array where the data is stored.
 */
void _ix_assign_search_outputs(
    _ix_lkup_swlpmv4 hLpmv4,
    ix_uint32 ipAddr,
    int maskLen,
    ix_uint16 data,
    ix_uint8 *pKey,
    ix_uint8 *pMask,
    ix_uint8 *pData)
{
    hLpmv4->iteratorKey = ipAddr;
    hLpmv4->iteratorMaskLen = maskLen;

    *(ix_uint32*) pKey = IX_HTON32(ipAddr);
    *(ix_uint32*) pMask = IX_HTON32(0xFFFFFFFFUL << (32 - maskLen));
    *(ix_uint16*) pData = IX_HTON16(data);
}

/*
 * NAME: _ix_get_next_chained_trie
 *
 * DESCRIPTION: Given an ipNibble, retrieves the next non-zero trie index in
 * the local trie.  Reports the ipNibble where the next trie was found.  The
 * maskNibble isn't considered because it is guaranteed to be 0xF.
 *
 * If the local trie contains no chained tries, *pTrieIndex is zero.
 *
 * @Param:      - IN    pLocal
 *                      Pointer to the local trie to examine.
 *
 * @Param:      - IN    continueSearch
 *                      Indicates whether to continue an existing search
 *                      (start after *pIpNibble) or start a new search
 *                      (ignore *pIpNibble).
 *
 * @Param:      - INOUT pIpNibble
 *                      On input, *pIpNibble indicates where to start the
 *                      search when continueSearch is true.  *pIpNibble is
 *                      ignored when continueSearch is false.  On output,
 *                      pIpNibble contains nibble of the IP address where
 *                      another trie was found.
 *
 * @Param:      - OUT   pTrieIndex
 *                      The next trie to search (indexed from trieTableBase).
 */
ix_error _ix_get_next_chained_trie(
    const _ix_local_trie *pLocal,
    int continueSearch,
    ix_uint8 *pIpNibble,
    ix_uint32 *pTrieIndex)  
{
    ix_uint8 ipNibble = continueSearch ? (*pIpNibble) + 1 : 0;
    ix_uint32 trieIndex = 0;	

    /*
     * Search for the next non-zero entry in the array.
     */
    while ((trieIndex == 0) && (ipNibble <= 0xF))
    {
        trieIndex = _IX_TRIE_INDEX(pLocal->values[ipNibble++]);
    }

    /*
     * The loop increments beyond what was really found.  Report the actual
     * value.
     */
    *pIpNibble  = ipNibble - 1;
    *pTrieIndex = trieIndex;

    return IX_SUCCESS;
}

/*
 * NAME: _ix_search_trie_data
 *
 * DESCRIPTION: Searches a trie for data, and returns the first entry found.
 *
 * @Param:      - IN    hLpmv4
 *                      Handle to the SW LPM v4.
 *
 * @Param:      - IN    pLocal
 *                      Pointer to the trie to search
 *
 * @Param:      - IN    continueSearch
 *                      Indicates whether to start a new search (zero) or
 *                      continue where a previous search ended (non-zero).
 *
 * @Param:      - INOUT pIpNibble
 *                      Used as an input when continueSearch is true: on
 *                      input, *pIpNibble is the last location where data
 *                      was found.  On output, is the IP nibble of the 
 *                      data that was found.
 *
 * @Param:      - INOUT pMaskNibble
 *                      Used as an input when continueSearch is true: on
 *                      input, *pIpNibble is the mask associated with the
 *                      last data that was found. On output, is the mask
 *                      for the data that was found.
 *
 * @Param:      - OUT   pData
 *                      Retrieved data.
 */
ix_error _ix_search_trie_data(
    _ix_lkup_swlpmv4 hLpmv4,
    _ix_local_trie *pLocal,
    int continueSearch,
    ix_uint8 *pIpNibble,
    ix_uint8 *pMaskNibble,
    ix_uint16 *pData)
{
    ix_uint8 ipNibble = *pIpNibble;
    ix_uint8 maskNibble = *pMaskNibble;
    ix_uint16 data;

    /*
     * If we're continuing the search, immediately increment the ipNibble /
     * maskNibble pair.
     */
    if (continueSearch)
    {
        if (_ix_get_next_prefix(&ipNibble, &maskNibble))
        {
            *pData = 0;
            return IX_SUCCESS;
        }
    }
    else
    {
        /*
         * Since we're starting a new search, initialize to the first prefix
         * in the sequence.
         */
        ipNibble = 0;
        maskNibble = 8;
    }

    /*
     * For each prefix in the sequence, see if there is non-zero data in the
     * local trie.  Stop as soon as a value is detected.
     */
    do {
        IX_ERROR_CR(
            _ix_lookup_local(hLpmv4, ipNibble, maskNibble, pLocal, &data));
    } while ((data == 0) && (!_ix_get_next_prefix(&ipNibble, &maskNibble)));

    *pIpNibble = ipNibble;
    *pMaskNibble = maskNibble;
    *pData = data;
    return IX_SUCCESS;
}

/*
 * NAME: _ix_get_next_prefix
 *
 * DESCRIPTION: Takes in the current ipNibble and maskNibble, and returns the
 * next ip/mask nibbles in the sequence.  The sequence follows these rules:
 * 1) Start with the least-specific netMask (8)
 * 2) For each netMask, report all unique addresses
 * 3) Repeat step 2 for each valid netMask
 *
 * The sequence is as follows:
 * 0/8 8/8
 * 0/C 4/C 8/C C/C
 * 0/E 2/E 4/E 6/E 8/E A/E C/E E/E
 * 0/F 1/F 2/F 3/F 4/F 5/F 6/F 7/F 8/F 9/F A/F B/F C/F D/F E/F F/F
 *
 * @Param:      - INOUT pIpNibble
 *                      On input, the current value in the sequence; on
 *                      output, the next value in the sequence.
 *
 * @Param:      - INOUT pMaskNibble
 *                      On input, the mask associated with the current
 *                      value in the sequence; on output, the mask associated
 *                      with the next value in the sequence.
 *
 * @Returns:
 * zero if the results are good, or 1 if there is an overflow.
 */
int _ix_get_next_prefix(
    ix_uint8 *pIpNibble,
    ix_uint8 *pMaskNibble)
{
    ix_uint8 ipNibble = *pIpNibble;
    ix_uint8 maskNibble = *pMaskNibble;

    /*
     * The next ipNibble in the sequence is found by incrementing an amount
     * based on the mask.  The table below shows the increment amount for
     * each possible mask, in binary.
     * MASK  INCR
     * 1000  1000
     * 1100  0100
     * 1110  0010
     * 1111  0001
     *
     * The increment amount is the two's complement of the mask:  Toggle all
     * the bits of the mask and add one.
     */
    ipNibble += ((~maskNibble) + 1) & 0xF;

    /*
     * If the ipNibble has gone beyond the limit of 16 elements, restart
     * with the next more specific mask.
     */
    if (ipNibble > 0xF)
    {
        ipNibble = 0;

        /*
         * If the maskNibble was already 0xF, we've reached the end of the
         * sequence.
         */
        if (maskNibble == 0xF)
        {
            return 1;
        }

        /*
         * Set one bit lower.  I.e. 0x8 becomes 0xC, 0xC becomes 0xE,
         * 0xE becomes 0xF.
         */
        maskNibble |= maskNibble >> 1;
    }

    *pIpNibble = ipNibble;
    *pMaskNibble = maskNibble;
    return 0;
}

/*
 * NAME: _ix_search_hi256_table_info
 *
 * DESCRIPTION: Searches the hi256 table for routes; this function will
 * find routes with a netmask length of 1 to 8 bits.  The iteratorMaskLen
 * indicates whether to start a new search (== 0) or continue an existing
 * search (> 0).
 *
 * @Param:      - IN    hLpmv4
 *                      Handle to the SW LPM v4.
 *
 * @Param:      - OUT   pFound
 *                      Indicates whether a route was found.
 *
 * @Param:      - OUT   pKey
 *                      Byte array where the resulting key is stored.
 *
 * @Param:      - OUT   pMask
 *                      Byte array where the resulting mask is stored.
 *
 * @Param:      - OUT   pData
 *                      Byte array where the resulting data is stored.
 */
ix_error _ix_search_hi256_table_info(
    _ix_lkup_swlpmv4 hLpmv4,
    int *pFound,
    ix_uint8 *pKey,
    ix_uint8 *pMask,
    ix_uint8 *pData)
{
    ix_uint16 ipByte;
    ix_uint8 maskLen;

    /*
     * Generally, we start where we left off.  I.e., increment the IP address
     * and see if it's the one.  However if the iteratorMaskLen is zero, we're
     * starting a new search.  Look in the first slot too.
     */
    if (hLpmv4->iteratorMaskLen == 0)
    {
        /*
         * Starting over at 0x00000000/1.
         */
        maskLen = 1;
        ipByte = 0;
    }
    else
    {
        maskLen = hLpmv4->iteratorMaskLen;
        ipByte = (hLpmv4->iteratorKey >> 24) & 0xFF;
        ipByte += (0x00FF >> maskLen) + 1;
    }

    for (; maskLen < 8; maskLen++)
    {
        for (; ipByte < 0xFF; ipByte += (0x00FF >> maskLen) + 1)
        {
            ix_uint16 data;

            ipByte &= 0xFF << (7-maskLen);
            data = *_ix_get_hi256_prefix_entry(hLpmv4, maskLen, ipByte);
            if (data != 0)
            {
                /*
                 * We found a real entry!  Report it.
                 */
                *pFound = 1;
                _ix_assign_search_outputs(hLpmv4, ipByte << 24,
                    maskLen, data, pKey, pMask, pData);
                return IX_SUCCESS;
            } /* if (data != 0) */
        } /* for (ipByte < 0xFF) */
        ipByte = 0;
    } /* for (maskLen < 8) */

    *pFound = 0;
    return IX_SUCCESS;
}

