/**
 * ===========================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      swlpmv4_hi256.c
 *
 * = DESCRIPTION
 *      Helper functions for manipulating the hi256 table.  This is part of
 *      the Lookup API's Software LPM algorithm for IPv4.
 *
 * ===========================================================================
 */

#define IX_ERROR_FILE_IDENT                                             \
        "$Id: swlpmv4_hi256.c,v 1.12 2003/01/23 01:25:26 lbmiller Exp $"

#include "ix_cc_error.h"
#include "lkup/internal/internal_swlpmv4.h"

/*
 * NAME: _ix_get_hi256_entry
 *
 * DESCRIPTION: Retrieves entries directly from the hi256 table, without any
 * tries.  This should only be called when the prefix has 8 or fewer
 * significant bits.
 *
 * @Param:      - IN    hLpmv4
 *                      The handle to the table.
 *
 * @Param:      - IN    pChain
 *                      Pointer to pre-allocated memory. On input, ipAddr,
 *                      netMask, and netMaskLen are expected to be valid.
 *
 * @Param:      - OUT   pData
 *                      Contains the resulting data in a format that can be
 *                      returned to the caller.
 *
 * Error Codes:
 * IX_LKUP_ERROR_NOTFOUND
 */
ix_error _ix_get_hi256_entry(
    const _ix_lkup_swlpmv4 hLpmv4,
    const _ix_trie_chain *pChain,
    ix_uint8* pData)
{
    ix_uint8 ipByte = pChain->ipAddr >> 24;
    ix_uint8 maskByte = pChain->netMask >> 24;
    ix_uint16 prefixData = 0;

    /*
     * Because there is no prefix8[] array, handle the 8-bit entries
     * on their own.
     */
    if (maskByte == 0xFF)
    {
        return _ix_get_hi256_8bit_entry(hLpmv4, pChain, pData);
    }

    prefixData = *_ix_get_hi256_prefix_entry(hLpmv4, pChain->netMaskLen, ipByte);
    if (prefixData == 0)
    {
        return _IX_ROUTE_NOTFOUND(pChain);
    }

    *(ix_uint16*) pData = IX_HTON16(prefixData);

    return IX_SUCCESS;
}

/*
 * NAME: _ix_add_hi256_entry
 *
 * DESCRIPTION: Adds entries directly into the hi256 table, without any
 * tries.  This should only be called when the prefix has 8 or fewer
 * significant bits.
 * 
 * @Param:      - INOUT hLpmv4
 *                      Handle to the SW LPM v4.
 *
 * @Param:      - IN    update
 *                      Indicates whether to expect an existing entry or not.
 *                      Zero if no existing entry, non-zero if updating.
 *
 * @Param:      - IN    pChain
 *                      Pointer to pre-allocated memory. On input, ipAddr,
 *                      netMask, netMaskLen, and data are expected to be valid.
 *
 * Error Codes:
 * IX_LKUP_ERROR_EXISTS
 *      Returned only when adding (not updating)
 *
 * IX_LKUP_ERROR_CONFLICT
 *      Returned only when adding (not updating)
 *
 * IX_LKUP_ERROR_NOTFOUND
 *      Returned only when updating (not adding)
 */
ix_error _ix_add_hi256_entry(
    _ix_lkup_swlpmv4 hLpmv4,
    int update,
    const _ix_trie_chain *pChain)
{
    ix_uint8 ipByte = pChain->ipAddr >> 24;
    ix_uint8 maskByte = pChain->netMask >> 24;
    ix_uint16 *pPrefixData = NULL;
    ix_uint8 entryIndex;

    /*
     * Because there is no prefix8[] array, handle the 8-bit entries
     * on their own.
     */
    if (maskByte == 0xFF)
    {
        return _ix_add_hi256_8bit_entry(hLpmv4, update, pChain);
    }

    /*
     * Since the maskByte wasn't 0xFF, look in the trie info's prefix arrays.
     */
    pPrefixData = _ix_get_hi256_prefix_entry(hLpmv4, pChain->netMaskLen, ipByte);

    if (!update)
    {
        /*
         * If we're not updating, and the route already exists, return an
         * error.
         */
        if (*pPrefixData != 0)
        {
            if (*pPrefixData == pChain->data)
            {
                return _IX_ROUTE_EXISTS(EXISTS, pChain);
            }
            else
            {
                return _IX_ROUTE_EXISTS(CONFLICT, pChain);
            }
        }
    }
    else
    {
        /*
         * If we're updating and the route doesn't exist, return an error.
         */
        if (*pPrefixData == 0)
        {
            return _IX_ROUTE_NOTFOUND(pChain);
        }
    }

    /*
     * The route is ok to add.  Update the appropriate element of the prefix
     * array.
     */
    *pPrefixData = pChain->data;

    /*
     * Assign the data into shared memory.
     */
    entryIndex = ipByte;
    do {
        /*
         * Only assign a given entry if it is at least as specific as an
         * existing entry.
         */
        if (hLpmv4->hi256TrieInfo.mask[entryIndex] <= maskByte)
        {
            /*
             * Note the significance of the new route.
             */
            hLpmv4->hi256TrieInfo.mask[entryIndex] = maskByte;

            /*
             * Assign the new value to shared memory.
             */
            _IX_SET_HI256_ENTRY_DATA(hLpmv4, entryIndex, pChain->data);
        }
    } while (entryIndex++ != (ipByte | (~maskByte & 0xFF)));

    return IX_SUCCESS;
}

/*
 * NAME: _ix_del_hi256_entry
 *
 * DESCRIPTION: Removes entries of 8 or fewer bits from the hi256 table. The
 * shared memory and the hi256TrieInfo structure might be modified, depending
 * on the key being removed and other existing keys.
 *
 * @Param:      - INOUT hLpmv4
 *                      The handle to the table.
 *
 * @Param:      - IN    pChain
 *                      Pointer to pre-allocated memory. On input, ipAddr,
 *                      netMask, and netMaskLen are expected to be valid.
 *
 * Error Codes:
 * IX_LKUP_ERROR_NOTFOUND
 */
ix_error _ix_del_hi256_entry(
    _ix_lkup_swlpmv4 hLpmv4,
    const _ix_trie_chain *pChain)
{
    ix_uint8 ipByte = pChain->ipAddr >> 24;
    ix_uint8 maskByte = pChain->netMask >> 24;
    ix_uint16 *pPrefixData = NULL;
    ix_uint8 entryIndex;
    ix_uint16 data;
    ix_uint8 mask;

    /*
     * Because there is no prefix8[] array, handle the 8-bit entries
     * on their own.
     */
    if (maskByte == 0xFF)
    {
        return _ix_del_hi256_8bit_entry(hLpmv4, pChain);
    }

    pPrefixData = _ix_get_hi256_prefix_entry(hLpmv4, pChain->netMaskLen,
        ipByte);

    if (*pPrefixData == 0)
    {
        return _IX_ROUTE_NOTFOUND(pChain);
    }

    /*
     * Clear out the appropriate element in the prefix arrays.
     */
    *pPrefixData = 0;

    /*
     * Determine which entries need to be back-filled.
     */
    entryIndex = ipByte;
    do {
        /*
         * We're removing an entry that has fewer than 8 bits.  If the mask
         * value is 0xFF, it is better than what is being removed;  skip to
         * the next value.
         */
        if (hLpmv4->hi256TrieInfo.mask[entryIndex] == 0xFF)
        {
            continue;
        }

        _ix_get_best_fit_hi256_data(hLpmv4, entryIndex, &data, &mask);

        /*
         * Only write the new value if it is changing.  Don't want to hit
         * SRAM unless we have to.
         */
        if (mask != hLpmv4->hi256TrieInfo.mask[entryIndex])
        {
            /*
             * Note which prefix (if any) has its data in this location.
             */
            hLpmv4->hi256TrieInfo.mask[entryIndex] = mask;

            /*
             * Assign the new value to shared memory.
             */
            _IX_SET_HI256_ENTRY_DATA(hLpmv4, entryIndex, data);
        }
    } while (entryIndex++ != (ipByte | (~maskByte & 0xFF)));

    return IX_SUCCESS;
}

/*
 * NAME: _ix_get_hi256_8bit_entry
 *
 * DESCRIPTION: Gets the data for an 8-bit entry. Because the hi256TrieInfo
 * structure doesn't contain the information for 8-bit entries, the data is
 * retrieved directly from shared memory.
 *
 * @Param:      - IN    hLpmv4
 *                      The handle to the table.
 *
 * @Param:      - IN    pChain
 *                      Pointer to pre-allocated memory. On input, ipAddr
 *                      is assumed to be valid, and netMaskLen is assumed
 *                      to be 8.
 *
 * @Param:      - OUT   pData
 *                      Contains the resulting data in a format that can be
 *                      returned to the caller.
 *
 * Error Codes:
 * IX_LKUP_ERROR_NOTFOUND
 */
ix_error _ix_get_hi256_8bit_entry(
    const _ix_lkup_swlpmv4 hLpmv4,
    const _ix_trie_chain *pChain,
    ix_uint8 *pData)
{
    ix_uint8 ipByte = pChain->ipAddr >> 24;
    ix_uint16 data;

    /*
     * Make sure there is an entry at this position.  We check for an existing
     * entry by looking at the corresponding location in the hi256 trie info's
     * mask array.
     */
    if (hLpmv4->hi256TrieInfo.mask[ipByte] != 0xFF)
    {
        return _IX_ROUTE_NOTFOUND(pChain);
    }

    /*
     * Extracting the data first suppresses some compiler warnings when
     * optimization is turned on; otherwise, ntoh/hton create conflicting
     * variable names.
     */
    data = _IX_GET_HI256_ENTRY_DATA(hLpmv4, ipByte);
    *(ix_uint16*) pData = IX_HTON16(data);

    return IX_SUCCESS;
}

/*
 * NAME: _ix_add_hi256_8bit_entry
 *
 * DESCRIPTION: Adds an 8-bit entry to the hi256 table.  The hi256TrieInfo's
 * mask array is updated indicating the new entry exists, and the shared
 * memory is modified to reflect the changes.
 *
 * @Param:      - INOUT hLpmv4
 *                      The handle to the table.
 *
 * @Param:      - IN    update
 *                      Indicates whether to expect an existing entry or not.
 *                      Zero if no existing entry, non-zero if updating.
 *
 * @Param:      - IN    pChain
 *                      Pointer to pre-allocated memory. On input, ipAddr,
 *                      netMaskLen, and data are expected to be valid.
 *
 * Error Codes:
 * IX_LKUP_ERROR_EXISTS
 *      Returned only when adding (not updating)
 *
 * IX_LKUP_ERROR_CONFLICT
 *      Returned only when adding (not updating)
 *
 * IX_LKUP_ERROR_NOTFOUND
 *      Returned only when updating (not adding)
 */
ix_error _ix_add_hi256_8bit_entry(
    _ix_lkup_swlpmv4 hLpmv4,
    int update,
    const _ix_trie_chain *pChain)
{
    ix_uint8 ipByte = pChain->ipAddr >> 24;

    if (!update)
    {
        /*
         * Make sure there isn't already an entry at this position.  Since the
         * mask of this entry is 0xFF, we check for an existing entry by
         * looking at the corresponding location in the hi256 trie info's mask
         * array.
         */
        if (hLpmv4->hi256TrieInfo.mask[ipByte] == 0xFF)
        {
            ix_uint32 existingData = _IX_GET_HI256_ENTRY_DATA(hLpmv4, ipByte);
            if (existingData == pChain->data)
            {
                return _IX_ROUTE_EXISTS(EXISTS, pChain);
            }
            else
            {
                return _IX_ROUTE_EXISTS(CONFLICT, pChain);
            }
        }
    }
    else
    {
        /*
         * If we're doing an update, the route must already exist.
         */
        if (hLpmv4->hi256TrieInfo.mask[ipByte] != 0xFF)
        {
            return _IX_ROUTE_NOTFOUND(pChain);
        }
    }

    /*
     * Note that there is now an 8-bit route in this location.
     */
    hLpmv4->hi256TrieInfo.mask[ipByte] = 0xFF;

    /*
     * Assign the new value to shared memory.
     */
    _IX_SET_HI256_ENTRY_DATA(hLpmv4, ipByte, pChain->data);
    return IX_SUCCESS;
}

/*
 * NAME: _ix_del_hi256_8bit_entry
 *
 * DESCRIPTION: Removes an 8-bit entry from the hi256 table.  Back-fills
 * the hi256 table entry with the next-most-specific value from the
 * hi256TrieInfo structure.  Modifies the hi256TrieInfo structure's mask
 * array with the mask of the value used to back-fill.  Both the hi256TrieInfo
 * and shared memory are modified.
 *
 * @Param:      - INOUT hLpmv4
 *                      The handle to the table.
 *
 * @Param:      - IN    pChain
 *                      Pointer to pre-allocated memory. On input, ipAddr
 *                      and netMaskLen are expected to be valid.
 *
 * Error Codes:
 * IX_LKUP_ERROR_NOTFOUND
 */
ix_error _ix_del_hi256_8bit_entry(
    _ix_lkup_swlpmv4 hLpmv4,
    const _ix_trie_chain *pChain)
{
    ix_uint8 ipByte = pChain->ipAddr >> 24;
    ix_uint16 data;
    ix_uint8 mask;

    if (hLpmv4->hi256TrieInfo.mask[ipByte] != 0xFF)
    {
        return _IX_ROUTE_NOTFOUND(pChain);
    }

    _ix_get_best_fit_hi256_data(hLpmv4, ipByte, &data, &mask);

    /*
     * Note which prefix (if any) has its data in this location.
     */
    hLpmv4->hi256TrieInfo.mask[ipByte] = mask;

    /*
     * Assign the new value to shared memory.
     */
    _IX_SET_HI256_ENTRY_DATA(hLpmv4, ipByte, data);
    return IX_SUCCESS;
}

/*
 * NAME: _ix_get_best_fit_hi256_data
 *
 * DESCRIPTION: Looks at one element in each prefix array, from most specific
 * to least (i.e. prefix7, prefix6, etc.), until a non-zero value is found.
 * Returns the value and mask corresponding to the position the value was
 * found in.
 *
 * @Param:      - IN    hLpmv4
 *                      The handle to the table.
 *
 * @Param:      - IN    entryIndex
 *                      Value from zero to 255, indicates the position in the
 *                      hi256 table.  This value is used to find an index in
 *                      each of the prefix arrays.
 *
 * @Param:      - OUT   pNewData
 *                      The data that was found, or zero if none.
 *
 * @Param:      - OUT   pMask
 *                      The mask indicating the position where the data was
 *                      found, or zero if none.
 */
void _ix_get_best_fit_hi256_data(
    const _ix_lkup_swlpmv4 hLpmv4,
    ix_uint8 entryIndex,
    ix_uint16 *pNewData,
    ix_uint8 *pMask)
{
    ix_uint16 data = 0;
    /*
     * Array of pointers to each of the prefix arrays.  Allows us to iterate
     * through each array until a non-zero value is found.
     */
    const ix_uint16 *pPrefixes[] = {
        hLpmv4->hi256TrieInfo.prefix7,
        hLpmv4->hi256TrieInfo.prefix6,
        hLpmv4->hi256TrieInfo.prefix5,
        hLpmv4->hi256TrieInfo.prefix4,
        hLpmv4->hi256TrieInfo.prefix3,
        hLpmv4->hi256TrieInfo.prefix2,
        hLpmv4->hi256TrieInfo.prefix1,
        /* NULL is the end-of-list marker */
        NULL
    };
    int prefixIndex;

    /*
     * Iterate through the array of prefix array pointers, stopping at the end
     * of the list -- marked by a NULL.
     */
    for (prefixIndex = 0; pPrefixes[prefixIndex] != NULL; prefixIndex++)
    {
        /*
         * prefixIndex tells us which of the prefix arrays to look at. It also
         * tells us how much of the IP address (i.e. entryIndex) to use when
         * indexing into the array.
         *
         * For example, when prefixIndex is zero, we're looking at the prefix7
         * array.  To convert from an 8-bit IP address to an index into the
         * prefix7 array, we right shift by one.
         */
        data = (pPrefixes[prefixIndex])[entryIndex >> (prefixIndex+1)];

        /*
         * Stop as soon as we find some data.
         */
        if (data != 0)
        {
            break;
        }
    }

    /*
     * At this point we either found some data (data != 0) or we didn't
     * (data == 0).  In either case, report what we found.
     */
    *pNewData = data;

    /*
     * The mask is based on the resulting prefixIndex:
     * 0 => data found in prefix7 => mask = 0xFE
     * 1 => data found in prefix6 => mask = 0xFC
     * 2 => data found in prefix5 => mask = 0xF8
     * ...
     * 7 => data not found        => mask = 0
     */
    *pMask = 0xFF << (prefixIndex + 1);
}

/*
 * NAME: _ix_get_hi256_prefix_entry
 *
 * DESCRIPTION: Retrieves the pointer to the element in one of the prefix
 * arrays, as determined by the netMaskLen and ipByte.
 *
 * @Param:      - IN    hLpmv4
 *                      The handle to the table.
 *
 * @Param:      - IN    netMaskLen
 *                      Value from 1 to 7.  Indicates whether to return a
 *                      pointer into prefix1, prefix2, ... prefix7.
 *
 * @Param:      - IN    ipByte
 *                      Value from 0 to 255.  Indicates which position in the
 *                      chosen prefix array to return a pointer to.
 *
 * @Returns:
 * Pointer into one of the hi256TrieInfo prefix arrays.
 */
ix_uint16 *_ix_get_hi256_prefix_entry(
    const _ix_lkup_swlpmv4 hLpmv4,
    ix_uint8 netMaskLen,
    ix_uint8 ipByte)
{
    /*
     * Array of pointers to each of the prefix arrays.  
     */
    ix_uint16 *pPrefixes[] = {
        hLpmv4->hi256TrieInfo.prefix1,
        hLpmv4->hi256TrieInfo.prefix2,
        hLpmv4->hi256TrieInfo.prefix3,
        hLpmv4->hi256TrieInfo.prefix4,
        hLpmv4->hi256TrieInfo.prefix5,
        hLpmv4->hi256TrieInfo.prefix6,
        hLpmv4->hi256TrieInfo.prefix7
    };

    return &pPrefixes[netMaskLen-1][ipByte >> (8-netMaskLen)];
}
