/**
 * ===========================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      swlpmv6.c
 *
 * = DESCRIPTION
 *      Main file implementing the Software Longest Prefix Match algorithm 
 *      for IPv6.  This implementation fulfills the requirements of the
 *      Lookup Library for a single algorithm.  Clients of this algorithm
 *      should not call these functions directly, but should instead use
 *      the functions and macros provided in ix_lkup.h.
 *
 * ===========================================================================
 */

#define IX_ERROR_FILE_IDENT "$Id: swlpmv6.c,v 1.11 2003/11/24 17:04:24 ttschlue Exp $";

#include "ix_cc_error.h"
#include "ix_cc_macros.h"
#include "ix_rm.h"
#include "ix_ossl.h"
#include "ix_netmacros.h"
#include "lkup/internal/internal_swlpmv6.h"

/*
 * NAME: _ix_alloc_swlpmv6
 *
 * DESCRIPTION: Allocates memory from the heap to store data associated with
 * the _ix_lkup_swlpmv6 handle.  All memory is pre-initialized to zero, and
 * all fields except those dealing with shared memory are initialized to their
 * appropriate values.
 *
 * @Param:      - IN    arg_NumTries
 *                      Number of tries to allocate
 *
 * @Param:      - OUT   arg_phLpmv6
 *                      Handle of the newly allocated structure
 *
 * Error Codes:
 * IX_LKUP_ERROR_RESOURCES
 */
static ix_error _ix_alloc_swlpmv6(
     ix_uint32 arg_NumTries,
     _ix_lkup_swlpmv6 *arg_phLpmv6)
{
    _ix_lkup_swlpmv6 hLpmv6 = NULL;
    ix_uint32 heapSize = 0;
    ix_error err;

    /*
     * The only memory allocated on the heap in this implementation 
     * is to maintain 1-bit trie nodes as we add routes. This is done
     * at run time. 
     * 
     */
    heapSize = sizeof(struct _ix_s_lkup_swlpmv6);

    hLpmv6 = (_ix_lkup_swlpmv6) ix_ossl_malloc(heapSize);
    if (hLpmv6 == NULL)
    {
        return IX_ERROR_LOCAL(IX_LKUP_ERROR_RESOURCES,
            ("Could not allocate %d bytes from the OSSL", heapSize));
    }

    /*
     * Initialize the SW LPM's memory (including the In Use arrays) to zero.
     */
    ix_ossl_memset(hLpmv6, 0, heapSize);

    /*
     * Fill in the members of the algorithm's structure
     */
    hLpmv6->numTries = arg_NumTries;
    hLpmv6->iteratorKey = 0;
    hLpmv6->iteratorMaskLen = 0;

    /* Create a pool of indexes. This is used when we want to allocate 
     * a 8-bit trie from the "free pool". Notice that trie #0 is reserved.
     */

    IX_ERROR_CGT(ix_cc_index_pool_init(arg_NumTries,
                                       1, /* zero is reserved and not used */
                                       &hLpmv6->hTrieIndexes),
                                       err, 
                                       IX_LKUP_ERROR_RESOURCES, 
                                       0, 
                                       clean_hTrieIndexes);
    *arg_phLpmv6 = hLpmv6;
    return IX_SUCCESS;

    /* Cleanup area */

clean_hTrieIndexes:
    (void) ix_cc_index_pool_fini(hLpmv6->hTrieIndexes);
    ix_ossl_free(hLpmv6);
    return err;
}

/*
 * NAME: _ix_alloc_shared_memory
 *
 * DESCRIPTION: Allocates memory from the Resource Manager that will be
 * shared with the IPv6 microblock.
 *
 * @Param:      - IN    arg_MemType
 *                      Type of memory to allocate (SRAM, DRAM, etc.)
 *
 * @Param:      - IN    arg_DataChannel
 *                      Channel the data will be used from
 *
 * @Param:      - IN    arg_TableSize
 *                      Roughly indicates the number of routes that will
 *                      be stored.
 *
 * @Param:      - OUT   arg_ppBase
 *                      Resulting pointer to the base of the shared memory.
 *
 * Error Codes:
 * IX_LKUP_ERROR_RESOURCES
 * IX_LKUP_ERROR_INVALID
 */
static ix_error _ix_alloc_shared_memory(
        ix_memory_type arg_MemType,
        ix_uint32 arg_DataChannel,
        ix_uint32 arg_TableSize,
        ix_uint32 **arg_ppBase)
{
    ix_error err = IX_SUCCESS;
    ix_uint8 *pSharedMemory = NULL;
    ix_uint32 i;

    err = ix_rm_mem_alloc_aligned(arg_MemType, 
                                  arg_DataChannel, 
                                  arg_TableSize, 
                                  IX_MEMORY_ALIGNMENT_TYPE_VIRTUAL,
                                  3, /* shift by 3 bytes, 
                                      * gives 64-bit alignment */
                                  (void**) &pSharedMemory);
    if (err != IX_SUCCESS)
    {
        if (IX_ERROR_GET_CODE(err) == IX_RM_ERROR_ASSERTION)
        {
            err = IX_ERROR_CHAIN(err,
                IX_LKUP_ERROR_INVALID, IX_ERROR_LEVEL_WARNING,
                ("Invalid channel: %ld", arg_DataChannel));
        }
        else
        {
            err = IX_ERROR_CHAIN(err,
                IX_LKUP_ERROR_RESOURCES, IX_ERROR_LEVEL_WARNING,
                ("Could not allocate %d bytes from the Resource Manager",
                 arg_TableSize));
        }

        return err;
    }

    if (pSharedMemory == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_RESOURCES,
            ("Resource Manager returned pSharedMemory == NULL"));
    }

    /*
     * Initialize the shared memory to zero 64 bits at a time.
     */
    for (i = 0; i < arg_TableSize; i += sizeof(ix_uint64))
    {
        IX_RM_MEM_UINT64_WRITE(pSharedMemory + i, 0);
    }

    *arg_ppBase = (ix_uint32 *)pSharedMemory;
    return IX_SUCCESS;
}


/*
 * NAME: _ix_lkup_swlpmv6_create_table
 *
 * DESCRIPTION: Invoked via IX_LKUP_CREATE_TABLE().  Creates a new table
 * using a software LPM algorithm.  This function is intended to be one of
 * several that might be called by the Lookup Library's more generic table
 * creation function.  The more generic function will call
 * _ix_lkup_swlpmv6_create_table() only when the table type is LPM and when
 * the key size indicates IPv6 (Key Size == 128 bits).
 *
 * @Param:      - IN    arg_hLkup
 *                      handle obtained by calling ix_lkup_sw_init().
 *
 * @Param:      - IN    arg_pTableConf
 *                      pointer to structure containing the table's desired
 *                      configuration.  This algorithm imposes these
 *                      restrictions:
 *                      * tableType == IX_LKUP_TABLE_LPM
 *                      * keySize == 128 bits
 *                      * dataSize <= 16 bits
 *                      * dataLocation == SRAM or DRAM
 *
 * @Param:      - IN    arg_pVndrConf
 *                      vendor-specific configuration; must be NULL
 *
 * @Param:      - OUT   arg_phTable
 *                      table handle to be used in future calls to the
 *                      algorithm
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The handle was invalid, or some of the table configuration information
 *      was invalid or not supported.
 *
 * IX_LKUP_ERROR_RESOURCES
 *      Not enough resources were available to create the requested table.
 */
ix_error _ix_lkup_swlpmv6_create_table(
        ix_lkup arg_hLkup,
        ix_lkup_table_conf *arg_pTableConf,
        void *arg_pVndrConf,
        ix_lkup_table *arg_phTable
        )
{
    _ix_lkup_swlpmv6 hLpmv6 = NULL;
    ix_lkup_sw_table hSwTable = NULL;
    ix_lkup_table hTable = NULL;
    ix_uint32 numTries = 0;
    ix_memory_type rmMemType = 0;
    ix_error err = IX_SUCCESS;
    ix_uint32 *pSharedMemory = NULL;
#ifdef SPLIT_V6_TRIE_TABLE
    ix_uint32 *pSharedMemory16BitTrie = NULL;
#endif /* SPLIT_V6_TRIE_TABLE */
    ix_uint32 tableSize = 0;

    const ix_uint32 hi64kTableSize = SIZEOF_16BIT_TRIE;
    const ix_uint32 trieTableSize = SIZEOF_8BIT_TRIE;

    (void) arg_hLkup;

    if ((arg_pTableConf->tableType !=  IX_LKUP_TABLE_LPM)               ||
        (arg_pTableConf->keySize   != _IX_LKUP_SWLPMV6_KEY_SIZE)        ||
        (arg_pTableConf->dataSize  >  _IX_LKUP_SWLPMV6_MAX_DATA_SIZE)   ||
        (arg_pVndrConf != NULL))
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
                ("Invalid table configuration"));
    }

    switch (arg_pTableConf->dataLocation)
    {
        case IX_LKUP_MEM_SRAM:
            rmMemType = IX_MEMORY_TYPE_SRAM;
            break;

        case IX_LKUP_MEM_DRAM:
            rmMemType = IX_MEMORY_TYPE_DRAM;
            break;

        default:
            return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
               ("Invalid dataLocation: %d", arg_pTableConf->dataLocation));
    }

    /* Calculate the number of tries to allocate. Round the number up to an
     * even 32. Because _IX_TRIE_PER_ROUTE_RATIO might be small numbers,
     * the calculation might result in zero. So we add 32 instead of the
     * usual 32, that way guaranteeing the number of tries will be non-zero.
     */ 

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    /*
     * In linux, no floating point support in kernel, so IX_TRIE_ROUTE_RATIO
     * is the number of bits to shift, effectively resulting in a divide 
     * by 2^IX_TRIE_ROUTE_RATIO.
     */
    numTries = arg_pTableConf->sizeHint >> _IX_TRIE_ROUTE_RATIO;
#else
    numTries = arg_pTableConf->sizeHint * _IX_TRIE_ROUTE_RATIO;
#endif

    numTries = (numTries + 32) & 0xFFFFFFFE0;

#ifdef SPLIT_V6_TRIE_TABLE

    /* Calculate only the 8-bit trie table size */
    tableSize = trieTableSize * numTries;

#else /* SPLIT_V6_TRIE_TABLE */

    /* Calculate total table size for both 16-bit trie and 8-bit trie tables */
    tableSize = hi64kTableSize + (trieTableSize * numTries);

#endif /* SPLIT_V6_TRIE_TABLE */

    /* Allocate shared memory for the route table */
    IX_ERROR_CR(_ix_alloc_swlpmv6(numTries, &hLpmv6));

    /* Initialize the 1-bit trie in memory */
    IX_ERROR_CR(init_trie_root(hLpmv6));

#ifdef SPLIT_V6_TRIE_TABLE

    /**
     * Place the initial 16-bit trie in an SRAM channel different from the
     * channel contaning all other tries 
     */
    /* Allocate shared memory for the Hi64K (initial 16-bit) Trie table */
    IX_ERROR_CG(_ix_alloc_shared_memory(rmMemType, 
                                        arg_pTableConf->dataChannel16BitTrie,
                                        hi64kTableSize,
                                        &pSharedMemory16BitTrie), err, clean_hLpmv6);


    /* Allocate shared memory for all the other 8-bit Tries */
    IX_ERROR_CG(_ix_alloc_shared_memory(rmMemType, 
                                        arg_pTableConf->dataChannel,
                                        tableSize,
                                        &pSharedMemory), err, clean_hLpmv6);

#else /* SPLIT_V6_TRIE_TABLE */

    /* Allocate shared memory for the Route Table */
    IX_ERROR_CG(_ix_alloc_shared_memory(rmMemType, 
                                        arg_pTableConf->dataChannel,
                                        tableSize,
                                        &pSharedMemory), err, clean_hLpmv6);

#endif /* SPLIT_V6_TRIE_TABLE */

    hSwTable = hLpmv6->hSwTable;
    hTable = hSwTable->hTable;

    /* 
     * Setup function pointers.
     */

    hSwTable->ltabDestroyTable = _ix_lkup_swlpmv6_destroy_table;

    hTable->ltabAddEntry  = _ix_lkup_swlpmv6_add_entry;
    hTable->ltabRemoveEntry = _ix_lkup_swlpmv6_remove_entry;
    hTable->ltabUpdateEntry = _ix_lkup_swlpmv6_update_entry;
    hTable->ltabSearchTable = _ix_lkup_swlpmv6_search_table;
    hTable->ltabFindEntry = _ix_lkup_swlpmv6_find_entry;
    hTable->ltabReadFirstEntry = _ix_lkup_swlpmv6_read_first_entry;
    hTable->ltabReadNextEntry = _ix_lkup_swlpmv6_read_next_entry;
    hTable->ltabResetTable = _ix_lkup_swlpmv6_reset_table;
    hTable->ltabSetProperty = _ix_lkup_swlpmv6_set_property;
    hTable->ltabGetProperty = _ix_lkup_swlpmv6_get_property;
    hTable->ltabGetTableInfo = _ix_lkup_swlpmv6_get_table_info;

    hLpmv6->memType = arg_pTableConf->dataLocation;

#ifdef SPLIT_V6_TRIE_TABLE

    hLpmv6->hi64kTableBase = (tableElement *)pSharedMemory16BitTrie;
    hLpmv6->trieTableBase = (tableElement *)pSharedMemory;

#else /* SPLIT_V6_TRIE_TABLE */

    hLpmv6->hi64kTableBase = (tableElement *)pSharedMemory;
    hLpmv6->trieTableBase = &hLpmv6->hi64kTableBase[0x10000];

#endif /* SPLIT_V6_TRIE_TABLE */

    *arg_phTable = hTable;

    return IX_SUCCESS;

    /* Cleanup area */
clean_hLpmv6:

    (void) ix_cc_index_pool_fini(hLpmv6->hTrieIndexes);
    hLpmv6->hTrieIndexes = NULL;

    ix_ossl_free(hLpmv6);
    return err;
}

/*
 * NAME: _ix_lkup_swlpmv6_destroy_table
 *
 * DESCRIPTION: Invoked via IX_LKUP_DESTROY_TABLE().  Destroys a table that
 * was created earlier with IX_LKUP_CREATE_TABLE().  The application should
 * remove all entries from the table before it is destroyed.
 *
 * @Param:      - IN    arg_hLkup
 *                      The handle that was obtained from ix_lkup_sw_init().
 *
 * @Param:      - IN    arg_hTable
 *                      The handle for the table to be destroyed.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      One of the handles was invalid
 */
ix_error _ix_lkup_swlpmv6_destroy_table(
        ix_lkup arg_hLkup,
        ix_lkup_table arg_hTable
        )
{
    _ix_lkup_swlpmv6 hLpmv6 = (_ix_lkup_swlpmv6) arg_hTable;

#if defined(IX_DEBUG)
    if (arg_hTable == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_hTable must not be NULL"));
    }
#endif /* IX_DEBUG */

    (void) arg_hLkup;

    /* cleanup the 1-bit trie in memory */
    (void) fini_trie_root(hLpmv6);

    (void) ix_rm_mem_free(hLpmv6->hi64kTableBase);
 
#ifdef SPLIT_V6_TRIE_TABLE

    /**
     * If trie table was split across 2 SRAM channels,
     * free the other chunk of memory (i.e. 8-bit trie).
     */
    (void) ix_rm_mem_free (hLpmv6->trieTableBase);

#endif /* SPLIT_V6_TRIE_TABLE */

    (void) ix_cc_index_pool_fini(hLpmv6->hTrieIndexes);

    ix_ossl_free(hLpmv6);

    return IX_SUCCESS;
}

/*
 * NAME: _ix_lkup_swlpmv6_add_entry
 *
 * DESTRIPTION: Invoked via IX_LKUP_ADD_ENTRY().  Adds an entry to a table that
 * was previously created by calling IX_LKUP_CREATE_TABLE().  This function
 * will return an error if it detects a duplicate entry.
 *
 * @Param:      - IN    arg_hTable
 *                      The handle to the table to update.
 *
 * @Param:      - IN    arg_pKey
 *                      A pointer to the key data.  The length of the key was
 *                      defined when the table was created (i.e. 128 bits).
 *                      The key for this algorithm is the IP prefix.
 *
 * @Param:      - IN    arg_pMask
 *                      A pointer to the mask data for this entry. The length
 *                      of the mask was defined when the table was created.
 *
 * @Param:      - IN    arg_Weight
 *                      unused
 *
 * @Param:      - IN    arg_pData
 *                      Pointer to the data to be associated with the key.
 *                      This data is opaque to the lookup library and is only
 *                      interpreted by the higher level appliciations.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      Some of the parameters passed are not valid.
 *
 * IX_LKUP_ERROR_EXISTS
 *      An entry with the same key, mask and data is already in the table.
 *      (This implementation ignores the weight.)
 *
 * IX_LKUP_ERROR_CONFLICT
 *      An entry with the same key and mask but different data is already in
 *      the table. (This implementation ignores the weight.)
 *
 * IX_LKUP_ERROR_RESOURCES
 *      Not enough resources are available to complete the operation. Either
 *      the maximum table size has been exceeded or other resources could not
 *      be allocated.
 *
 * IX_LKUP_ERROR_INTERNAL
 *      An internal error occurred, probably because of a bug in the
 *      algorithm.
 */
ix_error _ix_lkup_swlpmv6_add_entry(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8  *arg_Mask,
        int arg_Weight,
        ix_uint8 *arg_pData
        )
{
    _ix_lkup_swlpmv6 hLpmv6 = (_ix_lkup_swlpmv6) arg_hTable;
    ix_error error = IX_SUCCESS;
    ix_uint8 maskLength;
    ix_uint16 data;
    ix_uint128 ipv6Addr;

    (void) arg_Weight;

    maskLength = *arg_Mask;
    data = *(ix_uint16 *)arg_pData;

    IX_NTOH128((ix_uint128 *)arg_pKey, &ipv6Addr);

    data = IX_NTOH16(data);

#if defined(IX_DEBUG)

    if (arg_hTable == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_hTable must not be NULL"));
    }

    if (arg_pData == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_pData must not be NULL"));
    }

    _swlpmv6_print_ipv6_address(&ipv6Addr);

#endif /* IX_DEBUG */

    error = __swlpmv6_add_route(hLpmv6, 
                                ipv6Addr.aUint32, 
                                maskLength,
                                data);

    return error;
}

/*
 * NAME: _ix_lkup_swlpmv6_remove_entry
 *
 * DESCRIPTION: Invoked via IX_LKUP_REMOVE_ENTRY(). Removes an entry that was
 * previously added to the table. The entry is identified by the values used
 * to add the entry.
 *
 * @Param:      - IN    arg_hTable
 *                      Handle to the table to update.
 *
 * @Param:      - IN    arg_pKey
 *                      Pointer to the key data of the item to be removed. The
 *                      length of the data was defined when the table was
 *                      created.
 *
 * @Param:      - IN    arg_pMask
 *                      Pointer to the mask data of the entry to be removed.
 *                      The length of the mask data as defined when the table
 *                      was created.
 *
 * @Param:      - IN    arg_Weight
 *                      unused
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The table handle was not valid
 *
 * IX_LKUP_ERROR_NOTFOUND
 *      A matching entry was not found
 *
 * IX_LKUP_ERROR_INTERNAL
 *      An internal error occurred, probably because of a bug in the
 *      algorithm.
 */
ix_error _ix_lkup_swlpmv6_remove_entry(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int arg_Weight
        )
{
    _ix_lkup_swlpmv6 hLpmv6 = (_ix_lkup_swlpmv6) arg_hTable;
    ix_error error = IX_SUCCESS;
    ix_uint8 maskLength;
    ix_uint128 ipv6Addr;

    (void) arg_Weight;


    IX_NTOH128((ix_uint128 *)arg_pKey, &ipv6Addr);

    maskLength = *arg_pMask;

#if defined(IX_DEBUG)

    if (arg_hTable == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_hTable must not be NULL"));
    }

    _swlpmv6_print_ipv6_address(&ipv6Addr);

#endif /* IX_DEBUG */

    error = __swlpmv6_delete_route(hLpmv6, 
                                   ipv6Addr.aUint32, 
                                   maskLength);

    return error;
}

/*
 * NAME: _ix_lkup_swlpmv6_update_entry
 *
 * DESCRIPTION: Invoked via IX_LKUP_UPDATE_ENTRY(). Updates the data
 * associated with an element already in the table.
 *
 * @Param:      - IN    arg_hTable
 *                      Handle to the table to update.
 *
 * @Param:      - IN    arg_pKey
 *                      Pointer to the key data of the item to be updated. The
 *                      length of the data was defined when the table was
 *                      created.
 *
 * @Param:      - IN    arg_pMask
 *                      Pointer to the mask data of the entry to be updated.
 *                      The length of the mask data as defined when the table
 *                      was created.
 *
 * @Param:      - IN    arg_Weight
 *                      unused
 *
 * @Param:      - IN    arg_pNewData
 *                      Pointer to the new data to associate with the specific
 *                      key.  The size of the data was specified when the
 *                      table was created.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The table handle was not valid
 *
 * IX_LKUP_ERROR_NOTFOUND
 *      A matching entry was not found so the update was not performed
 *
 * IX_LKUP_ERROR_INTERNAL
 *      An internal error occurred, probably because of a bug in the
 *      algorithm.
 */
ix_error _ix_lkup_swlpmv6_update_entry(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int arg_Weight,
        ix_uint8 *arg_pNewData
        )
{
    _ix_lkup_swlpmv6 hLpmv6 = (_ix_lkup_swlpmv6) arg_hTable;
    ix_error error = IX_SUCCESS;
    ix_uint8 maskLength;
    ix_uint16 data, stored_data;
    ix_uint128 ipv6Addr;

    (void) arg_Weight;

    data = *(ix_uint16 *)arg_pNewData;
    maskLength = *arg_pMask;

    IX_NTOH128((ix_uint128 *)arg_pKey, &ipv6Addr);
    data = IX_NTOH16(data);


#if defined(IX_DEBUG)

    if (arg_hTable == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_hTable must not be NULL"));
    }

    _swlpmv6_print_ipv6_address(&ipv6Addr);

#endif /* IX_DEBUG */

    /* quick check - see if the route exists, it should! */
    error = __swlpmv6_find_in_trie_table(hLpmv6,
                                         ipv6Addr.aUint32,
					 maskLength,
                                         &stored_data);

    /* If data == NULL_DATA, no route exists */
    if ((error != IX_SUCCESS) || (stored_data == NULL_DATA))
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_NOTFOUND, 
	  ("swlpmv6_update_route failed, entry doesn't exits; use route_add"));
    }

    error = __swlpmv6_update_route(hLpmv6, 
                                   ipv6Addr.aUint32, 
                                   maskLength, 
                                   data);

    return error;
}

/*
 * NAME: _ix_lkup_swlpmv6_search_table
 *
 * DESCRIPTION: Invoked via IX_LKUP_SEARCH_TABLE(). Searches a specific table
 * and returns the associated data.  The search key is compared with the
 * entries and masks using the Longest Prefix Match algorithm.
 *
 * @Param:      - IN    arg_hTable
 *                      Handle to the table to search.
 *
 * @Param:      - IN    arg_pKey
 *                      Pointer to the key data used for the search. The
 *                      length of the data was defined when the table was
 *                      created.
 *
 * @Param:      - OUT   arg_pData
 *                      Pointer to a location where the associated data will
 *                      be stored.  The search is always considered
 *                      successful; in effect, the algorithm assumes a
 *                      'default route' with the data 'zero'.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The handle was invalid
 *
 * NOTE: This implementation does not return error codes in search_table if
 * a route was not found
 */
ix_error _ix_lkup_swlpmv6_search_table(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pData
        )
{
    _ix_lkup_swlpmv6 hLpmv6 = (_ix_lkup_swlpmv6) arg_hTable;
    ix_uint16 trieData;
    ix_uint128 ipv6Addr;

    IX_NTOH128((ix_uint128 *)arg_pKey, &ipv6Addr);

#if defined(IX_DEBUG)
    if (arg_hTable == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_hTable must not be NULL"));
    }

#endif /* IX_DEBUG */

    (void) __swlpmv6_find_in_route_table(hLpmv6,
                                         ipv6Addr.aUint32,
                                         &trieData);

    *(ix_uint16 *)arg_pData = IX_HTON16(trieData);

    return IX_SUCCESS;
}

/*
 * NAME: _ix_lkup_swlpmv6_find_entry
 *
 * DESCRIPTION: Invoked via IX_LKUP_FIND_ENTRY(). Searches a table for an
 * entry that matches the exact key and mask combination. This is used to
 * locate a specific entry for search types where multiple entries may match
 * a specific search.
 *
 * @Param:      - IN    arg_hTable
 *                      Handle to the table to search.
 *
 * @Param:      - IN    arg_pKey
 *                      Pointer to the key data used for the search. The
 *                      length of the data was defined when the table was
 *                      created.
 *
 * @Param:      - IN    arg_pMask
 *                      Pointer to the mask data to use for the search.
 *                      The length of the mask data as defined when the table
 *                      was created.
 *
 * @Param:      - IN    arg_Weight
 *                      unused
 *
 * @Param:      - OUT   arg_pData
 *                      Pointer to a location where the associated data will
 *                      be stored if the search was successful. It assumes
 *                      that the caller allocated sufficient space for the
 *                      data associated with the call.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The handle was invalid
 *
 * IX_LKUP_ERROR_NOTFOUND
 *      There was no entry that matched this search request
 *
 * IX_LKUP_ERROR_INTERNAL
 *      An internal error occurred, probably because of a bug in the
 *      algorithm.
 */
ix_error _ix_lkup_swlpmv6_find_entry(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int arg_Weight,
        ix_uint8 *arg_pData
        )
{
    _ix_lkup_swlpmv6 hLpmv6 = (_ix_lkup_swlpmv6) arg_hTable;
    ix_error error = IX_SUCCESS;
    ix_uint8 maskLength;
    ix_uint16 data;
    ix_uint128 ipv6Addr;

    (void) arg_Weight;

    IX_NTOH128((ix_uint128 *)arg_pKey, &ipv6Addr);

    maskLength = *arg_pMask;

#if defined(IX_DEBUG)
    if (arg_hTable == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_hTable must not be NULL"));
    }

    _swlpmv6_print_ipv6_address(&ipv6Addr);

#endif /* IX_DEBUG */

    error = __swlpmv6_find_in_trie_table(hLpmv6,
                                         ipv6Addr.aUint32,
					 maskLength,
                                         &data);

    *(ix_uint16 *)arg_pData = IX_HTON16(data);

    return error;;
}

/*
 * NAME: _ix_lkup_swlpmv6_read_first_entry
 *
 * DESCRIPTION: Invoked via IX_LKUP_READ_FIRST_ENTRY(). This is a debug
 * function that is used to read the items that are stored in the table. This
 * function will get the first item in the table. To continue iterating
 * through all the entries the programmer will call this function once and
 * then repeatedly call IX_LKUP_READ_NEXT_ENTRY until all the entries have
 * been enumerated.
 *
 * Any time a route is found, update the hLpmv6->iteratorKey and
 * iteratorMaskLen values.  _ix_swlpmv6_read_next_entry() continues the search
 * with that key and mask length; _ix_swlpmv6_read_first_entry() starts over
 * every time.  The cookie is not used.
 *
 * @Param:      - IN    arg_hTable
 *                      Handle to the table to search.
 *
 * @Param:      - OUT   arg_pKey
 *                      Pointer where the key data will be stored if the call
 *                      is successful. This caller should make sure there is
 *                      sufficient storage for the key size of the table.
 *
 * @Param:      - OUT   arg_pMask
 *                      Pointer where the mask data will be stored. The caller
 *                      should ensure that this structure is of sufficient
 *                      size to store the mask.
 *
 * @Param:      - OUT   arg_pWeight
 *                      unused
 *
 * @Param:      - OUT   arg_pData
 *                      Pointer to the location where the associated data will
 *                      be stored if the call is successful.
 *
 * @Param:      - OUT   arg_pCookie
 *                      Pointer to an ix_lkup_cookie where the cookie is
 *                      stored. The cookie will be passed to a subsequent call
 *                      to IX_LKUP_READ_NEXT_ENTRY. (not used)
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      Some of the parameters passed are not valid
 *
 * IX_LKUP_ERROR_NOTFOUND
 *      This error is returned when there are no entries in the table. In this
 *      case none of the data structures to be filled in have valid data.
 *
 * IX_LKUP_ERROR_INTERNAL
 *      An internal error occurred, probably because of a bug in the
 *      algorithm.
 */
ix_error _ix_lkup_swlpmv6_read_first_entry(
        ix_lkup_table arg_hTable,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int *arg_pWeight,
        ix_uint8 *arg_pData,
        ix_lkup_cookie *arg_pCookie
        )
{
    _ix_lkup_swlpmv6 hLpmv6 = (_ix_lkup_swlpmv6) arg_hTable;
    ix_uint128 ipv6Addr;
    ix_uint16 data = 0;
    ix_uint8 prefixLength = 0;
    ix_error err = IX_SUCCESS;

    hLpmv6->iteratorKey = 0;
    hLpmv6->iteratorMaskLen = 0;
    
    (void)arg_pCookie;
   
    /* 
     *  This routine will recurisively build a route in the IP address
     *  and return both the prefix length and the data associated with it.
     */
    err = __swlpmv6_read_first_entry(hLpmv6,
                                     ipv6Addr.aUint32,
                                     &prefixLength,      
                                     arg_pWeight, 
                                     &data);

    if(err != IX_SUCCESS)
    {
        return err;
    }

    /* 
     * Convert everything to network byte order 
     */

    IX_HTON128(&ipv6Addr, (ix_uint128 *)arg_pKey);
    *(ix_uint16 *)arg_pData = IX_HTON16(data);
    *arg_pMask = prefixLength;
  
    return IX_SUCCESS;
}

/*
 * NAME: _ix_lkup_swlpmv6_read_next_entry
 *
 * DESCRIPTION: Invoked via IX_LKUP_READ_NEXT_ENTRY(). This is a debug
 * function that is used to read the items that are stored in the table. This
 * function is repeatedly called after IX_LKUP_READ_FIRST_ENTRY to enumerate
 * the contents of the table. Each call will return a cookie that is passed
 * to the next call.
 *
 * @Param:      - IN    arg_hTable
 *                      Handle to the table to search.
 *
 * @Param:      - IN    arg_OldCookie
 *                      The cookie that was returned by the last successful
 *                      call to IX_LKUP_READ_FIRST_ENTRY or
 *                      IX_LKUP_READ_NEXT_ENTRY.
 *
 * @Param:      - OUT   arg_pKey
 *                      Pointer where the key data will be stored if the call
 *                      is successful. This caller should make sure there is
 *                      sufficient storage for the key size of the table.
 *
 * @Param:      - OUT   arg_pMask
 *                      Pointer where the mask data will be stored. The caller
 *                      should ensure that this structure is of sufficient
 *                      size to store the mask.
 *
 * @Param:      - OUT   arg_pWeight
 *                      unused
 *
 * @Param:      - OUT   arg_pData
 *                      Pointer to the location where the associated data will
 *                      be stored if the call is successful.
 *
 * @Param:      - OUT   arg_pNewCookie
 *                      Pointer to an ix_lkup_cookie where the cookie is
 *                      stored. The cookie will be passed to a subsequent call
 *                      to IX_LKUP_READ_NEXT_ENTRY. (not used)
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      Some of the parameters passed are not valid
 *
 * IX_LKUP_ERROR_NOTFOUND
 *      This error is returned when there are no more entries in the table.
 *
 * IX_LKUP_ERROR_INTERNAL
 *      An internal error occurred, probably because of a bug in the
 *      algorithm.
 */
ix_error _ix_lkup_swlpmv6_read_next_entry(
        ix_lkup_table arg_hTable,
        ix_lkup_cookie arg_OldCookie,
        ix_uint8 *arg_pKey,
        ix_uint8 *arg_pMask,
        int *arg_pWeight,
        ix_uint8 *arg_pData,
        ix_lkup_cookie *arg_pNewCookie
        )
{
    _ix_lkup_swlpmv6 hLpmv6 = (_ix_lkup_swlpmv6) arg_hTable;
    ix_uint128 ipv6Addr;
    ix_uint16 data = 0;
    ix_uint8 prefixLength = 0;
    ix_error err = IX_SUCCESS;

    hLpmv6->iteratorKey = 0;
    hLpmv6->iteratorMaskLen = 0;

    (void) arg_OldCookie;
    (void) arg_pNewCookie;
   
    err = __swlpmv6_read_next_entry(hLpmv6, 
                                    ipv6Addr.aUint32,
                                    &prefixLength,
                                    arg_pWeight,
                                    &data);
    if (err)
    {
        return err;
    }

    /* 
     * Convert everything to network byte order 
     */

    IX_HTON128(&ipv6Addr, (ix_uint128 *)arg_pKey);
    *(ix_uint16 *)arg_pData = IX_HTON16(data);
    *arg_pMask = prefixLength;
  
    return IX_SUCCESS;
}

/*
 * NAME: _ix_lkup_swlpmv6_reset_table
 *
 * DESCRIPTION: Invoked via IX_LKUP_RESET_TABLE(). Clears all the items in the
 * table and resets it to its initial state. 
 *
 * @Param:      - IN    arg_hTable
 *                      The handle to the table to update.
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The table handle was not valid
 *
 * IX_LKUP_ERROR_INTERNAL
 *      An internal error occurred, probably because of a bug in the
 *      algorithm.
 */
ix_error _ix_lkup_swlpmv6_reset_table(
        ix_lkup_table arg_hTable
        )
{
    _ix_lkup_swlpmv6 hLpmv6 = (_ix_lkup_swlpmv6) arg_hTable;
    ix_error error = IX_SUCCESS;

#if defined(IX_DEBUG)
    if (arg_hTable == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_hTable must not be NULL"));
    }
#endif /* IX_DEBUG */

    /* 
     * Start by cleaning out the trie blocks and returning them to the 
     * free pool. Also delete the entrie 1-bit trie in memory.
     */

    error = __swlpmv6_reset_table(hLpmv6);

    return error;
}

/*
 * NAME: _ix_lkup_swlpmv6_set_property
 *
 * DESCRIPTION: Invoked via IX_LKUP_SET_PROPERTY().  No properties are
 * supported in this implementation.
 *
 * @Param:      - IN    arg_hTable
 *                      The handle to the table to update.
 *
 * @Param:      - IN    arg_Property
 *                      unused
 *
 * @Param:      - IN    arg_pValue
 *                      unused
 *
 * Error Codes:
 * IX_LKUP_ERROR_UNSUPPORTED
 *      The defined property was not supported
 */
ix_error _ix_lkup_swlpmv6_set_property(
        ix_lkup_table arg_hTable,
        int arg_Property,
        ix_uint8 *arg_pValue
        )
{
    (void) arg_hTable;
    (void) arg_Property;
    (void) arg_pValue;
    return IX_ERROR_WARNING(IX_LKUP_ERROR_UNSUPPORTED,
        ("This version of the Lookup Library does not support properties"));
}

/*
 * NAME: _ix_lkup_swlpmv6_get_property
 *
 * DESCRIPTION: Invoked via IX_LKUP_GET_PROPERTY(). No properties are
 * supported in this implementation.
 *
 * @Param:      - IN    arg_hTable
 *                      The handle to the table.
 *
 * @Param:      - IN    arg_Property
 *                      unused
 *
 * @Param:      - OUT   arg_pValue
 *                      unused
 *
 * Error Codes:
 * IX_LKUP_ERROR_UNSUPPORTED
 *      The defined property was not supported
 */
ix_error _ix_lkup_swlpmv6_get_property(
        ix_lkup_table arg_hTable,
        int arg_Property,
        ix_uint8 *arg_pValue
        )
{
    (void) arg_hTable;
    (void) arg_Property;
    (void) arg_pValue;
    return IX_ERROR_WARNING(IX_LKUP_ERROR_UNSUPPORTED,
        ("This version of the Lookup Library does not support properties"));
}

/*
 * NAME: _ix_lkup_swlpmv6_get_table_info
 *
 * DESCRIPTION: Invoked via IX_LKUP_GET_TABLE_INFO(). Returns values that need
 * to be used by the microengines:
 *
 * @Param:      - IN    arg_hTable
 *                      The handle to the table.
 *
 * @Param:      - OUT   arg_pTableId
 *                      Encodes the information needed by the microengines to
 *                      perform the lookup. 
 *
 * @Param:      - OUT   arg_pTableData1
 *                      unused
 *
 * @Param:      - OUT   arg_pTableData2
 *                      unused
 *
 * Error Codes:
 * IX_LKUP_ERROR_INVALID
 *      The table handle was not valid
 *
 * IX_LKUP_ERROR_INTERNAL
 *      An internal error occurred, probably because of a bug in the
 *      algorithm.
 */
ix_error _ix_lkup_swlpmv6_get_table_info(
        ix_lkup_table arg_hTable,
        ix_uint32 *arg_pTableId,
        ix_uint32 *arg_pTableData1,
        ix_uint32 *arg_pTableData2
        )
{
    ix_memory_type memType = 0;
    ix_uint32 memChannel = 0;
    ix_uint32 tableId = 0;
    _ix_lkup_swlpmv6 hLpmv6 = (_ix_lkup_swlpmv6) arg_hTable;

#if defined(IX_DEBUG)
    if (arg_hTable == NULL)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID,
            ("arg_hTable must not be NULL"));
    }
#endif /* IX_DEBUG */

    /* 
     * This implementation uses the hi64k tables's base address as the
     * table ID.
     */

    IX_ERROR_CRT(ix_rm_get_phys_offset(hLpmv6->hi64kTableBase,
                                       &memType,
				       &memChannel,
				       &tableId,
				       NULL),
                 IX_LKUP_ERROR_INTERNAL, 0);

    if (memType == IX_MEMORY_TYPE_SRAM)
    {
        IX_CC_SET_SRAM_CHANNEL_INTO_OFFSET(memChannel, tableId);
    }

    *arg_pTableId = tableId;

#ifdef SPLIT_V6_TRIE_TABLE

    /* For split trie table, return the 8-bit trie table base via arg_pTableData1. */
    IX_ERROR_CRT(ix_rm_get_phys_offset(hLpmv6->trieTableBase,
                                       &memType,
				       &memChannel,
				       &tableId,
				       NULL),
                 IX_LKUP_ERROR_INTERNAL, 0);

    if (memType == IX_MEMORY_TYPE_SRAM)
    {
        IX_CC_SET_SRAM_CHANNEL_INTO_OFFSET(memChannel, tableId);
    }

    *arg_pTableData1 = tableId;

#else /* SPLIT_V6_TRIE_TABLE */

    /* Data 1 is unused if the trie table was not split */
    *arg_pTableData1 = 0;

#endif /* SPLIT_V6_TRIE_TABLE */
 
    /* Data 2 is unused */
    *arg_pTableData2 = 0;

    return IX_SUCCESS;
}
