/**
 * ===========================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      swlpmv6_trie.c
 *
 * = DESCRIPTION
 *      Helper functions for enumerating the LPM tables.  This is part of the
 *      Lookup API's Software LPM algorithm for IPv6.
 *
 * ===========================================================================
 */
#define IX_ERROR_FILE_IDENT "$Id: swlpmv6_trie.c,v 1.19 2003/11/11 17:47:02 ktseng Exp $";

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
#define COPY_PREV_ELE_CURRENT(_recArrayCurrEle, _recArrayPrevEle)\
{\
  _recArrayCurrEle.hLpmv6 = _recArrayPrevEle.hLpmv6;\
  _recArrayCurrEle.pNode = _recArrayPrevEle.pNode;\
  _recArrayCurrEle.pIpv6Addr = _recArrayPrevEle.pIpv6Addr;\
  _recArrayCurrEle.pPrefixLength =_recArrayPrevEle.pPrefixLength;\
  _recArrayCurrEle.bitPos = _recArrayPrevEle.bitPos;\
  _recArrayCurrEle.pData = _recArrayPrevEle.pData;\
  _recArrayCurrEle.err = _recArrayPrevEle.err;\
}
#endif
/* 
 * Source code to maintain a 1-bit trie in memory. This supports the route
 * add/delete/query operations for the IPv6 Lookup Library and RTMv6.
 *
 * An IPv6 Address is 128-bits long and is contained in an array of 
 * 4 ix_uint32's. For instance the IPv6 address 3ffe::1234 would be 
 * expanded out, using the usual rules, and would look like the following
 * in array form. This is in Host (Big Endian) Order.
 *
 *     IpAddrArray[0] = 0x3ffe0000
 *     IpAddrArray[1] = 0x0
 *     IpAddrArray[2] = 0x0
 *     IpAddrArray[3] = 0x1234
 */ 

#include "ix_cc_error.h"
#include "lkup/internal/internal_swlpmv6.h"
#include "ix_ossl.h"

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
ix_store_element recArray[256];
ix_uint32 g_index=0;
#endif

/*
 * NAME: bit_at_position
 *
 * DESCRIPTION: Returns the bit stored at a certain bit position in an
 * IPv6 address.
 *
 * @Param:      - IN    ipAddr
 *                      The bit position to be checked.
 *
 * @Param:      - IN    bitPos
 *                      The bit position to be checked.
 *
 * @Returns:
 * The value at bitPos (0/1)
 * 
 * EXAMPLE:
 * If the IPv6 Address is 3ffe::1234 ipAddr would look like:
 *      ipAddr[0] - 3ffe0000
 *      ipAddr[1] - 0
 *      ipAddr[0] - 0
 *      ipAddr[0] - 1234
 *
 * Now, if bitPos is 123 (bitPos has to be between 0 and 127), the 
 * result in this case is 1
 */

ix_uint8
bit_at_position(
    ix_uint32 *ipAddr,
    ix_uint8 bitPos)
{
    ix_uint8 arrayIndex = 0;
    ix_uint8 bit;

    /* Read the uint32 at the appropriate index in the array */
    INDEX_IN_ARRAY(bitPos, arrayIndex);

    /* Bit position in the uint32 */
    bitPos = bitPos % 32;

    /* Generate the bit */
    BIT_AT_POSITION_32(ipAddr[arrayIndex], bitPos, bit);

    return (bit);
}

/*
 * NAME: alloc_trie_node
 *
 * DESCRIPTION: This function allocates a trie node. The trie node
 * represents an element in the 1-bit trie.
 *
 * @Param:      - OUT   **ppNode
 *                      Pointer to the pointer to the allocated node.
 *
 * @Param:      - IN    arg_data
 *                      Data stored in the node.
 *
 * @Returns:
 * IX_SUCCESS: If the allocation of a trie node was successful.
 * IX_LKUP_ERROR_RESOURCES: If the allocation of a trie node failed.
 */
ix_error
alloc_trie_node(
    ix_1bit_trie_node **ppNode, 
    ix_uint16 arg_data)
{
    *ppNode = ix_ossl_malloc(sizeof(ix_1bit_trie_node));

    if (*ppNode == NULL)
    {
        return (IX_ERROR_WARNING(IX_LKUP_ERROR_RESOURCES,
                              ("Failed to allocate 1-bit trie node ")));
    }

    (*ppNode)->data = arg_data;
    (*ppNode)->left = (*ppNode)->right = NULL;
    (*ppNode)->block_id = 0;

    (*ppNode)->visited = 0;

    return (IX_SUCCESS);
}

/*
 * NAME: free_trie_node
 *
 * DESCRIPTION: This routine frees a trie node allocated by calling
 * alloc_trie_node(). There are no checks to see if the memory was in
 * fact allocated by alloc_trie_node. It is assumed that the caller
 * will return the right pointer.
 *
 * @Param:      - IN    *pNode
 *                      Pointer to the node.
 *
 * @Returns:
 * None.
 */
void
free_trie_node(
    ix_1bit_trie_node *pNode)
{
    ix_ossl_free(pNode);
}

/*
 * NAME: alloc_8bit_trie_block
 *
 * DESCRIPTION: This routine allocates a trie block from the free pool.
 *
 * @Param:      - IN    block_index
 *                      Index of the 8-bit trie this routing frees.
 *
 * @Returns:
 * The index of the allocated trie block (16 bits). This is assuming
 * that the maximum number of tries are 2^16 or 64K.
 */
static ix_uint16 alloc_8bit_trie_block(
    _ix_lkup_swlpmv6 hLpmv6)
{
    ix_uint32 index;
    ix_error err;

    /* "index" contains the index of the free trie blocks */
    err = ix_cc_index_pool_alloc(hLpmv6->hTrieIndexes, &index);
   
    if (err != IX_SUCCESS)
    {
        /* 0 is not a valid index, so return 0 to indicate error */
        return (0);
    }

#if defined(IX_DEBUG)
    ix_ossl_message_log("allocated block (%ld) \n", index);
#endif  /* IX_DEBUG */

    return index;
}
/*
 * NAME: free_8bit_trie_block
 *
 * DESCRIPTION: This routine frees a trie block previously allocated
 * by alloc_pool_index. When a block is freed, its entries are cleared
 * out before it is returned to the free pool.
 *
 * @Param:      - IN    block_index
 *                      Index of the 8-bit trie this routing frees.
 *
 * @Returns:
 * none.
 */
static void free_8bit_trie_block(
    _ix_lkup_swlpmv6 hLpmv6,
    ix_uint16 block_index)
{
    ix_uint16 entry;
    ix_error err;

    /* Each 8-bit trie has 256 entries, clear all of them */
    for (entry = 0; entry < 256; entry++)
    {
        _IX_SET_ENTRY_8(hLpmv6, block_index, entry, _IX_TRIE_PACK_LOCAL(0, 0));
    }

    err = ix_cc_index_pool_free(hLpmv6->hTrieIndexes, block_index);

    if (err != IX_SUCCESS)
    {
        ix_ossl_message_log("Failed to return block %d to free pool\n", 
	                                                         block_index);
    }
                     
    return;
}

/*
 * NAME: reset_1bit_trie
 *
 * DESCRIPTION: This routine frees a trie block previously allocated
 * by alloc_pool_index. When a block is freed, its entries are cleared
 * out before it is returned to the free pool.
 *
 * @Param:      - IN    block_index
 *                      Index of the 8-bit trie this routing frees.
 *
 * @Returns:
 * none.
 */
static void reset_1bit_trie(
    ix_1bit_trie_node *treeRoot)
{
    /* 
     * Start by walking down the trie and setting each 
     * nodes data to NULL_DATA. That way when it is time to 
     * delete the tree, each node will be deletable without any 
     * problems.
     */

    treeRoot->data = NULL_DATA;

    if (treeRoot->left != NULL)
    {
        reset_1bit_trie(treeRoot->left);
    }

    if (treeRoot->right != NULL)
    {
        reset_1bit_trie(treeRoot->right);
    }

    return;
}

/*
 * NAME: zero_hi64k_block
 *
 * DESCRIPTION: This routine zero's out the contents of the hi64K
 * block.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 *
 * @Returns:
 * none.
 */
static void zero_hi64k_block(
            _ix_lkup_swlpmv6 hLpmv6)
{
    ix_uint32 index, max = (1 << 16);


    /* Set all the values to 0 */
    for (index = 0; index < max; index++)
    {
        _IX_SET_ENTRY_16(hLpmv6, index, 0);
    }
}

/*
 * NAME: populate_blocks_in_trie
 *
 * DESCRIPTION: The block_id field in a trie node holds a value if the
 * trie node is at a boundary bit. So we store the block_id for bit positions
 * 23, 31...,127. When we delete a route, the block_id points to the block
 * that contains the bit.
 * 
 * @Param:      - IN    *ipAddr,
 *                      The IPv6 address.
 *
 * @Param:      - IN    *trie_chain
 *                      The chain (maximum of 14 entries) of 8-bit trie
 *                      indices. 
 *
 * @Param:      - IN    trie_count
 *                      count of the number of entries in "trie_chain".
 *
 * @Returns:
 * none.
 */
static
void
populate_blocks_in_trie(
    _ix_lkup_swlpmv6 hLpmv6,
    ix_uint32 *ipAddr,
    ix_uint16 *trie_chain, 
    ix_uint32 trie_count)  
{
    ix_uint32 bitPos, current_level;
    ix_1bit_trie_node *node = NULL;
    ix_uint16 bit;

    if (trie_count < 1)
    {
        /* Nothing to do */
	return;
    }

    node = hLpmv6->gRoot;
    current_level = 0;

    bitPos = 0;
    while (bitPos < MAX_IPV6_ADDR_LENGTH)
    {
        bit = bit_at_position(ipAddr, bitPos);

        node = (bit ? node->right : node->left);

	if (BOUNDARY_BIT_POSITION(bitPos) && (bitPos > 15))
	{
	    if ((node->block_id != 0) && 
	        (trie_chain[current_level] != node->block_id))
	    {
	        /* This is something that should never happen. Basically
		 * it means that there is already a route here and we're
		 * trying to add a new trie block when the block_id is non
		 * NULL. There is nothing much I can do. The return simply
		 * prevents it from overwriting the block_id.
		 */
		return;
	    }
	    else
	    {
	        node->block_id = trie_chain[current_level];
	    }

	    current_level++;
	    if(current_level == trie_count)
	    {
	       break;
	    }
	}

	bitPos++;
    }

    return;
}

/*
 * NAME: link_chain
 *
 * DESCRIPTION: This route links the 8-bit tries backwards (to form
 * a route). The 8-bit trie "chain" is finally linked to the appropriate
 * entry in the 16-bit trie table to finish constructing the route.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 * 
 * @Param:      - IN    ipAddr  
 *                      The IPv6 address.
 *
 * @Param:      - IN    trie_chain
 *                      An array of upto 14 8-bit trie indexes. 
 *
 * @Param:      - IN    trie_count
 *                      count of the number of elements in the "trie_chain" 
 *                      array. 
 *
 * @Param:      - IN    last_trie_index
 *                      the index of the last 8-bit trie in the chain.
 *
 * @Returns:
 * none.
 */

static
void
link_chain(
    _ix_lkup_swlpmv6 hLpmv6,
    ix_uint32 *ipAddr,
    ix_uint16 *trie_chain,
    ix_uint8 trie_count,
    ix_uint16 last_trie_index)
{
    ix_uint16 index, shift;
    ix_uint8 ipByte;
    ix_uint32 entry;
    ix_uint16 stored_data;
    ix_uint16 stored_next_trie;
    ix_int8 level;

    if (trie_chain == NULL)
    {
        ix_ossl_message_log ("link_chain: Null array for trie_chain? \n");
	return;
    }

    /* 
     * "Link" the trie chain backwards. This is required for synchronization
     * between the core and the microblock.
     *
     * Note that at this point "last_trie_index" still holds the index
     * of the last trie in the chain.
     */

    for (level = trie_count - 1; level >=0; level--)
    {
        /* The starting bit position of this 8-bit chunk is
         * (16 + 8 * level)
         */

        INDEX_IN_ARRAY((16 + 8 * level), index);
	SHIFT_BY((16 + 8 * level), shift);

	ipByte = (ipAddr[index] >> shift) & 0xFF;
	    
        entry = _IX_GET_ENTRY_8(hLpmv6, trie_chain[level], ipByte);
        stored_data = _IX_GET_DATA(entry);
        stored_next_trie = _IX_GET_NEXT_TRIE(entry);

	/* if there is a next_trie, it better be this one */
        if ((stored_next_trie != 0) && (stored_next_trie != last_trie_index))
        {
            ix_ossl_message_log("***>>>> Overwriting a route?? <<<<*** \n");
        }
        else
        {
	    if (stored_next_trie == 0)
	    {
                _IX_SET_ENTRY_8(hLpmv6, 
		                trie_chain[level], 
		                ipByte, 
			     _IX_TRIE_PACK_LOCAL(last_trie_index, stored_data));
            }
        }
        last_trie_index = trie_chain[level];
    }

    /* 
     * Link the first 8-bit trie to the 16-bit trie to complete
     * the operation 
     * Note: At this point "next_trie" holds the index of the first 
     *       8-bit trie in the chain.
     */
        
    entry = _IX_GET_ENTRY_16(hLpmv6, ipAddr[0] >> 16);
    stored_data = _IX_GET_DATA(entry);
    stored_next_trie = _IX_GET_NEXT_TRIE(entry);

    if ((stored_next_trie != 0) && (stored_next_trie != last_trie_index))
    {
        ix_ossl_message_log("***>>>> Overwriting a route?? <<<<*** \n");
    }
    else
    {
	if (stored_next_trie == 0)
	{
            _IX_SET_ENTRY_16(hLpmv6, ipAddr[0] >> 16, 
                           _IX_TRIE_PACK_LOCAL(last_trie_index, stored_data));
        }
    }

    /* 
     * Finally Store the 8-bit trie blocks in the 1-bit trie. They are 
     * useful when we are trying to delete a route. Notice that the last
     * trie is left out. This routine is invoked to store the trie 
     * numbers in the 1-bit trie which will be used to quickly locate
     * the tries when we delete this route. We don't need to store the
     * last trie, so we leave it out.
     */

    populate_blocks_in_trie(hLpmv6,
                            ipAddr, 
                            trie_chain, 
			    trie_count);

    return;
}

/*
 * NAME: delink_chain
 *
 * DESCRIPTION: This routine is used to cleanup a route. It delinks 
 * a chain of 8-bit tries. When a route is deleted, the chain is 
 * delinked starting with the 16-bit trie first working forward.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 * 
 * @Param:      - IN    ipAddr  
 *                      The IPv6 address.
 *
 * @Param:      - IN    bitPos
 *                      Bit position in the IPv6 address.
 *
 * @Param:      - IN    currentBlockId
 *                      Index of the current 8-bit trie block.
 *
 * @Returns:
 * none.
 */

static
void
delink_chain(
    _ix_lkup_swlpmv6 hLpmv6,
    ix_uint32 *ipAddr,
    ix_uint32 bitPos,
    ix_uint16 currentBlockId)
{
    ix_uint16 ipShort;
    ix_uint8 ipByte;
    ix_uint32 entry;
    ix_uint16 data_stored, next_trie_stored;
    ix_uint16 index, shift;

    if (bitPos < 16)
    {
        ipShort = ipAddr[0] >> 16;
	entry = _IX_GET_ENTRY_16(hLpmv6, ipShort);

	next_trie_stored = _IX_GET_NEXT_TRIE(entry);
	data_stored = _IX_GET_DATA(entry);

	/* Clear out the next trie */
	_IX_SET_ENTRY_16(hLpmv6, ipShort,  _IX_TRIE_PACK_LOCAL(0, data_stored));
    }
    else
    {
        /* we need to update an 8-bit trie entry */
	INDEX_IN_ARRAY(bitPos, index);
	SHIFT_BY(bitPos, shift);

	ipByte = ipAddr[index] >> shift;

	entry = _IX_GET_ENTRY_8(hLpmv6, currentBlockId, ipByte);

	next_trie_stored = _IX_GET_NEXT_TRIE(entry);
	data_stored = _IX_GET_DATA(entry);

        _IX_SET_ENTRY_8(hLpmv6, currentBlockId, ipByte,  
	                _IX_TRIE_PACK_LOCAL(0, data_stored));
    }

    if (next_trie_stored != 0)
    {
#if defined(IX_DEBUG)
	ix_ossl_message_log("freeing 8-bit trie blocks (%d) \n",
	                                                   next_trie_stored);
#endif  /* IX_DEBUG */

        free_8bit_trie_block(hLpmv6, next_trie_stored);
    }

    return;
}


/*
 * NAME: init_trie_root
 *
 * DESCRIPTION: This route initializes the root of the 1-bit trie.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 * @Returns:
 * IX_SUCCESS: If the trie root was initialized successfully.
 * IX_LKUP_ERROR_RESOURCES: If the initialization failed.
 */
ix_error
init_trie_root(
    _ix_lkup_swlpmv6 hLpmv6)
{
    return (alloc_trie_node(&(hLpmv6->gRoot), NULL_DATA));
}

/*
 * NAME: fini_trie_root
 *
 * DESCRIPTION: This route initializes the root of the 1-bit trie.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 * @Returns:
 * none.
 */
void
fini_trie_root(
    _ix_lkup_swlpmv6 hLpmv6)
{
    return (free_trie_node(hLpmv6->gRoot));
}

/*
 * NAME: pattern_generate_8
 *
 * DESCRIPTION: This routine generates all combinations given an 
 * 8-bit portion of an IP address and a bit position. The indexes generated
 * are used to write values in a 8-bit trie block.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 * 
 * @Param:      - IN    ipAddr
 *                      8-bit portion of an IPv6 address.
 * 
 * @Param:      - IN    bitPos
 *                      Bit position in "ipAddr"
 * 
 * @Param:      - IN    block_index
 *                      The index of the 8-bit trie block used to write values
 *                      after all possible offset combinations are generated.
 * 
 * @Param:      - IN    data
 *                      Data to write
 * @Returns:
 * none.
 */
static
void
pattern_generate_8(
    _ix_lkup_swlpmv6 hLpmv6,
    ix_uint8 ipAddr,
    ix_uint8 bitPos,
    ix_uint32 block_index,
    ix_uint16 data)
{
    ix_uint8 mask, ipByte;
    ix_uint8 current_index;
    ix_uint8 maskLength = (bitPos + 1) % 8;
    ix_uint16 stored_trie;

    if (BOUNDARY_BIT_POSITION(bitPos))
    {
        mask = 0xFF; 
    }
    else
    {
       GENERATE_MASK(maskLength, 8, mask);
    }

    ipByte = ipAddr & mask;
    current_index = ipByte;

    do 
    {
	stored_trie = 
            _IX_GET_NEXT_TRIE(_IX_GET_ENTRY_8(hLpmv6, block_index, 
	                                                     current_index));

        _IX_SET_ENTRY_8(hLpmv6, block_index, current_index, 
	                 _IX_TRIE_PACK_LOCAL(stored_trie, data));

    } while (current_index++ < (ix_uint8)(ipByte | ~mask)); 


    return;
}

/*
 * NAME: pattern_generate_16
 *
 * DESCRIPTION: This routine generates all possible combinations given
 * a 16-bit portion of an IP address and a bit position to start from.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 *
 * @Param:      - IN    ipAddr
 *                      16-bit portion of an IPv6 address
 *
 * @Param:      - IN    bitPos
 *                      Bit position to use to generate a mask
 * 
 * @Param:      - IN    data
 *                      Data to write in the 16-bit trie at the offsets
 *                      generated by this routine.
 *
 * @Returns:
 * none.
 */
static
void
pattern_generate_16(
    _ix_lkup_swlpmv6 hLpmv6,
    ix_uint16 ipAddr,
    ix_uint8 bitPos,
    ix_uint16 data)
{
    ix_uint16 mask, ipShort;
    ix_uint8 maskLength = bitPos + 1;
    ix_uint16 current_index;
    ix_uint16 stored_trie;

    if (bitPos == 15)
    {
        mask = 0xFFFF;
    }
    else
    {
        GENERATE_MASK(maskLength, 16, mask);
    }

    ipShort = ipAddr & mask;
    current_index = ipShort;

    do 
    {
	stored_trie = 
              _IX_GET_NEXT_TRIE(_IX_GET_ENTRY_16(hLpmv6, current_index));

	_IX_SET_ENTRY_16(hLpmv6, current_index, 
	                 _IX_TRIE_PACK_LOCAL(stored_trie, data));

    } while (current_index++ < (ix_uint16)(ipShort | ~mask)); 

    return;
}

/*
 * NAME: generate_all_combinations
 *
 * DESCRIPTION: This route generates all combinations of offsets given
 * an IPv6 address and a bit position. 
 * NOTE: The routine makes sure that it does not cross a trie boundary.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 *
 * @Param:      - IN    ipAddr
 *                      The IPv6 address
 *
 * @Param:      - IN    bitPos
 *                      the current bit position in the address.
 * 
 * @Param:      - IN    data
 *                      the data to write out
 *
 * @Param:      - IN    nextBit
 *                      This is used if something special needs to be 
 *                      done to the next bit before generating combinations.
 *                      if nextBit == -1, nothing special needs to be
 *                      done.
 *                      if nextBit is zero or one, the next bit (bitPos + 1)
 *                      is cleared or set and combinations are generated.
 * @Returns:
 * IX_LKUP_ERROR_INTERNAL: SW LPM algorithm internal error.
 * IX_LKUP_ERROR_RESOURCES: Failed to allocate resources needed.
 * IX_SUCCESS: 
 */
static
ix_error
generate_all_combinations(
    _ix_lkup_swlpmv6 hLpmv6,
    ix_uint32 *ipAddr,
    ix_uint8 bitPos,
    ix_uint16 data,
    ix_int8 nextBit)
{
    ix_uint16 index = 0, shift = 0;
    ix_error error = IX_SUCCESS;

    if(!LEGAL_BIT_POSITION(bitPos))
    {
        return (IX_ERROR_WARNING(IX_LKUP_ERROR_INTERNAL,
                                ("invalid bit position %d", bitPos)));
    }


    /* Where does this bitPos fall? */
    INDEX_IN_ARRAY(bitPos, index);
    SHIFT_BY(bitPos, shift);

    if (bitPos < 16)
    {
        /* 
         * Extract the top 16 bits of the IPv6 address and generate
         * all possible combinations of it.
         */
        ix_uint16  ipShort;

        ipShort = ipAddr[index] >> shift;

        /* 
         * Anything special to do with the next bit before passing it on to
         * the 16-bit-pattern-generator function 
         */
        if (nextBit != -1)
        {
           if (nextBit == 0)
           {
	       /* 
	        * Clear the next bit in the 16 bit chunk and generate 
		* patterns.
		*/
		CLEAR_BIT_AT_POSITION_16(ipShort, (bitPos + 1));
                pattern_generate_16(hLpmv6, ipShort, (bitPos + 1), data);
	    }
	    else
	    {
	        /* 
		 * Set the next bit in the 16 bit chunk and generate 
		 * patterns.
		 */
		SET_BIT_AT_POSITION_16(ipShort, (bitPos + 1));
                pattern_generate_16(hLpmv6, ipShort, (bitPos + 1), data);
	    }
        }
	else
	{
	    pattern_generate_16(hLpmv6, ipShort, bitPos, data);
	}
    }
    else
    {
        /* 
         * Extract the appropriate 8 bit part of the IPv6 address and generate
         * all possible combinations of it.
         */

        ix_uint8 ipByte;
	ix_uint16 next_trie;

	/* 
	 * trie_chain is an array of all trie blocks along the way to 
	 * the "target block", i.e. the target block is the last block 
	 * in the chain. If blocks don't exits they are allocated as the 
	 * trie traversal is done. 
	 * You'll notice that MAX_8BIT_TRIES_PER_ROUTE is defined as 14. 
	 * This is because with the 16-8-8-8... the maximum number of 
	 * blocks (8-bit tries) can be 14 (with a full 128 bit route).
	 */

        ix_uint16 trie_chain[MAX_8BIT_TRIES_PER_ROUTE] = {0};
	ix_uint8 trie_count;  /* total number of tries in trie_chain */
        ix_int8 level;

	/* 
	 * Start by walking the trie: 
	 * We know we're at a 8-bit trie, somewhere. Walk down the route 
	 * table, starting with the 16-bit trie. It is possible that
	 * some 8-bit tries haven't been allocated yet. We allocate them.  
         *
	 * NOTE: Since the microblock could be looking up a route when the 
	 *       core component is in the middle of adding a route, the 
	 *       allocated tries are "linked" backwards. We start from the 
	 *       last trie and work backwards to the 16-bit trie.
	 */

        ipByte = ipAddr[index] >> shift;

        trie_count = (bitPos - 16)/8;

        if (trie_count > MAX_8BIT_TRIES_PER_ROUTE)
        {
	    /* Anyway, we haven't allocated an array bigger than this, so 
	     * something is really wrong. Return error.
	     */
            return (IX_ERROR_WARNING(IX_LKUP_ERROR_INTERNAL,
                                ("invalid trie_count (%d)", trie_count)));
        }

	next_trie = _IX_GET_NEXT_TRIE(_IX_GET_ENTRY_16(hLpmv6, 
	                                               ipAddr[0] >> 16));

	for (level = 0; level < trie_count; level++)
	{
	    if (next_trie == 0)
	    {
	       trie_chain[level] = alloc_8bit_trie_block(hLpmv6);
	       if (trie_chain[level] <= 0)
	       {
                   return (IX_ERROR_WARNING (IX_LKUP_ERROR_RESOURCES,
                    ("Failed to allocate 8-bit trie, free pool empty?")));
	       }
	    }
	    else
	    {
	        /* Next .. */
	        INDEX_IN_ARRAY((16 + 8 * level), index);
	        SHIFT_BY((16 + 8 * level), shift);

                trie_chain[level] = next_trie;
	        next_trie = _IX_GET_NEXT_TRIE(_IX_GET_ENTRY_8(hLpmv6,
		                           next_trie, 
		                          ((ipAddr[index] >> shift) & 0xFF)));
            }
	}

	/* Allocate the last 8-bit trie in the chain. The reason
	 * this is done outside the loop is because it handles the case
	 * whenre the prefix length is less than 24 (i.e, bitPos is less 
	 * than 23)
	 */
	if (next_trie == 0)
	{
	    next_trie = alloc_8bit_trie_block(hLpmv6);
            if (next_trie <= 0)
            {
                return (IX_ERROR_WARNING (IX_LKUP_ERROR_RESOURCES,
                    ("Failed to allocate 8-bit trie, free pool empty?")));
            }

	}

#if defined(IX_DEBUG)
	ix_ossl_message_log("This is what the chain looks like trie_count (%d)\n", trie_count);
	for (level = 0; level < trie_count; level++)
	{
	    ix_ossl_message_log ("Trie at level %d is %d \n", 
	                                           level, trie_chain[level]);
	}
#endif  /* IX_DEBUG */

	/* 
	 * At this point we've reached the target block and we have all
	 * the blocks allocated 
	 */
 
        /* 
         * Anything special to do with the next bit before passing it on to
         * the 8-bit-pattern-generator function 
         */
        if (nextBit != -1)
        {
           if (nextBit == 0)
           {
	       /* 
	        * Clear the next bit in the 8 bit chunk and generate patterns.
		*/

		CLEAR_BIT_AT_POSITION_8(ipByte, (bitPos + 1));
                pattern_generate_8(hLpmv6, 
		                   ipByte, 
		                   (bitPos + 1), 
				   next_trie, 
				   data);
	    }
	    else
	    {
	        /* 
		 * Set the next bit in the 8 bit chunk and generate patterns.
		 */

		SET_BIT_AT_POSITION_8(ipByte, (bitPos + 1));

                pattern_generate_8(hLpmv6,
		                   ipByte, 
		                   (bitPos + 1), 
				   next_trie, 
				   data);
	    }
        }
	else
	{
	    pattern_generate_8(hLpmv6,
	                       ipByte, 
	                       bitPos, 
			       next_trie,
			       data);
	}


	/* link all the tries together properly */
        link_chain(hLpmv6, 
	           ipAddr, 
	           trie_chain, 
		   trie_count, 
		   next_trie);
    }
   return error;
}


/*
 * NAME: controlled_prefix_expansion
 *
 * DESCRIPTION: controlled prefix expansion needs to be performed when 
 * the mask length (prefix length) specified during route add is not on 
 * a trie boundary (i.e, it is not 16, 24, 32 etc.). If it is in between
 * a trie, multilple entries in the appropriate trie block need to be filled
 * with the same information. This routine traverses the 1-bit trie and 
 * figures out the entries that need to be filled with this data.
 *
 * It is obvious that not all combinations can be filled. Sometimes a specific
 * route might already be in place and this routine might get called during
 * the addition of a less specific route. In this case, some of the entries
 * need to be left as they are.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 * 
 * @Param:      - IN    ipAddr
 *                      The IPv6 address
 * 
 * @Param:      - IN    maskLength
 *                      Length of the prefix for this route.
 * 
 * @Param:      - IN    bitPos
 *                      the current bit position in the IP address
 * 
 * @Param:      - IN    treeRoot
 *                      The root of the subtree this invocation is checking
 * 
 * @Param:      - IN    data
 *                      the data to write out
 * 
 * @Returns:
 * IX_LKUP_ERROR_INTERNAL: SW LPM algorithm internal error.
 * IX_LKUP_ERROR_RESOURCES: Failed to allocate resources needed.
 * IX_SUCCESS: 
 */
static
ix_error
controlled_prefix_expansion(
    _ix_lkup_swlpmv6 hLpmv6,
    ix_uint32 *ipAddr,
    ix_uint8  maskLength,
    ix_uint8  bitPos,
    ix_1bit_trie_node *treeRoot,
    ix_uint16 data)
{
    ix_error error;

    /* 
     * There are three separate cases to handle;
     * 1. Both children of treeRoot are NULL
     * 2. Left child is NULL
     * 3. Right child is NULL
     */

    if (BOUNDARY_BIT_POSITION(bitPos))
    {
	/* generate combinations with the full 8 bit value */
        return (generate_all_combinations(hLpmv6, ipAddr, bitPos, data, -1));
    }

    if ((treeRoot->left == NULL) && (treeRoot->right == NULL))
    {
    	/* 
	 * safely generate all patterns starting with this prefix and 
    	 * overwrite all locations
    	 */

        return (generate_all_combinations(hLpmv6, ipAddr, bitPos, data, -1));

    } /* both children are null */
    else
    {
        if (treeRoot->left != NULL)
        {
            /* 
	     * If the data stored in the left child is "non-null", 
	     * we're done. We can stop going down this sub-tree. Move
	     * on to the right sub-tree.
             *
	     * If the data stored is "null", traverse down the tree.
	     */

            if ((treeRoot->left)->data == NULL_DATA)
	    {
	        CLEAR_BIT_AT_POSITION(ipAddr, (bitPos + 1));

	        error = controlled_prefix_expansion(hLpmv6,
		                                    ipAddr, 
	                                            maskLength,
                                                    (bitPos + 1), 
					            treeRoot->left,
					            data);
	       if (error != IX_SUCCESS)
	       {
		   return error;
	       }
            }
        }
	else
        {
    	    error = generate_all_combinations(hLpmv6, ipAddr, bitPos, data, 0);
            if (error != IX_SUCCESS)
            {
                return error;
            }
    	}

        /* 
	 * Check right sub-tree now. 
	 */

	/* Right child in not NULL */
        if (treeRoot->right != NULL)
        {
            /* 
	     * If the data stored in the right child is "non-null", 
	     * we're done. We can stop going down this sub-tree. 
             *
	     * If the data stored is "null", traverse down the tree.
	     */

	    /* 
	     * We've taken a right turn set the bit in the current bit position
	     */

	    SET_BIT_AT_POSITION(ipAddr, (bitPos + 1));

            if ((treeRoot->right)->data == NULL_DATA)
	    {
    	        error =  controlled_prefix_expansion(hLpmv6,
		                                     ipAddr, 
                                                     maskLength,
                                                     (bitPos + 1), 
                                                     treeRoot->right,
                                                     data);
	       if (error != IX_SUCCESS)
	       {
		   return error;
	       }
            }
        }
    	else
    	{
    	    error = generate_all_combinations(hLpmv6, ipAddr, bitPos, data, 1);
            if (error != IX_SUCCESS)
            {
                return error;
            }
    	}
    }

    return (IX_SUCCESS);
}

/*
 * NAME: traverse_and_build_tree
 *
 * DESCRIPTION: This routing traverses the 1-bit trie and extends it
 * if necessary based on the IPv6 address and the prefix length. 
 * This routine calls controlled_prefix_expansion to generate combinations
 * based on controlled prefix expansion. 
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 *
 * @Param:      - IN    ipAddr 
 *                      The IPv6 address
 *
 * @Param:      - IN    maskLength
 *                      Length of the prefix for the route
 *
 * @Param:      - IN    bitPos 
 *                      Bit position in the IPv6 address we're at
 *
 * @Param:      - IN    treeRoot
 *                      Pointer to the node in the 1-bit trie we're at
 *
 * @Param:      - IN    data
 *                      the data to write out
 *
 * @Param:      - IN    operation
 *                      operation to perform. This is to distinguish between
 *                      addition, deletion and update.
 * @Returns:
 * IX_LKUP_ERROR_INVALID: 
 * IX_LKUP_ERROR_INTERNAL
 * IX_LKUP_ERROR_RESOURCES
 * IX_LKUP_ERROR_CONFLICT
 * IX_LKUP_ERROR_EXISTS
 * IX_SUCCESS
 */
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
static
ix_error
traverse_and_build_tree(
    _ix_lkup_swlpmv6 hLpmv6,
    ix_uint32 *ipAddr,
    ix_uint8 maskLength,
    ix_uint8 bitPos,
    ix_1bit_trie_node *treeRoot,
    ix_uint16 data,
    ix_swlpmv6_op operation)
{
    ix_uint8 bit = 0;
    ix_uint8 tempBitPos = 0;
    ix_uint32 i;
    ix_error error = IX_SUCCESS;
    ix_1bit_trie_node *newNode = NULL;
    ix_1bit_trie_node *tempRoot = NULL;

    /* Init for the first time */
    tempRoot = treeRoot;
    tempBitPos = bitPos;
    
    if (maskLength > 128)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID, ("mask length %s is not valid \n", maskLength));
    }

    for(i=0; i<=(maskLength-1); i++)
    {
       if(!LEGAL_BIT_POSITION(tempBitPos))
       {
           return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID, ("bit position %d is not valid \n", tempBitPos));
       }

       bit = bit_at_position(ipAddr, tempBitPos);
    
       /* 
        * If bit == 0 and the left subtree doesn't exist, create a new
        * node and make that the root of the sub-tree to traverse further.
        * else simply go down the left subtree.
        * If bit == 1 do the same thing with the right sub-tree.
        */
       if (bit == 0)
       {
           if (tempRoot->left == NULL)
           {
               error = alloc_trie_node(&newNode, NULL_DATA);
               if (error != IX_SUCCESS)
               {
                   return error;
               }

               tempRoot->left = newNode;
           }
           else
   	   {
	       newNode = tempRoot->left;
	   }
       }
       else
       {
           if (tempRoot->right == NULL)
           {
               error = alloc_trie_node(&newNode, NULL_DATA);
               if (error != IX_SUCCESS)
               {
	           return error;
               }

               tempRoot->right = newNode;
           }
           else
	   {
	       newNode = tempRoot->right;
	   }
       }
       tempBitPos++;   
       tempRoot = newNode;
    } /*end of for */
 
    /* due to the traverse_and_build_tree is a nonrecursive function
     * when the for loop completes, the tempBitPos is extra incremented 
     * by one, when the entire tree is traversed depending prefix length
     * to overcome this problem tempBitPos value
     * is decremented by one.
     */
    
    tempBitPos = tempBitPos - 1;
    
 /* If data is present in the node and this is an "ADD" operation
  * then something isn't right. On the other hand, the same
  * function is called from the "UPDATE" function, in which
  * case, we'd expect data to be present
  */
    if (newNode->data != NULL_DATA)
    {
       if (operation == swlpmv6_add)
       {
            if (newNode->data == data) 
            {
               return (_IX_IPV6_ROUTE_EXISTS(EXISTS, ipAddr, data, 
	                                           maskLength));
            }
            else
            {
                return (_IX_IPV6_ROUTE_EXISTS(CONFLICT, ipAddr, data, 
	                                               maskLength));
            }
        }
	else if (operation == swlpmv6_update)
	{
	    if (newNode->data == NULL_DATA)
	    {
         	 /* trying to update an entry that does not exist */
	        return (_IX_IPV6_ROUTE_NOTFOUND(ipAddr, maskLength));
	    }
	    else
	    {
                newNode->data = data;
            }
	}
    }
    else
    {
        newNode->data = data;
    }

    /* 
     * Since we are the root of the tree, based on the mask length,
     * perform a Controlled Prefix Expansion to, if necessary, overwrite
     * some more general routes.
     */

     controlled_prefix_expansion(hLpmv6, ipAddr, maskLength, tempBitPos, 
		     newNode, data);

     return error;
}
#else

static
ix_error
traverse_and_build_tree(
    _ix_lkup_swlpmv6 hLpmv6,
    ix_uint32 *ipAddr,
    ix_uint8 maskLength,
    ix_uint8 bitPos,
    ix_1bit_trie_node *treeRoot,
    ix_uint16 data,
    ix_swlpmv6_op operation)
{
    ix_1bit_trie_node *newNode = NULL;
    ix_uint8 bit = 0;
    ix_error error = IX_SUCCESS;

    if(!LEGAL_BIT_POSITION(bitPos))
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID, ("bit position %d is not valid \n", bitPos));
    }

    if (maskLength > 128)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INVALID, ("mask length %s is not valid \n", maskLength));
    }

    bit = bit_at_position(ipAddr, bitPos);
    
    /* 
     * If bit == 0 and the left subtree doesn't exist, create a new
     * node and make that the root of the sub-tree to traverse further.
     * else simply go down the left subtree.
     * If bit == 1 do the same thing with the right sub-tree.
     */
    if (bit == 0)
    {
        if (treeRoot->left == NULL)
        {
            error = alloc_trie_node(&newNode, NULL_DATA);
            if (error != IX_SUCCESS)
            {
                return error;
            }

            treeRoot->left = newNode;
        }
	else
	{
	    newNode = treeRoot->left;
	}
    }
    else
    {
        if (treeRoot->right == NULL)
        {
            error = alloc_trie_node(&newNode, NULL_DATA);
            if (error != IX_SUCCESS)
            {
	        return error;
            }

            treeRoot->right = newNode;
        }
	else
	{
	    newNode = treeRoot->right;
	}
    }
   
    /* Have we reached the end of recursion ? */
    if (bitPos == (maskLength - 1))
    {
       if (newNode->data != NULL_DATA)
       {
	   /* If data is present in the node and this is an "ADD" operation
	    * then something isn't right. On the other hand, the same
	    * function is called from the "UPDATE" function, in which
	    * case, we'd expect data to be present
	    */
	   if (operation == swlpmv6_add)
	   {
               if (newNode->data == data) 
               {
                  return (_IX_IPV6_ROUTE_EXISTS(EXISTS, ipAddr, data, 
		                                             maskLength));
               }
               else
               {
                   return (_IX_IPV6_ROUTE_EXISTS(CONFLICT, ipAddr, data, 
	                                                  maskLength));
               }
           }
	   else if (operation == swlpmv6_update)
	   {
	       if (newNode->data == NULL_DATA)
	       {
		   /* trying to update an entry that does not exist */
	           return (_IX_IPV6_ROUTE_NOTFOUND(ipAddr, maskLength));
	       }
	       else
	       {
                   newNode->data = data;
               }
	   }
       }
       else
       {
           newNode->data = data;
       }

       /* 
        * Since we are the root of the tree, based on the mask length,
        * perform a Controlled Prefix Expansion to, if necessary, overwrite
        * some more general routes.
        */

       return (controlled_prefix_expansion(hLpmv6,
                                           ipAddr, 
                                           maskLength, 
				           bitPos, 
				           newNode, 
				           data));

    }

    error =  traverse_and_build_tree(hLpmv6,
                                     ipAddr,
                                     maskLength, 
                                     bitPos + 1, 
                                     newNode, 
                                     data, 
				     operation);

    return error;
}
#endif

/*
 * NAME: delete_subtree
 *
 * DESCRIPTION: This route checks to see if a subtree rooted at "treeRoot"
 * can be deleted. The way this is done is as follows:
 * 
 * - If the subtree has no children (it is a leaf node) it can be removed
 * without any problems.

 * - If it has any children, we traverse down the appropriate child and
 *   recursively check if the sub-tree rooted at that child can be deleted.
 * 
 * Also, to be able to cleanup trie blocks, we check to see if we're at a 
 * "boundary bit" (like bit 15, 23, 31 etc.). If we are and the subtree can
 * be deleted (it has no routes beyond it), it is time to free a 8-bit 
 * trie.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 * 
 * @Param:      - IN    ipAddr
 *                      IPv6 address of the route being deleted
 * 
 * @Param:      - IN    treeRoot
 *                      The node at which this tree is rooted
 * 
 * @Param:      - IN    bitPos
 *                      The bit position in the IPv6 address we're at
 *
 * @Param:      - IN    cleanup_trie_blocks
 *                      Flag indicating whether cleanup of 8-bit tries 
 *                      is desired or not. If cleanup_trie_block is set
 *                      to 1, we handle freeing up of 8-bit trie blocks.
 * 
 * @Returns:
 * 1: if the subtree rooted at "subtree" can be deleted
 * 0: otherwise.
 */

ix_int8
delete_subtree(
    _ix_lkup_swlpmv6 hLpmv6,
    ix_uint32 *ipAddr,
    ix_1bit_trie_node *treeRoot,
    ix_uint32 bitPos,
    ix_uint8 cleanup_trie_blocks)  
{
    ix_uint8 leftCanDelete = 1;
    ix_uint8 rightCanDelete = 1;

    if ((treeRoot->left == NULL) && (treeRoot->right == NULL))
    {
        if (treeRoot->data == NULL_DATA)
	{
	    /* This must be the node we're trying to delete */
	    return 1;
	}
	else
	{
            /* Cannot delete, simply return "NO" */
            return 0;
	}
    }

    /* Is the left sub-tree removable? */
    if (treeRoot->left != NULL)
    {
        leftCanDelete = delete_subtree(hLpmv6, ipAddr, treeRoot->left, 
                                       (bitPos + 1), cleanup_trie_blocks);

	if(leftCanDelete)
	{
            /* 
	     * The data in this node is "null-data" and the node has 
	     * no children 
	     */
            free_trie_node(treeRoot->left);
	    treeRoot->left = NULL;
	}
	else
	{
	    return 0;
	}
    }
	
    /* Is the right sub-tree removable ? */
    if (treeRoot->right != NULL)
    {
        rightCanDelete = delete_subtree(hLpmv6, ipAddr, treeRoot->right, 
                                        (bitPos + 1), cleanup_trie_blocks);

	if(rightCanDelete)
	{
            /* 
	     * The data in this node is "null-data" and the node 
	     * has no children 
	     */
            free_trie_node(treeRoot->right);
	    treeRoot->right = NULL;
	}
	else
	{
	    return 0;
	}
    }

    /* 
     * If I can delete the left and right sub-tree, it's time to 
     * delete this tree. Also note that if I am on the boundary of a bit,
     * it's time to return the 8-bit trie block to the free-pool.
     */
    if ((leftCanDelete && rightCanDelete) && (treeRoot != hLpmv6->gRoot))
    {
	if (cleanup_trie_blocks)
	{
            if (BOUNDARY_BIT_POSITION(bitPos))
	    {
	        /* Remeber that the above macro will return TRUE for 
    	         * bitPosition 7, in which case we don't want to do anything.
	         * Hence the check!
	         */
	        if (bitPos > 8)
	        {
	            delink_chain(hLpmv6, ipAddr, bitPos, treeRoot->block_id);
	        }
	    }
        }

        return 1;
    }

    return 0;
}

/*
 * NAME: zero_route
 *
 * DESCRIPTION: This routine "zero's" out a route given an IPv6 address
 * and a mask length.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 *
 * @Param:      - IN    ipAddr
 *                      The IPv6 address.
 *
 * @Param:      - IN    maskLength
 *                      The prefix length for the route being deleted.
 *
 * @Returns:
 * IX_LKUP_ERROR_INTERNAL: SW LPM algorithm internal error.
 */
ix_error
zero_route(
    _ix_lkup_swlpmv6 hLpmv6,      /* LPM IPV6 handle */
    ix_uint32 *ipAddr,            /* IPv6 Address */
    ix_uint8 maskLength)          /* Mask length for this route */
{
    ix_uint8 level;
    ix_uint16 next_trie, next_next_trie;
    ix_uint16 index, shift;
    ix_uint8 ipByte;

    if (maskLength <= 16)
    {
        pattern_generate_16(hLpmv6, 
	                    ipAddr[0] >> 16, 
	                    maskLength - 1, 
			    0);
    }
    else
    {
        /* Walk down the trie and zero out each element, start
	 * with the 16-bit trie block */

        next_trie = _IX_GET_NEXT_TRIE(_IX_GET_ENTRY_16(hLpmv6, 
	                                               ipAddr[0] >> 16));

	/* clear out the entry in the 16-bit trie */
	_IX_SET_ENTRY_16(hLpmv6, ipAddr[0] >> 16, 
	                 _IX_TRIE_PACK_LOCAL(next_trie, 0));

	if (next_trie == 0)
	{
	    return (IX_ERROR_WARNING(IX_LKUP_ERROR_INTERNAL,
	                  ("Is the route length really greater than 16?\n")));
	}

	level = 0;

	/* Keep walking as long as there is a trail ... */
	while (next_trie)
	{
	    INDEX_IN_ARRAY((16 + 8 * level), index);
	    SHIFT_BY((16 + 8 * level), shift);

	    ipByte = (ipAddr[index] >> shift) & 0xFF;

	    if ((maskLength - (16 + 8 * level)) <= 8)
	    {
	       /* bingo, we've reached our target block */
	       pattern_generate_8(hLpmv6,
	                          ipAddr[index] >> shift, 
	                          maskLength - 1, 
				  next_trie, 
				  0);
	       break;
	    }

	    next_next_trie = _IX_GET_NEXT_TRIE(_IX_GET_ENTRY_8(hLpmv6,
	                                                       next_trie, 
                                                              ipByte));

            _IX_SET_ENTRY_8(hLpmv6, next_trie, ipByte, 
	                           _IX_TRIE_PACK_LOCAL(next_next_trie, 0));

	    next_trie = next_next_trie;

	    level++;
	}
    }

    return (IX_SUCCESS);
}

/*
 * NAME: __swlpmv6_delete_route
 *
 * DESCRIPTION: This routine handles deletion routes given an IPv6 address
 * and a prefix length. There are a lot of different scenarios which 
 * are considered here.
 * 
 * The essential idea is as follows:
 * When a route is deleted and a "less specific" route exists, the route
 * being deleted is replaced with the less specific route.
 * For instance, if 3ffe:1234::/22 is added followed by 3ffe:1234::/25, 
 * the route with prefix length 25 is more specific and replaces the
 * data that was initially holding the data allocated when the 22 bit 
 * route was added. Now, when the 25 bit route is deleted we revert back
 * to the less specific route - the 22-bit route in this case.
 * 
 * It must be noted that when we revert back care must be taken to avoid
 * extending routes. If the less specific route (the one we're trying to 
 * revert back to)  doesn't lie in the same trie as the one we are deleting,
 * we simply zero out the locations.
 *
 * Also note that sometimes, there might not be a less specific route. 
 * For example, if we add 3ffe::/16 and try to delete it, there is nothing
 * to revert back to. In such a case we simply zero out the route. Care
 * must be taken in such cases to avoid deleting more specific routes that 
 * might exists. Like if we had added another route 3ffe:1234::/22 after 
 * adding the 16 bit route and we are trying to delete the 16-bit route.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 * 
 * @Param:      - IN    ipAddr
 *                      The IPv6 address
 * 
 * @Param:      - IN    maskLength
 *                      The prefix length for this route
 * @Returns:
 * IX_LKUP_ERROR_NOTFOUND
 * IX_LKUP_ERROR_INTERNAL
 * IX_LKUP_ERROR_RESOURCES
 * IX_SUCCESS
 */

ix_error
__swlpmv6_delete_route(
    _ix_lkup_swlpmv6 hLpmv6,
    ix_uint32 *ipAddr,
    ix_uint8 maskLength)
{
    ix_1bit_trie_node *node = NULL;              /* Node for this route */
    ix_1bit_trie_node *lastKnownNode = NULL;     /* Pointer to the node 
                                                  * containing the "less 
                                                  * specific" route.
                                                  */
    
    ix_uint16 lastKnownData = 0;                 /* Data value of the "less
                                                  * specific" route
                                                  */

    ix_int16 bitPos, lastKnownBitPos = 0, currentBitPos = 0;
    ix_int8 currentStride, lastKnownStride;
    ix_uint16 bit;
    ix_error error = IX_SUCCESS;

    /* Start with the root of the tree */
    node = hLpmv6->gRoot;
    lastKnownNode = node;

    /* Walk down the tree looking for the next best (less specific) data */
    for (bitPos = 0; bitPos < maskLength; bitPos++)
    {
        bit = bit_at_position(ipAddr, bitPos);

	/* 
	 * If bit == 0 go left else go right.
	 */

    	node = ((bit == 0) ? node->left : node->right);

    	if (node == NULL)
    	{
	    return (IX_ERROR_WARNING(IX_LKUP_ERROR_NOTFOUND,
	                    ("reached a null-node in the 1-bit-trie ? ")));
    	}

    	if ((node->data != NULL_DATA) && (bitPos != (maskLength - 1)))
    	{
    	    /* save this information */
    	    lastKnownNode = node;
	    lastKnownData = node->data;
            lastKnownBitPos = bitPos;
    	}
    }

    if (node->data == NULL_DATA)
    {
        return (IX_ERROR_WARNING(IX_LKUP_ERROR_NOTFOUND, 
	     ("The node does not contain valid data, does this route exist?")));
    }

    currentBitPos = bitPos - 1;	   /* the loop incremented the bitPos by 1 */

    /* This route is gone */
    node->data = NULL_DATA;

    /* 
     * This is the part where I have to do some tricky things. 
     * Couple of cases, 
     * 1. We could be deleting a specific route and expect to have it 
     *    restored by a more general route. However, if the more
     *    general route is in a different trie block, there is nothing 
     *    to do. We simply zero the entries and leave the more specific
     *    routes in the trie and return.
     *
     * The absolute bottom line, if the "last known good data" is in the
     * same 8-bit trie as the route we're trying to delete, simply restore.
     * Else clear out the locations from the current node down. (If there
     * are any specific routes, the routine will handle them automatically.
     */

    /* 
     * Cleanup the tree and syncup the route table to be consistent
     * with the tree. When we cleanup the route table, we make sure
     * that we do it walking forward, i.e. starting from the 16-bit 
     * trie. This is to maintain syncronization with the microblock.
     */

    if (lastKnownData != 0)
    {
        lastKnownStride = COMPUTE_STRIDE(lastKnownBitPos);
    	currentStride = COMPUTE_STRIDE(currentBitPos);

	if (lastKnownStride == currentStride)
	{
	   error = controlled_prefix_expansion(hLpmv6,
	                                       ipAddr,
	                                       maskLength,
				               lastKnownBitPos,
				               lastKnownNode,
				               lastKnownData);
	}
	else
	{
	   error = controlled_prefix_expansion(hLpmv6,
	                                       ipAddr,
	                                       maskLength,
				               currentBitPos,
				               node,
				               NULL_DATA);
        }
    }
    else
    {
        /* We pretty much need to clear out this route. This
	 * can be done by using controlled_prefix_expansion
	 * with NULL_DATA from the root, but, that would mean
	 * we would traverse a LOT of subtrees, which is a waste.
	 * Instead, traverse down the path you're trying to delete
	 * and clear it out.
	 */

	if ((node->left != NULL) || (node->right != NULL))
	{
	   error = controlled_prefix_expansion(hLpmv6,
	                                       ipAddr,
	                                       maskLength,
				               currentBitPos,
				               node,
				               NULL_DATA);
	}
	else
	{
            error = zero_route(hLpmv6, ipAddr, maskLength);
	}
    }

    if (error != IX_SUCCESS)
    {
        return error;
    }

    /* Since we're starting from the root, start with -1 */
    if (lastKnownNode == hLpmv6->gRoot)
    {
        lastKnownBitPos--;
    }

    delete_subtree(hLpmv6, ipAddr, lastKnownNode, lastKnownBitPos, 1);
    return (IX_SUCCESS);
}

/*
 * NAME: __swlpmv6_add_route
 *
 * DESCRIPTION: This routine adds a route to the routing table in shared
 * memory. Care is taken to avoid overwriting "more specific" routes, if 
 * they exist. 
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 * 
 * @Param:      - IN    ipAddr
 *                      The IPv6 address
 * 
 * @Param:      - IN    maskLength
 *                      Prefix length of the route we are trying to add
 * 
 * @Param:      - IN    data
 *                      Data for this route. Typically the next-hop ID
 * 
 * @Returns:
 * A valid error token. (see traverse_and_build_tree() above).
 */

ix_error
__swlpmv6_add_route(
    _ix_lkup_swlpmv6 hLpmv6,
    ix_uint32 *ipAddr,
    ix_uint8 maskLength,
    ix_uint16 data)
{
    ix_swlpmv6_op operation = swlpmv6_add;
    /* 
     * Start with gRoot and bit position 0 and traverse down the
     * tree building it as required.
     */
    return (traverse_and_build_tree(hLpmv6,
                                   ipAddr, 
                                   maskLength, 
				   0, 
				   hLpmv6->gRoot, 
				   data,
				   operation));
}

/*
 * NAME: __swlpmv6_update_route
 *
 * DESCRIPTION: This routine updates a route in the routing table in shared
 * memory. Care is taken to avoid overwriting "more specific" routes, if 
 * they exist. 
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 * 
 * @Param:      - IN    ipAddr
 *                      The IPv6 address
 * 
 * @Param:      - IN    maskLength
 *                      Prefix length of the route we are trying to add
 * 
 * @Param:      - IN    data
 *                      Data for this route. Typically the next-hop ID
 * 
 * @Returns:
 * A valid error token. (see traverse_and_build_tree() above).
 */

ix_error
__swlpmv6_update_route(
    _ix_lkup_swlpmv6 hLpmv6,
    ix_uint32 *ipAddr,
    ix_uint8 maskLength,
    ix_uint16 data)
{
    ix_swlpmv6_op operation = swlpmv6_update;

    /* 
     * Start with gRoot and bit position 0 and traverse down the
     * tree building it as required.
     */

    return (traverse_and_build_tree(hLpmv6,
                                    ipAddr, 
                                    maskLength, 
				    0, 
				    hLpmv6->gRoot, 
				    data,
				    operation));
}

/*
 * NAME: __swlpmv6_find_in_route_table
 *
 * DESCRIPTION: This routine finds the data (next_hop ID in this case)
 * given an IPv6 address. This is done based on the way the route table
 * is programmed. A "Longest Prefix Match (LPM)" scheme is used and the
 * data corresponding to the longest entry is retured.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 *
 * @Param:      - IN    ipAddr
 *                      The IPv6 address
 *
 * @Param:      - OUT   pData
 *                      The result of the search opeartion is returned here
 *
 * @Returns:
 * none.
 */

void
__swlpmv6_find_in_route_table(
    _ix_lkup_swlpmv6 hLpmv6,
    ix_uint32 *ipAddr,
    ix_uint16 *pData)
{
    ix_uint32 next_trie, entry, i;
    ix_uint32 index, shift;
    ix_uint16 data = 0;

    /* Start with the root of the tree */
    entry = _IX_GET_ENTRY_16(hLpmv6, ipAddr[0] >> 16);

    next_trie = _IX_GET_NEXT_TRIE(entry);
    data = _IX_GET_DATA(entry);

    *pData = data;

    if (next_trie == 0)
    {
	return;
    }

    for (i = 0; i < MAX_8BIT_TRIES_PER_ROUTE; i++)
    {
        INDEX_IN_ARRAY((16 + 8 * i), index);
        SHIFT_BY((16 + 8 * i), shift);

	entry = _IX_GET_ENTRY_8(hLpmv6, next_trie, 
	                        ((ipAddr[index] >> shift) & 0xFF));

        data = _IX_GET_DATA(entry);
	next_trie = _IX_GET_NEXT_TRIE(entry);

	if (data != 0)
	{
            *pData = data;
	}

        if (next_trie == 0)
        {
    	    return;
        }
    }

    return;
}


/*
 * NAME: __swlpmv6_find_in_trie_table
 *
 * DESCRIPTION: This routine finds corresponiding to a given IP address
 * and a prefix length. This is different from the routine above because
 * it is not a longest prefix match. This routine walks down the 1-bit trie
 * in memory and stops when it has traversed down to the right level.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 *
 * @Param:      - IN    ipAddr
 *                      The IPv6 address
 *
 * @Param:      - IN    maskLength
 *                      The mask length of the route we are interested in.
 *
 * @Param:      - OUT   pData
 *                      The result of the search opeartion is returned here
 *
 * @Returns:
 * IX_LKUP_ERROR_NOT_FOUND
 * IX_SUCCESS
 */
ix_error
__swlpmv6_find_in_trie_table(
    _ix_lkup_swlpmv6 hLpmv6,
    ix_uint32 *ipAddr,
    ix_uint8 maskLength,
    ix_uint16 *pData)
{
    ix_1bit_trie_node *node = NULL;
    ix_uint8 bitPos, bit;
    
    /* Start at the root */
    node = hLpmv6->gRoot;
    *pData = NULL_DATA;

    if (node == NULL)
    {
	return _IX_IPV6_ROUTE_NOTFOUND(ipAddr, maskLength);
    }
     
    for (bitPos = 0; bitPos < maskLength; bitPos++)
    {
        bit = bit_at_position(ipAddr, bitPos);

	/* if bit is 0, go left, else go right */
	node = ((bit == 0) ? node->left : node->right);

	if (node == NULL)
	{
	    return _IX_IPV6_ROUTE_NOTFOUND(ipAddr, maskLength);
	}
    }

    /* Check for the case where there is a route, but this just
     * happened to be a bit stream that doesn't land on real data 
     */
    if (node->data == NULL_DATA)
    {
        return _IX_IPV6_ROUTE_NOTFOUND(ipAddr, maskLength);
    }

    /* Here node should be the target node we're looking for */
    *pData = node->data;

    return IX_SUCCESS;
}

/*
 * NAME: __swlpmv6_reset_table
 *
 * DESCRIPTION: This routine resets a table. Clears all the trie blocks
 * and the 1-bit trie nodes and returns the table to its initial state.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 *
 * @Returns:
 * IX_LKUP_ERROR_INTERNAL
 * IX_SUCCESS
 */

ix_error __swlpmv6_reset_table(
    _ix_lkup_swlpmv6 hLpmv6)
{
    ix_uint16 trie_index;
    int in_use = 0;
    ix_error error = IX_SUCCESS;

    /* We do a couple of different things here:
     * 1. For each 8-bit trie block in use, free the trie, which
     *    includes zero'ing out its contents.
     *
     * 2. Mark all the 8-bit trie blocks free (this is done by 
     *    reseting the index pool).
     * 
     * 3. Zero out the contents of the hi64K trie block.
     *
     * 4. Walk the 1-bit trie in memory and set each node's data to 
     *    NULL_DATA. 
     * 
     * 5. Delete the subtree rooted at gRoot (the global root of the 
     *    1-bit trie in memory. Since all the node's data have been
     *    set to NULL_DATA, the entire tree should be deletable.
     *
     */

    /* Step 1: */
    for (trie_index = 1; trie_index < hLpmv6->numTries; trie_index++)
    {
        error = ix_cc_index_pool_in_use(hLpmv6->hTrieIndexes, 
	                                             trie_index, &in_use);
        if (error != IX_SUCCESS)
	{
            return IX_ERROR_WARNING(IX_LKUP_ERROR_INTERNAL,
                                 ("Failed ix_cc_index_pool_in_use \n"));
	}

        if (in_use)
	{
            free_8bit_trie_block(hLpmv6, trie_index);
        }
    }

    /* Step 2 */
    error = ix_cc_index_pool_reset(hLpmv6->hTrieIndexes);
    if (error != IX_SUCCESS)
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INTERNAL,
	               ("Failed to reset index pool for 8-bit trie blocks"));
    }

    /* Step 3 */
    zero_hi64k_block(hLpmv6);

    /* Step 4 */
    reset_1bit_trie(hLpmv6->gRoot);

    /* 
     * Step 5 : Notice that delete_subtree is called with the last parameter
     * set to 0
     */
    delete_subtree(hLpmv6, 0, hLpmv6->gRoot, -1, 0);

    /* Sanity check */
    if ((hLpmv6->gRoot->left != NULL) || (hLpmv6->gRoot->right != NULL))
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_INTERNAL,
                                 ("Failed to delete the entire trie "));
    }

    return IX_SUCCESS;
}


void
__swlpmv6_reset_visited(ix_1bit_trie_node *pRoot)
{

    if (pRoot == NULL)
    {
        return;
    }

   if (pRoot->visited) 
   {
       pRoot->visited = 0;
   }
   
   __swlpmv6_reset_visited(pRoot->left);
   __swlpmv6_reset_visited(pRoot->right);

   return;
}

/*
 * NAME: __swlpmv6_dfs
 *
 * DESCRIPTION: This routine performs a depth first search of the 1-bit
 * trie table and returns data and prefix length corresponding to the data.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 * @Param:      - OUT   pNode
 *                      The root of this subtree, here is where the search
 *                      begins in the current stack.
 * @Param:      - OUT   pIpv6Addr
 *                      The IPv6 address that is set as a result of the search.
 * @Param:      - OUT   pPrefixLength
 *                      The prefix Length that is set as a resulf of the search.
 * @Param:      - OUT   pData
 *                      The data resturned as a result of the search.
 *
 * @Returns:
 * IX_LKUP_ERROR_INTERNAL
 * IX_SUCCESS
 */
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
ix_error
__swlpmv6_dfs(_ix_lkup_swlpmv6 hLpmv6,
              ix_1bit_trie_node *pNode,
              ix_uint32 *pIpv6Addr,
              ix_uint8 *pPrefixLength,
              ix_uint16 bitPos,
              ix_uint16 *pData)
{
   ix_error err = IX_SUCCESS;
   
   recArray[g_index].hLpmv6 = hLpmv6;
   recArray[g_index].pNode = pNode;
   recArray[g_index].pIpv6Addr = pIpv6Addr;
   recArray[g_index].pPrefixLength = pPrefixLength;
   recArray[g_index].bitPos= bitPos;
   recArray[g_index].pData= pData;
   recArray[g_index].err = IX_SUCCESS;
   err = __swlpmv6_dfs_v();	

   return err;

}

/*
 * NAME: __swlpmv6_dfs_v
 *
 * DESCRIPTION: This routine performs a depth first search of the 1-bit
 * trie table,using global structure  and this function is called 
 * by __swlpmv6_dfs
 *
 * @Param:      - IN   none
 *
 * @Returns:
 */
ix_error
__swlpmv6_dfs_v(void)
{
    ix_error err = IX_SUCCESS;

    recArray[g_index].bitPos++;

    if (recArray[g_index].pNode->data != NULL_DATA)
    {
        if (!recArray[g_index].pNode->visited)
        {
            /* We've hit a case where we need to report this data */
            *(recArray[g_index].pData) = recArray[g_index].pNode->data;
            *(recArray[g_index].pPrefixLength) = recArray[g_index].bitPos - 1;
            recArray[g_index].pNode->visited = 1;  /* it is now */
            recArray[g_index].err = IX_SUCCESS ;  
            return IX_SUCCESS;
        }
    }
    
    /* No data found yet, check the left sub-tree first */
    if (recArray[g_index].pNode->left != NULL)
    {
        CLEAR_BIT_AT_POSITION(recArray[g_index].pIpv6Addr, recArray[g_index].bitPos - 1);
	g_index++;
        COPY_PREV_ELE_CURRENT(recArray[g_index],recArray[g_index-1]);
	recArray[g_index].pNode = recArray[g_index].pNode->left;
    
        err = __swlpmv6_dfs_v();
	g_index--;
        recArray[g_index].err = err;
    }

    if (recArray[g_index].err != IX_SUCCESS)
    { 
        return recArray[g_index].err;
    }
    else if (*(recArray[g_index].pData) != NULL_DATA)
    {
        /* return the data */
        recArray[g_index].err = IX_SUCCESS;  
        return IX_SUCCESS; 
    }
    
    /* 
     * If nothing is found yet and there is a right sub-tree, go down this
     * the right sub-tree.
     */
    if ((*recArray[g_index].pData == NULL_DATA) && 
		    (recArray[g_index].pNode->right != NULL))
    {
        SET_BIT_AT_POSITION(recArray[g_index].pIpv6Addr,recArray[g_index].bitPos - 1);
	g_index++;
        COPY_PREV_ELE_CURRENT(recArray[g_index],recArray[g_index-1]);
        recArray[g_index].pNode = recArray[g_index].pNode->right;
        /* Continue through the tree .. */
        err = __swlpmv6_dfs_v(); 
	g_index--;
        recArray[g_index].err = err;
    }
   
    if (recArray[g_index].err != IX_SUCCESS)
    {
        return recArray[g_index].err;
    }
    else if (*(recArray[g_index].pData) != NULL_DATA)
    {
        recArray[g_index].err = IX_SUCCESS;  
        return IX_SUCCESS;
    }

    /* 
     * If we've reached the top of the tree and we're here, clearly
     * no data is found 
     */
    if (recArray[g_index].pNode == recArray[g_index].hLpmv6->gRoot) 
    { 
        recArray[g_index].err = IX_LKUP_ERROR_NOTFOUND;  
        return IX_ERROR_WARNING(IX_LKUP_ERROR_NOTFOUND,
                                     ("Reached end of search \n"));
    }

    /* We haven't searched the entire tree yet, explore on.. */
    recArray[g_index].err = IX_SUCCESS;  
    return IX_SUCCESS;
}
#else
ix_error
__swlpmv6_dfs(_ix_lkup_swlpmv6 hLpmv6,
              ix_1bit_trie_node *pNode,
              ix_uint32 *pIpv6Addr,
              ix_uint8 *pPrefixLength,
              ix_uint16 bitPos,
              ix_uint16 *pData)
{

    ix_error err = IX_SUCCESS;

    bitPos++;

    if (pNode->data != NULL_DATA)
    {
        if (!pNode->visited)
        {
            /* We've hit a case where we need to report this data */
            *pData = pNode->data;
            *pPrefixLength = bitPos - 1;
            pNode->visited = 1;  /* it is now */
            return IX_SUCCESS;
        }
    }
    
    /* No data found yet, check the left sub-tree first */
    if (pNode->left != NULL)
    {
        CLEAR_BIT_AT_POSITION(pIpv6Addr, bitPos - 1);
    
        /* Continue through the tree .. */
        err = __swlpmv6_dfs(hLpmv6, pNode->left, 
                            pIpv6Addr, pPrefixLength, bitPos, pData);
    }

    if (err != IX_SUCCESS)
    { 
        return err;
    }
    else if (*pData != NULL_DATA)
    {
        return IX_SUCCESS; 
    }
    
    /* 
     * If nothing is found yet and there is a right sub-tree, go down this
     * the right sub-tree.
     */
    if ((*pData == NULL_DATA) && (pNode->right != NULL))
    {
        SET_BIT_AT_POSITION(pIpv6Addr, bitPos - 1);
    
        /* Continue through the tree .. */
        err = __swlpmv6_dfs(hLpmv6, pNode->right, 
                            pIpv6Addr, pPrefixLength, bitPos, pData);
    }
   
    if (err != IX_SUCCESS)
    {
        return err;
    }
    else if (*pData != NULL_DATA)
    {
        return IX_SUCCESS;
    }

    /* 
     * If we've reached the top of the tree and we're here, clearly
     * no data is found 
     */
    if (pNode == hLpmv6->gRoot) 
    {
        return IX_ERROR_WARNING(IX_LKUP_ERROR_NOTFOUND,
                                     ("Reached end of search \n"));
    }

    return IX_SUCCESS;
}
#endif
    
/* NAME: __swlpmv6_prune_ipv6_address
 *
 * DESCRIPTION: Internal routine to prune an IPv6 address for pretty 
 * display.
 *
 * @Param:      - IN    None.
 * @Param:      - OUT   None. 
 *
 * @Returns:
 * None.
 */
void
__swlpmv6_prune_ipv6_address(ix_uint32 *pIpv6Addr, ix_uint8 prefixLength)
{
    ix_uint32 mask;
    ix_uint8 shift, index;

    index = prefixLength / 32;
    shift = (32 - (prefixLength % 32));
    
    mask = ~((1 << shift) - 1);

    pIpv6Addr[index] &= mask;

    /* clear out the rest of the IPv6 address */
    for (shift = index + 1; shift < 4; shift++)
    {
        pIpv6Addr[shift] = 0x0;
    }

    return;
}

/*
 * NAME: __swlpmv6_read_first_entry
 *
 * DESCRIPTION: This routine reads the first entry.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 * @Param:      - OUT   pIpv6Addr
 *                      The IPv6 address that is set as a result of the search.
 * @Param:      - OUT   pPrefixLength
 *                      The prefix Length that is set as a resulf of the search.
 * @Param:      - OUT   pWeight
 *                      Unused.
 * @Param:      - OUT   pData
 *                      The data resturned as a result of the search.
 *
 * @Returns:
 * IX_LKUP_ERROR_INTERNAL
 * IX_SUCCESS
 */
ix_error 
__swlpmv6_read_first_entry(_ix_lkup_swlpmv6 hLpmv6,
                           ix_uint32 *pIpv6Addr,
                           ix_uint8 *prefixLength,
                           int *pWeight,
                           ix_uint16 *pData)
{
    /*
     * Start out by setting the "start_at" pointer to point the global 
     * root.
     */
    return (__swlpmv6_read_next_entry(hLpmv6, 
                                      pIpv6Addr, 
                                      prefixLength, 
                                      pWeight, 
                                      pData));

}

/*
 * NAME: __swlpmv6_read_next_entry
 *
 * DESCRIPTION: This routine reads the next entry from the 1-bit trie.
 *
 * @Param:      - IN    hLpmv6
 *                      Handle to the SW LPM V6 structure.
 * @Param:      - OUT   pIpv6Addr
 *                      The IPv6 address that is set as a result of the search.
 * @Param:      - OUT   pPrefixLength
 *                      The prefix Length that is set as a resulf of the search.
 * @Param:      - OUT   pWeight
 *                      Unused.
 * @Param:      - OUT   pData
 *                      The data resturned as a result of the search.
 *        
 * @Returns:
 * IX_LKUP_ERROR_INTERNAL
 * IX_SUCCESS
 */
ix_error 
__swlpmv6_read_next_entry(_ix_lkup_swlpmv6 hLpmv6,
                           ix_uint32 *pIpv6Addr,
                           ix_uint8 *pPrefixLength,
                           int *pWeight,
                           ix_uint16 *pData)
{
    ix_error err = IX_SUCCESS;
    ix_uint16 bitPos;

    /* 
     * Start out by zero'ing out the IPv6 address, data and the prefix 
     * length.
     */

    ix_ossl_memset(pIpv6Addr, 0, 4 * sizeof(ix_uint32));
    *pPrefixLength = 0;
    *pData = 0;

    /* Start from the top */
    bitPos = 0;

    /* 
     * Perform a depth starting at the gRoot 
     */
    err = __swlpmv6_dfs(hLpmv6,
                        hLpmv6->gRoot,
                        pIpv6Addr, 
                        pPrefixLength, 
                        bitPos,
                        pData);

    if (err != IX_SUCCESS)
    {
        __swlpmv6_reset_visited(hLpmv6->gRoot);

        /* This can happen if err is IX_LKUP_ERROR_NOTFOUND and it is
         * normal. Basically it indicates the end of recursion
         */
        *pPrefixLength = *pData = 0;
        return err;
    }
 
    if (*pPrefixLength > 128)
    {
         return IX_ERROR_WARNING(IX_LKUP_ERROR_INTERNAL, 
                ("Prefix length (%d) being returned > 128? \n", *pPrefixLength));
    }

    /* Make sure that the output looks pretty before we return it */
    __swlpmv6_prune_ipv6_address(pIpv6Addr, *pPrefixLength);
    
    return IX_SUCCESS;
}

