/*===========================================================================
 * = COPYRIGHT INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 * ===========================================================================
 */

/* =========================================================================*/
/**
 * @file   internal_msup.h
 *
 * Internal message support structures and functions
 **/
/*=========================================================================*/
/* $Id: internal_msup.h,v 1.5 2002/12/09 23:29:44 rjpetri Exp $ */



/** Ensure the file is only included once. */
#if !defined(__INTERNAL_MSUP_H_)
#define __INTERNAL_MSUP_H_

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

#include "ix_rm.h"
#include "cc/ix_cc_genex.h"
/** Starting size factor for the Genex */
#define IX_MSUP_GENEX_START_FACTOR 6

/* ix_msup_calltype */
/**
 * Call type enumeration.
 *
 * Contains values for the different call type that the message support
 * library supports.
 */
typedef enum ix_e_msup_calltype{
    IX_MSUP_CALLTYPE_FF,    /** fire and forget */
    IX_MSUP_CALLTYPE_ASYNC, /** asynchronous */
    IX_MSUP_CALLTYPE_SYNC,  /** synchronous */
    IX_MSUB_CALLTYPE_LAST   /** last and invalid */
}ix_msup_calltype;

/* ix_msup_err_data */
/**
 * ix_error data as stored in ix_buffers
 *
 * This structure is used as a format for each ix_error that is stored
 * in an ix_buffer
 */
typedef struct ix_s_msup_err_data{
    ix_error_level  level;
    ix_error_number code;
    ix_uint32 chained;
    size_t dataLen;
 }ix_msup_err_data;

/* ix_msup_msg_header */
/**
 * Header applied to messages.
 *
 * This header is at the beginning of every buffer sent by the message 
 * support library.
 */
typedef struct ix_s_msup_msg_header{
    /** COMMID for return messages */
    ix_uint32 returnId;
    /** Call ID, used for async and sync calls */
    ix_genex callId;
    /** Start of buffer chain containing encoded ix_error chain */
    ix_buffer_handle errBuff;
    /** Size of the user data */
    ix_uint32 dataSize;
}ix_msup_msg_header;

/** Minimum buffer size for messages */
#define IX_MSUP_MIN_BUFF_SIZE (sizeof(ix_msup_msg_header))

/* ix_msup_call */
/**
 * Call state
 *
 * State structure used to store information about each outstanding call.
 */
typedef struct ix_s_msup_call{
    /** Type of call */
    ix_msup_calltype callType; 

    /** User callback */
    ix_cc_msghlp_callback cb; 

    /** User context */
    void *context;

    /** Semaphore for synchronous calls */
    ix_ossl_sem_t syncSem;

    /** Start of buffer chain containing encoded ix_error chain
     *  This is used only by synchronous calls because the decoding
     *  of the error chain must be deferred to the calling thread. */
    ix_buffer_handle errBuff;

    /** Return message to be returned to the caller */
    void *retMsg;

    /** Length of the return message */
    ix_uint32 retMsgLen;
    
}ix_msup_call;


/* ix_msup_context */
/**
 * Message support context.
 *
 * The context structure contains context for a single instantiation 
 * of the message support library.  One of these will be allocated for 
 * each execution engine.
 */
typedef struct ix_s_msup_context{
    /** commID for return messages */
    ix_uint32	returnId; 

    /** Flag indicating whether calls can be made to this API */
    ix_uint32 	callsValid;
    
    /** the genex containing the call table */
    ix_genex_pool	callPool;	

    /** freelist for message buffers */
    ix_buffer_free_list_handle hMsgFreelist;

    /** size of message buffers */
    ix_uint32 msgBuffSize;
}ix_msup_context;

ix_error _ix_cc_msup_reply_msg_hdlr(ix_buffer_handle arg_mbuf,
         	                    ix_uint32 arg_userdata,      
            	  	            void *arg_cc_context);

ix_error _ix_cc_msup_write_buffs(ix_msup_context *arg_ms_ctx,
                                 ix_msup_msg_header *arg_header,
                                 void *arg_msg,
                                 ix_uint32 arg_msglen,
                                 ix_error arg_err,
                                 ix_buffer_handle *arg_head_buff);

ix_error _ix_cc_msup_encode_error_chain(ix_error arg_err,
                                  ix_buffer_handle *arg_mbuf);

ix_error _ix_cc_msup_write_arb_data(ix_buffer_handle *arg_pmbuf,
                                    ix_uint32 *arg_pOffset,
                                    ix_uint32 arg_dataSize,
                                    void *arg_data_area);

ix_error _ix_cc_msup_read_arb_data(ix_buffer_handle *arg_pmbuf,
                                   ix_uint32 *arg_pOffset,
                                   ix_uint32 arg_dataSize,
                                   void *arg_data_area);

ix_error _ix_cc_msup_reply_msg_hdlr(ix_buffer_handle arg_mbuf,
                                    ix_uint32 arg_userData,      
                                    void *arg_ccContext);

ix_error _ix_cc_msup_decode_error_chain(ix_buffer_handle arg_mbuf,
                                        ix_error *arg_err);



#if defined(IX_EXCLUDE_CCI)
extern void *g_Msup_Ms_Ctx;
typedef ix_uint32 ix_exe_handle;
#define IX_MSUP_EXE_GET_INFO(__hdl, __num, __ctx)\
  {                                     \
      *__hdl = IX_NULL_BUFFER_HANDLE;   \
      *__num = 0;                       \
      *__ctx = g_Msup_Ms_Ctx;           \
  }
#else
#define IX_MSUP_EXE_GET_INFO(__hdl, __num, __ctx)\
   IX_ERROR_CR(ix_cci_exe_get_info(__hdl, __num, __ctx));
#endif

#if defined(__cplusplus)
} /* extern "C" */
#endif /* __cplusplus */
#endif /* __INTERNAL_MSUP_H_ */
