/*===========================================================================
 * = COPYRIGHT INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a licence agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 * ===========================================================================
 */

/* =========================================================================*/
/**
 * @file   internal_msup.c
 *
 * Internal Message support functions.
 **/
/*=========================================================================*/
/* $Id: internal_msup.c,v 1.12 2003/07/15 23:11:43 ktseng Exp $ */

#include "ix_cc_error.h"
#include "ix_ossl.h"
#include "ix_rm.h"
#if !defined(IX_EXCLUDE_CCI)
#include "ix_cci.h"
#endif
#include "bindings.h"
#include "cc/ix_cc_genex.h"
#include "cc/ix_cc_msup.h"
#include "cc/internal/internal_msup.h"
#include "cc/internal/internal_error.h"


/*  _ix_cc_msup_write_buffs */
/*  ---------------------------------------- */
/**
 * Write header, message and errors to buffer chain.
 *
 * Allocates and chains as many buffers as are needed to store the header,
 * message data, and error.
 *
 * 
 * @param arg_ms_ctx Message support context
 * @param arg_header The message header
 * @param arg_msg The user message data
 * @param arg_msgLen Length of user message data
 * @param arg_err ix_error to encode or IX_SUCCESS
 * @param arg_headBuff Pointer to store the head of the new buffer chain
 * @return IX_SUCCESS, a pass-through ix_error, or an ix_error with one
 *         of the following values.
 * @retval NONE
**/
ix_error
_ix_cc_msup_write_buffs(ix_msup_context *arg_ms_ctx,
                        ix_msup_msg_header *arg_header,
                        void *arg_msg,
                        ix_uint32 arg_msgLen,
                        ix_error arg_err,
                        ix_buffer_handle *arg_headBuff)
{
    ix_error err = 0;
    ix_buffer_handle first_buff = IX_NULL_BUFFER_HANDLE;
    ix_buffer_handle prev_buff = IX_NULL_BUFFER_HANDLE;
    ix_buffer_handle errBuff = IX_NULL_BUFFER_HANDLE;
    void *pdata = NULL;
    void *pcurdata = NULL;
    void *pcurmsg = NULL;
    ix_uint32 space_left = 0;
    ix_uint32 offset = 0;

    IX_ERROR_CHECK_ARG_NULL(1, arg_ms_ctx);
    IX_ERROR_CHECK_ARG_NULL(2, arg_header);
    IX_ERROR_CHECK_ARG_NULL(6, arg_headBuff);

    *arg_headBuff = 0;

    /* allocate first buffer */
    IX_ERROR_CR(ix_rm_buffer_alloc(arg_ms_ctx->hMsgFreelist, &first_buff));
    /* setup buffer so the handle can be made safe later */
    IX_ERROR_CR(ix_rm_buffer_make_safe_single(first_buff));

#if !defined(IX_CONFIGURATION_unit_test)
    {
        ix_buffer_free_list_info flInfo;

        /* Initialize meta data for buffer. */
        IX_ERROR_CG(ix_rm_buffer_get_meta(first_buff, &pdata),
                    err, err_post_buff);
        if(pdata == NULL){
            err = IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_BUFF_DATA, 
                                  ("Meta data area of buffer was NULL."));
            goto err_post_buff;
        }

        /* Get info for msg freelist. */
        IX_ERROR_CG(ix_rm_buffer_free_list_get_info(arg_ms_ctx->hMsgFreelist, &flInfo),
            err, err_post_buff);

        /* Set the data size for this buffer. */
        ((ix_sw_buffer_meta*)pdata)->m_DataSize = flInfo.m_DataElementSize;
    }
#endif /* !defined(IX_CONFIGURATION_unit_test) */
    

    IX_ERROR_CG(ix_rm_buffer_get_data(first_buff, &pdata),
                err, err_post_buff);
    if(pdata == NULL){
        err = IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_BUFF_DATA, 
                              ("User data area of buffer was NULL."));
        goto err_post_buff;
    }
    prev_buff = first_buff;
    space_left = arg_ms_ctx->msgBuffSize;
    pcurdata = pdata;
    pcurmsg = arg_msg;

    arg_header->errBuff = IX_NULL_BUFFER_HANDLE;
    /* encode the error chain if there is one */
    if(arg_err){
        IX_ERROR_CG(_ix_cc_msup_encode_error_chain(arg_err,
                                                   &errBuff),
                    err, err_post_buff);

        arg_header->errBuff = errBuff;
    }
    /* write header information at the begining of first buffer */
    ix_ossl_memcpy(pcurdata, arg_header, sizeof(ix_msup_msg_header));
    pcurdata += sizeof(ix_msup_msg_header);
    offset = sizeof(ix_msup_msg_header);

    if(arg_msg == NULL){
        arg_msgLen = 0;
    }

    IX_ERROR_CG(_ix_cc_msup_write_arb_data(&prev_buff,
                                           &offset,
                                           arg_msgLen,
                                           arg_msg),
                err, err_post_buff);

    IX_ERROR_CR(ix_rm_buffer_make_safe_chained(&first_buff));
    *arg_headBuff = first_buff;
    return IX_SUCCESS;

 err_post_buff:
    *arg_headBuff = 0;
    /* free first_buff and all bufs chained to it */
    if(first_buff != IX_NULL_BUFFER_HANDLE){
        ix_rm_buffer_free_chain(first_buff);
    }
    if(arg_header->errBuff != IX_NULL_BUFFER_HANDLE){
        ix_rm_buffer_free_chain(arg_header->errBuff);
    }
    return err;
}


/*  _ix_cc_msup_encode_error */
/*  ---------------------------------------- */
/*
 * Recursivly encode each error in a chain.
 *
 * Recursivly encodes an error chain into ix_buffers
 * effectivly reversing the chain so the most signifigant
 * error is encoded first.
 *
 * 
 * @param 
 * @return IX_SUCCESS, a pass-through ix_error, or an ix_error with one
 *         of the following values.
 * @retval NONE
**/
ix_error
_ix_cc_msup_encode_error(ix_error arg_err,
                         ix_uint32 arg_buff_size,
                         ix_buffer_handle *arg_curbuff,
                         ix_uint32 *arg_offset,
                         ix_uint8 ischained)
{
    ix_error_hunk *phunk;
    size_t        size;
    ix_uint32     err_len;
    ix_uint32 ee_num;
    ix_exe_handle ee_handle;
    void *ee_context;
    ix_msup_err_data err_data;
    ix_msup_context *ms_ctx;
    ix_buffer_handle new_buff = IX_NULL_BUFFER_HANDLE;
    void *pdata = NULL;
    ix_uint32 data_left = 0;
    ix_uint32 space_left = 0;
    char *pcur_data = NULL;

    /* validate args */
    IX_ERROR_CHECK_ARG_NULL(3, arg_curbuff);
    IX_ERROR_CHECK_ARG_NULL(4, arg_offset);

    if(arg_err == IX_SUCCESS){
        return IX_ERROR_WARNING(IX_MSUP_ERROR_ENC_INVAL_ERR,
                              ("Cannot encode 0 error."));
    }
    if(arg_buff_size == 0){
        return IX_ERROR_WARNING(IX_MSUP_ERROR_ENC_INVAL_BUFFSIZE,
                              ("Cannot encode into 0 size buff."));
    }
    
    /* get context */
    IX_MSUP_EXE_GET_INFO(&ee_handle, &ee_num, &ee_context);
    if(ee_context != NULL){
        ms_ctx = (ix_msup_context*)ee_context;
    }else{
        return IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_EE_CONTEXT, 
                              ("execution engine context was NULL"));
    }
    
    /* Get error hunk pointer and data size */
    ix_error_find(arg_err, &phunk, &size);

    /* get the length of the data string */
    err_len = strlen(phunk->buf) + 1;

    /* if err_len is not long word aligned, align it (the err header
     * following this hunk needs to be long word aligned).
     */
    err_len = (((err_len + 3) >> 2) << 2);
    
    /* fill out the err_data structure in preparation for writing 
     * to the ix_buffer*/
    err_data.level = phunk->level;
    err_data.code  = phunk->code;
    err_data.dataLen = err_len;
    if(ischained != 0){
        err_data.chained = 1;
    }else{
        err_data.chained = 0;
    }
 
    /* If there is a higher signifigance error, recurse and
     * store it first.*/
    if(phunk->ref != 0){
        IX_ERROR_CR(_ix_cc_msup_encode_error(phunk->ref,
                                             arg_buff_size,
                                             arg_curbuff,
                                             arg_offset,
                                             1));
    }

    /* get current location in the buffer*/
    IX_ERROR_CR(ix_rm_buffer_get_data(*arg_curbuff, &pdata));
    if(pdata == NULL){
        return IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_BUFF_DATA, 
                              ("User data area of buffer was NULL."));
    }
    pdata += *arg_offset;

    /* store the err_data struct, allocating a new ix_buffer if needed */
    /* if there is not enough room left in the current buffer to store
     * the err_data struct, allocate a new one.
     * unlike the string data, the err_data structure does not span
     * buffer boundries.*/
    if(sizeof(err_data) > (arg_buff_size - *arg_offset)){
        /* allocate new buffer and get data pointer */
        IX_ERROR_CR(ix_rm_buffer_alloc(ms_ctx->hMsgFreelist, &new_buff));

#if !defined(IX_CONFIGURATION_unit_test)
        {
            ix_buffer_free_list_info flInfo;
            ix_error err = IX_SUCCESS;

            /* Initialize meta data for buffer. */
            IX_ERROR_CR(ix_rm_buffer_get_meta(new_buff, &pdata));
            if(pdata == NULL) {
                if(new_buff != IX_NULL_BUFFER_HANDLE) {
                    IX_ERROR_CP(ix_rm_buffer_free_chain(new_buff));
                }
                return IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_BUFF_DATA, 
                                      ("Meta data area of buffer was NULL."));
            }

            /* Get info for msg freelist. */
            err = ix_rm_buffer_free_list_get_info(ms_ctx->hMsgFreelist, &flInfo);
            if((err != IX_SUCCESS) && (new_buff != IX_NULL_BUFFER_HANDLE)) {
                IX_ERROR_CP(ix_rm_buffer_free_chain(new_buff));
                return IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
                    ("Could not get free list info\n"));
            }

            /* Set the data size for this buffer. */
            ((ix_sw_buffer_meta*)pdata)->m_DataSize = flInfo.m_DataElementSize;
        }
#endif /* !defined(IX_CONFIGURATION_unit_test) */

        IX_ERROR_CR(ix_rm_buffer_get_data(new_buff, &pdata));
        if(pdata == NULL){
            if(new_buff != IX_NULL_BUFFER_HANDLE){
                IX_ERROR_CP(ix_rm_buffer_free_chain(new_buff));
            }
            return IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_BUFF_DATA, 
                                  ("User data area of buffer was NULL."));
        }
        IX_ERROR_CR(ix_rm_buffer_link(arg_curbuff, &new_buff));
        *arg_curbuff = new_buff;
        *arg_offset = 0;
    }

    /* write the err_data */
    ix_ossl_memcpy(pdata, &err_data, sizeof(err_data));  
    *arg_offset += sizeof(err_data);

    /* set the amount of string data remaining, and the amount of space
     * left in the current buffer */
    data_left = err_len;
    space_left = arg_buff_size - *arg_offset;
    pcur_data = phunk->buf;

    IX_ERROR_CR(_ix_cc_msup_write_arb_data(arg_curbuff,
                                           arg_offset,
                                           err_len,
                                           pcur_data));
    return IX_SUCCESS;
}


/*  _ix_cc_msup_encode_error_chain */
/*  ---------------------------------------- */
/**
 * Encodes an ix_error chain into a buffer.
 *
 * Traverses the error chain storing all data from each ix_error
 * into the ix_buffer.
 *
 * 
 * @param arg_err The ix_error chain to encode
 * @param arg_mbuf pointer to store new error buffer
 * @return IX_SUCCESS, a pass-through ix_error, or an ix_error with one
 *         of the following values.
 * @retval IX_CC_ERROR_OOM_SYSTEM Out of memory during allocation attempt.
 * @retval IX_CC_ERROR_INTERNAL an internal error has occured.
 * @retval IX_CC_ERROR_NULL A value was null that should not have been.
**/
ix_error
_ix_cc_msup_encode_error_chain(ix_error arg_err,
                               ix_buffer_handle *arg_mbuf)
{
    ix_error err;
    ix_uint32 ee_num;
    ix_exe_handle ee_handle;
    void *ee_context;
    ix_msup_context *ms_ctx;
    ix_buffer_handle errBuff = IX_NULL_BUFFER_HANDLE;
    ix_buffer_handle errBstart = IX_NULL_BUFFER_HANDLE;
    ix_uint32 offset = 0;



    /* validate args */
    IX_ERROR_CHECK_ARG_NULL(2, arg_mbuf);

    if(arg_err == IX_SUCCESS){
        return IX_SUCCESS;
    }
    /* get context */
    IX_MSUP_EXE_GET_INFO(&ee_handle, &ee_num, &ee_context);
    if(ee_context != NULL){
        ms_ctx = (ix_msup_context*)ee_context;
    }else{
        return IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_EE_CONTEXT, 
                              ("execution engine context was NULL"));
    }
    *arg_mbuf = 0;

    IX_ERROR_CR(ix_rm_buffer_alloc(ms_ctx->hMsgFreelist, &errBuff));
    /* prepare to make the handle safe later */
    IX_ERROR_CR(ix_rm_buffer_make_safe_single(errBuff));

#if !defined(IX_CONFIGURATION_unit_test)
    {
        ix_buffer_free_list_info flInfo;
        void* pdata;

        /* Initialize meta data for buffer. */
        IX_ERROR_CG(ix_rm_buffer_get_meta(errBuff, &pdata),
                    err, err_post_alloc);
        if(pdata == NULL){
            err = IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_BUFF_DATA, 
                                  ("Meta data area of buffer was NULL."));
            goto err_post_alloc;
        }

        /* Get info for msg freelist. */
        IX_ERROR_CG(ix_rm_buffer_free_list_get_info(ms_ctx->hMsgFreelist, &flInfo),
            err, err_post_alloc);

        /* Set the data size for this buffer. */
        ((ix_sw_buffer_meta*)pdata)->m_DataSize = flInfo.m_DataElementSize;
    }
#endif /* !defined(IX_CONFIGURATION_unit_test) */

    errBstart = errBuff;

    IX_ERROR_CG(_ix_cc_msup_encode_error(arg_err,
                                         ms_ctx->msgBuffSize,
                                         &errBuff,
                                         &offset,
                                         0),
                err, err_post_alloc);

    /* Make the handle safe. */
    IX_ERROR_CR(ix_rm_buffer_make_safe_chained(&errBstart));
    *arg_mbuf = errBstart;
  
    return IX_SUCCESS;

 err_post_alloc:
    if(errBuff != 0){
        ix_rm_buffer_free_chain(errBuff);
    }
    return err;
}

/*  _ix_cc_msup_decode_error_chain */
/*  ---------------------------------------- */
/**
 * Decodes an ix_error chain from a reply buffer.
 *
 * Reads each encoded ix_error in the chain and reconstructs the ix_error
 * chain in the local ix_error system.  If an error exists in the buffer,
 * the chain is decoded and the ix_error handle is stored in arg_err.  If
 * no error exists, arg_err is assigned a value of IX_SUCCESS.
 *
 * 
 * @param arg_mbuf The buffer containing the error to decode
 * @param arg_err Pointer to store the new ix_error value in
 * @return IX_SUCCESS, a pass-through ix_error, or an ix_error with one
 *         of the following values.
 * @retval IX_CC_ERROR_OOM_SYSTEM Out of memory during allocation attempt.
 * @retval IX_CC_ERROR_INTERNAL an internal error has occured.
 * @retval IX_CC_ERROR_NULL A value was null that should not have been.
**/
ix_error
_ix_cc_msup_decode_error_chain(ix_buffer_handle arg_mbuf,
                               ix_error *arg_err)
{
    ix_uint32 ee_num;
    ix_exe_handle ee_handle;
    void *ee_context;
    ix_msup_context *ms_ctx = NULL;
    ix_uint32 buff_size;
    ix_msup_err_data *err_data;
    ix_error err_new = IX_SUCCESS;
    ix_error err_ref = IX_SUCCESS; 
    size_t size;
    size_t cursize = 0;
    ix_error_hunk *phunk = NULL;

    ix_uint32 offset = 0;

    ix_buffer_handle cur_buff = IX_NULL_BUFFER_HANDLE;
    /* current buffer pointers and counter */
    /* pointer to current location in buffer */
    void *pdata = NULL; 

    IX_ERROR_DCHECK_ARG_NULL(2, arg_err);



    /* get context */
    IX_MSUP_EXE_GET_INFO(&ee_handle, &ee_num, &ee_context);
    if(ee_context != NULL){
        ms_ctx = (ix_msup_context*)ee_context;
    }else{
        return IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_EE_CONTEXT, 
                              ("execution engine context was NULL"));
    }

    *arg_err = IX_SUCCESS;    

    /* If there is no buffer, then there is no error to decode */
    if(arg_mbuf == 0){
        return IX_SUCCESS;
    }


    buff_size = ms_ctx->msgBuffSize;

    cur_buff = arg_mbuf;


    do{
        /* get current location in the buffer*/
        IX_ERROR_CR(ix_rm_buffer_get_data(cur_buff, &pdata));
        if(pdata == NULL){
            return IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_BUFF_DATA, 
                                  ("User data area of buffer was NULL."));
        }
        /* jump to offset */
        pdata += offset;

        /* overlay error_data struct */
        err_data = (ix_msup_err_data*)pdata;
        offset += sizeof(ix_msup_err_data);
        /* build a new error */
        err_new = ix_error_build(&phunk, 
                                 &size, 
                                 err_data->code, 
                                 err_data->level, 
                                 err_ref);
        err_ref = err_new;
        /* make sure the string is 0 terminated */
        phunk->buf[size] = 0;
        /* set size or truncate to max error size */
        if(size < err_data->dataLen){
            cursize = size;
        }else{
            cursize = err_data->dataLen;
        }
        /* read error data */
        IX_ERROR_CR(_ix_cc_msup_read_arb_data(&cur_buff,
                                              &offset,
                                              cursize,
                                              &phunk->buf[0]));
        *arg_err = err_ref;
        /* if the buffer is chained figure out where next header is */
        if(err_data->chained){
            /* if there is not enough room for another header in this 
             * ix_buffer, then get the next buffer in the chain */
            if((buff_size - offset) < sizeof(ix_msup_err_data)){
                IX_ERROR_CR(ix_rm_buffer_get_next(cur_buff, &cur_buff));

                offset = 0;
            }
            /* else the offset returned by read_arb_data points to the 
             * next header */
        }
     }while(err_data->chained);

    return IX_SUCCESS;
}

/*  _ix_cc_msup_read_arb_data                 */
/*  ---------------------------------------- */
/**
 * Reads an arbitrary amount of data from an ix_buffer chain.
 *
 * Begins reading the arg_mbuf at the specified offset and continues until
 * the requested amount of data has been read into arg_data.  When the
 * end of a specific buffer is reached, read the next in the chain.
 *
 * 
 * @param arg_pmbuf Pointer to the buffer containing the data
 * @param arg_pOffset Offset to start of data in arg_pmbuf
 * @param arg_dataSize Size of the data to read
 *              (size of arg_data_area)
 * @param arg_data_area Pointer to the data area to store read data.
 * @return IX_SUCCESS, a pass-through ix_error, or an ix_error with one
 *         of the following values.
 * @retval IX_MSUP_ERROR_INTERNAL
 * @retval IX_MSUP_ERROR_CHAIN_TOO_SHORT
**/
ix_error
_ix_cc_msup_read_arb_data(ix_buffer_handle *arg_pmbuf,
                          ix_uint32 *arg_pOffset,
                          ix_uint32 arg_dataSize,
                          void *arg_data_area)
{
    ix_uint32 ee_num;
    ix_exe_handle ee_handle;
    void *ee_context;
    ix_msup_context *ms_ctx;

    void *pdata;
    /* pointer to current location in ix_buffer memory */
    void *pbuff = NULL;
    /* pointer to current location in destination memory */
    void *pdata_area;

    /* handle to current buffer */
    ix_buffer_handle cur_buff = IX_NULL_BUFFER_HANDLE;
    /* handle to next buffer */
    ix_buffer_handle next_buff= IX_NULL_BUFFER_HANDLE;
    /* data area left in current buffer */
    ix_uint32 buff_data_left = 0;
    /* space left in destination buffer */
    ix_uint32 space_left = 0;
    /* size of the ix_buffers */
    ix_uint32 buff_size;

    /* validate args */
    IX_ERROR_CHECK_ARG_NULL(1, arg_pmbuf);
    IX_ERROR_CHECK_ARG_NULL(2, arg_pOffset);
    IX_ERROR_CHECK_ARG_NULL(4, arg_data_area);

    /* get execution engine context (ms_ctx) */
    IX_MSUP_EXE_GET_INFO(&ee_handle, &ee_num, &ee_context);
    if(ee_context != NULL){
        ms_ctx = (ix_msup_context*)ee_context;
    }else{
        return IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_EE_CONTEXT, 
                              ("execution engine context was NULL"));
    }

    buff_size = ms_ctx->msgBuffSize;

    /* get data pointer from first buffer */
    IX_ERROR_CR(ix_rm_buffer_get_data(*arg_pmbuf, &pdata));
    if(pdata == NULL){
        return IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_BUFF_DATA, 
                              ("User data area of buffer was NULL."));
    }


    /* set the current ix_buffer */
    cur_buff = *arg_pmbuf;
    /* set space left in data buffer */
    space_left = arg_dataSize;
    /* set pointer to data buffer */
    pdata_area = arg_data_area;
    /* set pointer to ix_buffer data area */
    pbuff = pdata;


 

    /* jump to the offset and adjust the amount of payload space*/
    pbuff += *arg_pOffset;
    buff_data_left = buff_size - *arg_pOffset;

    /* do nothing if there is no data to copy */
    if(space_left > 0){
        if(space_left < buff_data_left){
            /* the number of bytes of payload needed are smaller in 
             * size than the buffer, so copy only the remaining payload
             * from this partially used buffer */
            ix_ossl_memcpy(pdata_area, pbuff, space_left);
            /* increment databuffer and ix_buffer pointers */
            pdata_area += space_left;
            pbuff += space_left;
            /* update buff_data_left and space_left counters */
            buff_data_left -= space_left;
            space_left = 0;
        }else{
            /* the entire payload area for this buffer must be copied into
             * the message buffer*/
            ix_ossl_memcpy(pdata_area, pbuff, buff_data_left);
            /* increment databuffer and ix_buffer pointers */
            pdata_area += buff_data_left;
            pbuff += buff_data_left;
            /* update buff_data_left and space_left counters */
            space_left -= buff_data_left;
            buff_data_left = 0;
            /* get next buffer (if one exists) (priming the loop) */
            IX_ERROR_CR(ix_rm_buffer_get_next(cur_buff, &next_buff));
        }

    }
    
    /* Traverse buffer chain while there are new buffers and 
     * more data to copy */
    while((next_buff != IX_NULL_BUFFER_HANDLE) && (space_left > 0)){
        /* swap next buff into the current buffer handle */
        cur_buff = next_buff;
        next_buff = IX_NULL_BUFFER_HANDLE;
        
        /* get buffer payload data pointer */
        IX_ERROR_CR(ix_rm_buffer_get_data(cur_buff, &pdata));
        if(pdata == NULL){
            return IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_BUFF_DATA, 
                                  ("User data area of buffer was NULL."));
        }
        /* set pointer to ix_buffer data and reset buff_data_left */
        pbuff = pdata;
        buff_data_left = buff_size;
        
        if(space_left < buff_data_left){
            /* the number of bytes of payload needed are smaller in 
             * size than the buffer, so copy only the remaining payload
             * from this partially used buffer */
            ix_ossl_memcpy(pdata_area, pbuff, space_left);
            /* increment databuffer and ix_buffer pointers */
            pdata_area += space_left;
            pbuff += space_left;
            /* update buff_data_left and space_left counters */
            buff_data_left -= space_left;
            space_left = 0;
        }else{
            /* the entire payload area fro this buffer must be copied into
             * the message buffer*/
            ix_ossl_memcpy(pdata_area, pbuff, buff_data_left);
            /* increment databuffer and ix_buffer pointers */            
            pdata_area += buff_data_left;
            pbuff += buff_data_left;
            /* update buff_data_left and space_left counters */
            space_left -= buff_data_left;
            buff_data_left -= buff_data_left;
        }
        
        /* get next chained buffer if one exists */
        IX_ERROR_CR(ix_rm_buffer_get_next(cur_buff, &next_buff));
    }

    /* Make sure that we didn't run out of buffers with data before we
     * filled the data area */
    if(space_left != 0){
        return IX_ERROR_WARNING(IX_MSUP_ERROR_MSG_SIZE_MISMATCH, 
                              ("The amount of data in the buffer chain does "
                               "not match the size in the arg_dataSize '%ld'",
                               arg_dataSize));
    }

    /* set output variables to the appropriate values */
    /* arg_pmbuf is set to the last buffer read from and the offset is
     * also set appropriatly.  This allows the calling function to read
     * furthur data from the buffer chain */
    *arg_pmbuf = cur_buff;
    *arg_pOffset = pbuff - pdata;
    return IX_SUCCESS;
}


/*  _ix_cc_msup_write_arb_data                 */
/*  ---------------------------------------- */
/**
 * Writes an arbitrary amount of data to an ix_buffer chain.
 *
 * Begins by writing the data in arg_pmbuf to the specified ix_buffer.  If
 * the amount of data exceeds the size of the initial buffer, then allocate
 * new buffers and chain them until all data has been stored.
 * 
 * Upon completion of the write, arg_pmbuf and arg_pOffset are set to the 
 * last ix_buffer and the offset to the end of writen data.  This will allow
 * the caller to write additional data.
 *
 * 
 * @param arg_pmbuf Pointer to the buffer containing the error to decode
 * @param arg_pOffset Offset in arg_pmbuf to start writing data
 * @param arg_dataSize Size of data to write from arg_data_area
 * @param arg_data_area Pointer to data to write into ix_buffer chain
 * @return IX_SUCCESS, a pass-through ix_error, or an ix_error with one
 *         of the following values.
 * @retval IX_MSUP_ERROR_INTERNAL
 * @retval IX_MSUP_ERROR_CHAIN_TOO_SHORT
**/
ix_error
_ix_cc_msup_write_arb_data(ix_buffer_handle *arg_pmbuf,
                          ix_uint32 *arg_pOffset,
                          ix_uint32 arg_dataSize,
                          void *arg_data_area)
{
    ix_uint32 ee_num;
    ix_exe_handle ee_handle;
    void *ee_context;
    ix_msup_context *ms_ctx;
    void *pdata = NULL;
    /* pointer to current location in ix_buffer memory */
    void *pbuff = NULL;
    /* pointer to current location in source memory */
    void *pdata_area;

    /* handle to current buffer */
    ix_buffer_handle cur_buff = IX_NULL_BUFFER_HANDLE;
    /* handle to new buffer */
    ix_buffer_handle new_buff= 0;
    /* data left in source memory */
    ix_uint32 data_left = 0;
    /* space left in current ix_buffer memory*/
    ix_uint32 space_left = 0;
    /* size of the ix_buffers */
    ix_uint32 buff_size;

    /* validate args */
    IX_ERROR_CHECK_ARG_NULL(1, arg_pmbuf);
    IX_ERROR_CHECK_ARG_NULL(2, arg_pOffset);

    /* get execution engine context (ms_ctx) */
    IX_MSUP_EXE_GET_INFO(&ee_handle, &ee_num, &ee_context);
    if(ee_context != NULL){
        ms_ctx = (ix_msup_context*)ee_context;
    }else{
        return IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_EE_CONTEXT, 
                              ("execution engine context was NULL"));
    }

    buff_size = ms_ctx->msgBuffSize;

    /* get data pointer from first buffer */
    IX_ERROR_CR(ix_rm_buffer_get_data(*arg_pmbuf, &pdata));
    if(pdata == NULL){
        return IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_BUFF_DATA, 
                              ("User data area of buffer was NULL."));
    }
    /* set the current ix_buffer */
    cur_buff = *arg_pmbuf;
    /* set space left in data buffer */
    data_left = arg_dataSize;
    /* set pointer to data buffer */
    pdata_area = arg_data_area;
    /* set pointer to ix_buffer data area */
    pbuff = pdata;
    /* set space left in current ix_buffer */
    space_left = buff_size;
 

    /* jump to the offset and adjust the amount of payload space*/
    pbuff += *arg_pOffset;
    space_left = buff_size - *arg_pOffset;

    /* do nothing if there is no data to copy */
    if(data_left > 0){
        if(data_left < space_left){
            /* the number of bytes of payload needed are smaller in 
             * size than the buffer, so copy only the remaining payload
             * to this partially used buffer */
            ix_ossl_memcpy(pbuff, pdata_area, data_left);
            /* increment databuffer and ix_buffer pointers */
            pdata_area += data_left;
            pbuff += data_left;
            /* update data_left and space_left counters */
            space_left -= data_left;
            data_left = 0;
        }else{
            /* the entire payload area for this buffer must be filled 
             * because there is more or equal data_left than the buffer
             * has remaining */
            ix_ossl_memcpy(pbuff, pdata_area, space_left);
            /* increment databuffer and ix_buffer pointers */
            pdata_area += space_left;
            pbuff += space_left;
            /* update data_left and space_left counters */
            data_left -= space_left;
            space_left = 0;
        }
    }

    /* while there is more data to copy, allocate buffers and chain */
    while(data_left > 0){
        IX_ERROR_CR(ix_rm_buffer_alloc(ms_ctx->hMsgFreelist, &new_buff));

#if !defined(IX_CONFIGURATION_unit_test)
        {
            ix_buffer_free_list_info flInfo;
            ix_error err = IX_SUCCESS;

            /* Initialize meta data for buffer. */
            IX_ERROR_CR(ix_rm_buffer_get_meta(new_buff, &pdata));
            if(pdata == NULL) {
                if(new_buff != IX_NULL_BUFFER_HANDLE) {
                    IX_ERROR_CP(ix_rm_buffer_free_chain(new_buff));
                }
                return IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_BUFF_DATA, 
                                      ("Meta data area of buffer was NULL."));
            }

            /* Get info for msg freelist. */
            err = ix_rm_buffer_free_list_get_info(ms_ctx->hMsgFreelist, &flInfo);
            if((err != IX_SUCCESS) && (new_buff != IX_NULL_BUFFER_HANDLE)) {
                IX_ERROR_CP(ix_rm_buffer_free_chain(new_buff));
                return IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
                    ("Could not get free list info\n"));
            }

            /* Set the data size for this buffer. */
            ((ix_sw_buffer_meta*)pdata)->m_DataSize = flInfo.m_DataElementSize;
        }
#endif /* !defined(IX_CONFIGURATION_unit_test) */

        IX_ERROR_CR(ix_rm_buffer_get_data(new_buff, &pdata));
        if(pdata == NULL){
            if(new_buff != IX_NULL_BUFFER_HANDLE){
                IX_ERROR_CP(ix_rm_buffer_free_chain(new_buff));
            }
            return IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_BUFF_DATA, 
                                  ("User buffer data was NULL."));
        }
        /* link new buffer to previous buffer */
        IX_ERROR_CR(ix_rm_buffer_link(&cur_buff, &new_buff));

        /* set pointer to ix_buffer data area */
        pbuff = pdata;

        cur_buff = new_buff;
        *arg_pmbuf = new_buff;
        *arg_pOffset = 0;

        space_left = buff_size;

        if(data_left < space_left){
            /* the number of bytes of payload needed are smaller in 
             * size than the buffer, so copy only the remaining payload
             * to this partially used buffer */
            ix_ossl_memcpy(pbuff, pdata_area, data_left);
            /* increment databuffer and ix_buffer pointers */
            pdata_area += data_left;
            pbuff += data_left;
            /* update data_left and space_left counters */
            space_left -= data_left;
            data_left = 0;
        }else{
            /* the entire payload area for this buffer must be filled 
             * because there is more or equal data_left than the buffer
             * has remaining */
            ix_ossl_memcpy(pbuff, pdata_area, space_left);
            /* increment databuffer and ix_buffer pointers */
            pdata_area += space_left;
            pbuff += space_left;
            /* update data_left and space_left counters */
            data_left -= space_left;
            space_left = 0;
        }
    }








    /* set output variables to the appropriate values */
    *arg_pmbuf = cur_buff;
    *arg_pOffset = pbuff - pdata;
    return IX_SUCCESS;
}




/*  _ix_cc_msup_reply_msg_hdlr */
/*  ---------------------------------------- */
/**
 * Message handler for reply messages.
 *
 * This message handler is registered for each execution engines reply CommID.
 * It handles replies to calls made from a given execution engine, eventually
 * calling the callback function.
 *
 * 
 * @param arg_mbuf The buffer containing the return message
 * @param arg_userData User value containing the callId
 * @param arg_ccContext The cc context (actually and EE context in this case)
 * @return IX_SUCCESS, a pass-through ix_error, or an ix_error with one
 *         of the following values.
 * @retval NONE
**/
ix_error
_ix_cc_msup_reply_msg_hdlr(ix_buffer_handle arg_mbuf,
                           ix_uint32 arg_userData,      
                           void *arg_ccContext)
{
    ix_error err = IX_SUCCESS;
    ix_error uerr = IX_SUCCESS;
    ix_msup_context *ms_ctx;

    ix_genex callID = 0;
    ix_msup_call *call;
    ix_msup_msg_header *header;
    void *pmsg;
    

    IX_ERROR_DCHECK_ARG_NULL(3, arg_ccContext);
    
    ms_ctx = (ix_msup_context*)arg_ccContext;

    callID = (ix_genex)arg_userData;

    /* lookup callID */
    IX_ERROR_CG(ix_genex_lookup(ms_ctx->callPool,
                                callID,
                                (void**)&call),
                err, err_post_alloc);
        
    IX_ERROR_CR(ix_cc_msup_extract_msg(arg_mbuf,
                                       &pmsg,
                                       (void**)&header));

    switch(call->callType){
    case IX_MSUP_CALLTYPE_ASYNC:
        /* if there is an error, extract it */
        if(header->errBuff != IX_NULL_BUFFER_HANDLE){
            IX_ERROR_CG(_ix_cc_msup_decode_error_chain(header->errBuff,
                                                       &uerr),
                        err, err_post_alloc);
        }
        /* invoke user callback */
        if(call->cb != NULL){
            IX_ERROR_CG(call->cb(uerr, call->context, pmsg, header->dataSize),
                        err, err_post_alloc);
        }
        if(pmsg != NULL){
            ix_ossl_free(pmsg);
            pmsg = NULL;
        }
        if(callID != 0){
            /* release the callID */
	    IX_ERROR_CG(ix_genex_del(ms_ctx->callPool, callID),
                        err, err_post_alloc);
        }
        break;


    case IX_MSUP_CALLTYPE_SYNC:
        call->retMsg = pmsg;
        call->retMsgLen = header->dataSize;
        call->errBuff = header->errBuff;
        IX_ERROR_CR(ix_ossl_sem_give(call->syncSem));
        break;

    default:
        return IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
                              ("Call type invalid during post-processing: 0x%08lx",
                               call->callType));
    }
    
    if(header->errBuff != IX_NULL_BUFFER_HANDLE){
        IX_ERROR_CR(ix_rm_buffer_free_chain(header->errBuff));
    }

    if(arg_mbuf != IX_NULL_BUFFER_HANDLE){
        IX_ERROR_CR(ix_rm_buffer_free_chain(arg_mbuf));
    }

    return IX_SUCCESS;

 err_post_alloc:

    if(header->errBuff != IX_NULL_BUFFER_HANDLE){
        IX_ERROR_CR(ix_rm_buffer_free_chain(header->errBuff));
    }

    if(arg_mbuf != IX_NULL_BUFFER_HANDLE){
        IX_ERROR_CR(ix_rm_buffer_free_chain(arg_mbuf));
    }

    if(pmsg != NULL){
        ix_ossl_free(pmsg);
        pmsg = NULL;
    }
    if(callID != 0){
        /* release the callID */
        IX_ERROR_CP(ix_genex_del(ms_ctx->callPool, callID));
    }
    return err;
}


