/*===========================================================================
 * = COPYRIGHT INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *       (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 * ===========================================================================
 */

/* =========================================================================*/
/**
 * @file   msup.c
 *
 * Contains all message support library APIs
 *
 * {{{detail  here}}}
 **/
/*=========================================================================*/
/* $Id: msup.c,v 1.12 2003/07/15 23:11:43 ktseng Exp $ */

#include "ix_cc_error.h"
#include "ix_ossl.h"
#include "ix_rm.h"
#if !defined(IX_EXCLUDE_CCI)
#include "ix_cci.h"
#endif
#include "bindings.h"
#include "cc/ix_cc_genex.h"
#include "cc/ix_cc_msup.h"
#include "cc/internal/internal_msup.h"


#if defined(IX_EXCLUDE_CCI)
void *g_Msup_Ms_Ctx;
#endif


/*  ix_cc_msup_init */
/*  ---------------------------------------- */
/**
 * Initialize the message support library for the current execution engine.
 *
 * The message support library must be initialized from each execution engine.
 * The local structures are allocated and a return path is initialized for
 * the execution engine by registering a message handler for all reply 
 * messages destined for this EE.
 *
 * @param arg_hMsgFreelist Message freelist.
 * @param arg_msgBuffSize Size of message buffers on the freelist
 * @param arg_context Pointer to a void* context.  The message support
 *                    state struct pointer is stored here.
 * @return IX_SUCCESS, a pass-through ix_error or an ix_error with one 
 *         of the following values.
 * @retval IX_CC_ERROR_OOM_SYSTEM Out of memory during allocation attempt.
 * @retval IX_CC_ERROR_INVAL_BUFF_SIZE Msg Buffer size is too small to use.
 * @retval IX_CC_ERROR_INTERNAL an internal error has occured.
**/
ix_error
ix_cc_msup_init(ix_buffer_free_list_handle arg_hMsgFreelist,
                ix_uint32 arg_msgBuffSize,
                void **arg_context)
{
    ix_uint32 ee_num;
    ix_exe_handle ee_handle;
    void *ee_context;
    ix_msup_context *ms_ctx;
    ix_error err;
    ix_uint32 msupInputIds[IX_CC_MSUP_NUM_EE_INPUTS] = {IX_CC_MSUP_EE_0_INPUT,
                                                        IX_CC_MSUP_EE_1_INPUT,
                                                        IX_CC_MSUP_EE_2_INPUT,
                                                        IX_CC_MSUP_EE_3_INPUT,
                                                        IX_CC_MSUP_EE_4_INPUT};

    /* validate args */
    IX_ERROR_DCHECK_ARG_NULL(1, arg_hMsgFreelist);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_context);

    if(arg_msgBuffSize < IX_MSUP_MIN_BUFF_SIZE){
        return IX_ERROR_WARNING(IX_MSUP_ERROR_INVAL_BUFF_SIZE,
                              ("msgBuffSize '%ld' is too small. "
                               "Must be at least '%ld'",
                               arg_msgBuffSize, 
                               IX_MSUP_MIN_BUFF_SIZE));
    }


    IX_MSUP_EXE_GET_INFO(&ee_handle, &ee_num, &ee_context);

    ms_ctx = (ix_msup_context*)ix_ossl_malloc(sizeof(ix_msup_context));
    if(ms_ctx == NULL){
        IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
                       ("OOM allocating ix_msup_context."));
    }
#if defined(IX_EXCLUDE_CCI)
    g_Msup_Ms_Ctx = (void*)ms_ctx;    
#endif


    ms_ctx->returnId = 0;
    ms_ctx->callsValid = 0;
    ms_ctx->callPool = 0;


    IX_ERROR_CG(ix_genex_init(IX_MSUP_GENEX_START_FACTOR, 
                              sizeof(ix_msup_call), 
                              &(ms_ctx->callPool)),
                err, error_post_alloc);

#if defined(IX_EXCLUDE_CCI)
    IX_ERROR_CG(ix_rm_message_handler_register(msupInputIds[ee_num],
                                               _ix_cc_msup_reply_msg_hdlr,
                                               (void*)ms_ctx),
                err, error_post_alloc);
#else
    IX_ERROR_CG(ix_cci_cc_add_message_handler(ee_handle, 
                                              msupInputIds[ee_num], 
                                              _ix_cc_msup_reply_msg_hdlr,
                                              IX_INPUT_TYPE_MULTI_SRC),
                err, error_post_genex);
#endif
    ms_ctx->returnId = msupInputIds[ee_num];
    ms_ctx->callsValid = 1;
    ms_ctx->hMsgFreelist = arg_hMsgFreelist;
    ms_ctx->msgBuffSize = arg_msgBuffSize;

    *arg_context = (void*)ms_ctx;    
    return IX_SUCCESS;

 error_post_genex:
    ix_genex_fini(ms_ctx->callPool);
 error_post_alloc:
    if(ms_ctx != NULL){
        ix_ossl_free(ms_ctx);
        ms_ctx = NULL;
    }
    return err;
}

/*  ix_cc_msup_fini */
/*  ---------------------------------------- */
/**
 * Shutdown message support for the current execution engine.
 *
 * Halt support for messaging within the current exeution engine.  State and
 * other memory allocated for this execution engine are freed and any 
 * outstanding calls are canceled.
 *
 * 
 * @param arg_context The void* context provided during init.  The message 
 *                    support state struct pointer is stored here. 
 * @return IX_SUCCESS or a pass-through ix_error.
 * @retval IX_CC_ERROR_INTERNAL an internal error has occured.
**/
ix_error
ix_cc_msup_fini(void *arg_context)
{
    ix_error err = IX_SUCCESS;
    ix_msup_context *ms_ctx;
    ix_genex_iter genex_iter;
    ix_genex genex;
    void *udata;
    ix_msup_call *ms_call;
    ix_error call_err;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_context);

    ms_ctx = (ix_msup_context*)arg_context;
    ms_ctx->callsValid = 0;

#if defined(IX_EXCLUDE_CCI)
    if(ms_ctx->returnId != 0){
        IX_ERROR_CP(ix_rm_message_handler_unregister(ms_ctx->returnId));
    }
#else
    if(ms_ctx->returnId != 0){
        IX_ERROR_CP(ix_cci_cc_remove_message_handler(ms_ctx->returnId));
    }
#endif
   
    IX_ERROR_CG(ix_genex_new_iter(ms_ctx->callPool, &genex_iter), err,
                post_calls);
    IX_ERROR_CG(ix_genex_get_first(genex_iter, &genex, &udata), err,
                post_calls);
    while(udata != NULL){
        ms_call = (ix_msup_call*)udata;
        if(ms_call->cb != NULL){
            call_err = IX_ERROR_WARNING(IX_MSUP_ERROR_SHUTDOWN, 
                                      ("The message support library is "
                                       "shutting down!"));
            ms_call->cb(call_err, ms_call->context, NULL, 0);
        }
        IX_ERROR_CG(ix_genex_del(ms_ctx->callPool, genex), err, post_iter);
        IX_ERROR_CG(ix_genex_get_next(genex_iter, &genex, &udata),err,post_iter);
    }
 post_iter:
    IX_ERROR_CP(ix_genex_del_iter(genex_iter));
 post_calls:
    IX_ERROR_CP(ix_genex_fini(ms_ctx->callPool));
    ms_ctx->callPool = 0;
    ix_ossl_free(ms_ctx);
    ms_ctx = NULL;

    return IX_SUCCESS;
}




/*  ix_cc_msup_send_msg */
/*  ---------------------------------------- */
/**
 * Send a fire-and-forget message.
 *
 * No response will be expected for this type of message.  arg_msg is copied
 * into an ix_buffer.
 * The buffer is then packed with control information and sent to the 
 * destination.
 *
 * @param arg_commID The communication ID to deliver this message to.
 * @param arg_msgType A message type that the destination CC will use to 
 *                    demux the message.
 * @param arg_msg Pointer to the message.
 * @param arg_msgLen Length of the message in bytes.
 * @return IX_SUCCESS, a pass-through ix_error, or an ix_error with one
 *         of the following values.
 * @retval IX_CC_ERROR_OOM_SYSTEM Out of memory during allocation attempt.
 * @retval IX_CC_ERROR_INTERNAL an internal error has occured.
 * @retval IX_CC_ERROR_NULL A value was null that should not have been.
 * @retval IX_RM_ERROR_COMM_QUEUE_FULL could not send message because the
 *              destinations queue was full.
**/
ix_error
ix_cc_msup_send_msg(ix_uint32 arg_commID,
                    ix_uint32 arg_msgType,
                    void *arg_msg,
                    ix_uint32 arg_msgLen)
{
    ix_error err = IX_SUCCESS;
    ix_uint32 ee_num;
    ix_exe_handle ee_handle;
    void *ee_context;
    ix_msup_context *ms_ctx;
    ix_buffer_handle first_buff = 0;
    ix_msup_msg_header header;
    
    /* validate args */

    /* get execution engine context (ms_ctx) */
    IX_MSUP_EXE_GET_INFO(&ee_handle, &ee_num, &ee_context);
    if(ee_context != NULL){
        ms_ctx = (ix_msup_context*)ee_context;
    }else{
        return IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_EE_CONTEXT, 
                              ("execution engine context was NULL"));
    }


    /* write header information at the begining of first buffer */
    header.returnId = 0;
    header.callId = 0;
    header.errBuff = 0;
    header.dataSize = arg_msgLen;
    IX_ERROR_CR(_ix_cc_msup_write_buffs(ms_ctx,
                                        &header,
                                        arg_msg,
                                        arg_msgLen,
                                        IX_SUCCESS,
                                        &first_buff));

    /* send the message */
    IX_ERROR_CG(ix_rm_message_send(arg_commID, first_buff, arg_msgType),
                err, err_post_buff);

    return IX_SUCCESS;

 err_post_buff:
    /* free first_buff and all bufs chained to it */
    if(first_buff != IX_NULL_BUFFER_HANDLE){
        ix_rm_buffer_free_chain(first_buff);
        first_buff = IX_NULL_BUFFER_HANDLE;
    }
    return err;
}


/*  ix_cc_msup_send_async_msg */
/*  ---------------------------------------- */
/**
 * Send's an asynchronous message to the destination.
 *
 * Fills in a call structure containing information about the specific call
 * such as callback information.  A unique identifier for this information is
 * generated by the GENEX and added to the header portion of the ix_buffer.
 * When the return message is recieved, this structure can be looked up allowing
 * for the correct callback and context to be invoked.
 *
 * 
 * @param arg_commID The communication ID to deliver this message to.
 * @param arg_pCallback Pointer to the callback function to be invoked when the
 *                      return message is recieved.
 * @param arg_pContext The context to be passed to the callback when invoked.
 * @param arg_msgType A message type that the destination CC will use to 
 *                    demux the message.
 * @param arg_msg Pointer to the message.
 * @param arg_msgLen Length of the message in bytes.
 * @return IX_SUCCESS, a pass-through ix_error, or an ix_error with one
 *         of the following values.
 * @retval IX_CC_ERROR_OOM_SYSTEM Out of memory during allocation attempt.
 * @retval IX_CC_ERROR_INTERNAL an internal error has occured.
 * @retval IX_CC_ERROR_NULL A value was null that should not have been.
 * @retval IX_RM_ERROR_COMM_QUEUE_FULL could not send message because the
 *              destinations queue was full.
**/
ix_error
ix_cc_msup_send_async_msg(ix_uint32 arg_commID,
                          ix_cc_msghlp_callback arg_pCallback,
                          void *arg_pContext,
                          ix_uint32 arg_msgType,
                          void *arg_msg,
                          ix_uint32 arg_msgLen)
{
    ix_error err = IX_SUCCESS;
    ix_uint32 ee_num;
    ix_exe_handle ee_handle;
    void *ee_context;
    ix_msup_context *ms_ctx;
    ix_buffer_handle first_buff = IX_NULL_BUFFER_HANDLE;
    ix_msup_msg_header header;
    ix_genex callId = 0;
    ix_msup_call *call;
    /* validate args */

    IX_ERROR_DCHECK_ARG_NULL(2, arg_pCallback);

    /* get execution engine context (ms_ctx) */
    IX_MSUP_EXE_GET_INFO(&ee_handle, &ee_num, &ee_context);
    if(ee_context != NULL){
        ms_ctx = (ix_msup_context*)ee_context;
    }else{
        return IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_EE_CONTEXT, 
                              ("execution engine context was NULL"));
    }

    /* get a free callId from the genex */
    IX_ERROR_CR(ix_genex_new(ms_ctx->callPool, 
                             &callId,
                             (void**)&call));

    /* populate call structure */
    call->callType = IX_MSUP_CALLTYPE_ASYNC;
    call->cb = arg_pCallback;
    call->context = arg_pContext;

    /* Mark the return id as remote if arg_commId is a remote id */
    if (IX_COMM_REMOTE_SUBSYSTEM_MASK & arg_commID)
    {
        header.returnId = ms_ctx->returnId | 
                          IX_MAKE_BIT_FIELD32(IX_PEER_SUBSYSTEM, 
                                              IX_COMM_ID_SUBSYSTEM_ID_LSB, 
                                              IX_COMM_ID_SUBSYSTEM_ID_MSB);
    }
    else
    {
        header.returnId = ms_ctx->returnId;
    }

    /* write header information at the begining of first buffer */
    header.callId = callId;
    header.errBuff = 0;
    header.dataSize = arg_msgLen;
    IX_ERROR_CR(_ix_cc_msup_write_buffs(ms_ctx,
                                        &header,
                                        arg_msg,
                                        arg_msgLen,
                                        IX_SUCCESS,
                                        &first_buff));

    /* send the message */
    IX_ERROR_CG(ix_rm_message_send(arg_commID, first_buff, arg_msgType),
                err, err_post_buff);

    return IX_SUCCESS;

 err_post_buff:
    /* free first_buff and all bufs chained to it */
    if(first_buff != IX_NULL_BUFFER_HANDLE){
        ix_rm_buffer_free_chain(first_buff);
    }

    return err;
}


/*  ix_cc_msup_send_sync_msg */
/*  ---------------------------------------- */
/**
 * Send's a synchronous message to the destination.
 *
 * Fills in a call structure containing information about the specific call
 * such as callback information.  A unique identifier for this information is
 * generated by the GENEX and added to the header portion of the ix_buffer.
 * When the return message is recieved, this structure can be looked up allowing
 * for the correct semaphore to be released.
 *
 * 
 * @param arg_commID The communication ID to deliver this message to.
 * @param arg_msgType A message type that the destination CC will use to 
 *                    demux the message.
 * @param arg_msg Pointer to the message.
 * @param arg_msgLen Length of the message in bytes.
 * @param arg_retMsg Pointer to a pointer to the return message(must be freed by caller)
 * @param arg_retMsgLen pointer to a uint32 where the return message len in stored.
 *
 * @return IX_SUCCESS, a pass-through ix_error from the remote function, 
 *         or an ix_error with one of the following values.
 * @retval IX_CC_ERROR_OOM_SYSTEM Out of memory during allocation attempt.
 * @retval IX_CC_ERROR_INTERNAL an internal error has occured.
 * @retval IX_CC_ERROR_NULL A value was null that should not have been.
 * @retval IX_RM_ERROR_COMM_QUEUE_FULL could not send message because the
 *              destinations queue was full.
**/
ix_error
ix_cc_msup_send_sync_msg(ix_uint32 arg_commID,
                         ix_uint32 arg_msgType,
                         void *arg_msg,
                         ix_uint32 arg_msgLen,
                         void **arg_retMsg,
                         ix_uint32 *arg_retMsgLen)
{
    ix_error err = IX_SUCCESS;
    ix_error uerr = IX_SUCCESS;
    ix_uint32 ee_num;
    ix_exe_handle ee_handle;
    void *ee_context;
    ix_msup_context *ms_ctx;
    ix_buffer_handle first_buff = IX_NULL_BUFFER_HANDLE;
    ix_msup_msg_header header;
    ix_genex callId = 0;
    ix_msup_call *call;
    
    /* validate args */
    IX_ERROR_DCHECK_ARG_NULL(5, arg_retMsg);
    IX_ERROR_DCHECK_ARG_NULL(6, arg_retMsgLen);

    /* get execution engine context (ms_ctx) */
    IX_MSUP_EXE_GET_INFO(&ee_handle, &ee_num, &ee_context);
    if(ee_context != NULL){
        ms_ctx = (ix_msup_context*)ee_context;
    }else{
        return IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_EE_CONTEXT, 
                              ("execution engine context was NULL"));
    }

    /* get a free callId from the genex */
    IX_ERROR_CR(ix_genex_new(ms_ctx->callPool, 
                             &callId,
                             (void**)&call));

    /* populate call structure */
    call->callType = IX_MSUP_CALLTYPE_SYNC;
    call->cb = NULL;
    call->context = NULL;
    call->errBuff = IX_NULL_BUFFER_HANDLE;
    call->retMsg = NULL;
    call->retMsgLen = 0;

    /* initialize semaphore */
    IX_ERROR_CR(ix_ossl_sem_init(IX_OSSL_SEM_UNAVAILABLE, &(call->syncSem)));

    if (IX_COMM_REMOTE_SUBSYSTEM_MASK & arg_commID)
    {
        header.returnId = ms_ctx->returnId | 
                          IX_MAKE_BIT_FIELD32(IX_PEER_SUBSYSTEM, 
                                              IX_COMM_ID_SUBSYSTEM_ID_LSB, 
                                              IX_COMM_ID_SUBSYSTEM_ID_MSB);
    }
    else
    {
        header.returnId = ms_ctx->returnId;
    }

    /* write header information at the begining of first buffer */
    header.callId = callId;
    header.errBuff = 0;
    header.dataSize = arg_msgLen;
    IX_ERROR_CR(_ix_cc_msup_write_buffs(ms_ctx,
                                        &header,
                                        arg_msg,
                                        arg_msgLen,
                                        IX_SUCCESS,
                                        &first_buff));

    /* send the message */
    IX_ERROR_CG(ix_rm_message_send(arg_commID, first_buff, arg_msgType),
                err, err_post_buff);

    IX_ERROR_CR(ix_ossl_sem_take(call->syncSem, IX_OSSL_WAIT_FOREVER));

    IX_ERROR_CR(ix_ossl_sem_fini(call->syncSem));

    /* Lookup the genex again to ensure it is still valid etc. */
    IX_ERROR_CR(ix_genex_lookup(ms_ctx->callPool,
                                callId,
                                (void**)&call));

    *arg_retMsg = call->retMsg;
    *arg_retMsgLen = call->retMsgLen;

    /* if there is an error, extract it */
    if(call->errBuff != IX_NULL_BUFFER_HANDLE){
        IX_ERROR_CR(_ix_cc_msup_decode_error_chain(call->errBuff,
                                                   &uerr));
        
        IX_ERROR_CR(ix_rm_buffer_free_chain(call->errBuff));
                   
        IX_ERROR_CR(ix_genex_del(ms_ctx->callPool, callId));
        return uerr;
    }
    IX_ERROR_CR(ix_genex_del(ms_ctx->callPool, callId));
    return IX_SUCCESS;

 err_post_buff:
    /* free first_buff and all bufs chained to it */
    if(first_buff != IX_NULL_BUFFER_HANDLE){
        ix_rm_buffer_free_chain(first_buff);
    }

    return err;
}




/*  ix_cc_msup_send_bcast_msg */
/*  ---------------------------------------- */
/**
 * Send a fire-and-forget message to a list of communication IDs.
 *
 * No response will be expected for this type of message.  arg_msg is copied
 * into an ix_buffer and must be freed by the caller when this call returns. 
 * The buffer is then packed with control information and sent to each of the
 * desination communication IDs.
 *
 * 
 * @param arg_commIdList The list of communication IDs to deliver this 
 *                        message to.  List is terminated with a 0 valued entry.
 * @param arg_msgType A message type that the destination CC will use to 
 *                    demux the message.
 * @param arg_msg Pointer to the message.
 * @param arg_msgLen Length of the message in bytes.
 * @return IX_SUCCESS, a pass-through ix_error, or an ix_error with one
 *         of the following values.
 * @retval IX_CC_ERROR_OOM_SYSTEM Out of memory during allocation attempt.
 * @retval IX_CC_ERROR_INTERNAL an internal error has occured.
 * @retval IX_CC_ERROR_NULL A value was null that should not have been.
 * @retval IX_RM_ERROR_COMM_QUEUE_FULL could not send message because the
 *              destinations queue was full.
**/
ix_error
ix_cc_msup_send_bcast_msg(ix_uint32 *arg_commIdList,
                          ix_uint32 arg_msgType,
                          void *arg_msg,
                          ix_uint32 arg_msgLen)
{
    ix_uint32 cur_idx = 0;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_commIdList);
    
    while(arg_commIdList[cur_idx] != 0){
        IX_ERROR_CR(ix_cc_msup_send_msg(arg_commIdList[cur_idx],
                                        arg_msgType,
                                        arg_msg,
                                        arg_msgLen));
        cur_idx++;
    }
    return IX_SUCCESS;
}



/*  ix_cc_msup_send_reply_msg */
/*  ---------------------------------------- */
/**
 * Send a reply message to the caller.
 *
 * Whan a non-fire-and-forget message is sent to a core component, a reply 
 * message is expected.  This function is used to generate such a reply.  
 * Before this function can be called, the IX_MSUP_EXTRACT_MSG macro must
 * be used on the original message.  This will provide the context passed into
 * arg #1.  If an error has occured during the call, an ix_error can be 
 * supplied.  This ix_error will be fully encoded within the message and 
 * reproduced for the caller.
 *
 * @param arg_pContext The private context retrieved using IX_MSUP_EXTRACT_MSG
 *                     on the original message.
 * @param arg_msg Pointer to the return message data.
 * @param arg_msgLen Length of the return message data.
 * @param arg_retError The ix_error to return along with the message.  This 
 *                      should be IX_SUCCESS if no error has occured.
 * @return IX_SUCCESS, a pass-through ix_error, or an ix_error with one
 *         of the following values.
 * @retval IX_CC_ERROR_OOM_SYSTEM Out of memory during allocation attempt.
 * @retval IX_CC_ERROR_INTERNAL an internal error has occured.
 * @retval IX_CC_ERROR_NULL A value was null that should not have been.
 * @retval IX_RM_ERROR_COMM_QUEUE_FULL could not send message because the
 *              destinations queue was full.
**/
ix_error
ix_cc_msup_send_reply_msg(void *arg_pContext,
                          void *arg_msg,
                          ix_uint32 arg_msgLen,
                          ix_error arg_retError)
{
    ix_error err = IX_SUCCESS;
    ix_uint32 ee_num;
    ix_exe_handle ee_handle;
    void *ee_context;
    ix_msup_context *ms_ctx;
    ix_buffer_handle first_buff = 0;
    ix_msup_msg_header header;
    ix_msup_msg_header *orig_header;

    /* validate args */
    IX_ERROR_DCHECK_ARG_NULL(1, arg_pContext);

    /* get execution engine context (ms_ctx) */
    IX_MSUP_EXE_GET_INFO(&ee_handle, &ee_num, &ee_context);
    if(ee_context != NULL){
        ms_ctx = (ix_msup_context*)ee_context;
    }else{
        return IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_EE_CONTEXT, 
                              ("execution engine context was NULL"));
    }

    orig_header = (ix_msup_msg_header*)arg_pContext;

    /* write header information at the begining of first buffer */
    header.returnId = 0;
    header.callId = orig_header->callId;
    header.errBuff = 0;
    header.dataSize = arg_msgLen;
    IX_ERROR_CR(_ix_cc_msup_write_buffs(ms_ctx,
                                        &header,
                                        arg_msg,
                                        arg_msgLen,
                                        arg_retError,
                                        &first_buff));

    /* send the message */
    IX_ERROR_CG(ix_rm_message_send(orig_header->returnId, 
                                    first_buff, 
                                    orig_header->callId),
                err, err_post_buff);

    return IX_SUCCESS;

 err_post_buff:
    /* if the error was already stored, free the buffer chain */
    if(header.errBuff != IX_NULL_BUFFER_HANDLE){
        ix_rm_buffer_free_chain(header.errBuff);
    }
    /* free first_buff and all bufs chained to it */
    if(first_buff != IX_NULL_BUFFER_HANDLE){
        ix_rm_buffer_free_chain(first_buff);
    }
    return err;
}


/*  ix_cc_msup_extract_msg */
/*  ---------------------------------------- */
/**
 * Extracts user info from a raw ix_buffer.
 *
 * Reads the header information from the ix_buffer and returns a pointer
 * to the user data area and an opaque context to be passed to 
 * ix_cc_msup_send_reply_msg.  This context is actually a pointer to the 
 * message support header information within the buffer.  The data stored 
 * in the chained buffers is copied to allocated storage for contiguous 
 * presentation to the caller.  The memory returned in arg_usrMsg must be
 * freed by the caller.
 *
 * @param arg_mbuf Input prameter of the ix_buffer recived by the message 
 *                 handler.
 * @param arg_usrMsg A pointer to a pointer to the message sent by the caller.
 * @param arg_context Opaque context that user will pass to 
 *                     ix_cc_msup_send_reply_msg.
 * @return IX_SUCCESS, a pass-through ix_error, or an ix_error with one
 *         of the following values.
 * @retval IX_CC_ERROR_OOM_SYSTEM Out of memory during allocation attempt.
 * @retval IX_CC_ERROR_INTERNAL an internal error has occured.
 * @retval IX_CC_ERROR_NULL A value was null that should not have been.
**/
ix_error
ix_cc_msup_extract_msg(ix_buffer_handle arg_mbuf,
                       void **arg_usrMsg,
                       void **arg_context)
{
    ix_error err = IX_SUCCESS;
    void *pdata = NULL;
    void *pmsg = NULL;
    ix_uint32 ee_num;
    ix_exe_handle ee_handle;
    void *ee_context;
    ix_msup_context *ms_ctx;
    ix_msup_msg_header *header;
    ix_buffer_handle cur_buff = 0;
    ix_uint32 space_left = 0;
    ix_uint32 offset = 0;

    /* validate args */
    IX_ERROR_DCHECK_ARG_NULL(2, arg_usrMsg);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_context);

    /* get execution engine context (ms_ctx) */
    IX_MSUP_EXE_GET_INFO(&ee_handle, &ee_num, &ee_context);
    if(ee_context != NULL){
        ms_ctx = (ix_msup_context*)ee_context;
    }else{
        return IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_EE_CONTEXT, 
                              ("execution engine context was NULL"));
    }

    *arg_usrMsg = NULL;
    *arg_context = NULL;

    /* get data pointer from first buffer */
    IX_ERROR_CR(ix_rm_buffer_get_data(arg_mbuf, &pdata));
    if(pdata == NULL){
        return IX_ERROR_WARNING(IX_MSUP_ERROR_NULL_BUFF_DATA, 
                              ("User data area of buffer was NULL."));
    }
    cur_buff = arg_mbuf;

    /* overlay header structure on beginiing of data */
    header = (ix_msup_msg_header *)pdata;
    if(header->dataSize != 0){
        /* there is payload data, so allocate memory to store it */
        pmsg = ix_ossl_malloc(header->dataSize);
        if(pmsg == NULL){
            return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM, 
                                  ("OOM allocating data extration area."));
        }
	space_left = header->dataSize;
    }else{
        /* there is no payload so set vars as appropriate */
        pmsg = NULL;
        space_left = 0;
    }

    /* jump past the header and adjust the amount of payload space*/
    offset = sizeof(ix_msup_msg_header);
    
    /* do nothing if there is no data to copy */
    if(space_left > 0){
        IX_ERROR_CG(_ix_cc_msup_read_arb_data(&cur_buff,
                                              &offset,
                                              header->dataSize,
                                              pmsg),
                    err, error_post_alloc);
    }

    /* set output variables to the appropriate values */
    *arg_usrMsg = pmsg;
    *arg_context = header;

    return IX_SUCCESS;

 error_post_alloc:
    if(pmsg != NULL){
        ix_ossl_free(pmsg);
        pmsg = NULL;
    }
    return err;
}


