/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation.All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.0 for the IXP2000 Network Processor, Release 6
 *
 * = FILENAME
 *      ix_cc_properties.c
 *
 * = DESCRIPTION
 *      Source file for shared properties functions.
 *
 * = AUTHOR
 *       Aaron Luk
 *       aaron.luk@intel.com
 *
 * = CHANGE HISTORY
 *      12/18/2002 12:01:27 PM - creation time
 *
 * ============================================================================
 * $Id: ix_cc_properties.c,v 1.1 2003/01/02 22:40:59 lbmiller Exp $
 * $History: $
 */

#include "ix_cc_error.h"
#include "ix_ossl.h"
#include "ix_rm.h"
#include "ix_cc.h"
#include "cc/ix_cc_msup.h"
#include "ix_cc_properties.h"
#include "bindings.h"
#include "ix_cc_prop_clients.h"

/**
 * MACRO NAME: _IX_CC_NOTIFY_PROPERTY_CLIENTS
 *
 * DESCRIPTION: Sends property updates messages to clients
 *              subscribed to a specific interface property.
 *
 * @Param:  - IN arg_propId - ID of the property being updated.
 * @Param:  - IN arg_SetPropId - ID of the property being checked - 
 *          arg_SetPropId, a bitfield, is compared against arg_propId,
 *          and if arg_SetPropId is set in arg_propId,
 *          then, we send property update messages for clients of
 *          the property identified by arg_SetPropId.
 * @Param:  - IN arg_pPropMsg - pointer to property update message.
 * @Param:  - IN arg_clientList - list of client commIDs.
 * @Param:  - OUT arg_err - error code to be returned.
 *
 * @Return: An appropriate error code if it fails to send a broadcast
 *          property update message to the property clients.
 */
#define _IX_CC_NOTIFY_PROPERTY_CLIENTS( \
                                       arg_propId, \
                                       arg_SetPropId, \
                                       arg_pPropMsg, \
                                       arg_clientList, \
                                       arg_err \
                                       ) \
    do \
    { \
        if(((arg_propId) & (arg_SetPropId)) != 0) \
        { \
            /* Set ID for property message. */ \
            (arg_pPropMsg)->propID = (arg_SetPropId); \
            /* Broadcast the update message to the clients. */ \
            IX_ERROR_C(ix_cc_msup_send_bcast_msg((arg_clientList), \
                            IX_CC_COMMON_MSG_ID_PROP_UPDATE, \
                            (arg_pPropMsg), sizeof(ix_cc_properties_msg)), \
                       (arg_err)); \
        } \
    } \
    while(0)


/**
 * NAME: ix_cc_async_set_property
 *
 * DESCRIPTION: This function sends update messages to the registered clients
 * who are interested in the respective property changes.
 * Note that although this is an asynchronous function, no callback will be
 * provided to indicate success or error of the actual property updates.
 * 
 * @Param:  - IN ix_uint32 arg_propId - id of the property to set.
 * @Param:  - IN ix_cc_properties* arg_pProperty - pointer to
 *          ix_cc_properties data structure.
 * @Return: IX_SUCCESS
 *          IX_CC_ERROR_OOM_SYSTEM
 */
ix_error ix_cc_async_set_property(
                                  ix_uint32 arg_propId,
                                  ix_cc_properties* arg_pProperty
                                  )
{
    ix_error err = IX_SUCCESS;

    ix_cc_properties_msg* pPropMsg = NULL;

    if(arg_pProperty == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_NULL, ("Function invoked with NULL property structure"));

    /* Allocate memory for new property update msg. */
    pPropMsg = ix_ossl_malloc(sizeof(ix_cc_properties_msg));
    if(pPropMsg == NULL)
        return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM, ("Could not allocate memory for property update msg"));

    /* Copy property data into property update msg. */
    pPropMsg->propData = *arg_pProperty;

    /* Notify clients for adding IPv4 interface. */
    _IX_CC_NOTIFY_PROPERTY_CLIENTS(arg_propId,
                                   IX_CC_SET_PROPERTY_ADD_INTERFACE_IPV4,
                                   pPropMsg,
                                   ix_ipv4_addr_clients,
                                   err);

    if(err != IX_SUCCESS)
        goto label_1;

    /* Notify clients for deleting IPv4 interface. */
    _IX_CC_NOTIFY_PROPERTY_CLIENTS(arg_propId,
                                   IX_CC_SET_PROPERTY_DEL_INTERFACE_IPV4,
                                   pPropMsg,
                                   ix_ipv4_addr_clients,
                                   err);

    if(err != IX_SUCCESS)
        goto label_1;

    /* Notify clients for adding IPv4 gateway. */
    _IX_CC_NOTIFY_PROPERTY_CLIENTS(arg_propId,
                                   IX_CC_SET_PROPERTY_ADD_GATEWAY_IPV4,
                                   pPropMsg,
                                   ix_ipv4_gway_clients,
                                   err);

    if(err != IX_SUCCESS)
        goto label_1;

    
    /* Notify clients for deleting IPv4 gateway. */
    _IX_CC_NOTIFY_PROPERTY_CLIENTS(arg_propId,
                                   IX_CC_SET_PROPERTY_DEL_GATEWAY_IPV4,
                                   pPropMsg,
                                   ix_ipv4_gway_clients,
                                   err);

    if(err != IX_SUCCESS)
        goto label_1;


    
    /* Notify clients for adding IPv4 broadcast. */
    _IX_CC_NOTIFY_PROPERTY_CLIENTS(arg_propId,
                                   IX_CC_SET_PROPERTY_ADD_BROADCAST_IPV4,
                                   pPropMsg,
                                   ix_ipv4_bcast_clients,
                                   err);

    if(err != IX_SUCCESS)
        goto label_1;

    
    /* Notify clients for deleting IPv4 broadcast. */
    _IX_CC_NOTIFY_PROPERTY_CLIENTS(arg_propId,
                                   IX_CC_SET_PROPERTY_DEL_BROADCAST_IPV4,
                                   pPropMsg,
                                   ix_ipv4_bcast_clients,
                                   err);

    if(err != IX_SUCCESS)
        goto label_1;


#if defined(IX_IPV6_SUPPORT)

    /* Notify clients for adding IPv6 interface. */
    _IX_CC_NOTIFY_PROPERTY_CLIENTS(arg_propId,
                                   IX_CC_SET_PROPERTY_ADD_INTERFACE_IPV6,
                                   pPropMsg,
                                   ix_ipv6_addr_clients,
                                   err);

    if(err != IX_SUCCESS)
        goto label_1;

    
    /* Notify clients for deleting IPv6 interface. */
    _IX_CC_NOTIFY_PROPERTY_CLIENTS(arg_propId,
                                   IX_CC_SET_PROPERTY_DEL_INTERFACE_IPV6,
                                   pPropMsg,
                                   ix_ipv6_addr_clients,
                                   err);

    if(err != IX_SUCCESS)
        goto label_1;

    /* Notify clients for adding IPv6 gateway. */
    _IX_CC_NOTIFY_PROPERTY_CLIENTS(arg_propId,
                                   IX_CC_SET_PROPERTY_ADD_GATEWAY_IPV6,
                                   pPropMsg,
                                   ix_ipv6_gway_clients,
                                   err);

    if(err != IX_SUCCESS)
        goto label_1;

    
    /* Notify clients for deleting IPv6 gateway. */
    _IX_CC_NOTIFY_PROPERTY_CLIENTS(arg_propId,
                                   IX_CC_SET_PROPERTY_DEL_GATEWAY_IPV6,
                                   pPropMsg,
                                   ix_ipv6_gway_clients,
                                   err);

    if(err != IX_SUCCESS)
        goto label_1;


    /* Notify clients for adding IPv6 broadcast. */
    _IX_CC_NOTIFY_PROPERTY_CLIENTS(arg_propId,
                                   IX_CC_SET_PROPERTY_ADD_BROADCAST_IPV6,
                                   pPropMsg,
                                   ix_ipv6_bcast_clients,
                                   err);

    if(err != IX_SUCCESS)
        goto label_1;

    
    /* Notify clients for deleting IPv6 broadcast. */
    _IX_CC_NOTIFY_PROPERTY_CLIENTS(arg_propId,
                                   IX_CC_SET_PROPERTY_DEL_BROADCAST_IPV6,
                                   pPropMsg,
                                   ix_ipv6_bcast_clients,
                                   err);

    if(err != IX_SUCCESS)
        goto label_1;

    
#endif /* defined(IX_IPV6_SUPPORT) */

        
    /* Notify clients for MTU update. */
    _IX_CC_NOTIFY_PROPERTY_CLIENTS(arg_propId,
                                   IX_CC_SET_PROPERTY_MTU,
                                   pPropMsg,
                                   ix_mtu_clients,
                                   err);

    if(err != IX_SUCCESS)
        goto label_1;

    
    /* Notify clients for if status update. */
    _IX_CC_NOTIFY_PROPERTY_CLIENTS(arg_propId,
                                   IX_CC_SET_PROPERTY_PHYSICAL_IF_STATUS,
                                   pPropMsg,
                                   ix_physical_if_status_clients,
                                   err);

    if(err != IX_SUCCESS)
        goto label_1;

    
    /* Notify clients for MTU update. */
    _IX_CC_NOTIFY_PROPERTY_CLIENTS(arg_propId,
                                   IX_CC_SET_PROPERTY_MAC_ADDR,
                                   pPropMsg,
                                   ix_mac_addr_clients,
                                   err);

    if(err != IX_SUCCESS)
        goto label_1;


    /* Notify clients for link speed update. */
    _IX_CC_NOTIFY_PROPERTY_CLIENTS(arg_propId,
                                   IX_CC_SET_PROPERTY_LINK_SPEED,
                                   pPropMsg,
                                   ix_link_speed_clients,
                                   err);

    if(err != IX_SUCCESS)
        goto label_1;

    /* Notify clients for link status update. */
    _IX_CC_NOTIFY_PROPERTY_CLIENTS(arg_propId,
                                   IX_CC_SET_PROPERTY_LINK_STATUS,
                                   pPropMsg,
                                   ix_link_status_clients,
                                   err);

    if(err != IX_SUCCESS)
        goto label_1;

label_1:
    /* Free memory allocated to property update message. */
    ix_ossl_free(pPropMsg);
    pPropMsg = NULL;

    return err;
}
