/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      ix_cc_rtmv4.h
 *
 * = DESCRIPTION
 *      Public header file for users of the Route Table Manager for the IPv4
 *      Forwarder Core Component (RTMv4).
 *
 * ============================================================================
 * $Id: ix_cc_rtmv4.h,v 1.18 2002/12/18 01:56:31 lbmiller Exp $
 */

#if !defined(__IX_CC_RTMV4_H__)
#define __IX_CC_RTMV4_H__

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

#include "ix_cc_error.h"
#include "ix_types.h"

/*
 * Opaque handle to the RTM, obtained by calling ix_cc_rtmv4_init().
 */
typedef struct ix_s_cc_rtmv4* ix_cc_rtmv4;

/*
 * Typedef to an intrinsic data type; represents a Next Hop ID.  Applications
 * at the highest levels allocate and assign the NHID values.
 */
typedef ix_uint32 ix_cc_rtmv4_nhid;

/*
 * The RTM defines a structure describing a Next Hop.  This structure is used
 * when adding a Next Hop (see ix_cc_rtmv4_add_next_hop(), retrieving Next Hop
 * Information (see ix_cc_rtmv4_get_next_hop()), or looking up a Next Hop by IP
 * address (see ix_cc_rtmv4_lookup()).
 *
 * Though this data is identical in meaning to the Next Hop Information used by
 * the IPv4 microblock [BLDBLK], the structure used to store the data is less
 * restrictive.  The RTM packs the Next Hop data into the format the
 * microengines expect.
 *
 * It is anticipated that the IPv4 Forwarder will expose this structure in its
 * interface, requiring Forwarding Plane Module to understand the structure.
 * It may be appropriate to move the definition to a more central location
 * later.  Doing so will affect a number of RTM interfaces.
 */
typedef struct ix_s_cc_rtmv4_next_hop_info
{
    ix_uint32 bladeID;
    ix_uint32 l2Index;
    ix_uint32 portID;
    ix_uint32 mtu;
    ix_uint32 flags;
    ix_uint32 ipAddr;
    ix_uint32 l2IndexType;
} ix_cc_rtmv4_next_hop_info;

/*
 * The RTM defines a structure containing values useful in the microengines.
 * This structure is obtained by calling ix_cc_rtmv4_get_symbols().
 */
typedef struct ix_s_cc_rtmv4_symbols
{
    ix_uint32   lkupTableID;
    ix_uint32   lkupTableData1;
    ix_uint32   lkupTableData2;
    ix_uint32   nextHopBase;
} ix_cc_rtmv4_symbols;                                                                

/*
 * The client may retrieve statistics about the RTM by calling
 * ix_cc_rtmv4_get_statistics().
 */
typedef struct ix_s_cc_rtmv4_statistics
{
    ix_uint32       numberOfRoutes;
    ix_uint32       numberOfNextHops;
} ix_cc_rtmv4_statistics;                                                             

/*
 * Enumeration of the supported algorithm types.
 */
typedef enum ix_e_cc_rtmv4_lkup_type
{
    IX_CC_RTMV4_LKUP_SOFTWARE,
    IX_CC_RTMV4_LKUP_TCAM,
    IX_CC_RTMV4_LKUP_LAST
} ix_cc_rtmv4_lkup_type;                                                              

/*
 * Enumeration of the supported memory types.
 */
typedef enum ix_e_cc_rtmv4_mem_type
{
    IX_CC_RTMV4_MEM_SDRAM,
    IX_CC_RTMV4_MEM_SRAM,
    IX_CC_RTMV4_MEM_TCAM,
    IX_CC_RTMV4_MEM_LAST
} ix_cc_rtmv4_mem_type;                                                               

/* IX_CC_RTMV4_DUMP_ROUTE_SIZE
 *
 * Description:
 * Macro for calculating the minimum size of a memory block useful for dumping
 * all the routes in the RTM.  The macro adds one because
 * ix_cc_rtmv4_dump_routes() always prints the default route as well.
 *
 * Syntax:
 * IX_CC_RTMV4_DUMP_ROUTE_SIZE(numberOfRoutes)
 *
 * Input Arguments:
 * numberOfRoutes  The number of routes in the RTM, obtained by calling
 *                 ix_cc_rtmv4_get_statistics().
 *
 * Returns (R-Value):
 * The number of bytes of memory required for ix_cc_rtmv4_dump_routes().
 */
#define IX_CC_RTMV4_DUMP_ROUTE_SIZE(__numberOfRoutes)                   \
    (_IX_CC_RTMAPI_TITLE_SIZE +                                         \
        (__numberOfRoutes) * _IX_CC_RTMAPI_LINE_SIZE + 1)

/*
 * IX_CC_RTMV4_DUMP_NEXT_HOP_SIZE
 *
 * Description:
 * Macro for calculating the minimum size of a memory block useful for dumping
 * all the next hops in the RTM.
 *
 * Syntax:
 * IX_CC_RTMV4_DUMP_NEXT_HOP_SIZE(numberOfNextHops)
 *
 * Input Arguments:
 * numberOfNextHops  The number of next hops in the RTM, obtained by
 *                   calling ix_cc_rtmv4_get_statistics().
 *
 * Returns (R-Value):
 * The number of bytes of memory required for ix_cc_rtmv4_dump_next_hops().
 */
#define IX_CC_RTMV4_DUMP_NEXT_HOP_SIZE(__numberOfNextHops)              \
    (_IX_CC_NHDB_TITLE_SIZE + (__numberOfNextHops) * _IX_CC_NHDB_LINE_SIZE + 1)

/*
 * IX_CC_RTMV4_NHID_NO_ROUTE
 *
 * Description:
 * The RTM provides a macro for a reserved Next Hop ID value.
 */
#define IX_CC_RTMV4_NHID_NO_ROUTE 0xFFFFFFFFUL

/*
 * IX_CC_RTMV4_L2INDEX_NO_ROUTE
 *
 * Description:
 * The RTM provides a macro for a reserved L2Index value.
 */
#define IX_CC_RTMV4_L2INDEX_NO_ROUTE 0x0000FFFFUL

typedef enum ix_e_cc_rtmv4_error
{
    IX_CC_RTMV4_ERROR_INVALID_NH_INFO = IX_ERROR_MAKE_GROUP(CC_RTMV4),
    IX_CC_RTMV4_ERROR_NHID_IN_USE,
    IX_CC_RTMV4_ERROR_INIT_LKUP,
    IX_CC_RTMV4_ERROR_CREATE_TABLE,
    IX_CC_RTMV4_ERROR_CREATE_NHDB,
    IX_CC_RTMV4_ERROR_LKUP_UNAVAILABLE,
    IX_CC_RTMV4_ERROR_INVALID_MASK,

    IX_CC_RTMV4_ERROR_LAST
} ix_cc_rtmv4_error;

/*
 * NAME: ix_cc_rtmv4_config
 *
 * DESCRIPTION: The ix_cc_rtmv4_config struct is used to indicate to
 * ix_cc_rtmv4_init() the desired configuration of the resulting route table
 * manager.
 *
 * routeTableSizeHint      Indicates approximately how many routes will be
 *                         supported.  The quantity indicated is not
 *                         guaranteed, but may be used to
 *                         determine the number of resources required.
 * lkupType        Type of lookup to use.  This must match the algorithm used
 *                 by the microengines.
 * lkupMemType     Type of memory to use in the lookup algorithm.  This must
 *                 match the memory type used by the microengines.
 * lkupChannel     Channel number where the microblock will access the LPM
 *                 tables.
 * nhdbSizeHint    Indicates approximately how many next hops will be
 *                 supported.  The quantity indicated is not guaranteed, but
 *                 may be used to determine the number of resources required.
 * nhdbMemType     Type of memory to use in the Next Hop Database.  This must
 *                 match the memory type used by the microengines.
 * nhdbChannel     Channel number where the microblock will access the Next
 *                 Hop table.
 */
typedef struct ix_s_cc_rtmv4_config
{
    ix_uint32 routeTableSizeHint;
    ix_cc_rtmv4_lkup_type lkupType;
    ix_cc_rtmv4_mem_type lkupMemType;
    ix_uint32 lkupChannel;
    ix_uint32 nhdbSizeHint;
    ix_cc_rtmv4_mem_type nhdbMemType;
    ix_uint32 nhdbChannel;
} ix_cc_rtmv4_config;

/*
 * NAME: ix_cc_rtmv4_init
 *
 * DESCRIPTION: Creates a new RTM for the client.  The client must initialize
 * the RTM before performing any other actions.
 *
 * @Param:      - IN    arg_pConfig
 *                      Configuration for the RTM to be created, indicating
 *                      the amount and type of memory to use.
 * @Param:      - OUT   pRtm
 *                      Handle to the RTM.
 *
 * Error codes:
 * IX_CC_ERROR_RANGE
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_RTMV4_ERROR_LKUP_UNAVAILABLE
 * IX_CC_RTMV4_ERROR_CREATE_NHDB
 * IX_CC_RTMV4_ERROR_INIT_LKUP
 */
ix_error ix_cc_rtmv4_init(
    const ix_cc_rtmv4_config *arg_pConfig,
    ix_cc_rtmv4 *arg_phRtm);

/*
 * NAME: ix_cc_rtmv4_fini
 *
 * DESCRIPTION: Destroys an existing RTM, freeing all previously allocated
 * resources.  The client must not finalize the RTM until all microblocks
 * have been disabled.  After returning, the arg_hRtm handle is no longer
 * valid.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init()
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_fini(
    ix_cc_rtmv4 arg_hRtm);

/*
 * NAME: ix_cc_rtmv4_add_next_hop
 *
 * DESCRIPTION: Adds Next Hops to the RTM's next hop database.  Typically,
 * clients would do this any time a new router is detected on a network.
 * Once a Next Hop is entered, many routes may be added referring to that
 * Next Hop.  Next Hops are uniquely identified by their NHID, as provided in
 * the Next Hop Information Structure.  Any attempt to add a duplicate NHID
 * will result in an error.
 *
 * A next hop value of IX_CC_RTMV4_NHID_NO_ROUTE is special and the client is
 * prevented from adding, deleting or changing it using
 * ix_cc_rtmv4_add_next_hop(), ix_cc_rtmv4_delete_next_hop(), or
 * ix_cc_rtmv4_update_next_hop().
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be added.
 * @Param:      - IN    arg_pNextHopInfo
 *                      pointer to Next Hop Information structure.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_DUPLICATE_ENTRY
 * IX_CC_ERROR_CONFLICTING_ENTRY
 * IX_CC_ERROR_FULL
 * IX_CC_RTMV4_ERROR_INVALID_NH_INFO
 */
ix_error ix_cc_rtmv4_add_next_hop(
    ix_cc_rtmv4 arg_hRtm,
    ix_cc_rtmv4_nhid arg_NextHopID,
    const ix_cc_rtmv4_next_hop_info *arg_pNextHopInfo);

/*
 * NAME: ix_cc_rtmv4_delete_next_hop
 *
 * DESCRIPTION: Removes a Next Hop from the RTM, provided no routes currently
 * refer to the Next Hop.  If a client wishes to remove a next hop, it must
 * first use ix_cc_rtmv4_delete_route() to remove all routes referring to the
 * next hop, then use ix_cc_rtmv4_delete_next_hop() to remove the next hop.
 * This may require the client to maintain a complete list of routes associated
 * with next hops.  A client would typically remove a Next Hop when a peer
 * router is removed from the network.
 *
 * This call will fail if any route, including the default route, uses the
 * indicated nextHopID.
 *
 * A next hop value of IX_CC_RTMV4_NHID_NO_ROUTE is special and the client is
 * prevented from adding, deleting or changing it using
 * ix_cc_rtmv4_add_next_hop(), ix_cc_rtmv4_delete_next_hop(), or
 * ix_cc_rtmv4_update_next_hop().
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be removed; network byte
 *                      order.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_RTMV4_ERROR_NHID_IN_USE
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_delete_next_hop(
    ix_cc_rtmv4 arg_hRtm,
    ix_cc_rtmv4_nhid arg_NextHopID);

/*
 * NAME: ix_cc_rtmv4_update_next_hop
 *
 * DESCRIPTION: Updates the information contained in a next hop.  The update
 * operation requires that the next hop already exists; next hops cannot be
 * created using the update operation, use ix_cc_rtmv4_add_next_hop() to add
 * next hops.
 *
 * A next hop value of IX_CC_RTMV4_NHID_NO_ROUTE is special and the client is
 * prevented from adding, deleting or changing it using
 * ix_cc_rtmv4_add_next_hop(), ix_cc_rtmv4_delete_next_hop(), or
 * ix_cc_rtmv4_update_next_hop().
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be removed
 * @Param:      - IN    arg_pNextHopInfo
 *                      pointer to the new Next Hop Information structure.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_RTMV4_ERROR_NHID_IN_USE
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_update_next_hop(
    ix_cc_rtmv4 arg_hRtm,
    ix_cc_rtmv4_nhid arg_NextHopID,
    const ix_cc_rtmv4_next_hop_info *arg_pNextHopInfo);

/*
 * NAME: ix_cc_rtmv4_get_next_hop
 *
 * DESCRIPTION: Retrieves a structure containing a copy of the information
 * originally given via ix_cc_rtmv4_add_next_hop().
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be retrieved
 * @Param:      - OUT   arg_pNextHopInfo
 *                      results are placed into the structure provided by the
 *                      caller.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_get_next_hop(
    ix_cc_rtmv4 arg_hRtm,
    ix_cc_rtmv4_nhid arg_NextHopID,
    ix_cc_rtmv4_next_hop_info *arg_pNextHopInfo);

/*
 * NAME: ix_cc_rtmv4_set_mtu
 *
 * DESCRIPTION: Updates the MTU for a given Next Hop.  If the default route
 * matches arg_NextHopID, the default route's MTU is affected as well.
 * Because the RTM does not interpret the meaning of the MTU, it does not
 * validate the value.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be modified
 * @Param:      - IN    arg_Mtu
 *                      new maximum transmission unit for the Next Hop
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_INVALID_NH_INFO
 */
ix_error ix_cc_rtmv4_set_mtu(
    ix_cc_rtmv4 arg_hRtm,
    ix_cc_rtmv4_nhid arg_NextHopID,
    ix_uint32 arg_Mtu);

/*
 * NAME: ix_cc_rtmv4_set_flags
 *
 * DESCRIPTION: Updates the Flags field for a given Next Hop.  If the default
 * route matches arg_NextHopID, the default route's flags are affected as well.
 * The meanings of the flags are determined by the caller.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be modified
 * @Param:      - IN    arg_Flags
 *                      new flags for the Next Hop
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_INVALID_NH_INFO
 */
ix_error ix_cc_rtmv4_set_flags(
    ix_cc_rtmv4 arg_hRtm,
    ix_cc_rtmv4_nhid arg_NextHopID,
    ix_uint32 arg_Flags);

/*
 * NAME: ix_cc_rtmv4_add_route
 *
 * DESCRIPTION: Adds routes, relating a range of network addresses to a Next
 * Hop.  In order to add a route, the client must first add the Next Hop ID
 * and its corresponding data into the RTM, using ix_cc_rtmv4_add_next_hop().
 *
 * After adding a route, lookup operations, either from Core Components through
 * the RTM or from microblocks, will use the newly added route.
 *
 * Adding a duplicate route to the same or different next hop is considered an
 * error condition.  Instead, use ix_cc_rtmv4_update_route() to modify an
 * existing route.
 *
 * The client may add a default route by setting the netAddr and netMask to
 * zero.  As always, the nextHopID must refer to a valid next hop.  Setting the
 * default route is a special case and the caller needn't remove an existing
 * default route before setting the default route to a new next hop.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 * @Param:      - IN    arg_NetAddr
 *                      network address identifying the route
 * @Param:      - IN    arg_NetMask
 *                      network mask for the route
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be used for the route
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_DUPLICATE_ENTRY
 * IX_CC_ERROR_CONFLICTING_ENTRY
 * IX_CC_RTMV4_ERROR_INVALID_MASK
 * IX_CC_ERROR_FULL
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_add_route(
    ix_cc_rtmv4 arg_hRtm,
    ix_uint32 arg_NetAddr,
    ix_uint32 arg_NetMask,
    ix_cc_rtmv4_nhid arg_NextHopID);

/*
 * NAME: ix_cc_rtmv4_delete_route
 *
 * DESCRIPTION: Deletes a route.  The caller is responsible for removing the
 * next hop; deleting the next hop is not an implicit operation of
 * ix_cc_rtmv4_delete_route().
 *
 * The client may remove the default route by setting the netAddr and netMask
 * to zero.  Removing the default route is a special case and the caller may do
 * so at any time, even if the default route did not previously exist.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 * @Param:      - IN    arg_NetAddr
 *                      Network address identifying the route
 * @Param:      - IN    arg_NetMask
 *                      Subnet mask for the route
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_delete_route(
    ix_cc_rtmv4 arg_hRtm,
    ix_uint32 arg_NetAddr,
    ix_uint32 arg_NetMask);

/*
 * NAME: ix_cc_rtmv4_update_route
 *
 * DESCRIPTION: Updates routes, relating an existing range of network
 * addresses to a different Next Hop.  In order to update a route, the client
 * must first add the Next Hop ID and its corresponding data into the RTM,
 * using ix_cc_rtmv4_add_next_hop().
 *
 * After adding a route, lookup operations, either from Core Components through
 * the RTM or from microblocks, will use the newly added route.
 *
 * The client should not use the update operation to modify teh default route.
 * Use ix_cc_rtmv4_add_route() or ix_cc_rtmv4_delete_route() instead.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 * @Param:      - IN    arg_NetAddr
 *                      network address identifying the route
 * @Param:      - IN    arg_NetMask
 *                      network mask for the route
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be used for the route
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_CONFLICTING_ENTRY
 * IX_CC_RTMV4_ERROR_INVALID_MASK
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_update_route(
    ix_cc_rtmv4 arg_hRtm,
    ix_uint32 arg_NetAddr,
    ix_uint32 arg_NetMask,
    ix_cc_rtmv4_nhid arg_NextHopID);

/*
 * NAME: ix_cc_rtmv4_get_route
 *
 * DESCRIPTION: Gets the Next Hop ID for a specific route.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 * @Param:      - IN    arg_NetAddr
 *                      network address identifying the route
 * @Param:      - IN    arg_NetMask
 *                      network mask for the route
 * @Param:      - OUT   arg_pNextHopID
 *                      identifier of the Next Hop used for the route
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_RTMV4_ERROR_INVALID_MASK
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_get_route(
    ix_cc_rtmv4 arg_hRtm,
    ix_uint32 arg_NetAddr,
    ix_uint32 arg_NetMask,
    ix_cc_rtmv4_nhid *arg_pNextHopID);

/*
 * NAME: ix_cc_rtmv4_lookup
 *
 * DESCRIPTION: Looks up routing information for a given IP address.  The RTM
 * uses an LPM algorithm to match the IP address with a network (as given by
 * calling ix_cc_rtmv4_add_route()) and thereby provide the information needed
 * to forward the packet to the appropriate next hop - i.e. the corresponding
 * Next Hop Information structure.
 *
 * This function is the workhorse of the RTM, and is expected to be the
 * most-used entry point; conceivably every packet arriving on the XScale core
 * will require a lookup.  Therefore, this function is designed with efficiency
 * in mind.
 *
 * If the nextHopID field in arg_pNextHopInfo is IX_CC_RTMV4_NHID_NO_ROUTE, all
 * other fields are invalid; the IP address has no route.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 * @Param:      - IN    arg_IpAddr
 *                      IP address to be looked up
 * @Param:      - OUT   arg_pNextHopInfo
 *                      results are placed into the structure provided by the
 *                      caller.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_lookup(
    ix_cc_rtmv4 arg_hRtm,
    ix_uint32 arg_IpAddr,
    ix_cc_rtmv4_next_hop_info *arg_pNextHopInfo);

/*
 * NAME: ix_cc_rtmv4_dump_next_hops
 *
 * DESCRIPTION: Prints all Next Hops to a block of memory for debugging
 * purposes.  The memory is provided by the caller, along with the size of the
 * memory block in bytes.  The caller should use IX_CC_RTMV4_DUMP_NEXT_HOP_SIZE
 * to determine how many bytes to allocate.
 *
 * If the function returns IX_CC_ERROR_FULL, the error should be considered
 * non-fatal.  The caller may allocate more memory and call the function again.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 * @Param:      - OUT   arg_pBuffer
 *                      pointer to a block of memory where the next hops are
 *                      to be stored; human-readable ASCII-Z string.
 * @Param:      - IN    arg_BufferSize
 *                      size in bytes of the buffer; host byte order.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_RANGE
 * IX_CC_ERROR_FULL
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_dump_next_hops(
    ix_cc_rtmv4 arg_hRtm,
    char *arg_pBuffer,
    ix_uint32 arg_BufferSize);

/*
 * NAME: ix_cc_rtmv4_dump_routes
 *
 * DESCRIPTION: Prints all routes to a block of memory for debugging purposes.
 * The memory is provided by the caller, along with the size of the memory
 * block in bytes.  The caller should use IX_CC_RTMV4_DUMP_ROUTE_SIZE
 * to determine how many bytes to allocate.
 *
 * If the function returns IX_CC_ERROR_FULL, the error should be considered
 * non-fatal.  The caller may allocate more memory and call the function again.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 * @Param:      - OUT   arg_pBuffer
 *                      pointer to a block of memory where the routes are
 *                      to be stored; human-readable ASCII-Z string.
 * @Param:      - IN    arg_BufferSize
 *                      size in bytes of the buffer; host byte order.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_RANGE
 * IX_CC_ERROR_FULL
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_dump_routes(
    ix_cc_rtmv4 arg_hRtm,
    char *arg_pBuffer,
    ix_uint32 arg_BufferSize);

/*
 * NAME: ix_cc_rtmv4_purge
 *
 * DESCRIPTION: Removes all routes and next hops.  Most likely, this would
 * only be done for debugging purposes.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_purge(
    ix_cc_rtmv4 arg_hRtm);

/*
 * NAME: ix_cc_rtmv4_purge_routes
 *
 * DESCRIPTION: Removes all routes.  Most likely, this would only be done
 * for debugging purposes.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_purge_routes(
    ix_cc_rtmv4 arg_hRtm);

/*
 * NAME: ix_cc_rtmv4_get_symbols
 *
 * DESCRIPTION: Retrieves symbols to patch to the micoengines.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 * @Param:      - OUT   arg_pSymbols
 *                      structure containing values that the client may patch
 *                      to the microengines.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERAL
 */
ix_error ix_cc_rtmv4_get_symbols(
    ix_cc_rtmv4 arg_hRtm,
    ix_cc_rtmv4_symbols *arg_pSymbols);

/*
 * NAME: ix_cc_rtmv4_get_statistics
 *
 * DESCRIPTION: Retrieves statistics of the RTM.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 *
 * @Param:      - OUT   arg_pStatistics
 *                      structure containing the statistics.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_get_statistics(
    ix_cc_rtmv4 arg_hRtm,
    ix_cc_rtmv4_statistics *arg_pStatistics);

/*****************************************************************************
 * The following macros are for internal use, and should not be used directly
 * by the client of the RTMv4.
 */
#define _IX_CC_NHDB_TITLE_STRING                                        \
    "   NHID    BladeID l2Index  Port   MTU Flags  IPAddr            Refcount\n"
#define _IX_CC_NHDB_TITLE_SIZE sizeof(_IX_CC_NHDB_TITLE_STRING)
#define _IX_CC_NHDB_TERM_STRING "...\n"
#define _IX_CC_NHDB_TERM_SIZE sizeof(_IX_CC_NHDB_TERM_STRING)
#define _IX_CC_NHDB_LINE_FMT "0x%08lX  0x%02X    0x%04X %5u %5d  0x%02X  %-15s %10lu\n"
/*
 * The line size is the number of characters in each line including the \n,
 * but excluding the null terminator.
 * 10 + 2 + 4 + 4 + 6 + 1 + 5 + 1 + 5 + 2 + 4 + 2 + 15 + 1 + 10 + 1 = 73
 */
#define _IX_CC_NHDB_LINE_SIZE 73

#define _IX_CC_RTMAPI_TITLE_STRING                                      \
    "   Prefix          nextHopID\n"
#define _IX_CC_RTMAPI_TITLE_SIZE sizeof(_IX_CC_RTMAPI_TITLE_STRING)
#define _IX_CC_RTMAPI_TERM_STRING "...\n"
#define _IX_CC_RTMAPI_TERM_SIZE sizeof(_IX_CC_RTMAPI_TERM_STRING)
#define _IX_CC_RTMAPI_LINE_FMT "%-18s 0x%08lX\n"
/*
 * The line size is the number of characters in each line including the \n,
 * but excluding the null terminator.
 * 18 + 1 + 10 + 1 = 30
 */
#define _IX_CC_RTMAPI_LINE_SIZE 30

#if defined(__cplusplus)
} /* extern "C" */
#endif /* __cplusplus */

#endif /*__IX_CC_RTMV4_H__*/
