/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      nhdb.h
 *
 * = DESCRIPTION
 *      Describes the interface to the Next Hop Database, a subcomponent
 *      of the RTMv4.
 *
 * ============================================================================
 * $Id: nhdb.h,v 1.14 2003/11/19 22:28:53 ttschlue Exp $
 */

#if !defined(__IX_NHDB_H__)
#define __IX_NHDB_H__

#include "ix_cc_error.h"
#include "ix_netmacros.h"
#include "cc/ix_cc_rtmv4.h"

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

/*
 * NAME: _ix_cc_nhdb
 *
 * DESCRIPTION: Handle for the Next Hop Database; returned by
 * _ix_cc_nhdb_init() and used in all other NHDB entry points. The handle
 * is opaque to the caller.
 */
typedef struct _ix_s_cc_nhdb* _ix_cc_nhdb;

/*
 * NAME: _ix_cc_nhdb_init
 *
 * DESCRIPTION: Creates a new database by allocating shared memory
 * accessible from the microblocks
 *
 * @Param:      - IN    arg_NhdbSizeHint
 *                      Approximate number of next hops that need to be stored.
 *
 * @Param:      - IN    arg_NhdbMemType
 *                      Type of memory to use for storing the next hops - SRAM
 *                      or DRAM.
 *
 * @Param:      - IN    arg_NhdbChannel
 *                      Channel to use for the memory. Must follow the rules set
 *                      by ix_rm_mem_alloc().
 *
 * @Param:      - OUT   arg_phNhdb
 *                      Handle to the NHDB, to be used in all other calls.
 *
 * Error Codes:
 * IX_CC_ERROR_RANGE
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_OOM_SYSTEM
 * IX_CC_ERROR_OOM
 * IX_CC_ERROR_INTERNAL
 */
ix_error _ix_cc_nhdb_init(
    ix_uint32 arg_NhdbSizeHint,
    ix_cc_rtmv4_mem_type arg_NhdbMemType,
    ix_uint32 arg_NhdbChannel,
    _ix_cc_nhdb *arg_phNhdb
    );

/*
 * NAME: _ix_cc_nhdb_fini
 *
 * DESCRIPTION: Deallocates all resources previously allocated using
 * _ix_cc_nhdb_init().
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error _ix_cc_nhdb_fini(
    _ix_cc_nhdb arg_hNhdb
    );

/*
 * NAME: _ix_cc_nhdb_add_next_hop
 *
 * DESCRIPTION: Makes a next hop usable for adding routes.  The nextHopID
 * field is stored and may be used as a handle for finding the record.  When
 * complete, the reference count for the newly added Next Hop is zero.  The
 * NHDB converts the data in arg_pNextHopInfo into a format the microengines
 * can understand.
 *
 * The NHDB prevents the caller from adding duplicate NHIDs.  The NHID value
 * of IX_CC_RTMV4_NHID_NO_ROUTE is a special case and may not be added using
 * _ix_cc_nhdb_add_next_hop().
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      Identifier of the Next Hop to be added
 *
 * @Param:      - IN    arg_pNextHopInfo
 *                      pointer to the Next Hop Information structure
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_RTMV4_ERROR_INVALID_NH_INFO
 * IX_CC_ERROR_CONFLICTING_ENTRY
 * IX_CC_ERROR_DUPLICATE_ENTRY
 * IX_CC_ERROR_FULL
 */
ix_error _ix_cc_nhdb_add_next_hop(
    _ix_cc_nhdb arg_hNhdb,
    ix_cc_rtmv4_nhid arg_NextHopID,
    const ix_cc_rtmv4_next_hop_info *arg_pNextHopInfo
    );

/*
 * NAME: _ix_cc_nhdb_update_next_hop
 *
 * DESCRIPTION: Updates an existing next hop with new information. If the
 * default next hop refers to this next hop, the default is updated as well.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be updated
 *
 * @Param:      - IN    arg_pNextHopInfo
 *                      pointer to the Next Hop Information structure
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_RTMV4_ERROR_INVALID_NH_INFO
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_INTERNAL
 */
ix_error _ix_cc_nhdb_update_next_hop(
    _ix_cc_nhdb arg_hNhdb,
    ix_cc_rtmv4_nhid arg_NextHopID,
    const ix_cc_rtmv4_next_hop_info *arg_pNextHopInfo
    );

/*
 * NAME: _ix_cc_nhdb_delete_next_hop
 *
 * DESCRIPTION: Removes a next hop.  Returns an error under the following
 * circumstances:
 *      + If arg_hNhdb is invalid;
 *      + If arg_NextHopID is IX_CC_RTMV4_NHID_NO_ROUTE;
 *      + If arg_NextHopID matches the default's arg_NextHopID;
 *      + If arg_nextHopID does not refer to a valid next hop;
 *      + If arg_NextHopID refers to a next hop with a non-zero reference
 *        count.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be removed
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_RTMV4_ERROR_NHID_IN_USE
 */
ix_error _ix_cc_nhdb_delete_next_hop(
    _ix_cc_nhdb arg_hNhdb,
    ix_cc_rtmv4_nhid arg_NextHopID
    );

/*
 * NAME: _ix_cc_nhdb_nhid_to_index
 *
 * DESCRIPTION: Gets the index associated with arg_NextHopID.  This index is
 * an intermediate reference to a Next Hop, and may be used to quickly look up
 * Next Hop Information.  This is a relatively slow operation compared to
 * _ix_cc_get_nh_by_index().
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop
 *
 * @Param:      - OUT   arg_pNextHopIndex
 *                      index identifying the Next Hop
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 */
ix_error _ix_cc_nhdb_nhid_to_index(
    _ix_cc_nhdb arg_hNhdb,
    ix_cc_rtmv4_nhid arg_NextHopID,
    ix_uint32 *arg_pNextHopIndex
    );

/*
 * NAME: _ix_cc_nhdb_index_to_nhid
 *
 * DESCRIPTION: Gets the nextHopID associated with arg_nextHopIndex.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * @Param:      - IN    arg_NextHopIndex
 *                      index identifying the Next Hop
 *
 * @Param:      - OUT   arg_pNextHopID
 *                      identifier of the Next Hop
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 */
ix_error _ix_cc_nhdb_index_to_nhid(
    _ix_cc_nhdb arg_hNhdb,
    ix_uint32 arg_NextHopIndex,
    ix_cc_rtmv4_nhid *arg_pNextHopID
    );

/*
 * NAME: _ix_cc_nhdb_get_nh_by_nhid
 *
 * DESCRIPTION: Given an NHID, gets a copy of the Next Hop Information.  This
 * is conceptually the same as calling _ix_cc_nhdb_nhid_to_index() followed by
 * _ix_cc_nhdb_get_nh_by_index().
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop
 *
 * @Param:      - OUT   arg_pNextHopInfo
 *                      Next Hop Information structure
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 */
ix_error _ix_cc_nhdb_get_nh_by_nhid(
    _ix_cc_nhdb arg_hNhdb,
    ix_cc_rtmv4_nhid arg_NextHopID,
    ix_cc_rtmv4_next_hop_info *arg_pNextHopInfo
    );

/*
 * NAME: _ix_cc_nhdb_get_nh_by_index
 *
 * DESCRIPTION: Get a copy of the Next Hop Information, given an index.  This
 * is a relatively fast operation compared to _ix_cc_nhdb_nhid_to_index() and
 * _ix_cc_nhdb_get_nh_by_index().
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * @Param:      - IN    arg_NextHopIndex
 *                      index as returned by _ix_cc_nhdb_nhid_to_index()
 *
 * @Param:      - OUT   arg_pNextHopInfo
 *                      Next Hop Information structure
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 */
ix_error _ix_cc_nhdb_get_nh_by_index(
    _ix_cc_nhdb arg_hNhdb,
    ix_uint32 arg_NextHopIndex,
    ix_cc_rtmv4_next_hop_info *arg_pNextHopInfo
    );

/*
 * NAME: _ix_cc_nhdb_inc_refcount
 *
 * DESCRIPTION: Adds one to the reference count of a next hop, indicating the
 * next hop is in use.  When a next hop's reference count is non-zero, the
 * next hop may not be deleted.  If incrementing the next hop would cause it to
 * wrap to zero, ix_cc_nhdb_inc_refcount() returns IX_CC_ERROR_INTERNAL.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * @Param:      - IN    arg_NextHopIndex
 *                      index as returned by _ix_cc_nhdb_nhid_to_index()
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_INTERNAL
 */
ix_error _ix_cc_nhdb_inc_refcount(
    _ix_cc_nhdb arg_hNhdb,
    ix_cc_rtmv4_nhid arg_NextHopID
    );

/*
 * NAME: _ix_cc_nhdb_dec_refcount
 *
 * DESCRIPTION: Decrements a next hop's reference count.  When the reference
 * count reaches zero, the client may delete the next hop using
 * _ix_cc_nhdb_delete_next_hop().  If decrementing the next hop would cause it
 * to go below zero, _ix_cc_nhdb_dec_refcount() returns IX_CC_ERROR_INTERNAL.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * @Param:      - IN    arg_NextHopIndex
 *                      index as returned by _ix_cc_nhdb_nhid_to_index()
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_INTERNAL
 */
ix_error _ix_cc_nhdb_dec_refcount(
    _ix_cc_nhdb arg_hNhdb,
    ix_cc_rtmv4_nhid arg_NextHopID
    );

/*
 * NAME: _ix_cc_nhdb_set_mtu
 *
 * DESCRIPTION: Changes the MTU for a next hop.  If the nextHopID is
 * IX_CC_RTMV4_NHID_NO_ROUTE, returns IX_CC_ERROR_ENTRY_NOT_FOUND.  If the
 * nextHopID matches the default route, the default route is updated as well.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be updated
 *
 * @Param:      - IN    arg_Mtu
 *                      new Maximum Transmission Unit
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_RTMV4_ERROR_INVALID_NH_INFO
 */
ix_error _ix_cc_nhdb_set_mtu(
    _ix_cc_nhdb arg_hNhdb,
    ix_cc_rtmv4_nhid arg_NextHopID,
    ix_uint32 arg_Mtu
    );

/*
 * NAME: _ix_cc_nhdb_set_flags
 *
 * DESCRIPTION: Changes the flags for a next hop.  If the nextHopID is 
 * IX_CC_RTMV4_NHID_NO_ROUTE, returns IX_CC_ERROR_ENTRY_NOT_FOUND.  If the
 * nextHopID matches the default route, the default route is updated as well.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be updated
 *
 * @Param:      - IN    arg_Flags
 *                      new Flags
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_RTMV4_ERROR_INVALID_NH_INFO
 */
ix_error _ix_cc_nhdb_set_flags(
    _ix_cc_nhdb arg_hNhdb,
    ix_cc_rtmv4_nhid arg_NextHopID,
    ix_uint32 arg_Flags
    );

/*
 * NAME: _ix_cc_nhdb_set_default
 *
 * DESCRIPTION: Indicates the default should either not exist or be a copy of
 * the provided nextHopID.  The nextHopID argument must either be
 * IX_CC_RTMV4_NHID_NO_ROUTE, or must refer to a next hop previously added
 * using _ix_cc_nhdb_add_next_hop().
 * 
 * If the client does not specifically set the default route, the No Route
 * case (IX_CC_RTMV4_NHID_NO_ROUTE) is implied.
 *
 * If arg_NextHopID is IX_CC_RTMV4_NHID_NO_ROUTE, the nextHopID for the
 * default route is set to IX_CC_RTMV4_NHID_NO_ROUTE.  All other fields are
 * invalid.
 *
 * If arg_NextHopID is not IX_CC_RTMV4_NHID_NO_ROUTE, looks for the indicated
 * next hop.  If it does not exist, returns an error.  If it does exist,
 * copies the structure into the default location.
 *
 * The reference count for the default is not affected.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be used for the default
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 */
ix_error _ix_cc_nhdb_set_default(
    _ix_cc_nhdb arg_hNhdb,
    ix_cc_rtmv4_nhid arg_NextHopID
    );

/*
 * NAME: _ix_cc_nhdb_get_default
 *
 * DESCRIPTION: Gets the default next hop.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * @Param:      - OUT   arg_pNextHopInfo
 *                      results are placed into the provided structure
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 */
ix_error _ix_cc_nhdb_get_default(
    _ix_cc_nhdb arg_hNhdb,
    ix_cc_rtmv4_next_hop_info *arg_pNextHopInfo
    );

/*
 * NAME: _ix_cc_nhdb_get_base_address
 *
 * DESCRIPTION: Because the NHDB shares the data structure with microblocks,
 * the client must have access to the base address of the memory block used to
 * store the Next Hops.  This function provides the base address.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * @Param:      - OUT   arg_pBaseAddress
 *                      base address of the database, to be patched to
 *                      microengines
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error _ix_cc_nhdb_get_base_address(
    _ix_cc_nhdb arg_hNhdb,
    ix_uint32 *arg_pBaseAddress
    );

/*
 * NAME: _ix_cc_nhdb_dump
 *
 * DESCRIPTION: Prints all Next Hop Information Structures contained in the
 * NHDB into a string buffer.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * @Param:      - OUT   arg_pBuffer
 *                      Buffer contains textual information about the contents
 *                      of the NHDB
 *
 * @Param:      - IN    arg_BufferSize
 *                      Number of bytes in the buffer
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_RANGE
 * IX_CC_ERROR_FULL
 */
ix_error _ix_cc_nhdb_dump(
    _ix_cc_nhdb arg_hNhdb,
    char *arg_pBuffer,
    ix_uint32 arg_BufferSize
    );

/*
 * NAME: _ix_cc_nhdb_purge
 *
 * DESCRIPTION: Removes all Next Hop Information Structures from the NHDB.
 * Sets the default's l2Index to IX_CC_RTMV4_L2INDEX_NO_ROUTE.  The reference
 * counts for the next hops are not checked before the next hops are deleted.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error _ix_cc_nhdb_purge(
    _ix_cc_nhdb arg_hNhdb
    );

/*
 * NAME: _ix_cc_nhdb_reset_refcounts
 *
 * DESCRIPTION: Sets all next hop reference counts to zero.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 */
ix_error _ix_cc_nhdb_reset_refcounts(
    _ix_cc_nhdb arg_hNhdb
    );

#if defined(__cplusplus)
} /* extern "C" */
#endif /* __cplusplus */

#endif /*__IX_NHDB_H__*/
