/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      rtmapi.c
 *
 * = DESCRIPTION
 *      Public entry points for the Route Table Manager for the IPv4 Forwarder
 *      Core Component.  This sub-module, called the RTM API, invokes the
 *      Lookup Library and Next Hop Database to provide most of the
 *      functionality of the RTM.
 *
 * ============================================================================
 */

#define IX_ERROR_FILE_IDENT                                             \
        "$Id: rtmapi.c,v 1.25 2003/12/12 18:38:32 ktseng Exp $"

#include "ix_cc_error.h"
#include "ix_ossl.h"
#include "cc/ix_cc_rtmv4.h"
#include "cc/rtmv4/nhdb.h"
#include "ix_lkup.h"
#include "cc/rtmv4/internal/internal_rtmapi.h"

static ix_error _get_netmask_len(ix_uint32 netMask, int *pNetMaskLen);

/*
 * NAME: ix_cc_rtmv4_init
 *
 * DESCRIPTION: Creates a new RTM for the client.  The client must initialize
 * the RTM before performing any other actions.
 *
 * @Param:      - IN    arg_pConfig
 *                      Configuration for the RTM to be created, indicating
 *                      the amount and type of memory to use.
 *
 * @Param:      - OUT   pRtm
 *                      Handle to the RTM.
 *
 * Error codes:
 * IX_CC_ERROR_RANGE
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_RTMV4_ERROR_LKUP_UNAVAILABLE
 * IX_CC_RTMV4_ERROR_CREATE_NHDB
 * IX_CC_RTMV4_ERROR_INIT_LKUP
 */
ix_error ix_cc_rtmv4_init(
    const ix_cc_rtmv4_config *arg_pConfig,
    ix_cc_rtmv4 *arg_phRtm)
{
    ix_error err;
    ix_cc_rtmv4 hRtm;
    ix_lkup_table_conf tableConfig;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_pConfig);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_pConfig, ix_cc_rtmv4_config);
    IX_ERROR_DCHECK_ARG_NULL(2, arg_phRtm);
    IX_ERROR_DCHECK_ARG_ALIGN(2, arg_phRtm, ix_cc_rtmv4);

    /*
     * The RTM uses the Lookup Library specifically to store and retrieve
     * IPv4 routes, referring them to next hop indexes.  Create the 
     * configuration structure accordingly.
     */
    tableConfig.tableType       = IX_LKUP_TABLE_LPM;
    tableConfig.keySize         = 32;   /* IPv4 uses 4-byte IP addresses */
    tableConfig.maxWeight       = 32;   /* Weights are 1 to 32 */
    tableConfig.sizeHint        = arg_pConfig->routeTableSizeHint;
    tableConfig.dataSize        = 16;   /* 16 bits for next hop indexes */
    tableConfig.dataChannel     = arg_pConfig->lkupChannel;

    switch (arg_pConfig->lkupMemType)
    {
        case IX_CC_RTMV4_MEM_SRAM:
            tableConfig.dataLocation    = IX_LKUP_MEM_SRAM;
            break;

        case IX_CC_RTMV4_MEM_SDRAM:
            tableConfig.dataLocation    = IX_LKUP_MEM_DRAM;
            break;

        case IX_CC_RTMV4_MEM_TCAM:
            tableConfig.dataLocation    = IX_LKUP_MEM_TCAM;
            break;

        default:
            return IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
                ("The lkupMemType %d is invalid", arg_pConfig->lkupMemType));
    }

    /*
     * Allocate memory on the heap for the structure associated with the
     * RTM's handle.
     */
    hRtm = ix_ossl_malloc(sizeof(struct ix_s_cc_rtmv4));
    if (hRtm == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
            ("Could not allocate %d bytes of memory from the OSSL",
             sizeof(struct ix_s_cc_rtmv4)));
    }
    ix_ossl_memset(hRtm, 0, sizeof(struct ix_s_cc_rtmv4));

    /*
     * Initialize the Lookup Library according to the type the caller wants:
     * Software or TCAM.
     */
    switch (arg_pConfig->lkupType)
    {
        case IX_CC_RTMV4_LKUP_SOFTWARE:
            IX_ERROR_CGT(ix_lkup_sw_init(
                        _IX_CC_RTMV4_LKUP_VERSION,
                        &hRtm->hLkup),
                err, IX_CC_RTMV4_ERROR_INIT_LKUP, 0, cleanup_hRtm);
            break;

        case IX_CC_RTMV4_LKUP_TCAM:
            err = IX_ERROR_WARNING(IX_CC_RTMV4_ERROR_LKUP_UNAVAILABLE,
                ("The TCAM is not currently supported"));
            goto cleanup_hRtm;
#if 0 /* This is removed for now because ix_lkup_tcam_init() isn't implemented */
            {
                ix_lkup_tcam_params tcamParams;
                tcamParams.qdrChannel = 0; /* Assign this! */
                tcamParams.qdrDeviceSelect = 0; /* Assign this! */
                IX_ERROR_CGT(ix_lkup_tcam_init(
                            _IX_CC_RTMV4_LKUP_VERSION,
                            &tcamParams,
                            NULL,       /* No special vendor params */
                            &hRtm->hLkup),
                        err, IX_CC_RTMV4_ERROR_INIT_LKUP, 0, cleanup_hRtm);
            }
            break;
#endif

        default:
            err = IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
                ("The value of lkupType (%d) is invalid",
                 arg_pConfig->lkupType));
            goto cleanup_hRtm;
    }

    /*
     * Create the Next Hop Database.
     */
    IX_ERROR_CGT(
        _ix_cc_nhdb_init(
            arg_pConfig->nhdbSizeHint,
            arg_pConfig->nhdbMemType,
            arg_pConfig->nhdbChannel,
            &hRtm->hNhdb),
        err, IX_CC_RTMV4_ERROR_CREATE_NHDB, 0, cleanup_hLkup);

    /*
     * Create the LPM table.
     */
    IX_ERROR_CGT(
        IX_LKUP_CREATE_TABLE(
            hRtm->hLkup,
            &tableConfig,
            NULL,       /* No special vendor configuration */
            &hRtm->hTable),
        err, IX_CC_RTMV4_ERROR_CREATE_TABLE, 0, cleanup_hNhdb);

    hRtm->numNextHops   = 0;
    hRtm->numRoutes     = 0;

    /*
     * Assign the output handle.
     */
    *arg_phRtm = hRtm;
    return IX_SUCCESS;

cleanup_hNhdb:
    (void) _ix_cc_nhdb_fini(hRtm->hNhdb);
cleanup_hLkup:
    (void) IX_LKUP_FINI(hRtm->hLkup);
cleanup_hRtm:
    ix_ossl_free(hRtm);
    return err;
}

/*
 * NAME: ix_cc_rtmv4_fini
 *
 * DESCRIPTION: Destroys an existing RTM, freeing all previously allocated
 * resources.  The client must not finalize the RTM until all microblocks
 * have been disabled.  After returning, the arg_hRtm handle is no longer
 * valid.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init()
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_fini(
    ix_cc_rtmv4 arg_hRtm)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_hRtm);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hRtm, struct ix_s_cc_rtmv4);

    IX_ERROR_CRT(IX_LKUP_DESTROY_TABLE(arg_hRtm->hLkup, arg_hRtm->hTable),
        IX_CC_ERROR_INTERNAL, 0);
    arg_hRtm->hTable = NULL;

    IX_ERROR_CRT(IX_LKUP_FINI(arg_hRtm->hLkup),
        IX_CC_ERROR_INTERNAL, 0);
    arg_hRtm->hLkup = NULL;

    IX_ERROR_CRT(_ix_cc_nhdb_fini(arg_hRtm->hNhdb),
        IX_CC_ERROR_INTERNAL, 0);
    arg_hRtm->hNhdb = NULL;

    ix_ossl_free(arg_hRtm);

    return IX_SUCCESS;
}

/*
 * NAME: ix_cc_rtmv4_add_next_hop
 *
 * DESCRIPTION: Adds Next Hops to the RTM's next hop database.  Typically,
 * clients would do this any time a new router is detected on a network.
 * Once a Next Hop is entered, many routes may be added referring to that
 * Next Hop.  Next Hops are uniquely identified by their NHID, as provided in
 * the Next Hop Information Structure.  Any attempt to add a duplicate NHID
 * will result in an error.
 *
 * A next hop value of IX_CC_RTMV4_NHID_NO_ROUTE is special and the client is
 * prevented from adding, deleting or changing it using
 * ix_cc_rtmv4_add_next_hop(), ix_cc_rtmv4_delete_next_hop(), or
 * ix_cc_rtmv4_update_next_hop().
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be added.
 *
 * @Param:      - IN    arg_pNextHopInfo
 *                      pointer to Next Hop Information structure.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_DUPLICATE_ENTRY
 * IX_CC_ERROR_CONFLICTING_ENTRY
 * IX_CC_ERROR_FULL
 * IX_CC_RTMV4_ERROR_INVALID_NH_INFO
 */
ix_error ix_cc_rtmv4_add_next_hop(
    ix_cc_rtmv4 arg_hRtm,
    ix_cc_rtmv4_nhid arg_NextHopID,
    const ix_cc_rtmv4_next_hop_info *arg_pNextHopInfo)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_hRtm);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hRtm, struct ix_s_cc_rtmv4);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pNextHopInfo);
    IX_ERROR_DCHECK_ARG_ALIGN(3, arg_pNextHopInfo, ix_cc_rtmv4_next_hop_info);

    IX_ERROR_CR(_ix_cc_nhdb_add_next_hop(arg_hRtm->hNhdb,
        arg_NextHopID, arg_pNextHopInfo));

    arg_hRtm->numNextHops++;

    return IX_SUCCESS;
}

/*
 * NAME: ix_cc_rtmv4_delete_next_hop
 *
 * DESCRIPTION: Removes a Next Hop from the RTM, provided no routes currently
 * refer to the Next Hop.  If a client wishes to remove a next hop, it must
 * first use ix_cc_rtmv4_delete_route() to remove all routes referring to the
 * next hop, then use ix_cc_rtmv4_delete_next_hop() to remove the next hop.
 * This may require the client to maintain a complete list of routes associated
 * with next hops.  A client would typically remove a Next Hop when a peer
 * router is removed from the network.
 *
 * This call will fail if any route, including the default route, uses the
 * indicated nextHopID.
 *
 * A next hop value of IX_CC_RTMV4_NHID_NO_ROUTE is special and the client is
 * prevented from adding, deleting or changing it using
 * ix_cc_rtmv4_add_next_hop(), ix_cc_rtmv4_delete_next_hop(), or
 * ix_cc_rtmv4_update_next_hop().
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be removed; network byte
 *                      order.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_RTMV4_ERROR_NHID_IN_USE
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_delete_next_hop(
    ix_cc_rtmv4 arg_hRtm,
    ix_cc_rtmv4_nhid arg_NextHopID)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_hRtm);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hRtm, struct ix_s_cc_rtmv4);

    IX_ERROR_CR(_ix_cc_nhdb_delete_next_hop(arg_hRtm->hNhdb, arg_NextHopID));

    arg_hRtm->numNextHops--;

    return IX_SUCCESS;
}

/*
 * NAME: ix_cc_rtmv4_update_next_hop
 *
 * DESCRIPTION: Updates the information contained in a next hop.  The update
 * operation requires that the next hop already exists; next hops cannot be
 * created using the update operation, use ix_cc_rtmv4_add_next_hop() to add
 * next hops.
 *
 * A next hop value of IX_CC_RTMV4_NHID_NO_ROUTE is special and the client is
 * prevented from adding, deleting or changing it using
 * ix_cc_rtmv4_add_next_hop(), ix_cc_rtmv4_delete_next_hop(), or
 * ix_cc_rtmv4_update_next_hop().
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be removed
 *
 * @Param:      - IN    arg_pNextHopInfo
 *                      pointer to the new Next Hop Information structure.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_RTMV4_ERROR_NHID_IN_USE
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_update_next_hop(
    ix_cc_rtmv4 arg_hRtm,
    ix_cc_rtmv4_nhid arg_NextHopID,
    const ix_cc_rtmv4_next_hop_info *arg_pNextHopInfo)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_hRtm);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hRtm, struct ix_s_cc_rtmv4);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pNextHopInfo);
    IX_ERROR_DCHECK_ARG_ALIGN(3, arg_pNextHopInfo, ix_cc_rtmv4_next_hop_info);

    IX_ERROR_CR(_ix_cc_nhdb_update_next_hop(arg_hRtm->hNhdb, arg_NextHopID,
                arg_pNextHopInfo));

    return IX_SUCCESS;
}

/*
 * NAME: ix_cc_rtmv4_get_next_hop
 *
 * DESCRIPTION: Retrieves a structure containing a copy of the information
 * originally given via ix_cc_rtmv4_add_next_hop().
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be retrieved
 *
 * @Param:      - OUT   arg_pNextHopInfo
 *                      results are placed into the structure provided by the
 *                      caller.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_get_next_hop(
    ix_cc_rtmv4 arg_hRtm,
    ix_cc_rtmv4_nhid arg_NextHopID,
    ix_cc_rtmv4_next_hop_info *arg_pNextHopInfo)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_hRtm);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hRtm, struct ix_s_cc_rtmv4);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pNextHopInfo);
    IX_ERROR_DCHECK_ARG_ALIGN(3, arg_pNextHopInfo, ix_cc_rtmv4_next_hop_info);

    IX_ERROR_CR(_ix_cc_nhdb_get_nh_by_nhid(arg_hRtm->hNhdb, arg_NextHopID,
                arg_pNextHopInfo));

    return IX_SUCCESS;
}

/*
 * NAME: ix_cc_rtmv4_set_mtu
 *
 * DESCRIPTION: Updates the MTU for a given Next Hop.  If the default route
 * matches arg_NextHopID, the default route's MTU is affected as well.
 * Because the RTM does not interpret the meaning of the MTU, it does not
 * validate the value.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be modified
 *
 * @Param:      - IN    arg_Mtu
 *                      new maximum transmission unit for the Next Hop
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_INVALID_NH_INFO
 */
ix_error ix_cc_rtmv4_set_mtu(
    ix_cc_rtmv4 arg_hRtm,
    ix_cc_rtmv4_nhid arg_NextHopID,
    ix_uint32 mtu)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_hRtm);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hRtm, struct ix_s_cc_rtmv4);
    
    IX_ERROR_CR(_ix_cc_nhdb_set_mtu(arg_hRtm->hNhdb, arg_NextHopID, mtu));

    return IX_SUCCESS;
}

/*
 * NAME: ix_cc_rtmv4_set_flags
 *
 * DESCRIPTION: Updates the Flags field for a given Next Hop.  If the default
 * route matches arg_NextHopID, the default route's flags are affected as well.
 * The meanings of the flags are determined by the caller.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be modified
 *
 * @Param:      - IN    arg_Flags
 *                      new flags for the Next Hop
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_INVALID_NH_INFO
 */
ix_error ix_cc_rtmv4_set_flags(
    ix_cc_rtmv4 arg_hRtm,
    ix_cc_rtmv4_nhid arg_NextHopID,
    ix_uint32 flags)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_hRtm);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hRtm, struct ix_s_cc_rtmv4);
    
    IX_ERROR_CR(_ix_cc_nhdb_set_flags(arg_hRtm->hNhdb, arg_NextHopID, flags));

    return IX_SUCCESS;
}

/*
 * NAME: ix_cc_rtmv4_add_route
 *
 * DESCRIPTION: Adds routes, relating a range of network addresses to a Next
 * Hop.  In order to add a route, the client must first add the Next Hop ID
 * and its corresponding data into the RTM, using ix_cc_rtmv4_add_next_hop().
 *
 * After adding a route, lookup operations, either from Core Components through
 * the RTM or from microblocks, will use the newly added route.
 *
 * Adding a duplicate route to the same or different next hop is considered an
 * error condition.  Instead, use ix_cc_rtmv4_update_route() to modify an
 * existing route.
 *
 * The client may add a default route by setting the netAddr and netMask to
 * zero.  As always, the nextHopID must refer to a valid next hop.  Setting the
 * default route is a special case and the caller needn't remove an existing
 * default route before setting the default route to a new next hop.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 *
 * @Param:      - IN    arg_NetAddr
 *                      network address identifying the route
 *
 * @Param:      - IN    arg_NetMask
 *                      network mask for the route
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be used for the route
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_DUPLICATE_ENTRY
 * IX_CC_ERROR_CONFLICTING_ENTRY
 * IX_CC_RTMV4_ERROR_INVALID_MASK
 * IX_CC_ERROR_FULL
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_add_route(
    ix_cc_rtmv4 arg_hRtm,
    ix_uint32 arg_NetAddr,
    ix_uint32 arg_NetMask,
    ix_cc_rtmv4_nhid arg_NextHopID)
{
    ix_uint32 nextHopIndex;
    ix_error err = IX_SUCCESS;
    ix_uint16 data;

    /*
     * The lookup API takes pointers to ix_uint8s.  Set up some helper
     * variables.
     */
    ix_uint32 netAddr  = IX_HTON32(arg_NetAddr);
    ix_uint32 netMask  = IX_HTON32(arg_NetMask);
    ix_uint8 *pNetAddr = (ix_uint8*) &netAddr;
    ix_uint8 *pNetMask = (ix_uint8*) &netMask;
    ix_uint8 *pData    = (ix_uint8*) &data;
    int       netMaskLen;
    int       weight;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hRtm);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hRtm, struct ix_s_cc_rtmv4);

    /*
     * The default route is identified as 0.0.0.0/0.  Forward requests for
     * the default route to the Next Hop Database.
     */
    if ((arg_NetAddr == 0) && (arg_NetMask == 0))
    {
        if (arg_NextHopID == IX_CC_RTMV4_NHID_NO_ROUTE)
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_CONFLICTING_ENTRY,
                ("Use ix_cc_rtmv4_delete_route to remove the default route"));
        }
        IX_ERROR_CRT(_ix_cc_nhdb_set_default(arg_hRtm->hNhdb, arg_NextHopID),
                     IX_CC_ERROR_INTERNAL, 0);
        return IX_SUCCESS;
    }

    IX_ERROR_CRT(_get_netmask_len(arg_NetMask, &netMaskLen),
        IX_CC_RTMV4_ERROR_INVALID_MASK, 0);
    weight = 32 - netMaskLen;

    /*
     * Request the index for the next hop id.  If the call fails, return
     * either ENTRY_NOT_FOUND or INTERNAL.
     */
    IX_ERROR_C(_ix_cc_nhdb_nhid_to_index(arg_hRtm->hNhdb, arg_NextHopID,
                                         &nextHopIndex), err);
    if (err != IX_SUCCESS)
    {
        if (IX_ERROR_GET_CODE(err) != IX_CC_ERROR_ENTRY_NOT_FOUND)
        {
            return IX_ERROR_SET_CODE(err, IX_CC_ERROR_INTERNAL);
        }
        return err;
    }

    /*
     * Add the entry into the LPM tables.  The NHDB always reports the index
     * in host order, and the Lookup API always takes the data in network
     * order.
     */
    data = IX_HTON16(nextHopIndex);
    err =
        IX_LKUP_ADD_ENTRY(arg_hRtm->hTable, pNetAddr, pNetMask, weight, pData);
    if (err != IX_SUCCESS)
    {
        ix_error_number errCode;
        switch ((int) IX_ERROR_GET_CODE(err))
        {
            case IX_LKUP_ERROR_EXISTS:
                errCode = IX_CC_ERROR_DUPLICATE_ENTRY;
                break;

            case IX_LKUP_ERROR_CONFLICT:
                errCode = IX_CC_ERROR_CONFLICTING_ENTRY;
                break;

            default:
                errCode = IX_CC_ERROR_INTERNAL;
                break;
        }
        return IX_ERROR_CHAIN(err, errCode, 0,
            ("Could not add prefix %d.%d.%d.%d/%d",
             pNetAddr[0], pNetAddr[1], pNetAddr[2], pNetAddr[3], netMaskLen));
    }

    /*
     * Tell the NHDB that its next hop is in use.
     */
    IX_ERROR_CG(_ix_cc_nhdb_inc_refcount(arg_hRtm->hNhdb, nextHopIndex),
                err, cleanup_Lkup);

    arg_hRtm->numRoutes++;
    return IX_SUCCESS;

cleanup_Lkup:
    (void) IX_LKUP_REMOVE_ENTRY(arg_hRtm->hTable, pNetAddr, pNetMask, weight);
    return err;
}

/*
 * NAME: ix_cc_rtmv4_delete_route
 *
 * DESCRIPTION: Deletes a route.  The caller is responsible for removing the
 * next hop; deleting the next hop is not an implicit operation of
 * ix_cc_rtmv4_delete_route().
 *
 * The client may remove the default route by setting the netAddr and netMask
 * to zero.  Removing the default route is a special case and the caller may do
 * so at any time, even if the default route did not previously exist.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 *
 * @Param:      - IN    arg_NetAddr
 *                      Network address identifying the route
 *
 * @Param:      - IN    arg_NetMask
 *                      Subnet mask for the route
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_INTERNAL
 * IX_CC_RTMV4_ERROR_INVALID_MASK
 */
ix_error ix_cc_rtmv4_delete_route(
    ix_cc_rtmv4 arg_hRtm,
    ix_uint32 arg_NetAddr,
    ix_uint32 arg_NetMask)
{
    ix_uint32 nextHopIndex;
    ix_error err;
    ix_uint16 data;

    /*
     * The lookup API takes pointers to ix_uint8s.  Set up some helper
     * variables.
     */
    ix_uint32 netAddr  = IX_HTON32(arg_NetAddr);
    ix_uint32 netMask  = IX_HTON32(arg_NetMask);
    ix_uint8 *pNetAddr = (ix_uint8*) &netAddr;
    ix_uint8 *pNetMask = (ix_uint8*) &netMask;
    ix_uint8 *pData    = (ix_uint8*) &data;
    int       netMaskLen;
    int       weight;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hRtm);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hRtm, struct ix_s_cc_rtmv4);

    if ((arg_NetAddr == 0) && (arg_NetMask == 0))
    {
        /*
         * The caller wants to clear the default route.  Forward it to
         * the Next Hop Database.
         */
        IX_ERROR_CRT(_ix_cc_nhdb_set_default(arg_hRtm->hNhdb,
                                             IX_CC_RTMV4_NHID_NO_ROUTE),
                     IX_CC_ERROR_INTERNAL, 0);
        return IX_SUCCESS;
    }

    IX_ERROR_CRT(_get_netmask_len(arg_NetMask, &netMaskLen),
        IX_CC_RTMV4_ERROR_INVALID_MASK, 0);
    weight = 32 - netMaskLen;

    /*
     * Get the route's data so the next hop's refcount can be decremented.
     */
    IX_ERROR_C(
        IX_LKUP_FIND_ENTRY(arg_hRtm->hTable, pNetAddr, pNetMask, weight,
                           pData), err);
    if (err != IX_SUCCESS)
    {
        if (IX_ERROR_GET_CODE(err) == IX_LKUP_ERROR_NOTFOUND)
        {
            return IX_ERROR_SET_CODE(err, IX_CC_ERROR_ENTRY_NOT_FOUND);
        }
        return IX_ERROR_SET_CODE(err, IX_CC_ERROR_INTERNAL);
    }
    nextHopIndex = IX_NTOH16(data);

    if (arg_hRtm->numRoutes == 0)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
            ("The RTM has no routes"));
    }

    IX_ERROR_CRT(_ix_cc_nhdb_dec_refcount(arg_hRtm->hNhdb, nextHopIndex),
                 IX_CC_ERROR_INTERNAL, 0);

    /*
     * Finally, remove the route.  This is done last because it is the
     * most time-intensive and we want to avoid undoing it if possible.
     */
    IX_ERROR_CGT(IX_LKUP_REMOVE_ENTRY(arg_hRtm->hTable, pNetAddr, pNetMask,
                                      weight),
                 err, IX_CC_ERROR_INTERNAL, 0, cleanup_refCount);

    arg_hRtm->numRoutes--;
    return IX_SUCCESS;

cleanup_refCount:
    (void) _ix_cc_nhdb_inc_refcount(arg_hRtm->hNhdb, nextHopIndex);
    return err;
}

/*
 * NAME: ix_cc_rtmv4_update_route
 *
 * DESCRIPTION: Updates routes, relating an existing range of network
 * addresses to a different Next Hop.  In order to update a route, the client
 * must first add the Next Hop ID and its corresponding data into the RTM,
 * using ix_cc_rtmv4_add_next_hop().
 *
 * After adding a route, lookup operations, either from Core Components through
 * the RTM or from microblocks, will use the newly added route.
 *
 * The client should not use the update operation to modify teh default route.
 * Use ix_cc_rtmv4_add_route() or ix_cc_rtmv4_delete_route() instead.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 *
 * @Param:      - IN    arg_NetAddr
 *                      network address identifying the route
 *
 * @Param:      - IN    arg_NetMask
 *                      network mask for the route
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be used for the route
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_CONFLICTING_ENTRY
 * IX_CC_RTMV4_ERROR_INVALID_MASK
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_update_route(
    ix_cc_rtmv4 arg_hRtm,
    ix_uint32 arg_NetAddr,
    ix_uint32 arg_NetMask,
    ix_cc_rtmv4_nhid arg_NextHopID)
{
    ix_uint32 nextHopIndex;
    ix_uint16 oldNextHopIndex;
    ix_error err;

    ix_uint32 ipAddr   = IX_HTON32(arg_NetAddr);
    ix_uint32 netMask  = IX_HTON32(arg_NetMask);
    ix_uint16 data;
    ix_uint8 *pKey     = (ix_uint8*) &ipAddr;
    ix_uint8 *pMask    = (ix_uint8*) &netMask;
    ix_uint8 *pData    = (ix_uint8*) &data;
    int       netMaskLen;
    int       weight;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hRtm);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hRtm, struct ix_s_cc_rtmv4);

    /*
     * Prevent updating the default route.
     */
    if ((arg_NetAddr == 0) && (arg_NetMask == 0))
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_CONFLICTING_ENTRY,
            ("Cannot update the default route"));
    }

    IX_ERROR_CRT(_get_netmask_len(arg_NetMask, &netMaskLen),
        IX_CC_RTMV4_ERROR_INVALID_MASK, 0);
    weight = 32 - netMaskLen;

    /*
     * Get the index for the new next hop.  This also verifies the next
     * hop exists.
     */
    IX_ERROR_C(_ix_cc_nhdb_nhid_to_index(arg_hRtm->hNhdb, arg_NextHopID,
               &nextHopIndex), err);
    if (err != IX_SUCCESS)
    {
        /*
         * Indicate either that the entry wasn't found or an internal
         * problem occurred.
         */
        if (IX_ERROR_GET_CODE(err) == IX_CC_ERROR_ENTRY_NOT_FOUND)
        {
            return err;
        }
        return IX_ERROR_SET_CODE(err, IX_CC_ERROR_INTERNAL);
    }

    /*
     * Updating the route means a previously-used next hop will have
     * one less reference.  Get its index so its refcount can be decreased.
     */
    IX_ERROR_C(IX_LKUP_FIND_ENTRY(arg_hRtm->hTable, pKey, pMask, weight,
               pData), err);
    if (err != IX_SUCCESS)
    {
        if (IX_ERROR_GET_CODE(err) == IX_LKUP_ERROR_NOTFOUND)
        {
            return IX_ERROR_SET_CODE(err, IX_CC_ERROR_ENTRY_NOT_FOUND);
        }
        else
        {
            return IX_ERROR_SET_CODE(err, IX_CC_ERROR_INTERNAL);
        }
    }
    oldNextHopIndex = IX_NTOH16(data);
    IX_ERROR_CRT(_ix_cc_nhdb_dec_refcount(arg_hRtm->hNhdb, oldNextHopIndex),
            IX_CC_ERROR_INTERNAL, 0);

    data = IX_HTON16(nextHopIndex);
    /*
     * Normally we might check for NOTFOUND, but FIND_ENTRY() would have
     * already reported that error if that was the case.  Simply convert all
     * errors to INTERNAL.
     */
    IX_ERROR_CGT(IX_LKUP_UPDATE_ENTRY(arg_hRtm->hTable, pKey, pMask, weight,
               pData),
        err, IX_CC_ERROR_ENTRY_NOT_FOUND, 0, cleanup_oldNextHop);

    IX_ERROR_CGT(_ix_cc_nhdb_inc_refcount(arg_hRtm->hNhdb, nextHopIndex),
            err, IX_CC_ERROR_INTERNAL, 0, cleanup_Route);

    return IX_SUCCESS;

cleanup_Route:
    (void) IX_LKUP_UPDATE_ENTRY(arg_hRtm->hTable, pKey, pMask, weight, pData);
cleanup_oldNextHop:
    (void) _ix_cc_nhdb_inc_refcount(arg_hRtm->hNhdb, oldNextHopIndex);
    return err;
}

/*
 * NAME: ix_cc_rtmv4_get_route
 *
 * DESCRIPTION: Gets the Next Hop ID for a specific route.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 * @Param:      - IN    arg_NetAddr
 *                      network address identifying the route
 * @Param:      - IN    arg_NetMask
 *                      network mask for the route
 * @Param:      - OUT   arg_pNextHopID
 *                      identifier of the Next Hop used for the route
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_RTMV4_ERROR_INVALID_MASK
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_get_route(
    ix_cc_rtmv4 arg_hRtm,
    ix_uint32 arg_NetAddr,
    ix_uint32 arg_NetMask,
    ix_cc_rtmv4_nhid *arg_pNextHopID)
{
    ix_uint32 nextHopIndex;
    ix_error err;

    ix_uint32 ipAddr   = IX_HTON32(arg_NetAddr);
    ix_uint32 netMask  = IX_HTON32(arg_NetMask);
    ix_uint16 data;
    ix_uint8 *pKey     = (ix_uint8*) &ipAddr;
    ix_uint8 *pMask    = (ix_uint8*) &netMask;
    ix_uint8 *pData    = (ix_uint8*) &data;
    int       netMaskLen;
    int       weight;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hRtm);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hRtm, struct ix_s_cc_rtmv4);
    IX_ERROR_DCHECK_ARG_NULL(4, arg_pNextHopID);
    IX_ERROR_DCHECK_ARG_ALIGN(4, arg_pNextHopID, ix_cc_rtmv4_nhid);

    IX_ERROR_CRT(_get_netmask_len(arg_NetMask, &netMaskLen),
        IX_CC_RTMV4_ERROR_INVALID_MASK, 0);
    weight = 32 - netMaskLen;

    IX_ERROR_C(IX_LKUP_FIND_ENTRY(arg_hRtm->hTable, pKey, pMask, weight,
               pData), err);
    if (err != IX_SUCCESS)
    {
        if (IX_ERROR_GET_CODE(err) == IX_LKUP_ERROR_NOTFOUND)
        {
            return IX_ERROR_SET_CODE(err, IX_CC_ERROR_ENTRY_NOT_FOUND);
        }
        else
        {
            return IX_ERROR_SET_CODE(err, IX_CC_ERROR_INTERNAL);
        }
    }
    nextHopIndex = IX_NTOH16(data);

    IX_ERROR_CRT(_ix_cc_nhdb_index_to_nhid(
                arg_hRtm->hNhdb, nextHopIndex, arg_pNextHopID),
            IX_CC_ERROR_INTERNAL, 0);

    return IX_SUCCESS;
}

/*
 * NAME: ix_cc_rtmv4_lookup
 *
 * DESCRIPTION: Looks up routing information for a given IP address.  The RTM
 * uses an LPM algorithm to match the IP address with a network (as given by
 * calling ix_cc_rtmv4_add_route()) and thereby provide the information needed
 * to forward the packet to the appropriate next hop - i.e. the corresponding
 * Next Hop Information structure.
 *
 * This function is the workhorse of the RTM, and is expected to be the
 * most-used entry point; conceivably every packet arriving on the XScale core
 * will require a lookup.  Therefore, this function is designed with efficiency
 * in mind.
 *
 * If the nextHopID field in arg_pNextHopInfo is IX_CC_RTMV4_NHID_NO_ROUTE, all
 * other fields are invalid; the IP address has no route.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 *
 * @Param:      - IN    arg_IpAddr
 *                      IP address to be looked up
 *
 * @Param:      - OUT   arg_pNextHopInfo
 *                      results are placed into the structure provided by the
 *                      caller.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_lookup(
    ix_cc_rtmv4 arg_hRtm,
    ix_uint32 arg_IpAddr,
    ix_cc_rtmv4_next_hop_info *arg_pNextHopInfo)
{
    ix_uint16 nextHopIndex;
    ix_uint32 ipAddr = IX_HTON32(arg_IpAddr);
    ix_uint16 data;
    ix_uint8 *pKey  = (ix_uint8*) &ipAddr;
    ix_uint8 *pData = (ix_uint8*) &data;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hRtm);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hRtm, struct ix_s_cc_rtmv4);
    IX_ERROR_DCHECK_ARG_NULL(2, arg_pNextHopInfo);
    IX_ERROR_DCHECK_ARG_ALIGN(2, arg_pNextHopInfo, ix_cc_rtmv4_next_hop_info);

    /*
     * Searching the LPM table should never fail, even if there is no
     * route for the ip address.  If it does fail, return INTERNAL.
     */
    IX_ERROR_CRT(
        IX_LKUP_SEARCH_TABLE(arg_hRtm->hTable, pKey, pData),
        IX_CC_ERROR_INTERNAL, 0);
    nextHopIndex = IX_NTOH16(data);

    /*
     * If the index was in the LPM, it must be in the NHDB.  Therefore, the
     * call shouldn't fail.  If it does, return INTERNAL.
     */
    IX_ERROR_CRT(
        _ix_cc_nhdb_get_nh_by_index(arg_hRtm->hNhdb, nextHopIndex,
            arg_pNextHopInfo),
        IX_CC_ERROR_INTERNAL, 0);

    return IX_SUCCESS;
}

/*
 * NAME: ix_cc_rtmv4_dump_next_hops
 *
 * DESCRIPTION: Prints all Next Hops to a block of memory for debugging
 * purposes.  The memory is provided by the caller, along with the size of the
 * memory block in bytes.  The caller should use IX_CC_RTMV4_DUMP_NEXT_HOP_SIZE
 * to determine how many bytes to allocate.
 *
 * If the function returns IX_CC_ERROR_FULL, the error should be considered
 * non-fatal.  The caller may allocate more memory and call the function again.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 *
 * @Param:      - OUT   arg_pBuffer
 *                      pointer to a block of memory where the next hops are
 *                      to be stored; human-readable ASCII-Z string.
 *
 * @Param:      - IN    arg_BufferSize
 *                      size in bytes of the buffer; host byte order.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_RANGE
 * IX_CC_ERROR_FULL
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_dump_next_hops(
    ix_cc_rtmv4 arg_hRtm,
    char *arg_pBuffer,
    ix_uint32 arg_BufferSize)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_hRtm);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hRtm, struct ix_s_cc_rtmv4);
    IX_ERROR_CR(_ix_cc_nhdb_dump(arg_hRtm->hNhdb, arg_pBuffer, arg_BufferSize));
    return IX_SUCCESS;
}

/*
 * NAME: _append_route
 *
 * DESCRIPTION: Helper function for dumping the routes.  Adds a single route
 * to the end of the buffer.
 */
static int _append_route(
        char *arg_pBuffer,
        unsigned int *pBufSizeLeft,
        ix_uint32 ipAddr,
        int netMaskLen,
        ix_cc_rtmv4_nhid nextHopID)
{
    char line[_IX_CC_RTMAPI_LINE_SIZE+1]; /* plus 1 for EOL */
    char ipAddrString[19];          /* nnn.nnn.nnn.nnn/nn\0 */

    /*
     * If we can't store more lines, drop on the terminator and quit.
     */
    if (*pBufSizeLeft < _IX_CC_RTMAPI_LINE_SIZE+1)
    {
        /*
         * If there's not enough room for the terminator, back the string
         * up by the line size before appending the terminator.
         */
        if (*pBufSizeLeft < _IX_CC_RTMAPI_TERM_SIZE+1)
        {
            arg_pBuffer[strlen(arg_pBuffer)-_IX_CC_RTMAPI_LINE_SIZE] = '\0';
        }
        strcat(arg_pBuffer, _IX_CC_RTMAPI_TERM_STRING);
        return 1;
    }

    sprintf(ipAddrString, "%ld.%ld.%ld.%ld/%d",
        (ipAddr >> 24) & 0xFF,
        (ipAddr >> 16) & 0xFF,
        (ipAddr >>  8) & 0xFF,
        (ipAddr      ) & 0xFF,
        netMaskLen);

    sprintf(line, _IX_CC_RTMAPI_LINE_FMT,
        ipAddrString, nextHopID);
    strcat(arg_pBuffer, line);

    *pBufSizeLeft -= _IX_CC_RTMAPI_LINE_SIZE;

    return 0;
}

/*
 * NAME: ix_cc_rtmv4_dump_routes
 *
 * DESCRIPTION: Prints all routes to a block of memory for debugging purposes.
 * The memory is provided by the caller, along with the size of the memory
 * block in bytes.  The caller should use IX_CC_RTMV4_DUMP_ROUTE_SIZE
 * to determine how many bytes to allocate.
 *
 * If the function returns IX_CC_ERROR_FULL, the error should be considered
 * non-fatal.  The caller may allocate more memory and call the function again.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 *
 * @Param:      - OUT   arg_pBuffer
 *                      pointer to a block of memory where the routes are
 *                      to be stored; human-readable ASCII-Z string.
 *
 * @Param:      - IN    arg_BufferSize
 *                      size in bytes of the buffer; host byte order.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_RANGE
 * IX_CC_ERROR_FULL
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_dump_routes(
    ix_cc_rtmv4 arg_hRtm,
    char *arg_pBuffer,
    ix_uint32 arg_BufferSize)
{
    unsigned int bufSizeLeft = arg_BufferSize;
    ix_lkup_cookie cookie;
    ix_uint32 key;
    ix_uint32 mask;
    int weight;
    ix_uint16 data;
    ix_uint8 *pKey  = (ix_uint8*) &key;
    ix_uint8 *pMask = (ix_uint8*) &mask;
    ix_uint8 *pData = (ix_uint8*) &data;
    ix_error err;
    ix_cc_rtmv4_nhid nextHopID;
    ix_uint32 netMask;
    int netMaskLen;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hRtm);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hRtm, struct ix_s_cc_rtmv4);
    IX_ERROR_DCHECK_ARG_NULL(2, arg_pBuffer);
    IX_ERROR_DCHECK_ARG_MIN(3, arg_BufferSize, 1UL, "%ld");

    if (bufSizeLeft < _IX_CC_RTMAPI_TITLE_SIZE + _IX_CC_RTMAPI_TERM_SIZE + 1)
    {
        arg_pBuffer[0] = '\0';
        return IX_ERROR_WARNING(IX_CC_ERROR_FULL,
            ("Buffer is too small (%ld bytes), should be at least %ld bytes\n"
             "\tUse IX_CC_RTMV4_DUMP_NEXT_HOP_SIZE()",
             arg_BufferSize,
             _IX_CC_RTMAPI_TITLE_SIZE + _IX_CC_RTMAPI_TERM_SIZE + 1
             ));
    }

    strcpy(arg_pBuffer, _IX_CC_RTMAPI_TITLE_STRING);
    bufSizeLeft -= _IX_CC_RTMAPI_TITLE_SIZE;

    /*
     * First entry is the default route.  Getting the index of the 0th entry
     * gives us the default.
     */
    IX_ERROR_CRT(_ix_cc_nhdb_index_to_nhid(arg_hRtm->hNhdb, 0, &nextHopID),
        IX_CC_ERROR_INTERNAL, 0);
    if(_append_route(arg_pBuffer, &bufSizeLeft, 0, 0,
            nextHopID))
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_FULL,
            ("Buffer is too small (%ld bytes)\n"
             "\tUse IX_CC_RTMV4_DUMP_ROUTE_SIZE()",
             arg_BufferSize));
    }

    /*
     * Iterate through the Lookup Library getting the entries one at a time
     * and adding them to the buffer.
     */
    err = IX_LKUP_READ_FIRST_ENTRY(arg_hRtm->hTable, pKey, pMask, &weight,
            pData, &cookie);

    while (err == IX_SUCCESS)
    {
        ix_uint32 ipAddr = IX_NTOH32(key);
        ix_uint16 nextHopIndex = IX_NTOH16(data);

        netMask = IX_HTON32(mask);
        IX_ERROR_CR(_get_netmask_len(netMask, &netMaskLen));

        IX_ERROR_CRT(_ix_cc_nhdb_index_to_nhid(arg_hRtm->hNhdb, nextHopIndex,
            &nextHopID), IX_CC_ERROR_INTERNAL, 0);

        if (_append_route(arg_pBuffer, &bufSizeLeft,
                ipAddr, netMaskLen, nextHopID))
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_FULL,
                ("Buffer is too small (%ld bytes)\n"
                 "\tUse IX_CC_RTMV4_DUMP_ROUTE_SIZE()",
                 arg_BufferSize));
        }

        err = IX_LKUP_READ_NEXT_ENTRY(arg_hRtm->hTable, cookie,
            pKey, pMask, &weight, pData, &cookie);
    }

    /*
     * There is always an error at this point.  Under normal conditions, it
     * means we've gotten all the routes.
     */
    if (IX_ERROR_GET_CODE(err) != IX_LKUP_ERROR_NOTFOUND)
    {
        return IX_ERROR_CHAIN(err, IX_CC_ERROR_INTERNAL, 0,
            ("Error while enumerating the routes"));
    }

    return IX_SUCCESS;
}

/*
 * NAME: ix_cc_rtmv4_purge
 *
 * DESCRIPTION: Removes all routes and next hops.  Most likely, this would
 * only be done for debugging purposes.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_purge(
    ix_cc_rtmv4 arg_hRtm)
{
    IX_ERROR_CR(ix_cc_rtmv4_purge_routes(arg_hRtm));

    /*
     * No condition should cause the NHDB to fail purging.
     */
    IX_ERROR_CRT(_ix_cc_nhdb_purge(arg_hRtm->hNhdb),
        IX_CC_ERROR_INTERNAL, 0);

    arg_hRtm->numNextHops = 0;
    return IX_SUCCESS;
}

/*
 * NAME: ix_cc_rtmv4_purge_routes
 *
 * DESCRIPTION: Removes all routes.  Most likely, this would only be done
 * for debugging purposes.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_purge_routes(
    ix_cc_rtmv4 arg_hRtm)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_hRtm);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hRtm, struct ix_s_cc_rtmv4);

    IX_ERROR_CRT(IX_LKUP_RESET_TABLE(arg_hRtm->hTable),
            IX_CC_ERROR_INTERNAL, 0);
    IX_ERROR_CRT(_ix_cc_nhdb_reset_refcounts(arg_hRtm->hNhdb),
            IX_CC_ERROR_INTERNAL, 0);
    IX_ERROR_CRT(_ix_cc_nhdb_set_default(arg_hRtm->hNhdb,
                IX_CC_RTMV4_NHID_NO_ROUTE),
            IX_CC_ERROR_INTERNAL, 0);

    arg_hRtm->numRoutes = 0;
    return IX_SUCCESS;
}

/*
 * NAME: ix_cc_rtmv4_get_symbols
 *
 * DESCRIPTION: Retrieves symbols to patch to the micoengines.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 *
 * @Param:      - OUT   arg_pSymbols
 *                      structure containing values that the client may patch
 *                      to the microengines.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERAL
 */
ix_error ix_cc_rtmv4_get_symbols(
    ix_cc_rtmv4 arg_hRtm,
    ix_cc_rtmv4_symbols *arg_pSymbols)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_hRtm);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hRtm, struct ix_s_cc_rtmv4);
    IX_ERROR_DCHECK_ARG_NULL(2, arg_pSymbols);
    IX_ERROR_DCHECK_ARG_ALIGN(2, arg_pSymbols, ix_cc_rtmv4_symbols);

    IX_ERROR_CRT(IX_LKUP_GET_TABLE_INFO(arg_hRtm->hTable,
                &arg_pSymbols->lkupTableID,
                &arg_pSymbols->lkupTableData1,
                &arg_pSymbols->lkupTableData2),
            IX_CC_ERROR_INTERNAL, 0);

    IX_ERROR_CRT(_ix_cc_nhdb_get_base_address(arg_hRtm->hNhdb,
                &arg_pSymbols->nextHopBase),
            IX_CC_ERROR_INTERNAL, 0);

    return IX_SUCCESS;
}

/*
 * NAME: ix_cc_rtmv4_get_statistics
 *
 * DESCRIPTION: Retrieves statistics of the RTM.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv4_init().
 *
 * @Param:      - OUT   arg_pStatistics
 *                      structure containing the statistics.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv4_get_statistics(
    ix_cc_rtmv4 arg_hRtm,
    ix_cc_rtmv4_statistics *arg_pStatistics)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_hRtm);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hRtm, struct ix_s_cc_rtmv4);
    IX_ERROR_DCHECK_ARG_NULL(2, arg_pStatistics);
    IX_ERROR_DCHECK_ARG_ALIGN(2, arg_pStatistics, ix_cc_rtmv4_statistics);

    /*
     * Report one extra route and next hop because of the default
     */
    arg_pStatistics->numberOfRoutes = arg_hRtm->numRoutes + 1;
    arg_pStatistics->numberOfNextHops = arg_hRtm->numNextHops + 1;

    return IX_SUCCESS;
}

/*
 * NAME _get_netmask_len
 *
 * DESCRIPTION: Given a network mask (e.g. 0xFFFF0000), return its length
 * (e.g. 16).
 */
ix_error _get_netmask_len(ix_uint32 netMask, int *pNetMaskLen)
{
    int bit;
    ix_uint32 mask = netMask;

    for (bit = 0; bit < 32; bit++)
    {
        if (mask & 1)
        {
            break;
        }
        mask >>= 1;
    }

    if (bit == 32)
    {
        bit = 0;
    }

    *pNetMaskLen = 32 - bit;
    if (netMask == (0xFFFFFFFFUL << bit))
    {
        return IX_SUCCESS;
    }

    return IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
        ("The netMask 0x%08lX is invalid", netMask));
}
