/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      ix_cc_rtmv6.h
 *
 * = DESCRIPTION
 *      Public header file for users of the Route Table Manager for the IPv6
 *      Forwarder Core Component (RTMv6).
 *
 * ============================================================================
 * $Id: ix_cc_rtmv6.h,v 1.14 2003/11/20 22:53:27 ttschlue Exp $
 */

#if !defined(__IX_CC_RTMV6_H__)
#define __IX_CC_RTMV6_H__

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

#include "cc/ix_cc_rtmv4.h"
#include "ix_cc_error.h"
#include "ix_types.h"

/*
 * Opaque handle to the RTM, obtained by calling ix_cc_rtmv6_init().
 */
typedef struct ix_s_cc_rtmv6* ix_cc_rtmv6;

/*
 * Typedef to an intrinsic data type; represents a Next Hop ID.  Applications
 * at the highest levels allocate and assign the NHID values.
 */
typedef ix_uint32 ix_cc_rtmv6_nhid;

/*
 * The RTM defines a structure describing a Next Hop.  This structure is used
 * when adding a Next Hop (see ix_cc_rtmv6_add_next_hop(), retrieving Next Hop
 * Information (see ix_cc_rtmv6_get_next_hop()), or looking up a Next Hop by IP
 * address (see ix_cc_rtmv6_lookup()).
 *
 * Though this data is identical in meaning to the Next Hop Information used by
 * the IPv6 microblock [BLDBLK], the structure used to store the data is less
 * restrictive.  The RTM packs the Next Hop data into the format the
 * microengines expect.
 *
 * It is anticipated that the IPv6 Forwarder will expose this structure in its
 * interface, requiring Forwarding Plane Module to understand the structure.
 * It may be appropriate to move the definition to a more central location
 * later.  Doing so will affect a number of RTM interfaces.
 */
typedef ix_cc_rtmv4_next_hop_info ix_cc_rtmv6_next_hop_info;

/*
 * The RTM defines a structure containing values useful in the microengines.
 * This structure is obtained by calling ix_cc_rtmv6_get_symbols().
 */
typedef ix_cc_rtmv4_symbols ix_cc_rtmv6_symbols;                                                                
/*
 * The client may retrieve statistics about the RTM by calling
 * ix_cc_rtmv6_get_statistics().
 */
typedef struct ix_s_cc_rtmv6_statistics
{
    ix_uint32       numberOfRoutes;
    ix_uint32       numberOfNextHops;
    ix_uint32       numberOfMultipleNextHops;
} ix_cc_rtmv6_statistics;                                                             

/*
 * Enumeration of the supported algorithm types.
 */
typedef enum ix_e_cc_rtmv6_lkup_type
{
    IX_CC_RTMV6_LKUP_SOFTWARE,
    IX_CC_RTMV6_LKUP_TCAM,
    IX_CC_RTMV6_LKUP_LAST
} ix_cc_rtmv6_lkup_type;                                                              

/*
 * Enumeration of the supported memory types.
 */
typedef enum ix_e_cc_rtmv6_mem_type
{
    IX_CC_RTMV6_MEM_SDRAM,
    IX_CC_RTMV6_MEM_SRAM,
    IX_CC_RTMV6_MEM_TCAM,
    IX_CC_RTMV6_MEM_LAST
} ix_cc_rtmv6_mem_type;                                                               

/* IX_CC_RTMV6_DUMP_ROUTE_SIZE
 *
 * Description:
 * Macro for calculating the minimum size of a memory block useful for dumping
 * all the routes in the RTM.  The macro adds one because
 * ix_cc_rtmv6_dump_routes() always prints the default route as well.
 *
 * Syntax:
 * IX_CC_RTMV6_DUMP_ROUTE_SIZE(numberOfRoutes)
 *
 * Input Arguments:
 * numberOfRoutes  The number of routes in the RTM, obtained by calling
 *                 ix_cc_rtmv6_get_statistics().
 *
 * Returns (R-Value):
 * The number of bytes of memory required for ix_cc_rtmv6_dump_routes().
 */
#define IX_CC_RTMV6_DUMP_ROUTE_SIZE(__numberOfRoutes)                   \
    (_IX_CC_RTMAPI6_TITLE_SIZE +                                         \
        (__numberOfRoutes) * _IX_CC_RTMAPI6_LINE_SIZE + 1)

/*
 * IX_CC_RTMV6_DUMP_NEXT_HOP_SIZE
 *
 * Description:
 * Macro for calculating the minimum size of a memory block useful for dumping
 * all the next hops in the RTM.
 *
 * Syntax:
 * IX_CC_RTMV6_DUMP_NEXT_HOP_SIZE(numberOfNextHops)
 *
 * Input Arguments:
 * numberOfNextHops  The number of next hops in the RTM, obtained by
 *                   calling ix_cc_rtmv6_get_statistics().
 *
 * Returns (R-Value):
 * The number of bytes of memory required for ix_cc_rtmv6_dump_next_hops().
 */
#define IX_CC_RTMV6_DUMP_NEXT_HOP_SIZE(__numberOfNextHops)              \
    (_IX_CC_NHDB_TITLE_SIZE + (__numberOfNextHops) * _IX_CC_NHDB_LINE_SIZE + 1)

/*
 * IX_CC_RTMV6_NHID_NO_ROUTE
 *
 * Description:
 * The RTM provides a macro for a reserved Next Hop ID value.
 */
#define IX_CC_RTMV6_NHID_NO_ROUTE IX_CC_RTMV4_NHID_NO_ROUTE

/*
 * IX_CC_RTMV6_L2INDEX_NO_ROUTE
 *
 * Description:
 * The RTM provides a macro for a reserved L2Index value.
 */
#define IX_CC_RTMV6_L2INDEX_NO_ROUTE IX_CC_RTMV4_L2INDEX_NO_ROUTE

typedef enum ix_e_cc_rtmv6_error
{
    IX_CC_RTMV6_ERROR_INVALID_NH_INFO = IX_ERROR_MAKE_GROUP(CC_RTMV6),
    IX_CC_RTMV6_ERROR_NHID_IN_USE,
    IX_CC_RTMV6_ERROR_INIT_LKUP,
    IX_CC_RTMV6_ERROR_CREATE_TABLE,
    IX_CC_RTMV6_ERROR_CREATE_NHDB,
    IX_CC_RTMV6_ERROR_LKUP_UNAVAILABLE,
    IX_CC_RTMV6_ERROR_INVALID_PREFIX,

    IX_CC_RTMV6_ERROR_LAST
} ix_cc_rtmv6_error;

/* 
 * RTMV6_NHDB_ERROR
 *
 * Description:
 * The NHDB code (in nhdb.c) returns RTMv4 errors, however, RTMv6 expects to 
 * report the above error codes. This macro map the returned error code to the
 * appropriate RTMv6 code and returns. The error code not listed here probably
 * do not require translation.
 */
#define RTMV6_NHDB_ERROR(_error, _level, _s) \
{\
    switch ((_error)) \
    { \
        case IX_CC_RTMV4_ERROR_INVALID_NH_INFO: \
            return IX_ERROR_CHAIN(0, IX_CC_RTMV6_ERROR_INVALID_NH_INFO, _level, _s); \
        case IX_CC_RTMV4_ERROR_NHID_IN_USE: \
            return IX_ERROR_CHAIN(0, IX_CC_RTMV6_ERROR_NHID_IN_USE, _level, _s); \
        case IX_CC_RTMV4_ERROR_CREATE_NHDB: \
	    return IX_ERROR_CHAIN(0, IX_CC_RTMV6_ERROR_CREATE_NHDB, _level, _s); \
        default: \
	    return IX_ERROR_CHAIN(0, _error, _level, _s); \
    } \
}

/*
 * NAME: ix_cc_rtmv6_config
 *
 * DESCRIPTION: The ix_cc_rtmv6_config struct is used to indicate to
 * ix_cc_rtmv6_init() the desired configuration of the resulting route table
 * manager.
 *
 * routeTableSizeHint      Indicates approximately how many routes will be
 *                         supported.  The quantity indicated is not
 *                         guaranteed, but may be used to
 *                         determine the number of resources required.
 * lkupType        	   Type of lookup to use.  This must match the 
 *                         algorithm used by the microengines.
 * lkupMemType             Type of memory to use in the lookup algorithm.  
 *                         This must match the memory type used by the 
 *                         microengines.
 * lkupChannel             Channel number where the microblock will access 
 *                         the LPM tables.
 * nhdbSizeHint            Indicates approximately how many next hops will be
 *                         supported.  The quantity indicated is not 
 *                         guaranteed, but may be used to determine the 
 *                         number of resources required.
 * nhdbMemType             Type of memory to use in the Next Hop Database.  
 *                         This must match the memory type used by the 
 *                         microengines.
 * nhdbChannel             Channel number where the microblock will access 
 *                         the Next Hop table.
 */
typedef struct ix_s_cc_rtmv6_config
{
    ix_uint32 routeTableSizeHint;
    ix_cc_rtmv6_lkup_type lkupType;
    ix_cc_rtmv6_mem_type lkupMemType;
    ix_uint32 lkupChannel;
    ix_uint32 nhdbSizeHint;
    ix_cc_rtmv6_mem_type nhdbMemType;
    ix_uint32 nhdbChannel;
#ifdef SPLIT_V6_TRIE_TABLE
    ix_uint32 lkupChannel16BitTrie;
#endif /* SPLIT_V6_TRIE_TABLE */
} ix_cc_rtmv6_config;

/*
 * NAME: ix_cc_rtmv6_init
 *
 * DESCRIPTION: Creates a new RTM for the client.  The client must initialize
 * the RTM before performing any other actions.
 *
 * @Param:      - IN    arg_pConfig
 *                      Configuration for the RTM to be created, indicating
 *                      the amount and type of memory to use.
 * @Param:      - OUT   pRtm
 *                      Handle to the RTM.
 *
 * Error codes:
 * IX_CC_ERROR_RANGE
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_RTMV6_ERROR_LKUP_UNAVAILABLE
 * IX_CC_RTMV6_ERROR_CREATE_NHDB
 * IX_CC_RTMV6_ERROR_INIT_LKUP
 */
ix_error ix_cc_rtmv6_init(
    const ix_cc_rtmv6_config *arg_pConfig,
    ix_cc_rtmv6 *arg_phRtm);

/*
 * NAME: ix_cc_rtmv6_fini
 *
 * DESCRIPTION: Destroys an existing RTM, freeing all previously allocated
 * resources.  The client must not finalize the RTM until all microblocks
 * have been disabled.  After returning, the arg_hRtm handle is no longer
 * valid.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv6_init()
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv6_fini(
    ix_cc_rtmv6 arg_hRtm);

/*
 * NAME: ix_cc_rtmv6_add_next_hop
 *
 * DESCRIPTION: Adds Next Hops to the RTM's next hop database.  Typically,
 * clients would do this any time a new router is detected on a network.
 * Once a Next Hop is entered, many routes may be added referring to that
 * Next Hop.  Next Hops are uniquely identified by their NHID, as provided in
 * the Next Hop Information Structure.  Any attempt to add a duplicate NHID
 * will result in an error.
 *
 * A next hop value of IX_CC_RTMV6_NHID_NO_ROUTE is special and the client is
 * prevented from adding, deleting or changing it using
 * ix_cc_rtmv6_add_next_hop(), ix_cc_rtmv6_delete_next_hop(), or
 * ix_cc_rtmv6_update_next_hop().
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv6_init().
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be added.
 * @Param:      - IN    arg_NumNextHops,
 *                      The number of next hops contained in the array below.
 * @Param:      - IN    arg_pNextHopInfo
 *                      Array of next hop infos.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_DUPLICATE_ENTRY
 * IX_CC_ERROR_CONFLICTING_ENTRY
 * IX_CC_ERROR_FULL
 * IX_CC_RTMV6_ERROR_INVALID_NH_INFO
 */
ix_error ix_cc_rtmv6_add_next_hop(
    ix_cc_rtmv6 arg_hRtm,
    ix_cc_rtmv6_nhid arg_NextHopID,
    ix_uint32 arg_NumNextHops,
    const ix_cc_rtmv6_next_hop_info *arg_pNextHopInfo);

/*
 * NAME: ix_cc_rtmv6_delete_next_hop
 *
 * DESCRIPTION: Removes a Next Hop from the RTM, provided no prefix currently
 * refer to the Next Hop.  If a client wishes to remove a next hop, it must
 * first use ix_cc_rtmv6_delete_prefix() to remove all prefix referring to the
 * next hop, then use ix_cc_rtmv6_delete_next_hop() to remove the next hop.
 * This may require the client to maintain a complete list of prefix associated
 * with next hops.  A client would typically remove a Next Hop when a peer
 * prefix is removed from the network.
 *
 * This call will fail if any prefix, including the default prefix, uses the
 * indicated nextHopID.
 *
 * A next hop value of IX_CC_RTMV6_NHID_NO_ROUTE is special and the client is
 * prevented from adding, deleting or changing it using
 * ix_cc_rtmv6_add_next_hop(), ix_cc_rtmv6_delete_next_hop(), or
 * ix_cc_rtmv6_update_next_hop().
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv6_init().
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be removed; network byte
 *                      order.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_RTMV6_ERROR_NHID_IN_USE
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv6_delete_next_hop(
    ix_cc_rtmv6 arg_hRtm,
    ix_cc_rtmv6_nhid arg_NextHopID);

/*
 * NAME: ix_cc_rtmv6_update_next_hop
 *
 * DESCRIPTION: Updates the information contained in a next hop.  The update
 * operation requires that the next hop already exists; next hops cannot be
 * created using the update operation, use ix_cc_rtmv6_add_next_hop() to add
 * next hops.
 *
 * A next hop value of IX_CC_RTMV6_NHID_NO_ROUTE is special and the client is
 * prevented from adding, deleting or changing it using
 * ix_cc_rtmv6_add_next_hop(), ix_cc_rtmv6_delete_next_hop(), or
 * ix_cc_rtmv6_update_next_hop().
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv6_init().
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be removed
 * @Param:      - IN    arg_pNextHopInfo
 *                      pointer to the new Next Hop Information structure.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_RTMV6_ERROR_NHID_IN_USE
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv6_update_next_hop(
    ix_cc_rtmv6 arg_hRtm,
    ix_cc_rtmv6_nhid arg_NextHopID,
    const ix_cc_rtmv6_next_hop_info *arg_pNextHopInfo);

/*
 * NAME: ix_cc_rtmv6_get_next_hop
 *
 * DESCRIPTION: Retrieves a structure containing a copy of the information
 * originally given via ix_cc_rtmv6_add_next_hop().
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv6_init().
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be retrieved
 * @Param:      - OUT   arg_pNextHopInfo
 *                      results are placed into the structure provided by the
 *                      caller. This structure has to be atleast 4 nexthops long,
 *                      since the NHID might be pointing to upto 4 equal cost
 *                      next hops.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv6_get_next_hop(
    ix_cc_rtmv6 arg_hRtm,
    ix_cc_rtmv6_nhid arg_NextHopID,
    ix_cc_rtmv6_next_hop_info *arg_pNextHopInfo);

/*
 * NAME: ix_cc_rtmv6_set_mtu
 *
 * DESCRIPTION: Updates the MTU for a given Next Hop.  If the default route
 * matches arg_NextHopID, the default route's MTU is affected as well.
 * Because the RTM does not interpret the meaning of the MTU, it does not
 * validate the value.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv6_init().
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be modified
 * @Param:      - IN    arg_Mtu
 *                      new maximum transmission unit for the Next Hop
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_INVALID_NH_INFO
 */
ix_error ix_cc_rtmv6_set_mtu(
    ix_cc_rtmv6 arg_hRtm,
    ix_cc_rtmv6_nhid arg_NextHopID,
    ix_uint32 arg_Mtu);

/*
 * NAME: ix_cc_rtmv6_set_flags
 *
 * DESCRIPTION: Updates the Flags field for a given Next Hop.  If the default
 * route matches arg_NextHopID, the default route's flags are affected as well.
 * The meanings of the flags are determined by the caller.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv6_init().
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be modified
 * @Param:      - IN    arg_Flags
 *                      new flags for the Next Hop
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_INVALID_NH_INFO
 */
ix_error ix_cc_rtmv6_set_flags(
    ix_cc_rtmv6 arg_hRtm,
    ix_cc_rtmv6_nhid arg_NextHopID,
    ix_uint32 arg_Flags);

/*
 * NAME: ix_cc_rtmv6_add_prefix
 *
 * DESCRIPTION: Adds prefixes, relating a range of network addresses to a Next
 * Hop.  In order to add a prefix, the client must first add the Next Hop ID
 * and its corresponding data into the RTM, using ix_cc_rtmv6_add_next_hop().
 *
 * After adding a prefix, lookup operations, either from Core Components through
 * the RTM or from microblocks, will use the newly added prefix.
 *
 * Adding a duplicate prefix to the same or different next hop is considered an
 * error condition.  Instead, use ix_cc_rtmv6_update_prefix() to modify an
 * existing prefix.
 *
 * The client may add a default prefix by setting the netAddr and netMask to
 * zero.  As always, the nextHopID must refer to a valid next hop.  Setting the
 * default prefix is a special case and the caller needn't remove an existing
 * default prefix before setting the default prefix to a new next hop.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv6_init().
 * @Param:      - IN    arg_NetAddr
 *                      network address identifying the prefix
 * @Param:      - IN    arg_PrefixLength
 *                      prefix length for the prefix.
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be used for the prefix
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_DUPLICATE_ENTRY
 * IX_CC_ERROR_CONFLICTING_ENTRY
 * IX_CC_RTMV6_ERROR_INVALID_MASK
 * IX_CC_ERROR_FULL
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv6_add_prefix(
    ix_cc_rtmv6 arg_hRtm,
    ix_uint128 arg_NetAddr,
    ix_uint8 arg_PrefixLength,
    ix_cc_rtmv6_nhid arg_NextHopID);

/*
 * NAME: ix_cc_rtmv6_delete_prefix
 *
 * DESCRIPTION: Deletes a prefix.  The caller is responsible for removing the
 * next hop; deleting the next hop is not an implicit operation of
 * ix_cc_rtmv6_delete_prefix().
 *
 * The client may remove the default prefix by setting the netAddr and netMask
 * to zero.  Removing the default prefix is a special case and the caller may do
 * so at any time, even if the default prefix did not previously exist.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv6_init().
 * @Param:      - IN    arg_NetAddr
 *                      Network address identifying the prefix
 * @Param:      - IN    arg_PrefixLength
 *                      prefix length for the prefix.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv6_delete_prefix(
    ix_cc_rtmv6 arg_hRtm,
    ix_uint128 arg_NetAddr,
    ix_uint8 arg_PrefixLength);

/*
 * NAME: ix_cc_rtmv6_update_prefix
 *
 * DESCRIPTION: Updates prefixes, relating an existing range of network
 * addresses to a different Next Hop.  In order to update a prefix, the client
 * must first add the Next Hop ID and its corresponding data into the RTM,
 * using ix_cc_rtmv6_add_next_hop().
 *
 * After adding a prefix, lookup operations, either from Core Components through
 * the RTM or from microblocks, will use the newly added prefix.
 *
 * The client should not use the update operation to modify teh default prefix.
 * Use ix_cc_rtmv6_add_prefix() or ix_cc_rtmv6_delete_prefix() instead.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv6_init().
 * @Param:      - IN    arg_NetAddr
 *                      network address identifying the prefix
 * @Param:      - IN    arg_PrefixLength
 *                      prefix length for the prefix.
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be used for the prefix
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_CONFLICTING_ENTRY
 * IX_CC_RTMV6_ERROR_INVALID_MASK
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv6_update_prefix(
    ix_cc_rtmv6 arg_hRtm,
    ix_uint128 arg_NetAddr,
    ix_uint8 arg_PrefixLength,
    ix_cc_rtmv6_nhid arg_NextHopID);


/*
 * NAME: ix_cc_rtmv6_get_prefix
 *
 * DESCRIPTION: Gets the next hop ID for a spefcific prefix.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv6_init().
 *
 * @Param:      - IN    arg_IpAddr
 *                      IP address to be looked up
 *
 * @Param:      - IN    arg_PrefixLength
 *                      PrefixLength of the IPv6 address.
 *
 * @Param:      - OUT   arg_pNextHopID
 *                      Identifier of the Next Hop used for the route. 
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv6_get_prefix(
    ix_cc_rtmv6 arg_hRtm,
    ix_uint128  arg_Ipv6Addr,
    ix_uint8    arg_PrefixLength,
    ix_cc_rtmv6_nhid *arg_pNextHopID);

/*
 * NAME: ix_cc_rtmv6_lookup
 *
 * DESCRIPTION: Looks up routing information for a given IP address.  The RTM
 * uses an LPM algorithm to match the IP address with a network (as given by
 * calling ix_cc_rtmv6_add_prefix()) and thereby provide the information needed
 * to forward the packet to the appropriate next hop - i.e. the corresponding
 * Next Hop Information structure.
 *
 * This function is the workhorse of the RTM, and is expected to be the
 * most-used entry point; conceivably every packet arriving on the XScale core
 * will require a lookup.  Therefore, this function is designed with efficiency
 * in mind.
 *
 * If the nextHopID field in arg_pNextHopInfo is IX_CC_RTMV6_NHID_NO_ROUTE, all
 * other fields are invalid; the IP address has no route.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv6_init().
 * @Param:      - IN    arg_IpAddr
 *                      IP address to be looked up
 * @Param:      - OUT   arg_pNextHopInfo
 *                      results are placed into the structure provided by the
 *                      caller.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv6_lookup(
    ix_cc_rtmv6 arg_hRtm,
    ix_uint128 arg_IpAddr,
    ix_cc_rtmv6_next_hop_info *arg_pNextHopInfo);

/*
 * NAME: ix_cc_rtmv6_match_prefix
 *
 * DESCRIPTION: Looks up routing information for a given IP address.  The RTM
 * uses an LPM algorithm to match the IP address with a network (as given by
 * calling ix_cc_rtmv6_add_route()) and thereby provide the information needed
 * to forward the packet to the appropriate next hop - i.e. the corresponding
 * Next Hop Information structure.
 *
 * This function defers from ix_cc_ipv6_lookup above in that it returns the 
 * Next Hop ID instead of the next hop information structure.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv6_init().
 *
 * @Param:      - IN    arg_IpAddr
 *                      IP address to be looked up
 *
 * @Param:      - OUT   arg_pNextHopID
 *                      Identifier of the Next Hop used for the route. 
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv6_match_prefix(
    ix_cc_rtmv6 arg_hRtm,
    ix_uint128  arg_Ipv6Addr,
    ix_cc_rtmv6_nhid *arg_pNextHopID);

/*
 * NAME: ix_cc_rtmv6_dump_next_hops
 *
 * DESCRIPTION: Prints all Next Hops to a block of memory for debugging
 * purposes.  The memory is provided by the caller, along with the size of the
 * memory block in bytes.  The caller should use IX_CC_RTMV6_DUMP_NEXT_HOP_SIZE
 * to determine how many bytes to allocate.
 *
 * If the function returns IX_CC_ERROR_FULL, the error should be considered
 * non-fatal.  The caller may allocate more memory and call the function again.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv6_init().
 * @Param:      - OUT   arg_pBuffer
 *                      pointer to a block of memory where the next hops are
 *                      to be stored; human-readable ASCII-Z string.
 * @Param:      - IN    arg_BufferSize
 *                      size in bytes of the buffer; host byte order.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_RANGE
 * IX_CC_ERROR_FULL
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv6_dump_next_hops(
    ix_cc_rtmv6 arg_hRtm,
    char *arg_pBuffer,
    ix_uint32 arg_BufferSize);

/*
 * NAME: ix_cc_rtmv6_dump_routes
 *
 * DESCRIPTION: Prints all routes to a block of memory for debugging purposes.
 * The memory is provided by the caller, along with the size of the memory
 * block in bytes.  The caller should use IX_CC_RTMV6_DUMP_ROUTE_SIZE
 * to determine how many bytes to allocate.
 *
 * If the function returns IX_CC_ERROR_FULL, the error should be considered
 * non-fatal.  The caller may allocate more memory and call the function again.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv6_init().
 * @Param:      - OUT   arg_pBuffer
 *                      pointer to a block of memory where the routes are
 *                      to be stored; human-readable ASCII-Z string.
 * @Param:      - IN    arg_BufferSize
 *                      size in bytes of the buffer; host byte order.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_RANGE
 * IX_CC_ERROR_FULL
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv6_dump_routes(
    ix_cc_rtmv6 arg_hRtm,
    char *arg_pBuffer,
    ix_uint32 arg_BufferSize);

/*
 * NAME: ix_cc_rtmv6_purge
 *
 * DESCRIPTION: Removes all routes and next hops.  Most likely, this would
 * only be done for debugging purposes.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv6_init().
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv6_purge(
    ix_cc_rtmv6 arg_hRtm);

/*
 * NAME: ix_cc_rtmv6_purge_routes
 *
 * DESCRIPTION: Removes all routes.  Most likely, this would only be done
 * for debugging purposes.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv6_init().
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv6_purge_routes(
    ix_cc_rtmv6 arg_hRtm);

/*
 * NAME: ix_cc_rtmv6_get_symbols
 *
 * DESCRIPTION: Retrieves symbols to patch to the micoengines.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv6_init().
 * @Param:      - OUT   arg_pSymbols
 *                      structure containing values that the client may patch
 *                      to the microengines.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERAL
 */
ix_error ix_cc_rtmv6_get_symbols(
    ix_cc_rtmv6 arg_hRtm,
    ix_cc_rtmv6_symbols *arg_pSymbols);

/*
 * NAME: ix_cc_rtmv6_get_statistics
 *
 * DESCRIPTION: Retrieves statistics of the RTM.
 *
 * @Param:      - IN    arg_hRtm
 *                      handle to the RTM, as returned by ix_cc_rtmv6_init().
 *
 * @Param:      - OUT   arg_pStatistics
 *                      structure containing the statistics.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error ix_cc_rtmv6_get_statistics(
    ix_cc_rtmv6 arg_hRtm,
    ix_cc_rtmv6_statistics *arg_pStatistics);

/*****************************************************************************
 * The following macros are for internal use, and should not be used directly
 * by the client of the RTMv6.
 */
#define _IX_CC_NHDB_TITLE_STRING                                        \
    "   NHID    BladeID l2Index  Port   MTU Flags  IPAddr            Refcount\n"
#define _IX_CC_NHDB_TITLE_SIZE sizeof(_IX_CC_NHDB_TITLE_STRING)
#define _IX_CC_NHDB_TERM_STRING "...\n"
#define _IX_CC_NHDB_TERM_SIZE sizeof(_IX_CC_NHDB_TERM_STRING)
#define _IX_CC_NHDB_LINE_FMT "0x%08lX  0x%02X    0x%04X %5u %5d  0x%02X  %-15s %10lu\n"
/*
 * The line size is the number of characters in each line including the \n,
 * but excluding the null terminator.
 * 10 + 2 + 4 + 4 + 6 + 1 + 5 + 1 + 5 + 2 + 4 + 2 + 15 + 1 + 10 + 1 = 73
 */
#define _IX_CC_NHDB_LINE_SIZE 73

#define _IX_CC_RTMAPI6_TITLE_STRING                                      \
    "    Prefix                                                         nextHopID\n"
#define _IX_CC_RTMAPI6_TITLE_SIZE sizeof(_IX_CC_RTMAPI6_TITLE_STRING)
#define _IX_CC_RTMAPI6_TERM_STRING "...\n"
#define _IX_CC_RTMAPI6_TERM_SIZE sizeof(_IX_CC_RTMAPI6_TERM_STRING)
#define _IX_CC_RTMAPI6_LINE_FMT "%-66s 0x%08lX\n"
/*
 * The line size is the number of characters in each line including the \n,
 * but excluding the null terminator.
 * 66 + 1 + 10 + 1 = 78
 */
#define _IX_CC_RTMAPI6_LINE_SIZE 78

#if defined(__cplusplus)
} /* extern "C" */
#endif /* __cplusplus */

#endif /*__IX_CC_RTMV6_H__*/
