/**
 * ===========================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      internal_nhdbv6.h
 *
 * = DESCRIPTION
 *      Private header file used only by the implementation (and some white-
 *      box tests) of the Next Hop v6 Database.
 *
 * ===========================================================================
 * $Id: internal_nhdbv6.h,v 1.2 2003/11/24 21:22:45 ttschlue Exp $
 */

#if !defined(__IX_NHDBV6_INTERNAL_H__)
#define __IX_NHDBV6_INTERNAL_H__

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

#include "cc/rtmv6/nhdbv6.h"
#include "cc/ix_cc_hash_table.h"
#include "cc/ix_cc_index_pool.h"

/*
 * NAME: _ix_next_hop_info
 *
 * DESCRIPTION: Data structure that is shared with the IPv4 Microblock. Though
 * this is defined by the microblock in the design documents, it is not
 * actually provided for use in the Next Hop Database in shared header files.
 * Therefore the NHDB defines it here.  This uncompressed next hop info is 
 * used by IPv6 (always) and IPv4 when it is operating with uncompressed
 * next hop table format.
 */
typedef struct _ix_s_next_hop_info
{
    /*
     * longWord0:
     *  31:16   (24 bits)       Reserved
     *  15: 0   ( 8 bits)       Flags
     */
    ix_uint16 res0;
    ix_uint16  flags;

    /*
     * longWord1:
     *  31:24   ( 8 bits)       Fabric Port ID
     *  23:20   ( 4 bits)       L2 Index Type
     *  19:16   ( 4 bits)       Reserved
     *  15: 0   (16 bits)       Nexhop ID
     */
    ix_uint8  bladeID;
    ix_uint8  l2IndexType;
    ix_uint16 l2Index;

    /*
     * longWord2:
     *  31:16   (16 bits)       MTU
     *  15: 0   (16 bits)       Out port ID
     */
    ix_uint16 mtu;
    ix_uint16 portID;

    /*
     * longWord3:
     *  31: 0   (32 bits)       Reserved
     */
    ix_uint32 res3;
} _ix_next_hop_info;


/*
 * NAME: _ix_intern_next_hop_info
 *
 * DESCRIPTION: Data structure describing the structure of an "intermediate"
 * next hop info.
 */
typedef struct _ix_s_inter_next_hop_info
{
    /*
     * longWord0:
     *  28:31   ( 4 bits)       Status
     *  24:27   ( 4 bits)       reserved
     *  16:23   ( 8 bits)       reserved
     *  15: 0   ( 8 bits)       Flags
     */
    ix_uint8  status;
    ix_uint8  res0;
    ix_uint16 flags;

    /*
     * longWord1:
     * 16:31    ( 16 bits)      NextHopID0
     * 0:15     ( 16 bits)      NextHopID1
     */
    ix_uint16 nhid0;
    ix_uint16 nhid1;

    /*
     * longWord2:
     *  31:16   (16 bits)       NextHopID2
     *  15: 0   (16 bits)       NextHopID3
     */
    ix_uint16 nhid2;
    ix_uint16 nhid3;

    /*
     * longWord3:
     *  31: 0   (32 bits)       Reserved
     */
    ix_uint32 res3;
} _ix_inter_next_hop_info;

/*
 * NAME: _ix_cc_nhdb_next_hop_meta
 *
 * DESCRIPTION: Additional information about a next hop that the microengines
 * don't need.  There should be one of these structures for each next hop in
 * shared memory, but this information is not kept in shared memory.
 */
typedef struct _ix_s_cc_nhdb_next_hop_meta
{
    /*
     * nextHopID: Identifier the caller recognizes for this next hop.
     */
    ix_uint32 nextHopID;

    /*
     * refCount: Number of references this next hop has.
     */
    ix_uint32 refCount;

    /*
     * ipAddr: IP address for the next hop (a.k.a. gateway IP address)
     */
    ix_uint32 ipAddr;
} _ix_cc_nhdb_next_hop_meta;

/*
 * NAME: _ix_s_cc_nhdb
 *
 * DESCRIPTION: Structure containing the state information for the Next Hop
 * Database. The _ix_cc_nhdb handle is a pointer to this structure.
 */
struct _ix_s_cc_nhdb
{
    /*
     * nextHopBase: Virtual address pointer to the next hop database.
     */
    _ix_next_hop_info *nextHopBase;

    /*
     * hIndexPool: Allows us to allocate a slot out of the next hop table.
     */
    ix_cc_index_pool hIndexPool;

    /*
     * pMetaData: Array of meta data structures; one element for each next
     * hop.
     */
    _ix_cc_nhdb_next_hop_meta   *pMetaData;

    /*
     * numElements: Number of next hop elements in the database.
     */
    ix_uint32 numElements;

    /*
     * hHash: Hash Table used for converting Next Hop IDs to indexes.
     */
    ix_cc_hash_table hHash;
};

/*
 * NAME: _IX_CC_WRAP
 *
 * DESCRIPTION: In some cases, wrap the statement in a do while zero.  This
 * ensures the macro is not "misused" -- i.e. is only used as a statement, not
 * as an expression.  Remove for TestRT because having the do while affects
 * the line coverage and code complexity scores.
 */
#if defined(OPTIONAL) && !defined(TESTRT) 
#define _IX_CC_WRAP(__s) do { __s; } while (0)
#else
#define _IX_CC_WRAP(__s) __s
#endif

/*
 * These macros are used to get and set data in shared memory.  Whenever 
 * possible, they use 64-bit versions of the Resource Manager copy macros 
 * just in case they are more efficient than several 32-bit copies.
 */

/*
 * NAME: _READ_NEXT_HOP()
 *
 * DESCRIPTION: Reads an entire next hop structure from shared memory and
 * places it into local memory.
 *
 * @Param:      - IN    __hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * @Param:      - IN    __index
 *                      index into nextHopBase of the next hop to be copied.
 *
 * @Param:      - OUT   __local
 *                      Local structure where the data will be placed.
 *                      
 */
#define _READ_NEXT_HOP(__hNhdb, __index, __local) _IX_CC_WRAP(          \
    *(ix_uint64*) &(__local) = IX_RM_MEM_UINT64_READ(                   \
        (ix_uint64*) &(__hNhdb)->nextHopBase[__index]);                 \
    *(((ix_uint64*) &(__local))+1) =                                    \
        IX_RM_MEM_UINT64_READ(                                          \
            ((ix_uint64*) &(__hNhdb)->nextHopBase[__index])+1))

/*
 * NAME: _CLEAR_NEXT_HOP()
 *
 * DESCRIPTION: Initializes a shared next hop structure to zero.
 *
 * @Param:      - IN    __hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * @Param:      - IN    __index
 *                      index into nextHopBase of the next hop to be cleared.
 */
#define _CLEAR_NEXT_HOP(__hNhdb, __index) _IX_CC_WRAP(                  \
    (void) IX_RM_MEM_UINT64_WRITE(                                      \
        (ix_uint64*) &(__hNhdb)->nextHopBase[__index], 0);              \
    (void) IX_RM_MEM_UINT64_WRITE(                                      \
        ((ix_uint64*) &(__hNhdb)->nextHopBase[__index])+1, 0))

/*
 * NAME: _WRITE_NEXT_HOP()
 * 
 * DESCRIPTION: Writes a local copy of a next hop structure to shared memory.
 *
 * @Param:      - IN    __hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * @Param:      - IN    __index
 *                      index into nextHopBase of the next hop to be written.
 *
 * @Param:      - IN    __local
 *                      Local structure to be copied into shared memory.
 */
#define _WRITE_NEXT_HOP(__hNhdb, __index, __local) _IX_CC_WRAP(         \
    (void) IX_RM_MEM_UINT64_WRITE(                                      \
        (ix_uint64*) &(__hNhdb)->nextHopBase[__index],                  \
        *(ix_uint64*) &((__local)));                                    \
    (void) IX_RM_MEM_UINT64_WRITE(                                      \
        ((ix_uint64*) &(__hNhdb)->nextHopBase[__index]) +1,             \
        *((ix_uint64*) &((__local))+1)))

/*
 * NAME: _WRITE_NEXT_HOP_MTU()
 *
 * DESCRIPTION: Writes the mtu field from a local copy of a next hop structure
 * to shared memory.  Note that only 16 bits are written.
 *
 * @Param:      - IN    __hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * @Param:      - IN    __index
 *                      index into nextHopBase of the next hop to be written.
 *
 * @Param:      - IN    __local
 *                      Local structure to be copied into shared memory.
 */
#define _WRITE_NEXT_HOP_MTU(__hNhdb, __index, __local) _IX_CC_WRAP(     \
    (void) IX_RM_MEM_UINT16_WRITE(                                      \
        &(__hNhdb)->nextHopBase[__index].mtu,                           \
        (__local).mtu))

/*
 * NAME: _WRITE_NEXT_HOP_FLAGS()
 *
 * DESCRIPTION: Writes the flags field from a local copy of a next hop
 * structure to shared memory.  Note that only 8 bits are written.
 *
 * @Param:      - IN    __hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdb_init().
 *
 * @Param:      - IN    __index
 *                      index into nextHopBase of the next hop to be written.
 *
 * @Param:      - IN    __local
 *                      Local structure to be copied into shared memory.
 */
#define _WRITE_NEXT_HOP_FLAGS(__hNhdb, __index, __local) _IX_CC_WRAP(   \
    (void) IX_RM_MEM_UINT16_WRITE(                                      \
        &(__hNhdb)->nextHopBase[__index].flags,                         \
        (__local).flags))

/*
 * Macros to get and set values in the local structure.
 */

/*
 * NAME: _CLEAR_LOCAL()
 *
 * DESCRIPTION: Initializes a local next hop structure to zero.
 *
 * @Param:      - OUT   __local
 *                      structure to be cleared.
 */
#define _CLEAR_LOCAL(__local) _IX_CC_WRAP(                              \
    (__local).res0    = 0; (__local).flags   = 0;   \
    (__local).bladeID = 0; (__local).l2IndexType = 0;                   \
    (__local).l2Index = 0; (__local).portID  = 0; (__local).mtu  = 0;   \
    (__local).res3  = 0)

/*
 * NAME: _SET_LOCAL()
 *
 * DESCRIPTION: Copies all fields from an ix_cc_rtmv4_next_hop_info structure
 * into a local copy of a next hop info structure.
 *
 * @Param:      - OUT   __local
 *                      Structure to be set. This is a mirror image of the
 *                      structure in shared memory.
 *
 * @Param:      - IN    __nhi
 *                      Next Hop Information structure as provided by the
 *                      caller.
 */
#define _SET_LOCAL(__local, __nhi) _IX_CC_WRAP(                         \
    _SET_LOCAL_FLAGS(__local, (__nhi).flags);                           \
    _SET_LOCAL_BLADE(__local, (__nhi).bladeID);                         \
    _SET_LOCAL_L2INDEX_TYPE(__local, (__nhi).l2IndexType);              \
    _SET_LOCAL_L2INDEX(__local, (__nhi).l2Index);                       \
    _SET_LOCAL_MTU(__local, (__nhi).mtu);                               \
    _SET_LOCAL_PORTID(__local, (__nhi).portID);                         \
    (__local).res3 = 0)

/*
 * NAME: _SET_INTER_LOCAL()
 *
 * DESCRIPTION: Copies all fields from an ix_cc_rtmv4_next_hop_info structure
 * into a local copy of an intermediate next hop info structure.
 *
 * @Param:      - OUT   __local
 *                      Structure to be set. This is a intermediate next hop
 *                      structure.
 *
 * @Param:      - IN    __nhi
 *                      Next Hop Information structure as provided by the
 *                      caller.
 */
#define _SET_INTER_LOCAL(__local, __flags, __status, __id0, __id1, __id2, __id3 ) _IX_CC_WRAP(                                                          \
    _SET_LOCAL_FLAGS_AND_STATUS(__local, __flags, __status);            \
    _SET_LOCAL_NHIDS(__local, __id0, __id1, __id2, __id3);              \
    (__local).res3 = 0)

/*
 * NAME: _GET_FROM_LOCAL
 *
 * DESCRIPTION: Given a local copy of a next hop info structure (shared with
 * the IPv4 microblock), copy its contents into an ix_cc_rtmv4_next_hop_info
 * structure.
 *
 * @Param:      - OUT   __nhi
 *                      Next Hop Information structure to set.
 *
 * @Param:      - IN    __local
 *                      Structure to use to set the __nhi.
 */
#define _GET_FROM_LOCAL(__nhi, __local) _IX_CC_WRAP(                    \
    (__nhi).flags = (__local).flags;                                    \
    (__nhi).bladeID = (__local).bladeID;                                \
    (__nhi).l2Index = IX_NTOH16((__local).l2Index);                     \
    (__nhi).l2IndexType = (((__local).l2IndexType) >> 4);               \
    (__nhi).mtu = IX_NTOH16((__local).mtu);                             \
    (__nhi).portID = IX_NTOH16((__local).portID))

/*
 * NAME: _SET_LOCAL_FLAGS
 *
 * DESCRIPTION: Sets the flags field of the local structure.  Also clears
 * res0 and res1.
 *
 * @Param:      - OUT   __local
 *                      Structure containing the flags to be set
 *
 * @Param:      - IN    __flags
 *                      Value to set into the flags field in __local.
 */
#define _SET_LOCAL_FLAGS(__local, __flags) _IX_CC_WRAP(                 \
    (__local).flags = (ix_uint8) ((__flags) & 0x000000FFUL))

/*
 * NAME: _SET_LOCAL_BLADE
 *
 * DESCRIPTION: Sets the bladeID field of the local structure.
 *
 * @Param:      - OUT   __local
 *                      Structure containing the bladeID to be set
 *
 * @Param:      - IN    __bladeID
 *                      Value to set into the bladeID field in __local.
 */
#define _SET_LOCAL_BLADE(__local, __bladeID) _IX_CC_WRAP(               \
    (__local).bladeID = (ix_uint8) (__bladeID))

/*
 * NAME: _SET_LOCAL_FLAGS_AND_STATUS
 *
 * DESCRIPTION: Sets the flag and status field of the inter local structure.
 *
 * @Param:      - OUT   __local
 *                      Structure containing the bladeID to be set
 *
 * @Param:      - IN    __flag
 *                      Value to set into the flag field in __local.
 *
 * @Param:      - IN    __status
 *                      Value to set into the status field in __local.
 */
#define _SET_LOCAL_FLAGS_AND_STATUS(__local, __flags, __status) _IX_CC_WRAP(                 \
    *(ix_uint32*) &(__local) =                                             \
         IX_HTON32(((__flags) & 0x0000FFFFUL)| ((__status & 0xF) << 28)))

/*
 * NAME: _SET_LOCAL_NHIDS
 *
 * DESCRIPTION: Sets the next hop ids in the local structure.
 *
 * @Param:      - OUT   __local
 *                      Structure containing the bladeID to be set
 *
 * @Param:      - IN    __id 0
 *                      Value to set into the nhid0 field in __local.
 *
 * @Param:      - IN    __id 1
 *                      Value to set into the nhid1 field in __local.
 *
 * @Param:      - IN    __id 2
 *                      Value to set into the nhid2 field in __local.
 *
 * @Param:      - IN    __id 3
 *                      Value to set into the nhid3 field in __local.
 */
#define _SET_LOCAL_NHIDS(__local, __id0, __id1, __id2, __id3) _IX_CC_WRAP(  \
       (__local).nhid0 = IX_HTON16(__id0);                               \
       (__local).nhid1 = IX_HTON16(__id1);                                \
       (__local).nhid2 = IX_HTON16(__id2);                                \
       (__local).nhid3 = IX_HTON16(__id3))
/*
 * NAME: _SET_LOCAL_L2INDEX_TYPE
 *
 * DESCRIPTION: Given an 'l2IndexType' value, set the local copy of the
 * next hop info structure's l2IndexType field.
 *
 * @Param:      - OUT   __local
 *                      Structure containing the l2IndexType to be set
 *
 * @Param:      - IN    __l2IndexType
 *                      Value to set into the l2Index field in __local
 */
#define _SET_LOCAL_L2INDEX_TYPE(__local, __l2IndexType) _IX_CC_WRAP(    \
    (__local).l2IndexType = (ix_uint8) (((__l2IndexType) & 0xF) << 4))

/*
 * NAME: _SET_LOCAL_L2INDEX
 *
 * DESCRIPTION: Given an 'l2Index' value, set the local copy of the next hop
 * info structure's l2Index field.
 *
 * @Param:      - OUT   __local
 *                      Structure containing the l2Index to be set
 *
 * @Param:      - IN    __l2Index
 *                      Value to set into the l2Index field in __local
 */
#define _SET_LOCAL_L2INDEX(__local, __l2Index) _IX_CC_WRAP(             \
    (__local).l2Index = IX_HTON16(__l2Index))

/*
 * NAME: _SET_LOCAL_PORTID
 *
 * DESCRIPTION: Given a 'portID' value, set the local copy of the next hop
 * info structure's portID field.
 *
 * @Param:      - OUT   __local
 *                      Structure containing the portID to be set
 *
 * @Param:      - IN    __portID
 *                      Value to set into the portID field in __local
 */
#define _SET_LOCAL_PORTID(__local, __portID) _IX_CC_WRAP(               \
    (__local).portID = IX_HTON16(__portID))

/*
 * NAME: _SET_LOCAL_MTU
 *
 * DESCRIPTION: Given a 'mtu' value, set the local copy of the next hop info
 * structure's mtu field.
 *
 * @Param:      - OUT   __local
 *                      Structure containing the mtu to be set
 *
 * @Param:      - IN    __mtu
 *                      Value to set into the mtu field in __local
 */
#define _SET_LOCAL_MTU(__local, __mtu) _IX_CC_WRAP(                     \
    (__local).mtu = IX_HTON16(__mtu))

/*
 * _MAX_REFCOUNT
 * Used to check before incrementing a reference count so it doesn't roll over
 * to zero.
 */
#define _MAX_REFCOUNT ((ix_uint16) -1)

/*
 * NAME: _IX_NHDB_DEFAULT
 *
 * DESCRIPTION: Index of the default next hop.
 */
#define _IX_NHDB_DEFAULT 0

ix_error _ix_cc_nhdb_validate_next_hop(
    ix_cc_rtmv4_nhid arg_NextHopID,
    const ix_cc_rtmv4_next_hop_info *arg_pNextHop);

/*
 * NAME: _IX_CC_NHDB_NUM_BUCKETS
 *
 * DESCRIPTION: Arbitrary number of buckets based on the size hint provided
 * by the caller. This can be tweaked to get better results.
 */
#define _IX_CC_NHDB_NUM_BUCKETS(__hint)                                 \
   ((__hint) / 10 + 1)

/*
 * NAME: _IX_TOO_SMALL
 *
 * DESCRIPTION: Creates an error with text stating the buffer was too
 * small.
 *
 * @Param:      - IN    __size
 *                      Size of the buffer
 */
#define _IX_TOO_SMALL(__size)                                           \
    IX_ERROR_WARNING(IX_CC_ERROR_FULL,                                  \
        ("Buffer is too small (%ld bytes)\n"                            \
         "\tUse IX_CC_RTMV4_DUMP_NEXT_HOP_SIZE()", __size))

#if defined(__cplusplus)
} /* extern "C" */
#endif /* __cplusplus */

#endif /*__IX_NHDB_INTERNAL_H__*/
