/**
 * ===========================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      nhdbv6.h
 *
 * = DESCRIPTION
 *      Next Hop Database -- A subcomponent of the Route Table Manager for
 *      the IPv6 Forwarder Core Component.  These routines manage the 
 *      next hops used by the IPv6 Forwarder microblock.
 *
 * ===========================================================================
 */

#define IX_ERROR_FILE_IDENT "$Id: nhdbv6.c,v 1.2 2003/11/24 21:22:46 ttschlue Exp $"

#include "ix_ossl.h"

#if (defined (__linux__) && defined (__KERNEL__))

#else

#include <string.h>

#endif /* end (defined (__linux__) && defined (__KERNEL__)) */

#include "ix_cc_error.h"
#include "ix_rm.h"
#include "ix_cc_macros.h"
#include "cc/rtmv6/nhdbv6.h"
#include "cc/rtmv6/internal/internal_nhdbv6.h"


/* Internal function declarations */
PRIVATE ix_error _ix_cc_nhdbv6_verify_index(_ix_cc_nhdb arg_hNhdb,
                                            ix_uint32 arg_NextHopIndex);
PRIVATE ix_error _ix_cc_nhdbv6_validate_next_hop(ix_cc_rtmv6_nhid arg_NextHopID,
                                         const ix_cc_rtmv6_next_hop_info *arg_pNextHop);


/*
 * NAME: _ix_cc_nhdbv6_init
 *
 * DESCRIPTION: Creates a new database by allocating shared memory
 * accessible from the microblocks
 *
 * @Param:      - IN    arg_NhdbSizeHint
 *                      Approximate number of next hops that need to be stored.
 *
 * @Param:      - IN    arg_NhdbMemType
 *                      Type of memory to use for storing the next hops - SRAM
 *                      or DRAM.
 *
 * @Param:      - IN    arg_NhdbChannel
 *                      Channel to use for the memory. Must follow the rules
 *                      set by ix_rm_mem_alloc().
 *
 * @Param:      - OUT   arg_phNhdb
 *                      Handle to the NHDB, to be used in all other calls.
 *
 * Error Codes:
 * IX_CC_ERROR_RANGE
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_OOM_SYSTEM
 * IX_CC_ERROR_OOM
 * IX_CC_ERROR_INTERNAL
 */
ix_error _ix_cc_nhdbv6_init(
    ix_uint32 arg_NhdbSizeHint,
    ix_cc_rtmv6_mem_type arg_NhdbMemType,
    ix_uint32 arg_NhdbChannel,
    _ix_cc_nhdb *arg_phNhdb
    )
{
    ix_memory_type rmMemType;
    _ix_cc_nhdb hNhdb;
    ix_error err;

    /*
     * numElements is the total number of entries in shared memory, derived
     * from arg_NhdbChannel.  We need one extra for the default next hop.
     */
    ix_uint32 numElements = arg_NhdbSizeHint + 1;

    /*
     * nhdbMemSize is the amount of shared memory (SRAM or DRAM) required to
     * support the requested number of next hops (arg_NhdbSizeHint).
     */
    ix_uint32 nhdbMemSize = numElements * sizeof (_ix_next_hop_info);

    /*
     * heapMemSize is the amount of memory required from the heap.  This
     * includes the structure referred to by the NHDB's handle, as well 
     * as enough for all the pointers in the structure, but not including
     * the memory shared with the microblock.
     *
     * The entire block of memory from the heap is allocated with one call.
     * The memory is broken up internally rather than using multiple calls
     * to malloc().
     */
    ix_uint32 heapMemSize =
        sizeof(struct _ix_s_cc_nhdb)      +              /* Main structure */
        sizeof(_ix_cc_nhdb_next_hop_meta) * numElements; /* Meta data      */

    IX_ERROR_CHECK_ARG_MIN(1, arg_NhdbSizeHint, 1UL, "%ld");
    IX_ERROR_CHECK_ARG_NULL(3, arg_phNhdb);
    IX_ERROR_CHECK_ARG_ALIGN(3, arg_phNhdb, _ix_cc_nhdb);

    switch (arg_NhdbMemType)
    {
        case IX_CC_RTMV6_MEM_SRAM:
            rmMemType   = IX_MEMORY_TYPE_SRAM;
            break;

        case IX_CC_RTMV6_MEM_SDRAM:
            rmMemType   = IX_MEMORY_TYPE_DRAM;
            break;

        default:
            return IX_ERROR_WARNING(IX_CC_ERROR_RANGE,
                ("Invalid memory type: %d", arg_NhdbMemType));
    }

    hNhdb = (_ix_cc_nhdb) ix_ossl_malloc(heapMemSize);
    if (hNhdb == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
            ("Could not allocate %d bytes", heapMemSize));
    }
    ix_ossl_memset(hNhdb, 0, heapMemSize);

    /*
     * pMetaData should point to the area immediately after the handle's
     * structure.
     */
    hNhdb->pMetaData = (_ix_cc_nhdb_next_hop_meta*) &hNhdb[1];

    /*
     * Allocate the shared memory (either SRAM or DRAM) from the Resource
     * Manager.
     */
    err = ix_rm_mem_alloc_aligned(rmMemType, arg_NhdbChannel, nhdbMemSize,
        IX_MEMORY_ALIGNMENT_TYPE_VIRTUAL,
        3,              /* shift by 3 bytes, gives 64-bit alignment */
        (void**) &hNhdb->nextHopBase);
    if (err != IX_SUCCESS)
    {
        switch ((int) IX_ERROR_GET_CODE(err))
        {
            case IX_RM_ERROR_ASSERTION:
                err = IX_ERROR_CHAIN(err, IX_CC_ERROR_RANGE, 0,
                    ("Invalid channel: %d", arg_NhdbChannel));
                break;

            case IX_RM_ERROR_CANNOT_ALLOCATE_SIZE:
                err = IX_ERROR_CHAIN(err, IX_CC_ERROR_OOM, 0,
                    ("Out of %s memory\n"
                     "\tCould not allocate %d bytes from the RM",
                     ((rmMemType == IX_MEMORY_TYPE_SRAM) ? "SRAM" : "DRAM"),
                     nhdbMemSize));
                break;

            default:
                err = IX_ERROR_CHAIN(err, IX_CC_ERROR_INTERNAL, 0,
                    ("Could not allocate %d bytes from the RM", nhdbMemSize));
                break;
        }
        goto clean_mem;
    }

    if (hNhdb->nextHopBase == NULL)
    {
        err =  IX_ERROR_LOCAL(IX_CC_ERROR_INTERNAL,
            ("Could not allocate %d bytes from the RM", nhdbMemSize));
        goto clean_mem;
    }
    ix_ossl_memset(hNhdb->nextHopBase, 0, nhdbMemSize);

    /*
     * The first next hop is reserved for the default.  Initialize it
     * appropriately. set_default() will not return
     * IX_CC_ERROR_ENTRY_NOT_FOUND when setting the route to NO_ROUTE.
     */
    IX_ERROR_CG(_ix_cc_nhdbv6_set_default(hNhdb, IX_CC_RTMV6_NHID_NO_ROUTE),
        err, clean_rm_mem);

    /*
     * Create a hash table to map next hop identifiers to indexes into the 
     * next hop table seen by the microengines.
     */
    IX_ERROR_CG(ix_cc_hash_table_init(
                _IX_CC_NHDB_NUM_BUCKETS(numElements),
                numElements,
                IX_CC_HASH_TABLE_ALGO_MAGIC,
                &hNhdb->hHash),
            err, clean_rm_mem);

    /*
     * Create the IndexPool.
     */
    IX_ERROR_CG(ix_cc_index_pool_init(
                arg_NhdbSizeHint,
                1,                      /* Reserve the zero spot */
                &hNhdb->hIndexPool),
        err, clean_hash_table);

    hNhdb->numElements = numElements;

    *arg_phNhdb = hNhdb;
    return IX_SUCCESS;

clean_hash_table:
    (void) ix_cc_hash_table_fini(hNhdb->hHash);
clean_rm_mem:
    (void) ix_rm_mem_free(hNhdb->nextHopBase);
clean_mem:
    ix_ossl_free(hNhdb);
    return err;
}

/*
 * NAME: _ix_cc_nhdbv6_fini
 *
 * DESCRIPTION: Deallocates all resources previously allocated using
 * _ix_cc_nhdbv6_init().
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error _ix_cc_nhdbv6_fini(
    _ix_cc_nhdb arg_hNhdb
    )
{
    IX_ERROR_CHECK_ARG_NULL(1, arg_hNhdb);
    IX_ERROR_CHECK_ARG_ALIGN(1, arg_hNhdb, struct _ix_s_cc_nhdb);

    IX_ERROR_CR(ix_cc_index_pool_fini(arg_hNhdb->hIndexPool));
    arg_hNhdb->hIndexPool = NULL;

    IX_ERROR_CR(ix_cc_hash_table_fini(arg_hNhdb->hHash));
    arg_hNhdb->hHash = NULL;

    IX_ERROR_CRT(ix_rm_mem_free(arg_hNhdb->nextHopBase),
        IX_CC_ERROR_INTERNAL, 0);
    arg_hNhdb->nextHopBase = NULL;

    ix_ossl_free(arg_hNhdb);
    return IX_SUCCESS;
}

/*
 * NAME: _ix_cc_nhdbv6_add_next_hop
 *
 * DESCRIPTION: Makes a next hop usable for adding routes.  The nextHopID
 * field is stored and may be used as a handle for finding the record.  When
 * complete, the reference count for the newly added Next Hop is zero.  The
 * NHDB converts the data in arg_pNextHopInfo into a format the microengines
 * can understand.
 *
 * The NHDB prevents the caller from adding duplicate NHIDs.  The NHID value
 * of IX_CC_RTMV6_NHID_NO_ROUTE is a special case and may not be added using
 * _ix_cc_nhdbv6_add_next_hop().
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      Identifier of the Next Hop to be added
 *
 * @Param:      - IN    arg_pNextHopInfo
 *                      pointer to the Next Hop Information structure
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_RTMV6_ERROR_INVALID_NH_INFO
 * IX_CC_ERROR_CONFLICTING_ENTRY
 * IX_CC_ERROR_DUPLICATE_ENTRY
 * IX_CC_ERROR_FULL
 */
ix_error _ix_cc_nhdbv6_add_next_hop(
    _ix_cc_nhdb arg_hNhdb,
    ix_uint32 arg_NextHopID,
    const ix_cc_rtmv6_next_hop_info *arg_pNextHopInfo
    )
{
    ix_uint32 nextHopIndex = 0;
    int exists = 0;
    _ix_next_hop_info local;
    ix_error err;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hNhdb);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hNhdb, struct _ix_s_cc_nhdb);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pNextHopInfo);
    IX_ERROR_DCHECK_ARG_ALIGN(3, arg_pNextHopInfo, ix_cc_rtmv6_next_hop_info);

#ifdef IX_DEBUG
    if (arg_NextHopID == IX_CC_RTMV6_NHID_NO_ROUTE)
    {
        return IX_ERROR_WARNING(IX_CC_RTMV6_ERROR_INVALID_NH_INFO,
            ("The NextHopID was invalid: 0x%08lX", arg_NextHopID));
    }

    IX_ERROR_CR(_ix_cc_nhdbv6_validate_next_hop(arg_NextHopID,
        arg_pNextHopInfo));
#endif /* IX_DEBUG */

    IX_ERROR_CR(ix_cc_hash_table_lookup(
        arg_hNhdb->hHash, arg_NextHopID,
        &exists, (void**) &nextHopIndex));
    if (exists != 0)
    {
#ifdef OPTIONAL
        /*
         * Check the existing next hop.  If any fields are different, report
         * CONFLICTING rather than DUPLICATE.
         */
        ix_cc_rtmv6_next_hop_info existing;

        IX_ERROR_CR(_ix_cc_nhdbv6_get_nh_by_index(
            arg_hNhdb, nextHopIndex, &existing));
        if ((existing.bladeID != arg_pNextHopInfo->bladeID) || 
            (existing.l2Index != arg_pNextHopInfo->l2Index) ||
            (existing.portID != arg_pNextHopInfo->portID) ||
            (existing.mtu != arg_pNextHopInfo->mtu) ||
            (existing.flags != arg_pNextHopInfo->flags) ||
            (existing.ipAddr != arg_pNextHopInfo->ipAddr))
        {
            return IX_ERROR_WARNING(IX_CC_ERROR_CONFLICTING_ENTRY,
                ("A next hop with ID = 0x%08X already exists",
                 arg_NextHopID));
        }
#endif /* OPTIONAL */

        return IX_ERROR_WARNING(IX_CC_ERROR_DUPLICATE_ENTRY,
            ("A next hop with ID = 0x%08X already exists",
             arg_NextHopID));
    }

    /*
     * Allocate a next hop from the table and map the next hop ID to
     * the index.
     */
    IX_ERROR_CRT(ix_cc_index_pool_alloc(arg_hNhdb->hIndexPool, &nextHopIndex),
        0, IX_ERROR_LEVEL_LOCAL);
    IX_ERROR_CGT(
        ix_cc_hash_table_add_entry(
            arg_hNhdb->hHash,
            arg_NextHopID,
            (void*) nextHopIndex),
        err, IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_LOCAL, clean_index);

    /*
     * Set the contents of the local copy of the next hop according to the
     * caller's information.
     */
    _SET_LOCAL(local, *arg_pNextHopInfo);

    /*
     * Remember the next hop ID according to the index of the next hop.
     */
    arg_hNhdb->pMetaData[nextHopIndex].nextHopID = arg_NextHopID;
    arg_hNhdb->pMetaData[nextHopIndex].ipAddr    = arg_pNextHopInfo->ipAddr;

    /*
     * Copy the local structure to shared memory.
     */
    _WRITE_NEXT_HOP(arg_hNhdb, nextHopIndex, local);

    IX_CC_OPTIONAL(arg_hNhdb->pMetaData[nextHopIndex].refCount = 0);
    return IX_SUCCESS;

clean_index:
    (void) ix_cc_index_pool_free(arg_hNhdb->hIndexPool, nextHopIndex);
    return err;
}

/*
 * NAME: _ix_cc_nhdbv6_add_multiple_next_hop
 *
 * DESCRIPTION: Makes a next hop usable for adding routes.  The nextHopID
 * field is stored and may be used as a handle for finding the record.  When
 * complete, the reference count for the newly added Next Hop is zero.  The
 * NHDB converts the data in arg_pNextHopInfo into a format the microengines
 * can understand.
 *
 * The NHDB prevents the caller from adding duplicate NHIDs.  The NHID value
 * of IX_CC_RTMV6_NHID_NO_ROUTE is a special case and may not be added using
 * _ix_cc_nhdb_add_next_hop().
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      Identifier of the Next Hop to be added
 *
 * @Param:      - IN    arg_pNextHopInfo
 *                      pointer to the Next Hop Information structure
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_RTMV6_ERROR_INVALID_NH_INFO
 * IX_CC_ERROR_CONFLICTING_ENTRY
 * IX_CC_ERROR_DUPLICATE_ENTRY
 * IX_CC_ERROR_FULL
 */
ix_error _ix_cc_nhdbv6_add_multiple_next_hop(
    _ix_cc_nhdb arg_hNhdb,
    ix_uint32 arg_NextHopID,
    ix_uint32 arg_NumNextHops,
    ix_cc_rtmv6_next_hop_info *arg_pNextHopInfo
    )
{
    ix_uint32 nextHopIndex[4];
    ix_uint32 interNextHopIndex;
    ix_uint8 index = 0;
    int exists = 0;
    _ix_next_hop_info local;
    _ix_inter_next_hop_info multiple;
    ix_error err;
    ix_cc_rtmv6_next_hop_info nextHopInfo[4];

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hNhdb);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hNhdb, struct _ix_s_cc_nhdb);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pNextHopInfo);
    IX_ERROR_DCHECK_ARG_ALIGN(3, arg_pNextHopInfo, ix_cc_rtmv6_next_hop_info);

    /* If count of Next Hops is equal to 1, simply call add_next_hop above
     * and return
     */
    if (arg_NumNextHops == 1)
    {
        /* Make sure that the flag field has the "MULTIPLE" flag turned off */
	arg_pNextHopInfo->flags &= 0xFFFFFFF7;
        return _ix_cc_nhdbv6_add_next_hop(arg_hNhdb, arg_NextHopID, arg_pNextHopInfo);
    }

#ifdef IX_DEBUG
    if (arg_NextHopID == IX_CC_RTMV6_NHID_NO_ROUTE)
    {
        return IX_ERROR_WARNING(IX_CC_RTMV6_ERROR_INVALID_NH_INFO,
            ("The NextHopID was invalid: 0x%08lX", arg_NextHopID));
    }

    for (index = 0; index < arg_NumNextHops; index++)
    {
        IX_ERROR_CR(_ix_cc_nhdbv6_validate_next_hop(arg_NextHopID,
                                                  &arg_pNextHopInfo[index]));
    }

#endif /* IX_DEBUG */

    IX_ERROR_CR(ix_cc_hash_table_lookup(
                            arg_hNhdb->hHash, arg_NextHopID,
                            &exists, (void**) &nextHopIndex));
    if (exists != 0)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_DUPLICATE_ENTRY,
            ("A next hop with ID = 0x%08X already exists", arg_NextHopID));
    }

    /* Before doing anything else, duplicate the contents if there are less
     * than 4 in the array 
     */
    if (arg_NumNextHops < 4)
    {
        if (arg_NumNextHops == 2)
	{
	   /* Copy the first two elements */
	   ix_ossl_memcpy (&nextHopInfo[0], &arg_pNextHopInfo[0],
	                   sizeof(ix_cc_rtmv6_next_hop_info));
	   ix_ossl_memcpy (&nextHopInfo[1], &arg_pNextHopInfo[1],
	                   sizeof(ix_cc_rtmv6_next_hop_info));

	   /* Duplicate the last two elements */
	   ix_ossl_memcpy (&nextHopInfo[2], &arg_pNextHopInfo[0],
	                   sizeof(ix_cc_rtmv6_next_hop_info));
	   ix_ossl_memcpy (&nextHopInfo[3], &arg_pNextHopInfo[1],
	                   sizeof(ix_cc_rtmv6_next_hop_info));
	}
	else
	{
	   /* Copy the first three elements */
	   ix_ossl_memcpy (&nextHopInfo[0], &arg_pNextHopInfo[0],
	                   sizeof(ix_cc_rtmv6_next_hop_info));
	   ix_ossl_memcpy (&nextHopInfo[1], &arg_pNextHopInfo[1],
	                   sizeof(ix_cc_rtmv6_next_hop_info));
	   ix_ossl_memcpy (&nextHopInfo[2], &arg_pNextHopInfo[2],
	                   sizeof(ix_cc_rtmv6_next_hop_info));
	   /* Duplicate the last element */
	   ix_ossl_memcpy (&nextHopInfo[3], &arg_pNextHopInfo[0],
	                   sizeof(ix_cc_rtmv6_next_hop_info));
	}
    }
    else
    {
        ix_ossl_memcpy (&nextHopInfo[0], &arg_pNextHopInfo[0],
	                sizeof(ix_cc_rtmv6_next_hop_info));
        ix_ossl_memcpy (&nextHopInfo[1], &arg_pNextHopInfo[1],
	                sizeof(ix_cc_rtmv6_next_hop_info));
        ix_ossl_memcpy (&nextHopInfo[2], &arg_pNextHopInfo[2],
	                sizeof(ix_cc_rtmv6_next_hop_info));
        ix_ossl_memcpy (&nextHopInfo[3], &arg_pNextHopInfo[3],
	                sizeof(ix_cc_rtmv6_next_hop_info));
    }

    for (index = 0; index < 4; index++)
    {
        /*
         * Allocate a next hop from the table and map the next hop ID to
         * the index.
         */
        IX_ERROR_CRT(ix_cc_index_pool_alloc(arg_hNhdb->hIndexPool, 
                                                    &nextHopIndex[index]),
                     0, IX_ERROR_LEVEL_LOCAL);

        /* Meta data set up */
        arg_hNhdb->pMetaData[nextHopIndex[index]].nextHopID = arg_NextHopID;
        arg_hNhdb->pMetaData[nextHopIndex[index]].ipAddr = 0;

        IX_CC_OPTIONAL(arg_hNhdb->pMetaData[nextHopIndex[index]].refCount = 0);

        /*
         * Set the contents of the local copy of the next hop according to the
         * caller's information.
         */
        _SET_LOCAL(local, nextHopInfo[index]);

        /*
         * Copy the local structure to shared memory.
         */
        _WRITE_NEXT_HOP(arg_hNhdb, nextHopIndex[index], local);
    }

    /* Finally create an intermediate entry and write it to the shared
     * table 
     */
    /*
     * Allocate a next hop from the table and map the next hop ID to
     * the index.
     */
    IX_ERROR_CRT(ix_cc_index_pool_alloc(arg_hNhdb->hIndexPool, 
                                                    &interNextHopIndex),
                                        0, IX_ERROR_LEVEL_LOCAL);

    /*
     * Set the contents of the local copy of the intermediate next hop 
     * according to the caller's information.
     */
    _SET_INTER_LOCAL(multiple, 0x8, 0xF, nextHopIndex[0], nextHopIndex[1],
                                         nextHopIndex[2], nextHopIndex[3]);

    /*
     * Copy the local structure to shared memory.
     */
    _WRITE_NEXT_HOP(arg_hNhdb, interNextHopIndex, multiple);

    IX_ERROR_CGT(
        ix_cc_hash_table_add_entry(
            arg_hNhdb->hHash,
            arg_NextHopID,
            (void*) interNextHopIndex),
        err, IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_LOCAL, clean_index);


    /*
     * Remember the next hop ID according to the index of the next hop.
     */
    arg_hNhdb->pMetaData[interNextHopIndex].nextHopID = arg_NextHopID;
    arg_hNhdb->pMetaData[interNextHopIndex].ipAddr = 0;

    IX_CC_OPTIONAL(arg_hNhdb->pMetaData[interNextHopIndex].refCount = 0);
    return IX_SUCCESS;

clean_index:

    for (index = 0; index < 4; index++)
    {
       (void) ix_cc_index_pool_free(arg_hNhdb->hIndexPool, nextHopIndex[index]);
    }

    (void) ix_cc_index_pool_free(arg_hNhdb->hIndexPool, interNextHopIndex);
    return err;
}
/*
 * NAME: _ix_cc_nhdbv6_update_next_hop
 *
 * DESCRIPTION: Updates an existing next hop with new information. If the
 * default next hop refers to this next hop, the default is updated as well.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be updated
 *
 * @Param:      - IN    arg_pNextHopInfo
 *                      pointer to the Next Hop Information structure
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_RTMV6_ERROR_INVALID_NH_INFO
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_INTERNAL
 */
ix_error _ix_cc_nhdbv6_update_next_hop(
    _ix_cc_nhdb arg_hNhdb,
    ix_cc_rtmv6_nhid arg_NextHopID,
    const ix_cc_rtmv6_next_hop_info *arg_pNextHopInfo
    )
{
    ix_uint32 nextHopIndex;
    _ix_next_hop_info local;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hNhdb);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hNhdb, struct _ix_s_cc_nhdb);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pNextHopInfo);
    IX_ERROR_DCHECK_ARG_ALIGN(3, arg_pNextHopInfo, ix_cc_rtmv6_next_hop_info);

#ifdef IX_DEBUG
    if (arg_NextHopID == IX_CC_RTMV6_NHID_NO_ROUTE)
    {
        return IX_ERROR_WARNING(IX_CC_RTMV6_ERROR_INVALID_NH_INFO,
            ("The NextHopID was invalid: 0x%08lX", arg_NextHopID));
    }

    IX_ERROR_CR(_ix_cc_nhdbv6_validate_next_hop(
        arg_NextHopID, arg_pNextHopInfo));
#endif /* IX_DEBUG */

    /*
     * Given the next hop ID, find its index in the table.
     */
    IX_ERROR_CR(_ix_cc_nhdbv6_nhid_to_index(
        arg_hNhdb,
        arg_NextHopID,
        &nextHopIndex
        ));

#ifdef IX_DEBUG
    IX_ERROR_CRT(_ix_cc_nhdbv6_verify_index(arg_hNhdb, nextHopIndex),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_LOCAL);
#endif /* IX_DEBUG */

    /*
     * Create a local copy of the shared next hop info structure using the
     * caller's data.
     */
    _SET_LOCAL(local, *arg_pNextHopInfo);

    /*
     * Write the local copy to shared memory.
     */
    _WRITE_NEXT_HOP(arg_hNhdb, nextHopIndex, local);
    arg_hNhdb->pMetaData[nextHopIndex].ipAddr = arg_pNextHopInfo->ipAddr;

    /*
     * If the indicated next hop is the default next hop, update the default
     * also.
     */
    if (arg_NextHopID == arg_hNhdb->pMetaData[_IX_NHDB_DEFAULT].nextHopID)
    {
        _WRITE_NEXT_HOP(arg_hNhdb, _IX_NHDB_DEFAULT, local);
        arg_hNhdb->pMetaData[_IX_NHDB_DEFAULT].ipAddr =
            arg_pNextHopInfo->ipAddr;
    }

    return IX_SUCCESS;
}

/*
 * NAME: _ix_cc_nhdbv6_delete_next_hop
 *
 * DESCRIPTION: Removes a next hop.  Returns an error under the following
 * circumstances:
 *      + If arg_hNhdb is invalid;
 *      + If arg_NextHopID is IX_CC_RTMV6_NHID_NO_ROUTE;
 *      + If arg_NextHopID matches the default's arg_NextHopID;
 *      + If arg_nextHopID does not refer to a valid next hop;
 *      + If arg_NextHopID refers to a next hop with a non-zero reference
 *        count.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be removed
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_RTMV6_ERROR_NHID_IN_USE
 */
ix_error _ix_cc_nhdbv6_delete_next_hop(
    _ix_cc_nhdb arg_hNhdb,
    ix_cc_rtmv6_nhid arg_NextHopID
    )
{
    ix_uint32 nextHopIndex;
    ix_uint16 refCount;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hNhdb);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hNhdb, struct _ix_s_cc_nhdb);

#if defined(IX_DEBUG)
    if (arg_NextHopID == IX_CC_RTMV6_NHID_NO_ROUTE)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
            ("Caller attempted to remove the NO_ROUTE next hop"));
    }
#endif /* IX_DEBUG */

    if (arg_hNhdb->pMetaData[_IX_NHDB_DEFAULT].nextHopID == arg_NextHopID)
    {
        return IX_ERROR_WARNING(IX_CC_RTMV6_ERROR_NHID_IN_USE,
            ("The requested nextHopID is the default and cannot be removed"));
    }

    IX_ERROR_CR(_ix_cc_nhdbv6_nhid_to_index(
        arg_hNhdb,
        arg_NextHopID,
        &nextHopIndex
        ));

    refCount = arg_hNhdb->pMetaData[nextHopIndex].refCount;
    if (refCount != 0)
    {
        return IX_ERROR_WARNING(IX_CC_RTMV6_ERROR_NHID_IN_USE,
            ("Cannot remove nextHopID 0x%08lX because it is in use",
             arg_NextHopID));
    }

    IX_ERROR_CRT(ix_cc_index_pool_free(arg_hNhdb->hIndexPool, nextHopIndex),
        IX_CC_ERROR_INTERNAL, 0);

    IX_ERROR_CR(
        ix_cc_hash_table_delete_entry(
            arg_hNhdb->hHash,
            arg_NextHopID));

#ifdef OPTIONAL
    _CLEAR_NEXT_HOP(arg_hNhdb, nextHopIndex);
    arg_hNhdb->pMetaData[nextHopIndex].ipAddr = 0;
#endif /* OPTIONAL */

    return IX_SUCCESS;
}

/*
 * NAME: _ix_cc_nhdbv6_delete_multiple_next_hop
 *
 * DESCRIPTION: Removes a next hop.  Returns an error under the following
 * circumstances:
 *      + If arg_hNhdb is invalid;
 *      + If arg_NextHopID is IX_CC_RTMV6_NHID_NO_ROUTE;
 *      + If arg_NextHopID matches the default's arg_NextHopID;
 *      + If arg_nextHopID does not refer to a valid next hop;
 *      + If arg_NextHopID refers to a next hop with a non-zero reference
 *        count.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be removed
 * @Param       - OUT   arg_flag 
 *                      If this was a multiple next hop 1 is returned. Otherwise
 *                      0 is returned.
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_RTMV6_ERROR_NHID_IN_USE
 */
ix_error _ix_cc_nhdbv6_delete_multiple_next_hop(
    _ix_cc_nhdb arg_hNhdb,
    ix_cc_rtmv6_nhid arg_NextHopID,
    ix_uint8 *arg_pFlag
    )
{
    ix_uint32 nextHopIndex;
    ix_uint16 refCount;
    _ix_inter_next_hop_info local;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hNhdb);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hNhdb, struct _ix_s_cc_nhdb);

#if defined(IX_DEBUG)
    if (arg_NextHopID == IX_CC_RTMV6_NHID_NO_ROUTE)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
            ("Caller attempted to remove the NO_ROUTE next hop"));
    }
#endif /* IX_DEBUG */

    if (arg_hNhdb->pMetaData[_IX_NHDB_DEFAULT].nextHopID == arg_NextHopID)
    {
        return IX_ERROR_WARNING(IX_CC_RTMV6_ERROR_NHID_IN_USE,
            ("The requested nextHopID is the default and cannot be removed"));
    }

    /* It is a multiple entry next hop */
    IX_ERROR_CR(_ix_cc_nhdbv6_nhid_to_index(
        arg_hNhdb,
        arg_NextHopID,
        &nextHopIndex
        ));

    refCount = arg_hNhdb->pMetaData[nextHopIndex].refCount;
    if (refCount != 0)
    {
        return IX_ERROR_WARNING(IX_CC_RTMV6_ERROR_NHID_IN_USE,
            ("Cannot remove nextHopID 0x%08lX because it is in use",
             arg_NextHopID));
    }

    if (!_ix_cc_is_nh_entry_multiple(arg_hNhdb, arg_NextHopID))
    {
	/* 
	 * It is a "real" entry, simply call the regular delete function 
	 * and return
	 */
	*arg_pFlag = 0;
        return _ix_cc_nhdbv6_delete_next_hop(arg_hNhdb, arg_NextHopID);
    }

    /* Set the flag indicating that is a multiple next hop entry */
    *arg_pFlag = 1;
    _READ_NEXT_HOP(arg_hNhdb, nextHopIndex, local);

    /* Free all four "real" next hops */
    IX_ERROR_CRT(ix_cc_index_pool_free(arg_hNhdb->hIndexPool, 
                                       IX_NTOH16(local.nhid0)),
                     IX_CC_ERROR_INTERNAL, 0);
        
    IX_ERROR_CRT(ix_cc_index_pool_free(arg_hNhdb->hIndexPool, 
                                       IX_NTOH16(local.nhid1)),
                     IX_CC_ERROR_INTERNAL, 0);

    IX_ERROR_CRT(ix_cc_index_pool_free(arg_hNhdb->hIndexPool, 
                                       IX_NTOH16(local.nhid2)),
                     IX_CC_ERROR_INTERNAL, 0);

    IX_ERROR_CRT(ix_cc_index_pool_free(arg_hNhdb->hIndexPool, 
                                       IX_NTOH16(local.nhid3)),
                     IX_CC_ERROR_INTERNAL, 0);

    /* Free the intermediate next hop */
    IX_ERROR_CRT(ix_cc_index_pool_free(arg_hNhdb->hIndexPool, nextHopIndex),
        IX_CC_ERROR_INTERNAL, 0);

    IX_ERROR_CR(
        ix_cc_hash_table_delete_entry(
            arg_hNhdb->hHash,
            arg_NextHopID));

#ifdef OPTIONAL
    _CLEAR_NEXT_HOP(arg_hNhdb, nextHopIndex);
    arg_hNhdb->pMetaData[nextHopIndex].ipAddr = 0;
#endif /* OPTIONAL */

    return IX_SUCCESS;
}


/*
 * NAME: _ix_cc_nhdbv6_nhid_to_index
 *
 * DESCRIPTION: Gets the index associated with arg_NextHopID.  This index is
 * an intermediate reference to a Next Hop, and may be used to quickly look up
 * Next Hop Information.  This is a relatively slow operation compared to
 * _ix_cc_get_nh_by_index().
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop
 *
 * @Param:      - OUT   arg_pNextHopIndex
 *                      index identifying the Next Hop
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 */
ix_error _ix_cc_nhdbv6_nhid_to_index(
    _ix_cc_nhdb arg_hNhdb,
    ix_cc_rtmv6_nhid arg_NextHopID,
    ix_uint32 *arg_pNextHopIndex
    )
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_hNhdb);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hNhdb, struct _ix_s_cc_nhdb);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pNextHopIndex);
    IX_ERROR_DCHECK_ARG_ALIGN(3, arg_pNextHopIndex, ix_uint32);

    IX_ERROR_CR(ix_cc_hash_table_lookup(
        arg_hNhdb->hHash, arg_NextHopID, NULL, (void**) arg_pNextHopIndex));

    return IX_SUCCESS;
}

/*
 * NAME: _ix_cc_nhdbv6_index_to_nhid
 *
 * DESCRIPTION: Gets the nextHopID associated with arg_nextHopIndex.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - IN    arg_NextHopIndex
 *                      index identifying the Next Hop
 *
 * @Param:      - OUT   arg_pNextHopID
 *                      identifier of the Next Hop
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 */
ix_error _ix_cc_nhdbv6_index_to_nhid(
    _ix_cc_nhdb arg_hNhdb,
    ix_uint32 arg_NextHopIndex,
    ix_cc_rtmv6_nhid *arg_pNextHopID
    )
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_hNhdb);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hNhdb, struct _ix_s_cc_nhdb);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pNextHopID);
    IX_ERROR_DCHECK_ARG_ALIGN(3, arg_pNextHopID, ix_cc_rtmv6_nhid);

    if (arg_NextHopIndex != _IX_NHDB_DEFAULT)
    {
        IX_ERROR_CR(_ix_cc_nhdbv6_verify_index(arg_hNhdb, arg_NextHopIndex));
    }

    *arg_pNextHopID = arg_hNhdb->pMetaData[arg_NextHopIndex].nextHopID;

    return IX_SUCCESS;
}

/*
 * NAME: _ix_cc_nhdbv6_get_nh_by_nhid
 *
 * DESCRIPTION: Given an NHID, gets a copy of the Next Hop Information.  This
 * is conceptually the same as calling _ix_cc_nhdbv6_nhid_to_index() followed by
 * _ix_cc_nhdbv6_get_nh_by_index().
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop
 *
 * @Param:      - OUT   arg_pNextHopInfo
 *                      Next Hop Information structure
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 */
ix_error _ix_cc_nhdbv6_get_nh_by_nhid(
    _ix_cc_nhdb arg_hNhdb,
    ix_cc_rtmv6_nhid arg_NextHopID,
    ix_cc_rtmv6_next_hop_info *arg_pNextHopInfo
    )
{
    ix_uint32 nextHopIndex;

    /*
     * Relies on _ix_cc_nhdbv6_nhid_to_index() to check arg_hNhdb.
     */
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pNextHopInfo);
    IX_ERROR_DCHECK_ARG_ALIGN(3, arg_pNextHopInfo, ix_cc_rtmv6_next_hop_info);

    IX_ERROR_CR(_ix_cc_nhdbv6_nhid_to_index(arg_hNhdb, arg_NextHopID,
                &nextHopIndex));
    IX_ERROR_CRT(_ix_cc_nhdbv6_get_nh_by_index(arg_hNhdb, nextHopIndex,
                arg_pNextHopInfo),
            IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_LOCAL);

    return IX_SUCCESS;
}

/*
 * NAME: _ix_cc_nhdbv6_get_nh_by_nhid_multiple
 *
 * DESCRIPTION: Given an NHID, gets a copy of the Next Hop Information.  This
 * is conceptually the same as calling _ix_cc_nhdbv6_nhid_to_index() followed by
 * _ix_cc_nhdbv6_get_nh_by_index(). Note that this routine is similar to the routine
 * above but it can deal with the fact that the NH at NHID might be pointing
 * to multiple next hop entries.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop
 *
 * @Param:      - OUT   arg_pNextHopInfo
 *                      An array of Next Hop Information structures (upto 4).
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 */
ix_error _ix_cc_nhdbv6_get_nh_by_nhid_multiple(
    _ix_cc_nhdb arg_hNhdb,
    ix_cc_rtmv6_nhid arg_NextHopID,
    ix_cc_rtmv6_next_hop_info *arg_pNextHopInfo
    )
{
    ix_uint32 nextHopIndex[4];
    ix_uint32 interNextHopIndex;
    _ix_inter_next_hop_info local;
    ix_uint8 count;

    /*
     * Relies on _ix_cc_nhdbv6_nhid_to_index() to check arg_hNhdb.
     */
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pNextHopInfo);
    IX_ERROR_DCHECK_ARG_ALIGN(3, arg_pNextHopInfo, ix_cc_rtmv6_next_hop_info);

    if (!_ix_cc_is_nh_entry_multiple(arg_hNhdb, arg_NextHopID))
    {
	/* It is a single next hop entry, simply read it and return */
        return _ix_cc_nhdbv6_get_nh_by_nhid(arg_hNhdb, arg_NextHopID, arg_pNextHopInfo);
    }

    /* 
     * It is a multiple next hop entry, read all four "real" entries and return it
     * in the array (arg_pNextHopInfo).
     */

    IX_ERROR_CR(_ix_cc_nhdbv6_nhid_to_index(
        arg_hNhdb,
        arg_NextHopID,
        &interNextHopIndex));

    /* Read the "intermediate" next hop and set up the indexes in the array */
    _READ_NEXT_HOP(arg_hNhdb, interNextHopIndex, local);

    nextHopIndex[0] =  IX_NTOH16(local.nhid0);
    nextHopIndex[1] =  IX_NTOH16(local.nhid1);
    nextHopIndex[2] =  IX_NTOH16(local.nhid2);
    nextHopIndex[3] =  IX_NTOH16(local.nhid3);


    for (count = 0; count < 4; count++)
    {
        IX_ERROR_CRT(_ix_cc_nhdbv6_get_nh_by_index(arg_hNhdb, nextHopIndex[count],
                                                    &arg_pNextHopInfo[count]),
                            IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_LOCAL);
    } 

    return IX_SUCCESS;
}

/*
 * NAME: _ix_cc_nhdbv6_get_nh_by_index
 *
 * DESCRIPTION: Get a copy of the Next Hop Information, given an index.  This
 * is a relatively fast operation compared to _ix_cc_nhdbv6_nhid_to_index() and
 * _ix_cc_nhdbv6_get_nh_by_index().
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - IN    arg_NextHopIndex
 *                      index as returned by _ix_cc_nhdbv6_nhid_to_index()
 *
 * @Param:      - OUT   arg_pNextHopInfo
 *                      Next Hop Information structure
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 */
ix_error _ix_cc_nhdbv6_get_nh_by_index(
    _ix_cc_nhdb arg_hNhdb,
    ix_uint32 arg_NextHopIndex,
    ix_cc_rtmv6_next_hop_info *arg_pNextHopInfo
    )
{
    volatile _ix_next_hop_info local;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hNhdb);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hNhdb, struct _ix_s_cc_nhdb);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pNextHopInfo);
    IX_ERROR_DCHECK_ARG_ALIGN(3, arg_pNextHopInfo, ix_cc_rtmv6_next_hop_info);

    if (arg_NextHopIndex != _IX_NHDB_DEFAULT)
    {
        IX_ERROR_CR(_ix_cc_nhdbv6_verify_index(arg_hNhdb, arg_NextHopIndex));
    }

    _READ_NEXT_HOP(arg_hNhdb, arg_NextHopIndex, local);
    _GET_FROM_LOCAL(*arg_pNextHopInfo, local);
    arg_pNextHopInfo->ipAddr = arg_hNhdb->pMetaData[arg_NextHopIndex].ipAddr;

    return IX_SUCCESS;
}

/*
 * NAME: _ix_cc_nhdbv6_inc_refcount
 *
 * DESCRIPTION: Adds one to the reference count of a next hop, indicating the
 * next hop is in use.  When a next hop's reference count is non-zero, the
 * next hop may not be deleted.  If incrementing the next hop would cause it
 * to wrap to zero, ix_cc_nhdbv6_inc_refcount() returns IX_CC_ERROR_INTERNAL.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - IN    arg_NextHopIndex
 *                      index as returned by _ix_cc_nhdbv6_nhid_to_index()
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_INTERNAL
 */
ix_error _ix_cc_nhdbv6_inc_refcount(
    _ix_cc_nhdb arg_hNhdb,
    ix_uint32 arg_NextHopIndex
    )
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_hNhdb);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hNhdb, struct _ix_s_cc_nhdb);

    if (arg_NextHopIndex == _IX_NHDB_DEFAULT)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
            ("NextHopIndex %d is not valid", arg_NextHopIndex));
    }

    IX_ERROR_CR(_ix_cc_nhdbv6_verify_index(arg_hNhdb, arg_NextHopIndex));

    if (arg_hNhdb->pMetaData[arg_NextHopIndex].refCount == _MAX_REFCOUNT)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
            ("Next Hop with ID 0x%08lX has too many references",
            arg_hNhdb->pMetaData[arg_NextHopIndex].nextHopID));
    }

    arg_hNhdb->pMetaData[arg_NextHopIndex].refCount++;

    return IX_SUCCESS;
}

/*
 * NAME: _ix_cc_nhdbv6_dec_refcount
 *
 * DESCRIPTION: Decrements a next hop's reference count.  When the reference
 * count reaches zero, the client may delete the next hop using
 * _ix_cc_nhdbv6_delete_next_hop().  If decrementing the next hop would cause it
 * to go below zero, _ix_cc_nhdbv6_dec_refcount() returns IX_CC_ERROR_INTERNAL.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - IN    arg_NextHopIndex
 *                      index as returned by _ix_cc_nhdbv6_nhid_to_index()
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_INTERNAL
 */
ix_error _ix_cc_nhdbv6_dec_refcount(
    _ix_cc_nhdb arg_hNhdb,
    ix_uint32 arg_NextHopIndex
    )
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_hNhdb);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hNhdb, struct _ix_s_cc_nhdb);

    if (arg_NextHopIndex == _IX_NHDB_DEFAULT)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
            ("NextHopIndex %d is not valid", arg_NextHopIndex));
    }

    IX_ERROR_CR(_ix_cc_nhdbv6_verify_index(arg_hNhdb, arg_NextHopIndex));

    if (arg_hNhdb->pMetaData[arg_NextHopIndex].refCount == 0)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
            ("Next Hop with ID 0x%08lX has no references",
            arg_hNhdb->pMetaData[arg_NextHopIndex].nextHopID));
    }

    arg_hNhdb->pMetaData[arg_NextHopIndex].refCount--;

    return IX_SUCCESS;
}

/*
 * NAME: _ix_cc_nhdbv6_set_mtu
 *
 * DESCRIPTION: Changes the MTU for a next hop.  If the nextHopID is
 * IX_CC_RTMV6_NHID_NO_ROUTE, returns IX_CC_ERROR_ENTRY_NOT_FOUND.  If the
 * nextHopID matches the default route, the default route is updated as well.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be updated
 *
 * @Param:      - IN    arg_Mtu
 *                      new Maximum Transmission Unit
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_RTMV6_ERROR_INVALID_NH_INFO
 */
ix_error _ix_cc_nhdbv6_set_mtu(
    _ix_cc_nhdb arg_hNhdb,
    ix_cc_rtmv6_nhid arg_NextHopID,
    ix_uint32 arg_Mtu
    )
{
    ix_uint32 nextHopIndex;
    _ix_next_hop_info local;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hNhdb);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hNhdb, struct _ix_s_cc_nhdb);

#ifdef IX_DEBUG
    if (0xFFFF0000UL & arg_Mtu)
    {
        return IX_ERROR_WARNING(IX_CC_RTMV6_ERROR_INVALID_NH_INFO,
            ("Invalid MTU (%ld) when setting for NextHopID 0x%08lX",
             arg_Mtu, arg_NextHopID));
    }

    if (arg_NextHopID == IX_CC_RTMV6_NHID_NO_ROUTE)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
            ("Could not set MTU (%ld) for NextHopID 0x%08lX",
             arg_Mtu, arg_NextHopID));
    }
#endif /* IX_DEBUG */

    IX_ERROR_CR(ix_cc_hash_table_lookup(arg_hNhdb->hHash,
        arg_NextHopID, NULL, (void**) &nextHopIndex));

    _CLEAR_LOCAL(local);
    _SET_LOCAL_MTU(local, arg_Mtu);
    _WRITE_NEXT_HOP_MTU(arg_hNhdb, nextHopIndex, local);

    /*
     * If the next hop being changed is the default route, change the
     * default as well.
     */
    if (arg_hNhdb->pMetaData[nextHopIndex].nextHopID ==
        arg_hNhdb->pMetaData[_IX_NHDB_DEFAULT].nextHopID)
    {
        _WRITE_NEXT_HOP_MTU(arg_hNhdb, _IX_NHDB_DEFAULT, local);
    }

    return IX_SUCCESS;
}

/*
 * NAME: _ix_cc_nhdbv6_set_flags
 *
 * DESCRIPTION: Changes the flags for a next hop.  If the nextHopID is 
 * IX_CC_RTMV6_NHID_NO_ROUTE, returns IX_CC_ERROR_ENTRY_NOT_FOUND.  If the
 * nextHopID matches the default route, the default route is updated as well.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be updated
 *
 * @Param:      - IN    arg_Flags
 *                      new Flags
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_RTMV6_ERROR_INVALID_NH_INFO
 */
ix_error _ix_cc_nhdbv6_set_flags(
    _ix_cc_nhdb arg_hNhdb,
    ix_cc_rtmv6_nhid arg_NextHopID,
    ix_uint32 arg_Flags
    )
{
    ix_uint32 nextHopIndex;
    _ix_next_hop_info local;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hNhdb);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hNhdb, struct _ix_s_cc_nhdb);

#ifdef IX_DEBUG
    if (0xFFFF0000UL & arg_Flags)
    {
        return IX_ERROR_WARNING(IX_CC_RTMV6_ERROR_INVALID_NH_INFO,
            ("Invalid flags (0x%02lX) when setting for NextHopID 0x%08lX",
            arg_Flags, arg_NextHopID));
    }

    if (arg_NextHopID == IX_CC_RTMV6_NHID_NO_ROUTE)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
            ("Could not set flags (%ld) for NextHopID 0x%08lX",
             arg_Flags, arg_NextHopID));
    }
#endif /* IX_DEBUG */

    IX_ERROR_CR(ix_cc_hash_table_lookup(arg_hNhdb->hHash,
        arg_NextHopID, NULL, (void**) &nextHopIndex));

    _CLEAR_LOCAL(local);
    _SET_LOCAL_FLAGS(local, arg_Flags);
    _WRITE_NEXT_HOP_FLAGS(arg_hNhdb, nextHopIndex, local);

    /*
     * If the next hop being changed is the default route, change the
     * default as well.
     */
    if (arg_hNhdb->pMetaData[nextHopIndex].nextHopID ==
        arg_hNhdb->pMetaData[_IX_NHDB_DEFAULT].nextHopID)
    {
        _WRITE_NEXT_HOP_FLAGS(arg_hNhdb, _IX_NHDB_DEFAULT, local);
    }

    return IX_SUCCESS;
}

/*
 * NAME: _ix_cc_nhdbv6_set_default
 *
 * DESCRIPTION: Indicates the default should either not exist or be a copy of
 * the provided nextHopID.  The nextHopID argument must either be
 * IX_CC_RTMV6_NHID_NO_ROUTE, or must refer to a next hop previously added
 * using _ix_cc_nhdbv6_add_next_hop().
 * 
 * If the client does not specifically set the default route, the No Route
 * case (IX_CC_RTMV6_NHID_NO_ROUTE) is implied.
 *
 * If arg_NextHopID is IX_CC_RTMV6_NHID_NO_ROUTE, the nextHopID for the
 * default route is set to IX_CC_RTMV6_NHID_NO_ROUTE.  All other fields are
 * invalid.
 *
 * If arg_NextHopID is not IX_CC_RTMV6_NHID_NO_ROUTE, looks for the indicated
 * next hop.  If it does not exist, returns an error.  If it does exist,
 * copies the structure into the default location.
 *
 * The reference count for the default is not affected.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop to be used for the default
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 */
ix_error _ix_cc_nhdbv6_set_default(
    _ix_cc_nhdb arg_hNhdb,
    ix_cc_rtmv6_nhid arg_NextHopID
    )
{
    _ix_next_hop_info local;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hNhdb);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hNhdb, struct _ix_s_cc_nhdb);

    if (arg_NextHopID == IX_CC_RTMV6_NHID_NO_ROUTE)
    {
        _CLEAR_LOCAL(local);
        _SET_LOCAL_L2INDEX(local, IX_CC_RTMV6_L2INDEX_NO_ROUTE);
        local.bladeID = 0xFF;
        arg_hNhdb->pMetaData[_IX_NHDB_DEFAULT].nextHopID =
            IX_CC_RTMV6_NHID_NO_ROUTE;
        arg_hNhdb->pMetaData[_IX_NHDB_DEFAULT].ipAddr = 0;
    }
    else
    {
        ix_uint32 nextHopIndex;
        IX_ERROR_CR(_ix_cc_nhdbv6_nhid_to_index(
            arg_hNhdb, arg_NextHopID, &nextHopIndex));

        _READ_NEXT_HOP(arg_hNhdb, nextHopIndex, local);
        arg_hNhdb->pMetaData[_IX_NHDB_DEFAULT].nextHopID = arg_NextHopID;
        arg_hNhdb->pMetaData[_IX_NHDB_DEFAULT].ipAddr =
            arg_hNhdb->pMetaData[nextHopIndex].ipAddr;
    }

    _WRITE_NEXT_HOP(arg_hNhdb, _IX_NHDB_DEFAULT, local);
    
    return IX_SUCCESS;
}

/*
 * NAME: _ix_cc_nhdbv6_get_default
 *
 * DESCRIPTION: Gets the default next hop.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - OUT   arg_pNextHopInfo
 *                      results are placed into the provided structure
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 */
ix_error _ix_cc_nhdbv6_get_default(
    _ix_cc_nhdb arg_hNhdb,
    ix_cc_rtmv6_next_hop_info *arg_pNextHopInfo
    )
{
    return _ix_cc_nhdbv6_get_nh_by_index(arg_hNhdb, _IX_NHDB_DEFAULT,
        arg_pNextHopInfo);
}

/*
 * NAME: _ix_cc_nhdbv6_get_base_address
 *
 * DESCRIPTION: Because the NHDB shares the data structure with microblocks,
 * the client must have access to the base address of the memory block used to
 * store the Next Hops.  This function provides the base address.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - OUT   arg_pBaseAddress
 *                      base address of the database, to be patched to
 *                      microengines
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error _ix_cc_nhdbv6_get_base_address(
    _ix_cc_nhdb arg_hNhdb,
    ix_uint32 *arg_pBaseAddress
    )
{
    ix_memory_type type;
    ix_uint32 channel;
    ix_error err;

    ix_uint32 baseAddress;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hNhdb);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hNhdb, struct _ix_s_cc_nhdb);
    IX_ERROR_DCHECK_ARG_NULL(2, arg_pBaseAddress);
    IX_ERROR_DCHECK_ARG_ALIGN(2, arg_pBaseAddress, ix_uint32);

    err = ix_rm_get_phys_offset(
        arg_hNhdb->nextHopBase,
        &type, &channel, &baseAddress, NULL);
    if (err != IX_SUCCESS)
    {
        IX_ERROR_CHAIN(err, IX_CC_ERROR_INTERNAL, 0, 
            ("Could not get Offset for address %p\n",
             arg_hNhdb->nextHopBase));
    }

    if (type == IX_MEMORY_TYPE_SRAM)
    {
        IX_CC_SET_SRAM_CHANNEL_INTO_OFFSET(channel, baseAddress);
    }

    *arg_pBaseAddress = baseAddress;

    return IX_SUCCESS;
}

/*
 * NAME: append_next_hop
 *
 * DESCRIPTION: Appends the text of one next hop to the existing string.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - IN    arg_NextHopIndex
 *                      Index of the next hop to append.
 *
 * @Param:      - INOUT arg_pBuffer
 *                      The running string buffer.
 *
 * @Param:      - INOUT arg_pBufSizeLeft
 *                      The amount of buffer remaining.
 *                      Use IX_CC_RTMV6_DUMP_NEXT_HOP_SIZE().
 *
 * @Returns:
 * zero if the line was added, or non-zero if the buffer was full
 */
static int append_next_hop(
    _ix_cc_nhdb arg_hNhdb,
    ix_uint32 arg_NextHopIndex,
    char *arg_pBuffer,
    ix_uint32 *arg_pBufSizeLeft)
{
    _ix_next_hop_info local;
    char line[_IX_CC_NHDB_LINE_SIZE+1]; /* plus 1 for EOL */
    ix_uint32 ipAddr;
    char ipAddrString[16];          /* nnn.nnn.nnn.nnn\0 */

    /*
     * If we can't store more lines, append the terminator and quit.
     */
    if (*arg_pBufSizeLeft < _IX_CC_NHDB_LINE_SIZE+1)
    {
        size_t length = strlen(arg_pBuffer);

        /*
         * If there's not enough room for the terminator, back the string
         * up by the line size before appending the terminator.
         */
        if (*arg_pBufSizeLeft < _IX_CC_NHDB_TERM_SIZE+1)
        {
            arg_pBuffer[length-_IX_CC_NHDB_LINE_SIZE] = '\0';
        }
        strcat(arg_pBuffer, _IX_CC_NHDB_TERM_STRING);
        *arg_pBufSizeLeft += _IX_CC_NHDB_LINE_SIZE - _IX_CC_NHDB_TERM_SIZE - 1;
        return 1;
    }

    _READ_NEXT_HOP(arg_hNhdb, arg_NextHopIndex, local);
    ipAddr = arg_hNhdb->pMetaData[arg_NextHopIndex].ipAddr;
    sprintf(ipAddrString, "%ld.%ld.%ld.%ld",
        (ipAddr >> 24) & 0xFF,
        (ipAddr >> 16) & 0xFF,
        (ipAddr >>  8) & 0xFF,
        (ipAddr      ) & 0xFF);

    sprintf(line, _IX_CC_NHDB_LINE_FMT,
        arg_hNhdb->pMetaData[arg_NextHopIndex].nextHopID,
        local.bladeID,
        IX_NTOH16(local.l2Index),
        IX_NTOH16(local.portID),
        IX_NTOH16(local.mtu),
        local.flags,
        ipAddrString,
        arg_hNhdb->pMetaData[arg_NextHopIndex].refCount);
    strcat(arg_pBuffer, line);

    *arg_pBufSizeLeft -= _IX_CC_NHDB_LINE_SIZE;
    return 0;
}

/*
 * NAME: _ix_cc_nhdbv6_dump
 *
 * DESCRIPTION: Prints all Next Hop Information Structures contained in the
 * NHDB into a string buffer.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - OUT   arg_pBuffer
 *                      Buffer contains textual information about the contents
 *                      of the NHDB
 *
 * @Param:      - IN    arg_BufferSize
 *                      Number of bytes in the buffer
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_RANGE
 * IX_CC_ERROR_FULL
 */
ix_error _ix_cc_nhdbv6_dump(
    _ix_cc_nhdb arg_hNhdb,
    char *arg_pBuffer,
    ix_uint32 arg_BufferSize
    )
{
    ix_uint32 nextHopIndex;
    ix_uint32 bufSizeLeft = arg_BufferSize;
    ix_error err;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hNhdb);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hNhdb, struct _ix_s_cc_nhdb);
    IX_ERROR_DCHECK_ARG_NULL(2, arg_pBuffer);
    IX_ERROR_DCHECK_ARG_MIN(3, arg_BufferSize, 1UL, "%ld");

    if (bufSizeLeft < _IX_CC_NHDB_TITLE_SIZE + _IX_CC_NHDB_TERM_SIZE + 1)
    {
        arg_pBuffer[0] = '\0';
        return _IX_TOO_SMALL(arg_BufferSize);
    }

    strcpy(arg_pBuffer, _IX_CC_NHDB_TITLE_STRING);
    bufSizeLeft -= _IX_CC_NHDB_TITLE_SIZE;

    if (append_next_hop(arg_hNhdb, _IX_NHDB_DEFAULT,
        arg_pBuffer, &bufSizeLeft))
    {
        return _IX_TOO_SMALL(arg_BufferSize);
    }
    err = ix_cc_index_pool_get_first(arg_hNhdb->hIndexPool, &nextHopIndex);
    while (err == IX_SUCCESS)
    {
        if (append_next_hop(arg_hNhdb, nextHopIndex,
            arg_pBuffer, &bufSizeLeft))
        {
            return _IX_TOO_SMALL(arg_BufferSize);
        }
        err = ix_cc_index_pool_get_next(arg_hNhdb->hIndexPool,
            nextHopIndex, &nextHopIndex);
    }
    if (IX_ERROR_GET_CODE(err) != IX_CC_ERROR_ENTRY_NOT_FOUND)
    {
        return IX_ERROR_CHAIN(err, IX_CC_ERROR_INTERNAL, 0,
            ("Internal error while enumerating the next hops"));
    }

    return IX_SUCCESS;
}

/*
 * NAME: append_next_hop_multiple
 *
 * DESCRIPTION: Appends the text of one next hop to the existing string. This
 * routine is slightly different from append_next_hop in that it can deal
 * with "intermediate" next hops.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - IN    arg_NextHopIndex
 *                      Index of the next hop to append.
 *
 * @Param:      - INOUT arg_pBuffer
 *                      The running string buffer.
 *
 * @Param:      - INOUT arg_pBufSizeLeft
 *                      The amount of buffer remaining.
 *                      Use IX_CC_RTMV6_DUMP_NEXT_HOP_SIZE().
 *
 * @Returns:
 * zero if the line was added, or non-zero if the buffer was full
 */
static int append_next_hop_multiple(
    _ix_cc_nhdb arg_hNhdb,
    ix_uint32 arg_NextHopIndex,
    char *arg_pBuffer,
    ix_uint32 *arg_pBufSizeLeft)
{
    _ix_next_hop_info local;
    char line[_IX_CC_NHDB_LINE_SIZE+1]; /* plus 1 for EOL */
    ix_uint32 ipAddr;
    char ipAddrString[16];          /* nnn.nnn.nnn.nnn\0 */

    /*
     * If we can't store more lines, append the terminator and quit.
     */
    if (*arg_pBufSizeLeft < _IX_CC_NHDB_LINE_SIZE+1)
    {
        size_t length = strlen(arg_pBuffer);

        /*
         * If there's not enough room for the terminator, back the string
         * up by the line size before appending the terminator.
         */
        if (*arg_pBufSizeLeft < _IX_CC_NHDB_TERM_SIZE+1)
        {
            arg_pBuffer[length-_IX_CC_NHDB_LINE_SIZE] = '\0';
        }
        strcat(arg_pBuffer, _IX_CC_NHDB_TERM_STRING);
        *arg_pBufSizeLeft += _IX_CC_NHDB_LINE_SIZE - _IX_CC_NHDB_TERM_SIZE - 1;
        return 1;
    }

    _READ_NEXT_HOP(arg_hNhdb, arg_NextHopIndex, local);

    /* Before appending it to the buffer, check to see if it is an 
     * intermediate next hop? If it is, simply skip it.
     */
    if (local.flags & 0x08)
    {
       /* The multiple flag is set, it must be an intermediate entry, skip it! */
       return 0;
    }
    
    ipAddr = arg_hNhdb->pMetaData[arg_NextHopIndex].ipAddr;
    sprintf(ipAddrString, "%ld.%ld.%ld.%ld",
        (ipAddr >> 24) & 0xFF,
        (ipAddr >> 16) & 0xFF,
        (ipAddr >>  8) & 0xFF,
        (ipAddr      ) & 0xFF);

    sprintf(line, _IX_CC_NHDB_LINE_FMT,
        arg_hNhdb->pMetaData[arg_NextHopIndex].nextHopID,
        local.bladeID,
        IX_NTOH16(local.l2Index),
        IX_NTOH16(local.portID),
        IX_NTOH16(local.mtu),
        local.flags,
        ipAddrString,
        arg_hNhdb->pMetaData[arg_NextHopIndex].refCount);
    strcat(arg_pBuffer, line);

    *arg_pBufSizeLeft -= _IX_CC_NHDB_LINE_SIZE;
    return 0;
}

/*
 * NAME: _ix_cc_nhdbv6_dump_multiple
 *
 * DESCRIPTION: Prints all Next Hop Information Structures contained in the
 * NHDB into a string buffer.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - OUT   arg_pBuffer
 *                      Buffer contains textual information about the contents
 *                      of the NHDB
 *
 * @Param:      - IN    arg_BufferSize
 *                      Number of bytes in the buffer
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_RANGE
 * IX_CC_ERROR_FULL
 */
ix_error _ix_cc_nhdbv6_dump_multiple(
    _ix_cc_nhdb arg_hNhdb,
    char *arg_pBuffer,
    ix_uint32 arg_BufferSize
    )
{
    ix_uint32 nextHopIndex;
    ix_uint32 bufSizeLeft = arg_BufferSize;
    ix_error err;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hNhdb);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hNhdb, struct _ix_s_cc_nhdb);
    IX_ERROR_DCHECK_ARG_NULL(2, arg_pBuffer);
    IX_ERROR_DCHECK_ARG_MIN(3, arg_BufferSize, 1UL, "%ld");

    if (bufSizeLeft < _IX_CC_NHDB_TITLE_SIZE + _IX_CC_NHDB_TERM_SIZE + 1)
    {
        arg_pBuffer[0] = '\0';
        return _IX_TOO_SMALL(arg_BufferSize);
    }

    strcpy(arg_pBuffer, _IX_CC_NHDB_TITLE_STRING);
    bufSizeLeft -= _IX_CC_NHDB_TITLE_SIZE;

    if (append_next_hop_multiple(arg_hNhdb, _IX_NHDB_DEFAULT,
        arg_pBuffer, &bufSizeLeft))
    {
        return _IX_TOO_SMALL(arg_BufferSize);
    }
    err = ix_cc_index_pool_get_first(arg_hNhdb->hIndexPool, &nextHopIndex);
    while (err == IX_SUCCESS)
    {
        if (append_next_hop_multiple(arg_hNhdb, nextHopIndex,
            arg_pBuffer, &bufSizeLeft))
        {
            return _IX_TOO_SMALL(arg_BufferSize);
        }
        err = ix_cc_index_pool_get_next(arg_hNhdb->hIndexPool,
            nextHopIndex, &nextHopIndex);
    }
    if (IX_ERROR_GET_CODE(err) != IX_CC_ERROR_ENTRY_NOT_FOUND)
    {
        return IX_ERROR_CHAIN(err, IX_CC_ERROR_INTERNAL, 0,
            ("Internal error while enumerating the next hops"));
    }

    return IX_SUCCESS;
}
/*
 * NAME: _ix_cc_nhdbv6_purge
 *
 * DESCRIPTION: Removes all Next Hop Information Structures from the NHDB.
 * Sets the default's l2Index to IX_CC_RTMV6_L2INDEX_NO_ROUTE.  The reference
 * counts for the next hops are not checked before the next hops are deleted.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_INTERNAL
 */
ix_error _ix_cc_nhdbv6_purge(
    _ix_cc_nhdb arg_hNhdb
    )
{
    ix_uint32 nextHopIndex;
    ix_error err;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hNhdb);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hNhdb, struct _ix_s_cc_nhdb);

    /*
     * First clear out shared memory.
     */
    ix_ossl_memset(arg_hNhdb->nextHopBase, 0,
            arg_hNhdb->numElements * sizeof(_ix_next_hop_info));

    /*
     * Reset the default
     */
    IX_ERROR_CR(_ix_cc_nhdbv6_set_default(arg_hNhdb, IX_CC_RTMV6_NHID_NO_ROUTE));

    err = ix_cc_hash_table_get_first_entry(arg_hNhdb->hHash, NULL,
        (void**) &nextHopIndex);
    while (err == IX_SUCCESS)
    {
        arg_hNhdb->pMetaData[nextHopIndex].refCount = 0;

        IX_ERROR_CRT(
            ix_cc_hash_table_delete_entry(arg_hNhdb->hHash,
                arg_hNhdb->pMetaData[nextHopIndex].nextHopID),
            IX_CC_ERROR_INTERNAL, 0);

        err = ix_cc_hash_table_get_first_entry(arg_hNhdb->hHash, NULL,
            (void**) &nextHopIndex);
    }
    if (IX_ERROR_GET_CODE(err) != IX_CC_ERROR_ENTRY_NOT_FOUND)
    {
        return IX_ERROR_CHAIN(err, IX_CC_ERROR_INTERNAL, 0,
            ("Error while enumerating next hops for deletion"));
    }

    IX_ERROR_CRT(ix_cc_index_pool_reset(arg_hNhdb->hIndexPool),
        IX_CC_ERROR_INTERNAL, IX_ERROR_LEVEL_LOCAL);

    return IX_SUCCESS;
}

/*
 * NAME: _ix_cc_nhdbv6_reset_refcounts
 *
 * DESCRIPTION: Sets all next hop reference counts to zero.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 */
ix_error _ix_cc_nhdbv6_reset_refcounts(
    _ix_cc_nhdb arg_hNhdb
    )
{
    ix_uint32 nextHopIndex;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hNhdb);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hNhdb, struct _ix_s_cc_nhdb);

    for (nextHopIndex = 0;
        nextHopIndex < arg_hNhdb->numElements;
        nextHopIndex++)
    {
        arg_hNhdb->pMetaData[nextHopIndex].refCount = 0;
    }

    return IX_SUCCESS;
}

/*
 * NAME: _ix_cc_nhdbv6_validate_next_hop
 *
 * DESCRIPTION: Examines a next hop information structure to ensure its
 * contents conform to the requirements.  If not, returns
 * IX_CC_RTMV6_ERROR_INVALID_NH_INFO.
 *
 * @Param:      - IN    arg_NextHopID
 *                      identifier of the Next Hop
 *
 * @Param:      - IN    arg_pNextHopInfo
 *                      Next Hop Information structure
 *
 * Error Codes:
 * IX_CC_RTMV6_ERROR_INVALID_NH_INFO
 */
PRIVATE ix_error _ix_cc_nhdbv6_validate_next_hop(
    ix_cc_rtmv6_nhid arg_NextHopID,
    const ix_cc_rtmv6_next_hop_info *arg_pNextHop)
{
    /*
     * l2Index, portID and mtu are limited to 16 bits; bladeID and flags are
     * 8 bits as is l2IndexType.
     */
    if ((0xFFFFFF00UL & arg_pNextHop->bladeID) ||
        (0xFFFF0000UL & arg_pNextHop->l2Index) ||
        (0xFFFF0000UL & arg_pNextHop->portID)  ||
        (0xFFFF0000UL & arg_pNextHop->mtu)     ||
        (0xFFFFFF00UL & arg_pNextHop->flags)   ||
        (0xFFFFFFF0UL & arg_pNextHop->l2IndexType))
    {
        return IX_ERROR_WARNING(IX_CC_RTMV6_ERROR_INVALID_NH_INFO,
            ("Invalid Info for NextHopID 0x%08lX:\n"
             "\tbladeID=0x%02lX l2Index=0x%04X portID=0x%04lX "
             "MTU=%ld flags=0x%02lX l2IndexType=0x%lX",
            arg_NextHopID,
            arg_pNextHop->bladeID,
            arg_pNextHop->l2Index,
            arg_pNextHop->portID,
            arg_pNextHop->mtu,
            arg_pNextHop->flags,
            arg_pNextHop->l2IndexType));
    }

    if (arg_pNextHop->l2Index == IX_CC_RTMV6_L2INDEX_NO_ROUTE)
    {
        return IX_ERROR_WARNING(IX_CC_RTMV6_ERROR_INVALID_NH_INFO,
            ("Invalid Info for NextHopID 0x%08lX:\n"
             "\tThe l2Index must not be 0x%08lX",
            IX_CC_RTMV6_L2INDEX_NO_ROUTE));
    }

    return IX_SUCCESS;
}

/*
 * NAME: _ix_cc_nhdbv6_verify_index
 *
 * DESCRIPTION: Internal function that verifies whether an index is correct.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - IN    arg_NextHopIndex
 *                      The index to verify.
 *
 * Error Codes:
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 * IX_CC_ERROR_INTERNAL
 */
ix_error _ix_cc_nhdbv6_verify_index(
    _ix_cc_nhdb arg_hNhdb,
    ix_uint32 arg_NextHopIndex)
{
    ix_error err;
    int inUse;

    err = ix_cc_index_pool_in_use(arg_hNhdb->hIndexPool, arg_NextHopIndex, &inUse);

    /*
     * If the pool index returns RANGE, it just means the caller sent in
     * an invalid index.  Translate it to ENTRY_NOT_FOUND.
     */
    if (err != IX_SUCCESS)
    {
        if (IX_ERROR_GET_CODE(err) == IX_CC_ERROR_RANGE)
        {
            return IX_ERROR_CHAIN(err, IX_CC_ERROR_ENTRY_NOT_FOUND, 0,
                ("Could not find next hop with index #%08ld",
                 arg_NextHopIndex));
        }
        return IX_ERROR_LOCAL(IX_CC_ERROR_INTERNAL,
            ("Internal error while validating index #%08ld",
             arg_NextHopIndex));
    }

    if (!inUse)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
            ("Could not find next hop with index #%08ld",
             arg_NextHopIndex));
    }

    return IX_SUCCESS;
}

/*
 * NAME: _ix_cc_is_nh_entry_multiple
 *
 * DESCRIPTION: Checks to see if an entry is intermediate.
 *
 * @Param:      - IN    arg_hNhdb
 *                      Handle to the NHDB, as returned by _ix_cc_nhdbv6_init().
 *
 * @Param:      - IN    arg_NextHopIndex
 *                      index as returned by _ix_cc_nhdbv6_nhid_to_index()
 *
 * Error Codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 */
ix_error _ix_cc_is_nh_entry_multiple(
    _ix_cc_nhdb arg_hNhdb,
    ix_cc_rtmv6_nhid arg_NextHopID)
{
    _ix_next_hop_info local;
    ix_uint32 nextHopIndex;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hNhdb);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hNhdb, struct _ix_s_cc_nhdb);

    IX_ERROR_CR(_ix_cc_nhdbv6_nhid_to_index(arg_hNhdb, arg_NextHopID,
                                                 &nextHopIndex));

    _READ_NEXT_HOP(arg_hNhdb, nextHopIndex, local);

    if (local.flags & 0x08)
    {
       /* The multiple flag is set, it must be an intermediate entry */
       return 1;
    }

    return 0;
}

