/*
 * ===========================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      internal_genex.h
 *
 * = DESCRIPTION
 *      Internal header file for the Generation/Index mechanism.  This header
 *      should only be included by the implementation of and white-box tests
 *      for the genex.
 *
 * ===========================================================================
 * $Id: internal_genex.h,v 1.2 2002/09/18 17:01:53 lbmiller Exp $
 */

#ifndef __IX_CC_INTERNAL_GENEX_H__
#define __IX_CC_INTERNAL_GENEX_H__

#include "cc/ix_cc_genex.h"

/*
 * Currently we are using a 32-bit number to hold the information for an
 * ix_genex.  16 of those bits are reserved for the generation, and the other
 * 16 are for the index.  Of the 16 for the index, four indicate the buffer
 * (for a total of 16 buffers).
 *
 */
#define IX_GENEX_BITS               (32UL)
#define IX_GENEX_ALL_SET            ((ix_genex) -1)

#define BITS_RESERVED_FOR_BUFFERS   (4)
#define INITIAL_TOTAL_BUFFERS       (1 << BITS_RESERVED_FOR_BUFFERS)

#define IX_GENEX_GENERATION_BITS    (16)
#define IX_GENEX_INDEX_BITS         (IX_GENEX_BITS - IX_GENEX_GENERATION_BITS)
#define IX_GENEX_GEN_MASK           (IX_GENEX_ALL_SET << IX_GENEX_INDEX_BITS)
#define IX_GENEX_INDEX_MASK         (~IX_GENEX_GEN_MASK)

/*
 * IX_GENEX_MAX_GENERATION is a bit pattern where the low order bits are
 * all set and the high order bits are not.  The number of bits set is
 * dependent on IX_GENEX_GENERATION_BITS.  For example, if
 * IX_GENEX_GENERATION_BITS is 16, IX_GENEX_MAX_GENERATION is 0x0000FFFF.
 * This allows a user of the #defines to maintain a large generation count
 * (for debugging purposes perhaps) but only a limited number of those bits
 * are meaningful when comparing with the generation portion of the ix_genex.
 * IX_GENEX_MAX_GENERATION can be used as a mask to find the meaningful bits.
 */
#define IX_GENEX_MAX_GENERATION     (IX_GENEX_ALL_SET >> IX_GENEX_INDEX_BITS)

/*
#define IX_GENEX_STRICT_CLEANUP
*/

/*
 * Calculate the buffer number and index into that buffer of the next
 * available genex.  This is more complicated than it might at first appear.
 * Start by finding which buffer contains the next available entry.  The
 * user said to use n bits for holding the index, meaning bit n+1 is the
 * first bit significant for finding the buffer index.  For example, if
 * the quantity_factor was first given as 6 and the next available genex
 * has an id of 0x0000020A.  Each buffer will contain 2^6 = 64 entries.
 * For our purposes only the bottom two bytes are important: 0x020A, or
 * 0000 0010 0000 1010.  The bottom six bits (00 1010) are the index into
 * the buffer, and the rest (0000 0010 00) are the buffer to use.  Finding
 * the buffer number is easy: shift right by 6.  This gives 00 0000 1000, or
 * 0x008.  However finding the index into the buffer is a bit more
 * difficult.  First we must generate a mask.  The mask we want to use has
 * 6 ones in the lowest bits and zeros in all other positions: 0x003F.  We
 * can get this by starting with 0xFFFF and right shifting it by all but
 * 6 positions: 0xFFFF >> (16-6).  Then mask 0x020A with 0x003F to get
 * 0x000A, our index into the buffer.
 *
 * The following macros do this for us.
 *  param __g       source genex number
 *  param __b       number of bits dedicated to the index
 */
#define BUFFER_NUMBER(__g, __b) ((__g & IX_GENEX_INDEX_MASK) >> __b)
#define BUFFER_OFFSET(__g, __b) \
    (__g & (IX_GENEX_INDEX_MASK >> (IX_GENEX_INDEX_BITS-__b)))

/*
 *  param __n       buffer number
 *  param __i       index into buffer
 *  param __b       number of bits dedicated to the index
 *  param __g       generation number (not ix_genex)
 */
#define PACK_GENEX(__n, __i, __b, __g) \
    ((__g << IX_GENEX_INDEX_BITS) | (__n << __b) | __i)

/*
 * The ix_genex_genptr structure is one member of a two-dimensional array.
 * Each member represents the data (internal to the genex and external for
 * the user's purposes) along with the generation count and a flag saying
 * whether it is currently in use or not.  The user data is actually a
 * pointer into one of several one-dimensional arrays of malloc'd memory.
 */
typedef struct ix_s_genex_genptr {
    ix_uint16 generation;
    ix_uint16 available;
    void *user_data;
} ix_genex_genptr;

struct ix_s_genex_pool {
    ix_uint16 element_size;
    ix_uint16 quantity_factor;
    ix_genex_genptr **buffers;
    ix_uint16 total_buffers;
    ix_uint16 used_buffers;
    ix_genex last_used;
};

struct ix_s_genex_iter {
    ix_genex_pool pool;
    ix_genex last_reported;
    int active;
};

#endif /*__IX_CC_INTERNAL_GENEX_H__*/
