/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      internal_hash_table.h
 *
 * = DESCRIPTION
 *      This is the internal header file for the hash table, and should not be
 *      included by anything except the hash table and white-box tests.  Use
 *      hash_table.h instead.
 *
 * ============================================================================
 * $Id: internal_hash_table.h,v 1.7 2003/08/19 23:12:12 ktseng Exp $
 */

#if !defined(__IX_CC_INTERNAL_HASH_TABLE_H__)
#define __IX_CC_INTERNAL_HASH_TABLE_H__

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

/*
 * The magic number is (SQRT(5) - 1)/2, and is used to derive an instance-
 * specific multiplier based on the number of buckets.  This comes from Knuth's
 * The Art of Computer Programming section 6.4, Theorem S.
 */
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
#define _IX_CC_HASH_TABLE_MAGIC 618
#else
#define _IX_CC_HASH_TABLE_MAGIC 0.61803398875
#endif /* end (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */ 
#define _IX_CC_HASH_TABLE_INVALID_ELEMENT ((ix_uint16) -1)

/*
 * NAME: _ix_cc_hash_table_link
 *
 * DESCRIPTION: Structure that describes a key and the data associated with
 * the key.  Instances of this structure belong in one of two linked lists:
 * 1) Free list:   Linked list of free elements that may be used when new
 *                 entries are added to the table.
 * 2) Bucket list: Linked list of keys whose values resolved to the same
 *                 bucket.  The keys in the list for each bucket are
 *                 sorted in ascending order.
 *
 * Rather than using 32 bits to store a pointer for the next and prev values,
 * the linked list is implemented using 16-bit offsets into the pLinks array.
 */
typedef struct _ix_s_cc_hash_table_link {
    /*
     * key: the 32-bit key added to the table using
     * ix_cc_hash_table_add_entry().
     */
    ix_uint32  key;

    /*
     * pData: the void pointer associated with the key.
     */
    void      *pData;

    /*
     * prev: the previous element in the linked list for the bucket the
     * element belongs in.
     */
    ix_uint16  prev;

    /*
     * next: the next element in the linked list for the bucket the element
     * belongs in.
     */
    ix_uint16  next;
} _ix_cc_hash_table_link;

/*
 * NAME: ix_s_cc_hash_table
 *
 * DESCRIPTION: The structure associated with the ix_cc_hash_table handle.
 * This structure contains all the data each instance requires to store and
 * manage a single hash table.  The structure and all associated storage is
 * allocated in ix_cc_hash_table_init(), and freed in ix_cc_hash_table_fini().
 */
struct ix_s_cc_hash_table {
    /*
     * pBuckets: Pointer to an array of buckets.  Each bucket points to the
     * head of a linked list of elements, or _IX_CC_HASH_TABLE_INVALID_ELEMENT
     * if empty.
     */
    ix_uint16                   *pBuckets;

    /*
     * numBuckets: The number of buckets used to store data in the hash
     * table.
     */
    ix_uint32                    numBuckets;

    /*
     * pLinks: Pointer to an array of links.  When in use, each link is a 
     * member of a linked list accessed from one of the buckets.  When not in
     * use, each link is a member of the free list.
     */
    _ix_cc_hash_table_link      *pLinks;

    /*
     * linkFreeList: Head of the free list.
     */
    ix_uint16                    linkFreeList;

    /*
     * algo: Algorithm used by this instance of the hash table, as the caller
     * specified in ix_cc_hash_table_init().
     */
    ix_cc_hash_table_algo        algo;

    /*
     * magic: Used only for the IX_CC_HASH_TABLE_ALGO_MAGIC algorithm.  This
     * is the multiplier used to determine which bucket a given key should be
     * placed into.
     */
    ix_uint32                    magic;
};

#if defined(__cplusplus)
} /* extern "C" */
#endif /* __cplusplus */

#endif /*__IX_CC_INTERNAL_HASH_TABLE_H__*/
