/**
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *    Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * $Id: internal_hash_table128.h,v 1.2 2003/01/25 00:00:40 lbmiller Exp $
 */

/**
 * @file internal_hash_table128.h
 *
 * This is the internal header file for the hash table, and should not be
 * included by anything except the hash table and white-box tests.  Use
 * ix_cc_hash_table128.h instead.
 */

#if !defined(__IX_CC_INTERNAL_HASH_TABLE128_H__)
#define __IX_CC_INTERNAL_HASH_TABLE128_H__

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

#include "cc/internal/internal_hash_table.h"
#include "cc/internal/internal_sig.h"

/*
 * These are the typedefs for the various enumerations and data structures.
 * They are listed outside of the group to clean up the generated
 * documentation.
 */
typedef struct _ix_s_cc_hash_table128_link _ix_cc_hash_table128_link;

/**
 * @defgroup HashTable128Internals 128-bit Hash Table Internals
 * Internal data structures needed by the 128-bit hash table.
 *
 * @{
 */

/**
 * Convert a 128-bit number into a 32-bit number.  Because the processor
 * cannot easily work with 128-bit numbers, cut the key from a 128-bit value
 * to a 32-bit value.  This is a part of the hashing algorithm, and results
 * in more overlap than is ideal, but makes the numbers reasonable to work
 * with.
 *
 * @param       arg_Key (IN)
 *              ix_uint128 key to be converted
 *
 * @returns 32-bit number based on the original 128-bit number.
 */
#define _IX_CC_HASH_TABLE128_MAKE_KEY32(arg_Key)                        \
    (arg_Key).aUint32[0] ^ (arg_Key).aUint32[1] ^                       \
    (arg_Key).aUint32[2] ^ (arg_Key).aUint32[3]

/**
 * Set a 128-bit value to zero.
 */
#define _IX_CC_HASH_TABLE128_CLEAR_KEY(arg_Key) IX_CC_WRAP(             \
    (arg_Key).aUint32[0] = 0;                                           \
    (arg_Key).aUint32[1] = 0;                                           \
    (arg_Key).aUint32[2] = 0;                                           \
    (arg_Key).aUint32[3] = 0)

/**
 * Verify a 128-bit Hash Table handle is valid.  Checks for such things as
 * NULL, alignment, and the proper signature.
 */
#define _IX_CC_HASH_TABLE128_CHECK_HANDLE(__hHash) IX_CC_WRAP(          \
    IX_ERROR_DCHECK_ARG_NULL(1, __hHash);                               \
    IX_ERROR_DCHECK_ARG_ALIGN(1, __hHash, struct _ix_s_cc_hash_table128); \
    _IX_CC_SIG_CHECK((__hHash)->sig, _ix_cc_hash_table128_sig))

/**
 * Structure that describes a key and the data associated with
 * the key.  Instances of this structure belong in one of two linked lists:
 * @li Free list:    Linked list of free elements that may be used when new
 *                   entries are added to the table.
 * @li Bucket list:  Linked list of keys whose values resolved to the same
 *                   bucket.
 *
 * Rather than using 32 bits to store a pointer for the next and prev values,
 * the linked list is implemented using 16-bit offsets into the pLinks array.
 *
 * @typename _ix_cc_hash_table128_link
 */
struct _ix_s_cc_hash_table128_link
{
    /**
     * The 128-bit key added to the table using
     * ix_cc_hash_table128_add_entry().
     */
    ix_uint128  key;

    /**
     * The void pointer associated with the key.
     */
    void      *pData;

    /**
     * The previous element in the linked list for the bucket the element
     * belongs in.
     */
    ix_uint16  prev;

    /**
     * The next element in the linked list for the bucket the element belongs
     * in.
     */
    ix_uint16  next;
};

/**
 * The structure associated with the ix_cc_hash_table128 handle.  This
 * structure contains all the data each instance requires to store and manage
 * a single hash table.  The structure and all associated storage is allocated
 * in ix_cc_hash_table128_init(), and freed in ix_cc_hash_table128_fini().
 */
struct _ix_s_cc_hash_table128 {
    _IX_CC_SIG_LOCAL(sig);

    /**
     * Pointer to an array of buckets.  Each bucket points to the head of a
     * linked list of elements, or _IX_CC_HASH_TABLE_INVALID_ELEMENT if empty.
     */
    ix_uint16                   *pBuckets;

    /**
     * The number of buckets used to store data in the hash table.
     */
    ix_uint32                    numBuckets;

    /**
     * Pointer to an array of links.  When in use, each link is a member of a
     * linked list accessed from one of the buckets.  When not in use, each
     * link is a member of the free list.
     */
    _ix_cc_hash_table128_link   *pLinks;

    /**
     * Head of the free list.
     */
    ix_uint16                    linkFreeList;

    /**
     * Algorithm used by this instance of the hash table, as the caller
     * specified in ix_cc_hash_table128_init().
     */
    ix_cc_hash_table_algo        algo;

    /**
     * Number of elements the hash table can contain, as the caller specified
     * in ix_cc_hash_table128_init().
     */
    ix_uint32                    numElements;

    /**
     * Used only for the IX_CC_HASH_TABLE_ALGO_MAGIC algorithm.  This is the
     * multiplier used to determine which bucket a given key should be placed
     * into.
     */
    ix_uint32                    magic;
};

/**
 * @} End of 128-bit Hash Table Internals
 */

#if defined(__cplusplus)
} /* extern "C" */
#endif /* __cplusplus */

/*vim:tw=78*/
#endif /*__IX_CC_INTERNAL_HASH_TABLE128_H__*/
