/*
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *    Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * $Id: internal_sig.h,v 1.2 2003/01/25 00:00:40 lbmiller Exp $
 *
 * Note that the format and location of the comments in this file are
 * significant, as they are used to generate documentation about the module.
 * Take care when modifying this file.
 */

/**
 * @file internal_sig.h
 * Private header file for clients using the signatures.
 */

#ifndef __IX_CC_SIGNATURES_H__
#define __IX_CC_SIGNATURES_H__

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

#ifdef IX_DEBUG
#define _IX_CC_ENABLE_SIGNATURES_
#endif /* IX_DEBUG */

/*
 * These are the typedefs for the various enumerations and data structures.
 * They are listed outside of the group to clean up the generated
 * documentation.
 */
typedef struct _ix_s_cc_sig _ix_cc_sig;

/**
 * @defgroup Signatures Signatures
 * Data structure and macros to create signatures. Signatures are unique
 * values that can be embedded in data structures and checked for validity,
 * ensuring a data structure returns to the caller as expected.  They are
 * of particular use for callback-style functions and opaque handles.
 *
 * @{
 */

/**
 * Structure for containing a signature. Compatible with Win32 GUIDs
 * (use guidgen.exe from Microsoft Visual Studio).
 *
 * @typename _ix_cc_sig
 */
struct _ix_s_cc_sig
{
    ix_uint32 data1;    /**< First data element */
    ix_uint16 data2;    /**< Second data element */
    ix_uint16 data3;    /**< Third data element */
    ix_uint8 data4[8];  /**< Fourth data element */
};

#ifndef _IX_CC_ENABLE_SIGNATURES_
/*
 * If signatures are disabled, provide empty macros.  Put the empty ones first
 * so Doxygen picks up the full implementations.
 */
#define _IX_CC_SIG_CHECK(__sig1, __sig2)
#define _IX_CC_SIG_GLOBAL(name, a, b, c, d0, d1, d2, d3, d4, d5, d6, d7)
#define _IX_CC_SIG_LOCAL(name)
#define _IX_CC_SIG_COPY(src, dest)

#else /* _IX_CC_ENABLE_SIGNATURES_ */

/**
 * Compares two signatures and returns an error if they don't match.
 * @param       __sig1
 *              Signature structure
 *
 * @param       __sig2
 *              Signature structure
 */
#define _IX_CC_SIG_CHECK(__sig1, __sig2)                                \
    if (((__sig1).data1    != (__sig2).data1)    ||                     \
        ((__sig1).data2    != (__sig2).data2)    ||                     \
        ((__sig1).data3    != (__sig2).data3)    ||                     \
        ((__sig1).data4[0] != (__sig2).data4[0]) ||                     \
        ((__sig1).data4[1] != (__sig2).data4[1]) ||                     \
        ((__sig1).data4[2] != (__sig2).data4[2]) ||                     \
        ((__sig1).data4[3] != (__sig2).data4[3]) ||                     \
        ((__sig1).data4[4] != (__sig2).data4[4]) ||                     \
        ((__sig1).data4[5] != (__sig2).data4[5]) ||                     \
        ((__sig1).data4[6] != (__sig2).data4[6]) ||                     \
        ((__sig1).data4[7] != (__sig2).data4[7]))                       \
        return IX_ERROR_LOCAL(IX_CC_ERROR_INTERNAL,                     \
            ("Signature mismatch (most likely a bad pointer)"));

/**
 * Defines a new signature.  This is the master signature that local copies
 * should be compared against.  A static variable is created with this name.
 *
 * @param       name
 *              Name of the signature to create.
 *
 * @param       a
 *              32-bit value to be assigned to data1
 *
 * @param       b
 *              16-bit value to be assigned to data2
 *
 * @param       c
 *              16-bit value to be assigned to data3
 *
 * @param       d0
 *              8-bit value to be assigned to data4[0]
 *
 * @param       d1
 *              8-bit value to be assigned to data4[1]
 *
 * @param       d2
 *              8-bit value to be assigned to data4[2]
 *
 * @param       d3
 *              8-bit value to be assigned to data4[3]
 *
 * @param       d4
 *              8-bit value to be assigned to data4[4]
 *
 * @param       d5
 *              8-bit value to be assigned to data4[5]
 *
 * @param       d6
 *              8-bit value to be assigned to data4[6]
 *
 * @param       d7
 *              8-bit value to be assigned to data4[7]
 */
#define _IX_CC_SIG_GLOBAL(name, a, b, c, d0, d1, d2, d3, d4, d5, d6, d7)\
    static _ix_cc_sig name = {                                          \
        a, b, c, {d0, d1, d2, d3, d4, d5, d6, d7}};

/**
 * Defines a local signature.  An example use would be to place a local
 * signature as a member of a context structure used for callbacks.
 *
 * @param       name
 *              Name of the local signature.  A variable is created with
 *              this name.
 */
#define _IX_CC_SIG_LOCAL(name)                                          \
    _ix_cc_sig name

/**
 * Assign a value to a local signature.
 */
#define _IX_CC_SIG_COPY(dest, src)                                      \
    (dest).data1 = (src).data1;                                         \
    (dest).data2 = (src).data2;                                         \
    (dest).data3 = (src).data3;                                         \
    (dest).data4[0] = (src).data4[0];                                   \
    (dest).data4[1] = (src).data4[1];                                   \
    (dest).data4[2] = (src).data4[2];                                   \
    (dest).data4[3] = (src).data4[3];                                   \
    (dest).data4[4] = (src).data4[4];                                   \
    (dest).data4[5] = (src).data4[5];                                   \
    (dest).data4[6] = (src).data4[6];                                   \
    (dest).data4[7] = (src).data4[7];

#endif /* _IX_CC_ENABLE_SIGNATURES_ */

/**
 * @} End of Signatures
 */

#if defined(__cplusplus)
} /* extern "C" */
#endif /* __cplusplus */

#endif /* __IX_CC_SIGNATURES_H__ */
