/**
 * ===========================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      hash_table.h
 *
 * = DESCRIPTION
 *      Public header for using the hash table.  This implementation stores
 *      void pointers.
 *
 * ===========================================================================
 * $Id: ix_cc_hash_table.h,v 1.2 2002/09/18 17:01:53 lbmiller Exp $
 */

#if !defined(__IX_CC_HASH_TABLE_H__)
#define __IX_CC_HASH_TABLE_H__

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

#include "ix_cc_error.h"

/*
 * NAME: ix_cc_hash_table
 *
 * DESCRIPTION: Public handle for using the hash table.
 */
typedef struct ix_s_cc_hash_table* ix_cc_hash_table;

/*
 * NAME: ix_cc_hash_table_algo
 *
 * DESCRIPTION: Enumeration of hashing algorithms supported by the hash
 * table.
 */
typedef enum ix_e_cc_hash_table_algo {
    /*
     * MAGIC: Based on a 'magic' number; multiplying a key by this number
     * results in a pseudo-random bucket position.
     */
    IX_CC_HASH_TABLE_ALGO_MAGIC,

    /*
     * NONE: No hashing, simply use the key modulo the number of buckets to
     * determine the bucket position.
     */
    IX_CC_HASH_TABLE_ALGO_NONE,

    /*
     * XOR8: Divide the 32-bit key into four 8-bit values and perform an
     * exclusive-or operation on all four values.
     */
    IX_CC_HASH_TABLE_ALGO_XOR8,

    /*
     * XOR16: Divide the 32-bit key into two 16-bit values and perform an
     * exclusive-or operation on them.
     */
    IX_CC_HASH_TABLE_ALGO_XOR16,

    IX_CC_HASH_TABLE_ALGO_LAST
} ix_cc_hash_table_algo;

/*
 * NAME: ix_cc_hash_table_init
 *
 * DESCRIPTION: Creates a new hash table.  All memory is allocated with a
 * single call, decreasing the chances for cache misses.
 *
 * @Param:      - IN arg_NumBuckets
 *                   Specifies how many buckets the hash table should allocate
 * @Param:      - IN arg_NumElements
 *                   Specifies how many elements the hash table will need to
 *                   hold
 * @Param:      - IN arg_Algo
 *                   Specifies the algorithm to use for this instantiation of
 *                   the hash table
 * @Param:      - OUT arg_phHashTable
 *                    handle to the newly-created hash table
 */
ix_error ix_cc_hash_table_init(
    ix_uint32             arg_NumBuckets,
    ix_uint32             arg_NumElements,
    ix_cc_hash_table_algo arg_Algo,
    ix_cc_hash_table     *arg_phHashTable);

/*
 * NAME: ix_cc_hash_table_fini
 *
 * DESCRIPTION: Destroys an existing hash table.
 *
 * @Param:      - IN arg_hHashTable
 *                   handle to the Hash Table, as returned by
 *                   ix_cc_hash_table_init()
 */
ix_error ix_cc_hash_table_fini(
    ix_cc_hash_table      arg_hHashTable);

/*
 * NAME: ix_cc_hash_table_add_entry
 *
 * DESCRIPTION: Adds an entry to the hash table.
 *
 * @Param:      - IN arg_hHashTable
 *                   handle to the Hash Table, as returned by
 *                   ix_cc_hash_table_init()
 * @Param:      - IN arg_Key
 *                   value to associate with the data; is run through the
 *                   hash algorithm
 * @Param:      - IN arg_pData
 *                   pointer to the data to associate with the key
 */
ix_error ix_cc_hash_table_add_entry(
    ix_cc_hash_table      arg_hHashTable,
    ix_uint32             arg_Key,
    void                 *arg_pData);

/*
 * NAME: ix_cc_hash_table_delete_entry
 *
 * DESCRIPTION: Removes an entry from the hash table.
 *
 * @Param:      - IN arg_hHashTable
 *                   handle to the Hash Table, as returned by
 *                   ix_cc_hash_table_init()
 * @Param:      - IN arg_Key
 *                   key of the entry to be deleted
 */
ix_error ix_cc_hash_table_delete_entry(
    ix_cc_hash_table      arg_hHashTable,
    ix_uint32             arg_Key);

/*
 * NAME: ix_cc_hash_table_lookup
 *
 * DESCRIPTION: Finds the data associated with a key.
 *
 * @Param:      - IN arg_hHashTable
 *                   handle to the Hash Table, as returned by
 *                   ix_cc_hash_table_init()
 * @Param:      - IN arg_Key
 *                   a value already added with ix_cc_hash_table_add_entry()
 * @Param:      - OUT arg_pExists
 *                   non-zero if the key has been added
 * @Param:      - OUT arg_ppData
 *                    If arg_pExists is non-zero, contains the data associated
 *                    with the key
 */
ix_error ix_cc_hash_table_lookup(
    ix_cc_hash_table      arg_hHashTable,
    ix_uint32             arg_Key,
    int                  *arg_pExists,
    void                **arg_ppData);

/*
 * NAME: ix_cc_hash_table_get_first_entry
 *
 * DESCRIPTION: Returns one entry from the hash table.  There is no guarantee
 * which specific entry is returned (i.e. first added vs. lowest number, etc.).
 *
 * @Param:      - IN arg_hHashTable
 *                   handle to the Hash Table, as returned by
 *                   ix_cc_hash_table_init()
 * @Param:      - OUT arg_pKey
 *                    a value already added with ix_cc_hash_table_add_entry()
 * @Param:      - OUT arg_ppData
 *                    pointer to the data associated with the key
 */
ix_error ix_cc_hash_table_get_first_entry(
    ix_cc_hash_table      arg_hHashTable,
    ix_uint32            *arg_pKey,
    void                **arg_ppData);

#if defined(__cplusplus)
} /* extern "C" */
#endif /* __cplusplus */

#endif /*__IX_CC_HASH_TABLE_H__*/
