/*
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *    Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * $Id: ix_cc_hash_table128.h,v 1.2 2003/01/25 00:00:40 lbmiller Exp $
 *
 * Note that the format and location of the comments in this file are
 * significant, as they are used to generate documentation about the module.
 * Take care when modifying this file.
 */

/**
 * @file ix_cc_hash_table128.h
 *
 * Public header for using the 128-bit hash table.
 */

#if !defined(__IX_CC_HASH_TABLE128_H__)
#define __IX_CC_HASH_TABLE128_H__

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

#include "ix_cc_error.h"
#include "ix_cc_hash_table.h"

/**
 * @defgroup HashTable128 128-bit Hash Table
 * The 128-bit Hash Table is nearly identical to the ix_cc_hash_table module,
 * except that it uses a 128-bit key instead of a 32-bit key.  Because a
 * 128-bit key is difficult to use, part of the hashing algorithm is to shorten
 * it to 32 bits.  This has the disadvantage of causing more clashes.
 *
 * @{
 */

/**
 * Public handle for using the hash table.
 */
typedef struct _ix_s_cc_hash_table128* ix_cc_hash_table128;

/**
 * Creates a new 128-bit hash table.  All memory is allocated with a single
 * call, decreasing the chances for cache misses.
 *
 * @param       arg_NumBuckets (IN)
 *              Specifies how many buckets the hash table should allocate
 *
 * @param       arg_NumElements (IN)
 *              Specifies how many elements the hash table will need to hold
 *
 * @param       arg_Algo (IN)
 *              Specifies the algorithm to use for this instantiation of the
 *              hash table.  See ix_cc_hash_table.h.
 *
 * @param       arg_phHashTable (OUT)
 *              Handle to the newly-created hash table
 */
ix_error ix_cc_hash_table128_init(
    ix_uint32             arg_NumBuckets,
    ix_uint32             arg_NumElements,
    ix_cc_hash_table_algo arg_Algo,
    ix_cc_hash_table128  *arg_phHashTable);

/**
 * Destroys an existing hash table.
 *
 * @param      arg_hHashTable (IN)
 *             handle to the Hash Table, as returned by
 *             ix_cc_hash_table128_init()
 */
ix_error ix_cc_hash_table128_fini(
    ix_cc_hash_table128   arg_hHashTable);

/**
 * Adds an entry to the hash table.
 *
 * @param       arg_hHashTable (IN)
 *              handle to the Hash Table, as returned by
 *              ix_cc_hash_table128_init()
 *
 * @param       arg_Key (IN)
 *              value to associate with the data; is run through the hash
 *              algorithm
 *
 * @param       arg_pData (IN)
 *              pointer to the data to associate with the key
 */
ix_error ix_cc_hash_table128_add_entry(
    ix_cc_hash_table128   arg_hHashTable,
    ix_uint128            arg_Key,
    void                 *arg_pData);

/**
 * Removes an entry from the hash table.
 *
 * @param       arg_hHashTable (IN)
 *              handle to the Hash Table, as returned by
 *              ix_cc_hash_table128_init()
 *
 * @param       arg_Key (IN)
 *              key of the entry to be deleted
 */
ix_error ix_cc_hash_table128_delete_entry(
    ix_cc_hash_table128   arg_hHashTable,
    ix_uint128            arg_Key);

/**
 * Finds the data associated with a key.
 *
 * @param       arg_hHashTable (IN)
 *              handle to the Hash Table, as returned by
 *              ix_cc_hash_table128_init()
 *
 * @param       arg_Key (IN)
 *              a value already added with ix_cc_hash_table128_add_entry()
 *
 * @param       arg_pExists (OUT)
 *              non-zero if the key has been added
 *
 * @param       arg_ppData (OUT)
 *              If arg_pExists is non-zero, contains the data associated with
 *              the key
 */
ix_error ix_cc_hash_table128_lookup(
    ix_cc_hash_table128   arg_hHashTable,
    ix_uint128            arg_Key,
    int                  *arg_pExists,
    void                **arg_ppData);

/**
 * Returns one entry from the hash table.  There is no guarantee which specific
 * entry is returned (i.e. first added vs. lowest number, etc.).
 *
 * @param       arg_hHashTable (IN)
 *              handle to the Hash Table, as returned by
 *              ix_cc_hash_table128_init()
 *
 * @param       arg_pKey (OUT)
 *              a value already added with ix_cc_hash_table128_add_entry()
 *
 * @param       arg_ppData (OUT)
 *              pointer to the data associated with the key
 */
ix_error ix_cc_hash_table128_get_first_entry(
    ix_cc_hash_table128 arg_hHashTable,
    ix_uint128          *arg_pKey,
    void                **arg_ppData);

/**
 * Clears out the contents of the hash table.
 *
 * @param       arg_hHashTable (IN)
 *              handle to the Hash Table, as returned by
 *              ix_cc_hash_table128_init()
 */
ix_error ix_cc_hash_table128_purge(
    ix_cc_hash_table128   arg_hHashTable);

/**
 * @} End of 128-bit Hash Table
 */

#if defined(__cplusplus)
} /* extern "C" */
#endif /* __cplusplus */

/*vim:tabstop=8:expandtab:shiftwidth=4:tw=78*/
#endif /*__IX_CC_HASH_TABLE128_H__*/
