/**
 * ===========================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      ix_cc_index_pool.h
 *
 * = DESCRIPTION
 *      Public header for using the Index Pool.  The pool allows the caller
 *      to allocate and free a number of indexes.  The index pool does not
 *      specifically provide its own thread safety; a single handle should not
 *      be used from multiple threads.  However, it is acceptable for several
 *      threads to each create and use their own index pools without
 *      interfering with other thread pools.
 *
 * ===========================================================================
 * $Id: ix_cc_index_pool.h,v 1.4 2002/09/27 16:41:18 lbmiller Exp $
 */

#if !defined(__IX_CC_INDEX_POOL_H__)
#define __IX_CC_INDEX_POOL_H__

#if defined(__cplusplus)
extern "C" {
#endif /* __cplusplus */

#include "ix_cc_error.h"

/*
 * NAME: ix_cc_index_pool
 *
 * DESCRIPTION: Public handle for using the bit array.
 */
typedef struct ix_s_cc_index_pool* ix_cc_index_pool;

/*
 * NAME: ix_cc_index_pool_init
 *
 * DESCRIPTION: Creates an index pool with the requested number of elements.
 *
 * @Param:      - IN    arg_NumElements
 *                      Number of indexes the ix_cc_index_pool should manage.
 *                      This number is the quantity beyond arg_Offset.
 *
 * @Param:      - IN    arg_Offset 
 *                      Number of reserved indexes; e.g. if arg_Offset == 2,
 *                      ix_cc_index_pool_alloc() would never return an index
 *                      of zero or one.
 *
 * @Param:      - OUT   arg_phIndexPool
 *                      Handle that should be used in future calls to the 
 *                      ix_cc_index_pool module.
 *
 * @Returns:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_OOM_SYSTEM
 */
ix_error ix_cc_index_pool_init(
    ix_uint32           arg_NumElements,
    ix_uint32           arg_Offset,
    ix_cc_index_pool   *arg_phIndexPool);

/*
 * NAME: ix_cc_index_pool_fini
 *
 * DESCRIPTION: Destroys an index pool.
 *
 * @Param:      - IN    arg_hIndexPool
 *                      Handle to the ix_cc_index_pool as returned by
 *                      ix_cc_index_pool_init().
 *
 * @Returns:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 */
ix_error ix_cc_index_pool_fini(
    ix_cc_index_pool    arg_hIndexPool);

/*
 * NAME: ix_cc_index_pool_alloc
 *
 * DESCRIPTION: Allocates an index out of the pool.
 *
 * @Param:      - IN    arg_hIndexPool
 *                      Handle to the ix_cc_index_pool as returned by
 *                      ix_cc_index_pool_init().
 *
 * @Param:      - OUT   arg_pIndex
 *                      Location where the allocated index is stored.
 *
 * @Returns:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_FULL
 */
ix_error ix_cc_index_pool_alloc(
    ix_cc_index_pool    arg_hIndexPool,
    ix_uint32          *arg_pIndex);

/*
 * NAME: ix_cc_index_pool_free
 *
 * DESCRIPTION: Frees a previously allocated index back into the pool; the
 * index may be allocated again later.
 *
 * @Param:      - IN    arg_hIndexPool
 *                      Handle to the ix_cc_index_pool as returned by
 *                      ix_cc_index_pool_init().
 *
 * @Param:      - IN    arg_Index
 *                      The index to free back into the pool.
 *
 * @Returns:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_RANGE
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 */
ix_error ix_cc_index_pool_free(
    ix_cc_index_pool    arg_hIndexPool,
    ix_uint32           arg_Index);

/*
 * NAME: ix_cc_index_pool_in_use
 *
 * DESCRIPTION: Indicates whether a given index has already been allocated.
 *
 * @Param:      - IN    arg_hIndexPool
 *                      Handle to the ix_cc_index_pool as returned by
 *                      ix_cc_index_pool_init().
 *
 * @Param:      - IN    arg_Index
 *                      The index being queried.
 *
 * @Param:      - OUT   arg_pInUse
 *                      Zero if the index has not been allocated, non-zero if
 *                      the index has been allocated.
 *
 * @Returns:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_RANGE
 */
ix_error ix_cc_index_pool_in_use(
    ix_cc_index_pool    arg_hIndexPool,
    ix_uint32           arg_Index,
    int                *arg_pInUse);

/*
 * NAME: ix_cc_index_pool_reset
 *
 * DESCRIPTION: Resets the contents of the pool so that all are unallocated.
 *
 * @Param:      - IN    arg_hIndexPool
 *                      Handle to the ix_cc_index_pool as returned by
 *                      ix_cc_index_pool_init().
 *
 * @Returns:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 */
ix_error ix_cc_index_pool_reset(
    ix_cc_index_pool    arg_hIndexPool);

/*
 * NAME: ix_cc_index_pool_get_first
 *
 * DESCRIPTION: Gets the first allocated index from the pool.
 *
 * @Param:      - IN    arg_hIndexPool
 *                      Handle to the ix_cc_index_pool as returned by
 *                      ix_cc_index_pool_init().
 *
 * @Param:      - OUT   arg_pIndex
 *                      The first allocated index in the pool.
 *
 * @Returns:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 */
ix_error ix_cc_index_pool_get_first(
    ix_cc_index_pool    arg_hIndexPool,
    ix_uint32          *arg_pIndex);

/*
 * NAME: ix_cc_index_pool_get_next
 *
 * DESCRIPTION: Gets the next allocated index from the pool, following the
 * provided index
 *
 * @Param:      - IN    arg_hIndexPool
 *                      Handle to the ix_cc_index_pool as returned by
 *                      ix_cc_index_pool_init().
 *
 * @Param:      - IN    arg_Previous
 *                      The index previously returned by
 *                      ix_cc_index_pool_get_first() or
 *                      ix_cc_index_pool_get_next().
 *
 * @Param:      - OUT   arg_pIndex
 *                      The next allocated index in the pool.
 *
 * @Returns:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 */
ix_error ix_cc_index_pool_get_next(
    ix_cc_index_pool    arg_hIndexPool,
    ix_uint32           arg_Previous,
    ix_uint32          *arg_pIndex);

#if defined(__cplusplus)
} /* extern "C" */
#endif /* __cplusplus */

#endif /* __IX_CC_INDEX_POOL_H__ */
