/**
 * ===========================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      cc_buffer.c
 *
 * = DESCRIPTION
 *      Core Components helper function(s) for dealing with buffers.
 *
 * ===========================================================================
 * $Id: cc_buffer.c,v 1.4 2003/09/30 20:16:09 ktseng Exp $
 */

#include "ix_ossl.h"
#include "ix_rm.h"
#include "ix_cc.h"
#include "ix_cc_error.h"

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <asm/io.h>
 
#endif /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

/*
 * NAME: ix_cc_hw_get_packet_data
 *
 * DESCRIPTION: Returns the pointer to the start of the packet data in a
 * hardware buffer. This is the defined as the start of the data plus the
 * offset obtained from the buffer's meta data.
 *
 * @Param:  - IN    arg_hBuffer
 *                  Handle of a hardware buffer
 *
 * @Param:  - OUT   arg_ppPacketData
 *                  Pointer to the start of packet data.
 *
 * Error codes:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_INTERNAL
 * IX_RM_ERROR_BUF_HANDLE_INVALID
 * IX_RM_ERROR_BUF_FREE_LIST_INVALID
 * IX_RM_ERROR_BUF_FREE_LIST_ID_INVALID
 */
ix_error ix_cc_hw_get_packet_data(
    ix_buffer_handle    arg_hBuffer,
    void              **arg_ppPacketData)
{
    void *pData = NULL;
    void *pMeta = NULL;
    ix_hw_buffer_meta *pHwMeta = NULL;

    IX_ERROR_DCHECK_ARG_NULL(2, arg_ppPacketData);

#if defined(IX_DEBUG)
    {
        ix_buffer_type bufferType;
        IX_ERROR_CR(ix_rm_buffer_get_type(arg_hBuffer, &bufferType));
        if (bufferType != IX_BUFFER_TYPE_HARDWARE)
        {
            return IX_ERROR_LOCAL(IX_CC_ERROR_INTERNAL,
                ("The provided buffer (%08lX) was not a hardware buffer",
                 arg_hBuffer));
        }
    }
#endif /* IX_DEBUG */

    IX_ERROR_CR(ix_rm_buffer_get_meta(arg_hBuffer, &pMeta));
    pHwMeta = (ix_hw_buffer_meta*) pMeta;

    IX_ERROR_CR(ix_rm_buffer_get_data(arg_hBuffer, &pData));
    *arg_ppPacketData = pData + pHwMeta->m_Offset;

    return IX_SUCCESS;
}

/**
 * NAME: ix_rm_cc_calculate_cell_count
 *
 * DESCRIPTION: This function will calculate the required POS cell count expected by hardware
 *  given the cell count and the buffer type: SOP or not. For the SOP buffers there are 8 extra
 *  bytes that need to go into the first cell, so the amount of data that will get into the first
 *  cell will be up to 112 bytes. For non SOP buffers all cells will get up to 120 bytes of data.
 *  The returned cell count would be what hardware expects, which would be the real cell count
 *  less 1.
 * 
 * @Param:  - IN arg_BufferSize - the size of the data buffer
 * @Param:  - IN arg_IsSOP - flag that specifies if the buffer is SOP or not. If the value is different from
 *	0, then the buffer is an SOP buffer.
 *
 * @Return: The cell count expected by hardware which is the reall cell count less 1.
 */
ix_uint32 ix_cc_calculate_cell_count(ix_uint32 arg_BufferSize, ix_uint32 arg_IsSOP)
{
    return (arg_BufferSize == 0) ? 0 : ((arg_IsSOP != 0) ? ((arg_BufferSize+7)/120) : (arg_BufferSize - 1)/120);
}


#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

/* utility functions to read/write register */
unsigned int get32(unsigned int phys_addr)
{
    void *va;
    unsigned int val;

    va = ioremap(phys_addr, 4);
    if (va == NULL) 
    {
        printk("ioremap failed: address %x\n", phys_addr);
        return 0;
    } 

    val = *(volatile unsigned int *)(va);
    iounmap(va);
    return val;

} /* get32 */


void put32(unsigned int phys_addr, unsigned int val)
{
    void *va;

    va = ioremap(phys_addr, 4);

    if (va == NULL)
    {
        printk("ioremap failed: address %x\n", phys_addr);
        return;
    }

    *(volatile unsigned int *)(va) = val;
    iounmap(va);

} /* put32 */

#endif /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

