/**
 * ===========================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      cc_net.c
 *
 * = DESCRIPTION
 *      Core Components helper function(s) for Internet 
 *      address manipulation.
 *
 * ===========================================================================
 * $Id: cc_net.c,v 1.4.2.2 2004/03/22 18:35:09 rranjeet Exp $
 */

#include "ix_ossl.h"
#include "ix_rm.h"
#include "ix_cc.h"
#include "ix_cc_error.h"

/*
* Copyright (C) 1996-2001  Internet Software Consortium.
*
* Permission to use, copy, modify, and distribute this software for any
* purpose with or without fee is hereby granted, provided that the above
* copyright notice and this permission notice appear in all copies.
*
* THE SOFTWARE IS PROVIDED "AS IS" AND INTERNET SOFTWARE CONSORTIUM
* DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL
* INTERNET SOFTWARE CONSORTIUM BE LIABLE FOR ANY SPECIAL, DIRECT,
* INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
* FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
* NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
* WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <errno.h>

#define NS_INT16SZ       2
#define NS_INADDRSZ      4
#define NS_IN6ADDRSZ    16

/*
* WARNING: Don't even consider trying to compile this on a system where
* sizeof(int) < 4.  sizeof(int) > 4 is fine; all the world's not a VAX.
*/
static int inet_pton4(const char *src, unsigned char *dst);
static int inet_pton6(const char *src, unsigned char *dst);

/* int
* ix_inet_pton(af, src, dst)
*      convert from presentation format (which usually means ASCII printable)
*      to network format (which is usually some kind of binary format).
* return:
*      1 if the address was valid for the specified address family
*      0 if the address wasn't valid (`dst' is untouched in this case)
*      -1 if some other error occurred (`dst' is untouched in this case, too)
* author:
*      Paul Vixie, 1996.
*/
int ix_inet_pton(int af, const char *src, void *dst)
{
	switch (af) {
	       	case IX_AF_INET:
		       	return (inet_pton4(src, dst));
	       	case IX_AF_INET6:
		       	return (inet_pton6(src, dst));
	       	default:

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
#else
		       	errno = EAFNOSUPPORT;
#endif
		       	return (-1);
}
/* NOTREACHED */
}
			
/* int
* inet_pton4(src, dst)
*      like inet_aton() but without all the hexadecimal and shorthand.
* return:
*      1 if `src' is a valid dotted quad, else 0.
* notice:
*      does not touch `dst' unless it's returning 1.
* author:
*      Paul Vixie, 1996.
*/
static int
inet_pton4(src, dst)
	const char *src;
	unsigned char *dst;
{
		static const char digits[] = "0123456789";
		int saw_digit, octets, ch;
		unsigned char tmp[NS_INADDRSZ], *tp;
		
		saw_digit = 0;
		octets = 0;
		*(tp = tmp) = 0;
		while ((ch = *src++) != '\0') {
			const char *pch;
			
			if ((pch = strchr(digits, ch)) != NULL) {
				unsigned int new = *tp * 10 + (pch - digits);
				if (new > 255)
					return (0);
				*tp = new;
				if (! saw_digit) {
					if (++octets > 4)
						return (0);
					saw_digit = 1;
				}
			} else if (ch == '.' && saw_digit) {
				if (octets == 4)
					return (0);
				*++tp = 0;
				saw_digit = 0;
			} else
				return (0);
		}
		if (octets < 4)
			return (0);
		ix_ossl_memcpy(dst, tmp, NS_INADDRSZ);
		return (1);
}

/* int
* inet_pton6(src, dst)
*      convert presentation level address to network order binary form.
* return:
*      1 if `src' is a valid [RFC1884 2.2] address, else 0.
* notice:
*      (1) does not touch `dst' unless it's returning 1.
*      (2) :: in a full address is silently ignored.
* credit:
*      inspired by Mark Andrews.
* author:
*      Paul Vixie, 1996.
*/
static int
inet_pton6(src, dst)
	const char *src;
	unsigned char *dst;
{
	static const char xdigits_l[] = "0123456789abcdef",
			  xdigits_u[] = "0123456789ABCDEF";
	unsigned char tmp[NS_IN6ADDRSZ], *tp, *endp, *colonp;
	const char *xdigits, *curtok;
	int ch, saw_xdigit;
	unsigned int val;
	
	ix_ossl_memset((tp = tmp), '\0', NS_IN6ADDRSZ);
	endp = tp + NS_IN6ADDRSZ;
	colonp = NULL;
	/* Leading :: requires some special handling. */
	if (*src == ':')
		if (*++src != ':')
			return (0);
	curtok = src;
	saw_xdigit = 0;
	val = 0;
	while ((ch = *src++) != '\0') {
		const char *pch;
		
		if ((pch = strchr((xdigits = xdigits_l), ch)) == NULL)
			pch = strchr((xdigits = xdigits_u), ch);
		if (pch != NULL) {
			val <<= 4;
			val |= (pch - xdigits);
			if (val > 0xffff)
				return (0);
			saw_xdigit = 1;
			continue;
		}
		if (ch == ':') {
			curtok = src;
			if (!saw_xdigit) {
				if (colonp)
					return (0);
				colonp = tp;
				continue;
			}
			if (tp + NS_INT16SZ > endp)
				return (0);
			*tp++ = (unsigned char) (val >> 8) & 0xff;
			*tp++ = (unsigned char) val & 0xff;
			saw_xdigit = 0;
			val = 0;
			continue;
		}
		if (ch == '.' && ((tp + NS_INADDRSZ) <= endp) &&
			inet_pton4(curtok, tp) > 0) {
			tp += NS_INADDRSZ;
			saw_xdigit = 0;
			break;  /* '\0' was seen by inet_pton4(). */
		}
		return (0);
	}
	if (saw_xdigit) {
		if (tp + NS_INT16SZ > endp)
			return (0);
		*tp++ = (unsigned char) (val >> 8) & 0xff;
		*tp++ = (unsigned char) val & 0xff;
	}
	if (colonp != NULL) {
	/*
	* Since some memmove()'s erroneously fail to handle
	* overlapping regions, we'll do the shift by hand.
	*/
	
		const int n = tp - colonp;
		int i;
		
		for (i = 1; i <= n; i++) {
			endp[- i] = colonp[n - i];
			colonp[n - i] = 0;
		}
		tp = endp;
	}
	if (tp != endp)
		return (0);
	
	ix_ossl_memcpy(dst, tmp, NS_IN6ADDRSZ);
	return (1);
}

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
/**
 * TYPENAME: ix_cc_rand
 *
 * DESCRIPTION: This is the function used to replace the rand() function
 *              used by VxWorks code.
 *
 * @Return: The random number generated.
 *
 */

ix_uint32 ix_cc_rand(void)
{
   struct timeval tv;
   ix_uint8 logic = 0;
   ix_uint32 number = 0;
   static ix_uint32 randNum = 0;
   static ix_uint8 flag = 0;

/* 
* IX_CC_SWITCH_MASK is used to obtain the switch case.
* IX_CC_NUM_MASK is used to obtain the first number based on time stamp 
*/
#define IX_CC_SWITCH_MASK 0x0000000000000007L
#define IX_CC_NUM_MASK  0x0000000FFFFFFFFL

   /* Obtain timestamp information */
   do_gettimeofday(&tv);

   /* Obtain the starting number and switch logic */
   logic = tv.tv_usec & IX_CC_SWITCH_MASK;
   number = tv.tv_usec & IX_CC_NUM_MASK;

   /* 
    * Keep changing the original number based in alternate generations 
    * of random number 
    */
   if (flag)
   {
       number = ~number;
       flag = 0;
   }
   else
   {
       flag = 1;
   }

   /* 
    * Based on the switch logic, perform different mathematical operations 
    * on the number. After different mathematical operations have been 
    * performed, make it more random my using the last generated random number 
    */
   switch(logic)
   {
      case 0:
         randNum = ((number/421) + 7127) ^ randNum;
         break;
      case 1:
         randNum = (number | 569) ^ randNum;
         break;
      case 2:
         randNum = ((number/421) * 43) ^ randNum;
         break;
      case 3:
         randNum = ((number % 967) + 1871) ^ randNum;
         break;
      case 4:
         randNum = (number + 6359) ^ randNum;
         break;
      case 5:
         randNum = (number + 17467) ^ randNum;
         break;
      case 6:
         randNum = (number ^ 2753) ^ randNum;
         break;
      case 7:
         randNum = (number & 16573) ^ randNum;
         break;
      default:
         /* This case should never happen */
#ifdef IX_DEBUG
         ix_ossl_message_log("ix_cc_rand() failed to generate the rand function\n");
#endif
         return 0;
         break;
   }
   return randNum;
} /* End of ix_cc_rand */
#endif /* _IX_OS_LINUX_KERNEL_ */
