/*
 * ===========================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      genex.c
 *
 * = DESCRIPTION
 *      Implementation of the generation/index storage mechanism.
 *
 * ===========================================================================
 */

#define IX_ERROR_FILE_IDENT                                             \
        "$Id: genex.c,v 1.4 2003/04/04 00:48:37 ageorges Exp $"

#include "ix_ossl.h"
#include "cc/internal/internal_genex.h"

static ix_error
initialize_buffer(
    ix_genex_genptr *genptrs[],
    ix_uint32 num_elements,
    ix_uint32 element_size
    )
{
    ix_error err;
    ix_uint32 ix;
    char *buffer;
    char *next_element;

    *genptrs = (ix_genex_genptr*)
        ix_ossl_malloc(sizeof(ix_genex_genptr) * num_elements);
    if (*genptrs == NULL) {
        return IX_ERROR_LOCAL(IX_CC_ERROR_OOM_SYSTEM,
            ("Unable to allocate %d bytes for %d genptrs",
            sizeof(ix_genex_genptr) * num_elements,
            num_elements));
    }

    buffer = (char*) ix_ossl_malloc(num_elements * element_size);
    if (buffer == NULL) {
        err = IX_ERROR_LOCAL(IX_CC_ERROR_OOM_SYSTEM,
            ("Could not allocate %d bytes for a buffer with\n"
            "\t%d elements %d bytes long",
            num_elements * element_size, num_elements, element_size));
        goto clean_genptrs;
    }

    next_element = buffer;
    for (ix = 0; ix < num_elements; ix++) {
        (*genptrs)[ix].generation = 1;
        (*genptrs)[ix].available = 1;
        (*genptrs)[ix].user_data = next_element;
        next_element += element_size;
    }

    return IX_SUCCESS;

clean_genptrs:
    ix_ossl_free(*genptrs);
    return err;
}

static ix_error
increase_list_size(
    ix_genex_pool pool
    )
{
    ix_uint32 new_total;
    ix_uint32 max_total;
    ix_uint32 ix;
    ix_genex_genptr **new_buffers;

    /*
     * Ideally we want to double the number of supported pools.  However,
     * because a 32-bit number must always be able to uniquely identify an
     * element both in space and time (i.e. identify both the buffer/offset
     * location and the generation count), we need to be careful not
     * to try to use too many bits.
     */
    new_total = pool->total_buffers * 2;

    max_total = (ix_uint32)
        (IX_GENEX_INDEX_MASK >> pool->quantity_factor) + 1;

    /*
     * If the total is overflowing, limit it at the mask.
     */
    if (new_total > max_total) {
        new_total = max_total;
    }

    /*
     * If it didn't grow at all, return an error indicating we're out of room
     */
    if (new_total == pool->total_buffers) {
        return IX_ERROR_WARNING(IX_CC_ERROR_FULL, 
            ("increase_list_size: no room to grow\n"
            "\tAll bits in the genex are already in use"));
    }

    /*
     * Make room to hold the list of buffers.
     */
    new_buffers = (ix_genex_genptr**)
        ix_ossl_malloc(sizeof(ix_genex_genptr*) * new_total);
    if (new_buffers == NULL) {
        return IX_ERROR_LOCAL(IX_CC_ERROR_OOM_SYSTEM,
            ("Could not allocate %d bytes for the list of %d buffers",
            sizeof(ix_genex_genptr*) * new_total,
            new_total));
    }

    /*
     * Copy the original list of buffers to the new array.
     */
    for (ix = 0; ix < pool->total_buffers; ix++) {
        new_buffers[ix] = pool->buffers[ix];
    }

    /*
     * Initialize the rest of the newly allocated array to NULL.
     */
    for (; ix < new_total; ix++) {
        new_buffers[ix] = NULL;
    }

    /*
     * Free the old array and start using the new one.
     */
    ix_ossl_free(pool->buffers);
    pool->buffers = new_buffers;
    pool->total_buffers = new_total;

    return IX_SUCCESS;
}

static ix_error
find_next_available(
    ix_genex_pool pool,
    ix_uint32 *nump,
    ix_uint32 *offp
    )
{
    ix_uint32 buffer_num, buffer_num_ix;
    ix_uint32 buffer_offset, buffer_offset_ix;
    ix_error err;
    ix_genex_genptr *genptr;

    buffer_num = BUFFER_NUMBER(pool->last_used, pool->quantity_factor);
    buffer_offset = BUFFER_OFFSET(pool->last_used, pool->quantity_factor);

    /*
     * First search from the last used element to the end of the list.
     */
    for (buffer_num_ix = buffer_num;
        buffer_num_ix < pool->used_buffers;
        buffer_num_ix++)
    {
        genptr = pool->buffers[buffer_num_ix];
        for (buffer_offset_ix = buffer_offset;
            buffer_offset_ix < (ix_uint32) 1<<pool->quantity_factor;
            buffer_offset_ix++)
        {
            if (genptr[buffer_offset_ix].available != 0) {
                *nump = buffer_num_ix;
                *offp = buffer_offset_ix;
                return IX_SUCCESS;
            }
        }
        buffer_offset = 0;
    }

    /*
     * Now search from the beginning up to the buffer containing
     * the last used element.
     */
    for (buffer_num_ix = 0;
        buffer_num_ix < buffer_num;
        buffer_num_ix++)
    {
        genptr = pool->buffers[buffer_num_ix];
        for (buffer_offset_ix = 0;
            buffer_offset_ix < (ix_uint32) 1<<pool->quantity_factor;
            buffer_offset_ix++)
        {
            if (genptr[buffer_offset_ix].available != 0) {
                *nump = buffer_num_ix;
                *offp = buffer_offset_ix;
                return IX_SUCCESS;
            }
        }
    }

    /*
     * Finally, search the through the remaining unsearched portion of the
     * buffer up to the last used.  Note we're only searching through one
     * buffer in this case.
     */
    buffer_num = BUFFER_NUMBER(pool->last_used, pool->quantity_factor);
    buffer_offset = BUFFER_OFFSET(pool->last_used, pool->quantity_factor);
    buffer_num_ix = buffer_num;
    genptr = pool->buffers[buffer_num_ix];
    for (buffer_offset_ix = 0;
        buffer_offset_ix < buffer_offset;
        buffer_offset_ix++)
    {
        if (genptr[buffer_offset_ix].available != 0) {
            *nump = buffer_num_ix;
            *offp = buffer_offset_ix;
            return IX_SUCCESS;
        }
    }

    /*
     * If all the existing elements are allocated, make another buffer and
     * pick the first element in the list as the next to use.
     */
    if (pool->total_buffers == pool->used_buffers) {
        IX_ERROR_CR(increase_list_size(pool));
    }

    if (pool->buffers[pool->used_buffers] != NULL) {
        return IX_ERROR_LOCAL(IX_CC_ERROR_INTERNAL,
            ("find_next_available: pointer is non-NULL\n"
            "\tThe buffer after the last one (position %d in\n"
            "\tthe array starting at %p for pool %p) should\n"
            "\tbe NULL, but is %p.",
            pool->used_buffers, pool->buffers, pool,
            pool->buffers[pool->used_buffers]));
    }

    err = initialize_buffer(&pool->buffers[pool->used_buffers],
        1 << pool->quantity_factor,
        pool->element_size);
    if (err != 0) {
        return IX_ERROR_LOCAL(IX_CC_ERROR_OOM_SYSTEM,
            ("find_next_available: couldn't make more\n"
            "\ttried to make a new buffer, but was unable"));
    }

    *nump = pool->used_buffers;
    *offp = 0;
    pool->used_buffers++;
    return IX_SUCCESS;
}

static ix_error
verify_index(
    ix_genex_pool pool,
    ix_genex genex,
    ix_uint32 *nump,
    ix_uint32 *offp,
    unsigned short *availp,
    unsigned short *matchp
    )
{
    ix_uint32 buffer_num, buffer_offset;
    unsigned short avail, match;

    buffer_num = BUFFER_NUMBER(genex, pool->quantity_factor);
    buffer_offset = BUFFER_OFFSET(genex, pool->quantity_factor);
    if (buffer_num >= pool->used_buffers) {
        /*
         * If the buffer_num is out of range, we return immediately.  If the
         * caller didn't give the opportunity to report the results in an
         * output parameter, report it as an error.
         */
        if ((matchp == NULL) || (availp == NULL)) {
            return IX_ERROR_LOCAL(IX_CC_GENEX_ERROR_FINGERPRINT,
                ("verify_index: invalid genex\n"
                "\tgenex %08X for pool %p indicated using a buffer that\n"
                "\tdoesn't exist (0x%08X)", genex, pool, buffer_num));
        }
        match = 0;
        avail = 0;
    } else {
        if (pool->buffers[buffer_num] == NULL) {
            return IX_ERROR_LOCAL(IX_CC_ERROR_INTERNAL,
               ("verify_index: inconsistent internal data\n"
               "\tbuffer number %d for pool %p is NULL, but there should\n"
               "be %d valid ones", buffer_num, pool, pool->used_buffers));
        }
        avail = pool->buffers[buffer_num][buffer_offset].available;
        match = (genex >> IX_GENEX_INDEX_BITS == 
                 pool->buffers[buffer_num][buffer_offset].generation) ? 1 : 0;
    }

    /*
     * Assignments and return.
     */
    if (nump != NULL) {
        *nump = buffer_num;
    }
    if (offp != NULL) {
        *offp = buffer_offset;
    }
    if (availp != NULL) {
        *availp = avail;
    }
    if (matchp != NULL) {
        *matchp = match;
    }
    return IX_SUCCESS;
}

static ix_error
verify_genex(
    ix_genex_pool pool,
    ix_genex genex,
    ix_uint32 *nump,
    ix_uint32 *offp,
    unsigned short *availp
    )
{
    ix_uint32 num, off;
    unsigned short avail, match;
    IX_ERROR_CR(verify_index(pool, genex, &num, &off, &avail, &match));
    if (!match) {
        return IX_ERROR_LOCAL(IX_CC_GENEX_ERROR_FINGERPRINT,
            ("verify_genex: invalid fingerprint\n"
            "\t(stale handle?) ix_genex %08X in pool %p\n"
            "\thas a fingerprint of %04X",
            genex, pool, pool->buffers[num][off].generation));
    }
    if (nump != NULL) {
        *nump = num;
    }
    if (offp != NULL) {
        *offp = off;
    }
    if (availp != NULL) {
        *availp = avail;
    }
    return IX_SUCCESS;
}

/*
 * Bump the generation count but don't let useful bits ever be zero.
 */
static void
bump_generation(
    unsigned short *generationp
    )
{
    unsigned short generation = *generationp;

    generation++;
    if ((generation & IX_GENEX_MAX_GENERATION) == 0) {
        generation++;
    }

    *generationp = generation;
}

ix_error
ix_genex_init(
    ix_uint32 quantity_factor,
    ix_uint32 element_size,
    ix_genex_pool *poolp
    )
{
    ix_error err;
    ix_genex_pool pool;
    ix_uint32 ix;

    IX_ERROR_DCHECK_ARG_NULL(3, poolp);
    IX_ERROR_DCHECK_ARG_RANGE(1, quantity_factor, 1UL,
            IX_GENEX_INDEX_BITS - BITS_RESERVED_FOR_BUFFERS, "%lu");

    /*
     * The pool is the repository for all data relating to a genex.  Make one.
     */
    pool = (ix_genex_pool) ix_ossl_malloc(sizeof(struct ix_s_genex_pool));
    if (pool == NULL) {
        return IX_ERROR_LOCAL(IX_CC_ERROR_OOM_SYSTEM,
            ("Could not allocate %d bytes for an ix_genex_pool",
            sizeof(struct ix_s_genex_pool)));
    }

    /*
     * Store the user's requested element size so we can do easy indexing later.
     */
    pool->element_size = element_size;

    /*
     * Need to know how big each buffer is
     * a) when searching through them and
     * b) when making new buffers.
     */
    pool->quantity_factor = quantity_factor;

    /*
     * It is much easier to allocate new buffers and use them as needed than to
     * grow one buffer (would require copying large blocks of data).  Start with
     * one buffer and support for up to INITIAL_TOTAL_BUFFERS.  If the genex
     * should grow so much that we need more buffers than the initial total,
     * it is easy to make a new array.  Then we are only copying small blocks
     * of data pointing to the large blocks.
     */
    pool->total_buffers = INITIAL_TOTAL_BUFFERS;

    /*
     * Make room to hold the list of buffers.
     */
    pool->buffers = (ix_genex_genptr**)
        ix_ossl_malloc(sizeof(ix_genex_genptr*) * pool->total_buffers);
    if (pool->buffers == NULL) {
        err = IX_ERROR_LOCAL(IX_CC_ERROR_OOM_SYSTEM,
            ("Could not allocate %d bytes for the list of %d buffers",
            sizeof(ix_genex_genptr*) * pool->total_buffers,
            pool->total_buffers));
        goto clean_pool;
    }

    /*
     * Allocate just one buffer to begin with.
     */
    pool->used_buffers = 1;
    err = initialize_buffer(&pool->buffers[0],
        1 << pool->quantity_factor,
        pool->element_size);
    if (err != 0) {
        err = IX_ERROR_LOCAL(IX_CC_ERROR_OOM_SYSTEM,
            ("Could not initialize the first buffers"));
        goto clean_buffers;
    }
    pool->last_used = 0;

    /*
     * Initialize the other buffers to zero of course.
     */
    for (ix = 1; ix < pool->total_buffers; ix++) {
        pool->buffers[ix] = NULL;
    }

    *poolp = pool;
    return IX_SUCCESS;

clean_buffers:
    ix_ossl_free(pool->buffers);
clean_pool:
    ix_ossl_free(pool);
    return err;
}

ix_error
ix_genex_new(
    ix_genex_pool pool,
    ix_genex *genexp,
    void **voidp
    )
{
    ix_uint32 buffer_num;
    ix_uint32 buffer_offset;
    unsigned short avail;
    ix_error err;

    IX_ERROR_DCHECK_ARG_NULL(1, pool);
    IX_ERROR_DCHECK_ARG_NULL(2, genexp);

    if (pool->last_used != 0) {
        err = verify_index(pool, pool->last_used,
            &buffer_num, &buffer_offset, &avail, NULL);
        if (err != 0) {
            return IX_ERROR_CHAIN(err, 0, 0,
                ("ix_genex_new: last used (%08X) was invalid",
                 pool->last_used));
        }
    } else {
        /*
         * If the last_used is some invalid number, just say it's unavailable
         * and search for another.
         */
        avail = 0;
    }

    if (avail == 0) {
        err = find_next_available(pool, &buffer_num, &buffer_offset);
        if (err != 0) {
            return err;
        }
    }

    /*
     * Reserve the existing one for the current caller's use.
     */
    pool->buffers[buffer_num][buffer_offset].available = 0;
    if (voidp != NULL) {
        *voidp = pool->buffers[buffer_num][buffer_offset].user_data;
    }
    *genexp = PACK_GENEX(buffer_num, buffer_offset, pool->quantity_factor,
        pool->buffers[buffer_num][buffer_offset].generation);
    pool->last_used = *genexp;

    return IX_SUCCESS;
}

ix_error
ix_genex_lookup(
    ix_genex_pool pool,
    ix_genex genex,
    void **voidp
    )
{
    ix_error err;
    ix_uint32 buffer_num;
    ix_uint32 buffer_offset;
    unsigned short avail;

    IX_ERROR_DCHECK_ARG_NULL(1, pool);

    err = verify_genex(pool, genex, &buffer_num, &buffer_offset, &avail);
    if (err != 0) {
        return IX_ERROR_CHAIN(err, 0, 0,
            ("ix_genex_lookup: invalid genex %08X", genex));
    }

    if (avail != 0) {
        return IX_ERROR_LOCAL(IX_CC_GENEX_ERROR_FINGERPRINT,
            ("ix_genex_lookup: data inconsistency\n"
            "\tgenex %08X in pool %p is valid and available\n"
            "\twhen it should be just valid",
            genex, pool));
    }

    if (voidp != NULL) {
        *voidp = pool->buffers[buffer_num][buffer_offset].user_data;
    }
    return IX_SUCCESS;
}

ix_error
ix_genex_bump(
    ix_genex_pool pool,
    ix_genex old,
    ix_genex *newp
    )
{
    ix_error err;
    ix_uint32 buffer_num;
    ix_uint32 buffer_offset;
    unsigned short avail;

    IX_ERROR_DCHECK_ARG_NULL(1, pool);
    IX_ERROR_DCHECK_ARG_NULL(2, old);
    IX_ERROR_DCHECK_ARG_NULL(3, newp);

    err = verify_genex(pool, old, &buffer_num, &buffer_offset, &avail);
    if (err != 0) {
        return IX_ERROR_CHAIN(err, 0, 0,
            ("ix_genex_bump: invalid genex %08X", old));
    }

    bump_generation(&pool->buffers[buffer_num][buffer_offset].generation);
    *newp = PACK_GENEX(buffer_num, buffer_offset, pool->quantity_factor,
        pool->buffers[buffer_num][buffer_offset].generation);
    return IX_SUCCESS;
}

ix_error
ix_genex_exists(
    ix_genex_pool pool,
    ix_genex genex,
    int *existsp
    )
{
    unsigned short avail;
    unsigned short match;

    IX_ERROR_DCHECK_ARG_NULL(1, pool);
    IX_ERROR_DCHECK_ARG_ALIGN(1, pool, struct ix_s_genex_pool);
    IX_ERROR_DCHECK_ARG_NULL(3, existsp);
    IX_ERROR_DCHECK_ARG_ALIGN(3, existsp, int);

    IX_ERROR_CR(verify_index(pool, genex, NULL, NULL, &avail, &match));
    *existsp = (!avail && match);

    return IX_SUCCESS;
}

ix_error
ix_genex_new_iter(
    ix_genex_pool pool,
    ix_genex_iter *iterp
    )
{
    ix_genex_iter iter;

    IX_ERROR_DCHECK_ARG_NULL(1, pool);
    IX_ERROR_DCHECK_ARG_NULL(2, iterp);

    iter = (ix_genex_iter) ix_ossl_malloc(sizeof (struct ix_s_genex_iter));
    if (iter == NULL) {
        return IX_ERROR_LOCAL(IX_CC_ERROR_OOM_SYSTEM,
            ("Could not allocate %d bytes for an iterator",
            sizeof (struct ix_s_genex_iter)));
    }

    iter->pool              = pool;
    iter->last_reported     = IX_GENEX_NULL;
    iter->active            = 0;

    *iterp = iter;
    return IX_SUCCESS;
}

ix_error
ix_genex_get_first(
    ix_genex_iter iter,
    ix_genex *genexp,
    void **voidp
    )
{
    ix_error err;

    IX_ERROR_DCHECK_ARG_NULL(1, iter);

    iter->active            = 1;
    iter->last_reported     = IX_GENEX_NULL;

    err = ix_genex_get_next(iter, genexp, voidp);
    if (err != 0) {
        err = IX_ERROR_CHAIN(err, 0, 0,
            ("ix_genex_get_first: get next returned an error"));
    }

    return err;
}

ix_error
ix_genex_get_next(
    ix_genex_iter iter,
    ix_genex *genexp,
    void **voidp
    )
{
    ix_uint32 buffer_num;
    ix_uint32 buffer_offset;
    ix_uint32 quantity_factor;
    int found               = 0;
    ix_genex genex          = 0;
    void *user_data         = NULL;

    IX_ERROR_DCHECK_ARG_NULL(1, iter);

    if (iter->active == 0) {
        return IX_ERROR_WARNING(IX_CC_GENEX_ERROR_ITER,
            ("ix_genex_get_next: iterator %p is inactive\n"
            "\tUse ix_genex_get_first to activate it or\n"
            "\tix_genex_del_iter to free it.",
            iter));
    }

    quantity_factor = iter->pool->quantity_factor;
    buffer_num      = BUFFER_NUMBER(iter->last_reported, quantity_factor);
    buffer_offset   = BUFFER_OFFSET(iter->last_reported, quantity_factor);

    /*
     * If last_reported is IX_GENEX_NULL, it is not really the last one
     * reported, it is the beginning of the list.  Otherwise, increment one
     * past the last one reported to find the next to report.
     */
    if (iter->last_reported != IX_GENEX_NULL) {
        buffer_offset++;
    }

    for (; buffer_num < iter->pool->used_buffers; buffer_num++) {
        for (; buffer_offset < (ix_uint32) (1 << quantity_factor);
            buffer_offset++) {
            ix_genex_genptr *ptr =
                &iter->pool->buffers[buffer_num][buffer_offset];
            if (ptr->available == 0) {
                found = 1;
                genex = PACK_GENEX(buffer_num, buffer_offset,
                    quantity_factor, ptr->generation);
                user_data = ptr->user_data;
                break;
            }
        }
        if (found != 0) {
            break;
        }
        buffer_offset = 0;
    }

    if (found == 0) {
        iter->active = 0;
    }

    iter->last_reported = genex;

    if (genexp != NULL) {
        *genexp = genex;
    }

    if (voidp != NULL) {
        *voidp = user_data;
    }
    return IX_SUCCESS;
}

ix_error
ix_genex_del_iter(
    ix_genex_iter iter
    )
{
    IX_ERROR_DCHECK_ARG_NULL(1, iter);

    iter->pool              = NULL;
    iter->last_reported     = IX_GENEX_NULL;
    iter->active            = 0;
    ix_ossl_free(iter);

    return IX_SUCCESS;
}

ix_error
ix_genex_del(
    ix_genex_pool pool,
    ix_genex genex
    )
{
    ix_uint32 buffer_num;
    ix_uint32 buffer_offset;
    unsigned short avail;

    IX_ERROR_DCHECK_ARG_NULL(1, pool);

    IX_ERROR_CRT(
        verify_genex(pool, genex, &buffer_num, &buffer_offset, &avail),
        IX_CC_ERROR_INTERNAL, 0);

    if (avail != 0) {
        return IX_ERROR_LOCAL(IX_CC_GENEX_ERROR_PARAM,
            ("ix_genex_del: genex is available\n"
            "\tsince the genex %08X was already available,\n"
            "\tit cannot be deleted", genex));
    }

    pool->buffers[buffer_num][buffer_offset].available = 1;
    bump_generation(&pool->buffers[buffer_num][buffer_offset].generation);
    return IX_SUCCESS;
}

ix_error
ix_genex_fini(
    ix_genex_pool pool
    )
{
    ix_uint32 ix;

    IX_ERROR_DCHECK_ARG_NULL(1, pool);

#ifdef IX_GENEX_STRICT_CLEANUP
    for (ix = 0; ix < pool->used_buffers; ix++) {
        int j;

        for (j = 0; j < (1 << pool->quantity_factor); j++) {
            if ((pool->buffers[ix] != NULL) &&
                (pool->buffers[ix][j].available == 0)) {
                return IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
                    ("ix_genex_fini: element is still allocated\n"
                    "\tElement in buffer #%d, index %d in pool %p\n"
                    "\t(genex 0x%08X) is still allocated. This could\n"
                    "\tindicate a resource leak.",
                    ix, j, pool, pool->buffers[ix][j].generation));
            }
        }
    }
#endif /* IX_GENEX_STRICT_CLEANUP */

    for (ix = 0; ix < pool->used_buffers; ix++) {
        if (pool->buffers[ix] == NULL) {
            ix_ossl_message_log(
                "ix_genex_fini: buffer number %ld should be valid, "
                "but is NULL\n", ix);
        }
        if (pool->buffers[ix]->user_data == NULL) {
            ix_ossl_message_log(
                "ix_genex_fini: first pointer to user data is NULL\n");
        }
        ix_ossl_free(pool->buffers[ix]->user_data);
        ix_ossl_free(pool->buffers[ix]);
        pool->buffers[ix] = NULL;
    }

    ix_ossl_free(pool->buffers);
    pool->buffers = NULL;
    ix_ossl_free(pool);
    return IX_SUCCESS;
}

