/*
 * ===========================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      genexts.c
 *
 * = DESCRIPTION
 *      Implementation of the threadsafe extensions to the generation/index
 *      storage mechanism.
 *
 * ===========================================================================
 */

#define IX_ERROR_FILE_IDENT                                             \
        "$Id: genexts.c,v 1.3 2003/04/04 00:48:37 ageorges Exp $"

#include "ix_cc_error.h"
#include "ix_ossl.h"
#include "cc/internal/internal_genex.h"
#include "cc/ix_cc_genexts.h"

struct ix_s_genex_ts_pool {
    ix_genex_pool genex_pool;
    ix_ossl_mutex_t mutex;
};

struct ix_s_genex_ts_iter {
    ix_genex_iter genex_iter;
    ix_genex_ts_pool genex_pool;
};

#ifdef DEBUG
#define STRICT_LOCKS
#endif

#ifdef STRICT_LOCKS
#define VERIFY_LOCK(__p, __f)                                           \
    do {                                                                \
        int __b;                                                        \
        ix_error __e;                                                   \
        __e = ix_genex_ts_trylock(__p, &__b);                           \
        if (__e != 0) {                                                 \
            return __e;                                                 \
        }                                                               \
        if (__b == 0) {                                                 \
            __e = ix_genex_ts_unlock(__p);                              \
            return IX_ERROR_CHAIN(__e, IX_CC_ERROR_INTERNAL, 0,         \
                ("mutex was not locked!\n"                              \
                "\tThe mutex for pool %p should be locked.",            \
                __p));                                                  \
        }                                                               \
    } while (0);
#else
#define VERIFY_LOCK(__p, __f)
#endif /* STRICT_LOCKS */

ix_error
ix_genex_ts_init(
    unsigned int quantity_factor,
    unsigned int element_size,
    ix_genex_ts_pool *poolp
    )
{
    ix_error err;
    ix_genex_ts_pool pool;

    IX_ERROR_DCHECK_ARG_NULL(3, poolp);

    pool = (ix_genex_ts_pool) ix_ossl_malloc(sizeof(struct ix_s_genex_ts_pool));
    if (pool == NULL) {
        return IX_ERROR_LOCAL(IX_CC_ERROR_OOM_SYSTEM,
            ("Could not allocate %d bytes for an ix_genex_ts_pool",
            sizeof(struct ix_s_genex_ts_pool)));
    }

    IX_ERROR_CG(
        ix_genex_init(quantity_factor, element_size, &pool->genex_pool),
        err, clean_memory);

    IX_ERROR_CG(
        ix_ossl_mutex_init(IX_OSSL_MUTEX_UNLOCK, &pool->mutex),
        err, clean_pool);

    *poolp = pool;

    return IX_SUCCESS;

clean_pool:
    (void) ix_genex_fini((ix_genex_pool) pool);
    return err;

clean_memory:
    ix_ossl_free(pool);
    return err;
}

ix_error
ix_genex_ts_new(
    ix_genex_ts_pool pool,
    ix_genex *genexp,
    void **voidp
    )
{
    IX_ERROR_DCHECK_ARG_NULL(1, pool);
    VERIFY_LOCK(pool, ix_genex_ts_new);

    return ix_genex_new(pool->genex_pool, genexp, voidp);
}

ix_error
ix_genex_ts_lookup(
    ix_genex_ts_pool pool,
    ix_genex genex,
    void **voidp
    )
{
    IX_ERROR_DCHECK_ARG_NULL(1, pool);
    VERIFY_LOCK(pool, ix_genex_ts_lookup);

    return ix_genex_lookup(pool->genex_pool, genex, voidp);
}

ix_error
ix_genex_ts_bump(
    ix_genex_ts_pool pool,
    ix_genex old,
    ix_genex *newp
    )
{
    IX_ERROR_DCHECK_ARG_NULL(1, pool);
    VERIFY_LOCK(pool, ix_genex_ts_bump);

    return ix_genex_bump(pool->genex_pool, old, newp);
}

ix_error
ix_genex_ts_exists(
    ix_genex_ts_pool pool,
    ix_genex genex,
    int *existsp
    )
{
    IX_ERROR_DCHECK_ARG_NULL(1, pool);
    IX_ERROR_DCHECK_ARG_ALIGN(1, pool, 
                              struct ix_s_genex_ts_pool);
    IX_ERROR_DCHECK_ARG_NULL(3, pool);
    IX_ERROR_DCHECK_ARG_ALIGN(3, pool, int);
    VERIFY_LOCK(pool, ix_genex_ts_exists);

    IX_ERROR_CR(ix_genex_exists(pool->genex_pool, genex, existsp));
    return IX_SUCCESS;
}

ix_error
ix_genex_ts_new_iter(
    ix_genex_ts_pool pool,
    ix_genex_ts_iter *iterp
    )
{
    ix_genex_ts_iter iter;
    ix_error err;

    IX_ERROR_DCHECK_ARG_NULL(1, pool);
    IX_ERROR_DCHECK_ARG_NULL(2, iterp);
    VERIFY_LOCK(pool, ix_genex_ts_new_iter);

    iter = (ix_genex_ts_iter) ix_ossl_malloc(sizeof(struct ix_s_genex_ts_iter));
    if (iter == NULL) {
        return IX_ERROR_LOCAL(IX_CC_ERROR_OOM_SYSTEM,
            ("Could not allocate %d bytes for ix_genex_ts_iter",
            sizeof(struct ix_s_genex_ts_iter)));
    }

    err = ix_genex_new_iter(pool->genex_pool, &iter->genex_iter);
    if (err != 0) {
        goto clean_iter;
    }

    iter->genex_pool = pool;

    *iterp = iter;
    return IX_SUCCESS;

clean_iter:
    ix_ossl_free(iter);
    return err;
}

ix_error
ix_genex_ts_get_first(
    ix_genex_ts_iter iter,
    ix_genex *genexp,
    void **voidp
    )
{
    IX_ERROR_DCHECK_ARG_NULL(1, iter);
    VERIFY_LOCK(iter->genex_pool, ix_genex_ts_get_first);

    return ix_genex_get_first(iter->genex_iter, genexp, voidp);
}

ix_error
ix_genex_ts_get_next(
    ix_genex_ts_iter iter,
    ix_genex *genexp,
    void **voidp
    )
{
    IX_ERROR_DCHECK_ARG_NULL(1, iter);
    VERIFY_LOCK(iter->genex_pool, ix_genex_ts_get_next);

    return ix_genex_get_next(iter->genex_iter, genexp, voidp);
}

ix_error
ix_genex_ts_del_iter(
    ix_genex_ts_iter iter
    )
{
    ix_error err;

    IX_ERROR_DCHECK_ARG_NULL(1, iter);
    VERIFY_LOCK(iter->genex_pool, ix_genex_ts_get_next);

    err = ix_genex_del_iter(iter->genex_iter);
    if (err != 0) {
        return err;
    }

    ix_ossl_free(iter);
    return IX_SUCCESS;
}

ix_error
ix_genex_ts_del(
    ix_genex_ts_pool pool,
    ix_genex genex
    )
{
    IX_ERROR_DCHECK_ARG_NULL(1, pool);
    VERIFY_LOCK(pool, ix_genex_ts_get_next);

    return ix_genex_del(pool->genex_pool, genex);
}

ix_error
ix_genex_ts_lock(
    ix_genex_ts_pool pool
    )
{
    IX_ERROR_DCHECK_ARG_NULL(1, pool);

    IX_ERROR_CRT(ix_ossl_mutex_lock(pool->mutex, IX_OSSL_WAIT_FOREVER),
        IX_CC_ERROR_INTERNAL, 0);
    return IX_SUCCESS;
}

ix_error
ix_genex_ts_trylock(
    ix_genex_ts_pool pool,
    int *busyp
    )
{
    ix_error err;
    IX_ERROR_DCHECK_ARG_NULL(1, pool);
    IX_ERROR_DCHECK_ARG_NULL(2, busyp);

    err = ix_ossl_mutex_lock(pool->mutex, IX_OSSL_WAIT_NONE);
    if (err == IX_SUCCESS)
    {
        *busyp = 0;
    }
    else
    {
        if (IX_ERROR_GET_CODE(err) != IX_OSSL_ERROR_MUTEX_LOCK_FAILED)
        {
            return IX_ERROR_CHAIN(err, IX_CC_ERROR_INTERNAL, 0,
                ("Could not get the mutex"));
        }
        else
        {
            *busyp = 1;
        }
    }
    return IX_SUCCESS;
}

ix_error
ix_genex_ts_unlock(
    ix_genex_ts_pool pool
    )
{
    IX_ERROR_DCHECK_ARG_NULL(1, pool);

    IX_ERROR_CRT(ix_ossl_mutex_unlock(pool->mutex),
        IX_CC_ERROR_INTERNAL, 0);

    return IX_SUCCESS;
}

ix_error
ix_genex_ts_fini(
    ix_genex_ts_pool pool
    )
{
    ix_error err;

    IX_ERROR_DCHECK_ARG_NULL(1, pool);
    VERIFY_LOCK(pool, ix_genex_ts_get_next);

    err =  ix_genex_fini(pool->genex_pool);
    if (err != 0) {
        return err;
    }

    /*
     * Can't return an error, just have to ignore it.
     */
    (void) ix_ossl_mutex_fini(pool->mutex);
    ix_ossl_free(pool);

    return IX_SUCCESS;
}

