/**
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *    Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 */

/**
 * @file hash_table128.c
 *
 * Implementation of the ix_cc_hash_table128 module.
 */

#define IX_ERROR_FILE_IDENT "$Id: hash_table128.c,v 1.5 2003/08/19 23:12:05 ktseng Exp $"

#include "ix_cc_error.h"
#include "ix_ossl.h"
#include "ix_cc_macros.h"
#include "cc/ix_cc_hash_table128.h"
#include "cc/internal/internal_sig.h"
#include "cc/internal/internal_hash_table128.h"

/**
 * Unique signature for the 128-bit Hash Table.
 * {1C8D3310-03B4-408c-9208-DCAB1274FF2B}
 */
_IX_CC_SIG_GLOBAL(_ix_cc_hash_table128_sig,
    0x1c8d3310, 0x3b4, 0x408c, 0x92, 0x8, 0xdc, 0xab, 0x12, 0x74, 0xff, 0x2b);

ix_error ix_cc_hash_table128_init(
    ix_uint32           arg_NumBuckets,
    ix_uint32           arg_NumElements,
    ix_cc_hash_table_algo arg_Algo,
    ix_cc_hash_table128 *arg_phHashTable)
{
    ix_cc_hash_table128 ht;
    ix_uint16 *pBuckets;
    _ix_cc_hash_table128_link *pLinks;
    unsigned int dataBlockSize;
    ix_error err;

    IX_ERROR_CHECK_ARG_MIN(1, arg_NumBuckets, 1UL, "%ld");
    IX_ERROR_CHECK_ARG_RANGE(2, arg_NumElements, 1UL,
            (ix_uint32) _IX_CC_HASH_TABLE_INVALID_ELEMENT-1, "%ld");
    IX_ERROR_CHECK_ARG_MAX(3, arg_Algo, IX_CC_HASH_TABLE_ALGO_LAST-1, "%d");
    IX_ERROR_DCHECK_ARG_NULL(4, arg_phHashTable);
    IX_ERROR_DCHECK_ARG_ALIGN(4, arg_phHashTable, ix_cc_hash_table128);

    /*
     * Calculate how much memory needs to be allocated.
     */
    dataBlockSize = 
        sizeof(struct _ix_s_cc_hash_table128) +
        sizeof(_ix_cc_hash_table128_link) * arg_NumElements +
        sizeof(ix_uint16) * arg_NumBuckets;

    /*
     * Allocate the memory.
     */
    ht = (ix_cc_hash_table128) ix_ossl_malloc(dataBlockSize);
    if (ht == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
            ("Could not allocate %d bytes from the OSSL", dataBlockSize));
    }

    /*
     * Start the linked list elements immediately after the initial structure.
     */
    pLinks = (_ix_cc_hash_table128_link*) &ht[1];

    /*
     * Start the buckets immediately after the linked list elements.
     */
    pBuckets = (ix_uint16*) &pLinks[arg_NumElements];

    /*
     * Initialize the data associated with the handle.
     */
    _IX_CC_SIG_COPY(ht->sig, _ix_cc_hash_table128_sig);
    ht->pBuckets        = pBuckets;
    ht->pLinks          = pLinks;
    ht->numBuckets      = arg_NumBuckets;
    ht->algo            = arg_Algo;
    ht->linkFreeList    = 0;
    ht->numElements     = arg_NumElements;
    ht->magic           = 0;

    /*
     * Use ix_cc_hash_table128_purge() to initialize the structures.
     */
    IX_ERROR_CG(ix_cc_hash_table128_purge(ht), err, clean_mem);

    if (arg_Algo == IX_CC_HASH_TABLE_ALGO_MAGIC)
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ht->magic = (arg_NumBuckets * _IX_CC_HASH_TABLE_MAGIC) / 1000;
#else
        ht->magic       = _IX_CC_HASH_TABLE_MAGIC * arg_NumBuckets;
#endif /* end (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */
    }

    /*
     * Assign the output parameter.
     */
    *arg_phHashTable = ht;
    return IX_SUCCESS;

clean_mem:
    ix_ossl_free(ht);
    return err;
}

ix_error ix_cc_hash_table128_fini(
    ix_cc_hash_table128 arg_hHashTable)
{
    _IX_CC_HASH_TABLE128_CHECK_HANDLE(arg_hHashTable);

#if defined(OPTIONAL)
    /*
     * Search through the contents looking for any entries.  If some are found,
     * return an error. This could potentially prevent memory leaks, depending
     * on how the hash table is used.
     */
    {
        unsigned int i;
        for (i = 0; i < arg_hHashTable->numBuckets; i++)
        {
            if (arg_hHashTable->pBuckets[i] != _IX_CC_HASH_TABLE_INVALID_ELEMENT)
            {
                return IX_ERROR_WARNING(IX_CC_ERROR_INTERNAL,
                    ("One or more keys were not removed before calling fini:\n"
                     "\tData = %p",
                     arg_hHashTable->pLinks[arg_hHashTable->pBuckets[i]].pData));
            }
        }
    }
#endif /* OPTIONAL */

    /*
     * Memset clears the signature, which is required in order to detect that
     * a handle is no longer valid.
     */
    ix_ossl_memset(arg_hHashTable, 0, sizeof(struct _ix_s_cc_hash_table128));

    /*
     * Freeing this one pointer deallocates all the memory because it was all
     * allocated with a single call.
     */
    ix_ossl_free(arg_hHashTable);
    return IX_SUCCESS;
}

/*
 * NAME: _ix_cc_hash_table128_hash
 *
 * DESCRIPTION: Hashes the provided key according to the chosen algorithm and
 * returns the bucket for that key.
 *
 * @Param:      - IN    arg_hHashTable
 *                      handle to the Hash Table, as returned by
 *                      ix_cc_hash_table128_init()
 *
 * @Param:      - IN    arg_Key
 *                      value associated with some data
 *
 * @Return: Index of the bucket for the key.
 */
static ix_uint32 _ix_cc_hash_table128_hash(
    ix_cc_hash_table128 arg_hHashTable,
    ix_uint128          arg_Key)
{
    ix_uint32 bucketIndex;
    ix_uint32 key32 = _IX_CC_HASH_TABLE128_MAKE_KEY32(arg_Key);

    switch (arg_hHashTable->algo)
    {
        case IX_CC_HASH_TABLE_ALGO_MAGIC:
            bucketIndex = key32 * arg_hHashTable->magic;
            break;

        case IX_CC_HASH_TABLE_ALGO_XOR8:
            bucketIndex = 
                ((key32 >> 24) & 0xFF) ^
                ((key32 >> 16) & 0xFF) ^
                ((key32 >>  8) & 0xFF) ^
                ((key32      ) & 0xFF);
            break;

        case IX_CC_HASH_TABLE_ALGO_XOR16:
            bucketIndex = 
                ((key32 >> 16) & 0xFFFF) ^
                ((key32      ) & 0xFFFF);
            break;


        case IX_CC_HASH_TABLE_ALGO_NONE:
        /*
         * The 'default' case should be impossible.
         */
        default:
            bucketIndex = key32;
            break;
    }
    return bucketIndex % arg_hHashTable->numBuckets;
}

/*
 * NAME: _ix_cc_hash_table128_find
 *
 * DESCRIPTION: This function is intended to be used to either:
 * 1) Find the requested key, or
 * 2) Find where the key should be inserted.
 *
 * If the requested key is in the list, this function returns a pointer to
 * the link for that key.  If the key does not exist, returns a pointer to the
 * link before the location where the link would be.  This makes an
 * insertion easier (only one traversal is required).  However, the caller
 * must evaluate the returned item:
 * 1) If the key was found, the result is non-NULL and points to a
 *    link with key == arg_Key.
 * 2) If the link was not found, the result indicates where the key would
 *    be inserted, and is either:
 *    a) NULL (insert the key as the first element in the list), or
 *    b) the element that should refer to the key.
 *
 * Note that the links for each bucket are sorted numerically by key value.
 *
 * @Param:      - IN    arg_hHashTable
 *                      handle to the Hash Table, as returned by
 *                      ix_cc_hash_table128_init()
 *
 * @Param:      - IN    arg_BucketIndex
 *                      Bucket to search
 *
 * @Param:      - IN    arg_Key
 *                      key to search for
 *
 * @Return: pointer to the link
 */
static _ix_cc_hash_table128_link *_ix_cc_hash_table128_find(
    ix_cc_hash_table128 arg_hHashTable,
    ix_uint16           arg_BucketIndex,
    ix_uint128          arg_Key)
{
    ix_uint16 elementIndex = arg_hHashTable->pBuckets[arg_BucketIndex];
    _ix_cc_hash_table128_link *pLink = NULL;
    _ix_cc_hash_table128_link *pLinks = arg_hHashTable->pLinks;

    /*
     * Walk down the linked list until:
     * 1) The link is found; or
     * 2) We reach the end of the list;
     */
    while (elementIndex != _IX_CC_HASH_TABLE_INVALID_ELEMENT)
    {
        pLink = &pLinks[elementIndex];
        if (IX_128_BIT_BINARY_OPERATION_EQ(pLink->key, arg_Key))
        {
            break;
        }
        else
        {
            elementIndex = pLink->next;
        }
    }

    return pLink;
}

ix_error ix_cc_hash_table128_add_entry(
    ix_cc_hash_table128 arg_hHashTable,
    ix_uint128          arg_Key,
    void                *arg_pData)
{
    ix_uint16 bucketIndex;
    _ix_cc_hash_table128_link *pLink;
    ix_error err = IX_SUCCESS;

    _IX_CC_HASH_TABLE128_CHECK_HANDLE(arg_hHashTable);

    bucketIndex = _ix_cc_hash_table128_hash(arg_hHashTable, arg_Key);

    /*
     * Search the bucketIndex to see if the key already exists.
     */
    pLink = _ix_cc_hash_table128_find(arg_hHashTable, bucketIndex, arg_Key);

    if ((pLink == NULL) || (IX_128_BIT_BINARY_OPERATION_NEQ(pLink->key, arg_Key)))
    {
        ix_uint16 elementIndex;
        _ix_cc_hash_table128_link *pLinks  = arg_hHashTable->pLinks;

        /*
         * Make sure there's at least one free element.
         */
        elementIndex = arg_hHashTable->linkFreeList;
        if (elementIndex != _IX_CC_HASH_TABLE_INVALID_ELEMENT)
        {
            /*
             * Remove the element from the free list.
             */
            arg_hHashTable->linkFreeList = pLinks[elementIndex].next;

            pLinks[elementIndex].key    = arg_Key;
            pLinks[elementIndex].pData  = arg_pData;

            /*
             * Insert the new element in the list, sorted by key.  If pLink
             * is NULL, insert it as the first element in the bucket.
             * Otherwise, put it after pLink.
             */
            if (pLink == NULL)
            {
                ix_uint16 *pBuckets = arg_hHashTable->pBuckets;
                ix_uint16 oldTop = pBuckets[bucketIndex];
                pLinks[elementIndex].next = oldTop;
                pLinks[elementIndex].prev =
                    _IX_CC_HASH_TABLE_INVALID_ELEMENT;
                pBuckets[bucketIndex] = elementIndex;

                /*
                 * If the bucket already had elements, back-link the old
                 * top of the list to the new top of the list.
                 */
                if (oldTop != _IX_CC_HASH_TABLE_INVALID_ELEMENT)
                {
                    pLinks[oldTop].prev = elementIndex;
                }
            }
            else
            {
                /*
                 * The link belongs in the middle of the list.  Insert it.
                 */
                pLinks[elementIndex].next = pLink->next;

                /*
                 * The value pLink - pLinks is the index of the pLink element
                 * into the pLinks array.
                 */
                pLinks[elementIndex].prev = pLink - pLinks;
                if (pLink->next != _IX_CC_HASH_TABLE_INVALID_ELEMENT)
                {
                    pLinks[pLink->next].prev = elementIndex;
                }
                pLink->next = elementIndex;
            }
        }
        else
        {
            /*
             * The free list was empty.
             */
            err = IX_ERROR_WARNING(IX_CC_ERROR_FULL,
                ("No more available hash table elements"));
        }
    }
    else
    {
        /*
         * Return an error stating whether the new entry is a duplicate or
         * conflicts with the original.
         */
        if (pLink->pData == arg_pData)
        {
            err = IX_ERROR_WARNING(IX_CC_ERROR_DUPLICATE_ENTRY,
                ("A duplicate entry has the data %p", pLink->pData));
        }
        else
        {
            err = IX_ERROR_WARNING(IX_CC_ERROR_CONFLICTING_ENTRY,
                ("A conflicting entry has the data %p", pLink->pData));
        }
    }

    return err;
}

ix_error ix_cc_hash_table128_delete_entry(
    ix_cc_hash_table128 arg_hHashTable,
    ix_uint128          arg_Key)
{
    ix_uint16 bucketIndex;
    _ix_cc_hash_table128_link *pLink;

    _IX_CC_HASH_TABLE128_CHECK_HANDLE(arg_hHashTable);

    bucketIndex = _ix_cc_hash_table128_hash(arg_hHashTable, arg_Key);

    /*
     * Attempt to find the element.
     */
    pLink = _ix_cc_hash_table128_find(arg_hHashTable, bucketIndex, arg_Key);
    if ((pLink == NULL) || (IX_128_BIT_BINARY_OPERATION_NEQ(pLink->key, arg_Key)))
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
            ("Could not find an element with the key 0x%08lX", arg_Key));
    }

    /*
     * Unlink the element.
     */
    if (pLink->prev == _IX_CC_HASH_TABLE_INVALID_ELEMENT)
    {
        arg_hHashTable->pBuckets[bucketIndex] = pLink->next;
    }
    else
    {
        arg_hHashTable->pLinks[pLink->prev].next = pLink->next;
    }

    if (pLink->next != _IX_CC_HASH_TABLE_INVALID_ELEMENT)
    {
        arg_hHashTable->pLinks[pLink->next].prev = pLink->prev;
    }

    /*
     * Invalidate the element's contents.
     */
    IX_CC_OPTIONAL(_IX_CC_HASH_TABLE128_CLEAR_KEY(pLink->key));
    IX_CC_OPTIONAL(pLink->pData = NULL);

    /*
     * Add the element to the top of the free list.
     */
    pLink->next = arg_hHashTable->linkFreeList;
    IX_CC_OPTIONAL(pLink->prev = _IX_CC_HASH_TABLE_INVALID_ELEMENT);
    arg_hHashTable->linkFreeList = pLink - arg_hHashTable->pLinks;

    return IX_SUCCESS;
}

ix_error ix_cc_hash_table128_lookup(
    ix_cc_hash_table128 arg_hHashTable,
    ix_uint128          arg_Key,
    int                *arg_pExists,
    void              **arg_ppData)
{
    ix_uint16 bucketIndex;
    _ix_cc_hash_table128_link *pLink;

    _IX_CC_HASH_TABLE128_CHECK_HANDLE(arg_hHashTable);
    IX_ERROR_DCHECK_ARG_ALIGN(3, arg_pExists, int);
    IX_ERROR_DCHECK_ARG_ALIGN(4, arg_ppData, void*);

    bucketIndex = _ix_cc_hash_table128_hash(arg_hHashTable, arg_Key);

    /*
     * Attempt to find the element.
     */
    pLink = _ix_cc_hash_table128_find(arg_hHashTable, bucketIndex, arg_Key);
    if ((pLink == NULL) || (IX_128_BIT_BINARY_OPERATION_NEQ(pLink->key, arg_Key)))
    {
        if (arg_pExists != NULL)
        {
            *arg_pExists = 0;
            return IX_SUCCESS;
        }

        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
            ("Could not find an element with the key 0x%08lX", arg_Key));
    }

    if (arg_pExists != NULL)
    {
        *arg_pExists = 1;
    }

    if (arg_ppData != NULL)
    {
        *arg_ppData = pLink->pData;
    }

    return IX_SUCCESS;
}

ix_error ix_cc_hash_table128_get_first_entry(
    ix_cc_hash_table128 arg_hHashTable,
    ix_uint128          *arg_pKey,
    void                **arg_ppData)
{
    ix_uint16 bucketIndex;

    _IX_CC_HASH_TABLE128_CHECK_HANDLE(arg_hHashTable);

    if (arg_pKey != NULL)
    {
        IX_ERROR_DCHECK_ARG_ALIGN(2, arg_pKey, ix_uint128);
    }
    if (arg_ppData != NULL)
    {
        IX_ERROR_DCHECK_ARG_ALIGN(3, arg_ppData, void*);
    }

    for (bucketIndex = 0;
        bucketIndex < arg_hHashTable->numBuckets;
        bucketIndex++)
    {
        ix_uint16 elementIndex = arg_hHashTable->pBuckets[bucketIndex];
        if (elementIndex != _IX_CC_HASH_TABLE_INVALID_ELEMENT)
        {
            _ix_cc_hash_table128_link *pLink =
                &arg_hHashTable->pLinks[elementIndex];

            if (arg_pKey != NULL)
            {
                *arg_pKey = pLink->key;
            }

            if (arg_ppData != NULL)
            {
                *arg_ppData = pLink->pData;
            }

            return IX_SUCCESS;
        }
    }

    /*
     * Use IX_ERROR_NEW() rather than IX_ERROR_WARNING() because it is
     * lower overhead. This is a rather normal condition, so low overhead
     * is good.
     */
    return IX_ERROR_NEW(IX_CC_ERROR_ENTRY_NOT_FOUND, IX_ERROR_LEVEL_WARNING);
}

ix_error ix_cc_hash_table128_purge(
    ix_cc_hash_table128 arg_hHashTable)
{
    _ix_cc_hash_table128_link *pLinks;
    unsigned int i;

    _IX_CC_HASH_TABLE128_CHECK_HANDLE(arg_hHashTable);

    pLinks = arg_hHashTable->pLinks;

    /*
     * Create a linked list of unallocated elements.  When in the free list,
     * the prev field is unused.
     */
    for (i = 0; i < arg_hHashTable->numElements; i++)
    {
        pLinks[i].next  = i + 1;
        IX_CC_OPTIONAL(
            pLinks[i].key.aUint32[0]    = 0;
            pLinks[i].key.aUint32[1]    = 0;
            pLinks[i].key.aUint32[2]    = 0;
            pLinks[i].key.aUint32[3]    = 0;
            pLinks[i].prev              = _IX_CC_HASH_TABLE_INVALID_ELEMENT;
            pLinks[i].pData             = NULL;
        )
    }
    /*
     * The last iteration of the loop assigns the next field to an undefined
     * element of the array.  Set it to the 'invalid' flag.
     */
    pLinks[i-1].next      = _IX_CC_HASH_TABLE_INVALID_ELEMENT;

    /*
     * Initialize the buckets by setting them all to an invalid state
     * (_IX_CC_HASH_TABLE_INVALID_ELEMENT).
     */
    for (i = 0; i < arg_hHashTable->numBuckets; i++)
    {
        arg_hHashTable->pBuckets[i] = _IX_CC_HASH_TABLE_INVALID_ELEMENT;
    }

    arg_hHashTable->linkFreeList    = 0;

    return IX_SUCCESS;
}
