/**
 * ===========================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      index_pool.c
 *
 * = DESCRIPTION
 *      Source file for the Index Pool.  The pool allows the caller to
 *      allocate and free a number of indexes.  The index pool does not
 *      specifically provide its own thread safety; a single handle should not
 *      be used from multiple threads.  However, it is acceptable for several
 *      threads to each create and use their own index pools without
 *      interfering with other thread pools.
 *
 * ===========================================================================
 * $Id: index_pool.c,v 1.8 2003/10/31 17:41:00 ktseng Exp $
 */

#include "cc/ix_cc_index_pool.h"
#include "ix_ossl.h"
#include "ix_cc_macros.h"

PRIVATE ix_error _ix_cc_index_pool_get(
    ix_cc_index_pool    arg_hIndexPool,
    ix_uint32           arg_StartPos,
    ix_uint32          *arg_pIndex);

/*
 * NAME: ix_s_cc_index_pool
 *
 * DESCRIPTION: Structure containing the information required by the
 * implementation of the index_pool.
 */
struct ix_s_cc_index_pool
{
    /*
     * pArray is a pointer to the array of elements used to track which
     * indexes have been allocated.
     */
    ix_uint32   *pArray;

    /*
     * numElements is the number of indexes managed by this instance of the
     * index_pool.
     */
    ix_uint32   numElements;

    /*
     * arraySize is the number of 32-bit elements used to hold the numElements
     * number of index allocation flags.
     */
    ix_uint32   arraySize;

    /*
     * searchHint stores a hint for where to next search for an empty element.
     * This helps prevent looping over a large sections of the array each time
     * a new index is allocated.
     */
    ix_uint32   searchHint;

    /*
     * offset is an adjustment value, allowing the caller to prevent certain
     * indexes from being allocated.  This value is only meaningful to the
     * client, and does not affect how the indexes are managed within the
     * index pool.
     */
    ix_uint32   offset;
};

/*
 * NAME: _IX_BITS_PER_ARRAY_ELEMENT
 *
 * DESCRIPTION: The number of bits in each element of the array
 */
#define _IX_BITS_PER_ARRAY_ELEMENT (sizeof(ix_uint32) * 8)

/*
 * NAME: _IX_GET_ELEMENT_INDEX
 *
 * DESCRIPTION: Returns the offset into the array of the indicated index.
 *
 * @Param:      - IN    arg_InternalIndex
 */
#define _IX_GET_ELEMENT_INDEX(arg_InternalIndex)                        \
        IX_GET_BIT_FIELD32(arg_InternalIndex, 5, 31)

/*
 * NAME: _IX_GET_BIT_OF_ELEMENT
 *
 * DESCRIPTION: Returns the bit position of the indicated index.
 *
 * @Param:      - IN    arg_InternalIndex
 */
#define _IX_GET_BIT_OF_ELEMENT(arg_InternalIndex)                       \
        IX_GET_BIT_FIELD32(arg_InternalIndex, 0, 4)

/*
 * NAME: ix_cc_index_pool_init
 *
 * DESCRIPTION: Creates an index pool with the requested number of elements.
 *
 * @Param:      - IN    arg_NumElements
 *                      Number of indexes the ix_cc_index_pool should manage.
 *                      This number is the quantity beyond arg_Offset.
 *
 * @Param:      - IN    arg_Offset 
 *                      Number of reserved indexes; e.g. if arg_Offset == 2,
 *                      ix_cc_index_pool_alloc() would never return an index
 *                      of zero or one.
 *
 * @Param:      - OUT   arg_phIndexPool
 *                      Handle that should be used in future calls to the 
 *                      ix_cc_index_pool module.
 *
 * @Returns:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_OOM_SYSTEM
 */
ix_error ix_cc_index_pool_init(
    ix_uint32           arg_NumElements,
    ix_uint32           arg_Offset,
    ix_cc_index_pool   *arg_phIndexPool)
{
    ix_uint32 memSize;
    ix_uint32 arraySize;
    ix_cc_index_pool hIndexPool;
    int numUsedInLastElement;

    IX_ERROR_DCHECK_ARG_MIN(1, arg_NumElements, 1UL, "%ld");
    IX_ERROR_DCHECK_ARG_NULL(3, arg_phIndexPool);
    IX_ERROR_DCHECK_ARG_ALIGN(3, arg_phIndexPool, ix_cc_index_pool);

    /*
     * Calculate the number of array elements required to manage the requested
     * number of indexes (arg_NumElements)
     */
    arraySize = (arg_NumElements + _IX_BITS_PER_ARRAY_ELEMENT - 1) /
                    _IX_BITS_PER_ARRAY_ELEMENT;

    /*
     * To improve performance a bit, allocate both the structure and the array
     * in the same call to ix_ossl_malloc.  This helps during initialization
     * (only one call) as well as at runtime; it is more likely that when the
     * structure is brought into the cache, the array is brought in as well.
     */
    memSize = sizeof(struct ix_s_cc_index_pool) +
              (sizeof(ix_uint32) * arraySize);

    hIndexPool = (ix_cc_index_pool) ix_ossl_malloc(memSize);
    if (hIndexPool == NULL)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_OOM_SYSTEM,
            ("Could not allocate %d bytes from the OSSL", memSize));
    }
    /*
     * Set the entire block to zero: the data structure, as well as all the
     * bits in the array.
     */
    ix_ossl_memset(hIndexPool, 0, memSize);

    /*
     * The array begins at the end of the structure.
     */
    hIndexPool->pArray          = (ix_uint32*) &hIndexPool[1];
    hIndexPool->numElements     = arg_NumElements;
    hIndexPool->arraySize       = arraySize;
    hIndexPool->offset          = arg_Offset;
    hIndexPool->searchHint      = 0;

    /*
     * Some of the bits in the last array element might not be used.
     * Initialize them to 1 so we don't try to allocate one by accident.
     */
    numUsedInLastElement = arg_NumElements % 32;
    if (numUsedInLastElement > 0) /* zero means they're all used */
    {
        hIndexPool->pArray[arraySize-1] = 0xFFFFFFFFUL << numUsedInLastElement;
    }

    *arg_phIndexPool = hIndexPool;
    return IX_SUCCESS;
}

/*
 * NAME: ix_cc_index_pool_fini
 *
 * DESCRIPTION: Destroys an index pool.
 *
 * @Param:      - IN    arg_hIndexPool
 *                      Handle to the ix_cc_index_pool as returned by
 *                      ix_cc_index_pool_init().
 *
 * @Returns:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 */
ix_error ix_cc_index_pool_fini(
    ix_cc_index_pool    arg_hIndexPool)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_hIndexPool);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hIndexPool, struct ix_s_cc_index_pool);

    ix_ossl_free(arg_hIndexPool);
    return IX_SUCCESS;
}

/*
 * NAME: ix_cc_index_pool_alloc
 *
 * DESCRIPTION: Allocates an index out of the pool.
 *
 * @Param:      - IN    arg_hIndexPool
 *                      Handle to the ix_cc_index_pool as returned by
 *                      ix_cc_index_pool_init().
 *
 * @Param:      - OUT   arg_pIndex
 *                      Location where the allocated index is stored.
 *
 * @Returns:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_FULL
 */
ix_error ix_cc_index_pool_alloc(
    ix_cc_index_pool    arg_hIndexPool,
    ix_uint32          *arg_pIndex)
{
    ix_uint32 curr;
    int found = 0;
    ix_uint32 element = 0;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hIndexPool);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hIndexPool, struct ix_s_cc_index_pool);
    IX_ERROR_DCHECK_ARG_NULL(2, arg_pIndex);
    IX_ERROR_DCHECK_ARG_ALIGN(2, arg_pIndex, ix_uint32);

    /*
     * Start at the searchHint and continue to the end of the list.  If none
     * of those are free, start at the beginning and continue until
     * searchHint.
     */
    for (curr = arg_hIndexPool->searchHint;
         curr < arg_hIndexPool->arraySize;
         curr++)
    {
        element = arg_hIndexPool->pArray[curr];

        /*
         * If the element is not 0xFFFFFFFF, at least one of the bits is not
         * set; stop the search and figure out which bit it is.
         */
        if (element != 0xFFFFFFFFUL)
        {
            found = 1;
            break;
        }
    }

    if (!found)
    {
        for (curr = 0;
             curr < arg_hIndexPool->searchHint;
             curr++)
        {
            element = arg_hIndexPool->pArray[curr];

            /*
             * If the element is not 0xFFFFFFFF, at least one of the bits is not
             * set; stop the search and figure out which bit it is.
             */
            if (element != 0xFFFFFFFFUL)
            {
                found = 1;
                break;
            }
        }
    }

    if (found)
    {
        /*
         * Found an entry that shows some unused elements.
         */
        ix_uint8 bit = 0;

        while ((element & 1) == 1)
        {
            element >>= 1;
            bit++;
        }
        arg_hIndexPool->searchHint = curr;
        arg_hIndexPool->pArray[curr] |= 1 << bit;
        *arg_pIndex = (curr * 32 + bit) + arg_hIndexPool->offset;
        return IX_SUCCESS;
    }

    return IX_ERROR_WARNING(IX_CC_ERROR_FULL,
        ("Cannot allocate more indexes"));
}

/*
 * NAME: ix_cc_index_pool_free
 *
 * DESCRIPTION: Frees a previously allocated index back into the pool; the
 * index may be allocated again later.
 *
 * @Param:      - IN    arg_hIndexPool
 *                      Handle to the ix_cc_index_pool as returned by
 *                      ix_cc_index_pool_init().
 *
 * @Param:      - IN    arg_Index
 *                      The index to free back into the pool.
 *
 * @Returns:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_RANGE
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 */
ix_error ix_cc_index_pool_free(
    ix_cc_index_pool    arg_hIndexPool,
    ix_uint32           arg_Index)
{
    ix_uint32 *pElement;
    ix_uint32 internalIndex;
    ix_uint32 arrayIndex;
    ix_uint32 bitOffset;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hIndexPool);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hIndexPool, struct ix_s_cc_index_pool);

    IX_ERROR_CHECK_ARG_RANGE(2, arg_Index,
        arg_hIndexPool->offset,                                     /* min */
        arg_hIndexPool->numElements + arg_hIndexPool->offset - 1,   /* max */
        "%ld");

    internalIndex = arg_Index - arg_hIndexPool->offset;
    arrayIndex = _IX_GET_ELEMENT_INDEX(internalIndex);
    bitOffset = _IX_GET_BIT_OF_ELEMENT(internalIndex);

    pElement = &arg_hIndexPool->pArray[arrayIndex];
    if ((*pElement & (1 << bitOffset)) == 0)
    {
        return IX_ERROR_WARNING(IX_CC_ERROR_ENTRY_NOT_FOUND,
            ("The element at %d was not allocated", arg_Index));
    }

    *pElement &= ~(1 << bitOffset);

    /*
     * Note that the searchHint could be reset to arrayIndex, resulting
     * in a guaranteed hit next time an item is allocated.  However, there
     * is no guarantee that the arrayIndex is any better than the current
     * value of searchHint; in fact, it may be a much worse hint overall.
     */
    return IX_SUCCESS;
}

/*
 * NAME: ix_cc_index_pool_in_use
 *
 * DESCRIPTION: Indicates whether a given index has already been allocated.
 *
 * @Param:      - IN    arg_hIndexPool
 *                      Handle to the ix_cc_index_pool as returned by
 *                      ix_cc_index_pool_init().
 *
 * @Param:      - IN    arg_Index
 *                      The index being queried.
 *
 * @Param:      - OUT   arg_pInUse
 *                      Zero if the index has not been allocated, non-zero if
 *                      the index has been allocated.
 *
 * @Returns:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_RANGE
 */
ix_error ix_cc_index_pool_in_use(
    ix_cc_index_pool    arg_hIndexPool,
    ix_uint32           arg_Index,
    int                *arg_pInUse)
{
    ix_uint32 internalIndex;
    ix_uint32 arrayIndex;
    ix_uint32 bitOffset;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hIndexPool);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hIndexPool, struct ix_s_cc_index_pool);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pInUse);
    IX_ERROR_DCHECK_ARG_ALIGN(3, arg_pInUse, int);

    IX_ERROR_CHECK_ARG_RANGE(2, arg_Index,
        arg_hIndexPool->offset,                                      /* min */
        arg_hIndexPool->numElements + arg_hIndexPool->offset - 1,    /* max */
        "%ld");

    internalIndex = arg_Index - arg_hIndexPool->offset;
    arrayIndex = _IX_GET_ELEMENT_INDEX(internalIndex);
    bitOffset = _IX_GET_BIT_OF_ELEMENT(internalIndex);

    *arg_pInUse =
        arg_hIndexPool->pArray[arrayIndex] & (1 << bitOffset)
        ? 1 : 0;

    return IX_SUCCESS;
}

/*
 * NAME: ix_cc_index_pool_reset
 *
 * DESCRIPTION: Resets the contents of the pool so that all are unallocated.
 *
 * @Param:      - IN    arg_hIndexPool
 *                      Handle to the ix_cc_index_pool as returned by
 *                      ix_cc_index_pool_init().
 *
 * @Returns:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 */
ix_error ix_cc_index_pool_reset(
    ix_cc_index_pool    arg_hIndexPool)
{
    ix_uint32 arraySize;
    int numUsedInLastElement;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hIndexPool);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hIndexPool, struct ix_s_cc_index_pool);

    ix_ossl_memset(arg_hIndexPool->pArray, 0,
            sizeof(ix_uint32) * arg_hIndexPool->arraySize);
    arg_hIndexPool->searchHint = 0;

    /*
     * Some of the bits in the last array element might not be used.
     * Initialize them to 1 so we don't try to allocate one by accident.
     */

    arraySize = arg_hIndexPool->arraySize;
    numUsedInLastElement = arg_hIndexPool->numElements % 32;

    if (numUsedInLastElement > 0) /* zero means they're all used */
    {
        arg_hIndexPool->pArray[arraySize-1] = 0xFFFFFFFFUL << numUsedInLastElement;
    }

    return IX_SUCCESS;
}

/*
 * NAME: _ix_cc_index_pool_get
 *
 * DESCRIPTION: Gets the first allocated index, starting the search at
 * arg_StartPos
 *
 * @Param:      - IN    arg_hIndexPool
 *                      Handle to the ix_cc_index_pool as returned by
 *                      ix_cc_index_pool_init().
 *
 * @Param:      - IN    arg_StartPos
 *                      The first value to search for.  Value from zero
 *                      to arg_hIndexPool->numElements.
 *
 * @Param:      - OUT   arg_pIndex
 *                      The first allocated index in the pool.
 *
 * @Returns:
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 */
PRIVATE ix_error _ix_cc_index_pool_get(
    ix_cc_index_pool    arg_hIndexPool,
    ix_uint32           arg_StartPos,
    ix_uint32          *arg_pIndex)
{
    ix_uint32 curr;
    ix_uint32 arrayIndex;
    ix_uint32 bitOffset;

    for (curr = arg_StartPos; curr < arg_hIndexPool->numElements; curr++)
    {
        arrayIndex = _IX_GET_ELEMENT_INDEX(curr);
        bitOffset = _IX_GET_BIT_OF_ELEMENT(curr);

        if (arg_hIndexPool->pArray[arrayIndex] & (1 << bitOffset))
        {
            *arg_pIndex = curr + arg_hIndexPool->offset;
            return IX_SUCCESS;
        }
    }

    /*
     * Report that none could be found. Use IX_ERROR_NEW to avoid the overhead
     * of the error extensions.
     */
    return IX_ERROR_NEW(IX_CC_ERROR_ENTRY_NOT_FOUND, IX_ERROR_LEVEL_WARNING);
}

/*
 * NAME: ix_cc_index_pool_get_first
 *
 * DESCRIPTION: Gets the first allocated index from the pool.
 *
 * @Param:      - IN    arg_hIndexPool
 *                      Handle to the ix_cc_index_pool as returned by
 *                      ix_cc_index_pool_init().
 *
 * @Param:      - OUT   arg_pIndex
 *                      The first allocated index in the pool.
 *
 * @Returns:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 */
ix_error ix_cc_index_pool_get_first(
    ix_cc_index_pool    arg_hIndexPool,
    ix_uint32          *arg_pIndex)
{
    IX_ERROR_DCHECK_ARG_NULL(1, arg_hIndexPool);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hIndexPool, struct ix_s_cc_index_pool);
    IX_ERROR_DCHECK_ARG_NULL(2, arg_pIndex);
    IX_ERROR_DCHECK_ARG_ALIGN(2, arg_pIndex, ix_uint32);

    return _ix_cc_index_pool_get(arg_hIndexPool, 0, arg_pIndex);
}

/*
 * NAME: ix_cc_index_pool_get_next
 *
 * DESCRIPTION: Gets the next allocated index from the pool, following the
 * provided index
 *
 * @Param:      - IN    arg_hIndexPool
 *                      Handle to the ix_cc_index_pool as returned by
 *                      ix_cc_index_pool_init().
 *
 * @Param:      - IN    arg_Previous
 *                      The index previously returned by
 *                      ix_cc_index_pool_get_first() or
 *                      ix_cc_index_pool_get_next().
 *
 * @Param:      - OUT   arg_pIndex
 *                      The next allocated index in the pool.
 *
 * @Returns:
 * IX_CC_ERROR_NULL
 * IX_CC_ERROR_ALIGN
 * IX_CC_ERROR_ENTRY_NOT_FOUND
 */
ix_error ix_cc_index_pool_get_next(
    ix_cc_index_pool    arg_hIndexPool,
    ix_uint32           arg_Previous,
    ix_uint32          *arg_pIndex)
{
    ix_uint32 startPos;

    IX_ERROR_DCHECK_ARG_NULL(1, arg_hIndexPool);
    IX_ERROR_DCHECK_ARG_ALIGN(1, arg_hIndexPool, struct ix_s_cc_index_pool);
    IX_ERROR_DCHECK_ARG_NULL(3, arg_pIndex);
    IX_ERROR_DCHECK_ARG_ALIGN(3, arg_pIndex, ix_uint32);

    startPos = arg_Previous + 1 - arg_hIndexPool->offset;
    return _ix_cc_index_pool_get(arg_hIndexPool, startPos, arg_pIndex);
}

