/**
 * ===========================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *     Copyright (C) 2002-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      l2_config.c
 *
 * = DESCRIPTION
 *      Entry points for controlling the L2 Table Manager from the VxWorks
 *      command line.
 * ===========================================================================
 */

#define IX_ERROR_FILE_IDENT "$Id: l2_config.c,v 1.16 2003/12/12 00:41:36 ktseng Exp $"

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
#include "ix_ossl.h"
#include <stdio.h>
#endif

#if (_IX_OS_TYPE_ != _IX_OS_LINUX_KERNEL_)
#include <string.h>
#endif

#include "ix_cc_error.h"
#include "cc/ix_cc_l2tm.h"
#include "ix_cc.h"

#define ADD_V4_ETH_ENTRY_USAGE                                              \
     "Adds a complete IPv4/Ethernet address to the L2 Table.\n"              \
    "addV4EthEntry usage:\n"                                                \
    "\taddV4EthEntry \"<l2Index> <ipAddr> <DestMAC> <SourceMAC> [name]\","  \
        " where:\n"                                                         \
    "\tl2Index:\tIndex into the L2 Table to be added, in Hex\n"             \
    "\tipAddr:\t\tIP address in dotted-decimal form\n"                      \
    "\tDestMAC:\tDestination MAC address (hex byte array) \n"               \
    "\tSourceMAC:\tSource MAC address (hex byte array)\n"                   \
    "\tname:\t\tOptional named table (or \"DEFAULT\" if left blank)\n"      \
    "Example:\n\taddV4EthEntry \"1 10.3.19.10 0a:0c:14:37:20:a0 "           \
        "cf:30:25:d6:4e:2f DEFAULT\"\n"

#define ADD_V6_ETH_ENTRY_USAGE                                              \
    "Adds a complete IPv6/Ethernet address to the L2 Table.\n"              \
    "addV6EthEntry usage:\n"                                                \
    "\taddV6EthEntry \"<l2Index> <IPv6Addr> <DestMAC> <SrcMAC> [name]\","   \
        " where:\n"                                                         \
    "\tl2Index:\tIndex into the L2 Table to be added, in Hex\n"             \
    "\tIPv6Addr:\t\tIPv6 address in colon separated form\n"                 \
    "\tDestMAC:\tDestination MAC address (hex byte array) \n"               \
    "\tSrcMAC:\tSource MAC address (hex byte array)\n"                      \
    "\tname:\t\tOptional named table (or \"DEFAULT\" if left blank)\n"      \
    "  Example:\n\taddV6EthEntry \"1 3FFE:0:0:100::1 0a:0c:14:37:20:a0 cf:30:25:d6:4e:2f DEFAULT\"\n"

#define ADD_V4_L3_INFO_USAGE                                                \
    "Adds an IPv4 address to the L2 Table.\n"                               \
    "addV4L3Info usage:\n"                                                  \
    "\taddV4L3Info \"<l2Index> <ipAddr> [name]\", where:\n"                 \
    "\tl2Index:\tIndex into the L2 Table to be added, in Hex\n"             \
    "\tipAddr:\t\tIP address in dotted-decimal form\n"                      \
    "\tname:\t\tOptional named table (or \"DEFAULT\" if left blank)\n"      \
    "Example:\n\taddV4L3Info \"2 10.3.19.11 DEFAULT\"\n"

#define ADD_V4_ATM_ENTRY_USAGE                                              \
    "Adds a complete IPv4/ATM address to the L2 Table.\n"                   \
    "addV4AtmEntry usage:\n"                                                \
    "\taddV4AtmEntry \"<l2Index> <ipAddr> <vpiVci> <vcq> <llcSnap>"         \
        " [name]\", where:\n"                                               \
    "\tl2Index:\tIndex into the L2 Table to be added, in Hex\n"             \
    "\tipAddr:\t\tIP address in dotted-decimal form\n"                      \
    "\tvpiVci:\t\tVirtual Path Id / Virtual Channel Id in Hex\n"            \
    "\tvcq:\t\tVirtual Channel Queue in Hex\n"                              \
    "\tllcSnap:\t64-bit LLC / SNAP Header in Hex\n"                         \
    "\tname:\t\tOptional named table (or \"DEFAULT\" if left blank)\n"      \
    "Example:\n\taddV4AtmEntry \"3 10.3.19.12 0x01234567 0x4321 0x123456789ABCDEF0 DEFAULT\"\n"

#define DELETE_L2_ENTRY_USAGE                                               \
    "Completely deletes an entry from the L2 Table.\n"                      \
    "deleteL2Entry usage:\n"                                                \
    "\tdeleteL2Entry \"<l2Index> [name]\", where:\n"                        \
    "\tl2Index:\tIndex into the L2 Table to be removed, in Hex\n"           \
    "\tname:\t\tOptional named table (or \"DEFAULT\" if left blank)\n"      \
    "Example:\n\tdeleteL2Entry \"3 DEFAULT\"\n"

#define CLEAR_L2_INFO_USAGE                                                 \
    "Removes L2 Information from the L2 Table (IP address is unchanged).\n" \
    "clearL2Info usage:\n"                                                  \
    "\tclearL2Info \"<l2Index> [name]\", where:\n"                          \
    "\tl2Index:\tIndex into the L2 Table to be cleared, in Hex\n"           \
    "\tname:\t\tOptional named table (or \"DEFAULT\" if left blank)\n"      \
    "Example:\n\tclearL2Info \"1 DEFAULT\"\n"

extern int ix_inet_pton(int af, const char *src, void *dst);

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

/* IGK0100023119: There are some problems with sscanf function in kernel, %hX doesn't accept letters */

static short int str_to_hex(char c)
{
    switch(c)
    {
    case '0':return 0;
    case '1':return 1;
    case '2':return 2;
    case '3':return 3;
    case '4':return 4;
    case '5':return 5;
    case '6':return 6;
    case '7':return 7;
    case '8':return 8;
    case '9':return 9;
    case 'a':
    case 'A':
        return 10;
    case 'b':
    case 'B':
        return 11;
    case 'c':
    case 'C':
        return 12;
    case 'd':
    case 'D':
        return 13;
    case 'e':
    case 'E':
        return 14;
    case 'f':
    case 'F':
        return 15;
    default:return 0;
    }
}

static int parse_mac(char* s,unsigned short int* a)
{
    char * chrptr = NULL;
    int i = 0;
    chrptr=strtok(s,":");
    while(chrptr != NULL)
    {
        a[i] = (str_to_hex(*chrptr) * 16);
        chrptr++;
        a[i] += str_to_hex(*chrptr);
        chrptr=strtok(NULL,":");
        i++;
    }
    if(i!=6)
        return (-1);
    return 0;
}

static unsigned int parse_alfa_to_hex(char* str)
{
    char* chr;
    unsigned int res = 0;
    chr = str;
    while(*chr != 0)
    {
        res = (res << 4) + str_to_hex(*chr);
        chr++;
    }
    return res;
}

#endif /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */



int addV4EthEntry(const char *argv)
{
    ix_cc_l2tm hL2tm;
    unsigned int ipA, ipB, ipC, ipD;
    ix_uint16 l2Index;
    ix_cc_l2tm_entry entry;
    int numParams;
    ix_cc_l2tm_ether_header *pEther = &entry.header.ether;
    char name[100];
    ix_error err;

    /*
     * These are because sscanf() for VxWorks doesn't support %hhX.
     */
    ix_uint16 sourceMAC[6], destMAC[6];
    int i;

    if (argv == NULL)
    {

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        ix_ossl_message_log(ADD_V4_ETH_ENTRY_USAGE);
#else
        printf(ADD_V4_ETH_ENTRY_USAGE);
#endif
        return -1;
    }

    /*
     * L2 Index
     * IPv4 address
     * Destination MAC address
     * Source MAC address
     * (name)
     */
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

/* IGK0100023119: There are some problems with sscanf function in kernel, %hX doesn't accept letters */
/* This code may be removed once the kernel sscanf problem is fixed */
    {
      char deststr[20];
      char srcstr[20];
      char l2Index_str[20];
      numParams = sscanf(argv,
			 "%s %d.%d.%d.%d %s %s %s",
			 l2Index_str,
			 &ipA, &ipB, &ipC, &ipD,
			 deststr,
			 srcstr,
			 name);
      if ((numParams != 7) && (numParams != 8))
	{
	  ix_ossl_message_log("numParams:%d\n",numParams);
	  ix_ossl_message_log(ADD_V4_ETH_ENTRY_USAGE);
	  return -1;
	}
      if(parse_mac(&deststr[0],&destMAC[0]))
	{
	  ix_ossl_message_log("Error in destination MAC Address\n");
	  ix_ossl_message_log(ADD_V4_ETH_ENTRY_USAGE);
	  return -1;
	}
      if(parse_mac(&srcstr[0],&sourceMAC[0]))
	{
	  ix_ossl_message_log("Error in source MAC Address\n");
	  ix_ossl_message_log(ADD_V4_ETH_ENTRY_USAGE);
	  return -1;
	}
      l2Index = parse_alfa_to_hex(l2Index_str);
    }

#else /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

    numParams = sscanf(argv,
        /*
         * The VxWorks port of sscanf() doesn't accept %hhX (unsigned char in
         * hex) so we have to scan in using %hX and move it into the array.
         */
        "%hX %d.%d.%d.%d %hX:%hX:%hX:%hX:%hX:%hX %hX:%hX:%hX:%hX:%hX:%hX %s",
        &l2Index,
        &ipA, &ipB, &ipC, &ipD,
        &destMAC[0], &destMAC[1], &destMAC[2],
        &destMAC[3], &destMAC[4], &destMAC[5],
        &sourceMAC[0], &sourceMAC[1], &sourceMAC[2],
        &sourceMAC[3], &sourceMAC[4], &sourceMAC[5],
        name);
    if ((numParams != 17) && (numParams != 18))
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_ossl_message_log(ADD_V4_ETH_ENTRY_USAGE);
#else
        printf(ADD_V4_ETH_ENTRY_USAGE);
#endif
        return -1;
    }

#endif /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

    for (i = 0; i < 6; i++)
    {
        pEther->destMAC[i] = destMAC[i];
        pEther->sourceMAC[i] = sourceMAC[i];
    }
    pEther->protocol = 0x0800;  /* IPv4 */

    if ((ipA > 255) || (ipB > 255) || (ipC > 255) | (ipD > 255))
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_ossl_message_log("Invalid IP address: %d.%d.%d.%d\n",
            ipA, ipB, ipC, ipD);
#else
        printf("Invalid IP address: %d.%d.%d.%d\n",
            ipA, ipB, ipC, ipD);

#endif
        return -1;
    }

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    if (numParams == 7)
#else
    if (numParams == 17)
#endif /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */
    {
        strcpy(name, "DEFAULT");
    }

    entry.l2AddrType = IX_CC_L2TM_L2ADDR_ETHER;
    entry.ipAddrType = IX_CC_L2TM_IPADDR_IPV4;
    entry.l3Address.ipv4Address = (ipA << 24) | (ipB << 16) | (ipC << 8) | ipD;

    IX_ERROR_CG(ix_cc_l2tm_create(name, &hL2tm), err, report);
    IX_ERROR_CG(ix_cc_l2tm_add_entry(hL2tm, l2Index, &entry), err, clean_hL2tm);
    IX_ERROR_CG(ix_cc_l2tm_destroy(hL2tm), err, report);

    return 0;

clean_hL2tm:
    (void) ix_cc_l2tm_destroy(hL2tm);
report:

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif

    return -1;
}

int addV6EthEntry(const char *argv)
{
    ix_cc_l2tm hL2tm;
    char v6addrStr[50];
    ix_uint128 ipv6Addr;
    ix_uint16 l2Index;
    ix_cc_l2tm_entry entry;
    int numParams;
    ix_cc_l2tm_ether_header *pEther = &entry.header.ether;
    char name[100];
    ix_error err;

    /*
     * These are because sscanf() for VxWorks doesn't support %hhX.
     */
    ix_uint16 sourceMAC[6], destMAC[6];
    int i;

    if (argv == NULL)
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        ix_ossl_message_log(ADD_V6_ETH_ENTRY_USAGE);
        
#else
        printf(ADD_V6_ETH_ENTRY_USAGE);
#endif
        return -1;
    }
 
    /*
     * L2 Index
     * IPv4 address
     * Destination MAC address
     * Source MAC address
     * (name)
     */
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

/* IGK0100023119: There are some problems with sscanf function in kernel, %hX doesn't accept letters */
/* This code may be removed once the kernel sscanf problem is fixed */
    {
      char deststr[20];
      char srcstr[20];
      char l2Index_str[20];
      numParams = sscanf(argv,
        "%s %s %s %s %s",
        l2Index_str, v6addrStr,
        deststr, srcstr,
        name);
      if ((numParams != 4) && (numParams != 5))
        {
	  ix_ossl_message_log("numParams:%d\n",numParams);
	  ix_ossl_message_log(ADD_V6_ETH_ENTRY_USAGE);
	  return -1;
	}
      if(parse_mac(&deststr[0],&destMAC[0]))
	{
	  ix_ossl_message_log("Error in destination MAC Address\n");
	  ix_ossl_message_log(ADD_V6_ETH_ENTRY_USAGE);
	  return -1;
	}
      if(parse_mac(&srcstr[0],&sourceMAC[0]))
	{
	  ix_ossl_message_log("Error in source MAC Address\n");
	  ix_ossl_message_log(ADD_V6_ETH_ENTRY_USAGE);
	  return -1;
	}
      l2Index = parse_alfa_to_hex(l2Index_str);
    }

#else /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

    numParams = sscanf(argv,
        /*
         * The VxWorks port of sscanf() doesn't accept %hhX (unsigned char in
         * hex) so we have to scan in using %hX and move it into the array.
         */
        "%hX %s %hX:%hX:%hX:%hX:%hX:%hX %hX:%hX:%hX:%hX:%hX:%hX %s",
        &l2Index, v6addrStr,
        &destMAC[0], &destMAC[1], &destMAC[2],
        &destMAC[3], &destMAC[4], &destMAC[5],
        &sourceMAC[0], &sourceMAC[1], &sourceMAC[2],
        &sourceMAC[3], &sourceMAC[4], &sourceMAC[5],
        name);
    if ((numParams != 14) && (numParams != 15))
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_ossl_message_log(ADD_V6_ETH_ENTRY_USAGE);
#else
        printf(ADD_V6_ETH_ENTRY_USAGE);
#endif
        return -1;
    }

#endif /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

    if (ix_inet_pton(IX_AF_INET6, v6addrStr, &ipv6Addr) != 1)
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_ossl_message_log("IPv6 Address is Invalid!\n");
#else
        printf("IPv6 Address is Invalid!\n");
#endif
        return -1;
    }

    for (i = 0; i < 6; i++)
    {
        pEther->destMAC[i] = destMAC[i];
        pEther->sourceMAC[i] = sourceMAC[i];
    }
    pEther->protocol = 0x86DD;  /* IPv6 */

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    if (numParams == 4)
#else
    if (numParams == 14)
#endif /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */
    {
        strcpy(name, "DEFAULT");
    }

    entry.l2AddrType = IX_CC_L2TM_L2ADDR_ETHER;
    entry.ipAddrType = IX_CC_L2TM_IPADDR_IPV6;
    entry.l3Address.ipv6Address = ipv6Addr;

    IX_ERROR_CG(ix_cc_l2tm_create(name, &hL2tm), err, report);
    IX_ERROR_CG(ix_cc_l2tm_add_entry(hL2tm, l2Index, &entry), err, clean_hL2tm);
    IX_ERROR_CG(ix_cc_l2tm_destroy(hL2tm), err, report);

    return 0;

clean_hL2tm:
    (void) ix_cc_l2tm_destroy(hL2tm);
report:

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif

    return -1;
}

int addV4AtmEntry(const char *argv)
{
    ix_cc_l2tm hL2tm;
    unsigned int ipA, ipB, ipC, ipD;
    ix_uint16 l2Index;
    ix_cc_l2tm_entry entry;
    int numParams;
    ix_cc_l2tm_atm_header *pAtm = &entry.header.atm;
    char name[100];
    ix_error err;

    if (argv == NULL)
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        ix_ossl_message_log(ADD_V4_ATM_ENTRY_USAGE);
#else
        printf(ADD_V4_ATM_ENTRY_USAGE);
#endif
        return -1;
    }

    /*
     * L2 Index
     * IPv4 address
     * vpiVci
     * vcq
     * llcSnap1 / llcSnap0
     * (name)
     */
    numParams = sscanf(argv,
        /*
         * The VxWorks port of sscanf() doesn't accept %hhX (unsigned char in
         * hex) so we have to scan in using %hX and move it into the array.
         */
        "%hX %d.%d.%d.%d %lX %lX %llX %s",
        &l2Index,
        &ipA, &ipB, &ipC, &ipD,
        &pAtm->vpiVci, &pAtm->vcq, &pAtm->llcSnapHeader,
        name);
    if ((numParams != 9) && (numParams != 10))
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        ix_ossl_message_log(ADD_V4_ATM_ENTRY_USAGE);
#else
        printf(ADD_V4_ATM_ENTRY_USAGE);
#endif
        return -1;
    }

    if ((ipA > 255) || (ipB > 255) || (ipC > 255) | (ipD > 255))
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_ossl_message_log("Invalid IP address: %d.%d.%d.%d\n",
            ipA, ipB, ipC, ipD);

#else
        printf("Invalid IP address: %d.%d.%d.%d\n",
            ipA, ipB, ipC, ipD);
#endif
        return -1;
    }

    if (numParams == 9)
    {
        strcpy(name, "DEFAULT");
    }

    entry.l2AddrType = IX_CC_L2TM_L2ADDR_ATM;
    entry.ipAddrType = IX_CC_L2TM_IPADDR_IPV4;
    entry.l3Address.ipv4Address = (ipA << 24) | (ipB << 16) | (ipC << 8) | ipD;
    /*pAtm->llcSnapHeader = (((ix_uint64) llcSnap1) << 32) | llcSnap0;*/

    IX_ERROR_CG(ix_cc_l2tm_create(name, &hL2tm), err, report);
    IX_ERROR_CG(ix_cc_l2tm_add_entry(hL2tm, l2Index, &entry), err, clean_hL2tm);
    IX_ERROR_CG(ix_cc_l2tm_destroy(hL2tm), err, report);

    return 0;

clean_hL2tm:
    (void) ix_cc_l2tm_destroy(hL2tm);
report:

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif

    return -1;
}

int addV4L3Info(const char *argv)
{
    ix_cc_l2tm hL2tm;
    unsigned int ipA, ipB, ipC, ipD;
    ix_uint16 l2Index;
    int numParams;
    char name[100];
    ix_error err;
    ix_cc_l2tm_ipaddr ipAddress;

    if (argv == NULL)
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        ix_ossl_message_log(ADD_V4_L3_INFO_USAGE);
#else
        printf(ADD_V4_L3_INFO_USAGE);
#endif
        return -1;
    }

    /*
     * L2 Index
     * IPv4 address
     * (name)
     */
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

/* IGK0100023119: There are some problems with sscanf function in kernel, %hX doesn't accept letters */
/* This code may be removed once the kernel sscanf problem is fixed */
    {
      char l2Index_str[20];

      numParams = sscanf(argv,
        "%s %d.%d.%d.%d %s",
        l2Index_str,
        &ipA, &ipB, &ipC, &ipD,
        name);
      if ((numParams != 5) && (numParams != 6))
        {
	  ix_ossl_message_log("numParams:%d\n",numParams);
          ix_ossl_message_log(ADD_V4_L3_INFO_USAGE);
	  return -1;
	}
      l2Index = parse_alfa_to_hex(l2Index_str);
    }

#else /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

    numParams = sscanf(argv,
        /*
         * The VxWorks port of sscanf() doesn't accept %hhX (unsigned char in
         * hex) so we have to scan in using %hX and move it into the array.
         */
        "%hX %d.%d.%d.%d %s",
        &l2Index,
        &ipA, &ipB, &ipC, &ipD,
        name);
    if ((numParams != 5) && (numParams != 6))
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        ix_ossl_message_log(ADD_V4_L3_INFO_USAGE);
#else
        printf(ADD_V4_L3_INFO_USAGE);
#endif
        return -1;
    }

#endif /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

    if ((ipA > 255) || (ipB > 255) || (ipC > 255) | (ipD > 255))
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_ossl_message_log("Invalid IP address: %d.%d.%d.%d\n",
            ipA, ipB, ipC, ipD);

#else
        printf("Invalid IP address: %d.%d.%d.%d\n",
            ipA, ipB, ipC, ipD);
#endif
        return -1;
    }

    if (numParams == 5)
    {
        strcpy(name, "DEFAULT");
    }

    ipAddress.ipv4Address = (ipA << 24) | (ipB << 16) | (ipC << 8) | ipD;

    IX_ERROR_CG(ix_cc_l2tm_create(name, &hL2tm), err, report);
    IX_ERROR_CG(ix_cc_l2tm_add_l3_info(hL2tm, l2Index, &ipAddress,
            IX_CC_L2TM_IPADDR_IPV4), err, clean_hL2tm);
    IX_ERROR_CG(ix_cc_l2tm_destroy(hL2tm), err, report);

    return 0;

clean_hL2tm:
    (void) ix_cc_l2tm_destroy(hL2tm);
report:

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif

    return -1;
}

int deleteL2Entry(const char *argv)
{
    ix_cc_l2tm hL2tm;
    unsigned short int l2Index;
    char name[100];
    int numParams;
    ix_error err;

    if (argv == NULL)
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        ix_ossl_message_log(DELETE_L2_ENTRY_USAGE);
#else
        printf(DELETE_L2_ENTRY_USAGE);
#endif
        return -1;
    }

    /*
     * L2 Index
     * (name)
     */
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

/* IGK0100023119: There are some problems with sscanf function in kernel, %hX doesn't accept letters */
/* This code may be removed once the kernel sscanf problem is fixed */
    {
      char l2Index_str[20];
      numParams = sscanf(argv, "%s %s", l2Index_str, name);
      if ((numParams != 1) && (numParams != 2))
        {
	  ix_ossl_message_log("numParams:%d\n",numParams);
          ix_ossl_message_log(DELETE_L2_ENTRY_USAGE);
	  return -1;
	}
      l2Index = parse_alfa_to_hex(l2Index_str);
    }

#else /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

    numParams =
        sscanf(argv, "%hX %s", &l2Index, name);
    if ((numParams != 1) && (numParams != 2))
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        ix_ossl_message_log(DELETE_L2_ENTRY_USAGE);
#else
        printf(DELETE_L2_ENTRY_USAGE);
#endif
        return -1;
    }

#endif /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

    if (numParams == 1)
    {
        strcpy(name, "DEFAULT");
    }

    IX_ERROR_CG(ix_cc_l2tm_create(name, &hL2tm), err, report);
    IX_ERROR_CG(ix_cc_l2tm_delete_entry(hL2tm, l2Index), err, clean_hL2tm);
    IX_ERROR_CG(ix_cc_l2tm_destroy(hL2tm), err, report);

    return 0;

clean_hL2tm:
    (void) ix_cc_l2tm_destroy(hL2tm);
report:

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif

    return -1;
}

int clearL2Info(const char *argv)
{
    ix_cc_l2tm hL2tm;
    unsigned short int l2Index;
    char name[100];
    int numParams;
    ix_error err;

    if (argv == NULL)
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        ix_ossl_message_log(CLEAR_L2_INFO_USAGE);
#else
        printf(CLEAR_L2_INFO_USAGE);
#endif
        return -1;
    }

    /*
     * L2 Index
     * (name)
     */
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

/* IGK0100023119: There are some problems with sscanf function in kernel, %hX doesn't accept letters */
/* This code may be removed once the kernel sscanf problem is fixed */
    {
      char l2Index_str[20];
      numParams = sscanf(argv, "%s %s", l2Index_str, name);
      if ((numParams != 1) && (numParams != 2))
        {
	  ix_ossl_message_log("numParams:%d\n",numParams);
          ix_ossl_message_log(CLEAR_L2_INFO_USAGE);
	  return -1;
	}
      l2Index = parse_alfa_to_hex(l2Index_str);
    }

#else /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

    numParams =
        sscanf(argv, "%hX %s", &l2Index, name);
    if ((numParams != 1) && (numParams != 2))
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        ix_ossl_message_log(CLEAR_L2_INFO_USAGE);
#else
        printf(CLEAR_L2_INFO_USAGE);
#endif
        return -1;
    }

#endif /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

    if (numParams == 1)
    {
        strcpy(name, "DEFAULT");
    }

    IX_ERROR_CG(ix_cc_l2tm_create(name, &hL2tm), err, report);
    IX_ERROR_CG(ix_cc_l2tm_clear_l2_info(hL2tm, l2Index), err, clean_hL2tm);
    IX_ERROR_CG(ix_cc_l2tm_destroy(hL2tm), err, report);

    return 0;

clean_hL2tm:
    (void) ix_cc_l2tm_destroy(hL2tm);
report:

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif

    return -1;
}

int purgeL2tm(char *argv)
{
    ix_cc_l2tm hL2tm;
    ix_error err;

    if (argv == NULL)
    {
        IX_ERROR_CG(ix_cc_l2tm_create("DEFAULT", &hL2tm), err, report);
    }
    else
    {
        IX_ERROR_CG(ix_cc_l2tm_create(argv, &hL2tm), err, report);
    }

    IX_ERROR_CG(ix_cc_l2tm_purge(hL2tm), err, clean_hL2tm);
    IX_ERROR_CG(ix_cc_l2tm_destroy(hL2tm), err, report);

    return 0;

clean_hL2tm:
    (void) ix_cc_l2tm_destroy(hL2tm);
report:

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif

    return -1;
}

int dumpL2tm(char *argv)
{
    ix_cc_l2tm hL2tm;
    ix_error err;

    if (argv == NULL)
    {
        IX_ERROR_CG(ix_cc_l2tm_create("DEFAULT", &hL2tm), err, report);
    }
    else
    {
        IX_ERROR_CG(ix_cc_l2tm_create(argv, &hL2tm), err, report);
    }

    IX_ERROR_CG(ix_cc_l2tm_dump(hL2tm), err, clean_hL2tm);
    IX_ERROR_CG(ix_cc_l2tm_destroy(hL2tm), err, report);

    return 0;

clean_hL2tm:
    (void) ix_cc_l2tm_destroy(hL2tm);
report:

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif

    return -1;
}

int l2_config(void)
{
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_ossl_message_log("Commands for controlling the L2 Table Manager:\n");
    ix_ossl_message_log("\taddV4AtmEntry\taddV4EthEntry\taddV6EthEntry\taddV4L3Info\n");
    ix_ossl_message_log("\tdeleteL2Entry\tclearL2Info\tpurgeL2tm\tdumpL2tm\n");
    ix_ossl_message_log("Type the name of the command to get usage information\n");
    ix_ossl_message_log("(NOTE: purgeL2tm and dumpL2tm accept a name, but do not "
           "require parameters.)\n");


#else
    printf("Commands for controlling the L2 Table Manager:\n");
    printf("\taddV4AtmEntry\taddV4EthEntry\taddV6EthEntry\taddV4L3Info\n");
    printf("\tdeleteL2Entry\tclearL2Info\tpurgeL2tm\tdumpL2tm\n");
    printf("Type the name of the command to get usage information\n");
    printf("(NOTE: purgeL2tm and dumpL2tm accept a name, but do not "
           "require parameters.)\n");
#endif
    return 0;
}
