/**
 * ===========================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      mac_config.c
 *
 *= AUTHOR
 *      Ranjeeta Singh
 * = DESCRIPTION
 *      Entry points for controlling the MAC filter table from the VxWorks
 *      command line.
 * ===========================================================================
 */

#define IX_ERROR_FILE_IDENT "$Id: mac_config.c,v 1.5 2003/05/08 22:06:53 rranjeet Exp $"

#if (_IX_OS_TYPE_ != _IX_OS_LINUX_KERNEL_)
#include <string.h>
#endif

#include "ix_cc_error.h"
#include "cc/ix_cc_eth_rx.h"

#define MAC_CONFIG_TIMEOUT 10000

#define ADD_MAC_USAGE                                              \
    "addMac usage:\n"                                               \
    "addMac \"<mac addr> <portID> "      \
    " where <mac addr> is in hex format 0a:01:02:03:0d:0a\n" \
    "  portId is in hexidecimal.\n"                         \
    "Example:\n\taddMac \"0a:0b:0c:01:02:03 0\"\n"

#define DEL_MAC_USAGE                                              \
    "delMac usage: delMac \"<mac addr>\",\n"                   \
    "\twhere <mac addr> is the mac addr to be deleted\n"        \
    "Example:\n\tdelMac \"0a:0b:0c:04:0c:0a\"\n"                                   \
    
#define LOOKUP_PORT_USAGE                                              \
    "lookupPort usage: lookupPort \"<mac addr>\",\n"                   \
    "\twhere <mac addr> is the mac addr to be looked up\n"        \
    "Example:\n\tlookupPort \"0a:0b:0c:04:0c:0a\"\n"                                   \


/* 
 * Services callbacks from add Mac, del Mac
 * Wakes up the caller when the callback comes in.
 */
static ix_error mac_config_cb(ix_error arg_Result, void *arg_pUserCtx)
{

    if (arg_Result != IX_SUCCESS)
    {
        /*
         * Print out the error for the user's benefit.
         */
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_error_dump(0, arg_Result);
#else
        ix_error_dump(stdout, arg_Result);
#endif
    }

    return IX_SUCCESS;
}

/* 
 * Services callbacks from lookup port
 * Wakes up the caller when the callback comes in.
 */
static ix_error mac_lookup_cb(ix_error arg_Result, void *arg_pUserCtx,
            ix_uint32 *arg_pPortNum)
{
   

    if (arg_Result != IX_SUCCESS)
    {
        /*
         * Print out the error for the user's benefit.
         */
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_error_dump(0, arg_Result);
#else
        ix_error_dump(stdout, arg_Result);
#endif
    }
    if (arg_Result == IX_SUCCESS)
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

		ix_ossl_message_log("Port ID got from MAC Filter table is %ld\n",*arg_pPortNum);
#else
		printf("Port ID got from MAC Filter table is %ld\n",*arg_pPortNum);
#endif	
    }
   

    return IX_SUCCESS;
}


ix_error ix_add_mac(
    ix_uint8           arg_destMac[6],
    ix_uint32           arg_portId)
{
    ix_error err;

    ix_ossl_sem_t sem;

    IX_ERROR_CG(ix_ossl_sem_init(IX_OSSL_SEM_UNAVAILABLE, &sem),
            err, report);

    IX_ERROR_CG(ix_cc_eth_rx_async_add_mac_addr(
                        arg_destMac, arg_portId,
                        mac_config_cb, &sem),
                err, clean_sem);

    IX_ERROR_CG(ix_ossl_sem_fini(sem), err, report);

    return IX_SUCCESS;

clean_sem:
    (void) ix_ossl_sem_fini(sem);
report:
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif
    return err;
}

ix_error ix_del_mac(
    ix_uint8           arg_destMac[6])
{
    ix_error err;
    ix_ossl_sem_t sem;

    IX_ERROR_CR(ix_ossl_sem_init(IX_OSSL_SEM_UNAVAILABLE, &sem));
    IX_ERROR_CG(ix_cc_eth_rx_async_delete_mac_addr(
                        arg_destMac,
                        mac_config_cb, &sem),
                err, clean_sem);
    
    IX_ERROR_CR(ix_ossl_sem_fini(sem));

    return IX_SUCCESS;

clean_sem:
    (void) ix_ossl_sem_fini(sem);
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif
    return err;
}

ix_error ix_lookup_port(
    ix_uint8           arg_destMac[6])
    
{
    ix_error err;
    ix_ossl_sem_t sem;

    IX_ERROR_CG(ix_ossl_sem_init(IX_OSSL_SEM_UNAVAILABLE, &sem),
            err, report);
    IX_ERROR_CG(ix_cc_eth_rx_async_lookup_port(
                        arg_destMac,
                        mac_lookup_cb, &sem),
                err, clean_sem);
   
    IX_ERROR_CG(ix_ossl_sem_fini(sem), err, report);

    return IX_SUCCESS;

clean_sem:
    (void) ix_ossl_sem_fini(sem);
report:
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif
    return err;
}

int addMac(const char *argv)
{
    int ret;
    unsigned int preA, preB, preC, preD;
    unsigned int preE, preF;
    unsigned int port;
	ix_uint8 destMac[6];

    if (argv == NULL)
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        ix_ossl_message_log(ADD_MAC_USAGE);
#else
        printf(ADD_MAC_USAGE);
#endif
        return 1;
    }

    if (sscanf(argv, "%x:%x:%x:%x:%x:%x %d",
            &preA, &preB, &preC, &preD,
            &preE, &preF, &port) != 7)
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        ix_ossl_message_log(ADD_MAC_USAGE);
#else
        printf(ADD_MAC_USAGE);
#endif
        return 1;
    }

    destMac[0] = preA;
    destMac[1] = preB; 
	destMac[2] = preC;
	destMac[3] = preD;
	destMac[4] = preE;
	destMac[5] = preF;
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_ossl_message_log("Adding Mac:\n");
    ix_ossl_message_log("Mac addr = %x:%x:%x:%x:%x:%x\n",
           preA, preB, preC, preD,
           preE, preF);
    ix_ossl_message_log("Port Id = %d\n", port);


#else	
    printf("Adding Mac:\n");
    printf("Mac addr = %x:%x:%x:%x:%x:%x\n",
           preA, preB, preC, preD,
           preE, preF);
    printf("Port Id = %d\n", port);
#endif
    ret = ix_add_mac(destMac, port);
    return ret == IX_SUCCESS ? 0 : 1;
}

int delMac(const char *argv)
{
    int ret;
    unsigned int preA, preB, preC, preD;
    unsigned int preE, preF;
    ix_uint8 destMac[6];

    if (argv == NULL)
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        ix_ossl_message_log(DEL_MAC_USAGE);
#else
        printf(DEL_MAC_USAGE);
#endif
        return 1;
    }

    if (sscanf(argv, "%x:%x:%x:%x:%x:%x",
            &preA, &preB, &preC, &preD,
            &preE, &preF) != 6)
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        ix_ossl_message_log(DEL_MAC_USAGE);
#else
        printf(DEL_MAC_USAGE);
#endif
        return 1;
    }

    destMac[0] = preA;
    destMac[1] = preB;
    destMac[2] = preC;
    destMac[3] = preD;
    destMac[4] = preE;
    destMac[5] = preF;
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    ix_ossl_message_log("Deleting Mac addr:\n");
    ix_ossl_message_log("Mac = %x:%x:%x:%x:%x:%x\n",
           preA, preB, preC, preD,
           preE, preF);
#else 
    printf("Deleting Mac addr:\n");
    printf("Mac = %x:%x:%x:%x:%x:%x\n",
           preA, preB, preC, preD,
           preE, preF);
#endif
    ret = ix_del_mac(destMac);
    return ret == IX_SUCCESS ? 0 : 1;
}

int lookupPort(const char *argv)
{
    int ret;
    unsigned int preA, preB, preC, preD;
    unsigned int preE, preF;
    ix_uint8 destMac[6];

    if (argv == NULL)
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        ix_ossl_message_log(LOOKUP_PORT_USAGE);
#else
        printf(LOOKUP_PORT_USAGE);
#endif
        return 1;
    }

    if (sscanf(argv, "%x:%x:%x:%x:%x:%x",
            &preA, &preB, &preC, &preD,
            &preE, &preF) != 6)
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        ix_ossl_message_log(LOOKUP_PORT_USAGE);
#else
        printf(LOOKUP_PORT_USAGE);
#endif
        return 1;
    }

    destMac[0] = preA;
    destMac[1] = preB;
    destMac[2] = preC;
    destMac[3] = preD;
    destMac[4] = preE;
    destMac[5] = preF;
   
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_ossl_message_log("Looking up Mac addr:\n");
    ix_ossl_message_log("Mac = %x:%x:%x:%x:%x:%x\n",
           preA, preB, preC, preD,
           preE, preF);
#else
    printf("Looking up Mac addr:\n");
    printf("Mac = %x:%x:%x:%x:%x:%x\n",
           preA, preB, preC, preD,
           preE, preF);


#endif
    ret = ix_lookup_port(destMac);
    return ret == IX_SUCCESS ? 0 : 1;
}


int mac_config(void)
{
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_ossl_message_log("Commands for controlling the MAC filter table:\n");
    ix_ossl_message_log("\taddMac\t\tdelMac\n");
    ix_ossl_message_log("\tlookupPort\n");

    ix_ossl_message_log("Type the name of the command to get usage information\n");


#else
    printf("Commands for controlling the MAC filter table:\n");
    printf("\taddMac\t\tdelMac\n");
    printf("\tlookupPort\n");

    printf("Type the name of the command to get usage information\n");
#endif
    return 0;
}
