/**
 * ===========================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2002 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK for the IXP2X00 Network Processor
 *
 * = FILENAME
 *      route_config.c
 *
 * = DESCRIPTION
 *      Entry points for controlling the IPv4 Route Table from the VxWorks
 *      command line.
 * ===========================================================================
 */


#define IX_ERROR_FILE_IDENT "$Id: route_config.c,v 1.27 2003/12/12 00:44:02 ktseng Exp $"

#if (_IX_OS_TYPE_ != _IX_OS_LINUX_KERNEL_)
#include <string.h>
#endif

#include "ix_cc_error.h"
#include "ix_cc.h"
#include "cc/ix_cc_ipv4.h"
#if defined(IX_CONFIGURATION_oc48_ethernet_ipv6_ingress) || defined(IX_CONFIGURATION_oc192_ethernet_ipv6_ingress)
#include "cc/ix_cc_ipv6.h"
#endif

extern int ix_inet_pton(int, const char *, void *);

#define ROUTE_CONFIG_TIMEOUT 10000

#define ADD_NEXT_HOP_USAGE                                              \
    "addNextHop usage:\n"                                               \
    "addNextHop \"<nextHopID> <bladeID> <l2Index> <portID> <mtu> "      \
    "<flags> <ipAddr> <l2IndexType>\"\n"                                \
    "  where  <l2Index> is  hexadecimal,\n"      \
    "  where <nextHopID> <flags> <l2IndexType> are hexadecimal,\n"      \
    "  <ipAddr> is dotted-decimal, and all other are decimal " \
    "\n"                                               \
    "Example:\n\taddNextHop \"20 2 14 0 1500 0 10.100.30.1 0\"\n"

#define DEL_NEXT_HOP_USAGE                                              \
    "delNextHop usage: delNextHop \"<nextHopID>\",\n"                   \
    "\twhere <nextHopID> is the hex identifier of the next hop "        \
    "to be removed.\n"                                                  \
    "Example:\n\tdelNextHop \"20\"\n"                                   \

#define ADD_ROUTE_USAGE                                                 \
    "addRoute usage: addRoute \"<prefix> <netmask> <nextHopID>\",\n"    \
    "\twhere <prefix> and <netmask> are in dotted-decimal notation,\n"  \
    "\tand <nextHopID> is hexadecimal.\n"                               \
    "Example:\n\taddRoute \"192.168.0.0 255.255.0.0 20\"\n"

#define DEL_ROUTE_USAGE                                                 \
    "delRoute usage: delRoute \"<prefix> <netmask>\",\n"                \
    "\t\twhere <prefix> and <netmask> are given in dotted-decimal "     \
    "format.\n"                                                         \
    "Example:\n\tdelRoute \"192.168.0.0 255.255.0.0\"\n"

#define LOOKUP_ROUTE_USAGE                                              \
    "lookup usage: lookup \"<ip addr>\",\n"                   \
    "\twhere <ip addr> is the dest ip addr to be looked up\n"        \
    "Example:\n\tlookup \"192.168.0.0\"\n"                                   \

/*
** For V6
*/

#if defined(IX_CONFIGURATION_oc48_ethernet_ipv6_ingress) || defined(IX_CONFIGURATION_oc192_ethernet_ipv6_ingress)

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
#define  NUMBER_OF_PORTS 3
#define  NUMBER_OF_BLADES 4 
#endif
    
#define ADD_NEXT_HOP_V6_USAGE                                              \
    "addNextHopV6 usage:\n"                                               \
    "addNextHopV6 \"<nextHopID> <bladeID> <l2Index> <portID> <mtu> "      \
    "<flags> <l2IndexType>\"\n"                                \
    "  where <mtu> is in decimal format,\n" \
    "  and all other values are hexadecimal.\n"                         \
    "Example:\n\taddNextHopV6 \"20 2 14 0 1500 0 0\"\n"

#define DEL_NEXT_HOP_V6_USAGE                                              \
    "delNextHopV6 usage: delNextHopV6 \"<nextHopID>\",\n"                   \
    "\twhere <nextHopID> is the hex identifier of the next hop "        \
    "to be removed.\n"                                                  \
    "Example:\n\tdelNextHopV6 \"20\"\n"                                   \
    "\tdelNextHopV6 \"0x20\"\n"

#define ADD_ROUTE_V6_USAGE                                                 \
    "addRouteV6 usage: addRouteV6 \"<prefix> <prefixlen> <nextHopID>\",\n"    \
    "\twhere <prefix> is in colon separated notation,\n"  \
    "\t<prefixlen> is in decimal and <nextHopID> is hexadecimal.\n"        \
    "Example:\n\taddRouteV6 \"3ffe:0:0:100::0 64 0x20\"\n"

#define DEL_ROUTE_V6_USAGE                                                 \
    "delRouteV6 usage: delRouteV6 \"<prefix> <prefixlen>\",\n"                \
    "\twhere <prefix> is in colon separated notation,\n"  \
    "\t<prefixlen> is in decimal format.\n"                               \
    "Example:\n\tdelRouteV6 \"3ffe:0:0:100::0 64\"\n"

#endif

/* 
 * Services callbacks from add route, del route, add next hop, and del
 * next hop.
 * Wakes up the caller when the callback comes in.
 */
static ix_error route_config_cb(ix_error arg_Result, void *arg_pUserCtx)
{
    ix_ossl_sem_t *pSem = (ix_ossl_sem_t*) arg_pUserCtx;

    if (arg_Result != IX_SUCCESS)
    {
        /*
         * Print out the error for the user's benefit.
         */
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_error_dump(0, arg_Result);
#else
        ix_error_dump(stdout, arg_Result);
#endif
    }


    if (pSem != NULL)
    {
        /*
         * If we're not able to give the semaphore, print the error.  Always
         * return success.
         */
        IX_ERROR_CP(ix_ossl_sem_give(*pSem));
    }


    return IX_SUCCESS;
}

/* 
 * Services callbacks from lookup port
 * Wakes up the caller when the callback comes in.
 */
static ix_error route_lookup_cb(ix_error arg_Result, void *arg_pUserCtx,
            ix_cc_rtmv4_next_hop_info *arg_pNextHopInfo)
{
    if (arg_Result != IX_SUCCESS)
    {
        /*
         * Print out the error for the user's benefit.
         */
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_error_dump(0, arg_Result);
#else
        ix_error_dump(stdout, arg_Result);
#endif
    }
    if (arg_Result == IX_SUCCESS)
    {

        ix_ossl_message_log("Got the following next hop from Route Table:\n");
        ix_ossl_message_log("bladeID: %ld\n", arg_pNextHopInfo->bladeID);
        ix_ossl_message_log("l2Index: 0x%lx\n", arg_pNextHopInfo->l2Index);
        ix_ossl_message_log("portID:  %ld\n", arg_pNextHopInfo->portID);
        ix_ossl_message_log("MTU:     %ld\n", arg_pNextHopInfo->mtu);
        ix_ossl_message_log("flags:   0x%lx\n", arg_pNextHopInfo->flags);
        ix_ossl_message_log("ipAddr:  %u.%u.%u.%u\n",
               (ix_uint8)((arg_pNextHopInfo->ipAddr >> 24) & 0xff),
               (ix_uint8)((arg_pNextHopInfo->ipAddr >> 16) & 0xff),
               (ix_uint8)((arg_pNextHopInfo->ipAddr >> 8)  & 0xff),
               (ix_uint8)((arg_pNextHopInfo->ipAddr >> 0)  & 0xff));
        ix_ossl_message_log("l2IndexType: 0x%lx\n", arg_pNextHopInfo->l2IndexType);

    }
    return IX_SUCCESS;
}


ix_error ix_lookup_route(
    ix_uint32 arg_IpAddr)
    
{
    ix_error err;
    ix_ossl_sem_t sem;

    IX_ERROR_CG(ix_ossl_sem_init(IX_OSSL_SEM_UNAVAILABLE, &sem),
            err, report);
    IX_ERROR_CG(ix_cc_ipv4_async_lookup_route(
                        arg_IpAddr,
                        route_lookup_cb, &sem),
                err, clean_sem);
   
    IX_ERROR_CG(ix_ossl_sem_fini(sem), err, report);

    return IX_SUCCESS;

clean_sem:
    (void) ix_ossl_sem_fini(sem);
report:
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif
    return err;
}



ix_error ix_add_next_hop(
    ix_cc_rtmv4_nhid    arg_NextHopID,
    ix_uint32           arg_BladeID,
    ix_uint32           arg_L2Index,
    ix_uint32           arg_PortID,
    ix_uint32           arg_Mtu,
    ix_uint32           arg_Flags,
    ix_uint32           arg_IpAddr,
    ix_uint32           arg_L2IndexType)
{
    ix_error err;
    ix_ossl_sem_t sem;
    ix_cc_rtmv4_next_hop_info nhInfo;

    nhInfo.bladeID      = arg_BladeID;
    nhInfo.l2Index      = arg_L2Index;
    nhInfo.portID       = arg_PortID;
    nhInfo.mtu          = arg_Mtu;
    nhInfo.flags        = arg_Flags;
    nhInfo.ipAddr       = arg_IpAddr;
    nhInfo.l2IndexType  = arg_L2IndexType;

    IX_ERROR_CR(ix_ossl_sem_init(IX_OSSL_SEM_UNAVAILABLE, &sem));
    IX_ERROR_CG(ix_cc_ipv4_async_add_next_hop(
                                    arg_NextHopID, &nhInfo,
                                    route_config_cb, &sem),
                err, clean_sem);
    IX_ERROR_CG(ix_ossl_sem_take(sem, ROUTE_CONFIG_TIMEOUT), err, clean_sem);
    IX_ERROR_CR(ix_ossl_sem_fini(sem));

    return IX_SUCCESS;

clean_sem:
    (void) ix_ossl_sem_fini(sem);

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif

    return err;
}

ix_error ix_add_route(
    ix_uint32           arg_Prefix,
    ix_uint32           arg_NetMask,
    ix_uint32           arg_NextHopID)
{
    ix_error err;
    ix_ossl_sem_t sem;

    IX_ERROR_CG(ix_ossl_sem_init(IX_OSSL_SEM_UNAVAILABLE, &sem),
            err, report);
    IX_ERROR_CG(ix_cc_ipv4_async_add_route(
                        arg_Prefix, arg_NetMask, arg_NextHopID,
                        route_config_cb, &sem),
                err, clean_sem);
    IX_ERROR_CG(ix_ossl_sem_take(sem, ROUTE_CONFIG_TIMEOUT), err, clean_sem);
    IX_ERROR_CG(ix_ossl_sem_fini(sem), err, report);

    return IX_SUCCESS;

clean_sem:
    (void) ix_ossl_sem_fini(sem);
report:

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif

    return err;
}

ix_error ix_del_route(
    ix_uint32           arg_Prefix,
    ix_uint32           arg_NetMask)
{
    ix_error err;
    ix_ossl_sem_t sem;

    IX_ERROR_CR(ix_ossl_sem_init(IX_OSSL_SEM_UNAVAILABLE, &sem));
    IX_ERROR_CG(ix_cc_ipv4_async_delete_route(
                        arg_Prefix, arg_NetMask,
                        route_config_cb, &sem),
                err, clean_sem);
    IX_ERROR_CG(ix_ossl_sem_take(sem, ROUTE_CONFIG_TIMEOUT), err, clean_sem);
    IX_ERROR_CR(ix_ossl_sem_fini(sem));

    return IX_SUCCESS;

clean_sem:
    (void) ix_ossl_sem_fini(sem);

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif

    return err;
}

ix_error ix_del_next_hop(
    ix_uint32           arg_NextHopID)
{
    ix_error err;
    ix_ossl_sem_t sem;

    IX_ERROR_CR(ix_ossl_sem_init(IX_OSSL_SEM_UNAVAILABLE, &sem));
    IX_ERROR_CG(ix_cc_ipv4_async_delete_next_hop(
                        arg_NextHopID,
                        route_config_cb, &sem),
                err, clean_sem);
    IX_ERROR_CG(ix_ossl_sem_take(sem, ROUTE_CONFIG_TIMEOUT), err, clean_sem);
    IX_ERROR_CR(ix_ossl_sem_fini(sem));

    return IX_SUCCESS;

clean_sem:
    (void) ix_ossl_sem_fini(sem);

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif

    return err;
}

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

/* IGK0100023119: There are some problems with sscanf function in kernel, %hX doesn't accept letters */

static short int str_to_hex(char c)
{
    switch(c)
    {
    case '0':return 0;
    case '1':return 1;
    case '2':return 2;
    case '3':return 3;
    case '4':return 4;
    case '5':return 5;
    case '6':return 6;
    case '7':return 7;
    case '8':return 8;
    case '9':return 9;
    case 'a':
    case 'A':
        return 10;
    case 'b':
    case 'B':
        return 11;
    case 'c':
    case 'C':
        return 12;
    case 'd':
    case 'D':
        return 13;
    case 'e':
    case 'E':
        return 14;
    case 'f':
    case 'F':
        return 15;
    default:return 0;
    }
}

static unsigned int parse_alfa_to_hex(char* str)
{
    char* chr;
    unsigned int res = 0;
    chr = str;
    while(*chr != 0)
    {
        res = (res << 4) + str_to_hex(*chr);
        chr++;
    }
    return res;
}

#endif /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */


int addNextHop(const char *argv)
{
    int ret;
    ix_cc_rtmv4_nhid nextHopID;
    ix_uint32 bladeID;
    ix_uint32 l2Index;
    ix_uint32 portID;
    ix_uint32 mtu;
    ix_uint32 flags;
    ix_uint32 ipAddr;
    ix_uint32 l2IndexType;
    unsigned int ipA, ipB, ipC, ipD;

    if (argv == NULL)
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        ix_ossl_message_log(ADD_NEXT_HOP_USAGE);
#else
        printf(ADD_NEXT_HOP_USAGE);
#endif
        return 1;
    }

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

/* IGK0100023119: There are some problems with sscanf function in kernel, %hX doesn't accept letters */
/* This code may be removed once this kernel sscanf problem s fixed */
    {
      char nextHopID_str[10];
      char flags_str[10];
      char l2IndexType_str[10];
      char l2Index_str[10];

      if (sscanf(argv, "%s %li %s %li %li %s %d.%d.%d.%d %s",
		 nextHopID_str, &bladeID, l2Index_str, &portID, &mtu, flags_str,
		 &ipA, &ipB, &ipC, &ipD, l2IndexType_str) != 11)
	{
	  ix_ossl_message_log(ADD_NEXT_HOP_USAGE);
	  return 1;
	}
      nextHopID = parse_alfa_to_hex(nextHopID_str);
      flags = parse_alfa_to_hex(flags_str);
      l2IndexType = parse_alfa_to_hex(l2IndexType_str);
      l2Index = parse_alfa_to_hex(l2Index_str);
    }

#else /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

    if (sscanf(argv, "%lX %li %lx %li %li %lX %d.%d.%d.%d %lx",
            &nextHopID, &bladeID, &l2Index, &portID, &mtu, &flags,
            &ipA, &ipB, &ipC, &ipD, &l2IndexType) != 11)
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_ossl_message_log(ADD_NEXT_HOP_USAGE);
#else
        printf(ADD_NEXT_HOP_USAGE);
#endif
        return 1;
    }

#endif /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

    if ((ipA > 255) || (ipB > 255) || (ipC > 255) | (ipD > 255))
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        ix_ossl_message_log("Invalid IP address: %d.%d.%d.%d\n",
            ipA, ipB, ipC, ipD);
#else
        printf("Invalid IP address: %d.%d.%d.%d\n",
            ipA, ipB, ipC, ipD);
#endif
        return 1;
    }

    ipAddr = (ipA << 24) | (ipB << 16) | (ipC << 8) | ipD;

    ix_ossl_message_log("Adding next hop:\n");
    ix_ossl_message_log("nextHopID = 0x%08lX\n"
           "bladeID = 0x%02lX, l2Index = 0x%04lX, portID = 0x%04lX\n"
           "MTU = %ld, flags = 0x%02lX, ipAddr = %d.%d.%d.%d, l2IndexType = 0x%lx\n",
           nextHopID, bladeID, l2Index, portID, mtu, flags,
           ipA, ipB, ipC, ipD, l2IndexType);


    ret = ix_add_next_hop(nextHopID, bladeID, l2Index, portID,
            mtu, flags, ipAddr, l2IndexType);
    return ret == IX_SUCCESS ? 0 : 1;
}

int addRoute(const char *argv)
{
    int ret;
    ix_uint32 prefix;
    ix_uint32 netMask;
    ix_cc_rtmv4_nhid nextHopID;
    unsigned int preA, preB, preC, preD;
    unsigned int netA, netB, netC, netD;

    if (argv == NULL)
    {

        ix_ossl_message_log(ADD_ROUTE_USAGE);

        return 1;
    }

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

/* IGK0100023119: There are some problems with sscanf function in kernel, %hX doesn't accept letters */
/* This code may be removed once this kernel sscanf problem s fixed */
    {
      char nextHopID_str[10];

      if (sscanf(argv, "%d.%d.%d.%d %d.%d.%d.%d %s",
		 &preA, &preB, &preC, &preD,
		 &netA, &netB, &netC, &netD,
		 nextHopID_str) != 9)
	{
	  ix_ossl_message_log(ADD_ROUTE_USAGE);
	  return 1;
	}
      nextHopID = parse_alfa_to_hex(nextHopID_str);
    }

#else /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

    if (sscanf(argv, "%d.%d.%d.%d %d.%d.%d.%d %lX",
            &preA, &preB, &preC, &preD,
            &netA, &netB, &netC, &netD,
            &nextHopID) != 9)
    {
        ix_ossl_message_log(ADD_ROUTE_USAGE);

        return 1;
    }

#endif /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

    if ((preA > 255) || (preB > 255) || (preC > 255) | (preD > 255))
    {

        ix_ossl_message_log("Invalid Prefix address: %d.%d.%d.%d\n",
            preA, preB, preC, preD);

        return 1;
    }

    if ((netA > 255) || (netB > 255) || (netC > 255) | (netD > 255))
    {
        ix_ossl_message_log("Invalid Network Mask: %d.%d.%d.%d\n",
            netA, netB, netC, netD);

        return 1;
    }

    prefix = (preA << 24) | (preB << 16) | (preC << 8) | preD;
    netMask = (netA << 24) | (netB << 16) | (netC << 8) | netD;

    ix_ossl_message_log("Adding route:\n");
    ix_ossl_message_log("Prefix = %d.%d.%d.%d, NetMask = %d.%d.%d.%d, nextHopID = 0x%08lX\n",
           preA, preB, preC, preD,
           netA, netB, netC, netD,
           nextHopID);

    ret = ix_add_route(prefix, netMask, nextHopID);
    return ret == IX_SUCCESS ? 0 : 1;
}

int delRoute(const char *argv)
{
    int ret;
    ix_uint32 prefix;
    ix_uint32 netMask;
    unsigned int preA, preB, preC, preD;
    unsigned int netA, netB, netC, netD;

    if (argv == NULL)
    {

        ix_ossl_message_log(DEL_ROUTE_USAGE);

        return 1;
    }

    if (sscanf(argv, "%d.%d.%d.%d %d.%d.%d.%d",
            &preA, &preB, &preC, &preD,
            &netA, &netB, &netC, &netD) != 8)
    {

        ix_ossl_message_log(DEL_ROUTE_USAGE);
        return 1;
    }


    if ((preA > 255) || (preB > 255) || (preC > 255) | (preD > 255))
    {

        ix_ossl_message_log("Invalid Prefix address: %d.%d.%d.%d\n",
            preA, preB, preC, preD);

        return 1;
    }

    if ((netA > 255) || (netB > 255) || (netC > 255) | (netD > 255))
    {
        ix_ossl_message_log("Invalid Network Mask: %d.%d.%d.%d\n",
            netA, netB, netC, netD);

        return 1;
    }

    prefix = (preA << 24) | (preB << 16) | (preC << 8) | preD;
    netMask = (netA << 24) | (netB << 16) | (netC << 8) | netD;

    ix_ossl_message_log("Deleting route:\n");
    ix_ossl_message_log("Prefix = %d.%d.%d.%d, NetMask = %d.%d.%d.%d\n",
           preA, preB, preC, preD,
           netA, netB, netC, netD);

    ret = ix_del_route(prefix, netMask);
    return ret == IX_SUCCESS ? 0 : 1;
}


int delNextHop(const char *argv)
{
    int ret;
    ix_cc_rtmv4_nhid nextHopID;

    if (argv == NULL)
    {

        ix_ossl_message_log(DEL_NEXT_HOP_USAGE);

        return 1;
    }

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

/* IGK0100023119: There are some problems with sscanf function in kernel, %hX doesn't accept letters */
/* This code may be removed once this kernel sscanf problem s fixed */
    {
      char nextHopID_str[10];

      if (sscanf(argv, "%s", nextHopID_str) != 1)
	{
	  ix_ossl_message_log(DEL_NEXT_HOP_USAGE);
	  return 1;
	}
      nextHopID = parse_alfa_to_hex(nextHopID_str);
    }

#else /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */

    if (sscanf(argv, "%lX", &nextHopID) != 1)
    {

        ix_ossl_message_log(DEL_NEXT_HOP_USAGE);
        return 1;
    }

#endif /* (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_) */


    ix_ossl_message_log("Deleting nextHop:\n");
    ix_ossl_message_log("nextHopID = 0x%08lX\n", nextHopID);

    ret = ix_del_next_hop(nextHopID);
    return ret == IX_SUCCESS ? 0 : 1;
}


int lookupRoute(const char *argv)
{
    ix_uint32 prefix;
    unsigned int preA, preB, preC, preD;
    int ret;
    if (argv == NULL)
    {

        ix_ossl_message_log(LOOKUP_ROUTE_USAGE);
        return 1;
    }

    if (sscanf(argv, "%d.%d.%d.%d",
            &preA, &preB, &preC, &preD) != 4)
    {

        ix_ossl_message_log(LOOKUP_ROUTE_USAGE);
        return 1;
    }


    if ((preA > 255) || (preB > 255) || (preC > 255) | (preD > 255))
    {

        ix_ossl_message_log("Invalid Prefix address: %d.%d.%d.%d\n",
            preA, preB, preC, preD);
        return 1;
    }

    
    prefix = (preA << 24) | (preB << 16) | (preC << 8) | preD;
    ix_ossl_message_log("Looking up route:\n");
    ix_ossl_message_log("Prefix = %d.%d.%d.%d\n",
           preA, preB, preC, preD);

    ret = ix_lookup_route(prefix);
    return ret == IX_SUCCESS ? 0 : 1;
}

static ix_error dump_rtm_cb(
    ix_error arg_Result, void *arg_pUserCtx,
    ix_cc_ipv4_dump_data *arg_pDumpData)
{
    ix_ossl_sem_t *pSem = (ix_ossl_sem_t*) arg_pUserCtx;
    char line[81];
    ix_uint32 i, copyLen;
    ix_int32 len = arg_pDumpData->size;

    if (arg_Result == IX_SUCCESS)
    {
        /*
         * The loop below is for displaying dumpData in fragments, because
         * the printf buffer could be smaller than the size of dumpData and
         * not all routes are displayed in that case.
         */
        i = 0;
        while (len > 0)
        {
            copyLen = (len > 80) ? 80 : len;
            
            strncpy(line,
                    (char *)((ix_uint32)arg_pDumpData->pBuffer + i*80),
                    copyLen);
            line[copyLen] = '\0';
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

            ix_ossl_message_log("%s", line);
#else
            printf("%s", line);
#endif
            len -= 80;
            i++;
        }
    }
    else
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_error_dump(0, arg_Result);
#else
        ix_error_dump(stdout, arg_Result);
#endif
    }

    /*
     * Signal the thread waiting for dump finish
     */
    if (pSem != NULL)
    {
        /*
         * If we're not able to give the semaphore, print the error.  Always
         * return success.
         */
        IX_ERROR_CP(ix_ossl_sem_give(*pSem));
    }
    
    return IX_SUCCESS;
}

int dumpRtm(void)
{
    ix_ossl_sem_t sem;
    ix_error err;

    IX_ERROR_CG(ix_ossl_sem_init(IX_OSSL_SEM_UNAVAILABLE, &sem), err, report);

    IX_ERROR_CG(ix_cc_ipv4_async_dump_next_hops(dump_rtm_cb, &sem),
                err, clean_sem);

    IX_ERROR_CG(ix_ossl_sem_take(sem, OS_WAIT_FOREVER),
                err, clean_sem);

    ix_ossl_message_log("\n\n");

    IX_ERROR_CG(ix_cc_ipv4_async_dump_routes(dump_rtm_cb, &sem),
                err, clean_sem);

    IX_ERROR_CG(ix_ossl_sem_take(sem, OS_WAIT_FOREVER),
                err, clean_sem);
    
    IX_ERROR_CG(ix_ossl_sem_fini(sem), err, report);

    return 0;

clean_sem:
    (void) ix_ossl_sem_fini(sem);
report:

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif

    return 1;
}

int purgeRtm(void)
{
    ix_error err;

    err = ix_cc_ipv4_async_purge_rtm();
    if (err != IX_SUCCESS)
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_error_dump(0, err);
#else
        ix_error_dump(stdout, err);
#endif
        return 1;
    }
    return 0;
}

int purgeRoutes(void)
{
    ix_error err;

    err = ix_cc_ipv4_async_purge_routes();
    if (err != IX_SUCCESS)
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_error_dump(0, err);
#else
        ix_error_dump(stdout, err);
#endif
        return 1;
    }
    return 0;
}

int route_config(void)
{

    ix_ossl_message_log("Commands for controlling the IPv4's Route Table Manager:\n");
    ix_ossl_message_log("\taddNextHop\t\tdelNextHop\n");
    ix_ossl_message_log("\taddRoute\t\tdelRoute\n");
    ix_ossl_message_log("\tpurgeRoutes\t\tpurgeRtm\n");
    ix_ossl_message_log("\tdumpRtm\n");
    ix_ossl_message_log("\tlookupRoute\n");
#if defined(IX_CONFIGURATION_oc48_ethernet_ipv6_ingress) || defined(IX_CONFIGURATION_oc192_ethernet_ipv6_ingress)
    ix_ossl_message_log("Commands for controlling the IPv6 Route Table Manager:\n");
    ix_ossl_message_log("\taddNextHopV6\t\tdelNextHopV6\n");
    ix_ossl_message_log("\taddRouteV6\t\tdelRouteV6\n");
#endif

    ix_ossl_message_log("Type the name of the command to get usage information\n");

    return 0;
}

#if defined(IX_CONFIGURATION_oc48_ethernet_ipv6_ingress) || defined(IX_CONFIGURATION_oc192_ethernet_ipv6_ingress)

/* 
 * Services callbacks from add route, del route, add next hop, and del
 * next hop.
 * Wakes up the caller when the callback comes in.
 */
static ix_error route_config_v6_cb(ix_error arg_Result, void *arg_pUserCtx)
{
    ix_ossl_sem_t *pSem = (ix_ossl_sem_t*) arg_pUserCtx;
    ix_ossl_message_log("In the IPv6 Callback; arg_Result = %lx\n", arg_Result);

    if (arg_Result != IX_SUCCESS)
    {
        /*
         * Print out the error for the user's benefit.
         */
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_error_dump(0, arg_Result);
#else
        ix_error_dump(stdout, arg_Result);
#endif
    }

    if (pSem != NULL)
    {
        /*
         * If we're not able to give the semaphore, print the error.  Always
         * return success.
         */
        IX_ERROR_CP(ix_ossl_sem_give(*pSem));
    }

    return IX_SUCCESS;
}

ix_error ix_add_next_hop_v6(
    ix_cc_rtmv6_nhid    arg_NextHopID,
    ix_uint32           arg_BladeID,
    ix_uint32           arg_L2Index,
    ix_uint32           arg_PortID,
    ix_uint32           arg_Mtu,
    ix_uint32           arg_Flags,
    ix_uint32           arg_IpAddr,
    ix_uint32           arg_L2IndexType)
{
    ix_error err;
    ix_ossl_sem_t sem;
    ix_cc_rtmv6_next_hop_info nhInfo;
    
    nhInfo.bladeID      = arg_BladeID;
    nhInfo.l2Index      = arg_L2Index;
    nhInfo.portID       = arg_PortID;
    nhInfo.mtu          = arg_Mtu;
    nhInfo.flags        = arg_Flags;
    nhInfo.ipAddr       = arg_IpAddr;
    nhInfo.l2IndexType  = arg_L2IndexType;

    IX_ERROR_CR(ix_ossl_sem_init(IX_OSSL_SEM_UNAVAILABLE, &sem));
    IX_ERROR_CG(ix_cc_ipv6_async_add_next_hop(
                                    arg_NextHopID, 1, &nhInfo,
                                    route_config_v6_cb, &sem),
                err, clean_sem);
    IX_ERROR_CG(ix_ossl_sem_take(sem, ROUTE_CONFIG_TIMEOUT), err, clean_sem);
    IX_ERROR_CR(ix_ossl_sem_fini(sem));

    return IX_SUCCESS;

clean_sem:
    (void) ix_ossl_sem_fini(sem);

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif

    return err;
}

ix_error ix_del_next_hop_v6(
    ix_uint32           arg_NextHopID)
{
    ix_error err;
    ix_ossl_sem_t sem;

    IX_ERROR_CR(ix_ossl_sem_init(IX_OSSL_SEM_UNAVAILABLE, &sem));
    IX_ERROR_CG(ix_cc_ipv6_async_delete_next_hop(
                        arg_NextHopID,
                        route_config_v6_cb, &sem),
                err, clean_sem);
    IX_ERROR_CG(ix_ossl_sem_take(sem, ROUTE_CONFIG_TIMEOUT), err, clean_sem);
    IX_ERROR_CR(ix_ossl_sem_fini(sem));

    return IX_SUCCESS;

clean_sem:
    (void) ix_ossl_sem_fini(sem);

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif

    return err;
}

ix_error ix_add_route_v6(
    ix_uint128          arg_Prefix,
    ix_uint32           arg_PrefixLen,
    ix_uint32           arg_NextHopID)
{
    ix_error err;
    ix_ossl_sem_t sem;


    IX_ERROR_CG(ix_ossl_sem_init(IX_OSSL_SEM_UNAVAILABLE, &sem),
            err, report);
    IX_ERROR_CG(ix_cc_ipv6_async_add_prefix(
                        arg_Prefix, arg_PrefixLen, arg_NextHopID,
                        route_config_v6_cb, &sem),
                err, clean_sem);
    IX_ERROR_CG(ix_ossl_sem_take(sem, ROUTE_CONFIG_TIMEOUT), err, clean_sem);
    IX_ERROR_CG(ix_ossl_sem_fini(sem), err, report);

    return IX_SUCCESS;

clean_sem:
    (void) ix_ossl_sem_fini(sem);
report:

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif

    return err;
}

ix_error ix_del_route_v6(
    ix_uint128          arg_Prefix,
    ix_uint32           arg_PrefixLen)
{
    ix_error err;
    ix_ossl_sem_t sem;

    IX_ERROR_CR(ix_ossl_sem_init(IX_OSSL_SEM_UNAVAILABLE, &sem));
    IX_ERROR_CG(ix_cc_ipv6_async_delete_prefix(
                        arg_Prefix, arg_PrefixLen,
                        route_config_cb, &sem),
                err, clean_sem);
    IX_ERROR_CG(ix_ossl_sem_take(sem, ROUTE_CONFIG_TIMEOUT), err, clean_sem);
    IX_ERROR_CR(ix_ossl_sem_fini(sem));

    return IX_SUCCESS;

clean_sem:
    (void) ix_ossl_sem_fini(sem);

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif

    return err;
}

int addNextHopV6(const char *argv)
{
    int ret;
    ix_cc_rtmv6_nhid nextHopID;
    ix_uint32 bladeID;
    ix_uint32 l2Index;
    ix_uint32 portID;
    ix_uint32 mtu;
    ix_uint32 flags;
    ix_uint32 l2IndexType;

    if (argv == NULL)
    {

        ix_ossl_message_log(ADD_NEXT_HOP_V6_USAGE);

        return 1;
    }

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
/* IGK0100023119: There are some problems with sscanf function in kernel, 
 * %hX doesn't accept letters */
/* This code may be removed once this kernel sscanf problem s fixed */
    {
      char nextHopID_str[10];
      char flags_str[10];
      char l2IndexType_str[10];
      char l2Index_str[10];

      if (sscanf(argv, "%s %li %s %li %li %s %s",
		 nextHopID_str, &bladeID, l2Index_str, &portID, &mtu, flags_str,
		 l2IndexType_str) != 7)
	{
          ix_ossl_message_log(ADD_NEXT_HOP_V6_USAGE);
  	      return 1;
	}
      nextHopID = parse_alfa_to_hex(nextHopID_str);
      flags = parse_alfa_to_hex(flags_str);
      l2IndexType = parse_alfa_to_hex(l2IndexType_str);
      l2Index = parse_alfa_to_hex(l2Index_str);
    }

#else /* _IX_OS_LINUX_KERNEL_ */

    if (sscanf(argv, "%lX %lX %lX %lX %ld %lX %lx",
            &nextHopID, &bladeID, &l2Index, &portID, &mtu, &flags,
            &l2IndexType) != 7)
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_ossl_message_log(ADD_NEXT_HOP_V6_USAGE);
#else
        printf(ADD_NEXT_HOP_V6_USAGE);
#endif
        return 1;
    }
#endif /* _IX_OS_LINUX_KERNEL_ */


    ix_ossl_message_log("Adding next hop:\n");
    ix_ossl_message_log("nextHopID = 0x%08lX\n"
           "bladeID = 0x%02lX, l2Index = 0x%04lX, portID = 0x%04lX\n"
           "MTU = %ld, flags = 0x%02lX, l2IndexType = %lx\n",
           nextHopID, bladeID, l2Index, portID, mtu, flags,
           l2IndexType);

    ret = ix_add_next_hop_v6(nextHopID, bladeID, l2Index, portID,
            mtu, flags, 0, l2IndexType);
    return ret == IX_SUCCESS ? 0 : 1;
}

int delNextHopV6(const char *argv)
{
    int ret;
    ix_cc_rtmv6_nhid nextHopID;

    if (argv == NULL)
    {
        ix_ossl_message_log(DEL_NEXT_HOP_V6_USAGE);

        return 1;
    }

    if (sscanf(argv, "%lX", &nextHopID) != 1)
    {

        ix_ossl_message_log(DEL_NEXT_HOP_V6_USAGE);
        return 1;
    }

    ix_ossl_message_log("Deleting nextHop:\n");
    ix_ossl_message_log("nextHopID = 0x%08lX\n", nextHopID);

    ret = ix_del_next_hop_v6(nextHopID);
    return ret == IX_SUCCESS ? 0 : 1;
}

int addRouteV6(const char *argv)
{
    int ret;
    int i;
    int prefixLen;
    ix_cc_rtmv6_nhid nextHopID;
    ix_uint128 ipv6Addr;
    char v6addrStr[50];

    if (argv == NULL)
    {

        ix_ossl_message_log(ADD_ROUTE_V6_USAGE);
        return 1;
    }

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

/* IGK0100023119: There are some problems with sscanf function in kernel, %hX doesn't accept letters */
/* This code may be removed once this kernel sscanf problem s fixed */
    {
       char nextHopID_str[10];

       if (sscanf(argv, "%s %d %s", v6addrStr, &prefixLen, nextHopID_str) != 3)
       {
          ix_ossl_message_log(ADD_ROUTE_V6_USAGE);
          return 1;
       }
       nextHopID = parse_alfa_to_hex(nextHopID_str);
    }
#else /* _IX_OS_LINUX_KERNEL_ */
    if (sscanf(argv, "%s %d %lX", v6addrStr, &prefixLen, &nextHopID) != 3)
    {

        ix_ossl_message_log(ADD_ROUTE_V6_USAGE);
        return 1;
    }
#endif /* _IX_OS_LINUX_KERNEL_ */

    if (ix_inet_pton(IX_AF_INET6, v6addrStr, &ipv6Addr) != 1)
    {

        ix_ossl_message_log("IPv6 Address is Invalid!\n");

        return -1;
    }

    for (i = 0; i < 4; i++)
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        ix_ossl_message_log("ipaddr[%d] = %lx\n", i, ipv6Addr.aUint32[i]);
#else
        printf("ipaddr[%d] = %lx\n", i, ipv6Addr.aUint32[i]);
#endif
    }

    ix_ossl_message_log("Adding route:\n");
    ix_ossl_message_log("Prefix = %s, Prefix Len = %d, nextHopID = 0x%08lX\n",
           v6addrStr, prefixLen, nextHopID);

    ret = ix_add_route_v6(ipv6Addr, prefixLen, nextHopID);
    return ret == IX_SUCCESS ? 0 : 1;
}

int delRouteV6(const char *argv)
{
    int ret;
    int prefixLen;
    ix_uint128 ipv6Addr;
    char v6addrStr[50];

    if (argv == NULL)
    {

        ix_ossl_message_log(DEL_ROUTE_V6_USAGE);
        return 1;
    }

    if (sscanf(argv, "%s %d",
            v6addrStr, &prefixLen) != 2)
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        ix_ossl_message_log(DEL_ROUTE_V6_USAGE);
#else
        printf(DEL_ROUTE_V6_USAGE);
#endif
        return 1;
    }

    if (ix_inet_pton(IX_AF_INET6, v6addrStr, &ipv6Addr) != 1)
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

        ix_ossl_message_log("IPv6 Address is Invalid!\n");
#else
        printf("IPv6 Address is Invalid!\n");
#endif
        return -1;
    }

    ix_ossl_message_log("Deleting route:\n");
    ix_ossl_message_log("Prefix = %s, PrefixLen = %d\n", v6addrStr, prefixLen);

    ret = ix_del_route_v6(ipv6Addr, prefixLen);

    return ret == IX_SUCCESS ? 0 : 1;
}

static ix_error dump_rtmv6_cb(
    ix_error arg_Result, void *arg_pUserCtx,
    ix_cc_ipv6_dump_data *arg_pDumpData)
{
    ix_ossl_sem_t *pSem = (ix_ossl_sem_t*) arg_pUserCtx;
    char line[81];
    ix_uint32 i, copyLen;
    ix_int32 len = arg_pDumpData->size;

    if (arg_Result == IX_SUCCESS)
    {
        /*
         * The loop below is for displaying dumpData in fragments, because
         * the printf buffer could be smaller than the size of dumpData and
         * not all routes are displayed in that case.
         */
        i = 0;
        while (len > 0)
        {
            copyLen = (len > 80) ? 80 : len;
            
            strncpy(line,
                    (char *)((ix_uint32)arg_pDumpData->pBuffer + i*80),
                    copyLen);
            line[copyLen] = '\0';
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
            ix_ossl_message_log("%s", line);
#else
            printf("%s", line);
#endif
            len -= 80;
            i++;
        }
    }
    else
    {
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_error_dump(0, arg_Result);
#else
        ix_error_dump(stdout, arg_Result);
#endif
    }

    if (pSem != NULL)
    {
        /*
         * If we're not able to give the semaphore, print the error.  Always
         * return success.
         */
        IX_ERROR_CP(ix_ossl_sem_give(*pSem));
    }

    return IX_SUCCESS;
}

int dumpRtmV6(void)
{
    ix_ossl_sem_t sem;
    ix_error err;

    IX_ERROR_CG(ix_ossl_sem_init(IX_OSSL_SEM_UNAVAILABLE, &sem), err, report);

    IX_ERROR_CG(ix_cc_ipv6_async_dump_next_hops(dump_rtmv6_cb, &sem),
                err, clean_sem);
    IX_ERROR_CG(ix_ossl_sem_take(sem, ROUTE_CONFIG_TIMEOUT), err, clean_sem);

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

    ix_ossl_message_log("\n\n");
#else
    printf("\n\n");
#endif
    IX_ERROR_CG(ix_cc_ipv6_async_dump_prefixes(dump_rtmv6_cb, &sem),
                err, clean_sem);
    IX_ERROR_CG(ix_ossl_sem_take(sem, ROUTE_CONFIG_TIMEOUT), err, clean_sem);

    IX_ERROR_CG(ix_ossl_sem_fini(sem), err, report);

    return 0;

clean_sem:
    (void) ix_ossl_sem_fini(sem);
report:

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif

    return 1;
}

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
/**
 * NAME: get_statistics_callback
 *
 * DESCRIPTION: statistics callback 
 *
 * @Param:  - IN  arg_Result,arg_Result                             
 * @Param: - OUT arg_pBuffer
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 *
 *
 */

/* get_statistics_callback */
static ix_error get_statistics_callback(ix_error arg_Result, void *arg_pUserCtx,
		                       ix_cc_ipv6_stats_data *arg_pBuffer)
{
    ix_uint32 i =0, bladeIndex =0, portIndex =0;
    ix_ossl_sem_t *pSem = (ix_ossl_sem_t*) arg_pUserCtx;

    ix_ossl_message_log("inside get_statistics_callback \n");
    if (arg_Result != IX_SUCCESS)
    {
        /*
         * Print out the error for the user's benefit.
         */
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        ix_error_dump(0, arg_Result);
#else
        ix_error_dump(stdout, arg_Result);
#endif
    }
    else
    {
	  for(i=0; i< NUMBER_OF_PORTS;i++)
	  {
  ix_ossl_message_log("INCOMMING STATISTICS---  %ld\n", i );
		  ix_ossl_message_log("ipv6MbInReceives =%lld \n", arg_pBuffer->ipv6InStats[i].ipv6MbInReceives);
		  ix_ossl_message_log("ipv6MbInOctets =%lld \n", arg_pBuffer->ipv6InStats[i].ipv6MbInOctets);
		  ix_ossl_message_log("ipv6MbInForwDatagrams =%ld \n", arg_pBuffer->ipv6InStats[i].ipv6MbInForwDatagrams);
		  ix_ossl_message_log("ipv6MbInAddrErrors =%ld \n", arg_pBuffer->ipv6InStats[i].ipv6MbInAddrErrors);
		  ix_ossl_message_log("ipv6MbInTruncated =%ld \n", arg_pBuffer->ipv6InStats[i].ipv6MbInTruncated);
		  ix_ossl_message_log("ipv6CoreMbInHdrErrors =%ld \n", arg_pBuffer->ipv6InStats[i].ipv6CoreMbInHdrErrors);
		  ix_ossl_message_log("ipv6CoreInNoRoutes =%ld \n", arg_pBuffer->ipv6InStats[i].ipv6CoreInNoRoutes);
		  ix_ossl_message_log("ipv6CoreInUnknownProtos =%ld \n", arg_pBuffer->ipv6InStats[i].ipv6CoreInUnknownProtos);
		  ix_ossl_message_log("ipv6CoreInDiscards =%ld \n", arg_pBuffer->ipv6InStats[i].ipv6CoreInDiscards);
		  ix_ossl_message_log("ipv6CoreInDelivers =%ld \n", arg_pBuffer->ipv6InStats[i].ipv6CoreInDelivers);
		  ix_ossl_message_log("ipv6CoreInMulticast =%lld \n", arg_pBuffer->ipv6InStats[i].ipv6CoreInMulticast);
		  ix_ossl_message_log("ipv6CoreInMulticastOctets =%lld \n", arg_pBuffer->ipv6InStats[i].ipv6CoreInMulticastOctets);
	  }

  ix_ossl_message_log("OUTGOING STATISTICS  \n" );
    for (bladeIndex = 0; bladeIndex < NUMBER_OF_BLADES; bladeIndex++)
    {
  ix_ossl_message_log("BLADE %ld  \n",bladeIndex);
        /* Get out stats for all ports on this bladeIndex */
        for (portIndex = 0; portIndex < NUMBER_OF_PORTS;
                                                             portIndex++)
        {
            i = (NUMBER_OF_BLADES  * bladeIndex) + portIndex;

            ix_ossl_message_log("ipv6MbOutForwDatagrams = %ld \n", arg_pBuffer->ipv6OutStats[i].ipv6MbOutForwDatagrams);
            ix_ossl_message_log("ipv6MbOutTransmits =%lld \n", arg_pBuffer->ipv6OutStats[i].ipv6MbOutTransmits);
            ix_ossl_message_log("ipv6MbOutOctets =%lld \n", arg_pBuffer->ipv6OutStats[i].ipv6MbOutOctets);
            ix_ossl_message_log("ipv6CoreOutRequests =%ld \n", arg_pBuffer->ipv6OutStats[i].ipv6CoreOutRequests);
            ix_ossl_message_log("ipv6CoreOutNoRoutes =%ld \n", arg_pBuffer->ipv6OutStats[i].ipv6CoreOutNoRoutes);
            ix_ossl_message_log("ipv6CoreOutDiscards  =%ld \n", arg_pBuffer->ipv6OutStats[i].ipv6CoreOutDiscards);
            ix_ossl_message_log("ipv6CoreOutMulticast  =%lld \n", arg_pBuffer->ipv6OutStats[i].ipv6CoreOutMulticast);
            ix_ossl_message_log("ipv6CoreOutMulticastOctets =%lld \n", arg_pBuffer->ipv6OutStats[i].ipv6CoreOutMulticastOctets);
       }
       ix_ossl_message_log("TOTAL  STATISTICS NUMBER %ld  \n", i );
   }

   ix_ossl_message_log("ipv6NumberOfRoutes =%ld \n", arg_pBuffer->ipv6NumberOfRoutes);
   ix_ossl_message_log("ipv6NumberOfNextHops =%ld \n", arg_pBuffer->ipv6NumberOfNextHops);

    }

    ix_ossl_message_log("get_statistics_callback SUCCESS \n");
    if (pSem != NULL)
    {
        /*
         * If we're not able to give the semaphore, print the error.  Always
         * return success.
         */
        IX_ERROR_CP(ix_ossl_sem_give(*pSem));
    }
    return IX_SUCCESS;
}

/**
 * NAME: getStatsV6
 *
 * DESCRIPTION: This function calls ipv6 statistics callback 
 *
 *
 * @Return: IX_SUCCESS if successful or a valid error token for failure.
 *
 *
 *
 */
ix_error getStatsV6(void)
{
    ix_error err;
    ix_ossl_sem_t sem;

    ix_ossl_message_log("inside getStatsV6 FNNN\n");
    IX_ERROR_CG(ix_ossl_sem_init(IX_OSSL_SEM_UNAVAILABLE, &sem),
            err, report);
    IX_ERROR_CG(ix_cc_ipv6_async_get_statistics(
                        get_statistics_callback, &sem),
                err, clean_sem);

    ix_ossl_message_log("AFTER CALL back completed \n");
    IX_ERROR_CG(ix_ossl_sem_take(sem, ROUTE_CONFIG_TIMEOUT), err, clean_sem);
    IX_ERROR_CG(ix_ossl_sem_fini(sem), err, report);

    return IX_SUCCESS;

clean_sem:
    (void) ix_ossl_sem_fini(sem);
report:

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
    ix_error_dump(0, err);
#else
    ix_error_dump(stdout, err);
#endif

    return err;
}
#endif /* IX_OS_LINUX_KERNEL */

#endif
