/**
 * ============================================================================
 * = COPYRIGHT
 *              INTEL CORPORATION PROPRIETARY INFORMATION
 *   This software is supplied under the terms of a license agreement or
 *   nondisclosure agreement with Intel Corporation and may not be copied 
 *   or disclosed except in accordance with the terms in that agreement.
 *      Copyright (C) 2000-2003 Intel Corporation. All rights reserved.
 *
 * = PRODUCT
 *      Intel(r) IXA SDK 3.1 for the IXP2000 Network Processor
 *
 * = LIBRARY
 *      
 *
 * = MODULE
 *      Linux kernel module file.
 *
 * = FILENAME
 *      route_config_linux_kernel_module.c
 *
 * = DESCRIPTION
 *      The file defines the init and cleanup functions for Linux kernel module.
 *
 * = CREATION TIME
 *      05/06/2003
 *
 * = CHANGE HISTORY
 *
 * ============================================================================
 * $Id: route_config_linux_kernel_module.c,v 1.9 2003/12/12 19:45:10 ktseng Exp $ 
 */

#include <linux/module.h>
#include "ix_ossl.h"
#include "route_config_linux_kernel_module.h"
#include "ix_cc.h"

#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)

MODULE_LICENSE("Proprietary");

int route_config_open(struct inode *, struct file *); 
int route_config_close(struct inode *, struct file *);
int route_config_ioctl(struct inode *, struct file *, unsigned int, unsigned long);

static struct file_operations route_config_fops = {
    ioctl: route_config_ioctl,
    open: route_config_open,
    release: route_config_close,
};      

/* dynamic allocation of the Major number from the kernel */
ix_int32 g_MajorNumber = DEV_NUMBER_RCONFIG;



/**
 * This function gets called on the insmod system call.
 */
int init_module(void)
{

#ifdef IX_DEBUG
    ix_ossl_message_log("Loading route_config utility driver.\n");
#endif

    if (register_chrdev(g_MajorNumber, DEV_NAME_RCONFIG, &route_config_fops) == 0)
    {
#ifdef IX_DEBUG
        ix_ossl_message_log("route_config utility driver is registered with major:%ld\n",g_MajorNumber);
        ix_ossl_message_log("USAGE:\n");
        ix_ossl_message_log("mknod /dev/%s c %ld <minor>\n", DEV_NAME_RCONFIG, g_MajorNumber);
        ix_ossl_message_log("with different minor numbers.\n\n");
#endif
    }
    else
    {
#ifdef IX_DEBUG
        ix_ossl_message_log("route_config utility driver cannot be registered!!!!.\n");
#endif
        return -1;
    }

    return 0;
} /* init_module */



/**
 * This function gets called on the rmmod system call.
 */
void cleanup_module(void)
{

#ifdef IX_DEBUG
    ix_ossl_message_log("Unloading route_config utility driver.\n");
#endif

    if (unregister_chrdev(g_MajorNumber, DEV_NAME_RCONFIG) == 0)
    {
#ifdef IX_DEBUG
        ix_ossl_message_log("route_config utility driver cleaned successfully...\n");
#endif
    }
    else
    {
#ifdef IX_DEBUG
        ix_ossl_message_log("route_config utility driver COULD NOT BE CLEANED UP !!!\n");
#endif
    }

    return;

} /* cleanup_module */



/**
 * This function is called internally, when the user application calls the open
 * function.
 */
int route_config_open(
    struct inode *route_config_inode,
    struct file *route_config_file)
{
#ifdef IX_DEBUG
    ix_ossl_message_log("route_config_open called.\n");
#endif

    return 0;

} /* route_config_open */



/**
 * This function is invoked, when the user application calls the close 
 * funtion.
 */
int route_config_close(
    struct inode *route_config_inode,
    struct file *route_config_file)
{
#ifdef IX_DEBUG
    ix_ossl_message_log("route_config_close called.\n");
#endif

    return 0;

} /* route_config_close */



/**
 * This function is called by the ioctl command.
 */
int route_config_ioctl(
    struct inode *route_config_inode,
    struct file *route_config_file,
    unsigned int cmd,
    unsigned long arg_IoctlPtr)
{
    ix_uint32 s_Result;
    ix_uint8 cmdString[MAX_NUM_CHAR_CMD_STRING];

#ifdef IX_DEBUG
    ix_ossl_message_log("route_config_ioctl called.\n");
#endif

    if (route_config_inode == NULL)
    {
#ifdef IX_DEBUG
        ix_ossl_message_log("route_config_inode is NULL.\n");
#endif
        return -1;
    }
  
    if (route_config_file == NULL)
    {
#ifdef IX_DEBUG
        ix_ossl_message_log("route_config_file is NULL.\n");
#endif
        return -1;
    }        

    /* Copy argument string to kernel space */
    if ((void *)arg_IoctlPtr != NULL)
    {
        strcpy(cmdString, (void *)arg_IoctlPtr);
    }

    switch(cmd)
    {
        case ADD_NEXT_HOP:
            s_Result = addNextHop(cmdString);
            break;
        case DEL_NEXT_HOP:
            s_Result = delNextHop(cmdString);
            break;
        case ADD_ROUTE:
            s_Result = addRoute(cmdString);
            break;
        case DEL_ROUTE:
            s_Result = delRoute(cmdString);
            break;
        case LOOKUP_ROUTE:
            s_Result = lookupRoute(cmdString);
            break;
        case DUMP_RTM:
            s_Result = dumpRtm();
            break;
        case PURGE_ROUTES:
            s_Result = purgeRoutes();
            break;
        case ROUTE_CONFIG:
            s_Result = route_config();
            break;

#if defined(IX_CONFIGURATION_oc48_ethernet_ipv6_ingress) || defined(IX_CONFIGURATION_oc192_ethernet_ipv6_ingress)
        case ADD_NEXT_HOP_V6:
            s_Result = addNextHopV6(cmdString);
            break;
        case DEL_NEXT_HOP_V6:
            s_Result = delNextHopV6(cmdString);
            break;
        case ADD_ROUTE_V6:
            s_Result = addRouteV6(cmdString);
            break;
        case DEL_ROUTE_V6:
            s_Result = delRouteV6(cmdString);
            break;
        case DUMP_RTM_V6:
            s_Result = dumpRtmV6();
            break;
#if (_IX_OS_TYPE_ == _IX_OS_LINUX_KERNEL_)
        case GET_STAT_V6:
            s_Result = getStatsV6();
            break;
#endif /* _IX_OS_LINUX_KERNEL */
#endif
        default:
#ifdef IX_DEBUG
            ix_ossl_message_log("Undefined ioctl command %d.\n", cmd);
#endif
            s_Result = -1;
            break;
    }

    if (s_Result != 0)
    {
#ifdef IX_DEBUG
        ix_ossl_message_log("route_config_ioctl failed.\n");
#endif
    }
    else
    {
#ifdef IX_DEBUG
        ix_ossl_message_log("route_config_ioctl succeeded.\n");
#endif
    }

    return s_Result;

} /* route_config_ioctl */

#endif
