#ifndef _AAL5_RX_UTIL_UC_
#define _AAL5_RX_UTIL_UC_

/*******************************************************************************
                             Intel Proprietary

 Copyright (c) 1998-2002 By Intel Corporation.  All rights reserved.  
 No part of this program or publication may be reproduced, transmited,
 transcribed, stored in a retrieval system, or translated into any language
 or computer language in any form or by any means, electronic, mechanical,
 magnetic, optical, chemical, manual, or otherwise, without the prior
 written permission of:
                         Intel Corporation
                         2200 Mission College Blvd.
                         Santa Clara, CA  95052-8119
*******************************************************************************/

/*
 *
 *      File Name: 	aal5_rx_util.uc
 *
 *      Purpose: 	This file contains utility macros for AAL5 RX microblock.
 *					
 *
 *      History:
 *
 *
 *      Date            Comment                         By
 *      ------------------------------------------------------------------------
 *
 *		04/16/2002		Created	
 *      06/20/2003		Merged code with ITP								
 *
 */


// To avoid warnings we have to perform tricky definition of parameter passed to
// dl_meta_... macros
#ifdef OPTIONAL_L2PORT

    // Pass the register containing the L2 port number
	#define L2PORT_NUMBER l2port

#else // OPTIONAL_L2PORT

    // If L2 port is not important we pass the ZERO to DL
    #define L2PORT_NUMBER zero

#endif // OPTIONAL_L2PORT

#ifdef OPTIONAL_SUID

    // Pass the register containing the SUID number
	#define SUID_NUMBER suid

#else // OPTIONAL_SUID

    // If SUID is not important we pass the ZERO to DL
    #define SUID_NUMBER zero

#endif // OPTIONAL_SUID

///////////////////////////////////////////////////////////////////////////////
// _aal5_rx_free_rbuf_element:
//
// 		Description: Free up the rbuf element, meaning put the rbuf element onto 
//					 the rbuf free list.
//					
//		Input:		
//			rbuf_element_no:		rbuf_element_no to be freed
//
//		Output:
//
//		Size: 2 cycles.
//			
//////////////////////////////////////////////////////////////////////////////			
#macro _aal5_rx_free_rbuf_element(ele_num)
.begin

	.reg ele_num_16

	; Free the rbuf element.

	; msf[fast_wr,...] requires 
	; rbuf ele no in upper 16 bits  and address in lower 16 bits.

	alu[ele_num_16, --, B, ele_num, <<16] 	       	
	msf[fast_wr, --, ele_num_16, RBUF_ELEMENT_DONE] 	

.end
#endm   //   _aal5_rx_free_rbuf_element()


////////////////////////////////////////////////////////////////////////////////
// _aal5_rx_read_rsw_inputports16:
//
// Description:
//
//		This macro is used when the number of inputput ports supported are 16.
//
// 		Macro to extract element_number, RBUF_offset, input_port,
//      vc_key(vpi/vci/port), 3 bits of PTI from $rsw0 and $rsw1.
//
//	Input:
//
//		RSW sxfer register 0.
//		RSW sxfer register 1.
//	
//	Output:
//
//		input_port: refers to the input port.
//
//	    vc_key:		refers to vci + vpi + input_port (32 bit).
//
//		eop_bit: 	refers to the third bit of PTI of the AAL5 header, which
//					says whether this rbuf element has EOP cell or not.
//
//		pti:		refers to the 3 bits of PTI of the AAL5 header, 
//					bit1|bit2|bit3=mgt_cell/user_cell | congestion ctrl | eop bit
//
//		Note: 
//			For hash lookup, to generate hash_key the init of CRC_REMAINDER and
//			crc_32 over vc_key is done here. This is to avoid extra nops required 
//			between them, which can be filled with useful instructions.
//
//		Size:   				
//		8 instr + 2 instr(for hash_lookup - init CRC_remainder & crc_be on vc_key)
//
////////////////////////////////////////////////////////////////////////////////
#macro _aal5_rx_read_rsw_inputports16(input_port, vc_key, pti, $rsw0, $rsw1)
.begin

	.reg vpi, vci

		#ifdef USE_HASH_LOOKUP
			; Init crc with 0xffff ffff, used to generate hash_key.
			// Ideally this should be done in _aal5_rx_get_vc_ptr[], but doing here to 
			// avoid 3 nops between crc init and crc instr (instr below)
			local_csr_wr[CRC_REMAINDER, lm_crc_init]
		#endif


		; Extract the VPI. 
		alu_shf[vpi, 0xff, AND, $rsw1, >>20]		
	
		; Extract the VCI.
		alu_shf[vci, mask_lower_half, AND, $rsw1, >>4]			
	
		; Extract input_port.
		alu_shf[input_port, 0xf, AND, $rsw0]  

		; Create a key VPI/VCI/port#(32)=(8/16/4).
		alu[vc_key, input_port, OR, vpi, <<20]
		alu[vc_key, vc_key, OR, vci, <<4]

		#ifdef USE_HASH_LOOKUP
			; Use CRC32 to get the index to the bucket data structure.
			// Ideally this should be done in _aal5_rx_get_vc_ptr[], but doing here to 
			// avoid 5 nops reqd between crc instr and crc read. (done in 
			// _aal5_rx_get_vc_ptr[])
			crc_be[crc_32,  --,  vc_key] 
		#endif
															 
		; Extract all 3 bits of PTI.
		alu_shf[pti, 0x7, AND, $rsw1, >>1]	 		

.end
#endm   //   _aal5_rx_read_rsw_inputports16()

////////////////////////////////////////////////////////////////////////////////
// _aal5_rx_read_rsw:
//
// Description:
//
//		This macro is used when the number of inputput ports supported are more
//		than 16.
//
// 		Macro to extract element_number, RBUF_offset, input_port,
//      vc_key(vpi/vci/port), 3 bits of PTI from $rsw0 and $rsw1.
//
//	Input:
//
//		RSW sxfer register 0.
//		RSW sxfer register 1.
//	
//	Output:
//		element_num:refers to the rbuf element_number.
//
//		rbuf_offset:refers to the address of the rbuf element.
//
//		input_port: refers to the input port.
//
//	    vc_key:		refers to vci + vpi + input_port (32 bit).
//
//		eop_bit: 	refers to the third bit of PTI of the AAL5 header, which
//					says whether this rbuf element has EOP cell or not.
//
//		pti:		refers to the 3 bits of PTI of the AAL5 header, 
//					bit1|bit2|bit3=mgt_cell/user_cell | congestion ctrl | eop bit
//
//		Note: 
//			For hash lookup, to generate hash_key the init of CRC_REMAINDER and
//			crc_32 over vc_key is done here. This is to avoid extra nops required 
//			between them, which can be filled with useful instructions.
//
//		Size:   				
//		8 instr + 2 instr(for USE_HASH_LOOKUP - init CRC_remainder & crc_be on vc_key)
//
////////////////////////////////////////////////////////////////////////////////
#macro _aal5_rx_read_rsw(FPGA_header, input_port, vc_key, pti, rsw0, rsw1, cell_0, clp)
.begin

	.reg vpi, vci

		#ifdef USE_HASH_LOOKUP
			; Init crc with 0xffff ffff, used to generate hash_key.
			// Ideally this should be done in _aal5_rx_get_vc_ptr[], but doing here to 
			// avoid 3 nops between crc init and crc instr (instr below)
//			local_csr_wr[CRC_REMAINDER, lm_crc_init] 
			local_csr_wr[CRC_REMAINDER, 0]  ; this is init value used by scripts
		#endif

		; Extract the VPI and VCI
		#ifdef IXP2800

			alu_shf[vpi, 0xff, AND, cell_0, >>20]
			alu_shf[vci, mask_lower_half, AND, cell_0, >>4]
			alu_shf[clp, 0x1, AND, cell_0]

		#else /* IXP2800 */

		    #ifdef FPGA_HEADER

			    alu_shf[vpi, 0xff, AND, cell_0, >>20]
    			alu_shf[vci, mask_lower_half, AND, cell_0, >>4]
	    		alu_shf[clp, 0x1, AND, cell_0]

		    #else /* FPGA_HEADER */

		        alu_shf[vpi, 0xff, AND, rsw1, >>20]		
		        alu_shf[vci, mask_lower_half, AND, rsw1, >>4]			
			    alu_shf[clp, 0x1, AND, rsw1]

		    #endif /* FPGA_HEADER */

		#endif /* IXP2800 */
	

	    ; Extract Protocol Port
		#ifdef IXP2800
	
		    ; Extract Port Address from Status in SPI4 mode.
			alu_shf[input_port, 0xff, AND, rsw0]  

		#else /* IXP2800 */

		    ; Extract Channel from RSW in UTOPIA mode.
		    alu_shf[input_port, 0xf, AND, rsw0]  

		#endif /* IXP2800 */


        #ifdef FPGA_HEADER
		
            ;extract FPGA port.
		    move[FPGA_port, 0x7FF]
		    alu[FPGA_port, --, B, FPGA_port, >>AAL5_PHY_IF_BITS]
		    alu[FPGA_port, FPGA_port, AND, FPGA_header, >>2]
		    alu[input_port,  FPGA_port, OR, input_port, <<AAL5_FPGA_PORT_BITS]	;contcatenate FPGA port and protocol port

        #else /* FPGA_HEADER */

		    alu[input_port, --, B, input_port, <<AAL5_FPGA_PORT_BITS]	;set FPGA port as zero. Only protocol port is valid

        #endif /* FPGA_HEADER */	

		; Create a key VPI/VCI(32)=(16/16).
		alu[vc_key, vci, OR, vpi, <<16]

		#ifdef USE_HASH_LOOKUP

			; Use CRC32 to get the index to the bucket data structure.
			// Ideally this should be done in _aal5_rx_get_vc_ptr[], but doing here to 
			// avoid 5 nops reqd between crc instr and crc read. (done in 
			// _aal5_rx_get_vc_ptr[])
			crc_be[crc_32,  --,  vc_key]    ;first hash uses VCI & VPI
			nop
			crc_be[crc_32,  --,  input_port];second hash uses SPI4port & FPGAport

		#endif

		; Extract all 3 bits of PTI.
		#ifdef IXP2800

				alu_shf[pti, 0x7, AND, cell_0, >>1]

		#else /* IXP2800 */

			#ifdef FPGA_HEADER

				alu_shf[pti, 0x7, AND, cell_0, >>1]

			#else /* FPGA_HEADER */

		        alu_shf[pti, 0x7, AND, rsw1, >>1]	 		

			#endif /* FPGA_HEADER */

		#endif /* IXP2800 */

.end
#endm   //   _aal5_rx_read_rsw()


////////////////////////////////////////////////////////////////////////////////
// _aal5_rx_get_vc_ptr_inputports16:
//
// Description:
//
//		This macro is used when the number of inputput ports supported are 16.
//
//		Get VC flowid by doing VC lookup. VC Info has both RXC and VC Stat 
//		counters. Thus, it returns index of RXC in the range of [0, 64K-1].
//
//		There are two ways of getting RXC ptr
//
//		[1] Simple lookup  - based on 16bit VCI of the vc_key ( 2^16 = 64K VCs)
//			- compiler option of SIMPLE_LOOKUP
//
//		[2] Hash lookup - using CRC to generate an index to the RXC table
//			- compiler option of USE_HASH_LOOKUP
//
//		Note: Simple lookups require that the vc_id corresponding to RXC(VC Info)
//			  is equal to the VCI of the VC. So while initializing the hash tables
//			  this point should be considered.
//
//			  For Hash lookup, the above condition is not required. Its based on 
//			  LS 16 bits result got by computing CRC-32 over the 32 bit VC_key
// 
//      Input:  key : 	 	vc_key (vpi:vci)
//
//		Output: vc_id :		RXC index(vc_ptr) into the RXC table(VC Info table) 
//							pertaining to vc_key in the range of [0, 64K-1].
//
//		Size:   14 cycles (USE_HASH_LOOKUP - hash_hit against first element)				
//				1 cycle  (SIMPLE_LOOKUP)
//			
////////////////////////////////////////////////////////////////////////////////

#macro _aal5_rx_get_vc_ptr_inputports16(vc_id, key)
.begin

	#ifdef SIMPLE_LOOKUP
	;----------------------------------------------------------------------------
	; Simple lookup in SRAM for VC ptr (= RXC_ptr)
	;----------------------------------------------------------------------------

		; Simple lookup is done based on 16bit VCI of vc_key ( 2^16 = 64K VCs).

		; key = vc_key (vpi:vci:port#).
	
		; Extract VCI from vc_key, VCI = vc_ptr would be in range of [0, 64K-1].
		alu[vc_id, mask_lower_half, AND, key, >>4] ; Extract VCI from vc_key


	#else   // USE_HASH_LOOKUP


	;----------------------------------------------------------------------------
	; Hash lookup in SRAM for VC ptr (= RXC_ptr)
	;----------------------------------------------------------------------------
	//
	// Lets say we have 64K connections and 64K Buckets in primary hash table each 
	// bucket having following structure.
	//

    //  Primary Bucket structure
	//  |=========================|
	//  |    	key1(32)		  |	// 32 bit first key
	//  |-------------------------|
	//  |vc_ptr2(16)|vc_ptr1(16)  |	// range [0, 64K-1]
	// 	|-------------------------|
	//  |		key1(32)  		  |	// 32 bit second key
	//	|-------------------------|	
	//	|res(16)  |nxt_bkt_idx(16)|	// index of next sec bucket in the chain(16)
	//	|=========================| // - range [0, 64K-1]


	//   The buckets in the secondary hash table has following structure

	//  Secondary Bucket structure
	//	|=========================|
	//	|    	key1(32)		  |	// 32 bit first key
	//	|-------------------------|
	//	|vc_ptr2(16) | vc_ptr1(16)|	// index range [0, 64K-1]
	//	|-------------------------|
	//	|	    key2(32)		  | // 32 bit second key
	//	|-------------------------|
	//	|	  	key3(32)		  |	// 32 bit third key
	//	|-------------------------|
	//	| vc_ptr4(16)| vc_ptr3(16)|	// index range [0, 64K-1]
	//	|-------------------------|	
	//	| 		key4(32)		  |	// 32 bit fourth key
	//	|-------------------------|	
	//	|	   reserved(32)	      |	// reserved(32)
	//	|-------------------------|	
	//	|res(16)  |nxt_bkt_idx(16)|	// index of next bucket in the chain (16)
	//	|=========================| // - range [0, 64K-1]
	//
	//---------------------------------------------------------------------------
	;
	; We use 32 bit VPI/VCI/input_port as an input to CRC unit, to get 32 bit 
	; crc_result.
	;
	; We use 16 bits of returned crc_result to index hash table to get an 
	; appropriate bucket. (2^16 = 64K buckets).
	;
	; Traverse the bucket comparing the key in the bucket against the
	; VPI/VCI/port# key till you find the hit or end of the bucket. In the later
	; case, if there is no next bucket in the chain then it is a hash_miss and we
	; will drop the packet.
	;
	; If it has next bucket in the chain, get it from SRAM, traverse the bucket, 
	; so on)
	;
	;----------------------------------------------------------------------------

	.reg bkt_index
	.reg bkt_offset
	.reg set_xfer_index
	.reg loop

	.reg bkt_key1, bkt_key2, bkt_key3, bkt_key4

	.sig bkt_read_sig

		// To get rid of these 3 NOPs, this CRC init is done outside this macro,
		// in _aal5_rx_read_rsw[] 

		// Init crc with 0xffff ffff
	    // local_csr_wr[CRC_REMAINDER, lm_crc_init]
		// nop
	 	// nop
	 	// nop 	

		// To get rid of these 5 NOPs, this CRC instr is placed outside this macro,
		// in _aal5_rx_read_rsw[] 

		// Use CRC32 to get the index to the bucket data structure
		// crc_be[crc_32,  --,  key] 	
	    // nop
	 	// nop
	 	// nop
	 	// nop
	 	// nop
	
		// This requires 5 nops between crc_be[] and local_csr_rd[] 

		; Read the actual CRC remainder.
		local_csr_rd[CRC_REMAINDER]
		immed[bkt_index,0]

	 	; bucket_index = first 16 bits of hash_result, index = [0, 64K-1].
		alu[bkt_index, lm_vc_index_mask, AND, bkt_index]  

		xbuf_alloc($bkt_in, 8, read)

		; Get the bucket from SRAM.

		; bucket_offset_sram = (bucket_index * 4) LWs. 
		; Each primary bucket is 4LWs, so <<2 and Offset should be in bytes, so <<4.
		alu_shf[bkt_offset, --, B, bkt_index, <<VC_HASH_PRI_SHF_VAL]         

		; Get the bucket from sram to sxfer reg.
		sram[read, $bkt_in[0], hash_pri_base_sram, bkt_offset, 4], 
				 sig_done[bkt_read_sig]
		ctx_arb[bkt_read_sig]


	; Traversing the bucket.

		// Instead of sxfer reg indexing, we can use sxfer reg directly, 
		// it saves 7 instr.

		br_bclr[$bkt_in[0], 31, HASH_MISS_ERR#]

		alu[bkt_key1, $bkt_in[0], AND~, 1, <<31] 
		alu[--, key, -, bkt_key1]
		beq[HASH_HIT1#]


		br_bclr[$bkt_in[2], 31, HASH_MISS_ERR#]

		alu[bkt_key2, $bkt_in[2], AND~, 1, <<31] 
		alu[--, key, -, bkt_key2]
		beq[HASH_HIT2#]

		; Here, we will get the index of the secondary bucket in the chain.
		; (4th LW)
		alu[bkt_index, mask_lower_half, AND, $bkt_in[3]]
		beq[HASH_MISS_ERR#] 		; If there is no bucket next in the chain, 
									; its hash_miss

GET_SEC_BUCKET#:

		;Get the next bucket. (each secondary bucket is 8LWs)
		alu_shf[bkt_offset, --, B, bkt_index, <<VC_HASH_SEC_SHF_VAL]         

		; Get the bucket from sram to sxfer reg.
		sram[read, $bkt_in[0], hash_sec_base_sram, bkt_offset, 8], \
				sig_done[bkt_read_sig]
		ctx_arb[bkt_read_sig]

		; Traversing the bucket.

		// Instead of sxfer reg indexing, we use sxfer reg directly.

		br_bclr[$bkt_in[0], 31, HASH_MISS_ERR#]

		alu[bkt_key1, $bkt_in[0], AND~, 1, <<31] 
		alu[--, key, -, bkt_key1]
		beq[S_HASH_HIT1#]

		br_bclr[$bkt_in[2], 31, HASH_MISS_ERR#]

		alu[bkt_key2, $bkt_in[2], AND~, 1, <<31] 
		alu[--, key, -, bkt_key2]
		beq[S_HASH_HIT2#]

		br_bclr[$bkt_in[3], 31, HASH_MISS_ERR#]

		alu[bkt_key3, $bkt_in[3], AND~, 1, <<31] 
		alu[--, key, -, bkt_key3]
		beq[S_HASH_HIT3#]
	
		br_bclr[$bkt_in[5], 31, HASH_MISS_ERR#]
	
		alu[bkt_key4, $bkt_in[5], AND~, 1, <<31] 
		alu[--, key, -, bkt_key4]
		beq[S_HASH_HIT4#]

		; Here, we will get the index of the secondary bucket in the chain,
		; (4th LW)
		alu[bkt_index, mask_lower_half, AND, $bkt_in[7]]
		beq[HASH_MISS_ERR#] 		; If there is no bucket next in the chain, 
									; its hash_miss
	    br[GET_SEC_BUCKET#]			; Else get next bucket in the chain

		; Secondary hash table hit

S_HASH_HIT1#:
	
		br[HASH_END#], defer[1]
			alu_shf[vc_id, mask_lower_half, AND, $bkt_in[1]]
	
S_HASH_HIT2#:

		br[HASH_END#], defer[1]
			alu_shf[vc_id, --, b, $bkt_in[1], >>16]

S_HASH_HIT3#:

		br[HASH_END#], defer[1]
			alu_shf[vc_id, mask_lower_half, AND, $bkt_in[4]]

S_HASH_HIT4#:
		br[HASH_END#], defer[1]
			alu_shf[vc_id, --, b, $bkt_in[4], >>16]



		; Primary hash table hit

HASH_HIT2#:
		br[HASH_END#], defer[1]
			alu_shf[vc_id, --, b, $bkt_in[1], >>16]

HASH_HIT1#:

		alu[vc_id, mask_lower_half, AND, $bkt_in[1]]


HASH_END#:

		xbuf_free($bkt_in)

	#endif  // USE_HASH_LOOKUP

.end
#endm   //   _aal5_rx_get_vc_ptr_inputports16()

////////////////////////////////////////////////////////////////////////////////
// _aal5_rx_get_vc_ptr:
//
// Description:
//
//		This macro is used when the number of inputput ports supported are more
//		than 16.
//
//		Get VC flowid by doing VC lookup. VC Info has both RXC and VC Stat 
//		counters. Thus, it returns index of RXC in the range of [0, 64K-1].
//
//		There are two ways of getting RXC ptr
//
//		[1] Simple lookup  - based on 16bit VCI of the vc_key ( 2^16 = 64K VCs)
//			- compiler option of SIMPLE_LOOKUP
//
//		[2] Hash lookup - using CRC to generate an index to the RXC table
//			- compiler option of USE_HASH_LOOKUP
//
//		Note: Simple lookups require that the vc_id corresponding to RXC(VC Info)
//			  is equal to the VCI of the VC. So while initializing the hash tables
//			  this point should be considered.
//
//			  For Hash lookup, the above condition is not required. Its based on 
//			  LS 16 bits result got by computing CRC-32 over the 32 bit VC_key
// 
//      Input:  key : 	 	vc_key (vpi:vci)
//
//		Output: vc_id :		RXC index(vc_ptr) into the RXC table(VC Info table) 
//							pertaining to vc_key in the range of [0, 64K-1].
//
//		Size:   14 cycles (USE_HASH_LOOKUP - hash_hit against first element)				
//				1 cycle  (SIMPLE_LOOKUP)
//			
////////////////////////////////////////////////////////////////////////////////

#macro _aal5_rx_get_vc_ptr(vc_id, key)
.begin

	#ifdef SIMPLE_LOOKUP
	;----------------------------------------------------------------------------
	; Simple lookup in SRAM for VC ptr (= RXC_ptr)
	;----------------------------------------------------------------------------

		; Simple lookup is done based on 16bit VCI of vc_key ( 2^16 = 64K VCs).

		; key = vc_key (vpi:vci).
	
		; Extract VCI from vc_key, VCI = vc_ptr would be in range of [0, 64K-1].
//		alu[vc_id, mask_lower_half, AND, key, >>4] ; Extract VCI from vc_key
		alu[vc_id, mask_lower_half, AND, key] ; Extract VCI from vc_key


	#else   // USE_HASH_LOOKUP


	;----------------------------------------------------------------------------
	; Hash lookup in SRAM for VC ptr (= RXC_ptr)
	;----------------------------------------------------------------------------
	//
	// Lets say we have 64K connections and 64K Buckets in primary hash table each 
	// bucket having following structure.
	//

    //  Primary Bucket structure
	//  |=========================|
	//  |    	key1(32)		  |	// 32 bit first key
	//  |-------------------------|
	//  |vc_ptr2(16)|vc_ptr1(16)  |	// range [0, 64K-1]
	// 	|-------------------------|
	//  |		key1(32)  		  |	// 32 bit second key
	//	|-------------------------|	
	//	|res(16)  |nxt_bkt_idx(16)|	// index of next sec bucket in the chain(16)
	//	|=========================| // - range [0, 64K-1]


	//   The buckets in the secondary hash table has following structure

	//  Secondary Bucket structure
	//	|=========================|
	//	|    	key1(32)		  |	// 32 bit first key
	//	|-------------------------|
	//	|vc_ptr2(16) | vc_ptr1(16)|	// index range [0, 64K-1]
	//	|-------------------------|
	//	|	    key2(32)		  | // 32 bit second key
	//	|-------------------------|
	//	|	  	key3(32)		  |	// 32 bit third key
	//	|-------------------------|
	//	| vc_ptr4(16)| vc_ptr3(16)|	// index range [0, 64K-1]
	//	|-------------------------|	
	//	| 		key4(32)		  |	// 32 bit fourth key
	//	|-------------------------|	
	//	|	   reserved(32)	      |	// reserved(32)
	//	|-------------------------|	
	//	|res(16)  |nxt_bkt_idx(16)|	// index of next bucket in the chain (16)
	//	|=========================| // - range [0, 64K-1]
	//
	//---------------------------------------------------------------------------
	;
	; We use 32 bit VPI/VCI/input_port as an input to CRC unit, to get 32 bit 
	; crc_result.
	;
	; We use 16 bits of returned crc_result to index hash table to get an 
	; appropriate bucket. (2^16 = 64K buckets).
	;
	; Traverse the bucket comparing the key in the bucket against the
	; VPI/VCI/port# key till you find the hit or end of the bucket. In the later
	; case, if there is no next bucket in the chain then it is a hash_miss and we
	; will drop the packet.
	;
	; If it has next bucket in the chain, get it from SRAM, traverse the bucket, 
	; so on)
	;
	;----------------------------------------------------------------------------

	.reg bkt_index
	.reg bkt_offset
	.reg set_xfer_index
	.reg loop tmp2

	.reg bkt_key1, bkt_key2, bkt_key3, bkt_key4
    .reg key2

	.sig bkt_read_sig

		// To get rid of these 3 NOPs, this CRC init is done outside this macro,
		// in _aal5_rx_read_rsw[] 

		// Init crc with 0xffff ffff
	    // local_csr_wr[CRC_REMAINDER, lm_crc_init]
		// nop
	 	// nop
	 	// nop 	

		// To get rid of these 5 NOPs, this CRC instr is placed outside this macro,
		// in _aal5_rx_read_rsw[] 

		// Use CRC32 to get the index to the bucket data structure
		// crc_be[crc_32,  --,  key] 	
	    // nop
	 	// nop
	 	// nop
	 	// nop
	 	// nop
	
		// This requires 5 nops between crc_be[] and local_csr_rd[] 

		alu[key2, --, B, input_port, >>7]			; Create addition key as 4MSB input_port

		; Read the actual CRC remainder.
		local_csr_rd[CRC_REMAINDER]
		immed[bkt_index,0]

	 	; bucket_index = index mask is calculated based on number of hash_table entries.
		alu[bkt_index, lm_vc_index_mask, AND, bkt_index]  

		xbuf_alloc($bkt_in, 8, read)

		; Get the bucket from SRAM.

		; bucket_offset_sram = (bucket_index * 4) LWs. 
		; Each primary bucket is 4LWs, so <<2 and Offset should be in bytes, so <<4.
		alu_shf[bkt_offset, --, B, bkt_index, <<AAL5_RX_VC_HASH_PRI_BKT_ENTRY_SIZE_LOG2]         

		; Get the bucket from sram to sxfer reg.
		sram[read, $bkt_in[0], hash_pri_base_sram, bkt_offset, 4], 
				 sig_done[bkt_read_sig]

		// It is necessary, because exist risk that for two different
		// port and the same VPI, VCI we can obtine the same bucket, so we must 
		// expand validation of bucket
		alu[key, key, OR, input_port, <<24]	; Add 7LSB of input_port to vc_key

		// we must left last bit 
		alu[key, key, AND~, 1, <<31]	

		ctx_arb[bkt_read_sig]

	; Traversing the bucket.

		// Instead of sxfer reg indexing, we can use sxfer reg directly, 
		// it saves 7 instr.

		br_bclr[$bkt_in[0], 31, HASH_MISS_ERR#]

   		alu[bkt_key1, $bkt_in[0], AND~, 1, <<31] 
		alu[bkt_key2, 0xF, AND, $bkt_in[3], >>28]
		alu[tmp2, key, -, bkt_key1]
		alu[tmp, key2, -, bkt_key2]
		alu[--, tmp, OR, tmp2]
		beq[HASH_HIT1#]
		
		alu[bkt_key1, $bkt_in[2], AND~, 1, <<31] 
		alu[bkt_key2, 0xF, AND, $bkt_in[3], >>24]
		alu[tmp2, key, -, bkt_key1]
		alu[tmp, key2, -, bkt_key2]
		alu[--, tmp, OR, tmp2]
		beq[HASH_HIT2#]

		; Here, we will get the index of the secondary bucket in the chain.
		; (4th LW)
		alu[bkt_index, mask_lower_half, AND, $bkt_in[3]]
		beq[HASH_MISS_ERR#] 		; If there is no bucket next in the chain, 
									; its hash_miss

GET_SEC_BUCKET#:

		;Get the next bucket. (each secondary bucket is 8LWs)
		alu_shf[bkt_offset, --, B, bkt_index, <<AAL5_RX_VC_HASH_SEC_BKT_ENTRY_SIZE_LOG2]         

		; Get the bucket from sram to sxfer reg.
		sram[read, $bkt_in[0], hash_sec_base_sram, bkt_offset, 8], sig_done[bkt_read_sig]
		ctx_arb[bkt_read_sig]

		; Traversing the bucket.

		// Instead of sxfer reg indexing, we use sxfer reg directly.

		br_bclr[$bkt_in[0], 31, HASH_MISS_ERR#]

		alu[bkt_key1, $bkt_in[0], AND~, 1, <<31] 
		alu[bkt_key2, 0xF, AND, $bkt_in[7], >>28]
		alu[tmp2, key, -, bkt_key1]
		alu[tmp, key2, -, bkt_key2]
		alu[--, tmp, OR, tmp2]
		beq[S_HASH_HIT1#]

		alu[bkt_key1, $bkt_in[2], AND~, 1, <<31] 
		alu[bkt_key2, 0xF, AND, $bkt_in[7], >>24]
		alu[tmp2, key, -, bkt_key1]
		alu[tmp, key2, -, bkt_key2]
		alu[--, tmp, OR, tmp2]
		beq[S_HASH_HIT2#]

		alu[bkt_key1, $bkt_in[3], AND~, 1, <<31] 
		alu[bkt_key2, 0xF, AND, $bkt_in[7], >>20]
		alu[tmp2, key, -, bkt_key1]
		alu[tmp, key2, -, bkt_key2]
		alu[--, tmp, OR, tmp2]
		beq[S_HASH_HIT3#]
	
		alu[bkt_key1, $bkt_in[5], AND~, 1, <<31] 
		alu[bkt_key2, 0xF, AND, $bkt_in[7], >>16]
		alu[tmp2, key, -, bkt_key1]
		alu[tmp, key2, -, bkt_key2]
		alu[--, tmp, OR, tmp2]
		beq[S_HASH_HIT4#]

		; Here, we will get the index of the secondary bucket in the chain,
		; (4th LW)
		alu[bkt_index, mask_lower_half, AND, $bkt_in[7]]
		beq[HASH_MISS_ERR#] 		; If there is no bucket next in the chain, 
									; its hash_miss
	    br[GET_SEC_BUCKET#]			; Else get next bucket in the chain

		; Secondary hash table hit

S_HASH_HIT1#:
		alu_shf[vc_id, mask_lower_half, AND, $bkt_in[1]]
		br[HASH_END#]
	
S_HASH_HIT2#:
		alu_shf[vc_id, --, b, $bkt_in[1], >>16]
		br[HASH_END#]

S_HASH_HIT3#:
		alu_shf[vc_id, mask_lower_half, AND, $bkt_in[4]]
		br[HASH_END#]

S_HASH_HIT4#:
		alu_shf[vc_id, --, b, $bkt_in[4], >>16]
   		br[HASH_END#]

		; Primary hash table hit
HASH_HIT2#:
		alu_shf[vc_id, --, b, $bkt_in[1], >>16]
   		br[HASH_END#]

HASH_HIT1#:

		alu[vc_id, mask_lower_half, AND, $bkt_in[1]]


HASH_END#:

		xbuf_free($bkt_in)

	#endif  // USE_HASH_LOOKUP

.end
#endm   //   _aal5_rx_get_vc_ptr()

///////////////////////////////////////////////////////////////////////////////
//  _aal5_rx_phase1_inputports16
//
//		Description:
//
//		This macro is used when the number of inputput ports supported are 16.
//		(The read of RSW and hashlookup is done accordingly)
//						
//		This part of code is executed when the thread wakes up with the 
//		rsw_rcvd_sig (ie when RBUF is assigned to it)
//
//		It checks RSW for any errors, reads the RBUF into Xfer regs of ME.
//		It then does (hashlookup/simple lookup) to get the vc_ptr into
//		RXC table for that VC
//
//	 	Outputs:
//
//		Inputs: 
//
//		Size: 		
//
///////////////////////////////////////////////////////////////////////////////

#macro _aal5_rx_phase1_inputports16()
.begin

	// For hash lookup, to generate hash_key, the init of CRC_REMAINDER and
	// crc_32 over vc_key is done inside this macro. This is to avoid extra nops 
	// required between them.
	_aal5_rx_read_rsw_inputports16(input_port, vc_key, pti_3bits, $rsw0, $rsw1)

	; Check for parity error, null/idle cell, cell_size less than 52B error
	; We assume we get AAL5 traffic here, so we dont check for port_type.

	; Check for error bit in RSW(13th bit), null/idle cell Rx (9th bit) and
	; cell_size from the RSW(byte_count[16:23], should be 0x34 = 52B)
	; So combined mask becomes 0xff22.

	alu[tmp, lm_rsw_err_chk_mask, AND, $rsw0, >>8]			
	alu[--, lm_rsw_err_chk_cell_sz, -, tmp]
	bne[RSW_ERR#]

						
	; Read the whole cell (header + payload) from rbuf into sxfer reg.
	; Set ref_cnt as 13(0xC) and OV[25] bit to override reference count.
	alu_shf[--, --, B, 0x1C, <<21]
	msf[read, $$cell_in0, rbuf_offset, rbuf_base, max_13], indirect_ref, \
			sig_done[rbuf_rd_sig]

		; Get VC flowid by doing VC lookup into SRAM (flowid - [0, 64K-1])
    _aal5_rx_get_vc_ptr_inputports16(vc_flowid, vc_key)


	#ifdef ONE_ME_RX_FP

		ctx_arb[rbuf_rd_sig, next_thread_sig]

	#else // TWO_ME_RX_FP

		br!=ctx[0, NOT_THD_0#]
	
		; ctx0 waits for [enter_cs1_inter_me_sig, next_thread_sig]
		; to enter critical section#1 - access to RXC1.
		ctx_arb[rbuf_rd_sig, enter_cs1_inter_me_sig, next_thread_sig]
		br[PHASE2#], defer[2]

			; Signal the next thread.
			signal_ctx(next_thread_sig_val)

			; clear CAM for critical section1.
			cam_clear

	NOT_THD_0#:

		ctx_arb[rbuf_rd_sig, next_thread_sig]

		; Signal the next thread.
		signal_ctx(next_thread_sig_val)

	#endif   // TWO_ME_RX_FP
.end
#endm // _aal5_rx_phase1_inputports16

///////////////////////////////////////////////////////////////////////////////
//  _aal5_rx_phase1
//	
//		Description:
//
//		This macro is used when the number of inputput ports supported are more
//		than 16.
//		(The read of RSW and hashlookup is done accordingly)
//						
//		This part of code is executed when the thread wakes up with the 
//		rsw_rcvd_sig (ie when RBUF is assigned to it)
//
//		It reads the RBUF into Xfer regs of ME. Then it checks RSW for any errors.
//		It then does (hashlookup/simple lookup) to get the vc_ptr into
//		RXC table for that VC
//
//	 	Outputs:
//
//		Inputs: 
//
//		Size: 		
//
///////////////////////////////////////////////////////////////////////////////

#macro _aal5_rx_phase1()
.begin

	#ifdef FPGA_HEADER

		; Read the whole cell (header + payload) from rbuf into sxfer reg.
		; Set ref_cnt as 14(0xD) and OV[25] bit to override reference count.
		alu_shf[--, --, B, 0x1D, <<21]
		msf[read, $$FPGA_header, rbuf_offset, rbuf_base, max_14], indirect_ref, \
				sig_done[rbuf_rd_sig]

	#else /* FPGA_HEADER */

		; Read the whole cell (header + payload) from rbuf into sxfer reg.
		; Set ref_cnt as 13(0xC) and OV[25] bit to override reference count.
		alu_shf[--, --, B, 0x1C, <<21]
		msf[read, $$cell_in[0], rbuf_offset, rbuf_base, max_13], indirect_ref, \
				 sig_done[rbuf_rd_sig]

	#endif /* FPGA_HEADER */

		// For hash lookup, to generate hash_key, the init of CRC_REMAINDER and
		// crc_32 over vc_key is done here. This is to avoid extra nops required 
		// between them.

		ctx_arb[rbuf_rd_sig]

	#ifdef FPGA_HEADER

		_aal5_rx_read_rsw($$FPGA_header, input_port, vc_key, pti_3bits, 
					$rsw0, $rsw1, $$cell_in[0], clp)

	#else /* FPGA_HEADER */

		_aal5_rx_read_rsw($$cell_in0, input_port, vc_key, pti_3bits, 
					$rsw0, $rsw1, $$cell_in[0], clp)

	#endif /* FPGA_HEADER */  

		; Check for parity error, null/idle cell, cell_size less than 52B error
		; We assume we get AAL5 traffic here, so we dont check for port_type.

		; Check for error bit in RSW(13th bit), null/idle cell Rx (9th bit) and
		; cell_size from the RSW(byte_count[16:23], should be 0x34 = 52B)
		; So combined mask becomes 0xff22.

		alu[tmp, lm_rsw_err_chk_mask, AND, $rsw0, >>8]			
		alu[--, lm_rsw_err_chk_cell_sz, -, tmp]
		bne[RSW_ERR#]

		; Get VC flowid by doing VC lookup into SRAM (flowid - [0, 64K-1])
	    _aal5_rx_get_vc_ptr(vc_flowid, vc_key)


	#ifdef ONE_ME_RX_FP

		ctx_arb[next_thread_sig]

	#else // TWO_ME_RX_FP

		br!=ctx[0, NOT_THD_0#]
	
		; ctx0 waits for [enter_cs1_inter_me_sig, next_thread_sig]
		; to enter critical section#1 - access to RXC1.
		ctx_arb[enter_cs1_inter_me_sig, next_thread_sig]

		; Signal the next thread.
		signal_ctx(next_thread_sig_val)

		; clear CAM for critical section1.
		cam_clear
	   	br[PHASE2#]

	NOT_THD_0#:

		ctx_arb[next_thread_sig]

		; Signal the next thread.
		signal_ctx(next_thread_sig_val)

	#endif   // TWO_ME_RX_FP

.end
#endm // _aal5_rx_phase1()

///////////////////////////////////////////////////////////////////////////////
//  _aal5_rx_eop_sop_proc
//
//		Description:
//						
//			Process a cell containing both SOP & EOP . i.e packet size
//			less than or equal to 40B data (40B data + 8B trailer = 48B cell).
//
//	 	Outputs:
//
//		Inputs: 
//			sig_mask :   mask to be set with various I/O signals.
//
//			rxc_adr :    absolute rxc addr used to write back RXC to SRAM.
//
//			$$cell_in :  dram xfer regs used to write to DRAM.
//
//			input_port : The port through which the packet came in.
//
//			vc_id : 	 index of RXC into RXC table (VC info table)
//
//			pti :  		 3 bits of PTI of ATM header.
//
//			$prefetch_buf_handle : already prefetched new buffer handle.
//
//		Size: 		
//
//		 ONE_ME_RX_FP  -  71 cycles.
//		 TWO_ME_RX_FP	 - 111 cycles.						
//
///////////////////////////////////////////////////////////////////////////////

#macro _aal5_rx_eop_sop_proc(sig_mask, rxc_addr, $$cell_in, input_port, vc_id,\
		pti, $prefetch_buf_handle)
.begin

	.reg buf_offset
	.reg final_crc_remainder
	.reg pkt_length
	.reg hdr_type

	.reg dram_data_addr
	.reg dram_ref_cnt


	.reg excep_bit	


	// CRC -init in the defer slot of jump table in aal5_rx() needs 3 instr 
	// before we can use crc instr. Because of defer slot in br[EOP_SOP#], 
	// it is taken care of.

	#ifdef TWO_ME_RX_FP

		// Allocate xfer reg for writing RXC1 back to SRAM.
		xbuf_alloc($rxc_out, 2, write)

	#endif


	; Check if prefetch buffer handle is zero.
	alu[--, --, B, $prefetch_buf_handle, <<8]	// [23:0] contains buf handle.
	beq[ES_NULL_BUF_PROC#]						// if null, error condition.

	; We got a valid buffer here.
	
	; Allocate sxfer reg to read in data from the RBUF element.
	xbuf_alloc($$cell_out, 12, write)


	; Compute CRC
	; This requires 1 nop/instr between two crc instr.

	crc_be[crc_32,  $$cell_out[0],  $$cell_in[1]]           
		; Get the packet length from 3rd and 4th B of 8B trailer (length includes 
		; whole of payload, including LLC/SNAP hdr, but doesnt include padding).
		alu[pkt_length, mask_lower_half, AND, $$cell_in[11]]

	crc_be[crc_32,  $$cell_out[1],  $$cell_in[2]]           
		; Actual no of 8B words = (bytes_in_last_cell - 1)/8 + 1 
		; Indirect ref cnt for DRAM write requires one less, so it would be 
		; (bytes_in_last_cell - 1)/8.
		; Calculate number of 8-byte words for writing to DRAM.
		alu[dram_ref_cnt, pkt_length, -, 1]

	crc_be[crc_32,  $$cell_out[2],  $$cell_in[3]]           
		alu[dram_ref_cnt, --, b, dram_ref_cnt, >>3]

	crc_be[crc_32,  $$cell_out[3],  $$cell_in[4]]           
		; Set OV[25] bit to override ref_cnt.
		alu[dram_ref_cnt, dram_ref_cnt, OR, 1, <<4]

	crc_be[crc_32,  $$cell_out[4],  $$cell_in[5]]           
		; Set the dl_buf_handle.
		; The buffer handle returned, always has EOP, SOP bit set & seg_cnt ==0,
		alu[dl_buf_handle, --, B, $prefetch_buf_handle]					

	crc_be[crc_32,  $$cell_out[5],  $$cell_in[6]]           
		dl_meta_set_packet_size(pkt_length)

	crc_be[crc_32,  $$cell_out[6],  $$cell_in[7]]           
		nop
	crc_be[crc_32,  $$cell_out[7],  $$cell_in[8]]           
		nop
	crc_be[crc_32,  $$cell_out[8],  $$cell_in[9]]           
		nop
	crc_be[crc_32,  $$cell_out[9],  $$cell_in[10]]           
		nop
	crc_be[crc_32,  $$cell_out[10], $$cell_in[11]]
		nop
	crc_be[crc_32,  $$cell_out[11], $$cell_in[12]]        

	; This requires 5 nops/instr between crc_be[] and local_csr_rd[] 
	
	; Get random offset used to do DRAM bank scheduling. 
	_aal5_rx_get_rand_buf_offset(buf_offset)

	; Get dram address for the next buffer.
	dl_buf_get_data(dram_data_addr, $prefetch_buf_handle) 

	; Read the actual CRC remainder.
	local_csr_rd[CRC_REMAINDER]
	immed[final_crc_remainder,0]

	#ifdef TWO_ME_RX_FP

		; In TWO_ME_RX_FP, we have to update two fields - crc_residue, exp_SOP
		; else the NEXT ME might get the stale state from SRAM, specifically
		; when there was a long packet ahead of it spread over TWO MEs.

		; Write back RXC1 to SRAM (fill in the defer slots with RXC1 updates).

ES_REF_CNT_CHK#:

		alu[--, *l$index0[RXC_LM_REF_CNT_OFFSET], -, 0]
		bne[ES_REF_CNT_NONZERO#], defer[2]
		; Set RXC:exp_SOP bit.
		alu_shf[lm_sop_alloc_valid_aal_flist, lm_sop_alloc_valid_aal_flist, OR, 1, <<31]
		; Reset RXC:crc.
		alu[lm_crc_residue, --, b, lm_crc_init]

		; Here the ref_cnt is zero.
		_aal5_rx_write_rxc_to_sram_cs1($rxc_out, rxc_addr, rxc_write_sig) 

		br!=ctx[7, ES_REF_CNT_NONZERO#], defer[1]
		alu[sig_mask, sig_mask, OR, 1, <<(&rxc_write_sig)]
		

		; Signal next ME with enter_cs1_inter_me_sig.
	   	alu[--, --, B, lm_cs1_inter_me_sig_val]
		cap[fast_wr, ALU, INTERTHREAD_SIG]	

		; Clear CAM for critical section2.
		cam_clear

ES_REF_CNT_NONZERO#:

		br!=ctx[0, ES_THD_NOT_0#]

		alu[sig_mask, sig_mask, OR, 1, <<ENTER_CS2_INTER_ME_SIGNAL]

ES_THD_NOT_0#:

		; Ctx0 waits [CS2_inter_me_sig, next_thread_sig, rxc_write_sig (if any)]
		; other ctx waits  [next_thread_sig, rxc_write_sig (if any)].
		local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
		ctx_arb[--] 

ES_PHASE4#:

		; Signal the next thread.
		signal_ctx(next_thread_sig_val)

		; Get RXC2 from sram
		_aal5_rx_get_rxc_cs2(rxc_addr)
	
ES_PHASE5#:

		; Decrement the ref cnt.
		alu[*l$index0[RXC_LM_REF_CNT_OFFSET], *l$index0[RXC_LM_REF_CNT_OFFSET], -, 1]
	
		br!=ctx[7, ES_NOT_THD_7#], defer[2]
		; Signal the next thread.
		signal_ctx(next_thread_sig_val)
		alu[sig_mask, --, B, 1, <<&next_thread_sig]

		; Signal next ME with CS2_inter_me_sig, here as there is no need to write
		; back RXC2 to SRAM
	   	alu[--,--,B,lm_cs2_inter_me_sig_val]
		cap[fast_wr, ALU, INTERTHREAD_SIG]	

ES_NOT_THD_7#:

	#endif  // TWO_ME_RX_FP


	; Compare the actual CRC remainder with the fixed remainder.
	alu[--,final_crc_remainder, -, lm_fixed_crc_remainder]
	bne[ES_CRC_ERR#] 

	; Check if pkt_len = 0, if yes then error.
	alu[--, pkt_length, -, 0]
	beq[ES_PKT_LEN_ERR#]					

	; Update counters.
	_aal5_rx_incr_vc_counter[vc_id, AAL5_RX_AAL5_PDU_COUNT]
	_aal5_rx_incr_port_counter[input_port, AAL5_RX_PKT_COUNT]


	// What part of meta data AAL5 RX writes out to SRAM and passes to the next 
	// block.

	// AAL5 RX updates LW 0,1,2,3(31:16) of meta data. The rest of the meta data
	// are not touched by AAL5 RX. (Refer Buffer Descriptors for more details.)
	//
	//
	// LW 0,1 are written to SRAM by AAL5 RX on EOP/buffer_overflow.
	//
	// On EOP, AAL5 RX sets following values using dl macros
	//
 	// And dl_next_block is set to AAL5_RX_NEXT1 only on EOP along with following
	// values.
	//
	// If dl_next_block = AAL5_RX_NEXT1. following 6 LWs are put onto the 
	// scratch ring in dl_sink[] for the next block. 
	//
	// (1) dl_buf_handle
	// (2) dl_eop_buf_handle
	// (3) meta1 (sop_buf_size, offset)
	// (4) meta2 (pkt_size, freelist_id, rx_stat, header_type, clp)	
	// (5) meta3 (input_port(ms 16bit))
	// (6) meta5 (vc_id(ls 16bit))

	; Write META LW0, LW1 of the current SOP_EOP buffer to SRAM.

	xbuf_alloc($eop_sop_meta,6, write)					// no code generated.


#ifdef OPTIONAL_L2PORT
	alu[l2port, lm_optional_info, AND, mask_lower_half]		;extract L2port
#endif

#ifdef OPTIONAL_SUID
	alu[suid, --, B, lm_optional_info]		;extract SUID
#endif


#ifdef METADATA_2LW

	dl_meta_write($eop_sop_meta, $prefetch_buf_handle, IX_NULL, buf_offset, pkt_length,\
					  0, 0, 0, 0, 0, \
					  0, 0, 0, 0, 0, \
					  0, meta1_sig, SIG_NONE, 0, 2)  
#else

	;prepare clp bit to put in meta data
	alu[clp, --, B, clp, <<CLP_BIT_OFFSET_IN_DL_META4]

	#ifdef DL_SPLIT_META_DATA

		dl_meta_write($eop_sop_meta, $prefetch_buf_handle, IX_NULL, buf_offset, pkt_length,\
						  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
						  0, clp, 0, SUID_NUMBER, 0, \
						  0, meta1_lw0_sig, SIG_NONE, 0, 1)  
		dl_meta_write($eop_sop_meta, $prefetch_buf_handle, IX_NULL, buf_offset, pkt_length,\
						  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
						  0, clp, 0, SUID_NUMBER, 0, \
						  0, meta1_sig, SIG_NONE, 1, 5)  
	#else /* DL_SPLIT_META_DATA */

		dl_meta_write($eop_sop_meta, $prefetch_buf_handle, IX_NULL, buf_offset, pkt_length,\
						  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
						  0, clp, 0, SUID_NUMBER, 0, \
						  0, meta1_sig, SIG_NONE, 0, 6)  

	#endif

#endif  // METADATA_2LW

					
	//	Cell Count Calculation. These are the possibilities for transmit.

	//[1] For CSIX TX, for every packet prepend hdr is added.
	//	  (apart from a cell hdr for every cell). Include this hdr in
	//	  cell count calculation. This per packet prepend header is added
	//    on fly and so its not counted in buffer size/packet_size.
	//
	//    CSIX cell is 120B. For first CSIX cell of the packet, we
	//    have packet prepend hdr of 8B. So first CSIX cell is 112B.
	//
	//	  No need to compute cell_count here, as cell_count here would be 1.
	//	  Queueing hardware requires it to be 1 less than actual.
	//
	//[2] For AAL5 TX, the cell_size is 48B. - no need to compute cell_cnt


	; Write this cell(includes LLC/SNAP hdr, if any) to dram, after rounding it off 
	; to 8B words.
	alu_shf[dram_ref_cnt, --, b, dram_ref_cnt, <<21]
	dram[write, $$cell_out[0], dram_data_addr, buf_offset, max_5],indirect_ref, \
				sig_done[data_write_sig]


	; Prefetch a buffer handle.
    dl_buf_alloc($prefetch_buf_handle, BUF_FREE_LIST0, bd_prefetch_sig, SIG_NONE) 

CHK_EXCEP_BIT#:

	; Check if RXC2:exception bit is set

	br_bset[lm_drp_ex_hdr_prv_prv_buf_addr, 30, ES_EXCEPTION#], defer[3]
	; Set other meta fields for the next block.
	dl_meta_set_offset(buf_offset)

.reg tmp
	alu[tmp, --, B, pkt_length]   				; To avoid reg conflict
	dl_meta_set_buffer_size(tmp)

	
#ifdef OPTIONAL_L2PORT

	alu[l2port, lm_optional_info, AND, mask_lower_half]		;extract L2port
	dl_meta_set_input_port(l2port) 
#else
	dl_meta_set_input_port(input_port) 

#endif

   	br[ES_SET_META#], defer[1]
	alu[dl_next_block, --, B,AAL5_RX_NEXT1]		; set next_block AAL5_RX_NEXT1
				
ES_EXCEPTION#: 

	alu[dl_next_block, --, B, IX_EXCEPTION]	

#ifdef OPTIONAL_L2PORT

	alu[l2port, lm_optional_info, AND, mask_lower_half]		;extract L2port
	dl_meta_set_input_port(l2port) 
#else
	dl_meta_set_input_port(input_port) 

#endif

ES_SET_META#:

	;	The way packets are sent to next block in the pipeline is thro'
	;	dl_sink. (see the outer loop in aal5_rx.uc). 
	;	As part of optimisation, 4 LWs ( 1, 2, 3(31:16), 5) from meta data are sent 
	;	in the scratch ring. dl_sink takes these meta data from the 
	;	previous dl_meta_set_* calls. These meta data should belong to SOP buffer.

	; In our case the prev dl_set_* calls belong to the SOP_EOP buffer. 

	; Some of the meta data is set in the defer slots above.

	dl_meta_set_free_list(FREE_LIST_ID)						
	dl_meta_set_rx_stat(0)					
	_aal5_rx_get_rxc_header_type(hdr_type, lm_drp_ex_hdr_prv_prv_buf_addr)	
	dl_meta_set_header_type(hdr_type) 	


	#ifdef ONE_ME_RX_FP

		alu[sig_mask, sig_mask, OR, 1, <<(&meta1_sig)]

        #ifdef DL_SPLIT_META_DATA

		    alu[sig_mask, sig_mask, OR, 1, <<(&meta1_lw0_sig)]

        #endif

		alu[sig_mask, sig_mask, OR, 3, <<(&data_write_sig)]
		alu[sig_mask, sig_mask, OR, 1, <<(&bd_prefetch_sig)]

	#else   //  TWO_ME_RX_FP

		alu[sig_mask, sig_mask, OR, 1, <<(&meta1_sig)]

        #ifdef DL_SPLIT_META_DATA

		    alu[sig_mask, sig_mask, OR, 1, <<(&meta1_lw0_sig)]

        #endif

		
		br!=ctx[0, ES_THD_NOT_0_2#],defer[2]
		alu[sig_mask, sig_mask, OR, 3, <<(&data_write_sig)]
		alu[sig_mask, sig_mask, OR, 1, <<(&bd_prefetch_sig)]

		; Add CS3_inter_me_sig to sig_mask.
		alu[sig_mask, sig_mask, OR, 1, <<ENTER_CS3_INTER_ME_SIGNAL]

ES_THD_NOT_0_2#:	

	#endif // TWO_ME_RX_FP


	; Ctx0 waits [CS3_inter_me_sig, meta1_sig, bd_prefetch_sig, data_write_sig,
	;			 next_thread_sig]
	; other ctx waits [meta1_sig, bd_prefetch_sig, data_write_sig, next_thread_sig]

	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
	ctx_arb[--], defer[2], br[ES_END#]
	alu[dl_eop_buf_handle, --, B, DL_BUF_NULL]  ; set dl_eop_buf_handle NULL
	dl_meta_set_flow_id(vc_id)

    	
	xbuf_free($eop_sop_meta)                 // no code generated
	
;-----------------------------------------------------------------------------
ES_CRC_ERR#:
;-----------------------------------------------------------------------------

	; We came here because of CRC error.

	_aal5_rx_incr_vc_counter[vc_id, AAL5_RX_AAL5_PDU_CRC_BAD]
	_aal5_rx_incr_port_counter[input_port, AAL5_RX_PKT_DROP_COUNT]

	br[ES_ERR_PROC#]

;-----------------------------------------------------------------------------
ES_PKT_LEN_ERR#:
;-----------------------------------------------------------------------------

	; We came here because pkt_len = 0.

	_aal5_rx_incr_vc_counter[vc_id, AAL5_RX_AAL5_PDU_LEN_BAD]
	_aal5_rx_incr_port_counter[input_port, AAL5_RX_PKT_DROP_COUNT]

ES_ERR_PROC#:


	; Set the next block ID to BID_NULL, no buffer used till now, so no 
	; buffer to drop.

	#ifdef TWO_ME_RX_FP

		br!=ctx[0, ES_THD_NOT_0_3#]

		alu[sig_mask, sig_mask, OR, 1, <<ENTER_CS3_INTER_ME_SIGNAL]

ES_THD_NOT_0_3#:	

	#endif

	; Ctx0 waits [CS3_inter_me_sig, next_thread_sig]
	; other ctx waits [next_thread_sig].
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
	ctx_arb[--], defer[1], br[ES_END#]
	alu[dl_next_block, --, B, BID_NULL]		 	; set next_block NULL


;-----------------------------------------------------------------------------		
ES_NULL_BUF_PROC#:
;-----------------------------------------------------------------------------

	;	We got a null buffer. So nothing to do for current cell. 

	; Update Stat counter.
	_aal5_rx_incr_port_counter[input_port, AAL5_RX_PKT_DROP_COUNT]



	#ifdef TWO_ME_RX_FP

		; In TWO_ME_RX_FP, we have to write back two fields - crc_residue, 
		; exp_SOP else the NEXT ME might get the stale state from SRAM, 
		; specifically when there was a long packet ahead of it spread over 
		; TWO MEs.

		; Write back RXC1 to SRAM (using defer slots for RXC1 updates)

ES_REF_CNT_CHK2#:

		alu[--, *l$index0[RXC_LM_REF_CNT_OFFSET], -, 0]
		bne[ES_REF_CNT_NONZERO2#], defer[2]
		; Set RXC:exp_SOP bit.
		alu_shf[lm_sop_alloc_valid_aal_flist, lm_sop_alloc_valid_aal_flist, OR, 1, <<31]
		; Reset RXC:crc.
		alu[lm_crc_residue, --, b, lm_crc_init]


		; Here the ref_cnt is zero.
		_aal5_rx_write_rxc_to_sram_cs1($rxc_out, rxc_addr, rxc_write_sig) 

		br!=ctx[7, ES_REF_CNT_NONZERO2#], defer[1]
		alu[sig_mask, sig_mask, OR, 1, <<(&rxc_write_sig)]


		; Signal next ME with CS1_inter_me_sig.
	   	alu[--, --, B, lm_cs1_inter_me_sig_val]
		cap[fast_wr, ALU, INTERTHREAD_SIG]	

		; Not getting RXC2 from SRAM, as we dont use it here in EOP_SOP case
		; So no writing back RXC2 to SRAM. So we can signal next ME here.

		; Signal next ME with CS2_inter_me_sig.
	   	alu[--, --, B, lm_cs2_inter_me_sig_val]
		cap[fast_wr, ALU, INTERTHREAD_SIG]	

		; Clear CAM for critical section2.
		cam_clear

ES_REF_CNT_NONZERO2#:

		br!=ctx[0, ES_THD_NOT_0_4#]

		alu[sig_mask, sig_mask, OR, 1, <<ENTER_CS2_INTER_ME_SIGNAL]

ES_THD_NOT_0_4#:

		; Ctx0 waits [CS2_inter_me_sig, next_thread_sig]
		; other ctx waits  [next_thread_sig, rxc_write_sig (if any)].
		local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
		ctx_arb[--] 

		; Signal the next thread.
		signal_ctx(next_thread_sig_val)

		; Not getting/using RXC2 from sram, so just waiting for next_thread_sig
		// _aal5_rx_get_rxc_cs2[rxc_addr]
	
		; So no need to decr the ref cnt
		// alu[*l$index0[RXC_LM_REF_CNT_OFFSET], *l$index0[RXC_LM_REF_CNT_OFFSET], -, 1]

		; To be in sync with other threads as we are not getting RXC2 from sram.
		ctx_arb[next_thread_sig]

		; Signal the next thread.
		signal_ctx(next_thread_sig_val)

		; Adding next_thread_sig to sig_mask
		alu[sig_mask, --, b, 1, <<&next_thread_sig]

		; Ctx0 waits for CS3_inter_me_sig
		br!=ctx[0, ES_THD_NOT_0_5#]

		alu[sig_mask, sig_mask, OR, 1, <<ENTER_CS3_INTER_ME_SIGNAL]

ES_THD_NOT_0_5#:	


	#endif  // TWO_ME_RX_FP


	; Get a new prefetch buffer so that it can used next time around.
	dl_buf_alloc[$prefetch_buf_handle, BUF_FREE_LIST0, bd_prefetch_sig, SIG_NONE]

	alu[sig_mask, sig_mask, OR, 1, <<(&bd_prefetch_sig)]



	; Ctx0 waits [CS3_inter_me_sig, bd_prefetch_sig, next_thread_sig]
	; other ctx waits  [next_thread_sig, bd_prefetch_sig].
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
	ctx_arb[--], defer[1]
	alu[dl_next_block, --, B, BID_NULL]		 	; set next_block NULL

 
ES_END#:

	.io_completed bd_prefetch_sig, data_write_sig, meta1_sig

#ifdef DL_SPLIT_META_DATA

	.io_completed meta1_lw0_sig

#endif

	#ifdef TWO_ME_RX_FP

		.io_completed rxc_write_sig
		xbuf_free($rxc_out)

	#endif


	xbuf_free($$cell_out)

.end
#endm   //   _aal5_rx_eop_sop_proc()


///////////////////////////////////////////////////////////////////////////////
//
// _aal5_rx_sop_only_proc
//
//	 	Description: 
//
//			Processes a cell containing only SOP.
//
//	 	Inputs:
//			sig_mask :   mask to be set with various I/O signals.
//
//			rxc_adr :    absolute rxc addr used to write back RXC to SRAM.
//
//			$$cell_in : dram xfer regs used to write to DRAM.
//
//			$prefetch_buf_handle - already prefetched new buffer handle.
//
//		Outputs:
//
//		Size: 
//
//			ONE_ME_RX_FP  -   52 cycles
//			TWO_ME_RX_FP  -  102 cycles
//
///////////////////////////////////////////////////////////////////////////////
#macro _aal5_rx_sop_only_proc(sig_mask, rxc_addr, $$cell_in, $prefetch_buf_handle)
.begin

	

	.reg buf_offset
	.reg offset
	.reg final_crc_remainder
	.reg dram_data_addr
	.reg tmp_byte_count


	// CRC -init in the defer slot of jump table in aal5_rx[] needs 3 instr before 
	// we can use crc instr. Because of defer slot in br[SOP#], it is taken care of.


	#ifdef TWO_ME_RX_FP

		// Allocate xfer reg for writing RXC back to SRAM.
		xbuf_alloc($rxc_out, 8, write)

	#endif


	// Allocate sxfer reg to write data out to DRAM.
	xbuf_alloc($$cell_out, 12, write)				// no code generated.

		
	; CRC compute.

	; This requires 1 instr/nop between two crc instr.

	crc_be[crc_32,  $$cell_out[0],  $$cell_in[1]]           
		; Clear RXC:exp_SOP bit.
		_aal5_rx_clear_rxc_exp_sop(lm_sop_alloc_valid_aal_flist)

	crc_be[crc_32,  $$cell_out[1],  $$cell_in[2]]           
		nop
	crc_be[crc_32,  $$cell_out[2],  $$cell_in[3]]          
		nop
	crc_be[crc_32,  $$cell_out[3],  $$cell_in[4]]           
		nop
	crc_be[crc_32,  $$cell_out[4],  $$cell_in[5]]           
		nop
	crc_be[crc_32,  $$cell_out[5],  $$cell_in[6]]           
		nop
	crc_be[crc_32,  $$cell_out[6],  $$cell_in[7]]           
		nop
	crc_be[crc_32,  $$cell_out[7],  $$cell_in[8]]           
		nop
	crc_be[crc_32,  $$cell_out[8],  $$cell_in[9]]           
		nop
	crc_be[crc_32,  $$cell_out[9],  $$cell_in[10]]           
		nop
	crc_be[crc_32,  $$cell_out[10], $$cell_in[11]]
		nop
	crc_be[crc_32,  $$cell_out[11], $$cell_in[12]]        

	; This requires 5 nops/useful instr between crc_be[] and local_csr_rd[] 
	
	; Get random offset used to do DRAM bank scheduling.	 
	_aal5_rx_get_rand_buf_offset(buf_offset)

	; Get dram address for the next buffer 
	dl_buf_get_data(dram_data_addr, $prefetch_buf_handle)


	#ifdef ONE_ME_RX_FP  


		; Read the actual CRC remainder.
		local_csr_rd[CRC_REMAINDER]
		immed[final_crc_remainder,0]

		; Update the RXC:crc_residue.
		alu[lm_crc_residue, --, B, final_crc_remainder]


	#else	// TWO_ME_RX_FP


		; Write back RXC1 to SRAM (fill in the defer slots with RXC1 updates).

S_REF_CNT_CHK#:

		alu[--, *l$index0[RXC_LM_REF_CNT_OFFSET], -, 0]
		bne[S_REF_CNT_NONZERO#], defer[3]
		; Read the actual CRC remainder.
		local_csr_rd[CRC_REMAINDER]
		immed[final_crc_remainder,0]
		; Update the RXC:crc_residue.
		alu[lm_crc_residue, --, B, final_crc_remainder]

	
		; Here the ref_cnt is zero.
		_aal5_rx_write_rxc_to_sram_cs1($rxc_out, rxc_addr, rxc_write_sig)

		br!=ctx[7, S_REF_CNT_NONZERO#], defer[1]
			alu[sig_mask, sig_mask, OR, 1, <<(&rxc_write_sig)]

		; Signal next ME CS1_inter_me_sig.
	   	alu[--, --, B, lm_cs1_inter_me_sig_val]
		cap[fast_wr, ALU, INTERTHREAD_SIG]	

		; Clear CAM for critical section2.
		cam_clear


S_REF_CNT_NONZERO#:

		br!=ctx[0, S_THD_NOT_0#]

		alu[sig_mask, sig_mask, OR, 1, <<ENTER_CS2_INTER_ME_SIGNAL]

S_THD_NOT_0#:


		; Ctx0 waits [CS2_inter_me_sig, next_thread_sig]
		; other ctx waits  [next_thread_sig, rxc_write_sig (if any)].
		local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
		ctx_arb[--] 

		.io_completed rxc_write_sig

S_PHASE4#:

		; Signal the next thread.
		signal_ctx(next_thread_sig_val)


		; Get RXC2 from SRAM  
		; rxc_addr is the absolute address of RXC, which is used as a cam tag.
		_aal5_rx_get_rxc_cs2(rxc_addr)

S_PHASE5#:

		; Decrement the reference count by 1.
		alu[*l$index0[RXC_LM_REF_CNT_OFFSET], *l$index0[RXC_LM_REF_CNT_OFFSET], -, 1]

		; Signal the next thread.
		signal_ctx(next_thread_sig_val)

		; Add next_thread_sig to sig_mask.
		alu[sig_mask, --, B, 1, <<&next_thread_sig]

	#endif // TWO_ME_RX_FP


	; Check if prefetch buffer handle is zero.
	alu[--,  --,  B, $prefetch_buf_handle, <<8]   ; [23:0] contains buf handle
	beq[S_NULL_BUF_PROC#]						  ; if null, error condition.

	; We got a valid buffer here.

	; AAL5 RX updates LW 0,1,2,3(31:16),5 of meta data. The rest of the meta 
	; data are not touched by AAL5 RX.
	
	; LW 0,1 are written to SRAM by AAL5 RX on EOP/buffer_overflow.
	; So here on SOP_only there is no meta write to SRAM, but meta info is 
	; set in the RXC.


	#ifdef NON_AAL5_TX

		;	Cell Count Calculation. 
	
			// There are two possibilities for output.

			//[1] For CSIX TX, for every packet prepend hdr is added.
			//	  (apart from a cell hdr for every cell). Include this hdr in
			//	  cell count calculation. This per packet prepend header is added
			//    on fly and so its not counted in buffer size/packet_size. This is
			// 	  applicable only for first CSIX frame, ie SOP_only case.

			//    CSIX cell is 120B. For first CSIX cell of the packet, we
			//    have packet prepend hdr of 8B. So first CSIX cell is 112B.

			//[2] For AAL5 TX, the cell_size is 48B.(we dont have to compute 
			//    cell_cnt here)

	
		; Set the RXC:byte_remainder = 40B + Per_pkt_prepend_bytes.
		alu[tmp_byte_count, lm_non_eop_cell_data_sz, +, PER_PKT_PREPEND_BYTES]
		ld_field[lm_prv_bf_sz_remain_cell_cnt, 0010, tmp_byte_count, <<8]

		; Clear RXC:cell count.
		ld_field[lm_prv_bf_sz_remain_cell_cnt, 0001, 0x0]  	

	#endif
	

   ; Update RXC

   	; Clear the drop_bit which may have been set for the previous frame.
	; Ideally we would do this at the end of EOP case, but as we dont write 
	; back RXC2 to SRAM on EOP case, we unclear the bit here.


	_aal5_rx_clear_rxc_drop_bit(lm_drp_ex_hdr_prv_prv_buf_addr)

	; Update RXC:offset to point to the header.
	_aal5_rx_set_rxc_offset(lm_sop_buf_sz_offset, buf_offset)


	; Update RXC:curr_buf_handle.
	
	; New buffer handle always has EOP bit and SOP bit set, with seg_cnt = 0
	; No need to clear the EOP bit here. We would clear EOP bit when we hit 
	; buffer overflow without getting EOP.
	alu[lm_curr_buf_handle, --, b, $prefetch_buf_handle]

	#ifdef ONE_ME_RX_FP


		; Update RXC:buf_offset for the next cell (next cell would be written at
		; location).
		alu[offset, buf_offset, +, NONEOP_CELL_DATA_SZ]
		_aal5_rx_set_rxc_buf_offset(lm_buf_offset_curr_pkt_sz, offset)

		; Update RXC:curr_pkt_size = 48 (this includes LLC/SNAP header, as the final 
		; pkt_len obtained from trailer includes LLC/SNAP header, Data, padding).
		_aal5_rx_set_rxc_curr_pkt_size(lm_buf_offset_curr_pkt_sz,  NONEOP_CELL_DATA_SZ)

		alu[sig_mask, sig_mask, OR, 3, <<(&data_write_sig)]

		alu[sig_mask, sig_mask, OR, 1, <<(&bd_prefetch_sig)]


	#else  // TWO_ME_RX_FP

		; Write back RXC2 to SRAM (fill in the defer slots with RXC2 updates).

S_REF_CNT_CHK2#:

		alu[--, *l$index0[RXC_LM_REF_CNT_OFFSET], -, 0]
		bne[S_REF_CNT_NONZERO2#], defer[3]
		; Update RXC:buf_offset for the next cell (next cell would be written 
		; at location).
		alu[offset, buf_offset, +, NONEOP_CELL_DATA_SZ]
	    _aal5_rx_set_rxc_buf_offset(lm_buf_offset_curr_pkt_sz, offset)
		; Update RXC:curr_pkt_size = 48 (this includes LLC/SNAP header, as final
		; pkt_len obtained from trailer includes LLC/SNAP header, Data, padding).
		_aal5_rx_set_rxc_curr_pkt_size(lm_buf_offset_curr_pkt_sz,  NONEOP_CELL_DATA_SZ)


		; Here the ref_cnt is zero.
		_aal5_rx_write_rxc_to_sram_cs2($rxc_out, rxc_addr, rxc_write_sig)

		br!=ctx[7, S_REF_CNT_NONZERO2#], defer[1]
		alu[sig_mask, sig_mask, OR, 1, <<(&rxc_write_sig)]

		; Signal next ME CS2_inter_me_sig.
	   	alu[--, --, B, lm_cs2_inter_me_sig_val]
		cap[fast_wr, ALU, INTERTHREAD_SIG]	

S_REF_CNT_NONZERO2#:


		br!=ctx[0, S_NOT_THD_0_2#], defer[2]
		alu[sig_mask, sig_mask, OR, 3, <<(&data_write_sig)]
		alu[sig_mask, sig_mask, OR, 1, <<(&bd_prefetch_sig)]


		; Add CS3_inter_me_sig to sig_mask.
		alu[sig_mask, sig_mask, OR, 1, <<ENTER_CS3_INTER_ME_SIGNAL]

S_NOT_THD_0_2#:


	#endif  //  TWO_ME_RX_FP

	; Data write to DRAM  (8B * 6 = 48B, including LLC/SNAP hdr, if any).
	dram[write, $$cell_out[0], dram_data_addr, buf_offset,6], \
				sig_done[data_write_sig] 


	; Prefetch the buffer handle ( 2-3 instr).
	dl_buf_alloc($prefetch_buf_handle, BUF_FREE_LIST0, bd_prefetch_sig, \
				SIG_NONE)

#ifdef DL_SPLIT_META_DATA
	alu[sig_mask, sig_mask, OR, 1, <<(&bd_prefetch_sig)]
#endif

	; Ctx0 waits [CS3_inter_me_sig, bd_prefetch_sig, data_write_sig, 
	;			  next_thread_sig, rxc_write_sig (if any)]
	; other ctx waits [bd_prefetch_sig, data_write_sig, next_thread_sig,
	;				   rxc_write_sig (if any)].

	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
	ctx_arb[--] , defer[1], br[S_END#]
	alu[dl_next_block, --, B, BID_NULL]		 	; set next_block NULL

;------------------------------------------------------------------------------
S_NULL_BUF_PROC#:
;------------------------------------------------------------------------------

	; We got a null buffer. So nothing to do. 


    ; Set RXC2:drop_bit so that subsequent cells of the same packet are dropped.
	_aal5_rx_set_rxc_drop_bit(lm_drp_ex_hdr_prv_prv_buf_addr)

	; Update Stat counter.
	_aal5_rx_incr_port_counter[input_port, AAL5_RX_PKT_DROP_COUNT]


	#ifdef TWO_ME_RX_FP

		; So write back only the RXC2:LW containing drop_bit
S_REF_CNT_CHK3#:

		alu[--, *l$index0[RXC_LM_REF_CNT_OFFSET], -, 0]
		bne[S_REF_CNT_NONZERO3#]

		; Here the ref_cnt is zero
		alu[$rxc_out[0], --, B, lm_drp_ex_hdr_prv_prv_buf_addr]
		sram[write,$rxc_out0, rxc_addr, 24, 1], sig_done[rxc_write_sig]

		br!=ctx[7, S_REF_CNT_NONZERO3#], defer[1]
		alu[sig_mask, sig_mask, OR, 1, <<(&rxc_write_sig)]

		; Signal next ME with CS2_inter_me_sig.
		alu[--, --, B, lm_cs2_inter_me_sig_val]
		cap[fast_wr, ALU, INTERTHREAD_SIG]	

S_REF_CNT_NONZERO3#:

		br!=ctx[0, S_NOT_THD_0_3#]

		; Add CS3_inter_me_sig to sig_mask
		alu[sig_mask, sig_mask, OR, 1, <<ENTER_CS3_INTER_ME_SIGNAL]


S_NOT_THD_0_3#:


	#endif // TWO_ME_RX_FP


	; Get a new prefetch buffer so that it can used next time around.
	dl_buf_alloc($prefetch_buf_handle, BUF_FREE_LIST0, bd_prefetch_sig, \
				SIG_NONE)
	alu[sig_mask, sig_mask, OR, 1, <<(&bd_prefetch_sig)]


	; Ctx0 waits [CS3_inter_me_sig, bd_prefetch_sig, next_thread_sig]
	; other ctx waits [bd_prefetch_sig, next_thread_sig, rxc_write_sig (if any)].
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
	ctx_arb[--] , defer[1]
	alu[dl_next_block, --, B, BID_NULL]		 ; set next_block NULL.

 
S_END#:


	.io_completed bd_prefetch_sig, data_write_sig, rxc_write_sig  

	#ifdef TWO_ME_RX_FP
		
		xbuf_free($rxc_out)

	#endif

	xbuf_free($$cell_out)							// no code generated	


.end
#endm   //  _aal5_rx_sop_only_proc()


///////////////////////////////////////////////////////////////////////////////
//
// _aal5_rx_mop_only_proc
//
//	 	Description: 
//
//			Processes a cell containing only MOP.
//
//	 	Inputs:
//			sig_mask :   mask to be set with various I/O signals.
//
//			rxc_adr :    absolute rxc addr used to write back RXC to SRAM.
//
//			$$cell_in : dram xfer regs used to write to DRAM.
//
//			input_port : The port through which the packet came in.
//
//			$prefetch_buf_handle - already prefetched new buffer handle.
//
//		Outputs:
//
//		Size: 
//			Buff_not_full :
//				ONE_ME_RX_FP    -  62 cycles
//				TWO_ME_RX_FP	 -  113 cycles
//			
//			Buff full_SOP :
//				ONE_ME_RX_FP    -  80 cycles
//				TWO_ME_RX_FP	 -  128 cycles		 						
//
//			Buff full_!SOP :
//				ONE_ME_RX_FP    -  86 cycles
//				TWO_ME_RX_FP	 -  134 cycles							
//
//
///////////////////////////////////////////////////////////////////////////////

#macro _aal5_rx_mop_only_proc(sig_mask, rxc_addr, $$cell_in, input_port, \
			$prefetch_buf_handle)
.begin

	

	.reg buf_offset
	.reg new_buf_offset
	.reg t_buf_offset

	.reg prev_buf_size
	.reg prv_buf_handle

	.reg offset

	.reg final_crc_remainder
	.reg curr_pkt_size

	.reg dram_buf_addr
	.reg cell_count

	.reg curr_handle
	.reg sop_handle
	.reg prefetch_handle

	.reg tmp

	// CRC -init in the defer slot of jump table in aal5_rx[] needs 3 instr before 
	// we can use crc instr. Because of defer slot in br[MOP#], it is taken care of.

	//	Allocate sxfer registers to read in data from the RBUF element.
	xbuf_alloc($$cell_out, 12, write)					// no code generated


	; CRC compute.

	; This requires 1 instr/nop between two crc instr.

	crc_be[crc_32,  $$cell_out[0],  $$cell_in[1]]           
		nop
	crc_be[crc_32,  $$cell_out[1],  $$cell_in[2]]           
		nop
	crc_be[crc_32,  $$cell_out[2],  $$cell_in[3]]           
		nop
	crc_be[crc_32,  $$cell_out[3],  $$cell_in[4]]           
		nop
	crc_be[crc_32,  $$cell_out[4],  $$cell_in[5]]           
		nop
	crc_be[crc_32,  $$cell_out[5],  $$cell_in[6]]           
		nop
	crc_be[crc_32,  $$cell_out[6],  $$cell_in[7]]           
		nop
	crc_be[crc_32,  $$cell_out[7],  $$cell_in[8]]           
		nop
	crc_be[crc_32,  $$cell_out[8],  $$cell_in[9]]           
		nop
	crc_be[crc_32,  $$cell_out[9],  $$cell_in[10]]           
		nop
	crc_be[crc_32,  $$cell_out[10], $$cell_in[11]]
		nop
	crc_be[crc_32,  $$cell_out[11], $$cell_in[12]]        

	nop
	nop
	nop
	nop
	nop
	// This requires 5 nops/useful instr between crc_be[] and local_csr_rd[] 

	#ifdef ONE_ME_RX_FP

		; Read the actual CRC remainder.
		local_csr_rd[CRC_REMAINDER]
		immed[final_crc_remainder,0]

		; Update the RXC:crc_residue.
		alu[lm_crc_residue, --, B, final_crc_remainder] 


	#else 	// TWO_ME_RX_FP


		// Allocate xfer registers for writing context back to SRAM.
		xbuf_alloc($rxc_out, 8, write)

	    ; Write back RXC1 to SRAM ( use defer slots to update RXC1).

M_REF_CNT_CHK#:

		alu[--, *l$index0[RXC_LM_REF_CNT_OFFSET], -, 0]
		bne[M_REF_CNT_NONZERO#], defer[3]
		; Read the actual CRC remainder.
		local_csr_rd[CRC_REMAINDER]
		immed[final_crc_remainder,0]
		; Update the RXC:crc_residue.
		alu[lm_crc_residue, --, B, final_crc_remainder] 


		; Here the ref_cnt is zero.

		_aal5_rx_write_rxc_to_sram_cs1($rxc_out, rxc_addr, rxc_write_sig) 

		br!=ctx[7, M_REF_CNT_NONZERO#], defer[1]
		alu[sig_mask, sig_mask, OR, 1, <<(&rxc_write_sig)]


		; Signal NEXT ME CS1_inter_me_sig.
	   	alu[--,--,B,lm_cs1_inter_me_sig_val]
		cap[fast_wr, ALU, INTERTHREAD_SIG]	

		; Clear CAM for critical section2.
		cam_clear

	
M_REF_CNT_NONZERO#:

		br!=ctx[0, M_THD_NOT_0#]

		alu[sig_mask, sig_mask, OR, 1, <<ENTER_CS2_INTER_ME_SIGNAL]

M_THD_NOT_0#:

		; Ctx0 waits [CS2_inter_me_sig, next_thread_sig]
		; other ctx waits  [next_thread_sig, rxc_write_sig (if any)].
		local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
		ctx_arb[--] 

		.io_completed rxc_write_sig

M_PHASE4#:

		; Signal the next thread.
		signal_ctx(next_thread_sig_val)

		; Get RXC2 from SRAM (15 - cam_miss).
		; rxc_addr is the absolute addr of RXC, which is used as cam tag.
		_aal5_rx_get_rxc_cs2(rxc_addr)

M_PHASE5#:

		; Decrement the reference count by 1.
		alu[*l$index0[RXC_LM_REF_CNT_OFFSET], *l$index0[RXC_LM_REF_CNT_OFFSET], \
				-, 1]

		; Signal the next thread.
		signal_ctx(next_thread_sig_val)

		alu[sig_mask, --, B, 1, <<&next_thread_sig]

	#endif // TWO_ME_RX_FP


	; Check if RXC2:drop_bit (bit 31) is set.
	br_bset[lm_drp_ex_hdr_prv_prv_buf_addr, 31, M_DROP#]

	; AAL5 RX updates LW 0,1,2,3(31:16),5 of meta data. The rest of the meta data 
	; are not touched by AAL5 RX.

	; LW 0,1 are written to SRAM by AAL5 RX on EOP/buffer_overflow.


	; RXC updates

	; Increment the RXC:curr_pkt_size with the bytes of the current cell.
	alu[curr_pkt_size, mask_lower_half, AND, lm_buf_offset_curr_pkt_sz]
	alu[curr_pkt_size, curr_pkt_size, +, NONEOP_CELL_DATA_SZ]
	ld_field[lm_buf_offset_curr_pkt_sz, 0011, curr_pkt_size]


	; Get the RXC:buf_offset.
	; Buf_offset is where the current cell will be stored in DRAM.
	ld_field_w_clr[buf_offset, 0011, lm_buf_offset_curr_pkt_sz, >>16]


	; Increment the RXC:buf_offset with the bytes of the current cell.
	; This is our updated buffer size.
	alu[new_buf_offset, NONEOP_CELL_DATA_SZ, +, buf_offset]


	; Buffer always has the space for the current cell. So we can write the
	; current cell. Check whether the next cell can be accomodated in
	; the current buffer.
	alu[t_buf_offset, NONEOP_CELL_DATA_SZ, +, new_buf_offset]

    alu[--, t_buf_offset, -, lm_max_buf_size_aal5]

	ble[M_BUF_NOT_FULL#], defer[1]
	alu[curr_handle, --, b, lm_curr_buf_handle]    


M_BUF_OVERFLOW#:

	;	At this point buffer will be full after storing the current
	;	cell. See if a new buffer (prefetched buffer) is available.

	alu[--, --, B, $prefetch_buf_handle, <<8]	; [23:0] contains buf handle
	beq[M_NULL_BUF_PROC#]

	
	#ifdef NON_AAL5_TX

		; Increment cell count depending on the number of bytes received.	
		_aal5_rx_incr_cell_count(cell_count, lm_prv_bf_sz_remain_cell_cnt, 
				NONEOP_CELL_DATA_SZ)

		; Embed this cell count in the buffer handle, [29:24] are cell count.
		alu[curr_handle, curr_handle, OR, cell_count, <<24]

		; Clear cell count and byte reminder for the new buffer.
		ld_field[lm_prv_bf_sz_remain_cell_cnt, 0001, 0x0]  				
		ld_field[lm_prv_bf_sz_remain_cell_cnt, 0010, 0x0, <<8] 	

	#endif
		
	; Check whether the current buffer is SOP buffer or not
	; (i.e a buffer containing start of packet.)           
		
	br_bset[curr_handle, 30, M_BUF_FULL_SOP#], defer[1]	; [30] is SOP bit
	; Clear EOP [31] bit, because this buffer will not contain EOP.
	alu_shf[curr_handle, curr_handle, AND~, 0x1, <<31]	


;---------------------------------------------------------------------------------
M_BUF_FULL_NOT_SOP#:
;---------------------------------------------------------------------------------

	;	Current buffer is full and it does not contain SOP cell.

	;	Two different cases to handle here. 
	;	[1] Previous Buffer is SOP. ( this has some headroom at the top)
	;	[2] Previous Buffer is not SOP. (this doesnt)


	;	Check if the previous buffer is a SOP buffer.
	;	If it is not a SOP buffer, the offset is set as NON_SOP_BUFF_OFFSET(0).
	;	If it is a SOP buffer we have stored the offset in RXC2:offset.

	xbuf_alloc($mop_meta,6, write)
	
	alu[prv_buf_handle, --, B, lm_prv_buf_handle]	

	br_bclr[prv_buf_handle,30,M_BUF_FULL_NOT_SOP_META_WR#], defer[3]
	; RXC:prev_to_prev_buf_handle = RXC:prev_buf_handle
	ld_field[lm_drp_ex_hdr_prv_prv_buf_addr, 0111, prv_buf_handle]
	; Get prev_buff_size.
	ld_field_w_clr[prev_buf_size, 0011, lm_prv_bf_sz_remain_cell_cnt, >>16]
	; Now make the prefetch buffer as current buffer. Any new
	; cell received will be stored in this new buffer. 
	alu[lm_curr_buf_handle, $prefetch_buf_handle, AND~, 0x1, <<30] ; clear sop bit



	; Here, previous buffer is a SOP buffer.

	; Get offset from RXC:offset.
	alu[offset, lm_sop_buf_sz_offset, AND, mask_lower_half]	
	
	; Flush META data of previous buffer to SRAM.

 
#ifdef OPTIONAL_L2PORT
	alu[l2port, lm_optional_info, AND, mask_lower_half]		;extract L2port
#endif

#ifdef OPTIONAL_SUID
	alu[suid, --, B, lm_optional_info]		;extract SUID
#endif


#ifdef METADATA_2LW

	dl_meta_write($mop_meta, prv_buf_handle, curr_handle, offset, \
	   			  prev_buf_size, 0, 0, 0, 0, 0, \
				  0, 0, 0, 0, 0, \
				  0, meta1_sig, SIG_NONE, 0, 2)

#else

	;prepare clp bit to put in meta data
	alu[clp, --, B, clp, <<CLP_BIT_OFFSET_IN_DL_META4]


    #ifdef DL_SPLIT_META_DATA
	    dl_meta_write($mop_meta, prv_buf_handle, curr_handle, offset, \
		   			  prev_buf_size, zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_lw0_sig, SIG_NONE, 0, 1)
	    dl_meta_write($mop_meta, prv_buf_handle, curr_handle, offset, \
		   			  prev_buf_size, zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_sig, SIG_NONE, 1, 5)

    #else /* DL_SPLIT_META_DATA */

	    dl_meta_write($mop_meta, prv_buf_handle, curr_handle, offset, \
		   			  prev_buf_size, zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_sig, SIG_NONE, 0, 6)

    #endif /* DL_SPLIT_META_DATA */

#endif // METADATA_2LW


	br[M_RXC_UPD#], defer[2]
	; We are done with current buffer. Make it previous buffer.
	alu_shf[lm_prv_buf_handle, --, B, curr_handle]
	; Set the RXC:prev_buf_size = new_buf_offset.
	ld_field[lm_prv_bf_sz_remain_cell_cnt, 1100, new_buf_offset, <<16]

M_BUF_FULL_NOT_SOP_META_WR#:

	; Here, previous buffer is a NOT SOP buffer.

	; Flush META data of previous buffer to SRAM.	


#ifdef OPTIONAL_L2PORT
	alu[l2port, lm_optional_info, AND, mask_lower_half]		;extract L2port
#endif

#ifdef OPTIONAL_SUID
	alu[suid, --, B, lm_optional_info]		;extract SUID
#endif


#ifdef METADATA_2LW

	dl_meta_write($mop_meta, prv_buf_handle, curr_handle, \
				  NON_SOP_BUFF_OFFSET, prev_buf_size, 0, 0, 0, 0, 0, \
				  0, 0, 0, 0, 0, \
				  0, meta1_sig, SIG_NONE, 0, 2)
#else

	;prepare clp bit to put in meta data
	alu[clp, --, B, clp, <<CLP_BIT_OFFSET_IN_DL_META4]

    #ifdef DL_SPLIT_META_DATA

	    dl_meta_write($mop_meta, prv_buf_handle, curr_handle, \
					  NON_SOP_BUFF_OFFSET, prev_buf_size, zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_lw0_sig, SIG_NONE, 0, 1)
	    dl_meta_write($mop_meta, prv_buf_handle, curr_handle, \
					  NON_SOP_BUFF_OFFSET, prev_buf_size, zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_sig, SIG_NONE, 1, 5)

    #else /* DL_SPLIT_META_DATA */

	    dl_meta_write($mop_meta, prv_buf_handle, curr_handle, \
					  NON_SOP_BUFF_OFFSET, prev_buf_size, zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_sig, SIG_NONE, 0, 6)

    #endif /* DL_SPLIT_META_DATA */

#endif // METADATA_2LW

	; We are done with current buffer. Make it previous buffer.
	alu_shf[lm_prv_buf_handle, --, B, curr_handle]

	; Set the RXC:prev_buf_size = new_buf_offset.
	ld_field[lm_prv_bf_sz_remain_cell_cnt, 1100, new_buf_offset, <<16]


M_RXC_UPD#:
	
	#ifdef ONE_ME_RX_FP

		; Set RXC:buf_offset as 0.
		ld_field[lm_buf_offset_curr_pkt_sz, 1100, NON_SOP_BUFF_OFFSET, <<16]

		alu[sig_mask, sig_mask, OR, 1, <<(&meta1_sig)]

        #ifdef DL_SPLIT_META_DATA

		    alu[sig_mask, sig_mask, OR, 1, <<(&meta1_lw0_sig)]		

        #endif        		

		alu[sig_mask, sig_mask, OR, 3, <<(&data_write_sig)]

		alu[sig_mask, sig_mask, OR, 1, <<(&bd_prefetch_sig)]


	#else 	// TWO_ME_RX_FP


		; Write back RXC2 to SRAM ( use defer slots for RXC2 updates) 

M_REF_CNT_CHK2#:

		alu[--, *l$index0[RXC_LM_REF_CNT_OFFSET], -, 0]
        
        #ifdef DL_SPLIT_META_DATA

    		bne[M_REF_CNT_NONZERO2#], defer[3]

    		; Set RXC:buf_offset as 0.
    		ld_field[lm_buf_offset_curr_pkt_sz, 1100, NON_SOP_BUFF_OFFSET, <<16]	
    		alu[sig_mask, sig_mask, OR, 1, <<(&meta1_sig)]	
    		alu[sig_mask, sig_mask, OR, 1, <<(&meta1_lw0_sig)]

        #else /* DL_SPLIT_META_DATA */

    		bne[M_REF_CNT_NONZERO2#], defer[2]
    		; Set RXC:buf_offset as 0.
    		ld_field[lm_buf_offset_curr_pkt_sz, 1100, NON_SOP_BUFF_OFFSET, <<16]	
    		alu[sig_mask, sig_mask, OR, 1, <<(&meta1_sig)]		

        #endif /* DL_SPLIT_META_DATA */
	
		; Here the ref_cnt is zero.

		_aal5_rx_write_rxc_to_sram_cs2($rxc_out, rxc_addr, rxc_write_sig) 

		br!=ctx[7, M_REF_CNT_NONZERO2#], defer[1]
		alu[sig_mask, sig_mask, OR, 1, <<(&rxc_write_sig)]


		; Signal NXT ME CS2_inter_me_sig.
	   	alu[--, --, B, lm_cs2_inter_me_sig_val]
		cap[fast_wr, ALU, INTERTHREAD_SIG]	

M_REF_CNT_NONZERO2#:

		br!=ctx[0, M_NOT_THD_0_2#], defer[2]
		alu[sig_mask, sig_mask, OR, 3, <<(&data_write_sig)]
		alu[sig_mask, sig_mask, OR, 1, <<(&bd_prefetch_sig)]

		// Add CS3_inter_me_sig to sig_mask.
		alu[sig_mask, sig_mask, OR, 1, <<ENTER_CS3_INTER_ME_SIGNAL]

M_NOT_THD_0_2#:

	#endif // TWO_ME_RX_FP


	; Write  the received cell to DRAM  (8B * 6 = 48B).
	dl_buf_get_data(dram_buf_addr, curr_handle) 		
	dram[write, $$cell_out[0], dram_buf_addr, buf_offset, 6],sig_done[data_write_sig]



	; Prefetch a new buffer so that it can be used next time around.
	dl_buf_alloc($prefetch_buf_handle, BUF_FREE_LIST0, bd_prefetch_sig, SIG_NONE) 

    #ifdef DL_SPLIT_META_DATA
	
    alu[sig_mask, sig_mask, OR, 1, <<(&bd_prefetch_sig)]
    
    #endif


	; Ctx0 waits [CS3_inter_me_sig, data_write_sig, next_thread_sig,
	;              meta1_sig, bd_prefetch_sig, rxc_write_sig (if any)]
	; other ctx waits [data_write_sig, next_thread_sig, meta1_sig, 
	;					bd_prefetch_sig, rxc_write_sig (if any)].
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
	ctx_arb[--] , defer[1], br[M_END#]
	alu[dl_next_block, --,B, BID_NULL]			; set next_block NULL.

	xbuf_free($mop_meta)

;---------------------------------------------------------------------------------	
M_BUF_FULL_SOP#:
;---------------------------------------------------------------------------------

	;  Here, the current buffer is full. This buffer contains SOP. 
	;  A new valid buffer is available.

	;  At this point, no meta data needs to be flushed to SRAM
	;  Store the values in the RXC.

	;  We are done with current buffer. Make it previous buffer as well as
	;  SOP Buffer.
	alu[lm_sop_buf_handle, --, B, curr_handle]					
	alu[lm_prv_buf_handle, --, B, curr_handle]	


	;	Now make the prefetch buffer as current buffer. Any new
	;	cell received will be stored in this new buffer.
	alu[lm_curr_buf_handle, $prefetch_buf_handle, AND~, 0x1, <<30]	; clear sop bit


	; Set RXC:buf_offset as 0 for the next buffer.
	ld_field[lm_buf_offset_curr_pkt_sz, 1100, NON_SOP_BUFF_OFFSET, <<16]

	alu[offset, mask_lower_half, AND, lm_sop_buf_sz_offset]

	#ifdef ONE_ME_RX_FP

		;  Set RXC:prev_buf_size = (new_buf_offset - RXC:offset).
		alu[prev_buf_size, new_buf_offset, -, offset]
		ld_field[lm_prv_bf_sz_remain_cell_cnt, 1100, prev_buf_size, <<16]
		;  Set RXC:sop_buf_size	 
		ld_field[lm_sop_buf_sz_offset, 1100, prev_buf_size, <<16]	 

		alu[sig_mask, sig_mask, OR, 3, <<(&data_write_sig)]

		alu[sig_mask, sig_mask, OR, 1, <<(&bd_prefetch_sig)]

	#else  // TWO_ME_RX_FP


		; Write back RXC2 to SRAM, (use defer slots for RXC2 updates).

M_REF_CNT_CHK3#:

		alu[--, *l$index0[RXC_LM_REF_CNT_OFFSET], -, 0]
		bne[M_REF_CNT_NONZERO3#], defer[3]
		;  Set RXC:prev_buf_size = (new_buf_offset - RXC:offset).
		alu[prev_buf_size, new_buf_offset, -, offset]
		ld_field[lm_prv_bf_sz_remain_cell_cnt, 1100, prev_buf_size, <<16]	 
		;  Set RXC:sop_buf_size	 
		ld_field[lm_sop_buf_sz_offset, 1100, prev_buf_size, <<16]	 


		; Here the ref_cnt is zero.
		_aal5_rx_write_rxc_to_sram_cs2($rxc_out, rxc_addr, rxc_write_sig) 

		br!=ctx[7, M_REF_CNT_NONZERO3#], defer[1]
			alu[sig_mask, sig_mask, OR, 1, <<(&rxc_write_sig)]
	
		; Signal next ME CS2_inter_me_sig.
	   	alu[--, --, B, lm_cs2_inter_me_sig_val]
		cap[fast_wr, ALU, INTERTHREAD_SIG]	

M_REF_CNT_NONZERO3#:

		br!=ctx[0, M_NOT_THD_0_3#], defer[2]

		alu[sig_mask, sig_mask, OR, 3, <<(&data_write_sig)]
		alu[sig_mask, sig_mask, OR, 1, <<(&bd_prefetch_sig)]

		; Add CS3_inter_me_sig to sig_mask.
		alu[sig_mask, sig_mask, OR, 1, <<ENTER_CS3_INTER_ME_SIGNAL]

M_NOT_THD_0_3#:


	#endif  // TWO_ME_RX_FP


	; Write the received cell to DRAM  (8B * 6 = 48B).
	dl_buf_get_data(dram_buf_addr, curr_handle) 				
	dram[write, $$cell_out[0], dram_buf_addr, buf_offset, 6],sig_done[data_write_sig]


	; Prefetch a new buffer so that it can be used next time around. 
	dl_buf_alloc($prefetch_buf_handle, BUF_FREE_LIST0, bd_prefetch_sig, SIG_NONE)

    #ifdef DL_SPLIT_META_DATA

	    alu[sig_mask, sig_mask, OR, 1, <<(&bd_prefetch_sig)]

    #endif

	; Ctx0 waits [CS3_inter_me_sig, data_write_sig, next_thread_sig,bd_prefetch_sig
	;				rxc_write_sig (if any)]
	; other ctx waits [data_write_sig, next_thread_sig,	bd_prefetch_sig, 
	;			rxc_write_sig (if any)].

	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
	ctx_arb[--] , defer[1], br[M_END#]
	alu[dl_next_block, --, B, BID_NULL] 		; set next_block NULL
	

;-----------------------------------------------------------------------------------
M_BUF_NOT_FULL#:
;-----------------------------------------------------------------------------------

	;  MOP received and current buffer is not FULL. (i.e it can take current 
	;  cell + atleast one more cell).

	#ifdef NON_AAL5_TX

		;Increment cell count depending on the number of bytes received.  
		 _aal5_rx_incr_cell_count(cell_count, lm_prv_bf_sz_remain_cell_cnt, \
		 			NONEOP_CELL_DATA_SZ)

	#endif


	#ifdef ONE_ME_RX_FP

		; Update RXC:buf_offset for the next_cell.
		ld_field[lm_buf_offset_curr_pkt_sz, 1100, new_buf_offset, <<16]

		alu[sig_mask, sig_mask, OR, 3, <<(&data_write_sig)]

	#else    // TWO_ME_RX_FP


		; Write back RXC2 to SRAM, use defer slots for RXC2 updates.

M_REF_CNT_CHK4#:

		alu[--, *l$index0[RXC_LM_REF_CNT_OFFSET], -, 0]
		bne[M_REF_CNT_NONZERO4#], defer[1]
		
		; Update RXC:buf_offset for the next_cell.
		ld_field[lm_buf_offset_curr_pkt_sz, 1100, new_buf_offset, <<16]

		; Here the ref_cnt is zero.

		_aal5_rx_write_rxc_to_sram_cs2($rxc_out, rxc_addr, rxc_write_sig) 

		br!=ctx[7, M_REF_CNT_NONZERO4#], defer[1]
		alu[sig_mask, sig_mask, OR, 1, <<(&rxc_write_sig)]

		; Signal next ME enter_cs2_inter_me_sig.
	   	alu[--, --, B, lm_cs2_inter_me_sig_val]
		cap[fast_wr, ALU, INTERTHREAD_SIG]	

M_REF_CNT_NONZERO4#:

		br!=ctx[0, M_NOT_THD_0_4#], defer[1]
		alu[sig_mask, sig_mask, OR, 3, <<(&data_write_sig)]

		; Add CS3_inter_me_sig to sig_mask.
		alu[sig_mask, sig_mask, OR, 1, <<ENTER_CS3_INTER_ME_SIGNAL]

M_NOT_THD_0_4#:


	#endif // TWO_ME_RX_FP


	; Write  the received cell to DRAM  (8B * 6 = 48B).
	dl_buf_get_data(dram_buf_addr, lm_curr_buf_handle) 			
	dram[write, $$cell_out[0], dram_buf_addr, buf_offset, 6],sig_done[data_write_sig]

	; Ctx0 waits [CS3_inter_me_sig, data_write_sig, next_thread_sig, 
	;				rxc_write_sig (if any)]
	; other ctx waits [data_write_sig, next_thread_sig,rxc_write_sig (if any)].
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
	ctx_arb[--] , defer[1], br[M_END#]
	alu[dl_next_block, --, B, BID_NULL] 		; set next_block to NULL
	
;-------------------------------------------------------------------------------
M_NULL_BUF_PROC#:
;-------------------------------------------------------------------------------

	; We encountered buffer overflow for the next cell and we got a null buffer. 

   
	; Set RXC2:drop_bit.  
	_aal5_rx_set_rxc_drop_bit(lm_drp_ex_hdr_prv_prv_buf_addr)

	; So DROP the packet reassembled till now. Packets are dropped in dl_sink
	; as we should wait for the DRAM write of previous thread to complete.

	; Update Stat counter.
	_aal5_rx_incr_port_counter[input_port, AAL5_RX_PKT_DROP_COUNT]

	#ifdef TWO_ME_RX_FP

		 ; Write back only the RXC2:LW containing drop_bit.
M_REF_CNT_CHK5#:

		alu[--, *l$index0[RXC_LM_REF_CNT_OFFSET], -, 0]
		bne[M_REF_CNT_NONZERO5#]

		; Here the ref_cnt is zero.

		alu[$rxc_out[0], --, B, lm_drp_ex_hdr_prv_prv_buf_addr]

		sram[write,$rxc_out[0], rxc_addr, 24, 1], sig_done[rxc_write_sig]

		br!=ctx[7, M_REF_CNT_NONZERO5#], defer[1]
			alu[sig_mask, sig_mask, OR, 1, <<(&rxc_write_sig)]

		; Signal next ME with CS2_inter_me_sig.
		alu[--, --, B, lm_cs2_inter_me_sig_val]
		cap[fast_wr, ALU, INTERTHREAD_SIG]	

M_REF_CNT_NONZERO5#:


		br!=ctx[0, M_NOT_THD_0_5#]

		; Add CS3_inter_me_sig to sig_mask.
		alu[sig_mask, sig_mask, OR, 1, <<ENTER_CS3_INTER_ME_SIGNAL]

M_NOT_THD_0_5#:


	#endif

	; In case of non_sop_buffer_overflow we have to update the meta data of 
	; the previous buffer with the current one. (we havent done this till now, 
	; as we were waiting to get EOP/to fill up the current buffer. But we would 
	; need this to drop the buffer chain using DROP queue.)

	; In case of sop_buffer_overflow, we just have one buffer to drop, so
	; no need to write meta data to SRAM.

	; The curr_handle should have EOP bit set, as it is the last buffer of
	; the chain to be dropped. Here curr_handle already has EOP bit set.

	; Check whether the current buffer is SOP buffer or not
	; (i.e a buffer containing start of packet.)        

	br_bset[curr_handle, 30, M_NULL_BUF_FULL_SOP#]			; [30] is SOP bit

	xbuf_alloc($prev_meta, 1, write)

	alu[prv_buf_handle, --, B, lm_prv_buf_handle]

	; Flush META data of the previous buffer to update just LW0 with
	; the curr_handle.
	dl_meta_write($prev_meta, prv_buf_handle, curr_handle, 0, \
		   			  0, 0, 0, 0, 0, 0, \
					  0, 0, 0, 0, 0, \
					  0, meta1_sig, SIG_NONE, 0, 1)

    alu[sig_mask, sig_mask, OR, 1, <<(&meta1_sig)]

	br[M_BD_ALLOC#], defer[2]
	; Set variables to drop buffer chain.
	alu[dl_buf_handle, --, b, lm_sop_buf_handle]
	alu[dl_eop_buf_handle, --, B, curr_handle]	 


M_NULL_BUF_FULL_SOP#:

	; Set variables to drop single buffer.
	alu[dl_buf_handle, --, B, curr_handle]	; dl_buf_handle = curr_handle = sop_handle 
	alu[dl_eop_buf_handle, --, B, DL_BUF_NULL]

M_BD_ALLOC#:

	; Get a new prefetch buffer so that it can used next time around.
	dl_buf_alloc($prefetch_buf_handle, BUF_FREE_LIST0, bd_prefetch_sig, SIG_NONE)
	alu[sig_mask, sig_mask, OR, 1, <<(&bd_prefetch_sig)]

	; Ctx0 waits [CS3_inter_me_sig, next_thread_sig,bd_prefetch_sig, 
	;				rxc_write_sig (if any)]
	; other ctx waits [next_thread_sig, bd_prefetch_sig, rxc_write_sig (if any)].
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
	ctx_arb[--] , defer[1], br[M_END#]
	alu[dl_next_block, --, B, BID_DROP]		 	; set next_block BID_DROP

	xbuf_free($prev_meta) 
		
;---------------------------------------------------------------------------------
M_DROP#:
;---------------------------------------------------------------------------------

	; Here, drop bit was set from previous iteration so nothing to do.

	#ifdef TWO_ME_RX_FP


	 ; Write back only the RXC2:LW containing drop_bit. For large frames
	 ; spread over more than one ME, its required that we write out
	 ; the RXC2:drop bit, so that next ME doesnt get stale data from SRAM.

M_REF_CNT_CHK6#:

		alu[--, *l$index0[RXC_LM_REF_CNT_OFFSET], -, 0]
		bne[M_REF_CNT_NONZERO6#]

		; Here the ref_cnt is zero.

		alu[$rxc_out[0], --, B, lm_drp_ex_hdr_prv_prv_buf_addr]

		sram[write,$rxc_out[0], rxc_addr, 24, 1], sig_done[rxc_write_sig]

		br!=ctx[7, M_REF_CNT_NONZERO6#], defer[1]
			alu[sig_mask, sig_mask, OR, 1, <<(&rxc_write_sig)]

		; Signal next ME with CS2_inter_me_sig.
		alu[--, --, B, lm_cs2_inter_me_sig_val]
		cap[fast_wr, ALU, INTERTHREAD_SIG]	

M_REF_CNT_NONZERO6#:


		br!=ctx[0, M_NOT_THD_0_6#]

		; Add CS3_inter_me_sig to sig_mask.
		alu[sig_mask, sig_mask, OR, 1, <<ENTER_CS3_INTER_ME_SIGNAL]

M_NOT_THD_0_6#:


	#endif // TWO_ME_RX_FP


	; Ctx0 waits [CS3_inter_me_sig, next_thread_sig]
	; other ctx waits [next_thread_sig].
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
	ctx_arb[--] , defer[1], br[M_END#]
	alu[dl_next_block, --, B, BID_NULL]		 	; set next_block NULL


M_END#:

	 .io_completed bd_prefetch_sig, data_write_sig, meta1_sig, rxc_write_sig  

    #ifdef DL_SPLIT_META_DATA

	    .io_completed meta1_lw0_sig

    #endif

	  xbuf_free($$cell_out)


	#ifdef TWO_ME_RX_FP
	
		xbuf_free($rxc_out)	

	#endif

.end
#endm   //   _aal5_rx_mop_only_proc()

///////////////////////////////////////////////////////////////////////////////
//
// _aal5_rx_eop_only_proc
//
//	 	Description: 
//
//			Processes a cell containing only EOP. It also processes when
//			there is an extra_cell containing just the trailer.
//
//	 	Inputs:
//			sig_mask :   mask to be set with various I/O signals.
//
//			rxc_adr :    absolute rxc addr used to write back RXC to SRAM.
//
//			$$cell_in :  dram xfer regs used to write to DRAM.
//
//			input_port : The port through which the packet came in.
//
//			vc_id :      index of RXC into RXC table (VC info table)
//
//			pti :        3 bits of PTI of ATM header.
//
//
//		Outputs:
//
//		Size: 
//			pkt in one buf  
//				ONE_ME_RX_FP    -  92 cycles
//				TWO_ME_RX_FP	 -  131 cycles							
//
//			pkt in one+ buf  
//				ONE_ME_RX_FP    - 101 cycles
//				TWO_ME_RX_FP	 -  140 cycles								
//
///////////////////////////////////////////////////////////////////////////////
#macro _aal5_rx_eop_only_proc(sig_mask, rxc_addr, $$cell_in, input_port, vc_id, pti)

.begin

	
	.reg buf_offset
	.reg offset

	.reg final_crc_remainder
	.reg curr_pkt_size

	.reg curr_handle
	.reg sop_handle
	.reg prefetch_handle

	.reg prv_buf_handle
	.reg prv_buf_size
	.reg sop_buf_size

	.reg buf_size

	.reg tmp

	.reg pkt_length

	.reg bytes_in_last_cell

	.reg dram_buf_addr
	.reg dram_ref_cnt
	.reg cell_count

	.reg excep_bit
	.reg hdr_type


	//	Allocate sxfer reg to read in data from the RBUF element.
	xbuf_alloc($$cell_out, 12, write)				

	; CRC compute.

	; This requires 1 instr/nop between two crc instr.
	crc_be[crc_32,  $$cell_out[0],  $$cell_in[1]]           
		; Get the packet length from 3rd and 4th B of 8B trailer 
		; (includes LLC/SNAP hdr too).
		alu[pkt_length, mask_lower_half, AND, $$cell_in[11]]

	crc_be[crc_32,  $$cell_out[1],  $$cell_in[2]]           
		dl_meta_set_packet_size(pkt_length)	 	    // Set packet size.
		
	crc_be[crc_32,  $$cell_out[2],  $$cell_in[3]]           
		dl_meta_set_flow_id(vc_id)
	crc_be[crc_32,  $$cell_out[3],  $$cell_in[4]]           
		nop	
	crc_be[crc_32,  $$cell_out[4],  $$cell_in[5]]           
		nop
	crc_be[crc_32,  $$cell_out[5],  $$cell_in[6]]           
		nop
	crc_be[crc_32,  $$cell_out[6],  $$cell_in[7]]           
		nop
	crc_be[crc_32,  $$cell_out[7],  $$cell_in[8]]           
		nop
	crc_be[crc_32,  $$cell_out[8],  $$cell_in[9]]           
		nop
	crc_be[crc_32,  $$cell_out[9],  $$cell_in[10]]           
		nop
	crc_be[crc_32,  $$cell_out[10], $$cell_in[11]]
		nop
	crc_be[crc_32,  $$cell_out[11], $$cell_in[12]]        

	; This requires 5 nops between crc_be[] and local_csr_rd[] 
	dl_meta_set_free_list(FREE_LIST_ID)			
	dl_meta_set_rx_stat(0)	
	
	#ifdef ONE_ME_RX_FP
		
		nop

		; Set RXC:exp_SOP bit.
		alu_shf[lm_sop_alloc_valid_aal_flist, lm_sop_alloc_valid_aal_flist, OR, 1, <<31]

		; Reset RXC:crc.
		alu[lm_crc_residue, --, b, lm_crc_init]

		; Read the actual CRC remainder.
		local_csr_rd[CRC_REMAINDER]
		immed[final_crc_remainder,0]

		; Compare the actual CRC remainder with the fixed remainder.
		alu[--,final_crc_remainder, -, lm_fixed_crc_remainder]
		bne[E_CRC_ERROR#]
	

	#else  // TWO_ME_RX_FP


		; Set RXC:exp_SOP bit.
		alu_shf[lm_sop_alloc_valid_aal_flist, lm_sop_alloc_valid_aal_flist, OR, 1, <<31]

		// Allocate xfer registers for writing context back to SRAM.
		xbuf_alloc($rxc_out, 8, write)

		; Write back RXC1 to SRAM ( use defer slots for RXC1 updates).

E_REF_CNT_CHK#:

		alu[--, *l$index0[RXC_LM_REF_CNT_OFFSET], -, 0]
		bne[E_REF_CNT_NONZERO#], defer[3]
		; Reset RXC:crc.
		alu[lm_crc_residue, --, b, lm_crc_init]
		; Read the actual CRC remainder.
		local_csr_rd[CRC_REMAINDER]
		immed[final_crc_remainder,0]


		; Here the ref_cnt is zero.

		_aal5_rx_write_rxc_to_sram_cs1($rxc_out, rxc_addr, rxc_write_sig)

		br!=ctx[7, E_REF_CNT_NONZERO#], defer[1]
			alu[sig_mask, sig_mask, OR, 1, <<(&rxc_write_sig)]

		; Signal NEXT ME enter_cs1_inter_me_sig.
	   	alu[--, --, B, lm_cs1_inter_me_sig_val]
		cap[fast_wr, ALU, INTERTHREAD_SIG]

		; CAM clear for critical section2.
		cam_clear

E_REF_CNT_NONZERO#:

		br!=ctx[0, E_THD_NOT_0#]

		alu[sig_mask, sig_mask, OR, 1, <<&enter_cs2_inter_me_sig]

E_THD_NOT_0#:

		; Ctx0 waits [CS2_inter_me_sig, next_thread_sig,rxc_write_sig (if any)]
		; other ctx waits  [next_thread_sig, rxc_write_sig (if any)].
		local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
		ctx_arb[--] 
	
		.io_completed rxc_write_sig

E_PHASE4#:

		; Signal the next thread.
		signal_ctx(next_thread_sig_val)

		; Compare the actual CRC remainder with the fixed remainder.
		alu[--,final_crc_remainder, -, lm_fixed_crc_remainder]
		bne[E_CRC_ERROR#] 

		; Get RXC2 from SRAM (14 - cam_miss).
		; rxc_addr is the absolute address of RXC, which is used as cam tag.
		_aal5_rx_get_rxc_cs2(rxc_addr)

E_PHASE5#:

		; On EOP, no need to flush RXC2, as the state is not required to be passed
		; to next ME, for the next new packet. State required is in RXC1 which has 
		; already been flushed in critical section 1. RXC2:Drop bit would be reset on
		; SOP case (There is no need to reset RXC2:drop bit in EOP_SOP case, as its
		; not using it)

		; So just signal next ME here, if thread 7.

		; Decrement the reference count by 1
		alu[*l$index0[RXC_LM_REF_CNT_OFFSET], *l$index0[RXC_LM_REF_CNT_OFFSET], -, 1]

		br!=ctx[7, E_NOT_THD_7#], defer[2]
	
		; Signal the next thread.
		signal_ctx(next_thread_sig_val)
		alu[sig_mask, --, B, 1, <<&next_thread_sig]

		; Signal next ME CS2_inter_me_sig, after getting RXC2 from SRAM.
		alu[--, --, B, lm_cs2_inter_me_sig_val]
		cap[fast_wr, ALU, INTERTHREAD_SIG]	

E_NOT_THD_7#:

	#endif   // TWO_ME_RX_FP



	#ifdef ONE_ME_RX_FP

		; Get RXC:curr_pkt_size (this also includes LLC/SNAP header, if any),
		; this curr_pkt_size doesnt include curr_cell rcvd.
		alu[curr_pkt_size, mask_lower_half, AND, lm_buf_offset_curr_pkt_sz]

	#else  // TWO_ME_RX_FP

		br!=ctx[0, E_NOT_THD_0_2#], defer[1]
		
		; Get RXC:curr_pkt_size (this also includes LLC/SNAP header, if any).
		alu[curr_pkt_size, mask_lower_half, AND, lm_buf_offset_curr_pkt_sz]



		; Add CS3_inter_me_sig to sig_mask.
		alu[sig_mask, sig_mask, OR, 1, <<ENTER_CS3_INTER_ME_SIGNAL]

E_NOT_THD_0_2#:

	#endif  // TWO_ME_RX_FP


	; Check if RXC2:drop_bit (bit 31) is set.
	br_bset[lm_drp_ex_hdr_prv_prv_buf_addr, 31, E_DROP#]

	; Check if the (curr_pkt_size - pkt_length) > 7.
	; This check takes care of two conditions
	;	(1)   pkt_length = 0
	;   (2)   if two or more AAL5 frames are combined into one because of 
	;		  loss of EOP cell in between (where somehow CRC check has passed).
	alu[tmp, curr_pkt_size, -, pkt_length]
	alu[--,tmp, -, 0x7]
	bgt[E_INV_FR_SZ_ERROR#]

	; AAL5 RX updates LW 0,1,2,3(31:16),5 of meta data. The rest of the meta 
	; data are not touched by AAL5 RX.

	; LW 0,1 are written to SRAM by AAL5 RX on EOP/buffer_overflow.

	; Pkt_length includes payload including LLC/SNAP header, but no padding.
	; curr_pkt_size doesnt include curr_cell rcvd, still if it is > than 
	; pkt_length, then its an EXTRA_CELL with just trailer.

	; Check for an Extra last cell(having AAL5 trailer only).
	alu[--, curr_pkt_size, -, pkt_length]
	bge[E_EXTRA_CELL#], defer[1] 
	alu[curr_handle, --, b, lm_curr_buf_handle]

;------------------------------------------------------------------------------
E_NORMAL_EOP#:
;------------------------------------------------------------------------------	

	; Here its Normal EOP cell.

	; Check if pkt_length is within MAX AAL5 frame size.
	alu[--, pkt_length, -, lm_max_aal5_fr_sz]
	bgt[E_INV_FR_SZ_ERROR#]	

	_aal5_rx_incr_vc_counter[vc_id, AAL5_RX_AAL5_PDU_COUNT]
	_aal5_rx_incr_port_counter[input_port, AAL5_RX_PKT_COUNT]


	; Check whether the exception bit is set or not.
	; If yes, then next block = IX_EXCEPTION.

	br_bset[lm_drp_ex_hdr_prv_prv_buf_addr, 30, E_EXCEPTION#], defer[3]
	; Calculate bytes in the last cell - this doesnt include padding & AAL5 
	; Trailer, as packet_length includes only payload excluding padding.
	alu[bytes_in_last_cell, pkt_length, -, curr_pkt_size]
	; Get RXC:buf_offset (this is where current cell will be written to DRAM).
	ld_field_w_clr[buf_offset, 0011, lm_buf_offset_curr_pkt_sz, >>16]
	_aal5_rx_get_rxc_header_type(hdr_type, lm_drp_ex_hdr_prv_prv_buf_addr)	
		

	br[E_CHECK_BUF#], defer[2]
	alu[dl_next_block, --, B,AAL5_RX_NEXT1]	; set next_block AAL5_RX_NEXT1
	dl_meta_set_header_type(hdr_type) 				

	
E_EXCEPTION#: 

	alu[dl_next_block, --, B, IX_EXCEPTION]	; set next_block IX_EXCEPTION
	dl_meta_set_header_type(hdr_type) 	

E_CHECK_BUF#:


	#ifdef MPLS_OVER_ATM
		.if (hdr_type == PKT_PROTO_MPLS)
			.reg tos_lbl
			; For MPLS over ATM case the aal5 Rx block needs to convey topmost label 
			; information ie, vpi/vci fields to MPLS blocks. dl_flow_id field is used
			; to convey this information
			; This is required only in eop only and not required in sop-eop case.
			; Because an MPLS over ATM frame will require at least two cells.
			; 4B Shim header +40B IPv4 header + 8B aal5 tailor which is greater than 48B

			; Update only VPI/VCI fields
			ld_field_w_clr[tos_lbl, 0111, $$cell_in0, >>4]
			dl_meta_set_flow_id (tos_lbl)
		.endif
	#endif

	#ifdef NON_AAL5_TX

		; Increment cell count depending on the number of bytes received.
		_aal5_rx_incr_cell_count(cell_count, lm_prv_bf_sz_remain_cell_cnt, \
					bytes_in_last_cell)

		; Handle already has EOP bit set. Embed cell count.
		alu[curr_handle, curr_handle, OR, cell_count, <<24]   

	#endif



#ifdef OPTIONAL_L2PORT

	alu[l2port, lm_optional_info, AND, mask_lower_half]		;extract L2port
	dl_meta_set_input_port(l2port) 		
#else
	dl_meta_set_input_port(input_port) 

#endif /* OPTIONAL_L2PORT */

	; Check if current buffer is a sop buffer.  
    br_bclr[curr_handle, 30, E_NOT_SOP_BUF#],defer[2]	; [30] is SOP bit.
	;  This is the curr buffer_size.
	alu[buf_size, buf_offset, +, bytes_in_last_cell]	
	; Get RXC2:offset.
	alu[offset, mask_lower_half, AND, lm_sop_buf_sz_offset]




;--------------------------------------------------------------------------------
E_SOP_BUF#:
;--------------------------------------------------------------------------------

	;	Yes, this buffer has SOP. i.e this buffer now contains
	;	a complete packet (SOP ...EOP). 

	;  So we need to handle only a single buffer. 

	alu[buf_size, buf_size, -, offset]

	;	Write the meta data for the SOP-EOP buffer to SRAM. 

	xbuf_alloc($eop_sop_meta, 6, write)	


#ifdef OPTIONAL_L2PORT
	alu[l2port, lm_optional_info, AND, mask_lower_half]		;extract L2port
#endif

#ifdef OPTIONAL_SUID
	alu[suid, --, B, lm_optional_info]		;extract SUID
#endif


#ifdef METADATA_2LW

	dl_meta_write($eop_sop_meta, curr_handle, IX_NULL, offset, buf_size, \
					  0, 0, 0, 0, 0, \
					  0, 0, 0, 0, 0, \
					  0, meta1_sig, SIG_NONE, 0, 2)
#else

	;prepare clp bit to put in meta data
	alu[clp, --, B, clp, <<CLP_BIT_OFFSET_IN_DL_META4]

    #ifdef DL_SPLIT_META_DATA

    	dl_meta_write($eop_sop_meta, curr_handle, IX_NULL, offset, buf_size, \
    					  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
    					  0, clp, 0, SUID_NUMBER, 0, \
    					  0, meta1_lw0_sig, SIG_NONE, 0, 1)
    	dl_meta_write($eop_sop_meta, curr_handle, IX_NULL, offset, buf_size, \
    					  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
    					  0, clp, 0, SUID_NUMBER, 0, \
    					  0, meta1_sig, SIG_NONE, 1, 5)

    	alu[sig_mask, sig_mask, OR, 1, <<(&meta1_lw0_sig)]

    #else /* DL_SPLIT_META_DATA */

    	dl_meta_write($eop_sop_meta, curr_handle, IX_NULL, offset, buf_size, \
					  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_sig, SIG_NONE, 0, 6)

    #endif /* DL_SPLIT_META_DATA */

#endif // METADATA_2LW

	 alu[sig_mask, sig_mask, OR, 1, <<(&meta1_sig)]

	; Flushing of RXC2 is not required on EOP.



	; Write last cell to dram, after rounding it off to 8B words
		
	dl_buf_get_data(dram_buf_addr, curr_handle) 

	; Calculate number of 8-byte words for writing to DRAM
	; Actual no of 8B words = (bytes_in_last_cell - 1)/8 + 1 
	; Indirect ref cnt for DRAM write requires one less, so it would be 
	; (bytes_in_last_cell - 1)/8.
	alu[dram_ref_cnt, bytes_in_last_cell, -, 1]
	alu[dram_ref_cnt, --, b, dram_ref_cnt, >>3]
	; set OV[25] bit to override ref_cnt
	alu[dram_ref_cnt, dram_ref_cnt, OR, 1, <<4]

	alu_shf[dram_ref_cnt, --, b, dram_ref_cnt, <<21]
	dram[write, $$cell_out[0], dram_buf_addr, buf_offset, max_5],indirect_ref,\
				 sig_done[data_write_sig]
	
	alu[sig_mask, sig_mask, OR, 3, <<(&data_write_sig)]


	;	The way packets are sent to next block in the pipeline is thro'
	;	dl_sink. (see the outer loop in aal5_rx.uc). 
	;	As part of optimisation, 4 LWs ( 1, 2, 3, 5) from meta data are sent 
	;	in the scratch ring. dl_sink takes these meta data from the 
	;	previous dl_meta_set_* calls. These meta data should belong to SOP buffer.

	; In our case the prev dl_set_* calls belong to the EOP buffer. 
	; So we	need to do a few dl_meta_set_* calls for SOP buffer here.

	; See the dl_meta_set_* in some of the defer slots of CRC and branch instr.

	dl_meta_set_offset(offset)
	alu[tmp, --, B, buf_size] 			; to avoid reg conflict.
	dl_meta_set_buffer_size(tmp)   		; set sop_buf_size.
		
	; Ctx0 waits [CS3_inter_me_sig, data_write_sig, next_thread_sig, meta1_sig]
	; other ctx waits  [data_write_sig, next_thread_sig, meta1_sig].
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
	ctx_arb[--] , defer[2], br[E_END#]
	alu[dl_eop_buf_handle, --, B, DL_BUF_NULL] 	; set eop_handle = NULL
	alu[dl_buf_handle, --, B, curr_handle] 		; Set dl_buf_handle = sop_handle

	xbuf_free[$eop_sop_meta] 				

;-------------------------------------------------------------------------------
E_NOT_SOP_BUF#:
;-------------------------------------------------------------------------------

	;	At this point, EOP and SOP are in different buffers i.e large packet.
	;   So we need to write previous buffer's meta data & current buffer's (EOP) 
	;	meta data to SRAM.
	;	There are two cases to handle.

	;	[1] Previous Buffer is SOP Buffer.
	;	[2] Previous Buffer is not SOP Buffer.

	;  curr_handle contains current buffer handle. cell count
	;  had already been updated in the buffer handle.

	xbuf_alloc($prev_meta1, 6, write)
	xbuf_alloc($eop_meta2, 6, write) 					

	;	Check if the previous buffer is a SOP buffer. 
	;  If it is not a SOP buffer, the offset is set as NON_SOP_OFFSET(0) 
	;  If it is a SOP buffer we have stored the offset in RXC:offset.

	alu[prv_buf_handle, --, B, lm_prv_buf_handle] 

	br_bclr[prv_buf_handle, 30, E_PRV_BUF_NOT_SOP#], defer[3] ; [30] is SOP bit
	; Get RXC2:prv_buf_size
	ld_field_w_clr[prv_buf_size, 0011, lm_prv_bf_sz_remain_cell_cnt, >>16]
	alu[dl_buf_handle, --, B, lm_sop_buf_handle] 	; set dl_buf_handle = sop_handle
	alu[sig_mask, sig_mask, OR, 3, <<(&data_write_sig)]	


	; Here previous buffer is SOP buffer.

	; Write META data of previous buffer to SRAM

#ifdef OPTIONAL_L2PORT
	alu[l2port, lm_optional_info, AND, mask_lower_half]		;extract L2port
#endif

#ifdef OPTIONAL_SUID
	alu[suid, --, B, lm_optional_info]		;extract SUID
#endif


#ifdef METADATA_2LW

	dl_meta_write($prev_meta1, prv_buf_handle, curr_handle, offset, \
					  prv_buf_size, 0, 0, 0, 0, 0, \
					  0, 0, 0, 0, 0, \
					  0, meta1_sig, SIG_NONE, 0, 2)
#else

	;prepare clp bit to put in meta data
	alu[clp, --, B, clp, <<CLP_BIT_OFFSET_IN_DL_META4]

    #ifdef DL_SPLIT_META_DATA

	    dl_meta_write($prev_meta1, prv_buf_handle, curr_handle, offset, \
					  prv_buf_size, zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_lw0_sig, SIG_NONE, 0, 1)
	    dl_meta_write($prev_meta1, prv_buf_handle, curr_handle, offset, \
					  prv_buf_size, zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_sig, SIG_NONE, 1, 5)

	    alu[sig_mask, sig_mask, OR, 1, <<(&meta1_lw0_sig)]
	    alu[sig_mask, sig_mask, OR, 1, <<(&meta2_lw0_sig)]

    #else /* DL_SPLIT_META_DATA */

	    dl_meta_write($prev_meta1, prv_buf_handle, curr_handle, offset, \
					  prv_buf_size, zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_sig, SIG_NONE, 0, 6)

    #endif	/* DL_SPLIT_META_DATA */

#endif  // METADATA_2LW


	br[E_EOP_META_WR#], defer[2]
	alu[sig_mask, sig_mask, OR, 1, <<(&meta1_sig)]
	alu[sig_mask, sig_mask, OR, 1, <<(&meta2_sig)]


E_PRV_BUF_NOT_SOP#:

	; Here previous buffer is NOT SOP buffer.

	; Write META data of previous buffer to SRAM.

#ifdef OPTIONAL_L2PORT
	alu[l2port, lm_optional_info, AND, mask_lower_half]		;extract L2port
#endif

#ifdef OPTIONAL_SUID
	alu[suid, --, B, lm_optional_info]		;extract SUID
#endif


#ifdef METADATA_2LW

	dl_meta_write($prev_meta1, prv_buf_handle, curr_handle, NON_SOP_BUFF_OFFSET, \
				      prv_buf_size, 0, 0, 0, 0, 0, \
					  0, 0, 0, 0, 0, \
					  0, meta1_sig, SIG_NONE, 0, 2)
#else

	;prepare clp bit to put in meta data
	alu[clp, --, B, clp, <<CLP_BIT_OFFSET_IN_DL_META4]

    #ifdef DL_SPLIT_META_DATA

    	dl_meta_write($prev_meta1, prv_buf_handle, curr_handle, NON_SOP_BUFF_OFFSET, \
    				      prv_buf_size, zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
    					  0, clp, 0, SUID_NUMBER, 0, \
    					  0, meta1_lw0_sig, SIG_NONE, 0, 1)
    	dl_meta_write($prev_meta1, prv_buf_handle, curr_handle, NON_SOP_BUFF_OFFSET, \
    				      prv_buf_size, zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
    					  0, clp, 0, SUID_NUMBER, 0, \
    					  0, meta1_sig, SIG_NONE, 1, 5)
	
    	alu[sig_mask, sig_mask, OR, 1, <<(&meta1_lw0_sig)]
    	alu[sig_mask, sig_mask, OR, 1, <<(&meta2_lw0_sig)]



    #else /* DL_SPLIT_META_DATA */

    	dl_meta_write($prev_meta1, prv_buf_handle, curr_handle, NON_SOP_BUFF_OFFSET, \
				      prv_buf_size, zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_sig, SIG_NONE, 0, 6)

    #endif /* DL_SPLIT_META_DATA */

#endif  // METADATA_2LW

	alu[sig_mask, sig_mask, OR, 1, <<(&meta1_sig)]
	alu[sig_mask, sig_mask, OR, 1, <<(&meta2_sig)]


E_EOP_META_WR#:

	; Write META data of EOP buffer to SRAM.


#ifdef OPTIONAL_L2PORT
	alu[l2port, lm_optional_info, AND, mask_lower_half]		;extract L2port
#endif

#ifdef OPTIONAL_SUID
	alu[suid, --, B, lm_optional_info]		;extract SUID
#endif


#ifdef METADATA_2LW

	dl_meta_write($eop_meta2, curr_handle, IX_NULL, NON_SOP_BUFF_OFFSET, buf_size, \
					  0, 0, 0, 0, 0, \
					  0, 0, 0, 0, 0, \
					  0, meta2_sig, SIG_NONE, 0, 2)

#else

	;prepare clp bit to put in meta data
	alu[clp, --, B, clp, <<CLP_BIT_OFFSET_IN_DL_META4]

    #ifdef DL_SPLIT_META_DATA

    	dl_meta_write($eop_meta2, curr_handle, IX_NULL, NON_SOP_BUFF_OFFSET, buf_size, \
					  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta2_lw0_sig, SIG_NONE, 0, 1)
    	dl_meta_write($eop_meta2, curr_handle, IX_NULL, NON_SOP_BUFF_OFFSET, buf_size, \
					  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta2_sig, SIG_NONE, 1, 5)

    #else /* DL_SPLIT_META_DATA */

	    dl_meta_write($eop_meta2, curr_handle, IX_NULL, NON_SOP_BUFF_OFFSET, buf_size, \
					  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta2_sig, SIG_NONE, 0, 6)

    #endif /* DL_SPLIT_META_DATA */

#endif  // METADATA_2LW

	; Flushing of RXC2 is not required on EOP.


	; Write last cell to dram, after rounding it off to 8B words.

	dl_buf_get_data(dram_buf_addr, curr_handle) ; (2-3)

	; Calculate number of 8-byte words for writing to DRAM
	; Actual no of 8B words = (bytes_in_last_cell - 1)/8 + 1 
	; Indirect ref cnt for DRAM write requires one less, so it would be 
	; (bytes_in_last_cell - 1)/8.
	alu[dram_ref_cnt, bytes_in_last_cell, -, 1]
	alu[dram_ref_cnt, --, b, dram_ref_cnt, >>3]
	; set OV[25] bit to override ref_cnt
	alu[dram_ref_cnt, dram_ref_cnt, OR, 1, <<4]


	alu_shf[dram_ref_cnt, --, b, dram_ref_cnt, <<21]
	dram[write, $$cell_out[0], dram_buf_addr, buf_offset, max_5],indirect_ref, \
			sig_done[data_write_sig]
	

	;	The way packets are sent to next block in the pipeline is thro'
	;	dl_sink. (see the outer loop in aal5_rx.uc). 
	;	As part of optimisation, 4 LWs ( 1, 2, 3, 5) from meta data are sent 
	;	in the scratch ring. dl_sink takes these meta data from the 
	;	previous dl_meta_set_* calls. These meta data should belong to SOP buffer.

	; In our case the prev dl_set_* calls belong to the EOP buffer. 
	; So we	need to do a few dl_meta_set_* calls for SOP buffer here.

	; See the dl_meta_set_* in some of the defer slots of CRC and branch instr.

	dl_meta_set_offset(offset)
	alu[sop_buf_size, --, b, lm_sop_buf_sz_offset, >>16]  ; get sop_buf_size

	; Ctx0 waits [CS3_inter_me_sig, data_write_sig, next_thread_sig,
	;              meta1_sig, meta2_sig]
	; other ctx waits  [data_write_sig, next_thread_sig, meta1_sig, meta2_sig].
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
	ctx_arb[--] , defer[2], br[E_END#]
	alu[dl_eop_buf_handle, --, B, curr_handle]	; set dl_eop_handle = last buf
	dl_meta_set_buffer_size(sop_buf_size)

	xbuf_free($prev_meta1)
	xbuf_free($eop_meta2)							

;--------------------------------------------------------------------------------
E_EXTRA_CELL#:
;--------------------------------------------------------------------------------

	; This time, we have got an extra_cell without data, with just 40B of padding and
	; 8B of trailer. 

	; Check if pkt_length is within MAX AAL5 frame size.
	alu[--, pkt_length, -, lm_max_aal5_fr_sz]
	bgt[E_INV_FR_SZ_ERROR#]	

	; Update stat counters.
	_aal5_rx_incr_vc_counter[vc_id, AAL5_RX_AAL5_PDU_COUNT]
	_aal5_rx_incr_port_counter[input_port, AAL5_RX_PKT_COUNT]


	; Check whether the exception bit is set or not.
	; If yes, then next block = IX_EXCEPTION.
	
	br_bset[lm_drp_ex_hdr_prv_prv_buf_addr, 30, E_EX_EXCEPTION#], defer[2]
	; The cell previous to this cell was the last cell with the
	; valid data and some padding too. So the actual bytes_in_last_cell for the 
	; previous iteration would be pkt_length - (curr_pkt_size - noneop_cell_size(48B)).
	; Calc bytes_in_last_cell.
	alu[tmp, curr_pkt_size, -, NONEOP_CELL_DATA_SZ]
	alu[bytes_in_last_cell, pkt_length, -, tmp]

	
	br[E_EX_CELL_HANDLING#], defer[2]
	alu[dl_next_block, --, B,AAL5_RX_NEXT1]	; set next_block AAL5_RX_NEXT1
	alu[buf_offset, --, b, lm_buf_offset_curr_pkt_sz]  // ld_field_w_clr wont work here			
	
E_EX_EXCEPTION#: 

	alu[dl_next_block, --, B, IX_EXCEPTION]		; set next_block IX_EXCEPTION
	alu[buf_offset, --, b, lm_buf_offset_curr_pkt_sz]  // ld_field_w_clr wont work here

E_EX_CELL_HANDLING#:


	; But NO DRAM WRITE to the current buffer (no need to write trailer to DRAM).

	;  Check whether a new buffer was allocated in the previous iteration 
	;  only for this extra_cell. If yes, then drop this buffer. In this case
	;  we have only one META to write ie for the PREV BUFFER.
	
	; Check if the buff_offset is 0, if yes, then we have allocated a new buffer in
	; the previous iteration for the extra_cell. So we will have to drop this buffer.

	_aal5_rx_get_rxc_header_type(hdr_type, lm_drp_ex_hdr_prv_prv_buf_addr)	

#ifdef MPLS_OVER_ATM
	.if (hdr_type == PKT_PROTO_MPLS)
		; For MPLS over ATM case the aal5 Rx block needs to convey topmost label 
		; information ie, vpi/vci fields to MPLS blocks. dl_flow_id field is used
		; to convey this information
		; This is required only in eop only and not required in sop-eop case.
		; Because an MPLS over ATM frame will require at least two cells.
		; 4B Shim header +40B IPv4 header + 8B aal5 tailor which is greater than 48B

		; Update only VPI/VCI fields
		ld_field_w_clr[tos_lbl, 0111, $$cell_in0, >>4]
		dl_meta_set_flow_id (tos_lbl)
	.endif 
#endif	

#ifdef OPTIONAL_L2PORT

	alu[l2port, lm_optional_info, AND, mask_lower_half]		;extract L2port
	dl_meta_set_input_port(l2port)         
#else
	dl_meta_set_input_port(input_port) 

#endif /* OPTIONAL_L2PORT */

	alu[buf_offset, --, b, buf_offset, >>16]
	beq[E_EX_CELL_NEW_BUF#], defer[1]
	dl_meta_set_header_type(hdr_type) 	
	


;---------------------------------------------------------------------------------
E_EX_CELL_NO_NEW_BUF#:
;---------------------------------------------------------------------------------

	; Here the new buffer wasnt allocated in the previous iteration for the extra 
	; cell.

	
	#ifdef NON_AAL5_TX

		; Get RXC:cell_count.
		alu[cell_count, 0xFF, AND, lm_prv_bf_sz_remain_cell_cnt]

		; Handle already has EOP bit set. Embed cell count.
		alu[curr_handle, curr_handle, OR, cell_count, <<24] 

	#endif
		
	
	;	Check if current buffer is a sop buffer
	br_bclr[curr_handle, 30, E_NO_NEW_BUF_NOT_SOP_BUF#], defer[3] ; [30] is SOP bit.
	; curr buf size = buf offset - noneop_cell_size + bytes_in_last_cell
	alu[buf_offset, buf_offset, -, NONEOP_CELL_DATA_SZ]	
	alu[buf_offset, buf_offset, +, bytes_in_last_cell]	; This is curr buf size.
	alu[offset, mask_lower_half, AND, lm_sop_buf_sz_offset] ; Get RXC:offset.


	
;---------------------------------------------------------------------------------
E_NO_NEW_BUF_SOP_BUF#:
;---------------------------------------------------------------------------------

	;	Yes, this buffer has SOP. i.e this buffer now contains
	;	a complete packet (SOP ...EOP). So we need to handle only
	;	a single buffer. Set all META data fields and write the 
	;	meta data to SRAM.

	alu[buf_offset, buf_offset, -, offset]		;  This is the sop buffer_size

	;	Write the meta data for the SOP-EOP buffer to SRAM.

	xbuf_alloc($ex_eop_sop_meta, 6, write)		//  alloc xfer regs for flush 

#ifdef OPTIONAL_L2PORT
	alu[l2port, lm_optional_info, AND, mask_lower_half]		;extract L2port
#endif

#ifdef OPTIONAL_SUID
	alu[suid, --, B, lm_optional_info]		;extract SUID
#endif


#ifdef METADATA_2LW

	dl_meta_write($ex_eop_sop_meta, curr_handle, IX_NULL, offset, buf_offset, \
					  0, 0, 0, 0, 0, \
					  0, 0, 0, 0, 0, \
					  0, meta1_sig, SIG_NONE, 0, 2)
#else

	;prepare clp bit to put in meta data
	alu[clp, --, B, clp, <<CLP_BIT_OFFSET_IN_DL_META4]

    #ifdef DL_SPLIT_META_DATA

    	dl_meta_write($ex_eop_sop_meta, curr_handle, IX_NULL, offset, buf_offset, \
					  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_lw0_sig, SIG_NONE, 0, 1)
	    dl_meta_write($ex_eop_sop_meta, curr_handle, IX_NULL, offset, buf_offset, \
					  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_sig, SIG_NONE, 1, 5)

	alu[sig_mask, sig_mask, OR, 1, <<(&meta1_lw0_sig)]

    #else /* DL_SPLIT_META_DATA */

	    dl_meta_write($ex_eop_sop_meta, curr_handle, IX_NULL, offset, buf_offset, \
					  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_sig, SIG_NONE, 0, 6)

    #endif /* DL_SPLIT_META_DATA */

#endif  // METADATA_2LW

	alu[sig_mask, sig_mask, OR, 1, <<(&meta1_sig)]

	; Flushing of RXC2 is not required on EOP.

	;	The way packets are sent to next block in the pipeline is thro'
	;	dl_sink. (see the outer loop in aal5_rx.uc). 
	;	As part of optimisation, 4 LWs ( 1, 2, 3, 5) from meta data are sent 
	;	in the scratch ring. dl_sink takes these meta data from the 
	;	previous dl_meta_set_* calls. These meta data should belong to SOP buffer.

	; In our case the prev dl_set_* calls belong to the EOP buffer. 
	; So we	need to do a few dl_meta_set_* calls for SOP buffer here.


	; Set other meta data for the next block
	dl_meta_set_offset(offset)
	alu[tmp, --, B, pkt_length]
	dl_meta_set_buffer_size(tmp)			; sop_buf_size = pkt_len


	; Ctx0 waits [CS3_inter_me_sig, next_thread_sig, meta1_sig]
	; other ctx waits  [next_thread_sig, meta1_sig].
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
	ctx_arb[--] , defer[2], br[E_END#]
	alu[dl_eop_buf_handle, --, B, DL_BUF_NULL] 	; set dl_eop_handle = NULL
	alu[dl_buf_handle, --, B, curr_handle] 		; set dl_buf_handle = sop_buffer 	


	xbuf_free($ex_eop_sop_meta)

;---------------------------------------------------------------------------------
E_NO_NEW_BUF_NOT_SOP_BUF#:
;---------------------------------------------------------------------------------
	; Here the new buffer was not allocated for the extra cell in the previous 
	; iteration. And the current buffer is not a SOP buffer.

	;   Two cases
	;	[1] Previous Buffer is SOP Buffer.
	;	[2] Previous Buffer is not SOP Buffer.

	;   curr_handle contains current buffer handle. cell count
	;	had already been updated in the buffer handle.

	;	Check if the previous buffer is a SOP buffer. 
	;   If it is not a SOP buffer, the offset is set as NON_SOP_OFFSET(0).
	;	If it is a SOP buffer we have stored the offset in RXC:offset.


	alu[prv_buf_handle, --, B, lm_prv_buf_handle] 

	xbuf_alloc($ex_prev_meta1, 6, write) 				//allocate xfer regs

	br_bclr[prv_buf_handle, 30, E_NO_NEW_BUF_PRV_BUF_NOT_SOP#], defer[2]; [30] is SOP bit
	ld_field_w_clr[prv_buf_size, 0011, lm_prv_bf_sz_remain_cell_cnt, >>16]
	alu[dl_buf_handle, --, B, lm_sop_buf_handle]   ; set dl_buf_handle = sop_handle


	;	Previous buffer is a SOP buffer. 

	;	Write META data of previous buffer to SRAM.


#ifdef OPTIONAL_L2PORT
	alu[l2port, lm_optional_info, AND, mask_lower_half]		;extract L2port
#endif

#ifdef OPTIONAL_SUID
	alu[suid, --, B, lm_optional_info]		;extract SUID
#endif


#ifdef METADATA_2LW

	dl_meta_write($ex_prev_meta1, prv_buf_handle, curr_handle, offset, prv_buf_size, \
					  0, 0, 0, 0, 0, \
					  0, 0, 0, 0, 0, \
					  0, meta1_sig, SIG_NONE, 0, 2)
#else

	;	prepare clp bit to put in meta data
	alu[clp, --, B, clp, <<CLP_BIT_OFFSET_IN_DL_META4]

    #ifdef DL_SPLIT_META_DATA

	    dl_meta_write($ex_prev_meta1, prv_buf_handle, curr_handle, offset, prv_buf_size, \
					  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_lw0_sig, SIG_NONE, 0, 1)
    	dl_meta_write($ex_prev_meta1, prv_buf_handle, curr_handle, offset, prv_buf_size, \
					  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_sig, SIG_NONE, 1, 5)

    	alu[sig_mask, sig_mask, OR, 1, <<(&meta1_lw0_sig)]
	    alu[sig_mask, sig_mask, OR, 1, <<(&meta2_lw0_sig)]

    #else /* DL_SPLIT_META_DATA */

    	dl_meta_write($ex_prev_meta1, prv_buf_handle, curr_handle, offset, prv_buf_size, \
					  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_sig, SIG_NONE, 0, 6)

    #endif /* DL_SPLIT_META_DATA */

#endif  // METADATA_2LW

	br[E_EX_EOP_META_WR#], defer[2]
	alu[sig_mask, sig_mask, OR, 1, <<(&meta1_sig)]
	alu[sig_mask, sig_mask, OR, 1, <<(&meta2_sig)]

;---------------------------------------------------------------------------------
E_NO_NEW_BUF_PRV_BUF_NOT_SOP#:
;---------------------------------------------------------------------------------
	;   Previous buffer is NOT SOP buffer.

	;	Write META data of previous buffer to SRAM.

#ifdef OPTIONAL_L2PORT
	alu[l2port, lm_optional_info, AND, mask_lower_half]		;extract L2port
#endif

#ifdef OPTIONAL_SUID
	alu[suid, --, B, lm_optional_info]		;extract SUID
#endif


#ifdef METADATA_2LW

	dl_meta_write($ex_prev_meta1, prv_buf_handle, curr_handle, NON_SOP_BUFF_OFFSET, \
					  prv_buf_size, 0, 0, 0, 0, 0, \
					  0, 0, 0, 0, 0, \
					  0, meta1_sig, SIG_NONE, 0, 2)
#else

    ; prepare clp bit to put in meta data
	alu[clp, --, B, clp, <<CLP_BIT_OFFSET_IN_DL_META4]


    #ifdef DL_SPLIT_META_DATA

    	dl_meta_write($ex_prev_meta1, prv_buf_handle, curr_handle, NON_SOP_BUFF_OFFSET, \
					  prv_buf_size, zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_lw0_sig, SIG_NONE, 0, 1)
	    dl_meta_write($ex_prev_meta1, prv_buf_handle, curr_handle, NON_SOP_BUFF_OFFSET, \
					  prv_buf_size, zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_sig, SIG_NONE, 1, 5)

    	alu[sig_mask, sig_mask, OR, 1, <<(&meta1_lw0_sig)]	
	    alu[sig_mask, sig_mask, OR, 1, <<(&meta2_lw0_sig)]
    
    #else /* DL_SPLIT_META_DATA */

    	dl_meta_write($ex_prev_meta1, prv_buf_handle, curr_handle, NON_SOP_BUFF_OFFSET, \
					  prv_buf_size, zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_sig, SIG_NONE, 0, 6)
    
    #endif /* DL_SPLIT_META_DATA */

#endif // METADATA_2LW
	
	alu[sig_mask, sig_mask, OR, 1, <<(&meta1_sig)]
	alu[sig_mask, sig_mask, OR, 1, <<(&meta2_sig)]

E_EX_EOP_META_WR#:

	xbuf_alloc($ex_eop_meta1, 6, write)					// allocate xfer regs

	;	Write META data of current/EOP buffer to SRAM.

#ifdef OPTIONAL_L2PORT
	alu[l2port, lm_optional_info, AND, mask_lower_half]		;extract L2port
#endif

#ifdef OPTIONAL_SUID
	alu[suid, --, B, lm_optional_info]		;extract SUID
#endif


#ifdef METADATA_2LW

	dl_meta_write($ex_eop_meta1, curr_handle, IX_NULL, NON_SOP_BUFF_OFFSET, buf_offset, \
					  0, 0, 0, 0, 0, \
					  0, 0, 0, 0, 0, \
					  0, meta2_sig, SIG_NONE, 0, 2)
#else

	;	prepare clp bit to put in meta data
	alu[clp, --, B, clp, <<CLP_BIT_OFFSET_IN_DL_META4]

    #ifdef DL_SPLIT_META_DATA

	    dl_meta_write($ex_eop_meta1, curr_handle, IX_NULL, NON_SOP_BUFF_OFFSET, buf_offset, \
					  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta2_lw0_sig, SIG_NONE, 0, 1)
    	dl_meta_write($ex_eop_meta1, curr_handle, IX_NULL, NON_SOP_BUFF_OFFSET, buf_offset, \
					  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta2_sig, SIG_NONE, 1, 5)

    #else /* DL_SPLIT_META_DATA */

	    dl_meta_write($ex_eop_meta1, curr_handle, IX_NULL, NON_SOP_BUFF_OFFSET, buf_offset, \
					  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta2_sig, SIG_NONE, 0, 6)

    #endif /* DL_SPLIT_META_DATA */

#endif  // METADATA_2LW

	;  Flushing of RXC2 is not required on EOP.

	;	The way packets are sent to next block in the pipeline is thro'
	;	dl_sink. (see the outer loop in aal5_rx.uc). 
	;	As part of optimisation, 4 LWs ( 1, 2, 3, 5) from meta data are sent 
	;	in the scratch ring. dl_sink takes these meta data from the 
	;	previous dl_meta_set_* calls. These meta data should belong to SOP buffer.

	; In our case the prev dl_set_* calls belong to the EOP buffer. 
	; So we	need to do a few dl_meta_set_* calls for SOP buffer here.

	dl_meta_set_offset(offset)						// set offset
	alu[sop_buf_size, --, B, lm_sop_buf_sz_offset, >>16]

	; Ctx0 waits [CS3_inter_me_sig, next_thread_sig, meta1_sig, meta2_sig]
	; other ctx waits  [next_thread_sig, meta1_sig, meta2_sig].
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
	ctx_arb[--] , defer[2], br[E_END#]
	alu[dl_eop_buf_handle, --, B, curr_handle]	; set dl_eop_handle = last buffer
	dl_meta_set_buffer_size(sop_buf_size)	 	; set sop buf size
			
	xbuf_free($ex_prev_meta1)
	xbuf_free($ex_eop_meta1)							

	
;---------------------------------------------------------------------------------
E_EX_CELL_NEW_BUF#:
;---------------------------------------------------------------------------------

	;  If we are here, it means we allocated a new buffer for the extra_cell.

	;  Two cases
	;  [1]  Previous buffer is SOP buffer ( one META write, curr buf dropped)
	;  [2]  Previous buffer is not SOP buffer ( two META Write, curr buf dropped)
		
	;  Check if the previous buffer is a SOP buffer. 

	; If it is not a SOP buffer, the offset is set as NON_SOP_OFFSET(0) 
	; If it is a SOP buffer we have stored the offset in RXC:offset
		
    ; Get the prev_handle
	alu[prv_buf_handle, --, B, lm_prv_buf_handle]

	br_bclr[prv_buf_handle, 30, E_NEW_BUF_PRV_BUF_NOT_SOP#], defer[3] ; [30] is SOP bit
	; Get RXC:prv_buf_size
	ld_field_w_clr[prv_buf_size, 0011, lm_prv_bf_sz_remain_cell_cnt, >>16]
	; The prev buffer size = prev buf size - noneop_cell_size + bytes_in_last_cell
	alu[prv_buf_size, prv_buf_size, -, NONEOP_CELL_DATA_SZ]
	alu[prv_buf_size, prv_buf_size, +, bytes_in_last_cell] 

;---------------------------------------------------------------------------------
E_NEW_BUF_PRV_BUF_SOP#:
;---------------------------------------------------------------------------------
	
	; Previous buffer is SOP buffer means, SOP and EOP in the same buffer.

	; So the new buffer allocated for the extra cell will be dropped here. 
	
	; Set the EOP bit in the prev buff handle (which was cleared before)
	
	; alu_shf requires unrestricted operand and LM[index 8 to 15] is restricted 
	; operand. So we have to move to register first
	; And we cannot have two LM as operands too.

	alu[prv_buf_handle, --,b, lm_prv_buf_handle]
	alu[prv_buf_handle, prv_buf_handle, OR, 0x1, <<31]

	alu[offset, mask_lower_half, AND, lm_sop_buf_sz_offset]

	; Write the meta data for the SOP-EOP buffer to SRAM.

	xbuf_alloc($ex_prev_meta, 6, write) 					; allocate xfer regs

#ifdef OPTIONAL_L2PORT
	alu[l2port, lm_optional_info, AND, mask_lower_half]		;extract L2port
#endif

#ifdef OPTIONAL_SUID
	alu[suid, --, B, lm_optional_info]		;extract SUID
#endif


#ifdef METADATA_2LW

	dl_meta_write($ex_prev_meta, prv_buf_handle, IX_NULL, offset, prv_buf_size, \
					  0, 0, 0, 0, 0, \
					  0, 0, 0, 0, 0, \
					  0, meta1_sig, SIG_NONE, 0, 2)

#else

	; prepare clp bit to put in meta data
	alu[clp, --, B, clp, <<CLP_BIT_OFFSET_IN_DL_META4]

    #ifdef DL_SPLIT_META_DATA

    	dl_meta_write($ex_prev_meta, prv_buf_handle, IX_NULL, offset, prv_buf_size, \
					  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_lw0_sig, SIG_NONE, 0, 1)
	    dl_meta_write($ex_prev_meta, prv_buf_handle, IX_NULL, offset, prv_buf_size, \
					  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_sig, SIG_NONE, 1, 5)

	    alu[sig_mask, sig_mask, OR, 1, <<(&meta1_sig)]
    	alu[sig_mask, sig_mask, OR, 1, <<(&meta1_lw0_sig)]

    #else /* DL_SPLIT_META_DATA */

    	dl_meta_write($ex_prev_meta, prv_buf_handle, IX_NULL, offset, prv_buf_size, \
					  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_sig, SIG_NONE, 0, 6)

	    alu[sig_mask, sig_mask, OR, 1, <<(&meta1_sig)]

    #endif /* DL_SPLIT_META_DATA */

#endif  // METADATA_2LW

	; Flushing of RXC2 is not required on EOP.

	;	The way packets are sent to next block in the pipeline is thro'
	;	dl_sink. (see the outer loop in aal5_rx.uc). 
	;	As part of optimisation, 4 LWs ( 1, 2, 3, 5) from meta data are sent 
	;	in the scratch ring. dl_sink takes these meta data from the 
	;	previous dl_meta_set_* calls. These meta data should belong to SOP buffer.

	; In our case the prev dl_set_* calls belong to the EOP buffer. 
	; So we	need to do a few dl_meta_set_* calls for SOP buffer here.


	; Set meta data for the next block.
	dl_meta_set_offset(offset)
	; Here SOP buffer size would be equal to pkt_len
	; Here, we can also get sop buffer size from RXC:sop_buf_size, but would
	; require an extra instruction here to extract its value. 

	alu[tmp, --, B, pkt_length]					; to avoid reg conflict
	dl_meta_set_buffer_size(tmp)				; set sop_buffer size 

	; Drop the last buffer allocated for the extra_cell.
	dl_buf_drop(curr_handle)

	; Ctx0 waits [CS3_inter_me_sig, next_thread_sig, meta1_sig]
	; other ctx waits  [next_thread_sig, meta1_sig].
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
	ctx_arb[--] , defer[2], br[E_END#]
	alu[dl_eop_buf_handle, --, B, DL_BUF_NULL] 	 ;set dl_eop_handle = NULL
	alu[dl_buf_handle, --, B, prv_buf_handle]  ; set dl_buf_handle = prev handle(sop buf)

	xbuf_free($ex_prev_meta)

;---------------------------------------------------------------------------------
E_NEW_BUF_PRV_BUF_NOT_SOP#:
;---------------------------------------------------------------------------------

	; Here Previous buffer is not SOP buffer.

	
	xbuf_alloc($ex_prev_meta2, 6, write)

	; Write META data of previous to previous buffer to SRAM.

	; Set the EOP bit in the prev buff handle (which was cleared before)

	; alu_shf requires unrestricted operand and LM[index 8 to 15] is restricted 
	; operand. So we have to move to register first
	; And we cannot have two LM as operands
	
	alu[prv_buf_handle, --,b, lm_prv_buf_handle]
	alu[prv_buf_handle, prv_buf_handle, OR, 0x1, <<31]

	; Write META data of previous to previous buffer to SRAM.

	dl_meta_write($ex_prev_meta2, lm_drp_ex_hdr_prv_prv_buf_addr, prv_buf_handle, 0, \
				  0, 0, FREE_LIST_ID, 0, 0, 0, \
				  0, 0, 0, 0, 0, \
				  0, meta1_sig, SIG_NONE, 0, 1)

	alu[sig_mask, sig_mask, OR, 1, <<(&meta1_sig)]

	; Write META data of previous buffer to SRAM.

	xbuf_alloc($ex_eop_meta2, 6, write) 					

#ifdef OPTIONAL_L2PORT
	alu[l2port, lm_optional_info, AND, mask_lower_half]		;extract L2port
#endif

#ifdef OPTIONAL_SUID
	alu[suid, --, B, lm_optional_info]		;extract SUID
#endif


#ifdef METADATA_2LW

	dl_meta_write($ex_eop_meta2, prv_buf_handle, IX_NULL, NON_SOP_BUFF_OFFSET, prv_buf_size, \
				  0, 0, 0, 0, 0, \
				  0, 0, 0, 0, 0, \
				  0, meta2_sig, SIG_NONE, 0, 2)

#else

	; prepare clp bit to put in meta data
	alu[clp, --, B, clp, <<CLP_BIT_OFFSET_IN_DL_META4]

    #ifdef DL_SPLIT_META_DATA
    
    	dl_meta_write($ex_eop_meta2, prv_buf_handle, IX_NULL, NON_SOP_BUFF_OFFSET, prv_buf_size, \
				  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
				  0, clp, 0, SUID_NUMBER, 0, \
				  0, meta2_lw0_sig, SIG_NONE, 0, 1)
	    dl_meta_write($ex_eop_meta2, prv_buf_handle, IX_NULL, NON_SOP_BUFF_OFFSET, prv_buf_size, \
				  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
				  0, clp, 0, SUID_NUMBER, 0, \
				  0, meta2_sig, SIG_NONE, 1, 5)

    	alu[sig_mask, sig_mask, OR, 1, <<(&meta2_sig)]
	    alu[sig_mask, sig_mask, OR, 1, <<(&meta2_lw0_sig)]
    
    #else /* DL_SPLIT_META_DATA */

    	dl_meta_write($ex_eop_meta2, prv_buf_handle, IX_NULL, NON_SOP_BUFF_OFFSET, prv_buf_size, \
				  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
				  0, clp, 0, SUID_NUMBER, 0, \
				  0, meta2_sig, SIG_NONE, 0, 6)

	    alu[sig_mask, sig_mask, OR, 1, <<(&meta2_sig)]
    
    #endif /* DL_SPLIT_META_DATA */

#endif  // METADATA_2LW

	; Drop the last buffer allocated for the extra_cell. 
	dl_buf_drop(curr_handle)


	; Flushing of RXC2 is not required on EOP.

	;	The way packets are sent to next block in the pipeline is thro'
	;	dl_sink. (see the outer loop in aal5_rx.uc). 
	;	As part of optimisation, 4 LWs ( 1, 2, 3, 5) from meta data are sent 
	;	in the scratch ring. dl_sink takes these meta data from the 
	;	previous dl_meta_set_* calls. These meta data should belong to SOP buffer.

	; In our case the prev dl_set_* calls belong to the EOP buffer. 
	; So we	need to do a few dl_meta_set_* calls for SOP buffer here.

	alu[offset, mask_lower_half, AND, lm_sop_buf_sz_offset]
	dl_meta_set_offset(offset)

	alu[sop_buf_size, --, B, lm_sop_buf_sz_offset, >>16]
	dl_meta_set_buffer_size(sop_buf_size)	   ; set sop buffer size

	; Ctx0 waits [CS3_inter_me_sig, next_thread_sig, meta1_sig, meta2_sig]
	; other ctx waits  [next_thread_sig, meta1_sig, meta2_sig].
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
	ctx_arb[--] , defer[2], br[E_END#]
	; Set dl_eop_handle = prv_buf_handle (last buf).
	alu[dl_eop_buf_handle, --, B, prv_buf_handle]
	alu[dl_buf_handle, --, B, lm_sop_buf_handle] ; set dl_buf_handle = sop_handle
	
	xbuf_free($ex_prev_meta2)
	xbuf_free($ex_eop_meta2) 						


;---------------------------------------------------------------------------------
E_CRC_ERROR#:
;---------------------------------------------------------------------------------
	

	; Here there was crc_error 

	; Update stat counters.
  	_aal5_rx_incr_vc_counter[vc_id, AAL5_RX_AAL5_PDU_CRC_BAD]
	_aal5_rx_incr_port_counter[input_port, AAL5_RX_PKT_DROP_COUNT]

	#ifdef TWO_ME_RX_FP
	
		; In sync with RXC2 read from SRAM
		ctx_arb[next_thread_sig]

		; Signal the next thread
		signal_ctx(next_thread_sig_val)
		alu[sig_mask, --, B, 1, <<&next_thread_sig]

		; Signal next ME with CS2_inter_me_sig.
		br!=ctx[7, E_THD_NOT_7_2#]
	   	alu[--, --, B, lm_cs2_inter_me_sig_val]
		cap[fast_wr, ALU, INTERTHREAD_SIG]	

E_THD_NOT_7_2#:

		br!=ctx[0, E_NOT_THD_0_3#]

		; Add CS3_inter_me_sig to sig_mask
		alu[sig_mask, sig_mask, OR, 1, <<ENTER_CS3_INTER_ME_SIGNAL]


E_NOT_THD_0_3#:

	#endif


	alu[curr_handle, --, b, lm_curr_buf_handle]

	; Drop the packet. We will be dropping packets in dl_sink[], as we want
	; to ensure that DRAM write of previous thread is complete.

	; Check if curr_handle has SOP bit set, if yes, then there is only one buffer
	; to drop.

	br_bclr[curr_handle, 30, E_DROP_BUF_CHAIN#]  ; [30] is SOP bit
	
	; Set variables to dop single buffer.
	alu[dl_buf_handle, --, B, curr_handle]	; dl_buf_handle = curr_handle = sop_handle 
	alu[dl_eop_buf_handle, --, B, DL_BUF_NULL]

	br[E_WAIT#]
	
E_DROP_BUF_CHAIN#:

	; Set variables to drop buffer chain.
	alu[dl_buf_handle, --, b, lm_sop_buf_handle]
	alu[dl_eop_buf_handle, --, B, curr_handle]	 
	

E_WAIT#:
	
	; Ctx0 waits [CS3_inter_me_sig, next_thread_sig]
	; other ctx waits  [next_thread_sig].
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
	ctx_arb[--] , defer[1], br[E_END#]
	alu[dl_next_block, --, B, BID_DROP]		 ; set dl_next_block = BID_DROP

;---------------------------------------------------------------------------------	
E_INV_FR_SZ_ERROR#:
;---------------------------------------------------------------------------------

	; We came here because of one of the two cases.
	; (1) (curr_pkt_size - pkt_length) > 7
	; (2) AAL5 frame size > MAX_FR_SZ (64KB)

	; Update stat counters.
	_aal5_rx_incr_vc_counter[vc_id, AAL5_RX_AAL5_PDU_LEN_BAD]
	_aal5_rx_incr_port_counter[input_port, AAL5_RX_PKT_DROP_COUNT]


	; Drop the packet. We will be dropping packets in dl_sink[], as we want
	; to ensure that DRAM write of previous thread is complete.
	alu[curr_handle, --, b, lm_curr_buf_handle]
	
	; Check if curr_handle has SOP bit set, if yes, then there is only one buffer
	; to drop.

	br_bclr[curr_handle, 30, E_DROP_BUF_CHAIN2#]  ; [30] is SOP bit
	
	; Set variables to drop single buffer.
	alu[dl_buf_handle, --, B, curr_handle] ; dl_buf_handle = curr_handle = sop_handle 
	alu[dl_eop_buf_handle, --, B, DL_BUF_NULL]

	br[E_WAIT2#]
	
E_DROP_BUF_CHAIN2#:

	; Set variables to drop buffer chain.
	alu[dl_buf_handle, --, b, lm_sop_buf_handle]
	alu[dl_eop_buf_handle, --, B, curr_handle]	 
	

E_WAIT2#:

	; Ctx0 waits [CS3_inter_me_sig, next_thread_sig]
	; other ctx waits  [next_thread_sig].
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
	ctx_arb[--] , defer[1], br[E_END#]
	alu[dl_next_block, --, B, BID_DROP]		 ; set dl_next_block = BID_DROP

;--------------------------------------------------------------------------------
E_DROP#:
;-------------------------------------------------------------------------------

	; Nothing to do.

	
	; Clear RXC2:drop_bit as it is the end of the packet.
	_aal5_rx_clear_rxc_drop_bit(lm_drp_ex_hdr_prv_prv_buf_addr)

	; Ctx0 waits [CS3_inter_me_sig, next_thread_sig]
	; other ctx waits  [next_thread_sig].
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
	ctx_arb[--] , defer[1], br[E_END#]
	alu[dl_next_block, --, B, BID_NULL]		 ; set dl_next_block = NULL


E_END#:

	.io_completed bd_prefetch_sig, data_write_sig, meta1_sig, meta2_sig, rxc_write_sig  		

    #ifdef DL_SPLIT_META_DATA

	    .io_completed meta1_lw0_sig, meta2_lw0_sig

    #endif
	xbuf_free($$cell_out)


	#ifdef TWO_ME_RX_FP
	
		xbuf_free($rxc_out)

	#endif


.end
#endm   //   _aal5_rx_eop_only_proc()

///////////////////////////////////////////////////////////////////////////////
//
// _aal5_rx_incr_cell_count
//
//	 	Description: 
//
//			It calculates how many TX cells are required to transmit a given number
//			of bytes and increment cell count accordingly. It does this 
//			incrementally, when this macro is called, by keeping track of bytes 
//			remainder.
//
//	 	Outputs:
//
//			cell_count(tmp_c):	(old) cell count + cells computed from byte_count
//								(one cell_cnt less than actual)
//		Inputs:
//
//			cell_count		:	existing cell count (taken from lm_index_var)
//			
//			byte_remainder	:	existing byte remainder (lm_index_var)
//
//			byte_count		:	Number of bytes to use in this cell count
//								calculation.
//		Size: 
//
//			8 cycles at a minimum. ( if no incr in cell_cnt)
//			10 cycles (when there is an incr in cell_cnt)
//
///////////////////////////////////////////////////////////////////////////////

#macro _aal5_rx_incr_cell_count(tmp_c, lm_index_var, byte_count)
.begin

	.reg tmp_r

	alu[tmp_c, lm_index_var,AND, 0xFF]
	alu[tmp_r, 0xFF, AND, lm_index_var, >>8]

	alu[tmp_r, tmp_r, +, byte_count]

	; Check whether remainder is GT CELL_SIZE
	alu[--, tmp_r, -, TX_CELL_SZ]		// minus cell size 
	ble[CELL_CNT_DONE#], defer[1]	
	// Having defer slot here, as for big cell_size (say 120B)
	// 2 out of 3 times we will be just incrementing remainder, 
	// and we are done(aal5 cell cell size being 48B). 
	// Every third time we will incr cell_cnt
	ld_field[lm_index_var, 0010, tmp_r, <<8]  

	
	// If we assume here that CELL_SIZE is it atleast 48B,
	// then there is at the most only one cell_cnt 
	// increment in one round. So we dont need to loop till 
	// remainder is GTE CELL_SIZE.
	
	; Here remainder is GT CELL_SIZE	
	alu[tmp_c, tmp_c, +, 0x1]			// incr cell count
	alu[tmp_r, tmp_r, -, TX_CELL_SZ]		// new remainder -= cell size

	ld_field[lm_index_var, 0001, tmp_c]	 
	ld_field[lm_index_var, 0010, tmp_r, <<8]  

	; computed cell_cnt is one less than actual, as required by hardware.
	; So on EOP/buffer overflow we dont have to worry about remainder bytes.

CELL_CNT_DONE#:

.end
#endm   //   _aal5_rx_incr_cell_count()


///////////////////////////////////////////////////////////////////////////////
// _aal5_rx_get_rand_buf_offset
//
// Description:
//		This is used for dram bank scheduling.
//		This gives the offset for the SOP buffer, to write the first cell of the 
//		packet.
//		First time the offset would be 128B, then 256B, then 384B, then 512B.
//		It would again start with 128B, 256B .....
//
// Input:
//
// Output: 
// 	 buf_offset: offset for the SOP buffer
//
// Size: 
//		3 cycles
///////////////////////////////////////////////////////////////////////////////
#macro	_aal5_rx_get_rand_buf_offset(buf_offset)
.begin

	alu[@rand_buf_offset, @rand_buf_offset, AND, lm_rand_offset_mask]
	alu[@rand_buf_offset, @rand_buf_offset, +, DDR_BANK_SIZE]
	alu[buf_offset, --, B, @rand_buf_offset]


.end
#endm   //   _aal5_rx_get_rand_buf_offset()

////////////////////////////////////////////////////////////////////////////////
// _aal5_rx_incr_port_counter
//
// Description:
//		It increments following port counters according to the offset
//		given as one of the parameters
//
//		AAL5_RX_PKT_COUNT 			0			
//		AAL5_RX_PKT_DROP_COUNT		4		
//		AAL5_RX_HW_ERROR			8		
//		AAL5_RX_VC_ERROR			12		
//
// Input: 
//	 	Counter Offset:  offset for the particular counter
//
//	 	port number : The port number [0-15]
//		
// Output: 
//		N/A
//
// Size: 
//		3 cycles
///////////////////////////////////////////////////////////////////////////////

#macro _aal5_rx_incr_port_counter[port_n, COUNTER_OFFSET]
.begin

#ifdef	AAL5_RX_COUNTERS

	.reg	offset

	alu[offset, --, B, port_n, <<AAL5_RX_PORT_STAT_ENTRY_SIZE_LOG2]	
	alu[offset, offset, +, COUNTER_OFFSET]				

	sram[incr, --, port_stat_base, offset]				; incr counter for this port

#endif

.end
#endm

////////////////////////////////////////////////////////////////////////////////
// _aal5_rx_incr_vc_counter
//
// Description:
//		It increments following VC counters according to the offset
//		given as one of the parameters
//
//		AAL5_RX_CELL_CNT				40
//		AAL5_RX_MGT_CELL_COUNT			44
//		AAL5_RX_AAL5_PDU_COUNT			48
//		AAL5_RX_AAL5_PDU_CRC_BAD		52
//		AAL5_RX_AAL5_PDU_LEN_BAD		56				
//
// Input: 
//	 	Counter Offset:  offset for the particular counter
//
//	 	VC ID : VC id [0 thru 64K-1]
//		
// Output: 
//		N/A
//
// Size: 
//		3 cycles
///////////////////////////////////////////////////////////////////////////////


#macro _aal5_rx_incr_vc_counter[vc_id, COUNTER_OFFSET]
.begin

#ifdef	AAL5_RX_COUNTERS

	.reg	offset

	alu[offset, --, B, vc_id, <<AAL5_RX_VC_INFO_ENTRY_SIZE_LOG2]	
	alu[offset, offset, +, COUNTER_OFFSET]						; 

	sram[incr, --, vc_info_sram_base, offset]		; incr counter for this VC

#endif

.end
#endm

///////////////////////////////////////////////////////////////////////////////
//_aal5_rx_clear_rxc_exp_sop		
//		
//Description:
//
//		It clears the expect SOP bit of RXC
//
//
//Input:  corresponding local memory LW.
//
//Output: N/A
//
//Size:   1 cycle
///////////////////////////////////////////////////////////////////////////////

#macro _aal5_rx_clear_rxc_exp_sop(lm_index_var)
.begin

	alu[lm_index_var, lm_index_var, AND~, 0x1, <<31]

.end
#endm   //   _aal5_rx_clear_rxc_exp_sop()

///////////////////////////////////////////////////////////////////////////////
//_aal5_rx_set_rxc_drop_bit		
//		
//Description:
//
//		It sets the DROP bit of RXC
//
//
//Input:  corresponding local memory LW.
//
//Output: N/A
//
//Size:   1 cycle
///////////////////////////////////////////////////////////////////////////////

#macro _aal5_rx_set_rxc_drop_bit(lm_index_var)
.begin

	 alu_shf[lm_index_var, lm_index_var, OR, 1, <<31]

.end
#endm   //   _aal5_rx_set_rxc_drop_bit()

///////////////////////////////////////////////////////////////////////////////
//_aal5_rx_clear_rxc_drop_bit		
//		
//Description:
//
//		It clears the DROP bit of RXC
//
//
//Input:  corresponding local memory LW.
//
//Output: N/A
//
//Size:   1 cycle
///////////////////////////////////////////////////////////////////////////////

#macro _aal5_rx_clear_rxc_drop_bit(lm_index_var)
.begin

	alu[lm_index_var, lm_index_var, AND~, 0x1, <<31]

.end
#endm   //   _aal5_rx_clear_rxc_drop_bit()

///////////////////////////////////////////////////////////////////////////////
//_aal5_rx_set_rxc_curr_pkt_size		
//		
//Description:
//
//		It sets the current pkt size in RXC
//
//
//Input:  corresponding local memory LW.
//
//Output: N/A
//
//Size:   1 cycle
///////////////////////////////////////////////////////////////////////////////

#macro _aal5_rx_set_rxc_curr_pkt_size(lm_index_var, size)
.begin

	ld_field[lm_index_var, 0011, size]	

.end
#endm   //   _aal5_rx_set_rxc_curr_pkt_size()

///////////////////////////////////////////////////////////////////////////////
//_aal5_rx_set_rxc_buf_offset		
//		
//Description:
//
//		It sets the buffer offset in RXC
//
//
//Input:  corresponding local memory LW.
//
//Output: N/A
//
//Size:   1 cycle
///////////////////////////////////////////////////////////////////////////////

#macro _aal5_rx_set_rxc_buf_offset(lm_index_var, buf_off_set)
.begin

	ld_field[lm_index_var, 1100, buf_off_set, <<16]

.end
#endm   //   _aal5_rx_set_rxc_buf_offset()

///////////////////////////////////////////////////////////////////////////////
//_aal5_rx_set_rxc_offset		
//		
//Description:
//
//		It sets the offset in RXC
//
//
//Input:  corresponding local memory LW.
//
//Output: N/A
//
//Size:   1 cycle
///////////////////////////////////////////////////////////////////////////////

#macro _aal5_rx_set_rxc_offset(lm_index_var, off_set)
.begin

	ld_field[lm_index_var, 0011, off_set]

.end
#endm   //   _aal5_rx_set_rxc_buf_offset()

///////////////////////////////////////////////////////////////////////////////
//_aal5_rx_get_rxc_header_type		
//		
//Description:
//
//		It gets the header type from RXC
//
//
//Input:  corresponding local memory LW.
//
//Output: N/A
//
//Size:   1 cycle
///////////////////////////////////////////////////////////////////////////////

#macro _aal5_rx_get_rxc_header_type(h_type, lm_index_var)
.begin

	alu[h_type, 0xF, AND, lm_index_var, >>24]

.end
#endm   //   _aal5_rx_get_rxc_header_type()


///////////////////////////////////////////////////////////////////////////////
//_aal5_rx_err_phase1		
//		
//Description:
//
//		It does necessary signalling and wait (error processing)
//		for phase 3 for ONE_ME_RX_FP and TWO_ME_RX_FP.
//
//
//Input:  N/A
//
//Output: N/A
//
//Size:   N/A
//
///////////////////////////////////////////////////////////////////////////////


#macro _aal5_rx_err_phase1()
.begin


	#ifdef ONE_ME_RX_FP

		; To be in sync with ctx_arb[rbuf_read, next_thread] 
		ctx_arb[next_thread_sig]

	#else  // TWO_ME_RX_FP

		br!=ctx[0, THD_NOT_0#]
	
		; To be in sync with ctx_arb[rbuf_read, next_thread, CS1_inter_me_sig] 
		ctx_arb[next_thread_sig, enter_cs1_inter_me_sig]

		br[END#], defer[1]
			cam_clear

THD_NOT_0#:

		; To be in sync with ctx_arb[rbuf_read, next_thread] 
		ctx_arb[next_thread_sig]

END#:

	#endif   // TWO_ME_RX_FP


.end
#endm   //   _aal5_rx_err_phase1()


///////////////////////////////////////////////////////////////////////////////
//_aal5_rx_err_phase2		
//		
//Description:
//
//		It does necessary signalling and wait (error processing)
//		for phase 2 for ONE_ME_RX_FP and TWO_ME_RX_FP
//
//Input:  N/A
//
//Output: N/A
//
//Size:   N/A
//
///////////////////////////////////////////////////////////////////////////////


#macro _aal5_rx_err_phase2()
.begin

	// Same for both ONE_ME_RX_FP and TWO_ME_RX_FP

	; signal next thread
	signal_ctx(next_thread_sig_val)

	; Free up the rbuf element (2)
	_aal5_rx_free_rbuf_element(element_num)

	; To be in sync with ctx_arb[rxc_rd, nxt_thread]
	ctx_arb[next_thread_sig]


.end
#endm   //   _aal5_rx_err_phase2()

///////////////////////////////////////////////////////////////////////////////
//_aal5_rx_err_phase2_no_rbuf_free		
//		
//Description:
//
//		It does necessary signalling and wait (error processing)
//		for phase 2 for ONE_ME_RX_FP and TWO_ME_RX_FP
//
//		Note: We dont have RBUF element allocated for RSW_idle condition,
//			  so we should not free RBUF element, else it turns out into
//			  pkt loss later in the pipeline.		
//
//Input:  N/A
//
//Output: N/A
//
//Size:   N/A
//
///////////////////////////////////////////////////////////////////////////////


#macro _aal5_rx_err_phase2_no_rbuf_free()
.begin

	// Same for both ONE_ME_RX_FP and TWO_ME_RX_FP

	; signal next thread
	signal_ctx(next_thread_sig_val)

	; To be in sync with ctx_arb[rxc_rd, nxt_thread]
	ctx_arb[next_thread_sig]


.end
#endm   //   _aal5_rx_err_phase2_no_rbuf_free()

///////////////////////////////////////////////////////////////////////////////
//_aal5_rx_err_phase3		
//		
//Description:
//
//		It does necessary signalling and wait (error processing)
//		for phase 3 for ONE_ME_RX_FP.
//
//		It does necessary signalling and wait (error processing)
//		for phase 3, 4, 5 for TWO_ME_RX_FP.
//
//		This is applicable to RSW errors and Hash miss error (dont have RXC in LM).
//
//
//Input:  
//		  N/A
//
//Output: 
//		  N/A
//
//Size:   N/A
//
///////////////////////////////////////////////////////////////////////////////

#macro _aal5_rx_err_phase3()
.begin

	#ifdef ONE_ME_RX_FP

		; To be in sync with ctx_arb[meta_wr, data_wr, bd_prefetch, nxt_thread]
		ctx_arb[next_thread_sig]


	#else   // TWO_ME_RX_FP

		; At this time, we dont have RXC in LM. No need to write back RXC1 to SRAM.
		; Signal next ME with enter_cs1_inter_me_sig and enter_cs2_inter_me_sig.

		; We dont care about RXC:exp_SOP bit here, as finally CRC_err or 
		; invalid_fr_sz_err(in case of two AAL5 frames merge because of loss of 
		; EOP cell) condtion would be hit.
		br!=ctx[7, NOT_THD_7#]

	  	alu[--, --, B, lm_cs1_inter_me_sig_val]
		cap[fast_wr, ALU, INTERTHREAD_SIG]	

	  	alu[--, --, B, lm_cs2_inter_me_sig_val]
		cap[fast_wr, ALU, INTERTHREAD_SIG]	

		cam_clear

NOT_THD_7#:

		br!=ctx[0, NOT_THD_0#]
	
		; To be in sync with entering CS2
		ctx_arb[enter_cs2_inter_me_sig, next_thread_sig]

		br[PHASE4#]

NOT_THD_0#:
	
		ctx_arb[next_thread_sig]

PHASE4#:

		; signal next thread 
		signal_ctx(next_thread_sig_val)

		; To be in sync with getting RXC2 from SRAM
		ctx_arb[next_thread_sig]


PHASE5#:

		; signal next thread 
		signal_ctx(next_thread_sig_val)

		; We dont have RXC2 to write back to SRAM. We have already signalled
		; next ME with enter_cs2_inter_me_sig
	
		br!=ctx[0, NOT_THD_0_2#]
	
		; To be in sync with entering CS3
		ctx_arb[enter_cs3_inter_me_sig, next_thread_sig]

		br[PHASE6#]

NOT_THD_0_2#:
	
		ctx_arb[next_thread_sig]

PHASE6#:

	#endif  // TWO_ME_RX_FP

	
.end
#endm   //   _aal5_rx_err_phase3()

///////////////////////////////////////////////////////////////////////////////
//_aal5_rx_hash_miss_inputports16		
//		
//Description:
//
//		This macro is used when the number of inputput ports supported are 16.
//
//		It does necessary signalling and wait (error processing)
//		for phase 1 for hashmiss case (for both ONE_ME_RX_FP and TWO_ME_RX_FP)
//
//Input:  N/A
//
//Output: N/A
//
//Size:   N/A
//
///////////////////////////////////////////////////////////////////////////////

#macro  _aal5_rx_hash_miss_inputports16()
.begin

	; Hash Miss Err Phase1.

	#ifdef ONE_ME_RX_FP

		ctx_arb[rbuf_rd_sig, next_thread_sig]

	#else // TWO_ME_RX_FP

	br!=ctx[0, NOT_THD_0_2#]
	
	; ctx0 waits for [enter_cs1_inter_me_sig, rbuf_rd_sig, next_thread_sig].
	ctx_arb[rbuf_rd_sig, enter_cs1_inter_me_sig, next_thread_sig]
	br[END_PHASE1#], defer[1]
	; clear CAM for critical section1.
	cam_clear


NOT_THD_0_2#:

	ctx_arb[rbuf_rd_sig, next_thread_sig]

END_PHASE1#:


	#endif // TWO_ME_RX_FP
.end

#endm  // 	_aal5_rx_hash_miss_inputports16()

///////////////////////////////////////////////////////////////////////////////
//_aal5_rx_hash_miss		
//		
//Description:
//
//		This macro is used when the number of inputput ports supported are more 
//		than 16.
//
//		It does necessary signalling and wait (error processing)
//		for phase 1 for hashmiss case (for both ONE_ME_RX_FP and TWO_ME_RX_FP)
//
//Input:  N/A
//
//Output: N/A
//
//Size:   N/A
//
///////////////////////////////////////////////////////////////////////////////

#macro 	_aal5_rx_hash_miss()
.begin

	#ifdef ONE_ME_RX_FP

		ctx_arb[next_thread_sig]

	#else // TWO_ME_RX_FP

	br!=ctx[0, NOT_THD_0_2#]
	
	; ctx0 waits for [enter_cs1_inter_me_sig, next_thread_sig].
	ctx_arb[enter_cs1_inter_me_sig, next_thread_sig]
	br[END_PHASE1#], defer[1]
	; clear CAM for critical section1.
	cam_clear


NOT_THD_0_2#:

	ctx_arb[next_thread_sig]

END_PHASE1#:


	#endif // TWO_ME_RX_FP
.end
#endm  // _aal5_rx_hash_miss()


///////////////////////////////////////////////////////////////////////////////
//_aal5_rx_mgt_cell_proc		
//		
//Description:
//
//		It does necessary mgt cell processing.
//
//
//Input:  
//		sig_mask :   mask to be set with various I/O signals.
//
//		rxc_adr :    absolute rxc addr used to write back RXC to SRAM.
//
//		$$cell_in : dram xfer regs used to write to DRAM.
//
//		$prefetch_handle : the already prefetched buffer handle.
//
//Output: 
//		  N/A
//
//Size:   N/A
//
//Note: This macro is called from _aal5_rx_oam_pm_processing[] of aal5_rx_oam.uc
//		only if ATM_OAM_ENABLED is defined.
// 
///////////////////////////////////////////////////////////////////////////////

#macro _aal5_rx_mgt_cell_proc(sig_mask, rxc_addr, $$cell_out, $prefetch_buf_handle)
.begin

	.reg dram_buf_addr
	.reg buf_offset


	#ifdef TWO_ME_RX_FP

		// Allocate xfer reg for writing RXC1 back to SRAM.
		xbuf_alloc($rxc_out, 2, write)

	#endif

	; Check if prefetch buffer handle is zero.
	alu[--, --, B, $prefetch_buf_handle, <<8]	; [23:0] contains buf handle.
	beq[MGT_NULL_BUF_PROC#]						; if null, error condition.


	; Is this required for mgt cell??
	; Get random offset used to do DRAM bank scheduling. 
	_aal5_rx_get_rand_buf_offset(buf_offset)


	#ifdef TWO_ME_RX_FP

		; In TWO_ME_RX_FP, we have to update two fields - crc_residue, exp_SOP
		; else the NEXT ME might get the stale state from SRAM, specifically
		; when there was a long packet ahead of it spread over TWO MEs.

		; Write back RXC1 to SRAM (using defer slots for RXC1 updates)

MGT_REF_CNT_CHK#:

		alu[--, *l$index0[RXC_LM_REF_CNT_OFFSET], -, 0]
		bne[MGT_REF_CNT_NONZERO#], defer[2]
		; Set RXC:exp_SOP bit.
		alu_shf[lm_sop_alloc_valid_aal_flist, lm_sop_alloc_valid_aal_flist, OR, 1, <<31]
		; Reset RXC:crc.
		alu[lm_crc_residue, --, b, lm_crc_init]


		; Here the ref_cnt is zero.
		_aal5_rx_write_rxc_to_sram_cs1($rxc_out, rxc_addr, rxc_write_sig) 

		br!=ctx[7, MGT_REF_CNT_NONZERO#], defer[1]
		alu[sig_mask, sig_mask, OR,  1, <<(&rxc_write_sig)]

		; Signal next ME with CS1_inter_me_sig.
	   	alu[--, --, B, lm_cs1_inter_me_sig_val]
		cap[fast_wr, ALU, INTERTHREAD_SIG]	

		; Not getting RXC2 from SRAM, as we dont use it here in EOP_SOP case
		; So no writing back RXC2 to SRAM. So we can signal next ME here.

		; Signal next ME with CS2_inter_me_sig.
	   	alu[--, --, B, lm_cs2_inter_me_sig_val]
		cap[fast_wr, ALU, INTERTHREAD_SIG]	

		; Clear CAM for critical section2.
		cam_clear

MGT_REF_CNT_NONZERO#:

		br!=ctx[0, MGT_THD_NOT_0#]

		alu[sig_mask, sig_mask, OR, 1, <<ENTER_CS2_INTER_ME_SIGNAL]

MGT_THD_NOT_0#:

		; Ctx0 waits [CS2_inter_me_sig, next_thread_sig]
		; other ctx waits  [next_thread_sig, rxc_write_sig (if any)].
		local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
		ctx_arb[--] 

		; Signal the next thread.
		signal_ctx(next_thread_sig_val)

		; Not getting/using RXC2 from sram, so just waiting for next_thread_sig
		// _aal5_rx_get_rxc_cs2[rxc_addr]
	
		; So no need to decr the ref cnt
		// alu[*l$index0[RXC_LM_REF_CNT_OFFSET], *l$index0[RXC_LM_REF_CNT_OFFSET], -, 1]

		; To be in sync with other threads as we are not getting RXC2 from sram.
		ctx_arb[next_thread_sig]

		; Signal the next thread.
		signal_ctx(next_thread_sig_val)

		; Adding next_thread_sig to sig_mask
		alu[sig_mask, --, b, 1, <<&next_thread_sig]

		; Ctx0 waits for CS3_inter_me_sig
		br!=ctx[0, MGT_THD_NOT_0_2#]

		alu[sig_mask, sig_mask, OR, 1, <<ENTER_CS3_INTER_ME_SIGNAL]

MGT_THD_NOT_0_2#:	


	#endif  // TWO_ME_RX_FP

	; Write META LW0, LW1 of the current this buffer to SRAM.
	; Setting buffer_next = 0
	; Setting buffer_size = 48
	; Setting offset = buf_offset

	xbuf_alloc($mgt_meta,6, write)					// no code generated.

#ifdef OPTIONAL_L2PORT
	alu[l2port, lm_optional_info, AND, mask_lower_half]		;extract L2port
#endif

#ifdef OPTIONAL_SUID
	alu[suid, --, B, lm_optional_info]		;extract SUID
#endif


#ifdef METADATA_2LW

	dl_meta_write($mgt_meta, $prefetch_buf_handle, IX_NULL, buf_offset, NONEOP_CELL_DATA_SZ,\
				  0, 0, 0, 0, 0, \
				  0, 0, 0, 0, 0, \
				  0, meta1_sig, SIG_NONE, 0, 2)  

#else

	; prepare clp bit to put in meta data
	alu[clp, --, B, clp, <<CLP_BIT_OFFSET_IN_DL_META4]

    #ifdef DL_SPLIT_META_DATA

	    dl_meta_write($mgt_meta, $prefetch_buf_handle, IX_NULL, buf_offset, NONEOP_CELL_DATA_SZ,\
					  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_lw0_sig, SIG_NONE, 0, 1)  
	    dl_meta_write($mgt_meta, $prefetch_buf_handle, IX_NULL, buf_offset, NONEOP_CELL_DATA_SZ,\
					  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_sig, SIG_NONE, 1, 5)  

	    alu[sig_mask, sig_mask, OR, 1, <<(&meta1_sig)]
    	alu[sig_mask, sig_mask, OR, 1, <<(&meta1_lw0_sig)]
    
    #else /* DL_SPLIT_META_DATA */

    	dl_meta_write($mgt_meta, $prefetch_buf_handle, IX_NULL, buf_offset, NONEOP_CELL_DATA_SZ,\
					  zero, FREE_LIST_ID, zero, 0, L2PORT_NUMBER, \
					  0, clp, 0, SUID_NUMBER, 0, \
					  0, meta1_sig, SIG_NONE, 0, 6)  

    	alu[sig_mask, sig_mask, OR, 1, <<(&meta1_sig)]
    
    #endif /* DL_SPLIT_META_DATA */

#endif // METADATA_2LW

	alu[dl_buf_handle, --, B, $prefetch_buf_handle]	   	; set dl_buf_handle

	; Write  the received cell to DRAM  (8B * 6 = 48B).
	; Later, modify this to write required data bytes.
	dl_buf_get_data(dram_buf_addr, $prefetch_buf_handle) 			
	dram[write, $$cell_out[0], dram_buf_addr, buf_offset, 6],sig_done[data_write_sig]

	alu[sig_mask, sig_mask, OR, 3, <<(&data_write_sig)]


	; Get a new prefetch buffer so that it can used next time around.
	dl_buf_alloc[$prefetch_buf_handle, BUF_FREE_LIST0, bd_prefetch_sig, SIG_NONE]

	alu[sig_mask, sig_mask, OR, 1, <<(&bd_prefetch_sig)]


	; Ctx0 waits [CS3_inter_me_sig, data_write_sig, bd_prefetch_sig, next_thread_sig]
	; other ctx waits [data_write_sig, bd_prefetch_sig, next_thread_sig].
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
	ctx_arb[--] , defer[2], br[MGT_END#]
	alu[dl_next_block, --, B, IX_EXCEPTION] 	; set next_block to IX_EXCEPTION
	alu[dl_eop_buf_handle, --, B, DL_BUF_NULL]	; set dl_eop_buf_handle to NULL	


;-----------------------------------------------------------------------------		
MGT_NULL_BUF_PROC#:
;-----------------------------------------------------------------------------

	;	We got a null buffer. So nothing to do for current cell. 

	#ifdef TWO_ME_RX_FP

		; In TWO_ME_RX_FP, we have to update two fields - crc_residue, exp_SOP
		; else the NEXT ME might get the stale state from SRAM, specifically
		; when there was a long packet ahead of it spread over TWO MEs.

		; Write back RXC1 to SRAM (using defer slots for RXC1 updates)

MGT_REF_CNT_CHK2#:

		alu[--, *l$index0[RXC_LM_REF_CNT_OFFSET], -, 0]
		bne[MGT_REF_CNT_NONZERO2#], defer[2]
		; Set RXC:exp_SOP bit.
		alu_shf[lm_sop_alloc_valid_aal_flist, lm_sop_alloc_valid_aal_flist, OR, 1, <<31]
		; Reset RXC:crc.
		alu[lm_crc_residue, --, b, lm_crc_init]


		; Here the ref_cnt is zero.
		_aal5_rx_write_rxc_to_sram_cs1($rxc_out, rxc_addr, rxc_write_sig) 

		br!=ctx[7, MGT_REF_CNT_NONZERO2#], defer[1]
		alu[sig_mask, sig_mask, OR, 1, <<(&rxc_write_sig)]


		; Signal next ME with CS1_inter_me_sig.
	   	alu[--, --, B, lm_cs1_inter_me_sig_val]
		cap[fast_wr, ALU, INTERTHREAD_SIG]	

		; Not getting RXC2 from SRAM, as we dont use it here in EOP_SOP case
		; So no writing back RXC2 to SRAM. So we can signal next ME here.

		; Signal next ME with CS2_inter_me_sig.
	   	alu[--, --, B, lm_cs2_inter_me_sig_val]
		cap[fast_wr, ALU, INTERTHREAD_SIG]	

		; Clear CAM for critical section2.
		cam_clear

MGT_REF_CNT_NONZERO2#:

		br!=ctx[0, MGT_THD_NOT_0_3#]

		alu[sig_mask, sig_mask, OR, 1, <<ENTER_CS2_INTER_ME_SIGNAL]

MGT_THD_NOT_0_3#:

		; Ctx0 waits [CS2_inter_me_sig, next_thread_sig]
		; other ctx waits  [next_thread_sig, rxc_write_sig (if any)].
		local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
		ctx_arb[--] 

		; Signal the next thread.
		signal_ctx(next_thread_sig_val)

		; Not getting/using RXC2 from sram, so just waiting for next_thread_sig
		// _aal5_rx_get_rxc_cs2[rxc_addr]
	
		; So no need to decr the ref cnt
		// alu[*l$index0[RXC_LM_REF_CNT_OFFSET], *l$index0[RXC_LM_REF_CNT_OFFSET], -, 1]

		; To be in sync with other threads as we are not getting RXC2 from sram.
		ctx_arb[next_thread_sig]

		; Signal the next thread.
		signal_ctx(next_thread_sig_val)

		; Adding next_thread_sig to sig_mask
		alu[sig_mask, --, b, 1, <<&next_thread_sig]

		; Ctx0 waits for CS3_inter_me_sig
		br!=ctx[0, MGT_THD_NOT_0_4#]

		alu[sig_mask, sig_mask, OR, 1, <<ENTER_CS3_INTER_ME_SIGNAL]

MGT_THD_NOT_0_4#:	


	#endif  // TWO_ME_RX_FP


	; Get a new prefetch buffer so that it can used next time around.
	dl_buf_alloc[$prefetch_buf_handle, BUF_FREE_LIST0, bd_prefetch_sig, SIG_NONE]

	alu[sig_mask, sig_mask, OR, 1, <<(&bd_prefetch_sig)]


	; Ctx0 waits [CS3_inter_me_sig, bd_prefetch_sig, next_thread_sig]
	; other ctx waits  [next_thread_sig, bd_prefetch_sig].
	local_csr_wr[ACTIVE_CTX_WAKEUP_EVENTS, sig_mask]
	ctx_arb[--], defer[1]
	alu[dl_next_block, --, B, BID_NULL]		 	; set next_block NULL


MGT_END#:

	.io_completed bd_prefetch_sig, data_write_sig, meta1_sig
    #ifdef DL_SPLIT_META_DATA
	    
        .io_completed meta1_lw0_sig

    #endif		

	#ifdef TWO_ME_RX_FP

		xbuf_free($rxc_out)
		.io_completed rxc_write_sig

	#endif

	xbuf_free($mgt_meta)



.end
#endm

#ifdef AAL2_INCLUDED

//-----------------------------------------------------------------------------
// _aal5_rx_to_atm_rx()
//
//    Description: 
//		This macro passes the received cell information to aal2_rx microblock which
//		is on the next microengine. Next Neighbour ring is used to pass the msg.
//
//    Parameters:
//      Outputs:
//      In/Outs:
//      Inputs: in_rsw_sxfer0, in_rsw_sxfer1 - xfers containing RSW
//		        in_usr_cntxt - vcid
//		        in_suid - SUID
//				ring_full_label - label to which branch should be performed if NN ring is full
//
//    Size:
//
//    See also:	_atm_rx_from_aal5_rx()
//
#macro _aal5_rx_to_atm_rx(in_rsw_sxfer0, in_rsw_sxfer1, in_usr_cntxt, in_suid, ring_full_label)

/////////////////////////////////////////////////////////
;	ATM_RX uEng to AAL5_RX uEng NN Ring message layout
;			 	3 3 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1
;			 	1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
;
;				+---------------------------------------------------------------+
;   nn lw 0		||   elem_num  |     reserved        |     ccid        |  port  |	//rsw0 + ccid
;				+---------------------------------------------------------------+
;   nn lw 1		|gfcvpi|     vpi       |            vci                | pti  |c|	//rsw1
;				+---------------------------------------------------------------+
;   nn lw 2		|                          vc_table_indx                        |
;				+---------------------------------------------------------------+
;   nn lw 3		|       					vc_flowid					        |
;				+---------------------------------------------------------------+
.begin
.reg _nn_lw_tmp	_tmp _tmp1
	
    // check osf parity
    alu[_tmp1, in_usr_cntxt, and~, 1, <<31]
    .begin
    	.reg  x y

    	alu[y, --, B, $$cell_in[1], >>1]
    	alu[x, y, xor, $$cell_in[1]] 
	
    	alu[y, --, B, x, >>2]
    	alu[x, y, xor, x]
	
    	alu[y, --, B, x, >>4]
    	alu[x, y, xor, x]

        br_bset[x, 24, parity1#]

        alu[_tmp1, _tmp1, or, 1, <<31]
parity1#:
    .end

	
    // check whether next ME's next neighbor ring full, if it is not full, 
	// enqueue one data to next ME's next neighbor ring ( Note that there is a 
	// latency from doing a put until the NN_DFULL status flag is valid. The 
	// NN_FULL status asserts when there are 32 entries still free, such that it 
	// is safe to put 8 items  before testing NN_FULL again without overflowing), 
	// increase both seed and enqueue_count by one
	br_inp_state[NN_FULL, ring_full_label]			; whether next ME's next neighbor ring full?

	alu[*n$index++, --, b, in_rsw_sxfer0]
	alu[*n$index++, --, b, in_rsw_sxfer1]
	alu[*n$index++, --, b, _tmp1]	; user context from VC tbl lookup
	alu[*n$index++, --, b, in_suid]

.end
#endm	//_aal5_rx_to_atm_rx()

#endif //AAL2_INCLUDED

#endif  //  _AAL5_RX_UTIL_UC_